/**
 * @license
 * Copyright (c) 2014, 2018, Oracle and/or its affiliates.
 * The Universal Permissive License (UPL), Version 1.0
 */
"use strict";
(function(){
ES6Promise['polyfill']();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/
/*global define: false,goog: true,self:true*/

/**
 * Defines the oj namespace
 */

/**
 * @private
 */
var _scope = {};

//  - check if the window object is available
// Note that the 'typeof' check  is required
if (typeof window !== 'undefined')
{
  _scope = window;
}
else if (typeof self !== 'undefined')
{
  _scope = self;
}

/**
 * @private
 */
var _oldVal = _scope['oj'];

var oj = _scope['oj'] =
{
  /**
   * @global
   * @member {string} version JET version numberr
   */
  'version': "5.0.0",
  /**
   * @global
   * @member {string} revision JET source code revision number
   */
  'revision': "2018-04-04_15-21-24",
          
  // This function is only meant to be used outside the library, so quoting the name
  // to avoid renaming is appropriate
  'noConflict': function()
  {
    _scope['oj'] = _oldVal;
  }

};
// Copyright (c) 2011, 2013, Oracle and/or its affiliates.
// All rights reserved.

/*jslint browser: true*/

/**
 * @class
 * @name oj.Logger
 * @hideconstructor
 * @since 1.0.0
 *
 * @classdesc
 * <h3>JET Logger</h3>
 *
 * <p>Logger object writes into the native browser console or a custom writer, if a custom writer is set as an option.
 * To use a custom writer, implement the following writer methods: log(), info(), warn(), error()
 *
 * <p>When any of the logging methods is called, it compares the requested log level with the value of a log level option
 * and logs the message if the log level is sufficient.
 *
 * <p>If the logging options are changed at a later point, the Logger will use the modified options for the subsequent log operations.
 *
 * <p>All the logging methods support string formatting, accept variable number of arguments and accept a function as a parameter.
 * When a callback function is specified as a parameter the function will be called if the log level is sufficient.
 *
 * <h3> Usage : </h3>
 * <pre class="prettyprint">
 * <code>
 * //optional calls, see defaults
 * oj.Logger.option("level",  oj.Logger.LEVEL_INFO);
 * oj.Logger.option("writer",  customWriter);  //an object that implements the following methods: log(), info(), warn(), error()
 *
 * // logging a message
 * oj.Logger.info("My log level is %d", oj.Logger.option("level"));  // string formatting
 * oj.Logger.warn("Beware of bugs", "in the above code");            // multiple parameters
 *
 * // using a callback function as a parameter
 * oj.Logger.info(function(){
 *    var foo = "This ";
 *    var bar = "is ";
 *    var zing = "a function";
 *    return foo + bar + zing;
 * });
 * </code></pre>
 *
 * @desc
 * oj.Logger cannot be instantiated
 * @export
 */
oj.Logger = {};
/**
 * Log level none
 * @const
 * @export
 * @type {number}
 * @memberof oj.Logger
 */
oj.Logger.LEVEL_NONE = 0;
/**
 * Log level error
 * @const
 * @type {number}
 * @export
 * @memberof oj.Logger
 */
oj.Logger.LEVEL_ERROR = 1;
/**
 * Log level warning
 * @const
 * @type {number}
 * @export
 * @memberof oj.Logger
 */
oj.Logger.LEVEL_WARN = 2;
/**
 * Log level info
 * @const
 * @type {number}
 * @export
 * @memberof oj.Logger
 */
oj.Logger.LEVEL_INFO = 3;
/**
 * Log level - general message
 * @const
 * @type {number}
 * @export
 * @memberof oj.Logger
 */
oj.Logger.LEVEL_LOG = 4;

/* private constants*/
oj.Logger._METHOD_ERROR = "error";
oj.Logger._METHOD_WARN = "warn";
oj.Logger._METHOD_INFO = "info";
oj.Logger._METHOD_LOG = "log";
oj.Logger._defaultOptions = {'level': oj.Logger.LEVEL_ERROR, 'writer': null};
oj.Logger._options = oj.Logger._defaultOptions;


/*public members*/
/**
 * Writes an error message.
 * @param {...(Object|string|number)} args The method supports a variable number of arguments, string substitutions and accepts a function as a parameter.
 *                                See examples in the overview section above.
 * @return {void}
 * @export
 * @memberof oj.Logger
 * @since 1.0.0
 */
oj.Logger.error = function(args)
{
  oj.Logger._write(oj.Logger.LEVEL_ERROR, oj.Logger._METHOD_ERROR, arguments);
};

/**
 * Writes an informational  message.
 * @param {...(Object|string|number)} args The method supports a variable number of arguments, string substitutions and accepts a function as a parameter.
 *                                See examples in the overview section above.
 * @return {void}
 * @export
 * @memberof oj.Logger
 * @since 1.0.0
 */
oj.Logger.info = function(args)
{
  oj.Logger._write(oj.Logger.LEVEL_INFO, oj.Logger._METHOD_INFO, arguments);
};

/**
 * Writes a warning message.
 * @param {...(Object|string|number)} args The method supports a variable number of arguments, string substitutions and accepts a function as a parameter.
 *                                See examples in the overview section above.
 * @export
 * @return {void}
 * @memberof oj.Logger
 * @since 1.0.0
 */
oj.Logger.warn = function(args)
{
  oj.Logger._write(oj.Logger.LEVEL_WARN, oj.Logger._METHOD_WARN, arguments);
};

/**
 * Writes a general message.
 * @param {...(Object|string|number)} args The method supports a variable number of arguments, string substitutions and accepts a function as a parameter.
 *                                See examples in the overview section above.
 * @return {void}
 * @export
 * @memberof oj.Logger
 * @since 1.0.0
 */
oj.Logger.log = function(args)
{
  oj.Logger._write(oj.Logger.LEVEL_LOG, oj.Logger._METHOD_LOG, arguments);
};

/**
 * Method for setting and getting logger option/options
 * <p>Sets/gets logger configuration - level and/or writer. Accepts variable number of arguments.
 * <p><h5>Defaults:</h5>
 * Default level: oj.Logger.LEVEL_ERROR<br/>
 * Default writer: null; writes to the console
 * <p><h5>Usages:</h5>
 * <i>oj.Logger.option(optionName)</i> gets the value associated the the specified optionName<br/>
 * <i>oj.Logger.option()</i> gets an object containing key/value pairs representing the logger options hash<br/>
 * <i>oj.Logger.option(optionName, value)</i> sets  the option value associated with optionName<br/>
 * <i>oj.Logger.option(options)</i> sets  one or more options for the logger
 *
 * @example <caption>Overriding default options</caption>
 * oj.Logger.option("level",  oj.Logger.LEVEL_INFO);
 * oj.Logger.option("writer",  customWriter);  //an object that implements the following methods: log(), info(), warn(), error()
 *
 * @param {Object|string} [key]
 * @param {*} [value]
 * @ojsignature {target: "Type", for: "returns", value: "any"}
 * @export
 * @memberof oj.Logger
 * @since 1.0.0
 */
oj.Logger.option = function (key, value)
{
  //getters
  var ret = {}, opt;
  if (arguments.length == 0) {
     for (opt in oj.Logger._options) {
      if (oj.Logger._options.hasOwnProperty(opt)) {
        ret[opt]=oj.Logger._options[opt];
      }
     }
     return ret;
  }
  if (typeof key === "string" && value === undefined) {
     return oj.Logger._options[key] === undefined ? null : oj.Logger._options[key];
  }

  //setters
  if (typeof key === "string") {
    oj.Logger._options[key] = value;
  }
  else { // case when all options are set in one call
    var options = key;
    for (opt in options) {
      if (options.hasOwnProperty(opt)) {
        oj.Logger.option(opt, options[opt]);
      }
    }
  }
};

/* private members*/
/*
 * Helper method - calls a specified method on the available writer (console or custom)
 * if the logging level is sufficient
 */
oj.Logger._write = function(level, method, args)
{
  if (oj.Logger.option("level") < level) {
    return;
  }

  var writer = oj.Logger._getWriter();
  if (writer != null) {
    if (args.length == 1 && (args[0] instanceof Function)) {
      var msg = args[0]();
      args = [msg];
    }
    if (writer[method] && writer[method].apply) {
      writer[method].apply(writer, args);
    }
    else if (writer[method]) {
      writer[method] = Function.prototype.bind.call(writer[method], writer);
      oj.Logger._write(level, method, args);
    }
  }
};

/*
 * Helper method - returns available writer (console or custom)
 */
oj.Logger._getWriter = function()
{
  var writer = null;
  if (oj.Logger.option("writer")) {
    writer =  oj.Logger.option("writer");
  }
  else if (typeof window !== 'undefined' && window.console !== undefined) {
    writer = window.console;
  }
  return writer;
};

/*
 * Helper method - validates an option against default options
 * Returns true if the option(key) is a default option
 */
oj.Logger._validateOption = function(key)
{
  return oj.Logger._defaultOptions[key] !== undefined;
};
/*
** Copyright (c) 2008, 2014, Oracle and/or its affiliates. All rights reserved.
**
**34567890123456789012345678901234567890123456789012345678901234567890123456789
*/

/**
 * @private
 */
var _checkpointManagerDelegate = _scope['__ojCheckpointManager'];

/**
 * Global Checkpoint Manager Instance
 * @const
 * @export
 * @ignore
 */
oj.CHECKPOINT_MANAGER = {};

/**
 * Starts a checkpoint
 * @param {!string} name - the name of the checkpoint
 * @param {string=} description - optional description of the checkpoint
 * @export
 * @memberof! oj.CHECKPOINT_MANAGER
 */
oj.CHECKPOINT_MANAGER.startCheckpoint = function (name, description)
{
  if (_checkpointManagerDelegate)
  {
    _checkpointManagerDelegate['startCheckpoint'](name, description);
  }
};

/**
 * Ends a checkpoint
 * @param {!string} name - the name of the checkpoint
 * @export
 * @memberof! oj.CHECKPOINT_MANAGER
 */
oj.CHECKPOINT_MANAGER.endCheckpoint = function (name)
{
  if (_checkpointManagerDelegate)
  {
    _checkpointManagerDelegate['endCheckpoint'](name);
  }
};

/**
 * Retrieves a checkpoint record for a given name
 * @param {!string} name - the name of the checkpoint
 * @return {undefined|{start: number, end: number, duration: number, name: {string}, description: (string|undefined)}}
 * @export
 * @memberof! oj.CHECKPOINT_MANAGER
 */
oj.CHECKPOINT_MANAGER.getRecord = function (name)
{
  return _checkpointManagerDelegate ? _checkpointManagerDelegate['getRecord'](name) : undefined;
};

/**
 * Retrieves all checkpoint records matching a regular expression
 * @param {!RegExp} regexp - regular expression to match.
 * @return Array.{{start: number, end: number, duration: number, name: {string}, description: (string|undefined)}}
 * @export
 * @memberof! oj.CHECKPOINT_MANAGER
 */
oj.CHECKPOINT_MANAGER.matchRecords = function (regexp)
{
  return _checkpointManagerDelegate ? _checkpointManagerDelegate['matchRecords'](regexp) : [];
};

/**
 * Dumps matched records into oj.Logger
 * @param {!RegExp} regexp - regular expression for the records to dump.
 * @export
 * @memberof! oj.CHECKPOINT_MANAGER
 */
oj.CHECKPOINT_MANAGER.dump = function (regexp)
{
  oj.Logger.info(
    function()
    {
      var logMsg = "Checkpoint Records:";
      var records = oj.CHECKPOINT_MANAGER.matchRecords(regexp);
      for (var i=0; i<records.length; i++)
      {
        var record = records[i];
        logMsg = logMsg + '\n' + record['name'];
        var desc = record['description'];
        if (desc != null)
        {
          logMsg = logMsg + ' (' + desc + ')';
        }
        logMsg += ':\n';
        logMsg  = logMsg + 'start: ' + record['start'] + '\t' + 'duration: ' + record['duration'];
      }
      return logMsg;
    }
  );
}
/*
** Copyright (c) 2008, 2012, Oracle and/or its affiliates. All rights reserved.
**
**34567890123456789012345678901234567890123456789012345678901234567890123456789
*/
/*jslint browser: true*/
/*global define: false,goog: true*/

/**
 * Base class of all OJET Objects.
 * <p>
 * To create a subclass of another oj.Object, use oj.Object.createSubclass.
 * The subclass can specify class-level initialization by implementing an
 * <code>InitClass()</code> method on its constructor.  <code>InitClass</code>
 * is guaranteed to be called only once per class.  Further, a class'
 * <code>InitClass</code> method is guranteed to be called only after its
 * superclass' class initialization has been called.  When <code>InitClass</code>
 * is called, <code>this</code> is the class' constructor.  This allows class
 * initialization implementations to be shared in some cases.
 * </p>
 */


/**
 * @constructor oj.Object
 * @since 1.0
 * @export
 */
oj.Object = function()
{
  this.Init();
};

oj.Object.superclass = null;

/**
 * @private
 */
oj.Object._typeName = "oj.Object";

// regular expressicloneon for stripping out the name of a function
/**
 * @private
 */
oj.Object._GET_FUNCTION_NAME_REGEXP = /function\s+([\w\$][\w\$\d]*)\s*\(/;
// oj.Object._TRIM_REGEXP = /(^\s*)|(\s*$)/g; this.replace(/(^\s*)|(\s*$)/g, "");

oj.Object.prototype = {};
oj.Object.prototype.constructor = oj.Object;


/**
 * Creates a subclass of a baseClass
 * @method createSubclass
 * @memberof oj.Object
 * @param {Object} extendingClass The class to extend from the base class
 * @param {Object} baseClass class to make the superclass of extendingClass
 * @param {string=} typeName to use for new class.  If not specified, the typeName will be extracted from the
 * baseClass's function if possible
 * @return {void}
 * @export
 */
oj.Object.createSubclass = function(
  extendingClass,
  baseClass,
  typeName)  // optional name to name this class
{
  oj.Assert.assertFunction(extendingClass);
  oj.Assert.assertFunctionOrNull(baseClass);
  oj.Assert.assertStringOrNull(typeName);

  if (baseClass === undefined)
  {
    // assume oj.Object
    baseClass = oj.Object;
  }

  oj.Assert.assert(extendingClass !== baseClass, "Class can't extend itself");

  // use a temporary constructor to get our superclass as our prototype
  // without out having to initialize the superclass
  /**
   * @private
   * @constructor
   */
  var TempConstructor = oj.Object._tempSubclassConstructor;

  TempConstructor.prototype = baseClass.prototype;
  extendingClass.prototype = new TempConstructor();

  extendingClass.prototype.constructor = extendingClass;
  extendingClass.superclass = extendingClass["superclass"] = baseClass.prototype;

  if (typeName) {
    extendingClass._typeName = typeName;
  }
};

/**
 * Copies properties from the source object to the prototype of the target class
 * Only properties 'owned' by the source object will be copied, i.e. the properties
 * from the source object's prototype chain will not be included.
 * To copy properties from another class with methods defined on the prototype, pass
 * otherClass.prototype as the source.
 * @method copyPropertiesForClass
 * @memberof oj.Object
 * @param {Object} targetClass - the function whose prototype will be used a
 * copy target
 * @param {Object} source - object whose properties will be copied
 * @return {void}
 * @export
 */
oj.Object.copyPropertiesForClass = function(targetClass, source)
{
  var prop;
  oj.Assert.assertFunction(targetClass);
  oj.Assert.assert(source != null, "source object cannot be null");

  for(prop in source)
  {
    if(source.hasOwnProperty(prop))
    {
      targetClass.prototype[prop] = source[prop];
    }
  }
};

/**
 * @private
 */
oj.Object._tempSubclassConstructor = function(){};




/**
 * Returns the class object for the instance
 * @method getClass
 * @memberof oj.Object
 * @instance
 * @param {Object=} otherInstance - if specified, the instance whose type
 * should be returned. Otherwise the type of this instance will be returned
 * @return {Object} the class object for the instance
 * @final
 * @export
 */
oj.Object.prototype.getClass = function(
  otherInstance)
{
  if (otherInstance === undefined) {
    otherInstance = this;
  }
  else if (otherInstance === null)
  {
    return null;
  }
  return otherInstance["constructor"];
};


/**
 * Returns a clone of this object.  The default implementation is a shallow
 * copy.  Subclassers can override this method to implement a deep copy.
 * @method clone
 * @memberof oj.Object
 * @instance
 * @return {Object} a clone of this object
 * @export
 */
oj.Object.prototype.clone = function()
{
  var clone = new this.constructor();

  oj.CollectionUtils.copyInto(clone, this);

  return clone;
};

/**
 * @export
 * @method toString
 * @memberof oj.Object
 * @instance
 * @return {string}
 */
oj.Object.prototype.toString = function()
{
  return this.toDebugString();
};

/**
 * @export
 * @method toDebugString
 * @memberof oj.Object
 * @instance
 * @return {string}
 */
oj.Object.prototype.toDebugString = function()
{
  return this.getTypeName() + " Object";
};


/**
 * Returns the type name for a class derived from oj.Object
 * @method getTypeName
 * @memberof oj.Object
 * @instance
 * @param {Object|null} clazz Class to get the name of
 * @return {string} name of the Class
 * @export
 */
oj.Object.getTypeName = function(clazz)
{
  oj.Assert.assertFunction(clazz);

  var typeName = clazz._typeName, constructorText, matches;

  if (typeName == null)
  {
    constructorText = clazz.toString();
    matches = oj.Object._GET_FUNCTION_NAME_REGEXP.exec(constructorText);

    if (matches)
    {
      typeName = matches[1];
    }
    else
    {
      typeName = "anonymous";
    }

    // cache the result on the function
    clazz._typeName = typeName;
  }

  return typeName;
};

/**
 * Returns the type name for this instance
 * @method getTypeName
 * @memberof oj.Object
 * @return {string} name of the Class
 * @final
 * @export
 */
oj.Object.prototype.getTypeName = function()
{
  return oj.Object.getTypeName(this.constructor);
};

/**
 * Initializes the instance.  Subclasses of oj.Object must call
 * their superclass' Init
 * @export
 * @method Init
 * @return {void}
 * @memberof oj.Object
 * @instance
 */
oj.Object.prototype.Init = function()
{
  if (oj.Assert.isDebug()) {
    oj.Assert.assert(this["getTypeName"], "Not an oj.Object");
  }

  // do any class initialization.  This code is duplicated from
  // oj.Object.ensureClassInitialization()

  var currClass = this.constructor;
  if (!currClass._initialized) {
    oj.Object._initClasses(currClass);
  }
};

/**
 * Ensures that a class is initialized.  Although class initialization occurs
 * by default the first time that an instance of a class is created, classes that
 * use static factory methods to create their instances may
 * still need to ensure that their class has been initialized when the factory
 * method is called.
 *
 * @method ensureClassInitialization
 * @memberof oj.Object
 * @param {Object} clazz The class to ensure initialization of
 * @return {void}
 * @export
 */
oj.Object.ensureClassInitialization = function(clazz)
{
  oj.Assert.assertFunction(clazz);

  if (!clazz._initialized) {
    oj.Object._initClasses(clazz);
  }
};


/**
 * Indicates whether some other oj.Object is "equal to" this one.
 * Method is equivalent to java ".equals()" method.
 * @method equals
 * @memberof oj.Object
 * @instance
 * @param {Object} object - comparison target
 * @return {boolean} true if if the comparison target is equal to this object, false otherwise
 * @export
 */
oj.Object.prototype.equals = function(
  object)
{
  return this === object;
};

/**
 * Binds the supplied callback function to an object
 * @method createCallback
 * @memberof oj.Object
 * @param {!Object} obj - object that will be available to the supplied callback
 * function as 'this'
 * @param {!Object} func - the original callback
 * @return {function()} a function that will be invoking the original callback with
 * 'this' object assigned to obj
 * @ojsignature {target: "Type", for: "returns", value: "()=>any"}
 * @export
 */
oj.Object.createCallback = function(obj, func)
{
  oj.Assert.assertFunction(func);

  // All browsers supported by JET support bind() method
  return func.bind(obj);
};


/**
 * @private
 */
oj.Object._initClasses = function(currClass)
{
  if (oj.Assert.isDebug())
  {
    oj.Assert.assertFunction(currClass);
    oj.Assert.assert(!currClass._initialized);
  }

  currClass._initialized = true;

  var superclass = currClass.superclass, superclassConstructor, typeName, InitClassFunc;

  // initialize the superclass if necessary
  if (superclass)
  {
    superclassConstructor = superclass.constructor;

    if (superclassConstructor && !superclassConstructor._initialized) {
      oj.Object._initClasses(superclassConstructor);
    }

  }


  // if the class has an initialization function, call it
  InitClassFunc = currClass.InitClass

  if (InitClassFunc)
  {
    InitClassFunc.call(currClass);
  }
};

/**
 * Compares 2 values using strict equality except for the case of
 * <ol>
 *   <li> Array [order matters]; will traverse through the arrays and compare oj.Object.compareValues(array[i], array2[i]) </li>
 *   <li> Instances that support valueOf [i.e. Boolean, String, Number, Date, and etc] will be compared by usage of that function </li>
 * </ol>
 * @param {*} obj1 The first value to compare.
 * @param {*} obj2 The second value to compare.
 * @return {boolean}
 * @public
 * @export
 * @method compareValues
 * @memberof oj.Object
 */
oj.Object.compareValues = function (obj1, obj2)
{
  if (obj1 === obj2)
  {
    return true;
  }

  var obj1Type = typeof obj1,
      obj2Type = typeof obj2;

  if (obj1Type !== obj2Type)
  {
    //of different type so consider them unequal
    return false;
  }

  //At this point means the types are equal

  //note that if the operand is an array or a null then typeof is an object
  //check if either is null and if so return false [i.e. case where one might be a null and another an object]
  //and one wishes to avoid the null pointer in the following checks. Note that null === null has been already tested
  if(obj1 === null || obj2 === null)
  {
    return false;
  }

  //now check for constructor since I think by here one has ruled out primitive values and if the constructors
  //aren't equal then return false
  if(obj1.constructor === obj2.constructor)
  {

    //these are special cases and will need to be modded on a need to have basis
    if(Array.isArray(obj1))
    {
      return oj.Object._compareArrayValues(obj1, obj2);
    }
    else if(obj1.constructor === Object)
    {
      //for now invoke innerEquals and in the future if there are issues then resolve them
      return oj.Object.__innerEquals(obj1, obj2);
    }
    else if(obj1["valueOf"] && typeof obj1["valueOf"] === "function")
    {
      //test cases for Boolean, String, Number, Date
      //Note if some future JavaScript constructors
      //do not impl it then it's their fault
      return obj1.valueOf() === obj2.valueOf();
    }

  }

  return false;
};

oj.Object._compareArrayValues = function (array1, array2)
{
  if (array1.length !== array2.length)
  {
    return false;
  }

  for (var i = 0, j = array1.length;i < j;i++)
  {
    //recurse on each of the values, order does matter for our case since do not wish to search
    //for the value [expensive]
    if (!oj.Object.compareValues(array1[i], array2[i]))
    {
      return false;
    }
  }
  return true;
}

//Comparion of two Objects containing id and or index properties. 
//Note: it returns false if one is an id and other is an index
//if ids are the same, index will be ignored if there is only one is provided
oj.Object._compareIdIndexObject = function (obj1, obj2)
{
  if ((typeof obj1 === "number" && typeof obj2 === "number") ||
      (typeof obj1 === "string" && typeof obj2 === "string"))
  {
    return obj1 == obj2;
  }

  if (typeof obj1 === "object" && typeof obj2 === "object")
  {
    if (obj1["id"] && obj2["id"])
    {
      if (obj1["id"] != obj2["id"])
        return false;

      if (obj1["index"] && obj2["index"])
        return obj1["index"] == obj2["index"];

      return true;
    }
    else if (obj1["index"] && obj2["index"])
    {
      return obj1["index"] == obj2["index"];
    }
  }

  return false;  
};


//Comparion of two arrays containing ids, indexes, or objects where each object has id, 
//index or both properties.
//order needn't be same but no duplicates
oj.Object._compareArrayIdIndexObject = function (array1, array2)
{
  //null and [] are equals
  if (!array1)
    return (!array2 || array2.length == 0);

  if (!array2)
    return (!array1 || array1.length == 0);

  if (array1.length != array2.length)
    return false;

  for (var i = 0; i < array1.length; i++)
  {
    var found = false;
    for (var j = 0; j < array2.length; j++)
    {
      if (oj.Object._compareIdIndexObject (array1[i], array2[j]))
      {
        found = true;
        break;
      }
    }
    if (! found)
      return false;
  }

  return true;
};


oj.Object.__innerEquals = function (obj1, obj2) {
  var prop, hasProperties = false;

  if (obj1 === obj2) {
    return true;
  }

  if (!(obj1 instanceof Object) || !(obj2 instanceof Object)) {
    return false;
  }

  if (obj1.constructor !== obj2.constructor)
  {
    return false;
  }

  for (prop in obj1)
  {
    if (!hasProperties)
    {
      hasProperties = true;
    }
    if (obj1.hasOwnProperty(prop)) {
      if (!obj2.hasOwnProperty(prop))
      {
        return false;
      }

      if (obj1[prop] !== obj2[prop])
      {
        if (typeof(obj1[prop]) !== 'object') {
          return false;
        }

        if (!oj.Object.__innerEquals(obj1[prop], obj2[prop]))
        {
          return false;
        }
      }
    }
  }

  for (prop in obj2)
  {
    if (!hasProperties)
    {
      hasProperties = true;
    }

    if (obj2.hasOwnProperty(prop) && !obj1.hasOwnProperty(prop)) {
      return false;
    }
  }

  if (!hasProperties)
  {
    // we are dealing with objects that have no properties like Number or Date.
    return JSON.stringify(obj1) === JSON.stringify(obj2);
  }

  return true;
};

/**
 * @method isEmpty
 * @return {boolean}
 * @memberof oj.Object
 */
oj.Object.isEmpty = function(object) {
    var prop;
    // Test if an object is empty
    if (object === undefined || object === null) {
        return true;
    }

    for (prop in object) {
        if (object.hasOwnProperty(prop)) {
            return false;
        }
    }
    return true;
};


/**
 * @private
 * @return  {boolean} true if AMD Loader (such as Require.js) is present,
 *                    false otherwise
 */
oj.__isAmdLoaderPresent = function()
{
  return (typeof define === 'function' && define['amd']);
};



/*
** Copyright (c) 2008, 2011, Oracle and/or its affiliates. All rights reserved.
*/

/*jslint browser: true*/

/**
 * Assertion utilities.
 * The container is expected to have already initialized the oj.Assert Object before this
 * code is executed and initialized the oj.Assert.DEBUG flag/
 * @class
 * @export
 * @ignore
 */
oj.Assert = {};

/**
 * @private
 */
var _DEBUG = 'DEBUG';

/**
 * Forces DEBUG to be set to true
 * @export
 * @memberof oj.Assert
 */
oj.Assert.forceDebug = function()
{
  oj.Assert[_DEBUG] = true;
};

/**
 * Forces DEBUG to be set to false
 * @export
 * @memberof oj.Assert
 */
oj.Assert.clearDebug = function() {
  oj.Assert[_DEBUG] = false;
};

/**
 * Determines whether oj.Assert is running in debug mode
 * @return {boolean} true for debug mode, false otherwise
 * @export
 * @memberof oj.Assert
 */
oj.Assert.isDebug = function()
{
  return oj.Assert[_DEBUG] == true;
}


/**
 * Asserts that a condition is true.  If the condition does not
 * evaluate to true, an exception is thrown with the optional message
 * and reason
 * @param {boolean} condition condition to test
 * @param {string=} message message to display
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assert = function(
  condition,
  message
  )
{
  if (oj.Assert[_DEBUG] && !condition)
  {
    var myMessage = message || "", i;
    if (arguments.length > 2)
    {      
      myMessage += "(";
      for(i=2; i<arguments.length; i=i+1)
      {
        myMessage += arguments[i];
      }
      myMessage += ")";
    }
    oj.Assert.assertionFailed(myMessage, 1);
  }
};

/**
 * Convenience function for asserting when an abstact function is called
 * @export
 * @memberof oj.Assert
 */
oj.Assert.failedInAbstractFunction = function()
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertionFailed("Abstract function called", 1);
  }
};



/**
 * Asserts that the the target object has the same prototype as the example
 * type
 * @param {Object} target description
 * @param {Function} theConstructor
 * @param {string=} reason
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertPrototype = function(
  target,
  theConstructor,
  reason
  )
{
  var thePrototype;
  
  if (oj.Assert[_DEBUG])
  {
    if (target != null)
    {
      oj.Assert.assertType(theConstructor, "function", null, 1, false);
      thePrototype = theConstructor.prototype;
    
      if (!thePrototype.isPrototypeOf(target))
      {
        oj.Assert.assertionFailed("object '" + target + "' doesn't match prototype "
                               + thePrototype,
                               1,
                               reason);
      }
    }
    else
    {
      oj.Assert.assertionFailed("null object doesn't match prototype " + thePrototype, 1, reason);
    }
  }
};

/**
 * Asserts that the the target object has the same prototype as the example
 * type or is null.
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertPrototypeOrNull = function(
  target,
  theConstructor,
  reason
  )
{
  var thePrototype;
  
  if (oj.Assert[_DEBUG] && (target != null))
  {
    if (target != null)
    {
      oj.Assert.assertType(theConstructor, "function", null, 1, false);
      thePrototype = theConstructor.prototype;
    
      if (!thePrototype.isPrototypeOf(target))
      {
        oj.Assert.assertionFailed("object '" + target + "' doesn't match prototype "
                               + thePrototype,
                               1,
                               reason);
      }
    }
    else
    {
      oj.Assert.assertionFailed("null object doesn't match prototype " + thePrototype, 1, reason);
    }
  }
};

/**
 * Asserts that the the target object has the same prototype as the example
 * types
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertPrototypes = function(
  target,
  instanceOne,
  instanceTwo,
  reason
  )
{
  if (oj.Assert[_DEBUG])
  {
    var thePrototype = instanceOne.prototype, thePrototypeTwo = instanceTwo.prototype;
    
    if (!(thePrototype.isPrototypeOf(target) ||
          thePrototypeTwo.isPrototypeOf(target)))
    {
      oj.Assert.assertionFailed("object '" + target + "' doesn't match prototype "
                             + thePrototype + " or " + thePrototypeTwo,
                             1,
                             reason);
    }
  }
};


/**
 * Asserts that the target is a DOM Node or Null
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertDomNodeOrNull = function(target, depth)
{
  if (oj.Assert[_DEBUG] && target)
  {
    if (target["nodeType"] === undefined)
    {
      oj.Assert.assertionFailed(target + " is not a DOM Node", depth + 1);
    }
  }
};

/**
 * Asserts that the target is a DOM Node
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertDomNode = function(target, depth)
{
  if (oj.Assert[_DEBUG])
  {
    if (!target || (target["nodeType"] === undefined))
    {
      oj.Assert.assertionFailed(target + " is not a DOM Node", depth + 1);
    }
  }
};

/**
 * Asserts that the target is a DOM Element and optionally has the specified
 * element name
 * @param {Object} target target object
 * @param {string=} nodeName name of the element
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertDomElement = function(target, nodeName)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertDomNode(target, 1);

    if (target.nodeType !== 1)
    {
      oj.Assert.assertionFailed(target + " is not a DOM Element", 1);
    }
    else if (nodeName && (target.nodeName !== nodeName))
    {      
      oj.Assert.assertionFailed(target + " is not a " + nodeName + " Element", 1);
    }
  }
};

/**
 * Asserts that the target is a DOM Element and optionally has the specified
 * element name
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertDomElementOrNull = function(target, nodeName)
{
  if (oj.Assert[_DEBUG] && (target != null))
  {
    oj.Assert.assertDomNode(target, 1);

    if (target.nodeType !== 1)
    {
      oj.Assert.assertionFailed(target + " is not a DOM Element", 1);
    }
    else if (nodeName && (target.nodeName !== nodeName))
    {
      oj.Assert.assertionFailed(target + " is not a " + nodeName + " Element", 1);
    }
  }
};


/**
 * Asserts that the target object has the typeof specified
 * 
 * @param {Object|null|string|undefined} target 
 * @param {string} type typeof type that statisfies this condition
 * @param {string|undefined|null} prefix
 * @param {number} depth stack depth to skip when printing stack traces
 * @param {boolean} nullOK true if a null value satisfies this condition
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertType = function(target,type,prefix,depth, nullOK)
{
  if (oj.Assert[_DEBUG])
  {
    // either the target is null and null is OK, or the target better
    // be of the correct type
    var message, targetType = typeof target;
    if (!(((target == null) && nullOK) || (targetType === type)))
    {
      message = target + " is not of type " + type;
      
      if (prefix) {
        message = prefix + message;
      }
        
      if (!depth) {
        depth = 0;
      }
        
      oj.Assert.assertionFailed(message, depth + 1);
    }
  }
};

/**
 * Asserts that the target is an Object
 * @param {Object} target description
 * @param {string=} prefix
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertObject = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "object", prefix, 1, false);
  }
};

/**
 * Asserts that the target is an Object or null
 * @param {Object} target description
 * @param {string=} prefix
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertObjectOrNull = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "object", prefix, 1, true);
  }
};

/**
 * Asserts that the target is a non-empty String
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertNonEmptyString = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "string", prefix, 1, false);
    oj.Assert.assert(target.length > 0, "empty string"); 
  }
};

/**
 * Asserts that the target is a String
 * @param target target object
 * @param {string=} prefix prefix string
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertString = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "string", prefix, 1, false);
  }
};

/**
 * Asserts that the target is a String or null
 * @param {string|null|undefined|Object} target target object
 * @param {string=} prefix prefix string
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertStringOrNull = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "string", prefix, 1, true);
  }
};

/**
 * Asserts that the target is a Function
 * @param {Object} target target object
 * @param {string=} prefix prefix string
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertFunction = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "function", prefix, 1, false);
  }
};

/**
 * Asserts that the target is a Function or null
 * @param {Object} target target object
 * @param {string=} prefix prefix
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertFunctionOrNull = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "function", prefix, 1, true);
  }
};

/**
 * Asserts that the target is a boolean 
 * @param {Object} target description
 * @param {string=} prefix
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertBoolean = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "boolean", prefix, 1, false);
  }
};

/**
 * Asserts that the target is a number
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertNumber = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "number", prefix, 1, false);
  }
};

/**
 * Asserts that the target is a number or Null
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertNumberOrNull = function(target, prefix)
{
  if (oj.Assert[_DEBUG])
  {
    oj.Assert.assertType(target, "number", prefix, 1, true);
  }
};


/**
 * Asserts that the target object is an Array
 * @param {Object} target target object
 * @param {string=} message optional message
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertArray = function(
  target,
  message
  )
{
  if (oj.Assert[_DEBUG])
  {
    if (!Array.isArray(target))
    {
      if (message === undefined) {
        message = target + " is not an array";
      }
        
      oj.Assert.assertionFailed(message, 1);
    }
  }
};

/**
 * Asserts that the target object is an Array or null
 * @param {Object} target target object
 * @param {string=} message optional message
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertArrayOrNull = function(
  target,
  message)
{
  if (oj.Assert[_DEBUG] && (target != null))
  {
    if (!Array.isArray(target))
    {
      if (message === undefined) {
        message = target + " is not an array";
      }
        
      oj.Assert.assertionFailed(message, 1);
    }
  }
};


/**
 * Asserts that the target object is not either a number, or convertible to a number
 * @param {Object} target target object
 * @param {string=} message optional message
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertNonNumeric = function(
  target,
  message
  )
{
  if (oj.Assert[_DEBUG])
  { 
    if (!isNaN(target))
    {
      if (message === undefined) {
        message = target + " is convertible to a number";
      }
        
      oj.Assert.assertionFailed(message, 1);
    }
  }
};

/**
 * Asserts that the target object is either a number, or convertible to a number
 * @param {Object} target target object
 * @param {string=} message optional message
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertNumeric = function(
  target,
  message
  )
{
  if (oj.Assert[_DEBUG])
  {
    if (isNaN(target))
    {
      if (message === undefined) {
        message = target + " is not convertible to a number";
      }
        
      oj.Assert.assertionFailed(message, 1);
    }
  }
};

/**
 * Asserts that value String is in the Set
 * @param {Object} value value to check
 * @param {Object} set set to check
 * @param {string=} message optional message
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertInSet = function(
  value,
  set,
  message)
{
  var keyString, k;
  if ((value == null) || (set[value.toString()] === undefined))
  {
    if (message === undefined)
    {
      keyString = " is not in set: {";
      
      for (k in set)
      {
        if (set.hasOwnProperty(k)) {
            keyString += k;
            keyString += ",";
        }
      }
      
      keyString += "}";
      
      message = value + keyString;
    }
      
    oj.Assert.assertionFailed(message, 1);
  }
};

/**
 * Base assertion failure support that supports specifying the stack skipping
 * level
 * @param {string} message Message to display
 * @param {number} skipLevel assertion level
 * @param {string=} reason reason to display
 * @export
 * @memberof oj.Assert
 */
oj.Assert.assertionFailed = function(
  message,
  skipLevel,
  reason)
{  
  if (!skipLevel) {
    skipLevel = 0;
  }

  var errorMessage = "Assertion", stackTrace, stackTraceString, error;
  
  if (reason)
  {
    errorMessage += " (" + reason + ")";
  }
  
  errorMessage += " failed: ";
  
  if (message !== undefined)
  {
    errorMessage += message;
  }
    
  error = new Error(errorMessage);
  

  throw error;
};

/**
 * @private
 * @memberof oj.Assert
 */
var _assertSetting = _scope['__oj_Assert_DEBUG'];

if (_assertSetting !== undefined)
{
  oj.Assert[_DEBUG] = _assertSetting;
}

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true*/

/**
 * @export
 * @class oj.EventSource
 * @classdesc Object which supports subscribing to and firing events
 * @constructor
 * @since 1.1
 */
oj.EventSource = function()
{
    this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.EventSource, oj.Object, "oj.EventSource");

/**
 * Initializes the instance.
 * @export
 */
oj.EventSource.prototype.Init = function()
{
    this._eventHandlers = [];
    oj.EventSource.superclass.Init.call(this);
};

/**
 * Attach an event handler.
 * <p>Application can call this if it wants to be notified of an event.  It can call the <code class="prettyprint">off</code> method to detach the handler when it no longer wants to be notified.</p>
 * @param {string} eventType eventType
 * @param {function(Object)} eventHandler event handler function
 * @return {void}
 * @memberof oj.EventSource
 * @export
 */
oj.EventSource.prototype.on = function(eventType, eventHandler)
{
    var foundEventHandler = false, i;
    for (i = 0; i < this._eventHandlers.length; i++)
    {
        if (this._eventHandlers[i]['eventType'] == eventType && 
            this._eventHandlers[i]['eventHandlerFunc'] == eventHandler)
        {
            foundEventHandler = true;
            break;
        }
    }
    if (!foundEventHandler) {
        this._eventHandlers.push({'eventType': eventType, 'eventHandlerFunc': eventHandler});
    }
};

/**
 * Detach an event handler.
 * <p>Application can call this if it no longer wants to be notified of an event that it has attached an handler to using the <code class="prettyprint">on</code> method.</p>
 * @param {string} eventType eventType
 * @param {function(Object)} eventHandler event handler function
 * @return {void}
 * @memberof oj.EventSource
 * @export
 */
oj.EventSource.prototype.off = function(eventType, eventHandler)
{
    var i;
    for (i = this._eventHandlers.length-1; i >= 0; i--)
    {
        if (this._eventHandlers[i]['eventType'] == eventType && 
            this._eventHandlers[i]['eventHandlerFunc'] == eventHandler)
        {
            this._eventHandlers.splice(i, 1);
            break;
        }
    }
};

/**
 * Handle the event
 * @param {string} eventType  event type
 * @param {Object} event  event
 * @return {boolean} Returns false if event is cancelled
 * @memberof oj.EventSource
 * @export
 */
oj.EventSource.prototype.handleEvent = function(eventType, event)
{
    var i, returnValue;
    for (i = 0; i < this._eventHandlers.length; i++)
    {
        var eventHandler = this._eventHandlers[i];
        if (eventHandler['eventType'] == eventType)
        {
            returnValue = eventHandler['eventHandlerFunc'].apply(this, Array.prototype.slice.call(arguments).slice(1));
            
            if (returnValue === false)
            {
                // event cancelled
                return false;
            }
        }
    }
    
    return true;
};
/*
** Copyright (c) 2008, 2013, Oracle and/or its affiliates. All rights reserved.
**
**34567890123456789012345678901234567890123456789012345678901234567890123456789
*/
/*jslint browser: true*/
/*global require:false,ojt:true */

/**
 * @class oj.Config
 * @hideconstructor
 * @classdesc Services for setting and retrieving configuration options
 * @since 1.0
 * @export
 */
oj.Config = {};

/**
 * Retrieves the current locale
 * @memberof oj.Config
 * @method getLocale
 * @return {string} current locale
 * @export 
 */
oj.Config.getLocale = function()
{
  var rl, loc;
  if (oj.__isAmdLoaderPresent())
  {
    oj.Assert.assert(typeof ojt !== 'undefined', "ojtranslations module must be defined");
    rl = ojt['_ojLocale_'];
    
    // If Require.js internationalziation plugin resolved the locale to "root" (presumably because "lang" attribute was not
    // set, and neither navigator.language or navigator.userLanguage were not available), return "en"
    return (rl == "root") ? "en" : rl;
  }
    loc = oj.Config._locale;
    if (loc == null)
    {
      loc = document.documentElement.lang;
      if (!loc)
      {
         loc = navigator === undefined ? "en" :
                            (navigator['language'] ||
                             navigator['userLanguage'] || "en").toLowerCase();
      }
      oj.Config._locale = loc = loc.toLowerCase();
    }
    return loc;
};

/**
 * Changes the current locale
 * @method setLocale
 * @param {string} locale (language code and subtags separated by dash)
 * @param {function(): void} [callback] - for applications running with an AMD Loader (such as Require.js), this optional callback 
 * will be invoked when the framework is done loading its translated resources and Locale Elements for the newly specified locale. 
 * For applications running without an AMD loader, this optional callback will be invoked immediately
 * @return {undefined}
 * @export
 * @memberof oj.Config
 */
oj.Config.setLocale = function(locale, callback)
{
  if (oj.__isAmdLoaderPresent())
  {
    var prefix = "ojL10n!ojtranslations/nls/",
        requestedBundles = [prefix + locale + "/ojtranslations"];
    
    var timezoneBundleCount = 0;
    
    // Request LocaleElements only if ojlocaledata module is loaded
    if (oj.LocaleData) 
    {
      requestedBundles.push(prefix + locale + "/localeElements");
      
      if (oj.TimezoneData)
      {
        var tzBundles = oj.TimezoneData.__getBundleNames();
        timezoneBundleCount = tzBundles.length;
        tzBundles.forEach(
          function(bundle)
          {
            requestedBundles.push(prefix + locale + bundle)
          }
        );
      }
    }
    
    require(requestedBundles,
      function(translations, localeElements)
      {
        ojt = translations;
        
        if (localeElements)
        {
          oj.LocaleData.__updateBundle(localeElements);
        }
        
        for (var i=0; i<timezoneBundleCount; i++)
        {
          var tzBundle = arguments[i+2];
          oj.TimezoneData.__mergeIntoLocaleElements(tzBundle);
        }
        
        if (callback)
        {
          callback();
        }
      }
    );
  }
  else
  {
    oj.Config._locale = locale;
    if (callback)
    {
      callback();
    }
  }
};

/**
 * Retrieves a URL for loading a component-specific resource.
 * The URL is resolved as follows:
 * 1. If the application has specified a base URL with setResourceBaseUrl(), the return values will be
 * a relative path appended to the base URL.
 * 2. Otherwise, if the application running with an AMD Loader (such as Require.js), the parent folder of a 
 * module with ojs/ mapping will be used as a base URL.
 * 3. Otherwise, the original relative path will be returned.
 * @method getResourceUrl
 * @param {string} relativePath resource path
 * @return {string} resource URL
 * @see oj.Config.setResourceBaseUrl
 * @export
 * @memberof oj.Config
 */
oj.Config.getResourceUrl = function(relativePath)
{
  // Returning null and full URLs (containing protocol or a leading slash) as is
  var fullUrlExp = /^\/|:/, base, modulePath;
  if (relativePath == null || fullUrlExp.test(relativePath))
  {
    return relativePath;
  }
  
  base = oj.Config._resourceBaseUrl;
  
  if (base)
  {
    return base + (base.charAt(base.length-1) == '/' ? "" : '/') + relativePath;
  }
  
  if (oj.__isAmdLoaderPresent())
  {
    // : use ojs/_foo_ instead of ojs/ojcore to handle the case when ojs.core ends up in a partition bundle
    // in a different location
    modulePath = require.toUrl("ojs/_foo_");
    return modulePath.replace(/[^\/]*$/, "../" + relativePath);
  }
  
  return relativePath;
};

/**
 * Sets the base URL for retrieving component-specific resources
 * @method setResourceBaseUrl
 * @param {string} baseUrl base URL
 * @return {undefined}
 * @see oj.Config.getResourceUrl
 * @export
 * @memberof oj.Config
 */
oj.Config.setResourceBaseUrl = function(baseUrl)
{
  oj.Config._resourceBaseUrl = baseUrl;  
};

/**
 * Sets the automation mode.
 * @method setAutomationMode
 * @param {string} mode "enabled" for running in automation mode
 * @return {undefined}
 * @see oj.Config.getAutomationMode
 * @export
 * @memberof oj.Config
 */
oj.Config.setAutomationMode = function(mode)
{
  oj.Config._automationMode = mode;
};

/**
 * Gets the automation mode.
 * @method getAutomationMode
 * @return {string} automation mode
 * @see oj.Config.setAutomationMode
 * @export
 * @memberof oj.Config
 */
oj.Config.getAutomationMode = function()
{
  return oj.Config._automationMode;
}; 

/**
 * Return a string containing important version information about JET and the libraries
 * it has loaded
 * @method getVersionInfo 
 * @return {string}
 * @export
 * @memberof oj.Config
 */
oj.Config.getVersionInfo = function()
{
    // JET information
    var info = "Oracle JET Version: " + oj['version'] + '\n';
    info += "Oracle JET Revision: " + oj['revision'] + '\n';
    
    var windowDefined = (typeof window !== 'undefined');
    
    // Browser information
    if (windowDefined && window.navigator) {
        info += "Browser: " + window.navigator.userAgent +'\n';
        info += "Browser Platform: " + window.navigator.platform +'\n';
    }
    
    // 3rd party libraries
    if ($) {
        if ($.fn) {
            info += "jQuery Version: " + $.fn.jquery + '\n';
        }
        if ($.ui && $.ui['version']) {
            info += "jQuery UI Version: " + $.ui['version'] + '\n';
        }
    }
    if (oj.ComponentBinding) {
        info += "Knockout Version: " + oj.ComponentBinding.__getKnockoutVersion() + '\n';
    }
    
    // Local require doesn't have version #
    if (windowDefined && window.require) {
        info += "Require Version: " + window.require['version'] + '\n';
    }
    
    return info;
};

/**
 * Dump information to the browser's console containing important version information about JET and
 * the libraries it has loaded
 * @method logVersionInfo 
 * @return {undefined}
 * @memberof oj.Config
 * @export
 */
oj.Config.logVersionInfo = function()
{
    console.log(oj.Config.getVersionInfo());
};

/**
 * Retrives JET's template engine for dealing with inline templates (currently internal only)
 * @ignore
 * @memberof oj.Config
 * @private
 */
oj.Config.__getTemplateEngine = function()
{
  if (!oj.Config._templateEnginePromise)
  {
    if (!oj.__isAmdLoaderPresent())
    {
      throw "JET Temlate engine cannot be loaded with an AMD loader";
    }
    oj.Config._templateEnginePromise = new Promise(
      function(resolve, reject)
      {
        require(['./ojtemplateengine'], resolve, reject);
      }
    );
  }
  return oj.Config._templateEnginePromise;
};
/* The custom element (webcomponents) support requires the native CustomEvent
 * object.  This polyfill provides CustomEvent implementation for browsers that
 * don't support it yet.
 */
(function () {

  if (typeof window === 'undefined') {
    return;
  }

  // defaultPrevented is broken in IE.
  // https://connect.microsoft.com/IE/feedback/details/790389/event-defaultprevented-returns-false-after-preventdefault-was-called
  var workingDefaultPrevented = (function() {
    var e = document.createEvent('Event');
    e.initEvent('foo', true, true);
    e.preventDefault();
    return e.defaultPrevented;
  })();

  if (!workingDefaultPrevented) {
    var origPreventDefault = Event.prototype.preventDefault;
    Event.prototype.preventDefault = function() {
      if (!this.cancelable) {
        return;
      }

      origPreventDefault.call(this);

      Object.defineProperty(this, 'defaultPrevented', {
        get: function() {
          return true;
        },
        configurable: true
      });
    };
  }

  if (typeof window['CustomEvent'] === "function") {
    return;
  }

  function CustomEvent (event, params) {
    params = params || {bubbles: false, cancelable: false, detail: undefined};

    var evt = document.createEvent('CustomEvent');

    evt.initCustomEvent(event, params.bubbles, params.cancelable, params.detail );
    return evt;
   }

  CustomEvent.prototype = window.Event.prototype;

  window['CustomEvent'] = CustomEvent;
})();

/* This polyfill implements a proposed Microsoft standard [1] for effective yielding.
 * With the setImmediate global function, developers can yield control flow to the
 * user agent before running script.  The yield is similar to the setTimeout function
 * in that it is evaluated in the macrotask queue.  However, the setTimeout often has
 * a minimum delay and is also subject to long delays when the browser is placed in the
 * background.  The setImmediate function implemented by this polyfill invokes the
 * callback in the "next-tick" after the current macrotask queue has been exhausted.
 *
 * [1] https://dvcs.w3.org/hg/webperf/raw-file/tip/specs/setImmediate/Overview.html
 *
 * The strategy for the polyfill implementation uses the window.postMessage API for
 * creating a context for calling the target function evaulated as a macrotask. This
 * plugin will not work in a webworker where the window object doesn't exist.
 */
(function ()
{

  if (typeof window === 'undefined' || window["setImmediate"] || !window["postMessage"])
  {
    return;
  }

  var _setImmediateMap;

  var _setImmediateCounter;
  function _nextId()
  {
    if (isNaN(_setImmediateCounter))
      _setImmediateCounter = 0;

    return ++_setImmediateCounter;
  }

  // postMessage "message" event listener for the setImmediate impl
  function _nextTickHandler (event)
  {
    var data = event.data;
    if (!data || "oj-setImmeidate" !== data.message)
      return;

    var id = data.id;
    var entry = _setImmediateMap.get(id);
    clearImmediateImpl(id);

    if (entry)
    {
      var callback = entry.callback;
      var args = entry.args;
      callback.apply(window, args);
    }
  };

  function setImmediateImpl ()
  {
    var callback = arguments[0];
    var slice = Array.prototype.slice;
    var args = slice.call(arguments, 1);

    if (typeof callback !== "function")
      callback = new Function(callback.toString());

    var id = _nextId();

    if (!_setImmediateMap)
      _setImmediateMap = new Map();

    _setImmediateMap.set(id, {callback: callback, args: args});

    if (_setImmediateMap.size === 1)
      window.addEventListener("message", _nextTickHandler);

    window.postMessage({id: id, message: "oj-setImmeidate"}, "*");
    return id;
  }

  function clearImmediateImpl(id)
  {
    if (!_setImmediateMap)
      return;

    _setImmediateMap.delete(id);

    if (_setImmediateMap.size < 1)
    {
      window.removeEventListener("message", _nextTickHandler);
      _setImmediateMap = null;
    }
  }

  window["setImmediate"] = setImmediateImpl;
  window["clearImmediate"] = clearImmediateImpl;
})();

(function ()
{
  if (typeof window === 'undefined' || window["__extends"])
  {
    return;
  }
  
  var __extends = (this && this.__extends) || (function () {
      var extendStatics = Object.setPrototypeOf ||
          ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
          function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
      return function (d, b) {
          extendStatics(d, b);
          /**
           * @constructor
           */
          function __() { this.constructor = d; }
          d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
      };
  })();
  
  window["__extends"] = __extends; 
})();

(function ()
{
  if (typeof window === 'undefined') {
    return;
  }
  if (window['Symbol']) {
    if (!window['Symbol']['asyncIterator']) {
      window['Symbol']['asyncIterator'] = 'asyncIterator';
    }

    if (!window['Symbol']['iterator']) {
      window['Symbol']['iterator'] = 'iterator';
    }
  } else {
    window['Symbol'] = {};
    window['Symbol']['asyncIterator'] = 'asyncIterator';
    window['Symbol']['iterator'] = 'iterator';
  }
})();



/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/


/**
 * @ojtsignore
 * @class oj.AgentUtils
 * @classdesc Utilities for qualifying the user agent string.
 * @public
 * @ignore
 */
oj.AgentUtils = function () {};

/**
 * Identity of the target browser.
 * @enum {string}
 * @public
 * @memberof oj.AgentUtils
 */
oj.AgentUtils.BROWSER =
{
  IE: "ie",
  FIREFOX: "firefox",
  SAFARI: "safari",
  CHROME: "chrome",
  EDGE: "edge",
  UNKNOWN: "unknown"
};
/**
 * Browser layout engine identity.
 * @enum {string}
 * @public
 * @memberof oj.AgentUtils
 */
oj.AgentUtils.ENGINE =
{
  TRIDENT: "trident",
  WEBKIT: "webkit",
  GECKO: "gecko",
  BLINK: "blink",
  EDGE_HTML: "edgehtml",
  UNKNOWN: "unknown"
};
/**
 * Operating system identity.
 * @enum {string}
 * @public
 * @memberof oj.AgentUtils
 */
oj.AgentUtils.OS =
{
  WINDOWS: "Windows",
  SOLARIS: "Solaris",
  MAC: "Mac",
  UNKNOWN: "Unknown",
  ANDROID: "Android",
  IOS: "IOS",
  WINDOWSPHONE: "WindowsPhone",
  LINUX: "Linux"
};
/**
 * Parses the browser user agent string determining what browser and layout engine
 * is being used.
 *
 * @param {Object|null|string=} userAgent a specific agent string but defaults to navigator userAgent if not provided
 * @return {{os: oj.AgentUtils.OS, browser: oj.AgentUtils.BROWSER, browserVersion: number,
 *          engine: oj.AgentUtils.ENGINE, engineVersion: number, hashCode: number}}
 * @public
 * @memberof oj.AgentUtils
 */
oj.AgentUtils.getAgentInfo = function (userAgent)
{
  if (oj.StringUtils.isEmptyOrUndefined(userAgent))
    userAgent = navigator.userAgent;
  userAgent = userAgent.toLowerCase();
  /** @type {number} */
  var hashCode = oj.StringUtils.hashCode(userAgent);
  var currAgentInfo = oj.AgentUtils._currAgentInfo;
  if (currAgentInfo && currAgentInfo["hashCode"] === hashCode)
    return {
      'os': currAgentInfo["os"],
      'browser': currAgentInfo["browser"],
      'browserVersion': currAgentInfo["browserVersion"],
      'engine': currAgentInfo["engine"],
      'engineVersion': currAgentInfo["engineVersion"],
      'hashCode': currAgentInfo["hashCode"]
    };
  /** @type {oj.AgentUtils.OS} */
  var os = oj.AgentUtils.OS.UNKNOWN;
  /** @type {oj.AgentUtils.BROWSER} */
  var browser = oj.AgentUtils.BROWSER.UNKNOWN;
  /** @type {number} */
  var browserVersion = 0;
  /** @type {oj.AgentUtils.ENGINE} */
  var engine = oj.AgentUtils.ENGINE.UNKNOWN;
  /** @type {number} */
  var engineVersion = 0;
  if (userAgent.indexOf("iphone") > -1 || userAgent.indexOf("ipad") > -1)
    os = oj.AgentUtils.OS.IOS;
  else if (userAgent.indexOf("mac") > -1)
    os = oj.AgentUtils.OS.MAC;
  else if (userAgent.indexOf("sunos") > -1)
    os = oj.AgentUtils.OS.SOLARIS;
  else if (userAgent.indexOf("android") > -1)
    os = oj.AgentUtils.OS.ANDROID;
  else if (userAgent.indexOf("linux") > -1)
    os = oj.AgentUtils.OS.LINUX;
  else if (userAgent.indexOf("windows phone") > -1)
    os = oj.AgentUtils.OS.WINDOWSPHONE;
  else if (userAgent.indexOf("win") > -1)
    os = oj.AgentUtils.OS.WINDOWS;
  if (userAgent.indexOf("msie") > -1)
  {
    browser = oj.AgentUtils.BROWSER.IE;
    browserVersion = oj.AgentUtils._parseFloatVersion(userAgent, /msie (\d+[.]\d+)/);
    if (userAgent.indexOf("trident"))
    {
      engine = oj.AgentUtils.ENGINE.TRIDENT;
      engineVersion = oj.AgentUtils._parseFloatVersion(userAgent, /trident\/(\d+[.]\d+)/);
    }
  }
  else if (userAgent.indexOf("trident") > -1)
  {
    browser = oj.AgentUtils.BROWSER.IE;
    browserVersion = oj.AgentUtils._parseFloatVersion(userAgent, /rv:(\d+[.]\d+)/);
    if (userAgent.indexOf("trident"))
    {
      engine = oj.AgentUtils.ENGINE.TRIDENT;
      engineVersion = oj.AgentUtils._parseFloatVersion(userAgent, /trident\/(\d+[.]\d+)/);
    }
  }
  else if (userAgent.indexOf("edge") > -1)
  {
    browser = oj.AgentUtils.BROWSER.EDGE;
    browserVersion = engineVersion  = oj.AgentUtils._parseFloatVersion(userAgent, /edge\/(\d+[.]\d+)/);
    engine = oj.AgentUtils.ENGINE.EDGE_HTML;
  }
  else if (userAgent.indexOf("chrome") > -1)
  {
    browser = oj.AgentUtils.BROWSER.CHROME;
    browserVersion = oj.AgentUtils._parseFloatVersion(userAgent, /chrome\/(\d+[.]\d+)/);
    if (browserVersion >= 28)
    {
      engine = oj.AgentUtils.ENGINE.BLINK;
      engineVersion = browserVersion;
    }
    else
    {
      engine = oj.AgentUtils.ENGINE.WEBKIT;
      engineVersion = oj.AgentUtils._parseFloatVersion(userAgent, /applewebkit\/(\d+[.]\d+)/);
    }
  }
  else if (userAgent.indexOf("safari") > -1)
  {
    browser = oj.AgentUtils.BROWSER.SAFARI;
    browserVersion = oj.AgentUtils._parseFloatVersion(userAgent, /version\/(\d+[.]\d+)/);
    engine = oj.AgentUtils.ENGINE.WEBKIT;
    engineVersion = oj.AgentUtils._parseFloatVersion(userAgent, /applewebkit\/(\d+[.]\d+)/);
  }
  else if (userAgent.indexOf("firefox") > -1)
  {
    browser = oj.AgentUtils.BROWSER.FIREFOX;
    browserVersion = oj.AgentUtils._parseFloatVersion(userAgent, /rv:(\d+[.]\d+)/);
    engine = oj.AgentUtils.ENGINE.GECKO;
    engineVersion = oj.AgentUtils._parseFloatVersion(userAgent, /gecko\/(\d+)/);
  }

  currAgentInfo = oj.AgentUtils._currAgentInfo =
  {
    'hashCode': hashCode,
    'os': os,
    'browser': browser,
    'browserVersion': browserVersion,
    'engine': engine,
    'engineVersion': engineVersion
  };
  return {
    'os': currAgentInfo["os"],
    'browser': currAgentInfo["browser"],
    'browserVersion': currAgentInfo["browserVersion"],
    'engine': currAgentInfo["engine"],
    'engineVersion': currAgentInfo["engineVersion"],
    'hashCode': currAgentInfo["hashCode"]
  };
};
/**
 * @param {string!} userAgent
 * @param {RegExp!} versionNumberPattern
 * @return {number}
 * @private
 * @memberof oj.AgentUtils
 */
oj.AgentUtils._parseFloatVersion = function (userAgent, versionNumberPattern)
{
  var matches = userAgent.match(versionNumberPattern);
  if (matches)
  {
    var versionString = matches[1];
    if (versionString)
      return parseFloat(versionString);
  }

  return 0;
};



/*
** Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
*/
/*jslint browser: true*/

/**
 * @class oj.ThemeUtils
 * @classdesc Services for getting information from the theme
 * @since 1.2.0
 * @export
 */
oj.ThemeUtils = function(){};

/**
 * get the name of the current theme
 * @method getThemeName
 * @memberof oj.ThemeUtils
 * @export
 * @static
 * @hideconstructor
 *
 * @return {string|null} the name of the theme
 */
oj.ThemeUtils.getThemeName = function()
{
  
  // get the map of theme info
  var themeMap = oj.ThemeUtils.parseJSONFromFontFamily("oj-theme-json") || {};

  return themeMap['name'];
}


/**
 * <p>Get the target platform of the current theme. </p>
 * <p>This API does not look at the user agent and therefore it 
 *    tells you nothing about the current platform you are actually on.  
 *    Instead it tells you the target platform the theme was written 
 *    for so that programmatic behaviors that match the theme's UI can be written.
 *    This can be useful when writing a cross platform hybrid mobile app.  </p>
 * 
 * <p>Example</p>
 * <pre class="prettyprint">
 * <code>
 * var themeTargetPlatform = oj.ThemeUtils.getThemeTargetPlatform();
 *
 * if (themeTargetPlatform == 'ios')
 *    // code for a behavior familiar in ios
 * else if (themeTargetPlatform == 'android')
 *    // code for a behavior familiar on android
 * else
 *    // code for the default behavior
 * </code></pre>
 * @export
 * @static
 * @method getThemeTargetPlatform
 * @memberof oj.ThemeUtils
 * @return {string|null} the target platform can be any string the theme 
 * wants to send down, but the usual values are 'web', 'ios', 'android', 'windows'
 */
oj.ThemeUtils.getThemeTargetPlatform = function()
{
  
  // get the map of theme info
  var themeMap = oj.ThemeUtils.parseJSONFromFontFamily("oj-theme-json") || {};

  return themeMap['targetPlatform'];
}



/**
 * clear values cached in  [oj.ThemeUtils.parseJSONFromFontFamily]{@link oj.ThemeUtils.parseJSONFromFontFamily}
 * @export
 * @static
 * @method clearCache
 * @return {void}
 * @memberof oj.ThemeUtils
 */
oj.ThemeUtils.clearCache = function()
{
  this._cache = null;
}


/**
 *
 * <p>json can be sent down as the font family in classes 
 * that look something like this 
 * (on the sass side of things see the file 
 * scss/utilities/_oj.utilities.json.scss
 * for information on jet mixins available to generate json):<p>
 *
 * <p>Example CSS</p>
 * <pre class="prettyprint">
 * <code>
 * .demo-map-json {
 *    font-family: '{"foo":"bar", "binky": 4}';
 * }
 *
 * .demo-list-json {
 *    font-family: '["foo","bar","binky"}';
 * }
 * </code></pre>
 * <p>Example Usage</p>
 * <pre class="prettyprint">
 * <code>
 * var mymap = oj.ThemeUtils.parseJSONFromFontFamily("demo-map-json");
 * var myarray = oj.ThemeUtils.parseJSONFromFontFamily("demo-list-json");
 * </code></pre>
 *
 * This function 
 * <ul>
 *   <li>Gets the font family string by creating a dom element, 
 *      applying the selector passed in, calling getcomputedstyle, 
 *      and then reading the value for font-family.
 *   </li>
 *   <li>Parses the font family value by calling JSON.pars.</li>
 *   <li>Caches the parsed value because calling getComputedStyle is a perf hit. 
 *       Subsequent requests for the same selector will return the cached value. 
 *       Call [oj.ThemeUtils.clearCache]{@link oj.ThemeUtils.clearCache} if new css is loaded.</li>
 *   <li>Return the parsed value.</li>
 * </ul>
 * 
 * <p>
 * If new css is loaded call oj.ThemeUtils.clearCache to clear the cache</p>
 *
 * @method parseJSONFromFontFamily
 * @memberof oj.ThemeUtils
 * @param {string} selector a class selector name, for example 'demo-map-json';
 * @return {*} the result of parsing the font family with JSON.parse. 
 *      The returned value is cached, so if you modify the returned 
 *      value it will be reflected in the cache.
 * @throws {SyntaxError} If JSON.parse throws a SyntaxError exception we will log an error and rethrow 
 * @export
 * @static
 */
oj.ThemeUtils.parseJSONFromFontFamily = function(selector)
{

 
 // NOTE: I first tried code inspired by 
 // https://css-tricks.com/making-sass-talk-to-javascript-with-json/
 // so I was using :before and content, for example
 //   .oj-button-option-defaults:before {
 //     content: '{"foo":"bar", "binky": 4}';
 //    }
 // 
 //  however IE 11 has a bug where the computed style doesn't actually 
 //  seem to be computed when it comes to :before,
 //  so if you set a class that affects :before after the page loads 
 //  on IE getComputedStyle doesn't work.
 //  See the pen below, the yellow box has the class applied in js,
 //  computedstyle works on chrome, doesn't work on IE11 for 
 //  class applied after page load.
 //     http://codepen.io/gabrielle/pen/OVOwev

  // if needed create the cache and initialize some things
  if (this._cache == null)
  {
    this._cache = {};

    // magic value that means null in the cache
    this._null_cache_value = {};

    // font family is inherited, so even if no selector/json 
    // is sent down we will get a string like 
    // 'HelveticaNeue',Helvetica,Arial,sans-serif' off of our generated element. 
    // So save off the font family from the head 
    // element to compare to what we read off our generated element. 
    this._headfontstring = window.getComputedStyle(document.head).getPropertyValue('font-family');
  }

  // see if we already have a map for this component's option defaults
  var jsonval =  this._cache[selector];

  // if there's something already cached return it
  if (jsonval === this._null_cache_value)
  {
    return null;
  }
  else if (jsonval != null)
  {
    return jsonval;
  }
    

  // there's nothing cached, so we need to create a dom element to apply the class to. 
  // We're creating a meta element, 
  // the hope is that the browser is smart enough to realize the 
  // meta element isn't visible and therefore we avoid perf issues of calling 
  // getcomputedstyle
  var elem = document.createElement("meta");
  elem.className = selector;
  document.head.appendChild(elem); // @HTMLUpdateOK 
  var rawfontstring = window.getComputedStyle(elem).getPropertyValue('font-family');

  if (rawfontstring != null)
  {

    // if the raw font string is the same value as the saved header 
    // font value then log a warning that no value was sent down.

    if (rawfontstring == this._headfontstring)
    {
      oj.Logger.warn("parseJSONFromFontFamily: When the selector ", selector, 
        " is applied the font-family read off the dom element is ", rawfontstring, 
        ". The parent dom elment has the same font-family value.",
        " This is interpreted to mean that no value was sent down for selector ", 
        selector, ". Null will be returned.");
    }
    else
    {
      // remove inconsistent quotes
      var fontstring = rawfontstring.replace(/^['"]+|\s+|\\|(;\s?})+|['"]$/g, '');
      //console.log("json fontstring for selector " + selector + ': ' + fontstring);

      if (fontstring)
      {

        try {

          jsonval = JSON.parse(fontstring);
        }
        catch(e)
        {

          // In Firefox you can turn off the page font
          // Options -> Content -> Fonts and Colors -> Advanced
          // Uncheck the "Allow pages to choose their own fonts, instead of my selections above" 
          // In that case they stick something like 'serif,"'at the front of the font family,
          // so search for the first comma, add 2, and try parsing again.
          var commaindex = fontstring.indexOf(',');
          var reparseSuccess = false;

          if (commaindex > -1)
          {
            fontstring = fontstring.substring(commaindex + 2);

            try {

              jsonval = JSON.parse(fontstring);
              reparseSuccess = true;
            }
            catch(e2)
            {
            }
          }

          if (reparseSuccess == false)
          {
            oj.Logger.error("Error parsing json for selector " + selector + 
                            ".\nString being parsed is " + fontstring + ". Error is:\n", e);
            
            // remove the meta tag
            document.head.removeChild(elem); // @HTMLUpdateOK 
            throw e;
          }

        }
      }
    }
  }

  // remove the meta tag
  document.head.removeChild(elem); // @HTMLUpdateOK 

  // cache the result 
  if (jsonval == null)
  {
    this._cache[selector] = this._null_cache_value;
  }
  else
  {
    this._cache[selector] = jsonval;
  }
  
  //console.log(this._cache);

  return jsonval;
}




/*jslint browser: true*/
/*
** Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
*/


/**
 * @class oj.ResponsiveUtils
 * @classdesc Utilities for working with the framework's responsive screen widths 
 * and ranges. Often used in conjunction with {@link oj.ResponsiveKnockoutUtils} 
 * to create knockout observables that can be used to drive responsive page behavior. 
 * See the method doc below for specific examples.
 * 
 * @since 1.1.0
 * @hideconstructor
 * @export
 */
oj.ResponsiveUtils = function() {};


/**
 * <p>In the jet sass files there are variables for
 * responsive screen widths, these look something like</p>
 *  <ul>
 *    <li>$screenSmallRange:  0, 767px;</li>
 *    <li>$screenMediumRange: 768px, 1023px;</li>
 *    <li>$screenLargeRange:  1024px, 1280px;</li>
 *    <li>$screenXlargeRange: 1281px, null;</li>
 *  </ul>
 *
 * <p>These constants are used to identify these ranges.</p>
 * @enum {string}
 * @memberof oj.ResponsiveUtils
 * @constant
 * @export
 */
oj.ResponsiveUtils.SCREEN_RANGE ={
   /**
   * @expose
   * @constant
   */
   SM: "sm",
   /**
   * @expose
   * @constant
   */
   MD: "md",
   /**
   * @expose
   * @constant
   */
   LG: "lg",
   /**
   * @expose
   * @constant
   */
   XL: "xl",
   /**
   * @expose
   * @constant
   */
   XXL: "xxl"

};


/**
 * <p>In the jet sass files there are variables for
 * responsive screen widths,
 * see {@link oj.ResponsiveUtils.SCREEN_RANGE} for details.
 * The jet sass files also has variables for
 * responsive queries like $responsiveQuerySmallUp,
 * $responsiveQuerySmallOnly, $responsiveQueryMediumUp, etc.</p>
 *
 * <p>These constants are used to identify these queries.</p>
 * @enum {string}
 * @memberof oj.ResponsiveUtils
 * @constant
 * @export
 */
oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY = {
   /**
   * Matches screen width small and wider
   * @expose
   * @constant
   */
   SM_UP: "sm-up",
   /**
   * matches screen width medium and wider
   * @expose
   * @constant
   */
   MD_UP: "md-up",
   /**
   * matches screen width large and wider
   * @expose
   * @constant
   */
   LG_UP: "lg-up",
   /**
   * matches screen width extra-large and wider
   * @expose
   * @constant
   */
   XL_UP: "xl-up",
   /**
   * matches screen width extra-extra-large and wider
   * @expose
   * @constant
   */
   XXL_UP: "xxl-up",

   /**
   * matches screen width small only
   * @expose
   * @constant
   */
   SM_ONLY: "sm-only",
   /**
   * matches screen width medium only
   * @expose
   * @constant
   */
   MD_ONLY: "md-only",
   /**
   * matches screen width large only
   * @expose
   * @constant
   */
   LG_ONLY: "lg-only",
   /**
   * matches screen width extra-large only
   * @expose
   * @constant
   */
   XL_ONLY: "xl-only",

   /**
   * matches screen width medium and narrower
   * @expose
   * @constant
   */
   MD_DOWN: "md-down",
   /**
   * matches screen width large and narrower
   * @expose
   * @constant
   */
   LG_DOWN: "lg-down",
   /**
   * matches screen width extra-large and narrower
   * @expose
   * @constant
   */
   XL_DOWN: "xl-down",
   /**
   * matches high resolution screens
   * @expose
   * @constant
   */
   HIGH_RESOLUTION: "high-resolution"

};

// used by the compare function
oj.ResponsiveUtils._RANGE = {};
oj.ResponsiveUtils._RANGE[oj.ResponsiveUtils.SCREEN_RANGE.SM]   = 0;
oj.ResponsiveUtils._RANGE[oj.ResponsiveUtils.SCREEN_RANGE.MD]   = 1;
oj.ResponsiveUtils._RANGE[oj.ResponsiveUtils.SCREEN_RANGE.LG]   = 2;
oj.ResponsiveUtils._RANGE[oj.ResponsiveUtils.SCREEN_RANGE.XL]   = 3;
oj.ResponsiveUtils._RANGE[oj.ResponsiveUtils.SCREEN_RANGE.XXL]  = 4;



/**
 * This idea/code is from zurb foundation, thanks zurb!
 *
 * In the jet sass files there are variables for
 * responsive screen sizes, these look something like

 *  <ul>
 *    <li>$screenSmallRange:  0, 767px;</li>
 *    <li>$screenMediumRange: 768px, 1023px;</li>
 *    <li>$screenLargeRange:  1024px, 1280px;</li>
 *    <li>$screenXlargeRange: 1281px, null;</li>
 *  </ul>
 *
 * <p>These variables in turn are used to generate responsive media queries in variables like
 * $responsiveQuerySmallUp, $responsiveQueryMediumUp, etc.</p>
 *
 * <p>we send down these media queries as the font family in classes
 * that look something like this:<p>
 *
 * <pre class="prettyprint">
 * <code>
 * .oj-mq-md {
 *    font-family: "/screen and (min-width: 768px)/";
 * }
 * </code></pre>
 *
 * <p>This function applies the class and then reads the font family off a dom
 * element to get the media query string</p>
 *
 * @param {string} selector a class selector name, for example 'oj-mq-md';
 * @return {string} the media query sent down for that class
 * @private
 */
oj.ResponsiveUtils._getMediaQueryFromClass = function(selector)
{

  var elem = /** @type {(Element | null)} */ (document.getElementsByClassName(selector).item(0));

  if(elem === null) {
    elem = document.createElement("meta");
    elem.className = selector;
    document.head.appendChild(elem); // @HTMLUpdateOK
  }

  var fontFamily= window.getComputedStyle(elem).getPropertyValue('font-family');

  return fontFamily.replace(/^[\/\\'"]+|(;\s?})+|[\/\\'"]+$/g, '');
}

/**
 * Get a framework (built in) media query string, 
 * see {@link oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY} for details on framework queries. 
 * The media query string returned can be passed to 
 * {@link oj.ResponsiveKnockoutUtils.createMediaQueryObservable} to create a knockout
 * observable, which in turn can be used to drive responsive page behavior.
 * 
 * <p>Example:</p>
 * <pre class="prettyprint">
 * <code>
 *
 *     var lgQuery = oj.ResponsiveUtils.getFrameworkQuery(
 *                             oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.LG_UP);
 *        
 *     self.large = oj.ResponsiveKnockoutUtils.createMediaQueryObservable(lgQuery);
 * </code></pre>
 * 
 *
 * @method getFrameworkQuery
 * @memberof oj.ResponsiveUtils
 * @param {oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY} frameworkQueryKey one of the FRAMEWORK_QUERY_KEY constants,
 *                       for example oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.MD_UP
 * @return {string | null} the media query to use for the framework query key passed in
 * @export
 * @static
 */
oj.ResponsiveUtils.getFrameworkQuery = function(frameworkQueryKey)
{
  var selector = "oj-mq-" + frameworkQueryKey;
  var query = oj.ResponsiveUtils._getMediaQueryFromClass(selector);

  if (query == "null")
    return null;
  else
    return query;
}


/**
 * <p> Compare can be used in conjunction with 
 * {@link oj.ResponsiveKnockoutUtils.createScreenRangeObservable}</p>
 *
 * 
 * <p>Example:</p>
 * <pre class="prettyprint">
 * <code>
 *        // create an observable which returns the current screen range
 *        self.screenRange = oj.ResponsiveKnockoutUtils.createScreenRangeObservable();
 *
 *        self.label2 = ko.computed(function() {
 *          var range = self.screenRange();
 * 
 *          if ( oj.ResponsiveUtils.compare( 
 *                       range, oj.ResponsiveUtils.SCREEN_RANGE.MD) <= 0)
 *          {
 *            // code for when screen is in small or medium range
 *          }
 *          else if (range == oj.ResponsiveUtils.SCREEN_RANGE.XL)
 *          {
 *            // code for when screen is in XL range
 *          }
 *        });
 * </code></pre>
 * 
 * @method compare
 * @memberof oj.ResponsiveUtils
 * @param {oj.ResponsiveUtils.SCREEN_RANGE} size1 one of the screen size constants,
 * for example oj.ResponsiveUtils.SCREEN_RANGE.MD
 * @param {oj.ResponsiveUtils.SCREEN_RANGE} size2 one of the screen size constants,
 * for example oj.ResponsiveUtils.SCREEN_RANGE.LG
 * @return {number} a negative integer if the first
 * argument is less than the second. Zero if the two are equal.
 * 1 or greater if the first argument is more than the second.
 *
 * @export
 * @static
 */
oj.ResponsiveUtils.compare = function(size1, size2)
{
  var range1 = oj.ResponsiveUtils._RANGE[size1];
  var range2 = oj.ResponsiveUtils._RANGE[size2];

  if (range1 == undefined)
    throw "size1 param " + size1 +
          " illegal, please use one of the screen size constants like oj.ResponsiveUtils.SCREEN_RANGE.MD";
  if (range2 == undefined)
    throw "size2 param " + size2 +
          " illegal, please use one of the screen size constants like oj.ResponsiveUtils.SCREEN_RANGE.MD";

  return range1 - range2;
}

/*
** Copyright (c) 2004, 2012, Oracle and/or its affiliates. All rights reserved.
*/
/**
 * String utilities.
 * @class oj.StringUtils
 * @export
 * @ignore
 */
oj.StringUtils = {};

oj.StringUtils._TRIM_ALL_RE = /^\s*|\s*$/g;

/**
  * Returns true if the value is null or if the trimmed value is of zero length.
  *
  * @param {Object|string|null} value
  * @returns {boolean} true if the string or Object (e.g., Array) is of zero length.
  * @export
  * @memberof oj.StringUtils
  */
 oj.StringUtils.isEmpty = function(value)
 {
   if (value === null)
   {
     return true;
   }

   var trimValue = oj.StringUtils.trim(value);

   return (trimValue.length === 0);
 };

 /**
  * Returns true if the value is null, undefined or if the trimmed value is of zero length.
  *
  * @param {Object|string|null=} value
  * @returns {boolean} true if the string or Object (e.g., Array) is of zero length.
  * @export
  * @memberof oj.StringUtils
  */
 oj.StringUtils.isEmptyOrUndefined = function (value)
 {
   if (value === undefined || oj.StringUtils.isEmpty(value))
   {
     return true;
   }

   return false;
 };

/**
 * Test if an object is a string (either a string constant or a string object)
 * @param {Object|string|null} obj object to test
 * @return {boolean} true if a string constant or string object
 * @export
 * @memberof oj.StringUtils
 */
oj.StringUtils.isString = function(obj)
{
  return obj !== null && ((typeof obj === 'string') || obj instanceof String);
};

/**
 * Remove leading and trailing whitespace
 * @param {Object|string|null} data to trim
 * @return {Object|string|null}
 * @export
 * @memberof oj.StringUtils
 */
oj.StringUtils.trim = function(data)
{
  if (oj.StringUtils.isString(data))
  {
    return data.replace(oj.StringUtils._TRIM_ALL_RE, '');
  }

  return data;
};

/**
 * Port of the Java String.hashCode method.
 * http://erlycoder.com/49/javascript-hash-functions-to-convert-string-into-integer-hash-
 *
 * @param {string} str
 * @returns {number}
 * @public
 * @memberof oj.StringUtils
 */
oj.StringUtils.hashCode = function(str)
{
  var hash = 0;
  if (str.length === 0)
    return hash;

  for (var i = 0; i < str.length; i++)
  {
    var c = str.charCodeAt(i);
    hash = ((hash<<5)-hash)+c;
    hash = hash & hash;
  }
  return hash;
};


//Polyfills for IE11
(function () {  
  // String.startsWith requires for IE11
  if (!String.prototype.startsWith) {
    String.prototype.startsWith = function(searchString, position){
      position = position || 0;
      return this.substr(position, searchString.length) === searchString;
    };
  }

  // String.endsWith requires for IE11
  if (!String.prototype.endsWith) {
    String.prototype.endsWith = function(searchString, position) {
      var subjectString = this.toString();
      if (typeof position !== 'number' || !isFinite(position) || 
          Math.floor(position) !== position || position > subjectString.length) {
        position = subjectString.length;
      }
      position -= searchString.length;
      var lastIndex = subjectString.lastIndexOf(searchString, position);
      return lastIndex !== -1 && lastIndex === position;
    };
  }
})();

/*
** Copyright (c) 2004, 2012, Oracle and/or its affiliates. All rights reserved.
*/

/**
 * Utilities for working with collections
 * @export
 */
oj.CollectionUtils = {};


/**
 * Copies all of the properties of source into the target and returns the target
 * 
 * @param {Object} target - target collection
 * @param {Object} source - source collection
 * @param {function(string)=} keyConverter a callback for converting the key
 * @param {boolean=} recurse - true if this method should recurse into plain Javascript object properties
 * @param {number=} maxRecursionDepth - the maximum depth of the recursion into plain Javascript object properties
 * @return target collection
 * @export
 * @memberof! oj.CollectionUtils
 */
oj.CollectionUtils.copyInto = function(
  target,
  source,
  keyConverter,
  recurse,
  maxRecursionDepth)
{
  return oj.CollectionUtils._copyIntoImpl(
                                          target,
                                          source,
                                          keyConverter,
                                          recurse,
                                          maxRecursionDepth,
                                          0);
};


/**
 * Checks whether the object is a direct instance of Object
 * @param {Object} obj - object to test
 * 
 * @return {boolean} true if the object is a direct instance of Object, false otherwise
 * @export
 * @memberof! oj.CollectionUtils
 */
oj.CollectionUtils.isPlainObject = function(obj)
{
  if (obj !== null && typeof obj === 'object')
  {
    try
    {
      if (obj.constructor && obj.constructor.prototype.hasOwnProperty("isPrototypeOf"))
      {
        return true;
      }
    }
    catch(e){}
  }
  
  return false;
};


/**
 * @private
 * @memberof! oj.CollectionUtils
 */
oj.CollectionUtils._copyIntoImpl = function(
  target,
  source,
  keyConverter,
  recurse,
  maxRecursionDepth,
  currentLevel)
{
  var k, targetKey, keys;
  
  if (maxRecursionDepth === undefined || maxRecursionDepth === null)
  {
    maxRecursionDepth = Number.MAX_VALUE;
  }
  
  if (target && source && (target !== source))
  {    
    keys = Object.keys(source);
    for (var i=0; i<keys.length; i++)
    {      
      k = keys[i];
      // allow the key mapping to be overridden
      if (keyConverter)
      {
        targetKey = keyConverter(k);
      }
      else
      {
        targetKey = k;
      }
      
      var sourceVal = source[k];
      
      var recursed = false;
      
      if (recurse && currentLevel < maxRecursionDepth)
      {
        var targetVal = target[targetKey];
        if (oj.CollectionUtils.isPlainObject(sourceVal) && 
           (targetVal == null || oj.CollectionUtils.isPlainObject(targetVal))) 
        {
          recursed = true;
          target[targetKey] = targetVal || {};
          oj.CollectionUtils._copyIntoImpl(
                                          target[targetKey],
                                          sourceVal,
                                          keyConverter,
                                          true,
                                          maxRecursionDepth,
                                          currentLevel + 1);
        }
      }
      if (!recursed)
      {
        target[targetKey] = sourceVal;
      }

    }
  }
  
  return target;
};

/*
** Copyright (c) 2008, 2013, Oracle and/or its affiliates. All rights reserved.
**
**34567890123456789012345678901234567890123456789012345678901234567890123456789
*/

/*global ojt:false*/

/**
 * @class oj.Translations
 * @classdesc Services for Retrieving Translated Resources
 * @export
 * @since 1.0
 * @hideconstructor
 */
oj.Translations = {};

/**
 * Sets the translation bundle used by JET
 * If an AMD loader (such as Require.js) is not present, this method should be called by the application to provide
 * translated strings for JET.
 * This method may also be used by an application that wants to completely replace the resource bundle that is automatically
 * fetched by an AMD loader.
 * @method setBundle
 * @memberof oj.Translations
 * @param {Object} bundle resource bundle that should be used by the framework
 * @return {void}
 * @export
 */
oj.Translations.setBundle = function(bundle)
{
  oj.Translations._bundle = bundle;
};

/**
 * Retrives a translated resource for a given key
 * @method getResource
 * @memberof oj.Translations
 * @param {string} key resource key The dot character (.) within the key string
 * is interpreted as a separator for the name of a sub-section within the bundle.
 * For example, 'components.chart', would be read as the 'chart' section within
 * 'components'. Thus the key name for an individual section should never contain a dot.
 * @return {Object|string|null} resource associated with the key or null if none was found
 * @export
 */
oj.Translations.getResource = function(key)
{
  return oj.Translations._getResourceString(key);
};

/**
 * Applies parameters to a format pattern
 * @method applyParameters
 * @memberof oj.Translations
 * @param {string} pattern pattern that may contain tokens like {0}, {1}, {name}. These tokens
 * will be used to define string keys for retrieving values from the parameters
 * object. Token strings should not contain comma (,) 
 * or space characters, since they are reserved for future format type enhancements. 
 * The reserved characters within a pattern are:
 * $ { } [ ]  
 * These characters will not appear in the formatted output unless they are escaped
 * with a dollar character ('$').
 * 
 * @param {Object|Array} parameters parameters to be inserted into the string. Both arrays and
 * Javascript objects with string keys are accepted.
 * 
 * @return {string|null} formatted message or null if the pattern argument was null
 * @export
 */
oj.Translations.applyParameters = function(pattern, parameters)
{
  return (pattern == null) ? null : oj.Translations._format(pattern, parameters);
};

/**
 * Retrieves a translated string after inserting optional parameters. 
 * The method uses a key to retrieve a format pattern from the resource bundle.
 * Tokens like {0}, {1}, {name} within the pattern will be used to define placement
 * for the optional parameters.  Token strings should not contain comma (,) 
 * or space characters, since they are reserved for future format type enhancements.
 * The reserved characters within a pattern are:
 * $ { } [ ]  
 * These characters will not appear in the formatted output unless they are escaped
 * with a dollar character ('$').
 * @method getTranslatedString
 * @memberof oj.Translations
 * @param {string} key  translations resource key. The dot character (.) within the key string
 * is interpreted as a separator for the name of a sub-section within the bundle.
 * For example, 'components.chart', would be read as the 'chart' section within
 * 'components'. Thus the key name for an individual section should never contain a dot.
 * 
 * @param {...(string|Object|Array)} var_args  - optional parameters to be inserted into the 
 * translated pattern.
 * 
 * If more than one var_args arguments are passed, they will be treated as an array 
 * for replacing positional tokens like {0}, {1}, etc.
 * If a single argument is passed, it will be treated as a Javascript Object whose
 * keys will be matched to tokens within the pattern. Note that an Array is just
 * a special kind of such an Object.
 * 
 * For backward compatibility, a var_args argument whose type is neither 
 * Object or Array will be used to replace {0} in the pattern.
 * 
 * @return formatted translated string
 * @ojsignature {target: "Type", for:"returns", value: "string"}
 * @export
 */
oj.Translations.getTranslatedString = function(key, var_args)
{  
  var val = oj.Translations._getResourceString(key);
  
  if (val == null)
  {
    return key;
  }
  
  var params = {};
  
  if (arguments.length > 2)
  {
    params = Array.prototype.slice.call(arguments, 1);
  }
  else if (arguments.length == 2)
  {
    params = arguments[1];
    if (typeof params !== 'object' && !(params instanceof Array))
    {
      params = [params];
    }
      
  }
  
  return oj.Translations.applyParameters(val, params);
};


/**
 * Provides a key-to-value map of the translated resources for a given component name
 * @method getComponentTranslations
 * @memberof oj.Translations
 * @param {string} componentName name of the component
 * @return {Object} a map of translated resources
 * @export
 */
oj.Translations.getComponentTranslations = function(componentName)
{
  var bundle = oj.Translations._getBundle()[componentName], translations, k;
  
  if (bundle == null)
  {
    return {};
  }
  
  // Assume that the set of keys remains constant regardless of the current locale
  translations = {};
  for(k in bundle)
  {
    if (bundle.hasOwnProperty(k)) {
        translations[k] = bundle[k];
    }
  }
  return translations;
};

/**
 * Retrives a translated resource for a given key, accounting for nested keys
 * @param {string} key
 * @return {string|null} resource associated with the key or null if none was found
 * @private
 */
oj.Translations._getResourceString = function(key)
{
  // Account for dot separated nested keys
  var keys = key ? key.split(".") : [], bundle = oj.Translations._getBundle(), 
          iteration = keys.length, index = 0, subkey = keys[index];
  oj.Assert.assertObject(bundle);
  
  // even though we start with a valid bundle it's possible that part or all of the key is invalid, 
  // so check we have a valid bundle in the while loop
  while (--iteration > 0 && bundle) 
  {
    // if we have a key like a.b.c
    bundle = bundle[subkey];
    index++;
    subkey = keys[index];
  }
    
  return bundle ? (bundle[subkey] || null) : null;
};

oj.Translations._format = function(formatString, parameters)
{
  var formatLength = formatString.length;
  
  // Use the javascript StringBuffer technique.
  var buffer = [];
  
  var token = null;
  
  
  var escaped = false;
  var isToken = false;
  var isGroup = false;
  var isExcluded = false;
  
  var tokenTerminated; // this will be set to true when a comma or space is 
                       // encountered in teh token
  var i;
  
  for (i = 0; i < formatLength; i++)
  {
    var ch = formatString.charAt(i);
    
    var accumulate = false;
    
    if (!escaped)
    {
      switch(ch)
      {
        case '$':
          escaped = true;
          break;
          
        case '{':
          if (!isExcluded)
          {
            if (!isToken)
            {
              tokenTerminated = false;
              token = [];
            }
            isToken = true;
          }
          break;
          
        case '}':
          if (isToken && token.length > 0)
          {
            var val = parameters[token.join('')];
            buffer.push((val === undefined) ? "null" : val);
          }
          isToken = false;
          break;
          
        case '[':
          if (!isToken)
          {
            if (isGroup)
            {
              isExcluded = true;
            }
            else
            {
              isGroup = true;
            }
          }
          break;
          
        case ']':
          if (isExcluded)
          {
            isExcluded = false;
          }
          else
          {
            isGroup = false;
          }
          break;
        
        default:
          accumulate = true;
      }
    }
    else
    {
      accumulate = true;
      escaped = false;  
    }
    
    if (accumulate)
    {
      if (isToken)
      {
        if (ch == ',' || ch ==' ')
        {
          tokenTerminated = true;
        }
        else if (!tokenTerminated)
        {
          token.push(ch);
        }
      }
      else if (!isExcluded)
      {
        buffer.push(ch);
      }
    } 
  }

  // Use the javascript StringBuffer technique for toString()
  return buffer.join("");
};


oj.Translations._getBundle = function()
{
  var b = oj.Translations._bundle;
  if (b)
  {
    return b;
  }
  
  if (oj.__isAmdLoaderPresent()) {
    oj.Assert.assert(ojt !== undefined, "ojtranslations module must be defined");
    return ojt;
  }
  return {};
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * Internally used by the {@link oj.BusyContext} to track a components state
 * while it is performing a task such as animation or fetching data.
 * 
 * @hideconstructor
 * @ignore
 * @protected
 * @constructor
 * @param {Function|Object|undefined} description of the component and cause
 *        of the busy state
 */
oj.BusyState = function (description)
{
  /**
   * @ignore
   * @private
   * @type {?}
   */
  this._description = description;

  /**
   * @ignore
   * @private
   * @type {number}
   */
  this._addedWaitTs = oj.BusyState._getTs();

  /**
   * @ignore
   * @private
   * @type {string}
   */
  this._id = this._addedWaitTs.toString(36) + "_" + Math.random().toString(36); //@RandomNumberOk -
  // random number concatinated to the current timestamp is used for a unique id for a local Map
  // key. This random number is not used use as a cryptography key.

};

Object.defineProperties(oj.BusyState.prototype,
  {
    /**
     * Identifies the usage instance of a busy state.
     * @memberof oj.BusyState
     * @instance
     * @property {!string} id
     */
    "id" : {
      "get" : function ()
      {
        return this._id;
      },
      "enumerable" : true
    },
    /**
     * Further definition of the busy state instance.
     * @memberof oj.BusyState
     * @instance
     * @property {?string} description
     */
    "description" :
      {
        "get" : function ()
        {
          if (this._description)
          {
            if (this._description instanceof Function)
              return this._description();
            else
              return this._description.toString();
          }

        },
        "enumerable" : true
      }
  });

/**
 * @override
 * @returns {string} returns the value of the object as a string
 */
oj.BusyState.prototype.toString = function ()
{
  var buff = "Busy state: [description=";

  var description = this["description"];

  if (description !== null)
  {
    buff += description;
  }

  var elapsed = oj.BusyState._getTs() - this._addedWaitTs;
  buff += ", elapsed=" + elapsed + "]";

  return buff;
};


/**
 * @private
 * @returns {number} current date represented by a number
 */
oj.BusyState._getTs = function ()
{
  // Safari V9.1.1 doesn't yet support performance.now
  return window["performance"] ? window["performance"].now() : new Date().getTime();
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * <p>The purpose of the BusyContext API is to accommodate sequential dependencies of asynchronous
 * operations. A common use cases defining the API is for automation testing (qunit and webdriver).
 * Automation test developers can use this API to wait until components finish animation effects
 * or data fetch before trying to interact with components. The BusyContext is not limited to
 * test automation developers usages. It is also needed by page developers for waiting on run-time
 * operation readiness.</p>
 *
 * The Busy Context API will wait until busy states have resolved or a timeout period has elapsed.
 * There are several primary wait scenarios:
 * <ol>
 *   <li>Component creation and page bindings applied.</li>
 *   <li>Components that implement animation effects.</li>
 *   <li>Components that must fetch data from a REST endpoint.</li>
 *   <li>General wait conditions that are not specific to the Jet framework. The customer might
 *       choose to register some busy condition associated with application domain logic such
 *       as REST endpoints.</li>
 *   <li>Wait until the bootstrap of the page has completed - jet libraries loaded via requireJS.</li>
 * </ol>
 *
 * <p>The first step for waiting on a busy context is to determine what conditions are of interest
 *   to wait on. The granularity of a busy context can be scoped for the entirety of the page or
 *   for a specific DOM element. Busy contexts have hierarchical dependencies mirroring the
 *   document's DOM structure with the root being the page context. Depending on the particular
 *   scenario, developers might need to target one of the following busy context scopes:</p>
 * <ul>
 *   <li>Scoped for the Page - Automation test developers will more commonly choose the page busy
 *     context. This context represents the page as a whole. Automation developers commonly need
 *     to wait until the page is fully loaded before starting automation. More commonly, automation
 *     developers are interesting in testing the functionality of an application having multiple
 *     JET components versus just a single component.
 *
 *     <pre class="prettyprint">
 *     <code>
 *     var busyContext = oj.Context.getPageContext().getBusyContext();
 *     </code></pre>
 *
 *   </li>
 *   <li>Scoped for the nearest DOM Element - Application developers sometime need a mechanism to
 *     wait until a specific component operation has complete. For example, it might be desirable
 *     to wait until a component has been created and bindings applied before setting a property or
 *     calling a method on the component. Another scenario, waiting for a popup to finish
 *     open or close animation before initiating the next action in their application flow.
 *     For this problem space developers would need to obtain a busy context scoped for a DOM node.
 *     The "data-oj-context" marker attribute is used to define a busy context for a dom subtree.
 *
 *     <pre class="prettyprint">
 *     <code>
 *     <!-- subtree assigned a marker 'data-oj-context' attribute -->
 *     &lt;div id="mycontext" data-oj-context&gt;
 *        ...
 *        &lt;!-- JET content --&gt;
 *        ...
 *     &lt;/div&gt;
 *
 *     var node = document.querySelector("#mycontext");
 *     var busyContext = oj.Context.getContext(node).getBusyContext();
 *     busyContext.whenReady().then(function ()
 *     {
 *       var component = document.querySelector("#myInput");
 *       component.value = "foo";
 *       component.validate().then(function (isValid)
 *       {
 *         if (!isValid)
 *           component.value = "foobar";
 *       });
 *     });
 *     </code></pre>
 *
 *   </li>
 * </ul>
 *
 * The BusyContext API utilizes {@link oj.Logger.LEVEL_LOG} to log detail busy state activity.
 * <pre class="prettyprint">
 * <code>
 *  oj.Logger.option("level", oj.Logger.LEVEL_LOG);
 * </code></pre>
 *
 * <b>This constructor should never be invoked by the application code directly.</b>
 * @hideconstructor
 * @param {Element=} hostNode DOM element associated with this busy context
 * @export
 * @constructor oj.BusyContext
 * @since 2.1.0
 * @classdesc Framework service for querying the busy state of components on the page.
 */
oj.BusyContext = function (hostNode)
{
  this.Init(hostNode);
};

oj.Object.createSubclass(oj.BusyContext, oj.Object, "oj.BusyContext");


/**
 * see oj.BusyContext#setDefaultTimeout
 * @type {number}
 * @ignore
 * @private
 */
oj.BusyContext._defaultTimeout = Number.NaN;

/**
 * Sets a default for the optional <code>timeout</code> argument of the {@link oj.BusyContext#whenReady}
 * for all BusyContext instances. The default value will be implicitly used if a timeout argument is not
 * provided.
 *
 * @export
 * @see oj.BusyContext#whenReady
 * @since 3.1.0
 * @memberof oj.BusyContext
 * @method setDefaultTimeout
 * @param {number} timeout in milliseconds
 * @return {undefined}
 */
oj.BusyContext.setDefaultTimeout = function(timeout)
{
  if (!isNaN(timeout))
    oj.BusyContext._defaultTimeout = timeout;
};

/**
 * @param {Element=} hostNode DOM element associated with this busy context
 * @instance
 * @protected
 */
oj.BusyContext.prototype.Init = function (hostNode)
{
  oj.BusyContext.superclass.Init.call(this);

  this._hostNode = hostNode;

  /**
   * Busy states cache.
   *
   * @type {?}
   * @ignore
   * @private
   */
  this._statesMap = new Map();

  /**
   * Coordinates resolution of the master when ready promise with one or more slave
   * when ready promises having a timeout period.
   *
   * @type {Object}
   * @ignore
   * @private
   */
  this._mediator = {
    /**
     * Returns a master primise that will resolve when all busy states have been resolved.
     *
     * @returns {Promise}
     * @ignore
     * @private
     */
    getMasterWhenReadyPromise : function ()
    {
      if (!this._masterWhenReadyPromise)
        this._masterWhenReadyPromise = new Promise(this._captureWhenReadyPromiseResolver.bind(this));
      return this._masterWhenReadyPromise;
    },
    /**
     * Triggers resolution of the master promise and clears all timeouts associated with slave
     * when ready promises.
     *
     * @returns {void}
     * @ignore
     * @private
     */
    resolveMasterWhenReadyPromise : function ()
    {
      if (this._masterWhenReadyPromiseResolver)
        this._masterWhenReadyPromiseResolver(true);
      this._masterWhenReadyPromise = null;
      this._masterWhenReadyPromiseResolver = null;
    },
    /**
     * Returns a promise that will resolve when the master promise resolves or reject when
     * the slave timeout promise rejects.
     *
     * @param {Promise} master
     * @param {Function} generateErrorCallback
     * @param {number} timeout
     * @returns {Promise}
     * @ignore
     * @private
     */
    getSlaveTimeoutPromise : function (master, generateErrorCallback, timeout)
    {
      var timer;
      var slaveTimeoutPromise = new Promise(function (resolve, reject)
      {
        timer = window.setTimeout(function ()
        {
          reject(generateErrorCallback());
        }, timeout);
      });
      this._slaveTimeoutPromiseTimers.push(timer);

      // When the master promise is resolved, all timers may be cleared
      return Promise.race([master, slaveTimeoutPromise])
                                  .then(this._clearAllSlaveTimeouts.bind(this));
    },

    /**
     * @private
     * @ignore
     * @return {Promise} resolves on the next-tick using setImmediate.
     */
    getNextTickPromise: function()
    {
      if (!this._nextTickPromise)
      {
        this._nextTickPromise = new Promise(function (resolve)
        {
          window.setImmediate(function ()
          {
            this._nextTickPromise = null;
            resolve(true);
          }.bind(this));
        }.bind(this));
      }

      return this._nextTickPromise;
    },

    /**
     * Clears all window timeout timeers that are slave when ready promises.
     *
     * @returns {boolean}
     * @ignore
     * @private
     */
    _clearAllSlaveTimeouts : function ()
    {
      var slaveTimeoutPromiseTimers = this._slaveTimeoutPromiseTimers;
      this._slaveTimeoutPromiseTimers = [];
      for (var i = 0; i < slaveTimeoutPromiseTimers.length; i++)
        window.clearTimeout(slaveTimeoutPromiseTimers[i]);

      return true;
    },
    /**
     * Promise executor function passed as the single master promise constructor.  Captures the
     * promise resolve callback function.  The resolve promise function will be called when all the
     * busy states have been removed.
     *
     * @param {Function} resolve
     * @param {Function} reject
     * @returns {void}
     * @ignore
     * @private
     */
    _captureWhenReadyPromiseResolver : function (resolve, reject)
    {
      this._masterWhenReadyPromiseResolver = resolve;
    },
    /**
     * Array of setTimeout timers that should be cancled when the busy state resolves.
     *
     * @type {Array.<number>}
     * @ignore
     * @private
     */
    _slaveTimeoutPromiseTimers : []
      /**
       * The master when ready promise that will resovle when all busy states resolve.
       *
       * @type {Promise|undefined}
       * @ignore
       * @private
       */
      //_masterWhenReadyPromise : undefined,
      /**
       * The resolve function of the masterWhenReadyPromise.
       *
       * @type {Function|undefined}
       * @ignore
       * @private
       */
      //_masterWhenReadyPromiseResolver : undefined,
      /**
       * Promise evaluated next-tick.
       *
       * @type {Promise|undefined}
       * @ignore
       * @private
       */
      //_nextTickPromise : undefined
  };
};

/**
 * Logs the current registered busy states ordered acceding by the order they were added.
 * The cost of compiling the list is only made if the logger level is oj.Logger.LEVEL_LOG.
 * @param {?} statesMap busy states
 * @returns {void}
 * @private
 */
oj.BusyContext._log = function (statesMap)
{
  if (oj.Logger.option('level') !== oj.Logger.LEVEL_LOG)
    return;

  oj.Logger.log(">> Busy states: %d", statesMap.size);

  var busyStates = oj.BusyContext._values(statesMap);
  if (busyStates.length > 0)
    oj.Logger.log(busyStates.join("\n"));
};

/**
 * @param {?} statesMap busy states
 * @return {Array.<oj.BusyState>} Returns an array of busy states entries from the states map
 * @private
 */
oj.BusyContext._values = function (statesMap)
{
  var busyStates = [];
  statesMap.forEach(function (value)
  {
    busyStates.push(value);
  });

  return busyStates;
};

/**
 * <p>Called by components or services performing a task that should be considered
 * in the overall busy state of the page. An example would be animation or fetching data.</p>
 *
 * Caveats:
 * <ul>
 *   <li>Busy context dependency relationships are determined at the point the first busy state
 *       is added.  If the DOM node is re-parented after a busy context was added, the context will
 *       maintain dependencies with any parent DOM contexts.</li>
 *   <li>The application logic creating busy states is responsible for ensuring these busy states
 *       are resolved. Busy states added internally by JET are automatically accounted for.
 *       The busy states added by the application logic must manage a reference to the resolve
 *       function associated with a busy state and it must be called to release the busy state.</li>
 * </ul>
 *
 * <pre class="prettyprint">
 * <code>// apply the marker attribute to the element
 * &lt;div id="context1" data-oj-context ... &gt;&lt;/&gt;
 * ...
 * ...
 * var context1 = document.querySelector("#context1");
 *
 * // obtain a busy context scoped for the target node
 * var busyContext1 = oj.Context.getContext(context1).getBusyContext();
 * // add a busy state to the target context
 * var options = {"description": "#context1 fetching data"};
 * var resolve = busyContext1.addBusyState(options);
 * ...
 * ...  // perform asynchronous operation that needs guarded by a busy state
 * ...
 * // resolve the busy state after the operation completes
 * resolve();
 * </code></pre>
 *
 * @since 2.1.0
 * @export
 * @memberof oj.BusyContext
 * @instance
 * @method addBusyState
 * @param {Object} options object that describes the busy state being registered.<br/>
 * @param {Object|function():string} options.description
 *         description: Option additional information of what is registering a busy state. Added to
 *                      logging and handling rejected status. Can be supplied as a Object or a
 *                      function.  If the type is an object the toString function needs to be
 *                      implemented.
 * @returns {function():void} resolve function called by the registrant when the busy state completes.
 *                     The resultant function will throw an error if the busy state is no longer
 *                     registered.
 */
oj.BusyContext.prototype.addBusyState = function (options)
{
  oj.Logger.log("BusyContext.addBusyState: start scope='%s'", this._getDebugScope());

  var statesMap = this._statesMap;

  /** @type {oj.BusyState} */
  var busyState = new oj.BusyState(options[oj.BusyContext._DESCRIPTION]);

  oj.Logger.log(">> " + busyState);

  statesMap.set(busyState.id, busyState);

  this._addBusyStateToParent();

  oj.Logger.log("BusyContext.addBusyState: end scope='%s'", this._getDebugScope());

  return this._removeBusyState.bind(this, busyState);
};

/**
 * Logs all active busy states to the {@link oj.Logger} at {oj.Logger.LEVEL_INFO}.
 * <pre class="prettyprint">
 * <code>
 *  oj.Logger.option("level", oj.Logger.LEVEL_INFO);
 *  oj.Context.getPageContext().getBusyContext().dump("before popup open");
 * </code></pre>
 *
 * @export
 * @since 3.1.0
 * @memberof oj.BusyContext
 * @instance
 * @method dump
 * @param {string=} message optional text used to further denote a debugging point
 * @return {undefined}
 */
oj.BusyContext.prototype.dump = function (message)
{
  oj.Logger.info("BusyContext.dump: start scope='%s' %s", this._getDebugScope(),
                 message ? message : "");

  var statesMap = this._statesMap;
  oj.Logger.info(">> Busy states: %d", statesMap.size);

  var busyStates = oj.BusyContext._values(statesMap);
  if (busyStates.length > 0)
    oj.Logger.info(busyStates.join("\n"));

  oj.Logger.info("BusyContext.dump: start scope='%s' %s", this._getDebugScope(),
                 message ? message : "");
};

/**
 * Returns an array of states representing the active busy states managed by the instance.
 *
 * @export
 * @since 3.1.0
 * @method getBusyStates
 * @memberof oj.BusyContext
 * @instance
 * @return {Array.<{id:string, description:string}>} active busy states managed by the context
 *         instance
 */
oj.BusyContext.prototype.getBusyStates = function ()
{
  var statesMap = this._statesMap;

   /** @type {?} */
  var busyStates = oj.BusyContext._values(statesMap);
  return busyStates;
};

/**
 * Forces all busy states per context instance to release.
 * Use with discretion - last course of action.
 *
 * @since 3.1.0
 * @method clear
 * @memberof oj.BusyContext
 * @instance
 * @export
 * @return {undefined}
 */
oj.BusyContext.prototype.clear = function ()
{
  oj.Logger.log("BusyContext.clear: start scope='%s'", this._getDebugScope());

  var statesMap = this._statesMap;
  var busyStates = oj.BusyContext._values(statesMap);
  for (var i = 0; i < busyStates.length; i++)
  {
    /** @type {?} **/
    var busyState = busyStates[i];
    try
    {
      this._removeBusyState(busyState);
    }
    catch (e)
    {
      oj.Logger.log("BusyContext.clear: %o", e);
    }

    Object.defineProperty(busyState, oj.BusyContext._OJ_RIP, {'value': true, 'enumerable': false});
  }

  oj.Logger.log("BusyContext.clear: end scope='%s'", this._getDebugScope());
};

/**
 * <p>Returns a Promise that will resolve when all registered busy states have completed or a maximum
 * timeout period has elapsed. The promise will be rejected if all the busy states are not resolved
 * within the timeout period. The busyness of the whenReady promsie is evaluated in the next-tick
 * of resolving a busy state.</p>
 *
 * "next-tick" is at the macrotask level. "whenReady" is waiting for the microtask queue to be exhausted,
 * yielding control flow to the user agent, before resolving busyness.
 *
 * @see oj.BusyContext#applicationBootstrapComplete
 * @since 2.1.0
 * @export
 * @memberof oj.BusyContext
 * @instance
 * @method whenReady
 * @param {number=} timeout "optional" maximum period in milliseconds the resultant promise
 *        will wait. Also see {@link oj.BusyContext.setDefaultTimeout}.
 * @returns {Promise.<boolean|Error>}
 */
oj.BusyContext.prototype.whenReady = function (timeout)
{
  var debugScope = this._getDebugScope();

  oj.Logger.log("BusyContext.whenReady: start, scope='%s', timeout=%d", debugScope, timeout);
  /** @type {?} */
  var statesMap = this._statesMap;

  var mediator = this._mediator;
  var nextTickPromise = mediator.getNextTickPromise();
  var bootstrapPromise = oj.BusyContext._BOOTSTRAP_MEDIATOR.whenReady();
  var promise = Promise.all([nextTickPromise, bootstrapPromise]).then(
    function()
    {
      oj.Logger.log("BusyContext.whenReady: bootstrap mediator ready scope=%s", debugScope);

      // Since we are executing this code on 'next tick', it is safe to flush any JET throttled updates.
      // Doing so will allow us to take into account any busy states added in response to the pending updates
      oj.BusyContext._deliverThrottledUpdates();

      if (statesMap.size === 0 && !this._waitingOnNextTickBusynessEval)
      {
        // no busy states, promise resolves immediately
        oj.Logger.log("BusyContext.whenReady: resolved no busy states scope=%s", debugScope);
        return Promise.resolve(true);
      }
      else
      {
        oj.Logger.log("BusyContext.whenReady: busy states returning master scope=%s", debugScope);
        return mediator.getMasterWhenReadyPromise();
      }
    }.bind(this)
  );

  // if a timeout argument is not provided, check the default timeout
  if (isNaN(timeout) && !isNaN(oj.BusyContext._defaultTimeout))
    timeout = oj.BusyContext._defaultTimeout;


  if (!isNaN(timeout))
  {
    var handleTimeout = function()
    {
      var error;
      var expiredText = "whenReady timeout of " + timeout + "ms expired ";

      oj.BusyContext._log(statesMap);
      var busyStates = oj.BusyContext._values(statesMap);

      if (!oj.BusyContext._BOOTSTRAP_MEDIATOR.isReady())
      {
        error = new Error(expiredText + 'while the application is loading.' +
          ' Busy state enabled by setting the "window.oj_whenReady = true;" global variable.' +
          ' Application bootstrap busy state is released by calling' +
          ' "oj.Context.getPageContext().getBusyContext().applicationBootstrapComplete();".');
      }
      else
      {
        error = new Error(expiredText + "with the following busy states: " +
                            busyStates.join(", "));
      }

      error["busyStates"] = busyStates;

      oj.Logger.log("BusyContext.whenReady: rejected scope='%s'\n%s", debugScope, error.message);
      return error;

    };
    promise = mediator.getSlaveTimeoutPromise(promise, handleTimeout, timeout);
  }


  oj.Logger.log("BusyContext.whenReady: end scope='%s'", this._getDebugScope());
  return promise;

};

/**
 * <p>Describes the busyness of the context. The busyness is evaluated in the "next-tick" of a busy
 * state being resolved, meaning the number of busy states doesn't necessarily equate to readiness.
 * The readiness is in sync with the {@link oj.BusyContext#whenReady} resultant promise resolution.</p>
 *
 * @see oj.BusyContext#getBusyStates
 * @since 2.1.0
 * @export
 * @memberof oj.BusyContext
 * @instance
 * @method isReady
 * @returns {boolean} <code>true</code> if the context is not busy
 */
oj.BusyContext.prototype.isReady = function ()
{
  oj.Logger.log("BusyContext.isReady: start scope='%s'", this._getDebugScope());
  var rtn = false;

  if (oj.BusyContext._BOOTSTRAP_MEDIATOR.isReady() && !this._waitingOnNextTickBusynessEval)
  {
    var statesMap = this._statesMap;

    rtn = statesMap.size === 0;
    oj.BusyContext._log(statesMap);
  }

  oj.Logger.log("BusyContext.isReady: end scope='%s'", this._getDebugScope());
  return rtn;
};

/**
 * @private
 * @param {oj.BusyState} busyState added busy state
 * @returns {void}
 * @throws {Error} Busy state has already been resolved
 */
oj.BusyContext.prototype._removeBusyState = function (busyState)
{
  var debugScope = this._getDebugScope();

  oj.Logger.log("BusyContext._removeBusyState: start scope='%s'", debugScope);

  // The BusyState object is passed here instead of just the generated id to provide a more
  // descriptive message when the busy state is removed twice. The description (if provided) of
  // the busy state will be captured in the error message.

  var statesMap = this._statesMap;

  if (busyState[oj.BusyContext._OJ_RIP])
  {
    oj.Logger.log("Busy state has been forcefully resolved via clear:\n" + busyState);
    return;
  }
  else if (!statesMap["delete"](busyState.id)) // quoted to make the closure compiler happy
  {
    throw new Error("Busy state has already been resolved:\n" + busyState);
  }

  // no more busy states; evaluate busyness in the next tick
  if (statesMap.size === 0 && !this._waitingOnNextTickBusynessEval)
  {
    this._waitingOnNextTickBusynessEval = true;
    window.setImmediate(this._evalBusyness.bind(this));
  }
  else
  {
    oj.Logger.log("BusyContext._removeBusyState: resolving busy state:\n" + busyState);
  }

  oj.Logger.log("BusyContext._removeBusyState: end scope='%s'", debugScope);
};

/**
 * Evaluates the busyness of the context.
 * @private
 */
oj.BusyContext.prototype._evalBusyness = function()
{
  var debugScope = this._getDebugScope();

  oj.Logger.log("BusyContext._evalBusyness: begin scope='%s'", debugScope);

  // Since we are executing this code on 'next tick', it is safe to flush any JET throttled updates.
  // Doing so will allow us to take into account any busy states added in response to the pending updates
  oj.BusyContext._deliverThrottledUpdates();

  var statesMap = this._statesMap;
  var mediator = this._mediator;

  if (statesMap.size === 0)
  {
    oj.Logger.log("BusyContext._evalBusyness: resolving whenReady promises");

    mediator.resolveMasterWhenReadyPromise();
    this._resolveBusyStateForParent();
  }
  else
  {
    oj.BusyContext._log(statesMap);
  }

  this._waitingOnNextTickBusynessEval = false;

  oj.Logger.log("BusyContext._evalBusyness: end scope='%s'", debugScope);
};

/**
 * <p>This function should be invoke by application domain logic to indicate all application
 * libraries are loaded and bootstrap processes complete.  The assumed strategy is that the
 * application will set a single global variable "oj_whenReady" from a inline script from the
 * document header section indicating the {@link oj.BusyContext#whenReady}
 * should {@link oj.BusyContext#addBusyState} until the application determines its bootstrap
 * sequence has completed.</p>
 *
 * Inline Script Example:
 * <pre class="prettyprint">
 * <code>
 * &lt;head&gt;
 *   &lt;script type=&quot;text/javascript&quot;&gt;
 *     // The "oj_whenReady" global variable enables a strategy that the busy context whenReady,
 *     // will implicitly add a busy state, until the application calls applicationBootstrapComplete
 *     // on the busy state context.
 *     window["oj_whenReady"] = true;
 *   &lt;/script&gt;
 * ...
 * ...
 * </code></pre>
 *
 * Requirejs callback Example:
 * <pre class="prettyprint">
 * <code>
 * require(['knockout', 'jquery', 'app', 'ojs/ojknockout', 'ojs/ojselectcombobox' ...],
 *   function(ko, $, app)
 *   {
 *     // release the application bootstrap busy state
 *     oj.Context.getPageContext().getBusyContext().applicationBootstrapComplete();
 *     ...
 *     ...
 *   });
 * </code></pre>
 *
 * @since 3.2.0
 * @export
 * @memberof oj.BusyContext
 * @method applicationBootstrapComplete
 * @returns {undefined}
 */
oj.BusyContext.prototype.applicationBootstrapComplete = function ()
{
  var debugScope = this._getDebugScope();
  oj.Logger.log("BusyContext.applicationBootstrapComplete: begin scope='%s'", debugScope);

  oj.BusyContext._BOOTSTRAP_MEDIATOR.notifyComplete();

  oj.Logger.log("BusyContext.applicationBootstrapComplete: end scope='%s'", debugScope);
};

/**
 * @ignore
 * @private
 * @return {oj.BusyContext} returns the nearest parent context
 */
oj.BusyContext.prototype._getParentBusyContext = function()
{
  if (this._hostNode)
  {
    var parentContext =  oj.Context.getContext(oj.Context.getParentElement(this._hostNode));
    if (parentContext)
    {
      return parentContext.getBusyContext();
    }
  }

  return null;
}

/**
 * Links a child context to its parent by registering a busy state with the parent
 * that will recursively register with its parent.
 *
 * @ignore
 * @private
 */
oj.BusyContext.prototype._addBusyStateToParent = function()
{
  if (!this._parentNotified)
  {
    this._parentNotified = true;

    var parentContext = this._getParentBusyContext();
    if (parentContext)
    {
      var opts = {};
      opts[oj.BusyContext._DESCRIPTION] = this._getCompoundDescription.bind(this);
      this._parentResolveCallback = parentContext.addBusyState(opts);
    }
  }
}

/**
 * Resolves the busy state linking a child context with its parent.
 *
 * @ignore
 * @private
 */
oj.BusyContext.prototype._resolveBusyStateForParent = function()
{
  this._parentNotified = false;
  if (this._parentResolveCallback)
  {
    this._parentResolveCallback();
    this._parentResolveCallback = null;
  }
}

/**
 * @private
 * @ignore
 * @return {string} description of all active busy states held by the context.
 */
oj.BusyContext.prototype._getCompoundDescription = function()
{
  var busyStates = oj.BusyContext._values(this._statesMap);
  return ('[' + busyStates.join(", ") + ']');
}

/**
 * @private
 * @ignore
 * @return {string} context debug scope
 */
oj.BusyContext.prototype._getDebugScope = function()
{
  function toSelector(node)
  {
    var selector = "undefined";
    if (node)
    {
      if (node.id && node.id.length > 0)
        selector = "#" + node.id;
      else
      {
        selector = node.nodeName;
        if (node.hasAttribute("data-oj-context"))
          selector += '[data-oj-context]';

        var clazz = node.getAttribute("class");
        if (clazz)
          selector += "." + clazz.split(" ").join(".");
      }
    }

    return selector;
  }

  if (!this._debugScope)
  {
    if (this._hostNode)
      this._debugScope = toSelector(this._hostNode.parentElement) + " > " +
                         toSelector(this._hostNode);
    else
      this._debugScope = "page";
  }

  return this._debugScope;
};

/**
 * @since 3.1.0
 * @override
 * @memberof oj.BusyContext
 * @instance
 * @method toString
 * @returns {string} returns the value of the object as a string
 */
oj.BusyContext.prototype.toString = function ()
{
  var msg = "Busy Context: [scope=";
  msg += this._getDebugScope();
  msg += " states=" + this._getCompoundDescription() + "]"
  return msg;
};

/**
 * @ignore
 * @private
 */
oj.BusyContext._deliverThrottledUpdates = function()
{
  // Dynamically check for the presence of ojs/ojknockout
  if (oj.ComponentBinding)
  {
    oj.ComponentBinding.deliverChanges();
  }
}

/**
 * @private
 * @ignore
 * @const
 * attribute name describing a busystate
 * @type {string}
 */
oj.BusyContext._DESCRIPTION = "description";

/**
 * @ignore
 * @private
 * @constant
 * {@link oj.BusyState} property name indicating the instance is dead
 * @type {string}
 */
oj.BusyContext._OJ_RIP = "__ojRip";

/**
 * @ojtsignore
 * @private
 * @ignore
 */
oj.BusyContext._BOOTSTRAP_MEDIATOR = new /** @constructor */(function()
{
  var _tracking;
  var _readyPromise;
  var _resolveCallback;

  if (typeof window !== 'undefined')
  {
    _tracking = window["oj_whenReady"];
  }

  this.whenReady = function()
  {
    if (_readyPromise)
    {
      return _readyPromise;
    }

    if (!_tracking)
    {
      _readyPromise = Promise.resolve(true);
    }
    else
    {
      _readyPromise = new Promise(
        function(resolve, reject)
        {
          _resolveCallback = resolve;
        }
      );
    }
    return _readyPromise;
  }

  this.isReady = function()
  {
    return !_tracking;
  }

  this.notifyComplete = function()
  {

    if (_resolveCallback)
    {
      // resovle the promise in the next-tick.
      window.setImmediate(function ()
      {
        _tracking = false;
        _resolveCallback(true);
        _resolveCallback = null;
      });
    }
    else
    {
      _tracking = false;
    }
  }

})();

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * <b>The constructor should never be invoked by an application directly</b>. Use
 * {@link oj.Context.getPageContext} and {@link oj.Context.getContext} APIs to
 * retrieve an instance of the context.
 * @param {Element=} node DOM node where the context should be created
 * @export
 * @hideconstructor
 * @constructor oj.Context
 * @since 2.1.0
 * @classdesc This is a general purpose context. Initially it only exposes the BusyContext
 * that keeps track of components that are currently animating or fetching data.
 * In the future this context might be expanded for other purposes.
 */
oj.Context = function (node)
{
  this.Init(node);
};

oj.Object.createSubclass(oj.Context, oj.Object, "oj.Context");

/**
 * @method Init
 * @param {Element=} node DOM node where the context should be created
 * @instance
 * @memberof oj.Context
 * @instance
 * @protected
 */
oj.Context.prototype.Init = function (node)
{
  oj.Context.superclass.Init.call(this);
  this._node = node;
};

/**
 * Returns the closest enclosing JET context for a node.
 * Any DOM element may be designated by the page author as a host of JET context.
 * The designation must be expressed in HTML markup by specifying the "data-oj-context"
 * attribute on the host element:

 * <pre class="prettyprint">
 * &lt;div data-oj-context>&lt;div>
 * </pre>
 *
 * <p>This method will walk up the element hierarchy starting with the source node to
 * find an element that has the data-oj-context attribute. If no such element is found,
 * the page context will be returned.</p>
 *
 * If the JET context is established on a particular element, the {@link oj.BusyContext}
 * associated with that context will be tracking busy states for that element and
 * its subtree
 *
 * @see oj.BusyContext for code examples
 * @method getContext
 * @memberof oj.Context
 * @param {Element} node DOM element whose enclosing context will be provided
 * @return {oj.Context} context object scoped per the target node
 * @since 2.2.0
 * @export
 */
oj.Context.getContext = function(node)
{
  while(node)
  {
    var context = node[oj.Context._OJ_CONTEXT_INSTANCE];
    if (context)
    {
      return context;
    }
    if (node.hasAttribute(oj.Context._OJ_CONTEXT_ATTRIBUTE))
    {
      context = new oj.Context(node);
      Object.defineProperty(node, oj.Context._OJ_CONTEXT_INSTANCE,
                                                        {'value': context});
      return context;
    }

    node = oj.Context.getParentElement(node);
  }

  return oj.Context.getPageContext();
};

/**
 * Static factory method that returns the page context.
 * @see oj.BusyContext for code examples
 * @export
 * @since 2.1.0
 * @method getPageContext
 * @return {oj.Context} context scoped for the page
 * @memberof oj.Context
 */
oj.Context.getPageContext = function ()
{
  if (!oj.Context._pageContext)
    oj.Context._pageContext = new oj.Context();

  return oj.Context._pageContext;
};

/**
 * @see oj.BusyContext for code examples
 * @since 2.1.0
 * @export
 * @method getBusyContext
 * @memberof oj.Context
 * @instance
 * @returns {oj.BusyContext} busy state context
 */
oj.Context.prototype.getBusyContext = function ()
{
  if (!this._busyContext)
    this._busyContext = new oj.BusyContext(this._node);

  return this._busyContext;
};

/**
 * @ignore
 * @private
 * @constant
 * Element marker attribute defining a context
 * @type {string}
 */
oj.Context._OJ_CONTEXT_ATTRIBUTE = "data-oj-context";

/**
 * @ignore
 * @private
 * @constant
 * Element property name for a context
 * @type {string}
 */
oj.Context._OJ_CONTEXT_INSTANCE = "__ojContextInstance";

/**
 * @ignore
 * @private
 * @constant
 * attribute identifying an open popup
 * @type {string}
 */
oj.Context._OJ_SURROGATE_ATTR = "data-oj-surrogate-id";

/**
 * @ignore
 * @public
 * @param {Element} element target
 * @return {Element} the logical parent of an element accounting for open popups
 * @memberof oj.Context
 */
oj.Context.getParentElement = function (element)
{
  // @see oj.ZOrderUtils._SURROGATE_ATTR in "ojpopupcore/PopupService.js" for the details on how
  // this attribute is used by the popup service. The constant was re-declared to simplify module
  // dependencies.

  if (element && element.hasAttribute(oj.Context._OJ_SURROGATE_ATTR))
  {
    var surrogate = document.getElementById(element.getAttribute(oj.Context._OJ_SURROGATE_ATTR));
    if (surrogate)
      return surrogate.parentElement;
  }

  return element.parentElement;
};

(function()
{

  /**
   * @ignore
   * @private
   */
  oj.__AttributeUtils = {};
  
  /**
   * @ignore
   * @return {{expr: (null|string), downstreamOnly: boolean}} 
   * @private
   */
  oj.__AttributeUtils.getExpressionInfo = function(attrValue)
  {
    var info = {};
    if (attrValue)
    {
      var trimmedVal = attrValue.trim();
      var exp = _ATTR_EXP.exec(trimmedVal);
      exp = exp ? exp[1] : null;
      if (!exp)
      {
       info.downstreamOnly = true;
       exp = _ATTR_EXP_RO.exec(trimmedVal);
       exp = exp ? exp[1] : null;
      }
      info.expr = exp;
    }

    return info;
  }
  
  /**
   * @ignore
   * @param {string} attr attribute name
   * @return {string} property name
   * @private
   */
  oj.__AttributeUtils.attributeToPropertyName = function(attr)
  {
    return attr.toLowerCase().replace(/-(.)/g,
      function(match, group1)
      {
        return group1.toUpperCase();
      }
    );
  }
  
  /**
   * @ignore
   * @param {string} name property name
   * @return {string} attribute name
   * @private
   */
  oj.__AttributeUtils.propertyNameToAttribute = function(name)
  {
    return name.replace(/([A-Z])/g,
      function(match)
      {
        return '-' + match.toLowerCase();
      }
    );
  }

  /**
   * @ignore
   * @param {string} type event type (e.g. ojBeforeExpand)
   * @return {string} event listener property name (e.g. onOjBeforeExpand)
   * @private
   */
  oj.__AttributeUtils.eventTypeToEventListenerProperty = function(type)
  {
    return 'on' + type.substr(0,1).toUpperCase() + type.substr(1);
  }

  /**
   * @ignore
   * @param {string} property event listener property name (e.g. onOjBeforeExpand)
   * @return {string|null} event type (e.g. ojBeforeExpand)
   * @private
   */
  oj.__AttributeUtils.eventListenerPropertyToEventType = function(property)
  {
    if (/^on[A-Z]/.test(property)) {
      return property.substr(2,1).toLowerCase() + property.substr(3);
    }
    return null;
  }

  /**
   * @ignore
   * @param {string} name property name (e.g. expanded)
   * @return {string} change event type (e.g. expandedChanged)
   * @private
   */
  oj.__AttributeUtils.propertyNameToChangeEventType = function(name)
  {
    return name + 'Changed';
  }

  /**
   * @ignore
   * @param {string} type change event type (e.g. expandedChanged)
   * @return {string|null} propertyName (e.g. expanded)
   * @private
   */
  oj.__AttributeUtils.changeEventTypeToPropertyName = function(type)
  {
    if (/Changed$/.test(type)) {
      return type.substr(0, type.length - 7);
    }
    return null;
  }

  /**
   * @ignore
   * @param {string} trigger event trigger (e.g. beforeExpand)
   * @return {string} event type (e.g. ojBeforeExpand)
   * @private
   */
  oj.__AttributeUtils.eventTriggerToEventType = function(trigger)
  {
    return 'oj' + trigger.substr(0,1).toUpperCase() + trigger.substr(1);
  }

  /**
   * @ignore
   * @param {string} type event type (e.g. ojBeforeExpand)
   * @return {string|null} event trigger (e.g. beforeExpand)
   * @private
   */
  oj.__AttributeUtils.eventTypeToEventTrigger = function(type)
  {
    if (/^oj[A-Z]/.test(type)) {
      return type.substr(2,1).toLowerCase() + type.substr(3);
    }
    return null;
  }
  
  /**
   * Parses attribute values to the specified metadata type. Throws
   * an error if the value cannot be parsed to the metadata type
   * or if no type was provided.
   * @ignore
   * @param {Element} elem The element whose value we are parsing
   * @param {string} attr attribute
   * @param {string} value attribute value
   * @param {string} type property type
   * @return {*} coerced value
   * @private
   */
  oj.__AttributeUtils.coerceValue = function(elem, attr, value, type)
  {
    if (!type)
    {
      throw "Unable to parse " + attr + "='" + value + "' for " + elem.tagName + " with id " + elem.id + 
      " . This attribute only supports data bound values. Check the API doc for supported types";
    }
    
    // We only support primitive types and JSON objects for coerced properties.
    // Generally, we support parsing of a single type except for Object|string or Array|string cases
    // defined in metadata.
    var typeLwr = type.toLowerCase();
    // The below checks ignore the couble {{}} [[]] cases since expression checking occurs
    // before attribute value coercion
    // Tests to see if the value starts and ends with matched [...] ignoring whitespace
    var isValueArray = _ARRAY_VALUE_EXP.test(value);
    // Tests to see if the value starts and ends with matched {...} ignoring whitespace
    var isValueObj = _OBJ_VALUE_EXP.test(value);

    if ((_ARRAY_TYPE_EXP.test(typeLwr) && isValueArray) ||
        (_OBJ_TYPE_EXP.test(typeLwr) && isValueObj) ||
        (typeLwr === 'any' && (isValueArray || isValueObj)))
    {
      try
      {
        return JSON.parse(value);
      }
      catch (ex) 
      {
        throw "Unable to parse " + attr + "='" + value + "' for " + elem.tagName + " with id " + elem.id +
        " to a JSON Object. Check the value for correct JSON syntax, e.g. double quoted strings. " + ex;
      }
    }
    else if (typeLwr === "boolean")
    {
      // Boolean attributes are considered true if the attribute is:
      // 1) Set to the empty string
      // 2) Present in the DOM without a value assignment
      // 3) Set to the 'true' string
      // 4) Set to the case-insensitive attribute name
      // Boolean values are considered false if set to the false string.
      // An error is thrown for all other values and the attribute value will not be set.
      if (value == null || value === "true" || value === "" || value.toLowerCase() === attr)
        return true;
      else if (value === "false")
        return false;
    }
    else if (typeLwr === "number")
    {
      if (!isNaN(value))
        return Number(value);
    }
    else
    {
      var typeAr = type.split('|');
      // The any type will return a string if not matched as an object or array in first check
      if (typeAr.indexOf('string') !== -1 || typeLwr === 'any')
        return value;
    }

    throw "Unable to parse " + attr + "='" + value + "' for " + elem.tagName + " with id " + elem.id + " to a " + type + ".";
  }
  
  // Checks that a string either starts with 'array' or contains '|array'
  var _ARRAY_TYPE_EXP = /(^array)|(\|array)/;
  // Checks that a string either starts with 'object' or contains '|object'
  var _OBJ_TYPE_EXP = /(^object)|(\|object)/;
  
  var _ARRAY_VALUE_EXP = /\s*\[[^]*\]\s*/;
  var _OBJ_VALUE_EXP = /\s*\{[^]*\}\s*/;

  // Check for {{..}} and [[..]] at the beginning of strings to avoid matching
  // any usages mid string
  var _ATTR_EXP = /^(?:\{\{)([^]+)(?:\}\})$/;
  var _ATTR_EXP_RO = /^(?:\[\[)([^]+)(?:\]\])$/;
  
})();


(function () {  
  if (typeof window === 'undefined') {
    return;
  }
  // polyfill for Element.closest()
  if (window['Element'] && !Element.prototype.closest) {
    Element.prototype.closest =
        function(s) {
      var matches = (this.document || this.ownerDocument).querySelectorAll(s),
              i,
              el = this;
      do {
        i = matches.length;
        while (--i >= 0 && matches.item(i) !== el) {}
      } while ((i < 0) && (el = el.parentElement));
      return el;
    };
  }
})();
/**
 * Utility class with functions for preprocessing HTML data  
 * to avoid a problem using a custom tag within HTML table.
 * Used as a part of ModuleElementUtils.createView() call
 *
 * @constructor
 * @private
 */
oj.__HtmlUtils = (function() {
  
  //replace '<from' tag with '<oj-bind-replace-from' globally in the provided string in order to identify and reverse it later
  var escapeTag = function (from, str) {
    var startTag = new RegExp('<' + from + '(?=\\s|>)', 'gi');
    var endTag = new RegExp('</' + from + '(?=\\s|>)', 'gi');
    return str.replace(startTag, '<' + 'oj-bind-replace-' + from).replace(endTag, '</' + 'oj-bind-replace-' + from);
  }
  
  //replace '<oj-bind-replace-tag' with the original <tag in the array of DOM nodes
  var unescapeTag = function (parent) {
    var children = parent.childNodes;
    var len = children.length;
    for (var i = 0; i < len; i++) {
      var child = children[i];
      unescapeTag(child);
      var nodeName = child.nodeName.toLowerCase();
      if (nodeName.substr(0,16) === 'oj-bind-replace-') {
        var replName = nodeName.substr(16);
        var replNode = document.createElement(replName);
        for (var j = 0; j < child.attributes.length; j++) {
          var attr = child.attributes[j];
          replNode.setAttribute(attr.name, attr.value);
        }
        var childHolder = replNode.content ? replNode.content : replNode;
        for (var j = 0; child.childNodes.length > 0; ) {
          childHolder.appendChild(child.childNodes[0]);
        }
        parent.replaceChild(replNode, child);
      }
      else if(nodeName === "script" || nodeName === "style") {
        var replNode = document.createElement(nodeName);
        for (var j = 0; j < child.attributes.length; j++) {
          var attr = child.attributes[j];
          replNode.setAttribute(attr.name, attr.value);
        }
        var origHTML = child.innerHTML;
        replNode.innerHTML = origHTML.replace(new RegExp('oj-bind-replace-', 'g'), '');
        parent.replaceChild(replNode, child);
      }
      else if(child.nodeType === 8) { //comment node
        var origValue = child.nodeValue;
        child.nodeValue = origValue.replace(new RegExp('oj-bind-replace-', 'g'), '');
      }      
    }
  }

  return {
    stringToNodeArray: function(html) {
      // escape html for the predefined tags
      var tags = ["table","caption","colgroup","col","thead","tfoot","th","tbody","tr","td","template"];
      for (var i = 0; i < tags.length; i++) {
        html = escapeTag(tags[i], html);
      }
      //convert tags into DOM structure
      var container = document.createElement("div");
      container.innerHTML = html; //@HTMLUpdateOK html is the oj-module or composite View which does not come from the end user
      if (html.indexOf("<oj-bind-replace-") !== -1)
        unescapeTag(container);
      
      //convert child nodes to nodes array accepted by oj-module element
      var childList = container.childNodes;
      var nodesArray = [];
      for(var i = childList.length; i--; nodesArray.unshift(childList[i]));
      return nodesArray;
    }
  };
  
})();
/**
 * Timing related utilities
 * @namespace
 * @name oj.TimerUtils
 * @since 4.1.0
 * @ojstatus preview
 */
oj.TimerUtils = {};

/**
 * A Timer encapsulates a Promise associated with a deferred function execution
 * and the ability to cancel the timer before timeout.
 * @interface Timer
 */
 function Timer() {};
 /**
  * Get the Promise assocaited with this timer.  Promise callbacks will be
  * passed a single boolean value indicating if the timer's timeout expired
  * normally (without being canceled/cleared).  If the timer is left to expire
  * after its configured timeout has been exceeded, then it will pass
  * boolean(true) to the callbacks.  If the timer's {@link #clear} method is
  * called before its configured timeout has been reached, then the callbacks
  * will receive boolean(false).
  * @memberof Timer
  * @return {Promise.<boolean>} This timer's Promise
  */
 Timer.prototype.getPromise = function() {};
 /**
  * Clears the timer and resolves the Promise.  If the normal timeout hasn't
  * yet expired, the value passed to the Promise callbacks will be boolean(false).
  * If the timeout has already expired, this function will do nothing, and all of
  * its Promise callbacks will receive boolean(true).
  * @return {void}
  * @memberof Timer
  */
 Timer.prototype.clear = function() {};

/**
 * Get a Timer object with the given timeout in milliseconds.  The Promise
 * associated with the timer is resolved when the timeout window expires, or if
 * the clear() function is called.
 * This is useful for when code needs to be executed on timeout (setTimeout) and
 * must handle cleanup tasks such as clearing {@link BusyState} when the timer
 * expires or is canceled.
 *
 * @param  {number} timeout The timeout value in milliseconds to wait before the
 * promise is resolved.
 * @return {Timer}
 * A Timer object which encapsulates the Promise that will be
 * resolved once the timeout has been exceeded or cleared.
 * @export
 * @memberof oj.TimerUtils
 * @example <caption>Get a timer to execute code on normal timeout and
 * cancelation.  If the timeout occurs normally (not canceled), both
 * callbacks are executed and the value of the 'completed' parameter will be
 * true.</caption>
 * var timer = oj.TimerUtils.getTimer(1000);
 * timer.getPromise().then(function(completed) {
 *     if (completed) {
 *       // Delayed code
 *     }
 *   })
 * timer.getPromise().then(function() {
 *   // Code always to be run
 * })
 *
 * @example <caption>Get a timer to execute code on normal timeout and cancelation.
 * In this example, the timer is canceled before its timeout expires, and the
 * value of the 'completed' parameter will be false.</caption>
 * var timer = oj.TimerUtils.getTimer(1000);
 * timer.getPromise()
 *   .then(function(completed) {
 *     if (completed) {
 *       // Delayed code
 *     }
 *   })
 * timer.getPromise()
 *   .then(function() {
 *     // Code always to be run
 *   })
 * ...
 * timer.clear(); // timer cleared before timeout expires
 */
oj.TimerUtils.getTimer = function(timeout) {
    return new oj.TimerUtils._TimerImpl(timeout);
}

/**
 * @constructor
 * @implements {Timer}
 * @param  {number} timeout The timeout value in milliseconds.
 * @private
 */
oj.TimerUtils._TimerImpl = function(timeout) {
    var _promise;
    var _resolve;
    var _timerId;

    this.getPromise = function() {
        return _promise;
    }
    this.clear = function() {
        window.clearTimeout(_timerId);
        _timerId = null;
        _timerDone(false);
    }

    /**
     * Called on normal and early timeout (cancelation)
     */
    function _timerDone(completed) {
        _timerId = null;
        _resolve(completed);
    }

    if (typeof window === 'undefined') {
        _promise = Promise.reject();
    }
    else {
        _promise = new Promise(function(resolve) {
            _resolve = resolve;
            _timerId = window.setTimeout(_timerDone.bind(null, true), timeout);
        });
    }
};

/**
 * Copyright (c) 2014, 2015 Oracle and/or its affiliates.
 * All rights reserved.
 */
/* jslint browser: true*/
/**
 * @constructor
 * @class oj.Events
 * @classdesc Supports event system for the common model ([oj.Collection]{@link oj.Collection} and
 * [oj.Model]{@link oj.Model})
 * @since 1.0.0
 * @ojtsignore
 */
oj.Events = oj.Events = // eslint-disable-line no-multi-assign
/** @lends oj.Events */
{
    /**
     * Add an event handler for an event type to the model or collection object.
     * @param {string|Object} eventType Types of event handlers to add (may be a single event type, a
     * space-delimited set of event types, or an object mapping events to callbacks).
     * @param {function(string, Object)} callback User's event handler callback function (called with the
     * eventType and model or collection object as parameters--the context will be the model or collection unless specified by context, below).
     * @param {Object=} context A context for the event
     * @return {undefined}
     * @since 1.0.0
     * @memberof oj.Events
     */
  on: function (eventType, callback, context) {
    return this.OnInternal(eventType, callback, context, false, false);
  },

    /**
     * Remove an event handler for an event type from the model or collection object.
     * @param {(string|Object)=} eventType Types of event handlers to remove (may be a single event type, a
     * space-delimited set of event types, or a map of events to callbacks). If omitted, remove all event handlers.
     * @param {function(string, Object)=} callback If provided, remove handlers only for eventType events with the
     * given callback function.
     * @param {Object=} context If provided, remove handlers only for eventType events with the given callback
     * function and context object.
     * @return {undefined}
     * @since 1.0.0
     * @memberof oj.Events
    */
  off: function (eventType, callback, context) {
    return this._offInternal(eventType, callback, context, false);
  },

    /**
     * Fire the given event type(s) for all registered handlers.
     * @param {string} eventType Types of event handlers to fire (may be a single event type or a space-delimited
     * set of event types).
     * @return {undefined}
     * @since 1.0.0
     * @memberof oj.Events
     */
  trigger: function (eventType) { // eslint-disable-line no-unused-vars
    var args = Array.prototype.slice.call(arguments);
                 // Inject a silent setting in there: if this is being called outside we want to fire all relevant
                 // events
    args.unshift(false);
    return oj.Events.TriggerInternal.apply(this, args);
  },

    /**
     * Add an event handler for an event type to the model or collection object, but only fire it once, then remove
     * it from the list of handlers.
     * @param {string} eventType Types of event handlers to add (may be a single event type or a space-delimited
     * set of event types).
     * @param {function(string, Object)} callback User's event handler callback function (called with the
     * eventType and model or collection object as parameters--the context will be the model or collection unless
     * specified by context, below).
     * @param {Object=} context A context for the event
     * @return {undefined}
     * @since 1.0.0
     * @memberof oj.Events
     */
  once: function (eventType, callback, context) {
    return this._onceInternal(eventType, callback, context, false, null);
  },

    /**
     * Add an event handler for an event type to a second model or collection object ("otherObj"), but track it on
     * the called object.
     * @param {Object} otherObj Model or collection object on which to add this event handler.
     * @param {string} eventType Types of event handlers to add (may be a single event type or a space-delimited
     * set of event types).
     * @param {function(string, Object)} callback User's event handler callback function (called with the
     * eventType and model or collection object as parameters--the context will be the model or collection
     * unless specified by context, below).
     * @return {undefined}
     * @since 1.0.0
     * @memberof oj.Events
     */
  listenTo: function (otherObj, eventType, callback) {
    var eventArray;
    var e;
    var event;
    var attr;
    var eventString;
    var listenerObj;
    var eventMap = {};
    var prop;

    if (eventType.constructor === String) {
      // Create a map out of it
      eventMap[eventType] = callback;
    } else {
      eventMap = eventType;
    }

    for (prop in eventMap) { // eslint-disable-line no-restricted-syntax
      if (eventMap.hasOwnProperty(prop)) { // eslint-disable-line no-prototype-builtins
        eventArray = oj.Events._getEvents(prop);
        for (e = 0; e < eventArray.length; e += 1) {
          event = eventArray[e].event;
          attr = eventArray[e].attribute;
          listenerObj = { event: event,
            attribute: attr,
            object: otherObj,
            callback: eventMap[prop] };
          eventString = attr ? event + ':' + attr : event;
          if (this._listeningTo === undefined) {
            this._listeningTo = [];
          }
          this._listeningTo.push(listenerObj);
          // fire
          otherObj.OnInternal(eventString, eventMap[prop], null, true, false);
        }
      }
    }
    return this;
  },

    /**
     * Add an event handler for an event type to a second model or collection object ("otherObj"), but track it on
     * the called object.  Only fire once.
     * @param {Object} otherObj Model or collection object on which to add this event handler.
     * @param {string} eventType Types of event handlers to add (may be a single event type or a space-delimited
     * set of event types).
     * @param {function(string, Object)} callback User's event handler callback function (called with the
     * eventType and model or collection object as parameters--the context will be the model or collection unless
     * specified by context, below).
     * @return {undefined}
     * @since 1.0.0
     * @memberof oj.Events
     */
  listenToOnce: function (otherObj, eventType, callback) {
    var eventArray;
    var e;
    var event;
    var attr;
    var eventString;
    var listenerObj;
    var eventMap = {};
    var prop;

    if (eventType.constructor === String) {
      // Create a map out of it
      eventMap[eventType] = callback;
    } else {
      eventMap = (eventType);
    }

    for (prop in eventMap) { // eslint-disable-line no-restricted-syntax
      if (eventMap.hasOwnProperty(prop)) { // eslint-disable-line no-prototype-builtins
        eventArray = oj.Events._getEvents(prop);
        for (e = 0; e < eventArray.length; e += 1) {
          event = eventArray[e].event;
          attr = eventArray[e].attribute;
          listenerObj = { event: event,
            attribute: attr,
            object: otherObj,
            callback: eventMap[prop] };
          eventString = attr ? event + ':' + attr : event;
          if (this._listeningTo === undefined) {
            this._listeningTo = [];
          }
          this._listeningTo.push(listenerObj);
          // fire
          otherObj._onceInternal(eventString, eventMap[prop], null, true, this);
        }
      }
    }
    return this;
  },

    /**
     * Remove event handlers from a model or collection object. If the arguments are omitted, removes all event
     * handlers from the model or collection.
     * @param {Object=} otherObj If specified, remove event handlers that target otherObj from this model or
     * collection.
     * @param {string=} eventType If specified, remove the event handlers for the given event types from this
     * model or collection
     * @param {function(string, Object)=} callback If specified, remove event handlers that call the given user
     * callback function from this model or collection
     * @return {undefined}
     * @since 1.0.0
     * @memberof oj.Events
     */
  stopListening: function (otherObj, eventType, callback) {
    var eventArray;
    var actualType;
    var eventMap = {};
    var e;
    var oneEvent;
    var oneAttr;
    var event;
    var objEqual;
    var eventEqual;
    var callbackEqual;
    var attrEqual;
    var i;
    var len;
    var cb;
    var prop;

    if (arguments == null || arguments.length <= 1) {
      len = this._listeningTo ? this._listeningTo.length : 0;
      // Remove everything
      for (i = 0; i < len; i++) {
        event = this._listeningTo[i];
        // If we have an "otherObj" argument, make sure that passes muster
        objEqual = otherObj ? otherObj === event.object : true;
        if (objEqual) {
          cb = event.object._offInternal;
          cb.apply(event.object, [event.event, event.callback, event.context, true]);
        }
      }
      this._listeningTo = [];
      return this;
    }

    actualType = eventType;
    // Account for missing otherObj
    if (otherObj && otherObj.constructor === String) {
      actualType = otherObj;
    }

    if (actualType.constructor === String) {
      // Create a map out of it
      eventMap[actualType] = callback;
    } else {
      eventMap = actualType;
    }

    for (prop in eventMap) { // eslint-disable-line no-restricted-syntax
      if (eventMap.hasOwnProperty(prop)) { // eslint-disable-line no-prototype-builtins
        eventArray = oj.Events._getEvents(prop);
        for (e = 0; e < eventArray.length; e += 1) {
          oneEvent = eventArray[e].event;
          oneAttr = eventArray[e].attribute;
          len = this._listeningTo ? this._listeningTo.length : 0;
          for (i = len - 1; i >= 0; i -= 1) {
            event = this._listeningTo[i];
            objEqual = otherObj ? otherObj === event.object : true;
            eventEqual = oneEvent ? oneEvent === event.event : true;
            callbackEqual = callback ? eventMap[prop] === event.callback : true;
            attrEqual = oneAttr ? oneAttr === event.attribute : true;
            if (objEqual && eventEqual && callbackEqual && attrEqual) {
              cb = this._listeningTo[i].object._offInternal;
              cb.apply(this._listeningTo[i].object, [this._listeningTo[i].event,
                this._listeningTo[i].callback,
                this._listeningTo[i].context, true]);
              this._listeningTo.splice(i, 1);
            }
          }
        }
      }
    }
    return this;
  }
};

// Aliases for backward compatibility
oj.Events.bind = oj.Events.on;
oj.Events.unbind = oj.Events.off;

/**
 * @export
 * Event types
 * @enum {string}
 * @memberof oj.Events
 */
oj.Events.EventType = {
        /** Triggered when a model is added to a collection<p>
         *  The event passes these arguments to the handler: <br>
         *  <ul>
         *  <b>model</b>: the model being added to the collection<br>
         *  <b>collection</b>: the collection to which the model has been added<br>
         *  <b>options</b>: any options passed in to the add call that triggered the event
         *  </ul>
         */
  ADD: 'add',
        /** Triggered by a collection during an add call once all models passed in have been added<p>
         * The event passes these arguments to the handler:<br>
         * <ul>
         * <b>collection</b>: the collection to which the models have been added<br>
         * <b>models</b>: the array of models that have been added <br>
         * <b>options</b>: any options passed in to the add call
         * </ul>
         */
  ALLADDED: 'alladded',
        /** Triggered when a model is removed from a collection<p>
         * The event passes these arguments to the handler: <br>
         * <ul>
         * <b>model</b>: the model being removed from the collection<br>
         * <b>collection</b>: the collection from which the model was removed<br>
         * <b>options</b>: <b>index</b>: the index of the model being removed
         * </ul>
         */
  REMOVE: 'remove',
        /** Triggered when a collection is reset (see oj.Collection.reset)<p>
         *  The event passes these arguments to the handler:<br>
         *  <ul>
         *  <b>collection</b>: the collection being reset<br>
         *  <b>options</b>: any options passed in to the reset call
         *  </ul>
         */
  RESET: 'reset',
        /** Triggered when a collection is refreshed (see oj.Collection.refresh)<p>
         *  The event passes these arguments to the handler: <br>
         *  <ul>
         *  <b>collection</b>: the collection being refreshed<br>
         *  <b>options</b>: any options passed in to the refresh call
         *  </ul>
         */
  REFRESH: 'refresh',
        /** Triggered when a collection is sorted.  If the second argument to the callback is set (options) and
         * 'add' is true, it means this sort event was triggered as a result of an add <p>
         *  The event passes these arguments to the handler:<br>
         *  <ul>
         *  <b>collection</b>: the collection being sorted<br>
         *  <b>options</b>: <b>add</b>: true if this sort event was triggered as the result of an add call,
         *  undefined or false if not
         *  </ul>
         */
  SORT: 'sort',
        /** Triggered when a model's attributes are changed.  This can be the result of a clear call on a model;
         * a property set call on a model; an unset call on a model; or the changing of properties due to the
         * merging of models (in an add, for example) <p>
         * The event passes these arguments to the handler:<br>
         * <ul>
         * <b>model</b>: the model on which the change occurred<br>
         * <b>value</b>: for property-specific change events, the new value of the property being changed<br>
         * <b>options</b>: any options passed in to the call that triggered the change event.  This is the second
         * argument passed for overall change events, and the third parameter (after value) for property-specific
         * change events.
         * </ul>
         */
  CHANGE: 'change',
        /** Triggered when a model is deleted from the data service (and thus from its Collection), due to a model
         * destroy call<p>
         * The event passes these arguments to the handler:<br>
         * <ul>
         * <b>model</b>: the model being deleted<br>
         * <b>collection</b>: the deleted model's collection, if any
         * </ul>
         */
  DESTROY: 'destroy',
        /** Triggered by a collection during a remove call once all models passed in have been removed and
         * destroyed<p>
         * The event passes these arguments to the handler:<br>
         * <ul>
         * <b>collection</b>: the collection from which the models have been removed<br>
         * <b>models</b>: the array of models that have been removed <br>
         * <b>options</b>: any options passed in to the remove call
         * </ul>
         */
  ALLREMOVED: 'allremoved',
        /** Triggered when a model or collection has sent a request to the data service <p>
         *  The event passes these arguments to the handler:<br>
         *  <ul>
         *  <b>collection or model</b>: the collection or model triggering the request<br>
         *  <b>xhr</b>: the xhr argument for the request<br>
         *  <b>options</b>: any options passed as part of the request
         *  </ul>
         */
  REQUEST: 'request',
        /** Triggered when a model or collection has been updated from the data service<p>
         *  The event passes these arguments to the handler:<br>
         *  <ul>
         *  <b>collection or model</b>: the collection or model that triggered the update<br>
         *  <b>response</b>: the response object from the data service<br>
         *  <b>options</b>: any options passed in to the call that triggered the update
         *  </ul>
         */
  SYNC: 'sync',
        /** Triggered when a model has failed to update on the data service<p>
         *  The event passes these arguments to the handler:<br>
         *  <b>collection or model</b>: the collection or model that made the call that resulted in the error<br>
         *  <b>xhr</b>: the xhr argument for the failing request, if any<br>
         *  <b>options</b>: any options passed in to the call that triggered the failing request, plus the status
         *  and error as textStatus and errorThrown
         *  </ul>
         */
  ERROR: 'error',
        /** Triggered on an error with data source interactions <p>
         * The event passes these arguments to the handler:<br>
         * <ul>
         * <b>model</b>: the model (or collection) on which the error operation happened <br>
         * <b>xhr</b>: the xhr involved, if relevant<br>
         * <b>options</b>: any options passed in to the call that triggered the invalid event
         * </ul>
         */
  INVALID: 'invalid',
        /** Triggered when all pending promises from Collection API calls have been resolved<p>
         * The event passes these arguments to the handler:<br>
         * <ul>
         * <b>collection</b>: the collection on which the promises have been resolved
         * </ul>
         */
  READY: 'ready',
        /** Triggered for any of the above events <p>
         * The event passes the name of the actual event and then any arguments normally passed to that event
         * following the name
         */
  ALL: 'all'
};


/**
 * @private
 * @param {Object} myClass
 * @param {Object=} source
 */
oj.Events.Mixin = function (myClass, source) {
  var methodName;
  var obj = source || this;
  for (methodName in obj) { // eslint-disable-line no-restricted-syntax
    if (typeof obj[methodName] === 'function') {
      myClass[methodName] = obj[methodName]; // eslint-disable-line no-param-reassign
    }
  }
    // Make sure actual vars are own copies
  myClass.eventHandlers = {}; // eslint-disable-line no-param-reassign
  myClass._listeningTo = []; // eslint-disable-line no-param-reassign
};

/**
 * @private
 */
oj.Events._onceInternal = function (eventType, callback, context, listenTo, otherObj) {
  var eventArray;
  var e;
  var event;
  var attr;
  var eventMap;
  var obj;
  var cxt = context;

  obj = this._getEventMap(eventType, callback, context);
  eventMap = obj.map;
  cxt = obj.context;

  var self = this;
  Object.keys(eventMap || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(eventMap, prop)) {
      eventArray = self._getEvents(prop);

      for (e = 0; e < eventArray.length; e += 1) {
        event = eventArray[e].event;
        attr = eventArray[e].attribute;
        if (self.eventHandlers === undefined) {
          self.eventHandlers = [];
        }
        if (self.eventHandlers[event] === undefined) {
          self.eventHandlers[event] = [];
        }

        self.eventHandlers[event].push({ callback: eventMap[prop],
          context: cxt,
          attribute: attr,
          once: true,
          fired: false,
          listen: listenTo,
          otherObj: otherObj });
      }
    }
  });
  return this;
};

/**
 * @private
 */
oj.Events._shouldFire = function (handler) {
  if (handler.once) {
    if (!handler.fired) {
      handler.fired = true; // eslint-disable-line no-param-reassign
      return true;
    }
    return false;
  }
  return true;
};

/**
 * @private
 */
oj.Events._getContext = function (obj, handler) {
  return handler.context || handler.otherObj || obj;
};

/**
 * @protected
 */
oj.Events.TriggerInternal = function (silent, eventType) {
  var eventArray = this._getEvents(eventType);
  var e;
  var event;
  var attr;
  var eventsToFire;
  var handlers;
  var i;
  var args;
  var allHandlers;
  var callback;

  eventsToFire = [];
  for (e = 0; e < eventArray.length; e += 1) {
    event = eventArray[e].event;
    attr = eventArray[e].attribute;
    // Do specific event...
    eventsToFire.push({ event: event, attribute: attr });
  }
  for (e = 0; e < eventsToFire.length; e += 1) {
    allHandlers = this._getHandlers(this.eventHandlers, oj.Events.EventType.ALL);
    handlers = oj.Events._getHandlers(this.eventHandlers, eventsToFire[e].event, false);
    for (i = 0; i < (handlers ? handlers.length : 0); i += 1) {
      if (handlers[i].attribute === eventsToFire[e].attribute && handlers[i].callback) {
        args = Array.prototype.slice.call(arguments);
        if (handlers && handlers[i] && handlers[i].once) {
          // Remove it: only want to fire once--make sure we remove it from the original
          this._removeHandler(oj.Events._getHandlers(this.eventHandlers, eventsToFire[e].event,
            true), handlers[i]);
          // Now take it out of the other object's "listen to" list, if relevant
          if (handlers[i].otherObj) {
            // Clean up the "other object" if this was a triggered listenOnce
            handlers[i].otherObj.stopListening(this, eventType, handlers[i].callback);
          }
        }
        if (handlers && handlers[i] && this._shouldFire(handlers[i])) {
          callback = handlers[i].callback;
          // If this isn't a silent firing or this handler always wants to be called, make the call
          if (!silent || handlers[i].ignoreSilent) {
            callback.apply(oj.Events._getContext(this, handlers[i]), args.slice(2));
          }
        }
      }
    }
    // Handle all
    for (i = 0; i < (allHandlers ? allHandlers.length : 0); i += 1) {
      args = Array.prototype.slice.call(arguments);
      if (args.length > 0) {
        if (eventsToFire[e].attribute) {
          args[1] = eventsToFire[e].event + ':' + eventsToFire[e].attribute;
        } else {
          args[1] = eventsToFire[e].event;
        }
      }
      // All case--make sure to pass event name
      if (allHandlers && allHandlers[i] && allHandlers[i].callback &&
        this._shouldFire(allHandlers[i])) {
        callback = allHandlers[i].callback;
        // If this isn't a silent firing or this handler always wants to be called, make the call
        if (!silent || allHandlers[i].ignoreSilent) {
          callback.apply(oj.Events._getContext(this, allHandlers[i]), args.slice(1));
        }
      }
      if (allHandlers && allHandlers[i] && allHandlers[i].once) {
        // Remove it: only want to fire once
        this._removeHandler(this._getHandlers(this.eventHandlers, oj.Events.EventType.ALL, true),
                                       allHandlers[i]);
        // Now take it out of the other object's "listen to" list, if relevant
        if (allHandlers[i].otherObj) {
          // Clean up the "other object" if this was a triggered listenOnce
          allHandlers[i].otherObj.stopListening(this, oj.Events.EventType.ALL,
                                                                 allHandlers[i].callback);
        }
      }
    }
  }
  return this;
};

/**
 * @protected
 */
oj.Events.OnInternal = function (eventType, callback, context, listenTo, ignoreSilent) {
  var eventMap;
  var eventArray;
  var i;
  var event;
  var attr;
  var eventObj;
  var cxt = context;
  var prop;

  var obj = this._getEventMap(eventType, callback, context);
  eventMap = obj.map;
  cxt = obj.context;

  for (prop in eventMap) { // eslint-disable-line no-restricted-syntax
    if (eventMap.hasOwnProperty(prop)) { // eslint-disable-line no-prototype-builtins
      eventArray = this._getEvents(prop);

      for (i = 0; i < eventArray.length; i += 1) {
        event = eventArray[i].event;
        attr = eventArray[i].attribute;
        if (this.eventHandlers === undefined) {
          this.eventHandlers = [];
        }
        if (this.eventHandlers[event] === undefined) {
          this.eventHandlers[event] = [];
        }

        eventObj = { callback: eventMap[prop],
          context: cxt,
          attribute: attr,
          listen: listenTo,
          ignoreSilent: ignoreSilent };
        if (this._checkForHandler(this.eventHandlers[event], eventObj,
          oj.Events._handlersIdentical) === -1) {
          this.eventHandlers[event].push(eventObj);
        }
      }
    }
  }
  return this;
};

/**
 * @private
 */
oj.Events._offInternal = function (eventType, callback, context, listen) {
  var eventMap;
  var obj;
  var cxt = context;
  var prop;

  if (arguments === null || arguments.length === 0) {
         // Remove everything
    this.eventHandlers = {};
    return this;
  }

  if (eventType == null) {
    this._removeEvent(eventType, callback, context, listen);
    return this;
  }

  obj = this._getEventMap(eventType, callback, context);
  eventMap = obj.map;
  cxt = obj.context;

  for (prop in eventMap) { // eslint-disable-line no-restricted-syntax
    if (eventMap.hasOwnProperty(prop)) { // eslint-disable-line no-prototype-builtins
      this._removeEvent(prop, eventMap[prop], cxt, listen);
    }
  }
  return this;
};


/**
 * @private
 */
oj.Events._getEventMap = function (eventType, callback, context) {
  var eventMap = {};

  if (eventType.constructor === String) {
    // Create a map out of it
    eventMap[eventType] = callback;
  } else {
    eventMap = eventType;
    // If eventType is a map of events->callbacks, then the callback argument is now context
    return { map: eventMap, context: callback };
  }
  return { map: eventMap, context: context };
};

/**
 * @private
 */
oj.Events._removeEvent = function (eventType, callback, context, listen) {
  var eventArray = [];
  var e;
  var evt;
  var i;
  var attr;
  var handlers;
  var callbacks;
  var contexts;
  var attrs;
  var listenEq;

  if (eventType) {
    eventArray = oj.Events._getEvents(eventType);
  } else if (this.eventHandlers !== undefined) {
    // Walk entire eventHandlers property list
    var self = this;
    Object.keys(this.eventHandlers || {}).forEach(function (event) {
      if (Object.prototype.hasOwnProperty.call(self.eventHandlers, event)) {
        eventArray.push({ event: event });
      }
    });
  }

  for (e = 0; e < eventArray.length; e += 1) {
    evt = eventArray[e].event;
    attr = eventArray[e].attribute;
    if (this.eventHandlers !== undefined && this.eventHandlers[evt] instanceof Array) {
      handlers = this.eventHandlers[evt];
      for (i = handlers.length - 1; i >= 0; i -= 1) {
        callbacks = (callback === undefined || callback === null || handlers[i].callback ===
          callback);

        contexts = (context === undefined || context === null || handlers[i].context === context);
        attrs = (attr === undefined || attr === null || handlers[i].attribute === attr);
        listenEq = (listen === undefined || listen === null || handlers[i].listen === listen);
        if (callbacks && contexts && attrs && listenEq) {
          handlers.splice(i, 1);
        }
      }
      if (handlers.length === 0) {
        // Delete the entry
        delete this.eventHandlers[evt];
      }
    }
  }
};

/**
 * @private
 */
oj.Events._removeHandler = function (handlers, handler) {
  var i;
  var callbacks;
  var contexts;
  var attrs;
  var listenEq;
  var onceEq;

  if (handlers) {
    for (i = handlers.length - 1; i >= 0; i -= 1) {
      callbacks = (handler.callback === undefined || handler.callback === null ||
                         handlers[i].callback === handler.callback);

      contexts = (handler.context === undefined || handler.context === null ||
                        handlers[i].context === handler.context);
      attrs = (handler.attribute === undefined || handler.attribute === null ||
                     handlers[i].attribute === handler.attribute);
      listenEq = (handler.listen === undefined || handler.listen === null ||
                        handlers[i].listen === handler.listen);
      onceEq = (handler.once === undefined || handler.once === null ||
        handlers[i].once === handler.once);
      if (callbacks && contexts && attrs && listenEq && onceEq) {
        handlers.splice(i, 1);
      }
    }
  }
};

/**
 * @private
 */
oj.Events._getEvents = function (eventString) {
  var eventList = eventString ? eventString.split(' ') : [];
  var retList = [];
  var i;
  var eventWithAttr;
  var name;
  var attr;

  for (i = 0; i < eventList.length; i += 1) {
    eventWithAttr = eventList[i].split(':');
    name = eventWithAttr[0];
    attr = eventWithAttr.length > 1 ? eventWithAttr[1] : null;
    retList.push({ event: name, attribute: attr });
  }
  return retList;
};

/**
 * @private
 */
oj.Events._handlersIdentical = function (handler1, handler2) {
  return (handler1.callback === handler2.callback) && (handler1.attribute === handler2.attribute) &&
          (handler1.context === handler2.context) && (handler1.listen === handler2.listen) &&
          (handler1.once === handler2.once);
};

/**
 * @private
 */
oj.Events._listenersIdentical = function (listener1, listener2) {
  return (listener1.event === listener2.event) && (listener1.attribute === listener2.attribute) &&
          (listener1.context === listener2.context) && (listener1.object === listener2.object);
};

/**
 * @private
 */
oj.Events._checkForHandler = function (handlerList, handler, handlerTest) {
  var i;
  if (handlerList === undefined) {
    return -1;
  }

  for (i = 0; i < handlerList.length; i += 1) {
    if (handlerTest(handlerList[i], handler)) {
      return i;
    }
  }
  return -1;
};

/**
 * @private
 */
oj.Events._getHandlers = function (handlers, eventType, original) {
  if (handlers && handlers[eventType] instanceof Array) {
    if (original) {
      return handlers[eventType];
    }
    // Make a copy
    var handlerReturn = [];
    var i;
    for (i = 0; i < handlers[eventType].length; i++) {
      handlerReturn.push(handlers[eventType][i]);
    }
    return handlerReturn;
  }
  return null;
};

/**
 * Copyright (c) 2014, 2015 Oracle and/or its affiliates.
 * All rights reserved.
 */
/* jslint browser: true*/
/* global Promise:false */

/**
 * @export
 * @class oj.Collection
 * @classdesc Collection of Model objects
 *
 * @param {Array.<oj.Model>=} models Set of model objects to put into collection at construction
 *                 time.  If models contain actual
 *                 oj.Model objects, then any custom parse callback set on the collection must
 *                 be able to handle oj.Model objects as a possible argument
 * @param {Object=} options Passed through to the user's initialize routine, if any, upon
 *                  construction
 * @constructor
 * @since 1.0
 * @mixes oj.Events
 * @ojsignature {target: "Type", value: "class Collection"}
 */
oj.Collection = function (models, options) {
  if (oj.Collection._justExtending) {
    return;
  }

    // Initialize
  oj.Collection._init(this, models, options, null);
};


/**
 * Subclass from oj.Object
 * @private
 */
oj.Object.createSubclass(oj.Collection, oj.Object, 'oj.Collection');

/**
 * @desc Property specifying the [model]{@link oj.Model} class object used by the collection
 * @memberof oj.Collection
 * @type Object
 * @export
 */
oj.Collection.prototype.model = null;


/**
 * @desc Function specifying how to construct the id for models in this collection.  Override to
 * change the construction of the id.
 * @memberof oj.Collection
 * @param {Object} attrs attributes of a model
 * @returns {null|string}
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.modelId = function (attrs) {
  var model = this.model;
  if (model && attrs) {
    return attrs[model.idAttribute || model.prototype.idAttribute || 'id'];
  }
  return null;
};

/**
 * @export
 * @desc Total number of models in the collection.  When the collection is virtual, not all of the
 * models may be locally available.
 * @memberof oj.Collection
 *
 * @type number
 * @since 1.0.0
 */
oj.Collection.prototype.length = undefined;

/**
 * @export
 * @desc Direct access to the collection's list of models objects<br/>
 * Note that this property should not be used directly when a collection is virtual, as
 * automatic fetches will not be triggered for undefined elements in the model.  Use at()
 * instead.
 * @memberof oj.Collection
 *
 * @type {Array.<oj.Model>}
 * @since 1.0.0
 */
oj.Collection.prototype.models = undefined;

/**
 * Tracking indices used
 * @private
 */
oj.Collection.prototype._modelIndices = [];

/**
 * @export
 * @desc The data service's server URL.
 * @memberof oj.Collection
 *
 * @type {null|string|function():string}
 * @since 1.0.0
 */
oj.Collection.prototype.url = null;

/**
 * @export
 * @desc Changes that have occured due to adds/removes since the last fetch.  This is a list of
 * indicies that have changed (location at which a model was added, deleted, or set).  They do
 * not shift with subsequent operations
 * @memberof oj.Collection
 * @type {Array.<number>}
 * @since 1.0.0
 */
oj.Collection.prototype.changes = [];


/**
 * A callback to allow users to customize the data service URLs.  The callback should accept
 * these parameters:<p>
 * <b>operation</b>: one of "create", "read", "update", "patch", or "delete", indicating the
 * type of operation for which to return the URL<p>
 * <b>collection</b>: the oj.Collection object requesting the URL<p>
 * <b>options</b>: any of the following properties:<br>
 * <ul>
 * <b>recordID</b>: id of the record involved, if relevant<br>
 * <b>fetchSize</b>: how many records to return.  If not set, return all.<br>
 * <b>startIndex</b>: Starting record number of the set to return.<br>
 * <b>startID</b>: Retrieve records starting with the record with the given unique ID.<br>
 * <b>since</b>: Retrieve records with timestamps after the given timestamp.<br>
 * <b>until</b>: Retrieve records with timestamps up to the given timestamp. Default is "until"<br>
 * <b>sort</b>:  field(s) by which to sort, if set<br>
 * <b>sortDir</b>: sort ascending or descending (asc/dsc)<br>
 * <b>query</b>: a set of attributes indicating filtering that should be done on the server.  See
 * [where]{@link oj.Collection#where} for complete documentation of query values<br>
 * <b>all</b>: true (along with 'query', above) indicates that this is a findWhere or where type
 * call that is expecting all models meeting the query condition to be returned<br>
 * </ul>
 * <p>
 * customURL callbacks should return either: null, in which case the default will be used; a
 * url string, which will be used with the standard HTTP method for the type of operation, or
 * an Object with any other attributes that should be passed to the ajax call.<br>
 * This object must at minimum include the URL, and other attributes as follows:<br>
 * <ul>
 * <b>url</b>: the custom URL string<br>
 * <b>type</b>: (optional) a string indicating the type of HTTP method to use (GET, POST,
 *              DELETE, etc.)<br>
 * <b>(other)</b>: (optional) any other attributes to pass in the ajax call<br>
 * </ul>
 * <p>
 * @type {function(string,oj.Collection,Object):(string|Object|null)|null}
 * @memberof oj.Collection
 * @export
 * @since 1.0.0
 */
oj.Collection.prototype.customURL = null;

/**
 * A callback function allowing users to extract their own paging/virtualization
 * return values from the server's response.<p>
 * It should accept these parameters:<p>
 * <b>response</b>: the Object data response coming back from the fetch call<p>
 * <p>
 * The callback should return either null, in which case the collection will look for the
 * default properties, or an object containing one or more of the following attribute/value
 * pairs (note that the Collection will look back to the response for default paging return
 * properties if not found in the returned object):<br>
 * <ul>
 * <b>totalResults</b>: the total number of records available on the server side, not just
 * in the current result.  By default the collection looks in the response for "totalResults"<br>
 * <b>limit</b>: the actual fetchSize from the server.  This may not be the client's fetchSize
 * or the number of records in the current result.  By default the collection looks in the
 * response for "limit".  This becomes the collection's "lastFetchSize" property<br>
 * <b>count</b>: the actual number of records returned by the server in the last result.  This
 * becomes the collection's "lastFetchCount".  By default the collection looks in the response
 * for "count".<br>
 * <b>offset</b>: the actual starting record number of the current result.  By default the
 * collection looks in the response for "offset"<br>
 * <b>hasMore</b>: boolean indicating whether or not there are more records available beyond
 * the current result.  By default the collection looks in the response for "hasMore"<br>
 * </ul>
 * <p>
 *
 * @type {(function(Object):(Object|null)|null)}
 * @memberof oj.Collection
 * @ojsignature {target: "Type",
 * value: "((response: object)=> oj.Collection.CustomPagingOptionsReturn|null)|null"}
 * @export
 * @since 1.0.0
 */
oj.Collection.prototype.customPagingOptions = null;

/**
 * @typedef {Object} oj.Collection.CustomPagingOptionsReturn
 * @property {number=} totalResults
 * @property {number=} limit
 * @property {number=} count
 * @property {number=} offset
 * @property {boolean=} hasMore
*/

/**
 * @export
 * @desc The server's fetch size.  This may not match [fetchSize]{@link oj.Collection#fetchSize}.
 * @memberof oj.Collection
 *
 * @type number
 * @since 1.0.0
 */
oj.Collection.prototype.lastFetchSize = undefined;

/**
 * @export
 * @desc Indicates whether or not there are more records available on the server, at indices
 * beyond the last fetch.
 * @memberof oj.Collection
 *
 * @type boolean
 * @since 1.0.0
 */
oj.Collection.prototype.hasMore = false;

/**
 * @export
 * @desc The total number of records available for this collection regardless of whether they
 * have been fetched or not.  For non-virtual collections this will equal the length.
 * @memberof oj.Collection
 *
 * @type number
 */
oj.Collection.prototype.totalResults = undefined;

/**
 *
 * @export
 * @desc The number of records actually fetched the last time the collection fetched from the
 * server.  This may or may not match [fetchSize]{@link oj.Collection#fetchSize} or
 * [lastFetchSize]{@link oj.Collection#lastFetchSize}
 * @memberof oj.Collection
 *
 * @type number
 */
oj.Collection.prototype.lastFetchCount = undefined;

/**
 * @export
 * @desc For virtual collections, the number of records to be kept in memory at any one
 * time.  The default of -1 indicates that no records are thrown out
 * @memberof oj.Collection
 *
 * @type number
 */
oj.Collection.prototype.modelLimit = -1;

/**
 * @export
 * @desc The actual starting index number at which models from the last server fetch
 * were inserted into the collection.
 * @memberof oj.Collection
 *
 * @type number
 */
oj.Collection.prototype.offset = undefined;

/**
 * @export
 * @desc The number of records to be fetched from the server in any one round trip.
 * The server's fetch size comes back as the "limit" property.  The default value
 * of -1 indicates that virtualization is not being used or is not available,
 * and all records will be fetched.<br>
 * The number of records actually fetched comes back as [count]{@link oj.Collection#count}<br>
 * @memberof oj.Collection
 *
 * @type number
 */
oj.Collection.prototype.fetchSize = -1;

/**
 * @export
 * @desc Sort direction for string-based sort comparators (model attribute names).  A value
 * of 1 indicates ascending sorts, -1 indicates descending.  The default is 1 (ascending).<br>
 * Users should call sort() after changing sort direction to ensure that models in the
 * collection are sorted correctly, or, for virtual collections, that there are no left
 * over models in an incorrect order.
 * @memberof oj.Collection
 *
 * @type number
 */
oj.Collection.prototype.sortDirection = 1;

/**
 * @export
 * @desc If set to a string, sort the collection using the given attribute of a model.<p>
 * If set to a function(Model), the function should return a string giving the model
 * attribute by which the sort should take place.<p>
 * If set to a function(Model1, Model2), then this function is called comparing Model1
 * and Model2 (see the JavaScript array.sort() for details)<p>
 * In the virtual case, comparator must be a string-based field comparator, which will
 * be passed to the server.<p>
 * Users should call sort() after making any changes to the comparator to ensure that
 * the models are correctly sorted, or that there are no leftover models sorted incorrectly
 * in the virtual case.
 * @memberof oj.Collection
 *
 * @type {null|string|function(oj.Model,oj.Model=):number}
 * @since 1.0.0
 */
oj.Collection.prototype.comparator = null;

/**
 * @export
 * @ojstatus preview
 * @memberof oj.Collection
 * @type {boolean}
 * @desc If true, do not insert the JET locale-based Accept-Language header.  If false,
 * let the Ajax system set the header.
 * @since 5.0.0
 */
oj.Collection.prototype.omitLanguageHeader = false;

/**
 * @private
 */
oj.Collection.prototype.Init = function () {
  oj.Collection.superclass.Init.call(this);
};

/**
 * Create a new, specific type of Collection object to represent a collection of records
 * from a JSON data set.
 * @param {Object=} properties Properties for the new Collection class.<p>
 * <b>parse</b>: a user callback function to allow parsing of the JSON collection after it's
 * returned from the data service.  If a collection is initialized with actual oj.Models or
 * collection.set is used with actual oj.Models, the parse callback must expect that the argument
 * passed to it may contain raw JSON data *or* oj.Model objects<br>
 * <b>model</b>: the specific type of [model]{@link oj.Model} object to use for each member of
 * the collection<br>
 * <b>url</b>: the URL string to use to get records from the data service<br>
 * <b>initialize</b>: a user callback function to be called when this collection is created.
 * Called in the context of the collection and passed: models, options.<br>
 * <b>comparator</b>: a user callback used on sort calls. May also be set to false to prevent
 * sorting.  See [comparator]{@link oj.Collection#comparator}<br>
 * <b>fetchSize</b>: the number of records to be fetched on each round trip to the server.
 * Overrides [fetchSize]{oj.Collection#fetchSize}. If not set, the collection will not consider
 * itself virtual<br>
 * <b>modelLimit</b>: the number of records to be held in memory at any one time, if
 * virtualization is in force.  The default is all records.  This uses an LRU algorithm to
 * determine which to roll off as more records are added.<br>
 * @param {Object=} classProperties optional properties that get attached to the constructor
 * of the extended Collection
 * @return {oj.Collection} new Collection object
 * @ojsignature [{target: "Type",
 *                value: "any",
 *                for: "returns"},
 * {target: "Type", value:"{parse?: (data: any)=> any, model?: oj.Model, url?: string,
 * initialize?: (models: Array<oj.Model>, options: Object)=> void,
 * comparator?: null|string|((model1: oj.Model, model2?: oj.Model)=> number),
 * fetchSize?: number, modelLimit?: number, [propName: string]: any}", for: "properties"}]
 * @memberof oj.Collection
 * @this {oj.Collection}
 * @since 1.0.0
 * @export
 */
oj.Collection.extend = function (properties, classProperties) {
  var obj = null;
  oj.Collection._justExtending = true;
  obj = new oj.Collection();
  oj.Collection._justExtending = false;

    // Add regular properties from this "parent"
    // oj.Events.Mixin(obj, this.prototype);
  $.extend(obj, this.prototype);

  var Collection;
  if (properties && properties.constructor && Object.prototype.hasOwnProperty.call(properties, 'constructor')) {
    Collection = properties.constructor;
  } else {
    Collection = function (models, options) {
      oj.Collection._init(this, models, options, properties);
    };
  }

  if (classProperties) {
    Object.keys(classProperties).forEach(function (prop) {
      if (Object.prototype.hasOwnProperty.call(classProperties, prop)) {
        Collection[prop] = classProperties[prop];
      }
    });
  }

  if (properties) {
    Object.keys(properties).forEach(function (prop) {
      if (Object.prototype.hasOwnProperty.call(properties, prop)) {
        obj[prop] = properties[prop];
      }
    });
  }

  $.extend(Collection, this);
  Collection.prototype = obj;

    // Allow extending resulting obj
  Collection.extend = oj.Collection.extend;

  Collection.prototype.constructor = Collection;

  return Collection;
};


/**
 * @private
 */
oj.Collection._init = function (collection, models, options, properties) {
  var i;
  var optionlist;
  var modelList;
  var prop;

  collection.Init();

    // Augment with Event
  oj.Events.Mixin(collection);

    // First, copy all properties passed in
  if (properties) {
    for (prop in properties) { // eslint-disable-line no-restricted-syntax
      if (Object.prototype.hasOwnProperty.call(properties, prop)) {
        collection[prop] = properties[prop];  // eslint-disable-line no-param-reassign
      }
    }
  }

    // Check options
  var opt = options || {};
  optionlist = ['comparator', 'customPagingOptions', 'customURL',
    oj.Collection._FETCH_SIZE_PROP, 'model', 'modelLimit', 'sortDirection', 'url'];
  for (i = 0; i < optionlist.length; i++) {
    if (Object.prototype.hasOwnProperty.call(opt, optionlist[i]) &&
        opt[optionlist[i]] !== undefined) {
      collection[optionlist[i]] = opt[optionlist[i]]; // eslint-disable-line no-param-reassign
    }
  }
  if (collection._getFetchSize(null) === undefined) {
    collection.setFetchSize(-1);
  }
  if (collection.modelLimit === undefined) {
    collection.setModelLimit(-1);
  }
  collection.hasMore = false; // eslint-disable-line no-param-reassign
  collection.lruCount = 0; // eslint-disable-line no-param-reassign

  collection._setModels([], true);
  var localModels = models;
  if (opt.parse) {
    localModels = collection.parse(models);
  }
  if (localModels != null) {
    opt.noparse = true;
    modelList = (localModels instanceof Array) ? localModels : [localModels];
    collection._addInternal(modelList, opt, true, false);
  }
  collection._setLength();
  if (!localModels) {
        // Make sure totalResults is uninitialized at first, though, for non virtual case--0 could be the length of the first fetch
    collection.totalResults = undefined; // eslint-disable-line no-param-reassign
  }

  if (properties && properties.initialize) {
    properties.initialize.call(collection, localModels, opt);
  }
};


// Placeholder for event mixins
oj.Collection.prototype.on = function (event, callback) {}; // eslint-disable-line no-unused-vars
oj.Collection.prototype.OnInternal = function (event, callback, context, listenTo, ignoreSilent) {}; // eslint-disable-line no-unused-vars
oj.Collection.prototype.TriggerInternal = function (silent, event, arg1, arg2, options) {}; // eslint-disable-line no-unused-vars

/**
 * Fire request event
 * @private
 */
oj.Collection.prototype._fireRequest = function (collection, xhr, options, silent) {
  this.TriggerInternal(silent, oj.Events.EventType.REQUEST, collection, xhr, options);
};

/**
 * @private
 */
oj.Collection.prototype._resetChanges = function () {
  this.changes = [];
};

/**
 * @private
 */
oj.Collection.prototype._setChangeAt = function (start, count) {
  for (var at = start; at < start + count; at++) {
    if (this.changes.indexOf(at) === -1) {
      this.changes.push(at);
            // this.changes.sort(function(a,b) { return a-b;});
    }
  }
};


/**
 * @private
 */
oj.Collection.prototype._setModels = function (models, clearing) {
  this.models = models;
  if (clearing) {
    this._modelIndices = [];
    this._resetChanges();
  } else {
    for (var i = 0; i < models.length; i++) {
      if (models[i]) {
        this._modelIndices.push(i);
      }
    }
  }
};

/**
 * @private
 */
oj.Collection.prototype._getModels = function () {
  return this.models;
};

/**
 * @private
 */
oj.Collection.prototype._getModelsLength = function () {
  return this._getModels().length;
};

/**
 * Designed to check if index exceeds the length of the models.  If we're in a virtual and "no totalResults" case,
 * we're never over the upper limit
 * @private
 */
oj.Collection.prototype._overUpperLimit = function (index) {
  if (index < this._getModelsLength()) {
    return false;
  }
  if (this.IsVirtual()) {
    if (!this._hasTotalResults() || this._getModelsLength() === 0) {
      return false;
    }
  }
  return true;
};

/**
 * @private
 */
oj.Collection.prototype._hasTotalResults = function () {
  return oj.Collection._defined(this.totalResults);
};

/**
 * @private
 */
oj.Collection._defined = function (value) {
  return value != null;
};

/**
 * @private
 */
oj.Collection.prototype._pushModel = function (model) {
  this._getModels().push(model);
  this._modelIndices.push(this._getModelsLength() - 1);
  this._setChangeAt(this._getModelsLength() - 1, 1);
};

/**
 * @private
 */
oj.Collection.prototype._pushModels = function (model) {
    // Model is being added to the end, it should be made the head
  this._makeModelHead(model);
  this._pushModel(model);
  this.lruCount = this.lruCount + 1;
  model.SetIndex(this._getModelsLength() - 1);
};

/**
 * @private
 */
oj.Collection.prototype._reduceLRU = function (removed) {
  if (removed) {
    for (var i = 0; i < removed.length; i++) {
      if (removed[i]) {
        this.lruCount = this.lruCount - 1;
      }
    }
  }
};

/**
 * @param {number} start
 * @param {number} count
 * @param {Object=} model
 * @private
 */
oj.Collection.prototype._spliceModels = function (start, count, model) {
    // Clean up prev/next links for models being removed
  for (var i = start; i < start + count; i++) {
    this._removePrevNext(this._getModel(i));
  }
  if (model === undefined) {
    this._reduceLRU(this._getModels().splice(start, count));
    this._spliceModelIndices(start, (start + count) - 1);
  } else {
    this._reduceLRU(this._getModels().splice(start, count, model));
    this._spliceModelIndices(start, (start + count) - 1);
    this._insertModelIndex(start);
    this._makeModelHead(model);
  }
  this._setChangeAt(start, count);
  if (this.lruCount < 0) {
    this.lruCount = 0;
  }
  this._realignModelIndices(start);
};

/**
 * @private
 */
oj.Collection.prototype._getModel = function (index) {
  return this._getModels()[index];
};

/**
 * Realign all the indices of the models (after sort for example)
 * @private
 */
oj.Collection.prototype._realignModelIndices = function (start) {
  var index;
  for (var i = 0; i < this._modelIndices.length; i++) {
    index = this._modelIndices[i];
    if (index >= start && this._getModel(index)) {
      this._getModel(index).SetIndex(index);
    }
  }
};

/**
 * Update next/prev pointers as though the given model were being removed
 * @private
 */
oj.Collection.prototype._removePrevNext = function (model) {
  if (!model) {
    return;
  }

  var oldPrev = model.GetPrevious();
  var oldNext = model.GetNext();
    // Link the two surrounding previous/next elements to each other, because this one is being replaced and moved
    // to the head
  if (oldPrev) {
    oldPrev.SetNext(oldNext);
  } else {
        // This element used to be the head
    this.head = oldNext;
  }
  if (oldNext) {
    oldNext.SetPrevious(oldPrev);
  } else {
        // This element used to be the tail
    this.tail = oldPrev;
  }
};

/**
 * @private
 */
oj.Collection.prototype._makeModelHead = function (model) {
    // Make this new model the most recently used: the head
  model.SetNext(this.head);
  if (this.head) {
    this.head.SetPrevious(model);
  } else {
        // No head: list is empty-->make tail the same element
    this.tail = model;
  }
  model.SetPrevious(null);
  this.head = model;
};

/**
 * Mark the model index tracker as having a used slot
 * @private
 */
oj.Collection.prototype._setModelIndex = function (index) {
  if (this._modelIndices.indexOf(index) === -1) {
    this._modelIndices.push(index);
  }
};


/**
 * Insert the given model at the given index
 * @private
 */
oj.Collection.prototype._insertModelIndex = function (start) {
    // Up all the indices of models with index greater than start
  for (var i = 0; i < this._modelIndices.length; i++) {
    if (this._modelIndices[i] >= start) {
      this._modelIndices[i] = this._modelIndices[i] + 1;
    }
  }
    // Now add the new one
  this._modelIndices.push(start);
};

/**
 * Splice out the given model index
 * @private
 */
oj.Collection.prototype._spliceModelIndices = function (start, end) {
  var localEnd = (end === undefined) ? start : end;
  this._clearModelIndices(start, localEnd);

  var count = (localEnd - start) + 1;
    // Reduce the indexes of any models above the endpoint by the number of models removed
  for (var i = 0; i < this._modelIndices.length; i++) {
    if (this._modelIndices[i] > localEnd) {
      this._modelIndices[i] -= count;
    }
  }
};

/**
 * Clear the given model index
 * @private
 */
oj.Collection.prototype._clearModelIndices = function (start, end) {
  var localEnd = (end === undefined) ? start : end;
    // Knock out any of the deleted model's indexes from the list
  for (var i = start; i <= localEnd; i++) {
    var idx = this._modelIndices.indexOf(start);
    if (idx > -1) {
      this._modelIndices.splice(idx, 1);
    }
  }
};

/**
 * @private
 */
oj.Collection.prototype._setModel = function (index, model) {
  var oldModel = this._getModel(index);
  this._removePrevNext(oldModel);
  if (!oldModel) {
        // Newly "inserted" model
    this.lruCount = this.lruCount + 1;
  }
  this._setChangeAt(index, 1);
  this._getModels()[index] = model;
  if (model) {
    this._setModelIndex(index);
    model.SetIndex(index);
    this._makeModelHead(model);
  }
};

/**
 * Clean off n models from tail (oldest) of prev/next list
 * @private
 */
oj.Collection.prototype._clearOutModels = function (n) {
  var current = this.tail;
  var index;
  var model;
  var i = 0;

  this.tail = null;
  while (current && i < n) {
        // Erase this model from collection, iff it hasn't changed
    index = current.GetIndex();
    model = this._getModel(index);
    if (!(model && model.hasChanged())) {
      this.lruCount = this.lruCount - 1;
      if (index > -1) {
                // this._getModels()[index] = undefined;
        this._setModel(index, undefined);
        this._clearModelIndices(index, index);
      }

            // Clear its pointers
      current.SetNext(null);
      current = current.SetPrevious(null);
      i += 1;
    } else {
            // Lock down the tail to this one we're not deleting
      if (!this.tail) {
        this.tail = current;
      }
      current = current.GetPrevious();
    }
  }
    // Make sure we set tail if not already set
  if (!this.tail) {
    this.tail = current;
  }
  if (this.lruCount < 0) {
    this.lruCount = 0;
  }
  if (this.lruCount === 0) {
    this.head = null;
    this.tail = null;
  }
};


/**
 * Reset the LRU list
 * @private
 */
oj.Collection.prototype._resetLRU = function () {
  this.lruCount = 0;
  this.head = null;
  this.tail = null;
};

/**
 * Make sure we have room in the LRU
 * @private
 */
oj.Collection.prototype._manageLRU = function (incoming) {
  if (this.IsVirtual()) {
    var limit = this._getModelLimit();
    if (limit > -1) {
      if (this.lruCount + incoming > limit) {
                // Must flush the amount over the limit
        this._clearOutModels((this.lruCount + incoming) - limit);
      }
    }
  }
};

/**
 * @export
 * @desc Return a copy of the Collection
 * @memberof oj.Collection
 *
 * @return {oj.Collection} copy of the Collection
 */
oj.Collection.prototype.clone = function () {
  return this._cloneInternal(true);
};

/**
 * @private
 */
oj.Collection.prototype._cloneInternal = function (withProperties) {
  var c = new this.constructor();
  var i;

    // Only copy locally if virtual
  var model;
  if (this.IsVirtual()) {
    c = this._copyFetchProperties(c);
    c._resetModelsToFullLength();
  }

  c = this._copyProperties(c);
  if (withProperties) {
        // Try to copy models only if told to--we may only need the shell of the collection with properties
        // Make a copy of the model indices (it will be modified by this process)
    var indices = [];
    for (i = 0; i < this._modelIndices.length; i++) {
      indices.push(this._modelIndices[i]);
    }
        // Sort them in reverse order so we eliminate the higher indexes first so as not to disrupt the position of
        // the earlier ones
    indices.sort(function (a, b) {
      return a - b;
    });

    var index;
    for (i = 0; i < indices.length; i++) {
      index = indices[i];
      model = this._atInternal(index, null, true, false);
      if (model) {
        c._addInternal(model.clone(), { at: index }, true, false);
      }
    }
  }
  return c;
};

/**
 * Copy critical properties in clone
 * @private
 */
oj.Collection.prototype._copyProperties = function (collection) {
  var props = ['comparator', 'model', 'modelId'];
  var prop;
  var i;

  for (i = 0; i < props.length; i++) {
    prop = props[i];
    collection[prop] = this[prop]; // eslint-disable-line no-param-reassign
  }
  return collection;
};

/**
 * Copy critical fetch properties in clone
 * @private
 */
oj.Collection.prototype._copyFetchProperties = function (collection) {
  var props = ['totalResults', 'hasMore', oj.Collection._FETCH_SIZE_PROP];
  var prop;
  var i;

  for (i = 0; i < props.length; i++) {
    prop = props[i];
    collection[prop] = this[prop]; // eslint-disable-line no-param-reassign
  }
  return collection;
};

/**
 * @private
 */
oj.Collection.prototype._getLength = function () {
  return this.length;
};

/**
 * @private
 */
oj.Collection.prototype._setLength = function () {
  var modelsLen = this._getModelsLength();
  this.length = modelsLen;
  if (!this.IsVirtual()) {
    this.totalResults = modelsLen;
  }
};

/**
 * Create a model instance using the model property if set
 * @param {Object} collection
 * @param {Object=} attrs
 * @param {Object=} options
 * @returns {null|Object}
 * @private
 */
oj.Collection._createModel = function (collection, attrs, options) {
  if (collection.model) {
    return $.isFunction(collection.model) ? new collection.model(attrs, options) : // eslint-disable-line new-cap
                                                   new collection.model.constructor(attrs, options); // eslint-disable-line new-cap
  }
  return null;
};

/**
 * @private
 */
oj.Collection.prototype._newModel = function (m, parse, options, ignoreDefaults) {
  var newModel = null;
  var validationValue;
  var opt = options || {};

//    opt.noparse = !parse;
  opt.ignoreDefaults = ignoreDefaults;

  if (m instanceof oj.Model) {
    newModel = m;
  } else if (this.model) {
            // model is defined
    newModel = oj.Collection._createModel(this, m, opt);
  } else {
            // Set this collection on the model
    opt.collection = this;
    newModel = new oj.Model(m, opt);
  }
    // Validate
  if (opt.validate && newModel.validate) {
    validationValue = newModel.validate(newModel.attributes);
    if (validationValue) {
      opt.validationError = validationValue;
      this.TriggerInternal(false, oj.Events.EventType.INVALID, newModel, validationValue, opt);
      return null;
    }
  }
  return newModel;
};

/**
 * Add a model or models to the end of the collection.<br>
 * Events:<p>
 * <ul>
 * <b>add</b>: fired for each model added, passing the collection, model added, and options<br>
 * <b>alladded</b>: fired after all models have been added, passing the collection, array of models added, and
 * options<br>
 * </ul>
 * <p>
 * Note that for virtual collections, if a new model is added after being saved up to the server, no add event
 * will be fired as the
 * collection will already "see" the model as existing.  Note that a warning will be logged if this add is not
 * a force, not merging, and duplicate IDs are found.
 * @param {oj.Model|Object|Array.<Object>|Array.<oj.Model>} m Model object (or array of models) to add. These can be already-created instance of
 * the oj.Model object, or sets of attribute/values, which will be wrapped by add() using the collection's model.
 * @param {Object=} options <b>silent</b>: if set, do not fire events<br>
 *                          <b>at</b>: splice the new model into the collection at the value given (at:index)<br>
 *                          <b>merge</b>: if set, and if the given model already exists in the collection (matched
 *                          by [id]{@link oj.Model#id}), then merge the attribute/value sets, firing change
 *                          events<br>
 *                          <b>sort</b>: if set, do not re-sort the collection even if the comparator is set.<br>
 *                          <b>force</b>: if set to true, do an add to the collection no matter whether the item is
 *                          found or not<br>
 *                          <b>deferred</b>: if true, return a promise as though this collection were virtual
 *                          whether it is or not<br>
 *
 * @returns {Promise.<Array>|Array.<oj.Model>} The model or models added to the collection (or found/merged if appropriate).  If
 * deferred or virtual, return the model or models added in a promise when the set has completed
 * @memberof oj.Collection
 * @ojsignature [{target: "Type", value:"{silent?: boolean, at?: number, merge?: boolean, sort?: boolean, force?: boolean, deferred?: boolean,
 *               [propName: string]: any}", for: "options"},
 *               {target: "Type", value: "Promise<Array<oj.Model>>|Array<oj.Model>", for: "returns"}]
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.add = function (m, options) {
  this._manageLRU(1);
  var opt = options || {};
  return this._handlePromise(this._addInternal(m, options, false, opt.deferred));
};

/**
 * fillIn: true indicates that we're just trying to use add() after a fetch to
 * insert into a preallocated list of models, not truly do an add/merge from the API
 * @private
 */
oj.Collection.prototype._addInternal = function (m, options, fillIn, deferred) {
    // Get options
  var opt = options || {};
  var modelArray = [];
  var at = opt.at;
  var silent = opt.silent;
  var force = opt.force;
  var i;
  var index;
  var cid;
  var merge = opt.merge || false;
  var sort = opt.sort;
  var needSort = true;
  var added = false;
  var addedModels = [];
  var modelReturnList = [];

  if (at !== undefined && at < 0) {
        // Normalize it using the length-- another BackboneJS test case
    at += this._getLength() + 1;
  }

  if (m instanceof Array) {
    modelArray = m;
  } else {
    modelArray.push(m);
  }

  function addToCollection(collection, newModel) {
    if (at === undefined) {
      collection._pushModels(newModel);
      index = collection._getModelsLength() - 1;
      collection._getModel(index).SetCid();
    } else {
      index = at;
      if (collection.IsVirtual() && fillIn) {
                // Array has been preallocated in this case
        collection._setModel(index, newModel);
      } else {
        collection._spliceModels(index, 0, newModel);
      }
      collection._getModel(index).SetCid();
            // Increment at so that later models will be added right after their predecessors, if an array is
            // passed in
      at += 1;
    }
    if (newModel.GetCollection() === undefined) {
      newModel.SetCollection(collection);
    }
    collection._setLength();
    collection._listenToModel(newModel);
    added = true;
  }

  function resortAndFireEvents(collection, existingModel, modelFoundInCollection,
    newModel, resortDeferred) {
    // Now resort if required (don't resort if either told not to, or if 'at' option is set)
    // and if there's more than one model
    var resortOpt = opt || {};
    if (fillIn) {
      resortOpt.fillIn = true;
    }
    var comparator = resortOpt.comparator || collection._hasComparator();
        // If we're filling in to a blank, also check that there's no comparator
    var fillInSort = !fillIn || (fillIn && comparator);
    if (fillInSort && needSort && existingModel === undefined && !sort &&
        at === undefined && collection._getLength() > 1) {
      if (index > -1) {
        cid = collection._getModel(index).cid;
      }
      var sortOpt = {};
      oj.CollectionUtils.copyInto(sortOpt, resortOpt);
      sortOpt.add = true;
      collection.sort(sortOpt);
            // Reset index--can't get it back if virtual--set to -1
      if (index > -1) {
        if (collection.IsVirtual()) {
          index = -1;
        } else {
          index = collection.indexOf(collection.getByCid(cid), resortDeferred);
        }
      }
    }

    if (added) {
      // Pass index property in resortOpt, if at is specified
      if (resortOpt.at) {
        resortOpt.index = index;
      }
      if (newModel) {
        newModel.TriggerInternal(silent, oj.Events.EventType.ADD, newModel, collection, resortOpt);
        addedModels.push(newModel);
      } else {
        modelFoundInCollection.TriggerInternal(silent, oj.Events.EventType.ADD,
          modelFoundInCollection, collection, resortOpt);
        addedModels.push(modelFoundInCollection);
      }
    }
  }

  function mergeAttrs(collection, modelToTryAndMerge, modelFoundInCollection, newModel,
    mergeDeferred) {
    var existingModel;
    var modelAdded = null;

    if (!force && merge && modelFoundInCollection) {
            // Try to merge the attributes--we're merging and the model (by id) was already in the collection
      needSort = modelFoundInCollection.Merge(modelToTryAndMerge, collection.comparator, silent);
      modelAdded = modelFoundInCollection;
    } else {
            // Make sure model is not already in there
      if (!force) {
        if (fillIn) {
                  // Only bother if we have a real id set--comparing cids on a fill in is useless
          existingModel = !newModel.isNew() ? collection._getLocal(newModel) : undefined;
        } else {
          existingModel = collection._getLocal(newModel);
        }
        if (existingModel && fillIn && at !== existingModel.index) {
                    // We're filling in a virtual collection: we should *not* be finding the new model already in
                    // the collection if we're not merging and not forcing: this indicates duplicate ids
                    // throw new Error("Duplicate IDs fetched or added without merging");
          oj.Logger.warn('Duplicate ID fetched or added without merging, the id = ' +
                                   existingModel.GetId());
        }
      }

      if (existingModel === undefined) {
        addToCollection(collection, newModel);
        modelAdded = newModel;
      } else {
        modelAdded = existingModel;
      }
    }

    resortAndFireEvents(collection, existingModel, modelFoundInCollection, newModel, mergeDeferred);

    return modelAdded;
  }

  function doAdd(collection, model, addDeferred) {
    added = false;
    var newModel = collection._newModel(model, true, opt, false);
    var modelToTryAndMerge = null;
    var modelFoundInCollection = null;
    if (newModel != null) {
      index = -1;
            // Make sure ID is up to date
      newModel.SetupId();

            // Use original model array not cloned model if merging--otherwise we won't find the model in the
            // collection
      modelToTryAndMerge = model instanceof oj.Model ? model : newModel;
      if (addDeferred) {
        if (force) {
          return new Promise(function (resolve) {
            var mergedModel = mergeAttrs(collection, modelToTryAndMerge, undefined, newModel,
              addDeferred);
            modelReturnList.push(mergedModel);
            resolve(mergedModel);
          });
        }
        return collection._getInternal(modelToTryAndMerge, { silent: true },
          addDeferred, true).then(
            function (modInfo) {
              modelFoundInCollection = modInfo.m;
              var mod = mergeAttrs(collection,
                   modelToTryAndMerge, modelFoundInCollection,
                   newModel, addDeferred);
              modelReturnList.push(mod);
            });
      }
      if (!force && merge) {
                // Grab the actual model we want to merge from the collection, if the caller has indicated that
                // we aren't forcing an add and we want to merge
        modelFoundInCollection = fillIn ? collection._getLocal(modelToTryAndMerge) :
                                                  collection.get(modelToTryAndMerge);
      }
      var modelAdded = mergeAttrs(collection, modelToTryAndMerge, modelFoundInCollection, newModel,
                                  addDeferred);
      if (modelAdded) {
        modelReturnList.push(modelAdded);
      }
    } else {
            // Add boolean falses for invalid models
      modelReturnList.push(false);
    }
    return Promise.resolve();
  }

  function _parse(collection, array) {
        // Must stop parsing if coming in from reset or constructor
    if (collection.parse && opt.parse && !opt.noparse) {
      return collection.parse(array);
    }
    return array;
  }

  if (!fillIn && (this.IsVirtual() || deferred)) {
    var self = this;
    return new Promise(function (allResolve, allReject) {
      var doTask = function (loc) {
        return new Promise(function (resolve, reject) {
          doAdd(self, modelArray[loc], true).then(function () {
            resolve(loc + 1);
          }, reject);
        });
      };

      var currentStep = Promise.resolve(0);

      modelArray = _parse(self, modelArray);
      for (i = 0; i < modelArray.length; i++) {
        currentStep = currentStep.then(doTask);
      }
      currentStep.then(function () {
        if (addedModels.length > 0) {
          self.TriggerInternal(opt.silent, oj.Events.EventType.ALLADDED, self, addedModels,
                                         opt);
        }
        allResolve(oj.Collection._returnModels(modelReturnList));
      }, allReject);
    });
  }

  modelArray = _parse(this, modelArray);
  for (i = 0; i < modelArray.length; i++) {
    doAdd(this, modelArray[i], false);
  }
  if (addedModels.length > 0) {
    this.TriggerInternal(opt.silent, oj.Events.EventType.ALLADDED, this, addedModels, opt);
  }
  return oj.Collection._returnModels(modelReturnList);
};

/**
 * @private
 */
oj.Collection._returnModels = function (modelReturnList) {
  if (modelReturnList.length === 1) {
    return modelReturnList[0];
  }
  return modelReturnList;
};

/**
 * @private
 */
oj.Collection.prototype._hasComparator = function () {
  return oj.Collection._defined(this.comparator);
};

/**
 * Sort the models in the collection.  For virtual collections, any locally stored models are cleaned out in
 * preparation for new fetches sending server-sorted models down to the client.  No fetch is automatically
 * performed.<p>
 * For non-virtual collections, the models are sorted based on the comparator property.<p>
 * See [comparator}{@link oj.Collection#comparator}<p>
 * For virtual collections, all sorting must be handled by the server.  If the string (attribute) comparator and
 * sortDirection are set, then this information
 * will be passed back via URL parameters, or passed to the customURL callback for application-construction of the
 * appropriate URL.  Function-based custom comparators are ignored in virtual collections.
 * Events:<p>
 * Fires a sort event, passing the collection and options<p>
 * @param {Object=} options <b>silent</b>: if true, do not fire the sort event<br>
 * <b>startIndex</b>: if provided, and if collection is virtual, do a fetch of startIndex+fetchSize immediately
 * after the sort and return a promise.  See [comparator}{@link oj.Collection#setRangeLocal}<p>
 * @return {Promise|null} if virtual and if startIndex is provided in options, a promise Object that resolves upon
 * completion.  The promise will be passed an Object containing start and count properties that represent
 * the *actual* starting position (start), count (count), and array (models) of the Models fetched, which may be
 * fewer than what was requested.  The promise will be rejected on an error and will pass the ajax status,
 * xhr object, error, and collection, if relevant.
 *
 * @memberof oj.Collection
 * @ojsignature  [{target: "Type", value:"{silent?: boolean, startIndex?: number, [propName: string]: any}", for: "options"},
 *                {target: "Type", value: "Promise<oj.Collection.SetRangeLocalPromise>|null", for: "returns"}]
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.sort = function (options) {
  var opt = options || {};
  var silent = opt.silent;
  var comparator = this.comparator;
  var self;
  var eventOpts;

    // Check for comparator
  if (!this._hasComparator()) {
    return null;
  }

    // This is a no-op in case of virtualization: we should just clear things out so that
    // any elements will be refetched
  if (this.IsVirtual()) {
    var totalResults = this.totalResults;
    if (this._hasTotalResults()) {
            // Make sure to set up the array if the length changes (i.e., from 0 to totalResults--need to
            // preallocate)
      this._setModels(new Array(totalResults), true);
    } else {
            // No totalresults
      this._setModels([], true);
      this._resetLRU();
      this._setLength();
    }
    eventOpts = opt.add ? { add: true } : null;
    this.TriggerInternal(silent, oj.Events.EventType.SORT, this, eventOpts, null);
    var startIndex = opt.startIndex;
    if (startIndex !== undefined && startIndex !== null) {
      return this.setRangeLocal(startIndex, this._getFetchSize(opt));
    }
    return null;
  }

  self = this;
  this._getModels().sort(function (a, b) {
    return oj.Collection.SortFunc(a, b, comparator, self, self);
  });
  this._realignModelIndices(0);
    // Indicate this sort is due to an add
  eventOpts = opt.add ? { add: true } : null;
  this.TriggerInternal(silent, oj.Events.EventType.SORT, this, eventOpts, null);
  return null;
};

/**
 * @private
 */
oj.Collection._getKey = function (val, attr) {
  if (val instanceof oj.Model) {
    return val.get(attr);
  }
  return oj.Model.GetPropValue(val, attr);
};

/**
 * @private
 */
oj.Collection.SortFunc = function (a, b, comparator, collection, self) {
  var keyA;
  var keyB;
  var i;
  var retVal;

  if ($.isFunction(comparator)) {
        // How many args?
    if (comparator.length === 1) {
            // "sortBy" comparator option
      keyA = comparator.call(self, a);
      keyB = comparator.call(self, b);
      var attrs1 = oj.StringUtils.isString(keyA) ? keyA.split(',') : [keyA];
      var attrs2 = oj.StringUtils.isString(keyB) ? keyB.split(',') : [keyB];
      for (i = 0; i < attrs1.length; i++) {
        retVal = oj.Collection._compareKeys(attrs1[i], attrs2[i], collection.sortDirection);
        if (retVal !== 0) {
          return retVal;
        }
      }
    }
        // "sort" comparator option
    return comparator.call(self, a, b);
  }
    // String option
  if (oj.StringUtils.isString(comparator)) {
    var attrs = comparator.split(',');

    for (i = 0; i < attrs.length; i++) {
      keyA = oj.Collection._getKey(a, attrs[i]);
      keyB = oj.Collection._getKey(b, attrs[i]);
      retVal = oj.Collection._compareKeys(keyA, keyB, collection.sortDirection);
      if (retVal !== 0) {
        return retVal;
      }
    }
  }
  return 0;
};

/**
 * Return the index at which the given model would be inserted, using the collection comparator See
 * [sort}{@link oj.Collection#sort}.  Not supported for virtual collections.
 *
 * @param {string|function(oj.Model,oj.Model=):Object=} comparator optional comparator to override the default
 * @returns {number} index at which model would be inserted.  -1 if no comparator
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @export
 * @since 1.0.0
 */
oj.Collection.prototype.sortedIndex = function (model, comparator) {
  var comp = comparator || this.comparator;
  var self;
  var test;

    // Check for comparator
  if (!comp) {
    return -1;
  }

  this._throwErrIfVirtual('sortedIndex');

  self = this;
  test = function (a, b) {
    var keyA;
    var keyB;

    if ($.isFunction(comp)) {
                // How many args?
      if (comp.length === 1) {
                    // "sortBy" comparator option
        keyA = comp.call(self, a);
        keyB = comp.call(self, b);
        var attrs1 = oj.StringUtils.isString(keyA) ? keyA.split(',') : [keyA];
        var attrs2 = oj.StringUtils.isString(keyB) ? keyB.split(',') : [keyB];
        var retVal;
        var i;
        for (i = 0; i < attrs1.length; i++) {
          retVal = oj.Collection._compareKeys(attrs1[i], attrs2[i], self.sortDirection);
          if (retVal !== 0) {
            return retVal;
          }
        }
      }
                // "sort" comparator option
      return comp.call(self, a, b);
    }
            // String option
    if (oj.StringUtils.isString(comp)) {
      keyA = a.get(comp);
      keyB = b.get(comp);
      return oj.Collection._compareKeys(keyA, keyB, self.sortDirection);
    }
    return 0;
  };
  return oj.Collection._find(this._getModels(), model, test);
};

/**
 * Binary search and return the index at which model would be inserted into sorted modelArray
 * @private
 */
oj.Collection._find = function (modelArray, model, comparator) {
  function search(min, max) {
    var cid;
    var id;
    var mid;

    if (min > max) {
      return -1;
    }

    cid = model.GetCid();
    id = model.GetId();
    if (modelArray[min].Match(id, cid)) {
      return min;
    }
    if (modelArray[max].Match(id, cid)) {
      return max;
    }

    mid = Math.floor((max + min) / 2);
    if (comparator(modelArray[mid], model) === -1) {
      return search(min + 1, mid);
    }
    if (comparator(modelArray[mid], model) === 1) {
      return search(mid, max - 1);
    }
    return mid;
  }

  return search(0, modelArray.length - 1);
};

/**
 * @private
 */
oj.Collection._compareKeys = function (keyA, keyB, sortDirection) {
  if (sortDirection === -1) {
    if (keyA < keyB) {
      return 1;
    }
    if (keyB < keyA) {
      return -1;
    }
  } else {
    if (keyA > keyB) {
      return 1;
    }
    if (keyB > keyA) {
      return -1;
    }
  }
  return 0;
};


/**
 * Add the given model to the front of the collection<br>
 * For events that may be fired, see [add]{@link oj.Collection#add}.<br>
 * @param {Object} m model (or set of attribute/value pairs) to add to the beginning of the collection
 * @param {Object=} options See [add]{@link oj.Collection#add}
 * @returns {Promise.<Array>|Array.<oj.Model>} The model or models added to the collection.  If
 * deferred or virtual, return the model or models added in a promise when the set has completed
 * @memberof oj.Collection
 * @ojsignature [{target: "Type", value:"{silent?: boolean, at?: number, merge?: boolean, sort?: boolean, force?: boolean, deferred?: boolean,
 *               [propName: string]: any}", for: "options"},
 *               {target: "Type", value: "Promise<Array<oj.Model>>|Array<oj.Model>", for: "returns"}]
 * @export
 * @since 1.0.0
 */
oj.Collection.prototype.unshift = function (m, options) {
    // Like an add but set 'at' to zero if not specified
  var opt = {};
  oj.CollectionUtils.copyInto(opt, options || {});
  if (!opt.at) {
    opt.at = 0;
  }
  this._manageLRU(1);
  return this._handlePromise(this._addInternal(m, opt, false, opt.deferred));
};

/**
 * @private
 */
oj.Collection.prototype._handlePromise = function (result) {
  if ($.isFunction(result.then)) {
    return this._addPromise(function () {
      return result;
    });
  }
  return result;
};

/**
 * Remove the first model from the collection and return it.<br>
 * For events that may be fired, see [remove]{@link oj.Collection#remove}.<br>
 * @param {Object=} options same as remove
 * @property {boolean=} silent if set, do not fire events
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual whether it is or not
 * @return {oj.Model|Promise.<oj.Model>|null} model that was removed.  If this is a virtual collection, this will return a promise
 *                  which will resolve passing the model value that was removed
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.shift = function (options) {
  var opt = options || {};
  var deferred = this._getDeferred(opt);
  var retVal;
  if (this.IsVirtual() || deferred) {
    var self = this;
    return this._atInternal(0, opt, false, true).then(function (model) {
      retVal = self._removeInternal(model, 0, opt);
      self.TriggerInternal(opt.silent, oj.Events.EventType.ALLREMOVED, self, [retVal], opt);

      return retVal;
    });
  }
  retVal = this._removeInternal(this.at(0), 0, opt);
  this.TriggerInternal(opt.silent, oj.Events.EventType.ALLREMOVED, this, [retVal], opt);
  return retVal;
};

/**
 * @export
 * @desc Return an array of models found in the Collection, excepting the last n.
 * @memberof oj.Collection
 * @param {number=} n number of models to leave off the returned array; defaults to 1
 * @return {Array.<oj.Model>} array of models from 0 to the length of the collection - n - 1
 * @throws {Error} when called on a virtual collection
 * @since 1.0.0
 */
oj.Collection.prototype.initial = function (n) {
  var index = (n === undefined) ? 1 : n;

  this._throwErrIfVirtual('initial');

  var array = [];
  var i;
  for (i = 0; i < this._getLength() - index; i += 1) {
    array.push(this.at(i));
  }
  return array;
};

/**
 * @private
 */
oj.Collection.prototype._getDeferred = function (options) {
  var opt = options || {};
  return opt.deferred;
};

/**
 * Return the last model in the collection.  If n is passed in, then the last n models are returned as an array
 * Note that if the collection is virtual, and totalResults is not returned by the server, the results returned
 * by last can be difficult to predict.  They depend on the fetch sizes, last known offset of a fetch, etc.
 * If code is using a server that does not return totalResults the use of last is not recommended.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 * @param {number=} n number of models to return.  Defaults to 1
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual whether it is or not
 * @return {Promise.<oj.Model>|Array.<oj.Model>|null} array of n models from the end of the Collection.  If this is a virtual collection,
 *                             this will return a promise which will resolve passing the array or single model
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.last = function (n, options) {
  var deferred = this._getDeferred(options);
  var index = (n === undefined) ? 1 : n;
  var len;

  if (index === 1) {
    len = this._getModelsLength();
    if (len === 0) {
      len = index;
    }
    if (len > 0) {
      return this._atInternal(len - 1, options, false, deferred);
    }
    return null;
  }

  var array = [];
  var i;
  len = this._getLength();
  if (deferred || this.IsVirtual()) {
        // Loop using deferred
    var start = len - n;

        // Handle edge or no totalResults cases
    if (start < 0) {
      start = 0;
    }
    if (len === 0) {
            // No totalresults, probably
      len = n;
    }

    var self = this;
    return this._addPromise(function () {
      return self.IterativeAt(start, len);
    });
  }

  for (i = len - n; i < len; i += 1) {
    array.push(this.at(i));
  }
  return array;
};

/**
 * Loop calling at() in a deferred way and return a promise to be resolved when all the elements are sequentially
 * fetched
 * @protected
 */
oj.Collection.prototype.IterativeAt = function (start, end) {
  var array = [];
  var i;
  var self = this;
  return new Promise(function (allResolve, allReject) {
    var doTask = function (index) {
            // Make sure we're not asking beyond what we know the server can deliver, if virtual
      if (self.IsVirtual() && self._hasTotalResults() && index >= self.totalResults) {
        return Promise.resolve(index + 1);
      }
      return new Promise(function (resolve, reject) {
        self._deferredAt(index, null).then(function (model) {
          array.push(model);
          resolve(index + 1);
        }, reject);
      });
    };

    var currentStep = Promise.resolve(start);
    for (i = start; i < end; i++) {
      currentStep = currentStep.then(doTask);
    }
    currentStep.then(function () {
      allResolve(array);
    }, allReject);
  });
};

/**
 * @private
 */
oj.Collection.prototype._getDefaultFetchSize = function (n) {
  if (n === undefined || n === null) {
    return this[oj.Collection._FETCH_SIZE_PROP];
  }

  return n;
};

/**
 * @private
 */
oj.Collection.prototype._calculateNextStart = function () {
  var lastFetch = this.lastFetchCount;
  if (lastFetch === undefined || lastFetch === null) {
    lastFetch = this[oj.Collection._FETCH_SIZE_PROP];
  }
  if (this.offset === undefined || this.offset === null) {
        // Assume zero offset (0+lastFetch)
    return lastFetch;
  }
  return this.offset + lastFetch;
};

/**
 * Fetch the next set of models from the server.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 *
 * @param {number} n number of models to fetch.  If undefined or null, the collection will attempt to use the
 * overall [fetchSize]{@link oj.Collection#fetchSize} property value
 * @param {Object=} options Options to control next<p>
 *                  <b>success</b>: a user callback called when the fetch has completed successfully. This makes
 *                  the fetch an asynchronous process. The callback is called passing the Collection object, raw
 *                  response, and the options argument.<br>
 *                  <b>error</b>: a user callback function called if the fetch fails. The callback is called
 *                  passing the collection object, xhr, and options arguments.<br>
 * @return {Object|null} xhr ajax object, by default.  null if nothing to fetch (the success callback will still be
 * called).  Note if [sync]{@link oj.Collection#sync} has been replaced, this would be the value returned by the
 * custom implementation.
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{success?: (collection: oj.Collection, response: Object, options: Object)=> void,
 *                                                  error?: (collection: oj.Collection, xhr: Object, options: Object)=> void, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.next = function (n, options) {
  var opt = options || {};
  opt[oj.Collection._FETCH_SIZE_PROP] = this._getDefaultFetchSize(n);

  var start = this._calculateNextStart();
  var length = this._getLength();
  if (length === 0 && opt[oj.Collection._FETCH_SIZE_PROP] > 0) {
        // If we have a fetch size and we have no length let next() do a fetchSize fetch starting at zero to
        // kick things off
    start = 0;
  } else if (start >= length) {
        // No op -- still call success because the items are already fetched.
    var self = this;
    if (opt.success) {
      opt.success.call(oj.Model.GetContext(opt, self), self, null, opt);
    }

    return null;
  }
  opt.startIndex = start;
  return this.fetch(opt);
};

/**
 * @private
 */
oj.Collection.prototype._calculatePrevStart = function (n) {
  if (this.offset === undefined || this.offset === null) {
        // Assume zero: we can't back up beyond that so if the offset wasn't set there's nothing to do
    return 0;
  }
  return this.offset - n;
};

/**
 * Fetch the previous set of models from the server.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 * @param {number} n number of models to fetch.  If undefined or null, the collection will attempt to use the
 * overall fetchSize property value
 * @param {Object=} options Options to control previous<p>
 *                  <b>success</b>: a user callback called when the fetch has completed successfully. This makes
 *                  the fetch an asynchronous process. The callback is called passing the Collection object, raw
 *                  response, and the options argument.<p>
 *                  <b>error</b>: a user callback function called if the fetch fails. The callback is called
 *                  passing the collection object, xhr, and options arguments.<p>
 * @return {Object} xhr ajax object, by default. null if there is nothing earlier to fetch (no fetch has happened
 * or the last fetch started at 0).  The success callback will still be called.  Note if
 * [sync]{@link oj.Collection#sync} has been replaced, this would be the value returned by the custom
 * implementation.
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{success?: (collection: oj.Collection, response: Object, options: Object)=> void,
 *                                                  error?: (collection: oj.Collection, xhr: Object, options: Object)=> void, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.previous = function (n, options) {
  var opt = options || {};
  if (this.offset === 0) {
        // No op -- still call success (if we've fetched before--lastFetchCount is other than zero) because the
        // items are already fetched.
    var self = this;
    if (opt.success && this.lastFetchCount) {
      opt.success.call(oj.Model.GetContext(opt, self), self, null, opt);
    }

    return null;
  }
  opt[oj.Collection._FETCH_SIZE_PROP] = this._getDefaultFetchSize(n);
  var start = this._calculatePrevStart(opt[oj.Collection._FETCH_SIZE_PROP]);
  if (start < 0) {
        // Only fetch from 0 to the last fetch's starting point...
    opt[oj.Collection._FETCH_SIZE_PROP] = this.offset;
    start = 0;
  }
  opt.startIndex = start;
  return this.fetch(opt);
};


/**
 * Set or change the number of models held at any one time
 *
 * @param {number} n maximum number of models to keep at a time
 * @return {undefined}
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.setModelLimit = function (n) {
  this.modelLimit = n;
    // Clean out down to the new limit, if necessary
  this._manageLRU(0);
};

/**
 * @private
 */
oj.Collection.prototype._getModelLimit = function () {
  return this.modelLimit;
};

/**
 * Set or change the number of models to fetch with each server request
 *
 * @param {number} n number of models to fetch with each request
 * @return {undefined}
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.setFetchSize = function (n) {
  this[oj.Collection._FETCH_SIZE_PROP] = n;
};


/**
 * Return the array of models found in the Collection starting with index n.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 * @param {number=} n index at which to start the returned array of models.  Defaults to 1.
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual whether it is or not
 * @return {Array.<Object>|Promise} array of models from the collection.  If this is a virtual collection, or
 *                  if deferred is passed as true, return a promise which resolves passing the array of models.
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.rest = function (n, options) {
  var deferred = this._getDeferred(options);
  var index = (n === undefined) ? 1 : n;

  var array = [];
  var i;
    // TODO
  if (this.IsVirtual() || deferred) {
    var self = this;
    return this._addPromise(function () {
      return self.IterativeAt(index, self._getLength());
    });
  }

  for (i = index; i < this._getLength(); i += 1) {
    array.push(this.at(i));
  }
  return array;
};

/**
 * Remove a model from the collection, if found.<br>
 * Events:<p>
 * <ul>
 * <b>remove</b>: fired for each model removed, passing the model removed, collection, and options<br>
 * <b>allremoved</b>: fired after all models have been removed, passing the collection, array of models removed,
 * and options<br>
 * </ul>
 * <p>
 * @param {oj.Model|Array.<oj.Model>} m model object or array of models to remove.
 * @param {Object=} options
 * @property {boolean=} silent if set, do not fire events
 * @return {Array.<oj.Model>|Object} an array of models or single model removed
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.remove = function (m, options) {
  var opt = options || {};
  var modArray = [];
  var mod;

  if (m instanceof Array) {
    modArray = m;
  } else {
    modArray.push(m);
  }

  var modsRemoved = [];
  for (mod = modArray.length - 1; mod >= 0; mod -= 1) {
        // Done to keep array in order matching one passed in--we do removal in reverse
    modsRemoved.unshift(this._removeInternal(modArray[mod], -1, opt));
  }
  this.TriggerInternal(opt.silent, oj.Events.EventType.ALLREMOVED, this, modArray, opt);
  return oj.Collection._returnModels(modsRemoved);
};

/**
 * @private
 */
oj.Collection.prototype._removeInternal = function (model, index, options) {
  var optCopy = options || {};
  var modInfo = index === -1 ? this._getInternal(model) : oj.Collection._getModinfo(index, model);
  var silent = optCopy.silent;

  var n = modInfo.index;
  if (n > -1) {
    var mod = modInfo.m;
        // only unset the collection setting if it's mine
    if (mod !== undefined && mod.GetCollection() === this) {
      mod.SetCollection(null);
    }
    this._spliceModels(n, 1);
    this._setLength();
        // if (!silent) {
    var opt = {};
    oj.CollectionUtils.copyInto(opt, optCopy);
    opt.index = n;
    if (mod !== undefined) {
      mod.TriggerInternal(silent, oj.Events.EventType.REMOVE, mod, this, opt);
    }
      //  }
        // Unlisten after event fired
    this._unlistenToModel(mod);
  }
  return modInfo.m;
};


/**
 * @private
 */
oj.Collection.prototype._unlistenToModel = function (m) {
  if (m !== undefined) {
    m.off(null, null, this);
  }
};

/**
 * @private
 */
oj.Collection.prototype._listenToModel = function (m) {
  m.OnInternal(oj.Events.EventType.ALL, this._modelEvent, this, false, true);
};

/**
 * Handle model destroyed events via the all listener
 * @private
 */
oj.Collection.prototype._modelEvent = function (event, model, collection, options) {
  if (event === oj.Events.EventType.DESTROY) {
    this.remove(model);
  }

    // Don't process general events if we're not the target
  if (collection !== undefined && collection instanceof oj.Collection && collection !== this) {
    return;
  }

    // Throw up to the collection
  var silent = options && options.silent;
  this.TriggerInternal(silent, event, model, collection, options);
};

/**
 * Clear all data from the collection and refetch (if non-virtual).  If virtual, clear all data.
 * If fetch() cannot be called (e.g., there is no URL) then the caller must reload the collection wtih data.
 * The existing data is still cleared.
 * Events (if silent is not set):<p>
 * <ul>
 * <b>refresh</b>: fired passing the collection and options<br>
 * For more events that may be fired if the collection is not virtual, see [fetch]{@link oj.Collection#fetch}.
 * </ul>
 * <p>
 * @param {Object=} options user options<p>
 *                          <b>silent</b>: if set, do not fire a refresh event<p>
 * <b>startIndex</b>: if provided, and if collection is virtual, do a fetch of startIndex+fetchSize immediately
 * after the refresh and return a promise.  See [comparator}{@link oj.Collection#setRangeLocal}<p>
 * @return {Promise.<undefined|Object>} promise object resolved when complete (in case there is a fetch for non-virtual mode).  If
 * startIndex is provided as an option, the returned promise resolution is the same as setRangeLocal.
 * @ojsignature  {target: "Type", value: "Promise<oj.Collection.SetRangeLocalPromise|undefined>", for: "returns"}
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{silent?: boolean, startIndex?: number, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.refresh = function (options) {
  var optCopy = options || {};

  var self = this;
  var silent;
  return this._addPromise(function () {
    return new Promise(function (resolve, reject) {
      if (!self.IsVirtual()) {
        silent = optCopy.silent !== undefined && optCopy.silent;
        try {
                    // Do a reset, with silent
          self.reset(null, { silent: true });
                    // Local: do a fetch to fill back up
                    // In case options are passed to refresh-->fetch
          var opt = {};
          Object.keys(optCopy).forEach(function (prop) {
            if (Object.prototype.hasOwnProperty.call(optCopy, prop)) {
              opt[prop] = optCopy[prop];
            }
          });
          opt.success = function (collection, response, successOpts) {
            self.TriggerInternal(silent, oj.Events.EventType.REFRESH, self, successOpts, null);
            resolve({ collection: collection, response: response, options: successOpts });
          };
          opt.error = function (xhr, status, error) {
            reject(oj.Collection._createRejectionError(xhr, status, error, self, optCopy, false));
          };
          self.fetch(opt);
          return;
        } catch (e) {
                    // This is OK if it's a URLError: just fire the event: local collection without custom sync
          if (e instanceof oj.URLError) {
                        // if it's a completely local collection, it's a no-op other than the event
            self.TriggerInternal(silent, oj.Events.EventType.REFRESH, self, optCopy, null);
            resolve({ collection: self, options: optCopy });
            return;
          }
          throw e;
        }
      }
            // Virtual
      var startIndex = optCopy.startIndex;

      self._setModels([], true);
      self._resetLRU();

            // Clear totalresults.
      self.totalResults = undefined;
      self._setLength();

      silent = optCopy.silent !== undefined && optCopy.silent;
      self.TriggerInternal(silent, oj.Events.EventType.REFRESH, self, optCopy, null);
      if (startIndex === undefined || startIndex === null) {
        startIndex = 0;
      }
      if (startIndex !== undefined && startIndex !== null) {
        // Do a set range local
        self._setRangeLocalInternal(startIndex, self._getFetchSize(optCopy)).then(
        function (actual) {
          resolve(actual);
        }, function (err) {
          reject(err);
        });
      } else {
        resolve(undefined);
      }
    });
  });
};

/**
 * Remove and replace the collection's entire list of models with a new set of models, if provided. Otherwise,
 * empty the collection.  totalResults is reset when no new data is passed in, set to the new length in the
 * non-virtual case if data is passed in, and left as was in the virtual case if data is passed in.<br>
 * Events (if silent is not set):<p>
 * <ul>
 * <b>reset</b>: fired passing the collection and options.  The new length of the collection should be in effect
 * when the event is fired<br>
 * For events that may be fired if data is passed in, see [add]{@link oj.Collection#add}.
 * </ul>
 * <p>
 * @param {Object=} data Array of model objects or attribute/value pair objects with which to replace the
 * collection's data.
 * @param {Object=} options user options, passed to event, unless silent<br>
 *                          <b>silent</b>: if set, do not fire events<p>
 * @returns {oj.Model|Array.<oj.Model>} The model or models added to the collection, if passed in
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{silent?: boolean, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.reset = function (data, options) {
  var opts = {};
  oj.CollectionUtils.copyInto(opts, options || {});

  opts.previousModels = this._getModels();

  var index;
  var model;
  for (var i = 0; i < this._modelIndices.length; i++) {
    index = this._modelIndices[i];
    model = this._getModel(index);
    if (model) {
      this._unlistenToModel(model);
      model.SetCollection(null);
    }
  }
  this._setModels([], true);
  this._resetLRU();

  var silent = opts.silent !== undefined && opts.silent;
  if (!data) {
    this._setLength();
        // Clear totalresults
    this.totalResults = undefined;
    this.TriggerInternal(silent, oj.Events.EventType.RESET, this, opts, null);
    return null;
  }

  var retObj = null;
  // Parse collection
  var newData = data;
  if (opts.parse) {
    newData = this.parse(data);
  }

  this._manageLRU((newData instanceof Array) ? newData.length : 1);
  opts.noparse = true;
  retObj = this._addInternal(newData, opts, true, false);
  this._setLength();
  this.TriggerInternal(silent, oj.Events.EventType.RESET, this, opts, null);

  return this._handlePromise(retObj);
};

/**
 * Return the model object found at the given index of the collection, or a promise object that will pass the model
 * as an argument when it resolves.<p>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.
 * @param {number} index Index for which to return the model object.
 * @param {Object=} options <b>fetchSize</b>: fetch size to use if the call needs to fetch more records from the
 * server, if virtualized.  Overrides the overall fetchSize setting <br>
 *                  <b>deferred</b>: if true, return a deferred/promise object as described below.  If not
 *                  specified, the type of return value will be determined by whether or not the collection is
 *                  virtual<br>
 *
 * @return {oj.Model|Promise.<oj.Model>|null} Model [model]{@link oj.Model} located at index. If index is out of range, returns null.
 * If this is a virtual collection, or if deferred is specified and true, at will return a Promise object which
 * will resolve passing the model at the given index (or null if out of range)
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{fetchSize?: number, deferred?: boolean, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.at = function (index, options) {
  var deferred = this._getDeferred(options);
  return this._atInternal(index, options, false, deferred);
};

/**
 * Local indicates that only what's stored locally should be returned (if true)--no fetching
 * @private
 */
oj.Collection.prototype._atInternal = function (index, options, local, deferred) {
  var n = index;
  if (n < 0) {
        // Normalize it using the length-- another BackboneJS test case
    n += this._getLength();
  }

  if (n < 0 || this._overUpperLimit(n)) {
    if (!local && (this.IsVirtual() || deferred)) {
      return this._addPromise(function () {
        return Promise.resolve(null);
      });
    }
    return null;
  }
  var self = this;
  if (!local && (this.IsVirtual() || deferred)) {
    return this._addPromise(function () {
      return self._deferredAt(n, options);
    });
  }
  return this._getModel(n);
};

/**
 * Return a promise, the resolution of which indicates that all promise-returning API in process have completed.
 *
 * @returns {Promise} a promise that when resolved, indicates that all unresolved promise-returning API calls
 * made at the moment of the whenReady have completed
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.whenReady = function () {
  if (this._promises) {
        // If we have an active chain, return it
    return this._promises;
  }
    // Otherwise return an immediately resolved promise that means nothing...
  return Promise.resolve();
};

/**
 * Add a task to a chained list of Promises
 * @private
 */
oj.Collection.prototype._addPromise = function (promiseTask) {
  var self = this;
    // Undefined, so set it up initially
  if (this._promises === undefined) {
    this._promiseCount = 0;
    this._promises = Promise.resolve();
  }
    // Track the number we have left to resolve
  this._promiseCount = this._promiseCount + 1;
    // Chain this new promise callback task to the end of the list
  this._promises = this._promises.then(promiseTask.bind(self)).then(
        function (arg) {
            // Resolved successfully--decrement our count and clean up if we have none left to resolve
          self._promiseCount -= 1;
          if (self._promiseCount === 0) {
            self._promises = undefined;
                // Fire the ready event
            self.TriggerInternal(false, oj.Events.EventType.READY, self, null, null);
          }
            // Resolve the true promise with the value we're given
          return arg;
        },
        function (error) {
            // Rejected--decrement our count and clean up if we have none left to resolve
          self._promiseCount -= 1;
          if (self._promiseCount === 0) {
            self._promises = undefined;
          }
            // Reject the promise
          return Promise.reject(error);
        });

    // Return the chain with the new promise at the end
  return this._promises;
};

/**
 * Add an xhr to a list of active xhrs
 * @private
 */
oj.Collection.prototype._addxhr = function (xhr) {
  if (xhr && xhr.abort) {
    if (this._xhrs === undefined) {
      this._xhrs = [];
    }
        // Listen to this xhr to know when to remove it
    var self = this;
    this._xhrs.push(xhr);
    xhr.done(function () {
            // Find the xhr
      var loc = self._xhrs ? self._xhrs.indexOf(xhr) : -1;
      if (loc > -1) {
                // Remove it from the list
        self._xhrs.splice(loc, 1);
      }
    });
  }
};

/**
 * Cancel all xhr requests known to this collection as still pending.  Return a promise that is resolved when all of the requests abort, and all
 * the promises tied to those requests resolve with an error
 *
 * @returns {Promise.<null>} a promise that when resolved, indicates that all the open xhr objects generated by this collection have aborted
 * @memberof oj.Collection
 * @since 2.1.0
 * @export
 */
oj.Collection.prototype.abort = function () {
  // Abort all pending XHR requests
  var self = this;
  function createPromise(index, resolve) {
    self._xhrs[index].then(function (data, status) {
      if (status === 'abort') {
                    // Remove from list
        self._xhrs.splice(index, 1);
                    // If this is the last one, resolve the promise we returned
        if (self._xhrs.length === 0) {
          self.whenReady().then(function () {
            resolve(null);
          }, function () {
            resolve(null);
          });
        }
      }
    }, function () {
      // Remove from list
      self._xhrs.splice(index, 1);
      // If this is the last one, resolve the promise we returned
      if (self._xhrs.length === 0) {
        self.whenReady().then(function () {
          resolve(null);
        }, function () {
          resolve(null);
        });
      }
    });
  }

  if (this._xhrs && this._xhrs.length > 0) {
    return new Promise(function (resolve) {
      // Count down so we can remove them
      for (var i = self._xhrs.length - 1; i >= 0; i--) {
        createPromise(i, resolve);
        self._xhrs[i].abort();
      }
    });
  }
  return Promise.resolve();
};

/**
 * @private
 */
oj.Collection.prototype._deferredAt = function (index, options) {
  var self = this;
    // If it's virtual, we need to see if this item has been fetched or not: if not, we need to fetch it + fetchSize
  var model = self._getModel(index);
  if (model === undefined) {
    return new Promise(function (resolve, reject) {
            // Go fetch
      var opts = {};
      oj.CollectionUtils.copyInto(opts, options || {});
      opts.context = self;
      opts.startIndex = index;
      opts.error = function (xhr, status, error) {
                // Handle potential errors
        reject(oj.Collection._createRejectionError(xhr, status, error, self, options, false));
      };

      opts.success = function () {
        resolve(self._getModel(index));
      };

      self._fetchInternal(opts, -1, false);
    });
  }
  return new Promise(function (resolve) {
    resolve(model);
  });
};

/**
 * Return the first model object from the collection whose client ID is the given model cid
 * @param {string} clientId Client ID (see Model cid) for which to return the model object, if found.
 * @return {oj.Model|null} First model object in the collection where model.cid = clientId. If none are found,
 * returns null.
 *
 * @throws {Error} when called on a virtual Collection if the item isn't found in memory
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.getByCid = function (clientId) {
  var models = this._getModels();
  var index;
  var foundModel = null;
  for (var i = 0; i < this._modelIndices.length; i++) {
    index = this._modelIndices[i];
    if (models[index] && clientId === models[index].cid) {
      foundModel = models[index];
      break;
    }
  }
  if (foundModel) {
    return foundModel;
  }
  if (this.IsVirtual()) {
    throw new Error('Not found locally and not supported by server for virtual collections');
  }
  return null;
};


/**
 * Return the first model object from the collection whose model id value is the given id or cid, or the id or
 * cid from a passed in model
 * Note this method will not function as expected if the id or cid is not set<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 *
 * @param {Object|string} id ID, cid, or Model (see Model id or cid) for which to return the model object, if found.
 * @param {Object=} options
 * @property {number=} fetchSize fetch size to use if the call needs to fetch more records from the server,
 *                  if virtualized.  Overrides the overall fetchSize setting<p>
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual whether it
 *                  is or not
 * @return {oj.Model|null|Promise.<oj.Model>} First model object in the collection where model.id = id or model.cid = id. If
 *                  none are found, returns null.  If deferred or virtual, return a promise passing the model
 *                  when done
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.get = function (id, options) {
  var deferred = this._getDeferred(options);
  var internalGet = this._getInternal(id, options, deferred);
  if (internalGet) {
        // Is this a deferred object?
    if ($.isFunction(internalGet.then)) {
      return this._addPromise(function () {
        return new Promise(function (resolve, reject) {
          internalGet.then(function (modInfo) {
            resolve(modInfo.m);
          },
                    function (err) {
                      reject(err);
                    });
        });
      });
    }
    if (this.IsVirtual()) {
      return this._addPromise(function () {
        return new Promise(function (resolve) {
          resolve(internalGet.m);
        });
      });
    }
    if (Object.prototype.hasOwnProperty.call(internalGet, 'm')) {
      return internalGet.m;
    }
  }
  return null;
};

/**
 * @private
 */
oj.Collection.prototype._getLocal = function (id) {
  var internalGet = this._getLocalInternal(id);
  if (internalGet) {
    return internalGet.m;
  }
  return null;
};

/**
 * @private
 */
oj.Collection.prototype._getLocalInternal = function (id) {
  var cid = id;
  var localId = id;
  if (id instanceof oj.Model) {
        // Get the cid
    cid = id.GetCid();
        // Get the id
    localId = id.GetId();
  } else if (oj.Collection._defined(id) && id.id !== undefined) {
    localId = id.id;
  }
  var foundObj = null;
  var len = this._modelIndices.length;
  var model;
  var models = this._getModels();
  var index;
  for (var i = 0; i < len; i++) {
    index = this._modelIndices[i];
    model = models[index];
    if (model !== undefined && model.Match(localId, cid)) {
      foundObj = oj.Collection._getModinfo(index, model);
      break;
    }
  }
  if (foundObj) {
    return foundObj;
  }
  return oj.Collection._getModinfo(-1, undefined);
};

/**
 * @private
 * @param {Object|null|string} id
 * @param {Object=} options
 * @param {boolean=} deferred
 * @param {boolean=} fillIn
 * @returns {Object}
 */
oj.Collection.prototype._getInternal = function (id, options, deferred, fillIn) {
  var cid = id;
  var localId = id;
  var fill = fillIn === undefined ? false : fillIn;
  if (id instanceof oj.Model) {
    // Get the cid
    cid = id.GetCid();
    // Get the id
    localId = id.GetId();
  } else if (oj.Collection._defined(id) && id.id !== undefined) {
    localId = id.id;
  }

  var foundObj = null;
  var models = this._getModels();
  var index;
  for (var i = 0; i < this._modelIndices.length; i++) {
    index = this._modelIndices[i];
    if (models[index] && models[index].Match(localId, cid)) {
      var retObj = oj.Collection._getModinfo(index, models[index]);
      foundObj = retObj;
      break;
    }
  }

  if (foundObj) {
    if (deferred) {
      return new Promise(function (resolve) {
        resolve(foundObj);
      });
    }
    return foundObj;
  }
    // If virtual, might be undefined because it needs to be fetched
  if (this.IsVirtual()) {
        // Try to fetch using start ID.  cid not supported
    if (localId === undefined && cid !== undefined) {
      return new Promise(function (resolve) {
        resolve(oj.Collection._getModinfo(-1, undefined));
      });
    }
    var self = this;
    return new Promise(function (resolve, reject) {
      var resp = function (response) {
        if (response != null) {
          var ind = self._getOffset();
          // Check that the model at index is the right one
          var model = self._getModel(ind);
          if (model !== undefined && model.Match(localId, cid)) {
            resolve(oj.Collection._getModinfo(ind, model));
          } else {
            resolve(oj.Collection._getModinfo(-1, undefined));
          }
        } else {
          resolve(oj.Collection._getModinfo(-1, undefined));
        }
      };

            // Go fetch
      var opts = {};
      oj.CollectionUtils.copyInto(opts, options || {});
      opts.context = self;
      opts.startID = localId;
      opts.error = function (xhr, status, error) {
                // Handle potential errors
        reject(oj.Collection._createRejectionError(xhr, status, error, self, options, false));
      };
      opts.success = function (collection, response) {
        resp(response);
      };

      self._fetchInternal(opts, -1, fill);
    });
  }

  var undefinedModInfo = oj.Collection._getModinfo(-1, undefined);
  if (deferred) {
    return new Promise(function (resolve) {
      resolve(undefinedModInfo);
    });
  }
  return undefinedModInfo;
};

/**
 * @private
 */
oj.Collection._getModinfo = function (index, model) {
  return { index: index, m: model };
};

/**
 * @private
 */
oj.Collection.prototype._parseImpl = function (response) {
  // Try to interpret ADFbc like controls where a collection is hanging off a property
  if (response instanceof Array) {
    return response;
  }

  if (!response) {
    return response;
  }

  // See if any of the properties contain arrays
  var prop;
  for (prop in response) { // eslint-disable-line no-restricted-syntax
    if (Object.prototype.hasOwnProperty.call(response, prop)) {
      if (response[prop] instanceof Array) {
        return response[prop];
      }
    }
  }
  return response;
};

/**
 * Optional callback to parse responses from the server.  It is called with the server's response and should return a response (possibly modified) for processing
 * @ojstatus preview
 * @type {function(Object):Object}
 * @since 1.0.0
 * @memberof oj.Collection
 * @export
 */
oj.Collection.prototype.parse = oj.Collection.prototype._parseImpl;

/**
 * Determine if actual means we are complete
 * @private
 */
oj.Collection.prototype._checkActual = function (start, count, actual) {
    // Are we at the end with what actually came back?  Then this request should satisfy the setLocalRange
  if (this._hasTotalResults() && (actual.start + actual.count >= this.totalResults)) {
    return true;
  }
  return (actual.start === start && actual.count === count);
};

/**
 *
 * Tell the collection to try and ensure that the given range is available locally.  Note that if the collection
 * is virtual, setRangeLocal may make repeated fetch calls to the server to satisfy the request, stopping only
 * when the range has been made local or the end of the records on the server is reached.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 *
 * @param {number} start starting index to make local
 * @param {number} count number of elements to make local
 * @return {Promise} a promise Object that resolves upon completion.  The promise will be passed an Object
 * containing start and count properties that represent the *actual* starting position (start), count (count),
 * and array (models) of the Models fetched, which may be fewer than what was requested.  The promise will be
 * rejected on an error and will pass the ajax status, xhr object, error, and collection, if relevant.
 * @memberof oj.Collection
 * @ojsignature  {target: "Type", value: "Promise<oj.Collection.SetRangeLocalPromise>", for: "returns"}
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.setRangeLocal = function (start, count) {
  var self = this;
  return this._addPromise(function () {
    return self._setRangeLocalInternal(start, count);
  });
};

/**
 * @typedef {Object} oj.Collection.SetRangeLocalPromise
 * @property {number} start starting index of fetched models
 * @property {number} count number of models fetched
 * @property {Array.<oj.Model>} models array of models fetched
 */

/**
 * @private
 */
oj.Collection.prototype._setRangeLocalInternal = function (start, count) {
  if (this.IsVirtual()) {
        // make sure we reconcile the length to what we think the totalresults are--if there have been any non
        // fetched changes in length we don't want to be placing things wrong
    this._resetModelsToFullLength();
  }
  var actual = this._getLocalRange(start, count);
  var self = this;
  if (this._checkActual(start, count, actual)) {
    return new Promise(function (resolve) {
      resolve(actual);
    });
  }

    // Manage the LRU - set model limit at least as high as the count we're trying to fetch
  var modelLimit = this._getModelLimit();
  if (modelLimit > -1 && modelLimit < count) {
    this.modelLimit = count;
  }
  return new Promise(function (resolve, reject) {
    self._setRangeLocalFetch(start, count, -1, { start: start, count: count }, resolve,
                             reject, true);
  });
};

/**
 * @private
 */
oj.Collection.prototype._setRangeLocalFetch = function (start, count, placement, original,
  resolve, reject, fill) {
  var self = this;
  var localStart = start;
  var resp = function () {
    var actual = self._getLocalRange(original.start, original.count);
    if (fill && self._hasTotalResults() && actual.count < original.count) {
      // The range wasn't fulfilled: try again
      var newPlacement = actual.start + actual.count;
      // Try the next block...don't repeat the request
      var newStart = localStart + (self.lastFetchCount ? self.lastFetchCount : count);
      if (newStart < self.totalResults) {
        self._setRangeLocalFetch(newStart, count, newPlacement, original, resolve, reject, fill);
      } else {
        // Can't go any further
        resolve(actual);
      }
    } else {
      resolve(actual);
    }
  };

    // Go fetch
  var limit = localStart + count;
    // Get the greater of the limit-localStart or fetchSize
  if (this[oj.Collection._FETCH_SIZE_PROP] && this[oj.Collection._FETCH_SIZE_PROP] > count) {
    limit = this[oj.Collection._FETCH_SIZE_PROP] + localStart;
  }

    // Now, to optimize, move localStart up to the first undefined model in the sequence
  var opts = null;
  if (this.IsVirtual()) {
    var newStart = this._getFirstMissingModel(localStart, limit);
    if (newStart > localStart) {
      localStart = newStart;
      // Recheck the limit
      limit = localStart + count;
      // Get the greater of the limit-localStart or fetchSize
      if (this[oj.Collection._FETCH_SIZE_PROP] && this[oj.Collection._FETCH_SIZE_PROP] > count) {
        limit = this[oj.Collection._FETCH_SIZE_PROP] + localStart;
      }
    }
    opts = { context: this, startIndex: localStart, fetchSize: limit - localStart };
  } else {
    opts = { context: this };
  }
  opts.error = function (xhr, status, error) {
        // Handle potential errors
    reject(oj.Collection._createRejectionError(xhr, status, error, self, null, false));
  };
  opts.success = function () {
    resp();
  };

  try {
    this._fetchInternal(opts, placement, placement > -1);
  } catch (e) {
        // This is OK if it's a URLError: local collection with no means of fetching: just resolve
    if (e instanceof oj.URLError) {
      var actual = self._getLocalRange(localStart, count);
      resolve(actual);
    }
  }
};

/**
 * @private
 */
oj.Collection._createRejectionError = function (xhr, status, error, collection, options,
  fireError) {
  var silent = false;
  if (options && options.silent) {
    silent = options.silent;
  }
    // To avoid duplication in many cases...
  if (fireError) {
    oj.Model._triggerError(collection, silent, options, status, error, xhr);
  }
  var err = new Error(status);
  err.xhr = xhr;
  err.error = error;
  err.collection = collection;
  err.status = status;
  return err;
};

/**
 * @private
 */
oj.Collection.prototype._getMaxLength = function (start, count) {
  var len = this._getModelsLength();
  if (len === 0) {
        // This is an exception: could be uninitialized
    return start + count;
  }
  return start + count > len ? len : start + count;
};

/**
 * Determine if every element of the given range is filled in locally
 *
 * @param {number} start starting index to check
 * @param {number} count number of elements to check
 * @return {boolean} true if all elements are local, false otherwise
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.isRangeLocal = function (start, count) {
  var localRange = this._getLocalRange(start, count);
  if (this._getModelsLength() === 0) {
    // If we have no models length, then range cannot be thought of as local unless the count
    // is zero--edge case
    return count === 0;
  }
  return start === localRange.start && (count === localRange.count ||
         start + count > this._getModelsLength());
};

/**
 * @private
 */
oj.Collection.prototype._getModelArray = function (start, count) {
  var retArr = [];
  var models = this._getModels();
  var end = start + count;
  for (var i = start; i < end; i++) {
    retArr.push(models[i]);
  }
  return retArr;
};

/**
 * For a given range of models, return the actual range which are local within that set.
 * @private
 */
oj.Collection.prototype._getLocalRange = function (start, count) {
    // Not virtual, local if there are any models
  if (!this.IsVirtual()) {
    if (this._getModelsLength() > 0) {
      if (start + count > this._getModelsLength()) {
                // Over the boundary
        var c = this._getModelsLength() - start;
        return { start: start, count: c, models: this._getModelArray(start, c) };
      }

      return { start: start, count: count, models: this._getModelArray(start, count) };
    }
    return { start: start, count: 0, models: [] };
  }
  var limit = this._getMaxLength(start, count);
    // Adjust for no totalResults
  if (!this._hasTotalResults() && limit < start + count) {
        // We don't know if it's local or not
    return { start: start,
      count: (limit - start),
      models: this._getModelArray(start, limit - start) };
  }
  if (limit === 0) {
        // There nothing here
    return { start: start, count: 0, models: [] };
  }
  var firstMissingModel = this._getFirstMissingModel(start, limit);
  if (firstMissingModel > -1) {
    return { start: start,
      count: (firstMissingModel - start),
      models: this._getModelArray(start, firstMissingModel - start) };
  }
    // Make sure start doesn't overrun the end!
  var localCount = count;
  if (start > limit) {
    localCount = 0;
  } else if (start + localCount > limit) {
    // Make sure that start+count doesn't overrun the end
    localCount = limit - start;
  }
  return { start: start, count: localCount, models: this._getModelArray(start, localCount) };
};

/**
 * Return the first model between start and limit that's undefined
 * @private
 */
oj.Collection.prototype._getFirstMissingModel = function (start, limit) {
  for (var i = start; i < limit; i++) {
    if (this._getModel(i) === undefined) {
      return i;
    }
  }
  return -1;
};

/**
 * Loads the Collection object from the data service URL. Performs a data "read."<p>
 * Events:<p>
 * <ul>
 * <b>request</b>: fired when the request to fetch is going to the server, passing the collection, xhr object, and
 * options<br>
 * <b>sync</b>: fired when the collection is fetched from the data service, passing the collection and the raw
 * response<br>
 * <b>error</b>: fired if there is an error during the fetch, passing the collection, xhr object, options<br>
 * </ul>
 * <p>
 * @param {Object=} options Options to control fetch<p>
 *                  <b>success</b>: a user callback called when the fetch has completed successfully. This makes
 *                  the fetch an asynchronous process. The callback is called passing the Collection object, raw
 *                  response, and the fetch options argument.<br>
 *                  <b>error</b>: a user callback function called if the fetch fails. The callback is called
 *                  passing the collection object, xhr, and options arguments.<br>
 *                  <b>add</b>: if set, new records retrieved from the data service will be added to those models
 *                  already in the collection. If not set, the records retrieved will be passed to the reset()
 *                  method, effectively replacing the previous contents of the collection with the new data.
 *                  Not supported for virtual/paging cases.<br>
 *                  <b>set</b>: if true, fetch will try to use the set function to try and merge the fetched models
 *                  with those already in the collection, on the client.  The default behavior is to reset the
 *                  collection before adding back in the fetched models.  This default is the reverse of
 *                  Backbone's current default.<br>
 *                  <b>startIndex</b>: numeric index with which to start fetching Models from the server.  The
 *                  page setting controls the number of Models to be fetched.  startID takes precedence over
 *                  startIndex if both are specified.  If both are specified and startID isn't supported then
 *                  startIndex will be used instead.<br>
 *                  <b>startID</b>: unique ID of the Model to start fetching from the server.  The page setting
 *                  controls the number of Models to be fetched.  Note if this is not supported by the server
 *                  then startID will be ignored.<br>
 *                  <b>since</b>: fetch records having a timestamp since the given UTC time<br>
 *                  <b>until</b>: fetch records having a timestamp up to the given UTC time<br>
 *                  <b>fetchSize</b>: use specified page size instead of collection's setting<br>
 * @return {Object} xhr ajax object, by default.  If [sync]{@link oj.Collection#sync} has been replaced, this
 * would be the value returned by the custom implementation.
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{success?: (collection: oj.Collection, response: Object, options: Object)=> void,
 *                                                  error?: (collection: oj.Collection, xhr: Object, options: Object)=> void,
 *                                                  add?: boolean, set?: boolean, startIndex?: number, startID?: Object, since?: Object,
 *                                                  until?: Object, fetchSize?: number, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.fetch = function (options) {
  var xhr = this._fetchInternal(options, -1, false);
  this._addPromise(function () {
    return Promise.resolve(xhr);
  });
  return xhr;
};

/**
 * fillIn is used to indicate that this fetch is just the result of a get() or part of an add(), etc., when virtual
 * @private
 */
oj.Collection.prototype._fetchInternal = function (options, placement, fillIn) {
  function doReset(collection, opt, fill) {
    if (!collection.IsVirtual()) {
      // If we're not doing a "fetch add", delete all the current models
      if (!opt.add && !opt.useset) {
        // Reset with internal model
        collection.reset(null, { silent: true });
      }
    } else if (!fill) {
      // If we're not infilling based on an at, get, etc., delete all the current local models
      collection._resetModelsToFullLength();
    }
  }

  var opt = options || {};
  var success = opt.success;
  var self;
  var errFunc = opt.error;

  if (opt.set) {
    opt.useset = !!opt.set;
  }

  // Set up the parsing option
  if (opt.parse === undefined) {
    opt.parse = true;
  }
  self = this;

  opt.error = function (xhr, status, error) {
    oj.Model._triggerError(self, false, options, status, error, xhr);
    if (errFunc) {
      errFunc.call(oj.Model.GetContext(options, self), xhr, status, error);
    }
  };

  opt.success = function (response) {
    // Run the entire returned dataset through the collection's parser (either default no-op or user specified)
    var data;
    try {
      data = self.parse(response, options);
    } catch (e) {
      oj.Collection._reportError(self, e, opt.error, options);
      return;
    }

    // Pull any virtualization properties out of the response
    self._setPagingReturnValues(response, options, data, fillIn);

    var dataList = null;

    var manageLRU = false;
    var locPlace = placement;
    if (!opt.add && !self.model) {
      // We're not doing a "fetch add" and we don't have a model set on the collection
      if (!fillIn) {
        if (self.IsVirtual()) {
          // Virtual case only
          // Clean out the collection
          doReset(self, opt, fillIn);

          // Check for passed in offset
          if (placement === -1) {
            manageLRU = true;
            locPlace = self._getOffset();
          }
          // Put the new data in
          dataList = self._putDataIntoCollection(data, locPlace, manageLRU);
        } else if (opt.useset) {
          // New backbone option to not reset
          self._setInternal(data, false, opt, false);
        } else {
          // Replace the data in the collection with a new set (non virtual)
          self.reset(data, { silent: true });
        }
      }
    } else {
      // We have a model and/or we're doing a "fetch add"
      // Clean out the old models if we're not "adding" or infilling for virtual
      doReset(self, opt, fillIn);

      // Parse each returned model (if appropriate), and put it into the collection, either from the
      // zeroth offset if non-virtual or using the appropriate offset if virtual
      try {
        // Check for passed in offset
        if (placement === -1) {
          manageLRU = true;
          locPlace = self._getOffset();
        }
        dataList = self._fillInCollectionWithParsedData(data, locPlace, manageLRU, opt);
      } catch (e) {
        oj.Collection._reportError(self, e, opt.error, options);
        return;
      }
    }
    if (self.IsVirtual()) {
      // Take in the number of records actually fetched
      if (dataList) {
        self.lastFetchCount = dataList.length;
      }
    }

    // Fire the sync event
    var silent = !!opt.silent;
    self.TriggerInternal(silent, oj.Events.EventType.SYNC, self, response, opt);
    // Call the caller's success callback, if specified
    if (success) {
      success.call(oj.Model.GetContext(options, self), self, response, opt);
    }
  };
  // Make the actual fetch call using ajax, etc.
  return this._fetchCall(opt);
};

/**
 * Puts server data into an empty virtual collection using a "silent add"
 * @private
 */
oj.Collection.prototype._putDataIntoCollection = function (data, placement, manageLRU) {
  var dataList;

  if (data) {
    dataList = (data instanceof Array) ? data : [data];

    var addOpt = {};
    // Only manage the LRU if we're not trying to achieve a range
    if (manageLRU) {
      this._manageLRU(dataList.length);
    }
    var insertPos = placement;
    var virtual = this.IsVirtual();
    var prevItem = null;
    for (var i = 0; i < dataList.length; i += 1) {
      if (virtual) {
        addOpt = { at: insertPos };
        prevItem = this._atInternal(insertPos, null, true, false);
      }
      // Don't fire add events
      addOpt.silent = true;
      this._addInternal(dataList[i], addOpt, true, false);
      // If virtual, make sure the item was really added where we thought--in other words, what's there now
      // shouldn't match what was there otherwise could be duplicate id and don't increment counter
      if (this._atInternal(insertPos, null, true, false) !== prevItem) {
        insertPos += 1;
      }
    }
  }
  return dataList;
};

/**
 * Parse each model returned, if appropriate, and push them into the (empty) collection with appropriate offsets
 * if virtual
 * @private
 */
oj.Collection.prototype._fillInCollectionWithParsedData = function (data, placement, manageLRU,
  options) {
  var opt = options || {};
  var parse = opt.parse;
  var modelInstance = oj.Collection._createModel(this);
  var dataList = null;

  if (data) {
    dataList = (data instanceof Array) ? data : [data];

    var addOpt = {};
    var parsedModel;

    // Only manage the LRU if we're not trying to achieve a range
    if (manageLRU) {
      this._manageLRU(dataList.length);
    }
    var virtual = this.IsVirtual();
    var i;
    if (opt.useset && !virtual) {
      // New backbone option
      for (i = 0; i < dataList.length; i += 1) {
        if (modelInstance && parse) {
          parsedModel = modelInstance.parse(dataList[i]);
        } else {
          parsedModel = dataList[i];
        }

        dataList[i] = parsedModel;
      }
      this._setInternal(dataList, false, opt, false);
    } else {
      var prevItem = null;
      var insertPos = placement;
      for (i = 0; i < dataList.length; i += 1) {
        if (modelInstance && parse) {
          parsedModel = modelInstance.parse(dataList[i]);
        } else {
          parsedModel = dataList[i];
        }

        if (virtual) {
          addOpt = { at: insertPos };
          prevItem = this._atInternal(insertPos, addOpt, true, false);
        }
                // Don't fire add events
        addOpt.silent = true;
        this._addInternal(parsedModel, addOpt, true, false);
                // If virtual, make sure the item was really added where we thought--in other words, what's there
                // now shouldn't match what was there otherwise could be duplicate id and don't increment counter
        if (this._atInternal(insertPos, null, true, false) !== prevItem) {
          insertPos += 1;
        }
      }
    }
  }
  return dataList;
};

/**
 * @private
 */
oj.Collection._reportError = function (collection, e, errorFunc, options) {
  oj.Logger.error(e.toString());
  if (errorFunc) {
    errorFunc.call(oj.Model.GetContext(options, collection), collection, e, options);
  }
};

/**
 * Used in virtualization to conduct server-based searches: returns list of fetched models via a promise but does
 * not add them to the collection model list
 * @private
 */
oj.Collection.prototype._fetchOnly = function (options) {
  var opt = options || {};
  var success = opt.success;
  var parsedModel;
  var self;

  if (opt.parse === undefined) {
    opt.parse = true;
  }
  self = this;

  opt.success = function (response) {
    var i;
    var modelInstance;
    var dataList = null;
    var fetchedModels = [];

    var data;
    try {
      data = self.parse(response, options);
    } catch (e) {
      oj.Collection._reportError(self, e, opt.error, options);
      return;
    }

    if (!opt.add && !self.model) {
      dataList = (data instanceof Array) ? data : [data];
    } else {
      modelInstance = oj.Collection._createModel(self);

      if (data) {
        dataList = (data instanceof Array) ? data : [data];

        for (i = 0; i < dataList.length; i += 1) {
          if (modelInstance && opt.parse) {
            try {
              parsedModel = modelInstance.parse(dataList[i]);
            } catch (e) {
              oj.Collection._reportError(self, e, opt.error, options);
              return;
            }
          } else {
            parsedModel = dataList[i];
          }
          fetchedModels.push(self._newModel(parsedModel));
        }
      }
    }

    self.TriggerInternal(false, oj.Events.EventType.SYNC, self, response, opt);
    if (success) {
      success.call(oj.Model.GetContext(options, self), self, fetchedModels, opt);
    }
  };
  return this._fetchCall(opt);
};

/**
 * @private
 */
oj.Collection.prototype._fetchCall = function (opt) {
  try {
    return oj.Model._internalSync('read', this, opt);
  } catch (e) {
         // Some kind of error: trigger an error event
    oj.Model._triggerError(this, false, opt, null, e, null);
    throw e;
  }
};

/**
 * @private
 */
oj.Collection.prototype._resetModelsToFullLength = function () {
  var totalResults = this.totalResults;
  if (totalResults !== undefined && this._getModelsLength() !== totalResults) {
        // Make sure to set up the array if the length changes (i.e., from 0 to totalResults--need to preallocate)
    this._setModels(new Array(totalResults), true);
    this._resetLRU();
    this._setLength();
    return true;
  }
  return false;
};

/**
 * @private
 */
oj.Collection.prototype._getFetchSize = function (options) {
  var opt = options || {};
  return opt[oj.Collection._FETCH_SIZE_PROP] || this[oj.Collection._FETCH_SIZE_PROP];
};

/**
 * Are we doing virtualization/paging?
 * @protected
 */
oj.Collection.prototype.IsVirtual = function () {
  return this._getFetchSize(null) > -1;
};

/**
 * @private
 */
oj.Collection.prototype._getReturnProperty = function (customObj, response, property,
  optionValue, defaultValue) {
  var value = parseInt(oj.Collection._getProp(customObj, response, property), 10);
  if (value === undefined || value === null || isNaN(value)) {
        // use fetchsize
    return optionValue || defaultValue;
  }
  return value;
};

/**
 * @private
 */
oj.Collection.prototype._cleanTotalResults = function (totalResults) {
    // In case server (incorrectly) passes back a -1, treat it as undefined
  if (totalResults === -1) {
    return undefined;
  }
  return totalResults;
};

/**
 * Parse out some of the paging return values we might expect in a virtual response
 * @private
 */
oj.Collection.prototype._setPagingReturnValues = function (response, options, data, fillIn) {
  var customObj = {};
  // See if there's a custom call out
  if (this.customPagingOptions) {
    customObj = this.customPagingOptions.call(this, response);
    if (!customObj) {
      customObj = {};
    }
  }
  // What limit was actually used to generate this response?
  var opt = options || {};
  this.lastFetchSize = this._getReturnProperty(customObj, response, 'limit', opt.fetchSize,
                                                    this.fetchSize);

  // What offset was actually used to generate this response?
  this.offset = this._getReturnProperty(customObj, response, 'offset', opt.startIndex, 0);

  // How many records actually came back?
  this.lastFetchCount = this._getReturnProperty(customObj, response, 'count', this.lastFetchCount,
                                                     this.lastFetchCount);

  // What is the total number of records possible for this collection?
  this.totalResults = this._cleanTotalResults(this._getReturnProperty(customObj, response, 'totalResults',
                                                   this.totalResults, this.totalResults));

  // Is there more?
  this.hasMore = this._getHasMore(oj.Collection._getProp(customObj, response, 'hasMore'),
                                       this.offset, this.lastFetchSize, this.totalResults);

  // Adjust total results to account for the case where the server tells us there's no more data, and
  // totalResults wasn't set by the server...but don't do it for simple gets/adds
  if (!fillIn) {
        // We want to know if the server *actually* returned values for these things, not if they defaulted above
    var totalResultsReturned = this._cleanTotalResults(parseInt(oj.Collection._getProp(customObj,
      response, 'totalResults'), 10));
    var lastFetchCountReturned = parseInt(oj.Collection._getProp(customObj, response, 'count'),
      10);
    this.totalResults = this._adjustTotalResults(totalResultsReturned, this.hasMore,
      this.offset, lastFetchCountReturned, data && Array.isArray(data) ? data.length : 0);
  }

  // Was fetchSize set?  If not, set it to limit
  if (!this.IsVirtual() && this.totalResults && this.totalResults !== this.lastFetchCount &&
        this.lastFetchSize) {
    this.setFetchSize(this.lastFetchSize);
  }
};

/**
 * @private
 */
oj.Collection.prototype._adjustTotalResults = function (totalResultsReturned, hasMore, offset,
  lastFetchCount, dataLength) {
    // Fix for : if hasMore is false, and totalResults wasn't set by the server, we should set it to
    // the last thing fetched here.  There is no more.
  if (!hasMore) {
        // No more data, per server
    if (isNaN(totalResultsReturned)) {
            // TotalResults was not returned by the server here
            // If lastFetchCount was set, use that as the final total.  Otherwise, use the length of data passed
            // in...all + offset
      var count = !isNaN(lastFetchCount) ? lastFetchCount : dataLength;
      return count + offset;
    }
  }
  return this.totalResults;
};

/**
 * @private
 */
oj.Collection.prototype._getHasMore = function (hasMore, offset, lastFetchSize, totalResults) {
  if (oj.Collection._defined(hasMore)) {
    return hasMore;
  }
    // Special case: return true if totalResults not defined
  if (totalResults === undefined || totalResults === null) {
    return true;
  }
    // Not there: figure it out.  It's true unless we're walking off the end
  return !((offset + lastFetchSize > totalResults));
};

/**
 * @private
 */
oj.Collection._getProp = function (custom, response, prop) {
  if (Object.prototype.hasOwnProperty.call(custom, prop)) {
    return custom[prop];
  }
  return response ? response[prop] : undefined;
};

/**
 * @private
 */
oj.Collection.prototype._getOffset = function () {
  return (oj.Collection._defined(this.offset) ? this.offset : 0);
};

/**
 * Creates a new model, saves it to the data service, and adds it on to the collection.<p>
 * Events:<p>
 * <ul>
 * <b>add</b>: fired when the model is added, passing the collection, model added, and options, if not silent<br>
 * <b>alladded</b>: fired after all models have been added, passing the collection, array of models added, and
 * options<br>
 * <b>request:</b>: fired when the request to save is going to the server<br>
 * <b>sync:</b>: fired when the model is saved to the data service, passing the model and the raw response<br>
 * </ul>
 * @param {Object=} attributes Set of attribute/value pairs with which to initialize the new model object, or a new
 * Model object
 * @param {Object=} options Options to control save (see [save]{@link oj.Model#save}), or add (see
 * [add]{@link oj.Collection#add}).  Plus:<p>
 *                          <b>deferred</b>: if true, return a promise as though this collection were virtual
 *                          whether it is or not<br>
 * @return {oj.Model|boolean|Promise.<oj.Model>} new model or false if validation failed.  If virtual, returns a promise that
 * resolves with the new model
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{silent?: boolean, at?: number, merge?: boolean, sort?: boolean, force?: boolean, deferred?: boolean,
 *               [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.create = function (attributes, options) {
  var self = this;
  var opt = options || {};
  var deferred = this._getDeferred(opt);

  function doSave(collection, newModel, validate, localOpt) {
    newModel.save(attributes instanceof oj.Model ? null : attributes, localOpt);

    return newModel;
  }

  function doAdd(newModel, addOpts) {
    if (opt.wait) {
            // Don't do adding now
      if (self.IsVirtual() || deferred) {
        return self._addPromise(function () {
          return Promise.resolve(undefined);
        });
      }
      return null;
    }
    return self.add(newModel, addOpts);
  }

    // Save the user's context and callback, if any
  var newModel = this._newModel(attributes, true, opt, false);
  var callback = opt.success;
  var context = opt.context;
  var validate = opt.validate;
  opt.context = this;
  opt.success = function (model, resp, successOpts) {
    // Make sure we pass xhr
    if (successOpts.xhr) {
      opt.xhr = successOpts.xhr;
    }
    if (opt.wait) {
      // Trigger delayed add events
      self.add(newModel, opt);
    }
    if (callback) {
      callback.call(context != null ? context : self, model, resp, opt);
    }
  };

  // Did validation pass?
  if (newModel == null) {
    return false;
  }

  // Force a save in case user has set value of idAttribute on the new Model
  opt.forceNew = newModel.GetId() != null;

  var addOpts = oj.Model._copyOptions(opt);

  newModel.SetCollection(this);
  if (deferred || this.IsVirtual()) {
    return new Promise(function (resolve) {
      addOpts.merge = true;
      addOpts.deferred = true;
      doAdd(newModel, addOpts)
                    .then(function () {
                      opt.success = function (model, resp, successOpts) {
                                    // Make sure we pass xhr
                        if (successOpts.xhr) {
                          opt.xhr = successOpts.xhr;
                        }
                        if (opt.wait) {
                                        // Trigger delayed add event
                                        // Force the add if virtual because we know it was successfully saved: we
                                        // don't want to go back to the server to check
                          if (self.IsVirtual()) {
                            addOpts.force = true;
                          }
                          self.add(newModel, addOpts).then(function () {
                            if (callback) {
                              callback.call(context != null ? context : self, model, resp, opt);
                            }
                            resolve(model);
                          });
                        } else {
                          if (callback) {
                            callback.call(context != null ? context : self, model, resp, opt);
                          }
                          resolve(model);
                        }
                      };
                      var model = doSave(self, newModel, validate, opt);
                                // make sure that success is called first if successful...promise resolved
                                // second
                      if (!model) {
                                    // Failed: make sure we resolve the promise.  Otherwise promise will
                                    // be resolved by success call, above
                        resolve(model);
                      }
                    });
    });
  }

  addOpts.merge = true;
  doAdd(newModel, addOpts);
  var model = doSave(this, newModel, validate, opt);
  return model;
};

/**
 * Return a list of all the values of attr found in the collection
 *
 * @param {string} attr attribute to return
 * @return {Array.<Object>} array of values of attr
 *
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.pluck = function (attr) {
  var arr = [];
  var i;

  this._throwErrIfVirtual('pluck');

  for (i = 0; i < this._getLength(); i += 1) {
    arr.push(this.at(i).get(attr));
  }
  return arr;
};

/**
 * Return an array of models that contain the given attribute/value pairs.  Note that this function, along with
 * findWhere, expects server-resolved filtering to return *all* models that meet the criteria, even in virtual
 * cases.  The fetchSize will be set to the value of totalResults for this call to indicate that all should
 * be returned.<br>
 * Events: for events, if virtual, see [fetch]{@link oj.Collection#fetch}
 *
 * @param {Object|Array.<Object>} attrs attribute/value pairs to find.  The attribute/value pairs are ANDed together.  If
 *                             attrs is an array of attribute/value pairs, then these are ORed together
 *                             If the value is an object (or an array of objects, in which case the single
 *                             attribute must meet all of the value/comparator conditions), then if it has both
 *                             'value' and 'comparator' parameters these will be interpreted as
 *                             expressions needing custom commparisons.  The comparator value may either be a
 *                             string or a comparator callback function.  Strings are only valid where the
 *                             filtering is sent back to the data service (virtual collections).  In the case of
 *                             a comparator function, the function always takes the signature function(model,
 *                             attr, value), and for non-virtual collections, is called for each
 *                             Model in the collection with the associated attribute and value.  The function
 *                             should return true if the model meets the attribute/value condition, and false if
 *                             not.  For cases where the filtering is to be done on the server, the function will
 *                             be called once per attr/value pair with a null model, and the function should
 *                             return the string to pass as the comparison in the expression for the filtering
 *                             parameter in the URL sent back to the server.  Note that the array of value object
 *                             case is really only meaningful for server-evaluated filters where a complex
 *                             construction on a single attribute might be needed (e.g., x>v1 && x <=v2)
 *                             For example:<p>
 *                             {Dept:53,Name:'Smith'}<br>
 *                             will return an array of models that have a Dept=53 and a Name=Smith, or, for
 *                             server-filtered collections, a ?q=Dept=53+Name=Smith parameter will be sent with
 *                             the URL.<p>
 *                             [{Dept:53},{Dept:90}]<br>
 *                             will return all models that have a Dept of 53 or 90.  Or, ?q=Dept=53,Dept=90 will
 *                             be sent to the server.<p>
 *                             {Dept:{value:53,comparator:function(model, attr, value) { return model.get(attr) !==
 *                             value;}}}<br>
 *                             will return all models that do not have a Dept of 53.<p>
 *                             {Dept:{value:53,comparator:'<>'}}<br>
 *                             For server-evaluated filters, a parameter ?q=Dept<>53 will be sent with the URL.
 *                             This form is an error on locally-evaluated colleection filters<p>
 *                             {Dept:{value:53,comparator:function(model, attr, value) { return "<>";}}}<br>
 *                             expresses the same thing for server-evaluated filters<p>
 *                             {Dept:[{value:53,comparator:'<'},{value:90,comparator:'<'}]}<br>
 *                             For server-evaluated filters, a parameter ?q=Dept>53+Dept<93 will be sent to the
 *                             server<p>
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual
 * whether it is or not<p>
 *
 * @return {Array.<oj.Model>|Promise.<Array>} array of models.  If virtual or deferred, a promise that resolves with the returned
 * array from the server
 * @ojsignature  {target: "Type", value: "Promise<Array<oj.Model>>|Array<oj.Model>", for: "returns"}
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.where = function (attrs, options) {
  return this._handlePromise(this._whereInternal(attrs, options));
};

/**
 * @private
 */
oj.Collection.prototype._whereInternal = function (attrs, options) {
  var opt = options || {};
  var deferred = this._getDeferred(opt);
  var self = this;
  if (this.IsVirtual()) {
    return new Promise(function (resolve, reject) {
      var success = function (collection, fetchedModels) {
        resolve(fetchedModels);
      };
            // Send the attributes for a server-based filter; also indicate that we need *all* the attributes.
            // In the standard REST URL construction this is accomplished by leaving off fetchSize/start indices,
            // etc.
      var opts = { query: attrs,
        all: true,
        success: success,
        error: function (xhr, status, error) {
          reject(oj.Collection._createRejectionError(xhr, status, error, self, opt, true));
        } };
      self._fetchOnly(opts);
    });
  }

  var arr = [];
  var i;
  var m;
  for (i = 0; i < this._getLength(); i += 1) {
    m = this.at(i);
    if (m.Contains(attrs)) {
      arr.push(m);
    }
  }
  if (deferred) {
    return new Promise(function (resolve) {
      resolve(arr);
    });
  }
  return arr;
};

/**
 * Return a collection of models that contain the given attribute/value pairs.
 * Note that this returns a non-virtual collection with all the models returned by the server,
 * even if the original collection is virtual.  Virtual collections doing filtering on the server should return
 * all models that meet
 * the critera.  See [where]{@link oj.Collection#where}
 * See [where]{@link oj.Collection#where} for complete documentation of events and parameters
 * @param {Object|Array.<Object>} attrs attribute/value pairs to find.  The attribute/value pairs are ANDed together.  If
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual
 * whether it is or not<p>
 *
 * @return {oj.Collection|Promise.<oj.Collection>} A collection containing models with given attribute/value pairs.  If virtual or
 * deferred, a promise that resolves with the collection returned by the server
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.whereToCollection = function (attrs, options) {
  var opt = options || {};
  var deferred = this._getDeferred(opt);
  var self = this;
  if (this.IsVirtual() || deferred) {
    return self._addPromise(function () {
      return new Promise(function (resolve, reject) {
        return self._whereInternal(attrs, opt).then(function (models) {
          var collection = self._makeNewCollection(models);
          resolve(collection);
        }, function (err) {
          reject(err);
        });
      });
    });
  }

  var models = this._whereInternal(attrs, opt);
  var newCollection = this._makeNewCollection(models);
  newCollection[oj.Collection._FETCH_SIZE_PROP] = -1;
  newCollection._setLength();
  return newCollection;
};

/**
 * @private
 */
oj.Collection.prototype._makeNewCollection = function (models) {
  var collection = this._cloneInternal(false);
  collection._setModels(models, false);
  collection._resetLRU();
  collection._setLength();
  return collection;
};

/**
 * @private
 */
oj.Collection.prototype._throwErrIfVirtual = function (func) {
  if (this.IsVirtual()) {
    throw new Error(func + ' not valid on a virtual Collection');
  }
};

/**
 * Return an array whose entries are determined by the results of calling the passed iterator function.  The
 * iterator will be called for each model in the collection
 *
 * @param {function(oj.Model):Object} iterator function to determine the mapped value for each model
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {Array.<Object>} array of values determined by the return value of calls made to iterator for each model
 *
 * @throws {Error} when called on a virtual Collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.map = function (iterator, context) {
  var retArr = [];
  var value;

  this._throwErrIfVirtual('map');

  this._getModels().forEach(function (model) {
    value = iterator.call(context || this, model);
    retArr.push(value);
  });
  return retArr;
};

/**
 * @desc Iterates over the models in the collection and calls the given iterator function
 *
 * @param {function(oj.Model)} iterator function to call for each model
 * @param {Object=} context context with which to make the calls on iterator
 * @return {undefined}
 *
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.each = function (iterator, context) {
  this._throwErrIfVirtual('each');

  this._getModels().forEach(iterator, context);
};

/**
 * Return the length of the collection
 * @returns {number} length of the collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.size = function () {
  return this._getLength();
};

/**
 * Return the models sorted determined by the iterator function (or property, if a string value).  If a function,
 * the function should return the attribute by which to sort.
 *
 * @param {string|function(oj.Model):Object} iterator method called or property used to get the attribute to sort by.
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {Array.<oj.Model>} models sorted using iterator
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.sortBy = function (iterator, context) {
  var retArr = [];
  var self;

  this._throwErrIfVirtual('sortBy');

  this._getModels().forEach(function (model) {
    retArr.push(model);
  });
  self = this;

  retArr.sort(function (a, b) {
    var keyA;
    var keyB;

    if ($.isFunction(iterator)) {
                // "sortBy" comparator option
      keyA = iterator.call(context || self, a);
      keyB = iterator.call(context || self, b);
      return oj.Collection._compareKeys(keyA, keyB, self.sortDirection);
    }
            // String option
    keyA = a.get(iterator);
    keyB = b.get(iterator);
    return oj.Collection._compareKeys(keyA, keyB, self.sortDirection);
  });

  return retArr;
};

/**
 * @desc Return the collection with models grouped into sets determined by the iterator function (or property, if
 * a string value)
 *
 * @param {string|function(oj.Model):Object} iterator method called or property (if a string) used to get the group key
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {Object} models grouped into sets
 *
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.groupBy = function (iterator, context) {
  var retObj = {};
  var groupVal;

  this._throwErrIfVirtual('groupBy');

  this._getModels().forEach(function (model) {
    if ($.isFunction(iterator)) {
      groupVal = iterator.call(context || this, model);
    } else {
      groupVal = model.get(iterator);
    }
    if (retObj[groupVal] === undefined) {
      retObj[groupVal] = [];
    }
    retObj[groupVal].push(model);
  }, this);
  return retObj;
};

/**
 * @desc Return an object with models as values for their properties determined by the iterator function or
 * property string
 *
 * @param {string|function(oj.Model)} iterator method called or property (if a string) used to get the index attribute
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {Object} models listed as property values where the properties are the values returned by iterator or
 * the attribute value given by the iterator string
 *
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.indexBy = function (iterator, context) {
  var retObj = {};
  var index;

  this._throwErrIfVirtual('indexBy');

  this._getModels().forEach(function (model) {
    if ($.isFunction(iterator)) {
      index = iterator.call(context || this, model);
    } else {
      index = model.get(iterator);
    }
    retObj[index] = model;
  }, this);
  return retObj;
};

/**
 * Return the "minimum" model in the collection, as determined by calls to iterator.  The return value of
 * iterator (called with a model passed in) will be compared against the current minimum
 *
 * @param {function(Object)} iterator function to determine a model's value for checking for the minimum
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {Object} "Minimum" model in the collection
 *
 * @throws {Error} when called on a virtual Collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.min = function (iterator, context) {
  var minModel = {};
  var minModelValue;
  var currValue;

  this._throwErrIfVirtual('min');

  if (this._getModelsLength() === 0) {
    return null;
  }
    // Get vals started
  minModel = this._getModel(0);
  minModelValue = iterator.call(context || this, this._getModel(0));

  this._getModels().forEach(function (model, i) {
    if (i >= 1) {
      currValue = iterator.call(context || this, model);
      if (currValue < minModelValue) {
        minModel = model;
        minModelValue = currValue;
      }
    }
  }, this);
  return minModel;
};

/**
 * Return the "maximum" model in the collection, as determined by calls to iterator.  The return value of
 * iterator (called with a model passed in) will be compared against the current maximum
 *
 * @param {function(oj.Model):Object} iterator function to determine a model's value for checking for the maximum
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {Object} "Maximum" model in the collection
 *
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.max = function (iterator, context) {
  var maxModel = {};
  var maxModelValue;
  var currValue;

  this._throwErrIfVirtual('max');
  if (this._getModelsLength() === 0) {
    return null;
  }
  // Get vals started
  maxModel = this._getModel(0);
  maxModelValue = iterator.call(context, this._getModel(0));

  this._getModels().forEach(function (model, i) {
    if (i >= 1) {
      currValue = iterator.call(context || this, model);
      if (currValue > maxModelValue) {
        maxModel = model;
        maxModelValue = currValue;
      }
    }
  }, this);
  return maxModel;
};

/**
 * @export
 * @desc Return an array of models that cause passed-in iterator to return true
 * @memberof oj.Collection
 *
 * @param {function(oj.Model)} iterator function to determine if a model should be included or not.  Should return
 * true or false
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {Array.<oj.Model>} array of models that caused iterator to return true
 *
 * @throws {Error} when called on a virtual Collection
 * @since 1.0.0
 */
oj.Collection.prototype.filter = function (iterator, context) {
  var retArr = [];

  this._throwErrIfVirtual('filter');

  this._getModels().forEach(function (model) {
    if (iterator.call(context || this, model)) {
      retArr.push(model);
    }
  });
  return retArr;
};

/**
 * Return an array of models minus those passed in as arguments
 * @param {...oj.Model} var_args models models to remove from the returned array
 * @returns {Array.<oj.Model>} array of models from the collection minus those passed in to models
 *
 * @throws {Error} when called on a virtual Collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.without = function (var_args) { // eslint-disable-line
  var retArr = [];
  var j;
  var id;
  var cid;
  var add;

  this._throwErrIfVirtual('without');

  var model;
    // Test each model in the collection
  for (var i = 0; i < this._getModels().length; i++) {
    add = true;
    model = this._getModel(i);
    for (j = 0; j < arguments.length; j += 1) {
            // Get the cid
      cid = arguments[j].GetCid();
            // Get the id
      id = arguments[j].GetId();
      if (model.Match(id, cid)) {
                // If it's found, don't return it--we're "subtracting" those from the return value, which starts
                // as all models in the collection
        add = false;
        break;
      }
    }
    if (add) {
      retArr.push(model);
    }
  }
  return retArr;
};

/**
 * Return an array of models in the collection but not passed in the array arguments
 * @param {...Array.<oj.Model>} var_args models arrays of models to check against the collection
 * @returns {Array.<oj.Model>} array of models from the collection not passed in as arguments
 *
 * @throws {Error} when called on a virtual Collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.difference = function (var_args) { // eslint-disable-line
  var retArr = [];
  var j;
  var k;
  var id;
  var cid;
  var add;

  this._throwErrIfVirtual('difference');

  var model;
  for (var i = 0; i < this._getModels().length; i++) {
    add = true;
    model = this._getModel(i);
    for (j = 0; j < arguments.length; j += 1) {
            // Each argument is assumed to be an array of oj.Models
      for (k = 0; k < arguments[j].length; k++) {
                // Get the cid
        cid = arguments[j][k].GetCid();
                // Get the id
        id = arguments[j][k].GetId();
        if (model.Match(id, cid)) {
          add = false;
          break;
        }
      }
      if (!add) {
                // We found this model somewhere in the arguments--we're not going to add it so get out
        break;
      }
    }
    if (add) {
      retArr.push(model);
    }
  }
  return retArr;
};

/**
 * Determine if the collection has any models
 *
 * @returns {boolean} true if collection is empty
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.isEmpty = function () {
  return this._getLength() === 0;
};

/**
 * @export
 * @desc Return true if any of the models in the collection pass the test made by calling the iterator function
 * parameter
 * @memberof oj.Collection
 *
 * @param {function(Object)} iterator function called with each model to determine if it "passes".  The function
 * should return true or false.
 * @param {Object=} context context with which to make the calls on iterator
 * @returns {boolean} true if any of the models cause the iterator function to return true
 *
 * @throws {Error} when called on a virtual collection
 * @since 1.0.0
 */
oj.Collection.prototype.any = function (iterator, context) {
  this._throwErrIfVirtual('any');

  var model;
  for (var i = 0; i < this._getModelsLength(); i += 1) {
    model = this._getModel(i);
    if (iterator.call(context || this, model)) {
      return true;
    }
  }
  return false;
};

/**
 * @export
 * @desc A version of where that only returns the first element found<br>
 * Events: for events, if virtual, see [fetch]{@link oj.Collection#fetch}<br>
 * @memberof oj.Collection
 * @param {Object|Array.<Object>} attrs attribute/value pairs to find.
 * See [where]{@link oj.Collection#where} for more details and examples.
 * @param {Object=} options <b>deferred</b>: if true, return a promise as though this collection were virtual
 * whether it is or not<p>
 *
 * @returns {oj.Model|Promise.<oj.Model>} first model found with the attribute/value pairs.  If virtual or deferred, a promise
 * that resolves with the returned model from the server
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{deferred?: boolean, [propName: string]: any}", for: "options"}
 */
oj.Collection.prototype.findWhere = function (attrs, options) {
  var deferred = this._getDeferred(options);
  var self = this;
  if (this.IsVirtual() || deferred) {
    return this._addPromise(function () {
      return new Promise(function (resolve) {
        self._whereInternal(attrs, options).then(function (modelList) {
          if (modelList && modelList.length > 0) {
            resolve(modelList[0]);
          }
          resolve(null);
        });
      });
    });
  }

  var arr = this._whereInternal(attrs, options);
  if (arr.length > 0) {
    return arr[0];
  }
  return null;
};

/**
 * Return a shallow copy of the models from start to end (if specified), in an array<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 *
 * @param {number} start model to start the return array with
 * @param {number=} end model to end the return array with, if specified (not inclusive).  If not, returns to the
 * end of the collection
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual
 * whether it is or not
 * @return {Array.<oj.Model>|Promise} array of model objects from start to end, or a promise that resolves specifying the
 * returned array when done
 * @memberof oj.Collection
 * @ojsignature  {target: "Type", value: "Promise<Array<oj.Model>>|Array<oj.Model>", for: "returns"}
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.slice = function (start, end, options) {
  var deferred = this._getDeferred(options);
  var ret = [];
  var i;
  var localEnd = end;
  if (localEnd === undefined) {
    if (this.IsVirtual() && !this._hasTotalResults()) {
      // We can't set the end: throw an error
      throw new Error('End must be set for virtual collections with no totalResults');
    }
    localEnd = this._getModelsLength();
  }

  if (deferred || this.IsVirtual()) {
    var self = this;
    return this._addPromise(function () {
      // Loop using deferred
      return self.IterativeAt(start, localEnd);
    });
  }

  for (i = start; i < localEnd; i += 1) {
    ret.push(this._getModel(i));
  }
  return ret;
};

/**
 * Update the collection with a model or models.  Depending on the options, new models will be added, existing
 * models will be merged, and unspecified models will be removed.
 * The model cid is used to determine whether a given model exists or not.<br>
 * May fire events as specified by [add]{@link oj.Collection#add} or [remove]{@link oj.Collection#remove},
 * depending on the options.
 *
 * @param {Object} models an array of or single model with which to update the collection.  If models contains
 * actual oj.Model objects, then any custom 'parse' function set on the collection needs to take that into
 * account and be prepared to handle an array of oj.Model.
 * @param {Object=} options <b>add</b>:false stops the addition of new models<br>
 *                          <b>remove</b>: false stops the removal of missing models<br>
 *                          <b>merge</b>: false prevents the merging of existing models<br>
 *                          <b>silent</b>: true prevents notifications on adds, removes, etc.<br>
 *                          <b>deferred</b>: if true, return a promise as though this collection were virtual
 *                          whether it is or not
 *
 * @returns {Promise|null} if deferred or virtual, return a promise that resolves when the set has completed
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{add?: boolean, remove?: boolean, merge?: boolean, silent?: boolean, deferred?: boolean,
 *               [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.set = function (models, options) {
  var deferred = this._getDeferred(options);
  return this._setInternal(models, true, options, deferred || this.IsVirtual());
};

oj.Collection._removeAfterSet = function (collection, models, remove, foundModels, options) {
    // Now remove models that weren't found
    // get an array of all models

    // Can't avoid looping over everything because we *have* to clean up even unfetched models, in order to fire
    // events, etc.
  if (remove) {
    for (var i = models.length - 1; i >= 0; i -= 1) {
      if (foundModels.indexOf(i) === -1) {
        collection._removeInternal(models[i], i, options);
      }
    }
  }
};

/**
 * Swap two models, and indicate if anything was actually swapped
 * @private
 */
oj.Collection.prototype._swapModels = function (oldIndex, newIndex, remove, add) {
  if (this._hasComparator() || !remove || !add) {
    return { index: oldIndex, swapped: false };
  }
    // Make sure in range
  var len = this._getModelsLength();
  if (oldIndex >= len || newIndex >= len) {
    return { index: oldIndex, swapped: false };
  }
    // Swap
  var oldModel = this._getModel(oldIndex);
  var newModel = this._getModel(newIndex);
    // this._getModels()[oldIndex] = newModel;
  this._setModel(oldIndex, newModel);
  newModel.SetIndex(oldIndex);
    // this._getModels()[newIndex] = oldModel;
  this._setModel(newIndex, oldModel);
  oldModel.SetIndex(newIndex);

  return { index: newIndex, swapped: (newIndex !== oldIndex) };
};

/**
 * @private
 */
oj.Collection.prototype._setInternal = function (models, parse, opt, deferred) {
    // Determine if any of the options are set
  var options = opt || {};
  var add = options.add === undefined ? true : options.add;
  var remove = options.remove === undefined ? true : options.remove;
  var merge = options.merge === undefined ? true : options.merge;
  var foundModels = [];
  var currModel = null;
  var i;
  var modelList;

  var mods = parse ? this.parse(models) : models;

  modelList = Array.isArray(mods) ? mods : [mods];

  if (deferred) {
    var self = this;
    return this._addPromise(function () {
      return self._deferredSet(modelList, self._getModels(), options, remove, add, merge, parse);
    });
  }

    // Go through the passed in models and determine what to do
  var swapped = false;
  for (i = 0; i < modelList.length; i += 1) {
    currModel = this._updateModel(this._newModel(modelList[i], parse, options, true), add,
      merge, deferred);
    if (currModel !== -1) {
            // And swap it into the position as passed in, if no comparator and we're removing, so that the slots
            // are exact
      var obj = this._swapModels(currModel, i, remove, add);
      var newLoc = obj.index;
      if (obj.swapped) {
        swapped = true;
      }
            // Save its new location as found
      if (foundModels.indexOf(newLoc) === -1) {
        foundModels.push(newLoc);
      }
    }
  }
  if (swapped) {
    var eventOpts = options.add ? { add: true } : null;
    this.TriggerInternal(options.silent, oj.Events.EventType.SORT, this, eventOpts, null);
  }

  oj.Collection._removeAfterSet(this, this._getModels(), remove, foundModels, options);
  return null;
};

/**
 * Handle the updates/removes on virtual collections
 * @private
 */
oj.Collection.prototype._deferredSet = function (modelList, modelsCopy, options, remove, add,
  merge, parse) {
  var foundModels = [];
  var i;

    // Go through the passed in models and determine what to do
  var self = this;
  return new Promise(function (allResolve, allReject) {
    var doTask = function (index) {
      return new Promise(function (resolve, reject) {
        self._updateModel(self._newModel(modelList[index], parse, options, true), add, merge,
                                                             true).then(function (currModel) {
                                                               if (currModel !== -1) {
                                                                 foundModels.push(currModel);
                                                               }
                                                               resolve(index + 1);
                                                             }, reject);
      });
    };

    var currentStep = Promise.resolve(0);

    for (i = 0; i < modelList.length; i += 1) {
      currentStep = currentStep.then(doTask);
    }
    currentStep.then(function () {
      oj.Collection._removeAfterSet(self, modelsCopy, remove, foundModels, options);
      allResolve(undefined);
    }, allReject);
  });
};


/**
 * Return the index of the given model after updating it, if it was found.  Otherwise it is added and a -1 is
 * returned
 * @private
 */
oj.Collection.prototype._updateModel = function (model, add, merge, deferred) {
  function update(collection, found, def) {
    var index = found ? found.index : -1;
    var foundModel = found ? found.m : null;

    var opt = {};
    if (foundModel) {
      if (merge) {
        // Do merge if not overridden
        opt = { merge: merge };
        if (def) {
          return new Promise(function (resolve) {
            collection._addInternal(model, opt, false, true).then(function () {
              resolve(index);
            });
          });
        }
        collection.add(model, opt);
      }
    } else if (add) {
      if (def) {
        return new Promise(function (resolve) {
          collection._addInternal(model, opt, false, true).then(function () {
            resolve(collection._getLength() - 1);
          });
        });
      }
      collection.add(model);
      index = collection._getLength() - 1;
    }
    return index;
  }

  // Check to see if this model is in the collection
  if (deferred || this.IsVirtual()) {
    var self = this;

    return new Promise(function (resolve) {
      self._getInternal(model, { silent: true }, deferred).then(function (found) {
        update(self, found, true).then(function (index) {
          resolve(index);
        });
      });
    });
  }

  var found = this._getInternal(model);
  return update(this, found, false);
};

/**
 * Return a copy of the Collection's list of current attribute/value pairs.
 *
 * @return {Array.<Object>} an array containing all the Collection's current sets of attribute/value pairs.
 * @throws {Error} when called on a virtual collection
 *
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.toJSON = function () {
  var retArr = [];

  this._throwErrIfVirtual('toJSON');

  this._getModels().forEach(function (model) {
    retArr.push(model.toJSON());
  });
  return retArr;
};

/**
 * Return the first model object in the collection, or an array of the first n model objects from the collection.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 * @param {number=} n Number of model objects to include in the array, starting with the first.
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual whether it is or not
 * @return {Array.<oj.Model>|null|Promise} An array of n model objects found in the collection, starting with the first. If n
 * is not included, returns all of the collection's models as an array.  If deferred or virtual, returns a promise
 * that resolves with the array or model
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.first = function (n, options) {
  var deferred = this._getDeferred(options);
  var elementCount = this._getLength();
  var retArray = [];
  var i;
  var self = this;

  var num = n;
  if (num) {
    elementCount = num;
  } else {
    num = 1;
  }

  var virtual = this.IsVirtual() || deferred;

  if (num === 1) {
    if (virtual) {
      return this._addPromise(function () {
        return self._deferredAt(0, null);
      });
    }

    if (this._getModelsLength() > 0) {
      return this._getModel(0);
    }
    return null;
  }

  if (elementCount > this._getModelsLength()) {
    if (this.IsVirtual() && !this._hasTotalResults()) {
      // Virtual, no total results: don't restrict elementCount
    } else {
      elementCount = this._getModelsLength();
    }
  }

  if (virtual) {
    return this._addPromise(function () {
      return self.IterativeAt(0, elementCount);
    });
  }

  for (i = 0; i < elementCount; i += 1) {
    retArray.push(this._getModel(i));
  }
  return retArray;
};

/**
 * Return the array index location of the given model object.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 * @param {Object} model Model object (or Model id) to locate
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual
 * whether it is or not
 * @return {number|Promise.<number>} The index of the given model object, or a promise that will resolve with the index
 * when complete.  If the object is not found, returns -1.
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.indexOf = function (model, options) {
  var location;
  var deferred = this._getDeferred(options);

  if (this.IsVirtual() || deferred) {
    var self = this;
    return this._addPromise(function () {
      return self._getInternal(model, null, true).then(function (loc) {
        return loc.index;
      });
    });
  }
  location = this._getInternal(model);

  return location.index;
};

/**
 * Determine if the given model object is present in the collection.<br>
 * For events that may be fired if the collection is virtual, see [fetch]{@link oj.Collection#fetch}.<br>
 * @param {Object} model Model object (or Model id) to locate
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual
 * whether it is or not
 * @return {boolean|Promise.<boolean>} true if the model is contained in the collection, false if not. If deferred, a
 * promise that will resolve with true or false when complete.
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.contains = function (model, options) {
  var location;
  var deferred = this._getDeferred(options);

  if (this.IsVirtual() || deferred) {
    var self = this;
    return this._addPromise(function () {
      return self._getInternal(model, null, true).then(function (loc) {
        return loc.index > -1;
      });
    });
  }
  location = this._getInternal(model);

  return location.index > -1;
};

/**
 * An alias for [contains]{@link oj.Collection#contains}
 * @param {Object} model Model object (or Model id) to locate
 * @param {Object=} options
 * @property {boolean=} deferred if true, return a promise as though this collection were virtual
 * whether it is or not
 * @return {boolean|Promise.<boolean>} true if the model is contained in the collection, false if not. If deferred, a
 * promise that will resolve with true or false when complete.
 * @kind function
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.include = oj.Collection.prototype.contains;

// Only look on models already fetched
oj.Collection.prototype._localIndexOf = function (model) {
  var location = this._getLocalInternal(model);

  return location !== undefined ? location.index : -1;
};

/**
 * Remove the last model from the collection and return it<br>
 * For events that may be fired if the collection is virtual, see [remove]{@link oj.Collection#remove}.<br>
 * @param {Object=} options Options for the method:<p>
 * <b>silent</b>: if set, do not fire a remove event <br>
 * <b>deferred</b>: if true, return a promise as though this collection were virtual whether it is or not<br>
 * @return {oj.Model|Promise.<oj.Model>} the model that was removed, or a promise that will resolve with the model that was
 * removed when complete
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{silent?: boolean, deferred?: boolean, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Collection.prototype.pop = function (options) {
  var deferred = this._getDeferred(options);
  if (this.IsVirtual() || deferred) {
    var self = this;
    return this._atInternal(this._getLength() - 1, options, false, true).then(function (model) {
      self.remove(model, options);
      return model;
    });
  }

  var m = /** @type {oj.Model} */ (this.at(this._getLength() - 1));
  this.remove(m, options);
  return m;
};

/**
 * Add the given model to the end of the Collection<br>
 * For events that may be fired if the collection is virtual, see [add]{@link oj.Collection#add}.<br>
 * @param {oj.Model|Object} m model to add to the end of the Collection (or a set of attribute value pairs)
 * @param {Object=} options same options as [add]{@link oj.Collection#add}
 * @return {Promise.<Array>|undefined} if deferred or virtual, a promise that will be resolved when the function is done.  Otherwise
 * undefined
 * @memberof oj.Collection
 * @ojsignature [{target: "Type", value:"{silent?: boolean, at?: number, merge?: boolean, sort?: boolean,
 *                                        force?: boolean, deferred?: boolean, [propName: string]: any}", for: "options"},
 *               {target: "Type", value: "Promise<Array<oj.Model>>|undefined", for: "returns"}]
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.push = function (m, options) {
  var deferred = this._getDeferred(options);
  this._manageLRU(1);
  return this._handlePromise(this._addInternal(m, options, false, deferred));
};

/**
 * Returns the index of the last location of the given model.  Not supported in virtual cases.
 * @param {oj.Model} model Model object to locate
 * @param {number=} fromIndex optionally start search at the given index
 * @return {number} The last index of the given model object.  If the object is not found, returns -1.
 * @throws {Error} when called on a virtual collection
 * @memberof oj.Collection
 * @since 1.0.0
 * @export
 */
oj.Collection.prototype.lastIndexOf = function (model, fromIndex) {
  var i;
  var locIndex = fromIndex;

  this._throwErrIfVirtual('lastIndexOf');

  if (locIndex === undefined) {
    locIndex = 0;
  }

  for (i = this._getLength() - 1; i >= locIndex; i -= 1) {
    if (oj.Object.__innerEquals(model, this.at(i))) {
      return i;
    }
  }
  return -1;
};

/**
 * @private
 */
oj.Collection.prototype._getSortAttrs = function (sortStr) {
  if (sortStr === undefined) {
    return [];
  }
  return sortStr.split(',');
};

/**
 * Return a URL query string based on an array of or a single attr/value pair set
 * @private
 */
oj.Collection._getQueryString = function (q) {
  function expression(left, right, compare) {
    return left + compare + right;
  }

  function processQuery(query, input) {
    var exp;
    var str = input;
    Object.keys(query || {}).forEach(function (prop) {
      if (Object.prototype.hasOwnProperty.call(query, prop)) {
        var val = Array.isArray(query[prop]) ? query[prop] : [query[prop]];
        for (var j = 0; j < val.length; j++) {
          if (oj.Model.IsComplexValue(val[j])) {
            var value = val[j].value;
            var compare = null;
            var comparator = val[j].comparator;
            if ($.isFunction(comparator)) {
              compare = comparator(null, prop, value);
            } else {
              compare = comparator;
            }
            exp = expression(prop, value, compare);
          } else {
            exp = expression(prop, query[prop], '=');
          }
          str += exp + '+';
        }
      }
    });
    // Remove trailing '+'
    str = str.substring(0, str.length - 1) + ',';
    return str;
  }

  var queries = Array.isArray(q) ? q : [q];
  var str = '';
  var i;
  for (i = 0; i < queries.length; i++) {
    str = processQuery(queries[i], str);
  }
    // Remove trailing ','
  if (str.substring(str.length - 1) === ',') {
    return str.substring(0, str.length - 1);
  }
  return str;
};

/**
 * @protected
 */
oj.Collection.prototype.ModifyOptionsForCustomURL = function (options) {
  var opt = {};
  Object.keys(options || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(options, prop)) {
      opt[prop] = options[prop];
    }
  });
  var comparator = this.comparator;
  if (comparator && oj.StringUtils.isString(comparator)) {
    var attrs = this._getSortAttrs(comparator);
    for (var i = 0; i < attrs.length; i++) {
      if (i === 0) {
        opt.sort = attrs[i];
      } else {
        opt.sort += ',' + attrs[i];
      }
    }
    opt.sortDir = this._getSortDirStr();
  }
    // Put fetchSize on if appropriate, and not already set
  if (this.IsVirtual()) {
    opt[oj.Collection._FETCH_SIZE_PROP] = this._getFetchSize(opt);
  }
  return opt;
};

/**
 * Determine if this collection is URL-based
 * @private
 */
oj.Collection.prototype.IsUrlBased = function (options) {
  var customURL = this.customURL;
  if ($.isFunction(customURL)) {
    return true;
  }
  var url = this.GetCollectionFetchUrl(options);
  return oj.Collection._defined(url);
};

/**
 * Build a URL with parameters for the collection fetch
 * @protected
 */
oj.Collection.prototype.GetCollectionFetchUrl = function (opt) {
  var url = oj.Model.GetPropValue(this, 'url');

    // Adorn it with options, if any
  if (this.IsVirtual()) {
    var options = opt || {};
    var all = options.all;

        // Put in page size
    var limit = null;
    if (all) {
      var totalResults = this.totalResults;
      limit = totalResults || this._getFetchSize(options);
    } else {
      limit = this._getFetchSize(options);
    }
    if (url && url.indexOf('?') > -1) {
            // Already have a param coming in
      url += '&';
    } else {
      url += '?';
    }
    url += 'limit=' + limit;

    if (!all) {
      if (oj.Collection._defined(options.startIndex)) {
        url += '&offset=' + options.startIndex;
      }
      if (options.startID) {
        url += '&fromID=' + options.startID;
      }
      if (options.since) {
        url += '&since=' + options.since;
      }
      if (options.until) {
        url += '&until=' + options.until;
      }
    }
        // Query
    if (options.query) {
      var queryString = oj.Collection._getQueryString(options.query);
      if (queryString && queryString.length > 0) {
        url += '&q=' + queryString;
      }
    }

        // Add sorting
    var comparator = this.comparator;
    if (comparator && oj.StringUtils.isString(comparator)) {
      var attrs = this._getSortAttrs(comparator);
      var sortDirStr = this._getSortDirStr();
      var i;
      for (i = 0; i < attrs.length; i++) {
        if (i === 0) {
          url += '&orderBy=' + attrs[i] + ':' + sortDirStr;
        } else {
          url += ',' + attrs[i] + ':' + sortDirStr;
        }
      }
    }
        // Always ask for totalresults
    url += '&totalResults=true';
  }
  return url;
};

/**
 * @private
 */
oj.Collection.prototype._getSortDirStr = function () {
  if (this.sortDirection === -1) {
    return 'desc';
  }
  return 'asc';
};

/**
 * @export
 * Called to perfrom server interactions, such as reading the collection.  Designed to be overridden by users
 *
 * @param {string} method "read"
 * @param {oj.Collection} collection the collection to be read (fetched)
 * @param {Object=} options to control sync<br>
 * <b>success</b>: called if sync succeeds.  Called with the array of JSON data being fetched<br>
 * <b>error</b>: called if sync fails.  Called with xhr, status, and error info, per jQuery Ajax (all if
 * available)<br>
 * @return {Object} xhr response from ajax by default
 * @memberof oj.Collection
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{success?: (json?: Array<Object>)=> void,
 *                                                  error?: (xhr: Object, status: Object, error: Object)=> void, [propName: string]: any}", for: "options"}
 * @alias oj.Collection.prototype.sync
 */
oj.Collection.prototype.sync = function (method, collection, options) {
  return oj.sync(method, collection, options);
};

/**
 * Constants
 * @private
 */
oj.Collection._FETCH_SIZE_PROP = 'fetchSize';

/**
 * Copyright (c) 2014, 2015 Oracle and/or its affiliates.
 * All rights reserved.
 */

/* jslint browser: true*/

/**
 * @export
 * @class oj.Model
 * @classdesc Object representing name/value pairs for a data service record
 *
 * @param {Object=} attributes Initial set of attribute/value pairs with which to seed this Model object
 * @param {Object=} options
 *                  collection: collection for this model
 * @constructor
 * @since 1.0
 * @mixes oj.Events
 * @ojsignature {target: "Type", value: "class Model"}
 */
oj.Model = function (attributes, options) {
  oj.Model._init(this, attributes, options, null);
};


/**
 * Subclass from oj.Object
 * @private
 */
oj.Object.createSubclass(oj.Model, oj.Object, 'oj.Model');

/**
 * @private
 */
oj.Model.prototype.Init = function () {
  oj.Model.superclass.Init.call(this);
};

/**
 *
 * @export
 * @desc Attribute/value pairs in the model.
 * @memberof oj.Model
 *
 * @type Object
 * @since 1.0.0
 */
oj.Model.prototype.attributes = {};

/**
 * @export
 * @desc The set of attribute/value pairs that serve as default values when new Model objects are created.
 * @memberof oj.Model
 *
 * @type Object
 * @since 1.0.0
 */
oj.Model.prototype.defaults = {};

/**
 * @export
 * @desc The model's unique ID.  This can be set by the application or retrieved from the data service. This ID
 * will be appended to the URL for single-record data operations (update, delete).
 * @memberof oj.Model
 *
 * @type {string|null}
 * @since 1.0.0
 */
oj.Model.prototype.id = null;

/**
 * @desc The name of the model property to be used as the unique ID. See [id]{@link oj.Model#id}. This defaults to
 * a value of "id".
 * @memberof oj.Model
 *
 * @type {string|null}
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.idAttribute = 'id';

/**
 * @export
 * @desc The base url on the data service used to perform CRUD operations on models.  If not defined, the model
 * will look to its collection.  One or the other must be defined before CRUD operations can succeed.
 * @memberof oj.Model
 *
 * @type {string|null}
 * @since 1.0.0
 */
oj.Model.prototype.urlRoot = null;

/**
 * @export
 * @desc A callback to allow users to completely customize the data service URLs
 * The callback should accept these parameters:<p>
 * <b>operation (string)</b>: one of "create", "read", "update", "patch", or "delete", indicating the type of
 * operation for which to return the URL<br>
 * <b>model (Object)</b>: the oj.Model object requesting the URL<br>
 * <b>options (Object)</b>: one or more of the following properties:<br>
 * <ul>
 * <b>recordID</b>: id of the record involved, if relevant<br>
 * </ul>
 *
 * customURL callbacks should return either: null, in which case the default will be used; a url string, which
 * will be used with the standard HTTP method for the type of operation, or an Object with with any other
 * attributes that should be passed to the ajax call.<br>
 * This object must at minimum include the URL, and other attributes as follows:<br>
 * <ul>
 * <b>url</b>: giving the custom URL string<br>
 * <b>type</b>: (optional) a string indicating the type of HTTP method to use (GET, POST, DELETE, etc.)<br>
 * <b>(other)</b>: (optional) any other ajax attributes to pass in the ajax call
 * </ul>
 * <p>
 *
 * @memberof oj.Model
 * @type {function(string,oj.Model,Object):(string|Object|null)|null}
 * @since 1.0.0
 */
oj.Model.prototype.customURL = null;


/**
 * @export
 * @desc A callback to allow optional custom validation of model attributes during save, set, etc.
 * The callback should accept these parameters:<p>
 * <b>attributes (Object)</b>: the attributes to validation<br>
 * <b>options (Object)</b>: the options passed in to the model call making the validation check<br>
 *
 * The validate callback should return nothing if the attributes are valid, or an error string or object if the validation fails<br>
 * <p>
 *
 * @memberof oj.Model
 * @type {function(Object,Object):(string|Object|null)|null}
 * @since 2.3.0
 */
oj.Model.prototype.validate = null;

/**
 * @export
 * @memberof oj.Model
 * @type {string|Object|null}
 * @desc The last value returned from a validate callback
 * @since 2.3.0
 */
oj.Model.prototype.validationError = null;

/**
 * @export
 * @memberof oj.Model
 * @ojstatus preview
 * @type {boolean}
 * @desc If true, do not insert the JET locale-based Accept-Language header.  If false, let the Ajax system set the header.
 * @since 5.0.0
 */
oj.Model.prototype.omitLanguageHeader = false;

/**
 * @private
 */
oj.Model._idCount = 0;

/**
 * @private
 */
oj.Model._init = function (model, attributes, opt, properties) {
  var parse;
  var attrCopy;
  var prop;

  if (oj.Model._justExtending) {
    return;
  }

  model.Init(); // eslint-disable-line no-param-reassign

  // Augment with Event
  oj.Events.Mixin(model);

  model._clearChanged(); // eslint-disable-line no-param-reassign
  model.previousAttrs = {}; // eslint-disable-line no-param-reassign
  model.nestedSet = false; // eslint-disable-line no-param-reassign
  model.index = -1; // eslint-disable-line no-param-reassign

  var options = opt || {};

  // Deep copy actual data if found
  model.attributes = {}; // eslint-disable-line no-param-reassign
  if (model.defaults && !options.ignoreDefaults) {
    model.attributes = oj.Model._cloneAttributes($.isFunction(model.defaults) ? // eslint-disable-line no-param-reassign
      model.defaults() : model.defaults, null);
  }

  // First, copy all properties passed in
  for (prop in properties) { // eslint-disable-line no-restricted-syntax
    if (Object.prototype.hasOwnProperty.call(properties, prop)) {
      model[prop] = properties[prop]; // eslint-disable-line no-param-reassign
    }
  }

  if (attributes) {
    parse = options.parse;
    if ($.isFunction(parse)) {
      model.parse = parse; // eslint-disable-line no-param-reassign
    }

    attrCopy = oj.Model._cloneAttributes(attributes, model.attributes);
    model.attributes = attrCopy; // eslint-disable-line no-param-reassign

    attrCopy = parse ? model.parse(attrCopy) : attrCopy;
    if (attrCopy == null || attrCopy === undefined) {
      // Reset it
      model.attributes = {}; // eslint-disable-line no-param-reassign
    } else {
      // Move them in
      for (prop in attrCopy) { // eslint-disable-line no-restricted-syntax
        if (Object.prototype.hasOwnProperty.call(attrCopy, prop)) {
          model._setProp(prop, attrCopy[prop], false, false, options, true);
        }
      }
    }
  }

  model.SetCid();

    // Grab collection option, if there
  model.SetCollection(options.collection);

  if (options.customURL) {
    model.customURL = options.customURL; // eslint-disable-line no-param-reassign
  }

    // If URL is set, use that
  if (options.url) {
    model.url = options.url; // eslint-disable-line no-param-reassign
  }

  if (options.urlRoot) {
    model.urlRoot = options.urlRoot; // eslint-disable-line no-param-reassign
  }

  if (model.initialize) {
    model.initialize(attributes, options);
  }

  model.SetupId();
};


/**
 * Create a new, specific type of model object to represent single records from a JSON data set.
 * @param {Object=} properties Properties for the new Model class.<br>
 *                  <b>defaults</b>: an Object containing starting attribute/value pairs for some or all of the
 *                  model's potential attributes<br>
 *                  <b>parse</b>: a user callback function to allow parsing of JSON record objects as they are
 *                  returned from the data service<br>
 *                  <b>parseSave</b>: a user callback function to allow conversion of models back into a format
 *                  appropriate for the data service on save calls<br>
 *                  <b>urlRoot</b>: the URL to use to get records from the data service in the abscence of a
 *                  collection (when an id is appended)<br>
 *                  <b>initialize</b>: a user callback function to be called when this model is created<br>
 *                  <b>validate</b>: a user callback function that will be called before a save to the data
 *                  service occurs. The callback is passed the current set of attributes and save options.
 *                  <br>
 * @param {Object=} classProperties properties that attach to the whole class
 * @return {oj.Model} new Model object
 * @export
 * @ojsignature [{target: "Type",
 *                value: "any",
 *                for: "returns"},
 *               {target: "Type", value:"{parse?: (data: any)=> any, parseSave?: (data: any)=> any, urlRoot?: string,
 *                                        initialize?: (models: Array<oj.Model>, options: Object)=> void,
 *                                        validate?: null|object|string|((attributes: Object, options?: oj.Model)=> number), [propName: string]: any}", for: "properties"}]
 * @memberof oj.Model
 * @this {oj.Model}
 * @since 1.0.0
 */
oj.Model.extend = function (properties, classProperties) {
  oj.Model._justExtending = true;
  var obj;

  obj = new oj.Model();
  oj.Model._justExtending = false;

    // Add regular properties from this "parent"
    // oj.Events.Mixin(obj, this.prototype);
  $.extend(obj, this.prototype);

    // Grab properties
  var props = properties || {};
  Object.keys(props).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(props, prop)) {
      obj[prop] = props[prop];
    }
  });

  var Model;

  if (props && props.constructor && Object.prototype.hasOwnProperty.call(props,
    'constructor')) {
    Model = props.constructor;
  } else {
    Model = function (attributes, options) {
      oj.Model._init(this, attributes, options, props);
    };
  }

  $.extend(Model, this);
  Model.prototype = obj;

    // Allow extending resulting obj
  Model.extend = oj.Model.extend;

  Model.prototype.constructor = Model;

    // Add class properties from this
  oj.Events.Mixin(Model, this);

  if (classProperties) {
    Object.keys(classProperties).forEach(function (prop) {
      if (Object.prototype.hasOwnProperty.call(classProperties, prop)) {
        Model[prop] = classProperties[prop];
      }
    });
  }


  return Model;
};

/**
 * Placeholder for event mixins
 * @private
 */
oj.Model.prototype.TriggerInternal = function (silent, event, arg1, arg2, options) {}; // eslint-disable-line no-unused-vars

/**
 * @protected
 */
oj.Model.prototype.SetCid = function () {
    // Create cid property if necessary
  if (!this.GetCid()) {
    this.cid = 'id' + oj.Model._idCount;
    oj.Model._idCount += 1;
  }
};

/**
 * @protected
 */
oj.Model.prototype.GetCid = function () {
  return this.cid;
};

/**
 * Index within collection
 * @protected
 */
oj.Model.prototype.SetIndex = function (index) {
  this.index = index;
};

/**
 * @protected
 */
oj.Model.prototype.GetIndex = function () {
  return this.index;
};

/**
 * LRU functions
 * @protected
 */
oj.Model.prototype.SetNext = function (model) {
  var retVal = this.nextModel;
  this.nextModel = model;
  return retVal;
};

/**
 * @protected
 */
oj.Model.prototype.GetNext = function () {
  return this.nextModel;
};

/**
 * @protected
 */
oj.Model.prototype.SetPrevious = function (model) {
  var retVal = this.previousModel;
  this.previousModel = model;
  return retVal;
};

/**
 * @protected
 */
oj.Model.prototype.GetPrevious = function () {
  return this.previousModel;
};

/**
 * Merge the given model's attributes with this model's attributes
 * @protected
 */
oj.Model.prototype.Merge = function (model, comparator, silent) {
  var needSort = false;
  var isStringComparator = oj.StringUtils.isString(comparator);
  var valueChange;
  var changes = false;

  var self = this;
  Object.keys(model.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(model.attributes, prop)) {
      valueChange = (self.attributes[prop] !== model.attributes[prop]);
      if (isStringComparator) {
                // We have a string comparator--does it match this property?  If we hit a property that doesn't
                // match, we need sort
        if (prop === comparator) {
                    // The property matches the comparator property: are we changing the value?
          if (valueChange) {
            needSort = true;
          }
        }
      } else if (valueChange) {
        needSort = true;
      }
      if (valueChange) {
        changes = true;
        self.attributes[prop] = model.attributes[prop];
        self._addChange(prop, model.attributes[prop]);
        self._fireAttrChange(prop, self.attributes[prop], null, silent);
      }
    }
  });
  this.SetupId();
    // Only fire master change if there were any changes
  if (changes) {
    this._fireChange(null, silent);
  }
  return needSort;
};

/**
 * @private
 */
oj.Model._hasProperties = function (object) {
  if (object && object instanceof Object) {
    var prop;
    for (prop in object) { // eslint-disable-line no-restricted-syntax
      if (Object.prototype.hasOwnProperty.call(object, prop)) {
        return true;
      }
    }
  }
  return false;
};

/**
 * @protected
 */
oj.Model.prototype.SetCollection = function (coll) {
  if (coll == null) {
    delete this.collection;
    return;
  }
  this.collection = coll;
    // This can depend on the collection
  this.SetupId();
};

/**
 * @protected
 */
oj.Model.prototype.GetCollection = function () {
  return this.collection;
};

/**
 * @private
 */
oj.Model.prototype._fireAttrChange = function (prop, value, options, silent) {
  if (prop != null) {
    this.TriggerInternal(silent, oj.Events.EventType.CHANGE + ':' + prop, this, value, options);
  }
};

/**
 * @private
 */
oj.Model.prototype._fireChange = function (options, silent) {
  this.TriggerInternal(silent, oj.Events.EventType.CHANGE, this, options, null);
};

/**
 * @protected
 */
oj.Model.prototype.SetupId = function () {
    // Replicate id attribute at top level
  var id = null;
    // Ask for collection's function if available
  if (this.collection && this.collection.modelId) {
    var modFunc = this.collection.modelId;
    id = $.isFunction(modFunc) ? modFunc.call(this.collection, this.attributes) : modFunc;
  }
  if (!id) {
    var idAttr = this._getIdAttr();
    id = this.attributes != null ? this.attributes[idAttr] : null;
  }
    // Supposedly this should always be model.id...who knew?
  this.id = id;
};

/**
 * @private
 */
oj.Model.prototype._setPropInternal = function (prop, value, copyRegardless) {
  var equality = oj.Object.__innerEquals(this.attributes[prop], value);
  if (copyRegardless || !equality) {
    this.attributes[prop] = value;
    this.SetupId();
        // Return value management here seems bizarre due to backbone tests: do the direct set if copyRegardless,
        // but only return if the inner equals was different
    return !equality;
  }
  return false;
};

/**
 * @private
 */
oj.Model.prototype._clearChanged = function () {
  this.changed = {};
};

/**
 * @private
 */
oj.Model.prototype._addChange = function (property, value) {
  this.changed[property] = value;
};

/**
 * @ignore
 * @private
 * @param {Object|string} prop
 * @param {Object|null|undefined} value
 * @param {boolean} copyRegardless
 * @param {boolean} propertyBag
 * @param {Object} options
 * @param {boolean} init
 * @returns {boolean}
 */
oj.Model.prototype._setProp = function (prop, value, copyRegardless, propertyBag, options, init) {
  if (prop == null) {
    return true;
  }

  var attrs = {};
  var isNested = this.nestedSet;
  var opts;

  if (!propertyBag) {
    attrs[prop] = value;
  } else {
    // We've passed in a whole property bag at once: validate all together
    Object.keys(prop).forEach(function (p) {
      if (Object.prototype.hasOwnProperty.call(prop, p)) {
        attrs[p] = prop[p];
      }
    });
  }
  opts = options || {};

  if (!this._checkValid(attrs, { validate: opts.validate }, false)) {
    return false;
  }

  if (!isNested) {
    this._clearChanged();
    this.changes = [];
  }

    // Store old value
  if (!this.nestedSet && !init) {
    this.previousAttrs = oj.Model._cloneAttributes(this.attributes, null);
  }

  this.nestedSet = true;
  var self = this;
  Object.keys(attrs).forEach(function (p) {
    if (Object.prototype.hasOwnProperty.call(attrs, p)) {
      if (self._setPropInternal(p, attrs[p], copyRegardless)) {
        // Trigger changes
        self._addChange(p, attrs[p]);
        self.changes.push(p);
      } else {
        delete attrs[p];
      }
    }
  });
    // Fire events: don't fire if silent
  var silent = opts.silent;
  Object.keys(attrs).forEach(function (p) {
    if (Object.prototype.hasOwnProperty.call(attrs, p)) {
      if (!silent && (self.changes.length > 0 || (isNested && self.changes.indexOf(p) === -1))) {
        self.pendingChanges = true;
        self.pendingOpts = opts;
      }
      self._fireAttrChange(p, attrs[p], opts, silent);
    }
  });

  if (isNested) {
    return true;
  }
  if (!silent && !isNested) {
    while (this.pendingChanges) {
      this.pendingChanges = false;
      this._fireChange(this.pendingOpts, silent);
    }
  }

  this.nestedSet = false;
  return true;
};


/**
 * Clears all attributes from the model<br>
 * Events:<p>
 * <ul>
 * <b>change:attr</b>: fired for each attribute cleared, passing the model, name of the changed property, and
 * options<br>
 * <b>change:all</b>: fired after all attributes have been cleaered, passing the model and options<br>
 * </ul>
 * <p>
 *
 * @param {Object=} options
 * @property {boolean=} silent if true, do not fire events
 * @property {boolean=} validate if true, validate the unsetting of all properties
 * @return {oj.Model|boolean} the model, or false if validation on clear fails
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.clear = function (options) {
  // Use unset to silently clear, to track changes to attributes
  var unsetOpt = { silent: true };
  var silent;
  var p;
  var opt = options || {};
  silent = opt.silent;
  unsetOpt.validate = opt.validate;
  this._clearChanged();

  var self = this;
  for (p in self.attributes) { // eslint-disable-line no-restricted-syntax
    if (Object.prototype.hasOwnProperty.call(self.attributes, p)) {
      if (!this._unsetInternal(p, unsetOpt, true)) {
        return false;
      }
      this.TriggerInternal(silent, oj.Events.EventType.CHANGE + ':' + p, this, undefined, null);
    }
  }
  this.attributes = {};
  this.SetupId();

  this._fireAttrChange(null, null, null, silent);
  this._fireChange(null, silent);
  return this;
};

/**
 * @private
 */
oj.Model._cloneAttributes = function (oldData, nd) {
  var newData = nd || {};

  // Handle not overwriting defaults with undefined
  // IE11 issue with Object.keys
  var newDataKeys = typeof newData === 'object' ? Object.keys(newData) : [];

  var canUseJson = true;

  var prop;
  if (newDataKeys.length > 0) {
    for (prop in newData) { // eslint-disable-line no-restricted-syntax
      if (Object.prototype.hasOwnProperty.call(newData, prop) &&
          Object.prototype.hasOwnProperty.call(oldData, prop)) {
        // They both have this: now is oldData undefined?
        if (oldData[prop] === undefined) {
          // Remove it so it doesn't get copied/overwritten
          delete oldData[prop]; // eslint-disable-line no-param-reassign
        }
      }
    }
    oj.CollectionUtils.copyInto(newData, oldData, undefined, true, 10000);
    return newData;
  }
  var type;
  for (prop in oldData) { // eslint-disable-line no-restricted-syntax, guard-for-in
    type = $.type(oldData[prop]);
    if (type === 'function' || type === 'undefined' || type === 'date' || type === 'array' || type === 'object') {
      canUseJson = false;
      break;
    }
  }

  if (canUseJson) {
    newData = JSON.parse(JSON.stringify(oldData));
  } else if (typeof oldData === 'object') {
    // IE11 issue with Object.keys
    oj.CollectionUtils.copyInto(newData, oldData, undefined, true, 10000);
  }
  return newData;
};

/**
 * Return a copy of the model with identical attributes and settings
 * @return {oj.Model} copy of the model
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.clone = function () {
  var c = new this.constructor();
  var prop;

  for (prop in this) { // eslint-disable-line no-restricted-syntax
        // Shallow copy all but data
    if (Object.prototype.hasOwnProperty.call(this, prop) && this[prop] !== this.attributes) {
      c[prop] = this[prop];
    }
  }
    // Deep copy data
  c.attributes = oj.Model._cloneAttributes(this.attributes, null);

    // Remove the cid--this should be unique
  delete c.cid;
    // Set a new cid
  c.SetCid();

  c.SetupId();

  return c;
};

/**
 * Does this model match the given id or cid?
 * @protected
 */
oj.Model.prototype.Match = function (id, cid) {
  var modId = this.GetId();
  var modCid;
  if (modId !== undefined && modId == id) { // eslint-disable-line eqeqeq
    return true;
  }
  modCid = this.cid;
  if (modCid !== undefined && modCid == cid) { // eslint-disable-line eqeqeq
    return true;
  }
  return false;
};

/**
 * Set the value(s) of one or more attributes of the model, and fire events.
 * Events:<p>
 * <ul>
 * <b>change:attr</b>: fired for each attribute set, passing the model, name of the changed property, and options<br>
 * <b>change:all</b>: fired after all attributes have been set, passing the model and options<br>
 * </ul>
 * <p>
 * @param {string|Object} property Property attribute name to set, or an Object containing attribute/value pairs
 * @param {Object=} value Value for property if property is not an Object containing attribute/value pairs
 * @param {Object=} options
 * @property {boolean=} silent prevent events from firing
 * @property {boolean=} unset delete all the properties passed in rather than setting them<br>
 * @returns {oj.Model|boolean} the model itself, false if validation failed on set
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.set = function (property, value, options) {
  var opts = options || {};
  var prop;
  var valid = true;

  if (arguments && arguments.length > 0) {
    // Check if first arg is not a string (property name)
    if (!oj.StringUtils.isString(property)) {
      // Options, if present, must be second argument...value, because a string/value
      // pair wasn't what was passed in
      opts = value || {};
      // For set, pass entire thing to setProp
      if (opts.unset) {
        for (prop in property) { // eslint-disable-line no-restricted-syntax
          if (Object.prototype.hasOwnProperty.call(property, prop)) {
            this._unsetInternal(prop, null, false);
          }
        }
      } else if (!this._setProp(property, null, true, true, opts, false)) {
        valid = false;
      }
    } else if (opts.unset) {
      // Not a property bag?  We assume it's a property/value argument
      this._unsetInternal(property, null, false);
    } else if (!this._setProp(property, value, false, false, opts, false)) {
      valid = false;
    }
  }
  if (valid) {
    return this;
  }
  return false;
};

/**
 * Deletes the given property from the model.<br>
 * Events:<p>
 * <ul>
 * <b>change:attr</b>: fired for each attribute unset: passing the model, name of the changed property, and
 * options<br>
 * <b>change:all</b>: fired after all attributes have been unset: passing the model and options<br>
 * @param {string} property Property to remove from model
 * @param {Object=} options
 * @property {boolean=} silent do not fire change events if true
 * @returns {boolean} false if validation of the unset fails
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.unset = function (property, options) {
  return this._unsetInternal(property, options, false);
};


/**
 * @private
 */
oj.Model.prototype._unsetInternal = function (property, opts, clear) {
  var options = opts || {};
  var silent = options.silent;
  var attrs = {};

  if (this.has(property)) {
    if (!this._checkValid(attrs, options, false)) {
      return false;
    }
    if (!clear) {
      this._clearChanged();
    }

        // attrs[property] = undefined;
    delete this.attributes[property];
    this._addChange(property, undefined);
        // if (!silent) {
    this._fireAttrChange(property, null, null, silent);
    this._fireChange(null, silent);
        // }
  }
  this.SetupId();
  return true;
};

/**
 * Returns the value of a property from the model.
 * @param {string} property Property to get from model
 * @return {Object} value of property
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.get = function (property) {
  return this.attributes[property];
};

/**
 * Determines if the Model has a certain property set, vs. undefined.
 * @param {string} property Property to check for
 * @return {boolean} true if the model contains the given property, false if undefined.
 * @memberof oj.Model
 * @export
 */
oj.Model.prototype.has = function (property) {
  return oj.Collection._defined(this.attributes[property]);
};

/**
 * Loads the Model object from the data service URL. Performs a data "read."<br>
 * Events:<p>
 * <ul>
 * <b>request</b>: fired when the request to fetch is going to the server, passing the model, xhr object, and
 * options<br>
 * <b>sync</b>: fired when the model is fetched from the data service, passing the model and the raw response<br>
 * <b>error</b>: fired if there is an error during the fetch, passing the model, xhr ajax object, and options<br>
 * </ul>
 * <p>
 * @param {Object=} options Options to control fetch<p>
 * <b>success</b>: a user callback called when the fetch has completed successfully. This makes the fetch an
 * asynchronous process. The callback is called passing the Model object, raw response, and the fetch options
 * argument.<p>
 * <b>error</b>: a user callback function called if the fetch fails. The callback is called passing the model
 * object, xhr, and options arguments.<p>
 * @return {Object} xhr ajax object, by default.  If [sync]{@link oj.Model#sync} has been replaced, this would be
 * the value returned by the custom implementation.
 * @memberof oj.Model
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{success?: (model: oj.Model, response: Object, options: Object)=> void,
 *                                                  error?: (model: oj.Model, xhr: Object, options: Object)=> void,
 *                                                  [propName: string]: any}", for: "options"}
 * @export
 */
oj.Model.prototype.fetch = function (options) {
  var tempOpts = options || {};
  var success = tempOpts.success;
  var userErr = tempOpts.error;
  var self = this;
  var opts;

  opts = oj.Model._copyOptions(tempOpts);
  opts.error = function (xhr, status, err) {
    // Trigger an error event
    oj.Model._triggerError(self, false, tempOpts, status, err, xhr);

    if (userErr) {
      userErr.apply(self, arguments);
    }
  };

  opts.success = function (response) {
    // Make sure we pass xhr
    if (opts.xhr) {
      tempOpts.xhr = opts.xhr;
    }
    oj.Model._fireSyncEvent(self, response, opts, false);

    if ($.isFunction(self.parse)) {
      self.set(self.parse(response), opts);
    }
    if (success) {
      success.call(oj.Model.GetContext(opts, self), self, response, tempOpts);
    }
  };
  return oj.Model._internalSync('read', this, opts);
};


/**
 * @private
 */
oj.Model.prototype._parseImpl = function (rawData) {
  return rawData;
};

/**
 * Optional callback to parse responses from the server.  It is called with the server's response with a model's data and should return a response (possibly modified) for processing
 * @type {function(Object):Object}
 * @ojstatus preview
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.parse = oj.Model.prototype._parseImpl;

/**
 * Return the URL used to access this model in the data source
 *
 * @returns {string|null} url to access this model in the data source
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.url = function () {
  var urlRoot = this._getUrlRoot();
  var id = this.GetId();
  var coll;
  var collUrl;
  var slash;

  if (urlRoot) {
    return id ? urlRoot + '/' + encodeURIComponent(id) : urlRoot;
  }

  coll = this.collection;
  if (coll) {
    collUrl = oj.Model.GetPropValue(coll, 'url');
    if (id && collUrl) {
      slash = oj.Model._getLastChar(collUrl) === '/' ? '' : '/';
      return collUrl + slash + encodeURIComponent(this.GetId());
    }
    return collUrl;
  }

  throw new oj.URLError();
};


/**
 * Return all of the model's attributes as an array
 *
 * @returns {Array.<Object>} array of all the model's attributes
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.keys = function () {
  var retArray = [];
  var self = this;

  Object.keys(self.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(self.attributes, prop)) {
      retArray.push(prop);
    }
  });
  return retArray;
};


/**
 * Return all of the model's attributes values as an array
 *
 * @returns {Array.<Object>} array of all the model's attributes values
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.values = function () {
  var retArray = [];
  var self = this;

  Object.keys(self.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(self.attributes, prop)) {
      retArray.push(self.get(prop));
    }
  });
  return retArray;
};

/**
 * Return an array of attributes/value pairs found in the model
 *
 * @returns {Object} returns the model's attribute/value pairs as an array
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.pairs = function () {
  var retObj = [];
  var item;
  var self = this;

  Object.keys(self.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(self.attributes, prop)) {
      item = [];
      item.push(prop);
      item.push(self.get(prop));
      retObj.push(item);
    }
  });
  return retObj;
};

/**
 * Return attribute/value pairs for the model minus those attributes listed in keys
 *
 * @param {Array.<Object>|Object} keys keys to exclude from the returned attribute/value pairs
 *
 * @returns {Object} array of the model's attribute/value pairs except those listed in keys
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.omit = function (keys) {
  var keyArr = [];
  var i;
  var retObj = {};

  if (keys instanceof Array) {
    keyArr = keys;
  } else {
    for (i = 0; i < arguments.length; i++) {
      keyArr.push(arguments[i]);
    }
  }
  var self = this;

  Object.keys(self.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(self.attributes, prop)) {
      if (keyArr.indexOf(prop) === -1) {
        retObj[prop] = self.get(prop);
      }
    }
  });
  return retObj;
};

/**
 * Return attribute/value pairs for the model for the keys
 *
 * @param {Array.<Object>|Object} keys keys for which to return attribute/value pairs
 *
 * @returns {Object} array of the model's attribute/value pairs filtered by keys
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.pick = function (keys) {
  var keyArr = [];
  var i;
  var retObj = {};

  if (keys instanceof Array) {
    keyArr = keys;
  } else {
    for (i = 0; i < arguments.length; i++) {
      keyArr.push(arguments[i]);
    }
  }
  for (i = 0; i < keyArr.length; i++) {
    if (Object.prototype.hasOwnProperty.call(this.attributes, keyArr[i])) {
      retObj[keyArr[i]] = this.get(keyArr[i]);
    }
  }
  return retObj;
};

/**
 * Return an array of value/attribute pairs found in the model
 *
 * @returns {Object} returns the model's value/attribute pairs as an array
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.invert = function () {
  var retObj = {};
  var val;
  var self = this;

  Object.keys(self.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(self.attributes, prop)) {
      val = self.get(prop);
      retObj[val] = prop;
    }
  });
  return retObj;
};

/**
 * @private
 */
oj.Model._getLastChar = function (str) {
  return str.charAt(str.length - 1);
};

/**
 * @private
 */
oj.Model.prototype._saveUrl = function () {
  var urlRoot = this._getUrlRoot();
  if (urlRoot) {
    return urlRoot;
  }

  if (this.GetCollection()) {
    return this.GetCollection().url;
  }

  return null;
};

/**
 * @private
 */
oj.Model.prototype._getUrlRoot = function () {
  return oj.Model.GetPropValue(this, 'urlRoot');
};

/**
 * @private
 */
oj.Model.prototype._parseSaveImpl = function (modelData) {
  return modelData;
};

/**
 * Callback function when writing a model to the server
 * @type {function(Object):Object}
 * @since 1.0.0
 * @memberof oj.Model
 * @ojstatus preview
 * @export
 */
oj.Model.prototype.parseSave = oj.Model.prototype._parseSaveImpl;

/**
 * Check to see if the model is valid by running the validate callback, if it is set
 *
 * @returns {boolean} true if validate passes or if no validate callback
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.isValid = function () {
  var options = {};
  options.validate = this.validate;
  return this._checkValid(this.attributes, options, false);
};

/**
 * @private
 */
oj.Model._isValidateSet = function (opt, save) {
  var options = opt || {};
  if (options.validate !== undefined && options.validate !== null) {
    return options.validate;
  }
    // The "default" is different for save vs. set
  return save;
};

/**
 * @private
 */
oj.Model.prototype._checkValid = function (attributes, opt, save) {
  var options = opt || {};
  var validate = this.validate;
  if (validate && oj.Model._isValidateSet(options, save)) {
        // If we have a validate override and it returns something, don't save
    this.validationError = validate.call(this, attributes, options);
    if (this.validationError) {
      this.TriggerInternal(false, oj.Events.EventType.INVALID, this, this.validationError, options);
      return false;
    }
  }
  return true;
};

/**
 * @private
 */
oj.Model._processArgs = function (args) {
  var ignoreLastArg = false;
  var options = {};
  var attributes = {};
  var i;

  if (args) {
    if (args.length > 0) {
            // Check if the last argument is not the first argument
      if (args.length > 1) {
        if (args[args.length - 1] && oj.Model._hasProperties(args[args.length - 1])) {
                    // Last arg is options: ignore later
          ignoreLastArg = true;
          options = args[args.length - 1] || {};
        }
      }
      if (args[0] == null) {
        return { attributes: null, options: options };
      }

      // Check if first arg is property bag
      if (oj.Model._hasProperties(args[0]) || oj.Object.isEmpty(args[0])) {
        Object.keys(args[0]).forEach(function (prop) {
          if (Object.prototype.hasOwnProperty.call(args[0], prop)) {
            attributes[prop] = args[0][prop];
          }
        });
      } else {
        // Not a property bag?  We assume arguments are a series of attr/values
        for (i = 0; i < args.length; i += 2) {
          // Process the arg as long as its: defined, and isn't the last argument where we're supposed to
          // ignore the last argument due to it being 'options'
          if (args[i] !== undefined || i < args.length - 1 || (!ignoreLastArg && i ===
            args.length - 1)) {
            attributes[args[i]] = args[i + 1];
          }
        }
      }
    }
  }
  return { attributes: attributes, options: options };
};

/**
 * @private
 */
oj.Model._copyOptions = function (opt) {
  var optReturn = {};
  var options = opt || {};

  Object.keys(options).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(options, prop)) {
      optReturn[prop] = options[prop];
    }
  });
  return optReturn;
};

/**
 * @private
 */
oj.Model._triggerError = function (self, silent, opt, status, err, xhr) {
  var options = opt || {};
  options.textStatus = status;
  options.errorThrown = err;
  self.TriggerInternal(silent, oj.Events.EventType.ERROR, self, xhr, options);
};


/**
 * Saves the current Model object to the data service. Performs a data "update."<br>
 * Events:<p>
 * <ul>
 * <b>change:attr</b>: fired for each attribute changed, if passed in as part of the save: passing the model, name
 * of the changed property, and options<br>
 * <b>change:all</b>: fired after all attributes have been set, if passed in as part of the save: passing the
 * model and options<br>
 * <b>request</b>: fired when the request to save is going to the server, passing the model, xhr object, and
 * options<br>
 * <b>sync</b>: fired when the model is saved to the data service, passing the model and the raw response<br>
 * <b>error</b>: fired if there is an error during the save, passing the model, xhr ajax object, and options<br>
 * </ul>
 * <p>
 * @param {Object=} attributes One or more attribute name/value pairs to set on the model before the save.
 * @param {Object=} options Options to control save<br>
 * <b>success</b>: a user callback called when the save has completed successfully. This makes the save an
 * asynchronous process. The callback is called passing the Model object, response from the AJAX call, and the
 * fetch options argument.<p>
 * <b>error</b>: a user callback function called if the save fails. <p>
 * <b>contentType</b>: in case the user's REST service requires a different POST content type than the default,
 * 'application/json'<p>
 * <b>validate</b>: should the validation routine be called if available<p>
 * <b>wait</b>: if true, wait for the server call before setting the attributes on the model<p>
 * <b>patch</b>: should only changed attributes be sent via a PATCH?<p>
 * <b>attrs</b>: pass a set of attributes to completely control the set of attributes that are saved to the
 * server (generally used with patch)
 * @return {Object|boolean} returns false if validation failed, or the xhr object
 * @memberof oj.Model
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{success?: (model: oj.Model, response: Object, options: Object)=> void,
 *                                                  error?: (model: oj.Model, xhr: Object, options: Object)=> void,
 *                                                  contentType?: string, valdiate?: boolean, wait?: boolean, patch?: boolean, attrs?: Object,
 *                                                  [propName: string]: any}", for: "options"}
 * @export
 */
oj.Model.prototype.save = function (attributes, options) {
  var forceNew;
  var success;
  var callback;
  var self;
  var userErr;
  var patch;
  var argResults = oj.Model._processArgs(arguments);
  var opts;
  var oldAttrs;
  var attrArgs;
  attrArgs = attributes === undefined ? undefined : argResults.attributes;

  var tempOpts = options || {};
  opts = oj.Model._copyOptions(argResults.options);

  var validAttrs = $.extend(true, {}, this.attributes, attrArgs);
  if (!this._checkValid(validAttrs, opts, true)) {
    return false;
  }

  if (!opts.wait) {
    this.set(attrArgs);
  }

  forceNew = opts.forceNew === undefined ? false : opts.forceNew;
  self = this;
  userErr = opts.error;
  patch = opts.patch;

  opts.error = function (xhr, status, err) {
                            // Trigger an error event
    oj.Model._triggerError(self, false, tempOpts, status, err, xhr);

    if (userErr) {
      userErr.apply(self, arguments);
    }
  };

  opts.saveAttrs = opts.wait ? this._attrUnion(attrArgs) : this.attributes;

    // Must temporarily at least set attrs for toJSON()
  oldAttrs = this.attributes;
    // Swap in what's to be saved and call toJSON()
  this.attributes = opts.saveAttrs;
  opts.saveAttrs = this.toJSON();
  this.attributes = oldAttrs;

  if (!forceNew && !this.isNew()) {
    success = opts.success;
    opts.success =
            function (resp) {
              var attrs;

                // Make sure we pass xhr
              if (opts.xhr) {
                tempOpts.xhr = opts.xhr;
              }

              if (resp && !oj.Object.isEmpty(resp)) {
                if ($.isFunction(self.parse)) {
                  attrs = self.parse(resp);
                } else {
                  attrs = resp;
                }

                self.attributes = $.extend(true, self.attributes, attrs);

                // Merge attrs from response/parse into arg attrs if different--server takes priority in case
                // of 'wait'
                if (opts.wait) {
                  Object.keys(attrArgs || {}).forEach(function (prop) {
                    if (Object.prototype.hasOwnProperty.call(attrs, prop)) {
                      // Prioritize the one in attrs
                      attrArgs[prop] = attrs[prop];
                    }
                  });
                }
                self.SetupId();
              }
              oj.Model._fireSyncEvent(self, resp, opts, false);

              if (opts.wait) {
                self.set(attrArgs);
              }

              if (success) {
                success.call(oj.Model.GetContext(opts, self), self, resp, tempOpts);
              }
              self._clearChanged();
            };
        // If caller passes in attrs themselves, just use those
    if (!opts.attrs) {
      if (attrArgs === undefined) {
        opts.attrs = undefined;
      } else {
        opts.attrs = patch ? attrArgs : opts.saveAttrs;
      }
    }
    return oj.Model._internalSync(patch ? 'patch' : 'update', this, opts);
  }

  callback = oj.Model._getSuccess(opts);
  opts.success = function (resp) {
    var attrs;
       // Make sure we pass xhr
    if (opts.xhr) {
      tempOpts.xhr = opts.xhr;
    }
    if (resp && !oj.Object.isEmpty(resp)) {
      if ($.isFunction(self.parse)) {
        attrs = self.parse(resp);
      } else {
        attrs = resp;
      }
      if (!self._checkValid(attrs, opts, true)) {
        return;
      }

      self.attributes = $.extend(true, self.attributes, attrs);

      // Merge attrs from response/parse into arg attrs if different--server takes priority in case of 'wait'
      if (opts.wait) {
        Object.keys(attrArgs || {}).forEach(function (prop) {
          if (Object.prototype.hasOwnProperty.call(attrs, prop)) {
            // Prioritize the one in attrs
            attrArgs[prop] = attrs[prop];
          }
        });
      }
      self.SetupId();
    }
    oj.Model._fireSyncEvent(self, resp, opts, false);
    if (opts.wait) {
      self.set(attrArgs);
    }

    if (callback) {
      callback.call(oj.Model.GetContext(opts, self), self, resp, tempOpts);
    }
    self._clearChanged();
  };

    // If caller passed in attrs, just use those
  if (!opts.attrs) {
    opts.attrs = opts.saveAttrs;
  }

    // Turn on parse flag
  opts.parse = true;

    // Bizarre case tested by backboneJS--if this is a new model, but we're patching, make sure we only save the
    // explicit attrs if passed in by user
  if (patch) {
    opts.saveAttrs = opts.attrs;
  }
  return oj.Model._internalSync('create', this, opts);
};

/**
 * @private
 */
oj.Model.prototype._attrUnion = function (attrs) {
  var attrReturn = {};
  var self = this;

  Object.keys(self.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(self.attributes, prop)) {
      attrReturn[prop] = self.attributes[prop];
    }
  });
  Object.keys(attrs || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(attrs, prop)) {
      attrReturn[prop] = attrs[prop];
    }
  });
  return attrReturn;
};

/**
 * @protected
 */
oj.Model.GetPropValue = function (obj, property) {
  if (obj) {
    if ($.isFunction(obj[property])) {
      return obj[property]();
    }

    return obj[property];
  }
  return $.isFunction(property) ? property() : property;
};

/**
 * @protected
 */
oj.Model.IsComplexValue = function (val) {
  return val && Object.prototype.hasOwnProperty.call(val, 'value') && Object.prototype.hasOwnProperty.call(val, 'comparator');
};

/**
 * Does this model contain all of the given attribute/value pairs?
 * @private
 */
oj.Model.prototype._hasAttrs = function (attrs) {
  var prop;

  for (prop in attrs) { // eslint-disable-line no-restricted-syntax
    if (Object.prototype.hasOwnProperty.call(attrs, prop)) {
      if (!Object.prototype.hasOwnProperty.call(this.attributes, prop)) {
        return false;
      }

      var val = Array.isArray(attrs[prop]) ? attrs[prop] : [attrs[prop]];
      for (var i = 0; i < val.length; i++) {
        if (oj.Model.IsComplexValue(val[i])) {
          var comparator = val[i].comparator;
          var value = val[i].value;
          if (oj.StringUtils.isString(comparator)) {
            throw new Error('String comparator invalid for local where/findWhere');
          }
          if (!comparator(this, prop, value)) {
            return false;
          }
        } else if (attrs[prop] !== this.attributes[prop]) {
          // Array case meaningless here.  Model can't be == value1 and value2
          return false;
        }
      }
    }
  }
  return true;
};

/**
 * Return a function that determines if this model contains all of the property/values in attrs
 *
 * @param {Object} attrs property/value pairs
 * @returns {function(oj.Model):boolean} function taking an oj.Model that returns true if all of attrs are contained within it
 * @memberof oj.Model
 * @since 1.1.0
 * @export
 */
oj.Model.prototype.matches = function (attrs) {
  return (function (model) {
    for (var prop in attrs) { // eslint-disable-line no-restricted-syntax
      if (model.get(prop) !== attrs[prop]) {
        return false;
      }
    }
    return true;
  }(this));
};

/**
 * See if this model contains any of the given attribute/value pairs
 * @protected
 */
oj.Model.prototype.Contains = function (attrs) {
  var attrList = (attrs.constructor === Array) ? attrs : [attrs];
  var i;

  for (i = 0; i < attrList.length; i++) {
    if (this._hasAttrs(attrList[i])) {
      return true;
    }
  }
  return false;
};

/**
 * @private
 */
oj.Model._getSuccess = function (options) {
  return options != null && options.success ? options.success : null;
};

/**
 * @protected
 */
oj.Model.GetContext = function (options, model) {
  if (options !== undefined && options.context !== undefined) {
    return options.context;
  }
  return model;
};

/**
 * Determines if this model object has been assigned an id value yet. This indicates whether or not the model's
 * data has been saved to or fetched from the data service at any point.
 * @returns {boolean} true if the Model object has not had its id set yet, false if not.
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.isNew = function () {
  return this.GetId() === undefined;
};

/**
 * @private
 */
oj.Model.prototype._getIdAttr = function () {
  return this.idAttribute || 'id';
};

/**
 * @protected
 */
oj.Model.prototype.GetId = function () {
  return this.id;
};

/**
 * Return the set of attributes and values that have changed since the last set.  Note that a Model.fetch() uses
 * set to store the returned attribute data. If attribute/value pairs are passed in, check those to see if they're
 * different than the model.
 * Return false if there were no changes
 * @param {Object=} attributes One or more attribute/value pairs to check against the model for changes
 * @return {Object|boolean} the set of all attribute value pairs that have changed since last set, if no
 * attributes passed in; the set of all attribute value pairs that are different than those listed in the
 * attributes parameter, if present.  False if no changes
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.changedAttributes = function (attributes) {
  if (attributes) {
    var internalChanges = {};
    var self = this;
    Object.keys(attributes).forEach(function (prop) {
      if (Object.prototype.hasOwnProperty.call(attributes, prop)) {
        if (!oj.Object.__innerEquals(attributes[prop], self.attributes[prop])) {
          internalChanges[prop] = attributes[prop];
        }
      }
    });
    return oj.Object.isEmpty(internalChanges) ? false : internalChanges;
  }
  return oj.Object.isEmpty(this.changed) ? false : this.changed;
};

/**
 * Return true if the Model object has had any changes made to its values, or if any changes have been made to
 * the optional attribute passed in.
 * @param {string=} attribute attribute to check for changes
 * @returns {boolean} true if the Model object has had any changes since retrieval or last update at all (if no
 * attributes parameter); true if the Model object has had changes to the passed-in attribute
 * since retrieval or last update (if attribute parameter present).
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.hasChanged = function (attribute) {
  if (attribute !== undefined) {
    return oj.Model._hasProperties(this.changed) &&
      Object.prototype.hasOwnProperty.call(this.changed, attribute);
  }
  return oj.Model._hasProperties(this.changed);
};


/**
 * Delete the record represented by this model object from the data service.  If the server responds with virtual
 * mode response properties (such as totalResults), these will be picked up at the end of the delete by the
 * model's collection.  Note that by default, the delete isn't sent with an HTTP data type so the object
 * returning these properties needs to be a string version of the JSON object.<br>
 * Events:<p>
 * <ul>
 * <b>destroy</b>: fired when the model is destroyed, either before or after the server call, depending on the
 * setting of wait.  The model and model's collection are passed in<br>
 * <b>request</b>: fired right as the call is made to request the server delete the model.  Passes the model,
 * xhr ajax object, and options.<br>
 * <b>sync</b>: fired after the server succeeds in destroying the model.  Passes the model, the raw response data,
 * and options.
 * </ul>
 * @param {Object=} options Options for the destroy operation:<br>
 * <b>wait</b>: if true, wait until the server is done to fire the destroy event.  Otherewise, fire immediately
 * and regardless of success or failure<br>
 * <b>success</b>: callback function called if the server call succeeds, passing the model, response data, and
 * options<br>
 * <b>error</b>: callback function on failure of the server call, firing an error event and passing the model,
 * xhr, status, and error values.<br>
 * @return {boolean}
 * @memberof oj.Model
 * @since 1.0.0
 * @ojsignature {target: "Type", value:"{success?: (model: oj.Model, response: Object, options: Object)=> void,
 *                                                  error?: (model: oj.Model, xhr: Object, options: Object)=> void,
 *                                                  wait?: boolean, [propName: string]: any}", for: "options"}
 * @export
 */
oj.Model.prototype.destroy = function (options) {
  var tempOpt = options || {};
  var isWait = tempOpt.wait;
  var callback;
  var userErr = tempOpt.error;
  var self = this;
  var xhr;
  var opts;

  opts = oj.Model._copyOptions(tempOpt);
  callback = oj.Model._getSuccess(opts);
    // Grab the collection off the model in case we need to update
  var collection = this.GetCollection();

  opts.success = function (data) {
    // Make sure we pass xhr
    if (opts.xhr) {
      tempOpt.xhr = opts.xhr;
    }

    // Give an opportunity to update any collection paging properties, like totalResults due to this destroy
    if (collection) {
      // Make sure to parse the data if necessary
      var props = oj.StringUtils.isString(data) && !oj.StringUtils.isEmpty(data) ?
        JSON.parse(data) : data;

      collection._setPagingReturnValues(props, null, data, true);
    }
    if (isWait) {
      self._fireDestroy(false);
    }
    oj.Model._fireSyncEvent(self, data, opts, false);

    if (callback) {
      callback.call(oj.Model.GetContext(opts, self), self, data, tempOpt);
    }
  };
  opts.error = function (xhrParam) {
    // Trigger an error event
    self.TriggerInternal(false, oj.Events.EventType.ERROR, self, xhrParam, tempOpt);

    if (userErr) {
      userErr.apply(self, arguments);
    }
  };

  if (!this.isNew()) {
    xhr = oj.Model._internalSync('delete', this, opts);
    if (!isWait) {
      this._fireDestroy(false);
    }
    return xhr;
  }
  if (!isWait) {
    this._fireDestroy(false);
  }
  if (callback) {
    callback.call(oj.Model.GetContext(opts, self), self, null, tempOpt);
  }
  return false;
};

/**
 * Fire request event
 * @private
 */
oj.Model.prototype._fireRequest = function (model, xhr, options, silent) {
  this.TriggerInternal(silent, oj.Events.EventType.REQUEST, model, xhr, options);
};

/**
 * Fire destroy event to all listeners
 * @private
 */
oj.Model.prototype._fireDestroy = function (silent) {
  this.TriggerInternal(silent, oj.Events.EventType.DESTROY, this, this.collection, null);
};

/**
 * Fire sync event to all listeners
 * @private
 */
oj.Model._fireSyncEvent = function (model, resp, options, silent) {
  model.TriggerInternal(silent, oj.Events.EventType.SYNC, model, resp, options);
};

/**
 * Return a copy of Model's current attribute/value pairs
 * @return {Object} a copy of the Model's current set of attribute/value pairs.
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.toJSON = function () {
  var retObj = {};
  var self = this;

  Object.keys(self.attributes || {}).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(self.attributes, prop)) {
      if (Array.isArray(self.attributes[prop])) {
        retObj[prop] = self.attributes[prop].slice(0);
      } else {
        retObj[prop] = self.attributes[prop];
      }
    }
  });
  return retObj;
};

/**
 * Return the previous value of the given attribute, if any.
 *
 * @param {string} attr
 * @returns {Object} previous value of attr, if any.  If the attribute has not changed, returns undefined
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.previous = function (attr) {
  return this.previousAttrs[attr];
};

/**
 * Return a copy of the model's previously set attributes
 *
 * @returns {Object} a copy of the model's previous attributes
 * @memberof oj.Model
 * @since 1.0.0
 * @export
 */
oj.Model.prototype.previousAttributes = function () {
  return this.previousAttrs;
};

/**
 * @export
 * Performs communications with the server.  Can be overridden/replaced by clients
 *
 * @param {string} method "create", "read", "update", or "delete"
 * @param {oj.Model} model Model to be read/saved/deleted/created
 * @param {Object=} options to control sync:<br>
 * <b>success</b>: called if sync succeeds:<br>
 * <ul>
 * For create, called with the response (attribute/value pairs); status (Ajax by default, but user-defined);
 * xhr object (Ajax--if available)<br>
 * For read, called with the response (attribute/value pairs being fetched)<br>
 * For update, same as create<br>
 * For delete, called with the oj.Model deleted, the data response (ignored), and the options passed to the
 * destroy call<br>
 * </ul>
 * <b>error</b>: called if sync fails.  Called with xhr, status, and error info, per jQuery Ajax (all if
 * available)<p>
 *
 * @return {Object} xhr object
 * @memberof oj.Model
 * @instance
 * @since 1.0.0
 * @alias sync
 */
oj.Model.prototype.sync = function (method, model, options) {
  return oj.sync(method, model, options);
};

/**
 * Internal processing before sync-- we want this stuff to happen even if user replaces sync
 * @private
 */
oj.Model._internalSync = function (method, model, opt) {
  var options = opt || {};
  // If Model/Collection has OAuth object, then create Authorization header (see oj.RestImpl.addOptions)
  if (model.oauth) {
    options.oauthHeader = model.oauth.getHeader();
  }

  // Make sure to transfer the data type if it's set on the calling object
  if (!options.dataType && model.dataType) {
    options.dataType = model.dataType;
  }
  if (!options.jsonpCallback && model.jsonpCallback) {
    options.jsonpCallback = model.jsonpCallback;
  }

  // Do parsing if necessary and tuck it on options
  if (method === 'create' || method === 'patch' || method === 'update') {
    options.parsedData = model.parseSave.call(model, method === 'patch' ? options.attrs :
                                                                               options.saveAttrs);
  }
  var recordId = null;
  if (model instanceof oj.Model) {
    recordId = model.GetId();
  }
  var newOpt = {};
  Object.keys(options || {}).forEach(function (prop) {
    newOpt[prop] = options[prop];
  });
  var urlOpt = oj.Model.SetCustomURLOptions(recordId, model, options);
  Object.keys(urlOpt || {}).forEach(function (prop) {
    newOpt[prop] = urlOpt[prop];
  });
    // Make sure we send back xhr in options-- can come from return value or passed back through options
  options.xhr = model.sync(method, model, newOpt);
  if (newOpt.xhr) {
    options.xhr = newOpt.xhr;
  }
  return options.xhr;
};

/**
 * Get all custom URL options
 * @protected
 */
oj.Model.SetCustomURLOptions = function (recordID, context, opt) {
  var options = context instanceof oj.Collection ? context.ModifyOptionsForCustomURL(opt) : {};
  if (recordID) {
    options.recordID = recordID;
  }
  return options;
};

/**
 * @export
 * @desc Master server access method for all models and collections.  Replace oj.sync with a new implementation
 * to customize all oj.Model and oj.Collection server interactions
 *
 * @param {string} method "create", "read", "update", "patch", or "delete"
 * @param {oj.Model|oj.Collection} model Model (or Collection to be read) to be read/saved/deleted/created
 * @param {Object=} options to control sync<br>
 * success: called if sync succeeds<br>
 * error: called if sync fails<br>
 * others are passed to jQuery<br>
 * Options that would normally be passed to a customURL callback are also included<br>
 *
 *
 * @return {Object} xhr object
 * @memberof oj
 * @global
 * @since 1.0.0
 * @alias oj.sync
 */
oj.sync = function (method, model, options) {
  var tempOpt = options || {};
  function _fireAndReturn(xhr) {
    model._fireRequest(model, xhr, tempOpt, tempOpt.silent);
    return xhr;
  }

  var restService;
  var success = tempOpt.success;
  var error = tempOpt.error;
  var url;

  if (method.valueOf() === 'create') {
    url = model._saveUrl();
    url = url || oj.Model.GetPropValue(model, 'url');
    restService = new oj.RestImpl(url, model);
    return _fireAndReturn(restService.addRecord(tempOpt.parsedData, error, tempOpt, model));
  }

  if (method.valueOf() === 'read') {
    if (model instanceof oj.Model) {
      url = tempOpt.url ? tempOpt.url : oj.Model.GetPropValue(model, 'url');
      restService = new oj.RestImpl(url, model);
      return _fireAndReturn(restService.getRecord(success, error, model.GetId(), tempOpt,
                                  oj.Model.GetContext(tempOpt, model)));
    }
    // Collection fetch
    url = model.GetCollectionFetchUrl(tempOpt);
    restService = new oj.RestImpl(url, model);
    return _fireAndReturn(restService.getRecords(success, error, tempOpt, model));
  }

  restService = new oj.RestImpl(oj.Model.GetPropValue(model, 'url'), model);
  var recordId = null;
  if (model instanceof oj.Model) {
    recordId = model.GetId();
  }
  if (method.valueOf() === 'update') {
    return _fireAndReturn(restService.updateRecord(success, recordId, tempOpt.parsedData,
      error, tempOpt, model, false));
  }
  if (method.valueOf() === 'patch') {
    return _fireAndReturn(restService.updateRecord(success, recordId, tempOpt.parsedData,
      error, tempOpt, model, true));
  }
  if (method.valueOf() === 'delete') {
    return _fireAndReturn(restService.deleteRecord(recordId, error, tempOpt, model));
  }
  return null;
};


/**
 * @private
 */
oj.Model._urlError = function (ajaxOptions) {
  if (!ajaxOptions.url) {
    throw new Error('The url property or function must be specified');
  }
};

/**
 * @export
 * @desc Master Ajax entry point for all oj.Model and oj.Collection server interactions, when they are using the
 * default sync implementations.  oj.ajax passes through to jQuery ajax by default.
 * See {@link http://api.jquery.com/jquery.ajax/} for expected parameters and return value.
 * @param {Object=} settings optional ajax settings
 *
 * @return {Object} xhr object
 * @memberof oj
 * @global
 * @since 1.0.0
 * @alias oj.ajax
 */
oj.ajax = function (settings) { // eslint-disable-line no-unused-vars
  if (arguments && arguments.length > 0) {
    oj.Model._urlError(arguments[0]);
  }
  return $.ajax.apply(oj, arguments);
};

/**
 * Copyright (c) 2014, 2015 Oracle and/or its affiliates.
 * All rights reserved.
 */

/* jslint browser: true*/
/* global jQuery:false*/

/**
 * @export
 * @class oj.OAuth
 * @classdesc Member of Model objects. Object representing name/value pairs for a data service record
 *
 * @param {Object} attributes Initial set of attribute/value pairs with which to seed this OAuth object
 * @param {string} header Actual name for the Authorization header (default 'Authorization')
 * @example <caption>Initialize OAuth with client credentials</caption>
 * var myOAuth = new oj.OAuth('X-Authorization', {...Client Credentials ...});
 *
 * @example <caption>Initialize OAuth with access_token</caption>
 * var myOAuth = new oj.OAuth('X-Authorization', {...Access Token...});
 *
 * @example <caption>Initialize empty OAuth and set access_token</caption>
 * var myOAuth = new oj.OAuth();
 * myOAuth.setAccessTokenResponse({...Access Token...});
 *
 * @constructor
 * @since 1.0.0
*/
oj.OAuth = function (header, attributes) {
  oj.OAuth._init(this, attributes || {}, header || 'Authorization');
};

// Subclass from oj.Object
oj.Object.createSubclass(oj.OAuth, oj.Object, 'oj.OAuth');
oj.OAuth.prototype.Init = function () {
  oj.OAuth.superclass.Init.call(this);
};

/**
 * Calculates Authorization header based on client credentials or access_token
 * @return {Object} OAuth 2.0 Authorization header
 * @example <caption>Get Authorization header</caption>
 * myOAuth.getHeader();
 *
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.getHeader = function () {
  var headers = {};
  if (!this.accessTokenResponse.access_token) {
    this.clientCredentialGrant();
  }
  headers[this.accessTokenRequest.auth_header] = 'Bearer ' + this.accessTokenResponse.access_token;
  return headers;
};

/**
 * Check is OAuth initialized (not null access_token).
 * @return {boolean} true/false
 * @example <caption>Check if OAuth initialized</caption>
 * if(myOAuth.isInitialized()) console.log('Initialized');
 *
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.isInitialized = function () {
  if (this.accessTokenResponse && this.accessTokenResponse.access_token) {
    return true;
  }
  return false;
};

/**
 * Request for access_token(bearer token) using Client Credential Authorization Grant.
 * Initialize response part of the OAuth object (access_token, e.t.c.)
 * @return {undefined}
 * @example <caption>Set/Re-set response part of the OAuth object using Client Credentials</caption>
 * myOAuth.clientCredentialGrant();
 *
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.clientCredentialGrant = function () {
  var headers = {};
  var self = this;
  headers[self.accessTokenRequest.auth_header] = 'Basic ' +
    oj.OAuth._base64_encode(self.accessTokenRequest.client_id + ':' +
    self.accessTokenRequest.client_secret);

  $.ajax({
    type: 'POST',
    async: false,
    url: this.accessTokenRequest.bearer_url,
    data: 'grant_type=client_credentials',
    headers: headers,
    success: function (data) {
      self.accessTokenResponse = oj.OAuth._initAccessToken(self.accessTokenResponse, data);
    },
    error: function (jqXHR) {
      throw new Error(jqXHR.responseText);
    }
  });
};

/**
 * Set response part of the OAuth object (access_token, e.t.c.)
 * @param {Object} data current response
 * @return {undefined}
 * @example <caption>'Initialize' response part of the OAuth object with access_token</caption>
 * myOAuth.setAccessTokenResponse({...Access Token...});
 *
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.setAccessTokenResponse = function (data) {
  this.accessTokenResopnse = oj.OAuth._initAccessToken(this.accessTokenResponse, data);
};

/**
 * Get response part of the OAuth object (access_token, e.t.c.)
 * @return {Object} cached response
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.getAccessTokenResponse = function () {
  return this.accessTokenResponse;
};

/**
 * Clean response part of the OAuth object (access_token, e.t.c.)
 * Null and remove all data from response part of the OAuth object
 * @return {undefined}
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.cleanAccessTokenResponse = function () {
  oj.OAuth._cleanAccessToken(this.accessTokenResponse);
};

/**
 * Set request part of the OAuth object (client credentials, uri endpoint)
 * @param {Object} data current client credentials and uri
 * @return {undefined}
 * @example <caption>'Initialize' request part of the OAuth object with client credentials and calculate
 * access_token</caption>
 * myOAuth.setAccessTokenRequest({...Client Credentials ...});
 * myOAuth.clientCredentialGrant();
 *
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.setAccessTokenRequest = function (data) {
  this.accessTokenRequest = oj.OAuth._initAccessToken(this.accessTokenRequest, data);
};

/**
 * Get request part of the OAuth object (client credentials, uri endpoint)
 * @return {Object} cached request
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.getAccessTokenRequest = function () {
  return this.accessTokenRequest;
};

/**
 * Clean request part of the OAuth object (client credentials, uri endpoint)
 * Null and remove all data from request part of the OAuth object
 * @return {undefined}
 * @memberof oj.OAuth
 * @export
 */
oj.OAuth.prototype.cleanAccessTokenRequest = function () {
  oj.OAuth._cleanAccessToken(this.accessTokenRequest);
};

/**
 * @private
 * @param {Object} oauth
 * @param {Object} attributes
 * @param {string|null} header
 */
oj.OAuth._init = function (oauth, attributes, header) {
  var oa = oauth;
  oa.Init();
  oa.accessTokenRequest = {};
  oa.accessTokenResponse = {};

  if (attributes.access_token) { // access_token has higher preference
    oa.accessTokenResponse = oj.OAuth._initAccessToken(oa.accessTokenResponse, attributes);
  } else if (attributes.client_id && attributes.client_secret && attributes.bearer_url) {
    // Client Credential Grant
    oa.accessTokenResponse = oj.OAuth._initAccessToken(oa.accessTokenRequest, attributes);
  }
  oa.accessTokenRequest.auth_header = header;
};

/**
 * @private
 * @param {Object} oauthObj - Request/Response object to deal with
 * @param {Object} data - object to populate
 */
oj.OAuth._initAccessToken = function (oauthObj, data) {
  var dat = data || {};
  var obj = oauthObj || {};
  Object.keys(dat).forEach(function (prop) {
    if (Object.prototype.hasOwnProperty.call(dat, prop)) {
      obj[prop] = dat[prop];
    }
  });
  return obj;
};

/**
 * @private
 * @param {Object} oauthObj - Request/Response object to deal with
 */
oj.OAuth._cleanAccessToken = function (oauthObj) {
  var obj = oauthObj || {};

  Object.keys(obj).forEach(function (key) {
    if (Object.prototype.hasOwnProperty.call(obj, key)) {
      if (key !== 'auth_header') {
        obj[key] = null;
        delete obj[key];
      }
    }
  });
};

/**
 * @private
 * @param {string} a The data to calculate the base64 representation from
 * @return {string} The base64 representation
 */
oj.OAuth._base64_encode = function (a) {
  var d;
  var e;
  var f;
  var b;
  var g = 0;
  var h = 0;
  var i = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=';
  var c = [];

  do {
    d = a.charCodeAt(g);
    g += 1;
    e = a.charCodeAt(g);
    g += 1;
    f = a.charCodeAt(g);
    g += 1;
    b = (d << 16) | (e << 8) | f; // eslint-disable-line no-bitwise
    d = (b >> 18) & 63; // eslint-disable-line no-bitwise
    e = (b >> 12) & 63; // eslint-disable-line no-bitwise
    f = (b >> 6) & 63; // eslint-disable-line no-bitwise
    b &= 63; // eslint-disable-line no-bitwise
    c[h] = i.charAt(d) + i.charAt(e) + i.charAt(f) + i.charAt(b);
    h += 1;
  } while (g < a.length);
  c = c.join('');
  d = a.length % 3;
  return (d ? c.slice(0, d - 3) : c) + '==='.slice(d || 3);
};

/**
 * Copyright (c) 2014, 2015 Oracle and/or its affiliates.
 * All rights reserved.
 */

/* jslint browser: true*/
/* global jQuery:false*/
/**
 * @private
 * @constructor
 */
oj.RestImpl = function (rootURL, model) {
  this.rootURL = rootURL;
  this.model = model;
  this.customURL = model.customURL;
  $.support.cors = true;
};

oj.RestImpl._HEADER_PROP = 'headers';

// Add the properties in options to starter, if not already there
oj.RestImpl.addOptions = function (starter, options, customOptions) {
  var initial = $.extend(true, starter, customOptions);
  if (options) {
    Object.keys(options || {}).forEach(function (prop) {
      if (Object.prototype.hasOwnProperty.call(options, prop) && prop !== 'oauthHeader') {
        if (!Object.prototype.hasOwnProperty.call(initial, prop)) {
          initial[prop] = options[prop];
        }
        if (prop === oj.RestImpl._HEADER_PROP) {
                  // Deep merge
          initial[prop] = $.extend(true, initial[prop], options[prop]);
        }
      }
    });
  }

  if (options.oauthHeader) {
        // if there are no any headers then create a new one.
    if (!initial[oj.RestImpl._HEADER_PROP]) initial[oj.RestImpl._HEADER_PROP] = {};
    Object.keys(options.oauthHeader || {}).forEach(function (prop) {
      if (Object.prototype.hasOwnProperty.call(options.oauthHeader, prop)) {
        if (!Object.prototype.hasOwnProperty.call(initial[oj.RestImpl._HEADER_PROP], prop)) {
          initial[oj.RestImpl._HEADER_PROP][prop] = options.oauthHeader[prop];
        }
      }
    });
  }

  return initial;
};

oj.RestImpl.prototype.getRecords = function (callback, errFunc, options, context) {
  var opt = options || {};
  var isJsonp = opt.dataType === 'jsonp';
  var urlInfo = this._getURL('read', this.rootURL, this.customURL, null, context, opt);
  var ajaxOptions = {
    crossDomain: opt.crossDomain || !isJsonp,
    dataType: this._getDataType(opt),
    jsonpCallback: opt.jsonpCallback,
    context: context !== null ? context : this,
    success: callback,
    error: errFunc
  };
  ajaxOptions = this._addHeaderProp(ajaxOptions);
  ajaxOptions = oj.RestImpl.addOptions(ajaxOptions, opt, urlInfo);
  opt.xhr = this.ajax(ajaxOptions, context);
  return opt.xhr;
};

oj.RestImpl.prototype._addHeaderProp = function (options) {
  var opt = options || {};
  if (!(this.model && this.model.omitLanguageHeader)) {
    opt[oj.RestImpl._HEADER_PROP] = { 'Accept-Language': this.getLocale() };
  }
  return opt;
};

oj.RestImpl.prototype.getRecord = function (success, error, recordID, options, context) {
  var opt = options || {};
  var isJsonp = opt.dataType === 'jsonp';
  var urlInfo = this._getURL('read', this.rootURL, this.customURL, recordID, context, opt);
  var ajaxOptions = {
    crossDomain: opt.crossDomain || !isJsonp,
    dataType: this._getDataType(opt),
    jsonpCallback: opt.jsonpCallback,
    context: context !== null ? context : this,
    success: success,
    error: error
  };
  ajaxOptions = this._addHeaderProp(ajaxOptions);
  ajaxOptions = oj.RestImpl.addOptions(ajaxOptions, opt, urlInfo);
  opt.xhr = this.ajax(ajaxOptions, context);
  return opt.xhr;
};


oj.RestImpl.prototype.updateRecord = function (callback, recordID, record, error,
  options, context, patch) {
  var opt = options || {};
  var isJsonp = opt.dataType === 'jsonp';
  var urlInfo = this._getURL(patch ? 'patch' : 'update', this.rootURL, this.customURL, recordID, context, opt);
  var emulateHTTP = oj.RestImpl._emulateHTTP(opt);
  var ajaxOptions = {
    crossDomain: opt.crossDomain || !isJsonp,
    contentType: this._getContentType(opt),
    dataType: this._getDataType(opt),
    jsonpCallback: opt.jsonpCallback,
    data: this._getData(JSON.stringify(record), opt, urlInfo),
    emulateHTTP: emulateHTTP,
    emulateJSON: oj.RestImpl._emulateJSON(opt),
    success: callback,
    error: error,
    context: context !== null ? context : this
  };
  ajaxOptions = this._addHeaderProp(ajaxOptions);
  ajaxOptions = oj.RestImpl.addOptions(ajaxOptions, opt, urlInfo);
  ajaxOptions = oj.RestImpl._beforeSendMod(emulateHTTP, ajaxOptions);
  opt.xhr = this.ajax(ajaxOptions, context);
  return opt.xhr;
};

oj.RestImpl._beforeSendMod = function (emulateHTTP, options) {
  var opt = options || {};
  if (emulateHTTP) {
        // Do a before send xhr mod for this case
    var beforeSend = opt.beforeSend;
    opt.beforeSend = function (xhr) {
      xhr.setRequestHeader('X-HTTP-Method-Override', opt._method);
      if (beforeSend) {
        return beforeSend.apply(this, arguments);
      }
      return null;
    };
  }
  return opt;
};

oj.RestImpl.prototype._getData = function (data, options, urlInfo) {
  if (oj.RestImpl._emulateJSON(options)) {
        // Push record and _method into an object
    var retObj = { _method: urlInfo._method ? urlInfo._method : urlInfo.type };
    if (data) {
      retObj.model = data;
    }
    return retObj;
  }
  return data;
};

oj.RestImpl.prototype._getHTTPMethod = function (operation, options) {
  if (options.type) {
    return { method: options.type };
  }
  var method = null;
  if (operation === 'create') {
    method = 'POST';
  }
  if (operation === 'delete') {
    method = 'DELETE';
  }
  if (operation === 'patch') {
    method = 'PATCH';
  }
  if (operation === 'update') {
    method = 'PUT';
  }
  if (oj.RestImpl._emulateHTTP(options)) {
        // Convert method to POST, put original method under data._method
    var retObj = {};
    retObj.method = 'POST';
    retObj._method = method;
    return retObj;
  }
  if (method === null) {
    method = 'GET';
  }
  return { method: method };
};

oj.RestImpl._emulateHTTP = function (options) {
  return options.emulateHTTP || oj.emulateHTTP;
};

oj.RestImpl._emulateJSON = function (options) {
  return options.emulateJSON || oj.emulateJSON;
};

oj.RestImpl.prototype._getURL = function (operation, rootURL, customURL, recordID,
  context, options) {
  var httpMethod = this._getHTTPMethod(operation, options);
  if ($.isFunction(customURL)) {
    var result = customURL.call(this, operation, context,
                                   oj.Model.SetCustomURLOptions(recordID, context, options));
    if (oj.StringUtils.isString(result)) {
      var ret = { url: result, type: httpMethod.method };
      if (httpMethod._method) {
        ret._method = httpMethod._method;
      }
      return ret;
    } else if (result) {
      result.url = Object.prototype.hasOwnProperty.call(result, 'url') ? result.url : rootURL;
      if (!Object.prototype.hasOwnProperty.call(result, 'type')) {
        result.type = httpMethod.method;
      }
      if (!Object.prototype.hasOwnProperty.call(result, 'data')) {
        if (httpMethod._method) {
          result._method = httpMethod._method;
        }
      }
      return result;
    }
  }
  var retObj = { url: oj.Model.GetPropValue(null, rootURL), type: httpMethod.method };
  if (httpMethod._method) {
    retObj._method = httpMethod._method;
  }
  return retObj;
};

oj.RestImpl.prototype.deleteRecord = function (recordID, error, options, context) {
  var opt = options || {};
  var isJsonp = opt.dataType === 'jsonp';
  var urlInfo = this._getURL('delete', this.rootURL, this.customURL, recordID, context, opt);

  var emulateHTTP = oj.RestImpl._emulateHTTP(opt);
  var emulateJSON = oj.RestImpl._emulateJSON(opt);
  var ajaxOptions = {
    crossDomain: opt.crossDomain || !isJsonp,
    success: opt.success,
    error: error,
    context: context !== null ? context : this,
    emulateHTTP: emulateHTTP,
    emulateJSON: emulateJSON
  };
  var data = this._getData(null, opt, urlInfo);
  if (data) {
    ajaxOptions.data = data;
  }
  ajaxOptions = oj.RestImpl.addOptions(ajaxOptions, opt, urlInfo);
  ajaxOptions = oj.RestImpl._beforeSendMod(emulateHTTP, ajaxOptions);
  opt.xhr = this.ajax(ajaxOptions, context);
  return opt.xhr;
};

oj.RestImpl.prototype.addRecord = function (record, error, options, context) {
  var opt = options || {};
  var recordStr = JSON.stringify(record);
  var isJsonp = opt.dataType === 'jsonp';
  var urlInfo = this._getURL('create', this.rootURL, this.customURL, null, context, opt);

  var emulateHTTP = oj.RestImpl._emulateHTTP(opt);
  var ajaxOptions = {
    crossDomain: opt.crossDomain || !isJsonp,
    contentType: opt.contentType || 'application/json',
    dataType: this._getDataType(opt),
    jsonpCallback: opt.jsonpCallback,
    data: this._getData(recordStr, opt, urlInfo),
    success: opt.success,
    error: error,
    emulateHTTP: emulateHTTP,
    emulateJSON: oj.RestImpl._emulateJSON(opt),
    context: context !== null ? context : this
  };
  ajaxOptions = this._addHeaderProp(ajaxOptions);
  ajaxOptions = oj.RestImpl.addOptions(ajaxOptions, opt, urlInfo);
  opt.xhr = this.ajax(ajaxOptions, context);

  return opt.xhr;
};

oj.RestImpl.prototype._getDataType = function (options) {
  if (oj.RestImpl._emulateJSON(options) && !oj.RestImpl._emulateHTTP(options)) {
    return 'application/x-www-form-urlencoded';
  }
  return options.dataType || 'json';
};

oj.RestImpl.prototype._getContentType = function (options) {
  if (oj.RestImpl._emulateJSON(options) && !oj.RestImpl._emulateHTTP(options)) {
    return 'application/x-www-form-urlencoded';
  }

  return options.contentType || 'application/json';
};

oj.RestImpl.prototype.getLocale = function () {
  return oj.Config.getLocale();
};

oj.RestImpl.prototype.ajax = function (settings, collection) {
  if (settings.url === null || settings.url === undefined) {
    throw new oj.URLError();
  }

  var xhr = oj.ajax(settings);
  if (collection._addxhr) {
    collection._addxhr(xhr);
  }
  return xhr;
};


/**
 * Copyright (c) 2014, 2015 Oracle and/or its affiliates.
 * All rights reserved.
 */

/* jslint browser: true*/

/**
 * @constructor
 * @class oj.URLError
 * @classdesc Constructs a URLError, thrown when API calls are made that require a URL but no URL is
 * defined.
 * @since 1.0.0
 * @export
 */
oj.URLError = function () {
  this.name = 'URLError';
  this.message = 'No URL defined';
};
oj.URLError.prototype = new Error();
oj.URLError.constructor = oj.URLError;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Constructs a message object.
 * <p><strong>NOTE:</strong>  Instead of using the constructor, please use an Object
 * that duck-types oj.Message - has summary, detail, and severity properties. 
 * Creating an oj.Message Object provides no additional value than
 * the duck-typed Object.
 * </p>
 * @param {string} summary - Localized summary message text
 * @param {string} detail - Localized detail message text 
 * @param {(number|string)=} severity - An optional severity for this message. Use constants 
 * oj.Message.SEVERITY_LEVEL for number types and oj.Message.SEVERITY_TYPE for string types. Default 
 * is SEVERITY_ERROR if no severity is specified
 * @constructor
 * @export
 * @since 0.6
 * @example <caption>Set application messages using the 
 * <code class="prettyprint">messages-custom</code> attribute. This example creates messages
 * the recommended way, by creating an Object that duck-types oj.Message.</caption>
 * --- HTML ---
 * &lt;oj-input-text messages-custom="{{appMessages}}");>&lt;/oj-input-text>
 * 
 * --- Javascript ---
 * // for messagesCustom property
 * self.appMessages = ko.observable();
 * var msgs = [];
 * msgs.push({summary: "Error Summary", detail: "Error Detail", 
 *  severity: oj.Message.SEVERITY_TYPE['CONFIRMATION']}); 
 * self.appMessages(msgs);
 */
oj.Message = function(summary, detail, severity)
{
  this.Init(summary, detail, severity);
};

/**
 * Indicates the type of severity that the message represents. 
 * @enum {string}
 * @export 
 */
oj.Message.SEVERITY_TYPE = {
  /**
   * Indicates a confirmation that an operation or task was completed. This is the lowest severity 
   * level.
   */
  'CONFIRMATION': 'confirmation',

  /**
   * Indicates information or operation messages. This has a lower severity level than warning.
   */
  'INFO':'info',

  /**
   * Indicates an application condition or situation that might require users' attention. This has a 
   * lower severity than error.
   */
  'WARNING': 'warning',

  /**
   * Used when data inaccuracies occur when completing a field and that needs fixing before user can 
   * continue. This has a lower severity level than fatal.
   * fatal.
   */
  'ERROR': 'error',

  /**
   * Used when a critical application error or an unknown failure occurs. This is the highest 
   * severity level. 
   * @const
   * @export 
   */
  'FATAL': 'fatal'
};


/**
 * Message severity level
 * @enum {number}
 * @export
 */
oj.Message.SEVERITY_LEVEL = {
  'FATAL' : 5,
  'ERROR' : 4,
  'WARNING' : 3,
  'INFO' : 2,
  'CONFIRMATION' : 1
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.Message, oj.Object, "oj.Message");

/**
 * Localized summary text.
 * 
 * @member
 * @name summary
 * @memberof oj.Message
 * @instance
 * @type {string}
 * @default ""
 */

/**
 * Localized detail text.
 * 
 * @member
 * @name detail
 * @memberof oj.Message
 * @instance
 * @type {string}
 * @default ""
 */

/**
 * Severity type of message. See oj.Message.SEVERITY_TYPE for string types and oj.Message.SEVERITY_LEVEL for number types.
 * 
 * @member
 * @name severity
 * @memberof oj.Message
 * @instance
 * @type {string|number}
 * @default oj.Message.SEVERITY_TYPE.ERROR
 */

/**
 * Initializes Message instance with the set options
 * @param {string} summary a localized summary message text
 * @param {string} detail a localized detail message text 
 * @param {number|string=} severity - An optional severity for this message.  Use constants 
 * oj.Message.SEVERITY_LEVEL for number types and oj.Message.SEVERITY_TYPE for string types.
 *                   
 * @export
 * @ignore
 */
oj.Message.prototype.Init = function(summary, detail, severity) 
{
  oj.Message.superclass.Init.call(this);
  this['summary'] = summary;
  this['detail'] = detail;
  this['severity'] = severity || oj.Message.SEVERITY_TYPE.ERROR; // defaults to ERROR
};

/**
 * Whether a message can display on the UI. 
 * 
 * @memberof! oj.Message
 * @returns {boolean} always returns true; 
 * @instance
 * @protected 
 * @since 0.7
 */
oj.Message.prototype.canDisplay = function ()
{
  return true;
};

/**
 * Indicates whether some other oj.Message instance or Object that duck-types it
 * is "equal to" this one. It compares 'severity', 'summary' and 'detail'.
 * 
 * @param {Object} msg 
 * @memberof oj.Message
 * @export
 * @deprecated Since we recommend using duck-typing oj.Message instead of using the constructor,
 *   we also recommend not using this instance-based equals method.
 * @ignore
 */
oj.Message.prototype.equals = function (msg)
{
  if (msg)
  {
    if ((oj.Message.getSeverityLevel(this['severity']) === 
            oj.Message.getSeverityLevel(msg['severity'])) && 
        this['summary'] === msg['summary'] && 
        this['detail'] === msg['detail'])
    {
      return true;
    }
  }
  
  return false;
};

/**
 * Returns an Object that is a clone of this message. 
 * 
 * @return {Object} instance of oj.Message
 * @export
 * @memberof oj.Message
 * @since 0.7
 * @deprecated Since we recommend using duck-typing oj.Message instead of using the constructor,
 *   we also recommend not using this instance-based clone method.
 * @ignore
 */
oj.Message.prototype.clone = function ()
{
  return new oj.Message(this['summary'], this['detail'], this['severity']);
};

/**
 * A convenience method that returns the severity level when given either a severity level of type 
 * number or a severity type of string. 
 * If severity level is not provided or is not valid this returns a severity error.
 * @param {(string|number|undefined)=} severity 
 * @return {number}
 * @memberof oj.Message
 * @public
 * @export
 */
oj.Message.getSeverityLevel = function (severity)
{
  var index;
  if (severity)
  {
    if (typeof severity === "string")
    {
      index = oj.Message._LEVEL_TO_TYPE.indexOf(severity, 1);
      if (index === -1)
      {
        severity = oj.Message.SEVERITY_LEVEL['ERROR'];
      }
      else
      {
        severity = index;
      }
    }
    else if (typeof severity === "number" && (severity < oj.Message.SEVERITY_LEVEL['CONFIRMATION'] || 
          severity > oj.Message.SEVERITY_LEVEL['FATAL']))
    {
      severity = oj.Message.SEVERITY_LEVEL['ERROR'];
    }
  }
  
  return !severity ? oj.Message.SEVERITY_LEVEL['ERROR'] : severity;
};

/**
 * A convenience method that returns the severity type when given either a severity level of type 
 * number or a severity type of string. 
 * If severity level is not provided or is not valid this return a severity error.
 * @param {(string|number|undefined)=} level 
 * @return {string}
 * @memberof oj.Message
 * @public
 * @export
 */
oj.Message.getSeverityType = function (level) 
{
  var index;
  if (level)
  {
    if (typeof level === "string")
    {
      index = oj.Message._LEVEL_TO_TYPE.indexOf(level, 1);
      if (index === -1)
      {
        // when given an unrecognized type return "error"
        level = oj.Message.SEVERITY_TYPE['ERROR'];
      }
    }
    else if (typeof level === "number")
    {
      if (level < oj.Message.SEVERITY_LEVEL['CONFIRMATION'] ||
          level > oj.Message.SEVERITY_LEVEL['FATAL'])
      {
        level = oj.Message.SEVERITY_TYPE['ERROR'];
      }
      else
      {
        level = oj.Message._LEVEL_TO_TYPE[level];
      }
    }
  }
  return level || oj.Message.SEVERITY_TYPE['ERROR'];
};

/**
 * Returns the max severity level in a array of message objects. 
 * 
 * @param {Array.<oj.Message>=} messages an array of message instances or duck typed messages
 * @returns {number} -1 if none can be determined; otherwise a severity level as defined by 
 * oj.Message.SEVERITY_LEVEL.
 * @export
 * @memberof oj.Message
 * @public
 */
oj.Message.getMaxSeverity = function (messages)
{
  var maxLevel = -1, currLevel, message, i;
  if (messages && messages.length > 0)
  {
    $.each(messages, function (i, message)
      {
        if (message)
        {
          currLevel = oj.Message.getSeverityLevel(message['severity']);
        }
        maxLevel = maxLevel < currLevel ? currLevel : maxLevel;
      });
    
  }
  
  return maxLevel;
};

/**
 * Returns false if messages are of severity error or greater.
 * 
 * @param {Array.<oj.Message>} messages an array of message instances or duck-typed messages
 * @returns {boolean} true if none of the messages are of severity error or greater. false otherwise
 * @export
 * @memberof oj.Message
 * @public
 */
oj.Message.isValid = function (messages)
{
  var maxSeverity = oj.Message.getMaxSeverity(messages);
  if (maxSeverity >= oj.Message.SEVERITY_LEVEL.ERROR)
  {
    return false;
  }
  
  return true;
};

/**
 * @private
 * @type Array
 */
oj.Message._LEVEL_TO_TYPE = ['none', // this can never be set
                             oj.Message.SEVERITY_TYPE['CONFIRMATION'], 
                             oj.Message.SEVERITY_TYPE['INFO'],
                             oj.Message.SEVERITY_TYPE['WARNING'],
                             oj.Message.SEVERITY_TYPE['ERROR'],
                             oj.Message.SEVERITY_TYPE['FATAL']];

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Extends oj.Message to represent a component specific message, this defines options that control 
 * how the message will display.
 * 
 * @param {string} summary - Localized summary message text
 * @param {string} detail - Localized detail message text 
 * @param {number|string} severity - An optional severity for this message. Use constants 
 * oj.Message.SEVERITY_LEVEL for number types and oj.Message.SEVERITY_TYPE for string types. Default 
 * is SEVERITY_ERROR if no severity is specified
 * @param {Object} options - an Object literal that defines the following properties
 * @property {string} display - whether the message needs to be displayed immediately or not. 
 * Accepted values are 'immediate', 'deferred'. The default is 'immediate'.
 * @property {string} context - the context the component was in when the validation messages was 
 * added.<p>
 * 
 * NOTE: messages added to the component directly by applications are unknown context and always 
 * shown {display: 'immediate', context: ''}. </p>
 * @private
 * @constructor
 * @since 0.7
 */
oj.ComponentMessage = function (summary, detail, severity, options) 
{
  this.Init(summary, detail, severity, options);
};

// Subclass from oj.Message 
oj.Object.createSubclass(oj.ComponentMessage, oj.Message, "oj.ComponentMessage");

/**
 * Determines whether the message is displayed immediately or not. Deferred messages are not shown 
 * to the user right away but are deferred until component explicitly does. See 
 * {@link oj.editableValue#showMessages}.
 * @private
 * @const
 * @type {Object}
 */
oj.ComponentMessage.DISPLAY = {SHOWN: 'shown', HIDDEN: 'hidden'};

/**
 * The default display options to use when none set.
 * 
 * @type {Object}
 * @private
 */
oj.ComponentMessage._DEFAULT_OPTIONS = {
  'display': oj.ComponentMessage.DISPLAY.SHOWN, 
  'context' : ''};

/**
 * Initializes the strategy based on the display options that specify the messaging artifacts that 
 * will be displayed by this strategy.
 * 
 * @param {string} summary - Localized summary message text
 * @param {string} detail - Localized detail message text 
 * @param {number|string} severity - An optional severity for this message. Use constants 
 * oj.Message.SEVERITY_LEVEL for number types and oj.Message.SEVERITY_TYPE for string types. Default 
 * is SEVERITY_ERROR if no severity is specified
 * @param {Object} options - an Object literal that defines the following properties
 * @memberof! oj.ComponentMessage
 * @instance
 * @protected
 * @ignore
 */
oj.ComponentMessage.prototype.Init = function (summary, detail, severity, options)
{
  oj.ComponentMessage.superclass.Init.call(this, summary, detail, severity);
  
  this._options = $.extend({}, oj.ComponentMessage._DEFAULT_OPTIONS, options);
};
  
/**
 * Indicates whether some other oj.Message instance - msg,  is "equal to" this one.
 * Method is equivalent to java ".equals()" method.
 * 
 * @param {Object} msg 
 * @memberof! oj.ComponentMessage
 * @instance
 * @protected
 * @ignore
 */
oj.ComponentMessage.prototype.equals = function (msg)
{
  var result = oj.ComponentMessage.superclass.equals.call(this, msg);

  // we don;t care what context the messages was added in, as long as they are the same display type
  return result && (this._options['display'] === msg._options['display']);
};

/**
 * Clones this and returns the new instance.
 * 
 * @memberof! oj.ComponentMessage
 * @instance
 * @protected
 * @ignore
 * @returns {Object} 
 */
oj.ComponentMessage.prototype.clone = function ()
{
  return new oj.ComponentMessage(this['summary'], this['detail'], this['severity'], this._options);
};

/**
 * Whether a message can display on the UI.
 * 
 * @memberof! oj.ComponentMessage
 * @returns {boolean} true if messages can be displayed; false if marked as deferred by component.
 * @instance
 * @protected 
 * @ignore
 */
oj.ComponentMessage.prototype.canDisplay = function ()
{
  return !(this._options && this._options['display'] ? 
            this._options['display'] === oj.ComponentMessage.DISPLAY.HIDDEN : false);
};

/**
 * 
 * Called by framework when the message needs to be shown
 * 
 * @returns {boolean} true if shown; false if message display was already shown
 * 
 * @memberof! oj.ComponentMessage
 * @instance
 * @private
 */
oj.ComponentMessage.prototype._forceDisplayToShown = function ()
{
  if (this._options && oj.ComponentMessage.DISPLAY.HIDDEN === this._options['display'])
  {
    this._options['display'] = oj.ComponentMessage.DISPLAY.SHOWN;
    return true;
  }
  
  return false;
};

/**
 * Called by framework to determine if message was added by component versus app.
 * 
 * @returns {boolean} true if messages was added by component. Usually the context is set.
 * 
 * @memberof! oj.ComponentMessage
 * @instance
 * @private
 */
oj.ComponentMessage.prototype._isMessageAddedByComponent = function ()
{
  if (this._options && this._options['context'])
  {
    return true;
  }
  
  return false;
};
/**
 * Custom element bridge prototype.
 * 
 * @class
 * @ignore
 */
oj.BaseCustomElementBridge = {};

/**
 * Prototype for subclasses
 * @ignore
 */
oj.BaseCustomElementBridge.proto = 
{
  getClass: function(descriptor) {
    var proto = Object.create(HTMLElement.prototype);
    this.InitializePrototype(proto);

    var metadata = this.GetMetadata(descriptor);
    // Enumerate metadata to define the prototype properties, methods, and events 
    oj.BaseCustomElementBridge._enumerateMetadataForKey(proto, metadata, 'properties', this.DefinePropertyCallback.bind(this));
    oj.BaseCustomElementBridge._enumerateMetadataForKey(proto, metadata, 'methods', this.DefineMethodCallback.bind(this));
    
    // Add additional element methods not defined in metadata, e.g. getNodeBySubId/getSubIdByNode or get/setProperty
    this.AddComponentMethods(proto);
    
    proto['setProperties'] = function(props){oj.BaseCustomElementBridge.getInstance(this)._setProperties(this, props);}; 

    // The set/unset methods are used for TypeScript only so we should define these as non enumerated properties
    Object.defineProperty(proto, 'set', { value: function(prop, value) { this.setProperty(prop, value); } });
    Object.defineProperty(proto, 'unset', { value: function(prop) { this.setProperty(prop, undefined); } });

    // Add lifecycle listeners
    proto['attributeChangedCallback'] = this._attributeChangedCallback;
    proto['connectedCallback'] = this._connectedCallback;
    proto['disconnectedCallback'] = this._detachedCallback;
    
    Object.defineProperty(proto, 'resolveBP', 
        {value: function(bp){oj.BaseCustomElementBridge.getInstance(this)._resolveBindingProvider(bp);}});
    
    var constructorFunc = function()
    { 
      var reflect = window['Reflect'];
      var ret;
      if (typeof reflect !== 'undefined')
      {
        ret = reflect['construct'](HTMLElement, [], this['constructor']);
      }
      else
      {
        ret = HTMLElement.call(this);
      }
      return ret;
    };
    
    var bridge = this;
    Object.defineProperty(constructorFunc, 'observedAttributes', 
    {
      'get': function() 
      { 
        return bridge.GetAttributes(metadata);
      }
    });

    Object.defineProperty(proto,  'constructor', { 'value': constructorFunc, 'writable': true, 'configurable': true });
    constructorFunc.prototype = proto;
    Object.setPrototypeOf(constructorFunc, HTMLElement);

    return constructorFunc;
  },

  resolveDelayedReadyPromise: function(element)
  {
    this.GetDelayedReadyPromise().resolvePromise();
  },  

  AddComponentMethods: function(proto) {},

  CreateComponent: function(element) {},

  DefineMethodCallback: function (proto, method, methodMeta) {},
  
  DefinePropertyCallback: function (proto, property, propertyMeta) {},

  /**
   * Returns a promise that will be resolved when the component has been initialized.
   * @return {Promise}
   */
  GetDelayedReadyPromise: function()
  {
    if (!this._delayedReady)
      this._delayedReady = new oj.BaseCustomElementBridge.__DelayedPromise();
    return this._delayedReady;
  },

  GetAttributes: function(metadata)
  {
    return oj.BaseCustomElementBridge.getAttributes(metadata['properties']);
  },

  GetMetadata: function(descriptor)
  {
    return descriptor[oj.BaseCustomElementBridge.DESC_KEY_META];
  },

  /**
   * Returns the aliased component property for a given custom element property,
   * e.g. readOnly for oj-switch's readonly custom element property. 
   * Will return the original property if there is no aliasing.
   * @param {string} property The property to check
   */
  GetAliasForProperty: function(property)
  {
    return property;
  },

  HandleAttributeChanged: function(element, attr, oldValue, newValue) {},

  HandleBindingsApplied: function(element, bindingContext) {},

  HandleDetached: function(element) {
    this._bConnected = false;
    if (!this._complete)
    {
      this._resolveBusyState(element);
    }
  },

  HandleReattached: function(element) {},

  InitializeElement: function(element) {},

  InitializePrototype: function(proto) {},
  
  BatchedPropertySet: function(elem, props) 
  {
    var keys = Object.keys(props);
    
    for (var i=0; i<keys.length; i++)
    {
      var key = keys[i];
      elem['setProperty'](key, props[key]);
    }
  },

  CanSkipTypeCheck: function(propAr)
  {
    // TODO: Remove method post 4.0.0 after metadata is generated from jsDoc.
    // Temporary fix for JET component translation subproperties that are not specified in metadata.
    return false;
  },

  GetProperty: function(element, prop, props)
  {
    var event = oj.__AttributeUtils.eventListenerPropertyToEventType(prop);
    var meta = oj.BaseCustomElementBridge.__GetPropertyMetadata(prop, oj.BaseCustomElementBridge.getProperties(this, element));

    // For event listener and non component properties, retrieve the value directly stored on the element.
    // For top level properties, this will delegate to our 'set' methods so we can handle default values.
    // props is the properties object we pass the definitional element or composite ViewModel
    if (event || !meta || prop.indexOf('.') === -1)
      return props[prop];
    else
      return oj.BaseCustomElementBridge.__GetProperty(props, prop);
  },

  InitializeBridge: function(element)
  {
    // Initialize property storage and other variables needed for property sets. 
    // Since early property sets can occur before the element's connected callback
    // is triggered we can't rely on performing this logic there. The cases where
    // the connected callback isn't called before a property set can occur if the 
    // custom element is programatically created and sets are done before adding 
    // the element to the DOM or if the element is stamped by knockout and its expressions
    // processed disconnected as in the case for oj-bind-for-each.
    var descriptor = oj.BaseCustomElementBridge.__GetDescriptor(element.tagName);
    this.METADATA = this.GetMetadata(descriptor);
    this._eventListeners = {};
  },

  PlaybackEarlyPropertySets: function(element)
  {
    this._bCanSetProperty = true;
    if (this._earlySets)
    {
      while (this._earlySets.length)
      {
        var setObj = this._earlySets.shift();
        element.setProperty(setObj.property, setObj.value);
      }
    }
  },

  /**
   * Returns true if this property set was handled as an early property set.
   * An early property set is any set that occurs before the component is created.
   * These sets do not trigger [property]Changed events.
   * @param {string} prop
   * @param {*} value
   * @return {boolean}
   */
  SaveEarlyPropertySet: function(prop, value)
  {
    // Do not save sets that occur during oj.BaseCustomElementBridge.__InitProperties 
    // or expression evaluation. We can process these as normal. Playback occurs before the 
    // binding provider promise is resolved leading to component creation.
    if (this.__INITIALIZING_PROPS || this._bCanSetProperty)
      return false;

    if (!this._earlySets)
      this._earlySets = [];

    this._earlySets.push({property: prop, value: value});
    return true;
  },

  SetEventListenerProperty: function(element, property, value) {
    var event = oj.__AttributeUtils.eventListenerPropertyToEventType(property);
    // Get event listener
    var eventListener = this._eventListeners[event];
    if (!eventListener) {
      // Create the wrapper
      eventListener = this._createEventListenerWrapper();      
      this._eventListeners[event] = eventListener;
      element.addEventListener(event, eventListener);
    }
    if (value == null || value instanceof Function) {
      eventListener.setListener(value);
    }
    else {
      oj.BaseCustomElementBridge.__ThrowTypeError(element, property, value, 'function');
    }
  },

  SetProperty: function(element, prop, value, props, bOuter)
  {      
    // Check value against any defined enums
    var event = oj.__AttributeUtils.eventListenerPropertyToEventType(prop);
    var meta = oj.BaseCustomElementBridge.__GetPropertyMetadata(prop, oj.BaseCustomElementBridge.getProperties(this, element));
    if (event || !meta) 
    {
      element[prop] = value;
    }
    else
    {
      // props is the properties object we pass the definitional element or composite ViewModel
      var previousValue = element['getProperty'](prop);
      var propPath = prop.split('.');
      var topProp = propPath[0]
      // If the top level property is an object, make a copy otherwise the old/new values 
      // will be the same.
      var topPropPrevValue = props[topProp];
      if (oj.CollectionUtils.isPlainObject(topPropPrevValue))
        topPropPrevValue = oj.CollectionUtils.copyInto({}, topPropPrevValue, undefined, true);

      if (!oj.BaseCustomElementBridge.__CompareOptionValues(prop, meta, value, previousValue))
      {
        if (bOuter)
        {
          // This ultimately triggers our element defined property setter
          this.ValidateAndSetProperty(this.GetAliasForProperty.bind(this), props, prop, value, element);
        }
        else
        {
          // Skip validation for inner sets so we don't throw an error when updating readOnly writeable properties
          oj.BaseCustomElementBridge.__SetProperty(this.GetAliasForProperty.bind(this), props, prop, value);
        }

        // Property change events for top level properties will be triggered by ValidateAndSetProperty so avoid firing twice
        if (prop.indexOf('.') !== -1)
        {
          var subprop = {};
          subprop['path'] = prop;
          subprop['value'] = value;
          subprop['previousValue'] = previousValue;
          // Pass the top level property value/previousValues
          var updatedFrom = bOuter ? 'external' : 'internal';
          oj.BaseCustomElementBridge.__FirePropertyChangeEvent(element, topProp, props[topProp], topPropPrevValue, updatedFrom, subprop);
        }
      }
    }
  },

  ValidateAndSetProperty: function(propNameFun, componentProps, property, value, element)
  {
    value = this.ValidatePropertySet(element, property, value);
    oj.BaseCustomElementBridge.__SetProperty(propNameFun, componentProps, property, value);
  },

  ValidatePropertySet: function(element, property, value)
  {
    var propsMeta = oj.BaseCustomElementBridge.getProperties(this, element);
    var propMeta = oj.BaseCustomElementBridge.__GetPropertyMetadata(property, propsMeta);
    var propAr = property.split('.');
     
    // TODO: Remove canSkip checks post 4.0.0 after metadata is generated from jsDoc.
    // Temporary fix for JET component translation subproperties that are not specified in metadata.
    var canSkipTypeCheck = this.CanSkipTypeCheck(propAr);
    if (!propMeta && !canSkipTypeCheck) 
    {
      oj.Logger.warn(oj.BaseCustomElementBridge.getElementInfo(element) + ": Ignoring property set for undefined property '" + property + "'.");
      return;
    }

    // Check readOnly property for top level property
    if (propsMeta[propAr[0]]['readOnly'])
    {
      this._throwError(element, "Read-only property '" + property + "' cannot be set.");
    }

    oj.BaseCustomElementBridge.__CheckEnumValues(element, property, value, propMeta);
    
    // TODO support checking for null values once we generate metadata from jsDoc and have accurate info
    // about component support for undefined/null
    if (!canSkipTypeCheck && value != null)
      return oj.BaseCustomElementBridge.__CheckType(element, property, value, propMeta);

    return value;
  },

  _attributeChangedCallback: function(attr, oldValue, newValue) 
  {
    var bridge = oj.BaseCustomElementBridge.getInstance(this);

    // The 1.0 custom element polyfill calls attribute changed callback on object creation which we want to ignore.
    // Also skip setProperty if the method isn't defined which should never be the case, but has occured in 
    // ojdatepicker Safari tests () so add this safety check to be sure. Best guess is test setup/polyfill issue.
    if (bridge._bCreateCalled && this['setProperty'])
    {
      var prop = oj.__AttributeUtils.attributeToPropertyName(attr);
      var propMeta = oj.BaseCustomElementBridge.__GetPropertyMetadata(prop, oj.BaseCustomElementBridge.getProperties(bridge, this));

      oj.BaseCustomElementBridge.__CheckOverlappingAttribute(this, attr);

      // removeAttribute calls return null as the newValue which we want to treat as
      // a property unset and convert to undefined. We allow property null sets as an
      // actual property override.
      if (newValue === null)
        newValue = undefined;

      var params = {
        'detail': {'attribute': attr, 'value': newValue, 'previousValue': oldValue}
      };
      this.dispatchEvent(new CustomEvent('attribute-changed', params));

      var expression = oj.__AttributeUtils.getExpressionInfo(newValue).expr;
      if (!expression)
      {
        if (propMeta)
          this['setProperty'](prop, oj.BaseCustomElementBridge.__ParseAttrValue(this, attr, prop, newValue, propMeta));

        // This allows subclasses to handle special cases like global transfer 
        // attributes for JET components
        bridge.HandleAttributeChanged(this, attr, oldValue, newValue);
      }
    }
  },

  _connected: function(element)
  {
    this._bConnected = true;
    if (!this._bCreateCalled) // initial attach
    {
      this._bCreateCalled = true;
      
      this._registerBusyState(element);
      
      this._monitorReadyPromise(element);
      
      this.InitializeElement(element);
      
      var self = this;
      
      this._getBindingProvider(element).then(
        function() 
        {
          self.CreateComponent(element);
        
        });
    }
    else
    {
	  // If the component had been previosly disconnected, and the 'ready'
	  // promise is still not resolved, we need to re-register the busy state
      if (!this._complete)
      {
        this._registerBusyState(element);
      }
      this.HandleReattached(element);
    }
  },

  _connectedCallback: function()
  {
    var bridge = oj.BaseCustomElementBridge.getInstance(this);
    bridge._connected(this);
  },

  _detachedCallback: function()
  {
    var bridge = oj.BaseCustomElementBridge.getInstance(this);
    bridge.HandleDetached(this);
  },

  // This wrapper does not provide any additional functionality to event listener functions
  // It exists solely to preserve event listener order
  _createEventListenerWrapper: function() {
    var eventListener;
    var domListener = function(event) {
      if (eventListener) {
        eventListener(event);
      }
    };
    domListener.setListener = function(listener) {
      eventListener = listener;
    }
    return domListener;    
  },

  _monitorReadyPromise: function(element)
  {
    var self = this;
    
    var completeHandler = function()
    {
  	  // If the component is disconnected, the busy state
  	  // must be already resolved
      if (self._bConnected)
      {
         self._resolveBusyState(element);
      }
      self._complete = true;
    };
    
    this.GetDelayedReadyPromise().getPromise().then(
      function() {
        // Add marker class to unhide components
        element.classList.add('oj-complete');
        completeHandler();
      },
      function() {
        // Add marker class to mark that there was an error duing upgrade so consumers like
        // VBCS can apply their own styling to incorrectly setup custom elements.
        element.classList.add('oj-incomplete');
        completeHandler();
      }
    );
  },

  _registerBusyState: function(element)
  {
    var busyContext = oj.Context.getContext(element).getBusyContext();
    this._initCompleteCallback = busyContext.addBusyState({'description' : oj.BaseCustomElementBridge.getElementInfo(element) + " is being upgraded."});
  },

  _resolveBusyState: function(element)
  {
    var callback = this._initCompleteCallback;
    if (!callback)
    {
      this._throwError(element, "Unexpected call to _resolveBusyState().");
    }
    
    this._initCompleteCallback = null;
    callback();
  },

  _setProperties: function(elem, props)
  {
    var mutationKeys = []; // keys for the 'dot mutation' properties
    var regularProps = {}; // the rest of the properties
    var hasRegularProps = false;
    
    var keys = Object.keys(props);
    for (var i=0; i<keys.length; i++)
    {
      var key = keys[i];
      if (key.indexOf('.') >= 0)
      {
        mutationKeys.push(key);
      }
      else
      {
        regularProps[key] = props[key];
        hasRegularProps = true;
      }
    }
    
    // Regular property updates may be batched
    if (hasRegularProps)
    {
      this.BatchedPropertySet(elem, regularProps);
    }
    
    // 'Dot notation' properties can only be set individually for now
    for (var p=0; p<mutationKeys.length; p++)
    {
      var mkey = mutationKeys[p];
      elem['setProperty'](mkey, props[mkey]);
    }
  },

  _throwError: function (elem, msg)
  {
    this.GetDelayedReadyPromise().rejectPromise();
    throw new Error(oj.BaseCustomElementBridge.getElementInfo(elem) + ": " + msg);
  },
  
  _resolveBindingProvider: function(provider)
  {
    if (this._bpResolve)
    {
      this._bpResolve(provider);
    }
    else
    {
      this._bpInst = provider;
    }
  },
  
  _setBpResolver: function(resolve)
  {
    this._bpResolve = resolve;
  },
  
  _getBindingProvider: function(element)
  { 
    var name = this._getBindingProviderName(element);
    
    if (name === oj.BaseCustomElementBridge._NO_BINDING_PROVIDER)
    {
      this.PlaybackEarlyPropertySets(element);
  
      return Promise.resolve(null);
    }
    else if (name == "knockout")
    {
      if (this._bpInst)
      {
        return Promise.resolve(this._bpInst);
      }
      else
      {
        return new Promise(this._setBpResolver.bind(this));
      }
    }
    else
    {
      this._throwError(element, "Unknown binding provider '" + name + "'.");
    }
  },
  
  _getBindingProviderName:function(element)
  {
    var cachedProp = "_ojBndgPrv";
    
    var name = element[cachedProp];
    if (name)
    {
      return name;
    }
    
    name = element.getAttribute("data-oj-binding-provider") || 
                oj.BaseCustomElementBridge._getCompositeBindingProviderName(element);
    
    if (!name)
    {
      var parent = element.parentElement;
      if (parent == null)
      {
        if (element === document.documentElement)
        {
          name = "knockout"; // the default
        }
        else
        {
          this._throwError(element, "Cannot determine binding provider for a disconnected subtree.");
        }
      }
      else
      {
        name = this._getBindingProviderName(parent);
      }
    }
    // cache provider name as a non-enumerable property
    Object.defineProperty(element, cachedProp, {'value': name});
    
    return name;
  }
 
  
};

/*************************/
/* PUBLIC STATIC METHODS */
/*************************/

/**
 * Returns the attributes including the dot notation versions of all complex properties
 * not including readOnly properties.
 * @param {Object} props The properties object
 * @return {Array}
 * @ignore
 */
oj.BaseCustomElementBridge.getAttributes = function(props) 
{
  var attrs = [];
  oj.BaseCustomElementBridge._getAttributesFromProperties('', props, attrs);
  return attrs;
};

/**
 * Helper method for Returns the attributes including the dot notation versions of all complex attributes
 * stored on a bridge instance
 * @param {string} propName The property to evaluate
 * @param {Object} props The properties object
 * @param {Array} attrs The attribute array to add to
 * @ignore
 */
oj.BaseCustomElementBridge._getAttributesFromProperties = function(propName, props, attrs) 
{
  if (props)
  {
    for (var prop in props)
    {
      var propMeta = props[prop];
      if (!propMeta['readOnly']) 
      {
        var concatName = propName + prop;
        attrs.push(oj.__AttributeUtils.propertyNameToAttribute(concatName));
        var type = propMeta['type'];
        if (type && type.toLowerCase() === 'object')
        {
          oj.BaseCustomElementBridge._getAttributesFromProperties(concatName + '.', propMeta['properties'], attrs);
        }
      }
    }
  }
};

/**
 * Returns a string including the element tag name and id for use in error messages and logging.
 * @param {Element} element The element to get the information for.
 * @return {string}
 * @ignore
 */
oj.BaseCustomElementBridge.getElementInfo = function(element) 
{
  if (element)
    return element.tagName.toLowerCase() + " with id '" + element.id + "'";
  return "";
};

/**
 * Returns the bridge instance for an element.
 * @ignore
 */
oj.BaseCustomElementBridge.getInstance = function(element)
{
  var instance =  element[oj.BaseCustomElementBridge._INSTANCE_KEY];
  if (!instance)
  {
    var info = oj.BaseCustomElementBridge._registry[element.tagName.toLowerCase()];
    if (!info)
    {
      throw new Error(oj.BaseCustomElementBridge.getElementInfo(element) + " Attempt to interact with the custom element before it has been registered.");
    }
    
    instance = Object.create(info.bridgeProto);
    instance.InitializeBridge(element);
    Object.defineProperty(element, oj.BaseCustomElementBridge._INSTANCE_KEY, {'value': instance});
  }
  
  return instance;
};

/**
 * Returns the properties stored on a bridge instance
 * @param  {Object} bridge The bridge instance
 * @param  {Object} element The element instance
 * @return {Object}
 * @ignore
 */
oj.BaseCustomElementBridge.getProperties = function(bridge, element) 
{
  return bridge.METADATA['properties'];
};

/**
 * Returns an object if JET component tag has been registered, null otherwise.
 * @param  {string}  tagName The tag name to look up
 * @return {Object|null} True if the component module has been loaded and registered
 * @ignore
 */
oj.BaseCustomElementBridge.getRegistered = function(tagName) 
{
  if (tagName)
  {
    var info = oj.BaseCustomElementBridge._registry[tagName.toLowerCase()];
    if (info)
    {
      return {composite: info.composite};
    }
  }
  
  return null;
};

/**
 * Returns the slot map of slot name to slotted child elements for a given custom element.
 * If the given element has no children, this method returns an empty object.
 * Note that the default slot name is mapped to the empty string.
 * @param  {Element} element The custom element
 * @param {boolean} isComposite True if we are returning the slot map for a composite
 * @return {Object} A map of the child elements for a given custom element.
 * @ignore
 */
oj.BaseCustomElementBridge.getSlotMap = function (element, isComposite) {
  var slotMap = {};
  var childNodeList = element.childNodes;
  for (var i = 0; i < childNodeList.length; i++) {
    var child = childNodeList[i];
    // Only assign Text and Element nodes to a slot
    if (oj.BaseCustomElementBridge.isSlotAssignable(child)) {
      // Ignore text nodes that only contain whitespace
      if (child.nodeType !== 3 || child.nodeValue.trim()) {
        // Text nodes and elements with no slot attribute map to the default slot
        var savedSlot = isComposite ? child.__oj_slots : null;
        var slot = savedSlot != null ? savedSlot : child.getAttribute && child.getAttribute('slot');
        if (!slot) {
          slot = '';
        }
 
        if (!slotMap[slot]) {
          slotMap[slot] = [];
        }
        slotMap[slot].push(child);
      }
    }
  }
  return slotMap;
};

/**
 * Returns true if an element is slot assignable.
 * @param {Element} node The element to check
 * @return {boolean}
 * @ignore
 */
oj.BaseCustomElementBridge.isSlotAssignable = function(node)
{
  return node.nodeType === 1 || node.nodeType === 3;
};

/*****************************/
/* NON PUBLIC STATIC METHODS */
/*****************************/

/**
 * @ignore
 */
oj.BaseCustomElementBridge._NO_BINDING_PROVIDER = "none";

/**
 * @ignore
 */
oj.BaseCustomElementBridge._enumerateMetadataForKey = function(proto, metadata, key, callback)
{
  if (!metadata || !metadata[key])
    return;

  var values = metadata[key];
  var names = Object.keys(values);
  names.forEach(
    function(name) { 
      callback(proto, name, values[name]); 
    }
  );
};



/**
 * Returns a binding provder name if the element is managed by a JET composite
 * @ignore
 */
oj.BaseCustomElementBridge._getCompositeBindingProviderName = function(element)
{
  var name = oj.Composite ? oj.Composite.getBindingProviderName(element.parentElement) : null;
  return name;
};


/**
 * Verify metadata for required properties
 * @ignore
 */
oj.BaseCustomElementBridge._verifyMetadata = function(tagName, metadata)
{
  if (metadata)
  {
    // Verify that declared properties don't override any global HTML element properties
    var properties = metadata['properties'];
    if (properties)
    {
      // We are not currently checking for redefined aria-*, data-*, or event handler attributes, e.g. onclick.
      var globals = oj.BaseCustomElementBridge._GLOBAL_PROPERTIES;
      for (var i = 0; i < globals.length; i++)
      {
        if (properties[globals[i]])
          oj.Logger.error("Error registering composite %s. Redefined global HTML element attribute '%s' in metadata.", tagName, globals[i]);
      }
    }
  }
};

/**
 * Checks to see whether a value is valid for an element property and throws an error if not.
 * @param  {Element}  element The custom element
 * @param  {string}  property The property to check
 * @param  {string}  value The property value
 * @param  {Object}  metadata The property metadata
 * @ignore
 */
oj.BaseCustomElementBridge.__CheckEnumValues = function(element, property, value, metadata)
{
  // Only check enum values for string types
  if (typeof value == 'string' && metadata)
  {
    var enums = metadata["enumValues"];
    if (enums && enums.indexOf(value) === -1)
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(element);
      bridge._throwError(element, "Invalid value '" + value + "' found for property '" + property + 
        "'. Expected one of the following '" + enums.toString() + "'.");
    }
  }
};


/**
 * @ignore
 */
oj.BaseCustomElementBridge.__CheckType = function(element, property, value, metadata)
{
  // We currently support checking of single typed properties of type: string, number,
  // boolean, Array, Object OR properties w/ two possible types where
  // the value can either be of type string|Array, string|Object, string|function, Array|Promise.
  // Any other types are currently skipped, but can be validated by the component in a future ER.
  var type = metadata['type'];
  if (type)
  {
    type = type.toLowerCase();

    var typeAr = type.split('|');
    var typeOf = typeof value;
    if (typeAr.length === 1)
    {
      if ((type.substring(0, 5) === 'array' && !Array.isArray(value)) ||
          (type.substring(0, 6) === 'object' && typeOf !== 'object') ||
          (type === 'number' && !(typeof value === 'number' && isFinite(value))) || // Number.isFinite isn't availabe on IE11
          (type === 'string' && typeOf !== 'string'))
      {
        oj.BaseCustomElementBridge.__ThrowTypeError(element, property, value, type);
      }

      // Treat boolean property sets like the DOM does where any value that passes 
      // 'if (boolVal)' results in a true prop set 
      if (type === 'boolean')
        value = !!value;
    }
    else if (typeAr.length === 2)
    {
      var strIdx = typeAr.indexOf('string');
      var promiseIdx = typeAr.indexOf('promise');
      if (strIdx !== -1 && typeOf !== 'string')
      {
        var otherType = strIdx === 0 ? typeAr[1] : typeAr[0];
        if ((otherType === 'function' && typeOf !== 'function') ||
            (otherType.substring(0, 5) === 'array' && !Array.isArray(value)) ||
            (otherType.substring(0, 6) === 'object' && typeOf !== 'object'))
        {
          oj.BaseCustomElementBridge.__ThrowTypeError(element, property, value, type);
        }
      }
      else if (promiseIdx !== -1 && !(value instanceof Promise))
      {
        var otherType = promiseIdx === 0 ? typeAr[1] : typeAr[0];
        if ((otherType.substring(0, 5) === 'array' && !Array.isArray(value)))
        {
          oj.BaseCustomElementBridge.__ThrowTypeError(element, property, value, type);
        }
      }
    }
  }
  return value;
};

/**
 * Compares two values, returning true if they are equal. Does a deeper check for writeback values
 * because we can't prevent knockout from triggering a second property set with the same values
 * when writing back, but we do want to prevent the addtional update and property changed event.
 * @ignore
 */
oj.BaseCustomElementBridge.__CompareOptionValues = function(property, metadata, value1, value2)
{
  if (metadata['writeback'])
    return oj.Object.compareValues(value1, value2);
  else
    return value1 === value2;
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__ThrowTypeError = function(element, property, value, type)
{
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  bridge._throwError(element, "Invalid type '" + (typeof value) + "' found for property '" + 
    property + "'. Expected value of type '" + type + "'.");
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__GetDescriptor = function(tagName) 
{
  return oj.BaseCustomElementBridge._registry[tagName.toLowerCase()].descriptor;
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__GetCache = function(tagName) 
{
  if (tagName)
    return oj.BaseCustomElementBridge._registry[tagName.toLowerCase()].cache;
  return null;
};

/**
 * Checks to see if there are any overlapping attribute for the given element and attribute
 * @ignore
 */
oj.BaseCustomElementBridge.__CheckOverlappingAttribute = function(element, attr) 
{
  var attrPath = attr.split('.');
  if (attrPath.length > 1)
  {
    attrPath.pop();
    while (attrPath.length)
    {
      var attrSubPath = attrPath.join('.');
      if (element.hasAttribute(attrSubPath))
      {
        var bridge = oj.BaseCustomElementBridge.getInstance(element);
        bridge._throwError(element, "Cannot set overlapping attributes '" + attr + "' and '" + attrSubPath + "'.");
      }
      attrPath.pop();
    }
  }
};

/**
  * Returns the metadata for the property, walking down the metadata hierarchy
  * for subproperties.
  * @param {string} prop The property including dot notation if applicable
  * @param {Object} metadata The component metadata
  * @return {Object|null} The metadata for the property or subproperty or 
  *                       null if not a component property, e.g. a global attribute
  * @ignore
  */
oj.BaseCustomElementBridge.__GetPropertyMetadata = function(prop, metadata) 
{
  var meta = metadata;
  var propAr = prop.split('.');
  for (var i = 0; i < propAr.length; i++)
  {
    meta = meta[propAr[i]];
    if (!meta)
      break;
    
    if (propAr.length > 1 && i < propAr.length -1)
    {
      meta = meta['properties'];
      if (!meta)
        break;
    }
  }
  return meta;
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__InitProperties = function(element, componentProps)
{
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  bridge.__INITIALIZING_PROPS = true;
  var metaProps = oj.BaseCustomElementBridge.getProperties(bridge, element);
  if (metaProps)
  {
    var attrs = element.attributes; // attrs is a NodeList
    for (var i = 0; i < attrs.length; i++)
    {
      var attr = attrs[i];
      var property = oj.__AttributeUtils.attributeToPropertyName(attr.nodeName);

      // See if attribute is a component property
      var meta = oj.BaseCustomElementBridge.__GetPropertyMetadata(property, metaProps);
      if (!meta || meta['readOnly'])
        continue;

      // If complex property, check if there are any overlapping attributes
      oj.BaseCustomElementBridge.__CheckOverlappingAttribute(element, attr.nodeName);

      var info = oj.__AttributeUtils.getExpressionInfo(attr.value);
      if (!info.expr)
      {
        var value = oj.BaseCustomElementBridge.__ParseAttrValue(element, attr.nodeName, property, attr.value, meta);
        bridge.ValidateAndSetProperty(bridge.GetAliasForProperty.bind(bridge), componentProps, property, value, element);
      }
    }
  }
  bridge.__INITIALIZING_PROPS = false;
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__SetProperty = function(propNameFun, componentProps, property, value)
{
  var propsObj = componentProps;
  var propPath = property.split('.');
  var branchedProps;
  // Set subproperty, initializing parent objects along the way unless the top level
  // property is not defined since setting it to an empty object will trigger a property changed
  // event. Instead, branch and set at the end. We only have listeners on top level properties
  // so setting a subproperty will not trigger a property changed event along the way.
  var topProp = propNameFun(propPath[0]);
  if (propPath.length > 1 && !componentProps[topProp])
  {
    branchedProps = {};
    propsObj = branchedProps;
  }

  // Walk to the correct location
  for (var i = 0; i < propPath.length; i++)
  {
    var subprop = propNameFun(propPath[i]);
    if (i === propPath.length - 1)
      propsObj[subprop] = value;
    else if (!propsObj[subprop]) {
      propsObj[subprop] = {};
    }
    propsObj = propsObj[subprop];
  }

  // Update the original component properties if we branched
  if (branchedProps)
    componentProps[topProp] = branchedProps[topProp];
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__GetProperty = function(componentProps, property)
{
  var propsObj = componentProps;
  var propPath = property.split('.');
  for (var i = 0; i < propPath.length; i++)
  {
    var subprop = propPath[i];
    // If no metadata is passed in, assume that the value has already been evaluated
    if (i === propPath.length - 1)
      return propsObj[subprop];
    else if (!propsObj[subprop])
      return undefined;
    propsObj = propsObj[subprop];
  }
};

/**
 * Returns the coerced attribute value using a custom parse function or the framework default.
 * @ignore
 */
oj.BaseCustomElementBridge.__ParseAttrValue = function(elem, attr, prop, val, metadata) 
{
  if (val == null)
    return val;
  
  var type = metadata['type'];
  var bridge = oj.BaseCustomElementBridge.getInstance(elem);
  function _coerceVal(value) {
    var coercedValue;
    try 
    {
      coercedValue = oj.__AttributeUtils.coerceValue(elem, attr, value, type);
    }
    catch (ex)
    {
      bridge._throwError(elem, ex);
    }
    return coercedValue;
  }

  var parseFunction = oj.BaseCustomElementBridge.__GetDescriptor(elem.tagName)['parseFunction'];
  if (parseFunction)
  {
    return parseFunction(val, prop, metadata, 
      function(value) {
        return _coerceVal(value); 
      }
    );
  }
  return _coerceVal(val);
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__ProcessEventListeners = function(metadata, bConvertName) 
{
  metadata = oj.CollectionUtils.copyInto({}, metadata, undefined, true, 1);
  metadata['properties'] = metadata['properties'] || {};
  oj.BaseCustomElementBridge._enumerateMetadataForKey(null, metadata, 'properties',
    function(proto, property, propertyMeta)
    {
      var eventName = oj.__AttributeUtils.propertyNameToChangeEventType(property);
      var eventListenerProperty = oj.__AttributeUtils.eventTypeToEventListenerProperty(eventName);
      metadata['properties'][eventListenerProperty] = {_derived:true, _eventListener:true};
    }
  );
  oj.BaseCustomElementBridge._enumerateMetadataForKey(null, metadata, 'events',
    function(proto, event, eventMeta)
    {
      var eventName = bConvertName ? oj.__AttributeUtils.eventTriggerToEventType(event) : event;
      var eventListenerProperty = oj.__AttributeUtils.eventTypeToEventListenerProperty(eventName);
      metadata['properties'][eventListenerProperty] = {_derived:true, _eventListener:true};
    }
  );
  return metadata;
};

/**
 * @ignore
 * @param {string} tagName
 * @param {Object} descriptor
 * @param {Object} bridgeProto
 * @param {boolean=} isComposite
 */
oj.BaseCustomElementBridge.__Register = function(tagName, descriptor, bridgeProto, isComposite)
{
  var name = tagName.toLowerCase();
  if (!oj.BaseCustomElementBridge._registry[name])
  {
    if (!descriptor)
      throw new Error("Cannot register " + tagName + ". Missing a descriptor.");

    oj.BaseCustomElementBridge._verifyMetadata(tagName, descriptor[oj.BaseCustomElementBridge.DESC_KEY_META]);
    oj.BaseCustomElementBridge._registry[name] = {descriptor: descriptor, bridgeProto: bridgeProto, composite: isComposite, cache: {}};
    return true;
  }
  return false;
};

/**
 * @ignore
 * @param {Element} element
 * @param {string} name
 * @param {Object} value
 * @param {Object} previousValue
 * @param {string} updatedFrom
 * @param {Object=} subprop
 */
oj.BaseCustomElementBridge.__FirePropertyChangeEvent = function(element, name, value, previousValue, updatedFrom, subprop)
{
  // The bridge sets the ready to fire flag after the component has been instantiated.
  // We shouldn't fire property changed events before then unless the update comes from internally
  // for cases like readOnly property updates.
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  if (updatedFrom === 'external' && !bridge.__READY_TO_FIRE)
    return;

  var detail = {};
  if (subprop)
    detail['subproperty'] = subprop;
  detail['value'] = value;
  detail['previousValue'] = previousValue;
  detail['updatedFrom'] = updatedFrom;

  // Check if the subclass needs to do anything before we fire the property change event,
  // e.g. composites that need to call the propertyChanged ViewModel callback.
  if (bridge.beforePropertyChangedEvent)
    bridge.beforePropertyChangedEvent(element, name, detail);
  element.dispatchEvent(new CustomEvent(name + "Changed", {'detail': detail}));
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__DefineDynamicObjectProperty = function(obj, property, getter, setter)
{
  Object.defineProperty(obj, property, { 'enumerable': true, 'get': getter, 'set': setter });
};

/**
 * @ignore
 */
oj.BaseCustomElementBridge.__GetUnique = function()
{
  return _UNIQUE + _UNIQUE_INCR++;
};

var _UNIQUE_INCR = 0;
var _UNIQUE = '_ojcustomelem';

/**
 * @ignore
 */
oj.BaseCustomElementBridge._registry = {};

/** @ignore */
oj.BaseCustomElementBridge.DESC_KEY_CSS = 'css';
/** @ignore */
oj.BaseCustomElementBridge.DESC_KEY_META = 'metadata';
/** @ignore */
oj.BaseCustomElementBridge.DESC_KEY_PARSE_FUN = 'parseFunction';
/** @ignore */
oj.BaseCustomElementBridge.DESC_KEY_VIEW = 'view';
/** @ignore */
oj.BaseCustomElementBridge.DESC_KEY_VIEW_MODEL = 'viewModel';
/** @ignore */
oj.BaseCustomElementBridge._GLOBAL_PROPERTIES = ['accesskey', 'class', 'contenteditable', 'contextmenu',
'dir', 'draggable', 'hidden', 'id', 'is', 'lang', 'style', 'tabindex', 'title'];

/**
 * @ignore
 */
oj.BaseCustomElementBridge._INSTANCE_KEY = "_ojBridge";

/**
 * Class used to track the create state.
 * @constructor
 * @protected
 * @ignore
 */
oj.BaseCustomElementBridge.__DelayedPromise = function()
{
  var _promise;
  var _resolve;
  var _reject;

  /**
   * Returns the create Promise, creating one as needed.
   * @return {Promise}
   * @ignore
   */
  this.getPromise = function()
  {
    if (!_promise)
    {
      _promise = new Promise(function (resolve, reject) 
      {
        _resolve = resolve;
        _reject = reject;
      });
    }
    return _promise;
  };
  
  /**
   * Rejects the create Promise if one exists.
   * @ignore
   */
  this.rejectPromise = function(reason)
  {
    if (_reject) 
    {
      _reject(reason);
    }
  };

  /**
   * Resolves the create Promise if one exists.
   * @ignore
   */
  this.resolvePromise = function(value)
  {
    if (_resolve) 
    {
      _resolve(value);
    }
  };
}




/** 
 * @ojoverviewdoc CustomElementOverview - JET Custom Elements
 * @classdesc
 * {@ojinclude "name":"customElementOverviewDoc"}
 */ 

/**
 *       <h3 id="ce-usage-section" class="subsection-title">
 *               JET Custom Element Usage
 *           <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-usage-section"></a>
 *        </h3>
 *        <p>
 *            JET components are implemented as 
 *            <a href="https://developer.mozilla.org/en-US/docs/Web/Web_Components/Custom_Elements">custom HTML elements</a>
 *            and programmatic access to these components is similar to interacting with regular HTML elements. 
 *            All JET custom elements have names starting with 'oj-'. A custom element will be recognized by the framework 
 *            only after its module is loaded by the application. Once the element is recognized, the framework will register 
 *            a busy state for the element and will begin the process of 'upgrading' the element. The element will not be 
 *            ready for interaction (e.g. retrieving properties or calling methods) until the upgrade process is complete with 
 *            the exception of property setters and the <code>setProperty</code> and <code>setProperties</code> methods. The application should 
 *            listen to either the page-level or an element-scoped BusyContext before attempting to interact with 
 *            any JET custom elements. See the <a href="oj.BusyContext.html">BusyContext</a> documentation
 *            on how BusyContexts can be scoped.
 *        </p>
 *        
 *        <h4 id="ce-attributes-section" class="subsection-title">Attributes
 *            <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-attributes-section"></a>
 *        </h4>
 *        <h5 id="ce-attrs-basics-section" class="subsection-title">Basics
 *            <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-attrs-basics-section"></a>
 *        </h5>
 *        <p>
 *          The next two sections will discuss how to work with custom element attributes and properties. Attributes
 *          are declared in the custom element's HTML while properties are accessed using JavaScript on the element
 *          instance. Attribute values set as string literals will be parsed and coerced to the property type. JET currently
 *          only supports the following string literal type coercions: boolean, number, string, Object and Array, where 
 *          Object and Array types must use JSON notation with double quoted strings. A special 'any' type is also supported
 *          and is described <a href="#ce-attrs-any-section">below</a>. All other types need to be set using 
 *          <a href="#ce-databind-section">expression syntax</a> in the DOM or using the element's property setters or <code>setProperty</code>
 *          and <code>setProperties</code> methods. Early property sets before the composite has been upgraded
 *          will not result in [property]Changed events and will be passed to the component as part of its initial state.
 *        </p>
 *        <p> 
 *          Attribute removals are treated as unsetting of a property where the component default value will be used if one exists. 
 *        </p>
 *        <p>
 *          As described <a href="#ce-databind-section">below</a>, JET uses [[...]] and {{...}} syntax to represent data bound expressions.
 *          JET does not currently provide any escaping syntax for "[[" or "{{" appearing at the beginning of the attribute
 *          value. You will need to add a space character to avoid having the string literal value interpreted as a binding
 *          expression (e.g. &lt;oj-some-element some-attribute='[ ["arrayValue1", "arrayValue2", ... ] ]'>&lt;/oj-some-element>).
 *        </p>
 *        <h5 id="ce-attrs-boolean-section" class="subsection-title">Boolean Attributes
 *            <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-attrs-boolean-section"></a>
 *        </h5>
 *        <p>
 *          JET custom elements treat boolean attributes differently than HTML5. Since a common application use case
 *          is to toggle a data bound boolean attribute, JET will coerce the string literal "false" to the boolean
 *          false for boolean attributes. The absence of a boolean attribute in the DOM will also be interpreted as false.
 *          JET will coerce the following string literal values to the boolean true and throw an Error for all other 
 *          invalid values.
 *          <ul>
 *            <li>No value assignment (e.g. &lt;oj-some-element boolean-attribute>&lt;/oj-some-element>)</li>
 *            <li>Empty string (e.g. &lt;oj-some-element boolean-attribute=''>&lt;/oj-some-element>)</li>
 *            <li>The 'true' string literal (e.g. &lt;oj-some-element boolean-attribute='true'>&lt;/oj-some-element>)</li>
 *            <li>The case-insensitive attribute name (e.g. &lt;oj-some-element boolean-attribute='boolean-attribute'>&lt;/oj-some-element>)</li>
 *          </ul>
 *        </p>
 *        <h5 id="ce-attrs-object-section" class="subsection-title">Object-Typed Attributes
 *            <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-attrs-object-section"></a>
 *        </h5>
 *        <p>
 *          Attributes that support Object type can be declaratively set using dot notation. 
 *          Note that applications should not set overlapping attributes as these will cause an error to be thrown.
 *          <pre class="prettyprint">
 *            <code>
 *              &lt;!-- person is an Object typed attribute with a firstName subproperty -->
 *              &lt;oj-some-element person.first-name="{{name}}">&lt;/oj-some-element>
 *              
 *              &lt;!-- overlapping attributes will throw an error -->
 *              &lt;oj-some-element person="{{personInfo}}" person.first-name="{{name}}">&lt;/oj-some-element>
 *            </code>
 *          </pre>
 *          If applications need to programmatically set subproperties, they can call  the JET custom element's <code>setProperty</code> 
 *          method with dot notation using the camelCased property syntax (e.g. element.setProperty("person.firstName", "Sally")).
 *        </p>
 *        <h5 id="ce-attrs-any-section" class="subsection-title">Any-Typed Attributes
 *            <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-attrs-any-section"></a>
 *        </h5>
 *        <p>
 *          Attributes that support any type are documented with type {*} in the API doc and will be coerced as
 *          Objects, Arrays, or strings. This limitation is due to the fact that we cannot determine whether value='2' 
 *          for a property supporting any type should be parsed as a string or a number. The application should use 
 *          data binding for all other value types.
 *        </p>
 *        <h4 id="ce-properties-section" class="subsection-title">Properties
 *            <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-properties-section"></a>
 *        </h4>
 *        <p>
 *            Every attribute listed in the component API doc will have a corresponding property
 *            on the HTML element and can be accessed once the component is fully upgraded. 
 *            See the <a href="#ce-proptoattr-section">property to attribute mapping</a>
 *            section below to see the syntax difference between setting attributes and properties. 
 *            Note that some properties are read only and updates to those properties using the property 
 *            setter will be ignored. The undefined value is treated as unsetting of a property when passed to the
 *            property setter and will result in the component using the default value if one exists. Unsetting of
 *            subproperties using the element's <code>setProperty</code> is not supported. Subproperties can only
 *            only be unset when the top level property is unset. 
 *            Property sets will not result in DOM attribute updates and after the custom 
 *            element is upgraded, the application should use the custom element properties, not attributes to check 
 *            the current value.
 *        </p>
 *        <p>
 *            When a property or attribute value changes, a [property]Changed <code>CustomEvent</code> 
 *            will be fired with the following properties in the event's detail property.
 *            <table class="props">
 *                <thead>
 *                    <tr>
 *                        <th>Name</th>
 *                        <th>Type</th>
 *                        <th>Description</th>
 *                    </tr>
 *                </thead>
 *                <tbody>
 *                    <tr>
 *                        <td>value</td>
 *                        <td>*</td>
 *                        <td>The current value of the property that changed.</td>
 *                    </tr>
 *                    <tr>
 *                        <td>previousValue</td>
 *                        <td>*</td>
 *                        <td>The previous value of the property that changed.</td>
 *                    </tr>
 *                    <tr>
 *                        <td>updatedFrom</td>
 *                        <td>string</td>
 *                        <td>
 *                          Where the property was updated from. Supported values are:
 *                          <ul>
 *                            <li>external - By the application, using either the element's property setter, setAttribute, or external data binding.</li>
 *                            <li>internal - By the component, e.g. after user interaction with a text field or selection.</li>
 *                          </ul>  
 *                        </td>
 *                    </tr>
 *                    <tr>
 *                        <td>subproperty</td>
 *                        <td>Object</td>
 *                        <td>An object holding information about the subproperty that changed.
 *                            <table class="props">
 *                                <thead>
 *                                    <tr>
 *                                        <th>Name</th>
 *                                        <th>Type</th>
 *                                        <th>Description</th>
 *                                    </tr>
 *                                </thead>
 *                                <tbody>
 *                                    <tr>
 *                                        <td>path</td>
 *                                        <td>string</td>
 *                                        <td>
 *                                          The subproperty path that changed, starting from the top level 
 *                                          property with subproperties delimited by '.'.
 *                                        </td>
 *                                    </tr>
 *                                    <tr>
 *                                        <td>value</td>
 *                                        <td>*</td>
 *                                        <td>The current value of the subproperty that changed.</td>
 *                                    </tr>
 *                                    <tr>
 *                                        <td>previousValue</td>
 *                                        <td>*</td>
 *                                        <td>The previous value of the subproperty that changed.</td>
 *                                    </tr>
 *                                </tbody>
 *                            </table>
 *                        </td>
 *                    </tr>
 *                </tbody>
 *            </table>
 *        </p>
 *        <p>
 *          Please note that in order for components to be notified of a property change for Array properties, the
 *          value should be data bound and updated using an expression, setting the property to an updated copy 
 *          by calling slice(), or by refreshing the component after an in place Array mutation.
 *        </p>
 *        <p>
 *          The application can listen to these [property]Changed events by adding a listener either declaratively:
 *          <pre class="prettyprint">
 *            <code>
 *              &lt;oj-some-element value="{{currentValue}}" on-value-changed="{{valueChangedListener}}">&lt;/oj-some-element>
 *            </code>
 *          </pre>
 *          or programmatically using the element property or the DOM addEventListener :
 *          <pre class="prettyprint">
 *            <code>
 *              someElement.addEventListener("valueChanged", function(event) {...});
 *            </code>
 *            <code>
 *              someElement.onValueChanged = function(event) {...};
 *            </code>
 *          </pre>
 *        </p>
 *        <p>
 *          Some JET components support complex properties. If the application needs to set a single subproperty
 *          instead of the entire complex property, the <code>setProperty</code> method should be used 
 *          instead to ensure that [property]Changed events will be fired correctly. Note that directly updating
 *          the subproperty via dot notation (e.g. element.topProp.subProp = newValue) will not result in a 
 *          [property]Changed event being fired.
 *        </p>
 *        <h4 id="ce-proptoattr-section" class="subsection-title">Property-to-Attribute Mapping
 *          <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-proptoattr-section"></a>
 *        </h4>
 *        <p>
 *          The following rules apply when mapping property to attribute names:
 *          <ul>
 *             <li>Attribute names are case insensitive. CamelCased properties are mapped to 
 *               kebab-cased attribute names by inserting a dash before the uppercase letter and converting that letter to lower case
 *               (e.g. a "chartType" property will be mapped to a "chart-type" attribute).</li>
 *             <li> The reverse occurs when mapping a property name from an attribute name.</li>
 *          </ul>
 *        </p>        
 *        <h4 id="ce-databind-section" class="subsection-title">Data Binding and Expression Writeback
 *          <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-databind-section"></a>
 *        </h4>
 *          The upgrade of JET custom elements relies on any data binding resolving, the management of which 
 *          is done by a binding provider. The binding provider is responsible for setting and updating attribute
 *          expressions and any custom elements within its managed subtree will not finish upgrading until it
 *          applies bindings on that subtree. By default, there is a single binding provider for a page, 
 *          but subtree specific binding providers can be added by using the <code>data-oj-binding-provider</code> 
 *          attribute with values of "none" and "knockout". The default binding provider is knockout, but if a 
 *          page or DOM subtree does not use any expression syntax or knockout, the application can set 
 *          <code>data-oj-binding-provider="none"</code> on that element so its dependent JET custom elements
 *          do not need to wait for bindings to be applied to finish upgrading. <b>When using the knockout binding
 *          provider, applications should require the ojknockout module.</b>
 *        <p>
 *          When a JET custom element is managed by a binding provider, its attributes may be set to a binding
 *          expression. Component attributes can be set directly using one-way [[...]] or two-way {{...}} data binding syntax
 *          whereas global attributes and attributes on any native HTML element may be one-way bound by prefixing the attribute
 *          name with ':' (e.g. :id='[[idVar]]'). The attribute binding syntax results in the attribute being set in the DOM
 *          with the evaluated expression. Since global HTML attributes are always string typed, expressions using the ':' 
 *          prefixing should resolve to strings with the exception of the style and class attributes which support additional types. 
 *          In the case of component attributes, applications are recommended to bind the attribute names directly and avoid the 
 *          use of the ":" prefix.
 *        </p>
 *        <p>
 *          The class attribute binding supports a space delimited string of classes, an Array of classes, or an Object to toggle
 *          classes (e.g. :class='[[{errorClass: hasErrors}]]' where the hasErrors expression determines whether the errorClass class
 *          is present in the DOM). Note that the Array and string types will override existing values in the class attribute 
 *          when updates occur, whereas the Object type will only add and remove the classes specified. Since JET custom elements 
 *          add their own classes, we recommend using the Object type when using the class attribute binding on JET custom elements.
 *        </p>
 *        <p>
 *          When using the style attribute binding with an Object type, the style Object names should be the JavaScript 
 *          names for that style (e.g. fontWeight instead of font-weight). Since the style attribute supports Object 
 *          types, it also supports dot notation for setting style subproperties directly (e.g. :style.font-weight='[[...]]').
 *        </p>
 *        <p>
 *          Applications can control expression writeback in the 
 *          custom element by using {{...}} syntax for two-way writable binding expressions or [[...]] for 
 *          one-way only expressions. Most component properties do not writeback and those that do will indicate
 *          it in their API doc. Please note that there should be no spaces between the braces when using the data 
 *          bind syntax (e.g. some-attribute='[ [...] ]').
 *        </p>
 *        <p>
 *          Certain properties such as 'value' on editable components support updating the associated expression 
 *          automatically whenever their value changes. This functionality is also known as 'writeback'. 
 *          Applications can control expression writeback by using  the {{...}} syntax for two-way writable 
 *          binding expressions or [[...]] for one-way only expressions. The one-way expressions should 
 *          be used when the application needs expressions strictly for 'downstream-only' purposes, e.g. only for 
 *          updating a component property. This is different from the read-only properties, which are 'upstream-only', 
 *          e.g. they are used only to monitor component state. Thus an expression associated with a read-only property 
 *          should always use the {{}} syntax. Most component properties do not writeback and those that do will 
 *          indicate it in their API doc. Please note that there should be no spaces between the braces when using 
 *          the data bind syntax (e.g. some-attribute='[ [...] ]').
 *        </p>
 *        <pre class="prettyprint">
 *          <code>
 *            &lt;oj-some-element value="[[currentValue]]" selection={{currentSelection}}>&lt;/oj-some-element>
 *          </code>
 *        </pre>        
 *        <h4 id="ce-methods-section" class="subsection-title">Methods
 *          <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-methods-section"></a>
 *        </h4>
 *        <p>
 *          Methods can be accessed on the JET custom HTML element after the element is fully upgraded. See
 *          the specific method doc for sample usages.
 *        </p>        
 *        <h4 id="ce-events-section" class="subsection-title">Events and Listeners
 *          <a class="bookmarkable-link" title="Bookmarkable Link" href="#ce-events-section"></a>
 *        </h4>
 *        <p>
 *          By default, JET components will fire [property]Changed (e.g. valueChanged) <code>CustomEvents</code>
 *          whenever a property is updated. These events, unlike other component events, are non bubbling.
 *          See the <a href="#ce-properties-section">properties section</a> above
 *          for details on the event payload.
 *        </p>
 *        <p>
 *          In addition to supporting event listeners via the standard <code>addEventListener</code>
 *          mechanism, both JET custom elements and native HTML elements support declarative specification of event
 *          listeners via <code>on-[event-name]</code> attributes (e.g. <code>on-click</code>,
 *          <code>on-value-changed</code> or <code>on-oj-expand</code>). The attributes ultimately delegate to the standard
 *          <code>addEventListener</code> mechanism and only support data bound expressions
 *          that evaluate to functions; arbitrary JavaScript will not be accepted.
 *          Sample usages can be seen in the attribute API doc.
 *        </p>
 *        <p>
 *          In addition to the event parameter, event listeners specified via <code>on-[event-name]</code>
 *          attributes will receive two additional parameters when they are invoked: <code>data</code> and <code>bindingContext</code>.
 *          The <code>bindingContext</code> parameter provides the listener with the entire data binding context that
 *          was applied to the element while the data parameter provides convenient access to relevant data.
 *          When in an iteration context (e.g. inside an <code>oj-bind-for-each</code>), the <code>data</code> parameter
 *          is equal to <code>bindingContext['$current']</code>; otherwise, it is equal to <code>bindingContext['$data']</code>.
 *          These event listeners should be written with signatures of <code>function(event, data, bindingContext)</code>.
 *        </p>
 *        <p>
 *          Please note that there is a
 *          current limitation where event listeners specified using this syntax can only be
 *          set during component initialization. Subsequent setAttribute calls for the
 *          event listener attributes will be ignored.
 *        </p>
 *        <p>
 *          Component events that are specifically documented in a JET custom element's API
 *          doc (including [property]Changed events) can also have listeners set via direct
 *          assignment of an on[EventName] property on the element (e.g. onClick or
 *          onOjExpand).  Note however that when these events bubble up through the DOM,
 *          they can only be listened to via addEventListener calls or on-[event-name]
 *          attributes on parent elements.  Event listeners that are specified via addEventListener
 *          or direct property assignment will not receive the additional parameters described above;
 *          they will be invoked with the single event parameter only.
 *        </p>
 * @ojfragment customElementOverviewDoc - General description doc fragment that shows up in every component's page via a link.
 * @memberof CustomElementOverview
 */
/*
** Copyright (c) 2008, 2013, Oracle and/or its affiliates. All rights reserved.
**
**34567890123456789012345678901234567890123456789012345678901234567890123456789
*/


/**
 * @class oj.Components
 * @classdesc JET Component services
 * @since 1.0
 * @export
 */
oj.Components = {};


/**
 * Sets default options values for JET components.
 * @param {!Object} options - property values that will be merged into the values
 * that were previously set using this method. The options object is expected to have the format demonstrated
 * by the following example:
 * <pre>
 * {
 *   'default': // properties for all JET components
 *   {
 *     'option1': 'somevalue'
 *   },
 *   'editableValue': // properties for editableValue components
 *   {
 *     'option1': 'somevalue1',
 *     'option2': oj.Components.createDynamicPropertyGetter(function(context){
 *                                 return context['containers'].indexOf('ojTable') >= 0 ? 'tableValue' : 'normalValue'})
 *   },
 *   'ojText': // properties for instances of ojText
 *   {
 *     'option1': 'somevalue2'
 *   }
 * }
 * </pre>
 * To specify a dynamic getter for the property, pass your callback to oj.Components.createDynamicPropertyGetter(). Note
 * that dynamic getters nested within a complex property value are not supported
 * @see oj.Components.createDynamicPropertyGetter
 * @return {void}
 * @export
 */
oj.Components.setDefaultOptions = function(options)
{
  var props = oj.Components._defaultProperties || {};

  var keys = Object.keys(options);

  keys.forEach(function(key)
    {
      var value = options[key];
      if (!oj.CollectionUtils.isPlainObject(value))
      {
        throw ("Invalid default options");
      }
      props[key] = _accumulateValues(props[key]||{}, value, false);
    }
  );

  oj.Components._defaultProperties = props;
};

/**
 * Retrieves default option values for JET components. This method should only be used internally by JET.
 * @deprecated since version 2.2
 * @ignore
 * @return {Object} default option values
 * @see oj.Components.setDefaultOptions
 * @export
 */
oj.Components.getDefaultOptions = function()
{
  return (oj.Components._defaultProperties || {});
};


/**
 * Creates a dynamic getter that can be used as a property value in oj.Components.setDefaultOptions()
 * @param {!Function} callback - dynamic property callback. The callback will receive a context object as a parameter.
 * The following properties are currently supported on the context object:
 * <ul>
 * <li>containers - an array of component names of the current component's containers that require special behavior from
 * their children</li>
 * <li>element - component's host DOM element</li>
 * </ul>
 * The callback should return the computed property value
 *
 * @return {Object} - dynamic property getter
 * @see oj.Components.setDefaultOptions
 * @export
 */
oj.Components.createDynamicPropertyGetter = function(callback)
{
  return new __ojDynamicGetter(callback);
};

/**
 * This method should only be used for JQueryUI components and will return null if used
 * with a custom element. Retrieves widget constructor associated with the HTML element
 * or null if none is found. The returned constructor is already bound to the associated
 * JQuery element, so it can be invoked as a function directly. For example:
 * <pre>
 * widgetConstructor("option", "label", "custom"); // sets label option
 * </pre>
 * If widgetName is not specified, and if more than one widget is associated with the element,
 * the method will a return the widget that was created first.
 * @param {?(Element|Node)} element - HTML element
 * @param {string=} widgetName - optional widget name
 * @return {Function|null} widget constructor
 * @export
 */
oj.Components.getWidgetConstructor = function(element, widgetName)
{
  if (element && !oj.BaseCustomElementBridge.getRegistered(element.tagName))
    return oj.Components.__GetWidgetConstructor(element, widgetName);
  return null;
};

/**
 * Internal version for components to call which won't return null for 
 * custom elements. See public method for jsDoc.
 * @param {?(Element|Node)} element - HTML element
 * @param {string=} widgetName - optional widget name
 * @return {Function|null} widget constructor
 * @ignore
 */
oj.Components.__GetWidgetConstructor = function(element, widgetName)
{
  var jelem = $(element);

  var data = jelem.data(_OJ_WIDGET_NAMES_DATA);
  if (data)
  {
    if (widgetName == null)
    {
      widgetName = data[0];
    }
    else if (data.indexOf(widgetName) < 0)
    {
      widgetName = undefined;
    }

    if (widgetName != null)
    {
      var func = jelem[widgetName];
      if ((typeof func) === "function")
      {
        return func.bind(jelem);
      }
    }
  }
  return null;
};

/**
 * Notifies JET framework that a subtree possibly containing JET components has been inserted
 * into the document programmatically.
 *
 * Note that there is no need to call this method when the new DOM is being inserted by the template engine
 * in Knockout.js
 * @param {!Element} node - the root of the subtree
 * @see oj.Components.subtreeDetached
 * @return {void}
 * @export
 */
oj.Components.subtreeAttached = function(node)
{
  oj.DomUtils.fixResizeListeners(node);
  _applyToComponents(node,
    function(instance)
    {
      instance.__handleSubtreeAttached();
    }
  );
};

/**
 * Notifies JET framework that a subtree possibly containing JET components has been removed
 * from the document programmatically.
 *
 * Note that calling this method is not needs after calling JQuery's .remove() because all JET components would have been
 * already destroyed in that case. Similarly, there is no need to call this method after the subtree has been removed by
 * Knockout.js
 * @param {!Element} node - the root of the subtree
 * @see oj.Components.subtreeAttached
 * @return {void}
 * @export
 */
oj.Components.subtreeDetached = function(node)
{
  _applyToComponents(node,
    function(instance)
    {
      instance.__handleSubtreeDetached();
    }
  );
};

/**
 * Notifies JET framework that a subtree possibly containing JET components is no longer hidden with display:none style
 * This method should be called by the application if the 'display' style is being changed from 'hidden' programmatically,
 * such as when JQuery's .show() method is called. 
 * For cases where subtree is shown on initial render, this method should be called with the options parameter set to
 * {'initialRender':true}, that will result in _NotifyInitShown() calls to the subtree components.
 * All oj-defer elements in the entire subtree will be activated. Note that subtreeShown currently notifies the entire
 * subtree as well. This generally means that only non-nested oj-defer elements make sense in a subtree.
 *
 * @param {!Element} node - the root of the subtree
 * @param {Object=} options Options to control subtreeShown
 * @param {boolean} options.initialRender The index at which to start fetching records.
 * @see oj.Components.subtreeHidden
 * @return {void}
 * @export
 */
oj.Components.subtreeShown = function(node, options)
{
  options = options || {};
  var isInitialRender = options['initialRender'];
  if (!isInitialRender)
  {
    oj.DomUtils.fixResizeListeners(node);
  }

  node = $(node)[0]; // Strip possible jQuery wrapper

  unmarkSubtreeHidden(node);

  _applyHideShowToComponents(node,
    function(instance)
    {
      if (isInitialRender)
      {
        instance._NotifyInitShown();
      }
      else
      {
        instance._NotifyShown();
      }
    }, true);
};

/**
 * Notifies JET framework that a subtree possibly containing JET components has been hidden  with display:none style
 * This method should be called by the application after the subtree has been hidden programmatically, such as
 * when JQuery's .hide() method is called.
 *
 * @param {!Element} node - the root of the subtree
 * @see oj.Components.subtreeShown
 * @return {void}
 * @export
 */
oj.Components.subtreeHidden = function(node)
{
  node = $(node)[0]; // Strip possible jQuery wrapper

  _applyHideShowToComponents(node,
    function(instance)
    {
      instance._NotifyHidden();
    }, false);

  markSubtreeHidden(node);
};

/**
 * Add a marker class indicating that this subtree is hidden.
 *
 * @ignore
 */
function markSubtreeHidden(element)
{
  element.classList.add(_OJ_SUBTREE_HIDDEN_CLASS);
}

/**
 * Remove the marker class indicating that this subtree is hidden.
 *
 * @ignore
 */
function unmarkSubtreeHidden(element)
{
  element.classList.remove(_OJ_SUBTREE_HIDDEN_CLASS);
}

/**
 * Called by CCAs and certain custom elements when they are first connected
 * to indicate that this component is initializing and will control
 * whether its child subtrees are hidden.
 *
 * @ignore
 */
oj.Components.markPendingSubtreeHidden = function(element)
{
  element.classList.add(_OJ_PENDING_SUBTREE_HIDDEN_CLASS);
}


/**
 * Called by CCAs and certain custom elements right before they are first rendered.
 * This component will control whether its child subtrees are hidden.
 *
 * @ignore
 */
oj.Components.unmarkPendingSubtreeHidden = function(element)
{
  element.classList.remove(_OJ_PENDING_SUBTREE_HIDDEN_CLASS);
}

/**
 * Determines if a component identified by the <code>widgetName</code> has been
 * bound and initialized on a given <code>jelement</code>.
 *
 * @param {jQuery} jelement to which the component is bound
 * @param {string} widgetName constructor name of the target component.
 * @return {boolean} <code>true</code> if the component identified by the widgetName
 *  has be bound and initialized to the target element.
 */
oj.Components.isComponentInitialized = function(jelement, widgetName)
{
  /** @type {?} */
  var widgets = jelement.data(_OJ_WIDGET_NAMES_DATA);
  if ($.isArray(widgets) && widgets.indexOf(widgetName) > -1 && jelement.is('.' + _OJ_COMPONENT_NODE_CLASS))
    return true;
  else
    return false;
};

/**
 * @ignore
 */
oj.Components.__getDefaultOptions = function(hierarchyNames)
{
  var defaults = {};
  var allProperties =  oj.Components.getDefaultOptions();

  for (var i=hierarchyNames.length-1; i>=0; i--)
  {
    var name = hierarchyNames[i];
    var props = allProperties[name];
    if (props !== undefined)
    {
      defaults = _accumulateValues(defaults, props, true);
    }
  }

  return defaults;
};

/**
 * Retrieves the JET component element that
 * the node is in.
 * @param {?(Element|Node)} node - DOM node
 * @return {?(Element|Node)} componentElement - JET component element
 * A component element is the DOM element on which the JET component is
 * initialized.
 * @export
*/
oj.Components.getComponentElementByNode = function(node)
{ 
  //Temporarily exposing this private flag in order to allow
  //MonkeyTalk to access JET components that are part of the 
  //composite's implementation. We are not exposing this flag
  //as a public API as a) accessing composite implementation
  //components is bad and b) the need for this method will soon
  //go away. Once our new automation API (recording adapters) are
  //in place, even MonkeyTalk will no longer need this flag. Adding
  //this as a stop-gap measure to allow MonkeyTalk to carry on in the
  //meantime. Callers other than MonkeyTalk must avoid specifying this flag. 
  var mtAccessCompositeInternals = arguments.length > 1 && arguments[1] ? true : false;
  return _getComponentElementByNode(node, mtAccessCompositeInternals);
};

/**
 * Private method implementing the functionality of
 * getComponentElementByNode. This was done because Closure
 * throws an error when the private, undocumented flag (mtAccessCompositeInternals)
 * is passed to recursive calls of getComponentElementByNode.
 * @private
*/ 
function _getComponentElementByNode(node, mtAccessCompositeInternals){
  if (node == null)
  {
    return null;
  }
  // node can be a Node or Element but we call some Element only APIs
  // so we need to do an additional isElement check first
  var isElement = node.nodeType === 1;

  var containingComposite = (oj.Composite && !mtAccessCompositeInternals ? oj.Composite.getContainingComposite(node) : null);
  if (containingComposite)
  { // node is in or is a composite, return composite
    return containingComposite;
  } 
  else if (isElement && node.hasAttribute('data-oj-internal'))
  {// node is an internal component
    if (node.parentNode instanceof Element && node.parentNode.hasAttribute('data-oj-surrogate-id'))
    { // internal component is a popup
      node = document.querySelector('[data-oj-popup-' + node.id + '-parent]'); // retrieves popups parent element
      return _getComponentElementByNode(node, mtAccessCompositeInternals);
    }
    return _getComponentElementByNode(node.parentNode, mtAccessCompositeInternals);
  } 
  else if (_isComponentElement(node))
  { // node is a component element
    return node;
  } 
  else if (isElement && node.classList.contains('oj-component'))
  { // node is component wrapper
    node = node.querySelector('.oj-component-initnode:not([data-oj-internal])') || node;
    if (_isJQueryUI(node))
    {
      return node;
    }
  } 
  else if (isElement && node.hasAttribute('data-oj-containerid'))
  { // node is non-internal component popup e.g listbox
      node = document.getElementById(node.getAttribute('data-oj-containerid'));
      return _getComponentElementByNode(node, mtAccessCompositeInternals);
  }
  return _getComponentElementByNode(node.parentNode, mtAccessCompositeInternals);
}

/**
 * Retrieves the subId of the node as
 * as part of a locator object i.e. at least
 * {subId: subIdOfNode}
 * @param {?Element} componentElement - JET component element
 * @param {?Element} node - DOM node
 * @return {*} locator - object with at least a subId
 * or null if the node does not have a subId
 * @export
*/
oj.Components.getSubIdByNode = function(componentElement, node)
{
  return oj.Components.callComponentMethod(componentElement, 'getSubIdByNode', node);
};

/**
 * Returns the component DOM node indicated
 * by the locator parameter.
 * @param {?Element} componentElement - JET component element
 * @param {Object} locator - Object containing, at minimum,
 * a subId property, whose value is a string that identifies
 * a particular DOM node in this component.
 * @return {*} node - The DOM node located by
 * the locator, or null if none is found
 * @export
*/
oj.Components.getNodeBySubId = function(componentElement, locator)
{
  return oj.Components.callComponentMethod(componentElement, 'getNodeBySubId', locator);
};

/**
 * Retrieves the specified option of
 * the specified JET component element
 * @param {?Element} componentElement - JET component element
 * @param {string} option - option to retrieve
 * @return {*} value of option
 * @export
*/
oj.Components.getComponentOption = function(componentElement, option)
{ 
  if (!_isComponentElement(componentElement))
  {
    throw new Error('node is not a component element');
  } else if (_isCompositeOrCustom(componentElement))
  {
    if (componentElement['getProperty'])
    {
      return componentElement['getProperty'].call(componentElement, option);
    }
  } else 
  {
    return oj.Components.__GetWidgetConstructor(componentElement)('option', option);
  }
};

/**
 * Sets the specified option of the specified
 * JET component element to the specified value
 * @param {?Element} componentElement - JET component element
 * @param {string} option - option to set
 * @param {*} value - value to set option to
 * @return {void}
 * @export
*/
oj.Components.setComponentOption = function(componentElement, option, value)
{ 
  if (!_isComponentElement(componentElement))
  {
    throw new Error('node is not a component element');
  } else if (_isCompositeOrCustom(componentElement))
  {
    if (componentElement['setProperty'])
    {
      componentElement['setProperty'].call(componentElement, option, value);
    }
  } else 
  {
    oj.Components.__GetWidgetConstructor(componentElement)('option', option, value);
  }
};

/**
 * Calls the specified JET component element's method
 * with the given arguments
 * @param {?Element} componentElement - JET component element
 * @param {string} method - name of JET component element method to call
 * @param {...*} methodArguments - list of arguments to pass to method call
 * @return {*}
 * @export
*/
oj.Components.callComponentMethod = function(componentElement, method, methodArguments)
{
  if (!_isComponentElement(componentElement))
  {
    throw new Error('node is not a component element');
  } else if (_isCompositeOrCustom(componentElement))
  {
    if (componentElement[method])
    {
      return componentElement[method].apply(componentElement, [].slice.call(arguments, 2));
    }
  } else 
  {
    return oj.Components.__GetWidgetConstructor(componentElement).apply($(componentElement), [].slice.call(arguments, 1));
  }
};

/**
 * @private
 */
function _applyToComponents(subtreeRoot, jqCallback)
{
  var processFunc = function()
  {
    var jelem =  $(this);
    var names = jelem.data(_OJ_WIDGET_NAMES_DATA);
    if (names != null)
    {
      for (var i=0; i < names.length; i++)
      {
        var instance =  jelem.data("oj-" + names[i]);
        if (instance != null)
        {
          jqCallback(instance);
        }
      }
    }
  };

  var locator = $(subtreeRoot);

  // Include the root node itself, and not just children ()
  if (locator.hasClass(_OJ_COMPONENT_NODE_CLASS))
  {
    processFunc.call(subtreeRoot);
  }

  locator.find('.' + _OJ_COMPONENT_NODE_CLASS).each(processFunc);
}

/**
 * @private
 */
function _applyHideShowToComponents(subtreeRoot, jqCallback, activateDefer)
{
  // Detect hidden without forcing a layout.
  function isHidden(node)
  {
    while (node)
    {
      if (node.nodeType === Node.DOCUMENT_NODE)
      {
        return false; // Walked up to document.  Not hidden
      }
      if (node.classList.contains(_OJ_SUBTREE_HIDDEN_CLASS))
      {
        return true;
      }
      node = node.parentNode;
    }
    return true;  // Didn't find document, so it must be detached and therefore hidden.
  }

  /**
   * Both node lists must be in document order.
   * Return new array containing nodes in 'allNodes' that are not in 'hiddenNodes'
   * @private
   */
  function filterHidden(allNodes, hiddenNodes)
  {
    var shownNodes = [];
    var i, j = 0;
    for (i = 0; i < hiddenNodes.length; i++)
    {
      var hidden = hiddenNodes[i];
      while (j < allNodes.length && allNodes[j] != hidden)
      {
        shownNodes.push(allNodes[j++])
      }
      j++;
    }
    while (j < allNodes.length)
    {
      shownNodes.push(allNodes[j++])
    }
    return shownNodes;
  }

  function processFunc(node)
  {
    if (jqCallback && node.classList.contains(_OJ_COMPONENT_NODE_CLASS))
    {
      var jelem =  $(node);
      var names = jelem.data(_OJ_WIDGET_NAMES_DATA);
      if (names != null)
      {
        for (var i=0; i < names.length; i++)
        {
          var instance =  jelem.data("oj-" + names[i]);
          if (instance != null)
          {
            jqCallback(instance);
          }
        }
      }
    }

    if (activateDefer && node.tagName.toLowerCase() === 'oj-defer')
    {
      node['_activate']();
    }
  };

  if (!isHidden(subtreeRoot))
  {
    processFunc(subtreeRoot);

    // Create selectors for jquery components and oj-defer as needed.
    var selectors = ['.' + _OJ_COMPONENT_NODE_CLASS];

    if (activateDefer)
    {
      selectors.push('oj-defer');
    }

    var hiddenSelectors = [];
    selectors.forEach(function(s) {
      hiddenSelectors.push('.' + _OJ_SUBTREE_HIDDEN_CLASS + ' ' + s);
      hiddenSelectors.push('.' + _OJ_PENDING_SUBTREE_HIDDEN_CLASS + ' ' + s);
    });

    // Create assemble a selector that gets all matches and the subset that are hidden
    var selector = selectors.join(',');
    var hiddenSelector = hiddenSelectors.join(',');
    
    // Fetch all matching elements and those that are hidden.
    // Use the second list to filter out hidden elements.
    var allNodes = subtreeRoot.querySelectorAll(selector);
    var hiddenNodes = subtreeRoot.querySelectorAll(hiddenSelector);
    var shownNodes = filterHidden(allNodes, hiddenNodes);
    
    for (var i = 0; i < shownNodes.length; i++)
    {
      processFunc(shownNodes[i]);
    }
  }
}

/**
 * @constructor
 * @param {!Function} callback
 * @private
 */
function __ojDynamicGetter(callback)
{
  this.getCallback = function()
  {
    return callback;
  }
};

/**
 * @ignore
 */
function _accumulateValues(target, source, valueInArray)
{
  var keys = Object.keys(source);

  keys.forEach(function(key)
    {
      var holder = target[key] || [];
      var sourceVal = source[key];
      if (valueInArray)
      {
        holder = holder.concat(sourceVal);
      }
      else
      {
        holder.push(sourceVal);
      }
      target[key] = holder;
    }
  );
  return target;
}

/**
 * @ignore
 */
 function _isCompositeOrCustom(node)
 {
  return oj.BaseCustomElementBridge.getRegistered(node.tagName);
 }

/**
 * @ignore
 */
 function _isJQueryUI(node)
 {
  return oj.Components.__GetWidgetConstructor(node) ? true: false;
 }


/**
 * @ignore
 */
 function _isComponentElement(node)
 {
  return (_isCompositeOrCustom(node) || _isJQueryUI(node));
 }

/**
 * @private
 */
oj.Components._OJ_CONTAINER_ATTR = "data-oj-container";

/**
 * @private
 */
var _OJ_WIDGET_NAMES_DATA = "oj-component-names";

/**
 * Marks the element which is a jQueryUI component
 * @private
 */
var _OJ_COMPONENT_NODE_CLASS = "oj-component-initnode";

/**
 * Marks an element as being hidden.
 *
 * @private
 */
var _OJ_SUBTREE_HIDDEN_CLASS = "oj-subtree-hidden";

/**
 * Marks an element as a container that will control hidden of its children
 * once it finishes initializing
 *
 * @private
 */
var _OJ_PENDING_SUBTREE_HIDDEN_CLASS = "oj-pending-subtree-hidden";

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/

(function() // BaseComponent wrapper function, to keep "private static members" private
{

/**
 * @private
 */
var _BASE_COMPONENT = 'baseComponent';

/**
 * @ojcomponent oj.baseComponent
 * @abstract
 * @since 0.6
 */
$.widget('oj.' + _BASE_COMPONENT,
{
  options:
  {
    /**
     * <p>There is no restriction on the order in which the JET Menu and the referencing component are initialized.  However, when specifying 
     * the Menu via the HTML attribute, the referenced DOM element must be in the document at the time that the referencing component is 
     * initialized.
     *
     * @ojfragment contextMenuInitOrderDoc - Decomped to fragment so Tabs, Tree, and MasonryLayout can override the fragment to convey their init order restrictions.
     * @memberof oj.baseComponent
     */
    /**
     * <p>To help determine whether it's appropriate to cancel the launch or customize the menu, the <code class="prettyprint">beforeOpen</code> 
     * listener can use component API's to determine which table cell, chart item, etc., is the target of the context menu. See the JSDoc and 
     * demos of the individual components for details.  Keep in mind that any such logic must work whether the context menu was launched via right-click, 
     * <kbd>Shift-F10</kbd>, <kbd>Press & Hold</kbd>, or component-specific touch gesture.
     *
     * @ojfragment contextMenuTargetDoc - Decomped to fragment so components can override the fragment to convey their specific API's for this.
     * @memberof oj.baseComponent
     */
    /**
     * <p>Identifies the [JET Menu]{@link oj.ojMenu} that the component should launch as a context menu on right-click, <kbd>Shift-F10</kbd>, <kbd>Press & Hold</kbd>,
     * or component-specific gesture. If specified, the browser's native context menu will be replaced by the specified JET Menu.
     *
     * <p>The value can be an HTML element, JQ selector, JQ object, NodeList, or array of elements.  In all cases, the first indicated element is used.
     *
     * <p>To specify a JET context menu on a DOM element that is not a JET component, see the <code class="prettyprint">ojContextMenu</code> binding.
     *
     * <p>To make the page semantically accurate from the outset, applications are encouraged to specify the context menu via the standard
     * HTML5 syntax shown in the below example.  When the component is initialized, the context menu thus specified will be set on the component.
     *
     * {@ojinclude "name":"contextMenuInitOrderDoc"}
     *
     * <p>After create time, the <code class="prettyprint">contextMenu</code> option should be set via this API, not by setting the DOM attribute.
     *
     * <p>The application can register a listener for the Menu's [beforeOpen]{@link oj.ojMenu#event:beforeOpen} event.  The listener can cancel the 
     * launch via <code class="prettyprint">event.preventDefault()</code>, or it can customize the menu contents by editing the menu DOM directly, 
     * and then calling [refresh()]{@link oj.ojMenu#refresh} on the Menu.  
     *
     * {@ojinclude "name":"contextMenuTargetDoc"}
     * 
     * @ojfragment contextMenuDoc - Decomped to fragment so subclasses can extend the verbiage as needed, by ojinclude'ing this fragment and then adding their own verbiage.
     * @memberof oj.baseComponent
     */
    /**
     * {@ojinclude "name":"contextMenuDoc"}
     *
     * @ignore
     * @expose
     * @memberof oj.baseComponent
     * @instance
     * @type {Element|Array.<Element>|string|jQuery|NodeList}
     * @default <code class="prettyprint">null</code>
     *
     * @example <caption>Initialize a JET component with a context menu:</caption>
     * // via recommended HTML5 syntax:
     * &lt;div id="myComponent" contextmenu="myMenu" data-bind="ojComponent: { ... }>
     *
     * // via JET initializer (less preferred) :
     * // Foo is the component, e.g., InputText, InputNumber, Select, etc.
     * $( ".selector" ).ojFoo({ "contextMenu": "#myMenu" });
     *
     * @example <caption>Get or set the <code class="prettyprint">contextMenu</code> option, after initialization:</caption>
     * // getter
     * // Foo is the component, e.g., InputText, InputNumber, Select, etc.
     * var menu = $( ".selector" ).ojFoo( "option", "contextMenu" );
     *
     * // setter
     * // Foo is the component, e.g., InputText, InputNumber, Select, etc.
     * $( ".selector" ).ojFoo( "option", "contextMenu", ".my-marker-class" );
     *
     * @example <caption>Set a JET context menu on an ordinary HTML element:</caption>
     * &lt;a href="#" id="myAnchor" contextmenu="myMenu" data-bind="ojContextMenu: {}">Some text</a>
     */
    contextMenu: null,

    /**
     * <p>Attributes specified here will be set on the component's root DOM element at creation time.
     * This is particularly useful for components like Dialog that wrap themselves in a new root element
     * at creation time.
     *
     * <p>The supported attributes are <code class="prettyprint">id</code>, which overwrites any existing value,
     * and <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended
     * to the current class and style, if any.
     *
     * <p>Setting this option after component creation has no effect.  At that time, the root element already
     * exists, and can be accessed directly via the <code class="prettyprint">widget</code> method, per the second example below.
     *
     * @example <caption>Initialize a JET component, specifying a set of attributes to be set
     * on the component's root DOM element:</caption>
     * // Foo is the component, e.g., Menu, Button, InputText, InputNumber, Select, etc.
     * $( ".selector" ).ojFoo({ "rootAttributes": {
     *   "id": "myId",
     *   "style": "max-width:100%; color:blue;",
     *   "class": "my-class"
     * }});
     *
     * @example <caption>After initialization, <code class="prettyprint">rootAttributes</code> should not be used.  It is
     * not needed at that time, as attributes of the root DOM element can simply be set directly, using
     * <code class="prettyprint">widget</code>:</caption>
     * // Foo is the component, e.g., Menu, Button, InputText, InputNumber, Select, etc.
     * $( ".selector" ).ojFoo( "widget" ).css( "height", "100px" );
     * $( ".selector" ).ojFoo( "widget" ).addClass( "my-class" );
     *
     * @ignore
     * @expose
     * @memberof oj.baseComponent
     * @instance
     * @type {?Object}
     * @default <code class="prettyprint">null</code>
     */
    rootAttributes: null,

    /**
     * <p>A collection of translated resources from the translation bundle, or <code class="prettyprint">null</code> if this
     * component has no resources.  Resources may be accessed and overridden individually or collectively, as seen in the examples.
     *
     * <p> If the component does not contain any translatable resource, the default value of this attribute will be 
     * <code class="prettyprint">null</code>. If not, an object containing all resources relevant to the component.
     * 
     * <p>If this component has translations, their documentation immediately follows this doc entry.
     *
     * @member
     * @name translations
     * @memberof oj.baseComponent
     * @instance
     * @ojtranslatable
     * @type {Object|null}
     * 
     *
     * @example <caption>Initialize the component, overriding some translated resources and leaving the others intact:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-some-element translations.some-key='some value' translations.some-other-key='some other value'>&lt;/oj-some-element>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-some-element translations='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-some-element>
     *
     * @example <caption>Get or set the <code class="prettyprint">translations</code> property after initialization:</caption>
     * // Get one
     * var value = myComponent.translations.someKey;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myComponent.setProperty('translations.someKey', 'some value');
     *
     * // Get all
     * var values = myComponent.translations;
     *
     * // Set all.  Must list every resource key, as those not listed are lost.
     * myComponent.translations = {
     *     someKey: 'some value',
     *     someOtherKey: 'some other value'
     * };
     *
     */
    // translations property is initialized programmatically, so this top-level API doc lives in this virtual comment.
    // Translations for all components are listed and JSDoc'ed in rt\src\main\resources\nls\root\ojtranslations.js.
    // That JSDoc appears in the same generated doc page as this top-level doc.


    // Events
    /**
     * Fired whenever a supported component option changes, whether due to user interaction or programmatic
     * intervention.  If the new value is the same as the previous value, no event will be fired.  The event 
     * listener will receive two parameters described below:
     *
     * @property {Event} event <code class="prettyprint">jQuery</code> event object
     * @property {Object} ui event payload
     * @property {string} ui.option the name of the option that changed.
     * @property {Object} ui.previousValue - an Object holding the previous value of the option.
     * When previousValue is not a primitive type, i.e., is an Object, it may hold the same value as
     * the value property.
     * @property {Object} ui.value - an Object holding the current value of the option.
     * @property {?Object} ui.subproperty - an Object holding information about the subproperty that changed.
     * @property {string} ui.subproperty.path - the subproperty path that changed.
     * @property {Object} ui.subproperty.previousValue - an Object holding the previous value of the subproperty.
     * @property {Object} ui.subproperty.value - an Object holding the current value of the subproperty.
     * @property {Object} ui.optionMetadata information about the option that changed
     * @property {string} ui.optionMetadata.writeback <code class="prettyprint">"shouldWrite"</code> or
     *  <code class="prettyprint">"shouldNotWrite"</code>. For use by the JET writeback mechanism;
     *  'shouldWrite' indicates that the value should be written to the observable.
     *
     * @example <caption>Initialize component with the <code class="prettyprint">optionChange</code> callback</caption>
     * // Foo is Button, InputText, etc.
     * $(".selector").ojFoo({
     *   'optionChange': function (event, ui) {}
     * });
     * @example <caption>Bind an event listener to the ojoptionchange event</caption>
     * $(".selector").on({
     *   'ojoptionchange': function (event, ui) {
     *       // verify that the component firing the event is a component of interest
     *       if ($(event.target).is(".mySelector")) {
     *           window.console.log("option that changed is: " + ui['option']);
     *       }
     *   };
     * });
     *
     * @ignore
     * @memberof oj.baseComponent
     * @expose
     * @event
     * @instance
     */
    optionChange: undefined,

    /**
     * <p>Triggered before the component is destroyed. This event cannot be canceled; the
     * component will always be destroyed regardless.
     *
     * @example <caption>Initialize component with the <code class="prettyprint">destroy</code> callback</caption>
     * // Foo is Button, InputText, etc.
     * $(".selector").ojFoo({
     *   'destroy': function (event, data) {}
     * });
     * @example <caption>Bind an event listener to the destroy event</caption>
     * $(".selector").on({
     *   'ojdestroy': function (event, data) {
     *       // verify that the component firing the event is a component of interest
     *       if ($(event.target).is(".mySelector")) {
     *           window.console.log("The DOM node id for the destroyed component is : %s", event.target.id);
     *       }
     *   };
     * });
     *
     * @ignore
     * @memberof oj.baseComponent
     * @expose
     * @event
     * @instance
     */
    destroy: undefined
  },

  // TODO: flesh out JSDoc verbiage, re: call after dom changes underneath component...
  /**
   * Refreshes the component.
   * @return {void}
   * @expose
   * @memberof oj.baseComponent
   * @instance
   */
  refresh: function()
  {
    this._propertyContext = null;
    // if application sets the context menu after initialization, it must refresh the component
    this._SetupContextMenu();
  },

  /**
   * <p>Overridden to save off component's default options and the options passed into the constructor (to be passed into
   * the _InitOptions() call).
   *
   * <p>This method is final. Components should instead override one or more of the overridable create-time methods
   * listed in <a href="#_ComponentCreate">_ComponentCreate</a>.
   *
   * @memberof oj.baseComponent
   * @instance
   * @private
   * @final
   */
  _createWidget: function(options, element)
  {
    // Save wrapper element
    if (options)
      this.OuterWrapper = options['_wrapper'];

    // There is no need to clone these objects since they are not modified by the _createWidget() in the base class
    this._originalDefaults = this.options || {};
    this._constructorOptions = options || {};

    this._super(options, element);
    this._AfterCreateEvent();
  },

  /**
   * <p>Reads the <code class="prettyprint">rootAttributes</code> option, and sets the root attributes on the
   * component's root DOM element.  See <a href="#rootAttributes">rootAttributes</a> for the set of supported
   * attributes and how they are handled.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @throws if unsupported attributes are supplied.
   */
  _SetRootAttributes : function ()
  {
    var value = this.options.rootAttributes;
    if (value)
    {
      var widget = this.widget();
      if (widget == null)
        return;

      var classValue = value["class"];

      if (classValue)
      {
        widget.addClass(classValue);
      }

      var styleValue = value["style"];

      if (styleValue)
      {
        var currStyle = widget.attr('style');

        if (currStyle)
        {
          widget.attr('style', currStyle + ';' + styleValue);
        }
        else
        {
          widget.attr('style', styleValue);
        }
      }

      // make shallow copy, remove class and style from the copy, and set all
      // remaining attrs on the element.  Currently id is the only remaining attr
      // that we support.
      value = $.extend({}, value);
      delete value['class'];
      delete value['style'];

      widget.attr(value);

      delete value['id']; // remove the remaining supported value
      var unsupportedAttrs = Object.keys(value);
      if (unsupportedAttrs.length)
        throw new Error("Unsupported values passed to rootAttributes option: " + unsupportedAttrs.toString());
    }
  },

  /**
   * <p>This method is final in JET. Components should instead override one or more of the overridable create-time methods
   * listed in <a href="#_ComponentCreate">_ComponentCreate</a>.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @final
   */
  _create : function()
  {
    this._SaveAttributes(this.element);
    this._InitOptions(this._originalDefaults, this._constructorOptions);

    delete this._originalDefaults;
    delete this._constructorOptions;

    this._ComponentCreate();
    this._AfterCreate();

    // allow subcomponent to setup needed resources
    // after the component is created.
    this._SetupResources();

    // Marker class for all JET components on the init node (as opposed to the outer node)
    // This marker class is used to:
    // 1) find all JET components within a subtree
    // 2) to prevent FOUC:  init nodes NOT yet having this class are hidden.
    this.element.addClass(_OJ_COMPONENT_NODE_CLASS);
  },

  /**
   * <p>This method is not used in JET. Components should instead override <a href="#_InitOptions">_InitOptions</a>.
   *
   * @method
   * @name oj.baseComponent#_getCreateOptions
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @final
   */

  /**
   * <p>This method is called before <a href="#_ComponentCreate">_ComponentCreate</a>, at which point
   * the component has not yet been rendered.  Component options should be initialized in this method,
   * so that their final values are in place when <a href="#_ComponentCreate">_ComponentCreate</a> is called.
   *
   * <p>This includes getting option values from the DOM, where applicable, and coercing option
   * values (however derived) to their appropriate data type if needed.
   *
   * <p>No work other than setting options should be done in this method.  In particular, nothing should be
   * set on the DOM until <a href="#_ComponentCreate">_ComponentCreate</a>, e.g. setting the <code class="prettyprint">disabled</code>
   * DOM attribute from the <code class="prettyprint">disabled</code> option.
   *
   * <p>A given option (like <code class="prettyprint">disabled</code>) appears in the <code class="prettyprint">constructorOptions</code>
   * param iff the app set it in the constructor:
   *
   * <ul>
   *   <li>If it appears in <code class="prettyprint">constructorOptions</code>, it should win over what's in the DOM
   *     (e.g. <code class="prettyprint">disabled</code> DOM attribute).  If for some reason you need to tweak the value
   *     that the app set, then enable writeback when doing so:
   *     <code class="prettyprint">this.option('foo', bar, {'_context': {writeback: true, internalSet: true}})</code>.</li>
   *   <li>If it doesn't appear in <code class="prettyprint">constructorOptions</code>, then that option definitely is not bound,
   *     so writeback is not needed.  So if you need to set the option (e.g. from a DOM attribute), use
   *     <code class="prettyprint">this.option('foo', bar, {'_context': {internalSet: true}})</code>.</li>
   * </ul>
   *
   * <p>Overrides of this method should call <code class="prettyprint">this._super</code> first.
   *
   * @param {!Object} originalDefaults - original default options defined on the component and its ancestors
   * @param {?Object} constructorOptions - options passed into the widget constructor
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _InitOptions : function (originalDefaults, constructorOptions)
  {
    this._setupDefaultOptions(originalDefaults, constructorOptions);
    this._initContextMenuOption(constructorOptions);
  },

  /**
   * <p>All component create-time initialization lives in this method, except the logic that specifically
   * needs to live in <a href="#_InitOptions">_InitOptions</a>, <a href="#_AfterCreate">_AfterCreate</a>,
   * or <a href="#_AfterCreateEvent">_AfterCreateEvent</a>,
   * per the documentation for those methods.  All DOM creation must happen here, since the intent of
   * <a href="#_AfterCreate">_AfterCreate</a>, which is called next, is to contain superclass logic that must
   * run after that DOM is created.
   *
   * <p>Overrides of this method should call <code class="prettyprint">this._super</code> first.
   *
   * <p>Summary of create-time methods that components can override, in the order that they are called:
   *
   * <ol>
   *   <li><a href="#_InitOptions">_InitOptions</a></li>
   *   <li><a href="#_ComponentCreate">_ComponentCreate</a> (this method)</li>
   *   <li><a href="#_AfterCreate">_AfterCreate</a></li>
   *   <li>(The <code class="prettyprint">create</code> event is fired here.)</li>
   *   <li><a href="#_AfterCreateEvent">_AfterCreateEvent</a></li>
   * </ol>
   *
   * <p>For all of these methods, the contract is that overrides must call <code class="prettyprint">this._super</code> <i>first</i>, so e.g., the
   * <code class="prettyprint">_ComponentCreate</code> entry means <code class="prettyprint">baseComponent._ComponentCreate</code>,
   * then <code class="prettyprint">_ComponentCreate</code> in any intermediate subclasses, then
   * <code class="prettyprint">_ComponentCreate</code> in the leaf subclass.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _ComponentCreate : function ()
  {
    // Store widget name, so that oj.Components.__GetWidgetConstructor() can get widget from the element
    _storeWidgetName(this.element, this.widgetName);
    
    // namespace facilitates removing activeable and hoverable handlers handlers separately
    this.activeableEventNamespace = this.eventNamespace + "activeable";
    this.hoverableEventNamespace = this.eventNamespace + "hoverable";
  },

  /**
   * <p>This method is called after <a href="#_ComponentCreate">_ComponentCreate</a>, but before the
   * <code class="prettyprint">create</code> event is fired.  The JET base component does
   * tasks here that must happen after the component (subclass) has created itself in its override of
   * <a href="#_ComponentCreate">_ComponentCreate</a>.  Notably, the base component handles the
   * <a href="#rootAttributes">rootAttributes</a> and <a href="#contextMenu">contextMenu</a> options here,
   * since those options operate on the component root node, which for some components is created in their override
   * of <a href="#_ComponentCreate">_ComponentCreate</a>.
   *
   * <p>Subclasses should override this method only if they have tasks that must happen after a superclass's
   * implementation of this method, e.g. tasks that must happen after the context menu is set on the component.
   *
   * <p>Overrides of this method should call <code class="prettyprint">this._super</code> first.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _AfterCreate : function ()
  {
    this._SetRootAttributes(); // do first, since has no dependencies, but other stuff might care about these attrs

    // namespace facilitates removing contextMenu handlers separately, if app clears the "contextMenu" option
    this.contextMenuEventNamespace = this.eventNamespace + "contextMenu";
    // same for activeable and hoverable handlers
    this.activeableEventNamespace = this.eventNamespace + "activeable";
    this.hoverableEventNamespace = this.eventNamespace + "hoverable";
  },

  /**
   * <p>This method is called after the <code class="prettyprint">create</code> event is fired.
   * Components usually should not override this method, as it is rarely correct to wait until after the
   * <code class="prettyprint">create</code> event to perform a create-time task.
   *
   * <p>An example of a correct usage of this method is [Dialog's auto-open behavior]{@link oj.ojDialog#initialVisibility},
   * which needs to happen after the <code class="prettyprint">create</code> event.
   *
   * <p>Only <i>behaviors</i> (like Dialog auto-open behavior) should occur in this method.  Component <i>initialization</i>
   * must occur earlier, before the <code class="prettyprint">create</code> event is fired, so that
   * <code class="prettyprint">create</code> listeners see a fully inited component.
   *
   * <p>Overrides of this method should call <code class="prettyprint">this._super</code> first.
   *
   * <p>Do not confuse this method with the <a href="#_AfterCreate">_AfterCreate</a> method, which is more commonly used.
   *
   * @method
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _AfterCreateEvent : $.noop,

  /**
   * <p>JET components should almost never implement this JQUI method.  Please consult an architect if you believe you have an exception.  Reasons:
   * <ul>
   *   <li>This method is called at create time, after the <code class="prettyprint">create</code> event is fired.  It is rare
   *       for that to be the appropriate time to perform a create-time task.  For those rare cases, we have the
   *       <a href="#_AfterCreateEvent">_AfterCreateEvent</a> method, which is preferred over this method since it is called only
   *       at that time, not also at re-init time (see next).</li>
   *   <li>This method is also called at "re-init" time, i.e. when the initializer is called after the component has already been created.
   *       JET has not yet identified any desired semantics for re-initing a component.</li>
   * </ul>
   *
   * @method
   * @name oj.baseComponent#_init
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */

   //  - remove JQUI memory leaks and CSS cruft introduced in 1.12 and 1.12.1
  _setOptionClasses:  function() {}, 
  _setOptionDisabled: function() {}, 
  _classes:     function() {return "";},
  _removeClass: function() {return this;},
  _addClass:    function() {return this;},
  _toggleClass: function() {return this;},

  /**
   * <p>Saves the element's attributes. This is called during _create.
   * <a href="#_RestoreAttributes">_RestoreAttributes</a> will restore all these attributes
   * and is called during _destroy.
   * </p>
   * <p> This base class default implementation does nothing.
   * </p>
   * <p>We also have <a href="#_SaveAllAttributes">_SaveAllAttributes</a> and
   * <a href="#_RestoreAllAttributes">_RestoreAllAttributes</a> methods
   *  that save and restore <i>all</i> the attributes on an element.
   *  Component subclasses can opt into these _SaveAllAttributes/_RestoreAllAttributes
   *  implementations by overriding _SaveAttributes and _RestoreAttributes to call
   *  _SaveAllAttributes/_RestoreAllAttributes. If the subclass wants a different implementation
   *  (like save only the 'class' attribute), it can provide the implementation itself in
   *  _SaveAttributes/_RestoreAttributes.
   *
   *
   * @param {Object} element - jQuery selection to save attributes for
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _SaveAttributes : function (element)
  {
    // default implementation does nothing.
  },
  /**
   * <p>Saves all the element's attributes within an internal variable.
   * <a href="#_RestoreAllAttributes">_RestoreAllAttributes</a> will restore the attributes
   * from this internal variable.</p>
   * <p>
   * This method is final in JET.
   * Subclasses can override _RestoreAttributes and call _RestoreAllAttributes.
   * </p>
   *
   * <p>The JSON variable will be held as:
   *
   * <pre class="prettyprint">
   * <code>[
   *   {
   *   "element" : element[i],
   *   "attributes" :
   *     {
   *       attributes[m]["name"] : {"attr": attributes[m]["value"]
   *     }
   *   }
   * ]
   * </code></pre>
   *
   * @param {Object} element - jQuery selection to save attributes for
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @final
   */
  _SaveAllAttributes : function (element)
  {
    var self = this;
    this._savedAttributes = [];

    $.each(element, function (index, ele)
    {
      //need to be able to save for multiple elements
      var saveAttributes = {},
          save = { "element" : ele, "attributes" : saveAttributes },
          attributes = ele.attributes;

      self._savedAttributes.push(save);

      $.each(attributes, function (index, attr)
      {
        // for proper access certain so called attributes should be accessed as properties
        // [i.e. required, disabled] so fetch them initially
        var attrName = attr["name"];

        saveAttributes[attrName] = { "attr" : attr["value"] };
      });

    });

  },

  /**
   * <p>Gets the saved attributes for the provided element.
   *
   * <p>If you don't override <a href="#_SaveAttributes">_SaveAttributes</a> and
   * <a href="#_RestoreAttributes">_RestoreAttributes</a>, then this will return null.
   * <p>If you override _SaveAttributes to call <a href="#_SaveAllAttributes">_SaveAllAttributes</a>,
   * then this will return all the attributes.
   * If you override _SaveAttributes/_RestoreAttributes to do your own thing, then you may also have
   * to override _GetSavedAttributes to return whatever you saved if you need access to the saved
   * attributes.
   *
   * @param {Object} element - jQuery selection, should be a single entry
   * @return {Object|null} savedAttributes - attributes that were saved for this element
   * in <a href="#_SaveAttributes">_SaveAttributes</a>, or null if none were saved.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _GetSavedAttributes : function (element)
  {
    var savedAttributes = this._savedAttributes;

    // The component may not have saved any attributes. If so, return.
    if (savedAttributes === undefined)
      return null;

    element = element[0];

    for (var i = 0, j = savedAttributes.length;i < j;i++)
    {
      var curr = savedAttributes[i];

      if (curr["element"] === element)
      {
        return curr["attributes"];
      }
    }

    return {};
  },
  /**
   * <p>Restore the attributes saved in <a href="#_SaveAttributes">_SaveAttributes</a>.</p>
   * <p>
   * _SaveAttributes is called during _create. And _RestoreAttributes is called during _destroy.
   * </p>
   * <p> This base class default implementation does nothing.
   * </p>
   * <p>We also have <a href="#_SaveAllAttributes">_SaveAllAttributes</a> and
   * <a href="#_RestoreAllAttributes">_RestoreAllAttributes</a> methods
   *  that save and restore <i>all</i> the attributes on an element.
   *  Component subclasses can opt into these _SaveAllAttributes/_RestoreAllAttributes
   *  implementations by overriding _SaveAttributes and _RestoreAttributes to call
   *  _SaveAllAttributes/_RestoreAllAttributes. If the subclass wants a different implementation
   *  (like save only the 'class' attribute), it can provide the implementation itself in
   *  _SaveAttributes/_GetSavedAttributes/_RestoreAttributes.
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _RestoreAttributes : function ()
  {
    // default implementation does nothing.
  },
  /**
   * <p>Restores <i>all</i> the element's attributes which were saved in
   * <a href="#_SaveAllAttributes">_SaveAllAttributes</a>.
   * This method is final in JET.</p>
   * <p>
   * If a subclass wants to save/restore all attributes on create/destroy, then the
   * subclass can override <a href="#_SaveAttributes">_SaveAttributes</a>
   *  and call  <a href="#_SaveAllAttributes">_SaveAllAttributes</a> and also
   *  override <a href="#_RestoreAttributes">_RestoreAttributes</a>
   *  and call <a href="#_RestoreAllAttributes">_RestoreAllAttributes</a>.
   *
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @final
   */
  _RestoreAllAttributes : function ()
  {

    $.each(this._savedAttributes, function (index, savedAttr)
    {
      var element = $(savedAttr["element"]),
          attributes = savedAttr["attributes"];

      //sanity check
      if (element.length === 1)
      {
        var currAttr = savedAttr["element"].attributes,
            removeAttr = [];

        //request is to remove any attributes that didn't exist previously
        //need to store the attributes in an array and remove them afterwards as otherwise there are side affects
        for(var i=0, j=currAttr.length; i < j; i++)
        {
          if(!(currAttr[i]["name"] in attributes))
          {
            removeAttr.push(currAttr[i]["name"]);
          }
        }

        for(var i=0, j=removeAttr.length; i < j; i++)
        {
          element.removeAttr(removeAttr[i]);
        }

        for (var attribute in attributes)
        {
          element.attr(attribute, attributes[attribute]["attr"]);
        }
      }

    });

  },


  /**
   * <p>Determines the name of the translation bundle section for this component.
   *
   * @return {string} the name of this component's translations section
   * @memberof oj.baseComponent
   * @protected
   */
  _GetTranslationsSectionName: function()
  {
    return this.widgetFullName;
  },


  /**
   * Compares 2 option values for equality and returns true if they are equal; false otherwise.
   * This method is called before _setOptions()/_internalSetOptions() to prevent an extra call 
   * with the same values when observables are written back. Components should override this 
   * method for options with non primitive writeback values like Arrays or Objects and ensure 
   * their metadata has writeback properties correctly indicated.
   *
   * @param {String} option - the name of the option
   * @param {Object} value1 first value
   * @param {Object} value2 another value
   * @return {boolean}
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _CompareOptionValues : function (option, value1, value2)
  {
    // We process the metadata for custom elements for writeback properties and save them on the component. 
    // For jQuery syntax, components are expected to override this method to check writeback values since 
    // there's not always a straightforward mapping of custom element to jQuery widget name.
    if (this._IsCustomElement() && this._getWritebackOption(option))
    {
      return oj.Object.compareValues(value1, value2);
    }
    else
    {
      return value1 == value2;
    }
  },


  /**
   * <p>Retrieves a translated string after inserting optional parameters.
   *
   * @param {string} key the translations resource key
   * The key is used to retrieve a format pattern from the component options, or if none
   * is found - from the translated resource bundle.
   * Tokens like {0}, {1}, {name} within the pattern will be used to define placement
   * for the optional parameters.  Token strings should not contain comma (,)
   * or space characters, since they are reserved for future format type enhancements.
   * The reserved characters within a pattern are:
   * $ { } [ ]
   * These characters will not appear in the formatted output unless they are escaped
   * with a dollar character ('$').
   *
   * @param {...string|Object|Array} var_args  - optional parameters to be inserted into the
   * translated pattern.
   *
   * If more than one var_args arguments are passed, they will be treated as an array
   * for replacing positional tokens like {0}, {1}, etc.
   * If a single argument is passed, it will be treated as a Javascript Object whose
   * keys will be matched to tokens within the pattern. Note that an Array is just
   * a special kind of such an Object.
   *
   * For backward compatibility, a var_args argument whose type is neither
   * Object or Array will be used to replace {0} in the pattern.
   *
   * @return formatted translated string or the key argument if the resource for the
   * key was not found
   *
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  // TODO: non-public methods need to start with "_".  Pinged architect, who thinks this
  // method should become protected post-V1, which would imply a capital _GetTranslatedString
  getTranslatedString : function (key, var_args)
  {
    var params = {}, pattern;

    if (arguments.length > 2)
    {
      params = Array.prototype.slice.call(arguments, 1);
    }
    else if (arguments.length == 2)
    {
      params = arguments[1];
      if (typeof params !== 'object' && !(params instanceof Array))
      {
        params = [params];
      }

    }
    pattern = this.option(_OJ_TRANSLATIONS_PREFIX + key);
    // pattern could be undefined
    return (pattern == null) ? key : oj.Translations.applyParameters(pattern.toString(), params);
  },

  // Subclasses should doc their sub-id's in the Sub-ID's section, via the ojsubid tag, not by overriding
  // and extending this method doc, which should remain general purpose.
  /**
   * <p>Returns the DOM node indicated by the <code class="prettyprint">locator</code> parameter.
   *
   * <p>If the <code class="prettyprint">locator</code> or its <code class="prettyprint">subId</code> is
   * <code class="prettyprint">null</code>, then this method returns this element.
   *
   * <p>If a non-null <code class="prettyprint">subId</code> is provided but no corresponding node
   * can be located, then this method returns <code class="prettyprint">null</code>.
   *
   * <p>This method is intended for use in test automation only, and should not be used in a production environment.
   *
   * @expose
   * @memberof oj.baseComponent
   * @instance
   * @ignore
   *
   * @param {Object} locator An Object containing, at minimum, a <code class="prettyprint">subId</code>
   * property, whose value is a string that identifies a particular DOM node in this element.
   *
   * <p>If this component has any subIds, then they are documented in the
   * <a href="#subids-section">Sub-ID's</a> section of this document.
   *
   * <p>Some components may support additional fields of the
   * <code class="prettyprint">locator</code> Object, to further specify the desired node.
   *
   * @returns {Element|null} The DOM node located by the 
   * <code class="prettyprint">locator</code>, or <code class="prettyprint">null</code> if none is found.
   *
   * @example <caption>Get the node for a certain subId:</caption>
   * var node = myComponent.getNodeBySubId({'subId': 'oj-some-sub-id'});
   */
  getNodeBySubId: function(locator)
  {
    if (locator == null || locator['subId'] == null)
    {
      return this.element ? this.element[0] : null;
    }

    // Non-null locators have to be handled by the component subclasses
    return null;
  },

  /**
   * <p>Returns the subId string for the given DOM node in this element.  For details, see
   * <a href="#getNodeBySubId">getNodeBySubId</a> and the <a href="#subids-section">Sub-ID's</a>
   * section of this document.
   * 
   * <p>This method is intended for use in test automation only, and should not be used in a production environment.
   *
   * @ojfragment getSubIdByNodeDesc
   * @memberof oj.baseComponent
   */
  /**
   * DOM node in this element
   * 
   * @ojfragment getSubIdByNodeNodeParam
   * @memberof oj.baseComponent
   */
  /**
   * The subId for the DOM node, or <code class="prettyprint">null</code> if none is found.
   * 
   * @ojfragment getSubIdByNodeReturn
   * @memberof oj.baseComponent
   */
  /**
   * Get the subId for a certain DOM node:
   * 
   * @ojfragment getSubIdByNodeCaption
   * @memberof oj.baseComponent
   */
  /**
   * var locator = myComponent.getSubIdByNode(nodeInsideElement);
   * 
   * @ojfragment getSubIdByNodeExample
   * @memberof oj.baseComponent
   */

  // While a subclass could technically extend the verbiage by adding its own verbiage after the ojinclude, 
  // please doc sub-id's in the subid's section, not by extending this method doc.
  /**
   * {@ojinclude "name":"getSubIdByNodeDesc"}
   *
   * @expose
   * @memberof oj.baseComponent
   * @instance
   * @ignore
   *
   * @param {!Element} node {@ojinclude "name":"getSubIdByNodeNodeParam"}
   * @returns {Object|null} {@ojinclude "name":"getSubIdByNodeReturn"}
   * 
   * @example <caption>{@ojinclude "name":"getSubIdByNodeCaption"}</caption>
   * {@ojinclude "name":"getSubIdByNodeExample"}
   */
  getSubIdByNode: function(node)
  {
    return null;
  },

  // Overridden to set oj-hover and oj-focus classes
  // TODO: Move JSDoc from subclasses to here.  Don't include above internal comment.  Make at-final.
  destroy: function()
  {
    if (this._IsCustomElement())
      throw new Error("destroy cannot be called on a custom element");

    // Fire 'destroy' event
    this._trigger('destroy');
    
    // Since jQuery event listeners get removed before the destroy() method whne jQuery.clean() cleans up the subtree,
    // we need to fire a custom DOM event as well. This will allow component binding to execute destroy callbacks for
    // custom bindings and managed attributes.
    
    oj.DomUtils.dispatchEvent(this.element[0], new CustomEvent("_ojDestroy"));

    this._super();

    //remove hover and active listeners
//    this.widget().off(this.eventNamespace);

    // allow subcomponent to release resources they hold.
    this._ReleaseResources();

    // clean up states
    this.element.removeClass(_OJ_COMPONENT_NODE_CLASS);
    this.widget().removeClass( "oj-disabled" );
    
    // pass init node (this.element), not root node if different (this.widget()), since all elements in 
    // the root node subtree but not the init node subtree should have been removed by the call to _super.
    this._removeStateClasses( this.element );

    _removeWidgetName(this.element, this.widgetName);

    this._RestoreAttributes();

    // TODO: move this to _RestoreAttributes?
    if (this._initialCmDomAttr)
      this.element.attr("contextmenu", this._initialCmDomAttr);
    else
      this.element.removeAttr("contextmenu");

    this._propertyContext = null;
  },

  /*
   * Internal notes:
   * Overridden to pass extra flags to _setOption
   * param {...Object} var_args - key (or map), value, flags
   */
  /**
   * <p>This method has several overloads, which get and set component options and their fields.  The functionality is unchanged from
   * that provided by JQUI.  See the examples for details on each overload.
   *
   * @ignore
   * @expose
   * @memberof oj.baseComponent
   * @instance
   * @final
   *
   * @param {string|Object=} optionName the option name (string, first two overloads), or the map (Object, last overload).
   *        Omitted in the third overload.
   * @param {Object=} value a value to set for the option.  Second overload only.
   * @return {Object|undefined} The getter overloads return the retrieved value(s).  When called via the public jQuery syntax, the setter overloads
   *         return the object on which they were called, to facilitate method chaining.
   *
   * @example <caption>First overload: get one option:
   * <p>This overload accepts a (possibly dot-separated) <code class="prettyprint">optionName</code> param as a string, and returns
   * the current value of that option.</caption>
   * var isDisabled = $( ".selector" ).ojFoo( "option", "disabled" ); // Foo is Button, Menu, etc.
   *
   * // For object-valued options, dot notation can be used to get the value of a field or nested field.
   * var startIcon = $( ".selector" ).ojButton( "option", "icons.start" ); // icons is object with "start" field
   *
   * @example <caption>Second overload: set one option:
   * <p>This overload accepts two params: a (possibly dot-separated) <code class="prettyprint">optionName</code> string, and a new value to
   * which that option will be set.</caption>
   * $( ".selector" ).ojFoo( "option", "disabled", true ); // Foo is Button, Menu, etc.
   *
   * // For object-valued options, dot notation can be used to set the value
   * // of a field or nested field, without altering the rest of the object.
   * $( ".selector" ).ojButton( "option", "icons.start", myStartIcon ); // icons is object with "start" field
   *
   * @example <caption>Third overload: get all options:
   * <p>This overload accepts no params, and returns a map of key/value pairs representing all the component
   * options and their values.</caption>
   * var options = $( ".selector" ).ojFoo( "option" ); // Foo is Button, Menu, etc.
   *
   * @example <caption>Fourth overload: set one or more options:
   * <p>This overload accepts a single map of option-value pairs to set on the component.  Unlike the first two
   * overloads, dot notation cannot be used.</caption>
   * $( ".selector" ).ojFoo( "option", { disabled: true, bar: 42 } ); // Foo is Button, Menu, etc.
   */
  option: function(optionName, value) // actually varArgs per comment above the JSDoc, but GCC warns unless matches the @param that we wish to doc
  {
    if (arguments.length === 0)
    {
      // don't return a reference to the internal hash
      return $.widget.extend({}, this.options);
    }

    var key = arguments[0];

    var options = key;
    var subkey = null;

    var flags = {};

    if (typeof key === "string")
    {
      // handle nested keys, e.g., "foo.bar" => { foo: { bar: ___ } }
      options = {};
      var parts = key.split(".");
      key = parts.shift();
      if (parts.length)
      {
        subkey = parts.join(".");

        var curOption;
        try
        {
          // Inform dynamic getters that the subkey is being set
          if (arguments.length > 1)
          {
            this._settingNestedKey = subkey;
          }

          curOption = options[key] = $.widget.extend({}, this.options[key]);
        }
        finally
        {
          this._settingNestedKey = null;
        }

        for (var i = 0; i < parts.length - 1; i++)
        {
          curOption[parts[i]] = curOption[parts[i]] || {};
          curOption = curOption[parts[i]];
        }

        key = parts.pop();
        if (arguments.length === 1)
        {
          return curOption[key] === undefined ? null : curOption[key];
        }

        curOption[key] = value;
      }
      else
      {
        if (arguments.length === 1)
        {
          return this.options[key] === undefined ? null : this.options[key];
        }
        options[key] = value;
      }

      flags = arguments[2] || flags;
    }
    else
    {
      flags = arguments[1] || flags;
    }

    // Store subkey on the flags to let _setOption() know that dot notation was used
    if (subkey != null)
    {
      var subprop = {
        'path': optionName,
        'value': value
      }
      flags = $.widget.extend({}, flags, {'subkey': subkey, 'subproperty': subprop});
    }

    var context = flags ? flags['_context'] : null;
    var internalSet = context ? context.internalSet : false;

    // This method can be called twice with the same value for writeback properties
    // so we need to go through the options object and only pass through the changed values
    var newOptions = {};
    for (var option in options)
    {
      var newValue = options[option];
      var oldValue = this.options[option];
      // The changed flag is set when components have updated an object or array value in place
      var changed = flags && flags['changed'];
      if (changed || !this._CompareOptionValues(option, oldValue, newValue))
      {
        newOptions[option] = newValue;
      }
    }

    if (Object.keys(newOptions).length > 0)
    {
      // Avoid _setOption() calls for internal sets, since component's _setOption()
      // and setOptions() overrides do not expect to be called in that case
      if (internalSet)
      {
        this._internalSetOptions(newOptions, flags);
      }
      else
      {
        this._setOptions(newOptions, flags);
      }
    }

    return this;

  },

  /**
   * option() calls this rather than _setOption() if the caller was internal.
   *
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _internalSetOptions: function (options, flags)
  {
    for (var key in options)
    {
      var value = options[key];
      var oldValue = this.options[key];
      this.options[key] = value;
      this._optionChanged(key, value, oldValue, flags);
    }
  },

  /**
   * <p>Overridden to pass extra flags to _setOption.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @ignore
   */
  _setOptions : function (options, flags)
  {
    for (var key in options)
    {
      var value = options[key];
      this._setOption(key, value, flags);
    }

    return this;
  },

  /**
   * Overridden to set oj-hover and oj-focus classes.
   * Components should not call this method directly, but instead call option().
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @ignore
   */
  _setOption: function(key, value, flags)
  {
    var originalValue = this.options[key];

    if (key === "disabled" )
    {
      // The JQUI superclass method has hard-coded style classes in the 'if key === "disabled"' block, so unfortunately
      // we must copy that logic here with updated style classes, and NOT call _super() for the disabled case.
      // TBD: keep this logic updated if superclass method changes.
      this.options[ key ] = value;

      // TBD: widget() is not always the thing that should have aria-disabled on it.  E.g. for the checkbox/radio flavors of ojButton,
      // widget() returns the root node, but aria-disabled belongs on the <input>.  We fixed this JQUI bug in ojButton by having ojButton
      // override this method to remove it from the root node and add it to the input.  Would be better for each component to know which
      // element to apply that to, e.g. an overridable method returning that element, or copying "hoverable" paradigm if appropriate.
      // In the cases where this.element is different than widget(), this.element is more likely to be the right thing, so maybe change
      // default to that.
      // Update: this issue is getting even more awkward now that we have "effectively disabled".  Probably need to refactor this code!
      this.widget()
	  .toggleClass( "oj-disabled", !!value )
	  .attr( "aria-disabled", value );

      if (value)
      {
        this._removeStateClasses( this.widget() );
      }
    }
    else {
      try
      {
        var subkey = (flags == null)? null :flags['subkey'];
        if (subkey != null)
        {
          this._settingNestedKey = subkey;
        }

        this._super(key, value);
      }
      finally
      {
        this._settingNestedKey = null;
      }

      // if contextMenu option wasn't set before, we'll need to start detect gesture
      if ( key === "contextMenu" )
      {
        this._SetupContextMenu();
      }
    }

    this._optionChanged(key, value, originalValue, flags);

    return this;
  },

  /**
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _optionChanged: function(key, value, originalValue, flags)
  {
    // Assume that all values are different from originalValues as we do equality
    // checking before calling this method or trust components that have set the 
    // 'changed' flag to indicate that they updated in place for an Object or Array
    // since we won't see any difference for those cases.
    var changed = false;
    var context = null;

    var writeback = false;
    var readOnly = false;
    var originalEvent = null;
    var updatedFrom = 'external';

    var optionMetadata = null;

    var extraData;

    if (flags)
    {
      context = flags['_context'];

      if (context)
      {
        originalEvent = context.originalEvent;
        writeback = (context.writeback === undefined) ? originalEvent != null : context.writeback;
        readOnly = context.readOnly;
        optionMetadata = context.optionMetadata;
        extraData = context.extraData;
        if (context.internalSet)
          updatedFrom = 'internal';
      }
    }

    optionMetadata = optionMetadata || {};
    optionMetadata['writeback'] =  writeback ? "shouldWrite" : "shouldNotWrite";
    
    if (readOnly)
    {
      optionMetadata['readOnly'] = true;
    }

    var optionChangeData =
    {
      "option" : key,
      "previousValue" : originalValue,
      "value" : value,
      "optionMetadata" : optionMetadata,
      "updatedFrom": updatedFrom
    };

    var subkey = (flags == null) ? null : flags['subkey'];
    // Walk previousValue object and find the subproperty previousValue
    if (subkey)
    {
      var subprops = subkey.split('.');
      var originalSubpropValue = originalValue;
      subprops.forEach(function(subprop) {
        if (!originalSubpropValue)
          return;
        originalSubpropValue = originalSubpropValue[subprop];
      });
      var subproperty = flags['subproperty'];
      subproperty['previousValue'] = originalSubpropValue;
      optionChangeData['subproperty'] = subproperty;
    }

    if (extraData != null)
    {
      optionChangeData = $.extend({}, extraData, optionChangeData);
    }

    this._trigger("optionChange", originalEvent, optionChangeData);
  },

  /**
   * <p>Sets up needed resources for this component, for example, add
   * listeners. This is called during _create.
   * <a href="#_ReleaseResources">_ReleaseResources</a> will release resources
   * help by this component, and is called during destroy.
   * </p>
   *  Component subclasses can opt in by overriding _SetupResources and
   *   _ReleaseResources.
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _SetupResources: function ()
  {
    this._SetupContextMenu();
  },

  /**
   * <p>Release resources held by this component, for example, remove
   * listeners. This is called during destroy.
   * <a href="#_SetupResources">_SetupResources</a> will set up resources
   * needed by this component, and is called during _create.
   * </p>
   *  Component subclasses can opt in by overriding _SetupResources and
   *   _ReleaseResources.
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _ReleaseResources: function()
  {
    this._ReleaseContextMenu();
  },

  /**
   * <p>Overridden to change the way the component events are treating original events:
   *
   * 1) preventDefault(), stopPropagation() and stopImmediatePropagation() no longer invoke
   *    the corresponding methods on the .originalEvent
   * 2) Properties of the .originalEvent are no longer copied to the new event being
   *    triggered
   *
   * @param {string} type - component event type
   * @param {?Object} event - original event
   * @param {Object=} data - event data
   * @return {boolean} true if the default action has not been prevented, false otherwise
   *
   * @private
   */
  _trigger : function (type, event, data)
  {
    return this._trigger2(type, event, data)['proceed'];
  },

  /**
   * <p>Same as _trigger(), but returns an object containing both the "prevented" status and the event.
   * 
   * <p>This is useful for event chaining, so that the returned event (e.g. Menu's select event) can be 
   * passed as the originalEvent of a subsequent event caused by the first (e.g. Menu's close event).
   *
   * @param {string} type - component event type
   * @param {?Object} event - original event
   * @param {Object=} data - event data
   * @return {!{proceed: boolean, event: ($.Event|CustomEvent)}} 
   *     proceed is true if the default action has not been prevented, false otherwise
   *     event is the new event that was triggered
   *
   * @private
   */
  _trigger2 : function (type, event, data)
  {
    data = data || {};

    if (this._IsCustomElement()) {
      return this._triggerCustomEvent(type, event, data);
    }
    else {
      var callback = this.options[type];

      event = $.Event(event, _OJ_COMPONENT_EVENT_OVERRIDES);
      event.type = (this.widgetEventPrefix + type).toLowerCase();

      // the original event may come from any element
      // so we need to reset the target on the new event
      event.target = this.element[0];

      this.element.trigger(event, data);

      return {
        'proceed': !($.isFunction(callback) && callback.apply(this.element[0], [event].concat(data)) === false || event.isDefaultPrevented()), 
        'event': event
      }
    }
  },

  /**
   * <p>Fires a CustomEvent instead of a jQuery event for when this component is created as a custom element.
   * 
   * @param {string} type - component event type
   * @param {?Object} event - original event
   * @param {Object} data - event data
   * @return {!{proceed: boolean, event: CustomEvent}} 
   *     proceed is true if the default action has not been prevented, false otherwise
   *     event is the new event that was triggered
   *
   * @private
   */
  _triggerCustomEvent : function (type, event, data) {
    var eventName, detail = {}, bubbles, cancelable, rootElement = this._getRootElement();
    if (type === 'optionChange') {
      var property = oj.CustomElementBridge.getPropertyForAlias(rootElement, data['option']);
      if (!oj.CustomElementBridge.isKnownProperty(rootElement, property)) {
        return {'proceed':true, 'event':null};        
      }
      eventName = oj.__AttributeUtils.propertyNameToChangeEventType(property);

      // Copy over component specific optionChange event properties, promoting those exposed
      // in the optionMetadata to the top level alongside value/previousValue/subproperty. 
      var keys = Object.keys(data);
      for (var i = 0; i < keys.length; i++)
      {
        var key = keys[i];
        if (key === 'option')
        {
          continue;
        }
        else if (key === 'optionMetadata')
        {
          // Do not expose optionMetadata for custom element property change events
          // Instead, promote any component specific metadata to the top level.
          var metaKeys = Object.keys(data[key]);
          for (var j = 0; j < metaKeys.length; j++)
          {
            var metaKey = metaKeys[j];
            if (metaKey === 'writeback' || metaKey === 'component')
            {
              continue;
            }
            else
            {
              detail[metaKey] = data[key][metaKey];
            }
          }
        }
        else
        {
          detail[key] = data[key];
        }
      }
    }
    else {
      if (!oj.CustomElementBridge.isKnownEvent(rootElement, type)) {
        return {'proceed':true, 'event':null};
      }
      bubbles = true;
      cancelable = true;
      eventName = oj.__AttributeUtils.eventTriggerToEventType(type);
      detail = this._resolveJQueryObjects(data);
    }
    if (event) {
      detail['originalEvent'] = event instanceof $.Event ? event.originalEvent : event;
    }
    var params = {'detail': detail};
    if (bubbles)
      params['bubbles'] = true;
    if (cancelable)
      params['cancelable'] = true;
    
    var customEvent = new CustomEvent(eventName, params);
    rootElement.dispatchEvent(customEvent);
    return {'proceed':!customEvent.defaultPrevented, 'event':customEvent};
  },

  /**
   * <p>Creates a shallow copy of the passed in object where any top-level JQuery objects have been resolved to their underlying objects.
   * @param {Object} data the original object
   * @return {Object} the resolved object copy
   *
   * @private
   */
  _resolveJQueryObjects: function(data) {
    var resolved = oj.CollectionUtils.copyInto({}, data);
    var val, key, keys = Object.keys(resolved);
    for (var k = 0; k < keys.length; k++) {
      key = keys[k];
      val = resolved[key];
      if (val && val instanceof $) {
        if (val.length == 0) {
          resolved[key] = null;
        }          
        else if (val.length == 1) {
          resolved[key] = val[0];
        }
        else {
          resolved = val.toArray();
        }
      }
    }
    return resolved;
  },


  /**
   * <p>Sets contextMenu option from DOM if option not set.
   *
   * <p>Do not override.  To be called only from _InitOptions().
   *
   * @memberof oj.baseComponent
   * @instance
   * @private
   * @final
   */
  _initContextMenuOption: function(constructorOptions)
  {
    var contextMenu = this.element.attr("contextmenu");

    this._initialCmDomAttr = contextMenu; // TODO: remove this after the _RestoreAttributes() call in destroy() is uncommented

    if (contextMenu && !('contextMenu' in constructorOptions)) // if app set DOM attr but not option, then set the option from the DOM
    {
      this.option("contextMenu",
                  document.getElementById(contextMenu),
                  {'_context': {internalSet: true}}); // writeback not needed since "not in constructorOptions" means "not bound"
    }
  },

  /**
   * Handler 
   * @param {Element} contextMenu root element of context menu
   * @param {Event} event the dom event
   * @param {string} eventType the type of event
   * @private
   */
  _handleContextMenuGesture: function(contextMenu, event, eventType)
  {
      // For components like Button where "effectively disabled" --> "not focusable", keyboard CM launch is impossible, so
      // allowing right-click access would be an a11y issue.  If there's ever a need to enable this for focusable effectively
      // disabled components, we can always replace the _IsEffectivelyDisabled() call with a new protected method whose
      // baseComponent impl returns _IsEffectivelyDisabled().
      if (this._IsEffectivelyDisabled())
        return;

      // contextMenu should reference the latest one from the scope
      var menu;
      if (contextMenu.tagName === "OJ-MENU")
      {
        menu = contextMenu;
      }
      else
      {
        var constructor = oj.Components.__GetWidgetConstructor(contextMenu, "ojMenu");
        menu = constructor && constructor("instance");
        if (!menu)
          throw new Error('Invalid JET Menu.'); // keeping old behavior
      }

      this._NotifyContextMenuGesture(menu, event, eventType);

      // todo: modify NotifyContextMenuGesture contract so we don't need to check visible
      if ($(contextMenu).is(":visible"))
      {
        event.preventDefault(); // don't show native context menu
      }
  },

  /**
   * <p>Call this method from _SetupResources().  It sets up listeners needed to detect context menu gestures.
   *
   * <p>We don't look for the menu until context menu gesture has been detected on the first launch,
   * so that the menu needn't be inited before this component.
   *
   * <p>If needed, override <code class="prettyprint">_NotifyContextMenuGesture()</code>, not this private method.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _SetupContextMenu: function()
  {
    var contextMenu = this._GetContextMenu();
    if (!contextMenu)
    {
      contextMenu = this._GetDefaultContextMenu();
    }

    if (contextMenu && this._contextMenuGestureInit === undefined)
    {    
      this._contextMenuGestureInit = contextMenu;

      var self = this;
      oj.GestureUtils.startDetectContextMenuGesture(this.widget()[0], function(event, eventType)
      {
        self._handleContextMenuGesture(contextMenu, event, eventType);
      }); 
    }
  },

  /**
   * <p>This method removes contextMenu functionality from the component and specified menu.
   *
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _ReleaseContextMenu: function()
  {
    this._contextMenuGestureInit = undefined;

    // access menu elem directly, rather than using the widget() of the Menu component, so listener is cleared even if component no longer exists.
    // $(contextMenuOption).off( this.contextMenuEventNamespace );
    oj.GestureUtils.stopDetectContextMenuGesture(this.widget()[0]);
  },

  /**
   * Helper method to retrieve the context menu element
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _GetContextMenu: function()
  {
    if (this._IsCustomElement())
    {
      var slotMap = oj.BaseCustomElementBridge.getSlotMap(this._getRootElement());        
      var slot = slotMap['contextMenu'];
      if (slot && slot.length > 0)
        return slot[0];
    }
    else
    {
      if (this.options.contextMenu)
        return $(this.options.contextMenu).first()[0];
    }

    return null;
  },

  /**
   * <p>When the <a href="#contextMenu">contextMenu</a> option is set, this method is called when the user invokes the context menu via
   * the default gestures: right-click, <kbd>Press & Hold</kbd>, and <kbd>Shift-F10</kbd>.  Components should not call this method directly.
   *
   * <p>The default implementation simply calls <a href="#_OpenContextMenu">this._OpenContextMenu(event, eventType)</a>.
   * Overrides of this method should call that same method, perhaps with additional params, not [menu.open()]{@link oj.ojMenu#open}.
   *
   * <p>This method may be overridden by components needing to do things like the following:
   *
   * <ul>
   * <li>Customize the [launcher]{@link oj.ojMenu#openOptions.launcher} or [position]{@link oj.ojMenu#openOptions.position} passed to
   * <a href="#_OpenContextMenu">_OpenContextMenu()</a>.  See that method for guidance on these customizations.</li>
   *
   * <li>Customize the menu contents.  E.g. some components need to enable/disable built-in commands like <kbd>Cut</kbd> and <kbd>Paste</kbd>,
   * based on state at launch time.</li>
   *
   * <li>Bail out in some cases.  E.g. components with UX approval to use <kbd>PressHoldRelease</kbd> rather than <kbd>Press & Hold</kbd> can override this method
   * to say <code class="prettyprint">if (eventType !== "touch") this._OpenContextMenu(event, eventType);</code>.  When those components
   * detect the alternate context menu gesture (e.g. <kbd>PressHoldRelease</kbd>), that separate listener should call <a href="#_OpenContextMenu">this._OpenContextMenu()</a>,
   * not this method (<code class="prettyprint">_NotifyContextMenuGesture()</code>), and not [menu.open()]{@link oj.ojMenu#open}.  </li>
   * </ul>
   *
   * <p>Components needing to do per-launch setup like the above tasks should do so in an override of this method, <i>not</i> in
   * a [beforeOpen]{@link oj.ojMenu#event:beforeOpen} listener or an <a href="#_OpenContextMenu">_OpenContextMenu()</a> override.
   * This is discussed more fully <a href="#_OpenContextMenu">here</a>.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   *
   * @param {!Object} menu The JET Menu to open as a context menu.  Always non-<code class="prettyprint">null</code>.
   * @param {!Event} event What triggered the menu launch.  Always non-<code class="prettyprint">null</code>.
   * @param {string} eventType "mouse", "touch", or "keyboard".  Never <code class="prettyprint">null</code>.
   */
  _NotifyContextMenuGesture: function(menu, event, eventType)
  {
    this._OpenContextMenu(event, eventType);
  },

  /**
   * <p>The only correct way for a component to open its context menu is by calling this method, not by calling [Menu.open()]{@link oj.ojMenu#open} or
   * <a href="#_NotifyContextMenuGesture">_NotifyContextMenuGesture()</a>.  This method should be called in two cases:
   *
   * <ul>
   * <li>This method is called by <a href="#_NotifyContextMenuGesture">_NotifyContextMenuGesture()</a> and its overrides.  That method is
   * called when the baseComponent detects the default context menu gestures: right-click, <kbd>Press & Hold</kbd>, and <kbd>Shift-F10</kbd>.</li>
   *
   * <li>Components with UX-approved support for alternate context menu gestures like <kbd>PressHoldRelease</kbd> should call this method directly
   * when those gestures are detected.</li>
   * </ul>
   *
   * <p>Components needing to customize how the context menu is launched, or do any per-launch setup, should do so in the caller of this method,
   * (which is one of the two callers listed above), often by customizing the params passed to this method
   * (<code class="prettyprint">_OpenContextMenu</code>) per the guidance below.  This setup should <i>not</i> be done in the following ways:
   *
   * <ul>
   * <li>Components should not perform setup in a [beforeOpen]{@link oj.ojMenu#event:beforeOpen} listener, as this can cause a race
   * condition where behavior depends on who got their listener registered first: the component or the app.  The only correct component use
   * of a <code class="prettyprint">beforeOpen</code> listener is when there's a need to detect whether <i>something else</i> launched the menu.</li>
   *
   * <li>Components should not override this method (<code class="prettyprint">_OpenContextMenu</code>), as this method is final.  Instead, customize
   * the params that are passed to it.</li>
   * </ul>
   *
   * <p><b>Guidance on setting OpenOptions fields:</b>
   *
   * <p><b>Launcher:</b>
   *
   * <p>Depending on individual component needs, any focusable element within the component can be the appropriate
   * [launcher]{@link oj.ojMenu#openOptions.launcher} for this launch.
   *
   * <p>Browser focus returns to the launcher on menu dismissal, so the launcher must at least be focusable.  Typically a tabbable (not just
   * focusable) element is safer, since it just focuses something the user could have focused on their own.
   *
   * <p>By default (i.e. if <code class="prettyprint">openOptions</code> is not passed, or if it lacks a <code class="prettyprint">launcher</code>
   * field), the component init node is used as the launcher for this launch.  If that is not focusable or is suboptimal for a given
   * component, that component should pass something else.  E.g. components with a "roving tabstop" (like Toolbar) should typically choose the
   * current tabstop as their launcher.
   *
   * <p>The [:focusable]{@link http://api.jqueryui.com/focusable-selector/} and [:tabbable]{@link http://api.jqueryui.com/tabbable-selector/} selectors
   * may come in handy for choosing a launcher, e.g. something like <code class="prettyprint">this.widget().find(".my-class:tabbable").first()</code>.
   *
   * <p><b>Position:</b>
   *
   * <p>By default, this method applies [positioning]{@link oj.ojMenu#openOptions.position} that differs from Menu's default in the following ways:
   * (The specific settings are subject to change.)
   *
   * <ul>
   * <li>For mouse and touch events, the menu is positioned relative to the event, not the launcher.</li>
   *
   * <li>For touch events, <code class="prettyprint">"my"</code> is set to <code class="prettyprint">"start>40 center"</code>,
   * to avoid having the context menu obscured by the user's finger.</li>
   * </ul>
   *
   * <p>Usually, if <code class="prettyprint">position</code> needs to be customized at all, the only thing that needs changing is its
   * <code class="prettyprint">"of"</code> field, and only for keyboard launches (since mouse/touch launches should almost certainly keep
   * the default <code class="prettyprint">"event"</code> positioning).  This situation arises anytime the element relative to which the menu
   * should be positioned for keyboard launches is different than the <code class="prettyprint">launcher</code> element (the element to which
   * focus should be returned upon dismissal).  For this case, <code class="prettyprint">{ "position": {"of": eventType==="keyboard" ? someElement : "event"} }</code>
   * can be passed as the <code class="prettyprint">openOptions</code> param.
   *
   * <p>Be careful not to clobber useful defaults by specifying too much.  E.g. if you only want to customize <code class="prettyprint">"of"</code>,
   * don't pass other fields like <code class="prettyprint">"my"</code>, since your value will be used for all modalities (mouse, touch, keyboard),
   * replacing the modality-specific defaults that are usually correct.  Likewise, don't forget the
   * <code class="prettyprint">eventType==="keyboard"</code> check if you only want to customize <code class="prettyprint">"of"</code> for keyboard launches.
   *
   * <p><b>InitialFocus:</b>
   *
   * <p>This method forces [initialFocus]{@link oj.ojMenu#openOptions.initialFocus} to <code class="prettyprint">"menu"</code> for this
   * launch, so the caller needn't specify it.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @final
   *
   * @param {!Event} event What triggered the context menu launch.  Must be non-<code class="prettyprint">null</code>.
   * @param {string} eventType "mouse", "touch", or "keyboard".  Must be non-<code class="prettyprint">null</code>.  Passed explicitly since caller
   *        knows what it's listening for, and since events like <code class="prettyprint">contextmenu</code> and
   *        <code class="prettyprint">click</code> can be generated by various input modalities, making it potentially error-prone for
   *        this method to determine how they were generated.
   * @param {Object=} openOptions Options to merge with this method's defaults, which are discussed above.  The result will be passed to
   *        [Menu.open()]{@link oj.ojMenu#open}.  May be <code class="prettyprint">null</code> or omitted.  See also the
   *        <code class="prettyprint">shallow</code> param.
   * @param {Object=} submenuOpenOptions Options to be passed through to [Menu.open()]{@link oj.ojMenu#open}.  May be <code class="prettyprint">null</code>
   *        or omitted.
   * @param {boolean=} shallow Whether to perform a deep or shallow merge of <code class="prettyprint">openOptions</code> with this method's default
   *        value.  The default and most commonly correct / useful value is <code class="prettyprint">false</code>.
   *
   *        <ul>
   *        <li>If <code class="prettyprint">true</code>, a shallow merge is performed, meaning that the caller's <code class="prettyprint">position</code>
   *        object, if passed, will completely replace this method's default <code class="prettyprint">position</code> object.</li>
   *
   *        <li>If <code class="prettyprint">false</code> or omitted, a deep merge is performed.  For example, if the caller wishes to tweak
   *        <code class="prettyprint">position.of</code> while keeping this method's defaults for <code class="prettyprint">position.my</code>,
   *        <code class="prettyprint">position.at</code>, etc., it can pass <code class="prettyprint">{"of": anOfValue}</code> as the
   *        <code class="prettyprint">position</code> value.</li>
   *        </ul>
   *
   *        <p>The <code class="prettyprint">shallow</code> param is n/a for <code class="prettyprint">submenuOpenOptions</code>, since this method doesn't
   *        apply any defaults to that.  (It's a direct pass-through.)
   */
  _OpenContextMenu: function(event, eventType, openOptions, submenuOpenOptions, shallow)
  {
    var contextMenuNode = this._GetContextMenu();
    if (!contextMenuNode)
    {
      // the context menu node could have been currently opened with another launcher
      if (this._contextMenuGestureInit && $(this._contextMenuGestureInit).is(":visible"))
      {
        contextMenuNode = this._contextMenuGestureInit;
      }
      else
      {
        contextMenuNode = this._GetDefaultContextMenu();
      }
    }

    if (contextMenuNode)
    {
      // Note: our touch positioning is similar to that of the iOS touch callout (bubble with "Open in New Tab", etc.), which is offset from the pressHold location as follows:
      // - to the right, vertically centered.  (by default)
      // - to the left, vertically centered.  (if fits better)
      // - above or below, horizontally centered.  (if fits better)
      // An offset like 40 prevents it from opening right under your finger, and is similar to iOS's offset.  It also prevents the issue (on iOS7 at least)
      // where touchend after the pressHold can dismiss the CM b/c the menu gets the touchend.

      var position = {
        "mouse": {
          "my": "start top",
          "at": "start bottom",
          "of": event,
          "collision": "flipfit"
        },
        "touch": {
          "my": "start>40 center",
          "at": "start bottom",
          "of": event,
          "collision": "flipfit"
        },
        "keyboard": {
          "my": "start top",
          "at": "start bottom",
          "of": "launcher",
          "collision": "flipfit"
        }
      };

      var defaults = {"launcher": this.element, "position": position[eventType]}; // used for fields caller omitted
      var forcedOptions = {"initialFocus": "menu"};

      openOptions = (shallow)
        ? $.extend(defaults, openOptions, forcedOptions)
        : $.extend(true, defaults, openOptions, forcedOptions);

      contextMenuNode.__openingContextMenu = true; // Hack.  See todo on this ivar in Menu.open().
      if (contextMenuNode.tagName === "OJ-MENU")
      {
        contextMenuNode.open(event, openOptions, submenuOpenOptions);
      }
      else
      {
        var constructor = oj.Components.__GetWidgetConstructor(contextMenuNode, "ojMenu");
        var menu = constructor && constructor("instance");
        menu.open(event, openOptions, submenuOpenOptions);
      }
      contextMenuNode.__openingContextMenu = false;
    }
  },

  /**
   * Retrieve the default context menu.
   * @return {Element|null} the root element for the default context menu, or null if there is no default context menu.  The default is null.
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _GetDefaultContextMenu: function()
  {
    return null;
  },

  /**
   * <p>Removes the <code class="prettyprint">oj-hover</code>, <code class="prettyprint">oj-focus</code>, 
   * <code class="prettyprint">oj-focus-highlight</code>, and 
   * <code class="prettyprint">oj-active</code> classes from the specified element and its subtree.
   *
   * @memberof oj.baseComponent
   * @instance
   * @private
   *
   * @param {!jQuery} element The element from whose subtree the 3 classes should be removed
   */
  _removeStateClasses: function( element )
  {
    element.removeClass("oj-hover oj-focus oj-focus-highlight oj-active");
    element.find( ".oj-hover" ).removeClass( "oj-hover" );
    element.find( ".oj-focus" ).removeClass( "oj-focus" );
    element.find( ".oj-focus-highlight" ).removeClass( "oj-focus-highlight" );
    element.find( ".oj-active").removeClass( "oj-active" );
  },


  /**
   * @private
   * @return {boolean} true if there is no touch detected within the last 500 ms
   */
  _isRealMouseEvent: function(event)
  {
    return ! oj.DomUtils.recentTouchEnd();
  },

  /**
   * Add mouse listners to toggle oj-hover class
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected

   * @param {(!Object|!jQuery)} options This param can either be the element 
   * (convenience syntax for callers needing to 
   *   specify only the element(s) that would 
   *   otherwise have been passed as <code class="prettyprint">options.element</code>) 
   *   or an object supporting the following fields:
   * @param {jQuery} options.element The element(s) to receive the <code class="prettyprint">oj-hover</code> class on hover
   *   Required if <code class="prettyprint">afterToggle</code> is specified.
   * @param {?function(string)} options.afterToggle Optional callback function called each time the hover classes have been toggled, 
   *   after the toggle.  The string "mouseenter" or "mouseleave" is passed, indicating whether the classes were added or removed. 
   *   Components with consistency requirements, such as "<code class="prettyprint">oj-default</code> must be applied iff no state classes
   *   such as <code class="prettyprint">oj-hover</code> are applied," can enforce those rules in this callback.
   * @see #_RemoveHoverable
   */
  _AddHoverable: function(options)
  {
    var element;

    if ($.isPlainObject(options)) {
      element = options['element'];
    }
    else {
      element = options;
      options = {};
    }

    var afterToggle = options['afterToggle'] || $.noop;
    var markerClass = "oj-hover";

    element.on("mouseenter" + this.hoverableEventNamespace, 
               this._hoverStartHandler.bind(this, afterToggle))
           .on("mouseleave" + this.hoverableEventNamespace, 
               this._hoverAndActiveEndHandler.bind(this, markerClass, afterToggle));
  },

  /**
   * Remove mouse listners that were registered in _AddHoverable
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @param {!jQuery} element The same element passed to _AddHoverable
   * @see #_AddHoverable
   */
  _RemoveHoverable: function(element)
  {
    if (element)
    {
      element.off(this.hoverableEventNamespace);
    }
  },

  /**
   * Add touch and mouse listeners to toggle oj-active class
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected

   * @param {(!Object|!jQuery)} options This parameter can either be the element 
   * (convenience syntax for callers needing to 
   *   specify only the element(s) that would 
   *   otherwise have been passed as <code class="prettyprint">options.element</code>) 
   *   or an object supporting the following fields:
   * @param {jQuery} options.element The element(s) to receive the 
   * <code class="prettyprint">oj-active</code> class on active
   *   Required if <code class="prettyprint">afterToggle</code> is specified.
   * @param {?function(string)} options.afterToggle Optional callback function called each time 
   *   the active classes have been toggled, after the toggle.  The event.type string is passed 
   *   and indicates whether the classes were added or removed. The active classes are added on
   *   "touchstart" or "mousedown" or "mouseenter" and the active classes are removed on 
   *   "touchend" or "touchcancel" or "mouseup" or "mouseleave".
   *   Components with consistency requirements, such as 
   *   "<code class="prettyprint">oj-default</code> must be applied iff no state classes
   *   such as <code class="prettyprint">oj-active</code> are applied," 
   *   can enforce those rules in this callback.
   * @see #_RemoveActiveable
   */
  _AddActiveable: function(options)
  {
    var element;

    if ($.isPlainObject(options)) {
      element = options['element'];
    }
    else {
      element = options;
      options = {};
    }

    var afterToggle = options['afterToggle'] || $.noop;
    var markerClass = "oj-active";

    if (oj.DomUtils.isTouchSupported())
    {
      element.on("touchstart" + this.activeableEventNamespace,
                 this._activeStartHandler.bind(this, afterToggle))
             .on("touchend" + this.activeableEventNamespace + " " + 
                 "touchcancel" + this.activeableEventNamespace, 
                 this._hoverAndActiveEndHandler.bind(this, markerClass, afterToggle));
    }

    element.on("mousedown" + this.activeableEventNamespace, 
                 this._activeStartHandler.bind(this, afterToggle))
           .on("mouseup" + this.activeableEventNamespace, 
               this._hoverAndActiveEndHandler.bind(this, markerClass, afterToggle))
           // mouseenter/mouseleave is for the case where you mousedown, then move mouse
           // out of element, then move mouse back. We want oj-active to disappear when you move 
           // outside and reappear when you move back.
           .on("mouseenter" + this.activeableEventNamespace, 
                 this._activeStartHandler.bind(this, afterToggle))
           .on("mouseleave" + this.activeableEventNamespace, 
               this._hoverAndActiveEndHandler.bind(this, markerClass, afterToggle));

  },

  /**
   * Remove touch and mouse listeners that were registered in _AddActiveable
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @param {!jQuery} element The same element passed to _AddActiveable
   * @see #_AddActiveable
   */
  _RemoveActiveable: function(element)
  {
    if (element)
    {
      element.off(this.activeableEventNamespace);
      _lastActiveElement = null;
    }
  },

  /**
   * Add oj-active style class and call the afterToggleFunction.
   * Set _lastActiveElement to event.currentTarget on mousedown. Clear it on mouseup.
   * _lastActiveElement is used to remove and add back oj-active on mouseleave and mouseenter
   * if the mouse stays down.
   * @private
   */
  _activeStartHandler: function (afterToggleFunction, event)
  {

    var elem = $(event.currentTarget);
   
    // do nothing on mouseenter if _lastActiveElement ivar is not the currentTarget or child of
    // currentTarget. Checking for the child is needed in case there are two nested 
    // dom nodes that have set_AddActiveable. Since events bubble, the _lastActiveElement will
    // be the outermost dom that got the mousedown event.
    if (event.type === "mouseenter" && !this._isTargetInActiveElement(event.currentTarget))
    {
      return;
    }
    
    // do this for either touchstart or real mouse events, but not mouse compatibility event
    if (! elem.hasClass("oj-disabled") && 
     (event.type === "touchstart" || this._isRealMouseEvent(event)))
    {
      elem.addClass("oj-active");
      afterToggleFunction(event.type); 
       
      // If we get mousedown on the element, we want oj-active to be removed on mouseleave
      // and added back on mouseenter if the mouse stays down.
      if (event.type === "mousedown")
      {
        _lastActiveElement = event.currentTarget;
                       
        this.document.one( "mouseup", function() {
          _lastActiveElement = null;
        });

      }
    }
  },

  /**
   * @private
   */
  _hoverStartHandler: function (afterToggleFunction, event)
  {
    // do this for real mouseenter, but not mouse compatibility event
    var elem = $(event.currentTarget);
    if (! elem.hasClass("oj-disabled") && this._isRealMouseEvent(event))
    {     
      elem.addClass("oj-hover");
      afterToggleFunction(event.type); 
    }
  },

  /**
   * Remove markerClass and call the afterToggleFunction.
   * @private
   */
  _hoverAndActiveEndHandler: function (markerClass, afterToggleFunction, event)
  {
    // for oj-active we don't care about mouseleave unless it was triggered when _lastActiveElement
    // was set on the currentTarget. (see _activeStartHandler). If that's not the case, return.
    if (markerClass === "oj-active" && event.type === "mouseleave" &&  
        !this._isTargetInActiveElement(event.currentTarget))
    {
      return;
    }
    $(event.currentTarget).removeClass(markerClass);
    afterToggleFunction(event.type);
  },
  /**
   * Returns true if the event target is _lastActiveElement or a child of _lastActiveElement.
   * Checking for the child is needed in case there are two nested dom nodes that have set
   * _AddActiveable. For example, inputDateTime -> the trigger root container and the trigger icon.
   * We are only keeping track of the _lastActiveElement, which means the ancestor element will
   * be stored in _lastActiveElement because the mouseleave event bubbles.
   * @private
   */
  _isTargetInActiveElement: function (currentTarget)
  {
    return (_lastActiveElement === currentTarget || 
      (_lastActiveElement != null && $.contains(_lastActiveElement, currentTarget)));
  }, 

   // We no longer use _hoverable, but should still override it to ensure the JQUI impl is not called.
  _hoverable: function() {},

  // The internal JSDoc of the DomUtils version of this API refers to this doc, so if changes are made here, that doc 
  // must be updated as needed.
  /**
   * <p>Sets JET's "focus" CSS classes when the element is focused and removes them when focus is lost.
   *
   * <p>The <code class="prettyprint">oj-focus</code> class is set on all focuses.
   * 
   * <p>Some components additionally have an <code class="prettyprint">oj-focus-highlight</code> class, which applies a focus 
   * indicator that is appropriate on a subset of the occasions that <code class="prettyprint">oj-focus</code> is appropriate.  
   * Those components should pass <code class="prettyprint">true</code> for the <code class="prettyprint">applyHighlight</code> 
   * param, in which case the <code class="prettyprint">oj-focus-highlight</code> class is set if appropriate given the 
   * current focus highlight policy. 
   * 
   * <h5>Focus highlight policy</h5>
   * 
   * <p>The focus highlight policy supports the 3 values listed below.  By default, it is retrieved from the 
   * <code class="prettyprint">$focusHighlightPolicy</code> SASS variable, shared by many components and patterns.  Components 
   * with different needs, including those exposing a component-specific SASS variable or other API for this, should see the 
   * <code class="prettyprint">getFocusHighlightPolicy</code> parameter below.
   * 
   * Valid focus highlight policies:
   * 
   * <table class="generic-table">
   *   <thead>
   *     <tr>
   *       <th>Policy</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>"nonPointer"</td>
   *       <td>Indicates that the component should apply the <code class="prettyprint">oj-focus-highlight</code> 
   *           class only for focuses not resulting from pointer (touch or mouse) interaction.  (In the built-in themes, the 
   *           SASS variable defaults to this value.)</td>
   *     </tr>
   *     <tr>
   *       <td>"all"</td>
   *       <td>Indicates that the component should apply the class for all focuses.</td>
   *     </tr>
   *     <tr>
   *       <td>"none"</td>
   *       <td>Indicates that the component should never apply the class, because the application has taken responsibility
   *           for applying the class when needed for accessibility.</td>
   *     </tr>
   *   </tbody>
   * </table>
   * 
   * <h5>Toggling the classes</h5>
   * 
   * <p>Components that toggle these focus classes outside of this API must maintain the invariant that 
   * <code class="prettyprint">oj-focus-highlight</code> is applied to a given element in a (not necessarily strict) subset 
   * of cases that <code class="prettyprint">oj-focus</code> is applied to that element.
   * 
   * <p>Typically the specified element should be within the component subtree, in which case the classes will 
   * automatically be removed from the element when the component is destroyed, when its <code class="prettyprint">disabled</code>
   * option is set to true, and when <code class="prettyprint">_NotifyDetached()</code> is called.
   * 
   * <p>As a minor exception, for components that wrap themselves in a new root node at create time, if the specified 
   * element is within the root node's subtree but not within the init node's subtree, then at destroy time only, the 
   * classes will not be removed, since <code class="prettyprint">destroy()</code> is expected to remove such nodes.
   * 
   * <p>If the element is NOT in the component subtree, then the caller is responsible for removing the classes at the 
   * times listed above.
   * 
   * <h5>Listeners</h5>
   * 
   * <p>If <code class="prettyprint">setupHandlers</code> is not passed, or if <code class="prettyprint">setupHandlers</code> 
   * is passed and uses <code class="prettyprint">_on</code> to register its listeners as seen in the example, then 
   * the listeners are not invoked when the component is disabled, and the listeners are automatically cleaned up when the 
   * component is destroyed.  Otherwise, the caller is responsible for ensuring that the disabled state is handled correctly, 
   * and removing the listeners at destroy time.
   * 
   * <h5>Related API's</h5>
   * 
   * <p>Non-component internal callers should see oj.DomUtils.makeFocusable().  Per its JSDoc (unpublished; see the source), it 
   * has a couple of additional usage considerations.
   * 
   * @memberof oj.baseComponent
   * @instance
   * @protected
   *
   * @param {(!Object|!jQuery)} options This param can either be the element (convenience syntax for callers needing to 
   *   specify only the element(s) that would otherwise have been passed as <code class="prettyprint">options.element</code>) 
   *   or an object supporting the following fields:
   * @param {jQuery} options.element The element(s) to receive the <code class="prettyprint">oj-focus</code> classes on focus.
   *   Required if <code class="prettyprint">setupHandlers</code> not passed; ignored otherwise.
   * @param {boolean} options.applyHighlight <code class="prettyprint">true</code> if the <code class="prettyprint">oj-focus-highlight</code> 
   *   class should be applied when appropriate.  <code class="prettyprint">false</code> or omitted if that class should never be applied.
   * @param {?function(string)} options.afterToggle Optional callback function called each time the focus classes have been toggled, 
   *   after the toggle.  The 
   *   string "focusin" or "focusout" is passed, indicating whether the classes were added or removed.  Components 
   *   with consistency requirements, such as "<code class="prettyprint">oj-default</code> must be applied iff no state classes such 
   *   as <code class="prettyprint">oj-focus</code> are applied," can enforce those rules in this callback.
   * @param {?function()} options.getFocusHighlightPolicy Optional if <code class="prettyprint">applyHighlight</code> is 
   *   <code class="prettyprint">true</code>; ignored otherwise.  Components with a component-specific focus policy 
   *   mechanism should pass a function that always returns one of the three valid values listed above, keeping in mind 
   *   that this method can be called on every focus.  See the example.
   * @param {?function()} options.recentPointer Relevant iff <code class="prettyprint">applyHighlight</code> is 
   *   <code class="prettyprint">true</code> and the focus highlight policy is <code class="prettyprint">"nonPointer"</code>;
   *   ignored otherwise.  Recent pointer activity is considered to have occurred if (a) a mouse button or finger has
   *   recently been down or up, or (b) this optional callback function returns true.  Components wishing to additionally take into 
   *   account (say) recent pointer <i>movements</i> can supply a function returning true if those movements have been detected, 
   *   keeping in mind that this method can be called on every focus.  See the example.
   * @param {?function(function(!jQuery),function(!jQuery))} options.setupHandlers Can be omitted by components whose focus 
   *   classes need to be added and removed on focusin and focusout, respectively.  Components needing to add/remove those
   *   classes in response to other events should specify this parameter, which is called once, immediately.  See the examples.
   * 
   * @example <caption>Opt into the highlight behavior, and specify a function to be called every time the classes are toggled:</caption>
   * var self = this;
   * this._focusable({
   *     'element': this.element, 
   *     'applyHighlight': true, 
   *     'afterToggle' : function() {
   *         self._toggleDefaultClasses();
   *     }
   * });
   * 
   * @example <caption>Arrange for mouse movement to be considered <u>in addition to</u> mouse/finger up/down.
   *   Also supply a component-specific focusHighlightPolicy:</caption>
   * var self = this;
   * this._focusable({
   *     'element': someElement, 
   *     'applyHighlight': true, 
   *     'recentPointer' : function() {
   *         // A timestamp-based approach avoids the risk of getting stuck in an inaccessible 
   *         // state if (say) mouseenter is not followed by mouseleave for some reason.
   *         var millisSincePointerMove = Date.now() - _myPointerMoveTimestamp;
   *         var isRecent = millisSincePointerMove < myThreshold;
   *         return isRecent;
   *     },
   *     'getFocusHighlightPolicy' : function() {
   *         // Return the value of a component-specific SASS $variable, component option, or other 
   *         // component-specific mechanism, either "all", "none", or "nonPointer".  SASS variables
   *         // should be pulled into JS once statically on load, not per-instance or per-focus.
   *     }
   * });
   * 
   * @example <caption>Add/remove the focus classes in response to events other than focusin/focusout:</caption>
   * var self = this;
   * this._focusable({
   *     'applyHighlight': myBooleanValue, 
   *     'setupHandlers': function( focusInHandler, focusOutHandler) {
   *         self._on( self.element, {
   *             // This example uses focus/blur listeners, which don't bubble, rather than the 
   *             // default focusin/focusout (which bubble).  This is useful when one focusable  
   *             // element is a descendant of another.
   *             focus: function( event ) {
   *                 focusInHandler($( event.currentTarget ));
   *             },
   *             blur: function( event ) {
   *                 focusOutHandler($( event.currentTarget ));
   *             }
   *         });
   *     }
   * });
   * 
   * @example <caption>Alternate usage of <code class="prettyprint">setupHandlers</code>, which simply stashes the 
   *   handlers so they can be called from the component's existing handlers:</caption>
   * var self = this;
   * this._focusable({
   *     'applyHighlight': myBooleanValue, 
   *     'setupHandlers': function( focusInHandler, focusOutHandler) {
   *         self._focusInHandler = focusInHandler;
   *         self._focusOutHandler = focusOutHandler;
   *     }
   * });
   */
  _focusable: function( options )
  {
    if (!$.isPlainObject(options)) {
        options = {'element': options};
    }
    
    options['component'] = this;
    oj.DomUtils.makeFocusable(options);
  },

  /**
   * Remove all listener references that were attached to the element.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _UnregisterChildNode: function(element)
  {
    if (element) {
      $(element).off(this.eventNamespace);
      var bindings = this['bindings'];
      if (bindings) {
        this['bindings'] = $(bindings.not(element));
      }
    }
  },

  /**
   * <p>Determines whether the component is LTR or RTL.
   *
   * <p>Component responsibilities:
   *
   * <ul>
   * <li>All components must determine directionality exclusively by calling this protected superclass method.
   *     (So that any future updates to the logic can be made in this one place.)</li>
   * <li>Components that need to know the directionality must call this method at create-time
   *     and from <code class="prettyprint">refresh()</code>, and cache the value.
   * <li>Components should not call this at other times, and should instead use the cached value.  (This avoids constant DOM
   *     queries, and avoids any future issues with component reparenting (i.e. popups) if support for directional islands is added.)</li>
   * </ul>
   *
   * <p>App responsibilities:
   *
   * <ul>
   * <li>The app specifies directionality by setting the HTML <code class="prettyprint">"dir"</code> attribute on the
   *     <code class="prettyprint">&lt;html></code> node.  When omitted, the default is <code class="prettyprint">"ltr"</code>.
   *     (Per-component directionality / directional islands are not currently supported due to inadequate CSS support.)</li>
   * <li>As with any DOM change, the app must <code class="prettyprint">refresh()</code> the component if the directionality changes dynamically.
   *   (This provides a hook for component housekeeping, and allows caching.)</li>
   * </ul>
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @return {string} the reading direction, either <code class="prettyprint">"ltr"</code> or <code class="prettyprint">"rtl"</code>
   * @default <code class="prettyprint">"ltr"</code>
   */
  _GetReadingDirection: function( )
  {
    return oj.DomUtils.getReadingDirection();
  },

  /**
   * <p>Notifies the component that its subtree has been connected to the document programmatically after the component has
   * been created.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _NotifyAttached: function()
  {
    this._propertyContext = null;
  },

  /**
   * <p>Notifies the component that its subtree has been removed from the document programmatically after the component has
   * been created.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _NotifyDetached: function()
  {
    this._propertyContext = null;
    this._removeStateClasses( this.widget() );
  },

  /**
   * <p>Notifies the component that its subtree is initially visible after the component has
   * been created.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _NotifyInitShown: function()
  {

  },
  
    /**
   * <p>Notifies the component that its subtree has been made visible programmatically after the component has
   * been created.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _NotifyShown: function()
  {

  },

  /**
   * <p>Notifies the component that its subtree has been made hidden programmatically after the component has
   * been created.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _NotifyHidden: function()
  {

  },

  /**
   * <p>Determines whether this component is effectively disabled, i.e. it has its 'disabled' attribute set to true
   * or it has been disabled by its ancestor component.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   * @return {boolean} true if the component has been effectively disabled, false otherwise
   */
  _IsEffectivelyDisabled: function()
  {
    return (this.options['disabled'] || this._ancestorDisabled) ? true : false;
  },

  /**
   * <p>Sets the ancestor-provided disabled state on this component.
   *
   * @memberof oj.baseComponent
   * @instance
   * @private
   * @param {boolean} disabled - true if this component is being disabled by its ancestor component, false otherwise
   */
  __setAncestorComponentDisabled: function(disabled)
  {
    this._ancestorDisabled = disabled;
  },


  /**
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _getTranslationSectionLoader: function()
  {
    var sectionNames = [];

    var self = this;

    var index = 0;

    this._traverseWidgetHierarchy(
      function(proto)
      {
        // retrive translation section name for the widget and all of its ancestors

        // Since _GetTranslationsSectionName() is a protected method, we can only call it on the widget instance.
        // For superclases, we will assume that their section names can only be their full widget name

        var name = (index == 0) ? self._GetTranslationsSectionName() : proto.widgetFullName;
        index++;

        var section = oj.Translations.getComponentTranslations(name);

        if (section != null && !$.isEmptyObject(section))
        {
          sectionNames.push(name);
        }
      }
    );

    var count = sectionNames.length;

    if (count > 0)
    {
      return function()
            {
              // Optimize for the most common case where superclasses do not define translations
              if (count == 1)
              {
                return oj.Translations.getComponentTranslations(sectionNames[0]);
              }
              else
              {
                var trs = {};

                for (var i = count-1; i>=0; i--)
                {
                  $.widget.extend(trs, oj.Translations.getComponentTranslations(sectionNames[i]));
                }

                return trs;
              }
            };
    }
    return null;
  },

  /**
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _getDynamicPropertyContext: function()
  {
    if (!this._propertyContext)
    {
      var c = {};
      this._propertyContext = c;
      var element = this.element[0];
      c['containers'] = _getSpecialContainerNames(element);
      c['element'] = element;
      c['isCustomElement'] = this._IsCustomElement();
    }
    return this._propertyContext;
  },

  /**
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _setupDefaultOptions: function(originalDefaults, constructorOptions)
  {
    var options = this.options;

    // Load component translations
    var translationLoader = this._getTranslationSectionLoader();

    var currVal = constructorOptions[_OJ_TRANSLATIONS_OPTION];

    if (translationLoader != null && (currVal === undefined || $.isPlainObject(currVal) ))
    {
      _defineDynamicProperty(this, undefined, constructorOptions[_OJ_TRANSLATIONS_OPTION],
                             options,  _OJ_TRANSLATIONS_OPTION, translationLoader);
    }


    // Load options specified with oj.Components.setDefaultOptions()
    this._loadGlobalDefaultOptions(originalDefaults, constructorOptions)

  },

  /**
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _loadGlobalDefaultOptions: function(originalDefaults, constructorOptions)
  {
    var options = this.options;

    var widgetHierNames = [];

    // walk up the widget hierarchy
    this._traverseWidgetHierarchy(
              function(proto)
              {
                 widgetHierNames.push(proto.widgetName);
              }
    );

    var allProperties = oj.Components.getDefaultOptions();
    widgetHierNames.push('default');


    // get properties applicable to this component
    var defaults = oj.Components.__getDefaultOptions(widgetHierNames);

    if ($.isEmptyObject(defaults))
    {
      return;
    }

    var self = this;

    var contextCallback = function()
    {
      return self._getDynamicPropertyContext();
    };


    for (var prop in defaults)
    {
      var val = constructorOptions[prop];

      if (val === undefined || $.isPlainObject(val))
      {
        var defaultValueList = defaults[prop];
        if (defaultValueList)
        {
          var callback = _getCompoundDynamicGetter(defaultValueList);
          if(callback)
          {
            _defineDynamicProperty(this, originalDefaults[prop], val, options, prop, callback, contextCallback);
          }
          else
          {
            var list = [originalDefaults[prop]].concat(defaultValueList);
            list.push(val);
            options[prop] = _mergeOptionLayers(list);
          }
        }
      }
    }
  },

  /**
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _traverseWidgetHierarchy: function(callback)
  {
    var proto = this.constructor.prototype;
    while (proto != null && 'oj' === proto['namespace'])
    {
      callback(proto);
      proto = Object.getPrototypeOf(proto);
    }
  },

  /**
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _getRootElement: function() 
  {
    return this.OuterWrapper || this.element[0];
  },

  /**
   * Determines whether the component is being rendered as a custom element.
   * @return {boolean} True if the component is being rendered as a custom element
   * 
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _IsCustomElement: function()
  {
    return oj.BaseCustomElementBridge.getRegistered(this._getRootElement().tagName) != null;
  },

  /**
   * Prepares a custom renderer context object for either the JQuery or custom element syntax,
   * removing and exposing keys as needed.
   * @param {Object} context The renderer context object.
   * @return {Object} The cleaned up renderer context.
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _FixRendererContext: function(context)
  {
    if (this._IsCustomElement())
    {
      // Do a shallow copy to avoid setter/getters from being lost
      var contextCopy =  oj.CollectionUtils.copyInto({}, context);
      // remove component or widget constructor references and expose element reference instead
      delete contextCopy['component'];
      contextCopy['componentElement'] = this._getRootElement();
      return contextCopy;
    }
    return context;
  },

    /**
   * Returns a wrapper function for custom elements that converts an object
   * returned by a custom renderer into an old format supported by widgets 
   * @param {Function} origRenderer Renderer function called to create custom content
   * @return {Function} A wrapper function that will used to convert result into toolkit format
   * @protected
   * @memberof oj.baseComponent
   */
  _WrapCustomElementRenderer: function (origRenderer)
  {
    if (this._IsCustomElement() && typeof origRenderer === 'function')
    {
      var customRenderer = function (context)
      {
        var obj = origRenderer(context);
        return obj && obj['insert'] ? obj['insert'] : null;
      }
      return customRenderer;
    }
    return origRenderer;
  },

  /**
   * Stores a map of writeback options that we reference during option comparison.
   * Package private method called from the CustomElementBridge.
   * @param  {Object} options The writeback options map
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  __saveWritebackOptions: function(options)
  {
    this._writebackOptions = options;
  },

  /**
   * Returns true if an option should be written back.
   * @param  {string} option The option to lookup
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  _getWritebackOption: function(option)
  {
    if (this._writebackOptions && this._writebackOptions[option])
    {
      return true;
    }
    return false;
  },

  /**
   * Called by oj.Components.subtreeAttached and will only call _NotifyAttached
   * for non custom elements. Custom elements are notified when they are 
   * attached from the DOM so oj.Components.subtreeAttached is unnecessary.
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  __handleSubtreeAttached: function() {
    if (!this._IsCustomElement())
      this._NotifyAttached();
  },

  /**
   * Called by oj.Components.subtreeAttached and will only call _NotifyDetached
   * for non custom elements. Custom elements are notified when they are 
   * detached from the DOM so oj.Components.subtreeDetached is unnecessary.
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  __handleSubtreeDetached: function() {
    if (!this._IsCustomElement())
      this._NotifyDetached();
  },

  /**
   * Called by the CustomElementBridge when the custom element is attached
   * to the DOM.
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  __handleConnected: function() {
    this._NotifyAttached();
    this._SetupResources();
  },

  /**
   * Called by the CustomElementBridge when the custom element is detached
   * from the DOM.
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  __handleDisconnected: function() {
    this._ReleaseResources();
    this._NotifyDetached();
  },

  /**
   * Method called by the CustomElementBridge to get the element to call focus on for this custom element
   * which can be the root custom element or an HTML element like an input or select.
   * @return {Element}
   * @memberof oj.baseComponent
   * @instance
   * @private
   */
  __getFocusElement: function()
  {
    return this.GetFocusElement();
  },

  /**
   * Returns the current focusable element for this component which can be the root custom element 
   * or an HTML element like an input or select.
   * @return {Element}
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  GetFocusElement: function()
  {
    return this.element[0];
  },
  
  /**
   * <p>The following CSS classes can be applied by the page author as needed.
   * 
   * @ojfragment ojStylingDocIntro - For use in the Styling section of the JSDoc, above the table of CSS classes .
   * @memberof oj.baseComponent
   */

  /**
   * Class
   * 
   * @ojfragment ojStylingDocClassHeader - For use in the "Class" <th> in the Styling section of the JSDoc.
   * @memberof oj.baseComponent
   */

  /**
   * Can be applied to
   * 
   * @ojfragment ojStylingDocApplyHeader - For use in the "Can be applied to" <th> in the Styling section of the JSDoc.
   * @memberof oj.baseComponent
   */

  /**
   * Description
   * 
   * @ojfragment ojStylingDocDescriptionHeader - For use in the "Description" <th> in the Styling section of the JSDoc.
   * @memberof oj.baseComponent
   */

  /**
   * Example
   * 
   * @ojfragment ojStylingDocExampleHeader - For use in the "Example" <th> in the Styling section of the JSDoc.
   * @memberof oj.baseComponent
   */

  /**
   * Under normal circumstances this class is applied automatically. It is documented here for the rare cases that an app 
   * developer needs per-instance control.
   * 
   * <p>The <code class="prettyprint">oj-focus-highlight</code> class applies focus styling that may not be desirable when 
   * the focus results from pointer interaction (touch or mouse), but which is needed for accessibility when the focus 
   * occurs by a non-pointer mechanism, for example keyboard or initial page load.
   * 
   * <p>The application-level behavior for this component is controlled in the theme by the 
   * <code class="prettyprint">$focusHighlightPolicy</code> SASS variable; however, note that this same variable controls 
   * the focus highlight policy of many components and patterns. The values for the variable are:
   * 
   * <ul>
   *   <li><code class="prettyprint">nonPointer</code>: <code class="prettyprint">oj-focus-highlight</code> is applied only 
   *       when focus is not the result of pointer interaction. Most themes default to this value.</li>
   *   <li><code class="prettyprint">all</code>: <code class="prettyprint">oj-focus-highlight</code> is applied regardless 
   *       of the focus mechanism.</li>
   *   <li><code class="prettyprint">none</code>: <code class="prettyprint">oj-focus-highlight</code> is never applied. This
   *       behavior is not accessible, and is intended for use when the application wishes to use its own event listener to 
   *       precisely control when the class is applied (see below). The application must ensure the accessibility of the result.</li>
   * </ul>
   * 
   * <p>To change the behavior on a per-instance basis, the application can set the SASS variable as desired and then use 
   * event listeners to toggle this class as needed.
   * 
   * @ojfragment ojFocusHighlightDoc - For use in the Styling table of components using the oj-focus-highlight class with the 
   *    $focusHighlightPolicy var.  Components using that class with a component-specific mechanism instead of that $var will need 
   *    different verbiage, which could be decomped to another baseComponent fragment if shareable by multiple components.
   * @memberof oj.baseComponent
   */
});

// Remove base component from the jQuery prototype, so it could not be created
// directly by page authors

delete $['fn'][_BASE_COMPONENT];

// -----------------------------------------------------------------------------
// "private static members" shared by all components
// -----------------------------------------------------------------------------

var _lastActiveElement;

}() ); // end of BaseComponent wrapper function


// -----------------------------------------------------------------------------
// End of baseComponent, start of other content
// -----------------------------------------------------------------------------

/**
 * <p>This method is our version of $.widget, i.e. the static initializer of a component such as ojButton.
 * It calls that method, plus does any other static init we need.
 *
 * TODO:
 * - Consider moving this method into its own file.
 * - For base param, make the type oj.baseComponent rather than Object, but need to declare that as a type first.  Review how that's done.
 *
 * @private
 * @param {string} name typically of the form "oj.ojMenu"
 * @param {Object} base NOT optional (unlike JQUI)
 * @param {Object} prototype
 * @param {boolean=} isHidden - if true, indicates that the component name should not
 * be available on jQuery prototype
 */
oj.__registerWidget = function(name, base, prototype, isHidden)
{
  $.widget( name, base, prototype );

  if (isHidden)
  {
    var globalName = name.split('.')[1];
    delete $['fn'][globalName];
  }

  // create single-OJ pseudo-selector for component, e.g. ":oj-menu", in addition to the ":oj-ojMenu" that $.widget() creates.
  // for private components it will begin with an underscore, e.g.,  ":_oj-radio"
  if (name.substring(0, 5) === "oj.oj" || name.substring(0, 6) === "oj._oj")
  {
    var nameArray = name.split( "." ); // ["oj", "ojMenu"], ["oj", "_ojRadio"]
    var namespace = nameArray[ 0 ];    // "oj"
    var simpleName = nameArray [ 1 ];  // "ojMenu", "_ojRadio"
    var fullName = namespace + "-" + simpleName; // "oj-ojMenu", "oj-_ojRadio"
    var isPrivate = simpleName.substring(0, 1) === "_";
    // if private, make the single-oj pseudo-selector start with an underscore, like this -> "_oj-radio"
    var modifiedFullName; // "oj-Menu", "_oj-Radio".  Lowercased below.
    if (isPrivate)
    {
      modifiedFullName = "_" + namespace + "-" + simpleName.substring(3);
    }
    else
    {
      modifiedFullName = namespace + "-" + simpleName.substring(2);
    }

    // Capitalization doesn't seem to matter with JQ pseudos, e.g. for the existing double-oj pseudo, both $(":oj-ojMenu") and $(":oj-ojmenu") work.
    // So, follow JQUI's pattern of using toLowerCase here, which will lowercase not only the "M' in "Menu", but also any camelcased chars after that.
    $.expr[ "pseudos" ][ modifiedFullName.toLowerCase() ] = function( elem ) {
      return !!$.data( elem, fullName );
    };
  }
};



/**
 * @param {Object} self
 * @param {Object|undefined} originalDefaultValue
 * @param {?Object} constructorValue
 * @param {!Object} options
 * @param {string} prop
 * @param {Function} getter
 * @param {Function=} contextCallback
 * @private
 */
 function _defineDynamicProperty(self, originalDefaultValue, constructorValue, options, prop, getter, contextCallback)
 {
   var override = constructorValue;
   var replaced = false;
   var overriddenSubkeys = {};

   delete options[prop];

   Object.defineProperty(options, prop,
     {
       'get': function()
              {
                // Once the option is replaced, we no longer merge in defaults
                if (replaced)
                {
                  return override;
                }

                if (self._settingNestedKey != null)
                {
                  // The getter is getting called from the option() method that will be mutating the current
                  // object. We need to return only the override portion in this case to avoid the defaults being
                  // reapplied as an override

                  return override;

                }

                var defaultVal = getter(contextCallback? contextCallback() : prop);

                return _mergeOptionLayers([originalDefaultValue, defaultVal, override], overriddenSubkeys)
              },
       'set': function(value)
              {
                override = value;

                if (self._settingNestedKey != null)
                {
                  overriddenSubkeys[self._settingNestedKey] = true;
                }
                else // The entire option has been replaced
                {
                  replaced = true;
                }
              },
       'enumerable' : true
     }
   );
 }
 
 /**
  * @ignore
  */
 function _getCompoundDynamicGetter(values)
 {
    if (values.length === 1)
    {
      var val = values[0];
      return (val instanceof __ojDynamicGetter) ? val.getCallback(): null;
    }
    
    var hasGetters = false;
    for (var i=0; i<values.length && !hasGetters; i++)
    {
      var value = values[i]
      if (value != null && value instanceof __ojDynamicGetter)
      {
        hasGetters = true;
      }
    }
    
    if (hasGetters)
    {
      var getter = function(context)
      {
        var resolvedVals = [];
        values.forEach(
          function(value)
          {
            if (value != null && value instanceof __ojDynamicGetter)
            {
              resolvedVals.push(value.getCallback()(context));
            }
            else
            {
              resolvedVals.push(value);
            }
          }
        );
        
        return _mergeOptionLayers(resolvedVals);
      };
      return getter;
    }
    
    return null;
    
 }

 /**
  * @private
  */
 function _getSpecialContainerNames(elem)
 {
    var containers = [];
    while (elem)
    {
      var ga =  elem.getAttribute;
      var name = ga ? ga.call(elem, oj.Components._OJ_CONTAINER_ATTR) : null;
      if (name != null)
      {
        containers.push(name);
      }
      elem = elem.parentNode;
    }

    return containers;
 };

 /**
  * @private
  */
 function _storeWidgetName(element, widgetName)
 {
   var data = element.data(_OJ_WIDGET_NAMES_DATA);
   if (!data)
   {
     data = [];
     element.data(_OJ_WIDGET_NAMES_DATA, data);
   }
   if (data.indexOf(widgetName) < 0)
   {
     data.push(widgetName);
   }
 }

 /**
  * @private
  */
 function _removeWidgetName(element, widgetName)
 {
   var data = element.data(_OJ_WIDGET_NAMES_DATA);
   if (data)
   {
     var index = data.indexOf(widgetName);
     if (index >= 0)
     {
       data.splice(index, 1);
       if (data.length === 0)
       {
         element.removeData(_OJ_WIDGET_NAMES_DATA);
       }
     }
   }
 }

 /**
  * @private
  * @param {Array} values - values to merge
  * @param {Object=} overriddenSubkeys subkeys where the merging should not occur, i.e.
  * the value from corresponsing subkey on the last element of values array should win
  */
 function _mergeOptionLayers(values, overriddenSubkeys)
 {
   var result = undefined;
   for (var i=0; i<values.length; i++)
   {
      var value = values[i];
      if (value !== undefined )
      {
        if ($.isPlainObject(value))
        {
          var input =  $.isPlainObject(result) ? [result, value] : [value];
          // The last object (overrides) is always fully merged in
          result = _mergeObjectsWithExclusions({}, input, (i == values.length - 1)? null: overriddenSubkeys, null);
        }
        else
        {
          result = value;
        }
      }
    }
    return result;
 }

 /**
 * @private
 */
function _mergeObjectsWithExclusions(target, input, ignoreSubkeys, basePath)
{
  var inputLength = input.length;

  for (var inputIndex = 0; inputIndex < inputLength; inputIndex++)
  {
    var source = input[inputIndex];
    var keys = Object.keys(source);
    for (var i=0; i<keys.length; i++)
    {
      var key = keys[i];
      var path = (ignoreSubkeys == null)? null:  (basePath == null ? key : basePath + '.' + key);
      // Ignore all sources when the current path is registered in ignoreSubkeys
      if (ignoreSubkeys == null || !ignoreSubkeys[path])
      {
        var value = source[key];
        if (value !== undefined)
        {
          if ($.isPlainObject(value))
          {
            var params = $.isPlainObject(target[key])? [target[key], value]: [value];
            target[key] = _mergeObjectsWithExclusions({}, params, ignoreSubkeys, path);
          }
          else
          {
            target[key] = value;
          }
        }
      }
    }
  }
  return target;
}


 /**
  * @private
  */
  var _OJ_TRANSLATIONS_OPTION = "translations";

  /**
   * @private
   */
  var _OJ_TRANSLATIONS_PREFIX = _OJ_TRANSLATIONS_OPTION + ".";


  /**
   * @private
   */
  function _returnTrue()
  {
    return true;
  };


  /**
   * @private
   */
  var _OJ_COMPONENT_EVENT_OVERRIDES =
  {
    isDefaultPrevented : function(){
        return false;
    },
    preventDefault : function ()
    {
      this.isDefaultPrevented = _returnTrue;
    },
    stopPropagation : function ()
    {
      this.isPropagationStopped = _returnTrue;
    },
    stopImmediatePropagation : function ()
    {
      this.isImmediatePropagationStopped = _returnTrue;
    }
  };

/**
 * Returns an object with context for the given child DOM node. This will always contain the subid for the node,
 * defined as the 'subId' property on the context object. Additional component specific information may also be included.
 *
 * For more details on returned objects, see <a href="#contextobjects-section">context objects</a>.
 *
 * @ojfragment nodeContextDoc
 * @memberof oj.baseComponent
 */

/**
 * The child DOM node
 *
 * @ojfragment nodeContextParam
 * @memberof oj.baseComponent
 */

/**
 * The context for the DOM node, or <code class="prettyprint">null</code> when none is found.
 *
 * @ojfragment nodeContextReturn
 * @memberof oj.baseComponent
 */

/**
 * // Returns {'subId': 'oj-some-sub-id', 'componentSpecificProperty': someValue, ...}
 * var context = myComponent.getContextByNode(nodeInsideElement);
 *
 * @ojfragment nodeContextExample
 * @memberof oj.baseComponent
 */

/**
 * <p>The contextMenu slot is set on the <code class="prettyprint">oj-menu</code> within this element.  
 * This is used to designate the JET Menu that this component should launch as a context menu on right-click, Shift-F10, Press & Hold, or component-specific gesture. 
 * If specified, the browser's native context menu will be replaced by the JET Menu specified in this slot.
 * <p>
 * The application can register a listener for the Menu's ojBeforeOpen event. The listener can cancel the launch via event.preventDefault(), 
 * or it can customize the menu contents by editing the menu DOM directly, and then calling refresh() on the Menu.
 * <p>
 * To help determine whether it's appropriate to cancel the launch or customize the menu, the ojBeforeOpen listener can use component API's to determine which 
 * table cell, chart item, etc., is the target of the context menu. See the JSDoc and demos of the individual components for details. 
 * <p>
 * Keep in mind that any such logic must work whether the context menu was launched via right-click, Shift-F10, Press & Hold, or component-specific touch gesture.
 *
 * @ojslot contextMenu
 * @memberof oj.baseComponent
 *
 * @example <caption>Initialize the component with a context menu:</caption>
 * &lt;oj-some-element>
 *     &lt;-- use the contextMenu slot to designate this as the context menu for this component -->
 *     &lt;oj-menu slot="contextMenu" style="display:none" aria-label="Some element's context menu">
 * ...
 *     &lt;/oj-menu>
 * &lt;/oj-some-element>
 */

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 * 
 * @function setProperty
 * @since 4.0.0
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 * 
 * @expose
 * @memberof oj.baseComponent
 * @instance
 * 
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */ 
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @since 4.0.0
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * 
 * @expose
 * @memberof oj.baseComponent
 * @instance
 * 
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @since 4.0.0
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 * 
 * @expose
 * @memberof oj.baseComponent
 * @instance
 * 
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */ 

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Component Messaging Utilities.
 * @param {Object} component instance
 * @protected
 * @constructor
 * @since 0.6
 * @ignore
 */
oj.ComponentMessaging = function (component) 
{
  this.Init(component);
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.ComponentMessaging, oj.Object, "oj.ComponentMessaging");

/**
 * Default display types supported for component messaging.
 * @memberof! oj.ComponentMessaging
 * @const
 * @protected
 * @ignore
 */
oj.ComponentMessaging._DISPLAY_TYPE = {
  NONE: "none",
  NOTEWINDOW: "notewindow", 
  PLACEHOLDER: "placeholder",
  INLINE: "inline"
};

/**
 * Tracks the messaging strategy callback function by type, used to instantiate messaging strategies.
 * Implementations register by type. 
 * @memberof! oj.ComponentMessaging
 * @const
 * @protected
 * @ignore
 */
oj.ComponentMessaging._DISPLAY_TYPE_TO_CALLBACK = {};
  
/**
 * Stores the constructor function callback object used to constuct a strategy object for the 
 * specified type.
 * 
 * @param {string} type
 * @param {Function} strategyConstructorCallback a constructor callback that can be used to constuct
 * a strategy object for the specified type
 * @ignore
 */
oj.ComponentMessaging.registerMessagingStrategy = function (type, strategyConstructorCallback)
{
  if (type && typeof strategyConstructorCallback === "function")
  {
    oj.ComponentMessaging._DISPLAY_TYPE_TO_CALLBACK[type] = strategyConstructorCallback;
  }
};

/**
 * Initializes the strategy based on the display options that specify the messaging artifacts that 
 * will be displayed by this strategy.
 * 
 * @param {Object} component widget instance
 * @memberof! oj.ComponentMessaging
 * @instance
 * @protected
 */
oj.ComponentMessaging.prototype.Init = function (component)
{
  oj.ComponentMessaging.superclass.Init.call(this);
  
  this._component = component;
  this._activated = false;
  
  // sets up this._strategies, which is a list of what we will display in each type of 
  // displayOption. E.g., 'inline': 'messages', 'notewindow': converterHint, validatorHint, etc.
  // See EditableValues displayOption attribute jsdoc for more details.
  this._initializeMessagingStrategies();
};

/**
 * Utility function that activates messaging on the component using the strategy provided.
 * @param {Object} launcher element(s) to which messaging applies
 * @param {Object} content 
 * @private
 */
oj.ComponentMessaging.prototype.activate = function (launcher, contentElement, content)
{
  var that = this;
  oj.Assert.assertObject(content);
  this._launcher = launcher;
  this._contentElement = contentElement;
  
  this._messagingContent = oj.CollectionUtils.copyInto(this._messagingContent || {}, content);
  
  // if already active, reinitialize strategies based on new messagingDisplay preferences.
  if (!this._isActive())
  {
    // for each 'messaging strategy' (e.g., inline == InlineMessagingStrategy, 
    // notewindow == PopupMessagingStrategy, etc), call .activate which initializes
    // the strategy.
    $.each(this._strategies, function(i, strategy) 
    {
      strategy.activate(that);
    });
    this._activated = true;
  }
  else
  {
    this._reactivate();
  }
};

/**
 * Utility function that updates messaging on the component for the content provided, using the 
 * strategies.
 * 
 * @param {Object} content 
 * @private
 */
// TODO: component messaging could take the component instance
oj.ComponentMessaging.prototype.update = function (content)
{
  oj.Assert.assertObject(content);
  oj.Assert.assertBoolean(this._activated);
  
  this._messagingContent = oj.CollectionUtils.copyInto(this._messagingContent || {}, content);

  if (this._activated)
  {
    $.each(this._strategies, function(i, strategy) 
      {
        if (strategy.shouldUpdate(content))
          strategy.update();
      });
  }
};

/**
 * Utility function that deactivates messaging on the component using the strategy provided.
 * @private
 */
oj.ComponentMessaging.prototype.deactivate = function ()
{
  $.each(this._strategies, function(i, strategy) 
    {
      strategy.deactivate();
    });
    
  this._activated = false;
  this._component = null;
  this._launcher = null;
  this._contentElement = null;
  this._strategies = {};
};

/**
 * Utility function that closes anything that needs to be closed when oj.Components.subtreeHidden
 * is called. e.g, popup.
 * @private
 */
oj.ComponentMessaging.prototype.close =  function ()
{
  $.each(this._strategies, function(i, strategy) 
  {
    strategy.close();
  }); 
}

/**
 * Creates a messaging strategy for the specified type, initializing it with the options provided.
 * @param {string|number} type defined by oj.ComponentMessaging._DISPLAY_TYPE. For example,
 * a displayType of 'notewindow' creates a PopupComponentMessaging strategy. See 
 * registerMessagingStrategy where we register the type and the callback to call for a given type.
 * We currently have PopupMessagingStrategy, DefaultMessagingStrategy, PlaceholderMessagingStrategy,
 * and InlineComponentStrategy.
 * e.g., In PopupComponentMessaging.js:
 *  oj.ComponentMessaging.registerMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.NOTEWINDOW, 
 *                              oj.PopupMessagingStrategy
 * @param {Array.<string>|undefined} artifactsForType (e.g., 'messages', 'title', 'validatorHints')
 * 
 * @private
 * @instance
 * @memberOf !oj.ComponentMessaging
 */
oj.ComponentMessaging.prototype._createMessagingStrategy = function (type, artifactsForType)
{
  var callback = oj.ComponentMessaging._DISPLAY_TYPE_TO_CALLBACK[type] || 
          oj.ComponentMessaging._DISPLAY_TYPE_TO_CALLBACK[oj.ComponentMessaging._DISPLAY_TYPE.NONE];
  
  // dynamically instantiate the strategy objects.
  return new callback(artifactsForType);
},

/**
 * Returns the component instance or null
 * 
 * @return {Object|null} 
 * @private
 * @instance
 * @memberOf !oj.ComponentMessaging
 */
oj.ComponentMessaging.prototype._getComponent = function ()     
{
  return this._component || null;
};

/**
 * Returns the launcher jquery element. This is the element on the component to which messaging 
 * applies. 
 * 
 * @return {Object|null} null if messaging is not activated.
 * @private
 * @instance
 * @memberOf !oj.ComponentMessaging
 */
oj.ComponentMessaging.prototype._getLauncher = function ()     
{
  return this._launcher || null;
};


/**
 * Returns the jquery element on the component to which aria-invalid
 * applies. This is either the launcher itself or the inputs
 * within the launcher dom.  JAWS only reads aria-invalid when it is on the input/textara/select.
 * <p>
 * In the case of radioset/checkboxset, for example, where the launcher
 * is the root dom element and the inputs are with it, we return the inputs.
 * 
 * </p>
 * 
 * @return {Object|null} null if launcher is null
 * @private
 * @instance
 * @memberOf !oj.ComponentMessaging
 */
oj.ComponentMessaging.prototype._getContentElement = function ()
{
  return this._contentElement || null;
};

/**
 * Returns the last saved messagingContent object.
 * 
 * @return {Object} 
 * @private
 * @instance
 * @memberOf !oj.ComponentMessaging
 */
oj.ComponentMessaging.prototype._getMessagingContent = function ()     
{
  return this._messagingContent || {};
};

/**
 * Whether the component messaging is activated. 
 * @return {boolean}
 * @private
 */
oj.ComponentMessaging.prototype._isActive = function ()
{
  return this._activated;
};

/**
 * Returns a key/value array: displayTypes -> array of artifacts using that displayType.
 * where artifacts is 'messages', 'converterHint', 'validatorHint', 'title'; 
 * e.g., 
 * artifactsByDisplayType[oj.ComponentMessaging._DISPLAY_TYPE.NOTEWINDOW] = ['messages', 'converterHints']
 * artifactsByDisplayType[oj.ComponentMessaging._DISPLAY_TYPE.NONE] = ['validatorHints']
 * The types of messaging content for which displayOptions can be configured include 
 * messages, converterHint, validatorHint and title.
 * The displayOptions for each type is specified either as an array of strings or a string. 
 * When an array is specified the first display option takes precedence over the second and so on,
 * so we will only have ONE display type per artifact.
 */
oj.ComponentMessaging.prototype._getResolvedMessagingDisplayOptions = function ()
{
    var artifactsByDisplayType = {}, artifactDisplayTypeResolved = false, displayTypes, displayType,
      compPH = this._component.options['placeholder'], artifact, keys, key, i, j,
      messagingPreferences = this._component.options['displayOptions'] || {},
      $messagingPreferences = {}, self = this;
      
    // first resolve primary display options for each artifact.
    // E.g. at the end of this loop you should have something like this
    // {messages: 'notewindow', converterHint: 'placeholder', validatorHint: 'notewindow', title: 'none'}
    keys = Object.keys(messagingPreferences);
    for (i = 0; i < keys.length; i++)
    {
      key = keys[i];
      displayTypes = messagingPreferences[key];
      // loop over array of displayTypes preferred for artifact.
      // artifacts are 'messages', 'converterHint', 'validatorHint', 'title'
      artifactDisplayTypeResolved = false;
      artifact = key + "";
      // we take either array or string values for displayOptions.
      if (Array.isArray(displayTypes))
      {
        for (j = 0; j < displayTypes.length; j++) 
        {
          displayType = displayTypes[j];
          if (!artifactDisplayTypeResolved)
          {
            artifactDisplayTypeResolved = 
              self._resolveDisplayTypeForArtifact(artifact, displayType, compPH, $messagingPreferences);
          }
        }
      }
      else if (typeof displayTypes === "string")
      {
        if (!artifactDisplayTypeResolved)
        {
          artifactDisplayTypeResolved = 
            self._resolveDisplayTypeForArtifact(artifact, displayTypes, compPH, $messagingPreferences);
        }
      }

      // if we couldn't resolve then use "none". E.g., validationHint: ['none']
      if (!artifactDisplayTypeResolved)
      {
        $messagingPreferences[artifact] = oj.ComponentMessaging._DISPLAY_TYPE.NONE;
      }
    }
    
    // collate by displayType -> artifact. but first reset 
    $.each(oj.ComponentMessaging._DISPLAY_TYPE, function(type, name){
      artifactsByDisplayType[name] = [];
    });
    
    $.each($messagingPreferences, function(artifact, displayType)
    {
      // an artifact eventually resolves to one displayType. 
      artifactsByDisplayType[displayType].push(artifact);        
    });
    
    return artifactsByDisplayType;
},
        
oj.ComponentMessaging.prototype._resolveDisplayTypeForArtifact = function(
  artifact, 
  displayType, 
  compPH,
  $messagingPreferences)
{
  var artifactDisplayTypeResolved = false;
  switch (displayType)
  {
    // placeholder display is special in that it's only supported on 'converterHint'.
    case oj.ComponentMessaging._DISPLAY_TYPE.PLACEHOLDER :

      if (artifact === "converterHint")
      {
        // if placeholder is the first preference for converterHint, it's used under certain 
        // conditions
        // if options.placeholder is not set then use 'converterHint' as the default 
        // 'placeholder'
        // alternately if (options.placeholder), i.e., a custom placeholder is set, then 
        // ignore the placeholder displayType and use the next display type as the default 
        // for the artifact. We may have a fallback displayType in which case we use it, 
        // otherwise we use 'none'. E.g., 
        // {'converterHint': ['placeholder', 'notewindow']} // use notewindow
        // {'converterHint': ['placeholder']}               // use none

        if (!artifactDisplayTypeResolved)
        {
          if (!compPH)
          {
            $messagingPreferences[artifact] = displayType;
            artifactDisplayTypeResolved = true;
            
          }
        }
      }
      else
      {
        // displayType 'placeholder' is not supported on other artifacts
        // ignore if present
        // TODO: In the future we may want to support configuring validatorHint ot title as 
        // placeholder as well.
      }

      break;
    // inline display is special in that it's only supported on 'messages'.
    case oj.ComponentMessaging._DISPLAY_TYPE.INLINE :

      if (artifact === "messages")
      {

        if (!artifactDisplayTypeResolved)
        {
          $messagingPreferences[artifact] = displayType;
          artifactDisplayTypeResolved = true;
        }
      }
      else
      {
        // displayType 'inline' is not supported on other artifacts
        // ignore if present
      }

      break;

    default:
      if (!artifactDisplayTypeResolved)
      {
        $messagingPreferences[artifact] = displayType;
        artifactDisplayTypeResolved = true;
      }
      break;
  }
  
  return artifactDisplayTypeResolved;
  
},

/**
 * Creates messaging strategies for the component based on the displayOptions.
 * @private
 */
oj.ComponentMessaging.prototype._initializeMessagingStrategies = function ()
{
  var artifactsByDisplayType = this._getResolvedMessagingDisplayOptions(), 
    displayInNoteWindow = artifactsByDisplayType[oj.ComponentMessaging._DISPLAY_TYPE.NOTEWINDOW], 
    displayNone = artifactsByDisplayType[oj.ComponentMessaging._DISPLAY_TYPE.NONE], 
    displayInPlaceholder = artifactsByDisplayType[oj.ComponentMessaging._DISPLAY_TYPE.PLACEHOLDER],
    displayInline = artifactsByDisplayType[oj.ComponentMessaging._DISPLAY_TYPE.INLINE],
    messagingStrategies = {};
    
  if (displayInNoteWindow.length > 0)
  {
    // displayInNoteWindow is an array of the artifacts that want to be displayed in the note window
    // e.g., 'messages', 'converterHints', etc.
    messagingStrategies[oj.ComponentMessaging._DISPLAY_TYPE.NOTEWINDOW] = 
      this._createMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.NOTEWINDOW, displayInNoteWindow);
  }
  
  if (displayInPlaceholder.length > 0)
  {
    // displayInPlaceholder is an array of the artifacts that want to be displayed in placeholder
    // e.g., 'converterHints'
    messagingStrategies[oj.ComponentMessaging._DISPLAY_TYPE.PLACEHOLDER] = 
      this._createMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.PLACEHOLDER, displayInPlaceholder);
  }
  
  if (displayInline.length > 0)
  {
    // displayInPlaceholder is an array of the artifacts that want to be displayed in placeholder
    // e.g., 'converterHints'
    messagingStrategies[oj.ComponentMessaging._DISPLAY_TYPE.INLINE] = 
      this._createMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.INLINE, displayInline);
  }
  
  messagingStrategies[oj.ComponentMessaging._DISPLAY_TYPE.NONE] = 
    this._createMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.NONE, displayNone);

  this._strategies = messagingStrategies;
};

/**
 * Reinitializes component messaging with new messagingDisplayOptions.
 * 
 * @private
 */
oj.ComponentMessaging.prototype._reactivate = function ()
{
  var artifactsByDisplayType = this._getResolvedMessagingDisplayOptions(), strategy, cm = this; 
  
  // for every displayType being requested either create the messaging strategy for the type or 
  // reuse existing strategy if it has already been created. 
  $.each(artifactsByDisplayType, function(type, artifactsForType) 
    {
      type = type + ""; // coerce to avoid GCC warning
      strategy = cm._strategies[type];
      if (artifactsForType && artifactsForType.length > 0)
      {
        if (!strategy)
        {
          // create a strategy if one doesn't exist for the type
          strategy = cm._createMessagingStrategy(type, artifactsForType);
          cm._strategies[type] = strategy;
          strategy.activate(cm);
        }
        else if (strategy)
        {
          // update strategy with the latest displayOptions if already present. we don;t 
          // want to remove it once activated.
          strategy.reactivate(artifactsForType);
        }
      }
      else
      {
        if (strategy && oj.ComponentMessaging._DISPLAY_TYPE.NONE !== type)
        {
          // if we have no artifacts to show for a type, then remove the strategy.
          // only if its other than the DefaultMessagingStrategy as it's always needed to theme 
          // component.
          strategy.deactivate();
          delete cm._strategies[type];
        }
      }
    });
};

/**
 * A base messaging strategy class that is initialized with a set of displayOptions. This object 
 * also provides helper methods for its subclasses.
 * 
 * @param {Array.<string>} displayOptions an array of messaging artifacts displayed.
 * 
 * @constructor
 * @class oj.MessagingStrategy
 * @private
 */
oj.MessagingStrategy = function (displayOptions)
{
  this.Init(displayOptions);
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.MessagingStrategy, oj.Object, "oj.MessagingStrategy");

/**
 * Initializes the strategy based on the display options that specify the messaging artifacts that 
 * will be displayed by this strategy.
 * 
 * @param {Array.<string>} displayOptions an array of messaging artifacts displayed.
 * @private
 */
oj.MessagingStrategy.prototype.Init = function (displayOptions)
{
  oj.Assert.assertArray(displayOptions);
  oj.MessagingStrategy.superclass.Init.call(this);
  
  this._displayOptions = displayOptions;
};

oj.MessagingStrategy.prototype.activate = function (cm)
{
  this._componentMessaging = cm;
};

/**
 * Cleans up messaging artifacts that were created on the component instance. E.g., destroys any 
 * widgets it created, removes styles added etc.
 * 
 * @private
 */
oj.MessagingStrategy.prototype.deactivate = function ()
{
};

/**
 * Utility function that closes anything that needs to be closed when oj.Components.subtreeHidden
 * is called. e.g, popup.
 * 
 * @private
 */
oj.MessagingStrategy.prototype.close = function ()
{
};

/**
 * Reinitializes with the new display options and updates component messaging using the new content. 
 * 
 * @param {Array.<string>} newDisplayOptions
 * @private
 */
oj.MessagingStrategy.prototype.reactivate = function (newDisplayOptions)
{
  this.Init(newDisplayOptions);
};

/**
 * Returns true always. Subclasses can override to ignore updates. 
 * 
 * @param {Object=} content the messaging content that is being updated 
 * @return {boolean} 
 * 
 * @private
 */
oj.MessagingStrategy.prototype.shouldUpdate = function (content)
{
  return true;
};


/**
 * Updates component with instance using the content provided.
 * 
 * @private
 */
oj.MessagingStrategy.prototype.update = function ()
{
};

// P R O T E C T E D  M E T H O D S 
/**
 * Gets the launcher element for which the messaging is applied.
 * @return {Object} the jquery element of the form element.
 * @private
 */
oj.MessagingStrategy.prototype.GetLauncher = function ()
{
  return this._componentMessaging._getLauncher();
};

/**
 * @return {Object} the jquery element of the form element.
 * @private
 */
oj.MessagingStrategy.prototype.GetContentElement = function ()
{
  return this._componentMessaging._getContentElement();
};

/**
 * Gets the component (widget).
 * @return {Object} the jet component instance
 * @private
 */
oj.MessagingStrategy.prototype.GetComponent = function ()
{
  return this._componentMessaging._getComponent();
};

/**
 * Returns an array of messages.
 * 
 * @return {Array} of messages each an instance of oj.Message
 * @private
 */
oj.MessagingStrategy.prototype.GetMessages = function ()
{
  return this.GetValidityState().getMessages();
};

oj.MessagingStrategy.prototype.GetMaxSeverity = function ()
{
  return this.GetValidityState().getMaxSeverity();
};

/**
 * Gets the converter hint.
 * 
 * @return {Array} an array of hints, each a string.
 * @private
 */
oj.MessagingStrategy.prototype.GetConverterHint = function ()
{
  var hints = [], mc = this._getMessagingContent(), converterHint = mc && mc.converterHint;
  if (converterHint)
  {
    hints.push(converterHint);
  }
  
  return hints;
};

oj.MessagingStrategy.prototype.GetValidatorHints = function ()
{
  var hints = [], mc = this._getMessagingContent(), vHints = mc && mc.validatorHint || [], 
      index, hint;
  $.each(vHints, function (index, hint)
    {
      hints.push(hint);
    }
  );
  
  return hints;
};


/**
 * Gets the short description.
 * @return {string} title or ""
 * @private
 */
oj.MessagingStrategy.prototype.GetTitle = function ()
{
  var mc = this._getMessagingContent();
  return mc && mc.title || ""; 
};

/**
 * Gets the validityState, an instance of oj.ComponentValidity or null.
 * @private
 */
oj.MessagingStrategy.prototype.GetValidityState = function ()
{
  var mc = this._getMessagingContent(); 
  return mc && mc.validityState || null;
};

/**
 * Whether the strategy is displaying messages or not.
 * @return {boolean} true if strategy has messages to display
 * @private
 */
oj.MessagingStrategy.prototype.HasMessages = function ()
{
  var messages = this.GetMessages();
  return (messages && messages.length > 0) ? true: false;
};

oj.MessagingStrategy.prototype.ShowMessages = function ()
{
  return this._displayOptions.indexOf('messages') !== -1 ? true : false;
};

oj.MessagingStrategy.prototype.ShowConverterHint = function ()
{
  return this._displayOptions.indexOf('converterHint') !== -1 ? true : false;
};

oj.MessagingStrategy.prototype.ShowValidatorHint = function ()
{
  return this._displayOptions.indexOf('validatorHint') !== -1 ? true : false;
};

oj.MessagingStrategy.prototype.ShowTitle = function ()
{
  return this._displayOptions.indexOf('title') !== -1 ? true : false;
};

/**
 * Returns true if we have invalid messages; false otherwise.
 * 
 * @return {boolean}
 * @private
 */
oj.MessagingStrategy.prototype.IsInvalid = function ()
{
  return this.GetValidityState().isInvalid();
};


/**
 * Gets the messagingContent stored in ComponentMessaging instance
 * @return {Object}
 * @private
 */
oj.MessagingStrategy.prototype._getMessagingContent = function ()
{
  if (this._componentMessaging)
  {
    return this._componentMessaging._getMessagingContent();
  }
  
  return {};
};


/**
 * A messaging strategy that updates the component theming and accessibility attributes.
 * 
 * @param {Array.<string>} displayOptions .
 * @constructor
 * @extends {oj.MessagingStrategy}
 * @private
 */
oj.DefaultMessagingStrategy = function(displayOptions)
{
  this.Init(displayOptions);
};

oj.ComponentMessaging.registerMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.NONE, 
                                                oj.DefaultMessagingStrategy);

// TODO: Need to retrieve style selectors from a Style Manager
oj.DefaultMessagingStrategy._SELECTOR_STATE_INVALID = "oj-invalid";
oj.DefaultMessagingStrategy._SELECTOR_STATE_WARNING = "oj-warning";

oj.Object.createSubclass(oj.DefaultMessagingStrategy, oj.MessagingStrategy, "oj.DefaultMessagingStrategy");

/**
 * Updates component theming, a11y attributes using the latest component state and its messaging 
 * content.
 * 
 * @private
 */
oj.DefaultMessagingStrategy.prototype.update = function ()
{
  oj.DefaultMessagingStrategy.superclass.update.call(this);

  var launcher = this.GetLauncher(), maxSeverity = this.GetMaxSeverity(), removeClasses = [], 
      addClasses=[], invalid = false, component = this.GetComponent(), jqRoot = component.widget();
  
  if (!launcher)
  {
    return;
  }
  
  // apply element error styling if invalid
  if (this.IsInvalid())
  {
    // enable tooltip; set invalid class and aria invalid
    // TODO: oj classes should be set on the root DOM
    removeClasses.push(oj.DefaultMessagingStrategy._SELECTOR_STATE_WARNING);
    addClasses.push(oj.DefaultMessagingStrategy._SELECTOR_STATE_INVALID);
    invalid = true;
  }
  else
  {
    // TODO: add warning or other severity state
    if (this.HasMessages() && maxSeverity === oj.Message.SEVERITY_LEVEL['WARNING'])
    {
      removeClasses.push(oj.DefaultMessagingStrategy._SELECTOR_STATE_INVALID);
      addClasses.push(oj.DefaultMessagingStrategy._SELECTOR_STATE_WARNING);
    }
    else
    {
      // for all other messages we remove selectors
      removeClasses.push(oj.DefaultMessagingStrategy._SELECTOR_STATE_INVALID);
      removeClasses.push(oj.DefaultMessagingStrategy._SELECTOR_STATE_WARNING);
    }
  }
  
  jqRoot.removeClass(removeClasses.join(" "))
          .addClass(addClasses.join(" ")); // classes added to root
  // aria-invalid needs to be on an input/textarea
  this.GetContentElement().attr({"aria-invalid": invalid});  
  
};

/**
 * Cleans up messaging artifacts that were created on the component instance. E.g., destroys any 
 * widgets it created, removes styles added etc.
 * 
 * @private
 */
oj.DefaultMessagingStrategy.prototype.deactivate = function ()
{
  var jqRoot = this.GetComponent().widget();
  
  jqRoot.removeClass(oj.DefaultMessagingStrategy._SELECTOR_STATE_INVALID)
                  .removeClass(oj.DefaultMessagingStrategy._SELECTOR_STATE_WARNING);
  this.GetContentElement().removeAttr("aria-invalid");
  oj.DefaultMessagingStrategy.superclass.deactivate.call(this);
};


/**
 * A messaging strategy that uses html5 placeholder (for now) to set/remove placeholder content.
 * 
 * @param {Array.<string>} displayOptions an array of messaging artifacts displayed in the placeholder.
 * @constructor
 * @extends {oj.MessagingStrategy}
 * @private
 */
oj.PlaceholderMessagingStrategy = function(displayOptions)
{
  this.Init(displayOptions);
};

oj.ComponentMessaging.registerMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.PLACEHOLDER, 
                                                oj.PlaceholderMessagingStrategy);

// Subclass from oj.MessagingStrategy 
oj.Object.createSubclass(oj.PlaceholderMessagingStrategy, oj.MessagingStrategy, "oj.PlaceholderMessagingStrategy");

/**
 * Initializer
 *  
 * @param {Array.<string>} displayOptions an array of messaging artifacts displayed in the notewindow.
 * @private
 */
oj.PlaceholderMessagingStrategy.prototype.Init = function (displayOptions) 
{
  oj.PlaceholderMessagingStrategy.superclass.Init.call(this, displayOptions);
};

/**
 * Sets up a placeholder for the component instance using the converter hint.
 * 
 * @param {Object} cm a reference to an instance of oj.ComponentMessaging that provides access to 
 * the latest messaging content.
 * 
 * @private
 */
oj.PlaceholderMessagingStrategy.prototype.activate = function (cm)
{
  oj.PlaceholderMessagingStrategy.superclass.activate.call(this, cm);
  this._refreshPlaceholder();
};

oj.PlaceholderMessagingStrategy.prototype.reactivate = function (newDisplayOptions)
{
  oj.PlaceholderMessagingStrategy.superclass.reactivate.call(this, newDisplayOptions);
  this._refreshPlaceholder();
};

/**
 * Returns true if the content being updated includes converterHint prop. This method is an 
 * optimization because the update() method is called too often and any time any content changes. 
 * The only time PlaceholderMessagingStrategy#update needs to execute is when the converter hint 
 * changes.
 * 
 * @param {Object=} content the messaging content that is being updated 
 * @return {boolean} 
 * 
 * @private
 */
oj.PlaceholderMessagingStrategy.prototype.shouldUpdate = function (content)
{
  return content && content.converterHint !== undefined ? true : false;
};

oj.PlaceholderMessagingStrategy.prototype.update = function ()
{
  oj.PlaceholderMessagingStrategy.superclass.update.call(this);
  this._refreshPlaceholder();
};

// a default placeholder is set on the component, and that is typically the converter hint
oj.PlaceholderMessagingStrategy.prototype._refreshPlaceholder = function()
{
  var launcher = this.GetLauncher(), content, hints;
  
  if (this.ShowPlaceholderContent() && launcher)
  {
    hints = this.GetConverterHint();
    content = hints.length ? hints[0] : "";
    var context = {};
    context.internalMessagingSet = true; // to indicate to component that placeholder is being 
                                       	 // set from messaging module 
    this.GetComponent().option({'placeholder': content}, {'_context': context});
  }
};

oj.PlaceholderMessagingStrategy.prototype.ShowPlaceholderContent = function ()
{
  // we have a placeholder to set/show if we have converterHint set.
  return this.ShowConverterHint();
};

/**
 * The ComponentValidity object represent a component's current validity state. The instance 
 * provides specific methods to retrieve info such as <p>
 *  - whether the component is valid <p>
 *  - the messages currently tracked on the component.<p>
 *  - the max severity level of the messages, e.g., fatal, error etc. See oj.Message for details
 * 
 * @param {boolean} valid
 * @param {Array} messages
 * @constructor
 * @private
 */
oj.ComponentValidity = function(valid, messages)
{
  // TODO: provide methods that allow model implementations to instruct the elements to showMessages, 
  // especially the ones marked for 'lazy' notification. 
  this.Init(valid, messages);
};

/**
 * whether there are invalid messages among the list of messages.
 * 
 * @param {Array} messages list of messages
 * @returns {boolean} true if we have invalid messages; false otherwise
 */
oj.ComponentValidity.isInvalid = function (messages)
{
  var maxLevel = oj.Message.getMaxSeverity(messages);
  if (maxLevel >= oj.Message.SEVERITY_LEVEL['ERROR'])
  {
    return true;
  }

  return false;
  
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.ComponentValidity, oj.Object, "oj.ComponentValidity");

/**
 * The jquery element whose validity this object describes
 * @param {boolean} valid
 * @param {Array} messages instances of oj.Message
 */
oj.ComponentValidity.prototype.Init = function (valid, messages) 
{
  oj.ComponentValidity.superclass.Init.call(this);
  this._initialize(valid, messages);
};

/**
 * Returns a boolean true if valid; false if element not valid
 * @returns {boolean}
 * @private
 */
oj.ComponentValidity.prototype.isInvalid = function () 
{
  return this._invalid;
};

/**
 * Returns an Array or messages that we are marked for immediate display or an empty array.
 * @private
 * @returns {Array}
 */
oj.ComponentValidity.prototype.getMessages = function () 
{
  return this._messages;
};

/**
 * Returns the max severity level.
 * @return {number}
 * @private
 */
oj.ComponentValidity.prototype.getMaxSeverity = function () 
{
  return this._maxSeverity;
};

/**
 * Updates the validity state for the component.
 * 
 * @param {boolean} valid
 * @param {Array} messages instances of oj.Message
 * @private
 */
oj.ComponentValidity.prototype.update = function (valid, messages)
{
  this._initialize(valid, messages);
};

oj.ComponentValidity.prototype._initialize = function (valid, messages)
{
  this._compValid = valid;
  this._compMessages = messages;
  
  this._messages = this._getImmediateMessages(); // messages currently showing
  this._maxSeverity = oj.Message.getMaxSeverity(this._messages); // max severity of messages currently showing
  this._invalid = oj.ComponentValidity.isInvalid(this._messages); 
};

/**
 * Returns an array of messages that are marked for immediate display.
 * 
 * @return {Array} of messages each an instance of oj.Message
 * @private
 */
oj.ComponentValidity.prototype._getImmediateMessages = function ()
{
  var messages = this._compMessages || [], immediateMsgs = [], msg;
  for (var index = 0; index < messages.length; index++)
  {
    msg = messages[index];
    // gather component messages marked for immediate display 
    if (msg instanceof oj.ComponentMessage && !msg.canDisplay())
    {
      continue;
    }
    immediateMsgs.push(msg);
  }
  
  return immediateMsgs;
};

/**
 * JET component custom element bridge.
 * 
 * This bridge ensures that JET components with child JET custom elements 
 * can access child properties before the child busy state resolves. 
 * This bridge does not guarantee that all properties for the child 
 * will be available to the application before its busy states resolves, 
 * e.g data bound attribute values. 
 * 
 * Applications should still wait on the element or page level 
 * busy context before accessing properties or methods.
 * 
 * @class
 * @ignore
 */
oj.CustomElementBridge = {};

/**
 * Prototype for the JET component custom element bridge instance
 */
oj.CustomElementBridge.proto = Object.create(oj.BaseCustomElementBridge.proto);

oj.CollectionUtils.copyInto(oj.CustomElementBridge.proto,
{
  AddComponentMethods: function(proto) 
  {
    // Add subproperty getter/setter
    proto['setProperty'] = function(prop, value) 
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      if (!bridge.SaveEarlyPropertySet(prop, value))
      {
        if (!bridge._setEventProperty(this, prop, value) && 
            !bridge._validateAndSetCopyProperty(this, prop, value, null)) 
        {
          // If not an event or copy property, check to see if it's a component specific property
          var meta = oj.BaseCustomElementBridge.__GetPropertyMetadata(prop, oj.BaseCustomElementBridge.getProperties(bridge, this));
          // This check is currently used for translation subproperties which don't all have metadata. We will eventually
          // remove this check once our metadata is generated from our jsDoc, but need this for now in 4.0 so component owners
          // don't need to all update their translations metadata.
          var canSkipTypeCheck = bridge.CanSkipTypeCheck(prop.split('.'));
          
          // For non component specific properties, just set directly on the element instead.
          if (!meta && !canSkipTypeCheck)
            this[prop] = value;
          else
            oj.CustomElementBridge._getPropertyAccessor(this, prop)(value); 
        }
      }
    };
    proto['getProperty'] = function(prop) 
    { 
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      var meta = oj.BaseCustomElementBridge.__GetPropertyMetadata(prop, oj.BaseCustomElementBridge.getProperties(bridge, this));
      var canSkipTypeCheck = bridge.CanSkipTypeCheck(prop.split('.'));
      var event = oj.__AttributeUtils.eventListenerPropertyToEventType(prop);

      // For event listeners and non component specific properties, return the property from the element.
      // Otherwise, return the widget property and let the widget handle dot notation for subproperties.
      if (event || (!meta && !canSkipTypeCheck))
        return this[prop];
      else
      {
        var ext = meta ? meta["extension"] : null;

        if (ext && ext._COPY_TO_INNER_ELEM)
          return bridge._getCopyProperty(this, prop, meta);
        else
          return oj.CustomElementBridge._getPropertyAccessor(this, prop)();
      }
    };
    // Override HTMLELement's focus/blur methods so we can call focus/blur on an inner element if needed.
    proto['focus'] = function()
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      // If focus is called before the component has been created there
      // will be no saved widget instance yet so call the HTMLElement
      // focus instead.
      if (bridge._WIDGET_INSTANCE)
      {
        var focusElem = bridge._WIDGET_INSTANCE.__getFocusElement();
        if (focusElem !== this)
          focusElem.focus();
        else
          HTMLElement.prototype.focus.call(this);
      }
      else
      {
        HTMLElement.prototype.focus.call(this);
      }
    };
    proto['blur'] = function()
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      if (bridge._WIDGET_INSTANCE)
      {
        var focusElem = bridge._WIDGET_INSTANCE.__getFocusElement();
        if (focusElem !== this)
          focusElem.blur();
        else
          HTMLElement.prototype.blur.call(this);
      }
      else
      {
        HTMLElement.prototype.focus.call(this);
      }
    };
  },
  
  BatchedPropertySet: function(elem, props)
  {
    var keys = Object.keys(props);
    var processedMap  = {};
    
    for (var i=0; i<keys.length; i++)
    {
      var property = keys[i];
      var value = props[property];
      
      // exclude event proprties and transfer attributes from batch updates
      if (!this._setEventProperty(elem, property, value) && 
          !this._validateAndSetCopyProperty(elem, property, value, null))
      {
        value = this.ValidatePropertySet(elem, property, value);
        
        property = this.GetAliasForProperty(property);
        processedMap[property] = value;
      }
    }
    // Skip batched property sets if widget constructor isn't available meaning
    // the widget wasn't instantiated due to an error on creation or destroyed.
    var widgetConstructor = oj.Components.__GetWidgetConstructor(this._WIDGET_ELEM);
    if (widgetConstructor)
      widgetConstructor('option', processedMap);
    else
    {
      for (var i=0; i<keys.length; i++)
      {
        var key = keys[i];
        elem['setProperty'](key, props[key]);
      }
    } 
  },

  CanSkipTypeCheck: function(propAr)
  {
    // TODO: Remove method post 4.0.0 after metadata is generated from jsDoc.
    // Temporary fix for JET component translation subproperties that are not specified in metadata.
    return propAr.length > 1 && propAr[0] === 'translations';
  },

  CreateComponent: function(element)
  {
    var innerDomFun = this._INNER_DOM_FUNCTION;
    this._WIDGET_ELEM = oj.CustomElementBridge._getWidgetElement(element, innerDomFun ? innerDomFun(element) : this._EXTENSION._INNER_ELEM);

    // Transfer global attributes and copy tagged properties to child element if one exists
    if (this._WIDGET_ELEM !== element) 
    {
      var transferAttrs = this._EXTENSION._GLOBAL_TRANSFER_ATTRS || [];
      for (var i = 0; i < transferAttrs.length; i++)
      {
        var attr = transferAttrs[i];
        if (element.hasAttribute(attr))
        {
          this._WIDGET_ELEM.setAttribute(attr, element.getAttribute(attr));
          // Remove attribute from custom element after transfering value to inner element
          // Set a flag so we know that we're removing the attribute, not app so
          // that on attribute changed we don't remove it again
          this._removingTransfer = true;
          element.removeAttribute(attr);
        }
      }

      this._copyProperties();

      // Setup blur/focus listeners on inner element so we can trigger on the root custom element for 
      var getFocusEventPropagator = function(type) {
        return function() {
          // Ensure that the target is the custom element, not the inner element, so create
          // a new event and dispatch on the custom element.
          var focusEvent = document.createEvent('UIEvent');
          focusEvent.initEvent(type, false, false);
          element.dispatchEvent(focusEvent);
        };
      }
      this._WIDGET_ELEM.addEventListener("focus", getFocusEventPropagator("focus"));
      this._WIDGET_ELEM.addEventListener("blur", getFocusEventPropagator("blur"));
    }

    oj.Components.unmarkPendingSubtreeHidden(element);

    // Initialize jQuery object with options and pass element as wrapper if needed
    var locator = $(this._WIDGET_ELEM);
    var widgetConstructor = $(this._WIDGET_ELEM)[this._EXTENSION._WIDGET_NAME].bind(locator);
    widgetConstructor(this._PROPS);
    this._WIDGET = widgetConstructor;
    this._WIDGET_INSTANCE = widgetConstructor('instance');

    if (this._WRITEBACK_PROPS)
      this._WIDGET_INSTANCE.__saveWritebackOptions(this._WRITEBACK_PROPS);


    // After parsing the DOM attribute values and initializing properties, remove the disabled
    // property if it exists due to 
    if (element.hasAttribute('disabled') && !this._disabledProcessed)
      oj.CustomElementBridge._removeDisabledAttribute(element);

    // Set flag when we can fire property change events
    this.__READY_TO_FIRE = true;

    // Resolve the component busy state 
    this.resolveDelayedReadyPromise();
  },

  DefineMethodCallback: function (proto, method, methodMeta) 
  {
    proto[method] = function()
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      var methodName = methodMeta['internalName'] || method;
      // Pass in null as thisArg to apply since the widget constructor is prebound to the jQuery element
      return bridge._WIDGET.apply(null, [methodName].concat([].slice.call(arguments)));
    };
  },
  
  DefinePropertyCallback: function (proto, property, propertyMeta) 
  {
    var listener;
    var ext = propertyMeta["extension"];
    Object.defineProperty(proto, property, 
    {
      'enumerable': true,
      'get': function() 
      { 
        if (propertyMeta._eventListener)
        {
          return listener;
        }
        else if (ext && ext._COPY_TO_INNER_ELEM)
        {
          var bridge = oj.BaseCustomElementBridge.getInstance(this);
          return bridge._getCopyProperty(this, property, propertyMeta);
        }
        else
        {
          return oj.CustomElementBridge._getPropertyAccessor(this, property)();
        }
      },
      'set': function(value) 
      {       
        var bridge = oj.BaseCustomElementBridge.getInstance(this);
        // Properties can be set before the component is created. These early
        // sets are actually saved until after component creation and played back.
        if (!bridge.SaveEarlyPropertySet(property, value))
        {
          if (propertyMeta._eventListener) 
          {
            bridge.SetEventListenerProperty(this, property, value);
            listener = value;
          }
          else
          {
            if (!bridge._validateAndSetCopyProperty(this, property, value, propertyMeta))
              oj.CustomElementBridge._getPropertyAccessor(this, property)(value);
          }
        }
      }
    });
  },

  GetAttributes: function(metadata)
  {
    var attrs = oj.BaseCustomElementBridge.getAttributes(metadata['properties']);
    if (metadata['extension']._GLOBAL_TRANSFER_ATTRS)
      return attrs.concat(metadata['extension']._GLOBAL_TRANSFER_ATTRS);
    return attrs;
  },

  GetAliasForProperty: function(property)
  {
    // Aliasing only supported for top level properties
    var alias = this._EXTENSION._ALIASED_PROPS;
    if (alias && alias[property])
      return alias[property];
    return property;
  },

  InitializeElement: function(element)
  {  
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializeElement.call(this, element);

    if (this._EXTENSION._CONTROLS_SUBTREE_HIDDEN)
      oj.Components.markPendingSubtreeHidden(element);

    oj.BaseCustomElementBridge.__InitProperties(element, this._PROPS);
  },

  HandleAttributeChanged: function(element, attr, oldValue, newValue) 
  {
    var transferAttrs = this._EXTENSION._GLOBAL_TRANSFER_ATTRS;
    var transfer = transferAttrs && transferAttrs.indexOf(attr) !== -1;
    if (!this._removingTransfer && transfer && this._WIDGET_ELEM)
    {
      if (newValue == null || newValue === false)
        this._WIDGET_ELEM.removeAttribute(attr);
      else
        this._WIDGET_ELEM.setAttribute(attr, newValue);
      // Remove attribute from custom element after transfering value to inner element
      // Set a flag so we know that we're removing the attribute, not app so
      // that on attribute changed we don't remove it again
      this._removingTransfer = true;
      element.removeAttribute(attr);
    }
    else if (this._removingTransfer && transfer)
    {
      this._removingTransfer = false;
    }
  },

  HandleDetached: function(element) 
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.HandleDetached.call(this, element);

    // Only call __handleDisconnected if the component hasn't previously
    // been destroyed which we can check by seeing if the widget constructor is null
    if (oj.Components.__GetWidgetConstructor(this._WIDGET_ELEM) && this._WIDGET_INSTANCE)
      this._WIDGET_INSTANCE.__handleDisconnected();
  },

  HandleReattached: function(element) {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.HandleReattached.call(this, element);

    if (this._WIDGET_INSTANCE)
      this._WIDGET_INSTANCE.__handleConnected();
  },

  InitializeBridge: function(element)
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializeBridge.call(this, element);
    
    var descriptor = oj.BaseCustomElementBridge.__GetDescriptor(element.tagName);
    this._INNER_DOM_FUNCTION = descriptor['innerDomFunction'];

    this._EXTENSION = this.METADATA['extension'] || {};

    this._PROPS = (this._EXTENSION._INNER_ELEM || this._INNER_DOM_FUNCTION) ? {'_wrapper': element} : {};
    this._setupPropertyAccumulator(element, this._PROPS);

    // Checks metadata for copy and writeback properties
    this._processProperties(element);
  },

  _attributeChangedCallback: function(attr, oldValue, newValue) 
  {
    var bridge = oj.BaseCustomElementBridge.getInstance(this);

    // Due to  where IE11 disables child inputs for a parent with the disabled attribute,
    // we will remove the disabled attribute after we save the value and will ignore all disabled 
    // attribute sets after component initialization when the application can just as easily use the property
    // setter instead. Expressions will be handled in the CustomElementBinding.
    if (attr === "disabled" && bridge._disabledProcessed)
    {
      // Always remove the disabled attribute even after component initialization and log warning.
      // A null value indicates that the value was removed already.
      if (newValue != null)
      {
        oj.Logger.warn("Ignoring 'disabled' attribute change after component initialization. Use element property setter instead.");
        oj.CustomElementBridge._removeDisabledAttribute(this);
      }
      return;
    }

    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto._attributeChangedCallback.call(this, attr, oldValue, newValue);
  },

  _copyProperties: function()
  {
    // Copies properties from the bridge _PROPS before the widget is instantiated
    // removing copied props from the object
    if (this._COPY_ATTRS)
    {
      for (var i = 0; i < this._COPY_ATTRS.length; i++) {
        var attr = this._COPY_ATTRS[i];
        var propName = oj.__AttributeUtils.attributeToPropertyName(attr);
        if (this._PROPS.hasOwnProperty(propName))
        {

          this._WIDGET_ELEM.setAttribute(attr, this._PROPS[propName]);
          // Delete the attribute we just copied from the options that we 
          // instantiate the widget with
          delete this._PROPS[propName];
        }
      }
    }
  },

  _getCopyProperty: function(elem, prop, propMeta)
  {
    var attrName = oj.__AttributeUtils.propertyNameToAttribute(prop);
    var ext = propMeta["extension"];
    if (ext._ATTRIBUTE_ONLY)
    {
      if (this._WIDGET_ELEM.hasAttribute(attrName))
      {
        var value = this._WIDGET_ELEM.getAttribute(attrName);
        var coercedValue;
        try 
        {
          coercedValue = oj.__AttributeUtils.coerceValue(elem, attrName, value, propMeta["type"]);
        }
        catch (ex)
        {
          this._throwError(elem, ex);
        }
        return coercedValue;
      }
      return null;
    } else {
      return this._WIDGET_ELEM[prop];
    }
  },

  _processProperties: function(elem)
  {
    var props = oj.BaseCustomElementBridge.getProperties(this, elem);
    if (props)
    {
      var propKeys = Object.keys(props);
      for (var i = 0; i < propKeys.length; i++)
      {
        var propName = propKeys[i];
        var propMeta = props[propName];
        // Store writeback properties on the bridge and set on widget when we instantiate it later
        if (propMeta['writeback'])
        {
          if (!this._WRITEBACK_PROPS)
            this._WRITEBACK_PROPS = {};
          this._WRITEBACK_PROPS[propName] = true;
        }
        // Store properties to copy to inner element for easy lookup
        var ext = propMeta['extension'];
        if (ext && ext._COPY_TO_INNER_ELEM)
        {
          if (!this._COPY_ATTRS)
            this._COPY_ATTRS = [];
          this._COPY_ATTRS.push(propName);
        }
      }
    }
  },

  _setupPropertyAccumulator: function(element, widgetOptions)
  {
    // Add an element function that will track property values until expressions are all evaluated.
    // This object will be replaced with the actual widget constructor.
    this._WIDGET = function(method, prop, value)
    {
      // Allow property access before widget is created for element binding and dynamic element creation
      if (method === 'option') 
      {
        if (!prop && !value)
        {
          return widgetOptions;
        }
        else
        {
          oj.BaseCustomElementBridge.__SetProperty(this.GetAliasForProperty.bind(this), widgetOptions, prop, value);
          return widgetOptions[prop];
        }
      }
      else
      {
        this._throwError(element, "Cannot access methods before element is upgraded.");
      }
    }
  },

  _validateAndSetCopyProperty: function(elem, prop, value, propMeta)
  {
    // propMeta is could be null so we should retrieve it if not passed in

    var attrName = oj.__AttributeUtils.propertyNameToAttribute(prop);
    var isCopy = this._COPY_ATTRS && this._COPY_ATTRS.indexOf(attrName) !== -1;
    // If widget hasn't been instantiated skip setting until CreateComponent
    if (isCopy) 
    {
      // We need to validate the value so that we don't copy an invalid value. 
      value = this.ValidatePropertySet(elem, prop, value);

      if (this._WIDGET_ELEM)
      {
        if (!propMeta)
          propMeta = oj.BaseCustomElementBridge.__GetPropertyMetadata(prop, oj.BaseCustomElementBridge.getProperties(this, elem));

        var previousValue = this._getCopyProperty(elem, prop, propMeta);
        if (value == null || value === false)
          this._WIDGET_ELEM.removeAttribute(attrName);
        else if (value === true)
          this._WIDGET_ELEM.setAttribute(attrName, '');
        else
          this._WIDGET_ELEM.setAttribute(attrName, value);

        // Fire a property change event for the copy properties since we don't actually pass
        // these to the widget. The widget will never update these properties themselves so 
        // all updates are external.
        oj.BaseCustomElementBridge.__FirePropertyChangeEvent(elem, prop, this._getCopyProperty(elem, prop, propMeta), previousValue, 'external');
      }
      else
      {
        // Save the value until inner widget is created and we can copy them over
        this._PROPS[attrName] = value;
      }
    }
    return isCopy;
  },

  _setEventProperty: function(elem, prop, value)
  {
    var isEvent = false;
    var event = oj.__AttributeUtils.eventListenerPropertyToEventType(prop);
    if (event) 
    {
      elem[prop] = value;
      isEvent = true;
    }
    return isEvent;
  },



});

/*************************/
/* PUBLIC STATIC METHODS */
/*************************/

/**
 * Returns the metadata object for the given component.
 * @param  {string} tagName        The component tag name
 * @return {Object}                The component metadata object
 * @ignore
 */
oj.CustomElementBridge.getMetadata = function(tagName) 
{
  return oj.CustomElementBridge._METADATA_MAP[tagName.toLowerCase()];
};

/**
 * Checks whether the specified event type was declared in the metadata for this custom element
 * @param {Element} element the custom element
 * @param {string} type the event type (e.g. "beforeExpand")
 * @return {boolean} true if the event type was declared in the metadata, false otherwise
 * @ignore
 */
oj.CustomElementBridge.isKnownEvent = function(element, type) 
{
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  if (bridge)
    return (bridge.METADATA['events'] && bridge.METADATA['events'][type]) != null;
  return false;
};

/**
 * Checks whether the specified property was declared in the metadata for this custom element
 * @param {Element} element the custom element
 * @param {string} prop the property name (e.g. "selection")
 * @return {boolean} true if the property was declared in the metadata, false otherwise
 * @ignore
 */
oj.CustomElementBridge.isKnownProperty = function(element, prop) 
{
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  if (bridge)
    return (bridge.METADATA['properties'] && bridge.METADATA['properties'][prop]) != null;
  return false;
};

/**
 * Returns the custom element property for a given aliased component property which can be used
 * for converting an internal optionChange event, e.g. returning readonly for oj-switch's readOnly
 * property so we can fire a readonly-changed event instead of readOnly-changed.
 * Will return the original property if there is no aliasing.
 * @param {Element} element The custom element
 * @param {string} property The component property
 * @return {string}
 * @ignore
 */
oj.CustomElementBridge.getPropertyForAlias = function(element, property) 
{
  // Aliasing only supported for top level properties
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  var alias = bridge._EXTENSION._COMPONENT_TO_ELEMENT_ALIASES;
  if (alias && alias[property])
    return alias[property];
  return property;
};

/**
 * Registers a component as a custom element.
 * @param {string} tagName The component tag name (all lower case), which should contain a dash '-' and not be a reserved tag name.     
 * @param {Object} descriptor The registration descriptor. The descriptor will contain keys for metadata and other component overrides.
 * @param {Object} descriptor.metadata The JSON object containing info like the widget name, whether component has an inner element, an outer wrapper, and the component metadata.
 * @param {function(string, string, Object, function(string))} descriptor.parseFunction The function that will be called to parse attribute values.
 * Note that this function is only called for non bound attributes. The parseFunction will take the following parameters:
 * <ul>
 *  <li>{string} value: The value to parse.</li>
 *  <li>{string} name: The name of the attribute.</li>
 *  <li>{Object} meta: The metadata object for the property which can include its type, default value, 
 *      and any extensions that the composite has provided on top of the required metadata.</li>
 *  <li>{function(string)} defaultParseFunction: The default parse function for the given attribute 
 *      type which is used when a custom parse function isn't provided and takes as its parameters 
 *      the value to parse.</li>
 * </ul>
 * @param {Element} descriptor.innerDomFunction The function that will be called to return the tag name of the inner DOM element, e.g. 'button' or 'a' 
 * The innerDomFunction will take the following parameters:
 * <ul>
 *  <li>{Element} element: The component custom element.</li>
 * </ul>
 * @ignore
 */
oj.CustomElementBridge.register = function(tagName, descriptor)
{
  // Use the simple definitional element prototype if no real widget is associated with this custom element
  var ext = descriptor[oj.BaseCustomElementBridge.DESC_KEY_META]['extension'];
  var proto = ext && ext._WIDGET_NAME ? oj.CustomElementBridge.proto : oj.DefinitionalElementBridge.proto;

  // Create component to element property alias mapping for easy optionChange lookup and stash it in the extension object
  var aliasMap = ext._ALIASED_PROPS;
  if (aliasMap)
  {
    ext._COMPONENT_TO_ELEMENT_ALIASES = {};
    var aliases = Object.keys(aliasMap);
    aliases.forEach(function(alias) {
      ext._COMPONENT_TO_ELEMENT_ALIASES[aliasMap[alias]] = alias;
    });
  }

  if (oj.BaseCustomElementBridge.__Register(tagName, descriptor, proto))
  {
    customElements.define(tagName.toLowerCase(), proto.getClass(descriptor));
  }
};

/**
 * Registers component metadata merging it with the metadata from its superclass hierarchy as needed.
 * @param  {string} tagName       The component tag name
 * @param  {string?} superclassName The component's superclass name
 * @param  {Object} metadata        The component metadata object
 * @ignore
 */
oj.CustomElementBridge.registerMetadata = function(tagName, superclassName, metadata) 
{
  metadata = oj.BaseCustomElementBridge.__ProcessEventListeners(metadata, true);
  if (superclassName) 
  {
    var superMeta = oj.CollectionUtils.copyInto({}, oj.CustomElementBridge.getMetadata(superclassName), undefined, true);
    oj.CustomElementBridge._METADATA_MAP[tagName.toLowerCase()] = oj.CollectionUtils.copyInto(superMeta, metadata, undefined, true);
  } 
  else 
  {
    oj.CustomElementBridge._METADATA_MAP[tagName.toLowerCase()] = metadata;
  }
};

/*****************************/
/* NON PUBLIC STATIC METHODS */
/*****************************/

/**
 * Returns a property accessor for setting/getting options
 * @private
 */
oj.CustomElementBridge._getPropertyAccessor = function(element, property) 
{
  var optionAccessor = function(value)
  {
    var bridge = oj.BaseCustomElementBridge.getInstance(element);

    if (arguments.length === 1)
    {
      value = bridge.ValidatePropertySet(element, property, value);
      property = bridge.GetAliasForProperty(property);
      bridge._WIDGET('option', property, value);
    }
    else
    {
      property = bridge.GetAliasForProperty(property);
      return bridge._WIDGET('option', property);
    }
  };
  return optionAccessor.bind(element);
};

/**
 * Returns the element that the widget constructor will be instantiated on which can be the custom element or a child element.
 * @private
 */
oj.CustomElementBridge._getWidgetElement = function(element, innerTagName) 
{
  // If component widget is bound to an inner child element like <ul> for <oj-list-view>, 
  // create one only if the application does not provide it.
  var widgetElem = element;
  if (innerTagName)
  {
    var firstChild = element.firstElementChild;
    if (firstChild && firstChild.tagName.toLowerCase() === innerTagName)
    {
      widgetElem = firstChild;
    }
    else
    {
      widgetElem = document.createElement(innerTagName);
      // Make a copy of the custom element children before appending the inner element
      var children = [];
      var nodeList = element.childNodes;
      for (var i = 0; i < nodeList.length; i++)
        children.push(nodeList[i]);
      
      element.appendChild(widgetElem); // @HtmlUpdateOk
      // If we create the inner child element, check to see if there are any children
      // to move like for <oj-button> which can have a child elements that should be moved to
      // the newly created inner <button> element.
      while (children.length)
      { 
        var child = children.shift();
        // do not move slot children to inner child element
        // component should decide whether they should be moved to its inner child element
        // for example, it does not make sense for <oj-list-view> to move contextMenu slot to its inner <ul> element.
        if (!child.hasAttribute || !child.hasAttribute("slot"))
          widgetElem.appendChild(child);
      }
    }
    // add data-oj-internal attribute for automation tests
    widgetElem.setAttribute('data-oj-internal', '');
  }
  return widgetElem;
};

/**
 * Removes the disabled attribute from an element and marks the bridge as having
 * processed the value to prevent evaluation of additional attribute sets.
 * @param  {Element} element The custom element
 * @private
 */
oj.CustomElementBridge._removeDisabledAttribute = function(element)
{
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  bridge._disabledProcessed = true;
  element.removeAttribute('disabled');
};

/**
 * Map of registered custom element names
 * @private
 */
oj.CustomElementBridge._METADATA_MAP = {};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
var DataProviderFeatureChecker = /** @class */ (function () {
    function DataProviderFeatureChecker() {
    }
    DataProviderFeatureChecker.isIteratingDataProvider = function (dataprovider) {
        if (dataprovider['fetchFirst']) {
            return true;
        }
        return false;
    };
    DataProviderFeatureChecker.isFetchByKeys = function (dataprovider) {
        if (dataprovider['fetchByKeys']) {
            return true;
        }
        return false;
    };
    DataProviderFeatureChecker.isFetchByOffset = function (dataprovider) {
        if (dataprovider['fetchByOffset']) {
            return true;
        }
        return false;
    };
    return DataProviderFeatureChecker;
}());
oj.DataProviderFeatureChecker = DataProviderFeatureChecker;

/**
 * A bridge for a custom element that renders using a constructor
 * function. Note that when a constructor function is provided, the new instance isn't
 * created until the CreateComponent method so property changes that occur before the
 * component instance is created will no-op.
 *
 * Components that provide a constructor function should implement the following methods:
 * createDOM - Called when the component is instantiated
 * updateDOM - Called after createDOM and when the component needs to do a full render on 
 * refresh and property changes if the component is not handling them separately.
 * handlePropertyChanged - (optional) Called when properties change and should return true if
 * the component has handled the property change and does not need to do a full render. If 
 * false is returned, updateDOM will be called to do a full render.
 *
 * When the constructor function is called, the bridge will pass a context object
 * with the following keys:
 * element - The custom element
 * props - A proxy for the element properties with setter/getter and setProperty APIs allowing the
 *         component to control writeback.
 * unique - A unique ID that the component can append to the custom element ID to generate unique IDs
 * 
 * Note that components supporting the constructor function approach may eventually
 * be refactored into composites once composites support non template rendering.
 * 
 * This bridge ensures that JET components with child JET custom elements 
 * can access child properties before the child busy state resolves. 
 * This bridge does not guarantee that all properties for the child 
 * will be available to the application before its busy states resolves, 
 * e.g data bound attribute values. 
 * 
 * Applications should still wait on the element or page level 
 * busy context before accessing properties or methods.
 * 
 * @class
 * @ignore
 */
oj.DefinitionalElementBridge = {};

/**
 * Prototype for the JET component definitional bridge instance
 */
oj.DefinitionalElementBridge.proto = Object.create(oj.BaseCustomElementBridge.proto);

oj.CollectionUtils.copyInto(oj.DefinitionalElementBridge.proto,
{
  beforePropertyChangedEvent: function(element, property, detail)
  {
    // Call the renderer function so the definitional element can refresh its UI
    var changedProp = property;
    var value = detail.value;
    if (detail.subproperty)
    {
      changedProp = detail.subproperty.path;
      value = detail.subproperty.value;
    }
    this._partialRender(element, changedProp, value);
  },

  AddComponentMethods: function(proto) 
  {
    // Add refresh and subproperty getter/setter methods for all definitional elements
    proto['refresh'] = function() 
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      bridge._fullRender(this);
    };
    proto['setProperty'] = function(prop, value) {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      if (!bridge.SaveEarlyPropertySet(prop, value))
        bridge.SetProperty(this, prop, value, this, true);
    };
    proto['getProperty'] = function(prop) {
      // 'this' is the property object we pass to the definitional element contructor to track internal property changes
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      return bridge.GetProperty(this, prop, this);
    };
    proto['_propsProto']['setProperty'] =  function(prop, value) {
      // 'this' is the property object we pass to the definitional element contructor to track internal property changes
      this._BRIDGE.SetProperty(this._ELEMENT, prop, value, this, false);
    };
    proto['_propsProto']['getProperty'] = function(prop) {
      return this._BRIDGE.GetProperty(this, prop, this);
    };
  },

  CreateComponent: function(element)
  {
    oj.Components.unmarkPendingSubtreeHidden(element);

    if (!this._INSTANCE && this._EXTENSION._CONSTRUCTOR)
    {
      // We expose a similar set of properties as composites except that props is
      // not a Promise and we don't expose any slot information.
      // At the moment some definitional elements have mutation observers so they don't need
      // to rely on refresh being called to be alerted of new children so any cached slotMap
      // can become out of sync. We should add this once we build in support to auto detect 
      // added/removed children to custom elements.
      this._CONTEXT = {
        'element': element,
        'props': this._PROPS_PROXY,
        'unique': oj.BaseCustomElementBridge.__GetUnique(),
      };
      this._CONTEXT['uniqueId'] = element.id ? element.id : this._CONTEXT['unique'];
      this._INSTANCE = new this._EXTENSION._CONSTRUCTOR(this._CONTEXT);
      // Let the component initialize any additional DOM and then do a full render
      if (this._INSTANCE.createDOM)
        this._INSTANCE.createDOM();
      if (this._INSTANCE.updateDOM)
        this._INSTANCE.updateDOM();
    }

    // Set flag when we can fire property change events
    this.__READY_TO_FIRE = true;

    // Resolve the component busy state 
    this.resolveDelayedReadyPromise();

  },
  
  DefineMethodCallback: function (proto, method, methodMeta) 
  {
    proto[method] = function()
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      if (bridge._INSTANCE)
      {
        var methodName = methodMeta['internalName'] || method;
        return bridge._INSTANCE[methodName].apply(bridge._INSTANCE, arguments);
      }
    };
  },

  DefinePropertyCallback: function (proto, property, propertyMeta) 
  {
    var set = function(value, bOuterSet)
    {
      // Properties can be set before the component is created. These early
      // sets are actually saved until after component creation and played back.
      if (!this._BRIDGE.SaveEarlyPropertySet(property, value))
      {
        var previousValue = this._BRIDGE._PROPS[property];
        if (!oj.BaseCustomElementBridge.__CompareOptionValues(property, propertyMeta, value, previousValue)) 
        {
          // Skip validation for inner sets so we don't throw an error when updating readOnly writeable properties
          if (bOuterSet)
              value = this._BRIDGE.ValidatePropertySet(this._ELEMENT, property, value)
          if (propertyMeta._eventListener)
          {
            this._BRIDGE.SetEventListenerProperty(this._ELEMENT, property, value);
            this._BRIDGE._PROPS[property] = value;
          }
          else 
          {
            this._BRIDGE._PROPS[property] = value;
            oj.BaseCustomElementBridge.__FirePropertyChangeEvent(this._ELEMENT, property, value, previousValue, bOuterSet ? 'external' : 'internal');
          }
        }
      }
    }

    var innerSet = function(value)
    {
      set.bind(this)(value, false);
    }

    // Called on the custom element
    var outerSet = function(value)
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      set.bind(bridge._PROPS_PROXY)(value, true);
    }

    var get = function()
    {
      var value = this._BRIDGE._PROPS[property];
      // If the attribute has not been set, return the default value
      if (value === undefined)
      {
        value = propertyMeta['value'];
        // Make a copy if the default value is an Object or Array to prevent modification
        // of the metadata copy and store in the propertyTracker so we have a copy
        // to modify in place for the object case
        if (Array.isArray(value))
          value = value.slice();
        else if (typeof value === 'object')
          value = oj.CollectionUtils.copyInto({}, value, undefined, true);
        this._BRIDGE._PROPS[property] = value;
      }
      return value;
    }

    var innerGet = function()
    {
      return get.bind(this)();
    }

    // Called on the custom element
    var outerGet = function()
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      return get.bind(bridge._PROPS_PROXY)();
    }

    // Don't add event listener properties for inner props
    if (!propertyMeta._derived)
      oj.BaseCustomElementBridge.__DefineDynamicObjectProperty(proto['_propsProto'], property, innerGet, innerSet);
    oj.BaseCustomElementBridge.__DefineDynamicObjectProperty(proto, property, outerGet, outerSet);
  },

  InitializeElement: function(element)
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializeElement.call(this, element);

    if (this._EXTENSION._CONTROLS_SUBTREE_HIDDEN)
      oj.Components.markPendingSubtreeHidden(element);

    oj.BaseCustomElementBridge.__InitProperties(element, this._PROPS_PROXY);  
  },

  InitializePrototype: function(proto)
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializePrototype.call(this, proto);
    
    Object.defineProperty(proto, '_propsProto', {value: {}});
  },

  InitializeBridge: function(element)
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializeBridge.call(this, element);

    this._EXTENSION = this.METADATA['extension'] || {};

    // For tracking outer property sets
    this._PROPS = {};

    /// For tracking inner property sets
    if (element['_propsProto'])
    {
      this._PROPS_PROXY = Object.create(element['_propsProto']);
      this._PROPS_PROXY._BRIDGE = this;
      this._PROPS_PROXY._ELEMENT = element;
    }
  },

  _fullRender: function(element)
  {
    if (this._INSTANCE && this._INSTANCE.updateDOM)
      this._INSTANCE.updateDOM();
  },

  _partialRender: function(element, property, value)
  {
    if (this._INSTANCE)
    {
      // For partial renders, check to see if the component is handling the property change
      // or if it should do a full render
      var handlePropChangedFun = this._INSTANCE.handlePropertyChanged;
      var fullRender = !handlePropChangedFun || !handlePropChangedFun(property, value);
      if (fullRender && this._INSTANCE.updateDOM)
        this._INSTANCE.updateDOM();
    }
  }

});

/*jslint browser: true*/
/*
** Copyright (c) 2004, 2012, Oracle and/or its affiliates. All rights reserved.
*/
/**
 * DOM utilities.
 * @ignore
 */
oj.DomUtils = {};

oj.DomUtils._HTML_START_TAG = "\x3chtml\x3e";
oj.DomUtils._HTML_END_TAG = "\x3c/html\x3e";
oj.DomUtils._LEGAL_ELEMENTS = {"SPAN":1, "B":1, "I":1, "EM":1, "BR":1, "HR":1, "LI":1, "OL":1, "UL":1, "P":1, "TT":1, "BIG":1, "SMALL":1, "PRE":1};
oj.DomUtils._LEGAL_ATTRIBUTES = {"class":1, "style":1};

/**
 * Returns true if the value is null or if the trimmed value is of zero length.
 *
 * @param {string|null} content
 * @return {boolean} true if the string is wrapped in <html> tag.
 */
oj.DomUtils.isHTMLContent = function(content)
{
  if (content.indexOf(oj.DomUtils._HTML_START_TAG) === 0 &&
      content.lastIndexOf(oj.DomUtils._HTML_END_TAG) === content.length - 7)
  {
    return true;
  }

  return false;
};

oj.DomUtils.cleanHtml = function (value)
{
  var offSpan = $(document.createElement("span")).get(0);
  offSpan.innerHTML = value;  // @HTMLUpdateOK safe manipulation
  if (value && value.indexOf("\x3c") >= 0)
  {
    oj.DomUtils._cleanElementHtml(offSpan);
  }
  return offSpan;
};

oj.DomUtils._cleanElementHtml = function(node)
{
  var children = node.childNodes, child, attrs, attr, childHasAttr, i;
  var count = children.length - 1;
  while (count >= 0)
  {
    child = children.item(count);
    if (child && child.nodeType === 1)
    {
      if (oj.DomUtils._LEGAL_ELEMENTS[child.nodeName])
      {
        attrs = child.attributes;
        for (i = attrs.length - 1;i >= 0;i--)
        {
          attr = attrs[i];
          // jquery - the .attr() method returns undefined for attributes that have not been set.
          childHasAttr = $(child).attr(attr.name) !== undefined;
          if (childHasAttr)
          {
            if (!oj.DomUtils._LEGAL_ATTRIBUTES[attr.name])
            {
              child.removeAttribute(attr.nodeName);
            }
          }
        }
        oj.DomUtils._cleanElementHtml(child);
      }
      else
      {
        if (child)
        {
          node.removeChild(child);
        }
      }
    }

    count--;
  }
};

/**
 * Checks to see if the "ancestorNode" is a ancestor of "node".
 *
 * @param {!Element} ancestorNode dom subtree to check to see if the target node exists
 * @param {!Element} node target node to check to see if it exists within a subtree rooted at the ancestorNode
 * @return {boolean} <code>true</code> if the "ancestorNode" is a ancestor of "node".
 */
oj.DomUtils.isAncestor = function (ancestorNode, node)
{
  // These can cause problems in IE11: sometimes the node is just an "empty" object
  //oj.Assert.assertDomElement(ancestorNode);
  //oj.Assert.assertDomElement(node);

  var parentNode = node.parentNode;

  while (parentNode)
  {
    if (parentNode == ancestorNode)
      return true;

    parentNode = parentNode.parentNode;
  }

  return false;
}

/**
 * Checks to see if the "ancestorNode" is a ancestor of "node" or if they are the same.
 *
 * @param {!Element} ancestorNode dom subtree to check to see if the target node exists
 * @param {!Element} node target node to check to see if it exists within a subtree rooted at the ancestorNode
 * @return {boolean} <code>true</code> if the "ancestorNode" is a ancestor of "node" or if they are the same
 */
oj.DomUtils.isAncestorOrSelf = function (ancestorNode, node)
{
  // These can cause problems in IE11: sometimes the node is just an "empty" object
  //oj.Assert.assertDomElement(ancestorNode);
  //oj.Assert.assertDomElement(node);

  return (node == ancestorNode) ?
          true :
          oj.DomUtils.isAncestor(ancestorNode, node);
};


/**
 * Adds a resize listener for a block or inline-block element
 * @param {!Element} elem - node where the listener should be added
 * @param {!Function} listener - listener to be added. The listener will receive
 * two parameters: 1) the new width in pixels; 2) the new height in pixels
 * @param {number=} collapseEventTimeout - timeout in milliseconds for collapsing
 * multiple resize events into one
 * @export
 */
oj.DomUtils.addResizeListener = function(elem, listener, collapseEventTimeout)
{
  var jelem = $(elem);
  var tracker = jelem.data(oj.DomUtils._RESIZE_TRACKER_KEY);
  if (tracker == null)
  {
    tracker = new oj.DomUtils._ResizeTracker(elem);
    jelem.data(oj.DomUtils._RESIZE_TRACKER_KEY, tracker);
    tracker.start();
  }
  tracker.addListener(listener, collapseEventTimeout);
}

/**
 * Removes a resize listener
 * @param {!Element} elem - node whose listener should be removed
 * @param {!Function} listener - listener to be removed
 * @export
 */
oj.DomUtils.removeResizeListener = function(elem, listener)
{
  var jelem = $(elem);
  var tracker = jelem.data(oj.DomUtils._RESIZE_TRACKER_KEY);
  if (tracker != null)
  {
    tracker.removeListener(listener);
    if (tracker.isEmpty())
    {
      tracker.stop();
      jelem.removeData(oj.DomUtils._RESIZE_TRACKER_KEY);
    }
  }
};


/**
 * Fixes resize listeners after a subtree has been connected to the DOM or after
 * its display:none stayle has been removed
 * @param {!Element} subtreeRoot - subtree root
 */
oj.DomUtils.fixResizeListeners = function(subtreeRoot)
{
  $(subtreeRoot).find(".oj-helper-detect-expansion").parent().each(
    function(index, div)
    {
      var tracker = $(div).data(oj.DomUtils._RESIZE_TRACKER_KEY);
      if (tracker != null)
      {
        tracker.init(true);
      }
    }
  );
};

/**
 * Determines whether a special 'meta' key was pressed when the event was fired.
 * For Mac OS, the 'meta' key is mapped to the 'Command' key, for all other platforms it is mapped
 * to the 'Control' key.
 * Note that this method will only work for the events that support .ctrlKey and .metaKey fields.
 * @param {!Object} evt - the event
 * @return true if the meta key is pressed, false otherwise
 */
oj.DomUtils.isMetaKeyPressed = function(evt)
{
  var agentInfo = oj.AgentUtils.getAgentInfo();
  return (oj.AgentUtils.OS.MAC === agentInfo["os"] ? evt.metaKey : evt.ctrlKey);
};

/**
 * Dispatches an event on the element
 * @param {!Element} element
 * @param {!Event} evt event object
 */
oj.DomUtils.dispatchEvent = function(element, evt)
{
  // Workaround for Mozilla issue #329509 - dispatchEvent() throws an error if
  // the element is disabled and disconnected
  // Also, IE simply ignores the .dispatchEvent() call for disabled elements
  var dis = 'disabled';
  var oldDisabled = element[dis];
  try
  {
    element[dis] = false;
    element.dispatchEvent(evt);
  }
  finally
  {
    element[dis] = oldDisabled;
  }
};

/**
 * @private
 */
oj.DomUtils._invokeAfterPaint =  (window['requestAnimationFrame'] || window['mozRequestAnimationFrame'] ||
                                  window['webkitRequestAnimationFrame'] ||
                                  function(fn){ return window.setTimeout(fn, 0); }
                                  ).bind(window);

/**
 * @private
 */
oj.DomUtils._cancelInvokeAfterPaint =  (window['cancelAnimationFrame'] || window['mozCancelAnimationFrame'] ||
                                  window['webkitCancelAnimationFrame'] ||
                                  function(id){ return window.clearTimeout(id); }
                                  ).bind(window);

/**
 * Utility class for tracking resize events for a given element and  sispatching them
 * to listeners
 * @constructor
 * @ignore
 * @private
 */
oj.DomUtils._ResizeTracker = function(div)
{
  var _listeners = $.Callbacks();
  var _collapsingManagers = [];
  var _collapsingListeners = [];

  var _RETRY_MAX_COUNT = 2;
  var _retrySetScroll = 0;
  var _invokeId = null;
  var _oldWidth  = null;
  var _oldHeight = null;
  var _detectExpansion = null;
  var _detectContraction = null;
  var _resizeListener = null;
  var _scrollListener = null;

  this.addListener = function(listener, collapseEventTimeout)
  {
    if (collapseEventTimeout === undefined || collapseEventTimeout === 0)
    {
      _listeners.add(listener);
    }
    else
    {
      _collapsingManagers.push(
              new oj.DomUtils._collapsingListenerManager(listener, collapseEventTimeout));
      _collapsingListeners.push(listener);
    }
  };

  this.removeListener = function(listener)
  {
    var index = _collapsingListeners.indexOf(listener);
    if (index >=0 )
    {
      _collapsingListeners.splice(index, 1);
      var removed = _collapsingManagers.splice(index, 1);
      removed[0].stop();
    }
    else
    {
      _listeners.remove(listener);
    }
  };

  this.isEmpty = function()
  {
    return !_listeners.has() && _collapsingListeners.length == 0;
  };

  this.start = function()
  {
    _scrollListener = _handleScroll.bind(this);

    // : Use native onresize support on teh DIV in IE9/10 and  since no scroll events are fired on the
    // contraction/expansion DIVs in IE9
    if (div.attachEvent)
    {
      _resizeListener = _handleResize.bind(this);
      div.attachEvent('onresize', _resizeListener);
    }

    else
    {
      var firstChild = div.childNodes[0];

      // This child DIV will track expansion events. It is meant to be 1px taller and wider than the DIV
      // whose resize events we are tracking. After we set its scrollTop and scrollLeft to 1, any increate in size
      // will fire a scroll event
      _detectExpansion = document.createElement("div");
      _detectExpansion.className = "oj-helper-detect-expansion";
      var expansionChild = document.createElement("div");
      _detectExpansion.appendChild(expansionChild); //@HTMLUpdateOK; expansionChild constructed by the code above
      if (firstChild != null)
      {
        div.insertBefore(_detectExpansion, firstChild);//@HTMLUpdateOK; _detectExpansion constructed by the code above
      }
      else
      {
        div.appendChild(_detectExpansion);//@HTMLUpdateOK; _detectExpansion constructed by the code above
      }

      _detectExpansion.addEventListener("scroll", _scrollListener, false);

      // This child DIV will track contraction events. Its height and width are set to 200%. After we set its scrollTop and
      // scrollLeft to the current height and width of its parent, any decrease in size will fire a scroll event
      _detectContraction = document.createElement("div");
      _detectContraction.className = "oj-helper-detect-contraction";

      var contractionChild = document.createElement("div");
      contractionChild.style.width = "200%";
      contractionChild.style.height = "200%";
      _detectContraction.appendChild(contractionChild); //@HTMLUpdateOK; contractionChild constructed by the code above
      div.insertBefore(_detectContraction, _detectExpansion); //@HTMLUpdateOK; _detectContraction constructed by the code above

      _detectContraction.addEventListener("scroll", _scrollListener, false);

      this.init(false);
      }
  };

  this.stop = function()
  {
    if (_invokeId != null)
    {
      oj.DomUtils._cancelInvokeAfterPaint(_invokeId);
      _invokeId = null;
    }
    if (_detectExpansion != null)
    {
      _detectExpansion.removeEventListener("scroll", _scrollListener);
      _detectContraction.removeEventListener("scroll", _scrollListener);
      // Check before removing to prevent CustomElement polyfill from throwing
      // a NotFoundError when removeChild is called with an element not in the DOM
      if (_detectExpansion.parentNode)
        div.removeChild(_detectExpansion);
      if (_detectContraction.parentNode)
        div.removeChild(_detectContraction);
    }
    else
    {
      // assume IE9/10
      div.detachEvent('onresize', _resizeListener);
    }
  };

  this.init = function(isFixup)
  {
    var adjusted = _checkSize(isFixup);
    if (isFixup && !adjusted && _detectExpansion.offsetParent != null)
    {
      _adjust(_oldWidth, _oldHeight);
    }
  };


  function _checkSize(fireEvent)
    {
    var adjusted = false;
    if (_detectExpansion.offsetParent != null)
    {
      var newWidth = _detectExpansion.offsetWidth;
      var newHeight = _detectExpansion.offsetHeight;

      if (_oldWidth !== newWidth || _oldHeight !== newHeight)
    {
        _retrySetScroll = _RETRY_MAX_COUNT;
        _adjust(newWidth, newHeight);
        adjusted = true;

        if (fireEvent)
      {
          _notifyListeners(true);
      }
      }
    }

    return adjusted;
  };

  function _notifyListeners(useAfterPaint)
  {
    var newWidth = div.offsetWidth;
    var newHeight = div.offsetHeight;
    if (_listeners.has())
    {
      if (!useAfterPaint)
      {
        _listeners.fire(newWidth, newHeight);
    }
      else
    {
        if (_invokeId !== null)
      {
          oj.DomUtils._cancelInvokeAfterPaint(_invokeId);
      }

        _invokeId = oj.DomUtils._invokeAfterPaint(
        function()
        {
            _invokeId = null;
            _listeners.fire(newWidth, newHeight);
        }
      );
    }
    }

    for (var i=0; i < _collapsingManagers.length; i++)
    {
      _collapsingManagers[i].getCallback()(newWidth, newHeight);
    }
  };

  function _handleScroll(evt)
  {
    evt.stopPropagation();
    if (!_checkSize(true))
    {
      // Workaround for the WebKit issue where scrollLeft gets reset to 0 without the DIV being expanded
      // We will retry to the set the scrollTop only twice to avoid infinite loops
      if (_retrySetScroll > 0 && _detectExpansion.offsetParent != null &&
            (_detectExpansion.scrollLeft == 0 || _detectExpansion.scrollTop == 0))
      {
        _retrySetScroll--;
        _adjust(_oldWidth, _oldHeight);
      }
    }
  };

  function _handleResize()
  {
     _notifyListeners(false);
  };

  function _adjust(width, height)
  {
    _oldWidth = width;
    _oldHeight = height;

    var expansionChildStyle = _detectExpansion.firstChild.style;

    var delta = 1;

    // The following loop is a workaround for the WebKit issue with zoom < 100% -
    // the scrollTop/Left gets reset to 0 because it gets computed to a value less than 1px.
    // We will try up to the delta of 5 to support scaling down to 20% of the original size
    do
    {
      expansionChildStyle.width = width + delta + 'px';
      expansionChildStyle.height = height + delta + 'px';
      _detectExpansion.scrollLeft = _detectExpansion.scrollTop = delta;
      delta++;
    } while ((_detectExpansion.scrollTop == 0 || _detectExpansion.scrollLeft == 0) && delta <= 5);


    _detectContraction.scrollLeft = width;
    _detectContraction.scrollTop = height;
  };
}

oj.DomUtils._RESIZE_TRACKER_KEY = "_ojResizeTracker";


/**
 * Returns true if the name is a valid identifier
 *
 * @param {string} name
 * @return {boolean} true if the name is a valid identifier
 */
oj.DomUtils.isValidIdentifier = function (name)
{
  return /^[A-Za-z][0-9A-Z_a-z-]*$/.test(name);
};


/**
 * @constructor
 * @ignore
 */

oj.DomUtils._collapsingListenerManager = function(originalCallback, timeout)
{
  var _lastArgs = null;
  var _timerId = null;

  var _timerCallback = function()
  {
    originalCallback.apply(null, _lastArgs);
    _timerId = null;
  };

  var _callback = function()
  {
    _lastArgs = Array.prototype.slice.call(arguments);
    if (_timerId == null)
    {
      _timerId = window.setTimeout(_timerCallback, timeout);
    }
  };

  this.getCallback = function()
  {
    return _callback;
  };

  this.stop = function()
  {
    if (_timerId != null)
    {
      window.clearTimeout(_timerId);
      _timerId = null;
    }
  }

};

/**
 * @return {boolean} true if touch is supported
 */
oj.DomUtils.isTouchSupported = function ()
{
  return ('ontouchstart' in window) // C, FF, Safari, Edge
    || (navigator.msMaxTouchPoints > 0) // IE10
    || (navigator.maxTouchPoints > 0);  // IE11
};

/**
 * @ignore
 */
oj.DomUtils.setInKoCleanExternal = function(node)
{
  oj.DomUtils._koCleanNode = node;
};

/**
 * Delegates to JQuery's unwrap() if the component's node is not currently
 * being removed by Knockout
 * @param {Object} locator
 * @param {Object=} replaceLocator - locator to be replaced. I fthis parameter is ommitted,
 * the parent node will be replaced
 * @ignore
 */
oj.DomUtils.unwrap = function(locator, replaceLocator)
{
  var koCleanNode = oj.DomUtils._koCleanNode;
  
  
  if (koCleanNode)
  {
    if (locator.get(0) === koCleanNode) // skip unwrap
    {
      return;
    }
  }
  
  if (replaceLocator)
  {
    replaceLocator.replaceWith(locator); // @HtmlUpdateOk
  }
  else
  {
    locator.unwrap();
  }
  
};

/**
 * Determines if the mouse event target is on browser chrome - i.e. "scrollbar".
 * If the event is not a mouse event with a clientX and clientY, the resultant will
 * be false.
 *
 * @param {Event} event native dom event
 * @returns {boolean} <code>true</code> if the target of the mouse event is browser
 *          chrome such as scrollbars.
 * @public
 */
oj.DomUtils.isChromeEvent = function(event)
{

  /**
   * @param {Event} event
   * @return {boolean}
   */
  function _isChromeEventGecko(event) {
    // assume that if we can't access the original target of the event, then it's because
    // the target was implemented in XUL and is part of the chrome;
    try
    {
      return event.originalTarget.localName ? false : true;
    }
    catch (e)
    {
      return true;
    }
  };

  /**
   * @param {Event} event
   * @return {boolean}
   */
  function _isChromeEventIE(event)
  {
    /*
      //IE has a specific API for this but doesn't seem to want to work in automation.
      //The webkit method works in IE too.  Using that over componentFromPoint but leaving
      //the code for future reference.
      //
      var target = event.target;
      var chromePart = target.componentFromPoint(event.clientX, event.clientY);
      if (oj.StringUtils.isEmpty(chromePart))
        return false;
      else
        return true;
    */
    return _isChromeEventWebkit(event);
  };

  /**
   * @param {Event} event
   * @return {boolean}
   */
  function _isChromeEventWebkit(event)
  {
    var domTarget = event.target;
    var target = $(domTarget);


    var pos = domTarget.getBoundingClientRect();
    var sbw = oj.DomUtils.getScrollBarWidth();
    var isLTR = oj.DomUtils.getReadingDirection() === "ltr";
    if (isLTR && (("HTML" === domTarget.nodeName || "visible" !== target.css("overflow-x")) && event.clientX > (pos["right"] - sbw)))
      return true; // ltr scrollbar is always on the right
    else if (!isLTR && "HTML" === domTarget.nodeName  && event.clientX > (pos["left"] - sbw))
      return true; // RTL scrollbar on the document is still on the right
    else if (!isLTR && "visible" !== target.css("overflow-x") && event.clientX < (pos["left"] + sbw))
      return true; // RTL scrollbar not on the document is on the left
    else if (("HTML" === domTarget.nodeName || "visible" !== target.css("overflow-y")) && event.clientY > (pos["bottom"] - sbw))
      return true; // below the scrollbar
    else
      return false;
  };

  // verify event is a mouse event
  if (!('clientX' in event) || !("clientY" in event))
    return false;

  var agentInfo = oj.AgentUtils.getAgentInfo();

  if (oj.AgentUtils.OS.ANDROID === agentInfo["os"] || oj.AgentUtils.OS.IOS === agentInfo["os"])
    return false;

  if (oj.AgentUtils.ENGINE.GECKO === agentInfo["engine"])
    return _isChromeEventGecko(event);
  else if (oj.AgentUtils.ENGINE.WEBKIT === agentInfo["engine"] ||
           oj.AgentUtils.ENGINE.BLINK === agentInfo["engine"])
    return _isChromeEventWebkit(event);
  if (oj.AgentUtils.BROWSER.IE === agentInfo["browser"])
    return _isChromeEventIE(event);
  else
    return false;
};

/**
 * @returns {number} width of the browser scrollbar
 */
oj.DomUtils.getScrollBarWidth = function()
{
  var scrollBarWidth = oj.DomUtils._scrollBarWidth;
  if ($.isNumeric(scrollBarWidth))
    return scrollBarWidth;

  /** @type {jQuery} **/
  var scrollBarMeasure = $("<div />");
  $(document.body).append(scrollBarMeasure); //@HTMLUpdateOK; scrollBarMeasure constructed by the code above
  scrollBarMeasure.width(50).height(50)
    .css({
            'overflow': 'scroll',
            'visibility': 'hidden',
            'position': 'absolute'
        });

  /** @type {jQuery} **/
  var scrollBarMeasureContent = $("<div />");
  scrollBarMeasureContent.height(1);
  scrollBarMeasure.append(scrollBarMeasureContent);  //@HTMLUpdateOK; scrollBarMeasureContent constructed by the code above

  var insideWidth = scrollBarMeasureContent.width();
  var outsideWitdh = scrollBarMeasure.width();
  scrollBarMeasure.remove();

  scrollBarWidth = oj.DomUtils._scrollBarWidth = outsideWitdh - insideWidth;
  return scrollBarWidth;
};

/**
 * @returns {string!} "rtl" or "ltr"
 */
oj.DomUtils.getReadingDirection = function()
{
  var dir = document.documentElement.getAttribute("dir");
  if (dir)
    dir = dir.toLowerCase();
  return (dir === "rtl") ? "rtl" : "ltr";
};

/**
 * Retrieve the bidi independent position of the horizontal scroll position that
 * is consistent across all browsers.
 * @param {Element} elem the element to retrieve the scrollLeft from
 * @return {number} the element's scrollLeft
 */
oj.DomUtils.getScrollLeft = function(elem)
{
  if (oj.DomUtils.getReadingDirection() === "rtl")
  {
    var browser = oj.AgentUtils.getAgentInfo()['browser'];
    if (browser === oj.AgentUtils.BROWSER.FIREFOX || browser === oj.AgentUtils.BROWSER.IE || browser === oj.AgentUtils.BROWSER.EDGE)
    {
      return Math.abs(elem.scrollLeft);
    }
    else
    {
      // webkit
      return Math.max(0, elem.scrollWidth - elem.clientWidth - elem.scrollLeft);
    }
  }
  else
  {
    return elem.scrollLeft;
  }
};

/**
 * Sets the bidi independent position of the horizontal scroll position that
 * is consistent across all browsers.
 * @param {Element} elem the element to set the scrollLeft on
 * @param {number} scrollLeft the element's new scrollLeft
 */
oj.DomUtils.setScrollLeft = function(elem, scrollLeft)
{
  if (oj.DomUtils.getReadingDirection() === "rtl")
  {
    var browser = oj.AgentUtils.getAgentInfo()['browser'];
    if (browser === oj.AgentUtils.BROWSER.FIREFOX)
    {
      // see mozilla bug 383026, even though it's marked as fixed, they basically
      // did not change anything.  It still expects a negative value for RTL
      elem.scrollLeft = -scrollLeft;
    }
    else if (browser === oj.AgentUtils.BROWSER.IE || browser === oj.AgentUtils.BROWSER.EDGE)
    {
      elem.scrollLeft = scrollLeft;
    }
    else
    {
      // webkit
      elem.scrollLeft = Math.max(0, elem.scrollWidth - elem.clientWidth - scrollLeft);
    }
  }
  else
  {
    elem.scrollLeft = scrollLeft;
  }        
};

/**
 * Converts a CSS length attribute into a integer value.
 * Conversion errors or non-number will result in a zero
 * resultant.
 *
 * @param {?} cssLength style attribute
 * @return {number} value as integer
 */
oj.DomUtils.getCSSLengthAsInt = function(cssLength)
{
  if (!isNaN(cssLength))
    return parseInt(cssLength, 10);

  if (cssLength && cssLength.length > 0 && cssLength != "auto")
  {
    var intLength = parseInt(cssLength, 10);

    if (isNaN(intLength))
      intLength = 0;

    return intLength;
  }
  else
  {
    return 0;
  }
};

/**
 * Converts a CSS attribute into a float value.
 * Conversion errors or non-number will result in a zero
 * resultant.
 *
 * @param {?} cssLength style attribute
 * @return {number} value as integer
 */
oj.DomUtils.getCSSLengthAsFloat = function(cssLength)
{
  if (!isNaN(cssLength))
    return parseFloat(cssLength);

  if (cssLength && cssLength.length > 0)
  {
    var floatLength = parseFloat(cssLength);

    if (isNaN(floatLength))
      floatLength = 0;

    return floatLength;
  }
  else
  {
    return 0;
  }
};

/**
 * Key used to store the logical parent of the popup element
 * as a jQuery data property. The logical parent refers the launcher of a popup.
 * @const
 * @private
 * @type {string}
 */
oj.DomUtils._LOGICAL_PARENT_DATA = "oj-logical-parent";

/**
 * This method returns the launcher of a popup when it's open.
 * Returns undefined otherwise.
 *
 * @param {jQuery} element jquery element
 * @returns {*}
 * @see #setLogicalParent
 */
oj.DomUtils.getLogicalParent = function(element)
{
  if (element)
    return element.data(oj.DomUtils._LOGICAL_PARENT_DATA);

  return undefined;
};

/**
 * Set the logical parent as a jQuery data property
 *
 * @param {jQuery} element jquery element
 * @param {jQuery | null} parent jquery element
 * @see #getLogicalParent
 */
oj.DomUtils.setLogicalParent = function(element, parent)
{
  if (! element)
    return;

  if (parent === null)
    element.removeData(oj.DomUtils._LOGICAL_PARENT_DATA);
  else
    element.data(oj.DomUtils._LOGICAL_PARENT_DATA, parent);
};

/**
 * Checks to see if the "ancestorNode" is a logical ancestor of "node"
 *
 * @param {!Element} ancestorNode dom subtree to check to see if the target node exists
 * @param {!Element} node target node to check to see if it exists within a subtree rooted at the ancestorNode
 * @return {boolean} <code>true</code> if the "ancestorNode" is a logical ancestor of "node" or if they are the same
 */
oj.DomUtils.isLogicalAncestorOrSelf = function (ancestorNode, node)
{
  oj.Assert.assertDomElement(ancestorNode);
  oj.Assert.assertDomElement(node);

  var parentNode = node;
  while (parentNode)
  {
    if (parentNode == ancestorNode)
      return true;

    var logicalParent = oj.DomUtils.getLogicalParent($(parentNode));
    if (logicalParent)
      parentNode = logicalParent[0];
    else
      parentNode = parentNode.parentNode;
  }

  return false;

};


/**
 * Checks whether the href represents a safe URL
 * @param {!string} href - HREF to test
 * @param {Array=} whitelist - optional list of the allowed protocols. Protocol name has to use lowercase latters and
 * be followed by a ':'. If the parameter is ommitted, ['http:', 'https:'] will be used
 * @throws {Exception} an error if the HREF represents an invalid URL
 * @ignore
 */
oj.DomUtils.validateURL = function(href, whitelist)
{
  var allowed = whitelist || ['http:', 'https:'];

  var link = document.createElement('a');
  link.href = href;

  var protocol = link.protocol;
  if (protocol != null)
  {
    protocol = protocol.toLowerCase();
  }

  if (allowed.indexOf(protocol) < 0)
  {
    throw protocol + " is not a valid URL protocol";
  }
};

/**
 * Cancels native context menu events for hybrid mobile applications.
 * @private
 */
oj.DomUtils._supressNativeContextMenu = function()
{
  if ($(document.body).hasClass("oj-hybrid")) {
    document.body.addEventListener('contextmenu',
      function(event) {
        event.preventDefault();
      }, true);
  }
}
oj.DomUtils._supressNativeContextMenu();

// standard duration of a pressHold gesture.  Point of reference: default
// JQ Mobile threshold to be a press-and-hold is 750ms.
oj.DomUtils.PRESS_HOLD_THRESHOLD = 750;


// ------------------------------------------------------------------------------------------------
// Recent touch end
// ------------------------------------------------------------------------------------------------

/**
 * Returns true if a touchend or touchcancel has been detected anywhere in the document in the last 500 ms.
 * Note: This function adds event listeners only once per document load.
 *
 * @return {boolean} boolean indicating whether a touch has recently been detected
 */
oj.DomUtils.recentTouchEnd = (function()
{
  // This function is immediately executed and returns the recentTouchEnd function
  // and therefore only execute once per document load.

  var touchTimestamp = 0;
  var TOUCH_THRESHOLD = 500;

  function _touchEndHandler() {
    touchTimestamp = Date.now();
  };

  // --- Document listeners ---
  document.addEventListener("touchend", _touchEndHandler, true);
  document.addEventListener("touchcancel", _touchEndHandler, true);

  // --- The function assigned to oj.DomUtils.recentTouchEnd ---

  return function()
  {
    // must be at least 300 for the "300ms" delay
    return (Date.now() - touchTimestamp) < TOUCH_THRESHOLD;
  };
})();

/**
 * Returns true if a touchstart has been detected anywhere in the document in the last 800 ms.
 * Note: This function adds event listeners only once per document load.
 *
 * @return {boolean} boolean indicating whether a touch has recently been detected
 */
oj.DomUtils.recentTouchStart = (function()
{
  // This function is immediately executed and returns the recentTouchStart function
  // and therefore only execute once per document load.

  var touchTimestamp = 0;
  // 800 because this is used to ignore mouseenter and focusin on 'press', and a 'press'
  // is usually detected after 750ms.
  var TOUCH_THRESHOLD = oj.DomUtils.PRESS_HOLD_THRESHOLD + 50;

  function _touchStartHandler() {
    touchTimestamp = Date.now();
  };

  // --- Document listeners ---
  document.addEventListener("touchstart", _touchStartHandler, true);

  // --- The function assigned to oj.DomUtils.recentTouchStart ---

  return function()
  {
    // must be at least TOUCH_THRESHOLD for the  delay
    return (Date.now() - touchTimestamp) < TOUCH_THRESHOLD;
  };
})();

// ------------------------------------------------------------------------------------------------
// Recent pointer
// ------------------------------------------------------------------------------------------------

/**
 * Returns true if a touchstart, touchend, mousedown, or mouseup has been detected anywhere in the
 * document in the last n ms, where n is calibrated across a variety of platforms to make this API
 * a maximally reliable indicator of whether the code now running was likely "caused by" the
 * specified touch and mouse interaction, vs. some other thing (e.g. mousemove, keyboard, or page
 * load).  E.g. the makeFocusable() / _focusable() mechanism uses this API to vary the focus theming
 * depending on whether the element was focused via keyboard or pointer.
 *
 * @return {boolean} boolean indicating whether a mouse button or finger has recently been down or up
 */
oj.DomUtils.recentPointer = (function()
{
  // The comments in this function are tailored to the makeFocusable() usage.

  // - Let "pointer down" mean mousedown or touchstart, and "pointer up" likewise.  (Not MS pointer events.)
  // - Event order can be 1) mousedown>focus>mouseup (like push buttons) or 2) mousedown>mouseup>focus (like toggle buttons).
  // - For 2, semantics for "focus caused by pointer" must be "if pointer interaction in last n ms," rather than "if pointer is currently down".
  // - Those "last n ms" semantics are preferred for 1 as well, rather than relying on pointer up to cancel a state set by pointer down,
  //   since if the pointer up is never received, we'd get stuck in an inaccessible state.
  // - So both pointer down and pointer up set a timestamp, and recentPointer() returns true if Date.now() is within n ms of that timestamp,
  //   where n is higher for touchstart per below.

  // Timestamp of last mousedown/up or touchstart/end. Initial value of 0 (1/1/1970) guarantees that if element is focused before any
  // mouse/touch interaction, then recentPointer() is false, so focus ring appears as desired.
  var pointerTimestamp = 0;

  var pointerTimestampIsTouchStart; // whether the latest timestamp is for touchstart vs. touchend/mouse

  // On Edge (Surface Win10), the lag from the up event to resulting programmatic focus is routinely ~350ms, even when the 300ms "tap delay" has
  // been prevented and confirmed to be absent.  (In Chrome on same device the same lag is ~10 ms.)  So use 600ms to be safe.  Even on Chrome,
  // the lag from the down/up event to natively induced focus can routinely be well into the 1xx ms range. Can exceed 600 if needed. There is no
  // need for a tight bound; if there was pointer interaction in the last second or so, it's perfectly reasonable to suppress the focus ring.
  var POINTER_THRESHOLD_CUSHION = 600;

  // If the number of millis since the last pointer down or up is < this threshold, then recentPointer() considers it recent and returns true.
  // See also TOUCHSTART_THRESHOLD.
  var POINTER_THRESHOLD = POINTER_THRESHOLD_CUSHION;

  // For touchstart only, use 750+600ms so that focus set by a 750ms pressHold gesture (e.g. context menu) is recognized as touch-related.  Same
  // 600ms padding as for POINTER_THRESHOLD.  A high threshold is OK, as it is used only for actual pressHolds (and the unusual case where the
  // pointer up is never received), since for normal clicks and taps, the pointerUp replaces the "1350ms after touchstart" policy with a "600ms
  // after pointerUp" policy. On Edge and desktop FF (desktop version runs on hybrid devices like Surface), which lack touchstart, context menus
  // are launched by the contextmenu event, which happen after the pointer up in both browsers, so the fact that we're using the higher
  // threshold only for touchstart should not be a problem there.
  var TOUCHSTART_THRESHOLD = oj.DomUtils.PRESS_HOLD_THRESHOLD + POINTER_THRESHOLD_CUSHION;


  // --- Document listeners ---

  // Use capture phase to make sure we hear the events before someone cancels them
  document.addEventListener("mousedown", function() {
    // If the mousedown immediately follows a touchstart, i.e. if it seems to be the compatibility mousedown
    // corresponding to the touchstart, then we want to consider it a "recent pointer activity" until the end time
    // that is max(touchstartTime + TOUCHSTART_THRESHOLD, now + POINTER_THRESHOLD), where now is mousedownTime in this
    // case.  (I.e. it would defeat the purpose if the inevitable mousedown replaced the longer touchstart threshold with
    // a shorter one.)  We don't do this in the touchend/mouseup listeners, as those obviously happen after the pressHold
    // is over, in which case the following analysis applies:
    // - If the pressHold was < PRESS_HOLD_THRESHOLD ms,
    // - then the higher TOUCHSTART_THRESHOLD is not needed or relevant, since anything focused on pressHold
    //   (like a context menu) never happened,
    // - else the touchend/mouseup happened > PRESS_HOLD_THRESHOLD ms after the touchstart, so in the max() above,
    //   the 2nd quantity is always bigger (later).
    var now = Date.now();
    if ((!pointerTimestampIsTouchStart) || (now > pointerTimestamp + oj.DomUtils.PRESS_HOLD_THRESHOLD)) {
      pointerTimestamp = now;
      pointerTimestampIsTouchStart = false;
    }
  }, true);

  document.addEventListener("touchstart", function() {
    pointerTimestamp = Date.now();
    pointerTimestampIsTouchStart = true;
  }, true);

  document.addEventListener("mouseup", function() {
    pointerTimestamp = Date.now();
    pointerTimestampIsTouchStart = false;
  }, true);

  document.addEventListener("touchend", function() {
    pointerTimestamp = Date.now();
    pointerTimestampIsTouchStart = false;
  }, true);


  // --- The function assigned to oj.DomUtils.recentPointer ---

  return function()
  {
    var millisSincePointer = Date.now() - pointerTimestamp;
    var threshold = pointerTimestampIsTouchStart ? TOUCHSTART_THRESHOLD : POINTER_THRESHOLD;
    var isRecent = millisSincePointer < threshold;
    return isRecent;
  };
})();


// ------------------------------------------------------------------------------------------------
// Utility for suppressing focus ring for mouse/touch interaction, but not KB or other interaction:
// ------------------------------------------------------------------------------------------------

/**
 * This API works like baseComponent's _focusable() API (see its detailed JSDoc), with the
 * similarities and differences listed below.  This API is intended for non-component callers;
 * components should typically call the baseComponent API via this._focusable().
 *
 * Comparison to baseComponent._focusable() :
 *
 * - This function's "options" param must be an object.  Only baseComponent._focusable()
 *   supports the backward-compatibility syntax where the options param can be the element.
 * - Same usage of oj-focus, oj-focus-highlight, and $focusHighlightPolicy.
 * - Same required invariant that oj-focus-highlight must not be set if oj-focus is not set.
 * - Same parameters with same semantics, plus the additional "component" and "remove" params
 *   discussed below.
 * - New options.component param, which takes a JET component instance.  (When a component is
 *   involved, typically that component should call this._focusable() rather than calling this
 *   version of the method directly.)
 *
 * If options.component is specified, then the following things work like the baseComponent
 * version of this API:
 *
 * - If the specified element is in the component subtree,
 *   then the classes will automatically be removed when the component is
 *   destroyed/disabled/detached, as detailed in the baseComponent JSDoc,
 *   else the caller has the same responsibility to remove the classes at those times.
 * - Same rules as to whether listeners are automatically cleaned up, or suppressed when the
 *   component is disabled, vs. being the caller's responsibility to handle those things.
 *
 * If options.component is NOT specified (for non-component callers), then those things are
 * the caller's responsibility.  Specifically:
 *
 * - Class removal can be done directly, as needed.
 * - To remove the listeners, see the following.
 *
 * Listener removal:
 *
 * - If options.component was specified, see above.
 * - Else if options.setupHandlers was specified, then only the caller knows what listeners were
 *   registered and how, so it is the caller's responsibility to remove them directly when needed.
 * - The remaining case is that options.component and options.setupHandlers were not specified.
 *   To remove from element e both the 2 classes and all listeners applied to e by all previous
 *   invocations of makeFocusable() where these options were not specified,
 *   call makeFocusable( {'element': e, 'remove': true} ).
 */
// If this is named focusable(), Closure Compiler generates a warning, and fails to rename the function in minified code,
// which suggests that focusable (not just _focusable) is apparently externed somewhere (although not in
// 3rdparty\jquery\externs\jquery-1.8.js, main\javascript\externs.js, or build\tools\closure\compiler.jar\externs.zip\),
// perhaps for JQUI's :focusable selector.  So name it makeFocusable().
oj.DomUtils.makeFocusable = (function()
{
  var nextId = 0; // used for unique namespace, for "remove" functionality

  // This private var is shared by all callers that use makeFocusable() and don't supply their own focus highlight policy.
  // If the oj-focus-config SASS object ever acquires a 2nd field, should continue to call pJFFF() only once, statically.
  var FOCUS_HIGHLIGHT_POLICY = (oj.ThemeUtils.parseJSONFromFontFamily('oj-focus-config') || {})['focusHighlightPolicy'];

  /**
   * @param {function()} focusPolicyCallback Optional getter passed to makeFocusable() by callers wishing to get use a caller-
   *   specific focus policy mechanism instead of the built-in mechanism.
   * @param {function()} recentPointerCallback Optional function passed to makeFocusable() by callers wishing to use a caller-
   *   specific mechanism in addition to the built-in mechanism.
   * @return {boolean} boolean indicating whether it is appropriate to apply the <code class="prettyprint">oj-focus-highlight</code>
   *   CSS class for a focus happening at the time of this method call.
   */
  var shouldApplyFocusHighlight = function(focusPolicyCallback, recentPointerCallback)
  {
    var focusHighlightPolicy = focusPolicyCallback ? focusPolicyCallback() : FOCUS_HIGHLIGHT_POLICY;
    switch (focusHighlightPolicy)
    {
      case "all":
        return true;
      case "none":
        return false;
      default: // "nonPointer" or no value provided (e.g. SASS var missing)
        return !( oj.DomUtils.recentPointer() || (recentPointerCallback && recentPointerCallback()) );
    }
  };

  // the function assigned to oj.DomUtils.makeFocusable
  var makeFocusable = function( options )
  {
    var element = options['element'];

    var dataKey = "ojFocusable";
    var namespacePrefix = "." + dataKey;
    var namespaceSeparator = " " + namespacePrefix;

    if (options['remove']) {
      element.removeClass("oj-focus oj-focus-highlight");

      // id's of listeners needing removal
      var ids = element.data(dataKey);
      if (ids == undefined)
          return;

      // map ids to namespaces.  "2" -> ".ojFocusable2".  "2,7" -> ".ojFocusable2 .ojFocusable7"
      var namespaces = namespacePrefix + (""+ids).split(",").join(namespaceSeparator);
      element.off(namespaces) // remove the listeners
             .removeData(dataKey); // clear list of listener id's needing removal
      return;
    }

    var afterToggle = options['afterToggle'] || $.noop;

    var applyOnlyFocus = function( element ) {
      element.addClass( "oj-focus" );
      afterToggle("focusin");
    };

    var applyBothClasses = function( element ) {
      element.addClass( "oj-focus" );
      if (shouldApplyFocusHighlight(options['getFocusHighlightPolicy'], options['recentPointer'])) {
        element.addClass( "oj-focus-highlight" );
      }
      afterToggle("focusin");
    };

    var addClasses = options['applyHighlight'] ? applyBothClasses : applyOnlyFocus;

    var removeClasses = function( element ) {
      element.removeClass( "oj-focus oj-focus-highlight" );
      afterToggle("focusout");
    };

    var setupHandlers = options['setupHandlers'] || function( focusInHandler, focusOutHandler) {
      var component = options['component'];
      var focusInListener = function( event ) {
        focusInHandler($( event.currentTarget ));
      };
      var focusOutListener = function( event ) {
        focusOutHandler($( event.currentTarget ));
      }

      if (component) {
        component._on( element, {
          focusin: focusInListener,
          focusout: focusOutListener
        });
      } else {
        // neither options.component nor options.setupHandlers were passed, so we must provide a
        // way for the caller to remove the listeners.  That's done via the "remove" param, which
        // uses the namespaces that we stash via data().
        var id = nextId++;

        // list of id's of existing listeners needing removal
        var ids = element.data(dataKey);

        // append id to that list, or start new list if first one
        element.data(dataKey, ids == undefined ? id : ids + "," + id);

        // add listeners namespaced by that id
        var handlers = {};
        var namespace = namespacePrefix + id;
        handlers["focusin" + namespace] = focusInListener;
        handlers["focusout" + namespace] = focusOutListener;
        element.on(handlers);
      }
    };

    setupHandlers(addClasses, removeClasses);
  };

  return makeFocusable;
})();
/*jslint browser: true*/
/*
** Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
*/
/**
 * Focus utilities.
 * @ignore
 */
oj.FocusUtils = {};

oj.FocusUtils._TABBABLE = ":tabbable";

//These functions inspired by AdfFocusUtils

/**
 * Tests whether the specified element contains the keyboard focus.
 * @param {!Element} element Element for which to check if it contains focus.
 * @returns {boolean} True if the element contains focus, false otherwise.
 */
oj.FocusUtils.containsFocus = function(element)
{
  var activeElem = document.activeElement;
  //FIX : if either elem is undefined, just return false
  if (!element || !activeElem)
    return false;
  
  return oj.DomUtils.isAncestorOrSelf(element, activeElem);
};

/**
 * Sets focus to the specified element.
 * @param {!Element} element Element to focus.
 */
oj.FocusUtils.focusElement = function(element)
{
  element.focus();
};

/**
 * Sets focus to the first tabbable element inside the given element, which
 * may be the given element itself.
 * @param {!Element} element Element to start searching for a tabbable element in.
 * @returns {Element} The DOM element that was focused, if any.
 */
oj.FocusUtils.focusFirstTabStop = function(element)
{
  var focusElement = oj.FocusUtils.getFirstTabStop(element);

  if (focusElement)
    oj.FocusUtils.focusElement(focusElement);

  return focusElement;
};

/**
 * Get the first tabbable element inside the given element, which may be the
 * given element itself.
 * @param {!Element} element Element to start searching for a tabbable element in.
 * @returns {Element} The first tabbable element inside the given element.
 */
oj.FocusUtils.getFirstTabStop = function(element)
{
  var jqElem = $(element);
  if (jqElem.is(oj.FocusUtils._TABBABLE))
    return element;
  
  var jqFocusables = jqElem.find(oj.FocusUtils._TABBABLE);
  if (jqFocusables && jqFocusables.length > 0)
    return jqFocusables[0];

  return null;
};

/**
 * Extends the jquery ":focusable" pseudo selector check for a Safari browser specific
 * exception - an anchor element not having a tabindex attribute.
 *
 * @param {Element} element target dom element to test if it will take focus
 * @returns {boolean} <code>true</code> if the target element is focusable
 */
oj.FocusUtils.isFocusable = function (element)
{
  if ($(element).is(":focusable"))
  {
    // An anchor element in safari will not take focus unless it has a tabindex.
    // http://developer.mozilla.org/en-US/docs/Web/HTML/Element/a#Clicking_and_focus
    if (element.nodeName === "A" && !element.hasAttribute("tabindex") &&
      oj.AgentUtils.getAgentInfo()["browser"] === oj.AgentUtils.BROWSER.SAFARI)
      return false;
    else
      return true;
  }

  return false;
};
/*jslint browser: true*/
/*
** Copyright (c) 2017, Oracle and/or its affiliates. All rights reserved.
*/
/**
 * Gesture utilities provided internally for JET components, currently only context menu gesture are available.
 * Moved from ojcomponentcore and made into static methods.
 * @ignore
 */
oj.GestureUtils = {};

/**
 * Event namespace used by context menu internal event registration.
 * Previously we got the namespace from the widget.
 */
oj.GestureUtils._EVENT_NAMESPACE = ".contextMenu";

/**
 * Utility method to tear down any artifacts created by GestureUtils.startDetectContextMenuGesture
 * @param {Element} rootNode the root element of the component
 */
oj.GestureUtils.stopDetectContextMenuGesture = function(rootNode)
{
    $(rootNode).off( oj.GestureUtils._EVENT_NAMESPACE )
               .removeClass("oj-menu-context-menu-launcher")
        [0].removeEventListener("click", rootNode._clickListener, true);

    // the other 2 contextMenu timeouts don't need to be cleared here
    clearTimeout(rootNode._contextMenuPressHoldTimer);
};

/**
 * Utility method to setup context menu gesture detection on a component
 * @param {Element} rootNode the root node of the component
 * @param {function(Event, string)} callback callback to invoke on the component when context menu gesture is detected
 */
oj.GestureUtils.startDetectContextMenuGesture = function(rootNode, callback)
{
    // Note: Whether or not we use Hammer to detect press-hold, this code would need to do the following things seen below:
    //
    // (1) Prevent the compatibility mousedown event from triggering Menu's clickAway logic.
    // (2) Prevent press-hold from also generating a click (unless Hammer does this automatically; I'm guessing it doesn't).
    // (3) Ensure we don't respond to *both* press-hold and contextmenu events on Android.
    //
    // So the only thing that Hammer would replace is:
    //
    // (4) Detecting the press-hold.
    //
    // Not currently using Hammer for (4), since:
    //
    // - This code predates Hammer, and was already stable after extensive iteration / fine-tuning.
    // - We use the same listeners for parts of 1-4. If moved 4 off to Hammer (separate listener), just need to ensure that
    //   we don't introduce any race conditions, etc.  (May be easy or hard, just need to look.)
    // - Hammer only wants to have one instance per DOM node, else they fight to control some things like touch-action. So
    //   a prereq for having this baseComponent logic put Hammer on components is to work out a protocol for super- and sub-
    //   classes to share the same instance and not step on each other.  Not insurmountable; just need to have the conversation.
    //   Tracked by ER 21357133, which links to detailed wiki.

    var pressHoldThreshold = oj.DomUtils.PRESS_HOLD_THRESHOLD; // launch CM at 750ms per UX spec
    var isPressHold = false; // to prevent pressHold from generating a click
    var contextMenuPressHoldTimer;

    var touchInProgress = false;

    // 5px is Hammer default.  (Didn't check whether they apply that separately to x and y like us, or to the hypotenuse,
    // but it's within a couple px either way -- think 3-4-5 triangle.)
    var maxAllowedMovement = 5;
    var touchPageX;
    var touchPageY;

    var doubleOpenTimer; // to prevent double open.  see usage below.
    var doubleOpenThreshold = 300; // made up this number.  TBD: Tweak as needed to make all platforms happy.
    var doubleOpenType = null; // "touchstart" or "contextmenu"

    var namespace = oj.GestureUtils._EVENT_NAMESPACE;

    var contextMenuPressHoldJustEnded = false;
    var launch = function(event, eventType, pressHold) 
    {
        // ensure that pressHold doesn't result in a click.  Set this before the bailouts below.
        isPressHold = pressHold;

        // In Mobile Safari only, mousedown fires *after* the touchend, which causes at least 2 problems:
        // 1) CM launches after 750ms (good), then disappears when lift finger (bad), because touchend -->
        // mousedown, which calls Menu's "clikAway" mousedown listener, which dismisses Menu.
        // 2) The isPressHold logic needs to reset the isPressHold ivar on any event that can start a click,
        // including mousedown.  This problem causes the mousedown listener to incorrectly clear the ivar
        // after a pressHold, which broke the whole mechanism.
        // SOLUTION FOR 1-2:  On each launch (at 750ms), set a one-time touchend listener that will set a
        // var and clear it 50ms later.  While the var is set, both mousedown listeners can disregard the
        // mousedown.  Make the var a static var in Menu, since Menu's listener is static, and since this
        // launcher component can get/set it via an (effectively static) menu method.
        // NON-SOLUTIONS:  Cancelling touchstart or touchend, via pD() and sP(), doesn't cancel iPad's mousedown.
        // Cancelling mousedown from here doesn't work even if capture phase, since ojMenu's listener is capture phase.
        // TIMING: The following block should be before the doubleOpen bailout.
        if (isPressHold)
        {
            $(rootNode).one( "touchend" + namespace, function( event ) {
                var touchendMousedownThreshold = 50; // 50ms.  Make as small as possible to prevent unwanted side effects.
                contextMenuPressHoldJustEnded = true;
                setTimeout(function() {
                    contextMenuPressHoldJustEnded = false;
                }, touchendMousedownThreshold);
            });
        }

        // On platforms like Android Chrome where long presses already fire the contextmenu event, the pressHold
        // logic causes the menu to open twice, once for the pressHold, once for the contextmenu.  There's no
        // guarantee which will happen first, but as long as they happen within doubleOpenThreshold ms
        // of each other, this logic should prevent the double open.
        // Note: Another option is a platform-specific solution where we only use pressHold for platforms that need
        // it (that don't already fire a contextmenu event for pressHold), but architectural preference is to avoid
        // platform-specific solutions if possible.
        if ((doubleOpenType === "touchstart" && event.type === "contextmenu")
                || (doubleOpenType === "contextmenu" && event.type === "touchstart"))
        {
            doubleOpenType = null;
            clearTimeout(doubleOpenTimer);
            return;
        }

        // If a nested element or component already showed a JET context menu for this event, don't replace it with ours.
        // Hack: must check defaultPrevented on the nested event too, because for touchstart events on iOS7 at least, when
        // the outer component reaches this point, event is a different JQ wrapper event than the one on which the inner
        // component previously called preventDefault, although they both wrap the same native originalEvent.  The new wrapper
        // never had its isDefaultPrevented field set to the returnTrue method, so must check the nested originalEvent.
        // This never seems to happen with right-click and Shift-F10 events.  Has nothing to do with the setTimeout: the events
        // received by the rootNode.on("touchstart"...) code are different (firstWrapper==secondWrapper returns false).
        // TODO: link to JQ bug once filed.
        if (event.isDefaultPrevented() || (event.originalEvent && event.originalEvent.defaultPrevented))
            return;

        callback(event, eventType);

        // if _NotifyContextMenuGesture() (or subclass override of it) actually opened the CM, and if that launch wasn't
        // cancelled by a beforeOpen listener...
        if (event.isDefaultPrevented())
        {
            // see double-open comments above
            if (event.type === "touchstart" || event.type === "contextmenu")
            {
                doubleOpenType = event.type;
                doubleOpenTimer = setTimeout(function(){
                    doubleOpenType = null;
                }, doubleOpenThreshold);
            }
        }
    };

    // At least some of the time, the pressHold gesture also fires a click event same as a short tap.  Prevent that here.
    var _clickListener = function( event ) {
        if (isPressHold) {
            // For Mobile Safari capture phase at least, returning false doesn't work; must use pD() and sP() explicitly.
            event.preventDefault();
            event.stopPropagation();
            isPressHold = false;
        }
    };

    rootNode._clickListener = _clickListener;

    // Use capture phase to make sure we cancel it before any regular bubble listeners hear it.
    rootNode.addEventListener("click", _clickListener, true);

    $(rootNode)
        .on( "touchstart" + namespace + " " +
             "mousedown" + namespace + " " +
             "keydown" + namespace + " ", function( event ) {
            // for mousedown-after-touchend Mobile Safari issue explained above where __contextMenuPressHoldJustEnded is set.
            if (event.type === "mousedown" && contextMenuPressHoldJustEnded)
                return;

            // reset isPressHold flag for all events that can start a click.
            isPressHold = false;

            // start a pressHold timer on touchstart.  If not cancelled before 750ms by touchend/etc., will launch the CM.
            // isolate the context menu long tap to a single touch point.
            if (event.type === "touchstart" && event.originalEvent["touches"].length === 1) {
                // note starting position so touchmove handler can tell if touch moved too much
                var firstTouch = event.originalEvent["touches"][0];
                touchPageX = firstTouch["pageX"];
                touchPageY = firstTouch["pageY"];

                touchInProgress = true;
                contextMenuPressHoldTimer = setTimeout(launch.bind(undefined, event, "touch", true), pressHoldThreshold);
                rootNode._contextMenuPressHoldTimer = contextMenuPressHoldTimer;
            }

            return true;
        })

        // if the touch moves too much, it's not a pressHold
        .on( "touchmove" + namespace, function( event ) {
            var firstTouch = event.originalEvent["touches"][0];
            if (Math.abs(touchPageX - firstTouch["pageX"]) > maxAllowedMovement
                || Math.abs(touchPageY - firstTouch["pageY"]) > maxAllowedMovement)
            {
                touchInProgress = false;
                clearTimeout(contextMenuPressHoldTimer);
            }
            return true;
        })

        // if the touch ends before the 750ms is up, it's not a long enough pressHold to show the CM
        .on( "touchend" + namespace + " " +
             "touchcancel" + namespace, function( event ) {
            touchInProgress = false;
            clearTimeout(contextMenuPressHoldTimer);
            return true;
        })
        .on( "keydown" + namespace + " " +
             "contextmenu" + namespace, function( event ) {
            if (event.type === "contextmenu" // right-click.  pressHold for Android but not iOS
                    || (event.which == 121 && event.shiftKey)) // Shift-F10
            {
                var eventType = touchInProgress ? "touch" : event.type === "keydown" ? "keyboard" : "mouse";
                launch(event, eventType, false);
            }

            return true;
        });

    // Does 2 things:
    // 1) Prevents native context menu / callout from appearing in Mobile Safari.  E.g. for links, native CM has "Open in New Tab".
    // 2) In Mobile Safari and Android Chrome, prevents pressHold from selecting the text and showing the selection handles and (in Safari) the Copy/Define callout.
    // In UX discussion, we decided to prevent both of these things for all JET components for now.  If problems, can always, say, add protected method allowing
    // subclass to opt out (e.g. if they need 1 and/or 2 to work).
    // Per discussion with architects, do #2 only for touch devices, so that text selection isn't prevented on desktop.  Since #1
    // is a no-op for non-touch, we can accomplish this by omitting the entire style class, which does 1 and 2, for non-touch.
    // Per comments in scss file, the suppression of 1 and 2 has issues in old versions of Mobile Safari.
    if (oj.DomUtils.isTouchSupported())
        $(rootNode).addClass("oj-menu-context-menu-launcher");
};
/*jslint browser: true*/
/**
 * in some OS/browser combinations you can attempt to detect high contrast mode
 * in javascript, go to the url below and look for "High Contrast"
 * http://www.w3.org/TR/wai-aria-practices/
 * 
 * This function uses a variation of the code in the "High Contrast" section of  
 * the site above to try and detect high contrast mode
 * by script, but it by no means definitively tells you whether or not you
 * are actually in high contrast mode. As discussed at the url above you 
 * may need to have a user preference setting for high contrast.
 * 
 * If the script is able to detect high contrast mode it sets the class 
 * "oj-hicontrast" on the body tag. When "oj-high-contrast" is present 
 * JET provides alternate informational images that are specially designed 
 * for high contrast users. 
 * @private
 */
function _ojHighContrast()
{
  // using a data uri, I googled for shortest uri to get this one since 
  // I don't care about the actual image, but I do want a legit image
  // otherwise I see an error in chrome and I don't want users to be
  // confused by seeing any error.

  var div = $("<div style='border: 1px solid;border-color:red green;position: absolute;top: -999px;background-image: url(data:image/gif;base64,R0lGODlhAQABAIAAAP///wAAACwAAAAAAQABAAACAkQBADs=);'></div>"), bki;
  div.appendTo("body");     // @HTMLUpdateOK safe manipulation

  bki = div.css("backgroundImage");
  //console.log("background-image:" + bki);
  //console.log("borderTopColor == borderRightColor: ", div.css("borderTopColor") == div.css("borderRightColor"));
  if (div.css("borderTopColor") == div.css("borderRightColor") ||
      (bki != null && (bki == 'none' || bki == 'url (invalid-url:)')))
  {
    $('body').addClass("oj-hicontrast");
  }

  div.remove();
}

$(document).ready(function() {
  _ojHighContrast();
});
// Copyright (c) 2013, Oracle and/or its affiliates.
// All rights reserved.

/*jslint browser: true*/

/**
 * @export
 * @class
 * @since 1.0
 * @classdesc Common test support in JavaScript
 */
oj.Test = {};


/**
 * A global application flag that can be set by a test to indicate that all page startup processing is done
 * and an external automated test can begin
 * @export
 * @type {boolean}
 */
oj.Test.ready = false;

/**
 * @export
 * Return the node found given the locator
 * @param {Object|string} locator A locator which is either a JSON string (to be parsed using $.parseJSON), or an Object with the following properties:
 *                                             element: the component's selector, determined by the test author when laying out the page
 *                                             subId: the string, documented by the component, that the component expects in getNodeBySubId to locate a particular subcomponent
 *  @returns {*} the subcomponent located by the subId string passed in locator, if found.
 */
oj.Test.domNodeForLocator = function (locator) {
  var locObj = locator;
  if (oj.StringUtils.isString(locator)) {
    var locStr = /** @type {string} */ (locator);
    try {
      locObj = JSON.parse(locStr);
    }
    catch (e) {
      return null;
    }
  }
  if (locObj && locObj['element']) {
    var element = $(locObj['element']);
    if (element) {
      delete locObj['element'];
      var id = /** @type {Object} */ (locObj);
      return oj.Components.getNodeBySubId(element[0], id);
    }
  }
  return null;
};

/**
 * @return {number} total number of open popups
 * @export
 * @since 1.1.0
 */
oj.Test.getOpenPopupCount = function ()
{
  return oj.ZOrderUtils.getOpenPopupCount();
};

/**
 * Returns a jQuery set of popup root elements that are open and actively
 * managed by the popup framework.
 *
 * @return {!jQuery}
 * @export
 * @since 1.1.0
 */
oj.Test.findOpenPopups = function ()
{
  return oj.ZOrderUtils.findOpenPopups();
};

/**
 * Utility used for testing. Compares two jQuery singleton wappered elements
 * determining which element has the greatest stacking context.
 *
 * @export
 * @param {jQuery} el1 first element to compare
 * @param {jQuery} el2 second element to compare
 * @return {number} 0 if elements have the same stacking context;
 *                  1 if the first element has a greater stacking context;
 *                 -1 when the second element has a greater stacking context;
 * @since 1.1.0
 */
oj.Test.compareStackingContexts = function (el1, el2)
{
  return oj.ZOrderUtils.compareStackingContexts(el1, el2);
};
// override jQuery's cleanData method to bypass cleanup of custom elements and composites
$['cleanData'] = (function(orig) 
{
  return function(elems) 
  {
    var nonCustomElements = [];
    var i, elem
    for (i = 0; (elem = elems[i]) != null; i++ ) 
    {
      // Skip cleaning any elements that are custom elements or are created by a custom element
      var bSkip = false;
      var constr = oj.Components.__GetWidgetConstructor(elem);
      if (constr) 
      {
        bSkip = constr('instance')._IsCustomElement();
        if (!bSkip)
        {
          var parent = oj.Components.getComponentElementByNode(elem);
          bSkip = parent && oj.BaseCustomElementBridge.getRegistered(parent.tagName);
        }
      }
      if (!bSkip)
        nonCustomElements.push(elem);
    }
    if (nonCustomElements.length > 0)
      orig(nonCustomElements);
  };
})($['cleanData']);
(function() {
var baseComponentMeta = {
  "properties": {
    "translations": {
      "type": "Object"
    }
  },
  "methods": {
    "getNodeBySubId": {},
    "getSubIdByNode": {},
    "refresh": {}
  },
  "extension": {
    _WIDGET_NAME: "baseComponent"
  }
};
oj.CustomElementBridge.registerMetadata('baseComponent', null, baseComponentMeta);
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */


/**
 * @class oj.EditableValueUtils
 * @classdesc JET Editable Component Utils
 * @export
 * @since 0.6
 * @hideconstructor
 * @ignore
 * 
 */
oj.EditableValueUtils = {};

// S T A T I C    V A R S   
/**
 * The various contexts under which validation can be run by component.
 * @ignore
 */ 
oj.EditableValueUtils.validationContext = {
  COMPONENT_CREATE : 1,
  CONVERTER_OPTION_CHANGE : 2,
  DISABLED_OPTION_CHANGE : 3,
  READONLY_OPTION_CHANGE: 4,
  REFRESH_METHOD: 5,
  REQUIRED_OPTION_CHANGE : 6,
  RESET_METHOD : 7,
  USER_ACTION: 8,
  VALIDATE_METHOD: 9,
  VALIDATORS_OPTION_CHANGE : 10,
  VALUE_OPTION_CHANGE: 11
};

/**
 * Default validation options used by validate method.
 * @ignore
 */ 
oj.EditableValueUtils.validateMethodOptions = {doValueChangeCheck: false,
                               validationContext: oj.EditableValueUtils.validationContext.VALIDATE_METHOD};

/**
 * Default validation options used when converter option changes
 * @ignore
 */
oj.EditableValueUtils.converterOptionOptions = {doNotClearMessages: true,
                                validationContext: oj.EditableValueUtils.validationContext.CONVERTER_OPTION_CHANGE};
                              
/**
 * Default validation options used when disabled option changes
 * @ignore
 */
oj.EditableValueUtils.disabledOptionOptions = {doNotClearMessages: true,
                               validationContext: oj.EditableValueUtils.validationContext.DISABLED_OPTION_CHANGE};

/**
 * Default validation options used when required option changes
 * @ignore
 */
oj.EditableValueUtils.requiredOptionOptions = {doNotClearMessages: true,
                               validationContext: oj.EditableValueUtils.validationContext.REQUIRED_OPTION_CHANGE};

/**
 * Default validation options used when readOnly option changes
 * @ignore
 */
oj.EditableValueUtils.readOnlyOptionOptions = {doNotClearMessages: true,
                               validationContext: oj.EditableValueUtils.validationContext.READONLY_OPTION_CHANGE};

/**
 * Default validation options used when refresh method is called.
 * @ignore
 */
oj.EditableValueUtils.refreshMethodOptions = {doNotClearMessages: true,
                               validationContext: oj.EditableValueUtils.validationContext.REFRESH_METHOD};
/**
 * Default validation options used when validators option changes
 * @ignore
 *  */
oj.EditableValueUtils.validatorsOptionOptions = {doNotClearMessages: true,
                                 validationContext: oj.EditableValueUtils.validationContext.VALIDATORS_OPTION_CHANGE};                      

/**
* String used in the id on the span that surrounds the help icon.
* @const
* @private
* @ignore
* @type {string}
*/
var _REQUIRED_ICON_ID = "_requiredIcon";

/**
* String used in the label element's id for custom &lt;oj-label>
* @const
* @ignore
* @type {string}
*/
oj.EditableValueUtils.CUSTOM_LABEL_ELEMENT_ID = "|label";

/**
* Enum for validate() return values
* @const
* @ignore
*/
oj.EditableValueUtils.VALIDATE_VALUES = {
  VALID : "valid",
  INVALID : "invalid"
};

/**
 * This method is called during _InitOptions() to initialize a component option value from DOM. This 
 * uusally is the case when the option value is undefined, 
 * i.e., this.options.optionName === undefined.
 * <br/>
 * Returns the attribute value for the given attribute on the element appropriately converted, or 
 * the default, if the attribute isn't set on the element.<br/>
 * 
 * @param {Object} element the element the component is initialized with.
 * @param {string} attribute the name of the element's attribtue. Example, value, disabled etc.
 * 
 * @returns {Object} a JSON object containing the following properties - <br/>
 * <ul>
 * <li><b>fromDom</b> - whether the option value was initialized from DOM. When true the option's 
 * value is written back (to observable).</li> 
 * <li><b>value</b> - the option value. the attribute value or the default if the attribute isn't 
 * set on the element.</li> 
 * </ul>
 * 
 * @private
 */
oj.EditableValueUtils.getAttributeValue = function (element, attribute)
{
  var result, attrVal, propVal, returnVal = {}; 
  
  if (element && attribute)
  {
    switch (attribute)
    {
      case "disabled" :
        result = element.attr("disabled") !== undefined ? 
          !!element.prop("disabled") : undefined;
        break;
        
      case "pattern":
        result = element.prop("pattern") || undefined; 
        break; 
        
      case "placeholder":
        result = element.prop("placeholder") || undefined;
        break;
        
      case "readonly": 
        result = element.attr("readonly") !== undefined ? 
          !!element.prop("readonly") : undefined; 
        break;
      
      case "required":
        attrVal = element.attr("required");
        propVal = element.prop("required");

        // If attribute is present and not undefined
        //   - In IE9, required attribute is not supported at all, so attr() is defined, prop() is 
        //   undefined. In such cases set default to !!attrVal
        //   - Otherwise set to !!propVal
        // If neither attr is undefined then use defaultValue
        // TODO: review needed 
        result = (attrVal !== undefined) ? 
          ((propVal !== undefined) ? !!propVal : !!attrVal) : undefined;
        
        break;
      
      case "title":
        result = element.attr("title") !== undefined ? 
          element.prop("title") : undefined;
        break;
        
      case "value":
        // element attribute may not be set, in which case default to null
        result = element.val() || undefined;
        break;
      
      case "min": 
        //same logic for min + max as in default
      case "max":
      
      default: 
        result = element.attr(attribute) || undefined;
        break;
    }
  }
  
  if (result !== undefined)
  {
    returnVal.fromDom = true;
    returnVal.value = result;
  }
  else
  {
    returnVal.fromDom = false;
    // returnVal.value = defaultValue; 
  }
  return returnVal;
};

/**
 * NOTE: This is unnecessary to call for custom elements.
 * Called from component._InitOptions() with an array of options
 * that the component might need to initialize from DOM (e.g.,
 * disabled, required, title, etc). This function loops through each of these and if the 
 * constructorOptions[option] is undefined, it tries to get the option from DOM.
 * The constructorOptions hold the options that the page author sets on the component, usually via
 * knockout bindings, and that takes precedence over DOM.
 * e.g., <input id="id" type="text" required data-bind="ojComponent: {component: 'ojInputText', 
                                  value: value}"/>
 * value is a constructorOption and required is a DOM option in this example. If you have:
 * e.g., <input id="id" type="text" required data-bind="ojComponent: {component: 'ojInputText', 
                                  value: value, required: false}"/>
 * required is both a constructorOption of false and a DOM of true. The constructorOption takes
 * precedence.
 * <p>
 * IMPORTANT: Do not call this method after component has been created, since option values are 
 * mutated directly after that point.</p>
 * 
 * The 'final' value an option uses/initializes from, can come from these places (in order of least 
 * to most likely) - <br/>
 * <ol>
 * <li>component default - this is the widget default </li><br/>
 * <li>app default - this is what a page author defines for the value in the page/app</li> <br/>
 * <li>dom value - if your option also has a dom attribute, this is the value set on element for 
 * component. </li> <br/>
 * <li>constructor value - this is the value page author sets on the component binding </li><br/>
 * </ol>
 * 
 * At the time _InitOptions is called, (1), (2) and (4) are merged, but this may not be the value a 
 * component wants for the option, especially when (4) is undefined. For example, if these values 
 * were set for a component - <br/>
 * (1) - 'foo'<br/>
 * (2) - 'bar'<br/>
 * (3) - 'lucy'<br/>
 * (4) - undefined<br/>
 * <p>
 * at the time _InitOptions is called, this.options.option is set to 'bar'. But because DOM value 
 * wins over app default or component default, the component needs to check if the constructor value was 
 * undefined and if so, set option to the dom value which is 'lucy' in this example. This is what
 * this function does.<br/>
 * This method always defaults the value to be - this.options.option -  
 * because we think if neither (3) nor (4) is set, then the value from (2) should win. <br/>
 * </p>
 * 
 * @param {Object} props Array holding Object-literal that a component provides 
 * with the following properties that helps determine the final value for one or more options.
 * 
 * @property {string} props.attribute - name of DOM attribute
 * @property {string|undefined} props.option - name of the option if different from attribute name.
 * 
 * @property {Function|boolean|undefined} props.coerceDomValue - if the DOM value is set and 
 * coercing the dom value is needed, then either set to boolean true, which uses the default 
 * coercion rules for common attributes (a), or provide a custom callback (b). <p>
 * E.g., 'value' option for input number, input date etc. have special rules for coercing the value,
 *  so thse provide a custom callback. For common attributes like required and disabled, set the 
 *  value to true so the default oj.EditableValueUtils#coerceDomValueForOption method gets used.
 *  
 * @property {boolean|undefined} props.validateOption - if set to true, then it calls 
 * oj.EditableValueUtils.validateValueForOption method to validate the option.
 * 
 * @param {Object} constructorOptions the options set on the component instance, often using 
 * component binding. (this is the value page author sets on the component binding)
 * @param {Object} comp component instance.
 * @param {Function=} postprocessCallback - optional callback that will receive a map of initialized
 * options for post-processing
 * @ignore
 * @public
 */
oj.EditableValueUtils.initializeOptionsFromDom = function (props, constructorOptions, comp, postprocessCallback)
{
  var initializedOptions = {};
  
  // Loop through props to initialize option 
  for (var i = 0; i < props.length; i++)
  {
    var finalValue, result, prop = props[i], 
        attribute = prop.attribute, 
        option = prop.option || attribute, 
        coerceDomValue = prop.coerceDomValue,
        validateOption = prop.validateOption,
        element = comp.element, 
        previousValue = comp.options[option];

    /* The precedence for the value that an option uses is as follows from lowest to highest -
     *
     * (1) component default - this is the widget default, already merged in to comp.options
     * (2) app default - this is what a page author defines for the value in the page / app,
     * already merged in to comp.options
     * (3) dom value - if your option also has a dom attribute, this is the value set on element. 
     * (4) constructor value - this is the value page author sets on the component binding, already
     * merged in to comp.options.
     *      
     * When (4) is undefined then attempt to default from (3). 
     */

    // Step 1: use DOM value
    if (constructorOptions[option] === undefined)
    {
      previousValue = comp.options[option];
      result = oj.EditableValueUtils.getAttributeValue(element, attribute);

      // if we are using domValue then coerce the dom value before writing to options and trigegr 
      // option change so the value is written back (to ko)
      if (result.fromDom)
      {
        finalValue = result.value;

        // only required needs coercing so not bad
        if (coerceDomValue) 
        {
          if (typeof coerceDomValue === "boolean")
          {
            finalValue = oj.EditableValueUtils.coerceDomValueForOption(option, finalValue);
          }
          else if (typeof coerceDomValue === "function")
          {
            finalValue = coerceDomValue.call(comp, finalValue);
          }
        }
        initializedOptions[option] = finalValue;
      }
    }
    
    var valueToValidate = (option in initializedOptions) ? initializedOptions[option] : previousValue;

    // Step 2: validate the option value if needed
    if (validateOption)
    {
      if (typeof validateOption === "boolean")
      {
        oj.EditableValueUtils.validateValueForOption(option, valueToValidate);
      }
    }
  }
  
  if (postprocessCallback != null)
  {
    postprocessCallback(initializedOptions);
  }
  
  comp.option(initializedOptions, {'_context': {writeback: true, internalSet: true}});
};

  /**
   * Validates value set for the option and throws error if invalid.
   * 
   * @param {string} option name of the option. Validates options common to all edtiableValue 
   * holders.
   * @param {string|Object|boolean|number|undefined} value of the option that is validated
   * 
   * @throws {Error} if option value is invalid
   * @public
   * @ignore
   */
  oj.EditableValueUtils.validateValueForOption = function (option, value)
  {
    var error = false;
    
    switch (option)
    {
      case 'required' :
        if (value !== null &&  typeof value !== "boolean")
        {
          error = true; 
        }
        break;
      
      case "readOnly":
      case "disabled" :
        if (value !== null &&  typeof value !== "boolean")
        {
          error = true;
        }
        
        break;
    }
    
    if (error)
    {
      throw "Option '" + option + "' has invalid value set: " + value;
    }
  };
  
  
  /**
   * Coerces the dom value being used for the option, and throws error if invalid.
   * 
   * @param {string} option name of the option.
   * @param {string|Object|boolean|number|null} domValue dom value that is being coerced to the 
   * option value
   * @throws {Error} if domValue cannot be coerced appropriately
   * @public
   * @ignore
   */
  oj.EditableValueUtils.coerceDomValueForOption = function(option, domValue) 
  { 
    var coerced = domValue;
    switch (option)
    {
      case 'required' :
        coerced = domValue ? true : false;
        break;
    }
    
    return coerced;
  };


/**
 * set pickerAttributes on a popup picker
 *
 * @param {jQuery} picker popup picker
 * @param {Object} pickerAttributes supported attributes are class and style, which are appended to the picker class and style, if any.
 *
 * @ignore
 */
oj.EditableValueUtils.setPickerAttributes = function (picker, pickerAttributes)
{
  // - let the popup picker accept the custom css class name from the component
  if (picker && pickerAttributes)
  {
    var classValue = pickerAttributes["class"];
    if (classValue)
      picker.addClass(classValue);

    var styleValue = pickerAttributes["style"];
    if (styleValue)
    {
      var currStyle = picker.attr("style");
      if (currStyle)
      {
        picker.attr("style", currStyle + ';' + styleValue);
      }
      else
      {
        picker.attr("style", styleValue);
      }
    }
  }
};

/**
 * Use this to get the oj-label's label element's id when there is a for/id relationship
 * between the oj-label and the form component. Some components need this information to
 * use as their aria-labelledby on their dom element(s) that takes focus.
 * @param {jQuery} widget The custom form component jQuery object. 
 * We use this to find the oj-label element
 * by looking at oj-label's for attribute matching the id.
 * @param {string} defaultLabelId. the value we want to add to oj-label's label-id attribute if 
 * we can't find an existing id to use as aria-labelledby on the form component. 
 * @return {string|null} return the string to use as the aria-labelledby on the form component's
 * focusable element. If oj-label doesn't exist, this will return null.
 * @ignore
 */
oj.EditableValueUtils.getOjLabelId = function (widget, defaultLabelId)
{

  var formCompId;
  var $labelElement;
  var ojLabelCustom;
  var id;
  var labelElement;
  var labelElementId;
  var labelId;

  formCompId = widget[0].id;
  // oj-label and the form component as siblings is the most common case, so check for that first.
  $labelElement = widget.siblings("[for='" + formCompId + "']");
  
  if ($labelElement.length === 0)
  {
    ojLabelCustom =  document.querySelector("oj-label[for='" + formCompId + "']");
  }
  else
    ojLabelCustom = $labelElement[0];
  
  if (ojLabelCustom)
  {
    labelId = ojLabelCustom.getAttribute("label-id");
    if (labelId)
    {
      // if label-id is set, oj-label writes it directly on the label element's id
      return labelId;      
    }
    else
    {
      id = ojLabelCustom.id;
      if (id)
      {
        // the contract is for the label element's id, it's the oj-label id + this suffix 
        // (if oj-label doesn't have the label-id attribute set)
        return id + oj.EditableValueUtils.CUSTOM_LABEL_ELEMENT_ID;
      }
      else 
      {
        // oj-label has no label-id or id, so get oj-label's label element, get its id and use that.
        labelElement = ojLabelCustom.querySelector("label");
        if (labelElement)
          labelElementId = labelElement.id;
        if (labelElementId)
          return labelElementId;
        else
        {
          // if we get here, we did find oj-label, but it did not have label-id nor id,
          // and we also didn't find its label element's id. 
          // What we need to do now is write label-id onto oj-label, given the defaultLabelId parameter.
          if (defaultLabelId)
          {
            ojLabelCustom.setAttribute("label-id", defaultLabelId);
            return defaultLabelId;
          }
 
        }  
      }
    }
  }
  return null;
  

};

/**
 * Validates the component's display value using the converter and all validators registered on 
 * the component and updates the <code class="prettyprint">value</code> option by performing the 
 * following steps. 
 * 
 * <p>
 * <ol> 
 * <li>All messages are cleared, including custom messages added by the app. </li>
 * <li>If no converter is present then processing continues to next step. If a converter is 
 * present, the UI value is first converted (i.e., parsed). If there is a parse error then 
 * messages are shown.</li>
 * <li>If there are no validators setup for the component the <code class="prettyprint">value</code> 
 * option is updated using the display value. Otherwise all 
 * validators are run in sequence using the parsed value from the previous step. The implicit 
 * required validator is run first if the component is marked required. When a validation error is 
 * encountered it is remembered and the next validator in the sequence is run. </li>
 * <li>At the end of validation if there are errors, the messages are shown. 
 * If there were no errors, then the 
 * <code class="prettyprint">value</code> option is updated.</li>
 * </ol>
 * 
 * @example <caption>Validate component using its current value.</caption>
 * myComp.validate();
 * 
 * @example <caption>Validate component and use the Promise's resolved state.</caption>
 * myComp.validate().then(
 *  function(result) {
 *    if(result === "valid")
 *    {
 *      submitForm();
 *    }
 *  });
 *  
 *  
 * @return {Promise} Promise resolves to "valid" if there were no converter parse errors and
 * the component passed all validations. 
 * The Promise resolves to "invalid" if there were converter parse errors or 
 * if there were validation errors.
 * 
 * @ignore
 * @private
 */
oj.EditableValueUtils.validate =  function ()
{
  var self = this;
  if (this._IsCustomElement())
  {
    // clear all messages; run full validation on display value
    var validateFunction = function() {
      return self._ValidateReturnBoolean() ? 
        oj.EditableValueUtils.VALIDATE_VALUES.VALID : 
        oj.EditableValueUtils.VALIDATE_VALUES.INVALID;;
    };

    return Promise.resolve(validateFunction());

  }
  else
  {
    // clear all messages; run full validation on display value
    return this._ValidateReturnBoolean();
  }
};

/**
 * 
 * @returns boolean true if component passed validation, 
 * false if there were validation errors.
 * 
 * @example <caption>Validate component using its current value.</caption>
 * // validate display value. 
 * var rslt = myComp.validate();
 * 
 * @ignore
 * @protected
 */
oj.EditableValueUtils._ValidateReturnBoolean =  function ()
{
  // clear all messages; run full validation on display value
  var result = this._SetValue(this._GetDisplayValue(), null, this._VALIDATE_METHOD_OPTIONS);
  return result;

};

/**
 * Called from _AfterCreate for IsCustomElement form components so that they will
 * be assocatied with their oj-label element correctly.
 * Set the sub-id on the element so if they have a oj-label
 * pointing to it with the 'for' attrbiute, JAWS will read the label.
 * @param {Element} element
 * @param {string} widgetId
 * @ignore
 * @private
 */
oj.EditableValueUtils.setSubIdForCustomLabelFor = function(element, widgetId)
{
  element.setAttribute("id", widgetId + "|input");
};
/**For custom element only.
 * When labelledBy changes, we need to update the aria-labelledby attribute.
 * Note: If labelledBy changes from a value to null, we should still remove the oldValue from
 * aria-labelledby.
 * @param {string|null} originalValue the old value of the labelledBy option
 * @param {string|null} value the new value of the labelledBy option. 
 * @param {jQuery} $elems jquery Object containing the node(s) to add/remove aria-labelledby to.
 * @private
 * @ignore
 */
oj.EditableValueUtils._updateLabelledBy = function(originalValue, value, $elems)
{
  var suffix = oj.EditableValueUtils.CUSTOM_LABEL_ELEMENT_ID;
  
  if (!this._IsCustomElement())
    return;

  if (originalValue)
    originalValue += suffix;
  if (value)
    value += suffix;
  
  if (originalValue)
    oj.EditableValueUtils._removeAriaLabelledBy($elems, originalValue);
  if (value)
    oj.EditableValueUtils._addAriaLabelledBy($elems, value);    
};

/**
 * Refresh everything that needs refreshing when the required option is toggled:
 * refreshes theming, aria-required, label.
 *  
 * Used by EditableValue components that support the required option.
 * The component links to this function like this:
 *   _refreshRequired : oj.EditableValueUtils._refreshRequired,
 * @param {Object=} value the current value of the required option
 * @private
 * @ignore
 */
oj.EditableValueUtils._refreshRequired = function(value)
{
  var id;
  var contentNode;
  var ariaValue;
  var ariaRequiredUnsupported = this._AriaRequiredUnsupported();
  
  this._refreshTheming("required", value);
  // refresh aria-required
  // Most inputs/selects need aria-required on the input element (aka 'content')
  // But it is not legal to have aria-required on radio/checkboxes.
  if (!ariaRequiredUnsupported)
  { 
    contentNode = this._GetContentElement();

    ariaValue = value; //(value == "required") ? true : false;
    if (ariaValue && contentNode) 
      contentNode.attr("aria-required", ariaValue);
    else
    contentNode.removeAttr("aria-required");
  }

  if (!this._IsCustomElement())
  {
    if (!this.$label)
      this._createOjLabel();
    
    // need to keep the label's required in sync with the input's required
    if (this.$label)
    {
      this.$label.ojLabel("option", "showRequired", value);
      // in most cases aria-required is supported and that is what we do to get JAWS to say
      // "Required" on focus of the input. But in the case of a 'set' of items where one is required,
      // say radioset/checkboxset, what do we do? aria-required doesn't make sense (nor is it valid
      // as it fails validation in some accessibility validators) on each input, when really it is
      // one in the set that is required, not each one. This is what we are doing from v1 on: we
      // put aria-describedby to point to the required icon text.
      if (ariaRequiredUnsupported)
      {
        // if aria-labelledby is set, 
        // add/remove aria-describedby to the inputs pointing to
        // the label+"_requiredIcon". 
        id = this._getAriaLabelledById(this.element);
        if (id)
        {   
          if (value)
            this._addAriaDescribedBy(id + _REQUIRED_ICON_ID);
          else
            this._removeAriaDescribedBy(id + _REQUIRED_ICON_ID);
        }
      } 
    }
  }
};
 
/**
 * Performs post processing after required option is set by taking the following steps.
 * 
 * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
 * run full validation with UI value (we don't know if the UI error is from a required validator 
 * or something else);<br/>
 * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
 * updated<br/>
 * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
 * listen to optionChange(value) to clear custom errors.<br/>
 * 
 * - if component is invalid and has messagesHidden -> required: false -> clear component 
 * errors; no deferred validation is run.<br/>
 * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
 * errors unnecessarily)<br/>
 * - messagesCustom is never cleared<br/>
 * 
 * @param {string} option
 * 
 * @private
 * @ignore
 */  
oj.EditableValueUtils._AfterSetOptionRequired = function (option)
{
  // refresh hints, theming and aria to reflect new state
  this._refreshRequired(this._IsRequired());
  this._runMixedValidationAfterSetOption(oj.EditableValueUtils.requiredOptionOptions);
};

/**
   * When validators option changes, take the following steps.
   * 
   * - Clear the cached normalized list of all validator instances. push new hints to messaging.<br/>
   * - if component is valid -> validators changes -> no change<br/>
   * - if component is invalid has messagesShown -> validators changes -> clear all component 
   * messages and re-run full validation on displayValue. if there are no errors push value to 
   * model;<br/>
   * - if component is invalid has messagesHidden -> validators changes -> do nothing; doesn't change 
   * the required-ness of component <br/>
   * - messagesCustom is not cleared.<br/>
   * 
   * NOTE: The behavior applies to any option that creates implicit validators - min, max, pattern, 
   * etc. Components can call this method when these options change.
   * 
   * @returns {undefined}
   * @private
   * @ignore
   */
  oj.EditableValueUtils._AfterSetOptionValidators = function ()
  {
    var runFullValidation = false;
    
    // resets all validators and pushes new hints to messaging
    this._ResetAllValidators();
    
    if (this._hasInvalidMessagesShowing())
    {
      runFullValidation = true;
    }
    
    if (runFullValidation)
    {
      this._clearComponentMessages();
      this._updateValue(oj.EditableValueUtils.validatorsOptionOptions);
    }
  };
  /**
   * Performs post processing after converter option changes by taking the following steps.
   * 
   * - always push new converter hint to messaging <br/>
   * - if component has no errors -> refresh UI value<br/>
   * - if component is invalid has messagesShown -> clear all component errors and run full 
   * validation using display value. <br/>
   * &nbsp;&nbsp;- if there are validation errors, value is not pushed to model; messagesShown is 
   * updated.<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to optionChange(value) to clear custom errors.<br/>
   * - if component is invalid has messagesHidden -> refresh UI value. no need to run deferred 
   * validations. <br/>
   * - messagesCustom is never cleared<br/>
   * 
   * @param {String} option
   * 
   * @returns {undefined}
   * @private
   * @ignore
   */
    // called when 'converter' option changed, usually from option/setOption calls
  oj.EditableValueUtils._AfterSetOptionConverter = function (option)
  {
    var runFullValidation = false;
    
    // clear the cached converter instance and push new hint to messaging
    this._ResetConverter();

    if (this._hasInvalidMessagesShowing())
    {
      runFullValidation = true;
    }
    
    if (runFullValidation)
    {
      this._clearComponentMessages();
      this._updateValue(oj.EditableValueUtils.converterOptionOptions);
    }
    else
    {
      // refresh UI display value when there are no errors or where there are only deferred errors 
      this._Refresh(option, this.options[option], true);
    }
  };
  /**
   * Clears the cached converter stored in _converter and pushes new converter hint to messaging.
   * Called when convterer option changes 
   * 
   * @private
   * @ignore
   */
  oj.EditableValueUtils._ResetConverter = function ()
  {
    this._converter = null;
    this._getComponentMessaging().update(
            this._getMessagingContent(this._MESSAGING_CONTENT_UPDATE_TYPE.CONVERTER_HINT));
  };
  /**
   * Returns an array of all validators built by merging the validators option set on the component 
   * and the implicit validators setup by the component. <br/>
   * This does not include the implicit required validator. Components can override to add to this 
   * array of validators.
   * 
   * @return {Array} of validators
   * 
   * @private
   * @ignore
   */
  oj.EditableValueUtils._GetNormalizedValidatorsFromOption = function ()
  {
    var i;
    var isValidatorInstance = true;
    var normalizedValidators = [];
    var validator;
    var validatorsOption;
    var vOptions;
    var vType;
    var vTypeStr;
          

    validatorsOption = this.options['validators']; 

    if (validatorsOption)
    {
      // Normalize validators 
      for (i = 0; i < validatorsOption.length; i++)
      {
        validator = validatorsOption[i];
        if (typeof validator === "object") 
        {
          // check if we have an actual validator instance that implements the validate() method
          // oj.Validation.__doImplementsCheck(validator, oj.Validator);
          if (!(validator['validate'] && typeof validator['validate'] === "function"))
          {
            isValidatorInstance = false;
          }

          if (!isValidatorInstance)
          {
            // we maybe dealing with an object literal
            // {'type': 'numberRange', 'options': { 'min': 100, 'max': 1000,
            //                                    'hint': {'min': 'some hint about min'}}}
            vTypeStr = validator['type'];
            if (vTypeStr && typeof vTypeStr === "string")
            {
              vType = oj.Validation.validatorFactory(vTypeStr);
              if (vType)
              {
                vOptions = oj.CollectionUtils.copyInto({}, validator['options']) || {};
                // we push converter into the options if not provided explicitly. This is to allow
                // validators to format values shown in the hint and messages
                vOptions['converter'] = vOptions['converter'] || this._GetConverter();
                vOptions['label'] = vOptions['label'] || this._getLabelText();
                validator = vType.createValidator(vOptions);
              }
              else
              {
                oj.Logger.error("Unable to locate a validatorFactory for the requested type: " + vTypeStr);
              }
            }
          }

          normalizedValidators.push(validator);
        }
        else
        {
          oj.Logger.error("Unable to parse the validator provided:" + validator);
        }
      }

    }
    return normalizedValidators;
  };
  /**
   * Returns the normalized converter instance.
   * 
   * @return {Object} a converter instance or null
   * 
   * @private
   * @ignore
   */
  oj.EditableValueUtils._GetConverter = function () 
  {
    var converterOption;
    
    // this._converter holds the instance
    if (!this._converter)
    {
      converterOption = this.options['converter'];
      this._converter = oj.IntlConverterUtils.getConverterInstance(converterOption);
    }
    
    return this._converter || null;
  };
   
  /**.
   * Add the id to the widget's aria-labelledby attribute.
   * @param {jQuery} $elems the jquery element(s) that represents the node on which aria-labelledby is
   * @param {string} id id to add to aria-labelledby
   * @private
   * @ignore
   */
  oj.EditableValueUtils._addAriaLabelledBy = function ($elems, id)
  {
    var index;
    
    
    $elems.each(function() {
      var labelledBy = $(this).attr("aria-labelledby");
      var tokens;

      tokens = labelledBy ? labelledBy.split(/\s+/) : [];
      // Get index that id is in the tokens, if at all.
      index = $.inArray(id, tokens);
      // add id if it isn't already there
      if (index === -1)
        tokens.push(id);
      labelledBy = $.trim(tokens.join(" "));
      $(this).attr("aria-labelledBy", labelledBy);
    });   
  };
  /**.
   * Remove the id from the widget's aria-labelledby attribute.
   * @param {jQuery} $elems the jquery element(s) that represents the node on which aria-labelledby is
   * @param {string} id id to remove from aria-labelledby
   * @private
   * @ignore
   */
  oj.EditableValueUtils._removeAriaLabelledBy = function ($elems, id)
  {
    var labelledBy;
    
    $elems.each(function() {

      var index;
      var tokens;
      
      // get aria-labelledby that is on the element(s)
      labelledBy = $(this).attr("aria-labelledby");
      // split into tokens
      tokens = labelledBy ? labelledBy.split(/\s+/) : [];
      // Get index that id is in the tokens, if at all.
      index = $.inArray(id, tokens);
      // remove that from the tokens array
      if (index !== -1)
        tokens.splice(index, 1);
      // join the tokens back together and trim whitespace
      labelledBy = $.trim(tokens.join(" "));
      if (labelledBy)
        $(this).attr("aria-labelledby", labelledBy);
      else
        $(this).removeAttr("aria-labelledby");
     });
  };
  /**
   * Set the type of the input element based on virtualKeyboard option.
   * 
   * @param {Array.<string>} allowedTypes an array of allowed types
   * 
   * @protected
   * @ignore
   */
  oj.EditableValueUtils._SetInputType = function (allowedTypes)
  {
    // Default input type is text
    var inputType = 'text';
    var agentInfo = oj.AgentUtils.getAgentInfo();

    // Only change the type on mobile browsers    
    if (agentInfo['os'] === oj.AgentUtils.OS.ANDROID ||
        agentInfo['os'] === oj.AgentUtils.OS.IOS ||
        agentInfo['os'] === oj.AgentUtils.OS.WINDOWSPHONE)
    {
      // Get input type from component's virtualKeyboard option
      if (allowedTypes.indexOf(this.options['virtualKeyboard']) >= 0)
      {
        inputType = this.options['virtualKeyboard'];
      }
/* For future support
      else
      {
        // Get input type from converter's virtualKeyboardHint option
        var converter = this._GetConverter();
        if (converter && converter['resolvedOptions'])
        {
          var resOptions = converter['resolvedOptions']();
          if (allowedTypes.indexOf(resOptions['virtualKeyboardHint']) >= 0)
          {
            inputType = resOptions['virtualKeyboardHint'];
          }
        }
      }
*/
    }
    
    this.element.attr('type', inputType);
  };
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * The various validation modes
 * @ignore
 */ 

var _sValidationMode = {
  FULL : 1, 
  VALIDATORS_ONLY : 2, 
  REQUIRED_VALIDATOR_ONLY : 3
};

/**
* String used in the id on the span that surrounds the help icon.
* @const
* @private
* @type {string}
*/
var _HELP_ICON_ID = "_helpIcon";


/**
* String used in the label element's id for custom &lt;oj-label>
* @const
* @private
* @type {string}
*/
var _CUSTOM_LABEL_ELEMENT_ID = "|label";


// E D I T A B L E V A L U E    A B S T R A C T   W I D G E T  
/**
 * @ojcomponent oj.editableValue
 * @augments oj.baseComponent
 * @ojsignature [{
 *                target: "Type",
 *                value: "abstract class editableValue<V, SP extends editableValueSettableProperties<V, SV, RV>, SV= V, RV= V> extends baseComponent<SP>"
 *               },
 *               {
 *                target: "Type",
 *                value: "editableValueSettableProperties<V, SV=V, RV=V> extends baseComponentSettableProperties",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @ojtsimport ojmessaging
 * @abstract
 * @since 0.6
 * @ojshortdesc Abstract EditableValue element
 * @ojrole input
 * @hideconstructor
 * 
 * @classdesc
 * Abstract base class for all editable components that are value holders and that require 
 * validation and messaging capabilities. <br/>
 * 
 * {@ojinclude "name":"validationAndMessagingDoc"}
 *  * <p>
 * Note: The <code class="prettyprint">required</code>, <code class="prettyprint">validators</code>,
 * <code class="prettyprint">converter</code> properties and the <code class="prettyprint">validate</code> 
 * method are not on all EditableValue components so they are not on the EditableValue class. 
 * See the EditableValue subclasses for which ones have which of these properties. For example,
 * oj-switch, oj-slider, oj-color-palette, and oj-color-spectrum do not have the 
 * <code class="prettyprint">validate</code> method nor do they have the 
 * <code class="prettyprint">required</code>, <code class="prettyprint">validators</code>,
 * <code class="prettyprint">converter</code> properties since the components 
 * wouldn't do anything with these properties anyway. A user can't type into these components and there
 * is no visual representation for 'nothing is set' on these components. Whereas InputBase, inputNumber,
 * inputSearch and combobox do have these properties since a user can type into the field (so you may
 * need to convert it and validate it) and also blank it out (so you may need to mark it required and
 * run the required validator).
 * </p>

 * <p>
 * <h3 id="declarative-binding-section">
 * Declarative Binding 
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#declarative-binding-section"></a>
 * </h3>
 * When the component's <code class="prettyprint">value</code> property is bound to a Knockout 
 * observable and when the value changes, whether the observable is updated or not, and whether a 
 * 'writeback' to the observable happens or not, depends on the action that caused the value to 
 * change.
 * <ul>
 * <li>when the value changes as a result of user interaction </li>
 * <li>when the value changes because normal validation was run as a result of these properties 
 * being changed by the app - <code class="prettyprint">converter</code>, <code class="prettyprint">disabled</code>, 
 * <code class="prettyprint">required</code>, <code class="prettyprint">validators</code>, then the 
 * value is written to the observable. See the specific property docs for details.</li>
 * <li>when the value changes because normal validation was run as a result of these methods being 
 * called by the app - 
 * <code class="prettyprint">refresh</code>, <code class="prettyprint">validate</code>, 
 * then the value is written to the observable. See the specific method docs for details.</li>
 * <li>when the value changes due to programmatic intervention by app then the value is not written 
 * back to observable. This is based on the assumption that the app has mutated the observable 
 * already. In this case updating the component's <code class="prettyprint">value</code> property 
 * alone will not propagate the change automatically to the observable. Updating the observable is 
 * recommended as this will propagate the change automatically to the component.
 * </li>
 * </ul>
 * </p>
 * 
 * @example <caption>Initialize component</caption>
 * &lt;input id="foo" type="text"/&gt;
 * &lt;script&gt;
 * &nbsp;&nbsp;$('#foo").ojInputText({'value': 'abc'});
 * &lt;/script&gt;
 * // using knockout ojComponent binding
 * &lt;input id="foo" data-bind="ojComponent: {component: 'ojInputText', value: 'abc'}"/&gt;
 * @example <caption>Initialize component value using ko observable</caption>
 * &lt;input id="foo" data-bind="ojComponent: {component: 'ojInputText', value: salary}"/&gt;
 * &lt;script&gt;
 * &nbsp;&nbsp;var salary = ko.observable('abc');
 * &lt;/script&gt;
 * @example <caption>Initialize component value using element value</caption>
 * &lt;input id="foo" data-bind="ojComponent: {component: 'ojInputText'}" value='abc'/&gt;
 */
oj.__registerWidget('oj.editableValue', $['oj']['baseComponent'], 
{
  widgetEventPrefix: "oj",
  
  options: 
  {
    /** 
     * It is used to establish a relationship between this component and another element.
     * Typically this is not used by the application developer, but by the oj-label custom element's
     * code. One use case is where the oj-label custom element code writes described-by
     * on its form component for accessibility reasons.
     * To facilitate correct screen reader behavior, the described-by attribute is
     * copied to the aria-describedby attribute on the component's dom element.
     * @example <caption>Initialize component with the <code class="prettyprint">described-by</code> attribute specified:</caption>
     * &lt;oj-some-element described-by="someId">&lt;/oj-some-element>
     *
     * @example <caption>Get or set the <code class="prettyprint">describedBy</code> property after initialization:</caption>
     * // getter
     * var descById = myComp.describedBy;
     *
     * // setter
     * myComp.describedBy = "someId";
     *
     * @ojshortdesc described the relationship between this component and another element.
     * @expose 
     * @type {?string}
     * @public
     * @instance
     * @memberof oj.editableValue
     * @since 4.0.0
     */
    describedBy : null,
    /** 
     * Whether the component is disabled. The default is false.
     * 
     * <p>
     * When the <code class="prettyprint">disabled</code> property changes due to programmatic 
     * intervention, the component may clear messages and run validation in some cases. </br>
     * <ul>
     * <li>when a required component is initialized as disabled 
     * <code class="prettyprint">value="{{currentValue}}" required disabled</code>, 
     * deferred validation is skipped.</li>
     * <li>when a disabled component is enabled, 
     *  <ul>
     *  <li>if component is invalid and showing messages then all component messages are cleared, 
     *  and full validation run using the display value.
     *   <ul>
     *    <li>if there are validation errors, they are shown.</li>
     *    <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *    property is updated. Page authors can listen to the <code class="prettyprint">onValueChanged</code> 
     *    event to clear custom errors.</li>
     *   </ul>
     *  </li>
     *  
     *  <li>if component is valid and has no errors then deferred validation is run.
     *    <ul>
     *    <li>if there is a deferred validation error, then the valid property is updated. </li>
     *    </ul>
     *  </li>
     *  <li>if component is invalid and deferred errors then component messages are cleared and 
     *  deferred validation re-run.
     *    <ul>
     *    <li>if there is a deferred validation error, then the valid property is updated.</li>
     *    </ul>
     *  </li>
     *  </ul>
     * </li>
     * <li>when enabled component is disabled then no validation is run and the component appears 
     * disabled.</li>
     * </ul>
     * </p>
     * 
     * @example <caption>Initialize component with <code class="prettyprint">disabled</code> attribute:</caption>
     * &lt;oj-some-element disabled>&lt;/oj-some-element>
     *
     * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
     * // getter
     * var disabled = myComp.disabled;
     *
     * // setter
     * myComp.disabled = false;
     *
     * @ojshortdesc Whether the component is disabled. The default is false.
     * @expose 
     * @type {boolean}
     * @default false
     * @public
     * @instance
     * @memberof oj.editableValue
     * @since 0.7
     */
    disabled : false,
    /**
     * Display options for auxilliary content that determines where it should be displayed 
     * in relation to the component. 
     * 
     * <p>
     * The types of messaging content for which display options can be configured include 
     * <code class="prettyprint">messages</code>, <code class="prettyprint">converterHint</code>, 
     * <code class="prettyprint">validatorHint</code> and <code class="prettyprint">helpInstruction</code>.<br/>
     * The display options for each type is specified either as an array of strings or a string. When 
     * an array is specified the first display option takes precedence over the second and so on. 
     * </p>
     * <p>
     * JET editable components set defaults that apply to the entire app/page. 
     * It is possible to override the defaults on 
     * a per instance basis as explained in the examples below or change defaults for the entire
     * application using 
     * <a href="oj.Components.html#setDefaultOptions"><code class="prettyprint">oj.Components#setDefaultOptions</code></a> method.
     * It is much easier to change the defaults using setDefaultOptions once rather than putting
     * the displayOptions option on every component instance.<br/>
     * </p>
     * <p>
     * When display-options changes due to programmatic intervention, the component updates its 
     * display to reflect the updated choices. For example, if 'help.instruction' property goes from 
     * 'notewindow' to 'none' then it no longer shows in the notewindow.
     * </p>
     * <p>
     * A side note: help.instruction and message detail text can include formatted HTML text, whereas hints and 
     * message summary text cannot. If you use formatted text, it should be accessible 
     * and make sense to the user if formatting wasn't there.
     * To format the help.instruction, you could do this:
     * <pre class="prettyprint"><code>&lt;html>Enter &lt;b>at least&lt;/b> 6 characters&lt;/html></code></pre>
     * </p>
     *  
     * @property {(Array.<'placeholder'|'notewindow'|'none'>|'placeholder'|'notewindow'|'none')=} converterHint - supported values are <code class="prettyprint">'placeholder'</code>, 
     * <code class="prettyprint">'notewindow'</code>, <code class="prettyprint">'none'</code>. The 
     * default value is <code class="prettyprint">['placeholder', 'notewindow']</code>. When there 
     * is already a placeholder set on the component, the converter hint falls back to display 
     * type of 'notewindow'.
     * To change the default value you can do this - <br/> 
     * E.g. <code class="prettyprint">{'displayOptions: {'converterHint': ['none']}}</code>
     * @property {(Array.<'notewindow'|'none'>|'notewindow'|'none')=} validatorHint - supported values are <code class="prettyprint">'notewindow'</code>, 
     * <code class="prettyprint">'none'</code>.
     * To change the default value you can do this - <br/>
     * <code class="prettyprint">{'displayOptions: {'validatorHint': ['none']}}</code>
     * @property {(Array.<'inline'|'notewindow'|'none'>|'inline'|'notewindow'|'none')=} messages - supported values are <code class="prettyprint">'notewindow'</code>, 
     * <code class="prettyprint">'inline'</code>,
     * <code class="prettyprint">'none'</code>. The default is 'inline'. 
     * To change the default value you can do this - <br/>
     * E.g. <code class="prettyprint">{'displayOptions: {'messages': 'none'}}</code>
     * @property {(Array.<'notewindow'|'none'>|'notewindow'|'none')=} helpInstruction - supported values are <code class="prettyprint">'notewindow'</code>, 
     * <code class="prettyprint">'none'</code>.
     * To change the default value you can do this - <br/>
     * E.g. <code class="prettyprint">displayOptions='{"helpInstruction": "none"}'</code>
     * 
     * @example <caption>Override default values for <code class="prettyprint">displayOptions</code> 
     *  for messages for the entire application:</caption>
     * // messages will be shown in the notewindow for the application.
     * oj.Components.setDefaultOptions({
     *    'editableValue':
     *    {
     *      'displayOptions': 
     *      {
     *        'messages': ['notewindow']
     *      }
     *    }
     * });
     * 
     * @example <caption>Override default values for <code class="prettyprint">display-options</code> 
     * for one component:</caption>
     * // In this example, the display-options are changed from the defaults.
     * // The 'converterHint' is none, the 'validatorHint' is none and the 'helpInstruction' is none,
     * // so only the 'messages' will display in its default state.
     * // For most apps, you will want to change the displayOptions app-wide
     * // for all EditableValue components, so you should use the
     * // oj.Components#setDefaultOptions function instead (see previous example).
     * //
     * &lt;oj-some-element display-options='{"converterHint": "none",
     *                                     "validatorHint": "none",
     *                                     "helpInstruction": "none"}'>&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">displayOptions</code> property after initialization:</caption>
     * // Get one subproperty
     * var hint = myComp.displayOptions.converterHint;
     * 
     * // Set one, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myComp.setProperty("displayOptions.converterHint", "none");
     * 
     * // get all
     * var options = myComp.displayOptions;
     *
     * // set all.  Must list every resource key, as those not listed are lost.
     * myComp.displayOptions = {converterHint: "none", validatorHint: "none", helpInstruction: "none"};
     *
     * @ojshortdesc Customize how to display to the user the
     *  form field's messages, converter and validator hints and help instruction text.
     * @expose 
     * @access public
     * @instance
     * @default {'messages': ['inline'],'converterHint': ['placeholder','notewindow'],'validatorHint': ['notewindow'],'helpInstruction': ['notewindow']}
     * @memberof oj.editableValue
     * @type {Object|undefined}
     * @since 0.7
     */
    displayOptions : undefined,    
    /**
     * Form component help information.
     * <p>
     * The properties supported on the <code class="prettyprint">help</code> option are:
     * 
     * @property {string=} instruction this represents advisory information for the component
     * The default value is <code class="prettyprint">null</code>.
     * 
     * @expose 
     * @memberof oj.editableValue
     * @instance
     * @public
     * @type {Object.<string, string>}
     * @default {'help' : {'instruction': null}}
     * @since 0.7
     */
    help: undefined,
    /**
     * <p>help definition text.  See the top-level <code class="prettyprint">help</code> option for details.
     * 
     * @expose
     * @alias help.definition
     * @memberof! oj.editableValue
     * @instance
     * @type {?string}
     * @ignore
     * @default null
     * 
     * @example <caption>Get or set the <code class="prettyprint">help.definition</code> sub-option, after initialization:</caption>
     * // getter
     * var definitionText = $( ".selector" ).ojFoo( "option", "help.definition" );
     * 
     * // setter:
     * $( ".selector" ).ojFoo( "option", "help.definition", "Enter your name" );
     */     
    /**
     * <p>help source url.  See the top-level <code class="prettyprint">help</code> option for details.
     * 
     * @expose
     * @alias help.source
     * @memberof! oj.editableValue
     * @instance
     * @ignore
     * @type {?string}
     * @default null
     * 
     * @example <caption>Get or set the <code class="prettyprint">help.source</code> sub-option, after initialization:</caption>
     * // getter
     * var helpSource = $( ".selector" ).ojFoo( "option", "help.source" );
     * 
     * // setter:
     * $( ".selector" ).ojFoo( "option", "help.source", "www.abc.com" );
     */
    /**
     * Represents hints for oj-form-layout element to render help information on the label of the editable component. 
     * <p>This is used only if the editable component is added as a direct child to an oj-form-layout element, and the labelHint property is also specified.</p>
     * 
     * <p>
     * The helpHints object contains a definition property and a source property.
     * </p>
     * <ul>
     * <li><code class="prettyprint">definition</code> - hint for help definition text.</li>
     * <li><code class="prettyprint">source</code> - hint for help source URL.</li>
     * </ul>
     *
     * @example <caption>Initialize the component with help hints:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-some-element help-hints.definition='some value' help-hints.source='some-url'>&lt;/oj-some-element>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-some-element help-hints='{"definition":"some value", "source":"some-url"}'>&lt;/oj-some-element>
     *
     * @example <caption>Get or set the <code class="prettyprint">helpHints</code> property after 
     * initialization:</caption>
     * 
     * // Get one
     * var value = myComponent.helpHints.definition;
     * 
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myComponent.setProperty('helpHints.definition', 'some new value');
     *
     * // Get all
     * var values = myComponent.helpHints;
     *
     * // Set all.  Must list every subproperty, as those not listed are lost.
     * myComponent.helpHints = {
     *     definition: 'some new value',
     *     source: 'some-new-url'
     * };
     *
     * @ojshortdesc Represents hints for oj-form-layout element to 
     * render help information on the label of the editable component.
     * @expose
     * @access public
     * @memberof oj.editableValue
     * @ojtranslatable
     * @instance
     * @type {Object.<string, string>}
     * @default {'definition': "", 'source': ""}
     * @since 4.1.0
     */
     helpHints:
     {
    /**
     * Hint for help definition text associated with the label. 
     * <p>It is what shows up when the user hovers over the help icon, or tabs into the help icon, or press and holds the help icon on a mobile device. No formatted text is available for help definition attribute.</p>
     * 
     * <p>See the <a href="#helpHints">help-hints</a> attribute for usage examples.</p>
     *
     * @expose
     * @alias helpHints.definition
     * @memberof! oj.editableValue
     * @instance
     * @type {string}
     * @ojsignature {target:"Type", value: "?"}
     * @default ""
     * @since 4.1.0
     */
      definition: "",
    /**
     * Hint for help source URL associated with the label. 
     * <p>If present, a help icon will render next to the label. For security reasons we only support urls with protocol http: or https:. If the url doesn't comply we ignore it and throw an error. 
     * Pass in an encoded URL since we do not encode the URL.</p>
     * 
     * <p>See the <a href="#helpHints">help-hints</a> attribute for usage examples.</p>
     *
     * @expose
     * @alias helpHints.source
     * @memberof! oj.editableValue
     * @instance
     * @type {string}
     * @ojsignature {target:"Type", value: "?"}
     * @default ""
     * @since 4.1.0
     */
      source: ""
    },
    /** 
     * Represents a hint for oj-form-layout element to render a label on the editable component. 
     * <p>This is used only if the editable component is added as a direct child to an oj-form-layout element.</p>
     * 
     * <p>
     * When labelHint is present it gives a hint to the oj-form-layout element to create an oj-label element for the editable component.
     * When the <code class="prettyprint">labelHint</code> property changes oj-form-layout element refreshes to 
     * display the updated label information. 
     * </p>
     *  
     * @example <caption>Initialize the component with the <code class="prettyprint">label-hint</code> attribute specified:</caption>
     * &lt;oj-some-element label-hint='input label'>&lt;/oj-some-element>
     *
     * @example <caption>Get or set the <code class="prettyprint">labelHint</code> property after 
     * initialization:</caption>
     * 
     * // getter
     * var value = myComponent.labelHint;
     * 
     * // setter
     * myComponent.labelHint = 'some new value'
     * 
     * @ojshortdesc Represents a hint for oj-form-layout element to render a label on the editable component.
     * @expose 
     * @access public
     * @instance
     * @alias labelHint
     * @ojtranslatable
     * @default ""
     * @memberof oj.editableValue
     * @type {string}
     * @since 4.1.0
     */  
    labelHint: "",
    /**
     * List of messages an app would add to the component when it has business/custom validation 
     * errors that it wants the component to show. This allows the app to perform further validation 
     * before sending data to the server. When this option is set the message shows to the 
     * user right away. To clear the custom message, set <code class="prettyprint">messagesCustom</code>
     * back to an empty array.<br/>
     * <p>Each message in the array is an object that duck types oj.Message.      
     * See {@link oj.Message} for details.
     * </p>
     * <p>
     * See the <a href="#validation-section">Validation and Messages</a> section
     * for details on when the component clears <code class="prettyprint">messagesCustom</code>; 
     * for example, when full validation is run.
     * </p>
     * 
     *
     * @example <caption>Get or set the <code class="prettyprint">messagesCustom</code> property after initialization:</caption>
     * // getter
     * var customMsgs = myComp.messagesCustom;
     * 
     * // setter
     * myComp.messagesCustom = [{summary:"hello", detail:"detail", severity:oj.Message.SEVERITY_LEVEL.INFO}];
     * 
     * @example <caption>Set messagesCustom when there are cross-validation errors:</caption>
     * --- HTML ---
     * &lt;oj-some-element messages-custom='{{messagesCustom}}'>&lt;/oj-some-element> 
     * 
     * --- ViewModel code ---
     * self.messagesCustom = ko.observableArray([]);
     * 
     * // the app's function that gets called when the user presses the submit button
     * if (!myValidateCrossValidationFields())
     * {
     *   // the app adds a custom messages to the component and it is displayed right away
     *   var msgs = [];
     *   msgs.push({'summary':'Cross field error', 'detail':'Field 1 needs to be less than Field 2'});
     *   self.messagesCustom(msgs);
     * }
     * else
     * {
     *   // submit data to the server
     * }
     * 
     * @ojshortdesc List of messages an app would add to the component
     * @expose 
     * @access public
     * @instance
     * @memberof oj.editableValue
     * @default []
     * @type {Array.<Object>}
     * @ojsignature {target: "Type", value: "Array<oj.Message>"}
     * @since 0.7
     * @ojwriteback
     */    
    messagesCustom : [],
    
    /**
     * List of messages currently hidden on component, these are added by component when it runs 
     * deferred validation. Each message in the array is an object that duck types oj.Message. 
     * See {@link oj.Message} for 
     * details. <br/>
     * 
     * <p>
     * This is a read-only option so page authors cannot set or change it directly.
     * </p>
     * 
     * <p>
     * These messages are not shown to the end-user by default, but page author 
     * can show hidden messages using the {@link showMessages} method. 
     * </p>
     * 
     * @example <caption>Get <code class="prettyprint">messagesShown</code> for the component:</caption>
     * // Foo is InputText, InputNumber, Select, etc.
     * var messages = $(".selector").ojFoo("option", "messagesShown"); 
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof oj.editableValue
     * @default []
     * @type {Array.<Object>|undefined}
     * @since 0.7
     * @see #showMessages
     * @readonly
     * @ignore
     * @ojwriteback
     */    
    messagesHidden : undefined,     
    
    /**
     * List of messages currently shown on component, these include messages generated both by the 
     * component and ones provided by app using <code class="prettyprint">messagesCustom</code>. 
     * Each message in the array is an object that duck types oj.Message.
     * See {@link oj.Message} for details. <br/>
     * 
     * <p>
     * This is a read-only option so page authors cannot set or change it directly.
     * </p>
     * 
     * <p>
     * Messages retrieved using the <code class="prettyprint">messagesShown</code> option are by 
     * default shown inline, but this can be controlled using the 'messages' property of 
     * the <code class="prettyprint">displayOptions</code> option. 
     * </p>
     * 
     * @example <caption>Get <code class="prettyprint">messagesShown</code> for the component:</caption>
     * // Foo is InputText, InputNumber, Select, etc.
     * var messages = $(".selector").ojFoo("option", "messagesShown"); 
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof oj.editableValue
     * @default []
     * @type {Array.<Object>|undefined}
     * @since 0.7
     * @readonly
     * @ignore
     * @ojwriteback
     */    
    messagesShown : undefined,    

    
    /** 
     * Represents advisory information for the component, such as would be appropriate for a tooltip. 
     * 
     * <p>
     * When help.instruction is present it is by default displayed in the notewindow, or as determined by the 
     * 'helpInstruction' property set on the <code class="prettyprint">displayOptions</code> attribute. 
     * When the <code class="prettyprint">help.instruction</code> property changes the component refreshes to 
     * display the updated information. 
     * </p>
     * 
     * <p>
     * JET takes the help instruction text and creates a notewindow with the text.
     * The help instruction only shows up as a tooltip on mouse over, not on keyboard and not in a mobile
     * device. So help instruction would only be for text that is not important enough to show all users, or
     * for text that you show the users in another way as well, like in the label.
     * Also you cannot theme the native browser's title window like you can the JET
     * notewindow, so low vision users may have a hard time seeing it. 
     * For these reasons, the JET EditableValue components do not use the HTML's title
     * attribute.
     * </p>
     * 
     * <p>
     * To include formatted text in the help.instruction, format the string using html tags. 
     * For example the 
     * help.instruction might look like: 
     * <pre class="prettyprint"><code>&lt;oj-some-element help.instruction="&lt;html>Enter &lt;b>at least&lt;/b> 6 characters&lt;/html>">&lt;/oj-some-element></code></pre>
     * If you use formatted text, it should be accessible 
     * and make sense to the user if formatting wasn't there.
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">help.instruction</code> sub-attribute:</caption>
     * &lt;oj-some-element help.instruction="some tooltip">&lt;/oj-some-element>
     *  
     * @example <caption>Get or set the <code class="prettyprint">help.instruction</code> property after initialization:</caption>
     * // Get one subproperty
     * var instr = myComp.help.instruction;
     *
     * // Set one subproperty, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myComponent.setProperty('help.instruction', 'some new value');
     * 
     * // Get all
     * var values = myComponent.help;
     * 
     * // Set all.  Must list every resource key, as those not listed are lost.
     * myComponent.help = {
     *   instruction: 'some new value'
     * };
     * 
     * @ojshortdesc Represents advisory information for the component
     * @expose 
     * @access public
     * @instance
     * @alias help.instruction
     * @ojtranslatable
     * @default ""
     * @memberof! oj.editableValue
     * @ojtsignore
     * @type {string}
     * @since 4.0.0
     */    
    title: "",
    
    /**
     * <p> 
     * The current valid state of the component. It is evaluated on initial render.
     * It is re-evaluated 
     * <ul>
     *   <li>after validation is run (full or deferred)</li> 
     *   <li>when messagesCustom is updated, 
     *   since messagesCustom can be added by the app developer any time.</li>
     *   <li>when showMessages() is called. Since showMessages() moves the 
     *   hidden messages into messages shown,
     *   if the valid state was "invalidHidden" then it would become "invalidShown".</li>
     *   <li>when the required property has changed. If a component is empty and has required
     *   set, the valid state may be "invalidHidden" (if no invalid messages are being shown as well). 
     *   If required property is removed, the valid state would change to "valid".</li> 
     * </ul>
     * </p>
     * <p>
     *  Note: New valid states may be added to the list of valid values in future releases.
     *  Any new values will start with "invalid" 
     *  if it is an invalid state, "pending" if it is pending state, 
     *  and "valid" if it is a valid state.
     * </p>
     * @example <caption>Get <code class="prettyprint">valid</code> attribute, after initialization:</caption>
     * // Getter:
     * var valid = myComp.valid;
     * @example <caption>Set the <code class="prettyprint">on-valid-changed</code>
     *  listener so you can do work in the ViewModel based on the 
     *  <code class="prettyprint">valid</code> property:</caption>
     * &lt;oj-some-element id='username' on-valid-changed='[[validChangedListener]]'>
     * &lt;/oj-some-element>
     * &lt;oj-some-element id='password' on-valid-changed='[[validChangedListener]]'>
     * &lt;/oj-some-element>
     * &lt;oj-button disabled='[[componentDisabled]]' on-click='[[submit]]'>Submit&lt;/oj-button>
     * -- ViewModel --
     * self.validChangedListener = function (event) {
     *   var enableButton;
     *   var usernameValidState;
     *   var passwordValidState;
     *   
     *   // update the button's disabled state.
     *   usernameValidState = document.getElementById("username").valid;
     *   passwordValidState = document.getElementById("password").valid;
     *   
     *   // this updates the Submit button's disabled property's observable based
     *   // on the valid state of two components, username and password.
     *   // It is up to the application how it wants to handle the “pending” state 
     *   // but we think that in general buttons don’t need to be 
     *   // enabled / disabled based on the "pending" state.
     *   enableButton = 
     *    (usernameValidState !== "invalidShown") &&
     *    (passwordValidState !== "invalidShown");
     *   self.componentDisabled(!enableButton);;
     * };
     * 
     * @ojshortdesc The validity state of the component
     * @expose 
     * @access public
     * @instance
     * @type {string}
     * @ojvalue {string} "valid" The component is valid
     * @ojvalue {string} "pending" The component is waiting for the validation state to be determined.
     * The "pending" state is never set in this release of JET. It will be set in a future release.
     * @ojvalue {string} "invalidHidden" The component has invalid messages hidden
     *    and no invalid messages showing. An invalid message is one with severity "error" or higher.
     * @ojvalue {string} "invalidShown" The component has invalid messages showing.
     *  An invalid message is one with severity "error" or higher.
     * @ojwriteback
     * @readonly
     * @memberof oj.editableValue
     * @since 4.2.0 
     * @ojstatus preview
     */
    valid: undefined,
    
    /** 
     * The value of the component. 
     * 
     * <p>
     * When <code class="prettyprint">value</code> property changes due to programmatic 
     * intervention, the component always clears all messages
     * including <code class="prettyprint">messagesCustom</code>, runs deferred validation, and 
     * always refreshes UI display value.</br>
     * 
     * <h4>Running Validation</h4>
     * <ul>
     * <li>component always runs deferred validation; the 
     * <code class="prettyprint">valid</code> property is updated with the result.</li>
     * </ul>
     * </p>
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">value</code> attribute specified:</caption>
     * &lt;oj-some-element value='10'>&lt;/oj-some-element>
     * @example <caption>Get or set <code class="prettyprint">value</code> attribute, after initialization:</caption>
     * // Getter: returns '10'
     * var val = myComp.value;
     * // Setter: sets '20'
     * myComp.value = '20';
     * 
     * @ojshortdesc The value of the editablevalue component
     * @expose 
     * @access public
     * @instance
     * @default null
     * @ojwriteback
     * @memberof oj.editableValue
     * @since 0.6
     * @type {Object|undefined}
     * @ojsignature {
     *                 target: "Accessor",
     *                 value: {
     *                          GetterType: "V",
     *                          SetterType: "SV"}
     *              }
     */
    value: undefined,

    // Events
    /**
     * Triggered when a default animation is about to start on an element owned by the component.
     *
     * <p>The default animation can be cancelled by calling <code class="prettyprint">event.preventDefault</code>, followed by
     * a call to <code class="prettyprint">event.detail.endCallback</code>.  <code class="prettyprint">event.detail.endCallback</code> should be
     * called immediately after <code class="prettyprint">event.preventDefault</code> if the application merely wants to cancel animation, 
     * or it should be called when the custom animation ends if the application is invoking another animation function.  Failure to
     * call <code class="prettyprint">event.detail.endCallback</code> may prevent the component from working properly.</p>
     * <p>For more information on customizing animations, see the documentation of 
     * <a href="oj.AnimationUtils.html">oj.AnimationUtils</a>.</p>
     *
     * <caption>The default animations are controlled via the theme (SCSS) :</caption>
     * <pre class="prettyprint"><code>
     * // default animations for "notewindow" display option
     * $noteWindowOpenAnimation: (effect: "zoomIn", transformOrigin: "#myPosition") !default;
     * $noteWindowCloseAnimation: (effect: "none") !default;
     *
     * // default animations for "inline" display option
     * $messageComponentInlineOpenAnimation: (effect: "expand", startMaxHeight: "#oldHeight") !default;
     * $messageComponentInlineCloseAnimation: (effect: "collapse", endMaxHeight: "#newHeight") !default;
     * </code></pre>
     *
     * @ojshortdesc Triggered when a default animation is about to start, such as when the component is
     * being opened/closed or a child item is being added/removed. The default animation can
     * be cancelled by calling event.preventDefault.
     *
     * @expose
     * @event
     * @memberof oj.editableValue
     * @since 4.0.0
     * @ojbubbles
     * @ojcancelable
     * @instance
     * @property {string} action The action that triggers the animation. Supported values are:
     *                    <ul>
     *                      <li>"inline-open" - when an inline message container opens or increases in size</li>
     *                      <li>"inline-close" - when an inline message container closes or decreases in size</li>
     *                      <li>"notewindow-open" - when a note window opens</li>
     *                      <li>"notewindow-close" - when a note window closes</li>
     *                    </ul>
     * @property {Element} element The element being animated.
     * @property {function():void} endCallback If the event listener calls event.preventDefault to
     *            cancel the default animation, it must call the endCallback function when it
     *            finishes its own animation handling and any custom animation has ended.
     *
     * @example <caption>Define an event listener for the
     *          <code class="prettyprint">ojAnimateStart</code> event to override the default
     *          "inline-open" animation:</caption>
     * var listener = function( event )
     *   {
     *     // Change the "inline-open" animation for inline message
     *     if (event.detail.action == "inline-open") {
     *       // Cancel default animation
     *       event.preventDefault();
     *       // Invoke new animation and call endCallback when the animation ends
     *       oj.AnimationUtils.fadeIn(event.detail.element).then(event.detail.endCallback);
     *     }
     *   };
     *
     * @example <caption>Define an event listener for the
     *          <code class="prettyprint">ojAnimateStart</code> event to cancel the default
     *          "notewindow-close" animation:</caption>
     * var listener = function( event )
     *   {
     *     // Change the "notewindow-close" animation for note window
     *     if (ui.action == "notewindow-close") {
     *       // Cancel default animation
     *       event.preventDefault();
     *       // Call endCallback immediately to indicate no more animation
     *       event.detail.endCallback();
     *   };
     */
    animateStart : null,
    /**
     * Triggered when a default animation has ended.
     *
     * @expose
     * @event
     * @ojbubbles
     * @ojcancelable
     * @memberof oj.editableValue
     * @since 4.0.0
     * @instance
     * @property {string} action The action that triggers the animation. Supported values are:
     *                    <ul>
     *                      <li>"inline-open" - when an inline message container opens or increases in size</li>
     *                      <li>"inline-close" - when an inline message container closes or decreases in size</li>
     *                      <li>"notewindow-open" - when a note window opens</li>
     *                      <li>"notewindow-close" - when a note window closes</li>
     *                    </ul>
     * @property {Element} element The element being animated.
     * @example <caption>Define an event listener for the
     *          <code class="prettyprint">ojAnimateEnd</code> event to add any processing after the end of
     *          "inline-open" animation:</caption>
     * var listener = function( event )
     * {
     *   // Check if this is the end of "inline-open" animation for inline message
     *   if (event.detail.action == "inline-open") {
     *     // Add any processing here
     *   }
     * };
     */
    animateEnd : null
  },
  
  // P U B L I C    M E T H O D S
  
  // @inheritdoc
  getNodeBySubId: function(locator)
  {
    var node;
    var subId;

    node = this._super(locator);
    
    // this subId is only for non-custom element components so skip if custom element
    if (!node && !this._IsCustomElement())
    {
      subId = locator['subId'];

      if (subId === "oj-label-help-icon")
      {
        var label = this._GetLabelElement();

        if (label)
          node = label.parent().find(".oj-label-help-icon");
      }
    }
    // Non-null locators have to be handled by the component subclasses
    return node || null;
  },
  //** @inheritdoc */
  getSubIdByNode : function(elem)
  {
    var $node,
        anchor,
        div,
        label,
        subId = null;
    
    if (elem != null)
    {
      $node = $(elem);
      anchor = $node.closest("a.oj-label-help-icon");
       
      if (anchor != null)
      {
        // Go up to the top level element of the label
        div = anchor.closest(".oj-label");
        
        if (div != null)
        {
          // Now find the actual label
          label = div.find("label")[0];
          
          if (label)
          {
            // Make sure the label is the one associated with this component
            if (label == this._GetLabelElement()[0])
            {
               subId = {"subId":"oj-label-help-icon"};
            }
          }
        }
      }
    }
     
    return subId;
  },
  /**
   * whether the component is currently valid.  It is valid if it doesn't have any errors.
   * This method is final; do not override.
   * Currently this is for the widget components and not custom elements.
   * @example <caption>Check whether the component is valid:</caption>
   * var valid = myInputElement.isValid();
   * @returns {boolean}
   * @access public
   * @instance
   * @expose
   * @memberof oj.editableValue
   * @final
   * @ignore
   */
  isValid : function ()
  {
    if (this._valid === undefined)
    {
      this._valid = !this._hasInvalidMessages();
    }
    
    return this._valid;
  },
  
  /**
   * Called when the DOM underneath the component changes requiring a re-render of the component. An 
   * example is when the <code class="prettyprint">id</code> for the input changes. <br/>
   * <p>
   * Another time when refresh might be called is when the locale for the page changes. When it 
   * changes, attributes used by its converter and validator that are locale specific, its hints, 
   * messages and translations will be updated. 
   * </p>
   * 
   * <p>
   * When <code class="prettyprint">refresh</code> method is called, the component may take various 
   * steps such as clearing messages, running validation etc., based on the state it is in. </br>
   * 
   * <h4>Steps Performed Always</h4>
   * <ul>
   * <li>The converter and validators used by the component are reset, and new converter and 
   * validator hints is pushed to messaging. E.g., notewindow displays the new hint(s).
   * </li>
   * </ul>
   *  
   * <h4>Running Validation</h4>
   * <ul>
   * <li>if component is valid when refresh() is called, the display value is refreshed if component 
   * has a converter set.</li>
   * <li>if component is invalid and is showing messages when 
   * <code class="prettyprint">refresh()</code> is called, then all component messages are cleared 
   * and full validation run using the display value on the component. 
   * <ul>
   *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
   *   attribute is not updated and the error is shown. 
   *   </li>
   *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
   *   attribute is updated; page author can listen to the <code class="prettyprint">onValueChanged</code> 
   *   event to clear custom errors.</li>
   * </ul>
   * </li>
   * <li>if component is invalid and has deferred messages when <code class="prettyprint">refresh()</code> 
   * is called, then all component messages are cleared and deferred validation is run.</li>
   * </ul>
   * </p>
   * 
   * <h4>Clearing Messages</h4>
   * <ul>
   * <li>If clearing messages only those created by the component are cleared.
   * <li><code class="prettyprint">messagesCustom</code> attribute is not cleared.</li>
   * </ul>
   * </p>
   * 
   * @example <caption>Redraw the component element.</caption>
   * myComp.refresh();
   * 
   * @access public
   * @instance
   * @expose
   * @return {void}
   * @memberof oj.editableValue
   * @since 0.7
   */
  refresh : function ()
  {
    this._super();
    this._doRefresh(true);
  },
  
  /**
   * Resets the component by clearing all messages and messages attributes - 
   * <code class="prettyprint">messagesCustom</code> -
   * and updates the component's display value using the attribute value. User entered values will be 
   * erased when this method is called.
   * 
   * @example <caption>Reset component</caption>
   * myComp.reset(); <br/>
   * 
   * @access public
   * @instance
   * @expose
   * @return {void}
   * @memberof oj.editableValue
   * @since 0.7
   */
  reset : function ()
  {
    this._clearAllMessages();
    // since we are pushing component value to UI, only deferred validation need to be run; this is
    // same as setting value option.
    this._runDeferredValidation(this._VALIDATION_CONTEXT.RESET_METHOD);
    this._refreshComponentDisplayValue(this.options['value'], true);
  },
  
  /**
   * Takes all deferred messages and shows them. 
   * It then updates the valid property; e.g.,
   * if the valid state was "invalidHidden"
   * before showMessages(), the valid state will become "invalidShown" after showMessages(). 
   * <p>
   * If there were no deferred messages this method simply returns. 
   * </p>
   * 
   * 
   * @example <caption>Display all messages including deferred ones.</caption>
   * myComp.showMessages();
   * @access public
   * @instance
   * @return {void}
   * @expose
   * @memberof oj.editableValue
   * @since 0.7
   */
  showMessages : function ()
  {
    var clonedMsgs = [];
    var i;
    var msgHidden;
    var msgsHidden = this.options['messagesHidden'];
    var hasMessagesHidden = msgsHidden.length > 0;
    var clonedMsg;
    
    // showMessages() copies messagesHidden into clonedMsgs, 
    // then clears messagesHidden, and updates messagesShown with the clonedMsgs. 
    // It then updates the valid property; e.g.,
    // if the valid state was "invalidHidden"
    // before showMessages(), the valid state will become "invalidShown" after showMessages(). 
    for (i = 0; i < msgsHidden.length; i++)
    {
      msgHidden = msgsHidden[i];
      
      // The oj.Message and oj.ComponentMessage distinction is important in a few places in
      // the EV code. For example, when messagesCustom property changes, we keep existing
      // messagesShown only if they are ComponentMessage added by component.
      // if (msg instanceof oj.ComponentMessage && msg._isMessageAddedByComponent())
      if (msgHidden instanceof oj.ComponentMessage)
      {
        // change oj.ComponentMessage's display state to oj.ComponentMessage.DISPLAY.SHOWN
        msgHidden._forceDisplayToShown();
        

        // 
        // .clone() clones the message and the options that were passed in when the message
        // was originally created.
        clonedMsg = msgHidden.clone();
      }
      else
        // NOTE: oj.Message is a public class and oj.ComponentMessage is private.
        // oj.Message's .clone is deprecated, so we purposely don't call it here.
        clonedMsg = new oj.Message(msgHidden['summary'], msgHidden['detail'], msgHidden['severity']);
      
      clonedMsgs.push(clonedMsg);
    }
    
    if (hasMessagesHidden)
    {

      this._clearMessages('messagesHidden');
      
      this._updateMessagesOption('messagesShown', clonedMsgs);
      
      this._setValidOption(null);
    }
  },

  
  // P R O T E C T E D    C O N S T A N T S   A N D   M E T H O D S

  // *********** START WIDGET FACTORY METHODS (they retain _camelcase naming convention) **********
  
  /**
   * Validation mode specifying the kind of validation that gets run.
   * <ul>
   *   <li>FULL - the default and runs both the converter and all validators. </li>
   *   <li>VALIDATORS_ONLY - runs all validators including the required validator is run.</li>
   *   <li>REQUIRED_VALIDATOR_ONLY - runs just the required validator. </li>
   * </ul>  
   * @protected
   * @const
   * @type {Object}
   * @memberof oj.editableValue
   */       
  _VALIDATION_MODE : _sValidationMode,
  
  /**
   * The context the component can be in when validation is run. 
   * <ul>
   * <li>COMPONENT_CREATE - when component is created and we run validators. usually messages are 
   * not displayed right away, i.e, are 'deferred'. </li>
   * <li>VALUE_OPTION_CHANGE - when component's value is updated programmatically. messages are  
   * deferred.</li>
   * <li>REQUIRED_OPTION_CHANGE - when component's required option is updated programmatically. messages are
   * deferred.</li>
   * <li>USER_ACTION - when component runs validation as a result of user interating with component.
   * messages are displayed immediately.</li>
   * <li>VALIDATE_METHOD - when component's validate() method is called explicitly. messages are 
   * displayed immediately.</li>
   * </ul>
   * 
   * @protected
   * @const
   * @type {Object}
   * @memberof oj.editableValue
   * 
   */
  _VALIDATION_CONTEXT : oj.EditableValueUtils.validationContext,
  
  /**
   * Default options used by validate method. 
   * 
   * @protected
   * @const
   * @type {Object}
   * @memberof oj.editableValue
   * @see #validate
   */  
  _VALIDATE_METHOD_OPTIONS : oj.EditableValueUtils.validateMethodOptions,  
  
  /**
   * Called at component create time primarily to initialize options, often using DOM values. This 
   * method is called before _ComponentCreate is called, so components that override this method 
   * should be aware that the component has not been rendered yet. The element DOM is available and 
   * can be relied on to retrieve any default values. <p>
   * 
   * This method sets defaults for its options that have a DOM namesake. E.g., value, required, 
   * disabled etc. Subclasses can override this method to set their own defaults for these options.
   * Example, the value option is often not set on this.element for components like radioset, which
   * walk the sub-tree to determine the value.
   * 
   * @param {!Object} originalDefaults - original default options defined on the widget and its ancestors
   * @param {?Object} constructorOptions - options passed into the wiget constructor
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _InitOptions : function(originalDefaults, constructorOptions)
  {
    this._super(originalDefaults, constructorOptions); 
    
    // We need to do this for custom element components since
    // help.instruction is used instead of title.
    if (this._IsCustomElement())
    {
      // update the title option if help.instruction option exists
      var help = this.options["help"];

      if (help != null)
      {
        var helpInstruction = help["instruction"];

        if (helpInstruction != null)
        {
          this.options["title"] = helpInstruction;
        }
      }
      
      // update displayOptions.title from displayOptions.helpInstruction
      var dispOpts = this.options["displayOptions"];
      
      if (dispOpts != null)
      {
        var helpInstruction = dispOpts["helpInstruction"];

        if (helpInstruction != null)
        {
          dispOpts["title"] = helpInstruction;
          this.options["displayOptions"] = dispOpts;
        }
      }
    }
    
    
  },

  /**
   * Initializes options defined by this base class.
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _ComponentCreate : function ()
  {
    //remove attributes that trigger html5 validation + inline bubble
    var attrsToRemove = ["required", "title", "pattern"]; 
    var node = this.element;
    var savedAttributes = this._GetSavedAttributes(node); 
    
    this._super();
    
    this.options['messagesCustom'] = this.options['messagesCustom'] || [];
    this.options['messagesHidden'] = []; 
    this.options['messagesShown'] = this.options['messagesCustom'].length > 0 ? 
      this._cloneMessagesBeforeSet(this.options['messagesCustom']) : [];
    
    // update element DOM for disabled. TODO: say why
    this._SetDisabledDom(node);
    
    // we do this here instead of in _InitOptions because here we have the final value.
    // - an empty placeholder shows up if data changed after first binding 
    if (this._HasPlaceholderSet())
    {
      // update element placeholder
      this._SetPlaceholder(this.options['placeholder']);
      this._customPlaceholderSet = true;
    }
    
    // remove html5 validation attributes; it's safe to remove these here because components should 
    // have already initialized options based on DOM in _InitOptions().
    // Only need to do for <input> elements
    // custom elements create their input/textarea dom, so no need to do that for them.
    if (savedAttributes && !this._IsCustomElement())
    {
      var tagName = node[0].tagName.toLowerCase();
      if (tagName === 'input' || tagName === 'textarea')
      {
        $.each(attrsToRemove, function (index, value)
        {
          if (value in savedAttributes)
          {
            node.removeAttr(value);
          }
        });
      }
    }
  },
  
  /**
   * The value option alone is initialized here since it requires the component to be fully
   * created. Calling this.options.value before this method does not guarantee the correct 
   * value to be returned.
   *
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _AfterCreate : function ()
  {
    var describedBy;
    
    this._super();
    
    // refresh value, theming and aria attributes
    this._doRefresh(false);

    // create an ojLabel only if this isn't a custom element. For example, ojInputText will
    // create an ojLabel, but <oj-input-text> will not. Instead the app dev uses <oj-label>.
    if (!this._IsCustomElement())
      this._createOjLabel();
    else
    {
      // If this is a custom element, we need to write the default style class + "-label" 
      // onto the oj-label. The reason is so
      // we can theme the oj-label based on what form control it is labeling. We use this in
      // our themes to line up the label to the form control, 
      // e.g., oj-label-inline.oj-radioset-label {margin-top:.5em;}.
      if (this.customOjLabelElement === undefined)
        this.customOjLabelElement = this._getCustomOjLabelElement();
      if (this.customOjLabelElement)
        $(this.customOjLabelElement).addClass(this._GetDefaultStyleClass()+"-label");
    }


    // set describedby on the element as aria-describedby
    describedBy = this.options['describedBy'];

    if (describedBy)
    {
      this._addAriaDescribedBy(describedBy);
    }

    // initialize component messaging
    this._initComponentMessaging();
    
    // run deferred validation 
    this._runDeferredValidation(this._VALIDATION_CONTEXT.COMPONENT_CREATE);
    
    // trigger messagesShownChanged for messagesShown if it's non-empty. 
    // this.options['messagesShown'] would have been 
    // updated in _ComponentCreate if messagesCustom was non-empty. Because we are setting
    // the 'changed' flag to true, the messagesShownChanged event will be fired, and that's what we want.
    if (this.options['messagesShown'].length > 0)
    {
      this._setMessagesOption('messagesShown', this.options['messagesShown'], null, true);
    }

    this.widget().addClass("oj-form-control");
  },
  /**
   * <p>Saves all the element's attributes. In _destroy all attributes will be restored.
   * </p>
   *
   * @param {Object} element - jQuery selection to save attributes for
   * @protected
   * @memberof oj.editableValue
   * @instance
   * @override
   */
  _SaveAttributes : function (element)
  {
    if (!this._IsCustomElement())
      this._SaveAllAttributes(element);
  },
  /**
   * @protected
   * @memberof oj.editableValue
   * @instance
   * @override
   */
  _RestoreAttributes : function (element)
  {
    if (!this._IsCustomElement())
      this._RestoreAllAttributes(element);
  },
  /**
   * Performs post processing after _SetOption() calls _superApply(). Different options, when changed, perform
   * different tasks. 
   * 
   * @param {string} option
   * @param {Object=} flags 
   * @protected
   * @memberof oj.editableValue
   * @instance
   */
  _AfterSetOption : function (option, flags)
  {
    switch (option)
    {
      case "disabled":
        this._AfterSetOptionDisabledReadOnly(option, oj.EditableValueUtils.disabledOptionOptions);
        break;

      case "displayOptions" :
        // For custom element, we use displayOptions.helpInstruction instead of displayOptions.title
        // so internally we need to update title with the value of helpInstruction
        if (this._IsCustomElement())
        {
          var dispOpts = this.options["displayOptions"];
          
          if (dispOpts != null)
          {
            var helpInstruction = dispOpts["helpInstruction"];

            if (helpInstruction != null)
            {
              dispOpts["title"] = helpInstruction;
              this.options["displayOptions"] = dispOpts;
            }
          }
        }
        
        // clear the cached merged options; the getter setup for this.options['displayOptions']
        // will merge the new value with the defaults
        this._initComponentMessaging();
        break;

      case "help":
        if (!this._IsCustomElement())
          this._Refresh(option, this.options[option]);
        
        // For custom element components, we use help.instruction instead of title
        // so we need to update the internal title option with the value of
        // help.instruction
        if (this._IsCustomElement())
        {
          this.options["title"] = this.options.help["instruction"];
          this._titleOptionChanged();
        }
        
        break;

      case "messagesCustom":
        this._messagesCustomOptionChanged(flags);
        this._setValidOption(null);
        break; 
        
      case "placeholder":
        this._placeholderOptionChanged(flags);
        break;

      case "title":
        // no reason to refresh component when title changes.
        this._titleOptionChanged();
        break;
        
      case "translations": 
        this.refresh();
        break;

      case 'value':
        this._AfterSetOptionValue(option, flags);
        break;
  
      default:
        break;
    }
    
  },
  
  /**
   * Performs post processing after disabled or readOnly option changes by taking the following 
   * steps. (Steps are same for readOnly option).
   * <p>
   * if disabled component is enabled then, <br/>
   * - if there are no errors, run deferred validation. component could have been initialized with 
   * empty value and disabled.<br/>
   * - if component is invalid and showing messages clear component error, grab UI value and run 
   * full validation.<br/>
   * - if component is invalid and has hidden messages; do nothing. <br/>
   * </p>
   * <p>
   * if enabled component is disabled no validation is run.<br/>
   * </p>
   * 
   * @param {String} option
   * @param {Object} validationOptions
   * 
   * @protected
   * @memberof oj.editableValue
   * @instance
   */
  _AfterSetOptionDisabledReadOnly : function (option, validationOptions)
  {
    var isEnabled = !(this.options[option] || false);
    
    // always refresh
    this._Refresh(option, this.options[option]);
    if (isEnabled)
    {
      this._runMixedValidationAfterSetOption(validationOptions);
    }
  },

  /**
   * Performs post processing after value option changes by taking the following steps.
   * 
   * - triggers a valueChanged event and does writeback if required.<br/>
   * - if setOption was from programmatic intervention, <br/>
   * &nbsp;&nbsp;- clear custom messages and component messages; <br/>
   * &nbsp;&nbsp;- run deferred validation. if there is an error, updates messagesHidden. <br/>
   * - always refreshes UI display <br/>
   * 
   * @param {string} option
   * @param {Object=} flags
   * 
   * @protected
   * @memberof oj.editableValue
   * @instance
   * 
   */
  _AfterSetOptionValue : function(option, flags)
  {
    var context = flags ? flags['_context'] : null;
    var doNotClearMessages;
    var isUIValueChange = false;

    if (context)
    {
      isUIValueChange = context.originalEvent ? true : false;
      doNotClearMessages = context.doNotClearMessages || false;
    }

    if (!isUIValueChange)
    {
      // value option can be updated directly (i.e., programmatically or through user interaction) 
      // or updated indirectly as a result of some other option changing - e.g., converter, 
      // validators, required etc. See _updateValue() method for details.
      // When value changes directly due to programatic intervention (usually page author does this) 
      // then clear all messages and run deferred validation.
      // If value changes indirectly do not clear custom messages (component messages are already 
      // cleared) and run deferred validation.
      if (!doNotClearMessages) 
      {
        this._clearAllMessages(null);
      }
      this._runDeferredValidation(this._VALIDATION_CONTEXT.VALUE_OPTION_CHANGE);
    }
    
    // refresh UI display value
    this._Refresh(option, this.options[option], true);
  },
  
  /**
   * Whether the a value can be set on the component. For example, if the component is 
   * disabled then setting value on component is a no-op. 
   * 
   * @see #_SetValue
   * @return {boolean}
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _CanSetValue: function ()
  {
    var disabled = this.options['disabled'] || false;
    
    return (disabled) ? false : true;
  },
  
  /**
   * Detaches the widget from the element and restores element exactly like it was before the widget 
   * was attached.
   * @protected
   * @expose
   * @memberof oj.editableValue
   * @instance
   * 
   */
  _destroy : function ()
  {
    var labelIndex;
    var labelLength;
    var ret = this._super();
    
    this._clearAllMessages(null, true);
    this._getComponentMessaging().deactivate();

    // make sure the label is still "alive". Otherwise we could get error when we try to 
    // destroy it if the dom was removed first and ojLabel was destroyed directly.
    // also make sure to check if there is more than one label and destroy them individually.
    if (this.$label)
    {
      labelLength = this.$label.length;
      for (labelIndex = 0; labelIndex < labelLength; labelIndex++) 
     {
       if (this.$label[labelIndex] && 
           oj.Components.__GetWidgetConstructor(this.$label[labelIndex]) != null)
       {
         $(this.$label[labelIndex]).ojLabel( "destroy" );
       } 
     }     
    }


    return ret;
  },

  /**
   * @memberof oj.editableValue
   * @instance
   * @override
   * @protected
   * @since 5.0.0
   */
  GetFocusElement : function ()
  {
    return this._GetContentElement()[0];
  },
  
  /**
   * Called (by the widget factory) when the option changes, this method responds to the change 
   * by refreshing the component if needed. This method is not called for the options passed in 
   * during the creation of the widget.
   * 
   * @param {string} name of the option
   * @param {Object|string} value
   * @param {Object?} flags - optional flags. The following flags are currently supported:
   * <ul>
   *  <li>changed - true if the caller wants to indicate the value has changed, so no comparison is necessary</li>
   * </ul>
   * 
   * @expose
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _setOption : function (name, value, flags) 
  {
    var retVal;
    var skipSetOption = false;
    var oldValue;
    var newValue;
    
    // Step 1: Remember previous values
    if (typeof name === "string" && value !== undefined)
    {
      switch (name)
      {
        case "messagesHidden":
          // this option can never be set programmatically by page author
          skipSetOption = true;
          break;          
          
        case "messagesShown":
          // this option can never be set programmatically by page author
          skipSetOption = true;
          break;
        
        case "rawValue":
          // rawValue is readOnly, so throw an error here.
          skipSetOption = true;
          break;
                
        case "describedBy":
          // This sets the aria-describedby on the correct dom node
          oldValue = this.options['describedBy'];
          newValue = value;

          if (oldValue)
          {
            this._removeAriaDescribedBy(oldValue);
          }
          if (newValue)
          {
            this._addAriaDescribedBy(newValue);
          }
    
          break;
        }
    }
    
    
    if (skipSetOption)
    {
      oj.Logger.error(name + " option cannot be set");
      return this;
    }
    
    // Step 2: Update option value 
    retVal = this._superApply(arguments);
    
    // Step 3: Do post processing like triggering events, refreshing component DOM etc.
    this._AfterSetOption(name, flags);

    return retVal;
  },
  
  // *********** END WIDGET FACTORY METHODS **********
  
  /**
   * Returns a jquery object of the element representing the content node. This could be a jQuery 
   * object of the element the widget was invoked on - typically this is an input or select or 
   * textarea element for which a value can be set.
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   * @return {jQuery} the jquery element that represents the editable content. E.g., an input
   */
  _GetContentElement : function ()
  {
    return this.element;
  },

  /**
   * Returns a jquery object of the element(s) representing the label node(s) for the input 
   * component. 
   * First we look for the aria-labelledby attribute on the input.
   * If that's not found, we look for the label with 'for' attribute 
   * pointing to input.
   * If that's not found, we walk up the dom looking for aria-labelledby.
   * Note: multiple labels for one input is legal in html-5.
   * @memberof oj.editableValue
   * @instance
   * @protected
   * @return {Object} the jquery element that represents the input component's label.
   *  return null if it can't find anything.
   */
  _GetLabelElement : function ()
  {
    var ariaElement;
    var labelQuery;
    
    if (this.$label)
      return this.$label;
    
    // If input has aria-labelledby set, then look for label it is referring to.
    var queryResult = this._getAriaLabelledByElement(this.element);
    if (queryResult !== null && queryResult.length !== 0)
    {
      return queryResult;
    }
    
    // if no aria-labelledby is on the input, then look for a label with 'for'
    // set.
    var id = this.element.prop("id");
    if (id !== undefined)
    {
      labelQuery = "label[for='" + id + "']";
      queryResult = $(labelQuery);
      if (queryResult.length !== 0)
      {
        return queryResult;
      }
    }
 
    // if no aria-labelledby on input and no label with 'for' pointing to input,
    // then as a final step we walk up the dom to see if aria-labelledby is set.
    // If so, then we find the label it is referring to.
    // This would be the case when you have multiple inputs grouped in a div 
    // <label id="grouplabel">Address</label>
    // <div aria-labelledby="grouplabel"><input/><input/><input/></div>
    ariaElement = this.element.closest("[aria-labelledby]");
    if (ariaElement.length !== 0)
    {
      // Element has aria-labelledby set, so look for label it is referring to.
      queryResult = this._getAriaLabelledByElement(ariaElement);
      if (queryResult !== null && queryResult.length !== 0)
      {
        return queryResult;
      }
    }
    return null;

  },
  
  
  /**
   * Returns the element's value. Normally, this is a call to this.element.val(), but for some 
   * components, it could be something else. E.g., for ojRadioset the element's value is really the 
   * value of the selected radio in the set. 
   * 
   * @override
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _GetElementValue : function () 
  {
    return this.element.val();
  },
  

/**
   * Returns a jquery object of the element that triggers messaging behavior. The trigger element 
   * is usually an input or select or textarea element for which a value can be set/retrieved and 
   * validated. 
   * 
   * @return {jQuery} jquery object 
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _GetMessagingLauncherElement : function ()
  {
    return this._GetContentElement();
  },
  
  /**
   * Returns the normalized converter instance. Since EditableValue does not have a converter 
   * option, this returns null.
   * 
   * @return {Object} null
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _GetConverter : function () 
  {
    return null;
  },

  /**
   * Returns an array of implicit validators setup by component. This list contains validators for 
   * the internal use of the component and are not a part of this.options.validators. <br/>
   * E.g., if the pattern attribute or option is set, a RegExpValidator instance is automatically 
   * created and added to this list. <br/>
   * RequiredValidator is tracked separately from the default validators.
   * 
   * @return {Object} a map of string name to the validator instance. 
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _GetImplicitValidators : function ()
  {
    if (!this._implicitValidators)
    {
      this._implicitValidators = {};
    }
    
    return this._implicitValidators;
  },
          
  /**
   * Returns the display value that is ready to be passed to the converter.
   * 
   * @param {Object} value the stored value if available that needs to be formatted for display
   * @return {string} usually a string display value
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _GetDisplayValue : function (value) 
  {
    return this._GetContentElement().val();
  },
  /**
   * For components that have the 'validators' option, 
   * this returns an array of all validators 
   * normalized from the validators option set on the component. <br/>
   * Since EditableValue does not include the 'validators' option, this returns [].
   * 
   * @return {Array} of validators. 
   * Since EditableValue does not include the 'validators' option, this returns [].
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _GetNormalizedValidatorsFromOption : function ()
  {
    return [];
  },
  /**
   * Returns an array of all validators built by merging the validators option set on the component 
   * and the implicit validators setup by the component. <br/>
   * This does not include the implicit required validator. Components can override to add to this 
   * array of validators.
   * 
   * @return {Array} of validators
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _GetAllValidators : function ()
  {
    var allValidators;
    var idx;
    var implicitValidatorMap;
    var implicitValidators;
    var normalizedValidators;
          
    if (!this._allValidators)
    {
      allValidators = [];
      implicitValidatorMap = this._GetImplicitValidators(), 
      implicitValidators = [];

      // combine public and implicit validators to get the combined list
      var keys = Object.keys(implicitValidatorMap), valType;
      var len  = keys.length;
      if (len > 0)
      {
        for (idx = 0; idx < len; idx++) 
        {
          valType = keys[idx];
          implicitValidators.push(implicitValidatorMap[valType]);
        }
        allValidators = allValidators.concat(implicitValidators);
      }
        
      normalizedValidators = this._GetNormalizedValidatorsFromOption();
      if (normalizedValidators.length > 0)
      {
        // Add normalize validators 
        allValidators = allValidators.concat(normalizedValidators);
      }
      
      this._allValidators = allValidators;
    }
    
    return this._allValidators || [];
  },
  
  /**
   * EditableValue caches the validators to be run, within this._allValidators variable.
   * This is great; however when the implicit validator needs to be reset [i.e. min + max changing] 
   * or the validators option changes, then the cached this._allValidators needs to be cleared. 
   * This method also updates the messaging strategies as hints associated with validators could 
   * have changed.
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _ResetAllValidators : function () 
  {
    if (this._allValidators)
    {
      this._allValidators.length = 0;
    }
    this._allValidators = null;
    
    // update messagingstrategy as hints associated with validators could have changed
    this._getComponentMessaging().update(this._getMessagingContent(
            this._MESSAGING_CONTENT_UPDATE_TYPE.VALIDATOR_HINTS));

  },
  
  /**
   * Whether the component is required. EditableValue doesn't support the required option,
   * so return false.  This allows us to keep some of the required code in EditableValue.
   * 
   * @return {boolean} true if required; false
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _IsRequired : function () 
  {
    return false;
  },
  
  /**
   * Convenience handler for the DOM 'change' event. Subclasses are expected to wire up event 
   * handlers for DOM events that they wish to handle.<br/>
   * 
   * The implementation retrieves the display value for the component by calling _GetDisplayValue() 
   * and calls _SetValue(), with full validation.
   * 
   * @param {Event} event DOM event 
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _HandleChangeEvent : function (event) 
  {
    var submittedValue = this._GetDisplayValue();
    // run full validation
    this._SetValue(submittedValue, event);
  }, 
  /**
   * Convenience function to set the rawValue option. Called by subclasses
   * 
   * @param {String} val value to set rawValue to
   * @param {Event} event DOM event 
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _SetRawValue : function (val, event)
  {
    var flags = {};
    flags['_context'] = {originalEvent: event, writeback: true, internalSet: true, readOnly: true};

    if (!this._CompareOptionValues('rawValue', this.options['rawValue'], val))
      this.option("rawValue", val, flags);
  },

  /**
   * Called in response to a change in the options set for this component, this method refreshes the 
   * component display value. Subclasses can override to provide custom refresh behavior.
   * 
   * @param {String=} name the name of the option that was changed
   * @param {Object=} value the current value of the option
   * @param {boolean=} forceDisplayValueRefresh
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _Refresh : function (name, value, forceDisplayValueRefresh)
  {
    var helpDef;
    var helpSource;
    var id;

    switch (name)
    {
      case "converter":
        value = this.options['value'];
        this._refreshComponentDisplayValue(value, forceDisplayValueRefresh);
        break;
        
      case "disabled":
        this._refreshTheming("disabled", this.options['disabled']);
        break;
        
      case "help":
        if (!this._IsCustomElement())
        {
          // refresh the help - need to keep the label in sync with the input.
          if (this.$label)
          {
            helpDef = this.options.help["definition"];
            helpSource = this.options.help["source"];
            this.$label.ojLabel("option", "help", 
                                 {"definition" : helpDef, "source" : helpSource});
            this._refreshDescribedByForLabel();

          }
        }
        break;
     
      case "value":
        this._refreshComponentDisplayValue(value, forceDisplayValueRefresh); 
        break;
        
    }
  },

  /**
   * <p>Notifies the component that its subtree has been made hidden programmatically 
   * after the component has been created.
   *
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _NotifyHidden: function()
  {
    this._superApply(arguments);
    this._getComponentMessaging().close();
  },
  /**
   * <p>Notifies the component that its subtree has been removed from the 
   * document programmatically after the component has
   * been created.
   *
   * @memberof oj.baseComponent
   * @instance
   * @protected
   */
  _NotifyDetached: function()
  {
    this._superApply(arguments);
    this._getComponentMessaging().close();
  },
  /**
   * Called anytime the label DOM changes requiring a reset of any dependent feature that caches the 
   * label, including all validators.
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected   
   */
  _ResetComponentState : function ()
  {
    // the DOM for the label and its text could have changed. 
    if (this.$label)
    {
      this.$label.ojLabel("refresh");
    }

    // reset all validators when label changes
    this._implicitReqValidator = null;
    this._converter = null;
    this._ResetAllValidators();
  },  
  
  /**
   * Called when the display value on the element needs to be updated. This method updates the 
   * (content) element value. Widgets can override this method to update the element appropriately. 
   * 
   * @param {String} displayValue of the new string to be displayed
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
  */  
  _SetDisplayValue : function (displayValue) 
  {
    var contentElem = this._GetContentElement();
    if (contentElem.val() !== displayValue)
      contentElem.val(displayValue);
  },
  /**
   * Sets the disabled option onto the dom. 
   * Component subclasses can override this method to not do this in cases where it is invalid, 
   * like on a div (e.g., radioset's root dom element is a div).
   * @param {Object} node - dom node
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   * @since 1.0.0
   */
  _SetDisabledDom : function(node)
  {
    if (typeof this.options['disabled'] === "boolean")
    {
      node.prop("disabled", this.options['disabled']);
    }
  },
  
  /**
   * Sets the placeholder text on the content element by default. It sets the placeholder attribute
   * on the element. Component subclasses can override this method to control where placeholder text
   * gets set.
   * @param {string} value
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _SetPlaceholder : function(value)
  {
    this._GetContentElement().attr("placeholder", value);
  },
  
  /**
   * Sets the placeholder option with the value. 
   * 
   * @param {string} value
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */  
  _SetPlaceholderOption : function (value)
  {
    this.options['placeholder'] = value;
  },
  
  /**
   * whether the placeholder option is set
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _HasPlaceholderSet : function()
  {
    // - an empty placeholder shows up if data changed after first binding 
    return this.options['placeholder'];
  },  
    
  /**
   * Clear the placeholder option
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _ClearPlaceholder : function()
  {
    // - an empty placeholder shows up if data changed after first binding 
    this._SetPlaceholderOption("");
    this._SetPlaceholder("");
  },

  /**
   * Runs full validation on the newValue (usually the display value) and sets the parsed value on 
   * the component if value passes basic checks and there are no validation errors. <br/>
   * If the newValue is undefined or if it differs from the last saved displayValue this method 
   * skips validation and does not set value (same as ADF).<br/>
   * 
   * @param {string|Object} newValue the ui value that needs to be parsed before it's set.
   * @param {Object=} event an optional event if this was a result of ui interaction. For user 
   * initiated actions that trigger a DOM event, passing this event is required. E.g., if user action 
   * causes a 'blur' event.
   * @param {Object=} options - an Object literal that callers pass in to determine how validation 
   * gets run. 
   * @param {boolean=} options.doValueChangeCheck - if set to true compare newValue with last 
   * displayValue before running validation; if false, always run validation. E.g., set to false 
   * when validate() is called.
   * @param {boolean=} options.doNotClearMessages - if set method will not clear all messages. This 
   * is provided for callers that may want to clear only some of the messages. E.g., when required 
   * option changes, it clears only component messages, not custom.
   * @param {number=} options.validationMode - accepted values (defined in _VALIDATION_MODE) are:
   * <ul>
   *   <li>FULL - the default and runs both the converter and all validators. </li>
   *   <li>VALIDATORS_ONLY - runs all validators including the required validator is run.</li>
   *   <li>REQUIRED_VALIDATOR_ONLY - runs just the required validator. </li>
   * </ul>
   * @return {boolean} false if value was not set due to validation error. 
   * @example  <caption>Widget subclasses can use this convention to run full validation</caption>
   * this._SetValue(value, event);
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   * 
   */
  _SetValue : function (newValue, event, options)
  {
    var doValueChangeCheck = (options && typeof options.doValueChangeCheck === "boolean") ? 
                            options.doValueChangeCheck : true, parsed;
    
    // disallow setting a value of undefined by widgets
    if (newValue === undefined)
    {
      oj.Logger.warn("Attempt to set a value of undefined");
      return false;
    }

    if (!doValueChangeCheck || newValue !== this._getLastDisplayValue()) 
    {
      parsed = this._Validate(newValue, event, options);
           
      this._setValidOption(event);
      
      // if validation passed
      if (parsed !== undefined && this.isValid())
      {
        var updateContext;
        if (options && options['_context'])
        {
          updateContext = options['_context'];
        }
        // update value option
        this._updateValueOption(parsed, event, options && options.validationContext,
            updateContext);
        return true;
      }
    }
    else
    {
      if (oj.Logger.level > oj.Logger.LEVEL_WARN)
      {
        oj.Logger.info("Validation skipped and value option not updated as submitted value '" + 
                (newValue.toString) ? newValue.toString() : newValue + " same as previous.");
      }
    }
    
    return false;
  },
  
  /**
   * Runs full validation on the value. If value fails basic checks 
   * (see <a href="#_CanSetValue">_CanSetValue</a>, or if value failed validation, this method 
   * returns false. Otherwise it returns true.
   * 
   * <p>
   * Components should call this method if they know UI value has changed and want to set the 
   * new component value. 
   * </p>
   * 
   * @param {string|Object} newValue the actual value to be set. Usually this is the string display value
   * @param {Object=} event an optional event if this was a result of ui interaction. For user 
   * initiated actions that trigger a DOM event, passing this event is required. E.g., if user action 
   * causes a 'blur' event.
   * @param {{doNotClearMessages:boolean,validationContext:number,validationMode:number}=} options 
   * an Object literal that callers pass in to determine how validation gets run. 
   * @param {boolean=} options.doNotClearMessages - if set method will not clear all messages. This 
   * is provided for callers that may want to clear only some of the messages. E.g., when required 
   * option changes, it clears only component messages, not custom.
   * @param {number=} options.validationContext - the context this method was called. When not set it 
   * defaults to _VALIDATION_CONTEXT.USER_ACTION.
   * @param {number=} options.validationMode - accepted values defined in _VALIDATION_MODE
   * 
   * @return {Object|string|undefined} the parsed value or undefined if validation failed
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   */
  _Validate : function (newValue, event, options)
  {
    var mode = options && options.validationMode ? options.validationMode : 
                this._VALIDATION_MODE.FULL; 
    var context = options && options.validationContext ? options.validationContext : 
                this._VALIDATION_CONTEXT.USER_ACTION;
    var doNotClearMessages = options && options.doNotClearMessages || false, result;

    // disallow setting a value of undefined by widgets
    if (newValue === undefined)
    {
      oj.Logger.warn("Attempt to set a value of undefined");
    }
    else if (this._CanSetValue())
    {
      if (!doNotClearMessages)
      {
        this._clearAllMessages(event);
      };
      
      this._setLastSubmittedValue(newValue);
      
      try
      {
        return this._runNormalValidation(newValue, mode, context, event);
      }
      catch(e)
      {
        // validation failed
      }
    }
    else
    {
      if (oj.Logger.level > oj.Logger.LEVEL_WARN)
      {
        oj.Logger.info("Validation skipped and value option not set as component state does not " + 
                " allow setting value. For example if the component is readonly or disabled.");
      }
    }
    
    return result;
  },  
  
  //** @inheritdoc */
  _CompareOptionValues : function (option, value1, value2)
  {
    if (option === 'value' || option === 'rawValue')
    {
      return oj.Object.compareValues(value1, value2);
    }
    else if (option.indexOf('messages') === 0)
    {
      return this._messagesEquals(value1, value2);
    }
    
    return this._superApply(arguments);
  },
  
  /**
   * Returns the default styleclass for the component. All input components must override.
   * 
   * @return {string}
   * 
   * @memberof oj.editableValue
   * @instance
   * @protected
   * @abstract
   */
  _GetDefaultStyleClass : function ()
  {
    oj.Assert.failedInAbstractFunction();
    return "";
  },
    
  
  // I N T E R N A L   P R I V A T E   C O N S T A N T S    A N D   M E T H O D S 
  // Subclasses should not override or call these methods
  
  /**
   * Types of messaging content to update.
   * <ul>
   * <li>'ALL' - builds all messaging content</li>
   * <li>'VALIDITY_STATE' - updates only validityState every time validation runs and there are 
   * new messages or when the messages option changes.</li>
   * <li>'CONVERTER_HINT' - updates only converter hints, this is used when converter option 
   * changes</li>
   * <li>'VALIDATOR_HINTS' - updates only validator hints, this is used when validators option 
   * changes</li>
   * <li>'TITLE' - updates only title, when the title property changes</li>
   * </ul>
   * @private
   * @memberof oj.editableValue
   */
  _MESSAGING_CONTENT_UPDATE_TYPE : {ALL : 1, 
                                    VALIDITY_STATE : 2, 
                                    CONVERTER_HINT : 3, 
                                    VALIDATOR_HINTS : 4, 
                                    TITLE : 5},
                                   
  /**
   * when below listed options are passed to the component, corresponding CSS will be toggled
   * @private
   * @const
   * @type {Object}
   * @memberof oj.editableValue
   */
  _OPTION_TO_CSS_MAPPING: {
    "disabled": "oj-disabled",
    "required": "oj-required"
  },
  
    
  /**
   * Clears all messages for this component. Today this only happens when <br/>
   *  - the component's value changes as a result of user interaction, <br/>
   *  - the reset method is called, <br/>
   *  - component is destroyed.<br/>
   * 
   * @param {Event=} event the original event (for user initiated actions that trigger a DOM event,
   * like blur) or undefined.
   * @param {boolean=} doNotSetOption default value is false; a true value clears the option 
   * directly without using the public option method, causing no events to be fired. 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _clearAllMessages : function(event, doNotSetOption)
  {
    doNotSetOption = doNotSetOption ? true : false;
    if (!doNotSetOption)
    {
      this._clearMessages('messagesHidden', event);
      this._clearMessages('messagesShown', event);
      this._clearMessages('messagesCustom', event);
    }
    else
    {
      this.options['messagesHidden'] = [];
      this.options['messagesShown'] = [];
      this.options['messagesCustom'] = [];
    }
  },
  
  /**
   * Clears all messages that were added by component. These includes all messages in messagesHidden 
   * option and all messages except custom in messagesShown. 
   * 
   * Called when these options change - validators, disabled, readonly, required, and methods - 
   * refresh are called. 
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */  
  _clearComponentMessages : function ()
  { 
    var beforeLen;
    var msg;
    var shownMsgs = this.options['messagesShown'];

    beforeLen = shownMsgs.length;
    this._clearMessages('messagesHidden'); 
    
    // remove component messages in messagesShown. Custom messges are kept intact.
    for (var i = beforeLen -1; i >= 0; i--)
    {
      // NOTE: shownMsgs is this.options['messagesShown'] 
      // so we are modifying this.options['messagesShown'] here.
      msg = shownMsgs[i];
      if (msg instanceof oj.ComponentMessage)
      {
        shownMsgs.splice(i, 1);
      }
    }
    
    if (shownMsgs.length !== beforeLen)
    {
      // Setting 'changed' flag to true means that although we have already 
      // updated this.options['messagesShown'], we still want to fire a messagesShownChanged event.
      this._setMessagesOption('messagesShown', shownMsgs, null, true);
    }    
  },
  
    
  /**
   * Sets the messages option with the new value.
   * Setting 'changed' flag to true means we want to fire a property changed event
   * without checking that the option value to what you are setting it to. Useful
   * if the option is an array or we have already updated the option directly.
   * 
   * This method updates the option directly without invoking setOption() method. This is done by 
   * setting the following property in flags parameter of the option() method - 
   * <code class="prettyprint">{'_context': {internalSet: true}}</code>
   * 
   * @param {string} key
   * @param {Array} value
   * @param {Event=} event the event like the user blurred to trigger a messages option change
   * or undefined.
   * @param {Boolean=} changed when this is true, then we set the 'changed' flag to true, and having
   * the changed flag be true will guarantee that the property changed event is fired even if the
   * the property value is equal to what you are setting it to.
   * @private
   * @memberof oj.editableValue
   */
  _setMessagesOption : function (key, value, event, changed)
  {
    var flags = {};
    
    // Optimize for the common 'clear' operation
    var bothEmpty = value.length === 0 && this.options[key].length === 0;
    
    if (changed || !bothEmpty)
    {
      // 'messagesCustom' is not read-only, but 'messagesShown' and 'messagesHidden' are
      flags['_context'] = {originalEvent: event, writeback: true, internalSet: true};
      if (key !== "messagesCustom")
        flags['_context']['readOnly'] = true;

      flags['changed'] = changed || !bothEmpty;
      
      this._resetValid();
      
      this.option(key, value, flags); 
      
      this._updateMessagingContent();
    }
  },  
  
  /**
   * Sets the valid option with the new value.
   * 
   * This method updates the option directly without invoking setOption() method. This is done by 
   * setting the following property in flags parameter of the option() method - 
   * <code class="prettyprint">{'_context': {internalSet: true}}</code>
   * 
   * @param {Event=} event the original event (for user initiated actions that trigger a DOM event,
   * like blur) or undefined. The custom element bridge creates a CustomEvent out of this when 
   * it sends the property changed event.
   * @private
   */
  _setValidOption : function (event)
  {
    var flags = {};

    // 'valid' is read-only
    flags['_context'] = {originalEvent: event, writeback: true, internalSet: true, readOnly: true};

    this.option("valid", this._determineValid(), flags);

  },  
  
  /**
   * Clears the messages and message options - <code class="prettyprint">messagesHidden</code>,
   * <code class="prettyprint">messagesShown</code>, <code class="prettyprint">messagesCustom</code>.
   * 
   * @param {String} option messages option that is being cleared.
   * @param {Event=} event the original event (for user initiated actions that trigger a DOM event,
   * like blur) or undefined.
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _clearMessages : function (option, event)
  {
    this._setMessagesOption(option, [], event);
  },
  
  /**
   * Clones messages before it's set. <br/>
   * 
   * @param {Array=} value
   * @returns {Array} of cloned messages
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */    
  _cloneMessagesBeforeSet : function (value)
  {
    var i;
    var msg;
    var msgsClone = [];
    var val;
    
    // we want all messages to be an instance of oj.Message. So clone array
    if (value && value.length > 0)
    {
      for (i = 0; i < value.length; i++)
      {
        val = value[i];
        msg = new oj.Message(val['summary'], val['detail'], val['severity']);
        msg =  Object.freeze ? Object.freeze(msg) : msg;
        msgsClone.push(msg);
      }
    }
    
    return msgsClone;
  },  
  
 
  /** 
   * Create the ojLabel component with help (required is done in the components that support
   * required) see oj.EditableValueUtils._refreshRequired.
   * This is not supported for custom elements. For pages with custom elements, the app dev
   * uses the public &lt;oj-label> component.
   * @private
   * @memberof oj.editableValue
   * @instance
   * 
   */
  _createOjLabel : function ()
  {
    var helpDef;
    var helpSource;
    
    if (this._IsCustomElement())
      return;
    
    this.$label = this._GetLabelElement();
    if (this.$label)
    {
      helpDef = this.options['help']['definition'];
      helpSource = this.options['help']['source'];

      // create the ojLabel component 
      this.$label.ojLabel(
        {rootAttributes:{'class': this._GetDefaultStyleClass()+"-label"},
        help:{'definition': helpDef, 
              'source': helpSource}});
      this._createDescribedByForLabel();
    }    
  },
 /**
   * Refreshes the aria-describedby for label element's helpIcon
   * @protected
   * @memberof oj.editableValue
   * @instance
   */
  _createDescribedByForLabel : function ()
  {
    var helpDef = this.options['help']['definition'];
    var helpSource = this.options['help']['source'];
    var labelId;
    
    if ((helpSource != null) || (helpDef != null))
    {
      // get label's helpIconSpan get the id and add it here.
      labelId = this.$label.attr("id");
      if (labelId) 
      {
        this._addAriaDescribedBy(labelId + _HELP_ICON_ID);
      }        
    }
  },
    /**
   * Refreshes the aria-describedby for label element's helpIcon
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _refreshDescribedByForLabel : function ()
  {
    var helpDef = this.options['help']['definition'];
    var helpSource = this.options['help']['source'];
    // if aria-labelledby is set, 
    // add/remove aria-describedby to the inputs pointing to
    // the label+"_helpIcon".
    var labelId = this.$label.attr("id");
    if (labelId) 
    {
      if ((helpSource != null) || (helpDef != null))
        this._addAriaDescribedBy(labelId + _HELP_ICON_ID);
      else
        this._removeAriaDescribedBy(labelId + _HELP_ICON_ID);
    }  
  },
  /**
   * Refreshes the component to respond to DOM changes, in which case fullRefresh=true. 
   * @param {boolean} fullRefresh true if a full refresh of the component is desired.
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _doRefresh : function (fullRefresh)
  {
    var runFullValidation = false;
    
    fullRefresh = fullRefresh || false;
    if (fullRefresh)
    {
      // reset state and re-initialize component messaging, since refresh() can be called when 
      // locale changes, requiring component to show messaging artifacts for current locale. 
      this._ResetComponentState();
      this._initComponentMessaging();
      
      if (this._hasInvalidMessagesShowing())
      {
        runFullValidation = true;
      }

      // clear component messages always
      this._clearComponentMessages();
      if (runFullValidation)
      {
        // run full validation using the current display value; 
        // show messages immediately if there are errors otherwise set value to option
        this._updateValue(oj.EditableValueUtils.refreshMethodOptions);
      }
      else
      {
        // run deferred validation if comp is either showing a deferred error or has no errors. 
        // But only when required is true. 
        if (this._IsRequired())
        {
          this._runDeferredValidation(oj.EditableValueUtils.refreshMethodOptions.validationContext);
        }
        // refresh UI display value when there are no errors or where there are only deferred errors 
        this._Refresh("value", this.options['value'], true);
      }
    }
    else
    {
      // we call this._doRefresh(false); in component create
      // we call this._doRefresh(true); from refresh.
      this._Refresh("value", this.options['value']);
    }
    // always refresh these
    this._Refresh("disabled", this.options['disabled']);
  },
  
  /**
   * Gets the last stored model value
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getLastModelValue : function ()
  {
    return this._oj_lastModelValue;
  },
  
  /**
   * Gets the last display value
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getLastDisplayValue : function () 
  {
    // initially, _oj_lastElementValue is undefined. But the browser returns "" for the 
    // displayValue. 
    if (this._oj_lastElementValue === undefined)
      this._oj_lastElementValue = "";
    return this._oj_lastElementValue;
  },
  /**
   * Get the element whose id matches the elem's aria-labelledby value, if any.
   * @param {Object} elem the dom element from which you want to get the 
   * aria-labelledby property value
   * @return {jQuery|null} if element does not have aria-labelledby defined, then
   *    returns null. If it  does, then return a new jQuery object with the 
   *    label with an id equal to the aria-labelledby value. If no match, then
   *    the jQuery object will be empty.
   *    
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getAriaLabelledByElement: function (elem)
  {
    // look for a label with an id equal to the value of aria-labelledby. 
    // .prop does not work for aria-labelledby. Need to use .attr to find
    // aria-labelledby.
    var ariaId = elem.attr("aria-labelledby");
    var labelQuery;

    if (ariaId !== undefined )
    {
      labelQuery = "label[id='" + ariaId + "']"; 
      return $(labelQuery);
    }    
    else
      return null;
  },
  /**
   * Get the id of the element whose aria-labelledby is set
   * @param {Object} elem the dom element from which you want to get the 
   * aria-labelledby property value
   * @return {string} id or null
   *    
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getAriaLabelledById: function (elem)
  {
    var id = null;
    
    var ariaLabelledByElem = this._getAriaLabelledByElement(elem);
    if (ariaLabelledByElem !== null && ariaLabelledByElem.length !== 0)
    {
      id = ariaLabelledByElem.attr("id");
    }
    return id;
  },
  /**  
   * Add the aria-describedby on the content element(s) if it isn't already there.
   * 
   * @param {string} id the id for aria-describedby
   * @private
   * @memberof oj.editableValue
   * @instance
   * 
   */
  _addAriaDescribedBy : function (id)
  {
    var contentElements = this._GetContentElement();
    var index;
    
    contentElements.each(function() {
      var describedby = $(this).attr("aria-describedby");
      var tokens;

      tokens = describedby ? describedby.split(/\s+/) : [];
      // Get index that id is in the tokens, if at all.
      index = $.inArray(id, tokens);
      // add id if it isn't already there
      if (index === -1)
        tokens.push(id);
      describedby = $.trim(tokens.join(" "));
      $(this).attr("aria-describedby", describedby);
    });
      
  },
  /**  
   * Remove the aria-describedby from the content element(s).
   * 
   * @param {string} id the id for aria-describedby
   * @private
   * @memberof oj.editableValue
   * @instance
   * 
   */
  _removeAriaDescribedBy : function (id)
  {

    var contentElements = this._GetContentElement();        

    contentElements.each(function() {

      var describedby;
      var index;
      var tokens;
      
      // get aria-describedby that is on the content element(s)
      describedby = $(this).attr("aria-describedby");
      // split into tokens
      tokens = describedby ? describedby.split(/\s+/) : [];
      // Get index that id is in the tokens, if at all.
      index = $.inArray(id, tokens);
      // remove that from the tokens array
      if (index !== -1)
        tokens.splice(index, 1);
      // join the tokens back together and trim whitespace
      describedby = $.trim(tokens.join(" "));
      if (describedby)
        $(this).attr("aria-describedby", describedby);
      else
        $(this).removeAttr("aria-describedby");
     });

  },

  /**
   * Returns a concat of messagesShown and messagesHidden.
   * 
   * @returns {Array}
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getMessages : function()
  {
    return this.options['messagesShown'].concat(this.options['messagesHidden']); // todo: revisit
  }, 
  /**
   * Finds the oj-label element associated with this form component and returns it.
   * @returns {Element|null}
   * @private
   * @memberof oj.editableValue
   * @instance
   */  
  _getCustomOjLabelElement: function ()
  {
    var labelElement = null;
    var $labelElement;
    var labelledBy = this.options['labelledBy'];
    var widget;
    var widgetId;
    
    if (labelledBy)
    {
      // First, if 'labelledBy', look for the oj-label by its 'id' since that will be fastest.
      labelElement = document.getElementById(labelledBy);
    }
      
    // if there is no label element found, then look at the 'id'/'for' combination.
    if (labelElement === null)
    {
      // If 'id', look for the label by its 'for' attribute. (First look for sibling 
      // since that is the most common way and attribute selector searches perform better than
      // looking at the entire document.)  
      // widget will be the JET form element, like oj-input-text, in jquery format.
      widget = this.widget();
      widgetId = widget[0].id;
      if (widgetId)
      {
        // $labelElement will be the <oj-label>.
        $labelElement = widget.siblings("[for='" + widgetId + "']");
        if ($labelElement.length === 0)
        {
          // a sibling label is not found, so look for it from the document level.
          $labelElement = $("[for='" + this.element.id + "']");

        }
        if ($labelElement.length > 0)
        {
          // if it is still 0, we couldn't find the label. If it has a length, we return
          // the first one.
          labelElement = $labelElement[0];
        }
      }        
    }
    return labelElement;
  },
  /**
   * Helper method to retrieve the label text. Needed for required translation. 
   * Returns the form component's label or oj-label's textContent, if the label was found.
   * 
   * @return {string|null}
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getLabelText : function ()
  {

    if (this.$label)
    {
      return this.$label.text();
    }
    else
    {
      if (this.customOjLabelElement === undefined)
        this.customOjLabelElement = this._getCustomOjLabelElement();
      return this.customOjLabelElement ? this.customOjLabelElement.textContent : null;
    }
  },

  /**        
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getValidityState : function ()
  {
    if (this._validityState)
    {
      return this._validityState;
    }
    else
    {
      
      this._validityState = new oj.ComponentValidity(this.isValid(), this._getMessages());
    }
    return this._validityState;
  },

  
  /**        
   * Whether component has invalid messages.
   * 
   * @return {boolean} true if invalid; false otherwise
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _hasInvalidMessages : function ()
  {
    return !oj.Message.isValid(this._getMessages());
  },
  
  /**        
   * Whether there are invalid messages, that are currently showing.
   * 
   * @return {boolean} 
   * @private
   * @memberof oj.editableValue
   * @instance
   */  
  _hasInvalidMessagesShowing : function ()
  {
    return (!this.isValid() && this.options['messagesShown'].length > 0);
  },
  
  /**        
   * Whether component has invalid messages added by component, that are currently showing.
   * 
   * @return {boolean} 
   * @private
   * @memberof oj.editableValue
   * @instance
   */  
  _hasInvalidComponentMessagesShowing : function ()
  {
    var compMsgs;
    var msg;
    var shown = this.options['messagesShown'];
    
    for (var i = 0; i < shown.length; i++)
    {
      msg = shown[i];
      if (msg instanceof oj.ComponentMessage && msg._isMessageAddedByComponent())
      {
        compMsgs = compMsgs || [];
        compMsgs.push(msg);
      }
    }
    
    return (compMsgs === undefined) ? false : !oj.Message.isValid(compMsgs);
  },  
  
  /**
   * Initializes component messaging both when component is initialized or when displayOptions is 
   * set/changed. 
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _initComponentMessaging : function()
  {
    var compMessaging = this._getComponentMessaging();
    var messagingLauncher = this._GetMessagingLauncherElement();
    var compContentElement = this._GetContentElement();
    var messagingContent = this._getMessagingContent(this._MESSAGING_CONTENT_UPDATE_TYPE.ALL);
    
    // if default placeholder is currently set then it needs to be cleared here. This is needed for 
    // the following reasons
    // i. a component is reinitialized when the locale changed, requiring the converter hint for 
    // new locale to be set as placeholder.
    // ii. or a component's placeholder option or displayOptions option, could have changed 
    // requiring the placeholder to be reset if it's currently set to the default.
    // 
    if (!this._customPlaceholderSet)
    {
      this._ClearPlaceholder();
    }
    
    compMessaging.activate(messagingLauncher, compContentElement, messagingContent);
  },
  
  
  /**
   * Called after messagesCustom option changed. This method pushes custom messages to the 
   * messagesShown option.
   * 
   * @param {Object} flags
   * @returns {undefined}
   * @private
   */  
  _messagesCustomOptionChanged : function (flags)
  {
    var context = flags ? flags['_context'] : null;
    var customMsgs = this.options['messagesCustom'];
    var i;
    var msg;
    var previousShown = this.options['messagesShown'];
    var shownMsgs = [];
    
    // remove old custom messages from messagesShown array
    for (i = 0; i < previousShown.length; i++)
    {
      msg = previousShown[i];
      if (msg instanceof oj.ComponentMessage && msg._isMessageAddedByComponent())
      {
        shownMsgs.push(msg);
      }
    }
    
    // add new customMsgs to messagesShown
    for (i = 0; i < customMsgs.length; i++)
    {
      shownMsgs.push(customMsgs[i]);
    }
    
    // set 'messagesShown' option as an internal set
    this._setMessagesOption('messagesShown', shownMsgs, 
                            context ? context.originalEvent : null,
                            flags && flags['changed']);
  },
  /**        
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _placeholderOptionChanged : function (flags)
  {
    var context = flags && flags['_context'] || {};
    var refreshMessagingOptions = 
        //  internalMessagingSet indicates whether the current change is from the messaging module.
        // see ComponentMessaging for details
        context.internalMessagingSet ? false : true;
    var value = this.options['placeholder'];

    this._SetPlaceholder(value);
    if (refreshMessagingOptions)
    {
      // if placeholder was set and it's not from messaging code, then the messaging display options 
      // may need to re-evaluated. E.g., the default display for 
      // converterHint: ['placeholder', 'notewindow'] is 'placeholder', but if user were to set a 
      // custom placeholder, this changes the default display for convererHint from 'placeholder'
      // to 'notewindow'. 
      this._customPlaceholderSet = true;
      if (this._GetConverter())
      {
        this._initComponentMessaging();
      }
    }
    else
    {
      this._customPlaceholderSet = false;
    }
  },
  /**        
   * @private
   * @memberof oj.editableValue
   * @instance
   */  
  _setLastModelValue : function (value)
  {
    this._oj_lastModelValue = value;
  },
  /**        
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _setLastSubmittedValue : function (value) 
  {
    this._oj_lastElementValue = value;
  },
  /**        
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _titleOptionChanged : function ()
  {
    // when title changes push new title to messaging
    this._getComponentMessaging().update(
            this._getMessagingContent(this._MESSAGING_CONTENT_UPDATE_TYPE.TITLE));
  },
  
  /**
   * Adds messages to the option specified - <code class="prettyprint">messagesShown</code> or 
   * <code class="prettyprint">messagesHidden</code>. 
   * 
   * @param {String} option name of the option
   * @param {Object|Array} newMsgs an Array of one or more oj.Message object. 
   * @param {Event=} event - the event like the user blurred to trigger a messages option change
   * or undefined.
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _updateMessagesOption : function(option, newMsgs, event) 
  {
    var i;
    var len;
    var msgs;

    if (typeof newMsgs === "object" && Array.isArray(newMsgs)) 
    {
      // update this.options[option] directly by pushing any new messages into it.
      msgs = this.options[option];

      len = newMsgs.length;
      for (i = 0; i < len; i++)
      {
        msgs.push(newMsgs[i]);
      }
    }
    // Setting 'changed' flag to true means that although we have already 
    // updated this.options[option], we still want to fire a property changed event.
    this._setMessagesOption(option, msgs, event, true);
  },
  
  
  /**
   * Called after the messages option has changed to update internal valid property and messaging 
   * display.
   * 
   * @private
   * @memberof oj.editableValue
   */
  _updateMessagingContent : function() 
  {
    // update component messaging
    this._getComponentMessaging().update(this._getMessagingContent());
  },  
  
  /**
   * Writes the value into the option by calling the option method.  
   * 
   * @param {Object|string} newValue the new value to be written to option
   * @param {Object=} event the original event that triggered this
   * @param {number=} validationContext the context in which validation was run that resulted in 
   * value being updated.
   * @param {Object=} updateContext the context for updating the value option.
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   * @see #_setOption
   * @see #_AfterSetOptionValue
   */
  _updateValueOption : function (newValue, event, validationContext, updateContext)
  {
    var context = updateContext || {};
    
    // set dom event
    if (event)
    {  
      context.originalEvent = event;
    }
    
    // set writeback flag that determines whether value is written back. 
    switch (validationContext)
    {
      // value is written back outside of normal UI interaction in the following cases.
      case this._VALIDATION_CONTEXT.CONVERTER_OPTION_CHANGE:
      case this._VALIDATION_CONTEXT.DISABLED_OPTION_CHANGE:
      case this._VALIDATION_CONTEXT.READONLY_OPTION_CHANGE:
      case this._VALIDATION_CONTEXT.REFRESH_METHOD:
      case this._VALIDATION_CONTEXT.REQUIRED_OPTION_CHANGE:
      case this._VALIDATION_CONTEXT.VALIDATE_METHOD:
      case this._VALIDATION_CONTEXT.VALIDATORS_OPTION_CHANGE:
        
        context.writeback = true;
        
        // when the above options change or methods are called, and full validation is run the 
        // current display value is parsed, and set on the value option if all validations pass. 
        // Typically when the value option changes - either programmatically or user changes it, we 
        // end up clearing all 3 messages options.
        // But when value changes indirectly as a result of the above cases, then we do not clear 
        // custom messages. So a special flag is set so _AfterSetOptionValue can do the right 
        // thing. Component messages are already cleared when this method is called.
        context.doNotClearMessages = true;
        break;
    }

    context.internalSet = true;
    this.option({'value': newValue}, {'_context': context});
    // When internalSet is true _setOption->_AfterSetOptionValue->_Refresh isn't called. 
    // We still need the converter to run and the displayValue to be refreshed, so we
    // call this._AfterSetOptionValue ourselves
    this._AfterSetOptionValue('value', {'_context': context});
    

  },
  
  /**
   * Resets the internal property so that the next call to this.isValid() re-evaluates the correct 
   * value.
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _resetValid : function ()
  {
    this._valid = undefined;
  },  

  /**
   * Determines the validity of component based on current value of the messages* options.
   * This is what the 'valid' property should be set to.
   * 
   * Now when we start the lifecycle we clearAllMessages, then we continue with validation, etc.
   * But we don't want to update the valid property on clearAllMessages and again when validation
   * has failed. We want to update it only once.
   * @return {string} "valid", "invalidShown", "invalidHidden"
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _determineValid : function ()
  {
    var msgsHidden = this.options['messagesHidden'];
    var msgsShown = this.options['messagesShown'];

    var valid = "valid";
    
    // When new messages are written update the valid property
    if ( msgsShown && msgsShown.length !== 0 && !oj.Message.isValid(msgsShown))
    {
      valid = "invalidShown";
    }
    else if(msgsHidden && msgsHidden.length !== 0 && !oj.Message.isValid(msgsHidden))
    {
      valid = "invalidHidden";
    }
    
    return valid;
  },
  
  /**
   * Formats the value for display, based on the converter options. If no converter is set then 
   * returns the value as is.
   * 
   * @param {string} value value to be formatted
   * 
   * @return {string} formatted value
   * @throws {Error} when an error occurs during formatting
   * @protected
   * @memberof oj.editableValue
   * @instance
   */
  _formatValue : function (value)
  {
    var formattedValue = value;
    var converter = this._GetConverter();
    
    // don't clear messages here because we clear messages only when direct user interaction with 
    // component changes value. All other usecases we expect page authors to clear messages.

    if (converter)
    {
      // TODO: We should support chaining converters but for now we use just the first converter 
      // to parse.

      // Check if we have a converter instance
      if (typeof converter === "object") 
      {
        if (converter['format'] && typeof converter['format'] === "function")
        {
          formattedValue = converter['format'](value);
        }
        else
        {
          if (oj.Logger.level > oj.Logger.LEVEL_WARN)
          {
            oj.Logger.info("converter does not support the format method.");
          }
        }
      }
    }
    
    return formattedValue;
  },
  
  /**
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getComponentMessaging : function ()
  {
    if (!this._componentMessaging)
    {
      this._componentMessaging = new oj.ComponentMessaging(this);
    }
    
    return this._componentMessaging;
  },
  

  /**
   * Returns an array of validator hints.
   * @param {Array} allValidators
   * @private
   * @memberof oj.editableValue
   * @instance
   */        
  _getHintsForAllValidators : function(allValidators)
  {
    var i;
    var validator;
    var validatorHints = [];
    var vHint = "";

    if (this._IsRequired())
    {
      // get the hint for the default required validator and push into array if it's not already 
      // present in the validators array
      validator = this._getImplicitRequiredValidator();
      if (validator['getHint'] && typeof validator['getHint'] === "function")
      {
        vHint = validator['getHint']();
        if (vHint)
        {
          validatorHints.push(vHint);
        }
      }
    }

    // loop through all remaining validators to gather hints
    for (i = 0; i < allValidators.length; i++)
    {
      validator = allValidators[i], vHint = "";
      if (typeof validator === "object") 
      {
        if (validator['getHint'] && typeof validator['getHint'] === "function")
        {
          vHint = validator['getHint']();
          if (vHint)
          {
            validatorHints.push(vHint);
          }
        }
      }
    }

    return validatorHints;
  },
  
  /**
   * Returns the required validator instance or creates it if needed and caches it.
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getImplicitRequiredValidator : function ()
  {
    var vf;
    var reqTrans = {};
    var reqValOptions;
    
    if (this._implicitReqValidator == null) // falsey check
    {
      reqTrans = 
            this.options['translations'] ? this.options['translations']['required'] || {} : {} ;

      // TODO: cache required validator; purged when its options change, i.e., translations or label 
      // DOM changes
      reqValOptions = {
        'hint': reqTrans['hint'] || null,
        'label': this._getLabelText(), 
        'messageSummary': reqTrans['messageSummary'] || null,
        'messageDetail': reqTrans['messageDetail'] || null
      };

      vf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED);
      this._implicitReqValidator = vf ? vf.createValidator(reqValOptions) : null;
    }
    
    return this._implicitReqValidator;
  },

  
  /**
   * Returns content that will be used by messaging strategies.
   * 
   * @param {number} updateType of messaging content to update. Accepted values are defined by 
   * this._MESSAGING_CONTENT_UPDATE_TYPE.
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _getMessagingContent : function (updateType)
  {
    var messagingContent = {};

    var allValidators;
    var converter;
    var converterHint = "";
    var messages;
    var validatorHints = [];
    
    updateType = updateType || this._MESSAGING_CONTENT_UPDATE_TYPE.VALIDITY_STATE;

    // Add validityState which includes messages, valid and severity
    if (updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.ALL || 
        updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.VALIDITY_STATE)
    {
      messages = this._getMessages();
      
      // update validityState before packaging it
      this._getValidityState().update(this.isValid(), messages);
      messagingContent.validityState = this._getValidityState();
    }
    
    if (updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.ALL || 
        updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.CONVERTER_HINT)
    {
      converter = this._GetConverter();
      if (converter)
      {
        if (typeof converter === "object") 
        {
          if (converter['getHint'] && typeof converter['getHint'] === "function")
          {
            converterHint = converter['getHint']() || "";
          }
        }
      }
      messagingContent.converterHint = converterHint;
    }
    
    if (updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.ALL || 
        updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.VALIDATOR_HINTS)
    {
      allValidators = this._GetAllValidators();
      validatorHints = this._getHintsForAllValidators(allValidators) || [];
      messagingContent.validatorHint = validatorHints;
    }
    
    if (updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.ALL || 
        updateType === this._MESSAGING_CONTENT_UPDATE_TYPE.TITLE)
    {
      messagingContent.title = this.options['title'] || "";
    }
    
    return messagingContent;
  },
  
  /**
   * Compares the messages arrays for equality.
   * 
   * @param {Array} pm previous messages
   * @param {Array} m new messages
   * @returns {boolean} true if equal false otherwise
   * @private
   * @memberof oj.editableValue
   * @instance
   * 
   */
  _messagesEquals : function (pm, m)
  {
    var match = -1;
    var pmo;
    var passed = true;
    // $.extend merges the contents of two or more objects together into the first object
    var previousMsgs = $.extend([], pm);
    var msgs = $.extend([], m);
    
    
    if (previousMsgs.length !== msgs.length)
    {
      return false;
    }
    
    // one way it gets here is if there is one messages-custom message on initialization and
    // after the busyContext is complete meaning the page is rendered, we set a different
    // messages-custom message.
    $.each(previousMsgs, function (i, pMsg) 
    {
      if (!(pMsg instanceof oj.Message))
      {
        // freeze message instance once its created
        pmo = new oj.Message(pMsg['summary'], pMsg['detail'], pMsg['severity']);
        pmo = Object.freeze ? Object.freeze(pmo) : pmo;
      }
      else
      {
        pmo = pMsg;
      }
      
      match = -1;
      $.each(msgs, function(j, msg) {
        {
          if((oj.Message.getSeverityLevel(pmo['severity']) === 
             oj.Message.getSeverityLevel(msg['severity'])) && 
             pmo['summary'] === msg['summary'] && 
             pmo['detail'] === msg['detail'])
          {
            match = j;
            return; // found a match, so break out of loop
          }
        }
      });
      
      // remove entry at 'match' index from msgs
      if (match > -1)
      {
        msgs.splice(match, 1);
      }
      else
      {
        // we found no match so no need to loop
        passed = false;
        return;
      }
    });
    
    return passed;
  },
  
  /**
   * Parses the value using the converter set and returns the parsed value. If parsing fails the 
   * error is written into the element 
   * 
   * @param {string=} submittedValue to parse
   * @return {Object} parsed value 
   * @throws {Error} an Object with message
   * @protected
   * @memberof oj.editableValue
   * @instance
   */
  _parseValue: function(submittedValue) 
  {
    var converter = this._GetConverter();
    var parsedValue = submittedValue;
    
    if (converter)
    {
      // TODO: We should support chaining converters but for now we use just the first converter 
      // to parse.

      if (typeof converter === "object") 
      {
        if (converter['parse'] && typeof converter['parse'] === "function")
        {
          // we are dealing with a converter instance
          parsedValue = converter['parse'](submittedValue);
        }
        else
        {
          if (oj.Logger.level > oj.Logger.LEVEL_WARN)
          {
            oj.Logger.info("converter does not support the parse method.");
          }
        }
      }
    }
    
    return parsedValue;
  },
  /**
   * @private
   * @memberof oj.editableValue
   */
  _addValidationError : function(e, msgs)
  {
    var detail;
    var severity;
    var ojmessage;
    var summary;

    if (e instanceof oj.ConverterError || e instanceof oj.ValidatorError)
    {
      ojmessage = e.getMessage();

      severity = ojmessage['severity'] || oj.Message.SEVERITY_LEVEL['ERROR'];
      summary = ojmessage['summary'] || oj.Translations.getTranslatedString("oj-message.error");;
      detail = ojmessage['detail'] || oj.Translations.getTranslatedString("oj-converter.detail");
    }
    else if (e['summary'] || e['detail'])
    {
      severity = oj.Message.SEVERITY_LEVEL['ERROR'];
      summary = e['summary'] || oj.Translations.getTranslatedString("oj-message.error");;
      detail = e['detail'] || oj.Translations.getTranslatedString("oj-converter.detail");
    }
    else
    {
      // TODO: is this error message generic enough to use for both converter and validator errors?
      severity = oj.Message.SEVERITY_LEVEL['ERROR'];
      summary = oj.Translations.getTranslatedString("oj-message.error");
      detail = e['message'] || oj.Translations.getTranslatedString("oj-converter.detail");
    }
    
    msgs.push({summary: summary, detail: detail, severity: severity});
  },

  /**
   * Processes the error information for one or more errors and returns an Array of 
   * oj.ComponentMessage instances.
   * 
   * @param {Error} e instance of Error
   * @param {number=} context the context in which the validation error was thrown
   * @param {String=} display whether message is shown or hidden
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _processValidationErrors : function (e, context, display)
  {
    var componentMsgs = [];
    var msg;
    var msgs = e._messages || [];
    var options = {};
    
    options['context'] = context || 0;
    options['display'] = display || oj.ComponentMessage.DISPLAY.SHOWN;
    
    if (msgs.length === 0)
    {
      this._addValidationError(e, msgs);
    }
    
    for (var i = 0; i < msgs.length; i++)
    {
      msg = msgs[i];
      componentMsgs.push(this._createComponentMessage(msg.summary, msg.detail, msg.severity, options));
    }
    
    return componentMsgs || null;
  },
  /**
   * @private
   * @memberof oj.editableValue
   */
  _createComponentMessage : function (summary, detail, severity, options)
  {
    var cMsg;
    // new properties can't be added but existing properties can be changed
    cMsg = new oj.ComponentMessage(summary, detail, severity, options);
    cMsg = Object.seal ? Object.seal(cMsg) : cMsg;
    return cMsg;
  },
          
  /**
   * Refreshes the component display value, only when the current value is different from the last 
   * saved value, unless asked to always refresh the display value.
   * 
   * @param {Object|undefined} value the changed value that needs to be updated on UI
   * @param {boolean=} fullRefresh false is the default; true means always refresh component 
   * display value using the current option value. This overwrites any UI value, the user may have 
   * entered.
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */        
  _refreshComponentDisplayValue : function (value, fullRefresh)
  {
    if (fullRefresh || (value !== this._getLastModelValue()))
    {
      this._updateElementDisplayValue(value);
    }
  },


  /**
   * Toggles css selector on the widget. E.g., when required option changes, the oj-required 
   * selector needs to be toggled.
   * @param {string} option
   * @param {Object|string} value 
   * @private
   * @memberof oj.editableValue
   * @instance
   */        
  _refreshTheming : function (option, value)
  {
    if (Object.keys(this._OPTION_TO_CSS_MAPPING).indexOf(option) !== -1) 
    {
      this.widget().toggleClass(this._OPTION_TO_CSS_MAPPING[option], !!value);
    }
  },
 
  /**
   * Runs validators in deferred mode using the option value. Any validation error thrown is 
   * deferred, or hidden by component, until explicitly asked to show them (see showMessages()). 
   * Deferred error is pushed to <code class="prettyprint">messagesHidden</code> option.
   * 
   * @param {number} context in which validation was run. 
   * 
   * @see #showMessages
   * @private 
   * @memberof oj.editableValue
   * @instance
   */
  _runDeferredValidation : function (context) 
  {
    var newMsgs;
    var newValue;
    
    if (this._CanSetValue())
    {
      try
      {
        // TODO: Today required is the only validator that runs deferred. We need a generic way to 
        // retrieve deferred validators. 
        newValue = this._validateValue(this.options['value'], 
                      this._VALIDATION_MODE.REQUIRED_VALIDATOR_ONLY);
      }
      catch (ve)
      {
        newMsgs = this._processValidationErrors(ve, context, oj.ComponentMessage.DISPLAY.HIDDEN);
        if (newMsgs)
        {
          this._updateMessagesOption('messagesHidden', newMsgs);
        }    
      }
    }
    else
    {
      if (oj.Logger.level > oj.Logger.LEVEL_WARN)
      {
        oj.Logger.info("Deferred validation skipped as component is readonly or disabled.");
      }
    }
    // set valid option whether or not we can set a value in the field. It won't trigger
    // an option changed event if the valid value is the same as it was before.
    this._setValidOption(null);
  },
    
  /**
   * Runs validation based on the mode settings. 
   * 
   * @param {Object} value to parse and/or validate
   * @param {number=} mode determines how validation is run. see _VALIDATION_MODE
   * @param {number=} context determines when validation is being run.
   * @param {Event=} event the original event (for user initiated actions that trigger a DOM event,
   * like blur) or undefined.
   * 
   * @return {Object|string} parsed value
   * @throws {Error} when validation fails.
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _runNormalValidation : function (value, mode, context, event) 
  {
    var newMsgs;
    var newValue = value;

    // callers of this function should clear messages 
    try
    {
      // Step 1: only when "full" validation is requested converters get run
      if (mode === this._VALIDATION_MODE.FULL)
      {
        // Step1: Parse value using converter
        newValue = this._parseValue(value);
      }

      // Step 2: Run validators
      this._validateValue(newValue, mode === this._VALIDATION_MODE.REQUIRED_VALIDATOR_ONLY);
      
    }
    catch (ve)
    {
      newMsgs = this._processValidationErrors(ve, context);
      this._updateMessagesOption ('messagesShown', newMsgs, event);
      // update valid option before we throw the error
      throw ve;
    }
    
    return newValue;
  },
  
  /**
   * Runs either deferred or normal validation based on the state component is in. This method is 
   * called when certain options change - required, disabled etc.
   * 
   * <p>
   * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
   * run full validation with UI value (we don't know if the UI error is from a required validator 
   * or something else);<br/>
   * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
   * updated<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to valueChanged to clear custom errors.<br/>
   * 
   * - if component is invalid and has messagesHidden -> required: false -> clear component 
   * errors; no deferred validation is run.<br/>
   * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
   * errors unnecessarily)<br/>
   * - messagesCustom is never cleared<br/>
   * </p>
   * 
   * @param {Object} validationOptions
   * 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _runMixedValidationAfterSetOption : function (validationOptions)
  {
    var runDeferredValidation = true;
    
    if (this._hasInvalidMessagesShowing())
    {
      runDeferredValidation = false;
    }

    // clear component messages always
    this._clearComponentMessages();
    if (!runDeferredValidation)
    {
      // run full validation using the current display value; 
      // show messages immediately if there are errors otherwise set value to option
      this._updateValue(validationOptions);
    }
    else
    {
      // run deferred validation if comp is either showing a deferred error or has no errors. 
      // But only when required is true. 
      // we update the valid option within _runDeferredValidation
      if (this._IsRequired())
      {
        this._runDeferredValidation(validationOptions.validationContext);
      }
      else
      {
        this._setValidOption(null);
      }
    }
  },
    
  /**
   * @private
   * @memberof oj.editableValue
   */
  _updateElementDisplayValue : function (modelValue, event)
  {
    var actualDisplayValue;
    var displayValue;
    var newMsgs;
    
    // cache the new model value on the element 
    this._setLastModelValue(modelValue);

    // Update element with the displayValue
    displayValue = modelValue;

    try
    {
      displayValue = this._formatValue(modelValue);
    }
    catch (e)
    {
      newMsgs = this._processValidationErrors(e);
      this._updateMessagesOption ('messagesShown', newMsgs, event);
    }
    
    this._SetDisplayValue(displayValue); 
    // getting the display value right after we set it is probably not necessary,  but just in
    // case a subclass did something to it, we do.
    actualDisplayValue = this._GetDisplayValue();
    this._setLastSubmittedValue(actualDisplayValue);
     // update rawValue option to keep it in sync with the display value
    this._SetRawValue(actualDisplayValue, null);
  },
  
    
  /**
   * Runs normal validation when certain options change - converter, required, validators etc. - and 
   * updates (re-sets) the value if no new errors are found. This is different from _SetValue()
   * which is called when end-user changes the value as a result of interacting with the component. 
   * Additionally _SetValue() clears all messages always, whereas callers of this method determine 
   * which messages they want cleared.
   * 
   * @param {Object} options - name value properties used when validation is run. 
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _updateValue : function (options)
  {
    var newValue;
    
    // run full validation using the current display value; 
    // show messages immediately if there are errors otherwise set value to option
    newValue = this._Validate(this._GetDisplayValue(), null, options);
           
    this._setValidOption(null);

    // if validation was a success and component is valid, or it's invalid with no new component 
    // messages then update value option. 
    // NOTE: often custom msgs are not cleared when this method is called.
    if (newValue !== undefined && 
        (this.isValid() || !this._hasInvalidComponentMessagesShowing()))
    {
      this._updateValueOption(newValue, null, options.validationContext);
    }
  }, 
  
  /**
   * Validates the value by running through the list of regsitered validators. The algorithm is as 
   * follows -
   * 1. check to see if we are currently valid, if not return
   * 2. run required check. 
   * 3. if value is not empty get all the validators and validate in sequence. 
   * 4. if all validators pass return.
   * 
   * Callers can rely on the 'valid' options property to determine the validity state of the 
   * component after calling this method
   * 
   * @param {Object|string} value to be validated
   * @param {boolean} requiredOnly true only runs required validation
   * @throws {Error} when validation fails.
   * @private
   * @memberof oj.editableValue
   * @instance
   */
  _validateValue : function (value, requiredOnly)
  {
    var allValidators = this._GetAllValidators();
    var i;
    var validator;
    var valMsgs = [];

    // run required validation first; 
    if (this._IsRequired())
    {
      validator = this._getImplicitRequiredValidator();
      try
      {
        // check if trimmed value is empty. See AdfUIEditableValue.prototype.ValidateValue
        validator.validate(oj.StringUtils.trim(value));
      }
      catch (e)
      {
        // save all validation errors
        this._addValidationError(e, valMsgs);
      }
    }
      
    // run other validators when requested. 
    // latest review - we want to do exactly as ADF does and it validates empty field values by 
    // default. The condition always resolves to true in JET!
    if (!requiredOnly) // && (!isEmptyValue || this._shouldValidateEmptyFields()))
    {
      for (i = 0; i < allValidators.length; i++)
      {
        validator = allValidators[i];
        if (typeof validator === "object") 
        {
          if (validator['validate'] && typeof validator['validate'] === "function")
          {
            try
            {
              validator['validate'](value);
            }
            catch (e)
            {
              // save all validation errors
              this._addValidationError(e, valMsgs);
            }              
          }
          else
          {
            if (oj.Logger.level > oj.Logger.LEVEL_WARN)
            {
              oj.Logger.info("validator does not support the validate method.");
            }
          }
        }
      }
    }

    // throw error if there were validation failures
    if (valMsgs.length > 0)
    {
      var ve = new Error();
      ve._messages = valMsgs;
      throw ve;
    }
  } 
}, true);

oj.Components.setDefaultOptions(
  {

    'editableValue': // properties for all editableValue components 
    {
      'displayOptions': oj.Components.createDynamicPropertyGetter(function(context){
        var displayOptions = {
          'messages': context['containers'].indexOf('ojDataGrid') >= 0 || 
                      context['containers'].indexOf('ojTable') >= 0 ? ['notewindow'] : ['inline'],
          'converterHint': ['placeholder', 'notewindow'], 
          'validatorHint': ['notewindow']
        };
        displayOptions[context['isCustomElement'] ? 'helpInstruction' : 'title'] = ['notewindow'];
        return displayOptions;
      }),
      'help': oj.Components.createDynamicPropertyGetter(function(context) {
        // Conditionally set the defaults for custom element vs widget syntax since we expose different APIs
        if (context['isCustomElement']) {
          return {'instruction': ''};
        } else {
          return {'definition': null, 'source': null};
        }
      }),
    }
  }
 

);
//////////////////     SUB-IDS     //////////////////
/**
 * <p>Sub-ID for the help icon element used by EditableValue components.</p>
 * 
 * @ojsubid oj-label-help-icon
 * @memberof oj.editableValue
 * @ignore
 *
 * @example <caption>Get the help icon element associated with an editable value component:</caption>
 * var node = myComp.getNodeBySubId("oj-label-help-icon");
 */

//////////////// fragments /////////////////
/**
 * <p>
 * <h3 id="validation-section">
 * Validation and Messaging
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#validation-section"></a>
 * </h3>
 * An editable component runs validation (normal or deferred) based on the action performed on it 
 * (either by end-user or page author), and the state it was in when the action occurred. Examples 
 * of actions are - creating a component, user changing the value of the component by interacting 
 * with it, the app setting a value programmatically, the app calling the validate() method etc. At 
 * the time the action occurs, the component could already be showing errors, or can have a deferred 
 * error or have no errors. 
 * <p>
 * These factors also determine whether validation errors/messages get shown to the user immediately 
 * or get deferred. The following sections highlight the kinds of validation that are run and how 
 * messages get handled.
 * </p>
 * <h4 id="normal-validation-section">Normal Validation 
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#normal-validation-section"></a></h4>
 * Normal validation is run in the following cases on the display value, using the converter and 
 * validators set on the component, and validation errors are reported to user immediately.
 * <ul>
 * <li>When value changes as a result of user interaction all messages are cleared, including custom 
 * messages added by the app, and full validation is run on the UI value. The steps performed are 
 * outlined below.
 * <ol> 
 * <li>All messages are cleared and <code class="prettyprint">messagesCustom</code> property is cleared</li>
 * <li>If no converter is present then processing continues to next step. If a converter is 
 * present, the UI value is first converted (i.e., parsed). If there is a parse error then 
 * the messages are shown and processing returns.</li>
 * <li>If there are no validators setup for the component then the value is set on the component. 
 * Otherwise all validators are run in sequence using the parsed value from the previous step. The 
 * implicit required is run first if the component is marked required. When a validation error is 
 * encountered it is remembered and the next validator in the sequence is run. 
 * <ul><li>NOTE: The value is trimmed before required validation is run</li></ul>
 * </li>
 * <li>At the end of the validation run if there are errors, the messages are shown
 * and processing returns. If there are no errors, then the 
 * <code class="prettyprint">value</code> property is updated and the formatted value displayed on the 
 * UI.</li>
 * </ol>
 * </li>
 * <li>When the <code class="prettyprint">validate</code> method is called by app, all messages are 
 * cleared and full validation run using the display value. See <code class="prettyprint">validate</code>
 * method on the sub-classes for details. Note: JET validation is designed to catch user input errors, and not invalid
 * data passed from the server; this should be caught on the server.</li>
 * <li>When certain properties change through programmatic intervention by app, the component 
 * determines whether it needs to run normal validation based on the state the component is in. 
 * Refer to the <a href="#mixed-validation-section">Mixed Validation</a> section below for details. </li>
 * </ul>
 * 
 * <h4 id="deferred-validation-section">Deferred Validation
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#deferred-validation-section"></a>
 * </h4>
 * Deferred validation is run in the following cases on the component value using the implicit 
 * required validator if required is true, and validation errors are deferred, i.e., not shown to user immediately. 
 * Refer to the <a href="#deferred-messages-section">Showing Deferred Messages</a> section to 
 * understand how deferred messages can be shown.
 * <ul>
 *  <li>When a component is created and it is required deferred validation is run and no messages are cleared 
 *  prior to running validation.  
 *  Refer to the <a href="#deferred-validators-section">Validators 
 *  Participating in Deferred Validation</a> section for details.</li> 
 *  <li>When the <code class="prettyprint">value</code> property changes due to programmatic 
 *  intervention deferred validation is run, after all messages and messagesCustom property are cleared.</li>  
 *  <li>When the <code class="prettyprint">reset</code> method is called, deferred validation is run 
 *   after all messages and messagesCustom property are cleared.</li>  
 *  <li>When certain properties change through programmatic intervention by app, the component 
 *  determines whether it needs to run deferred validation based on the state the component is in. 
 *  Refer to the <a href="#mixed-validation-section">Mixed Validation</a> section below for details.</li>
 * </ul>
 * 
 * <h4 id="mixed-validation-section">Mixed Validation
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#mixed-validation-section"></a>
 * </h4>
 * Either deferred or normal validation is run in the following cases based on the state the 
 * component is in and any validation errors encountered are either hidden or shown to user.
 * <ul>
 *  <li>when disabled property changes. See <a href="#disabled">disabled</a> property for details.</li>
 *  <li>when refresh method is called. See <a href="#refresh">refresh</a> method for details.</li> 
 *  <li>when converter property changes. Not all EditableValue components have the converter property. See
 *  the sub-classes that have the converter property for details, e.g., {@link oj.inputBase#converter}.</li>
 *  <li>when required property changes. Not all EditableValue components have the required property. See
 *  the sub-classes that have the required property for details, e.g., {@link oj.inputBase#required}.</li>
 *  <li>when validators property changes. Not all EditableValue components have the validators property. See
 *  the sub-classes that have the validators property for details, e.g., {@link oj.inputBase#validators}.</li>
 *   
 * </ul>
 * </p>
 * 
 * <p>
 * <h3 id="deferred-messages-section">
 * Showing Deferred Messages
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#deferred-messages-section"></a>
 * </h3>
 * Deferred validation messages are displayed only when page author requests for it explicitly in 
 * one of the following ways: 
 * <ul>
 * <li>calls the <a href="#showMessages"><code class="prettyprint">showMessages</code></a> method on the component</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <h3 id="deferred-validators-section">
 * Validators Participating in Deferred Validation
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#deferred-validators-section"></a>
 * </h3>
 * The required validator is the only validator type that participates in deferred validation.
 * The required property needs to be set to true for the required validator to run.
 * </p> 
 * @ojfragment validationAndMessagingDoc - Used in the general section of classdesc
 * @memberof oj.editableValue
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/

/**
 * A messaging strategy that places the messaging content inline (underneath) the editableValue
 * component.
 *
 * @param {Array.<string>} displayOptions an array of messaging artifacts displayed inline. e.g,
 * 'messages' (for now only 'messages' are supported inline, not converterHints, and others)
 * @constructor
 * @extends {oj.MessagingStrategy}
 * @private
 */
oj.InlineMessagingStrategy = function (displayOptions)
{
  this.Init(displayOptions);
};

/**
 * Registers the InlineMessagingStrategy constructor function with oj.ComponentMessaging.
 *
 * @private
 */
oj.ComponentMessaging.registerMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.INLINE,
                                                oj.InlineMessagingStrategy);

// Subclass from oj.MessagingStrategy
oj.Object.createSubclass(oj.InlineMessagingStrategy,
  oj.MessagingStrategy, "oj.InlineMessagingStrategy");

/**
 * Reinitializes with the new display options and updates component messaging using the new content.
 *
 * @param {Array.<string>} newDisplayOptions
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 * @override
 *
 */
oj.InlineMessagingStrategy.prototype.reactivate = function (newDisplayOptions)
{
  oj.InlineMessagingStrategy.superclass.reactivate.call(this, newDisplayOptions);
  this._updateInlineMessage();
};

/**
 * Returns true if the messaging content should update. This method is an
 * optimization because the update() method is called too often and any time any content changes.
 * The only time InlineMessagingStrategy#update needs to execute is when the oj.ComponentValidity
 * object is in the content because we don't add this unless there are messages.
 *
 * @param {Object=} content the messaging content object. If it contains validityState, then
 * this means the component has messaging content.
 * @return {boolean}
 * @memberof oj.InlineMessagingStrategy
 * @instance
 * @private
 * @override
 */
oj.InlineMessagingStrategy.prototype.shouldUpdate = function (content)
{
  // content is messaging content, and in EditableValue we add a validityState for
  // messages, valid, and severity, so if validityState is there, we know messages are there.
  return content && content.validityState !== undefined ? true : false;
};

/**
 * Updates component with instance using the content provided.
 * @memberof oj.InlineMessagingStrategy
 * @instance
 * @private
 * @override
 */
oj.InlineMessagingStrategy.prototype.update = function ()
{
  oj.InlineMessagingStrategy.superclass.update.call(this);
  this._updateInlineMessage();
};

/**
 * Cleans up messages on the component and destroys any widgets it created.
 *
 * @memberof oj.InlineMessagingStrategy
 * @instance
 * @private
 * @override
 */
oj.InlineMessagingStrategy.prototype.deactivate = function ()
{
  this._removeMessagingContentRootDom();
  oj.InlineMessagingStrategy.superclass.deactivate.call(this);
};

/**
 * Get the default animation.
 *
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._getDefaultAnimation = function()
{
  // Load the default animation once per page scope
  if (!oj.InlineMessagingStrategy._defaultAnimation)
  {
    var animation = (oj.ThemeUtils.parseJSONFromFontFamily('oj-messaging-inline-option-defaults') || {})["animation"];
    animation = animation || {};
    oj.InlineMessagingStrategy._defaultAnimation = animation; 
  }
  
  return oj.InlineMessagingStrategy._defaultAnimation;
};

/**
 * Replace animation options with runtime values, such as oldHeight and 
 * newHeight, which are specified as placeholders in the default animations
 * but their values are not known until runtime.
 *
 * @param {string|Object|Array} effects - The animation options.
 * @param {Object} map - An object containing the runtime property key-value map.
 * @return {string|Object|Array} The resolved animation effects
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._replaceAnimationOptions = function (effects, map)
{
  var effectsAsString;
  var isEffectsTypeofString;

  if (!oj.StringUtils.isString(effects))
  {
    isEffectsTypeofString = false;
    effectsAsString = JSON.stringify(effects);
  }
  else
  {
    isEffectsTypeofString = true;
    effectsAsString = effects + ""; // append "" to get around a closure compiler warning
  }

  for (var key in map)
  {
    effectsAsString = effectsAsString.replace(new RegExp(key, 'g'), map[key]);
  }

  effects = isEffectsTypeofString ? effectsAsString :
    /** @type {Object} */ (JSON.parse(effectsAsString));

  return effects;
};

/**
 * Determine the animation for displaying new messaging content.
 *
 * @param {jQuery} rootElem - The root element for inline message.
 * @param {string} newContent - The new content to display.
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._determineAnimation = function(rootElem, newContent)
{
  var action;
  var effect;
  var parsedEffect;
  var defaults = this._getDefaultAnimation();
  if (defaults)
  {
    var oldContent = rootElem[0].innerHTML;// @HTMLUpdateOK
    var oldHeight = rootElem.outerHeight();
    var newHeight;
    
    rootElem[0].innerHTML = newContent;// @HTMLUpdateOK
    newHeight = rootElem.outerHeight();
    rootElem[0].innerHTML = oldContent;// @HTMLUpdateOK

    action = newHeight > oldHeight ? 'open' : 'close';
    effect = defaults[action];

    if (effect)
    {
      parsedEffect = this._replaceAnimationOptions(effect, {'#oldHeight': oldHeight + 'px',
                                                            '#newHeight': newHeight + 'px'});
    }
  }
  
  return {action: action,
          effect: parsedEffect};
};

/**
 * Set busy state before opening or closing inline message.
 *
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._setBusyState = function()
{
  // Set a page-level busy state if not already set  
  if (!this._resolveBusyState)
  {
    var component = this.GetComponent();
    var jElem = component ? component.element : null;
    var domElem = jElem ? jElem[0] : null;
    var busyContext = oj.Context.getContext(domElem).getBusyContext();
    var description = 'The page is waiting for inline message ';
    
    if (domElem && domElem.id)
    {
      description += 'for "' + domElem.id + '" ';
    }
    description += 'to open/close';

    this._resolveBusyState = busyContext.addBusyState({'description': description});
  }  
};

/**
 * Clear busy state after opening or closing inline message.
 *
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._clearBusyState = function()
{
  if (this._resolveBusyState)
  {
    this._resolveBusyState();
    this._resolveBusyState = null;
  }
};

/**
 * Queue up inline messaging open/close actions so that it only animate once
 * for multiple updates within the same tick.
 *
 * @param {string} contentToShow - The messaging content to show.
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._queueAction = function(contentToShow)
{
  var self = this;
  var rootElem = this.$messagingContentRoot;

  this._setBusyState();
  
  // If there is a pending timeout, clear it and set a new one so that only the
  // last animation queued within the same tick will be run.  Otherwise we will
  // end up with too many animation since the messaging framework keeps clearing
  // and updating the message display during validation, etc.
  if (this._timeoutId)
  {
    clearTimeout(this._timeoutId);
  }

  this._timeoutId = setTimeout(function() {
    self._timeoutId = null;
    
    // Make sure $messagingContentRoot is still there.  It could have been
    // removed by the time the timeout function is run.
    if (rootElem && rootElem[0])
    {
      // Parse and substitute runtime values in animation options
      var actionEffect = self._determineAnimation(rootElem, contentToShow);
      var action = actionEffect.action;
      var effect = actionEffect.effect;
      
      // Set the new content first if we're opening
      if (action == 'open')
      {
        rootElem[0].innerHTML = contentToShow;// @HTMLUpdateOK
      }

      // Invoke animation
      oj.AnimationUtils.startAnimation(rootElem[0], 'inline-' + action, effect, self.GetComponent()).then(function() {
        // Set the new content last if we're closing
        if (action == 'close')
        {
          rootElem[0].innerHTML = contentToShow;// @HTMLUpdateOK
        }

        // Clear busy state when animation end
        self._clearBusyState();        
      });
    }
    else
    {
      // Just clear the busy state if $messagingContentRoot no longer exists
      self._clearBusyState();
    }
  }, 0);
};

/**
 * If the inline message is already open its contents need to updated when update() or
 * reactivate() is called.
 *
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._updateInlineMessage = function()
{
  var contentToShow;
  var domNode;

  // contentToShow will be "" (a falsey) if there are no messages to show.
  contentToShow = this._buildInlineHtml();

  // create the inline messaging dom if there is content to show and the dom hasn't been created.
  if (contentToShow && this.$messagingContentRoot == null)
  {
    this._createInlineMessage();
  }

  if (this.$messagingContentRoot && this.$messagingContentRoot[0])
  {
    // Only enable default animation for custom elements so that automated tests
    // for legacy components are not affected    
    if (this.GetComponent()._IsCustomElement())
    {
      // This may be called multiple times within the same event cycle because the
      // old message is cleared before validation and the message is
      // reconstructed.  Since we now have animation for inline message, we don't 
      // want to update the DOM every single time.  Instead we queue up the 
      // updates and will only show the last one within the same event cyle.    
      this._queueAction(contentToShow);
    }
    else
    {
      // Legacy components don't have animation so just update the DOM
      if (contentToShow)
      {
        // push new content into inline message dom
        domNode = this.$messagingContentRoot[0];

        // contentToShow includes content that may come from app. It is scrubbed for illegal tags
        // before setting to innerHTML
        domNode.innerHTML = contentToShow;  // @HTMLUpdateOK

      }
      else
      {
        // if there is no content to show and inline message dom is currently there, remove the dom.
          // NOTE: If you see that a button seems to be losing its click event
          // after inline messaging validation or after a reset
          // it may be because the button is moving as a result of the inline messaging appearing
          // and/or disappearing. The workaround for the user is to use 'mousedown' event instead
          // of the 'click' event.
        this._removeMessagingContentRootDom();
      }
    }
  }
};

oj.InlineMessagingStrategy.prototype._createInlineMessage = function ()
{
  var widget;
  
  this.$messagingContentRoot = $(this._getInlineContentHtml());
  this._addAriaDescribedBy(this.$messagingContentRoot);
  this._addAriaLive(this.$messagingContentRoot);
  // append content that goes in inline messaging div
  // make it the very LAST child of the widget.
  widget = this.GetComponent().widget();
  widget.append(this.$messagingContentRoot); // @HTMLUpdateOK
};

/**
 * Returns the dom for the messaging-inline-container.
 *
 * @return {string}
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._getInlineContentHtml = function ()
{
  return "<div class='oj-messaging-inline-container'></div>";
};

/**
 * Removes the messaging content root dom and anything else that was adding during the
 * creation of the messaging content root dom.
 *
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._removeMessagingContentRootDom = function ()
{
  if (this.$messagingContentRoot != null)
  {
    this._removeAriaDescribedBy(this.$messagingContentRoot);
    this.$messagingContentRoot.remove();
    this.$messagingContentRoot = null;
  }
};

/**
 * Create an id to put on the root dom element that holds the inline messaging content,
 * then add aria-describedby on the launcher (this is what PopupMessaging does as well).
 * This makes it so the screen reader user knows the messaging content is connected to the launcher.
 * @memberof oj.InlineMessagingStrategy
 * @instance
 * @private
 */
oj.InlineMessagingStrategy.prototype._addAriaDescribedBy = function (messagingRoot)
{
  var describedby;
  var launcher;
  var messagingRootId;
  var tokens;

  // create an id on the div holding the inline messaging.
  // add aria-describedby to the launcher to associate the launcher and the inline message
  launcher = this.GetLauncher();

  oj.Assert.assertPrototype(launcher, $);
  oj.Assert.assertPrototype(messagingRoot, $);

  messagingRootId = messagingRoot.uniqueId().attr("id");
  describedby = launcher.attr("aria-describedby");
  tokens = describedby ? describedby.split(/\s+/) : [];
  tokens.push(messagingRootId);
  describedby = $.trim(tokens.join(" "));
  launcher.attr("aria-describedby", describedby);
};

/**
 * aria-live: polite
 * This is needed so a screen reader will read the inline message without the user needing to
 * set focus to the input field.
 * @param {jQuery} messagingRoot
 * @memberof oj.InlineMessagingStrategy
 * @instance
 * @private
 */
oj.InlineMessagingStrategy.prototype._addAriaLive = function (messagingRoot)
{
  oj.Assert.assertPrototype(messagingRoot, $);
  messagingRoot.attr("aria-live", "polite");
};

/**
 * Removes the aria-describedby from the launcher that was added by _addAriaDescribedBy
 * @param {jQuery} messagingRoot
 * @memberof oj.InlineMessagingStrategy
 * @instance
 * @private
 */
oj.InlineMessagingStrategy.prototype._removeAriaDescribedBy = function(messagingRoot)
{
  var describedby;
  var index;
  var launcher;
  var messagingRootId;
  var tokens;

  launcher = this.GetLauncher();
  oj.Assert.assertPrototype(launcher, $);
  oj.Assert.assertPrototype(messagingRoot, $);

  messagingRootId = messagingRoot.attr("id");
  describedby = launcher.attr("aria-describedby");
  tokens = describedby ? describedby.split(/\s+/) : [];
  index = $.inArray(messagingRootId, tokens);
  if (index !== -1)
    tokens.splice(index, 1);
  describedby = $.trim(tokens.join(" "));

  if (describedby)
    launcher.attr("aria-describedby", describedby);
  else
    launcher.removeAttr("aria-describedby");
};

/**
 * Returns the content to show inside the inline message html.
 * @return {string} content to show, else "". "" is a falsey.
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._buildInlineHtml = function ()
{
  var document;

  // For now anyway, ShowMessages is always true since the inlineMessaging case is for messaging.
  if (this.ShowMessages())
  {
    document = this.GetComponent().document[0];
    // returns messages or "" if there are none
    return this._buildMessagesHtml(document);
  }
  else
    return "";
};

/**
 * Returns the content to show inside messages (not hints)
 * @param {Document} document
 * @return {string} content if there are messages, else "". "" is a falsey.
 * @private
 * @memberof oj.InlineMessagingStrategy
 * @instance
 */
oj.InlineMessagingStrategy.prototype._buildMessagesHtml = function (document)
{
  var content = "";
  var maxSeverity;
  var messages;
  var renderSeveritySelectors = true;

  if (this.HasMessages())
  {
    messages = this.GetMessages();
    maxSeverity = this.GetMaxSeverity();
    content =
      oj.PopupMessagingStrategyUtils.buildMessagesHtml(
        document, messages, maxSeverity, renderSeveritySelectors);
  }
  return content;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/

/**
 * A messaging strategy that uses an instance of ojPopup component to show and hide messaging content.
 *
 * @param {Array.<string>} displayOptions an array of messaging artifacts displayed in the popup. e.g,
 * 'messages', 'converterHints', 'validationHints', 'title'.
 * @constructor
 * @extends {oj.MessagingStrategy}
 * @private
 */
oj.PopupMessagingStrategy = function (displayOptions)
{
  this.Init(displayOptions); 
};

/**
 * Registers the PopupMessagingStrategy constructor function with oj.ComponentMessaging.
 *
 * @private
 */
oj.ComponentMessaging.registerMessagingStrategy(oj.ComponentMessaging._DISPLAY_TYPE.NOTEWINDOW,
oj.PopupMessagingStrategy);

// Subclass from oj.MessagingStrategy
oj.Object.createSubclass(oj.PopupMessagingStrategy, oj.MessagingStrategy, "oj.PopupMessagingStrategy");

/**
 * Messaging popup defaults for components, by component type. A special 'default' type defines the
 * defaults for most editableValue components.
 * The following properties are available -
 * 'events' - these specify the on handlers for events that are setup to open and close popups
 * 'position' - specifies the type of element the popup is positioned against.
 * @private
 */
oj.PopupMessagingStrategy._DEFAULTS_BY_COMPONENT =
{
  // mouseenter and mouseleave is what you want instead of mouseover/mouseout when the launcher
  // isn't a simple input. In the case of radioset and checkboxset, the launcher is the widget
  // which is the div that contains all the rows, inputs and labels. If we use mouseover/mouseout
  // in this case we are constantly opening and closing the popup (not really visible to the user,
  // but still not good for performance I'm sure) if the user moves the mouse around the different
  // dom elements within the widget.
  //
  // on touch devices: the "press" event name maps to Hammer's press event, so a touch and hold
  // will open the popup.
  "ojRadioset":
  {
    position: 'launcher',
    // when press opens popup, the user taps elsewhere to dismiss popup
    events: {open: "focusin mouseenter press", close: "mouseleave"}
  },
  "ojCheckboxset":
  {
    position: 'launcher',
    // when press opens popup, the user taps elsewhere to dismiss popup
    events: {open: "focusin mouseenter press", close: "mouseleave"}
  },
  // Since we now add extra dom on the input components for inline messages, we don't want to
  // position on the tip of the component root. Instead we want to position on the main part of the
  // component, which is in a lot of cases the launcher. In the case of inputDate/Time/Number,
  // it's the launcher's parent (inputDate/Time/Number wrap input and buttons with a parent).
  "ojInputText":
  {
    position: 'launcher',
    events: {open: "focusin"}
  },
  "ojTextArea":
  {
    position: 'launcher',
    events: {open: "focusin"}
  },
  "ojInputPassword":
  {
    position: 'launcher',
    events: {open: "focusin"}
  },
  "ojSwitch":
  {
    position: 'launcher',
    events: {open: "focusin mouseenter", close: "mouseleave"}
  },
  "ojSlider":
  {
    position: 'launcher',
    events: {open: "focusin mouseenter", close: "mouseleave"}
  },
  "ojColorSpectrum":
  {
    position: 'launcher',
    events: {open: "focusin mouseenter", close: "mouseleave"}
  },
  "ojColorPalette":
  {
    position: 'launcher',
    events: {open: "focusin mouseenter", close: "mouseleave"}
  },
  "default":
  {
    position: 'launcher-wrapper',
    events: {open: "focusin"}
  }
};

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT = "oj-form-control-hint";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT_CONVERTER = "oj-form-control-hint-converter";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT_VALIDATOR = "oj-form-control-hint-validator";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT_TITLE = "oj-form-control-hint-title";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategy._OPEN_NAMESPACE = ".ojPopupMessagingOpen";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategy._CLOSE_NAMESPACE = ".ojPopupMessagingClose";

/**
 * Sets up a tooltip for the component instance using the messaging content provided.
 *
 * @param {Object} cm a reference to an instance of oj.ComponentMessaging that provides access to
 * the latest messaging content.
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @override
 * @instance
 */
oj.PopupMessagingStrategy.prototype.activate = function (cm)
{
  oj.PopupMessagingStrategy.superclass.activate.call(this, cm);
  this._initMessagingPopup();
};

/**
 * Reinitializes with the new display options and updates component messaging using the new content.
 *
 * @param {Array.<string>} newDisplayOptions
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 * @override
 */
oj.PopupMessagingStrategy.prototype.reactivate = function (newDisplayOptions)
{
  oj.PopupMessagingStrategy.superclass.reactivate.call(this, newDisplayOptions);
  this._updatePopupIfOpen();
};

oj.PopupMessagingStrategy.prototype.update = function ()
{
  oj.PopupMessagingStrategy.superclass.update.call(this);
  this._updatePopupIfOpen();
};

/**
 * Cleans up messages on the component and destroys any widgets it created.
 *
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 * @override
 */
oj.PopupMessagingStrategy.prototype.deactivate = function ()
{
  this._unregisterLauncherEvents();
  this._destroyTooltip();
  oj.PopupMessagingStrategy.superclass.deactivate.call(this);
};
/**
 * Close the popup if it is open. EditableValue calls this from _NotifyHidden and _NotifyDetached
 * so that we don't have an open popup if the app dev hides a subtree the component is within.
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype.close = function ()
{
  this._closePopup();
};
/**
 * Closes the associated notewindow popup
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._closePopup = function ()
{
  function doClose(resolve)
  {
    if (this._isPopupInitialized())
    {
      if (resolve)
      {
        // Add an event listener to resolve the promise
        this._setActionResolver(this.$messagingContentRoot, "close", resolve);
      }
          
      this.$messagingContentRoot.ojPopup("close");

      // Just return if we call ojPopup close.  The promise will be resolved
      // by the ojclose event listener.
      return;
    }
    
    if (resolve)
    {
      // Resolve the promise immediately if we didn't call ojPopup close
      resolve(true);
    }
  }
  
  this._queueAction(doClose.bind(this));
};

/**
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._initMessagingPopup = function ()
{
  if (!this._openPopupCallback)
    this._registerLauncherEvents();
};

/**
 * Add listeners for animation events.
 * We use this to delegate animation events to the editableValue component since
 * the original events are triggered on the popup, which is created internally 
 * and the application cannot bind listeners to it.  By delegating the events, 
 * application can bind the listeners to the component.
 * 
 * @param {jQuery} messagingContentRoot - The jQuery object for the messaging root node
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._addAnimateEventListeners = function(messagingContentRoot)
{
  var delegateEvent = function(newEventType, event, ui) {
    var component = this.GetComponent();
    if (component && component._trigger)
    {
      // always stop propagation if we have a component to delegate to
      event.stopPropagation();

      // prevent default only if the component handler says so, as indicated by
      // a return value of false.
      if (!component._trigger(newEventType, null, ui))
      {
        event.preventDefault();
      }
    }
  };

  // Add animation event listeners to delegate the events to the component
  messagingContentRoot.on('ojanimatestart.notewindow', delegateEvent.bind(this, 'animateStart'));
  messagingContentRoot.on('ojanimateend.notewindow', delegateEvent.bind(this, 'animateEnd'));
};

/**
 * Remove listeners for animation events.
 * 
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._removeAnimateEventListeners = function(messagingContentRoot)
{
  messagingContentRoot.off('ojanimatestart.notewindow');
  messagingContentRoot.off('ojanimateend.notewindow');
};

/**
 * Set busy state on the component that invokes the notewindow.
 * 
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._setBusyState = function(eventType)
{
  var component = this.GetComponent();
  var jElem = component ? component.element : null;
  var domElem = jElem ? jElem[0] : null;
  var busyContext = oj.Context.getContext(domElem).getBusyContext();
  var description = 'The page is waiting for note window ';
  
  if (domElem && domElem.id)
  {
    description += 'for "' + domElem.id + '" ';
  }
  description += 'to ' + eventType;

  return busyContext.addBusyState({'description': description});
};

/**
 * Set an event listener to resolve promise when popup open/close action ends.
 * 
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._setActionResolver = function(messagingContentRoot, eventType, resolvePromise)
{
  var animationOption;

  // Disable animation if there are other queued actions.  Otherwise we will end
  // up with too many animation since the messaging framework keeps clearing and
  // updating the message display during validation, etc.
  if (this._actionCount > 1)
  {
    // Remember the original animation so that we can restore it later
    animationOption = messagingContentRoot.ojPopup("option", "animation");
    messagingContentRoot.ojPopup("option", "animation", null);
  }

  // Add a busy state for the component.  Even though ojpopup add busy state,
  // it is in the scope of the popup element.
  var resolveBusyState = this._setBusyState(eventType);

  // Add an one-time listener to resolve the promise
  messagingContentRoot.one('oj' + eventType, function() {
    // Restore any saved animation option
    if (animationOption)
    {
      messagingContentRoot.ojPopup("option", "animation", animationOption);
    }
    
    resolveBusyState();
    resolvePromise(true);
  });
};

/**
 * Queue up popup open and close actions so that they are executed in the
 * correct order.
 *
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._queueAction = function(task)
{
  if (this.GetComponent()._IsCustomElement())
  {
    // Queue up the action for custom elements to avoid animation overlapping each other
    var self = this;

    var createActionPromise = function(task)
    {
      var promise = new Promise(task);
      promise.then(function() {
        --self._actionCount;
      });
      return promise;
    };
    
    if (!this._actionCount)
    {
      // If there is no action in progress, create a new promise directly instead
      // of chaining to any resolved promise to avoid an extra wait state.
      this._actionCount = 1;
      this._actionPromise = createActionPromise(task);
    }
    else
    {
      ++this._actionCount;
      this._actionPromise = this._actionPromise.then(function() {
        return createActionPromise(task);
      });
    }
  }
  else
  {
    // Invoke the action immediately for legacy components since there is no animation
    task(null);
  }
};

/**
 * Opens a popup. This handler is called in the context of the launcher usually the this.element or
 * some relevant node the messaging popup is associated to.
 *
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._openPopup = function (event)
{
  function doOpen(resolve)
  {
    var domNode;
    var latestContent;
    var $launcher;

    
    if (this._canOpenPopup())
    {

      latestContent = this._buildPopupHtml();
      if (!oj.StringUtils.isEmptyOrUndefined(latestContent))
      {
        var messagingContentRoot = this._getPopupElement();
        var isPopupOpen = messagingContentRoot.ojPopup("isOpen");

        // replace popup messaging content with new content
        domNode = oj.PopupMessagingStrategyPoolUtils.getPopupContentNode(messagingContentRoot);

        // latestContent is includes content that may come from app. It is scrubbed for illegal tags
        // before setting to innerHTML
        domNode.innerHTML = ""; // @HTMLUpdateOK
        domNode.innerHTML = latestContent; // @HTMLUpdateOK

        if (!isPopupOpen)
        {
          $launcher = this.GetLauncher();
          if (event.type === "press")
          {
            this._openPopupOnPressEvent($launcher);
          }

          if (resolve)
          {
            // Add an event listener to resolve the promise
            this._setActionResolver(messagingContentRoot, "open", resolve);
          }
          
          messagingContentRoot.ojPopup("open", $launcher);
          
          // Just return if we call ojPopup open.  The promise will be resolved
          // by the ojopen event listener.
          return;
        }
        else if (isPopupOpen)
        {
          messagingContentRoot.ojPopup("refresh");
        }
      }
    }
    
    if (resolve)
    {
      // Resolve the promise immediately if we didn't call ojPopup open
      resolve(true);
    }
  }
  
  this._queueAction(doOpen.bind(this));
};

/**
 * This is called to open the popup on the 'press' event. E.g., ojCheckboxset and ojRadioset
 * use press to open the popup.
 * @param {Object|null} jqLauncher
 */
oj.PopupMessagingStrategy.prototype._openPopupOnPressEvent = function (jqLauncher)
{
  this._inPressEvent = true;

  // We add these event listeners when we open the popup as a result of the 'press' event
  // and we are going to remove them when we close the popup, as well as when we unregister
  // launcher events to make doubly sure they aren't lying around.
  /// Use capture phase to make sure we cancel it before any regular bubble listeners hear it.
  jqLauncher[0].addEventListener("click", this._eatChangeAndClickOnPress, true);
  // need to eat 'change' as well. Otherwise the dialog will close on press up, and the input
  // stays unchecked.
  // This is because when the input  gets the 'change' event, it calls validate,
  // which then updates messages, and if there is no message,
  // then calls _updatePopupIfOpen, contentToShow = "", then it closes the popup.
  jqLauncher[0].addEventListener("change", this._eatChangeAndClickOnPress, true);

  // touchend/mousedown/change/click happen in fast succession on tap or press.
  // Android never fires a click event on press up, so after 50ms we clear the inPressEvent flag
  // since the _eatChangeAndClickOnPress callback never gets called for Android.
  jqLauncher.one("touchend", function (event)
  {
    // 50ms.  Make as small as possible to prevent unwanted side effects.
    setTimeout(function ()
    {
      this._inPressEvent = false;
    }, 50);
  });
};

/**
 * The pressHold gesture fires a click and change event on ios after touchend.  Prevent that here.
 * @private
 */
oj.PopupMessagingStrategy.prototype._eatChangeAndClickOnPress = function (event)
{
      // on ios:
      // if I tap quickly on an input, I get on div: touchstart/touchend/mousedown/change/click
      // if I tap and hold on an input, I get: touchstart
      // when I let up, I get: touchend/mousedown/change/click
      // on android:
      // if I tap quickly on an input, I get touchstart touchend mousedown click change
      // if I tap and hold on an input, I get touchstart/mousedown
      // when I let up, I get touchend. (no change or click like I do for ios)

      // After 'press' release of a radio or checkbox if we do not eat the the click and change events,
      // the dialog closes.
      if (this._inPressEvent)
      {
        // For Mobile Safari capture phase at least,
        // returning false doesn't work; must use pD() and sP() explicitly.
        event.preventDefault();
        event.stopPropagation();
        // the event order is first change, then click.
        // so when we get the click, clear the inPressEvent flag.
        if (event.type === "click")
          this._inPressEvent = false;
      }
};
    
/**
 * Determines whether the messaging popup can be opened.
 * @return {boolean}
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._canOpenPopup = function ()
{
  var options = this.GetComponent().options;
  var isDisabled = options['disabled'] || false;
  var isReadOnly = options['readOnly'] || false;

  return !(isDisabled || isReadOnly);
};

/**
 * If the popup is already open its contents need to updated when update() or reactivate() is called.
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._updatePopupIfOpen = function ()
{
  var contentToShow;
  var domNode;
  var isLauncherActiveElement;
  var isPopupOpen = false;
  var launcher;
  var messagingContentRoot;

  if (this._isPopupInitialized())
  {
    messagingContentRoot = this._getPopupElement();
    isPopupOpen = messagingContentRoot.ojPopup("isOpen");
    contentToShow = this._buildPopupHtml();
    launcher = this.GetLauncher();
    if (launcher == null)
      return;
    isLauncherActiveElement = document.activeElement === this.GetLauncher()[0] ? true : false;
    if (isPopupOpen)
    {
      if (contentToShow)
      {
        // push new content into popup
        domNode = oj.PopupMessagingStrategyPoolUtils.getPopupContentNode(messagingContentRoot);

        // contentToShow is includes content that may come from app. It is scrubbed for illegal tags
        // before setting to innerHTML
        domNode.innerHTML = ""; // @HTMLUpdateOK
        domNode.innerHTML = contentToShow; // @HTMLUpdateOK
        messagingContentRoot.ojPopup("refresh");
      }
      else
      {
        // if there is no content to show and popup is currently open, close it.
        messagingContentRoot.ojPopup("close");
      }
    }
    else if (isLauncherActiveElement && contentToShow)
    {
      // if popup is closed but focus is on activeElement re-open it
      this._openPopup(undefined);
    }
  }
};

/**
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._unregisterLauncherEvents = function ()
{
  var compDefaults;
  var events;
  var jqLauncher = this.GetLauncher();

  compDefaults = oj.PopupMessagingStrategy._DEFAULTS_BY_COMPONENT[this.GetComponent().widgetName];
  events = compDefaults ?
  compDefaults.events : oj.PopupMessagingStrategy._DEFAULTS_BY_COMPONENT["default"].events;

  // Remove event handlers setup on launcher
  jqLauncher.off(oj.PopupMessagingStrategy._OPEN_NAMESPACE);
  jqLauncher.off(oj.PopupMessagingStrategy._CLOSE_NAMESPACE);
  jqLauncher[0].removeEventListener("click", this._eatChangeAndClickOnPress, true);
  jqLauncher[0].removeEventListener("change", this._eatChangeAndClickOnPress, true);

  if (oj.DomUtils.isTouchSupported())
  {
    jqLauncher.ojHammer().off("press");
    jqLauncher.ojHammer("destroy");
    jqLauncher.off("contextmenu", this._eatContextMenuOnOpenPopupListener);
    this._eatContextMenuOnOpenPopupListener = null;
    this._inPressEvent = null;
  }
  this._openPopupCallback = null;
  this._closePopupCallback = null;
};

/**
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._registerLauncherEvents = function ()
{
  var closeEvents;
  var closePopupCallback;
  var compDefaults;
  var events;
  var hammerOptions;
  var jqLauncher = this.GetLauncher();
  var nonPressOpenEvents;
  var openPopupCallback;
  var pressEventIndex;

  compDefaults = oj.PopupMessagingStrategy._DEFAULTS_BY_COMPONENT[this.GetComponent().widgetName];
  events = compDefaults ? compDefaults.events :
           oj.PopupMessagingStrategy._DEFAULTS_BY_COMPONENT["default"].events;

  // 1. associate the ojPopup component to wrapper <div> for popup content
  // 2. wire up on() event handlers for registered events that open and close popup. E.g., focusin.
  // 3. autoDismissal happens automatically when focus leaves component. For other events like
  // mouseover it's required to call off()
  if (events['open'])
  {
    openPopupCallback = this._openPopupCallback;
    if (!openPopupCallback)
      openPopupCallback = this._openPopupCallback = this._openPopup.bind(this);
    
    // separate out press event, namespace the events string, and attach event handler
    pressEventIndex = events['open'].indexOf("press");
    nonPressOpenEvents = 
    this._getNamespacedEvents(events['open'].replace("press", ''), 
                              oj.PopupMessagingStrategy._OPEN_NAMESPACE);
    jqLauncher.on(nonPressOpenEvents, openPopupCallback);
    
    // The pressHold gesture also fires a contextmenu event on Windows 10 touch.  
    // Prevent that here.
    if (oj.DomUtils.isTouchSupported())
    {
      this._eatContextMenuOnOpenPopupListener = function (event)
      {
        return false;
      };

      jqLauncher.on("contextmenu", this._eatContextMenuOnOpenPopupListener);

      // for radios and checkboxes, on ios, press hold brings up popup, but release closes it
      // and checks it, so in this case we have to eat the click/change events. this happens
      // in the openPopupCallback
      if (pressEventIndex !== -1)
      {
        hammerOptions = {
          "recognizers": [
            [Hammer.Press, {time: 750}]
          ]};
        jqLauncher.ojHammer(hammerOptions).on("press", openPopupCallback);
      }
    }
  }

  if (events['close'])
  {
    closePopupCallback = this._closePopupCallback;
    if (!closePopupCallback)
      closePopupCallback = this._closePopupCallback = this._closePopup.bind(this);

    closeEvents = this._getNamespacedEvents(events['close'], 
      oj.PopupMessagingStrategy._CLOSE_NAMESPACE);
    jqLauncher.on(closeEvents, closePopupCallback);
  }
};

/**
 * Turn the events string into an array, add namespace, and turn it back into a string.
 * @param {string} events e.g., "focusin mousedown"
 * @param {string} namespace the namespace that starts with a dot
 * @return {string|null} the events string that is namespaced. 
 * e.g., "focusin.ojPopupMessagingOpen mousedown.ojPopupMessagingOpen"
 * @private
 */
oj.PopupMessagingStrategy.prototype._getNamespacedEvents = function(events, namespace)
{
  var eventsArray;
  var namespacedEventsArray;
  var length;
      
  if (events === "" || namespace === "")
    return events;

  eventsArray = events.split(" ");
  length = eventsArray.length;
  namespacedEventsArray = [];
  
  for (var i=0; i < length; i++)
  {
    // ignore ""
    if (eventsArray[i])
      namespacedEventsArray.push(eventsArray[i] + namespace);
  }
  
  return namespacedEventsArray.join(" ");
};

/**
 * Returns the popup position options.
 * Components like radio and checkboxset use the launcher, which is the inputs.
 * Since we now add extra dom for inline messages, we don't want to position
 * on the tip of the component root. Instead we want to position on the main part of the component,
 * which is in a lot of cases the launcher. In the case of inputDate/Time/Number, it's the launcher's
 * parent.
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._getPopupPosition = function ()
{
  var compDefaultPosition;
  var compDefaults;
  var launcher;
  var popupPositionOptions;

  compDefaults =
  oj.PopupMessagingStrategy._DEFAULTS_BY_COMPONENT[this.GetComponent().widgetName];
  compDefaultPosition = compDefaults ? compDefaults.position :
  oj.PopupMessagingStrategy._DEFAULTS_BY_COMPONENT["default"].position;

  if (compDefaultPosition)
  {
    if (compDefaultPosition === "launcher")
    {
      launcher = this.GetLauncher();
    }
    else if (compDefaultPosition === "launcher-wrapper")
    {
      launcher = this.GetLauncher().parent();
    }
  }
  // should never get here since the _DEFAULTS_BY_COMPONENTS["default"] should cover it.
  if (!launcher)
    launcher = this.GetComponent().widget();

  popupPositionOptions = {
    'my': 'start bottom',
    'at': 'end top',
    'collision': 'flipcenter',
    'of': launcher
  };
  return popupPositionOptions;

};

/**
 * Returns a jquery element that a messaging popup is bound to.
 *
 * @return {jQuery!} messaging popup pool container
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._getPopupElement = function ()
{
  var popup;
  var position;

  if (this.$messagingContentRoot)
    return this.$messagingContentRoot;

  popup = oj.PopupMessagingStrategyPoolUtils.getNextFreePopup();
  position = this._getPopupPosition();
  popup.ojPopup("option", "position", position);
  popup.ojPopup("option", "beforeClose", this._popupBeforeCloseCallback.bind(this));
  popup.ojPopup("option", "close", this._popupCloseCallback.bind(this));
  popup.ojPopup("option", "open", this._popupOpenCallback.bind(this));
  
  // Use default animation only for custom elements
  if (this.GetComponent()._IsCustomElement())
  {
    // Get the default animation
    var defaultAnimations = (oj.ThemeUtils.parseJSONFromFontFamily('oj-messaging-popup-option-defaults') || {})["animation"];
    defaultAnimations["actionPrefix"] = "notewindow";
    popup.ojPopup("option", "animation", defaultAnimations);
  
    this._addAnimateEventListeners(popup);
  }
  else
  {
    popup.ojPopup("option", "animation", null);
  }

  this.$messagingContentRoot = popup;
  return this.$messagingContentRoot;
};

/**
 * Popup open event listener that changes the popups autoDismiss to focusLoss
 * in a timeout of 10ms.  This timeout period gives the browser time to fire
 * events that might follow a click such as a focus event.  This is to allow
 * validation by a button versus a component instance.
 *
 * @param {jQuery.event=} event
 * @memberof! oj.PopupMessagingStrategy
 * @private
 */
oj.PopupMessagingStrategy.prototype._popupOpenCallback = function (event)
{
  var target = $(event.target);
  var self = this;
  window.setTimeout(function ()
  {
    if (oj.Components.isComponentInitialized(target, "ojPopup"))
      target.ojPopup("option", "autoDismiss", "focusLoss");
    else
      delete self.$messagingContentRoot;
  }, 10);
};

/**
 * Popup beforeClose event listener that will add busy state to the component
 * @param {jQuery.event=} event
 * @memberof! oj.PopupMessagingStrategy
 * @private
 */
oj.PopupMessagingStrategy.prototype._popupBeforeCloseCallback = function (event)
{
  this._resolveBusyState = this._setBusyState('close');
};

/**
 * Popup closed event listener that will reset the popups state and free it into the
 * pool of available messaging popups.
 * @param {jQuery.event=} event
 * @memberof! oj.PopupMessagingStrategy
 * @private
 */
oj.PopupMessagingStrategy.prototype._popupCloseCallback = function (event)
{
  var jqLauncher, popupContent, target;
  jqLauncher = this.GetLauncher();

  target = $(event.target);
  
  this._removeAnimateEventListeners(target);

  if (oj.Components.isComponentInitialized(target, "ojPopup"))
  {
    target.ojPopup("option", "autoDismiss", "none");
    target.ojPopup("option", "open", null);
    target.ojPopup("option", "close", null);
    target.ojPopup("option", "beforeClose", null);
  }

  // Check that the launcher is still there when removing listeners  
  if (jqLauncher && jqLauncher[0])
  {
    jqLauncher[0].removeEventListener("click", this._eatChangeAndClickOnPress, true);
    jqLauncher[0].removeEventListener("change", this._eatChangeAndClickOnPress, true);
  }
    
  this.$messagingContentRoot = null;
  this._inPressEvent = null;

  popupContent = $(oj.PopupMessagingStrategyPoolUtils.getPopupContentNode(target));
  popupContent.empty();
  
  if (this._resolveBusyState)
  {
    this._resolveBusyState();
    this._resolveBusyState = null;
  }
};

/**
 * @memberof! oj.PopupMessagingStrategy
 * @private
 * @instance
 */
oj.PopupMessagingStrategy.prototype._destroyTooltip = function ()
{
  this._closePopup();
  oj.PopupMessagingStrategyPoolUtils.destroyFreePopup();
};

/**
 * Returns the content to show inside popup.
 * @private
 * @return {String|string} content
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._buildPopupHtml = function ()
{
  var addSeparator = false;
  var document = this.GetComponent().document[0];
  var nwContent = [];
  var nwHtml = "";

  if (this.ShowMessages())
  {
    nwContent.push(this._buildMessagesHtml(document));
  }

  if (this.ShowConverterHint() || this.ShowValidatorHint() || this.ShowTitle())
  {
    nwContent.push(this._buildHintsHtml(document));
  }

  $.each(nwContent, function (i, content)
  {
    if (content)
    {
      if (addSeparator)
      {
        nwHtml = nwHtml.concat(oj.PopupMessagingStrategyUtils.getSeparatorHtml(document));
      }
      else
      {
        addSeparator = true;
      }

      nwHtml = nwHtml.concat(content);
    }
  });

  return nwHtml;
};

/**
 * Returns the messages html (e.g., error messages, confirmation messages), not hints
 * @param {Document} document
 * @return {string} content
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._buildMessagesHtml = function (document)
{
  var content = "";
  var maxSeverity = this.GetMaxSeverity();
  var messages;
  var renderSeveritySelectors = false;

  if (this.HasMessages())
  {
    messages = this.GetMessages();
    content =
    oj.PopupMessagingStrategyUtils.buildMessagesHtml(
    document, messages, maxSeverity, renderSeveritySelectors);
  }
  return content;
};

/**
 * All hints including title
 * @param {Document} document
 * @return {string} html content for all hints.
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._buildHintsHtml = function (document)
{
  var hint;
  var hints = [];
  var hintsHtml = "";
  var i;

  if (this.ShowConverterHint())
  {
    hints = this.GetConverterHint();
    hint = hints.length ? hints[0] : "";
    hintsHtml += oj.PopupMessagingStrategyUtils.buildHintHtml(document,
    oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT_CONVERTER,
    hint, false,
    oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT);
  }

  if (this.ShowValidatorHint())
  {
    hints = this.GetValidatorHints();
    for (i = 0; i < hints.length; i++)
    {
      hintsHtml += oj.PopupMessagingStrategyUtils.buildHintHtml(document,
      oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT_VALIDATOR,
      hints[i], false, oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT);
    }
  }

  if (this.ShowTitle())
  {
    hintsHtml += oj.PopupMessagingStrategyUtils.buildHintHtml(document,
    oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT_TITLE,
    this.GetTitle(), true, oj.PopupMessagingStrategy._SELECTOR_FORMCONTROL_HINT);
  }

  return hintsHtml ? "<div class='oj-form-control-hints'>" + hintsHtml + "</div>" : "";
};

/**
 * Determines if there is a message popup currently associated with the component
 * strategy.
 * @return {boolean}
 * @private
 * @memberof oj.PopupMessagingStrategy
 * @instance
 */
oj.PopupMessagingStrategy.prototype._isPopupInitialized = function ()
{
  // is(":oj-popup") finds the popup component if it exists
  return (this.$messagingContentRoot) ?
  oj.Components.isComponentInitialized(this.$messagingContentRoot, "ojPopup") : false;
};

/**
 * @ignore
 */
oj.PopupMessagingStrategyUtils = {};

/**
 * Renders the html content for a single hint.
 * @param {Document} document
 * @param {string} selector
 * @param {string} hintText
 * @param {boolean} htmlAllowed
 * @param {string} formControlSelectors
 *
 * @return {string} html content for a single hint.
 * @public
 */
oj.PopupMessagingStrategyUtils.buildHintHtml = function (document, selector, hintText, htmlAllowed, formControlSelectors)
{
  var jTitleDom;
  if (hintText)
  {
    jTitleDom = $(document.createElement("div"));
    formControlSelectors += " " + selector;
    jTitleDom.addClass(formControlSelectors);

    jTitleDom.append(oj.PopupMessagingStrategyUtils._getTextDom(document, hintText, htmlAllowed)); // @HTMLUpdateOK
  }

  return jTitleDom ? jTitleDom.get(0).outerHTML : "";// @HTMLUpdateOK
};

/**
 * @param {number} severity
 * @returns (string} translated string for the severity
 * @public
 */
oj.PopupMessagingStrategyUtils.getSeverityTranslatedString = function (severity)
{
  var sevTypeStr;
  // get the translated string for the severity
  switch (severity)
  {
    case oj.Message.SEVERITY_LEVEL['FATAL']:
      sevTypeStr = oj.Translations.getTranslatedString('oj-message.fatal');
      break;
    case oj.Message.SEVERITY_LEVEL['ERROR']:
      sevTypeStr = oj.Translations.getTranslatedString('oj-message.error');
      break;
    case oj.Message.SEVERITY_LEVEL['WARNING']:
      sevTypeStr = oj.Translations.getTranslatedString('oj-message.warning');
      break;
    case oj.Message.SEVERITY_LEVEL['INFO']:
      sevTypeStr = oj.Translations.getTranslatedString('oj-message.info');
      break;
    case oj.Message.SEVERITY_LEVEL['CONFIRMATION']:
      sevTypeStr = oj.Translations.getTranslatedString('oj-message.confirmation');
      break;
  }

  return sevTypeStr;
};

/**
 * @param {Document} document
 * @returns {string}
 * @public
 */
oj.PopupMessagingStrategyUtils.getSeparatorHtml = function (document)
{
  var jSeparatorDom;
  jSeparatorDom = $(document.createElement("hr"));

  return jSeparatorDom ? jSeparatorDom.get(0).outerHTML : "";// @HTMLUpdateOK
};

/**
 * Returns the messages html (e.g., error messages, confirmation messages), not hints
 * @param {Document} document
 * @param {Array} messages
 * @param {number} maxSeverity
 * @param {boolean} renderSeveritySelectors
 * @return {string} content
 * @private
 * @memberof oj.PopupMessagingStrategyUtils
 * @instance
 */
oj.PopupMessagingStrategyUtils.buildMessagesHtml =
function (document, messages, maxSeverity, renderSeveritySelectors)
{
  var content = "";
  var detail;
  var i;
  var j;
  var message;
  var messagesByType = [];
  var messagesByTypes = {};
  var messageObj;
  var severityLevel;
  var severityStr;
  var summary;


  // Step1: build an indexed array of messages by severity level.
  for (i = 0; i < messages.length; i++)
  {
    message = messages[i];

    if (!(message instanceof oj.Message))
    {
      messageObj = new oj.Message(message['summary'], message['detail'], message['severity']);
    }
    else
    {
      messageObj = message;
    }

    severityLevel = oj.Message.getSeverityLevel(messageObj['severity']);
    if (!messagesByTypes[severityLevel])
    {
      messagesByTypes[severityLevel] = [];
    }

    messagesByTypes[severityLevel].push(messageObj);
  }

  // Step 2: starting with maxSeverity level build messages with decreasing severity
  for (i = maxSeverity; i >= oj.Message.SEVERITY_LEVEL['CONFIRMATION']; i--)
  {
    messagesByType = messagesByTypes[i] || [];

    for (j = 0; j < messagesByType.length; j++)
    {
      message = messagesByType[j];

      severityLevel = oj.Message.getSeverityLevel(message['severity']);
      severityStr = oj.PopupMessagingStrategyUtils.getSeverityTranslatedString(severityLevel);
      summary = message['summary'] || severityStr;

      // if detail is empty we don't care to duplicate summary. also detail if present can be
      // formatted html content (ADF feature)
      detail = message['detail'] || "";
      content = content.concat(
      oj.PopupMessagingStrategyUtils.buildMessageHtml(
      document, summary, detail, severityLevel, renderSeveritySelectors));
    }
  }
  return content;

};

/**
 * Builds the HTML content for a single message
 * @param {Document} document
 * @param {string} summary
 * @param {string} detail
 * @param {number} severityLevel
 * @returns {string}
 * @public
 */
oj.PopupMessagingStrategyUtils.buildMessageHtml =
function (document, summary, detail, severityLevel, addSeverityClass)
{
  var $msgContent;
  var $msgDetail;
  var $msgDom;
  var $msgIcon;
  var $msgSummary;
  var severityStr = oj.PopupMessagingStrategyUtils.getSeverityTranslatedString(severityLevel);

  // build message
  // (x) <Summary Text>
  // <Detail Text>
  $msgDom = $(document.createElement("div"));
  $msgDom.addClass(oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE);

  if (addSeverityClass)
    $msgDom.addClass(oj.PopupMessagingStrategyUtils._getSeveritySelector(severityLevel));

  // build msg icon
  $msgIcon = $(document.createElement("span"));
  $msgIcon.addClass(oj.PopupMessagingStrategyUtils._getSeverityIconSelector(severityLevel))
  .attr("title", severityStr)
  .attr("role", 'img');

  $msgDom.append($msgIcon); // @HTMLUpdateOK

  // build msg content which includes summary and detail
  $msgContent = $(document.createElement("span"));
  $msgContent.addClass(oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_CONTENT);

  $msgSummary = $(document.createElement("div"));
  $msgSummary.addClass(oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_SUMMARY).text(summary);

  $msgContent.append($msgSummary); // @HTMLUpdateOK

  if (detail)
  {
    // detail text allows html content. So scrub it before setting it.
    var detailDom = oj.PopupMessagingStrategyUtils._getTextDom(document, detail, true);
    $msgDetail = $(document.createElement("div"));

    $msgDetail.addClass(oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_DETAIL).append(detailDom); // @HTMLUpdateOK
    $msgContent.append($msgDetail);// @HTMLUpdateOK
  }

  $msgDom.append($msgContent); // @HTMLUpdateOK

  return $msgDom ? $msgDom.get(0).outerHTML : "";// @HTMLUpdateOK
};

/**
 * @param {number} severity
 * @return {string} the icon selector for the severity
 * @private
 */
oj.PopupMessagingStrategyUtils._getSeverityIconSelector = function (severity)
{
  var sevIconStr;
  // get the icon selector for the severity
  switch (severity)
  {
    case oj.Message.SEVERITY_LEVEL['FATAL']:
      sevIconStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_ERROR_ICON;
      break;
    case oj.Message.SEVERITY_LEVEL['ERROR']:
      sevIconStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_ERROR_ICON;
      break;
    case oj.Message.SEVERITY_LEVEL['WARNING']:
      sevIconStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_WARNING_ICON;
      break;
    case oj.Message.SEVERITY_LEVEL['INFO']:
      sevIconStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_INFO_ICON;
      break;
    case oj.Message.SEVERITY_LEVEL['CONFIRMATION']:
      sevIconStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_CONFIRMATION_ICON;
      break;
  }

  return oj.PopupMessagingStrategyUtils._DEFAULT_STATUS_ICON_SELECTORS + sevIconStr;
};

/**
 * @param {number} severity
 * @return {string} the style selector for the severity
 * @private
 */
oj.PopupMessagingStrategyUtils._getSeveritySelector = function (severity)
{
  var sevSelectorStr;
  // get the icon selector for the severity
  switch (severity)
  {
    case oj.Message.SEVERITY_LEVEL['FATAL']:
      sevSelectorStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_ERROR;
      break;
    case oj.Message.SEVERITY_LEVEL['ERROR']:
      sevSelectorStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_ERROR;
      break;
    case oj.Message.SEVERITY_LEVEL['WARNING']:
      sevSelectorStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_WARNING;
      break;
    case oj.Message.SEVERITY_LEVEL['INFO']:
      sevSelectorStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_INFO;
      break;
    case oj.Message.SEVERITY_LEVEL['CONFIRMATION']:
    default:
      sevSelectorStr = oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_CONFIRMATION;
      break;
  }

  return sevSelectorStr;
};


/**
 *  if content is html clean html by allowing only legal tags before returning, to safeguard from
 *  script injection errors.
 *
 * @param {Document} document
 * @param {string} value
 * @param {boolean=} htmlAllowed if value can have html content
 *
 * @return {Element} dom node containing the scrubbed hint
 * @private
 */
oj.PopupMessagingStrategyUtils._getTextDom = function (document, value, htmlAllowed)
{
  var textDom = null;

  if (oj.StringUtils.isString(value))
  {
    if (htmlAllowed && oj.DomUtils.isHTMLContent(value))
    {
      // strip out html start/end tags
      textDom = oj.DomUtils.cleanHtml(value.substring(6, value.length - 7));
    }
    else
    {
      textDom = document.createElement("span");
      textDom.textContent = value;
    }
  }

  return textDom;
};

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._DEFAULT_STATUS_ICON_SELECTORS = "oj-component-icon oj-message-status-icon ";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE = "oj-message";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_SUMMARY = "oj-message-summary";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_DETAIL = "oj-message-detail";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_CONTENT = "oj-message-content";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_ERROR_ICON = "oj-message-error-icon";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_WARNING_ICON = "oj-message-warning-icon";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_INFO_ICON = "oj-message-info-icon";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_CONFIRMATION_ICON = "oj-message-confirmation-icon";

// new theming keys so that we can style the different types of messages differently. Like,
// the background-color can be light red for error. This style will go on the same dom node
// as the oj-message selector.
/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_ERROR = "oj-message-error";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_WARNING = "oj-message-warning";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_INFO = "oj-message-info";

/**
 * @private
 * @const
 * @type {string}
 */
oj.PopupMessagingStrategyUtils._SELECTOR_MESSAGE_CONFIRMATION = "oj-message-confirmation";

/**
 * @ignore
 */
oj.PopupMessagingStrategyPoolUtils = {};

/**
 * @public
 * @returns {jQuery} popup taken or created from the free pool
 */
oj.PopupMessagingStrategyPoolUtils.getNextFreePopup = function ()
{
  var pool = oj.PopupMessagingStrategyPoolUtils._getPool();
  var popups = pool.find("." + oj.PopupMessagingStrategyPoolUtils._SELECTOR_MESSAGING);
  var popup;

  if (popups.length === 0)
  {
    popup = $(oj.PopupMessagingStrategyPoolUtils._getPopupContentHtml()).hide();
    // popup is an empty div
    popup.appendTo(pool); // @HTMLUpdateOk    
    var popupOptions =
    {
      'initialFocus': 'none',
      'tail': 'simple',
      'autoDismiss': 'none',
      'modality': 'modeless',
      'animation': {'open': null, 'close': null}
    };
    popup.ojPopup(popupOptions);
  }
  else
    popup = $(popups[0]);

  return popup;
};

/**
 * Passed in the root dom element of the message popup and returns the content element.
 *
 * @param {jQuery} popup root element
 * @returns {Element} content element of message popup
 */
oj.PopupMessagingStrategyPoolUtils.getPopupContentNode = function (popup)
{
  return popup.find("." + oj.PopupMessagingStrategyPoolUtils._SELECTOR_MESSAGING_CONTAINER)[0];
};

/**
 * @public
 */
oj.PopupMessagingStrategyPoolUtils.destroyFreePopup = function ()
{
  var popup;
  if (oj.PopupMessagingStrategyPoolUtils._getFreePoolCount() > 0)
  {
    // if the message popup is open, remove it.
    // if there is at least one popup in the pool, remove it.
    popup = oj.PopupMessagingStrategyPoolUtils.getNextFreePopup();
    popup.ojPopup("destroy");
    popup.remove();
  }
};

/**
 * Returns a div appended to the body that is a common pool of notewindow popups
 * used internally by editable value holders.
 *
 * @return {jQuery!} messaging popup pool container
 * @private
 */
oj.PopupMessagingStrategyPoolUtils._getPool = function ()
{
  /** @type {jQuery!} */
  var pool = $("#" + oj.PopupMessagingStrategyPoolUtils._MESSAGING_POPUP_POOL_ID);
  if (pool.length > 0)
    return pool;

  pool = $("<div>");
  pool.attr("id", oj.PopupMessagingStrategyPoolUtils._MESSAGING_POPUP_POOL_ID);
  pool.attr("role", "presentation");
  pool.appendTo($(document.body)); // @HTMLUpdateOk

  return pool;
};

/**
 * @return {number} number of unused popup in the pool
 * @private
 */
oj.PopupMessagingStrategyPoolUtils._getFreePoolCount = function ()
{
  var pool = oj.PopupMessagingStrategyPoolUtils._getPool();
  var popups = pool.find("." + oj.PopupMessagingStrategyPoolUtils._SELECTOR_MESSAGING);
  return popups.length;
};

/**
 * @return {string} messaging popup html
 * @private
 */
oj.PopupMessagingStrategyPoolUtils._getPopupContentHtml = function ()
{
  return '<div class="' + oj.PopupMessagingStrategyPoolUtils._SELECTOR_MESSAGING + '">' +
    '<div class="' + oj.PopupMessagingStrategyPoolUtils._SELECTOR_MESSAGING_CONTAINER + '"></div>' +
    '</div>';
};

/**
 * @const
 * @private
 * @type {string}
 */
oj.PopupMessagingStrategyPoolUtils._SELECTOR_MESSAGING_CONTAINER = "oj-messaging-popup-container";

/**
 * @const
 * @private
 * @type {string}
 */
oj.PopupMessagingStrategyPoolUtils._SELECTOR_MESSAGING = "oj-messaging-popup";

/**
 * @const
 * @private
 * @type {string}
 */
oj.PopupMessagingStrategyPoolUtils._MESSAGING_POPUP_POOL_ID = "__oj_messaging_popup_pool";

(function() {
var editableValueMeta = {
  "properties": {
    "describedBy": {
      "type": "string"
    },
    "disabled": {
      "type": "boolean"
    },
    "displayOptions": {
      "type": "Object",
      "properties": {
        "converterHint": {
          "type": "string"
        },
        "helpInstruction": {
          "type": "string"
        },
        "messages": {
          "type": "string"
        },
        "validatorHint": {
          "type": "string"
        }
      }
    },
    "help": {
      "type": "Object<string, string>",
      "properties": {
        "instruction": {
          "type": "string"
        }
      }
    }, 
    "helpHints": {
      "type": "Object",
      "properties": {
        "definition": {
          "type": "string"
         },
         "source": {
           "type": "string"
         }
       }
    },
    "labelHint": {
      "type": "string"
    },
    "messagesCustom": {
      "type": "Array",
      "writeback": true
    },
    "valid": {
      "type": "string",
      "writeback": true,
      "readOnly": true,
      "enumValues": ["valid", "invalidShown", "invalidHidden","pending"]
    },
    "value": {
      "type": "Object",
      "writeback": true
    }
  },
  "methods": {
    "reset": {},
    "showMessages": {}
  },
  "events": {
    "animateStart": {},
    "animateEnd": {}
  },
  "extension": {
    _WIDGET_NAME: "editableValue"
  }
};
oj.CustomElementBridge.registerMetadata('editableValue', 'baseComponent', editableValueMeta);
})();

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.inputBase
 * @augments oj.editableValue
 * @abstract
 * @ojsignature [{
 *                target: "Type",
 *                value: "abstract class inputBase<V, SP extends inputBaseSettableProperties<V, SV>, SV= V, RV= V> extends editableValue<V, SP, SV, RV>"
 *               },
 *               {
 *                target: "Type",
 *                value: "inputBaseSettableProperties<V, SV=V, RV= V> extends editableValueSettableProperties<V, SV, RV>",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @since 0.6
 * @ojshortdesc Abstract InputBase element
 * @ojrole input
 * @hideconstructor
 * @ojtsimport ojvalidation-base
 * @ojtsimport ojvalidation-datetime
 * @ojtsimport ojvalidation-number
 *
 * @classdesc
 * <h3 id="inputBaseOverview-section">
 *   Abstract inputBase component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputBaseOverview-section"></a>
 * </h3>
 * 
 * <p>Description: The inputBase component takes care of general needs of other input components [i.e. text + password]
 * 
 * @param {Object=} options a map of option-value pairs to set on the component
 */
oj.__registerWidget("oj.inputBase", $['oj']['editableValue'], 
{
  version : "1.0.0", 
  widgetEventPrefix : "oj",
  
  /**
   * Convenience Array which one can extend to set the attribute to a mandatory value if it doesn't exist or is set to something else 
   * [
   * {
   *    "attr"              : string - attribute associated to the task
   *    "setMandatory"      : value it must be set to [i.e. type="text"]
   * }
   * ]
   * 
   * Examples:
   * 1) [{"attr": "type", "setMandatory": "text"}]
   * 
   * @expose
   * @memberof! oj.inputBase
   * @private
   */
  _ATTR_CHECK : [],
  
  /** 
   * Class names to be applied to this.element()
   * 
   * @expose
   * @memberof! oj.inputBase
   * @private
   */
  _CLASS_NAMES : "",
  
  /** 
   * Class names to be applied to this.widget()
   * 
   * Note that if this value is defined then the this.element will be wrapped with a root dom element
   * 
   * @expose
   * @memberof! oj.inputBase
   * @private
   */
  _WIDGET_CLASS_NAMES : "",
  
  /** 
   * Class names to be applied to the dom that wraps the input and trigger elements. It is a child
   * of the root dom element - this.widget().
   * 
   * An element can be wrapped by a root dom element OR by both a root dom element and this extra
   * wrapper dom. the time and date pickers have two wrappers since we want to do an extra
   * wrapping when an input has trigger elements (date icon, for example). This is needed so we
   * can add extra dom to the root dom element for inline messaging.
   * 
   * @expose
   * @memberof! oj.inputBase
   * @private
   */
  _ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES : "",
  
  /** 
   * Array to be used for oj.EditableValueUtils.initializeOptionsFromDom attribute is the html-5 dom attribute name.
   * If option is different, like in the case of readonly (readonly html vs readOnly 
   * (camelcase) component option), specify both attribute and option.
   * NOTE: This is for the widget components only. Do not add custom element attributes here.
   * @expose
   * @memberof! oj.inputBase
   * @private
   */
  _GET_INIT_OPTIONS_PROPS_FOR_WIDGET: [{attribute: 'disabled', validateOption: true},
                            {attribute: 'pattern'},
                            {attribute: 'placeholder'},
                            {attribute: 'value'},
                            {attribute: 'readonly', option: 'readOnly', 
                             validateOption: true},
                            {attribute: 'required', coerceDomValue: true, validateOption: true},
                            {attribute: 'title'}],
  /**
   * If defined will append a div element containing text to be read out by Jaws when focus is placed on the input element 
   * and the value will be used to locate the translated text to be read out by Jaws.
   * 
   * Note the component must also be wrapped
   * 
   * @expose
   * @memberof! oj.inputBase
   * @private
   */
  _INPUT_HELPER_KEY: "",
  
  _BLUR_HANDLER_KEY: "blur",
  _KEYDOWN_HANDLER_KEY: "keydown",
  _KEYUP_HANDLER_KEY: "keyup",
  _INPUT_HANDLER_KEY: "input",
  _DROP_HANDLER_KEY: "drop",
  
  options : 
  {
    /**
     * <p>
     * When <code class="prettyprint">converter</code> property changes due to programmatic
     * intervention, the element performs various tasks based on the current state it is in. </br>
     *
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>Any cached converter instance is cleared and new converter created. The converter hint is
     * pushed to messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if element is valid when <code class="prettyprint">converter</code> property changes, the
     * display value is refreshed.</li>
     * <li>if element is invalid and is showing messages when
     * <code class="prettyprint">converter</code> property changes then all element messages are
     * cleared and full validation run using the current display value on the element.
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code>
     *   property is not updated, and the error is shown. 
     *   The display value is not refreshed in this case. </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code>
     *   property is updated; page author can listen to the <code class="prettyprint">onValueChanged</code>
     *   event to clear custom errors. The
     *   display value is refreshed with the formatted value provided by converter.</li>
     * </ul>
     * </li>
     * <li>if element is invalid and has deferred messages when converter property changes, the
     *   display value is again refreshed with the formatted value provided by converter.</li>
     * </ul>
     *
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the element are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.
     * Page authors can 
     * choose to clear it explicitly when setting the converter option.</li>
     * </ul>
     * </p>
     * @ojfragment inputBaseConverterOptionDoc
     * @memberof oj.inputBase
     **/
    /** 
     * Dictates component's autocomplete state. 
     * This attribute indicates whether the value of the control can be automatically 
     * completed by the browser.
     * 
     * @example <caption>Initialize component with <code class="prettyprint">autocomplete</code> attribute:</caption>
     * &lt;oj-some-element autocomplete="on">&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">autocomplete</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.autocomplete;
     *
     * // setter
     * myComp.autocomplete = "on";
     * 
     * @expose 
     * @type {string}
     * @ojvalue {string} "on" enable autofill
     * @ojvalue {string} "off" disable autofill
     * @alias autocomplete
     * @default "on"
     * @instance
     * @since 4.0.0
     * @memberof oj.inputBase
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    autocomplete: undefined,
     /** 
     * Autofocus is a Boolean that reflects the autofocus attribute, If it is set to true 
     * then the associated component  will get input focus when the page is loaded.
     * Setting this property doesn't set the focus to the component: 
     * it tells the browser to focus to it when the element is inserted in the document. 
     * 
     * @example <caption>Initialize component with <code class="prettyprint">autofocus</code> attribute:</caption>
     * &lt;oj-some-element autofocus>&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">autofocus</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.autofocus;
     *
     * // setter
     * myComp.autofocus = false;
     * 
     * @expose 
     * @type {boolean}
     * @alias autofocus
     * @default false
     * @instance
     * @since 4.0.0
     * @memberof oj.inputBase
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    autofocus: false,    
    /** 
     * It indicates the name of the component. 
     * 
     * @example <caption>Initialize component with <code class="prettyprint">name</code> attribute:</caption>
     * &lt;oj-some-element name="myName">&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">name</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.name;
     *
     * // setter
     * myComp.name = myName;
     * 
     * @expose 
     * @type {string}
     * @default ""
     * @alias name
     * @instance
     * @since 4.0.0
     * @memberof oj.inputBase
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    name: "",
    /**
     * The placeholder text to set on the element.
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">placeholder</code> attribute:</caption>
     * &lt;oj-some-element placeholder="User Name">&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">placeholder</code> property after initialization:</caption>
     * // getter
     * var myPh = myComp.placeholder;
     *
     * // setter
     * myComp.placeholder = myPlaceholder;
     * 
     * If the attribute is not set and if a converter is set then the
     * converter hint is used. See displayOptions for details.
     * 
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof! oj.inputBase
     * @type {string}
     */    
    placeholder: "",    
     /**
     * <p>The  <code class="prettyprint">rawValue</code> is the read-only property for retrieving 
     * the current value from the input field in string form. The main consumer of
     * <code class="prettyprint">rawValue</code> is a converter.</p>
     * <p>
     * The <code class="prettyprint">rawValue</code> updates on the 'input' javascript event, 
     * so the <code class="prettyprint">rawValue</code> changes as the value of the input is changed. 
     * If the user types in '1,200' into the field, the rawValue will be '1', then '1,', then '1,2', 
     * ..., and finally '1,200'. Then when the user blurs or presses 
     * Enter the <code class="prettyprint">value</code> property gets converted and validated
     * (if there is a converter or validators) and then gets updated if valid.
     * </p>
     * <p>This is a read-only attribute so page authors cannot set or change it directly.</p>
     * @expose 
     * @access public
     * @instance
     * @memberof! oj.inputBase
     * @type {string}
     * @ojsignature {target: "Type", value: "RV"}
     * @since 1.2.0
     * @readonly
     * @ojwriteback
     */
    rawValue: "",     
    /** 
     * Dictates component's readonly state. 
     * 
     * @example <caption>Initialize component with <code class="prettyprint">readonly</code> attribute:</caption>
     * &lt;oj-some-element readonly>&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">readonly</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.readonly;
     *
     * // setter
     * myComp.readonly = false;
     * 
     * @expose 
     * @type {boolean}
     * @alias readonly
     * @default false
     * @instance
     * @memberof! oj.inputBase
     */
    readOnly: false,
    /** 
     * Whether the component is required or optional. When required is set to true, an implicit 
     * required validator is created using the validator factory - 
     * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED).createValidator()</code>.
     * 
     * Translations specified using the <code class="prettyprint">translations.required</code> attribute 
     * and the label associated with the component, are passed through to the options parameter of the 
     * createValidator method. 
     * 
     * <p>
     * When <code class="prettyprint">required</code> property changes due to programmatic intervention, 
     * the component may clear messages and run validation, based on the current state it's in. </br>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when required is set to true, then it runs deferred validation on 
     * the value property. This is to ensure errors are not flagged unnecessarily.
     * </li>
     * <li>if component is invalid and has deferred messages when required is set to false, then 
     * component messages are cleared but no deferred validation is run.
     * </li>
     * <li>if component is invalid and currently showing invalid messages when required is set, then 
     * component messages are cleared and normal validation is run using the current display value. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   property is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code> 
     *   event on the component to clear custom errors.</li>
     * </ul>
     * </li>
     * </ul>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the component are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * 
     * </p>
     * 
     * @ojvalue {boolean} false - implies a value is not required to be provided by the user. 
     * This is the default.
     * @ojvalue {boolean} true - implies a value is required to be provided by user and the 
     * input's label will render a required icon. Additionally a required validator - 
     * {@link oj.RequiredValidator} - is implicitly used if no explicit required validator is set. 
     * An explicit required validator can be set by page authors using the validators attribute. 
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">required</code> attribute:</caption>
     * &lt;oj-some-element required>&lt;/oj-some-element><br/>
     * 
     * @example <caption>Customize messages and hints used by implicit required validator when 
     * <code class="prettyprint">required</code> attribute is set:</caption> 
     * &lt;oj-some-element required translations='{"required": {
     *                 "hint": "custom: enter at least 3 alphabets",
     *                 "messageSummary": "custom: \'{label}\' is Required", 
     *                 "messageDetail": "custom: please enter a valid value for \'{label}\'"}}'>
     * &lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
     * // getter
     * var rq = myComp.required;
     *
     * // setter
     * myComp.required = false;
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof oj.inputBase
     * @type {boolean}
     * @default false
     * @since 0.7
     * @see #translations
     */
    required: false,

    /** 
     * List of validators used by component when performing validation. Each item is either an 
     * instance that duck types {@link oj.Validator}, or is an Object literal containing the 
     * properties listed below. Implicit validators created by a component when certain attributes 
     * are present (e.g. <code class="prettyprint">required</code> attribute), are separate from 
     * validators specified through this attribute. At runtime when the component runs validation, it 
     * combines the implicit validators with the list specified through this attribute. 
     * <p>
     * Hints exposed by validators are shown in the notewindow by default, or as determined by the 
     * 'validatorHint' property set on the <code class="prettyprint">display-options</code> 
     * attribute. 
     * </p>
     * 
     * <p>
     * When <code class="prettyprint">validators</code> property changes due to programmatic 
     * intervention, the component may decide to clear messages and run validation, based on the 
     * current state it is in. </br>
     * 
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>The cached list of validator instances are cleared and new validator hints is pushed to 
     * messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when validators changes, component does nothing other than the 
     * steps it always performs.</li>
     * <li>if component is invalid and is showing messages when 
     * <code class="prettyprint">validators</code> changes then all component messages are cleared 
     * and full validation run using the display value on the component. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   property is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code> 
     *   event to clear custom errors.</li>
     * </ul>
     * </li>
     * <li>if component is invalid and has deferred messages when validators changes, it does 
     * nothing other than the steps it performs always.</li>
     * </ul>
     * </p>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the component are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * </p>
     * 
     * @property {string} type - the validator type that has a {@link oj.ValidatorFactory} that can 
     * be retrieved using the {@link oj.Validation} module. For a list of supported validators refer 
     * to {@link oj.ValidatorFactory}. <br/>
     * E.g., <code class="prettyprint">{validators: [{type: 'regExp'}]}</code>
     * @property {Object=} options - optional Object literal of options that the validator expects. 
     * <br/>
     * E.g., <code class="prettyprint">{validators: [{type: 'regExp', options: {pattern: '[a-zA-Z0-9]{3,}'}}]}</code>

     * 
     * @example <caption>Initialize the component with validator object literal:</caption>
     * &lt;oj-some-element validators='[{"type": "regExp", "options": {
     *                     "pattern": "[a-zA-Z0-9]{3,}", 
     *                     "messageDetail": "You must enter at least 3 letters or numbers"}}]'>
     * &lt;/oj-some-element>      
     * 
     * NOTE: oj.Validation.validatorFactory('numberRange') returns the validator factory that is used 
     * to instantiate a range validator for numbers.
     * 
     * @example <caption>Initialize the component with multiple validator instances:</caption>
     * var validator1 = new MyCustomValidator({'foo': 'A'}); 
     * var validator2 = new MyCustomValidator({'foo': 'B'});
     * var validators = [validator1, validator2]<br/>
     * ...
     * &lt;oj-some-element validators='[[validators]]'>
     * &lt;/oj-some-element>      
     * 
     * @example <caption>Get or set the <code class="prettyprint">validators</code> property after initialization:</caption>
     * // getter
     * var validators = myComp.validators;
     *
     * // setter
     * myComp.validators = myValidators;
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof oj.inputBase
     * @ojsignature  { target: "Type", value: "Array<oj.Validator<V>|oj.Validation.FactoryRegisteredValidatorOrConverter>|null"}
     * @type {Array|null}
     */    
    validators: undefined
  },
  
  /**
   * The base method needs to be overriden so that one can perform attribute check/set [i.e. ojInputText can only have type="text"] 
   * 
   * @protected
   * @override
   * @param {Object} element - jQuery selection to save attributes for
   * @instance
   * @memberof! oj.inputBase
   */
  _SaveAttributes : function (element)
  {
    var ret = this._superApply(arguments);
    
    this._processAttrCheck();
    
    return ret;
  },
  
  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.inputBase
   */
  _InitOptions: function(originalDefaults, constructorOptions)
  {
    this._super(originalDefaults, constructorOptions);
    if (!this._IsCustomElement())
    {
      oj.EditableValueUtils.initializeOptionsFromDom(
        this._GET_INIT_OPTIONS_PROPS_FOR_WIDGET, constructorOptions, this);
    }
  },

  /**
   * 1) Initializes the options
   * 2) If needed wraps the input element, 
   * 
   * @protected
   * @override
   * @instance
   * @memberof! oj.inputBase
   */
  _ComponentCreate : function()
  {
    var node = this.element, 
        ret = this._superApply(arguments),
        savedAttributes = this._GetSavedAttributes(node),
        readOnly = this.options['readOnly'];
    
    this._rtl = this._GetReadingDirection() === "rtl";
    
    // update element state using options
    if (typeof readOnly === "boolean") 
    {
      this.element.prop("readonly", readOnly);
    }
    
    if (this._DoWrapElement())
    {
      this._wrapElementInRootDomElement();
      // may need to do an extra wrapping if the element has triggers
      if (this._DoWrapElementAndTriggers())
      {
        this._WrapElement();
      }
      this._focusable({
        'element': this._wrapper, 
        'applyHighlight': true
      });
    }
    else
    {
      this._focusable({
        'element': this.element, 
        'applyHighlight': true
      });
    }

    
    // remove pattern attribute to not trigger html5 validation + inline bubble
//    if ('pattern' in savedAttributes)
//    {
//      node.removeAttr('pattern');
//    }
    
    this._defaultRegExpValidator = {};
    this._eventHandlers = null;
    return ret;
  },
  
  /**
   * 1) Updates component state based on the option values
   * 2) Adds the classname to the element [intentionally postponing till this process since the component might need to 
   *    reset this.element for some reason]
   * 3) Hooks up the blur handler
   * 4) If necessary appends an input helper to be read out by Jaws accessibility reader
   * 
   * @protected
   * @override
   * @instance
   * @memberof! oj.inputBase
   */        
  _AfterCreate : function () 
  {
    var ret = this._superApply(arguments),
        options = ["disabled", "readOnly"],
        self = this;
    var widgetId;
    
    this._refreshRequired(this.options['required']);
    
    if(this._CLASS_NAMES) 
    {
      this.element.addClass(this._CLASS_NAMES);
    }
    
    //attach handlers such as blur, keydown, and drop. placed in a function so to detach the handlers as well
    //when the options change
    this._attachDetachEventHandlers();
    
    this._AppendInputHelper();
    
    $.each(options, function(index, ele)
    {
      if(self.options[ele]) 
      {
        self._processDisabledReadOnly(ele, self.options[ele]);
      }
    });
    
    if (this._IsCustomElement())
    {
      // if it is a custom element, then set the sub-id on the input so if they have a oj-label
      // pointing to it with the 'for' attrbiute, JAWS will read the label.
      widgetId = this.widget().attr("id");
      if (widgetId)
        oj.EditableValueUtils.setSubIdForCustomLabelFor(this._GetContentElement()[0], widgetId);
    }
    return ret;
  },
  /**
   * Whether the component is required.
   * 
   * @return {boolean} true if required; false
   * 
   * @memberof! oj.inputBase
   * @instance
   * @protected
   * @override
   */
  _IsRequired : function () 
  {
    return this.options['required'];
  },
  /**
   * Performs post processing after required option is set by taking the following steps.
   * 
   * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
   * run full validation with UI value (we don't know if the UI error is from a required validator 
   * or something else);<br/>
   * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
   * updated<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to valueChanged to clear custom errors.<br/>
   * 
   * - if component is invalid and has messagesHidden -> required: false -> clear component 
   * errors; no deferred validation is run.<br/>
   * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
   * errors unnecessarily)<br/>
   * - messagesCustom is never cleared<br/>
   * 
   * @param {string} option
   * 
   * @memberof! oj.inputBase
   * @instance
   * @protected
   */
  _AfterSetOptionRequired : oj.EditableValueUtils._AfterSetOptionRequired,
  
  /**
   * When validators option changes, take the following steps.
   * 
   * - Clear the cached normalized list of all validator instances. push new hints to messaging.<br/>
   * - if component is valid -> validators changes -> no change<br/>
   * - if component is invalid has messagesShown -> validators changes -> clear all component 
   * messages and re-run full validation on displayValue. if there are no errors push value to 
   * model;<br/>
   * - if component is invalid has messagesHidden -> validators changes -> do nothing; doesn't change 
   * the required-ness of component <br/>
   * - messagesCustom is not cleared.<br/>
   * 
   * NOTE: The behavior applies to any option that creates implicit validators - min, max, pattern, 
   * etc. Components can call this method when these options change.
   * 
   * @memberof! oj.inputBase
   * @instance
   * @protected
   */
  _AfterSetOptionValidators : oj.EditableValueUtils._AfterSetOptionValidators,
  /**
   * Performs post processing after converter option changes by taking the following steps.
   * 
   * - always push new converter hint to messaging <br/>
   * - if component has no errors -> refresh UI value<br/>
   * - if component is invalid has messagesShown -> clear all component errors and run full 
   * validation using display value. <br/>
   * &nbsp;&nbsp;- if there are validation errors, value is not pushed to model; messagesShown is 
   * updated.<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to valueChanged to clear custom errors.<br/>
   * - if component is invalid has messagesHidden -> refresh UI value. no need to run deferred 
   * validations. <br/>
   * - messagesCustom is never cleared<br/>
   * 
   * @memberof! oj.inputBase
   * @instance
   * @protected
   */
  _AfterSetOptionConverter : oj.EditableValueUtils._AfterSetOptionConverter, 
  /**
   * Clears the cached converter stored in _converter and pushes new converter hint to messaging.
   * Called when convterer option changes 
   * @memberof! oj.inputBase
   * @instance
   * @protected 
   */
  _ResetConverter : oj.EditableValueUtils._ResetConverter,  
  /**
   * Returns the normalized converter instance.
   * 
   * @return {Object} a converter instance or null
   * @memberof! oj.inputBase
   * @instance
   * @protected 
   */
  _GetConverter : oj.EditableValueUtils._GetConverter,    
  /**
   * This returns an array of all validators 
   * normalized from the validators option set on the component. <br/>
   * @return {Array} of validators. 
   * @memberof! oj.inputBase
   * @instance
   * @protected
   */
  _GetNormalizedValidatorsFromOption : oj.EditableValueUtils._GetNormalizedValidatorsFromOption,
  
  _processDisabledReadOnly : function (key, value) 
  {
    if (key === "disabled")
    {
      this.element.prop("disabled", value);
    }
    
    if (key === "readOnly")
    {
      this.element.prop("readonly", value);
      this._refreshStateTheming("readOnly", value);
    }
    
    if(key === "disabled" || key === "readOnly") 
    {
      this._attachDetachEventHandlers();
    }
  },
  
  /**
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @override
   */
  _setOption : function (key, value, flags)
  {
    var retVal = this._superApply(arguments);
    
    if(key === "disabled" || key === "readOnly") 
    {
      this._processDisabledReadOnly(key, value);
    }

    if (key === "pattern")
    {
      this._defaultRegExpValidator[oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP] = this._getImplicitRegExpValidator();
      this._AfterSetOptionValidators();
    }
    

    
    return retVal;
  },
  /**
   * Performs post processing after _SetOption() is called. Different options when changed perform
   * different tasks. See _AfterSetOption[OptionName] method for details.
   *
   * @param {string} option
   * @param {Object|string=} previous
   * @param {Object=} flags
   * @protected
   * @memberof! oj.inputBase
   * @instance
   */
  _AfterSetOption : function (option, previous, flags)
  {
    this._superApply(arguments);
    switch (option)
    {         
      case "readOnly":
        this._AfterSetOptionDisabledReadOnly(option, oj.EditableValueUtils.readOnlyOptionOptions);
        break;
      case "required":
        this._AfterSetOptionRequired(option);
        break;
      case "validators":
        this._AfterSetOptionValidators(option);
        break;
      case "converter":
        this._AfterSetOptionConverter(option);
        break;           
      default:
        break;
    }
  },
  /**
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @override
   */
  _destroy : function ()
  {
    var ret = this._superApply(arguments);

    this.element.off("blur drop keydown keyup input");
    
    if(this._inputHelper) 
    {
      this._inputHelper.remove();
    }
    
    //unwrap only for non custom elements.
    // for custom elements, the wrapper is created by apps. So no need to replace it with input.
    if(this._DoWrapElement() && !this._IsCustomElement())
    {
      //  - DomUtils.unwrap() will avoid unwrapping if the node is being destroyed by Knockout
      if (this._DoWrapElementAndTriggers())
        oj.DomUtils.unwrap(this.element, this._wrapper);
      else
        oj.DomUtils.unwrap(this.element);
    }
 
    return ret;
  },
  
  _attachDetachEventHandlers: function () 
  {
    
    if(!this.options["readOnly"] && !this.options["disabled"]) {
      
      this._eventHandlers = {};
      
      var blurHandler = $.proxy(this._onBlurHandler, this),
          keyDownHandler = $.proxy(this._onKeyDownHandler, this),
          keyUpHandler = $.proxy(this._onKeyUpHandler, this),
          inputHandler = $.proxy(this._onInputHandler, this),
          dropHandler = function() 
                        {
                          this.focus();
                        };
      
      this.element.on(this._BLUR_HANDLER_KEY, blurHandler);
      this.element.on(this._KEYDOWN_HANDLER_KEY, keyDownHandler);
      this.element.on(this._KEYUP_HANDLER_KEY, keyUpHandler);
      this.element.on(this._INPUT_HANDLER_KEY, inputHandler);
      
      //other than FF when a drop is dispatched focus is placed back on the element
      //this would cause difference in behavior of the observable change [as set within blur], so in order to provide
      //consisteny placing the focus on the element after the drop
      this.element.on(this._DROP_HANDLER_KEY, dropHandler);
      
      this._eventHandlers[this._BLUR_HANDLER_KEY] = blurHandler;
      this._eventHandlers[this._KEYDOWN_HANDLER_KEY] = keyDownHandler;
      this._eventHandlers[this._KEYUP_HANDLER_KEY] = keyUpHandler;
      this._eventHandlers[this._INPUT_HANDLER_KEY] = inputHandler;
      this._eventHandlers[this._DROP_HANDLER_KEY] = dropHandler;
    }
    else 
    {
      
      //meaning either it is readOnly or is disabled, remove the handlers if they were attached previously
      if(this._eventHandlers) 
      {
        var eventEntries = [this._BLUR_HANDLER_KEY, this._KEYDOWN_HANDLER_KEY, this._KEYUP_HANDLER_KEY,
          this._INPUT_HANDLER_KEY, this._DROP_HANDLER_KEY];
        
        for(var i=0, j=eventEntries.length; i < j; i++) 
        {
          if(this._eventHandlers[eventEntries[i]]) 
          {
            this.element.off(eventEntries[i], this._eventHandlers[eventEntries[i]]);
            delete this._eventHandlers[eventEntries[i]];
          }
        }
      }
    }
    
  },
  
   /**
   * when below listed options are passed to the component, corresponding CSS will be toggled
   * @private
   * @memberof! oj.inputBase
   * @const
   * @type {Object}
   */
  _OPTION_TO_CSS_MAPPING: {
    "readOnly": "oj-read-only"
  }, 
  
  /**
   * Performs the attribute check/set by using _ATTR_CHECK variable [i.e. ojInputText must have type be set to "text"].
   * 
   * @private
   * @memberof! oj.inputBase
   */
  _processAttrCheck : function ()
  {
    
    var attrCheck = this._ATTR_CHECK;
    
    for(var i=0, j=attrCheck.length; i < j; i++) 
    {
      var attr = attrCheck[i]["attr"],
          setMandatoryExists = "setMandatory" in attrCheck[i];
      
      //if it doesn't exist just have to check whether one should set it to a mandatory value
      if(setMandatoryExists)
      {
        this.element.attr(attr, attrCheck[i]["setMandatory"]);
      }
    }
  },
  
  /**
   * Invoked when blur is triggered of the this.element
   * 
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @param {Event} event
   */
  _onBlurHandler : function (event) 
  {
    this._SetValue(this._GetDisplayValue(), event);
  },
  
  /**
   * Invoked when keydown is triggered of the this.element
   * 
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @param {Event} event
   */
  _onKeyDownHandler : function (event) 
  {
  },

  /**
   * Invoked when keydown is triggered of the this.element
   * 
   * When of keyCode is of Enter, invoke _SetValue on it
   * 
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @param {Event} event
   */
  _onKeyUpHandler : function (event) 
  {
    if(event.keyCode === $.ui.keyCode.ENTER) 
    {
      this._SetValue(this._GetDisplayValue(), event);
    }
  },
    
  /**
   * Invoked when the input event happens
   * 
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @param {Event} event
   */
  _onInputHandler : function (event) 
  {
    this._SetRawValue(this._GetContentElement().val(), event);
  },
  
  /**
   * Whether the this.element should be wrapped in a root dom element. 
   * Method so that additional conditions can be placed
   * 
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @return {boolean}
   */
  _DoWrapElement : function ()
  {
    return this._WIDGET_CLASS_NAMES;
  },

  /**
   * Whether the this.element and triggers should be wrapped. 
   * Method so that additional conditions can be placed
   * 
   * @ignore
   * @protected
   * @memberof! oj.inputBase
   * @return {boolean}
   */
  _DoWrapElementAndTriggers : function ()
  {
    return this._ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES;
  },
  
  /**
   * Wraps the this.element and adds _WIDGET_CLASS_NAMES classes to the wrapped element
   * 
   * @private
   * @memberof! oj.inputBase
   */
  _wrapElementInRootDomElement : function () 
  {
    if (this.OuterWrapper) {
      this._wrapper = $(this.OuterWrapper).addClass(this._WIDGET_CLASS_NAMES);
      this._wrapper.append(this.element); //@HTMLUpdateOK this is the original input element so ok, nothing special
    } else {
      $(this.element).wrap( $("<div>").addClass(this._WIDGET_CLASS_NAMES) ); //@HTMLUpdateOK original input element so ok, nothing special
      this._wrapper = this.element.parent();
    }
  },
          
  /**
   * Wraps the this.element and adds _ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES classes to the wrapped element.
   * We might need this extra wrapper if the component has input+triggers (like inputDate).
   * 
   * @protected
   * @ignore
   * @memberof! oj.inputBase
   * @return {jQuery}
   */
  _WrapElement : function () 
  {
    return $(this.element).wrap( $("<div>").addClass(this._ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES) ).parent(); //@HTMLUpdateOK original input element so ok, nothing special
  },
  
  /**
   * In some complex components [i.e. datepicker], when the input element receives focus we wish to have Jaws read 
   * out some content.
   * 
   * For those case does this method exist.
   *  
   * @protected
   * @instance
   * @memberof! oj.inputBase
   */
  _AppendInputHelper : function ()
  {
    if(this._INPUT_HELPER_KEY && this._DoWrapElement()) 
    {
      var describedBy = this.element.attr("aria-describedby") || "",
          helperDescribedById = this._GetSubId(this._INPUT_HELPER_KEY);
      
      describedBy += " " + helperDescribedById;
      this.element.attr("aria-describedby", describedBy);
      this._inputHelper = $("<div class='oj-helper-hidden-accessible' aria-hidden='true' id='" + helperDescribedById + "'>" + this._EscapeXSS(this.getTranslatedString(this._INPUT_HELPER_KEY)) + "</div>");
      
      this._AppendInputHelperParent().append(this._inputHelper);//@HTMLUpdateOK append action of the div element created with escaped translated text, so ok
    }
  },
  
  /** 
   * Helper function to escape Cross site script text
   * 
   * @param {string} escapeMe
   * @return {jQuery|string}
   * @memberof! oj.inputBase
   * @ignore
   */
  _EscapeXSS : function (escapeMe) 
  {
    return $("<span>" + escapeMe + "</span>").text();
  },
  
  /**
   * Which parent node the inputHelper should be appended to. Usually do not need to override.
   *  
   * @protected
   * @instance
   * @memberof! oj.inputBase
   */
  _AppendInputHelperParent : function () 
  {
    return this.widget();
  },
  
  /**
   * Sets up the default regExp validator.
   * 
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.inputBase
   */
  _GetImplicitValidators : function ()
  {
    var ret = this._superApply(arguments);

    // register a default RegExp validator if we have a valid pattern
    if (this.options['pattern'])
    {
      // add validator to the special internalValidators list. These are validators created by 
      // the framework. We don't want these cleared using the option - 'validators'
      this._defaultRegExpValidator[oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP] = this._getImplicitRegExpValidator();
    }
    
    return $.extend(this._defaultRegExpValidator, ret);
  },
  /**
   * Whether the a value can be set on the component. For example, if the component is 
   * disabled or readOnly then setting value on component is a no-op. 
   * 
   * @see #_SetValue
   * @return {boolean}
   * @memberof! oj.inputBase
   * @override
   * @instance
   * @protected
   */
    _CanSetValue: function ()
    {
      var readOnly;
      var superCanSetValue = this._super();

      if (!superCanSetValue)
        return false;

      readOnly = this.options['readOnly'] || false;
      return (readOnly) ? false : true;
    },
    /**
   * Toggles css selector on the widget. E.g., when readonly property changes, 
   * the oj-read-only selector needs to be toggled.
   * @param {string} option
   * @param {Object|string} value 
   * @memberof! oj.inputBase
   * @private
   */        
  _refreshStateTheming : function (option, value)
  {
    if (Object.keys(this._OPTION_TO_CSS_MAPPING).indexOf(option) != -1) 
    {
      // value is a boolean
      this.widget().toggleClass(this._OPTION_TO_CSS_MAPPING[option], !!value);
    }
  },
  
  /**
   * Returns the regexp validator instance or creates it if needed and caches it.
   * @private
   * @memberof! oj.inputBase
   */
  _getImplicitRegExpValidator : function ()
  {
    if (!this.options['pattern']) 
    {
      return null;
    }
    var regexpOptions = {'pattern': this.options['pattern'], 
                                   'label': this._getLabelText()};
    
    $.extend(regexpOptions, this.options['translations']['regexp'] || {});
    return oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP).createValidator(regexpOptions);
  },
  
  /**
   * This helper function will generate ids using widget's uuid as unique identifier for 
   * wai-aria and other necessary ids
   * 
   * @ignore
   * @protected
   * @param {string} sub
   * @return {string}
   * @memberof! oj.inputBase
   */
  _GetSubId : function (sub)
  {
    return this["uuid"] + "_" + sub;
  },
  
  /**
   * @ignore
   * @protected
   * @return {boolean}
   * @memberof! oj.inputBase
   */
  _IsRTL : function ()
  {
    return this._rtl;
  },
  
  /**
   * <p>Refreshes the component.  Usually called after dom changes have been made.
   *
   * @example <caption>Refresh component after dome changes have been made.</caption>
   * // refresh component. 
   * myComp.refresh();

   * @expose
   * @memberof oj.inputBase
   * @public
   * @return {void}
   * @instance
   */
  refresh: function()
  {
    var retVal = this._superApply(arguments);
    
    this._rtl = this._GetReadingDirection() === "rtl";
    
    this._refreshRequired(this.options['required']);
     
    return retVal;
  },
  /**
   * @memberof! oj.inputBase
   * @instance
   * @private
   */
  _refreshRequired : oj.EditableValueUtils._refreshRequired,
  /**
   * the validate method from v3.x that returns a boolean
   * @memberof! oj.inputBase
   * @instance
   * @protected
   * @ignore
   */
  _ValidateReturnBoolean: oj.EditableValueUtils._ValidateReturnBoolean,

  getNodeBySubId: function(locator)
  {
    return this._super(locator);
  },

    
  /**
   * Validates the component's display value using the converter and all validators registered on 
   * the component and updates the <code class="prettyprint">value</code> option by performing the 
   * following steps. 
   * 
   * <p>
   * <ol> 
   * <li>All messages are cleared, including custom messages added by the app. </li>
   * <li>If no converter is present then processing continues to next step. If a converter is 
   * present, the UI value is first converted (i.e., parsed). If there is a parse error then 
   * the messages are shown.</li>
   * <li>If there are no validators setup for the component the <code class="prettyprint">value</code> 
   * option is updated using the display value. Otherwise all 
   * validators are run in sequence using the parsed value from the previous step. The implicit 
   * required validator is run first if the component is marked required. When a validation error is 
   * encountered it is remembered and the next validator in the sequence is run. </li>
   * <li>At the end of validation if there are errors, the messages are shown. 
   * If there were no errors, then the 
   * <code class="prettyprint">value</code> option is updated.</li>
   * </ol>
   * 
   * @example <caption>Validate component using its current value.</caption>
   * // validate display value and shows messages if there are any to be shown.
   * myComp.validate();
   * @example <caption>Validate component and use the Promise's resolved state.</caption>
   * myComp.validate().then(
   *  function(result) {
   *    if(result === "valid")
   *    {
   *      submitForm();
   *    }
   *  });
   * @return {Promise.<string>} Promise resolves to "valid" if there were no converter parse errors and
   * the component passed all validations. 
   * The Promise resolves to "invalid" if there were converter parse errors or 
   * if there were validation errors.
   * @ojsignature {target: "Type", value: "Promise<'valid'|'invalid'>", for : "returns"}
   * 
   * 
   * @method
   * @access public
   * @expose
   * @memberof oj.inputBase
   * @instance
   * @since 4.0.0
   * @ojstatus preview
   */
  validate : oj.EditableValueUtils.validate,

  /**
   * Called to find out if aria-required is unsupported.
   * @memberof! oj.inputBase
   * @instance
   * @protected
   */
  _AriaRequiredUnsupported : function()
  {
    return false;
  },
  /**
   * Returns a <code class="prettyprint">jQuery</code> object containing the element visually 
   * representing the component, excluding the label associated with the it. 
   * 
   * <p>This method does not accept any arguments.</p>
   * 
   * @expose
   * @memberof! oj.inputBase
   * @instance
   * @return {jQuery} the root of the component
   * @ignore
   * 
   * @example <caption>Invoke the <code class="prettyprint">widget</code> method:</caption>
   * var widget = $( ".selector" ).ojFoo( "widget" ); // Foo is InputText, InputPassword, TextArea
   */       
  widget : function () 
  {
    return this._DoWrapElement() ? this._wrapper : this.element;
  }

}, true);

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojInputPassword
 * @augments oj.inputBase
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojInputPassword extends inputBase<string, ojInputPasswordSettableProperties>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojInputPasswordSettableProperties extends inputBaseSettableProperties<string>",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @since 0.6
 * @ojshortdesc Provides basic support for specifying a text value of type 'password'.
 * @ojrole textbox
 * @ojstatus preview
 *
 * @classdesc
 * <h3 id="inputPasswordOverview-section">
 *   JET InputPassword Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputPasswordOverview-section"></a>
 * </h3>
 *
 * <p>Description: The oj-input-password component enhances a browser input type="password" element.
 *
 * <pre class="prettyprint"><code>&lt;oj-input-password>&lt;/oj-input-password></code></pre>
 *
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <h3 id="label-section">
 *   Label and InputPassword
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate an oj-label to the oj-input-password component.
 * For accessibility, you should associate an oj-label element with the oj-input-password component
 * by putting an <code>id</code> on the oj-input-password element, and then setting the
 * <code>for</code> attribute on the oj-label to be the component's id.
 * </p>
 *
 * @example <caption>Initialize the oj-input-password element with no attributess specified:</caption>
 * &lt;oj-input-password>&lt;/oj-input-password>
 *
 * @example <caption>Initialize the oj-input-password element with some attributes:</caption>
 * &lt;oj-input-password id="pwdId" disabled>&lt;/oj-input-password>
 *
 * @example <caption>Initialize a component attribute via component binding:</caption>
 * &lt;oj-input-password id="pwdId" value="{{currentValue}}">&lt;/oj-input-password>
 */
oj.__registerWidget("oj.ojInputPassword", $['oj']['inputBase'],
{
  version : "1.0.0",
  defaultElement : "<input>",
  widgetEventPrefix : "oj",

  /**
   * @expose
   * @private
   * @memberof! oj.ojInputPassword
   */
  _ATTR_CHECK : [{"attr": "type", "setMandatory": "password"}],

  /**
   * @expose
   * @private
   * @memberof! oj.ojInputPassword
   */
  _CLASS_NAMES : "oj-inputpassword-input",

  /**
   * @expose
   * @private
   * @memberof! oj.ojInputPassword
   */
  _WIDGET_CLASS_NAMES : "oj-inputpassword oj-form-control oj-component",

  options :
  {
    /**
     * @expose 
     * @access public
     * @instance
     * @memberof! oj.ojInputPassword
     * @type {Object|undefined}
     * @ignore
     */    
    converter: undefined,
    /**
     * Regular expression pattern which will be used to validate the component's value. 
     * <p>
     * When pattern is set to true, an implicit regExp validator is created using the validator
     * factory -
     * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP).createValidator()</code>.
     * </p>
     *
     * @example <caption>Initialize the component with the <code class="prettyprint">pattern</code> attribute:</caption>
     * &lt;oj-input-password pattern="[a-zA-Z0-9]{3,}">&lt;/oj-input-password><br/>
     *
     * @example <caption>Get or set the <code class="prettyprint">pattern</code> property after initialization:</caption>
     * // getter
     * var pattern = myComp.pattern;
     *
     * // setter
     * myComp.pattern = "[0-9]{3,}";
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputPassword
     * @type {string|undefined}
     * @ignore
     */
    pattern: ""

    // Events

    /**
     * Triggered when the ojInputPassword is created.
     *
     * @event
     * @name create
     * @memberof oj.ojInputPassword
     * @instance
     * @property {Event} event event object
     * @property {Object} ui Currently empty
     * @ignore
     *
     * @example <caption>Initialize the ojInputPassword with the <code class="prettyprint">create</code> callback specified:</caption>
     * $( ".selector" ).ojInputPassword({
     *     "create": function( event, ui ) {}
     * });
     *
     * @example <caption>Bind an event listener to the <code class="prettyprint">ojcreate</code> event:</caption>
     * $( ".selector" ).on( "ojcreate", function( event, ui ) {} );
     */
    // create event declared in superclass, but we still want the above API doc
  },

  getNodeBySubId: function(locator)
  {
    var node = this._superApply(arguments), subId;
    if (!node)
    {
      subId = locator['subId'];
      if (subId === "oj-inputpassword-input") {
        node = this.element ? this.element[0] : null;
      }
    }
    // Non-null locators have to be handled by the component subclasses
    return node || null;
  },

  /**
   * @override
   * @instance
   * @memberof! oj.ojInputPassword
   * @protected
   * @return {string}
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-inputpassword";
  }

});

// Fragments:

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *    <tr>
 *       <td>Input</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Sets focus to input. If hints, help.instruction or messages exist in a notewindow,
 *       popup the notewindow.</td>
 *     </tr>
 *   </tbody>
 *  </table>
 *
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputPassword
 */

 /**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>Tab In</kbd></td>
 *       <td>Set focus to the input. 
 *       If hints, help.instruction or messages exist in a notewindow, 
 *        pop up the notewindow.</td>
 *     </tr> 
 *   </tbody>
 * </table>
 *
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputPassword
 */


//////////////////     SUB-IDS     //////////////////
/**
 * <p>Sub-ID for the ojInputPassword component's input element.</p>
 *
 * @ojsubid oj-inputpassword-input
 * @ignore
 * @memberof oj.ojInputPassword
 * @deprecated 4.0.0 Since the application supplies this element, it can supply a unique ID by which the element can be accessed.
 *
 * @example <caption>Get the node for the input element:</caption>
 * var node = myComp.getNodeBySubId("oj-inputpassword-input");
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojInputText
 * @augments oj.inputBase
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojInputText extends inputBase<any, ojInputTextSettableProperties>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojInputTextSettableProperties extends inputBaseSettableProperties<any>",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @since 0.6
 * @ojshortdesc Provides basic support for specifying a text value.
 * @ojrole textbox
 * @ojstatus preview
 *
 * @classdesc
 * <h3 id="inputTextOverview-section">
 *   JET InputText Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputTextOverview-section"></a>
 * </h3>
 *
 * <p>Description: The oj-input-text component enhances a browser input type="text" element.
 * 
 * <pre class="prettyprint"><code>&lt;oj-input-text>&lt;/oj-input-text></code></pre>
 *
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="label-section">
 *   Label and InputText
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate an oj-label with the oj-input-text component.
 * For accessibility, you should associate an oj-label element to the oj-input-text component
 * by putting an <code>id</code> on the oj-input-text element, and then setting the
 * <code>for</code> attribute on the oj-label to be the component's id.
 * </p>
 *
 * @example <caption>Declare the oj-input-text component with no attributes specified:</caption>
 * &lt;oj-input-text>&lt;/oj-input-text>
 *
 * @example <caption>Initialize the component with some attributes:</caption>
 * &lt;oj-input-text id="textId" disabled>&lt;/oj-input-text>
 *
 * @example <caption>Initialize a component attribute via component binding:</caption>
 * &lt;oj-input-text id="textId" value="{{currentValue}}">&lt;/oj-input-text>
 */
oj.__registerWidget("oj.ojInputText", $['oj']['inputBase'],
{
  version : "1.0.0",
  defaultElement : "<input>",
  widgetEventPrefix : "oj",

  /**
   * @expose
   * @private
   */
  _ATTR_CHECK : [{"attr": "type", "setMandatory": "text"}],

  /**
   * @expose
   * @private
   */
  _CLASS_NAMES : "oj-inputtext-input",

  /**
   * @expose
   * @private
   */
  _WIDGET_CLASS_NAMES : "oj-inputtext oj-form-control oj-component",
  
  /**
   * @private
   */
  _ALLOWED_TYPES : ['email', 'number', 'search', 'tel', 'text', 'url'],

  _CLICK_HANDLER_KEY: "click",

  options :
  {
    /**
     * @expose
     * @memberof! oj.ojInputText
     * @instance
     * @type {string}
     * @ojvalue {string} "never" The clear icon is never visible
     * @ojvalue {string} "always" The clear icon is always visible
     * @ojvalue {string} "conditional" The clear icon is visible under the following conditions:
     * if the component has a non-empty value AND it either has focus OR the mouse is over the field.
     * @default "never"
     * @desc Specifies if the clear icon should be visible.
     *
     * @example <caption>Initialize the oj-input-text with the <code class="prettyprint">clear-icon</code> attribute specified:</caption>
     * &lt;oj-input-text clear-icon="conditional" id="inputcontrol">&lt;/oj-input-text>
     *
     * @example <caption>Get or set the <code class="prettyprint">clearIcon</code> property after initialization:</caption>
     * // getter
     * var clearIcon = myInputText.clearIcon;
     *
     * // setter
     * myInputText.clearIcon = 'conditional';
     */
    clearIcon: "never",
    /**
     * a converter instance that duck types {@link oj.Converter}. Or an object literal containing 
     * the following properties.
     * {@ojinclude "name":"inputBaseConverterOptionDoc"} 
     * 
     * @property {string} type - the converter type registered with the oj.ConverterFactory. 
     * Supported types are 'number' and 'datetime'. See {@link oj.ConverterFactory} for details. 
     * <br/>
     * E.g., <code class="prettyprint">converter='{"type": "number"}'</code>
     * @property {Object=} options - optional Object literal of options that the converter expects. 
     * See {@link oj.IntlNumberConverter} for options supported by the number converter. See 
     * {@link oj.IntlDateTimeConverter} for options supported by the date time converter. <br/>
     * E.g., <code class="prettyprint">converter='{"type": "number", "options": {"style": "decimal"}}'</code>
     * 
     * @example <caption>Initialize the component with a number converter instance:</caption>
     * &lt;oj-input-text converter="[[salaryConverter]]">&lt;/oj-input-text><br/>
     * // Initialize converter instance using currency options
     * var options = {style: 'currency', 'currency': 'USD', maximumFractionDigits: 0};
     * var numberConverterFactory = oj.Validation.converterFactory("number");
     * var salaryConverter = numberConverterFactory.createConverter(options);
     * 
     * @example <caption>Initialize the component with converter object literal:</caption>
     * &lt;oj-input-text converter='{"type": "number", "options": {"style": "decimal"}}'>&lt;/oj-input-text>
     * 
     * @example <caption>Get or set the <code class="prettyprint">converter</code>
     *  property after initialization:</caption>
     * // getter
     * var converter = myComp.converter;
     *
     * // setter
     * myComp.converter = myConverter;
     * @expose 
     * @access public
     * @instance
     * @memberof! oj.ojInputText
     * @ojsignature {
     *    target: "Type",
     *    value: "oj.Converter<any>|oj.Validation.FactoryRegisteredValidatorOrConverter|null"}
     * @type {Object|undefined}
     */    
    converter: undefined,
    /** 
     * Indicates a list of pre-defined options to suggest to the user. 
     * The value must be the id of a &lt;datalist> element in the same page.
     * This attribute is ignored when the type attribute's value is hidden. 
     * 
     * @example <caption>Initialize component with <code class="prettyprint">list</code> attribute:</caption>
     * &lt;oj-some-element list="sampleDataList">&lt;/oj-some-element>
     * 
     * <p>Example for datalist:
     * <pre class="prettyprint">
     * <code>
     * &lt;datalist id="sampleDataList">
     *   &lt;option value="item 1">item 1&lt;/option>
     *   &lt;option value="item 2">item 2&lt;/option>
     *   &lt;option value="item 3">item 3&lt;/option>
     *   &lt;option value="item 4">item 4&lt;/option>
     * &lt;/datalist> 
     * </code></pre>
     *
     * @example <caption>Get or set the <code class="prettyprint">list</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.list;
     *
     * // setter
     * myComp.list = "myDataList";
     * 
     * @expose
     * @instance
     * @memberof! oj.ojInputText
     * @type {string}
     * @public
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    list: "",
    /**
     * Regular expression pattern which will be used to validate the component's value.
     * <p>
     * When pattern is set to a non-empty string value, an implicit regExp validator is created using the validator
     * factory -
     * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP).createValidator()</code>.
     * </p>
     *
     * @example <caption>Initialize the component with the <code class="prettyprint">pattern</code> attribute:</caption>
     * &lt;oj-input-text pattern="[a-zA-Z0-9]{3,}">&lt;/oj-input-text><br/>
     * 
     * @example <caption>Get or set the <code class="prettyprint">pattern</code> property after initialization:</caption>
     * // getter
     * var pattern = myComp.pattern;
     *
     * // setter
     * myComp.pattern = "[0-9]{3,}";
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputText
     * @type {string|undefined}
     * @ignore
     */
    pattern: "",
    /** 
     * Setting the value of this attribute to true indicates that the element needs to have its 
     * spelling and grammar checked. The value default indicates that the element is to act according
     * to a default behavior, possibly based on the parent element's own spellcheck value. 
     * The value false indicates that the element should not be checked.
     * 
     * @example <caption>Initialize component with <code class="prettyprint">spellcheck</code> attribute:</caption>
     * &lt;oj-some-element spellcheck>&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">spellcheck</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.spellcheck;
     *
     * // setter
     * myComp.spellcheck = false;
     * 
     * @expose
     * @instance
     * @memberof! oj.ojInputText
     * @default false
     * @type {boolean}
     * @public
     * @ojextension {_ATTRIBUTE_ONLY: true, _COPY_TO_INNER_ELEM: true}
     */
    spellcheck: false,
    /**
     * The type of virtual keyboard to display for entering value on mobile browsers.  This attribute has no effect on desktop browsers.
     *
     * @example <caption>Initialize the component with the <code class="prettyprint">virtual-keyboard</code> attribute:</caption>
     * &lt;oj-input-text virtual-keyboard="number">&lt;/oj-input-text>
     * 
     * @example <caption>Get or set the <code class="prettyprint">virtualKeyboard</code> property after initialization:</caption>
     * // Getter
     * var virtualKeyboard = myComp.virtualKeyboard;
     * 
     * // Setter
     * myComp.virtualKeyboard = "number";
     *
     * @expose
     * @instance
     * @memberof oj.ojInputText
     * @type {string}
     * @ojvalue {string} "auto" The component will determine the best virtual keyboard to use.
     *                          <p>This is always "text" for this release but may change in future
     *                          releases.</p>
     * @ojvalue {string} "email" Use a virtual keyboard for entering email.
     * @ojvalue {string} "number" Use a virtual keyboard for entering number.
     *                            <p>Note that on Android and Windows Mobile, the "number" keyboard does
     *                            not contain the minus sign.  This value should not be used on fields that
     *                            accept negative values.</p>
     * @ojvalue {string} "search" Use a virtual keyboard for entering search terms.
     * @ojvalue {string} "tel" Use a virtual keyboard for entering telephone number.
     * @ojvalue {string} "text" Use a virtual keyboard for entering text.
     * @ojvalue {string} "url" Use a virtual keyboard for entering URL.
     * @default "auto"
     * @since 5.0.0
     */
    virtualKeyboard: "auto"

    // Events

    /**
     * Triggered when the ojInputText is created.
     *
     * @event
     * @name create
     * @memberof oj.ojInputText
     * @instance
     * @property {Event} event event object
     * @property {Object} ui Currently empty
     * @ignore
     *
     * @example <caption>Initialize the ojInputText with the <code class="prettyprint">create</code> callback specified:</caption>
     * $( ".selector" ).ojInputText({
     *     "create": function( event, ui ) {}
     * });
     *
     * @example <caption>Bind an event listener to the <code class="prettyprint">ojcreate</code> event:</caption>
     * $( ".selector" ).on( "ojcreate", function( event, ui ) {} );
     */
    // create event declared in superclass, but we still want the above API doc
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputText
   */
  _ComponentCreate : function ()
  {
    var retVal = this._super();

    // add in the clear icon if needed
    var clearIconAttr = this.options["clearIcon"];
    
    this._processClearIconAttr(clearIconAttr);
    
    this._AddHoverable(this._wrapper);

    // Set the input type attribute based on virtualKeyboard property
    this._SetInputType(this._ALLOWED_TYPES);

    return retVal;
  },
  /**
   * Render or remove the clear icon
   * @ignore
   * @private
   * @param {string} clearIconAttr
   * @memberof oj.ojInputText
   * @instance
   */
  _processClearIconAttr : function (clearIconAttr)
  {
    var wrapperElem = this._wrapper[0];
    var clearIconBtn = wrapperElem.querySelector("a.oj-inputtext-clear-icon-btn");

    if (clearIconAttr === "never" || this.options["disabled"] || this.options["readOnly"])
    {
      // remove the icon if it is there
      if (clearIconBtn)
      {
        wrapperElem.removeChild(clearIconBtn);
      }
      
      // if the clearIcon is not rendered, we shouldn't have these classes
      wrapperElem.classList.remove('oj-inputtext-clearicon-visible');
      wrapperElem.classList.remove("oj-inputtext-clearicon-conditional");
    }
    else
    {
      if (clearIconBtn === null)
      {
        var clearIcon;
        clearIconBtn = document.createElement('a');
        clearIconBtn.className = "oj-inputtext-clear-icon-btn oj-component-icon oj-clickable-icon-nocontext";
        clearIconBtn.setAttribute("tabindex", "-1");
        clearIconBtn.setAttribute("target", "_blank");
        clearIcon = document.createElement('span');
        clearIcon.className =
                            "oj-inputtext-clear-icon";
   
        clearIconBtn.appendChild(clearIcon);
        wrapperElem.appendChild(clearIconBtn);

        clearIconBtn.addEventListener(this._CLICK_HANDLER_KEY, this._onClearIconClickHandler.bind(this));
      }
      
      // If clear-icon = "always", we want the icon visible all the time
      if (clearIconAttr === "always")
        wrapperElem.classList.add('oj-inputtext-clearicon-visible');
      else
      {
        wrapperElem.classList.remove('oj-inputtext-clearicon-visible');
      
        // for the conditional case, we add oj-inputtext-clearicon-hidden
        if (clearIconAttr === "conditional")
        {
          // For the conditional case, we render oj-form-control-empty-clearicon if the input doesn't
          // have a value, as we always want the clear icon hidden for this case.  When the input does
          // have a value, then we use oj-inputtext-clearicon-conditional, which has selectors for
          // oj-hover and oj-focus to determine when the icon is visible.
          wrapperElem.classList.add("oj-inputtext-clearicon-conditional");
          
          var val;
          
          // if the component is not fully rendered, then we need to use the value option's value
          // instead of the input element's value
          if (wrapperElem.classList.contains("oj-complete"))
            val = this.element[0].value;
          else
            val = this.options["value"];
          
          if (val !== "")
          {
            wrapperElem.classList.remove("oj-form-control-empty-clearicon");
          }
          else
          {
            wrapperElem.classList.add("oj-form-control-empty-clearicon");
          }
        }
        else
        {
          wrapperElem.classList.remove("oj-inputtext-clearicon-conditional");
          wrapperElem.classList.remove("oj-form-control-empty-clearicon");
        }
      }
    }
  },
  /**
   * Performs post processing after _SetOption() calls _superApply(). Different options, when changed, perform
   * different tasks. 
   * 
   * @param {string} option
   * @param {Object=} flags 
   * @protected
   * @memberof oj.ojInputText
   * @instance
   */
  _AfterSetOption : function (option, flags)
  {
    this._super(option, flags);
    
    switch (option)
    {
      // all of these options potentially affect the visiblity and/or rendering of the icon
      // so we need to process the icon if any of these change.
      case "clearIcon" :
      case "value" :
      case "disabled" :
      case "readOnly" :
        this._processClearIconAttr(this.options["clearIcon"]);
        break;
      
      case "virtualKeyboard":
        this._SetInputType(this._ALLOWED_TYPES);
        break;

      default:
        break;
    }
  },
  /**
   * Invoked when the input event happens
   * 
   * @ignore
   * @protected
   * @memberof! oj.ojInputText
   * @param {Event} event
   */
  _onInputHandler : function (event) 
  {
    this._super(event);
    var inputNode = event.target;
    var wrapperNode = this._wrapper[0];
    var clearIconAttr = this.options["clearIcon"];
    
    if (clearIconAttr === "conditional")
    {
      if (inputNode.value !== "")
      {
        wrapperNode.classList.remove("oj-form-control-empty-clearicon");
      }
      else
      {
        wrapperNode.classList.add("oj-form-control-empty-clearicon");
      }
    }
  },  
  /**
   * The handler clears the value of the input element and sets the focus on the element
   * NOTE: this handler expects the input text component to be bound to this via .bind()
   * 
   * @ignore
   * @private
   * @memberof! oj.ojInputText
   * @param {Event} event
   */
  _onClearIconClickHandler : function (event)
  {
    var elem = this.element[0];

    elem.value = "";
    // we need to update the raw value to keep it in sync
    this._SetRawValue(elem.value, event);
    elem.focus();
    this._wrapper[0].classList.add("oj-form-control-empty-clearicon");
  },
  /**
   * Invoked when blur is triggered of the this.element
   * We don't want to set the value if the event.relatedTarget is the clear icon button
   * 
   * @ignore
   * @protected
   * @memberof! oj.inputText
   * @param {Event} event
   */
  _onBlurHandler : function (event) 
  {
    var wrapperNode = this._wrapper[0];
    var target = event.relatedTarget;

    // if this is the clear icon, skip the blur handler if it is an ancestor of the input text
    if (!(target && 
          target.classList.contains("oj-inputtext-clear-icon-btn") &&
          target.parentElement &&
          oj.DomUtils.isAncestorOrSelf(wrapperNode, target.parentElement)))
    {
      this._super(event);
    }
    else
    {
      // We need to put the oj-focus back on the wrapperNode so that the icon
      // doesn't disappear on iOS and make it so that the click handler will fire.
      wrapperNode.classList.add("oj-focus");
    }
  },
  getNodeBySubId: function(locator)
  {
    var node = this._superApply(arguments), subId;
    if (!node)
    {
      subId = locator['subId'];
      if (subId === "oj-inputtext-input") {
        node = this.element ? this.element[0] : null;
      }
    }
    // Non-null locators have to be handled by the component subclasses
    return node || null;
  },

  /**
   * @override
   * @instance
   * @memberof! oj.ojInputText
   * @protected
   * @return {string}
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-inputtext";
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputText
   */
  _GetTranslationsSectionName: function()
  {
    return "oj-inputBase";
  },
  
  /**
   * Set the type of the input element based on virtualKeyboard option.
   * @memberof oj.ojInputText
   * @instance
   * @protected
   * @ignore
   */
  _SetInputType: oj.EditableValueUtils._SetInputType

});

  // Fragments:

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Gesture</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *    <tr>
   *       <td>Input</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Sets focus to input. If hints, help.instruction or messages exist in a notewindow,
   *       popup the notewindow.</td>
   *     </tr>
   *   </tbody>
   *  </table>
   *
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojInputText
   */

   /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>Input</td>
   *       <td><kbd>Tab In</kbd></td>
   *       <td>Set focus to the input. 
   *       If hints, help.instruction or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr> 
   *   </tbody>
   * </table>
   *
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojInputText
   */
  /**
   * {@ojinclude "name":"ojStylingDocIntro"}
   * 
   * <table class="generic-table styling-table">
   *   <thead>
   *     <tr>
   *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
   *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>oj-form-control-text-align-right</td>
   *       <td>Aligns the text to the right regardless of the reading direction,
               this is normally used for right aligning numbers 
   *     </td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
   * @memberof oj.ojInputText
   */

//////////////////     SUB-IDS     //////////////////
/**
 * <p>Sub-ID for the ojInputText component's input element.</p>
 * 
 * @ojsubid oj-inputtext-input
 * @memberof oj.ojInputText
 *
 * @example <caption>Get the node for the input element:</caption>
 * var node = myComp.getNodeBySubId("oj-input-text-input");
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojTextArea
 * @augments oj.inputBase
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojTextArea extends inputBase<any, ojTextAreaSettableProperties>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojTextAreaSettableProperties extends inputBaseSettableProperties<any>",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @since 0.6
 * @ojshortdesc Provides basic support for specifying a multi-line text value.
 * @ojrole textbox
 * @ojstatus preview
 *
 * @classdesc
 * <h3 id="textAreaOverview-section">
 *   JET TextArea Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#textAreaOverview-section"></a>
 * </h3>
 *
 * <p>Description: The oj-text-area component enhances a browser textarea element.
 *
 * <pre class="prettyprint"><code>&lt;oj-text-area>&lt;/oj-text-area></code></pre>
 *
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <h3 id="label-section">
 *   Label and TextArea
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate an oj-label to the oj-text-area component.
 * For accessibility, you should associate an oj-label element with the oj-text-area component
 * by putting an <code>id</code> on the oj-text-area element, and then setting the
 * <code>for</code> attribute on the oj-label to be the component's id.
 * </p>
 * <h3 id="binding-section">
 *   Declarative Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#binding-section"></a>
 * </h3>
 *
 * @example <caption>Initialize the textarea with no attributes specified:</caption>
 * &lt;oj-text-area>&lt;/oj-text-area>
 *
 * * @example <caption>Initialize the textarea with some attributes:</caption>
 * &lt;oj-text-area id="txtAreaId" disabled>&lt;/oj-text-area>
 *
 * @example <caption>Initialize the textarea via the JET component binding:</caption>
 * &lt;oj-text-area id="txtAreaId" value="{{currentValue}}">&lt;/oj-text-area>
 */
oj.__registerWidget("oj.ojTextArea", $['oj']['inputBase'],
{
  version : "1.0.0",
  defaultElement : "<textarea>",
  widgetEventPrefix : "oj",

  /**
   * @expose
   * @private
   * @memberof! oj.ojTextArea
   */
  _ATTR_CHECK : [],

  /**
   * @expose
   * @private
   * @memberof! oj.ojTextArea
   */
  _CLASS_NAMES : "oj-textarea-input",

  /**
   * @expose
   * @private
   * @memberof! oj.ojTextArea
   */
  _WIDGET_CLASS_NAMES : "oj-textarea oj-form-control oj-component",

  options :
  {
    /**
     * a converter instance that duck types {@link oj.Converter}. Or an object literal containing 
     * the following properties.
     * {@ojinclude "name":"inputBaseConverterOptionDoc"} 
     * 
     * @property {string} type - the converter type registered with the oj.ConverterFactory. 
     * Supported types are 'number' and 'datetime'. See {@link oj.ConverterFactory} for details. 
     * <br/>
     * E.g., <code class="prettyprint">converter='{"type": "number"}'</code>
     * @property {Object=} options - optional Object literal of options that the converter expects. 
     * See {@link oj.IntlNumberConverter} for options supported by the number converter. See 
     * {@link oj.IntlDateTimeConverter} for options supported by the date time converter. <br/>
     * E.g., <code class="prettyprint">converter='{"type": "number", "options": {"style": "decimal"}}'</code>
     * 
     * @example <caption>Initialize the component with a number converter instance:</caption>
     * &lt;oj-text-area converter="[[salaryConverter]]">&lt;/oj-text-area><br/>
     * // Initialize converter instance using currency options
     * var options = {style: 'currency', 'currency': 'USD', maximumFractionDigits: 0};
     * var numberConverterFactory = oj.Validation.converterFactory("number");
     * var salaryConverter = numberConverterFactory.createConverter(options);
     * 
     * @example <caption>Initialize the component with converter object literal:</caption>
     * &lt;oj-text-area converter='{"type": "number", "options": {"style": "decimal"}}'>&lt;/oj-text-area>
     * 
     * @example <caption>Get or set the <code class="prettyprint">converter</code> 
     * property after initialization:</caption>
     * // getter
     * var converter = myComp.converter;
     *
     * // setter
     * myComp.converter = myConverter;
     * @expose 
     * @access public
     * @instance
     * @memberof! oj.ojTextArea
     * @ojsignature {
     *    target: "Type",
     *    value: "oj.Converter<any>|oj.Validation.FactoryRegisteredValidatorOrConverter|null"}
     * @type {Object|undefined}
     */    
    converter: undefined,
    /**
     * Regular expression pattern which will be used to validate the component's value.
     * <p>
     * When pattern is set to true, an implicit regExp validator is created using the validator
     * factory -
     * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP).createValidator()</code>.
     * </p>
     *
     *
     * @example <caption>Initialize the component with the <code class="prettyprint">pattern</code> attribute:</caption>
     * &lt;oj-text-area pattern="[a-zA-Z0-9]{3,}">&lt;/oj-text-area><br/>
     * 
     * @example <caption>Get or set the <code class="prettyprint">pattern</code> property after initialization:</caption>
     * // getter
     * var pattern = myComp.pattern;
     *
     * // setter
     * myComp.pattern = "[0-9]{3,}";
     *
     * @expose
     * @instance
     * @memberof! oj.ojTextArea
     * @type {string|undefined}
     * @ignore
     */
    pattern: "",
     /** 
     * The number of visible text lines in the textarea. It can also be used to 
     * give specific height to the textarea.
     * 
     * @example <caption>Initialize component with <code class="prettyprint">rows</code> attribute:</caption>
     * &lt;oj-some-element rows="5">&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">rows</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.rows;
     *
     * // setter
     * myComp.rows = 5;
     * 
     * @expose
     * @instance
     * @memberof! oj.ojTextArea
     * @type {number}
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    rows: undefined,
     /** 
     * Setting the value of this attribute to true indicates that the element needs to have its 
     * spelling and grammar checked. The value default indicates that the element is to act according
     * to a default behavior, possibly based on the parent element's own spellcheck value. 
     * The value false indicates that the element should not be checked.
     * 
     * @example <caption>Initialize component with <code class="prettyprint">spellcheck</code> attribute:</caption>
     * &lt;oj-some-element spellcheck>&lt;/oj-some-element>
     * 
     * @example <caption>Get or set the <code class="prettyprint">spellcheck</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.spellcheck;
     *
     * // setter
     * myComp.spellcheck = false;
     * 
     * @expose
     * @instance
     * @memberof! oj.ojTextArea
     * @default false
     * @type {boolean}
     * @ojextension {_ATTRIBUTE_ONLY: true, _COPY_TO_INNER_ELEM: true}
     */
    spellcheck: false

    // Events

    /**
     * Triggered when the ojTextArea is created.
     *
     * @event
     * @name create
     * @memberof oj.ojTextArea
     * @instance
     * @property {Event} event event object
     * @property {Object} ui Currently empty
     * @ignore
     *
     * @example <caption>Initialize the ojTextArea with the <code class="prettyprint">create</code> callback specified:</caption>
     * $( ".selector" ).ojTextArea({
     *     "create": function( event, ui ) {}
     * });
     *
     * @example <caption>Bind an event listener to the <code class="prettyprint">ojcreate</code> event:</caption>
     * $( ".selector" ).on( "ojcreate", function( event, ui ) {} );
     */
    // create event declared in superclass, but we still want the above API doc
  },

  getNodeBySubId: function(locator)
  {
    var node = this._superApply(arguments), subId;
    if (!node)
    {
      subId = locator['subId'];
      if (subId === "oj-textarea-input") {
        node = this.element ? this.element[0] : null;
      }
    }
    // Non-null locators have to be handled by the component subclasses
    return node || null;
  },

  /**
   * @instance
   * @memberof! oj.ojTextArea
   * @override
   * @protected
   * @return {string}
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-textarea";
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojTextArea
   */
  _GetTranslationsSectionName: function()
  {
    return "oj-inputBase";
  }

});

  // Fragments:

	/**
	 * <table class="keyboard-table">
	 *   <thead>
	 *     <tr>
	 *       <th>Target</th>
	 *       <th>Gesture</th>
	 *       <th>Action</th>
	 *     </tr>
	 *   </thead>
	 *   <tbody>
	 *    <tr>
	 *       <td>TextArea</td>
	 *       <td><kbd>Tap</kbd></td>
	 *       <td>Sets focus to textarea. If hints, help.instruction or messages exist in a notewindow,
   *       popup the notewindow.</td>
	 *     </tr>
	 *   </tbody>
	 *  </table>
	 *
	 *
	 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
	 * @memberof oj.ojTextArea
	 */

   /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>TextArea</td>
   *       <td><kbd>Tab In</kbd></td>
   *       <td>Set focus to the textarea. 
   *       If hints, help.instruction or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr> 
   *   </tbody>
   * </table>
   *
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojTextArea
   */


//////////////////     SUB-IDS     //////////////////
/**
 * <p>Sub-ID for the ojTextArea component's textarea element.</p>
 * 
 * @ojsubid oj-textarea-input
 * @ignore
 * @memberof oj.ojTextArea
 * @deprecated 4.0.0  Since the application supplies this element, it can supply a unique ID by which the element can be accessed.
 *
 * @example <caption>Get the node for the input element:</caption>
 * var node = myComp.getNodeBySubId("oj-textarea-input");
 */

(function() {
var inputBaseMeta = {
  "properties": {
    "autocomplete": {
      "type": "string",
      "enumValues": ["off", "on"],
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "autofocus": {
      "type": "boolean",
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "name": {
      "type": "string",
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "placeholder": {
      "type": "string"
    },
    "rawValue": {
      "type": "string",
      "readOnly": true,
      "writeback": true
    },
    "readonly": {
      "type": "boolean"
    },
    "required": {
      "type": "boolean"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "regexp": {
          "type": "Object",
          "properties": {
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        }
      }
    },
    "validators": {
      "type": "Array"
    }
  },
  "methods": {
    "validate": {}
  },
  "extension": {
    _INNER_ELEM: 'input',
    _WIDGET_NAME: "inputBase",
    _GLOBAL_TRANSFER_ATTRS: ["accesskey", "aria-label", "tabindex"],
    _ALIASED_PROPS: {"readonly": "readOnly"}
  }
};
oj.CustomElementBridge.registerMetadata('inputBase', 'editableValue', inputBaseMeta);
})();

(function() {
var ojInputPasswordMeta = {
  "properties": {
    "value": {
      "type": "string",
      "writeback": true
    }
  },
  "methods": {},
  "extension": {
    _WIDGET_NAME: "ojInputPassword"
  }
};
oj.CustomElementBridge.registerMetadata('oj-input-password', 'inputBase', ojInputPasswordMeta);
oj.CustomElementBridge.register('oj-input-password', {'metadata': oj.CustomElementBridge.getMetadata('oj-input-password')});
})();

(function() {
var ojInputTextMeta = {
  "properties": {
    "clearIcon": {
      "type": "string",
      "enumValues": ["always", "conditional", "never"]
    },
    "converter": {
      "type": "Object"
    },
    "list": {
      "type": "string",
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "spellcheck": {
      "type": "boolean",
      "extension": {
        _ATTRIBUTE_ONLY: true,
        _COPY_TO_INNER_ELEM: true
      }
    },
    "value": {
      "type": "any",
      "writeback": true
    },
    "virtualKeyboard": {
      "type": "string",
      "enumValues": ["auto", "email", "number", "search", "tel", "text", "url"]
    }
  },
  "methods": {},
  "extension": {
    _WIDGET_NAME: "ojInputText"
  }
};
oj.CustomElementBridge.registerMetadata('oj-input-text', 'inputBase', ojInputTextMeta);
oj.CustomElementBridge.register('oj-input-text', {'metadata': oj.CustomElementBridge.getMetadata('oj-input-text')});
})();

(function() {
var ojTextAreaMeta = {
  "properties": {
    "converter": {
      "type": "Object"
    },
    "rows": {
      "type": "number",
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "spellcheck": {
      "type": "boolean",
      "extension": {
        _ATTRIBUTE_ONLY: true,
        _COPY_TO_INNER_ELEM: true
      }
    },
    "value": {
      "type": "any",
      "writeback": true
    }
  },
  "methods": {},
  "extension": {
    _INNER_ELEM: 'textarea',
    _WIDGET_NAME: "ojTextArea"
  }
};
oj.CustomElementBridge.registerMetadata('oj-text-area', 'inputBase', ojTextAreaMeta);
oj.CustomElementBridge.register('oj-text-area', {'metadata': oj.CustomElementBridge.getMetadata('oj-text-area')});
})();
/*
** Copyright (c) 2008, 2013, Oracle and/or its affiliates. All rights reserved.
**
**34567890123456789012345678901234567890123456789012345678901234567890123456789
*/

/*global ojld:true*/

/**
 * @class oj.LocaleData
 * @classdesc Locale Data Services
 * @export
 * @since 0.6
 * @hideconstructor
 */
oj.LocaleData = {};

/**
 * Sets the Locale Elements bundle used by JET
 * If an AMD loader (such as Require.js) is not present, this method should be called by the application to provide
 * a Locale Elements for JET.
 * This method may also be used by an application that wants to completely replace the Locale Elements bundle that is automatically
 * fetched by an AMD loader.
 * @param {Object} bundle resource bundle that should be used by the framework
 * @return {void}
 * @export
 */
oj.LocaleData.setBundle = function(bundle)
{
  oj.LocaleData._bundle = bundle;
};

/**
 * Retrieves the first day of week for the current locale's region
 * @return {number} a numeric representation of the first week day of the week: 
 * 0 for Sunday, 1 for Monday, etc.
 * @export
 */
oj.LocaleData.getFirstDayOfWeek = function()
{
  return oj.LocaleData._getWeekData("firstDay");
};

/**
 * Retrieves the first weekend day for the current locale's region
 * @return {number} a numeric representation of the first weekend day: 
 * 0 for Sunday, 1 for Monday, etc.
 * @export
 */
oj.LocaleData.getWeekendStart = function()
{
  return oj.LocaleData._getWeekData("weekendStart");
};

/**
 * Retrieves the last weekend day for the current locale's region
 * @return {number} a numeric representation of the last weekend day: 
 * 0 for Sunday, 1 for Monday, etc.
 * @export
 */
oj.LocaleData.getWeekendEnd = function()
{
  return oj.LocaleData._getWeekData("weekendEnd");
};

/**
 * Retrieves locale-specific names of the days of the week
 * @return {Array.<string>} names of the days from Sunday through Sturday
 * @param {string} [type] - the type of the name. Currently, "abbreviated", "narrow" and "wide" are supported
 * @ojsignature { target:"Type", for: "type", value: "'abbreviated'|'narrow'|'wide'"}
 * @export
 */
oj.LocaleData.getDayNames = function(type)
{
  if (type== null || (type !== "abbreviated" && type !== "narrow"))
  {
    type = "wide";
  }
  var days = oj.LocaleData._getCalendarData()["days"]["stand-alone"][type];
  
  return [days["sun"], days["mon"], days["tue"], days["wed"], days["thu"],  days["fri"],  days["sat"]];
};

/**
 * Retrieves locale-specific names of months
 * @return {Array.<string>} names of months from January through December
 * @param {string} [type] - the type of the name. Currently, "abbreviated", "narrow" and "wide" are supported
 * @ojsignature { target:"Type", for: "type", value: "'abbreviated'|'narrow'|'wide'"}
 * @export
 */
oj.LocaleData.getMonthNames = function(type)
{
  if (type== null || (type !== "abbreviated" && type !== "narrow"))
  {
    type = "wide";
  }
  var months = oj.LocaleData._getCalendarData()["months"]["stand-alone"][type];
  
  return [months["1"], months["2"], months["3"], months["4"], months["5"], months["6"],
           months["7"], months["8"], months["9"], months["10"], months["11"], months["12"]];
};

/**
 * Retrieves whether month is displayed prior to year
 * @return {boolean} whether month is prior to year
 * @export
 */
oj.LocaleData.isMonthPriorToYear = function() 
{
  var longDateFormat = oj.LocaleData._getCalendarData()["dateFormats"]["long"].toUpperCase(),
      monthIndex = longDateFormat.indexOf("M"),
	  yearIndex = longDateFormat.indexOf("Y");
  
  return monthIndex < yearIndex;
}

/**
 * @hidden
 * @private
 */
oj.LocaleData._getWeekData = function(key)
{
  var b = oj.LocaleData.__getBundle();
  var defRegion = "001";
  var region = oj.LocaleData._getRegion() || defRegion;
  
  var data = b["supplemental"]["weekData"][key];
  
  var val = data[region];
  
  if (val === undefined)
  {
    val = data[defRegion];
  }
  
  return val;
};

/**
 * @hidden
 * @private
 */
oj.LocaleData._getCalendarData = function()
{
   var b = oj.LocaleData.__getBundle();
   var main  = b['main'];
   
   // skip one level (the name of the locale)
   var data, p;
   for (p in main)
   {
     if (main.hasOwnProperty(p)) {
        data = main[p];
        break;
     }
   }
   return data['dates']['calendars']['gregorian'];
};

/**
 * @hidden
 * @private
 */
oj.LocaleData._getRegion = function()
{
  var locale = oj.Config.getLocale();
  if (locale)
  {
    var tokens = locale.toUpperCase().split(/-|_/);
    if (tokens.length >= 2)
    {
      var tag = tokens[1];
      if (tag.length == 4) // this is a script tag
      {
        if (tokens.length >= 3)
        {
          return tokens[2];
        }
      }
      else
      {
        return tag;
      }
    }
  }
  return null;
};

/**
 * @hidden
 * @private
 */
oj.LocaleData.__getBundle = function()
{
  var b = oj.LocaleData._bundle;
  if (b)
  {
    return b;
  }
  
  if (oj.__isAmdLoaderPresent()) {
    oj.Assert.assert(ojld !== undefined, "LocaleElements module must be loaded");
    return ojld;
  }
  return {};
};

/**
 * Called from oj.Config after AMD loader fetches LocaleElements for the new locale.
 *
 * @hidden
 * @private
 * 
 */
oj.LocaleData.__updateBundle = function(bundle)
{
  ojld = bundle;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*
 DESCRIPTION
 oj.OraI18nUtils provides helper functions for converter objects.
 
 PRIVATE CLASSES
 <list of private classes defined - with one-line descriptions>
 
 NOTES
 <other useful comments, qualifications, etc.>
 
 MODIFIED    (MM/DD/YY)
        05/13/14 - Creation
 */


oj.OraI18nUtils = {};
//supported numbering systems
oj.OraI18nUtils.numeringSystems = {
  'latn': "\u0030\u0031\u0032\u0033\u0034\u0035\u0036\u0037\u0038\u0039",
  'arab': "\u0660\u0661\u0662\u0663\u0664\u0665\u0666\u0667\u0668\u0669",
  'thai': "\u0e50\u0e51\u0e52\u0e53\u0e54\u0e55\u0e56\u0e57\u0e58\u0e59"
};

oj.OraI18nUtils.regexTrim = /^\s+|\s+$|\u200f|\u200e/g;
oj.OraI18nUtils.regexTrimNumber = /\s+|\u200f|\u200e/g;
oj.OraI18nUtils.regexTrimRightZeros = /0+$/g;
oj.OraI18nUtils.zeros = ["0", "00", "000"];
//ISO 8601 string accepted values:
//date only: YYYY or YYYY-MM or YYYY-MM-dd or YYYYMM or YYYYMMdd
//time only without timezone: Thh:mm or Thh:mm:ss or Thh:mm:ss.SSS or Thhmm or Thhmmss or Thhmmss.SSS
//time only with timezone: any of the time values above followed by any of the following: Z or +/-hh:mm or extended timezone like @America/Los_Angeles
//date time: any of the date values followed by any of the time values
oj.OraI18nUtils._ISO_DATE_REGEXP = /^\d{4}(?:-?\d{2}(?:-?\d{2})?)?(?:T\d{2}:?\d{2}(?::?\d{2}(?:\.\d{1,3})?)?(Z|[+-]\d{2}:\d{2})?)?$|^T\d{2}:?\d{2}(?::?\d{2}(?:\.\d{1,3})?)?(Z|[+-]\d{2}:\d{2})?$/;

/**
 * Returns the timezone offset between UTC and the local time in Etc/GMT[+-]hh:mm syntax.
 * The offset is positive if the local timezone is behind UTC and negative if
 * it is ahead. The offset range is between Etc/GMT-14 and Etc/GMT+12 (UTC-12 and UTC+14)
 * Examples:
 * 1- The local time is UTC-7 (Pacific Daylight Time):
 * oj.OraI18nUtils. getLocalTimeZoneOffset() will return the string "Etc/GMT+07:00" 
 * 2- The local time is UTC+1 (Central European Standard Time):
 * oj.OraI18nUtils. getLocalTimeZoneOffset() will return the string "Etc/GMT-01:00"  
 * @returns {string}  
 */
oj.OraI18nUtils.getLocalTimeZoneOffset = function () {
  var d = new Date();
  var offset = d.getTimezoneOffset();
  return oj.OraI18nUtils.getTimeStringFromOffset("Etc/GMT", offset, false, false);
};


/*
 * Will return timezone if it exists.
 */
oj.OraI18nUtils._getTimeZone = function (isoString)
{
  if (!isoString || typeof isoString !== "string")
  {
    return null;
  }
  var match = oj.OraI18nUtils._ISO_DATE_REGEXP.exec(isoString);
  //make sure it is iso string
  if (match === null)
  {
    oj.OraI18nUtils._throwInvalidISOString(isoString);
  }
  if (match[1] !== undefined)
    return match[1];
  return null;
};

/*
 * Will return local isoString provided a date.
 */
oj.OraI18nUtils.dateToLocalIso = function (date)
{
  var isoStr = oj.OraI18nUtils.padZeros(date.getFullYear(), 4) + "-" + oj.OraI18nUtils.padZeros((date.getMonth() + 1), 2) + "-" + oj.OraI18nUtils.padZeros(date.getDate(), 2) + "T" +
      oj.OraI18nUtils.padZeros((date.getHours()), 2) + ":" + oj.OraI18nUtils.padZeros((date.getMinutes()), 2) + ":" +
      oj.OraI18nUtils.padZeros((date.getSeconds()), 2);
  if (date.getMilliseconds() > 0) {
    isoStr += "." + oj.OraI18nUtils.trimRightZeros(oj.OraI18nUtils.padZeros(date.getMilliseconds(), 3));
  }
  return isoStr;
};

oj.OraI18nUtils.partsToIsoString = function (parts)
{
  var isoStr = oj.OraI18nUtils.padZeros(parts[0], 4) + "-" + oj.OraI18nUtils.padZeros(parts[1], 2) + "-" + oj.OraI18nUtils.padZeros(parts[2], 2) + "T" +
      oj.OraI18nUtils.padZeros(parts[3], 2) + ":" + oj.OraI18nUtils.padZeros(parts[4], 2) + ":" + oj.OraI18nUtils.padZeros(parts[5], 2);
  if (parts[6] > 0) {
    isoStr += "." + oj.OraI18nUtils.trimRightZeros(oj.OraI18nUtils.padZeros(parts[6], 3));
  }
  return isoStr;
};

oj.OraI18nUtils.isoToLocalDate = function (isoString)
{
  if (!isoString || typeof isoString !== "string")
  {
    return null;
  }
  return this._isoToLocalDateIgnoreTimezone(isoString);
};

oj.OraI18nUtils._isoToLocalDateIgnoreTimezone = function (isoString) {
  var datetime = oj.OraI18nUtils._IsoStrParts(isoString);
  return new Date(datetime[0], datetime[1] - 1, datetime[2], datetime[3], datetime[4], datetime[5], datetime[6]);
};

oj.OraI18nUtils._IsoStrParts = function (isoString) {
  var splitted = isoString.split("T"),
      tIndex = isoString.indexOf("T"),
      today = new Date(), i,
      datetime = [today.getFullYear(), today.getMonth() + 1, today.getDate(), 0, 0, 0, 0];

  if (splitted[0] !== "")
  {
    //contains date portion
    var dateSplitted = splitted[0].split("-");
    for (i = 0; i < dateSplitted.length; i++) {
      datetime[i] = parseInt(dateSplitted[i], 10);
    }
  }

  if (tIndex !== -1) {
    var milliSecSplitted = splitted[1].split("."), //contain millseconds
        timeSplitted = milliSecSplitted[0].split(":"); //contain hours, minutes, seconds

    for (i = 0; i < timeSplitted.length; i++)
    {
      datetime[3 + i] = parseInt(timeSplitted[i], 10);
    }

    if (milliSecSplitted.length === 2 && milliSecSplitted[1])
    {
      datetime[6] = parseInt(oj.OraI18nUtils.zeroPad(milliSecSplitted[1], 3, false), 10);
    }
  }
  return datetime;
};

oj.OraI18nUtils.getISOStrFormatInfo = function (isoStr) {
  var res = {
    'format': null,
    'dateTime': null,
    'timeZone': "",
    'isoStrParts': null
  };
  var exe = oj.OraI18nUtils._ISO_DATE_REGEXP.exec(isoStr);
  if (exe === null) {
    oj.OraI18nUtils._throwInvalidISOString(isoStr);
  }
  if (exe[1] === undefined && exe[2] === undefined) {
    res['format'] = 'local';
    res['dateTime'] = isoStr;
    res['isoStrParts'] = oj.OraI18nUtils._IsoStrParts(res['dateTime']);
    return res;
  }
  res['timeZone'] = (exe[1] !== undefined) ? exe[1] : exe[2];
  if (res['timeZone'] === 'Z')
    res['format'] = 'zulu';
  else
    res['format'] = 'offset';
  res['dateTime'] = isoStr.substring(0, isoStr.indexOf(res['timeZone']));
  res['isoStrParts'] = oj.OraI18nUtils._IsoStrParts(res['dateTime']);
  return res;
};

oj.OraI18nUtils._throwTimeZoneNotSupported = function () {
  var msg, error, errorInfo;
  msg = "time zone is not supported";
  error = new Error(msg);
  errorInfo = {
    'errorCode': 'timeZoneNotSupported'
  };
  error['errorInfo'] = errorInfo;
  throw error;
};

oj.OraI18nUtils._throwInvalidISOString = function (str) {
  var msg, error, errorInfo;
  msg = "The string " + str + " is not a valid ISO 8601 string.";
  error = new Error(msg);
  errorInfo = {
    'errorCode': 'invalidISOString',
    'parameterMap': {
      'isoStr': str
    }
  };
  error['errorInfo'] = errorInfo;
  throw error;
};

oj.OraI18nUtils.trim = function (value) {
  return (value + "").replace(oj.OraI18nUtils.regexTrim, "");
};

oj.OraI18nUtils.trimRightZeros = function (value) {
  return (value + "").replace(oj.OraI18nUtils.regexTrimRightZeros, "");
};


oj.OraI18nUtils.trimNumber = function (value) {
  var s = (value + "").replace(oj.OraI18nUtils.regexTrimNumber, "");
  return s;
};

oj.OraI18nUtils.startsWith = function (value, pattern) {
  return value.indexOf(pattern) === 0;
};

oj.OraI18nUtils.toUpper = function (value) {
  // "he-IL" has non-breaking space in weekday names.
  return value.split("\u00A0").join(" ").toUpperCase();
};

oj.OraI18nUtils.padZeros = function (num, c) {
  var r, s = num + "";
  if (c > 1 && s.length < c) {
    r = (oj.OraI18nUtils.zeros[c - 2] + s);
    return r.substr(r.length - c, c);
  }
  else {
    r = s;
  }
  return r;
};

oj.OraI18nUtils.zeroPad = function (str, count, left) {
  str = "" + str;
  var l;
  for (l = str.length; l < count; l += 1) {
    str = (left ? ("0" + str) : (str + "0"));
  }
  return str;
};

oj.OraI18nUtils.getTimeStringFromOffset = function (prefix, offset, reverseSign, alwaysMinutes) {
  var isNegative = reverseSign ? offset >= 0 : offset < 0;
  offset = Math.abs(offset);
  var hours = (offset / 60) << 0;
  var minutes = offset % 60;
  var sign = isNegative ? "-" : "+";
  if (alwaysMinutes) {
    hours = oj.OraI18nUtils.zeroPad(hours, 2, true);
  }
  var str = prefix + sign + hours;
  if (minutes > 0 || alwaysMinutes) {
    str += ":" + oj.OraI18nUtils.zeroPad(minutes, 2, true);
  }
  return str;
};

//get the numbering system key from the locale's unicode extension.
//Verify that the locale data has a numbers entry for it, if not return latn as default.
oj.OraI18nUtils.getNumberingSystemKey = function (localeElements, locale) {
  if (locale === undefined)
    return 'latn';
  var numberingSystemKey = oj.OraI18nUtils.getNumberingExtension(locale);
  var symbols = "symbols-numberSystem-" + numberingSystemKey;
  if (localeElements['numbers'][symbols] === undefined)
    numberingSystemKey = 'latn';
  return numberingSystemKey;
};

//return the language part
oj.OraI18nUtils.getBCP47Lang = function (tag) {
  var arr = tag.split("-");
  return arr[0];
};

//return the region part. tag is lang or lang-region or lang-script or
//lang-script-region
oj.OraI18nUtils.getBCP47Region = function (tag) {
  var arr = tag.split("-");
  if (arr.length === 3)
    return arr[2];
  if (arr.length === 2) {
    if (arr[1].length === 2)
      return arr[1];
  }
  return '001';
};


//get the unicode numbering system extension.
oj.OraI18nUtils.getNumberingExtension = function (locale) {
  locale = locale || "en-US";
  var idx = locale.indexOf("-u-nu-");
  var numbering = 'latn';
  if (idx !== -1) {
    numbering = locale.substr(idx + 6, 4);
  }
  return numbering;
};

oj.OraI18nUtils.haveSamePropertiesLength = function (obj) {
  var count = 0;
  for (var n in obj)
  {
    count++;
  }
  return count;
};

//cldr locale data start with "main" node.
//return the subnode under main.
oj.OraI18nUtils.getLocaleElementsMainNode = function (bundle) {
  var mainNode = bundle['main'];
  var subnode;
  for (var n in mainNode)
  {
    subnode = n;
    break;
  }
  return mainNode[subnode];
};

//get the locale which is a subnode of "main".
oj.OraI18nUtils.getLocaleElementsMainNodeKey = function (bundle) {
  var mainNode = bundle['main'];
  var subnode;
  for (var n in mainNode)
  {
    subnode = n;
    break;
  }
  return subnode;
};

oj.OraI18nUtils._toBoolean = function (value) {
  if (typeof value === "string") {
    var s = value.toLowerCase().trim();
    switch (s) {
      case "true":
      case "1":
        return true;
      case "false":
      case "0":
        return false;
      default:
        return value;
    }
  }
  return value;
};
//Return a function getOption.
//The getOption function extracts the value of the property named 
//property from the provided options object, converts it to the required type,
// checks whether it is one of a List of allowed values, and fills in a 
// fallback value if necessary.
oj.OraI18nUtils.getGetOption = function (options, getOptionCaller) {
  if (options === undefined) {
    throw new Error('Internal ' + getOptionCaller +
        ' error. Default options missing.');
  }

  var getOption = function getOption(property, type, values, defaultValue) {
    if (options[property] !== undefined) {
      var value = options[property];
      switch (type) {
        case 'boolean':
          value = oj.OraI18nUtils._toBoolean(value);
          break;
        case 'string':
          value = String(value);
          break;
        case 'number':
          value = Number(value);
          break;
        default:
          throw new Error('Internal error. Wrong value type.');
      }
      if (values !== undefined && values.indexOf(value) === -1) {
        var expectedValues = [];
        for (var i = 0; i < values.length; i++) {
          expectedValues.push(values[i]);
        }
        var msg = "The value '" + options[property] +
            "' is out of range for '" + getOptionCaller +
            "' options property '" + property + "'. Valid values: " +
            expectedValues;
        var rangeError = new RangeError(msg);
        var errorInfo = {
          'errorCode': 'optionOutOfRange',
          'parameterMap': {
            'propertyName': property,
            'propertyValue': options[property],
            'propertyValueValid': expectedValues,
            'caller': getOptionCaller
          }
        };
        rangeError['errorInfo'] = errorInfo;
        throw rangeError;
      }

      return value;
    }
    return defaultValue;
  };

  return getOption;
};

/**
 * matches a string to a reference string and returns the start and end indexes
 * of the match in the referensed string. The locale and options arguments let 
 * applications specify the language whose sort order should be used and customize 
 * the behavior of the function.
 * 
 * @param {string} str the reference string 
 * @param {string} pat The string against which the reference string is compared
 * @param {string} locale a BCP 47 language tag
 * @param {Object=} options Optional. An object with the following property:
 * sensitivity: 
 *   Which differences in the strings should lead to non-zero result values. Possible values are:
 *   "base": Only strings that differ in base letters compare as unequal. Examples: a ? b,  a = A.
 *   "accent": Only strings that differ in base letters or accents and other diacritic marks compare as unequal. Examples: a ? b, , a = A.
 *   "case": Only strings that differ in base letters or case compare as unequal. Examples: a ? b, a ? A.
 *   "variant": Strings that differ in base letters, accents and other diacritic marks, or case compare as unequal. 
 *   The default is base.
 * @return {Array|null} an array containing the start and end indexes of the match or null if there is no match.
 */
oj.OraI18nUtils.matchString = function (str, pat, locale, options) {
  if (options === undefined) {
    options = {sensitivity: 'base', usage: 'sort'};
  }
  var getOption = oj.OraI18nUtils.getGetOption(options, "oj.OraI18nUtils.matchString");
  options['usage'] = getOption('usage', 'string', ['sort', 'search'], 'sort');
  options['sensitivity'] = getOption('sensitivity', 'string', ['base', 'accent', 'case', 'variant'], 'base');
  var i, j;
  var len = str.length;
  var patLen = pat.length - 1;
  for (i = 0; i < len; i++) {
    for (j = 0; j < 3; j++) {
      var len2 = len - i;
      len2 = Math.min(len2, patLen + j);
      var str2 = str.substr(i, len2);
      var res = str2.localeCompare(pat, locale, options);
      if (res === 0) {
        var end = i + len2 - 1;
        var ret = [i, end];
        return ret;
      }
    }
  }
  return null;
};

var _DEFAULT_TIME_PORTION = "T00:00:00.000";
var _DATE_TIME_KEYS = {"fullYear": {pos: 0, pad: 4}, "month": {pos: 1, pad: 2}, "date": {pos: 2, pad: 2},
  "hours": {pos: 3, pad: 2}, "minutes": {pos: 4, pad: 2}, "seconds": {pos: 5, pad: 2},
  "milliseconds": {pos: 6, pad: 3}, "timeZone": {pos: 7}};

/**
 * Parses the isoString and returns a JavaScript Date object
 * 
 * @param {string} isoString isoString to parse and to return Date of
 * @return {Date} the parsed JavaScript Date Object
 */
oj.OraI18nUtils.isoToDate = function (isoString)
{
  //note new Date w/ isoString in IE fails so need to use parsing from momentjs support
  return new Date(this._normalizeIsoString(isoString));
};

/**
 * Will return an updated toIsoString using the timePortion from the fromIsoString or from the default 
 * oj.OraI18nUtils.DEFAULT_TIME_PORTION
 * 
 * @private
 * @param {string} fromIsoString isoString that may not be a complete isoString
 * @param {string} toIsoString isoString that may not be a complete isoString
 * @returns {string} modified toIsoString with original date portion and the time portion from the fromIsoString
 * @since 1.1
 */
oj.OraI18nUtils._copyTimeOver = function (fromIsoString, toIsoString)
{
  if (!fromIsoString || !toIsoString)
  {
    throw new Error("Provided invalid arguments");
  }

  //need to only normalize toIsoString, since copying only time from fromIsoString
  toIsoString = this._normalizeIsoString(toIsoString);

  var fromTimeIndex = fromIsoString.indexOf("T"),
      toTimeIndex = toIsoString.indexOf("T"),
      toDatePortion = toIsoString.substring(0, toTimeIndex),
      fromTimePortion = fromTimeIndex !== -1 ? fromIsoString.substring(fromTimeIndex) : _DEFAULT_TIME_PORTION;

  return toDatePortion + fromTimePortion;
};

/**
 * Clears the time portion of the isoString
 * 
 * @private
 * @param {string} isoString isoString that may not be a complete isoString
 * @returns {string} an updated isoString
 * @since 1.1
 */
oj.OraI18nUtils._clearTime = function (isoString)
{
  return this._dateTime(isoString, {"hours": 0, "minutes": 0, "seconds": 0, "milliseconds": 0});
};

/**
 * Will accept an isoString and perform a get operation or a set operation depending on whether param is an Array 
 * or a JSON 
 * 
 * The keys for the get and set operation are defined in _DATE_TIME_KEYS.
 * 
 * Note the handling of month starting with 0 in Date object and being 1 based in isoString will be handled by the function 
 * with the usage of doParseValue. Meaning when you doParseValue and you are getting the value it will automatically 
 * decrement the value and when you are setting the param it will check if the value is of number and if so will 
 * increment it.
 * 
 * @private
 * @param {string} isoString isoString that may not be a complete isoString
 * @param {Array|Object} actionParam if an Array will be a get operation, if a JSON will be a set operation
 * @param {boolean=} doParseValue whether one should parseInt the value during the get request
 * @returns {Object|string} an Object when a get operation and a string when a set operation
 * @since 1.1
 */
oj.OraI18nUtils._dateTime = function (isoString, actionParam, doParseValue)
{
  if (!isoString || !actionParam)
  {
    throw new Error("Invalid argument invocation");
  }

  var pos, value, dtKey,
      retVal = null,
      key = null,
      dateTimeKeys = _DATE_TIME_KEYS,
      oraUtilsPadZero = this.padZeros,
      isoStringNormalized = this._normalizeIsoString(isoString), //note intentionally normalizing 
      captured = /(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):?(\d{2})?\.?(\d{3})?(.*)?/.exec(isoStringNormalized);

  if (!captured)
  {
    throw new Error("Unable to capture anything");
  }

  captured = captured.slice(1);

  if (Array.isArray(actionParam))
  {
    retVal = {};

    //means an array so perform a get operation
    for (var i = 0, len = actionParam.length; i < len; i++)
    {
      key = actionParam[i];

      if (key in dateTimeKeys)
      {
        pos = dateTimeKeys[key].pos;
        value = captured[pos];

        if (doParseValue && "timeZone" === key)
        {
          throw new Error("Dude you tried to ask timezone to be parsed");
        }

        if (doParseValue)
        {
          var parsed = parseInt(value, 10);
          retVal[key] = pos === 1 ? parsed - 1 : parsed; //since month is 0 based, though about having a callback but month only special
        }
        else
        {
          retVal[key] = value;
        }

      }
    }
  }
  else if ($.isPlainObject(actionParam))
  {

    for (var keys in actionParam)
    {
      dtKey = dateTimeKeys[keys];
      pos = dtKey.pos;
      value = actionParam[keys];

      //special case for month again, 0 based so check if number and if so increment it
      if (pos === 1 && typeof value === "number")
      {
        value++;
      }
      captured[pos] = dtKey.pad ? oraUtilsPadZero(value, dtKey.pad) : value;
    }
    //"2015-02-02T21:12:30.255Z"
    retVal = captured[0] + "-" + captured[1] + "-" + captured[2] + "T" + captured[3] + ":" + captured[4] + ":" + captured[5] +
        (captured.length > 6 && captured[6] ? ("." + captured[6] + (captured.length === 8 && captured[7] ? captured[7] : "")) : "");
  }

  return retVal;
};

/**
 * So the problem is Jet uses incomplete isoString which causes issues in different browsers. 
 * 
 * For instance for a new Date().toISOString() => 2015-02-02T18:00:37.007Z
 * ojInputDate stores 2015-02-02
 * ojInputTime stores T18:00:37.007Z
 * 
 * yet constructing new Date(val) on above causes different results or errors in different browsers, so 
 * this function is to normalize them. Note it is assumed that the point is creating the Date object from the 
 * normalized isoString. Meaning if both contain only the time portion today's date will appended to it.
 * 
 * Here are the use cases
 * 
 * @private
 * @param {string} isoString isoString that may not be a complete isoString
 * @returns {string} a normalized isoString
 * @since 1.1
 */
oj.OraI18nUtils._normalizeIsoString = function (isoString)
{
  if (!isoString)
  {
    throw new Error("Provided invalid arguments");
  }

  var checkTime = function (timeValue)
  {
    var splitted = timeValue.split(":");
    if (splitted.length > 1)
    {
      return timeValue;
    }
    //need at least hour + minute for proper parsing on browser except IE
    return timeValue + ":00";
  },
      todayIsoString = new Date().toISOString(),
      todayDatePortion = todayIsoString.substring(0, todayIsoString.indexOf("T")),
      timeIndex = isoString.indexOf("T"),
      datePortion = timeIndex === -1 ? isoString : isoString.substring(0, timeIndex),
      timePortion = timeIndex !== -1 ? checkTime(isoString.substring(timeIndex)) : _DEFAULT_TIME_PORTION;

  datePortion = datePortion || todayDatePortion;

  return datePortion + timePortion;
};

/**
 * formats a parameterized string
 * 
 * @param {string} str string to be formatted. EX: "{0} bytes"
 * @param {Array} params array of parameters to be substituted in the string
 * @returns {string} The formatted string
 * @since 4.0.0
 */
oj.OraI18nUtils.formatString = function (str, params)
{
  var i;
  var token;
  var len = params.length;
  
  for (i = 0; i < len; i++)
  {
    token = '{' + i + '}';
    str = str.replace(token, params[i]);
  }
  return str;
};

/**
 * Returns a person's initials
 * 
 * @param {string=} firstName first name
 * @param {string=} lastName last name or surname
 * @returns {string|undefined} uppercase concatenation of first letter of first name and first letter
 * of last name.
 * There are the following special cases:
 * - If the name is Arabic characters, it returns empty string.
 * - If the name is Hindi characters, it returns the first letter of the first name.
 * - If the name is Thai characters, it returns the first letter of the first name.
 * - If the name is Korean characters, it returns the first name.
 * - If the name is Japanese or Chinese characters, it returns the last name.
 *   
 * @since 4.0.0
 */
oj.OraI18nUtils.getInitials = function (firstName, lastName)
{
  //We assume the names are valid. We test the first character only.
  var c = 0; 
  var c1 = 0;
  var u;
  if(firstName !== undefined && firstName.length > 0)
    c = firstName.charCodeAt(0);
  
  // Arabic characters. Return empty string
  if(c >= 0x0600 && c <= 0x06FF)
    return "";
  
  // Hindi characters. Return first letter of the first name
  if(c >= 0x0900 && c <= 0x097F)
    return firstName.charAt(0);
  
  // Thai characters. Return first letter of the first name
  if(c >= 0x0E00 && c <= 0x0E7F)
    return firstName.charAt(0);
  
  // Korean characters. Return first name
  if((c >= 0x1100 && c <= 0x11FF) || (c >= 0x3130 && c <= 0x318F) || 
     (c >= 0xA960 && c <= 0xA97F) || (c >= 0xAC00 && c <= 0xD7FF))
    return firstName;
  
  // Japanese and Chinese characters. Return last name
  if(lastName !== undefined && lastName.length > 0)
    c = lastName.charCodeAt(0);
  if((c >= 0x2E80 && c <= 0x2FDF) || (c >= 0x3000 && c <= 0x312F) ||
     (c >= 0x3190 && c <= 0x31FF) || (c >= 0x3300 && c <= 0x4DBF) ||
     (c >= 0x4E00 && c <= 0x9FFF) || (c >= 0xF900 && c <= 0xFAFF))
     return lastName;
   
  //Handle surrogate characters for Japanese and Chinese characters.  
  if(c >= 0xD800 && c <= 0xDBFF) {
    if(lastName.length < 2)
      return "";
    c1 = lastName.charCodeAt(1);
    //c1 must be in DC00-DFFF range
    if(c1 < 0xDC00 || c1 > 0xDFFF)
      return "";
    //Convert high and low surrogates into unicode scalar.
    u =  ((c - 0xD800) * 0x400) + (c1 - 0xDC00) + 0x10000;
    //test the blocks
    if((u >= 0x1B000 && u <= 0x1B0FF) || (u >= 0x1F200 && u <= 0x1F2FF) ||
       (u >= 0x20000 && u <= 0x2A6DF) || (u >= 0x2A700 && u <= 0x2B73F) ||
       (u >= 0x2B740 && u <= 0x2B81F) || (u >= 0x2B820 && u <= 0x2CEAF) ||
       (u >= 0x2F800 && u <= 0x2FA1F))
      return lastName;
  }
  
  //return default
  c = "";
  c1 = "";
  if(firstName !== undefined && firstName.length > 0)
    c = firstName.charAt(0).toUpperCase();
  if(lastName !== undefined && lastName.length > 0)
    c1 = lastName.charAt(0).toUpperCase();
  return c + c1;
};


/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Used to register and retrieve converterFactories and validationFactories.
 * There is no need to call the constructor.
 * @see oj.Validation.converterFactory
 * @see oj.Validation.validatorFactory
 * @see oj.ConverterFactory
 * @see oj.ValidatorFactory
 * @name oj.Validation
 * @class 
 * @hideconstructor
 * @export
 * @since 0.6
 * 
 */
oj.Validation = {};

/**
 * Internal properties to hold all factory provider callbacks or instances by name
 * @private
 */
oj.Validation._converterFactories = {}; oj.Validation._validatorFactories = {};

/**
 * Internal properties to hold the default factory instances.
 * @private
 */
oj.Validation._defaultConverterFactories = {}; oj.Validation._defaultValidatorFactories = {};

/**
 * Internal property that identifies the type that is the contract for conveters and validators.
 * @private
 */
oj.Validation._CONTRACTS = {'converter' : {name: "oj.ConverterFactory",  type: oj.ConverterFactory},
                            'validator': {name: "oj.ValidatorFactory", type: oj.ValidatorFactory}};

/**
 * Method to register and retrieve converter factory instances by name. 
 * When passed only the name, an existing factory (registered for the name) is returned. Callers can 
 * expect to get back the default 'number', 'datetime', or 'color' converters. 
 * When passed two arguments, a new factory for the name is registered. If the name already exists 
 * the new instance replaces the old one. 
 * 
 * @param {string} type a case insensitive name of the converter factory. e.g., 'number' (or
 * oj.ConverterFactory.CONVERTER_TYPE_NUMBER), 'datetime' (or oj.ConverterFactory.CONVERTER_TYPE_DATETIME),
 * or 'color' (or oj.ConverterFactory.CONVERTER_TYPE_COLOR).
 * @param {Object=} instance the instance of the factory that implements the contract for 
 * oj.ConverterFactory.
 * 
 * @export
 * @ojsignature {target: "Type", 
 *                value: "<V, CF extends oj.ConverterFactory<V>>(type:'number'|'color'|'datetime'|string, instance?:CF): CF|null"}
 * @see oj.ConverterFactory
 * @example <caption>Initialize a JET datetime converter instance using options
 * and use it in a knockout-binding and bind it to the
 * converter attribute on the JET form element.</caption>
 * ----- Javascript ViewModel----
 * self.converter = ko.observable(
    oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).
      createConverter({formatType: 'date', dateFormat: 'full'}));
 * ----- HTML -----
 * &lt;oj-input-date value="{{endDate}}" converter="[[converter]]">&lt/oj-input-date>
 * @example <caption>Create your own converter, register it, and use it when displaying information
 * on the page to the user.</caption>
 * ----- Javascript -----
 * RelativeDateTimeConverterFactory = (function () {
 *  function _createRelativeDateTimeConverter(options)
 *  {
 *    // this is a custom converter (code not shown here) that formats the datetime into "Today" or 
 *    // "Tomorrow", or "This Week"/"Next Week", etc.
 *    return new RelativeDateTimeConverter(options);
 *  }
 *  return {
 *    'createConverter': function (options) {
 *      return _createRelativeDateTimeConverter(options);
 *    }
 *  };
 *  }());
 * oj.Validation.converterFactory("relativeDate", RelativeDateTimeConverterFactory);
 *  // Use our custom relativeDate converter.
 *  // In this demo, we want to see in the Schedule For column the words 
 *  // Today or Tomorrow so we set the relativeField option's value
 *  // to 'day'. If we want to see This Week, we'd set it to 'week', etc.
 *  var rdConverter =  oj.Validation.converterFactory("relativeDate")
 *  .createConverter({relativeField: 'day', year: "numeric", month: "numeric", day: "numeric"});
 *  ...
 *  // Our custom converter's format function returns an object with 'value' and 'title'.
 *  // We put the 'value' in innerHTML so the user can read it. E.g., Today or Tomorrow.
 *  // And we put the actual date in the title. The user can read it when they hover over
 *  // the word Today or Tomorrow.
 *  content = rdConverter.format(context.row.ScheduleFor);
 *  span.setAttribute('title', content.title);
 *  span.innerHTML = content.value;
 * ...
 * @example <caption>On your JET form component that has a 'converter' attribute,
 * set the 'converter' attribute using "type" of a registered converter. This 
 * could be a JET Converter or a oj.Converter 
 * (or Object that duck-types it) that you wrote. (it must implement parse and format
 * since our JET form components call format and parse.</caption>
 * &lt;oj-input-date value="{{date}}" 
 * converter= '{"type":"datetime", 
 * "options": {"year": "numeric", "month": "long", "day": "numeric"}}'> &lt;/oj-input-date>
 */
oj.Validation.converterFactory = function (type, instance)
{
  var retValue;
  if (type && !instance)
  {
    // getter
    retValue = oj.Validation._getFactory(type, oj.Validation._converterFactories);
  }
  else if (type && instance)
  {
    // setter
    retValue = oj.Validation._registerFactory(type, 
                                              instance, 
                                              oj.Validation._converterFactories,
                                              oj.Validation._CONTRACTS['converter']);
  }
  
  return retValue;
};

/**
 * Method to register and retrieve validator factory instances by name. 
 * When passed only the name, an existing factory (registered for the name) is returned. 
 * When passed two arguments, a new factory for the name is registered. If the name already exists 
 * the new instance replaces the old one. 
 * 
 * @param {string} type a case insensitive name of the validator factory. 
 * @param {Object=} instance the instance of the factory that implements the contract for 
 * oj.ValidatorFactory.
 * @example <caption>Initialize a JET validator instance using options
 * and use it in a knockout-binding and bind it to the
 * validators attribute on the JET form element.</caption>
 * ----- Javascript ViewModel----
 * var options = {min: 5, max: 10};
 * var validator =
 * oj.Validation.validatorFactory(
 * oj.ValidatorFactory.VALIDATOR_TYPE_LENGTH).createValidator(options);
 * self.validator = ko.observable([validator]);
 * ----- HTML -----
 * &lt;oj-input-text validators="[[validator]]">&lt/oj-input-text>
 * @example <caption>An easier usage is to return an Object since the validators option
 * takes an array where each item is either an instance that duck types oj.Validator, 
 * or is an Object literal containing type and options.</caption>
 * ----- Javascript ViewModel -----
 * self.validators = ko.computed(function() {
 * return [{type: 'numberRange', options: { min: 10000.05, max: 25000.95, 
 * hint: {inRange: 'Enter a value between {min} and {max}.'}}}];
 * });
 * ----- HTML -----
 * &lt;oj-input-text id="numberRange1" value="{{numberValue1}}" validators="[[validators]]">
 * &lt;/oj-input-text>
 * @example <caption>Get a registered validator with your options, and calls its validate method
 * on your own value.</caption>
 * var reqValOptions = {
 *   'hint': reqTrans['hint'] || null,
 *   'label': this._getLabelText(), 
 *   'messageSummary': reqTrans['messageSummary'] || null,
 *   'messageDetail': reqTrans['messageDetail'] || null
 * };
 * var vf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED);
 * var requiredValidator = vf.createValidator(reqValOptions);
 * var isValid = validator.validate(value);
 * @export
 * @since 0.6
 * @see oj.ValidatorFactory
 * @ojsignature {target: "Type", 
 *   value: "<V, VF extends oj.ValidatorFactory<V>>(type:'required'|'regexp'|'numberRange'|'length'|'dateTimeRange'|'dateRestriction'|string, instance?:VF): VF|null"}
 */
oj.Validation.validatorFactory = function (type, instance)
{
  var retValue;
  if (type && !instance)
  {
    // getter
    retValue = oj.Validation._getFactory(type, oj.Validation._validatorFactories);
  }
  else if (type && instance)
  {
    // setter
    retValue = oj.Validation._registerFactory(type, 
                                              instance, 
                                              oj.Validation._validatorFactories, 
                                              oj.Validation._CONTRACTS['validator']);
  }
  
  return retValue;
};

/**
 * Returns the default converter factory instances for the supported types as defined by the 
 * oj.ConverterFactory. Instead of calling this method, it should be 
 * sufficient to use oj.Validation.converterFactory.
 * 
 * @param {string} type The default converter factory for the type. Supported types are 'number' and 
 * 'datetime'
 * @return {Object|null} an instance of oj.ConverterFactory or null if an unknown type is requested.
 * 
 * @export
 * @ojsignature {target:"Type", 
 *    value: "<V>(type: 'number'|'color'|'datetime'): oj.ConverterFactory<V>|null"}
 * @see oj.ConverterFactory
 * @see oj.Validation.converterFactory
 * 
 */
oj.Validation.getDefaultConverterFactory = function (type)
{
  return oj.Validation._getFactory(type, oj.Validation._defaultConverterFactories);
};

/**
 * Returns the default validator factory instance for the requested types as defined by the 
 * oj.ValidatorFactory.  Instead of calling this method, it should be 
 * sufficient to use oj.Validation.validatorFactory.
 * 
 * @param {string} type The default converter factory for the type. Supported types are 'number' and 
 * 'datetime'
 * @return {Object|null} an instance of oj.ConverterFactory or null if an unknown type is requested.
 * 
 * @export
 * @ojsignature {target:"Type", 
 *    value: "<V>(type: 'required'|'regexp'|'numberRange'|'length'|'dateTimeRange'|'dateRestriction'): oj.ValidatorFactory<V>|null"}
 * @see oj.ValidatorFactory
 * @see oj.Validation.validatorFactory
 */
oj.Validation.getDefaultValidatorFactory = function (type)
{
  return oj.Validation._getFactory(type, oj.Validation._defaultValidatorFactories);
};

// PACKAGE PRIVATE METHODS
/**
 * Called only by internal jet converter factory implementations.
 * 
 * @param {string} name
 * @param {Object} instance
 * @private
 */
oj.Validation.__registerDefaultConverterFactory = function (name, instance)
{
  // save to both factories
  var contractDef = oj.Validation._CONTRACTS['converter'];
  oj.Validation._registerFactory (name, instance, oj.Validation._defaultConverterFactories, contractDef);
  oj.Validation._registerFactory(name, instance, oj.Validation._converterFactories, contractDef);
};

/**
 * Called only by internal jet validator factory implementations.
 * 
 * @param {string} name of the validator factory
 * @param {Object} instance of the validator factory that creates instances of the validator
 * @private
 */
oj.Validation.__registerDefaultValidatorFactory = function (name, instance)
{
  // save to both factories
  var contractDef = oj.Validation._CONTRACTS['validator'];
  oj.Validation._registerFactory (name, instance, oj.Validation._defaultValidatorFactories, contractDef);
  oj.Validation._registerFactory(name, instance, oj.Validation._validatorFactories, contractDef);
};


/**
 * Checks that the instance implements the interface type. If it doesn't it throws an error.
 * @param {Object} instance
 * @param {Object} type
 * @param {string} typeName 
 * @throws {Error} if instance does not implement the methods defined on type.  
 * @private
 */
oj.Validation._doImplementsCheck = function (instance, type, typeName)
{
  if (type)
  {
    // Check that instance duck types providerType
    if (!oj.Validation._quacksLike(instance, type))
    {
      throw new Error("Factory instance does not implement the methods expected by the factory of type " + typeName);
    }
  }
};

/**
 * Retrieves the converter factory by name from the provided factories.
 * @private
 */
oj.Validation._getFactory = function(name, factories)
{
  oj.Assert.assertString(name);
  var cachedInstance = null;
  
  if (name)
  {
    name = name.toLowerCase();

    // getter called to retrieve the factory instance 
    var providerProps = factories[name] || {}; 
    cachedInstance = providerProps['instance'] || null;
  }
  // TODO: log a warning that name is null
  return cachedInstance;
};

/**
 * Tests whether an object 'quacks like a duck'. Returns `true` if the thingie has all of the 
 * methods of the second, parameter 'duck'. Returns `false` otherwise. 
 *
 * @param {Object} thingie the object to test.
 * @param {Object} duck The archetypal object, or 'duck', that the test is against.
 * @private
*/
oj.Validation._quacksLike = function(thingie, duck) 
{
  var valid = true, property;

  oj.Assert.assertObject(thingie);
  oj.Assert.assertObject(duck);

  for (property in duck) 
  {
    // Ensure that thingie defines the same functions as duck. We don't care about other properties
    if (duck.hasOwnProperty(property)) 
    {
      if (typeof duck[property] === "function" && 
              !thingie[property] && typeof thingie[property] !== "function") 
      {
        valid = false;
        break;
      }
    }
  }
  
  return valid;
};

/**
 * Registers the factory instance by the name, storing it into the factories object, after ensuring 
 * that the instance duck types the specified contract.
 * 
 * @private
 */
oj.Validation._registerFactory = function(name, instance, factories, contractDef)
{
  oj.Assert.assertString(name);
  oj.Assert.assertObject(instance);

  if (name)
  {
    // set new provider factory function clearing out the previously stored instance
    var props = {};
    props['instance'] = instance;
    oj.Validation._doImplementsCheck(instance, contractDef.type, contractDef.name);

    // Save to default and public factories
    factories[name.toLowerCase()] = props;
  }
};

/**
 * Contract for a ConverterFactory that provides a factory method to create a converter instance for 
 * the requested type. JET provides three factory implementations for number and datetime 
 * and color types that 
 * implement this contract. Customers can register custom converter factories for the supported 
 * types or create and register factories for new types altogether.
 * 
 * @name oj.ConverterFactory
 * @abstract
 * @class
 * @export
 * @since 0.6
 * @ojsignature {target: "Type", value: "class ConverterFactory<V>"}
 * @see oj.Validation
 * @see oj.NumberConverterFactory
 * @see oj.DateTimeConverterFactory
 * @see oj.ColorConverterFactory
 */
oj.ConverterFactory = 
{
  /**
   * Default type for a factory used to create number converters. This type is passed to the 
   * [Validation.converterFactory]{@link oj.Validation#converterFactory} method to retrieve the 
   * number converter factory of type {@link oj.NumberConverterFactory}. 
   * @example <caption>Create a JET number converter with options</caption>
   * var convFactory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER);
   * var converter  =  convFactory.createConverter({
   *  style: 'currency', 
   *  currency: 'USD', 
   *  currencyDisplay: 'symbol', 
   *  pattern: '¤ ##,##0.00'});
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "CONVERTER_TYPE_NUMBER" : 'number',
  
  /**
   * Default type for a factory used to create datetime converters. This type is passed to the 
   * [Validation.converterFactory]{@link oj.Validation#converterFactory} method to retrieve the 
   * datetime converter factory of type {@link oj.DateTimeConverterFactory}. 
   * @example <caption>Create a JET dateTime converter with options</caption>
   * var dateTimeCvtr = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME);
   * var dateOptions = {day: 'numeric', month: 'numeric'};
   * var dayMonthConverter = dateTimeCvtr.createConverter(dateOptions);
   * 
   * @expose
   * @const
   * @member
   * @type {string}
   */

  "CONVERTER_TYPE_DATETIME" : 'datetime',
  
  /**
   * Default type for a factory used to create color converters. This type is passed to the 
   * [Validation.converterFactory]{@link oj.Validation#converterFactory} method to retrieve the 
   * color converter factory of type {@link oj.ColorConverterFactory}.
   * @example <caption>Create a JET color converter with options</caption>
   * this._convFactory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_COLOR);
   * this._convHex  =  this._convFactory.createConverter({"format": "hex"})
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "CONVERTER_TYPE_COLOR" : 'color',
  
  /**
   * Creates an immutable converter instance of the type the factory implements. 
   * 
   * @param {(Object|null)} options an object literal containing properties required by the converter 
   * for its initialization. The properties provided in the options is implementation specific.
   * 
   * @return {Object} a converter instance.
   * @ojsignature { target: "Type", 
   *                value: "oj.Converter<V>",
   *                for: "returns"}
   * @instance
   * @throws {TypeError} if an unrecognized type was provided 
   * @expose
   * 
   * @example <caption>Create a JET dateTime converter with options</caption>
   * var dateTimeCvtr = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME);
   * var dateOptions = {day: 'numeric', month: 'numeric'};
   * var dayMonthConverter = dateTimeCvtr.createConverter(dateOptions);
   * @example <caption>Create your own ConverterFactory and Converter, register the Converter on your
   * ConverterFactory, and use it when displaying information
   * on the page to the user.</caption>
   * ----- Javascript -----
   * RelativeDateTimeConverterFactory = (function () {
   *  function _createRelativeDateTimeConverter(options)
   *  {
   *    // this is a custom converter (code not shown here) that formats the datetime into "Today" or 
   *    // "Tomorrow", or "This Week"/"Next Week", etc.
   *    return new RelativeDateTimeConverter(options);
   *  }
   *  return {
   *    'createConverter': function (options) {
   *      return _createRelativeDateTimeConverter(options);
   *    }
   *  };
   *  }());
   * oj.Validation.converterFactory("relativeDate", RelativeDateTimeConverterFactory);
   *  // Use our custom relativeDate converter.
   *  // In this demo, we want to see in the Schedule For column the words 
   *  // Today or Tomorrow so we set the relativeField option's value
   *  // to 'day'. If we want to see This Week, we'd set it to 'week', etc.
   *  var rdConverter =  oj.Validation.converterFactory("relativeDate")
   *  .createConverter({relativeField: 'day', year: "numeric", month: "numeric", day: "numeric"});
   *  ...
   *  // Our custom converter's format function returns an object with 'value' and 'title'.
   *  // We put the 'value' in innerHTML so the user can read it. E.g., Today or Tomorrow.
   *  // And we put the actual date in the title. The user can read it when they hover over
   *  // the word Today or Tomorrow.
   *  content = rdConverter.format(context.row.ScheduleFor);
   *  span.setAttribute('title', content.title);
   *  span.innerHTML = content.value;
   * ...
   */
  createConverter : function(options) {}
};

/**
 * Contract for a ValidatorFactory that provides a factory method to create a validator instance for 
 * the requested type. JET provides several factory implementations that implement this contract - 
 * for example dateRestriction, dateTimeRange, numberRange, length, required, regexp. Customers can 
 * register custom validator factories for the supported types or create and register factories for 
 * new types altogether.
 * 
 * @name oj.ValidatorFactory
 * @abstract
 * @class
 * @ojsignature {target: "Type", value: "class ValidatorFactory<V>"}
 * @export
 * @since 0.6
 * @see oj.Validation
 * @see oj.DateRestrictionValidatorFactory
 * @see oj.DateTimeRangeValidatorFactory
 * @see oj.LengthValidatorFactory
 * @see oj.NumberRangeValidatorFactory
 * @see oj.RegExpValidatorFactory
 * @see oj.RequiredValidatorFactory
 */
oj.ValidatorFactory = 
{
  /**
   * Default type for a factory used to create required validators. This type is passed to the 
   * [Validation.validatorFactory]{@link oj.Validation#validatorFactory} method to retrieve the 
   * required validator factory of type {@link oj.RequiredValidatorFactory}.
   * @example <caption>Create a JET required validator</caption>
   * var rvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED); 
   * var options = {'hint' : 'a value is required for this field'};
   * var requiredValidator = rvf.createValidator(options);
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "VALIDATOR_TYPE_REQUIRED" : 'required',
          
  /**
   * Default type for a factory used to create regExp validators. This type is passed to the 
   * [Validation.validatorFactory]{@link oj.Validation#validatorFactory} method to retrieve the 
   * regExp validator factory of type {@link oj.RegExpValidatorFactory}.
   * @example <caption>Create a JET regexp validator</caption>
   * var validatorFactory = 
   * oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP);
   * var options =  
   * {pattern: '[a-zA-Z0-9]{3,}', 
   * hint: 'enter at least 3 letters or numbers.',
   * messageDetail: 'You must enter at least 3 letters or numbers.'}
   * var validator = validatorFactory.createValidator(options);
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "VALIDATOR_TYPE_REGEXP" : 'regexp',

  /**
   * Default type for a factory used to create numberRange validators. This type is passed to the 
   * [Validation.validatorFactory]{@link oj.Validation#validatorFactory} method to retrieve the 
   * numberRange validator factory of type {@link oj.NumberRangeValidatorFactory}.
   * @example <caption>Create a JET numberrange validator</caption>
   * var validatorFactory = 
   * oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_NUMBERRANGE); 
   * var options = {min: 10000.05, max: 25000.95, 
   * hint: {inRange: 'Enter a value between {min} and {max}.'}};
   * var validator = validatorFactory.createValidator(options);
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "VALIDATOR_TYPE_NUMBERRANGE" : 'numberRange',

  /**
   * Default type for a factory used to create length validators. This type is passed to the 
   * [Validation.validatorFactory]{@link oj.Validation#validatorFactory} method to retrieve the 
   * length validator factory of type {@link oj.LengthValidatorFactory}.
   * @example <caption>Create a JET length validator</caption>
   * var validatorFactory = 
   * oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_LENGTH); 
   * var options = {min: 5, max: 10};
   * var validator = validatorFactory.createValidator(options);
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "VALIDATOR_TYPE_LENGTH" : 'length',

  /**
   * Default type for a factory used to create required validators. This type is passed to the 
   * [Validation.validatorFactory]{@link oj.Validation#validatorFactory} method to retrieve the 
   * dateTimeRange validator factory of type {@link oj.DateTimeRangeValidatorFactory}. 
   * @example <caption>Create a JET datetime validator</caption>
   * var validator = 
   * oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE)
   * .createValidator({{max: oj.IntlConverterUtils.dateToLocalIso(new Date()),
   * min: oj.IntlConverterUtils.dateToLocalIso(new Date(2000, 00, 01)),
   * hint: {'inRange': 'Enter a date that falls in the current millennium.'}}); 
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "VALIDATOR_TYPE_DATETIMERANGE" : 'dateTimeRange',
  
  /**
   * Default type for a factory used to create date restriction validators. This type is passed to 
   * the [Validation.validatorFactory]{@link oj.Validation#validatorFactory} method to retrieve the 
   * dateRestriction validator factory of type {@link oj.DateRestrictionValidatorFactory}.
   * @example <caption>Create a JET dateRestriction validator</caption>
   * var validatorFactory = 
   * oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATERESTRICTION);
   * var options = {dayFormatter : self.aprilFoolsFormatter, // your own formatter code
   * message : {messageDayMetaData : 'You can\'t fool me! Try a different date.'}};
   * var validator = validatorFactory.createValidator(options);
   * 
   * @expose
   * @const
   * @member
   * @type {string}
   */
  "VALIDATOR_TYPE_DATERESTRICTION" : 'dateRestriction',
          
  /**
   * Creates an immutable validator instance of the type the factory implements. 
   * For the specific options parameters, see the Validator jsdoc for the validator you
   * are creating, e.g., oj.RequiredValidator if you are creating a required validator.
   * 
   * @example <caption>create an instance of the required validator using the factory</caption>
   * var rvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED); 
   * var options = {'hint' : 'a value is required for this field'};
   * var requiredValidator = rvf.createValidator(options);
   * @param {(Object|null)} options an object literal containing properties required by the validator 
   * for its initialization. The properties provided in the options is implementation specific.
   * 
   * @return {Object} a validator instance.
   * @throws {TypeError} if an unrecognized type was provided 
   * @ojsignature { target: "Type", 
   *                value: "oj.Validator<V>",
   *                for: "returns"}
   * @instance
   * @expose
   * @see oj.DateRestrictionValidator
   * @see oj.DateTimeRangeValidator
   * @see oj.LengthValidator
   * @see oj.NumberRangeValidator
   * @see oj.RegExpValidator
   * @see oj.RequiredValidator
   */
  createValidator : function(options) {} 

  /**
   * A ducktype for accessing converter or validator created and/ registered by referring to {@link oj.Validation.validatorFactory} or {@link oj.Validation.converterFactory}
   * @typedef {object} oj.Validation.FactoryRegisteredValidatorOrConverter
   * @property {string} type the name of the factory registered
   * @property {Object=} options options to the validator or converter
   */
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Converter Contract
 */

/**
 * Constructs an immutable instance of Converter.
 * 
 * @param {Object=} options an object literal used to provide an optional information to 
 * initialize the converter.<p>
 * @export
 * @ojsignature {target: "Type", value: "class Converter<V>"}
 * @abstract
 * @class
 * @since 0.6
 */
oj.Converter = function(options)
{
  this.Init(options);
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.Converter, oj.Object, "oj.Converter");


/**
 * Initializes converter instance with the set options
 * @param {Object=} options an object literal used to provide an optional information to 
 * initialize the converter.<p>
 * @export
 * @ignore
 */
oj.Converter.prototype.Init = function(options) 
{
  oj.Converter.superclass.Init.call(this);
  // should we make options truly immutable? non-configurable, non-enumerable, non-writable
  // Object.defineProperty(oj.Converter.prototype, "_options", {value: options});
  this._options = options;
};


/**
 * Returns a hint that describes the converter format expected.
 * @method getHint
 * @return {string|null} a hint describing the format the value is expected to be in.
 * @memberof oj.Converter
 * @ojsignature {target: "Type", value: "?(): string|null"}
 * @instance
 */

/**
 * Returns the options called with converter initialization.
 * @return {Object} an object of options.
 * @ojsignature {target: "Type", value: "?(): object"}
 * @export
 */
oj.Converter.prototype.getOptions = function () 
{
  return (this._options || {});
};


/**
 * Parses a String value using the options provided. 
 * @method parse
 * @param {string} value to parse
 * @return {*} the parsed value. 
 * @ojsignature [
 *                {target: "Type", value: "V", for: "returns"}
 *              ]
 * @throws {Error} if parsing fails
 * @memberof oj.Converter
 * @instance
 */

/**
 * Formats the value using the options provided. 
 * 
 * @param {*} value the value to be formatted for display
 * @return {(string|null)} the localized and formatted value suitable for display
 * @throws {Error} if formatting fails.
 * @method format
 * @ojsignature {target: "Type", value: "V", for: "value"}
 * @memberof oj.Converter
 * @instance
 */

/**
 * Returns an object literal with locale and formatting options computed during initialization of 
 * the object. If options was not provided at the time of initialization, the properties will be 
 * derived from the locale defaults.
 * @return {Object} an object of resolved options.
 * @ojsignature {target: "Type", value: "?(): object"}
 * @export
 */
oj.Converter.prototype.resolvedOptions = function ()
{
  var resolved = {};
  // returns a clone of this._options
  $.extend(resolved, this._options);
  
  return resolved;
};


// oj.ConverterError

/**
 * Constructs a ConverterError instance from a summary and detail 
 * 
 * @param {string} summary a localized String that provides a summary of the error
 * @param {string} detail a localized String that provides a detail of the error
 * @constructor
 * @ojtsimport ojmessaging
 * @since 0.6
 * @export
 */
oj.ConverterError = function (summary, detail)
{
  var message = {summary: summary, 
                 detail: detail, 
                 severity: oj.Message.SEVERITY_LEVEL['ERROR']};
  this.Init(message); 
};

oj.ConverterError.prototype = new Error();

/**
 * Initializes the instance. 
 * @param {Object} message an Object that duck-types oj.Message which is an 
 * Object with summary, detail, and severity properties.
 * @export
 * @ignore
 */
oj.ConverterError.prototype.Init = function (message)
{
  var detail = message['detail'];
  var summary = message['summary'];
  this._message = message;

  // so browser can get to e.name and e.message 
  this.name = 'Converter Error';
  this.message = detail || summary;
};

/**
 * Returns an Object that duck-types oj.Message which is an 
 * Object with summary, detail, and severity properties.
 * 
 * @return {Object} an Object that duck-types oj.Message which is an 
 * Object with summary, detail, and severity properties.
 * @ojsignature {target: "Type", value: "oj.Message", for: "returns"}
 * @export
 */
oj.ConverterError.prototype.getMessage = function ()
{
  return this._message;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Validator Contract
 */

/**
 *  @example <caption>Create a Validator and implement its methods.</caption>
 * 
 *  //Validator to ensure that the selected time is a multiple of 15 minute interval.
 *  //This converter takes in a converter option which will be used to format the hint/error
 *  //message shown to the user. It is mandatory to pass the converter option.
 *  var TimeIncrementValidator = function (options) { 
 *    if(options && options.converter){
 *      this._converter = options.converter;
 *    }
 *  };
 *  //Need to be a subclass of oj.Validator
 *  oj.Object.createSubclass(TimeIncrementValidator, oj.Validator, "TimeIncrementValidator");
 * 
 *  // Validates if the passed in value is a multiple of 15 minute interval.
 *  // Throws an error if the validation fails.
 *  TimeIncrementValidator.prototype.validate = function (value)
 *  {
 *    if (value)
 *    {
 *      var currentTime = oj.IntlConverterUtils.isoToLocalDate(value);
 *      var previousValidValue, nextValidValue, sampleMinutes;
 *      var minutes = currentTime.getMinutes();
 *      //Check if the minute is in increment of 15 by taking a modulo
 *      if ((minutes % 15) !== 0)
 *      {
 *        sampleMinute = Math.floor((minutes / 15))*15;
 *        currentTime.setMinutes(sampleMinute);
 *        previousValidValue = oj.IntlConverterUtils.dateToLocalIso(currentTime);
 *        sampleMinute = sampleMinute+ 15;
 *        if(sampleMinute >= 60){
 *          sampleMinute = 0;
 *          currentTime.setTime(currentTime.getTime() + (60*60*1000));
 *        }
 *        currentTime.setMinutes(sampleMinute);
 *        nextValidValue = oj.IntlConverterUtils.dateToLocalIso(currentTime);
 *        throw new oj.ValidatorError("Only multiples of 15 minute intervals are allowed.", "For example, " + this._converter.format(previousValidValue) +
 *                                     " or "+ this._converter.format(nextValidValue));
 *      }
 *    }
 *  };
 *  
 *  //Generates a hint message with 4 different samples of valid values.
 *  TimeIncrementValidator.prototype.getHint = function ()
 *  {
 *    var currentTime = new Date();
 *    var hintMessage = "Only 15 minute intervals accepted, for example";
 *    //generate 4 sample values
 *    for (var i = 0; i < 4; i++) {
 *      currentTime.setMinutes(i * 15);
 *      hintMessage = hintMessage + ", " + this._converter.format(oj.IntlConverterUtils.dateToLocalIso(currentTime));
 *    }
 *    return hintMessage;
 *  };
 * @example <caption>Create a Validator and implement its methods. Bind it to the
 * JET form component which will call this 'validate' to validate the user's input.</caption>
 *  // Validator that ensures endDate is never less than start date
 *  self.endDateValidator = {  
 *    'validate' : function(value)
 *    {
 *      var value =  oj.IntlConverterUtils.isoToLocalDate(value); 
 *      var start = oj.IntlConverterUtils.isoToLocalDate(self.startDate());
 *      if (value && !(value.getFullYear() > start.getFullYear() || 
 *      {
 *        throw new oj.ValidatorError('End Date cannot be less than Start Date');
 *      }
 *    },
 *    'getHint' : function() { return "End Date needs to be greater than Start Date");}
 *  };
 *  -- HTML --
 *  &lt;oj-input-date id="nextday" value="{{endDate}}" 
 *  validators="{{[weekendDateValidator, endDateValidator]}}">&lt;/oj-input-date>
 * @abstract
 * @class
 * @param {Object=} options An object which contains the options for the validator
 * @ojsignature {target: "Type", value: "class Validator<V>"}
 * @export
 * @since 0.6
 * 
 */
oj.Validator = function(options)
{
  this.Init(options);
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.Validator, oj.Object, "oj.Validator");

/**
 * Initializes validator instance with the set options
 * @export
 * @ignore
 */
oj.Validator.prototype.Init = function(options) 
{
  oj.Validator.superclass.Init.call(this);
};


/**
 * Validates the value.
 * The function typically returns true if the validation passes and throws an error if it fails.
 * However some of the implementations (like LengthValidator) returns the original input if the validation is successful.
 * 
 * @example <caption>Create a Validator and implement the validate method. Bind it to the
 * JET form component which will call this 'validate' to validate the user's input.</caption>
 *  // Validator that ensures endDate is never less than start date
 *  self.endDateValidator = {  
 *    'validate' : function(value)
 *    {
 *      var value =  oj.IntlConverterUtils.isoToLocalDate(value); 
 *      var start = oj.IntlConverterUtils.isoToLocalDate(self.startDate());
 *      if (value && !(value.getFullYear() > start.getFullYear() || 
 *      {
 *        throw new oj.ValidatorError('End Date cannot be less than Start Date');
 *      }
 *    },
 *    'getHint' : function() { return "End Date needs to be greater than Start Date");}
 *  };
 *  -- HTML --
 *  &lt;oj-input-date id="nextday" value="{{endDate}}" 
 *  validators="{{[weekendDateValidator, endDateValidator]}}">&lt;/oj-input-date>
 * @param {*} value to be validated
 * @return {*} a boolean true or the original value if validation passes.
 * @throws {Error} if validation fails
 * @method validate
 * @ojsignature [{target: "Type", 
 *                value: "V",
 *                for: "value"},
 *                {target: "Type", 
 *                value: "void",
 *                for: "returns"}]
 * @memberof oj.Validator
 * @instance
 */


/**
 * Returns a hint that describes the validator rule.
 * @example <caption>Create a Validator and implement the getHint method. Bind it to the
 * JET form component which will show the hint.</caption>
 *  // Validator that ensures endDate is never less than start date
 *  self.endDateValidator = {  
 *    'validate' : function(value)
 *    {
 *      ...
 *    },
 *    'getHint' : function() { return "End Date needs to be greater than Start Date");}
 *  };
 *  -- HTML --
 *  &lt;oj-input-date id="nextday" value="{{endDate}}" 
 *  validators="{{[weekendDateValidator, endDateValidator]}}">&lt;/oj-input-date>
 * @returns {string|null} a hint string or null
 * @method getHint
 * @memberof oj.Validator
 * @ojsignature {target: "Type", value: "?(): string|null"}
 * @instance
 */

// ValidatorError

/**
 * Constructs a ValidatorError instance from a summary and detail 
 * 
 * @param {string} summary a localized String that provides a summary of the error
 * @param {string} detail a localized String that provides a detail of the error
 * @example <caption>throw new oj.ValidationError from custom validator's validate method</caption> 
 *  // A custom validator whose validate method ensures that the value is not 'junk'.
 *  self.noJunkValidator = {
 *  'validate' : function (value)
 *  {
 *    value = value + "";
 *    if (value && value.toLowerCase()  === "junk")
 *    {
 *      throw new oj.ValidatorError("Invalid value", "You cannot enter a value that is 'junk'!!");
 *    }
 *    return true;
 *    }
 *  };
 * @constructor
 * @ojtsimport ojmessaging
 * @since 0.6
 * @export
 */
oj.ValidatorError = function (summary, detail)
{
  var message = {summary: summary, 
                 detail: detail, 
                 severity: oj.Message.SEVERITY_LEVEL['ERROR']};
  this.Init(message); 
};

oj.ValidatorError.prototype = new Error();

/**
 * Initializes the instance. 
 * @param {Object} message an Object that duck-types oj.Message which is an 
 * Object with summary, detail, and severity keys. The severity is
 * oj.Message.SEVERITY_LEVEL['ERROR'].
 * @export
 * @ignore
 */
oj.ValidatorError.prototype.Init = function (message)
{
  var detail = message['detail'];
  var summary = message['summary'];
  this._message = message;

  // so browser can get to e.name and e.message 
  this.name = 'Validator Error';
  this.message = detail || summary;
};

/**
 * Returns an Object that duck-types oj.Message which is an 
 * Object with summary, detail, and severity properties. The severity is
 * oj.Message.SEVERITY_LEVEL['ERROR']
 * 
 * @example <caption>get the oj.ValidationError that was thrown and get the message</caption> 
 * var ojmessage = valError.getMessage();
 * var severity = ojmessage['severity'] || oj.Message.SEVERITY_LEVEL['ERROR'];
 * var summary = ojmessage['summary'];
 * var detail = ojmessage['detail'];
 * @returns {Object} an Object that duck-types oj.Message which is an 
 * Object with summary, detail, and severity properties.
 * The severity is
 * oj.Message.SEVERITY_LEVEL['ERROR']
 * @ojsignature {target: "Type", value: "oj.Message", for: "returns"}
 * @export
 */
oj.ValidatorError.prototype.getMessage = function ()
{
  return this._message;
};
/**
 * Copyright (c) 2016, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * oj.ColorConverter Contract. 
 */


(function ()
{

  /**
   * @export
   * @constructor
   * @augments oj.Converter 
   * @name oj.ColorConverter
   * @ojtsimport ojcolor
   * @ojsignature [{target: "Type", 
   *                value: "class ColorConverter extends Converter<oj.Color>"},
   *               {target: "Type",
   *                value: "oj.ColorConverter.ConverterOptions",
   *                for: "options",
   *                jsdocOverride: true}
   *              ]
   
   * @classdesc An {@link oj.Color} object format converter.
   * @desc Creates a Converter that allows any color format to be obtained from an {@link oj.Color} object.
   * @since 0.6
   * @param {Object=} options - an object literal used to provide optional information to 
   * initialize the converter.
   * @example <caption>Create a color converter to convert an rgb specification to hsl format</caption>
   * var cvFactory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_COLOR);
   * var cv        = cv.createConverter({format: "hsl");
   * var color     = new oj.Color("rgb(30, 87, 236)") ;
   * var hsl       = cv.format(color);   -->  "hsl(223, 84%, 52%)"
   * 
   */
  oj.ColorConverter = function (options)
  {
    this.Init(options);
  };

  /**
   * @typedef {object} oj.ColorConverter.ConverterOptions
   * @property {('rgb'|'hsl'|'hsv'|'hex'|'hex3')=} format - sets the format of the converted color specification.
   * Allowed values are "rgb" (the default, if omitted), "hsl", "hsv" "hex", and "hex3". "hex" returns six
   * hex digits ('#rrggbb'), and "hex3" returns three hex digits if possible ('#rgb') or six hex
   * digits if the value cannot be converted to three.
   */

// Subclass from oj.Converter 
  oj.Object.createSubclass(oj.ColorConverter, oj.Converter, "oj.ColorConverter");

  /**
   * Initializes the color converter instance with the set options.
   * @param {Object=} options an object literal used to provide an optional information to 
   * initialize the converter.<p>
   * @export
   * @ignore
   */
  oj.ColorConverter.prototype.Init = function (options)
  {
    options = options || {};
    options["format"] = options["format"] || "rgb";
    oj.ColorConverter.superclass.Init.call(this, options);
  };

  /**
   * Formats the color using the options provided into a string.
   * 
   * @param {oj.Color} color the {@link oj.Color} instance to be formatted to a color specification string
   * @return {(string | null)} the color value formatted to the color specification defined in the options.
   * @throws {Error} a ConverterError if formatting fails, or the color option is invalid.
   * @export
   */
  oj.ColorConverter.prototype.format = function (color)
  {
    var fmt = this._getFormat(),
    ret = null;

    if (fmt === "rgb")
    {
      ret = color.toString();
    }
    else if (fmt === "hsl")
    {
      ret = _toHslString(color);
    }
    else if (fmt === "hex")
    {
      ret = _toHexString(color);
    }
    else if (fmt === "hex3")
    {
      ret = _toHexString(color, true);
    }
    else if (fmt === "hsv")
    {
      ret = _toHsvString(color);
    }
    else
    {
      _throwInvalidColorFormatOption();
    }

    return ret ? ret : oj.ColorConverter.superclass.format.call(this, color);
  };


  /**
   * Parses a CSS3 color specification string and returns an oj.Color object.</br>
   * (Note that the "format" option used to create the Converter is not used
   * by this method, since the oj.Color object created is color agnostic.)
   * @param {string} value The color specification string to parse.
   * @return {oj.Color} the parsed value as an {@link oj.Color} object.
   * @throws {Error} a ConverterError if parsing fails
   * @export
   */
  oj.ColorConverter.prototype.parse = function (value)
  {
    try
    {
      return  new oj.Color(value);   //throws error if invalid
    }
    catch (e)
    {
      _throwInvalidColorSyntax();
    }
  };


  /**
   * Returns a hint that describes the color converter format.
   * @return {string} The expected format of a converted color.
   * @export
   */
  oj.ColorConverter.prototype.getHint = function ()
  {
    return this._getFormat();
  };


  /**
   * Returns an object literal with properties reflecting the color formatting options computed based 
   * on the options parameter.
   * 
   * @ojsignature {target: "Type", for: "returns", value: "oj.ColorConverter.ConverterOptions"}
   * @return {Object} An object literal containing the resolved values for the following options.
   * <ul>
   * <li><b>format</b>: A string value with the format of the color specification.
   * for formatting.</li>
   * </ul>
   * @export
   */
  oj.ColorConverter.prototype.resolvedOptions = function ()
  {
    return {
      "format": this._getFormat()
    };
  };


  /**
   *   @private
   */
  oj.ColorConverter.prototype._getFormat = function ()
  {
    return  oj.ColorConverter.superclass.getOptions.call(this)["format"];
  }



  /**-------------------------------------------------------------*/
  /*   Helpers                                                    */
  /**-------------------------------------------------------------*/


  /**
   *  Converts an oj.Color object to a 3 or 6 hex character string 
   *  @param {Object} color  The oj.Color object to be converted to a hex string.
   *  @param {boolean=} allow3Char  If true the representation is 3 hex characters
   *  (if possible). If false, or omitted, 6 hex characters are used.
   *  @return {string} The hex string representation of the color object.
   *  @private
   */
  function _toHexString(color, allow3Char)
  {
    return '#' + _toHex(color, allow3Char);
  };


  /**
   *  Converts an oj.Color object to an hsl/hsla string 
   *  @param {Object} color  The oj.Color object to be converted to an hsl/hsla string.
   *  @return {string} The hsl/hsla representation of the color object.
   *  @private
   */
  function _toHslString(color)
  {
    var hsl = _rgbToHsl(color._r, color._g, color._b);
    var h = Math.round(hsl.h * 360), s = Math.round(hsl.s * 100), l = Math.round(hsl.l * 100);

    return (color._a == 1) ? "hsl(" + h + ", " + s + "%, " + l + "%)" :
    "hsla(" + h + ", " + s + "%, " + l + "%, " + color._a + ")";
  };

  /**
   *  Converts an oj.Color object to a 3 or 6 hex character string 
   *  @param {Object} color  The oj.Color object to be converted to a hex string.
   *  @param {boolean=} allow3Char  If true the representation is 3 hex characters
   *                   (if possible). If false, or omitted, 6 hex characters are used.
   *  @return {string} The hex string representation of the color object.
   *  @private
   */
  function _toHex(color, allow3Char)
  {
    return _rgbToHex(color._r, color._g, color._b, allow3Char);
  };

  /**
   *  Converts an oj.Color object to an hsv/hsva string 
   *  @param {Object} color  The oj.Color object to be converted to an hsv/hsva string.
   *  @return {string} The hsv/hsva representation of the color object.
   *  @private
   */
  function _toHsvString(color)
  {
    var hsv = _rgbToHsv(color._r, color._g, color._b);

    var h = Math.round(hsv.h * 360),
    s = Math.round(hsv.s * 100),
    v = Math.round(hsv.v * 100);

    return (color._a == 1) ?
    "hsv(" + h + ", " + s + "%, " + v + "%)" :
//   "hsva(" + h + ", " + s + "%, " + v + "%, "+ this._roundA + ")";
    "hsva(" + h + ", " + s + "%, " + v + "%, " + color._a + ")";
  };


  /**
   * Converts RGB color values to hex
   * @param {number} r the red value in the set [0,255]
   * @param {number} g the green value in the set [0,255]
   * @param {number} b the blue value in the set [0,255]
   * @param {boolean=} allow3Char  If true the representation is 3 hex characters
   *                   (if possible). If false, or omitted, 6 hex characters are used.
   * @returns {string} a 3 or 6 hex character string.
   * @private
   */
  function _rgbToHex(r, g, b, allow3Char)
  {
    var hex = [
      _pad2(Math.round(r).toString(16)),
      _pad2(Math.round(g).toString(16)),
      _pad2(Math.round(b).toString(16))
    ];

    // Return a 3 character hex if possible
    if (allow3Char && hex[0].charAt(0) == hex[0].charAt(1) &&
    hex[1].charAt(0) == hex[1].charAt(1) && hex[2].charAt(0) == hex[2].charAt(1))
    {
      return hex[0].charAt(0) + hex[1].charAt(0) + hex[2].charAt(0);
    }

    return hex.join("");
  };


  /**
   * Converts an RGB color value to HSL.
   * Handle bounds/percentage checking to conform to CSS color spec, and returns
   * an object containg the h,s,l values.
   * <http://www.w3.org/TR/css3-color/>
   * Assumes:  r, g, b in [0, 255] or [0, 1]
   * @param {number} r the red value
   * @param {number} g the green value
   * @param {number} b the blue value
   * @returns {Object} Object with properties h, s, l, in [0, 1].
   * @private
   */
  function _rgbToHsl(r, g, b)
  {
    r = _bound01(r, 255);
    g = _bound01(g, 255);
    b = _bound01(b, 255);

    var max = Math.max(r, g, b), min = Math.min(r, g, b);
    var h, s, l = (max + min) / 2;

    if (max == min)
    {
      h = s = 0; // achromatic
    }
    else
    {
      var d = max - min;
      s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
      switch (max)
      {
        case r:
          h = (g - b) / d + (g < b ? 6 : 0);
          break;
        case g:
          h = (b - r) / d + 2;
          break;
        case b:
          h = (r - g) / d + 4;
          break;
      }

      h /= 6;
    }

    return {h: h, s: s, l: l};
  };

  /**
   * Converts an RGB color value to HSV.
   * Handle bounds/percentage checking to conform to CSS color spec, and returns
   * an object containg the h,s,v values.
   * <http://www.w3.org/TR/css3-color/>
   * Assumes:  r, g, and b are contained in the set [0,255] or [0,1]
   * @param {number} r the red value
   * @param {number} g the green value
   * @param {number} b the blue value
   * @returns {Object} Object with properties h, s, v, in [0,1].
   * @private
   */
  function _rgbToHsv(r, g, b)
  {
    r = _bound01(r, 255);
    g = _bound01(g, 255);
    b = _bound01(b, 255);

    var max = Math.max(r, g, b), min = Math.min(r, g, b);
    var h, s, v = max;

    var d = max - min;
    s = max === 0 ? 0 : d / max;

    if (max == min)
    {
      h = 0; // achromatic
    }
    else
    {
      switch (max)
      {
        case r:
          h = (g - b) / d + (g < b ? 6 : 0);
          break;
        case g:
          h = (b - r) / d + 2;
          break;
        case b:
          h = (r - g) / d + 4;
          break;
      }
      h /= 6;
    }
    return {h: h, s: s, v: v};
  };


  /**
   * Converts an RGBA color plus alpha transparency to hex
   * Assumes r, g, b and a are contained in the set [0, 255]
   * @param {number} r the red value in the set [0, 255]
   * @param {number} g the green value in the set [0, 255]
   * @param {number} b the blue value in the set [0, 255]
   * @param {number} a the alpha value in the set [0,1]
   * Returns an 8 character hex string
   */
  /*    NOT USED currently
   function rgbaToHex(r, g, b, a)
   {
   var hex = [
   pad2(convertDecimalToHex(a)),
   pad2(mathRound(r).toString(16)),
   pad2(mathRound(g).toString(16)),
   pad2(mathRound(b).toString(16))
   ];
   
   return hex.join("");
   }
   */


  /**
   * Take input from [0, n] and return it as [0, 1]
   * @private
   */
  function _bound01(n, max)
  {
    if (_isOnePointZero(n))
    {
      n = "100%";
    }

    var processPercent = _isPercentage(n);
    n = Math.min(max, Math.max(0, parseFloat(n)));

    // Automatically convert percentage into number
    if (processPercent)
    {
      n = parseInt(n * max, 10) / 100;
    }

    // Handle floating point rounding errors
    if ((Math.abs(n - max) < 0.000001))
    {
      return 1;
    }

    // Convert into [0, 1] range if it isn't already
    return (n % max) / parseFloat(max);
  };


  /**
   *   Need to handle 1.0 as 100%, since once it is a number, there is no difference between it and 1
   *   <http://stackoverflow.com/questions/7422072/javascript-how-to-detect-number-as-a-decimal-including-1-0>
   *   @private
   */
  function _isOnePointZero(n)
  {
    return typeof n == "string" && n.indexOf('.') != -1 && parseFloat(n) === 1;
  };

  /**
   *  Check to see if string passed in is a percentage
   *  @param {string}  n  The number string
   *  @return {boolean}  True if the string contains a '%' character.
   *  @private
   */
  function _isPercentage(n)
  {
    return typeof n === "string" && n.indexOf('%') != -1;
  }
  ;


  /**
   *  Force a hex value string to have 2 characters by inserting a preceding zero
   *  if neccessary.  e.g. 'a' -> '0a'
   *  @param {string} c  The hex character(s) to be tested.
   *  @return {string} A two character hex string.
   *  @private
   */
  function _pad2(c)
  {
    return c.length == 1 ? '0' + c : '' + c;
  }
  ;


  /*
   *   Throw an invalid color specfication error.
   */
  function _throwInvalidColorSyntax()
  {
    var summary, detail, ce;

//  summary =  oj.Translations.getTranslatedString("oj-converter.color.invalidFormat.summary") ;
//  detail  = oj.Translations.getTranslatedString("oj-converter.color.invalidFormat.detail") ;
    summary = "Invalid color specification";
    detail = "Color specification does not conform to CSS3 standard";

    ce = new oj.ConverterError(summary, detail);

    throw ce;
  }
  ;


  /*
   *   Throw an invalid converter specfication error.
   */
  function _throwInvalidColorFormatOption()
  {
    var summary, detail, ce;

// summary =  oj.Translations.getTranslatedString("oj-converter.color.invalidFormat.summary") ;
// detail  = oj.Translations.getTranslatedString("oj-converter.color.invalidFormat.detail") ;
    summary = "Invalid color format";
    detail = "Invalid color format option specification";
    ce = new oj.ConverterError(summary, detail);

    throw ce;
  }
  ;


})();





/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @export
 * @class oj.IntlConverterUtils
 * @classdesc Utility function for converters
 * @hideconstructor
 * @since 0.7
 */
oj.IntlConverterUtils = {};

/**
 * Parses the isoString and returns a JavaScript Date object
 * 
 * @export
 * @param {string} isoString isoString to parse and to return Date of
 * @return {Date} the parsed JavaScript Date Object
 */
oj.IntlConverterUtils.isoToDate = function(isoString) 
{
  return oj.OraI18nUtils.isoToDate(isoString);
};

/**
 * Returns a local Date object provided a local isoString
 * 
 * @param {string} isoString date in isoString format
 * @returns {Date} localDate 
 * @export
 * @since 0.7
 */
oj.IntlConverterUtils.isoToLocalDate = function(isoString) 
{
  return oj.OraI18nUtils.isoToLocalDate(isoString);
};

/**
 * Returns a local isoString provided a Date object
 * 
 * @param {Date} date
 * @returns {string} isoString 
 * @export
 * @since 0.7
 */
oj.IntlConverterUtils.dateToLocalIso = function(date) 
{
  return oj.OraI18nUtils.dateToLocalIso(date);
};

oj.IntlConverterUtils._getTimeZone = function(isoString) 
{
  return oj.OraI18nUtils._getTimeZone(isoString);
};

/**
 * Returns the timezone offset between UTC and the local time in Etc/GMT[+-]hh:mm syntax.
 * The offset is positive if the local timezone is behind UTC and negative if
 * it is ahead. The offset range is between Etc/GMT-14 and Etc/GMT+12 (UTC-12 and UTC+14)
 * 
 * @example <caption>The local time is UTC-7 (Pacific Daylight Time)</caption>
 * oj.IntlConverterUtils.getLocalTimeZoneOffset() will return the string "Etc/GMT+07:00" 
 *
 * @example <caption>The local time is UTC+1 (Central European Standard Time)</caption>
 * oj.IntlConverterUtils.getLocalTimeZoneOffset() will return the string "Etc/GMT-01:00" 
 * 
 * @export 
 * @returns {string}
 */
oj.IntlConverterUtils.getLocalTimeZoneOffset = function() 
{
  return oj.OraI18nUtils.getLocalTimeZoneOffset();
};

/**
 * Given either an Object literal representing a 'converter' option (used in components) or a 
 * converter instance of type oj.Converter, this method returns the converter instance.
 * You can also pass in a string. In this case, it will return you an instance of a converter
 * registered with that type. 
 * 
 * @param {string| Object} converterOption
 * @returns {Object|null} converterInstance or null if a converter cannot be determined 
 * @ojsignature {
 *        target: "Type", value: "<T>(converterOption: string|oj.Validation.FactoryRegisteredValidatorOrConverter|oj.Converter<T>): oj.Converter<T>|null"}
 * @export
 * @since 0.6
 */
oj.IntlConverterUtils.getConverterInstance = function (converterOption)
{
  var cTypeStr = "", cOptions = {}, converterInstance = null, cf;
  
  if (converterOption)
  {
    if (typeof converterOption === "object")
    {
      // TODO: Should we check that it duck types oj.Converter?
      if (converterOption instanceof oj.Converter || 
          (converterOption['parse'] && typeof converterOption['parse'] === "function") || 
          (converterOption['format'] && typeof converterOption['format'] === "function"))
      {
        // we are dealing with a converter instance
        converterInstance = converterOption;
      }
      else 
      {
        // check if there is a type set
        cTypeStr = converterOption['type'];
        cOptions = converterOption['options'] || {};
      }
    }

    if (!converterInstance)
    {
      // either we have an object literal or just plain string.
      cTypeStr = cTypeStr || converterOption;
      if (cTypeStr && typeof cTypeStr === "string")
      {
        // if we are passed a string get registered type. 
        cf = oj.Validation.converterFactory(cTypeStr);
        converterInstance = cf.createConverter(cOptions);
      }
    }
  }
  
  return converterInstance;
};

/**
 * So the requirement is if min or max lacks date portion and value contains it, then min + max should use 
 * value's date portion
 * 
 * @param {string} minMax date in isoString format
 * @param {string} value date in isoString format
 * @returns {string} merged date in isoString format
 * @export
 * @ignore
 * @since 1.2
 */
oj.IntlConverterUtils._minMaxIsoString = function(minMax, value) 
{
  if(minMax) 
  {
    value = value || this.dateToLocalIso(new Date());
    
    var vTindex = value.indexOf("T");

    if(minMax.indexOf("T") === 0 && vTindex > 0) 
    {
      //meaning only time exists for minMax and value contains date
      minMax = value.substring(0, vTindex) + minMax;
    }
  }
  
  return minMax;
};

// PACKAGE PRIVATE

/**
 * Processes an converter option error and returns a oj.ConverterERror instance.
 * @param {string} errorCode
 * @param {Object} parameterMap
 * @return {Object} an oj.ConverterError instance
 * @private
 */
oj.IntlConverterUtils.__getConverterOptionError = function(errorCode, parameterMap)
{
  oj.Assert.assertObject(parameterMap);
  var summary = "", detail = "", propName = parameterMap['propertyName'], reqPropName, 
          propValueValid;
  
  if (errorCode === "optionTypesMismatch")
  {
    reqPropName = parameterMap['requiredPropertyName'];
    propValueValid = parameterMap['requiredPropertyValueValid'];
    // Summary: A value for the property '{requiredPropertyName}' is required when the property 
    // '{propertyName}' is set to '{propertyValue}'.
    summary = oj.Translations.getTranslatedString("oj-converter.optionTypesMismatch.summary", 
      {'propertyName': propName,
       'propertyValue': parameterMap['propertyValue'],
       'requiredPropertyName': reqPropName});

    detail = oj.IntlConverterUtils._getOptionValueDetailMessage(reqPropName, propValueValid);
  }
  else if (errorCode === "optionTypeInvalid")
  {
    // Summary: A value of the expected type was not provided for '{propertyName}'.
    propName = parameterMap['propertyName'];
    propValueValid = parameterMap['propertyValueValid'];
    summary = oj.Translations.getTranslatedString("oj-converter.optionTypeInvalid.summary", 
      {'propertyName': propName});

    detail = oj.IntlConverterUtils._getOptionValueDetailMessage(propName, propValueValid);
  }
  else if (errorCode === "optionOutOfRange")
  {
    // The value {propertyValue} is out of range for the option '{propertyName}'.
    summary = oj.Translations.getTranslatedString("oj-converter.optionOutOfRange.summary", 
      {'propertyName': propName,
       'propertyValue': parameterMap['propertyValue']});

    propValueValid = parameterMap['propertyValueValid'];
    detail = oj.IntlConverterUtils._getOptionValueDetailMessage(propName, propValueValid);
  }
  else if (errorCode === "optionValueInvalid")
  {
    // An invalid value '{propertyValue}' was specified for the option '{propertyName}'.. 
    summary = oj.Translations.getTranslatedString("oj-converter.optionValueInvalid.summary", 
      {'propertyName': propName,
       'propertyValue': parameterMap['propertyValue']});
    
    propValueValid = parameterMap['propertyValueHint'];
    detail = oj.IntlConverterUtils._getOptionValueDetailMessage(propName, propValueValid);
  }
  
  return new oj.ConverterError(summary, detail);

};

/**
 * Builds the detail message for possible converter option values. Only applicable when errorInfo is 
 * returned from JET converter implementation.
 * 
 * @param {string} propName name of the property 
 * @param {Array|string} propValueValid valid value or values expected.
 * 
 * @return {string} the localized message
 * @private
 */
oj.IntlConverterUtils._getOptionValueDetailMessage = function (propName, propValueValid)
{
  // Detail: An accepted value for '{propertyName}' is '{propertyValueValid}'. or 
  // Accepted values for '{propertyName}' are '{propertyValueValid}'.
  var resourceKey;
  
  if (propValueValid)
  {
    if (typeof propValueValid === "string")
    {
      resourceKey = "oj-converter.optionHint.detail";
    }
    else
    {
      // we have an array of values
      resourceKey = "oj-converter.optionHint.detail-plural";
      propValueValid = 
         propValueValid.join(oj.Translations.getTranslatedString("oj-converter.plural-separator"));
    }
    return oj.Translations.getTranslatedString(resourceKey, 
      {'propertyName': propName,
       'propertyValueValid': propValueValid});

  }
  
  return "";
};

/**
 * Returns the default value for non-truthy values.
 * 
 * @returns {string} an empty string
 * @private
 */
oj.IntlConverterUtils.__getNullFormattedValue = function ()
{
  return "";
};

/**
 * Will return an updated toIsoString using the timePortion from the fromIsoString or from the default 
 * oj.OraI18nUtils.DEFAULT_TIME_PORTION
 * 
 * @private
 * @expose
 * @param {string} fromIsoString isoString that may not be a complete isoString
 * @param {string} toIsoString isoString that may not be a complete isoString
 * @returns {string} modified toIsoString with original date portion and the time portion from the fromIsoString
 * @since 1.1
 */
oj.IntlConverterUtils._copyTimeOver = function(fromIsoString, toIsoString) 
{
  return oj.OraI18nUtils._copyTimeOver(fromIsoString, toIsoString);
};

/**
 * Clears the time portion of the isoString
 * 
 * @private
 * @expose
 * @param {string} isoString isoString that may not be a complete isoString
 * @returns {string} an updated isoString
 * @since 1.1
 */
oj.IntlConverterUtils._clearTime = function(isoString) 
{
  return oj.OraI18nUtils._clearTime(isoString);
};

/**
 * Will accept an isoString and perform a get operation or a set operation depending on whether param is an Array 
 * or a JSON 
 * 
 * The keys for the get and set operation are defined in oj.OraI18nUtils's _DATE_TIME_KEYS.
 * 
 * Note the handling of month starting with 0 in Date object and being 1 based in isoString will be handled by the function 
 * with the usage of doParseValue. Meaning when you doParseValue and you are getting the value it will automatically 
 * decrement the value and when you are setting the param it will check if the value is of number and if so will 
 * increment it.
 * 
 * @private
 * @expose
 * @param {string} isoString isoString that may not be a complete isoString
 * @param {Array|Object} actionParam if an Array will be a get operation, if a JSON will be a set operation
 * @param {boolean=} doParseValue whether one should parseInt the value during the get request
 * @returns {Object|string} an Object when a get operation and a string when a set operation
 * @since 1.1
 */
oj.IntlConverterUtils._dateTime = function(isoString, actionParam, doParseValue) 
{
  return oj.OraI18nUtils._dateTime(isoString, actionParam, doParseValue);
};

/**
 * So the problem is Jet uses incomplete isoString which causes issues in different browsers. 
 * 
 * For instance for a new Date().toISOString() => 2015-02-02T18:00:37.007Z
 * ojInputDate stores 2015-02-02
 * ojInputTime stores T18:00:37.007Z
 * 
 * yet constructing new Date(val) on above causes different results or errors in different browsers, so 
 * this function is to normalize them. Note it is assumed that the point is creating the Date object from the 
 * normalized isoString. Meaning if both contain only the time portion today's date will appended to it.
 * 
 * @private
 * @expose
 * @param {string} isoString isoString that may not be a complete isoString
 * @returns {string} a normalized isoString
 * @since 1.1
 */
oj.IntlConverterUtils._normalizeIsoString = function(isoString) 
{
  return oj.OraI18nUtils._normalizeIsoString(isoString);
};

/**
 * Returns a person's initials
 * 
 * @param {string=} firstName first name
 * @param {string=} lastName last name
 * @returns {string|undefined} uppercase concatenation of first letter of first name and first letter
 * of last name.
 * There are the following special cases:
 * - If the name is Arabic characters, it returns empty string.
 * - If the name is Hindi characters, it returns the first letter of the first name.
 * - If the name is Thai characters, it returns the first letter of the first name.
 * - If the name is Korean characters, it returns the first name.
 * - If the name is Japanese or Chinese characters, it returns the last name.
 * 
 * @export  
 * @since 4.0.0
 */
oj.IntlConverterUtils.getInitials = function(firstName, lastName) 
{
  return oj.OraI18nUtils.getInitials(firstName, lastName);
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Constructs a LengthValidator that ensures the value entered is within a given length.
 * <p> 
 * By default this uses Javascript's String length property 
 * which counts a UTF-16 surrogate pair as length === 2. 
 * If you need to count surrogate pairs as length === 1, then set the 
 * <code class="prettyprint">countBy</code> option to <code class="prettyprint">"codePoint"</code>
 * or use
 * <code class="prettyprint">oj.LengthValidator.defaults.countBy = "codePoint";</code>
 * to set the page-wide default.
 * </p>
 * <p> 
 * You can customize the default messages of all validators including this one
 * using the messageDetail and messageSummary options.
 * <p>
 * @param {Object=} options an object literal used to provide:<p>
 * @export
 * @constructor
 * @augments oj.Validator
 * @ojsignature [{target: "Type", 
 *                value: "class LengthValidator extends Validator<number|string>"},
 *               {target: "Type",
 *                value: "oj.LengthValidator.ValidatorOptions",
 *                for: "options",
 *                jsdocOverride: true}
 *              ]
 * @since 0.7
 */
oj.LengthValidator = function (options)
{
  this.Init(options);
};

/**
 * @typedef {object} oj.LengthValidator.ValidatorOptions
 * @property {('codeUnit'|'codePoint')=} countBy - A string that specifies how to count the length. Valid values are
 * <code class="prettyprint">"codeUnit"</code> and <code class="prettyprint">"codePoint"</code>.
 * Defaults to <code class="prettyprint">oj.LengthValidator.defaults.countBy</code> which defaults
 * to <code class="prettyprint">"codeUnit"</code>.<br/>
 * <code class="prettyprint">"codeUnit"</code> uses javascript's length function which counts the 
 * number of UTF-16 code units. Here a Unicode surrogate pair has a length of two. <br/>
 * <code class="prettyprint">"codePoint"</code> 
 * counts the number of Unicode code points. 
 * Here a Unicode surrogate pair has a length of one.<br/>
 * @property {number=} min - a number 0 or greater that is the minimum length of the value.
 * @property {number=} max - a number 1 or greater that is the maximum length of the value.
 * @property {Object=} hint - an optional object literal of hints to be used. 
 * <p>The hint strings (e.g., hint.min) are  passed as the 'pattern' parameter to
 * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
 * that documentation, if you are using a reserved character, you need to escape it with 
 * a dollar character ('$').
 * </p> 
 * @property {string=} hint.max - a hint message to be used to indicate the allowed maximum. 
 * When not present, the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.hint.max</code>.<p>
 * Tokens: <br/>
 * {max} - the maximum<p>
 * Usage: <br/>
 * Enter {max} or fewer characters
 * @property {string=} hint.min - a hint message to be used to indicate the allowed minimum. 
 * When not present, the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.hint.min</code>.<p>
 * Tokens: <br/>
 * {min} the minimum<p>
 * Usage: <br/>
 * Enter {min} or more characters 
 * @property {string=} hint.inRange - a hint message to be used to indicate the allowed range. 
 * When not present, the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.hint.inRange</code>.<p>
 * Tokens: <br/>
 * {min} the minimum<p>
 * {max} - the maximum<p>
 * Usage: <br/>
 * Enter between {min} and {max} characters
 * @property {string=} hint.exact - a hint message to be used, to indicate the exact length. 
 * When not present, the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.hint.exact</code>.<p>
 * Tokens: <br/>
 * {length} the length<p>
 * Usage: <br/>
 * Enter {length} characters
 * @property {Object=} messageDetail - an optional object literal of custom error messages to 
 * be used.
 * <p>The messageDetail strings (e.g., messageDetail.tooLong) are  passed as the 'pattern' 
 * parameter to [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
 * that documentation, if you are using a reserved character, you need to escape it with 
 * a dollar character ('$').
 * </p> 
 * @property {string=} messageDetail.tooLong - the detail error message to be used as the error 
 * message, when the length of the input value exceeds the maximum value set. When not present, the 
 * default detail message is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.messageDetail.tooLong</code>.<p>
 * Tokens:<br/>
 * {value} - value entered by the user<br/>
 * {max} - the maximum allowed value<p>
 * Usage: <br/>
 * The {value} has too many characters. Enter {max} or fewer characters, not more.
 * @property {string=} messageDetail.tooShort - the detail error message to be used as the error 
 * message, when the length of the input value is less the minimum value set. When not present, the 
 * default detail message is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.messageDetail.tooShort</code>.<p>
 * Tokens:<br/>
 * {value} - value entered by the user<br/>
 * {min} - the minimum allowed value<p>
 * Usage: <br/>
 * The {value} has too few characters. Enter {min} or more characters, not less.
 * @property {Object=} messageSummary - optional object literal of custom error summary message 
 * to be used. 
 * <p>The messageSummary strings (e.g., messageSummary.tooLong) are  passed as the 'pattern' 
 * parameter to [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
 * that documentation, if you are using a reserved character, you need to escape it with 
 * a dollar character ('$').
 * </p> 
 * @property {string=} messageSummary.tooLong - the message to be used as the summary error 
 * message, when the length of the input value exceeds the maximum value set. When not present, the 
 * default message summary is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.messageSummary.tooLong</code>.
 * @property {string=} messageSummary.tooShort - the message to be used as the summary error 
 * message, when input value is less than the set minimum value. When not present, the default 
 * message summary is the resource defined with the key 
 * <code class="prettyprint">oj-validator.length.messageSummary.tooShort</code>.
 */

/**
 * The set of attribute/value pairs that serve as default values 
 * when new LengthValidator objects are created.
 * <p>
 * LengthValidator's <code class="prettyprint">countBy</code> option may be changed
 * for the entire application after the 'ojs/ojvalidation' module is loaded 
 * (each form control module includes the 'ojs/ojvalidation' module). If the 
 * <code class="prettyprint">options.countBy</code> is specifically set, 
 * it will take precedence over this default.
 * </p>
 * <p>For example:
 * <code class="prettyprint">
 * oj.LengthValidator.defaults.countBy = 'codePoint';
 * </code></p>
 * @property {string} countBy count the length by <code class="prettyprint">"codeUnit"</code> or 
 * <code class="prettyprint">"codePoint"</code>. 
 * Defaults to <code class="prettyprint">"codeUnit"</code>.
 * @export
 * @memberof oj.LengthValidator
 * @since 2.1.0
 */
oj.LengthValidator.defaults =
{
  'countBy': 'codeUnit'
};

// Subclass from oj.Validator 
oj.Object.createSubclass(oj.LengthValidator, oj.Validator, "oj.LengthValidator");

/**
 * Initializes validator instance with the set options
 * @param {Object=} options
 * @ignore
 */
oj.LengthValidator.prototype.Init = function (options)
{
  var countByOptions = options["countBy"];

  oj.LengthValidator.superclass.Init.call(this);

  this._min = options["min"] !== undefined ? parseInt(options["min"], 10) : null;
  this._max = options["max"] !== undefined ? parseInt(options["max"], 10) : null;
  
  // check that the min/max make sense, otherwise throw an error
  if (isNaN(this._min))
    throw new Error("length validator's min option is not a number. min option is " + options["min"]);
  if (isNaN(this._max))
    throw new Error("length validator's max option is not a number. max option is " + options["min"]);
  if (this._min !== null && this._min < 0)
    throw new Error("length validator's min option cannot be less than 0. min option is " + 
      options["min"]);
  if (this._max !== null && this._max < 1)
    throw new Error("length validator's max option cannot be less than 1. max option is " + 
      options["max"]);
  
  this._countBy = (countByOptions === undefined) ?
  oj.LengthValidator.defaults['countBy'] : countByOptions;

  if (options)
  {
    this._hint = options['hint'] || {};
    this._customMessageSummary = options['messageSummary'] || {};
    this._customMessageDetail = options['messageDetail'] || {};

  }
};

/**
 * A message to be used as hint, when giving a hint about the expected length. There is no default 
 * hint for this property.
 * 
 * @returns {string|null} a hint message or null if no hint is available in the options
 * @memberof oj.LengthValidator
 * @instance
 * @export
 */
oj.LengthValidator.prototype.getHint = function ()
{
  var hint = null;
  var hints = this._hint;
  var hintExact = hints["exact"];
  var hintInRange = hints["inRange"];
  var hintMaximum = hints["max"];
  var hintMinimum = hints["min"];

  var max = this._max;
  var min = this._min;
  var params;
  var translations = oj.Translations;

  if (min !== null && max !== null)
  {
    if (min !== max)
    {
      params = {"min": min, "max": max};
      hint = hintInRange ? translations.applyParameters(hintInRange, params) :
      translations.getTranslatedString('oj-validator.length.hint.inRange', params);
    }
    else
    {
      params = {'length': min};
      hint = hintExact ? translations.applyParameters(hintExact, params) :
      translations.getTranslatedString('oj-validator.length.hint.exact', params);
    }
  }
  else if (min !== null)
  {
    params = {"min": min};
    hint = hintMinimum ? translations.applyParameters(hintMinimum, params) :
    translations.getTranslatedString('oj-validator.length.hint.min', params);
  }
  else if (max !== null)
  {
    params = {"max": max};
    hint = hintMaximum ? translations.applyParameters(hintMaximum, params) :
    translations.getTranslatedString('oj-validator.length.hint.max', params);
  }

  return hint;
};

/**
 * Validates the length of value is greater than minimum and/or less than maximum.
 *
 * @param {string|number} value that is being validated
 * @returns {string} original if validation was successful
 * @ojsignature {target: "Type", for: "returns",
 *                value: "void"}
 * @throws {Error} when the length is out of range.
 * @export
 */
oj.LengthValidator.prototype.validate = function (value)
{

  var customMessageDetail = this._customMessageDetail;
  var customMessageSummary = this._customMessageSummary;
  var detail = "";
  var length;
  var max = this._max;
  var messageSummaryTooLong = customMessageSummary["tooLong"];
  var messageSummaryTooShort = customMessageSummary["tooShort"];
  var messageTooLong = customMessageDetail["tooLong"];
  var messageTooShort = customMessageDetail["tooShort"];
  var min = this._min;
  var params;
  var string;
  var summary = "";
  var translations = oj.Translations;

  string = "" + value;
  length = this._getLength(string);

  // If only min is set and length is at least min, or 
  // if only max is set and length is at most max, or
  // if length is between min and max or
  // if neither min or max is set return with no error.
  if ((min === null || length >= this._min) &&
  ((max === null) || (length <= this._max)))
  {
    return string;
  }
  else
  {
    if (length < this._min) //too short
    {
      params = {"value": value, "min": min};
      summary = messageSummaryTooShort ? translations.applyParameters(messageSummaryTooShort, params) :
      translations.getTranslatedString('oj-validator.length.messageSummary.tooShort');
      detail = messageTooShort ? translations.applyParameters(messageTooShort, params) :
      translations.getTranslatedString('oj-validator.length.messageDetail.tooShort', params);
    }
    else // too long
    {
      params = {"value": value, "max": max};
      summary = messageSummaryTooLong ? translations.applyParameters(messageSummaryTooLong, params) :
      translations.getTranslatedString('oj-validator.length.messageSummary.tooLong');
      detail = messageTooLong ? translations.applyParameters(messageTooLong, params) :
      translations.getTranslatedString('oj-validator.length.messageDetail.tooLong', params);
    }

    throw new oj.ValidatorError(summary, detail);
  }
};

/**
 * @returns {number} the length of the text counted by UTF-16 codepoint
 *  or codeunit as specified in the countBy option.
 * @private
 */
oj.LengthValidator.prototype._getLength = function (text)
{

  var countBy = this._countBy.toLowerCase();
  var codeUnitLength = text.length;
  var i;
  var length;
  var surrogateLength = 0;

  switch (countBy)
  {
    case "codepoint" :
      // if countBy is "codePoint", then count supplementary characters as length of one
      // For UTF-16, a "Unicode  surrogate pair" represents a single supplementary character. 
      // The first (high) surrogate is a 16-bit code value in the range U+D800 to U+DBFF. 
      // The second (low) surrogate is a 16-bit code value in the range U+DC00 to U+DFFF.
      // This code figures out if a charCode is a high or low surrogate and if so, 
      // increments surrogateLength
      for (i = 0; i < codeUnitLength; i++)
      {
        if ((text.charCodeAt(i) & 0xF800) === 0xD800)
        {
          surrogateLength++;
        }
      }
      // e.g., if the string is two supplementary characters, codeUnitLength is 4, and the 
      // surrogateLength is 4, so we will return two.
      oj.Assert.assert(surrogateLength%2 === 0, 
        "the number of surrogate chars must be an even number.");
      length = (codeUnitLength - surrogateLength / 2);
      break;
    case "codeunit" :
    default :
      // Javascript's length function counts # of code units. 
      // A supplementary character has a length of 2 code units.
      length = codeUnitLength;
  }
  return length;
};
/**
 * Copyright (c) 2008, 2013, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Constructs a RegExpValidator that ensures the value matches the provided pattern
 * @param {Object=} options an object literal used to provide the pattern, an optional hint and error 
 * message.
 * @export
 * @constructor
 * @augments oj.Validator 
 * @ojsignature [{target: "Type", 
 *                value: "class RegExpValidator extends Validator<string|number>"},
 *               {target: "Type",
 *                value: "oj.RegExpValidator.ValidatorOptions",
 *                for: "options", jsdocOverride: true}
 *              ]
 * @since 0.6
 */
oj.RegExpValidator = function (options)
{
  this.Init(options);
};

// Subclass from oj.Object or oj.Validator. It does not matter
oj.Object.createSubclass(oj.RegExpValidator, oj.Validator, "oj.RegExpValidator");

// key to access required validator specific resources in the bundle 
oj.RegExpValidator._BUNDLE_KEY_DETAIL = "oj-validator.regExp.detail";
oj.RegExpValidator._BUNDLE_KEY_SUMMARY = "oj-validator.regExp.summary";

/**
  * @typedef {object} oj.RegExpValidator.ValidatorOptions
  * @property {string=} pattern - a regexp pattern that the validator matches a value against.<p>
  * Example:<br/>
  * '\\d{10}'
  * @property {string=} hint - an optional hint text. There is no default hint provided by the 
  * validator. It is generally not recommended to show the actual pattern in the hint as it might be 
  * confusing to end-user, but if you do, you can use the {pattern} token.<p>
  * <p>The hint string is passed as the 'pattern' parameter to
  * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
  * that documentation, if you are using a reserved character, you need to escape it with 
  * a dollar character ('$').
  * </p>
  * Tokens: <br/>
  * {pattern} - the pattern to enforce<p>
  * Example:<br/>
  * "value must meet this pattern {pattern}" 
  * @property {string=} messageSummary - a custom error message summarizing the error when the 
  * users input does not match the specified pattern. When not present, the default summary is the 
  * resource defined with the key <code class="prettyprint">oj-validator.regExp.summary</code>. 
  * It is generally not recommended to show the actual pattern in the message as it might be 
  *  confusing to end-user. <p>
  * <p>The messageSummary string is passed as the 'pattern' parameter to
  * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
  * that documentation, if you are using a reserved character, you need to escape it with 
  * a dollar character ('$').
  * </p>
  * Tokens: <p>
  * {label} - label of the component for which this message applies. The label may not always be 
  * available depending on the usage of the validator. <br/>
  * {pattern} - the pattern the value should match<br/>
  * {value} - value entered by user<p>
  * Examples:<br/>
  * "'{label}' Format Incorrect" // translating to 'Phone Number' Format Incorrect
  * @property {string=} messageDetail - a custom error message to be used for creating detail 
  * part of message, when the users input does not match the specified pattern. When not present, the 
  * default detail message is the resource defined with the key 
  * <code class="prettyprint">oj-validator.regExp.detail</code>.<p>
  * <p>The messageDetail string is passed as the 'pattern' parameter to
  * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
  * that documentation, if you are using a reserved character, you need to escape it with 
  * a dollar character ('$').
  * </p>
  * Tokens:<br/>
  * {label} - label text of the component for which this message applies. <br/>
  * {pattern} the 'pattern' that the value should match <br/>
  * {value} value entered by the user <p>
  * Examples:<br/>
  * "The value {value} must contain at least 3 alphanumeric characters"<br/>
  */
/**
 * Initializes validator instance with the set options
 * @param {Object=} options
 * @memberof oj.RegExpValidator
 * @instance
 * @ignore
 */
oj.RegExpValidator.prototype.Init = function (options)
{
  oj.RegExpValidator.superclass.Init.call(this);
  this._options = options;
};

/**
 * Validates value for matches using the regular expression provided by the pattern. This method 
 * does not raise an error when value is the empty string or null; the method returns true indicating
 * that the validation was successful. If the application wants the empty string to fail validation, 
 * then the application should chain in the required validator (e.g., set required on the input). 
 *  
 * @param {string|number} value that is being validated 
 * @returns {boolean} true if validation was successful
 * @ojsignature {target: "Type", for: "returns",
 *                value: "void"}
 * 
 * @throws {Error} when there is no match
 * @memberof oj.RegExpValidator
 * @instance
 * @export
 */
oj.RegExpValidator.prototype.validate = function (value)
{
  var detail;
  var label;
  var params;
  var pattern = (this._options && this._options['pattern']) || "";

  var summary;


  // don't validate null or empty string; per 
  // There are one of two ways we could handle the empty string:
  // 1) blow up on null and then require that customers wrap the validator with one that 
  // succeeds on null if they don’t like the behavior 
  // 2) Accept null and expect that the application will chain in the required checked if necessary
  // As a team we decided 2) was better than 1).
  if (value === null || value === undefined || value === "")
  {
    return true;
  }

  // when using digits as input values parseString becomes a integer type, so get away with it.
  value = (value || value === 0) ? value.toString() : value;

  // We intend that the pattern provided is matched exactly
  var exactPattern = "^(" + pattern + ")$", valid = false,
  localizedDetail, localizedSummary, matchArr;

  matchArr = value.match(exactPattern);
  if ((matchArr !== null) && (matchArr[0] === value))
  {
    valid = true;
  }
  else
  {
    if (this._options)
    {
      summary = this._options['messageSummary'] || null;
      detail = this._options['messageDetail'] || null;
      label = this._options && this._options['label'] || "";
    }

    params = {'label': label, 'pattern': pattern, 'value': value};
    localizedSummary = summary ?
    oj.Translations.applyParameters(summary, params) :
    oj.Translations.getTranslatedString(this._getSummaryKey(), params);
    localizedDetail = (detail) ?
    oj.Translations.applyParameters(detail, params) :
    oj.Translations.getTranslatedString(this._getDetailKey(), params);

    throw new oj.ValidatorError(localizedSummary, localizedDetail);
  }

  return valid;
};

/**
 * A message to be used as hint, when giving a hint on the expected pattern. There is no default 
 * hint for this property.
 * 
 * @returns {string|null} a hint message or null if no hint is available in the options
 * @memberof oj.RegExpValidator
 * @instance
 * @export
 */
oj.RegExpValidator.prototype.getHint = function ()
{
  var hint = null;
  var params = {};
  if (this._options && (this._options['hint']))
  {
    params = {'pattern': this._options['pattern']};
    hint = oj.Translations.applyParameters(this._options['hint'], params);
  }

  return hint;
};

oj.RegExpValidator.prototype._getSummaryKey = function ()
{
  return oj.RegExpValidator._BUNDLE_KEY_SUMMARY;
};

oj.RegExpValidator.prototype._getDetailKey = function ()
{
  return oj.RegExpValidator._BUNDLE_KEY_DETAIL;
};
/**
 * Copyright (c) 2008, 2013, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Constructs a RequiredValidator that ensures that the value provided is not empty
 * @param {Object=} options an object literal used to provide an optional hint and error message.<p>
 * 
 * @export
 * @constructor
 * @augments oj.Validator
 * @ojsignature [{target: "Type", value: "class RequiredValidator extends Validator<Object|string|number>"},
 *               {target: "Type", value: "oj.RequiredValidator.ValidatorOptions", for: "options", jsdocOverride: true}]
 * @since 0.6
 * 
 */
oj.RequiredValidator = function (options)
{
  this.Init(options);
};

// Subclass from oj.Object or oj.Validator. It does not matter
oj.Object.createSubclass(oj.RequiredValidator, oj.Validator, "oj.RequiredValidator");

// key to access required validator specific resources in the bundle 
oj.RequiredValidator._BUNDLE_KEY_DETAIL = "oj-validator.required.detail";
oj.RequiredValidator._BUNDLE_KEY_SUMMARY = "oj-validator.required.summary";

/**
 * @typedef {object} oj.RequiredValidator.ValidatorOptions
 * @property {string=} hint an optional hint text. There is no default hint provided by this 
 * validator.
 * @property {string=} messageSummary - an optional custom error message summarizing the 
 * error. When not present, the default message summary is the resource defined with the key 
 * <code class="prettyprint">oj-validator.required.summary</code>.<p>
 * <p>The messageSummary string is passed as the 'pattern' parameter to
 * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
 * that documentation, if you are using a reserved character, you need to escape it with 
 * a dollar character ('$').
 * </p>
 * Tokens: {label} - this token can be used to substitute the label of the component at runtime. </p>
 * <p>
 * Example:<br/>
 * "'{label}' Required"<br/>
 * </p>
 * @property {string=} messageDetail - a custom error message used for creating detail part 
 * of message, when the value provided is empty. When not present, the default message detail is the 
 * resource defined with the key <code class="prettyprint">oj-validator.required.detail</code>.
 * <p>The messageDetail string is passed as the 'pattern' parameter to
 * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
 * that documentation, if you are using a reserved character, you need to escape it with 
 * a dollar character ('$').
 * </p>
 * <p>Tokens: {label} - this token can be used to substitute the label of the component at runtime.</p>
 * <p>
 * Example:<br/>
 * "A value is required for the field '{label}'."<br/>
 * </p>
 */
/**
 * Initializes validator instance with the set options
 * @param {Object=} options
 * @memberof oj.RequiredValidator
 * @instance
 * @ignore
 */
oj.RequiredValidator.prototype.Init = function (options)
{
  oj.RequiredValidator.superclass.Init.call(this);
  this._options = options;
};

/**
 * Validates value to be non-empty
 * 
 * @param {Object|string|number} value that is being validated 
 * @returns {boolean} true if validation was was successful the value is non-empty
 * 
 * @throws {Error} when fails required-ness check
 * @ojsignature {target: "Type", for: "returns",
 *                value: "void"}
 * @memberof oj.RequiredValidator
 * @instance
 * @export
 */
oj.RequiredValidator.prototype.validate = function (value)
{
  var detail;
  var label = "";
  var localizedDetail;
  var localizedSummary;
  var summary;
  var params = {};

  // checks for empty arrays and String. Objects are considered non-null.
  // Need to specifically test for if value is 0 first if number is passed on.
  if ((typeof value === "number" && value === 0) || (value && value.length !== 0))
  {
    return true;
  }
  else
  {
    if (this._options)
    {
      // we have deprecated support for message param and instead use messageDetail.
      detail = this._options['messageDetail'] || this._options['message'] || null;
      summary = this._options['messageSummary'] || null;
      label = this._options['label'] || "";
    }
    params = {'label': label};
    localizedSummary = (summary) ? oj.Translations.applyParameters(summary, params) :
    oj.Translations.getTranslatedString(this._getSummaryKey(), params);
    localizedDetail = (detail) ?
    oj.Translations.applyParameters(detail, params) :
    oj.Translations.getTranslatedString(this._getDetailKey(), params);

    throw new oj.ValidatorError(localizedSummary, localizedDetail);
  }

};

/**
 * A message to be used as hint, when giving a hint on the expected pattern. There is no default 
 * hint for this property.
 * 
 * @returns {string|null} a hint message or null if no hint is available in the options
 * @memberof oj.RequiredValidator
 * @instance
 * @export
 */
oj.RequiredValidator.prototype.getHint = function ()
{
  var hint = "";
  if (this._options && (this._options['hint']))
  {
    hint = oj.Translations.getTranslatedString(this._options['hint']);
  }

  return hint;
};

oj.RequiredValidator.prototype._getSummaryKey = function ()
{
  return oj.RequiredValidator._BUNDLE_KEY_SUMMARY;
};

oj.RequiredValidator.prototype._getDetailKey = function ()
{
  return oj.RequiredValidator._BUNDLE_KEY_DETAIL;
};
/**
 * Copyright (c) 2008, 2013, Oracle and/or its affiliates. 
 * All rights reserved.
 */
/**
 * A factory implementation to create the built-in color converter of type 
 * {@link oj.ColorConverter}. 
 * 
 * @name oj.ColorConverterFactory
 * @public
 * @class
 * @hideconstructor
 * @example <caption>create an instance of the jet color converter using the options provided</caption>
 * var ccf = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_COLOR);  
 * var options = {format: 'hsl'};
 * var colorConverter = ccf.createConverter(options);
 * @since 0.6
 * 
 */
oj.ColorConverterFactory = (function () 
{
  function _createColorConverter(options) 
  {
    return new oj.ColorConverter(options);
  }
  
  /**
   * 
   * @public
   */
  return {
    /**
     * Creates an immutable (jet) color converter instance. 
     * 
     * @param {Object=} options an object literal used to provide an optional information to 
     * initialize the jet color converter. For details on what to pass for options, refer to 
     * {@link oj.ColorConverter}.
     * 
     * @return {oj.ColorConverter} 
     * @ojsignature {target: "Type", for: "options", value: "oj.ColorConverter.ConverterOptions"}
     * @instance
     * @memberOf oj.ColorConverterFactory
     * @public
     */
    'createConverter' : function(options) {
                             return _createColorConverter(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultConverterFactory(oj.ConverterFactory.CONVERTER_TYPE_COLOR, // factory name
                                                oj.ColorConverterFactory);

// JET VALIDATOR FACTORIES 

/**
 * A factory implementation to create an instance of the built-in required validator of type 
 * {@link oj.RequiredValidator}. 
 * 
 * @example <caption>create an instance of the required validator using the factory</caption>
 * var rvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED); 
 * var options = {'hint' : 'a value is required for this field'};
 * var requiredValidator = rvf.createValidator(options);
 * 
 * @name oj.RequiredValidatorFactory
 * @hideconstructor
 * @class
 * @public
 * @since 0.6
 * 
 */
oj.RequiredValidatorFactory = (function () 
{
  
  function _createRequiredValidator(options) 
  {
    return new oj.RequiredValidator(options);
  }
  
  return {
    /**
     * Creates an immutable validator instance of type @link oj.RequiredValidator that ensures that 
     * the value provided is not empty.  
     * 
     * @param {Object=} options an object literal used to provide an optional hint and error 
     * message. See {@link oj.RequiredValidator} for details.<p>
     * 
     * @return {oj.RequiredValidator}
     * @ojsignature {target: "Type", for: "options", value: "oj.RequiredValidator.ValidatorOptions"}
     * @memberOf oj.RequiredValidatorFactory
     * @instance
     * @public
     */
    'createValidator': function(options) {
      return _createRequiredValidator(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultValidatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED,
                                                oj.RequiredValidatorFactory);
                               
/**
 * A factory implementation that creates an instance of the built-in regExp validator of type 
 * {@link oj.RegExpValidator}.
 * 
 * @example <caption>create an instance of the regExp validator using the factory</caption>
 * var rvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP);  
 * var usernameValidator = rvf.createValidator(
 *  {
 *    'pattern': '[a-zA-Z0-9]{3,}', 
 *    'messageDetail': 'You must enter at least 3 letters or numbers'}
 *  });
 * 
 * @name oj.RegExpValidatorFactory
 * @class
 * @public
 * @hideconstructor
 * @since 0.6
 * 
 */
oj.RegExpValidatorFactory = (function () 
{
  
  function _createRegExpValidator(options) 
  {
    return new oj.RegExpValidator(options);
  }
  
  return {
    /**
     * Creates an immutable validator instance of type {@link oj.RegExpValidator} that ensures the value 
     * matches the provided pattern. 
     * 
     * @param {Object} options an object literal used to provide the pattern, an optional hint, error 
     * message among others. See {@link oj.RegExpValidator} for details.<p>
     * 
     * 
     * @return {oj.RegExpValidator} 
     * @memberOf oj.RegExpValidatorFactory
     * @ojsignature {target: "Type", for: "options", value: "oj.RegExpValidator.ValidatorOptions"}
     * @instance
     * @public
     */
    'createValidator': function(options) {
      return _createRegExpValidator(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultValidatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REGEXP,
                                                oj.RegExpValidatorFactory);
                                        
                              
/**
 * Returns an instance of oj.LengthValidatorFactory that provides a factory method to create an 
 * instance of a length validator. 
 * 
 * @example <caption>create an instance of the length validator using the factory</caption>
 * var lvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_LENGTH);
 * var options = {hint: {max: 'Enter {max} or fewer characters'}, max: 10};
 * var lValidator = lvf.createValidator(options);
 * 
 * @name oj.LengthValidatorFactory
 * @hideconstructor
 * @class
 * @public
 * @since 0.6
 * 
 */
oj.LengthValidatorFactory = (function () 
{
  
  function _createLengthValidator(options) 
  {
    return new oj.LengthValidator(options);
  }
  
  return {
    /**
     * Creates an immutable validator instance of type oj.LengthValidator that ensures that the 
     * value provided is withing a given length.
     * 
     * @param {Object=} options an object literal used to provide the 'minimum', 'maximum' and other 
     * optional values. See {@link oj.LengthValidator} for details.<p>
     * 
     * @return {oj.LengthValidator}
     * @memberOf oj.LengthValidatorFactory
     * @ojsignature {target: "Type", for: "options", value: "oj.LengthValidator.ValidatorOptions"}
     * @instance
     * @public
     */
    'createValidator': function(options) {
      return _createLengthValidator(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultValidatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_LENGTH,
                                                oj.LengthValidatorFactory);
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * oj.NumberConverter Contract. 
 */

/**
 * @export
 * @constructor
 * @augments oj.Converter
 * @ojsignature {target: "Type", 
 *                value: "class NumberConverter extends Converter<number>"}
 *              
 * @name oj.NumberConverter
 * @abstract
 * @since 0.6
 */
oj.NumberConverter = function()
{
  this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.NumberConverter, oj.Converter, "oj.NumberConverter");

/**
 * Initializes the number converter instance with the set options.
 * @param {Object=} options an object literal used to provide an optional information to 
 * initialize the converter.<p>
 * @export
 * @ignore
 */
oj.NumberConverter.prototype.Init = function(options) 
{
  oj.NumberConverter.superclass.Init.call(this, options);
};

/**
 * Formats the Number value using the options provided and returs a String value.
 * 
 * @param {number} value the value to be formatted for display
 * @return {(string|null)} the localized and formatted value suitable for display
 * @throws {Error} a ConverterError if formatting fails.
 * @export
 */
oj.NumberConverter.prototype.format = function (value) 
{
  return oj.NumberConverter.superclass.format.call(this, value);
};

/**
 * Parses the value using the options provided and returns a Number object.
 * 
 * @param {string} value to parse
 * @return {number} the parsed value as a Number object.
 * @throws {Error} a ConverterError if parsing fails
 * @export
 */
oj.NumberConverter.prototype.parse = function (value) 
{
  return oj.NumberConverter.superclass.parse.call(this, value);
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*global OraNumberConverter:true*/
/**
 * @export
 * Placeholder here as closure compiler objects to export annotation outside of top level
 */

/**
 * @constructor
 * @classdesc Constructs an immutable instance and initializes it with the options provided. When initialized 
 * with no options, the default options for the current locale are assumed. The converters by 
 * default use the current page locale (returned by oj.Config.getLocale()). There are several ways 
 * to initialize the converter.
 * <p>
 * <ul>
 * <li>Using options defined by the ECMA 402 Specification, these would be the properties style, 
 * currency, currencyDisplay, minimumIntegerDigits, minimumFractionDigits, maximumFractionDigits, 
 * useGrouping. NOTE: minimumSignificantDigits and maximumSignificantDigits are not supported.</li>
 * <li>Using a custom decimal, currency or percent format pattern. specified using the 'pattern' property</li>
 * <li>Using the decimalFormat option to define a compact pattern, such as "1M" and "1 million".</li>
 * <li>Using the currencyFormat option to define a compact pattern, such as "$1M" and "$1 million".</li>
 * <li>Using the roundingMode and roundDuringParse options to round the number HALF_UP, HALF_DOWN, or HALF_EVEN.</li>
 * </ul>
 * <p>
 * 
 * The converter provides leniency when parsing user input value to a number in the following ways:<br/>
 * 
 * <ul>
 * <li>Prefix and suffix that do not match the pattern, are removed. E.g., when pattern is 
 * "#,##0.00%" (suffix is the % character), a value of "abc-123.45xyz", will be leniently parsed to 
 * -123.45</li>
 * <li>When a value includes a symbol but the pattern doesn't require it.  E.g., the options are 
 * {pattern: "###", currency: 'USD'}, then values ($123), (123) and -123 will be leniently parsed as 
 * -123.</li>
 * </ul>
 * <p>
 * Lenient parse can be disabled by setting the property lenientParse to "none". In which case the user input must
 * be an exact match of the expected pattern and all the leniency described above will be disabled.
 * <p>
 * @param {Object=} options - an object literal used to provide optional information to 
 * initialize the converter.
 * 
 * @example <caption>Create a number converter for currencies</caption>
 * var converterFactory = oj.Validation.converterFactory("number");
 * var options = {style: "currency", currency: "USD", minimumIntegerDigits: 2};
 * converter = converterFactory.createConverter(options);
 * converter.format(9); --> "$09.00" if page locale is 'en-US'
 * converter.format(9); --> "09,00 $US" if page locale is 'fr-FR'<br/>
 * 
 * @example <caption>A number converter for percent values using a custom (CLDR) pattern</caption>
 * var converterFactory = oj.Validation.converterFactory("number");
 * var options = {pattern: '#,##0%'};
 * converter = converterFactory.createConverter(options);<br/>
 * 
 * @example <caption>To parse a value as percent but format it without displaying the percent character</caption>
 * var options = {style: 'percent', pattern: '#,##0'};<br/>
 * 
 * @example <caption>To parse a value as currency using a custom (CLDR) pattern</caption>
 * var options = {pattern: '¤#,##0', currency: 'USD'};
 * 
 * @example <caption>To format a value as digital bit unit</caption>
 * var options = {style:'unit', unit:'bit'};
 * converter = converterFactory.createConverter(options);
 * var nb = 1024;
 * converter.format(nb, localeElements, options);--> 1Kb<br/>
 * 
 * @example <caption>To format a value as digital byte unit</caption>
 * var options = {style:'unit', unit:'byte'};
 * converter = converterFactory.createConverter(options);
 * var nb = 1024;
 * converter.format(nb, localeElements, options);--> 1KB<br/>
 *
 * @example <caption>The following decimalFormat examples are in en locale.
 * To format a value as short (default for fraction digits is based on the locale)</caption>
 * var options = {style:’decimal’, decimalFormat:’short’};
 * converter = converterFactory.createConverter(options);
 * converter.format(12345);--> 12.354K<br/>
 * 
 * @example <caption>Same as above for currencyFormat.
 * To format a value as short (default for fraction digits is based on the locale)</caption>
 * var options = {style:'currency', currency: 'USD', currencyFormat:'short'};
 * converter = converterFactory.createConverter(options);
 * converter.format(1234);--> $1.23K<br/>
 * 
 * @example <caption>To format a value as long (default for fraction digits is based on the locale):</caption>
 * var options = {style:’decimal’, decimalFormat:’long’};
 * converter = converterFactory.createConverter(options);
 * converter.format(12345);--> 12.345 thousand<br/>
 * 
 * @example <caption>To format a value as long currency format:</caption>
 * var options = {style:'currency',  currency: 'USD', currencyFormat:'long'};
 * converter = converterFactory.createConverter(options);
 * converter.format(1234);--> $1.23 thousand<br/>
 * 
 * @example <caption>To format a value as short with minimum fraction digits:</caption>
 * options = { style:’decimal’, decimalFormat:’short’, 
 * minimumFractionDigits:4};
 * converter = converterFactory.createConverter(options);
 * converter.format(1234);--> 1.2340K<br/>
 * 
 * @example <caption>To format a value as short with maximum fraction digits:</caption>
 * options = { style:’decimal’, decimalFormat:’short’, 
 * maximumFractionDigits:0};
 * converter = converterFactory.createConverter(options);
 * converter.format(12345);--> 12K<br/>
 * 
 * @example <caption>To format a value as long with minimum and maximum fraction digits:</caption>
 * options = { style:’decimal’, decimalFormat:’long', 
 * minimumFractionDigits:2, maximumFractionDigits:4};
 * converter = converterFactory.createConverter(options);
 * converter.format(12000);--> 12.00 thousand<br/>
 * 
 * @example <caption>To format a value as short with minimum and maximum fraction digits:</caption>
 * options = { style:’decimal’, decimalFormat:’long', 
 * minimumFractionDigits:2, maximumFractionDigits:4};
 * converter = converterFactory.createConverter(options);
 * converter.format(12345678);--> 12.345 million<br/>
 * 
 * @example <caption>decimal style default is standard:</caption>
 * options = { style:’decimal’, decimalFormat:’standard’}; 
 * converter = converterFactory.createConverter(options);
 * converter.format(12345);--> 12,345<br/>
 * 
 * @example <caption>decimal round HALF_DOWN:</caption>
 * options = { style:’decimal’,  maximumFractionDigits:2, roundingMode:'HALF_DOWN'}; 
 * converter = converterFactory.createConverter(options);
 * converter.format(0.225);--> 0.22
 * converter.parse(0.225);-->0.225 //doesn't round during parse by default<br/>
 * 
 * @example <caption>decimal round HALF_UP:</caption>
 * options = { style:’decimal’,  maximumFractionDigits:2, roundingMode:'HALF_UP'}; 
 * converter = converterFactory.createConverter(options);
 * converter.format(0.225);--> 0.23
 * converter.parse(0.225);--> 0.225 //doesn't round during parse by default<br/>
 * 
 * @example <caption>decimal round HALF_EVEN:</caption>
 * options = { style:’decimal’,  maximumFractionDigits:2, roundingMode:'HALF_EVEN'}; 
 * converter = converterFactory.createConverter(options);
 * converter.format(0.225);--> 0.22
 * converter.format(0.235);--> 0.24
 * converter.parse(0.225);--> 0.225 //doesn't round during parse by default
 * converter.parse(0.235);--> 0.235 //doesn't round during parse by default<br/>
 * 
 * @example <caption>decimal round HALF_DOWN and roundDuringParse:</caption>
 * options = { style:’decimal’, maximumFractionDigits:2, 
 *             roundingMode:'HALF_DOWN', roundDuringParse: true}; 
 * converter = converterFactory.createConverter(options);
 * converter.format(0.225);--> 0.22
 * converter.parse(0.225);-->0.22<br/>
 * 
 * @example <caption>decimal round HALF_UP and roundDuringParse:</caption>
 * options = { style:’decimal’,  maximumFractionDigits:2, 
 *             roundingMode:'HALF_UP', roundDuringParse: true}; 
 * converter = converterFactory.createConverter(options);
 * converter.format(0.225);--> 0.23
 * converter.parse(0.225);--> 0.23<br/>
 * 
 * @example <caption>decimal round HALF_EVEN and roundDuringParse:</caption>
 * options = { style:’decimal’,  maximumFractionDigits:2, 
 *             roundingMode:'HALF_EVEN', roundDuringParse: true}; 
 * converter = converterFactory.createConverter(options);
 * converter.format(0.225);--> 0.22
 * converter.format(0.235);--> 0.24
 * converter.parse(0.225);--> 0.22
 * converter.parse(0.235);--> 0.24<br/>
 *
 * @example <caption>Use custom decimal and grouping separators:</caption>
 * options = { style:'decimal', separators: {decimal: ',',  group: '.'}};
 * converter = converterFactory.createConverter(options);
 * var nb = 1234567.89; 
 * converter.format(nb);--> 1.234.567,89
 * converter.parse("1.234.567,89");--> 1234567.89
 * <br/>
 *
 * @example <caption>Disable lenient parse:</caption>
 * options = { style:'decimal',  lenientParse: 'none'}; 
 * converter = converterFactory.createConverter(options);
 * converter.parse("abc-123.45xyz");--> Error: Unparsable number abc-123.45xyz The expected number pattern is #,##0.###
 * <br/>
 * 
 * @export
 * @augments oj.NumberConverter 
 * @name oj.IntlNumberConverter
 * @ojsignature [{target: "Type", 
 *                value: "class IntlNumberConverter extends NumberConverter"},
 *               {target: "Type",
 *                value: "oj.IntlNumberConverter.ConverterOptions",
 *                for: "options", jsdocOverride: true}
 *              ]
 * @since 0.6
 */
oj.IntlNumberConverter = function(options)
{
  this.Init(options);
};

/**
 * @typedef {object} oj.IntlNumberConverter.ConverterOptions
 * @property {('decimal'|'currency'|'percent'|'unit')=} style - sets the style of number formatting. Allowed values are "decimal" 
 * (the default), "currency", "percent" or "unit". When a number is formatted as a decimal, the decimal 
 * character is replaced with the most appropriate symbol for the locale. In English this is a 
 * decimal point ("."), while in many locales it is a decimal comma (","). If grouping is enabled the 
 * locale dependent grouping separator is also used. These symbols are also used for numbers 
 * formatted as currency or a percentage, where appropriate.
 * @property {string=} currency - specifies the currency that will be used when formatting the 
 * number. The value should be a ISO 4217 alphabetic currency code. If the style is set to currency, 
 * it's required that the currency property also be specified. This is because there is no default 
 * currency associated with the current locale. The user must always specify the currency code 
 * to be shown, otherwise an error will be thrown. The current page locale 
 * (returned by oj.Config.getLocale()) determines the formatting elements of the number 
 * like grouping separator and decimal separator. The currency code tells us which currency to 
 * display in current page locale. JET has translations for currency names.
 * <p>
 * As an example if we want to format 1000.35 EURO and the page locale is "en-US", 
 * we pass {style:'currency', currency:'EUR', currencyDisplay:'symbol'} and we will get "€1,000.35"
 * If the page locale is "fr-FR", with the same options, we will get: "1 000,35 €"
 * </p>
 * @property {('byte'|'bit')=} unit - Mandatory when style is "unit". Allowed values: 
 * "byte" or "bit". It is used for formatting only. It can not be used for parsing.
 * <p> 
 * It is used to format digital units like 10Mb for bit unit or 10MB for byte unit. 
 * There is no need to specify the scale of the unit. We automatically detect it.
 * For example 1024 is formatted as 1KB and ?1048576? as 1MB.
 * The user can also specify 'minimumFractionDigits' and  'maximumFractionDigits' to be displayed, 
 * otherwise we use the locale's default max and min fraction digits.
 * </p> 
 * @property {('code'|'symbol'|'name')=} currencyDisplay - if the number is using currency formatting, specifies 
 * if the currency will be displayed using its "code" (as an ISO 4217 alphabetic currency code), 
 * "symbol" (a localized currency symbol (e.g. $ for US dollars, £ for Great British pounds, and so 
 * on), or "name" (a localized currency name. Allowed values are "code", "symbol" and "name". 
 * The default is "symbol".
 * @property {('standard'|'short'|'long')=} decimalFormat -
 * specifies the decimal format length to use when style is set to "decimal". 
 * Allowed values are : "standard"(default), "short" and "long". 'standard' is equivalent to not 
 * specifying the 'decimalFormat' attribute, in that case the locale’s default decimal pattern 
 * is used for formatting.
 * <p>
 * The user can also specify 'minimumFractionDigits' and  'maximumFractionDigits' to display. 
 * When not present we use the locale's default max and min fraction digits.
 * </p>
 * <p>
 * There is no need to specify the scale; we automatically detect greatest scale that is less or 
 * equal than the input number. For example  1000000 is formatted as "1M" or "1 million" and
 * 1234 is formatted, with zero fractional digits, as "1K" or " 1 thousand" for 
 * short and long formats respectively. The pattern for the short and long number is locale dependent 
 * and uses plural rules for the particular locale.
 * </p>
 * <p>
 * NOTE: Currently this option formats a value (e.g., 2000 -> 2K), but it does not parse a value 
 * (e.g., 2K -> 2000), so it can only be used
 * in a readOnly EditableValue because readOnly EditableValue components do not call
 * the converter's parse function.
 * </p>
 * @property {('standard'|'short'|'long')=} currencyFormat -
 * specifies the currency format length to use when style is set to "currency". 
 * Allowed values are : "standard"(default), "short" and "long". 'standard' is equivalent to not 
 * specifying the 'currencyFormat' attribute, in that case the locale's default currency pattern 
 * is used for formatting.
 * Similar to decimalFormat, currencyFormat can only be used for formatting. It can not be used for parsing.
 * <p>
 * The user can also specify 'minimumFractionDigits' and  'maximumFractionDigits' to display. 
 * When not present we use the locale's default max and min fraction digits.
 * </p>
 * <p>
 * There is no need to specify the scale; we automatically detect greatest scale that is less or 
 * equal than the input number. For example  1000000 is formatted as "$1M" or "1 million dollar" and
 * 1000 is formatted as "$1K" or " 1 thousand dollar" for short and long formats respectively.
 * The pattern for the short and long number is locale dependent and uses plural rules for the particular locale.
 * </p>
 * @property {number=} minimumIntegerDigits - sets the minimum number of digits before the 
 * decimal place (known as integer digits). The number is padded with leading zeros if it would not 
 * otherwise have enough digits. The value must be an integer between 1 and 21.
 * @property {number=} minimumFractionDigits - similar to 'minimumIntegerDigits', except it 
 * deals with the digits after the decimal place (fractional digits). It must be an integer between 
 * 0 and 20. The fractional digits will be padded with trailing zeros if they are less than the minimum.
 * @property {number=} maximumFractionDigits - follows the same rules as 'minimumFractionDigits', 
 * but sets the maximum number of fractional digits that are allowed. The value will be rounded if 
 * there are more digits than the maximum specified.
 * @property {boolean=} useGrouping - when the value is truthy, the locale dependent grouping 
 * separator is used when formatting the number. This is often known as the thousands separator, 
 * although it is up to the locale where it is placed. The ‘useGrouping’ is set to true by default.
 * @property {string=} pattern an optional localized pattern, where the characters used in 
 * pattern are as defined in the Unicode CLDR for numbers, percent or currency formats. When present 
 * this will override the other "options". <p>
 * 
 * &nbsp;&nbsp;- When the pattern represents a currency style the 'currency' property is required to 
 * be set, as not setting this will throw an error. The 'currencyDisplay' is optional. <br/>Example: 
 * {pattern: '¤#,##0', currency: 'USD'}. <p>
 * 
 * &nbsp;&nbsp;- It's not mandatory for the pattern to have the special character '¤' (currency sign) 
 * be present. When not present, values are treated as a currency value, but are not formatted to 
 * show the currency symbol. <br/>Example: {pattern: '#,##0', currency: 'USD'} <p>
 * 
 * &nbsp;&nbsp;- When the pattern represents a percent style, the percent special character ('%') needs to be 
 * explicitly specified in the pattern, e.g., {pattern: "#,##0%"}. If the pattern does not contain 
 * the percent character it's treated as a decimal pattern, unless the style is set to percent, 
 * in which case the value is treated as a percent value, but not formatted to show the percent symbol. 
 * <br/>Example: {style: 'percent', pattern: "#,##0"}. <p>
 * 
 * &nbsp;&nbsp;- A decimal pattern or exponent pattern is specified in the pattern using the CLDR 
 * conventions. <br/>Example: {pattern: "#,##0.00"} or {pattern: "0.##E+0"}. <p>
 * 
 * NOTE: 'pattern' is provided for backwards compatibility with existing apps that may want the 
 * convenience of specifying an explicit format mask. Setting a pattern will override the default 
 * locale specific format. <br/>
 * 
 * @property {('HALF_UP'|'HALF_DOWN'|'HALF_EVEN')=} roundingMode - specifies the rounding behavior. 
 * This follows the Java.Math.RoundingMode behavior.
 * Currently we support the options: HALF_UP, HALF_DOWN, and HALF_EVEN 
 * 
 * @property {boolean=} roundDuringParse - Specifies whether or not to round during
 * parse. Defaults to false; the number converter rounds during format but not during parse.
 * 
 * @property {Object=} separators - An object with 2 fields: 'decimal' and 'group'.
 * It allows the user to provide custom decimal and grouping separators. It is accepted for both
 * format and parse methods. 
 * <br/>
 * 
 * @property {('full'|'none')=} lenientParse - The lenientParse property can be used to enable or disable leninet parsing.
 *  Allowed values: "full" (default), "none". 
 * <p style='padding-left: 5px;'>
 * By default the lenient parse is enabled and the leniency rules descibed above will be used. When lenientParse is
 * set to "none" the lenient parse is disabled and the user input must match the expected input otherwise an exception will 
 * be thrown.<br/><br/>
 */
oj.Object.createSubclass(oj.IntlNumberConverter, oj.NumberConverter, "oj.IntlNumberConverter");

/**
 * Initializes the number converter instance with the set options.
 * @param {Object=} options an object literal used to provide an optional information to 
 * initialize the converter.<p>
 * @export
 * @ignore
 */
oj.IntlNumberConverter.prototype.Init = function(options) 
{
  oj.IntlNumberConverter.superclass.Init.call(this, options);
};


// Returns the wrapped number converter implementation object.
oj.IntlNumberConverter.prototype._getWrapped = function ()
{
  if (!this._wrapped)
  {
    this._wrapped = OraNumberConverter.getInstance();
  }
  
  return this._wrapped;
};

/**
 * Formats a Number and returns the formatted string, using the options this converter was 
 * initialized with.
 * 
 * @param {number} value to be formatted for display
 * @return {string} the localized and formatted value suitable for display. When the value is 
 * formatted as a percent it's multiplied by 100.
 * 
 * @throws {Error} a ConverterError both when formatting fails, or if the options provided during 
 * initialization cannot be resolved correctly. 
 * 
 * @export
 */
oj.IntlNumberConverter.prototype.format = function (value) 
{
  var converterError;
  var locale;
  var localeElements;
  var resolvedOptions;

  // undefined, null and empty string values all return null. If value is NaN then return "".
  if (value == null || 
      (typeof value === "string" && (oj.StringUtils.trim("" + value)).length === 0) ||
      (typeof value === "number" && isNaN(value))) 
  {
    return oj.IntlConverterUtils.__getNullFormattedValue();
  }
  
  locale = oj.Config.getLocale();
  localeElements = oj.LocaleData.__getBundle();
  resolvedOptions = this.resolvedOptions();

  
  try
  {
    return this._getWrapped().format(value, localeElements, resolvedOptions, locale);
  }
  catch (e)
  {
    converterError = this._processConverterError(e, value);
    throw converterError;
  }
};

/**
 * In general, returns hint for the converter. For a IntlNumberConverter returned value is always null.
 * 
 * @return {null} a hint describing the format the value is expected to be in.
 * @export
 */
oj.IntlNumberConverter.prototype.getHint = function ()
{
  // UX does not want any hint for numbers. 
  // return oj.Translations.getTranslatedString("oj-converter.hint.summary",
  //        {'exampleValue': this._getHintValue()}); 
  //return oj.IntlNumberConverter.superclass.getHint.call(this); // this asserts, and we don't want that.
  return null;
};

/**
 * Returns the options called with converter initialization.
 * @return {Object} an object of options.
 * @ojsignature {target:"Type", for: "returns",
 *    value: "oj.IntlNumberConverter.ConverterOptions"}
 * @export
 */
oj.IntlNumberConverter.prototype.getOptions = function () 
{
  return oj.IntlNumberConverter.superclass.getOptions.call(this);
};

/**
 * Parses a string value to return a Number, using the options this converter was initialized with. 
 * 
 * @param {string} value to parse
 * @return {number|null} the parsed number or null if the value was null or an empty string. When 
 * the value is parsed as a percent its 1/100th part is returned.
 * 
 * @throws {Error} a ConverterError both when parsing fails, or if the options provided during 
 * initialization cannot be resolved correctly. 
 *  
 * @export
 */
oj.IntlNumberConverter.prototype.parse = function (value) 
{
  var converterError;
  var locale; 
  var localeElements;
  var resolvedOptions;

  // null and empty string values are ignored and not parsed. It
  // undefined.
  if (value == null || value === "") // check for undefined, null and ""
  {
    return null;
  }
  
  locale = oj.Config.getLocale(); 
  localeElements = oj.LocaleData.__getBundle();
  resolvedOptions = this.resolvedOptions();
  
  try
  {
    // we want to trim the value for leading spaces before and after
    return this._getWrapped().parse(oj.StringUtils.trim(value), 
                                    localeElements, 
                                    resolvedOptions, 
                                    locale);
  }
  catch (e)
  {
    converterError = this._processConverterError(e, value);
    throw converterError;
  }
  
};

/**
 * Returns an object literal with properties reflecting the number formatting options computed based 
 * on the options parameter. If options (or pattern) is not provided, the properties will be derived 
 * from the locale defaults.
 * 
 * @return {Object} An object literal containing the resolved values for the following options. Some 
 * of these properties may not be present, indicating that the corresponding components will not be 
 * represented in the formatted output.
 * <ul>
 * <li><b>locale</b>: a String value with the language tag of the locale whose localization is used 
 * for formatting.</li>
 * <li><b>style</b>: a String value. One of the allowed values - "decimal", "currency" or "percent".</li>
 * <li><b>currency</b>: a String value.  an ISO 4217 alphabetic currency code. May be present only 
 *  when style is currency.</li>
 * <li><b>currencyDisplay</b>: a String value. One of the allowed values - "code", "symbol", or 
 *  "name".</li>
 * <li><b>numberingSystem</b>: a String value of the numbering system used. E.g. latn</li>
 * <li><b>minimumIntegerDigits</b>: a non-negative integer Number value indicating the minimum 
 *  integer digits to be used.</li>
 * <li><b>minimumFractionDigits</b>: a non-negative integer Number value indicating the minimum 
 *  fraction digits to be used.</li>
 * <li><b>maximumFractionDigits</b>: a non-negative integer Number value indicating the maximum 
 *  fraction digits to be used.</li>
 * <li><b>useGrouping</b>: a Boolean value indicating whether a grouping separator is used.</li>
 * <li><b>lenientParse</b>: specifies if lenient parse is enabled or disabled. Allowed values: "full", "none".
 * default is "full" which means lenient parse is enabled.</li>
 * <li><b>separators</b>: - An object with 2 fields: 'decimal' and 'group'.</li>
 * 
 * @throws a oj.ConverterError when the options that the converter was initialized with are invalid. 
 * @ojsignature {target:"Type", for: "returns",
 *    value: "oj.IntlNumberConverter.ConverterOptions"}
 * @export
 */
oj.IntlNumberConverter.prototype.resolvedOptions = function()
{
  var converterError;
  var locale = oj.Config.getLocale();
  var localeElements;
  
  // options are resolved and cached for the current locale. when locale changes resolvedOptions 
  // is reevaluated as it contains locale specific info.
  if ((locale !== this._locale) || !this._resolvedOptions)
  {
    localeElements = oj.LocaleData.__getBundle();
    try
    {
      if (!localeElements)
      {
        oj.Logger.error("locale bundle for the current locale %s is unavailable", locale);
        return {};
      }
      
      // cache if successfully resolved
      this._resolvedOptions = this._getWrapped().resolvedOptions(localeElements, 
                                                                 this.getOptions(), 
                                                                 locale);
      this._locale = locale;
    }
    catch (e)
    {
      converterError = this._processConverterError(e);
      throw converterError;
    }
  }
  
  return this._resolvedOptions;
};

/**
 * Processes the error returned by the converter implementation and throws a oj.ConverterError 
 * instance.
 * 
 * @param {Error} e
 * @param {String|string|Number|number|Object=} value
 * @throws an instance of oj.ConverterError
 * @private
 */
oj.IntlNumberConverter.prototype._processConverterError = function (e, value)
{
  var converterError;
  var errorCode;
  var errorInfo = e['errorInfo'];
  var detail;
  var parameterMap;
  var propName;
  var resourceKey;
  var summary;

  if (errorInfo)
  {
    errorCode = errorInfo['errorCode'];
    parameterMap = errorInfo['parameterMap'];
    oj.Assert.assertObject(parameterMap);
    propName = parameterMap['propertyName'];
    
    switch (errorCode)
    {
      case "optionTypesMismatch":
      case "optionTypeInvalid":
        converterError = oj.IntlConverterUtils.__getConverterOptionError(errorCode, parameterMap);
        break;
      case "optionOutOfRange":
        converterError = oj.IntlConverterUtils.__getConverterOptionError(errorCode, parameterMap);
        break;
      case "optionValueInvalid":
        converterError = oj.IntlConverterUtils.__getConverterOptionError(errorCode, parameterMap);
        break;
      case "decimalFormatMismatch":
        // The '{value}' does not match the expected decimal format
        resourceKey = "oj-converter.number.decimalFormatMismatch.summary";
        break;
      case "currencyFormatMismatch":
        // The {value} does not match the expected currency format
        resourceKey = "oj-converter.number.currencyFormatMismatch.summary";
        break;
      case "percentFormatMismatch":
        // The {value} does not match the expected currency format
        resourceKey = "oj-converter.number.percentFormatMismatch.summary";
        break;  
      case "unsupportedParseFormat":
        // TODO: We'll be able to remove this exception when this bug is fixed post V1.1:
        //  - implement parse() for short number converter
        //  
        summary =  oj.Translations.getTranslatedString(
          "oj-converter.number.shortLongUnsupportedParse.summary");
        detail = oj.Translations.getTranslatedString(
          "oj-converter.number.shortLongUnsupportedParse.detail");
        converterError = new oj.ConverterError(summary, detail);
    }

    // The formatMismatch errors need a hint
    if (resourceKey)
    {
      summary = oj.Translations.getTranslatedString(resourceKey, 
        {'value': value || parameterMap['value'],
         'format': parameterMap['format']});

      // _getHintValue is smart. It uses the converter's 'format' function
      //  to get the example format to show the end user.
      detail = oj.Translations.getTranslatedString("oj-converter.hint.detail",
        {'exampleValue': this._getHintValue()}); 

      converterError = new oj.ConverterError(summary, detail);
    }

  }
  
  if (!converterError)
  {
    // An error we are unfamiliar with. Get the message and set as detail
    summary = e.message; // TODO: What should the summary be when it's missing??
    detail = e.message;
    converterError = new oj.ConverterError(summary, detail);
  }
  
  return converterError;
};

// Returns the hint value. It uses the converter's format function to return a formatted
// example. For example, if the converter's style is decimal and decimalFormat is short,
// this.format(12345.98765) returns 12K, and we show 12K in the error message as an example
// of what they should type in.
oj.IntlNumberConverter.prototype._getHintValue = function()
{
  var value = "";
  try
  {
    // use .format to get a real example to show the user what format they can type in to the field.
    value =  this.format(12345.98765);
  }
  catch (e)
  {
    if (e instanceof oj.ConverterError)
    {
      // Something went wrong and we don't have a way to retrieve a valid value.    
      value = "";
      oj.Logger.error("error retrieving hint value in format");
    }
  }
  finally
  {
    // returns the formatted value of 12345.98765
    return value;
  }
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Constructs a NumberRangeValidator that ensures the value provided is within a given range
 * @param {Object=} options an object literal used to provide the following properties
 * @export
 * @constructor
 * @augments oj.Validator
 * @ojsignature [{target: "Type", 
 *                value: "class NumberRangeValidator extends Validator<string|number>"},
 *               {target: "Type",
 *                value: "oj.NumberRangeValidator.ValidatorOptions",
 *                for: "options", jsdocOverride: true}
 *              ]
 * @since 0.7
 * 
 */
oj.NumberRangeValidator = function _NumberRangeValidator(options)
{
  this.Init(options);
};

// Subclass from oj.Validator 
oj.Object.createSubclass(oj.NumberRangeValidator, oj.Validator, "oj.NumberRangeValidator");

/**
 * @typedef {object} oj.NumberRangeValidator.ValidatorOptions
 * @property {number=} min - the minimum number value of the entered value.
 * @property {number=} max - the maximum number value of the entered value.
 * @property {Object=} hint - an optional object literal of hints to be used.
 * <p>The hint strings (e.g., hint.min) are  passed as the 'pattern' parameter to
 * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
 * that documentation, if you are using a reserved character, you need to escape it with 
 * a dollar character ('$').
 * </p> 
 * @property {string=} hint.max - a hint used to indicate the allowed maximum. When not present, 
 * the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.hint.max</code>.<p>
 * Tokens: <br/>
 * {max} - the maximum<p>
 * Usage: <br/>
 * Enter a number less than or equal to {max}
 * @property {string=} hint.min - a hint used to indicate the allowed minimum. When not present, 
 * the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.hint.min</code>.<p>
 * Tokens: <br/>
 * {min} the minimum <p>
 * Usage: <br/>
 * Enter a number greater than or equal to {min}</li>
 * @property {string=} hint.inRange - a hint used to indicate the allowed range. When not 
 * present, the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.hint.inRange</code>.<p>
 * Tokens:<br/>
 * {min} the minimum<br/>
 * {max} the maximum<p>
 * Usage: <br/>
 * Enter a number between {min} and {max}
 * @property {string=} hint.exact - a hint used to indicate the allowed value. 
 * This is used when min and max are non-null and are equal to each other.  
 * When not present, the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.hint.exact</code>.<p>
 * Tokens:<br/>
 * {num} the number allowed<br/>
 * Usage: <br/>
 * Enter the number {num}
 * @since 3.0.0 
 * @property {Object=} messageDetail - an optional object literal of custom error messages to 
 * be used.
 * <p>The messageDetail strings (e.g., messageDetail.rangeUnderflow) are  passed as the 'pattern' 
 * parameter to [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
 * that documentation, if you are using a reserved character, you need to escape it with 
 * a dollar character ('$').
 * </p> 
 * @property {string=} messageDetail.rangeUnderflow - the detail error message to be used when 
 * input value is less than the set minimum value. When not present, the default detail message is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.messageDetail.rangeUnderflow</code>.<p>
 * Tokens:<br/>
 * {value} - value entered by the user<br/>
 * {min} - the minimum allowed value<p>
 * Usage: <br/>
 * The number must be greater than or equal to {min}.
 * @property {string=} messageDetail.rangeOverflow - the detail error message to be used when 
 * input value exceeds the maximum value set. When not present, the default detail message is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.messageDetail.rangeOverflow</code>.<p>
 * Tokens:<br/>
 * {value} - value entered by the user<br/>
 * {max} - the maximum allowed value<p>
 * Usage: <br/>
 * The number must be less than or equal to {max}.   
 * @property {string=} messageDetail.exact - the detail error message to be used when the
 * input value is not between min and max when min and max are both non-null and equal.
 *  When not present, the default detail message is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.messageDetail.exact</code>.<p>
 * Tokens:<br/>
 * {num} - the allowed value<p>
 * Usage: <br/>
 * The number must be {num}.
 * @since 3.0.0 
 * @property {Object=} messageSummary - optional object literal of custom error summary message 
 * to be used. 
 * @property {string=} messageSummary.rangeUnderflow - the summary of the error message when 
 * input value is less than the set minimum value. When not present, the default message summary is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.messageSummary.rangeUnderflow</code>.
 * @property {string=} messageSummary.rangeOverflow - the summary of the error message when 
 * input value exceeds the maximum value set.  When not present, the default message summary is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.number.messageSummary.rangeOverflow</code>.
 */

/**
 * Initializes validator instance with the set options
 * @param {Object=} options
 * @memberof oj.NumberRangeValidator
 * @instance
 * @export
 * @ignore
 */
oj.NumberRangeValidator.prototype.Init = function (options)
{
  oj.NumberRangeValidator.superclass.Init.call(this);

  if (options)
  {
    this._min = options["min"];
    this._max = options["max"];
    this._converter = oj.IntlConverterUtils.getConverterInstance(options['converter']);
    this._hint = options['hint'] || {};
    this._customMessageSummary = options['messageSummary'] || {};
    this._customMessageDetail = options['messageDetail'] || {};
  }
};

/**
 * Validates the minimum + maximum conditions
 *
 * @param {string|number} value that is being validated
 * @returns {string|null} original if validation was successful
 *
 * @throws {Error} when value is out of range
 * @ojsignature {target: "Type", for: "returns",
 *                value: "void"}
 * @memberof oj.NumberRangeValidator
 * @instance
 * @export
 */
oj.NumberRangeValidator.prototype.validate = function (value)
{
  var string = value ? value.toString() : value;
  var numberValue = parseFloat(string); 
  var customMessageSummary = this._customMessageSummary;
  var customMessageDetail = this._customMessageDetail;
  var messageDetailRangeOverflow = customMessageDetail["rangeOverflow"]; 
  var messageDetailRangeUnderflow = customMessageDetail["rangeUnderflow"]; 
  var messageDetailExact = customMessageDetail["exact"];
  var messageSummaryRangeOverflow = customMessageSummary["rangeOverflow"];
  var messageSummaryRangeUnderflow = customMessageSummary["rangeUnderflow"];
  var min = this._min !== undefined ? parseFloat(this._min) : null; 
  var max = this._max !== undefined ? parseFloat(this._max) : null; 
  var minStr = min && this._converter ? this._converter['format'](min) : min;
  var maxStr = max && this._converter ? this._converter['format'](max) : max;
  var summary = ""; 
  var detail = ""; 
  var params = null;
  var translations = oj.Translations;
  
  if(value === null) 
  {
    // request to not throw an error when value being passed is of null
    return value;
  }
  
  if (min !== null && max !== null)
  {
    //range
    if ((numberValue >= min && numberValue <= max) || min > max)
    {
      return string;
    }
  }
  else 
  {
    //only min
    if (min !== null)
    {
      if (numberValue >= min)
      {
        return string;
      }
	  
    }
    //max only or no min or max
    else 
    {
      if (max === null || numberValue <= max)
      {
        return string;
      }
    }
  }
  
  // if we haven't returned with an OK, then we need to throw a ValidatorError
  // 
  // First check if we have both a max and a min and if they are equal. If so the message will
  // be the messageDetail.exact message, like "Enter the number 1"
  if (max !== null && min !== null && min === max)
  {
    params = {"value": value, "num": maxStr};
    detail = messageDetailExact ? 
      translations.applyParameters(messageDetailExact, params) : 
      translations.getTranslatedString('oj-validator.range.number.messageDetail.exact', params);
    // if number is greater than max, the summary may say "The number is too high"
    if (numberValue > max)
    {
      summary = messageSummaryRangeOverflow ? 
        messageSummaryRangeOverflow : 
        translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeOverflow');            
    } 
    // if number is less than min, the summary may say "The number is too low"
    else if (numberValue < min)
    {
      summary = messageSummaryRangeOverflow ? 
        messageSummaryRangeUnderflow : 
        translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeUnderflow');
    }
  }
  // Next check if we have a max, and the number we are validating is greater than the max
  // throw an error, 
  // like "The number is too high." and "The number must be less than or equal to {max}"
  else if (max !== null && numberValue > max)
  {
	  params = {"value": value, "max": maxStr};
    summary = messageSummaryRangeOverflow ?
      messageSummaryRangeOverflow : 
      translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeOverflow');
    detail = messageDetailRangeOverflow ?
      translations.applyParameters(messageDetailRangeOverflow, params) : 
      translations.getTranslatedString('oj-validator.range.number.messageDetail.rangeOverflow', params);
  }
  // 
  // Else check if we have a min, and the number we are validating is less than the min
  // throw an error, 
  // like "The number is too low." and "The number must be greater than or equal to {min}"
  else if (min !== null && numberValue < min)
  {
 	  params = {"value": value, "min": minStr};
    summary = messageSummaryRangeUnderflow ? 
      messageSummaryRangeUnderflow : 
      translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeUnderflow');
    detail = messageDetailRangeUnderflow ? 
      translations.applyParameters(messageDetailRangeUnderflow, params) : 
      translations.getTranslatedString('oj-validator.range.number.messageDetail.rangeUnderflow', params);   
  }

  throw new oj.ValidatorError(summary, detail);
};

/**
 * @returns {string|null} a hint message or null if no hint is available in the options.
 * A hint message may be like "Enter a value between {min} and {max}"
 * or "Enter a number greater than or equal to {min}"
 * @memberof oj.NumberRangeValidator
 * @instance
 * @export
 */
oj.NumberRangeValidator.prototype.getHint = function ()
{
  var hint = null;
  var hints = this._hint; 
  var hintInRange = hints["inRange"];
  var hintExact = hints["exact"];
  var hintMinimum = hints["min"];
  var hintMaximum = hints["max"];
  var translations = oj.Translations;
  var min = this._min !== undefined ? parseFloat(this._min) : null; 
  var max = this._max !== undefined ? parseFloat(this._max) : null;
  var minStr = min && this._converter ? this._converter['format'](min) : min;
  var maxStr = max && this._converter ? this._converter['format'](max) : max;
	 
  // if both min and max are specified, the hint may say something like "Enter a value
  // between {min} and {max}".
  if (min !== null && max !== null) 
  {
    if (min !== max)
    {
      // if hintInRange is specified (validator's hint.inRange option is set), 
      // use that string, else use the default.
      hint = hintInRange ?
              translations.applyParameters(hintInRange, {"min": minStr, "max": maxStr}) : 
              translations.getTranslatedString('oj-validator.range.number.hint.inRange', 
                {"min": minStr, "max": maxStr});
    }
    else
    {
      // if hintExact is specified (validator's hint.exact option is set), 
      // use that string, else use the default.
      hint = hintExact ?
              translations.applyParameters(hintExact, {"num": minStr}) : 
              translations.getTranslatedString('oj-validator.range.number.hint.exact', 
                {"num": minStr});
    }
  }
  // else if min is specified, the hint may say something like "Enter a value
  // greater than or equal to {min}".
  else if (min !== null)
  {
    // if hintMinimum is specified (validator's hint.min option is set), 
    // use that string, else use the default.
    hint = hintMinimum ? 
             translations.applyParameters(hintMinimum, {"min": minStr}) :
	           translations.getTranslatedString('oj-validator.range.number.hint.min', {"min": minStr});
  }
  // else if max is specified, the hint may say something like "Enter a value
  // less than or equal to {max}".
  else if (max !== null)
  {
    // if hintMaximum is specified (validator's hint.max option is set), 
    // use that string, else use the default.
    hint = hintMaximum ?  
            translations.applyParameters(hintMaximum, {"max": maxStr}) :
            translations.getTranslatedString('oj-validator.range.number.hint.max', {"max": maxStr});
  }

  return hint;
};
/**
 * Copyright (c) 2008, 2013, Oracle and/or its affiliates. 
 * All rights reserved.
 */



/**
 * A factory implementation to create the built-in number converter of type 
 * {@link oj.IntlNumberConverter}. 
 * 
 * @name oj.NumberConverterFactory
 * @hideconstructor
 * @class
 * 
 * @example <caption>create an instance of the jet datetime converter using the options provided</caption>
 * var ncf = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER); 
 * var salaryOptions = {currency: "EUR" , pattern: "¤#,##0.00;(¤#,##0.00)"};
 * var salaryConverter = ncf.createConverter(salaryOptions);
 * @public
 * @since 0.6
 * 
 */
oj.NumberConverterFactory = (function () 
{
  
  function _createNumberConverter(options) 
  {
    return new oj.IntlNumberConverter(options);
  }
  
  return {
    /**
     * Creates an immutable (jet) number converter instance.
     * 
     * @param {Object=} options an object literal used to provide optional information to initialize 
     * the jet number converter with. For details on what to pass for options, refer to 
     * {@link oj.IntlNumberConverter}
     * 
     * @return {oj.IntlNumberConverter}
     * @memberOf oj.NumberConverterFactory
     * @ojsignature {target: "Type", for: "options", value: "oj.IntlNumberConverter.ConverterOptions"}
     * @instance
     * @public
     */
    'createConverter': function(options) {
      return _createNumberConverter(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultConverterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER, // factory name
                               oj.NumberConverterFactory);


// JET VALIDATOR FACTORIES 
                                        
/**
 * a factory method to create an instance of a built-in numberRange validator of type 
 * {@link oj.NumberRangeValidator}. 
 * 
 * @example <caption>create an instance of the numberRange validator using the factory</caption>
 * var lrvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_NUMBER_RANGE);
 * var options = {hint: {min: 'Enter a value greater than {min}'}, min: 100};
 * var lrValidator = lrvf.createValidator(options);
 * 
 * @name oj.NumberRangeValidatorFactory
 * @hideconstructor
 * @class
 * @public
 * @since 0.6
 * 
 */
oj.NumberRangeValidatorFactory = (function () 
{
  
  function _createNumberRangeValidator(options) 
  {
    return new oj.NumberRangeValidator(options);
  }
  
  return {
    /**
     * Creates an immutable validator instance of type {@link oj.NumberRangeValidator} that ensures 
     * that the value provided is within a given range.
     * 
     * @param {Object=} options an object literal used to provide the minimum, maximum and other 
     * optional values. See {@link oj.NumberRangeValidator} for details.<p>
     * 
     * @return {oj.NumberRangeValidator}
     * @memberOf oj.NumberRangeValidatorFactory
     * @instance
     * @ojsignature {target: "Type", for: "options", value: "oj.NumberRangeValidator.ValidatorOptions"}
     * @public
     */
    'createValidator': function(options) {
      return _createNumberRangeValidator(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultValidatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_NUMBERRANGE,
                                                oj.NumberRangeValidatorFactory);
    
/**
 * Copyright (c) 2016, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * This is a forked version of globalize.js
 */
/*
 * Copyright Software Freedom Conservancy, Inc.
 * Dual licensed under the MIT or GPL Version 2 licenses.
 * http://jquery.org/license
 */
/**
 * @constructor
 * 
 * @classdesc OraNumberConverter object implements number parsing and formatting for
 * decimal, currency, percent and perMill types. It supports ECMA-402 options 
 * and user defined pattern. The user defined pattern is parsed in order to 
 * derive the options that can be specified as ECMA options.
 * There are several ways to use the converter.
 * <p>
 * <ul>
 * <li>Using options defined by the ECMA 402 Specification, these would be the properties style, 
 * currency, currencyDisplay, minimumIntegerDigits, minimumFractionDigits, maximumFractionDigits, 
 * useGrouping. NOTE: minimumSignificantDigits and maximumSignificantDigits are not supported.</li>
 * <li>Using a custom decimal, currency or percent format pattern. specified using the 'pattern' property</li>
 * <li>Using the decimalFormat option to define a compact pattern, such as "1M" and "1 million".</li>
 * <li>Using the currencyFormat option to define a compact pattern, such as "$1M" and "$1 million".</li>
 * <li>Using the roundingMode and roundDuringParse options to round the number HALF_UP, HALF_DOWN, or HALF_EVEN.</li>
 * </ul>
 * <p>
 * 
 * The converter provides leniency when parsing user input value to a number in the following ways:<br/>
 * 
 * <ul>
 * <li>Prefix and suffix that do not match the pattern, are removed. E.g., when pattern is 
 * "#,##0.00%" (suffix is the % character), a value of "abc-123.45xyz", will be leniently parsed to 
 * -123.45</li>
 * <li>When a value includes a symbol but the pattern doesn't require it.  E.g., the options are 
 * {pattern: "###", currency: 'USD'}, then values ($123), (123) and -123 will be leniently parsed as 
 * -123.</li>
 * </ul>
 * <p>
 * Lenient parse can be disabled by setting the property lenientParse to "none". In which case the user input must
 * be an exact match of the expected pattern and all the leniency described above will be disabled.
 * <p>
 * @property {Object=} options - an object literal used to provide optional information to 
 * initialize the converter.
 * @property {string=} options.style - sets the style of number formatting. Allowed values are "decimal" 
 * (the default), "currency", "percent" or "unit". When a number is formatted as a decimal, the decimal 
 * character is replaced with the most appropriate symbol for the locale. In English this is a 
 * decimal point ("."), while in many locales it is a decimal comma (","). If grouping is enabled the 
 * locale dependent grouping separator is also used. These symbols are also used for numbers 
 * formatted as currency or a percentage, where appropriate.
 * @property {string=} options.currency - specifies the currency that will be used when formatting the 
 * number. The value should be a ISO 4217 alphabetic currency code. If the style is set to currency, 
 * it's required that the currency property also be specified. This is because there is no default 
 * currency associated with the current locale. The user must always specify the currency code 
 * to be shown, otherwise an error will be thrown. The current page locale 
 * (returned by oj.Config.getLocale()) determines the formatting elements of the number 
 * like grouping separator and decimal separator. The currency code tells us which currency to 
 * display in current page locale. JET has translations for currency names.
 * <p>
 * As an example if we want to format 1000.35 EURO and the page locale is "en-US", 
 * we pass {style:'currency', currency:'EUR', currencyDisplay:'symbol'} and we will get "€1,000.35"
 * If the page locale is "fr-FR", with the same options, we will get: "1 000,35 €"
 * </p>
 * @property {string=} options.unit - Mandatory when style is "unit". Allowed values: 
 * "byte" or "bit". It is used for formatting only. It can not be used for parsing.
 * <p> 
 * It is used to format digital units like 10Mb for bit unit or 10MB for byte unit. 
 * There is no need to specify the scale of the unit. We automatically detect it.
 * For example 1024 is formatted as 1KB and 1048576 as 1MB.
 * The user can also specify 'minimumFractionDigits' and  'maximumFractionDigits' to be displayed, 
 * otherwise we use the locale's default max and min fraction digits.
 * </p> 
 * @property {string=} options.currencyDisplay - if the number is using currency formatting, specifies 
 * if the currency will be displayed using its "code" (as an ISO 4217 alphabetic currency code), 
 * "symbol" (a localized currency symbol (e.g. $ for US dollars, £ for Great British pounds, and so 
 * on), or "name" (a localized currency name. Allowed values are "code", "symbol" and "name". 
 * The default is "symbol".
 * @property {string=} options.decimalFormat -
 * specifies the decimal format length to use when style is set to "decimal". 
 * Allowed values are : "standard"(default), "short" and "long". "standard" is equivalent to not 
 * specifying the 'decimalFormat' attribute, in that case the locale's default decimal pattern 
 * is used for formatting.
 * <p>
 * The user can also specify 'minimumFractionDigits' and  'maximumFractionDigits' to display. 
 * When not present we use the locale's default max and min fraction digits.
 * </p>
 * <p>
 * There is no need to specify the scale; we automatically detect greatest scale that is less or 
 * equal than the input number. For example  1000000 is formatted as "1M" or "1 million" and
 * 1234 is formatted, with zero fractional digits, as "1K" or " 1 thousand" for 
 * short and long formats respectively. The pattern for the short and long number is locale dependent 
 * and uses plural rules for the particular locale.
 * </p>
 * <p>
 * NOTE: Currently this option formats a value (e.g., 2000 -> 2K), but it does not parse a value 
 * (e.g., 2K -> 2000), so it can only be used
 * in a readOnly EditableValue because readOnly EditableValue components do not call
 * the converter's parse function.
 * </p>
 * @property {string=} options.currencyFormat -
 * specifies the currency format length to use when style is set to "currency". 
 * Allowed values are : "standard"(default), "short" and "long". 'standard' is equivalent to not 
 * specifying the 'currencyFormat' attribute, in that case the locale's default currency pattern 
 * is used for formatting.
 * Similar to decimalFormat, currencyFormat can only be used for formatting. It can not be used for parsing.
 * <p>
 * The user can also specify 'minimumFractionDigits' and  'maximumFractionDigits' to display. 
 * When not present we use the locale's default max and min fraction digits.
 * </p>
 * <p>
 * There is no need to specify the scale; we automatically detect greatest scale that is less or 
 * equal than the input number. For example  1000000 is formatted as "$1M" or "1 million dollar" and
 * 1000 is formatted as "$1K" or " 1 thousand dollar" for short and long formats respectively.
 * The pattern for the short and long number is locale dependent and uses plural rules for the particular locale.
 * </p>
 * @property {number=} options.minimumIntegerDigits - sets the minimum number of digits before the 
 * decimal place (known as integer digits). The number is padded with leading zeros if it would not 
 * otherwise have enough digits. The value must be an integer between 1 and 21.
 * @property {number=} options.minimumFractionDigits - similar to 'minimumIntegerDigits', except it 
 * deals with the digits after the decimal place (fractional digits). It must be an integer between 
 * 0 and 20. The fractional digits will be padded with trailing zeros if they are less than the minimum.
 * @property {number=} options.maximumFractionDigits - follows the same rules as 'minimumFractionDigits', 
 * but sets the maximum number of fractional digits that are allowed. The value will be rounded if 
 * there are more digits than the maximum specified.
 * @property {boolean=} options.useGrouping - when the value is truthy, the locale dependent grouping 
 * separator is used when formatting the number. This is often known as the thousands separator, 
 * although it is up to the locale where it is placed. The 'useGrouping' is set to true by default.
 * @property {string=} options.pattern an optional localized pattern, where the characters used in 
 * pattern are as defined in the Unicode CLDR for numbers, percent or currency formats. When present 
 * this will override the other "options". <p>
 * 
 * &nbsp;&nbsp;- When the pattern represents a currency style the 'currency' property is required to 
 * be set, as not setting this will throw an error. The 'currencyDisplay' is optional. <br/>Example: 
 * {pattern: '¤#,##0', currency: 'USD'}. <p>
 * 
 * &nbsp;&nbsp;- It's not mandatory for the pattern to have the special character '¤' (currency sign) 
 * be present. When not present, values are treated as a currency value, but are not formatted to 
 * show the currency symbol. <br/>Example: {pattern: '#,##0', currency: 'USD'} <p>
 * 
 * &nbsp;&nbsp;- When the pattern represents a percent style, the percent special character ('%') needs to be 
 * explicitly specified in the pattern, e.g., {pattern: "#,##0%"}. If the pattern does not contain 
 * the percent character it's treated as a decimal pattern, unless the style is set to percent, 
 * in which case the value is treated as a percent value, but not formatted to show the percent symbol. 
 * <br/>Example: {style: 'percent', pattern: "#,##0"}. <p>
 * 
 * &nbsp;&nbsp;- A decimal pattern or exponent pattern is specified in the pattern using the CLDR 
 * conventions. <br/>Example: {pattern: "#,##0.00"} or {pattern: "0.##E+0"}. <p>
 * 
 * NOTE: 'pattern' is provided for backwards compatibility with existing apps that may want the 
 * convenience of specifying an explicit format mask. Setting a pattern will override the default 
 * locale specific format. <br/>
 * 
 * @property {string=} options.roundingMode - specifies the rounding behavior. 
 * This follows the Java.Math.RoundingMode behavior.
 * Currently we support the options: HALF_UP, HALF_DOWN, and HALF_EVEN 
 * 
 * @property {boolean=} options.roundDuringParse - Specifies whether or not to round during
 * parse. Defaults to false; the number converter rounds during format but not during parse.
 * 
 * @property {Object=} options.separators - An object with 2 fields: 'decimal' and 'group'.
 * It allows the user to override the locale's default decimal and grouping separators. It is accepted for both
 * format and parse methods. 
 * <br/>
 * 
 * @property {string=} options.lenientParse - The lenientParse property can be used to enable or disable leninet parsing.
 *  Allowed values: "full" (default), "none". 
 * <p style='padding-left: 5px;'>
 * By default the lenient parse is enabled and the leniency rules descibed above will be used. When lenientParse is
 * set to "none" the lenient parse is disabled and the user input must match the expected input otherwise an exception will
 * be thrown.<br/><br/>
 * 
 * @example <caption>Create a number converter for currencies</caption>
 * var converter = OraNumberConveter.getInstance();
 * var options = {style: "currency", currency: "USD", minimumIntegerDigits: 2};
 * var localeElements;
 * var nb = 9;
 * converter.format(nb, localeElements, options); --> "$09.00" if page locale is 'en-US'
 * converter.format(nb, localeElements, options); --> "09,00 $US" if page locale is 'fr-FR'<br/>
 * 
 * @example <caption>Options for percent values using a custom (CLDR) pattern</caption>
 * var options = {pattern: '#,##0%'};
 *converter = converterFactory.createConverter(options);<br/>
 * 
 * @example <caption>To parse a value as percent but format it without displaying the percent character</caption>
 * var options = {style: 'percent', pattern: '#,##0'};<br/>
 * 
 * @example <caption>To parse a value as currency using a custom (CLDR) pattern</caption>
 * var options = {pattern: '¤#,##0', currency: 'USD'};
 * 
 * @example <caption>To format a value as digital bit unit</caption>
 * var options = {style:'unit', unit:'bit'};
 * var nb = 1024;
 * converter.format(nb, localeElements, options);--> 1Kb<br/>
 * 
 * @example <caption>To format a value as digital byte unit</caption>
 * var options = {style:'unit', unit:'byte'};
 * var nb = 1024;
 * converter.format(nb, localeElements, options);--> 1KB<br/>
 * 
 * @example <caption>The following decimalFormat examples are in en locale.
 * To format a value as short (default for fraction digits is based on the locale)</caption>
 * var options = {style:'decimal', decimalFormat:'short'};
 * var nb = 12345
 * converter.format(nb, localeElements, options);--> 12.354K<br/>
 * 
 * @example <caption>To format a value as long (default for fraction digits is based on the locale):</caption>
 * var options = {style:'decimal', decimalFormat:'long'};
 * var nb = 12345;
 * converter.format(nb, localeElements, options);--> 12.345 thousand<br/>
 * 
 * @example <caption>To format a value as short with minimum fraction digits:</caption>
 * options = { style:'decimal', decimalFormat:'short', minimumFractionDigits:4};
 * var nb = 1234;
 * converter.format(nb, localeElements, options);--> 1.2340K<br/>
 * 
 * @example <caption>To format a value as short with maximum fraction digits:</caption>
 * options = { style:'decimal', decimalFormat:'short', maximumFractionDigits:0};
 *  var nb = 1234;
 * converter.format(nb, localeElements, options);--> 12K<br/>
 * 
 * @example <caption>To format a value as long with minimum and maximum fraction digits:</caption>
 * options = { style:'decimal', decimalFormat:'long', 
 * minimumFractionDigits:2, maximumFractionDigits:4};
 * var nb = 12000;
 * converter.format(nb, localeElements, options);--> 12.00 thousand<br/>
 * 
 * @example <caption>To format a value as short with minimum and maximum fraction digits:</caption>
 * options = { style:'decimal', decimalFormat:'long', 
 * minimumFractionDigits:2, maximumFractionDigits:4};
 * var nb = 12345678;
 * converter.format(nb, localeElements, options);--> 12.345 million<br/>
 * 
 * @example <caption>decimal style default is standard:</caption>
 * options = { style:'decimal', decimalFormat:'standard'}; 
 * var nb = 12345;
 * converter.format(nb, localeElements, options);--> 12,345<br/>
 * 
 * @example <caption>decimal round HALF_DOWN:</caption>
 * options = { style:'decimal',  maximumFractionDigits:2, roundingMode:'HALF_DOWN'}; 
 * var nb = 0.225;
 * converter.format(nb, localeElements, options);--> 0.22
 * var str = "0.225";
 * converter.parse(str, localeElements, options);-->0.225 //doesn't round during parse by default<br/>
 * 
 * @example <caption>decimal round HALF_UP:</caption>
 * options = { style:'decimal',  maximumFractionDigits:2, roundingMode:'HALF_UP'}; 
 * var nb = 0.225;
 * converter.format(nb, localeElements, options);--> 0.23
 * var str = "0.225";
 * converter.parse(str, localeElements, options);--> 0.225 //doesn't round during parse by default<br/>
 * 
 * @example <caption>decimal round HALF_EVEN:</caption>
 * options = { style:'decimal',  maximumFractionDigits:2, roundingMode:'HALF_EVEN'}; 
 * converter.format(0.225, localeElements, options);--> 0.22
 * converter.format(0.235, localeElements, options);--> 0.24
 * converter.parse("0.225", localeElements, options);--> 0.225 //doesn't round during parse by default
 * converter.parse("0.235", localeElements, options);--> 0.235 //doesn't round during parse by default<br/>
 * 
 * @example <caption>decimal round HALF_DOWN and roundDuringParse:</caption>
 * options = { style:'decimal', maximumFractionDigits:2, 
 *             roundingMode:'HALF_DOWN', roundDuringParse: true};
 * var nb = 0.225; 
 * converter.format(nb, localeElements, options);--> 0.22
 * var str = "0.225";
 * converter.parse(str, localeElements, options);-->0.22<br/>
 * 
 * @example <caption>decimal round HALF_UP and roundDuringParse:</caption>
 * options = { style:'decimal',  maximumFractionDigits:2, 
 *             roundingMode:'HALF_UP', roundDuringParse: true}; 
 * var nb = 0.225;
 * converter.format(nb, localeElements, options);--> 0.23
 * var str = "0.225";
 * converter.parse(str, localeElements, options);--> 0.23<br/>
 * 
 * @example <caption>decimal round HALF_EVEN and roundDuringParse:</caption>
 * options = { style:'decimal',  maximumFractionDigits:2, 
 *             roundingMode:'HALF_EVEN', roundDuringParse: true}; 
 * converter.format(0.225, localeElements, options);--> 0.22
 * converter.format(0.235, localeElements, options);--> 0.24
 * converter.parse("0.225", localeElements, options);--> 0.22
 * converter.parse("0.235", localeElements, options);--> 0.24<br/>
 * 
 * @example <caption>Override locale's decimal and grouping separators:</caption>
 * in en-US locale, the decimal separator is '.' and grouping separator is ','. In this example we will swap them.
 * options = { style:'decimal', separators: {decimal: ',',  group: '.'}};
 * var nb = 1234567.89; 
 * converter.format(nb, localeElements, options);--> 1.234.567,89
 * converter.parse("1.234.567,89", localeElements, options);--> 1234567.89
 * <br/>
 * 
 * @example <caption>Disable lenient parse:</caption>
 * options = { style:'decimal',  lenientParse: 'none'}; 
 * converter.parse("abc-123.45xyz", localeElements, options);--> Error: Unparsable number abc-123.45xyz The expected number pattern is #,##0.###
 * <br/>
 * 
 * @name OraNumberConverter
 * @ignore
 */

/**
 * @ignore
 */
var OraNumberConverter;

OraNumberConverter = (function () {
  var _zeroPad;
  var _formatNumberImpl;
  var _parseNumberImpl;
  var _getLatnDigits;
  var _getNumberParts;
  var _throwNaNException;
  var _applyPatternImpl;
  var _parseNegativePattern;
  var _lenientParseNumber;
  var _parseNegativeExponent;
  var _getNumberSettings;
  var _validateNumberOptions;
  var _getRoundedNumber;
  var _throwMissingCurrency;
  var _getNumberingSystemKey;
  var _getBCP47Lang;
  var _throwNumberOutOfRange;
  var _roundNumber;
  var _decimalAdjust;
  var _getRoundingMode;
  var _getNumberOption;
  var _getNumberingExtension;
  var _adjustRoundingMode;
  var _getParsedValue;
  var _throwUnsupportedParseOption;
  var _toRawFixed;
  var _toExponentialPrecision;
  var _toCompactNumber;
  var _toDigitalByte;
 var _throwMissingUnit;
  var instance;
  var _regionMatches;
  var _expandAffix;
  var _expandAffixes;
  var _throwSyntaxError;
  var _resolveNumberSettings;
  var _resolveOptions;

  var _REGEX_INFINITY = /^[+\-]?infinity$/i;
  var _REGEX_PARSE_FLOAT = /^[+\-]?\d*\.?\d*(e[+\-]?\d+)?$/;
  var _LENIENT_REGEX_PARSE_FLOAT = /([^+-.0-9]*)([+\-]?\d*\.?\d*(E[+\-]?\d+)?).*$/;
  var _ESCAPE_REGEXP = /([\^\$\.\*\+\?\|\[\]\(\)\{\}])/g;
  var _REGEX_TRIM_ZEROS = /(^0\.0*)([^0].*$)/;

  var _decimalTypeValues = {
    'trillion': [100000000000000, 10000000000000, 1000000000000],
    'billion': [100000000000, 10000000000, 1000000000],
    'million': [100000000, 10000000, 1000000],
    'thousand': [100000, 10000, 1000]
  };

  var _decimalTypeValuesMap = {
    'trillion': 1000000000000,
    'billion': 1000000000,
    'million': 1000000,
    'thousand': 1000
  };

  //maps roundingMode attributes to Math rounding modes. 
  var _roundingModeMap = {
    'HALF_UP': 'ceil',
    'HALF_DOWN': 'floor',
    'DEFAULT': 'round'
  };

  var _DIGITAL_KILO = 1024;
  var _DIGITAL_MEGA = 1024 * 1024;
  var _DIGITAL_GIGA = 1024 * 1024 * 1024;
  var _DIGITAL_TERA = 1024 * 1024 * 1024 * 1024;


  //prepend or append count zeros to a string.
  _zeroPad = function (str, count, left) {
    var l;
    for (l = str.length; l < count; l += 1) {
      str = (left ? ("0" + str) : (str + "0"));
    }
    return str;
  };

  _throwNumberOutOfRange = function (value, minimum, maximum, property) {
    var msg = value +
        " is out of range.  Enter a value between " + minimum +
        " and " + maximum + " for " + property;
    var rangeError = new RangeError(msg);
    var errorInfo = {
      'errorCode': "numberOptionOutOfRange",
      'parameterMap': {
        'value': value,
        'minValue': minimum,
        'maxValue': maximum,
        'propertyName': property
      }
    };
    rangeError['errorInfo'] = errorInfo;
    throw rangeError;
  };

  _getNumberOption = function (options, property, minimum, maximum, fallback) {
    var value = options[property];
    if (value !== undefined) {
      value = Number(value);
      if (isNaN(value) || value < minimum || value > maximum) {
        _throwNumberOutOfRange(value, minimum, maximum, property);
      }
      return Math.floor(value);
    }
    else {
      return fallback;
    }
  };

  //get the numbering system key from the locale's unicode extension.
  //Verify that the locale data has a numbers entry for it, if not return latn as default.
  _getNumberingSystemKey = function (localeElements, locale) {
    if (locale === undefined)
      return 'latn';
    var numberingSystemKey = _getNumberingExtension(locale);
    var symbols = "symbols-numberSystem-" + numberingSystemKey;
    if (localeElements['numbers'][symbols] === undefined)
      numberingSystemKey = 'latn';
    return numberingSystemKey;
  };

  //return the language part
  _getBCP47Lang = function (tag) {
    var arr = tag.split("-");
    return arr[0];
  };

  //get the unicode numbering system extension.
  _getNumberingExtension = function (locale) {
    locale = locale || "en-US";
    var idx = locale.indexOf("-u-nu-");
    var numbering = 'latn';
    if (idx !== -1) {
      numbering = locale.substr(idx + 6, 4);
    }
    return numbering;
  };

  /*return the properties for a number such as minimum and maximum fraction 
   *digits, decimal separator, grouping separator.
   *-If no user defined pattern is provided, get the pattern from the locale
   *  data and parse it to extrcat the number properties. If ecma options are
   *  present, override the corresponding default properties.
   *-If a user defined pattern is provided, parse it and extrcat the number
   *  properties. Ignore ecma ptions if present.
   */

  _getNumberSettings = function (localeElements, numberSettings,
      options, locale) {
    var pat;
    var localeElementsMainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var numberingSystemKey = _getNumberingSystemKey(localeElementsMainNode, locale);
    numberSettings['numberingSystemKey'] = numberingSystemKey;
    numberSettings['numberingSystem'] = "symbols-numberSystem-" +
        numberingSystemKey;
    var lenient = options['lenientParse'];
    numberSettings['lenientParse'] = lenient || 'full';
    //pattern passed in options
    if (options['pattern'] !== undefined && options['pattern'].length > 0) {
      pat = options['pattern'];
    }
    else
    {
      var key;
      switch (options['style'])
      {
        case "decimal" :
          key = "decimalFormats-numberSystem-";
          break;
        case "currency" :
          key = "currencyFormats-numberSystem-";
          break;
        case "percent" :
          key = "percentFormats-numberSystem-";
          break;
        default:
          key = "decimalFormats-numberSystem-";
          break;
      }
      key += numberSettings['numberingSystemKey'];
      pat = localeElementsMainNode['numbers'][key]['standard'];
      //check if decimalFormat is set
      var decFormatLength = options['decimalFormat'];
      //if not, check for currencyFormat
      if (decFormatLength === undefined)
        decFormatLength = options['currencyFormat'];
      //if either decimalFormat or currencyFormat is set, save it in number settings
      if (decFormatLength !== undefined && (options['style'] === 'decimal' || options['style'] === 'currency')) {
        numberSettings['shortDecimalFormat'] = localeElementsMainNode['numbers']['decimalFormats-numberSystem-latn'][decFormatLength]['decimalFormat'];
      }
    }
    var decimalSeparator = localeElementsMainNode['numbers'][numberSettings['numberingSystem']]['decimal'];
    var groupSeparator = localeElementsMainNode['numbers'][numberSettings['numberingSystem']]['group'];
    var separators = options['separators'];
    if (separators !== undefined) {
      numberSettings['separators'] = separators;
      var dec = separators['decimal'];
      var grp = separators['group'];
      if (dec !== undefined && dec !== '')
        decimalSeparator = separators['decimal'];
      if (grp !== undefined)
        groupSeparator = separators['group'];
    }
    var mainNodeKey = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
    var lang = _getBCP47Lang(mainNodeKey);
    numberSettings['plurals'] = localeElements['supplemental']['plurals'];
    numberSettings['lang'] = lang;
    numberSettings['pat'] = pat;
    numberSettings['minusSign'] = localeElementsMainNode['numbers']
    [numberSettings['numberingSystem']]['minusSign'];
    numberSettings['decimalSeparator'] = decimalSeparator;
    numberSettings['exponential'] = localeElementsMainNode['numbers']
    [numberSettings['numberingSystem']]['exponential'];
    numberSettings['groupingSeparator'] = groupSeparator;
    numberSettings['currencyDisplay'] = options['currencyDisplay'];
    if (options['currency'] !== undefined)
      numberSettings['currencyCode'] = options['currency'].toUpperCase();
    if (options['unit'] !== undefined)
      numberSettings['unit'] = options['unit'].toLowerCase();
    numberSettings['style'] = options['style'];
    _applyPatternImpl(options, pat, localeElementsMainNode, numberSettings);
    if (options['pattern'] === undefined) {
      numberSettings['minimumIntegerDigits'] = _getNumberOption(options,
          'minimumIntegerDigits', 1, 21,
          numberSettings['minimumIntegerDigits']);
      if (options['maximumFractionDigits'] !== undefined) {
        numberSettings['maximumFractionDigits'] = _getNumberOption(options,
            'maximumFractionDigits', 0, 20, numberSettings['maximumFractionDigits']);
        if (numberSettings['maximumFractionDigits'] < numberSettings['minimumFractionDigits']) {
          numberSettings['minimumFractionDigits'] = numberSettings['maximumFractionDigits'];
        }
      }
      if (options['minimumFractionDigits'] !== undefined) {
        numberSettings['minimumFractionDigits'] = _getNumberOption(options,
            'minimumFractionDigits', 0, 20,
            numberSettings['minimumFractionDigits']);
      }
      if (numberSettings['maximumFractionDigits'] < numberSettings['minimumFractionDigits']) {
        numberSettings['maximumFractionDigits'] = numberSettings['minimumFractionDigits'];
        oj.Logger.info(
        "maximumFractionDigits is less than minimumFractionDigits, so maximumFractionDigits will be set to minimumFractionDigits");
      }
    }
  };

  _throwMissingCurrency = function (prop) {
    var typeError = new TypeError('The property "currency" is required when' +
        ' the property "' + prop + '" is "currency". An accepted value is a ' +
        'three-letter ISO 4217 currency code.');
    var errorInfo = {
      'errorCode': 'optionTypesMismatch',
      'parameterMap': {
        'propertyName': prop, // the driving property
        'propertyValue': 'currency', // the driving property's value
        'requiredPropertyName': 'currency', // the required property name
        'requiredPropertyValueValid': 'a three-letter ISO 4217 currency code'
      }
    };
    typeError['errorInfo'] = errorInfo;
    throw typeError;
  };

  _throwMissingUnit = function (prop) {
    var typeError = new TypeError('The property "unit" is required when' +
        ' the property "' + prop + '" is "unit". An accepted value is  ' +
        '"byte" or "bit".');
    var errorInfo = {
      'errorCode': 'optionTypesMismatch',
      'parameterMap': {
        'propertyName': prop, // the driving property
        'propertyValue': 'unit', // the driving property's value
        'requiredPropertyName': 'unit', // the required property name
        'requiredPropertyValueValid': 'byte or bit'
      }
    };
    typeError['errorInfo'] = errorInfo;
    throw typeError;
  };

  _throwUnsupportedParseOption = function (val) {
    var error, errorInfo;
    var code = "unsupportedParseFormat";
    var msg = "long and short " + val + " are not supported for parsing";
    error = new Error(msg);
    errorInfo = {
      'errorCode': code,
      'parameterMap': {
        'shortFormats': val
      }
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };

  //If the user specifies currency as a style, currency option must also be
  // provided. parse does not support short and long decimalFormat.
  _validateNumberOptions = function (options, caller) {
    var getOption = oj.OraI18nUtils.getGetOption(options, caller);
    var s = getOption('style', 'string', ['currency', 'decimal', 'percent', 'unit', 'perMill'],
        'decimal');
    if (s === 'decimal' || s === 'currency') {
      var fmt = (s === 'decimal') ? 'decimalFormat' : 'currencyFormat';
      s = getOption(fmt, 'string', ['standard', 'short', 'long']);
      if (caller === 'OraNumberConverter.parse' && s !== undefined && s !== 'standard') {
        _throwUnsupportedParseOption(fmt);
      }
    }
    var c = getOption('currency', 'string');
    if (s === 'currency' && c === undefined) {
      _throwMissingCurrency("style");
    }

    c = getOption('unit', 'string');
    if (s === 'unit' && c === undefined) {
      _throwMissingUnit("style");
    }

    var roundingMode = getOption('roundingMode', 'string', ['HALF_UP', 'HALF_DOWN', 'HALF_EVEN'],
        'DEFAULT');
    var lenientParse = getOption('lenientParse', 'string', ['none', 'full'], 'full');
  };

  //_toDigitalByte does compact formatting like 300MB, 300Mb
  _toDigitalByte = function (number, options, numberSettings, localeElements) {
    var scale;
    var count;
    
    if (number >= _DIGITAL_TERA) {
      scale = "digital-tera";
      count = number / _DIGITAL_TERA;
    }
    else if (number >= _DIGITAL_GIGA) {
      scale = "digital-giga";
      count = number / _DIGITAL_GIGA;
    }
    else if (number >= _DIGITAL_MEGA) {
      scale = "digital-mega";
      count = number / _DIGITAL_MEGA;
    }
    else if (number >= _DIGITAL_KILO) {
      scale = "digital-kilo";
      count = number / _DIGITAL_KILO;
    }
    else {
      scale = "digital-";
      count = number;
    }
    //Find the corresponding entry in resource budle under units section
    //scale -> 'digital-kilo-bit' or 'digital-kilo-byte'
    scale = scale + numberSettings['unit'];
    var lang = numberSettings['lang'];
    //get plural rule: one, many, etc..
    var plural = numberSettings['plurals'][lang](count);
    //plural -> 'unitPattern-count-one' or 'unitPattern-count-many'
    plural = "unitPattern-count-" + plural;
    //format the number
    var fmt = _toRawFixed(count, options, numberSettings);
    //format the number based on plural rule: "{0} Gb", etc..
    var entry = localeElements['units']['narrow'][scale][plural];
    fmt = oj.OraI18nUtils.formatString(entry, [fmt]);
    return fmt;
  };

  //_toCompactNumber does compact formatting like 3000->3K for short
  //and "3 thousand" for long
  _toCompactNumber = function (number, options, numberSettings) {

    function _getZerosInPattern(s) {
      var i = 0, n = 0, idx = 0, prefix = '';
      if (s[0] !== '0') {
        while (s[i] !== '0' && i < s.length) {
          i++;
        }
        prefix = s.substr(0, i);
        idx = i;
      }
      for (i = idx; i < s.length; i++) {
        if (s[i] === '0')
          n++;
        else
          break;
      }
      return [prefix, n];
    }

    /* To format a number N, the greatest type less than or equal to N is used, with
     * the appropriate plural category. N is divided by the type, after removing the
     * number of zeros in the pattern, less 1.
     * APIs supporting this format should provide control over the number of
     * significant or fraction digits.
     *Thus N=12345 matches <pattern type="10000" count="other">00 K</pattern>.
     *N is divided by 1000 (obtained from 10000 after removing "00" and restoring
     *one "0". The result is formatted according to the normal decimal pattern.
     *With no fractional digits, that yields "12 K".
     */
    function _matchTypeValue(n) {
      var i, j, len;
      for (i in _decimalTypeValues) {
        len = _decimalTypeValues[i].length;
        for (j = 0; j < len; j++) {
          if (_decimalTypeValues[i][j] <= n)
            return [i, _decimalTypeValues[i][j]];
        }
      }
      return [n, null];
    }


    var typeVal = _matchTypeValue(number);
    var prefix = '';
    if (typeVal[1] !== null) {
      var lang = numberSettings['lang'];
      var plural = numberSettings['plurals'][lang](Math.floor(number / _decimalTypeValuesMap[typeVal[0]]));
      var decimalFormatType = "" + typeVal[1] + "-count-" + plural;
      decimalFormatType = numberSettings['shortDecimalFormat'][decimalFormatType];
      if (decimalFormatType === undefined) {
        plural = "other";
        decimalFormatType = "" + typeVal[1] + "-count-" + plural;
        decimalFormatType = numberSettings['shortDecimalFormat'][decimalFormatType];
      }
      var tokens = _getZerosInPattern(decimalFormatType);
      var zeros = tokens[1];
      prefix = tokens[0];
      if (zeros < decimalFormatType.length) {
        var i = (1 * Math.pow(10, zeros));
        i = (typeVal[1] / i) * 10;
        number = number / i;
      }
    }
    var s = "";
    var fmt;
    if (decimalFormatType !== undefined)
      s = decimalFormatType.substr(zeros + tokens[0].length);
    fmt = _toRawFixed(number, options, numberSettings);
    var regExp = /'\.'/g;
    s = s.replace(regExp, ".");
    s = prefix + fmt + s;
    return s;
  };

  //_toExponentialPrecision does the formatting when the pattern contain E,
  //for example #.#E0  
  _toExponentialPrecision = function (number, numberSettings) {
    var numStr0 = number + "";
    var trimExp = 0;
    var split = numStr0.split(/e/i);
    var numStr = split[0];
    _REGEX_TRIM_ZEROS.lastIndex = 0;
    var match = _REGEX_TRIM_ZEROS.exec(numStr);
    if (match !== null) {
      trimExp = match[1].length - 1;
      numStr = match[2];
    }
    else {
      numStr = numStr.replace(".", "");
    }
    var exponent = split.length > 1 ? parseInt(split[1], 10) : 0;
    var numStr1 = parseInt(numStr, 10);
    var len = numberSettings['minimumIntegerDigits'] + numberSettings['maximumFractionDigits'];
    if (numStr.length > len) {
      len -= numStr.length;
      var factor = Math.pow(10, len);
      numStr1 = Math.round(numStr1 * factor);
    }
    var padLen = numberSettings['minimumIntegerDigits'] + numberSettings['minimumFractionDigits'];
    numStr1 = numStr1 + "";
    numStr1 = _zeroPad(numStr1, padLen, false);
    if (numStr0.indexOf('.') !== -1) {
      exponent -= numberSettings['minimumIntegerDigits'] - numStr0.indexOf('.') + trimExp;
    }
    else {
      exponent -= padLen - numStr.length - numberSettings['minimumFractionDigits'];
    }
    var posExp = Math.abs(exponent);
    posExp = _zeroPad(posExp + "", numberSettings['minExponentDigits'], true);
    if (exponent < 0)
      posExp = numberSettings['minusSign'] + posExp;
    var str1 = numStr1.slice(0, numberSettings['minimumIntegerDigits']);
    var str2 = numStr1.slice(numberSettings['minimumIntegerDigits']);
    if (str2.length > 0) {
      str1 += numberSettings['decimalSeparator'] + numStr1.slice(numberSettings['minimumIntegerDigits']) +
          numberSettings['exponential'] + posExp;
    }
    else {
      str1 += numberSettings['exponential'] + posExp;
    }
    return str1;
  };

  //_toRawFixed does the formatting based on
  //minimumFractionDigits and maximumFractionDigits.
  _toRawFixed = function (number, options, numberSettings) {
    var curSize = numberSettings['groupingSize'];
    var curSize0 = numberSettings['groupingSize0'];
    var decimalSeparator = numberSettings['decimalSeparator'];
    //First round the number based on maximumFractionDigits
    var numberString = number + "";
    var split = numberString.split(/e/i);
    var exponent = split.length > 1 ? parseInt(split[1], 10) : 0;
    numberString = split[ 0 ];
    split = numberString.split('.');
    var right = split.length > 1 ? split[ 1 ] : "";
    var precision;
    //round the number only if it has decimal points
    if (split.length > 1 && right.length > exponent)
    {
      precision = Math.min(numberSettings['maximumFractionDigits'],
          right.length - exponent);
      var mode = options['roundingMode'] || 'DEFAULT';
      number = _roundNumber(number, precision, mode);
    }
    //split the number into integer, fraction and exponent parts.
    numberString = number + "";
    split = numberString.split(/e/i);
    exponent = split.length > 1 ? parseInt(split[1], 10) : 0;
    numberString = split[ 0 ];
    split = numberString.split('.');
    numberString = split[ 0 ];
    right = split.length > 1 ? split[ 1 ] : "";
    //pad zeros based on the exponent value and minimumFractionDigits 
    if (exponent > 0) {
      right = _zeroPad(right, exponent, false);
      numberString += right.slice(0, exponent);
      right = right.substr(exponent);
    }
    else if (exponent < 0) {
      exponent = -exponent;
      numberString = _zeroPad(numberString, exponent + 1, true);
      right = numberString.slice(-exponent, numberString.length) + right;
      numberString = numberString.slice(0, -exponent);
    }
    if (precision > 0 && right.length > 0) {
      right = decimalSeparator +
          ((right.length > precision) ? right.slice(0, precision) :
              _zeroPad(right, precision, false));
    }
    else {
      if (numberSettings['minimumFractionDigits'] > 0) {
        right = decimalSeparator;
      }
      else {
        right = "";
      }
    }
    //insert grouping separator in the integer part based on groupingSize
    var padLen = decimalSeparator.length +
        numberSettings['minimumFractionDigits'];
    right = _zeroPad(right, padLen, false);
    var sep = numberSettings['groupingSeparator'],
        ret = "";
    if (options['useGrouping'] === false && options['pattern'] === undefined)
      sep = '';
    numberString = _zeroPad(numberString,
        numberSettings['minimumIntegerDigits'], true);
    var stringIndex = numberString.length - 1;
    right = right.length > 1 ? right : "";
    var rets;
    while (stringIndex >= 0) {
      if (curSize === 0 || curSize > stringIndex) {
        rets = numberString.slice(0, stringIndex + 1) +
            (ret.length ? (sep + ret + right) : right);
        return rets;
      }
      ret = numberString.slice(stringIndex - curSize + 1, stringIndex + 1) +
          (ret.length ? (sep + ret) : "");
      stringIndex -= curSize;
      if (curSize0 > 0) {
        curSize = curSize0;
      }
    }
    rets = numberString.slice(0, stringIndex + 1) + sep + ret + right;
    return rets;
  };

  //HALF_DOWN behaves as HALF_UP if the discarded fraction is > 0.5
  _adjustRoundingMode = function (value, maxDigits, mode) {
    if (mode === 'HALF_DOWN' || mode === 'HALF_EVEN') {
      var n = value.substr(maxDigits);
      n = parseInt(n, 10);
      if (n > 5)
        mode = 'HALF_UP';
    }
    return mode;
  };

  _roundNumber = function (value, scale, mode) {
    var parts = value.toString().split('.');
    if (parts[1] === undefined)
      return value;
    if (parts[1][scale] === '5' && mode !== 'DEFAULT') {
      var adjustedMode = _adjustRoundingMode(parts[1], scale, mode);
      adjustedMode = _getRoundingMode(parts, adjustedMode, scale);
      return _decimalAdjust(value, -scale, adjustedMode, parts);
    }
    else {
      var factor = Math.pow(10, scale),
          rounded = Math.round(value * factor) / factor;
      if (!isFinite(rounded)) {
        return value;
      }
      return rounded;
    }
  };

  _getRoundingMode = function (parts, rMode, scale) {
    var mode = _roundingModeMap[rMode];
    if (rMode === 'HALF_EVEN') {
      var c;
      if (scale === 0) {
        var len = parts[0].length;
        c = parseInt(parts[0][len - 1], 10);
      }
      else {
        c = parseInt(parts[1][scale - 1], 10);
      }
      if (c % 2 === 0) {
        mode = _roundingModeMap['HALF_DOWN'];
      }
      else {
        mode = _roundingModeMap['HALF_UP'];
      }
    }
    return mode;
  };

  /**
   * This function does the actual rounding of the number based on the rounding
   * mode:
   * value is the number to be rounded.
   * scale is the maximumFractionDigits.
   * mode is the rounding mode: ceil, floor, round. 
   * parts is the integer and fraction parts of the value.
   */
  _decimalAdjust = function (value, scale, mode, parts) {
    if (scale === 0) {
      if (parts[1][0] === '5') {
        return Math[mode](value);
      }
      return Math['round'](value);
    }
    var strValue = value.toString().split('e');
    var v0 = strValue[0];
    var v1 = strValue[1];
    //shift the decimal point based on the scale so that we can apply ceil or floor
    //scale is a number, no need to parse it, just parse v1.
    var s = v0 + 'e' + (v1 ? (parseInt(v1, 10) - scale) : -scale);
    var num = parseFloat(s);
    value = Math[mode](num);
    strValue = value.toString().split('e');
    //need to extract v0 and v1 again because value has chnaged after applying Math[mode].
    v0 = strValue[0];
    v1 = strValue[1];
    //shift the decimal point back to its original position
    s = v0 + 'e' + (v1 ? (parseInt(v1, 10) + scale) : scale);
    num = parseFloat(s);
    return num;
  };

  //first call _toRawFixed then add prefixes and suffixes. Display the 
  //number using native digits based on the numbering system
  _formatNumberImpl = function (value, options, localeElements,
      numberSettings, locale) {
    var localeElementsMainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    if (!isFinite(value)) {
      if (value === Infinity) {
        return localeElementsMainNode['numbers'][numberSettings['numberingSystem']]['infinity'];
      }
      if (value === -Infinity) {
        return localeElementsMainNode['numbers'][numberSettings['numberingSystem']]['infinity'];
      }
      return localeElementsMainNode['numbers'][numberSettings['numberingSystem']]['nan'];
    }
    var number = Math.abs(value);
    if (numberSettings['isPercent'] === true ||
        options['style'] === 'percent')
      number *= 100;
    else if (numberSettings['isPerMill'] === true)
      number *= 1000;
    //expand the number
    var formatType = options['decimalFormat'];
    if (formatType === undefined)
      formatType = options['currencyFormat'];
    var optStyle = options['style'];
    if ((optStyle === 'decimal' || optStyle === 'currency')
        && formatType !== undefined
        && formatType !== 'standard')
      number = _toCompactNumber(number, options, numberSettings);
    else if (numberSettings['useExponentialNotation'] === true)
      number = _toExponentialPrecision(number, numberSettings);
    else if (optStyle === 'unit')
      number = _toDigitalByte(number, options, numberSettings, localeElementsMainNode);
    else
      number = _toRawFixed(number, options, numberSettings);
    var ret = "";
    //add negative prefix and suffix if number is negative
    //and the new formatted value isn't zero
    if (value < 0 && (number - 0 !== 0)) {
      ret += numberSettings['negativePrefix'] + number +
          numberSettings['negativeSuffix'];
    }
    //add positive prefix and suffix if number is positive
    else {
      ret += numberSettings['positivePrefix'] + number +
          numberSettings['positiveSuffix'];
    }
    //display the digits based on the numbering system
    var numberingSystemKey = _getNumberingExtension(locale);
    if (oj.OraI18nUtils.numeringSystems[numberingSystemKey] === undefined)
      numberingSystemKey = 'latn';
    if (numberingSystemKey !== 'latn') {
      var idx;
      var nativeRet = [];
      for (idx = 0; idx < ret.length; idx++)
      {
        if (ret[idx] >= '0' && ret[idx] <= '9')
          nativeRet.push(oj.OraI18nUtils.numeringSystems[numberingSystemKey][ret[idx]]);
        else
          nativeRet.push(ret[idx]);

      }
      return nativeRet.join("");
    }
    return ret;
  };

  //remove prefix and suffix, return a sign and value. First try to extract
  //a number using exact match. If it fails try lenient parsing.
  _parseNegativePattern = function (value, options, numberSettings,
      localeElements) {
    var ret;
    var num = oj.OraI18nUtils.trimNumber(value);
    var sign = "";
    var exactMatch = false;
    var posSign = localeElements['numbers'][numberSettings['numberingSystem']]['plusSign'];
    var posSignRegExp = new RegExp("^" + posSign.replace(_ESCAPE_REGEXP, "\\$1"));
    num = num.replace(posSignRegExp, "");
    var nbSettingPosPrefix = oj.OraI18nUtils.trimNumber(numberSettings['positivePrefix']),
        nbSettingPosSuffix = oj.OraI18nUtils.trimNumber(numberSettings['positiveSuffix']),
        nbSettingNegPrefix = oj.OraI18nUtils.trimNumber(numberSettings['negativePrefix']),
        nbSettingNegSuffix = oj.OraI18nUtils.trimNumber(numberSettings['negativeSuffix']);
    //try exact match of negative prefix and suffix
    var posPrefRegexp = new RegExp("^" + (nbSettingPosPrefix ||
        "").replace(_ESCAPE_REGEXP, "\\$1"));
    var posSuffRegexp = new RegExp((nbSettingPosSuffix || "").
        replace(_ESCAPE_REGEXP, "\\$1") + "$");
    var negPrefRegexp = new RegExp("^" + (nbSettingNegPrefix ||
        "").replace(_ESCAPE_REGEXP, "\\$1"));
    var negSuffRegexp = new RegExp((nbSettingNegSuffix ||
        "").replace(_ESCAPE_REGEXP, "\\$1") + "$");

    if (negPrefRegexp.test(num) === true && negSuffRegexp.test(num) === true) {
      num = num.replace(negPrefRegexp, "");
      num = num.replace(negSuffRegexp, "");
      sign = "-";
      exactMatch = true;
    }
    //try exact match of positive prefix and suffix
    else if (posPrefRegexp.test(num) === true && posSuffRegexp.test(num) === true) {
      num = num.replace(posPrefRegexp, "");
      num = num.replace(posSuffRegexp, "");
      sign = "+";
      exactMatch = true;
    }
    //if style is currency, remove currency symbol from prefix and suffix 
    //and try a match
    else if (options['style'] === 'currency') {
      var code = numberSettings['currencyCode'], symbol = code;
      var posPrefix, posSuffix, negPrefix, negSuffix, repStr;
      if (localeElements['numbers']['currencies'][code] !== undefined) {
        symbol = localeElements['numbers']['currencies'][code]['symbol'];
      }
      if (numberSettings['currencyDisplay'] === undefined ||
          numberSettings['currencyDisplay'] === "symbol") {
        repStr = symbol;
      }
      else if (numberSettings['currencyDisplay'] === "code") {
        repStr = code;
      }
      if (repStr !== undefined) {
        posPrefix = (nbSettingPosPrefix || "").replace(
            repStr, "");
        posSuffix = (nbSettingPosSuffix || "").replace(
            repStr, "");
        negPrefix = (nbSettingNegPrefix || "").replace(
            repStr, "");
        negSuffix = (nbSettingNegSuffix || "").replace(
            repStr, "");
        posPrefRegexp = new RegExp(("^" + posPrefix).replace(
            _ESCAPE_REGEXP, "\\$1"));
        posSuffRegexp = new RegExp(posSuffix.replace(
            _ESCAPE_REGEXP, "\\$1") + "$");
        negPrefRegexp = new RegExp(("^" + negPrefix).replace(
            _ESCAPE_REGEXP, "\\$1"));
        negSuffRegexp = new RegExp(negSuffix.replace(
            _ESCAPE_REGEXP, "\\$1") + "$");

        //try  match of positive prefix and suffix
        if (negPrefRegexp.test(num) === true && negSuffRegexp.test(num) === true) {
          num = num.replace(negPrefRegexp, "");
          num = num.replace(negSuffRegexp, "");
          sign = "-";
          exactMatch = true;
        }
        //try exact match of positive prefix and suffix
        else if (posPrefRegexp.test(num) === true && posSuffRegexp.test(num) === true) {
          num = num.replace(posPrefRegexp, "");
          num = num.replace(posSuffRegexp, "");
          sign = "+";
          exactMatch = true;
        }
      }
    }
    if (!exactMatch) {
      if (numberSettings['lenientParse'] === 'full') {
        ret = _lenientParseNumber(num, numberSettings);
        ret[2] = true;
      }
      else {
        _throwNaNException(options['style'], numberSettings, value)
      }
    }
    else
      ret = [sign, num];
    return ret;
  };

  _lenientParseNumber = function (num, numberSettings) {
    // Try to extract the number accoring to the following pattern:
    // optional +- followed by one or many digits followed by optional
    // fraction part followed by optional exponential.
    // use localized +, -, decimal separator, exponential
    // [+-]?\d+(?:\.\d+)?(?:E[+-]?\d+)?/;
    //remove grouping deparator from string
    var groupingSeparator = numberSettings['groupingSeparator'];
    var decimalSeparator = numberSettings['decimalSeparator'];
    var localeMinusSign = numberSettings['minusSign'];
    var plusSign = "+";
    var minusSign = "-";
    var sign = "";
    var dot = "";
    var exponential =
        oj.OraI18nUtils.toUpper(numberSettings['exponential']);
    num = oj.OraI18nUtils.toUpper(num);
    num = num.split(exponential).join("E");
    //remove grouping separator from string
    var groupSep = groupingSeparator;
    num = num.split(groupSep).join("");
    var altGroupSep = groupSep.replace(/\u00A0/g, " ");
    if (groupSep !== altGroupSep) {
      num = num.split(altGroupSep).join("");
    }
    num = num.split(decimalSeparator).join(".");
    if (num.charAt(0) === ".") {
      num = num.substr(1);
      dot = ".";
    }
    //replace localized minus with minus
    num = num.replace(localeMinusSign, minusSign);
    var match = _LENIENT_REGEX_PARSE_FLOAT.exec(num);
    var resNum = dot + match[2];
    if (oj.OraI18nUtils.startsWith(resNum, minusSign)) {
      resNum = resNum.substr(minusSign.length);
      sign = "-";
    }
    else if (oj.OraI18nUtils.startsWith(num, plusSign)) {
      resNum = resNum.substr(plusSign.length);
      sign = "+";
    }
    return [sign, resNum];
  };

  //parse the exponent part of a number
  _parseNegativeExponent = function (value, numberSettings) {
    var neg = numberSettings['minusSign'];
    var pos = numberSettings['plusSign'];
    var ret;
    value = oj.OraI18nUtils.trimNumber(value);
    neg = oj.OraI18nUtils.trimNumber(neg);
    pos = oj.OraI18nUtils.trimNumber(pos);
    if (oj.OraI18nUtils.startsWith(value, neg)) {
      ret = ["-", value.substr(neg.length)];
    }
    else if (oj.OraI18nUtils.startsWith(value, oj.OraI18nUtils.trimNumber(pos))) {
      ret = ["+", value.substr(pos.length)];
    }
    return ret || ["", value];
  };

  _getLatnDigits = function (str, locale) {
    var numberingSystemKey = _getNumberingExtension(locale);
    if (oj.OraI18nUtils.numeringSystems[numberingSystemKey] === undefined)
      return str;
    var idx;
    var latnStr = [];
    for (idx = 0; idx < str.length; idx++) {
      var pos = oj.OraI18nUtils.numeringSystems[numberingSystemKey].indexOf(str[idx]);
      if (pos !== -1)
        latnStr.push(pos);
      else
        latnStr.push(str[idx]);
    }
    var ret = latnStr.join("");
    return ret;
  };

  //split the number into integer, fraction and exponential parts
  _getNumberParts = function (num, numberSettings) {
    var parts = {};
    var decimalSeparator = numberSettings['decimalSeparator'];
    var groupSep = numberSettings['groupingSeparator'];
    num = num.replace(/ /g, "");
    // determine exponent and number
    var exponentSymbol = numberSettings['exponential'];
    var integer;
    var intAndFraction;
    var exponentPos = num.indexOf(exponentSymbol.toLowerCase());
    if (exponentPos < 0)
      exponentPos = num.indexOf(oj.OraI18nUtils.toUpper(exponentSymbol));
    if (exponentPos < 0) {
      intAndFraction = num;
      parts['exponent'] = null;
    }
    else {
      intAndFraction = num.substr(0, exponentPos);
      parts['exponent'] = num.substr(exponentPos + exponentSymbol.length);
    }
    // determine decimal position
    var decSep = decimalSeparator;
    var decimalPos = intAndFraction.indexOf(decSep);
    if (decimalPos < 0) {
      integer = intAndFraction;
      parts['fraction'] = null;
    }
    else {
      integer = intAndFraction.substr(0, decimalPos);
      parts['fraction'] = intAndFraction.substr(decimalPos + decSep.length);
    }
    // handle groups (e.g. 1,000,000)
    integer = integer.split(groupSep).join("");
    var altGroupSep = groupSep.replace(/\u00A0/g, " ");
    if (groupSep !== altGroupSep) {
      integer = integer.split(altGroupSep).join("");
    }
    parts['integer'] = integer;
    return parts;
  };

  _getParsedValue = function (ret, options, numberSettings, errStr) {
    if (isNaN(ret)) {
      _throwNaNException(options['style'], numberSettings, errStr)
    }
    if (numberSettings['isPercent'] === true || options['style'] ===
        'percent')
      ret /= 100;
    else if (numberSettings['isPerMill'] === true)
      ret /= 1000;
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraNumberConverter.parse");
    var roundDuringParse = getOption('roundDuringParse', 'boolean', [true, false], false);
    if (roundDuringParse) {
      ret = _getRoundedNumber(ret, numberSettings, options);
    }
    return ret;
  };

  _throwNaNException = function (style, numberSettings, errStr) {
    var msg, error, errorInfo, code;
    msg = "Unparsable number " + errStr + " The expected number " +
        "pattern is " + numberSettings['pat'];
    switch (style)
    {
      case "decimal" :
        code = "decimalFormatMismatch";
        break;
      case "currency" :
        code = "currencyFormatMismatch";
        break;
      case "percent" :
        code = "percentFormatMismatch";
        break;
    }
    error = new Error(msg);
    errorInfo = {
      'errorCode': code,
      'parameterMap': {
        'value': errStr,
        'format': numberSettings['pat']
      }
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };

  _parseNumberImpl = function (str, localeElements, options, locale) {
    var localeElementsMainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var numberSettings = {};
    var numStr = _getLatnDigits(str, locale);
    _getNumberSettings(localeElements, numberSettings, options, locale);
    var ret = NaN;
    var value1 = numStr.replace(/ /g, "");
    // allow infinity or hexidecimal
    if (_REGEX_INFINITY.test(value1)) {
      ret = parseFloat(numStr);
      return ret;
    }
    var signInfo = _parseNegativePattern(numStr, options, numberSettings,
        localeElementsMainNode);
    var sign = signInfo[ 0 ];
    var num = signInfo[ 1 ];
    sign = sign || "+";
    if (signInfo[2]) {
      ret = parseFloat(sign + num);
      return  _getParsedValue(ret, options, numberSettings, str);
    }

    var parts = _getNumberParts(num, numberSettings);
    var integer = parts['integer'];
    var fraction = parts['fraction'];
    var exponent = parts['exponent'];

    // build a natively parsable number string
    var p = sign + integer;
    if (fraction !== null) {
      p += "." + fraction;
    }
    if (exponent !== null) {
      // exponent itself may have a number pattern
      var expSignInfo = _parseNegativeExponent(exponent, numberSettings);
      p += "e" + (expSignInfo[0] || "+") + expSignInfo[ 1 ];
    }
    if (_REGEX_PARSE_FLOAT.test(p)) {
      ret = parseFloat(p);
    }
    else {
      if (numberSettings['lenientParse'] === 'full') {
        p = _lenientParseNumber(numStr, numberSettings);
        ret = parseFloat(p[0] + p[1]);
      }
      else {
        _throwNaNException(options['style'], numberSettings, str);
      }
    }
    return  _getParsedValue(ret, options, numberSettings, str);
  };

  /* This module handles the  parsing of a number pattern.
   * It sets prefix, suffix, minimum and maximum farcation digits, 
   * miimum  integer digits and grouping size. 
   */

  var _ZERO_DIGIT = '0',
      _GROUPING_SEPARATOR = ',',
      _DECIMAL_SEPARATOR = '.',
      _PERCENT = '%',
      _PER_MILL = '\u2030',
      _DIGIT = '#',
      _SEPARATOR = ';',
      _EXPONENT = "E",
      _MINUS = '-',
      _QUOT = '\'',
      _CURRENCY = '\u00A4';

  var posPrefixPattern,
      posSuffixPattern,
      negPrefixPattern,
      negSuffixPattern;

  var _MAXIMUM_INTEGER_DIGITS = 0x7fffffff;
  var _MAXIMUM_FRACTION_DIGITS = 0x7fffffff;

  _throwSyntaxError = function (pattern) {
    var msg, syntaxError, errorInfo, samplePattern = "#,##0.###";
    msg = "Unexpected character(s) encountered in the pattern \"" +
        pattern + " An example of a valid pattern is \"" + samplePattern +
        '".';
    syntaxError = new SyntaxError(msg);
    errorInfo = {
      'errorCode': 'optionValueInvalid',
      'parameterMap': {
        'propertyName': 'pattern',
        'propertyValue': pattern,
        'propertyValueHint': samplePattern
      }
    };
    syntaxError['errorInfo'] = errorInfo;
    throw syntaxError;
  };

  _regionMatches = function (str1, offset1, str2) {
    var sub1 = str1.substr(offset1, str2.length);
    var regExp = new RegExp(str2, "i");
    return (regExp.exec(sub1) !== null);
  };

  _expandAffixes = function (localeElements, numberSettings) {
    var curDisplay = {};
    if (posPrefixPattern !== null) {
      numberSettings['positivePrefix'] = _expandAffix(posPrefixPattern,
          localeElements, numberSettings, curDisplay);
    }
    if (posSuffixPattern !== null) {
      numberSettings['positiveSuffix'] = _expandAffix(posSuffixPattern,
          localeElements, numberSettings, curDisplay);
    }
    if (negPrefixPattern !== null) {
      numberSettings['negativePrefix'] = _expandAffix(negPrefixPattern,
          localeElements, numberSettings, curDisplay);
    }
    if (negSuffixPattern !== null) {
      numberSettings['negativeSuffix'] = _expandAffix(negSuffixPattern,
          localeElements, numberSettings, curDisplay);
    }
    if (curDisplay['name'] !== undefined) {
      numberSettings['positiveSuffix'] = "\u00a0" + curDisplay['name'];
      numberSettings['positivePrefix'] = "";
      if (numberSettings['lang'] === 'ar') {
        numberSettings['negativeSuffix'] = localeElements['numbers'][numberSettings['numberingSystem']]['minusSign'] + "\u00a0" + curDisplay['name'];
        numberSettings['negativePrefix'] = "";
      }
      else {
        numberSettings['negativeSuffix'] = "\u00a0" + curDisplay['name'];
        numberSettings['negativePrefix'] = localeElements['numbers'][numberSettings['numberingSystem']]['minusSign'];
      }
    }
  };

  _expandAffix = function (pattern, localeElements, numberSettings,
      currencyDisplay) {
    var buffer = "";
    for (var i = 0; i < pattern.length; ) {
      var c = pattern.charAt(i++);
      if (c === _QUOT)// {
        continue;
      //c = pattern.charAt(i++);
      switch (c) {
        case _CURRENCY:
          var code = numberSettings['currencyCode'];
          var name = code, symbol = code;
          if (localeElements['numbers']['currencies'][code] !== undefined) {
            name = localeElements['numbers']['currencies'][code]['displayName'];
            symbol = localeElements['numbers']['currencies'][code]['symbol'];
          }
          if (numberSettings['currencyDisplay'] === undefined ||
              numberSettings['currencyDisplay'] === "symbol")
            c = symbol;
          else if (numberSettings['currencyDisplay'] === "code") {
            c = code;
          }
          else {
            c = name;
            currencyDisplay['name'] = c;
          }
          break;
        case _PERCENT:
          c = localeElements['numbers'][numberSettings['numberingSystem']]['percentSign'];
          break;
        case _PER_MILL:
          c = localeElements['numbers'][numberSettings['numberingSystem']]['perMille'];
          break;
        case _MINUS:
          c = localeElements['numbers'][numberSettings['numberingSystem']]['minusSign'];
          break;
      }
      //}
      buffer = buffer.concat(c);
    }
    return buffer;
  };

  _applyPatternImpl = function (options, pattern, localeElements,
      numberSettings) {

    var gotNegative = false,
        useExponentialNotation = false;
    var phaseOneLength = 0;
    var start = 0;
    var isPrefix = true;

    for (var j = 1; j >= 0 && start < pattern.length; --j) {
      var inQuote = false;
      var prefix = "";
      var suffix = "";
      var decimalPos = -1;
      var multiplier = 1;
      var digitLeftCount = 0, zeroDigitCount = 0, digitRightCount = 0,
          groupingCount = -1, groupingCount0 = -1;
      var minExponentDigits;
      var phase = 0;

      isPrefix = true;
      for (var pos = start; pos < pattern.length; ++pos) {
        var ch = pattern.charAt(pos);
        switch (phase) {
          case 0:
          case 2:
            // Process the prefix / suffix characters
            if (inQuote) {
              if (ch === _QUOT) {
                if ((pos + 1) < pattern.length && pattern.charAt(pos + 1) ===
                    _QUOT) {
                  ++pos;
                  if (isPrefix)
                    prefix = prefix.concat("''");
                  else
                    suffix = suffix.concat("''");
                }
                else {
                  inQuote = false; // 'do'
                }
                continue;
              }
            }
            else {
              // Process unquoted characters seen in prefix or suffix phase.
              if (ch === _DIGIT ||
                  ch === _ZERO_DIGIT ||
                  ch === _GROUPING_SEPARATOR ||
                  ch === _DECIMAL_SEPARATOR) {
                phase = 1;
                --pos; // Reprocess this character
                continue;
              }
              else if (ch === _CURRENCY) {
                if (options['currency'] === undefined)
                  _throwMissingCurrency("style");
                // Use lookahead to determine if the currency sign
                // is doubled or not.
                options['style'] = 'currency';
                var doubled = (pos + 1) < pattern.length &&
                    pattern.charAt(pos + 1) === _CURRENCY;
                if (doubled) { // Skip over the doubled character
                  ++pos;
                }
                if (isPrefix)
                  prefix = prefix.concat(doubled ? "'\u00A4\u00A4" :
                      "'\u00A4");
                else
                  suffix = suffix.concat(doubled ? "'\u00A4\u00A4" :
                      "'\u00A4");
                continue;
              }
              else if (ch === _QUOT) {
                if (ch === _QUOT) {
                  if ((pos + 1) < pattern.length &&
                      pattern.charAt(pos + 1) === _QUOT) {
                    ++pos;
                    if (isPrefix)
                      prefix = prefix.concat("''");// o''clock
                    else
                      suffix = suffix.concat("''");
                  }
                  else {
                    inQuote = true; // 'do'
                  }
                  continue;
                }
              }
              else if (ch === _SEPARATOR) {
                if (phase === 0 || j === 0) {
                  _throwSyntaxError(pattern);
                }
                start = pos + 1;
                pos = pattern.length;
                continue;
              }

              // Next handle characters which are appended directly.
              else if (ch === _PERCENT) {
                options['style'] = 'percent';
                if (multiplier !== 1) {
                  _throwSyntaxError(pattern);
                }
                numberSettings['isPercent'] = true;
                multiplier = 100;
                if (isPrefix)
                  prefix = prefix.concat("'%");
                else
                  suffix = suffix.concat("'%");
                continue;
              }
              else if (ch === _PER_MILL) {
                if (multiplier !== 1) {
                  _throwSyntaxError(pattern);
                }
                options['style'] = 'perMill';
                numberSettings['isPerMill'] = true;
                multiplier = 1000;
                if (isPrefix)
                  prefix = prefix.concat("'\u2030");
                else
                  suffix = suffix.concat("'\u2030");
                continue;
              }
              else if (ch === _MINUS) {
                if (isPrefix)
                  prefix = prefix.concat("'-");
                else
                  suffix = suffix.concat("'-");
                continue;
              }
            }
            if (isPrefix)
              prefix = prefix.concat(ch);
            else
              suffix = suffix.concat(ch);
            break;

          case 1:
            if (j === 1) {
              ++phaseOneLength;
            }
            else {
              if (--phaseOneLength === 0) {
                phase = 2;
                isPrefix = false;
              }
              continue;
            }

            if (ch === _DIGIT) {
              if (zeroDigitCount > 0) {
                ++digitRightCount;
              }
              else {
                ++digitLeftCount;
              }
              if (groupingCount >= 0 && decimalPos < 0) {
                ++groupingCount;
              }
            }
            else if (ch === _ZERO_DIGIT) {
              if (digitRightCount > 0) {
                _throwSyntaxError(pattern);
              }
              ++zeroDigitCount;
              if (groupingCount >= 0 && decimalPos < 0) {
                ++groupingCount;
              }
            }
            else if (ch === _GROUPING_SEPARATOR) {
              groupingCount0 = groupingCount;
              groupingCount = 0;
            }
            else if (ch === _DECIMAL_SEPARATOR) {
              if (decimalPos >= 0) {
                _throwSyntaxError(pattern);
              }
              decimalPos = digitLeftCount + zeroDigitCount +
                  digitRightCount;
            }
            else if (_regionMatches(pattern, pos, _EXPONENT)) {
              if (useExponentialNotation) {
                _throwSyntaxError(pattern);
              }
              useExponentialNotation = true;
              minExponentDigits = 0;
              pos = pos + _EXPONENT.length;
              while (pos < pattern.length && pattern.charAt(pos) ===
                  _ZERO_DIGIT) {
                ++minExponentDigits;
                ++phaseOneLength;
                ++pos;
              }

              if ((digitLeftCount + zeroDigitCount) < 1 ||
                  minExponentDigits < 1) {
                _throwSyntaxError(pattern);
              }
              phase = 2;
              isPrefix = false;
              --pos;
              continue;
            }
            else {
              phase = 2;
              isPrefix = false;
              --pos;
              --phaseOneLength;
              continue;
            }
            break;
        }
      }


      if (zeroDigitCount === 0 && digitLeftCount > 0 && decimalPos >= 0) {
        // Handle "###.###" and "###." and ".###"
        var n = decimalPos;
        if (n === 0) { // Handle ".###"
          ++n;
        }
        digitRightCount = digitLeftCount - n;
        digitLeftCount = n - 1;
        zeroDigitCount = 1;
      }

      // Do syntax checking on the digits.
      if ((decimalPos < 0 && digitRightCount > 0) ||
          (decimalPos >= 0 && (decimalPos < digitLeftCount ||
              decimalPos > (digitLeftCount + zeroDigitCount))) ||
          groupingCount === 0 || inQuote) {
        _throwSyntaxError(pattern);
      }

      if (j === 1) {
        posPrefixPattern = prefix;
        posSuffixPattern = suffix;
        negPrefixPattern = posPrefixPattern;
        negSuffixPattern = posSuffixPattern;
        var digitTotalCount = digitLeftCount + zeroDigitCount +
            digitRightCount;
        // The effectiveDecimalPos is the position the decimal is at or
        //would be at if there is no decimal. Note that if decimalPos<0,
        // then digitTotalCount == digitLeftCount + zeroDigitCount.
        var effectiveDecimalPos = decimalPos >= 0 ?
            decimalPos : digitTotalCount;
        numberSettings['minimumIntegerDigits'] = (effectiveDecimalPos -
            digitLeftCount);
        numberSettings['maximumIntegerDigits'] = (useExponentialNotation ?
            digitLeftCount + numberSettings['minimumIntegerDigits'] :
            _MAXIMUM_INTEGER_DIGITS);
        numberSettings['maximumFractionDigits'] = (decimalPos >= 0 ?
            (digitTotalCount - decimalPos) : 0);
        numberSettings['minimumFractionDigits'] = (decimalPos >= 0 ?
            (digitLeftCount + zeroDigitCount - decimalPos) : 0);
        numberSettings['groupingSize'] = (groupingCount > 0) ?
            groupingCount : 0;
        numberSettings['groupingSize0'] = groupingCount0;
      }
      else {
        negPrefixPattern = prefix;
        negSuffixPattern = suffix;
        gotNegative = true;
      }
    }

    if (pattern.length === 0) {
      posPrefixPattern = posSuffixPattern = "";
      numberSettings['minimumIntegerDigits'] = 0;
      numberSettings['maximumIntegerDigits'] = _MAXIMUM_INTEGER_DIGITS;
      numberSettings['minimumFractionDigits'] = 0;
      numberSettings['maximumFractionDigits'] = _MAXIMUM_FRACTION_DIGITS;
    }
    numberSettings['useExponentialNotation'] = useExponentialNotation;
    numberSettings['minExponentDigits'] = minExponentDigits;
    // If there was no negative pattern, or if the negative pattern is
    // identical to the positive pattern, then prepend the minus sign to
    // the positive pattern to form the negative pattern.
    if (!gotNegative ||
        ((negPrefixPattern.localeCompare(posPrefixPattern) === 0)
            && (negSuffixPattern.localeCompare(posSuffixPattern) === 0))) {
      if (options['style'] === 'currency' && numberSettings['lang'] === 'ar') {
        negSuffixPattern = posSuffixPattern + "'\u200f-";
        negPrefixPattern = posPrefixPattern;
      }
      else {
        negSuffixPattern = posSuffixPattern;
        negPrefixPattern = "'-" + posPrefixPattern;

      }
    }
    _expandAffixes(localeElements, numberSettings);
  };

  _getRoundedNumber = function (ret, numberSettings, options) {
    var precision = numberSettings['maximumFractionDigits'];
    var isNegative = ret < 0;
    var mode = options['roundingMode'] || 'DEFAULT';
    var roundedNumber = _roundNumber(Math.abs(ret), precision, mode);
    return isNegative ? -roundedNumber : roundedNumber;
  };

  _resolveNumberSettings = function (localeElements, options, locale) {
    var numberSettings = {};
    _validateNumberOptions(options, "OraNumberConverter.resolvedOptions");
    _getNumberSettings(localeElements, numberSettings, options, locale);
    numberSettings['numberingSystemKey'] = _getNumberingExtension(locale);
    if (oj.OraI18nUtils.numeringSystems[numberSettings['numberingSystemKey']] ===
        undefined)
      numberSettings['numberingSystemKey'] = 'latn';
    return numberSettings;
  };

  _resolveOptions = function (numberSettings, options, locale) {
    var resOptions = {
      'locale': locale,
      'style': (options['style'] === undefined) ? 'decimal' : options['style'],
      'useGrouping': (options['useGrouping'] === undefined) ? true : options['useGrouping'],
      'numberingSystem': numberSettings['numberingSystemKey']
    };
    resOptions['minimumIntegerDigits'] = numberSettings['minimumIntegerDigits'];
    resOptions['minimumFractionDigits'] = numberSettings['minimumFractionDigits'];
    resOptions['maximumFractionDigits'] = numberSettings['maximumFractionDigits'];
    if (options['style'] === 'decimal' && options['decimalFormat'] !== undefined) {
      resOptions['decimalFormat'] = options['decimalFormat'];
    }
    if (options['style'] === 'currency' && options['currencyFormat'] !== undefined) {
      resOptions['currencyFormat'] = options['currencyFormat'];
    }
    if (options['style'] === 'currency') {
      resOptions['currency'] = options['currency'];
      resOptions['currencyDisplay'] = (options['currencyDisplay'] ===
          undefined) ? 'symbol' : options['currencyDisplay'];
    }
    if (options['unit'] !== undefined)
      resOptions['unit'] = options['unit'];
    if (options['pattern'] !== undefined)
      resOptions['pattern'] = options['pattern'];
    var roundingMode = options['roundingMode'];
    var roundDuringParse = options['roundDuringParse'];
    if (roundingMode !== undefined)
      resOptions['roundingMode'] = roundingMode;
    if (roundDuringParse !== undefined)
      resOptions['roundDuringParse'] = roundDuringParse;
    var leneint = numberSettings['lenientParse'];
    if (leneint !== undefined)
      resOptions['lenientParse'] = leneint;
    var sep = numberSettings['separators'];
    if (sep !== undefined)
      resOptions['separators'] = sep;
    return resOptions;
  };

  function _init()
  {

    return {
      /**
       * Format a number.
       * @memberOf OraNumberConverter
       * @param {number} value - Number object to be formatted.
       * @param {Object} localeElements - the instance of LocaleElements  
       * bundle
       * @param {Object=} options - Containing the following properties:<br>
       * - <b>style.</b>  is one of the String values "decimal", "currency",  
       * "percent" or "unit". The default is "decimal".<br>
       * - <b>decimalFormat.</b> is used in conjuction with "decimal" style. 
       * It can have one of the string values "short", "long", "standard". "standard"
       * is the default. It is used for compact number formatting. For example 3000 is displayed
       *  as 3K for "short" and 3 thousand for "long". We take into consideration
       *  the locale's plural rules for the compact pattern.<br> 
       * - <b>currency.</b> An ISO 4217 alphabetic currency code. Mandatory 
       *  when style is "currency".<br>
       * - <b>unit.</b> Mandatory when style is "unit". Allowed values are "byte" or "bit".<br> 
       * - <b>currencyFormat.</b> is used in conjuction with "currency" style. 
       * It can have one of the string values "short", "long", "standard". "standard"
       * is the default. It is used for compact currency formatting. For example $3000 is displayed
       *  as $3K for "short" and 3 thousand US Dollar for "long".<br> 
       * - <b>currencyDisplay.</b> is one of the String values "code", 
       * "symbol", or "name", specifying whether to display the currency as  
       * an ISO 4217 alphabetic currency code, 
       * a localized currency symbol, or a localized currency name if 
       * formatting with the "currency" style. It is only present when style 
       * has the value "currency". The default is "symbol".<br>
       * - <b>minimumIntegerDigits.</b> is a non-negative integer Number value 
       * indicating the minimum integer digits to be used. Numbers will be 
       * padded with leading zeroes if necessary.<br>
       * - <b>minimumFractionDigits.</b> a non-negative integer Number value 
       * indicating the minimum fraction digits to be used. Numbers will be 
       * padded with trailing zeroes if necessary.<br>
       * - <b>maximumFractionDigits.</b> a non-negative integer Number value 
       * indicating the maximum fraction digits to be used. Numbers will be 
       * rounded if necessary.<br>
       * - <b>roundingMode.</b> specifies the rounding behavior. This follows the
       *  Java.Math.RoundingMode behavior. Currently we support the options : 
       *  HALF_UP, HALF_DOWN, and HALF_EVEN<br>
       * - <b>useGrouping.</b> is a Boolean value indicating whether a 
       * grouping separator should be used. The default is true.<br>
       * - <b>separators.</b> - An object with 2 fields: 'decimal' and 'group'.
       * It allows the user to override the locale's default decimal and grouping separators.<br>
       * - <b>pattern.</b> custom pattern. Will override above options 
       * when present.
       * @param {string=} locale - A BCP47 compliant language tag. it is only 
       * used to extract the unicode extension keys. 
       * @return {string} formatted number.
       * @throws {RangeError} If a property value of the options parameter is 
       * out of range.
       * @throws {TypeError} If the style is currency and currency code is 
       * missing.
       * @throws {SyntaxError} If an unexpected character is encountered in 
       * the pattern.
       */
      format: function (value, localeElements, options, locale) {
        if (arguments.length <= 2 || options === undefined) {
          options = {
            'useGrouping': true,
            'style': 'decimal'
          };
        }
        _validateNumberOptions(options, "OraNumberConverter.format");
        var numberSettings = {};
        _getNumberSettings(localeElements, numberSettings, options, locale);
        return _formatNumberImpl(value, options, localeElements,
            numberSettings, locale);
      },
      /**
       * Parse a number.
       * @memberOf OraNumberConverter
       * @param {string|number} str - string to be parsed.
       * @param {Object} localeElements - the instance of LocaleElements 
       * bundle
       * @param {Object=} options - Containing the following properties:<br>
       * - <b>style.</b>  is one of the String values "decimal", "currency" or 
       * "percent". The default is "decimal".<br>
       * - <b>currency.</b> An ISO 4217 alphabetic currency code. Mandatory 
       * when style is "currency".<br>
       * - <b>currencyDisplay.</b> is one of the String values "code", 
       * "symbol", or "name", specifying whether to display the currency as 
       * an ISO 4217 alphabetic currency code,
       *  a localized currency symbol, or a localized currency name if 
       *  formatting with the "currency" style. It is only considered when 
       *  style has the value "currency". The default is "symbol".<br>
       * - <b>pattern.</b> custom pattern. Will override above options when 
       * present.<br>
       * - <b>roundingMode.</b> specifies the rounding behavior. This follows the
       *  Java.Math.RoundingMode behavior. Currently we support the options : 
       *  HALF_UP, HALF_DOWN, and HALF_EVEN<br>
       *  - <b>roundDuringParse.</b> Boolean value. Specifies whether or not to round during parse.
       *  by default the number converter rounds during format but not during parse.<br>
       *  - <b>separators.</b> - An object with 2 fields: 'decimal' and 'group'.
       * It allows the user to override the locale's default decimal and grouping separators.<br>
       *  - <b>lenientParse.</b> specifies if lenient parse is enabled or disabled. Allowed values: "full", "none".
       *  default is "full" which means lenient parse is enabled.<br>
       * @param {string=} locale - A BCP47 compliant language tag. it is only 
       * used to extract the unicode extension keys. 
       * @return {number} a number object parsed from the string. In case of 
       * error, returns null.
       * @throws {RangeError} If a property value of the options parameter is 
       * out of range.
       * @throws {TypeError} If the style is currency and currency code is 
       * missing.
       * @throws {SyntaxError} If an unexpected character is encountered in 
       * the pattern.
       * @throws {Error} If the <i>str</i> parameter does not match the number 
       * pattern.
       */
      parse: function (str, localeElements, options, locale) {
        if (typeof str === "number")
          return str;
        if (Object.prototype.toString.call(str) === '[object Number]')
          return  Number(str);
        if (arguments.length <= 2 || options === undefined) {
          options = {
            'useGrouping': true,
            'style': 'decimal'
          };
        }
        _validateNumberOptions(options, "OraNumberConverter.parse");
        return _parseNumberImpl(str, localeElements, options, locale);
      },
      /**
       * Resolve options.
       * Returns a new object with properties reflecting the number formatting 
       * options computed based on the options parameter.
       * If options is not provided, the properties will be derived from the 
       * locale defaults.
       * @memberOf OraNumberConverter
       * @param {Object} localeElements - the instance of LocaleElements 
       * bundle
       * @param {Object=} options containing the following properties:<br>
       * - <b>style.</b> "decimal", "currency", "percent" or "unit". The default is 
       * "decimal".<br>
       * - <b>unit.</b> one of the strings "byte" or "bit" when the style is "unit".<br>
       * - <b>decimalFormat.</b> It can have one of the string values "short", "long", "standard".
       * "standard" is the default. It is used for compact number formatting. For example 3000 is displayed
       *  as 3K for "short" and 3 thousand for "long". We take into consideration
       *  the locale's plural rules for the compact pattern.<br> 
       * - <b>currency.</b> An ISO 4217 alphabetic currency code. Mandatory 
       * when when style is "currency".<br>
       * - <b>currencyFormat.</b> is used in conjuction with "currency" style. 
       * It can have one of the string values "short", "long", "standard". "standard"
       * is the default. It is used for compact currency formatting. For example $3000 is displayed
       *  as $3K for "short" and 3 thousand US Dollar for "long".<br> 
       * - <b>currencyDisplay.</b> is one of the String values "code", 
       * "symbol", or "name", specifying whether to display the currency as 
       * an ISO 4217 alphabetic currency code,
       *   a localized currency symbol, or a localized currency name if 
       *   formatting with the "currency" style. It is only present 
       *   when style has the value "currency". The default is "symbol".<br>
       * - <b>minimumIntegerDigits.</b> is a non-negative integer Number value 
       * indicating the minimum integer digits to be used. Numbers will be 
       * padded with leading zeroes if necessary.<br>
       * - <b>minimumFractionDigits.</b> a non-negative integer Number value 
       * indicating the minimum fraction digits to be used. Numbers will be 
       * padded with trailing zeroes if necessary.<br>
       * - <b>maximumFractionDigits.</b> a non-negative integer Number value 
       * indicating the maximum fraction digits to be used. Numbers will be 
       * rounded if necessary.<br>
       * - <b>numberingSystem</b>. The numbering system.<br>
       * - <b>useGrouping.</b> is a Boolean value indicating whether a 
       * grouping separator should be used. The default is true.<br>
       * - <b>pattern.</b> custom pattern. Will override above options when 
       * present.<br>
       * - <b>roundingMode.</b> specifies the rounding behavior. This follows the
       *  Java.Math.RoundingMode behavior. Currently we support the options : 
       *  HALF_UP, HALF_DOWN, and HALF_EVEN<br>
       *  - <b>roundDuringParse.</b> Boolean value. Specifies whether or not to round during parse.
       *  by default the number converter rounds during format but not during parse.<br>
       *  - <b>separators.</b> - An object with 2 fields: 'decimal' and 'group'.
       * It allows the user to override the locale's default decimal and grouping separators.<br>
       *  - <b>lenientParse.</b> specifies if lenient parse is enabled or disabled. Allowed values: "full", "none".
       * default is "full" which means lenient parse is enabled.<br>
       * @param {string=} locale - A BCP47 compliant language tag. it is only 
       * used to extract the unicode extension keys. 
       * @return {Object} Resolved options object.
       * @throws {RangeError} If a property value of the options parameter is 
       * out of range.
       * @throws {TypeError} If the style is currency and currency code is 
       * missing. 
       */
      resolvedOptions: function (localeElements, options, locale) {
        if (arguments.length < 3 || locale === undefined) {
          locale = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
        }
        if (arguments.length < 2 || options === undefined) {
          options = {
            'useGrouping': true,
            'style': 'decimal'
          };
        }
        var numberSettings = _resolveNumberSettings(localeElements, options, locale);
        return _resolveOptions(numberSettings, options, locale);
      }
    };
  }

  return {
    /**
     * getInstance.
     * Returns the singleton instance of OraNumberConverter class.  
     * @memberOf OraNumberConverter
     * @return {Object} The singleton OraNumberConverter instance.
     */
    getInstance: function () {
      if (!instance) {
        instance = _init();
      }
      return instance;
    }
  };
}());
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * oj.DateTimeConverter Contract. 
 */

/**
 * @class
 * @param {Object=} options an object literal used to provide an optional information to 
 * @augments oj.Converter 
 * @ojsignature {target: "Type", 
 *                value: "abstract class DateTimeConverter extends Converter<string>"}
 *       
 * @name oj.DateTimeConverter
 * @abstract
 * @export
 * @since 0.6
 */
oj.DateTimeConverter = function(options)
{
  this.Init(options);
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.DateTimeConverter, oj.Converter, "oj.DateTimeConverter");

/**
 * Initializes the date time converter instance with the set options.
 * 
 * @param {Object=} options an object literal used to provide an optional information to 
 * initialize the converter.<p>
 * @export
 * @ignore
 */
oj.DateTimeConverter.prototype.Init = function(options) 
{
  oj.DateTimeConverter.superclass.Init.call(this, options);
};

/**
 * Formats the local isoString value using the options provided and returns a string value. Note that if previous application 
 * code was passing a JavaScript Date object which is no longer supported, one can use the utility function oj.IntlConverterUtils.dateToLocalIso 
 * to get the proper isoString value.
 * 
 * @example <caption>For example <code class="prettyprint">converter.format(oj.IntlConverterUtils.dateToLocalIso(new Date()))</code></caption>
 * @see oj.IntlConverterUtils.dateToLocalIso
 * @param {string} value to be formatted for display which should be a local isoString
 * @return {(string|null)} the localized and formatted value suitable for display
 * @throws {Error} a ConverterError if formatting fails.
 * @export
 */
oj.DateTimeConverter.prototype.format = function (value) 
{
  return oj.DateTimeConverter.superclass.format.call(this, value);
};


/**
 * Returns true if a 24-hour format is set; false otherwise.
 * @export
 * @abstract
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 */
oj.DateTimeConverter.prototype.isHourInDaySet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if 12-hour is set; false otherwise.
 * @export
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 * @abstract
 */
oj.DateTimeConverter.prototype.isHourInAMPMSet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if minutes are shown in the time portion; false otherwise.
 * @export
 * @abstract
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 */
oj.DateTimeConverter.prototype.isMinuteSet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if seconds are shown in the time portion; false otherwise.
 * @export
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 * @abstract
 */
oj.DateTimeConverter.prototype.isSecondSet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if milliseconds are shown in the time portion; false otherwise.
 * @export
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 * @abstract
 */
oj.DateTimeConverter.prototype.isMilliSecondSet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if year is shown in the date portion; false otherwise.
 * @export
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 * @abstract
 */
oj.DateTimeConverter.prototype.isYearSet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if month is shown in the date portion; false otherwise.
 * @export
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 * @abstract
 */
oj.DateTimeConverter.prototype.isMonthSet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if day is shown in the date portion; false otherwise.
 * @export
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 * @abstract
 */
oj.DateTimeConverter.prototype.isDaySet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns true if the day name is shown in the date portion; false otherwise.
 * @export
 * @ojsignature {target: "Type", 
 *                value: "():boolean"}
 * @abstract
 */
oj.DateTimeConverter.prototype.isDayNameSet = function()
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Returns the calculated week for the isoString value.
 * @param {string} value to return the calculated week of
 * @return {number|undefined} calculated week.
 * @export
 * @abstract
 */
oj.DateTimeConverter.prototype.calculateWeek = function(value)
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Parses the value using the options provided and returns a local isoString value. For convenience if one wishes to 
 * retrieve a JavaScript Date object from the local isoString an utility function oj.IntlConverterUtils.isoToLocalDate is 
 * provided.
 * 
 * @example <caption>For example <code class="prettyprint">oj.IntlConverterUtils.isoToLocalDate(converter.parse(isoString))</code></caption>
 * @see oj.IntlConverterUtils.isoToLocalDate
 * @param {string} value to parse
 * @return {string} the parsed value as a local isoString value
 * @throws {Error} a ConverterError if parsing fails
 * @export
 */
oj.DateTimeConverter.prototype.parse = function (value) 
{
  return oj.DateTimeConverter.superclass.parse.call(this, value);
};

/**
 * Compares 2 ISO 8601 strings, returning the time difference between the two
 * 
 * @param {string} isoStr first iso string
 * @param {string} isoStr2 second iso string
 * @return {number} the time difference between isoStr and isoStr2
 * @export
 */
oj.DateTimeConverter.prototype.compareISODates = function (isoStr, isoStr2)
{
  return oj.DateTimeConverter.superclass.compareISODates.call(this, isoStr, isoStr2);
};

/**
 * Gets the supported timezones for the converter.<br/>
 *
 * @return {Array} supported timezones
 * @since 4.0.0
 * @export
 */
oj.DateTimeConverter.prototype.getAvailableTimeZones = function ()
{
  return oj.DateTimeConverter.superclass.getAvailableTimeZones.call(this);
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * Constructs a DateRestrictionValidator that ensures the value provided is not in a disabled entry of dayMetaData
 * @param {Object=} options an object literal used to provide the following properties
 * @export
 * @constructor
 * @since 0.6
 * @augments oj.Validator
 * @ojsignature [{target: "Type", 
 *                value: "class DateRestrictionValidator extends Validator<string>"},
 *               {target: "Type",
 *                value: "oj.DateRestrictionValidator.ValidatorOptions",
 *                for: "options", jsdocOverride: true}
 *              ]
 */
oj.DateRestrictionValidator = function _DateRestrictionValidator(options)
{
  this.Init(options);
};

/**
 * Input type for the dayFormatter option call back function
 * @typedef {object} oj.DateRestrictionValidator.DayFormatterInput
 * @property {number} fullYear
 * @property {number} month
 * @property {number} date
 */
/**
 * Output type for the dayFormatter option call back function
 * @typedef {object} oj.DateRestrictionValidator.DayFormatterOutput
 * @property {boolean=} disabled
 * @property {string=} className
 * @property {string=} tooltip
 */
/**
 * @typedef {object} oj.DateRestrictionValidator.ValidatorOptions
 * @property {(function(oj.DateRestrictionValidator.DayFormatterInput): (oj.DateRestrictionValidator.DayFormatterOutput|null|'all'))=} dayFormatter - Additional info to be used when rendering the day. This 
 * should be a JavaScript Function callback which accepts as its argument the following JSON format 
 * <code class="prettyprint">{fullYear: Date.getFullYear(), month: Date.getMonth()+1, date: Date.getDate()}</code>
 * and returns <code class="prettyprint">null</code> or all or partial JSON data of the form 
 * <code class="prettyprint">{disabled: true|false, className: "additionalCSS", tooltip: 'Stuff to display'}</code>
 * @property {string=} messageSummary - an optional custom error message summarizing the 
 * error. When not present, the default message summary is the resource defined with the key 
 * <code class="prettyprint">oj-validator.restriction.date.messageSummary</code>.
 * Tokens: {value} - value entered by user<p>. 
 * Example:<br/>
 * "Value {value} is disabled."<br/>
 * <p>
 * @property {string=} messageDetail - a custom error message used for creating detail part 
 * of message. When not present, the default message detail is the 
 * resource defined with the key <code class="prettyprint">oj-validator.restriction.date.messageDetail</code>.
 * Tokens: {value} - value entered by user<p>. 
 * Example:<br/>
 * "Value {value} is a disabled entry. Please select a different date."<br/>
 * </p>
 */

// Subclass from oj.Validator 
oj.Object.createSubclass(oj.DateRestrictionValidator, oj.Validator, "oj.DateRestrictionValidator");

/**
 * Initializes validator instance with the set options
 * @param {Object=} options
 * @memberof oj.DateRestrictionValidator
 * @instance
 * @ignore
 * @export
 */
oj.DateRestrictionValidator.prototype.Init = function (options)
{
  oj.DateRestrictionValidator.superclass.Init.call(this);
  this._dayFormatter = options["dayFormatter"];
  this._converter = oj.IntlConverterUtils.getConverterInstance(options["converter"]);
  if (options)
  {
    this._messageSummary = options['messageSummary'] || null;
    this._messageDetail = options['messageDetail'] || null;
  }
};

/**
 * Validates whether the date provided is part of disabled date
 *
 * @private
 * @ignore
 * @param {Object|string} valueDateParam that is being validated. Note it is set to string as well to keep gc happy
 * @returns {boolean} boolean of whether it is a disabled date
 */
oj.DateRestrictionValidator.prototype._inDisabled = function(valueDateParam) 
{
  var dayFormatter = this._dayFormatter;
  
  if(dayFormatter) {
    var fullYear = valueDateParam["fullYear"],
        month = valueDateParam["month"] + 1, //request to start from 1 rather than 0
        date = valueDateParam["date"],
        metaData = dayFormatter({"fullYear": fullYear, "month": month, "date": date});
    
    return metaData && metaData.disabled;
  }
  
  return false;
};

/**
 * Validates whether the date provided is part of disabled date
 *
 * @param {string} value that is being validated
 * @returns {string} original if validation was successful
 *
 * @throws {Error} when there is no match
 * @ojsignature {target: "Type", for: "returns",
 *                value: "void"}
 * @memberof oj.DateRestrictionValidator
 * @instance
 * @export
 */
oj.DateRestrictionValidator.prototype.validate = function (value)
{
  var summary = "", 
      detail = "", 
      translations = oj.Translations, 
      messageSummary = this._messageSummary,
      messageDetail = this._messageDetail,
      valueStr = value ? this._converter['format'](value) : value,
      valueDateParam = value ? oj.IntlConverterUtils._dateTime(value, ["fullYear", "month", "date"], 
                        true) : null;
  
  if(value === null) 
  {
    return value;
  }
  
  if(this._inDisabled(valueDateParam)) {
    
    summary = messageSummary ? translations.applyParameters(messageSummary, {"value": valueStr}) : 
                translations.getTranslatedString('oj-validator.restriction.date.messageSummary', {"value": valueStr});
    detail = messageDetail ? translations.applyParameters(messageDetail, {"value": valueStr}) : 
                translations.getTranslatedString('oj-validator.restriction.date.messageDetail', {"value": valueStr});
    throw new oj.ValidatorError(summary, detail);
  }
  
  return value;
};

/**
 * A message to be used as hint. As there exists no hint for DateRestrictionValidator, default is to return null.
 *
 * @returns {string|null} a hint message or null if no hint is available in the options
 * @memberof oj.DateRestrictionValidator
 * @instance
 * @export
 */
oj.DateRestrictionValidator.prototype.getHint = function ()
{
  return null;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * Constructs a DateTimeRangeValidator that ensures the value provided is within a given range
 * @param {Object=} options an object literal used to provide the following properties
 * @export
 * @constructor
 * @augments oj.Validator
 * @ojsignature [{target: "Type", 
 *                value: "class DateTimeRangeValidator extends Validator<string>"},
 *               {target: "Type",
 *                value: "oj.DateTimeRangeValidator.ValidatorOptions",
 *                for: "options", jsdocOverride: true}
 *              ]
 * @since 0.6
*/
oj.DateTimeRangeValidator = function _DateTimeRangeValidator(options)
{
  this.Init(options);
};

/**
 * @typedef {Object} oj.DateTimeRangeValidator.ValidatorOptions
 * @property {oj.DateTimeConverter} converter - an instance implementation of oj.DateTimeConverter (i.e. oj.IntlDateTimeConverter). 
 * In order to compare isoString value with the timeZone options in the converter, it is necessary for the validator that 
 * a converter is passed in.
 * @property {string=} min - the minimum datetime value of the entered value. Should be ISOString.
 * @property {string=} max - the maximum datetime value of the entered value. Should be ISOString.
 * @property {Object=} hint - an optional object literal of hints to be used. 
 * @property {string=} hint.max - a hint used to indicate the allowed maximum. When not present, 
 * the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.datetime.hint.max</code>.<p>
 * Tokens: <br/>
 * {max} - the maximum<p>
 * Usage: <br/>
 * Enter a datetime less than or equal to {max}
 * @property {string=} hint.min - a hint used to indicate the allowed minimum. When not present, 
 * the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.datetime.hint.min</code>.<p>
 * Tokens: <br/>
 * {min} the minimum <p>
 * Usage: <br/>
 * Enter a datetime greater than or equal to {min}
 * @property {string=} hint.inRange - a hint used to indicate the allowed range. When not 
 * present, the default hint is the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.datetime.hint.inRange</code>.<p>
 * Tokens:<br/>
 * {min} the minimum<br/>
 * {max} the maximum<p>
 * Usage: <br/>
 * Enter a datetime between {min} and {max}
 * @property {string=} translationKey - an optional string for the default messages + hints. Note 
 * that this is required only when the validator is used standlone (i.e. if one uses min + max 
 * attribute for ojInputDate the validator will have this already be set to correct "date" value)
 * <ul>
 *  <li>"datetime"</li>
 *  <li>"date"</li>
 *  <li>"time"</li>
 * </ul>
 * <br/>
 * @property {Object=} messageDetail - an optional object literal of custom error messages to 
 * be used.
 * @property {string=} messageDetail.rangeUnderflow - the detail error message to be used when 
 * input value is less than the set minimum value. When not present, the default detail message is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.datetime.messagedetail.rangeUnderflow</code>.<p>
 * Tokens:<br/>
 * {value} - value entered by the user<br/>
 * {min} - the minimum allowed value<p>
 * Usage: <br/>
 * Entered {value} with min being {min}
 * @property {string=} messageDetail.rangeOverflow - the detail error message to be used when 
 * input value exceeds the maximum value set.  When not present, the default detail message is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.datetime.messagedetail.rangeOverflow</code>.<p>
 * Tokens:<br/>
 * {value} - value entered by the user<br/>
 * {max} - the maximum allowed value<p>
 * Usage: <br/>
 * Entered {value} with max being {max}
 * @property {Object=} messageSummary - optional object literal of custom error summary message 
 * to be used. 
 * @property {string=} messageSummary.rangeUnderflow - the summary of the error message when 
 * input value is less than the set minimum value. When not present, the default message summary is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.datetime.messageSummary.rangeUnderflow</code>.
 * @property {string=} messageSummary.rangeOverflow - the summary of the error message when 
 * input value exceeds the maximum value set.  When not present, the default message summary is 
 * the resource defined with the key 
 * <code class="prettyprint">oj-validator.range.datetime.messageSummary.rangeOverflow</code>.
 */

// Subclass from oj.Validator 
oj.Object.createSubclass(oj.DateTimeRangeValidator, oj.Validator, "oj.DateTimeRangeValidator");

/**
 * Initializes validator instance with the set options
 * @param {Object=} options
 * @memberof oj.DateTimeRangeValidator
 * @instance
 * @export
 * @ignore
 */
oj.DateTimeRangeValidator.prototype.Init = function (options)
{
  oj.DateTimeRangeValidator.superclass.Init.call(this);

  //if undefined set to null as they are equivalent in terms of logic
  //setting to null for the default validator [min + max option] is taken care of for ojInputDate 
  //in _InitOptions for min + max values [default validator]; however 
  //user can pass in the validators via validators option so taking care of it here
  this._converter = oj.IntlConverterUtils.getConverterInstance(options["converter"]);
  if(!this._converter) 
  {
    throw new Error("oj.DateTimeRangeValidator missing a converter option");
  }
  this._min = options["min"] || null;
  this._max = options["max"] || null;
  this._translationKey = options["translationKey"] || 'datetime';
  
  if (options)
  {
    this._hint = options['hint'] || {};
    this._customMessageSummary = options['messageSummary'] || {};
    this._customMessageDetail = options['messageDetail'] || {};
  }

  if(this._translationKey)
  {
    //random test to check that the specified translation identity exists; otherwise just use datetime
    if(!oj.Translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.messageSummary.rangeOverflow')) {
      this._translationKey = 'datetime';
    }
  }
  
};

/**
 * Validates the minimum + maximum conditions
 *
 * @param {string} value that is being validated
 * @returns {string|null} original if validation was successful
 * @ojsignature {target: "Type", for: "returns",
 *                value: "void"}
 * @throws {Error} when there is no match
 * @memberof oj.DateTimeRangeValidator
 * @instance
 * @export
 */
oj.DateTimeRangeValidator.prototype.validate = function (value)
{
  var customMessageSummary = this._customMessageSummary,
      customMessageDetail = this._customMessageDetail,
      messageDetailRangeOverflow = customMessageDetail["rangeOverflow"], 
      messageDetailRangeUnderflow = customMessageDetail["rangeUnderflow"], 
      messageSummaryRangeOverflow = customMessageSummary["rangeOverflow"],
      messageSummaryRangeUnderflow = customMessageSummary["rangeUnderflow"],
      converterUtils = oj.IntlConverterUtils,
      min = this._min, 
      max = this._max, 
      summary = "", 
      detail = "", 
      translations = oj.Translations, 
      params = null,
      valStr = value ? this._converter['format'](value) : value,
      minStr,
      maxStr;
  
  if(value === null) 
  {
    // request to not throw an error when value being passed is of null
    return value;
  }
  
  valStr = this._converter['format'](value);
  
  if(min) 
  {
    min = converterUtils._minMaxIsoString(min, value);
    minStr = this._converter ? this._converter['format'](min) : min;
  }
  
  if(max)
  {
    max = converterUtils._minMaxIsoString(max, value);
    maxStr = this._converter ? this._converter['format'](max) : max;
  }
  
  if (min !== null && max !== null)
  {
    //range
    if ((this._converter.compareISODates(value, min) >= 0 && this._converter.compareISODates(value, max) <= 0) || 
        this._converter.compareISODates(min, max) > 0)
    {
      return value;
    }
  }
  else 
  {
    //only min
    if (min !== null)
    {
      if (this._converter.compareISODates(value, min) >= 0)
      {
        return value;
      }
	  
    }
    //max only
    else 
    {
      if (max === null || this._converter.compareISODates(value, max) <= 0)
      {
        return value;
      }
      
    }
  }
  
  if (max !== null && this._converter.compareISODates(value, max) > 0)
  {
      params = {"value": valStr, "max": maxStr};
      summary = messageSummaryRangeOverflow ? messageSummaryRangeOverflow : 
        translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.messageSummary.rangeOverflow');
      detail = messageDetailRangeOverflow ? 
        translations.applyParameters(messageDetailRangeOverflow, params) : 
        translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.messageDetail.rangeOverflow', 
        params);
  }
  else if (min !== null && this._converter.compareISODates(value, min) < 0)
  {
      params = {"value": valStr, "min": minStr};
      summary = messageSummaryRangeUnderflow ? messageSummaryRangeUnderflow : 
        translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.messageSummary.rangeUnderflow');
      detail = messageDetailRangeUnderflow ?
        translations.applyParameters(messageDetailRangeUnderflow, params) : 
        translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.messageDetail.rangeUnderflow', 
        params);   
  }

  throw new oj.ValidatorError(summary, detail);
};

/**
 * A message to be used as hint.
 *
 * @returns {string|null} a hint message or null if no hint is available in the options
 * @memberof oj.DateTimeRangeValidator
 * @instance
 * @export
 */
oj.DateTimeRangeValidator.prototype.getHint = function ()
{
  var hint = null, hints = this._hint, 
      hintInRange = hints["inRange"], hintMinimum = hints["min"], 
      hintMaximum = hints["max"],
      min = this._min, 
      max = this._max, 
      minStr = min && this._converter ? this._converter['format'](min) : min,
      maxStr = max && this._converter ? this._converter['format'](max) : max,
      params = null,
      translations = oj.Translations;
  
  if (min !== null && max !== null) 
  {
    params = {"min": minStr, "max": maxStr};
    hint = hintInRange ? translations.applyParameters(hintInRange, params) : 
            translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.hint.inRange', params);
  }
  else if (min !== null)
  {
    params = {"min": minStr};
    hint = hintMinimum ?  translations.applyParameters(hintMinimum, params) :
            translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.hint.min', params);
  }
  else if (max !== null)
  {
    params = {"max": maxStr};
    hint = hintMaximum ?  translations.applyParameters(hintMaximum, params) :
            translations.getTranslatedString('oj-validator.range.' + this._translationKey + '.hint.max', params);
  }

  return hint;
};
/**
 * Copyright (c) 2008, 2013, Oracle and/or its affiliates. 
 * All rights reserved.
 */



/**
 * A factory implementation to create the built-in datetime converter of type 
 * {@link oj.IntlDateTimeConverter}. 
 * 
 * @name oj.DateTimeConverterFactory
 * @hideconstructor
 * @public
 * @class
 * @example <caption>create an instance of the jet datetime converter using the options provided</caption>
 * var dtcf = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME);  
 * var dateOptions = {year: '2-digit', month: 'numeric', day: 'numeric'};
 * var dateConverter = dtcf.createConverter(dateOptions);
 * @since 0.6
 * 
 */
oj.DateTimeConverterFactory = (function ()
{
  function _createDateTimeConverter(options)
  {
    return new oj.IntlDateTimeConverter(options);
  }

  /**
   * 
   * @public
   */
  return {
    /**
     * Creates an immutable (jet) datetime converter instance. 
     * 
     * @param {Object=} options an object literal used to provide an optional information to 
     * initialize the jet datetime converter. For details on what to pass for options, refer to 
     * @link oj.IntlDateTimeConverter.
     * 
     * @return {oj.IntlDateTimeConverter} 
     * @ojsignature {target: "Type", for: "options", value: "oj.IntlDateTimeConverter.ConverterOptions"}
     * @instance
     * @memberOf oj.DateTimeConverterFactory
     * @public
     */
    'createConverter': function (options)
    {
      return _createDateTimeConverter(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultConverterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME, // factory name
oj.DateTimeConverterFactory);


// JET VALIDATOR FACTORIES 

/**
 * a factory implementation to create an instance of the built-in dateTimeRange validator of type 
 * {@link oj.DateTimeRangeValidator}. 
 * 
 * @example <caption>create an instance of the dateTimeRange validator using the factory</caption>
 * var drvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE);  
 * var birthdateOptions = {min: new Date(1930, 00, 01), max: new Date(1995, 11,31)};
 * var birthdateValidator = drvf.createValidator(birthdateOptions);
 *  
 * @name oj.DateTimeRangeValidatorFactory
 * @hideconstructor
 * @class
 * @public
 * @since 0.6
 * 
 */
oj.DateTimeRangeValidatorFactory = (function () 
{
  
  function _createDateTimeRangeValidator(options) 
  {
    return new oj.DateTimeRangeValidator(options);
  }
  
  return {
    /**
     * Creates an immutable validator instance of type {@link oj.DateTimeRangeValidator} that ensures 
     * that the (datetime) value provided is within a given range.
     * 
     * @param {Object=} options an object literal used to provide the minimum, maximum and other 
     * optional values. See {@link oj.DateTimeRangeValidator} for details.<p>
     * 
     * @return {oj.DateTimeRangeValidator}
     * @memberOf oj.DateTimeRangeValidatorFactory
     * @instance
     * @ojsignature {target: "Type", for: "options", value: "oj.DateTimeRangeValidator.ValidatorOptions"}
     * @public
     */
    'createValidator': function(options) {
      return _createDateTimeRangeValidator(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultValidatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE,
                                                oj.DateTimeRangeValidatorFactory);
												
/**
 * a factory method to create an instance of the built-in dateRestriction validator of type 
 * {@link oj.DateRestrictionValidator}. 
 * 
 * @example <caption>create an instance of the dateRestriction validator using the factory </caption>
 * var drvf = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATERESTRICTION);
 * var drValidator = drvf.createValidator();
 *  
 * @name oj.DateRestrictionValidatorFactory
 * @class
 * @hideconstructor
 * @since 0.6
 * @public
 * 
 */
oj.DateRestrictionValidatorFactory = (function () 
{
  
  function _createDateRestrictionValidator(options) 
  {
    return new oj.DateRestrictionValidator(options);
  }
  
  return {
    /**
     * Creates an immutable validator instance of type oj.DateRestrictionValidator that ensures that the 
     * isoString value provided is not in a disabled entry of dayFormatter callback.
     * 
     * @param {Object=} options an object literal
     * See {@link oj.ojInputDate} and {@link oj.DateRestrictionValidator} for details.<p>
     * 
     * @return {oj.DateRestrictionValidator}
     * @memberOf oj.DateRestrictionValidatorFactory
     * @instance
     * @ojsignature {target: "Type", for: "options", value: "oj.DateRestrictionValidator.ValidatorOptions"}
     * @public
     */
    'createValidator': function(options) {
      return _createDateRestrictionValidator(options);
    }
  };
}()); // notice immediate invocation of anonymous function

/** Register the default factory provider function */
oj.Validation.__registerDefaultValidatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATERESTRICTION,
                                                oj.DateRestrictionValidatorFactory);
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*global OraDateTimeConverter:true*/
/**
 * @export
 * Placeholder here as closure compiler objects to export annotation outside of top level
 */

/**
 * @constructor
 * 
 * @classdesc Constructs an immutable instance and initializes it with the options provided. 
 * <p>
 *  The converter instance uses locale symbols for the locale set on the page (returned by 
 *  {@link oj.Config#getLocale}.
 *  </p>
 * There are several ways to initialize the converter.
 * <ul>
 * <li>Using options defined by the ECMA 402 Specification, these would be the properties year, 
 * month, day, hour, minute, second, weekday, era, timeZoneName, hour12</li>
 * <li>Using a custom date and/or time format pattern using the 'pattern' property</li>
 * <li>Using the standard date, datetime and time format lengths defined by Unicode CLDR, these 
 * would include the properties formaType, dateFormat, timeFormat.</li>
 * </ul>
 * 
 * <p>
 * The options when specified take precendence in the following order:<br>
 * 1. pattern.<br>
 * 2. ECMA options.<br>
 * 3. formatType/dateFormat/timeFormat.
 * <p>
 * The converter provides great leniency when parsing a user input value to a date in the following 
 * ways: <br/>
 * <ul>
 * <li>Allows use of any character for separators irrespective of the separator specified in the 
 * associated pattern. E.g., if pattern is set to 'y-M-d', the following values are all valid - 
 * 2013-11-16, 2013/11-16 and 2013aaa11xxx16.</li>
 * <li>Allows specifying 4 digit year in any position in relation to day and month. E.g., 11-2013-16 
 * or 16-11-2013</li>
 * <li>Supports auto-correction of value, when month and day positions are swapped as long as the 
 * date is > 12 when working with the Gregorian calendar. E.g., if the pattern is 'y-M-d', 
 * 2013-16-11 will be auto-corrected to 2013-11-16. However if both date and month are less or equal 
 * to 12, no assumptions are made about the day or month and the value parsed against the exact pattern.</li>
 * <li>Supports auto-correction of value, for the short and long types of weekday and month names. 
 * So they can are used anywhere in the value. E.g., if the expected pattern is E, MMM, d, y, all 
 * these values are acceptable - Tue, Nov 26 2013 or Nov, Tue 2013 26 or 2013 Tue 26 Nov. <br/>
 * NOTE: Lenient parsing of narrow era, weekday or month name is not supported because of ambiguity in 
 * choosing the right value. So we expect for narrow era, weekday or month option that values be 
 * provided either in their short or long forms. E.g., Sat, March 02, 2013.
 * </li>
 * <li>Specifying the weekday is optional. E.g., if the expected pattern is E, MMM, d, y; then 
 * entering Nov 26, 2013, is automatically turned to Tuesday Nov 26, 2013. But entering an invalid 
 * weekday, i.e., if the weekday does not match the date, an exception is thrown.</li>
 * <li>Leniency rules apply equally no matter which option is used - pattern, ECMA options or formatType</li>
 * </ul>
 * 
 * <p>
 * Lenient parse can be disabled by setting the property lenientParse to "none". In which case the user input must
 * be an exact match of the expected pattern and all the leniency described above will be disabled.
 *
 * @param {Object=} options - an object literal used to provide an optional information to 
 * initialize the converter.<p>
 * 
 * @example <caption>Create a date time converter using no options. This uses the default value for 
 * year, month, day properties</caption>
 * var converterFactory = oj.Validation.converterFactory("datetime");
 * converter = converterFactory.createConverter();
 * var resolved = converter.resolvedOpions();
 * // logs "day=numeric, month=numeric, year=numeric"
 * console.log("day=" + resolved.day + ", month=" + resolved.month + ", year=" + resolved.year);
 * <br/>
 * 
 * @example <caption>Create a date time converter using the ECMA options to represent date</caption>
 * var options = { year:'2-digit', month: '2-digit', day: '2-digit'};
 * var converterFactory = oj.Validation.converterFactory("datetime");
 * converter = converterFactory.createConverter(options);<br/>
 * 
 * @example <caption>Create a date time converter using the 'pattern' option</caption>
 * var options = {pattern: 'MM-dd-yyyy'}; 
 * var converterFactory = oj.Validation.converterFactory("datetime");
 * converter = converterFactory.createConverter(options);<br/>
 * 
 * @example <caption>Create a date time converter using the standard format length</caption>
 * var options = {formatType: 'date', dateFormat: 'medium'}; 
 * var converterFactory = oj.Validation.converterFactory("datetime");
 * converter = converterFactory.createConverter(options);<br/>
 *
 * @example <caption>Create a date time converter using specific pattern with IANA timezone ID with 
 * isoStrFormat of offset.</caption>
 * var options = {pattern: 'MM/dd/yy hh:mm:ss a Z', timeZone: 'America/Los_Angeles', isoStrFormat: 'offset'}; 
 * var converterFactory = oj.Validation.converterFactory("datetime");
 * converter = converterFactory.createConverter(options);<br/>
 *
 * @example <caption>Create a date time converter using specific pattern with Etc/GMT timezone ID with 
 * isoStrFormat of zulu.</caption>
 * var options = {pattern: 'MM/dd/yy hh:mm:ss a Z', timeZone: 'Etc/GMT-08:00', isoStrFormat: 'zulu'};  
 * var converterFactory = oj.Validation.converterFactory("datetime");
 * converter = converterFactory.createConverter(options);<br/>
 * 
 * @example <caption>Disable lenient parse.</caption>
 * var options = {pattern: 'MM/dd/yy', lenientParse:'none'};  
 * var converterFactory = oj.Validation.converterFactory("datetime");
 * converter = converterFactory.createConverter(options);
 * var str = "14/05/16";
 * var obj = converter.parse(str); --> RangeError: 13 is out of range.  Enter a value between 1 and 12 for month.<br/>
 *
 * @export
 * @augments oj.DateTimeConverter
 * @name oj.IntlDateTimeConverter
 * @ojsignature [{target: "Type", 
 *                value: "class IntlDateTimeConverter extends DateTimeConverter"},
 *               {target: "Type",
 *                value: "oj.IntlDateTimeConverter.ConverterOptions",
 *                for: "options", jsdocOverride: true}
 *              ]
 * @since 0.6
 */
oj.IntlDateTimeConverter = function(options)
{
  this.Init(options);
};

/**
 * @typedef {object} oj.IntlDateTimeConverter.ConverterOptions
 * @property {('2-digit'|'numeric')=} year - allowed values are "2-digit", "numeric". When no options are 
 * set the default value of "numeric" is used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the year, padded: 00-99.</td>
 *       <td>2001 => 01, 2016 => 16</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the year depending on the value.</td>
 *       <td>2010, 199</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {number=} two-digit-year-start - the 100-year period 2-digit year. 
 * During parsing, two digit years will be placed in the range two-digit-year-start to two-digit-year-start + 100 years. 
 * The default is 1950.
 * <p style='padding-left: 5px;'>
 * Example: if two-digit-year-start is 1950, 10 is parsed as 2010<br/><br/>
 * Example: if two-digit-year-start is 1900, 10 is parsed as 1910
 * </p>
 *
 * @property {('2-digit'|'numeric'|'narrow'|'short'|'long')=} month - specifies how the month is formatted. Allowed values are 
 * "2-digit", "numeric", "narrow", "short", "long". The last 3 values behave in the same way as for 
 * weekday, indicating the length of the string used. When no options are set the default value of 
 * "numeric" is used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the month, padded: 01-12.</td>
 *       <td>1 => 01, 12 => 12</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the month depending on the value.</td>
 *       <td>1, 11</td>
 *     </tr>
 *     <tr>
 *       <td>narrow</td>
 *       <td>narrow name of the month.</td>
 *       <td>J</td>
 *     </tr>
 *     <tr>
 *       <td>short</td>
 *       <td>abbreviated name of the month.</td>
 *       <td>Jan</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>wide name of the month.</td>
 *       <td>January</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {('2-digit'|'numeric')=} day - specifies how the day is formatted. Allowed values are "2-digit",
 *  "numeric". When no options are set the default value of "numeric" is used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the day in month, padded: 01-31.</td>
 *       <td>1 => 01, 27 => 27</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the day in month depending on the value.</td>
 *       <td>1, 31</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {('2-digit'|'numeric')=} hour - specifies how the hour is formatted. Allowed values are 
 * "2-digit" or "numeric". The hour is displayed using the 12 or 24 hour clock, depending on the 
 * locale. See 'hour12' for details.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the hour, padded: 01-24 depending on the locale.</td>
 *       <td>1 => 01, 24 => 24</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the day in month depending on the value.</td>
 *       <td>1, 24</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {('2-digit'|'numeric')=} minute - specifies how the minute is formatted. Allowed values are 
 * "2-digit", "numeric". Although allowed values for minute are numeric and 2-digit, minute is always 
 * displayed as 2 digits: 00-59.
 *
 * @property {('2-digit'|'numeric')=} second - specifies whether the second should be displayed as "2-digit" 
 * or "numeric". Although allowed values for second are numeric and 2-digit, second is always displayed 
 * as 2 digits: 00-59.
 * 
 * @property {('numeric')=} millisecond - specifies how the minute is formatted. Allowed 
 * value is "numeric". millisecond is always displayed as 3-digits except the case where only millisecond 
 * is present (hour and minute not specified) in which case we display it as no-padded number, example: .5
 *
 * @property {('narrow'|'short'|'long')=} weekday - specifies how the day of the week is formatted. If absent, it 
 * is not included in the date formatting. Allowed values are "narrow", "short", "long" indicating the 
 * length of the string used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>narrow</td>
 *       <td>narrow name of the day of week.</td>
 *       <td>M</td>
 *     </tr>
 *     <tr>
 *       <td>short</td>
 *       <td>abbreviated name of the day of week.</td>
 *       <td>Mon</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>wide name of the day of week.</td>
 *       <td>Monday</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {('narrow'|'short'|'long')=} era - specifies how the era is included in the formatted date. If 
 * absent, it is not included in the date formatting. Allowed values are "narrow", "short", "long".
 * Although allowed values are narrow, short, long, we only display era in abbreviated format: BC, AD.
 *
 * @property {('short'|'long')=} timeZoneName - allowed values are "short", "long".
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>short</td>
 *       <td>short name of the time zone.</td>
 *       <td>short: short name of the time zone: PDT, PST, EST, EDT. Note: Not all locales have 
 *           translations for short time zone names, in this case we display the English short name</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>short name of the time zone.</td>
 *       <td>Pacific Standard Time, Pacific Daylight Time.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} timeZone - The possible values of the timeZone property are valid IANA 
 * timezone IDs. If the users want to pass an offset, they can use one of the Etc/GMT timezone IDs.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>IANA ID</td>
 *       <td>America/Los_Angeles, Europe/Paris</td>
 *     </tr>
 *     <tr>
 *       <td>Offset</td>
 *       <td>Etc/GMT-8. The offset is positive if the local time zone is behind UTC and negative if it is ahead. 
 *           The offset range is between Etc/GMT-14 and Etc/GMT+12 (UTC-12 and UTC+14). Which means that Etc/GMT-8 
 *           is equivalent to UTC+08.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {('offset'|'zulu'|'local'|'auto')=} isoStrFormat - specifies in which format the ISO string is returned. 
 * The possible values of isoStrFormat are: "offset", "zulu", "local", "auto". 
 * The default format is auto.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>offset</td>
 *       <td>time zone offset from UTC.</td>
 *       <td>2016-01-05T11:30:00-08:00</td>
 *     </tr>
 *     <tr>
 *       <td>zulu</td>
 *       <td>zulu time or UTC time.</td>
 *       <td>2016-01-05T19:30:00Z</td>
 *     </tr>
 *     <tr>
 *       <td>local</td>
 *       <td>local time, does not contain time zone offset.</td>
 *       <td>2016-01-05T19:30:00</td>
 *     </tr>
 *     <tr>
 *       <td>auto</td>
 *       <td>auto means the returned ISO string depends on the input string and options</td>
 *       <td></td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {(boolean)=} dst - The dst option can be used for time only values in conjunction with offset. 
 * Setting dst to true indicates the time is in DST. By default the time is interpreted as standard time. 
 * The possible values of dst are: "true" or "false". Default is "false".
 * <p style='padding-left: 5px;'>
 * Due to Daylight Saving Time, there is a possibility that a time exists twice If the time falls in the duplicate window 
 * (switching from daylight saving time to standard time). The application can disambiguate the time in the overlapping 
 * period using the dst option. Setting dst to true indicates the time is in DST. By default the time is interpreted as 
 * standard time.<br/><br/>
 * Example: On November 1st, 2105 in US the time between 1 and 2 AM will be repeated. The dst option can indicate the 
 * distinction as follows. Initially the time is in DST, so dst:'true' is specified.<br/>
 * var options = {formatType:'datetime', dateFormat:'short', timeFormat:'medium', timeZone:'America/Los_Angeles', isoStrFormat: 'offset', dst : true};<br/>
 * var localeElements = oj.getLocaleElemnts();<br/>
 * var str= "11/1/15 1:59:59 AM";<br/>
 * cnv.parse(str, localeElements, options);-->2015-11-01T01:59:59-07:00
 * <br/><br/>
 * If the user does not pass the dst option, the time will be interpreted as standard time. 
 * var options = {formatType:'datetime', dateFormat:'short', timeFormat:'medium', timeZone:'America/Los_Angeles'};<br/>
 * var localeElements = oj.getLocaleElemnts();<br/>
 * var str= "11/1/15 1:59:59 AM";<br/>
 * cnv.parse(str, localeElements, options);-->2015-11-01T01:59:59-08:00
 * <br/><br/>
 * At 2AM, DST will be over and the clock brings back to 1AM. Then the dst option should be false or not specified.
 * var options = {formatType:'datetime', dateFormat:'short', timeFormat:'medium', timeZone:'America/Los_Angeles', isoStrFormat: 'offset'};<br/>
 * var localeElements = oj.getLocaleElemnts();<br/>
 * var str= "11/1/15 1:00:00 AM";<br/>
 * cnv.parse(str, localeElements, options);-->2015-11-01T01:00:00-08:00
 * </p>
 *
 * @property {boolean=} hour12 - specifies what time notation is used for formatting the time. 
 * A true value uses the 12-hour clock and false uses the 24-hour clock (often called military time 
 * in the US). This property is undefined if the hour property is not used when formatting the date.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>true</td>
 *       <td>T13:10  is formatted as “1:10 PM”</td>
 *     </tr>
 *     <tr>
 *       <td>false</td>
 *       <td>T13:10 is formatted as “13:10”</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 * 
 * @property {string=} pattern - a localized string pattern, where the the characters used in 
 * pattern conform to Unicode CLDR for date time formats. This will override all other options 
 * when present. <br/>
 * NOTE: 'pattern' is provided for backwards compatibility with existing apps that may want the 
 * convenience of specifying an explicit format mask. Setting a 'pattern' will override the default 
 * locale specific format.
 * NOTE: The supported tokens for timezone are of 'Z', 'VV', and 'X'.<br/><br/>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Letter</th>
 *       <th>Date or Time Component</th>
 *       <th>Examples</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>G, GG, GGG</td>
 *       <td>Era designator</td>
 *       <td>AD</td>
 *     </tr>
 *     <tr>
 *       <td>y</td>
 *       <td>numeric representation of year</td>
 *       <td>1, 2014</td>
 *     </tr>
 *     <tr>
 *       <td>yy</td>
 *       <td>2-digit representation of year</td>
 *       <td>01, 14</td>
 *     </tr>
 *     <tr>
 *       <td>yyyy</td>
 *       <td>4-digit representation of year</td>
 *       <td>0001, 2014</td>
 *     </tr>
 *     <tr>
 *       <td>M</td>
 *       <td>Numeric representation of month in year: (1-12)</td>
 *       <td>1, 12</td>
 *     </tr>
 *     <tr>
 *       <td>MM</td>
 *       <td>2-digit representation of month in year: (01-12)</td>
 *       <td>01, 12</td>
 *     </tr>
 *     <tr>
 *       <td>MMM</td>
 *       <td>Formatted  name of the month, abbreviated</td>
 *       <td>Jan</td>
 *     </tr>
 *     <tr>
 *       <td>MMMM</td>
 *       <td>Formatted  name of the month, wide</td>
 *       <td>January</td>
 *     </tr>
 *     <tr>
 *       <td>MMMMM</td>
 *       <td>Formatted  name of the month, narrow</td>
 *       <td>J</td>
 *     </tr>
 *     <tr>
 *       <td>LLL</td>
 *       <td>Stand-alone name of the month, abbreviated</td>
 *       <td>Jan</td>
 *     </tr>
 *     <tr>
 *       <td>LLLL</td>
 *       <td>Stand-alone name of the month, wide</td>
 *       <td>January</td>
 *     </tr>
 *     <tr>
 *       <td>LLLLL</td>
 *       <td>Stand-alone name of the month, narrow</td>
 *       <td>J</td>
 *     </tr>
 *     <tr>
 *       <td>d</td>
 *       <td>Numeric representation of  day in month (1-31)</td>
 *       <td>1, 21</td>
 *     </tr>
 *     <tr>
 *       <td>dd</td>
 *       <td>2-digit representation of  day in month (01-31)</td>
 *       <td>01, 21</td>
 *     </tr>
 *     <tr>
 *       <td>E, EE, EEE</td>
 *       <td>Formatted name of day in week, abbreviated</td>
 *       <td>Tue</td>
 *     </tr>
 *     <tr>
 *       <td>EEEE</td>
 *       <td>Formatted name of day in week, wide</td>
 *       <td>Tuesday</td>
 *     </tr>
 *     <tr>
 *       <td>EEEEE</td>
 *       <td>Formatted name of day in week, narrow</td>
 *       <td>T</td>
 *     </tr>
 *     <tr>
 *       <td>c, cc, ccc</td>
 *       <td>Stand-alone name of day in week, abbreviated</td>
 *       <td>Tue</td>
 *     </tr>
 *     <tr>
 *       <td>cccc</td>
 *       <td>Stand-alone name of day in week, wide</td>
 *       <td>Tuesday</td>
 *     </tr>
 *     <tr>
 *       <td>ccccc</td>
 *       <td>Stand-alone name of day in week, narrow</td>
 *       <td>T</td>
 *     </tr>
 *     <tr>
 *       <td>a</td>
 *       <td>am/pm marker</td>
 *       <td>PM</td>
 *     </tr>
 *     <tr>
 *       <td>H</td>
 *       <td>Numeric hour in day (0-23)</td>
 *       <td>1, 23</td>
 *     </tr>
 *     <tr>
 *       <td>HH</td>
 *       <td>2-digit hour in day (00-23)</td>
 *       <td>01, 23</td>
 *     </tr>
 *     <tr>
 *       <td>h</td>
 *       <td>Numeric  hour in am/pm (1-12)</td>
 *       <td>1, 12</td>
 *     </tr>
 *     <tr>
 *       <td>hh</td>
 *       <td>2-digit hour in day (01-12)</td>
 *       <td>01, 12</td>
 *     </tr>
 *     <tr>
 *       <td>k</td>
 *       <td>Numeric  hour in day (1-24)</td>
 *       <td>1, 24</td>
 *     </tr>
 *     <tr>
 *       <td>kk</td>
 *       <td>2-digit hour in day (1-24)</td>
 *       <td>01, 24</td>
 *     </tr>
 *     <tr>
 *       <td>K</td>
 *       <td>Numeric  hour in am/pm (0-11)</td>
 *       <td>1, 11</td>
 *     </tr>
 *     <tr>
 *       <td>KK</td>
 *       <td>2-digit hour in am/pm (0-11)</td>
 *       <td>01, 11</td>
 *     </tr>
 *     <tr>
 *       <td>m, mm</td>
 *       <td>2-digit  minute in hour (00-59)</td>
 *       <td>05, 59</td>
 *     </tr>
 *     <tr>
 *       <td>s, ss</td>
 *       <td>2-digit second in minute (00-59)</td>
 *       <td>01, 59</td>
 *     </tr>
 *     <tr>
 *       <td>S</td>
 *       <td>Numeric  Millisecond (0-999)</td>
 *       <td>1, 999</td>
 *     </tr>
 *     <tr>
 *       <td>SS</td>
 *       <td>2-digit Millisecond (00-999)</td>
 *       <td>01, 999</td>
 *     </tr>
 *     <tr>
 *       <td>SSS</td>
 *       <td>3-digit Millisecond (000-999)</td>
 *       <td>001, 999</td>
 *     </tr>
 *     <tr>
 *       <td>z, zz, zzz</td>
 *       <td>Abbreviated time zone name</td>
 *       <td>PDT, PST</td>
 *     </tr>
 *     <tr>
 *       <td>zzzz</td>
 *       <td>Full time zone name</td>
 *       <td>Pacific Standard Time, Pacific Daylight Time</td>
 *     </tr>
 *     <tr>
 *       <td>Z, ZZ, ZZZ</td>
 *       <td>Sign hours minutes</td>
 *       <td>-0800</td>
 *     </tr>
 *     <tr>
 *       <td>X</td>
 *       <td>Sign hours</td>
 *       <td>-08</td>
 *     </tr>
 *     <tr>
 *       <td>XX</td>
 *       <td>Sign hours minutes</td>
 *       <td>-0800</td>
 *     </tr>
 *     <tr>
 *       <td>XXX</td>
 *       <td>Sign hours:minutes</td>
 *       <td>-08:00</td>
 *     </tr>
 *     <tr>
 *       <td>VV</td>
 *       <td>Time zone ID</td>
 *       <td>Americs/Los_Angeles</td>
 *     </tr>
 *   </tbody>
 * </table>
 * 
 * @property {('date'|'time'|'datetime')=} formatType - determines the 'standard' date and/or time format lengths 
 * to use. Allowed values: "date", "time", "datetime". See 'dateFormat' and 'timeFormat' options. 
 * When set a value must be specified.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>datetime</td>
 *       <td>date and time portions are displayed.</td>
 *       <td>“September 20, 2015 12:04 PM”, “September 20, 2015 12:05:35 PM Pacific Daylight Time”</td>
 *     </tr>
 *     <tr>
 *       <td>date</td>
 *       <td>date portion only is displayed.</td>
 *       <td>“September 20, 2015”</td>
 *     </tr>
 *     <tr>
 *       <td>time</td>
 *       <td>time portion only is displayed.</td>
 *       <td>“12:05:35”</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {('short'|'medium'|'long'|'full')=} dateFormat - specifies the standard date format length to use when 
 * formatType is set to "date" or "datetime". Allowed values are : "short" (default), "medium", "long", 
 * "full". 
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>short</td>
 *       <td>9/20/15</td>
 *     </tr>
 *     <tr>
 *       <td>medium</td>
 *       <td>Sep 20, 2015</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>September 20, 2015</td>
 *     </tr>
 *     <tr>
 *       <td>full</td>
 *       <td>Sunday, September 20, 2015</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {('short'|'medium'|'long'|'full')=} timeFormat - specifies the standard time format length to use when 
 * 'formatType' is set to "time" or "datetime". Allowed values: "short" (default), "medium", "long", 
 * "full". 
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>short</td>
 *       <td>12:11 PM</td>
 *     </tr>
 *     <tr>
 *       <td>medium</td>
 *       <td>12:11:23 PM</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>12:12:19 PM PDT</td>
 *     </tr>
 *     <tr>
 *       <td>full</td>
 *       <td>12:12:37 PM Pacific Daylight Time</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 * 
 *  @property {('full'|'none')=} lenientParse - The lenientParse property can be used to enable or disable leninet parsing.
 *  Allowed values: "full" (default), "none". 
 * <p style='padding-left: 5px;'>
 * By default the lenient parse is enabled and the leniency rules descibed above will be used. When lenientParse is
 * set to "none" the lenient parse is disabled and the user input must match the expected input otherwise an exception will
 * be thrown.</p>
 */
// Subclass from oj.Object 
oj.Object.createSubclass(oj.IntlDateTimeConverter, oj.DateTimeConverter, "oj.IntlDateTimeConverter");
oj.IntlDateTimeConverter._DEFAULT_DATE = new Date(1998, 10, 29, 15, 45, 31);

/**
 * Initializes the date time converter instance with the set options.
 * @param {Object=} options an object literal used to provide an optional information to initialize 
 * the converter.<p>
 * 
 * @export
 * @ignore
 */
oj.IntlDateTimeConverter.prototype.Init = function(options) 
{
  oj.IntlDateTimeConverter.superclass.Init.call(this, options);
};


// Returns the wrapped date time converter implementation object.
oj.IntlDateTimeConverter.prototype._getWrapped = function ()
{
  if (!this._wrapped)
  {
    this._wrapped = OraDateTimeConverter.getInstance();
  }
  
  return this._wrapped;
};

/**
 * Formats the isoString value using the options provided and returns a string value. 
 * <p>
 * 
 * @param {string} value to be formatted for display which should be an isoString
 * @return {string|null} the formatted value suitable for display
 * 
 * @throws {Error} a ConverterError both when formatting fails, and if the options provided during 
 * initialization cannot be resolved correctly.
 * 
 * @example <caption>To convert Javascript Date to a local iso string before passing to <code class="prettyprint">format</code></caption>
 * var date = new Date();
 * var formatted = converter.format(oj.IntlConverterUtils.dateToLocalIso(date));
 *
 * @example <caption>Standard format invocation
 * var formatted = converter.format("2013-12-01T20:00:00-08:00");
 * 
 * @see oj.IntlConverterUtils.dateToLocalIso
 * @memberOf oj.IntlDateTimeConverter
 * @export
 */
oj.IntlDateTimeConverter.prototype.format = function (value) 
{
  var localeElements = oj.LocaleData.__getBundle(), locale = oj.Config.getLocale(), 
          resolvedOptions = this.resolvedOptions(), converterError;
  
  // undefined, null and empty string values all return null. If value is NaN then return "".
  // TODO: Should we automatically parse() the integer value representing the number of milliseconds 
  // since 1 January 1970 00:00:00 UTC (Unix Epoch)?
  if (value == null || 
      (typeof value === "string" && (oj.StringUtils.trim("" + value)).length === 0)) 
  {
    return oj.IntlConverterUtils.__getNullFormattedValue();
  }
  
  try
  {
    return this._getWrapped().format(value, localeElements, resolvedOptions, locale);
  }
  catch (e)
  {
    converterError = this._processConverterError(e, value);
    throw converterError;
  }
};


/**
 * Formats an ISOString as a relative date time, using the relativeOptions. 
 * <p>
 * 
 * @param {string} value - value to be formatted. This value is compared with the current date 
 * on the client to arrive at the relative formatted value.
 * @param {Object=} relativeOptions - an Object literal containing the following properties. The 
 * default options are ignored during relative formatting - 
 * @param {string=} relativeOptions.formatUsing - Specifies the relative formatting convention to. 
 * Allowed values are "displayName" and “calendar”. Setting value to 'displayName' uses the relative 
 * display name for the instance of the dateField, and one or two past and future instances. 
 * When omitted we use the implicit rules.
 * @param {string=} relativeOptions.dateField - To be used in conjunction of 'displayName'  value 
 * of formatUsing attribute.  Allowed values are: "day", "week", "month", "year", "hour", "minute", "second".
 * @param {string=} relativeOptions.relativeTime - Allowed values are: "fromNow", "toNow". 
 * "fromNow" means the system's current date is the reference and "toNow" means the value attribute 
 * is the reference. Default "fromNow".
 * @param {boolean=} relativeOptions.dateOnly - A boolean value that can be used in conjunction with 
 * “calendar” of formatUsing attribute.  When set to true date only format is used. Example: “Sunday”  
 * instead of “Sunday at 2:30 PM”. Default value is false.
 * @param {string=} relativeOptions.timeZone - The timeZone attribute can be used to specify the 
 * time zone of the  value parameter.  The system’s time zone is used for the current time. If timeZone 
 * attribute is not specified, we use the system’s time zone  for both. The value parameter, which is an 
 * iso string, can be Z or contain and offset, in this case  the timeZone attribute is overwritten.
 *
 * @return {string|null} relative date. null if the value falls out side the supported relative range.
 * @throws {Object} an instance of {@link oj.ConverterError}
 * 
 * @example <caption>Relative time in the future using implicit rules</caption>
 * var dateInFuture = new Date();
 * dateInFuture.setMinutes(dateInFuture.getMinutes() + 41);
 * var formatted = converter.formatRelative(oj.IntlConverterUtils.dateToLocalIso(dateInFuture)); -> in 41 minutes
 *
 * @example <caption>Relative time in the past using implicit rules</caption>
 * var dateInPase = new Date();
 * dateInPase.setHours( dateInPast.getHours() - 20);
 * var formatted = converter.formatRelative(oj.IntlConverterUtils.dateToLocalIso(dateInPase)); -> 20 hours ago
 *
 * @example <caption>Relative time using dateField. Assuming system’s current date is 2016-07-28.</caption>
 * Format relative year:
 * var options = {formatUsing: ”displayName”, dateField: “year”};
 * var formatted = converter.formatRelative("2015-06-01T00:00:00", options); -> last year
 *
 * @example <caption>Relative time using relativeTime. Assuming system’s current date is 2016-07-28.</caption>
 * var options = {formatUsing: ”displayName”, dateField: “day”, relativeTime: ”fromNow”};
 * var formatted = converter.formatRelative("2016-07-28T00:00:00", options); -> tomorrow
 * options = {formatUsing: ”displayName”, dateField: “day”, relativeTime: toNow};
 * formatted = converter.formatRelative("2016-07-28T00:00:00", options); -> yesterday
 *
 * @example <caption>Relative time using calendar. Assuming system’s current date is 2016-07-28.</caption>
 * var options = {formatUsing: ”calendar”};
 * var formatted = converter.formatRelative("2016-07-28T14:15:00", options); -> tomorrow at 2:30 PM
 *
 * 
 * @example <caption>Relative time using timeZone. Assuming that the system’s time zone is America/Los_Angeles.</caption>
 * var options = {timeZone:”America/New_York”};
 * var nyDateInFuture = new Date();
 * nyDateInFuture.setHours(nyDateInFuture.getHours() + 6);
 * var formatted = converter.formatRelative(oj.IntlConverterUtils.dateToLocalIso(nyDateInFuture), options); -> in 3 hours
 * 
 * @see oj.IntlConverterUtils.dateToLocalIso
 * 
 * @memberOf oj.IntlDateTimeConverter
 * @export
 */
oj.IntlDateTimeConverter.prototype.formatRelative = function(value, relativeOptions) 
{
  var localeElements = oj.LocaleData.__getBundle(), locale = oj.Config.getLocale(), converterError;
  try
  {
    return this._getWrapped().formatRelative(value, localeElements, relativeOptions, locale);
  }
  catch (e)
  {
    converterError = this._processConverterError(e, value);
    throw converterError;
  }
};

/**
 * It returns null for the placeholder hint. There is no default placeholder hint when using our IntlDateTimeConverter. 
 * 
 * @return {null} hint for the converter.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 */
oj.IntlDateTimeConverter.prototype.getHint = function ()
{ 
  //do not return any hint.
  return null;
};

// Returns the hint value.
oj.IntlDateTimeConverter.prototype._getHintValue = function()
{
  var value = "";
  try
  {
    // example date 
    value =  this.format(oj.IntlConverterUtils.dateToLocalIso(oj.IntlDateTimeConverter._DEFAULT_DATE));
  }
  catch (e)
  {
    if (e instanceof oj.ConverterError)
    {
      // Something went wrong and we don't have a way to retrieve a valid value.
      // TODO: Log an error
      value = "";
    }
  }
  finally
  {
    return value;
  }
};


/**
 * Returns the options called with converter initialization.
 * @return {Object} an object of options.
 * @ojsignature {target: "Type", for: "returns",
 *                value: "oj.IntlDateTimeConverter.ConverterOptions"}
 * @export
 * @memberOf oj.IntlDateTimeConverter
 * 
 */
oj.IntlDateTimeConverter.prototype.getOptions = function () 
{
  return oj.IntlDateTimeConverter.superclass.getOptions.call(this);
};

/**
 * Returns an object literal with locale and formatting options computed during initialization of 
 * the object. If options was not provided at the time of initialization, the properties will be 
 * derived from the locale defaults.
 * @return {Object} an object of resolved options. Properties whose corresponding internal 
 * properties are not present are not assigned. 
 * @ojsignature {target: "Type", for: "returns",
 *                value: "oj.IntlDateTimeConverter.ConverterOptions"}
 * @throws a oj.ConverterError when the options that the converter was initialized with are invalid. 
 * @memberOf oj.IntlDateTimeConverter
 * @export
 */
oj.IntlDateTimeConverter.prototype.resolvedOptions = function ()
{
  var localeElements, locale = oj.Config.getLocale(), converterError, options = this.getOptions();
  // options are resolved and cached for a locale
  if ((locale !== this._locale) || !this._resolvedOptions)
  {
    localeElements = oj.LocaleData.__getBundle();
    try
    {
      if (!localeElements)
      {
        oj.Logger.error("locale bundle for the current locale %s is unavailable", locale);
        return {};
      }
      // cache if successfully resolved
      this._resolvedOptions = this._getWrapped().resolvedOptions(localeElements, 
                                                                 options, 
                                                                 locale);
      this._locale = locale;
    }
    catch (e)
    {
      converterError = this._processConverterError(e);
      throw converterError;
    }
  }
  
  return this._resolvedOptions;
};

/**
 * Returns true if a 24-hour format is set; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isHourInDaySet = function()
{
  var ro = this.resolvedOptions(), hour = ro['hour'], hour12 = ro['hour12'];
  if (hour && !hour12)
  {
    // if hour12=false or not set and hour is set to some value
    return true;
  }
  
  return false;
};

/**
 * Returns true if 12-hour is set; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isHourInAMPMSet = function()
{
  var ro = this.resolvedOptions(), hour = ro['hour'], hour12 = ro['hour12'];  
  if (hour && hour12)
  {
    // if hour12==true and hour is set to some value
    return true;
  }
  
  return false;
  
};

/**
 * Returns true if minutes are shown in the time portion; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isMinuteSet = function()
{
  return this._isOptionSet('minute');
};

/**
 * Returns true if seconds are shown in the time portion; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isSecondSet = function()
{
  return this._isOptionSet('second');
};

/**
 * Returns true if milliseconds are shown in the time portion; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isMilliSecondSet = function()
{
  return this._isOptionSet('millisecond');
};

/**
 * Returns true if year is shown in the date portion; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isYearSet = function()
{
  return this._isOptionSet('year');
};

/**
 * Returns true if month is shown in the date portion; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isMonthSet = function()
{
  return this._isOptionSet('month');
};

/**
 * Returns true if day is shown in the date portion; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isDaySet = function()
{
  return this._isOptionSet('day');
};

/**
 * Returns true if the day name is shown in the date portion; false otherwise.
 * @memberOf oj.IntlDateTimeConverter
 * @export
 * @return {boolean}
 */
oj.IntlDateTimeConverter.prototype.isDayNameSet = function()
{
  return this._isOptionSet('weekday');
};

/**
 * Returns the calculated week for the isoString value
 * 
 * @param {string} value to return the calculated week of
 * @return {number} calculated week.
 * 
 * @memberOf oj.IntlDateTimeConverter
 * @export
 */
oj.IntlDateTimeConverter.prototype.calculateWeek = function(value)
{
  return this._getWrapped().calculateWeek(value, oj.LocaleData.__getBundle(), oj.Config.getLocale());
};

/**
 * Parses the value using the options provided and returns the date and time as a string 
 * expressed using the ISO-8601 format (http://en.wikipedia.org/wiki/ISO_8601).
 * 
 * <p>
 * For converter options specific to a date, the iso date representation alone is returned. <br/>
 * For time only options, the iso time representation alone is returned. <br/>
 * For options that include both date and time, the iso date and time representation is 
 * returned.<br/>
 * </p>
 * 
 * <p>
 * For convenience, if one wishes to retrieve a JavaScript Date object from the local isoString, a 
 * utility function oj.IntlConverterUtils.isoToLocalDate is provided.
 * 
 * Or oj.IntlConverterUtils.isoToDate if one wish to utilize the timezone of the isoString.
 * </p>
 * 
 * @param {string} value to parse
 * @return {string|null} the parsed value as an ISOString.
 * 
 * @throws {Error} a ConverterError both when parsing fails, and if the options provided during 
 * initialization cannot be resolved correctly. Parsing can also fail when the value includes a time
 *  zone. 
 * 
 * @example <caption>Parse date, time and date & time values using <code class="prettyprint">parse</code> method.</caption>
 * &lt;!-- For date-time values  -->
 * var options = {pattern: 'MM/dd/yy hh:mm:ss a'};
 * var conv = oj.Validation.converterFactory('datetime').createConverter(options);
 * cnv.parse('09/11/14 03:02:01 PM'); // '2014-09-11T15:02:01'
 * 
 * &lt;!-- For date values -->
 * var options = {pattern: 'MM/dd/yy'};
 * cnv.parse('09/11/14'); // '2014-09-11'
 * 
 * &lt;!-- For time values -->
 * var options = {pattern: 'hh:mm:ss a'};
 * cnv.parse('03:02:01 PM'); // 'T15:02:01'
 * 
 * @example <caption>Convert from iso string to Javascript Date object</caption>
 * var isoString = '2014-10-20T15:02:01';
 * var date = oj.IntlConverterUtils.isoToLocalDate(converter.parse(isoString));
 * 
 * @see oj.IntlConverterUtils.isoToLocalDate
 * @see oj.IntlConverterUtils.isoToDate
 *  
 * @export
 */
oj.IntlDateTimeConverter.prototype.parse = function (value) 
{
  var result, parsed;
  // undefined, null and empty string values are ignored and not parsed. 
  if (value == null || value === "") 
  {
    return null;
  }
  
  var localeElements = oj.LocaleData.__getBundle(), locale = oj.Config.getLocale(), 
          resolvedOptions = this.resolvedOptions(), converterError;

  try
  {
    // date converter parses the value and returns an Object with 2 fields - 'value' and 'warning'
    result = this._getWrapped().parse(value, localeElements, resolvedOptions, locale); 
    parsed = result['value'];
    if (parsed)
    {
      // TODO: For now log a warning when we leniently parse a value; later we plan to flash the 
      // field.
      if (result['warning'])
      {
        oj.Logger.warn("The value " + value + " was leniently parsed to represent a date " + 
                (parsed.toString) ? parsed.toString() : parsed);
      }
      
    }
    return parsed;
  }
  catch (e)
  {
    converterError = this._processConverterError(e, value);
    throw converterError;
  }
};

/**
 * Compares 2 ISO 8601 strings, returning the time difference between the two
 * 
 * @param {string} isoStr first iso string
 * @param {string} isoStr2 second iso string
 * @return {number} the time difference between isoStr and isoStr2
 * @export
 */
oj.IntlDateTimeConverter.prototype.compareISODates = function (isoStr, isoStr2)
{
  var stringChecker = oj.StringUtils.isString;

  if(!stringChecker(isoStr) || !stringChecker(isoStr2)) 
  {
    throw new Error("Invalid arguments for compareISODates ", isoStr, isoStr2);
  }

  return this._getWrapped().compareISODates(isoStr, isoStr2, oj.LocaleData.__getBundle());
};

/**
 * Processes the error returned by the converter implementation and throws a oj.ConverterError 
 * instance.
 * @param {Error} e
 * @param {String|string|Date|Object=} value 
 * @throws an instance of oj.ConverterError
 * @private
 */
oj.IntlDateTimeConverter.prototype._processConverterError = function (e, value)
{
  var errorInfo = e['errorInfo'], summary, detail, errorCode, parameterMap, converterError, 
          propName, resourceKey;
  if (errorInfo)
  {
    errorCode = errorInfo['errorCode'];
    parameterMap = errorInfo['parameterMap'] || {};
    oj.Assert.assertObject(parameterMap);
    propName = parameterMap['propertyName'];
    propName = oj.Translations.getTranslatedString("oj-converter.datetime.datetimeOutOfRange."
                 + propName);
    // update the map back so that functions like __getConverterOptionError use new value
    parameterMap['propertyName'] = propName;
    
    if (e instanceof TypeError)
    {
      if (errorCode === "optionTypesMismatch" || errorCode === "optionTypeInvalid")
      {
        converterError = oj.IntlConverterUtils.__getConverterOptionError(errorCode, parameterMap);
      }
    }
    else if (e instanceof RangeError)
    {
      if (errorCode === "optionOutOfRange")
      {
        converterError = oj.IntlConverterUtils.__getConverterOptionError(errorCode, parameterMap);
      }
      else if (errorCode === 'datetimeOutOfRange') // TODO: NLS should use lower case time
      {
        // The '{value}' is out of range. Enter a value between '{minValue}' and '{maxValue}' for 
        // '{propertyName}'.
        summary = oj.Translations.getTranslatedString("oj-converter.datetime.datetimeOutOfRange.summary", 
          {'propertyName': propName,
           'value': parameterMap['value']});
        detail = oj.Translations.getTranslatedString("oj-converter.datetime.datetimeOutOfRange.detail",
          {'minValue': parameterMap['minValue'],
           'maxValue': parameterMap['maxValue']});
         
        converterError = new oj.ConverterError(summary, detail);
      }
    }
    else if (e instanceof SyntaxError)
    {
      if (errorCode === "optionValueInvalid")
      {
        converterError = oj.IntlConverterUtils.__getConverterOptionError(errorCode, parameterMap);
      }
    }
    else if (e instanceof Error)
    {
      if (errorCode === "dateFormatMismatch")
      {
        // The '{value}' does not match the expected date format '{format}'.
        resourceKey = "oj-converter.datetime.dateFormatMismatch.summary";
      }
      else if (errorCode === "timeFormatMismatch")
      {
        // The {value} does not match the expected time format {format}.
        resourceKey = "oj-converter.datetime.timeFormatMismatch.summary";
      }
      else if (errorCode === "datetimeFormatMismatch")
      {
        resourceKey = "oj-converter.datetime.datetimeFormatMismatch.summary";
      }
      else if (errorCode === "invalidTimeZoneID")
      {
        summary = oj.Translations.getTranslatedString("oj-converter.datetime.invalidTimeZoneID.summary", 
          {'timeZoneID': parameterMap['timeZoneID']});
        detail = oj.Translations.getTranslatedString("oj-converter.hint.detail",
          {'exampleValue': this._getHintValue()}); 
        
        converterError = new oj.ConverterError(summary, detail);
      }
      else if (errorCode === "nonExistingTime")
      {
        resourceKey = "oj-converter.datetime.nonExistingTime.summary";
      }
      else if (errorCode === "missingTimeZoneData")
      {
        resourceKey = "oj-converter.datetime.missingTimeZoneData.summary";
      }
      else if (errorCode === "dateToWeekdayMismatch")
      {
        summary = oj.Translations.getTranslatedString("oj-converter.datetime.dateToWeekdayMismatch.summary", 
          {'date': parameterMap['date'], 'weekday': parameterMap['weekday']});
        detail = oj.Translations.getTranslatedString("oj-converter.datetime.dateToWeekdayMismatch.detail");
        converterError = new oj.ConverterError(summary, detail);
      }
      else if (errorCode === "invalidISOString")
      {
        summary = oj.Translations.getTranslatedString("oj-converter.datetime.invalidISOString.summary", 
          {'isoStr': parameterMap['isoStr']});
        detail = oj.Translations.getTranslatedString("oj-converter.datetime.invalidISOString.detail");
        converterError = new oj.ConverterError(summary, detail);
      }

      if (resourceKey)
      {
        summary = oj.Translations.getTranslatedString(resourceKey, 
          {'value': value || parameterMap['value'],
           'format': parameterMap['format']});
        
        detail = oj.Translations.getTranslatedString("oj-converter.hint.detail",
          {'exampleValue': this._getHintValue()}); 
          
        converterError = new oj.ConverterError(summary, detail);
      }
    }
  }
  
  if (!converterError)
  {
    // An error we are unfamiliar with. Get the message and set as detail
    summary = e.message; // TODO: What should the summary be when it's missing??
    detail = e.message;
    converterError = new oj.ConverterError(summary, detail);
  }
  
  return converterError;
};

/**
 * Checks to see if an option is present in the resolved options.
 * @param {string} optionName
 * @returns {boolean} true if optionName is present.
 * @private
 */
oj.IntlDateTimeConverter.prototype._isOptionSet = function (optionName)
{
  var ro = this.resolvedOptions(), hasOption = ro[optionName] ? true : false;
  return hasOption;
};

/**
 * Gets the supported timezones for the converter. The returned value is an array of objects. Each object represents a timezone
 * and contains 2 properties: <br/>
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Property</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>id</td>
 *       <td>IANA timezone ID</td>
 *     </tr>
 *     <tr>
 *       <td>displayName</td>
 *       <td><ul>It is the concatenation of 3 string:
 *              <li>UTC timezone offset</li>
 *              <li>City name</li>
 *              <li>Generic time zone name</li>
 *           </ul>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 * @example <caption>Example of an array entry in en-US locale</caption>
 * {id: 'America/Edmonton', displayName: '(UTC-07:00) Edmonton - Mountain Time'} <br/>
 * 
 * @example <caption>Example of above entry in fr-FR locale</caption>
 * {id: 'America/Edmonton', displayName: '(UTC-07:00) Edmonton - heure des Rocheuses' } <br/> 
 * 
 * @return {Array} supported timezones
 * @since 4.0.0
 * @export
 */
oj.IntlDateTimeConverter.prototype.getAvailableTimeZones = function ()
{
  return this._getWrapped().getAvailableTimeZones(oj.LocaleData.__getBundle());
};

/**
 * Internal API to getTimePositioning.
 * @returns {Object} json object of the positioning with h + m as keys
 * @export
 * @ignore
 */
oj.IntlDateTimeConverter.prototype._getTimePositioning = function ()
{
  return this._getWrapped().getTimePositioning(oj.LocaleData.__getBundle(), this.resolvedOptions(), oj.Config.getLocale());
};


/**
 * Copyright (c) 2016, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * This is a forked version of globalize.js
 */
/*
 * Copyright Software Freedom Conservancy, Inc.
 * Dual licensed under the MIT or GPL Version 2 licenses.
 * http://jquery.org/license
 */
/**
 * @constructor
 * 
 * @classdesc OraDateTimeConverter object implements date-time parsing and formatting and 
 * relative date formatting. 
 * <p>
 * There are several ways to use the converter.
 * <ul>
 * <li>Using options defined by the ECMA 402 Specification, these would be the properties year, 
 * month, day, hour, minute, second, weekday, era, timeZone, hour12</li>
 * <li>Using a custom date and/or time format pattern using the 'pattern' property</li>
 * <li>Using the standard date, datetime and time format lengths defined by Unicode CLDR, these 
 * would include the properties formaType, dateFormat, timeFormat.</li>
 * </ul>
 * 
 * <p>
 * The options when specified take precendence in the following order:<br>
 * 1. pattern.<br>
 * 2. ECMA options.<br>
 * 3. formatType/dateFormat/timeFormat.
 * <p>
 * The converter provides great leniency when parsing a user input value to a date in the following 
 * ways: <br/>
 * <ul>
 * <li>Allows use of any character for separators irrespective of the separator specified in the 
 * associated pattern. E.g., if pattern is set to 'y-M-d', the following values are all valid - 
 * 2013-11-16, 2013/11-16 and 2013aaa11xxx16.</li>
 * <li>Allows specifying 4 digit year in any position in relation to day and month. E.g., 11-2013-16 
 * or 16-11-2013</li>
 * <li>Supports auto-correction of value, when month and day positions are swapped as long as the 
 * date is > 12 when working with the Gregorian calendar. E.g., if the pattern is 'y-M-d', 
 * 2013-16-11 will be auto-corrected to 2013-11-16. However if both date and month are less or equal 
 * to 12, no assumptions are made about the day or month and the value parsed against the exact pattern.</li>
 * <li>Supports auto-correction of value, for the short and long types of weekday and month names. 
 * So they can are used anywhere in the value. E.g., if the expected pattern is E, MMM, d, y, all 
 * these values are acceptable - Tue, Nov 26 2013 or Nov, Tue 2013 26 or 2013 Tue 26 Nov. <br/>
 * NOTE: Lenient parsing of narrow era, weekday or month name is not supported because of ambiguity in 
 * choosing the right value. So we expect for narrow era, weekday or month option that values be 
 * provided either in their short or long forms. E.g., Sat, March 02, 2013.
 * </li>
 * <li>Specifying the weekday is optional. E.g., if the expected pattern is E, MMM, d, y; then 
 * entering Nov 26, 2013, is automatically turned to Tuesday Nov 26, 2013. But entering an invalid 
 * weekday, i.e., if the weekday does not match the date, an exception is thrown.</li>
 * <li>Leniency rules apply equally no matter which option is used - pattern, ECMA options or formatType</li>
 * </ul>
 * <p>
 * Lenient parse can be disabled by setting the property lenientParse to "none". In which case the user input must
 * be an exact match of the expected pattern and all the leniency described above will be disabled.
 * @property {Object=} options - an object literal used to provide  optional information to 
 * the converter.<p>
 * @property {string=} options.year - allowed values are "2-digit", "numeric". When no options are 
 * set the default value of "numeric" is used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the year, padded: 00-99.</td>
 *       <td>2001 => 01, 2016 => 16</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the year depending on the value.</td>
 *       <td>2010, 199</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {number=} options.two-digit-year-start - the 100-year period 2-digit year. 
 * During parsing, two digit years will be placed in the range two-digit-year-start to two-digit-year-start + 100 years. 
 * The default is 1950.
 * <p style='padding-left: 5px;'>
 * Example: if two-digit-year-start is 1950, 10 is parsed as 2010<br/><br/>
 * Example: if two-digit-year-start is 1900, 10 is parsed as 1910
 * </p>
 *
 * @property {string=} options.month - specifies how the month is formatted. Allowed values are 
 * "2-digit", "numeric", "narrow", "short", "long". The last 3 values behave in the same way as for 
 * weekday, indicating the length of the string used. When no options are set the default value of 
 * "numeric" is used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the month, padded: 01-12.</td>
 *       <td>1 => 01, 12 => 12</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the month depending on the value.</td>
 *       <td>1, 11</td>
 *     </tr>
 *     <tr>
 *       <td>narrow</td>
 *       <td>narrow name of the month.</td>
 *       <td>J</td>
 *     </tr>
 *     <tr>
 *       <td>short</td>
 *       <td>abbreviated name of the month.</td>
 *       <td>Jan</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>wide name of the month.</td>
 *       <td>January</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.day - specifies how the day is formatted. Allowed values are "2-digit",
 *  "numeric". When no options are set the default value of "numeric" is used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the day in month, padded: 01-31.</td>
 *       <td>1 => 01, 27 => 27</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the day in month depending on the value.</td>
 *       <td>1, 31</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.hour - specifies how the hour is formatted. Allowed values are 
 * "2-digit" or "numeric". The hour is displayed using the 12 or 24 hour clock, depending on the 
 * locale. See 'hour12' for details.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>2-digit</td>
 *       <td>2 digit representation of the hour, padded: 01-24 depending on the locale.</td>
 *       <td>1 => 01, 24 => 24</td>
 *     </tr>
 *     <tr>
 *       <td>numeric</td>
 *       <td>variable digit representation of the day in month depending on the value.</td>
 *       <td>1, 24</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.minute - specifies how the minute is formatted. Allowed values are 
 * "2-digit", "numeric". Although allowed values for minute are numeric and 2-digit, minute is always 
 * displayed as 2 digits: 00-59.
 *
 * @property {string=} options.second - specifies whether the second should be displayed as "2-digit" 
 * or "numeric". Although allowed values for second are numeric and 2-digit, second is always displayed 
 * as 2 digits: 00-59.
 *
 * @property {string=} options.millisecond - specifies whether the millisecond should be displayed.
 * Allowed value is "numeric".
 *
 * @property {string=} options.weekday - specifies how the day of the week is formatted. If absent, it 
 * is not included in the date formatting. Allowed values are "narrow", "short", "long" indicating the 
 * length of the string used.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>narrow</td>
 *       <td>narrow name of the day of week.</td>
 *       <td>M</td>
 *     </tr>
 *     <tr>
 *       <td>short</td>
 *       <td>abbreviated name of the day of week.</td>
 *       <td>Mon</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>wide name of the day of week.</td>
 *       <td>Monday</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.era - specifies how the era is included in the formatted date. If 
 * absent, it is not included in the date formatting. Allowed values are "narrow", "short", "long".
 * Although allowed values are narrow, short, long, we only display era in abbreviated format: BC, AD.
 *
 * @property {string=} options.timeZoneName - allowed values are "short", "long".
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>short</td>
 *       <td>short name of the time zone.</td>
 *       <td>short: short name of the time zone: PDT, PST, EST, EDT. Note: Not all locales have 
 *           translations for short time zone names, in this case we display the English short name</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>short name of the time zone.</td>
 *       <td>Pacific Standard Time, Pacific Daylight Time.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.timeZone - The possible values of the timeZone property are valid IANA 
 * timezone IDs. If the users want to pass an offset, they can use one of the Etc/GMT timezone IDs.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>IANA ID</td>
 *       <td>America/Los_Angeles, Europe/Paris</td>
 *     </tr>
 *     <tr>
 *       <td>Offset</td>
 *       <td>Etc/GMT-8. The offset is positive if the local time zone is behind UTC and negative if it is ahead. 
 *           The offset range is between Etc/GMT-14 and Etc/GMT+12 (UTC-12 and UTC+14). Which means that Etc/GMT-8 
 *           is equivalent to UTC+08.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.isoStrFormat - specifies in which format the ISO string is returned. 
 * The possible values of isoStrFormat are: "offset", "zulu", "local", "auto". 
 * The default format is auto.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>offset</td>
 *       <td>time zone offset from UTC.</td>
 *       <td>2016-01-05T11:30:00-08:00</td>
 *     </tr>
 *     <tr>
 *       <td>zulu</td>
 *       <td>zulu time or UTC time.</td>
 *       <td>2016-01-05T19:30:00Z</td>
 *     </tr>
 *     <tr>
 *       <td>local</td>
 *       <td>local time, does not contain time zone offset.</td>
 *       <td>2016-01-05T19:30:00</td>
 *     </tr>
 *     <tr>
 *       <td>auto</td>
 *       <td>auto means the returned ISO string depends on the input string and options</td>
 *       <td></td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.dst - The dst option can be used for time only values in conjunction with offset. 
 * Setting dst to true indicates the time is in DST. By default the time is interpreted as standard time. 
 * The possible values of dst are: "true" or "false". Default is "false".
 * <p style='padding-left: 5px;'>
 * Due to Daylight Saving Time, there is a possibility that a time exists twice If the time falls in the duplicate window 
 * (switching from daylight saving time to standard time). The application can disambiguate the time in the overlapping 
 * period using the dst option. Setting dst to true indicates the time is in DST. By default the time is interpreted as 
 * standard time.<br/><br/>
 * Example: On November 1st, 2105 in US the time between 1 and 2 AM will be repeated. The dst option can indicate the 
 * distinction as follows. Initially the time is in DST, so dst:'true' is specified.<br/>
 * var options = {formatType:'datetime', dateFormat:'short', timeFormat:'medium', timeZone:'America/Los_Angeles', isoStrFormat: 'offset', dst : true};<br/>
 * var localeElements = oj.getLocaleElemnts();<br/>
 * var str= "11/1/15 1:59:59 AM";<br/>
 * cnv.parse(str, localeElements, options);-->2015-11-01T01:59:59-07:00
 * <br/><br/>
 * If the user does not pass the dst option, the time will be interpreted as standard time. 
 * var options = {formatType:'datetime', dateFormat:'short', timeFormat:'medium', timeZone:'America/Los_Angeles'};<br/>
 * var localeElements = oj.getLocaleElemnts();<br/>
 * var str= "11/1/15 1:59:59 AM";<br/>
 * cnv.parse(str, localeElements, options);-->2015-11-01T01:59:59-08:00
 * <br/><br/>
 * At 2AM, DST will be over and the clock brings back to 1AM. Then the dst option should be false or not specified.
 * var options = {formatType:'datetime', dateFormat:'short', timeFormat:'medium', timeZone:'America/Los_Angeles', isoStrFormat: 'offset'};<br/>
 * var localeElements = oj.getLocaleElemnts();<br/>
 * var str= "11/1/15 1:00:00 AM";<br/>
 * cnv.parse(str, localeElements, options);-->2015-11-01T01:00:00-08:00
 * </p>
 *
 * @property {boolean=} options.hour12 - specifies what time notation is used for formatting the time. 
 * A true value uses the 12-hour clock and false uses the 24-hour clock (often called military time 
 * in the US). This property is undefined if the hour property is not used when formatting the date.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>true</td>
 *       <td>T13:10  is formatted as 1:10 PM</td>
 *     </tr>
 *     <tr>
 *       <td>false</td>
 *       <td>T13:10 is formatted as 13:10</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 * 
 * @property {string=} options.pattern - a localized string pattern, where the the characters used in 
 * pattern conform to Unicode CLDR for date time formats. This will override all other options 
 * when present. <br/>
 * NOTE: 'pattern' is provided for backwards compatibility with existing apps that may want the 
 * convenience of specifying an explicit format mask. Setting a 'pattern' will override the default 
 * locale specific format.
 * NOTE: The supported tokens for timezone are of 'Z', 'VV', and 'X'.<br/><br/>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Letter</th>
 *       <th>Date or Time Component</th>
 *       <th>Examples</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>G, GG, GGG</td>
 *       <td>Era designator</td>
 *       <td>AD</td>
 *     </tr>
 *     <tr>
 *       <td>y</td>
 *       <td>numeric representation of year</td>
 *       <td>1, 2014</td>
 *     </tr>
 *     <tr>
 *       <td>yy</td>
 *       <td>2-digit representation of year</td>
 *       <td>01, 14</td>
 *     </tr>
 *     <tr>
 *       <td>yyyy</td>
 *       <td>4-digit representation of year</td>
 *       <td>0001, 2014</td>
 *     </tr>
 *     <tr>
 *       <td>M</td>
 *       <td>Numeric representation of month in year: (1-12)</td>
 *       <td>1, 12</td>
 *     </tr>
 *     <tr>
 *       <td>MM</td>
 *       <td>2-digit representation of month in year: (01-12)</td>
 *       <td>01, 12</td>
 *     </tr>
 *     <tr>
 *       <td>MMM</td>
 *       <td>Formatted  name of the month, abbreviated</td>
 *       <td>Jan</td>
 *     </tr>
 *     <tr>
 *       <td>MMMM</td>
 *       <td>Formatted  name of the month, wide</td>
 *       <td>January</td>
 *     </tr>
 *     <tr>
 *       <td>MMMMM</td>
 *       <td>Formatted  name of the month, narrow</td>
 *       <td>J</td>
 *     </tr>
 *     <tr>
 *       <td>LLL</td>
 *       <td>Stand-alone name of the month, abbreviated</td>
 *       <td>Jan</td>
 *     </tr>
 *     <tr>
 *       <td>LLLL</td>
 *       <td>Stand-alone name of the month, wide</td>
 *       <td>January</td>
 *     </tr>
 *     <tr>
 *       <td>LLLLL</td>
 *       <td>Stand-alone name of the month, narrow</td>
 *       <td>J</td>
 *     </tr>
 *     <tr>
 *       <td>d</td>
 *       <td>Numeric representation of  day in month (1-31)</td>
 *       <td>1, 21</td>
 *     </tr>
 *     <tr>
 *       <td>dd</td>
 *       <td>2-digit representation of  day in month (01-31)</td>
 *       <td>01, 21</td>
 *     </tr>
 *     <tr>
 *       <td>E, EE, EEE</td>
 *       <td>Formatted name of day in week, abbreviated</td>
 *       <td>Tue</td>
 *     </tr>
 *     <tr>
 *       <td>EEEE</td>
 *       <td>Formatted name of day in week, wide</td>
 *       <td>Tuesday</td>
 *     </tr>
 *     <tr>
 *       <td>EEEEE</td>
 *       <td>Formatted name of day in week, narrow</td>
 *       <td>T</td>
 *     </tr>
 *     <tr>
 *       <td>c, cc, ccc</td>
 *       <td>Stand-alone name of day in week, abbreviated</td>
 *       <td>Tue</td>
 *     </tr>
 *     <tr>
 *       <td>cccc</td>
 *       <td>Stand-alone name of day in week, wide</td>
 *       <td>Tuesday</td>
 *     </tr>
 *     <tr>
 *       <td>ccccc</td>
 *       <td>Stand-alone name of day in week, narrow</td>
 *       <td>T</td>
 *     </tr>
 *     <tr>
 *       <td>a</td>
 *       <td>am/pm marker</td>
 *       <td>PM</td>
 *     </tr>
 *     <tr>
 *       <td>H</td>
 *       <td>Numeric hour in day (0-23)</td>
 *       <td>1, 23</td>
 *     </tr>
 *     <tr>
 *       <td>HH</td>
 *       <td>2-digit hour in day (00-23)</td>
 *       <td>01, 23</td>
 *     </tr>
 *     <tr>
 *       <td>h</td>
 *       <td>Numeric  hour in am/pm (1-12)</td>
 *       <td>1, 12</td>
 *     </tr>
 *     <tr>
 *       <td>hh</td>
 *       <td>2-digit hour in day (01-12)</td>
 *       <td>01, 12</td>
 *     </tr>
 *     <tr>
 *       <td>k</td>
 *       <td>Numeric  hour in day (1-24)</td>
 *       <td>1, 24</td>
 *     </tr>
 *     <tr>
 *       <td>kk</td>
 *       <td>2-digit hour in day (1-24)</td>
 *       <td>01, 24</td>
 *     </tr>
 *     <tr>
 *       <td>K</td>
 *       <td>Numeric  hour in am/pm (0-11)</td>
 *       <td>1, 11</td>
 *     </tr>
 *     <tr>
 *       <td>KK</td>
 *       <td>2-digit hour in am/pm (0-11)</td>
 *       <td>01, 11</td>
 *     </tr>
 *     <tr>
 *       <td>m, mm</td>
 *       <td>2-digit  minute in hour (00-59)</td>
 *       <td>05, 59</td>
 *     </tr>
 *     <tr>
 *       <td>s, ss</td>
 *       <td>2-digit second in minute (00-59)</td>
 *       <td>01, 59</td>
 *     </tr>
 *     <tr>
 *       <td>S</td>
 *       <td>Numeric  Millisecond (0-999)</td>
 *       <td>1, 999</td>
 *     </tr>
 *     <tr>
 *       <td>SS</td>
 *       <td>2-digit Millisecond (00-999)</td>
 *       <td>01, 999</td>
 *     </tr>
 *     <tr>
 *       <td>SSS</td>
 *       <td>3-digit Millisecond (000-999)</td>
 *       <td>001, 999</td>
 *     </tr>
 *     <tr>
 *       <td>z, zz, zzz</td>
 *       <td>Abbreviated time zone name</td>
 *       <td>PDT, PST</td>
 *     </tr>
 *     <tr>
 *       <td>zzzz</td>
 *       <td>Full time zone name</td>
 *       <td>Pacific Standard Time, Pacific Daylight Time</td>
 *     </tr>
 *     <tr>
 *       <td>Z, ZZ, ZZZ</td>
 *       <td>Sign hours minutes</td>
 *       <td>-0800</td>
 *     </tr>
 *     <tr>
 *       <td>X</td>
 *       <td>Sign hours</td>
 *       <td>-08</td>
 *     </tr>
 *     <tr>
 *       <td>XX</td>
 *       <td>Sign hours minutes</td>
 *       <td>-0800</td>
 *     </tr>
 *     <tr>
 *       <td>XXX</td>
 *       <td>Sign hours:minutes</td>
 *       <td>-08:00</td>
 *     </tr>
 *     <tr>
 *       <td>VV</td>
 *       <td>Time zone ID</td>
 *       <td>Americs/Los_Angeles</td>
 *     </tr>
 *   </tbody>
 * </table>
 * 
 * @property {string=} options.formatType - determines the 'standard' date and/or time format lengths 
 * to use. Allowed values: "date", "time", "datetime". See 'dateFormat' and 'timeFormat' options. 
 * When set a value must be specified.
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>datetime</td>
 *       <td>date and time portions are displayed.</td>
 *       <td>September 20, 2015 12:04 PM September 20, 2015 12:05:35 PM Pacific Daylight Time</td>
 *     </tr>
 *     <tr>
 *       <td>date</td>
 *       <td>date portion only is displayed.</td>
 *       <td>September 20, 2015</td>
 *     </tr>
 *     <tr>
 *       <td>time</td>
 *       <td>time portion only is displayed.</td>
 *       <td>12:05:35</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.dateFormat - specifies the standard date format length to use when 
 * formatType is set to "date" or "datetime". Allowed values are : "short" (default), "medium", "long", 
 * "full". 
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>short</td>
 *       <td>9/20/15</td>
 *     </tr>
 *     <tr>
 *       <td>medium</td>
 *       <td>Sep 20, 2015</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>September 20, 2015</td>
 *     </tr>
 *     <tr>
 *       <td>full</td>
 *       <td>Sunday, September 20, 2015</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * @property {string=} options.timeFormat - specifies the standard time format length to use when 
 * 'formatType' is set to "time" or "datetime". Allowed values: "short" (default), "medium", "long", 
 * "full". 
 * <p style='padding-left: 5px;'>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Option</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>short</td>
 *       <td>12:11 PM</td>
 *     </tr>
 *     <tr>
 *       <td>medium</td>
 *       <td>12:11:23 PM</td>
 *     </tr>
 *     <tr>
 *       <td>long</td>
 *       <td>12:12:19 PM PDT</td>
 *     </tr>
 *     <tr>
 *       <td>full</td>
 *       <td>12:12:37 PM Pacific Daylight Time</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 * 
 *  @property {string=} options.lenientParse - The lenientParse property can be used to enable or disable leninet parsing.
 *  Allowed values: "full" (default), "none". 
 * <p style='padding-left: 5px;'>
 * By default the lenient parse is enabled and the leniency rules descibed above will be used. When lenientParse is
 * set to "none" the lenient parse is disabled and the user input must match the expected input otherwise an exception will
 * be thrown.<br/><br/>
 * 
 * </p>
 * 
 * @example <caption>Create a date time converter using no options. This uses the default value for 
 * year, month, day properties</caption>
 * var converter = OraDateTimeConverter.getInstance();
 * var localeElements;
 * var resolved = converter.resolvedOptions(localeElements);
 * // logs "day=numeric, month=numeric, year=numeric"
 * console.log("day=" + resolved.day + ", month=" + resolved.month + ", year=" + resolved.year);
 * <br/>
 * 
 * @example <caption>Use a converter using the ECMA options to represent date</caption>
 * var options = { year:'2-digit', month: '2-digit', day: '2-digit'};
 * var localeElements;
 * var converter = OraDateTimeConverter.getInstance();
 * var date = "2016-04-17";
 * var str = converter.format(date, localeElements, options);<br/>
 * 
 * 
 * @example <caption>Converter using the 'pattern' option</caption>
 * var options = {pattern: 'MM-dd-yyyy'}; 
 * var localeElements;
 * var converter = OraDateTimeConverter.getInstance();
 * var date = "2016-04-17";
 * var str = converter.format(date, localeElements, options);<br/>
 * 
 * @example <caption>Converter using predefined style</caption>
 * var options = {formatType: 'date', dateFormat: 'medium'}; 
 * var localeElements;
 * var converter = OraDateTimeConverter.getInstance();
 * var date = "2016-04-17";
 * var str = converter.format(date, localeElements, options);<br/>
 *
 * @example <caption>Create a date time converter using specific pattern with IANA timezone ID with 
 * isoStrFormat of offset.</caption>
 * var options = {pattern: 'MM/dd/yy hh:mm:ss a Z', timeZone: 'America/Los_Angeles', isoStrFormat: 'offset'};
 * var localeElements;
 * var converter = OraDateTimeConverter.getInstance();
 * var date = "2016-04-17T13:05:30";
 * var str = converter.format(date, localeElements, options);<br/>
 *
 * @example <caption>Create a date time converter using specific pattern with Etc/GMT timezone ID with 
 * isoStrFormat of zulu.</caption>
 * var options = {pattern: 'MM/dd/yy hh:mm:ss a Z', timeZone: 'Etc/GMT-08:00', isoStrFormat: 'zulu'};  
 * var localeElements;
 * var converter = OraDateTimeConverter.getInstance();
 * var str = "01/05/16 01:01:01 AM +0800";
 * var obj = converter.parse(str, localeElements, options);<br/>
 * 
 * @example <caption>Disable lenient parse.</caption>
 * var options = {pattern: 'MM/dd/yy', lenientParse:'none'};  
 * var localeElements;
 * var converter = OraDateTimeConverter.getInstance();
 * var str = "14/05/16";
 * var obj = converter.parse(str, localeElements, options); --> RangeError: 13 is out of range.  Enter a value between 1 and 12 for month.<br/>
 * @name OraDateTimeConverter
 * @ignore
 */

/**
 * @ignore
 */
var OraDateTimeConverter;

OraDateTimeConverter = (function () {
  var instance;
  var _appendPreOrPostMatch;
  var _expandFormat;
  var _expandISOFormat;
  var _parseExact;
  var _formatImpl;
  var _parseImpl;
  var _formatRelativeImpl;
  var _throwInvalidDateFormat;
  var _getResolvedOptionsFromPattern;
  var _dateTimeStyle;
  var _isoStrDateTimeStyle;
  var _get2DigitYearStart;
  var _isHour12;
  var _dateTimeStyleFromPattern;
  var _expandPredefinedStylesFormat;
  var _localeDataCache = {};
  var _timeZoneDataCache = {};
  var _isLeapYear;
  var _getDaysInMonth;
  var _toAvailableFormatsKeys;
  var _expandAvailableDateFormatsPattern;
  var _expandAvailableTimeFormatsPattern;
  var _basicFormatMatcher;
  var _appendToKey;
  var _getDaysDif;
  var _getDayIndex;
  var _isSameYear;
  var _isNextYear;
  var _isPrevYear;
  var _isSameMonth;
  var _isNextMonth;
  var _isPrevMonth;
  var _isSameWeek;
  var _isNextWeek;
  var _isPrevWeek;
  var _isSameDay;
  var _isNextDay;
  var _isPrevDay;
  var _expandYear;
  var _getTokenIndex;
  var _parseLenient;
  var _parseLenientyMEd;
  var _parseLenientyMMMEd;
  var _parseLenienthms;
  var _getDayIndex1;
  var _getResolvedDefaultOptions;
  var _resolvedOptionsImpl;
  var _getMonthIndex;
  var _getPredefinedStylesResolvedOptions;
  var _getResolvedOptions;
  var _getParseRegExp;
  var _getPatternResolvedOptions;
  var _getECMAResolvedOptions;
  var _getRelativeTimeResolvedOptions;
  var _validateRange;
  var _getCompareISODatesOptions;
  var _arrayIndexOfDay;
  var _getLocaleDecimalSeparator;
  var _arrayIndexOfMonth;
  var _getIsoStrStyle;
  var _throwDateFormatMismatch;
  var _getPatternFromSingleToken;
  var _throwWeekdayMismatch;
  var _createParseISOStringFromDate;
  var _getParseISOStringOffset;
  var _createISOStrParts;
  var _adjustHours;
  var _getStdOffset;
  var _availableTimeZonesImpl;
  var _getTimePart;
  var _getNameIndex;
  var _getWeekdayName;
  var _getMetazone;
  var _parseMetaDate;
  var _getTimezoneName;
  var _getTimeZone;
  var _parseZone;
  var _convertToLocalDate;
  var _getTimeDiff;
  var _replaceParams;
  var _formatRelativeCalendar;
  var _formatRelativeDisplayName;
  var _formatRelativeImplicit;
  var _getBCP47Lang;
  var _getUnits;
  var _daysToMonths;

  var _YEAR_AND_DATE_REGEXP = /(\d{1,4})\D+?(\d{1,4})/g;
  var _YMD_REGEXP = /(\d{1,4})\D+?(\d{1,4})\D+?(\d{1,4})/g;
  var _MONTH_REGEXP_FMT = /^[M][^M]|[^M]M[^M]|[^M]M$|^M$/g;
  var _MONTH_REGEXP_STD = /^[L][^L]|[^L]L[^L]|[^L]L$|^L$/g;
  var _DAY_REGEXP = /^[d][^d]|[^d]d[^d]|[^d]d$|^d$|^d[^d]/;
  var _HOUR_REGEXP = /(?:^|[^h])h[^h]|[^H]H[^H]|[^k]k[^k]|[^K]K[^K]|^H[^H]|^H$|[^h]h$/;
  var _COMMENT_REGEXP = /'.*'/;
  var _DAY_COMMENT_REGEXP = /'[^d]*d[^d]*'/;
  var _TIME_REGEXP = /(\d{1,2})(?:\D+?(\d{1,2}))?(?:\D+?(\d{1,2}))?(?:\D+?(\d{1,3}))?/g;
  var _TIME_FORMAT_REGEXP = /h|H|K|k/g;
  var _ESCAPE_REGEXP = /([\^\$\.\*\+\?\|\[\]\(\)\{\}])/g;
  var _TOKEN_REGEXP = /ccccc|cccc|ccc|cc|c|EEEEE|EEEE|EEE|EE|E|dd|d|MMMMM|MMMM|MMM|MM|M|LLLLL|LLLL|LLL|LL|L|yyyy|yy|y|hh|h|HH|H|KK|K|kk|k|mm|m|ss|s|aa|a|SSS|SS|S|zzzz|zzz|zz|z|v|ZZZ|ZZ|Z|XXX|XX|X|VV|GGGGG|GGGG|GGG|GG|G/g;
  var _TIME_FORMATS_Z_TOKENS = /\s?(?:\(|\[)?z{1,4}(?:\)|\])?/g;
  var tzNotSupported = false;
  var _DAYS_INDEXES = {
    0: "sun",
    1: "mon",
    2: "tue",
    3: "wed",
    4: "thu",
    5: "fri",
    6: "sat"
  };

  var _THRESHOLDS = {
    s: 46, // seconds to minute
    m: 46, // minutes to hour
    h: 23, // hours to day
    d: 7, // days to week
    w: 4, //weeks to 
    M: 12   // months to year
  };

  var _ZULU = 'zulu';
  var _LOCAL = 'local';
  var _AUTO = 'auto';
  var _INVARIANT = 'invariant';
  var _OFFSET = 'offset';
  var _UTC = 'UTC';

  var _ALNUM_REGEXP = '(\\D+|\\d\\d?\\D|\\d\\d?|\\D+\\d\\d?)',
      _NON_DIGIT_REGEXP = '(\\D+|\\D+\\d\\d?)',
      _NON_DIGIT_OPT_REGEXP = '(\\D*)',
      _STR_REGEXP = '(.+?)',
      _TWO_DIGITS_REGEXP = '(\\d\\d?)',
      _THREE_DIGITS_REGEXP = '(\\d{1,3})',
      _FOUR_DIGITS_REGEXP = '(\\d{1,4})',
      _TZ_HM_REGEXP = '([+-]?\\d{1,4})',
      _TZ_H_REGEXP = '([+-]?\\d{1,2})',
      _TZ_H_SEP_M_REGEXP = '([+-]?\\d{1,2}:\\d{1,2})',
      _SLASH_REGEXP = '(\\/)';

  var _PROPERTIES_MAP = {
    'MMM': {
      'token': 'months',
      'style': 'format',
      'mLen': 'abbreviated',
      'matchIndex': 0,
      'key': 'month',
      'value': 'short',
      'regExp': _ALNUM_REGEXP
    },
    'MMMM': {
      'token': 'months',
      'style': 'format',
      'mLen': 'wide',
      'matchIndex': 0,
      'key': 'month',
      'value': 'long',
      'regExp': _ALNUM_REGEXP
    },
    'MMMMM': {
      'token': 'months',
      'style': 'format',
      'mLen': 'narrow',
      'matchIndex': 0,
      'key': 'month',
      'value': 'narrow',
      'regExp': _ALNUM_REGEXP
    },
    'LLL': {
      'token': 'months',
      'style': 'stand-alone',
      'mLen': 'abbreviated',
      'matchIndex': 1,
      'key': 'month',
      'value': 'short',
      'regExp': _ALNUM_REGEXP
    },
    'LLLL': {
      'token': 'months',
      'style': 'stand-alone',
      'mLen': 'wide',
      'matchIndex': 1,
      'key': 'month',
      'value': 'long',
      'regExp': _ALNUM_REGEXP
    },
    'LLLLL': {
      'token': 'months',
      'style': 'stand-alone',
      'mLen': 'narrow',
      'matchIndex': 1,
      'key': 'month',
      'value': 'narrow',
      'regExp': _ALNUM_REGEXP
    },
    'E': {
      'token': 'days',
      'style': 'format',
      'dLen': 'abbreviated',
      'matchIndex': 0,
      'key': 'weekday',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'EE': {
      'token': 'days',
      'style': 'format',
      'dLen': 'abbreviated',
      'matchIndex': 0,
      'key': 'weekday',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'EEE': {
      'token': 'days',
      'style': 'format',
      'dLen': 'abbreviated',
      'matchIndex': 0,
      'key': 'weekday',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'EEEE': {
      'token': 'days',
      'style': 'format',
      'dLen': 'wide',
      'matchIndex': 0,
      'key': 'weekday',
      'value': 'long',
      'regExp': _NON_DIGIT_REGEXP
    },
    'EEEEE': {
      'token': 'days',
      'style': 'format',
      'dLen': 'narrow',
      'matchIndex': 0,
      'key': 'weekday',
      'value': 'narrow',
      'regExp': _NON_DIGIT_REGEXP
    },
    'c': {
      'token': 'days',
      'style': 'stand-alone',
      'dLen': 'abbreviated',
      'matchIndex': 1,
      'key': 'weekday',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'cc': {
      'token': 'days',
      'style': 'stand-alone',
      'dLen': 'abbreviated',
      'matchIndex': 1,
      'key': 'weekday',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'ccc': {
      'token': 'days',
      'style': 'stand-alone',
      'dLen': 'abbreviated',
      'matchIndex': 1,
      'key': 'weekday',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'cccc': {
      'token': 'days',
      'style': 'stand-alone',
      'dLen': 'wide',
      'matchIndex': 1,
      'key': 'weekday',
      'value': 'long',
      'regExp': _NON_DIGIT_REGEXP
    },
    'ccccc': {
      'token': 'days',
      'style': 'stand-alone',
      'dLen': 'narrow',
      'matchIndex': 1,
      'key': 'weekday',
      'value': 'narrow',
      'regExp': _NON_DIGIT_REGEXP
    },
    'h': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 11,
      'start2': 1,
      'end2': 12,
      'key': 'hour',
      'value': 'numeric',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'hh': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 11,
      'start2': 1,
      'end2': 12,
      'key': 'hour',
      'value': '2-digit',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'K': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 11,
      'start2': 0,
      'end2': 11,
      'key': 'hour',
      'value': 'numeric',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'KK': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 11,
      'start2': 0,
      'end2': 11,
      'key': 'hour',
      'value': '2-digit',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'H': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 23,
      'start2': 0,
      'end2': 23,
      'key': 'hour',
      'value': 'numeric',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'HH': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 23,
      'start2': 0,
      'end2': 23,
      'key': 'hour',
      'value': '2-digit',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'k': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 23,
      'start2': 1,
      'end2': 24,
      'key': 'hour',
      'value': 'numeric',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'kk': {
      'token': 'time',
      'timePart': 'hour',
      'start1': 0,
      'end1': 23,
      'start2': 1,
      'end2': 24,
      'key': 'hour',
      'value': '2-digit',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'm': {
      'token': 'time',
      'timePart': 'minute',
      'start1': 0,
      'end1': 59,
      'start2': 0,
      'end2': 59,
      'key': 'minute',
      'value': 'numeric',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'mm': {
      'token': 'time',
      'timePart': 'minute',
      'start1': 0,
      'end1': 59,
      'start2': 0,
      'end2': 59,
      'key': 'minute',
      'value': '2-digit',
      'regExp': _TWO_DIGITS_REGEXP
    },
    's': {
      'token': 'time',
      'timePart': 'second',
      'start1': 0,
      'end1': 59,
      'start2': 0,
      'end2': 59,
      'key': 'second',
      'value': 'numeric',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'ss': {
      'token': 'time',
      'timePart': 'second',
      'start1': 0,
      'end1': 59,
      'start2': 0,
      'end2': 59,
      'key': 'second',
      'value': '2-digit',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'S': {
      'token': 'time',
      'timePart': 'millisec',
      'start1': 0,
      'end1': 999,
      'start2': 0,
      'end2': 999,
      'key': 'millisecond',
      'value': 'numeric',
      'regExp': _THREE_DIGITS_REGEXP
    },
    'SS': {
      'token': 'time',
      'timePart': 'millisec',
      'start1': 0,
      'end1': 999,
      'start2': 0,
      'end2': 999,
      'key': 'millisecond',
      'value': 'numeric',
      'regExp': _THREE_DIGITS_REGEXP
    },
    'SSS': {
      'token': 'time',
      'timePart': 'millisec',
      'start1': 0,
      'end1': 999,
      'start2': 0,
      'end2': 999,
      'key': 'millisecond',
      'value': 'numeric',
      'regExp': _THREE_DIGITS_REGEXP
    },
    'd': {
      'token': 'dayOfMonth',
      'key': 'day',
      'value': 'numeric',
      'getPartIdx': 2,
      'regExp': _TWO_DIGITS_REGEXP
    },
    'dd': {
      'token': 'dayOfMonth',
      'key': 'day',
      'value': '2-digit',
      'getPartIdx': 2,
      'regExp': _TWO_DIGITS_REGEXP
    },
    'M': {
      'token': 'monthIndex',
      'key': 'month',
      'value': 'numeric',
      'getPartIdx': 1,
      'regExp': _TWO_DIGITS_REGEXP
    },
    'MM': {
      'token': 'monthIndex',
      'key': 'month',
      'value': '2-digit',
      'getPartIdx': 1,
      'regExp': _TWO_DIGITS_REGEXP
    },
    'L': {
      'token': 'monthIndex',
      'key': 'month',
      'value': 'numeric',
      'getPartIdx': 1,
      'regExp': _TWO_DIGITS_REGEXP
    },
    'LL': {
      'token': 'monthIndex',
      'key': 'month',
      'value': '2-digit',
      'getPartIdx': 1,
      'regExp': _TWO_DIGITS_REGEXP
    },
    'y': {
      'token': 'year',
      'key': 'year',
      'value': 'numeric',
      'regExp': _FOUR_DIGITS_REGEXP
    },
    'yy': {
      'token': 'year',
      'key': 'year',
      'value': '2-digit',
      'regExp': _TWO_DIGITS_REGEXP
    },
    'yyyy': {
      'token': 'year',
      'key': 'year',
      'value': 'numeric',
      'regExp': _FOUR_DIGITS_REGEXP
    },
    'a': {
      'token': 'ampm',
      'key': 'hour12',
      'value': true,
      'regExp': _NON_DIGIT_OPT_REGEXP
    },
    'z': {
      'token': 'tzAbbrev',
      'regExp': _STR_REGEXP
    },
    'v': {
      'token': 'tzAbbrev',
      'key': 'timeZoneName',
      'value': 'short',
      'regExp': _STR_REGEXP
    },
    'zz': {
      'token': 'tzAbbrev',
      'regExp': _STR_REGEXP
    },
    'zzz': {
      'token': 'tzAbbrev',
      'regExp': _STR_REGEXP
    },
    'zzzz': {
      'token': 'tzFull',
      'regExp': _STR_REGEXP
    },
    'Z': {
      'token': 'tzhm',
      'regExp': _TZ_HM_REGEXP
    },
    'ZZ': {
      'token': 'tzhm',
      'regExp': _TZ_HM_REGEXP
    },
    'ZZZ': {
      'token': 'tzhm',
      'regExp': _TZ_HM_REGEXP
    },
    'X': {
      'token': 'tzh',
      'regExp': _TZ_H_REGEXP
    },
    'XX': {
      'token': 'tzhm',
      'key': 'XX',
      'regExp': _TZ_HM_REGEXP
    },
    'XXX': {
      'token': 'tzhsepm',
      'regExp': _TZ_H_SEP_M_REGEXP
    },
    'VV': {
      'token': 'tzid',
      'regExp': _STR_REGEXP
    },
    'G': {
      'token': 'era',
      'key': 'era',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'GG': {
      'token': 'era',
      'key': 'era',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'GGG': {
      'token': 'era',
      'key': 'era',
      'value': 'short',
      'regExp': _NON_DIGIT_REGEXP
    },
    'GGGG': {
      'token': 'era',
      'key': 'era',
      'value': 'long',
      'regExp': _NON_DIGIT_REGEXP
    },
    'GGGGG': {
      'token': 'era',
      'key': 'era',
      'value': 'narrow',
      'regExp': _NON_DIGIT_REGEXP
    },
    '/': {
      'token': 'slash',
      'regExp': _SLASH_REGEXP
    }
  };


  /*
   *Helper functions
   */
  _get2DigitYearStart = function (options) {
    var option = options['two-digit-year-start'];
    if (option === undefined || isNaN(option))
      option = 1950;
    option = parseInt(option, 10);
    return option;
  };

  //Each locale has 12 or 24 hour preferred format
  _isHour12 = function (localeElements) {
    var locale = localeElements['_ojLocale_']
    var territory = oj.OraI18nUtils.getBCP47Region(locale);
    var prefferedHours = localeElements['supplemental']['prefferedHours'];
    var hour12 = prefferedHours[territory];
    return hour12 === 'h';
  };

  _isLeapYear = function (y) {
    if (y % 400 === 0)
      return true;
    else if (y % 100 === 0)
      return false;
    else if (y % 4 === 0)
      return true;
    else
      return false;
  };

  _getDaysInMonth = function (y, m) {
    switch (m) {
      case 0 :
      case 2 :
      case 4 :
      case 6 :
      case 7 :
      case 9 :
      case 11 :
        return 31;
      case 1:
        if (_isLeapYear(y))
          return 29;
        else
          return 28;
      default:
        return 30;
    }
  };

  //returns the locale's pattern from the predefined styles.
  //EX: for en-US dateFormat:"full" returns the pattern "EEEE, MMMM d, y".
  _expandPredefinedStylesFormat = function (options, localeElements, caller) {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var getOption = oj.OraI18nUtils.getGetOption(options, caller);
    var fmtType = getOption('formatType', 'string', ['date', 'time',
      'datetime'], 'date');
    var dStyle = getOption('dateFormat', 'string',
        ['short', 'medium', 'long', 'full'], 'short');
    var tStyle = getOption('timeFormat', 'string',
        ['short', 'medium', 'long', 'full'], 'short');
    var cal = mainNode['dates']['calendars']['gregorian'];
    var dateFormats = cal['dateFormats'];
    var timeFormats = cal['timeFormats'];

    var dStyleFmt, tStyleFmt, format;
    switch (dStyle)
    {
      case "full" :
        dStyleFmt = dateFormats['full'];
        break;
      case "long" :
        dStyleFmt = dateFormats['long'];
        break;
      case "medium" :
        dStyleFmt = dateFormats['medium'];
        break;
      case "short" :
        dStyleFmt = dateFormats['short'];
        break;
      default:
        break;
    }
    switch (tStyle)
    {
      case "full" :
        tStyleFmt = timeFormats['full'];
        break;
      case "long" :
        tStyleFmt = timeFormats['long'];
        break;
      case "medium" :
        tStyleFmt = timeFormats['medium'];
        break;
      case "short" :
        tStyleFmt = timeFormats['short'];
        break;
      default:
        break;
    }
    if (dStyleFmt !== undefined && (fmtType === "datetime" ||
        fmtType === "date"))
      format = dStyleFmt;
    if (tStyleFmt !== undefined && (fmtType === "datetime" ||
        fmtType === "time")) {
      if (tzNotSupported) {
        tStyleFmt = tStyleFmt.replace(_TIME_FORMATS_Z_TOKENS, '');

      }
      if (format)
        format = format + " " + tStyleFmt;
      else
        format = tStyleFmt;
    }
    return format;
  };

  // appends pre- and post- token match strings while removing escaped 
  // characters.
  // Returns a single quote count which is used to determine if the 
  // token occurs
  // in a string literal.
  _appendPreOrPostMatch = function (preMatch, strings) {
    var quoteCount = 0;
    var escaped = false;
    for (var i = 0, il = preMatch.length; i < il; i++) {
      var c = preMatch.charAt(i);
      switch (c) {
        case "\'":
          if (escaped) {
            strings.push("\'");
          }
          else {
            quoteCount++;
          }
          escaped = false;
          break;
        case "\\":
          if (escaped) {
            strings.push("\\");
          }
          escaped = !escaped;
          break;
        default:
          strings.push(c);
          escaped = false;
          break;
      }
    }
    return quoteCount;
  };

  //Throw an exception if date-time pattern is invalid
  _throwInvalidDateFormat = function (format, options, m) {
    var msg;
    var error;
    var errorInfo;
    var samplePattern;
    var isDate = options['year'] !== undefined || options['month'] !==
        undefined ||
        options['weekday'] !== undefined || options['day'] !== undefined;
    var isTime = options['hour'] !== undefined || options['minute'] !==
        undefined ||
        options['second'] !== undefined;
    if (isDate && isTime) {
      samplePattern = "MM/dd/yy hh:mm:ss a";
    }
    else if (isDate) {
      samplePattern = "MM/dd/yy";
    }
    else {
      samplePattern = "hh:mm:ss a";
    }
    msg = "Unexpected character(s) " + m + " encountered in the pattern \"" +
        format + " An example of a valid pattern is \"" + samplePattern + '".';
    error = new SyntaxError(msg);
    errorInfo = {
      'errorCode': 'optionValueInvalid',
      'parameterMap': {
        'propertyName': 'pattern',
        'propertyValue': format,
        'propertyValueHint ': samplePattern
      }
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };

  //implementation of basic forat matcher algorithm from ECMA spec.
  //This impelmentation takes into consideration hour12, For example if 
  //hour12=true, H entries are removed from availableFormats in order to
  //avoid wrong matching for hour.
  _basicFormatMatcher = function (dateTimeKeys, localeElements, isDate,
      hour12)
  {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var availableFormats = mainNode['dates']['calendars']['gregorian']
    ['dateTimeFormats']['availableFormats'];
    var dateTimeFormats = ["weekday", "era", "year", "month", "day",
      "hour", "minute", "second", "timeZoneName"];
    var values = {
      '2-digit': 0,
      'numeric': 1,
      'narrow': 2,
      'short': 3,
      'long': 4
    };

    var removalPenalty = 120;
    var additionPenalty = 20;
    var longLessPenalty = 8;
    var longMorePenalty = 6;
    var shortLessPenalty = 6;
    var shortMorePenalty = 3;
    var bestScore = -Infinity;
    var bestFormat = undefined;
    var match;
    var skip;
    for (var f in availableFormats) {
      skip = false;
      var format = {};
      format['pattern'] = availableFormats[f];
      var score = 0;
      while ((match = _TOKEN_REGEXP.exec(f)) !== null) {
        var m = match[0];
        if ((m === 'h' || m === 'hh') && !hour12) {
          skip = true;
          continue;
        }
        else if ((m === 'H' || m === 'HH') && hour12) {
          skip = true;
          continue;
        }
        if (_PROPERTIES_MAP[m] !== undefined) {
          format[_PROPERTIES_MAP[m]['key']] =
              _PROPERTIES_MAP[m]['value'];
        }
      }
      if (skip)
        continue;
      for (var property in dateTimeFormats) {
        var optionsProp = dateTimeKeys[dateTimeFormats[property]];
        var formatProp = format[dateTimeFormats[property]];
        if (optionsProp === undefined && formatProp !== undefined) {
          score -= additionPenalty;
        }
        else if (optionsProp !== undefined && formatProp === undefined) {
          score -= removalPenalty;
        }
        else if (optionsProp !== undefined && formatProp !== undefined) {

          var optionsPropIndex = values[optionsProp];
          var formatPropIndex = values[formatProp];
          var delta = Math.max(Math.min(formatPropIndex -
              optionsPropIndex, 2), -2);
          if (delta === 2) {
            score -= longMorePenalty;
          }
          else if (delta === 1) {
            score -= shortMorePenalty;
          }
          else if (delta === -1) {
            score -= shortLessPenalty;
          }
          else if (delta === -2) {
            score -= longLessPenalty;
          }
        }
      }
      if (score > bestScore) {
        bestScore = score;
        bestFormat = format;
      }
    }
    if (bestFormat !== undefined) {
      return bestFormat['pattern'];
    }
    return null;
  };

  //Return a format key from ecma options. For example:
  //{year:"2-digit", month:"long", day:"numeric", weekday:"long"};
  //will return "yyMMMMdEEEE"
  _toAvailableFormatsKeys = function (options, localeElements, caller) {
    var dateKey = '';
    var timeKey = '';
    var dateOptions = {};
    var timeOptions = {};

    var getOption = oj.OraI18nUtils.getGetOption(options, caller);

    //date key
    var option = getOption('era', 'string', ['narrow', 'short', 'long']);
    dateKey += _appendToKey(dateOptions, 'era',
        option, {
          'narrow': 'GGGGG',
          'short': 'GGG',
          'long': 'GGGG'
        });

    option = getOption('year', 'string', ['2-digit', 'numeric']);
    dateKey += _appendToKey(dateOptions, 'year',
        option, {
          '2-digit': 'yy',
          'numeric': 'y'
        });

    option = getOption('month', 'string', ['2-digit', 'numeric', 'narrow',
      'short', 'long']);
    dateKey += _appendToKey(dateOptions, 'month',
        option, {
          '2-digit': 'MM',
          'numeric': 'M',
          'narrow': 'MMMMM',
          'short': 'MMM',
          'long': 'MMMM'
        });

    option = getOption('weekday', 'string', ['narrow', 'short', 'long']);
    dateKey += _appendToKey(dateOptions, 'weekday',
        option, {
          'narrow': 'EEEEE',
          'short': 'E',
          'long': 'EEEE'
        });

    option = getOption('day', 'string', ['2-digit', 'numeric']);
    dateKey += _appendToKey(dateOptions, 'day',
        option, {
          '2-digit': 'dd',
          'numeric': 'd'
        });

    //time key
    var hr12 = getOption('hour12', 'boolean', [true, false]);
    if (hr12 === undefined)
      hr12 = _isHour12(localeElements);
    option = getOption('hour', 'string', ['2-digit', 'numeric']);
    if (hr12 === true) {
      timeKey += _appendToKey(timeOptions, 'hour',
          option, {
            '2-digit': 'hh',
            'numeric': 'h'
          });
    }
    else {
      timeKey += _appendToKey(timeOptions, 'hour',
          option, {
            '2-digit': 'HH',
            'numeric': 'H'
          });
    }

    option = getOption('minute', 'string', ['2-digit', 'numeric']);
    timeKey += _appendToKey(timeOptions, 'minute',
        option, {
          '2-digit': 'mm',
          'numeric': 'm'
        });

    option = getOption('second', 'string', ['2-digit', 'numeric']);
    timeKey += _appendToKey(timeOptions, 'second',
        option, {
          '2-digit': 'ss',
          'numeric': 's'
        });
    option = getOption('timeZoneName', 'string', ['short', 'long']);
    timeKey += _appendToKey(timeOptions, 'timeZoneName',
        option, {
          'short': 'v',
          'long': 'v'
        });

    return [dateKey, timeKey, dateOptions, timeOptions];
  };

  _appendToKey = function (obj, prop, option, pairs) {
    if (option !== undefined) {
      obj[prop] = option;
      return pairs[option];
    }
    else {
      return '';
    }
  };


  /*
   *This function is used by the munger algorith. It expands a pattern
   *in order to match the user's ECMA options.
   *For example if the user provided options is 
   *options = {year: 'numeric', month: 'long', weekday: 'long',
   *day : '2-digit'}; 
   *The key for these options is: yMMMMEEEEdd. under availableFormats, 
   *we have the following entry: "yMMMEd": "E, MMM d, y".
   *The _basicFormatMatcher algorithm will find "yMMMEd" key as the 
   *closest match to yMMMMEEEEdd. If formatMatcher ="munger", we will 
   *expand the corresponding pattern to match the options. 
   *So "E, MMM d, y" will be expanded to "EEEE, MMMM dd y".
   */
  _expandAvailableDateFormatsPattern = function (formatTemplate, options,
      caller)
  {
    var datePat = formatTemplate;
    var match;
    var getOption = oj.OraI18nUtils.getGetOption(options, caller);
    //year
    var option = getOption('year', 'string', ['2-digit', 'numeric']);
    var pairs = {
      '2-digit': 'yy',
      'numeric': 'yyyy'
    };
    if (option !== undefined)
      datePat = datePat.replace(/y{1,4}/, pairs[option]);

    //month
    option = getOption('month', 'string', ['2-digit', 'numeric', 'narrow',
      'short', 'long']);
    if (option !== undefined) {
      pairs = {
        '2-digit': 'MM',
        'numeric': 'M',
        'narrow': 'MMMMM',
        'short': 'MMM',
        'long': 'MMMM'
      };
      var pairsL = {
        '2-digit': 'LL',
        'numeric': 'L',
        'narrow': 'LLLLL',
        'short': 'LLL',
        'long': 'LLLL'
      };
      if (pairs[option] !== undefined && pairs[option].length > 2) {
        datePat = datePat.replace(/M{3,5}/, pairs[option]);
        datePat = datePat.replace(/L{3,5}/, pairsL[option]);
      }
      else if (option === '2-digit') {
        _MONTH_REGEXP_FMT.lastIndex = 0;
        match = _MONTH_REGEXP_FMT.test(formatTemplate);
        if (match) {
          datePat = datePat.replace('M', 'MM');
        }
        match = _MONTH_REGEXP_STD.test(formatTemplate);
        if (match) {
          datePat = datePat.replace('L', 'LL');
        }
      }
    }

    //weekday
    option = getOption('weekday', 'string', ['narrow', 'short', 'long']);
    if (option !== undefined) {
      var pairsFormat = {
        'narrow': 'EEEEE',
        'short': 'EEE',
        'long': 'EEEE'
      };
      var pairsStandalone = {
        'narrow': 'ccccc',
        'short': 'ccc',
        'long': 'cccc'
      };
      datePat = datePat.replace(/E{1,5}/, pairsFormat[option]);
      datePat = datePat.replace(/c{1,5}/, pairsStandalone[option]);
    }
    //day
    option = getOption('day', 'string', ['2-digit', 'numeric']);
    if (option !== undefined) {
      if (option === '2-digit') {
        _DAY_REGEXP.lastIndex = 0;
        _DAY_COMMENT_REGEXP.lastIndex = 0;
        var match1 = _DAY_COMMENT_REGEXP.test(formatTemplate);
        match = _DAY_REGEXP.test(formatTemplate);
        if (match == true && match1 === false) {
          datePat = datePat.replace('d', 'dd');
        }
      }
    }
    return datePat;
  };

  //Same as above for time entries
  _expandAvailableTimeFormatsPattern = function (formatTemplate, options,
      caller) {
    var timePat = formatTemplate;
    var getOption = oj.OraI18nUtils.getGetOption(options, caller);
    var option = getOption('hour', 'string', ['2-digit', 'numeric']);
    if (option === '2-digit') {
      _HOUR_REGEXP.lastIndex = 0;
      _COMMENT_REGEXP.lastIndex = 0;
      var formatTemplate1 = formatTemplate.replace(_COMMENT_REGEXP, '');
      var match = _HOUR_REGEXP.exec(formatTemplate1);
      if (match !== null) {
        _TIME_FORMAT_REGEXP.lastIndex = 0;
        var match1 = _TIME_FORMAT_REGEXP.exec(match[0]);
        var ext = match1[0] + match1[0];
        timePat = formatTemplate.replace(match1[0], ext);
      }
    }
    return timePat;
  };

  _getPatternFromSingleToken = function (token, tokenObj, availableFormats) {
    var i;
    var count = 0;
    for (i in tokenObj) {
      count++;
    }
    if (count > 1) {
      return null;
    }
    var pattern;
    for (i = token.length; i > 0; i--) {
      pattern = availableFormats[token.substr(0, i)];
      if (pattern !== undefined)
        return pattern;
    }
    return token;
  };

  _expandISOFormat = function (options) {
    var format = 'yyyy-MM-ddTHH:mm:ss';
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraDateTimeConverter.format");
    var isoFormat = getOption('isoStrFormat', 'string',
        [_ZULU, _OFFSET, _INVARIANT, _LOCAL, _AUTO], _AUTO);
    switch (isoFormat) {
      case _ZULU:
        options['timeZone'] = 'Etc/GMT+0';
        format += "'Z'";
        break;
      case _OFFSET:
        format += 'XXX';
        break;
      default:
        break;
    }
    return format;
  };

  //Returns the localized decimal separator. Used for milliseconds
  _getLocaleDecimalSeparator = function (localeElements, locale) {
    var numberingSystemKey = oj.OraI18nUtils.getNumberingSystemKey(localeElements, locale);
    var numberingSystem = "symbols-numberSystem-" + numberingSystemKey;
    return localeElements['numbers'][numberingSystem]['decimal'];
  }

  //Returns a pattern corresponding to user's options.
  //Cache the entries for which we already found a pattern 
  _expandFormat = function (options, localeElements, mlocale, caller) {
    function substituteTokens(cachedVal) {
      if (msOption !== undefined) {
        var sep = _getLocaleDecimalSeparator(mainNode, mlocale);
        if (hOption === undefined && mOption === undefined && sOption === undefined) {
          sep = "S";
        }
        else {
          sep = "ss" + sep + "SSS";
        }
        cachedVal = cachedVal.replace("ss", sep);
        if (sOption === undefined) {
          options['second'] = undefined;
        }
      }
      //substitute time zone token v  based on short or long
      var option = getOption('timeZoneName', 'string', ['short', 'long']);
      var pairs = {
        'short': 'z',
        'long': 'zzzz'
      };
      if (option !== undefined) {
        cachedVal = cachedVal.replace(/v/, pairs[option]);
      }
      return cachedVal;
    }

    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var locale = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
    var getOption = oj.OraI18nUtils.getGetOption(options, caller);
    var pattern;
    var matcher = getOption('formatMatcher', 'string',
        ['basic', 'munger'], 'munger');
    var count = 0;
    for (count in options) {
      count++;
    }
    if (count === 0) {
      options = {
        'year': 'numeric',
        'month': 'numeric',
        'day': 'numeric'
      };
    }
    var hOption;
    var mOption;
    var sOption;
    var msOption;
    //append millisecnods to pattern
    msOption = getOption('millisecond', 'string', ['numeric', '2-digit']);
    sOption = getOption('second', 'string', ['numeric', '2-digit']);
    if (msOption !== undefined) {
      hOption = getOption('hour', 'string', ['numeric', '2-digit']);
      mOption = getOption('minute', 'string', ['numeric', '2-digit']);
      options['second'] = '2-digit';
    }

    var dateTimeKeys = _toAvailableFormatsKeys(options, localeElements, caller);
    //First try to get the pattern from cache
    if (_localeDataCache[locale] !== undefined) {
      var cachedPattern = _localeDataCache[locale]['dates']['calendars']['gregorian']
      ['dateTimeFormats'][dateTimeKeys[0] + dateTimeKeys[1]];
      if (cachedPattern !== undefined) {
        cachedPattern = substituteTokens(cachedPattern);
        return cachedPattern;
      }
    }
    if (dateTimeKeys[0] === '' && dateTimeKeys[1] === '') {
      return _expandPredefinedStylesFormat(options, localeElements,
          caller);
    }
    var availableFormats = mainNode['dates']['calendars']['gregorian']
    ['dateTimeFormats']['availableFormats'];
    var datePattern = availableFormats[dateTimeKeys[0]];
    var hour12 = getOption('hour12', 'boolean', [true, false]);
    if (hour12 === undefined)
      hour12 = _isHour12(localeElements);
    if (datePattern === undefined && dateTimeKeys[0] !== '') {
      datePattern = _getPatternFromSingleToken(dateTimeKeys[0], dateTimeKeys[2],
          availableFormats);
      if (datePattern === null)
        datePattern = _basicFormatMatcher(dateTimeKeys[2],
            localeElements, true, hour12);
      if (datePattern !== null) {
        if (matcher !== 'basic') {
          datePattern = _expandAvailableDateFormatsPattern(
              datePattern, options, caller);
        }
      }
      else
        datePattern = dateTimeKeys[0];
    }
    var timePattern = availableFormats[dateTimeKeys[1]];
    if (timePattern === undefined && dateTimeKeys[1] !== '') {
      timePattern = _getPatternFromSingleToken(dateTimeKeys[1], dateTimeKeys[3],
          availableFormats);
      if (timePattern === null)
        timePattern = _basicFormatMatcher(dateTimeKeys[3],
            localeElements, true, hour12);
      if (timePattern !== null) {
        if (matcher !== 'basic') {
          timePattern = _expandAvailableTimeFormatsPattern(
              timePattern, options, caller);
        }
      }
      else
        timePattern = dateTimeKeys[1];
    }
    pattern = datePattern || '';
    if (timePattern !== undefined)
    {
      if (pattern !== '')
        pattern += ' ' + timePattern;
      else
        pattern = timePattern;
    }

    //cache the pattern
    if (_localeDataCache[locale] === undefined) {
      _localeDataCache[locale] = {};
      _localeDataCache[locale]['dates'] = {};
      _localeDataCache[locale]['dates']['calendars'] = {};
      _localeDataCache[locale]['dates']['calendars']['gregorian'] = {};
      _localeDataCache[locale]['dates']['calendars']['gregorian']
      ['dateTimeFormats'] = {};
    }
    _localeDataCache[locale]['dates']['calendars']['gregorian']
    ['dateTimeFormats'][dateTimeKeys[0] + dateTimeKeys[1]] = pattern;
    pattern = substituteTokens(pattern);
    return pattern;
  };

  _parseMetaDate = function (str) {
    var parts = str.split(' ');
    var dParts = parts[0].split('-');
    var d = new Date(dParts[0], dParts[1] - 1, dParts[2]);
    if (parts.length > 1) {
      dParts = parts[1].split(':');
      d.setHours(dParts[0]);
      d.setMinutes(dParts[1]);
    }
    return d.getTime();
  };

  _getMetazone = function (value, zoneName, metazones) {
    var now = new Date(value[0], value[1] - 1, value[2],
        value[3], value[4], value[5]);
    now = now.getTime();
    var parts = zoneName.split('/');
    var country = parts[0];
    var city = parts[1];
    var zone = metazones[country];
    if (zone === undefined)
      return null;
    zone = zone[city];
    if (zone === undefined) {
      return null;
    }
    var i, length = zone.length;
    var mzoneStartTime;
    var mzoneEndTime;
    var mzoneStart;
    var mzoneEnd;
    var mzoneName;

    for (i = 0; i < length; i++) {
      mzoneStart = zone[i]['usesMetazone']['_from'];
      mzoneEnd = zone[i]['usesMetazone']['_to'];
      mzoneName = zone[i]['usesMetazone']['_mzone'];
      //time zone never chnaged. _to and _from undefined
      if (mzoneStart === undefined && mzoneEnd === undefined) {
        return mzoneName;
      }
      //_from undefined. check if  now <= _to
      if (mzoneStart === undefined && mzoneEnd !== undefined) {
        mzoneEndTime = _parseMetaDate(mzoneEnd);
        if (now <= mzoneEndTime)
          return mzoneName;
      }
      //_to undefined. check if  now >= _from
      if (mzoneStart !== undefined && mzoneEnd === undefined) {
        mzoneStartTime = _parseMetaDate(mzoneStart);
        if (now >= mzoneStartTime)
          return mzoneName;
      }
      // find interval where now falls between _to and _from
      if (mzoneStart !== undefined && mzoneEnd !== undefined) {
        mzoneStartTime = _parseMetaDate(mzoneStart);
        mzoneEndTime = _parseMetaDate(mzoneEnd);
        if (now >= mzoneStartTime && now < mzoneEndTime) {
          return mzoneName;
        }
      }
    }
  };

  //value is the  utc date. Used to determinse if dst or not
  //options, param passed to the converter
  //len = 0, return short timezone name; 1, return long timezone
  //isTimeOnly is used to determine of we should igonore dst
  _getTimezoneName = function (localeElements, value, options, len, isTimeOnly) {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var zoneName = "";
    var timeZone = options['timeZone'];
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraDateTimeConverter");
    var dst = getOption('dst', 'boolean', [true, false], false);
    if (timeZone === undefined) {
      return zoneName;
    }
    var metazones = localeElements['supplemental']['metazones'];
    var metaZone = _getMetazone(value, timeZone, metazones);
    var zoneNameEntry0;
    var tzLong = 'long';
    var tzShort = 'short';
    var during = 'standard';
    var ignoreDst = isTimeOnly ? false : true;
    var zone = _getTimeZone(timeZone, localeElements);
    var index = _parseZone(zone, value, dst, ignoreDst, true);
    if (mainNode['dates']['timeZoneNames']['metazone'] !== undefined) {
      zoneNameEntry0 = mainNode['dates']['timeZoneNames']['metazone'][metaZone];
    }
    if (zoneNameEntry0 === undefined) {
      var offset = zone.ofset(index);
      return oj.OraI18nUtils.getTimeStringFromOffset(_UTC, offset, true, true);
    }
    var zoneNameEntry;
    if (len === 1) {
      zoneNameEntry = zoneNameEntry0[tzLong];
    }
    else {
      zoneNameEntry = zoneNameEntry0[tzShort];
    }
    var offset1 = zone.ofset(index);
    var offset2 = zone.ofset(index + 1);
    if (offset1 < offset2) {
      during = 'daylight';
    }
    if (zoneNameEntry !== undefined) {
      zoneName = zoneNameEntry[during];
      if (zoneName !== undefined)
        return zoneName;
    }
    //return UTC offset if we can not find a timezone name.
    return oj.OraI18nUtils.getTimeStringFromOffset(_UTC, offset1, true, true);
  };

  _getTimeZone = function (timeZoneId, localeElements) {
    var tz = OraTimeZone.getInstance();
    var zone = tz.getZone(timeZoneId, localeElements);
    return zone;
  };

  _parseZone = function (zone, parts, dst, ignoreDst, dateTime) {
    var utcDate = Date.UTC(parts[0], parts[1] - 1, parts[2],
        parts[3], parts[4], parts[5]);
    var index = zone.parse(utcDate, dst, ignoreDst, dateTime);
    return index;
  };

  _formatImpl = function (localeElements, options, isoStrInfo, locale) {
    var ret;
    var format;
    var value = isoStrInfo['isoStrParts'];
    var isTimeOnly = isoStrInfo['dateTime'].indexOf('T') === 0;
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraDateTimeConverter.format");
    //get the pattern from options
    format = options['pattern'] || _expandFormat(options, localeElements,
        locale, "OraDateTimeConverter.format");
    // Start with an empty string
    ret = [];
    var part,
        quoteCount = 0,
        cal = mainNode['dates']['calendars']['gregorian'];

    function _getPart(date, part) {
      switch (part) {
        case 0:
          return date[0];
        case 1:
          return date[1];
        case 2:
          return date[2];
        case 3:
          var dt = new Date(date[0], date[1] - 1, date[2],
              date[3], date[4], date[5]);
          return _DAYS_INDEXES[dt.getDay()];
      }
    }

    function _getPaddedPart(ret, value, idx, len) {
      var val = _getPart(value, idx);
      ret.push(len > 1 ? oj.OraI18nUtils.padZeros(val, len) : val);
    }

    function _getTimeParts(ret, value, len, currentPart, current) {
      var val;
      switch (currentPart['timePart']) {
        case 'hour' :
          if (currentPart['end1'] === 11)
            val = value[3] % 12;
          else
            val = value[3];
          if (current === 'h' || current === 'hh') {
            if (val === 0)
              val = 12;
          }
          else if (current === 'k' || current === 'kk') {
            if (val === 0)
              val = 24;
          }
          break;
        case 'minute' :
          val = value[4];
          break;
        case 'second' :
          if (options['second'] === undefined && options['millisecond'] !== undefined)
            val = 0;
          else
            val = value[5];
          break;
        case 'millisec' :
          //val = oj.OraI18nUtils.zeroPad("" + value[6], 3, true);          
          val = value[6];
          break;
      }
      ret.push(len > 1 ? oj.OraI18nUtils.padZeros(val, len) : val);
    }

    function _getTimezoneOffset(value, options, isTimeOnly) {
      var offset;
      var zone;
      var index;
      var timeZone = options['timeZone'];
      var dst = getOption('dst', 'boolean', [true, false], false);
      var ignoreDst = isTimeOnly ? false : true;
      if (timeZone !== undefined) {
        zone = _getTimeZone(timeZone, localeElements);
        index = _parseZone(zone, value, dst, ignoreDst, true);
        offset = -zone.ofset(index);
      }
      else if (isoStrInfo['format'] !== _LOCAL) {
        switch (isoStrInfo['format']) {
          case _OFFSET :
            offset = isoStrInfo['timeZone'].split(":");
            var hoursOffset = parseInt(offset[0], 10);
            var minOffset = parseInt(offset[1], 10);
            offset = (hoursOffset * 60) +
                (oj.OraI18nUtils.startsWith(isoStrInfo['timeZone'], "-") ? -minOffset : minOffset);
            break;
          case _ZULU :
            offset = 0;
            break;
          default:
            offset = 0;
            break;
        }
      }
      return offset;
    }

    //adjust the offset based on the offset of input iso str and timeZone attribute
    //EX: if the input iso str is one of the following:
    //2014-06-01T16:00:00-07:00
    //2014-06-01T16:00:00Z
    var timeZone = options['timeZone'];
    if (isoStrInfo['format'] !== _LOCAL && timeZone !== undefined) {
      _adjustHours(isoStrInfo, options, localeElements);
    }

    for (; ; ) {
      // Save the current index
      var index = _TOKEN_REGEXP.lastIndex,
          // Look for the next pattern
          ar = _TOKEN_REGEXP.exec(format);

      // Append the text before the pattern (or the end of the string if 
      // not found)
      var preMatch = format.slice(index, ar ? ar.index : format.length);
      quoteCount += _appendPreOrPostMatch(preMatch, ret);

      if (!ar) {
        break;
      }

      // do not replace any matches that occur inside a string literal.
      if (quoteCount % 2) {
        ret.push(ar[0]);
        continue;
      }

      var current = ar[ 0 ],
          clength = current.length,
          currentPart = _PROPERTIES_MAP[current];
      switch (currentPart['token']) {
        case 'days':
          part = cal[currentPart['token']][currentPart['style']][currentPart['dLen']];
          ret.push(part[_getPart(value, 3)]);
          break;
        case 'months':
          part = cal[currentPart['token']][currentPart['style']][currentPart['mLen']];
          ret.push(part[_getPart(value, 1)]);
          break;
        case 'dayOfMonth' :
        case 'monthIndex' :
          ret.push(_getPaddedPart(ret, value, currentPart['getPartIdx'], clength));
          break;
        case 'year':
          // Year represented by four full digits
          part = value[0];
          if (clength === 2) {
            part = part % 100;
          }
          ret.push(oj.OraI18nUtils.padZeros(part, clength));
          break;
        case 'time':
          _getTimeParts(ret, value, clength, currentPart, current);
          break;
        case "ampm":
          // Multicharacter am/pm indicator
          part = value[3] < 12 ?
              cal['dayPeriods']['format']['wide']['am'] :
              cal['dayPeriods']['format']['wide']['pm'];
          ret.push(part);
          break;
        case "tzhm":
          // Time zone hours minutes: -0800 
          part = _getTimezoneOffset(value, options, isTimeOnly);
          if (part === 0) {
            if (currentPart['key'] === 'XX')
              ret.push('Z');
            else
              ret.push("+0000");
          }
          else {
            ret.push(
                (part <= 0 ? "-" : "+") +
                oj.OraI18nUtils.padZeros(Math.floor(Math.abs(part / 60)), 2) +
                oj.OraI18nUtils.padZeros(Math.floor(Math.abs(part % 60)), 2)
                );
          }
          break;
        case "tzhsepm":
          // Time zone hours minutes: -08:00 
          part = _getTimezoneOffset(value, options, isTimeOnly);
          if (part === 0) {
            ret.push("Z");
          }
          else {
            ret.push(
                (part <= 0 ? "-" : "+") +
                oj.OraI18nUtils.padZeros(Math.floor(Math.abs(part / 60)), 2) +
                ":" +
                oj.OraI18nUtils.padZeros(Math.floor(Math.abs(part % 60)), 2)
                );
          }
          break;
        case "tzh":
          // Time zone hours: -08 
          part = _getTimezoneOffset(value, options, isTimeOnly);
          if (part === 0) {
            ret.push("Z");
          }
          else {
            ret.push(
                (part <= 0 ? "-" : "+") +
                oj.OraI18nUtils.padZeros(Math.floor(Math.abs(part / 60)), 2));
          }
          break;
        case "tzid":
          // Time zone ID: America/Los_Angeles 
          if (timeZone !== undefined)
            part = timeZone;
          else if (isoStrInfo['format'] === _ZULU) {
            part = _UTC;
          }
          else
            part = "";
          ret.push(part);
          break;
        case "tzAbbrev":
          // Time zone abbreviation: PDT, PST 
          part = _getTimezoneName(localeElements, value, options, 0, isTimeOnly);
          ret.push(part);
          break;
        case "tzFull":
          // Time zone Full: Pacific Standard Time 
          part = _getTimezoneName(localeElements, value, options, 1, isTimeOnly);
          ;
          ret.push(part);
          break;
        case "era":
          part = cal['eras']['eraAbbr'];
          ret.push(part['1']);
          break;
        case "slash":
          ret.push("/");
          break;
        default:
          _throwInvalidDateFormat(format, options, current);
      }
    }
    return ret.join("");
  };


  //_formatRelativeImpl

  //d1 and d2 same year
  _isSameYear = function (d1, d2) {
    return d1.getFullYear() === d2.getFullYear();
  };

  //d2 is next year
  _isNextYear = function (d1, d2) {
    return d2.getFullYear() - d1.getFullYear() === 1;
  };

  //d2 is previous year
  _isPrevYear = function (d1, d2) {
    return _isNextYear(d2, d1);
  };

  //d2 and d1 same month 
  _isSameMonth = function (d1, d2) {
    return _isSameYear(d1, d2) && (d1.getMonth() === d2.getMonth());
  };

  //d2 is next month
  _isNextMonth = function (d1, d2) {
    if (_isSameYear(d1, d2))
      return (d2.getMonth() - d1.getMonth()) === 1;
    else if (_isNextYear(d1, d2)) {
      return d1.getMonth() === 11 && (d2.getMonth() === 0);
    }
    return false;
  };

  //d2 is previous month
  _isPrevMonth = function (d1, d2) {
    return _isNextMonth(d2, d1);
  };

  //difference in days between d2 and d1. Only valid if d2 is same or 
  //next month of d1
  _getDaysDif = function (d1, d2) {
    var day1 = d1.getDate();
    var day2 = d2.getDate();
    if (_isNextMonth(d1, d2)) {
      day2 += _getDaysInMonth(d1.getFullYear, d1.getMonth());
    }
    return day2 - day1;
  };

  _getDayIndex = function (localeElements, idx) {
    var locale = localeElements['_ojLocale_'];
    var territory = oj.OraI18nUtils.getBCP47Region(locale);
    var firstDayNode = localeElements['supplemental']['weekData']['firstDay'];
    var firstDayOfweek = firstDayNode[territory];
    if (firstDayOfweek === undefined)
      firstDayOfweek = firstDayNode['001'];
    var ret = idx - firstDayOfweek;
    if (ret < 0)
      ret += 7;
    return ret;
  };

  //d1 and d2 same week
  _isSameWeek = function (localeElements, d1, d2) {
    if (d1 > d2) {
      //swap dates to make sure we work with positive numbers
      var tmp = d1;
      d1 = d2;
      d2 = tmp;
    }
    if ((!_isSameMonth(d1, d2)) && (!_isNextMonth(d1, d2)))
      return false;
    var dif = _getDaysDif(d1, d2) +
        _getDayIndex(localeElements, d1.getDay());
    return dif >= 0 && dif <= 6;
  };

  //d2 is next week
  _isNextWeek = function (localeElements, d1, d2) {
    if ((!_isSameMonth(d1, d2)) && (!_isNextMonth(d1, d2)))
      return false;
    var dif = _getDaysDif(d1, d2) +
        _getDayIndex(localeElements, d1.getDay());
    return dif >= 7 && dif <= 13;
  };

  //d2 is previous week
  _isPrevWeek = function (localeElements, d1, d2) {
    return _isNextWeek(localeElements, d2, d1);
  };

  //d1 and d2 same day
  _isSameDay = function (d1, d2) {
    return _isSameYear(d1, d2) && _isSameMonth(d1, d2) &&
        (d1.getDate() === d2.getDate());
  };

  //d2 is next day
  _isNextDay = function (d1, d2) {
    if ((!_isSameMonth(d1, d2)) && (!_isNextMonth(d1, d2))) {
      return false;
    }
    return (_getDaysDif(d1, d2) === 1);
  };

  //d2 is previous day
  _isPrevDay = function (d1, d2) {
    return _isNextDay(d2, d1);
  };

  _convertToLocalDate = function (d, localeElements, options) {
    var srcTimeZone = options['timeZone'];
    var isoInfo = oj.OraI18nUtils.getISOStrFormatInfo(d);
    var isoInfoFormat = isoInfo['format'];
    //first test id d is local
    if (isoInfoFormat === _LOCAL && srcTimeZone === undefined) {
      return d;
    }
    //first, convert to zulu
    var tzOptions = {isoStrFormat: 'zulu'};
    if (srcTimeZone !== undefined) {
      tzOptions['timeZone'] = srcTimeZone;
      tzOptions['dst'] = true;
    }
    var zuluDate = _parseImpl(d, localeElements, tzOptions, "en-US");
    //second, convert to local date
    var localOffset = oj.OraI18nUtils.getLocalTimeZoneOffset();
    tzOptions = {timeZone: localOffset, isoStrFormat: 'local'};
    var localDate = _parseImpl(zuluDate['value'], localeElements, tzOptions, "en-US");
    return localDate['value'];
  };

  _getTimeDiff = function (d1, d2, isCalendar) {
    var datetime1 = oj.OraI18nUtils._IsoStrParts(d1);
    var datetime2 = oj.OraI18nUtils._IsoStrParts(d2);

    if (isCalendar) {
      //for calendar, normalize the times to midnight so that the diff is the same
      // regardless of time of day.
      datetime1 = Date.UTC(datetime1[0], datetime1[1] - 1, datetime1[2], 0, 0, 0, 0);
      datetime2 = Date.UTC(datetime2[0], datetime2[1] - 1, datetime2[2], 0, 0, 0, 0);
    }
    else {
      datetime1 = Date.UTC(datetime1[0], datetime1[1] - 1, datetime1[2], datetime1[3], datetime1[4], datetime1[5], datetime1[6]);
      datetime2 = Date.UTC(datetime2[0], datetime2[1] - 1, datetime2[2], datetime2[3], datetime2[4], datetime2[5], datetime2[6]);
    }
    return  (datetime1 - datetime2);
  };

  _replaceParams = function (string, replacements) {
    return string.replace(/\{(\d+)\}/g, function () {
      return replacements[arguments[1]];
    });
  };

  _formatRelativeCalendar = function (now, relativeDate, localeElements, dateOnly) {
    var datePart;
    var timePart;
    var relativeDay;
    var dayIndex;
    var timePartOptions = {hour: 'numeric', minute: 'numeric'};
    var elseOptions = {year: 'numeric', month: 'numeric', day: 'numeric'};
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var pattern = mainNode['dates']['calendars']['gregorian']['dateTimeFormats']['long'];
    var days = mainNode['dates']['calendars']['gregorian']['days']['format']['wide'];
    var fields = mainNode['dates']['fields'];

    var value = oj.OraI18nUtils.isoToLocalDate(relativeDate);
    var localNow = oj.OraI18nUtils.isoToLocalDate(now);
    var isoStrInfo = oj.OraI18nUtils.getISOStrFormatInfo(relativeDate);

    if (_isSameDay(localNow, value))
      datePart = fields['day']['relative-type-0'];
    else if (_isNextDay(localNow, value))
      datePart = fields['day']['relative-type-1'];
    else if (_isPrevDay(localNow, value))
      datePart = fields['day']['relative-type--1'];
    else {
      relativeDay = value.getDay();
      dayIndex = _DAYS_INDEXES[relativeDay];
      var diff = _getTimeDiff(relativeDate, now, true);
      diff = diff / 864e5; //number of days
      if (diff < -1 && diff > -7) {
        datePart = fields[dayIndex]['relative-type--1'];
      }
      //next week
      else if (diff > 1 && diff < 7) {
        datePart = days[dayIndex];
      }
      //everything else
      else {
        return _formatImpl(localeElements, elseOptions, isoStrInfo, "en-US");
      }
    }
    if (dateOnly) {
      return datePart;
    }
    timePart = _formatImpl(localeElements, timePartOptions, isoStrInfo, "en-US");
    pattern = pattern.replace(/'/g, "");
    pattern = _replaceParams(pattern, [timePart, datePart]);
    return pattern;
  };

  _formatRelativeDisplayName = function (isoNow, isoValue, options, localeElements) {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var fields = mainNode['dates']['fields'];
    var getOption = oj.OraI18nUtils.getGetOption(options,
        "OraDateTimeConverter.formatRelative");
    var option = getOption('dateField', 'string',
        ['day', 'week', 'month', 'year', 'hour', 'minute', 'second']);
    var now = oj.OraI18nUtils.isoToLocalDate(isoNow);
    var value = oj.OraI18nUtils.isoToLocalDate(isoValue);
    switch (option) {
      case "day" :
        if (_isSameDay(now, value))
          return fields['day']['relative-type-0'];
        if (_isNextDay(now, value))
          return fields['day']['relative-type-1'];
        if (_isPrevDay(now, value))
          return fields['day']['relative-type--1'];
        return _formatRelativeImplicit(isoNow, isoValue, localeElements, 'day');
      case "week" :
        if (_isSameWeek(localeElements, now, value))
          return fields['week']['relative-type-0'];
        if (_isNextWeek(localeElements, now, value))
          return fields['week']['relative-type-1'];
        if (_isPrevWeek(localeElements, now, value))
          return fields['week']['relative-type--1'];
        return _formatRelativeImplicit(isoNow, isoValue, localeElements, 'week');
      case "month" :
        if (_isSameMonth(now, value))
          return fields['month']['relative-type-0'];
        if (_isNextMonth(now, value))
          return fields['month']['relative-type-1'];
        if (_isPrevMonth(now, value))
          return fields['month']['relative-type--1'];
        return _formatRelativeImplicit(isoNow, isoValue, localeElements, 'month');
      case "year" :
        if (_isSameYear(now, value))
          return fields['year']['relative-type-0'];
        if (_isNextYear(now, value))
          return fields['year']['relative-type-1'];
        if (_isPrevYear(now, value))
          return fields['year']['relative-type--1'];
        return _formatRelativeImplicit(isoNow, isoValue, localeElements, 'year');
      case "hour" :
        return _formatRelativeImplicit(isoNow, isoValue, localeElements, 'hour');
      case "minute" :
        return _formatRelativeImplicit(isoNow, isoValue, localeElements, 'minute');
      case "second" :
        return _formatRelativeImplicit(isoNow, isoValue, localeElements, 'second');
      default :
        break;
    }
    return null;
  };

  _daysToMonths = function (days) {
    // 400 years have 146097 days (taking into account leap year rules)
    return days * 4800 / 146097;
  };

  _getUnits = function (milliseconds) {
    var days = milliseconds / 864e5;
    var months = _daysToMonths(days);
    var years = months / 12;
    var obj = {
      'year': Math.round(years),
      'month': Math.round(months),
      'week': Math.round(milliseconds / 6048e5),
      'day': Math.round(milliseconds / 864e5),
      'hour': Math.round(milliseconds / 36e5),
      'minute': Math.round(milliseconds / 6e4),
      'second': Math.round(milliseconds / 1000),
      'millisecond': milliseconds
    };
    return obj;
  };

  //return the language part
  _getBCP47Lang = function (tag) {
    var arr = tag.split("-");
    return arr[0];
  };

  _formatRelativeImplicit = function (now, relativeDate, localeElements, field) {
    var future = "relativeTime-type-future";
    var past = "relativeTime-type-past";
    var nowNodeKey = "relative-type-0";
    var mainNodeKey = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var lang = _getBCP47Lang(mainNodeKey);
    var plurals = localeElements['supplemental']['plurals'];
    var fields = mainNode['dates']['fields'];
    var pluralKey = "relativeTimePattern-count-";

    var diff = _getTimeDiff(relativeDate, now, false);
    var absdiff = Math.abs(diff);
    var units = _getUnits(absdiff);
    if (field === null) {
      field = units['second'] < _THRESHOLDS.s && 'second' ||
          units['minute'] < _THRESHOLDS.m && 'minute' ||
          units['hour'] < _THRESHOLDS.h && 'hour' ||
          units['day'] < _THRESHOLDS.d && 'day' ||
          units['week'] < _THRESHOLDS.w && 'week' ||
          units['month'] < _THRESHOLDS.M && 'month' ||
          'year';
    }
    //when seconds <= 45 display 'now' instead of number of seconds
    if (field === 'second' && units['second'] < _THRESHOLDS.s) {
      return fields[field][nowNodeKey];
    }
    var plural = plurals[lang](units[field]);
    var pluralEntry = pluralKey + plural;
    var x = diff < 0 ? past : future;
    var format = fields[field][x][pluralEntry];
    //some locales only have other plural entry
    if (format === undefined) {
      pluralEntry = pluralKey + 'other';
      format = fields[field][x][pluralEntry];
    }
    format = _replaceParams(format, [units[field]]);
    return format;
  };

  _formatRelativeImpl = function (value, localeElements, options) {
    var now = oj.OraI18nUtils.dateToLocalIso(new Date());
    if (typeof value === "number") {
      value = oj.OraI18nUtils.dateToLocalIso(new Date(value));
    }
    else if (typeof value === "string") {
      if (oj.OraI18nUtils.trim(value) === '')
        return null;
    }
    else {
      return null;
    }
    if (options === undefined) {
      options = {'formatUsing': 'displayName'};
    }
    var getOption = oj.OraI18nUtils.getGetOption(options,
        "OraDateTimeConverter.formatRelative");
    var relativeTime = getOption('relativeTime', 'string',
        ['fromNow', 'toNow'], 'fromNow');
    var fieldOption = getOption('dateField', 'string',
        ['day', 'week', 'month', 'year', 'hour', 'minute', 'second']);

    value = _convertToLocalDate(value, localeElements, options);
    var toNow = (relativeTime === 'toNow');
    if (toNow) {
      var tmp = now;
      now = value;
      value = tmp;
    }
    var formatUsing = getOption('formatUsing', 'string',
        ['displayName', 'calendar'], 'displayName');
    if (formatUsing === 'calendar') {
      var dateOnly = getOption('dateOnly', 'boolean', [true, false], false);
      return _formatRelativeCalendar(now, value, localeElements, dateOnly);
    }
    if (fieldOption !== undefined)
      return _formatRelativeDisplayName(now, value, options, localeElements);
    return _formatRelativeImplicit(now, value, localeElements, null);
  };


  // parse functions

  _throwWeekdayMismatch = function (weekday, day) {
    var msg;
    var error;
    var errorInfo;
    msg = "The weekday " + weekday + " does not match the date " + day;
    error = new Error(msg);
    errorInfo = {
      'errorCode': 'dateToWeekdayMismatch',
      'parameterMap': {
        'weekday': weekday,
        'date': day
      }
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };

  _throwDateFormatMismatch = function (value, format, style) {
    var msg;
    var error;
    var errorInfo;
    var errorCodeType;
    if (style === 2) {
      msg = "The value \"" + value +
          "\" does not match the expected date-time format \"" + format + '"';
      errorCodeType = "datetimeFormatMismatch";
    }
    else if (style === 0) {
      msg = "The value \"" + value +
          "\" does not match the expected date format \"" + format + '"';
      errorCodeType = "dateFormatMismatch";

    }
    else {
      msg = "The value \"" + value +
          "\" does not match the expected time format \"" + format + '"';
      errorCodeType = "timeFormatMismatch";

    }
    error = new Error(msg);
    errorInfo = {
      'errorCode': errorCodeType,
      'parameterMap': {
        'value': value,
        'format': format
      }
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };

  _expandYear = function (start2DigitYear, year) {
    // expands 2-digit year into 4 digits.
    if (year < 100) {
      var ambiguousTwoDigitYear = start2DigitYear % 100;
      year += Math.floor((start2DigitYear / 100)) * 100 +
          (year < ambiguousTwoDigitYear ? 100 : 0);
    }
    return year;
  };

  _arrayIndexOfDay = function (daysObj, item) {
    var days = {
      "sun": 0,
      "mon": 1,
      "tue": 2,
      "wed": 3,
      "thu": 4,
      "fri": 5,
      "sat": 6
    };
    for (var d in daysObj) {
      if (oj.OraI18nUtils.trim(oj.OraI18nUtils.toUpper(daysObj[d])) === oj.OraI18nUtils.trim(item)) {
        return days[d];
      }
    }
    return -1;
  };

  _arrayIndexOfMonth = function (monthsObj, item) {
    for (var m in monthsObj) {
      if (oj.OraI18nUtils.trim(oj.OraI18nUtils.toUpper(monthsObj[m])) === oj.OraI18nUtils.trim(item)) {
        return (m - 1);
      }
    }
    return -1;
  };

  _getDayIndex1 = function (localeElements, value, fmt) {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var ret;
    var days;
    var calDaysFmt =
        mainNode['dates']['calendars']['gregorian']['days']['format'];
    var calDaysStdAlone =
        mainNode['dates']['calendars']['gregorian']['days']['stand-alone'];
    switch (fmt) {
      case 0:
        days = [
          calDaysFmt['abbreviated'],
          calDaysFmt['wide']
        ];
        break;
      case 1:
        days = [
          calDaysStdAlone['abbreviated'],
          calDaysStdAlone['wide']
        ];
        break;
      default:
        break;
    }
    value = oj.OraI18nUtils.toUpper(value);
    ret = _arrayIndexOfDay(days[0], value);
    if (ret === -1) {
      ret = _arrayIndexOfDay(days[1], value);
    }
    return ret;
  };

  //fmt:0 for format, 1 for stand-alone, 2 for lenient parse
  _getMonthIndex = function (localeElements, value, fmt) {
    var ret = -1;
    var months;
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var calMonthsFmt =
        mainNode['dates']['calendars']['gregorian']['months']['format'];
    var calMonthsStdAlone =
        mainNode['dates']['calendars']['gregorian']['months']['stand-alone'];
    switch (fmt) {
      case 0:
        months = [
          calMonthsFmt['wide'],
          calMonthsFmt['abbreviated']
        ];
        break;
      case 1:
        months = [
          calMonthsStdAlone['wide'],
          calMonthsStdAlone['abbreviated']
        ];
        break;
      case 2:
        months = [
          calMonthsFmt['wide'],
          calMonthsFmt['abbreviated'],
          calMonthsStdAlone['wide'],
          calMonthsStdAlone['abbreviated']
        ];
        break;
      default:
        return -1;
    }
    value = oj.OraI18nUtils.toUpper(value);
    for (var m in months) {
      ret = _arrayIndexOfMonth(months[m], value);
      if (ret !== -1) {
        return ret;
      }
    }
    return ret;
  };

  // converts a format string into a regular expression with groups that
  // can be used to extract date fields from a date string.
  // check for a cached parse regex.
  _getParseRegExp = function (format, options) {
    var re = {};


    // expand single digit formats, then escape regular expression
    //  characters.
    var expFormat = format.replace(_ESCAPE_REGEXP, "\\\\$1");
    var regexp = ["^"];
    var groups = [];
    var index = 0;
    var quoteCount = 0;
    var match;

    // iterate through each date token found.
    while ((match = _TOKEN_REGEXP.exec(expFormat)) !== null) {
      var preMatch = expFormat.slice(index, match.index);
      index = _TOKEN_REGEXP.lastIndex;

      // don't replace any matches that occur inside a string literal.
      quoteCount += _appendPreOrPostMatch(preMatch, regexp);
      if (quoteCount % 2) {
        regexp.push(match[0]);
        continue;
      }

      // add a regex group for the token.
      var m = match[ 0 ];
      var add;
      if (_PROPERTIES_MAP[m] !== undefined) {
        add = _PROPERTIES_MAP[m]['regExp'];
      }
      else
        _throwInvalidDateFormat(format, options, m);
      if (add) {
        regexp.push(add);
      }
      groups.push(match[0]);
    }
    _appendPreOrPostMatch(expFormat.slice(index), regexp);
    regexp.push("$");

    // allow whitespace to differ when matching formats.
    var regexpStr = regexp.join("").replace(/\s+/g, "\\s+");
    var parseRegExp = {
      'regExp': regexpStr,
      'groups': groups
    };
    // cache the regex for this format.
    return re[ format ] = parseRegExp;
  };

  _validateRange = function (name, value, low, high, displayValue,
      displayLow, displayHigh) {
    if (value < low || value > high) {
      var msg = displayValue +
          " is out of range.  Enter a value between " + displayLow +
          " and " + displayHigh + " for " + name;
      var rangeError = new RangeError(msg);
      var errorInfo = {
        'errorCode': "datetimeOutOfRange",
        'parameterMap': {
          'value': displayValue,
          'minValue': displayLow,
          'maxValue': displayHigh,
          'propertyName': name
        }
      };
      rangeError['errorInfo'] = errorInfo;
      throw rangeError;
    }
  };

  _getTokenIndex = function (arr, token) {
    for (var i = 0; i < arr.length; i++) {
      for (var n in arr[i])
      {
        if (n === token)
          return parseInt(i, 10);
      }
    }
    return 0;
  };

  //time lenient parse
  _parseLenienthms = function (result, timepart, format, localeElements, dtype) {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var calPM = mainNode['dates']['calendars']['gregorian']['dayPeriods']['format']['wide']['pm'];
    //hour, optional minutes and optional seconds
    _TIME_REGEXP.lastIndex = 0;
    var hour = 0, minute = 0, second = 0, msec = 0, idx;
    var match = _TIME_REGEXP.exec(timepart);
    if (match === null) {
      _throwDateFormatMismatch(timepart, format, dtype);
    }
    if (match[1] !== undefined)
      hour = parseInt(match[1], 10);
    if (match[2] !== undefined)
      minute = parseInt(match[2], 10);
    if (match[3] !== undefined)
      second = parseInt(match[3], 10);
    if (match[4] !== undefined)
      msec = parseInt(match[4], 10);

    _TIME_FORMAT_REGEXP.lastIndex = 0;
    match = _TIME_FORMAT_REGEXP.exec(format);
    switch (match[0]) {
      case "h":
        // Hour in am/pm (1-12)
        if (hour === 12)
          hour = 0;
        _validateRange("hour", hour, 0, 11, hour, 1, 12);
        idx = (oj.OraI18nUtils.toUpper(timepart)).indexOf(oj.OraI18nUtils.toUpper(calPM));
        if (idx !== -1 && hour < 12)
          hour += 12;
        break;
      case "K":
        // Hour in am/pm (0-11)
        _validateRange("hour", hour, 0, 11, hour, 0, 11);
        idx = (oj.OraI18nUtils.toUpper(timepart)).indexOf(oj.OraI18nUtils.toUpper(calPM));
        if (idx !== -1 && hour < 12)
          hour += 12;
        break;
      case "H":
        _validateRange("hour", hour, 0, 23, hour, 0, 23);
        break;
      case "k":
        if (hour === 24)
          hour = 0;
        _validateRange("hour", hour, 0, 23, hour, 1, 24);
        break;
      default:
        break;
    }
    // Minutes.
    _validateRange("minute", minute, 0, 59, minute, 0, 59);
    // Seconds.
    _validateRange("second", second, 0, 59, second, 0, 59);
    //millisec
    _validateRange("millisec", msec, 0, 999, msec, 0, 999);

    result.setHours(hour, minute, second, msec);
  };

  _getWeekdayName = function (value, localeElements) {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var calDaysFmt =
        mainNode['dates']['calendars']['gregorian']['days']['format'];
    var calDaysStandAlone =
        mainNode['dates']['calendars']['gregorian']['days']['stand-alone'];
    var days = [
      calDaysFmt['wide'],
      calDaysFmt['abbreviated'],
      calDaysStandAlone['wide'],
      calDaysStandAlone['abbreviated']
    ];

    var foundMatch = false;
    var dName, i, j;
    for (i in days) {
      for (j in days[i]) {
        dName = days[i][j];
        var dRegExp = new RegExp(dName + "\\b", 'i');
        if (dRegExp.test(value)) {
          foundMatch = true;
          break;
        }
      }
      if (foundMatch)
        break;
      dName = null;
    }
    return dName;
  };

  //lenient parse yMd and yMEd patterm. Must have year, moth, 
  //date all numbers. Ex: 5/3/2013
  //weekday is optional. If present it must match date. 
  //Ex:  Tuesday 11/19/2013
  //if year is 3-digits it can be anywhere in the string. 
  //Otherwise assume its position based on pattern
  //if date > 12 it can be anywhere in the string. 
  //Otherwise assume its position based on pattern 
  //separators can be any non digit characters
  _parseLenientyMEd = function (value, format, options, localeElements,
      isDateTime)
  {
    _YMD_REGEXP.lastIndex = 0;
    var match = _YMD_REGEXP.exec(value);
    if (match === null) {
      var dtype = isDateTime ? 2 : 0;
      _throwDateFormatMismatch(value, format, dtype);
    }
    var tokenIndexes = [{
        'y': format.indexOf("y")
      }, {
        'M': format.indexOf("M")
      }, {
        'd': format.indexOf("d")
      }];
    tokenIndexes.sort(function (a, b) {
      for (var n1 in a)
      {
        break;
      }
      for (var n2 in b)
      {
        break;
      }
      return a[n1] - b[n2];
    });
    var year, month, day, yearIndex, foundDayIndex, i, j;
    var dayIndex = _getTokenIndex(tokenIndexes, 'd');
    var foundYear = false, foundDay = false;
    for (i = 1; i <= 3; i++)
    {
      var tokenMatch = match[i];
      //find year if year is yyy|yyyy
      if (tokenMatch.length > 2 || tokenMatch > 31)
      {
        year = tokenMatch;
        foundYear = true;
        yearIndex = i - 1;
      }
    }
    if (!foundYear) {
      yearIndex = _getTokenIndex(tokenIndexes, 'y');
      year = match[_getTokenIndex(tokenIndexes, 'y') + 1];
    }
    //find day if day value > 12
    for (i = 0; i < 3; i++) {
      if (i !== yearIndex && match[i + 1] > 12) {
        day = match[i + 1];
        foundDay = true;
        foundDayIndex = i;
        break;
      }
    }
    if (!foundDay) {
      if (yearIndex === _getTokenIndex(tokenIndexes, 'd'))
      {
        day = match[_getTokenIndex(tokenIndexes, 'y') + 1];
        month = match[_getTokenIndex(tokenIndexes, 'M') + 1];
      }
      else if (yearIndex === _getTokenIndex(tokenIndexes, 'M'))
      {
        day = match[_getTokenIndex(tokenIndexes, 'd') + 1];
        month = match[_getTokenIndex(tokenIndexes, 'y') + 1];
      }
      else {
        day = match[_getTokenIndex(tokenIndexes, 'd') + 1];
        month = match[_getTokenIndex(tokenIndexes, 'M') + 1];

      }
    }
    else {
      for (i = 0; i < 3; i++) {
        if (i !== foundDayIndex && i !== yearIndex) {
          month = match[i + 1];
          break;
        }
      }
      if (month === undefined) {
        month = match[_getTokenIndex(tokenIndexes, 'M') + 1];
      }
    }
    month -= 1;
    var daysInMonth = _getDaysInMonth(year, month);
    //if both month and day > 12 and swapped, throw exception
    // based on original order
    if (foundDay && dayIndex !== foundDayIndex && month > 12) {
      _validateRange("month", day, 0, 11, day, 1, 12);
    }
    _validateRange("month", month, 0, 11, month + 1, 1, 12);
    _validateRange("day", day, 1, daysInMonth, day, 1, daysInMonth);
    var start2DigitYear = _get2DigitYearStart(options);
    year = _expandYear(start2DigitYear, parseInt(year, 10));
    _validateRange("year", year, 0, 9999, year, 0, 9999);
    var parsedDate = new Date(year, month, day);
    //locate weekday
    var dName = _getWeekdayName(value, localeElements);
    if (dName !== null) {
      var weekDay = _getDayIndex1(localeElements, dName, 0);
      // day of week does not match date
      if (parsedDate.getDay() !== weekDay) {
        _throwWeekdayMismatch(dName, parsedDate.getDate());
      }
    }
    var result =
        {
          'value': parsedDate,
          'offset': '',
          'warning': 'lenient parsing was used'
        };
    if (isDateTime) {
      var timepart = value.substr(_YMD_REGEXP.lastIndex);
      if (timepart.length === 0)
        result['value'].setHours(0, 0, 0, 0);
      else
        _parseLenienthms(result['value'], timepart, format, localeElements, 2);
    }
    result['value'] = oj.OraI18nUtils.dateToLocalIso(result['value']);
    return result;
  };

  //lenient parse yMMMd and yMMMEd patterns. Must have year, date as numbers 
  //and month name.
  //weekday is optional. If present it must match date.
  // Ex:  Monday Nov, 11 2013
  //weekday and month name can be anywhere in the string.
  //if year > 2-digits it can be anywhere in the string. 
  //Otherwise assume its position based on pattern
  //separators can be any non digit characters
  _parseLenientyMMMEd = function (value, format, options, localeElements,
      isDateTime) {
    var origValue = value;
    value = oj.OraI18nUtils.toUpper(value);
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    //locate month name
    var calMonthsFmt =
        mainNode['dates']['calendars']['gregorian']['months']['format'];
    var calMonthsStandAlone =
        mainNode['dates']['calendars']['gregorian']['months']['stand-alone'];
    var months = [
      calMonthsFmt['wide'],
      calMonthsFmt['abbreviated'],
      calMonthsStandAlone['wide'],
      calMonthsStandAlone['abbreviated']
    ];
    var foundMatch = false;
    var reverseMonth;
    var i, j, mName;
    for (i in months) {
      reverseMonth = [];
      for (j in months[i]) {
        mName = oj.OraI18nUtils.toUpper(months[i][j]);
        reverseMonth.push({
          'idx': j,
          'name': mName
        });
      }
      reverseMonth.sort(function (a, b) {
        return b['idx'] - a['idx'];
      });

      for (j in reverseMonth) {
        mName = reverseMonth[j]['name'];
        if (value.indexOf(mName) !== -1) {
          foundMatch = true;
          value = value.replace(mName, "");
          break;
        }
      }
      if (foundMatch)
        break;
    }
    //There is no month name. Try yMEd lenient parse.
    if (!foundMatch) {
      return _parseLenientyMEd(origValue, format, options, localeElements,
          isDateTime);
    }

    var month = _getMonthIndex(localeElements, mName, 2);
    _validateRange("month", month, 0, 11, month, 1, 12);

    //locate weekday
    var dName = _getWeekdayName(origValue, localeElements);
    var dRegExp = new RegExp(dName + "\\W", 'i');
    if (dName !== null) {
      value = value.replace(dRegExp, "");
    }
    //find year and date
    _YEAR_AND_DATE_REGEXP.lastIndex = 0;
    var match = _YEAR_AND_DATE_REGEXP.exec(value);
    if (match === null) {
      var dtype = isDateTime ? 2 : 0;
      _throwDateFormatMismatch(origValue, format, dtype);
    }
    var tokenIndexes = [{
        'y': format.indexOf("y")
      }, {
        'd': format.indexOf("d")
      }];

    tokenIndexes.sort(function (a, b) {
      for (var n1 in a)
      {
        break;
      }
      for (var n2 in b)
      {
        break;
      }
      return a[n1] - b[n2];
    });

    var year, day, yearIndex;
    var foundYear = false;
    for (i = 1; i <= 2; i++)
    {
      var tokenMatch = match[i];
      //find year if year is yyy|yyyy
      if (tokenMatch.length > 2 || tokenMatch > 31)
      {
        year = tokenMatch;
        foundYear = true;
        yearIndex = i - 1;
      }
    }
    if (!foundYear) {
      yearIndex = _getTokenIndex(tokenIndexes, 'y');
      year = match[_getTokenIndex(tokenIndexes, 'y') + 1];
    }
    if (yearIndex === _getTokenIndex(tokenIndexes, 'd'))
    {
      day = match[_getTokenIndex(tokenIndexes, 'y') + 1];
    }
    else {
      day = match[_getTokenIndex(tokenIndexes, 'd') + 1];
    }

    var start2DigitYear = _get2DigitYearStart(options);
    year = _expandYear(start2DigitYear, parseInt(year, 10));
    _validateRange("year", year, 0, 9999, year, 0, 9999);
    var parsedDate = new Date(year, month, day);
    if (dName !== null) {
      var weekDay = _getDayIndex1(localeElements, dName, 0);
      // day of week does not match date
      if (parsedDate.getDay() !== weekDay) {
        _throwWeekdayMismatch(dName, parsedDate.getDate());
      }
    }
    var daysInMonth = _getDaysInMonth(year, month);
    _validateRange("day", day, 1, daysInMonth, day, 1, daysInMonth);
    var result =
        {
          'value': parsedDate,
          'offset': '',
          'warning': 'lenient parsing was used'
        };
    if (isDateTime) {
      var timepart = value.substr(_YEAR_AND_DATE_REGEXP.lastIndex);
      if (timepart.length === 0)
        result['value'].setHours(0, 0, 0, 0);
      else
        _parseLenienthms(result['value'], timepart, format, localeElements, 2);
    }
    result['value'] = oj.OraI18nUtils.dateToLocalIso(result['value']);
    return result;
  };

  _parseLenient = function (value, format, options, localeElements) {
    var dtStyle = _dateTimeStyle(options, "OraDateTimeConverter.parse");
    switch (dtStyle) {
      //date style
      case 0 :
        return _parseLenientyMMMEd(value, format, options, localeElements,
            false);
        break
        //time style
      case 1 :
        var d = new Date();
        _parseLenienthms(d, value, format, localeElements, 1);
        var isoStr = oj.OraI18nUtils.dateToLocalIso(d);
        var result =
            {
              'value': isoStr,
              'offset': '',
              'warning': 'lenient parsing was used'
            };
        return result;
        break;
        //date-time style
      case 2 :
        return _parseLenientyMMMEd(value, format, options, localeElements,
            true);
        break;
      default:
        break;
    }
    return null;
  };

  _getNameIndex = function (localeElements, datePart, matchGroup, mLength,
      style, matchIndex, start1, end1, start2, end2, name) {
    var index;
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var monthsFormat =
        mainNode['dates']['calendars']['gregorian'][datePart][style];
    var startName;
    var endName;
    if (datePart === 'months') {
      index = _getMonthIndex(localeElements, matchGroup, matchIndex);
    }
    else {
      index = _getDayIndex1(localeElements, matchGroup, matchIndex);
    }
    startName = monthsFormat[mLength][start2];
    endName = monthsFormat[mLength][end2];
    _validateRange(name, index, start1, end1, matchGroup,
        startName, endName);
    return index;
  };

  _getTimePart = function (matchInt, timeObj, objMap, timeToken) {
    timeObj[objMap['timePart']] = matchInt;
    if (timeToken === 'h' || timeToken === 'hh') {
      if (matchInt === 12)
        timeObj[objMap['timePart']] = 0;
    }
    else if (timeToken === 'k' || timeToken === 'kk') {
      if (matchInt === 24)
        timeObj[objMap['timePart']] = 0;
    }
    _validateRange(objMap['timePart'], timeObj[objMap['timePart']],
        objMap['start1'], objMap['end1'], matchInt,
        objMap['start2'], objMap['end2']);
  };

  //exact match parsing for date-time. If it fails, try lenient parse.
  _parseExact = function (value, format, options, localeElements) {
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    // remove spaces from era.
    var cal = mainNode['dates']['calendars']['gregorian'];
    var eraPart = cal['eras']['eraAbbr']['1'];
    var trimEraPart = oj.OraI18nUtils.trimNumber(eraPart);
    value = value.replace(eraPart, trimEraPart);
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraDateTimeConverter.parse");
    var lenientParse = getOption('lenientParse', 'string', ['none', 'full'], 'full');
    var dtStyle = _dateTimeStyle(options, "OraDateTimeConverter.parse");
    // convert date formats into regular expressions with groupings.
    // use the regexp to determine the input format and extract the date
    //  fields.
    var parseInfo = _getParseRegExp(format, options);
    var match = new RegExp(parseInfo['regExp']).exec(value);
    if (match === null) {
      if (lenientParse === 'full')
        return _parseLenient(value, format, options, localeElements);
      else
        _throwDateFormatMismatch(value, format, dtStyle);
    }
    // found a date format that matches the input.
    var groups = parseInfo['groups'],
        year = null, month = null, date = null, weekDay = null,
        hourOffset = "", tzID = null,
        pmHour = false, weekDayName,
        timeObj = {
          'hour': 0,
          'minute': 0,
          'second': 0,
          'millisec': 0
        },
        calPM = mainNode['dates']['calendars']['gregorian']['dayPeriods']['format']['wide']['pm'];
    var start2DigitYear = _get2DigitYearStart(options);
    // iterate the format groups to extract and set the date fields.
    for (var j = 0, jl = groups.length; j < jl; j++) {
      var matchGroup = match[ j + 1 ];
      if (matchGroup) {
        var current = groups[ j ],
            matchInt = parseInt(matchGroup, 10),
            currentGroup = _PROPERTIES_MAP[current];
        switch (currentGroup['token']) {
          case 'months':
            month = _getNameIndex(localeElements, currentGroup['token'],
                matchGroup, currentGroup['mLen'], currentGroup['style'],
                currentGroup['matchIndex'], 0, 11, '1', '12', 'month name');
            break;
          case 'days':
            weekDayName = matchGroup;
            weekDay = _getNameIndex(localeElements, currentGroup['token'],
                matchGroup, currentGroup['dLen'], currentGroup['style'],
                currentGroup['matchIndex'], 0, 6, 'sun', 'sat', 'weekday');
            break;
          case 'time':
            _getTimePart(matchInt, timeObj, currentGroup, current);
            break;
          case 'dayOfMonth':
            date = matchInt;
            //try leneient parse for date style only
            if (date > 31 && lenientParse === 'full')
              return _parseLenient(value, format, options, localeElements);
            break;
          case 'monthIndex':
            // Month.
            month = matchInt - 1;
            //try leneient parse for date style only
            if (month > 11 && lenientParse === 'full')
              return _parseLenient(value, format, options, localeElements);
            break;
          case 'year':
            year = _expandYear(start2DigitYear, matchInt);
            _validateRange("year", year, 0, 9999, year, 0, 9999);
            break;
          case "ampm":
            pmHour = (oj.OraI18nUtils.toUpper(matchGroup)).indexOf(oj.OraI18nUtils.toUpper(calPM)) !== -1;
            break;
          case "tzhm":
            // Time zone hours minutes: -0800 
            hourOffset = matchGroup.substr(-2);
            hourOffset = matchGroup.substr(0, 3) + ":" + hourOffset;
            break;
          case "tzhsepm":
            // Time zone hours minutes: -08:00 
            hourOffset = matchGroup;
            break;
          case "tzh":
            // Time zone hours minutes: -08 
            hourOffset = matchGroup + ":00";
            break;
          case "tzid":
            // Time zone ID: America/Los_Angeles 
            tzID = matchGroup;
            break;
        }
      }
    }
    var parsedDate = new Date();
    if (year === null) {
      year = parsedDate.getFullYear();
    }
    // if day and month are unspecified,the defaults are current 
    // day and month.
    if (month === null && date === null) {
      month = parsedDate.getMonth();
      date = parsedDate.getDate();
    }
    // if day is unspecified, default 1st day of month.
    else if (date === null) {
      date = 1;
    }

    //validate month
    _validateRange("month", month, 0, 11, month + 1, 1, 12);

    //validate day range, depending on the month and year
    var daysInMonth = _getDaysInMonth(year, month);
    _validateRange("day", date, 1, daysInMonth, date, 1, daysInMonth);
    parsedDate.setFullYear(year, month, date);

    // day of week does not match date
    if (weekDay !== null && parsedDate.getDay() !== weekDay) {
      _throwWeekdayMismatch(weekDayName, parsedDate.getDate());
    }
    // if pm designator token was found make sure the hours fit the 
    // 24-hour clock.
    if (pmHour && timeObj['hour'] < 12) {
      timeObj['hour'] += 12;
    }
    var parts = [year, month + 1, date, timeObj['hour'], timeObj['minute'],
      timeObj['second'], timeObj['millisec']];
    var isoParsedDate = oj.OraI18nUtils.partsToIsoString(parts);
    if (tzID !== null) {
      var zone = _getTimeZone(tzID, localeElements);
      var index = _parseZone(zone, parts, false, true, true);
      hourOffset = -zone.ofset(index);
      hourOffset = oj.OraI18nUtils.getTimeStringFromOffset("", hourOffset, false, true)
    }
    if (hourOffset !== "") {
      isoParsedDate += hourOffset;
    }
    var result =
        {
          'value': isoParsedDate
        };
    return result;
  };

  //given a user defined pattern, derive the ecma options that will
  //be returned by getResolvedOptions method
  _getResolvedOptionsFromPattern = function (locale, numberingSystemKey,
      pattern) {
    // expand single digit formats, then escape regular expression 
    // characters.
    var expFormat = pattern.replace(_ESCAPE_REGEXP, "\\\\$1");
    var regexp = ["^"];
    var quoteCount = 0;
    var index = 0;
    var match;
    var result = {
      'locale': locale,
      'numberingSystem': numberingSystemKey,
      'calendar': 'gregorian'
    };
    // iterate through each date token found.
    while ((match = _TOKEN_REGEXP.exec(expFormat)) !== null) {
      var preMatch = expFormat.slice(index, match.index);
      index = _TOKEN_REGEXP.lastIndex;

      // skip matches that occur inside a string literal.
      quoteCount += _appendPreOrPostMatch(preMatch, regexp);
      if (quoteCount % 2) {
        continue;
      }

      // add a regex group for the token.
      var m = match[ 0 ];
      if (m === '/' || m === 'zzzz' || m === 'zzz' || m === 'zz' || m === 'z') {
        continue;
      }
      if (_PROPERTIES_MAP[m] !== undefined) {
        if (_PROPERTIES_MAP[m]['key'] !== undefined)
          result[_PROPERTIES_MAP[m]['key']] = _PROPERTIES_MAP[m]['value'];
        if (m === 'kk' || m === 'HH' || m === 'H' || m === 'k') {
          result['hour12'] = false;
        }
        else if (m === 'KK' || m === 'hh' || m === 'h' || m === 'K') {
          result['hour12'] = true;
        }
      }
      else {
        _throwInvalidDateFormat(pattern, result, m);
      }
    }
    return result;
  };
  //test if the pattern is date, time or date-time
  //0: date, 1:time, 2:date-time
  _dateTimeStyleFromPattern = function (pattern) {
    var result = _getResolvedOptionsFromPattern('', '', pattern);
    var isDate = (result['year'] !== undefined || result['month'] !==
        undefined ||
        result['weekday'] !== undefined || result['day'] !== undefined);
    var isTime = (result['hour'] !== undefined || result['minute'] !==
        undefined || result['second'] !== undefined ||
        result['millisecond'] !== undefined);
    if (isDate && isTime)
      return 2;
    else if (isTime)
      return 1;
    else
      return 0;
  };

  //test if the isoStr is date, time or date-time
  //0: date, 1:time, 2:date-time
  _isoStrDateTimeStyle = function (isoStr) {
    var timeIndex = isoStr.indexOf("T");
    if (timeIndex === -1)
      return 0;
    if (timeIndex > 0)
      return 2;
    return 1;
  };

  //test if the pattern/options is date, time or date-time
  //0: date, 1:time, 2:date-time
  _dateTimeStyle = function (options, caller) {
    //try pattern
    if (options['pattern'] !== undefined) {
      return _dateTimeStyleFromPattern(options['pattern']);
    }

    //try ecma options
    var getOption = oj.OraI18nUtils.getGetOption(options, caller);
    var isTime = (getOption('hour', 'string', ['2-digit', 'numeric']) !==
        undefined ||
        getOption('minute', 'string', ['2-digit', 'numeric']) !== undefined ||
        getOption('second', 'string', ['2-digit', 'numeric']) !== undefined ||
        getOption('millisecond', 'string', ['numeric']) !== undefined);
    var isDate = (getOption('year', 'string', ['2-digit', 'numeric']) !==
        undefined ||
        getOption('month', 'string',
            ['2-digit', 'numeric', 'narrow', 'short', 'long']) !== undefined ||
        getOption('day', 'string', ['2-digit', 'numeric']) !== undefined ||
        getOption('weekday', 'string', ['narrow', 'short', 'long']) !==
        undefined);
    if (isDate && isTime)
      return 2;
    else if (isTime)
      return 1;
    else if (isDate)
      return 0;

    //try predefined style
    var option = getOption('formatType', 'string',
        ['date', 'time', 'datetime'], 'date');
    if (option === 'datetime')
      return 2;
    else if (option === 'time')
      return 1;
    return 0;
  };

  _getStdOffset = function (zone, value) {
    var index = _parseZone(zone, value, false, true, false);
    var offset0 = zone.ofset(index);
    var offset1 = zone.ofset(index + 1);
    return Math.max(offset0, offset1);
  };

  _adjustHours = function (isoStrInfo, options, localeElements) {
    var value = isoStrInfo['isoStrParts'];
    var isoStrFormat = isoStrInfo['format'];
    var timeZone = options['timeZone'];
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraDateTimeConverter.parse");
    var zone = _getTimeZone(timeZone, localeElements);
    var utcd;
    var origOffset;
    var newOffset;
    var index;
    utcd = Date.UTC(value[0], value[1] - 1, value[2],
        value[3], value[4], value[5]);
    switch (isoStrFormat) {
      case _OFFSET :
        var tzPart = isoStrInfo['timeZone'];
        origOffset = tzPart.split(":");
        var hoursOffset = parseInt(origOffset[0], 10);
        var minOffset = parseInt(origOffset[1], 10);
        origOffset = (hoursOffset * 60) +
            (oj.OraI18nUtils.startsWith(tzPart, "-") ? -minOffset : minOffset);
        break;
      case _ZULU :
        origOffset = 0;
        break;
      default :
        break;
    }
    //get target zone offset:
    //1.get target zone standard time offset
    newOffset = _getStdOffset(zone, value);
    //2.adjust utcd to target zone
    utcd -= (newOffset + origOffset) * 60000;
    //3.get target zone offset
    index = zone.parse(utcd, false, true, false);
    newOffset = -zone.ofset(index);
    //adjust the offset
    newOffset -= origOffset;
    //Do the offset math through the Date object.
    var adjustD = new Date(Date.UTC(value[0], value[1] - 1, value[2],
        value[3], value[4], value[5]));

    var adjustedMin = adjustD.getUTCMinutes() + newOffset;
    adjustD.setUTCHours(adjustD.getUTCHours() + ((adjustedMin / 60) << 0),
        adjustedMin % 60);
    value[0] = adjustD.getUTCFullYear();
    value[1] = adjustD.getUTCMonth() + 1;
    value[2] = adjustD.getUTCDate();
    value[3] = adjustD.getUTCHours();
    value[4] = adjustD.getUTCMinutes();
    value[5] = adjustD.getUTCSeconds();

  };

  //test if isoStrInfo is date, time or date-time
  //0: date, 1:time, 2:date-time
  _getIsoStrStyle = function (isoStrInfo) {
    var dt = isoStrInfo['dateTime'];
    var dtStyle = 2;
    var dtParts = dt.split('T');
    //time only Thh:mm:ss
    if (dtParts[0] === "")
      dtStyle = 1;
    //date only yy-MM-dd
    else if (dtParts[1] === undefined)
      dtStyle = 0;
    //date-time    
    return dtStyle;
  };

  //Returns a time-only, date-only or date-time ISO string based on dtStyle.
  _createISOStrParts = function (dtStyle, d) {
    var ms;
    var val = "";
    switch (dtStyle) {
      //Date only
      case 0 :
        val = oj.OraI18nUtils.padZeros(d[0], 4) + "-" + oj.OraI18nUtils.padZeros(d[1], 2) + "-" +
            oj.OraI18nUtils.padZeros(d[2], 2);
        break;
        //Time only
      case 1 :
        val = "T" + oj.OraI18nUtils.padZeros(d[3], 2) + ":" +
            oj.OraI18nUtils.padZeros(d[4], 2) + ":" +
            oj.OraI18nUtils.padZeros(d[5], 2);
        ms = d[6];
        if (ms > 0) {
          val += "." + oj.OraI18nUtils.trimRightZeros(oj.OraI18nUtils.padZeros(ms, 3));
          ;
        }
        break;
        //Date-Time  
      default :
        val = oj.OraI18nUtils.padZeros(d[0], 4) + "-" +
            oj.OraI18nUtils.padZeros(d[1], 2) + "-" +
            oj.OraI18nUtils.padZeros(d[2], 2) +
            "T" + oj.OraI18nUtils.padZeros(d[3], 2) + ":" +
            oj.OraI18nUtils.padZeros(d[4], 2) + ":" +
            oj.OraI18nUtils.padZeros(d[5], 2);
        ms = d[6];
        if (ms > 0) {
          val += "." + oj.OraI18nUtils.trimRightZeros(oj.OraI18nUtils.padZeros(ms, 3));
        }
        break;
    }
    return val;
  };

  _getParseISOStringOffset = function (tzName, parts, dst, ignoreDst, localeElements, thowException) {
    var zone = _getTimeZone(tzName, localeElements);
    var index = _parseZone(zone, parts, dst, ignoreDst, thowException);
    //hours
    var offset = zone.ofset(index);
    return oj.OraI18nUtils.getTimeStringFromOffset('', offset, true, true);
  };

  _createParseISOStringFromDate = function (dtStyle, isoStrInfo, options, localeElements) {
    var zone;
    var index;
    var offset;
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraDateTimeConverter.parse");
    var isoFormat = getOption('isoStrFormat', 'string',
        [_ZULU, _OFFSET, _INVARIANT, _LOCAL, _AUTO], _AUTO);
    var dst = getOption('dst', 'boolean', [true, false], false);
    var ignoreDst = true;
    var parts = isoStrInfo['isoStrParts'];
    var dTimeZone = isoStrInfo['timeZone'];
    var tzName = options['timeZone'];
    var isoStrFormat = isoStrInfo['format'];
    var optionsFormat = options['isoStrFormat'];
    var val = _createISOStrParts(dtStyle, parts);
    //do not include timezone if date-only or local
    if (dtStyle === 0 || optionsFormat === 'local') {
      return val;
    }
    //when iso string is time-only, do not ignore dst outside ambiguous intervals.
    if (dtStyle === 1) {
      ignoreDst = false;
    }
    switch (isoFormat) {
      case _OFFSET :
        if (tzName === undefined && isoStrFormat === _OFFSET) {
          val += dTimeZone;
        }
        else if (tzName === undefined && isoStrFormat === _LOCAL) {
          val += "";
        }
        else if (tzName === undefined && isoStrFormat === _ZULU) {
          val += "+00:00";
        }
        else if (tzName !== undefined) {
          offset = _getParseISOStringOffset(tzName, parts, dst, ignoreDst, localeElements, true);
          val += offset;
        }
        break;
      case _ZULU :
        var adjustedMin = 0;
        if (tzName === undefined) {
          if (isoStrFormat === _OFFSET) {
            offset = dTimeZone.split(":");
            var offsetHours = parseInt(offset[0], 10);
            var offsetMinutes = parseInt(offset[1], 10);
            adjustedMin = (offsetHours * 60) +
                (oj.OraI18nUtils.startsWith(offset[0], "-") ? -offsetMinutes : offsetMinutes);
            adjustedMin = -adjustedMin;
          }
        }
        else {
          zone = _getTimeZone(tzName, localeElements);
          index = _parseZone(zone, parts, dst, ignoreDst, true);
          offset = zone.ofset(index);
          adjustedMin = offset;
        }
        if (adjustedMin !== 0) {
          //Do the offset math through date object.
          var adjustD = new Date(Date.UTC(parts[0], parts[1] - 1, parts[2],
              parts[3], parts[4], parts[5], parts[6]));
          adjustedMin = adjustD.getUTCMinutes() + adjustedMin;
          adjustD.setUTCHours(adjustD.getUTCHours() + ((adjustedMin / 60) << 0),
              adjustedMin % 60);
          parts[0] = adjustD.getUTCFullYear();
          parts[1] = adjustD.getUTCMonth() + 1;
          parts[2] = adjustD.getUTCDate();
          parts[3] = adjustD.getUTCHours();
          parts[4] = adjustD.getUTCMinutes();
          parts[5] = adjustD.getUTCSeconds();
          val = _createISOStrParts(dtStyle, parts);
        }
        val += "Z";
        break;
      case _AUTO :
        if (tzName !== undefined) {
          offset = _getParseISOStringOffset(tzName, parts, dst, ignoreDst, localeElements, true);
          val += offset;
        }
        else {
          offset = dTimeZone;
          if (offset)
            val += offset;
        }
        break;
      default :
        break;
    }
    return val;
  };

  _parseImpl = function (str, localeElements, options, locale) {
    var numberingSystemKey = oj.OraI18nUtils.getNumberingExtension(locale);
    if (oj.OraI18nUtils.numeringSystems[numberingSystemKey] === undefined)
      numberingSystemKey = 'latn';
    if (numberingSystemKey !== 'latn') {
      var idx;
      var latnStr = [];
      for (idx = 0; idx < str.length; idx++)
      {
        var pos = oj.OraI18nUtils.numeringSystems[numberingSystemKey].indexOf(str[idx]);
        if (pos !== -1)
          latnStr.push(pos);
        else
          latnStr.push(str[idx]);
      }
      str = latnStr.join("");
    }
    if (arguments.length <= 2 || options === undefined) {
      //default is yMd
      options = {
        'year': 'numeric',
        'month': 'numeric',
        'day': 'numeric'
      };
    }
    var dtStyle;
    var formats;
    //First try if str is an iso 8601 string 
    var testIsoStr = oj.OraI18nUtils._ISO_DATE_REGEXP.test(str);
    var parsedIsoStr;
    var isoStrInfo;
    var res = {};
    if (testIsoStr === true) {
      parsedIsoStr = str;
      dtStyle = _isoStrDateTimeStyle(str);
    }
    else {
      formats = options['pattern'] || _expandFormat(options, localeElements,
          locale, "OraDateTimeConverter.parse");
      dtStyle = _dateTimeStyle(options, "OraDateTimeConverter.parse");
      res = _parseExact(str, formats, options, localeElements);
      parsedIsoStr = res['value'];
    }
    isoStrInfo = oj.OraI18nUtils.getISOStrFormatInfo(parsedIsoStr);
    if (options['timeZone'] !== undefined && isoStrInfo['format'] !== _LOCAL) {
      _adjustHours(isoStrInfo, options, localeElements);
    }
    parsedIsoStr = _createParseISOStringFromDate(dtStyle, isoStrInfo, options, localeElements);
    res['value'] = parsedIsoStr;
    return res;
  };

  //If one of the iso strings is local/invariant parse both strings as local.
  //Else, normalize both strings to zulu
  _getCompareISODatesOptions = function (isoStr1, isoStr2) {
    var options = {'isoStrFormat': _LOCAL};
    var isoInfo1 = oj.OraI18nUtils.getISOStrFormatInfo(isoStr1);
    var isoInfo2 = oj.OraI18nUtils.getISOStrFormatInfo(isoStr2);
    var isoInfo1Format = isoInfo1['format'];
    var isoInfo2Format = isoInfo2['format'];
    if (isoInfo1Format === _LOCAL || isoInfo2Format === _LOCAL) {
      return options;
    }
    options['isoStrFormat'] = _ZULU;
    return options;
  };

  _getPatternResolvedOptions = function (isoFormat, tz, dst, locale,
      numberingSystemKey, options, getOption) {
    var result = _getResolvedOptionsFromPattern(locale, numberingSystemKey,
        options['pattern']);
    result['pattern'] = options['pattern'];
    if (isoFormat !== undefined)
      result['isoStrFormat'] = isoFormat;
    if (tz !== undefined)
      result['timeZone'] = tz;
    if (dst !== undefined)
      result['dst'] = dst;
    result['two-digit-year-start'] = _get2DigitYearStart(options);
    result['lenientParse'] = getOption('lenientParse', 'string', ['none', 'full'], 'full');
    return result;
  };

  _getECMAResolvedOptions = function (getOption, result, dst, localeElements) {
    var ecma = false;
    var option;
    if (dst !== undefined)
      result['dst'] = dst;
    option = getOption('year', 'string', ['2-digit', 'numeric']);
    if (option !== undefined) {
      result['year'] = option;
      ecma = true;
    }
    option = getOption('era', 'string', ['narrow', 'short', 'long']);
    if (option !== undefined) {
      result['era'] = option;
      ecma = true;
    }
    option = getOption('month', 'string', ['2-digit', 'numeric',
      'narrow', 'short', 'long']);
    if (option !== undefined) {
      result['month'] = option;
      ecma = true;
    }
    option = getOption('day', 'string', ['2-digit', 'numeric']);
    if (option !== undefined) {
      result['day'] = option;
      ecma = true;
    }
    option = getOption('weekday', 'string', ['narrow', 'short', 'long']);
    if (option !== undefined) {
      result['weekday'] = option;
      ecma = true;
    }
    option = getOption('hour', 'string', ['2-digit', 'numeric']);
    if (option !== undefined) {
      result['hour'] = option;
      ecma = true;
      option = getOption('hour12', 'boolean', [true, false]);
      if (option === undefined)
        option = _isHour12(localeElements);
      result['hour12'] = option;
    }
    option = getOption('minute', 'string', ['2-digit', 'numeric']);
    if (option !== undefined) {
      result['minute'] = option;
      ecma = true;
    }
    option = getOption('second', 'string', ['2-digit', 'numeric']);
    if (option !== undefined) {
      result['second'] = option;
      ecma = true;
    }
    option = getOption('millisecond', 'string', ['numeric']);
    if (option !== undefined) {
      result['millisecond'] = option;
      ecma = true;
    }
    return ecma;
  };

  _getPredefinedStylesResolvedOptions = function (result, options, localeElements, getOption) {
    var format = _expandPredefinedStylesFormat(options,
        localeElements, OraDateTimeConverter.resolvedOptions);
    var fmtType = getOption('formatType', 'string',
        ['date', 'time', 'datetime'], 'date');
    var dStyle = getOption('dateFormat', 'string',
        ['short', 'medium', 'long', 'full'], 'short');
    var tStyle = getOption('timeFormat', 'string',
        ['short', 'medium', 'long', 'full'], 'short');
    result['formatType'] = fmtType;
    if (fmtType === 'datetime' || fmtType === 'date') {
      result['dateFormat'] = dStyle;
    }
    if (fmtType === 'datetime' || fmtType === 'time') {
      result['timeFormat'] = tStyle;
    }
    result['patternFromOptions'] = format;
    result['lenientParse'] = getOption('lenientParse', 'string', ['none', 'full'], 'full');
  };

  _getRelativeTimeResolvedOptions = function (getOption, result) {
    var option = getOption('formatUsing', 'string', ['displayName', 'calendar']);
    if (option !== undefined) {
      result['formatUsing'] = option;
    }
    option = getOption('dateField', 'string',
        ['day', 'week', 'month', 'year', 'hour', 'minute', 'second']);
    if (option !== undefined) {
      result['dateField'] = option;
    }
    option = getOption('relativeTime', 'string',
        ['fromNow', 'toNow']);
    if (option !== undefined) {
      result['relativeTime'] = option;
    }
    option = getOption('dateOnly', 'boolean', [true, false]);
    if (option !== undefined) {
      result['dateOnly'] = option;
    }
  };

  _getResolvedOptions = function (options, getOption, tz, dst, isoFormat,
      localeElements, numberingSystemKey, locale) {
    var result = {
      'locale': locale,
      'numberingSystem': numberingSystemKey,
      'calendar': 'gregorian'
    };
    var option;
    var count = 0;
    for (option in options) {
      count++;
    }
    if (count === 0) {
      result['year'] = 'numeric';
      result['month'] = 'numeric';
      result['day'] = 'numeric';
      result['lenientParse'] = 'full';
      return result;
    }
    if (tz !== undefined) {
      result['timeZone'] = tz;
      if (isoFormat !== undefined)
        result['isoStrFormat'] = isoFormat;
    }

    _getRelativeTimeResolvedOptions(getOption, result);

    var ecma = _getECMAResolvedOptions(getOption, result, dst, localeElements);
    result['two-digit-year-start'] = _get2DigitYearStart(options);
    if (!ecma) {
      _getPredefinedStylesResolvedOptions(result, options, localeElements, getOption);
      return result;
    }
    if (tz !== undefined) {
      option = getOption('timeZoneName', 'string', ['short', 'long']);
      if (option !== undefined) {
        result['timeZoneName'] = option;
      }
    }
    result['lenientParse'] = getOption('lenientParse', 'string', ['none', 'full'], 'full');
    result['patternFromOptions'] = _expandFormat(result,
        localeElements, locale, "OraDateTimeConverter.resolvedOptions");
    return result;
  };

  _getResolvedDefaultOptions = function (localeElements, locale, numberingSystemKey) {
    var defaultOptions = {
      'year': 'numeric',
      'month': 'numeric',
      'day': 'numeric'
    };
    var patternFromOptions = _expandFormat(defaultOptions,
        localeElements, locale, "OraDateTimeConverter.resolvedOptions");
    return {
      'calendar': 'gregorian',
      'locale': locale,
      'numberingSystem': numberingSystemKey,
      'year': 'numeric',
      'month': 'numeric',
      'day': 'numeric',
      'patternFromOptions': patternFromOptions,
      'lenientParse': 'full'
    };
  };

  _resolvedOptionsImpl = function (localeElements, options, locale) {
    if (arguments.length < 3 || locale === undefined) {
      locale = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
    }
    if (arguments.length < 2 || options === undefined)
    {
      //default is yMd
      options = {
        'year': 'numeric',
        'month': 'numeric',
        'day': 'numeric'
      };
    }
    var tz;
    var isoFormat;
    var dst;
    var getOption = oj.OraI18nUtils.getGetOption(options, "OraDateTimeConverter.resolvedOptions");
    if (options !== undefined) {
      isoFormat = getOption('isoStrFormat', 'string',
          [_ZULU, _OFFSET, _INVARIANT, _LOCAL, _AUTO], _AUTO);
      dst = getOption('dst', 'boolean', [true, false], false);
      tz = options['timeZone'];
    }
    var numberingSystemKey = oj.OraI18nUtils.getNumberingExtension(locale);
    if (oj.OraI18nUtils.numeringSystems[numberingSystemKey] === undefined)
      numberingSystemKey = 'latn';
    if (options !== undefined && options['pattern'] !== undefined) {
      return _getPatternResolvedOptions(isoFormat, tz, dst, locale,
          numberingSystemKey, options, getOption);
    }
    if (options !== undefined) {
      return _getResolvedOptions(options, getOption, tz, dst, isoFormat,
          localeElements, numberingSystemKey, locale);
    }
    return _getResolvedDefaultOptions(localeElements, locale, numberingSystemKey);
  };

  _availableTimeZonesImpl = function (localeElements) {

    function getLocalizedName(id, offset, metaZones, cities) {
      var parts = id.split('/');
      var region = parts[0];
      var city = parts[1];
      var locCity = "";
      var locZone = "";
      var nameObject = {};
      var metaRegion = cities[region];
      if (metaRegion !== undefined) {
        locCity = metaRegion[city];
        if (locCity !== undefined) {
          locCity = locCity['exemplarCity'];
          if (locCity !== undefined) {
            locCity = " " + locCity;
          }
        }
      }
      id = region + "/" + city;
      var metazones = localeElements['supplemental']['metazones'];
      var metaZone = _getMetazone(dParts, id, metazones);
      if (metaZones !== undefined)
        metaZone = metaZones[metaZone];
      if (metaZone !== undefined && metaZone !== null && metaZone['long'] !== undefined) {
        locZone = metaZone['long']['generic'];
        //some metazones do not have generic. Use standard 
        if (locZone === undefined) {
          locZone = metaZone['long']['standard'];
        }
        if (locZone !== undefined) {
          locZone = " - " + locZone;
        }
      }
      var locName = '(' + _UTC + ')';
      if (offset !== 0) {
        locName = oj.OraI18nUtils.getTimeStringFromOffset(_UTC, offset, true, true);
        locName = "(" + locName + ")";
      }
      if (locCity === undefined || locZone === undefined) {
        return null;
      }
      nameObject['offsetLocName'] = locName + locCity + locZone;
      nameObject['locName'] = locCity + locZone;
      return nameObject;
    }

    function pushZoneNameObject(zones) {
      var i;
      var zone;
      var offset;
      for (i in zones) {
        zone = tz.getZone(i, localeElements);
        offset = _getStdOffset(zone, dParts);
        var localizedName = getLocalizedName(i, offset, metaZones, cities);
        if (localizedName === null) {
          continue;
        }
        sortedZones.push({
          'id': i,
          'displayName': localizedName
        });
        offsets[i] = offset;
      }
    }
    //return cahched array if available
    var locale = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
    if (_timeZoneDataCache[locale] !== undefined) {
      var ret = _timeZoneDataCache[locale]['availableTimeZones'];
      if (ret !== undefined) {
        return ret;
      }
    }
    var tz = OraTimeZone.getInstance();
    var sortOptions = {sensitivity: 'variant'};
    var sortLocale = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
    var mainNode = oj.OraI18nUtils.getLocaleElementsMainNode(localeElements);
    var metaZones = mainNode['dates']['timeZoneNames']['metazone'];
    var cities = mainNode['dates']['timeZoneNames']['zone'];
    var sortedZones = [];
    var offsets = {};
    var d = new Date();
    var dParts = [d.getFullYear(), d.getMonth() + 1, d.getDate(),
      d.getHours(), d.getMinutes(), d.getSeconds()];
    var tzData = localeElements['supplemental']['timeZoneData'];

    pushZoneNameObject(tzData['zones']);
    //add the links
    pushZoneNameObject(tzData['links']);
    sortedZones.sort(function (a, b) {
      var res1 = (offsets[b['id']] - offsets[a['id']]);
      var res2 = a['displayName']['locName'].localeCompare(b['displayName']['locName'], sortLocale, sortOptions);
      return (res1 + res2);
    });
    var len = sortedZones.length;
    //return an array with "display name with offset" instead of the
    //object localizedName which was only used for sorting
    for (var j = 0; j < len; j++) {
      sortedZones[j]['displayName'] = sortedZones[j]['displayName']['offsetLocName'];
    }
    //cache the sorted zones
    if (_timeZoneDataCache[locale] === undefined) {
      _timeZoneDataCache[locale] = {};
      _timeZoneDataCache[locale]['availableTimeZones'] = sortedZones;
    }
    return sortedZones;
  };

  function _init() {
    return {
      /**
       * Format a date.
       * @memberOf OraDateTimeConverter
       * @param {string} value - an iso 8601 string to be formatted. It  may be in 
       * extended or non-extended form. http://en.wikipedia.org/wiki/ISO_8601
       * @param {Object} localeElements - the instance of LocaleElements bundle.
       * @param {Object=} options - Containing the following properties:<br>
       * - <b>weekday.</b> Allowed values: "narrow", "short", "long".<br>
       * - <b>era.</b> Allowed values: "narrow", "short", "long".<br>
       * - <b>year.</b> Allowed values:"2-digit", "numeric".<br>
       * - <b>month.</b> Allowed values: "2-digit", "numeric", "narrow", 
       * "short", "long".<br>
       * - <b>day.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>hour.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>minute.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>second.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>millisecond.</b> Allowed values: "numeric".<br>
       * - <b>timeZone.</b> The possible values of the timeZone property are valid IANA timezone IDs.
       * If the users want to pass an offset, they can use one of the Etc/GMT timezone IDs. 
       *  yet.<br>
       * - <b>timeZoneName.</b> allowed values are "short", "long". </b> <br>
       * - <b>dst.</b> is a Boolean value. Setting dst to true indicates the time is in DST. By default the
       * time is interpreted as standard time. The possible values of dst are: "true" or "false". Default is "false".<br>
       * - <b>hour12.</b> is a Boolean value indicating whether 12-hour format 
       * (true) or 24-hour format (false) should be used. It is only relevant 
       * when hour is also present.<br>
       * - <b>pattern.</b> custom String pattern as defined by Unicode CLDR.<br>
       * - <b>formatType.</b> a predefined formatting type. Allowed values: 
       * "date", "time", "datetime".<br>
       * - <b>dateFormat.</b> optional, specifies the date format field. 
       * Allowed values: "short", "medium", "long", "full". It is only 
       * considered when formatType is present. The default value 
       * is "short".<br>
       * - <b>timeFormat.</b> optional, specifies the time format field. 
       * Allowed values: "short", "medium", "long", "full". It is only 
       * considered when formatType is present. The default value 
       * is "short".<br><br>
       * The order of precedence is the following:<br>
       * 1. pattern.<br>
       * 2. ECMA options.<br>
       * 3. formatType.<br>
       * If options is ommitted, the default will be the following object:<br>
       * {<br>
       * year:"numeric",<br> 
       * month:"numeric",<br> 
       * day:"numeric"<br>
       * };
       * @param {string=} locale - A BCP47 compliant language tag. it is only 
       * used to extract the unicode "nu" extension key. We currently support "arab", "latn" and "thai" 
       * numbering systems. EX: locale: 'en-US-u-nu-latn' 
       * @return {string|null} formatted date.
       * @throws {RangeError} If a propery value of the options parameter is 
       * out of range.
       * @throws {SyntaxError} If an Unexpected token is encountered in the 
       * pattern.
       * @throws {invalidISOString} if the ISO string is not valid.
       
       */
      format: function (value, localeElements, options, locale) {
        var val;
        var isoStrInfo;
        if (typeof value === "number") {
          val = oj.OraI18nUtils.dateToLocalIso(new Date(value));
        }
        else if (typeof value === "string") {
          val = oj.OraI18nUtils.trim(value);
          if (val === '')
            return null;
        }
        else {
          return null;
        }
        if (arguments.length <= 2 || options === undefined)
        {
          //default is yMd
          options = {
            'year': 'numeric',
            'month': 'numeric',
            'day': 'numeric'
          };
        }
        isoStrInfo = oj.OraI18nUtils.getISOStrFormatInfo(val);
        var ret = _formatImpl(localeElements, options, isoStrInfo, locale);
        var numberingSystemKey = oj.OraI18nUtils.getNumberingExtension(locale);
        if (oj.OraI18nUtils.numeringSystems[numberingSystemKey] === undefined)
          numberingSystemKey = 'latn';
        if (numberingSystemKey !== 'latn') {
          var idx;
          var nativeRet = [];
          for (idx = 0; idx < ret.length; idx++)
          {
            if (ret[idx] >= '0' && ret[idx] <= '9')
              nativeRet.push(oj.OraI18nUtils.numeringSystems[numberingSystemKey][ret[idx]]);
            else
              nativeRet.push(ret[idx]);

          }
          return nativeRet.join("");
        }
        return ret;
      },
      /**
       * Formats an ISO String as a relative date time. 
       * <p>
       * 
       * @param {string} value - iso 8601 string to be formatted. This value is compared with the current date 
       * on the client to arrive at the relative formatted value.
       *  @param {Object} localeElements - the instance of LocaleElements bundle.
       * @param {Object=} options - an Object literal containing the following properties. The 
       * default options are ignored during relative formatting <br> 
       * <b>formatUsing:</b> - Specifies the relative formatting convention. 
       * Allowed values are "displayName" and "calendar". Setting value to "displayName" uses the relative 
       * display name for the instance of the dateField, and one or two past and future instances. 
       * When omitted we use the implicit rules.<br>
       * <b>dateField:</b> - To be used in conjunction of 'displayName'  value 
       * of formatUsing attribute.  Allowed values are: "day", "week", "month", "year", "hour", "minute", "second".<br>
       * relativeTime:</b> - Allowed values are: "fromNow", "toNow". "fromNow" means the system's 
       * current date is the reference and "toNow" means the value attribute is the reference. Default "fromNow".<br>
       * <b>dateOnly:</b> - A boolean value that can be used in conjunction with 
       * calendar of formatUsing attribute.  When set to true date only format is used. Example: Sunday 
       * instead of Sunday at 2:30 PM. Default value is false.<br>
       * <b>timeZone:</b> - The timeZone attribute can be used to specify the 
       * time zone of the  value parameter.  The system's time zone is used for the current time. If timeZone 
       * attribute is not specified, we use the system's time zone  for both. The value parameter, which is an 
       * iso string, can be Z or contain and offset, in this case  the timeZone attribute is overwritten.
       *
       * @return {string} relative date.
       * 
       * @example <caption>Relative time in the future using implicit rules</caption>
       * var localeElements;
       * var dateInFuture = new Date();
       * dateInFuture.setMinutes(dateInFuture.getMinutes() + 41);
       * dateInFuture = oj.OraI18nUtils.dateToLocalIso(dateInFuture);
       * var formatted = converter.formatRelative(dateInFuture, localeElements); -> in 41 minutes
       *
       * @example <caption>Relative time in the past using implicit rules</caption>
       * var localeElements;
       * var dateInPast = new Date();
       * dateInPast.setHours(dateInPast.getHours() - 20);
       * dateInPast = oj.OraI18nUtils.dateToLocalIso(dateInPast);
       * var formatted = converter.formatRelative(dateInPast, localeElements); -> 20 hours ago
       *
       * @example <caption>Relative time using dateField. Assuming system's current date is 2016-07-28.</caption>
       * Format relative year:
       * var localeElements;
       * var options = {formatUsing: "displayName", dateField: "year"};
       * var formatted = converter.formatRelative("2015-06-01T00:00:00", localeElements, options); -> last year
       *
       * @example <caption>Relative time using relativeTime. Assuming system's current date is 2016-07-28.</caption>
       * var localeElements;
       * var options = {formatUsing: "displayName", dateField: "day", relativeTime: "fromNow"};
       * var formatted = converter.formatRelative("2016-07-28T00:00:00", localeElements, options); -> tomorrow
       * options = {formatUsing: "displayName", dateField: "day", relativeTime: "toNow"};
       * formatted = converter.formatRelative("2016-07-28T00:00:00", localeElements, options); -> yesterday
       *
       * @example <caption>Relative time using calendar. Assuming system's current date is 2016-07-28.</caption>
       * var localeElements;
       * var options = {formatUsing: "calendar"};
       * var formatted = converter.formatRelative("2016-07-28T14:15:00", localeElements, options); -> tomorrow at 2:30 PM
       *
       * 
       * @example <caption>Relative time using timeZone. Assuming that the system's time zone is America/Los_Angeles.</caption>
       * var localeElements;
       * var options = {timeZone:"America/New_York"};
       * var nyDateInFuture = new Date();
       * nyDateInFuture.setHours(nyDateInFuture.getHours() + 6);
       * nyDateInFuture = oj.OraI18nUtils.dateToLocalIso(nyDateInFuture);
       * var formatted = converter.formatRelative(nyDateInFuture, localeElements, options); -> in 3 hours
       * 
       * @memberOf OraDateTimeConverter
       * @export
       */
      formatRelative: function (value, localeElements, options) {
        return _formatRelativeImpl(value, localeElements, options);
      },
      /**
       * Parse a date. It also support lenient parse when input does not match
       * the pattern.<br>
       * We first try to match month a pattern where we have month and weekday 
       * names Ex:  Monday Nov, 11 2013
       * weekday name and month name can be anywhere in the string.
       * if year > 2-digits it can be anywhere in the string. Otherwise we assume
       * its position based on pattern. Separators can be any non digit characters
       * <br>If month name is not present, we try lenient parse yMd and yMEd
       * pattern. Must have year, moth and date all numbers. Ex: 5/3/2013 weekday
       * is optional. If present it must match date. Ex:  Tuesday 11/19/2013
       * if year > 2-digits it can be anywhere in the string. Otherwise assume its
       * position based on pattern if date > 12 it can be anywhere in the string. 
       * Otherwise assume its position based on pattern separators can be any
       * non digit characters.<br><br>
       * @memberOf OraDateTimeConverter
       * @param {string} str - a String to be parsed. it can be an iso 8601 string
       * or a formatted string.
       * @param {Object} localeElements - The instance of LocaleElements bundle
       * @param {Object=} options - Containing the following properties:<br>
       * - <b>weekday.</b> Allowed values: "narrow", "short", "long".<br>
       * - <b>era.</b> Allowed values: "narrow", "short", "long".<br>
       * - <b>year.</b> Allowed values:"2-digit", "numeric".<br>
       * - <b>month.</b> Allowed values: "2-digit", "numeric", "narrow", 
       * "short", "long".<br>
       * - <b>day.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>hour.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>minute.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>second.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>millisecond.</b> Allowed values: "numeric".<br>
       * - <b>timeZone.</b> The possible values of the timeZone property are valid IANA timezone IDs.
       * If the users want to pass an offset, they can use one of the Etc/GMT timezone IDs. 
       *  yet.<br>
       * - <b>timeZoneName.</b> allowed values are "short", "long". </b> <br>
       * - <b>dst</b> is a Boolean value. Setting dst to true indicates the time is in DST. By default the
       * time is interpreted as standard time. The possible values of dst are: "true" or "false". Default is "false".<br>
       * -<b>isoStrFormat.</b>specifies in which format the ISO string is returned. 
       * The possible values of isoStrFormat are: "offset", "zulu", "local", "auto". The default format is auto.<br>
       * - <b>hour12.</b> is a Boolean value indicating whether 12-hour format 
       * (true) or 24-hour format (false) should be used. It is only relevant 
       * when hour is also present.<br>
       * - <b>pattern.</b> custom String pattern as defined by Unicode CLDR.<br>
       * - <b>two-digit-year-start.</b> the 100-year period 2-digit year. 
       * During parsing, two digit years will be placed in the range 
       * two-digit-year-start to two-digit-year-start + 100 years. 
       * The default is 1950.<br>
       * - <b>formatType.</b> a predefined formatting type. Allowed values: 
       * "date", "time", "datetime".<br>
       * - <b>dateFormat.</b> optional, specifies the date format. Allowed 
       * values: "short", "medium", "long", "full". It is only considered when 
       * formatType is present. The default value is "short".<br>
       * - <b>timeFormat.</b> optional, specifies the time format. Allowed 
       * values: "short", "medium", "long", "full". It is only considered when 
       * formatType is present. The default value is "short".<br><br>
       * The order of precedence is the following:<br>
       * 1. pattern.<br>
       * 2. ECMA options.<br>
       * 3. formatType.<br>          
       * If options is ommitted, the default will be the following object:<br>
       * {<br>
       * year:"numeric",<br> 
       * month:"numeric",<br> 
       * day:"numeric"<br>
       * };<br>
       * - <b>lenientParse.</b> specifies if lenient parse is enabled or disabled. Allowed values: "full", "none".
       * default is "full" which means lenient parse is enabled.<br>
       * @param {string=} locale - A BCP47 compliant language tag. it is only 
       * used to extract the unicode "nu" extension key. We currently support "arab", "latn" and "thai" 
       * numbering systems. EX: locale: 'en-US-u-nu-latn' 
       * @return {string} an iso 8601 extended String. http://en.wikipedia.org/wiki/ISO_8601.
       * If the patern is a date only, returns the date part of the iso string.
       * If the pattern is time only, returns the time part of the iso string. 
       * If the pattern is date-time, returns the date-time iso string.
       * <br>Example1:
       * <br> var pattern = 'MM/dd/yy hh:mm:ss a';
       * <br> cnv.parse('09/11/14 03:02:01 PM', localeElems, pattern);
       * <br> The return value is '2014-10-20T15:02:01'; 
       * <br>Example2:
       * <br> var pattern = 'MM/dd/yy';
       * <br> cnv.parse('09/11/14', localeElems, pattern);
       * <br> The return value is '2014-10-20'; 
       * <br>Example3:
       * <br> var pattern = 'hh:mm:ss a';
       * <br> cnv.parse('03:02:01 PM', localeElems, pattern);
       * <br> The return value is 'T15:02:01'; 
       * @throws {RangeError} If a property value of the options parameter is 
       * out of range.
       * @throws {SyntaxError} If an Unexpected token is encountered in the 
       * pattern.
       * @throws {Error} If the <i>str</i> parameter does not match the format 
       * pattern.
       * @throws {RangeError} if one of the date fields is out of range.
       * @throws {invalidISOString} if the string to be parsed is an invalid ISO string.
       */
      parse: function (str, localeElements, options, locale) {
        return _parseImpl(str, localeElements, options, locale);
      },
      /**
       * Resolve options.
       * Returns a new object with properties reflecting the date and time 
       * formatting options computed based on the options parameter. 
       * If the options parameter is ommitted, the following object will be 
       * returned:<br>
       * {<br>
       * calendar: "gregorian"<br>
       * numberingSystem: "latn"<br>
       * locale: &lt;locale parameter&gt;,<br>
       * day: "numeric",<br>
       * month: "numeric",<br>
       * year: "numeric"<br>
       * };
       * @memberOf OraDateTimeConverter
       * @param {Object} localeElements - The instance of LocaleElements bundle                           
       * @param {Object=} options - Containing the following properties:<br>
       * - <b>calendar.</b> The calendar system.<br>
       * - <b>weekday.</b> Allowed values: "narrow", "short", "long".<br>
       * - <b>era.</b> Allowed values: "narrow", "short", "long".<br>
       * - <b>year.</b> Allowed values:"2-digit", "numeric".<br>
       * - <b>month.</b> Allowed values: "2-digit", "numeric", "narrow", 
       * "short", "long".<br>
       * - <b>day.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>hour.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>minute.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>second.</b> Allowed values: "2-digit", "numeric".<br>
       * - <b>millissecond.</b> Allowed values: "numeric".<br>
       * - <b>timeZone.</b> The possible values of the timeZone property are valid IANA timezone IDs.<br>
       * - <b>timeZoneName.</b> allowed values are "short", "long". </b> <br>
       * - <b>dst.</b> is a Boolean value. The possible values of dst are: "true" or "false". Default is "false".<br>
       * - <b>isoStrFormat.</b> specifies in which format the ISO string is returned. 
       * The possible values of isoStrFormat are: "offset", "zulu", "local", "auto". The default format is auto.<br> 
       * - <b>hour12.</b> is a Boolean value indicating whether 12-hour format 
       * (true) or 24-hour format (false) should be used. It is only relevant 
       * when hour is also present.<br>
       * - <b>pattern.</b> custom String pattern as defined by Unicode CLDR. 
       * Will override above options when present.<br>
       * - <b>two-digit-year-start.</b> the 100-year period 2-digit year. 
       * During parsing, two digit years will be placed in the range 
       * two-digit-year-start to two-digit-year-start + 100 years. 
       * The default is 1950.<br>
       * - <b>formatType.</b> predefined format type.  Allowed values: 
       * "datetime", "date", "time"<br>
       * - <b>dateFormat.</b> format of date field.  Allowed values: "short", 
       * "medium", "long", "full". It is only relevant when formatType is also 
       * present<br>
       * - <b>timeFormat.</b> format of time field.  Allowed values: "short", 
       * "medium", "long", "full". It is only relevant when formatType is also 
       * present<br>
       * - <b>lenientParse.</b> specifies if lenient parse is enabled or disabled. Allowed values: "full", "none".
       * default is "full" which means lenient parse is enabled.<br>
       * @param {string=} locale - A BCP47 compliant language tag. it is only 
       * used to extract the unicode extension keys.
       * @return {Object} Resolved options object.
       * @throws {RangeError} If a property value of the options parameter is 
       * out of range.
       * @throws {Error} if weekday does not match the date.
       */

      resolvedOptions: function (localeElements, options, locale) {
        return _resolvedOptionsImpl(localeElements, options, locale);
      },
      /**
       * Returns the current week in the year when provided a date.
       * @memberOf OraDateTimeConverter
       * @param {string} date - iso 8601 string. It may be in extended or 
       * non-extended form. http://en.wikipedia.org/wiki/ISO_8601
       * @return {number}. The current week in the year when provided a date.
       */
      calculateWeek: function (date) {
        var d = oj.OraI18nUtils.isoToLocalDate(date);
        var time;
        var checkDate = new Date(d.getTime());

        // Find Thursday of this week starting on Monday
        checkDate.setDate(checkDate.getDate() + 4 - (checkDate.getDay() || 7));
        time = checkDate.getTime();
        checkDate.setMonth(0);// Compare with Jan 1
        checkDate.setDate(1);
        return Math.floor(Math.round((time - checkDate) / 86400000) / 7) + 1;
      },
      /**
       * Returns the available timeZones.
       * @memberOf OraDateTimeConverter
       * @param {Object} localeElements - The instance of LocaleElements bundle
       * @return {Object}. An array of objects. Each object represents a timezone
       * and contains 2 attributes:<br>
       * 'id': IANA timezone ID<br>
       * 'displayName': It is the concatenation of 3 strings:<br>
       * 1- UTC timezone offset.<br>
       * 2- City name.<br>
       * 3- Generic time zone name
       * The displayName attribute is localized based on the current locale.
       * Example of an array entry in en-US locale:
       * {id: 'America/Edmonton',<br>
       *  displayName: '(UTC-07:00) Edmonton - Mountain Time'<br>
       * }<br>
       * The same array entry is the following in fr-FR locale:
       * {id: 'America/Edmonton',<br>
       *  displayName: '(UTC-07:00) Edmonton - heure des Rocheuses'<br>
       * }<br> 
       * The array is sorted by offsets in ascending order. Within the same offset,
       * the entries by displayName in ascending order.
       */
      getAvailableTimeZones: function (localeElements) {
        return _availableTimeZonesImpl(localeElements)
      },
      /**
       * Compares 2 ISO 8601 strings.
       * @memberOf OraDateTimeConverter
       * @param {string} isoStr1 isoString that may be date, time or date-time and possible timezone info
       * @param {string} isoStr2 isoString that may be date, time or date-time and with possible timezone info
       * @param {Object} localeElements - The instance of LocaleElements bundle                           
       * @returns {number} 0 if isoStr1 is equal to this isoStr2;
       *  a value less than 0 if isoStr1 is before isoStr2; and a value 
       *  greater than 0 if isoStr1 is after isoStr2.
       */
      compareISODates: function (isoStr1, isoStr2, localeElements) {
        var options = _getCompareISODatesOptions(isoStr1, isoStr2);
        var iso1 = _parseImpl(isoStr1, localeElements, options, "en-US");
        var iso2 = _parseImpl(isoStr2, localeElements, options, "en-US");
        var str1 = iso1['value'].replace("Z", "");
        var str2 = iso2['value'].replace("Z", "");
        var datetime1 = oj.OraI18nUtils._IsoStrParts(str1);
        datetime1 = Date.UTC(datetime1[0], datetime1[1] - 1, datetime1[2], datetime1[3], datetime1[4], datetime1[5], datetime1[6]);
        var datetime2 = oj.OraI18nUtils._IsoStrParts(str2);
        datetime2 = Date.UTC(datetime2[0], datetime2[1] - 1, datetime2[2], datetime2[3], datetime2[4], datetime2[5], datetime2[6]);
        return  (datetime1 - datetime2);
      },
      /**
       * returns if a locale is 12 or 24 hour format.
       * @memberOf OraDateTimeConverter
       * @param {Object} localeElements - The instance of LocaleElements bundle 
       * @returns {boolean} true if the locale's preferred hour format is 12, false for 24 hour format.
       * @since 2.2
       */
      isHour12: function (localeElements) {
        return _isHour12(localeElements);
      },
      /**
       * returns time tokens positions
       * @memberOf OraDateTimeConverter
       * @param {Object} localeElements - the instance of LocaleElements bundle.
       * @param {Object=} options - ECMA options or a pattern<br>
       * @param {string=} locale - A BCP47 compliant language tag. it is only 
       * used to extract the unicode extension keys.
       * @return {Object}. time tokens positions. For example  if the pattern is 'hh:mm a'
       * the return value is { 'h': 0, 'm': 1, 'a': 2 }, fot RTL locales the order 
       * of the tokens is reversed for 'h' and 'm' { 'h':1, 'm': 0, 'a': 2 } 
       */
      getTimePositioning: function (localeElements, options, locale) {
        var resOptions = _resolvedOptionsImpl(localeElements, options, locale);
        var pattern = resOptions['pattern'] || resOptions['patternFromOptions'];
        pattern = pattern.replace(/\'[^']*\'/g, "").replace(/[^hHkKma]*/g, "");
        pattern = pattern.replace(/(h|H|k|K)+/, "h").replace(/m+/, "m");
        var mainNodeKey = oj.OraI18nUtils.getLocaleElementsMainNodeKey(localeElements);
        var lang = _getBCP47Lang(mainNodeKey);
        var isRTL = (lang === 'ar') || (lang === 'he');
        var positioning = {};
        var i = 0;
        var c;
        var len = pattern.length;
        for (i = 0; i < len; i++) {
          c = pattern.charAt(i);
          positioning[c] = i;
        }
        if (isRTL) {
          // for RTL locales reverse h and m
          i = positioning['h'];
          positioning['h'] = positioning['m'];
          positioning['m'] = i;
        }
        return positioning;
      }
    };
  }

  return {
    /**
     * getInstance.
     * Returns the singleton instance of OraDateTimeConverter class.  
     * @memberOf OraDateTimeConverter
     * @return {Object} The singleton OraDateTimeConverter instance.
     */
    getInstance: function () {
      if (!instance) {
        instance = _init();
      }
      return instance;
    }
  };
}());

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/** This is a forked version of moment-timezone.js
 * The MIT License (MIT)
 * Copyright (c) 2014 Tim Wood
 * https://github.com/moment/moment-timezone/blob/develop/LICENSE
 */

/*
 DESCRIPTION
 OraTimeZone object implements timeZone support.
 
 PRIVATE CLASSES
 <list of private classes defined - with one-line descriptions>
 
 NOTES
 <other useful comments, qualifications, etc.>
 
 MODIFIED    (MM/DD/YY)
        02/01/15 - Creation
 */



/**
 * @ignore
 */
var OraTimeZone;

OraTimeZone = (function () {

  var _zones = {};
  var instance;
  var _charCodeToInt;
  var _unpackBase60;
  var _arrayToInt;
  var _intToUntil;
  var _mapIndices;
  var _unpack;
  var _packBase60;
  var _packBase60Fraction;
  var  _normalizeName;
  var _getZone;
  var _addZone;
  var _throwInvalidtimeZoneID;
  var _throwNonExistingTime;
  var _throwMissingTimeZoneData;
  var _GMT_REGEXP = /^Etc\/GMT/i;
  var _SECOND = 1000;
  var _MINUTE = 60 * _SECOND;
  var _HOUR = 60 * _MINUTE;
  var _MIN_OFFSET = -14 * 60;
  var _MAX_OFFSET = +12 * 60;

  /************************************
   Unpacking
   ************************************/

  var __BASE60 = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWX';
  var  _EPSILON = 0.000001; // Used to fix floating point rounding errors

  _packBase60Fraction = function (fraction, precision) {
    var buffer = '.',
      output = '',
      current;

    while (precision > 0) {
      precision -= 1;
      fraction *= 60;
      current = Math.floor(fraction + _EPSILON);
      buffer += __BASE60[current];
      fraction -= current;

      // Only add buffer to output once we have a non-zero value.
      // This makes '.000' output '', and '.100' output '.1'
      if (current) {
        output += buffer;
        buffer = '';
      }
    }

    return output;
  };

  _packBase60 = function (number, precision) {
    var output = '';
    var  absolute = Math.abs(number);
    var  whole = Math.floor(absolute);
    var  fraction = _packBase60Fraction(absolute - whole, Math.min(~~precision, 10));

    while (whole > 0) {
      output = __BASE60[whole % 60] + output;
      whole = Math.floor(whole / 60);
    }

    if (number < 0) {
      output = '-' + output;
    }

    if (output && fraction) {
      return output + fraction;
    }

    if (!fraction && output === '-') {
      return '0';
    }

    return output || fraction || '0';
  };

  /************************************
   Unpacking
   ************************************/
  _charCodeToInt = function (charCode) {
    if (charCode > 96) {
      return charCode - 87;
    }
    else if (charCode > 64) {
      return charCode - 29;
    }
    return charCode - 48;
  };

  _unpackBase60 = function (string) {
    var i = 0;
    var parts = string.split('.');
    var whole = parts[0];
    var fractional = parts[1] || '';
    var multiplier = 1;
    var num;
    var out = 0;
    var sign = 1;

    // handle negative numbers
    if (string.charCodeAt(0) === 45) {
      i = 1;
      sign = -1;
    }
    // handle digits before the decimal
    for (; i < whole.length; i++) {
      num = _charCodeToInt(whole.charCodeAt(i));
      out = 60 * out + num;
    }
    // handle digits after the decimal
    for (i = 0; i < fractional.length; i++) {
      multiplier = multiplier / 60;
      num = _charCodeToInt(fractional.charCodeAt(i));
      out += num * multiplier;
    }
    return out * sign;
  };


  _arrayToInt = function (array) {
    for (var i = 0; i < array.length; i++) {
      array[i] = _unpackBase60(array[i]);
    }
  };

  _intToUntil = function (array, length) {
    for (var i = 0; i < length; i++) {
      array[i] = Math.round((array[i - 1] || 0) + (array[i] * _MINUTE)); // minutes to milliseconds
    }

    array[length - 1] = Infinity;
  };

  _mapIndices = function (source, indices) {
    var out = [], i;
    for (i = 0; i < indices.length; i++) {
      out[i] = source[indices[i]];
    }
    return out;
  };

  _unpack = function (id, string) {
    var data = string.split('|');
    var  offsets = data[1].split(' ');
    var  indices = data[2].split('');
    var  untils = data[3].split(' ');

    _arrayToInt(offsets);
    _arrayToInt(indices);
    _arrayToInt(untils);
    _intToUntil(untils, indices.length);
    return {
      name : id,
      abbrs : _mapIndices(data[0].split(' '), indices),
      offsets : _mapIndices(offsets, indices),
      untils : untils
    };
  };

  /************************************
   Exceptions
   ************************************/
  _throwInvalidtimeZoneID = function (str) {
    var msg, error, errorInfo;
    msg = "invalid timeZone ID: " + str;
    error = new Error(msg);
    errorInfo = {
      'errorCode' : 'invalidTimeZoneID',
      'parameterMap' : {
        'timeZoneID' : str
      }
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };

  _throwNonExistingTime = function () {
    var msg, error, errorInfo;
    msg = "The input time does not exist because it falls during the transition to daylight saving time.";
    error = new Error(msg);
    errorInfo = {
      'errorCode' : 'nonExistingTime'
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };
  
  _throwMissingTimeZoneData = function () {
    var msg, error, errorInfo;
    msg = "TimeZone data is missing. Please call require 'ojs/ojtimezonedata' in order to load the TimeZone data.";
    error = new Error(msg);
    errorInfo = {
      'errorCode' : 'missingTimeZoneData'
    };
    error['errorInfo'] = errorInfo;
    throw error;
  };

  /************************************
   Zone object
   ************************************/

  /**
   * @ignore
   * @constructor
   */
  function Zone (name, tzData) {
    var data = tzData['zones'][name];
    //Try  if name matches Etc/GMT offset
    if (_GMT_REGEXP.test(name)) {
      var hours, minutes = 0;
      var offset = name.replace(_GMT_REGEXP, "");
      var parts = offset.split(":");
      hours = parseInt(parts[0], 10) * 60;
      if (isNaN(hours))
        return;
      if (parts.length === 2) {
        minutes = parseInt(parts[1], 10);
        if (isNaN(minutes))
          return;
      }
      hours += (hours >= 0) ? minutes : -minutes;
      //offset must be between -14 and +12
      if (hours < _MIN_OFFSET || hours > _MAX_OFFSET)
        return;
      hours = _packBase60(hours, 1);
      var gmtName = name.replace("/etc\//i", "").toUpperCase();
      data = gmtName + "|" + hours + "|" + "0|";
    }
    if (data !== undefined)
      this._set(_unpack(name, data));
  }

  Zone.prototype = {
    _set : function (unpacked) {
      this.name = unpacked.name;
      this.abbrs = unpacked.abbrs;
      this.untils = unpacked.untils;
      this.offsets = unpacked.offsets;
    },
    parse : function (target, dst, ignoreDst, throwException) {
      var offsets = this.offsets;
      var  untils = this.untils;
      var  max = untils.length - 1;
      var  offset;
      var offset1;
      var i;
      var transitionTime;
      var gapTime;
      var dupTime;
      var until;
      for (i = 0; i < max; i++) {
        offset = offsets[i];
        offset1 = offsets[i + 1];
        until = untils[i];
        transitionTime = until - (offset * _MINUTE);
        gapTime = transitionTime + _HOUR;
        dupTime = transitionTime - _HOUR;
        //Transition to dst:
        //Test if the time falls during the non existing hour when trasition to
        //dst happens. The missing hour is between transitionTime and gapTime.
        //If we are converting from source timezone to target timezone, we do not
        //throw an exception if target timezone falls in non existing window, 
        //we just skip one hour, throwException is passed as false in this scenario.
        if (target >= transitionTime && target < gapTime && offset > offset1) {
          if(throwException === true) {
            _throwNonExistingTime();
          }
          else {
            return (i + 1);
          }
        }
        //Test if the time falls during the duplicate hour when dst ends.
        //The duplicate hour is between dupTime and transitionTime.
        //if dst is set to true, return dst offset.
        if (target >= dupTime && target < transitionTime && offset < offset1) {
          if (dst) {
            return i;
          }
          return (i + 1);
        }
        //Time is outside transtition times.
        if (target < until - (offset * _MINUTE)) {
          if (ignoreDst === false) {
            if (dst) {
              if (offset < offset1) {
                return i;
              }
              return (i + 1);
            }
            else {
              if (offset < offset1) {
                return (i +1);
              }
              return i;
            }
          }
          return i;
        }
      }
      return max;
    },
    //user first need to call pasre to get the index, then pass it to the 
    //2 functions below
    abbr : function (idx) {
      return this.abbrs[idx];
    },
    ofset : function (idx) {
      var len = this.offsets.length;
      if(idx >= 0 && idx < len) {
        return parseInt(this.offsets[idx], 10);
      }  
      return parseInt(this.offsets[len -1], 10);
    },
    len : function () {
      return this.offsets.length;
    }
  };

  /************************************
   timeZOne functions
   ************************************/
  _normalizeName = function (name) {
    return (name || '').toLowerCase().replace(/\//g, '_');
  };

  _addZone = function (name, tzData) {
    var zone, zoneName;
    zone = new Zone(name, tzData);
    zoneName = _normalizeName(zone['name']);
    _zones[zoneName] = zone;
  };

  _getZone = function (name, tzData) {
    var zoneName = _normalizeName(name);
    if (_zones[zoneName] === undefined)
      _addZone(name, tzData);
    return _zones[_normalizeName(name)] || null;
  };


  function _init () {
    return {
      getZone : function (name, localeElements) {
        var tzData = localeElements['supplemental']['timeZoneData'];
        if(tzData === undefined) {
          _throwMissingTimeZoneData();
        }
        var s = _getZone(name, tzData);
        //try the links
        if (!s) {          
          var link = tzData['links'][name];
          if (link) {
            s = _getZone(link, tzData);
          }
        }
        if (!s) {
          _throwInvalidtimeZoneID(name);
        }
        return s;
      }
    };
  }

  return {
    /**
     * getInstance.
     * Returns the singleton instance of OraTimeZone class. 
     * @ignore
     * @memberOf OraTimeZone
     * @return {Object} The singleton OraTimeZone instance.
     */
    getInstance : function () {
      if (!instance) {
        instance = _init();
      }
      return instance;
    }
  };
}());
/**
 * @ignore
 * @constructor
 */
function _KoCustomBindingProvider()
{

  this.install = function()
  {
    var provider = ko.bindingProvider;
    var instance_name = "instance";
    var wrapped = provider[instance_name];

    var getAccessors = 'getBindingAccessors';
    if (!wrapped[getAccessors])
    {
      oj.Logger.error("JET's Knockout bindings are not compatible with the current binding provider since it does " +
                         "not implement getBindingAccessors()");
      return this;
    }

    var custom = provider[instance_name] = {};

    var methodsToWrap = [];
    methodsToWrap.push(getAccessors, "nodeHasBindings", "getBindings", "preprocessNode");

    methodsToWrap.forEach(
      function(name)
      {
        _wrap(wrapped, custom, name);
      }
    );

    return this;

  };

  this.addPostprocessor = function(postprocessor)
  {
    var keys = Object.keys(postprocessor);
    keys.forEach(
      function(key)
      {
        _postprocessors[key] = _postprocessors[key] || [];
        _postprocessors[key].push(postprocessor[key]);
      }
    );
  }

  var _postprocessors = {};

  function _wrap(wrapped, target, name)
  {
    var delegate = wrapped[name];
    target[name] = function()
    {
      var ret = delegate ? delegate.apply(wrapped, arguments) : null;
      var postprocessHandlers = _postprocessors[name];

      if (postprocessHandlers != null)
      {
        var originalArgs = arguments;
        postprocessHandlers.forEach(
          function(handler)
          {
            var args = Array.prototype.slice.call(originalArgs);
            args.push(ret, wrapped);
            ret = handler.apply(null, args);
          }
        );
      }
      return ret;
    }
  }
}

/**
 * @ignore
 * @private
 */
oj.__KO_CUSTOM_BINDING_PROVIDER_INSTANCE = new _KoCustomBindingProvider().install();
/**
 * @private
 * @constructor
 * Global Change Queue Implementation
 * The queue is used to delay component updates until all model changes have been propagated
 * This is a private class that does not need to be xported
 */
function GlobalChangeQueue()
{
  this.Init();
}

// Subclass from oj.Object 
oj.Object.createSubclass(GlobalChangeQueue, oj.Object, "ComponentBinding.GlobalChangeQueue");

GlobalChangeQueue.prototype.Init = function()
{
  GlobalChangeQueue.superclass.Init.call(this);
  this._trackers = [];
  this._queue = [];
};

GlobalChangeQueue.prototype.registerComponentChanges = function(tracker)
{
  if (this._trackers.indexOf(tracker) === -1)
  {
    this._trackers.push(tracker);
    if (!this._delayTimer)
    {
      this._delayTimer = setTimeout(oj.Object.createCallback(this, this._deliverChangesImpl), 1);//@HTMLUpdateOK; delayting our own callback
    }
  }
};


GlobalChangeQueue.prototype.deliverChanges = function()
{
  if (this._delayTimer)
  {
    clearTimeout(this._delayTimer);
  }
  this._deliverChangesImpl();
};

GlobalChangeQueue.prototype._deliverChangesImpl = function()
{
  var i;
  this._delayTimer = null;
  var trackers = this._trackers;
  this._trackers = [];
  
 
  for (i=0; i<trackers.length; i++)
  {
    var tracker = trackers[i];
    this._queue.push({tracker: tracker, changes: tracker.flushChanges()});
  }
  
  while (this._queue.length > 0)
  {
    var record = this._queue.shift();
    record.tracker.applyChanges(record.changes);
  }
};
/*jslint browser: true, devel: true*/
/*global ComponentChangeTracker:true*/


/**
 * To create a custom binding,
 * use oj.ComponentBinding.create(). Using prototypal inheritance to extend
 * oj.ComponentBinding is not supported.
 * @export
 * @class oj.ComponentBinding 
 * @classdesc JQueryUI component binding for Knockout.js. 
 * @param {string|Array.<string>} name - the name of the binding or an
 * array of strings in case the binding needs to be registered under several names
 * @param {?(Object|string)=} options - property object
 * @see oj.ComponentBinding.create
 * @constructor
 * @since 1.0
 * @hideconstructor
 */
oj.ComponentBinding = function(name, options)
{
  this.Init(name, options);
};

oj.Object.createSubclass(oj.ComponentBinding, oj.Object, "oj.ComponentBinding");


/**
 * Creates a binding instance and registers it with Knockout.js
 * @export
 * @param {string|Array.<string>} name - the name of the binding or an
 * array of strings in case the binding needs to be registered under several names
 * @param {?(Object|string)=} options - property object with the following fields:
 * <ul>
 *   <li>{string} 'componentName' - name of the component 
 *       Not specifying this parameter indicates that the binding should use the 'component' attribute 
 *       on itself to determine the constructor name
 *   </li>
 *   <li>{Function} 'afterCreate'- a callback invoked after the component instance has been created.
 *        The function will receive the following parameters:
 *        <ul>
 *           <li>{Element} element - DOM element associated with this binding</li>
 *           <li>{Function} widgetConstructor - constructor for the JQueryUI widget created
 *            by this binding. The constructor is already bound to the associated 
 *            JQuery element</li>
 *           <li>{Function} valueAccessor - a JavaScript function that you can call to 
 *           get a map of current binding attributes</li>
 *           <li>{Object} allBindings -  a JavaScript object that you can use to access all the model values bound 
 *           to this DOM element</li>
 *           <li>{Object} bindingContext -  an object that holds the binding context available to this element's bindings. 
 *           This object includes special properties including $parent, $parents, and $root that can be used to access
 *           data that is bound against ancestors of this context</li>
 *        </ul>
 *   </li>
 *   <li>{Function} 'beforeDestroy'- a callback invoked before the component instance is detroyed
 *        The function will receive the same parameters as the 'afterCreate' callback above.
 *   </li>
 * </ul>
 * When this parameter is specified as a string, it will be interpreted as a single 'componentName' option
 * @return {oj.ComponentBinding} binding instance
 */
oj.ComponentBinding.create = function(name, options)
{
  if (name == null)
  {
    throw "Binding name is required!";
  }
  
  var instance = new oj.ComponentBinding(name, options);
  
  var componentName = instance._getBindingOptions()['componentName'];
  
  var handlers = ko.bindingHandlers, i;
  var names = Array.isArray(name) ? name : [name];
  for (i=0; i<names.length; i++)
  {
    var nm = names[i];
    oj.ComponentBinding._REGISTERED_NAMES.push(nm);
    
    handlers[nm] = instance;
  }
  
  return instance;
};

/**
 * Retrieves the default component binding instance registered with Knockout.js
 * @return {oj.ComponentBinding} default binding instance
 * @export
 */
oj.ComponentBinding.getDefaultInstance = function()
{
  return oj.ComponentBinding._INSTANCE;
};

/**
 * Sets up custom handling for attributes that will be managed by this binding 
 * instance
 * @param {Object} opts - property object with the following fields:
 * <ul>
 *   <li>'attributes' - string array of attribue names</li>
 *   <li>{Function} 'init' - a function called when a managed attribute is initialized.
 *        The function will receive the following parameters:
 *        <ul>
 *           <li>{string} name - attribute name</li>
 *           <li>{Object} value - attribute value</li>
 *           <li>{Element} element - DOM element where this binding is being attached</li>
 *           <li>{Function} widgetConstructor - constructor for the JQueryUI widget created
 *            by this binding. The constructor is already bound to the associated 
 *            JQuery element</li>
 *           <li>{Function} valueAccessor - a JavaScript function that you can call to 
 *           get a map of current binding attributes</li>
 *           <li>{Object} allBindings -  a JavaScript object that you can use to access all the model values bound 
 *           to this DOM element</li>
 *           <li>{Object} bindingContext -  an object that holds the binding context available to this element's bindings. 
 *           This object includes special properties including $parent, $parents, and $root that can be used to access
 *           data that is bound against ancestors of this context</li>
 *        </ul>
 *        The optional return value of the function is a name-to-value map of
 *        properties that should be set on the component
 *   </li>
 *   <li>{Function} 'update' - a function called when a managed attribute is updated.
 *        The function will receive the same parameters as the 'init' callback above.
 *        The optional return value of the function is a name-to-value map of
 *        properties that should be set on the component at the time when other
 *        accumulated changes are delivered
 *   </li>
 *   <li>{Function} 'afterCreate'- a callback invoked after the component instance has been created.
 *        The function will receive the following parameters:
 *        <ul>
 *           <li>{string} name - attribute name</li>
 *           <li>{Element} element - DOM element associated with this binding</li>
 *           <li>{Function} widgetConstructor - constructor for the JQueryUI widget created
 *            by this binding. The constructor is already bound to the associated 
 *            JQuery element</li>
 *           <li>{Function} valueAccessor - a JavaScript function that you can call to 
 *           get a map of current binding attributes</li>
 *           <li>{Object} allBindings -  a JavaScript object that you can use to access all the model values bound 
 *           to this DOM element</li>
 *           <li>{Object} bindingContext -  an object that holds the binding context available to this element's bindings. 
 *           This object includes special properties including $parent, $parents, and $root that can be used to access
 *           data that is bound against ancestors of this context</li>
 *        </ul>
 *   </li>
 *   <li>{string} 'for' (optional) - a string representing component type or constructor 
 *        name restricting the applicability of these managed attributes
 *   </li>
 *   <li>{Array.<string>} 'use' (optional) - an string array of component types whose managed attribute behavior
 *       should be used for the component type specified with the 'for' attribute
 *   </li>
 *   <li>{Function} 'beforeDestroy'- a callback invoked before the component instance is detroyed
 *        The function will receive the same parameters as the 'afterCreate' callback above.
 *   </li>
 * </ul>
 * @return {void}
 * @export
 */
oj.ComponentBinding.prototype.setupManagedAttributes = function(opts)
{
  var forName = opts['for'];
  forName = forName == null ? '@global' : forName;
  
  var managers = this._managedAttrOptions[forName] || [];
  
  managers.push(opts);
  
  this._managedAttrOptions[forName] = managers;
};



/**
 * Delivers all accumulated component changes across all instances of this binding.
 * Calling this method is optional - the changes will be delivered after a 1ms timeout
 * if this method is never invoked. However, you may call this method to speed up
 * component updates when the aplication code is done updating the view models.
 * @return {void}
 * @export
 */
oj.ComponentBinding.deliverChanges = function()
{
  oj.ComponentBinding._changeQueue.deliverChanges();
};

/**
 * @private
 */
oj.ComponentBinding.prototype.Init = function(names, options)
{
  oj.ComponentBinding.superclass.Init.call(this);
  
  if (typeof options === "string")
  {
    options = {'componentName': options};
  }
  
  this._bindingOptions = options || {};
  
  
  this._bindingNames = Array.isArray(names) ? names : [names];
  
  this['init'] = this._init.bind(this);
  this['update'] = this._update.bind(this);
  
  this._managedAttrOptions = {};
};

/**
 * @private
 */
oj.ComponentBinding.prototype._getBindingOptions = function()
{
  return this._bindingOptions
}



/**
 * @private
 */
oj.ComponentBinding.prototype._init = function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) 
{
  //Invoke child bindings first to allow on-the-fly generation of child content
  ko.applyBindingsToDescendants(bindingContext, element);
  
    
  return {'controlsDescendantBindings': true};
};


/**
 * @private
 */
oj.ComponentBinding.prototype._update = function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext)
{
  // an array of ko.computed() that would need cleaning up
  var componentComputeds = [];
  var stage = 0; //init
  var component;
  var changeTracker;
  
  var jelem = $(element);
  
  function cleanup(destroyComponent)
  {
    componentComputeds.forEach(function(computed){ computed['dispose'](); });
    componentComputeds = [];
    
    // when cleanup() is called by the node disposal, there is no need to destroy the component because cleanNode() will
    // do it for us
    if (destroyComponent && component)
    {
      component('destroy');
      component = null;
    }
    if (changeTracker)
    {
      changeTracker.dispose();
      changeTracker = null;
    }
    
    jelem.off(oj.ComponentBinding._HANDLER_NAMESPACE);
  };
  
  function createComponent(componentName, nameOption, bindingValue)
  { 
    if (componentName == null)
    {
      // null component name is legal - we just won't create a components
      return;
    }
  
    var comp = jelem[componentName];
    
    if ((typeof comp) !== "function")
    {
      oj.Logger.error("Component %s is not found", componentName);
      return;
    }
    
    comp = comp.bind(jelem);
    
    var updaterCallback = function(changes) 
    {
      var mutationOptions = oj.ComponentBinding.__removeDotNotationOptions(changes);
    
      comp("option", changes);
      var keys  = Object.keys(mutationOptions);
      
      for (var k=0; k<keys.length; k++)
      {
        var key = keys[k];
        comp("option", key, mutationOptions[key]);
      }
    };
    
    changeTracker = new ComponentChangeTracker(updaterCallback, oj.ComponentBinding._changeQueue);
    
    // Create a compatible valueAccessor for backward compatibility with the managed attributes and custom bindings that
    // expect the ojComponent value to be defined inline. We can use already-resolved binding value here because the component
    // will be recreated whenever the binding value changes, and this method will be called with the new binding value
    
    var compatibleValueAccessor = function(){return bindingValue;};
                                        
    var specifiedOptions = Object.keys(bindingValue).filter(
        function(value)
        {
          return !(value == null || value === nameOption);
        }
    );
    
    
    var componentContext = 
    {
      component: comp,
      changeTracker: changeTracker,
      componentName: componentName,
      specifiedOptions: specifiedOptions,
      computeds: componentComputeds, 
      valueAccessor: compatibleValueAccessor, 
      allBindingsAccessor: allBindingsAccessor, 
      bindingContext: bindingContext,
      destroyCallback: function(){component = null;},
      readOnlyProperties: {}
    };
    
    component = this._initComponent(element, componentContext);
  }
  
  // Since we want the update() method to be invoked by Knockout only once, we are suspending dependency detection
  ko.ignoreDependencies(
    function()
    {
      ko.computed(
        function()
        {
          // invoking valueAccessor() adds a dependency on the bidning value and the possible observable ViewModel
          // to this computed, and unwrapping the observable adds a dependency on the observable binding value
          var bindingValue = ko.utils.unwrapObservable(valueAccessor());
          
          if (typeof bindingValue !== 'object')
          {
            oj.Logger.error("ojComponent binding should evaluate to an object");                
          }
          
          var componentName = this._bindingOptions['componentName'];
          
          var nameOpt;
          var nameOptionFound = false;
          
          if (componentName == null & bindingValue != null)
          {
            var name_attrs = [oj.ComponentBinding._COMPONENT_OPTION, 'role'];
            
            for(var n=0; !nameOptionFound && n<name_attrs.length; n++)
            {
              nameOpt = name_attrs[n];
              if (nameOpt in bindingValue)
              {
                nameOptionFound = true;
                componentName = bindingValue[nameOpt];
              }
            }
            
            if (!nameOptionFound)
            {
              oj.Logger.error("component attribute is required for the ojComponent binding");
            }
            
            componentName = ko.utils.unwrapObservable(componentName);  
          }
          
          
          if (stage == 0)
          {
            stage = 1;
          }
          else
          {
            // Suspend dependency detection for this computed during cleanup()
            ko.ignoreDependencies(cleanup, this, [true]);
          }
          
          // Suspend dependency detection for this computed during createComponent(). Changes to the component options
          // will be tracked separately
          ko.ignoreDependencies(createComponent, this, [componentName, nameOpt, bindingValue]);
        },
        this,
        {'disposeWhenNodeIsRemoved' : element}
      );
    },
    this
  );
    
  // cleanup() called by the node disposal should not be destroying components, as they will be destroyed by .cleanNode()
  ko.utils.domNodeDisposal.addDisposeCallback(element, cleanup.bind(this, false /*detsroyComponent flag*/));
};


/**
 * @private
 */
oj.ComponentBinding.prototype._initComponent = function(element, ctx)
{
  var disposed = false;
  var stage = 0; //init
  
  var specifiedManagedAttrs = {};
  
  var jelem = $(element);
  
  var comp = ctx.component;
  var componentName = ctx.componentName;
  
  var self = this;
  
  var destroyEventType = "_ojDestroy";

      
  var destroyListener = 
    function(evt)
    {
      if (evt.target && evt.target == element)
      {
        // ensure that we do not try to destroy the component again during node cleanup
        ctx.destroyCallback();
        
        var destroyCallback = self._bindingOptions['beforeDestroy'];
        if (destroyCallback)
        {
          destroyCallback(element, comp, ctx.valueAccessor, ctx.allBindingsAccessor, ctx.bindingContext);
        }
        
        oj.ComponentBinding._deliverCreateDestroyEventToManagedProps(false, 
                      specifiedManagedAttrs, element, comp, ctx.valueAccessor, 
                      ctx.allBindingsAccessor, ctx.bindingContext);
        
        disposed = true;
        ctx.changeTracker.dispose();
        element.removeEventListener(destroyEventType, destroyListener);
      }
    };
  
  element.addEventListener(destroyEventType, destroyListener);
      
  var managedAttrMap = oj.ComponentBinding._resolveManagedAttributes(this._managedAttrOptions, ctx.specifiedOptions, componentName);
      
  // Always use managed attribute behavior from the default instance
  var defaultInstance = oj.ComponentBinding.getDefaultInstance();
  if (this !== defaultInstance)
  {
    var defaultManagedMap = defaultInstance._getManagedAttributes(ctx.specifiedOptions, componentName);
    // Override default managed attribute map with values from this binding's map.
    // Note that there is no need to clone defaultManagedMap because a new instance gets created
    // every time _getManagedAttributes() is called
    oj.CollectionUtils.copyInto(defaultManagedMap, managedAttrMap);
    managedAttrMap = defaultManagedMap;
  }
      
  var resolvedInitialOptions = {};
      
  var propertyReadFunc = function()
  {
    var prop = this._property;
   
    var value = oj.ComponentBinding._toJS(ctx.valueAccessor()[prop]);

    
    if (stage === 0) // init, no change
    {
      var managedPropEntry = managedAttrMap[prop];
      if (managedPropEntry != null)
      {
        specifiedManagedAttrs[prop] = managedPropEntry;
        var initFunc = managedPropEntry.init;
        if (initFunc != null)
        {
          var initProps = initFunc(prop, value, element, comp, ctx.valueAccessor, 
                                   ctx.allBindingsAccessor, ctx.bindingContext) || {};
                                   
          var initKeys = Object.keys(initProps);
          
          // ensure that the array properties returned by the init function are cloned 
          for (var ip = 0; ip<initKeys.length; ip++)
          {
            var ikey = initKeys[ip];
            resolvedInitialOptions[ikey] = oj.ComponentBinding.__cloneIfArray(initProps[ikey]);
          }
        }
      }
      else
      {
        resolvedInitialOptions[prop] = oj.ComponentBinding.__cloneIfArray(value);
      }
    }
    // this is a post-init change
    else if (!disposed)
    { 
      if (managedAttrMap[prop] != null)
      {
        var updateFunc = managedAttrMap[prop].update;
        if (updateFunc != null) 
        {
          var updateProps = updateFunc(prop, value, element, comp, ctx.valueAccessor, 
                                        ctx.allBindingsAccessor, ctx.bindingContext) || {};
          
          var updateKeys = Object.keys(updateProps);
          
          for (var k = 0; k<updateKeys.length; k++)            
          {
            var p = updateKeys[k];
            ctx.changeTracker.addChange(p, oj.ComponentBinding.__cloneIfArray(updateProps[p]));
          }
        }
      }
      else if (!ctx.readOnlyProperties[prop]) // ignore chnages to read-only properties
      {
        ctx.changeTracker.addChange(prop, oj.ComponentBinding.__cloneIfArray(value));
      }
    }
  };
      
     
  for (var k=0; k < ctx.specifiedOptions.length; k++)
  {   
    // ko.computed is used to set up dependency tracking for the bindings's attribute
    // Any observable evaulated during the initial invocation of the function is going to be treated as a dependency
    // by Knockout. Once that dependency changes, the fuction below will be called again, in which case we will know
    // to deliver the change
    ctx.computeds.push(
        ko.computed(propertyReadFunc, 
                    {_property: ctx.specifiedOptions[k]}/* 'this' object for the 'read' function*/
                   )
    );
  }
  
  stage = 1; // post-init
  
      
  oj.ComponentBinding._registerWritebacks(jelem, ctx);
      
  var mutationOptions = oj.ComponentBinding.__removeDotNotationOptions(resolvedInitialOptions);
  
  // Initialization of the component happens here
  comp(resolvedInitialOptions);
  
  Object.keys(mutationOptions).forEach(
    function(mo)
    {
      comp('option', mo, mutationOptions[mo]);
    }
  );
  
  var createCallback = this._bindingOptions['afterCreate'];
  if (createCallback)
  {
    createCallback(element, comp, ctx.valueAccessor, ctx.allBindingsAccessor, ctx.bindingContext);
  }
  
  oj.ComponentBinding._deliverCreateDestroyEventToManagedProps(true, specifiedManagedAttrs, 
              element, comp, ctx.valueAccessor, ctx.allBindingsAccessor, ctx.bindingContext);
  
  
  resolvedInitialOptions = null;
  
  return comp;
};



/**
 * @ignore
 */
oj.ComponentBinding.__CreateEvaluator = function(expr)
{
  /*jslint evil:true */
  return new Function("$context", "with($context){with($data||{}){return " + expr + ";}}"); //@HTMLUpdateOK; binding expression evaluation
}

/**
 * @private 
 */
function _extractValueFromChangeEvent(event, eventData) 
{
  var prop = 'value';
  var nameVal = {};
  nameVal[prop] = eventData[prop];
  return nameVal;
};

/**
 * @private 
 */
function _extractOptionChange(ctx, event, eventData)
{
  var nameVal = {};
  var metadata = eventData['optionMetadata'];
  if (metadata)
  {
    var shouldWrite = ("shouldWrite" === metadata['writeback']);
    if (shouldWrite)
    {
      var name = eventData['option'];
      nameVal[name] = eventData['value'];
      if (metadata['readOnly'])
      {
        ctx.readOnlyProperties[name] = true;
      }
    }
  }
  return nameVal;
};

/**
 * @private
 */
oj.ComponentBinding.prototype._getManagedAttributes = function(specifiedOptions, componentName)
{
  return oj.ComponentBinding._resolveManagedAttributes(this._managedAttrOptions, specifiedOptions, componentName);
}

/**
 * @private
 */
oj.ComponentBinding._resolveManagedAttributes = function(optionMap, specifiedOptions, componentName)
{
  var managedAttrMap = {};
  
  var applicableOptions = [];
  
  var attrs_field = 'attributes';
  
  var traverseOptions = function(name, followLinks)
  {
    var managers = optionMap[name];
    if (managers != null)
    {
      for (var n=managers.length-1; n>=0; n--)
      {
        var opt = managers[n];
        
        if (opt[attrs_field] != null)
        {
          applicableOptions.push(opt);
        }
        if (followLinks)
        {
          var parents = opt['use'];
          if (parents != null)
          {
            parents = Array.isArray(parents) ? parents : [parents];
            for (var i=0; i<parents.length; i++)
            {
              traverseOptions(parents[i], true);
            }
          }
        }
      }
    }
  };
  
  traverseOptions(componentName, true);
  
  // If this is a JET component, check managed options registered for the ancestors
  var ojNamespace = 'oj';
  var widgetClass = $[ojNamespace][componentName];
  
  if (widgetClass != null)
  {
    var proto = Object.getPrototypeOf(widgetClass.prototype);
    while (proto != null && ojNamespace === proto['namespace'])
    {
      traverseOptions(proto['widgetName'], true)
      proto = Object.getPrototypeOf(proto);
    }
  }
  
  traverseOptions('@global', false);
  
  if (applicableOptions.length > 0)
  { 
    for (var k=0; k<specifiedOptions.length; k++)
    {
      var attr = specifiedOptions[k];
   
      for (var l=0; l<applicableOptions.length; l++)
      {
        var opts  = applicableOptions[l];
        
        var attributes = opts[attrs_field];
        
        if (attributes.indexOf(attr) >= 0)
        {
          managedAttrMap[attr] = {init: opts['init'], update:opts['update'], 
                                  afterCreate:opts['afterCreate'], beforeDestroy:opts['beforeDestroy']};
          break;
        }
      }
    }
  }
  
  return managedAttrMap;
};

/**
 * @private
 */
oj.ComponentBinding._HANDLER_NAMESPACE = ".oj_ko";


/**
 * @private
 */
oj.ComponentBinding._registerWritebacks = function(jelem, ctx)
{
  var writablePropMap = 
  {
    '^slider$' : [{'event': 'slidechange', 'getter': _extractValueFromChangeEvent}],
    '^oj*': [{'event': 'ojoptionchange', 'getter': _extractOptionChange.bind(undefined, ctx)}]
  };
  
  var cachedWriterFunctionEvaluators = {};
  
  var keys = Object.keys(writablePropMap);
  
  for (var k=0; k < keys.length; k++)
  {
    var pattern = keys[k];
    if (ctx.componentName.match(pattern))
    {
      var eventInfos = writablePropMap[pattern];
      for (var i=0; i<eventInfos.length; i++)
      {
        var info = eventInfos[i];
        
        jelem.on(
          info['event'] + oj.ComponentBinding._HANDLER_NAMESPACE,
          { //JQuery will pass this object as event.data
            getter: info['getter']
          },
          function(evt, data)
          {
            // Ignore optionChange events that bubbled up from child components
            if(evt.target !== jelem[0])
            {
              return;
            }
            
            var nameValues = evt.data.getter(evt, data);       
           
            var accessor = ctx.valueAccessor();
            
            var names = Object.keys(nameValues);
            for (var nm=0; nm<names.length; nm++)
            {
              var name = names[nm];
              ctx.changeTracker.suspend(name);
              try
              {
                if (ctx.specifiedOptions.indexOf(name) >= 0)
                {
                  
                  var optionMap = accessor[oj.ComponentBinding._OPTION_MAP];
                  var expr = optionMap == null ? null: optionMap[name];
                
                  var target = accessor[name];
                  
                  oj.ComponentBinding._writeValueToProperty(name, 
                                                            target, 
                                                            nameValues[name],
                                                            expr,
                                                            ctx.bindingContext,
                                                            cachedWriterFunctionEvaluators);
                }
              }
              finally
              {
                ctx.changeTracker.resume(name);
              }  
            }
            
          }
        );
      }
      break;
    }
  }
      
};



/**
 * @private
 */
oj.ComponentBinding._writeValueToProperty = function(name, target, value,
                                                     propertyExpression,
                                                     bindingContext, 
                                                     cachedWriterFunctionEvaluators)
{  
  if (target == null || !ko.isObservable(target))
  {    
    if (!(name in cachedWriterFunctionEvaluators))
    {
      var inContextWriter = null;
      var writerExpr = oj.__ExpressionUtils.getPropertyWriterExpression(propertyExpression);
      if (writerExpr != null)
      {
        inContextWriter = oj.ComponentBinding.__CreateEvaluator(writerExpr);
      }
      cachedWriterFunctionEvaluators[name] = inContextWriter;
    }
    
    var func = cachedWriterFunctionEvaluators[name];
    
    if (func)
    {
      var writer = func(bindingContext);
      writer(oj.ComponentBinding.__cloneIfArray(value));
    }
  
  }
  else if (ko.isWriteableObservable(target))
  {
    target(oj.ComponentBinding.__cloneIfArray(value));
  }
};


/**
 * @private
 */
oj.ComponentBinding._toJS = function(prop) 
{
  // ko.toJS creates a cloned object for both plain javascript objects and Object subclasses. We need to avoid it
  // for the latter case to ensure that complex Model objects can be used as binding properties without being cloned
  
  prop = ko.utils.unwrapObservable(prop);
  
  if ((Array.isArray(prop) || oj.CollectionUtils.isPlainObject(prop)) && prop['ojConvertToJS'])
  {
    prop = ko.toJS(prop);
  }

  return prop;
};


/**
 * @ignore
 */
oj.ComponentBinding.__cloneIfArray = function(value)
{
  if (Array.isArray(value))
  {
    value = value.slice();
  }
  return value;
}



/**
 * @private
 */
oj.ComponentBinding.__removeDotNotationOptions = function(options)
{
  var mutationOptions = {};
  
  var keys = Object.keys(options);
  
  for (var k=0; k<keys.length; k++)
  {
    var opt = keys[k];
    if (opt.indexOf('.') >= 0)
    {
      mutationOptions[opt] = options[opt];
      delete options[opt];
    }
  }
  
  return mutationOptions;
};


/**
 * @private
 */
oj.ComponentBinding._deliverCreateDestroyEventToManagedProps = function(isCreate, 
                    managedAttrMap, element, comp, valueAccessor, allBindingsAccessor, bindingContext)
{
  var props = Object.keys(managedAttrMap);
  for (var i=0; i<props.length; i++)
  {
    var prop = props[i];
    var entry = managedAttrMap[prop];
    var callback = isCreate ? entry.afterCreate : entry.beforeDestroy;
    
    if (callback)
    {
      callback(prop, element, comp, valueAccessor, allBindingsAccessor, bindingContext);
    }
  }
  
};

/**
 * @private
 */
oj.ComponentBinding._changeQueue = new GlobalChangeQueue();

/**
 * @private
 */
oj.ComponentBinding.__getKnockoutVersion = function() {
    if (oj.__isAmdLoaderPresent() && ko) {
        return ko.version;
    };
    return "";
};

/**
 * @private
 */
oj.ComponentBinding._isNameRegistered = function(bindingName)
{
  return oj.ComponentBinding._REGISTERED_NAMES.indexOf(bindingName) >=0;
}




/**
 * @ignore
 */
oj.ComponentBinding._REGISTERED_NAMES = [];


/**
 * @ignore
 */
oj.ComponentBinding._COMPONENT_OPTION = 'component';

/**
 * @ignore
 */
oj.ComponentBinding._OPTION_MAP = '_ojOptions';



/**
 * Redefine ko.utils.domNodeDisposal.cleanExternalData, so that JET components can avoid removing the wrapper node during
 * KO cleanup
 * @private
 */
 (function()
 {
    var name = 'cleanExternalData';
    var disposal = ko.utils.domNodeDisposal;
    var wrapped = disposal[name];
    
    disposal[name] = function(node) 
    {
      // This module doe snot have an explicit dependency on ojcomponentcore,
      // so check for the oj.Components instance dynamically
      var func = oj.DomUtils ? oj.DomUtils.setInKoCleanExternal: null;
      if (func)
      {
        func(node);
      }
      try
      {
        wrapped(node);
      }
      finally
      {
        if (func)
        {
          func(null);
        }
      }
    };
 
 })();
 
 /**
 * @private
 */
oj.ComponentBinding._INSTANCE = oj.ComponentBinding.create(["ojComponent", "jqueryUI"]);


/**
 * This is added so that we could cleanup any ko references on the element when it is removed.
 * @export
 */
$.widget("oj._ojDetectCleanData", 
{
  options: 
  {
    /**
     * @type {boolean}
     * @default <code class="prettyprint">false</code>
     */
    'cleanParent': false,
  },
  _destroy: function()
  {
    var disposal, cleanExternalData, oldCleanExternal;

    disposal = ko.utils.domNodeDisposal;
    cleanExternalData = "cleanExternalData";

    // need to temporarily short circuit the domNodeDisposal call otherwise
    // the _destroy override would be invoked again 
    oldCleanExternal = disposal[cleanExternalData];
    disposal[cleanExternalData] = function(){};

    try
    {
      // provide the option to clean from the parent node for components like ojdatagrid so that the comment
      // and text nodes aren't memory leaked with ko when remove/_destroy is not called on all node types
      if (this.options['cleanParent'] && this.element[0].parentNode != null)
      {
        ko.cleanNode(this.element[0].parentNode);
      }
      else
      {
        ko.cleanNode(this.element[0]);          
      }
    }
    finally
    {
      disposal[cleanExternalData] = oldCleanExternal;
    }
  }
 }
);
/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleManagedChartAttributes(name, value, bindingContext) {
  if (name === "pieCenter" && value['template']) {
    value['_renderer'] = _getDvtRenderer(bindingContext, value['template']);
  }
  return {'pieCenter': value};
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({
  'attributes': ["pieCenter"],
  'init': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedChartAttributes(name, value, bindingContext);
  },
  'update': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedChartAttributes(name, value, bindingContext);
  },
  'for': 'ojChart'
});

/**
 * Returns a renderer function executes the template specified in the binding attribute.
 * (for example, a knockout template).
 * @param {Object} bindingContext the ko binding context
 * @param {string} template the name of the template
 * @return {Function} the renderer function
 * @private
 */
function _getComboboxOptionRenderer(bindingContext, template)
{
  return function(context)
  {
    var parent, childContext;
    parent = context["parentElement"];

    // runs the template
    childContext = bindingContext["createChildContext"](context["data"], null,
      function(binding)
      {
        binding["$optionContext"] = context;
      }
    );
    ko["renderTemplate"](template, childContext, null, parent);

    // tell the combobox not to do anything
    return null;
  };
}

/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleComboboxManagedAttributes(name, value, bindingContext)
{
  if (name === "optionTemplate" && value !== null)
  {
    // find the option template and creates a renderer
    var template = String(value);
    var optionRenderer = _getComboboxOptionRenderer(bindingContext, template);

    return {"optionRenderer" : optionRenderer};
  }

  return null;
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
  {
    "attributes": ["optionTemplate"],
    "init": function(name, value, element, widgetConstructor, valueAccessor,
      allBindingsAccessor, bindingContext)
    {
      var result = _handleComboboxManagedAttributes(name, value, bindingContext);
      if (result !== null)
        return result;
    },
    "update": function(name, value, element, widgetConstructor, valueAccessor,
      allBindingsAccessor, bindingContext)
    {
      return _handleComboboxManagedAttributes(name, value, bindingContext);
    },
    "for": "ComboboxOptionRenderer"
  });

/**
 * Default declaration for ojCombobox
 */
oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
  {
    "for": "ojCombobox",
    "use": "ComboboxOptionRenderer"
  });

/**
 * Default declaration for ojSelect
 */
oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
  {
    "for": "ojSelect",
    "use": "ComboboxOptionRenderer"
  });

/**
 * Default declaration for ojInputSearch
 */
oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
  {
    "for": "ojInputSearch",
    "use": "ComboboxOptionRenderer"
  });

/**
 * @private
 * @constructor
 * Keeps track of changes for a single component
 */
function ComponentChangeTracker(component, queue)
{
  this.Init(component, queue);
}

// Subclass from oj.Object 
oj.Object.createSubclass(ComponentChangeTracker, oj.Object, "ComponentBinding.ComponentChangeTracker");

/**
 * @param {Function} updateCallback
 * @param {Object} queue
 */
ComponentChangeTracker.prototype.Init = function(updateCallback, queue)
{
  ComponentChangeTracker.superclass.Init.call(this);
  this._updateCallback = updateCallback;
  this._queue = queue;
  this._changes = {};
  this._suspendCountMap = {};
};


ComponentChangeTracker.prototype.addChange = function(property, value)
{
  if (this._isSuspended(property) || this._disposed)
  {
    return;
  }
  this._changes[property] = value;
  this._queue.registerComponentChanges(this);
};

ComponentChangeTracker.prototype.dispose = function()
{
  this._disposed = true;
};

ComponentChangeTracker.prototype.resume = function(option)
{
  var count = this._suspendCountMap[option] || 0;
  count--;
  if (count < 0)
  {
    oj.Logger.error("ComponentChangeTracker suspendCount underflow");
    return;
  }

  if (count == 0)
  {
    delete this._suspendCountMap[option];
  }
  else
  {
    this._suspendCountMap[option] = count;
  }
};

ComponentChangeTracker.prototype.suspend = function(option)
{
  var count = this._suspendCountMap[option] || 0;
  this._suspendCountMap[option] = count + 1;
};

ComponentChangeTracker.prototype.applyChanges = function(changes)
{
  if (!this._disposed)
  {
    this._updateCallback(changes);
  }
};

ComponentChangeTracker.prototype.flushChanges = function()
{
  var changes = this._changes;
  this._changes = {};
  return changes;
};


ComponentChangeTracker.prototype._isSuspended = function(option)
{
  var count = this._suspendCountMap[option] || 0;
  return (count >= 1);
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true, devel: true*/

// TODO: do we have JSDoc / API doc for bindings?  (Latest answer: no for now, just doc it briefly in baseComponent's contextMenu option for now.)
// TODO: split up init and update so get from DOM on init only, and update only sets it on DOM.  That way, 
//       can update observable to null, without having to additionally clear DOM attr to avoid having it restored from DOM attr.
//       Vet with Max first.
// TODO: keep binding and DOM in synch, a la disabled option in JQUI, similar to todo for contextMenu feature on JET base class.
// TODO: share code with baseComponent._SetupContextMenu?  Should this have any of the configurability of that method?
//       where would shared code live?
ko.bindingHandlers['ojContextMenu'] = 
{
  'update': function (element, valueAccessor, allBindings, viewModel, bindingContext) 
  {
    // 1) declare vars including functions: ---

    var eventNamespace = ".ojContextMenu";
    var $element = $(element);
    
    var pressHoldTimer;
    var pressHoldThreshold = 750; // per UX spec.  Point of reference: JQ Mobile uses 750ms by default.
    var isPressHold = false; // to prevent pressHold from generating a click

    var touchInProgress = false;

    var doubleOpenTimer; // to prevent double open.  see usage below.
    var doubleOpenThreshold = 300; // made up this number.  TBD: Tweak as needed to make all platforms happy.
    var doubleOpenType = null; // "touchstart" or "contextmenu"
    
    // At least some of the time, the pressHold gesture also fires a click event same as a short tap.  Prevent that here.
    var clickListener = function( event ) {
        if (isPressHold) {
            // For Mobile Safari capture phase at least, returning false doesn't work; must use pD() and sP() explicitly.
            // Since it's wonky, do both for good measure.
            event.preventDefault();
            event.stopPropagation();
            isPressHold = false;
            return false;
        }
    };

    //, on Chrome preventDefault on "keyup" will avoid triggering contextmenu event 
    //which will display native contextmenu.This also need to be added on document as event target is not menu launcher.
    var preventKeyUpEventIfMenuOpen = function(event){
      if (event.which == 121 && event.shiftKey) // Shift-F10
      {
        if (getContextMenuNode().is(":visible"))
          event.preventDefault();
      }
    };

    // lazily get the Menu component in case it is inited after the ojContextMenu binding is applied to launcher.
    var getContextMenu = function()
    {
      var constructor = oj.Components.__GetWidgetConstructor(getContextMenuNode()[0], "ojMenu");

      // The JET Menu of the first element found.
      // Per architect discussion, get it every time rather than caching the menu.
      var contextMenu = constructor && constructor("instance");

      // if no element found, or if element has no JET Menu
      if (!contextMenu)
        throw new Error('ojContextMenu binding bound to "' + (menuId ? menuId : menuSelector) + '", which does not reference a valid JET Menu.');

      if (!contextMenuListenerSet) {
        // must use "on" syntax rather than clobbering whatever "close" handler the app may have set via the menu's "option" syntax
        contextMenu.widget().on( "ojclose" + eventNamespace , function( event, ui ) {
          document.removeEventListener("keyup", preventKeyUpEventIfMenuOpen);
        });

        contextMenuListenerSet = true;
      }

      return contextMenu;
    };
    
    // All gets of the menu element must use this function, rather than calling $(menuSelector) directly, 
    // in case menuId was provided instead.
    var getContextMenuNode = function()
    {
      return _getContextMenuNode(menuSelector, menuId);
    };

    var _getContextMenuNode = function(selector, id)
    {
      return (id)
        ? $(document.getElementById(id)) // Do NOT use $('#' + id), due to escaping issues
        : $(selector).first();
    };

    var launch = function(event, eventType, pressHold) {
        // ensure that pressHold doesn't result in a click.  Set this before the bailouts below.
        isPressHold = pressHold;

        var menu = getContextMenu();

        // In Mobile Safari only, mousedown fires *after* the touchend, which causes at least 2 problems:
        // 1) CM launches after 750ms (good), then disappears when lift finger (bad), because touchend --> 
        // mousedown, which calls Menu's "clikAway" mousedown listener, which dismisses Menu.  
        // 2) The isPressHold logic needs to reset the isPressHold ivar on any event that can start a click, 
        // including mousedown.  This problem causes the mousedown listener to incorrectly clear the ivar 
        // after a pressHold, which broke the whole mechanism.
        // SOLUTION FOR 1-2:  On each launch (at 750ms), set a one-time touchend listener that will set a 
        // var and clear it 50ms later.  While the var is set, both mousedown listeners can disregard the 
        // mousedown.  Make the var a static var in Menu, since Menu's listener is static, and since this 
        // launcher component can get/set it via an (effectively static) menu method.
        // NON-SOLUTIONS:  Cancelling touchstart or touchend, via pD() and sP(), doesn't cancel iPad's mousedown.  
        // Cancelling mousedown from here doesn't work even if capture phase, since ojMenu's listener is capture phase.
        // TIMING: The following block should be before the doubleOpen bailout.
        if (isPressHold)
        {
            $element.one( "touchend" + eventNamespace, function( event ) {
                var touchendMousedownThreshold = 50; // 50ms.  Make as small as possible to prevent unwanted side effects.
                menu.__contextMenuPressHoldJustEnded(true);
                setTimeout(function() { 
                    menu.__contextMenuPressHoldJustEnded(false);
                }, touchendMousedownThreshold); //@HTMLUpdateOK; delaying our own callback
            });
        }

        // On platforms like Android Chrome where long presses already fire the contextmenu event, the pressHold 
        // logic causes the menu to open twice, once for the pressHold, once for the contextmenu.  There's no 
        // guarantee which will happen first, but as long as they happen within doubleOpenThreshold ms 
        // of each other, this logic should prevent the double open.
        // Note: Another option is a platform-specific solution where we only use pressHold for platforms that need 
        // it (that don't already fire a contextmenu event for pressHold), but architectural preference is to avoid 
        // platform-specific solutions if possible.
        if ((doubleOpenType === "touchstart" && event.type === "contextmenu")
                || (doubleOpenType === "contextmenu" && event.type === "touchstart")) 
        {
            doubleOpenType = null;
            clearTimeout(doubleOpenTimer);
            return;
        }
        
        // if a nested element or component already showed a JET context menu for this event, don't replace it with ours.
        if (event.isDefaultPrevented()) 
          return;
        
        var position = {
          "mouse": {
            "my": "start top", 
            "at": "start bottom", 
            "of": event 
          }, 
          "touch": {
            "my": "start>40 center", 
            "at": "start bottom", 
            "of": event,
            "collision": "flipfit"
          }, 
          "keyboard": {
            "my": "start top", 
            "at": "start bottom", 
            "of": "launcher"
          }
        };

        var openOptions = {"launcher": $element, "initialFocus": "menu", "position": position[eventType]};

        menu.__openingContextMenu = true; // Hack.  See todo on this ivar in Menu.open().
        menu.open(event, openOptions);
        menu.__openingContextMenu = false;
        
        // if the launch wasn't cancelled by a beforeOpen listener
        if (menu.widget().is(":visible"))
        {  
          event.preventDefault(); // don't show native context menu
          document.addEventListener("keyup", preventKeyUpEventIfMenuOpen);
          
          // see double-open comments above
          if (event.type === "touchstart" || event.type === "contextmenu")
          {
            doubleOpenType = event.type;
            doubleOpenTimer = setTimeout(function(){ 
              doubleOpenType = null;
            }, doubleOpenThreshold); //@HTMLUpdateOK; delaying our own callback
          }
        }
    };


    // 2) Clean up from previously bound value, if any: ---
    
    $element
      .off( eventNamespace )
      .removeClass("oj-menu-context-menu-launcher")
      [0].removeEventListener("click", clickListener, true);
    
    clearTimeout(pressHoldTimer);
    
    var oldMenuData = $element.data("_ojLastContextMenu");

    // If binding's bound value is changing at RT, remove listener from old menu.
    // To avoid memory leak (capture of old menu node), don't assign the node to a var.
    if (oldMenuData) 
      _getContextMenuNode(oldMenuData.selector, oldMenuData.id).off( eventNamespace );
    
    var contextMenuListenerSet = false;


    // 3) Get menu selector/id: ---

    // Accepts anything $() accepts, i.e. same things component CM option accepts,
    // including selector like "#myMenuId", element, JQ object, etc. Also accepts 
    // {}, meaning "use the HTML contextmenu attribute.  Else malformed.
    var menuSelector = ko.utils.unwrapObservable(valueAccessor());

    // if menuSelector is {}, use the id from the contextmenu attr instead. See getContextMenuNode comments.
    var menuId = $.isPlainObject(menuSelector) ? element.getAttribute("contextmenu") : null;

    // Remember how to find menu, without caching menu itself, so that if binding's bound observable is later 
    // changed to point to some other menu, we can clean up the old menu.
    // If/when KO binding is disposed, .data data is discarded, so no leak.
    $element.data("_ojLastContextMenu", {selector: menuSelector, id: menuId});


    // 4) Add listeners to element having context menu (not menu element) : ---

    // Use capture phase to make sure we cancel it before any regular bubble listeners hear it.
    element.addEventListener("click", clickListener, true);

    $element
        .on( "touchstart" + eventNamespace + " " + 
             "mousedown" + eventNamespace + " " +  
             "keydown" + eventNamespace + " ", function( event ) {
            // for mousedown-after-touchend Mobile Safari issue explained above where __contextMenuPressHoldJustEnded is set.
            if (event.type === "mousedown" && getContextMenu().__contextMenuPressHoldJustEnded())
                return;

            // reset isPressHold flag for all events that can start a click.  
            isPressHold = false;

            // start a pressHold timer on touchstart.  If not cancelled before 750ms by touchend/etc., will launch the CM.
            if (event.type === "touchstart") {
                touchInProgress = true;
                pressHoldTimer = setTimeout(launch.bind(undefined, event, "touch", true), pressHoldThreshold);//@HTMLUpdateOK; delaying our own callback
            }

            return true;
        })

        // if the touch ends before the 750ms is up, it's not a long enough tap-and-hold to show the CM
        .on( "touchend" + eventNamespace + " " + 
             "touchcancel" + eventNamespace, function( event ) {
            touchInProgress = false;
            clearTimeout(pressHoldTimer);
            return true;
        })
        .on( "keydown" + eventNamespace + " " + 
             "contextmenu" + eventNamespace, function( event ) {
            if (event.type === "contextmenu" // right-click.  pressHold for Android but not iOS
                    || (event.which == 121 && event.shiftKey)) // Shift-F10
            {
                var eventType = touchInProgress ? "touch" : event.type === "keydown" ? "keyboard" : "mouse";
                launch(event, eventType, false);
            }

            return true;
        })

        // Does 2 things: 
        // 1) Prevents native context menu / callout from appearing in Mobile Safari.  E.g. for links, native CM has "Open in New Tab".
        // 2) In Mobile Safari and Android Chrome, prevents pressHold from selecting the text and showing the selection handles and (in Safari) the Copy/Define callout.
        // In UX discussion, we decided to prevent both of these things for now.  App can theme directly if they have specific needs.
        // Per discussion with architects, do #2 only for touch devices, so that text selection isn't prevented on desktop.  Since #1 
        // is a no-op for non-touch, we can accomplish this by omitting the entire style class, which does 1 and 2, for non-touch.
        // Per comments in scss file, the suppression of 1 and 2 has issues in old versions of Mobile Safari.
        .addClass(oj.DomUtils.isTouchSupported() ? "oj-menu-context-menu-launcher" : "");
  }
};

/**
 * Returns a header renderer function executes the template specified in the binding attribute.
 * (for example, a knockout template).
 * @param {Object} bindingContext the ko binding context
 * @param {Function|string} template the name of the template
 * @return {Function} the renderer function
 * @private
 */
function _getDataGridHeaderRenderer(bindingContext, template)
{
    var nodeStore = {};
    return function(context)
    {
        var parent, childContext;

        parent = context['parentElement'];
        // runs the template
        childContext = bindingContext['createChildContext'](context['data'], null, 
                           function(binding)
                           {
                               binding['$key'] = context['key'];
                               binding['$metadata'] = context['metadata'];
                               binding['$headerContext'] = context;
                           }
                       );
                       
        _executeTemplate(context, template, parent, childContext, nodeStore);

        // tell the datagrid not to do anything
        return null;
    };
};

/**
 * Returns a cell renderer function executes the template specified in the binding attribute.
 * (for example, a knockout template).
 * @param {Object} bindingContext the ko binding context
 * @param {Function|string} template the name of the template
 * @return {Function} the renderer function
 * @private
 */
function _getDataGridCellRenderer(bindingContext, template)
{
    var nodeStore = {};
    return function(context)
    {
        var parent, childContext;

        parent = context['parentElement'];
        // runs the template
        childContext = bindingContext['createChildContext'](context['data'], null, 
                           function(binding)
                           {
                               binding['$keys'] = context['keys'];
                               binding['$metadata'] = context['metadata'];
                               binding['$cellContext'] = context;
                               binding['$cell'] = context['cell'];
                           }
                       );

        _executeTemplate(context, template, parent, childContext, nodeStore);
        
        // tell the datagrid not to do anything
        return null;
    };
};


/**
 * Execute a template and applies relevant afterRender function
 * @param {Object} context the ko binding context
 * @param {Function|string} template the template function or string
 * @param {Element} parent the parent element of the binding
 * @param {Object} childContext the context applied to binding
 * @param {Object} nodeStore place to store the nodeArray
 * @private
 */
function _executeTemplate(context, template, parent, childContext, nodeStore)
{
    var nodesArray, i, useTemplate;
    
    useTemplate = _resolveDataGridTemplate(template, context);
    
    nodesArray = _getClonedNodeArray(useTemplate, nodeStore);
    
    ko.virtualElements.setDomNodeChildren(parent, nodesArray);
    ko.applyBindingsToDescendants(childContext, parent);
    
    for (i = 0; i < parent.childNodes.length; i++)
    {
        if (parent.childNodes[i].nodeType == 1)
        {
            $(parent.childNodes[i])['_ojDetectCleanData']({'cleanParent': true});
            break;
        }
    }
};

/**
 * Get an array of nodes from the template or cached template nodes
 * @param {string} useTemplate the template function or string
 * @param {Object} nodeStore place to store the nodeArray
 * @returns {Array} cloned node array from the template or node store
 * @private
 */
function _getClonedNodeArray(useTemplate, nodeStore)
{
    var nodesArray = nodeStore[useTemplate];
    if (nodesArray == null)
    {
        nodesArray = ko.utils.parseHtmlFragment(document.getElementById(useTemplate).innerHTML);
        nodeStore[useTemplate] = nodesArray;
    }
    
    var newNodesArray = nodesArray.map(function(obj){
        return obj.cloneNode(true);
    });
    
    return newNodesArray;
};

/**
 * If they specify a function for the template get the string of the template by
 * calling the function and passing the cell or header context
 * @param {Function|string} template
 * @param {Object} context
 * @private
 */
function _resolveDataGridTemplate(template, context)
{
    if (typeof template === 'function')
    {
        return template(context);
    }
    return template;
};

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
{
  'attributes': ["header", "cell"],
  'init': function(name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext)
  {
    var row, rowTemplate, column, columnTemplate, cellTemplate, rowEnd, rowEndTemplate, columnEnd, columnEndTemplate;
    if (name === "header")
    {
        // find row template and creates a renderer
        row = value['row'];
        if (row != null)
        {
            rowTemplate = row['template'];
            if (rowTemplate != null)
            {
                row['renderer'] = _getDataGridHeaderRenderer(bindingContext, rowTemplate);
            }
        }

        // find column template and creates a renderer
        column = value['column'];
        if (column != null)
        {
            columnTemplate = column['template'];
            if (columnTemplate != null)
            {
                column['renderer'] = _getDataGridHeaderRenderer(bindingContext, columnTemplate);
            }
        }
        
        // find column template and creates a renderer
        rowEnd = value['rowEnd'];
        if (rowEnd != null)
        {
            rowEndTemplate = rowEnd['template'];
            if (rowEndTemplate != null)
            {
                rowEnd['renderer'] = _getDataGridHeaderRenderer(bindingContext, rowEndTemplate);
            }
        }
        
        // find column template and creates a renderer
        columnEnd = value['columnEnd'];
        if (columnEnd != null)
        {
            columnEndTemplate = columnEnd['template'];
            if (columnEndTemplate != null)
            {
                columnEnd['renderer'] = _getDataGridHeaderRenderer(bindingContext, columnEndTemplate);
            }
        }        
        return {'header': value};
    }
    else if (name === "cell")
    {
        // find the cell template and creates a renderer
        cellTemplate = value['template'];
        if (cellTemplate != null)
        {
            value['renderer'] = _getDataGridCellRenderer(bindingContext, cellTemplate);
        }
        return {'cell': value};
    }
  },
  'update': function(name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext)
  {
    var row, rowTemplate, column, columnTemplate, cellTemplate, rowEnd, rowEndTemplate, columnEnd, columnEndTemplate;
    if (name === "header")
    {
        // find row template and creates a renderer
        row = value['row'];
        if (row != null)
        {
            rowTemplate = row['template'];
            if (rowTemplate != null)
            {
                row['renderer'] = _getDataGridHeaderRenderer(bindingContext, rowTemplate);
            }
        }

        // find column template and creates a renderer
        column = value['column'];
        if (column != null)
        {
            columnTemplate = column['template'];
            if (columnTemplate != null)
            {
                column['renderer'] = _getDataGridHeaderRenderer(bindingContext, columnTemplate);
            }
        }
        
        // find column template and creates a renderer
        rowEnd = value['rowEnd'];
        if (rowEnd != null)
        {
            rowEndTemplate = rowEnd['template'];
            if (rowEndTemplate != null)
            {
                rowEnd['renderer'] = _getDataGridHeaderRenderer(bindingContext, rowEndTemplate);
            }
        }
        
        // find column template and creates a renderer
        columnEnd = value['columnEnd'];
        if (columnEnd != null)
        {
            columnEndTemplate = columnEnd['template'];
            if (columnEndTemplate != null)
            {
                columnEnd['renderer'] = _getDataGridHeaderRenderer(bindingContext, columnEndTemplate);
            }
        }        
        return {'header': value};
    }
    else if (name === "cell")
    {
        // find the cell template and creates a renderer
        cellTemplate = value['template'];
        if (cellTemplate != null)
        {
            value['renderer'] = _getDataGridCellRenderer(bindingContext, cellTemplate);
        }
        return {'cell': value};
    }
    return null;
  },
      
  'for': 'ojDataGrid'
});

function _handleManagedDiagramAttributes(name, value, bindingContext) {
  if (name === "template") {
    return {'_templateFunction': _getDvtDataRenderer(bindingContext, value)};
  }
  return null;
};

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({
  'attributes': ['template'],
  'init': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedDiagramAttributes(name, value, bindingContext);
  },
  'update': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedDiagramAttributes(name, value, bindingContext);
  },
  'for': 'ojDiagram'
});
(function()
{

  /**
   * @ignore
   * @private
   */
  oj.__ExpressionUtils = {};
  
  oj.__ExpressionUtils.getPropertyWriterExpression = function(expression)
  { 
    var reserveddWords = ["true", "false", "null", "undefined"];
    
    if (expression == null || reserveddWords.indexOf(expression) >= 0)
    {
      return null;
    }
    
    // Remove the white space on both ends to ensure that the _ASSIGNMENT_TARGET_EXP regexp
    // is matched properly
    expression = expression.trim();
  
    // Matches something that can be assigned to--either an isolated identifier or something ending with a property accessor
    // This is designed to be simple and avoid false negatives, but could produce false positives (e.g., a+b.c).
    // This also will not properly handle nested brackets (e.g., obj1[obj2['prop']];).
    var match = expression.match(_ASSIGNMENT_TARGET_EXP);
    
    if (match === null)
    {
      return null;
    }
    
    var target = match[1] ? ('Object(' + match[1] + ')' + match[2]) : expression;
    
    return 'function(v){' + target + '=v;}';
  };
  
  var _ASSIGNMENT_TARGET_EXP = /^(?:[$_a-z][$\w]*|(.+)(\.\s*[$_a-z][$\w]*|\[.+\]))$/i;
})();


/**
 * @ignore
 */
oj.__KO_CUSTOM_BINDING_PROVIDER_INSTANCE.addPostprocessor({'getBindingAccessors': _replaceComponentBindingWithV2});



/**
 * This function modifies the return value of getBindingAccessors()
 * to separate bindings for each tracked attributess in the 'inline' ojComponent binding
 * @param node
 * @param bindingContext
 * @param accessorMap
 * @param wrapped
 * @ignore
 */
function _replaceComponentBindingWithV2(node, bindingContext, accessorMap, wrapped)
{
  if (accessorMap == null)
  {
    return null;
  }

  // Remove the old (pre-V2) binding from the accessor map
  var bindingName = _findOwnBinding(accessorMap)
  if (bindingName != null)
  {
    accessorMap = _modifyOjComponentBinding(node, bindingName, wrapped, bindingContext, accessorMap);
  }

  return accessorMap;
}

/**
 * @ignore
 */
function _findOwnBinding(accessorMap)
{
  var keys = Object.keys(accessorMap)
  for (var i=-0; i<keys.length; i++)
  {
    var key = keys[i];
    if (oj.ComponentBinding._isNameRegistered(key))
    {
      return key;
    }
  }
  return null;
}

/**
 * @param node
 * @param bindingName
 * @param wrapped
 * @param bindingContext
 * @param accessorMap
 * @ignore
 */
function _modifyOjComponentBinding(node, bindingName, wrapped, bindingContext, accessorMap)
{
  var info = _getBindingValueInfo(node, bindingName, wrapped, bindingContext);

  var bindingList = info.attrList;

  if (bindingList == null) // binding is specified externally (as opposed to an 'inline' object literal)
  {
    return accessorMap;
  }

  var bindingMap = {};
  _keyValueArrayForEach(bindingList,
    function(key, value)
    {
      bindingMap[key] = value;
    }
  );

  // clone the original accessor map
  accessorMap = oj.CollectionUtils.copyInto({}, accessorMap);

  //Add accessor for the V2 version of the Component bidning
  accessorMap[bindingName] =
                      _getOjComponent2BindingAccessor(bindingContext, bindingMap, info.bindingExpr);

  return accessorMap;

}

/**
 * @ignore
 */
function _getOjComponent2BindingAccessor(bindingContext, attributeMap, bindingExpr)
{
  var accessorFunc = function()
  {
    var accessor = {};
    Object.keys(attributeMap).forEach(
      function(option)
      {
        var expression = attributeMap[option];

        // bindingContext will be passed as as the first parameter to the evaluator
        var getter = oj.ComponentBinding.__CreateEvaluator(expression).bind(null, bindingContext);

        Object.defineProperty(accessor, option,
          {
            'get': getter,
            'enumerable': true
          }
        );
      }
    );

    Object.defineProperty(accessor, oj.ComponentBinding._OPTION_MAP,
      {
        'value': attributeMap
        /* not enumerable */
      }
    );

    return accessor;
  }

  // Define toString() for the custom accessor function since we do not want the entire function body to show up in
  // the log whenever a binding evaluation error occurs
  accessorFunc.toString = function(){return bindingExpr;};

  return accessorFunc;
}


/**
 * @param node
 * @param wrapped
 * @param bindingContext
 * @ignore
 */
function _getBindingString(node, wrapped, bindingContext)
{
  var func = wrapped['getBindingsString'];
  if (func)
  {
    return func.call(wrapped, node, bindingContext);
  }

  switch (node.nodeType)
  {
    case 1: // Element
      return node.getAttribute("data-bind");

    case 8: // Comment node
      var match = node.nodeValue.match(/^\s*ko(?:\s+([\s\S]+))?\s*$/);
      return  (match ? match[1]: null);
  }
  return null;
}

/**
 * @param node
 * @param bindingName
 * @param wrapped
 * @param bindingContext
 * @ignore
 */
function _getBindingValueInfo(node, bindingName, wrapped, bindingContext)
{
  var list = null;

  var bindingString = _getBindingString(node, wrapped, bindingContext);
  var keyValueArray = ko.jsonExpressionRewriting.parseObjectLiteral(bindingString);

  var selfVal = null;

  _keyValueArrayForEach(keyValueArray,
    function(key, value)
    {
      if (key === bindingName)
      {
        selfVal = value;
        return true;
      }
      return false;
    }
  );

  if (selfVal != null)
  {
    // check for object literal
    if (selfVal.indexOf('{') === 0)
    {
      list = ko.jsonExpressionRewriting.parseObjectLiteral(selfVal);
    }
  }

  return {attrList: list, bindingExpr: selfVal};
}

function _keyValueArrayForEach(array, callback)
{
  for(var i=0; i<array.length; i++)
  {
    var entry = array[i];
    var key = entry['key'];
    var value = entry['value'];
    if (key != null && value != null && callback(key.trim(), value.trim()))
    {
      break;
    }
  }
}


oj.__KO_CUSTOM_BINDING_PROVIDER_INSTANCE.addPostprocessor(
  {
    'nodeHasBindings': function(node, wrappedReturn)
    {
      if (!oj.BaseCustomElementBridge)
        return wrappedReturn;
      return wrappedReturn || (node.nodeType === 1 && oj.BaseCustomElementBridge.getRegistered(node.nodeName));
    },

    'getBindingAccessors': function(node, bindingContext, wrappedReturn)
    {
      if (node.nodeType === 1 && oj.BaseCustomElementBridge)
      {
        var name = node.nodeName;
        var info = oj.BaseCustomElementBridge.getRegistered(name);
        
        if (info)
        {
          wrappedReturn = wrappedReturn || {};

          wrappedReturn['_ojCustomElement'] = function() {return {composite: info.composite}}

        }
      }

      return wrappedReturn;
    }
  }
);

/**
 * @ignore
 */
(function()
{
  ko['bindingHandlers']['_ojCustomElement'] =
  {
    'after': ['attr'], // Ensure attr binding is processed first so to handle :disabled case since we only process on init
    'init': function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext)
    {
       // Apply child bindings first
       ko.applyBindingsToDescendants(bindingContext, element);
       
       return {'controlsDescendantBindings' : true};
    },
    
    'update': function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext)
    {
      var callbackId = 0;
      var _expressionHandler;
      var attributeListener;
      
      // This instance won't be created unless oj.BaseCustomElementBridge is defined so this call is safe
      var bridge = oj.BaseCustomElementBridge.getInstance(element);
      
      function cleanup()
      {
        if (_expressionHandler)
        {
          _expressionHandler.teardown();
          _expressionHandler = null;
        }
        
       
        if(attributeListener)
        {
          element.removeEventListener(_ATTRIBUTE_CHANGED, attributeListener);
          attributeListener = null;
        }
      };
      
      function setup(isComposite)
      {
        var metadataProps = oj.BaseCustomElementBridge.getProperties(bridge, element);
        _expressionHandler = new oj.__ExpressionPropertyUpdater(element, bindingContext, isComposite);

        // Dummy metadata for passing to the ExpressionPropertyUpdater for DOM listener attributes i.e. on-*
        var domListenerMetadata = {_domListener:true};
        
        // Set a flag on the bridge to indicate that we are initializing expressions from the DOM
        // to avoid overriding any property sets that could have occured after
        bridge.__INITIALIZING_PROPS = true;

        // setupExpression will only update properties defined in metadata so it's safe to iterate through all element attributes
        // including ones defined on the base HTML prototype
        var attrs = element.attributes; // attrs is a NamedNodeMap
        for (var i = 0; i < attrs.length; i++)
        {
          var attr = attrs[i];
          var propName = oj.__AttributeUtils.attributeToPropertyName(attr.nodeName);
          var eventName = oj.__AttributeUtils.eventListenerPropertyToEventType(propName);
          var isDomEvent = eventName && !metadataProps[propName];
          _expressionHandler.setupExpression(attr.value, propName, isDomEvent ? domListenerMetadata : metadataProps[propName.split('.')[0]]);
        }
        
        bridge.__INITIALIZING_PROPS = false;

        attributeListener = function(evt)
        {
          var detail = evt['detail'];
          var attr = detail['attribute'];
          var propName = oj.__AttributeUtils.attributeToPropertyName(attr);
          var eventName = oj.__AttributeUtils.eventListenerPropertyToEventType(propName);
          var isDomEvent = eventName && !metadataProps[propName];
          
          var metadata = isDomEvent ? domListenerMetadata : metadataProps[propName.split('.')[0]]; // send metadata for top level property
          _expressionHandler.setupExpression(detail['value'], propName, metadata);
         
        };
            
        element.addEventListener(_ATTRIBUTE_CHANGED, attributeListener);
      };
  
      // Since we are tracking all our dependencies explicitly, we are suspending dependency detection here.
      // update() will be called only once as a result
      ko.ignoreDependencies(
        function()
        {
          ko.computed(
            function()
            {
              // Access valueAccesor to ensure that the binding is re-initialzied when an
              // observable ViewModel is mutated
              var isComposite = valueAccessor().composite;
              
              if (!ko.computedContext.isInitial()) 
              {
                cleanup();
              }
              
              callbackId++;
              
              
              function _createCallbackWithIdMatching(currentId, callback)
              {
                var ret = function(id, value)
                {
                  if (id === callbackId)
                  {
                    callback(value);
                  }
                }.bind(undefined, currentId);
                return ret;
              }
              
              setup(isComposite);
              bridge.PlaybackEarlyPropertySets(element);
              oj._KnockoutBindingProvider.getInstance().__NotifyBindingsApplied(element);
              
            },
            null,
            {'disposeWhenNodeIsRemoved' : element}
          )
        }
      );
  
      ko.utils.domNodeDisposal.addDisposeCallback(element, cleanup);
    
  
  
    }
  }
  
  var _ATTRIBUTE_CHANGED = 'attribute-changed';
})();



/**
 * @ignore
 * @constructor
 * @private
 */
oj.__ExpressionPropertyUpdater = function(element, bindingContext, skipThrottling)
{ 
  // This function should be called when the bindings are applied initially and whenever the expression attribute changes
  this.setupExpression = function(attrVal, propName, metadata)
  {
    // If no metadata was passed in, just return bc this is not a component property.
    if (!metadata)
      return;

    // See if attribute is a component property/subproperty by checking for metadata
    var meta = _getPropertyMetadata(propName, metadata);
    if (!meta)
      return;

    var info = oj.__AttributeUtils.getExpressionInfo(attrVal);

    var oldListener = _expressionListeners[propName];
    if (oldListener)
    {
      oldListener['dispose']();
      _expressionListeners[propName] = null;
    }

    // Clean up property change listeners to handler the case when the type of the expression changes
    var changeListener  = _changeListeners[propName];
    if (changeListener)
    {
      element.removeEventListener(propName + _CHANGED_EVENT_SUFFIX, changeListener);
      _changeListeners[propName] = null;
    }

    if (metadata._domListener) {
      var event = oj.__AttributeUtils.eventListenerPropertyToEventType(propName);
      // If the attribute is removed, there won't be a new expression so the remove call
      // below can't be relied upon.  So clean up here as well to handle that case
      _setDomListener(bindingContext, event, null);
    }

    var readOnly = metadata['readOnly'];

    var expr = info.expr;

    if (expr)
    {
      // TODO: consider moving  __CreateEvaluator() to a more generic utility class
      var evaluator = oj.ComponentBinding.__CreateEvaluator(expr);

      if (!readOnly)
      {
        var initialRead = true;
        ko.ignoreDependencies(
          function()
          {
            _expressionListeners[propName] = ko.computed(
              // The read() function for the computed will be called when the computed is created and whenever any of
              // the expression's dependency changes
              function()
              {
                var value = evaluator(bindingContext);
                var unwrappedValue = ko.utils.unwrapObservable(value);

                if (metadata._domListener) {
                  var event = oj.__AttributeUtils.eventListenerPropertyToEventType(propName);
                  _setDomListener(bindingContext, event, unwrappedValue);
                }
                else
                {
                  // We cannot share the same storage with KO because we want to be able to compare values,
                  // so make a copy of the array before setting it.
                  if (Array.isArray(unwrappedValue))
                  {
                    unwrappedValue = unwrappedValue.slice();
                  }

                  if (metadata._eventListener) {
                    // Need to wrap the function so that it gets called with extra params
                    // No need to throw an error for the non-function case here because
                    // the bridge will throw it for us
                    if (unwrappedValue && unwrappedValue instanceof Function) {
                      unwrappedValue = _createSimpleEventListenerWrapper(bindingContext, unwrappedValue)
                    }
                  }
                  
                  if (!initialRead && _throttler)
                  {
                    _throttler.addChange(propName, unwrappedValue);
                  }
                  else
                  {
                    _setElementProperty(propName, unwrappedValue);
                  }
                }
              }
            );
          }
        );
        initialRead = false;
      }

      // Only listen for property changes for writeable properties
      if (metadata['writeback'] && !info.downstreamOnly)
        _changeListeners[propName] = _listenToPropertyChanges(propName, expr, evaluator);

      return true;
    }


    return false;
  }
  
  
  this.teardown = function()
  {
    var names = Object.keys(_expressionListeners);
    for (var i=0; i<names.length; i++)
    {
      var listener = _expressionListeners[names[i]];
      // listener might be null if this attribute changed from an expression to a literal
      if (listener)
        listener['dispose']();
    }
    _expressionListeners = {};


    // reset change listeners
    names = Object.keys(_changeListeners);
    for (i=0; i<names.length; i++)
    {
      var prop = names[i];
      element.removeEventListener(prop + _CHANGED_EVENT_SUFFIX, _changeListeners[prop]);
    }
    _changeListeners = {};

    // dom event listeners
    names = Object.keys(_domListeners);
    for (i=0; i<names.length; i++)
    {
      var event = names[i];
      element.removeEventListener(event, _domListeners[event]);
    }
    _domListeners = {};
    
    if (_throttler)
    {
      _throttler.dispose();
    }
  }

  function _getCurrent(bindingContext) {
    return bindingContext['$current'] || bindingContext['$data'];
  }

  function _createSimpleEventListenerWrapper(bindingContext, listener) {
    var current = _getCurrent(bindingContext);
    return function(event) {
      listener(event, current, bindingContext);
    };
  }

  function _createMutableEventListenerWrapper(bindingContext) {
    var current = _getCurrent(bindingContext);
    var eventListener;
    var domListener = function(event) {
      if (eventListener) {
        eventListener(event, current, bindingContext);
      }
    };
    domListener.setListener = function(listener) {
      eventListener = listener;
    }
    return domListener;    
  }

  function _setDomListener(bindingContext, event, listener) {
    var domListener = _domListeners[event];
    if (!domListener) {
      // Create the wrapper
      domListener = _createMutableEventListenerWrapper(bindingContext);
      _domListeners[event] = domListener;
      element.addEventListener(event, domListener);
    }
    // TODO: throw an Error in 5.0.0 if the listener is not a Function instead of failing silently
    if (listener == null || listener instanceof Function) {
      domListener.setListener(listener);
    }
  }

  function _listenToPropertyChanges(propName, expr, evaluator)
  {
    var listener = function(evt)
    {
      if (!_isSettingProperty(propName))
      {
        var written = false;
        var reason;
        ko.ignoreDependencies(
          function()
          {
            var value = evt['detail']['value'];

            var target = evaluator(bindingContext);

            if (ko.isObservable(target))
            {
              if (ko.isWriteableObservable(target))
              {
                target(oj.ComponentBinding.__cloneIfArray(value));
                written = true;
              }
              else
              {
                reason = "the observable is not writeable";
              }
            }
            else
            {
              var writerExpr = oj.__ExpressionUtils.getPropertyWriterExpression(expr);
              if (writerExpr != null)
              {
                // TODO: consider caching the evaluator
                var wrirerEvaluator = oj.ComponentBinding.__CreateEvaluator(writerExpr);
                wrirerEvaluator(bindingContext)(oj.ComponentBinding.__cloneIfArray(value));
                written = true;
              }
              else
              {
                reason = "the expression is not a valid update target";
              }
            }
          }
        );
        
        if (!written)
        {
          if (reason)
          {
            oj.Logger.info("The expression '%s' for property '%s' was not updated because %s.", expr, propName, reason);
          }
        }
      }

    };

    element.addEventListener(propName + _CHANGED_EVENT_SUFFIX, listener);
    return listener;
  }


  function _setElementProperty(propName, value)
  {
    _startSettingProperty(propName);
    try
    {
      element['setProperty'](propName, value);
    }
    finally
    {
      _endSettingProperty(propName);
    }
  }
  
  function _startSettingProperty(propName)
  {
    var count = _settingProperties[propName] || 0;
    count++;
    _settingProperties[propName] = count;
  }
  
  function _endSettingProperty(propName)
  {
    var count = _settingProperties[propName];
    if (!count)
    {
      oj.Logger.error("Property count undefrlow");
      return;
    }
    count--;
    if (count === 0)
    {
      _settingProperties[propName] = null;
    }
    else
    {
      _settingProperties[propName] = count;
    }
  }
  
  function _isSettingProperty(propName)
  {
    return _settingProperties[propName];
  }

  function _getPropertyMetadata(propName, metadata)
  {
    var propPath = propName.split('.');
    var meta = metadata;
    propPath.shift(); // current metadata is already for top level property
    for (var j = 0; j < propPath.length; j++)
    {
      meta = meta['properties'];
      if (!meta)
        break;
      meta = meta[propPath[j]];
    }
    return meta;
  }
  
  function _setupComponentChangeTracker()
  {
    
    //  We are not throttling property updates for JET composites and native HTML elements because:
    //  1) Existing customers may expect that properties be updated immediately
    //  2) KO templates do not have any optimizations for simulteneous property updates
    //  3) Updates to embeddded JET components will still be throttled
    if (skipThrottling)
    {
      return null;
    }
    
    var updater = function(changes)
    {
      var keys = Object.keys(changes);
      for (var i=0; i<keys.length; i++)
      {
        var key = keys[i];
      
        // We rely on the flag set by the _startSettingProperty() to determine whether
        // the property is being changed in response to a subscription notification
        // from KO. Change events from all sets made NOT in response to a KO update 
        // trigger an an attempt to write the value back into the associated expression.
        // If the writeback is not possible for any reason, we consider that the binding
        // has been 'obscured' by a local property set, and further update notifications
        // are are going to be ignored. This means that we are stuck with calling 
        // _startSettingProperty() for all properties in the batch update. 
        // The downside for this approach is that property changes made in response 
        // to a property set from within the same batch will NOT be written back
        // into the KO expression. The likelihood of the writeback bugs because 
        // of this is very low IMO: we have few properties with writeback enabled,
        // and the behavior for the component changing a property whose value is being set
        // in the same batch is already undefined. If some app runs into an issue, the
        // workaround will be to call oj.ComponentBinding.deliverChanges() after each observable
        // change.
        _startSettingProperty(key);
      }
      try
      {
        element['setProperties'](changes);
      }
      finally
      {
        for (var n=0; n<keys.length; n++)
        {
          var k = keys[n];
          _endSettingProperty(k);
        }
      }
    };
     
    return new ComponentChangeTracker(updater, oj.ComponentBinding._changeQueue);
  }
  
  
  var _throttler = _setupComponentChangeTracker();
  
  var _domListeners = {};
  var _expressionListeners = {};
  var _changeListeners = {};
  var _settingProperties = {};
  var _CHANGED_EVENT_SUFFIX = "Changed";
};

/**
 * @ignore
 */
(function () {
  oj.__KO_CUSTOM_BINDING_PROVIDER_INSTANCE.addPostprocessor
  (
    {
      'preprocessNode': function(node, wrappedReturn)
      {
        var newNodes;
        if (node.nodeType === 1)
        {
          var nodeName = node.nodeName.toLowerCase();

          if ('oj-bind-text' === nodeName)
          {
            newNodes = _replaceWithKo(node, 'text', 'value', true);
          }
          else if ('oj-bind-if' === nodeName)
          {
            newNodes = _replaceWithKo(node, 'if', 'test', false);
          }
          else if ('oj-bind-for-each' === nodeName) 
          {
            newNodes = _replaceWithKoForEach(node);
          }
        }
        return newNodes? newNodes: wrappedReturn;
      },
      'nodeHasBindings': function (node, wrappedReturn) {
        var bindings = _getBindings(node);
        return wrappedReturn || bindings._ATTR_BIND != null || bindings._STYLE_BIND != null || bindings._EVENT_BIND != null;
      },
      'getBindingAccessors': function (node, bindingContext, wrappedReturn) 
      {
        if (node.nodeType === 1) 
        {
          wrappedReturn = wrappedReturn || {};
          var bindings = _getBindings(node);

          // Style bindings
          var styleAttrs = bindings._STYLE_BIND;
          if (styleAttrs)
          {
            if (wrappedReturn['style'])
              throw new Error("Cannot have both style data-bind and JET style binding on " + node.tagName + " with id " + node.id);

            // If :style is set, any additional :style.* attributes would have thrown an error earlier
            if (styleAttrs === 'style')
            {
              wrappedReturn[styleAttrs] = _getValueEvaluator(node, styleAttrs,  node.getAttribute(_getBoundAttrName(styleAttrs)), bindingContext, 'object');
            }
            else
            {
              var styleEvaluators = {};
              for (var i = 0; i < styleAttrs.length; i++)
              {
                var attr = styleAttrs[i];
                var styleProp = oj.__AttributeUtils.attributeToPropertyName(attr.substring(6));
                styleEvaluators[styleProp] = _getValueEvaluator(node, attr,  node.getAttribute(_getBoundAttrName(attr)), bindingContext, 'string');
              }
              wrappedReturn['style'] = _getObjectEvaluator(styleEvaluators);
            }
          }

          // All other attribute bindings
          var boundAttrs = bindings._ATTR_BIND;
          if (boundAttrs)
          {
            if (wrappedReturn['attr'])
              throw new Error("Cannot have both attr data-bind and JET attribute binding on " + node.tagName + " with id " + node.id);
            
            var attrEvaluators = {};
            for (var i = 0; i < boundAttrs.length; i++)
            {
              var attr = boundAttrs[i];
              var evaluator;
              // For the class attribute, we also support array and object types so we should check
              // for arrays and convert to a computed that returns a string for knockout. For object
              // types we pass to knockout's css binding instead of attr so pass the attrEvaluators object
              // and the wrappedReturn to the method to determine which to use.
              if (attr === 'class')
                _setClassEvaluator(wrappedReturn, attrEvaluators, node, node.getAttribute(_getBoundAttrName(attr)), bindingContext);
              else
                attrEvaluators[attr] = _getValueEvaluator(node, attr, node.getAttribute(_getBoundAttrName(attr)), bindingContext, 'string');
            }

            wrappedReturn['attr'] = _getObjectEvaluator(attrEvaluators);
          }

          // Event bindings for native HTML elements
          var eventAttrs = bindings._EVENT_BIND;
          if (eventAttrs)
          {
            // Delegate to ExpressionPropertyUpdater, which is also responsible for
            // setting up the event bindings for custom elements.
            var expressionHandler = new oj.__ExpressionPropertyUpdater(node, bindingContext, true);

            for (var i = 0; i < eventAttrs.length; i++)
            {
              var attrNode = eventAttrs[i];
              var propName = oj.__AttributeUtils.attributeToPropertyName(attrNode.nodeName);
              expressionHandler.setupExpression(attrNode.value, propName, {_domListener: true});
            }

            ko.utils.domNodeDisposal.addDisposeCallback(node, function()
            {
              if (expressionHandler)
              {
                expressionHandler.teardown();
                expressionHandler = null;
              }
            });
          }
          
        }
        return wrappedReturn;
      }
    }
  );

  function _replaceWithKo(node, bindableAttr, nodeAttr, stringify)
  {
    var expr = _getExpression(node.getAttribute(nodeAttr), stringify);
    if (!expr)
      return;

    var binding = 'ko ' + bindableAttr + ':' + expr;
    return _getReplacementNodes(node, bindableAttr, binding);
  }
  
  function _replaceWithKoForEach(node)
  {
    var dataValue = node.getAttribute('data');
    var dataExp = oj.__AttributeUtils.getExpressionInfo(dataValue).expr;
    if (!dataExp) 
    {
      try 
      {
        var literalValue = JSON.parse(dataValue);
        if (Array.isArray(literalValue))
          dataExp = dataValue;
        else
          throw new Error("got value " + dataValue);
      }
      catch (e) 
      {
        throw new Error("The value on the oj-bind-for-each data attribute should be either a JSON array or an expression : " + e);
      }
    }

    var asExp = _getExpression(node.getAttribute('as'), true);
    if (!dataExp)
      return;
    
    var binding = "ko _ojBindForEach_:{data:" + dataExp;
    binding += asExp ? ",as:" + asExp + "}" : "}";
    return _getReplacementNodes(node, "_ojBindForEach_", binding);
  }
  
  function _getReplacementNodes (node, bindableAttr, binding){
    // Wrap the ko comment we will generate with an additional wrapper
    // with all of the attribute info that was on the original DOM node.
    // <!--oj-bind-for-each data='[[users]]' -->
    var nodeName = node.tagName.toLowerCase();
    var ojcommenttext = nodeName;
    var attrs = node.attributes;
    for (var i = 0; i < attrs.length; i++) {
      // jsperf says string concat is faster than array join, but
      // varies across browsers whether text += or text = text + is faster
      // the below method is fastest on Chrome and not bad on FF
      var attr = attrs[i];
      ojcommenttext = ojcommenttext + ' ';
      ojcommenttext = ojcommenttext + attr.name;
      ojcommenttext = ojcommenttext + "='";
      ojcommenttext = ojcommenttext + attr.value;
      ojcommenttext = ojcommenttext + "'"; 
    }

    var parent  = node.parentNode;
    var ojOpenComment = document.createComment(ojcommenttext);
    var ojCloseComment = document.createComment('/' + nodeName);
    parent.insertBefore(ojOpenComment, node); // @HTMLUpdateOK

    var koOpenComment = document.createComment(binding);
    var koCloseComment = document.createComment('/ko');
    parent.insertBefore(koOpenComment, node); // @HTMLUpdateOK

    var newNodes = [ojOpenComment, koOpenComment];
    // Copy oj-bind-x children into the comment node
    if (bindableAttr === 'if')
    {
      while (node.childNodes.length > 0)
      {
        var child = node.childNodes[0];
        parent.insertBefore(child, node); // @HTMLUpdateOK
        newNodes.push(child);
      }
    }
    else if (bindableAttr === '_ojBindForEach_') 
    {
      while (node.childNodes.length > 0)
      {
        var child = node.childNodes[0];
        parent.insertBefore(child, node); // @HTMLUpdateOK
        newNodes.push(child);
        if (child.nodeType === 1 && "template" === child.nodeName.toLowerCase()) 
        {
          if (!koOpenComment['_templateNode'])
            Object.defineProperty(koOpenComment, '_templateNode', { 'value': child, 'enumerable': false });
          else 
            throw new Error("Multiple templates found: oj-bind-for-each requires a single template element as its direct child");
        }
      }
      // check for the template since it is a required attribute for oj-bind-for-each.
      if (!koOpenComment['_templateNode'])
        throw new Error("Template not found: oj-bind-for-each requires a single template element as its direct child");
    }
    parent.insertBefore(koCloseComment, node); // @HTMLUpdateOK
    newNodes.push(koCloseComment);

    parent.replaceChild(ojCloseComment, node); // @HTMLUpdateOK
    newNodes.push(ojCloseComment);

    return newNodes;
  }

  function _getExpression(attrValue, stringify)
  {
    if (attrValue != null)
    {
      var exp = oj.__AttributeUtils.getExpressionInfo(attrValue).expr;
      if (exp == null)
      {
        if (stringify)
          exp = "'" + attrValue + "'";
        else
          exp = attrValue;
      }
      return exp;
    }

    return null;
  }

  function _setClassEvaluator(wrappedReturn, attrEvaluators, elem, value, bindingContext)
  {
    var evaluator, resolvedValue, isArray;
    var exp = oj.__AttributeUtils.getExpressionInfo(value).expr;
    if (exp == null)
    {
      resolvedValue = oj.__AttributeUtils.coerceValue(elem, 'class', value, 'any');
      isArray = Array.isArray(resolvedValue);
      evaluator = function() {
        return isArray ? resolvedValue.join(' ') : resolvedValue;
      };
    }
    else
    {
      evaluator = oj.ComponentBinding.__CreateEvaluator(exp).bind(null, bindingContext);
      var classObs = evaluator();

      if (typeof classObs === 'function')
      {
        resolvedValue = classObs();
        isArray = Array.isArray(resolvedValue);
        if (isArray)
        {
          evaluator = ko.pureComputed(function() {
            return classObs().join(' ');
          });
        }
      }
      else
      {
        resolvedValue = classObs;
      }
    }
    if (typeof resolvedValue === 'string' || isArray)
      attrEvaluators['class'] = evaluator;
    else
      wrappedReturn['css'] = evaluator
  }

  function _getValueEvaluator(elem, attr, value, bindingContext, type)
  {
    var exp = oj.__AttributeUtils.getExpressionInfo(value).expr;
    if (exp == null)
    {
      return function() { 
        return type === 'object' ? oj.__AttributeUtils.coerceValue(elem, attr, value, type) : value; 
      };
    }
    else
      return oj.ComponentBinding.__CreateEvaluator(exp).bind(null, bindingContext);
  }

  function _getObjectEvaluator(attrEvaluators) 
  {
    return function() {
      var evaluatedExprs = {};
      var attrs = Object.keys(attrEvaluators);
      for (var i = 0; i < attrs.length; i++)
      {
        var attr = attrs[i];
        evaluatedExprs[attr] = attrEvaluators[attr]();
      }
      return evaluatedExprs;
    };
  }

  function _getBindings(node)
  {
    if (node.nodeType !== 1)
      return {};

    if (!node[_BINDINGS])
    {
      // Using the weird notation below instead of:
      // var isCustomElement = oj.BaseCustomElementBridge && oj.BaseCustomElementBridge.getRegistered(node.nodeName);
      // because it causes minification error for some reason.
      var isCustomElement = oj.BaseCustomElementBridge ? !!oj.BaseCustomElementBridge.getRegistered(node.nodeName) : false;

      // Iterate through all attributes on the element and find those that are using our
      // data bind attibute syntax
      var bindings = {};
      var boundAttrs = [];
      var boundStyle = [];
      var boundEvents = [];
      var attrs = node.attributes; // attrs is a NamedNodeMap
      for (var i = 0; i < attrs.length; i++)
      {
        var attr = attrs[i];
        var unboundAttrName = _getUnboundAttrName(attr.name);
        if (unboundAttrName)
        {
          // An element can have a style binding or several style.* bindings.
          // Throw an error if both are present.
          if (unboundAttrName === 'style')
            bindings._STYLE_BIND = unboundAttrName;
          else if (unboundAttrName.substring(0,6) === 'style.')
            boundStyle.push(unboundAttrName);
          else
            boundAttrs.push(unboundAttrName);
        }

        // Handle event binding for native HTML elements.
        // The event binding for custom elements is handled by the bridge.
        if (!isCustomElement && attr.name.substring(0,3) === 'on-')
        {
          boundEvents.push(attr);
        }
      }

      if (boundStyle.length)
      {
        if (bindings._STYLE_BIND)
          throw new Error('Cannot have both style and style.* data bound attributes on ' + node.tagName + ' with id ' + node.id);
        else
          bindings._STYLE_BIND = boundStyle;
      }

      if (boundAttrs.length)
        bindings._ATTR_BIND = boundAttrs;

      if (boundEvents.length)
        bindings._EVENT_BIND = boundEvents;

      // Cache attribute map as a non-enumerable property so we don't have to loop again
      // in getBindingAccessors
      Object.defineProperty(node, _BINDINGS, {'value': bindings});
    }
    return node[_BINDINGS];
  }

  function _getUnboundAttrName(attr)
  {
    if (attr && attr.charAt(0) === ':')
      return attr.slice(1);
    return null;
  }

  function _getBoundAttrName(attr)
  {
    return ':' + attr;
  }

  var _BINDINGS = "_ojbindingsobj";
})();
/**
 * @ignore
 * @constructor
 * @private
 */
oj._KnockoutBindingProvider = function()
{
  

  /**
   * @ignore
   */
  this.__NotifyBindingsApplied = function(elem)
  {
    elem['resolveBP'](this);
  }

}

/**
 * @ignore
 */
oj._KnockoutBindingProvider.getInstance = function()
{
  return oj._KnockoutBindingProvider._instance;
}

/**
 * @ignore
 */
oj._KnockoutBindingProvider._instance = new oj._KnockoutBindingProvider();





/**
 * Utility methods for knockout templates.
 * 
 * @ojstatus preview
 * @since 4.0.0
 * @namespace
 * @export
 */
oj.KnockoutTemplateUtils = {};

/**
 * JET custom elements do not support template binding attributes, so applications using knockout templates 
 * should use this utility to convert their knockout templates to a renderer function for use in component 
 * renderer APIs instead.
 * 
 * @param {string} template The name of the knockout template to use.
 * @param {boolean=} bReplaceNode True if the entire target node should be replaced by the output of the template.
 *                                If false or omitted, the children of the target node will be replaced.
 * @return {function(Object)} A renderer function that takes a context object.
 * @export
 * 
 * @example <caption>Convert a knockout template to a custom tooltip renderer function:</caption>
 * &lt;oj-tag-cloud tooltip.renderer="[[oj.KnockoutTemplateUtils.getRenderer('tooltip_template')]]">&lt;/oj-tag-cloud>
 */
oj.KnockoutTemplateUtils.getRenderer = function(template, bReplaceNode) 
{
  var templateRenderer = function (context) {
    // For DVTs, the node to attach the template to is different than the context's parentElement key
    var parentElement = context['_parentElement'] || context['parentElement'];
    
    var bindingContext = ko.contextFor(context['componentElement']);

    var childContext = bindingContext['createChildContext'](context['data'], null, function(binding) { binding['$context'] = context; });
    ko['renderTemplate'](template, childContext, {'afterRender': function(renderedElement) { $(renderedElement)['_ojDetectCleanData'](); }}, 
      parentElement, bReplaceNode ? 'replaceNode' : 'replaceChildren');
    return null;
  };
  
  return function (context) {
    if (context['componentElement'].classList && context['componentElement'].classList.contains("oj-dvtbase")){
      // Create a dummy div
      var dummyDiv = document.createElement("div");
      dummyDiv.style.display = "none";
      dummyDiv._dvtcontext = context['_dvtcontext'];
      context['componentElement'].appendChild(dummyDiv);
      Object.defineProperty(context, '_parentElement', { 'value': dummyDiv, 'enumerable': false });
      Object.defineProperty(context, '_templateCleanup', { 'value': function(){$(dummyDiv).remove();}, 'enumerable': false });
      Object.defineProperty(context, '_templateName', { 'value': template, 'enumerable': false });

      templateRenderer(context);

      var elem = dummyDiv.children[0];
      if (elem) {
        if (elem.namespaceURI === 'http://www.w3.org/2000/svg') {
          dummyDiv.removeChild(elem);
          $(dummyDiv).remove();
        } 
        return {'insert':elem};
      }
      return {'preventDefault':true};
    }
    else {
      return templateRenderer(context);
    }
  };
};
/*jslint browser: true*/
/*
** Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
*/


/**
 * @class oj.ResponsiveKnockoutUtils
 * @classdesc Utilities for creating knockout observables to implement responsive pages. 
 * For example you could use oj.ResponsiveKnockoutUtils.createMediaQueryObservable to 
 * create an observable based on the screen width and then bind the tab bar's 
 * orientation attribute to it. See the method doc below for specific examples.
 * @since 1.1.0
 * @export
 * @hideconstructor
 * @ojtsimport knockout
 * @ojstatus preview
 */
oj.ResponsiveKnockoutUtils = {};

/**
 * <p>creates an observable that 
 * returns true or false based on a media query string. 
 * Can be used in conjuntion with {@link oj.ResponsiveUtils.getFrameworkQuery}
 * to create an observable based on a framework media query.</p>
 * 
 * <p>Example:</p>
 * <pre class="prettyprint">
 * <code>
 *    var customQuery = oj.ResponsiveKnockoutUtils.createMediaQueryObservable(
 *                                         '(min-width: 400px)');
 *
 *    var lgQuery = oj.ResponsiveUtils.getFrameworkQuery(
 *                             oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.LG_UP);
 *        
 *    self.large = oj.ResponsiveKnockoutUtils.createMediaQueryObservable(lgQuery);
 * </code></pre>
 * 
 * @param {string} queryString media query string, for example '(min-width: 400px)'
 * @return a knockout observable  that 
 *              returns true or false based on a media query string.
 * @ojsignature {target: "Type", for: "returns", value: "KnockoutObservable<boolean>"}
 * @export
 * @static
 */
oj.ResponsiveKnockoutUtils.createMediaQueryObservable = function(queryString)
{

  if (queryString == null)
  {
    throw new Error("oj.ResponsiveKnockoutUtils.createMediaQueryObservable: aborting, queryString is null");
  }

  var query = window.matchMedia(queryString);

  var observable = ko.observable(query.matches);

  // add a listener for future changes
  query.addListener(function(query) {
      observable(query.matches);
      //console.log("query listener called! query.matches = " + query.matches + ", size = " + (document.outerWidth || document.body.clientWidth));
  });


 
  // There is a major bug in webkit, tested on ios 7 going from
  // landscape to portrait.
  //    https://bugs.webkit.org/show_bug.cgi?id=123293
  // 
  // Basically if you use a media query in css
  // then the js matchmedia call won't work! 
  //
  // According to the bug this is known to be on webkit 538.4,
  // but I see it on 537.51 as well which is earlier, so 
  // assume the problem exists generally on safari.
  // Chrome now uses blink instead of webkit, but chrome
  // still has webkit in their user agent string, however they
  // now only change the number after "Chrome".
  
  if (navigator.userAgent.indexOf("WebKit") != -1 && 
      navigator.userAgent.indexOf("Chrome") == -1) 
  {
    $(window).resize(function() {
      //console.log("Resize called! Size = " + (document.outerWidth || document.body.clientWidth));

      // Somehow if I change some text in the dom on resize 
      // the query listener is called
      var selector = "oj-webkit-bug-123293";

      if($('body').has("." + selector).length === 0) 
      {
        // setting display: none doesn't work, so using 
        // oj-helper-hidden-accessible because this visually 
        // hides the content without using display:none. 
        // However we don't want screen readers to read 
        // this so setting aria-hidden to true.
        $('body').append('<div aria-hidden="true" class="oj-helper-hidden-accessible ' + selector + '">'); // @HTMLUpdateOK 
      }

      $("." + selector).text((new Date().getMilliseconds()).toString()); // @HTMLUpdateOK 
    });
  }
  
  return observable;
};


/**
 * This function creates a computed observable, the 
 * value of which is one of the {@link oj.ResponsiveUtils.SCREEN_RANGE} constants. 
 * For example when the width is in the 
 * range defined by the sass variable $mediumScreenRange then 
 * the observable returns <code>oj.ResponsiveUtils.SCREEN_RANGE.MD</code>, 
 * but if it's in the range defined by $largeScreenRange then 
 * it returns <code>oj.ResponsiveUtils.SCREEN_RANGE.LG</code>, etc. 
 * 
 * 
 * 
 * <p>Example:</p>
 * <pre class="prettyprint">
 * <code>
 *        // create an observable which returns the current screen range
 *        self.screenRange = oj.ResponsiveKnockoutUtils.createScreenRangeObservable();
 *
 *        self.label2 = ko.computed(function() {
 *          var range = self.screenRange();
 * 
 *          if ( oj.ResponsiveUtils.compare( 
 *                       range, oj.ResponsiveUtils.SCREEN_RANGE.MD) <= 0)
 *          {
 *            // code for when screen is in small or medium range
 *          }
 *          else if (range == oj.ResponsiveUtils.SCREEN_RANGE.XL)
 *          {
 *            // code for when screen is in XL range
 *          }
 *        });
 * </code></pre>
 *
 * @return a knockout observable the value of which is one of the
 *  screen range constants, for example oj.ResponsiveUtils.SCREEN_RANGE.MD
 * @ojsignature {target: "Type", for: "returns", value: "KnockoutObservable<string>"}
 * @export
 * @static
 */
oj.ResponsiveKnockoutUtils.createScreenRangeObservable = function()
{
  // queryies
  var xxlQuery = oj.ResponsiveUtils.getFrameworkQuery(
                                oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.XXL_UP);

  var xlQuery = oj.ResponsiveUtils.getFrameworkQuery(
                                oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.XL_UP);

  var lgQuery = oj.ResponsiveUtils.getFrameworkQuery(
                                oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.LG_UP);

  var mdQuery = oj.ResponsiveUtils.getFrameworkQuery(
                                oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.MD_UP);

  var smQuery = oj.ResponsiveUtils.getFrameworkQuery(
                                oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.SM_UP);



  // observables
  var xxlObservable = xxlQuery == null ? 
                      null : 
                      oj.ResponsiveKnockoutUtils.createMediaQueryObservable(xxlQuery);

  var xlObservable = xlQuery == null ? 
                     null : 
                     oj.ResponsiveKnockoutUtils.createMediaQueryObservable(xlQuery);

  var lgObservable = lgQuery == null ? 
                     null : 
                     oj.ResponsiveKnockoutUtils.createMediaQueryObservable(lgQuery);

  var mdObservable = mdQuery == null ? 
                     null : 
                     oj.ResponsiveKnockoutUtils.createMediaQueryObservable(mdQuery);

  var smObservable = smQuery == null ? 
                     null : 
                     oj.ResponsiveKnockoutUtils.createMediaQueryObservable(smQuery);


  return ko.computed(function() {
    if (xxlObservable && xxlObservable())
      return oj.ResponsiveUtils.SCREEN_RANGE.XXL; 

    if (xlObservable && xlObservable())
      return oj.ResponsiveUtils.SCREEN_RANGE.XL; 

    if (lgObservable && lgObservable())
      return oj.ResponsiveUtils.SCREEN_RANGE.LG; 

    if (mdObservable && mdObservable())
      return oj.ResponsiveUtils.SCREEN_RANGE.MD; 

    if (smObservable && smObservable())
      return oj.ResponsiveUtils.SCREEN_RANGE.SM; 

    throw new Error(" NO MATCH in oj.ResponsiveKnockoutUtils.createScreenRangeObservable");

  });
};

/**
 * Returns a renderer function and executes the template specified in the binding attribute. (for example, a knockout template).
 * @param {Object} bindingContext the ko binding context
 * @param {string} template the name of the template
 * @return {Function} the renderer function
 * @private
 */
function _getDvtRenderer(bindingContext, template) {
  return function (context) {
    var model = bindingContext['createChildContext'](context['context']);
    ko['renderTemplate'](template, model, {'afterRender': function(renderedElement) { $(renderedElement)['_ojDetectCleanData'](); }}, context['parentElement']);
    return null;
  };
};

/**
 * Returns a renderer function and executes the template specified in the binding attribute for data items. (for example, a knockout template).
 * @param {Object} bindingContext the ko binding context
 * @param {string} template the name of the template
 * @return {Function} the renderer function
 * @private
 */
function _getDvtDataRenderer(bindingContext, template) {
  return function (context) {
    var model = bindingContext['createChildContext'](context['data']);
    ko['renderTemplate'](template, model, {'afterRender': function(renderedElement) { $(renderedElement)['_ojDetectCleanData'](); }}, context['parentElement']);
    return null;
  };
};

/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleManagedTooltipAttributes(name, value, bindingContext) {
  if (name === "tooltip" && value['template']) {
     value['_renderer'] = _getDvtRenderer(bindingContext, value['template']);
  }   
  return {'tooltip': value};
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({
  'attributes': ["tooltip"],
  'init': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedTooltipAttributes(name, value, bindingContext);
  },
  'update': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedTooltipAttributes(name, value, bindingContext);
  },
  'for': "tooltipOptionRenderer"
});

// Default declarations for all components supporting tooltips
(function() {
  var componentsArray = ['ojChart', 'ojDiagram', 'ojNBox', 'ojPictoChart', 'ojSunburst', 'ojTagCloud', 'ojThematicMap', 'ojTreemap',
                         'ojDialGauge', 'ojLedGauge', 'ojRatingGauge','ojSparkChart', 'ojStatusMeterGauge', 'ojGantt'];
  for(var i = 0; i < componentsArray.length; i++) {
    oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({ 
      "for": componentsArray[i],
      "use": "tooltipOptionRenderer"
    });
  }                   
})();

/*!
  Knockout Fast Foreach v0.6.0 (2016-07-28T11:02:54.197Z)
  By: Brian M Hunt (C) 2015 | License: MIT

  Adds `fastForEach` to `ko.bindingHandlers`.
  
  Modification notice: The code is obtained from https://github.com/brianmhunt/knockout-fast-foreach 
  and modified by Oracle JET team to be included into Oracle JET project.
*/

(function() {
  "use strict";
// index.js
// --------
// Fast For Each
//
// Employing sound techniques to make a faster Knockout foreach binding.
// --------

//      Utilities
var MAX_LIST_SIZE = 9007199254740991;

// from https://github.com/jonschlinkert/is-plain-object
function isPlainObject(o) {
  return !!o && typeof o === 'object' && o.constructor === Object;
}

// From knockout/src/virtualElements.js
var commentNodesHaveTextProperty = document && document.createComment("test").text === "<!--test-->";
var startCommentRegex = commentNodesHaveTextProperty ? /^<!--\s*ko(?:\s+([\s\S]+))?\s*-->$/ : /^\s*ko(?:\s+([\s\S]+))?\s*$/;
var supportsDocumentFragment = document && typeof document.createDocumentFragment === "function";
function isVirtualNode(node) {
  return (node.nodeType === 8) && startCommentRegex.test(commentNodesHaveTextProperty ? node.text : node.nodeValue);
}


// Get a copy of the (possibly virtual) child nodes of the given element,
// put them into a container, then empty the given node.
function makeTemplateNode(sourceNode) {
  var container = document.createElement("div");
  var parentNode;
  if(sourceNode['_templateNode'] && sourceNode['_templateNode'].content) 
  {
    // For e.g. <template> tags
    parentNode = sourceNode['_templateNode'].content; //parentNode = sourceNode.content;
  }
  else if (sourceNode['_templateNode']) 
  {
    parentNode = sourceNode['_templateNode'];
  }
  else
  {
    //allow body of a script tag to be used as a template
    var scripts = sourceNode.parentNode.getElementsByTagName('SCRIPT');
    parentNode = scripts && scripts.length > 0 ? ko.utils.parseHtmlFragment(scripts[0].text) : sourceNode;
  }
  
  var nodes = Array.isArray(parentNode) ? parentNode : ko.virtualElements.childNodes(parentNode);
  ko.utils.arrayForEach(nodes, function (child) {
    // FIXME - This cloneNode could be expensive; we may prefer to iterate over the
    // parentNode children in reverse (so as not to foul the indexes as childNodes are
    // removed from parentNode when inserted into the container)
    if (child) {
      container.insertBefore(child.cloneNode(true), null); // @HTMLUpdateOK
    }
  });
  return container;
}

// Mimic a KO change item 'add'
function valueToChangeAddItem(value, index) {
  return {
    status: 'added',
    value: value,
    index: index
  };
}

// KO 3.4 doesn't seem to export this utility function so it's here just to be sure
function createSymbolOrString(identifier) {
  //return typeof Symbol === 'function' ? Symbol(identifier) : identifier;
  return identifier;
}

// Array.prototype.map doesn't execute the mapping function on indices that don't have assigned values
// so it would be a NOP for new Array(5).  Call this for loop based version instead
function arrayMap(array, mapping) {
  var result;
  if (array) {
    var length = array.length;
    result = new Array(length);
    for (var i = 0; i < length; i++) {
      result[i] = mapping(array[i], i, array);
    }
  }
  return result;
}

// store a symbol for caching the pending delete info index in the data item objects
var PENDING_DELETE_INDEX_KEY = createSymbolOrString("_ko_ffe_pending_delete_index");

/**
 * @constructor
 * @private
 * @since 4.1.0
 */
function OjForEach(element, value, context) {
  this.element = value.element || element;
  this.container = isVirtualNode(this.element) ?
                   this.element.parentNode : this.element;
  this.$context = context;
  this.data = value['data'];
  this.as = value['as'];
  this.templateNode = makeTemplateNode(this.element);
  this.changeQueue = [];
  this.firstLastNodesList = [];
  this.indexesToDelete = [];
  this.rendering_queued = false;
  this.pendingDeletes = [];

  // Remove existing content.
  ko.virtualElements.emptyNode(this.element);

  // Prime content
  var primeData = ko.unwrap(this.data);
  this.onArrayChange(arrayMap(primeData, valueToChangeAddItem));
  this.addSubscriptions();
}

OjForEach.PENDING_DELETE_INDEX_KEY = PENDING_DELETE_INDEX_KEY;

OjForEach.prototype.addSubscriptions = function () {
  // Watch for changes
  if (ko.isObservable(this.data)) {
    if (!this.data.indexOf) {
      // Make sure the observable is trackable.
      this.data = this.data['extend']({ 'trackArrayChanges': true });
    }
    this.changeSubs = this.data['subscribe'](this.onArrayChange, this, 'arrayChange');
  }
};

OjForEach.prototype.removeSubscriptions = function () {
  if (this.changeSubs) {
    this.changeSubs['dispose']();
  }
};


// If the array changes we register the change.
OjForEach.prototype.onArrayChange = function (changeSet) {
  
  var self = this;
  var changeMap = {
    'added': [],
    'deleted': []
  };
  var statusAdded = 'added', statusDeleted = 'deleted';
  
  // knockout array change notification index handling:
  // - sends the original array indexes for deletes
  // - sends the new array indexes for adds
  // - sorts them all by index in ascending order
  // because of this, when checking for possible batch additions, any delete can be between to adds with neighboring indexes, so only additions should be checked
  for (var i = 0, len = changeSet.length; i < len; i++) {

    if (changeMap[statusAdded].length && changeSet[i].status == statusAdded) {
      var lastAdd = changeMap[statusAdded][changeMap[statusAdded].length - 1];
      var lastIndex = lastAdd.isBatch ? lastAdd.index + lastAdd.values.length - 1 : lastAdd.index;
      if (lastIndex + 1 == changeSet[i].index) {
        if (!lastAdd.isBatch) {
          // transform the last addition into a batch addition object
          lastAdd = {
            isBatch: true,
            status: statusAdded,
            index: lastAdd.index,
            values: [lastAdd.value]
          };
          changeMap[statusAdded].splice(changeMap[statusAdded].length - 1, 1, lastAdd);
        }
        lastAdd.values.push(changeSet[i].value);
        continue;
      }
    }

    changeMap[changeSet[i].status].push(changeSet[i]);
  }

  if (changeMap[statusDeleted].length > 0) {
    this.changeQueue.push.apply(this.changeQueue, changeMap[statusDeleted]);
    this.changeQueue.push({ status: 'clearDeletedIndexes' });
  }
  this.changeQueue.push.apply(this.changeQueue, changeMap[statusAdded]);
  // Once a change is registered, the ticking count-down starts for the processQueue.
  if (this.changeQueue.length > 0 && !this.rendering_queued) {
    this.rendering_queued = true;
    self.processQueue();
  }
};


// Reflect all the changes in the queue in the DOM, then wipe the queue.
OjForEach.prototype.processQueue = function () {
  var self = this;
  var lowestIndexChanged = MAX_LIST_SIZE;

  ko.utils.arrayForEach(this.changeQueue, function (changeItem) {
    if (typeof changeItem.index === 'number') {
      lowestIndexChanged = Math.min(lowestIndexChanged, changeItem.index);
    }
    // console.log(self['data'](), "CI", JSON.stringify(changeItem, null, 2), JSON.stringify($(self.element).text()))
    self[changeItem.status](changeItem);
    // console.log("  ==> ", JSON.stringify($(self.element).text()))
  });
  this.flushPendingDeletes();
  this.rendering_queued = false;

  // Update our indexes.
  this.updateIndexes(lowestIndexChanged);

  this.changeQueue = [];
};


// Process a changeItem with {status: 'added', ...}
/**
 * @expose
 * @ignore 
 */
OjForEach.prototype.added = function (changeItem) {
  var index = changeItem.index;
  var valuesToAdd = changeItem.isBatch ? changeItem.values : [changeItem.value];
  var referenceElement = this.getLastNodeBeforeIndex(index);
  // gather all childnodes for a possible batch insertion
  var allChildNodes = [];

  for (var i = 0, len = valuesToAdd.length; i < len; ++i) {
    var childNodes;

    // we check if we have a pending delete with reusable nodesets for this data, and if yes, we reuse one nodeset
    var pendingDelete = this.getPendingDeleteFor(valuesToAdd[i]);
    if (pendingDelete && pendingDelete.nodesets.length) {
      childNodes = pendingDelete.nodesets.pop();
    } else {
      var templateClone = this.templateNode.cloneNode(true);
      var current = {
          'data': valuesToAdd[i],
          'index': index + i,
          'observableIndex': ko.observable()
        };
      var childContext = this.$context['extend']({
        '$root': undefined,
        '$parent': undefined,
        '$parents': undefined,
        '$current': current
      });
      if (this.as) {
        childContext[this.as] = current;
      }

      // apply bindings first, and then process child nodes, because bindings can add childnodes
      ko.applyBindingsToDescendants(childContext, templateClone);

      childNodes = ko.virtualElements.childNodes(templateClone);
    }

    // Note discussion at https://github.com/angular/angular.js/issues/7851
    allChildNodes.push.apply(allChildNodes, Array.prototype.slice.call(childNodes));
    this.firstLastNodesList.splice(index + i, 0, { first: childNodes[0], last: childNodes[childNodes.length - 1] });
  }

  this.insertAllAfter(allChildNodes, referenceElement);
};

OjForEach.prototype.getNodesForIndex = function (index) {
  var result = [],
    ptr = this.firstLastNodesList[index].first,
    last = this.firstLastNodesList[index].last;
  result.push(ptr);
  while (ptr && ptr !== last) {
    ptr = ptr.nextSibling;
    result.push(ptr);
  }
  return result;
};

OjForEach.prototype.getLastNodeBeforeIndex = function (index) {
  if (index < 1 || index - 1 >= this.firstLastNodesList.length)
    return null;
  return this.firstLastNodesList[index - 1].last;
};

OjForEach.prototype.insertAllAfter = function (nodeOrNodeArrayToInsert, insertAfterNode) {
  var frag, len, i,
    containerNode = this.element;

  // poor man's node and array check, should be enough for this
  if (nodeOrNodeArrayToInsert.nodeType === undefined && nodeOrNodeArrayToInsert.length === undefined) {
    throw new Error("Expected a single node or a node array");
  }
  if (nodeOrNodeArrayToInsert.nodeType !== undefined) {
    ko.virtualElements.insertAfter(containerNode, nodeOrNodeArrayToInsert, insertAfterNode); // @HTMLUpdateOK
    return [nodeOrNodeArrayToInsert];
  } else if (nodeOrNodeArrayToInsert.length === 1) {
    ko.virtualElements.insertAfter(containerNode, nodeOrNodeArrayToInsert[0], insertAfterNode); // @HTMLUpdateOK
  } else if (supportsDocumentFragment) {
    frag = document.createDocumentFragment();

    for (i = 0, len = nodeOrNodeArrayToInsert.length; i !== len; ++i) {
      frag.appendChild(nodeOrNodeArrayToInsert[i]);
    }
    ko.virtualElements.insertAfter(containerNode, frag, insertAfterNode); // @HTMLUpdateOK
  } else {
    // Nodes are inserted in reverse order - pushed down immediately after
    // the last node for the previous item or as the first node of element.
    for (i = nodeOrNodeArrayToInsert.length - 1; i >= 0; --i) {
      var child = nodeOrNodeArrayToInsert[i];
      if (!child) { break; }
      ko.virtualElements.insertAfter(containerNode, child, insertAfterNode); // @HTMLUpdateOK
    }
  }
  return nodeOrNodeArrayToInsert;
};

// checks if the deleted data item should be handled with delay for a possible reuse at additions
OjForEach.prototype.shouldDelayDeletion = function (data) {
  return data && (typeof data === "object" || typeof data === "function");
};

// gets the pending deletion info for this data item
OjForEach.prototype.getPendingDeleteFor = function (data) {
  var index = data && data[PENDING_DELETE_INDEX_KEY];
  if (index === undefined) return null;
  return this.pendingDeletes[index];
};

// tries to find the existing pending delete info for this data item, and if it can't, it registeres one
OjForEach.prototype.getOrCreatePendingDeleteFor = function (data) {
  var pd = this.getPendingDeleteFor(data);
  if (pd) {
    return pd;
  }
  pd = {
    data: data,
    nodesets: []
  };
  data[PENDING_DELETE_INDEX_KEY] = this.pendingDeletes.length;
  this.pendingDeletes.push(pd);
  return pd;
};

// Process a changeItem with {status: 'deleted', ...}
/**
 * @expose
 * @ignore
 */
OjForEach.prototype.deleted = function (changeItem) {
  // if we should delay the deletion of this data, we add the nodeset to the pending delete info object
  if (this.shouldDelayDeletion(changeItem.value)) {
    var pd = this.getOrCreatePendingDeleteFor(changeItem.value);
    pd.nodesets.push(this.getNodesForIndex(changeItem.index));
  } else { // simple data, just remove the nodes
    this.removeNodes(this.getNodesForIndex(changeItem.index));
  }
  this.indexesToDelete.push(changeItem.index);
};

// removes a set of nodes from the DOM
OjForEach.prototype.removeNodes = function (nodes) {
  if (!nodes.length) { return; }

  var removeFn = function () {
    var parent = nodes[0].parentNode;
    for (var i = nodes.length - 1; i >= 0; --i) {
      ko.cleanNode(nodes[i]);
      parent.removeChild(nodes[i]);
    }
  };

  removeFn();
};

// flushes the pending delete info store
// this should be called after queue processing has finished, so that data items and remaining (not reused) nodesets get cleaned up
OjForEach.prototype.flushPendingDeletes = function () {
  for (var i = 0, len = this.pendingDeletes.length; i != len; ++i) {
    var pd = this.pendingDeletes[i];
    while (pd.nodesets.length) {
      this.removeNodes(pd.nodesets.pop());
    }
    if (pd.data && pd.data[PENDING_DELETE_INDEX_KEY] !== undefined)
      delete pd.data[PENDING_DELETE_INDEX_KEY];
  }
  this.pendingDeletes = [];
};

// We batch our deletion of item indexes in our parallel array.
// See brianmhunt/knockout-fast-foreach#6/#8
/**
 * @expose
 * @ignore 
 */
OjForEach.prototype.clearDeletedIndexes = function () {
  // We iterate in reverse on the presumption (following the unit tests) that KO's diff engine
  // processes diffs (esp. deletes) monotonically ascending i.e. from index 0 -> N.
  for (var i = this.indexesToDelete.length - 1; i >= 0; --i) {
    this.firstLastNodesList.splice(this.indexesToDelete[i], 1);
  }
  this.indexesToDelete = [];
};


OjForEach.prototype.getContextStartingFrom = function (node) {
  var ctx;
  while (node) {
    ctx = ko.contextFor(node);
    if (ctx) { return ctx; }
    node = node.nextSibling;
  }
};


OjForEach.prototype.updateIndexes = function (fromIndex) {
  var ctx;
  for (var i = fromIndex, len = this.firstLastNodesList.length; i < len; ++i) {
    ctx = this.getContextStartingFrom(this.firstLastNodesList[i].first);
    if (ctx && ctx['$current']) {
      ctx['$current']['observableIndex'](i);
    }
  }
};

// Getter used in ko.computed callback that monitors array replacements 
OjForEach.prototype.getData = function () {
  return this.data;
};

// Setter used in ko.computed callback that monitors array replacements 
OjForEach.prototype.setData = function (data) {
  this.removeSubscriptions();
  this.data = data;
  this.addSubscriptions();
};

ko['bindingHandlers']['_ojBindForEach_'] = {
  // Valid valueAccessors:
  //    {data: array, as: string}
  'init': function init(element, valueAccessor, bindings, vm, context) {
    
    var ffe, value;
    
    ko.computed(function() {
      //: watch for array modifications, that are not covered by addSubscriptions() call:
      //- non-observable array replacement triggered by CCA
      //- array data defined with expression
      value = valueAccessor();
      if (ffe) {
        var updatedValues = value['data'];
        var currentValues = ffe.getData();
        var currentArray, updatedArray;
        //retrive the array data that might be defined as an observable 
        ko.ignoreDependencies(function() {
          currentArray = ko.unwrap(currentValues);
          updatedArray = ko.unwrap(updatedValues);
        });
        // initialize options for ko.utils.compareArrays
        // 'dontLimitMoves':true recommended for newer code, 'sparse':true is used for trackArray changes
        var compareArrayOptions = {'sparse':true,'dontLimitMoves':true};
        var changeSet = ko.utils.compareArrays(currentArray, updatedArray, compareArrayOptions);
        ffe.setData(updatedValues);
        ffe.onArrayChange(changeSet);
      }
    });
    
    ffe = new OjForEach(element, value, context);

    ko.utils.domNodeDisposal.addDisposeCallback(element, function () {
      ffe.removeSubscriptions();
    });
    return { 'controlsDescendantBindings': true };
  },
};

ko.virtualElements.allowedBindings['_ojBindForEach_'] = true;
})();
//
// Define a template source that allows the use of a knockout array (ko[])
// to provide storage for a template string.
//
// This simplifies template assignment and template usage for the user, as shown in the following example:
//
// Template Assignment:
//
//   ko.templates["myKey"] = templateText;
//
// Template Usage:
//
//   <div data-bind="template: {name: myKey}">
//
/*jslint browser: true*/

/**
 * @export
 */
oj.koStringTemplateEngine = {};

/**
 * @export
 */
oj.koStringTemplateEngine.install = function() 
{
    //define a template source that tries to key into an object first to find a template string

    if (ko.templates) return;

    var _templateText = {}; // Stores the text property for the template object.
    var _templateData = {}; // Stores the data property for the template object.

    // data = {},
    var _engine = new ko['nativeTemplateEngine']();

    /**
     *  @constructor
     *  @private
     */

    var StringTemplate = function(template) {

        this._templateName = template;

        this.text = function(value) 
	{
	    // When passed no parameters, return the template object.
            if (!value)
	    {
                return _templateText[this._templateName];
            }

            _templateText[this._templateName] = value;
        };

        this.data = function(key, value)
	{

            if (!_templateData[this._templateName]) {
		_templateData[this._templateName] = {};
            }

            if (arguments.length === 1) {
                return _templateData[this._templateName][key];
            }

            _templateData[this._templateName][key] = value;
        };
    };

    //
    // Override knockout's makeTemplateSource(), returning the new stringTemplate 
    //
    _engine['makeTemplateSource'] = function(template, templateDocument)
    {
	if (typeof template == "string") 
	{
            templateDocument = templateDocument || document;
            var elem = templateDocument.getElementById(template);

            if (elem) 
	    {
		return new ko['templateSources']['domElement'](elem);
	    }
            return new StringTemplate(template);
	}
        if ((template && (template.nodeType == 1)) || (template.nodeType == 8)) 
	{
            return new ko['templateSources']['anonymousTemplate'](template);
        }
    };

    //make the templates accessible
    // ko.templates = _templateText;
    ko.templates = _templateText;

    //make this new template engine our default engine
    ko['setTemplateEngine'](_engine);
};


/**
 * Returns a renderer function executes the template specified in the binding attribute.
 * (for example, a knockout template).
 * @param {Object} bindingContext the ko binding context
 * @param {string} template the name of the template
 * @return {Function} the renderer function
 * @private
 */
function _getListViewItemRenderer(bindingContext, template)
{
    return function(context)
    {
        var parent, childContext;

        parent = context['parentElement'];
        // runs the template
        childContext = bindingContext['createChildContext'](context['data'], null, 
                           function(binding)
                           {
                               binding['$itemContext'] = context;
                           }
                       );
        ko['renderTemplate'](template, childContext, {
                           'afterRender': function(renderedElement)
                           {
                               $(renderedElement)['_ojDetectCleanData']();
                           }}, parent, 'replaceNode');

        // tell the listview not to do anything
        return null;
    };
}

/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleListViewManagedAttributes(name, value, bindingContext)
{
    if (name == "item")
    {
      // find the cell template and creates a renderer
      var template = value['template'];
      if (template != null)
      {
        value['renderer'] = _getListViewItemRenderer(bindingContext, template);
      }

      return {'item': value};
    }

    return null;
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
{
  'attributes': ["item"],
  'init': function(name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext)
  {
    var result = _handleListViewManagedAttributes(name, value, bindingContext);
    if (result != null)
        return result;
  },
  'update': function(name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext)
  {
    return _handleListViewManagedAttributes(name, value, bindingContext);
  },
      
  'for': 'ojListViewRenderer'
});

/**
 * Default declaration for ojListView
 */
oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
{
  'for': 'ojListView',
  'use': 'ojListViewRenderer'
});

/**
 * Default declaration for ojListView
 */
oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
{
  'for': 'ojNavigationList',
  'use': 'ojListViewRenderer'
});


/**
 * @ojstatus preview
 * @ojcomponent oj.ojBindForEach
 * @ojshortdesc Binds items of an array to the specified markup section. The markup section is duplicated for each array item when element is rendered.
 * @ojbindingelement
 * @ojsignature {target: "Type", value:"class ojBindForEach extends JetElement<ojBindForEachSettableProperties>"}
 * @since 4.1.0
 *
 * @classdesc
 * <h3 id="oj-for-each-overview-section">
 *   ForEach Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#oj-for-each-overview-section"></a>
 * </h3>
 * <p>Use &lt;oj-bind-for-each&gt; to bind items of an array to the specified markup section. 
 * The markup section is duplicated for each array item when element is rendered. 
 * &lt;oj-bind-for-each&gt; requires the application to specify a single &lt;template&gt; element 
 * as its direct child.  The markup being stamped out should be placed inside of this &lt;template&gt; element. 
 * </p>
 * <p>During iteration, each item will have access to the same binding context that is applied to the 
 * &lt;oj-bind-for-each&gt; element. 
 * In addition the binding context will contain the following properties:
 *  <ul>
 *   <li><code>$current</code> - An object that contains information for the current array item being rendered.</li>
 *    <ul>
 *       <li><code>data</code> - The current array item being rendered.</li>
 *       <li><code>index</code> - Zero-based index of the current array item being rendered.
 *                                The index value is not updated in response to array additions and removals and
 *                                is only recommended for static arrays.</li>
 *       <li><code>observableIndex</code> - An observable that refers to the zero-based index of the current array item being rendered. 
 *                                          The <code>observableIndex</code> value is updated in response to array additions and removals
 *                                          and can be used for both static and dynamic arrays.</li>
 *    </ul>
 *   <li>alias - If <b>as</b> attribute was specified, the value will be used to provide an 
 *        application-named alias for <code>$current</code>. This can be especially useful 
 *        if multiple oj-bind-for-each elements are nested to provide access to the data 
 *        for each level of iteration.
 *   </li>
 *  </ul>
 * </p>
 * <p>Note that the &lt;oj-bind-for-each&gt; element will be removed from the DOM after binding is applied.
 * For slotting, applications need to wrap the oj-bind-for-each element inside another HTML element (e.g. &lt;span&gt;) with the slot attribute.
 * The oj-bind-for-each element does not support the slot attribute.</p>
 * <p>Also note that if the &lt;oj-bind-for-each&gt; element is being used to programmatically build an HTML table, 
 * it must be placed in the view of an oj-module and loaded via ModuleElementUtils.</p>
 *
 * @example <caption>Initialize the oj-bind-for-each - access data using <code>$current</code>:</caption>
 *  &lt;oj-bind-for-each data='[{"type":"Apple"},{"type":"Orange"}]'>
 *    &lt;template>
 *      &lt;p>&lt;oj-bind-text value='[[$current.data.type]]'>&lt;/oj-bind-text>&lt;/p>
 *    &lt;/template>
 *  &lt;/oj-bind-for-each>
 *
 * @example <caption>Initialize the oj-bind-for-each - access data using alias:</caption>
 *  &lt;oj-bind-for-each data="[[fruits]]" as="fruit">
 *    &lt;template>
 *      &lt;p>&lt;oj-bind-text value="[[fruit.data.type]]">&lt;/oj-bind-text>&lt;/p>
 *    &lt;/template>
 *  &lt;/oj-bind-for-each>
 */

/**
 * The array that you wish to iterate over. Required property. Note that the &lt;oj-bind-for-each&gt; will 
 * dynamically update the generated DOM in response to changes if the value is an observableArray.
 * @expose
 * @name data
 * @memberof oj.ojBindForEach
 * @instance
 * @type {array}
 */
 
 /**
 * An alias for the array item. This can be especially useful 
 * if multiple oj-bind-for-each elements are nested to provide access to the data 
 * for each level of iteration.
 * @expose
 * @name as
 * @memberof oj.ojBindForEach
 * @instance
 * @type {string}
 */

/**
 * @ojstatus preview
 * @ojcomponent oj.ojBindIf
 * @ojshortdesc Conditionally render its contents only if a provided test returns true.
 * @ojsignature {target: "Type", value:"class ojBindIf extends JetElement<ojBindIfSettableProperties>"}
 * @ojbindingelement
 * @since 4.1.0
 *
 * @classdesc
 * <h3 id="overview-section">
 *   If Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#overview-section"></a>
 * </h3>
 * <p>Use &lt;oj-bind-if&gt; to conditionally render its contents only if a provided test 
 * returns true. Note that the &lt;oj-bind-if&gt; element will be removed from the DOM 
 * after binding is applied. For slotting, applications need to wrap the oj-bind-if element 
 * inside another HTML element (e.g. &lt;span&gt;) with the slot attribute. The oj-bind-if element does not support 
 * the slot attribute.</p>
 *
 * @example <caption>Initialize the oj-bind-if:</caption>
 * &lt;oj-bind-if test='[[myTest]]'>
 *   &lt;div>My Contents&lt;/div>
 * &lt;/oj-bind-if>
 */

/**
 * The test condition for the if clause. The children of the element will 
 * only be rendered if the test is true.
 * @expose
 * @name test
 * @memberof oj.ojBindIf
 * @instance
 * @type {boolean}
 * @example <caption>Initialize the oj-bind-if:</caption>
 * &lt;oj-bind-if test='[[myTest]]'>
 *   &lt;div>My Contents&lt;/div>
 * &lt;/oj-bind-if>
 */

/**
 * @ojstatus preview
 * @ojcomponent oj.ojBindText
 * @ojshortdesc Binds a text node to an expression.
 * @ojsignature {target: "Type", value:"class ojBindText extends JetElement<ojBindTextSettableProperties>"}
 * @ojbindingelement
 * @since 4.1.0
 *
 * @classdesc
 * <h3 id="overview-section">
 *   Text Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#overview-section"></a>
 * </h3>
 * <p>Use &lt;oj-bind-text&gt; to bind a text node to a variable. Note that the
 * &lt;oj-bind-text&gt; element will be removed from the DOM after binding is
 * applied, and any child elements it has will be removed. For slotting, applications 
 * need to wrap the oj-bind-text element inside another HTML element (e.g. &lt;span&gt;) with the slot attribute. 
 * The oj-bind-text element does not support the slot attribute.</p>
 *
 * @example <caption>Initialize the oj-bind-text:</caption>
 * &lt;span>
 *   &lt;oj-bind-text value='[[myText]]'>&lt;/oj-bind-text>
 * &lt;/span>
 */

/**
 * The value of the text node.
 * @expose
 * @name value
 * @memberof oj.ojBindText
 * @instance
 * @type {string}
 * @example <caption>Initialize the oj-bind-text:</caption>
 * &lt;span>
 *   &lt;oj-bind-text value='[[myText]]'>&lt;/oj-bind-text>
 * &lt;/span>
 */

/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleManagedGaugeAttributes(name, value, bindingContext) {
  if (name === "center" && value['template']) {
    value['_renderer'] = _getDvtRenderer(bindingContext, value['template']);
  }
  return {'center': value};
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({
  'attributes': ["center"],
  'init': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedGaugeAttributes(name, value, bindingContext);
  },
  'update': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedGaugeAttributes(name, value, bindingContext);
  },
  'for': 'ojStatusMeterGauge'
});

/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleManagedSunburstAttributes(name, value, bindingContext) {
  if (name === "rootNodeContent" && value['template']) {
    value['_renderer'] = _getDvtRenderer(bindingContext, value['template']);
  }
  return {'rootNodeContent': value};
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({
  'attributes': ["rootNodeContent"],
  'init': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedSunburstAttributes(name, value, bindingContext);
  },
  'update': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedSunburstAttributes(name, value, bindingContext);
  },
  'for': 'ojSunburst'
});

/*jslint browser: true, devel: true*/

/** 
 * @private
 * @const
 */
var _COLUMNS_ATTR = 'columns';

/** 
 * @private
 * @const
 */
var _COLUMNS_DEFAULT_ATTR = 'columnsDefault';

/** 
 * The row template will be used to render the row elements. 
 * The row, status, and component objects will be available 
 * in the template context.
 * @private
 * @const
 */
var _ROW_TEMPLATE_ATTR = 'rowTemplate';

/**
 * Create and return a renderer which the component will call. That renderer
 * will render the template.
 * @param {Object} bindingContext  Binding Context
 * @param {string} type  'cell' or 'header' or 'row'
 * @param {string} template  template name
 * @return {Object} Renderer
 * @private
 */
function _getTableColumnTemplateRenderer(bindingContext, type, template)
{
  var rendererOption = {};
  (function(template, type) {
    rendererOption = function(params) {
      var childContext = null;
      var parentElement = null;
      if (type == 'header')
      {
        childContext = bindingContext['createChildContext'](null, null,
          function(binding)
          {
            binding['$columnIndex'] = params['columnIndex'];
            binding['$headerContext'] = params['headerContext'];
            binding['$data'] = params['data'];
          });
        parentElement = params['headerContext']['parentElement'];
      }
      else if (type == 'cell')
      {
        var childData = params['row'];
        childContext = bindingContext['createChildContext'](childData, null,
          function(binding)
          {
            binding['$columnIndex'] = params['columnIndex'];
            binding['$cellContext'] = params['cellContext'];
          }
        );
        parentElement = params['cellContext']['parentElement'];
      }
      if (type == 'footer')
      {
        childContext = bindingContext['createChildContext'](null, null,
          function(binding)
          {
            binding['$columnIndex'] = params['columnIndex'];
            binding['$footerContext'] = params['footerContext'];
          });
        parentElement = params['footerContext']['parentElement'];
      }
      ko['renderTemplate'](template, childContext, {
                           'afterRender': function(renderedElement)
                           {
                               $(renderedElement)['_ojDetectCleanData']();
                           }}, parentElement, 'replaceNode');
    };
  }(template, type));

  return rendererOption;
}

/**
 * Create and return a renderer which the component will call. That renderer
 * will render the template.
 * @param {string} template  template name
 * @return {Object} Renderer
 * @private
 */
function _getTableRowTemplateRenderer(bindingContext, template)
{
  return function(params) {
    var childData = params['row'];
    var childContext = bindingContext['createChildContext'](childData, null,
      function(binding)
      {
        binding['$rowContext'] = params['rowContext'];
      }
    );
    ko['renderTemplate'](template, childContext, {
                           'afterRender': function(renderedElement)
                           {
                               $(renderedElement)['_ojDetectCleanData']();
                           }}, params['rowContext']['parentElement'], 'replaceNode');

  };
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
  {
    'attributes': [_COLUMNS_ATTR, _COLUMNS_DEFAULT_ATTR, _ROW_TEMPLATE_ATTR],
    'init': function(name, value, element, widgetConstructor, valueAccessor,
      allBindingsAccessor, bindingContext)
    {
      if (name == _COLUMNS_ATTR || name == _COLUMNS_DEFAULT_ATTR)
      {
        var i, template, footerTemplate, headerTemplate;
        for (i = 0; i < value.length; i++)
        {
          var column = value[i];
          template = column['template'];
          footerTemplate = column['footerTemplate'];
          headerTemplate = column['headerTemplate'];

          if (template != null)
          {
            column['renderer'] = _getTableColumnTemplateRenderer(bindingContext, 'cell', template);
          }
          if (footerTemplate != null)
          {
            column['footerRenderer'] = _getTableColumnTemplateRenderer(bindingContext, 'footer', footerTemplate);
          }
          if (headerTemplate != null)
          {
            column['headerRenderer'] = _getTableColumnTemplateRenderer(bindingContext, 'header', headerTemplate);
          }
        }
        if (name == _COLUMNS_ATTR)
        {
          return {'columns': value};
        }
        else
        {
          return {'columnsDefault': value};
        }
      }
      else if (name == _ROW_TEMPLATE_ATTR)
      {
        return {'rowRenderer': _getTableRowTemplateRenderer(bindingContext, value)};
      }
    },
    'update': function(name, value, element, widgetConstructor, valueAccessor,
      allBindingsAccessor, bindingContext)
    {
      if (name == _COLUMNS_ATTR || name == _COLUMNS_DEFAULT_ATTR)
      {
        var i, template, footerTemplate, headerTemplate;
        for (i = 0; i < value.length; i++)
        {
          var column = value[i];
          template = column['template'];
          footerTemplate = column['footerTemplate'];
          headerTemplate = column['headerTemplate'];

          if (template != null)
          {
            column['renderer'] = _getTableColumnTemplateRenderer(bindingContext, 'cell', template);
          }
          if (footerTemplate != null)
          {
            column['footerRenderer'] = _getTableColumnTemplateRenderer(bindingContext, 'footer', footerTemplate);
          }
          if (headerTemplate != null)
          {
            column['headerRenderer'] = _getTableColumnTemplateRenderer(bindingContext, 'header', headerTemplate);
          }
        }
        if (name == _COLUMNS_ATTR)
        {
          widgetConstructor({'columns': value});
        }
        else
        {
          widgetConstructor({'columnsDefault': value});
        }
      }
      else if (name == _ROW_TEMPLATE_ATTR)
      {
        return {'rowRenderer': _getTableRowTemplateRenderer(bindingContext, value)};
      }
      return null;
    },
    'for': 'ojTable'
  });

/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleManagedMapAttributes(name, value, bindingContext) {
  if (name === "areaLayers") {
    for (var i = 0; i < value.length; i++) {
      var areaDataLayer = value[i]['areaDataLayer'];
      if (areaDataLayer) {
        var template = areaDataLayer['template'];
        if (template != null) {
          areaDataLayer['_templateRenderer'] = _getDvtDataRenderer(bindingContext, template);
        }
      }
    }
    return {'areaLayers': value};
  }
  else if (name === "pointDataLayers") {
    for (var i = 0; i < value.length; i++) {
      var template = value[i]['template'];
      if (template != null) {
        value[i]['_templateRenderer'] = _getDvtDataRenderer(bindingContext, template);
      }
    }
    return {'pointDataLayers': value};
  }
  return null;
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({
  'attributes': ["areaLayers", "pointDataLayers"],
  'init': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedMapAttributes(name, value, bindingContext);
  },
  'update': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedMapAttributes(name, value, bindingContext);
  },
  'for': 'ojThematicMap'
});

/**
 * Common method to handle managed attributes for both init and update
 * @param {string} name the name of the attribute
 * @param {Object} value the value of the attribute
 * @param {Object} bindingContext the ko binding context
 * @return {Object} the modified attribute
 * @private
 */
function _handleManagedTreemapAttributes(name, value, bindingContext) {
  if (name === "nodeContent" && value['template']) {
    value['_renderer'] = _getDvtRenderer(bindingContext, value['template']);
  }
  return {'nodeContent': value};
}

oj.ComponentBinding.getDefaultInstance().setupManagedAttributes({
  'attributes': ["nodeContent"],
  'init': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedTreemapAttributes(name, value, bindingContext);
  },
  'update': function (name, value, element, widgetConstructor, valueAccessor, allBindingsAccessor, bindingContext) {
    return _handleManagedTreemapAttributes(name, value, bindingContext);
  },
  'for': 'ojTreemap'
});

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true, devel: true*/

/**
 * Tracks the validity of a group of components bound to this observable. It also provides 
 * properties and methods that the page author can use to enforce form validation best practice.
 * 
 * <p>
 * Validity of components that are disabled or readOnly will not be tracked by this object.</li>
 * </p>
 * 
 * <p>
 * The <code class="prettyprint">invalidComponentTracker</code> binding attribute should be bound to 
 * a ko observable, and 'ojs/ojknockout-validation' should be included in the dependency list. 
 * At runtime the framework creates an instance of this type 
 * <code class="prettyprint">oj.InvalidComponentTracker</code> and sets it on the bound observable.<br/> 
 * This object can then be used by page authors to do the following - 
 * <ul>
 * <li>determine if there are invalid components tracked by this object that are currently showing 
 * errors.</li>
 * <li>determine if there are invalid components tracked by this object that are currently deferring 
 * showing errors.</li>
 * <li>set focus on the first invalid component in the tracked group</li>
 * <li>show all messages on all tracked components including deferred error messages, and set focus 
 * on the first invalid component.</li>
 * </ul>
 * </p>
 * 
 * @example <caption> Bind an observable to the <code class="prettyprint">invalidComponentTracker</code> and access oj.InvalidComponentTracker instance.</caption>
 * &lt;input id="username" type="text" required 
 *   data-bind="ojComponent: {component: 'ojInputText', value: userName, 
 *    invalidComponentTracker: tracker}">
 * &lt;input id="password" type="text" required 
 *   data-bind="ojComponent: {component: 'ojInputPassword', value: password, 
 *    invalidComponentTracker: tracker}">
 * 
 * &lt;script>
 * function MyViewModel() { 
 *   var self = this;
 *   self.tracker = ko.observable();
 *   
 *   log = function ()
 *   {
 *     var trackerObj = ko.utils.unwrapObservable(self.tracker);
 *     console.log(trackerObj instanceof oj.InvalidComponentTracker); // true, so safe to call InvalidComponentTracker methods and properties. 
 *   }
 *   
 *   self.focusOnFirstInvalid = function() 
 *   {
 *      var trackerObj = ko.utils.unwrapObservable(self.tracker);
 *      if (trackerObj !== undefined)
 *      {
 *        // make sure the trackerObj is an oj.InvalidComponentTracker 
 *        // before calling methods on it.
 *        if (trackerObj instanceof oj.InvalidComponentTracker)
 *        {
*           // showMessages first
*           // (this will show any hidden messages, if any)
*           trackerObj.showMessages();
*           // focusOnFirstInvalid will focus on the first component
*           // that is invalid, if any.      
*           trackerObj.focusOnFirstInvalid();
*         }
*       }
*     }
 * }
 * &lt;/script>
 * Note: Make sure you have included the 'ojs/ojknockout-validation' dependency in your require list,
 * otherwise you will see an error about InvalidComponentTracker being undefined.
 * 
 * @class oj.InvalidComponentTracker 
 * @constructor
 * @class
 * @export
 * @since 0.7
 * @ignore
 */
oj.InvalidComponentTracker = function()
{
  this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.InvalidComponentTracker, oj.Object, "oj.InvalidComponentTracker");

// DOCLETS for public properties

/**
 * Whether there is at least one component (tracked by this object) that is invalid and is currently 
 * showing messages. 
 * 
 * 
 * @example <caption>Disable button using <code class="prettyprint">invalidShown</code> property:</caption>
 * &lt;input id="username" type="text" required 
 *   data-bind="ojComponent: {component: 'ojInputText', value: userName, 
 *    invalidComponentTracker: tracker}">
 * &lt;input id="password" type="text" required 
 *   data-bind="ojComponent: {component: 'ojInputPassword', value: password, 
 *    invalidComponentTracker: tracker}">
 * &lt;button type="button" data-bind="ojComponent: {component: 'ojButton', label: 'Create',
 *   disabled: tracker()['invalidShown']}"></button>
 *   
 * &lt;script>
 * var userName = ko.observable();
 * var password = ko.observable();
 * var tracker = ko.observable();
 * &lt;/script>
 * 
 * @member 
 * @name invalidShown
 * @access public
 * @instance
 * @default false 
 * @type {boolean}
 * @expose
 * @memberof oj.InvalidComponentTracker
 */


/** 
 * prevent preceding jsdoc from applying to following line of code 
 * @ignore
 */
//Options we are interested in listening to changes for.
oj.InvalidComponentTracker._OPTION_MESSAGES_SHOWN = 'messagesShown';
oj.InvalidComponentTracker._OPTION_MESSAGES_HIDDEN = 'messagesHidden';
oj.InvalidComponentTracker._OPTION_DISABLED = 'disabled';
oj.InvalidComponentTracker._OPTION_READONLY = 'readOnly';

/**
 * Whether there is at least one component that is invalid with deferred messages, i.e., messages 
 * that are currently hidden. For example, when a component is 
 * required, deferred validation is run when the component is created. 
 * Any validation error raised is not shown to user right away, i.e., it is deferred.
 * 
 * @example <caption>Enable button using <code class="prettyprint">invalidHidden</code> property:</caption>
 * &lt;input id="username" type="text" required 
 *   data-bind="ojComponent: {component: 'ojInputText', value: userName, 
 *    invalidComponentTracker: tracker}">
 * &lt;input id="password" type="text" required 
 *   data-bind="ojComponent: {component: 'ojInputPassword', value: password, 
 *    invalidComponentTracker: tracker}">
 * &lt;br/>
 * &lt;button type="button" data-bind="ojComponent: {component: 'ojButton', label: 'Create',
 *   disabled: !tracker()['invalidHidden']}"></button>
 * 
 * &lt;script>
 * var userName = ko.observable();
 * var password = ko.observable();
 * var tracker = ko.observable();
 * &lt;/script>
 * 
 * @member 
 * @name invalidHidden
 * @access public
 * @instance
 * @default false 
 * @type {boolean}
 * @expose
 * @memberof oj.InvalidComponentTracker
 */
/**
 * Initializer
 * @protected
 * @memberof oj.InvalidComponentTracker
 * @instance
 */
oj.InvalidComponentTracker.prototype.Init = function()
{
  oj.InvalidComponentTracker.superclass.Init.call(this);

  // INTERNAL PROPERTIES
  // Array of Objects containing component and its element 
  // that are being tracked by invalidComponentTracker
  // [{'component': component, 'element': element}]
  this._tracked = [];

  // tracks invalid components showing messages. indices correspond to this_tracked.
  // e.g., [true, false, false, false, true]
  this._invalid = [];

  // tracks invalid components hiding messages. indices correspond to this_tracked.
  // e.g., [false, true, false, false, false]
  this._invalidHidden = [];

  // PUBLIC PROPERTIES
  // Whether there is at least one component 
  // (tracked by this object) that is invalid and is currently showing messages.
  this['invalidShown'] = false;
  // Whether there is at least one component that is invalid with deferred messages, 
  // i.e., messages that are currently hidden.
  this['invalidHidden'] = false;
};

/**
 * Sets focus on first invalid component currently showing an error. This method does not set focus 
 * on components that are invalid and have deferred messages. For example, when a component is 
 * required, deferred validation is run. Any validation error raised is not shown to user right away,
 *  i.e., it is deferred. 
 *  <p>
 *  To show hidden messages on all tracked components use showMessages() method. </p>
 * 
 * @return {boolean} true if there is at least one invalid component to set focus on; false if 
 * unable to locate a component to focus on or there are no invalid components. 
 * @export
 * @see #showMessages
 * @memberof oj.InvalidComponentTracker
 * @instance
 */
oj.InvalidComponentTracker.prototype.focusOnFirstInvalid = function()
{
  var firstInvalid = null;
  var self = this;
  var updateCounter = this._updateCounter;
  var busyContext;
  var resolveBusyState;
  
  // we don't always have a element associated, so this is on the page level context
  busyContext = oj.Context.getPageContext().getBusyContext();
  resolveBusyState = busyContext.addBusyState(
    {"description" : "Setting Focus to first invalid component."});

  if (this['invalidShown'])
  {
    firstInvalid = this._getFirstInvalidComponent();
  }

  // always call focus handler on a timer to give time for binding layer to
  // apply changes on the component.
  setTimeout(function () {

    // sometimes when this timer is called, firstInvalid may not have been determined 
    // yet. Or the invalid states could have changed in between the timer being set and the 
    // callback being called.
    // TODO: talk to Pavitra regarding this logic, it seems unnecessary. 
    firstInvalid = (updateCounter === self._updateCounter) ? 
                   firstInvalid || self._getFirstInvalidComponent() :
                   self._getFirstInvalidComponent(); 

    if (firstInvalid)
    {
      // Get the widget instance for the firstInvalid so we can call protected methods directly.
      firstInvalid('instance').GetFocusElement().focus();
    }

    resolveBusyState();
  }, 1);
  
  return firstInvalid ? true : false;
};

/**
 * Shows hidden messages on all tracked components by calling showMessages() method on each tracked 
 * editable component. 
 * 
 * @example <caption>Show all hidden messages on tracked components:</caption>
 *  function ViewModel ()
 *  {
 *    self = this;
 *    var tracker = ko.observable();
 *   // ...
 *   
 *   showAllMessages : function ()
 *   {
 *      var trackerObj = ko.utils.unwrapObservable(self.tracker);
        return trackerObj.showMessages();
 *    }
 *  }
 * 
 * @export
 * @memberof oj.InvalidComponentTracker
 * @instance
 * @see oj.editableValue#showMessages
 */
oj.InvalidComponentTracker.prototype.showMessages = function()
{
  var index;  
  var len;
  var tr;

  if (this['invalidHidden'])
  {
    len = this._invalidHidden.length;
    for (index = 0; index < len; index++)
    {
      if (this._invalidHidden[index])
      {
        tr = (this._tracked[index]['component']).call(tr, "showMessages");
      }
    }
  }
};

// P R I V A T E    M E T H O D S
/**
 * Gets the first invalid component and returns the component to focus on.
 * 
 * @returns the component instance that has focus or null
 * @private
 * @memberof oj.InvalidComponentTracker
 * @instance
 */
oj.InvalidComponentTracker.prototype._getFirstInvalidComponent = function()
{
  var isInvalid;
  var idx = 0;
  var invalidComponents = [];
  var len = this._invalid.length;

  // locate first invalid component and set focus on it
  // this._invalid is an array of booleans that tracks invalid components showing messages.
  // e.g., [false, true, true, false]. This maps to the this._tracked which holds the 
  // components. So the components showing messages is the ones at indices 1 and 2.
  // this._tracked may not be in DOM order if a new tracked component was added not
  // at the end of the DOM order.
  for (idx = 0; idx < len; idx++)
  {
    isInvalid = this._invalid[idx];
    if (isInvalid)
    {
      invalidComponents.push(this._tracked[idx]);
    }
  }
  
  if (invalidComponents.length === 0)
  {
    return null;
  }
  
  // sort the invalidComponents based on dom order
  invalidComponents.sort(function(a, b) {
    var elementA = a['element'];
    var elementB = b['element'];
    // If elementA precedes elementB in dom order, return -1
    return (elementA.compareDocumentPosition(elementB) & Node.DOCUMENT_POSITION_FOLLOWING) ? -1 : 1;
  });

  // invalidComponents is sorted now by document order, so return the first one.
  return invalidComponents[0]['component'];
};

/**
 * Removes the component from its tracked list. 
 * 
 * @param {Object} component being removed
 * @returns {boolean} if internal state mutated; false otherwise
 * @private
 * @memberof oj.InvalidComponentTracker
 * @instance
 */
oj.InvalidComponentTracker.prototype._remove = function(component)
{
  var trackedIndex = -1;
  var mutated = false;

  // locate the index in tracked, for the component that was updated 
  $.each(this._tracked, function(index, item)
  {
    if (trackedIndex < 0 && item['component'] === component)
    {
      trackedIndex = index;
      return;
    }
  });

  if (trackedIndex >= 0)
  {
    this._tracked.splice(trackedIndex, 1);
    // stop tracking them in the internal arrays 
    this._invalid.splice(trackedIndex, 1);
    this._invalidHidden.splice(trackedIndex, 1);
    this._updateInvalidProperties();
    mutated = true;
  }

  return mutated;
};


/**
 * Updates the internal properties to reflect the current validity state of the component, using 
 * new messages.
 * 
 * @param {Object} component the component that has the new messages
 * @param {Element} element the element that has the new messages (this is the component's element)
 * @param {string} option the option being updated. e.g., messagesShown, messagesHidden, disabled,
 *  readOnly
 * @param {Array.<string>| Array.<oj.Message>} value the value of the option. For example, if option is 'messagesShown'
 *   the value is the Array of the messagesShown, if empty value is [].
 * @returns {boolean} if internal state mutated; false otherwise
 * @private
 * @memberof oj.InvalidComponentTracker
 * @instance
 */
oj.InvalidComponentTracker.prototype._update = function(component, element, option, value)
{
  var compValid = component.call(component, "isValid");
  var isDisabled;
  var isReadOnly;
  var mutated = true;
  var result;
  var trackedIndex = -1;

  // locate the index in tracked, for the component that was updated.
  $.each(this._tracked, function(index, item)
  {
    if (trackedIndex < 0 && item['component'] === component)
    {
      trackedIndex = index;
      return;
    }
  });

  switch (option)
  {
    case oj.InvalidComponentTracker._OPTION_MESSAGES_SHOWN:
    case oj.InvalidComponentTracker._OPTION_MESSAGES_HIDDEN:

      result = false;
      if (value)
      {
        // start tracking component if not already doing it.
        if (trackedIndex < 0)
        {
          trackedIndex = this._tracked.push({'component': component, 'element': element}) - 1;
          // adds the trackedIndex/result to this._invalid and this._invalidHidden boolean arrays,
          // keeping this._tracked in sync with this._invalid and this._invalidHidden.
          // adds the trackedIndex/result to this._invalid and this._invalidHidden boolean arrays,
          // keeping this._tracked in sync with this._invalid and this._invalidHidden.
          this._initializeInvalidTrackers(trackedIndex, result);
        }

        if (!compValid)
        {
          if (oj.InvalidComponentTracker._hasInvalidMessages(value))
          {
            result = true;
            if (option === oj.InvalidComponentTracker._OPTION_MESSAGES_SHOWN)
            {
              // if component is disabled or readOnly but has messages showing, tracker stops 
              // tracking component in its 'invalidShown' list. We do this because if property is 
              // bound to a button, and if the only invalid component showing messages is disabled 
              // the button would appear disabled visually, confusing the end-user. 
              // E.g., disabled component can be initialized with messagesCustom making it invalid and 
              // disabled.

              isDisabled = component.call(component, "option", oj.InvalidComponentTracker._OPTION_DISABLED),
              isReadOnly = component.call(component, "option", oj.InvalidComponentTracker._OPTION_READONLY);
              result = isDisabled || isReadOnly ? false : true;
            }
          }
        }

        mutated = this._updateInvalidTracker(option, trackedIndex || 0, result);
        this._updateInvalidProperties();

        // update properties
        if (mutated)
        {
          // every time messages mutates, we track it. 
          if (this._updateCounter === undefined)
          {
            this._updateCounter = 0;
          }
          this._updateCounter++;
        }
      }
      break;

    case oj.InvalidComponentTracker._OPTION_DISABLED:
    case oj.InvalidComponentTracker._OPTION_READONLY:

      // when component goes from enabled to disabled (or to readOnly) tracker updates invalidShown 
      // to be false, since the component cannot be showing errors visually. Same goes for 
      // invalidHidden.
      // 
      // when component goes from disabled (or readOnly) to enabled validations are re-run and 
      // component's messagesHidden and messagesShown are updated which result in invalidShown and 
      // invalidHidden to be updated. This case is not handled here.

      mutated = false;
      if (value)
      {
        mutated = this._updateInvalidTracker(oj.InvalidComponentTracker._OPTION_MESSAGES_SHOWN,
                trackedIndex || 0, false);
        mutated = this._updateInvalidTracker(oj.InvalidComponentTracker._OPTION_MESSAGES_HIDDEN,
                trackedIndex || 0, false) || mutated;
        this._updateInvalidProperties();
      }
      break;
  }
  return mutated;
};

/**
 * This is called when we are starting to track a component. Before this function is called,
 * we push the component and element into the this._tracked array and then we call this 
 * with the trackedIndex
 * 
 * @param {number} trackedIndex
 * @param {boolean} result true or false. Since this is initializing the this._invalid and 
 * this._invalidHidden arrays, it really only makes sense if it is false but it 
 *  isn't enforced.
 * @private
 * @memberof oj.InvalidComponentTracker
 * @instance
 */
oj.InvalidComponentTracker.prototype._initializeInvalidTrackers = function(trackedIndex, result)
{
  if (this._invalid[trackedIndex] === undefined)
    this._updateInvalidTracker(
      oj.InvalidComponentTracker._OPTION_MESSAGES_SHOWN, trackedIndex, result);
  if (this._invalidHidden[trackedIndex] === undefined)
    this._updateInvalidTracker(
      oj.InvalidComponentTracker._OPTION_MESSAGES_HIDDEN, trackedIndex, result);
};

oj.InvalidComponentTracker.prototype._updateInvalidProperties = function()
{
  // updates public properties exposed by this object
  this['invalidShown'] = this._invalid.indexOf(true) >= 0;
  this['invalidHidden'] = this._invalidHidden.indexOf(true) >= 0;
};

/**
 * If option is "messagesShown", then update this._invalid boolean array, 
 * else if option is "messagesHidden", update this._invalidHidden boolean array
 * using the trackedIndex and the value.
 * this._invalid is a boolean array, e.g., [false, false, false, true, false], indicating
 * which components have invalid messages showing.
 * this._invalidHidden is also a boolean array indicating which components have invalid messages
 * hidden. The components are tracked in this._tracked. The indices correspond. For example,
 * the component at this._tracked[3]['component'] is showing invalid messages 
 * if this._invalid[3] is true.
 * @private
 * @memberof oj.InvalidComponentTracker
 * @instance
 * @param {string} option "messagesShown" or "messagesHidden".
 * @param {number} trackedIndex the index of the tracked component
 * @param {boolean} value true or false
 * @returns {boolean} whether or not this._invalid or this._invalidHidden was mutated
 */
oj.InvalidComponentTracker.prototype._updateInvalidTracker = function(option, trackedIndex, value)
{
  var arr;
  var mutated = false;
   
  arr = (option === oj.InvalidComponentTracker._OPTION_MESSAGES_SHOWN) ? 
        this._invalid : (option === oj.InvalidComponentTracker._OPTION_MESSAGES_HIDDEN) ? 
          this._invalidHidden : [];

  // updates or pushes the value into the appropriate array
  if (trackedIndex >= 0 && arr[trackedIndex] !== undefined)
  {
    // mark component as invalid or invalidHidden to match the trackedIndex; update only if value 
    // changes
    mutated = arr[trackedIndex] !== value ? true : false;
    if (mutated)
    {
      arr[trackedIndex] = value;
    }
  }
  else
  {
    // new
    arr.push(value);
    mutated = true;
  }

  return mutated;
};

/**
 * helper to determine if we have invalid messages among the list of messages that are currently 
 * showing i.e., that are showing.
 * 
 * @param {!Array.<oj.Message>} messages list of all messages associated with component
 * @returns {boolean}
 * @private
 * @memberof oj.InvalidComponentTracker
 * @instance
 */
oj.InvalidComponentTracker._hasInvalidMessages = function(messages)
{
  return !oj.Message.isValid(messages);
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true, devel: true*/

// private to prevent creating a JSDoc page for this class.  The only thing we wish 
// to JSDoc is the invalidComponentTracker, which we're putting in EditableValue's output.
/**
 * An extension to oj.ComponentBinding, properties exposed on this binding are available 
 * to jet components that extend from oj.editableValue. 
 * 
 * @private
 * @constructor oj.ValueBinding 
 * @see oj.ComponentBinding
 * @see oj.editableValue
 * @since 0.6
 */
oj.ValueBinding = function(){};


/** 
 * <p>When this attribute is bound to an observable, the framework pushes 
 * an object of type {@link oj.InvalidComponentTracker} onto the observable. 
 * The object itself tracks the validity of a group of editable components.
 * 
 * <p>When this attribute is present, the binding registers a listener 
 * for the <a href="#optionChange">optionChange</a>
 * event. This event is fired by JET editable components whenever its validity changes (i.e. when 
 * <a href="#messagesShown">messagesShown</a> or <a href="#messagesHidden">messagesHidden</a> 
 * options change). When the event is fired, the listener determines the current validity of the 
 * component and updates the tracker.
 * 
 * <p>
 * The observable bound to this attribute is often used with multiple component binding declarations 
 * as shown in the example below.
 * </p>
 * 
 * <p>
 * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and 
 * is not a component option.
 * </p>
 * 
 * @example <caption>Track validity of multiple components using a single observable 
 * bound to the <code class="prettyprint">invalidComponentTracker</code> attribute:</caption>
 * &lt;input id="username" type="text" name="username" required 
 *   data-bind="ojComponent: {component: 'ojInputText', value: userName, 
 *                            invalidComponentTracker: tracker}">
 * 
 * &lt;input id="password" type="password" name="password" required 
 *   data-bind="ojComponent: {component: 'ojInputPassword', value: password, 
 *                            invalidComponentTracker: tracker}"/>
 *   
 * // ViewModel that defines the tracker observable
 * &lt;script>
 * function MemberViewModel() 
 * { 
 *   var self = this;
 *   
 *   self.tracker = ko.observable();
 *   
 *   self.userName = ko.observable();   
 *   self.password = ko.observable();
 *   
 *   self.focusOnFirstInvalid = function() 
 *   {
 *      var trackerObj = ko.utils.unwrapObservable(self.tracker);
 *      if (trackerObj !== undefined)
 *      {
 *        // make sure the trackerObj is an oj.InvalidComponentTracker 
 *        // before calling methods on it.
 *        if (trackerObj instanceof oj.InvalidComponentTracker)
 *        {
*           // showMessages first
*           // (this will show any hidden messages, if any)
*           trackerObj.showMessages();
*           // focusOnFirstInvalid will focus on the first component
*           // that is invalid, if any.      
*           trackerObj.focusOnFirstInvalid();
*         }
*       }
*     }
 * }
 * &lt;/script>
 * 
 * @example <caption>Use tracker property <code class="prettyprint">invalid</code> to disable button:</caption>
 * // button is disabled if there are components currently showing errors
 * &lt;button type="button" data-bind="ojComponent: {component: 'ojButton', label: 'Submit',
 *                                                disabled: tracker()['invalidShown']}"></button>
 * 
 * @ojbindingonly
 * @member
 * @name invalidComponentTracker
 * @memberof oj.editableValue
 * @instance
 * @type {oj.InvalidComponentTracker}
 * @default <code class="prettyprint">null</code>
 * @since 0.7
 * @ignore
 */
/** prevent preceding jsdoc from applying to following line of code */

oj.ValueBinding._ATTRIBUTE_INVALID_COMPONENT_TRACKER = 'invalidComponentTracker';

// A listener is added for this event to listen to changes to the 'messagesHidden' or 
// 'messagesShown' options. The listener updates the InvalidComponentTracker.
oj.ValueBinding._EVENT_OPTIONCHANGE = "ojoptionchange";

// Options we are interested in listening to changes for.
oj.ValueBinding._OPTION_MESSAGES_SHOWN = 'messagesShown';
oj.ValueBinding._OPTION_MESSAGES_HIDDEN = 'messagesHidden';

// options that are managed primarily to detect changes for tracker to be notified.
oj.ValueBinding._OPTION_DISABLED = 'disabled';
oj.ValueBinding._OPTION_READONLY = 'readOnly';


// callback called when managed attribute is being updated
oj.ValueBinding._update = function(name, value, element, component, valueAccessor)
{
  var options = valueAccessor.call();
  var updateProps = {};
  var ictObs = options[oj.ValueBinding._ATTRIBUTE_INVALID_COMPONENT_TRACKER];
  var icTracker;
    
  if (name === oj.ValueBinding._OPTION_DISABLED || name === oj.ValueBinding._OPTION_READONLY)
  {
    icTracker = ictObs && ictObs.peek() || null; // don't add extra subscriptions
    // when either of these options are updated
    if (icTracker !== null && ko.isWriteableObservable(ictObs))
    {
      if (icTracker._update.call(icTracker, component, element, name, value))
      {
        // if _update mutates state
        ictObs.valueHasMutated();
      }
    }
    updateProps[name] = value;
    return updateProps;
  }
};

// init callback for managed attributes. When managing options like disabled, readOnly
// this method is required to return values.
oj.ValueBinding._init = function(name, value)
{
  var initProps = {};

  initProps[name] = value;
  return initProps;
};

/**
 * Called after component binding creates the component. 
 * @param {string} property 
 * @param {Element} element the element to which binding applied the componnet
 * @param {Function=} component the widget bridge 
 * @param {Object=} valueAccessor 
 * @private
 */
oj.ValueBinding._afterCreate = function(property, element, component, valueAccessor)
{
  var initProps = {};
  var optionsSet = valueAccessor.call();
  var isICTOptionSet;

  if (property ===  oj.ValueBinding._ATTRIBUTE_INVALID_COMPONENT_TRACKER)
  {
    isICTOptionSet = optionsSet[property] ? true : false;

    if (isICTOptionSet)
    {
      // register a writeback for invalidComponentTracker property by registering an event listener
      //  for the optionChange event.
      oj.ValueBinding._registerInvalidComponentTrackerWriteback(property, optionsSet, element, component);
    }
  }

  return initProps;
};

/**
 * Called right before component is destroyed.
 *  
 * @param {Element} element
 * @private
 */
oj.ValueBinding._beforeDestroy = function(property, element, component, valueAccessor)
{
  var jelem = $(element);
  var options = valueAccessor.call();
  var icTracker;
  var ictObs = options[property];

  if (property ===  oj.ValueBinding._ATTRIBUTE_INVALID_COMPONENT_TRACKER)
  {
    if (jelem)
    {
      jelem.off(oj.ValueBinding._EVENT_OPTIONCHANGE, oj.ValueBinding._updateInvalidComponentTracker);
      if (ictObs && ko.isWriteableObservable(ictObs))
      {
        icTracker = ictObs.peek(); 
        // remove component from tracker
        if (icTracker._remove.call(icTracker, component))
        {
          // if _remove mutates state, then components need to react to it.
          // example a button that binds to properties on invalidComponentTracker.
          ictObs.valueHasMutated();
        }
      }
    }
  }
};

/**
 * Listener for the optionChange event, it updates the invalidComponentTracker associated to the 
 * component that triggered the event.
 * 
 * @param {jQuery.event=} event
 * @private
 */
oj.ValueBinding._updateInvalidComponentTracker = function(event)
{
  var ictObs = event.data.tracker;
  var icTracker;
  var component = event.data.component;
  var element = event.data.element;
  var payload = arguments[1];
  var option = payload['option']; 
  var msgs = payload['value'];
    
  if (option === oj.ValueBinding._OPTION_MESSAGES_SHOWN || 
      option === oj.ValueBinding._OPTION_MESSAGES_HIDDEN)
  {
    if (ictObs && ko.isWriteableObservable(ictObs))
    {
      icTracker = ictObs.peek();
      if (icTracker && icTracker._update.call(icTracker, component, element, option, msgs))
      {
        // if _update mutates state
        ictObs.valueHasMutated();
      }
    }
  }
};


/**
 * Register a default callback for the 'optionChange' event. The callback writes the component and 
 * its validity to the invalidComponentTracker observable.
 * @param {string} property
 * @param {Object} options original options set on element 
 * @param {Element} element
 * @param {Function=} component
 * @private
 */
oj.ValueBinding._registerInvalidComponentTrackerWriteback = function(property, options, element, component)
{
  var ictObs = options[property];
  var messagesShown;
  var messagesHidden;
  var eventData;
  var icTracker;
  var jElem = $(element);

  // Create new intsance of InvalidComponentTracker if the observable is not set.
  if (ko.isObservable(ictObs))
  {
    icTracker = ictObs.peek();
    // push new instance of oj.InvalidComponentTracker onto observable if none present.
    if (icTracker == null) // null or undefined
    {
      icTracker = new oj.InvalidComponentTracker();
      ictObs(icTracker);
    }
  }
  else
  {
    // tracker object is not an observable. 
    throw new Error('Binding attribute ' + oj.ValueBinding._ATTRIBUTE_INVALID_COMPONENT_TRACKER +
            ' should be bound to a ko observable.');
  }

  if (icTracker !== null)
  {
    // update icTracker initial state using component's latest option values
    if (ko.isWriteableObservable(ictObs))
    {
      // EditableValue's messagesShown option is an Array of messages currently shown on component.
      messagesShown = component.call(component, "option", oj.ValueBinding._OPTION_MESSAGES_SHOWN);
      // EditableValue's messagesHidden option is an Array of messages currently hidden on component
      messagesHidden = component.call(component, "option", oj.ValueBinding._OPTION_MESSAGES_HIDDEN);

      icTracker._update.call(icTracker, component, element,
              oj.ValueBinding._OPTION_MESSAGES_SHOWN, messagesShown);
      icTracker._update.call(icTracker, component, element,
              oj.ValueBinding._OPTION_MESSAGES_HIDDEN, messagesHidden);
      ictObs.valueHasMutated();
    }

    // register listener for optionChange event for future changes to messages* options
    eventData = {tracker: ictObs, component: component, element: element};
    jElem.on(oj.ValueBinding._EVENT_OPTIONCHANGE, eventData,
            oj.ValueBinding._updateInvalidComponentTracker);
  }
};

/**
 * editableValue Behavior Definition and Injection
 */
oj.ComponentBinding.getDefaultInstance().setupManagedAttributes(
{
  'for': 'editableValue',
  'attributes': [oj.ValueBinding._ATTRIBUTE_INVALID_COMPONENT_TRACKER, 
                 oj.ValueBinding._OPTION_DISABLED,
                  oj.ValueBinding._OPTION_READONLY],
  'init': oj.ValueBinding._init,
  'update': oj.ValueBinding._update,
  'afterCreate': oj.ValueBinding._afterCreate,
  'beforeDestroy': oj.ValueBinding._beforeDestroy
});

var EventTargetMixin = /** @class */ (function () {
    function EventTargetMixin() {
    }
    EventTargetMixin.prototype.addEventListener = function (eventType, listener) {
        if (!this._eventListeners) {
            this._eventListeners = [];
        }
        this._eventListeners.push({
            'type': eventType.toLowerCase(),
            'listener': listener
        });
    };
    ;
    EventTargetMixin.prototype.removeEventListener = function (eventType, listener) {
        if (this._eventListeners) {
            var i = void 0;
            for (i = this._eventListeners.length - 1; i >= 0; i--) {
                if (this._eventListeners[i]['type'] == eventType &&
                    this._eventListeners[i]['listener'] == listener) {
                    this._eventListeners.splice(i, 1);
                }
            }
        }
    };
    ;
    EventTargetMixin.prototype.dispatchEvent = function (evt) {
        if (this._eventListeners) {
            var i, returnValue;
            //clone the eventListeners to isolate mutations that may occur during dispatching events
            var eventListeners = this._eventListeners.slice(0);
            for (i = 0; i < eventListeners.length; i++) {
                var eventListener = eventListeners[i];
                if (eventListener['type'] == evt.type) {
                    returnValue = eventListener['listener'].apply(this, [evt]);
                    if (returnValue === false) {
                        // event cancelled
                        return false;
                    }
                }
            }
        }
        return true;
    };
    ;
    EventTargetMixin.applyMixin = function (derivedCtor) {
        var baseCtors = [EventTargetMixin];
        baseCtors.forEach(function (baseCtor) {
            Object.getOwnPropertyNames(baseCtor.prototype).forEach(function (name) {
                if (name !== 'constructor') {
                    derivedCtor.prototype[name] = baseCtor.prototype[name];
                }
            });
        });
    };
    return EventTargetMixin;
}());
oj.EventTargetMixin = EventTargetMixin;

var GenericEvent = /** @class */ (function () {
    function GenericEvent(type, options) {
        this.type = type;
        this.options = options;
        if (options != null) {
            this['detail'] = options['detail'];
        }
    }
    return GenericEvent;
}());
oj.GenericEvent = GenericEvent;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.FetchByKeysParameters
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.FetchByKeysParameters
 * @ojsignature {target: "Type",
 *               value: "interface FetchByKeysParameters<K>"}
 */

/**
 * Keys for rows to fetch
 *
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchByKeysParameters
 * @instance
 * @name keys
 * @type {Set.<*>}
 * @ojsignature {target: "Type",
 *               value: "Set<K>"}
 */

/**
 * The interface for oj.FetchByKeysResults
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.FetchByKeysResults
 * @ojsignature {target: "Type",
 *               value: "interface FetchByKeysResults<K, D>"}
 */

/**
 * The parameters used for the fetch call.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchByKeysResults
 * @instance
 * @name fetchParameters
 * @type {oj.FetchByKeysParameters}
 * @ojsignature {target: "Type",
 *               value: "FetchByKeysParameters<K>"}
 */

/**
 * Map of keys and corresponding {@link oj.Item}.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchByKeysResults
 * @instance
 * @name results
 * @type {Map.<*, oj.Item>}
 * @ojsignature {target: "Type",
 *               value: "Map<K, Item<K, D>>"}
 */

/**
 * The interface for oj.ContainsKeysResults
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.ContainsKeysResults
 * @ojsignature {target: "Type",
 *               value: "interface ContainsKeysResults<K>"}
 */

/**
 * The parameters used for the containsKeys call.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.ContainsKeysResults
 * @instance
 * @name containsParameters
 * @type {oj.FetchByKeysParameters}
 * @ojsignature {target: "Type",
 *               value: "FetchByKeysParameters<K>"}
 */

/**
 * Requested keys subset which are contained in the DataProvider.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.ContainsKeysResults
 * @instance
 * @name results
 * @type {Set.<*>}
 * @ojsignature {target: "Type",
 *               value: "Set<K>"}
 */

 /**
  * end of jsdoc
  */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.SortCriterion
 * @ojsignature {target: "Type",
 *               value: "interface SortCriterion<D>"}
 */

/**
 * sort attribute upon which the data should be sorted.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.SortCriterion
 * @instance
 * @name attribute
 * @type {string}
 * @ojsignature {target: "Type",
 *               value: "keyof D"}
 */

/**
 * sort direction, either "ascending" or "descending".
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.SortCriterion
 * @instance
 * @name direction
 * @type {string}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.DataMapping
 * @ojsignature {target: "Type",
 *               value: "interface DataMapping <K, D, Kin, Din>"}
 */

/**
 * Field mapping function which takes an item and returns the mapped item
 * 
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.DataMapping
 * @instance
 * @name mapFields
 * @type {Function}
 * @ojsignature {target: "Type",
 *               value: "(item: Item<Kin, Din>) => Item<K, D>"}
 */

/**
 * Optional FilterCriterion mapping function which takes filterOperator and returns the mapped filterOperator
 * 
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.DataMapping
 * @instance
 * @name mapFilterCriterion
 * @type {Function}
 * @ojsignature {target: "Type",
 *               value: "?(filterCriterion: FilterOperator<D>[]) => FilterOperator<Din>[]"}
 */

/**
 * Optional SortCriterion mapping function which takes sortCriteria and returns the mapped sortCriteria
 * 
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.DataMapping
 * @instance
 * @name mapSortCriteria
 * @type {Function}
 * @ojsignature {target: "Type",
 *               value: "?(sortCriteria: SortCriterion<D>[]) => SortCriterion<Din>[]"}
 */

/**
 * Optional SortCriterion unmapping function which takes mapped sortCriteria and returns the unmapped sortCriteria
 * 
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.DataMapping
 * @instance
 * @name unmapSortCriteria
 * @type {Function}
 * @ojsignature {target: "Type",
 *               value: "?(sortCriteria: SortCriterion<Din>[]) => SortCriterion<D>[]"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.FetchFunc
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @ojsignature {target: "Type",
 *               value: "interface FetchFunc<K, D>"}
 */

/**
 * @ojsignature {target: "Type",
 *               value: "(FetchListParameters?): AsyncIterable<FetchListResult<K, D>>"}
 */

 /**
  * end of jsdoc
  */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.FetchListParameters
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.FetchListParameters
 * @ojsignature {target: "Type",
 *               value: "interface FetchListParameters<D>"}
 */

/**
 * Optional number of rows to fetch starting from offset.  If fewer than that number of rows exist, the fetch will succeed but be truncated.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchListParameters
 * @instance
 * @name size
 * @type {number}
 */

/**
 * Optional sort criteria to apply.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchListParameters
 * @instance
 * @name sortCriteria
 * @type {Array.<oj.SortCriterion>}
 * @ojsignature {target: "Type",
 *               value: "?SortCriterion<D>[]"}
 */

/**
 * Optional filter criterion to apply. The filter criterion would be composed of a
 * supported FilterOperator such as a AttributeFilterOperator or a CompoundFilterOperator.
 * 
 * @example
 * The following would be a valid filterCriterion:
 * {op: '$or', criteria: [{op: '$eq', attribute: 'name', value: 'Bob'}, {op: '$gt', attribute: 'level', value: 'Low'}]}
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchListParameters
 * @instance
 * @name filterCriterion
 * @type {oj.FilterOperator}
 * @ojsignature {target: "Type",
 *               value: "?FilterOperator<D>"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.FetchListResult
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.FetchListResult
 * @ojsignature {target: "Type",
 *               value: "interface FetchListResult<K, D>"}
 */

/**
 * The {@link oj.FetchListParameters} used for the fetch call
 *
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchListResult
 * @instance
 * @name fetchParameters
 * @type {oj.FetchListParameters}
 * @ojsignature {target: "Type",
 *               value: "FetchListParameters<D>"}
 */

/**
 * Array of data for each row
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchListResult
 * @instance
 * @name data
 * @type {Array.<Object>}
 * @ojsignature {target: "Type",
 *               value: "D[]"}
 */

/**
 * Array of {link@ oj.ItemMetadata} for each row
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchListResult
 * @instance
 * @name metadata
 * @type {Array.<oj.ItemMetadata>}
 * @ojsignature {target: "Type",
 *               value: "ItemMetadata<K>[]"}
 */
 
 /**
  * end of jsdoc
  */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.ItemMetadata
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.ItemMetadata
 * @ojsignature {target: "Type",
 *               value: "interface ItemMetadata<K>"}
 */

/**
 * The key for the row
 *
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.ItemMetadata
 * @instance
 * @name key
 * @type {any}
 * @ojsignature {target: "Type",
 *               value: "K"}
 */

/**
 * End of jsdoc
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
var AttributeFilterOperator;
(function (AttributeFilterOperator) {
    var AttributeOperator;
    (function (AttributeOperator) {
        AttributeOperator["$co"] = "$co";
        AttributeOperator["$eq"] = "$eq";
        AttributeOperator["$ew"] = "$ew";
        AttributeOperator["$pr"] = "$pr";
        AttributeOperator["$gt"] = "$gt";
        AttributeOperator["$ge"] = "$ge";
        AttributeOperator["$lt"] = "$lt";
        AttributeOperator["$le"] = "$le";
        AttributeOperator["$ne"] = "$ne";
        AttributeOperator["$regex"] = "$regex";
        AttributeOperator["$sw"] = "$sw";
    })(AttributeOperator = AttributeFilterOperator.AttributeOperator || (AttributeFilterOperator.AttributeOperator = {}));
    ;
})(AttributeFilterOperator || (AttributeFilterOperator = {}));
oj['AttributeFilterOperator'] = AttributeFilterOperator;
oj['AttributeFilterOperator']['AttributeOperator'] = AttributeFilterOperator.AttributeOperator;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The interface for oj.AttributeFilterOperator
 *
 * @ojstatus preview
 * @export
 * @interface oj.AttributeFilterOperator
 * @extends oj.FilterOperator
 * @since 5.0.0
 * @ojsignature {target: "Type",
 *               value: "interface AttributeFilterOperator<D>"}
 */

/**
 * Operator one of enum oj.AttributeOperator {$co, $eq, $ew, $pr, $gt, $ge, $lt, $le, $ne, $regex, $sw}
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.AttributeFilterOperator
 * @instance
 * @name op
 * @type {oj.AttributeFilterOperator.AttributeOperator}
 */

/**
 * Attribute
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.AttributeFilterOperator
 * @instance
 * @name attribute
 * @type {string}
 */

/**
 * Value
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.AttributeFilterOperator
 * @instance
 * @name value
 * @type {any}
 */

/**
 * oj.AttributeOperator enum
 * <br>
 * <p>The operators are based on the filtering spec of the RFC 7644 SCIM
 * protocol:
 * <br>
 * {@link https://tools.ietf.org/html/rfc7644#section-3.4.2|SCIM Filtering}
 * <br>
 * <p>
 * @ojstatus preview
 * @export
 * @enum {string}
 * @name AttributeOperator
 * @memberof oj.AttributeFilterOperator
 * @property {string} $co The entire operator value must be a substring of the attribute value for a match.
 * @property {string} $eq The attribute and operator values must be identical for a match.
 * @property {string} $ew The entire operator value must be a substring of the attribute value matching at the end of the attribute value.  This criterion is satisfied if the two strings are identical.     
 * @property {string} $pr If the attribute has a non-empty or non-null value, or if it contains a non-empty node for complex attributes, there is a match.
 * @property {string} $gt If the attribute value is greater than the operator value, there is a match.
 * @property {string} $ge If the attribute value is greater than or equal to the operator value, there is a match. 
 * @property {string} $lt If the attribute value is less than the operator value, there is a match.
 * @property {string} $le If the attribute value is less than or equal to the operator value, there is a match.
 * @property {string} $ne The attribute and operator values are not identical.  
 * @property {string} $regex If the attribute value satisfies the regular expression, there is a match.
 * @property {string} $sw The entire operator value must be a substring of the attribute value, starting at the beginning of the attribute value.  This criterion is satisfied if the two strings are identical.   
 * @since 5.0.0
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
var CompoundFilterOperator;
(function (CompoundFilterOperator) {
    var CompoundOperator;
    (function (CompoundOperator) {
        CompoundOperator["$and"] = "$and";
        CompoundOperator["$or"] = "$or";
    })(CompoundOperator = CompoundFilterOperator.CompoundOperator || (CompoundFilterOperator.CompoundOperator = {}));
    ;
})(CompoundFilterOperator || (CompoundFilterOperator = {}));
oj['CompoundFilterOperator'] = CompoundFilterOperator;
oj['CompoundFilterOperator']['CompoundOperator'] = CompoundFilterOperator.CompoundOperator;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The interface for oj.CompoundFilterOperator
 *
 * @ojstatus preview
 * @export
 * @interface oj.CompoundFilterOperator
 * @extends oj.FilterOperator
 * @since 5.0.0
 * @ojsignature {target: "Type",
 *               value: "interface CompoundFilterOperator<D>"}
 */

/**
 * Operator one of enum oj.CompoundOperator {$and, $or}
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.CompoundFilterOperator
 * @instance
 * @name op
 * @type {oj.CompoundFilterOperator.CompoundOperator}
 */

/**
 * Array of FilterOperators on which to apply the operator
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.CompoundFilterOperator
 * @instance
 * @name criteria
 * @type {Array.<oj.FilterOperator>}
 * @ojsignature {target: "Type",
 *               value: "FilterOperator<D>[]"}
 */

/**
 * oj.CompoundOperator enum
 * <br>
 * <p>The operators are based on the filtering spec of the RFC 7644 SCIM
 * protocol:
 * <br>
 * {@link https://tools.ietf.org/html/rfc7644#section-3.4.2|SCIM Filtering}
 * <br>
 * <p>
 * @ojstatus preview
 * @export
 * @enum {string}
 * @name CompoundOperator
 * @memberof oj.CompoundFilterOperator
 * @property {string} $and The filter is only a match if both expressions evaluate to true.
 * @property {string} $or The filter is a match if either expression evaluates to true.
 * @since 5.0.0
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @interface oj.DataProviderAddOperationEventDetail
 * @extends oj.DataProviderOperationEventDetail
 * @ojsignature {target: "Type",
 *               value: "interface DataProviderAddOperationEventDetail<K, D> extends DataProviderOperationEventDetail<K, D>"}
 */

/**
 * Optional keys of items located after the items involved in the operation. If null and
 * index not specified then insert at the end.
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderAddOperationEventDetail
 * @instance
 * @name afterKeys
 * @type {Set.<*>}
 * @ojsignature {target: "Type",
 *               value: "?Set<K>"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
var DataProviderMutationEvent = /** @class */ (function (_super) {
    __extends(DataProviderMutationEvent, _super);
    function DataProviderMutationEvent(detail) {
        var _this = this;
        var eventOptions = {};
        eventOptions[DataProviderMutationEvent._DETAIL] = detail;
        _this = _super.call(this, 'mutate', eventOptions) || this;
        return _this;
    }
    DataProviderMutationEvent._DETAIL = 'detail';
    return DataProviderMutationEvent;
}(GenericEvent));
oj.DataProviderMutationEvent = DataProviderMutationEvent;
oj['DataProviderMutationEvent'] = DataProviderMutationEvent;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * This interface specifies as one atomic event all the mutation operations which
 * occurred. The keys for each operation must be disjoint from each other, e.g. for example
 * an add and remove cannot occur on the same item. In addition, any indexes specified must
 * be monotonically increasing.
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @interface oj.DataProviderMutationEventDetail
 * @ojsignature {target: "Type",
 *               value: "interface DataProviderMutationEventDetail<K, D>"}
 */

/**
 * Optional add operation detail
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEventDetail
 * @instance
 * @name add
 * @type {oj.DataProviderAddOperationEventDetail}
 * @ojsignature {target: "Type",
 *               value: "?DataProviderAddOperationEventDetail<K, D>"}
 */

/**
 * Optional remove operation detail
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEventDetail
 * @instance
 * @name remove
 * @type {oj.DataProviderOperationEventDetail}
 * @ojsignature {target: "Type",
 *               value: "?DataProviderOperationEventDetail<K, D>"}
 */

/**
 * Optional update operation detail
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEventDetail
 * @instance
 * @name update
 * @type {oj.DataProviderOperationEventDetail}
 * @ojsignature {target: "Type",
 *               value: "?DataProviderOperationEventDetail<K, D>"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @class oj.DataProviderMutationEvent
 * @implements Event
 * @classdesc Mutation event dispatched by oj.DataProvider
 * @param {oj.DataProviderMutationEventDetail} detail the event detail
 * @ojsignature [{target: "Type",
 *               value: "class DataProviderMutationEvent<K, D> implements Event"},
 *               {target: "Type",
 *               value: "DataProviderMutationEventDetail<K, D>", 
 *               for: "detail"}]
 */

/**
 * Event detail
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name detail
 * @type {oj.DataProviderMutationEventDetail}
 * @ojsignature {target: "Type",
 *               value: "DataProviderMutationEventDetail<K, D>"}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name type
 * @type {string}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name bubbles
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name cancelable
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name cancelBubble
 * @type {boolean}
 */  

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name currentTarget
 * @type {EventTarget}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name defaultPrevented
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name eventPhase
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name isTrusted
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name returnValue
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name srcElement
 * @type {Element | null}
 */


/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name target
 * @type {EventTarget}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name timeStamp
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name scoped
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name initEvent
 * @ojsignature {target: "Type",
 *               value: "(eventTypeArg: string, canBubbleArg: boolean, cancelableArg: boolean) => void"}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name preventDefault
 * @ojsignature {target: "Type",
 *               value: "() => void"}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name stopImmediatePropagation
 * @ojsignature {target: "Type",
 *               value: "() => void"}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name stopPropagation
 * @ojsignature {target: "Type",
 *               value: "() => void"}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name deepPath
 * @ojsignature {target: "Type",
 *               value: "() => EventTarget[]"}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name AT_TARGET
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name BUBBLING_PHASE
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderMutationEvent
 * @instance
 * @name CAPTURING_PHASE
 * @type {number}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @interface oj.DataProviderOperationEventDetail
 * @ojsignature {target: "Type",
 *               value: "interface DataProviderOperationEventDetail<K, D>"}
 */

/**
 * keys of items involved in the operation
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderOperationEventDetail
 * @instance
 * @name keys
 * @type {Set.<*>}
 * @ojsignature {target: "Type",
 *               value: "Set<K>"}
 */

/**
 * Optional metadata of items involved in the operation
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderOperationEventDetail
 * @instance
 * @name metadata
 * @type {Array.<ItemMetadata>}
 * @ojsignature {target: "Type",
 *               value: "?ItemMetadata<K>[]"}
 */

/**
 * Optional data of items involved in the operation
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderOperationEventDetail
 * @instance
 * @name data
 * @type {Array.<Object>}
 * @ojsignature {target: "Type",
 *               value: "?D[]"}
 */

/**
 * Optional indexes of items involved in the operation
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.DataProviderOperationEventDetail
 * @instance
 * @name indexes
 * @type {Array.<number>}
 * @ojsignature {target: "Type",
 *               value: "?number[]"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
var DataProviderRefreshEvent = /** @class */ (function (_super) {
    __extends(DataProviderRefreshEvent, _super);
    function DataProviderRefreshEvent() {
        return _super.call(this, 'refresh') || this;
    }
    return DataProviderRefreshEvent;
}(GenericEvent));
oj.DataProviderRefreshEvent = DataProviderRefreshEvent;
oj['DataProviderRefreshEvent'] = DataProviderRefreshEvent;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The oj.DataProviderRefreshEvent class
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @class oj.DataProviderRefreshEvent
 * @implements Event
 * @classdesc Refresh Event dispatched by the DataProvider. This event is fired when
 * the data has been refreshed and components need to re-fetch the data.
 * @ojsignature {target: "Type",
 *               value: "class DataProviderRefreshEvent implements Event"}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name type
 * @type {string}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name bubbles
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name cancelable
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name cancelBubble
 * @type {boolean}
 */  

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name currentTarget
 * @type {EventTarget}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name defaultPrevented
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name eventPhase
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name isTrusted
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name returnValue
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name srcElement
 * @type {Element | null}
 */


/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name target
 * @type {EventTarget}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name timeStamp
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name scoped
 * @type {boolean}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name initEvent
 * @ojsignature {target: "Type",
 *               value: "(eventTypeArg: string, canBubbleArg: boolean, cancelableArg: boolean) => void"}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name preventDefault
 * @ojsignature {target: "Type",
 *               value: "() => void"}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name stopImmediatePropagation
 * @ojsignature {target: "Type",
 *               value: "() => void"}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name stopPropagation
 * @ojsignature {target: "Type",
 *               value: "() => void"}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name deepPath
 * @ojsignature {target: "Type",
 *               value: "() => EventTarget[]"}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name AT_TARGET
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name BUBBLING_PHASE
 * @type {number}
 */

/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.DataProviderRefreshEvent
 * @instance
 * @name CAPTURING_PHASE
 * @type {number}
 */

/**
 * End of jsdoc
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @interface oj.DataProvider
 * @extends EventTarget
 * @ojsignature {target: "Type",
 *               value: "interface DataProvider<K, D> extends EventTarget"}
 * @classdesc
 * DataProvider is the basic interface for getting runtime data which JET components that display list of items (such as [oj-table]{@link oj.ojTable} and [oj-list-view]{@link oj.ojListView}) can use.
 * <p>
 * JET provides some implementations of this interface, such as {@link oj.ArrayDataProvider}.
 * </p><p>
 * Applications can also create their own implementations of this interface and use them with JET components that support it.  For example, an application can create a DataProvider implementation
 * that fetches data from a REST endpoint.
 * </p><p>
 * Implementation class must implement all of the interface methods.  It should also fire the events described below when appropriate, so that JET components or other consumers can respond to data change accordingly. 
 * </p>
 *
 * <h3 id="events-section">
 *   Events
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#events-section"></a>
 * </h3>
 * Implementation can fire the following events by creating an instance of the event class and passing the event payload in the constructor.
 * <h4 id="event:DataProviderMutationEvent" class="name">
 *   {@link oj.DataProviderMutationEvent}
 * </h4>
 * This event is fired when items have been added or removed from the data.
 * <p>
 * Event payload should implement the {@link oj.DataProviderMutationEventDetail} interface.
 * </p><p>
 * Consumers can add an event listener for the "mutate" event type on the DataProvider object.
 * </p>
 * <h4 id="event:DataProviderRefreshEvent" class="name">
 *   {@link oj.DataProviderRefreshEvent}
 * </h4>
 * This event is fired when the data has been refreshed and components need to re-fetch the data.
 * <p>
 * This event contains no additional event payload.
 * </p><p>
 * Consumers can add an event listener for the "refresh" event type on the DataProvider object.
 * </p>
 *
 * <i>Example of implementation firing an oj.DataProviderMutationEvent for removed items:</i>
 * <pre class="prettyprint"><code>var removeDetail = {data: removedDataArray,
 *                     indexes: removedIndexArray,
 *                     keys: removedKeySet,
 *                     metadata: removedMetadataArray};
 * this.dispatchEvent(new oj.DataProviderMutationEvent({remove: removeDetail}));
 * </code></pre>
 *
 * <i>Example of consumer listening for the "mutate" event type:</i>
 * <pre class="prettyprint"><code>var listener = function(event) {
 *   if (event.detail.remove) {
 *     var removeDetail = event.detail.remove;
 *     // Handle removed items
 *   }
 * };
 * dataProvider.addEventListener("mutate", listener);
 * </code></pre>
 */
oj.DataProvider = function()
{
};

/**
 * Fetch the first block of data.
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @param {oj.FetchListParameters=} params fetch parameters
 * @return {AsyncIterable.<oj.FetchListResult>} AsyncIterable with {@link oj.FetchListResult}
 * @see {@link https://github.com/tc39/proposal-async-iteration} for further information on AsyncIterable.
 * @export
 * @expose
 * @memberof oj.DataProvider
 * @instance
 * @method
 * @name fetchFirst
 * @ojsignature {target: "Type",
 *               value: "(FetchListParameters?): AsyncIterable<FetchListResult<K, D>>"}
 */

/**
 * Determines whether this DataProvider supports a certain feature.
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @param {string} capabilityName capability name. Supported capability names are:
 *                  "fetchByKeys", "fetchByOffset", "sort", and "filter".
 * @return {Object} capability information or null if unsupported
 * <ul>
 *   <li>If capabilityName is "fetchByKeys", returns a {@link oj.FetchByKeysCapability} object.</li>
 *   <li>If capabilityName is "fetchByOffset", returns a {@link oj.FetchByOffsetCapability} object.</li>
 *   <li>If capabilityName is "sort", returns a {@link oj.SortCapability} object.</li>
 *   <li>If capabilityName is "filter", returns a {@link oj.FilterCapability} object.</li>
 * </ul>
 * @export
 * @expose
 * @memberof oj.DataProvider
 * @instance
 * @method
 * @name getCapability
 * @ojsignature {target: "Type",
 *               value: "(capabilityName: string): any"}
 */

/**
 * Return the total number of rows in this dataprovider
 * 
 * @ojstatus preview
 * @return {Promise.<number>} Returns a Promise which resolves to the total number of rows. -1 is unknown row count.
 * @export
 * @expose
 * @memberof oj.DataProvider
 * @instance
 * @method
 * @name getTotalSize
 */

/**
 * Fetch rows by keys
 * <p>
 * A generic implementation of this method is available from {@link oj.FetchByKeysMixin}.
 * It is for convenience and may not provide the most efficient implementation for your data provider.
 * Classes that implement the DataProvider interface are encouraged to provide a more efficient implementation.
 * </p>
 *
 * @ojstatus preview
 * @since 4.2.0
 * @param {oj.FetchByKeysParameters} parameters fetch by key parameters
 * @return {Promise.<oj.FetchByKeysResults>} Returns Promise which resolves to {@link oj.FetchByKeysResults}.
 * @export
 * @expose
 * @memberof oj.DataProvider
 * @instance
 * @method
 * @name fetchByKeys
 * @ojsignature {target: "Type",
 *               value: "(parameters : FetchByKeysParameters<K>) : Promise<FetchByKeysResults<K, D>>"}
 */

/**
 * Check if there are rows containing the specified keys
 * <p>
 * A generic implementation of this method is available from {@link oj.FetchByKeysMixin}.
 * It is for convenience and may not provide the most efficient implementation for your data provider.
 * Classes that implement the DataProvider interface are encouraged to provide a more efficient implementation.
 * </p>
 *
 * @ojstatus preview
 * @since 4.2.0
 * @param {oj.FetchByKeysParameters} parameters contains by key parameters
 * @return {Promise.<oj.ContainsKeysResults>} Returns Promise which resolves to {@link oj.ContainsKeysResults}.
 * @export
 * @expose
 * @memberof oj.DataProvider
 * @instance
 * @method
 * @name containsKeys
 * @ojsignature {target: "Type",
 *               value: "(parameters : FetchByKeysParameters<K>) : Promise<ContainsKeysResults<K>>"}
 */

/**
 * Fetch rows by offset
 * <p>
 * A generic implementation of this method is available from {@link oj.FetchByOffsetMixin}.
 * It is for convenience and may not provide the most efficient implementation for your data provider.
 * Classes that implement the DataProvider interface are encouraged to provide a more efficient implementation.
 * </p>
 *
 * @ojstatus preview
 * @since 4.2.0
 * @param {oj.FetchByOffsetParameters} parameters fetch by offset parameters
 * @return {Promise.<oj.FetchByOffsetResults>} Returns Promise which resolves to {@link oj.FetchByOffsetResults}.
 * @export
 * @expose
 * @memberof oj.DataProvider
 * @instance
 * @method
 * @name fetchByOffset
 * @ojsignature {target: "Type",
 *               value: "(parameters: FetchByOffsetParameters<D>): Promise<FetchByOffsetResults<K, D>>"}
 */

/**
 * Returns a string that indicates if this data provider is empty.  Valid values are:
 * <ul>
 * <li>"yes": this data provider is empty.</li>
 * <li>"no": this data provider is not empty.</li>
 * <li>"unknown": it is not known if this data provider is empty until a fetch is made.</li>
 * </ul>
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @return {"yes" | "no" | "unknown"} string that indicates if this data provider is empty
 * @export
 * @expose
 * @memberof oj.DataProvider
 * @instance
 * @method
 * @name isEmpty
 * @ojsignature {target: "Type",
 *               value: "(): 'yes' | 'no' | 'unknown'"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The interface for oj.FetchByKeysCapability
 *
 * @ojstatus preview
 * @export
 * @interface oj.FetchByKeysCapability
 * @since 4.2.0
 * @ojsignature {target: "Type",
 *               value: "interface FetchByKeysCapability<D>"}
 */

/**
 * The type of implementation for fetchByKeys and containsKeys methods.  Possible values are:
 * <ul>
 * <li>"iteration": the implementation uses fetchFirst iteratively to find the result</li>
 * <li>"lookup": the implementation uses direct lookup to find the result</li>
 * </ul>
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.FetchByKeysCapability
 * @instance
 * @name implementation
 * @type {"iteration" | "lookup"}
 */

/**
 * End of jsdoc
 */
var FetchByKeysMixin = /** @class */ (function () {
    function FetchByKeysMixin() {
    }
    /**
     * Fetch rows by keys
     */
    FetchByKeysMixin.prototype.fetchByKeys = function (params) {
        var fetched = 0;
        var limit = this['getIterationLimit'] ? this['getIterationLimit']() : -1;
        var options = {};
        options['size'] = 25;
        var resultMap = new Map();
        var dataProviderAsyncIterator = this['fetchFirst'](options)[Symbol.asyncIterator]();
        function _fetchNextSet(params, dataProviderAsyncIterator, resultMap) {
            return dataProviderAsyncIterator.next().then(function (result) {
                var value = result['value'];
                var data = value['data'];
                var metadata = value['metadata'];
                var keys = metadata.map(function (metadata) {
                    return metadata['key'];
                });
                var foundAllKeys = true;
                params['keys'].forEach(function (findKey) {
                    if (!resultMap.has(findKey)) {
                        keys.map(function (key, index) {
                            if (key == findKey) {
                                resultMap.set(key, { 'metadata': metadata[index], 'data': data[index] });
                            }
                        });
                    }
                    if (!resultMap.has(findKey)) {
                        foundAllKeys = false;
                    }
                });
                // Keep track of how many rows we have fetched
                fetched += data.length;
                // Keep iterating if we haven't found all keys and there are more data
                if (!foundAllKeys && !result['done']) {
                    if (limit != -1 && fetched >= limit) {
                        // If we have reached the limit, just return the results
                        return resultMap;
                    }
                    else {
                        return _fetchNextSet(params, dataProviderAsyncIterator, resultMap);
                    }
                }
                else {
                    return resultMap;
                }
            });
        }
        return _fetchNextSet(params, dataProviderAsyncIterator, resultMap).then(function (resultMap) {
            var mappedResultMap = new Map();
            resultMap.forEach(function (value, key) {
                var mappedItem = [value];
                mappedResultMap.set(key, mappedItem[0]);
            });
            return { 'fetchParameters': params, 'results': mappedResultMap };
        });
    };
    /**
     * Check if rows are contained by keys
     */
    FetchByKeysMixin.prototype.containsKeys = function (params) {
        return this.fetchByKeys(params).then(function (fetchByKeysResult) {
            var results = new Set();
            params['keys'].forEach(function (key) {
                if (fetchByKeysResult['results'].get(key) != null) {
                    results.add(key);
                }
            });
            return Promise.resolve({ 'containsParameters': params,
                'results': results });
        });
    };
    FetchByKeysMixin.prototype.getCapability = function (capabilityName) {
        if (capabilityName == 'fetchByKeys') {
            return { implementation: 'iteration' };
        }
        var cap = null;
        if (this['_ojSkipLastCapability'] !== true) {
            this['_ojSkipLastCapability'] = true;
            // Find the index for the very last _ojLastGetCapability
            var index = 1;
            while (this['_ojLastGetCapability' + index]) {
                ++index;
            }
            // Iterate through the _ojLastGetCapability(n) in reverse order
            for (--index; index > 0; index--) {
                cap = this['_ojLastGetCapability' + index](capabilityName);
                if (cap) {
                    break;
                }
            }
            delete this['_ojSkipLastCapability'];
        }
        return cap;
    };
    FetchByKeysMixin.applyMixin = function (derivedCtor) {
        // Save the current getCapability
        var _lastGetCapability = derivedCtor.prototype['getCapability'];
        var baseCtors = [FetchByKeysMixin];
        baseCtors.forEach(function (baseCtor) {
            Object.getOwnPropertyNames(baseCtor.prototype).forEach(function (name) {
                if (name !== 'constructor') {
                    derivedCtor.prototype[name] = baseCtor.prototype[name];
                }
            });
        });
        if (_lastGetCapability) {
            var index = 1;
            while (derivedCtor.prototype['_ojLastGetCapability' + index]) {
                ++index;
            }
            derivedCtor.prototype['_ojLastGetCapability' + index] = _lastGetCapability;
        }
    };
    return FetchByKeysMixin;
}());
oj['FetchByKeysMixin'] = FetchByKeysMixin;
oj['FetchByKeysMixin']['applyMixin'] = FetchByKeysMixin.applyMixin;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @export
 * @class oj.FetchByKeysMixin
 * @classdesc Mixin class to provide generic implementation of fetchByKeys and containsKeys
 * methods for the {@link oj.DataProvider} interface.
 * <p>
 * By default, the mixin implementation will iterate through all the rows to find
 * the result.  DataProvider implementations can implement a "getIterationLimit" function
 * that returns a row limit for the iteration:<br>
 * getIterationLimit() => number
 * </p>
 * <p>
 * This class cannot be instantiated.  You can only call the static applyMixin
 * method to add the implementation to another class.
 * </p>
 * @since 4.2.0
 * @hideconstructor
 */

/**
 * Apply this mixin to another class
 *
 * @ojstatus preview
 * @param {Function} derivedCtor the constructor of an existing class
 * @export
 * @expose
 * @memberof oj.FetchByKeysMixin
 * @method
 * @name applyMixin
 * @example <caption>Apply the mixin in Typescript:</caption>
 * class CustomDataProvider&lt;K, D> implements DataProvider&lt;K, D> {
 *   // Add stand-in properties to satisfy the compiler
 *   containsKeys: (parameters: FetchByKeysParameters&lt;K>) => Promise&lt;ContainsKeysResults&lt;K>>;
 *   fetchByKeys: (parameters: FetchByKeysParameters&lt;K>) => Promise&lt;FetchByKeysResults&lt;K, D>>;
 *
 *   constructor() {
 *     // Constructor implementation
 *   }
 * }
 *
 * oj.FetchByKeysMixin.applyMixin(CustomDataProvider);
 *
 * @example <caption>Apply the mixin in Javascript:</caption>
 * function CustomDataProvider() {
 *   // Constructor implementation
 * }
 *
 * oj.FetchByKeysMixin.applyMixin(CustomDataProvider);
 */

 /**
  * end of jsdoc
  */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The interface for oj.FetchByOffsetCapability
 *
 * @ojstatus preview
 * @export
 * @interface oj.FetchByOffsetCapability
 * @since 4.2.0
 * @ojsignature {target: "Type",
 *               value: "interface FetchByOffsetCapability<D>"}
 */

/**
 * The type of implementation for fetchByOffset method.  Possible values are:
 * <ul>
 * <li>"iteration": the implementation uses fetchFirst iteratively to find the result</li>
 * <li>"randomAccess": the implementation uses random access to find the result</li>
 * </ul>
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.FetchByOffsetCapability
 * @instance
 * @name implementation
 * @type {"iteration" | "randomAccess"}
 */

/**
 * End of jsdoc
 */
var FetchByOffsetMixin = /** @class */ (function () {
    function FetchByOffsetMixin() {
    }
    /**
     * Fetch rows by offset
     */
    FetchByOffsetMixin.prototype.fetchByOffset = function (params) {
        var size = params && params['size'] > 0 ? params['size'] : 25;
        var sortCriteria = params ? params['sortCriteria'] : null;
        var offset = params && params['offset'] > 0 ? params['offset'] : 0;
        var fetched = 0;
        var limit = this['getIterationLimit'] ? this['getIterationLimit']() : -1;
        var done = false;
        var options = {};
        options['size'] = size;
        options['sortCriteria'] = sortCriteria;
        var resultArray = new Array();
        var dataProviderAsyncIterator = this['fetchFirst'](options)[Symbol.asyncIterator]();
        function _fetchNextSet(params, dataProviderAsyncIterator, resultArray) {
            return dataProviderAsyncIterator.next().then(function (result) {
                done = result['done'];
                var value = result['value'];
                var data = value['data'];
                var metadata = value['metadata'];
                var dataLen = data.length;
                if (offset < (fetched + dataLen)) {
                    var start = (offset <= fetched) ? 0 : (offset - fetched);
                    for (var index = start; index < dataLen; index++) {
                        if (resultArray.length == size) {
                            break;
                        }
                        resultArray.push({ 'metadata': metadata[index], 'data': data[index] });
                    }
                }
                fetched += dataLen;
                if (resultArray.length < size && !done) {
                    if (limit != -1 && fetched >= limit) {
                        // If we have reached the limit, just return the results
                        return resultArray;
                    }
                    else {
                        return _fetchNextSet(params, dataProviderAsyncIterator, resultArray);
                    }
                }
                else {
                    return resultArray;
                }
            });
        }
        return _fetchNextSet(params, dataProviderAsyncIterator, resultArray).then(function (resultArray) {
            return { 'fetchParameters': params,
                'results': resultArray,
                'done': done };
        });
    };
    FetchByOffsetMixin.prototype.getCapability = function (capabilityName) {
        if (capabilityName == 'fetchByOffset') {
            return { implementation: 'iteration' };
        }
        var cap = null;
        if (this['_ojSkipLastCapability'] !== true) {
            this['_ojSkipLastCapability'] = true;
            // Find the index for the very last _ojLastGetCapability
            var index = 1;
            while (this['_ojLastGetCapability' + index]) {
                ++index;
            }
            // Iterate through the _ojLastGetCapability(n) in reverse order
            for (--index; index > 0; index--) {
                cap = this['_ojLastGetCapability' + index](capabilityName);
                if (cap) {
                    break;
                }
            }
            delete this['_ojSkipLastCapability'];
        }
        return cap;
    };
    FetchByOffsetMixin.applyMixin = function (derivedCtor) {
        // Save the current getCapability
        var _lastGetCapability = derivedCtor.prototype['getCapability'];
        var baseCtors = [FetchByOffsetMixin];
        baseCtors.forEach(function (baseCtor) {
            Object.getOwnPropertyNames(baseCtor.prototype).forEach(function (name) {
                if (name !== 'constructor') {
                    derivedCtor.prototype[name] = baseCtor.prototype[name];
                }
            });
        });
        if (_lastGetCapability) {
            var index = 1;
            while (derivedCtor.prototype['_ojLastGetCapability' + index]) {
                ++index;
            }
            derivedCtor.prototype['_ojLastGetCapability' + index] = _lastGetCapability;
        }
    };
    return FetchByOffsetMixin;
}());
oj['FetchByOffsetMixin'] = FetchByOffsetMixin;
oj['FetchByOffsetMixin']['applyMixin'] = FetchByOffsetMixin.applyMixin;

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @export
 * @class oj.FetchByOffsetMixin
 * @classdesc Mixin class to provide generic implementation of fetchByOffset
 * method for the {@link oj.DataProvider} interface.
 * <p>
 * By default, the mixin implementation will iterate through all the rows to find
 * the result.  DataProvider implementations can implement a "getIterationLimit" function
 * that returns a row limit for the iteration:<br>
 * getIterationLimit() => number
 * </p>
 * <p>
 * This class cannot be instantiated.  You can only call the static applyMixin
 * method to add the implementation to another class.
 * </p>
 * @since 4.2.0
 * @hideconstructor
 */

/**
 * Apply this mixin to another class
 *
 * @ojstatus preview
 * @param {Function} derivedCtor the constructor of an existing class
 * @export
 * @expose
 * @memberof oj.FetchByOffsetMixin
 * @method
 * @name applyMixin
 * @example <caption>Apply the mixin in Typescript:</caption>
 * class CustomDataProvider&lt;K, D> implements DataProvider&lt;K, D> {
 *   // Add a stand-in property to satisfy the compiler
 *   fetchByOffset: (parameters: FetchByOffsetParameters&lt;D>) => Promise&lt;FetchByOffsetResults&lt;K, D>>;
 *
 *   constructor() {
 *     // Constructor implementation
 *   }
 * }
 *
 * oj.FetchByOffsetMixin.applyMixin(CustomDataProvider);
 *
 * @example <caption>Apply the mixin in Javascript:</caption>
 * function CustomDataProvider() {
 *   // Constructor implementation
 * }
 *
 * oj.FetchByOffsetMixin.applyMixin(CustomDataProvider);
 */

 /**
  * end of jsdoc
  */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.FetchByOffsetParameters
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.FetchByOffsetParameters
 * @extends oj.FetchListParameters
 * @ojsignature {target: "Type",
 *               value: "interface FetchByOffsetParameters<D> extends FetchListParameters<D>"}
 */

/**
 * The offset used for the fetch call.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchByOffsetParameters
 * @instance
 * @name offset
 * @type {number}
 */

/**
 * The interface for oj.FetchByOffsetResults
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.FetchByOffsetResults
 * @ojsignature {target: "Type",
 *               value: "interface FetchByOffsetResults<K, D>"}
 */

/**
 * The parameters used for the fetch call.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchByOffsetResults
 * @instance
 * @name fetchParameters
 * @type {oj.FetchByOffsetParameters}
 * @ojsignature {target: "Type",
 *               value: "FetchByOffsetParameters<D>"}
 */


/**
 * Array of {@link oj.Item}.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchByOffsetResults
 * @instance
 * @name results
 * @type {Array.<oj.Item>}
 * @ojsignature {target: "Type",
 *               value: "Array<Item<K, D>>"}
 */

/**
 * Indicates whether there are more items which can be fetched.
 * <p>If this is true, fetching the next block will likely return an empty array as the result.  A DataProvider can potentially make a stronger guarantee (if the DataProvider is running against an immutable repository or the DataProvider doesn’t attempt to retrieve a subsequent block if the DataProvider believes it is complete).  We don’t generally make the stronger guarantee since the repository may have been mutated since the previous response with done:true, such that new records would be returned.</p>
 * <p>If this is false, fetching the next block may or may not return an empty array as a result.</p>
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.FetchByOffsetResults
 * @instance
 * @name done
 * @type {boolean}
 */

 /**
  * end of jsdoc
  */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The interface for oj.FilterCapability
 *
 * @ojstatus preview
 * @export
 * @interface oj.FilterCapability
 * @since 5.0.0
 * @ojsignature {target: "Type",
 *               value: "interface FilterCapability"}
 */

/**
 * An array of supported filter operator strings
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.FilterCapability
 * @instance
 * @name operators
 * @type {Array.<string>}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The interface for oj.FilterOperator
 *
 * @ojstatus preview
 * @export
 * @interface oj.FilterOperator
 * @since 5.0.0
 * @ojsignature {target: "Type",
 *               value: "interface FilterOperator<D>"}
 */

/**
 * The operator, either an oj.AttributeFilterOperator.AttributeOperator or oj.CompoundFilterOperator.CompoundOperator.
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.FilterOperator
 * @instance
 * @name op
 * @type {oj.AttributeFilterOperator.AttributeOperator | oj.CompoundFilterOperator.CompoundOperator}
 */

/**
 * Optional function which is used to locally filter the data.
 * 
 * @ojstatus preview
 * @since 5.0.0
 * @export
 * @expose
 * @memberof oj.FilterOperator
 * @instance
 * @name filter
 * @method
 * @param {Array} data The data to filter
 * @return {Array} filtered data
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.Item
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.Item
 * @ojsignature {target: "Type",
 *               value: "interface Item<K, D>"}
 */

/**
 * The metadata for the item
 *
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.Item
 * @instance
 * @name metadata
 * @type {oj.ItemMetadata}
 * @ojsignature {target: "Type",
 *               value: "ItemMetadata<K>"}
 */

/**
 * The data for the item
 *
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.Item
 * @instance
 * @name data
 * @type {Object}
 * @ojsignature {target: "Type",
 *               value: "D"}
 */

/**
 * End of jsdoc
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * The interface for oj.SortCapability
 *
 * @ojstatus preview
 * @export
 * @interface oj.SortCapability
 * @since 4.2.0
 * @ojsignature {target: "Type",
 *               value: "interface SortCapability<D>"}
 */

/**
 * Number of attributes that can be sorted at the same time.  Possible values are:
 * <ul>
 * <li>"none": no sorting is supported.</li>
 * <li>"single": only one attribute can be sorted at a time.</li>
 * <li>"multiple": more than one attribute can be sorted at a time.</li>
 * </ul>
 * 
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @expose
 * @memberof oj.SortCapability
 * @instance
 * @name attributes
 * @type {"none" | "single" | "multiple"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
var TableDataSourceAdapter = /** @class */ (function () {
    function TableDataSourceAdapter(tableDataSource) {
        this.tableDataSource = tableDataSource;
        this._KEY = 'key';
        this._KEYS = 'keys';
        this._AFTERKEYS = 'afterKeys';
        this._DIRECTION = 'direction';
        this._STARTINDEX = 'startIndex';
        this._ATTRIBUTE = 'attribute';
        this._SILENT = 'silent';
        this._SORT = 'sort';
        this._SORTCRITERIA = 'sortCriteria';
        this._PAGESIZE = 'pageSize';
        this._DATA = 'data';
        this._METADATA = 'metadata';
        this._INDEXES = 'indexes';
        this._OFFSET = 'offset';
        this._REFRESH = 'refresh';
        this._SIZE = 'size';
        this._VALUE = 'value';
        this._DONE = 'done';
        this._FETCHPARAMETERS = 'fetchParameters';
        this._CONTAINSPARAMETERS = 'containsParameters';
        this._RESULTS = 'results';
        this._ADD = 'add';
        this._REMOVE = 'remove';
        this._UPDATE = 'update';
        this._FETCHTYPE = 'fetchType';
        this.AsyncIterable = /** @class */ (function () {
            function class_1(_parent, _asyncIterator) {
                this._parent = _parent;
                this._asyncIterator = _asyncIterator;
                this[Symbol.asyncIterator] = function () {
                    return this._asyncIterator;
                };
            }
            return class_1;
        }());
        this.AsyncIterator = /** @class */ (function () {
            function class_2(_parent, _nextFunc, _params) {
                this._parent = _parent;
                this._nextFunc = _nextFunc;
                this._params = _params;
                this._fetchFirst = true;
            }
            class_2.prototype['next'] = function () {
                var fetchFirst = this._fetchFirst;
                this._fetchFirst = false;
                return this._nextFunc(this._params, fetchFirst);
            };
            return class_2;
        }());
        this.AsyncIteratorResult = /** @class */ (function () {
            function class_3(_parent, value, done) {
                this._parent = _parent;
                this.value = value;
                this.done = done;
                this[_parent._VALUE] = value;
                this[_parent._DONE] = done;
            }
            return class_3;
        }());
        this.FetchByKeysResults = /** @class */ (function () {
            function class_4(_parent, fetchParameters, results) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.results = results;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._RESULTS] = results;
            }
            return class_4;
        }());
        this.ContainsKeysResults = /** @class */ (function () {
            function class_5(_parent, containsParameters, results) {
                this._parent = _parent;
                this.containsParameters = containsParameters;
                this.results = results;
                this[_parent._CONTAINSPARAMETERS] = containsParameters;
                this[_parent._RESULTS] = results;
            }
            return class_5;
        }());
        this.FetchListResult = /** @class */ (function () {
            function class_6(_parent, fetchParameters, data, metadata) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.data = data;
                this.metadata = metadata;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._DATA] = data;
                this[_parent._METADATA] = metadata;
            }
            return class_6;
        }());
        this.Item = /** @class */ (function () {
            function class_7(_parent, metadata, data) {
                this._parent = _parent;
                this.metadata = metadata;
                this.data = data;
                this[_parent._METADATA] = metadata;
                this[_parent._DATA] = data;
            }
            return class_7;
        }());
        this.ItemMetadata = /** @class */ (function () {
            function class_8(_parent, key) {
                this._parent = _parent;
                this.key = key;
                this[_parent._KEY] = key;
            }
            return class_8;
        }());
        this.FetchByOffsetResults = /** @class */ (function () {
            function class_9(_parent, fetchParameters, results, done) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.results = results;
                this.done = done;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._RESULTS] = results;
                this[_parent._DONE] = done;
            }
            return class_9;
        }());
        this.FetchListParameters = /** @class */ (function () {
            function class_10(_parent, size, sortCriteria) {
                this._parent = _parent;
                this.size = size;
                this.sortCriteria = sortCriteria;
                this[_parent._SIZE] = size;
                this[_parent._SORTCRITERIA] = sortCriteria;
            }
            return class_10;
        }());
        this.SortCriterion = /** @class */ (function () {
            function class_11(_parent, attribute, direction) {
                this._parent = _parent;
                this.attribute = attribute;
                this.direction = direction;
                this[_parent._ATTRIBUTE] = attribute;
                this[_parent._DIRECTION] = direction;
            }
            return class_11;
        }());
        this.DataProviderMutationEventDetail = /** @class */ (function () {
            function class_12(_parent, add, remove, update) {
                this._parent = _parent;
                this.add = add;
                this.remove = remove;
                this.update = update;
                this[_parent._ADD] = add;
                this[_parent._REMOVE] = remove;
                this[_parent._UPDATE] = update;
            }
            return class_12;
        }());
        this.DataProviderOperationEventDetail = /** @class */ (function () {
            function class_13(_parent, keys, metadata, data, indexes) {
                this._parent = _parent;
                this.keys = keys;
                this.metadata = metadata;
                this.data = data;
                this.indexes = indexes;
                this[_parent._KEYS] = keys;
                this[_parent._METADATA] = metadata;
                this[_parent._DATA] = data;
                this[_parent._INDEXES] = indexes;
            }
            return class_13;
        }());
        this.DataProviderAddOperationEventDetail = /** @class */ (function () {
            function class_14(_parent, keys, afterKeys, metadata, data, indexes) {
                this._parent = _parent;
                this.keys = keys;
                this.afterKeys = afterKeys;
                this.metadata = metadata;
                this.data = data;
                this.indexes = indexes;
                this[_parent._KEYS] = keys;
                this[_parent._AFTERKEYS] = afterKeys;
                this[_parent._METADATA] = metadata;
                this[_parent._DATA] = data;
                this[_parent._INDEXES] = indexes;
            }
            return class_14;
        }());
        this._addTableDataSourceEventListeners(tableDataSource);
        this[this._OFFSET] = 0;
    }
    TableDataSourceAdapter.prototype.containsKeys = function (params) {
        var self = this;
        var resultsPromiseArray = [];
        params[this._KEYS].forEach(function (key) {
            resultsPromiseArray.push(self.tableDataSource.get(key));
        });
        return Promise.all(resultsPromiseArray).then(function (resultsArray) {
            var results = new Set();
            resultsArray.map(function (value) {
                results.add(value[self._KEY]);
            });
            return Promise.resolve(new self.ContainsKeysResults(self, params, results));
        });
    };
    TableDataSourceAdapter.prototype.fetchByKeys = function (params) {
        var self = this;
        var resultsPromiseArray = [];
        params[this._KEYS].forEach(function (key) {
            resultsPromiseArray.push(self.tableDataSource.get(key));
        });
        return Promise.all(resultsPromiseArray).then(function (resultsArray) {
            var results = new Map();
            resultsArray.map(function (value) {
                var key = value[self._KEY];
                var data = value[self._DATA];
                results.set(key, new self.Item(self, new self.ItemMetadata(self, key), data));
            });
            return Promise.resolve(new self.FetchByKeysResults(self, params, results));
        });
    };
    TableDataSourceAdapter.prototype.fetchByOffset = function (params) {
        var self = this;
        var size = params != null ? params[this._SIZE] : -1;
        var sortCriteria = params != null ? params[this._SORTCRITERIA] : null;
        var offset = params != null ? params[this._OFFSET] > 0 ? params[this._OFFSET] : 0 : 0;
        var fetchParams = new this.FetchListParameters(this, size, sortCriteria);
        var iteratorResults = this._getFetchFunc(fetchParams, offset);
        var value = iteratorResults[this._VALUE];
        var done = iteratorResults[this._DONE];
        var data = value[this._DATA];
        var keys = value[this._METADATA].map(function (value) {
            return value[self._KEY];
        });
        var resultsArray = data.map(function (value, index) {
            return new self.Item(self, value, keys[index]);
        });
        return Promise.resolve(new this.FetchByOffsetResults(this, params, resultsArray, done));
    };
    TableDataSourceAdapter.prototype.fetchFirst = function (params) {
        if (!this._isPagingModelTableDataSource()) {
            this._startIndex = 0;
        }
        return new this.AsyncIterable(this, new this.AsyncIterator(this, this._getFetchFunc(params), params));
    };
    TableDataSourceAdapter.prototype.getCapability = function (capabilityName) {
        if (capabilityName == this._SORT &&
            this.tableDataSource.getCapability(capabilityName) == 'full') {
            return { attributes: 'multiple' };
        }
        return null;
    };
    TableDataSourceAdapter.prototype.getTotalSize = function () {
        return Promise.resolve(this.tableDataSource.totalSize());
    };
    TableDataSourceAdapter.prototype.isEmpty = function () {
        return this.tableDataSource.totalSize() > 0 ? 'no' : 'yes';
    };
    // Start PagingModel APIs
    TableDataSourceAdapter.prototype.getPage = function () {
        if (this._isPagingModelTableDataSource()) {
            return this.tableDataSource.getPage();
        }
        return -1;
    };
    TableDataSourceAdapter.prototype.setPage = function (value, options) {
        if (this._isPagingModelTableDataSource()) {
            return this.tableDataSource.setPage(value, options);
        }
        return Promise.reject(null);
    };
    TableDataSourceAdapter.prototype.getStartItemIndex = function () {
        if (this._isPagingModelTableDataSource()) {
            return this.tableDataSource.getStartItemIndex();
        }
        return -1;
    };
    TableDataSourceAdapter.prototype.getEndItemIndex = function () {
        if (this._isPagingModelTableDataSource()) {
            return this.tableDataSource.getEndItemIndex();
        }
        return -1;
    };
    TableDataSourceAdapter.prototype.getPageCount = function () {
        if (this._isPagingModelTableDataSource()) {
            return this.tableDataSource.getPageCount();
        }
        return -1;
    };
    TableDataSourceAdapter.prototype.totalSize = function () {
        if (this._isPagingModelTableDataSource()) {
            return this.tableDataSource.totalSize();
        }
        return -1;
    };
    TableDataSourceAdapter.prototype.totalSizeConfidence = function () {
        if (this._isPagingModelTableDataSource()) {
            return this.tableDataSource.totalSizeConfidence();
        }
        return null;
    };
    // End PagingModel APIs
    /**
     * Get the function which performs the fetch
     */
    TableDataSourceAdapter.prototype._getFetchFunc = function (params, offset) {
        var self = this;
        if (params != null &&
            params[this._SORTCRITERIA] != null) {
            var attribute = params[this._SORTCRITERIA][0][this._ATTRIBUTE];
            var direction = params[this._SORTCRITERIA][0][this._DIRECTION];
            this._ignoreSortEvent = true;
            if (!this._isPagingModelTableDataSource()) {
                this._startIndex = 0;
            }
            return function (attribute, direction) {
                return function (params, fetchFirst) {
                    if (fetchFirst) {
                        var sortParam = {};
                        sortParam[self._KEY] = attribute;
                        sortParam[self._DIRECTION] = direction;
                        self[self._OFFSET] = 0;
                        return self.tableDataSource.sort(sortParam).then(function () {
                            self._ignoreSortEvent = false;
                            return self._getTableDataSourceFetch(params, offset)(params);
                        });
                    }
                    else {
                        return self._getTableDataSourceFetch(params, offset)(params);
                    }
                };
            }(attribute, direction);
        }
        else {
            return this._getTableDataSourceFetch(params, offset);
        }
    };
    /**
     * Get the function which invokes fetch() on TableDataSource
     */
    TableDataSourceAdapter.prototype._getTableDataSourceFetch = function (params, offset) {
        var self = this;
        return function (params, fetchFirst) {
            var options = {};
            offset = offset > 0 ? offset : 0;
            if (self._startIndex != null) {
                options[self._STARTINDEX] = self._startIndex + offset;
            }
            options[self._PAGESIZE] = params != null && params[self._SIZE] > 0 ? params[self._SIZE] : null;
            if (!self._isPagingModelTableDataSource()) {
                options[self._SILENT] = true;
            }
            if (self.tableDataSource[self._SORTCRITERIA] != null && params[self._SORTCRITERIA] == null) {
                params[self._SORTCRITERIA] = [];
                var sortCriterion = new self.SortCriterion(self, self.tableDataSource[self._SORTCRITERIA][self._KEY], self.tableDataSource[self._SORTCRITERIA][self._DIRECTION]);
                params[self._SORTCRITERIA].push(sortCriterion);
            }
            options[self._FETCHTYPE] = params[self._FETCHTYPE];
            self._isFetching = true;
            return new Promise(function (resolve, reject) {
                self._fetchResolveFunc = resolve;
                self._fetchRejectFunc = reject;
                self._fetchParams = params;
                if (!self._requestEventTriggered) {
                    self.tableDataSource.fetch(options).then(function (result) {
                        if (result !== null) {
                            self._isFetching = false;
                            if (result === undefined) {
                                // fetch was not executed due to startFetch='disabled'
                                result = {};
                                result[self._KEYS] = [];
                                result[self._DATA] = [];
                            }
                            var resultMetadata = [];
                            if (result[self._KEYS] != null) {
                                resultMetadata = result[self._KEYS].map(function (value) {
                                    return new self.ItemMetadata(self, value);
                                });
                            }
                            if (self._startIndex == null) {
                                self._startIndex = 0;
                            }
                            var done = false;
                            self._startIndex = self._startIndex + result[self._DATA].length;
                            if (self.tableDataSource.totalSizeConfidence() == 'actual' &&
                                self.tableDataSource.totalSize() > 0 &&
                                self._startIndex >= self.tableDataSource.totalSize()) {
                                done = true;
                            }
                            else if (options[self._PAGESIZE] > 0 &&
                                result[self._DATA].length < options[self._PAGESIZE]) {
                                done = true;
                            }
                            else if (result[self._DATA].length == 0) {
                                done = true;
                            }
                            self._fetchResolveFunc = null;
                            self._fetchParams = null;
                            resolve(new self.AsyncIteratorResult(self, new self.FetchListResult(self, params, result[self._DATA], resultMetadata), done));
                        }
                    }, function (error) {
                        reject(error);
                    });
                }
            });
        };
    };
    /**
     * Add event listeners to TableDataSource
     */
    TableDataSourceAdapter.prototype._addTableDataSourceEventListeners = function (tableDataSource) {
        var self = this;
        tableDataSource.on('sync', function (event) {
            self._startIndex = null;
            if (event[self._STARTINDEX] > 0) {
                self._startIndex = event[self._STARTINDEX];
                self[self._OFFSET] = self._startIndex;
            }
            if (self._fetchResolveFunc && event[self._KEYS] != null) {
                self._isFetching = false;
                var resultMetadata = event[self._KEYS].map(function (value) {
                    return new self.ItemMetadata(self, value);
                });
                var done = false;
                if (self.tableDataSource.totalSizeConfidence() == 'actual' &&
                    self.tableDataSource.totalSize() > 0 &&
                    self._startIndex >= self.tableDataSource.totalSize()) {
                    done = true;
                }
                self._fetchResolveFunc(new self.AsyncIteratorResult(self, new self.FetchListResult(self, self._fetchParams, event[self._DATA], resultMetadata), done));
                self._fetchResolveFunc = null;
                self._fetchParams = null;
            }
            else if (!self._requestEventTriggered) {
                self.dispatchEvent(new oj.DataProviderRefreshEvent());
            }
            self._requestEventTriggered = false;
        });
        tableDataSource.on('add', function (event) {
            var metadataArray = event[self._KEYS].map(function (value) {
                return new self.ItemMetadata(self, value);
            });
            var keySet = new Set();
            event[self._KEYS].map(function (key) {
                keySet.add(key);
            });
            var operationEventDetail = new self.DataProviderAddOperationEventDetail(self, keySet, null, metadataArray, event[self._DATA], event[self._INDEXES]);
            var mutationEventDetail = new self.DataProviderMutationEventDetail(self, operationEventDetail, null, null);
            self.dispatchEvent(new oj.DataProviderMutationEvent(mutationEventDetail));
        });
        tableDataSource.on('remove', function (event) {
            var metadataArray = event[self._KEYS].map(function (value) {
                return new self.ItemMetadata(self, value);
            });
            var keySet = new Set();
            event[self._KEYS].map(function (key) {
                keySet.add(key);
            });
            var operationEventDetail = new self.DataProviderOperationEventDetail(self, keySet, metadataArray, event[self._DATA], event[self._INDEXES]);
            var mutationEventDetail = new self.DataProviderMutationEventDetail(self, null, operationEventDetail, null);
            self.dispatchEvent(new oj.DataProviderMutationEvent(mutationEventDetail));
        });
        tableDataSource.on('reset', function (event) {
            if (!self._requestEventTriggered) {
                self._startIndex = 0;
                self.dispatchEvent(new oj.DataProviderRefreshEvent());
            }
        });
        tableDataSource.on(this._SORT, function (event) {
            if (!self._ignoreSortEvent) {
                self._startIndex = null;
                self.dispatchEvent(new oj.DataProviderRefreshEvent());
            }
        });
        tableDataSource.on('change', function (event) {
            var metadataArray = event[self._KEYS].map(function (value) {
                return new self.ItemMetadata(self, value);
            });
            var keySet = new Set();
            event[self._KEYS].map(function (key) {
                keySet.add(key);
            });
            var operationEventDetail = new self.DataProviderOperationEventDetail(self, keySet, metadataArray, event[self._DATA], event[self._INDEXES]);
            var mutationEventDetail = new self.DataProviderMutationEventDetail(self, null, null, operationEventDetail);
            self.dispatchEvent(new oj.DataProviderMutationEvent(mutationEventDetail));
        });
        tableDataSource.on(this._REFRESH, function (event) {
            if (!self._isFetching &&
                !self._requestEventTriggered) {
                if (event[self._OFFSET] != null) {
                    self._startIndex = event[self._OFFSET];
                }
                else {
                    self._startIndex = null;
                }
                self.dispatchEvent(new oj.DataProviderRefreshEvent());
            }
            self._requestEventTriggered = false;
        });
        tableDataSource.on('request', function (event) {
            if (typeof oj.Model !== "undefined" &&
                event instanceof oj.Model) {
                // ignore request events by oj.Model. Those will be followed by row
                // mutation events anyway
                return;
            }
            if (!self._isFetching) {
                if (event[self._STARTINDEX] > 0 &&
                    self.getStartItemIndex() == 0) {
                    self._startIndex = event[self._STARTINDEX];
                }
                // dispatch a refresh event which will trigger a the component to
                // do a fetchFirst. However, the fact that we are receiving a request
                // event means that a fetch was already done on the underlying TableDataSource.
                // So we don't need to do another fetch once a fetchFirst comes in, we can
                // just resolve with the results from the paired sync event.
                self._requestEventTriggered = true;
                self.dispatchEvent(new oj.DataProviderRefreshEvent());
            }
        });
        tableDataSource.on('error', function (event) {
            if (self._fetchRejectFunc) {
                self._fetchRejectFunc(event);
            }
            self._isFetching = false;
            self._requestEventTriggered = false;
        });
        tableDataSource.on('page', function (event) {
            self._isFetching = false;
            self._requestEventTriggered = false;
            var options = {};
            options['detail'] = event;
            self.dispatchEvent(new oj.GenericEvent(oj.PagingModel.EventType['PAGE'], options));
        });
    };
    /**
     * Check if it's a PagingModel TableDataSource
     */
    TableDataSourceAdapter.prototype._isPagingModelTableDataSource = function () {
        if (this.tableDataSource['getStartItemIndex'] != null) {
            return true;
        }
        return false;
    };
    return TableDataSourceAdapter;
}());
oj.EventTargetMixin.applyMixin(TableDataSourceAdapter);
oj['TableDataSourceAdapter'] = TableDataSourceAdapter;
oj.TableDataSourceAdapter = TableDataSourceAdapter;

var ArrayDataProvider = /** @class */ (function () {
    function ArrayDataProvider(data, options) {
        this.data = data;
        this.options = options;
        this._KEY = 'key';
        this._KEYS = 'keys';
        this._AFTERKEYS = 'afterKeys';
        this._DIRECTION = 'direction';
        this._ATTRIBUTE = 'attribute';
        this._SORT = 'sort';
        this._SORTCRITERIA = 'sortCriteria';
        this._DATA = 'data';
        this._METADATA = 'metadata';
        this._INDEXES = 'indexes';
        this._OFFSET = 'offset';
        this._SIZE = 'size';
        this._IDATTRIBUTE = 'idAttribute';
        this._IMPLICITSORT = 'implicitSort';
        this._SORTCOMPARATORS = 'sortComparators';
        this._COMPARATORS = 'comparators';
        this._COMPARATOR = 'comparator';
        this._RESULTS = 'results';
        this._CONTAINS = 'contains';
        this._FETCHPARAMETERS = 'fetchParameters';
        this._CONTAINSPARAMETERS = 'containsParameters';
        this._VALUE = 'value';
        this._DONE = 'done';
        this._ADD = 'add';
        this._REMOVE = 'remove';
        this._UPDATE = 'update';
        this.Item = /** @class */ (function () {
            function class_1(_parent, metadata, data) {
                this._parent = _parent;
                this.metadata = metadata;
                this.data = data;
                this[_parent._METADATA] = metadata;
                this[_parent._DATA] = data;
            }
            return class_1;
        }());
        this.ItemMetadata = /** @class */ (function () {
            function class_2(_parent, key) {
                this._parent = _parent;
                this.key = key;
                this[_parent._KEY] = key;
            }
            return class_2;
        }());
        this.FetchByKeysResults = /** @class */ (function () {
            function class_3(_parent, fetchParameters, results) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.results = results;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._RESULTS] = results;
            }
            return class_3;
        }());
        this.ContainsKeysResults = /** @class */ (function () {
            function class_4(_parent, containsParameters, results) {
                this._parent = _parent;
                this.containsParameters = containsParameters;
                this.results = results;
                this[_parent._CONTAINSPARAMETERS] = containsParameters;
                this[_parent._RESULTS] = results;
            }
            return class_4;
        }());
        this.FetchByOffsetResults = /** @class */ (function () {
            function class_5(_parent, fetchParameters, results, done) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.results = results;
                this.done = done;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._RESULTS] = results;
                this[_parent._DONE] = done;
            }
            return class_5;
        }());
        this.FetchListParameters = /** @class */ (function () {
            function class_6(_parent, size, sortCriteria) {
                this._parent = _parent;
                this.size = size;
                this.sortCriteria = sortCriteria;
                this[_parent._SIZE] = size;
                this[_parent._SORTCRITERIA] = sortCriteria;
            }
            return class_6;
        }());
        this.FetchListResult = /** @class */ (function () {
            function class_7(_parent, fetchParameters, data, metadata) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.data = data;
                this.metadata = metadata;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._DATA] = data;
                this[_parent._METADATA] = metadata;
            }
            return class_7;
        }());
        this.AsyncIterable = /** @class */ (function () {
            function class_8(_parent, _asyncIterator) {
                this._parent = _parent;
                this._asyncIterator = _asyncIterator;
                this[Symbol.asyncIterator] = function () {
                    return this._asyncIterator;
                };
            }
            return class_8;
        }());
        this.AsyncIterator = /** @class */ (function () {
            function class_9(_parent, _nextFunc, _params, _offset) {
                this._parent = _parent;
                this._nextFunc = _nextFunc;
                this._params = _params;
                this._offset = _offset;
                this._cachedOffset = _offset;
            }
            class_9.prototype['next'] = function () {
                var result = this._nextFunc(this._params, this._cachedOffset);
                this._cachedOffset = this._cachedOffset + result.value[this._parent._DATA].length;
                return Promise.resolve(result);
            };
            return class_9;
        }());
        this.AsyncIteratorResult = /** @class */ (function () {
            function class_10(_parent, value, done) {
                this._parent = _parent;
                this.value = value;
                this.done = done;
                this[_parent._VALUE] = value;
                this[_parent._DONE] = done;
            }
            return class_10;
        }());
        this.DataProviderMutationEventDetail = /** @class */ (function () {
            function class_11(_parent, add, remove, update) {
                this._parent = _parent;
                this.add = add;
                this.remove = remove;
                this.update = update;
                this[_parent._ADD] = add;
                this[_parent._REMOVE] = remove;
                this[_parent._UPDATE] = update;
            }
            return class_11;
        }());
        this.DataProviderOperationEventDetail = /** @class */ (function () {
            function class_12(_parent, keys, metadata, data, indexes) {
                this._parent = _parent;
                this.keys = keys;
                this.metadata = metadata;
                this.data = data;
                this.indexes = indexes;
                this[_parent._KEYS] = keys;
                this[_parent._METADATA] = metadata;
                this[_parent._DATA] = data;
                this[_parent._INDEXES] = indexes;
            }
            return class_12;
        }());
        this.DataProviderAddOperationEventDetail = /** @class */ (function () {
            function class_13(_parent, keys, afterKeys, metadata, data, indexes) {
                this._parent = _parent;
                this.keys = keys;
                this.afterKeys = afterKeys;
                this.metadata = metadata;
                this.data = data;
                this.indexes = indexes;
                this[_parent._KEYS] = keys;
                this[_parent._AFTERKEYS] = afterKeys;
                this[_parent._METADATA] = metadata;
                this[_parent._DATA] = data;
                this[_parent._INDEXES] = indexes;
            }
            return class_13;
        }());
        this._cachedIndexMap = [];
        this._sequenceNum = 0;
        this._subscribeObservableArray(data);
        if (options != null && options[this._KEYS] != null) {
            this._keysSpecified = true;
            this._keys = options[this._KEYS];
        }
    }
    ArrayDataProvider.prototype.containsKeys = function (params) {
        var self = this;
        return this.fetchByKeys(params).then(function (fetchByKeysResult) {
            var results = new Set();
            params[self._KEYS].forEach(function (key) {
                if (fetchByKeysResult[self._RESULTS].get(key) != null) {
                    results.add(key);
                }
            });
            return Promise.resolve(new self.ContainsKeysResults(self, params, results));
        });
    };
    ArrayDataProvider.prototype.fetchByKeys = function (params) {
        var self = this;
        this._generateKeysIfNeeded();
        var results = new Map();
        var keys = this._getKeys();
        var findKeyIndex, i = 0;
        params[this._KEYS].forEach(function (searchKey) {
            findKeyIndex = null;
            for (i = 0; i < keys.length; i++) {
                if (oj.Object.compareValues(keys[i], searchKey)) {
                    findKeyIndex = i;
                    break;
                }
            }
            if (findKeyIndex != null &&
                findKeyIndex >= 0) {
                results.set(searchKey, new self.Item(self, new self.ItemMetadata(self, searchKey), self._getRowData()[findKeyIndex]));
            }
        });
        return Promise.resolve(new this.FetchByKeysResults(this, params, results));
    };
    ArrayDataProvider.prototype.fetchByOffset = function (params) {
        var self = this;
        var size = params != null ? params[this._SIZE] : -1;
        var sortCriteria = params != null ? params[this._SORTCRITERIA] : null;
        var offset = params != null ? params[this._OFFSET] > 0 ? params[this._OFFSET] : 0 : 0;
        this._generateKeysIfNeeded();
        var fetchParams = new this.FetchListParameters(this, size, sortCriteria);
        var iteratorResults = this._fetchFrom(fetchParams, offset);
        var value = iteratorResults[this._VALUE];
        var done = iteratorResults[this._DONE];
        var data = value[this._DATA];
        var keys = value[this._METADATA].map(function (value) {
            return value[self._KEY];
        });
        var resultsArray = data.map(function (value, index) {
            return new self.Item(self, new self.ItemMetadata(self, keys[index]), value);
        });
        return Promise.resolve(new this.FetchByOffsetResults(this, params, resultsArray, done));
    };
    /**
     * Fetch the first block of data
     */
    ArrayDataProvider.prototype.fetchFirst = function (params) {
        var offset = 0;
        return new this.AsyncIterable(this, new this.AsyncIterator(this, this._fetchFrom.bind(this), params, offset));
    };
    /**
     * Determines whether this DataProvider supports certain feature.
     */
    ArrayDataProvider.prototype.getCapability = function (capabilityName) {
        // Just call the static version of getCapability
        return ArrayDataProvider.getCapability(capabilityName);
    };
    ArrayDataProvider.getCapability = function (capabilityName) {
        if (capabilityName == 'sort') {
            return { attributes: 'multiple' };
        }
        else if (capabilityName == 'fetchByKeys') {
            return { implementation: 'lookup' };
        }
        else if (capabilityName == 'fetchByOffset') {
            return { implementation: 'randomAccess' };
        }
        return null;
    };
    ArrayDataProvider.prototype.getTotalSize = function () {
        return Promise.resolve(this._getRowData().length);
    };
    ArrayDataProvider.prototype.isEmpty = function () {
        return this._getRowData().length > 0 ? 'no' : 'yes';
    };
    /**
     * Get the rows data, unwrapping observableArray if needed.
     */
    ArrayDataProvider.prototype._getRowData = function () {
        if (!(this[this._DATA] instanceof Array)) {
            return this[this._DATA]();
        }
        return this[this._DATA];
    };
    /**
     * Get the keys, unwrapping observableArray if needed.
     */
    ArrayDataProvider.prototype._getKeys = function () {
        if (this._keys != null &&
            !(this._keys instanceof Array)) {
            return this._keys();
        }
        return this._keys;
    };
    /**
     * If observableArray, then subscribe to it
     */
    ArrayDataProvider.prototype._subscribeObservableArray = function (data) {
        if (!(data instanceof Array)) {
            if (!this._isObservableArray(data)) {
                // we only support Array or ko.observableArray
                throw new Error('Invalid data type. ArrayDataProvider only supports Array or observableArray.');
            }
            // subscribe to observableArray arrayChange event to get individual updates
            var self = this;
            data['subscribe'](function (changes) {
                var i, id, dataArray = [], keyArray = [], indexArray = [], metadataArray = [], afterKeyArray = [];
                // first see if we have deletes and adds. If we do then just do a refresh
                var foundDelete = false;
                var foundAdd = false;
                var dispatchRefreshEvent = false;
                for (i = 0; i < changes.length; i++) {
                    if (changes[i]['status'] === 'deleted') {
                        foundDelete = true;
                        break;
                    }
                }
                for (i = 0; i < changes.length; i++) {
                    if (changes[i]['status'] === 'added') {
                        foundAdd = true;
                        break;
                    }
                }
                if (foundAdd && foundDelete) {
                    dispatchRefreshEvent = true;
                }
                if (foundDelete) {
                    for (i = 0; i < changes.length; i++) {
                        if (changes[i]['status'] === 'deleted') {
                            keyArray.push(self._getKeys()[changes[i].index]);
                            dataArray.push(changes[i].value);
                            indexArray.push(changes[i].index);
                        }
                    }
                    if (keyArray.length > 0) {
                        keyArray.map(function (key) {
                            var keyIndex = self._getKeys().indexOf(key);
                            self._keys.splice(keyIndex, 1);
                        });
                    }
                    if (keyArray.length > 0 && !dispatchRefreshEvent) {
                        metadataArray = keyArray.map(function (value) {
                            return new self.ItemMetadata(self, value);
                        });
                        var keySet_1 = new Set();
                        keyArray.map(function (key) {
                            keySet_1.add(key);
                        });
                        var operationEventDetail = new self.DataProviderOperationEventDetail(self, keySet_1, metadataArray, dataArray, indexArray);
                        var mutationEventDetail = new self.DataProviderMutationEventDetail(self, null, operationEventDetail, null);
                        self._mutationEvent = new oj.DataProviderMutationEvent(mutationEventDetail);
                    }
                }
                if (foundAdd) {
                    for (i = 0; i < changes.length; i++) {
                        if (changes[i]['status'] === 'added') {
                            self._generateKeysIfNeeded();
                            id = self._getId(changes[i].value);
                            if (id == null) {
                                id = self._getKeys()[changes[i].index];
                            }
                            if (id == null) {
                                id = self._sequenceNum++;
                                self._keys.splice(changes[i].index, 0, id);
                            }
                            if (self._getKeys().indexOf(id) == -1) {
                                self._keys.splice(changes[i].index, 0, id);
                            }
                            keyArray.push(id);
                            var afterKey = self._getKeys()[changes[i].index + 1];
                            afterKey = afterKey == null ? '' : afterKey;
                            afterKeyArray.push(afterKey);
                            dataArray.push(changes[i].value);
                            indexArray.push(changes[i].index);
                        }
                    }
                    if (keyArray.length > 0 && !dispatchRefreshEvent) {
                        metadataArray = keyArray.map(function (value) {
                            return new self.ItemMetadata(self, value);
                        });
                        var keySet_2 = new Set();
                        keyArray.map(function (key) {
                            keySet_2.add(key);
                        });
                        var afterKeySet_1 = new Set();
                        afterKeyArray.map(function (key) {
                            afterKeySet_1.add(key);
                        });
                        var operationEventDetail = new self.DataProviderAddOperationEventDetail(self, keySet_2, afterKeySet_1, metadataArray, dataArray, indexArray);
                        var mutationEventDetail = new self.DataProviderMutationEventDetail(self, operationEventDetail, null, null);
                        self._mutationEvent = new oj.DataProviderMutationEvent(mutationEventDetail);
                    }
                }
            }, null, 'arrayChange');
            data['subscribe'](function (changes) {
                if (self._mutationEvent) {
                    self.dispatchEvent(self._mutationEvent);
                }
                else {
                    self.dispatchEvent(new oj.DataProviderRefreshEvent());
                }
                self._mutationEvent = null;
            }, null, 'change');
        }
    };
    /**
     * Check if observableArray
     */
    ArrayDataProvider.prototype._isObservableArray = function (obj) {
        return ko.isObservable(obj) && !(obj['destroyAll'] === undefined);
    };
    /**
     * Generate keys array if it wasn't passed in options.keys
     */
    ArrayDataProvider.prototype._generateKeysIfNeeded = function () {
        if (this._keys == null) {
            var idAttribute = this.options != null ? this.options[this._IDATTRIBUTE] : null;
            this._keys = [];
            var rowData = this._getRowData();
            var id = void 0, i = 0;
            for (i = 0; i < rowData.length; i++) {
                id = this._getId(rowData[i]);
                if (id == null || idAttribute == '@index') {
                    id = this._sequenceNum++;
                }
                this._keys[i] = id;
            }
            return true;
        }
        return false;
    };
    /**
     * Get id value for row
     */
    ArrayDataProvider.prototype._getId = function (row) {
        var id;
        var idAttribute = this.options != null ? this.options[this._IDATTRIBUTE] : null;
        if (idAttribute != null) {
            if (Array.isArray(idAttribute)) {
                var i;
                id = [];
                for (i = 0; i < idAttribute.length; i++) {
                    id[i] = this._getVal(row, idAttribute[i]);
                }
            }
            else if (idAttribute == '@value') {
                id = this._getAllVals(row);
            }
            else {
                id = this._getVal(row, idAttribute);
            }
            return id;
        }
        else {
            return null;
        }
    };
    ;
    /**
     * Get value for attribute
     */
    ArrayDataProvider.prototype._getVal = function (val, attr) {
        if (typeof (val[attr]) == 'function') {
            return val[attr]();
        }
        return val[attr];
    };
    ;
    /**
     * Get all values in a row
     */
    ArrayDataProvider.prototype._getAllVals = function (val) {
        var self = this;
        return Object.keys(val).map(function (key) {
            return self._getVal(val, key);
        });
    };
    ;
    /**
     * Fetch from offset
     */
    ArrayDataProvider.prototype._fetchFrom = function (params, offset) {
        var self = this;
        this._generateKeysIfNeeded();
        var sortCriteria = params != null ? params[this._SORTCRITERIA] : null;
        var indexMap = this._getCachedIndexMap(sortCriteria);
        var mappedData = indexMap.map(function (index) {
            return self._getRowData()[index];
        });
        var mappedKeys = indexMap.map(function (index) {
            return self._getKeys()[index];
        });
        var fetchSize = params != null ? params[this._SIZE] > 0 ? params[this._SIZE] : params[this._SIZE] < 0 ? self._getKeys().length : 25 : 25;
        var resultData = mappedData.slice(offset, offset + fetchSize);
        var resultKeys = mappedKeys.slice(offset, offset + fetchSize);
        var resultMetadata = resultKeys.map(function (value) {
            return new self.ItemMetadata(self, value);
        });
        var hasMore = offset + fetchSize < mappedData.length ? true : false;
        var mergedSortCriteria = this._mergeSortCriteria(sortCriteria);
        if (mergedSortCriteria != null) {
            params = params || {};
            params[this._SORTCRITERIA] = mergedSortCriteria;
        }
        var result = new this.FetchListResult(this, params, resultData, resultMetadata);
        return new this.AsyncIteratorResult(this, result, !hasMore);
    };
    /**
     * Get cached index map
     */
    ArrayDataProvider.prototype._getCachedIndexMap = function (sortCriteria) {
        var dataIndexes = this._getRowData().map(function (value, index) {
            return index;
        });
        var indexMap = this._sortData(dataIndexes, sortCriteria);
        return indexMap;
    };
    /**
     * Sort data
     */
    ArrayDataProvider.prototype._sortData = function (indexMap, sortCriteria) {
        var self = this;
        var indexedData = indexMap.map(function (index) {
            return { index: index, value: self._getRowData()[index] };
        });
        if (sortCriteria != null) {
            indexedData.sort(this._getSortComparator(sortCriteria));
        }
        return indexedData.map(function (item) {
            return item.index;
        });
    };
    /**
     * Apply sort comparators
     */
    ArrayDataProvider.prototype._getSortComparator = function (sortCriteria) {
        var self = this;
        return function (x, y) {
            var sortComparators = self.options != null ? self.options[self._SORTCOMPARATORS] : null;
            var i, direction, attribute, comparator;
            for (i = 0; i < sortCriteria.length; i++) {
                direction = sortCriteria[i][self._DIRECTION];
                attribute = sortCriteria[i][self._ATTRIBUTE];
                comparator = null;
                if (sortComparators != null) {
                    comparator = sortComparators[self._COMPARATORS].get(attribute);
                }
                if (comparator != null) {
                    var descendingResult = direction == 'descending' ? -1 : 1;
                    var comparatorResult = comparator(x.value[attribute], y.value[attribute]) * descendingResult;
                    if (comparatorResult != 0) {
                        return comparatorResult;
                    }
                }
                else {
                    var compareResult = 0;
                    var strX = (typeof x.value[attribute] === 'string') ? x.value[attribute] : new String(x.value[attribute]);
                    var strY = (typeof y.value[attribute] === 'string') ? y.value[attribute] : new String(y.value[attribute]);
                    if (direction == 'ascending') {
                        compareResult = strX.localeCompare(strY, undefined, { 'numeric': true, 'sensitivity': 'base' });
                    }
                    else {
                        compareResult = strY.localeCompare(strX, undefined, { 'numeric': true, 'sensitivity': 'base' });
                    }
                    if (compareResult != 0) {
                        return compareResult;
                    }
                }
            }
            return 0;
        };
    };
    /**
     * Merge sort criteria
     */
    ArrayDataProvider.prototype._mergeSortCriteria = function (sortCriteria) {
        var implicitSort = this.options != null ? this.options[this._IMPLICITSORT] : null;
        if (implicitSort != null) {
            if (sortCriteria == null) {
                return implicitSort;
            }
            // merge
            var mergedSortCriteria = sortCriteria.slice(0);
            var i = void 0, j = void 0, found = void 0;
            for (i = 0; i < implicitSort.length; i++) {
                found = false;
                for (j = 0; j < mergedSortCriteria.length; j++) {
                    if (mergedSortCriteria[j][this._ATTRIBUTE] == implicitSort[i][this._ATTRIBUTE]) {
                        found = true;
                    }
                }
                if (!found) {
                    mergedSortCriteria.push(implicitSort[i]);
                }
            }
            return mergedSortCriteria;
        }
        else {
            return sortCriteria;
        }
    };
    return ArrayDataProvider;
}());
oj['ArrayDataProvider'] = ArrayDataProvider;
oj.EventTargetMixin.applyMixin(ArrayDataProvider);

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @class oj.ArrayDataProvider
 * @implements oj.DataProvider
 * @classdesc This class implements {@link oj.DataProvider}.  
 *            Object representing data available from an array.  This dataprovider can be used by [ListView]{@link oj.ojListView}, [NavigationList]{@link oj.ojNavigationList},
 *            [TabBar]{@link oj.ojTabBar}, and [Table]{@link oj.ojTable}.<br><br>
 *            See the <a href="../jetCookbook.html?component=table&demo=basicTable">Table - Base Table</a> demo for an example.<br><br>
 *            The default sorting algorithm used when a sortCriteria is passed into fetchFirst is natural sort.
 *
 * <h3 id="events-section">
 *   Events
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#events-section"></a>
 * </h3>
 * Consumers can add event listeners to listen for the following event types and respond to data change.
 * <h4 id="event:mutate" class="name">
 *   mutate
 * </h4>
 * This event is fired when items have been added or removed from the data.
 * <p>
 * Event payload is found under <code class="prettyprint">event.detail</code>, which implements the {@link oj.DataProviderMutationEventDetail} interface.
 * </p>
 *
 * <h4 id="event:refresh" class="name">
 *   refresh
 * </h4>
 * This event is fired when the data has been refreshed and components need to re-fetch the data.
 * <p>
 * This event contains no additional event payload.
 * </p>
 *
 * <i>Example of consumer listening for the "mutate" event type:</i>
 * <pre class="prettyprint"><code>var listener = function(event) {
 *   if (event.detail.remove) {
 *     var removeDetail = event.detail.remove;
 *     // Handle removed items
 *   }
 * };
 * dataProvider.addEventListener("mutate", listener);
 * </code></pre>
 *
 * @param {(Array|function():Array)} data data supported by the components
 *                                      <p>This can be either an Array, or a Knockout observableArray.</p>
 * @param {Object=} options Options for the ArrayDataProvider
 * @param {oj.SortComparators=} options.sortComparators Optional {@link oj.sortComparator} to use for sort.
 * @param {Array.<oj.SortCriterion>=} options.implicitSort Optional array of {@link oj.sortCriterion} used to specify sort information when the data loaded into the dataprovider is already sorted.
 * @param {(Array|function():Array)=} options.keys Optional keys for the data. If not supplied, then the keys are generated according options.idAttribute. If that is also not supplied then index is used as key. 
 * @param {(string | Array.<string>)=} options.idAttribute Optionally the field name which stores the id in the data. Can be a string denoting a single key attribute or an array
 *                                                  of strings for multiple key attributes. @index causes ArrayDataProvider to use index as key and @value will cause ArrayDataProvider to
 *                                                  use all attributes as key. @index is the default.
 * @ojsignature [{target: "Type",
 *               value: "class ArrayDataProvider<K, D> implements DataProvider<K, D>"},
 *               {target: "Type",
 *               value: "SortCriterion<D>[]", 
 *               for: "options.implicitSort"},
 *               {target: "Type",
 *               value: "SortComparators<D>", 
 *               for: "options.sortComparators"}]
 * @example
 * // First initialize an array
 * var deptArray = [{DepartmentId: 10, DepartmentName: 'Administration', LocationId: 200},
 *                  {DepartmentId: 20, DepartmentName: 'Marketing', LocationId: 200},
 *                  {DepartmentId: 30, DepartmentName: 'Purchasing', LocationId: 200}];
 * // Then create an ArrayDataProvider object with the array
 * var dataprovider = new oj.ArrayDataProvider(deptArray, {idAttribute: 'DepartmentId'});
 */

/**
 * Check if there are rows containing the specified keys
 *
 * @ojstatus preview
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.ContainsKeysResults>} Promise which resolves to {@link oj.ContainsKeysResults}
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name containsKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<ContainsKeysResults<K>>"}
 */

/**
 * Fetch rows by keys
 *
 * @ojstatus preview
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.FetchByKeysResults>} Promise which resolves to {@link oj.FetchByKeysResults}
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name fetchByKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<FetchByKeysResults<K, D>>"}
 */

/**
 * Fetch rows by offset
 *
 * @ojstatus preview
 * @param {oj.FetchByOffsetParameters} params Fetch by offset parameters
 * @return {Promise.<oj.FetchByOffsetResults>} Promise which resolves to {@link oj.FetchByOffsetResults}
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name fetchByOffset
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByOffsetParameters<D>): Promise<FetchByOffsetResults<K, D>>"}
 */

/**
 * Fetch the first block of data.
 * 
 * @ojstatus preview
 * @param {oj.FetchListParameters=} params Fetch parameters
 * @return {AsyncIterable.<oj.FetchListResult>} AsyncIterable with {@link oj.FetchListResult}
 * @see {@link https://github.com/tc39/proposal-async-iteration} for further information on AsyncIterable.
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name fetchFirst
 * @ojsignature {target: "Type",
 *               value: "<F extends FetchListResult<K, D>>(params?: FetchListParameters<D>): AsyncIterable<F>"}
 */

/**
 * Determines whether this DataProvider supports certain feature.
 * 
 * @ojstatus preview
 * @param {string} capabilityName capability name. Supported capability names are:
 *                  "fetchByKeys", "fetchByOffset", and "sort".
 * @return {Object} capability information or null if unsupported
 * <ul>
 *   <li>If capabilityName is "fetchByKeys", returns a {@link oj.FetchByKeysCapability} object.</li>
 *   <li>If capabilityName is "fetchByOffset", returns a {@link oj.FetchByOffsetCapability} object.</li>
 *   <li>If capabilityName is "sort", returns a {@link oj.SortCapability} object.</li>
 * </ul>
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name getCapability
 * @ojsignature {target: "Type",
 *               value: "(capabilityName?: string): any"}
 */

/**
 * Return the total number of rows in this dataprovider
 * 
 * @ojstatus preview
 * @return {Promise.<number>} Returns a Promise which resolves to the total number of rows. -1 is unknown row count.
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name getTotalSize
 */

/**
 * Return a string that indicates if this data provider is empty
 * 
 * @ojstatus preview
 * @return {"yes"|"no"|"unknown"} a string that indicates if this data provider is empty. Valid values are:
 *                  "yes": this data provider is empty.
 *                  "no": this data provider is not empty.
 *                  "unknown": it is not known if this data provider is empty until a fetch is made.
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name isEmpty
 */

/**
 * Add a callback function to listen for a specific event type.
 *
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name addEventListener
 * @param {string} eventType The event type to listen for.
 * @param {EventListener} listener The callback function that receives the event notification.
 * @ojsignature {target: "Type",
 *               value: "(eventType: string, listener: EventListener): void"}
 */

/**
 * Remove a listener previously registered with addEventListener.
 *
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name removeEventListener
 * @param {string} eventType The event type that the listener was registered for.
 * @param {EventListener} listener The callback function that was registered.
 * @ojsignature {target: "Type",
 *               value: "(eventType: string, listener: EventListener): void"}
 */

/**
 * Dispatch an event and invoke any registered listeners.
 *
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ArrayDataProvider
 * @instance
 * @method
 * @name dispatchEvent
 * @param {Event} event The event object to dispatch.
 * @return {boolean} Return false if a registered listener has cancelled the event. Return true otherwise.
 * @ojsignature {target: "Type",
 *               value: "(evt: Event): boolean"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @interface oj.SortComparators
 * @ojsignature {target: "Type",
 *               value: "interface SortComparators<D>"}
 */


/**
 * Sort comparators. Map of attribute to comparator function.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.SortComparator
 * @instance
 * @name comparators
 * @type {Map.<string, Function>}
 * @ojsignature {target: "Type",
 *               value: "Map<keyof D, (a: any, b: any) => number>"}
 */

/**
 * End of jsdoc
 */
var DeferredDataProvider = /** @class */ (function () {
    function DeferredDataProvider(_dataProvider, _capabilityFunc) {
        this._dataProvider = _dataProvider;
        this._capabilityFunc = _capabilityFunc;
        this._DATAPROVIDER = 'dataProvider';
        this.AsyncIterable = /** @class */ (function () {
            function class_1(_asyncIterator) {
                this._asyncIterator = _asyncIterator;
                this[Symbol.asyncIterator] = function () {
                    return this._asyncIterator;
                };
            }
            return class_1;
        }());
        this.AsyncIterator = /** @class */ (function () {
            function class_2(_asyncIteratorPromise) {
                this._asyncIteratorPromise = _asyncIteratorPromise;
            }
            class_2.prototype['next'] = function () {
                var self = this;
                return self._asyncIteratorPromise.then(function (asyncIterator) {
                    return asyncIterator['next']();
                });
            };
            return class_2;
        }());
    }
    /**
     * Fetch the first block of data
     */
    DeferredDataProvider.prototype.fetchFirst = function (params) {
        var asyncIteratorPromise = this._getDataProvider().then(function (dataProvider) {
            return dataProvider.fetchFirst(params)[Symbol.asyncIterator]();
        }.bind(params));
        return new this.AsyncIterable(new this.AsyncIterator(asyncIteratorPromise));
    };
    ;
    /**
     * Fetch rows by keys
     */
    DeferredDataProvider.prototype.fetchByKeys = function (params) {
        return this._getDataProvider().then(function (dataProvider) {
            return dataProvider.fetchByKeys(params);
        });
    };
    /**
     * Check if rows are contained by keys
     */
    DeferredDataProvider.prototype.containsKeys = function (params) {
        return this._getDataProvider().then(function (dataProvider) {
            return dataProvider.containsKeys(params);
        });
    };
    /**
     * Fetch rows by offset
     */
    DeferredDataProvider.prototype.fetchByOffset = function (params) {
        return this._getDataProvider().then(function (dataProvider) {
            return dataProvider.fetchByOffset(params);
        });
    };
    /**
     * Returns the total size of the data
     */
    DeferredDataProvider.prototype.getTotalSize = function () {
        return this._getDataProvider().then(function (dataProvider) {
            return dataProvider.getTotalSize();
        });
    };
    /**
     * Returns a string that indicates if this data provider is empty.
     * Returns "unknown" if the dataProvider has not resolved yet.
     */
    DeferredDataProvider.prototype.isEmpty = function () {
        if (!this[this._DATAPROVIDER])
            return "unknown";
        else
            return this[this._DATAPROVIDER].isEmpty();
    };
    /**
     * Determines whether this DataProvider supports certain feature.
     */
    DeferredDataProvider.prototype.getCapability = function (capabilityName) {
        if (this._capabilityFunc)
            return this._capabilityFunc(capabilityName);
        return null;
    };
    /** EVENT TARGET IMPLEMENTATION **/
    DeferredDataProvider.prototype.addEventListener = function (eventType, listener) {
        this._getDataProvider().then(function (dataProvider) {
            dataProvider.addEventListener(eventType, listener);
        });
    };
    ;
    DeferredDataProvider.prototype.removeEventListener = function (eventType, listener) {
        this._getDataProvider().then(function (dataProvider) {
            dataProvider.removeEventListener(eventType, listener);
        });
    };
    ;
    DeferredDataProvider.prototype.dispatchEvent = function (evt) {
        if (!this[this._DATAPROVIDER])
            return false;
        return this[this._DATAPROVIDER].dispatchEvent(evt);
    };
    ;
    /**
     * Returns the resolved dataProvider for this instance
     */
    DeferredDataProvider.prototype._getDataProvider = function () {
        var self = this;
        return this._dataProvider.then(function (dataProvider) {
            if (oj.DataProviderFeatureChecker.isIteratingDataProvider(dataProvider)) {
                if (!self[self._DATAPROVIDER])
                    self[self._DATAPROVIDER] = dataProvider;
                return dataProvider;
            }
            else
                throw new Error('Invalid data type. DeferredDataProvider takes a Promise<DataProvider>');
        });
    };
    return DeferredDataProvider;
}());
oj['DeferredDataProvider'] = DeferredDataProvider;

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.2.0
 * @export
 * @class oj.DeferredDataProvider
 * @implements oj.DataProvider
 * @classdesc This class implements {@link oj.DataProvider}.  
 *            This object represents a data provider that is created with deferred data and can be used by any component that requires a data provider that will be created with data from a Promise.
 * @param {Promise.<oj.DataProvider>} dataProvider A promise that resolves to an oj.DataProvider
 * @param {Function} capabilityFunc An function that implements {@link oj.DataProvider#getCapability}.
 * @ojsignature [{target: "Type",
 *               value: "class DeferredDataProvider<K, D> implements DataProvider<K, D>"},
 *               {target: "Type",
 *               value: "Promise<oj.DataProvider<K, D>>[]", 
 *               for: "dataProvider"}]
 * @example
 * // DeferredDataProvider is used in cases where the data for the data provider will be
 * // provided asynchronously. In the example below, let getDeferredData() be any function
 * // that returns a Promise that will resolve to the final data.
 * var deferredDataPromise = getDeferredData();
 *
 * // Create a Promise that will resolve to a data provider containing the resolved data
 * var dataProviderPromise = deferredDataPromise.then(function(resolvedData){
 *  return new oj.ArrayDataProvider(resolvedData);
 * });
 *
 * // Then create a DeferredDataProvider object with the promise that will resolve to a data provider,
 * // and an implemenation of {@link oj.DataProvider#getCapability}
 * var dataprovider = new oj.DeferredDataProvider(dataProviderPromise, capabilityFunc);
 */

/**
 * Check if there are rows containing the specified keys
 *
 * @ojstatus preview
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.ContainsKeysResults>} Promise which resolves to {@link oj.ContainsKeysResults}
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name containsKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<ContainsKeysResults<K>>"}
 */
  
/**
 * Fetch rows by keys
 *
 * @ojstatus preview
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.FetchByKeysResults>} Promise which resolves to {@link oj.FetchByKeysResults}
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name fetchByKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<FetchByKeysResults<K, D>>"}
 */

/**
 * Fetch rows by offset
 *
 * @ojstatus preview
 * @param {oj.FetchByOffsetParameters} params Fetch by offset parameters
 * @return {Promise.<oj.FetchByOffsetResults>} Promise which resolves to {@link oj.FetchByOffsetResults}
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name fetchByOffset
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByOffsetParameters<D>): Promise<FetchByOffsetResults<K, D>>"}
 */  
  
/**
 * Fetch the first block of data.
 * 
 * @ojstatus preview
 * @param {oj.FetchListParameters=} params Fetch parameters
 * @return {AsyncIterable.<oj.FetchListResult>} AsyncIterable with {@link oj.FetchListResult}
 * @see {@link https://github.com/tc39/proposal-async-iteration} for further information on AsyncIterable.
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name fetchFirst
 * @ojsignature {target: "Type",
 *               value: "<F extends FetchListResult<K, D>>(params?: FetchListParameters<D>): AsyncIterable<F>"}
 */
  
/**
 * Return the total number of rows in this dataprovider
 * 
 * @ojstatus preview
 * @return {Promise.<number>} Returns a Promise which resolves to the total number of rows. -1 is unknown row count.
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name getTotalSize
 */
  
/**
 * Determines whether this DataProvider supports certain feature.
 * 
 * @ojstatus preview
 * @param {string} capabilityName capability name. Supported capability names are:
 *                  "fetchByKeys", "fetchByOffset", and "sort"
 * @return {Object} capability information or null if unsupported
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name getCapability
 */

/**
 * Return a string that indicates if this data provider is empty 
 * @ojstatus preview
 * @return {"yes"|"no"|"unknown"} a string that indicates if this data provider is empty. Valid values are:
 *                  "yes": this data provider is empty.
 *                  "no": this data provider is not empty.
 *                  "unknown": it is not known if this data provider is empty until it has resolved, and a fetch is made.
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name isEmpty
 */
  
/**
 * @ojstatus preview
 * @param {string} eventType The event type to add listener to.
 * @param {EventListener} listener The event listener to add.
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name addEventListener
 */

/**
 * @ojstatus preview
 * @param {string} eventType The event type to remove listener from.
 * @param {EventListener} listener The event listener to remove.
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name removeEventListener
 */

/**
 * @ojstatus preview
 * @param {Event} evt The event to dispatch.
 * @return {boolean} false if the deferred data provider has not resolved or the event has been cancelled, and true otherwise.
 * @export
 * @expose
 * @memberof oj.DeferredDataProvider
 * @instance
 * @method
 * @name dispatchEvent
 */
 
/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true*/

/**
 * @export
 * @class oj.DataSource
 * @extends oj.EventSource
 * @classdesc Object representing data used by table and grid components
 * @param {Object} data data supported by the components
 * @constructor
 * @since 1.0
 * @abstract
 * @ojtsignore
 */
oj.DataSource = function(data)
{
    this.data = data;
    this.Init();
};

/**
 * Subclass from oj.Object 
 * @private
 */
oj.Object.createSubclass(oj.DataSource, oj.EventSource, "oj.DataSource");

/**
 * Initializes the instance.
 * @export
 * @memberof oj.DataSource
 * @return {undefined}
 * @ojtsignore
 */
oj.DataSource.prototype.Init = function()
{
    oj.DataSource.superclass.Init.call(this);
};

/**
 * Determines whether this DataSource supports the specified feature.
 * @method
 * @name getCapability
 * @memberof oj.DataSource
 * @instance
 * @param {string} feature the feature in which its capabilities is inquired. 
 * @return {string|null} the capability of the specified feature.  Returns null if the feature is not recognized.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * @class oj.TreeDataSource
 * @classdesc Abstract class representing hierarchical (tree) data that can be used by different components such as [Indexer]{@link oj.ojIndexer}, [ListView]{@link oj.ojListView}, [NavigationList]{@link oj.ojNavigationList},
 * and [TreeView]{@link oj.ojTreeView}.<br><br>
 * This class is not used directly and is used as the base to implement other subclasses.  Implementations of TreeDataSource must implement all of the methods documented here.<br><br>
 * JET provides the following implementations:<br><br>
 * <table class="generic-table">
 *   <thead>
 *     <th>Subclass</th>
 *     <th>When to Use</th>
 *     <th>Use with</th>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>{@link oj.CollectionTreeDataSource}</td>
 *       <td>When the data is available from an {@link oj.Collection} object, such as an external data source.</td>
 *       <td>ListView<br>NavigationList<br>TreeView<br></td>
 *     </tr>
 *     <tr>
 *       <td>{@link oj.IndexerModelTreeDataSource}</td>
 *       <td>When the data has a tree-like structure that is displayed in a ListView with Indexer.</td>
 *       <td>ListView (with Indexer)<br></td>
 *     </tr>
 *     <tr>
 *       <td>{@link oj.JsonTreeDataSource}</td>
 *       <td>When the data is available from an array of JSON objects that represent tree nodes.</td>
 *       <td>ListView<br>NavigationList<br>TreeView<br></td>
 *     </tr>
 *   </tbody>
 * </table>
 * <br>Refer to the documentation and demos of individual components for more information on how to use them with the TreeDataSource subclasses.</br><br>
 * In case specialized behavior is needed, new subclass can be created by using [oj.Object.createSubclass]{@link oj.Object#createSubclass}.  New subclass can be based on 
 * TreeDataSource, in which case all methods must be implemented, or it can be based on an existing subclass, in which case only methods that require different behavior need 
 * to be overridden.
 *            
 * @param {Object} data data supported by the component
 * @export
 * @extends oj.DataSource
 * @abstract
 * @constructor
 * @since 1.0
 * @ojtsignore
 */
oj.TreeDataSource = function(data)
{
    oj.TreeDataSource.superclass.constructor.call(this, data);
};


/**
 * Subclass TreeDataSource to DataSource
 * @private
 */
oj.Object.createSubclass(oj.TreeDataSource, oj.DataSource, "oj.TreeDataSource");

/**
 * Returns the number of children for a specified parent.  If the value returned is not >= 0 then it is automatically assumed
 * that the child count is unknown.
 * @param {*} parent the parent key.  Specify null if inquiring child count of the root.
 * @return {number} the number of children for the specified parent.
 * @method
 * @name getChildCount
 * @memberof oj.TreeDataSource
 * @instance
 */

/**
 * Fetch the children
 * @param {*} parent the parent key.  Specify null if fetching children from the root.
 * @param {Object} range information about the range, it must contain the following properties: start, count
 * @property {number} range.start the start index of the range in which the children are fetched
 * @property {number} range.count the size of the range in which the children are fetched
 * @param {Object} callbacks the callbacks to be invoke when fetch children operation is completed.  The valid callback
 *        types are "success" and "error"
 * @property {function(oj.NodeSet)} callbacks.success the callback to invoke when fetch completed successfully.
 * @property {function({status: *})=} callbacks.error the callback to invoke when fetch children failed.
 * @param {Object=} options optional parameters for this operation
 * @property {boolean=} options.queueOnly true if this fetch request is to be queued and not execute yet.  The implementation must maintain 
 *        the order of the fetch operations.  When queueOnly is false/null/undefined, any queued fetch operations are then
 *        flushed and executed in the order they are queued.  This flag is ignored if the datasource does not support batching
 * @return {void}
 * @method
 * @name fetchChildren
 * @memberof oj.TreeDataSource
 * @instance
 */

/**
 * Fetch all children and their children recursively from a specified parent.
 * @param {*} parent the parent key.  Specify null to fetch everything from the root (i.e. expand all)
 * @param {Object} callbacks the callbacks to be invoke when fetch children operation is completed.  The valid callback
 *        types are "success" and "error"
 * @property {function(oj.NodeSet)} callbacks.success the callback to invoke when fetch completed successfully.
 * @property {function({status: *})=} callbacks.error the callback to invoke when fetch children failed.
 * @param {Object=} options optional parameters for this operation
 * @property {number=} options.start the index related to parent in which to begin fetching descendants from.  If this is not specified, then value zero will be used
 * @property {number=} options.maxCount the maximum number of children to fetch.  If a non-positive number is specified, then the value is ignored and
 *        there is no maximum fetch count
 * @return {void}
 * @method
 * @name fetchDescendants
 * @memberof oj.TreeDataSource
 * @instance
 */

/**
 * Performs a sort operation on the tree data.
 * @param {Object} criteria the sort criteria.  It must contain the following properties: key, direction
 * @property {*} criteria.key the key identifying the attribute (column) to sort on
 * @property {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object} callbacks callbacks for the sort operation
 * @property {function():void} callbacks.success the callback to invoke when the sort completed successfully
 * @property {function({status: *})=} callbacks.error the callback to invoke when sort failed.
 * @return {void}
 * @method
 * @name sort
 * @memberof oj.TreeDataSource
 * @instance
 */

/**
 * Returns the current sort criteria of the tree data.
 * @return {Object} the current sort criteria.  It should contain the following properties: key, direction where
 *         criteria.key the key identifying the attribute (column) to sort on.  Value is null if it's not sorted.
 *         criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @ojsignature {target: "Type",
 *               value: "{key: any, direction: 'ascending'|'descending'|'none'}",
 *               for: "returns"}
 * @method
 * @name getSortCriteria
 * @memberof oj.TreeDataSource
 * @instance
 */

/**
 * Moves a row from one location to another (different position within the same parent or a completely different parent)
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference row's current children.
 * @param {Object} callbacks the callbacks for the move function
 * @property {function():void} callbacks.success the callback to invoke when the move completed successfully
 * @property {function({status: *})=} callbacks.error the callback to invoke when move failed.
 * @return {void}
 * @method
 * @name move
 * @memberof oj.TreeDataSource
 * @instance
 */ 

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference row's current children.
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @ojsignature {target: "Type",
 *               value: "'valid'|'invalid'",
 *               for: "returns"}
 * @method
 * @name moveOK
 * @memberof oj.TreeDataSource
 * @instance
 */ 

/**
 * Determines whether this TreeDataSource supports the specified feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the valid features "sort", 
 *        "move", "fetchDescendants", "batchFetch"
 * @return {string|null} the name of the feature.  Returns null if the feature is not recognized.
 *         For "sort", the valid return values are: "default", "none".  
 *         For "fetchDescendants", the valid return values are: "enable", "disable", "suboptimal".  
 *         For "move", the valid return values are: "default", "none".  
 *         For "batchFetch", the valid return values are: "enable", "disable".  
 * @method
 * @name getCapability
 * @memberof oj.TreeDataSource
 * @instance
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/

/**
 * @export
 * @class oj.TableDataSource
 * @abstract
 * @classdesc Abstract class representing tabular data that can be used by different components such as [ListView]{@link oj.ojListView}, [NavigationList]{@link oj.ojNavigationList},
 * [PagingControl]{@link oj.ojPagingControl}, [TabBar]{@link oj.ojTabBar}, and [Table]{@link oj.ojTable}.<br><br>
 * This class is not used directly and is used as the base to implement other subclasses.  Implementations of TableDataSource must implement all of the methods documented here.<br><br>
 * JET provides the following implementations:<br><br>
 * <table class="generic-table">
 *   <thead>
 *     <th>Subclass</th>
 *     <th>When to Use</th>
 *     <th>Use with</th>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>{@link oj.ArrayTableDataSource}</td>
 *       <td>When the data is available from an array.</td>
 *       <td>ListView<br>NavigationList<br>TabBar<br>Table<br></td>
 *     </tr>
 *     <tr>
 *       <td>{@link oj.CollectionTableDataSource}</td>
 *       <td>When the data is available from an {@link oj.Collection} object, such as an external data source.</td>
 *       <td>ListView<br>NavigationList<br>TabBar<br>Table<br></td>
 *     </tr>
 *     <tr>
 *       <td>{@link oj.FlattenedTreeTableDataSource}</td>
 *       <td>When the data has a tree-like structure that is displayed in a Table with RowExpander.</td>
 *       <td>Table (with RowExpander)<br></td>
 *     </tr>
 *     <tr>
 *       <td>{@link oj.PagingTableDataSource}</td>
 *       <td>When paging functionality is needed on top of other TableDataSource implementations.</td>
 *       <td>ListView<br>PagingControl<br>Table<br></td>
 *     </tr>
 *   </tbody>
 * </table>
 * <br>Refer to the documentation and demos of individual components for more information on how to use them with the TableDataSource subclasses.</br><br>
 * In case specialized behavior is needed, new subclass can be created by using [oj.Object.createSubclass]{@link oj.Object#createSubclass}.  New subclass can be based on 
 * TableDataSource, in which case all methods must be implemented, or it can be based on an existing subclass, in which case only methods that require different behavior need 
 * to be overridden.
 *            
 * @extends oj.DataSource
 * @param {Object} data data supported by the components
 * @param {Object=} options Options for the TableDataSource
 * @constructor
 * @since 1.0
 * @ojtsignore
 */
oj.TableDataSource = function(data, options)
{
  if (this.constructor == oj.TableDataSource)
  {
    // This should only be called by the constructors of the subclasses. If you need
    // to initialize a new TableDataSource then call the constructors of the subclasses such
    // as oj.ArrayTableDataSource or oj.CollectionTableDataSource.
    var errSummary = oj.TableDataSource._LOGGER_MSG._ERR_TABLE_DATASOURCE_INSTANTIATED_SUMMARY;
    var errDetail = oj.TableDataSource._LOGGER_MSG._ERR_TABLE_DATASOURCE_INSTANTIATED_DETAIL;
    throw new Error(errSummary + '\n' + errDetail);
  }
  // Initialize
  this.data = data;
  this.options = options;
  this.isFetching = false;
  this._startIndex = 0;
  this.Init();
};

/**
 * Subclass from oj.DataSource 
 * @private
 */
oj.Object.createSubclass(oj.TableDataSource, oj.DataSource, "oj.TableDataSource");

/**
 * Initializes the instance.
 * @memberof oj.TableDataSource
 * @instance
 * @override
 * @protected
 */
oj.TableDataSource.prototype.Init = function()
{
  oj.TableDataSource.superclass.Init.call(this);
};

/**
 * @export
 * @expose
 * @memberof oj.TableDataSource
 * @desc The sort criteria. Whenever sort() is called with the criteria parameter, that value is copied to this
 * property. If sort() is called with empty sort criteria then the criteria set in this property is used.
 * 
 * @type {Object} 
 * @property {*} criteria.key The key that identifies which field to sort
 * @property {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 */
oj.TableDataSource.prototype.sortCriteria = null;

/**
 * Return the row data found at the given index.
 * 
 * @param {number} index Index for which to return the row data. 
 * @param {Object=} options Options to control the at.
 * @return {Promise.<null|oj.TableDataSource.RowData>} Promise resolves to a compound object which has the structure below. If the index is out of range, Promise resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @method
 * @name at
 * @memberof oj.TableDataSource
 * @instance
 */

/**
 * Fetch the row data.
 * @param {Object=} options Options to control fetch
 * @property {number} [options.startIndex] The index at which to start fetching records.
 * @property {boolean} [options.silent] If set, do not fire a sync event.
 * @return {Promise.<null|oj.TableDataSource.RowDatas>} Promise object resolves to a compound object which contains an array of row data objects, an array of ids, and the startIndex triggering done when complete.<p>
 *         The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>startIndex</b></td><td>The startIndex for the returned set of rows</td></tr>
 * </tbody>
 * </table>  
 * @method
 * @name fetch
 * @memberof oj.TableDataSource
 * @instance
 */

/**
 * Return the first row data whose id value is the given id
 * @param {string} id ID for which to return the row data, if found. 
 * @param {Object=} options Options to control the get.
 * @return {Promise.<null|oj.TableDataSource.RowData>} Promise which resolves to a compound object which has the structure below where the id matches the given id. If none are found, resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @method
 * @name get
 * @memberof oj.TableDataSource
 * @instance
 */

/**
 * Determines whether this TableDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none".  
 *         Returns null if the feature is not recognized.
 * @memberof oj.TableDataSource
 * @instance
 * @method
 * @name getCapability
 */

/**
 * Performs a sort on the data source. Null criteria clears the existing sort.
 * @param {Object} [criteria] the sort criteria.
 * @property {*} criteria.key The key that identifies which field to sort
 * @property {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @return {Promise.<null>} promise object triggering done when complete.
 * @method
 * @name sort
 * @memberof oj.TableDataSource
 * @instance
 */


/**
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @method
 * @name totalSize
 * @memberof oj.TableDataSource
 * @instance
 */

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @ojsignature {target: "Type", for: "returns", value: "'actual'|'estimate'|'atLeast'|'unknown'"}
 * @export
 * @expose
 * @memberof oj.TableDataSource
 * @instance 
 */
oj.TableDataSource.prototype.totalSizeConfidence = function()
{ 
  return "actual";
};

 /**
 * Row Data returned by methods like, at/get.
 * @typedef {Object} oj.TableDataSource.RowData
 * @property {Object} data The raw row data. 
 * @property {number} index The index for the row.
 * @property {*} key The key value for the row.
 */

 /**
 * Row Datas returned by fetch method.
 * @typedef {Object} oj.TableDataSource.RowDatas
 * @property {Array.<Object>} data An array of raw row data. 
 * @property {Array.<*>} keys An array of key values for the rows.
 * @property {number} startIndex The startIndex for the returned set of rows.
 */

/**
 * Events which are published by TableDataSource. Any custom TableDataSource
 * implementation must publish the corresponding events.
 * @export
 * Event types
 * @enum {string}
 * @memberof oj.TableDataSource
 */
oj.TableDataSource.EventType =
  {
    /** Triggered when a Row has been added to a TableDataSource<p>
     * The event payload contains:<p>
     * <table cellspacing="0" style="border-collapse: collapse;">
     * <tbody>
     * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
     * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
     * <tr><td><b>indexes</b></td><td>An array of index values for the rows in post-insert sorted order</td></tr>
     * </tbody>
     * </table>
     */
    'ADD': "add",
    /** Triggered when a Row has been removed from a TableDataSource<p>
     * The event payload contains:<p>
     * <table cellspacing="0" style="border-collapse: collapse;">
     * <tbody>
     * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
     * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
     * <tr><td><b>indexes</b></td><td>An array of index values for the rows in pre-remove sorted order</td></tr>
     * </tbody>
     * </table>
     */
    'REMOVE': "remove",
    /** Triggered when a TableDataSource has been reset */
    'RESET': "reset",
    /** Triggered when a TableDataSource has been refreshed */
    'REFRESH': "refresh",
    /** Triggered when a TableDataSource has been sorted<p>
     * The event payload contains:<p>
     * <table cellspacing="0" style="border-collapse: collapse;">
     * <tbody>
     * <tr><td><b>header</b></td><td>the key of the header which was sorted on</td></tr>		 
     * <tr><td><b>direction</b></td><td>the direction of the sort ascending/descending</td></tr>
     * </tbody>
     * </table>
     */
    'SORT': "sort",
    /** Triggered when a Row's attributes have been changed<p>
     * The event payload contains:<p>
     * <table cellspacing="0" style="border-collapse: collapse;">
     * <tbody>
     * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
     * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
     * <tr><td><b>indexes</b></td><td>An array of index values for the rows</td></tr>
     * </tbody>
     * </table>
     */
    'CHANGE': "change",
    /** Triggered when a TableDataSource has sent a fetch request. It is expected that
     * a component using TableDataSource will go into a busy state upon receiving
     * this event.
     * The event payload contains:<p>
     * <table cellspacing="0" style="border-collapse: collapse;">
     * <tbody>
     * <tr><td><b>startIndex</b></td><td>The start index at which the fetch was invoked</td></tr>
     * </tbody>
     * </table>
     */
    'REQUEST': "request",
    /** Triggered when a TableDataSource has been updated by a fetch. It is expected that
     * a component using TableDataSource will exit busy state upon completion of rendering
     * after receiving this event.<p>
     * The event payload contains:<p>
     * <table cellspacing="0" style="border-collapse: collapse;">
     * <tbody>
     * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
     * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
     * <tr><td><b>startIndex</b></td><td>The start index at which the fetch occurred</td></tr>
     * </tbody>
     * </table>
     */
    'SYNC': "sync",
    /** Triggered when an error occurred on the TableDataSource */
    'ERROR': "error"
  };

/**
 * @private
 */
oj.TableDataSource._LOGGER_MSG =
  {
    '_ERR_TABLE_DATASOURCE_INSTANTIATED_SUMMARY': 'oj.TableDataSource constructor called.',
    '_ERR_TABLE_DATASOURCE_INSTANTIATED_DETAIL':  'Please do not instantiate oj.TableDataSource. Please use one of the subclasses instead such as oj.ArrayTableDataSource or oj.CollectionTableDataSource.',
    '_ERR_DATA_INVALID_TYPE_SUMMARY':             'Invalid data type.',
    '_ERR_DATA_INVALID_TYPE_DETAIL':              'Please specify the appropriate data type.'
  };
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.DataGridDataSource
 * @classdesc The base class for DataGridDataSource.<br>
 * DataGridDataSource implementations must implement all of the functions documented here.
 * @export
 * @extends oj.DataSource
 * @constructor
 * @since 1.0
 * @ojtsignore
 */
oj.DataGridDataSource = function(data)
{
    // oj.DataSource would calls Init
    oj.DataGridDataSource.superclass.constructor.call(this, data);
};

/**
 * Subclass DataGridDataSource to DataSource
 * @private
 */
oj.Object.createSubclass(oj.DataGridDataSource, oj.DataSource, "oj.DataGridDataSource");


/**
 * Returns the total number of rows or columns.  If the value return is not >= 0 then it is automatically assumed
 * that the total count is unknown.
 * @param {string} axis the axis in which we inquire for the total count.  Valid values are "row" and "column".
 * @return {number} the total number of rows/columns.
 * @method
 * @name getCount
 * @memberof oj.DataGridDataSource
 * @instance
 */

/**
 * Returns whether the total count returned in getCount function is an actual or an estimate.
 * @param {string} axis the axis in which we inquire whether the total count is an estimate.  Valid values are 
 *        "row" and "column".
 * @return {string} "exact" if the count returned in getCount function is the actual count, "estimate" if the
 *         count returned in getCount function is an estimate.  The default value is "exact".
 * @method
 * @name getCountPrecision
 * @memberof oj.DataGridDataSource
 * @instance
 */

/**
 * Fetch a range of headers from the data source.
 * @export
 * @method
 * @name fetchHeaders
 * @memberof oj.DataGridDataSource
 * @instance
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count.
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column".
 * @property {number} headerRange.start the start index of the range in which the header data are fetched.
 * @property {number} headerRange.count the size of the range in which the header data are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch headers operation is completed.
 * @property {function({startHeaderSet: Object}, {headerRange: Object}, {endHeaderSet: Object}):undefined} callbacks.success the callback to invoke when fetch headers completed successfully.
 *        The function takes three parameters: HeaderSet object representing start headers, headerRange object passed into the original fetchHeaders call,
 *        and a HeaderSet object representing the end headers along the axis.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @return {undefined}
 */

/**
 * Fetch a range of cells from the data source.
 * @method
 * @name fetchCells
 * @memberof oj.DataGridDataSource
 * @instance
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched. 
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.
 * @property {function({cellSet: Object}, {cellRanges: Array.<Object>}):undefined} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @return {undefined}
 */

/**
 * Returns the keys based on the indexes.
 * @method
 * @name keys
 * @memberof oj.DataGridDataSource
 * @instance
 * @param {Object} indexes the index for each axis
 * @property {number|null} indexes.row the index for the row axis
 * @property {number|null} indexes.column the index for the column axis
 * @return {Promise.<Object>} a Promise object which when resolved returns an object containing the keys for each axis
 */

/**
 * Returns the row and column index based on the keys.
 * @method
 * @name indexes
 * @memberof oj.DataGridDataSource
 * @instance
 * @param {Object} keys the key for each axis
 * @property {*} keys.row the key for the row axis
 * @property {*} keys.column the key for the column axis
 * @return {Promise.<Object>} a Promise object which when resolved returns an object containing the index for each axis
 */

/**
 * Performs a sort on the data source.
 * @method
 * @name sort
 * @memberof oj.DataGridDataSource
 * @instance
 * @param {Object|null} criteria the sort criteria. 
 * @property {string} criteria.axis The axis in which the sort is performed, valid values are "row", "column"
 * @property {*} criteria.key The key that identifies which header to sort
 * @property {string} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the sort operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.  
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @return {undefined}
 */

/**
 * Moves a row from one location to another.
 * @method
 * @name move
 * @memberof oj.DataGridDataSource
 * @instance
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {string} position The position of the moved row relative to the reference row.  
 *        Valid values are: "before", "after" 
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the move operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @return {undefined}
 */

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine
 *        the destination of where the row should moved to.
 * @param {string} position the position of the moved row relative to the reference row.
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @method
 * @name moveOK
 * @memberof oj.DataGridDataSource
 * @instance
 */

/**
 * Determines whether this DataGridDataSource supports certain feature.
 * @method
 * @name getCapability
 * @memberof oj.DataGridDataSource
 * @instance
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none", "row", "column".
 *         For "move", the valid return values are: "row", "none".
 *         Returns null if the feature is not recognized.
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * @class oj.DiagramDataSource
 * @classdesc Base class for Diagram DataSource.  Implementations must implement all of the functions documented here.
 * @param {Object} data data required by the DiagramDataSource implementation
 * @export
 * @extends oj.DataSource
 * @constructor
 * @since 3.0
 * @ojtsignore
 */
oj.DiagramDataSource = function(data)
{
    oj.DiagramDataSource.superclass.constructor.call(this, data);
};


/**
 * Subclass DiagramDataSource to DataSource
 * @private
 */
oj.Object.createSubclass(oj.DiagramDataSource, oj.DataSource, "oj.DiagramDataSource");

/**
 * Object that defines diagram node. The object might also have additional custom properties that can be 
 * mapped to node styles (see {@link oj.ojDiagram#nodeProperties}) 
 * or used during Diagram layout (see {@link oj.DvtDiagramLayoutContextNode#getData}).
 * @typedef {Object} NodeObject
 * @memberof oj.DiagramDataSource
 * @property {string} id Node id
 * @property {Array.<string>=} categories An optional array of additional category strings corresponding to the node.
 * @property {string=} label The label for the node. If the string contains new line characters, a multi-line label will be rendered.
 * @property {Array.<Object>=} nodes An array of objects with properties for the child nodes.
 *                     Set value to null to indicate leaf node, if child nodes could be fetched on container disclosure.
 *                     Set value to 'undefined' to indicate potential child nodes, if the child nodes should be fetched on container disclosure.
 * @property {string=} selectable Specifies whether or not the node will be selectable. Acceptable values are 'off' and 'auto'. Default value is 'auto'
 * @property {string=} draggable  Specifies whether or not the node will be draggable. Acceptable values are 'off' and 'auto'. Default value is 'auto'
 * @property {string=} shortDesc  The description of the node. This is used for accessibility and also for customizing the tooltip text.
 */
 
/**
 * Object that defines diagram link. The object might also have additional custom properties that can be 
 * mapped to link styles (see {@link oj.ojDiagram#linkProperties})
 * or used during Diagram layout (see {@link oj.DvtDiagramLayoutContextLink#getData}).
 * @typedef {Object} LinkObject
 * @memberof oj.DiagramDataSource
 * @property {string} id Link id
 * @property {string} startNode Start node id.
 * @property {string} endNode End node id.
 * @property {Array.<string>=} categories An optional array of additional category strings corresponding to the link.
 * @property {string=} label The label for the link. If the string contains new line characters, a multi-line label will be rendered.
 * @property {string=} selectable Specifies whether or not the node will be selectable. Acceptable values are 'off' and 'auto'. Default value is 'auto'
 * @property {string=} shortDesc  The description of the node. This is used for accessibility and also for customizing the tooltip text.
 */

/**
 * Returns child data for the given parent.
 * The data include all immediate child nodes along with links whose endpoints
 * both descend from the current parent node. 
 * If all the links are available upfront, they can be returned as part of the
 * top-level data (since all nodes descend from the diagram root).
 * If lazy-fetching links is desirable, the most
 * optimal way to return links is as part of the data of the
 * nearest common ancestor of the link's endpoints.
 *
 * @param {Object|null} parentData An object that contains data for the parent node. See {@link oj.DiagramDataSource.NodeObject} section.
 *                     If parentData is null, the method retrieves data for top level nodes.
 * @return {Promise} Promise resolves to a component object with the following structure:<p>
 * <table>
 * <tbody>
 * <tr><td><b>nodes</b></td><td>An array of objects for the child nodes for the given parent.
 *              See {@link oj.DiagramDataSource.NodeObject} section.</td></tr>
 * <tr><td><b>links</b></td><td>An array of objects for the links for the given parent.
 *              See {@link oj.DiagramDataSource.LinkObject} section.</td></tr>
 * </tbody>
 * </table>
 * @method
 * @name getData
 * @memberof oj.DiagramDataSource
 * @instance
 */
 
/**
 * Retrieves number of child nodes
 * @param {Object} nodeData A data object for the node in question.
 *                          See {@link oj.DiagramDataSource.NodeObject} section.
 * @return {number} Number of child nodes if child count is available.
 *                  The method returns 0 for leaf nodes.
 *                  The method returns -1 if the child count is unknown
 *                  (e.g. if the children have not been fetched).
 * @method
 * @name getChildCount
 * @memberof oj.DiagramDataSource
 * @instance
 */
 
/**
 * Indicates whether the specified object contains links
 * that should be discovered in order to display promoted links.
 *
 * @param {Object} nodeData A data object for the container node in question.
 *                          See {@link oj.DiagramDataSource.NodeObject} section
 * @return {string} the valid values are "connected", "disjoint", "unknown"
 * @method
 * @name getDescendantsConnectivity
 * @memberof oj.DiagramDataSource
 * @instance
 */
 
 /**
 * @export
 * Event types
 * @enum {string}
 * @memberof oj.DiagramDataSource
 */
oj.DiagramDataSource.EventType =
{
  /** 
   * Triggered when nodes or links are added to DiagramDataSource.<p>
   * The event payload contains.<p>
   *
   * <table cellspacing="0" style="border-collapse: collapse;">
   * <tbody>
   * <tr><td><b>data</b></td><td>Object</td><td>An object with the following properties:
   *  <ul>
   *    <li>nodes: An array of node objects. See {@link oj.DiagramDataSource.NodeObject} section.</li>
   *    <li>links: An array of link objects. See {@link oj.DiagramDataSource.LinkObject} section.</li>
   *  </ul>
   * </td></tr>
   * <tr><td><b>parentId</b></td><td>string</td><td>parent id for nodes and links</td></tr>
   * <tr><td><b>index</b></td><td>number</td><td>An index where the nodes should be added</td></tr>
   * </tbody>
   * </table>
   */
  'ADD': "add",
  /**
   * Triggered when nodes or links are removed from DiagramDataSource.
   * The event payload contains:<p>
   *
   * <table cellspacing="0" style="border-collapse: collapse;">
   * <tbody>
   * <tr><td><b>data</b></td><td>Object</td><td>An object with the following properties:
   *  <ul>
   *    <li>nodes: An array of node objects. See {@link oj.DiagramDataSource.NodeObject} section.</li>
   *    <li>links: An array of link objects. See {@link oj.DiagramDataSource.LinkObject} section.</li>
   *  </ul>
   * </td></tr>
   * <tr><td><b>parentId</b></td><td>string</td><td>parent id for nodes and links</td></tr>
   * </tbody>
   * </table>
   */
  'REMOVE': "remove",
  /**
   * Triggered when nodes or links are removed from DiagramDataSource.
   * The event payload contains:<p>
   *
   * <table cellspacing="0" style="border-collapse: collapse;">
   * <tbody>
   * <tr><td><b>data</b></td><td>Object</td><td>An object with the following properties:
   *  <ul>
   *    <li>nodes: An array of node objects. See {@link oj.DiagramDataSource.NodeObject} section.</li>
   *    <li>links: An array of link objects. See {@link oj.DiagramDataSource.LinkObject} section.</li>
   *  </ul>
   * </td></tr>
   * </tbody>
   * </table>
   */
  'CHANGE': "change"
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojPagingControl
 * @augments oj.baseComponent
 * @ojstatus preview
 * @since 0.7
 * @ojshortdesc Provides paging functionality for data collections.
 * @ojrole navigation
 * @ojrole button
 * @ojtsimport ojpagingtabledatasource
 * @ojtsignore
 * 
 * @classdesc
 * <h3 id="pagingcontrolOverview-section">
 *   JET PagingControl
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#pagingcontrolOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET PagingControl provides the ability to fetch and display a page of data at a time. The paging control will only fetch the items for the currently displayed page so it should be used for large datasets which can be fetched in pages. 
 * The number of items per page is uniform and configurable. The paging control can be used with any DataSource which implements the oj.PagingModel interface, such as oj.PagingTableDataSource and oj.PagingDataGridDataSource. That means that the Paging Control can be used with ojTable, ojDataGrid, and ojListView.</p>
 * 
 * <pre class="prettyprint"><code>&lt;oj-paging-control
 *   data='{{pagingModel}}'
 *   page-size='10'>
 * &lt;/oj-paging-control></code></pre>
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 * 
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *  
 */
(function() {
  oj.__registerWidget("oj.ojPagingControl", $['oj']['baseComponent'],
    {
      version: '1.0.0',
      defaultElement: '<div>',
      widgetEventPrefix: 'oj',
      options:
        {
          /** 
           * The data to bind to the PagingControl.
           * <p>
           * Must implement the oj.PagingModel interface {@link oj.PagingModel} 
           * @expose 
           * @public 
           * @instance
           * @memberof! oj.ojPagingControl
           * @type {oj.PagingModel}
           * @default null
           *
           * @example <caption>Initialize the PagingControl with the <code class="prettyprint">data</code> attribute specified:</caption>
           * &lt;oj-paging-control data='{{pagingDataSource}}'>&lt;/oj-paging-control>
           *
           * @example <caption>Get or set the <code class="prettyprint">data</code> property after initialization:</caption>
           * // getter
           * var pagingDataSource = myPagingControl.data;
           *
           * // setter
           * myPagingControl.data = pagingDataSource;
           */
          data: null,
          /** 
           * Options for when the PagingControl width is too narrow to accommodate the controls in the paging control
           * @expose 
           * @public 
           * @instance
           * @memberof! oj.ojPagingControl
           * @type {string}
           * @ojvalue {string} "fit" Display as many controls as can fit in the PagingControl width.
           * @ojvalue {string} "none" Display all controls. Controls which cannot fit will be truncated.
           * @default "fit"
           *
           * @example <caption>Initialize the PagingControl with the <code class="prettyprint">overflow</code> attribute specified:</caption>
           * &lt;oj-paging-control overflow='none'>&lt;/oj-paging-control>
           *
           * @example <caption>Get or set the <code class="prettyprint">overflow</code> property after initialization:</caption>
           * // getter
           * var overflowValue = myPagingControl.overflow;
           *
           * // setter
           * myPagingControl.overflow = 'none';
           */
          overflow: 'fit',
          /** 
           * Page size.
           * <p>
           * @expose 
           * @public 
           * @instance
           * @memberof! oj.ojPagingControl
           * @type {number}
           * @default 25
           * @ojmin 1
           *
           * @example <caption>Initialize the PagingControl with the <code class="prettyprint">page-size</code> attribute specified:</caption>
           * &lt;oj-paging-control page-size='50'>&lt;/oj-paging-control>
           *
           * @example <caption>Get or set the <code class="prettyprint">pageSize</code> property after initialization:</caption>
           * // getter
           * var pageSizeValue = myPagingControl.pageSize;
           *
           * // setter
           * myPagingControl.pageSize = 50;
           */
          pageSize: 25,
          /** 
           * Options for page mode. 
           * @expose 
           * @public 
           * @instance
           * @memberof! oj.ojPagingControl
           * @type {Object.<string, Array|number>}
           * @default {'layout': ['auto'], 'type': 'numbers', 'orientation': 'horizontal', 'maxPageLinks': 6}
           *
           * @example <caption>Initialize the PagingControl, overriding some page-options values and leaving the others intact:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-paging-control page-options.some-key='some value' page-options.some-other-key='some other value'>&lt;/oj-paging-control>
           * 
           * &lt;!-- Using JSON notation -->
           * &lt;oj-paging-control page-options='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-paging-control>
           *
           * @example <caption>Get or set the <code class="prettyprint">pageOptions</code> property after initialization:</caption>
           * // Get one
           * var value = myPagingControl.pageOptions.someKey;
           *
           * // Set one, leaving the others intact. Always use the setProperty API for 
           * // subproperties rather than setting a subproperty directly.
           * myPagingControl.setProperty('pageOptions.someKey', 'some value');
           *
           * // Get all
           * var values = myPagingControl.pageOptions;
           *
           * // Set all.  Must list every pageOptions key, as those not listed are lost.
           * myPagingControl.pageOptions = {
           *     someKey: 'some value',
           *     someOtherKey: 'some other value'
           * };
           */
          pageOptions: {
            /**
             * Array of paging navigation controls to be displayed (only applicable for numbers type).
             * <p>This is an array of one or more supported values.</p>
             * <p>See the <a href="#pageOptions">page-options</a> attribute for usage examples.</p>
             * @expose
             * @name pageOptions.layout
             * @public
             * @memberof! oj.ojPagingControl
             * @instance
             * @type {Array}
             * @ojvalue {string} 'auto' The PagingControl decides which controls to display
             * @ojvalue {string} 'all' Display all controls
             * @ojvalue {string} 'input' Display the page input control
             * @ojvalue {string} 'rangeText' Display the page range text control
             * @ojvalue {string} 'pages' Display the page links
             * @ojvalue {string} 'nav' Display the navigation arrows
             * @default ['auto']
             * @ojsignature { target: "Type",
             *                value: "?['auto'|'all'|'input'|'rangeText'|'pages'|'nav']"}
             */
            'layout': ['auto'], 
            /**
             * The type of page links.
             * <p>See the <a href="#pageOptions">page-options</a> attribute for usage examples.</p>
             * @expose
             * @name pageOptions.type
             * @public
             * @memberof! oj.ojPagingControl
             * @instance
             * @type {string}
             * @ojvalue {string} 'numbers' Render numeric page links
             * @ojvalue {string} 'dots' Render dots
             * @ojsignature { target: "Type",
             *                value: "?"}
             * @default "numbers"
             */
            'type': 'numbers', 
            /**
             * The maximum number of page links to display (only applicable for numbers type).
             * An ellipsis '...' will be displayed for pages which exceed the maximum.
             * maxPageLinks must be greater than 4.
             * <p>See the <a href="#pageOptions">page-options</a> attribute for usage examples.</p>
             * @expose
             * @name pageOptions.maxPageLinks
             * @public
             * @memberof! oj.ojPagingControl
             * @instance
             * @type {number}
             * @ojsignature { target: "Type",
             *                value: "?"}
             * @default 6
             * @ojmin 5
             */
            'maxPageLinks': 6, 
            /**
             * The orientation of the page links.
             * <p>See the <a href="#pageOptions">page-options</a> attribute for usage examples.</p>
             * @expose
             * @name pageOptions.orientation
             * @public
             * @memberof! oj.ojPagingControl
             * @instance
             * @type {string}
             * @ojvalue {string} 'horizontal'
             * @ojvalue {string} 'vertical'
             * @ojsignature { target: "Type",
             *                value: "?"}
             * @default "horizontal"
             */
            'orientation': 'horizontal'
          },
          /** 
           * Options for loadMore mode. 
           * @expose 
           * @public 
           * @instance
           * @memberof! oj.ojPagingControl
           * @type {Object.<string, number>}
           * @default {"maxCount": 500}
           *
           * @example <caption>Initialize the PagingControl, overriding load-more-options value:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-paging-control load-more-options.max-count='300'>&lt;/oj-paging-control>
           * 
           * &lt;!-- Using JSON notation -->
           * &lt;oj-paging-control load-more-options='{"maxCount":300}'>&lt;/oj-paging-control>
           *
           * @example <caption>Get or set the <code class="prettyprint">loadMoreOptions</code> property after initialization:</caption>
           * // Get one
           * var value = myPagingControl.loadMoreOptions.maxCount;
           *
           * // Set one. Always use the setProperty API for 
           * // subproperties rather than setting a subproperty directly.
           * myPagingControl.setProperty('loadMoreOptions.maxCount', 300);
           *
           * // Get all
           * var values = myPagingControl.loadMoreOptions;
           *
           * // Set all.  Must list every loadMoreOptions key, as those not listed are lost.
           * myPagingControl.loadMoreOptions = {
           *     maxCount: 300
           * };
           */
          loadMoreOptions: {
            /**
             * The maximum number items to display.
             * <p>See the <a href="#loadMoreOptions">load-more-options</a> attribute for usage examples.</p>
             * @expose
             * @name loadMoreOptions.maxCount
             * @memberof! oj.ojPagingControl
             * @instance
             * @type {number}
             * @default 500
             * @ojmin 0
             */
            'maxCount': 500
          },
          /** 
           * Paging mode.
           * @expose 
           * @public 
           * @instance
           * @memberof! oj.ojPagingControl
           * @type {string}
           * @ojvalue {string} "page" Display paging control in pagination mode.
           * @ojvalue {string} "loadMore" Display paging control in high watermark mode.
           * @default "page"
           *
           * @example <caption>Initialize the PagingControl with the <code class="prettyprint">mode</code> attribute specified:</caption>
           * &lt;oj-paging-control mode='loadMore'>&lt;/oj-paging-control>
           *
           * @example <caption>Get or set the <code class="prettyprint">mode</code> property after initialization:</caption>
           * // getter
           * var modeValue = myPagingControl.mode;
           *
           * // setter
           * myPagingControl.mode = 'loadMore';
           */
          mode: 'page',
          /**
            * Triggered when the paging control has finished rendering
            *
            * @expose 
            * @event 
            * @memberof! oj.ojPagingControl
            * @instance
            * 
            * @example <caption>Initialize the paging control with the <code class="prettyprint">ready</code> callback specified:</caption>
            * $( ".selector" ).ojPagingControl({
            *     "ready": function() {}
            * });
            *
            * @example <caption>Bind an event listener to the <code class="prettyprint">ojready</code> event:</caption>
            * $( ".selector" ).on( "ojready", function() {} );
            *
            * @ignore
            */
          ready: null
        },
      /**
       * @private
       */
      _BUNDLE_KEY:
        {
          _LABEL_ACC_PAGING:                              'labelAccPaging',
          _LABEL_ACC_NAV_FIRST_PAGE:                      'labelAccNavFirstPage',
          _LABEL_ACC_NAV_LAST_PAGE:                       'labelAccNavLastPage',
          _LABEL_ACC_NAV_NEXT_PAGE:                       'labelAccNavNextPage',
          _LABEL_ACC_NAV_PREVIOUS_PAGE:                   'labelAccNavPreviousPage',
          _LABEL_ACC_NAV_PAGE:                            'labelAccNavPage',
          _LABEL_LOAD_MORE:                               'labelLoadMore',
          _LABEL_LOAD_MORE_MAX_ROWS:                      'labelLoadMoreMaxRows',
          _LABEL_NAV_INPUT_PAGE:                          'labelNavInputPage',
          _LABEL_NAV_INPUT_PAGE_MAX:                      'labelNavInputPageMax',
          _LABEL_NAV_INPUT_PAGE_SUMMARY:                  'labelNavInputPageSummary',
          _MSG_ITEM_RANGE_CURRENT:                        'msgItemRangeCurrent',
          _MSG_ITEM_RANGE_CURRENT_SINGLE:                 'msgItemRangeCurrentSingle',
          _MSG_ITEM_RANGE_ITEMS:                          'msgItemRangeItems',
          _MSG_ITEM_RANGE_ATLEAST:                        'msgItemRangeOfAtLeast',
          _MSG_ITEM_RANGE_APPROX:                         'msgItemRangeOfApprox',
          _MSG_ITEM_RANGE_OF:                             'msgItemRangeOf',
          _TIP_NAV_INPUT_PAGE:                            'tipNavInputPage',
          _TIP_NAV_PAGE_LINK:                             'tipNavPageLink',
          _TIP_NAV_NEXT_PAGE:                             'tipNavNextPage',
          _TIP_NAV_PREVIOUS_PAGE:                         'tipNavPreviousPage',
          _TIP_NAV_FIRST_PAGE:                            'tipNavFirstPage',
          _TIP_NAV_LAST_PAGE:                             'tipNavLastPage',
          _ERR_PAGE_INVALID_SUMMARY:                      'pageInvalid.summary',
          _ERR_PAGE_INVALID_DETAIL:                       'pageInvalid.detail',
          _ERR_DATA_INVALID_TYPE_SUMMARY:                 'dataInvalidType.summary',
          _ERR_DATA_INVALID_TYPE_DETAIL:                  'dataInvalidType.detail',
          _ERR_MAXPAGELINKS_INVALID_SUMMARY:              'maxPageLinksInvalid.summary',
          _ERR_MAXPAGELINKS_INVALID_DETAIL:               'maxPageLinksInvalid.detail'
        },
      /**
       * @private
       */
      _MARKER_STYLE_CLASSES:
        {
          _WIDGET:                                        'oj-component',
          _ACTIVE:                                        'oj-active',
          _CLICKABLE_ICON:                                'oj-clickable-icon-nocontext',
          _DISABLED:                                      'oj-disabled',
          _ENABLED:                                       'oj-enabled',
          _FOCUS:                                         'oj-focus',
          _FOCUS_HIGHLIGHT:                               'oj-focus-highlight',
          _HOVER:                                         'oj-hover',
          _SELECTED:                                      'oj-selected'
        },
      /**
       * @private
       */
      _CSS_CLASSES:
        {
          _PAGING_CONTROL_CLASS:                          'oj-pagingcontrol',
          _PAGING_CONTROL_ACC_LABEL_CLASS:                'oj-pagingcontrol-acc-label',
          _PAGING_CONTROL_CONTENT_CLASS:                  'oj-pagingcontrol-content',
          _PAGING_CONTROL_LOAD_MORE_CLASS:                'oj-pagingcontrol-loadmore',
          _PAGING_CONTROL_LOAD_MORE_LINK_CLASS:           'oj-pagingcontrol-loadmore-link',
          _PAGING_CONTROL_LOAD_MORE_MAX_ROWS_CLASS:       'oj-pagingcontrol-loadmore-max-rows',
          _PAGING_CONTROL_LOAD_MORE_RANGE_CLASS:          'oj-pagingcontrol-loadmore-range',
          _PAGING_CONTROL_LOAD_MORE_RANGE_CURRENT_CLASS:  'oj-pagingcontrol-loadmore-range-current',
          _PAGING_CONTROL_LOAD_MORE_RANGE_MAX_CLASS:      'oj-pagingcontrol-loadmore-range-max',
          _PAGING_CONTROL_NAV_CLASS:                      'oj-pagingcontrol-nav',
          _PAGING_CONTROL_NAV_ARROW_CLASS:                'oj-pagingcontrol-nav-arrow',
          _PAGING_CONTROL_NAV_ARROW_SECTION_CLASS:        'oj-pagingcontrol-nav-arrow-section',
          _PAGING_CONTROL_NAV_PAGE_CLASS:                 'oj-pagingcontrol-nav-page',
          _PAGING_CONTROL_NAV_PAGE_ELLIPSIS_CLASS:        'oj-pagingcontrol-nav-page-ellipsis',
          _PAGING_CONTROL_NAV_DOT_CLASS:                  'oj-pagingcontrol-nav-dot',
          _PAGING_CONTROL_NAV_DOT_BULLET_CLASS:           'oj-pagingcontrol-nav-dot-bullet',
          _PAGING_CONTROL_NAV_PAGE_ACC_LABEL_CLASS:       'oj-pagingcontrol-nav-page-acc-label',
          _PAGING_CONTROL_NAV_LABEL_CLASS:                'oj-pagingcontrol-nav-label',
          _PAGING_CONTROL_NAV_INPUT_SECTION_CLASS:        'oj-pagingcontrol-nav-input-section',
          _PAGING_CONTROL_NAV_INPUT_CLASS:                'oj-pagingcontrol-nav-input',
          _PAGING_CONTROL_NAV_INPUT_MAX_CLASS:            'oj-pagingcontrol-nav-input-max',
          _PAGING_CONTROL_NAV_INPUT_SUMMARY_CLASS:        'oj-pagingcontrol-nav-input-summary',
          _PAGING_CONTROL_NAV_INPUT_SUMMARY_CURRENT_CLASS:'oj-pagingcontrol-nav-input-summary-current',
          _PAGING_CONTROL_NAV_INPUT_SUMMARY_MAX_CLASS:    'oj-pagingcontrol-nav-input-summary-max',
          _PAGING_CONTROL_NAV_PAGES_SECTION_CLASS:        'oj-pagingcontrol-nav-pages-section',
          _PAGING_CONTROL_NAV_PAGES_LINKS_CLASS:          'oj-pagingcontrol-nav-pages-links',
          _PAGING_CONTROL_NAV_FIRST_CLASS:                'oj-pagingcontrol-nav-first',
          _PAGING_CONTROL_NAV_FIRST_ACC_LABEL_CLASS:      'oj-pagingcontrol-nav-first-acc-label',
          _PAGING_CONTROL_NAV_PREVIOUS_CLASS:             'oj-pagingcontrol-nav-previous',
          _PAGING_CONTROL_NAV_PREVIOUS_ACC_LABEL_CLASS:   'oj-pagingcontrol-nav-previous-acc-label',
          _PAGING_CONTROL_NAV_NEXT_CLASS:                 'oj-pagingcontrol-nav-next',
          _PAGING_CONTROL_NAV_NEXT_ACC_LABEL_CLASS:       'oj-pagingcontrol-nav-next-acc-label',
          _PAGING_CONTROL_NAV_LAST_CLASS:                 'oj-pagingcontrol-nav-last',
          _PAGING_CONTROL_NAV_LAST_ACC_LABEL_CLASS:       'oj-pagingcontrol-nav-last-acc-label',
          _PAGING_CONTROL_NAV_FIRST_ICON_CLASS:           'oj-pagingcontrol-nav-first-icon',
          _PAGING_CONTROL_NAV_PREVIOUS_ICON_CLASS:        'oj-pagingcontrol-nav-previous-icon',
          _PAGING_CONTROL_NAV_NEXT_ICON_CLASS:            'oj-pagingcontrol-nav-next-icon',
          _PAGING_CONTROL_NAV_LAST_ICON_CLASS:            'oj-pagingcontrol-nav-last-icon',
          _PAGING_CONTROL_NAV_FIRST_VERTICAL_ICON_CLASS:  'oj-pagingcontrol-nav-first-vertical-icon',
          _PAGING_CONTROL_NAV_PREVIOUS_VERTICAL_ICON_CLASS:'oj-pagingcontrol-nav-previous-vertical-icon',
          _PAGING_CONTROL_NAV_NEXT_VERTICAL_ICON_CLASS:   'oj-pagingcontrol-nav-next-vertical-icon',
          _PAGING_CONTROL_NAV_LAST_VERTICAL_ICON_CLASS:   'oj-pagingcontrol-nav-last-vertical-icon',
          _WIDGET_ICON_CLASS:                             'oj-component-icon',
          _HIDDEN_CONTENT_ACC_CLASS:                      'oj-helper-hidden-accessible'
        },
      /**
       * @private
       * @type {string}
       */
      _DATA_ATTR_PAGE_NUM: 'data-oj-pagenum',
      /**
       * @private
       * @type {string}
       */
      _OPTION_ENABLED: 'enabled',
      /**
       * @private
       * @type {string}
       */
      _OPTION_DISABLED: 'disabled',
      /**
       * @private
       * @type {string}
       */
      _TAB_INDEX: 'tabindex',
      /**
       * @private
       */
      _MODE:
        {
          _LOAD_MORE:   'loadMore',
          _PAGE:        'page'
        },
      /**
       * @private
       */
      _PAGE_OPTION_LAYOUT:
        {
          _AUTO:  'auto',
          _ALL: 'all',
          _INPUT: 'input',
          _RANGE_TEXT: 'rangeText',
          _PAGES: 'pages',
          _NAV:   'nav'
        },
      /**
       * @private
       */
      _PAGE_OPTION_DEFAULT_MAX_PAGE_LINKS: 6,
      /**
       * @private
       */
      _TYPE:
        {
          _NUMBERS:   'numbers',
          _DOTS:      'dots'
        },
      /**** start Public APIs ****/

      /**
       * Load the first page of data
       * @expose
       * @memberof oj.ojPagingControl
       * @instance
       * @public
       * @return {Promise.<null>} promise object triggering done when complete.
       * @throws {Error}
       * @export
       * @example <caption>Invoke the <code class="prettyprint">firstPage</code> method:</caption>
       * myPagingControl.firstPage();
       */
      'firstPage': function()
      {
        var data = this._getData();
        if (data != null)
        {
          return this._invokeDataPage(0, false);
        }
        return this._getRejectPromise();
      },
      /**
       * Load the previous page of data
       * @expose
       * @memberof oj.ojPagingControl
       * @instance
       * @public
       * @return {Promise.<null>} promise object triggering done when complete.
       * @throws {Error}
       * @export
       * @example <caption>Invoke the <code class="prettyprint">previousPage</code> method:</caption>
       * myPagingControl.previousPage();
       */
      'previousPage': function()
      {
        var data = this._getData();
        if (data != null)
        {
          var page = this._getCurrentPage();
          // can only go to previous page if on 2nd page or greater
          if (page > 0)
          {
            return this._invokeDataPage(page - 1, false);
          }
        }
        return this._getRejectPromise();
      },
      /**
       * Load the next page of data
       * @expose
       * @memberof oj.ojPagingControl
       * @instance
       * @public
       * @return {Promise.<null>} promise object triggering done when complete.
       * @throws {Error}
       * @export
       * @example <caption>Invoke the <code class="prettyprint">nextPage</code> method:</caption>
       * myPagingControl.nextPage();
       */
      'nextPage': function()
      {
        var data = this._getData();
        if (data != null)
        {
          var page = this._getCurrentPage();
          if (this._isTotalSizeConfidenceActual() && page + 1 <= this._getTotalPages() - 1 || 
              this._getTotalPages() < 0 ||
              !this._isTotalSizeConfidenceActual()) 
          {
            return this._invokeDataPage(page + 1, false);
          }
        }
        return this._getRejectPromise();
      },
      /**
       * Load the last page of data
       * @expose
       * @memberof oj.ojPagingControl
       * @instance
       * @public
       * @return {Promise.<null>} promise object triggering done when complete.
       * @throws {Error}
       * @export
       * @example <caption>Invoke the <code class="prettyprint">lastPage</code> method:</caption>
       * myPagingControl.lastPage();
       */
      'lastPage': function()
      {
        var data = this._getData();
        if (data != null)
        {
          if (this._getTotalPages() > 0)
          {
            return this._invokeDataPage(this._getTotalPages() - 1, false);
          }
        }
        return this._getRejectPromise();
      },
      /**
       * Load the specified page of data
       * @expose
       * @memberof oj.ojPagingControl
       * @instance
       * @public
       * @param {number} page  Page number. 
       * @return {Promise.<null>} promise object triggering done when complete.
       * @throws {Error}
       * @export
       * @example <caption>Invoke the <code class="prettyprint">page</code> method:</caption>
       * myPagingControl.page(5);
       */
      'page': function(page)
      {
        var data = this._getData();
        if (data != null)
        {
          if ((this._isTotalSizeConfidenceActual() && page <= this._getTotalPages() - 1) || 
               this._getTotalPages() < 0 || 
               !this._isTotalSizeConfidenceActual()) 
          {
            return this._invokeDataPage(page, false);
          }
        }
        return this._getRejectPromise();
      },
      /**
       * Load the next set of data
       * @expose
       * @memberof oj.ojPagingControl
       * @instance
       * @public
       * @return {Promise.<null>} promise object triggering done when complete.
       * @throws {Error}
       * @export
       * @example <caption>Invoke the <code class="prettyprint">loadNext</code> method:</caption>
       * myPagingControl.loadNext();
       */
      'loadNext': function()
      {
        var data = this._getData();
        if (data != null)
        {
          return this._invokeDataFetchNext();
        }
        return this._getRejectPromise();
      },
      /**
       * Refresh the paging control.
       * @expose
       * @memberof oj.ojPagingControl
       * @instance
       * @public
       * @export
       * @return {void}
       * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
       * myPagingControl.refresh();
       */
      'refresh': function()
      {
        this._super();
        this._refresh();
      },
      // @inheritdoc
      'getNodeBySubId': function(locator)
      {
        if (locator == null)
        {
          return this.element ? this.element[0] : null;
        }

        var subId = locator['subId'];
        var retval = null;

        if (subId === 'oj-pagingcontrol-nav-input')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-input-max')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_MAX_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-input-summary')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-input-summary-current')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CURRENT_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-input-summary-max')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_MAX_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-first')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-next')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-previous')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-last')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-nav-page')
        {
          var index = locator['index'];
          retval = this._getPagingControlContainer().find("[" + this._DATA_ATTR_PAGE_NUM + "=" + index +"]")[0];
        }
        else if (subId === 'oj-pagingcontrol-load-more-link')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_LINK_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-load-more-range')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-load-more-range-current')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_CURRENT_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-load-more-range-max')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_MAX_CLASS)[0];
        }
        else if (subId === 'oj-pagingcontrol-load-more-max-rows')
        {
          retval = this._getPagingControlContainer().find('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_MAX_ROWS_CLASS)[0];
        }

        // Non-null locators have to be handled by the component subclasses
        if (retval === undefined)
          retval = null;

        return retval;
      },              
      // @inheritdoc
      'getSubIdByNode': function(node)
      {
        if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-input'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_MAX_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-input-max'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-input-summary'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CURRENT_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-input-summary-current'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_MAX_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-input-summary-max'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-first'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-next'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-previous'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-last'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGE_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-nav-page',
                  'index': $(node).attr(this._DATA_ATTR_PAGE_NUM)};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_LINK_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-load-more-link'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-load-more-range'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_CURRENT_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-load-more-range-current'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_MAX_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-load-more-range-max'};
        }
        else if ($(node).hasClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_MAX_ROWS_CLASS))
        {
          return {'subId': 'oj-pagingcontrol-load-more-max-rows'};
        }

        return null;
      },              
      /**** end Public APIs ****/

      /**** start internal widget functions ****/
      
      /**
       * @override
       * @protected
       * @instance
       * @memberof! oj.ojPagingControl
       */
      _ComponentCreate : function ()
      {
        this._super();
        this._draw();
        this._on(this._events);
      },
      /**
       * Initialize the paging control after creation
       * @protected
       * @override
       * @memberof! oj.ojPagingControl
       */        
      _AfterCreate: function () 
      {  
        this._super();
        this._registerSwipeHandler();
        this._isInitFetch = true;
      },
      /**
       * Sets up needed resources for paging control, for example, add
       * listeners.
       * @protected
       * @override
       * @memberof! oj.ojPagingControl
       */
      _SetupResources: function ()
      {
        this._super();
        this._registerResizeListener(this._getPagingControlContainer());
        this._registerDataSourceEventListeners();
        if (this._isInitFetch)
        {
          this._setInitialPage();
          this._isInitFetch = false;
        } 
        else
        {
          this._refresh();
        }
      },
      /**
       * Releases resources for paging control.
       * @protected
       * @override
       * @memberof! oj.ojPagingControl
       */
      _ReleaseResources: function ()
      {
        this._super();
        // unregister the listeners on the datasource
        this._unregisterDataSourceEventListeners();
        this._unregisterResizeListener();
      },
      /**
       * Called by component to add a busy state and return the resolve function
       * to call when the busy state can be removed.
       * @private
       */
      _addComponentBusyState: function(msg)
      {
        var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
        var options = {'description' : "The component identified by '" + this.element.attr('id') + "' " + msg};
        var resolver = busyContext.addBusyState(options);

        // Keep track of the busy state resolvers so that we can clean up later    
        if (!this._busyStateResolvers)
        {
          this._busyStateResolvers = [];
        }
        this._busyStateResolvers.push(resolver);

        return resolver;
      },
      /**
       * Called by component to remove a busy state.
       * @private
       */
      _removeComponentBusyState: function(resolver)
      {
        // Remove the busy state resolver from tracking
        if (this._busyStateResolvers)
        {
          var index = this._busyStateResolvers.indexOf(resolver);
          if (index >= 0)
          {
            this._busyStateResolvers.splice(index, 1);

            // Resolve the busy state only if it's found in the tracking list
            // to avoid busy state getting resolved twice.  e.g. some component
            // code may still be executed after all busy states are cleaned up
            // in _destroy.            
            resolver();
          }
        }
      },
      /**
       * @override
       * @private
       */
      _destroy: function()
      {
        this._unregisterDataSourceEventListeners();
        this._unregisterSwipeHandler();
        
        // Remove any pending busy states
        if (this._busyStateResolvers)
        {
          var resolver;
          while (resolver = this._busyStateResolvers.pop())
          {
            resolver();
          }
          this._busyStateResolvers = null;
        }
        
        // invalidate our refresh/fetch queue
        this._componentDestroyed = true;
      },
      /**
       * @override
       * @private
       */
      _draw: function()
      {
        var options = this.options;
        // add main css class to element
        this.element.addClass(this._CSS_CLASSES._PAGING_CONTROL_CLASS);
        this.element.addClass(this._MARKER_STYLE_CLASSES._WIDGET);

        this._createPagingControlAccLabel();
        this._createPagingControlContent();
        this._mode = options['mode'];
        
        if (options['mode'] == this._MODE._LOAD_MORE)
        {
          this._createPagingControlLoadMore();
          this._createPagingControlLoadMoreLink();
          this._createPagingControlLoadMoreRange();
        }
        else
        {
          this._createPagingControlNav();
        }
      },
      /**
       * @override
       * @private
       */
      _events:
        {
          /*
           * invoke loading next page of data
           */
          'mouseup .oj-pagingcontrol-loadmore-link': function(event)
          {
            this['loadNext']();
            // we need to listen to mouseup because click is not
            // reliable on MacOS X
            $(event.target).data('_mouseup', true);
            event.preventDefault();
          },
          /*
           * invoke loading next page of data
           */
          'click .oj-pagingcontrol-loadmore-link': function(event)
          {
            if ($(event.target).data('_mouseup'))
            {
              $(event.target).data('_mouseup', false);
            }
            else
            {
              this['loadNext']();
            }
            event.preventDefault();
          },
          /*
           * invoke loading page of data
           */
          'click .oj-pagingcontrol-nav-dot': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              var pageNum = $(event.currentTarget).attr('data-oj-pagenum');
              var self = this;
              this['page'](pageNum).then(
                function(result)
                {
                  self = null;
                  return;
                },
                function(err)
                {
                  var errSummary = self.getTranslatedString(self._BUNDLE_KEY._ERR_PAGE_INVALID_SUMMARY);
                  oj.Logger.error(errSummary + '\n' + err);
                  self = null;
                });  
            }
            event.preventDefault();
          },
          /*
           * invoke loading page of data
           */
          'click .oj-pagingcontrol-nav-page': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              var pageNum = $(event.currentTarget).attr('data-oj-pagenum');
              var self = this;
              this['page'](pageNum).then(
                function(result)
                {
                  self = null;
                  return;
                },
                function(err)
                {
                  var errSummary = self.getTranslatedString(self._BUNDLE_KEY._ERR_PAGE_INVALID_SUMMARY);
                  oj.Logger.error(errSummary + '\n' + err);
                  self = null;
                });  
            }
            event.preventDefault();
          },
          /*
           * invoke loading first page of data
           */
          'click .oj-pagingcontrol-nav-first': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              var self = this;
              this['firstPage']().then(
                function(result)
                {
                  self = null;
                  return;
                },
                function(err)
                {
                  var errSummary = self.getTranslatedString(self._BUNDLE_KEY._ERR_PAGE_INVALID_SUMMARY);
                  oj.Logger.error(errSummary + '\n' + err);
                  self = null;
                });  
            }
            event.preventDefault();
          },
          /*
           * invoke loading previous page of data
           */
          'click .oj-pagingcontrol-nav-previous': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              var self = this;
              this['previousPage']().then(
                function(result)
                {
                  self = null;
                  return;
                },
                function(err)
                {
                  var errSummary = self.getTranslatedString(self._BUNDLE_KEY._ERR_PAGE_INVALID_SUMMARY);
                  oj.Logger.error(errSummary + '\n' + err);
                  self = null;
                });  
            }
            event.preventDefault();
          },
          /*
           * invoke loading next page of data
           */
          'click .oj-pagingcontrol-nav-next': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              var self = this;
              this['nextPage']().then(
                function(result)
                {
                  self = null;
                  return;
                },
                function(err)
                {
                  var errSummary = self.getTranslatedString(self._BUNDLE_KEY._ERR_PAGE_INVALID_SUMMARY);
                  oj.Logger.error(errSummary + '\n' + err);
                  self = null;
                });  
            }
            event.preventDefault();
          },
          /*
           * invoke loading last page of data
           */
          'click .oj-pagingcontrol-nav-last': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              var self = this;
              this['lastPage']().then(
                function(result)
                {
                  self = null;
                  return;
                },
                function(err)
                {
                  var errSummary = self.getTranslatedString(self._BUNDLE_KEY._ERR_PAGE_INVALID_SUMMARY);
                  oj.Logger.error(errSummary + '\n' + err);
                  self = null;
                });  
            }
            event.preventDefault();
          },
          /*
           * prevent submission of form on enter
           */
          'keypress .oj-pagingcontrol-nav-input': function(event)
          {
            var keyCode = event.which;
            
            if (keyCode == 13)
            {
              event.preventDefault();
            }
          },
          /*
           * Add oj-active
           */
          'mousedown .oj-pagingcontrol-nav-first': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              $(event.target).addClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            }
            event.preventDefault();
          },
          /*
           * Add oj-active
           */
          'mousedown .oj-pagingcontrol-nav-previous': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              $(event.target).addClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            }
            event.preventDefault();
          },
          /*
           * Add oj-active
           */
          'mousedown .oj-pagingcontrol-nav-next': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              $(event.target).addClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            }
            event.preventDefault();
          },
          /*
           * Add oj-active
           */
          'mousedown .oj-pagingcontrol-nav-last': function(event)
          {
            if (!$(event.currentTarget).hasClass(this._MARKER_STYLE_CLASSES._DISABLED))
            {
              $(event.target).addClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            }
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseup .oj-pagingcontrol-nav-first': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseup .oj-pagingcontrol-nav-previous': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseup .oj-pagingcontrol-nav-next': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseup .oj-pagingcontrol-nav-last': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseleave .oj-pagingcontrol-nav-first': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseleave .oj-pagingcontrol-nav-previous': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseleave .oj-pagingcontrol-nav-next': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          },
          /*
           * Remove oj-active
           */
          'mouseleave .oj-pagingcontrol-nav-last': function(event)
          {
            $(event.target).removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            event.preventDefault();
          }
        },
      /**
       * @private
       */
      _refresh: function(resize)
      {
        if (this._data != this.options['data'])
        {
          this._clearCachedDataMetadata();
          this._setInitialPage();
        }
        
        var size = 0;
        var startIndex = 0;
        
        if (this._data != null)
        {
          startIndex = this._data.getStartItemIndex();
        }

        // if totalSize == 0 then size is just 0. If totalSize < 0 then
        // we have unknown row count, if totalSize > 0 then we have a known row count.
        // For non-zero unknown row count and non-zero known row count
        // we need to calculate the rows on the page.
        if (this._data != null && this._data.totalSize() != 0 && this._data.getEndItemIndex() >= 0)
        {
          // startIndex = 0 for loadMore mode
          size = this._data.getEndItemIndex() - startIndex + 1;
        }

        this._mode = this.options['mode'];
          
        if (this.options['mode'] == this._MODE._LOAD_MORE)
        {
          this._refreshPagingControlLoadMore(size, startIndex);
        }
        else
        {
          this._refreshPagingControlNav(size, startIndex);
        }
      },
      /**
       * @override
       * @private
       */
      _setOption: function(key, value)
      {
        this._superApply(arguments);
        this._invokeDataPage(0, true).then(
                function(result)
                {
                  return;
                },
                function(err)
                {
                  oj.Logger.error(err);
                });
                
        if (this.options['mode'] != this._MODE._LOAD_MORE && 
            key == 'pageOptions')
        {
          var pagingControlContent = this._getPagingControlContent();
          if (pagingControlContent != null)
          {
            this._unregisterChildStateListeners(pagingControlContent);
            this._unregisterSwipeHandler();
            pagingControlContent.empty();
          }
          this._clearCachedDomPagingControlNav();
          this._createPagingControlNav();
          this._registerSwipeHandler();
        }
        this._queueRefresh();
      },
      /**** end internal widget functions ****/

      /**** start internal functions ****/
      /**
       * Clear any cached data metadata
       * @private
       */
      _clearCachedDataMetadata: function()
      {
        if (this._data != null)
        {
          this._unregisterDataSourceEventListeners();
        }
        this._data = null;
      },
      /**
       * Clear cached range text DOM
       * @private
       */
      _clearCachedDomLoadMore: function()
      {
        this._cachedDomPagingControlLoadMore = null;
        this._cachedDomPagingControlLoadMoreLink = null;
        this._cachedDomPagingControlLoadMoreRange = null;
      },
      /**
       * Clear any cached DOM nav elements
       * @private
       */
      _clearCachedDomPagingControlNav: function()
      {
        this._cachedDomPagingControlNav = null;
        this._cachedDomPagingControlNavInput = null;
        this._cachedDomPagingControlNavInputSummary = null;
      },
      /**
       * Create navigation arrow element
       * @param {string} typeClass  CSS class for the arrow type
       * @param {string} iconClass  CSS class for the arrow icon
       * @param {string} tipKey  Bundle key for the tooltip text
       * @param {string} accLabelKey  Bundle key for the accessibility label text
       * @param {string} accLabelClass  CSS class for the accessibility label
       * @param {boolean} isVertical  Whether the paging control is vertical
       * @return {jQuery} The navigation arrow jQuery object
       * @private
       */
      _createNavArrow: function(typeClass, iconClass, tipKey, accLabelKey, accLabelClass, isVertical)
      {
        var navArrow = $(document.createElement('a'));

        // JAWS on Firefox does not read aria-disabled state on links, even 
        // though Chrome and IE do. Need to add a button role for it to be read.
        navArrow.attr('role', 'button');

        navArrow.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_ARROW_CLASS);
        navArrow.addClass(typeClass);
        navArrow.addClass(iconClass);
        navArrow.addClass(this._CSS_CLASSES._WIDGET_ICON_CLASS);
        navArrow.addClass(this._MARKER_STYLE_CLASSES._CLICKABLE_ICON);
        navArrow.addClass(this._MARKER_STYLE_CLASSES._DISABLED);

        // Use aria-disabled to specify the disabled state because there is
        // no disabled attribute on <a> element.
        navArrow.attr('aria-disabled', 'true');

        var navArrowTip = this.getTranslatedString(tipKey);
        this._AddHoverable(navArrow);
        this._focusable({'element': navArrow, 'applyHighlight': true});
        navArrow.attr('title', navArrowTip);
        navArrow.attr(this._TAB_INDEX, '0');
        navArrow.attr('href', '#');
        navArrow.attr('oncontextmenu', 'return false;');
        var accLabelText = this.getTranslatedString(accLabelKey);
        var accLabel = this._createAccLabelSpan(accLabelText, accLabelClass);
        navArrow.append(accLabel); //@HTMLUpdateOK
        if (isVertical)
        {
          navArrow.css('display', 'block');
        }
        return navArrow;
      },
      /**
       * Disable/enable navigation arrow element
       * @param {jQuery} navArrow  the navigation arrow jQuery object
       * @param {boolean} disable  true to disable; false to enable
       * @private
       */
      _disableNavArrow: function(navArrow, disable)
      {
        if (disable)
        {
          navArrow.addClass(this._MARKER_STYLE_CLASSES._DISABLED);
          navArrow.removeClass(this._MARKER_STYLE_CLASSES._ENABLED);
          navArrow.removeClass(this._MARKER_STYLE_CLASSES._FOCUS_HIGHLIGHT);
          navArrow.removeClass(this._MARKER_STYLE_CLASSES._FOCUS);
          navArrow.attr('aria-disabled', 'true');
          navArrow.attr('tabindex', '-1');
        }
        else
        {
          navArrow.addClass(this._MARKER_STYLE_CLASSES._ENABLED);
          navArrow.removeClass(this._MARKER_STYLE_CLASSES._DISABLED);
          navArrow.removeAttr('aria-disabled');
          navArrow.attr(this._TAB_INDEX, '0');
        }
      },
      /**
       * Return the current page
       * @return {number} Current page.
       * @throws {Error}
       * @private
       */
      _getCurrentPage: function()
      {
        var data = this._getData();
        var page = 0;
        if (data != null)
        {
          page = data.getPage();
        }
        return page;
      },
      /**
       * Return the datasource object defined for this paging control
       * @return {Object} Datasource object.
       * @throws {Error}
       * @private
       */
      _getData: function()
      {
        if (!this._data && this.options['data'] != null)
        {
          this._data = this.options['data'];
          this._dataMetadata = this.options['data'];
          // In case we get a delayed setting of the data property--check to rebind the listeners
          this._registerDataSourceEventListeners();
        }
        return this._data;
      },
      /**
       * Return the item range text span
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @return {Object} Item range text span element.
       * @throws {Error}
       * @private
       */
      _getItemRange: function(size, startIndex)
      {
        var pageFrom = startIndex >= 0 ? startIndex : 0;
        var itemRangeSpan = $(document.createElement('span'));
        var itemRangeCurrentSpan = $(document.createElement('span'));
        if (this.options['mode'] == this._MODE._LOAD_MORE)
        {
          itemRangeCurrentSpan.addClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_CURRENT_CLASS); 
        }
        else
        {
          itemRangeCurrentSpan.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CURRENT_CLASS); 
        }
        itemRangeSpan.append(itemRangeCurrentSpan); //@HTMLUpdateOK
        var data = this._getData();
        if (data != null && data.totalSize() != null && size != null)
        {
          var itemRangeCurrentText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_CURRENT_SINGLE, {'pageFrom': pageFrom});

          var pageTo = parseInt(startIndex, 10) + parseInt(size, 10);
          pageFrom = pageTo > 0 ? pageFrom + 1 : 0;

          if (data.totalSize() != -1)
          {
            pageTo = pageTo > data.totalSize() ? data.totalSize() : pageTo;
            if (pageFrom == pageTo)
            {
              itemRangeCurrentText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_CURRENT_SINGLE, {'pageFrom': pageFrom});
            }
            else if (pageFrom > pageTo)
            {
              // this can happen if we go from unknown to known row count and the last page has no rows. In that case don't return anything
              return itemRangeSpan;
            }
            else
            {
              itemRangeCurrentText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_CURRENT, {'pageFrom': pageFrom, 'pageTo': pageTo});
            }
            var itemRangeOfText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_OF);
            var itemRangeConf = null;

            if (data.totalSizeConfidence() == "atLeast")
            {
              itemRangeConf = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_ATLEAST);
            }
            else if (data.totalSizeConfidence() == "estimate")
            {
              itemRangeConf = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_APPROX);
            }

            var itemRangeOfSpan = $(document.createElement('span'));

            if (itemRangeConf == null)
            {
              itemRangeOfSpan.text(" " + itemRangeOfText + " ");
            }
            else
            {
              itemRangeOfSpan.text(" " + itemRangeConf + " ");
            }
            itemRangeSpan.append(itemRangeOfSpan); //@HTMLUpdateOK
            var itemRangeMaxSpan = $(document.createElement('span'));
            if (this.options['mode'] == this._MODE._LOAD_MORE)
            {
              itemRangeMaxSpan.addClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_MAX_CLASS); 
            }
            else
            {
              itemRangeMaxSpan.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_MAX_CLASS); 
            }
            itemRangeMaxSpan.text(data.totalSize());
            itemRangeSpan.append(itemRangeMaxSpan); //@HTMLUpdateOK
            var itemRangeItemsText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_ITEMS);
            var itemRangeItemsSpan = $(document.createElement('span'));
            itemRangeItemsSpan.text(" " + itemRangeItemsText);
            itemRangeSpan.append(itemRangeItemsSpan); //@HTMLUpdateOK
          }
          else
          {
            if (size == 0)
            {
              itemRangeCurrentText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_CURRENT_SINGLE, {'pageFrom': 0});
            }
            else
            {
              itemRangeCurrentText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_CURRENT, {'pageFrom': pageFrom, 'pageTo': pageTo});
            }
            var itemRangeItemsText = this.getTranslatedString(this._BUNDLE_KEY._MSG_ITEM_RANGE_ITEMS);
            var itemRangeItemsSpan = $(document.createElement('span'));
            itemRangeItemsSpan.text(" " + itemRangeItemsText);
            itemRangeSpan.append(itemRangeItemsSpan); //@HTMLUpdateOK
          }
          itemRangeCurrentSpan.text(itemRangeCurrentText);
        } 
        return itemRangeSpan;
      },
      /**
       * Return maximum number of page links
       * @return {number} Max page links.
       * @private
       */
      _getMaxPageLinks: function()
      {
        var maxPageLinks = this.options['pageOptions']['maxPageLinks'];
        
        if (this.options['pageOptions']['type'] == this._TYPE._DOTS)
        {
          maxPageLinks = Number.MAX_VALUE;
        }
        else if (!maxPageLinks)
        {
          maxPageLinks = this._PAGE_OPTION_DEFAULT_MAX_PAGE_LINKS;
        }
        
        return maxPageLinks;
      },
      /**
       * @param {number} size Number of rows
       * @private
       */
      _getMaxPageVal: function(size)
      {
        var maxPageVal = 0;
        
        if (this._getTotalPages() > 0 &&
            this._isTotalSizeConfidenceActual())
        {
          maxPageVal = this._getTotalPages();
        }
        else
        {
          if (size > 0)
          {
            var data = this._getData();
            if (data != null && (data.totalSizeConfidence() == "atLeast" ||
                                 data.totalSizeConfidence() == "estimate"))
            {
              maxPageVal = this._getTotalPages() + 1;
            }
            else
            {
              maxPageVal = this._getCurrentPage() + 2;
            }
          }
          else
          {
            maxPageVal = this._getCurrentPage() + 1;
          }
        }
        return maxPageVal;
      },
      /**
       * Return reject Promise
       * @return {Promise} promise object reject.
       * @private
       */
      _getRejectPromise: function()
      {
        return Promise.reject();
      },
      /**
       * Return the total number of pages
       * @return {number} Total pages.
       * @throws {Error}
       * @private
       */
      _getTotalPages: function()
      {
        var data = this._getData();
        var pageCount = 0;
        if (data != null)
        {
          pageCount = data.getPageCount();
        }
        // if pageCount is not valid then treat it as zero
        pageCount = pageCount >= -1 ? pageCount : 0;
        return pageCount;
      },
      /**
       * Callback handler for fetch completed in the datasource.
       * status message.
       * @param {Object} event 
       * @private
       */
      _handleDataFetchEnd: function(event)
      {
        // restore focus
        var activeElement = $(document.activeElement);

        if (activeElement.hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGE_CLASS))
        {
          var pageNum = parseInt(activeElement.attr('data-oj-pagenum'), 10);
          var self = this;
          setTimeout(function() 
          {
            if (pageNum >= 0)
            {
              // try to focus on the next page. If not the previous page
              var nextPageNum = pageNum + 1;
              var prevPageNum = pageNum - 1;
              var nextNavPage = self._getPagingControlContent().find('a[data-oj-pagenum=' + nextPageNum + ']');
              
              if (nextNavPage != null &&
                nextNavPage.length > 0)
              {
                nextNavPage.focus();
              }
              else
              {
                var prevNavPage = self._getPagingControlContent().find('a[data-oj-pagenum=' + prevPageNum + ']');
                
                if (prevNavPage != null &&
                  prevNavPage.length > 0)
                {
                  prevNavPage.focus();
                }
              }
            }
            pageNum = null;
            self = null;
          }, 100);
        }
        else if (activeElement.hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_CLASS))
        {
          var self = this;
          setTimeout(function() 
          {
            //since we are already in the first page, focus should be on next arrow.
            var nextArrow = self._getPagingControlContent().find('.' + self._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS);
            nextArrow.focus();
            self = null;
          }, 100);
        }
        else if (activeElement.hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS))
        {
          var self = this;
          setTimeout(function ()
          {
            var previousArrow = self._getPagingControlContent().find('.' + self._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS);
            if (!previousArrow.hasClass(self._MARKER_STYLE_CLASSES._DISABLED))
            {
              previousArrow.focus();
            }
            else
            {
              var nextArrow = self._getPagingControlContent().find('.' + self._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS);
              nextArrow.focus();
            }
            self = null;
          }, 100);
        }
        else if (activeElement.hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS))
        {
          var self = this;
          setTimeout(function ()
          {
            var nextArrow = self._getPagingControlContent().find('.' + self._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS);
            if (!nextArrow.hasClass(self._MARKER_STYLE_CLASSES._DISABLED))
            {
              nextArrow.focus();
            }
            else
            {
              var previousArrow = self._getPagingControlContent().find('.' + self._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS);
              previousArrow.focus();
            }
            self = null;
          }, 100);
        }
        else if (activeElement.hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_CLASS))
        {
          var self = this;
          setTimeout(function() 
          {
            var previousArrow = self._getPagingControlContent().find('.' + self._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS);
            previousArrow.focus();
            self = null;
          }, 100);
        }
        else if (activeElement.hasClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_CLASS))
        {
          var self = this;
          setTimeout(function() 
          {
            var navInput = self._getPagingControlContent().find('.' + self._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_CLASS);
            navInput.focus();
            self = null;
          }, 100);
        }
        this._queueRefresh();
      },
      /**
       * Callback handler for page event in the datasource.
       * status message.
       * @param {Object} event 
       * @private
       */
      _handleDataPage: function(event)
      {
        var page = event['page'];
        var previousPage = event['previousPage'];
        
        if (page != previousPage)
        {
          // only refresh if page has changed
          this._queueRefresh();
        }
      },
      /**
       * Callback handler for reset in the datasource.
       * status message.
       * @param {Object} event 
       * @private
       */
      _handleDataReset: function(event)
      {
        this._invokeDataPage(0, false).then(
                function(result)
                {
                  return;
                },
                function(err)
                {
                  oj.Logger.error(err);
                });    
      },
      /**
       * Callback handler for refresh in the datasource.
       * status message.
       * @param {Object} event 
       * @private
       */
      _handleDataRefresh: function(event)
      {
        this._queueRefresh();       
      },
      /**
       * Callback handler for sort in the datasource.
       * status message.
       * @param {Object} event 
       * @private
       */
      _handleDataSort: function(event)
      {
        // Do a reset if the paging mode is loadmore, otherwise just refresh
        if (this.options['mode'] == this._MODE._LOAD_MORE) 
        {
            this._handleDataReset(event);
        }
      },
      /**
       * Callback handler for row added into the datasource.
       * status message.
       * @param {Object} event
       * @private
       */
      _handleDataRowAdd: function(event)
      {
        if (this._isOperationOnCurrentPage(event))
        {
          if (this.options['mode'] == this._MODE._PAGE)
          {

              // this means that the add caused the pages to shift or
              // the row was added to the current page and the first page is full already
              // so we need to re-fetch the current page
              this._invokeDataPage(this._getCurrentPage(), true).then(
                function(result)
                {
                  return;
                },
                function(err)
                {
                  oj.Logger.error(err);
                });
              return;
          }
          else
          {
            this._invokeDataFetchCurrent();
            return;
          }
        }
        this._queueRefresh();
      },
      /**
       * Callback handler for row removed in the datasource.
       * status message.
       * @param {Object} event
       * @private
       */
      _handleDataRowRemove: function(event)
      {
        if (this.options['mode'] == this._MODE._PAGE)
        {
          if (this._getTotalPages() > 0 && this._getCurrentPage() > this._getTotalPages() - 1)
          {
            // if the number of pages decreased due to the removal, then
            // reset the page
            this._invokeDataPage(this._getTotalPages() - 1, true).then(
                function(result)
                {
                  return;
                },
                function(err)
                {
                  oj.Logger.error(err);
                });
            return;
          }
          else if (this._isOperationOnCurrentPage(event))
          {
            // this means that the remove caused the pages to shift or
            // the row was deleted from the current page
            // so we need to re-fetch the current page
            this._invokeDataPage(this._getCurrentPage(), true).then(
                function(result)
                {
                  return;
                },
                function(err)
                {
                  oj.Logger.error(err);
                });
            return;
          }
        }
        else if (this._isOperationOnCurrentPage(event))
        {
          this._invokeDataFetchCurrent();
          return;
        }
        this._queueRefresh();
      },
      /**
       * Callback handler for page change.
       * @param {Object} event
       * @private
       */
      _handlePageChange: function(event, data)
      {
        var option = data.option;
        
        if (option != 'value')
        {
          return;
        }
        
        var page = data.value;
        if (page != this._getCurrentPage() + 1 && !isNaN(page) && page > 0)
        {
          page = Math.round(page);
          var self = this;
          this['page'](page - 1).then(
              function(result)
              {
                self = null;
                return;
              },
              function(err)
              {
                var errSummary = self.getTranslatedString(self._BUNDLE_KEY._ERR_PAGE_INVALID_SUMMARY);
                oj.Logger.error(errSummary + '\n' + err);
                self = null;
              });  
        }
      },
      /**
       * Set page
       * @param {number} page Page
       * @param {boolean} async Asynchronous
       * @return (Promise} Page Promise 
       * @private
       */
      _invokeDataPage: function(page, async)
      {
        try
        {
          page = parseInt(page, 10);
        }
        catch (e)
        {
          return Promise.reject(e);
        }
        
        this._currentStartIndex = 0;
        this._resetPagingControlNavInput();
        
        if (async)
        {
          this._queuePageFetch(page);
          return Promise.resolve();
        }
        else
        {
          return this._invokeDataSetPage(page);
        }
      },
      /**
       * Set the current page in the datasource
       * @param {number} page Page
       * @return (Promise} Page Promise 
       * @private
       */
      _invokeDataSetPage: function(page)
      {
        var data = this._getData();
        var self = this;
        return new Promise(function(resolve, reject)
        {
          if (data != null)
          {
            var resolveBusyState = self._addComponentBusyState('is setting page.');
            data.setPage(page, {'pageSize' : self.options['pageSize']}).then(function(result)
            {
              self._removeComponentBusyState(resolveBusyState);
              self = null;
              resolveBusyState = null;
              resolve(null);
            }, function(error)
            {
              self._removeComponentBusyState(resolveBusyState);
              self = null;
              resolveBusyState = null;
              reject(error);
            });
            data = null;
          }
          else
          {
            self = null;
            resolve(null);
          }
        });
      },
      /**
       * Fetch the next set of rows
       * @return (Promise} Promise 
       * @private
       */
      _invokeDataFetchNext: function()
      {
        var pageSize = this.options['pageSize'];
        
        if (!this._currentStartIndex)
        {
          this._currentStartIndex = pageSize;
        }
        else
        {
          this._currentStartIndex = this._currentStartIndex + pageSize;
        }

        return this._invokeDataFetch({'startIndex': this._currentStartIndex, 'pageSize': pageSize});
      },
      /**
       * Fetch data rows using specified options
       * @param {Object} options Options to control fetch
       * @return (Promise} Promise 
       * @private
       */
      _invokeDataFetch: function(options)
      {
        var data = this._getData();

        if (!this._isTotalSizeConfidenceActual() || 
            (data.totalSize() > this._currentStartIndex && this._isTotalSizeConfidenceActual()))
        {
          var self = this;
          return new Promise(function(resolve, reject)
          {
            var resolveBusyState = self._addComponentBusyState('is fetching data.');
            data.fetch(options).then(function(result)
            {
              self._removeComponentBusyState(resolveBusyState);
              self = null;
              resolveBusyState = null;
              resolve(result);
            }, function (error)
            {
              self._removeComponentBusyState(resolveBusyState);
              self = null;
              resolveBusyState = null;
              reject(null);  
            });
            data = null;
          });
        }
        return Promise.resolve();
      },
      /**
       * Fetch the current set of rows in loadMore
       * @return (Promise} Promise 
       * @private
       */
      _invokeDataFetchCurrent: function()
      {
        var pageSize = this.options['pageSize'];

        return this._invokeDataFetch({'startIndex': 0, 'pageSize': this._currentStartIndex + pageSize});
      },
      /**
       * Check if the rowIdx is for the current page
       * @return {boolean} true or false
       * @private
       */
      _isOperationOnCurrentPage: function(event)
      {
        if (event == null)
        {
          return false;
        }
        
        var data = this._getData();
        var startIndex = data.getStartItemIndex();
        if (this.options['mode'] == this._MODE._LOAD_MORE)
        {
          startIndex = 0;
        }
          
        var endIndex = data.getEndItemIndex();
        var rowIdx;
        
        // check if one or more
        if (event['index'] != null)
        {
          rowIdx = event['index'];
          
          if (rowIdx >= startIndex && rowIdx <= endIndex)
          {
            return true;
          }
        }
        else if (event['indexes'] != null)
        {
          var i;
          for (i = 0; i < event['indexes'].length; i++)
          {
            rowIdx = event['indexes'][i];

            if (rowIdx >= startIndex && rowIdx <= endIndex)
            {
              return true;
            }
          }
        }
        return false;
      },
      /**
       * Check if the totalSize confidence is "actual"
       * @return {boolean} true or false
       * @private
       */
      _isTotalSizeConfidenceActual: function()
      {
        var data = this._getData();
        
        if (data != null && data.totalSizeConfidence() == "actual")
        {
          return true;
        }
        
        return false;
      },
      _queuePageFetch: function(page)
      {
        var self = this;
        if (!this._pendingPageFetch)
        {
          this._pageFetchCount = 0;
          this._pendingPageFetch = Promise.resolve();
        }
        this._pageFetchCount++;
        // keep track of the latest page. We only do a fetch on the latest page.
        this._pageFetchLatestPage = page;
        this._pendingPageFetch = this._pendingPageFetch
        .then(function()
        {
          self._pageFetchCount--;
          if (self._pageFetchCount == 0 && !self._componentDestroyed)
          {
            self._pendingPageFetch = undefined;
            self._invokeDataSetPage(self._pageFetchLatestPage).then(
              function()
              {
                self = null;
              },
              function(error)
              {
                self._pageFetchCount--;
                if (self._pageFetchCount <= 0)
                {
                  self._pendingPageFetch = undefined;
                  oj.Logger.error(error);
                  self = null;
                }
              }
            );
          }
        },
        function(error)
        {
          self._pageFetchCount--;
          if (self._pageFetchCount <= 0)
          {
            self._pendingPageFetch = undefined;
            oj.Logger.error(error);
            self = null;
          }
        });
      },
      _queueRefresh: function()
      {
        var self = this;
        if (!this._pendingRefreshes)
        {
          this._refreshCount = 0;
          this._pendingRefreshes = Promise.resolve();
          this._setComponentNotReady();
        }
        this._refreshCount++;
        this._pendingRefreshes = this._pendingRefreshes
        .then(function()
        {
          self._refreshCount--;
          if (self._refreshCount == 0)
          {
            self._pendingRefreshes = undefined;
            if (!self._componentDestroyed)
            {
              self._refresh();
              self._trigger('ready');
            }
            self._setComponentReady();
            self = null;
          }
        },
        function(error)
        {
          self._refreshCount--;
          if (self._refreshCount == 0)
          {
            self._pendingRefreshes = undefined;
            oj.Logger.error(error);
            self._setComponentReady();
            self = null;
          }
        });
      },
      /**
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @private
       */
      _refreshPagingControlLoadMore: function(size, startIndex)
      {
        var data = this._getData();
        var pagingControlLoadMore = this._getPagingControlLoadMore();
        var rowCount = startIndex + size;
        var needLoadMore = !(data != null && ((rowCount == data.totalSize() && this._isTotalSizeConfidenceActual()) || data.totalSize() == 0));

        // Recreate loadMore control if this is the first time or there is more data
        if (!pagingControlLoadMore || needLoadMore)
        {          
          var pagingControlContent = this._getPagingControlContent();
          if (pagingControlContent != null)
          {
            pagingControlContent.empty();
          }
          this._clearCachedDomLoadMore();
          pagingControlLoadMore = this._createPagingControlLoadMore();
          rowCount = -1;
          
          if (size != null)
          {
            rowCount = startIndex + size;
          }
          
          if (rowCount < 0 || rowCount < this.options['loadMoreOptions']['maxCount'])
          {
            this._createPagingControlLoadMoreLink();
            this._createPagingControlLoadMoreRange(size, startIndex);
          }
          else
          {
            this._createPagingControlLoadMoreMaxRows();
          }
        }
        
        // hide loadMore if there are no more rows to fetch
        if (!needLoadMore)
        {
          pagingControlLoadMore.css('display', 'none');
        }
      },
      /**
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @private
       */
      _refreshPagingControlNav: function(size, startIndex)
      {
        var overflowOption = this.options['overflow'];
        this._refreshPagingControlNavMaxPageVal(size, startIndex);
        this._refreshPagingControlNavLabel();
        this._refreshPagingControlNavInput();
        this._refreshPagingControlNavSummaryLabel(size, startIndex);
        this._refreshPagingControlNavPages(size, startIndex);
        this._refreshPagingControlNavArrows(size, startIndex);
        
        if (overflowOption == 'fit')
        {
          // dynamically hide controls based on available width
          var elementWidth = this.element.width();
          var pagingControlNavArrowSection = this._getPagingControlNavArrowSection();
          var pagingControlNavInputSection = this._getPagingControlNavInputSection();
          var pagingControlNavPageLinks = this._getPagingControlNavPageLinks();
          var pagingControlNavInputSummary = this._getPagingControlNavInputSummary();
          var pagingControlNavArrowSectionWidth = pagingControlNavArrowSection != null ? pagingControlNavArrowSection[0].offsetWidth:0;
          var pagingControlNavInputSectionWidth = pagingControlNavInputSection != null ? pagingControlNavInputSection[0].offsetWidth:0;
          var pagingControlNavPageLinksWidth = pagingControlNavPageLinks != null ? pagingControlNavPageLinks.width():0;
          var pagingControlNavInputSummaryWidth = pagingControlNavInputSummary != null ? pagingControlNavInputSummary.width():0;
          var pagingControlNavWidth = pagingControlNavArrowSectionWidth + pagingControlNavInputSectionWidth + pagingControlNavInputSummaryWidth;
          
          if (pagingControlNavWidth > elementWidth)
          {
            if (pagingControlNavWidth 
                - pagingControlNavPageLinksWidth <= elementWidth)
            {
              // hide only the page links
              pagingControlNavPageLinks.css('display', 'none'); 
            }
            else if (pagingControlNavWidth 
                     - pagingControlNavPageLinksWidth 
                     - pagingControlNavInputSummaryWidth <= elementWidth)
            {
              // hide the range text too
              if (pagingControlNavPageLinks != null)
              {
                pagingControlNavPageLinks.css('display', 'none');
              }
              if (pagingControlNavInputSummary != null)
              {
                pagingControlNavInputSummary.css('display', 'none');
              }
            }
            else
            {
              // hide the arrows too
              if (pagingControlNavPageLinks != null)
              {
                pagingControlNavPageLinks.css('display', 'none');
              }
              if (pagingControlNavInputSummary != null)
              {
                pagingControlNavInputSummary.css('display', 'none');
              }
              if (pagingControlNavArrowSection != null)
              {
                pagingControlNavArrowSection.css('display', 'none'); 
              }
            }
          }
          else if (pagingControlNavWidth > 0)
          {
            if (pagingControlNavPageLinks != null)
            {
              pagingControlNavPageLinks.css('display', '');
            }
            if (pagingControlNavInputSummary != null)
            {
              pagingControlNavInputSummary.css('display', '');
            }
            if (pagingControlNavArrowSection != null)
            {
              pagingControlNavArrowSection.css('display', '');
            }
          }
        }
      },
      /**
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @private
       */
      _refreshPagingControlNavArrows: function(size, startIndex)
      {
        var pageSize = this.options['pageSize'];
        var pagingControlNavArrowSection = this._getPagingControlNavArrowSection();
        var pagingControlNavFirst = pagingControlNavArrowSection.children('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_CLASS);
        if (pagingControlNavFirst && pagingControlNavFirst.length > 0)
        {
          pagingControlNavFirst = $(pagingControlNavFirst[0]);
          var navFirstPageTip = this.getTranslatedString(this._BUNDLE_KEY._TIP_NAV_FIRST_PAGE);
          pagingControlNavFirst.attr('title', navFirstPageTip);
          
          this._disableNavArrow(pagingControlNavFirst, this._getCurrentPage() == 0);
        }
        var pagingControlNavPrevious = pagingControlNavArrowSection.children('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS);
        if (pagingControlNavPrevious && pagingControlNavPrevious.length > 0)
        {
          pagingControlNavPrevious = $(pagingControlNavPrevious[0]);
          var navPreviousPageTip = this.getTranslatedString(this._BUNDLE_KEY._TIP_NAV_PREVIOUS_PAGE);
          pagingControlNavPrevious.attr('title', navPreviousPageTip);
          
          this._disableNavArrow(pagingControlNavPrevious, this._getCurrentPage() == 0);
        }
        var pagingControlNavLast = pagingControlNavArrowSection.children('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_CLASS);
        if (pagingControlNavLast && pagingControlNavLast.length > 0)
        {
          pagingControlNavLast = $(pagingControlNavLast[0]);
          var navLastPageTip = this.getTranslatedString(this._BUNDLE_KEY._TIP_NAV_LAST_PAGE);
          pagingControlNavLast.attr('title', navLastPageTip);
          
          this._disableNavArrow(pagingControlNavLast, this._getCurrentPage() == this._getTotalPages() - 1 || 
                                                      this._getTotalPages() <= 0 ||
                                                      !this._isTotalSizeConfidenceActual());
        }
        var pagingControlNavNext = pagingControlNavArrowSection.children('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS);
        if (pagingControlNavNext && pagingControlNavNext.length > 0)
        {
          pagingControlNavNext = $(pagingControlNavNext[0]);
          var navNextPageTip = this.getTranslatedString(this._BUNDLE_KEY._TIP_NAV_NEXT_PAGE);
          pagingControlNavNext.attr('title', navNextPageTip);
          
          this._disableNavArrow(pagingControlNavNext, (this._getCurrentPage() == this._getTotalPages() - 1 && this._isTotalSizeConfidenceActual()) || 
                                                      this._getTotalPages() === 0 ||
                                                      (this._getTotalPages() < 0 && size === 0) ||
                                                      (this._getTotalPages() < 0 && size < pageSize));
        }
      },
      /**
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @private
       */
      _refreshPagingControlNavPages: function(size, startIndex)
      {
        var pagingControlNavPagesSection = this._getPagingControlNav().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGES_SECTION_CLASS);
        
        if (pagingControlNavPagesSection != null && pagingControlNavPagesSection.length > 0)
        {
          pagingControlNavPagesSection = $(pagingControlNavPagesSection.get(0));
          this._unregisterChildStateListeners(pagingControlNavPagesSection);
          pagingControlNavPagesSection.empty();
          this._createPagingControlNavPages(pagingControlNavPagesSection, this._getMaxPageLinks(), size, startIndex);
        }
      },
      /**
       * @private
       */
      _refreshPagingControlNavLabel: function()
      {
        var pagingControlNavInputSection = this._getPagingControlNavInputSection();
        
        if (pagingControlNavInputSection != null)
        {
          var pagingControlNavLabel = pagingControlNavInputSection.children('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_LABEL_CLASS);

          if (pagingControlNavLabel != null && pagingControlNavLabel.length > 0)
          {
            pagingControlNavLabel = $(pagingControlNavLabel[0]);
            var navInputPageLabel = this.getTranslatedString(this._BUNDLE_KEY._LABEL_NAV_INPUT_PAGE);
            pagingControlNavLabel.text(navInputPageLabel);
          }
        }
      },
      /**
       * @private
       */
      _refreshPagingControlNavInput: function()
      {
        var pagingControlNavInput = this._getPagingControlNavInput();
        
        if (pagingControlNavInput != null)
        {
          var navInputPageTip = this.getTranslatedString(this._BUNDLE_KEY._TIP_NAV_INPUT_PAGE);
          pagingControlNavInput.attr('title', navInputPageTip);
          pagingControlNavInput.ojInputText('option', 'title', navInputPageTip);
        }
      },
      /**
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @private
       */
      _refreshPagingControlNavSummaryLabel: function(size, startIndex)
      {
        var pagingControlNavSummaryLabel = this._getPagingControlNav().children('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CLASS);
        
        if (pagingControlNavSummaryLabel != null && pagingControlNavSummaryLabel.length > 0)
        {
          var itemRange = this._getItemRange(size, startIndex);
          pagingControlNavSummaryLabel = $(pagingControlNavSummaryLabel.get(0));
          pagingControlNavSummaryLabel.empty();
          
          if (itemRange.text().length > 0)
          {
            pagingControlNavSummaryLabel.append("("); //@HTMLUpdateOK
            pagingControlNavSummaryLabel.append(itemRange); //@HTMLUpdateOK
            pagingControlNavSummaryLabel.append(")"); //@HTMLUpdateOK
          }
        }
      },
      /**
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @private
       */
      _refreshPagingControlNavMaxPageVal: function(size, startIndex)
      {
        var maxPageVal = this._getMaxPageVal(size);
        
        var pagingControlNavMaxLabel = this._getPagingControlNav().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_MAX_CLASS);

        if (pagingControlNavMaxLabel != null && pagingControlNavMaxLabel.length > 0)
        {
          pagingControlNavMaxLabel = $(pagingControlNavMaxLabel.get(0));
          
          if (this._getTotalPages() > 0 &&
            this._isTotalSizeConfidenceActual())
          {
            var navInputPageMaxLabel = this.getTranslatedString(this._BUNDLE_KEY._LABEL_NAV_INPUT_PAGE_MAX, {'pageMax': maxPageVal});
            pagingControlNavMaxLabel.text(navInputPageMaxLabel);
          }
          else
          {
            pagingControlNavMaxLabel.empty();
          }
        }
        else
        {
          var pagingControlNavInputSection = this._getPagingControlNavInputSection();
          
          if (pagingControlNavInputSection != null && 
            this._getTotalPages() > 0 &&
            this._isTotalSizeConfidenceActual() &&
            maxPageVal >= this._getCurrentPage() + 1)
          {
            pagingControlNavMaxLabel = $(document.createElement('span'));
            pagingControlNavMaxLabel.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_MAX_CLASS);
            pagingControlNavMaxLabel.addClass('oj-label-inline');
            pagingControlNavInputSection.append(pagingControlNavMaxLabel); //@HTMLUpdateOK
            var navInputPageMaxLabel = this.getTranslatedString(this._BUNDLE_KEY._LABEL_NAV_INPUT_PAGE_MAX, {'pageMax': maxPageVal});
            pagingControlNavMaxLabel.text(navInputPageMaxLabel);
          }
        }
          
        var pagingControlNavInput = this._getPagingControlNav().find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_CLASS);
        if (pagingControlNavInput != null && pagingControlNavInput.length > 0)
        {
          pagingControlNavInput = $(pagingControlNavInput.get(0));
          pagingControlNavInput.ojInputText();
          pagingControlNavInput.ojInputText('option', 'validators', [{'type': 'numberRange', 'options': {'min': 1, max: maxPageVal}}]);
          var messagesShown = pagingControlNavInput.ojInputText('option', 'messagesShown');
          
          if (messagesShown == null || messagesShown.length == 0)
          {
            // only reset the value to the current page if there is currently no validation message displayed
            this._resetPagingControlNavInput();
          }
          
          if (maxPageVal == 1)
          {
            // make readOnly if we only have one page
            pagingControlNavInput.ojInputText('option', 'readOnly', true);
          }
          else
          {
            pagingControlNavInput.ojInputText('option', 'readOnly', false);
          }
        }
      },
      /**
       * Register event listeners which need to be registered datasource. 
       * @private
       */
      _registerDataSourceEventListeners: function()
      {
        // register the listeners on the datasource
        var data = this._getData();
        if (data != null)
        {
          this._unregisterDataSourceEventListeners();
          
          this._dataSourceEventHandlers = [];
          this._dataSourceEventHandlers.push({'eventType': oj.PagingModel.EventType['PAGE'], 'eventHandler': this._handleDataPage.bind(this)});
          this._dataSourceEventHandlers.push({'eventType': oj.PagingModel.EventType['PAGECOUNT'], 'eventHandler': this._handleDataRefresh.bind(this)});
          this._dataSourceEventHandlers.push({'eventType': oj.PagingTableDataSource.EventType['ADD'], 'eventHandler': this._handleDataRowAdd.bind(this)});
          this._dataSourceEventHandlers.push({'eventType': oj.PagingTableDataSource.EventType['REMOVE'], 'eventHandler': this._handleDataRowRemove.bind(this)});
          this._dataSourceEventHandlers.push({'eventType': oj.PagingTableDataSource.EventType['RESET'], 'eventHandler': this._handleDataReset.bind(this)});
          this._dataSourceEventHandlers.push({'eventType': oj.PagingTableDataSource.EventType['REFRESH'], 'eventHandler': this._handleDataRefresh.bind(this)});
          this._dataSourceEventHandlers.push({'eventType': oj.PagingTableDataSource.EventType['SYNC'], 'eventHandler': this._handleDataFetchEnd.bind(this)});
          this._dataSourceEventHandlers.push({'eventType': oj.PagingTableDataSource.EventType['SORT'], 'eventHandler': this._handleDataSort.bind(this)});

          var i, ev;
          for (i = 0; i < this._dataSourceEventHandlers.length; i++) {
            ev = data.on(this._dataSourceEventHandlers[i]['eventType'], this._dataSourceEventHandlers[i]['eventHandler']);
            if (ev) {
                this._dataSourceEventHandlers[i]['eventHandler'] = ev;
            }
            
        }
        }
      },
      /**
       * Register event listeners for resize the container DOM element.
       * @param {jQuery} element  DOM element
       * @private
       */
      _registerResizeListener: function(element)
      { 
        if (!this._resizeListener)
        {
          var self = this;
          this._resizeListener = function (width, height) 
          {
            self._queueRefresh();
          };
        }
        if (!this._resizeListenerElement)
        {
          oj.DomUtils.addResizeListener(element[0], this._resizeListener, 50);
          this._resizeListenerElement = element;
        }
      },
      /**
       * Register swipe handler for DOM element.
       * @private
       */
      _registerSwipeHandler: function()
      {
        if (oj.DomUtils.isTouchSupported())
        {
          if (this.options['mode'] == this._MODE._PAGE)
          {
            var pagingControlNav = this._getPagingControlNav();

            if (pagingControlNav != null)
            {
              var isVertical = this.options['pageOptions']['orientation'] == 'vertical' ? true : false;
              var self = this;

              if (isVertical)
              {
                var options = {
                  'recognizers': [
                    [Hammer.Swipe, {'direction': Hammer['DIRECTION_VERTICAL']}]
                ]};

                this._hammerSwipeUp = pagingControlNav.ojHammer(options).on('swipeup', function(event)
                {
                  event.preventDefault();
                  self['nextPage']();
                });

                this._hammerSwipeDown = pagingControlNav.ojHammer(options).on('swipedown', function(event)
                {
                  event.preventDefault();
                  self['previousPage']();
                });
              }
              else
              {
                var options = {
                  'recognizers': [
                    [Hammer.Swipe, {'direction': Hammer['DIRECTION_HORIZONTAL']}]
                ]};

                this._hammerSwipeLeft = pagingControlNav.ojHammer(options).on('swipeleft', function(event)
                {
                  event.preventDefault();
                  self['nextPage']();
                });

                this._hammerSwipeRight = pagingControlNav.ojHammer(options).on('swiperight', function(event)
                {
                  event.preventDefault();
                  self['previousPage']();
                });
              }
            }
          }
        }
      },
      /**
       * @private
       */
      _resetPagingControlNavInput: function()
      {
        var pagingControlNavInput = this._getPagingControlNavInput();
        
        if (pagingControlNavInput != null && pagingControlNavInput.hasClass('oj-component-initnode'))
        {
          try
          {
            pagingControlNavInput.ojInputText('option', 'value', this._getCurrentPage() + 1);
          }
          catch(err)
          {};
        }
      },
      /**
       * Called by component to declare rendering is not finished. This method currently 
       * handles the ready state for the component page level BusyContext
       * @private
       */
      _setComponentNotReady : function() 
      {
        // For page level BusyContext
        // If we've already registered a busy state with the page's busy context, don't need to do anything further
        if (!this._readyResolveFunc) 
        {
          this._readyResolveFunc = this._addComponentBusyState('is being loaded.');
        }
      },
      /**
       * Called by component to declare rendering is finished. This method currently
       * handles the page level BusyContext.
       * @private
       */
      _setComponentReady : function() 
      {
        if (this._readyResolveFunc) 
        {
          this._removeComponentBusyState(this._readyResolveFunc);
          this._readyResolveFunc = null;
        }
      },
      /**
       * Set the initial page.
       * @private
       */
      _setInitialPage: function()
      {
        var currentPage = this._getCurrentPage();
        
        if (currentPage > 0)
        {
          this._invokeDataPage(currentPage, true).then(
                function(result)
                {
                  return;
                },
                function(err)
                {
                  oj.Logger.error(err);
                });
        }
        else
        {
          this._invokeDataPage(0, true).then(
                function(result)
                {
                  return;
                },
                function(err)
                {
                  oj.Logger.error(err);
                });
        }
      },
      /**
       * Unregister event listeners which are registered on datasource. 
       * @private
       */
      _unregisterDataSourceEventListeners: function()
      {
        var data = this._getData();
        // unregister the listeners on the datasource
        if (this._dataSourceEventHandlers != null && data != null)
        {
          var i;
          for (i = 0; i < this._dataSourceEventHandlers.length; i++)
            data.off(this._dataSourceEventHandlers[i]['eventType'], this._dataSourceEventHandlers[i]['eventHandler']);
        }
      },
      /**
       * Unregister _focusable(), etc, which were added to the child elements
       * @param {jQuery} parent jQuery div DOM element
       * @private
       */
      _unregisterChildStateListeners: function(parent)
      {
        var self = this;
        parent.find('*').each(function()
        {
          self._UnregisterChildNode(this);
        });
        self = null;
      },
      /**
       * Unregister event listeners for resize the container DOM element.
       * @private
       */
      _unregisterResizeListener: function()
      {
        if (this._resizeListenerElement != null)
        {
          oj.DomUtils.removeResizeListener(this._resizeListenerElement, this._resizeListener);
          this._resizeListenerElement = null;
        }
      },
      /**
       * Unregister swipe handler for DOM element.
       * @private
       */
      _unregisterSwipeHandler: function()
      {
        if (oj.DomUtils.isTouchSupported())
        {
          var isVertical = this.options['pageOptions']['orientation'] == 'vertical' ? true : false;
 
          if (isVertical)
          {
            if (this._hammerSwipeUp != null)
            {
              this._hammerSwipeUp.off('swipeup');
              this._hammerSwipeUp = null;
            }
            if (this._hammerSwipeDown != null)
            {
              this._hammerSwipeDown.off('swipedown');
              this._hammerSwipeDown = null;
            }
          }
          else
          {
            if (this._hammerSwipeLeft != null)
            {
              this._hammerSwipeLeft.off('swipeleft');
              this._hammerSwipeLeft = null;
            }
            if (this._hammerSwipeRight != null)
            {
              this._hammerSwipeRight.off('swiperight');
              this._hammerSwipeRight = null;
            }
          }
        }
      },
      /**** end internal functions ****/
      /**
       * Create a span element for acc purposes
       * @param {string} text span text
       * @param {string} className css class
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createAccLabelSpan: function(text, className)
      {
        var accLabel = $(document.createElement('span'));
        accLabel.addClass(className);
        accLabel.addClass(this._CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
        accLabel.text(text);

        return accLabel;
      },
      /**** start internal DOM functions ****/
      /**
       * Create the acc paging control label
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createPagingControlAccLabel: function()
      {
        var pagingControlContainer = this._getPagingControlContainer();
        var pagingControlAccLabelText = this.getTranslatedString(this._BUNDLE_KEY._LABEL_ACC_PAGING);
        var pagingControlAccLabel = this._createAccLabelSpan(pagingControlAccLabelText, this._CSS_CLASSES._PAGING_CONTROL_ACC_LABEL_CLASS);
        var pagingControlAccLabelId = this.element.attr('id') + '_oj_pgCtrl_acc_label';
        pagingControlAccLabel.attr('id', pagingControlAccLabelId);
        pagingControlContainer.append(pagingControlAccLabel); //@HTMLUpdateOK

        return pagingControlAccLabel;
      },
      /**
       * Create the acc page link label
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createPagingControlAccNavPageLabel: function()
      {
        var pagingControlAccNavPageLabelText = this.getTranslatedString(this._BUNDLE_KEY._LABEL_ACC_NAV_PAGE);
        var pagingControlAccNavPageLabel = this._createAccLabelSpan(pagingControlAccNavPageLabelText, this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGE_ACC_LABEL_CLASS);

        return pagingControlAccNavPageLabel;
      },
      /**
       * Create an paging content div
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createPagingControlContent: function()
      {
        var pagingControlContainer = this._getPagingControlContainer();
        var pagingControlContent = $(document.createElement('div'));
        pagingControlContent.addClass(this._CSS_CLASSES._PAGING_CONTROL_CONTENT_CLASS);
        var pagingControlAccLabelId = this._getPagingControlAccLabel().attr('id');
        pagingControlContent.attr('role', 'navigation');
        pagingControlContent.attr('aria-labelledby', pagingControlAccLabelId);
        pagingControlContainer.append(pagingControlContent); //@HTMLUpdateOK

        return pagingControlContent;
      },
      /**
       * Create an paging load more div
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createPagingControlLoadMore: function()
      {
        var pagingControlContent = this._getPagingControlContent();
        var pagingControlLoadMore = $(document.createElement('div'));
        pagingControlLoadMore.addClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_CLASS);
        pagingControlContent.append(pagingControlLoadMore); //@HTMLUpdateOK

        return pagingControlLoadMore;
      },
      /**
       * Create an paging load more link
       * @return {jQuery} jQuery a DOM element
       * @private
       */
      _createPagingControlLoadMoreLink: function()
      {
        var pagingControlLoadMore = this._getPagingControlLoadMore();
        var pagingControlLoadMoreLink = $(document.createElement('a'));
        pagingControlLoadMoreLink.addClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_LINK_CLASS);
        var loadMoreText = this.getTranslatedString(this._BUNDLE_KEY._LABEL_LOAD_MORE);
        pagingControlLoadMoreLink.text(loadMoreText);
        pagingControlLoadMoreLink.attr(this._TAB_INDEX, '0');
        pagingControlLoadMoreLink.attr('href', '#');
        pagingControlLoadMore.append(pagingControlLoadMoreLink); //@HTMLUpdateOK

        return pagingControlLoadMoreLink;
      },
      /**
       * Create an paging load more max page message
       * @return {jQuery} jQuery a DOM element
       * @private
       */
      _createPagingControlLoadMoreMaxRows: function()
      {
        var pagingControlLoadMore = this._getPagingControlLoadMore();
        var pagingControlLoadMoreMaxRows = $(document.createElement('span'));
        pagingControlLoadMoreMaxRows.addClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_MAX_ROWS_CLASS);
        var maxRowsText = this.getTranslatedString(this._BUNDLE_KEY._LABEL_LOAD_MORE_MAX_ROWS, {'maxRows': this.options['loadMoreOptions']['maxCount']});
        pagingControlLoadMoreMaxRows.text(maxRowsText);
        pagingControlLoadMore.append(pagingControlLoadMoreMaxRows); //@HTMLUpdateOK

        return pagingControlLoadMoreMaxRows;
      },
      /**
       * Create an paging load more link
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @return {jQuery} jQuery a DOM element
       * @private
       */
      _createPagingControlLoadMoreRange: function(size, startIndex)
      {
        var pagingControlLoadMore = this._getPagingControlLoadMore();
        var pagingControlLoadMoreRange = $(document.createElement('span'));
        pagingControlLoadMoreRange.addClass(this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_CLASS);
        var loadMoreRange = this._getItemRange(size, startIndex);
        pagingControlLoadMoreRange.append(loadMoreRange); //@HTMLUpdateOK
        pagingControlLoadMore.append(pagingControlLoadMoreRange); //@HTMLUpdateOK

        return pagingControlLoadMoreRange;
      },
      /**
       * Create the paging nav bar div
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createPagingControlNav: function(size, startIndex)
      {
        var options = this.options;
        var isVertical = this.options['pageOptions']['orientation'] == 'vertical' ? true : false;
        var isDot = this.options['pageOptions']['type'] == 'dots' ? true : false;
        var pageOptionLayout = options['pageOptions']['layout'];
        if (pageOptionLayout == null)
        {
          pageOptionLayout = [this._PAGE_OPTION_LAYOUT._AUTO];
        }
        var pagingControlContent = this._getPagingControlContent();
        var pagingControlNav = $(document.createElement('div'));
        pagingControlNav.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_CLASS);
        pagingControlContent.append(pagingControlNav); //@HTMLUpdateOK

        // page input section
        if (($.inArray(this._PAGE_OPTION_LAYOUT._AUTO, pageOptionLayout) != -1 && !isDot)||
            $.inArray(this._PAGE_OPTION_LAYOUT._ALL, pageOptionLayout) != -1 ||
            $.inArray(this._PAGE_OPTION_LAYOUT._INPUT, pageOptionLayout) != -1)
        {
          var pagingControlNavInputSection = $(document.createElement('div'));
          pagingControlNavInputSection.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SECTION_CLASS);
          pagingControlNav.append(pagingControlNavInputSection); //@HTMLUpdateOK
          var pagingControlNavLabel = $(document.createElement('label'));
          pagingControlNavLabel.attr('for', this.element.attr('id') + '_nav_input');
          pagingControlNavLabel.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_LABEL_CLASS);
          pagingControlNavLabel.addClass('oj-label-inline');
          var navInputPageLabel = this.getTranslatedString(this._BUNDLE_KEY._LABEL_NAV_INPUT_PAGE);
          pagingControlNavLabel.text(navInputPageLabel);
          pagingControlNavInputSection.append(pagingControlNavLabel); //@HTMLUpdateOK

          var pagingControlNavInput = $(document.createElement('input'));
          pagingControlNavInput.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_CLASS);
          var navInputPageTip = this.getTranslatedString(this._BUNDLE_KEY._TIP_NAV_INPUT_PAGE);
          pagingControlNavInput.attr('id', this.element.attr('id') + '_nav_input');
          pagingControlNavInput.attr('title', navInputPageTip);
          pagingControlNavInput.attr(this._TAB_INDEX, '0');
          pagingControlNavInput.val(this._getCurrentPage() + 1);
          pagingControlNavInputSection.append(pagingControlNavInput); //@HTMLUpdateOK
          var maxPageVal = this._getMaxPageVal(size);
          
          if (this._getTotalPages() > 0 &&
              this._isTotalSizeConfidenceActual())
          {
            var pagingControlNavMaxLabel = $(document.createElement('span'));
            pagingControlNavMaxLabel.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_MAX_CLASS);
            pagingControlNavMaxLabel.addClass('oj-label-inline');
            var navInputPageMaxLabel = this.getTranslatedString(this._BUNDLE_KEY._LABEL_NAV_INPUT_PAGE_MAX, {'pageMax': maxPageVal});
            pagingControlNavMaxLabel.text(navInputPageMaxLabel);
            pagingControlNavInputSection.append(pagingControlNavMaxLabel); //@HTMLUpdateOK
          }
          pagingControlNavInput.ojInputText({'displayOptions': {'messages': ['notewindow'], 'converterHint': ['notewindow'],  'validatorHint': ['notewindow']}, 'rootAttributes': {'style':"width: auto; min-width: 0;"}, 'optionChange': this._handlePageChange.bind(this), 'validators': [{'type': 'numberRange', 'options': {'min': 1, max: maxPageVal}}]}).attr('data-oj-internal', '');
        }
        
        if (($.inArray(this._PAGE_OPTION_LAYOUT._AUTO, pageOptionLayout) != -1 && !isDot) ||
            $.inArray(this._PAGE_OPTION_LAYOUT._ALL, pageOptionLayout) != -1 ||
            $.inArray(this._PAGE_OPTION_LAYOUT._RANGE_TEXT, pageOptionLayout) != -1)
        {
          var pagingControlNavSummaryLabel = $(document.createElement('span'));
          pagingControlNavSummaryLabel.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CLASS);
          pagingControlNavSummaryLabel.addClass('oj-label-inline');
          var itemRange = this._getItemRange(size, startIndex);
          if (itemRange.text().length > 0)
          {
            pagingControlNavSummaryLabel.append("("); //@HTMLUpdateOK
            pagingControlNavSummaryLabel.append(itemRange); //@HTMLUpdateOK
            pagingControlNavSummaryLabel.append(")"); //@HTMLUpdateOK
          }
          pagingControlNav.append(pagingControlNavSummaryLabel); //@HTMLUpdateOK
        }

        // nav arrow section
        var pagingControlNavArrowSection = $(document.createElement('div'));
        pagingControlNavArrowSection.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_ARROW_SECTION_CLASS);
        pagingControlNav.append(pagingControlNavArrowSection); //@HTMLUpdateOK
        
        if (($.inArray(this._PAGE_OPTION_LAYOUT._AUTO, pageOptionLayout) != -1 && !isDot) ||
            $.inArray(this._PAGE_OPTION_LAYOUT._ALL, pageOptionLayout) != -1 ||
            $.inArray(this._PAGE_OPTION_LAYOUT._NAV, pageOptionLayout) != -1)
        {
          var pagingControlNavFirst = this._createNavArrow(
            this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_CLASS,
            isVertical ? this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_VERTICAL_ICON_CLASS :
                         this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_ICON_CLASS,
            this._BUNDLE_KEY._TIP_NAV_FIRST_PAGE,
            this._BUNDLE_KEY._LABEL_ACC_NAV_FIRST_PAGE,
            this._CSS_CLASSES._PAGING_CONTROL_NAV_FIRST_ACC_LABEL_CLASS,
            isVertical);

          pagingControlNavArrowSection.append(pagingControlNavFirst); //@HTMLUpdateOK

          var pagingControlNavPrevious = this._createNavArrow(
            this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_CLASS,
            isVertical ? this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_VERTICAL_ICON_CLASS :
                         this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_ICON_CLASS,
            this._BUNDLE_KEY._TIP_NAV_PREVIOUS_PAGE,
            this._BUNDLE_KEY._LABEL_ACC_NAV_PREVIOUS_PAGE,
            this._CSS_CLASSES._PAGING_CONTROL_NAV_PREVIOUS_ACC_LABEL_CLASS,
            isVertical);

          pagingControlNavArrowSection.append(pagingControlNavPrevious); //@HTMLUpdateOK
        }

        // nav pages section
        if ($.inArray(this._PAGE_OPTION_LAYOUT._AUTO, pageOptionLayout) != -1 ||
            $.inArray(this._PAGE_OPTION_LAYOUT._ALL, pageOptionLayout) != -1 ||
            $.inArray(this._PAGE_OPTION_LAYOUT._PAGES, pageOptionLayout) != -1)
        {
          var pagingControlNavPagesSection = $(document.createElement('div'));
          pagingControlNavPagesSection.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGES_SECTION_CLASS);
          pagingControlNavArrowSection.append(pagingControlNavPagesSection); //@HTMLUpdateOK
          this._createPagingControlNavPages(pagingControlNavPagesSection, this._getMaxPageLinks(), size, startIndex);
        }

        if (($.inArray(this._PAGE_OPTION_LAYOUT._AUTO, pageOptionLayout) != -1 && !isDot) ||
            $.inArray(this._PAGE_OPTION_LAYOUT._ALL, pageOptionLayout) != -1 ||
            $.inArray(this._PAGE_OPTION_LAYOUT._NAV, pageOptionLayout) != -1)
        {
          var pagingControlNavNext = this._createNavArrow(
            this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_CLASS,
            isVertical ? this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_VERTICAL_ICON_CLASS :
                         this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_ICON_CLASS,
            this._BUNDLE_KEY._TIP_NAV_NEXT_PAGE,
            this._BUNDLE_KEY._LABEL_ACC_NAV_NEXT_PAGE,
            this._CSS_CLASSES._PAGING_CONTROL_NAV_NEXT_ACC_LABEL_CLASS,
            isVertical);

          pagingControlNavArrowSection.append(pagingControlNavNext); //@HTMLUpdateOK

          var pagingControlNavLast = this._createNavArrow(
            this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_CLASS,
            isVertical ? this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_VERTICAL_ICON_CLASS :
                         this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_ICON_CLASS,
            this._BUNDLE_KEY._TIP_NAV_LAST_PAGE,
            this._BUNDLE_KEY._LABEL_ACC_NAV_LAST_PAGE,
            this._CSS_CLASSES._PAGING_CONTROL_NAV_LAST_ACC_LABEL_CLASS,
            isVertical);

          pagingControlNavArrowSection.append(pagingControlNavLast); //@HTMLUpdateOK
        }

        return pagingControlNav;
      },
      /**
       * Create the page links
       * @param {jQuery} parentDiv parent element
       * @param {number} numLinks number of page links
       * @param {number} size Number of rows
       * @param {number} startIndex Start index
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createPagingControlNavPages: function(parentDiv, numLinks, size, startIndex)
      {
        if (numLinks < 5)
        {
          var errSummary = this.getTranslatedString(this._BUNDLE_KEY._ERR_MAXPAGELINKS_INVALID_SUMMARY);
          var errDetail = this.getTranslatedString(this._BUNDLE_KEY._ERR_MAXPAGELINKS_INVALID_DETAIL);
          throw new Error(errSummary + '\n' + errDetail);
        }
        var pagingControlNavPagesLinks = $(document.createElement('div'));
        pagingControlNavPagesLinks.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGES_LINKS_CLASS);
        parentDiv.append(pagingControlNavPagesLinks); //@HTMLUpdateOK
        var totalPages = this._getTotalPages();
        var currentPage = this._getCurrentPage();
        var pageSize = this.options['pageSize'];
        
        var numPagesToAdd = numLinks;
        // this will hold our page list
        var pageList = [];

        if (currentPage >= 0)
        {
          var i;
          if (this._isTotalSizeConfidenceActual() && 
              totalPages <= numPagesToAdd)
          {
            // always add the first page
            pageList[0] = 0;
            
            // just enumerate the pages
            for (i = 1; i < totalPages; i++)
            {
              pageList[i] = i;
            }
          }
          else
          {
            // add the first, current, and last page
            pageList.push(0);
            if (currentPage != 0)
            {
              pageList.push(currentPage);
            }
            // add last page if known row count
            if (currentPage != totalPages - 1 && 
                this._isTotalSizeConfidenceActual())
            {
              pageList.push(totalPages - 1);
            }
            numPagesToAdd = numPagesToAdd - pageList.length;
            // keep adding before the current page till we get to the
            // first page or we've added numPagesToAdd - 1.
            // If the last page or 2nd to last page then add until numPagesToAdd
            var pageBeforeCurrent = currentPage - 1;
            // number of pages to add after current
            var numPagesAfterCurrent = 1;
            // if at last page or second to last page then don't add any pages
            // after current
            if (this._isTotalSizeConfidenceActual() &&
                (currentPage == totalPages - 1 || currentPage == totalPages - 2))
            {
              numPagesAfterCurrent = 0;
            }
            while (numPagesToAdd > numPagesAfterCurrent && pageBeforeCurrent >= 1)
            {
              pageList.push(pageBeforeCurrent);
              pageBeforeCurrent--;
              numPagesToAdd--;
            }
            // keep adding after the current page
            var pageAfterCurrent = currentPage + 1;
            // if unknown row count, only add one page after current if there is data
            if (totalPages == -1)
            {
              if (size > 0 && size >= pageSize)
              {
                numPagesToAdd = 1;
              }
              else
              {
                numPagesToAdd = 0;
              }
            }
            while (numPagesToAdd > 0 && (pageAfterCurrent <= totalPages || totalPages == -1))
            {
              pageList.push(pageAfterCurrent);
              pageAfterCurrent++;
              numPagesToAdd--;
            }
          }

          // sort the pageList array
          var compareNumbers = function (a, b) {
            return a - b;
          };
          
          pageList.sort(compareNumbers);

          for (i = 0; i < pageList.length; i++)
          {
            var pageNum = pageList[i];
            this._createPagingControlNavPage(pagingControlNavPagesLinks, pageNum);
            // check if we have a gap
            if (i != pageList.length - 1)
            {
              if (pageNum != pageList[i + 1] - 1)
              {
                this._createPagingControlNavPage(pagingControlNavPagesLinks, -1);
              }
            }
          }
          if (!this._isTotalSizeConfidenceActual() && size >= pageSize)
          {
            this._createPagingControlNavPage(pagingControlNavPagesLinks, -1);
          }
        }
        return pagingControlNavPagesLinks;
      },
      /**
       * Create the page link or page gap
       * @param {jQuery} parentDiv parent element
       * @param {number} pageNum page number
       * @return {jQuery} jQuery div DOM element
       * @private
       */
      _createPagingControlNavPage: function(parentDiv, pageNum)
      {
        var currentPage = this._getCurrentPage();
        var pagingControlNavPage = null;
        var isRTL = (this._GetReadingDirection() === "rtl");
        var isVertical = this.options['pageOptions']['orientation'] == 'vertical' ? true : false;
        var isDot = this.options['pageOptions']['type'] == 'dots' ? true : false;
        if (pageNum == -1)
        {
          pagingControlNavPage = $(document.createElement('span'));
          pagingControlNavPage.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGE_ELLIPSIS_CLASS);
          pagingControlNavPage.text("...");
          parentDiv.append(pagingControlNavPage); //@HTMLUpdateOK
        }
        else
        {          
          if (currentPage == pageNum)
          {
             pagingControlNavPage = $(document.createElement('div'));
             pagingControlNavPage.addClass(this._MARKER_STYLE_CLASSES._SELECTED);
             pagingControlNavPage.addClass(this._MARKER_STYLE_CLASSES._ACTIVE);
             pagingControlNavPage.addClass(this._MARKER_STYLE_CLASSES._DISABLED);
             pagingControlNavPage.removeClass(this._MARKER_STYLE_CLASSES._ENABLED);
          }
          else
          {
            pagingControlNavPage = $(document.createElement('a'));
            pagingControlNavPage.removeClass(this._MARKER_STYLE_CLASSES._SELECTED);
            pagingControlNavPage.removeClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            pagingControlNavPage.removeClass(this._MARKER_STYLE_CLASSES._DISABLED);
            pagingControlNavPage.addClass(this._MARKER_STYLE_CLASSES._ENABLED);
            pagingControlNavPage.attr(this._TAB_INDEX, '0');
            pagingControlNavPage.attr('href', '#');
          }
          pagingControlNavPage.attr('data-oj-pagenum', pageNum);
          if (!isDot)
          {
            pagingControlNavPage.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGE_CLASS);
          }
          else
          {
            pagingControlNavPage.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_DOT_CLASS);
          }
          var pageTitle = this.getTranslatedString(this._BUNDLE_KEY._TIP_NAV_PAGE_LINK, {'pageNum': (pageNum + 1).toString()});
          this._AddHoverable(pagingControlNavPage);
          this._focusable({'element': pagingControlNavPage, 'applyHighlight': true});
          pagingControlNavPage.attr('title', pageTitle);
          pagingControlNavPage.attr('oncontextmenu', 'return false;');
          // create the acc label for the page link
          var accPageLabel = this._createPagingControlAccNavPageLabel();
          pagingControlNavPage.append(accPageLabel); //@HTMLUpdateOK
          var pagingControlNavPageSpan = $(document.createElement('span'));
          pagingControlNavPageSpan.append((pageNum + 1).toString()); //@HTMLUpdateOK
          
          if (isDot)
          {
            // hide the number
            pagingControlNavPageSpan.addClass(this._CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
            // add the dot
            var pagingControlNavPageDotBulletSpan = $(document.createElement('span'));
            pagingControlNavPageDotBulletSpan.addClass(this._CSS_CLASSES._PAGING_CONTROL_NAV_DOT_BULLET_CLASS);
            this._AddHoverable(pagingControlNavPageDotBulletSpan);
            
            if (currentPage == pageNum)
            {
               pagingControlNavPageDotBulletSpan.addClass(this._MARKER_STYLE_CLASSES._SELECTED);
               pagingControlNavPageDotBulletSpan.addClass(this._MARKER_STYLE_CLASSES._ACTIVE);
            }
            pagingControlNavPage.append(pagingControlNavPageDotBulletSpan); //@HTMLUpdateOK
          }
          else
          {
            this._AddHoverable(pagingControlNavPage);
          }
          var dirAttrVal = isRTL ? 'rtl' : 'ltr';
          pagingControlNavPageSpan.attr('dir', dirAttrVal);
          pagingControlNavPage.append(pagingControlNavPageSpan); //@HTMLUpdateOK
          if (isVertical)
          {
            pagingControlNavPage.css('display', 'block');
          }
          parentDiv.append(pagingControlNavPage); //@HTMLUpdateOK
        }
        return pagingControlNavPage;
      },
      /**
       * Return the paging content acc label
       * @return {jQuery|null} jQuery div DOM element
       * @private
       */
      _getPagingControlAccLabel: function()
      {
        var pagingControlContainer = this._getPagingControlContainer();
        var pagingControlContentAccLabel = null;
        
        if (pagingControlContainer)
        {
          pagingControlContentAccLabel = pagingControlContainer.find('.' + this._CSS_CLASSES._PAGING_CONTROL_ACC_LABEL_CLASS);
          if (pagingControlContentAccLabel && pagingControlContentAccLabel.length > 0)
          {
            pagingControlContentAccLabel = $(pagingControlContentAccLabel.get(0));
          }
        }
        
        return pagingControlContentAccLabel;
      },
      /**
       * Return the paging container
       * @return {jQuery|null} jQuery div DOM element
       * @private
       */
      _getPagingControlContainer: function()
      {
        return $(this.element);
      },
      /**
       * Return the paging content
       * @return {jQuery|null} jQuery div DOM element
       * @private
       */
      _getPagingControlContent: function()
      {
        if (!this._cachedDomPagingControlContent)
        {
          var pagingControlContainer = this._getPagingControlContainer();
          var pagingControlContent = null;
          if (pagingControlContainer)
          {
            pagingControlContent = pagingControlContainer.find('.' + this._CSS_CLASSES._PAGING_CONTROL_CONTENT_CLASS);
            if (pagingControlContent && pagingControlContent.length > 0)
            {
              this._cachedDomPagingControlContent = $(pagingControlContent.get(0));
            }
          }
        }

        return this._cachedDomPagingControlContent;
      },
      /**
       * Return the Load More div
       * @return {jQuery|null} jQuery a DOM element
       * @private
       */
      _getPagingControlLoadMore: function()
      {
        if (!this._cachedDomPagingControlLoadMore)
        {
          var pagingControlContent = this._getPagingControlContent();
          var pagingControlLoadMore = null;
          if (pagingControlContent)
          {
            pagingControlLoadMore = pagingControlContent.children('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_CLASS);
            if (pagingControlLoadMore && pagingControlLoadMore.length > 0)
            {
              this._cachedDomPagingControlLoadMore = $(pagingControlLoadMore.get(0));
            }
          }
        }

        return this._cachedDomPagingControlLoadMore;
      },
      /**
       * Return the Load More link
       * @return {jQuery|null} jQuery a DOM element
       * @private
       */
      _getPagingControlLoadMoreLink: function()
      {
        if (!this._cachedDomPagingControlLoadMoreLink)
        {
          var pagingControlLoadMore = this._getPagingControlLoadMore();
          var pagingControlLoadMoreLink = null;
          if (pagingControlLoadMore)
          {
            pagingControlLoadMoreLink = pagingControlLoadMore.children('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_LINK_CLASS);
            if (pagingControlLoadMoreLink && pagingControlLoadMoreLink.length > 0)
            {
              this._cachedDomPagingControlLoadMoreLink = $(pagingControlLoadMoreLink.get(0));
            }
          }
        }

        return this._cachedDomPagingControlLoadMoreLink;
      },
      /**
       * Return the Load More Range
       * @return {jQuery|null} jQuery span DOM element
       * @private
       */
      _getPagingControlLoadMoreRange: function()
      {
        if (!this._cachedDomPagingControlLoadMoreRange)
        {
          var pagingControlLoadMore = this._getPagingControlLoadMore();
          var pagingControlLoadMoreRange = null;
          if (pagingControlLoadMore)
          {
            pagingControlLoadMoreRange = pagingControlLoadMore.children('.' + this._CSS_CLASSES._PAGING_CONTROL_LOAD_MORE_RANGE_CLASS);
            if (pagingControlLoadMoreRange && pagingControlLoadMoreRange.length > 0)
            {
              this._cachedDomPagingControlLoadMoreRange = $(pagingControlLoadMoreRange.get(0));
            }
          }
        }

        return this._cachedDomPagingControlLoadMoreRange;
      },
      /**
       * Return the paging nav bar
       * @return {jQuery|null} jQuery a DOM element
       * @private
       */
      _getPagingControlNav: function()
      {
        if (!this._cachedDomPagingControlNav)
        {
          var pagingControlContent = this._getPagingControlContent();
          var pagingControlNav = null;
          if (pagingControlContent)
          {
            pagingControlNav = pagingControlContent.children('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_CLASS);
            if (pagingControlNav && pagingControlNav.length > 0)
            {
              this._cachedDomPagingControlNav = $(pagingControlNav.get(0));
            }
          }
        }

        return this._cachedDomPagingControlNav;
      },
      /**
       * Return the paging nav input
       * @return {jQuery|null} jQuery input DOM element
       * @private
       */
      _getPagingControlNavInput: function()
      {
        if (!this._cachedDomPagingControlNavInput)
        {
          var pagingControlNav = this._getPagingControlNav();
          var pagingControlNavInput = null;
          if (pagingControlNav)
          {
            pagingControlNavInput = pagingControlNav.find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_CLASS);
            if (pagingControlNavInput && pagingControlNavInput.length > 0)
            {
              this._cachedDomPagingControlNavInput = $(pagingControlNavInput.get(0));
            }
          }
        }

        return this._cachedDomPagingControlNavInput;
      },
      /**
       * Return the paging nav input summary
       * @return {jQuery|null} jQuery input DOM element
       * @private
       */
      _getPagingControlNavInputSummary: function()
      {
        if (!this._cachedDomPagingControlNavInputSummary)
        {
          var pagingControlNav = this._getPagingControlNav();
          var pagingControlNavInputSummary = null;
          if (pagingControlNav)
          {
            pagingControlNavInputSummary = pagingControlNav.find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SUMMARY_CLASS);
            if (pagingControlNavInputSummary && pagingControlNavInputSummary.length > 0)
            {
              this._cachedDomPagingControlNavInputSummary = $(pagingControlNavInputSummary.get(0));
            }
          }
        }

        return this._cachedDomPagingControlNavInputSummary;
      },
      /**
       * Return the page links
       * @private
       */
      _getPagingControlNavPageLinks: function()
      {
        var pagingControlNav = this._getPagingControlNav();
        var pagingControlNavPageLinks = null;
        if (pagingControlNav)
        {
          pagingControlNavPageLinks = pagingControlNav.find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_PAGES_LINKS_CLASS);
          if (pagingControlNavPageLinks && pagingControlNavPageLinks.length > 0)
          {
            pagingControlNavPageLinks = $(pagingControlNavPageLinks.get(0));
          }
        }

        return pagingControlNavPageLinks;
      },
      /**
       * Return the nav arrows
       * @private
       */
      _getPagingControlNavArrowSection: function()
      {
        var pagingControlNav = this._getPagingControlNav();
        var pagingControlNavArrowSection = null;
        if (pagingControlNav)
        {
          pagingControlNavArrowSection = pagingControlNav.find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_ARROW_SECTION_CLASS);
          if (pagingControlNavArrowSection && pagingControlNavArrowSection.length > 0)
          {
            pagingControlNavArrowSection = $(pagingControlNavArrowSection.get(0));
          }
          else
          {
            return null;
          }
        }

        return pagingControlNavArrowSection;
      }
      ,
      /**
       * Return the nav input section
       * @private
       */
      _getPagingControlNavInputSection: function()
      {
        var pagingControlNav = this._getPagingControlNav();
        var pagingControlNavInputSection = null;
        if (pagingControlNav)
        {
          pagingControlNavInputSection = pagingControlNav.find('.' + this._CSS_CLASSES._PAGING_CONTROL_NAV_INPUT_SECTION_CLASS);
          if (pagingControlNavInputSection && pagingControlNavInputSection.length > 0)
          {
            pagingControlNavInputSection = $(pagingControlNavInputSection.get(0));
          }
          else
          {
            return null;
          }
        }

        return pagingControlNavInputSection;
      }
      /**** end internal DOM functions ****/
    })
    //////////////////     FRAGMENTS    //////////////////
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Page Navigation Bar</td>
     *       <td><kbd>Swipe</kbd></td>
     *       <td>When mode='page', swiping left or right on the page navigation bar will either increment or decrement the page respectively.</td>
     *     </tr>
     *     <tr>
     *       <td>Page Number Input</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Set focus to the input.</td>
     *     </tr>
     *     <tr>
     *       <td>Arrow Page Navigation</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Navigates to the first, previous, next, or last page.</td>
     *     </tr>
     *     <tr>
     *       <td>Numbered Page Links</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Navigates to the page.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch section of classdesc, and standalone gesture doc
     * @memberof oj.ojPagingControl
     */  
    
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Page Number Input</td>
     *       <td><kbd>Tab In</kbd></td>
     *       <td>Set focus to the input.</td>
     *     </tr>
     *     <tr>
     *       <td>Arrow Page Navigation</td>
     *       <td><kbd>Tab</kbd></td>
     *       <td>Set focus to the first, previous, next, or last page arrow.</td>
     *     </tr>
     *     <tr>
     *       <td>Numbered Page Links</td>
     *       <td><kbd>Tab</kbd></td>
     *       <td>Set focus to to the page link.</td>
     *     </tr>
     *   </tbody> 
     * </table>
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojPagingControl
     */
    
    //////////////////     SUB-IDS     //////////////////
    /**
     * <p>Sub-ID for the PagingControl page number navigation input.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-input
     * @memberof oj.ojPagingControl
     *
     * @example <caption>Get the page number navigation input:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-input'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl current maximum page text.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-input-max
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the current maximum page text:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-input-max'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl summary items text.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-input-summary
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the summary items text:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-input-summary'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl summary current items text.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-input-summary-current
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the summary current items text:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-input-summary-current'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl summary max items text.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-input-summary-max
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the summary max items text:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-input-summary-max'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl first page button.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-first
     * @memberof oj.ojPagingControl
     *
     * @example <caption>Get the first page button:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-first'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl next page button.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-next
     * @memberof oj.ojPagingControl
     *
     * @example <caption>Get the next page button:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-next'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl previous page button.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-previous
     * @memberof oj.ojPagingControl
     *
     * @example <caption>Get the previous page button:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-previous'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl previous page button.</p>
     *
     * @ojsubid oj-pagingcontrol-nav-last
     * @memberof oj.ojPagingControl
     *
     * @example <caption>Get the last page button:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-last'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl page button.</p>
     * To lookup a page button the locator object should have the following:
     * <ul>
     * <li><b>index</b>zero-based index of page number node</li>
     * </ul>
     *
     * @ojsubid oj-pagingcontrol-nav-page
     * @memberof oj.ojPagingControl
     *
     * @example <caption>Get the page button:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-nav-page', 'index': 1} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl Show More link.</p>
     *
     * @ojsubid oj-pagingcontrol-load-more-link
     * @memberof oj.ojPagingControl
     *
     * @example <caption>Get the Show More link:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-load-more-link'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl load more range text.</p>
     *
     * @ojsubid oj-pagingcontrol-load-more-range
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the load more range text:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-load-more-range'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl load more range current items text.</p>
     *
     * @ojsubid oj-pagingcontrol-load-more-range-current
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the load more range current items text:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-load-more-range-current'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl load more range max items text.</p>
     *
     * @ojsubid oj-pagingcontrol-load-more-range-max
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the load more range max items text:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-load-more-range-max'} );
     */
    
    /**
     * <p>Sub-ID for the PagingControl load more max message.</p>
     *
     * @ojsubid oj-pagingcontrol-load-more-max-rows
     * @deprecated 2.0.2 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojPagingControl
     * @ignore
     *
     * @example <caption>Get the load more max message:</caption>
     * var node = myPagingControl.getNodeBySubId( {'subId': 'oj-pagingcontrol-load-more-max-rows'} );
     */

}());
(function() {
var ojPagingControlMeta = {
  "properties": {
    "data": {},
    "loadMoreOptions": {
      "type": "Object",
      "properties": {
        "maxCount": {
          "type": "number"
        }
      }
    },
    "mode": {
      "type": "string",
      "enumValues": ["page", "loadMore"]
    },
    "overflow": {
      "type": "string",
      "enumValues": ["fit", "none"]
    },
    "pageOptions": {
      "type": "Object",
      "properties": {
        "layout" : {
          "type": "Array<string>"
        },
        "maxPageLinks" : {
          "type": "number"
        },
        "orientation": {
          "type": "string",
          "enumValues": ["horizontal", "vertical"]
        },
        "type": {
          "type": "string",
          "enumValues": ["numbers", "dots"]
        }
      }
    },
    "pageSize": {
      "type": "number"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "labelAccNavFirstPage": {
          "type": "string",
          "value": "First Page"
        },
        "labelAccNavLastPage": {
          "type": "string",
          "value": "Last Page"
        },
        "labelAccNavNextPage": {
          "type": "string",
          "value": "Next Page"
        },
        "labelAccNavPage": {
          "type": "string",
          "value": "Page"
        },
        "labelAccNavPreviousPage": {
          "type": "string",
          "value": "Previous Page"
        },
        "labelAccPaging": {
          "type": "string",
          "value": "Pagination"
        },
        "labelLoadMore": {
          "type": "string",
          "value": "Show More..."
        },
        "labelLoadMoreMaxRows": {
          "type": "string",
          "value": "Reached Maximum Limit of {maxRows} rows"
        },
        "labelNavInputPage": {
          "type": "string",
          "value": "Page"
        },
        "labelNavInputPageMax": {
          "type": "string",
          "value": "of {pageMax}"
        },
        "maxPageLinksInvalid": {
          "type": "string",
          "value": "Value for maxPageLinks is invalid."
        },
        "msgItemRangeCurrent": {
          "type": "string",
          "value": "{pageFrom}-{pageTo}"
        },
        "msgItemRangeCurrentSingle": {
          "type": "string",
          "value": "{pageFrom}"
        },
        "msgItemRangeItems": {
          "type": "string",
          "value": "items"
        },
        "msgItemRangeOf": {
          "type": "string",
          "value": "of"
        },
        "msgItemRangeOfApprox": {
          "type": "string",
          "value": "approx."
        },
        "msgItemRangeOfAtLeast": {
          "type": "string",
          "value": "at least"
        },
        "pageInvalid": {
          "type": "string",
          "value": "The page value entered is invalid."
        },
        "tipNavFirstPage": {
          "type": "string",
          "value": "First"
        },
        "tipNavInputPage": {
          "type": "string",
          "value": "Go To Page"
        },
        "tipNavLastPage": {
          "type": "string",
          "value": "Last"
        },
        "tipNavNextPage": {
          "type": "string",
          "value": "Next"
        },
        "tipNavPageLink": {
          "type": "string",
          "value": "Go To Page {pageNum}"
        },
        "tipNavPreviousPage": {
          "type": "string",
          "value": "Previous"
        }
      }
    }
  },
  "methods": {
    "firstPage": {},
    "lastPage": {},
    "loadNext": {},
    "nextPage": {},
    "page": {},
    "previousPage": {}
  },
  "extension": {
    _WIDGET_NAME: "ojPagingControl"
  }
};
oj.CustomElementBridge.registerMetadata('oj-paging-control', 'baseComponent', ojPagingControlMeta);
oj.CustomElementBridge.register('oj-paging-control', {'metadata': oj.CustomElementBridge.getMetadata('oj-paging-control')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/
/**
 * @export
 * This class captures all translation resources and style classes used by the DataGrid.
 * This should be populated with information extracted through the framework and set on the DataGrid.
 * Internal.  Developers should never use this class.
 * @constructor
 * @ignore
 */
oj.DataGridResources = function(rtlMode, translationFunction, defaultOptions)
{
    this.rtlMode = rtlMode;
    this.translationFunction = translationFunction;
    this.defaultOptions = defaultOptions;
    this.styles = {};
    this.styles['datagrid'] = "oj-datagrid";
    this.styles['cell'] = "oj-datagrid-cell";
    this.styles['banded'] = "oj-datagrid-banded";
    this.styles['row'] = "oj-datagrid-row";
    this.styles['databody'] = "oj-datagrid-databody";
    this.styles['topcorner'] = "oj-datagrid-top-corner";
    this.styles['bottomcorner'] = "oj-datagrid-bottom-corner";
    this.styles['rowheaderspacer'] = "oj-datagrid-row-header-spacer";
    this.styles['colheaderspacer'] = "oj-datagrid-column-header-spacer";
    this.styles['status'] = "oj-datagrid-status";
    this.styles['loadingicon'] = "oj-icon oj-datagrid-loading-icon"
    this.styles['emptytext'] = "oj-datagrid-empty-text";
    this.styles['header'] = "oj-datagrid-header";
    this.styles['endheader'] = "oj-datagrid-end-header";
    this.styles['groupingcontainer'] = "oj-datagrid-header-grouping";
    this.styles['headercell'] = "oj-datagrid-header-cell";
    this.styles['rowheader'] = "oj-datagrid-row-header";
    this.styles['colheader'] = "oj-datagrid-column-header";
    this.styles['colheadercell'] = "oj-datagrid-column-header-cell";
    this.styles['rowheadercell'] = "oj-datagrid-row-header-cell";
    this.styles['endheadercell'] = "oj-datagrid-end-header-cell";
    this.styles['rowendheader'] = "oj-datagrid-row-end-header";
    this.styles['colendheader'] = "oj-datagrid-column-end-header";
    this.styles['colendheadercell'] = "oj-datagrid-column-end-header-cell";
    this.styles['rowendheadercell'] = "oj-datagrid-row-end-header-cell";
    this.styles['scroller-mobile'] = "oj-datagrid-scroller-touch";
    this.styles['scroller'] = "oj-datagrid-scroller";
    this.styles['scrollers'] = "oj-datagrid-scrollers";
    this.styles['focus'] = "oj-focus";
    this.styles['hover'] = "oj-hover";
    this.styles['active'] = "oj-active";
    this.styles['selected'] = "oj-selected";
    this.styles['topSelected'] = "oj-datagrid-selected-top";
    this.styles['bottomSelected'] = "oj-datagrid-selected-bottom";
    this.styles['disabled'] = "oj-disabled";
    this.styles['enabled'] = "oj-enabled";
    this.styles['default'] = "oj-default";
    this.styles['sortcontainer'] = "oj-datagrid-sort-icon-container";
    this.styles['sortascending'] = "oj-datagrid-sort-ascending-icon";
    this.styles['sortdescending'] = "oj-datagrid-sort-descending-icon";
    this.styles['icon'] = "oj-component-icon";
    this.styles['clickableicon'] = "oj-clickable-icon-nocontext";
    this.styles['info'] = "oj-helper-hidden-accessible";
    this.styles['rowexpander'] = "oj-rowexpander";
    this.styles['cut'] = "oj-datagrid-cut";
    this.styles['selectaffordancetop'] = "oj-datagrid-touch-selection-affordance-top";
    this.styles['selectaffordancebottom'] = "oj-datagrid-touch-selection-affordance-bottom";
    this.styles['toucharea'] = "oj-datagrid-touch-area";
    this.styles['readOnly'] = "oj-read-only";
    this.styles['editable'] = "oj-datagrid-editable";
    this.styles['cellEdit'] = "oj-datagrid-cell-edit";
    this.styles['draggable'] = "oj-draggable";
    this.styles['drag'] = "oj-drag";
    this.styles['drop'] = "oj-drop";
    this.styles['activedrop'] = "oj-active-drop";
    this.styles['validdrop'] = "oj-valid-drop";
    this.styles['invaliddrop'] = "oj-invalid-drop";
    this.styles['formcontrol'] = "oj-form-control-inherit";
    this.styles['borderHorizontalNone'] = "oj-datagrid-border-horizontal-none";
    this.styles['borderVerticalNone'] = "oj-datagrid-border-vertical-none";
    this.styles['borderHorizontalSmall'] = "oj-datagrid-small-content-border-horizontal";
    this.styles['borderVerticalSmall'] = "oj-datagrid-small-content-border-vertical";
    this.styles['offsetOutline'] = "oj-datagrid-focus-offset";
    this.styles['depth'] = "oj-datagrid-depth-";

    this.commands = {};
    this.commands['sortCol'] = "oj-datagrid-sortCol";
    this.commands['sortColAsc'] = "oj-datagrid-sortColAsc";
    this.commands['sortColDsc'] = "oj-datagrid-sortColDsc";
    this.commands['sortRow'] = "oj-datagrid-sortRow";
    this.commands['sortRowAsc'] = "oj-datagrid-sortRowAsc";
    this.commands['sortRowDsc'] = "oj-datagrid-sortRowDsc";
    this.commands['resize'] = "oj-datagrid-resize";
    this.commands['resizeWidth'] = "oj-datagrid-resizeWidth";
    this.commands['resizeHeight'] = "oj-datagrid-resizeHeight";
    this.commands['cut'] = "oj-datagrid-cut";
    this.commands['paste'] = "oj-datagrid-paste";
    this.commands['discontiguousSelection'] = "oj-datagrid-discontiguousSelection";

    this.attributes = {};
    this.attributes['busyContext'] = oj.Context._OJ_CONTEXT_ATTRIBUTE; // 'data-oj-context'
    this.attributes['context'] = "data-oj-cellContext";
    this.attributes['resizable'] = "data-oj-resizable";
    this.attributes['sortable'] = "data-oj-sortable";
    this.attributes['sortDir'] = "data-oj-sortdir";
    this.attributes['expander'] = "data-oj-expander";
    this.attributes['expanderIndex'] = "data-oj-expander-index";
    this.attributes['container'] = oj.Components._OJ_CONTAINER_ATTR;
    this.attributes['extent'] = "data-oj-extent";
    this.attributes['start'] = "data-oj-start";
    this.attributes['depth'] = "data-oj-depth";
    this.attributes['level'] = "data-oj-level";
    this.attributes['tabMod'] = "data-oj-tabmod";
};

/**
 * Whether the reading direction is right to left.
 * @return {boolean} true if reading direction is right to left, false otherwise.
 * @export
 */
oj.DataGridResources.prototype.isRTLMode = function()
{
    return (this.rtlMode === "rtl") ? true : false;
};

/**
 * Gets the translated text
 * @param {string} key the key to the translated text
 * @param {Array=} args optional arguments to format the translated text
 * @return {string|null} the translated text
 * @export
 */
oj.DataGridResources.prototype.getTranslatedText = function(key, args)
{
    return this.translationFunction(key, args);
};

/**
 * Gets the default option
 * @param {string} key the key to the default option
 * @return {Object|null} the value of the option
 * @export
 */
oj.DataGridResources.prototype.getDefaultOption = function(key)
{
    if (this.defaultOptions)
    {
        return this.defaultOptions[key];
    }
    return null;
};

/**
 * Gets the mapped style class
 * @param {string} key the key to the style class
 * @return {string|null} the style class
 * @export
 */
oj.DataGridResources.prototype.getMappedStyle = function(key)
{
    if (key != null)
    {
        return this.styles[key];
    }
    return null;
};

/**
 * Gets the mapped command class
 * @param {string} key the key to the command class
 * @return {string|null} the command class
 * @export
 */
oj.DataGridResources.prototype.getMappedCommand = function(key)
{
    if (key != null)
    {
        return this.commands[key];
    }
    return null;
};

/**
 * Gets the mapped attribute
 * @param {string} key the key to the attribute
 * @return {string|null} the attribute
 * @export
 */
oj.DataGridResources.prototype.getMappedAttribute = function(key)
{
    if (key != null)
    {
        return this.attributes[key];
    }
    return null;
};
/**
 * Creates a new DataGrid
 * @constructor
 * @private
 */
var DvtDataGrid = function(root)
{
    this.m_root = root;
    
    this.MAX_COLUMN_THRESHOLD = 20;
    this.MAX_ROW_THRESHOLD = 30;

    this.m_utils = new DvtDataGridUtils(this);

    this.m_discontiguousSelection = false;

    this.m_sizingManager = new DvtDataGridSizingManager();

    this.m_keyboardHandler = new DvtDataGridKeyboardHandler(this);

    this.m_rowHeaderWidth = null;
    this.m_rowEndHeaderWidth = null;
    this.m_colHeaderHeight = null;
    this.m_colEndHeaderHeight = null;

    // a mapping of style class+inline style combo to a dimension
    // to reduce the need to do excessive offsetWidth/offsetHeight
    this.m_styleClassDimensionMap = {'width':{}, 'height':{}};

    // cached whether row/column count are unknown
    this.m_isEstimateRowCount = undefined;
    this.m_isEstimateColumnCount = undefined;
    this.m_stopRowFetch = false;
    this.m_stopRowHeaderFetch = false;
    this.m_stopRowEndHeaderFetch = false;
    this.m_stopColumnFetch = false;
    this.m_stopColumnHeaderFetch = false;
    this.m_stopColumnEndHeaderFetch = false;

    // not done initializing until initial headers/cells are fetched
    this.m_initialized = false;
    this.m_shouldFocus = null;

    this.callbacks = {};

    this._setupActions();

    this.m_readinessStack = [];
    this.m_modelEvents = [];

    this.m_databodyMap = new Map();
};

// keyCodes for data grid keyboard
DvtDataGrid.keyCodes =
{
    TAB_KEY : 9,
    ENTER_KEY : 13,
    SHIFT_KEY : 16,
    CTRL_KEY : 17,
    ALT_KEY : 18,
    ESC_KEY : 27,
    SPACE_KEY : 32,
    PAGEUP_KEY : 33,
    PAGEDOWN_KEY : 34,
    END_KEY : 35,
    HOME_KEY : 36,
    LEFT_KEY : 37,
    UP_KEY : 38,
    RIGHT_KEY : 39,
    DOWN_KEY : 40,
    NUM5_KEY : 53,
    V_KEY : 86,
    X_KEY : 88,
    F1_KEY : 112,
    F2_KEY : 113,
    F8_KEY : 119,
    F10_KEY : 121,
    F15_KEY : 126,
    A_KEY : 65
};

// constants for update animation
DvtDataGrid.UPDATE_ANIMATION_FADE_INOUT = 1;
DvtDataGrid.UPDATE_ANIMATION_SLIDE_INOUT = 2;
DvtDataGrid.UPDATE_ANIMATION_DURATION = 250;

// constants for expand/collapse animation
DvtDataGrid.EXPAND_ANIMATION_DURATION = 500;
DvtDataGrid.COLLAPSE_ANIMATION_DURATION = 400;

// swipe gesture constants
DvtDataGrid.MAX_OVERSCROLL_PIXEL = 50;
DvtDataGrid.BOUNCE_ANIMATION_DURATION = 500;
DvtDataGrid.DECELERATION_FACTOR = 0.0006;
DvtDataGrid.TAP_AND_SCROLL_RESET = 300;
// related to timing and x/y position of events
DvtDataGrid.MIN_SWIPE_DURATION = 200;
DvtDataGrid.MAX_SWIPE_DURATION = 400;
DvtDataGrid.MIN_SWIPE_DISTANCE = 10;
// for the actual transition animation
DvtDataGrid.MIN_SWIPE_TRANSITION_DURATION = 100;
DvtDataGrid.MAX_SWIPE_TRANSITION_DURATION = 500;

// constants for touch gestures
DvtDataGrid.CONTEXT_MENU_TAP_HOLD_DURATION = 750;
DvtDataGrid.HEADER_TAP_SHORT_HOLD_DURATION = 300;

// when filling viewport fetch when this close to the edge
DvtDataGrid.FETCH_PIXEL_THRESHOLD = 5;

//visibility constants
DvtDataGrid.VISIBILITY_STATE_HIDDEN = 'hidden';

DvtDataGrid.VISIBILITY_STATE_REFRESH = 'refresh';

DvtDataGrid.VISIBILITY_STATE_RENDER = 'render';

DvtDataGrid.VISIBILITY_STATE_VISIBLE = 'visible';

/**
 * Sets options on DataGrid
 * @param {Object} options - the options to set on the data grid
 * @param {Function} rendererWrapperFunction - callback function used to fix renderer function for custom element
 */
DvtDataGrid.prototype.SetOptions = function(options, rendererWrapperFunction)
{
    this.m_options = new DvtDataGridOptions(options, rendererWrapperFunction);
};

/**
 * Sets the original event after a sort event for use in selectionChanged events
 * @param {Event} event - sort event to store
 */
DvtDataGrid.prototype.SetSortOriginalEvent = function(event){
    this.m_sortInfo["originalEvent"] = event;
};
/**
 * Update options on DataGrid
 * @param {Object} options - the options to set on the data grid
 * @param {Object=} flags - contains modified subkey
 */
DvtDataGrid.prototype.UpdateOptions = function(options, flags)
{
    var candidate;

    //We should check each updated option (candidate) from the list of updated options (first loop)
    //against options already presented in the internal DvtDataGridOptions (this.m_options) object in
    //purpose to keep them in sync.
    for (candidate in options)
    {
        if (candidate in this.m_options['options'])
        {
            if (this.m_options['options'][candidate] != options[candidate])
            {
                this.m_options['options'][candidate] = options[candidate];
            }
        }
    }
    //now update it
    for (candidate in options)
    {
        if (!this._updateDataGrid(candidate, flags))
        {
            this.empty();
            this.refresh(this.m_root);
            break;
        }
    }
};

/**
 * Partial update for DataGrid
 * @private
 * @param {string} option - the option to update the data grid based on
 * @param {Object=} flags - contains modified subKey if applicable
 * @return {boolean} true if redraw is not required otherwise return false
 */
DvtDataGrid.prototype._updateDataGrid = function(option, flags)
{
    var obj;

    switch (option)
    {
        // updates the data grid can make without refresh
        case "bandingInterval":
            this._removeBanding();
            this.updateColumnBanding();
            this.updateRowBanding();
            break;
        case "currentCell":
            obj = this.m_options.getCurrentCell();
            this._updateActive(obj, true, false);
            break;
        case "editMode":
            this.m_editMode = this.m_options.getEditMode();
            break;
        case "gridlines":
            this._updateGridlines();
            break;
        case "header":
            obj = this.m_options['options']['header'];
            this._updateHeaderOptions(obj, flags);
            break;
        case "scrollPosition":
            obj = this.m_options.getScrollPosition();
            this._updateScrollPosition(obj);
            break;
        case "selection":
            obj = this.m_options.getSelection();
            this._updateSelection(obj);
            break;
        case "selectionMode":
            this._clearSelection(null);
            break;

        // just refresh
        default:
            return false;
    }
    return true;
};

/**
 * Update selection from option
 * @private
 * @param {Object} selection the selection from options
 */
DvtDataGrid.prototype._updateSelection = function(selection)
{
    // selection should not be null
    if (selection == null)
    {
        return;
    }

    // check whether selection is enabled
    if (this._isSelectionEnabled())
    {
        // don't clear the selection so the old one can be passed in
        // sets the new selection
        this.SetSelection(selection);
    }
    else
    {
        // selection is not enable, still need to clear the selection in options
        selection.length = 0;
    }
};

/**
 * Update header resizable and sortable options, all others refresh the grid for now
 * @private
 * @param {Object} headerObject
 * @param {Object=} flags
 */
DvtDataGrid.prototype._updateHeaderOptions = function(headerObject, flags)
{
    var subKey, subKeyArray, axis, option, headers, header, headerContext, i, hasSortContainer, sortIcon;
    if (flags == null || flags['subkey'] == null)
    {
        return;
    }
    subKey = flags['subkey'];
    subKeyArray = subKey.split('.');
    axis = subKeyArray[0];
    option = subKeyArray[1];

    if (axis == 'column' && this.m_colHeader != null && this.m_colHeader['firstChild'] != null)
    {
        headers = this.m_colHeader['firstChild']['childNodes'];
    }
    else if (axis == 'row' && this.m_rowHeader != null && this.m_rowHeader['firstChild'] != null)
    {
        headers = this.m_rowHeader['firstChild']['childNodes'];
    }
    else if (axis == 'columnEnd' && this.m_colEndHeader != null && this.m_colEndHeader['firstChild'] != null)
    {
        headers = this.m_colEndHeader['firstChild']['childNodes'];
    }
    else if (axis == 'rowEnd' && this.m_rowEndHeader != null && this.m_rowEndHeader['firstChild'] != null)
    {
        headers = this.m_rowEndHeader['firstChild']['childNodes'];
    }

    if (headers != null)
    {
        for (i = 0; i < headers.length; i++)
        {
            header = headers[i];
            headerContext = header[this.getResources().getMappedAttribute('context')];
            headerContext['index'] = this.getHeaderCellIndex(header);

            if (option == 'resizable')
            {
                if (this._isHeaderResizeEnabled(axis, headerContext))
                {
                    this._setAttribute(header, option, 'true');
                }
                else
                {
                    this._setAttribute(header, option, 'false');
                }
            }
            else if (option == 'sortable')
            {
                hasSortContainer = this.m_utils.containsCSSClassName(header['lastChild'], this.getMappedStyle('sortcontainer'));
                if (this._isSortEnabled(axis, headerContext))
                {
                    if (!hasSortContainer)
                    {
                        sortIcon = this._buildSortIcon(headerContext);
                        header.appendChild(sortIcon); //@HTMLUpdateOK
                    }
                    this._setAttribute(header, option, 'true');
                }
                else
                {
                    if (hasSortContainer)
                    {
                        this._remove(header['lastChild']);
                    }
                    this._setAttribute(header, option, 'false');
                }
            }
        }
    }
};

/**
 * Update gridlines
 * @private
 */
DvtDataGrid.prototype._updateGridlines = function()
{
    var i, j, horizontalGridlines, verticalGridlines, dir, cell;

    horizontalGridlines = this.m_options.getHorizontalGridlines();
    verticalGridlines = this.m_options.getVerticalGridlines();
    dir = this.getResources().isRTLMode() ? "right" : "left";
    // i and j are absolute index
    for (i = this.m_startRow; i <= this.m_endRow; i++)
    {
        for (j = this.m_startCol; j <= this.m_endCol; j++)
        {
            cell = this._getCellByIndex(this.createIndex(i, j));
            if (verticalGridlines === 'hidden' || (this._isLastColumn(j) && this.getRowHeaderWidth() + this.getElementDir(cell, dir) + this.calculateColumnWidth(cell) >= this.getWidth()))
            {
                this.m_utils.addCSSClassName(cell, this.getMappedStyle('borderVerticalNone'));
            }
            else
            {
                this.m_utils.removeCSSClassName(cell, this.getMappedStyle('borderVerticalNone'));
            }

            if (horizontalGridlines === 'hidden' || (this._isLastRow(i) && this.getRowBottom(cell, null) >= this.getHeight()))
            {
                this.m_utils.addCSSClassName(cell, this.getMappedStyle('borderHorizontalNone'));
            }
            else
            {
                this.m_utils.removeCSSClassName(cell, this.getMappedStyle('borderHorizontalNone'));
            }
        }
    }
};

/**
 * Updates the borders of cells along the grid edges for focus ring spacing to be appropriated correctly
 * @param {string=} activeValue current border value
 * @param {string=} prevValue previous border value
 */
DvtDataGrid.prototype._updateEdgeCellBorders = function(activeValue, prevValue)
{
    var activeCell, prevActiveCell;
    if (this.m_active != null && this.m_active['type'] != 'header')
    {
        activeCell = this._getActiveElement();
        if (activeCell != null)
        {
            if (this._isLastRow(this.m_active['indexes']['row']))
            {
                if (activeValue == 'none')
                {
                    this.m_utils.addCSSClassName(activeCell, this.getMappedStyle('borderHorizontalNone'));
                }
                else
                {
                    this.m_utils.removeCSSClassName(activeCell, this.getMappedStyle('borderHorizontalNone'));
                }
            }

            if (this._isLastColumn(this.m_active['indexes']['column']))
            {
                if (activeValue == 'none')
                {
                    this.m_utils.addCSSClassName(activeCell, this.getMappedStyle('borderVerticalNone'));
                }
                else
                {
                    this.m_utils.removeCSSClassName(activeCell, this.getMappedStyle('borderVerticalNone'));
                }
            }
        }
    }

    if (prevValue != null && this.m_prevActive != null && this.m_prevActive['type'] != 'header')
    {
        prevActiveCell = this._getElementFromActiveObject(this.m_prevActive);
        if (prevActiveCell != null)
        {
            if (this._isLastRow(this.m_prevActive['indexes']['row']))
            {
                if (prevValue == 'none')
                {
                    this.m_utils.addCSSClassName(prevActiveCell, this.getMappedStyle('borderHorizontalNone'));
                }
                else
                {
                    this.m_utils.removeCSSClassName(prevActiveCell, this.getMappedStyle('borderHorizontalNone'));
                }
            }
            if (this._isLastColumn(this.m_prevActive['indexes']['column']))
            {
                if (prevValue == 'none')
                {
                    this.m_utils.addCSSClassName(prevActiveCell, this.getMappedStyle('borderVerticalNone'));
                }
                else
                {
                    this.m_utils.removeCSSClassName(prevActiveCell, this.getMappedStyle('borderVerticalNone'));
                }
            }
        }
    }
};

/**
 * Gets options on DataGrid
 * @return {Object} the options set on the data grid
 */
DvtDataGrid.prototype.GetOptions = function()
{
    if (this.m_options == null)
    {
        this.m_options = new DvtDataGridOptions();
    }

    return this.m_options;
};

/**
 * Sets resources on DataGrid
 * @param {Object} resources - the resources to set on the data grid
 */
DvtDataGrid.prototype.SetResources = function(resources)
{
    this.m_resources = resources;
};

/**
 * Gets resources from DataGrid
 * @return {Object} the resources set on the data grid
 */
DvtDataGrid.prototype.getResources = function()
{
    return this.m_resources;
};

/**
 * Gets start row index from DataGrid
 * @return {number} the start row index
 */
DvtDataGrid.prototype.getStartRow = function()
{
    return this.m_startRow;
};

/**
 * Gets start row header index from DataGrid
 * @return {number} the start row header index
 */
DvtDataGrid.prototype.getStartRowHeader = function()
{
    return this.m_startRowHeader;
};

/**
 * Gets start column index from DataGrid
 * @return {number} the start column index
 */
DvtDataGrid.prototype.getStartColumn = function()
{
    return this.m_startCol;
};

/**
 * Gets start column header index from DataGrid
 * @return {number} the start column header index
 */
DvtDataGrid.prototype.getStartColumnHeader = function()
{
    return this.m_startColHeader;
};

/**
 * Gets start row end header index from DataGrid
 * @return {number} the start row end header index
 */
DvtDataGrid.prototype.getStartRowEndHeader = function()
{
    return this.m_startRowEndHeader;
};

/**
 * Gets start column end header index from DataGrid
 * @return {number} the start column end header index
 */
DvtDataGrid.prototype.getStartColumnEndHeader = function()
{
    return this.m_startColEndHeader;
};

/**
 * Gets mapped style from the resources
 * @private
 * @param {string} key the key to get style on
 * @return {string} the style from the resources
 */
DvtDataGrid.prototype.getMappedStyle = function(key)
{
    return this.getResources().getMappedStyle(key);
};

/**
 * Sets the data source on DataGrid
 * @param {Object} dataSource - the data source to set on the data grid
 */
DvtDataGrid.prototype.SetDataSource = function(dataSource)
{
    // if we are setting a new data source be sure to clear out any old
    // model events
    this.m_modelEvents = [];

    this.m_dataSource = dataSource;
};

/**
 * Gets the data source from the DataGrid
 * @return {Object} the data source set on the data grid
 */
DvtDataGrid.prototype.getDataSource = function()
{
    return this.m_dataSource;
};

/**
 * Set the internal visibility of datagrid
 * @param {string} state a string for the visibility
 */
DvtDataGrid.prototype.setVisibility = function(state)
{
    this.m_visibility = state;
};

/**
 * Get the internal visibility of datagrid
 * @return {string} visibility
 */
DvtDataGrid.prototype.getVisibility = function()
{
    if (this.m_visibility == null)
    {
        if (this.m_root.offsetParent != null)
        {
            this.setVisibility(DvtDataGrid.VISIBILITY_STATE_VISIBLE);
        }
        else
        {
            this.setVisibility(DvtDataGrid.VISIBILITY_STATE_HIDDEN);
        }
    }
    return this.m_visibility;
};

/**
 * Set the callback for remove
 * @param {Function} callback a callback for the remove function
 */
DvtDataGrid.prototype.SetOptionCallback = function(callback)
{
    this.m_setOptionCallback = callback;
};

/**
 * Set the callback for remove
 * @param {Function} callback a callback for the remove function
 */
DvtDataGrid.prototype.SetContextCallback = function(callback)
{
    this.m_contextCallback = callback;
};

/**
 * Set the callback for remove
 * @param {Function} callback a callback for the remove function
 */
DvtDataGrid.prototype.SetRemoveCallback = function(callback)
{
    this.m_removeCallback = callback;
};

/**
 * Set the callback for add or remove of id
 * @param {Function} callback a callback for the unique id function
 */
DvtDataGrid.prototype.SetUniqueIdCallback = function(callback)
{
    this._uniqueIdCallback = callback;
};

/**
 * Set the callback for compare values
 * @param {Function} callback a callback for the compare values function
 */
DvtDataGrid.prototype.SetCompareValuesCallback = function(callback)
{
    this._compareValuesCallback = callback;
};

/**
 * Set the callback for subtreeAttached that should be called when adding content to the dom
 * @param {Function} callback a callback for the subtree attached function
 */
DvtDataGrid.prototype.SetSubtreeAttachedCallback = function(callback)
{
    this.m_subtreeAttachedCallback = callback;
};

/**
 * Remove an element from the DOM, if it is not being reattached
 * @param {Element} element the element to remove
 */
DvtDataGrid.prototype._remove = function(element)
{
    this._uniqueIdCallback(element, true);

    // check if there is a callback set for remove
    // callback allows jQuery to clean the node on a remove
    if (this.m_removeCallback != null)
    {
        this.m_removeCallback.call(null, element);
    }
    else
    {
        element['parentNode'].removeChild(element);
    }
};

/**
 * Remove all elements in an array of elements
 * @param {Array} elems an array of the elements that need to be removed
 */
DvtDataGrid.prototype._removeFromArray = function(elems)
{
    for (var i = 0; i < elems.length; i++)
    {
        this._remove(elems[i]);
    }
};

/**
 * Set the callback for signifying not ready
 * @param {Function} callback a callback for the not ready function
 */
DvtDataGrid.prototype.SetNotReadyCallback = function(callback)
{
    this.m_notReady = callback;
};

/**
 * Set the callback for signifying ready
 * @param {Function} callback a callback for the make ready function
 */
DvtDataGrid.prototype.SetMakeReadyCallback = function(callback)
{
    this.m_makeReady = callback;
};

/**
  * Invoke whenever a task is started. Moves the component out of the ready state if necessary.
  */
DvtDataGrid.prototype._signalTaskStart = function()
{
    if (this.m_readinessStack)
    {
        if (this.m_readinessStack.length == 0)
        {
            this.m_notReady();
        }
        this.m_readinessStack.push(1);
    }
};

/**
 * Invoke whenever a task finishes. Resolves the readyPromise if component is ready to move into ready state.
 */
DvtDataGrid.prototype._signalTaskEnd = function()
{
    if (this.m_readinessStack && this.m_readinessStack.length > 0)
    {
        this.m_readinessStack.pop();
        if (this.m_readinessStack.length == 0)
        {
            this.m_makeReady();
        }
    }
};

/**
 * Get the indexes from the data source and call back to a function once they are available.
 * The callback should be a function(keys, indexes)
 * @param {Object} keys the keys to find the index of with properties row, column
 * @param {Function} callback the callback to pass the keys back to
 * @private
 */
DvtDataGrid.prototype._indexes = function(keys, callback)
{
    var self = this, indexes;
    indexes = this.getDataSource().indexes(keys);
    if (typeof indexes['then'] === 'function')
    {
        // start async indexes call
        self._signalTaskStart();
        indexes.then(function(obj) {
            callback.call(self, obj, keys);
            // end async indexes call
            self._signalTaskEnd();
        }, function()
        {
            callback.call(self, {'row': -1, 'column': -1}, keys);
            // end async indexes call
            self._signalTaskEnd();
        });
    }
    else
    {
        callback.call(self, indexes, keys);
    }
};

/**
 * Get the keys from the data source and call back to a function once they are available.
 * The callback should be a function(indexes, keys)
 * @param {Object} indexes the indexes to find the keys of with properties row, column
 * @param {Function} callback the callback to pass the indexes back to
 * @private
 */
DvtDataGrid.prototype._keys = function(indexes, callback)
{
    var self = this, keys;
    keys = this.getDataSource().keys(indexes);
    if (typeof keys['then'] === 'function')
    {
        // start async call
        self._signalTaskStart();
        keys.then(function(obj) {
            callback.call(self, obj, indexes);
            // end async indexes call
            self._signalTaskEnd();
        }, function()
        {
            callback.call(self, {'row': null, 'column': null}, indexes);
            // end async indexes call
            self._signalTaskEnd();
        });
    }
    else
    {
        callback.call(self, keys, indexes);
    }
};

/**
 * Register a callback when creating the header context or cell context.
 * @param {function(Object)} callback the callback function to inject addition or modify
 *        properties in the context.
 */
DvtDataGrid.prototype.SetCreateContextCallback = function(callback)
{
    this.m_createContextCallback = callback;
};

/**
 * Register the focusable callbacks for handling focus classNames
 * @param {function()} focusInHandler
 * @param {function()} focusOutHandler
 */
DvtDataGrid.prototype.SetFocusableCallback =  function(focusInHandler, focusOutHandler)
{
    this.m_focusInHandler = focusInHandler;
    this.m_focusOutHandler = focusOutHandler;
};

/**
 * Register a callback when creating the header context or cell context.
 * @param {function(Object)} callback the callback function to inject addition or modify
 *        properties in the context.
 */
DvtDataGrid.prototype.SetFixContextCallback = function(callback)
{
    this.m_fixContextCallback = callback;
};

/**
 * Whether high watermark scrolling is used
 * @return {boolean} true if high watermark scrolling is used, false otherwise
 * @private
 */
DvtDataGrid.prototype._isHighWatermarkScrolling = function()
{
    return (this.m_options.getScrollPolicy() != "scroll");
};

/**
 * Destructor method that should be called when the widget is destroyed. Removes event
 * listeners on the document.
 */
DvtDataGrid.prototype.destroy = function()
{
    delete this.m_fetching;
    this._removeDataSourceEventListeners();
    this._removeDomEventListeners();
    delete this.m_styleClassDimensionMap;
    this.m_styleClassDimensionMap = {'width':{}, 'height':{}};
};

/**
 * Adds data source event listeners
 * @private
 */
DvtDataGrid.prototype._addDataSourceEventListeners = function()
{
    this._removeDataSourceEventListeners();

    if (this.m_dataSource != null)
    {
        this.m_handleModelEventListener = this.handleModelEvent.bind(this);
        this.m_handleExpandEventListener = this.handleExpandEvent.bind(this);
        this.m_handleCollapseEventListener = this.handleCollapseEvent.bind(this);

        this.m_dataSource.on("change", this.m_handleModelEventListener, this);
        // if it's not flattened datasource, these will be ignored
        this.m_dataSource.on("expand", this.m_handleExpandEventListener, this);
        this.m_dataSource.on("collapse", this.m_handleCollapseEventListener, this);
    }
};

/**
 * Remove data source event listeners
 * @private
 */
DvtDataGrid.prototype._removeDataSourceEventListeners = function()
{
    if (this.m_dataSource != null)
    {
        this.m_dataSource.off("change", this.m_handleModelEventListener);
        this.m_dataSource.off("expand", this.m_handleExpandEventListener);
        this.m_dataSource.off("collapse", this.m_handleCollapseEventListener);
    }
};

/**
 * Adds event listeners registered on the document or the root element
 * @private
 */
DvtDataGrid.prototype._addDomEventListeners = function()
{
    if (!this.m_handleDatabodyKeyDown)
    {
        this.m_handleDatabodyKeyDown = this.handleDatabodyKeyDown.bind(this);
    }
    if (!this.m_handleRootFocus)
    {
        this.m_handleRootFocus = this.handleRootFocus.bind(this);
    }
    if (!this.m_handleRootBlur)
    {
        this.m_handleRootBlur = this.handleRootBlur.bind(this);
    }

    this.m_root.addEventListener("keydown", this.m_handleDatabodyKeyDown, false);
    this.m_root.addEventListener('focus', this.m_handleRootFocus, true);
    this.m_root.addEventListener('blur', this.m_handleRootBlur, true);
};

/**
 * Remove dom event listeners
 * @private
 */
DvtDataGrid.prototype._removeDomEventListeners = function()
{
    document.removeEventListener("mousemove", this.m_docMouseMoveListener, false);
    document.removeEventListener("mouseup", this.m_docMouseUpListener, false);
    // unregister all listeners

    if (this.m_root != null)
    {
        if (this.m_handleDatabodyKeyDown)
        {
            this.m_root.removeEventListener("keydown", this.m_handleDatabodyKeyDown, false);
        }
        if (this.m_handleRootFocus)
        {
            this.m_root.removeEventListener("focus", this.m_handleRootFocus, true);
        }
        if (this.m_handleRootBlur)
        {
            this.m_root.removeEventListener("blur", this.m_handleRootBlur, true);
        }
    }
};

/**
 * Get the DataGrid root element
 * @return {Element} the root element
 */
DvtDataGrid.prototype.getRootElement = function()
{
    return this.m_root;
};

/**
 * Get the cached width of the root element. If not cached, sets the cached width.
 * @return {number} the cached width of the root element
 * @protected
 */
DvtDataGrid.prototype.getWidth = function()
{
    if (this.m_width == null)
    {
        //clientWidth since we use border box and care about the content of our root div
        this.m_width = this.getRootElement()['clientWidth'];
    }

    return this.m_width;
};

/**
 * Get the cached height of the root element. If not cached, sets the cached height.
 * @return {number} the cached height of the root element
 * @protected
 */
DvtDataGrid.prototype.getHeight = function()
{
    if (this.m_height == null)
    {
        //clientHeight since we use border box and care about the content of our root div
        this.m_height = this.getRootElement()['clientHeight'];
    }

    return this.m_height;
};

/**
 * Gets the scrollable width of the grid.
 * @return {number} the scrollable width of the grid
 * @protected
 */
DvtDataGrid.prototype.getScrollableWidth = function()
{
    var scrollerContent = this.m_databody['firstChild'];
    return this.getElementWidth(scrollerContent);
};

/**
 * Get the viewport width, which is defined as 1.5 the size of the width of Grid
 * @return {number} the viewport width
 */
DvtDataGrid.prototype.getViewportWidth = function()
{
    var width = this.getWidth();
    return Math.round(width * 1.5);
};

/**
 * Get the viewport height, which is defined as 1.5 the size of the height of Grid
 * @return {number} the viewport height
 */
DvtDataGrid.prototype.getViewportHeight = function()
{
    var height = this.getHeight();
    return Math.round(height * 1.5);
};

/**
 * Get viewport top
 * @return {number} the viewport top
 */
DvtDataGrid.prototype._getViewportTop = function()
{
    return this.m_currentScrollTop;
};

/**
 * Get viewport bottom
 * @return {number} the viewport bottom
 */
DvtDataGrid.prototype._getViewportBottom = function()
{
    var top = this._getViewportTop();
    var databodyHeight = this.getElementHeight(this.m_databody);
    var scrollbarSize = this.m_utils.getScrollbarSize();

    return this.m_hasHorizontalScroller ? top + databodyHeight - scrollbarSize : top + databodyHeight;
};

/**
 * Get viewport left
 * @return {number} the viewport left
 */
DvtDataGrid.prototype._getViewportLeft = function()
{
    return this.m_currentScrollLeft;
};

/**
 * Get viewport right
 * @return {number} the viewport right
 */
DvtDataGrid.prototype._getViewportRight = function()
{
    var left = this._getViewportLeft();
    var databodyWidth = this.getElementWidth(this.m_databody);
    var scrollbarSize = this.m_utils.getScrollbarSize();

    return this.m_hasVerticalScroller ? left + databodyWidth - scrollbarSize : left + databodyWidth;
};

/**
 * Calculate the fetch size for rows or columns
 * @param {string} axis - the axis 'row'/'column' to get fetch size on
 * @return {number} the fetch size
 */
DvtDataGrid.prototype.getFetchSize = function(axis)
{
    // get the cached fetch size, this should be clear when the size changes
    if (axis == "row")
    {
        if (this.m_rowFetchSize == null)
        {
            this.m_rowFetchSize = Math.max(1, Math.round(this.getViewportHeight() / this.getDefaultRowHeight()));
        }

        return this.m_rowFetchSize;
    }
    if (axis == "column")
    {
        if (this.m_columnFetchSize == null)
        {
            this.m_columnFetchSize = Math.max(1, Math.round(this.getViewportWidth() / this.getDefaultColumnWidth()));
        }
        return this.m_columnFetchSize;
    }

    return 0;
};

/**
 * If the empty text option is 'default' return default empty translated text,
 * otherwise return the emptyText set in the options
 * @return {string} the empty text
 */
DvtDataGrid.prototype.getEmptyText = function()
{
    var emptyText, resources;
    emptyText = this.m_options.getEmptyText();
    if (emptyText == null)
    {
        resources = this.getResources();
        emptyText = resources.getTranslatedText("msgNoData");
    }
    return emptyText;
};

/**
 * Build an empty text div and populate it with empty text
 * @return {Element} the empty text element
 * @private
 */
DvtDataGrid.prototype._buildEmptyText = function()
{
    var emptyText, empty, height, dir, left;
    emptyText = this.getEmptyText();
    empty = document.createElement("div");
    empty['id'] = this.createSubId("empty");
    empty['className'] = this.getMappedStyle("emptytext");
    height = this.m_endColHeader >= 0 ? this.getColumnHeaderHeight() : 0;
    this.setElementDir(empty, height, 'top');
    dir = this.getResources().isRTLMode() ? "right" : "left";
    left = this.m_endRowHeader >= 0 ? this.getRowHeaderWidth() : 0;
    this.setElementDir(empty, left, dir);
    empty.textContent = emptyText;
    this.m_empty = empty;
    return  empty;
};

/**
 * Determine the size of the buffer that triggers fetching of rows. For example,
 * if the size of the buffer is zero, then the fetch will be triggered when the
 * scroll position reached the end of where the current range ends
 * @return {number} the row threshold
 */
DvtDataGrid.prototype.getRowThreshold = function()
{
    return 0;
};

/**
 * Determine the size of the buffer that triggers fetching of columns. For example,
 * if the size of the buffer is zero, then the fetch will be triggered when the
 * scroll position reached the end of where the current range ends.
 * @return {number} the column threshold
 */
DvtDataGrid.prototype.getColumnThreshold = function()
{
    return 0;
};


/*
 * Caches the default datagrid dimensions located in the style sheet, creates
 * just one div to reduce createElement calls. This function should get called once on create.
 * Values found in style are:
 *  column width
 *  row height
 */
DvtDataGrid.prototype.setDefaultDimensions = function()
{
    var div, resources;
    div = document.createElement('div');
    div['style']['visibilty'] = 'hidden';

    resources = this.getResources();
    // we can avoid a repaint by using both row and headercell here because this isn't where the col height and row width are set
    div['className'] = resources.getMappedStyle("rowheadercell") + " " + resources.getMappedStyle("colheadercell") + " " + resources.getMappedStyle("headercell");
    this.m_root.appendChild(div); //@HTMLUpdateOK
    // Not using offsetWidth/Height due to 
    this.m_defaultColumnWidth = Math.round(div.getBoundingClientRect()['width']);
    this.m_defaultRowHeight = Math.round(div.getBoundingClientRect()['height']);

    this.m_root.removeChild(div);
};

/**
 * Get the default row height which comes from the style sheet
 * @return {number} the default row height
 */
DvtDataGrid.prototype.getDefaultRowHeight = function()
{
    if (this.m_defaultRowHeight == null)
    {
        this.setDefaultDimensions();
    }
    return this.m_defaultRowHeight;
};

/**
 * Get the default column width which comes from the stylesheet
 * @return {number} the default column width
 */
DvtDataGrid.prototype.getDefaultColumnWidth = function()
{
    if (this.m_defaultColumnWidth == null)
    {
        this.setDefaultDimensions();
    }
    return this.m_defaultColumnWidth;
};

/**
 * Get the default dimension on an axis
 * @param {string} axis
 * @returns {number|undefined}
 */
DvtDataGrid.prototype._getDefaultDimension = function(axis)
{
    if (axis == 'row')
    {
         return this.getDefaultRowHeight();
    }
    else if (axis == 'column')
    {
        return this.getDefaultColumnWidth();
    }
    return;
};

/**
 * Gets the header dimension for an axis, for rows this would be height, for columns, width
 * @param {Element} elem the header element to get dimension of
 * @param {string|null} key the row or column key
 * @param {string} axis row or column
 * @param {string} dimension width ro height
 * @returns {number}
 */
DvtDataGrid.prototype._getHeaderDimension = function(elem, key, axis, dimension)
{
    var value, className;

    value = this.m_sizingManager.getSize(axis, key);
    if (value != null)
    {
        return value;
    }

    // check if inline style set on element
    if (elem['style'][dimension] != '')
    {
        value = this.getElementDir(elem, dimension);
        //in the event that row height is set via an additional style only on row header store the value
        this.m_sizingManager.setSize(axis, key, value);
        return value;
    }

    // check style class mapping, mapping prevents multiple offsetHeight calls on headers with the same class name
    className = elem['className'];
    value = this.m_styleClassDimensionMap[dimension][className];
    if (value == null)
    {
        // exhausted all options, use offsetHeight then, remove element in the case of shim element
        value = this.getElementDir(elem, dimension);
    }

    //the value isn't the default the cell will use meaning it's from an external
    //class, so store it in the sizing manager cell can pick it up, header and cell dimension can vary on em
    this.m_sizingManager.setSize(axis, key, value);

    this.m_styleClassDimensionMap[dimension][className] = value;
    return value;
};

/**
 * Helper method to create subid based on the root element's id
 * @param {string} subId - the id to append to the root element id
 * @return {string} the subId to append to the root element id
 */
DvtDataGrid.prototype.createSubId = function(subId)
{
    var id = this.getRootElement()['id'];
    if (id == null)
    {
        id = "";
    }

    return [id, subId].join(":");
};

/**
 * Checks whether header fetching is completed
 * @return {boolean} true if header fetching completed, else false
 */
DvtDataGrid.prototype.isHeaderFetchComplete = function()
{
    return (this.m_fetching['row'] === false && this.m_fetching['column'] === false);
};

/**
 * Checks whether header AND cell fetching is completed
 * @return {boolean} true if header AND cell fetching completed, else false
 */
DvtDataGrid.prototype.isFetchComplete = function()
{
    return (this.m_fetching != null && this.isHeaderFetchComplete() && this.m_fetching['cells'] === false);
};

/**
 * Checks whether the index is the last row
 * @param {number} index
 * @return {boolean} true if it's the last row, false otherwise
 */
DvtDataGrid.prototype._isLastRow = function(index)
{
    if (this._isCountUnknown("row"))
    {
        // if row count is unknown, then the last row is if the index is before the last row fetched
        // and there's no more rows from datasource
        return (index === this.m_endRow && this.m_stopRowFetch);
    }
    else
    {
        // if column count is known, then just check the index with the total column count
        return (index + 1 === this.getDataSource().getCount("row"));
    }
};

/**
 * Checks whether the index is the last column
 * @param {number} index
 * @return {boolean} true if it's the last column, false otherwise
 */
DvtDataGrid.prototype._isLastColumn = function(index)
{
    if (this._isCountUnknown("column"))
    {
        // if column count is unknown, then the last column is if the index is the last column fetched
        // and there's no more columns from datasource
        return (index === this.m_endCol && this.m_stopColumnFetch);
    }
    else
    {
        // if column count is known, then just check the index with the total column count
        return (index + 1 === this.getDataSource().getCount("column"));
    }
};

/**
 * Removes all of the datagrid children built by DvtDataGrid, this excludes context menus/popups
 */
DvtDataGrid.prototype.empty = function()
{
    //remove everything that will be rebuilt
    if (this.m_empty)
    {
        this.m_root.removeChild(this.m_empty);
    }
    if (this.m_corner)
    {
        this.m_root.removeChild(this.m_corner);
    }
    if (this.m_bottomCorner)
    {
        this.m_root.removeChild(this.m_bottomCorner);
    }
    if (this.m_columnHeaderScrollbarSpacer)
    {
        this.m_root.removeChild(this.m_columnHeaderScrollbarSpacer);
    }
    if (this.m_rowHeaderScrollbarSpacer)
    {
        this.m_root.removeChild(this.m_rowHeaderScrollbarSpacer);
    }

    this.m_root.removeChild(this.m_placeHolder);
    this.m_root.removeChild(this.m_status);
    this.m_root.removeChild(this.m_accSummary);
    this.m_root.removeChild(this.m_accInfo);
    this.m_root.removeChild(this.m_stateInfo);
    this.m_root.removeChild(this.m_contextInfo);
    // elements that may contain other components
    this._remove(this.m_colHeader);
    this._remove(this.m_rowHeader);
    this._remove(this.m_colEndHeader);
    this._remove(this.m_rowEndHeader);
    this._remove(this.m_databody);
    this._clearDatabodyMap();
};

/**
 * Re-renders the data grid. Resets all the necessary properties.
 * @param {Element} root - the root dom element for the DataGrid.
 */
DvtDataGrid.prototype.refresh = function(root)
{
    this.resetInternal();
    this.render(root);
};

/**
 * Resets internal state of data grid.
 * @private
 */
DvtDataGrid.prototype.resetInternal = function()
{
    this.m_initialized = false;
    this.m_readinessStack = [];
    this._signalTaskStart();
    this._signalTaskEnd();

    //databody map
    this._clearDatabodyMap();

    //cursor
    this.m_cursor = null;

    //dom elements
    this.m_corner = null;
    this.m_bottomCorner = null;
    this.m_columnHeaderScrollbarSpacer = null;
    this.m_rowHeaderScrollbarSpacer = null;
    this.m_colHeader = null;
    this.m_colEndHeader = null;
    this.m_rowHeader = null;
    this.m_rowEndHeader = null;
    this.m_databody = null;
    this.m_empty = null;
    this.m_accInfo = null;
    this.m_accSummary = null;
    this.m_contextInfo = null;
    this.m_placeHolder = null;
    this.m_stateInfo = null;
    this.m_status = null;

    //fetching
    this.m_isEstimateRowCount = undefined;
    this.m_isEstimateColumnCount = undefined;
    this.m_stopRowFetch = false;
    this.m_stopRowHeaderFetch = false;
    this.m_stopRowEndHeaderFetch = false;
    this.m_stopColumnFetch = false;
    this.m_stopColumnHeaderFetch = false;
    this.m_stopColumnEndHeaderFetch = false;
    this.m_rowFetchSize = null;
    this.m_columnFetchSize = null;
    this.m_fetching = null;
    this.m_processingModelEvent = false;
    this.m_processingEventQueue = false;
    this.m_animating = false;

    //dimensions
    this.m_sizingManager.clear();
    this.m_styleClassDimensionMap = {'width':{}, 'height':{}};
    this.m_height = null;
    this.m_width = null;
    this.m_scrollHeight = null;
    this.m_scrollWidth = null;
    this.m_avgRowHeight = undefined;
    this.m_avgColWidth = undefined;
    this.m_defaultColumnWidth = null;
    this.m_defaultRowHeight = null;
    this.m_colHeaderHeight = null;
    this.m_colEndHeaderHeight = null;
    this.m_rowHeaderWidth = null;
    this.m_rowEndHeaderWidth = null;
    this.m_rowHeaderLevelWidths = [];
    this.m_rowEndHeaderLevelWidths = [];
    this.m_columnHeaderLevelHeights = [];
    this.m_columnEndHeaderLevelHeights = [];

    //active
    this.m_active = null;
    this.m_prevActive = null;
    this.m_trueIndex = {};

    //dnd
    this.m_headerDragState = false;
    this.m_databodyDragState = false;
    this.m_databodyMove = false;
    this.m_moveRow = null;
    this.m_moveRowHeader = null;
    this.m_dropTarget = null;
    this.m_dropTargetHeader = null;

    //selection
    this.m_discontiguousSelection = false;

    //event listeners
    this.m_docMouseMoveListener = null;
    this.m_docMouseUpListener = null;
    this.m_modelEvents = [];

    //scrolling
    this.m_hasHorizontalScroller = null;
    this.m_hasVerticalScroller = null;
    this.m_currentScrollLeft = null;
    this.m_currentScrollTop = null;
    this.m_prevScrollLeft = null;
    this.m_prevScrollTop = null;
    this.m_handleScrollOverflow = null;

    //resizing
    this.m_resizing = false;
    this.m_resizingElement = null;
    this.m_resizingElementMin = null;

    //data states
    this.m_startRow = null;
    this.m_startCol = null;
    this.m_endRow = null;
    this.m_endCol = null;
    this.m_startRowPixel = null;
    this.m_startColPixel = null;
    this.m_endRowPixel = null;
    this.m_endColPixel = null;
    this.m_startRowHeader = null;
    this.m_startColHeader = null;
    this.m_endRowHeader = null;
    this.m_endColHeader = null;
    this.m_startRowHeaderPixel = null;
    this.m_startColHeaderPixel = null;
    this.m_endRowHeaderPixel = null;
    this.m_endColHeaderPixel = null;
    this.m_rowHeaderLevelCount = null;
    this.m_columnHeaderLevelCount = null;
    this.m_startRowEndHeader = null;
    this.m_startColEndHeader = null;
    this.m_endRowEndHeader = null;
    this.m_endColEndHeader = null;
    this.m_startRowEndHeaderPixel = null;
    this.m_startColEndHeaderPixel = null;
    this.m_endRowEndHeaderPixel = null;
    this.m_endColEndHeaderPixel = null;
    this.m_rowEndHeaderLevelCount = null;
    this.m_columnEndHeaderLevelCount = null;
    this.m_sortInfo = null;
    this.m_resizeRequired = null;
    this.m_externalFocus = null;
    this.m_currentMode = null;
    this.m_editMode = null;

    this.m_hasCells = null;
    this.m_hasRowHeader = null;
    this.m_hasRowEndHeader = null;
    this.m_hasColHeader = null;
    this.m_hasColEndHeader = null;
    this.m_isLongScroll = null;

    this.m_addBorderBottom = null;
    this.m_addBorderRight = null;
};

/**
 * DataGrid should initialize if there's no outstanding fetch, it is unitialized
 * and the databody is attached to the root.
 * @private
 * @returns {boolean} true if we have the properties that signify an end to initialize
 */
DvtDataGrid.prototype._shouldInitialize = function()
{
    return (this.isFetchComplete() && !this.m_initialized && this.m_databody.parentNode != null);
};

/**
 * Handle the end of datagrid initialization whether at the end of rendering or fetching
 * @private
 * @param {boolean=} hasData false if there is no data and thus should skip resizing
 */
DvtDataGrid.prototype._handleInitialization = function(hasData)
{
    if (hasData == true)
    {
        this.resizeGrid();
        this.fillViewport();

        if (this.isFetchComplete())
        {
            this._updateActive(this.m_options.getCurrentCell(), false, false);
            this.m_initialized = true;
            this.fireEvent('ready', {});
            this._runModelEventQueue();
        }
    }
    else
    {
        this.m_initialized = true;
        this.fireEvent('ready', {});
        this._runModelEventQueue();
    }
};

/**
 * Run the events in the model event list
 * The queue shifts the first event and runs that.
 * The event is expected to call _runModelEventQueue 
 * once it completes
 * If the queue is empty, stop processing
 * Usage: After any animation related event chain 
 *        completes, run the event queue. 
 *        During the event queue, queued events
 *        should also have a call back to _runModelEventQueue
 *        at its completion. See handleExpandEvent for an example
 * @private
 */
DvtDataGrid.prototype._runModelEventQueue = function()
{
    var event;
    //Run the event queue generally after initialization 
    //or animations are complete. 
    //m_modelEvents acts as the queue and will
    //be initialized normally. In the case that
    //the m_modelEvent queue is accessed before 
    //it is initialized, it will act as length 0 queue
    if (this.m_modelEvents != null)
    {
        this.m_processingEventQueue = true;
        if (this.m_modelEvents.length === 0) {
            this.m_processingEventQueue = false;
            return;
        } 
        else 
        {
            event = this.m_modelEvents.shift();

            if (event['operation'] == 'expand')
            {
                this.handleExpandEvent(event, true);
            }
            else if (event['operation'] == 'collapse')
            {
                this.handleCollapseEvent(event, true);
            }
            else
            {
                this.handleModelEvent(event, true);
            }
        }
    }
    else
    {
        this.m_processingEventQueue = false;
    }
};

/**
 * Renders the DataGrid, initializes DataGrid properties.
 * @param {Element} root - the root dom element for the DataGrid.
 */
DvtDataGrid.prototype.render = function(root)
{
    // if this is the same instance and state wasn't clear out from last time
    if (this.m_databody != null)
    {
        this.destroy();
        this.resetInternal();
    }

    this.m_timingStart = new Date();
    this.m_fetching = {
    };

    // since headers and cells are independently fetched, they could be returned
    // at a different time, therefore we'll need var to keep track the current range
    // for each one of them
    this.m_startRow = 0;
    this.m_startCol = 0;
    this.m_endRow = -1;
    this.m_endCol = -1;
    this.m_startRowPixel = 0;
    this.m_startColPixel = 0;
    this.m_endRowPixel = 0;
    this.m_endColPixel = 0;

    this.m_startRowHeader = 0;
    this.m_startColHeader = 0;
    this.m_endRowHeader = -1;
    this.m_endColHeader = -1;
    this.m_startRowHeaderPixel = 0;
    this.m_startColHeaderPixel = 0;
    this.m_endRowHeaderPixel = 0;
    this.m_endColHeaderPixel = 0;

    this.m_startRowEndHeader = 0;
    this.m_startColEndHeader = 0;
    this.m_endRowEndHeader = -1;
    this.m_endColEndHeader = -1;
    this.m_startRowEndHeaderPixel = 0;
    this.m_startColEndHeaderPixel = 0;
    this.m_endRowEndHeaderPixel = 0;
    this.m_endColEndHeaderPixel = 0;

    this.m_currentScrollLeft = 0;
    this.m_currentScrollTop = 0;
    this.m_prevScrollLeft = 0;
    this.m_prevScrollTop = 0;
    this.m_handleScrollOverflow = false;

    this.m_rowHeaderLevelWidths = [];
    this.m_rowEndHeaderLevelWidths = [];
    this.m_columnHeaderLevelHeights = [];
    this.m_columnEndHeaderLevelHeights = [];

    this.buildGrid(root);
};

/**
 * Builds the DataGrid, adds root children (headers, databody, corners),
 * initializes event listeners, and sets inital scroll position.
 * @param {Element} root - the root dom element for the DataGrid.
 */
DvtDataGrid.prototype.buildGrid = function(root)
{
    var status, accSummary, accInfo, stateInfo, rtl, colHeader, rowHeader,
            databody, empty, contextInfo, placeHolder, colEndHeader, rowEndHeader, returnObj, mousewheelEvent;
    this.m_root = root;
    //class name set on component create
    this.m_root.setAttribute("role", "application");
    if (this._isCellEditable())
    {
        this.m_utils.addCSSClassName(this.m_root, this.getMappedStyle('editable'));
    }
    else
    {
        this.m_utils.addCSSClassName(this.m_root, this.getMappedStyle('readOnly'));
    }
    //this.m_root.setAttribute("aria-describedby", this.createSubId("summary"));

    // set a tab index so it can be focus and keyboard navigation to work
    root.tabIndex = 0;

    status = this.buildStatus();
    root.appendChild(status); //@HTMLUpdateOK
    this.m_status = status;

    accSummary = this.buildAccSummary();
    root.appendChild(accSummary); //@HTMLUpdateOK
    this.m_accSummary = accSummary;

    accInfo = this.buildAccInfo();
    root.appendChild(accInfo); //@HTMLUpdateOK
    this.m_accInfo = accInfo;

    stateInfo = this.buildStateInfo();
    root.appendChild(stateInfo); //@HTMLUpdateOK
    this.m_stateInfo = stateInfo;

    contextInfo = this.buildContextInfo();
    root.appendChild(contextInfo); //@HTMLUpdateOK
    this.m_contextInfo = contextInfo;

    placeHolder = this.buildPlaceHolder();
    root.appendChild(placeHolder); //@HTMLUpdateOK
    this.m_placeHolder = placeHolder;

    if (this.getDataSource() != null)
    {
        //in the event that the empty text was set when there was no datasource
        this.m_empty = null;

        rtl = this.getResources().isRTLMode();

        returnObj = this.buildHeaders("column", this.getMappedStyle("colheader"), this.getMappedStyle("colendheader"));
        colHeader = returnObj.root;
        colEndHeader = returnObj.endRoot;
        root.insertBefore(colHeader, status); //@HTMLUpdateOK
        root.insertBefore(colEndHeader, status); //@HTMLUpdateOK

        returnObj = this.buildHeaders("row", this.getMappedStyle("rowheader"), this.getMappedStyle("rowendheader"));
        rowHeader = returnObj.root;
        rowEndHeader = returnObj.endRoot;
        root.insertBefore(rowHeader, status); //@HTMLUpdateOK
        root.insertBefore(rowEndHeader, status); //@HTMLUpdateOK

        databody = this.buildDatabody();
        root.insertBefore(databody, status); //@HTMLUpdateOK

        if (rtl)
        {
            colHeader['style']['direction'] = "rtl";
            databody['style']['direction'] = "rtl";
        }

        this.m_isResizing = false;
        this.m_resizingElement = null;
        this.m_resizingElementMin = null;
        this.m_databodyDragState = false;

        // store the listeners so we can remove them later (bind creates a new function)
        this.m_docMouseMoveListener = this.handleMouseMove.bind(this);
        this.m_docMouseUpListener = this.handleMouseUp.bind(this);
        
        //touch event handling
        if (this.m_utils.isTouchDevice())
        {
            //databody touch listeners
            databody.addEventListener("touchstart", this.handleTouchStart.bind(this), false);
            databody.addEventListener("touchmove", this.handleTouchMove.bind(this), false);
            databody.addEventListener("touchend", this.handleTouchEnd.bind(this), false);
            databody.addEventListener("touchcancel", this.handleTouchCancel.bind(this), false);

            //column header listeners
            colHeader.addEventListener("touchstart", this.handleHeaderTouchStart.bind(this), false);
            colHeader.addEventListener("touchmove", this.handleHeaderTouchMove.bind(this), false);
            colHeader.addEventListener("touchend", this.handleHeaderTouchEnd.bind(this), false);
            colHeader.addEventListener("touchcancel", this.handleHeaderTouchCancel.bind(this), false);

            //row header listeners
            rowHeader.addEventListener("touchstart", this.handleHeaderTouchStart.bind(this), false);
            rowHeader.addEventListener("touchmove", this.handleHeaderTouchMove.bind(this), false);
            rowHeader.addEventListener("touchend", this.handleHeaderTouchEnd.bind(this), false);
            rowHeader.addEventListener("touchcancel", this.handleHeaderTouchCancel.bind(this), false);

            //column end header listeners
            colEndHeader.addEventListener("touchstart", this.handleHeaderTouchStart.bind(this), false);
            colEndHeader.addEventListener("touchmove", this.handleHeaderTouchMove.bind(this), false);
            colEndHeader.addEventListener("touchend", this.handleHeaderTouchEnd.bind(this), false);
            colEndHeader.addEventListener("touchcancel", this.handleHeaderTouchCancel.bind(this), false);

            //row end header listeners
            rowEndHeader.addEventListener("touchstart", this.handleHeaderTouchStart.bind(this), false);
            rowEndHeader.addEventListener("touchmove", this.handleHeaderTouchMove.bind(this), false);
            rowEndHeader.addEventListener("touchend", this.handleHeaderTouchEnd.bind(this), false);
            rowEndHeader.addEventListener("touchcancel", this.handleHeaderTouchCancel.bind(this), false);
        }
        else
        {
            //non-touch event listening

            mousewheelEvent = this.m_utils.getMousewheelEvent();
            //databody listeners
            databody.addEventListener(mousewheelEvent, this.handleDatabodyMouseWheel.bind(this), false);
            databody.addEventListener("mousedown", this.handleDatabodyMouseDown.bind(this), false);
            databody.addEventListener("mousemove", this.handleDatabodyMouseMove.bind(this), false);
            databody.addEventListener("mouseup", this.handleDatabodyMouseUp.bind(this), false);
            databody.addEventListener("mouseout", this.handleDatabodyMouseOut.bind(this), false);
            databody.addEventListener("mouseover", this.handleDatabodyMouseOver.bind(this), false);
            databody.addEventListener("dblclick", this.handleDatabodyDoubleClick.bind(this), false);

            //header listeners
            rowHeader.addEventListener(mousewheelEvent, this.handleDatabodyMouseWheel.bind(this), false);
            rowHeader.addEventListener("mousedown", this.handleHeaderMouseDown.bind(this), false);
            colHeader.addEventListener("mousedown", this.handleHeaderMouseDown.bind(this), false);
            rowHeader.addEventListener("mouseover", this.handleHeaderMouseOver.bind(this), false);
            colHeader.addEventListener("mouseover", this.handleHeaderMouseOver.bind(this), false);
            rowHeader.addEventListener("mousemove", this.handleRowHeaderMouseMove.bind(this), false);
            colHeader.addEventListener("mousemove", this.handleColumnHeaderMouseMove.bind(this), false);
            rowHeader.addEventListener("mouseup", this.handleHeaderMouseUp.bind(this), false);
            colHeader.addEventListener("mouseup", this.handleHeaderMouseUp.bind(this), false);
            rowHeader.addEventListener("mouseout", this.handleHeaderMouseOut.bind(this), false);
            colHeader.addEventListener("mouseout", this.handleHeaderMouseOut.bind(this), false);
            rowHeader.addEventListener("click", this.handleHeaderClick.bind(this), false);
            colHeader.addEventListener("click", this.handleHeaderClick.bind(this), false);

            //end header listeners
            rowEndHeader.addEventListener(mousewheelEvent, this.handleDatabodyMouseWheel.bind(this), false);
            rowEndHeader.addEventListener("mousedown", this.handleHeaderMouseDown.bind(this), false);
            colEndHeader.addEventListener("mousedown", this.handleHeaderMouseDown.bind(this), false);
            rowEndHeader.addEventListener("mouseover", this.handleHeaderMouseOver.bind(this), false);
            colEndHeader.addEventListener("mouseover", this.handleHeaderMouseOver.bind(this), false);
            rowEndHeader.addEventListener("mousemove", this.handleRowHeaderMouseMove.bind(this), false);
            colEndHeader.addEventListener("mousemove", this.handleColumnHeaderMouseMove.bind(this), false);
            rowEndHeader.addEventListener("mouseup", this.handleHeaderMouseUp.bind(this), false);
            colEndHeader.addEventListener("mouseup", this.handleHeaderMouseUp.bind(this), false);
            rowEndHeader.addEventListener("mouseout", this.handleHeaderMouseOut.bind(this), false);
            colEndHeader.addEventListener("mouseout", this.handleHeaderMouseOut.bind(this), false);
            rowEndHeader.addEventListener("click", this.handleHeaderClick.bind(this), false);
            colEndHeader.addEventListener("click", this.handleHeaderClick.bind(this), false);
        }

        // check if data is fetched and size the grid
        if (this._shouldInitialize())
        {
            this._handleInitialization(true);
        }
    }
    else
    {
        //if no datasource render empty text
        empty = this._buildEmptyText();
        this.m_root.appendChild(empty); //@HTMLUpdateOK
        this._handleInitialization(false);
    }
};

/**
 * Handle resize of grid to a new width and height.
 * @param {number} width the new width
 * @param {number} height the new height
 */
DvtDataGrid.prototype.HandleResize = function(width, height)
{
    //can either get the client width or subtract the border width.
    width = this.getRootElement()['clientWidth'];
    height = this.getRootElement()['clientHeight'];
    // don't do anything if nothing has changed
    if (width != this.m_width || height != this.m_height)
    {
        // assign new width and height
        this.m_width = width;
        this.m_height = height;

        this.m_rowFetchSize = null;
        this.m_columnFetchSize = null;

        // if it's not initialize (or fetching), then just skip now
        // handleCellsFetchSuccess will attempt to fill the viewport
        if (this.m_initialized)
        {
            // call resize logic
            this.resizeGrid();
            if (this.isFetchComplete())
            {
                this.m_resizeRequired = true;
                // check viewport
                this.fillViewport();
            }
        }
    }
};

/**
 * Size the headers, scroller, databody based on current width and height.
 * @private
 */
DvtDataGrid.prototype.resizeGrid = function()
{
    var width, height, colHeader, rowHeader, databody, columnHeaderWidth,
            colHeaderHeight, rowHeaderWidth, rowHeaderHeight, databodyContentWidth, databodyWidth, databodyContentHeight, databodyHeight,
            isTouchDevice, isDatabodyHorizontalScrollbarRequired, isDatabodyVerticalScrollbarRequired, scrollbarSize,
            dir, empty, endTime, colEndHeader, rowEndHeader, colEndHeaderHeight, rowEndHeaderWidth, emptyHeight, emptyWidth,
            availableHeight, availableWidth, rowEndHeaderDir, columnEndHeaderDir, isEmpty, databodyScroller;


    width = this.getWidth();
    height = this.getHeight();
    colHeader = this.m_colHeader;
    colEndHeader = this.m_colEndHeader;
    rowHeader = this.m_rowHeader;
    rowEndHeader = this.m_rowEndHeader;
    databody = this.m_databody;
    databodyScroller = databody['firstChild'];

    // cache these since they will be used in multiple places and we want to minimize reflow
    colHeaderHeight = this.getColumnHeaderHeight();
    colEndHeaderHeight = this.getColumnEndHeaderHeight();
    rowHeaderWidth = this.getRowHeaderWidth();
    rowEndHeaderWidth = this.getRowEndHeaderWidth();

    //adjusted to make the databody wrap the databody content, and the scroller to fill the remaing part of the grid
    //this way our scrollbars are always at the edges of our viewport
    availableHeight = height - colHeaderHeight - colEndHeaderHeight;
    availableWidth = width - rowHeaderWidth - rowEndHeaderWidth;

    isTouchDevice = this.m_utils.isTouchDevice();
    scrollbarSize = this.m_utils.getScrollbarSize();

    isEmpty = this._databodyEmpty();

    // check if there's no data
    if (isEmpty)
    {
        //could be getting here in the handle resize of an empty grid
        if (this.m_empty == null)
        {
            empty = this._buildEmptyText();
            this.m_root.appendChild(empty); //@HTMLUpdateOK
        }
        else
        {
            empty = this.m_empty;
        }
        emptyHeight = this.getElementHeight(empty);
        emptyWidth = this.getElementWidth(empty);

        if (emptyHeight > this.getElementHeight(databodyScroller))
        {
            this.setElementHeight(databodyScroller, emptyHeight);
        }
        if (emptyWidth > this.getElementWidth(databodyScroller))
        {
            this.setElementWidth(databodyScroller, emptyWidth);
        }
    }

    databodyContentWidth = this.getElementWidth(databody['firstChild']);
    databodyContentHeight = this.getElementHeight(databody['firstChild']);
    //determine which scrollbars are required, if needing one forces need of the other, allows rendering within the root div
    isDatabodyHorizontalScrollbarRequired = this.isDatabodyHorizontalScrollbarRequired(availableWidth);
    if (isDatabodyHorizontalScrollbarRequired)
    {
        isDatabodyVerticalScrollbarRequired = this.isDatabodyVerticalScrollbarRequired(availableHeight - scrollbarSize);
        databody['style']['overflow'] = "auto";
    }
    else
    {
        isDatabodyVerticalScrollbarRequired = this.isDatabodyVerticalScrollbarRequired(availableHeight);
        if (isDatabodyVerticalScrollbarRequired)
        {
            isDatabodyHorizontalScrollbarRequired = this.isDatabodyHorizontalScrollbarRequired(availableWidth - scrollbarSize);
            databody['style']['overflow'] = "auto";
        }
        else
        {
            // for an issue where same size child causes scrollbars (similar code used in resizing already)
            // Adding timeout to address firefox async scrolling and pixel perfect scroll bar issues:
            // If the datagrid doesn't need scrolling, ff will skip the async scroll event, so we need to manually handle
            // it with a timeout. Flag added here and handleScroll so that either will execute, but only once.
            this.m_handleScrollOverflow = false;
            var self = this;
            setTimeout(function(){
                // Need to check this in case the state has changed during the timeout period.
                if (!self.m_handleScrollOverflow && !self.m_hasVerticalScroller && !self.m_hasHorizontalScroller)
                {
                    databody['style']['overflow'] = "hidden";
                    self.m_handleScrollOverflow = true;
                }
            }, 10);                        
        }
    }

    this.m_hasHorizontalScroller = isDatabodyHorizontalScrollbarRequired;
    this.m_hasVerticalScroller = isDatabodyVerticalScrollbarRequired;

    if (this.m_endColEndHeader != -1)
    {
        databodyHeight = Math.min(databodyContentHeight + (isDatabodyHorizontalScrollbarRequired ? scrollbarSize : 0), availableHeight);
        rowHeaderHeight = isDatabodyHorizontalScrollbarRequired ? databodyHeight - scrollbarSize : databodyHeight;
    }
    else
    {
        databodyHeight = availableHeight;
        rowHeaderHeight = Math.min(databodyContentHeight, isDatabodyHorizontalScrollbarRequired ? databodyHeight - scrollbarSize : databodyHeight);
    }

    if (this.m_endRowEndHeader != -1)
    {
        databodyWidth = Math.min(databodyContentWidth + (isDatabodyVerticalScrollbarRequired ? scrollbarSize : 0), availableWidth);
        columnHeaderWidth = isDatabodyVerticalScrollbarRequired ? databodyWidth - scrollbarSize : databodyWidth;
    }
    else
    {
        databodyWidth = availableWidth;
        columnHeaderWidth = Math.min(databodyContentWidth, isDatabodyVerticalScrollbarRequired ? databodyWidth - scrollbarSize : databodyWidth);
    }

    rowEndHeaderDir = rowHeaderWidth + columnHeaderWidth + (isDatabodyVerticalScrollbarRequired ? scrollbarSize : 0);
    columnEndHeaderDir = colHeaderHeight + rowHeaderHeight + (isDatabodyHorizontalScrollbarRequired ? scrollbarSize : 0);

    dir = this.getResources().isRTLMode() ? "right" : "left";

    this.setElementDir(rowHeader, 0, dir);
    this.setElementDir(rowHeader, colHeaderHeight, 'top');
    this.setElementHeight(rowHeader, rowHeaderHeight);

    this.setElementDir(rowEndHeader, rowEndHeaderDir, dir);
    this.setElementDir(rowEndHeader, colHeaderHeight, 'top');
    this.setElementHeight(rowEndHeader, rowHeaderHeight);

    this.setElementDir(colHeader, rowHeaderWidth, dir);
    this.setElementWidth(colHeader, columnHeaderWidth);

    this.setElementDir(colEndHeader, rowHeaderWidth, dir);
    this.setElementDir(colEndHeader, columnEndHeaderDir, 'top');
    this.setElementWidth(colEndHeader, columnHeaderWidth);

    this.setElementDir(databody, colHeaderHeight, 'top');
    this.setElementDir(databody, rowHeaderWidth, dir);
    this.setElementWidth(databody, databodyWidth);
    this.setElementHeight(databody, databodyHeight);

    // cache the scroll width and height to minimize reflow
    this.m_scrollWidth = databodyContentWidth - columnHeaderWidth;
    this.m_scrollHeight = databodyContentHeight - rowHeaderHeight;

    this.buildCorners();

    // check if we need to remove border on the last column header/add borders to headers
    this._adjustHeaderBorders();

    // now we do not need to resize
    this.m_resizeRequired = false;

    endTime = new Date();
};

/**
 * Size the databody scroller based on whatever dimensions are available.
 * @private
 */
DvtDataGrid.prototype._sizeDatabodyScroller = function()
{
    var databody, scroller, isEmpty, isHWS, maxHeight, maxWidth, rowCount, colCount, totalHeight, totalWidth,
            endRowPixel, endColPixel;
    databody = this.m_databody;
    scroller = databody.firstChild;
    isEmpty = this._databodyEmpty();
    isHWS = this._isHighWatermarkScrolling();
    maxHeight = this.m_utils._getMaxDivHeightForScrolling();
    maxWidth = this.m_utils._getMaxDivWidthForScrolling();
    rowCount = this.getDataSource().getCount('row');
    colCount = this.getDataSource().getCount('column');
    totalHeight = 0;
    totalWidth = 0;
    endRowPixel = 0;
    endColPixel = 0;

    if (isEmpty)
    {
        // min is 1 so that the scrollbars show up
        endRowPixel = Math.max(Math.max(this.m_endRowHeaderPixel, this.m_endRowEndHeaderPixel), 1);
        endColPixel = Math.max(Math.max(this.m_endColHeaderPixel, this.m_endColEndHeaderPixel), 1);
    }
    else
    {
        endRowPixel = this.m_endRowPixel;
        endColPixel = this.m_endColPixel;
    }

    totalHeight = (rowCount != -1 && !isHWS) ? rowCount * this.m_avgRowHeight : endRowPixel;
    totalWidth = (colCount != -1 && !isHWS) ? colCount * this.m_avgColWidth : endColPixel;

    this.setElementHeight(scroller, Math.min(maxHeight, totalHeight));
    this.setElementWidth(scroller, Math.min(maxWidth, totalWidth));

    if (this.m_initialized)
    {
        this.m_scrollWidth = this.getElementWidth(scroller) - this.getElementWidth(databody) + (this.m_hasVerticalScroller ?  this.m_utils.getScrollbarSize() : 0);
        this.m_scrollHeight = this.getElementHeight(scroller) - this.getElementHeight(databody) + (this.m_hasHorizontalScroller ?  this.m_utils.getScrollbarSize() : 0);
    }
};

/**
 * Adjust the last header on specific axis properties
 * @private
 * @param {number} headerIndex
 * @param {number} headerLevels
 * @param {Element} container
 * @param {number} startIndex
 * @param {string} className
 * @param {boolean} remove
 */
DvtDataGrid.prototype._adjustLastHeadersAlongAxis = function(headerIndex, headerLevels, container, startIndex, className, remove)
{
    var i, lastHeader;
    i = 0;
    while (i < headerLevels)
    {
        lastHeader = this._getHeaderByIndex(headerIndex, i, container, headerLevels, startIndex);
        remove ? this.m_utils.removeCSSClassName(lastHeader, className) : this.m_utils.addCSSClassName(lastHeader, className);
        i += this.getHeaderCellDepth(lastHeader);
    }
};

/**
 * Adjust the last header and the spacer along a given axis
 *
 * @param {Element} container
 * @param {Function} lastFunction
 * @param {number} endHeaderIndex
 * @param {boolean} dimensionCheck
 * @param {Element} spacer
 * @param {string} className
 * @param {number} headerLevels
 * @param {number} startIndex
 */
DvtDataGrid.prototype._adjustHeaderBordersAlongAxis = function(container, lastFunction, endHeaderIndex, dimensionCheck, spacer, className, headerLevels, startIndex)
{
    if (container != null && endHeaderIndex >= 0)
    {
        dimensionCheck ? this.m_utils.addCSSClassName(spacer, className): this.m_utils.removeCSSClassName(spacer, className);
        if (lastFunction(endHeaderIndex))
        {
            this._adjustLastHeadersAlongAxis(endHeaderIndex, headerLevels, container, startIndex, className, dimensionCheck);
        }
    }
};

/**
 * Adjust the border style/width setting on the headers using classNames so that they can be overwritten
 * @private
 */
DvtDataGrid.prototype._adjustHeaderBorders = function()
{
    var colHeaderHeight, colHeaderWidth, widthCheck, heightCheck, scrollbarSize, style,
            colEndHeaderHeight, rowHeaderWidth, rowHeaderHeight, rowEndHeaderWidth, width, height, i, tags, bw, lastFunction;

    scrollbarSize = this.m_utils.getScrollbarSize();
    width = this.getWidth();
    height = this.getHeight();
    colHeaderHeight = this.getColumnHeaderHeight();
    colHeaderWidth = this.getElementWidth(this.m_colHeader);
    colEndHeaderHeight = this.getElementHeight(this.m_colEndHeader);
    rowHeaderWidth = this.getRowHeaderWidth();
    rowHeaderHeight = this.getElementHeight(this.m_rowHeader);
    rowEndHeaderWidth = this.getElementWidth(this.m_rowEndHeader);

    widthCheck = rowHeaderWidth + colHeaderWidth + rowEndHeaderWidth + (this.m_hasVerticalScroller ? scrollbarSize : 0) < width;
    heightCheck = colHeaderHeight + rowHeaderHeight + colEndHeaderHeight + (this.m_hasHorizontalScroller ? scrollbarSize : 0) < height;

    if (widthCheck && this.m_endRowEndHeader >= 0 )
    {
        bw = true;
        this.m_addBorderRight = true;
    }
    else if (this.m_addBorderRight == true)
    {
        bw = false;
    }

    if (bw != null)
    {
        style = this.getMappedStyle('borderVerticalSmall');
        if (this.m_columnHeaderScrollbarSpacer != null)
        {
            bw ? this.m_utils.addCSSClassName(this.m_columnHeaderScrollbarSpacer, style) : this.m_utils.removeCSSClassName(this.m_columnHeaderScrollbarSpacer, style);
        }
        if (this.m_bottomCorner != null)
        {
            bw ? this.m_utils.addCSSClassName(this.m_bottomCorner, style) : this.m_utils.removeCSSClassName(this.m_bottomCorner, style);
        }
        tags = this.m_rowEndHeader.firstChild.childNodes;
        for (i = 0; i < tags.length; i++)
        {
            bw ? this.m_utils.addCSSClassName(tags[i], style) : this.m_utils.removeCSSClassName(tags[i], style);
        }
    }
    else
    {
        style = this.getMappedStyle('borderVerticalNone');
        lastFunction = this._isLastColumn.bind(this);
        this._adjustHeaderBordersAlongAxis(this.m_colHeader, lastFunction, this.m_endColHeader, widthCheck, this.m_columnHeaderScrollbarSpacer, style,
                this.m_columnHeaderLevelCount, this.m_startColHeader);
        this._adjustHeaderBordersAlongAxis(this.m_colEndHeader, lastFunction, this.m_endColEndHeader, widthCheck, this.m_bottomCorner, style,
                this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
    }

    bw = null;

    if (heightCheck && this.m_endColEndHeader >= 0)
    {
        this.m_addBorderBottom = true;
        bw = true;
    }
    else if (this.m_addBorderBottom == true)
    {
        bw = false;
    }

    if (bw != null)
    {
        style = this.getMappedStyle('borderHorizontalSmall');
        if (this.m_rowHeaderScrollbarSpacer != null)
        {
            bw ? this.m_utils.addCSSClassName(this.m_rowHeaderScrollbarSpacer, style) : this.m_utils.removeCSSClassName(this.m_rowHeaderScrollbarSpacer, style);
        }
        if (this.m_bottomCorner != null)
        {
            bw ? this.m_utils.addCSSClassName(this.m_bottomCorner, style) : this.m_utils.removeCSSClassName(this.m_bottomCorner, style);
        }
        tags = this.m_colEndHeader.firstChild.childNodes;
        for (i = 0; i < tags.length; i++)
        {
            bw ? this.m_utils.addCSSClassName(tags[i], style) : this.m_utils.removeCSSClassName(tags[i], style);
        }
    }
    else
    {
        style = this.getMappedStyle('borderHorizontalNone');
        lastFunction = this._isLastRow.bind(this);
        this._adjustHeaderBordersAlongAxis(this.m_rowHeader, lastFunction, this.m_endRowHeader, heightCheck, this.m_rowHeaderScrollbarSpacer, style,
                this.m_rowHeaderLevelCount, this.m_startRowHeader);
        this._adjustHeaderBordersAlongAxis(this.m_rowEndHeader, lastFunction, this.m_endRowEndHeader, heightCheck, this.m_bottomCorner, style,
                this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
    }
};

/**
 * Build the corners of the grid. If they exist, removes them and rebuilds them.
 * @private
 */
DvtDataGrid.prototype.buildCorners = function()
{
    var scrollbarSize, colHeaderHeight, rowHeaderWidth, bottomCorner,
            corner, dir, rowHeaderScrollbarSpacer, columnHeaderScrollbarSpacer,
            colHeaderWidth, rowHeaderHeight, colEndHeaderHeight,
            rowEndHeaderWidth, width, height;

    scrollbarSize = this.m_utils.getScrollbarSize();
    width = this.getWidth();
    height = this.getHeight();
    colHeaderHeight = this.getColumnHeaderHeight();
    colHeaderWidth = this.getElementWidth(this.m_colHeader);
    colEndHeaderHeight = this.getColumnEndHeaderHeight();
    rowHeaderWidth = this.getRowHeaderWidth();
    rowEndHeaderWidth = this.getRowEndHeaderWidth();
    rowHeaderHeight = this.getElementHeight(this.m_rowHeader);
    dir = this.getResources().isRTLMode() ? "right" : "left";

    //rather than removing and appending the nodes every time just adjust the live ones

    if (this.m_endRowHeader != -1 && this.m_endColHeader != -1)
    {
        // render corner
        if (this.m_corner != null)
        {
            corner = this.m_corner;
        }
        else
        {
            corner = document.createElement("div");
            corner['id'] = this.createSubId("corner");
            corner['className'] = this.getMappedStyle("topcorner");
        }

        this.setElementWidth(corner, rowHeaderWidth);
        this.setElementHeight(corner, colHeaderHeight);

        if (this.m_corner == null)
        {
            this.m_root.appendChild(corner); //@HTMLUpdateOK
            this.m_corner = corner;
        }
    }
    if (this.m_corner != null && corner == null)
    {
        this.m_root.removeChild(this.m_corner);
        this.m_corner = null;
    }

    //corner listener:
    if (this.m_corner && this.m_corner !== null)
    {
        this.m_corner.addEventListener("click", this.handleCornerClick.bind(this), false);
    }

    //no bottom left corner if there are no row headers
    if (this.m_endRowHeader != -1)
    {
        if (this.m_hasHorizontalScroller || this.m_endColEndHeader != -1)
        {
            if (this.m_rowHeaderScrollbarSpacer != null)
            {
                rowHeaderScrollbarSpacer = this.m_rowHeaderScrollbarSpacer;
            }
            else
            {
                rowHeaderScrollbarSpacer = document.createElement("div");
                rowHeaderScrollbarSpacer['id'] = this.createSubId("rhSbSpacer");
                rowHeaderScrollbarSpacer['className'] = this.getMappedStyle("rowheaderspacer");
            }

            this.setElementDir(rowHeaderScrollbarSpacer, (rowHeaderHeight + colHeaderHeight), 'top');
            this.setElementDir(rowHeaderScrollbarSpacer, 0, dir);
            this.setElementWidth(rowHeaderScrollbarSpacer, rowHeaderWidth);
            if (this.m_endColEndHeader != -1)
            {
                this.setElementHeight(rowHeaderScrollbarSpacer, colEndHeaderHeight + (this.m_hasHorizontalScroller ? scrollbarSize : 0));
            }
            else
            {
                this.setElementHeight(rowHeaderScrollbarSpacer, height - rowHeaderHeight - colHeaderHeight);
            }

            if (this.m_rowHeaderScrollbarSpacer == null)
            {
                this.m_root.appendChild(rowHeaderScrollbarSpacer); //@HTMLUpdateOK
                this.m_rowHeaderScrollbarSpacer = rowHeaderScrollbarSpacer;
            }
        }
        else
        {
            if (this.m_rowHeaderScrollbarSpacer != null)
            {
                this.m_root.removeChild(this.m_rowHeaderScrollbarSpacer);
            }
            this.m_rowHeaderScrollbarSpacer = null;
        }
    }

    //no top right corner if there are no col headers
    if (this.m_endColHeader != -1)
    {
        if (this.m_hasVerticalScroller || this.m_endRowEndHeader != -1)
        {
            if (this.m_columnHeaderScrollbarSpacer != null)
            {
                columnHeaderScrollbarSpacer = this.m_columnHeaderScrollbarSpacer;
            }
            else
            {
                columnHeaderScrollbarSpacer = document.createElement("div");
                columnHeaderScrollbarSpacer['id'] = this.createSubId("chSbSpacer");
                columnHeaderScrollbarSpacer['className'] = this.getMappedStyle("colheaderspacer");
            }

            this.setElementDir(columnHeaderScrollbarSpacer, (rowHeaderWidth + colHeaderWidth), dir);
            this.setElementDir(columnHeaderScrollbarSpacer, 0, 'top');
            if (this.m_endRowEndHeader != -1)
            {
                this.setElementWidth(columnHeaderScrollbarSpacer, rowEndHeaderWidth + (this.m_hasVerticalScroller ? scrollbarSize : 0));
            }
            else
            {
                this.setElementWidth(columnHeaderScrollbarSpacer, width - colHeaderWidth - rowHeaderWidth);
            }
            this.setElementHeight(columnHeaderScrollbarSpacer, colHeaderHeight);

            if (this.m_columnHeaderScrollbarSpacer == null)
            {
                this.m_root.appendChild(columnHeaderScrollbarSpacer); //@HTMLUpdateOK
                this.m_columnHeaderScrollbarSpacer = columnHeaderScrollbarSpacer;
            }
        }
        else
        {
            if (this.m_columnHeaderScrollbarSpacer != null)
            {
                this.m_root.removeChild(this.m_columnHeaderScrollbarSpacer);
            }
            this.m_columnHeaderScrollbarSpacer = null;
        }
    }

    if ((this.m_hasHorizontalScroller && this.m_hasVerticalScroller) ||
        (this.m_hasVerticalScroller && this.m_endColEndHeader != -1) ||
        (this.m_hasHorizontalScroller && this.m_endRowEndHeader != -1) ||
        (this.m_endRowEndHeader != -1 && this.m_endColEndHeader != -1))
    {
    // render bottom corner (for both scrollbars) if needed
        if (this.m_bottomCorner != null)
        {
            bottomCorner = this.m_bottomCorner;
        }
        else
        {
            bottomCorner = document.createElement("div");
            bottomCorner['id'] = this.createSubId("bcorner");
            bottomCorner['className'] = this.getMappedStyle("bottomcorner");
        }

        this.setElementDir(bottomCorner, (rowHeaderHeight + colHeaderHeight), 'top');
        this.setElementDir(bottomCorner, (rowHeaderWidth + colHeaderWidth), dir);
        if (this.m_endRowEndHeader != -1)
        {
            this.setElementWidth(bottomCorner, rowEndHeaderWidth + (this.m_hasVerticalScroller ? scrollbarSize : 0));
        }
        else
        {
            this.setElementWidth(bottomCorner, width - colHeaderWidth - rowHeaderWidth);
        }

        if (this.m_endColEndHeader != -1)
        {
            this.setElementHeight(bottomCorner, colEndHeaderHeight + (this.m_hasHorizontalScroller ? scrollbarSize : 0));
        }
        else
        {
            this.setElementHeight(bottomCorner, height - rowHeaderHeight - colHeaderHeight);
        }

        if (this.m_bottomCorner == null)
        {
            this.m_root.appendChild(bottomCorner); //@HTMLUpdateOK
            this.m_bottomCorner = bottomCorner;
        }
    }
    // remove bottom corner on resize if not neccessary
    if (this.m_bottomCorner != null && bottomCorner == null)
    {
        this.m_root.removeChild(this.m_bottomCorner);
        this.m_bottomCorner = null;
    }
};

/**
 * Move to a desired scoll position object
 * @private
 */
DvtDataGrid.prototype._updateScrollPosition = function(scrollPositionObject)
{
    this._scrollToScrollPositionObject(scrollPositionObject);
};

/**
 * Set the scrollPosition attribute on the grid
 * @private
 */
DvtDataGrid.prototype._setScrollPosition = function()
{
    this.m_setOptionCallback("scrollPosition", this._createScrollPositionObject(this.m_currentScrollLeft, this.m_currentScrollTop), {'_context': {writeback: true, internalSet: true}});
};

/**
 * Set the scrollPosition minus the keys
 * @private
 */
DvtDataGrid.prototype._clearScrollPositionKeys = function()
{
    var newPos = this.m_options.getScrollPosition();
    newPos['rowKey'] = undefined;
    newPos['columnKey'] = undefined;
    // do not writeback it will be updated once the sync has completed
    this.m_setOptionCallback("scrollPosition", newPos, {'_context': {internalSet: true}});
};

/**
 * Create a scroll position object from x,y coordinates
 * @private
 */
DvtDataGrid.prototype._createScrollPositionObject = function(x, y)
{
    var scrollPositionObject = {'x': x, 'y': y};
    var dir = this.getResources().isRTLMode() ? "right" : "left";
    
    var cell = this._getCellAtPixel(x, y);
    if (cell != null)
    {
        scrollPositionObject['rowIndex'] = this._getIndex(cell, 'row');
        scrollPositionObject['columnIndex'] = this._getIndex(cell, 'column');
        scrollPositionObject['rowKey'] = this._getKey(cell, 'row');
        scrollPositionObject['columnKey'] = this._getKey(cell, 'column');
        scrollPositionObject['offsetX'] = x - this.getElementDir(cell, dir);
        scrollPositionObject['offsetY'] = y - this.getElementDir(cell, "top");
    }
    else
    {
        var rowHeader = this._getHeaderAtPixel(y, 'row');
        if (rowHeader != null)
        {
            scrollPositionObject['rowIndex'] = this._getIndex(rowHeader);            
            scrollPositionObject['rowKey'] = this._getKey(rowHeader);
            scrollPositionObject['offsetY'] = y - this.getElementDir(rowHeader, 'top');
        }
        
        var columnHeader = this._getHeaderAtPixel(x, 'column');
        if (columnHeader != null)
        {
            scrollPositionObject['columnIndex'] = this._getIndex(columnHeader);            
            scrollPositionObject['columnKey'] = this._getKey(columnHeader);
            scrollPositionObject['offsetX'] = x - this.getElementDir(columnHeader, dir);
        }        
    }
    
    return scrollPositionObject;
};

/**
 * Get a cell at x,y coordinates
 * @private
 */
DvtDataGrid.prototype._getCellAtPixel = function(x, y)
{
    var cells = this.m_databody['firstChild']['childNodes'];
    var dir = this.getResources().isRTLMode() ? "right" : "left";

    for (var i = 0; i < cells.length; i++)
    {
        var cell = cells[i];
        var cellLeft = this.getElementDir(cell, dir);
        var cellRight = cellLeft + this.getElementWidth(cell);
        
        if (cellLeft <= x && x< cellRight)
        {
            var cellTop = this.getElementDir(cell, 'top');
            var cellBottom = cellTop + this.getElementHeight(cell);
            if (cellTop <= y && y < cellBottom)
            {
                return cell;
            }
        }
    }
    return null;
};

/**
 * Get a header at a given coordinate and axis
 * @private
 */
DvtDataGrid.prototype._getHeaderAtPixel = function(pixel, axis)
{
    var self = this;
    if (axis == 'row')
    {
        var headers = this.m_rowHeader['firstChild']['childNodes'];
        var endheaders = this.m_rowEndHeader['firstChild']['childNodes'];
        var dir = 'top';
        var dimension = 'height';
    }
    else if (axis == 'column')
    {
        var headers = this.m_colHeader['firstChild']['childNodes'];
        var endheaders = this.m_colEndHeader['firstChild']['childNodes'];
        var dir = this.getResources().isRTLMode() ? "right" : "left";
        var dimension = 'width';
    }
    
    var loop = function(elements)
    {
        for (var i = 0; i < elements.length; i++)
        {
            var header = elements[i];
            var start = self.getElementDir(header, dir);
            var end = start + self.getElementDir(header, dimension);
            
            if (start <= pixel && pixel < end)
            {
                return header;
            }
        } 
    };
    
    var header = loop(headers);
    if (header == null)
    {
        header = loop(endheaders);
    }

    return header;
};

/**
 * See if we have reached our desired scroll position
 * @private
 */
DvtDataGrid.prototype._checkScrollPosition = function()
{
    if (this.m_desiredScrollPositionObject != null)
    {
        this._scrollToScrollPositionObject(this.m_desiredScrollPositionObject);
    }
    else if (!this.m_initialized)
    {
        this._updateScrollPosition(this.m_options.getScrollPosition());
    }
    else
    {
        this._setScrollPosition();
    }
};

/**
 * Call initiate scroll to scroll to a scroll position object
 */
DvtDataGrid.prototype._scrollToScrollPositionObject = function(scrollPositionObject)
{    
    var x = scrollPositionObject['x'];
    var y = scrollPositionObject['y'];
    var rowIndex = scrollPositionObject['rowIndex'];
    var columnIndex = scrollPositionObject['columnIndex'];
    var rowKey = scrollPositionObject['rowKey'];   
    var columnKey = scrollPositionObject['columnKey'];
    var offsetX = scrollPositionObject['offsetX'] ? scrollPositionObject['offsetX'] : 0;
    var offsetY = scrollPositionObject['offsetY'] ? scrollPositionObject['offsetY'] : 0;
    
    var newScrollX;
    var newScrollY;
    
    var dir = this.getResources().isRTLMode() ? "right" : "left";
    
    var self = this;
    
    var indexFromKeyPromise = new Promise(function(resolve, reject) {
         if (rowKey != null || columnKey != null)
         {
            self._indexes({'row': rowKey, 'column': columnKey}, function(indexes, keys){
                resolve({'rowIndexFromKey':  indexes['row'], 'columnIndexFromKey': indexes['column']});
            });
         }
         else
         {
            resolve({'rowIndexFromKey':  null, 'columnIndexFromKey': null});
         }
    });
    
    indexFromKeyPromise.then(function(returnObj)
    {
        var newScrollX = self._getPositionEstimate('column', dir, columnKey, returnObj['columnIndexFromKey'], columnIndex, x, offsetX, self.m_currentScrollLeft, self._getMaxRightPixel(), self.m_avgColWidth);        
        var newScrollY = self._getPositionEstimate('row', 'top', rowKey, returnObj['rowIndexFromKey'], rowIndex, y, offsetY, self.m_currentScrollTop, self._getMaxBottomPixel(), self.m_avgRowHeight);
        
        if (newScrollX != self.m_currentScrollLeft || newScrollY != self.m_currentScrollTop)
        {
            self.m_desiredScrollPositionObject = scrollPositionObject;
            self._initiateScroll(newScrollX, newScrollY);
        }
        else 
        {
            self.m_desiredScrollPositionObject = null;
            self._setScrollPosition();
        }        
    });
};

/**
 * Estimate the location of the scroll based on key/index/x,y following the jsdoc guidelines
 * @private
 */
DvtDataGrid.prototype._getPositionEstimate = function(axis, dir, key, indexFromKey, index, pos, offset, current, max, average)
{    
    var newScrollPos;
    var isHighWatermark = this._isHighWatermarkScrolling();
    
    // indexFromKey === -1 means that the key is definitely not in the data source
    // indexFromKey === null means that the key is possibly not in the data source
    if (key != null && indexFromKey != -1)
    {
        var element = this._getCellOrHeaderByKey(key, axis);
        if (element != null)
        {
            newScrollPos = this.getElementDir(element, dir) + offset;
        }
        else if (isHighWatermark)
        {
            newScrollPos = max;
        }
        else if (indexFromKey != null)
        {
            newScrollPos = (average * indexFromKey) + offset;
        } 
        
        if (newScrollPos != null)
        {
            return newScrollPos;
        }
        
        // if key specified and virtual scrolling but no index found, switch to index check
    }
    
    if (index != null)
    {
        var element = this._getCellOrHeaderByIndex(index, axis);
        if (element != null)
        {
           newScrollPos = this.getElementDir(element, dir) + offset;
        }            
        else if (isHighWatermark)
        {
           newScrollPos = max;
        }            
        else
        {
            newScrollPos = average * index;
        }
    }
    else if (pos != null)
    {
        newScrollPos = pos;
    }
    else
    {
        newScrollPos = current;
    }
    
    return newScrollPos;
};

/**
 * Check if the key is rendered in the viewport to avoid asking the ds
 * @private
 */
DvtDataGrid.prototype._isKeyRendered = function(key, axis)
{
    var cells = this.m_databody['firstChild']['childNodes'];
    for (var i = 0; i < cells.length; i++)
    {
        if (this._shallowThenDeepCompare(this._getKey(cells[i], axis), key))
        {
            return cells[i];
        }
    }
};

/**
 * Determine if horizontal scrollbar is needed
 * @param {number|null=} expectedWidth - databody width
 * @return {boolean} true if horizontal scrollbar required
 */
DvtDataGrid.prototype.isDatabodyHorizontalScrollbarRequired = function(expectedWidth)
{
    var databody, scroller, expected;
    // if expected width of the databody is not specified, extract from style
    databody = this.m_databody;
    if (expectedWidth == null)
    {
        expected = this.getElementWidth(databody);
    }
    else
    {
        expected = expectedWidth;
    }

    scroller = databody['firstChild'];
    if (this.getElementWidth(scroller) > expected)
    {
        return true;
    }
    return false;
};

/**
 * Determine if vertical scrollbar is needed
 * @param {number|null=} expectedHeight - databody height
 * @return {boolean} true if vertical scrollbar required
 * @private
 */
DvtDataGrid.prototype.isDatabodyVerticalScrollbarRequired = function(expectedHeight)
{
    var databody, scroller, expected;
    // if expected height of the databody is not specified, extract from style
    databody = this.m_databody;
    if (expectedHeight == null)
    {
        expected = this.getElementHeight(databody);
    }
    else
    {
        expected = expectedHeight;
    }

    scroller = databody['firstChild'];
    if (this.getElementHeight(scroller) > expected)
    {
        return true;
    }
    return false;
};

/**
 * todo: merge with buildAccInfo, we just need one status role div.
 * Build a status bar div
 * @return {Element} the root of the status bar
 * @private
 */
DvtDataGrid.prototype.buildStatus = function()
{
    var icon, root;

    icon = document.createElement("div");
    icon['className'] = this.getMappedStyle("loadingicon");

    root = document.createElement("div");
    root['id'] = this.createSubId("status");
    root['className'] = this.getMappedStyle("status");
    root.setAttribute("role", "status");
    root.appendChild(icon);

    return root;
};

/**
 * Build the offscreen div used by screenreader for action such as sort
 * @return {Element} the root of the accessibility info div
 */
DvtDataGrid.prototype.buildAccInfo = function()
{
    var root = document.createElement("div");
    root['id'] = this.createSubId("info");
    root['className'] = this.getMappedStyle("info");
    root.setAttribute("role", "status");

    return root;
};

/**
 * Build the offscreen div used by screenreader for summary description
 * @return {Element} the root of the accessibility summary div
 */
DvtDataGrid.prototype.buildAccSummary = function()
{
    var root = document.createElement("div");
    root['id'] = this.createSubId("summary");
    root['className'] = this.getMappedStyle("info");

    return root;
};

/**
 * Build the offscreen div used by screenreader for state information
 * @return {Element} the root of the accessibility state info div
 */
DvtDataGrid.prototype.buildStateInfo = function()
{
    var root = document.createElement("div");
    root['id'] = this.createSubId("state");
    root['className'] = this.getMappedStyle("info");

    return root;
};

/**
 * Build the offscreen div used by screenreader for cell context information
 * @return {Element} the root of the accessibility context info div
 */
DvtDataGrid.prototype.buildContextInfo = function()
{
    var root = document.createElement("div");
    root['id'] = this.createSubId("context");
    root['className'] = this.getMappedStyle("info");

    return root;
};

/**
 * Build the offscreen div used by screenreader used for reading current cell context information
 * @return {Element} the root of the accessibility current cell context info div
 */
DvtDataGrid.prototype.buildPlaceHolder = function()
{
    var root = document.createElement("div");
    root['id'] = this.createSubId("placeHolder");
    root['className'] = this.getMappedStyle("info");

    return root;
};

/**
 * Sets the text on the offscreen div.  The text contains a summary text describing the grid
 * including structure information
 * @private
 */
DvtDataGrid.prototype.populateAccInfo = function()
{
    var summary, summaryExpanded;

    summary = this.getResources().getTranslatedText('accessibleSummaryExact', {'rownum': (this.m_endRow + 1), 'colnum': (this.m_endCol + 1)});

    // if it's hierarchical, then include specific accessible info about what's expanded
    if (this.getDataSource().getExpandedKeys)
    {
        summaryExpanded = this.getResources().getTranslatedText('accessibleSummaryExpanded', {'num': this.getDataSource().getExpandedKeys().length});
        summary = summary + ". " + summaryExpanded;
    }

    // add instruction text
    summary += ". ";

    // set the summary text on the screen reader div
    this.m_accSummary.textContent = summary;
};

/**
 * Implements Accessible interface.
 * Sets accessible information on the DataGrid.
 * This is currently used by the Row Expander to alert screenreader of such
 * information as depth, expanded state, index etc
 * @param {Object} context an object containing attribute context or state to set m_accessibleContext/state
 */
DvtDataGrid.prototype.SetAccessibleContext = function(context)
{
    var label, ancestors, col, i, parent, rowCells, cell, text;

    if (context != null)
    {
        // got row context info
        if (context['context'] != null)
        {
            // save it for updateContextInfo to consume later
            this.m_accessibleContext = context['context'];
        }

        // got disclosure state info
        if (context['state'] != null)
        {
            this.m_stateInfo.textContent = context['state'];
        }

        // got ancestors info
        if (context['ancestors'] != null && this._isDatabodyCellActive())
        {
            label = '';
            ancestors = context['ancestors'];
            col = this.m_active['indexes']['column'];
            if (col != null && col >= 0)
            {
                // constructs the appropriate parent context info text
                for (i = 0; i < ancestors.length; i++)
                {
                    if (i > 0)
                    {
                        label = label.concat(', ');
                    }
                    parent = ancestors[i];
                    rowCells = this._getAxisCellsByKey(parent['key'], 'row');
                    if (rowCells != null)
                    {
                        cell = rowCells[0];
                        // we are just going to extract any text content (or find first aria-label if null?)
                        text = cell.textContent;
                        // remove any carriage return, tab etc.
                        if (text != null)
                        {
                            text = text.replace(/\n|<br\s*\/?>/gi, '').trim();
                        }
                        else
                        {
                            text = '';
                        }
                        label = label.concat(parent['label']).concat(' ').concat(text);
                    }
                }
            }

            // prepend to existing context info
            this.m_accessibleContext = label.concat(', ').concat(this.m_accessibleContext);
        }
    }
};

/**
 * Sets the accessibility state info text
 * @param {Array} items the message key
 * @private
 */
DvtDataGrid.prototype._updateStateInfo = function(items)
{
    var i, state, text;
    // allows pause after the data is read out before the state
    text = '. ';
    for (i = 0; i <items.length; i++)
    {
        state = this.getResources().getTranslatedText(items[i]['key'], items[i]['args']);
        if (state != null)
        {
            // for the original period we will just add text otherwise need a comma
            text = text.length === 2 ? text + state : text + ', ' + state;
        }
    }
    
    // end state with a period
    text = text.length === 0 ? text : text + '. ';    
    
    this.m_stateInfo.textContent = text;
};

/**
 * Sets the accessibility context info text
 * @param {Object} context the context info about the cell
 * @param {number=} context.row the row index
 * @param {number=} context.column the column index
 * @param {number=} context.level the level of the header if there is one
 * @param {number=} context.rowHeader the rowHeader index
 * @param {number=} context.columnHeader the rowHeader index
 * @param {string=} skip whether to skip row or column
 * @private
 */
DvtDataGrid.prototype._updateContextInfo = function(context, skip)
{
    var row, column, info, text, level, rowHeader, rowEndHeader, columnHeader, columnEndHeader, rowExtent, columnExtent, spacerText, endContextText;
    if (context['indexes'])
    {
        row = context['indexes']['row'];
        column = context['indexes']['column'];
    }

    level = context['level'];
    rowHeader = context['rowHeader'];
    rowEndHeader = context['rowEndHeader'];
    columnHeader = context['columnHeader'];
    columnEndHeader = context['columnEndHeader'];
    info = "";
    spacerText = ", ";
    endContextText = ". ";

    // row context.  Skip if there is an outstanding accessible row context
    if (this.m_accessibleContext == null && !isNaN(row) && skip != 'row')
    {
        text = this.getResources().getTranslatedText('accessibleRowContext', {'index': row + 1});
        if (text != null)
        {
            info = text;
        }   
    }

    // column context
    if (!isNaN(column) && skip != 'column')
    {
        text = this.getResources().getTranslatedText('accessibleColumnContext', {'index': column + 1});
        if (text != null)
        {
            info = info.length === 0 ? text : info + spacerText + text;
        } 
    }

    // rowHeader context
    if (!isNaN(rowHeader) && skip != 'rowHeader')
    {
        text = this.getResources().getTranslatedText('accessibleRowHeaderContext', {'index': rowHeader + 1});
        if (text != null)
        {
            info = info.length === 0 ? text : info + spacerText + text;
        }
    }

    // columHeader context
    if (!isNaN(columnHeader) && skip != 'columnHeader')
    {
        text = this.getResources().getTranslatedText('accessibleColumnHeaderContext', {'index': columnHeader + 1});
        if (text != null)
        {
            info = info.length === 0 ? text : info + spacerText + text;
        }
    }

    // rowHeader context
    if (!isNaN(rowEndHeader) && skip != 'rowEndHeader')
    {
        text = this.getResources().getTranslatedText('accessibleRowEndHeaderContext', {'index': rowEndHeader + 1});
        if (text != null)
        {
            info = info.length === 0 ? text : info + spacerText + text;
        }
    }

    // columHeader context
    if (!isNaN(columnEndHeader) && skip != 'columnEndHeader')
    {
        text = this.getResources().getTranslatedText('accessibleColumnEndHeaderContext', {'index': columnEndHeader + 1});
        if (text != null)
        {
            info = info.length === 0 ? text : info + spacerText + text;
        }
    }

    // level context
    if (!isNaN(level) && skip != 'level')
    {
        text = this.getResources().getTranslatedText('accessibleLevelContext', {'level': level + 1});
        if (text != null)
        {
            info = info.length === 0 ? text : info + spacerText + text;
        }
    }  
    
    // Put a period at the end of the context readout
    info = info.length === 0 ? info : info + endContextText;
    
    // merge with accesssible context (from row expander)
    if (this.m_accessibleContext != null)
    {
        info = info + this.m_accessibleContext;
        // reset
        this.m_accessibleContext = null;
    }

    this.m_contextInfo.textContent = info;
};

/**
 * Determine whether the row/column count is unknown.
 * @param {string} axis the row/column axis
 * @return {boolean|undefined} true if the count for the axis is unknown, false otherwise
 * @private
 */
DvtDataGrid.prototype._isCountUnknown = function(axis)
{
    var datasource, rowPrecision, colPrecision, rowCount, colCount;

    datasource = this.getDataSource();
    if (axis === 'row' || axis === 'rowEnd')
    {
        if (this.m_isEstimateRowCount === undefined)
        {
            rowPrecision = datasource.getCountPrecision('row');
            rowCount = datasource.getCount('row');
            if (rowPrecision === 'estimate' || rowCount < 0)
            {
                this.m_isEstimateRowCount = true;
            }
            else
            {
                this.m_isEstimateRowCount = false;
            }
        }

        return this.m_isEstimateRowCount;
    }
    else if (axis === 'column' || axis === 'columnEnd')
    {
        if (this.m_isEstimateColumnCount === undefined)
        {
            colPrecision = datasource.getCountPrecision('column');
            colCount = datasource.getCount('column');
            if (colPrecision === 'estimate' || colCount < 0)
            {
                this.m_isEstimateColumnCount = true;
            }
            else
            {
                this.m_isEstimateColumnCount = false;
            }
        }

        return this.m_isEstimateColumnCount;
    }

    // unrecognize axis, just assume the count is known
    return false;
};

/**
 * Convenient method which returns true if row count is unknown or high watermark scrolling is used.
 * @param {string} axis the row/column axis
 * @return {boolean} true if count is unknown or high watermark scrolling is used, false otherwise.
 * @private
 */
DvtDataGrid.prototype._isCountUnknownOrHighwatermark = function(axis)
{
    return (this._isCountUnknown(axis) || this._isHighWatermarkScrolling());
};

/**
 * Set display to none
 * @param {Element} root
 * @private
 */
DvtDataGrid.prototype._hideHeader = function(root)
{
    root['style']['display'] = 'none';
};

/**
 * Set display
 * @param {Element} root
 * @private
 */
DvtDataGrid.prototype._showHeader = function(root)
{
    root['style']['display'] = '';
};

/**
 * Build a header div
 * @param {string} axis - 'row' or 'column'
 * @param {string} styleClass - class to set on the header
 * @param {string} endStyleClass - class to set on the end header
 * @return {Object} contains the root and endRoot of the header axis
 */
DvtDataGrid.prototype.buildHeaders = function(axis, styleClass, endStyleClass)
{
    var root, endRoot, scroller, endScroller, scrollerClassName;

    scrollerClassName =  this.getMappedStyle("scroller") + (this.m_utils.isTouchDevice() ? " " + this.getMappedStyle("scroller-mobile") : "");

    root = document.createElement("div");
    root['id'] = this.createSubId(axis + "Header");
    root['className'] = styleClass + " " + this.getMappedStyle("header");
    scroller = document.createElement("div");
    scroller['className'] = scrollerClassName;
    root.appendChild(scroller); //@HTMLUpdateOK

    endRoot = document.createElement("div");
    endRoot['id'] = this.createSubId(axis + "EndHeader");
    endRoot['className'] = endStyleClass + " " + this.getMappedStyle("endheader");
    endScroller = document.createElement("div");
    endScroller['className'] = scrollerClassName;
    endRoot.appendChild(endScroller); //@HTMLUpdateOK

    if (axis === 'column')
    {
        this.m_colHeader = root;
        this.m_colEndHeader = endRoot;
    }
    else if (axis === 'row')
    {
        this.m_rowHeader = root;
        this.m_rowEndHeader = endRoot;
    }

    this.fetchHeaders(axis, 0, root, endRoot, null, null);

    return {root: root, endRoot:endRoot};
};

/**
 * Fetch the headers by calling the fetch headers method on the data source. Pass
 * callbacks for success and error to the data source.
 * @param {string} axis - 'row' or 'column'
 * @param {number} start - index to start fetching at
 * @param {Element|DocumentFragment} header - the root element of the axis header
 * @param {Element|DocumentFragment} endHeader - the root element of the axis endHeader
 * @param {number|null=} fetchSize - number of headers to fetch
 * @param {Object=} callbacks - the optional callbacks to invoke when the fetch success or fail
 * @protected
 */
DvtDataGrid.prototype.fetchHeaders = function(axis, start, header, endHeader, fetchSize, callbacks)
{
    var headerRange, successCallback;
    // check if we are already fetching
    if (this.m_fetching[axis])
    {
        return;
    }

    // fetch size could be explicitly specified.  If not, use the calculated one.
    if (fetchSize == undefined)
    {
        fetchSize = this.getFetchCount(axis, start);
    }

    headerRange = {
        "axis": axis, "start": start, "count": fetchSize, "header": header, "endHeader": endHeader
    };

    this.m_fetching[axis] = headerRange;

    // check if overriding callbacks are specified
    if (callbacks != null && callbacks['success'] != null)
    {
        successCallback = callbacks['success'];
    }
    else
    {
        successCallback = this.handleHeadersFetchSuccess;
    }

    this.showStatusText();
    // start fetch
    this._signalTaskStart();
    this.getDataSource().fetchHeaders(headerRange, {
        "success": successCallback, "error": this.handleHeadersFetchError
    }, {'success': this, 'error': this});
};

/**
 * Checks whether header fetch result match the request
 * @param {Object} headerRange the header range for the response
 * @protected
 */
DvtDataGrid.prototype.isHeaderFetchResponseValid = function(headerRange)
{
    var axis, responseCount;

    axis = headerRange['axis'];
    if (this.m_fetching == null)
    {
        return false;
    }
    responseCount = this.m_fetching[axis]['count'];

    // do object reference check, imagine fetching 20 2 consecutive times but
    // the data changed in bewteeen and we accidentally accept the first because
    // the counts are the same
    return (headerRange == this.m_fetching[axis]);
};

/**
 * Checks whether the result is within the current viewport
 * @param {Object} headerRange
 * @private
 */
DvtDataGrid.prototype.isHeaderFetchResponseInViewport = function(headerRange)
{
    var start, returnVal, axis;

    if (!this.m_initialized)
    {
        // initial scroll these are not defined so just return true, or if not inited or if no databody
        return true;
    }

    // the goal of this method is to make sure we haven't scrolled further since the last fetch
    // so our request is still valid, we run a massive risk of running loops if our logic is wrong otherwise
    // as in we continue to request the same thing but it is never valid.
    axis = headerRange['axis'];
    start = headerRange['start'];

    if (axis == 'row')
    {
        returnVal = this._getLongScrollStart(this.m_currentScrollTop, this.m_prevScrollTop, axis);
    }
    else
    {
        returnVal = this._getLongScrollStart(this.m_currentScrollLeft, this.m_prevScrollLeft, axis);
    }

    // return true if the viewport fits inside the fetched range
    return (returnVal['start'] == start);
};

/**
 * Handle a successful call to the data source fetchHeaders
 * @param {Object} startResults a headerSet object returned from the data source
 * @param {Object} headerRange {"axis":,"start":,"count":,"header":}
 * @param {Object} endResults a headerSet object returned from the data source
 * @param {boolean} rowInsert if this is triggered by a row insert event
 * @protected
 */
DvtDataGrid.prototype.handleHeadersFetchSuccess = function(startResults, headerRange, endResults, rowInsert)
{
    var axis, root, endRoot, start, count, scrollOptions, maxRowCount, maxColumnCount;
            
    scrollOptions = this.m_options.getScrollPolicyOptions();
    maxRowCount = scrollOptions ? scrollOptions['maxRowCount'] : null;
    maxColumnCount = scrollOptions ? scrollOptions['maxColumnCount'] : null;

    // validate result matches what we currently asks for
    if (!this.isHeaderFetchResponseValid(headerRange))
    {
        // end fetch
        this._signalTaskEnd();
        // not valid, so ignore result
        return;
    }

    axis = headerRange["axis"];

    // checks if the response covers the viewport
    if (this.isLongScroll() && !this.isHeaderFetchResponseInViewport(headerRange))
    {
        // clear cells fetching flag
        this.m_fetching[axis] = false;
        // store that the header is invalid for the case when there are no cells
        this.m_headerInvalid = true;
        // end fetch
        this._signalTaskEnd();
        return;
    }

    // remove fetching message
    this.m_fetching[axis] = false;

    root = headerRange["header"];
    endRoot = headerRange["endHeader"];
    start = headerRange["start"];
    count = this.getDataSource().getCount(axis);

    if (axis === "column")
    {
        if (startResults != null)
        {
            this.buildColumnHeaders(root, startResults, start, count, false, false);
            if (startResults.getCount() < headerRange['count'] || 
                (maxColumnCount && maxColumnCount > 0 && maxColumnCount === start + startResults.getCount()))
            {
                this.m_stopColumnHeaderFetch = true;
            }
        }
        if (this.m_endColHeader < 0)
        {
            this._hideHeader(root);
            this.m_stopColumnHeaderFetch = true;
        }
        else
        {
            this.m_hasColHeader = true;
        }

        if (endResults != null)
        {
            this.buildColumnEndHeaders(endRoot, endResults, start, count, false, false);
            if (endResults.getCount() < headerRange['count'] || 
                (maxColumnCount && maxColumnCount > 0 && maxColumnCount === start + endResults.getCount()))
            {
                this.m_stopColumnEndHeaderFetch = true;
            }
        }
        if (this.m_endColEndHeader < 0)
        {
            this._hideHeader(endRoot);
            this.m_stopColumnEndHeaderFetch = true;
        }
        else
        {
            this.m_hasColEndHeader = true;
        }
    }
    else if (axis === "row")
    {
        if (startResults != null)
        {
            this.buildRowHeaders(root, startResults, start, count, rowInsert, false);
            if (startResults.getCount() < headerRange['count'] || 
                (maxRowCount && maxRowCount > 0 && maxRowCount === start + startResults.getCount()))
            {
                this.m_stopRowHeaderFetch = true;
            }
        }
        if (this.m_endRowHeader < 0)
        {
            this._hideHeader(root);
            this.m_stopRowHeaderFetch = true;
        }
        else
        {
            this.m_hasRowHeader = true;
        }

        if (endResults != null)
        {
            this.buildRowEndHeaders(endRoot, endResults, start, count, rowInsert, false);
            if (endResults.getCount() < headerRange['count'] ||
                (maxRowCount && maxRowCount > 0 && maxRowCount === start + endResults.getCount()))
            {
                this.m_stopRowEndHeaderFetch = true;
            }
        }
        if (this.m_endRowEndHeader < 0)
        {
            this._hideHeader(endRoot);
            this.m_stopRowEndHeaderFetch = true;
        }
        else
        {
            this.m_hasRowEndHeader = true;
        }
    }

    if (this.isFetchComplete())
    {
        this.hideStatusText();
        if (this._shouldInitialize() && !rowInsert)
        {
            this._handleInitialization(true);
        }
    }

    if (this.m_initialized)
    {
        //if there are no cells and we are initialized then size the scroller
        this._sizeDatabodyScroller();

        // we cannot syncScroller here. On touch this will trigger a refetch before the fetchCells has been called and will
        // cause an infinite loop. We always call fetchCells after fetchHeaders which calls syncScroller
        // check if we need to sync header scroll position
    }

    // end fetch
    this._signalTaskEnd();
};

/**
 * Handle an unsuccessful call to the data source fetchHeaders
 * @param {Error} error - the error returned from the data source
 * @param {Object} headerRange - keys of {axis,start,count,header}
 */
DvtDataGrid.prototype.handleHeadersFetchError = function(error, headerRange)
{
    // remove fetching message
    var axis = headerRange["axis"];
    this.m_fetching[axis] = false;
    // end fetch
    this._signalTaskEnd();
};

/**
 * Build a header context object for a header and return it
 * The header elem and the data can be set to null for cases where there are no headers
 * but varying height and width are desired
 * @param {string} axis - 'row' or 'column'
 * @param {number} index - the index of the header
 * @param {Object|null} data - the data the cell contains
 * @param {Object} metadata - the metadata the cell contains
 * @param {Element|null} elem - the header element
 * @param {number} level - the header level
 * @param {number} extent - the header extent
 * @param {number} depth - the header depth
 * @return {Object} the header context object, keys of {axis,index,data,datagrid}
 */
DvtDataGrid.prototype.createHeaderContext = function(axis, index, data, metadata, elem, level, extent, depth)
{
    var headerContext, prop;
    headerContext = {
    };
    headerContext['axis'] = axis;
    headerContext['index'] = index;
    headerContext['data'] = data;
    headerContext['component'] = this;
    headerContext['datasource'] = this.getDataSource();
    headerContext['level'] = level;
    headerContext['depth'] = depth;
    headerContext['extent'] = extent;

    //set the parent element to the content div
    if (elem != null)
    {
        headerContext['parentElement'] = elem;
    }

    // merge properties from metadata into cell context
    // the properties in metadata would have precedence
    for (prop in metadata)
    {
        if (metadata.hasOwnProperty(prop))
        {
            headerContext[prop] = metadata[prop];
        }
    }

    // invoke callback to allow ojDataGrid to change datagrid reference
    if (this.m_createContextCallback != null)
    {
        this.m_createContextCallback.call(this, headerContext);
    }

    return this.m_fixContextCallback.call(this, headerContext);;
};

/**
 * Construct the column headers
 * @param {Element} headerRoot
 * @param {Object} headerSet
 * @param {number} start
 * @param {number} totalCount
 * @param {boolean} insert
 * @param {boolean} returnAsFragment
 * @returns {DocumentFragment|Object|undefined}
 */
DvtDataGrid.prototype.buildColumnHeaders = function(headerRoot, headerSet, start, totalCount, insert, returnAsFragment)
{
    var returnObj, axis, count, isAppend, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, totalColumnWidth, totalColumnHeight;
    if (this.m_columnHeaderLevelCount == null)
    {
        this.m_columnHeaderLevelCount = headerSet.getLevelCount();
    }
    if (this.m_columnHeaderLevelCount == 0)
    {
        return;
    }

    axis = 'column';
    count = headerSet.getCount();
    isAppend = start > this.m_endColHeader;
    insert = false;
    reference = null;
    atPixel = isAppend ? this.m_endColHeaderPixel : this.m_startColHeaderPixel;
    currentEnd = this.m_endColHeader;
    levelCount = this.m_columnHeaderLevelCount;
    rootClassName = this.getMappedStyle('colheader') + ' ' + this.getMappedStyle('header');
    cellClassName = this.getMappedStyle('headercell') + ' ' + this.getMappedStyle('colheadercell');
    returnAsFragment = false;

    returnObj = this.buildAxisHeaders(headerRoot, headerSet, axis, start, count, isAppend, insert, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, returnAsFragment);

    if (returnAsFragment)
    {
        return returnObj;
    }

    totalColumnWidth = returnObj.totalHeaderDimension;
    totalColumnHeight = returnObj.totalLevelDimension;

    if (totalColumnWidth != 0 && (this.m_avgColWidth == 0 || this.m_avgColWidth == undefined))
    {
        // the average column width should only be set once, it will only change when the column width varies between columns, but
        // in such case the new average column width would not be any more precise than previous one.
        this.m_avgColWidth = totalColumnWidth / count;
    }

    if (!this.m_colHeaderHeight)
    {
        this.m_colHeaderHeight = totalColumnHeight;
        this.setElementHeight(headerRoot, this.m_colHeaderHeight);
    }

    // whether this is adding columns to the left or right
    if (!isAppend)
    {
        // to the left
        this.m_startColHeader = this.m_startColHeader - count;
        this.m_startColHeaderPixel = this.m_startColHeaderPixel - totalColumnWidth;
    }
    else
    {
        // to the right, in case of long scroll this should alwats be the end header of the set
        this.m_endColHeader = start + count - 1;
        this.m_endColHeaderPixel = this.m_endColHeaderPixel + totalColumnWidth;
    }

    if (totalCount == -1)
    {
        totalCount = this.m_endColHeader;
    }

    // stop subsequent fetching if highwatermark scrolling is used and we have reach the last row, flag it.
    if (!this._isCountUnknown('column') && this._isHighWatermarkScrolling() && this.m_endColHeader + 1 >= totalCount)
    {
        this.m_stopColumnHeaderFetch = true;
    }
    else
    {
        this.m_stopColumnHeaderFetch = returnObj.stopFetch;
    }

    // if virtual scrolling may have to adjust at the beginning
    if (this.m_startColHeader == 0 && this.m_startColHeaderPixel != 0)
    {
        this._shiftHeadersAlongAxisInContainer(headerRoot['firstChild'], 0, this.m_startColHeaderPixel * -1, this.getResources().isRTLMode() ? "right" : "left", this.getMappedStyle('colheadercell'));
        this.m_startColHeaderPixel = 0;
    }
};

/**
 * Construct the column end headers
 * @param {Element} headerRoot
 * @param {Object} headerSet
 * @param {number} start
 * @param {number} totalCount
 * @param {boolean} insert
 * @param {boolean} returnAsFragment
 * @returns {DocumentFragment|Object|undefined}
 */
DvtDataGrid.prototype.buildColumnEndHeaders = function(headerRoot, headerSet, start, totalCount, insert, returnAsFragment)
{
    var returnObj, axis, count, isAppend, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, totalColumnWidth, totalColumnHeight;
    if (this.m_columnEndHeaderLevelCount == null)
    {
        this.m_columnEndHeaderLevelCount = headerSet.getLevelCount();
    }
    if (this.m_columnEndHeaderLevelCount == 0)
    {
        return;
    }

    axis = 'columnEnd';
    count = headerSet.getCount();
    isAppend = start > this.m_endColEndHeader;
    insert = false;
    reference = null;
    atPixel = isAppend ? this.m_endColEndHeaderPixel : this.m_startColEndHeaderPixel;
    currentEnd = this.m_endColEndHeader;
    levelCount = this.m_columnEndHeaderLevelCount;
    rootClassName = this.getMappedStyle('colendheader') + ' ' + this.getMappedStyle('endheader');
    cellClassName = this.getMappedStyle('endheadercell') + ' ' + this.getMappedStyle('colendheadercell');
    returnAsFragment = false;

    returnObj = this.buildAxisHeaders(headerRoot, headerSet, axis, start, count, isAppend, insert, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, returnAsFragment);

    if (returnAsFragment)
    {
        return returnObj;
    }

    totalColumnWidth = returnObj.totalHeaderDimension;
    totalColumnHeight = returnObj.totalLevelDimension;

    if (totalColumnWidth != 0 && (this.m_avgColWidth == 0 || this.m_avgColWidth == undefined))
    {
        // the average column width should only be set once, it will only change when the column width varies between columns, but
        // in such case the new average column width would not be any more precise than previous one.
        this.m_avgColWidth = totalColumnWidth / count;
    }

    if (!this.m_colEndHeaderHeight)
    {
        this.m_colEndHeaderHeight = totalColumnHeight;
        this.setElementHeight(headerRoot, this.m_colEndHeaderHeight);
    }

    // whether this is adding columns to the left or right
    if (!isAppend)
    {
        // to the left
        this.m_startColEndHeader = this.m_startColEndHeader - count;
        this.m_startColEndHeaderPixel = this.m_startColEndHeaderPixel - totalColumnWidth;
    }
    else
    {
        // to the right, in case of long scroll this should alwats be the end header of the set
        this.m_endColEndHeader = start + count - 1;
        this.m_endColEndHeaderPixel = this.m_endColEndHeaderPixel + totalColumnWidth;
    }

    if (totalCount == -1)
    {
        totalCount = this.m_endColEndHeader;
    }

    // stop subsequent fetching if highwatermark scrolling is used and we have reach the last row, flag it.
    if (!this._isCountUnknown('column') && this._isHighWatermarkScrolling() && this.m_endColEndHeader + 1 >= totalCount)
    {
        this.m_stopColumnEndHeaderFetch = true;
    }
    else
    {
        this.m_stopColumnEndHeaderFetch = returnObj.stopFetch;
    }

    // if virtual scrolling may have to adjust at the beginning
    if (this.m_startColEndHeader == 0 && this.m_startColEndHeaderPixel != 0)
    {
        this._shiftHeadersAlongAxisInContainer(headerRoot['firstChild'], 0, this.m_startColEndHeaderPixel * -1, this.getResources().isRTLMode() ? "right" : "left", this.getMappedStyle('colendheadercell'));
        this.m_startColEndHeaderPixel = 0;
    }
};

/**
 * Construct the row headers
 * @param {Element} headerRoot
 * @param {Object} headerSet
 * @param {number} start
 * @param {number} totalCount
 * @param {boolean} insert
 * @param {boolean} returnAsFragment
 * @returns {DocumentFragment|Object|undefined}
 */
DvtDataGrid.prototype.buildRowHeaders = function(headerRoot, headerSet, start, totalCount, insert, returnAsFragment)
{
    var returnObj, axis, count, isAppend, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, totalRowHeight, totalRowWidth, prev;
    if (this.m_rowHeaderLevelCount == null)
    {
        this.m_rowHeaderLevelCount = headerSet.getLevelCount();
    }
    if (this.m_rowHeaderLevelCount == 0)
    {
        return;
    }

    axis = 'row';
    count = headerSet.getCount();
    isAppend = start > this.m_endRowHeader;
    atPixel = isAppend ? this.m_endRowHeaderPixel : this.m_startRowHeaderPixel;
    if (insert)
    {
        reference = headerRoot['firstChild']['childNodes'][start - this.m_startRowHeader];
        atPixel = this.getElementDir(reference, 'top');
    }
    else
    {
        reference = null;
    }
    currentEnd = this.m_endRowHeader;
    levelCount = this.m_rowHeaderLevelCount;
    rootClassName = this.getMappedStyle('rowheader') + ' ' + this.getMappedStyle('header');
    cellClassName = this.getMappedStyle('headercell') + ' ' + this.getMappedStyle('rowheadercell');

    returnObj = this.buildAxisHeaders(headerRoot, headerSet, axis, start, count, isAppend, insert, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, returnAsFragment);

    totalRowHeight = returnObj.totalHeaderDimension;
    totalRowWidth = returnObj.totalLevelDimension;

    if (returnAsFragment)
    {
        return returnObj;
    }

    if (totalRowHeight != 0 && (this.m_avgRowHeight == 0 || this.m_avgRowHeight == undefined))
    {
        // the average row height should only be set once, it will only change when the row height varies between rows, but
        // in such case the new average row height would not be any more precise than previous one.
        this.m_avgRowHeight = totalRowHeight / count;
    }

    if (!this.m_rowHeaderWidth)
    {
        this.m_rowHeaderWidth = totalRowWidth;
        this.setElementWidth(headerRoot, this.m_rowHeaderWidth);
    }

    if (isAppend)
    {
        //if appending a row header, make sure the previous fragment has a bottom border if it was the last
        if (this.m_endRowHeader != -1 && count != 0)
        {
            //get the last header in the scroller
            prev = headerRoot['firstChild']['childNodes'][this.m_endRowHeader - this.m_startRowHeader];
            if (prev != null)
            {
                this.m_utils.removeCSSClassName(prev, this.getMappedStyle('borderHorizontalNone'));
            }
        }
        //in case of a long scroll the end should always be the start plus the count - 1 for 0 indexing
        this.m_endRowHeader = start + count - 1;
        this.m_endRowHeaderPixel = this.m_endRowHeaderPixel + totalRowHeight;
    }
    else if (insert)
    {
        if (start < this.m_startRowHeader)
        {
            // added before the start
            this.m_startRowHeader = start;
            this.m_startRowHeaderPixel = Math.max(0, this.m_startRowHeaderPixel - totalRowHeight);
        }
        //update the endRowHeader and endRowheaderPixel no matter where we insert
        this.m_endRowHeader = this.m_endRowHeader + count;
        this.m_endRowHeaderPixel = Math.max(0, this.m_endRowHeaderPixel + totalRowHeight);
        this.pushRowHeadersDown(reference, totalRowHeight);
    }
    else
    {
        this.m_startRowHeader = Math.max(0, this.m_startRowHeader - count);
        // zero maximum is handled below by realigning when appropriate
        this.m_startRowHeaderPixel = this.m_startRowHeaderPixel - totalRowHeight;
    }

    if (totalCount == -1)
    {
        totalCount = this.m_endRowHeader;
    }

    // stop subsequent fetching if highwatermark scrolling is used and we have reach the last row, flag it.
    if (!this._isCountUnknown('row') && this._isHighWatermarkScrolling() && this.m_endRowHeader + 1 >= totalCount)
    {
        this.m_stopRowHeaderFetch = true;
    }
    else
    {
        this.m_stopRowHeaderFetch = returnObj.stopFetch;
    }

    // if virtual scrolling may have to adjust at the beginning
    if (this.m_startRowHeader == 0 && this.m_startRowHeaderPixel != 0)
    {
        this._shiftHeadersAlongAxisInContainer(headerRoot['firstChild'], 0, this.m_startRowHeaderPixel * -1, "top", this.getMappedStyle('rowheadercell'));
        this.m_startRowHeaderPixel = 0;
    }
};

/**
 * Construct the row end headers
 * @param {Element} headerRoot
 * @param {Object} headerSet
 * @param {number} start
 * @param {number} totalCount
 * @param {boolean} insert
 * @param {boolean} returnAsFragment
 * @returns {DocumentFragment|Object|undefined}
 */
DvtDataGrid.prototype.buildRowEndHeaders = function(headerRoot, headerSet, start, totalCount, insert, returnAsFragment)
    {
    var returnObj, axis, count, isAppend, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, totalRowHeight, totalRowWidth, prev;
    if (this.m_rowEndHeaderLevelCount == null)
    {
        this.m_rowEndHeaderLevelCount = headerSet.getLevelCount();
    }
    if (this.m_rowEndHeaderLevelCount == 0)
    {
        return;
    }

    axis = 'rowEnd';
    count = headerSet.getCount();
    isAppend = start > this.m_endRowEndHeader;
    atPixel = isAppend ? this.m_endRowEndHeaderPixel : this.m_startRowEndHeaderPixel;
    if (insert)
    {
        reference = headerRoot['firstChild']['childNodes'][start - this.m_startRowEndHeader];
        atPixel = this.getElementDir(reference, 'top');
    }
    else
    {
        reference = null;
    }
    currentEnd = this.m_endRowEndHeader;
    levelCount = this.m_rowEndHeaderLevelCount;
    rootClassName = this.getMappedStyle('rowendheader') + ' ' + this.getMappedStyle('endheader');
    cellClassName = this.getMappedStyle('endheadercell') + ' ' + this.getMappedStyle('rowendheadercell');

    returnObj = this.buildAxisHeaders(headerRoot, headerSet, axis, start, count, isAppend, insert, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, returnAsFragment);

    if (returnAsFragment)
    {
        return returnObj;
    }

    totalRowHeight = returnObj.totalHeaderDimension;
    totalRowWidth = returnObj.totalLevelDimension;

    if (returnAsFragment)
    {
        return returnObj;
    }

    if (totalRowHeight != 0 && (this.m_avgRowHeight == 0 || this.m_avgRowHeight == undefined))
    {
        // the average row height should only be set once, it will only change when the row height varies between rows, but
        // in such case the new average row height would not be any more precise than previous one.
        this.m_avgRowHeight = totalRowHeight / count;
    }

    if (!this.m_rowEndHeaderWidth)
    {
        this.m_rowEndHeaderWidth = totalRowWidth;
        this.setElementWidth(headerRoot, this.m_rowEndHeaderWidth);
    }

    if (isAppend)
    {
        //if appending a row header, make sure the previous fragment has a bottom border if it was the last
        if (this.m_endRowEndHeader != -1 && count != 0)
        {
            //get the last header in the scroller
            prev = headerRoot['firstChild']['childNodes'][this.m_endRowEndHeader - this.m_startRowEndHeader];
            if (prev != null)
            {
                this.m_utils.removeCSSClassName(prev, this.getMappedStyle('borderHorizontalNone'));
            }
        }
        //in case of a long scroll the end should always be the start plus the count - 1 for 0 indexing
        this.m_endRowEndHeader = start + count - 1;
        this.m_endRowEndHeaderPixel = this.m_endRowEndHeaderPixel + totalRowHeight;
    }
    else if (insert)
    {
        if (start < this.m_startRowEndHeader)
        {
            // added before the start
            this.m_startRowEndHeader = start;
            this.m_startRowEndHeaderPixel = Math.max(0, this.m_startRowEndHeaderPixel - totalRowHeight);

        }
        //update the endRowEndHeader and endRowEndHeaderPixel no matter where we insert
        this.m_endRowEndHeader = this.m_endRowEndHeader + count;
        this.m_endRowEndHeaderPixel = Math.max(0, this.m_endRowEndHeaderPixel + totalRowHeight);
        this.pushRowHeadersDown(reference, totalRowHeight);
    }
    else
    {
        this.m_startRowEndHeader = Math.max(0, this.m_startRowEndHeader - count);
        // zero maximum is handled below by realigning
        this.m_startRowEndHeaderPixel = this.m_startRowEndHeaderPixel - totalRowHeight;
    }

    if (totalCount == -1)
    {
        totalCount = this.m_endRowEndHeader;
    }

    // stop subsequent fetching if highwatermark scrolling is used and we have reach the last row, flag it.
    if (!this._isCountUnknown('row') && this._isHighWatermarkScrolling() && this.m_endRowEndHeader + 1 >= totalCount)
    {
        this.m_stopRowEndHeaderFetch = true;
    }
    else
    {
        this.m_stopRowEndHeaderFetch = returnObj.stopFetch;
    }


    // if virtual scrolling may have to adjust at the beginning
    if (this.m_startRowEndHeader == 0 && this.m_startRowEndHeaderPixel != 0)
    {
        this._shiftHeadersAlongAxisInContainer(headerRoot['firstChild'], 0, this.m_startRowEndHeaderPixel * -1, "top", this.getMappedStyle('rowendheadercell'));
        this.m_startRowEndHeaderPixel = 0;
    }
};

/**
 * Build headers from the axis info provided
 * @param {Element} headerRoot
 * @param {Object} headerSet
 * @param {string} axis
 * @param {number} start
 * @param {number} count
 * @param {boolean} isAppend
 * @param {boolean} insert
 * @param {Element|null|undefined} reference
 * @param {number} atPixel
 * @param {number} currentEnd
 * @param {number} levelCount
 * @param {string} rootClassName
 * @param {string} cellClassName
 * @param {boolean} returnAsFragment
 * @returns {Object}
 */
DvtDataGrid.prototype.buildAxisHeaders = function(headerRoot, headerSet, axis, start, count, isAppend, insert, reference, atPixel, currentEnd, levelCount, rootClassName, cellClassName, returnAsFragment)
{
    var scroller, stopFetch, renderer, totalHeaderDimension, totalLevelDimension, fragment, x, styleDir, index, returnVal, top, left, columns;
    columns = axis.indexOf('column') != -1;
    styleDir = columns ? 'height': 'width';
    stopFetch = false;
    totalHeaderDimension = 0;
    totalLevelDimension = 0;
    left = 0;
    top = 0;

    if (!returnAsFragment)
    {
        // if unknown count and there's no more column, mark it so we won't try to fetch again
        if (count == 0 && this._isCountUnknown(axis))
        {
            stopFetch = true;
            return {totalHeaderDimension:totalHeaderDimension, totalLevelDimension:totalLevelDimension, stopFetch:stopFetch};
        }
        scroller = headerRoot['firstChild'];
        //add class name back if header populated later
        if (currentEnd == -1 && headerRoot['className'] == '')
        {
            headerRoot['className'] = rootClassName;
            headerRoot['style'][styleDir] = '';
            scroller['style'][styleDir] = '';
        }
    }

    renderer = this.m_options.getRenderer(axis);
    fragment = document.createDocumentFragment();
    x = 0;
    while (count - x > 0)
    {
        if (isAppend || insert)
        {
            index = start + x;
            left = columns ? atPixel + totalHeaderDimension : 0;
            top = columns ? 0 : atPixel + totalHeaderDimension;
        }
        else
        {
            index = start + (count - 1 - x);
            left = columns ? atPixel - totalHeaderDimension : 0;
            top = columns ? 0 : atPixel - totalHeaderDimension;
        }

        returnVal = this.buildLevelHeaders(fragment, index, 0, left, top, isAppend, insert, renderer, headerSet, axis, cellClassName, levelCount);
        //increment the count over the extent of the header
        x += returnVal.count;
        totalHeaderDimension += returnVal.totalHeaderDimension;
        if (returnVal.totalLevelDimension > totalLevelDimension)
        {
            totalLevelDimension = returnVal.totalLevelDimension;
        }
    }

    if (returnAsFragment)
    {
        return fragment;
    }

    if (isAppend)
    {
        scroller.appendChild(fragment); //@HTMLUpdateOK
    }
    else if (insert)
    {
        scroller.insertBefore(fragment, reference); //@HTMLUpdateOK
    }
    else
    {
        scroller.insertBefore(fragment, scroller['firstChild']); //@HTMLUpdateOK
    }

    if (!headerRoot.hasChildNodes() && !insert)
    {
        headerRoot.appendChild(scroller); //@HTMLUpdateOK
    }

    return {totalHeaderDimension:totalHeaderDimension, totalLevelDimension:totalLevelDimension, stopFetch:stopFetch};
};

/**
 * This method is used to call the renderer
 * @param {Function|undefined|null} renderer
 * @param {Object} context cellContext or headerContext
 * @param {Element} cell cell or header to append content to
 * @param {Object|string} data data for the cell
 * @private
 */
DvtDataGrid.prototype._renderContent = function(renderer, context, cell, data)
{
    var content;
    if (renderer != null)
    {
        content = renderer.call(this, context);
        if (content != null)
        {
            // allow return of document fragment from jquery create/js document.createDocumentFragment
            if (content['parentNode'] === null || content['parentNode'] instanceof DocumentFragment)
            {
                cell.appendChild(content); //@HTMLUpdateOK
                this.m_subtreeAttachedCallback(content);
            }
            else if (content['parentNode'] != null)
            {
                // parent node exists, do nothing
            }
            else if (content.toString)
            {
                cell.appendChild(document.createTextNode(content.toString())); //@HTMLUpdateOK
            }
        }

        // make all focusable elements non-focusable, since we want to manage tab stops
        var busyContext = this.m_contextCallback(cell).getBusyContext();
        var self=this;
        busyContext.whenReady().then(function () {
            self._disableAllFocusableElements(cell);
        });
    }
    else
    {
        if (data != null && typeof data === 'object' && data.hasOwnProperty('data'))
        {
            data = data['data'];
        }
        if (data == null)
        {
            data = "";
        }
        cell.appendChild(document.createTextNode(data.toString())); //@HTMLUpdateOK
    }
};

/**
 * Build headers along an axis recursively building them within the set
 * @param {DocumentFragment|Element|undefined} fragment the fragment to append the headers to
 * @param {number} index the index to begin rendering at
 * @param {number} level the level of the header to build at
 * @param {number} left the left value of the headers
 * @param {number} top the top value to start at
 * @param {boolean} isAppend is appending after
 * @param {boolean|undefined|null} insert is row or column insert
 * @param {Function|undefined|null} renderer header renderer
 * @param {Object} headerSet object
 * @param {string} axis column or row
 * @param {string} className string of the class names to be applied to the headers
 * @param {number} totalLevels the number of levels on the header
 * @returns {Object}
 */
DvtDataGrid.prototype.buildLevelHeaders = function(fragment, index, level, left, top, isAppend, insert, renderer, headerSet, axis, className, totalLevels)
{
    var levelDimension, headerDimension, dimensionToAdjust, dimensionToAdjustValue, dimensionSecond, dimensionSecondValue,
            start, end, extentInfo, headerExtent, patchBefore, patchAfter, groupingContainer, levelDimensionCache,
            header, returnVal, i, levelDimensionValue, headerDimensionValue, totalHeaderDimensionValue, headerCount, headerData, headerMetadata, headerContext,
            headerDepth, inlineStyle, styleClass, nextIndex, totalLevelDimensionValue, returnObj, sortIcon, d, groupingRoot, dimensionAxis;

    levelDimensionValue = 0;
    totalLevelDimensionValue = 0;
    totalHeaderDimensionValue = 0;
    headerCount = 0;

    if (axis === 'row')
    {
        dimensionAxis = 'row';
        groupingRoot = this.m_rowHeader;
        levelDimension = 'width';
        levelDimensionCache = this.m_rowHeaderLevelWidths;
        headerDimension = 'height';
        dimensionToAdjust = 'top';
        dimensionToAdjustValue = top;
        dimensionSecond = this.getResources().isRTLMode() ? "right" : "left";
        dimensionSecondValue = left;
        start = this.m_startRowHeader;
        end = this.m_endRowHeader;
    }
    else if (axis == 'rowEnd')
    {
        dimensionAxis = 'row';
        groupingRoot = this.m_rowEndHeader;
        levelDimension = 'width';
        levelDimensionCache = this.m_rowEndHeaderLevelWidths;
        headerDimension = 'height';
        dimensionToAdjust = 'top';
        dimensionToAdjustValue = top;
        dimensionSecond = this.getResources().isRTLMode() ? "left" : "right";
        dimensionSecondValue = left;
        start = this.m_startRowEndHeader;
        end = this.m_endRowEndHeader;
    }
    else if (axis == 'column')
    {
        dimensionAxis = 'column';
        groupingRoot = this.m_colHeader;
        levelDimension = 'height';
        levelDimensionCache = this.m_columnHeaderLevelHeights;
        headerDimension = 'width';
        dimensionToAdjust = this.getResources().isRTLMode() ? "right" : "left";
        dimensionToAdjustValue = left;
        dimensionSecond = 'top';
        dimensionSecondValue = top;
        start = this.m_startColHeader;
        end = this.m_endColHeader;
    }
    else
    {
        dimensionAxis = 'column';
        groupingRoot = this.m_colEndHeader;
        levelDimension = 'height';
        levelDimensionCache = this.m_columnEndHeaderLevelHeights;
        headerDimension = 'width';
        dimensionToAdjust = this.getResources().isRTLMode() ? "right" : "left";
        dimensionToAdjustValue = left;
        dimensionSecond = 'bottom';
        dimensionSecondValue = top;
        start = this.m_startColEndHeader;
        end = this.m_endColEndHeader;
    }

    //get the extent info
    extentInfo = headerSet.getExtent(index, level);
    headerExtent = extentInfo['extent'];
    patchBefore = extentInfo['more']['before'];
    patchAfter = extentInfo['more']['after'];
    headerDepth = headerSet.getDepth(index, level);

    // if the data source says to patch before this header
    // and the index is 1 more than what is currently in the viewport
    // get the groupingContainer and add to it
    if (patchBefore && index === end + 1)
    {
        // get the grouping of the container at the previous index
        groupingContainer = this._getHeaderContainer(index - 1, level, 0, null, groupingRoot, totalLevels);
        // increment the extent stored in the grouping container
        this._setAttribute(groupingContainer, 'extent', this._getAttribute(groupingContainer, 'extent', true) + headerExtent);
        header = groupingContainer['firstChild'];
        levelDimensionValue = this.getElementDir(header, levelDimension);
        // add columns to that grouping container
        for (i = 0; i < headerExtent;)
        {
            if (axis === 'column' || axis === 'columnEnd')
            {
                returnVal = this.buildLevelHeaders(groupingContainer, index + i, level + headerDepth, dimensionToAdjustValue, dimensionSecondValue + levelDimensionValue, isAppend, insert, renderer, headerSet, axis, className, totalLevels);
            }
            else
            {
                returnVal = this.buildLevelHeaders(groupingContainer, index + i, level + headerDepth, dimensionSecondValue + levelDimensionValue, dimensionToAdjustValue, isAppend, insert, renderer, headerSet, axis, className, totalLevels);
            }
            // increment the left and total and count and skip ahead to the next header
            dimensionToAdjustValue += returnVal.totalHeaderDimension;
            totalHeaderDimensionValue += returnVal.totalHeaderDimension;
            headerCount += returnVal.count;
            i += returnVal.count;
        }
        // adjust the header size based on the total of the new sizes passed back
        this.setElementDir(header, this.getElementDir(header, headerDimension) + totalHeaderDimensionValue, headerDimension);
    }
    // if the data source says to patch after this header
    // and the index is 1 less than what is currently in the viewport
    // get the groupingContainer and add to it
    else if (patchAfter && index === start - 1)
    {
        // get the grouping of the container at the previous index
        groupingContainer = this._getHeaderContainer(index + 1, level, 0, null, groupingRoot, totalLevels);
        // increment the extent stored in the grouping container
        this._setAttribute(groupingContainer, 'extent', this._getAttribute(groupingContainer, 'extent', true) + headerExtent);
        // decrement the start stored in the grouping container, since inserting before
        this._setAttribute(groupingContainer, 'start', this._getAttribute(groupingContainer, 'start', true) - headerExtent);
        header = groupingContainer['firstChild'];
        levelDimensionValue = this.getElementDir(header, levelDimension);
        for (i = 0; i < headerExtent;)
        {
            if (axis === 'column' || axis === 'columnEnd')
            {
                returnVal = this.buildLevelHeaders(groupingContainer, index - i, level + headerDepth, dimensionToAdjustValue, dimensionSecondValue + levelDimensionValue, isAppend, insert, renderer, headerSet, axis, className, totalLevels);
            }
            else
            {
                returnVal = this.buildLevelHeaders(groupingContainer, index - i, level + headerDepth, dimensionSecondValue + levelDimensionValue, dimensionToAdjustValue, isAppend, insert, renderer, headerSet, axis, className, totalLevels);
            }
            dimensionToAdjustValue -= returnVal.totalHeaderDimension;
            totalHeaderDimensionValue += returnVal.totalHeaderDimension;
            headerCount += returnVal.count;
            i += returnVal.count;
        }
        this.setElementDir(header, this.getElementDir(header, headerDimension) + totalHeaderDimensionValue, headerDimension);
        this.setElementDir(header, dimensionToAdjustValue, dimensionToAdjust);
    }
    else
    {
        //get the information from the headers
        headerData = headerSet.getData(index, level);
        headerMetadata = headerSet.getMetadata(index, level);

        //create the header element and append the content to it
        header = document.createElement("div");

        //build headerContext to pass to renderer
        headerContext = this.createHeaderContext(axis, index, headerData, headerMetadata, header, level, headerExtent, headerDepth);
        header.setAttribute(this.getResources().getMappedAttribute('busyContext'),'');
        this._createUniqueId(header);
        header[this.getResources().getMappedAttribute('context')] = headerContext;
        inlineStyle = this.m_options.getInlineStyle(axis, headerContext);
        styleClass = this.m_options.getStyleClass(axis, headerContext);

        //add inline styles to the column header cell
        if (inlineStyle != null)
        {
            header['style']['cssText'] = inlineStyle;
        }

        //add class names to the column header cell
        header['className'] = className;
        header['className'] += (" " +  this.getMappedStyle('depth') + headerDepth);
        if (styleClass != null)
        {
            header['className'] += " " + styleClass;
        }

        // position the element before getting its dimensions because of half pixel rounding issues in chrome/ie
        this.setElementDir(header, dimensionSecondValue, dimensionSecond);
        this.setElementDir(header, dimensionToAdjustValue, dimensionToAdjust);

        levelDimensionValue = this._getHeaderLevelDimension(level, header, levelDimensionCache, levelDimension, headerDepth);
        this.setElementDir(header, levelDimensionValue, levelDimension);

        //find the size in case it depends on the classes, will be set in the appropriate case
        headerDimensionValue = this._getHeaderDimension(header, headerContext['key'], dimensionAxis, headerDimension);
        this._setAttribute(header, 'depth', headerDepth);

        //if this is an outer level then add a groupingContainer around it
        if (level != totalLevels - 1)
        {
            groupingContainer = document.createElement("div");
            groupingContainer['className'] = this.getMappedStyle('groupingcontainer');
            groupingContainer.appendChild(header); //@HTMLUpdateOK
            this._setAttribute(groupingContainer, 'start', (isAppend || insert) ? index : index - headerExtent + 1);
            this._setAttribute(groupingContainer, 'extent', headerExtent);
            this._setAttribute(groupingContainer, 'level', level);
        }

        if (level + headerDepth == totalLevels)
        {
            // set the px width on the header regardless of unit type currently on it
            this.setElementDir(header, headerDimensionValue, headerDimension);
            totalHeaderDimensionValue += headerDimensionValue;
            headerCount++;
            totalLevelDimensionValue = levelDimensionValue;
            if (!(isAppend || insert))
            {
                dimensionToAdjustValue -= headerDimensionValue;
                this.setElementDir(header, dimensionToAdjustValue, dimensionToAdjust);
            }
        }
        else
        {
            for (i = 0; i < headerExtent; i++)
            {
                nextIndex = (isAppend || insert) ? index + i : index - i;
                if (axis === 'column' || axis === 'columnEnd')
                {
                    returnVal = this.buildLevelHeaders(groupingContainer, nextIndex, level + headerDepth, dimensionToAdjustValue, dimensionSecondValue + levelDimensionValue, isAppend, insert, renderer, headerSet, axis, className, totalLevels);
                }
                else
                {
                    returnVal = this.buildLevelHeaders(groupingContainer, nextIndex, level + headerDepth, dimensionSecondValue + levelDimensionValue, dimensionToAdjustValue, isAppend, insert, renderer, headerSet, axis, className, totalLevels);
                }
                headerDimensionValue = returnVal.totalHeaderDimension;
                dimensionToAdjustValue = (isAppend || insert) ? dimensionToAdjustValue + headerDimensionValue : dimensionToAdjustValue - headerDimensionValue;
                totalHeaderDimensionValue += headerDimensionValue;
                headerCount += returnVal.count;
                i += returnVal.count - 1;
            }
            totalLevelDimensionValue = levelDimensionValue + returnVal.totalLevelDimension;
            if (!(isAppend || insert))
            {
                this.setElementDir(header, dimensionToAdjustValue, dimensionToAdjust);
            }
            this.setElementDir(header, totalHeaderDimensionValue, headerDimension);
        }

        if (axis == 'columnEnd' && this.m_addBorderBottom)
        {
            this.m_utils.addCSSClassName(header, this.getMappedStyle('borderHorizontalSmall'));
        }

        if (axis == 'rowEnd' && this.m_addBorderRight)
        {
            this.m_utils.addCSSClassName(header, this.getMappedStyle('borderVerticalSmall'));
        }

        //add resizable attribute if resize enabled
        if (this._isHeaderResizeEnabled(axis, headerContext))
        {
            this._setAttribute(header, 'resizable', "true");
        }

        // Temporarily add groupingContainer or header into this.m_root to have them in active DOM before rendering contents
        if (groupingContainer != null)
        {
            this.m_root.appendChild(groupingContainer); //@HTMLUpdateOK
        }
        else
        {
            this.m_root.appendChild(header); //@HTMLUpdateOK
        }

        this._renderContent(renderer, headerContext, header, headerData);

        if (axis === 'column')
        {
            // check if we need to render sort icons
            if (this._isSortEnabled(axis, headerContext))
            {
                if (headerMetadata['sortDirection'] != null && this.m_sortInfo == null)
                {
                    this.m_sortInfo = {};
                    this.m_sortInfo['key'] = headerMetadata['key'];
                    this.m_sortInfo['direction'] = headerMetadata['sortDirection'];
                    this.m_sortInfo['axis'] = axis;
                }

                sortIcon = this._buildSortIcon(headerContext);
                header.appendChild(sortIcon); //@HTMLUpdateOK
                this._setAttribute(header, 'sortable', "true");
            }
        }

        // Moves groupingContainer/header from this.m_root to fragment
        if (isAppend || insert)
        {
            // if we are appending to the end, if there is a grouping container append that, if not just append the row header
            if (groupingContainer != null)
            {
                fragment.appendChild(groupingContainer); //@HTMLUpdateOK
            }
            else
            {
                fragment.appendChild(header); //@HTMLUpdateOK
            }
        }
        else
        {
            // if we are not appending to the end
            // if there is a grouping container append that to the fragment
            if (groupingContainer != null)
            {
                // if the fragment already has a firstChild we want to insert before it
                if (fragment['firstChild'])
                {
                    //if the firstChild is a groupingContainer just insert before it
                    if (this.m_utils.containsCSSClassName(fragment['firstChild'], this.getMappedStyle('groupingcontainer')))
                    {
                        fragment.insertBefore(groupingContainer, fragment['firstChild']); //@HTMLUpdateOK
                    }
                    //if the firstChild is a cell need to insert after it
                    else if (this.m_utils.containsCSSClassName(fragment['firstChild'], this.getMappedStyle('headercell')) ||
                        this.m_utils.containsCSSClassName(fragment['firstChild'], this.getMappedStyle('endheadercell')))
                    {
                        fragment.insertBefore(groupingContainer, fragment['firstChild']['nextSibling']); //@HTMLUpdateOK
                    }
                }
                else
                {
                    fragment.appendChild(groupingContainer); //@HTMLUpdateOK
                }
            }
            // if the fragment itself is a grouping container insert before the other grouping containers
            else if (this.m_utils.containsCSSClassName(fragment, this.getMappedStyle('groupingcontainer')))
            {
                fragment.insertBefore(header, fragment['firstChild']['nextSibling']); //@HTMLUpdateOK
            }
            // otherwise just insert the header at the beginning of the fragment
            else
            {
                fragment.insertBefore(header, fragment['firstChild']); //@HTMLUpdateOK
            }
        }
    }

    // do not put borders on last header cell, treat the index as the index + extent
    // needs to be here and not in loop in case of pactching nested headers
    if (axis === 'column' || axis === 'columnEnd')
    {
        if (this._isLastColumn(index + headerExtent - 1))
        {
            this.m_utils.addCSSClassName(header, this.getMappedStyle('borderVerticalNone'));
        }
    }
    //do not put bottom border on last row, pass the index + extent to see if it's the last index
    else if (this._isLastRow(index + headerExtent - 1) && !insert)
    {
        this.m_utils.addCSSClassName(header, this.getMappedStyle('borderHorizontalNone'));
    }

    // return value is the totalHeight of the rendered headers at this level,
    // the total count of headers rendered at that level,
    // and the totalWidth of the levels underneath it
    returnObj = {};
    returnObj.totalLevelDimension = totalLevelDimensionValue;
    returnObj.totalHeaderDimension = totalHeaderDimensionValue;
    returnObj.count = headerCount;
    return returnObj;
};

/**
 * Get the header dimension at a particular level, which will be cached if set once at that level
 * This permits the user to set the level width on the first row header at that width using renderers.
 * If it is not cached get the width
 * @param {number} level the level to get the dimension of
 * @param {Element} element the row header to get the dimension of if not cached
 * @param {Object} cache the  cache to look in
 * @param {string} dimension width/height
 * @param {number} depth
 * @returns {number} the dimension of that level
 * @private
 */
DvtDataGrid.prototype._getHeaderLevelDimension = function(level, element, cache, dimension, depth)
{
    var width = 0, cachedWidth = 0;

    for (var i = 0; i < depth; i++)
    {
        cachedWidth = cache[level + i];
        if (cachedWidth == null)
        {
            width = null;
            break;
        }
        width += cachedWidth;
    }

    if (width != null)
    {
        return width;
    }

    width = this.getElementDir(element, dimension);
    if (depth == 1)
    {
        cache[level] = width;
    }
    return width;
};

/**
 * Get the header container surrounding the headers.
 * The structure of a container is as follows:
 * firstChild: header at that level
 * subsequent children: grouping containers except at the innermost level
 * @param {number|string} index
 * @param {number|string} level
 * @param {number|string} currentLevel
 * @param {Element|Array} headers
 * @param {Element} root
 * @param {number} levelCount
 * @returns {Element|null}
 * @private
 */
DvtDataGrid.prototype._getHeaderContainer = function(index, level, currentLevel, headers, root, levelCount)
{
    var headerIndex, headerExtent, headerDepth, i;
    if (headers == null)
    {
        headers = root['firstChild']['childNodes'];
        // if we are on the scroller children there is no first header so start at the first header in the list
        i = 0;
    }
    else
    {
        // if we are on a groupingContainer skip the first header which should be a row header at that level
        i = 1;
    }
    // if at the innermost level just return the parent
    if (currentLevel === levelCount - 1)
    {
        return headers[0]['parentNode'];
    }

    // loop over all headers skipping firstChild of groups
    while (i < headers.length)
    {
        // if the index is between that header start and start+extent dig deeper
        headerIndex = this._getAttribute(headers[i], 'start', true);
        headerExtent = this._getAttribute(headers[i], 'extent', true);
        headerDepth = this._getAttribute(headers[i]['firstChild'], 'depth', true);
        if (index >= headerIndex && index < headerIndex + headerExtent)
        {
            if (level < currentLevel + headerDepth)
            {
                return headers[i];
            }
            return this._getHeaderContainer(index, level, currentLevel + headerDepth, headers[i]['childNodes'], root, levelCount);
        }
        i++;
    }
    return null;
};

/**
 * Get the header at a particular index and level for a root
 * @param {number|string} index
 * @param {number|string} level
 * @param {Element} root
 * @param {number} totalLevels
 * @param {number} startIndex for that level
 * @returns {Element|null}
 * @private
 */
DvtDataGrid.prototype._getHeaderByIndex = function(index, level, root, totalLevels, startIndex)
{
    var relativeIndex, headerContent, headerContainer, start;
    if (level < 0)
    {
        return null;
    }
    // if there is only one level just get the header by index in the row ehader
    if (totalLevels === 1)
    {
        headerContent = root['firstChild']['childNodes'];
        relativeIndex = index - startIndex;
        return headerContent[relativeIndex];
    }
    // otherwise get the column header container
    headerContainer = this._getHeaderContainer(index, level, 0, null, root, totalLevels);
    if (headerContainer == null)
    {
        return null;
    }

    if (level <= (this._getAttribute(headerContainer, 'level', true) + this._getAttribute(headerContainer['firstChild'], 'depth', true) - 1))
    {
        return headerContainer['firstChild'];
    }

    // if the innermost level then get the child of the container at the index
    start = this._getAttribute(headerContainer, 'start', true);
    relativeIndex = index - start + 1;
    return headerContainer['childNodes'][relativeIndex];
};

/**
 * Get the column header height at a particulat level, which will be cached if set once at that level
 * This permits the user to set the level height on the first column header at that height using renderers.
 * If it is not cached get the height
 * @param {number} level the level to get the height of
 * @param {Element} element the column header to get the height of if not cached
 * @returns {number} the height of that level
 * @private
 */
DvtDataGrid.prototype._getColumnHeaderLevelHeight = function(level, element)
{
    var height;
    if (this.m_columnHeaderLevelHeights == null)
    {
        this.m_columnHeaderLevelHeights = [];
    }

    height = this.m_columnHeaderLevelHeights[level];
    if (height != null)
    {
        return height;
    }
    height = this.getElementHeight(element);
    this.m_columnHeaderLevelHeights[level] = height;
    return height;
};

/**
 * Get the attribute value that we have set in our mapping attribute
 * @param {Element} element
 * @param {string} attributeKey
 * @param {boolean} parse
 * @returns {number|string}
 */
DvtDataGrid.prototype._getAttribute = function(element, attributeKey, parse)
{
    var value = element.getAttribute(this.getResources().getMappedAttribute(attributeKey));
    if (parse)
    {
        return parseInt(value, 10);
    }
    return value;
};

/**
 * Set a mapped attribute
 * @param {Element} element
 * @param {string} attributeKey
 * @param {string|number|boolean} value
 */
DvtDataGrid.prototype._setAttribute = function(element, attributeKey, value)
{
    element.setAttribute(this.getResources().getMappedAttribute(attributeKey), value);
};

/**
 * Build the databody, fetching cells as well
 * @return {Element} the root of databody
 */
DvtDataGrid.prototype.buildDatabody = function()
{
    var root, scroller;
    root = document.createElement("div");
    root['id'] = this.createSubId("databody");
    root['className'] = this.getMappedStyle("databody");
    // workaround for mozilla bug 616594, where overflow div would make it focusable
    root['tabIndex'] = '-1';
    this.m_databody = root;
    if (!root.addEventListener)
    {
        root.attachEvent("onscroll", this.handleScroll.bind(this));
    }
    else
    {
        root.addEventListener("scroll", this.handleScroll.bind(this), false);
    }

    scroller = document.createElement("div");
    scroller['className'] = this.getMappedStyle("scroller") + (this.m_utils.isTouchDevice() ? " " + this.getMappedStyle("scroller-mobile") : "");
    root.appendChild(scroller); //@HTMLUpdateOK

    this.fetchCells(root, 0, 0);

    return root;
};

/**
 * Get the fetch count which is limited by the scollPolicyOptions max count along the axis
 * @private
 */
DvtDataGrid.prototype.getFetchCount = function(axis, start)
{
    var count = this.getFetchSize(axis);
    if (this._isHighWatermarkScrolling())
    {
        var prop = axis === 'row' ? 'maxRowCount' : 'maxColumnCount';
        var maxCount = this.m_options.getScrollPolicyOptions();
        if (maxCount && maxCount[prop] != null && maxCount[prop] > 0)
        {
            count = Math.max(Math.min(count, maxCount[prop] - start), 0);
        }
    }
    return count
};

/**
 * Fetch cells to put in the databody. Calls fetch cells on the data source,
 * setting callbacks for success and failure.
 * @param {Element} databody - the root of the databody element
 * @param {number} rowStart - the row to start fetching at
 * @param {number} colStart - the column to start fetching at
 * @param {number|null=} rowCount - the total number of rows in the data source, if undefined then calculated
 * @param {number|null=} colCount - the total number of columns in the data source, if undefined then calculated
 * @param {Object=} callbacks - specifies success and error callbacks.  If undefined then default callbacks are used
 * @protected
 */
DvtDataGrid.prototype.fetchCells = function(databody, rowStart, colStart, rowCount, colCount, callbacks)
{
    var rowRange, columnRange, successCallback;

    // checks if we are already fetching cells
    if (this.m_fetching['cells'])
    {
        return;
    }

    if (rowCount == null)
    {
        rowCount = this.getFetchCount("row", rowStart);
    }

    if (colCount == null)
    {
        colCount = this.getFetchCount("column", colStart);
    }

    rowRange = {
        "axis": "row", "start": rowStart, "count": rowCount
    };
    columnRange = {
        "axis": "column", "start": colStart, "count": colCount, "databody": databody
    };

    this.m_fetching['cells'] = {'rowRange': rowRange, 'columnRange': columnRange};

    // if there is a override success callback specified, use it, otherwise use default one
    if (callbacks != null && callbacks['success'] != null)
    {
        successCallback = callbacks['success'];
    }
    else
    {
        successCallback = this.handleCellsFetchSuccess;
    }

    this.showStatusText();
    // start fetch
    this._signalTaskStart();
    this.getDataSource().fetchCells([rowRange, columnRange], {
        "success": successCallback, "error": this.handleCellsFetchError
    }, {'success': this, 'error': this});
};

/**
 * Checks whether the response matches the current request
 * @param {Object} cellRange the cell range of the response
 * @protected
 */
DvtDataGrid.prototype.isCellFetchResponseValid = function(cellRange)
{
    var responseRowRange, responseColumnRange, requestCellRanges;

    responseRowRange = cellRange[0];
    responseColumnRange = cellRange[1];

    if (this.m_fetching == null)
    {
        return false;
    }

    requestCellRanges = this.m_fetching['cells'];

    // do object reference check, imagine fetching 20 2 consecutive times but
    // the data changed in bewteeen and we accidentally accept the first because
    // the counts are the same
    return (responseRowRange == requestCellRanges['rowRange'] && responseColumnRange == requestCellRanges['columnRange']);
};

/**
 * Returns true if this is a long scroll (or initial scroll)
 * @return {boolean} true if it is a long or initial scroll, false otherwise
 */
DvtDataGrid.prototype.isLongScroll = function()
{
    return this.m_isLongScroll;
};

/**
 * Checks whether the result is within the current viewport
 * @param {Object} cellSet - a CellSet object which encapsulates the result set of cells
 * @param {Array.<Object>} cellRange - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 * @private
 */
DvtDataGrid.prototype.isCellFetchResponseInViewport = function(cellSet, cellRange)
{
    var rowRange, rowStart, columnRange, columnStart,  rowReturnVal, columnReturnVal;

    if (isNaN(this.m_avgRowHeight) || isNaN(this.m_avgColWidth) || this.m_empty != null || !this.m_initialized)
    {
        // initial scroll these are not defined so just return true, or if not inited or if no databody
        return true;
    }

    // the goal of this method is to make sure we haven't scrolled further since the last fetch
    // so our request is still valid, we run a massive risk of running loops if our logic is wrong otherwise
    // as in we continue to request the same thing but it is never valid.

    rowRange = cellRange[0];
    rowStart = rowRange['start'];

    columnRange = cellRange[1];
    columnStart = columnRange['start'];

    rowReturnVal = this._getLongScrollStart(this.m_currentScrollTop, this.m_prevScrollTop, 'row');
    columnReturnVal = this._getLongScrollStart(this.m_currentScrollLeft, this.m_prevScrollLeft, 'column');

    // return true if the viewport fits inside the fetched range
    return (rowReturnVal['start'] == rowStart && columnReturnVal['start'] == columnStart);
};

/**
 * Handle a successful call to the data source fetchCells. Create new row and
 * cell DOM elements when necessary and then insert them into the databody.
 * @param {Object} cellSet - a CellSet object which encapsulates the result set of cells
 * @param {Array.<Object>} cellRange - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":}]
 * @param {boolean=} rowInsert - if this is triggered by a row insert event
 * @protected
 */
DvtDataGrid.prototype.handleCellsFetchSuccess = function(cellSet, cellRange, rowInsert)
{
    var totalRowCount, totalColumnCount, rowRange, rowStart, rowCount,
            rowRangeNeedsUpdate, columnRange, columnStart, columnCount, columnRangeNeedsUpdate,
            databody, top, referenceCell, databodyContent, cleanDirection, left,
            isAppend, fragment, totalColumnWidth, totalRowHeight, i, avgHeight, avgWidth, duration, self,
            rows, prev, addResult, scrollOptions, maxRowCount, maxColumnCount;
            
    scrollOptions = this.m_options.getScrollPolicyOptions();
    maxRowCount = scrollOptions ? scrollOptions['maxRowCount'] : null;
    maxColumnCount = scrollOptions ? scrollOptions['maxColumnCount'] : null;
    totalRowCount = this.getDataSource().getCount("row");
    totalColumnCount = this.getDataSource().getCount("column");

    self = this;
    duration = DvtDataGrid.EXPAND_ANIMATION_DURATION;

    // if rowInsert is specified we can skip a couple of checks
    if (rowInsert === undefined)
    {
        rowInsert = false;

        // checks whether result matches what we requested
        if (!this.isCellFetchResponseValid(cellRange))
        {
            // end fetch
            this._signalTaskEnd();
            // ignore result if it is not valid
            return;
        }

        // checks if the response covers the viewport or the headers were invalid
        if (this.isLongScroll() && (!this.isCellFetchResponseInViewport(cellSet, cellRange) || this.m_headerInvalid))
        {
            // clear cells fetching flag
            this.m_fetching['cells'] = false;
            this.m_headerInvalid = false;

            // ignore the response and fetch another set for the current viewport
            this.handleLongScroll(this.m_currentScrollLeft, this.m_currentScrollTop);

            // end fetch
            this._signalTaskEnd();
            return;
        }
        else
        {
            this.m_isLongScroll = false;
        }
    }

    rowRange = cellRange[0];
    rowStart = rowRange['start'];
    rowCount = cellSet.getCount("row");

    // for short fetch it would be equal for long fetch it would be > (bottom) or < (top)
    rowRangeNeedsUpdate = rowCount > 0 && (rowStart > this.m_endRow || rowStart + rowCount <= this.m_startRow);

    // if no results returned and count is unknown, flag it so we won't try to fetch again
    // OR if highwater mark scrolling is used and count is known and we have reach the last row, stop fetching
    // OR if result set is less than what's requested, then assumes we have fetched the last row
    if ((rowCount == 0 && this._isCountUnknown('row') && rowRange['count'] > 0) ||
            (rowRangeNeedsUpdate && this._isHighWatermarkScrolling() && !this._isCountUnknown('row') && (this.m_endRow + rowCount + 1 >= totalRowCount)) ||
            (rowCount < rowRange['count']) || 
            (this._isHighWatermarkScrolling() && maxRowCount && maxRowCount > 0 && rowStart + rowCount === maxRowCount) )
    {
        this.m_stopRowFetch = true;
    }

    columnRange = cellRange[1];
    columnStart = columnRange['start'];
    columnCount = cellSet.getCount("column");

    columnRangeNeedsUpdate = columnCount > 0 && (columnStart > this.m_endCol || columnStart + columnCount == this.m_startCol);

    // if no results returned and count is unknown, flag it so we won't try to fetch again
    // OR if highwater mark scrolling is used and count is known and we have reach the last column, stop fetching
    // OR if result set is less than what's requested, then assumes we have fetched the last column
    if ((columnCount == 0 && this._isCountUnknown('column') && columnRange['count'] > 0) ||
            (columnRangeNeedsUpdate && this._isHighWatermarkScrolling() && !this._isCountUnknown('column') && (this.m_endCol + columnCount + 1 >= totalColumnCount)) ||
            (columnCount < columnRange['count']) ||
            (this._isHighWatermarkScrolling() && maxColumnCount && maxColumnCount > 0 && columnStart + columnCount === maxColumnCount) )
    {
        this.m_stopColumnFetch = true;
    }

    databody = this.m_databody;
    if (databody == null)
    {
        // try to search for it in the param
        databody = columnRange['databody'];
    }

    databodyContent = databody['firstChild'];

    // if these are new rows (append or insert in the middle)
    if (rowRangeNeedsUpdate || rowInsert)
    {
        // whether this is adding rows to bottom (append) or top (insert)
        isAppend = !rowInsert && rowStart >= this.m_startRow ? true : false;

        if (isAppend)
        {
            top = this.m_endRowPixel;
        }
        else
        {
            if (rowInsert)
            {
                referenceCell = this._getCellByIndex({'row': rowStart + rowCount, 'column': this.m_startCol});
                top = this.getElementDir(referenceCell, 'top');
            }
            else
            {
                top = this.m_startRowPixel;
            }
        }

        left = columnStart >= this.m_startCol ? this.m_startColPixel : this.m_endColPixel;

        fragment = document.createDocumentFragment();
        addResult = this._addCellsToFragment(fragment, cellSet, rowStart, top, columnStart, left);
        totalColumnWidth = addResult['totalColumnWidth'];
        totalRowHeight = addResult['totalRowHeight'];
        avgWidth = addResult['avgWidth'];
        avgHeight = totalRowHeight / rowCount;

        this._populateDatabody(databodyContent, fragment);

        if (isAppend)
        {
            // make sure there is a bottom border if adding a row to the bottom
            if (this.m_endRow != -1 && rowCount != 0)
            {
                //get the previous last row
                this._highlightCellsAlongAxis(this.m_endRow, 'row', 'index', 'remove', ['borderHorizontalNone']);
            }
            // update row range info if neccessary
            this.m_endRow = rowStart + rowCount - 1;
            this.m_endRowPixel = this.m_endRowPixel + totalRowHeight;
        }
        else if (rowInsert)
        {
            // update row range info if neccessary
            if (rowStart < this.m_startRow)
            {
                // added in the middle
                this.m_startRow = rowStart;
                this.m_startRowPixel = Math.max(0, this.m_startRowPixel - totalRowHeight);
            }
            //update the endRow and endRowPixel no matter where we insert
            this.m_endRow = this.m_endRow + rowCount;
            this.m_endRowPixel = this.m_endRowPixel + totalRowHeight;
            this.pushRowsDown(rowStart + rowCount, totalRowHeight);
        }
        else
        {
            // update row range info if neccessary
            this.m_startRow = this.m_startRow - rowCount;
            // zero maximum is handled by realigning
            this.m_startRowPixel = this.m_startRowPixel - totalRowHeight;
        }
    }
    else if (columnRangeNeedsUpdate)
    {
        // whether this is adding cols to right (append) or left (insert)
        isAppend = columnStart >= this.m_startCol ? true : false;

        left = isAppend ? this.m_endColPixel : this.m_startColPixel;
        top = rowStart >= this.m_startRow ? this.m_startRowPixel : this.m_endRowPixel;

        fragment = document.createDocumentFragment();
        addResult = this._addCellsToFragment(fragment, cellSet, rowStart, top, columnStart, left);

        this._populateDatabody(databodyContent, fragment);
        totalColumnWidth = addResult['totalColumnWidth'];
    }

    // added to only do this on initialization
    // check to see if the average width and height has change and update the canvas and the scroller accordingly
    if (avgWidth != undefined && (this.m_avgColWidth == 0 || this.m_avgColWidth == undefined))
    {
        // the average column width should only be set once, it will only change when the column width varies between columns, but
        // in such case the new average column width would not be any more precise than previous one.
        this.m_avgColWidth = avgWidth;
    }

    if (avgHeight != undefined && (this.m_avgRowHeight == 0 || this.m_avgRowHeight == undefined))
    {
        // the average row height should only be set once, it will only change when the row height varies between rows, but
        // in such case the new average row height would not be any more precise than previous one.
        this.m_avgRowHeight = avgHeight;
    }

    // update column range info if neccessary
    if (columnRangeNeedsUpdate)
    {
        // add to left or to right
        if (columnStart < this.m_startCol)
        {
            this.m_startCol = this.m_startCol - columnCount;
            this.m_startColPixel = this.m_startColPixel - totalColumnWidth;
        }
        else
        {
            //in virtual fetch end should always be set to last
            this.m_endCol = columnStart + columnCount - 1;
            this.m_endColPixel = this.m_endColPixel + totalColumnWidth;
        }
    }

    this._sizeDatabodyScroller();

    if (this.m_endCol >= 0 && this.m_endRow >= 0)
    {
        this.m_hasCells = true;
    }

    // if virtual scrolling we may need to adjust when the user hits the beginning
    if (this.m_startCol == 0 && this.m_startColPixel != 0)
    {
        this._shiftCellsAlongAxis('column', -this.m_startColPixel, 0, true)
        this.m_startColPixel = 0;
    }
    if (this.m_startRow == 0 && this.m_startRowPixel != 0)
    {
        this.pushRowsDown(0, -this.m_startRowPixel);
        this.m_startRowPixel = 0;
    }

    // fetch is done
    this.m_fetching['cells'] = false;
    
    // size the grid if fetch is done
    if (this.isFetchComplete())
    {
        this.hideStatusText();

        this._syncScroller();
        
        this._checkScrollPosition();
        
        // highlight focus cell or header if specified
        if (this.m_scrollIndexAfterFetch != null)
        {
            this.scrollToIndex(this.m_scrollIndexAfterFetch);
            //wait for the scroll event to be fired to avoid using cell.focus() to bring into view, the case where it's in the viewport but hasn't been scrolled to yet
        }
        else if (this.m_scrollHeaderAfterFetch != null)
        {
            // if the there is a header that needs to be scrolled to after fetch scroll to the header
            this.scrollToHeader(this.m_scrollHeaderAfterFetch);
        }
        else if (!this.isActionableMode() && this._getActiveElement() != null && !this.m_utils.containsCSSClassName(this._getActiveElement(), this.getMappedStyle('focus')))
        {
            //highliht the active cell if we are virtualized scroll and scrolled away from the active and came back
            //also on a move event insert this will preserve the active cell
            if (this.m_shouldFocus != true)
            {
                this.m_shouldFocus = false;
            }
            this._highlightActive();
            this._manageMoveCursor();
        }

        // apply current selection range to newly fetched cells
        // this is more efficient than looping over ranges when rendering cell
        if (this._isSelectionEnabled())
        {
            this.applySelection(rowStart, rowStart + rowCount, columnStart, columnStart + columnCount);
        }

        // update accessibility info
        this.populateAccInfo();

        // force bitmap (to GPU) to be generated now rather than when doing actual 3d translation to minimize
        // the delay
        if (this.m_utils.isTouchDevice() && window.hasOwnProperty('WebKitCSSMatrix'))
        {
            databody.style.webkitTransform = "translate3d(0, 0, 0)";
            if (this.m_rowHeader != null)
            {
                this.m_rowHeader.style.webkitTransform = "translate3d(0, 0, 0)";
            }
            if (this.m_colHeader != null)
            {
                this.m_colHeader.style.webkitTransform = "translate3d(0, 0, 0)";
            }
            if (this.m_rowEndHeader != null)
            {
                this.m_rowEndHeader.style.webkitTransform = "translate3d(0, 0, 0)";
            }
            if (this.m_colEndHeader != null)
            {
                this.m_colEndHeader.style.webkitTransform = "translate3d(0, 0, 0)";
            }
        }

        // initialize/resize/fillViewport/trigger ready event
        if (this._shouldInitialize())
        {
            this._handleInitialization(true);
        }
        else if (this.m_initialized)
        {
            // cases that require resize internally on fetch:
            // 1: the datagrid root node grew / resize required
            // 2: a delete triggered a fillViewport that shrunk and expanded the databody, the check that the end pixel was below the databody height then beyond it
            if (this.m_resizeRequired == true || (this.m_endRowPixel - totalRowHeight < this.getElementHeight(databody)))
            {
                this.resizeGrid();
            }

            // clean up rows outside of viewport (for non-highwatermark scrolling only)
            if (rowRangeNeedsUpdate)
            {
                if (isAppend)
                {
                    cleanDirection = 'top';
                }
                else if (!rowInsert)
                {
                    cleanDirection = 'bottom';
                }
            }
            else if (columnRangeNeedsUpdate)
            {
                // add to left or to right
                if (columnStart == this.m_startCol)
                {
                    cleanDirection = 'right';
                }
                else
                {
                    cleanDirection = 'left';
                }
            }
            this._cleanupViewport(cleanDirection);

            this.fillViewport();
            if (this.isFetchComplete())
            {
                this.fireEvent('ready', {});
            }
        }
    }

    // update any row/column selections if necessary. 
    if (this._isSelectionEnabled() && this.isMultipleSelection())
    {
        this.updateSelectionHeader();
    }

    // end fetch
    this._signalTaskEnd();
    //this.dumpRanges();
};

/**
 * Set the Z index values before a given row index
 * @param {string} axis
 * @param {number} index
 * @param {boolean} startHeaders
 * @param {boolean} endHeaders
 */
DvtDataGrid.prototype._setZIndexBefore = function(axis, index, startHeaders, endHeaders)
{
    var start, maxPixel, dir, headerContainer, levelCount, headerStart, endHeaderContainer, endLevelCount, endHeaderStart, i, cells, j, header;
    if (axis == 'row')
    {
        start = this.m_startRow;
        maxPixel = this.m_currentScrollTop;
        dir = 'top';
        headerContainer = this.m_rowHeader;
        levelCount = this.m_rowHeaderLevelCount;
        headerStart = this.m_startRowHeader;
        endHeaderContainer = this.m_rowEndHeader;
        endLevelCount = this.m_rowEndHeaderLevelCount;
        endHeaderStart = this.m_startRowEndHeader;
    }
    else
    {
        return;
    }
    
    for (i = index; i >= start; i--)
    {
        cells = this._getAxisCellsByIndex(i, axis);
        if (this.getElementDir(cells[0], dir) < maxPixel)
        {
            break;
        }

        for (j = 0; j < cells.length; j++)
        {
            this.changeStyleProperty(cells[j], this.getCssSupport('z-index'), 10);
        }
        
        if (startHeaders)
        {
            header = this._getHeaderByIndex(i, 0, headerContainer, levelCount, headerStart);
            this.changeStyleProperty(header, this.getCssSupport('z-index'), 10);    
        }

        if (endHeaders)
        {
            header = this._getHeaderByIndex(i, 0, endHeaderContainer, endLevelCount, endHeaderStart);
            this.changeStyleProperty(header, this.getCssSupport('z-index'), 10);    
        }
    }
};

DvtDataGrid.prototype._onTransitionEnd = function(target, handler, duration)
{
  var endEvents = "transitionend";
  var transitionTimer;
  var listener =
    function ()
    {
      if (transitionTimer)
      {
        clearTimeout(transitionTimer);
        transitionTimer = undefined;
      }
      //remove handler
      target.removeEventListener(endEvents, listener);

      handler();
    };

  //add transition end listener
  target.addEventListener(endEvents, listener);

  transitionTimer = 
    setTimeout(listener, duration + 100);
};

/**
 * Insert rows with animation.
 * @param {DocumentFragment|undefined} cellsFragment
 * @param {DocumentFragment|undefined} rowHeaderFragment
 * @param {number} rowStart the starting row index
 * @private
 */
DvtDataGrid.prototype._insertRowsWithAnimation = function(cellsFragment, rowHeaderFragment, rowEndHeaderFragment, rowStart, rowCount, totalRowHeight)
{
    var self, isAppend, databodyContent, rowHeaderSupport, rowHeaderContent, referenceCells, referenceRowHeader, referenceCellTop,
            insertStartPixel, i, row, rowHeader, newTop, deltaY, lastAnimatedElement, transitionListener, rowEndHeaderSupport, rowEndHeaderContent,
            referenceRowEndHeader, rowEndHeader, duration, cell, j, rowCells;

    self = this;
    // animation start
    self._signalTaskStart();
    isAppend = rowStart > this.m_endRow;
    databodyContent = this.m_databody['firstChild'];
    rowHeaderSupport = rowHeaderFragment == null ? false : true;
    rowEndHeaderSupport = rowEndHeaderFragment == null ? false : true;

    // row to be inserted after is the reference row
    var referenceCellsIndex = rowStart - 1;
    referenceCells = this._getAxisCellsByIndex(rowStart - 1, 'row');
    referenceCellTop = this.getElementDir(referenceCells[0], 'top');
    insertStartPixel = referenceCellTop + this.getElementHeight(referenceCells[0]);


    //all inherited animated rows should be hidden under previous rows in view
    this._setZIndexBefore('row', referenceCellsIndex, rowHeaderSupport, rowEndHeaderSupport)

    if (rowHeaderSupport)
    {
        rowHeaderContent = this.m_rowHeader['firstChild'];
        referenceRowHeader = rowHeaderContent['childNodes'][rowStart - this.m_startRow - 1];
    }

    if (rowEndHeaderSupport)
    {
        rowEndHeaderContent = this.m_rowEndHeader['firstChild'];
        referenceRowEndHeader = rowEndHeaderContent['childNodes'][rowStart - this.m_startRow - 1];
    }  
    
    // loop over the fragment and assign proper top values to the fragment and then hide them
    // with transform behind the reference row
    for (i = 0; i < cellsFragment.childNodes.length; i++)
    {
        cell = cellsFragment.childNodes[i];
        newTop = insertStartPixel + this.getElementDir(cell, 'top');
        deltaY = referenceCellTop - newTop - this.getElementHeight(cell);

        // move row to actual new position
        this.setElementDir(cell, newTop, 'top');

        // move row to behind reference row
        this.addTransformMoveStyle(cell, 0, 0, 'linear', 0, deltaY, 0);
    }

    if (rowHeaderSupport)
    {
        for (i = 0; i < rowHeaderFragment.childNodes.length; i++)
        {
            rowHeader = rowHeaderFragment.childNodes[i];
            newTop = insertStartPixel + this.getElementDir(rowHeader, 'top');
            deltaY = referenceCellTop - newTop - this.getElementHeight(rowHeader);
            this.setElementDir(rowHeader, newTop, 'top');
            this.addTransformMoveStyle(rowHeader, 0, 0, 'linear', 0, deltaY, 0);
        }
    }

    if (rowEndHeaderSupport)
    {
        for (i = 0; i < rowEndHeaderFragment.childNodes.length; i++)
        {
            rowEndHeader = rowEndHeaderFragment.childNodes[i];
            newTop = insertStartPixel + this.getElementDir(rowEndHeader, 'top');
            deltaY = referenceCellTop - newTop - this.getElementHeight(rowEndHeader);
            this.setElementDir(rowEndHeader, newTop, 'top');
            this.addTransformMoveStyle(rowEndHeader, 0, 0, 'linear', 0, deltaY, 0);
        }
    }

    // loop over the row after the insert point, assign new top values, but keep
    // them where they are using transforms
    for (i = rowStart - this.m_startRow; i <= this.m_endRow; i++)
    {
        rowCells = this._getAxisCellsByIndex(i, 'row');
        newTop = totalRowHeight + this.getElementDir(rowCells[0], 'top');
        deltaY = -totalRowHeight;

        for (j = 0; j < rowCells.length; j++)
        {
            // move row to actual new position
            this.setElementDir(rowCells[j], newTop, 'top');

            // move row to original position
            this.addTransformMoveStyle(rowCells[j], 0, 0, 'linear', 0, deltaY, 0);
        }

        if (rowHeaderSupport)
        {
            rowHeader = rowHeaderContent.childNodes[i];
            this.setElementDir(rowHeader, newTop, 'top');
            this.addTransformMoveStyle(rowHeader, 0, 0, 'linear', 0, deltaY, 0);
        }
        if (rowEndHeaderSupport)
        {
            rowEndHeader = rowEndHeaderContent.childNodes[i];
            this.setElementDir(rowEndHeader, newTop, 'top');
            this.addTransformMoveStyle(rowEndHeader, 0, 0, 'linear', 0, deltaY, 0);
        }
    }

    this._modifyAxisCellContextIndex('row', rowStart, this.m_endRow - rowStart + 1, rowCount);

    // need to resize first in order to ensure visible region is big enough to handle new rows
    this.m_endRow += rowCount;
    this.m_endRowPixel += totalRowHeight;
    if (rowHeaderSupport)
    {
        this.m_endRowHeader += rowHeaderFragment.childNodes.length;
        this.m_endRowHeaderPixel += totalRowHeight;
    }
    if (rowEndHeaderSupport)
    {
        this.m_endRowEndHeader += rowEndHeaderFragment.childNodes.length;
        this.m_endRowEndHeaderPixel += totalRowHeight;
    }

    // find the row in which the new rows will be inserted and insert
    databodyContent.appendChild(cellsFragment); //@HTMLUpdateOK
    if (isAppend)
    {
        if (rowHeaderSupport)
        {
            rowHeaderContent.appendChild(rowHeaderFragment); //@HTMLUpdateOK
        }
        if (rowEndHeaderSupport)
        {
            rowEndHeaderContent.appendChild(rowEndHeaderFragment); //@HTMLUpdateOK
        }
    }
    else
    {
        if (rowHeaderSupport)
        {
            rowHeaderContent.insertBefore(rowHeaderFragment, referenceRowHeader['nextSibling']); //@HTMLUpdateOK
        }
        if (rowEndHeaderSupport)
        {
            rowEndHeaderContent.insertBefore(rowEndHeaderFragment, referenceRowEndHeader['nextSibling']); //@HTMLUpdateOK
        }
    }
    this.setElementHeight(databodyContent, this.m_endRowPixel - this.m_startRowPixel);
    this.resizeGrid();
    this.updateRowBanding();

    this._refreshDatabodyMap();

    lastAnimatedElement = this._getCellByIndex(this.createIndex(rowStart + rowCount - 1, this.m_endCol));
    transitionListener = function()
    {   
        self.m_stopRowFetch = false;
        if (self.m_endRowHeader != -1)
        {
            self.m_stopRowHeaderFetch = false;
        }
        if (self.m_endRowEndHeader != -1)
        {
            self.m_stopRowEndHeaderFetch = false;
        }
        self._handleAnimationEnd();
    };
        
    this.m_animating = true;

    // must grab duration outside of timeout otherwise processingEventQueue flag would have been reset already
    // note we set the animation duration to 1 instead of 0 because some browsers don't invoke transition end listener if duration is 0
    duration = self.m_processingEventQueue ? 1 : DvtDataGrid.EXPAND_ANIMATION_DURATION;
    
    this._onTransitionEnd(lastAnimatedElement, transitionListener, duration);
    
    setTimeout(function()
    {
        var i, j, timing, rowCells;
        duration = DvtDataGrid.EXPAND_ANIMATION_DURATION;
        timing = 'ease-out';
        //add animation rules to the inserted rows
        for (i = rowStart; i <= self.m_endRow; i++)
        {
            rowCells = self._getAxisCellsByIndex(i, 'row');
            for (j=0; j < rowCells.length; j++)
            {
                self.addTransformMoveStyle(rowCells[j], duration + "ms", 0, timing, 0, 0, 0);
            }
            if (rowHeaderSupport)
            {
                self.addTransformMoveStyle(rowHeaderContent.childNodes[i], duration + "ms", 0, timing, 0, 0, 0);
            }
            if (rowEndHeaderSupport)
            {
                self.addTransformMoveStyle(rowEndHeaderContent.childNodes[i], duration + "ms", 0, timing, 0, 0, 0);
            }
        }        
    }, 0);
};

/**
 * Add cells to the fragment passed in
 * @param {DocumentFragment|undefined} fragment
 * @param {Object} cellSet
 * @param {number} rowStart the starting row index
 * @param {number} topStart the starting top value
 * @param {number} columnStart the starting column index
 * @param {number} leftStart the starting left (right) value
 * @private
 */
DvtDataGrid.prototype._addCellsToFragment = function(fragment, cellSet, rowStart, topStart, columnStart, leftStart)
{
    var renderer, columnBandingInterval, rowBandingInterval, horizontalGridlines, verticalGridlines, rowCount,
        columnCount, i, rowIndex, applyRowBanding, left, dir, totalWidth, j, columnIndex, indexes, cellData, cellMetadata, cellExtents,
        rowSpan, columnSpan, cell, cellContext, rowKey, columnAppend, cellExtent, rowAppend, columnKey, tempRowIndex, tempArray,
        shimHeaderContext, shimHeader, inlineStyle, styleClass, height, width, bottom, totalHeight, top, inlineStyleClass, avgWidth, extents,
        patched, k;

    dir = this.getResources().isRTLMode() ? "right" : "left";

    renderer = this.m_options.getRenderer("cell");
    columnBandingInterval = this.m_options.getColumnBandingInterval();
    rowBandingInterval = this.m_options.getRowBandingInterval();
    horizontalGridlines = this.m_options.getHorizontalGridlines();
    verticalGridlines = this.m_options.getVerticalGridlines();

    rowCount = cellSet.getCount('row');
    columnCount = cellSet.getCount('column');

    rowAppend = rowStart >= this.m_startRow;
    columnAppend = columnStart >= this.m_startCol;

    totalWidth = 0;
    totalHeight = 0;
    top = topStart;

    tempArray = [];

    for (i = 0; i < rowCount; i += 1)
    {
        left = leftStart;
        rowIndex = rowAppend ? rowStart + i : rowStart + (rowCount - 1 - i);
        applyRowBanding = (Math.floor(rowIndex / rowBandingInterval) % 2 === 1) ? true : false;

        for (j = 0; j < columnCount; j += extents['column'])
        {
            width = 0;
            height = 0;

            columnIndex = columnAppend ? columnStart + j : columnStart + (columnCount - 1 - j);

            if (cellSet.getExtent)
            {
                cellExtent = cellSet.getExtent(this.createIndex(rowIndex, columnIndex));
            }
            else
            {
                 cellExtent = {'row': {'extent':1, 'more': {'before':false, 'after':false}}, 
                               'column': {'extent':1, 'more': {'before':false, 'after':false}}};
            }
            extents = {"row": cellExtent['row']['extent'], "column": cellExtent['column']['extent']};

            if (!columnAppend)
            {
                columnIndex = columnIndex - extents['column'] + 1;
            }
            //cannot directly modify rowIndex because outside of this loop
            tempRowIndex =  rowAppend ? rowIndex : rowIndex - extents['row'] + 1;

            indexes = this.createIndex(tempRowIndex, columnIndex);

            if (tempArray[i] && tempArray[i][j])
            {
                // update the left value since that cell exists already
                width = tempArray[i][j].width;
                left = columnAppend ? left + width : left - width;
                continue;
            }

            patched = this._patchExistingCells(cellExtent, columnIndex, tempRowIndex, tempArray);
            if (patched)
            {
                width = patched['width'];
                cell = patched['cell'];
                cellContext = cell[this.getResources().getMappedAttribute('context')];
            }
            else
            {
                cellData = cellSet.getData(indexes);
                cellMetadata = cellSet.getMetadata(indexes);

                cell = document.createElement('div');
                cellContext = this.createCellContext(indexes, cellData, cellMetadata, cell, extents);
                cell.setAttribute(this.getResources().getMappedAttribute('busyContext'),'');
                this._createUniqueId(cell);
                cell[this.getResources().getMappedAttribute('context')] = cellContext;

                //before setting our own styles, else we will overwrite them
                inlineStyle = this.m_options.getInlineStyle("cell", cellContext);
                if (inlineStyle != null)
                {
                    cell['style']['cssText'] = inlineStyle;
                }

                //don't want developer setting height or width through inline styles on cell
                //should be done through header styles, or through the stylesheet
                if (cell['style']['height'] != '')
                {
                    cell['style']['height'] = '';
                }
                if (cell['style']['width'] != '')
                {
                    cell['style']['width'] = '';
                }

                this.m_utils.addCSSClassName(cell, this.getMappedStyle("cell"));
                this.m_utils.addCSSClassName(cell, this.getMappedStyle("formcontrol"));
                //determine if the newly fetched row should be banded
                if (applyRowBanding || (Math.floor(columnIndex / columnBandingInterval) % 2 === 1))
                {
                    this.m_utils.addCSSClassName(cell, this.getMappedStyle("banded"));
                }

                inlineStyleClass = this.m_options.getStyleClass("cell", cellContext);
                if (inlineStyleClass != null)
                {
                    this.m_utils.addCSSClassName(cell, inlineStyleClass);
                }

                // get the row height
                for (k = 0; k < extents['row']; k++)
                {
                    rowKey = k === 0 ? cellContext['keys']['row'] : this._getKey(this._getHeaderByIndex(tempRowIndex + k, 0, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader), 'row');
                    height += this._getCellDimension(cell, tempRowIndex, rowKey, 'row', 'height', true);
                }

                // set the px height on the cell
                this.setElementHeight(cell, height);

                for (k = 0; k < extents['column']; k++)
                {
                    columnKey = k === 0 ? cellContext['keys']['column'] : this._getKey(this._getHeaderByIndex(columnIndex + k, 0, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader), 'column');
                    width += this._getCellDimension(cell, columnIndex, columnKey, 'column', 'width', columnIndex === 0);
                }

                // set the px width on the cell regardless of unit type currently on it
                this.setElementWidth(cell, width)

                //do not put borders on far edge column, edge row, turn off gridlines
                if (verticalGridlines === 'hidden' || (this._isLastColumn(columnIndex + extents['column'] - 1) && ((this.getRowHeaderWidth() + left + width >= this.getWidth()) || this.m_endRowEndHeader != -1)))
                {
                    this.m_utils.addCSSClassName(cell, this.getMappedStyle('borderVerticalNone'));
                }

                if (horizontalGridlines === 'hidden')
                {
                    this.m_utils.addCSSClassName(cell, this.getMappedStyle('borderHorizontalNone'));
                }
                else if (this._isLastRow(tempRowIndex + extents['row'] - 1))
                {
                    if (this.getRowBottom(cell, top + height) >= this.getHeight() || this.m_endColEndHeader != -1)
                    {
                        this.m_utils.addCSSClassName(cell, this.getMappedStyle('borderHorizontalNone'));
                    }
                }

                if (rowAppend)
                {
                    this.setElementDir(cell, top, 'top');
                }
                else
                {
                    this.setElementDir(cell, top - height, 'top');
                }

                if (columnAppend)
                {
                    this.setElementDir(cell, left, dir);
                }
                else
                {
                    this.setElementDir(cell, left - width, dir);
                }


                //add cell to live DOM while rendering, row is now in live DOM so do this first
                this.m_root.appendChild(cell); //@HTMLUpdateOK
                this._renderContent(renderer, cellContext, cell, cellData);

                fragment.appendChild(cell);
            }

            // store the extents that were rendered with this
            this._updateTempArray(tempArray, {'width': width}, i ,j, extents['row'], extents['column']);

            left = columnAppend ? left + width : left - width;

            if (i == 0)
            {
                totalWidth += width;
            }
        }

        height = this._getCellDimension(cell, cellContext['indexes']['row'], cellContext['keys']['row'], 'row', 'height', false);
        top = rowAppend ? top + height : top - height;
        totalHeight += height;
    }

    avgWidth = totalWidth / columnCount;
    return {"avgWidth": avgWidth, "totalRowHeight": totalHeight, "totalColumnWidth": totalWidth};
};

/**
 * Change the appropriate cell dimension as patching occurs
 * @param {number} axisExtent
 * @param {number} otherAxisExtent
 * @param {number} columnIndex
 * @param {number} rowIndex
 * @param {Array} tempArray
 * @param {string} axis
 * @param {string} dimension
 * @param {boolean} isBefore
 * @private
 */
DvtDataGrid.prototype._updateCellDimension = function(axisExtent, otherAxisExtent, columnIndex, rowIndex, tempArray, axis, dimension, isBefore)
{
    var dimensionDelta, otherDimension, cell, cellDimension, cellContext, extent, axisIndex, header,
        headerLevelCount, startHeader, key, returnObj, k, newDimension, dir, addRowIndex, addColumnIndex, j;

    dimensionDelta = 0;
    otherDimension = dimension == 'width' ? 'height' : 'width';

    cell = this._getCellByIndex(this.createIndex(rowIndex, columnIndex));
    cellDimension = this.getElementDir(cell, dimension);
    cellContext = cell[this.getResources().getMappedAttribute('context')];
    cellContext['extents'][axis] = cellContext['extents'][axis] + axisExtent;

    if (axis == 'row')
    {
        axisIndex = rowIndex;
        header = this.m_rowHeader;
        headerLevelCount = this.m_rowHeaderLevelCount;
        startHeader = this.m_startRowHeader;
    }
    else
    {
        axisIndex = columnIndex;
        header = this.m_colHeader;
        headerLevelCount = this.m_columnHeaderLevelCount;
        startHeader = this.m_startColHeader;
    }

    for (k = 1; k <= axisExtent; k++)
    {
        key = this._getKey(this._getHeaderByIndex(isBefore ? axisIndex - k : axisIndex + k, 0, header, headerLevelCount, startHeader), axis);
        dimensionDelta += this._getCellDimension(cell, isBefore ? axisIndex - k : axisIndex + k, key, axis, dimension, true);
        for (j = 0; j < otherAxisExtent; j++)
        {
            if (axis == 'row')
            {
                addRowIndex = isBefore ? rowIndex - k : rowIndex + k;
                addColumnIndex = columnIndex + j;
            }
            else
            {
                addRowIndex = rowIndex + j;
                addColumnIndex = isBefore ? columnIndex - k : columnIndex + k;
            }
            this._addIndexToDatabodyMap(this.createIndex(addRowIndex, addColumnIndex), cell['id']);
        }
    }

    if (isBefore)
    {
        dir = dimension === 'width' ? this.getResources().isRTLMode() ? "right" : "left" : 'top';
        this.setElementDir(cell, this.getElementDir(cell, dir) - dimensionDelta, dir);
        cellContext['indexes'][axis] = cellContext['indexes'][axis] - axisExtent;
    }

    this.setElementDir(cell, cellDimension + dimensionDelta, dimension);

    returnObj = {};
    returnObj[dimension] = dimensionDelta;
    returnObj[otherDimension] = this.getElementDir(cell, otherDimension);
    returnObj['cell'] = cell;
    return returnObj;
};

/**
 * Patch cells as more data is fetched
 * @param {Object} cellExtent
 * @param {number} columnIndex
 * @param {number} rowIndex
 * @param {Array} tempArray
 * @private
 */
DvtDataGrid.prototype._patchExistingCells = function(cellExtent, columnIndex, rowIndex, tempArray)
{
    var extents, patchRowBefore, patchRowAfter, patchColumnBefore, patchColumnAfter, columnExtent, rowExtent;
    columnExtent = cellExtent['column']['extent'];
    rowExtent = cellExtent['row']['extent'];

    extents = {"row": cellExtent['row']['extent'], "column": cellExtent['column']['extent']};
    patchRowBefore = cellExtent['row']['more']['before'];
    patchRowAfter = cellExtent['row']['more']['after'];
    patchColumnBefore = cellExtent['column']['more']['before'];
    patchColumnAfter = cellExtent['column']['more']['after'];

    if (columnIndex - 1 == this.m_endCol && patchColumnBefore)
    {
        return this._updateCellDimension(columnExtent, rowExtent, this.m_endCol, rowIndex, tempArray, 'column', 'width', false);
    }
    else if (rowIndex - 1 == this.m_endRow && patchRowBefore)
    {
        return this._updateCellDimension(rowExtent, columnExtent, columnIndex, this.m_endRow, tempArray, 'row', 'height', false);
    }
    else if (columnIndex + columnExtent == this.m_startCol && patchColumnAfter)
    {
        return this._updateCellDimension(columnExtent, rowExtent, this.m_startCol, rowIndex, tempArray, 'column', 'width', true);
    }
    else if (rowIndex + rowExtent == this.m_startRow && patchRowAfter)
    {
        return this._updateCellDimension(rowExtent, columnExtent, columnIndex, this.m_startRow, tempArray, 'row', 'height', true);
    }

    return false;
};

/**
 * Fill spots in the temporary array with the tempObj
 * @param {Array} tempArray
 * @param {*} tempObj
 * @param {number} i
 * @param {number} j
 * @param {number} iExtent
 * @param {number} jExtent
 * @private
 */
DvtDataGrid.prototype._updateTempArray = function(tempArray, tempObj, i, j, iExtent, jExtent)
{
    // store the extents that were rendered with this
    for (var k = 0; k < iExtent; k++)
    {
        if (tempArray[i + k] == null)
        {
            tempArray[i+k] = [];
        }
        for (var l = 0; l < jExtent; l++)
        {
            tempArray[i + k][j + l] = tempObj;
        }
    }
};

/**
 * Get a cells dimension
 * @param {Element} cell
 * @param {number} index
 * @param {string|null} key
 * @param {string} axis
 * @param {string} dimension
 * @param {boolean} firstPass
 * @private
 */
DvtDataGrid.prototype._getCellDimension = function(cell, index, key, axis, dimension, firstPass)
{
    var headerClassName, endHeader, dimensionOf, shimHeaderContext, inlineStyle, styleClass;
    if (axis == 'row')
    {
        headerClassName = this.getMappedStyle('rowheadercell') + ' ' + this.getMappedStyle('headercell');
        endHeader = this.m_endRowHeader;
    }
    else if (axis == 'column')
    {
        headerClassName = this.getMappedStyle('colheadercell') + ' ' + this.getMappedStyle('headercell');
        endHeader = this.m_endColHeader;
    }

    // use a shim element so that we don't have to manage class name ordering
    // in the case of no headers this gets called everytime, so added firstPass to make sure it's only the first time
    if (endHeader == -1 && firstPass && !this.m_initialized)
    {
        shimHeaderContext = this.createHeaderContext(axis, index, null, {'key': key}, null, 0, 0, 1);
        inlineStyle = this.m_options.getInlineStyle(axis, shimHeaderContext);
        styleClass = this.m_options.getStyleClass(axis, shimHeaderContext);
        dimensionOf = document.createElement('div');
        dimensionOf['style']['cssText'] = inlineStyle;
        dimensionOf['className'] = headerClassName + ' ' + styleClass;
    }
    else
    {
        dimensionOf = cell;
    }

    return this._getHeaderDimension(dimensionOf, key, axis, dimension);
};

/**
 * Push the row and all of its next siblings down.
 * @param {number} rowIndex the starting row to push down.
 * @param {number} adjustment the amount in pixel to push down.
 * @private
 */
DvtDataGrid.prototype.pushRowsDown = function(rowIndex, adjustment)
{
    while (rowIndex <= this.m_endRow)
    {
        var cells = this._getAxisCellsByIndex(rowIndex, 'row');
        if (cells.length > 0)
        {
            for (var i = 0; i < cells.length; i++)
            {
                var cell = cells[i];
                var top = this.getElementDir(cell, 'top') + adjustment;
                cell['style']['top'] = top + 'px';
            }
        }
        rowIndex++;
    }
};

/**
 * Push the row header and all of its next siblings up.
 * @param {number} rowIndex the starting row to push up.
 * @param {number} adjustment the amount in pixel to push up.
 * @private
 */
DvtDataGrid.prototype.pushRowsUp = function(rowIndex, adjustment)
{
    this.pushRowsDown(rowIndex, -adjustment);
};

/**
 * Push the row header and all of its next siblings down.
 * @param {Element} rowHeader the starting rowHeader to push down.
 * @param {number} adjustment the amount in pixel to push down.
 * @private
 */
DvtDataGrid.prototype.pushRowHeadersDown = function(rowHeader, adjustment)
{
    while (rowHeader)
    {
        var top = this.getElementDir(rowHeader, 'top') + adjustment;
        rowHeader['style']['top'] = top + 'px';
        rowHeader = rowHeader['nextSibling'];
    }
};

/**
 * Push the row and all of its next siblings up.
 * @param {Element} rowHeader the starting rowHeader to push up.
 * @param {number} adjustment the amount in pixel to push up.
 * @private
 */
DvtDataGrid.prototype.pushRowHeadersUp = function(rowHeader, adjustment)
{
    this.pushRowHeadersDown(rowHeader, -adjustment);
};

/**
 * Build a cell context object for a cell and return it
 * @param {Object} indexes - the row and column index of the cell
 * @param {Object} data - the data the cell contains
 * @param {Object} metadata - the metadata the cell contains
 * @param {Element} elem - the cell element
 * @return {Object} the cell context object, keys of {indexes,data,keys,datagrid}
 */
DvtDataGrid.prototype.createCellContext = function(indexes, data, metadata, elem, extents)
{
    var cellContext, prop;

    cellContext = {};

    //set the parent to the cell content div
    cellContext['parentElement'] = elem;
    cellContext['indexes'] = indexes;
    cellContext['cell'] = data;
    cellContext['data'] = (data != null && typeof data === 'object' && data.hasOwnProperty('data')) ? data['data']:data;
    cellContext['component'] = this;
    cellContext['datasource'] = this.getDataSource();
    cellContext['mode'] = 'navigation';
    cellContext['extents'] = extents;

    // merge properties from metadata into cell context
    // the properties in metadata would have precedence
    for (prop in metadata)
    {
        if (metadata.hasOwnProperty(prop))
        {
            cellContext[prop] = metadata[prop];
        }
    }

    // invoke callback to allow ojDataGrid to change datagrid reference
    if (this.m_createContextCallback != null)
    {
        this.m_createContextCallback.call(this, cellContext);
    }

    return this.m_fixContextCallback.call(this, cellContext);
};

/**
 * Creates a unique ID
 * @private
 */
DvtDataGrid.prototype._createUniqueId = function(cell)
{
    return this._uniqueIdCallback(cell, false);
};

/**
 * Creates the cell element
 * @param {Object} metadata the metadata for the cell
 * @protected
 */
DvtDataGrid.prototype.createCellElement = function(metadata)
{
    return document.createElement("div");
};

/**
 * Gets the width of the row header
 * @return {number} the width of the row header in pixel.
 * @protected
 */
DvtDataGrid.prototype.getRowHeaderWidth = function()
{
    if (this.m_endRowHeader === -1)
    {
        // check if there's no row header
        return 0;
    }
    return this.m_rowHeaderWidth;
};

/**
 * Gets the height of the column header
 * @return {number} the height of the column header in pixel.
 * @protected
 */
DvtDataGrid.prototype.getColumnHeaderHeight = function()
{
    if (this.m_endColHeader === -1)
    {
        // check if there's no column header
        return 0;
    }
    return this.m_colHeaderHeight;
};

/**
 * Gets the width of the row end header
 * @return {number} the width of the row end header in pixel.
 * @protected
 */
DvtDataGrid.prototype.getRowEndHeaderWidth = function()
{
    if (this.m_endRowEndHeader === -1)
    {
        // check if there's no row header
        return 0;
    }
    return this.m_rowEndHeaderWidth;
};

/**
 * Gets the height of the column end header
 * @return {number} the height of the column end header in pixel.
 * @protected
 */
DvtDataGrid.prototype.getColumnEndHeaderHeight = function()
{
    if (this.m_endColEndHeader === -1)
    {
        // check if there's no column header
        return 0;
    }
    return this.m_colEndHeaderHeight;
};

/**
 * Gets the bottom value relative to the datagrid in pixel.
 * @param {Element} row the row element
 * @param {number|undefined|null} bottom the bottom value in pixel relative to the databody
 * @return {number} the bottom value relative to the datagrid in pixels.
 * @private
 */
DvtDataGrid.prototype.getRowBottom = function(row, bottom)
{
    var colHeaderHeight, top, height;

    // gets the height of the column header, if any
    colHeaderHeight = this.getColumnHeaderHeight();
    // if a bottom value is specified use that
    if (bottom != null)
    {
        return colHeaderHeight + bottom;
    }
    else
    {
        // otherwise try find it from the row element
        top = this.getElementDir(row, 'top');
        height = this.calculateRowHeight(row);
        if (!isNaN(top) && !isNaN(height))
        {
            return colHeaderHeight + top + height;
        }
    }

    return colHeaderHeight;
};

/**
 * Handle an unsuccessful call to the data source fetchCells
 * @param {Error} errorStatus - the error returned from the data source
 * @param {Array.<Object>} cellRange - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 */
DvtDataGrid.prototype.handleCellsFetchError = function(errorStatus, cellRange)
{
    var rowRange, columnRange;

    // remove fetch message
    this.m_fetching['cells'] = false;

    // hide status message
    this.hideStatusText();

    // update datagrid in responds to failed fetch
    if (this.m_databody['firstChild'] == null)
    {
        // if it's initial fetch, then show no data
        if (this._shouldInitialize())
        {
            this._handleInitialization(true);
        }
    }
    else
    {
        // failed while fetching more data.  stop any future fetching
        rowRange = cellRange[0];
        columnRange = cellRange[1];

        if (columnRange['start'] + columnRange['count'] - 1 > this.m_endCol)
        {
            this.m_stopColumnFetch = true;
            // stop header fetch as well
            this.m_stopColumnHeaderFetch = true;
            this.m_stopColumnEndHeaderFetch = true;
        }

        if (rowRange['start'] + rowRange['count'] - 1 > this.m_endRow)
        {
            this.m_stopRowFetch = true;
            // stop header fetch as well
            this.m_stopRowHeaderFetch = true;
            this.m_stopRowEndHeaderFetch = true;
        }
    }
};

/**
 * Display the 'fetching' status message
 */
DvtDataGrid.prototype.showStatusText = function()
{
    var left, top, msg, self = this;

    if (this.m_status['style']['display'] == "block" || this.m_showStatusTimeout)
    {
        return;
    }

    this.m_showStatusTimeout = setTimeout(function()
    {
        msg = self.getResources().getTranslatedText("msgFetchingData");
        self.m_status.setAttribute("aria-label", msg);
        self.m_status['style']['display'] = "block";

        left = self.getWidth() / 2 - self.m_status['offsetWidth'] / 2;
        top = self.getHeight() / 2 - self.m_status['offsetHeight'] / 2;
        self.m_status['style']['left'] = left + 'px';
        self.m_status['style']['top'] = top + 'px';
        self.m_showStatusTimeout = null;
    }, this.getShowStatusDelay());
};

/**
 * Retrieve the delay before showing status
 * @return {number} the delay in ms
 * @private
 */
DvtDataGrid.prototype.getShowStatusDelay = function()
{
    var delay = parseInt(this.getResources().getDefaultOption('showIndicatorDelay'), 10);
    return isNaN(delay) ? 0 : delay;
};

/**
 * Hide the 'fetching' status message
 */
DvtDataGrid.prototype.hideStatusText = function()
{
    if (this.m_showStatusTimeout)
    {
        clearTimeout(this.m_showStatusTimeout);
        this.m_showStatusTimeout = null;
    }
    this.m_status['style']['display'] = "none";
};

/********************* focusable/editable element related methods *****************/
/**
 * Finds all the focusable elements in a node
 * @param {Element|undefined|null} node
 * @param {boolean=} skipTabIndexCheck
 * @return {Array} An array of all of the focusable elements in a node
 */
DvtDataGrid.prototype.getFocusableElementsInNode = function(node, skipTabIndexCheck)
{
    var inputElems, nodes, elem, nodeCount, i, attr;
    inputElems = [];
    attr = this.getResources().getMappedAttribute('tabMod');

    if (document.evaluate)
    {
        // FF and IE are not case sensitive with x-path, but webkit browser are (GoogleChrome and Safari only recognize lower case)
        // to be safe, we check for both lower and upper case
        nodes = document.evaluate(".//*[@tabindex>=0]|.//*[@" + attr + ">=0]|.//input|.//select|.//textarea|.//button|.//a|.//INPUT|.//SELECT|.//TEXTAREA|.//BUTTON|.//A",
                node, null, XPathResult.ANY_TYPE, null);
        elem = nodes.iterateNext();
        while (elem)
        {
            if (!elem.disabled && elem.style.display != 'none' && (skipTabIndexCheck || !elem.tabIndex || elem.tabIndex > 0 || parseInt(elem.getAttribute(attr), 10) >= 0))
            {
                inputElems.push(elem);
            }

            elem = nodes.iterateNext();
        }
    }
    else
    {
        // use the same query as above which has proven to work on non-ie browsers
        // only difference being that tabIndex check can't be greater than or less than so do not -1 for most common case
        nodes = node.querySelectorAll("input, select, button, a, textarea, [tabIndex]:not([tabIndex='-1']), [" + attr + "]:not([" + attr + "='-1'])");
        nodeCount = nodes.length;
        // we don't want to use AdfDhtmlPivotTablePeer._INPUT_REGEXP because it has OPTION in the regexp
        // in IE, each 'option' after 'select' elem will be counted as an input element(and cause duplicate input elems returned)
        // this will cause problem with TAB/Shift-TAB (recognizing whether to go to next cell or to tab within the current cell
        for (i = 0; i < nodeCount; i += 1)
        {
            elem = nodes[i];
            if (!elem.disabled && elem.style.display != 'none' && (skipTabIndexCheck || !elem.tabIndex || elem.tabIndex >= 0 || parseInt(elem.getAttribute(attr), 10) >= 0))
            {
                inputElems.push(elem);
            }
        }
    }
    return inputElems;
};

/**
 * Whether an element is visible
 * @param {Element|undefined|null} elem
 * @param {Element|undefined|null} node
 * @private
 */
DvtDataGrid.prototype.isElementVisible = function(elem, node)
{
    return true;
};

/**
 * Make all focusable elements within the specified cell unfocusable
 * @param {Element} cell
 * @private
 */
DvtDataGrid.prototype._disableAllFocusableElements = function(cell)
{
    var focusElems, i, tabIndex, attr;
    attr = this.getResources().getMappedAttribute('tabMod');

    // make all focusable elements non-focusable, since we want to manage tab stops
    focusElems = this.getFocusableElementsInNode(cell);
    for (i = 0; i < focusElems.length; i++)
    {
        tabIndex = parseInt(focusElems[i].tabIndex, 10);
        if (isNaN(tabIndex) || tabIndex >= 0)
        {
            // store the tabindex as an attribute
            focusElems[i].setAttribute(attr, isNaN(tabIndex) ? -1 : tabIndex);
            focusElems[i].setAttribute('tabIndex', -1);
        }
    }
};

/**
 * Make all focusable elements within the specified cell that were made unfocusable before focusable again
 * @param {Element|undefined|null} cell
 * @private
 */
DvtDataGrid.prototype._enableAllFocusableElements = function(cell)
{
    var focusElems, i, tabIndex, attr;
    attr = this.getResources().getMappedAttribute('tabMod');

    // make all non-focusable elements focusable again
    focusElems = this.getFocusableElementsInNode(cell, false);
    for (i = 0; i < focusElems.length; i++)
    {
        tabIndex = parseInt(focusElems[i].getAttribute(attr), 10);
        focusElems[i].removeAttribute(attr);
        // restore tabIndex as needed
        if (tabIndex == -1)
        {
            focusElems[i].removeAttribute("tabIndex");
        }
        else
        {
            focusElems[i].setAttribute("tabIndex", tabIndex);
        }
    }
};

/**
 * Determine whether the element is a focusable element.
 * @param {Element} elem the element to check
 * @return {boolean} true if element is a focusable element, false otherwise.
 * @private
 */
DvtDataGrid.prototype._isFocusableElement = function(elem)
{
    var tagName = elem.tagName;
    return (tagName === 'INPUT' || tagName === 'TEXTAREA' || tagName === 'SELECT' || tagName === 'BUTTON' || tagName === 'A' || (elem.getAttribute("tabIndex") != null && parseInt(elem.getAttribute("tabIndex"), 10) >= 0 && this.findCell(elem) != elem));
};

DvtDataGrid.prototype._isFocusableElementBeforeCell = function(elem)
{
    // if element is null or if we reach the root of DataGrid or if it is the cell
    if (elem == null || elem == this.getRootElement() || this.m_utils.containsCSSClassName(elem, this.getMappedStyle('cell')))
    {
        return false;
    }

    var tagName = elem.tagName;
    if (tagName === 'INPUT' || tagName === 'TEXTAREA' || tagName === 'SELECT' || tagName === 'BUTTON' || tagName === 'A' || this.m_utils.containsCSSClassName(elem, this.getMappedStyle('active')) || (elem.getAttribute("tabIndex") != null && parseInt(elem.getAttribute("tabIndex"), 10) >= 0 && this.findCell(elem) != elem))
    {
        return true;
    }
    return this._isFocusableElementBeforeCell(elem['parentNode']);
};

/**
 * Sets focus on first focusable element contained by an element
 * @param {Element|undefined|null} elem
 * @return {boolean} true if focus is set successfully, false otherwise
 */
DvtDataGrid.prototype._setFocusToFirstFocusableElement = function(elem)
{
    var elems, firstElement;
    elems = this.getFocusableElementsInNode(elem, true);
    if (elems.length > 0)
    {
        firstElement = elems[0];
        firstElement.focus();
        if (firstElement.setSelectionRange && firstElement.value)
        {
            try
            {
                // ensure focus at the end
                firstElement.setSelectionRange(firstElement.value.length, firstElement.value.length);
            }
            catch (e)
            {
                // invalid state error
            }
        }
        if (this._overwriteFlag === true && typeof elems[0].select === 'function')
        {
            firstElement.select();
        }
        return true;
    }

    return false;
};

/************************************** scrolling/virtualization ************************************/

/**
 * Handle a scroll event calling scrollTo
 * @param {Event} event - the scroll event triggering the method
 */
DvtDataGrid.prototype.handleScroll = function(event)
{
    // Adding to address firefox async scrolling and pixel perfect scroll bar issues:
    // If the datagrid doesn't need scrolling, ff will skip the async scroll event 
    // If the handleScroll is called properly, it'll set the databody attribute as usual
    // Flag added here and resizeGrid setTimeout function so that either will execute, but only once.
    if (!this.m_handleScrollOverflow)
    {
        if (!this.m_hasVerticalScroller && !this.m_hasHorizontalScroller)
        {
            this.m_databody['style']['overflow'] = "hidden";
        }
        this.m_handleScrollOverflow = true;
    }

    // prevent scrolling when animating sort
    if (this.m_animating)
    {
        event.preventDefault();
        return;
    }

    // scroll on touch is handled directly by touch handlers
    if (this.m_utils.isTouchDevice())
    {
        return;
    }

    if (this.m_silentScroll == true)
    {
        this.m_silentScroll = false;
        return;
    }

    var scrollLeft, scrollTop, scroller;
    if (!event)
    {
        event = window['event'];
    }

    if (!event['target'])
    {
        scroller = event['srcElement'];
    }
    else
    {
        scroller = event['target'];
    }

    scrollLeft = this.m_utils.getElementScrollLeft(scroller);
    scrollTop = scroller['scrollTop'];

    this.scrollTo(scrollLeft, scrollTop);
};

/**
 * Retrieve the maximum scrollable width.
 * @return {number} the maximum scrollable width.  Returns MAX_VALUE
 *         if canvas size is unknown.
 * @private
 */
DvtDataGrid.prototype._getMaxScrollWidth = function()
{
    if (this._isCountUnknownOrHighwatermark('column') && !this.m_stopColumnFetch)
    {
        return Number.MAX_VALUE;
    }
    return this.m_scrollWidth;
};

/**
 * Retrieve the maximum scrollable height.
 * @return {number} the maximum scrollable width.  Returns MAX_VALUE
 *         if canvas size is unknown.
 * @private
 */
DvtDataGrid.prototype._getMaxScrollHeight = function()
{
    if (this._isCountUnknownOrHighwatermark('row') && !this.m_stopRowFetch)
    {
        return Number.MAX_VALUE;
    }
    return this.m_scrollHeight;
};

/**
 * Handle a programtic scroll
 * @param {Object} options an object containing the scrollTo information
 * @param {Object} options.position scroll to an x,y location which is relative to the origin of the grid
 * @param {Object} options.position.scrollX the x position of the scrollable region, this should always be positive
 * @param {Object} options.position.scrollY the Y position of the scrollable region, this should always be positive
 *
 */
DvtDataGrid.prototype.scroll = function(options)
{
    var scrollLeft, scrollTop;
    if (options['position'] != null)
    {
        scrollLeft = Math.max(0, Math.min(this._getMaxScrollWidth(), options['position']['scrollX']));
        scrollTop = Math.max(0, Math.min(this._getMaxScrollHeight(), options['position']['scrollY']));
        this._initiateScroll(scrollLeft, scrollTop);
    }
};

/**
 * Used by mouse wheel and touch scrolling to set the scroll position,
 * since the deltas are obtained instead of new scroll position.
 * @param {number} deltaX - the change in X position
 * @param {number} deltaY - the change in Y position
 */
DvtDataGrid.prototype.scrollDelta = function(deltaX, deltaY)
{
    var scrollLeft, scrollTop;
    // Make sure to adjust the scroller size in case the scroller is no longer the same size. 
    this._adjustScrollerSize();

    scrollLeft = Math.max(0, Math.min(this._getMaxScrollWidth(), this.m_currentScrollLeft - deltaX));
    scrollTop = Math.max(0, Math.min(this._getMaxScrollHeight(), this.m_currentScrollTop - deltaY));
    this._initiateScroll(scrollLeft, scrollTop);
};

/**
 * Used by touch scrolling to adjust the scroll position to prevent diagonal scrolling,
 * since the deltas are obtained instead of new scroll position.
 * @param {number} diffX - the change in X position
 * @param {number} diffY - the change in Y position
 * @returns {Array} adjusted diffX, diffY
 */
DvtDataGrid.prototype.adjustTouchScroll = function(diffX, diffY)
{
    // prevent 'diagonal' scrolling
    if (this.m_utils.isTouchDevice())
    {
        if (diffX != 0 && diffY != 0)
        {
            // direction depends on which way moves the most
            if (Math.abs(diffX) > Math.abs(diffY))
            {
                diffY = 0;
                this.m_extraScrollOverY = null;
            }
            else
            {
                diffX = 0;
                this.m_extraScrollOverX = null;
            }
        }        
    }

    return [diffX, diffY];
}

/**
 * Initiate a scroll, this will differentiate between scrolling on touch vs desktop
 * @param {number} scrollLeft
 * @param {number} scrollTop
 */
DvtDataGrid.prototype._initiateScroll = function(scrollLeft, scrollTop)
{
    if (!this.m_utils.isTouchDevice())
    {
        this.m_utils.setElementScrollLeft(this.m_databody, scrollLeft);
        this.m_databody['scrollTop'] = scrollTop;
    }
    else
    {
        // for touch we'll call scrollTo directly instead of relying on scroll event to fire due to performance
        // or if the scroll position of the databody was already set properly from mousewheel etc, then just sync everything up
        // in scrollTo
        this.scrollTo(scrollLeft, scrollTop);
    }
};

/**
 * Disable touch scroll animation by setting durations to 0
 * @private
 */
DvtDataGrid.prototype._disableTouchScrollAnimation = function()
{
    this.m_databody['firstChild'].style.webkitTransitionDuration = '0ms';
    this.m_rowHeader['firstChild'].style.webkitTransitionDuration = '0ms';
    this.m_colHeader['firstChild'].style.webkitTransitionDuration = '0ms';
    this.m_rowEndHeader['firstChild'].style.webkitTransitionDuration = '0ms';
    this.m_colEndHeader['firstChild'].style.webkitTransitionDuration = '0ms';
};

/**
 * Should the datagrid long scroll using appropriate params if no databody but headers.
 * @param {number} scrollLeft - the position the scroller left should be
 * @param {number} scrollTop - the position the scroller top should be
 * @returns {boolean} true if long scroll should init
 */
DvtDataGrid.prototype._shouldLongScroll = function(scrollLeft, scrollTop)
{
    // only long scroll if virtual scrolling
    if (this._isHighWatermarkScrolling())
    {
        return false;
    }

    return ((scrollLeft + this.getViewportWidth()) < this._getMaxLeftPixel() ||
            (scrollTop + this.getViewportHeight()) < this._getMaxTopPixel() ||
            scrollLeft > this._getMaxRightPixel() ||
            scrollTop > this._getMaxBottomPixel());
};


/**
 * Set the scroller position, using translate3d when permitted
 * @param {number} scrollLeft - the position the scroller left should be
 * @param {number} scrollTop - the position the scroller top should be
 */
DvtDataGrid.prototype.scrollTo = function(scrollLeft, scrollTop)
{
    if (scrollLeft != this.m_currentScrollLeft)
    {
        this.m_prevScrollLeft = this.m_currentScrollLeft;
        this.m_currentScrollLeft = scrollLeft;
    }
    if (scrollTop != this.m_currentScrollTop)
    {
        this.m_prevScrollTop = this.m_currentScrollTop;
        this.m_currentScrollTop = scrollTop;
    }

    // check if this is a long scroll
    // don't do this for touch, the check must be done AFTER transition ends otherwise
    // animation will become sluggish, see _syncScroller
    if (!this.m_utils.isTouchDevice())
    {
        if (this._shouldLongScroll(scrollLeft, scrollTop))
        {
            this.handleLongScroll(scrollLeft, scrollTop);
        }
        else
        {
            this.fillViewport();
        }
    }

    // update header and databody scroll position
    this._syncScroller();

    // check if we need to adjust scroller dimension
    this._adjustScrollerSize();

    // check if there's a cell to focus
    if (this.m_cellToFocus != null)
    {
        var cell = this.m_cellToFocus;
        this.m_cellToFocus = null;
        this._setActive(cell, this._createActiveObject(cell), null, false, false, true);
    }

    //if there's an index we wanted to sctoll to after fetch it has now been scrolled to by scrollToIndex, so highlight it
    if (this.m_scrollIndexAfterFetch != null)
    {
        if (this._isInViewport(this.m_scrollIndexAfterFetch) === DvtDataGrid.INSIDE)
        {
            if (this._isDatabodyCellActive() &&
                    this.m_scrollIndexAfterFetch['row'] == this.m_active['indexes']['row'] &&
                    this.m_scrollIndexAfterFetch['column'] == this.m_active['indexes']['column'])
            {
                this._highlightActive();
            }
            //should be able to scroll to index without highlighting it
            this.m_scrollIndexAfterFetch = null;
        }
    }

    // do the same for headers
    if (this.m_scrollHeaderAfterFetch != null)
    {
        if (!this._isDatabodyCellActive() &&
                this.m_scrollHeaderAfterFetch['axis'] == this.m_active['axis'] &&
                this.m_scrollHeaderAfterFetch['index'] == this.m_active['index'] &&
                this.m_scrollHeaderAfterFetch['level'] == this.m_active['level'])
        {
            this._highlightActive();
        }
        //should be able to scroll to index without highlighting it
        this.m_scrollHeaderAfterFetch = null;
    }

    if (!this.m_utils.isTouchDevice())
    {
        // If detect an actual scroll, fire scroll event
        if (this.m_prevScrollTop !== scrollTop || this.m_prevScrollLeft !== scrollLeft)
        {
            this.fireEvent('scroll', {'event': null, 'ui':{'scrollX': scrollLeft, 'scrollY': scrollTop}});
        }
    }
    this._setScrollPosition();
};

/**
 * Callback to run when the final transition ends
 * @private
 */
DvtDataGrid.prototype._scrollTransitionEnd = function()
{
    var databody;

    if (this.m_scrollTransitionEnd != null)
    {
        databody = this.m_databody['firstChild'];

        // remove existing listener
        databody.removeEventListener("webkitTransitionEnd", this.m_scrollTransitionEnd);
    }

    //center touch affordances if row selection multiple
    if (this._isSelectionEnabled())
    {
        this._scrollTouchSelectionAffordance();
    }

    // Fire scroll event after physical scrolling finishes
    this.fireEvent('scroll', {'event': null, 'ui':{'scrollX': this.m_currentScrollLeft, 'scrollY': this.m_currentScrollTop}});

    // check how the viewport needs to be filled, through long scroll or HWS fillViewport.
    // This should be replaced once we optimize sort going to the newly sorted location.
    if (this._shouldLongScroll(this.m_currentScrollLeft, this.m_currentScrollTop))
    {
        this.handleLongScroll(this.m_currentScrollLeft, this.m_currentScrollTop);
    }
    else
    {
        this.fillViewport();
    }
};

/**
 * Perform the bounce back animation when a swipe gesture causes over scrolling
 * @private
 */
DvtDataGrid.prototype._bounceBack = function()
{
    var scrollLeft, scrollTop, databody, colHeader, rowHeader, colEndHeader, rowEndHeader;

    scrollLeft = this.m_currentScrollLeft;
    scrollTop = this.m_currentScrollTop;

    databody = this.m_databody['firstChild'];
    colHeader = this.m_colHeader['firstChild'];
    rowHeader = this.m_rowHeader['firstChild'];
    colEndHeader = this.m_colEndHeader['firstChild'];
    rowEndHeader = this.m_rowEndHeader['firstChild'];

    // remove existing listener
    databody.removeEventListener("webkitTransitionEnd", this.m_bounceBack);

    databody.style.webkitTransitionDuration = DvtDataGrid.BOUNCE_ANIMATION_DURATION + 'ms';
    rowHeader.style.webkitTransitionDuration = DvtDataGrid.BOUNCE_ANIMATION_DURATION + 'ms';
    rowEndHeader.style.webkitTransitionDuration = DvtDataGrid.BOUNCE_ANIMATION_DURATION + 'ms';
    colHeader.style.webkitTransitionDuration = DvtDataGrid.BOUNCE_ANIMATION_DURATION + 'ms';
    colEndHeader.style.webkitTransitionDuration = DvtDataGrid.BOUNCE_ANIMATION_DURATION + 'ms';

    // process to run after bounce back animation ends
    if (this.m_scrollTransitionEnd == null)
    {
        this.m_scrollTransitionEnd = this._scrollTransitionEnd.bind(this);
    }
    databody.addEventListener("webkitTransitionEnd", this.m_scrollTransitionEnd);

    // scroll back to actual scrollLeft/scrollTop positions
    if (this.getResources().isRTLMode())
    {
        databody.style.webkitTransform = "translate3d(" + scrollLeft + "px, " + (-scrollTop) + "px, 0)";
        colHeader.style.webkitTransform = "translate3d(" + scrollLeft + "px, 0, 0)";
        colEndHeader.style.webkitTransform = "translate3d(" + scrollLeft + "px, 0, 0)";
    }
    else
    {
        databody.style.webkitTransform = "translate3d(" + (-scrollLeft) + "px, " + (-scrollTop) + "px, 0)";
        colHeader.style.webkitTransform = "translate3d(" + (-scrollLeft) + "px, 0, 0)";
        colEndHeader.style.webkitTransform = "translate3d(" + (-scrollLeft) + "px, 0, 0)";
    }
    rowHeader.style.webkitTransform = "translate3d(0, " + (-scrollTop) + "px, 0)";
    rowEndHeader.style.webkitTransform = "translate3d(0, " + (-scrollTop) + "px, 0)";

    // reset
    this.m_extraScrollOverX = null;
    this.m_extraScrollOverY = null;
};

/**
 * Make sure the databody/headers and the scroller are in sync, which could happen when scrolling
 * stopped awaiting fetch to complete.
 * @private
 */
DvtDataGrid.prototype._syncScroller = function()
{
    var scrollLeft, scrollTop, databody, colHeader, rowHeader, dir, colEndHeader, rowEndHeader;

    scrollLeft = this.m_currentScrollLeft;
    scrollTop = this.m_currentScrollTop;

    databody = this.m_databody['firstChild'];
    colHeader = this.m_colHeader['firstChild'];
    rowHeader = this.m_rowHeader['firstChild'];
    colEndHeader = this.m_colEndHeader['firstChild'];
    rowEndHeader = this.m_rowEndHeader['firstChild'];

    // use translate3d for smoother scrolling
    // this checks determine whether this is webkit and translated3d is supported
    if (this.m_utils.isTouchDevice() && window.hasOwnProperty('WebKitCSSMatrix'))
    {
        // check if the swipe gesture causes over scrolling of scrollable area
        if (this.m_extraScrollOverX != null || this.m_extraScrollOverY != null)
        {
            // swipe horizontal or vertical
            if (this.m_extraScrollOverX != null)
            {
                scrollLeft = scrollLeft + this.m_extraScrollOverX;
            }
            else
            {
                scrollTop = scrollTop + this.m_extraScrollOverY;
            }

            // bounce back animation function
            if (this.m_bounceBack == null)
            {
                this.m_bounceBack = this._bounceBack.bind(this);
            }

            databody.addEventListener("webkitTransitionEnd", this.m_bounceBack);
        }
        else
        {
            if (databody.style.webkitTransitionDuration == '0ms')
            {
                // no transition, just call the handler directly
                this._scrollTransitionEnd();
            }
            else
            {
                if (this.m_scrollTransitionEnd == null)
                {
                    this.m_scrollTransitionEnd = this._scrollTransitionEnd.bind(this);
                }
                databody.addEventListener("webkitTransitionEnd", this.m_scrollTransitionEnd);
            }
        }

        // actual scrolling of databody and headers
        if (this.getResources().isRTLMode())
        {
            databody.style.webkitTransform = "translate3d(" + scrollLeft + "px, " + (-scrollTop) + "px, 0)";
            colHeader.style.webkitTransform = "translate3d(" + scrollLeft + "px, 0, 0)";
            colEndHeader.style.webkitTransform = "translate3d(" + scrollLeft + "px, 0, 0)";
        }
        else
        {
            databody.style.webkitTransform = "translate3d(" + (-scrollLeft) + "px, " + (-scrollTop) + "px, 0)";
            colHeader.style.webkitTransform = "translate3d(" + (-scrollLeft) + "px, 0, 0)";
            colEndHeader.style.webkitTransform = "translate3d(" + (-scrollLeft) + "px, 0, 0)";
        }
        rowHeader.style.webkitTransform = "translate3d(0, " + (-scrollTop) + "px, 0)";
        rowEndHeader.style.webkitTransform = "translate3d(0, " + (-scrollTop) + "px, 0)";
    }
    else
    {
        dir = this.getResources().isRTLMode() ? "right" : "left";
        this.setElementDir(colHeader, -scrollLeft, dir);
        this.setElementDir(colEndHeader, -scrollLeft, dir);
        this.setElementDir(rowHeader, -scrollTop, 'top');
        this.setElementDir(rowEndHeader, -scrollTop, 'top');
    }
};

/**
 * Adjust the scroller when we scroll to the ends of the scroller.  The scroller dimension might
 * need adjustment due to 1) variable column width or row height due to custom sizing 2) the row
 * or column count is not exact.
 * @private
 */
DvtDataGrid.prototype._adjustScrollerSize = function()
{
    var scrollerContent, scrollerContentHeight, scrollerContentWidth;
    scrollerContent = this.m_databody['firstChild'];
    scrollerContentHeight = this.getElementHeight(scrollerContent);
    scrollerContentWidth = this.getElementWidth(scrollerContent);

    // if (1) actual content is higher than scroller (regardless of the current position) OR
    //    (2) we have reached the last row and the actual content is shorter than scroller
    if ((this._getMaxBottomPixel() > scrollerContentHeight) ||
            (this.getDataSource().getCount('row') == (this._getMaxBottom() + 1) && !this._isCountUnknown('row') && this._getMaxBottom() > -1))
    {
        this.setElementHeight(scrollerContent, this._getMaxBottomPixel());
    }

    // if (1) actual content is wider than scroller (regardless of the current position) OR
    //    (2) we have reached the last column and the actual content is narrower than scroller
    if ((this._getMaxRightPixel() > scrollerContentWidth) ||
            (this.getDataSource().getCount('column') == (this._getMaxRight() + 1) && !this._isCountUnknown('column') && this._getMaxRight() > -1))
    {
        this.setElementWidth(scrollerContent, this._getMaxRightPixel());
    }
};

/**
 * Get the starting position based on scroll
 * @param {number} scrollDir
 * @param {number} prevScrollDir
 * @param {string} axis
 * @returns {Object} contains start and startPixel
 */
DvtDataGrid.prototype._getLongScrollStart = function(scrollDir, prevScrollDir, axis)
{
    var start, startPixel, oversizeRatio, fetchSize, total, scrollerDimension, maxDimension, maxScroll, avgDimension, scrollbarSize;

    // totals must be 0 or higher for long scroll
    if (prevScrollDir != scrollDir)
    {
        if (axis == 'row')
        {
            scrollerDimension = this.getElementHeight(this.m_databody.firstChild);
            maxDimension = this.m_utils._getMaxDivHeightForScrolling();
            maxScroll = this._getMaxScrollHeight();
            avgDimension = this.m_avgRowHeight;
            scrollbarSize = this.m_hasHorizontalScroller ?  this.m_utils.getScrollbarSize() : 0;
        }
        else if (axis == 'column')
        {
            scrollerDimension = this.getElementWidth(this.m_databody.firstChild);
            maxDimension = this.m_utils._getMaxDivWidthForScrolling();
            maxScroll = this._getMaxScrollWidth();
            avgDimension = this.m_avgColWidth;
            scrollbarSize = this.m_hasVerticalScroller ?  this.m_utils.getScrollbarSize() : 0;
        }

        oversizeRatio =  Math.max(Math.min(scrollDir / scrollerDimension, 1), 0);
        total = this.getDataSource().getCount(axis);
        fetchSize = this.getFetchSize(axis);
        start = Math.floor(total * oversizeRatio);
        startPixel = maxDimension <= scrollerDimension ? Math.min(scrollDir, maxScroll) : start * avgDimension;

        if (oversizeRatio == 1 || (scrollDir + (fetchSize * avgDimension)) > (scrollerDimension - scrollbarSize))
        {
            start = Math.max(total - fetchSize, 0);
            startPixel = Math.max(scrollerDimension - (fetchSize * avgDimension), 0);
        }
    }
    else
    {
        if (axis == 'row')
        {
            start = this.m_startRow;
            startPixel = this.m_startRowPixel;
        }
        else if (axis == 'column')
        {
            start = this.m_startCol;
            startPixel = this.m_startColPixel;
        }
    }

    return {'start': start, 'startPixel': startPixel};
};

/**
 * Handle scroll to position that is completely outside of the current row/column range
 * For example, in Chrome it is possible to cause a "jump" back to the start position
 * This might also be needed if we decide to use delay scroll (to detect long scroll) to avoid
 * excessive fetching.
 * @param {number} scrollLeft - the position the scroller left should be
 * @param {number} scrollTop - the position the scroller top should be
 */
DvtDataGrid.prototype.handleLongScroll = function(scrollLeft, scrollTop)
{
    var startRow, startCol, startRowPixel, startColPixel, rowReturnVal, columnReturnVal;

    this.m_isLongScroll = true;

    if (this.isFetchComplete())
    {
        rowReturnVal = this._getLongScrollStart(scrollTop, this.m_prevScrollTop, 'row');
        startRow = rowReturnVal['start'];
        startRowPixel = rowReturnVal['startPixel'];

        columnReturnVal = this._getLongScrollStart(scrollLeft, this.m_prevScrollLeft, 'column');
        startCol = columnReturnVal['start'];
        startColPixel = columnReturnVal['startPixel'];

        // reset ranges, just cleaned up to only set if the header is present
        if (this.m_hasCells)
        {
            this.m_startRow = startRow;
            this.m_endRow = -1;
            this.m_startRowPixel = startRowPixel;
            this.m_endRowPixel = startRowPixel;
            this.m_startCol = startCol;
            this.m_endCol = -1;
            this.m_startColPixel = startColPixel;
            this.m_endColPixel = startColPixel;
        }

        if (this.m_hasRowHeader)
        {
            this.m_startRowHeader = startRow;
            this.m_endRowHeader = -1;
            this.m_startRowHeaderPixel = startRowPixel;
            this.m_endRowHeaderPixel = startRowPixel;
        }
        if (this.m_hasRowEndHeader)
        {
            this.m_startRowEndHeader = startRow;
            this.m_endRowEndHeader = -1;
            this.m_startRowEndHeaderPixel = startRowPixel;
            this.m_endRowEndHeaderPixel = startRowPixel;
        }
        if (this.m_hasColHeader)
        {
            this.m_startColHeader = startCol;
            this.m_endColHeader = -1;
            this.m_startColHeaderPixel = startColPixel;
            this.m_endColHeaderPixel = startColPixel;
        }
        if (this.m_hasColEndHeader)
        {
            this.m_startColEndHeader = startCol;
            this.m_endColEndHeader = -1;
            this.m_startColEndHeaderPixel = startColPixel;
            this.m_endColEndHeaderPixel = startColPixel;
        }

        this.m_stopRowFetch = false;
        this.m_stopRowHeaderFetch = false;
        this.m_stopRowEndHeaderFetch = false;
        this.m_stopColumnFetch = false;
        this.m_stopColumnHeaderFetch = false;
        this.m_stopColumnEndHeaderFetch = false;

        // custom success callback so that we can reset all ranges and fields
        // initiate fetch of headers and cells
        this.fetchHeaders("row", startRow, this.m_rowHeader, this.m_rowEndHeader, undefined, {'success': function(headerSet, headerRange, endHeaderSet)
            {
                this.handleRowHeadersFetchSuccessForLongScroll(headerSet, headerRange, endHeaderSet);
            }});
        this.fetchHeaders("column", startCol, this.m_colHeader, this.m_colEndHeader, undefined, {'success': function(headerSet, headerRange, endHeaderSet)
            {
                this.handleColumnHeadersFetchSuccessForLongScroll(headerSet, headerRange, endHeaderSet);
            }});
        this.fetchCells(this.m_databody, startRow, startCol, null, null, {'success': function(cellSet, cellRange)
            {
                this.handleCellsFetchSuccessForLongScroll(cellSet, cellRange, startRow, startCol, startRowPixel, startColPixel);
            }});
    }
};

/**
 * Handle a successful call to the data source fetchHeaders for long scroll
 * @param {Object} headerSet - the result of the fetch
 * @param {Object} headerRange - {"axis":,"start":,"count":,"header":}
 * @param {Object} endHeaderSet - the result of the fetch
 * @protected
 */
DvtDataGrid.prototype.handleRowHeadersFetchSuccessForLongScroll = function(headerSet, headerRange, endHeaderSet)
{
    var headerContent, endHeaderContent;
    headerContent = this.m_rowHeader['firstChild'];
    endHeaderContent = this.m_rowEndHeader['firstChild'];
    if (headerContent != null)
    {
        this.m_utils.empty(headerContent);
    }
    if (endHeaderContent != null)
    {
        this.m_utils.empty(endHeaderContent);
    }
    this.handleHeadersFetchSuccess(headerSet, headerRange, endHeaderSet, false);
};

/**
 * Handle a successful call to the data source fetchHeaders for long scroll
 * @param {Object} headerSet - the result of the fetch
 * @param {Object} headerRange - {"axis":,"start":,"count":,"header":}
 * @param {Object} endHeaderSet - the result of the fetch
 * @protected
 */
DvtDataGrid.prototype.handleColumnHeadersFetchSuccessForLongScroll = function(headerSet, headerRange, endHeaderSet)
{
    var headerContent, endHeaderContent;
    headerContent = this.m_colHeader['firstChild'];
    endHeaderContent = this.m_colEndHeader['firstChild'];
    if (headerContent != null)
    {
        this.m_utils.empty(headerContent);
    }
    if (endHeaderContent != null)
    {
        this.m_utils.empty(endHeaderContent);
    }
    this.handleHeadersFetchSuccess(headerSet, headerRange, endHeaderSet, false);
};
/**
 * Handle a successful call to the data source fetchCells. Create new row and
 * cell DOM elements when necessary and then insert them into the databody.
 * @param {Object} cellSet - a CellSet object which encapsulates the result set of cells
 * @param {Array.<Object>} cellRange - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 * @param {number} startRow the row to start insert at
 * @param {number} startCol the col to start insert at
 * @param {number} startRowPixel the row pixel to start insert at
 * @param {number} startColPixel the col pixel to start insert at
 * @protected
 */
DvtDataGrid.prototype.handleCellsFetchSuccessForLongScroll = function(cellSet, cellRange, startRow, startCol, startRowPixel, startColPixel)
{
    var databodyContent = this.m_databody['firstChild'];
    if (databodyContent != null)
    {
        this._emptyDatabody(databodyContent);
    }

    // now calls fetch success proc
    this.handleCellsFetchSuccess(cellSet, cellRange);
};

/**
 * Method to clean up the viewport in one direction, left cleans the first columns, top the first rows etc.
 * This is seperate from fill viewport so that in both the synchronus and asynchronus
 * fetch case the cleanuo happens after we get the data fpor the next area.
 * @param {string|null|undefined} direction left/right/top/bottom
 */
DvtDataGrid.prototype._cleanupViewport = function(direction)
{
    var viewportLeft, viewportRight, viewportTop, viewportBottom;

    if (this._isHighWatermarkScrolling() || !this._isScrollBackToEditable())
    {
        return;
    }

    // direction can be null if there are no cells fetched
    if (direction == null)
    {
        if (this.m_prevScrollLeft > this.m_currentScrollLeft)
        {
            direction = 'right';
        }
        else if (this.m_prevScrollLeft < this.m_currentScrollLeft)
        {
            direction = 'left';
        }
        else if (this.m_prevScrollTop > this.m_currentScrollTop)
        {
            direction = 'bottom';
        }
        else if (this.m_prevScrollTop < this.m_currentScrollTop)
        {
            direction = 'top';
        }
    }

    //the viewport is the scroller, width and height
    viewportLeft = this._getViewportLeft();
    viewportRight = this._getViewportRight();
    viewportTop = this._getViewportTop();
    viewportBottom = this._getViewportBottom();

    if (direction == 'top' && viewportTop > this._getMaxTopPixel())
    {
        this.removeRowsFromTop(this.m_databody);
        this.removeRowHeadersFromTop();
    }
    else if (direction == 'bottom' && viewportBottom < this._getMaxBottomPixel())
    {
        this.removeRowsFromBottom(this.m_databody);
        this.removeRowHeadersFromBottom();
    }
    else if (direction == 'left' && viewportLeft > this._getMaxLeftPixel())
    {
        this.removeColumnsFromLeft(this.m_databody);
        this.removeColumnHeadersFromLeft();
    }
    else if (direction == 'right' && viewportRight < this._getMaxRightPixel())
    {
        this.removeColumnsFromRight(this.m_databody);
        this.removeColumnHeadersFromRight();
    }
};

/**
 * Make sure the viewport is filled of cells, this method has been modified to just fill
 * and so that it will always follow a fetchHeaders call with a fetchCells call to keep them in sync.
 */
DvtDataGrid.prototype.fillViewport = function()
{
    var viewportLeft, viewportRight, viewportTop, viewportBottom, fetchStart, fetchSize;

    if (this.isFetchComplete())
    {
        //the viewport is the scroller, width and height
        // fetch slightly before the edge for the zoomed browser case as the pixel mapping isn't perfect
        viewportLeft = this._getViewportLeft();
        viewportRight = this._getViewportRight() + DvtDataGrid.FETCH_PIXEL_THRESHOLD;
        viewportTop = this._getViewportTop();
        viewportBottom = this._getViewportBottom() + DvtDataGrid.FETCH_PIXEL_THRESHOLD;

        if (this._getMaxBottomPixel() <= viewportBottom)
        {
            if (!this.m_stopRowHeaderFetch || !this.m_stopRowEndHeaderFetch || !this.m_stopRowFetch)
            {
                fetchStart = Math.max(0, this._getMaxBottom() + 1);
                fetchSize = Math.max(0, this.getFetchCount("row", fetchStart));
                this.fetchHeaders("row", fetchStart, this.m_rowHeader, this.m_rowEndHeader, fetchSize);
                this.fetchCells(this.m_databody, fetchStart, this.m_startCol, fetchSize, this.m_endCol - this.m_startCol + 1);
                return;
            }
        }

        if (this._getMaxTopPixel() > viewportTop)
        {
            fetchStart = Math.max(0, this._getMaxTop() - this.getFetchSize("row"));
            fetchSize = Math.max(0, this._getMaxTop() - fetchStart);
            this.fetchHeaders("row", fetchStart, this.m_rowHeader, this.m_rowEndHeader, fetchSize);
            this.fetchCells(this.m_databody, fetchStart, this.m_startCol, fetchSize, this.m_endCol - this.m_startCol + 1);
            return;
        }

        if (this._getMaxRightPixel() <= viewportRight)
        {
            if (!this.m_stopColumnHeaderFetch || !this.m_stopColumnEndHeaderFetch || !this.m_stopColumnFetch)
            {
                fetchStart = Math.max(0, this._getMaxRight() + 1);
                fetchSize = Math.max(0, this.getFetchCount("column", fetchStart));
                this.fetchHeaders("column", fetchStart, this.m_colHeader, this.m_colEndHeader, fetchSize);
                this.fetchCells(this.m_databody, this.m_startRow, fetchStart, this.m_endRow - this.m_startRow + 1, fetchSize);
                return;
            }
        }

        if (this._getMaxLeftPixel() > viewportLeft)
        {
            fetchStart = Math.max(0, this._getMaxLeft() - this.getFetchSize("column"));
            fetchSize = Math.max(0, this._getMaxLeft() - fetchStart);
            this.fetchHeaders("column", fetchStart, this.m_colHeader, this.m_colEndHeader, fetchSize);
            this.fetchCells(this.m_databody, this.m_startRow, fetchStart, this.m_endRow - this.m_startRow + 1, fetchSize);
            return;
        }
    }
};

/**
 * @returns {number} last column or column start or end header
 */
DvtDataGrid.prototype._getMaxRight = function()
{
    return Math.max(Math.max(this.m_endCol, this.m_endColHeader), this.m_endColEndHeader);
};

/**
 * @returns {number} first column or column start or end header
 */
DvtDataGrid.prototype._getMaxLeft = function()
{
    return Math.max(Math.max(this.m_startCol, this.m_startColHeader), this.m_startColEndHeader);
};

/**
 * @returns {number} last column or column start or end header pixel
 */
DvtDataGrid.prototype._getMaxRightPixel = function()
{
    return Math.max(Math.max(this.m_endColPixel, this.m_endColHeaderPixel), this.m_endColEndHeaderPixel);
};

/**
 * @returns {number} first column or column start or end header pixel
 */
DvtDataGrid.prototype._getMaxLeftPixel = function()
{
    return Math.max(Math.max(this.m_startColPixel, this.m_startColHeaderPixel), this.m_startColEndHeaderPixel);
};

/**
 * @returns {number} last row or row start or end header
 */
DvtDataGrid.prototype._getMaxBottom = function()
{
    return Math.max(Math.max(this.m_endRow, this.m_endRowHeader), this.m_endRowEndHeader);
};

/**
 * @returns {number} first row or row start or end header
 */
DvtDataGrid.prototype._getMaxTop = function()
{
    return Math.max(Math.max(this.m_startRow, this.m_startRowHeader), this.m_startRowEndHeader);
};

/**
 * @returns {number} last row or row start or end header pixel
 */
DvtDataGrid.prototype._getMaxBottomPixel = function()
{
    return Math.max(Math.max(this.m_endRowPixel, this.m_endRowHeaderPixel), this.m_endRowEndHeaderPixel);
};

/**
 * @returns {number} first row or row start or end header pixel
 */
DvtDataGrid.prototype._getMaxTopPixel = function()
{
    return Math.max(Math.max(this.m_startRowPixel, this.m_startRowHeaderPixel), this.m_startRowEndHeaderPixel);
};

/**
 * If we are about to remove a cell that is being edited, try to handle it first
 * @private
 */
DvtDataGrid.prototype._isScrollBackToEditable = function()
{
    var currentMode = this._getCurrentMode();
    var cell = this._getActiveElement();
    if (currentMode == 'edit' && this._isCellGoingToBeRemoved(cell))
    {
        return this._handleExitEdit(null, cell);
    }
    return true;
};

/**
 * Check if the cell is supposed to be removed
 * @private
 * @param {Element|null} cell
 */
DvtDataGrid.prototype._isCellGoingToBeRemoved = function(cell)
{
    var height, width, left, top;
    if (!this._isHighWatermarkScrolling())
    {
        if((this.m_endRow - this.m_startRow) > this.MAX_ROW_THRESHOLD)
        {
            top = this.getElementDir(cell['parentNode'], 'top');
            height = this.getElementHeight(cell);
            if (top + height < this.m_currentScrollTop || top < this.m_currentScrollTop + this.getViewportHeight())
            {
                return true;
            }
        }
        if ((this.m_endCol - this.m_startCol) > this.MAX_COLUMN_THRESHOLD)
        {
            left = this.getElementDir(cell, 'left');
            width = this.getElementWidth(cell);
            if (left + width < this.m_currentScrollLeft || left < this.m_currentScrollLeft + this.getViewportHeight())
            {
                return true;
            }
        }
    }

};

/**
 * Fetch cells and append results to the bottom
 * @protected
 */
DvtDataGrid.prototype.fetchCellsToBottom = function()
{
    this.fetchCells(this.m_databody, this.m_endRow + 1, this.m_startCol, null, this.m_endCol - this.m_startCol + 1);
};

/**
 * Fetch cells and insert results to the top
 * @protected
 */
DvtDataGrid.prototype.fetchCellsToTop = function()
{
    var fetchSize, fetchStartRow;
    // add rows to top
    fetchStartRow = Math.max(0, this.m_startRow - this.getFetchSize("row"));
    fetchSize = Math.max(0, this.m_startRow - fetchStartRow);
    this.fetchCells(this.m_databody, fetchStartRow, this.m_startCol, fetchSize, this.m_endCol - this.m_startCol + 1);
};

/**
 * Remove cells along a given axis
 * @param {string} axis
 * @param {number} threshold
 * @param {boolean} isFromEnd
 * @private
 */
DvtDataGrid.prototype._removeCellsAlongAxis = function(axis, threshold, isFromEnd)
{
    var i, j, totalDimensionChange, totalCountChange, axisStart, axisEnd, axisHeaders, 
    axisLevelCount, axisStartHeader, dimension, axisStartPixel, axisEndPixel, currentScroll,
    otherAxis, otherAxisStart, otherAxisEnd, dir, key, dimensionValue, cell, cellContext,
    axisExtent, otherExtent, k, index;
    totalDimensionChange = 0;
    totalCountChange = 0;
    if (axis == 'row')
    {
        axisStart = this.m_startRow;
        axisEnd = this.m_endRow;
        axisHeaders = this.m_rowHeader;
        axisLevelCount = this.m_rowHeaderLevelCount;
        axisStartHeader = this.m_startRowHeader;
        dimension = 'height';
        axisStartPixel = this.m_startRowPixel;
        axisEndPixel = this.m_endRowPixel;
        currentScroll = this.m_currentScrollTop;
        otherAxis = 'column';
        otherAxisStart = this.m_startCol;
        otherAxisEnd = this.m_endCol;
        dir = 'top';
        j = isFromEnd ? axisEnd : axisStart;
    }
    else
    {
        axisStart = this.m_startCol;
        axisEnd = this.m_endCol;
        axisHeaders = this.m_colHeader;
        axisLevelCount = this.m_columnHeaderLevelCount;
        axisStartHeader = this.m_startColHeader;
        dimension = 'width';
        axisStartPixel = this.m_startColPixel;
        axisEndPixel = this.m_endColPixel;
        currentScroll = this.m_currentScrollLeft;
        otherAxis = 'row';
        otherAxisStart = this.m_startRow;
        otherAxisEnd = this.m_endRow;
        dir = this.getResources().isRTLMode() ? 'right' : 'left';
        j = isFromEnd ? axisEnd : axisStart;
    }


    while (j <= axisEnd && j >= axisStart)
    {
        key = this._getKey(this._getHeaderByIndex(j, 0, axisHeaders, axisLevelCount, axisStartHeader), axis);
        if (key == null)
        {
            key = this._getKey(this._getCellByIndex(axis == 'column' ? this.createIndex(this.m_startRow, j) :this.createIndex(j, this.m_startCol)), axis);
        }
        dimensionValue = this._getCellDimension(null, j, key, axis, dimension, false);
        if (isFromEnd ? (axisEndPixel - dimensionValue - totalDimensionChange > threshold) : (axisStartPixel + dimensionValue + totalDimensionChange < currentScroll - threshold))
        {
            for (i = otherAxisStart; i <= otherAxisEnd; i += otherExtent)
            {
                cell = this._getCellByIndex(axis == 'column' ? this.createIndex(i, j) :this.createIndex(j, i));
                cellContext = cell[this.getResources().getMappedAttribute('context')];
                axisExtent = cellContext['extents'][axis];
                otherExtent = cellContext['extents'][otherAxis];

                if (axisExtent == 1)
                {
                    this._remove(cell);
                }
                else
                {
                    cellContext['extents'][axis] -= 1;
                    this.setElementDir(cell, this.getElementDir(cell, dimension) - dimensionValue, dimension);
                    if (!isFromEnd)
                    {
                        cellContext['indexes'][axis] += 1;
                        this.setElementDir(cell, this.getElementDir(cell, dir) + dimensionValue, dir);
                    }
                }

                for (k = 0; k < otherExtent; k++)
                {
                    index = axis == 'column' ? this.createIndex(i + k, j) : this.createIndex(j, i + k);
                    this._removeIndexFromDatabodyMap(index);
                }
            }
            totalDimensionChange += dimensionValue;
            totalCountChange += 1;
            j = isFromEnd ? j - 1 : j + 1;
        }
        else
        {
            break;
        }
    }

    return {dimensionChange: totalDimensionChange, extentChange: totalCountChange};
};

/**
 * Removes all of the headers in the containing div up until the right value is less than the scroll position minus the threshold.
 * It is recuresively called on inner levels in the multi-level header case.
 * @param {Element} headersContainer
 * @param {Element|null} firstChild
 * @param {number} startPixel
 * @param {number} threshold
 * @param {string} className
 * @param {string} dimension
 * @param {string} dir
 * @param {number} scrollPosition
 * @returns {Object} object with keys extentChange, which denotes how many header
 *      indexes were removed under the parent and dimensionChange which is the
 *      total dimensions of the headers removed
 */
DvtDataGrid.prototype.removeHeadersFromStartOfContainer = function(headersContainer, firstChild, startPixel, threshold, className, dimension, dir, scrollPosition)
{
    var headers, element, header, isHeader, dimensionValue, removedHeaders = 0, removedDimensionValue = 0, returnVal;

    headers = headersContainer['childNodes'];
    element = firstChild == null ? headersContainer['firstChild'] : firstChild['nextSibling'];
    if (element == null)
    {
        return {extentChange: 0, dimensionChange: 0};
    }
    isHeader = this.m_utils.containsCSSClassName(element, className);
    header = isHeader ? element : element['firstChild'];
    dimensionValue = this.getElementDir(header, dimension);

    while (startPixel + dimensionValue < scrollPosition - threshold)
    {
        this._remove(element);
        removedDimensionValue += dimensionValue;
        removedHeaders += isHeader ? 1 : this._getAttribute(element, 'extent', true);
        startPixel += dimensionValue;

        element = firstChild == null ? headersContainer['firstChild'] : firstChild['nextSibling'];
        if (element == null)
        {
            return {extentChange: removedHeaders, dimensionChange: removedDimensionValue};
        }
        isHeader = this.m_utils.containsCSSClassName(element, className);
        header = isHeader ? element : element['firstChild'];
        dimensionValue = this.getElementDir(header, dimension);
    }

    if (!isHeader)
    {
        returnVal = this.removeHeadersFromStartOfContainer(element, element['firstChild'], startPixel, threshold, className, dimension, dir, scrollPosition);
        this._setAttribute(element, 'start', this._getAttribute(element, 'start', true) + returnVal.extentChange);
        this._setAttribute(element, 'extent', this._getAttribute(element, 'extent', true) - returnVal.extentChange);
        this.setElementDir(header, this.getElementDir(header, dir) + returnVal.dimensionChange, dir);
        this.setElementDir(header, this.getElementDir(header, dimension) - returnVal.dimensionChange, dimension);

        removedHeaders += returnVal.extentChange;
        removedDimensionValue += returnVal.dimensionChange;
    }

    return {extentChange: removedHeaders, dimensionChange: removedDimensionValue};
};

/**
 * Removes all of the headers in the containing div up until the right value is less than the specified threshold.
 * It is recuresively called on inner levels in the multi-level header case.
 * @param {Element} headersContainer
 * @param {number} endPixel
 * @param {number} threshold
 * @param {string} className
 * @param {string} dimension
 * @returns {Object} object with keys extentChange, which denotes how many header
 *      indexes were removed under the parent and dimensionChange which is the
 *      total width of the headers removed
 */
DvtDataGrid.prototype.removeHeadersFromEndOfContainer = function(headersContainer, endPixel, threshold, className, dimension)
{
    var element, header, isHeader, dimensionValue, removedHeaders = 0, removedHeadersDimension = 0, returnVal;

    element = headersContainer['lastChild'];
    isHeader = this.m_utils.containsCSSClassName(element, className);
    header = isHeader ? element : element['firstChild'];
    dimensionValue = this.getElementDir(header, dimension);

    while (endPixel - dimensionValue > threshold)
    {
        this._remove(element);

        removedHeadersDimension += dimensionValue;
        removedHeaders += isHeader ? 1 : this._getAttribute(element, 'extent', true);

        endPixel -= dimensionValue;

        element = headersContainer['lastChild'];
        isHeader = this.m_utils.containsCSSClassName(element, className);
        header = isHeader ? element : element['firstChild'];
        dimensionValue = this.getElementDir(header, dimension);
    }

    if (!isHeader)
    {
        returnVal = this.removeHeadersFromEndOfContainer(element, endPixel, threshold, className, dimension);

        this._setAttribute(element, 'extent', this._getAttribute(element, 'extent', true) - returnVal.extentChange);
        this.setElementDir(header, this.getElementDir(header, dimension) - returnVal.dimensionChange, dimension);

        removedHeaders += returnVal.extentChange;
        removedHeadersDimension += returnVal.dimensionChange;
    }

    return {extentChange: removedHeaders, dimensionChange: removedHeadersDimension};
};

/**
 * Remove column start and end headers to the left of the current viewport
 */
DvtDataGrid.prototype.removeColumnHeadersFromLeft = function()
        {
    var colHeaderContent, colThreshold, returnVal, colEndHeaderContent;
    // clean up left column headers
    if ((this.m_endColHeader - this.m_startColHeader) > this.MAX_COLUMN_THRESHOLD)
    {
        colHeaderContent = this.m_colHeader['firstChild'];
        colThreshold = this.getColumnThreshold();
        if (this.m_startColHeaderPixel <= this.m_currentScrollLeft - colThreshold)
        {
            returnVal = this.removeHeadersFromStartOfContainer(colHeaderContent, null, this.m_startColHeaderPixel, colThreshold, this.getMappedStyle('colheadercell'), 'width', this.getResources().isRTLMode() ? "right" : "left", this.m_currentScrollLeft);

            this.m_startColHeaderPixel += returnVal.dimensionChange;
            this.m_startColHeader += returnVal.extentChange;
        }
    }

    if ((this.m_endColEndHeader - this.m_startColEndHeader) > this.MAX_COLUMN_THRESHOLD)
    {
        colEndHeaderContent = this.m_colEndHeader['firstChild'];
        colThreshold = this.getColumnThreshold();
        if (this.m_startColEndHeaderPixel < this.m_currentScrollLeft - colThreshold)
        {
            returnVal = this.removeHeadersFromStartOfContainer(colEndHeaderContent, null, this.m_startColEndHeaderPixel, colThreshold, this.getMappedStyle('colendheadercell'), 'width', this.getResources().isRTLMode() ? "right" : "left", this.m_currentScrollLeft);

            this.m_startColEndHeaderPixel += returnVal.dimensionChange;
            this.m_startColEndHeader += returnVal.extentChange;
        }
    }
};

/**
 * Remove cells to the left of the current viewport
 * @param {Element} databody - the root of the databody
 */
DvtDataGrid.prototype.removeColumnsFromLeft = function(databody)
{
    var databodyContent, cells, cell, width, colThreshold, i, j, index, returnVal;
    // clean up right column headers
    if ((this.m_endCol - this.m_startCol) > this.MAX_COLUMN_THRESHOLD)
    {
        databodyContent = databody['firstChild'];
        cells = databodyContent['childNodes'];
        colThreshold = this.getColumnThreshold();

        // no rows in databody, nothing to remove
        if (cells.length < 1)
        {
            return;
        }

        returnVal = this._removeCellsAlongAxis('column', colThreshold, false);
        this.m_startColPixel += returnVal.dimensionChange;
        this.m_startCol += returnVal.extentChange;
    }
};

/**
 * Remove column start and end headers to the right of the current viewport
 */
DvtDataGrid.prototype.removeColumnHeadersFromRight = function()
{
    var colHeaderContent, colThreshold, returnVal;
    colThreshold = this.m_currentScrollLeft + this.getViewportWidth() + this.getColumnThreshold();

    // clean up right column headers
    if ((this.m_endColHeader - this.m_startColHeader) > this.MAX_COLUMN_THRESHOLD)
    {
        colHeaderContent = this.m_colHeader['firstChild'];
        // don't clean up if end of row header is not below the bottom of viewport
        if (this.m_endColHeaderPixel > colThreshold)
        {
            if (this.m_stopColumnHeaderFetch)
            {
                this.m_stopColumnHeaderFetch = false;
            }

            returnVal = this.removeHeadersFromEndOfContainer(colHeaderContent, this.m_endColHeaderPixel, colThreshold, this.getMappedStyle('colheadercell'), 'width');

            this.m_endColHeaderPixel -= returnVal.dimensionChange;
            this.m_endColHeader -= returnVal.extentChange;
        }
    }

    // clean up right column end headers
    if ((this.m_endColEndHeader - this.m_startColEndHeader) > this.MAX_COLUMN_THRESHOLD)
    {
        colHeaderContent = this.m_colEndHeader['firstChild'];
        // don't clean up if end of row header is not below the bottom of viewport
        if (this.m_endColEndHeaderPixel > colThreshold)
        {
            if (this.m_stopColumnEndHeaderFetch)
            {
                this.m_stopColumnEndHeaderFetch = false;
            }

            returnVal = this.removeHeadersFromEndOfContainer(colHeaderContent, this.m_endColEndHeaderPixel, colThreshold, this.getMappedStyle('colendheadercell'), 'width');

            this.m_endColEndHeaderPixel -= returnVal.dimensionChange;
            this.m_endColEndHeader -= returnVal.extentChange;
        }
    }
};

/**
 * Remove cells to the right of the current viewport
 * @param {Element} databody - the root of the databody
 */
DvtDataGrid.prototype.removeColumnsFromRight = function(databody)
{
    var databodyContent, threshold, cells, cell, width, i, j, index, returnVal;
    // clean up right column headers
    if ((this.m_endCol - this.m_startCol) > this.MAX_COLUMN_THRESHOLD)
    {
        databodyContent = databody['firstChild'];
        cells = databodyContent['childNodes'];
        threshold = this.m_currentScrollLeft + this.getViewportWidth() + this.getColumnThreshold();

        // don't clean up if end of row header is not below the bottom of viewport
        // no rows in databody, nothing to remove
        if (this.m_endColPixel <= threshold || cells.length < 1)
        {
            return;
        }

        if (this.m_stopColumnFetch)
        {
            this.m_stopColumnFetch = false;
        }

        returnVal = this._removeCellsAlongAxis('column', threshold, true);
        this.m_endColPixel -= returnVal.dimensionChange;
        this.m_endCol -= returnVal.extentChange;
    }
};

/**
 * Remove row start and end headers above the current viewport
 */
DvtDataGrid.prototype.removeRowHeadersFromTop = function()
{
    var rowHeaderContent, rowThreshold, returnVal, rowEndHeaderContent;
    if ((this.m_endRowHeader - this.m_startRowHeader) > this.MAX_ROW_THRESHOLD)
    {
        rowHeaderContent = this.m_rowHeader['firstChild'];
        rowThreshold = this.getRowThreshold();
        if (!(this.m_startRowHeaderPixel >= this.m_currentScrollTop - rowThreshold))
        {
            returnVal = this.removeHeadersFromStartOfContainer(rowHeaderContent, null, this.m_startRowHeaderPixel, rowThreshold, this.getMappedStyle('rowheadercell'), 'height', 'top', this.m_currentScrollTop);

            this.m_startRowHeaderPixel += returnVal.dimensionChange;
            this.m_startRowHeader += returnVal.extentChange;
        }
    }

    if ((this.m_endRowEndHeader - this.m_startRowEndHeader) > this.MAX_ROW_THRESHOLD)
{
        rowEndHeaderContent = this.m_rowEndHeader['firstChild'];
        rowThreshold = this.getRowThreshold();
        if (!(this.m_startRowEndHeaderPixel >= this.m_currentScrollTop - rowThreshold))
        {
            returnVal = this.removeHeadersFromStartOfContainer(rowEndHeaderContent, null, this.m_startRowEndHeaderPixel, rowThreshold, this.getMappedStyle('rowendheadercell'), 'height', 'top', this.m_currentScrollTop);

            this.m_startRowEndHeaderPixel += returnVal.dimensionChange;
            this.m_startRowEndHeader += returnVal.extentChange;
        }
    }
};

/**
 * Remove rows/cells above the current viewport
 * @param {Element} databody - the root of the databody
 */
DvtDataGrid.prototype.removeRowsFromTop = function(databody)
{
    var databodyContent, rowThreshold, cell, height, i, j, index, returnVal;
    if ((this.m_endRow - this.m_startRow) > this.MAX_ROW_THRESHOLD)
    {
        databodyContent = databody['firstChild'];
        rowThreshold = this.getRowThreshold();
        if (this.m_startRowPixel >= this.m_currentScrollTop - rowThreshold)
        {
            return;
        }

        // remove all rows from top until the threshold is reached
         returnVal = this._removeCellsAlongAxis('row', rowThreshold, false);
        this.m_startRowPixel += returnVal.dimensionChange;
        this.m_startRow += returnVal.extentChange;
    }
};

/**
 * Remove row start and end headers below the current viewport
 */
DvtDataGrid.prototype.removeRowHeadersFromBottom = function()
{
    var rowHeaderContent, rowThreshold, returnVal, rowEndHeaderContent;
    rowThreshold = this.m_currentScrollTop + this.getViewportHeight() + this.getRowThreshold();

    // clean up bottom row headers
    if ((this.m_endRowHeader - this.m_startRowHeader) > this.MAX_ROW_THRESHOLD)
    {
        rowHeaderContent = this.m_rowHeader['firstChild'];
        // don't clean up if end of row header is not below the bottom of viewport
        if (!(this.m_endRowHeaderPixel <= rowThreshold))
        {
            if (this.m_stopRowHeaderFetch)
            {
                this.m_stopRowHeaderFetch = false;
            }

            returnVal = this.removeHeadersFromEndOfContainer(rowHeaderContent, this.m_endRowHeaderPixel, rowThreshold, this.getMappedStyle('rowheadercell'), 'height');

            this.m_endRowHeaderPixel -= returnVal.dimensionChange;
            this.m_endRowHeader -= returnVal.extentChange;
        }
    }

    // clean up bottom row headers
    if ((this.m_endRowEndHeader - this.m_startRowEndHeader) > this.MAX_ROW_THRESHOLD)
    {
        rowEndHeaderContent = this.m_rowEndHeader['firstChild'];
        // don't clean up if end of row header is not below the bottom of viewport
        if (!(this.m_endRowEndHeaderPixel <= rowThreshold))
        {
            if (this.m_stopRowEndHeaderFetch)
            {
                this.m_stopRowEndHeaderFetch = false;
            }

            returnVal = this.removeHeadersFromEndOfContainer(rowEndHeaderContent, this.m_endRowEndHeaderPixel, rowThreshold, this.getMappedStyle('rowendheadercell'), 'height');

            this.m_endRowEndHeaderPixel -= returnVal.dimensionChange;
            this.m_endRowEndHeader -= returnVal.extentChange;
        }
    }
};

/**
 * Remove rows/cells below the current viewport
 * @param {Element} databody - the root of the databody
 */
DvtDataGrid.prototype.removeRowsFromBottom = function(databody)
{
    var databodyContent, threshold, cell, i, j, height, index, returnVal;
    if ((this.m_endRow - this.m_startRow) > this.MAX_ROW_THRESHOLD)
    {
        databodyContent = databody['firstChild'];
        threshold = this.m_currentScrollTop + this.getViewportHeight() + this.getRowThreshold();

        // don't clean up if end of row header is not below the bottom of viewport
        if (this.m_endRowPixel <= threshold)
        {
            return;
        }

        if (this.m_stopRowFetch)
        {
            this.m_stopRowFetch = false;
        }

        returnVal = this._removeCellsAlongAxis('row', threshold, true);
        this.m_endRowPixel -= returnVal.dimensionChange;
        this.m_endRow -= returnVal.extentChange;
    }
};

/**
 * Debugging method to dump out current range info
 */
DvtDataGrid.prototype.dumpRanges = function()
{
    if (console != undefined && console.log)
    {
        console.log("=================");
        console.log("Start Row: " + this.m_startRow);
        console.log("  End Row: " + this.m_endRow);
        console.log("Start Column: " + this.m_startCol);
        console.log("  End Column: " + this.m_endCol);
        console.log("=================");
    }
};
/*********************************** end scrolling/virtualization ************************************/

/*********************************** start dom event handling ***************************************/
/**
 * Handle the context menu gesture
 * @param {Event} event the event of the context menu gesture
 * @param {string} eventType keyboard/touch/mouse
 * @param {Function} callback where to pass the data back
 */
DvtDataGrid.prototype.handleContextMenuGesture = function(event, eventType, callback)
{
    var index, element, capabilities, launcher, target;
    // if we are on a touch device and in a cell we need to set the correct active
    // and call focus before triggering the context menu to open. headers take
    // care of this by setting active in the 300ms callback for tap+short hold
    target = /** @type {Element} */ (event['originalEvent'].target);
    element = this.findCell(target);
    if (eventType === 'touch' && element != null)
    {
        index = this.getCellIndexes(element);
        // if right click and inside multiple selection or current active do not change anything
        if ((!this.isMultipleSelection() || !this._isContainSelection(index)) ||
                (this._isDatabodyCellActive() && index['row'] != this.m_active['indexes']['row'] && index['column'] != this.m_active['indexes']['column']))
        {
            if (this._isSelectionEnabled())
            {
                this.handleDatabodyClickSelection(event['originalEvent']);
            }
            else
            {
                //activate on a tap
                this.handleDatabodyClickActive(event['originalEvent']);
            }
        }
    }

    // first check if we are invoking on an editable or clickable element, if so bail
    if (this.m_utils._isNodeEditableOrClickable(target, this.m_root))
    {
        return;
    }

    // enable and disable context menu items depending on capability of the datasource and options
    // if the action was performed on a cell
    if (element != null)
    {
        index = this.getCellIndexes(element);
        // if fired from inside a multiple selection
        if (this.isMultipleSelection() && this._isContainSelection(index))
        {
            launcher = this._getActiveElement();
            // if there is an active cell we want that to be the launcher of the context menu so
            // that focus can be restored to it. If it fired form the keyboard open with launcher and context
            // of the active cell, if right click or touch open with the context of the clicked cell
            if (this._isDatabodyCellActive())
            {
                capabilities = eventType === 'keyboard' ? this._getCellCapability(launcher) : this._getCellCapability(launcher, element);
            }
            // there is the case where header is active and entire row/column selected
            else
            {
                // the launcher will be the active header, and the context of the menu will be relative to the active header
                capabilities = this._getHeaderCapability(launcher, element);
            }
        }
        else
        {
            // open on the cell with its context
            launcher = element;
            capabilities = this._getCellCapability(launcher);
        }
    }
    else
    {
        element = this.findHeader(target);
        if (element == null)
        {
            // not a header or cell don't do anything
            return;
        }
        capabilities = this._getHeaderCapability(element);
        launcher = element;
    }

    callback.call(null, {'capabilities': capabilities, 'launcher': launcher}, event, eventType);
};

/**
 * Get the capabilities for context menu opened on a cell
 * @param {Element} cell the cell whose context we want
 * @param {Element=} actualCell the cell with context menu opened on it
 * @return {Object} capabilities object with props resize, resizeWidth, resizeHeight, sortRow, sortCol, cut, paste
 * @private
 */
DvtDataGrid.prototype._getCellCapability = function(cell, actualCell)
{
    var capabilities, rowHeader, columnHeader, resizable, sortable,
            sameColumn = true, sameRow = true, disable, enable, sorted;
    disable = 'disable';
    enable = 'enable';
    capabilities = {'resize': disable, 'resizeWidth': disable, 'resizeHeight': disable,
        'sortRow': disable, 'sortCol': disable, 'cut': disable, 'paste': disable,
        'sortColAsc': disable, 'sortColDsc': disable};

    // if there is an actual cell that means we want the context relative to that cell,
    // so if it is the same column, our column operations (resize width, sort column) can
    // be utilized. If it's in the same row the row operations (resize height, sort row, cut, paste)
    // can be utilized
    if (actualCell != null)
    {
        sameColumn = this._getIndex(cell, 'column') === this._getIndex(actualCell, 'column');
        sameRow = this._getKey(cell, 'row') === this._getKey(actualCell, 'row');
        if (sameRow === false && sameColumn === false)
        {
            return capabilities;
        }
    }
    rowHeader = this.getHeaderFromCell(cell, 'row');
    columnHeader = this.getHeaderFromCell(cell, 'column');
    resizable = this.getResources().getMappedAttribute('resizable');
    sortable = this.getResources().getMappedAttribute('sortable');
    if (columnHeader != null && sameColumn)
    {
        if (columnHeader.getAttribute(resizable) === 'true')
        {
            capabilities['resize'] = enable;
            capabilities['resizeWidth'] = enable;
        }
        if (columnHeader.getAttribute(sortable) === 'true')
        {
            capabilities['sortCol'] = enable;
            capabilities['sortColAsc'] = enable;
            capabilities['sortColDsc'] = enable;
            sorted = columnHeader.getAttribute(this.getResources().getMappedAttribute('sortDir'));
            if (sorted === 'ascending')
            {
                capabilities['sortColAsc'] = disable;
            }
            else if (sorted === 'descending')
            {
                capabilities['sortColDsc'] = disable;
            }
        }
    }
    if (sameRow)
    {
        if (this._isMoveEnabled('row'))
        {
            capabilities['cut'] = enable;
            capabilities['paste'] = enable;
        }
        if (rowHeader != null)
        {
            if (rowHeader.getAttribute(resizable) === 'true')
            {
                capabilities['resize'] = enable;
                capabilities['resizeHeight'] = enable;
            }
            if (rowHeader.getAttribute(sortable) === 'true')
            {
                capabilities['sortRow'] = enable;
            }
        }
    }
    return capabilities;
};

/**
 * Get the capabilities for context menu opened on a header
 * @param {Element} header the header whose context we want
 * @param {Element=} actualCell the cell that we are actually opening on
 * @return {Object} capabilities object with props resizeWidth, resizeHeight, sortRow, sortCol
 * @private
 */
DvtDataGrid.prototype._getHeaderCapability = function(header, actualCell)
{
    var capabilities, resizable, sortable, sameColumn = true, sameRow = true, disable, enable, sorted, axis;
    disable = 'disable';
    enable = 'enable';
    capabilities = {'resize': disable, 'resizeWidth': disable, 'resizeHeight': disable,
        'sortRow': disable, 'sortCol': disable, 'cut': disable, 'paste': disable,
        'sortColAsc': disable, 'sortColDsc': disable};

    // if there is an actual cell that means we want the context relative to that cell,
    // so if it is the same column, our column operations (resize width, sort column) can
    // be utilized. If it's in the same row the row operations (resize height, sort row, cut, paste)
    // can be utilized
    if (actualCell != null)
    {
        sameColumn = this.getHeaderCellIndex(header) === this._getIndex(actualCell, 'column');
        sameRow = this._getKey(header, 'row') === this._getKey(actualCell, 'row');
        if (sameRow === false && sameColumn === false)
        {
            return capabilities;
        }
    }

    axis = this.getHeaderCellAxis(header);
    resizable = this.getResources().getMappedAttribute('resizable');
    sortable = this.getResources().getMappedAttribute('sortable');
    if (header !== null)
    {
        if ((axis == 'column' || axis =='columnEnd') && sameColumn)
        {
            if (header.getAttribute(resizable) === 'true')
            {
                capabilities['resizeWidth'] = enable;
                capabilities['resize'] = enable;
            }
            capabilities['resizeHeight'] = this.m_options.isResizable(axis, 'height');
            if (header.getAttribute(sortable) === 'true')
            {
                capabilities['sortCol'] = enable;
                capabilities['sortColAsc'] = enable;
                capabilities['sortColDsc'] = enable;
                sorted = header.getAttribute(this.getResources().getMappedAttribute('sortDir'));
                if (sorted === 'ascending')
                {
                    capabilities['sortColAsc'] = disable;
                }
                else if (sorted === 'descending')
                {
                    capabilities['sortColDsc'] = disable;
                }
            }
        }
        else if (sameRow)
        {
            if (this._isMoveEnabled('row'))
            {
                capabilities['cut'] = enable;
                capabilities['paste'] = enable;
            }
            if (header.getAttribute(resizable) === 'true')
            {
                capabilities['resize'] = enable;
                capabilities['resizeHeight'] = enable;
            }
            capabilities['resizeWidth'] = this.m_options.isResizable(axis, 'width');
            if (header.getAttribute(sortable) === 'true')
            {
                capabilities['sortRow'] = enable;
            }
        }
    }
    capabilities['resize'] = capabilities['resizeHeight'] === enable ||
            capabilities['resizeWidth'] === enable ? enable : disable;

    return capabilities;
};

/**
 * Handle the callback from the widget to resize or sort.
 * @param {Event} event - the original contextmenu event
 * @param {string} id - the id returned from the context menu
 * @param value - the value set in the dialog on resizing
 */
DvtDataGrid.prototype.handleContextMenuReturn = function(event, id, value)
{
    var target, direction;

    // the target is the active element at all times
    if (this.m_active != null)
    {
        target = this._getActiveElement();
    }

    if (id === this.m_resources.getMappedCommand('resizeHeight') || id === this.m_resources.getMappedCommand('resizeWidth'))
    {
        if ((this.isResizeEnabled()))
        {
            // target may not be (event.target)
            this.handleContextMenuResize(event, id, value, target);
        }
    }
    else if (id === this.m_resources.getMappedCommand('sortColAsc') || id === this.m_resources.getMappedCommand('sortColDsc'))
    {
        direction = id === this.m_resources.getMappedCommand('sortColAsc') ? 'ascending' : 'descending';
        if (this.m_utils.containsCSSClassName(target, this.getMappedStyle('cell')))
        {
            target = this.getHeaderFromCell(target, 'column');
        }
        if (this._isDOMElementSortable(target))
        {
            this._handleCellSort(event, direction, target);
        }
    }
    else if (id === this.m_resources.getMappedCommand('sortRowAsc') || id === this.m_resources.getMappedCommand('sortRowDsc'))
    {
        direction = id === this.m_resources.getMappedCommand('sortRowAsc') ? 'ascending' : 'descending';
        if (this.m_utils.containsCSSClassName(target, this.getMappedStyle('cell')))
        {
            target = this.getHeaderFromCell(target, 'row');
        }
        if (this._isDOMElementSortable(target))
        {
            this._handleCellSort(event, direction, target);
        }
    }
    else if (id === this.m_resources.getMappedCommand('cut'))
    {
        this._handleCut(event, target);
    }
    else if (id === this.m_resources.getMappedCommand('paste'))
    {
        this._handlePaste(event, target);
    }
    else if (id === this.m_resources.getMappedCommand('discontiguousSelection'))
    {
        // handle discontiguous selection context menu
        this.setDiscontiguousSelectionMode(value);
    }
};

/**
 * Determined if sort is supported for the specified axis.
 * @param {string} axis the axis which we check whether sort is supported.
 * @param {Object} headerContext the header context object
 * @private
 */
DvtDataGrid.prototype._isSortEnabled = function(axis, headerContext)
{
    var capability, sortable;
    capability = this.getDataSource().getCapability("sort");
    sortable = this.m_options.isSortable(axis, headerContext);
    if ((sortable === "enable" || sortable === "auto") && (capability === "full" || capability === axis))
    {
        return true;
    }

    return false;
};

/**
 * Determined if sort is supported for the specified element.
 * @param {Element|undefined} element to check if sorting should be on
 * @private
 */
DvtDataGrid.prototype._isDOMElementSortable = function(element)
{
    if (element == null)
    {
        return false;
    }
    var header = this.findHeader(element);
    if (header == null)
    {
        return false;
    }
    return header.getAttribute(this.getResources().getMappedAttribute('sortable')) == "true";
};

/**
 * Check if selection enabled by options on the grid
 * @return {boolean} true if selection enabled
 * @private
 */
DvtDataGrid.prototype._isSelectionEnabled = function()
{
    return (this.m_options.getSelectionCardinality() != "none");
};

/**
 * Check whether multiple row/cell selection is allowed by options on the grid
 * @return {boolean} true if multipl selection enabled
 */
DvtDataGrid.prototype.isMultipleSelection = function()
{
    return (this.m_options.getSelectionCardinality() == "multiple");
};

/**
 * Check if resizing enabled on any header by options on the grid
 * @return {boolean} true if resize enabled
 */
DvtDataGrid.prototype.isResizeEnabled = function()
{
    return (this.m_options.isResizable('row', 'width') || this.m_options.isResizable('row', 'height') ||
            this.m_options.isResizable('column', 'width') || this.m_options.isResizable('column', 'height')||
            this.m_options.isResizable('rowEnd', 'width') || this.m_options.isResizable('rowEnd', 'height')||
            this.m_options.isResizable('columnEnd', 'width') || this.m_options.isResizable('columnEnd', 'height'));
};

/**
 * Check if resizing enabled on a specific header
 * @param {string} axis the axis which we check whether sort is supported.
 * @param {Object} headerContext the header context object
 * @return {boolean} true if resize enabled
 */
DvtDataGrid.prototype._isHeaderResizeEnabled = function(axis, headerContext)
{
    var resizable;
    if (axis == 'column' || axis == 'columnEnd')
    {
        resizable = this.m_options.isResizable(axis, 'width', headerContext);
        return resizable == 'enable' ? true : false;
    }
    else if (axis == 'row' || axis == 'rowEnd')
    {
        resizable = this.m_options.isResizable(axis, 'height', headerContext);
        return resizable == 'enable' ? true : false;
    }
    return false;
};

/**
 * Handle mousemove events on the document
 * @param {Event} event - mousemove event on the document
 */
DvtDataGrid.prototype.handleMouseMove = function(event)
{
    if (this.isResizeEnabled() && (this.m_databodyDragState == false))
    {
        this.handleResize(event);
    }
};

/**
 * Handle row header mousemove events on the document
 * @param {Event} event - mousemove event on the document
 */
DvtDataGrid.prototype.handleRowHeaderMouseMove = function(event)
{
    if (this.m_databodyMove)
    {
        this._handleMove(event);
    }
    else if (this.m_headerDragState && this.m_selectionFrontier && this.m_selectionFrontier['axis'].indexOf('row') != -1)
    {
        this.extendSelectionHeader(event.target, event);
    }
    // if it is resizing use the mouse move on the document
    else if (!this.m_isResizing)
    {
        this.handleMouseMove(event);
    }
};

/**
 * Handle row header mousemove events on the document
 * @param {Event} event - mousemove event on the document
 */
DvtDataGrid.prototype.handleColumnHeaderMouseMove = function(event)
{
    if (this.m_headerDragState && this.m_selectionFrontier && this.m_selectionFrontier['axis'].indexOf('column') != -1)
    {
        this.extendSelectionHeader(event.target, event);
    }
    // if it is resizing use the mouse move on the document
    else if (!this.m_isResizing)
    {
        this.handleMouseMove(event);
    }
};

/**
 * Handle mousedown events on the headers
 * @param {Event} event - mousedown event on the headers
 */
DvtDataGrid.prototype.handleHeaderMouseDown = function(event)
{
    var cell, row, processed, target, header, cellContext, shiftKey, selectionMode;

    this._exitActionableMode();
    target = /** @type {Element} */ (event.target);
    shiftKey = event.shiftKey;

    if (this._isEditOrEnter())
    {
        cell = this._getActiveElement();
        if (this._leaveEditing(event, cell, false) == false)
        {
            return;
        }
    }

    //only perform events on left mouse, (right in rtl culture)
    if (event.button === 0)
    {
        //if mousedown in an icon it the click event will handle mousedown/up
        if ((this.m_utils.containsCSSClassName(target, this.getMappedStyle('sortascending')) ||
                this.m_utils.containsCSSClassName(target, this.getMappedStyle('sortdescending')))
                && this._isDOMElementSortable(target))
        {
            event.preventDefault();
            this._handleSortIconMouseDown(target);
            return;
        }

        //handle resize movements first if we're on the border
        if (this.isResizeEnabled())
        {
            processed = this.handleResizeMouseDown(event);
        }

        //if our move is enabled make sure our row has the active cell in it
        if (!this.m_isResizing && this._isMoveOnElementEnabled(this.findHeader(target)))
        {
            this.m_databodyMove = true;
            this.m_currentX = event['pageX'];
            this.m_currentY = event['pageY'];
            processed = true;
        }
    }
    // activate header on click or right click
    if (!this.m_isResizing)
    {
        if (!this.m_root.contains(document.activeElement) || document.activeElement === this.m_root)
        {
            this.m_externalFocus = true;
        }

        selectionMode = this.m_options.getSelectionMode();
        header = this.findHeader(target);
        cellContext = header[this.getResources().getMappedAttribute("context")];

        // if click or right click we want to adjust the selction
        // no else so that we can select a cell in the same row as long as no drag
        // check if selection is enabled
        if (this._isSelectionEnabled() && this.isMultipleSelection() && !(selectionMode == 'row' && cellContext['axis'].indexOf('row') == -1) && !this.m_databodyMove)
        {
            this.handleHeaderClickSelection(event);

            // only allow drag on left click
            if (event.button === 0)
            {
                this.m_headerDragState = true;
            }
        }
        // for single row based selection only
        else if (selectionMode == 'row' && cellContext['axis'].indexOf('row') != -1 && this._isSelectionEnabled())
        {
            this.handleHeaderClickSelection(event);            
        }
        // if not selecting, just make active.
        else
        {
            this.handleHeaderClickActive(event);
        }
    }

    if (processed === true)
    {
        event.preventDefault();
    }
};

/**
 * Handle mouseup events on the document
 * @param {Event} event - mouseup event on the document
 */
DvtDataGrid.prototype.handleMouseUp = function(event)
{
    // toggle off the drag state
    this.m_headerDragState = false;
    this.m_databodyDragState = false;

    //if we mouseup outside the grid we want to cancel the selection and return the row
    if (this.m_databodyMove)
    {
        this._handleMoveMouseUp(event, false);
    }
    else if (this.isResizeEnabled())
    {
        this.handleResizeMouseUp(event);
    }

    this.m_databodyMove = false;
};

DvtDataGrid.prototype.handleHeaderMouseOver = function(event)
{
    var target = /** @type {Element} */ (event.target);
    this.m_utils.addCSSClassName(this.findHeader(target), this.getMappedStyle('hover'));
    if (this._isDOMElementSortable(target))
    {
        this._handleSortMouseOver(event);
    }
};

DvtDataGrid.prototype.handleHeaderMouseOut = function(event)
{
    var target = /** @type {Element} */ (event.target);
    this.m_utils.removeCSSClassName(this.findHeader(target), this.getMappedStyle('hover'));
    if (this._isDOMElementSortable(target))
    {
        this._handleSortMouseOut(event);
    }
};

/**
 * Event handler for when header mouse up event
 * @protected
 * @param {Event} event - header mouse up event
 */
DvtDataGrid.prototype.handleHeaderMouseUp = function(event)
{
    // handle anchor change while in header drag mode
    this.handleDragAnchorChange(event);

    // toggle off the drag state
    this.m_headerDragState = false;
    this.m_databodyDragState = false;
    if (this.m_databodyMove)
    {
        this._handleMoveMouseUp(event, true);
    }
};

/**
 * Event handler for when the top left corner is clicked
 * @protected
 * @param {Event} event - click event on the top left corner
 */
DvtDataGrid.prototype.handleCornerClick = function(event)
{
    this._handleSelectAll(event);
}

/**
 * Event handler for when row/column header is clicked
 * @protected
 * @param {Event} event - click event on the headers
 */
DvtDataGrid.prototype.handleHeaderClick = function(event)
{
    var target = /** @type {Element} */ (event.target);
    if ((this.m_utils.containsCSSClassName(target, this.getMappedStyle('sortascending')) ||
            this.m_utils.containsCSSClassName(target, this.getMappedStyle('sortdescending')))
            && this._isDOMElementSortable(target))
    {
        this._removeTouchSelectionAffordance();
        this._handleHeaderSort(event);
        event.preventDefault();
    }
};

/**
 * Event handler for when mouse down anywhere in the databody
 * @protected
 * @param {Event} event - mousedown event on the databody
 */
DvtDataGrid.prototype.handleDatabodyMouseDown = function(event)
{
    var cell, activeCell, target;
    target = /** @type {Element} */ (event.target);
    cell = this.findCell(target);
    if (cell == null)
    {
        this.m_scrollbarFocus = true;
        return;
    }

    if (this._isEditOrEnter())
    {
        activeCell = this._getActiveElement();
        if (cell != activeCell)
        {
            if (this._leaveEditing(event, activeCell, false) == false)
            {
                return;
            }
        }
        else
        {
            return;
        }
    }
    else
    {
        // reset actionable mode whenever user clicks in the databody
        this._exitActionableMode();
    }

    //only perform events on left mouse, (right in rtl culture)
    if (event.button === 0)
    {
        if (this._isMoveOnElementEnabled(cell))
        {
            this.m_databodyMove = true;
            this.m_currentX = event['pageX'];
            this.m_currentY = event['pageY'];
        }
    }

    if (!this.m_root.contains(document.activeElement) || document.activeElement === this.m_root)
    {
        this.m_externalFocus = true;
    }

    if (this._isGridEditable())
    {
        this.m_shouldFocus = !this._isFocusableElementBeforeCell(target);
    }

    // if click or right click we want to adjust the selction
    // no else so that we can select a cell in the same row as long as no drag
    // check if selection is enabled
    if (this._isSelectionEnabled())
    {
        this.handleDatabodyClickSelection(event);

        // only allow drag on left click
        if (this.isMultipleSelection() && event.button === 0)
        {
            this.m_databodyDragState = true;
        }
    }
    else
    {
        // if selection is disable, we'll still need to highlight the active cell
        this.handleDatabodyClickActive(event);
    }
};

DvtDataGrid.prototype.handleDatabodyMouseOut = function(event)
{
    var target, targetCell;
    if (!this.m_databodyMove)
    {
        target = /** @type {Element} */ (event.target);
        targetCell = this.findCell(target);
        this._setCellHover(targetCell, 'remove');
    }
};

DvtDataGrid.prototype.handleDatabodyMouseOver = function(event)
{
    var target, targetCell;
    if (!this.m_databodyMove)
    {
        target = /** @type {Element} */ (event.target);
        targetCell = this.findCell(target);
        this._setCellHover(targetCell, 'add');
    }
};

DvtDataGrid.prototype._setCellHover = function(targetCell, addOrRemove)
{
    if (targetCell != null)
    {
        var selectionMode = this.m_options.getSelectionMode();
        if (selectionMode === 'cell')
        {
            addOrRemove == 'add' ? this.m_utils.addCSSClassName(targetCell, this.getMappedStyle('hover')) : this.m_utils.removeCSSClassName(targetCell, this.getMappedStyle('hover'));
        }
        else if (selectionMode === 'row')
        {
            var index = this._getIndex(targetCell, 'row');
            var returnObj = this._getSelectionStartAndEnd(this.createIndex(index, this.m_startCol), this.createIndex(index, this.m_endCol), 0);
            for (var i = returnObj['min']['row']; i <= returnObj['max']['row']; i++)
            {
                this._highlightCellsAlongAxis(i, 'row', 'index', addOrRemove, ['hover']);
            }
        }
    }
};

DvtDataGrid.prototype.handleDatabodyDoubleClick = function(event)
{
    var target, cell, currentMode, activeCell;
    if (this._isGridEditable())
    {
        target = event.target;
        cell = this.findCell(target);
        currentMode = this._getCurrentMode();
        if (currentMode == 'edit')
        {
            activeCell = this._getActiveElement();
            if (cell == activeCell)
            {
                // if the active cell is being edited and it is the target do not eat the double click
                return;
            }
            if (!this._handleExitEdit(event, activeCell))
            {
                return;
            }
        }
        this._handleEditable(event, cell);
        this._handleEdit(event, cell);
    }
};

/**
 * Event handler for when mouse move anywhere in the databody
 * @protected
 * @param {Event} event - mousemove event on the databody
 */
DvtDataGrid.prototype.handleDatabodyMouseMove = function(event)
{
    //handle move first because it should happen first on the second click
    if (this.m_databodyMove)
    {
        this._handleMove(event);
    }
    else if (this.m_databodyDragState)
    {
        this.handleDatabodySelectionDrag(event);
    }
};

/**
 * Event handler for when mouse down anywhere in the databody
 * @protected
 * @param {Event} event - mouseup event on the databody
 */
DvtDataGrid.prototype.handleDatabodyMouseUp = function(event)
{
    this.m_databodyDragState = false;
    this.m_headerDragState = false;
    if (this.m_databodyMove)
    {
        this._handleMoveMouseUp(event, true);
    }
};

/**
 * Event handler for when user press down a key in the databody
 * @protected
 * @param {Event} event - keydown event on the databody
 */
DvtDataGrid.prototype.handleDatabodyKeyDown = function(event)
{
    var action, element, target, keyCode, ctrlKey;
    target = /** @type {Element} */ (event.target);
    keyCode = event.keyCode;
    ctrlKey = this.m_utils.ctrlEquivalent(event);

    // no longer fire keydown, just check if row expander handled the event already
    // also ignore if the component is animating
    if ((event.defaultPrevented && ctrlKey && (DvtDataGrid.keyCodes.LEFT_KEY || DvtDataGrid.keyCodes.RIGHT_KEY)) || this.m_animating)
    {
        return;
    }

    // check if header is active
    if (this.m_active != null && this.m_active['type'] == 'header')
    {
        action = this._getActionFromKeyDown(event, this.m_active['axis']);
    }
    else
    {
        action = this._getActionFromKeyDown(event, 'cell');
    }

    element = this._getActiveElement();

    if (action != null)
    {
        if (action.call(this, event, element))
        {
            event.preventDefault();
        }
    }
};

/**
 * Handles when a key down is pressed on the databody or headers
 * @param {Event} event the DOM event
 * @private
 */
DvtDataGrid.prototype._fireKeyDownEvent = function(event)
{
    var rowKey, details;
    // keydown always acts on the active cell
    rowKey = this._getActiveKey('row');

    // the event contains the context info
    details = {
        'event': event, 'ui': {
            'rowKey': rowKey
        }
    };

    return this.fireEvent('keydown', details);
};

/**
 * Find top and left offset of an element relative to the (0,0) point on the page
 * @param {Element} element - the element to find left and top offset of
 * @return {Array.<number>} - [leftOffset, topOffset]
 */
DvtDataGrid.prototype.findPos = function(element)
{
    var parentPos, transform;
    if (element)
    {
        parentPos = this.findPos(element['offsetParent']);
        transform = this.getElementTranslationXYZ(element['offsetParent']);
        return [
            parseInt(parentPos[0], 10) + parseInt(element['offsetLeft'], 10) + transform[0],
            parseInt(parentPos[1], 10) + parseInt(element['offsetTop'], 10) + transform[1]
        ];
    }
    return [0, 0];
};

/**
 * Get an elements transform3d X,Y,Z
 * @param {Element} element - the element to find transform3d X,Y,Z of
 * @return {Array.<number>} - [transformX, transformY, transformZ]
 */
DvtDataGrid.prototype.getElementTranslationXYZ = function(element)
{
    var cs, transform, matrixArray, transformX, transformY, transformZ;
    if (element)
    {
        cs = document.defaultView.getComputedStyle(element, null);
        transform = cs.getPropertyValue("-webkit-transform") ||
                cs.getPropertyValue("-moz-transform") ||
                cs.getPropertyValue("-ms-transform") ||
                cs.getPropertyValue("-o-transform") ||
                cs.getPropertyValue("transform");
        matrixArray = transform.substr(7, transform.length - 8).split(', ');
        transformX = isNaN(parseInt(matrixArray[4], 10)) ? 0 : parseInt(matrixArray[4], 10);
        transformY = isNaN(parseInt(matrixArray[5], 10)) ? 0 : parseInt(matrixArray[5], 10);
        transformZ = isNaN(parseInt(matrixArray[6], 10)) ? 0 : parseInt(matrixArray[6], 10);
        return [transformX, transformY, transformZ];
    }
    return [0, 0, 0];
};



/**
 * Event handler for when mouse wheel is used on the databody
 * @param {Event} event - mousewheel event on the databody
 */
DvtDataGrid.prototype.handleDatabodyMouseWheel = function(event)
{
    var delta, deltaX, deltaY;
    // prevent scrolling of the page
    event.preventDefault();

    // prevent scroll if animating sort
    if (this.m_animating)
    {
        return;
    }


    delta = this.m_utils.getMousewheelScrollDelta(event);

    deltaX = delta['deltaX'];
    deltaY = delta['deltaY'];

    this.scrollDelta(deltaX, deltaY);
};

/**************** touch related methods ********************/

/**
 * Event handler for when touch is started on the databody
 * @param {Event} event - touchstart event on the databody
 */
DvtDataGrid.prototype.handleTouchStart = function(event)
{
    var fingerCount, dir, selection, target;

    fingerCount = event.touches.length;
    target = /** @type {Element} */ (event.touches[0].target);

    // move = one finger swipe (or two?)
    if (fingerCount == 1)
    {
        // get the coordinates of the touch
        this.m_startX = event.touches[0].pageX;
        this.m_startY = event.touches[0].pageY;

        // need these to detect whether touch is hold and move vs. swipe
        this.m_currentX = this.m_startX;
        this.m_currentY = this.m_startY;
        this.m_prevX = this.m_startX;
        this.m_prevY = this.m_startY;
        this.m_startTime = (new Date()).getTime();

        // flag it
        this.m_touchActive = true;

        //if multiple select enabled check to see if the touch start was on a select affordance
        if (this.isMultipleSelection())
        {
            //if the target is not the container, but rather the icon itself, choose the container instead
            if (target['className'] === this.getMappedStyle("selectaffordancetop") || target['className'] === this.getMappedStyle("selectaffordancebottom"))
            {
                target = target['parentNode'];
            }

            //determine which icon was clicked on
            dir = target === this.m_topSelectIconContainer ? 'top' : target === this.m_bottomSelectIconContainer ? 'bottom' : null;

            if (dir)
            {
                //keeps track of multiple select mode
                this.m_touchMultipleSelect = true;
                selection = this.GetSelection();
                if (dir === 'top')
                {
                    //anchor is bottom right of selection for selecting top affordance
                    this.m_touchSelectAnchor = selection[selection.length - 1]['endIndex'];
                }
                else
                {
                    //anchor is top left of selection for selecting bottom affordance
                    this.m_touchSelectAnchor = selection[selection.length - 1]['startIndex'];
                }
            }
        }

        //if not multiple select, check for row reorder
        if (!this.m_touchMultipleSelect && this._isMoveOnElementEnabled(this.findCell(target)))
        {
            this.m_databodyMove = true;
        }
    }
    else
    {
        // more than one finger touched so cancel
        this.handleTouchCancel(event);
    }
};

/**
 * Event handler for when touch moves on the databody
 * @param {Event} event - touchmove event on the databody
 */
DvtDataGrid.prototype.handleTouchMove = function(event)
{
    var diffX, diffY, target;
    target = /** @type {Element} */ (event.target);

    if (this.m_touchActive)
    {
        event.preventDefault();
        this.m_currentX = event.touches[0].pageX;
        this.m_currentY = event.touches[0].pageY;

        diffX = this.m_currentX - this.m_prevX;
        diffY = this.m_currentY - this.m_prevY;
        if (this.getResources().isRTLMode())
        {
            diffX = diffX * -1;
        }

        if (this.m_touchMultipleSelect)
        {
            this.handleDatabodySelectionDrag(event);
        }
        else if (this.m_databodyMove)
        {
            this._removeTouchSelectionAffordance();
            this._handleMove(event);
        }
        else
        {
            if (this._isEditOrEnter())
            {
                var cell = this._getActiveElement();
                if (this.findCell(target) != cell)
                {
                    this._handleNonSwipeScroll(diffX, diffY);
                }
            }
            else
            {
                this._handleNonSwipeScroll(diffX, diffY);
            }
        }

        this.m_prevX = this.m_currentX;
        this.m_prevY = this.m_currentY;
    }
    else
    {
        this.handleTouchCancel(event);
    }
};

/**
 * Event handler for when touch ends on the databody
 * @param {Event} event - touchend event on the databody
 */
DvtDataGrid.prototype.handleTouchEnd = function(event)
{
    var duration, cell, target;
    target = /** @type {Element} */ (event.target);

    if (this._isEditOrEnter())
    {
        cell = this._getActiveElement();
        if (this.findCell(target) != cell)
        {
            this._leaveEditing(event, cell, false);
        }
        else
        {
            this.handleTouchCancel(event);
            return;
        }
    }
    else
    {
        // reset actionable mode whenever user clicks in the databody
        this._exitActionableMode();
    }

    if (this.m_lastTapTime != null && (this.m_startTime - this.m_lastTapTime) < 250 && this.m_lastTapTarget === target)
    {
        this.m_lastTapTime = null;
        this.m_lastTapTarget = null;
        cell = this.findCell(target);
        if (cell != null)
        {
            this._handleEditable(event, cell);
            this._handleEdit(event, cell);
            event.preventDefault();
        }
    }
    else
    {
        this.m_lastTapTarget = event.target;
        this.m_lastTapTime = (new Date()).getTime();
    }

    if (this.m_touchActive && !event.defaultPrevented)
    {
        if (this.m_touchMultipleSelect)
        {
            event.preventDefault();
            this.m_touchMultipleSelect = false;
        }
        else
        {
            duration = this.m_lastTapTime - this.m_startTime;
            if (this.m_currentX == this.m_startX && this.m_currentY == this.m_startY)
            {
                // this means we performed a tap within the row with the active cell
                // and it wasn't actually a move, also only change selection on a tap
                // outside of the current selection, if it was longer than context menu the
                // handleContextMenuGesture will have changed this
                this.m_databodyMove = false;
                if (this._isSelectionEnabled() && duration < DvtDataGrid.CONTEXT_MENU_TAP_HOLD_DURATION)
                {
                    this.handleDatabodyClickSelection(event);
                    return;
                }
                else
                {
                    //activate on a tap
                    this.handleDatabodyClickActive(event);
                    return;
                }
            }

            if (this.m_databodyMove)
            {
                event.preventDefault();
                this.m_databodyMove = false;
                this._handleMoveMouseUp(event, true);
                return;
            }

            this._handleSwipe(event);
        }

    }

    this.handleTouchCancel(event);
};

/**
 * Calculate the momentum based on the distance and duration of the swipe
 * @param {number} current the current touch position
 * @param {number} start the start touch position
 * @param {number} time the duration of the swipe
 * @param {number} currentScroll the current scroll position
 * @param {number} maxScroll the maximum scroll position
 * @param {boolean=} rtl true if right to left, false if left to right, undefined if determining momentum in Y direction
 * @return {Object} an object with three keys:
 *                      destination - the point to scroll to with the momentum
 *                      overScroll - the pixel amount that is scrolled beyond the scrollable region
 *                      duration - the duration of the scroll to that destination
 * @private
 */
DvtDataGrid.prototype._calculateMomentum = function(current, start, time, currentScroll, maxScroll, rtl)
{
    var distance, speed, destination, duration, overScroll;

    distance = current - start;
    speed = Math.abs(distance) / time;
    destination = (speed * speed) / (2 * DvtDataGrid.DECELERATION_FACTOR) * (distance < 0 ? -1 : 1);
    duration = speed / DvtDataGrid.DECELERATION_FACTOR;

    if (rtl)
    {
        destination = destination * -1;
    }

    // if the distance overshoots, then we'll have to adjust and recalculate the duration
    if (currentScroll - destination > maxScroll)
    {
        // too far bottom/right
        overScroll = Math.max(DvtDataGrid.MAX_OVERSCROLL_PIXEL * -1, destination);
        destination = currentScroll - maxScroll;
        distance = maxScroll - currentScroll;
        duration = distance / speed;
    }
    else if (currentScroll - destination < 0)
    {
        // too far top/left
        overScroll = Math.min(DvtDataGrid.MAX_OVERSCROLL_PIXEL, destination);
        destination = currentScroll;
        distance = currentScroll;
        duration = distance / speed;
    }

    return {
        destination: Math.round(destination),
        // durations can be up to 4s currently let's cap them at 500ms
        duration: Math.min(Math.max(DvtDataGrid.MIN_SWIPE_TRANSITION_DURATION, duration), DvtDataGrid.MAX_SWIPE_TRANSITION_DURATION),
        overScroll: overScroll
    };
};

/**
 * Event handler for when touch is cancelled on the databody
 * @param {Event} event - touchcancel event on the databody
 */
DvtDataGrid.prototype.handleTouchCancel = function(event)
{
    if (this.m_databodyMove)
    {
        this._handleMoveMouseUp(event, false);
        this.m_databodyMove = false;
    }
    this.m_touchSelectAnchor = null;
    this.m_touchMultipleSelect = false;
    // reset the variables back to default values
    this.m_touchActive = false;
    this.m_startX = 0;
    this.m_startY = 0;
    this.m_prevX = 0;
    this.m_prevY = 0;
    this.m_currentX = 0;
    this.m_currentY = 0;
    this.m_startTime = 0;
};

/**
 * Event handler for when touch is started on the header
 * @param {Event} event - touchstart event on the header
 */
DvtDataGrid.prototype.handleHeaderTouchStart = function(event)
{
    var header, fingerCount, target;

    //store start time of touch
    this.m_touchStart = (new Date()).getTime();

    fingerCount = event.touches.length;
    target = /** @type {Element} */ (event.target);

    // move = one finger swipe (or two?)
    if (fingerCount == 1)
    {
        // get the coordinates of the touch
        this.m_startX = event.touches[0].pageX;
        this.m_startY = event.touches[0].pageY;

        // need these to detect whether touch is hold and move vs. swipe
        this.m_currentX = this.m_startX;
        this.m_currentY = this.m_startY;
        this.m_prevX = this.m_startX;
        this.m_prevY = this.m_startY;

        // flag it
        this.m_touchActive = true;
        header = this.findHeader(target);

        if (this.isResizeEnabled())
        {
            this.handleResize(event);
            this.handleResizeMouseDown(event);
        }

        //allow row reorder on headers if our move is enabled make sure our row has the active cell in it
        if (!this.m_isResizing && this._isMoveOnElementEnabled(header))
        {
            this.m_databodyMove = true;
        }
    }
    else
    {
        // more than one finger touched so cancel
        this.handleHeaderTouchCancel(event);
    }
};

/**
 * Event handler for when touch moves on the header
 * @param {Event} event - touchmove event on the header
 */
DvtDataGrid.prototype.handleHeaderTouchMove = function(event)
{
    var diffX, diffY, header, target, axis;

    if (this.m_touchActive)
    {
        event.preventDefault();

        this.m_currentX = event.touches[0].pageX;
        this.m_currentY = event.touches[0].pageY;

        diffX = this.m_currentX - this.m_prevX;
        diffY = this.m_currentY - this.m_prevY;

        if (this.m_isResizing && this.isResizeEnabled())
        {
            this.handleResize(event);
        }
        else if (this.m_databodyMove)
        {
            this._removeTouchSelectionAffordance();
            this._handleMove(event);
        }
        else
        {
            target = /** @type {Element} */ (event.target);
            // can't swipe column headers in Y and row headers in X
            header = this.findHeader(target);
            axis = this.getHeaderCellAxis(header);
            if (axis == 'column' || axis == 'columnEnd')
            {
                this._handleNonSwipeScroll(diffX, 0);
            }
            else
            {
                this._handleNonSwipeScroll(0, diffY);
            }
        }

        this.m_prevX = this.m_currentX;
        this.m_prevY = this.m_currentY;
    }
    else
    {
        this.handleTouchCancel(event);
    }
};

/**
 * Event handler for when touch ends on the header
 * @param {Event} event - touchend event on the header
 */
DvtDataGrid.prototype.handleHeaderTouchEnd = function(event)
{
    var touchEnd, touchLength, tapMax = DvtDataGrid.HEADER_TAP_SHORT_HOLD_DURATION, header, target, selectionMode, cellContext;

    //calculate the end of touch time for tap and hold
    touchEnd = (new Date()).getTime();
    touchLength = touchEnd - this.m_touchStart;

    if (this.m_touchActive && !event.defaultPrevented)
    {
        target = /** @type {Element} */ (event.target);
        // if resizing handle resize first so that we don't conflict and forget to end
        if (this.m_isResizing && this.isResizeEnabled())
        {
            this.handleResizeMouseUp(event);
            if (this.m_currentX != this.m_startX && this.m_currentY != this.m_startY)
            {
                event.preventDefault();
            }
        }
        // if a short tap select
        else if (this.m_currentX == this.m_startX && this.m_currentY == this.m_startY)
        {
            selectionMode = this.m_options.getSelectionMode();
            header = this.findHeader(target);
            cellContext = header[this.getResources().getMappedAttribute("context")];

            //if touch in an icon it the click event will handle mousedown/up
            if ((this.m_utils.containsCSSClassName(target, this.getMappedStyle('sortascending')) ||
                    this.m_utils.containsCSSClassName(target, this.getMappedStyle('sortdescending')))
                    && this._isDOMElementSortable(target))
            {
                event.preventDefault();
                this._removeTouchSelectionAffordance();
                this._handleSortIconMouseDown(target);
                this._handleHeaderSort(event);
            }
            // if click or right click we want to adjust the selction
            // no else so that we can select a cell in the same row as long as no drag
            // check if selection is enabled
            else if (this._isSelectionEnabled() && this.isMultipleSelection() && !(selectionMode == 'row' && cellContext['axis'].indexOf('row') == -1) && !this.m_databodyMove)
            {
                this.handleHeaderClickSelection(event);

                // only allow drag on left click
                if (event.button === 0)
                {
                    this.m_headerDragState = true;
                }
            }
            // for single row based selection only
            else if (selectionMode == 'row' && cellContext['axis'].indexOf('row') != -1 && this._isSelectionEnabled())
            {
                this.handleHeaderClickSelection(event);            
            }
            // if not selecting, just make active.
            else
            {
                this.handleHeaderClickActive(event);            
            }       
        }
        // if reordering a row
        else if (this.m_databodyMove)
        {
            event.preventDefault();
            this.m_databodyMove = false;
            this._handleMoveMouseUp(event, true);
        }
        //handle potential swipe
        else
        {
            header = this.findHeader(target);
            this._handleSwipe(event, this.getHeaderCellAxis(header));
        }
        //tap and long hold shows context menu, through the wrapper layer
    }
    this.handleHeaderTouchCancel(event);
};

/**
 * Event handler for when touch is cancelled on the header
 * @param {Event} event - touchcancel event on the header
 */
DvtDataGrid.prototype.handleHeaderTouchCancel = function(event)
{
    if (this.m_databodyMove)
    {
        this._handleMoveMouseUp(event, false);
        this.m_databodyMove = false;
    }
    // reset the variables back to default values
    this.m_touchActive = false;
    this.m_startX = 0;
    this.m_startY = 0;
    this.m_prevX = 0;
    this.m_prevY = 0;
    this.m_currentX = 0;
    this.m_currentY = 0;
};

/**
 * Handle a touch scroll that is a slow drag
 * @param {number} diffX
 * @param {number} diffY
 */
DvtDataGrid.prototype._handleNonSwipeScroll = function(diffX, diffY)
{
    var time = (new Date()).getTime();
    // for non-swipe scroll use 0ms to prevent jiggling
    this._disableTouchScrollAnimation();

    var diff = this.adjustTouchScroll(diffX, diffY);
    diffX = diff[0];
    diffY = diff[1];

    this.scrollDelta(diffX, diffY);

    // reset start position if this is a tap and scroll, so that we can handle
    // user doing a swipe at the end
    if (time - this.m_startTime > DvtDataGrid.TAP_AND_SCROLL_RESET)
    {
        this.m_startX = this.m_currentX;
        this.m_startY = this.m_currentY;
        this.m_startTime = (new Date()).getTime();
    }
};

/**
 * Event handler for when touch swipe may have been detected
 * @param {Event} event - touchcancel event on the header
 * @param {string|null=} axis - if a header the header axis so we don't swipe in the direction
 */
DvtDataGrid.prototype._handleSwipe = function(event, axis)
{
    var duration, rtl, diffX, diffY, momentumX, momentumY, transitionDuration;
    duration = (new Date()).getTime() - this.m_startTime;
    rtl = this.getResources().isRTLMode();
    diffX = this.m_currentX - this.m_startX;
    diffY = this.m_currentY - this.m_startY;
    // if right to left the difference is the opposite on swipe
    if (rtl)
    {
        diffX = diffX * -1;
    }
    // detect whether this is a swipe
    if (Math.abs(diffX) < DvtDataGrid.MIN_SWIPE_DISTANCE && Math.abs(diffY) < DvtDataGrid.MIN_SWIPE_DISTANCE && duration < DvtDataGrid.MIN_SWIPE_DURATION)
    {
        event.preventDefault();
        //center touch affordances if row selection multiple
        if (this._isSelectionEnabled())
        {
            this._scrollTouchSelectionAffordance();
        }
    }
    // swipe case
    else if (duration < DvtDataGrid.MAX_SWIPE_DURATION)
    {
        event.preventDefault();
        if (axis != 'row' && axis != 'rowEnd')
        {
            // calculate momentum
            momentumX = this._calculateMomentum(this.m_currentX, this.m_startX, duration, this.m_currentScrollLeft, this.m_scrollWidth, rtl);
            if (!isNaN(momentumX.overScroll))
            {
                // don't overscroll if there's more rows to fetch
                if (momentumX.overScroll > 0 || this.m_stopColumnFetch)
                {
                    this.m_extraScrollOverX = momentumX.overScroll * -1;
                }
            }
        }
        else
        {
            momentumX = {duration: 0, destination: 0};
            diffX = 0;
        }

        if (axis != 'column' && axis != 'columnEnd')
        {
            momentumY = this._calculateMomentum(this.m_currentY, this.m_startY, duration, this.m_currentScrollTop, this.m_scrollHeight);
            if (!isNaN(momentumY.overScroll))
            {
                // don't overscroll if there's more rows to fetch
                if (momentumY.overScroll > 0 || this.m_stopRowFetch)
                {
                    this.m_extraScrollOverY = momentumY.overScroll * -1;
                }
            }
        }
        else
        {
            momentumY = {duration: 0, destination: 0};
            diffY = 0;
        }

        transitionDuration = Math.max(momentumX.duration, momentumY.duration);
        this.m_databody['firstChild'].style.webkitTransitionDuration = transitionDuration + 'ms';
        this.m_rowHeader['firstChild'].style.webkitTransitionDuration = transitionDuration + 'ms';
        this.m_colHeader['firstChild'].style.webkitTransitionDuration = transitionDuration + 'ms';
        this.m_rowEndHeader['firstChild'].style.webkitTransitionDuration = transitionDuration + 'ms';
        this.m_colEndHeader['firstChild'].style.webkitTransitionDuration = transitionDuration + 'ms';

        diffX = diffX + momentumX.destination;
        diffY = diffY + momentumY.destination;
        var diff = this.adjustTouchScroll(diffX, diffY);
        diffX = diff[0];
        diffY = diff[1];

        this.scrollDelta(diffX, diffY);
    }
};

/************* end touch related methods ********************/

/**
 * Callback on a widget listener
 * @param {string} functionName - the function name to look up in the callbacks
 * @param {Object} details - the object to pass into the callback function
 * @return {boolean|undefined} true if event passes, false if vetoed
 */
DvtDataGrid.prototype.fireEvent = function(functionName, details)
{
    var callback;
    if (functionName == null || details == null)
    {
        return;
    }

    callback = this.callbacks[functionName];
    if (callback != null)
    {
        return callback(details);
    }
    return true;
};

/**
 * Add a callback function to the callbacks object
 * @param {string} functionName - the function name to callback on
 * @param {Object.<Function>} handler - the function to callback to
 */
DvtDataGrid.prototype.addListener = function(functionName, handler)
{
    this.callbacks[functionName] = handler;
};
/*********************************** end dom event handling ***************************************/

/**
 * Set the style height on an element in pixels
 * @param {Element} elem - the element to set height on
 * @param {number} height - the pixel height to set the element to
 */
DvtDataGrid.prototype.setElementHeight = function(elem, height)
{
    elem['style']['height'] = height + "px";
};

/**
 * Get a number of the style height of an element
 * @param {Element|undefined|null} elem - the element to get height on
 * @return {number} the style height of the element
 */
DvtDataGrid.prototype.getElementHeight = function(elem)
{
    return this.getElementDir(elem, 'height');
};

/**
 * Set the style width on an element in pixels
 * @param {Element} elem - the element to set width on
 * @param {number} width - the pixel width to set the element to
 */
DvtDataGrid.prototype.setElementWidth = function(elem, width)
{
    elem['style']['width'] = width + "px";
};

/**
 * Get a number of the style pixel width of an element
 * @param {Element|undefined|null} elem - the element to get width on
 * @return {number} the style width of the element
 */
DvtDataGrid.prototype.getElementWidth = function(elem)
{
    return this.getElementDir(elem, 'width');
};

/**
 * Set the style left/right/top/bottom on an element in pixels
 * @param {Element|undefined|null} elem - the element to set width on
 * @param {number} pix - the pixel width to set the element to
 * @param {string} dir - 'left','right','top,'bottom'
 * */
DvtDataGrid.prototype.setElementDir = function(elem, pix, dir)
{
    elem['style'][dir] = pix + "px";
};

/**
 * Get a number of the style left/right/top/bottom of an element
 * @param {Element|undefined|null} elem - the element to get style left/right/top/bottom on
 * @param {string} dir - 'left','right','top,'bottom'
 * @return {number} the style left/right/top/bottom of the element
 */
DvtDataGrid.prototype.getElementDir = function(elem, dir)
{
    var value;
    if (elem['style'][dir].indexOf('px') > -1 && elem['style'][dir].indexOf('e') == -1)
    {
        // parseFloat does better with big numbers
        return parseFloat(elem['style'][dir]);
    }

    if (!document.body.contains(elem))
    {
        elem['style']['visibility'] = "hidden";
        this.m_root.appendChild(elem); //@HTMLUpdateOK
        // Started using offset again because of how it handles large numbers and limits on BoundingClient
        // Note that on chrome and IE offsetHeight will round differently if the top value is at a decimal
        // pixel value greater or less than .5
        value = Math.round(elem['offset' + dir.charAt(0).toUpperCase() + dir.slice(1)]);
        this.m_root.removeChild(elem);
        elem['style']['visibility'] = "";
    }
    else
    {
        value = Math.round(elem['offset' + dir.charAt(0).toUpperCase() + dir.slice(1)]);
    }
    return value;
};

/************************* Model change event *****************************************/
/**
 * @private
 */
DvtDataGrid.BEFORE = 1;

/**
 * @private
 */
DvtDataGrid.AFTER = 2;

/**
 * @private
 */
DvtDataGrid.INSIDE = 3;

/**
 * Checks whether an index (row/column) is within the range of the current viewport.
 * @param {Object} indexes the row and column indexes
 * @return {number} BEFORE if the index is before the current viewport, AFTER if the index is after
 *         the current viewport, INSIDE if the index is within the current viewport
 * @private
 */
DvtDataGrid.prototype._isInViewport = function(indexes)
{
    var rowIndex, columnIndex;

    rowIndex = indexes['row'];
    columnIndex = indexes['column'];

    if (rowIndex === -1 && columnIndex === -1)
    {
        // actually, this is an invalid index... should throw an error?
        return -1;
    }

    // if row index wasn't specified, just verify the column range
    if (rowIndex === -1)
    {
        if (columnIndex < this.m_startCol)
        {
            return DvtDataGrid.BEFORE;
        }

        if (columnIndex > this.m_endCol)
        {
            return DvtDataGrid.AFTER;
        }

        // if it's not before or after, it must be inside
        return DvtDataGrid.INSIDE;
    }

    // if column index wasn't specified, just verify the row range
    if (columnIndex === -1)
    {
        if (rowIndex < this.m_startRow)
        {
            return DvtDataGrid.BEFORE;
        }

        if (rowIndex > this.m_endRow)
        {
            return DvtDataGrid.AFTER;
        }

        // if it's not before or after, it must be inside
        return DvtDataGrid.INSIDE;
    }

    // both row and column index are defined, then check both ranges
    if (columnIndex >= this.m_startCol && columnIndex <= this.m_endCol && rowIndex >= this.m_startRow && rowIndex <= this.m_endRow)
    {
        return DvtDataGrid.INSIDE;
    }

    // undefined
    return -1;
};

/**
 * Checks whether any part of the cell is in the viewport
 * @param {number} left
 * @param {number} right
 * @param {number} top
 * @param {number} bottom
 * @return {boolean} true if cell is in the viewport
 * @private
 */
DvtDataGrid.prototype._isCellBoundryInViewport = function(left, right, top, bottom)
{
    var viewportTop = this._getViewportTop();
    var viewportBottom = this._getViewportBottom();
    var viewportLeft = this._getViewportLeft();
    var viewportRight = this._getViewportRight();

    return ((bottom <= viewportBottom && bottom > viewportTop) || (top >= viewportTop && top < viewportBottom)) &&
                ((right <= viewportRight && right > viewportLeft) ||  (left >= viewportLeft && left < viewportRight));
};

/**
 * @param {Object} event the model event
 * @return {boolean} true if event is queued, false otherwise
 * @private
 */
DvtDataGrid.prototype.queueModelEvent = function(event)
{
    // in case if the model event arrives before the grid is fully rendered or the event arrives during processing
    // of model queue or we are in the middle of processing/animation model event, queue the event and handle it later
    if (!this.m_initialized || this.m_processingEventQueue || this.m_animating || this.m_processingModelEvent)
    {
        if (this.m_modelEvents == null)
        {
            this.m_modelEvents = [];
        }
        this.m_modelEvents.push(event);
        return true;
    }

    return false;
};

/**
 * Model event handler
 * @param {Object} event the model change event
 * @param {boolean} fromQueue whether this is invoked from model queue processing, optional
 * @protected
 */
DvtDataGrid.prototype.handleModelEvent = function(event, fromQueue)
{
    var operation, keys, cellSet, headerSet, endHeaderSet, indexes, source, silent, requiresAnimation;

    // in case if the model event arrives before the grid is fully rendered,
    // queue the event and handle it later
    if (fromQueue === undefined && this.queueModelEvent(event))
    {
        return;
    }

    operation = event['operation'];
    keys = event['keys'];
    source = event['source'];
    indexes = event['indexes'];
    cellSet = event['result'];
    headerSet = event['header'];
    endHeaderSet = event['endheader'];
    silent = event['silent'];
    requiresAnimation = false;

    this.m_processingModelEvent = true;

    if (operation === 'insert')
    {
        this._adjustActive(operation, indexes);
        this.m_shouldFocus = true;
        this._adjustSelectionOnModelChange(operation, keys, indexes);

        if (cellSet != null)
        {
            // range insert event with cellset returned
            this._handleModelInsertRangeEvent(cellSet, headerSet, endHeaderSet);
            requiresAnimation = true;
        }
        else
        {
            this._handleModelInsertEvent(indexes, keys);
        }
    }
    else if (operation === 'update')
    {
        this._handleModelUpdateEvent(indexes, keys);
        requiresAnimation = true;
    }
    else if (operation === 'delete')
    {
        // adjust selection if neccessary
        // do this before the rows in the databody is mutate
        // (easier this way because of animation delays, plus the selection is immediately updated
        // to reflect the updated state)
        this._adjustActive(operation, indexes);
        this._adjustSelectionOnModelChange(operation, keys, indexes);

        if (source && oj.FlattenedTreeDataGridDataSource && source instanceof oj.FlattenedTreeDataGridDataSource && this.m_utils.supportsTransitions())
        {
            this._handleModelDeleteEventWithAnimation(keys);
            requiresAnimation = true;
        }
        else
        {
            this._handleModelDeleteEvent(indexes, keys, silent);
        }
    }
    else if (operation === 'refresh' || operation === 'reset')
    {
        this._handleModelRefreshEvent();
    }
    else if (operation === 'sync')
    {
        this._handleModelSyncEvent(event);
    }

    this.m_processingModelEvent = false;

    //Need to rerun the queued events if 
    //coming from the queue. If no animation
    //was involved in the current event, 
    //we can directly call _runModelEventQueue
    //from here. Animation events will call
    //_runModelEventQueue at the end of their 
    //transition function.
    if (!requiresAnimation && fromQueue)
    {
        this._runModelEventQueue();
    }
};

/**
 * Adjust selection ranges if neccessary on insert or delete.
 * @param {string} operation the model event operation which triggers selection adjustment.
 * @param {Object} indexes the indexes that identify the rows that got inserted/deleted.
 * @private
 */
DvtDataGrid.prototype._adjustActive = function(operation, indexes)
{
    var activeRowIndex, i, rowIndex, activeHeader, adjustment = 0;

    if (this.m_active != null)
    {
        if(this.m_active['type'] == 'cell')
        {
            activeHeader = false;
            activeRowIndex = this.m_active['indexes']['row'];
        }
        else if (this.m_active['axis'] === 'row')
        {
            activeHeader = true;
            activeRowIndex = this.m_active['index'];
        }
        else
        {
            return;
        }
    }
    else
    {
        return;
    }

    if (!Array.isArray(indexes))
    {
        indexes = new Array(indexes);
    }

    //if we are getting this from a move event
    if (this.m_moveActive === true)
    {
        if (operation === 'insert')
        {
            if (!activeHeader)
            {
                this.m_active['indexes']['row'] = indexes[0]['row'];
            }
            else
            {
                this.m_active['index'] = indexes[0]['row'];
            }
            return;
        }
        else if (operation === 'delete' && indexes[0]['row'] === activeRowIndex)
        {
            // do not clear the active since we know the active should be the
            // same once the moved row is returned via insert
            return;
        }
    }

    adjustment = operation === 'insert' ? 1 : -1;

    for (i = 0; i < indexes.length; i++)
    {
        rowIndex = indexes[i]['row'];
        if (rowIndex < activeRowIndex)
        {
            if (!activeHeader)
            {
                this.m_active['indexes']['row'] += adjustment;
            }
            else
            {
                this.m_active['index'] += adjustment;
            }
        }
        else if (rowIndex === activeRowIndex && operation === 'delete')
        {
            this._setActive(null, null);
        }
    }
};

/**
 * Adjust selection ranges if neccessary on insert or delete.
 * @param {string} operation the model event operation which triggers selection adjustment.
 * @param {Object} keys the keys that identify the rows that got inserted/deleted.
 * @param {Object} indexes the indexes that identify the rows that got inserted/deleted.
 * @private
 */
DvtDataGrid.prototype._adjustSelectionOnModelChange = function(operation, keys, indexes)
{
    var selection, i, rowKey, rowIndex, j, range, startRowKey, endRowKey, startRowIndex, endRowIndex, newRowKey, adjustment, movedRow;

    // make it an array if it's a single entry event
    if (!Array.isArray(keys))
    {
        keys = new Array(keys);
    }

    if (!Array.isArray(indexes))
    {
        indexes = new Array(indexes);
    }

    selection = this.GetSelection();

    if (keys == null || indexes == null || keys.length != indexes.length || selection.length == 0)
    {
        // on a move reset the selection
        if (this.m_moveActive && operation == 'insert')
        {
            if (this._isSelectionEnabled() && this._isDatabodyCellActive())
            {
                if (this.m_options.getSelectionMode() == 'cell')
                {
                    movedRow = this.createRange(this.m_active['indexes'], this.m_active['indexes'], keys[0], keys[0]);
                }
                else
                {
                    movedRow = this.createRange(indexes[0], indexes[0], keys[0], keys[0]);
                }
                this.m_selectionFrontier = this.m_active['indexes'];
                selection.push(movedRow);
            }
            this.m_moveActive = false;
        }
        // we are done
        return;
    }

    adjustment = operation === 'insert' ? 1 : -1;

    for (i = 0; i < keys.length; i++)
    {
        rowKey = keys[i]['row'];
        rowIndex = indexes[i]['row'];

        // have to do this backwards since we'll be mutating the array at the same time
        j = selection.length;
        while (j--)
        {
            range = selection[j];
            startRowKey = range['startKey']['row'];
            endRowKey = range['endKey']['row'];
            startRowIndex = range['startIndex']['row'];
            endRowIndex = range['endIndex']['row'];

            if (startRowKey == rowKey)
            {
                if (endRowKey == rowKey)
                {
                    // single row in range, and it has been deleted, so remove from selection
                    if (operation == 'delete')
                    {
                        selection.splice(j, 1);
                        continue;
                    }
                }

                // adjust start key, index stays the same
                // adjust end index, end key stays the same
                // get the key of the next row, which will become the new start key
                newRowKey = this._getKey(this._getAxisCellsByIndex(range['startIndex']['row'] + 1, 'row')[0], 'row');
                range['startKey']['row'] = newRowKey;
                range['endIndex']['row'] += adjustment;
            }
            else if (endRowKey == rowKey)
            {
                // adjust end key and end index
                // get the key of the next row, which will become the new start key
                newRowKey = this._getKey(this._getAxisCellsByIndex(range['startIndex']['row'] - 1, 'row')[0], 'row');
                range['endKey']['row'] = newRowKey;
                range['endIndex']['row'] += adjustment;
            }
            else if (rowIndex <= startRowIndex)
            {
                // before start index, so adjust both start and end index
                range['startIndex']['row'] += adjustment;
                range['endIndex']['row'] += adjustment;
            }
            else if (rowIndex < endRowIndex)
            {
                // something in between start and end selection, adjust the end index
                range['endIndex']['row'] += adjustment;
            }
        }
    }
};

/**
 * Handles model insert event
 * @param {Object} indexes the indexes that identifies the row that got updated.
 * @param {Object} keys the key that identifies the row that got updated.
 * @private
 */
DvtDataGrid.prototype._handleModelInsertEvent = function(indexes, keys)
{
    var flag, row, rowHeader, rowEndHeader;
    // checks if the new row/column is in the viewport
    flag = this._isInViewport(indexes);
    //If the model inserted is just the next model fetch it
    if (flag === DvtDataGrid.INSIDE || (flag === DvtDataGrid.AFTER && indexes['row'] == (this.m_endRow + 1)))
    {
        // an insert can only be a insert new row or new column.  A cell insert is
        // automatically treated as row insert, keys['row'/'column'] can be the number 0
        if (keys['row'] != null)
        {
            //if we have added to an empty grid just refresh, so we can fetch all headers
            if (this._databodyEmpty())
            {
                this.empty();
                this.refresh(this.m_root);
            }
            else
            {
                // move all rows up an index
                this._modifyAxisCellContextIndex('row', indexes['row'], this.m_endRow - indexes['row'] + 1, 1);
                this._refreshDatabodyMap();

                this.fetchHeaders("row", indexes['row'], this.m_rowHeader, this.m_rowEndHeader, 1, {"success": this._handleHeaderInsertsFetchSuccess});
                this.fetchCells(this.m_databody, indexes['row'], this.m_startCol, 1, this.m_endCol - this.m_startCol + 1, {"success": this._handleCellInsertsFetchSuccess});
            }
        }
        //else if (keys['column'] != null)
        //{
            // todo: handle column insert
        //}
    }
    else
    {
        if (flag === DvtDataGrid.BEFORE)
        {
            // move all rows up an index
            this._modifyAxisCellContextIndex('row', 0, this.m_endRow + 1, 1);
            this._refreshDatabodyMap();

            this.m_startRow++;
            this.m_startRowHeader++;
            this.m_endRow++;
            this.m_endRowHeader++;
            this.m_startRowPixel += this.m_avgRowHeight;
            this.m_startRowHeaderPixel += this.m_avgRowHeight;
            this.m_endRowPixel += this.m_avgRowHeight;
            this.m_endRowHeaderPixel += this.m_avgRowHeight;
            row = this.m_databody['firstChild']['firstChild'];
            if (row != null)
            {
                this.pushRowsDown(row, this.m_avgRowHeight);
            }
            rowHeader = this.m_rowHeader['firstChild']['firstChild'];
            if (rowHeader != null)
            {
                this.pushRowsDown(rowHeader, this.m_avgRowHeight);
            }
            rowEndHeader = this.m_rowEndHeader['firstChild']['firstChild'];
            if (rowEndHeader != null)
            {
                this.pushRowsDown(rowEndHeader, this.m_avgRowHeight);
            }
        }

        this.scrollToIndex(indexes);
    }
};

/**
 * Handle a successful call to the data source fetchCells. Update the row and
 * cell DOM elements when necessary.
 * @param {Object} cellSet - a CellSet object which encapsulates the result set of cells
 * @param {Array.<Object>} cellRanges - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 */
DvtDataGrid.prototype._handleCellInsertsFetchSuccess = function(cellSet, cellRanges)
{
    var rowStart;

    // so that grid will be resize
    this.m_initialized = false;

    // insert the row
    this.handleCellsFetchSuccess(cellSet, cellRanges, this.m_endRow >= cellRanges[0]['start']);

    // make sure the new row is in range
    rowStart = cellRanges[0]['start'];
    this._scrollRowIntoViewport(rowStart);

    // clean up rows outside of viewport (for non-highwatermark scrolling only)
    if (!this._isHighWatermarkScrolling())
    {
        this._cleanupViewport('top');
    }
    this.updateRowBanding();
    this.m_stopRowFetch = false;
    if (this.m_endRowHeader != -1)
    {
        this.m_stopRowHeaderFetch = false;
    }
    if (this.m_endRowEndHeader != -1)
    {
        this.m_stopRowEndHeaderFetch = false;
    }
    // Need to fill viewport in the case of a silent delete of multiple records with an insert following.
    // i.e. a splice of the data which removes 2 models silently and adds 1 back in, need to add the last model to fill view
    this.fillViewport();
};

/**
 * Handle a successful call to the data source fetchHeaderss. Update the row header DOM elements when necessary.
 * @param {Object} headerSet - a HeaderSet object which encapsulates the result set of cells
 * @param {Object} headerRanges - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 */
DvtDataGrid.prototype._handleHeaderInsertsFetchSuccess = function(headerSet, headerRanges, endHeaderSet)
{
    // so that grid will be resize
    this.m_resizeRequired = true;
    // insert the row
    this.handleHeadersFetchSuccess(headerSet, headerRanges, endHeaderSet, this.m_endRowHeader >= headerRanges['start']);
};

/**
 * Scrolls the row with index into the viewport
 * @param {number} index the row index
 * @private
 */
DvtDataGrid.prototype._scrollRowIntoViewport = function(index)
{
    var absIndex, databodyContent, rowCells, viewportTop, viewportBottom, rowTop, diff;

    rowCells = this._getAxisCellsByIndex(index, 'row');
    if (rowCells == null)
    {
        // something is wrong the newly inserted row does not exists
        return;
    }

    viewportTop = this._getViewportTop();
    viewportBottom = this._getViewportBottom();

    rowTop = this.getElementDir(rowCells[0], 'top');
    diff = viewportTop - rowTop;
    if (diff > 0)
    {
        // row added to top, scroll up
        this.scrollDelta(0, diff);
    }
    else
    {
        diff = viewportBottom - rowTop;
        if (diff < 0)
        {
            // row added to bottom, scroll down
            this.scrollDelta(0, diff);
        }
    }
};

/**
 * Handles model range insert event
 * @param {Object} cellSet the range of cells inserted.
 * @param {Object=} headerSet the row headers.
 * @param {Object=} endHeaderSet the row end headers.
 * @private
 */
DvtDataGrid.prototype._handleModelInsertRangeEvent = function(cellSet, headerSet, endHeaderSet)
{
    var rowStart, rowCount, columnStart, columnCount, rowHeaderFragment, newCellElements,
            headerCount, c, index, totalRowHeight, returnVal, className, renderer, rowRange,
            columnRange, headerRange, rowEndHeaderFragment, headerEndCount;

    // reconstruct the cell ranges from result
    rowStart = cellSet.getStart("row");
    rowCount = cellSet.getCount("row");
    columnStart = cellSet.getStart("column");
    columnCount = cellSet.getCount("column");

    // if it should animate
    if (this.m_utils.supportsTransitions())
    {
        //create  a fragment with all of the row headers
        if (headerSet != null)
        {
            rowHeaderFragment = document.createDocumentFragment();
            headerCount = headerSet.getCount();
            // add the headers to the row header
            totalRowHeight = 0;
            c = 0;
            className = this.getMappedStyle("headercell") + " " + this.getMappedStyle("rowheadercell");
            renderer = this.m_options.getRenderer("row");
            while (headerCount - c > 0)
            {
                index = rowStart + c;
                returnVal = this.buildLevelHeaders(rowHeaderFragment, index, 0, 0, this.m_startRowPixel + totalRowHeight, true, rowStart != this.m_endRowHeader + 1, renderer, headerSet, 'row', className, this.m_rowHeaderLevelCount);
                c += returnVal['count'];
                totalRowHeight += returnVal.totalHeaderDimension;
            }
        }

        //create  a fragment with all of the row headers
        if (endHeaderSet != null)
        {
            rowEndHeaderFragment = document.createDocumentFragment();
            headerEndCount = endHeaderSet.getCount();
            // add the headers to the row header
            totalRowHeight = 0;
            c = 0;
            className = this.getMappedStyle("endheadercell") + " " + this.getMappedStyle("rowendheadercell");
            renderer = this.m_options.getRenderer("rowEnd");
            while (headerEndCount - c > 0)
            {
                index = rowStart + c;
                returnVal = this.buildLevelHeaders(rowEndHeaderFragment, index, 0, 0, this.m_startRowPixel + totalRowHeight, true, rowStart != this.m_endRowEndHeader + 1, renderer, endHeaderSet, 'rowEnd', className, this.m_rowEndHeaderLevelCount);
                c += returnVal['count'];
                totalRowHeight += returnVal.totalHeaderDimension;
            }
        }

        newCellElements = document.createDocumentFragment();
        returnVal = this._addCellsToFragment(newCellElements, cellSet, rowStart, this.m_startRowPixel, columnStart, this.m_startColPixel);
        if (newCellElements.childNodes.length === 0 && (rowHeaderFragment == null || rowHeaderFragment.childNodes.length === 0) && (rowEndHeaderFragment == null || rowEndHeaderFragment.childNodes.length === 0))
        {
            return;
        }
        this._insertRowsWithAnimation(newCellElements, rowHeaderFragment, rowEndHeaderFragment, rowStart, cellSet.getCount('row'), returnVal['totalRowHeight']);
    }
    else
    {
        rowRange = {"axis": "row", "start": rowStart, "count": rowCount};
        columnRange = {"axis": "column", "start": columnStart, "count": columnCount};
        if (headerSet != null)
        {
            headerRange = {'axis': 'row', 'header': this.m_rowHeader, 'endHeader':this.m_rowEndHeader, 'start': rowStart, 'count': headerSet.getCount()};
            this.m_fetching['row'] = headerRange;
            this._handleHeaderInsertsFetchSuccess(headerSet, headerRange, endHeaderSet);
        }

        this._modifyAxisCellContextIndex('row', rowStart, this.m_endRow - rowStart + 1, rowCount);
        this._refreshDatabodyMap();

        // insert the rows
        this._handleCellInsertsFetchSuccess(cellSet, [rowRange, columnRange]);
    }
};

/**
 * Handles model update event
 * @param {Object} indexes the indexes that identifies the row that got updated.
 * @param {Object} keys the key that identifies the row that got updated.
 * @private
 */
DvtDataGrid.prototype._handleModelUpdateEvent = function(indexes, keys)
{
    var flag;

    // if the new row/column is in the viewport
    flag = this._isInViewport(indexes);
    if (flag === DvtDataGrid.INSIDE)
    {
        //if there is a row header update it
        if (this.m_endRowHeader != -1)
        {
            // fetch the updated row header and row
            this.fetchHeaders("row", indexes['row'], this.m_rowHeader, this.m_rowEndHeader, 1, {
                "success": this._handleHeaderUpdatesFetchSuccess,
                "error": this.handleHeadersFetchError
            });
        }

        this.fetchCells(this.m_databody, indexes['row'], this.m_startCol, 1, this.m_endCol - this.m_startCol + 1, {
            "success": this._handleCellUpdatesFetchSuccess,
            "error": this.handleCellsFetchError
        });
    }

    // if it's not in range then do nothing
};

/**
 * Handle a successful call to the data source fetchHeaderss. Update the row header DOM elements when necessary.
 * @param {Object} headerSet - a HeaderSet object which encapsulates the result set of cells
 * @param {Array} headerRange - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 * @param {Object} endHeaderSet - a HeaderSet object which encapsulates the result set of cells
 * @private
 */
DvtDataGrid.prototype._handleHeaderUpdatesFetchSuccess = function(headerSet, headerRange, endHeaderSet)
{
    var axis, rowStart, row;

    axis = headerRange["axis"];
    this.m_fetching[axis] = false;
    rowStart = headerRange["start"];

    this._replaceHeaders(this.buildRowHeaders.bind(this), headerSet, this.m_rowHeader, rowStart, this.m_startRowHeader);
    this._replaceHeaders(this.buildRowEndHeaders.bind(this), endHeaderSet, this.m_rowEndHeader, rowStart, this.m_startRowEndHeader);

    row = this.m_rowHeader.firstChild.childNodes[rowStart - this.m_startRowHeader];

    if (this.m_active != null && this.m_active['type'] === 'header' && (this.m_active['axis'] === 'row' || this.m_active['axis'] === 'rowEnd') && this._getKey(row, 'row') === this.m_active['key'])
    {
        this._highlightActive();
    }
    // end fetch
    this._signalTaskEnd();
    // should animate the fragment in the future like updateCells
};

/**
 * Replace the headers on update
 * @param {Function} buildFunction
 * @param {Object|null|undefined} headerSet
 * @param {Element} root
 * @param {number} index
 * @param {number} start
 * @private
 */
DvtDataGrid.prototype._replaceHeaders = function(buildFunction, headerSet, root, index, start)
{
    var headerContent, fragment, row;
    if (headerSet != null)
    {
        fragment = buildFunction(root, headerSet, index, 1, true, true);
        headerContent = root['firstChild'];
        row = headerContent.childNodes[index - start];
        headerContent.replaceChild(fragment, row); //@HTMLUpdateOK
    }
};

/**
 * Handle a successful call to the data source fetchCells. Update the row and
 * cell DOM elements when necessary.
 * @param {Object} cellSet - a CellSet object which encapsulates the result set of cells
 * @param {Array.<Object>} cellRange - [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 * @private
 */
DvtDataGrid.prototype._handleCellUpdatesFetchSuccess = function(cellSet, cellRange)
{
    var rowStart, renderer, columnBandingInterval, rowBandingInterval;

    //fetch complete
    this.m_fetching['cells'] = false;

    rowStart = cellRange[0]['start'];

    renderer = this.m_options.getRenderer("cell");
    columnBandingInterval = this.m_options.getColumnBandingInterval();
    rowBandingInterval = this.m_options.getRowBandingInterval();

    // update the cells in the row
    this._updateCellsInRow(cellSet, rowStart, renderer, this.m_startCol, columnBandingInterval);

    // end fetch
    this._signalTaskEnd();
};

/**
 * Retrieves the type of update animation to use.
 * @return {number} the type of update animation.  See constants.
 * @private
 */
DvtDataGrid.prototype._getUpdateAnimation = function()
{
    return DvtDataGrid.UPDATE_ANIMATION_SLIDE_INOUT;
};

/**
 * Retrieves the update animation duration.
 * @return {number} the animation duration.
 * @private
 */
DvtDataGrid.prototype._getUpdateAnimationDuration = function()
{
    return DvtDataGrid.UPDATE_ANIMATION_DURATION;
};

/**
 * Adds cells to a row. Iterate over the cells passed in, create new div elements
 * for them settign appropriate styles, and append or prepend them to the row based on the start column.
 * @param {Object} cellSet - the result set of cell data
 * @param {number} rowIndex - the index of the row element
 * @param {function(Object)} renderer - the cell renderer
 * @param {number} columnStart - the index to start start adding at
 * @param {number} columnBandingInterval - the column banding interval
 * @private
 */
DvtDataGrid.prototype._updateCellsInRow = function(cellSet, rowIndex, renderer, columnStart, columnBandingInterval)
{
    var animationDuration, self, width, listener, cells, fragment, top, i;

    animationDuration = this._getUpdateAnimationDuration();

    cells = this._getAxisCellsByIndex(rowIndex, 'row');
    top = this.getElementDir(cells[0], 'top');

    // check whether animation should be used
    if (animationDuration === 0 || !this.m_utils.supportsTransitions())
    {
        // clear the content of the row first
        this._removeFromArray(cells);

        fragment = document.createDocumentFragment();
        this._addCellsToFragment(fragment, cellSet, rowIndex, top, columnStart, this.m_startColPixel);
        this._populateDatabody(this.m_databody['firstChild'], fragment);

        // re-apply selection and active cell since content changed
        if (this._isSelectionEnabled())
        {
            this.applySelection();
        }
        this._highlightActive();

        // hide fetching text now that we are done
        this.hideStatusText();
    }
    else
    {
        self = this;
        // animation start
        self._signalTaskStart();

        // clear the content of the row and refill it with new data
        this._removeFromArray(cells);

        fragment = document.createDocumentFragment();
        this._addCellsToFragment(fragment, cellSet, rowIndex, top, columnStart, this.m_startColPixel);
        cells = fragment.childNodes;

        //hide the row
        width = this.getElementWidth(this.m_databody);
        for (i = 0; i < cells.length; i++)
        {
            this.addTransformMoveStyle(cells[i], 0, 0, 'linear', width, 0, 0);
        }

        this._populateDatabody(this.m_databody['firstChild'], fragment);

        cells = this._getAxisCellsByIndex(rowIndex, 'row');

        // hide fetching text now that we are done
        this.hideStatusText();

        listener = function()
        {
            for (i = 0; i < cells.length; i++)
            {
                self.removeTransformMoveStyle(cells[i]);
            }

            // re-apply selection and active cell since content changed
            if (self._isSelectionEnabled())
            {
                self.applySelection();
            }
            self._highlightActive();
            
            // end animation
            self._signalTaskEnd();

            //runQueue if applicable
            self._runModelEventQueue();
        };
        
        this._onTransitionEnd(cells[cells.length -1], listener, animationDuration);

        setTimeout(function()
        {
            //kick off animation
            for (i = 0; i < cells.length; i++)
            {
                self.addTransformMoveStyle(cells[i], animationDuration + "ms", 0, 'linear', 0, 0, 0);
            }        
        }, 0);
    }
};


/**
 * Handles model delete event
 * @param {Array|Object} indexes the indexes that identifies the row that got deleted.
 * @param {Array|Object} keys the key that identifies the row that got deleted.
 * @param {boolean} silent true if the datagrid should not fill the databody
 * @private
 */
DvtDataGrid.prototype._handleModelDeleteEvent = function(indexes, keys, silent)
{
    var key, i, rowKey, row, height, referenceRow, databodyContent, beforeRowsHeight, insideRowsHeight, afterRowsHeight,
            databodyContentHeight, rowHeader, flag, index, beforeRowsDeleted, insideRowsDeleted, totalHeight, scrollerContent,
            rowEndHeader, cells;

    // make it an array if it's a single entry event
    if (!Array.isArray(keys))
    {
        keys = new Array(keys);
        indexes = new Array(indexes);
    }

    beforeRowsHeight = 0;
    insideRowsHeight = 0;
    afterRowsHeight = 0;
    beforeRowsDeleted = 0;
    insideRowsDeleted = 0;
    for (i = 0; i < keys.length; i++)
    {
        key = keys[i];
        index = indexes[i];
        if (key['row'] != null)
        {
            height = 0;
            rowKey = key['row'];
            flag = this._isInViewport(index);
            if (flag === DvtDataGrid.BEFORE)
            {
                //should only happen in virtual scrolling

                // adjust the cell context index
                this._modifyAxisCellContextIndex('row', this.m_startRow, this.m_endRow - this.m_startRow + 1, -1);

                beforeRowsDeleted++;
                beforeRowsHeight += this.m_avgRowHeight;
                this.m_startRowPixel -= this.m_avgRowHeight;
                this.m_endRowPixel -= this.m_avgRowHeight;
                if (this.m_endRowHeader != -1)
                {
                    this.m_startRowHeaderPixel -= this.m_avgRowHeight;
                    this.m_endRowHeaderPixel -= this.m_avgRowHeight;
                }
                cells = this._getAxisCellsByKey(rowKey, 'row');
                if (cells != null)
                {
                    this.pushRowsUp(this.m_startRow, this.m_avgRowHeight);
                }
                rowHeader = this.m_rowHeader['firstChild']['firstChild'];
                if (rowHeader != null)
                {
                    this.pushRowHeadersUp(rowHeader, this.m_avgRowHeight);
                }
                rowEndHeader = this.m_rowEndHeader['firstChild']['firstChild'];
                if (rowEndHeader != null)
                {
                    this.pushRowHeadersUp(rowEndHeader, this.m_avgRowHeight);
                }
            }
            else if (flag === DvtDataGrid.INSIDE)
            {
                insideRowsDeleted++;

                cells = this._getAxisCellsByKey(rowKey, 'row');
                if (cells != null)
                {
                    height = this.calculateRowHeight(cells[0]);

                    index = cells[0][this.getResources().getMappedAttribute('context')]['indexes']['row'];
                    this._modifyAxisCellContextIndex('row', index + 1, this.m_endRow - index, -1);

                    this._removeFromArray(cells);
                    this.pushRowsUp(index, height);
                    this.m_endRowPixel -= height;
                }
                rowHeader = this._findHeaderByKey(rowKey, this.m_rowHeader, this.getMappedStyle('rowheadercell'));
                if (rowHeader != null)
                {
                    height = this.calculateRowHeaderHeight(rowHeader);
                    referenceRow = rowHeader['nextSibling'];
                    this._remove(rowHeader);
                    this.pushRowHeadersUp(referenceRow, height);
                    this.m_endRowHeaderPixel -= height;
                }
                rowEndHeader = this._findHeaderByKey(rowKey, this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'));
                if (rowEndHeader != null)
                {
                    height = this.calculateRowHeaderHeight(rowEndHeader);
                    referenceRow = rowEndHeader['nextSibling'];
                    this._remove(rowEndHeader);
                    this.pushRowHeadersUp(referenceRow, height);
                    this.m_endRowEndHeaderPixel -= height;
                }
                insideRowsHeight = insideRowsHeight + height;
            }
            else //flag === DvtDataGrid.AFTER
            {
                //only include after rows if virtual scroll
                if (this.m_options.getScrollPolicy() === 'scroll')
                {
                    afterRowsHeight += this.m_avgRowHeight;
                }
            }
        }
    }

    this._refreshDatabodyMap();

    this.m_startRow -= beforeRowsDeleted;
    this.m_endRow = this.m_endRow - beforeRowsDeleted - insideRowsDeleted;
    if (this.m_endRowHeader != -1)
    {
        this.m_startRowHeader -= beforeRowsDeleted;
        this.m_endRowHeader = this.m_endRowHeader - beforeRowsDeleted - insideRowsDeleted;
    }
    if (this.m_endRowEndHeader != -1)
    {
        this.m_startRowEndHeader -= beforeRowsDeleted;
        this.m_endRowEndHeader = this.m_endRowEndHeader - beforeRowsDeleted - insideRowsDeleted;
    }
    totalHeight = beforeRowsHeight + insideRowsHeight + afterRowsHeight;

    // adjust the databody height
    databodyContent = this.m_databody['firstChild'];
    databodyContentHeight = this.getElementHeight(databodyContent) - totalHeight;
    this.setElementHeight(databodyContent, databodyContentHeight);
    this.resizeGrid();

    if (!silent && this.m_moveActive != true)
    {
        // so that grid will be resize
        this.m_resizeRequired = true;
        // check viewport to see if we need to fetch because of deleted row causing empty spaces
        this.m_stopRowFetch = false;
        if (this.m_endRowHeader != -1)
        {
            this.m_stopRowHeaderFetch = false;
        }
        if (this.m_endRowEndHeader != -1)
        {
            this.m_stopRowEndHeaderFetch = false;
        }
        this.fillViewport();
    }
    this.updateRowBanding();
};

/**
 * Handles model delete event with animation
 * @param {Array} keys the key that identifies the row that got deleted.
 * @private
 */
DvtDataGrid.prototype._handleModelDeleteEventWithAnimation = function(keys)
{
    this._collapseRowsWithAnimation(keys);
};

/**
 * Helper function to calculate gaps in the selection if any.
 * @param {Array.<number>} indices indices that identifies rows that got deleted.
 * @return {Array.<Array.<number>>} idxs.
 * @private
 */
DvtDataGrid.prototype._getSelectionGaps = function(indices)
{
    var i, idx, idxs, first;

    idx = [];
    idxs = [];
    first = true;

    for (i = 0; i < indices.length - 1; i++)
    {
        if (indices[i + 1] - indices[i] == 1)
        {
            idx.push(indices[i]);
            first = false;
        }
        else
        {
            if (first)
            {
                idx.push(indices[i]);
            }
            else
            {
                idx.push(indices[i]);
            }
            idxs.push(idx);
            idx = [];
            first = true;
        }
    }
    idx.push(indices[indices.length - 1]);
    idxs.push(idx);

    return idxs;
};

/**
 * Helper method to process animated rows in responce on the model delete event
 * @param {Object} keys set of keys that identifies rows that got deleted.
 * @private
 */
DvtDataGrid.prototype._collapseRowsWithAnimation = function(keys)
{
    var self, duration, databodyContent, referenceCellsIndex, referenceCells, referenceRowHeader, referenceRowEndHeader, lastAnimationElement,rowCells,
            i, j, rowKey, row, rowsToRemove, rowHeadersToRemove, rowEndHeadersToRemove, totalRowHeight, rowHeader, rowEndHeader,
            transitionListener, referenceCellsTop, referenceRowHeaderTop, rowHeaderSupport, referenceRowEndHeaderTop, rowEndHeaderSupport;

    if (keys.length == 0)
    {
        return;
    }

    self = this;
    // animation start
    self._signalTaskStart();
    // note we set the duration to 1 instead of 0 because some browsers do not invoke transition end listener if duration is 0
    duration = this.m_processingEventQueue ? 1 : DvtDataGrid.COLLAPSE_ANIMATION_DURATION;
    rowsToRemove = [];
    totalRowHeight = 0;
    rowHeaderSupport = this.m_endRowHeader == -1 ? false : true;
    rowEndHeaderSupport = this.m_endRowEndHeader == -1 ? false : true;
    databodyContent = this.m_databody['firstChild'];

    referenceCellsIndex = this._getIndex(this._getAxisCellsByKey(keys[0]['row'], 'row')[0] ,'row') - 1;
    referenceCells = this._getAxisCellsByIndex(referenceCellsIndex, 'row');
    referenceCellsTop = this.getElementDir(referenceCells[0], 'top');

    //all inherited animated rows should be hidden under previous rows in view
    for (i = referenceCellsIndex; i >= this.m_startRow; i--)
    {
        rowCells = this._getAxisCellsByIndex(i, 'row');
        if (this.getElementDir(rowCells[0], 'top') + this.getElementHeight(rowCells[0]) < this.m_currentScrollTop)
        {
            break;
        }

        for (j = 0; j < rowCells.length; j++)
        {
            this.changeStyleProperty(rowCells[j], this.getCssSupport('z-index'), 10);
        }
    }

    if (rowHeaderSupport)
    {
        rowHeadersToRemove = [];
        referenceRowHeader = this._findHeaderByKey(keys[0]['row'], this.m_rowHeader, this.getMappedStyle('rowheadercell'))['previousSibling'];
        referenceRowHeaderTop = this.getElementDir(referenceRowHeader, 'top');
        row = referenceRowHeader;
        while (row)
        {
            if (this.getElementDir(row, 'top') + this.getElementHeight(row) < this.m_currentScrollTop)
            {
                break;
            }
            this.changeStyleProperty(row, this.getCssSupport('z-index'), 10);
            row = row['previousSibling'];
        }
    }

    if (rowEndHeaderSupport)
    {
        rowEndHeadersToRemove = [];
        referenceRowEndHeader = this._findHeaderByKey(keys[0]['row'], this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'))['previousSibling'];
        referenceRowEndHeaderTop = this.getElementDir(referenceRowEndHeader, 'top');
        row = referenceRowEndHeader;
        while (row)
        {
            if (this.getElementDir(row, 'top') + this.getElementHeight(row) < this.m_currentScrollTop)
            {
                break;
            }
            this.changeStyleProperty(row, this.getCssSupport('z-index'), 10);
            row = row['previousSibling'];
        }
    }

    // get the rows we need to remove and set the new top to align row bottom with
    // the reference row bottom, but keep it where it is for the time being
    for (i = 0; i < keys.length; i++)
    {
        rowKey = keys[i]['row'];
        rowCells = this._getAxisCellsByKey(rowKey, 'row');
        if (rowCells != null)
        {
            rowsToRemove.push(rowCells);
            totalRowHeight += this.getElementHeight(rowCells[0]);
            for (j = 0; j < rowCells.length; j++)
            {
                this.setElementDir(rowCells[j], this.getElementDir(rowCells[j], 'top') - totalRowHeight, 'top');
                this.addTransformMoveStyle(rowCells[j], 0, 0, 'linear', 0, totalRowHeight, 0);
            }
        }
        if (rowHeaderSupport)
        {
            rowHeader = this._findHeaderByKey(rowKey, this.m_rowHeader, this.getMappedStyle('rowheadercell'));
            if (rowHeader != null)
            {
                rowHeadersToRemove.push(rowHeader);
                this.setElementDir(rowHeader, this.getElementDir(rowHeader, 'top') - totalRowHeight, 'top');
                this.addTransformMoveStyle(rowHeader, 0, 0, 'linear', 0, totalRowHeight, 0);
            }
        }
        if (rowEndHeaderSupport)
        {
            rowEndHeader = this._findHeaderByKey(rowKey, this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'));
            if (rowEndHeader != null)
            {
                rowEndHeadersToRemove.push(rowEndHeader);
                this.setElementDir(rowEndHeader, this.getElementDir(rowEndHeader, 'top') - totalRowHeight, 'top');
                this.addTransformMoveStyle(rowEndHeader, 0, 0, 'linear', 0, totalRowHeight, 0);
            }
        }
    }

    // for all the rows after the collapse change the top values appropriately
    for (i = referenceCellsIndex + keys.length + 1; i <= this.m_endRow; i++)
    {
        // change the row top but keep it where it is
        rowCells = this._getAxisCellsByIndex(i, 'row');
        for (j = 0; j < rowCells.length; j++)
        {
            this.setElementDir(rowCells[j], this.getElementDir(rowCells[j], 'top') - totalRowHeight, 'top');
            this.addTransformMoveStyle(rowCells[j], 0, 0, 'linear', 0, totalRowHeight, 0);
        }
        if (rowHeaderSupport)
        {
            rowHeader = rowHeader['nextSibling'];
            this.setElementDir(rowHeader, this.getElementDir(rowHeader, 'top') - totalRowHeight, 'top');
            this.addTransformMoveStyle(rowHeader, 0, 0, 'linear', 0, totalRowHeight, 0);
        }
        if (rowEndHeaderSupport)
        {
            rowEndHeader = rowEndHeader['nextSibling'];
            this.setElementDir(rowEndHeader, this.getElementDir(rowEndHeader, 'top') - totalRowHeight, 'top');
            this.addTransformMoveStyle(rowEndHeader, 0, 0, 'linear', 0, totalRowHeight, 0);
        }
    }

    // listen to the last rows transition end
    lastAnimationElement = this._getCellByIndex(this.createIndex(this.m_endRow, this.m_endCol));
    transitionListener = function()
    {
        var i, j;
        self._modifyAxisCellContextIndex('row', referenceCellsIndex + keys.length + 1, self.m_endRow - (referenceCellsIndex + keys.length), -1 * keys.length);

        for (i = 0; i < rowsToRemove.length; i++)
        {
            for (j = 0; j < rowsToRemove[i].length; j++)
            {
                self._remove(rowsToRemove[i][j]);
            }
            if (rowHeaderSupport)
            {
                self._remove(rowHeadersToRemove[i]);
            }
            if (rowEndHeaderSupport)
            {
                self._remove(rowEndHeadersToRemove[i]);
            }
        }

        self._refreshDatabodyMap();

        // clean up the variables no longer need because event animation handling
        self.m_endRow -= rowsToRemove.length;
        self.m_endRowPixel -= totalRowHeight;
        self.m_stopRowFetch = false;
        if (rowHeaderSupport)
        {
            self.m_endRowHeader -= rowHeadersToRemove.length;
            self.m_endRowHeaderPixel -= totalRowHeight;
            self.m_stopRowHeaderFetch = false;
        }
        if (rowEndHeaderSupport)
        {
            self.m_endRowEndHeader -= rowHeadersToRemove.length;
            self.m_endRowEndHeaderPixel -= totalRowHeight;
            self.m_stopRowEndHeaderFetch = false;
        }

        self.setElementHeight(databodyContent, self.m_endRowPixel - self.m_startRowPixel);
        self.resizeGrid();
        self.updateRowBanding();
        self.fillViewport();
        self._handleAnimationEnd();
    };
    
    self._onTransitionEnd(lastAnimationElement, transitionListener, duration);         

    // animate all rows
    this.m_animating = true;

    setTimeout(function() {
        for (i = referenceCellsIndex + 1; i <= self.m_endRow; i++)
        {
            // change the row top but keep it where it is
            rowCells = self._getAxisCellsByIndex(i, 'row');
            for (j = 0; j < rowCells.length; j++)
            {
                self.addTransformMoveStyle(rowCells[j], duration + "ms", 0, 'ease-out', 0, 0, 0);
            }
            if (rowHeaderSupport)
            {
                rowHeader = self._getHeaderByIndex(i, 0, self.m_rowHeader, self.m_rowHeaderLevelCount, self.m_startRowHeader);
                self.addTransformMoveStyle(rowHeader, duration + "ms", 0, 'ease-out', 0, 0, 0);
            }
            if (rowEndHeaderSupport)
            {
                rowEndHeader = self._getHeaderByIndex(i, 0, self.m_rowEndHeader, self.m_rowEndHeaderLevelCount, self.m_startRowEndHeader);
                self.addTransformMoveStyle(rowEndHeader, duration + "ms", 0, 'ease-out', 0, 0, 0);
            }
        }
    }, 0);
};

/**
 * Clean up the datagrid animations by resetting transform vars and z-index
 * @private
 */
DvtDataGrid.prototype._handleAnimationEnd = function()
{
    var i, databodyContent, rowHeaderContent, rowEndHeaderContent;
    // cleanRows
    databodyContent = this.m_databody['firstChild'];
    rowHeaderContent = this.m_rowHeader['firstChild'];
    rowEndHeaderContent = this.m_rowEndHeader['firstChild'];
    for (i = 0; i < databodyContent.childNodes.length; i++)
    {
        this.removeTransformMoveStyle(databodyContent.childNodes[i]);
        this.changeStyleProperty(databodyContent.childNodes[i], this.getCssSupport('z-index'), null, 'remove');
    }

    if (this.m_endRowHeader != -1)
    {
        for (i = 0; i < rowHeaderContent.childNodes.length; i++)
        {
            this.removeTransformMoveStyle(rowHeaderContent.childNodes[i]);
            this.changeStyleProperty(rowHeaderContent.childNodes[i], this.getCssSupport('z-index'), null, 'remove');
        }
    }

    if (this.m_endRowEndHeader != -1)
    {
        for (i = 0; i < rowEndHeaderContent.childNodes.length; i++)
        {
            this.removeTransformMoveStyle(rowEndHeaderContent.childNodes[i]);
            this.changeStyleProperty(rowEndHeaderContent.childNodes[i], this.getCssSupport('z-index'), null, 'remove');
        }
    }
    // end animation
    this.m_animating = false;

    // check event queue for outstanding model events
    this._runModelEventQueue();

    // if we signal ready before emptying the queue there may be outstanding events to immediately follow
    this._signalTaskEnd();
};

/**
 * Get a cell or header from a given key and axis, gets the first cell with that key
 * @private
 */
DvtDataGrid.prototype._getCellOrHeaderByKey = function(key, axis)
{
    var element = null;
    var cells = this._getAxisCellsByKey(key, axis, true);
    if (cells != null && cells.length > 0)
    {
        element = cells[0];
    }
    
    if (element == null)
    {
        if (axis === 'row')
        {
            element = this._findHeaderByKey(key, this.m_rowHeader, this.getMappedStyle('rowheadercell'));
            if (element == null)
            {
                element = this._findHeaderByKey(key, this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'));
            }
        }
        else if (axis === 'column')
        {
            element = this._findHeaderByKey(key, this.m_colHeader, this.getMappedStyle('colheadercell'));
            if (element == null)
            {
                element = this._findHeaderByKey(key, this.m_colEndHeader, this.getMappedStyle('colendheadercell'));
            }
        }
    }    
    return element;
};

/**
 * Find the header element by key inside a given root and className
 * @param {string|null} key the key
 * @param {DocumentFragment|Element} root
 * @param {string} className
 * @return {Element|null} the row element
 * @private
 */
DvtDataGrid.prototype._findHeaderByKey = function(key, root, className)
{
    var headers, header, i, headerKey;

    if (root == null)
    {
        return null;
    }

    //getElementsByClassName support is IE9 and up
    if (root.getElementsByClassName)
    {
        headers = root.getElementsByClassName(className);
    }
    else
    {
        headers = root.childNodes;
    }
    for (i = 0; i < headers.length; i++)
    {
        header = headers[i];
        headerKey = this._getKey(header);
        if (this._shallowThenDeepCompare(headerKey, key))
        {
            return header;
        }
    }

    // can't find it, the row is not in viewport
    return null;
};

/**
 * Handles model refresh event
 * @private
 */
DvtDataGrid.prototype._handleModelRefreshEvent = function()
{
    var visibility = this.getVisibility();

    // if we are visible, make sure we are visible, and just refresh the datagrid
    // if we are hidden we want to change the state to refresh so the wrapper know to call refresh when we are shown.
    // if we are already in state refresh we do not need to update.
    // if we are in state render we do not want to update that.
    if (visibility === DvtDataGrid.VISIBILITY_STATE_VISIBLE)
    {
        this.empty();
        // if the app developer doesn't notify the grid that it has become hidden
        // check to make sure, if it isn't hidden, refresh if it is
        // supported in IE9+
        if (this.m_root.offsetParent != null)
        {
            this.refresh(this.m_root);
        }
        else
        {
            this.setVisibility(DvtDataGrid.VISIBILITY_STATE_REFRESH);
        }

    }
    else if (visibility === DvtDataGrid.VISIBILITY_STATE_HIDDEN)
    {
        this.empty();
        this.setVisibility(DvtDataGrid.VISIBILITY_STATE_REFRESH);
    }
};

/**
 * Handles data source fetch end (model sync) event
 * @param {Object} event the model event
 * @private
 */
DvtDataGrid.prototype._handleModelSyncEvent = function(event)
{
    var startRow, pageSize, startRowPixel, startCol, startColPixel;
    //Currently these are set to zero for now, may come from the event later
    startRow = 0;
    startRowPixel = 0;
    startCol = 0;
    startColPixel = 0;
    pageSize = event['pageSize'];

    //cancel previous fetch calls
    this.m_fetching = {};

    // reset ranges
    this.m_startRow = startRow;
    this.m_endRow = -1;
    this.m_startRowHeader = startRow;
    this.m_endRowHeader = -1;
    this.m_startRowEndHeader = startRow;
    this.m_endRowEndHeader = -1;
    this.m_startRowPixel = startRowPixel;
    this.m_endRowPixel = startRowPixel;
    this.m_startRowHeaderPixel = startRowPixel;
    this.m_endRowHeaderPixel = startRowPixel;
    this.m_startRowEndHeaderPixel = startRowPixel;
    this.m_endRowEndHeaderPixel = startRowPixel;
    this.m_startCol = startCol;
    this.m_endCol = -1;

    this.m_startColHeader = startCol;
    this.m_endColHeader = -1;
    this.m_startColEndHeader = startCol;
    this.m_endColEndHeader = -1;

    this.m_startColPixel = startColPixel;
    this.m_endColPixel = startColPixel;

    this.m_startColHeaderPixel = startColPixel;
    this.m_endColHeaderPixel = startColPixel;

    this.m_startColEndHeaderPixel = startColPixel;
    this.m_endColEndHeaderPixel = startColPixel;

    this.m_rowHeaderLevelCount = undefined;
    this.m_columnHeaderLevelCount = undefined;
    this.m_rowEndHeaderLevelCount = undefined;
    this.m_columnEndHeaderLevelCount = undefined;

    this.m_avgRowHeight = undefined;
    this.m_avgColWidth = undefined;

    this.m_isEstimateRowCount = undefined;
    this.m_isEstimateColumnCount = undefined;
    this.m_stopRowFetch = false;
    this.m_stopRowHeaderFetch = false;
    this.m_stopRowEndHeaderFetch = false;
    this.m_stopColumnFetch = false;
    this.m_stopColumnHeaderFetch = false;
    this.m_stopColumnEndHeaderFetch = false;

    this._clearScrollPositionKeys();

    //clear selections
    this.m_selection = null;
    this.m_active = null;
    this.m_prevActive = null;
    this.m_trueIndex = {};

    if (this.m_empty != null)
    {
        this.m_root.removeChild(this.m_empty);
        this.m_empty = null;
    }

    this._showHeader(this.m_rowHeader);
    this._showHeader(this.m_colHeader);
    this._showHeader(this.m_rowEndHeader);
    this._showHeader(this.m_colEndHeader);

    this.m_initialized = false;
    this.fetchHeaders("row", startRow, this.m_rowHeader, this.m_rowEndHeader, pageSize, {'success': function(headerSet, headerRange, endHeaderSet)
        {
            this.handleRowHeadersFetchSuccessForLongScroll(headerSet, headerRange, endHeaderSet);
        }});
    this.fetchHeaders("column", startCol, this.m_colHeader, this.m_colEndHeader, undefined, {'success': function(headerSet, headerRange, endHeaderSet)
        {
            this.handleColumnHeadersFetchSuccessForLongScroll(headerSet, headerRange, endHeaderSet);
        }});

    this.fetchCells(this.m_databody, startRow, startCol, pageSize, null, {'success': function(cellSet, cellRange)
        {
            this.handleCellsFetchSuccessForLongScroll(cellSet, cellRange);
        }});
};

/************************************ active cell navigation ******************************/
/**
 * Sets the active cell by index
 * @param {Object} index row and column index
 * @param {Event=} event the DOM event causing the active cell change
 * @param {boolean=} clearSelection true if we should clear the selection on active change
 * @param {boolean=} silent
 * @param {boolean=} shouldNotScroll
 * @private
 * @return {boolean} true if active was changed, false if not
 */
DvtDataGrid.prototype._setActiveByIndex = function(index, event, clearSelection, silent, shouldNotScroll)
{
    return this._setActive(this._getCellByIndex(index), {'type':'cell', 'indexes': index}, event, clearSelection, silent, shouldNotScroll);
};

/**
 * Updates the active cell based on external set, do not fire events
 * @param {Object} activeObject set by application could be sparse
 * @param {boolean} shouldFocus
 * @param {boolean=} shouldNotScroll
 * @private
 */
DvtDataGrid.prototype._updateActive = function(activeObject, shouldFocus, shouldNotScroll)
{
    //the activeObject is potentially sparse, try to get an element from it
    var level, newActiveElement;
    if (activeObject == null)
    {
        this._setActive(null, null, null, true, false, shouldNotScroll);
    }
    else if (activeObject['keys'] != null)
    {
       newActiveElement = this._getCellByKeys(activeObject['keys']);
    }
    else if (activeObject['indexes'] != null)
    {
        newActiveElement = this._getCellByIndex(activeObject['indexes']);
    }
    else if (activeObject['axis'] != null)
    {
        level = activeObject['level'] == null ? 0 : activeObject['level'];
        if (activeObject['axis'] == 'column')
        {
            if (activeObject['key'] != null)
            {
                newActiveElement = this._findHeaderByKey(activeObject['key'], this.m_colHeader, this.getMappedStyle('colheadercell'));
            }
            else if (activeObject['index'] != null)
            {
                newActiveElement = this._getHeaderByIndex(activeObject['index'], level, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
            }
        }
        else if (activeObject['axis'] == 'row')
        {
            if (activeObject['key'] != null)
            {
                newActiveElement = this._findHeaderByKey(activeObject['key'], this.m_rowHeader, this.getMappedStyle('rowheadercell'));
            }
            else if (activeObject['index'] != null)
            {
                newActiveElement = this._getHeaderByIndex(activeObject['index'], level, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
            }
        }
        else if (activeObject['axis'] == 'columnEnd')
        {
            if (activeObject['key'] != null)
            {
                newActiveElement = this._findHeaderByKey(activeObject['key'], this.m_colEndHeader, this.getMappedStyle('colendheadercell'));
            }
            else if (activeObject['index'] != null)
            {
                newActiveElement = this._getHeaderByIndex(activeObject['index'], level, this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
            }
        }
        else if (activeObject['axis'] == 'rowEnd')
        {
            if (activeObject['key'] != null)
            {
                newActiveElement = this._findHeaderByKey(activeObject['key'], this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'));
            }
            else if (activeObject['index'] != null)
            {
                newActiveElement = this._getHeaderByIndex(activeObject['index'], level, this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
            }
        }
    }

    if (newActiveElement != null)
    {
        if (!shouldFocus)
        {
            this.m_shouldFocus = false;
        }
        this._setActive(newActiveElement, activeObject, null, true, false, shouldNotScroll);
    }
};

/**
 * Sets the active cell or header by element
 * @param {Element|null} element to set active to
 * @param {Event|null=} event the DOM event causing the active cell change
 * @param {boolean|null=} clearSelection true if we should clear the selection on active change
 * @param {boolean|null=} silent true if we should not fire events
 * @param {boolean|null=} shouldNotScroll true if we should not scroll before setting active (in case it came froma  scroll event, prevent loop)
 * @returns {boolean} true if active was changed, false if not
 */
DvtDataGrid.prototype._setActive = function(element, cellInfo, event, clearSelection, silent, shouldNotScroll)
{
    if (cellInfo != null && !shouldNotScroll)
    {
        this._scrollToActive(cellInfo);
    }

    if (element != null)
    {
        var active = this._createActiveObject(element);
        // see if the active cell is actually changing
        if (this._compareActive(active, this.m_active))
        {
            // fire vetoable beforeCurrentCell event
            if (silent || this._fireBeforeCurrentCellEvent(active, this.m_active, event))
            {
                this.m_prevActive = this.m_active;
                this.m_active = active;
                
                if (event && event.type === 'mousedown')
                {
                    this.m_trueIndex = null;
                }
                
                if (clearSelection && this._isSelectionEnabled())
                {
                    this._clearSelection(event);
                }
                this._unhighlightActiveObject(this.m_prevActive);
                this._highlightActiveObject(this.m_active, this.m_prevActive);
                this._manageMoveCursor();
                if (this._isGridEditable())
                {
                    if (active['type'] === 'cell')
                    {
                        this._setEditableClone(element);
                    }
                    this._updateEdgeCellBorders('');
                }
                if (!silent)
                {
                    this._fireCurrentCellEvent(active, event);
                }
                return true;
            }
        }
        else
        {
            // still wanted to make sure the cell was highlighted even if focus hasn't actually changed
            this._highlightActive();
        }
    }
    else if (!this.m_scrollIndexAfterFetch && !this.m_scrollHeaderAfterFetch)
    {
        if (silent || this._fireBeforeCurrentCellEvent(active, this.m_active, event))
        {
            this.m_prevActive = this.m_active;
            this.m_active = null;
            this._unhighlightActiveObject(this.m_prevActive);
            if (!silent)
            {
                this._fireCurrentCellEvent(active, event);
            }
        }
        return true;
    }
    return false;
};

/**
 * Create an active object from an element active object contains:
 * For header: type, axis, index, key, level
 * For cell: indexes, keys
 * @param {Element} element - the element to create an active object from
 * @return {Object} an active object
 */
DvtDataGrid.prototype._createActiveObject = function(element)
{
    var context = element[this.getResources().getMappedAttribute('context')];
    if (this.m_utils.containsCSSClassName(element, this.getMappedStyle('headercell')) ||
        this.m_utils.containsCSSClassName(element, this.getMappedStyle('endheadercell')))
    {
        return {
            'type': 'header',
            'axis': context['axis'],
            'index': this.getHeaderCellIndex(element),
            'key': context['key'],
            'level': context['level']
        };
    }
    else
    {
        return {
            'type': 'cell',
            'indexes': {
                'row': context['indexes']['row'],
                'column': context['indexes']['column'],
            },
            'keys': {
                'row': context['keys']['row'],
                'column': context['keys']['column']
            },
            'extents': {
                'row': context['extents']['row'],
                'column': context['extents']['column']
            }
        };
    }
};

/**
 * Retrieve the active element.
 * @return {Element|null} the active cell or header cell
 * @private
 */
DvtDataGrid.prototype._getActiveElement = function()
{
    return this._getElementFromActiveObject(this.m_active);
};

/**
 * Retrieve the element based on an active object.
 * @param {Object} active the object to get the element of
 * @return {Element|null} the active cell or header cell
 * @private
 */
DvtDataGrid.prototype._getElementFromActiveObject = function(active)
{
    var elements;
    if (active != null)
    {
        if (active['type'] == 'header')
        {
            if (active['axis'] === 'row')
            {
                return this._findHeaderByKey(active['key'], this.m_rowHeader, this.getMappedStyle('rowheadercell'));
            }
            else if (active['axis'] === 'column')
            {
                return this._findHeaderByKey(active['key'], this.m_colHeader, this.getMappedStyle('colheadercell'));
            }
            else if (active['axis'] === 'rowEnd')
            {
                return this._findHeaderByKey(active['key'], this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'));
        }
            else if (active['axis'] === 'columnEnd')
            {
                return this._findHeaderByKey(active['key'], this.m_colEndHeader, this.getMappedStyle('colendheadercell'));
            }
        }
        else
        {
            return this._getCellByIndex(active['indexes']);
        }
    }
    return null;
};

/**
 * Compare two active objects to see if they are equal
 * @param {Object} active1 an active object
 * @param {Object} active2 a comparison active object
 * @return {boolean} true if not equal
 */
DvtDataGrid.prototype._compareActive = function(active1, active2)
{
    if (active1 == null && active2 == null)
    {
        return false;
    }
    else if ((active1 == null && active2 != null) || (active1 != null && active2 == null))
    {
        return true;
    }
    else if (active1['type'] == active2['type'])
    {
        if (active1['type'] == 'header')
        {
            if (active1['index'] != active2['index'] ||
                    active1['key'] != active2['key'] ||
                    active1['axis'] != active2['axis'] ||
                    active1['level'] != active2['level'])
            {
                return true;
            }
        }
        else
        {
            if (active1['indexes']['row'] != active2['indexes']['row'] ||
                    active1['indexes']['column'] != active2['indexes']['column'] ||
                    active1['keys']['row'] != active2['keys']['row'] ||
                    active1['keys']['column']!= active2['keys']['column'])
            {
                return true;
            }
        }
    }
    else
    {
        return true;
    }
    return false;
};

/**
 * Fires an event before the current cell changes
 * @param {Object|undefined} newActive the new active information
 * @param {Object} oldActive the new active information
 * @param {Event|undefined} event the DOM event
 * @private
 * @return {boolean|undefined} true if event should continue
 */
DvtDataGrid.prototype._fireBeforeCurrentCellEvent = function(newActive, oldActive, event)
{
    // the event contains the context info
    var details =
    {
        'event': event,
        'ui':
        {
            'currentCell': newActive,
            'previousCurrentCell': oldActive
        }
    };

    return this.fireEvent('beforeCurrentCell', details);
};

/**
 * Fires an event to tell the datagrid to update the currentCell option
 * @param {Object|undefined} active the new active information
 * @param {Event|undefined} event the DOM event
 * @private
 */
DvtDataGrid.prototype._fireCurrentCellEvent = function(active, event)
{
    // the event contains the context info
    var details =
    {
        'event': event,
        'ui': active
    };

    return this.fireEvent('currentCell', details);
};

/**
 * Is the databody cell active
 * @return {boolean} true if active element is a cell
 * @private
 */
DvtDataGrid.prototype._isDatabodyCellActive = function()
{
    return (this.m_active != null && this.m_active['type'] == 'cell');
};

/**
 * Update the context info based on active changess
 * @param {Object} activeObject
 * @param {Object} prevActiveObject
 */
DvtDataGrid.prototype._updateActiveContext = function(activeObject, prevActiveObject)
{
    var axis, index, level, contextObj, skip;
    if (activeObject['type'] === 'header')
    {
        axis = activeObject['axis'];
        index = activeObject['index'];
        level = activeObject['level'];

        contextObj = {};
        if (activeObject['axis'] === 'row')
        {
            if (this.m_rowHeaderLevelCount > 1)
            {
                if (prevActiveObject == null ? true : !(level === prevActiveObject['level'] && axis === prevActiveObject['axis']))
                {
                    contextObj['level'] = level;
                }
            }
            if (prevActiveObject == null ? true : !(index === prevActiveObject['index'] && axis === prevActiveObject['axis']))
            {
                contextObj['rowHeader'] = index;
            }
        }
        else if (axis === 'column')
        {
            if (this.m_columnHeaderLevelCount > 1)
            {
                if (prevActiveObject == null ? true : !(level === prevActiveObject['level'] && axis === prevActiveObject['axis']))
                {
                    contextObj['level'] = level;
                }
            }
            if (prevActiveObject == null ? true : !(index === prevActiveObject['index'] && axis === prevActiveObject['axis']))
            {
                contextObj['columnHeader'] = index;
            }
        }
        else if (activeObject['axis'] === 'rowEnd')
        {
            if (this.m_rowEndHeaderLevelCount > 1)
            {
                if (prevActiveObject == null ? true : !(level === prevActiveObject['level'] && axis === prevActiveObject['axis']))
                {
                    contextObj['level'] = level;
                }
            }
            if (prevActiveObject == null ? true : !(index === prevActiveObject['index'] && axis === prevActiveObject['axis']))
            {
                contextObj['rowEndHeader'] = index;
            }
        }
        else if (axis === 'columnEnd')
        {
            if (this.m_columnEndHeaderLevelCount > 1)
            {
                if (prevActiveObject == null ? true : !(level === prevActiveObject['level'] && axis === prevActiveObject['axis']))
                {
                    contextObj['level'] = level;
                }
            }
            if (prevActiveObject == null ? true : !(index === prevActiveObject['index'] && axis === prevActiveObject['axis']))
            {
                contextObj['columnEndHeader'] = index;
            }
        }
        // update context info
        this._updateContextInfo(contextObj, skip);
    }
    else
    {
        // check whether the prev and current active cell is in the same row/column so that we can
        // skip row/column header info in aria-labelledby (to make the description more brief)
        if (prevActiveObject != null && prevActiveObject['type'] == 'cell' && activeObject != null && !this.m_externalFocus)
        {
           if (activeObject['indexes']['row'] === prevActiveObject['indexes']['row'])
           {
               skip = "row";
           }
           else if (activeObject['indexes']['column'] === prevActiveObject['indexes']['column'])
           {
               skip = "column";
           }
        }
        // update context info
        this._updateContextInfo(activeObject, skip);
    }
};

/**
 * Handles click to make a cell active
 * @param {Event} event
 * @private
 */
DvtDataGrid.prototype.handleDatabodyClickActive = function(event)
{
    var cell, target;
    target = /** @type {Element} */ (event.target);
    cell = this.findCell(target);
    if (cell != null)
    {
        this._setActive(cell, this._createActiveObject(cell), event);
    }
};

/**
 * Handles click to select a header
 * @param {Event} event
 */
DvtDataGrid.prototype.handleHeaderClickActive = function(event)
{
    var target, header;
    target = /** @type {Element} */ (event.target);
    header = this.findHeader(target);
    if (header != null)
    {
        this._clearSelection(event);
        this._setActive(header, this._createActiveObject(header), event);
    }
};

/**
 * Scroll to the active object
 * @param {Object} activeObject
 */
DvtDataGrid.prototype._scrollToActive = function(activeObject)
{
    if (activeObject['type'] === 'header')
    {
        this.scrollToHeader(activeObject);
    }
    else
    {
        this.scrollToIndex(activeObject['indexes']);
    }
};

/**
 * Retrieve cell by keys, this is rarely used, more common to look up cells by index within the data grid
 * @param {Object} keys
 * @return {Element|null} the active cell
 * @private
 */
DvtDataGrid.prototype._getCellByKeys = function(keys)
{
    var databodyContent, cells, cell, i, rowKey, columnKey;
    if (keys == null || keys['row'] == null || keys['column'] == null || 
        this.m_databody == null || this.m_databody['firstChild'] == null)
    {
        return null;
    }

    databodyContent = this.m_databody['firstChild'];
    cells = databodyContent.getElementsByClassName(this.getMappedStyle('cell'));
    for (i = 0; i < cells.length; i++)
    {
        cell = cells[i];
        rowKey = this._getKey(cell, 'row');
        if (this._shallowThenDeepCompare(rowKey, keys['row']))
        {
            columnKey = this._getKey(cell, 'column');
            if (this._shallowThenDeepCompare(columnKey, keys['column']))
            {
                return cell;
            }
        }
    }

    return null;
};

/**
 * @param {*} value1
 * @param {*} value2
 * @return {boolean}
 * @private
 */
DvtDataGrid.prototype._shallowThenDeepCompare = function(value1, value2)
{
    if (value1 === value2)
    {
        return true;
    }
    return this._compareValuesCallback(value1, value2);
};

/**
 * Retrieve the indexes of a cell
 * @param {Element} cell
 * @return {Object}
 */
DvtDataGrid.prototype.getCellIndexes = function(cell)
{
    var cellContext = cell[this.getResources().getMappedAttribute("context")];
    return this.createIndex(cellContext['indexes']['row'], cellContext['indexes']['column']);
};

/**
 * Retrieve the extents of a cell
 * @param {Element} cell
 * @return {Object}
 */
DvtDataGrid.prototype.getCellExtents = function(cell)
{
    var cellContext = cell[this.getResources().getMappedAttribute("context")];
    return this.createIndex(cellContext['extents']['row'], cellContext['extents']['column']);
};


/**
 * Retrieve the end indexes of a cell (index + extent - 1)
 * @param {Element} cell
 * @return {Object}
 */
DvtDataGrid.prototype.getCellEndIndexes = function(cell)
{
    var cellIndexes = this.getCellIndexes(cell);
    var cellExtents = this.getCellExtents(cell);
    return this.createIndex(cellIndexes['row'] + cellExtents['row'] - 1, cellIndexes['column'] + cellExtents['column'] - 1);
};

/**
 * Retrieve the index of a cell/header along a given axis
 * @param {Element|undefined|null} element
 * @param {string=} axis
 * @return {number|null}
 */
DvtDataGrid.prototype._getIndex = function(element, axis)
{
    if (element != null)
    {
        if (axis != null && this.m_utils.containsCSSClassName(element, this.getMappedStyle('cell')))
        {
            var cellContext = element[this.getResources().getMappedAttribute("context")];
            if (axis == 'row')
            {
                return cellContext['indexes']['row']
            }
            if (axis == 'column')
            {
                return cellContext['indexes']['column']
            }
        }
        else
        {
            return this.getHeaderCellIndex(element);
        }
    }
    return null;
};

/**
 * Retrieve the extent of a cell along a given axis
 * @param {Element|undefined|null} element
 * @param {string} axis
 * @return {number|null}
 */
DvtDataGrid.prototype._getExtent = function(element, axis)
{
    if (element != null)
    {
        if (axis != null && this.m_utils.containsCSSClassName(element, this.getMappedStyle('cell')))
        {
            var cellContext = element[this.getResources().getMappedAttribute("context")];
            if (axis == 'row')
            {
                return cellContext['extents']['row']
            }
            if (axis == 'column')
            {
                return cellContext['extents']['column']
            }
        }
        else
        {
            return parseInt(this._getAttribute(element, 'extent', true), 10);
        }
    }
    return null;
};

/**
 * Retrieve the index of a header cell
 * @param {Element} header header cell element
 * @return {number}
 */
DvtDataGrid.prototype.getHeaderCellIndex = function(header)
{
    var axis, levelCount, start, index;
    axis = this.getHeaderCellAxis(header);
    switch (axis)
    {
        case 'column':
            levelCount = this.m_columnHeaderLevelCount;
            start = this.m_startColHeader;
            break;
        case 'row':
            levelCount = this.m_rowHeaderLevelCount;
            start = this.m_startRowHeader;
            break;
        case 'columnEnd':
            levelCount = this.m_columnEndHeaderLevelCount;
            start = this.m_startColEndHeader;
            break;
        case 'rowEnd':
            levelCount = this.m_rowEndHeaderLevelCount;
            start = this.m_startRowEndHeader;
            break;
        default:
            return -1;
    }

    // if there are multiple levels on the row header
    if (levelCount > 1)
    {
        // get the groupingContainer's start value and set thtat to the index
        index = /** @type {number} */ (this._getAttribute(header['parentNode'], 'start', true));
        //if this is the groupingContainer's first child rturn that value
        if (header === header['parentNode']['firstChild'])
        {
            return index;
        }
        //decrement the index by one for the first header element at the level above it
        index--;
    }
    else
    {
        index = start;
    }

    while (header['previousSibling'])
    {
        index += 1;
        header = header['previousSibling'];
    }

    return index;
};

/**
 * Retrieve the axis of a header cell
 * @param {Element|undefined|null} header header cell element
 * @return {string|null} row or column
 */
DvtDataGrid.prototype.getHeaderCellAxis = function(header)
{
    if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('colheadercell')))
    {
        return 'column';
    }
    else if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('rowheadercell')))
    {
        return 'row';
    }
    else if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('rowendheadercell')))
    {
        return 'rowEnd';
    }
    else if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('colendheadercell')))
    {
        return 'columnEnd';
    }
    return null;
};

/**
 * Retrieve the level of a header cell
 * @param {Element} header header cell element
 * @return {number|string} row or column
 */
DvtDataGrid.prototype.getHeaderCellLevel = function(header)
{
    var level;
    if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('colheadercell')))
    {
        if (this.m_columnHeaderLevelCount === 1)
        {
            return 0;
        }

    }
    else if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('rowheadercell')))
    {
        if (this.m_rowHeaderLevelCount === 1)
        {
            return 0;
        }
    }
    else if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('colendheadercell')))
    {
        if (this.m_columnEndHeaderLevelCount === 1)
        {
            return 0;
        }
    }
    else if (this.m_utils.containsCSSClassName(header, this.getMappedStyle('rowendheadercell')))
    {
        if (this.m_rowEndHeaderLevelCount === 1)
        {
            return 0;
        }
    }

    level = this._getAttribute(header['parentNode'], 'level', true);
    if (header === header['parentNode']['firstChild'])
    {
        return level;
    }
    // plus one case is if we are on the innermost level the headers do not have their own
    // grouping containers so if it is the first child it is the level of the grouping container
    // but all subsequent children are the next level in
    return level + this.getHeaderCellDepth(header['parentNode']['firstChild']);
};

/**
 * Retrieve the depth of a header cell
 * @param {Element} header header cell element
 * @return {string|number|null} row or column depth
 */
DvtDataGrid.prototype.getHeaderCellDepth = function(header)
{
    return this._getAttribute(header, 'depth', true);
};

/**
 * Find the cell or header element (recursively if needed)
 * @private
 * @param {Element} elem
 * @return {Element|undefined|null}
 */
DvtDataGrid.prototype.findCellOrHeader = function(elem)
{
    var cell = this.findCell(elem);
    if (cell == null)
    {
        cell = this.findHeader(elem);
    }
    return cell;
};

/**
 * Find the cell element (recursively if needed)
 * @private
 * @param {Element} elem
 * @return {Element|undefined|null}
 */
DvtDataGrid.prototype.findCell = function(elem)
{
    return this.find(elem, "cell");
};

/**
 * Find the cell element (recursively if needed)
 * @param {Element|undefined|null} elem
 * @param {string} key
 * @param {string=} className
 * @return {Element|undefined|null}
 */
DvtDataGrid.prototype.find = function(elem, key, className)
{
    // if element is null or if we reach the root of DataGrid
    if (elem == null || elem == this.getRootElement())
    {
        return null;
    }

    // recursively walk up the element and find the class name that matches the cell class name
    if (className == undefined)
    {
        className = this.getMappedStyle(key);
    }

    if (className == null)
    {
        return null;
    }

    // if the element contains the cell class name, then it's a cell, otherwise go up
    if (this.m_utils.containsCSSClassName(elem, className))
    {
        return elem;
    }
    return this.find(elem['parentNode'], key, className);
};

/**
 * Highlight the current active element
 * @param {Array=} classNames string of classNames to add to active element
 * @private
 */
DvtDataGrid.prototype._highlightActive = function(classNames)
{
    this._highlightActiveObject(this.m_active, this.m_prevActive, classNames);
};


/**
 * Unhighlight the current active element
 * @param {Array=} classNames string of classNames to remove from active element
 * @private
 */
DvtDataGrid.prototype._unhighlightActive = function(classNames)
{
    this._unhighlightActiveObject(this.m_active, classNames);
};

/**
 * Highlight the specified object
 * @param {Object} activeObject active to unhighlight
 * @param {Object} prevActiveObject last active to base aria properties on
 * @param {Array=} classNames string of classNames to add to active element
 * @private
 */
DvtDataGrid.prototype._highlightActiveObject = function(activeObject, prevActiveObject, classNames)
{
    if (classNames == null && this.m_utils.shouldOffsetOutline())
    {
        classNames = ['offsetOutline'];
    }
    if (activeObject != null)
    {
        var element = this._getElementFromActiveObject(activeObject);
        //possible in the virtual case
        if (element != null)
        {
            this.m_focusInHandler(element);
            if (classNames != null)
            {
                this._highlightElement(element, classNames);
            }
            this._setAriaProperties(activeObject, prevActiveObject, element);
        }
    }
};

/**
 * Unhighlight the specified object
 * @param {Object} activeObject to unhighlight
 * @param {Array=} classNames string of classNames to remove from active element
 * @private
 */
DvtDataGrid.prototype._unhighlightActiveObject = function(activeObject, classNames)
{
    if (classNames == null && this.m_utils.shouldOffsetOutline())
    {
        classNames = ['offsetOutline'];
    }
    if (activeObject != null)
    {
        var element = this._getElementFromActiveObject(activeObject);
        if (element != null)
        {
            this.m_focusOutHandler(element);
            if (classNames != null)
            {
                this._unhighlightElement(element, classNames);
            }
            this._unsetAriaProperties(element);
        }
    }
};

/**
 * Highlight all the cells in a row
 * @param {number|string|null} value key/index
 * @param {string} axis
 * @param {string} indexOrKey
 * @param {string} addOrRemove
 * @param {Array} classNames
 */
DvtDataGrid.prototype._highlightCellsAlongAxis = function(value, axis, indexOrKey, addOrRemove, classNames)
{
    var i, cells;

    cells = indexOrKey == 'key' ? this._getAxisCellsByKey( /** @type {string} */ (value), axis) : this._getAxisCellsByIndex( /** @type {number} */ (value), axis);

    for (i = 0; i < cells.length; i++)
    {
        addOrRemove == 'add' ? this._highlightElement(cells[i], classNames) : this._unhighlightElement(cells[i], classNames);
    }
};

/**
 * Highlight an element adding classes in the provided array
 * @param {Element} element
 * @param {Array} classNames
 */
DvtDataGrid.prototype._highlightElement = function(element, classNames)
{
    var className, i;
    for (i = 0; i < classNames.length; i++)
    {
        className = this.getMappedStyle(classNames[i]);
        this.m_utils.addCSSClassName(element, className);
    }
};

/**
 * Unhighlight an element removing classes in the provided array
 * @param {Element} element
 * @param {Array} classNames
 */
DvtDataGrid.prototype._unhighlightElement = function(element, classNames)
{
    var className, i;
    for (i = 0; i < classNames.length; i++)
    {
        className = this.getMappedStyle(classNames[i]);
        this.m_utils.removeCSSClassName(element, className);
    }
};

/**
 * Reset all wai-aria properties on a cell or header.
 * @param {Object} activeObject active to unhighlight
 * @param {Object} prevActiveObject last active to base aria properties on
 * @param {Element} element the element to reset all wai-aria properties
 * @private
 */
DvtDataGrid.prototype._setAriaProperties = function(activeObject, prevActiveObject, element)
{
    var label;
    label = this.getLabelledBy(activeObject, prevActiveObject, element);
    this._updateActiveContext(activeObject, prevActiveObject);

    element.setAttribute("tabIndex", 0);
    element.setAttribute("aria-labelledby", label);

    // check to see if we should focus on the cell later
    if ((this.m_cellToFocus == null || this.m_cellToFocus != element) && this.m_shouldFocus !== false)
    {
        element.focus();
    }
    this.m_shouldFocus = null;
};

/**
 * Reset all wai-aria properties on a cell or header.
 * @param {Element} element the element to reset all wai-aria properties.
 */
DvtDataGrid.prototype._unsetAriaProperties = function(element)
{
    if (element != null)
    {
        // reset focus index
        element.setAttribute("tabIndex", -1);
        // remove aria related attributes
        element.removeAttribute("aria-labelledby");
    }
};

/**
 * Returns the wai-aria labelled by property for a cell
 * @param {Object} activeObject
 * @param {Object} prevActiveObject
 * @param {Element} element
 * @return {string} the wai-aria labelled by property for the cell
 */
DvtDataGrid.prototype.getLabelledBy = function(activeObject, prevActiveObject, element)
{
    var label, previousElement, direction, key, previousRowIndex, previousColumnIndex, row, rowEnd, column, columnEnd;
    label = "";

    if (activeObject['type'] == 'header')
    {
        // get the previous active header to compare what the screen reader needs to read for parent Ids,
        // should only need this if multi level header
        if (prevActiveObject != null && prevActiveObject['type'] == 'header' && !this.m_externalFocus)
        {
            if (prevActiveObject['axis'] === 'row' && this.m_rowHeaderLevelCount > 1)
            {
                previousElement = this._getHeaderByIndex(prevActiveObject['index'], prevActiveObject['level'], this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
            }
            else if (prevActiveObject['axis'] === 'column' && this.m_columnHeaderLevelCount > 1)
            {
                previousElement = this._getHeaderByIndex(prevActiveObject['index'], prevActiveObject['level'], this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
            }
            else if (prevActiveObject['axis'] === 'rowEnd' && this.m_rowEndHeaderLevelCount > 1)
            {
                previousElement = this._getHeaderByIndex(prevActiveObject['index'], prevActiveObject['level'], this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
        }
            else if (prevActiveObject['axis'] === 'columnEnd' && this.m_columnEndHeaderLevelCount > 1)
            {
                previousElement = this._getHeaderByIndex(prevActiveObject['index'], prevActiveObject['level'], this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
            }
        }

        label = [this.createSubId("context"), this._getHeaderAndParentIds(element, previousElement)].join(" ");
        direction = element.getAttribute(this.getResources().getMappedAttribute('sortDir'));
        if (direction === "ascending")
        {
            key = "accessibleSortAscending";
            label = label + " " + this.createSubId("state");
        }
        else if (direction === "descending")
        {
            key = "accessibleSortDescending";
            label = label + " " + this.createSubId("state");
        }

        if (this.m_externalFocus === true)
        {
            label = [this.createSubId("summary"), label].join(" ");
            this.m_externalFocus = false;
        }

        if (key != null)
        {
        this._updateStateInfo([{'key':key, 'args':{'id': ''}}]);
        }

        element.setAttribute("tabIndex", 0);
    }
    else
    {
        if (prevActiveObject != null)
        {
            if (prevActiveObject['type'] === 'header')
            {
                previousRowIndex = prevActiveObject['axis'] === 'row' ? prevActiveObject['index'] : null;
                previousColumnIndex = prevActiveObject['axis'] === 'column' ? prevActiveObject['index'] : null;
            }
            else
            {
                previousRowIndex = prevActiveObject['indexes']['row'];
                previousColumnIndex = prevActiveObject['indexes']['column'];
            }
        }

        // Add the header labels
        row = this._getHeaderLabel('row', this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader, this.m_endRowHeader, activeObject['indexes']['row'], previousRowIndex, element);
        rowEnd = this._getHeaderLabel('rowEnd', this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader, this.m_endRowEndHeader, activeObject['indexes']['row'], previousRowIndex, element);
        column = this._getHeaderLabel('column', this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader, this.m_endColHeader, activeObject['indexes']['column'], previousColumnIndex, element);
        columnEnd = this._getHeaderLabel('columnEnd', this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader, this.m_endColEndHeader, activeObject['indexes']['column'], previousColumnIndex, element);

        // we want extent information at the end of the readout so put it in the state info
        var rowExtent = activeObject['extents']['row'];
        var columnExtent = activeObject['extents']['column']; 
        var statesArray = [];
        if (rowExtent > 1)
        {
            statesArray.push({'key': 'accessibleRowSpanContext', 'args': {'extent': rowExtent}})
        }
        if (columnExtent > 1)
        {
            statesArray.push({'key': 'accessibleColumnSpanContext', 'args': {'extent': columnExtent}})
        }
        this._updateStateInfo(statesArray);

        label = [this.createSubId("context"), row, rowEnd, column, columnEnd, element['id'], this.createSubId("state")].join(" ");
        // remove double spaces rather than check everytime
        label = label.replace(/ +(?= )/g,'');

        if (this.m_externalFocus)
        {
            label = [this.createSubId("summary"), label].join(" ");
            this.m_externalFocus = false;
        }
    }
    return label;
};

/**
 * Returns the header that is in line with a cell along an axis.
 * Key Note: in the case of row, we return the row not the headercell
 * @param {Element|undefined|null} cell the element for the cell
 * @param {string} axis the axis along which to find the header, 'row', 'column'
 * @param {boolean=} lastContained true to get the last header contained by the cell along an axis
 * @return {Element} the header Element along the axis
 */
DvtDataGrid.prototype.getHeaderFromCell = function(cell, axis, lastContained)
{
    var index, extent, level, container, levelCount, start;

    if (axis === 'row')
    {
        if (this.m_rowHeader != null)
        {
            index = this._getIndex(cell, 'row');
            extent = this._getExtent(cell, 'row');
            level = this.m_rowHeaderLevelCount - 1;
            container = this.m_rowHeader;
            levelCount = this.m_rowHeaderLevelCount;
            start = this.m_startRowHeader;
        }
    }
    else if (axis === 'column')
    {
        if (this.m_colHeader != null)
        {
            index = this._getIndex(cell, 'column');
            extent = this._getExtent(cell, 'column');
            level = this.m_columnHeaderLevelCount - 1;
            container = this.m_colHeader;
            levelCount = this.m_columnHeaderLevelCount;
            start = this.m_startColHeader;
        }
    }
    else if (axis === 'rowEnd')
    {
        if (this.m_rowEndHeader != null)
        {
            index = this._getIndex(cell, 'row');
            extent = this._getExtent(cell, 'row');
            level = this.m_rowEndHeaderLevelCount - 1;
            container = this.m_rowEndHeader;
            levelCount = this.m_rowEndHeaderLevelCount;
            start = this.m_startRowEndHeader;
        }
    }
    else if (axis === 'columnEnd')
    {
        if (this.m_colEndHeader != null)
        {
            index = this._getIndex(cell, 'column');
            extent = this._getExtent(cell, 'column');
            level = this.m_columnEndHeaderLevelCount - 1;
            container = this.m_colEndHeader;
            levelCount = this.m_columnEndHeaderLevelCount;
            start = this.m_startColEndHeader;
        }
    }

    if (index != null && level != null && index > -1)
    {
        if (lastContained)
        {
            index += extent - 1;
        }
        return this._getHeaderByIndex(index, level, container, levelCount, start);
    }

    return null;
};

/**
 * Creates a range object given the start and end index, will add in keys if they are passed in
 * @param {Object} startIndex - the start index of the range
 * @param {Object=} endIndex - the end index of the range.  Optional, if not specified it represents a single cell/row
 * @param {Object=} startKey - the start key of the range.  Optional, if not specified it represents a single cell/row
 * @param {Object=} endKey - the end key of the range.  Optional, if not specified it represents a single cell/row
 * @return {Object} a range object representing the start and end index, along with the start and end key.
 */
DvtDataGrid.prototype.createRange = function(startIndex, endIndex, startKey, endKey)
{
    var startRow, endRow, startColumn, endColumn, startRowKey, endRowKey, startColumnKey, endColumnKey;
    if (endIndex)
    {
        // -1 means unbound
        if (startIndex['row'] < endIndex['row'] || endIndex['row'] == -1)
        {
            startRow = startIndex['row'];
            endRow = endIndex['row'];
            if (startKey)
            {
                startRowKey = startKey['row'];
                endRowKey = endKey['row'];
            }
        }
        else
        {
            startRow = endIndex['row'];
            endRow = startIndex['row'];
            if (startKey)
            {
                startRowKey = endKey['row'];
                endRowKey = startKey['row'];
            }
        }

        // row based selection does not have column specified for range
        if (!isNaN(startIndex['column']) && !isNaN(endIndex['column']))
        {
            // -1 means unbound
            if (startIndex['column'] < endIndex['column'] || endIndex['column'] == -1)
            {
                startColumn = startIndex['column'];
                endColumn = endIndex['column'];
                if (startKey)
                {
                    startColumnKey = startKey['column'];
                    endColumnKey = endKey['column'];
                }
            }
            else
            {
                startColumn = endIndex['column'];
                endColumn = startIndex['column'];
                if (startKey)
                {
                    startColumnKey = endKey['column'];
                    endColumnKey = startKey['column'];
                }
            }

            startIndex = {
                "row": startRow, "column": startColumn
            };
            endIndex = {
                "row": endRow, "column": endColumn
            };
            if (startKey)
            {
                startKey = {
                    "row": startRowKey, "column": startColumnKey
                };
                endKey = {
                    "row": endRowKey, "column": endColumnKey
                };
            }
        }
        else
        {
            startIndex = {
                "row": startRow
            };
            endIndex = {
                "row": endRow
            };
            if (startKey)
            {
                startKey = {
                    "row": startRowKey, "column": startColumnKey
                };
                endKey = {
                    "row": endRowKey, "column": endColumnKey
                };
            }
        }
    }

    if (startKey)
    {
        return {"startIndex": startIndex, "endIndex": endIndex, "startKey": startKey, "endKey": endKey};
    }

    return {"startIndex": startIndex, "endIndex": endIndex};
};


/**
 * Creates a range object given the start and end index
 * @param {Object} startIndex - the start index of the range
 * @param {Object|undefined|null} endIndex - the end index of the range.
 * @param {Function} callback - the callback for the range to call when its fully fetched
 * @private
 */
DvtDataGrid.prototype._createRangeWithKeys = function(startIndex, endIndex, callback)
{
    this._keys(startIndex, this._createRangeStartKeyCallback.bind(this, endIndex, callback));
};

/**
 * Creates a range object given the start and end index
 * @param {Object|null|undefined} endIndex - the end index of the range.
 * @param {Function} callback - the callback for the range to call when its fully fetched
 * @param {Object} startKey - the start key of the range
 * @param {Object} startIndex - the start index of the range
 * @private
 */
DvtDataGrid.prototype._createRangeStartKeyCallback = function(endIndex, callback, startKey, startIndex)
{
    //keys will be the same
    if (endIndex === startIndex)
    {
        this._createRangeEndKeyCallback(startKey, startIndex, callback, startKey, startIndex);
    }
    //new keys needed
    else if (endIndex)
    {
        this._keys(endIndex, this._createRangeEndKeyCallback.bind(this, startKey, startIndex, callback));
    }
    //create range from single key
    else
    {
        callback.call(this, {"startIndex": startIndex, "endIndex": startIndex, "startKey": startKey, "endKey": startKey});
    }
};

/**
 * Creates a range object given the start and end index
 * @param {Object} startKey - the start key of the range
 * @param {Object} startIndex - the start index of the range
 * @param {Function} callback - the callback for the range to call when its fully fetched
 * @param {Object} endKey - the end key of the range.
 * @param {Object} endIndex - the end index of the range.
 * @private
 */
DvtDataGrid.prototype._createRangeEndKeyCallback = function(startKey, startIndex, callback, endKey, endIndex)
{
    callback.call(this, this.createRange(startIndex, endIndex, startKey, endKey));
};

/**
 * Retrieve the end index of the range, return start index if end index is undefined
 * @param {Object} range
 * @return {Object}
 */
DvtDataGrid.prototype.getEndIndex = function(range)
{
    return (range['endIndex'] == null) ? range['startIndex'] : range['endIndex'];
};

/**
 * Grabs all the elements in the databody which are within the specified range.
 * @param {Object} range - the range in which to get the elements
 * @param {number=} startRow
 * @param {number=} endRow
 * @param {number=} startCol
 * @param {number=} endCol
 * @return {Array}
 */
DvtDataGrid.prototype.getElementsInRange = function(range, startRow, endRow, startCol, endCol)
{
    var startIndex, endIndex, rangeStartRow, rangeEndRow, rangeStartColumn, rangeEndColumn, nodes, databodyContent, rows, i, columns, j, cell, row;
    if (startRow == undefined)
    {
        startRow = this.m_startRow;
    }
    if (endRow == undefined)
    {
        endRow = this.m_endRow + 1;
    }

    startIndex = range['startIndex'];
    endIndex = this.getEndIndex(range);

    rangeStartRow = startIndex['row'];
    rangeEndRow = endIndex['row'];
    // index = -1 means unbounded index
    if (rangeEndRow == -1)
    {
        rangeEndRow = Number.MAX_VALUE;
    }

    // check if in the rendered range
    if (endRow < rangeStartRow || rangeEndRow < startRow)
    {
        return null;
    }

    if (!isNaN(startIndex['column']) && !isNaN(endIndex['column']))
    {
        rangeStartColumn = startIndex['column'];
        rangeEndColumn = endIndex['column'];
        // index = -1 means unbounded index
        if (rangeEndColumn == -1)
        {
            rangeEndColumn = Number.MAX_VALUE;
        }

        // check if in the rendered range
        if ((this.m_endCol + 1) < rangeStartColumn || rangeEndColumn < this.m_startCol)
        {
            return null;
        }
    }

    nodes = [];
    // now walk the databody to find the nodes in range
    databodyContent = this.m_databody['firstChild'];
    if (databodyContent == null)
    {
        return null;
    }

    // the range is within the databody, calculate the relative position
    rangeStartRow = Math.max(0, rangeStartRow);
    rangeEndRow = Math.min(this.m_endRow, rangeEndRow);

    // cell case
    if (!isNaN(rangeStartColumn) && !isNaN(rangeEndColumn))
    {
        rangeStartColumn = Math.max(0, rangeStartColumn);
        rangeEndColumn = Math.min(this.m_endCol, rangeEndColumn);

        for (i = rangeStartRow; i <= rangeEndRow; i += 1)
        {
            for (j = rangeStartColumn; j <= rangeEndColumn; j += 1)
            {
                cell = this._getCellByIndex(this.createIndex(i,j));
                if (cell != null && nodes.indexOf(cell) === -1)
                {
                    nodes.push(cell);
                }
            }
        }
    }
    else
    {
        // row case
        rangeStartColumn = Math.max(0, this.m_startCol);
        rangeEndColumn = Math.max(rangeStartColumn, this.m_endCol);
        for (i = rangeStartRow; i <= rangeEndRow; i += 1)
        {
            for (j = rangeStartColumn; j <= rangeEndColumn; j += 1)
            {
                cell = this._getCellByIndex(this.createIndex(i,j));
                if (cell != null && nodes.indexOf(cell) === -1)
                {
                    nodes.push(cell);
                }
            }
        }
    }

    return nodes;
};

/**
 * Read the full content of the active cell (or frontier cell) to the screen reader
 * @protected
 * @returns {boolean} true if there is content to read out
 */
DvtDataGrid.prototype.readCurrentContent = function()
{
    var current, range, cell, currentCell, subid, needToModify, labelledBy;

    if (this.m_active == null)
    {
        return false;
    }

    if (this.m_active['type'] == 'header')
    {
        current = {};
        if (this.m_active['axis'] === 'row')
        {
            if (this.m_rowHeaderLevelCount > 1)
            {
                current['level'] = this.m_active['level'];
            }
            current['rowHeader'] = this.m_active['index'];
        }
        else
        {
            if (this.m_columnHeaderLevelCount > 1)
            {
                current['level'] = this.m_active['level'];
            }
            current['columnHeader'] = this.m_active['index'];
        }
        currentCell = this._getActiveElement();
    }
    else
    {
        current = this.m_active['indexes'];
        if (this._isSelectionEnabled() && this.isMultipleSelection())
        {
            if (this.m_selectionFrontier != null)
            {
                current = this.m_selectionFrontier;
            }
        }
        // make sure there is an active cell or frontier cell
        if (current == null)
        {
            return false;
        }

        // find the cell div
        range = this.createRange(current);
        cell = this.getElementsInRange(range);
        if (cell == null || cell.length == 0)
        {
            return false;
        }

        currentCell = cell[0];
    }

    //Fill in the placeholder div aria-labelledby with the currentCell label so that it 
    //will read out the contents of the current cell when focus is shifted to the 
    //place holder div. Set tabIndex to -1 so that it can be focused but not tabbed into.
    this.m_placeHolder.tabIndex = -1;
    labelledBy = currentCell.getAttribute("aria-labelledby");
    this.m_placeHolder.setAttribute('aria-labelledby', labelledBy + ' ' + currentCell.id);

    //Since JAWS screen reader requires changing focus in order to 
    //trigger a read command, we are toggling between the placeholder
    //div and the currentCell.
    if (this.m_placeHolder === document.activeElement) {
        currentCell.focus();
    }
    else
    {
        this.m_placeHolder.focus();
    }
    return true;
};

/**
 * Enter actionable mode
 * @param {Element|undefined|null} element to set actionable
 * @returns {boolean} false
 */
DvtDataGrid.prototype._enterActionableMode = function(element)
{
    // enable all focusable elements
    this._enableAllFocusableElements(element);

    // focus on first focusable item in the cell
    if (this._setFocusToFirstFocusableElement(element))
    {
        this.m_focusOutHandler(element);
        this.setActionableMode(true);
    }
    return false;
};

/**
 * Exit actionable mode on the active cell if in actionable mode
 */
DvtDataGrid.prototype._exitActionableMode = function()
{
    var elem;
    if (this.isActionableMode())
    {
        elem = this._getActiveElement();
        this.setActionableMode(false);
        this._disableAllFocusableElements(elem);
        this.m_focusInHandler(elem);
    }
};

/**
 * Re render a cell
 * @param {Element|undefined|null} cell
 * @param {string} mode
 * @param {string} classToToggle class to toggle on or off before rerendering
 * @param {Element|null=} editableClone
 */
DvtDataGrid.prototype._reRenderCell = function(cell, mode, classToToggle, editableClone)
{
    var renderer, cellContext;
    renderer = this.m_options.getRenderer('cell');
    cellContext = cell[this.getResources().getMappedAttribute("context")];
    cellContext['mode'] = mode;

    // empty the cell
    this.m_utils.empty(cell);

    // now that the cell is empty toggle the appropraite edit classes so that alignment never has to shift
    if (this.m_utils.containsCSSClassName(cell, classToToggle))
    {
        this.m_utils.removeCSSClassName(cell, classToToggle);
    }
    else
    {
        this.m_utils.addCSSClassName(cell, classToToggle);
    }

    if (editableClone)
    {
        while (editableClone.hasChildNodes()) {
            cell.appendChild(editableClone.firstChild);
        }
        delete this.m_editableClone;
    }
    else
    {
        this._renderContent(renderer, cellContext, cell, cellContext['data']);
    }
};

/**
 * Set the editable clone property
 * @param {Element|undefined|null} element to clone
 */
DvtDataGrid.prototype._setEditableClone = function(element)
{
    delete this.m_editableClone;
    if (element != null)
    {
        var clone = element.cloneNode(false);
        this._createUniqueId(clone);
        clone[this.getResources().getMappedAttribute("context")] = element[this.getResources().getMappedAttribute("context")];
        clone[this.getResources().getMappedAttribute("context")]['parentElement'] = clone;
        this.m_root.appendChild(clone);
        this._reRenderCell(clone, 'edit', this.getMappedStyle('cellEdit'), null);
        this.m_root.removeChild(clone);
        clone[this.getResources().getMappedAttribute("context")]['parentElement'] = element;
        this.m_editableClone = clone;
    }
};

/**
 *
 * @param {number} keyCode
 * @return {boolean}
 */
DvtDataGrid.prototype.isNavigationKey = function(keyCode)
{
    return (this.isArrowKey(keyCode) || keyCode == DvtDataGrid.keyCodes.HOME_KEY || keyCode == DvtDataGrid.keyCodes.END_KEY || keyCode == DvtDataGrid.keyCodes.PAGEUP_KEY || keyCode == DvtDataGrid.keyCodes.PAGEDOWN_KEY);
};

/**
 *
 * @param {number} keyCode
 * @return {boolean}
 */
DvtDataGrid.prototype.isArrowKey = function(keyCode)
{
    return (keyCode == DvtDataGrid.keyCodes.UP_KEY || keyCode == DvtDataGrid.keyCodes.DOWN_KEY || keyCode == DvtDataGrid.keyCodes.LEFT_KEY || keyCode == DvtDataGrid.keyCodes.RIGHT_KEY);
};

/**
 * Creates an index object for the cell/row
 * @param {number|string|null=} row - the start index of the range
 * @param {number|string|null=} column - the end index of the range.  Optional, if not specified it represents a single cell/row
 * @return {Object} an index object
 */
DvtDataGrid.prototype.createIndex = function(row, column)
{
    if (row == null && column == null)
    {
        return null;
    }
    
    var returnObj = {};
    if (row != null)
    {
        returnObj['row'] = row;
    }
    if (column != null)
    {
        returnObj['column'] = column;
    }
    return returnObj;
};

/**
 * Checks if the corners of a selection matches the selection frontier index and parameter index
 * @param {string} axis - the axis of the selection if header
 * @param {number} index - one index of the selection
 * @param {Object} selection - selection being checked
 * @return {boolean} true or false whether or not the corners are matched
 */
DvtDataGrid.prototype.checkCorners = function(axis, index, selection)
{
    //ignore nested headers
    if ((this.m_selectionFrontier['axis'] == 'column' && this.m_columnHeaderLevelCount != this.m_selectionFrontier['level']) || (this.m_selectionFrontier['axis'] == 'row' && this.m_rowHeaderLevelCount != this.m_selectionFrontier['level']) || (this.m_selectionFrontier['axis'] == 'rowEnd' && this.m_rowEndHeaderLevelCount != this.m_selectionFrontier['level']) || (this.m_selectionFrontier['axis'] == 'columnEnd' && this.m_columnEndHeaderLevelCount != this.m_selectionFrontier['level']))
    {
        return true;
    }
    if (axis.indexOf('row') != -1)
    {
        return ((index == selection['startIndex']['row'] && this.m_selectionFrontier['end'] == selection['endIndex']['row']) || (index == selection['endIndex']['row'] && this.m_selectionFrontier['end'] == selection['startIndex']['row']))
    }
    else if (axis.indexOf('column') != -1)
    {
        return ((index == selection['startIndex']['column'] && this.m_selectionFrontier['end'] == selection['endIndex']['column']) || (index == selection['endIndex']['column'] && this.m_selectionFrontier['end'] == selection['startIndex']['column']))
    }
    else 
    {
        return false;
    }
};

/**
 * Handles arrow keys navigation on header
 * @param {number} keyCode description
 * @param {Event} event the DOM event that caused the arrow key
 * @param {boolean} isExtend boolean if we are extending a selection
 * @param {boolean} jumpToHeaders jump to opposite headers if possible
 * @return  boolean true if the event was processed
 */
DvtDataGrid.prototype.handleHeaderFocusChange = function(keyCode, event, isExtend, jumpToHeaders)
{
    var axis, index, level, elem, newCellIndex, newElement, newIndex, newLevel, depth,
            root, start, end, levelCount, stopFetch;

    // ensure that there's no outstanding fetch requests
    if (!this.isFetchComplete())
    {
        //act like it's processed until we finish the fetch
        return true;
    }

    if (this.getResources().isRTLMode())
    {
        if (keyCode == DvtDataGrid.keyCodes.LEFT_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.RIGHT_KEY;
        }
        else if (keyCode == DvtDataGrid.keyCodes.RIGHT_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.LEFT_KEY;
        }
    }

    axis = this.m_active['axis'];
    index = this.m_active['index'];
    level = this.m_active['level'];
    elem = this._getActiveElement();
    depth = elem != null ? this._getAttribute(elem, 'depth', true) : 1;
    
    // if cell mode, need to set values properly to check corners
    if (!axis && !index && this.m_active)
    {
        axis = this.m_selectionFrontier['axis'];                
        index = this.m_active['indexes'][axis];        
    }

    // if shiftkey active and we're navigating headers, as long as we don't leave headers, set isExtend to active.
    if (event.shiftKey && !isExtend && this.m_selectionFrontier && this.isMultipleSelection())
    {
        isExtend = this.checkHeaderToDatabody(axis, keyCode);
    }

    // if extending and anchors are the same, use the selectionFrontier
    if (isExtend && this.isArrowKey(keyCode) && this.isHeaderSelectionType(this.m_selectionFrontier) && this.checkCorners(axis, index, this.m_selection[this.m_selection.length - 1]))
    {
        axis = this.m_selectionFrontier['axis'];
        index = this.m_selectionFrontier['index'];
        level = this.m_selectionFrontier['level'];

        if (index == -1)
        {
            index = this.m_active['indexes'][axis];
        }

        // don't allow changing levels if anchor is a cell. 
        if (this.m_active['type'] == 'cell')
        {
            if (((keyCode === DvtDataGrid.keyCodes.LEFT_KEY || keyCode === DvtDataGrid.keyCodes.RIGHT_KEY) && axis.indexOf("row") != -1) || ((keyCode === DvtDataGrid.keyCodes.UP_KEY || keyCode === DvtDataGrid.keyCodes.DOWN_KEY) && axis.indexOf("column") != -1))
            {
                return;
            }
        }
    }

    if (axis == 'column')
    {
        root = this.m_colHeader;
        start = this.m_startColHeader;
        end = this.m_endColHeader;
        levelCount = this.m_columnHeaderLevelCount;
        stopFetch = this.m_stopColumnHeaderFetch;
    }
    else if (axis == 'row')
    {
        root = this.m_rowHeader;
        start = this.m_startRowHeader;
        end = this.m_endRowHeader;
        levelCount = this.m_rowHeaderLevelCount;
        stopFetch = this.m_stopRowHeaderFetch;
    }
    if (axis == 'columnEnd')
    {
        // treat up and down keys opposite of column Headers
        if (keyCode == DvtDataGrid.keyCodes.DOWN_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.UP_KEY;
        }
        else if (keyCode == DvtDataGrid.keyCodes.UP_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.DOWN_KEY;
        }
        root = this.m_colEndHeader;
        start = this.m_startColEndHeader;
        end = this.m_endColEndHeader;
        levelCount = this.m_columnEndHeaderLevelCount;
        stopFetch = this.m_stopColumnEndHeaderFetch;
   }
    if (axis == 'rowEnd')
    {
        // treat right and left oppostie of row headers
        if (keyCode == DvtDataGrid.keyCodes.LEFT_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.RIGHT_KEY;
        }
        else if (keyCode == DvtDataGrid.keyCodes.RIGHT_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.LEFT_KEY;
        }
        root = this.m_rowEndHeader;
        start = this.m_startRowEndHeader;
        end = this.m_endRowEndHeader;
        levelCount = this.m_rowEndHeaderLevelCount;
        stopFetch = this.m_stopRowEndHeaderFetch;
    }

    // get the selection frontier current header element
    if (isExtend && this.isArrowKey(keyCode) && this.isHeaderSelectionType(this.m_selectionFrontier))
    {
        elem = this._getHeaderByIndex(index, level, root, levelCount, start);
        depth = elem != null ? this._getAttribute(elem, 'depth', true) : 1;
    }

    switch (keyCode)
    {
        case DvtDataGrid.keyCodes.LEFT_KEY:
            if ((axis === 'column' || axis === 'columnEnd') && index > 0)
            {
                if (levelCount === 1)
                {
                    newIndex = index - 1;
                    newElement = elem != null ? elem['previousSibling'] : null;
                    newLevel = level;
                }
                else
                {
                    newElement = this._getHeaderByIndex(index - 1, level, root, levelCount, start);
                    newIndex = newElement != null ? this._getAttribute(newElement['parentNode'], 'start', true) : index - 1;
                    newLevel = newElement != null ? this.getHeaderCellLevel(newElement) : level;
                    if (newIndex < 0)
                    {
                        break;
                    }
                }
                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this._clearSelection(event);
                        this.m_selectionFrontier = {};
                    } 
                    this._setActive(newElement, {'type':'header', 'index':newIndex, 'level':newLevel, 'axis':axis}, event);
                    this._highlightActive();                    
                }
            }
            else if ((axis === 'row' || axis === 'rowEnd') && level > 0)
            {
                //moving down a level in the header
                newElement = this._getHeaderByIndex(index, level - 1, root, levelCount, start);
                newIndex = this._getAttribute(newElement['parentNode'], 'start', true);
                newLevel = this.getHeaderCellLevel(newElement);

                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this._clearSelection(event);
                        this.m_selectionFrontier = {};
                    }
                    this._setActive(newElement, {'type':'header', 'index':newIndex, 'level':newLevel, 'axis':axis}, event);
                    this._highlightActive();                    
                }
            }
            break;
        case DvtDataGrid.keyCodes.RIGHT_KEY:
            if (axis == 'rowEnd' && jumpToHeaders && this.m_endRowHeader != -1)
            {
                newElement = this._getHeaderByIndex(index, this.m_rowHeaderLevelCount, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this._clearSelection(event);
                        this.m_selectionFrontier = {};
                    }
                    this._setActive(newElement, {'type':'header', 'index':index, 'level':this.m_rowHeaderLevelCount, 'axis':axis}, event);
                    this._highlightActive();                
                }
            }
            else if (axis == 'row' && jumpToHeaders && this.m_endRowEndHeader != -1)
            {
                newElement = this._getHeaderByIndex(index, this.m_rowEndHeaderLevelCount, this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this._clearSelection(event);
                        this.m_selectionFrontier = {};
                    }
                    this._setActive(newElement, {'type':'header', 'index':index, 'level':this.m_rowEndHeaderLevelCount, 'axis':axis}, event);
                    this._highlightActive();                
                }
            }
            else if (axis === 'row' || axis === 'rowEnd')
            {
                if (level + depth >= levelCount && !isExtend)
                {
                    // row header, move to databody
                    // make the first cell of the current row active
                    // no need to scroll since it will be in the viewport
                    if (axis == 'row')
                    {
                        newCellIndex = this.createIndex(index, 0);
                    }
                    else if (this._isHighWatermarkScrolling())
                    {
                        newCellIndex = this.createIndex(index, this.m_endCol);
                    }
                    else
                    {
                        newCellIndex = this.createIndex(index, this.getDataSource().getCount("column") - 1);
                    }
                    
                    this.m_trueIndex = {'row': index};

                    if (this._isSelectionEnabled())
                    {
                        this.selectAndFocus(newCellIndex, event);
                    }
                    else
                    {
                        this._setActiveByIndex(newCellIndex, event);
                        this._highlightActive();
                    }
                }
                else
                {
                    //moving down a level in the header
                    newElement = this._getHeaderByIndex(index, level + depth, root, levelCount, start);
                    newIndex = this._getAttribute(newElement['parentNode'], 'start', true);
                    newLevel = this.getHeaderCellLevel(newElement);
                    if (isExtend)
                    {
                        this.extendSelectionHeader(newElement, event, true);
                    }
                    else
                    {
                        if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                        {
                            // unhighlight and clear selection
                            this._clearSelection(event);
                            this.m_selectionFrontier = {};
                        }
                        this._setActive(newElement, {'type':'header', 'index':newIndex, 'level': newLevel, 'axis':axis}, event);
                        this._highlightActive();                
                    }
                }
            }
            else
            {
                if (levelCount === 1)
                {
                    newIndex = index + 1;
                    newElement = elem != null ? elem['nextSibling'] : null;
                    newLevel = level;
                }
                else
                {
                    if (level === levelCount - 1)
                    {
                        newIndex = index + 1;
                        newElement = this._getHeaderByIndex(newIndex, level, root, levelCount, start);
                    }
                    else
                    {
                        newIndex = elem != null ? this._getAttribute(elem['parentNode'], 'start', true) + this._getAttribute(elem['parentNode'], 'extent', true) : index + 1;
                        newElement = this._getHeaderByIndex(newIndex, level, root, levelCount, start);
                    }
                    newLevel = newElement != null ? this.getHeaderCellLevel(newElement) : level;
                }

                if (!(newIndex > end && stopFetch) && (this._isCountUnknown("column") || newIndex < this.getDataSource().getCount("column")))
                {
                    if (isExtend)
                    {
                        this.extendSelectionHeader(newElement, event, true);
                    }
                    else
                    {
                        if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                        {
                            // unhighlight and clear selection
                            this._clearSelection(event);
                            this.m_selectionFrontier = {};
                        }
                        this._setActive(newElement, {'type':'header', 'index':newIndex, 'level':newLevel, 'axis':axis}, event);
                        this._highlightActive();                
                    }
                }
            }
            break;
        case DvtDataGrid.keyCodes.UP_KEY:
            if ((axis === 'row' || axis === 'rowEnd') && index > 0)
            {
                if (levelCount === 1)
                {
                    newIndex = index - 1;
                    newElement = elem != null ? elem['previousSibling'] : null;
                    newLevel = level;
                }
                else
                {
                    if (level === levelCount - 1)
                    {
                        newIndex = index - 1;
                        newElement = this._getHeaderByIndex(newIndex, level, root, levelCount, start);
                    }
                    else
                    {
                        newElement = this._getHeaderByIndex(this._getAttribute(elem['parentNode'], 'start', true) - 1, level, root, levelCount, start);
                        newIndex = newElement != null ? this._getAttribute(newElement['parentNode'], 'start', true) : index - 1;
                    }
                    newLevel = newElement != null ? this.getHeaderCellLevel(newElement) : level;
                    if (newIndex < 0)
                    {
                        break;
                    }
                }
                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this.m_selectionFrontier = {};
                        this._clearSelection(event);
                    }
                    this._setActive(newElement, {'type':'header', 'index':newIndex, 'level':newLevel, 'axis':axis}, event);
                    this._highlightActive();
                }
            }
            else if ((axis === 'column' || axis === 'columnEnd') && level > 0)
            {
                //moving down a level in the header
                newElement = this._getHeaderByIndex(index, level - 1, root, levelCount, start);
                newIndex = this._getAttribute(newElement['parentNode'], 'start', true);
                newLevel = this.getHeaderCellLevel(newElement);
                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this.m_selectionFrontier = {};
                        this._clearSelection(event);
                    }
                    this._setActive(newElement, {'type':'header', 'index':newIndex, 'level':newLevel, 'axis':axis}, event);
                    this._highlightActive();
                }
            }
            break;
        case DvtDataGrid.keyCodes.DOWN_KEY:
            if (axis == 'columnEnd' && jumpToHeaders && this.m_endColHeader != -1)
            {
                newElement = this._getHeaderByIndex(index, this.m_columnHeaderLevelCount, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this.m_selectionFrontier = {};
                        this._clearSelection(event);
                    }
                    this._setActive(newElement, {'type':'header', 'index':index, 'level': this.m_columnHeaderLevelCount, 'axis':axis}, event);
                    this._highlightActive();
                }
            }
            else if (axis == 'column' && jumpToHeaders && this.m_endColEndHeader != -1)
                {
                newElement = this._getHeaderByIndex(index, this.m_columnEndHeaderLevelCount, this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
                if (isExtend)
                {
                    this.extendSelectionHeader(newElement, event, true);
                }
                else
                {
                    if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                    {
                        // unhighlight and clear selection
                        this._clearSelection(event);
                        this.m_selectionFrontier = {};
                    }
                    this._setActive(newElement, {'type':'header', 'index':index, 'level': this.m_columnEndHeaderLevelCount, 'axis':axis}, event);
                    this._highlightActive();
                }
            }
            else if (axis === 'column' || axis === 'columnEnd')
            {
                if (level + depth >= levelCount && !isExtend)
                {
                    // column header, move to databody
                    // make the cell of the first row and current column active
                    // no need to scroll since it will be in the viewport
                    if (axis == 'column')
                    {
                        newCellIndex = this.createIndex(0, index);
                    }
                    else if (this._isHighWatermarkScrolling())
                    {
                        newCellIndex = this.createIndex(this.m_endRow, index);
                    }
                    else
                    {
                        newCellIndex = this.createIndex(this.getDataSource().getCount("row") - 1, index);
                    }
                    
                    this.m_trueIndex = {'column': index};

                    if (this._isSelectionEnabled())
                    {
                        this.selectAndFocus(newCellIndex, event);
                    }
                    else
                    {
                        this._setActiveByIndex(newCellIndex, event);
                        this._highlightActive();
                    }
                }
                else
                {
                    //moving down a level in the header
                    newElement = this._getHeaderByIndex(index, level + depth, root, levelCount, start);
                    newIndex = this._getAttribute(newElement['parentNode'], 'start', true);
                    newLevel = this.getHeaderCellLevel(newElement);
                    if (isExtend)
                    {
                        this.extendSelectionHeader(newElement, event, true);
                    }
                    else
                    {
                        if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                        {
                            // unhighlight and clear selection
                            this._clearSelection(event);
                            this.m_selectionFrontier = {};
                        }
                        this._setActive(newElement, {'type':'header', 'index':newIndex, 'level': newLevel, 'axis':axis}, event);
                        this._highlightActive();
                    }
                }
            }
            else
            {
                if (levelCount === 1)
                {
                    newIndex = index + 1;
                    newElement = elem != null ?  elem['nextSibling'] : null;
                    newLevel = level;
                }
                else
                {
                    if (level === levelCount - 1)
                    {
                        newIndex = index + 1;
                        newElement = this._getHeaderByIndex(newIndex, level, root, levelCount, start);
                    }
                    else
                    {
                        newIndex = elem != null ? this._getAttribute(elem['parentNode'], 'start', true) + this._getAttribute(elem['parentNode'], 'extent', true) : index + 1;
                        newElement = this._getHeaderByIndex(newIndex, level, root, levelCount, start);
                    }
                    newLevel = newElement != null ? this.getHeaderCellLevel(newElement) : level;
                }

                if (!(newIndex > end && stopFetch) && (this._isCountUnknown("row") || newIndex < this.getDataSource().getCount("row")))
                {
                    if (isExtend)
                    {
                        this.extendSelectionHeader(newElement, event, true);
                    }
                    else
                    {
                        if (this._isSelectionEnabled() && !this.m_discontiguousSelection)
                        {
                            // unhighlight and clear selection
                            this._clearSelection(event);
                            this.m_selectionFrontier = {};
                        }
                        this._setActive(newElement, {'type':'header', 'index':newIndex, 'level': newLevel, 'axis':axis}, event);
                        this._highlightActive();
                    }
                }
            }
            break;
        case DvtDataGrid.keyCodes.PAGEUP_KEY:
            if (axis === 'row' || axis === 'rowEnd')
            {
                // selects the first available row header
                elem = this._getHeaderByIndex(0, level, root, levelCount, start);
                this._setActive(elem, {'type':'header', 'index':0, 'level': level, 'axis':axis}, event);
            }
            break;
        case DvtDataGrid.keyCodes.PAGEDOWN_KEY:
            if (axis === 'row' || axis === 'rowEnd')
            {
                // selects the last available row header
                if (!this._isCountUnknown("row") && !this._isHighWatermarkScrolling())
                {
                    index = Math.max(0, this.getDataSource().getCount("row") - 1);
                }
                else
                {
                    index = Math.max(0, end);
                }
                elem = this._getHeaderByIndex(index, level, root, levelCount, start);
                this._setActive(elem, {'type':'header', 'index':index, 'level': level, 'axis':axis}, event);
            }
            break;
        case DvtDataGrid.keyCodes.HOME_KEY:
            if (axis === 'column' || axis === 'columnEnd')
            {
                // selects the first cell of the current row
                elem = this._getHeaderByIndex(0, level, root, levelCount, start);
                this._setActive(elem, {'type':'header', 'index':0, 'level': level, 'axis':axis}, event);
            }
            break;
        case DvtDataGrid.keyCodes.END_KEY:
            if (axis === 'column' || axis === 'columnEnd')
            {
                // selects the last cell of the current row
                if (!this._isCountUnknown("column") && !this._isHighWatermarkScrolling())
                {
                    index = Math.max(0, this.getDataSource().getCount("column") - 1);
                }
                else
                {
                    index = Math.max(0, end);
                }
                // selects the first cell of the current row
                elem = this._getHeaderByIndex(index, level, root, levelCount, start);
                this._setActive(elem, {'type':'header', 'index':index, 'level': level, 'axis':axis}, event);
            }
            break;
    }
    return true;
};

/**
 * Check if the focus is changing from header to databody
 * Get the label of the header
 * @param {string} axis 
 * @param {number} keyCode
 * @returns {boolean} True if the header is not leaving the databody
 */
DvtDataGrid.prototype.checkHeaderToDatabody = function(axis, keyCode)
{
    if (!((axis === 'row' && this.m_rowHeaderLevelCount === this.m_selectionFrontier['level'] && keyCode == DvtDataGrid.keyCodes.RIGHT_KEY) || (axis === 'rowEnd' && this.m_rowEndHeaderLevelCount === this.m_selectionFrontier['level'] && keyCode == DvtDataGrid.keyCodes.LEFT_KEY) || (axis === 'column' && this.m_columnHeaderLevelCount === this.m_selectionFrontier['level'] && keyCode == DvtDataGrid.keyCodes.DOWN_KEY) || (axis === 'columnEnd' && this.m_columnEndHeaderLevelCount === this.m_selectionFrontier['level'] && keyCode == DvtDataGrid.keyCodes.UP_KEY)))
    {
        return true;
    }
    else
    {
        return false;
    }
};



/**
 * Get the label of the header
 * @param {string} axis
 * @param {Element} root
 * @param {number} levelCount
 * @param {number} start
 * @param {number} end
 * @param {number} currentIndex
 * @param {number} previousIndex
 * @param {Element} element
 * @returns {string}
 */
DvtDataGrid.prototype._getHeaderLabel = function(axis, root, levelCount, start, end, currentIndex, previousIndex, element)
{
    var columnEndHeader, previousElement;
    if (end != -1 && (currentIndex != previousIndex || this.m_externalFocus))
    {
        columnEndHeader = this.getHeaderFromCell(element, axis);
        if (previousIndex != null)
        {
            previousElement = this._getHeaderByIndex(previousIndex, levelCount - 1, root, levelCount, start);
        }
        return this._getHeaderAndParentIds(columnEndHeader, previousElement);
    }
    return '';
};

/**
 * Get the Id's in a string to put in the accessibility labelledby
 * @param {Element=} header
 * @param {Element=} previousHeader
 * @returns {string}
 */
DvtDataGrid.prototype._getHeaderAndParentIds = function(header, previousHeader)
{
    var i, parents, idString = '', previousParents = [];
    if (header == null)
    {
        // header not rendered
        return '';
    }

    parents = this._getHeaderAndParents(header);
    if (previousHeader != null)
    {
        previousParents = this._getHeaderAndParents(previousHeader);
    }
    for (i = 0; i < parents.length; i++)
    {
        // always add the header that we are focusing
        if (previousParents[i] != parents[i] || i === parents.length - 1)
        {
            idString += (idString == '' ? '':' ') + parents[i]['id'];
        }
    }
    return idString;
};

/**
 * Get the nested headers above the header and including the header.
 * Puts them in an array starting with the outermost.
 * @param {Element} header
 * @returns {Array}
 */
DvtDataGrid.prototype._getHeaderAndParents = function(header)
{
    var axis, level, headerLevels, headers = [header];
    axis = this.getHeaderCellAxis(header);
    level = this.getHeaderCellLevel(header);
    if (axis === 'row')
    {
        headerLevels = this.m_rowHeaderLevelCount;
    }
    else if (axis === 'column')
    {
        headerLevels = this.m_columnHeaderLevelCount;
    }
    else if (axis === 'rowEnd')
    {
        headerLevels = this.m_rowEndHeaderLevelCount;
    }
    else if (axis === 'columnEnd')
    {
        headerLevels = this.m_columnEndHeaderLevelCount;
    }

    if (headerLevels === 1)
    {
        return headers;
    }
    else if (level === headerLevels - 1)
    {
        header = header['parentNode']['firstChild'];
        headers.unshift(header);
        level -= 1;
    }

    while (level > 0)
    {
        header = header['parentNode']['parentNode']['firstChild'];
        headers.unshift(header);
        level -= 1;
    }
    return headers;
};

/**
 * Checks if the input selection Frontier is type header
 * @param {Object} selectionFrontier
 * @returns {boolean} true if type header, false otherwise
 */
DvtDataGrid.prototype.isHeaderSelectionType = function(selectionFrontier)
{
    if (selectionFrontier && selectionFrontier['axis'])
    {
        return true;
    }
    else
    {
        return false;
    }
}

/**
 * Handles arrow keys navigation on cell
 * @param {number} keyCode description
 * @param {boolean} isExtend
 * @param {Event} event the DOM event causing the arrow keys
 * @param {boolean} changeRegions
 * @param {boolean} jumpToHeaders jump to headers if possible
 */
DvtDataGrid.prototype.handleFocusChange = function(keyCode, isExtend, event, changeRegions, jumpToHeaders)
{
    var currentCellIndex, row, column, newCellIndex, focusFunc, header, rowExtent = 1, columnExtent = 1, currentCell, cellContext;

    // ensure that there's no outstanding fetch requests
    if (!this.isFetchComplete())
    {
        //act as if processed to prevent page scrolling before fetch done
        return true;
    }

    if (isExtend)
    {
        currentCellIndex = this.m_selectionFrontier;
        // if extending and selection frontier has an axis component, we are in header realm 
        if (this.isHeaderSelectionType(this.m_selectionFrontier))
        {
            this.handleHeaderFocusChange(keyCode, event, isExtend, jumpToHeaders);
            return;
        }
    }
    else
    {
        currentCellIndex = this.m_active['indexes'];
    }

    if (currentCellIndex == null)
    {
        return;
    }

    if (this.m_trueIndex == null)
    {
        this.m_trueIndex = {};
    }

    if (this.getResources().isRTLMode())
    {
        if (keyCode == DvtDataGrid.keyCodes.LEFT_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.RIGHT_KEY;
        }
        else if (keyCode == DvtDataGrid.keyCodes.RIGHT_KEY)
        {
            keyCode = DvtDataGrid.keyCodes.LEFT_KEY;
        }
    }

    // invoke different function for handling focusing on active cell depending on whether selection is enabled
    focusFunc = this._isSelectionEnabled() ? this.selectAndFocus.bind(this) : this._setActiveByIndex.bind(this);
    row = currentCellIndex['row'];
    column = currentCellIndex['column'];
    currentCell = this._getCellByIndex(currentCellIndex);
    if (currentCell)
    {
        cellContext = currentCell[this.getResources().getMappedAttribute('context')];
        rowExtent = cellContext['extents']['row'];
        columnExtent = cellContext['extents']['column'];
    }

    // navigation to cell using arrow keys.  We are using index instead of dom element
    // because the dom element might not be there in all cases
    switch (keyCode)
    {
        case DvtDataGrid.keyCodes.LEFT_KEY:
            if (!this.m_trueIndex['row'] && !isExtend)
            {
                this.m_trueIndex = {'row': row};
            }
            if (column > 0 && !(jumpToHeaders && this.m_endRowHeader != -1))
            {
                // for left and right key in row selection mode, we'll be only shifting active cell and
                // selection will not be affected
                if (this.m_options.getSelectionMode() == "row")
                {
                    // ensure active cell index is used for row since it might use frontier if extended
                    newCellIndex = this.createIndex(this.m_trueIndex['row'], column - 1);
                    this._setActiveByIndex(newCellIndex, event);
                }
                else
                {
                    if (isExtend)
                    {
                        newCellIndex = this.createIndex(row, column - 1);                        
                        this.extendSelection(newCellIndex, event, keyCode);
                    }
                    else
                    {
                        newCellIndex = this.createIndex(this.m_trueIndex['row'], column - 1);
                        focusFunc(newCellIndex, event);
                    }

                    // announce to screen reader that we have reached first column
                    if (column - 1 === 0)
                    {
                        this._setAccInfoText('accessibleFirstColumn');
                    }
                }
            }
            else if (!isExtend && changeRegions)
            {
                // reached the first column, go to row header if available
                header = this._getHeaderByIndex(this.m_trueIndex['row'], this.m_rowHeaderLevelCount - 1, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
                if (this.m_discontiguousSelection)
                {
                    this.discontiguousHeaderSetActiveFromDatabody(event, 'row', header, this.m_rowHeaderLevelCount);
                }
                else
                {
                    this._setActive(header, {'type':'header', 'index':this.m_trueIndex['row'], 'level': this.m_rowHeaderLevelCount - 1, 'axis': 'row'}, event, true);                    
                }
            }
            break;
        case DvtDataGrid.keyCodes.RIGHT_KEY:
            if (!this.m_trueIndex['row'] && !isExtend)
            {
                this.m_trueIndex = {'row': row};
            }
            // if condition for unknown count and known count cases on whether we have reached the end
            if (!this._isLastColumn(column + columnExtent - 1) && !(jumpToHeaders && this.m_endRowEndHeader != -1))
            {
                // for left and right key in row selection mode, we'll be only shifting active cell and
                // selection will not be affected
                if (this.m_options.getSelectionMode() == "row")
                {
                    // ensure active cell index is used for row since it might use frontier if extended
                    newCellIndex = this.createIndex(this.m_trueIndex['row'], column + columnExtent);
                    this._setActiveByIndex(newCellIndex, event);
                }
                else
                {
                    if (isExtend)
                    {
                        newCellIndex = this.createIndex(row, column + 1);
                        this.extendSelection(newCellIndex, event, keyCode);
                    }
                    else
                    {
                        newCellIndex = this.createIndex(this.m_trueIndex['row'], column + columnExtent);                        
                        focusFunc(newCellIndex, event);
                    }

                    // announce to screen reader that we have reached last column
                    if (this._isLastColumn(column + columnExtent))
                    {
                        this._setAccInfoText('accessibleLastColumn');
                    }
                }
            }
            else if (this.m_endRowEndHeader != -1 && changeRegions)
            {
                // reached the last column, go to row end header if available
                header = this._getHeaderByIndex(this.m_trueIndex['row'], this.m_rowEndHeaderLevelCount - 1, this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
                if (this.m_discontiguousSelection)
                {
                    this.discontiguousHeaderSetActiveFromDatabody(event, 'rowEnd', header, this.m_rowEndHeaderLevelCount);
                }
                else
                {
                    this._setActive(header, {'type':'header', 'index':this.m_trueIndex['row'], 'level': this.m_rowEndHeaderLevelCount - 1, 'axis': 'rowEnd'}, event, true);
                }
            }
            else if (!isExtend)
            {
                // if anchor cell is in the last column, and they arrow right (without Shift), then collapse the range to just the focus cell.  (Matches Excel and intuition.)
                focusFunc(currentCellIndex, event);
            }
            break;
        case DvtDataGrid.keyCodes.UP_KEY:
            if (!this.m_trueIndex['column'] && !isExtend)
            {
                this.m_trueIndex = {'column': column};
            }
            if (row > 0 && !(jumpToHeaders && this.m_endColHeader != -1))
            {
                if (isExtend)
                {
                    newCellIndex = this.createIndex(row - 1, column);                    
                    this.extendSelection(newCellIndex, event, keyCode);
                }
                else
                {
                    newCellIndex = this.createIndex(row - 1, this.m_trueIndex['column']);                    
                    focusFunc(newCellIndex, event);
                }

                // announce to screen reader that we have reached first row
                if (row - 1 === 0)
                {
                    this._setAccInfoText('accessibleFirstRow');
                }
            }
            else
            {
                //if in multiple selection don't clear the selection
                if (!isExtend && changeRegions)
                {
                    // reached the first row, go to column header if available
                    header = this._getHeaderByIndex(this.m_trueIndex['column'], this.m_columnHeaderLevelCount - 1, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
                    if (this.m_discontiguousSelection)
                    {
                        this.discontiguousHeaderSetActiveFromDatabody(event, 'column', header, this.m_columnHeaderLevelCount);
                    }
                    else
                    {
                        this._setActive(header, {'type':'header', 'index':this.m_trueIndex['column'], 'level': this.m_columnHeaderLevelCount - 1, 'axis': 'column'}, event, true);
                    }
                }
            }
            break;
        case DvtDataGrid.keyCodes.DOWN_KEY:
            if (!this.m_trueIndex['column'] && !isExtend)
            {
                this.m_trueIndex = {'column': column};
            }
            if (!this._isLastRow(row + rowExtent - 1) && !(jumpToHeaders && this.m_endColEndHeader != -1))
            {
                if (isExtend)
                {
                    newCellIndex = this.createIndex(row + 1, column);
                    this.extendSelection(newCellIndex, event, keyCode);
                }
                else
                {
                    newCellIndex = this.createIndex(row + rowExtent, this.m_trueIndex['column']);                    
                    focusFunc(newCellIndex, event);
                }

                // announce to screen reader that we have reached last row
                if (this._isLastRow(row + rowExtent))
                {
                    this._setAccInfoText('accessibleLastRow');
                }
            }
            else if (this.m_endColEndHeader != -1 && changeRegions)
            {
                // reached the last column, go to column end header if available
                header = this._getHeaderByIndex(this.m_trueIndex['column'], this.m_columnEndHeaderLevelCount - 1, this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
                if (this.m_discontiguousSelection)
                {
                    this.discontiguousHeaderSetActiveFromDatabody(event, 'columnEnd', header, this.m_columnEndHeaderLevelCount);
                }
                else
                {
                    this._setActive(header, {'type':'header', 'index':this.m_trueIndex['column'], 'level': this.m_columnEndHeaderLevelCount - 1, 'axis': 'columnEnd'}, event, true);
                }
            }
            else if (!isExtend)
            {
                // if anchor cell is in the last row, and they arrow down (without Shift), then collapse the range to just the focus cell.  (Matches Excel and intuition.)
                focusFunc(currentCellIndex, event);
            }
            break;
        case DvtDataGrid.keyCodes.HOME_KEY:
            if (!this.m_trueIndex['row'])
            {
                this.m_trueIndex = {'row': row};
            }
            // selects the first cell of the current row
            newCellIndex = this.createIndex(this.m_trueIndex['row'], 0);
            focusFunc(newCellIndex, event);
            break;
        case DvtDataGrid.keyCodes.END_KEY:
            if (!this.m_trueIndex['row'])
            {
                this.m_trueIndex = {'row': row};
            }
            // selects the last cell of the current row
            if (!this._isCountUnknown("column") && !this._isHighWatermarkScrolling())
            {
                newCellIndex = this.createIndex(this.m_trueIndex['row'], Math.max(0, this.getDataSource().getCount("column") - 1));
            }
            else
            {
                newCellIndex = this.createIndex(this.m_trueIndex['row'], Math.max(0, this.m_endCol));
            }
            focusFunc(newCellIndex, event);
            break;
        case DvtDataGrid.keyCodes.PAGEUP_KEY:
            if (!this.m_trueIndex['column'])
            {
                this.m_trueIndex = {'column': column};
            }        
            // selects the first cell of the current column
            newCellIndex = this.createIndex(0, column);
            focusFunc(newCellIndex, event);
            break;
        case DvtDataGrid.keyCodes.PAGEDOWN_KEY:
            if (!this.m_trueIndex['column'])
            {
                this.m_trueIndex = {'column': column};
            }                
            // selects the last cell of the current column
            if (!this._isCountUnknown("column") && !this._isHighWatermarkScrolling())
            {
                newCellIndex = this.createIndex(Math.max(0, this.getDataSource().getCount("row") - 1), this.m_trueIndex['column']);
            }
            else
            {
                newCellIndex = this.createIndex(Math.max(0, this.m_endRow), this.m_trueIndex['column']);
            }
            focusFunc(newCellIndex, event);
            break;
    }

    return true;
};

/**
 * Scrolls to an  index
 * @param {Object} index - the end index of the selection.
 * @param {boolean|null=} ignoreHighlight - true if we want to ignore highlighting a cell
 * @param {boolean|null=} scrollToOrigin - true if we align the viewport with the origin
 */
DvtDataGrid.prototype.scrollToIndex = function(index, ignoreHighlight, scrollToOrigin)
{
    var row, column, deltaX, deltaY, scrollTop, viewportTop, viewportBottom, dir,
            rowTop, rowHeight, scrollLeft, cell, cellLeft, cellWidth, viewportLeft, viewportRight, scrollRows;
    row = index['row'];
    column = index['column'];
    if (ignoreHighlight)
    {
        this.m_shouldFocus = false;
    }
    if (scrollToOrigin)
    {
        index['scrollToOrigin'] = true;
    }

    dir = this.getResources().isRTLMode() ? "right" : "left";

    deltaX = 0;
    deltaY = 0;
    viewportTop = this._getViewportTop();
    viewportBottom = this._getViewportBottom();
    viewportLeft = this._getViewportLeft();
    viewportRight = this._getViewportRight();

    // check if index is completely outside of rendered
    if (row < this.m_startRow || row > this.m_endRow)
    {
        if (row < this.m_startRow)
        {
            scrollTop = this.m_avgRowHeight * row;
        }
        else
        {
            scrollTop = this.m_avgRowHeight * (row + 1) - viewportBottom + viewportTop;
        }
        deltaY = this.m_currentScrollTop - scrollTop;

        // remember to focus on the row after fetch
        this.m_scrollIndexAfterFetch = index;
        scrollRows = true;
    }
    else
    {
        // it's rendered, find location and scroll to it
        cell = this._getCellByIndex({'row':row, 'column': this.m_startCol});

        rowTop = this.getElementDir(cell, 'top');
        rowHeight = this.getElementHeight(cell);
        // If we are scrolling to a row position, align it to the top row of the viewport
        // if specified
        if (scrollToOrigin || index['scrollToOrigin'])
        {
            deltaY = viewportTop - rowTop;
        }
        else
        {
            if (rowTop + rowHeight > viewportBottom)
            {
                deltaY = viewportBottom - (rowTop + rowHeight);
            }
            else if (rowTop < viewportTop)
            {
                deltaY = viewportTop - rowTop;
            }            
        }

    }

    // if column is defined and it's not already a fetch outside of rendered
    // use scrollRows to know it was not pre-defined
    // if initial Scroll, we should adjust the column
    if (!isNaN(column) && scrollRows != true)
    {
        // check if index is completely outside of rendered
        // approximate scroll position
        if (column < this.m_startCol || column > this.m_endCol)
        {
            if (column < this.m_startCol)
            {
                scrollLeft = this.m_avgColWidth * column;
            }
            else
            {
                scrollLeft = this.m_avgColWidth * (column + 1) - viewportRight + viewportLeft;
            }
            deltaX = this.m_currentScrollLeft - scrollLeft;

            // remember to focus on the cell after fetch
            this.m_scrollIndexAfterFetch = index;
        }
        else
        {
            // it's rendered, find location and scroll to it
            cell = this._getCellByIndex({'row':this.m_startRow, 'column': column});

            cellLeft = this.getElementDir(cell, dir);
            cellWidth = this.getElementWidth(cell);

            if (scrollToOrigin || index['scrollToOrigin'])
            {
                deltaX = viewportLeft - cellLeft;
            }
            else
            {
                if (cellLeft < viewportLeft)
                {
                    deltaX = viewportLeft - cellLeft;
                }
                else if (cellLeft + cellWidth > viewportRight)
                {
                    deltaX = viewportRight - (cellLeft + cellWidth);
                }                
            }
        }
    }

    // scroll if either horiz or vert scroll pos has changed
    if (deltaX != 0 || deltaY != 0)
    {
        cell = this._getCellByIndex(index);

        // this.m_shouldFocus for second call after initial scroll. 
        if (cell != null && ignoreHighlight !== true && this.m_shouldFocus !== false)
        {
            // delay focus on cell until databody has scrolled (by the scroll event handler)
            // if we are not highlighting, ignore this
            this.m_cellToFocus = cell;
        }
        this.scrollDelta(deltaX, deltaY);
    }//if there's an index we wanted to scroll to after fetch it has now been scrolled to by scrollToIndex, so highlight it
    else if (this.m_scrollIndexAfterFetch != null)
    {
        // this.m_shouldFocus for second call after initial scroll. 
        if (!ignoreHighlight && this.m_shouldFocus !== false)
        {
            if (this._setActiveByIndex(this.m_scrollIndexAfterFetch, null, false, false, true))
            {
                this.m_scrollIndexAfterFetch = null;
            }
        }
        else
        {
            this.m_scrollIndexAfterFetch = null;            
        }
    }
};

/**
 * Scrolls to an  index
 * @param {Object} headerInfo
 * @param {string} headerInfo.axis
 * @param {number} headerInfo.index
 * @param {number} headerInfo.level
 */
DvtDataGrid.prototype.scrollToHeader = function(headerInfo)
{
    var delta, startIndex, endIndex, averageDiff, currentScroll, newScroll, headerMin, headerDiff,
            header, viewportMin, viewportMax, viewportDiff, axis, index, level;
    axis = headerInfo['axis'];
    index = headerInfo['index'];
    level = headerInfo['level'];
    delta = 0;

    if (axis === 'row')
    {
        startIndex = this.m_startRowHeader;
        endIndex = this.m_endRowHeader;
        averageDiff = this.m_avgRowHeight;
        currentScroll = this.m_currentScrollTop;
        viewportMin = this._getViewportTop();
        viewportMax = this._getViewportBottom();
    }
    else if (axis === 'column')
    {
        startIndex = this.m_startColHeader;
        endIndex = this.m_endColHeader;
        averageDiff = this.m_avgColWidth;
        currentScroll = this.m_currentScrollLeft;
        viewportMin = this._getViewportLeft();
        viewportMax = this._getViewportRight();
    }
    else if (axis === 'rowEnd')
    {
        startIndex = this.m_startRowEndHeader;
        endIndex = this.m_endRowEndHeader;
        averageDiff = this.m_avgRowHeight;
        currentScroll = this.m_currentScrollTop;
        viewportMin = this._getViewportTop();
        viewportMax = this._getViewportBottom();
    }
    else if (axis === 'columnEnd')
    {
        startIndex = this.m_startColEndHeader;
        endIndex = this.m_endColEndHeader;
        averageDiff = this.m_avgColWidth;
        currentScroll = this.m_currentScrollLeft;
        viewportMin = this._getViewportLeft();
        viewportMax = this._getViewportRight();
    }

    viewportDiff = viewportMax - viewportMin;

    // check if index is completely outside of rendered
    if (index < startIndex || index > endIndex)
    {
        if (index < startIndex)
        {
            newScroll = averageDiff * index;
        }
        else
        {
            newScroll = averageDiff * (index + 1) - viewportDiff;
        }
        delta = currentScroll - newScroll;

        // remember to focus on the row after fetch
        this.m_scrollHeaderAfterFetch = headerInfo;
    }
    else
    {
        if (axis === 'row' || axis === 'rowEnd')
        {
            if (axis === 'row')
            {
                header = this._getHeaderByIndex(index, level, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
            }
            else
            {
                header = this._getHeaderByIndex(index, level, this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
            }
            headerMin = this.getElementDir(header, 'top');
            headerDiff = this.getElementHeight(header);
        }
        else if (axis === 'column' || axis === 'columnEnd')
        {
            if (axis === 'column')
            {
                header = this._getHeaderByIndex(index, level, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
            }
            else
            {
                header = this._getHeaderByIndex(index, level, this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
            }
            headerMin = this.getElementDir(header, this.getResources().isRTLMode() ? "right" : "left");
            headerDiff = this.getElementWidth(header);
        }

        if (viewportDiff > headerDiff)
        {
            if (headerMin + headerDiff > viewportMax)
            {
                delta = viewportMax - (headerMin + headerDiff);
            }
            else if (headerMin < viewportMin)
            {
                delta = viewportMin - headerMin;
            }
        }
        else
        {
            delta = viewportMin - headerMin;
        }
    }

    // scroll if either horiz or vert scroll pos has changed
    if (delta != 0)
    {
        if (header != null && this.m_shouldFocus !== false)
        {
            // delay focus on cell until databody has scrolled (by the scroll event handler)
            this.m_cellToFocus = header;
        }
        (axis === 'row' || axis === 'rowEnd') ? this.scrollDelta(0, delta) : this.scrollDelta(delta, 0);
    }
    //if there's an index we wanted to sctoll to after fetch it has now been scrolled to by scrollToIndex, so highlight it
    else if (this.m_scrollHeaderAfterFetch != null)
    {
        this._updateActive(headerInfo, true, true);
        this.m_scrollHeaderAfterFetch = null;
    }
};

/**
 * Locate the header element.  Look up recursively from its parent if neccessary.
 * @param {Element|undefined|null} elem the starting point to locate the header element
 * @param {string=} headerCellClassName the name of the header cell class name
 * @param {string=} endHeaderCellClassName the name of the header cell class name
 * @return {Element|null|undefined} the header element
 * @private
 */
DvtDataGrid.prototype.findHeader = function(elem, headerCellClassName, endHeaderCellClassName)
{
    if (headerCellClassName == null)
    {
        headerCellClassName = this.getMappedStyle("headercell");
    }

    if (endHeaderCellClassName == null)
    {
        endHeaderCellClassName = this.getMappedStyle("endheadercell");
    }

    if (headerCellClassName != null)
    {
        if (this.m_utils.containsCSSClassName(elem, headerCellClassName) ||
            this.m_utils.containsCSSClassName(elem, endHeaderCellClassName))
        {
            // found header element
            return elem;
        }
        else if (elem['parentNode'])
        {
            // recursive call with parent node
            return this.findHeader(elem['parentNode'], headerCellClassName, endHeaderCellClassName);
        }
        else if (elem === this.m_root)
        {
            // short circuit to terminal when root is reached
            return null;
        }
    }

    // all other case returns null
    return null;
};

/**
 * Ensures row banding is set on the proper rows
 * @private
 */
DvtDataGrid.prototype.updateRowBanding = function()
{
    var rowBandingInterval, cells, i, index, bandingClass, cell;
    rowBandingInterval = this.m_options.getRowBandingInterval();
    if (rowBandingInterval > 0)
    {
        cells = this.m_databody['firstChild']['childNodes'];
        bandingClass = this.getMappedStyle("banded");
        for (i = 0; i < cells.length; i++)
        {
            cell = cells[i];
            index = this._getIndex(cell, 'row');
            if ((Math.floor(index / rowBandingInterval) % 2 === 1))
            {
                if (!this.m_utils.containsCSSClassName(cell, bandingClass))
                {
                    this.m_utils.addCSSClassName(cell, bandingClass);
                }
            }
            else
            {
                if (this.m_utils.containsCSSClassName(cell, bandingClass))
                {
                    this.m_utils.removeCSSClassName(cell, bandingClass);
                }
            }
        }
    }
};

/**
 * Ensures column banding is set on the proper rows
 * @private
 */
DvtDataGrid.prototype.updateColumnBanding = function()
{
    var columnBandingInterval, cells, i, index, bandingClass, cell;
    columnBandingInterval = this.m_options.getColumnBandingInterval();
    if (columnBandingInterval > 0)
    {
        cells = this.m_databody['firstChild']['childNodes'];
        bandingClass = this.getMappedStyle("banded");
        for (i = 0; i < cells.length; i += 1)
        {
            cell = cells[i];
            index = this._getIndex(cell, 'column');
            if ((Math.floor(index / columnBandingInterval) % 2 === 1))
            {
                if (!this.m_utils.containsCSSClassName(cell, bandingClass))
                {
                    this.m_utils.addCSSClassName(cell, bandingClass);
                }
            }
            else
            {
                if (this.m_utils.containsCSSClassName(cell, bandingClass))
                {
                    this.m_utils.removeCSSClassName(cell, bandingClass);
                }
            }
        }
    }
};

/**
 * Remove banding (both row and column)
 * @private
 */
DvtDataGrid.prototype._removeBanding = function()
{
    var cells, i, bandingClass;
    cells = this.m_databody['firstChild']['childNodes'];
    bandingClass = this.getMappedStyle("banded");
    for (i = 0; i < cells.length; i++)
    {
        if (this.m_utils.containsCSSClassName(cells[i], bandingClass))
        {
            this.m_utils.removeCSSClassName(cells[i], bandingClass);
        }
    }
};

/**
 * Sets the accessibility status text
 * @param {string} key the message key
 * @param {Object|Array|null=} args to pass into the translator
 * @private
 */
DvtDataGrid.prototype._setAccInfoText = function(key, args)
{
    var text = this.getResources().getTranslatedText(key, args);
    if (text != null)
    {
        this.m_accInfo.textContent = text;
    }
};

/**
 * Handles expand event from the flattened datasource.
 * @param {Object} event the expand event
 * @param {boolean} fromQueue whether this is invoked from processing the model event queue, optional.
 * @private
 */
DvtDataGrid.prototype.handleExpandEvent = function(event, fromQueue)
{
    var rowCells, rowKey, i;

    if (fromQueue === undefined && this.queueModelEvent(event))
    {
        // tag the event for discovery later
        event['operation'] = 'expand';
        return;
    }

    //rowKey = event['rowKey'];
    //rowCells = this._getAxisCellsByKey(rowKey, 'row');
    //for (i = 0; i < rowCells.length; i++)
    //{
    //    rowCells[i].setAttribute("aria-expanded", true);
    //}

    // update screen reader alert
    this._setAccInfoText("accessibleRowExpanded");
    this.populateAccInfo();
    if (fromQueue) 
    {
        this._runModelEventQueue();
    }
};

/**
 * Handles collapse event from the flattened datasource.
 * @param {Object} event the collapse event
 * @param {boolean} fromQueue whether this is invoked from processing the model event queue, optional.
 * @private
 */
DvtDataGrid.prototype.handleCollapseEvent = function(event, fromQueue)
{
    var rowCells, rowKey, i;

    if (fromQueue === undefined && this.queueModelEvent(event))
    {
        // tag the event for discovery later
        event['operation'] = 'collapse';
        return;
    }

    //rowKey = event['rowKey'];
    //rowCells = this._getAxisCellsByKey(rowKey, 'row');
    //for (i = 0; i < rowCells.length; i++)
    //{
    //    rowCells[i].setAttribute("aria-expanded", false);
    //}

    // update screen reader alert
    this._setAccInfoText("accessibleRowCollapsed");
    this.populateAccInfo();
    if (fromQueue) 
    {
        this._runModelEventQueue();
    }
};

/**
 * Retrieve the key from an element.
 * @param {Element|Node|undefined} element the element to retrieve the key from.
 * @param {string=} axis
 * @return {string|null} the key of the element
 * @private
 */
DvtDataGrid.prototype._getKey = function(element, axis)
{
    // make sure the element has a context
    if (element != null && element[this.getResources().getMappedAttribute("context")])
    {
        if (axis != null && this.m_utils.containsCSSClassName(element, this.getMappedStyle('cell')))
        {
            return element[this.getResources().getMappedAttribute("context")]['keys'][axis];
        }
        return element[this.getResources().getMappedAttribute("context")]['key'];
    }
    return null;
};

/**
 * Retrieve the active axis key.
 * @param {string} axis
 * @param {boolean=} prev if we want the previous row key instead
 * @return {string|null} the key of the active row
 * @private
 */
DvtDataGrid.prototype._getActiveKey = function(axis, prev)
{
    if (prev && this.m_prevActive != null)
    {
        if (this.m_prevActive['type'] == 'header' && (this.m_prevActive['axis'] === axis || this.m_prevActive['axis'] === axis + 'End'))
        {
            return this.m_prevActive['key'];
        }
        else if (this.m_prevActive['type'] == 'cell')
        {
            return this.m_prevActive['keys'][axis];
        }
    }
    else if (this.m_active != null)
    {
        if (this.m_active['type'] == 'header' && (this.m_active['axis'] === axis || this.m_active['axis'] === axis + 'End'))
        {
            return this.m_active['key'];
        }
        else if (this.m_active['type'] == 'cell')
        {
            return this.m_active['keys'][axis];
        }
    }
    return null;
};

/**
 * Retrieve the active index.
 * @param {string} axis
 * @param {boolean=} prev if we want the previous index instead
 * @return {string|null} the index of the active axis
 * @private
 */
DvtDataGrid.prototype._getActiveIndex = function(axis, prev)
{
    if (prev && this.m_prevActive != null)
    {
        if (this.m_prevActive['type'] == 'header' && (this.m_prevActive['axis'] === axis || this.m_prevActive['axis'] === axis + 'End'))
        {
            return this.m_prevActive['index'];
        }
        else if (this.m_prevActive['type'] == 'cell')
        {
            return this.m_prevActive['indexes'][axis];
        }
    }
    else if (this.m_active != null)
    {
        if (this.m_active['type'] == 'header' && (this.m_active['axis'] === axis || this.m_active['axis'] === axis + 'End'))
        {
            return this.m_active['index'];
        }
        else if (this.m_active['type'] == 'cell')
        {
            return this.m_active['indexes'][axis];
        }
    }
    return null;
};

///////////////////// move methods////////////////////////
/**
 * Handles cut event from the flattened datasource.
 * @param {Event} event the cut event
 * @param {Element=} target the target element
 * @return {boolean} true if the event was processed here
 * @private
 */
DvtDataGrid.prototype._handleCut = function(event, target)
{
    var rowKey, cell, i;
    if (target == null)
    {
        target = /** @type {Element} */ (event.target);
    }
    cell = this.findCellOrHeader(target);

    if (this._isMoveOnElementEnabled(cell))
    {
        if (this.m_cutCells != null)
        {
            for (i = 0; i < this.m_cutCells.length; i++)
            {
                this.m_utils.removeCSSClassName(this.m_cutCells[i], this.getMappedStyle('cut'));
            }
        }

        rowKey = this._getKey(cell, 'row');
        //cut row header with row
        this.m_cutCells = this._getAxisCellsByKey(rowKey, 'row');
        this.m_cutRowHeader = this._findHeaderByKey(rowKey, this.m_rowHeader, this.getMappedStyle('rowheadercell'));
        this.m_cutRowEndHeader = this._findHeaderByKey(rowKey, this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'));

        this._highlightCellsAlongAxis(rowKey, 'row', 'key', 'add', ['cut']);
        if (this.m_cutRowHeader !== null)
        {
            this.m_utils.addCSSClassName(this.m_cutRowHeader, this.getMappedStyle('cut'));
        }
        if (this.m_cutRowEndHeader !== null)
        {
            this.m_utils.addCSSClassName(this.m_cutRowEndHeader, this.getMappedStyle('cut'));
        }

        return true;
    }
    return false;
};

/**
 * Handles cut event from the flattened datasource.
 * @param {Event} event the cut event
 * @param {Element=} target the target element
 *
 * @private
 */
DvtDataGrid.prototype._handlePaste = function(event, target)
{
    var pasteRowKey, cutRowKey, i;
    if (target == null)
    {
        target = /** @type {Element} */ (event.target);
    }
    if (this.m_cutCells != null)
    {
        for (i = 0; i < this.m_cutCells.length; i++)
        {
            this.m_utils.removeCSSClassName(this.m_cutCells[i], this.getMappedStyle('cut'));
        }

        if (this.m_cutRowHeader !== null)
        {
            //remove css from row header too
            this.m_utils.removeCSSClassName(this.m_cutRowHeader, this.getMappedStyle('cut'));
            this.m_cutRowHeader = null;
        }
        if (this.m_cutRowEndHeader !== null)
        {
            //remove css from row header too
            this.m_utils.removeCSSClassName(this.m_cutRowEndHeader, this.getMappedStyle('cut'));
            this.m_cutRowEndHeader = null;
        }

        pasteRowKey = this._getKey(this.findCellOrHeader(target), 'row');
        cutRowKey = this._getKey(this.m_cutCells[0], 'row');
        if (cutRowKey !== pasteRowKey)
        {
            if (this._isSelectionEnabled())
            {
                // unhighlight and clear selection
                this._clearSelection(event);
            }
            if (this._isDatabodyCellActive())
            {
                this._unhighlightActive();
            }
            this.m_moveActive = true;
            this.getDataSource().move(cutRowKey, pasteRowKey);
        }
        this.m_cutCells = null;
        return true;
    }
    return false;
};

/**
 * Handles canceling a reorder
 * @param {Object} event the cut event
 * @param {Element=} target the target element
 *
 * @private
 */
DvtDataGrid.prototype._handleCancelReorder = function(event, target)
{
    var i;
    if (this.m_cutCells != null)
    {
        for (i = 0; i < this.m_cutCells.length; i++)
        {
            this.m_utils.removeCSSClassName(this.m_cutCells[i], this.getMappedStyle('cut'));
        }
        this.m_cutCells = null;

        if (this.m_cutRowHeader !== null)
        {
            this.m_utils.removeCSSClassName(this.m_cutRowHeader, this.getMappedStyle('cut'));
            this.m_cutRowHeader = null;
        }
        if (this.m_cutRowEndHeader !== null)
        {
            this.m_utils.removeCSSClassName(this.m_cutRowEndHeader, this.getMappedStyle('cut'));
            this.m_cutRowEndHeader = null;
        }
        return true;
    }
};

/**
 * Handles cut event from the flattened datasource.
 * @param {Object} event the cut event
 * @private
 */
DvtDataGrid.prototype._handleMove = function(event)
{
    var deltaY, height, cell, rowKey, target;

    //initialize the move
    if (this.m_moveCells == null)
    {
        target = /** @type {Element} */ (event.target);
        cell = this.findCellOrHeader(target);

        //get the move row key to set the move row/rowHeader
        rowKey = this._getKey(cell, 'row');
        this.m_originalMoveIndex = this._getIndex(cell, 'row');
        this.m_moveIndex = /** @type {number} */ (this.m_originalMoveIndex);

        this.m_moveCells = this._getAxisCellsByIndex(this.m_moveIndex, 'row');
        this.m_moveRowHeader = this._findHeaderByKey(rowKey, this.m_rowHeader, this.getMappedStyle('rowheadercell'));
        this.m_moveRowEndHeader = this._findHeaderByKey(rowKey, this.m_rowEndHeader, this.getMappedStyle('rowendheadercell'));

        //add the move style class to the css
        this._highlightCellsAlongAxis(this.m_moveIndex, 'row', 'index', 'add', ['drag']);

        this.m_originalTop = this.getElementDir(this.m_moveCells[0], 'top');

        this.m_dropTarget = document.createElement("div");
        this.m_utils.addCSSClassName(this.m_dropTarget, this.getMappedStyle('drop'));
        this.setElementHeight(this.m_dropTarget, this.calculateRowHeight(this.m_moveCells[0]));
        this.setElementDir(this.m_dropTarget, this.m_originalTop, 'top');
        this.m_databody['firstChild'].appendChild(this.m_dropTarget); //@HTMLUpdateOK

        this._addHeaderDropTarget(this.m_moveRowHeader, this.m_rowHeader, false);
        this._addHeaderDropTarget(this.m_moveRowEndHeader, this.m_rowEndHeader, true);
    }

    //calculate the change in Y direction
    if (!this.m_utils.isTouchDevice())
    {
        this.m_prevY = this.m_currentY;
        this.m_currentY = event['pageY'];
    }
    deltaY = this.m_currentY - this.m_prevY;
    height = this.calculateRowHeight(this.m_moveCells[0]);

    //adjust the top height of the moveRow and moveRowHeader
    for (var i = 0; i < this.m_moveCells.length; i++)
    {
        this.setElementDir(this.m_moveCells[i], (this.getElementDir(this.m_moveCells[i], 'top') + deltaY), 'top');
    }
    if (this.m_moveRowHeader !== null)
    {
        this.setElementDir(this.m_moveRowHeader, (this.getElementDir(this.m_moveRowHeader, 'top') + deltaY), 'top');
    }
    if (this.m_moveRowEndHeader !== null)
    {
        this.setElementDir(this.m_moveRowEndHeader, (this.getElementDir(this.m_moveRowEndHeader, 'top') + deltaY), 'top');
    }

    var nextSiblingIndex = this.m_moveIndex + 1;
    var previousSiblingIndex = this.m_moveIndex - 1;
    var nextSibling = this._getCellByIndex(this.createIndex(nextSiblingIndex, this.m_startCol));
    var previousSibling = this._getCellByIndex(this.createIndex(previousSiblingIndex, this.m_startCol));

    // see if the element has crossed the halfway point of the next row
    if (nextSibling != null && (this.getElementDir(nextSibling, 'top') < (this.getElementDir(this.m_moveCells[0], 'top') + (height / 2))))
    {
        this._moveDropRows('nextSibling', nextSiblingIndex);
    }
    else if (previousSibling != null && (this.getElementDir(previousSibling, 'top') > (this.getElementDir(this.m_moveCells[0], 'top') - (height / 2))))
    {
        this._moveDropRows('previousSibling', previousSiblingIndex);
    }
};

/**
 * Add drop header target
 * @param {Element|null|undefined} moveHeader
 * @param {boolean} isEnd
 * @private
 */
DvtDataGrid.prototype._addHeaderDropTarget = function(moveHeader, root, isEnd)
{
    var dropTarget;
    if (moveHeader !== null)
    {
        //need to store the height inline if not already because top values will be changing
        if (moveHeader['style']['height'] == null)
        {
            this.setElementHeight(moveHeader, this.calculateRowHeight(moveHeader));
        }
        this.m_utils.addCSSClassName(moveHeader, this.getMappedStyle('drag'));
        dropTarget = document.createElement("div");
        this.m_utils.addCSSClassName(dropTarget, this.getMappedStyle('drop'));
        this.setElementHeight(dropTarget, this.calculateRowHeight(moveHeader));
        this.setElementDir(dropTarget, this.m_originalTop, 'top');
        root['firstChild'].appendChild(dropTarget); //@HTMLUpdateOK

        if (isEnd)
        {
            this.m_dropTargetEndHeader = dropTarget;
        }
        else
        {
            this.m_dropTargetHeader = dropTarget;
        }
    }
};

/**
 * Determined if move is supported for the specified axis.
 * @param {string} sibling nextSibling/previosusSibling
 * @private
 */
DvtDataGrid.prototype._moveDropRows = function(sibling, index)
{
    var newTop, newSiblingTop, headerScroller, endHeaderScroller, siblingCells;

    //move the drop target and the adjacent row
    if (sibling == 'nextSibling')
    {
        siblingCells = this._getAxisCellsByIndex(index, 'row');
        newTop = this.m_originalTop + this.calculateRowHeight(siblingCells[0]);
        newSiblingTop = this.m_originalTop;
    }
    else
    {
        siblingCells = this._getAxisCellsByIndex(index, 'row');
        newTop = this.getElementDir(siblingCells[0], 'top');
        newSiblingTop = newTop + this.calculateRowHeight(siblingCells[0]);
    }

    this.setElementDir(this.m_dropTarget, newTop, 'top');

    for (var i = 0; i < siblingCells.length; i++)
    {
        this.setElementDir(siblingCells[i], newSiblingTop, 'top');
    }

    if (this.m_moveRowHeader !== null)
    {
        headerScroller = this.m_moveRowHeader['parentNode'];
        this.setElementDir(this.m_dropTargetHeader, newTop, 'top');
        this.setElementDir(this.m_moveRowHeader[sibling], newSiblingTop, 'top');
    }
    if (this.m_moveRowEndHeader !== null)
    {
        endHeaderScroller = this.m_moveRowEndHeader['parentNode'];
        this.setElementDir(this.m_dropTargetEndHeader, newTop, 'top');
        this.setElementDir(this.m_moveRowEndHeader[sibling], newSiblingTop, 'top');
    }

    //store the new top value
    this.m_originalTop = newTop;

    this._highlightCellsAlongAxis(this.m_moveIndex + 1, 'row', 'index', 'remove', ['activedrop']);

    //move the moveRow and rowHeader so we can continue to pull the adjacent header
    if (sibling === 'nextSibling')
    {
        this._modifyAxisCellContextIndex('row', this.m_moveIndex, 1, 1);
        this._modifyAxisCellContextIndex('row', this.m_moveIndex + 1, 1, -1);
        this.m_moveIndex += 1;

        if (this.m_moveRowHeader !== null)
        {
            headerScroller.insertBefore(this.m_moveRowHeader, this.m_moveRowHeader[sibling][sibling]); //@HTMLUpdateOK
        }
        if (this.m_moveRowEndHeader !== null)
        {
            endHeaderScroller.insertBefore(this.m_moveRowEndHeader, this.m_moveRowEndHeader[sibling][sibling]); //@HTMLUpdateOK
        }
    }
    else
    {
        this._modifyAxisCellContextIndex('row', this.m_moveIndex, 1, -1);
        this._modifyAxisCellContextIndex('row', this.m_moveIndex - 1, 1, 1);
        this.m_moveIndex -= 1;

        if (this.m_moveRowHeader !== null)
        {
            headerScroller.insertBefore(this.m_moveRowHeader, this.m_moveRowHeader[sibling]); //@HTMLUpdateOK
        }
        if (this.m_moveRowEndHeader !== null)
        {
            endHeaderScroller.insertBefore(this.m_moveRowEndHeader, this.m_moveRowEndHeader[sibling]); //@HTMLUpdateOK
        }
    }

    this._refreshDatabodyMap();

    this._highlightCellsAlongAxis(this.m_moveIndex + 1, 'row', 'index', 'add', ['activedrop']);
};

/**
 * Determined if move is supported for the specified axis.
 * @param {string} axis the axis which we check whether move is supported.
 * @private
 */
DvtDataGrid.prototype._isMoveEnabled = function(axis)
{
    var capability, moveable;
    capability = this.getDataSource().getCapability("move");
    moveable = this.m_options.isMoveable('row');
    if (moveable === "enable" && (capability === "full" || capability === axis))
    {
        return true;
    }

    return false;
};

/**
 * Handles a mouse up after move
 * @param {Event} event MouseUp Event
 * @param {boolean} validUp true if in the databody or rowHeader
 * @private
 */
DvtDataGrid.prototype._handleMoveMouseUp = function(event, validUp)
{
    if (this.m_moveCells != null)
    {
        //remove the the drop target div from the databody/rowHeader
        this._remove(this.m_dropTarget);
        if (this.m_moveRowHeader !== null)
        {
            this._remove(this.m_dropTargetHeader);
        }
        if (this.m_moveRowEndHeader !== null)
        {
            this._remove(this.m_dropTargetEndHeader);
        }
        if (this.m_active != null && this.m_active['axis'] != 'column')
        {
            this.m_moveActive = true;
        }

        //clear selection
        if (this._isSelectionEnabled())
        {
            // unhighlight and clear selection
            this._clearSelection(event);
        }

        var moveCell = this.m_moveCells[0];
        var moveCellKey = this._getKey(moveCell, 'row');

        //if the mousup was in the rowHeader or databody
        if (validUp == true)
        {
            var insertIndex = this.m_moveIndex + 1;
            var insertKey = this._getKey(this._getCellByIndex(this.createIndex(insertIndex, this.m_startCol)), 'row');
            this.getDataSource().move(moveCellKey, insertIndex === null ? null : insertKey);
        }
        else
        {
            this.getDataSource().move(moveCellKey, moveCellKey);
        }
        this.m_moveCells = null;
        this.m_originalMoveIndex = null;
        this.m_moveIndex = null;
    }
    this.m_databodyMove = false;
};

/**
 * Check if a row can be moved, meaning it is the active row and move is enabled
 * @param {Element|null|undefined} cell the row to move
 * @returns {boolean} true if the row can be moved
 */
DvtDataGrid.prototype._isMoveOnElementEnabled = function(cell)
{
    if (cell != null && this._isMoveEnabled('row'))
    {
        if (this._getActiveKey('row') === this._getKey(cell, 'row'))
        {
            return true;
        }
    }
    return false;
};

/**
 * Applies the draggable class to the new active row and row header, removes it if the active has changed
 */
DvtDataGrid.prototype._manageMoveCursor = function()
{
    var className, activeKey, prevActiveKey, activeRowHeader, prevActiveRowHeader, activeRowEndHeader,
        rowHederStyle, rowEndHeaderStyle, activeCells;

    activeKey = this._getActiveKey('row');
    prevActiveKey = this._getActiveKey('row', true);

    className = this.getMappedStyle('draggable');
    rowHederStyle = this.getMappedStyle('rowheadercell');
    rowEndHeaderStyle = this.getMappedStyle('rowendheadercell');

    if (prevActiveKey != null)
    {
        this._highlightCellsAlongAxis(prevActiveKey, 'row', 'key', 'remove', ['draggable']);

        prevActiveRowHeader = this._findHeaderByKey(prevActiveKey, this.m_rowHeader, rowHederStyle);
        if (this.m_utils.containsCSSClassName(prevActiveRowHeader, className))
        {
            this.m_utils.removeCSSClassName(prevActiveRowHeader, className);
        }

        prevActiveRowHeader = this._findHeaderByKey(prevActiveKey, this.m_rowEndHeader, rowEndHeaderStyle);
        if (this.m_utils.containsCSSClassName(prevActiveRowHeader, className))
        {
            this.m_utils.removeCSSClassName(prevActiveRowHeader, className);
        }
    }

    if (activeKey != null)
    {
        activeCells = this._getAxisCellsByKey(activeKey, 'row');
        //if move enabled and draggable class name
        if (this._isMoveOnElementEnabled(activeCells[0]))
        {
            this._highlightCellsAlongAxis(activeKey, 'row', 'key', 'add', ['draggable']);

            activeRowHeader = this._findHeaderByKey(activeKey, this.m_rowHeader, rowHederStyle);
            this.m_utils.addCSSClassName(activeRowHeader, className);

            activeRowEndHeader = this._findHeaderByKey(activeKey, this.m_rowEndHeader, rowEndHeaderStyle);
            this.m_utils.addCSSClassName(activeRowEndHeader, className);
        }
    }
};

/**
 * Handles focus on the root and its children by setting focus class on the root
 * @param {Event} event
 */
DvtDataGrid.prototype.handleRootFocus = function(event)
{
    var newCellIndex, selection;
    this.m_utils.addCSSClassName(this.m_root, this.getMappedStyle('focus'));

    // if nothing is active, and came from the outside of the datagrid, activate first cell
    if (!this.m_root.contains(document.activeElement) || (document.activeElement === this.m_root && this.m_root.tabIndex == 0)
            || (document.activeElement === this.m_databody && this.m_scrollbarFocus && this.m_root.tabIndex == 0))
    {
        this.m_externalFocus = true;

        if (this._isCellEditable())
        {
            this._setAccInfoText('accessibleEditableMode');
        }
        else if (this._isGridEditable())
        {
            this._setAccInfoText('accessibleNavigationMode');
        }

        if (this.m_active == null && !this._databodyEmpty())
        {
            newCellIndex = this.createIndex(0, 0);

            if (this.m_scrollbarFocus == true)
            {
                this.m_shouldFocus = false;
                this.m_scrollbarFocus = false;
            }
            else
            {
                // make sure it's visible
                this.scrollToIndex(newCellIndex);
            }

            selection = this.GetSelection();
            // select or focus it
            if (this._isSelectionEnabled() && (selection == null || selection.length == 0))
            {
                this.selectAndFocus(newCellIndex, event);
            }
            else
            {
                this._setActiveByIndex(newCellIndex);
            }
        }
        else if (this.m_active != null)
        {
            this._highlightActive();
        }
    }
    this.m_root.tabIndex = -1;
};

/**
 * Handles blur on the root and its children by removing focus class on the root
 * @param {Event} event
 */
DvtDataGrid.prototype.handleRootBlur = function(event)
{
    var active;
    // There is no cross-browser way to tell if the whole grid is out of focus on blur today.
    // document.activeElement returns null in chrome and firefox on blur events.
    // relatedTarget doesn't return a value in firefox and IE though there a tickets to fix.
    // We could implement a non-timeout solution that exiting and re-entering
    // the grid via tab key would not read the summary text upon re-entry (initial would work)
    setTimeout(function(){
        if (!this.m_root.contains(document.activeElement))
        {
            this.m_root.tabIndex = 0;
            active = this._getActiveElement();
            if (active != null)
            {
                this._unsetAriaProperties(active);
            }
        }
    }.bind(this), 100);

    //don't change the color on move
    if (this.m_moveRow == null)
    {
        this.m_utils.removeCSSClassName(this.m_root, this.getMappedStyle('focus'));
    }
};

/**
 * Calculate the a row's height using top or endRowPixel
 * @param {Element|undefined|null} row the row to calculate height on
 * @return {number} the row height
 */
DvtDataGrid.prototype.calculateRowHeight = function(row)
{
    if (row['style']['height'] != '')
    {
        return this.getElementHeight(row);
    }
    if (row['nextSibling'] != null)
    {
        return this.getElementDir(row['nextSibling'], 'top') - this.getElementDir(row, 'top');
    }
    return this.m_endRowPixel - this.getElementDir(row, 'top');
};

/**
 * Calculate the a row headers's height using top or endRowHeaderPixel
 * @param {Element|undefined|null} rowHeader the rowHeader to calculate height on
 * @return {number} the rowHeader height
 */
DvtDataGrid.prototype.calculateRowHeaderHeight = function(rowHeader)
{
    if (rowHeader['style']['height'] != '')
    {
        return this.getElementHeight(rowHeader);
    }
    if (rowHeader['nextSibling'] != null)
    {
        return this.getElementDir(rowHeader['nextSibling'], 'top') - this.getElementDir(rowHeader, 'top');
    }
    return this.m_endRowHeaderPixel - this.getElementDir(rowHeader, 'top');
};

/**
 * Calculate the a column's width using left/right or endColumnPixel
 * @param {Element|undefined|null} cell the cell to calculate width on
 * @return {number} the cell width
 */
DvtDataGrid.prototype.calculateColumnWidth = function(cell)
{
    if (cell['style']['width'] != '')
    {
        return this.getElementWidth(cell);
    }
    var dir = this.getResources().isRTLMode() ? "right" : "left";
    if (cell['nextSibling'] != null)
    {
        return this.getElementDir(cell['nextSibling'], dir) - this.getElementDir(cell, dir);
    }
    return this.m_endColPixel - this.getElementDir(cell, dir);
};

/**
 * Calculate the a column headers's width using left/right or endColumnHeaderPixel
 * @param {Element|undefined|null} columnHeader the columnHeader to calculate width on
 * @return {number} the columnHeader width
 */
DvtDataGrid.prototype.calculateColumnHeaderWidth = function(columnHeader)
{
    if (columnHeader['style']['width'] != '')
    {
        return this.getElementWidth(columnHeader);
    }
    var dir = this.getResources().isRTLMode() ? "right" : "left";
    if (columnHeader['nextSibling'] != null)
    {
        return this.getElementDir(columnHeader['nextSibling'], dir) - this.getElementDir(columnHeader, dir);
    }
    return this.m_endColHeaderPixel - this.getElementDir(columnHeader, dir);
};

/**
 * @return {boolean} true if the databody is empty
 */
DvtDataGrid.prototype._databodyEmpty = function()
{
    if (this.m_databody['firstChild'] == null || this.m_databody['firstChild']['firstChild'] == null)
    {
        return true;
    }
    return false;
};

/**
 * Change or add CSS property of element
 * @param {Element} target the element to which css property will be added
 * @param {string|undefined} prop the style property name
 * @param {string|number|null} value the value of css property
 * @param {string=} action the flag variable if it is require to remove css property
 * @private
 */
DvtDataGrid.prototype.changeStyleProperty = function(target, prop, value, action)
{
    if (typeof prop != "undefined")
    {
        target.style[prop] = (action == "remove") ? "" : value;
    }
};

/**
 * Add set of required animation rules to the element
 * @param {Element} target the element to which animation rules will be added
 * @param {number|string} duration the duration of animation
 * @param {number|string} delay the delay of animation
 * @param {string} timing the easing function
 * @param {number|string} x the final position (in pixels) of the current animation
 * @param {number|string} y the final position (in pixels) of the current animation
 * @param {number|string} z the final position (in pixels) of the current animation
 * @private
 */
DvtDataGrid.prototype.addTransformMoveStyle = function(target, duration, delay, timing, x, y, z)
{
    this.changeStyleProperty(target, this.getCssSupport('transition-delay'), delay);
    this.changeStyleProperty(target, this.getCssSupport('transition-timing-function'), timing);
    this.changeStyleProperty(target, this.getCssSupport('transition-duration'), duration);
    this.changeStyleProperty(target, this.getCssSupport('transform'), 'translate3d(' + x + 'px,' + y + 'px,' + z + 'px)');
};

/**
 * Add set of required animation rules to the element
 * @param {Element} target the element to which animation rules will be added
 * @private
 */
DvtDataGrid.prototype.removeTransformMoveStyle = function(target)
{
    this.changeStyleProperty(target, this.getCssSupport('transition-delay'), null, 'remove');
    this.changeStyleProperty(target, this.getCssSupport('transition-timing-function'), null, 'remove');
    this.changeStyleProperty(target, this.getCssSupport('transition-duration'), null, 'remove');
    this.changeStyleProperty(target, this.getCssSupport('transform'), null, 'remove');
};

/**
 * Check if CSS property is supported by appropriate vendors
 * @param {string} cssprop css property
 * @return {string|undefined} css property with appropiate vendor's prefix
 * @private
 */
DvtDataGrid.prototype.getCssSupport = function(cssprop)
{
    var vendors, root, i, css3mc;

    vendors = ['', '-moz-', '-webkit-', '-o-', '-ms-', '-khtml-'];
    root = document.documentElement;

    function toCamel(str)
    {
        return str.replace(/\-([a-z])/gi, function(match, val)
        {
            // convert first letter after "-" to uppercase
            return val.toUpperCase();
        });
    }

    for (i = 0; i < vendors.length; i++)
    {
        css3mc = toCamel(vendors[i] + cssprop);
        // if property starts with 'Ms'
        if (css3mc.substr(0, 2) == 'Ms')
        {
            // Convert 'M' to lowercase
            css3mc = 'm' + css3mc.substr(1);
        }
        if (css3mc in root.style)
        {
            return css3mc;
        }
    }

    return undefined;
};

/**
 * Clears the databody map and repopulates it based on what's in the databody
 * @private
 */
DvtDataGrid.prototype._refreshDatabodyMap = function()
{
    this._clearDatabodyMap();
    this._addNodesToDatabodyMap(this.m_databody.firstChild.childNodes)
};

/**
 * Adds a fragment to the databody content and fills the data body mapKey
 * @param {Element} databodyContent
 * @param {Element|DocumentFragment} fragment
 * @private
 */
DvtDataGrid.prototype._populateDatabody = function(databodyContent, fragment)
{
    this._addNodesToDatabodyMap(fragment.childNodes);
    databodyContent.appendChild(fragment); //@HTMLDUpdateOK
};

/**
 * Empties the databody and clears the databody map
 * @param {Element} databodyContent
 * @private
 */
DvtDataGrid.prototype._emptyDatabody = function(databodyContent)
{
    this._clearDatabodyMap();
    this.m_utils.empty(databodyContent);
};

/**
 * Adds an array of nodes to the databody map
 * @param {NodeList|Array} nodes
 * @private
 */
DvtDataGrid.prototype._addNodesToDatabodyMap = function(nodes)
{
    var node, indexes, id,extents, i;
    for (i = 0; i <nodes.length; i++)
    {
        node = nodes[i];
        if (this.m_utils.containsCSSClassName(node, this.getMappedStyle('cell')))
        {
            indexes = this.getCellIndexes(node);
            extents = this.getCellExtents(node);
            id = node.id;
            this._addToDatabodyMap(indexes, id, extents);
        }
    }
};

/**
 * Removes an array of nodes from the databody map
 * @param {Array} nodes
 * @private
 */
DvtDataGrid.prototype._removeNodesFromDatabodyMap = function(nodes)
{
    for (var i = 0; i <nodes.length; i++)
    {
        var indexes = this.getCellIndexes(nodes[i]);
        this._removeIndexFromDatabodyMap(indexes);
    }
};

/**
 * Adds an index, id pair to the databody map along with its extents
 * @param {Object} indexes
 * @param {string} id
 * @param {Object} extents
 * @private
 */
DvtDataGrid.prototype._addToDatabodyMap = function(indexes, id, extents)
{
    var rowExtent, columnExtent, i , j;
    rowExtent = extents['row'];
    columnExtent = extents['column'];

    for (i = 0; i < rowExtent; i++)
    {
        for (j = 0; j < columnExtent; j++)
        {
            this._addIndexToDatabodyMap(this.createIndex(indexes['row'] + i, indexes['column'] + j), id);
        }
    }
};

/**
 * Adds an index, id pair to the databody map
 * @param {Object} indexes
 * @param {string} id
 * @private
 */
DvtDataGrid.prototype._addIndexToDatabodyMap = function(indexes, id)
{
    var mapKey = 'r' + indexes['row'] + 'c' + indexes['column'];
    this.m_databodyMap["set"](mapKey, id); // quoted to make the closure compiler happy
};

/**
 * Removes an index, id pair from the databody map
 * @param {Object} indexes
 * @returns {boolean}
 * @private
 */
DvtDataGrid.prototype._removeIndexFromDatabodyMap = function(indexes)
{
    var mapKey = 'r' + indexes['row'] + 'c' + indexes['column'];
    return this.m_databodyMap["delete"](mapKey); // quoted to make the closure compiler happy
};

/**
 * Gets an id from the databody based on the index
 * @param {Object} indexes
 * @return the id at the index
 * @private
 */
DvtDataGrid.prototype._getFromDatabodyMap = function(indexes)
{
    var mapKey = 'r' + indexes['row'] + 'c' + indexes['column'];
    return this.m_databodyMap["get"](mapKey); // quoted to make the closure compiler happy
};

/**
 * Clears the databody map
 * @returns {boolean} the map
 * @private
 */
DvtDataGrid.prototype._clearDatabodyMap = function()
{
    return this.m_databodyMap["clear"](); // quoted to make the closure compiler happy
};

/**
 * Update the cellContext.indexes of a range of cells
 * @param {string} axis row/column
 * @param {number} atIndex startIndex along the axis
 * @param {number} count number of cells after the start to modify
 * @param {number} value value to increment/decremnt the index value by
 * @private
 */
DvtDataGrid.prototype._modifyAxisCellContextIndex = function(axis, atIndex, count, value)
{
    var i, j, axisCells, cell, cellContext;

    for (i = atIndex; i < atIndex + count; i++)
    {
        axisCells = this._getAxisCellsByIndex(i, axis);
        for (j = 0; j < axisCells.length; j++)
        {
            cell = axisCells[j];
            cellContext = cell[this.getResources().getMappedAttribute("context")];
            cellContext['indexes'][axis] += value;
        }
    }
};

/**
 * Get a cell or header by index along a given axis, will return first cell on that axis
 * @private
 */
DvtDataGrid.prototype._getCellOrHeaderByIndex = function(index, axis)
{
    var element = null;
    var cells = this._getAxisCellsByIndex(index, axis, true);
    if (cells != null && cells.length > 0)
    {
        element = cells[0];
    }
    if (element == null)
    {
        if (axis === 'row')
        {
            element = this._getHeaderByIndex(index, 0, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
            if (element == null)
            {
                element = this._getHeaderByIndex(index, 0, this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
            }
        }
        if (axis === 'column')
        {
            element = this._getHeaderByIndex(index, 0, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
            if (element == null)
            {
                element = this._getHeaderByIndex(index, 0, this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
            }
        }
    }
    return element;
};

/**
 * Get a cell by index
 * @param {Object} indexes
 * @returns {Element|null}
 */
DvtDataGrid.prototype._getCellByIndex = function(indexes)
{
    var id = this._getFromDatabodyMap(indexes);
    if (id != null)
    {
        return document.getElementById(id);
    }
    return null
};

/**
 * Get all the cells along an axis by index
 * @param {number} index
 * @param {string} axis row/column
 * @param {boolean=} breakOnFirstFind 
 * @returns {Array|null}
 * @private
 */
DvtDataGrid.prototype._getAxisCellsByIndex = function(index, axis, breakOnFirstFind)
{
    var i, cells, cell, start, end, axisExtent;
    cells = [];

    start = axis == 'row' ? this.m_startCol : this.m_startRow;
    end = axis == 'row' ? this.m_endCol : this.m_endRow;

    for (i = start; i <= end; i+= axisExtent)
    {
        cell = this._getCellByIndex(this.createIndex(axis == 'row' ? index : i, axis == 'row' ? i : index));
        if (cell != null)
        {
            axisExtent = this.getCellExtents(cell)[axis == 'row' ? 'column' : 'row'];
            cells.push(cell)
            if (breakOnFirstFind)
            {
                break;
            }
        }
        else
        {
            axisExtent = 1;
        }
    }
    return cells;
};

/**
 * Get all the cells along an axis by key
 * @param {string|null} key
 * @param {string} axis row/column
 * @param {boolean=} breakOnFirstFind 
 * @returns {Array|null}
 * @private
 */
DvtDataGrid.prototype._getAxisCellsByKey = function(key, axis, breakOnFirstFind)
{
    var databodyContent, cells, cell, i, axisKey, matchingCells = [];
    if (key == null || this.m_databody == null || this.m_databody['firstChild'] == null)
    {
        return null;
    }

    databodyContent = this.m_databody['firstChild'];
    cells = databodyContent['childNodes'];
    for (i = 0; i < cells.length; i++)
    {
        cell = cells[i];
        if (this.m_utils.containsCSSClassName(cell, this.getMappedStyle('cell')))
        {
            axisKey = this._getKey(cell, axis);
            if (axisKey === key)
            {
                matchingCells.push(cell);
                if (breakOnFirstFind)
                {
                    break;
                }
            }
        }
    }

    // can't find it, the row is not in viewport
    return matchingCells;
};
/**
 * @preserve Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Checks whether there are any focusable element inside a cell
 * @param {Element|undefined|null} elem the element to check focus inside
 * @return {boolean}
 */
DvtDataGrid.prototype._isContainFocusableElement = function(elem)
{
    var elems = this.getFocusableElementsInNode(elem);
    return (elems.length > 0);
};

/**
 * Unhighlights the selection.  Does not change selection, focus cell, anchor, or frontier
 */
DvtDataGrid.prototype.unhighlightSelection = function()
{
    var i, ranges;
    ranges = this.GetSelection();
    for (i = 0; i < ranges.length; i += 1)
    {
        this.unhighlightRange(ranges[i]);
    }
};

/**
 * Unhighlights the range.
 * @param {Object} range
 */
DvtDataGrid.prototype.unhighlightRange = function(range)
{
    var elems = this.getElementsInRange(range);
    this.unhighlightElems(elems);
};

/**
 * Highlights the range.
 * @param {Object} range
 * @param {boolean=} updateAccInfo
 */
DvtDataGrid.prototype.highlightRange = function(range, updateAccInfo)
{
    var elems, count;

    elems = this.getElementsInRange(range);
    this.highlightElems(elems);

    if (updateAccInfo)
    {
        // if there's islands of cells, then we'll have to count them
        if (this.GetSelection().length == 1)
        {
            count = elems.length;
        }
        else
        {
            count = this._getCurrentSelectionCellCount();
        }
        this._setAccInfoText('accessibleMultiCellSelected', {'num': count});
    }
};

/**
 * Calculate the total number of cells within the current selection ranges.
 * @private
 */
DvtDataGrid.prototype._getCurrentSelectionCellCount = function()
{
    var total, selection, elems, i;

    total = 0;
    selection = this.GetSelection();
    for (i = 0; i < selection.length; i++)
    {
        // count the number of elements in each selection range
        elems = this.getElementsInRange(selection[i]);
        if (elems != null)
        {
            total = total + elems.length;
        }
    }

    return total;
};

/**
 * Unhighlight elements
 * @param {Array} elems
 */
DvtDataGrid.prototype.unhighlightElems = function(elems)
{
    var i, elem;
    if (elems == null || elems.length == 0)
    {
        return;
    }

    for (i = 0; i < elems.length; i += 1)
    {
        elem = elems[i];
        this._unhighlightElement(elem, ['selected', 'topSelected', 'bottomSelected']);
    }
};

/**
 * Highlight elements
 * @param {Array} elems
 */
DvtDataGrid.prototype.highlightElems = function(elems)
{
    var i, elem, startRow, endRow, classArray, rowIndex;
    if (elems == null || elems.length == 0)
    {
        return;
    }

    elem = elems[0];
    startRow = this._getIndex(elem, 'row');

    elem = elems[elems.length - 1];
    endRow = this._getIndex(elem, 'row');

    for (i = 0; i < elems.length; i += 1)
    {
        classArray = ['selected'];
        elem = elems[i];
        rowIndex = this._getIndex(elem, 'row');
        if (rowIndex === startRow)
        {
            classArray.push('topSelected');
        }
        if (rowIndex === endRow)
        {
            classArray.push('bottomSelected');
        }
        this._highlightElement(elem, classArray);
    }
};

/**
 * Apply current selection to a range.  This is called when a newly set of cells are
 * rendered and selection needs to be applied on them.
 * @param {number=} startRow
 * @param {number=} endRow
 * @param {number=} startCol
 * @param {number=} endCol
 */
DvtDataGrid.prototype.applySelection = function(startRow, endRow, startCol, endCol)
{
    var i, ranges, elems;
    ranges = this.GetSelection();
    for (i = 0; i < ranges.length; i += 1)
    {
        elems = this.getElementsInRange(ranges[i], startRow, endRow, startCol, endCol);
        this.highlightElems(elems);
    }
};

/**
 * Handles click and drag to select multiple cells/rows
 * @param {Event} event
 */
DvtDataGrid.prototype.handleDatabodySelectionDrag = function(event)
{
    var index, cell, target;

    if (this.m_utils.isTouchDevice())
    {
        cell = this.findCell(document.elementFromPoint(event.touches[0].clientX, event.touches[0].clientY));
    }
    else
    {
        target = /** @type {Element} */ (event.target);
        cell = this.findCell(target);
    }

    if (cell != null)
    {
        if (this.isHeaderSelectionType(this.m_selectionFrontier))
        {
            this.extendSelectionHeader(cell, event);
        }
        else
        {
            index = this.getCellIndexes(cell);
            this.extendSelection(index, event);
        }
    }

};

/**
 * Checks whether a cell is selected.  This is used in touch logic in handleDatabodyClickSelection.
 * @param {Element} cell the cell element
 * @return {boolean} true if the cell is selected, false otherwise.
 * @private
 */
DvtDataGrid.prototype._isSelected = function(cell)
{
    var selectedClassName = this.getMappedStyle("selected");
    if (selectedClassName != null)
    {
        return this.m_utils.containsCSSClassName(cell, selectedClassName);
    }

    // should not end up here
    return false;
};

/**
 * Deselect a cell/row.  This is used in touch logic in handleDatabodyClickSelection.
 * @param {Object} index the cell index of the cell/row to deselect
 * @private
 */
DvtDataGrid.prototype._deselect = function(index)
{
    var rowIndex, columnIndex, indexToRemove, ranges, i, range, startIndex, endIndex, rangeStartRow,
            rangeEndRow, rangeStartColumn, rangeEndColumn;

    if (this.m_options.getSelectionMode() == "row")
    {
        // drop the column index
        index = this.createIndex(index['row']);
    }

    rowIndex = index['row'];
    columnIndex = index['column'];

    // find the range in current selection
    indexToRemove = -1;
    ranges = this.GetSelection();
    for (i = 0; i < ranges.length; i += 1)
    {
        range = ranges[i];
        startIndex = range['startIndex'];
        endIndex = this.getEndIndex(range);

        rangeStartRow = startIndex['row'];
        rangeEndRow = endIndex['row'];

        if (rangeStartRow != rowIndex || rangeEndRow != rowIndex)
        {
            continue;
        }

        if (!isNaN(startIndex['column']) && !isNaN(endIndex['column']))
        {
            rangeStartColumn = startIndex['column'];
            rangeEndColumn = endIndex['column'];

            if (rangeStartColumn != columnIndex || rangeEndColumn != columnIndex)
            {
                continue;
            }

            // both row and column index matches, we are done
            indexToRemove = i;
            break;
        }
        else
        {
            // if column index is defined in index, skip this range
            if (!isNaN(columnIndex))
            {
                continue;
            }

            // no column index, and row matches
            indexToRemove = i;
            break;
        }
    }

    // unhighlight index and remove from selection
    if (indexToRemove != -1)
    {
        this.unhighlightRange(ranges[indexToRemove]);
        ranges.splice(indexToRemove, 1);
    }
};

/**
 * Handles click to select header rows/columns
 * @param {Event} event
 */
DvtDataGrid.prototype.handleHeaderClickSelection = function(event)
{
    var axis, index, level, target, header, start, end, elem, shiftKey;

    target = /** @type {Element} */ (event.target);
    header = this.findHeader(target);
    shiftKey = event.shiftKey;

    if (!shiftKey && header)
    {
        this._setActive(header, this._createActiveObject(header), event);
    }

    // no selection if header was not selected or if cell is active and shiftkey is not selected when selectin gheader
    if (this.m_active == null || (this.m_active['type'] != 'header' && !event.shiftKey))
    {
        return false;
    }

    if (this.m_active['type'] == 'header')
    {
        axis = this.m_active['axis'];
        index = this.m_active['index'];
        level = this.m_active['level'];
    }

    if (this.m_utils.isTouchDevice())
    {
        //remove the touch affordance on a new tap, unhighlight the active cell, and select the new one
        this._removeTouchSelectionAffordance();
        this._selectHeader(axis, index, level, event)            
    }
    else
    {
        if (!shiftKey)
        {
            this._selectHeader(axis, index, level, event)            
        }
        else if (shiftKey)
        {
            var headerContext = target[this.getResources().getMappedAttribute("context")];
            //check if cell anchor, if so inherit row/column axis from target
            if (axis == null)
            {
                axis = headerContext['axis'];
            }
            //make sure we only handle row - row, column - column properly. row-column/column-row is ignored if shiftkey is active.
            if ((axis.indexOf('row') != -1 && headerContext['axis'].indexOf("row") != -1) || (axis.indexOf('column') != -1 && headerContext['axis'].indexOf("column") != -1))
            {
                this.extendSelectionHeader(target, event)
            }
            return;
        }
    }
}

/**
 * Handles discontiguous header set active from databody
 * @param {Event} event triggering the operation
 * @param {string} axis header axis
 * @param {Element|null} header to be set active
 * @param {number} axisLevelCount
 */
DvtDataGrid.prototype.discontiguousHeaderSetActiveFromDatabody = function(event, axis, header, axisLevelCount)
{
    var trueAxis = axis.replace("End", "");
    var currentSelection = this.m_selection[this.m_selection.length - 1];
    this.unhighlightSelection();
    if (currentSelection['startIndex']['row'] == this.m_active['indexes']['row'] && currentSelection['endIndex']['row'] == this.m_active['indexes']['row'] && currentSelection['startIndex']['column'] == this.m_active['indexes']['column'] && currentSelection['endIndex']['column'] == this.m_active['indexes']['column'])
    {
        this.m_selection.pop();
    }
    this._setActive(header, {'type':'header', 'index':this.m_trueIndex[trueAxis], 'level': axisLevelCount - 1, 'axis': axis}, event, false);
    this.rehighlightSelection();
};

/**
 * Handles header change due to mouse up after dragging between parent/child or child/parent header anchor and target.
 * @param {Event} event triggering the operation
 */
DvtDataGrid.prototype.handleDragAnchorChange = function(event)
{
    var targetElement, targetContext, targetStart, targetEnd, sourceParent, sourceParentContext, sourceParentStart, sourceParentEnd, extent;
    // If dragging and we land in a parent or child of the anchor, reset the anchor. 
    if (this.m_headerDragState)
    {
        targetElement = /** @type {Element} */ (event.target);
        targetContext = targetElement[this.getResources().getMappedAttribute("context")];
        sourceParent = this._getActiveElement();
        sourceParentContext = sourceParent[this.getResources().getMappedAttribute("context")];
        sourceParentStart = this._getAttribute(sourceParent['parentNode'], 'start', true);
        sourceParentEnd = sourceParentStart + this._getAttribute(sourceParent['parentNode'], 'extent', true) - 1;

        // if no extent, skip
        if (!targetContext['extent'])
        {
            return;
        }

        if (targetContext['extent'] == 1)
        {
            targetStart = targetContext['index'];
            targetEnd = targetContext['index'];
        }
        else
        {
            targetStart = this._getAttribute(targetElement['parentNode'], 'start', true);
            targetEnd = targetStart + this._getAttribute(targetElement['parentNode'], 'extent', true) - 1;
        }

        sourceParentStart = this._getAttribute(sourceParent['parentNode'], 'start', true);
        sourceParentEnd = sourceParentStart + this._getAttribute(sourceParent['parentNode'], 'extent', true) - 1;

        extent = sourceParentContext['extent'];

        if (extent != 1)
        {
            if ((targetStart <= sourceParentStart && targetEnd >= sourceParentEnd) || (targetStart >= sourceParentStart && targetEnd <= sourceParentEnd))
            {
                // if moving levels, and we are a parent or child of current cell, just reset active to the parent/child.
                this._setActive(targetElement, this._createActiveObject(targetElement), event);
            }
        }
        else
        {
            if (targetStart != targetEnd && (targetStart <= sourceParentStart && sourceParentStart <= targetEnd))
            {
                // if moving levels, and we are a parent or child of current cell, just reset active to the parent/child.
                this._setActive(targetElement, this._createActiveObject(targetElement), event);
            }             
        }
    }
}

/**
 * Handles click to select multiple cells/rows
 * @param {Event} event
 */
DvtDataGrid.prototype.handleDatabodyClickSelection = function(event)
{
    var index, cell, ctrlKey, shiftKey, target;
    target = /** @type {Element} */ (event.target);
    cell = this.findCell(target);
    if (cell != null)
    {
        index = this.getCellIndexes(cell);
    }

    if (index != null && index != undefined)
    {
        if (this.isMultipleSelection() && event.button === 2 && this._isContainSelection(index))
        {
            //if right click and inside multiple selection do not change anything
            return;
        }

        ctrlKey = this.m_utils.ctrlEquivalent(event);
        shiftKey = event.shiftKey;
        if (this.isMultipleSelection())
        {
            if (this.m_utils.isTouchDevice())
            {
                //remove the touch affordance on a new tap, unhighlight the active cell, and select the new one
                this._removeTouchSelectionAffordance();
                if (this.m_active != null)
                {
                    this._unhighlightActive();
                }
                this.selectAndFocus(index, event, false);
            }
            else
            {
                if (!ctrlKey && !shiftKey)
                {
                    this.selectAndFocus(index, event, false);
                }
                else if (!ctrlKey && shiftKey)
                {
                    this.extendSelection(index, event);
                    // active cell doesn't change in this case
                }
                else
                {
                    this.selectAndFocus(index, event, true);
                }
            }
        }
        else
        {
            this.selectAndFocus(index, event, false);
        }
    }
};

/**
 * Handle select all
 * @param {Event} event the event causing the action
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleSelectAll = function(event)
{
    if (this._isSelectionEnabled() && this.isMultipleSelection())
    {
        if (this.m_options.getSelectionMode() == "row")
        {
            // drop the column index
            this._selectRange(this.createIndex(0), this.createIndex(-1), event);
        }
        else
        {
            this._selectRange(this.createIndex(0, 0), this.createIndex(-1, -1), event);
        }
        // if affordances are active, remove them.
        if (this.m_utils.isTouchDevice() && this.m_topSelectIconContainer && this.m_bottomSelectIconContainer)
        {
            this.m_topSelectIconContainer['parentNode'].removeChild(this.m_topSelectIconContainer);
            this.m_bottomSelectIconContainer['parentNode'].removeChild(this.m_bottomSelectIconContainer);
        }
        return true;
    }
    else
    {
        return false;
    }
}

/**
 * Determine if the specified cell index is inside the current selection.
 * @param {Object} index the cell index
 * @param {Array=} ranges the selection to see if the index is in, allows us to check old ranges
 * @return {boolean} true is the cell index specified is inside the selection, false otherwise
 * @private
 */
DvtDataGrid.prototype._isContainSelection = function(index, ranges)
{
    var range, startIndex, endIndex, rangeStartRow, rangeEndRow, rangeStartColumn, rangeEndColumn, i;
    if (ranges == null)
    {
        ranges = this.GetSelection();
    }
    for (i = 0; i < ranges.length; i += 1)
    {
        range = ranges[i];
        startIndex = range['startIndex'];
        endIndex = this.getEndIndex(range);

        rangeStartRow = startIndex['row'];
        rangeEndRow = endIndex['row'];

        // checks if row outside of range
        if (index['row'] < rangeStartRow || (rangeEndRow != -1 && index['row'] > rangeEndRow))
        {
            // skip
            continue;
        }

        if (!isNaN(startIndex['column']) && !isNaN(endIndex['column']))
        {
            rangeStartColumn = startIndex['column'];
            rangeEndColumn = endIndex['column'];

            // checks if row outside of range
            if (index['column'] < rangeStartColumn || (rangeEndColumn != -1 && index['column'] > rangeEndColumn))
            {
                // skip
                continue;
            }

            // within range return immediately
            return true;
        }
        else
        {
            // no column specified, meaning all columns
            return true;
        }
    }

    return false;
};

/**
 * Compare the two selection to see if they are identical.
 * @param {Object} selection1 the first selection
 * @param {Object} selection2 the second selection
 * @return {boolean} true if the selections are identical, false otherwise
 * @private
 */
DvtDataGrid.prototype._compareSelections = function(selection1, selection2)
{
    var i, j, foundMatch;

    // currently assumes all selections will be the same if old and new selection are equal
    // now allows not to fire on every drag event
    // todo: needs to handle discontigous selection case

    if (selection1.length !== selection2.length)
    {
        return false;
    }

    for (i = 0; i < selection1.length; i += 1)
    {
        foundMatch = false;
        for (j = 0; j < selection2.length; j += 1)
        {
            if (this._compareIndividualSelectionObjects(selection1[i], selection2[j]))
            {
                foundMatch = true;
            }
        }
        if (foundMatch === false)
        {
            return false;
        }
    }

    return true;
};

/**
 * Compare the two selection to see if they are identical.
 * @param {Object} selection1 the first selection
 * @param {Object} selection2 the second selection
 * @return {boolean} true if the selections are identical, false otherwise
 * @private
 */
DvtDataGrid.prototype._compareIndividualSelectionObjects = function(selection1, selection2)
{
    if (selection1['startIndex']['row'] === selection2['startIndex']['row'] &&
            selection1['startIndex']['column'] === selection2['startIndex']['column'] &&
            selection1['endIndex']['row'] === selection2['endIndex']['row'] &&
            selection1['endIndex']['column'] === selection2['endIndex']['column'])
    {
        return true;
    }
    return false;
};

/**
 * Unhighlight and clear the current selection. If you are modifying the selection
 * object you should not call this method. It should only be used in the case of a
 * true clear where the selection winds up empty. This fires an event that the selection
 * has changed if it contained values beforehand.
 * @private
 * @param {Event=} event the event triggering the clear
 */
DvtDataGrid.prototype._clearSelection = function(event)
{
    var previous;

    // unhighlight previous selection
    this.unhighlightSelection();
    this._removeTouchSelectionAffordance();

    // clear the selection and fire the
    previous = this.GetSelection();
    this.m_selection = [];

    this._compareSelectionAndFire(event, previous);
};

/************************* key handler methods ************************************/
/**
 * Sets whether the data grid is in discontiguous selection mode
 * @param {boolean} flag true to set grid to discontiguous selection mode
 * @private
 */
DvtDataGrid.prototype.setDiscontiguousSelectionMode = function(flag)
{
    this.m_discontiguousSelection = flag;

    // announce to screen reader
    this._setAccInfoText(flag ? "accessibleRangeSelectModeOn" : "accessibleRangeSelectModeOff");
};

/**
 * Selects the entire row of cells
 * @param {number} rowStart the end row index
 * @param {number} rowEnd the start row index
 * @param {Event} event the dom event that triggers the selection
 * @private
 */
DvtDataGrid.prototype._selectEntireRow = function(rowStart, rowEnd, event)
{
    var startIndex, endIndex, returnObj;

    // create the start and end index then selects the range
    startIndex = this.createIndex(rowStart, 0);
    endIndex = this.createIndex(rowEnd, -1);
    returnObj = this._getSelectionStartAndEnd(startIndex, endIndex, 0);   
    
    if (this.m_options.getSelectionMode() == "row")
    {
        // drop the column index
        this._selectRange(this.createIndex(returnObj['min']['row']), this.createIndex(returnObj['max']['row']), event);
    }
    else
    {
        this._selectRange(this.createIndex(returnObj['min']['row'], 0), this.createIndex(returnObj['max']['row'], -1), event);        
    }
};

/**
 * Selects the entire column of cells
 * @param {number} columnStart the column start index
 * @param {number} columnEnd the column end index
 * @param {Event} event the dom event that triggers the selection
 * @private
 */
DvtDataGrid.prototype._selectEntireColumn = function(columnStart, columnEnd, event)
{
    var startIndex, endIndex, returnObj;

    // create the start and end index then selects the range
    startIndex = this.createIndex(0, columnStart);
    endIndex = this.createIndex(-1, columnEnd);
    returnObj = this._getSelectionStartAndEnd(startIndex, endIndex, 0);   
    
    this._selectRange(this.createIndex(0, returnObj['min']['column']), this.createIndex(-1, returnObj['max']['column']), event);
};

/**
 * Selects a range of cells.
 * @param {Object} startIndex the start row/column indexes
 * @param {Object} endIndex the end row/column indexes
 * @param {Event} event the dom event that triggers the selection
 * @private
 */
DvtDataGrid.prototype._selectRange = function(startIndex, endIndex, event)
{
    // no longer clear selection, if it is cleared here we can't return anything for previous selection
    this.unhighlightSelection();
    this._createRangeWithKeys(startIndex, endIndex, this._selectRangeCallback.bind(this, event));
};

/**
 * Callback for once the new range is constructed
 * @param {Event} event the dom event that triggers the selection
 * @param {Object} newRange the new range to be selected
 * @private
 */
DvtDataGrid.prototype._selectRangeCallback = function(event, newRange)
{
    var selection, previous;

    // We need to pass the option change event the previous selection.
    // We also need to overwrite the old selection instance with a new one
    // so clone the old one, update, and then replace so that the object passed
    // as the previous matches the old reference and the new selection is a new
    // reference, create a brand new selection
    previous = this.GetSelection();
    if (!this.m_discontiguousSelection && event && !(this.m_utils.ctrlEquivalent(event) && event.button == 0))
    {
        selection = [];        
    }
    else
    {
        selection = previous.slice(0);
    }
    selection.push(newRange);
    this.m_selection = selection;

    this.rehighlightSelection();

    if (this._isDatabodyCellActive() && event.target[this.getResources().getMappedAttribute("context")] && !event.target[this.getResources().getMappedAttribute("context")]['axis'] && !this.m_selectionFrontier['axis'])
    {
        // reset frontier to be the same as active
        this.m_selectionFrontier = this.m_active['indexes'];

        this._highlightActive();
    }

    // fire selection event if the selection has changed
    this._compareSelectionAndFire(event, previous);
};

/**
 * Highlight all ranges in the selection
 */
DvtDataGrid.prototype.rehighlightSelection = function()
{
    for (var i = 0; i < this.m_selection.length; i++)
    {
        this.highlightRange(this.m_selection[i]);
    }
}

/**
 * Retrieve the current selection
 * @return {Array} an array of ranges
 */
DvtDataGrid.prototype.GetSelection = function()
{
    if (this.m_selection == null)
    {
        this.m_selection = [];
    }
    return this.m_selection;
};

/**
 * Sets a range of selections
 * @param {Object} selection
 */
DvtDataGrid.prototype.SetSelection = function(selection)
{
    var previous;

    // it can be null but cannot be undefined
    if (selection != undefined)
    {
        if (selection == null)
        {
            selection = [];
        }

        //if we set the selection we should ungihlight the old one
        this.unhighlightSelection();

        previous = this.GetSelection();
        this.m_selection = selection;

        // if it's not render yet, don't apply selection
        if (this.m_databody != null)
        {
            this.applySelection(this.m_startRow, this.m_endRow, this.m_startCol, this.m_endCol);
        }
        // do not fire selection event when set on us externally, it will be taken
        // care of in the wrappers option layer
    }
};

/**
 * Fires selection event
 * @param {Event|undefined} event the dom event that triggers the selection
 * @param {Object} previousSelection
 * @protected
 */
DvtDataGrid.prototype.fireSelectionEvent = function(event, previousSelection)
{
    var details = {
        'event': event, 'ui': {
            'selection': this.GetSelection(),
            'previousSelection': previousSelection
        }
    };
    this.fireEvent('select', details);
};

/**
 * Shift+click to extend the selection
 * @param {Object} index - the end index of the selection.
 * @param {Event=} event - the DOM event causing the selection to to be extended
 * @param {number=} direction - the keystroke keyCode if applicable
 */
DvtDataGrid.prototype.extendSelection = function(index, event, direction)
{
    var anchor, activeRow, activeColumn, returnObj, minIndex, maxIndex;

    // prevent cell selection when extending a row/column selection
    if (this.m_active['type'] == 'header')
    {
        return;
    }

    // find the the top left index
    if (this.m_utils.isTouchDevice())
    {
        anchor = this.m_touchSelectAnchor;
    }
    else
    {
        // do not copy anchor object so we can directly modify it
        activeRow = this.m_active['indexes']['row'];
        activeColumn = this.m_active['indexes']['column'];
        anchor = {'row': activeRow, 'column': activeColumn};
    }
    
    if (anchor == null)
    {
        return;
    }
    
    // reset focus on previous selection frontier
    this._resetSelectionFrontierFocus();
    
    // update the selctionFrontier, more complicated dues to merged cells
    returnObj = this._updateSelectionFrontier(anchor, index, direction);
    minIndex = returnObj['min'];
    maxIndex = returnObj['max'];
    
    if (this.m_options.getSelectionMode() == "row")
    {
        // drop the column index
        minIndex = this.createIndex(minIndex['row']);
        maxIndex = this.createIndex(maxIndex['row']);
    }
   
    this._createRangeWithKeys(minIndex, maxIndex, this._extendSelectionCallback.bind(this, event, anchor));
};

/**
 * Extend the selection for headers
 * @param {Object} target - {axis, index} target of the selection.
 * @param {Event=} event - the DOM event causing the selection to to be extended
 * @param {boolean=} isExtend - boolean if we are extending a selection
 */
DvtDataGrid.prototype.extendSelectionHeader = function(target, event, isExtend)
{
    var anchor, axis, targetStart, targetEnd, sourceParent, sourceParentContext, sourceParentStart, sourceParentEnd, extent, endpoint, targetContext, targetElement, previousAnchor;
    // no target so just return
    if (target === null)
    {
        return;
    }

    targetElement = /** @type {Element} */ (target);
    targetContext = target[this.getResources().getMappedAttribute("context")];

    if (isExtend == null)
    {
        isExtend = this.m_discontiguousSelection;
    }

    // if no targetContext, break
    if (!targetContext)
    {
        return;
    }
    else if (targetContext['cell'])
    {
        if (this.m_selectionFrontier['axis'].indexOf("column") != -1)
        {
            targetStart = targetContext['indexes']['column'];
        }
        else
        {
            targetStart = targetContext['indexes']['row'];            
        }
        targetEnd = targetStart;
        endpoint = targetStart;
    }
    else
    {
        // check if target is the same as the current frontier target
        if (this.m_selectionFrontier && targetContext['index'] == this.m_selectionFrontier['index'] && targetContext['level'] == this.m_selectionFrontier['level'])
        {
            return;
        }

        if (targetContext['extent'] == 1)
        {
            targetStart = targetContext['index'];
            targetEnd = targetContext['index'];
            endpoint = targetContext['index']
        }
        else
        {
            targetStart = this._getAttribute(target['parentNode'], 'start', true);
            targetEnd = targetStart + this._getAttribute(target['parentNode'], 'extent', true) - 1;
        }
    }

    sourceParent = this._getActiveElement();
    sourceParentContext = sourceParent[this.getResources().getMappedAttribute("context")];

    // find the the top left index
    if (this.m_utils.isTouchDevice())
    {
        if (this.m_selectionFrontier['axis'].indexOf("column") != -1)
        {
            anchor = this.m_touchSelectAnchor['column'];
        }
        else
        {
            anchor = this.m_touchSelectAnchor['row'];            
        }

        if (anchor == -1)
        {
            anchor = 0;
        }
        sourceParentStart = anchor;
        sourceParentEnd = anchor;
        axis = this.m_selectionFrontier['axis'];
    }
    else
    {
        sourceParentStart = this._getAttribute(sourceParent['parentNode'], 'start', true);
        sourceParentEnd = sourceParentStart + this._getAttribute(sourceParent['parentNode'], 'extent', true) - 1;

        if (this.m_active['type'] === 'header')
        {
            axis = this.m_active['axis'];
            anchor = this.m_active['index'];
        }
    }

    if (this.m_active['type'] == 'cell')
    {
        extent = 1;
        axis = targetContext['axis'];
        if (axis.indexOf('row') != -1)
        {
            anchor = this.m_active['indexes']['row'];
        }
        else
        {
            anchor = this.m_active['indexes']['column'];
        }
    }
    else
    {
        extent = sourceParentContext['extent'];
    }

    // pop if anchors match so we don't duplicate a selection
    if (isExtend)
    {
        previousAnchor = this.m_selection[this.m_selection.length - 1];

        // unhighlight selection now 
        this.unhighlightSelection();

        // check that the anchor and the selection frontier index match the previous anchor start and end indices
        if (previousAnchor && this.checkCorners(axis, anchor, previousAnchor))
        {
            this.m_selection.pop();
        }
    }

    if (extent != 1)
    {
        if ((targetStart <= sourceParentStart && targetEnd >= sourceParentEnd) || (targetStart >= sourceParentStart && targetEnd <= sourceParentEnd))
        {
            // if moving levels, and we are a parent or child of current cell, just reset active to the parent/child.
            // unless we are click + drag, then we don't change the target
            if (!this.m_headerDragState)
            {
                this._setActive(targetElement, this._createActiveObject(targetElement), event);
            }

            anchor = targetStart;
            endpoint = targetEnd;
        }
        else if (targetStart > sourceParentStart && targetEnd > sourceParentEnd)
        {
            anchor = sourceParentStart;
            endpoint = targetEnd;
        }
        else if (targetStart < sourceParentStart && targetEnd < sourceParentEnd)
        {
            anchor = targetStart; 
            endpoint = sourceParentEnd;
        }
    }
    else if (extent == 1)
    {
        if (targetStart != targetEnd)
        {
            if (targetStart <= anchor && anchor <= targetEnd)
            {
                // if moving levels, and we are a parent or child of current cell, just reset active to the parent/child.
                // unless we are click + drag or have cell type anchor, then we don't change the target
                if (!this.m_headerDragState && this.m_active['type'] != 'cell')
                {
                    this._setActive(targetElement, this._createActiveObject(targetElement), event);
                }
                anchor = targetStart;
                endpoint = targetEnd;                    
            } 
            else if (targetStart <= anchor && targetEnd <= anchor)
            {
                endpoint = anchor
                anchor = targetStart;
            }
            else if (targetStart >= anchor && targetEnd >= anchor)
            {
                endpoint = targetEnd;
            }
        } 
    }

    if (targetContext['cell'])
    {
        this.setHeaderSelectionFrontier(axis, endpoint, endpoint, this.m_selectionFrontier['level'], /** @type {Element} */ (target), true);
    }
    else
    {
        this.setHeaderSelectionFrontier(axis, endpoint, targetContext['index'], targetContext['level'], /** @type {Element} */ (target), true);        
    }

    if(axis.indexOf('column') != -1)
    {
        if (event)
        {
            this._selectEntireColumn(anchor, endpoint, event);
        }
        else
        {
            this._selectEntireColumn(anchor, endpoint, null);            
        }
    }
    else if (axis.indexOf('row') != -1)
    {
        if (event)
        {
            this._selectEntireRow(anchor, endpoint, event);
        }
        else
        {
            this._selectEntireRow(anchor, endpoint, null);
        }
    }
}

/**
 * @param {Object} anchor - the current selection.
 * @param {Object} index - the anchor of the extend selection.
 * @param {number=} direction - keyCode
 * @returns {Object} with min and max index of the selection
 */
DvtDataGrid.prototype._updateSelectionFrontier = function(anchor, index, direction)
{
    var returnObj, minIndex, maxIndex, useIndex, previous, lastSelectionIndex, currentRange, startIndex, endIndex, tempAnchor, tempIndex;
    
    if (this.m_options.getSelectionMode() == "row")
    {
        // drop the column index
        tempAnchor = this.createIndex(anchor['row'], this.m_startCol);
        tempIndex = this.createIndex(index['row'], this.m_endCol);
    }    
    else
    {
        tempAnchor = anchor;
        tempIndex = index;
    }
    
    // keyboard specific
    if (direction != null)
    {
        previous = this.GetSelection();
        lastSelectionIndex = this._getLastAnchoredSelectionIndex(previous, tempAnchor);
        currentRange = previous[lastSelectionIndex];    
        startIndex = currentRange['startIndex'];
        endIndex = currentRange['endIndex'];
        
        // update the tempAnchor if it has extended due to a merge cell changing the corner of the tempAnchor
        if (startIndex['row'] === this.m_selectionFrontier['row'] && tempAnchor['row'] !== startIndex['row'])
        {
            tempAnchor['row'] = endIndex['row'];
        }
        else if (endIndex['row'] === this.m_selectionFrontier['row'] && tempAnchor['row'] !== endIndex['row'])
        {
            tempAnchor['row'] = startIndex['row']
        }
        
        if (startIndex['column'] === this.m_selectionFrontier['column'] && tempAnchor['column'] !== startIndex['column'])
        {
            tempAnchor['column'] = endIndex['column'];
        }
        else if (endIndex['column'] === this.m_selectionFrontier['column'] && tempAnchor['column'] !== endIndex['column'])
        {
            tempAnchor['column'] = startIndex['column']
        }
    }
    
    returnObj = this._getSelectionStartAndEnd(tempAnchor, tempIndex, 0);
    useIndex = tempIndex;    

    if (direction != null)
    {        
        minIndex = returnObj['min'];
        maxIndex = returnObj['max'];
        
        if (this.m_options.getSelectionMode() == "row")
        {
            // drop the column index
            minIndex = this.createIndex(minIndex['row']);
            maxIndex = this.createIndex(maxIndex['row']);
        }        
        
        // ensure the range actually grows as a result of the arrow key
        while (this._compareIndividualSelectionObjects(currentRange, this.createRange(minIndex, maxIndex)))
        {
            if (direction === DvtDataGrid.keyCodes.LEFT_KEY)
            {
                if (tempIndex['column'] === 0)
                {
                    break;
                }
                tempIndex['column'] -= 1;
            }
            else if (direction === DvtDataGrid.keyCodes.RIGHT_KEY)
            {
                if (tempIndex['column'] === this.m_endCol)
                {
                    break;
                }                    
                tempIndex['column'] += 1;
            }        
            else if (direction === DvtDataGrid.keyCodes.UP_KEY)
            {
                if (tempIndex['row'] === 0)
                {
                    break;
                }                     
                tempIndex['row'] -= 1;
            }              
            else if (direction === DvtDataGrid.keyCodes.DOWN_KEY)
            {
                if (tempIndex['row'] === this.m_endRow)
                {
                    break;
                }                      
                tempIndex['row'] += 1;
            }        
            else
            {
                break;
            }
            
            returnObj = this._getSelectionStartAndEnd(tempAnchor, tempIndex, 0);
            minIndex = returnObj['min'];
            maxIndex = returnObj['max'];        
            if (this.m_options.getSelectionMode() == "row")
            {
                // drop the column index
                minIndex = this.createIndex(minIndex['row']);
                maxIndex = this.createIndex(maxIndex['row']);
            } 
        }    
        
        // set the frontier appropriately
        if (direction === DvtDataGrid.keyCodes.LEFT_KEY)
        {
            useIndex['column'] = (minIndex['column'] < tempAnchor['column']) ? minIndex['column'] : maxIndex['column'];
            useIndex['row'] = (minIndex['row'] < tempAnchor['row']) ? minIndex['row'] : maxIndex['row'];
        }
        else if (direction === DvtDataGrid.keyCodes.RIGHT_KEY)
        {
            useIndex['column'] = (maxIndex['column'] > tempAnchor['column']) ? maxIndex['column'] : minIndex['column'];
            useIndex['row'] = (minIndex['row'] < tempAnchor['row']) ? minIndex['row'] : maxIndex['row'];
        }             
        else if (direction === DvtDataGrid.keyCodes.UP_KEY)
        {
            useIndex['row'] = (minIndex['row'] < tempAnchor['row']) ? minIndex['row'] : maxIndex['row'];
            useIndex['column'] = (minIndex['column'] < tempAnchor['column']) ? minIndex['column'] : maxIndex['column'];            
        }
        else if (direction === DvtDataGrid.keyCodes.DOWN_KEY)
        {
            useIndex['row'] = (maxIndex['row'] > tempAnchor['row']) ? maxIndex['row'] : minIndex['row'];
            useIndex['column'] = (minIndex['column'] < tempAnchor['column']) ? minIndex['column'] : maxIndex['column'];            
        }
    }
    else
    {
        useIndex = this.getCellIndexes(this._getCellByIndex(useIndex));        
    }
    
    if (this.m_options.getSelectionMode() == "row")
    {
        useIndex = this.createIndex(useIndex['row'], this.m_active['indexes']['column']);
    }
    
    this.m_selectionFrontier = useIndex;     
    
    return returnObj;
    
};

/**
 * Get the min or max between two numbers that could either be NaN
 * @param {number|undefined|null} val1 - val1
 * @param {number|undefined|null} val2 - val2
 * @param {Function} mathFunc - math function to apply to two values
 * @returns {number|undefined|null} 
 */
DvtDataGrid.prototype._getMinOrMax = function(val1, val2, mathFunc)
{
    if (isNaN(val1))
    {
        if (isNaN(val2))
        {
            return null;
        }
        return val2;
    }
    else if (isNaN(val2))
    {
        return val1;
    }
    return mathFunc(val1, val2);
};

/**
 * @param {Object} anchor - the current selection.
 * @param {Object} index - the anchor of the extend selection.
 * @param {number} numElems - the number of elements in the range used to break recursion
 * @returns {Object} with min and max index of the selection
 */
DvtDataGrid.prototype._getSelectionStartAndEnd = function(anchor, index, numElems)
{
    var cells, minIndex, maxIndex, i, startIndex, endIndex;
    cells = this.getElementsInRange(this.createRange(anchor, index));
    
    if (cells.length === numElems)
    {
        return {'min': anchor, 'max': index};
    }
    
    minIndex = {'row': this._getMinOrMax(anchor['row'], index['row'], Math.min), 'column': this._getMinOrMax(anchor['column'], index['column'], Math.min)};
    maxIndex = {'row': this._getMinOrMax(anchor['row'], index['row'], Math.max), 'column': this._getMinOrMax(anchor['column'], index['column'], Math.max)};
    
    for (i = 0; i < cells.length; i++)
    {
        startIndex = this.getCellIndexes(cells[i]);
        endIndex = this.getCellEndIndexes(cells[i]);
        
        if (startIndex['row'] < minIndex['row'] ||  minIndex['row'] == null)
        {
            minIndex['row'] = startIndex['row'];
        }
        if (startIndex['column'] < minIndex['column'] ||  minIndex['column'] == null)
        {
            minIndex['column'] = startIndex['column'];
        }  
        if (endIndex['row'] > maxIndex['row'] ||  maxIndex['row'] == null)
        {
            maxIndex['row'] = endIndex['row'];
        }
        if (endIndex['column'] > maxIndex['column'] ||  maxIndex['column'] == null)
        {
            maxIndex['column'] = endIndex['column'];
        }
    }
    
    return this._getSelectionStartAndEnd(minIndex, maxIndex, cells.length)
};

/**
 * @param {Object} selection - the current selection.
 * @param {Object} anchor - the anchor of the extend selection.
 * @returns {number} the index of the selection that contains the anchor cell as a corner
 * @private
 */
DvtDataGrid.prototype._getLastAnchoredSelectionIndex = function(selection, anchor)
{
    var i, rowStart, rowEnd, columnStart, columnEnd, rowAnchor, columnAnchor;
    for (i = selection.length - 1; i > -1; i -= 1)
    {
        // if row selection just compare rows, also check if its not select all type.
        if (this._isContainSelection(anchor, [selection[i]]))
        {
            return i;
        }
    }
    // return last index by default to keep old behavior as worst case
    return i - 1;
};

/**
 * Once the range is created from the index continue to extend the selection
 * @param {Event|null|undefined} event - the DOM event causing the selection to to be extended
 * @param {Object} anchor - the anchor cell
 * @param {Object} newRange - the new range of the selection.
 * @private
 */
DvtDataGrid.prototype._extendSelectionCallback = function(event, anchor, newRange)
{
    var selection, previous, currentRange, startIndexesMatch, endIndexesMatch, lastSelectionIndex;

    previous = this.GetSelection();
    lastSelectionIndex = this._getLastAnchoredSelectionIndex(previous, anchor);
    currentRange = previous[lastSelectionIndex];

    if (currentRange == null)
    {
        // Anchor cell removed from selection during marquee selection
        return;
    }

    // checks if selection has changed
    startIndexesMatch = (currentRange['startIndex']['row'] == newRange['startIndex']['row']);
    if (currentRange['startIndex']['column'] != null && newRange['startIndex']['column'] != null)
    {
        startIndexesMatch = startIndexesMatch && (currentRange['startIndex']['column'] == newRange['startIndex']['column']);
    }

    endIndexesMatch = (currentRange['endIndex']['row'] == newRange['endIndex']['row']);
    if (currentRange['endIndex']['column'] != null && newRange['endIndex']['column'] != null)
    {
        endIndexesMatch = endIndexesMatch && (currentRange['endIndex']['column'] == newRange['endIndex']['column']);
    }

    if (startIndexesMatch && endIndexesMatch)
    {
        return;
    }

    // if ctrl key is active, act as if discontiguous mode is on
    if (this.m_discontiguousSelection || (event && this.m_utils.ctrlEquivalent(event) && event.button == 0))
    {
        // We also need to overwrite the old selection instance with a new one
        // so clone the old one, update, and then replace so that the object passed
        // as the previous matches the old reference and the new selection is a new
        // reference
        selection = previous.slice(0);

        // unhighlight the last range
        this.unhighlightRange(currentRange);        

        // remove the current range and put it at the end
        selection.splice(lastSelectionIndex, 1);
    }
    else
    {
        // not keeping selections so clear selection and unhighlight.
        selection = [];
        this.unhighlightSelection();
    }
     

    selection.push(newRange);
    this.m_selection = selection;

    // if ctrl key is active, act as if discontiguous mode is on
    if (this.m_discontiguousSelection || (event && this.m_utils.ctrlEquivalent(event) && event.button == 0))
    {
        for (var i = 0; i < this.m_selection.length; i ++)
        {
            this.highlightRange(this.m_selection[i]);
        }
    }
    else
    {
        this.highlightRange(newRange, true);
    }

    // focus on the frontier cell
    this._makeSelectionFrontierFocus();

    this._compareSelectionAndFire(event, previous);
};

/**
 * Reset focus on selection frontier
 * @private
 */
DvtDataGrid.prototype._resetSelectionFrontierFocus = function()
{
    var range, cell;

    // make sure there is a selection frontier and it's not the same as the active cell
    if (this.m_selectionFrontier == null || (this._isDatabodyCellActive() && this.m_selectionFrontier['row'] == this.m_active['indexes']['row'] && this.m_selectionFrontier['column'] == this.m_active['indexes']['column']))
    {
        return;
    }

    range = this.createRange(this.m_selectionFrontier);
    cell = this.getElementsInRange(range);

    if (cell != null && cell.length > 0)
    {
        this._unsetAriaProperties(cell[0]);
    }
};

/**
 * Make the selection frontier focusable.
 * @private
 */
DvtDataGrid.prototype._makeSelectionFrontierFocus = function()
{
    var range, rowOrCell, cell;

    // make sure there is a selection frontier and it's not the same as the active cell
    if (this.m_selectionFrontier == null || (this._isDatabodyCellActive() && this.m_selectionFrontier['row'] == this.m_active['indexes']['row'] && this.m_selectionFrontier['column'] == this.m_active['indexes']['column']))
    {
        return;
    }

    // unset focus properties on active cell first
    if (this._isDatabodyCellActive())
    {
        range = this.createRange(this.m_active['indexes']);
        cell = this.getElementsInRange(range);

        if (cell != null && cell.length > 0)
        {
            this._unsetAriaProperties(cell[0]);
        }
    }

    range = this.createRange(this.m_selectionFrontier);
    rowOrCell = this.getElementsInRange(range);
    if (rowOrCell == null || rowOrCell.length == 0)
    {
        return;
    }

    // update context info
    this._updateContextInfo(this.m_selectionFrontier);

    // focus on the cell (or first cell in the row)
    cell = this.m_utils.containsCSSClassName(rowOrCell[0], this.getMappedStyle('row')) ? rowOrCell[0]['firstChild'] : rowOrCell[0];
    this._setAriaProperties(this._createActiveObject(cell), null, cell);

};

/**
 * Selects the row or column of the specified header,
 * @param {string} axis - the string axis of the header
 * @param {number} index - the index of the header selected.
 * @param {number} level - the level of the header selected.
 * @param {Event} event - the event causing the header selection
 */
DvtDataGrid.prototype._selectHeader = function(axis, index, level, event)
{
    var start, end, elem;

    // get indices and select rows/columns underneath header
    if ((axis.indexOf('row') != -1 && this.m_rowHeaderLevelCount - 1 === level) || (axis.indexOf('column') != -1 && this.m_columnHeaderLevelCount - 1 === level))
    {
        start = index;
        end = index;
    }
    else 
    {
        elem = this._getActiveElement();
        start = /** @type {number} */ (this._getAttribute(elem['parentNode'], 'start', true));
        end = start + this._getAttribute(elem['parentNode'], 'extent', true) - 1;
    }

    if (axis.indexOf('row') != -1)
    {
        // block selection on row based mode single
        if ((start == end && !this.isMultipleSelection()) || this.isMultipleSelection())
        {
            this.setHeaderSelectionFrontier(axis, end, index, level, /** @type {Element} */ (event.target), true);
    
            //handle the space key in headers for rows
            this._selectEntireRow(start, end, event);
        }

        // announce to screen reader, no need to include context info
        this._setAccInfoText('accessibleRowSelected', {'row': index + 1});
    }
    else if (axis.indexOf('column') != -1) 
    {
        this.setHeaderSelectionFrontier(axis, end, index, level, /** @type {Element} */ (event.target), true);

        //handle the space key in headers for columns
        this._selectEntireColumn(start, end, event);

        // announce to screen reader, no need to include context info
        this._setAccInfoText('accessibleColumnSelected', {'column': index + 1});        
    }

    // In case we are row based single selection mode
    if (this.isMultipleSelection())
    {
        this.m_headerDragState = true;
    }
}

/**
 * Sets the selection frontier for headers
 * @param {string} axis - the string axis of the header
 * @param {number} end - the index number of the far row/column from the anchor
 * @param {number} index - the index of the header selected.
 * @param {number} level - the level of the header selected.
 * @param {Element} target target cell or header of the event
 * @param {boolean} reset - whether or not to clear the selection frontier
 */
DvtDataGrid.prototype.setHeaderSelectionFrontier = function(axis, end, index, level, target, reset)
{
    if (reset)
    {
        this.m_selectionFrontier = {};
    }
    this.m_selectionFrontier['axis'] = axis;
    this.m_selectionFrontier['end'] = end;
    this.m_selectionFrontier['index'] = index;
    this.m_selectionFrontier['level'] = level;
    this.m_selectionFrontier['target'] = target;
}

/**
 * Updates the selection for headers after a fetch
 */
DvtDataGrid.prototype.updateSelectionHeader = function()
{
    if (this.m_selectionFrontier && this.m_selectionFrontier['target'])
    {
        if (this.m_utils.isTouchDevice())
        {
            this.m_touchSelectAnchor = this.m_selection[this.m_selection.length - 1]['startIndex'];
        }
        this.extendSelectionHeader(this.m_selectionFrontier['target'], null)
    }
};

/**
 * Selects the focus on the specified element, if ctrl+click to add cell/row to the current selection,
 * set the augment flag
 * Select and focus is an asynchronus call
 * @param {Object} index - the end index of the selection.
 * @param {Event=} event - the event causing the selection and setting active
 * @param {boolean=} augment - true if we are augmenting the selecition, default to false
 */
DvtDataGrid.prototype.selectAndFocus = function(index, event, augment)
{
    if (augment == null)
    {
        augment = false
    }

    // reset any focus properties set on frontier cell
    this._resetSelectionFrontierFocus();

    // update active cell
    // if virtual we will still want the new selection to be applied
    this._setActiveByIndex(index, event);

    if (this.m_options.getSelectionMode() == "row")
    {
        index = this.createIndex(index['row']);
    }    
    
    // need the selection frontier maintained until final callback
    var returnObj = this._getSelectionStartAndEnd(index, index, 0)
    var minIndex = returnObj['min'];
    var maxIndex = returnObj['max'];
    
    // update selection mode
    if (this.m_options.getSelectionMode() == "row")
    {
        // drop the column index
        minIndex = this.createIndex(minIndex['row']);
        maxIndex = this.createIndex(maxIndex['row']);
    }
    
    // ensure end index is specified when push to selection
    this._createRangeWithKeys(minIndex, maxIndex, this._selectAndFocusRangeCallback.bind(this, minIndex, event, augment));
};

/**
 * Continue to selectAndFocus and _selectAndFocusActiveCallback
 * @param {Object} index - the end index of the selection.
 * @param {Event|undefined} event - the event causing the selection to to be changed
 * @param {boolean} augment - true if selection being augmented
 * @param {Object} range - the range of the selection.
 * @private
 */
DvtDataGrid.prototype._selectAndFocusRangeCallback = function(index, event, augment, range)
{
    var selection, previous, classNames;

    previous = this.GetSelection();
    selection = previous.slice(0);
    if (!augment)
    {
        // if we are not augmenting the selection modify the old one appropriately
        if (!this.m_discontiguousSelection && event && !(this.m_utils.ctrlEquivalent(event) && event.button == 0))
        {
            this.unhighlightSelection();
            // this should be a new selection
            selection = [];
        }
        // this is for the Shift + F8 navigate case, we are adding to the selection on every arrow,
        // but if the user is trying to navigate away we are always popping the last selection off because
        // it was just used to navigate away, do not do this on touch because their is no navigation concept
        else if (this._isDatabodyCellActive() && this.m_prevActive != null && this.m_prevActive['type'] == 'cell' &&
                this.m_selectionFrontier['row'] == this.m_prevActive['indexes']['row'] &&
                this.m_selectionFrontier['column'] == this.m_prevActive['indexes']['column'] &&
                !this.m_utils.isTouchDevice() && (event.keyCode || !this.isMultipleSelection()))
        {
            // remove the last selection
            selection.pop();

            // unhighlight previous (active and selection)
            // only if it's not in an existing selection
            if (!this._isContainSelection(this.m_prevActive['indexes'], selection))
            {
                this._unhighlightElement(this._getCellByIndex(this.m_prevActive['indexes']), ['selected', 'topSelected', 'bottomSelected']);
            }
        }
    }

    this.m_selectionFrontier = index;

    // We need to overwrite the old selection instance with a new one
    // so clone the old one, update, and then replace so that the object passed
    // as the previous matches the old reference and the new selection is a new
    // reference
    selection.push(range);
    this.m_selection = selection;

    this.highlightRange(range);

    this._compareSelectionAndFire(event, previous);
};

/********************* end key handler methods ************************************/

/********************* focusable/editable element related methods *****************/
/**
 * Compare the selection to a clone and fire selection event if it has changed
 * @param {Event|undefined} event the DOM event to pass off in the selection event
 * @param {Object} clone the old selection object
 * @private
 */
DvtDataGrid.prototype._compareSelectionAndFire = function(event, clone)
{
    var selection = this.GetSelection();
    //only deal with touch affordances if multiple selection on touch
    if (this.m_utils.isTouchDevice() && this.isMultipleSelection() && selection.length > 0)
    {
        this._addTouchSelectionAffordance(event);
        this._moveTouchSelectionAffordance();
    }

    // fire event if selection has changed
    if (!this._compareSelections(selection, clone))
    {
        this.fireSelectionEvent(event, clone);
    }
};

/**
 * Add the touch affordance to the grid. It will be added to the row containing the active cell in row/cell selection mode.
 * Sets the position of the affordance to be on the corner of a cell in cell selection or the center of the viewport in row
 * selection.
 * @param {Event|undefined} event the event that drives the need for touch affordance
 * @private
 */
DvtDataGrid.prototype._addTouchSelectionAffordance = function(event)
{
    //icon in the corner
    var cell, iconSize, topIcon, bottomIcon, selectionMode, left, dir, target, cellContext, cellTop, cellBottom;
    if (this.m_topSelectIconContainer == null && this.m_bottomSelectIconContainer == null)
    {
        dir = this.getResources().isRTLMode() ? "right" : "left";
        iconSize = this._getTouchSelectionAffordanceSize();

        target = /** @type {Element} */ (event.target);
        cell = this.findCell(target);
        // if cell not found, try header
        if (!cell)
        {
            cell = this.findHeader(target);
        }
        // if no cell is present, it is not a header or a cell. Probably corner click to select all. 
        // in this case just return since we don't want affordances for select all (immutable selection).
        if (!cell)
        {
            return;
        }
        cellContext = cell[this.getResources().getMappedAttribute("context")];
        cellTop = this.getElementDir(cell, 'top');
        cellBottom = cellTop + this.getElementHeight(cell);

        //cache the containers so we always know where they are since selection object isn't always current
        //wrap the icon in a container so the touch area is larger than the icon
        this.m_topSelectIconContainer = document.createElement('div');
        this.m_topSelectIconContainer['className'] = this.getMappedStyle('toucharea');
        this.setElementDir(this.m_topSelectIconContainer, cellTop - (iconSize / 2), 'top');

        topIcon = document.createElement('div');
        topIcon['className'] = this.getMappedStyle('selectaffordancetop');
        topIcon.setAttribute('role', 'button');
        topIcon.setAttribute('aria-label', this.getResources().getTranslatedText('accessibleSelectionAffordanceTop'));
        this.m_topSelectIconContainer.appendChild(topIcon); //@HTMLUpdateOK

        this.m_bottomSelectIconContainer = document.createElement('div');
        this.m_bottomSelectIconContainer['className'] = this.getMappedStyle('toucharea');
        this.setElementDir(this.m_bottomSelectIconContainer, cellBottom - (iconSize / 2), 'top');
        bottomIcon = document.createElement('div');
        bottomIcon['className'] = this.getMappedStyle('selectaffordancebottom');
        bottomIcon.setAttribute('role', 'button');
        bottomIcon.setAttribute('aria-label', this.getResources().getTranslatedText('accessibleSelectionAffordanceBottom'));
        this.m_bottomSelectIconContainer.appendChild(bottomIcon); //@HTMLUpdateOK

        selectionMode = this.m_options.getSelectionMode();
        if (selectionMode === 'row')
        {
            left = (this.getElementWidth(this.m_databody) / 2) + this.m_currentScrollLeft - (iconSize / 2);
            this.setElementDir(this.m_topSelectIconContainer, left, dir);
            this.setElementDir(this.m_bottomSelectIconContainer, left, dir);
        }
        else if (cellContext['axis'] && cellContext['axis'].indexOf("row") != -1)
        {
            left = (this.getElementWidth(this.m_databody) / 2) + this.m_currentScrollLeft - (iconSize / 2);
            this.setElementDir(this.m_topSelectIconContainer, left, dir);
            this.setElementDir(this.m_bottomSelectIconContainer, left, dir);
        }
        else if (cellContext['axis'] && cellContext['axis'].indexOf("column") != -1)
        {
            cellTop = (this.getElementHeight(this.m_databody) / 2) + this.m_currentScrollTop - (iconSize / 2);
            this.setElementDir(this.m_topSelectIconContainer, cellTop, 'top');
            this.setElementDir(this.m_bottomSelectIconContainer, cellTop, 'top');

            left = this.getElementDir(cell, dir) - (iconSize / 2);
            this.setElementDir(this.m_topSelectIconContainer, left, dir);
            this.setElementDir(this.m_bottomSelectIconContainer, left + this.calculateColumnWidth(cell), dir);
        }
        else
        {
            left = this.getElementDir(cell, dir) - (iconSize / 2);
            this.setElementDir(this.m_topSelectIconContainer, left, dir);
            this.setElementDir(this.m_bottomSelectIconContainer, left + this.calculateColumnWidth(cell), dir);
        }

        this.m_databody.firstChild.appendChild(this.m_topSelectIconContainer); //@HTMLUpdateOK
        this.m_databody.firstChild.appendChild(this.m_bottomSelectIconContainer); //@HTMLUpdateOK
    }
};

/**
 * Finds and removes the touch selection icons from the DOM
 * @private
 */
DvtDataGrid.prototype._removeTouchSelectionAffordance = function()
{
    if (this._isDatabodyCellActive() && this.m_topSelectIconContainer && this.m_topSelectIconContainer['parentNode'])
    {
        this.m_topSelectIconContainer['parentNode'].removeChild(this.m_topSelectIconContainer);
        this.m_bottomSelectIconContainer['parentNode'].removeChild(this.m_bottomSelectIconContainer);
    }
};

/**
 * Finds and moves the touch selection affordances based on the old and new selection
 * @private
 */
DvtDataGrid.prototype._moveTouchSelectionAffordance = function()
{
    var selection, topRowCells, bottomRowCells, selectionMode, topIconCell, bottomIconCell, elementsInRange, dir, iconSize;
    dir = this.getResources().isRTLMode() ? "right" : "left";
    selection = this.GetSelection();
    if (selection.length > 0)
    {
        selectionMode = this.m_options.getSelectionMode();
        iconSize = this._getTouchSelectionAffordanceSize();

        if (this.m_topSelectIconContainer != null && this.m_bottomSelectIconContainer != null)
        {
            if (selectionMode === 'row')
            {
                topRowCells = this._getAxisCellsByKey(selection[selection.length - 1]['startKey']['row'], 'row');
                bottomRowCells = this._getAxisCellsByKey(selection[selection.length - 1]['endKey']['row'], 'row');

                this.setElementDir(this.m_topSelectIconContainer, this.getElementDir(topRowCells[0], 'top') - (iconSize / 2), 'top');
                this.setElementDir(this.m_bottomSelectIconContainer, this.getElementDir(bottomRowCells[0], 'top') + this.getElementHeight(bottomRowCells[0]) - (iconSize / 2), 'top');
            }
            else if (selection[selection.length -1]['endIndex']['row'] != -1 && selection[selection.length -1]['endIndex']['column'] == -1)
            {
                topRowCells = this._getAxisCellsByKey(selection[selection.length - 1]['startKey']['row'], 'row');
                bottomRowCells = this._getAxisCellsByKey(selection[selection.length - 1]['endKey']['row'], 'row');

                this.setElementDir(this.m_topSelectIconContainer, this.getElementDir(topRowCells[0], 'top') - (iconSize / 2), 'top');
                this.setElementDir(this.m_bottomSelectIconContainer, this.getElementDir(bottomRowCells[0], 'top') + this.getElementHeight(bottomRowCells[0]) - (iconSize / 2), 'top');

                var left = (this.getElementWidth(this.m_databody) / 2) + this.m_currentScrollLeft - (iconSize / 2);
                this.setElementDir(this.m_topSelectIconContainer, left, dir);
                this.setElementDir(this.m_bottomSelectIconContainer, left, dir);

            }
            else if (selection[selection.length -1]['endIndex']['column'] != -1 && selection[selection.length -1]['endIndex']['row'] == -1)
            {
                topRowCells = this._getAxisCellsByKey(selection[selection.length - 1]['startKey']['column'], 'column');
                bottomRowCells = this._getAxisCellsByKey(selection[selection.length - 1]['endKey']['column'], 'column');                

                var top = (this.getElementHeight(this.m_databody) / 2) + this.m_currentScrollTop - (iconSize / 2);
                this.setElementDir(this.m_topSelectIconContainer, top, 'top');
                this.setElementDir(this.m_bottomSelectIconContainer, top, 'top');

                this.setElementDir(this.m_topSelectIconContainer, this.getElementDir(topRowCells[0], 'left') - (iconSize / 2), 'left');
                this.setElementDir(this.m_bottomSelectIconContainer, this.getElementDir(bottomRowCells[0], 'left') + this.getElementWidth(bottomRowCells[0]) - (iconSize / 2), 'left');
            }
            else
            {
                //get the cells for left/right alignment
                topIconCell = this._getCellByIndex(selection[selection.length - 1]['startIndex']);
                bottomIconCell = this._getCellByIndex(selection[selection.length - 1]['endIndex']);

                this.setElementDir(this.m_topSelectIconContainer, this.getElementDir(topIconCell, 'top') - (iconSize / 2), 'top');
                this.setElementDir(this.m_bottomSelectIconContainer, this.getElementDir(bottomIconCell, 'top') + this.getElementHeight(bottomIconCell) - (iconSize / 2), 'top');
                this.setElementDir(this.m_topSelectIconContainer, this.getElementDir(topIconCell, dir) - (iconSize / 2), dir);
                this.setElementDir(this.m_bottomSelectIconContainer, this.getElementDir(bottomIconCell, dir) + this.getElementWidth(bottomIconCell) - (iconSize / 2), dir);
            }

            if (this.m_topSelectIconContainer['parentNode'] == null)
            {
                this.m_databody.firstChild.appendChild(this.m_topSelectIconContainer); //@HTMLUpdateOK
                this.m_databody.firstChild.appendChild(this.m_bottomSelectIconContainer); //@HTMLUpdateOK
            }
        }
    }
};
/**
 * Moves the touch selection affordances horizontally in the row to ensure they are in the viewport.
 * Only moved in row selection.
 * @private
 */
DvtDataGrid.prototype._scrollTouchSelectionAffordance = function()
{
    var selectionMode, newLeft, newTop, dir;
    selectionMode = this.m_options.getSelectionMode();
    if (selectionMode === 'row')
    {
        if (this.m_topSelectIconContainer != null)
        {
            dir = this.getResources().isRTLMode() ? "right" : "left";
            newLeft = (this.getElementWidth(this.m_databody) / 2) + this.m_currentScrollLeft;
            this.setElementDir(this.m_topSelectIconContainer, newLeft, dir);
            this.setElementDir(this.m_bottomSelectIconContainer, newLeft, dir);
        }
    }
    else if (selectionMode === 'cell' && this.isHeaderSelectionType(this.m_selectionFrontier))
    {
        if (this.m_topSelectIconContainer != null)
        {
            if (this.m_selectionFrontier['axis'].indexOf('row') != -1)
            {
                dir = this.getResources().isRTLMode() ? "right" : "left";
                newLeft = (this.getElementWidth(this.m_databody) / 2) + this.m_currentScrollLeft;
                this.setElementDir(this.m_topSelectIconContainer, newLeft, dir);
                this.setElementDir(this.m_bottomSelectIconContainer, newLeft, dir);
            }
            else
            {
                newTop = (this.getElementHeight(this.m_databody) / 2) + this.m_currentScrollTop;
                this.setElementDir(this.m_topSelectIconContainer, newTop, 'top');
                this.setElementDir(this.m_bottomSelectIconContainer, newTop, 'top');
            }
        }        
    }
};

/**
 * Get the touch affordance icon size
 * @return {number} the touch affordance icon size
 * @private
 */
DvtDataGrid.prototype._getTouchSelectionAffordanceSize = function()
{
    var div, divWidth;
    if (this.m_touchSelectionAffordanceSize == null)
    {
        div = document.createElement('div');
        div['className'] = this.getMappedStyle('toucharea');
        div['style']['visibilty'] = 'hidden';
        div['style']['top'] = '0px';
        div['style']['visibilty'] = '0px';
        this.m_root.appendChild(div); //@HTMLUpdateOK
        divWidth = div['offsetWidth'];
        this.m_root.removeChild(div);
        this.m_touchSelectionAffordanceSize = divWidth;
    }
    return this.m_touchSelectionAffordanceSize;
};
/**
 * @preserve Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

DvtDataGrid.SORT_ANIMATION_DURATION = 800;

/**
 * Event handler for handling mouse over event on headers.
 * @param {Event} event the DOM event
 * @private
 */
DvtDataGrid.prototype._handleSortMouseOver = function(event)
{
    var target, header;
    if (!this._databodyEmpty())
    {
        // checks if the mouse out is trigger by leaving the sort icons
        // the event can now enter the sort container
        target = /** @type {Element} */ (event.target);
        header = this.findHeader(target);
        if (header)
        {
            this._showOrHideSortIcon(header, false);
        }

        //if we are hovering the icon add hover class
        if (this.m_utils.containsCSSClassName(target, this.getMappedStyle("sortascending"))
                || this.m_utils.containsCSSClassName(target, this.getMappedStyle("sortdescending")))
        {
            this.m_utils.addCSSClassName(target, this.getMappedStyle("hover"));
        }
    }
};

/**
 * Event handler for handling mouse out event on headers.
 * @param {Event} event the DOM event
 * @private
 */
DvtDataGrid.prototype._handleSortMouseOut = function(event)
{
    var header, target, relatedTarget;
    if (!this._databodyEmpty())
    {
        target = /** @type {Element} */ (event.target);
        relatedTarget = /** @type {Element} */ (event.relatedTarget);
        header = this.findHeader(target);
        //if there is no header or we didn't just exit the content of the header
        if (header == null || relatedTarget == null ? true : header !== this.findHeader(relatedTarget))
        {
            this._showOrHideSortIcon(header, true);
        }

        //if we are done hovering the icon remove hover class
        if (this.m_utils.containsCSSClassName(target, this.getMappedStyle("sortascending"))
                || this.m_utils.containsCSSClassName(target, this.getMappedStyle("sortdescending")))
        {
            this.m_utils.removeCSSClassName(target, this.getMappedStyle("hover"));
            this.m_utils.removeCSSClassName(target, this.getMappedStyle("selected"));
        }
    }
};

/**
 * Add the selcted color on mousedown
 * @param {Element} icon the icon to set selected on
 * @private
 */
DvtDataGrid.prototype._handleSortIconMouseDown = function(icon)
{
    if (!this._databodyEmpty())
    {
        this.m_utils.addCSSClassName(icon, this.getMappedStyle("selected"));
    }
};

/**
 * Show or hide the sort indicator icons.
 * @param {Element} header the dom element of the header to switch icon direction in
 * @param {string} direction asecnding or descending to switch to
 * @private
 */
DvtDataGrid.prototype._toggleSortIconDirection = function(header, direction)
{
    var icon;
    if (header != null)
    {
        // shows the sort indicator
        icon = this._getSortIcon(header);
        if (direction === 'descending' && this.m_utils.containsCSSClassName(icon, this.getMappedStyle("sortascending")))
        {
            this.m_utils.removeCSSClassName(icon, this.getMappedStyle("sortascending"));
            this.m_utils.addCSSClassName(icon, this.getMappedStyle("sortdescending"));
        }
        else if (direction === 'ascending' && this.m_utils.containsCSSClassName(icon, this.getMappedStyle("sortdescending")))
        {
            this.m_utils.removeCSSClassName(icon, this.getMappedStyle("sortdescending"));
            this.m_utils.addCSSClassName(icon, this.getMappedStyle("sortascending"));
        }
    }
};

/**
 * Show or hide the sort indicator icons.
 * @param {Element|undefined|null} header the dom event
 * @param {boolean} hide true if hide the icons, false to show the icons
 * @private
 */
DvtDataGrid.prototype._showOrHideSortIcon = function(header, hide)
{
    var icon, sorted = false;
    // shows the sort indicator
    if (header != null)
    {
        icon = this._getSortIcon(header);
        if (this.m_sortInfo != null)
        {
            sorted = this.m_sortInfo['key'] === this._getKey(header);
        }
        if (hide === false && !sorted)
        {
            this.m_utils.removeCSSClassName(icon, this.getMappedStyle("disabled"));
            this.m_utils.addCSSClassName(icon, this.getMappedStyle("default"));
        }
        else if (hide === true && !sorted)
        {
            this.m_utils.removeCSSClassName(icon, this.getMappedStyle("default"));
            this.m_utils.addCSSClassName(icon, this.getMappedStyle("disabled"));
        }
    }
};

/**
 * Creates the sort indicator icons and the panel around them.
 * @param {Object} headerContext a header context object, contianing key
 * @return {Element} the sort indicator icons panel
 * @private
 */
DvtDataGrid.prototype._buildSortIcon = function(headerContext)
{
    var sortIcon, iconClassString, key, direction, sortContainer;
    //sort container is used to create fade effect
    sortContainer = document.createElement("div");
    sortContainer['className'] = this.getMappedStyle("sortcontainer");

    sortIcon = document.createElement("div");
    iconClassString = this.getMappedStyle("icon") + " " + this.getMappedStyle("clickableicon");
    key = this.m_sortInfo != null ? this.m_sortInfo['key'] : null;

    //handles the case where we scroll the header which was sorted on, off screen and come back to them
    if (headerContext['key'] === key)
    {
        direction = this.m_sortInfo != null ? this.m_sortInfo['direction'] : null;
        if (direction === 'ascending')
        {
            sortIcon['className'] = this.getMappedStyle("sortascending") + " " + iconClassString + " " + this.getMappedStyle("default");
        }
        else if (direction === 'descending')
        {
            sortIcon['className'] = this.getMappedStyle("sortdescending") + " " + iconClassString + " " + this.getMappedStyle("default");
        }
    }
    else
    {
        iconClassString += " " + this.getMappedStyle("disabled");
        sortIcon['className'] = this.getMappedStyle("sortascending") + " " + iconClassString;
    }
    sortContainer.appendChild(sortIcon); //@HTMLUpdateOK
    return sortContainer;
};

/**
 * Handles sorting using keyboard (enter key while focus on header).  See HandleHeaderKeyDown.
 * @param {Element} header header being sorted on
 * @param {Event} event DOM keyboard event triggering sort
 * @private
 */
DvtDataGrid.prototype._handleKeyboardSort = function(header, event)
{
    if (!this._databodyEmpty())
    {
        var direction = header.getAttribute(this.getResources().getMappedAttribute('sortDir'));
        if (direction == null || direction === "descending")
        {
            direction = "ascending";
        }
        else
        {
            direction = "descending";
        }

        this._doHeaderSort(event, header, direction);
    }
};

/**
 * Handles click on the header, this would perform the sort operation.
 * @param {Event} event the DOM event
 * @param {string=} direction asecnding or descending to sort on
 * @private
 */
DvtDataGrid.prototype._handleHeaderSort = function(event, direction)
{
    var target, header;
    if (!this._databodyEmpty())
    {
        target = /** @type {Element} */ (event.target);

        header = this.findHeader(target);
        if (header != null)
        {
            // use the class name to determine if it's asecnding or descending
            if (direction == null)
            {
                if (this.m_sortInfo != null && this.m_sortInfo['key'] === this._getKey(header))
                {
                    if (this.m_sortInfo['direction'] === 'ascending')
                    {
                        direction = 'descending';
                    }
                    else
                    {
                        direction = 'ascending';
                    }
                }
                else
                {
                    //we should get here on inital touch sort only
                    direction = 'ascending';
                }
            }
            this._doHeaderSort(event, header, direction);
        }
    }
};

/**
 * Handles click on the header, this would perform the sort operation.
 * @param {Event} event the DOM event
 * @param {string} direction asecnding or descending to switch to
 * @param {Element|undefined} header the header to sort on
 * @private
 */
DvtDataGrid.prototype._handleCellSort = function(event, direction, header)
{
    var target;
    target = /** @type {Element} */ (event.target);
    if (header != null && !this._databodyEmpty())
    {
        this._removeTouchSelectionAffordance();
        this._doHeaderSort(event, header, direction);
    }
};

/**
 * Handles click on the header, this would perform the sort operation.
 * @param {Event} event the DOM event
 * @param {Element} header the header element
 * @param {string} direction the sort direction
 * @private
 */
DvtDataGrid.prototype._doHeaderSort = function(event, header, direction)
{
    var key, axis, criteria;
    if (this.m_isSorting != true)
    {
        this.m_delayedSort = null;

        // get the key and axis
        key = this._getKey(header);
        axis = this._getAxis(header);

        this._removeSortSelection();

        // needed for toggle and screenreader
        header.setAttribute(this.getResources().getMappedAttribute('sortDir'), direction);
        this.m_sortInfo = {'event': event, 'key': key, 'axis': axis, 'direction': direction};

        //flip the icon direction
        this._toggleSortIconDirection(header, direction);
        this._addSortSelection();

        // creates the criteria object and invoke sort on the data source
        if (direction != null && key != null && axis != null)
        {
            this.m_isSorting = true;
            // show status message
            this.showStatusText();

            // invoke sort
            criteria = {"axis": axis, "key": key, "direction": direction};
            this.getDataSource().sort(criteria, {"success": this._handleSortSuccess.bind(this), "error": this._handleSortError.bind(this)});
        }

        // update screen reader alert
        this._setAccInfoText(direction === 'ascending' ? 'accessibleSortAscending' : 'accessibleSortDescending', {'id': key});
    }
    else
    {
        this.m_delayedSort = {'event': event, 'header': header, 'direction': direction};
    }
};

/**
 * Callback method invoked when the sort operation failed.
 * @private
 */
DvtDataGrid.prototype._handleSortError = function()
{
    this.hideStatusText();
};

/**
 * Remove the selected style class from the previous sorted sort icon, and add disabled back to it
 * @private
 */
DvtDataGrid.prototype._removeSortSelection = function()
{
    var oldSortedHeader, oldsortIcon;
    if (this.m_sortInfo != null)
    {
        //get the header that was sorted on and the icon within it based on the values stored in this.m_sortInfo
        oldSortedHeader = this._findHeaderByKey(this.m_sortInfo['key'], this.m_colHeader, this.getMappedStyle('colheadercell'));
        oldSortedHeader.removeAttribute(this.getResources().getMappedAttribute('sortDir'));
        oldsortIcon = this._getSortIcon(oldSortedHeader);
        //flip icon back to ascending
        this._toggleSortIconDirection(oldSortedHeader, 'ascending');
        if (this.m_sortInfo['direction'] === 'descending')
        {
            //switch back to the default ascending icon
            this.m_utils.removeCSSClassName(oldsortIcon, this.getMappedStyle("sortdescending"));
            this.m_utils.addCSSClassName(oldsortIcon, this.getMappedStyle("sortascending"));
        }
        //disable the icon to hide it, remove the selected style
        this.m_utils.addCSSClassName(oldsortIcon, this.getMappedStyle('disabled'));
        this.m_utils.removeCSSClassName(oldsortIcon, this.getMappedStyle('default'));
        this.m_utils.removeCSSClassName(this._getSortContainer(oldSortedHeader), this.getMappedStyle('enabled'));
    }
};

/**
 * Add the selected style class to the newly sorted sort icon and remove disabled from it
 * @private
 */
DvtDataGrid.prototype._addSortSelection = function()
{
    var sortedHeader, sortIcon;
    if (this.m_sortInfo != null)
    {
        //get the header that is sorted on and the icon within it based on the values stored in this.m_sortInfo
        sortedHeader = this._findHeaderByKey(this.m_sortInfo['key'], this.m_colHeader, this.getMappedStyle('colheadercell'));
        sortIcon = this._getSortIcon(sortedHeader);

        //select the icon to show it, remove the disabled style
        this.m_utils.addCSSClassName(sortIcon, this.getMappedStyle('default'));
        this.m_utils.removeCSSClassName(sortIcon, this.getMappedStyle('disabled'));
        this.m_utils.removeCSSClassName(sortIcon, this.getMappedStyle('selected'));
        this.m_utils.addCSSClassName(this._getSortContainer(sortedHeader), this.getMappedStyle('enabled'));
    }
};

/**
 * Determine the axis of the header.
 * @param {Element} header the header to determine the axis, returns either "row" or "column".
 * @return {string|null} the axis of the header
 * @private
 */
DvtDataGrid.prototype._getAxis = function(header)
{
    var columnHeaderCellClassName, rowHeaderCellClassName;

    columnHeaderCellClassName = this.getMappedStyle("colheadercell");
    rowHeaderCellClassName = this.getMappedStyle("rowheadercell");

    if (this.m_utils.containsCSSClassName(header, columnHeaderCellClassName))
    {
        return "column";
    }

    if (this.m_utils.containsCSSClassName(header, rowHeaderCellClassName))
    {
        return "row";
    }

    return null;
};

/**
 * Callback method invoked when the sort operation completed successfully.
 * @private
 */
DvtDataGrid.prototype._handleSortSuccess = function()
{
    // hide the message
    this.hideStatusText();

    // sort is completed successfully, now fetch the sorted data
    if (this._isDatabodyCellActive())
    {
        // scroll position should go to the new active cell location if virtual
        this._indexes({'row': this.m_active['keys']['row'], 'column': this.m_active['keys']['column']}, this._handlePreSortScrolling);
    }
    else
    {
        // scroll position should remain unchanged if high watermark or virtual without an active cell
        this._fetchForSort(this.m_startRow, this.m_endRow - this.m_startRow + 1, false);
    }
};

/**
 * Handle scrolling of the datagrid before fetching the data
 * @param {Object} indexes index of the new location of the active cell
 */
DvtDataGrid.prototype._handlePreSortScrolling = function(indexes)
{
    var rowIndex, startRow, cellTop, cellBottom, startRowPixel, isInRenderedRange, isHighWatermark, isInVisibleRange;
    rowIndex = indexes['row'] === -1 ? 0 : indexes['row'];
    cellTop = rowIndex * this.m_avgRowHeight;
    cellBottom = cellTop + this.m_avgRowHeight

    isHighWatermark = this._isHighWatermarkScrolling();
    isInRenderedRange = this._isInViewport(indexes) != -1;
    isInVisibleRange = this.m_currentScrollTop <= cellTop && cellBottom <= this.m_currentScrollTop + this.getElementHeight(this.m_databody);

    // cell is in rendered range and visible, or high watermark regardless of visibilty,
    // do a refetch of the current viewport and no scrolling
    if (isInVisibleRange || (isHighWatermark)) //&& !isInRenderedRange))
    {
        this._fetchForSort(this.m_startRow, this.m_endRow - this.m_startRow + 1, false);
    }

    // we have deceided not to prescroll on high watermark because the active cell
    // was hard to follow through the animation, if we wanted that behavior in the
    // future simlpy follow the format commented out below and above
    // cell is in rendered range but not visible on high watermak,
    // do a scroll to the new position with a refresh of the whole viewport
    //else if (isHighWatermark && isInRenderedRange)
    //{
    //    // set a new scrollTop to the top of that cell or the closest it can be
    //    this.m_currentScrollTop = Math.min(cellTop, this.m_scrollHeight);
    //    this._fetchForSort(this.m_startRow, this.m_endRow - this.m_startRow + 1, true);
    //}
    //
    // in virtual scrolling and not outside of the visible range scroll to the new location and refresh
    else
    {
        // get the scroll top that it will need to be
        this.m_currentScrollTop = Math.min(cellTop, this._getMaxScrollHeight());

        //find the start row we need to fetch at that scroll position
        startRow = Math.floor(this.m_currentScrollTop / this.m_avgRowHeight);

        startRowPixel = startRow * this.m_avgRowHeight;
        // reset ranges on rows
        this.m_startRow = startRow;
        this.m_endRow = -1;
        this.m_startRowHeader = startRow;
        this.m_endRowHeader = -1;
        this.m_startRowPixel = startRowPixel;
        this.m_endRowPixel = startRowPixel;
        this.m_startRowHeaderPixel = startRowPixel;
        this.m_endRowHeaderPixel = startRowPixel;

        this._fetchForSort(startRow, null, true);
    }
};

/**
 * A method to fetch data with the correct sort callbacks
 * @param {number} startRow
 * @param {number|null} rowCount
 * @param {boolean} scroll true if we need to pre scroll the datagrid
 */
DvtDataGrid.prototype._fetchForSort = function(startRow, rowCount, scroll)
{
    var rowHeaderFragment = document.createDocumentFragment();
    var endRowHeaderFragment = document.createDocumentFragment();
    this.fetchHeaders("row", startRow, rowHeaderFragment, endRowHeaderFragment, rowCount, {"success": this.handleHeadersFetchSuccessForSort.bind(this), "error": this.handleCellsFetchError});
    this.fetchCells(this.m_databody, startRow, this.m_startCol, rowCount, this.m_endCol - this.m_startCol + 1, {"success": this.handleCellsFetchSuccessForSort.bind(this, rowHeaderFragment, endRowHeaderFragment, scroll), "error": this.handleCellsFetchError});
};

/**
 * Handle a successful call to the data source fetchHeaders for sorting. Used to populate the new headers fragment.
 * @param {Object} headerSet - an array of headers returned from the dataSource
 * @param {Object} headerRange - {"axis":,"start":,"count":,"header":}
 * @param {boolean} rowInsert - if this is triggered by a row insert event
 * @protected
 */
DvtDataGrid.prototype.handleHeadersFetchSuccessForSort = function(headerSet, headerRange, endHeaderSet, rowInsert)
{
    var axis, headerFragment, start, headerCount, c, index, totalRowHeight, returnVal, className, renderer, endHeaderFragment;
    axis = headerRange["axis"];
    start = headerRange["start"];
    headerFragment = headerRange["header"];
    endHeaderFragment = headerRange["endHeader"];

    // remove fetching message
    this.m_fetching[axis] = false;

    if (headerSet != null)
    {
        // add the headers to the row header
        headerCount = headerSet.getCount();
        totalRowHeight = 0;
        c = 0;
        className = this.getMappedStyle("headercell") + " " + this.getMappedStyle("rowheadercell");
        renderer = this.m_options.getRenderer("row");
        while (headerCount - c > 0)
        {
            index = start + c;
            returnVal = this.buildLevelHeaders(headerFragment, index, 0, 0, this.m_startRowPixel + totalRowHeight, true, rowInsert, renderer, headerSet, 'row', className, this.m_rowHeaderLevelCount);
            c += returnVal.count;
            totalRowHeight += returnVal.totalHeaderDimension;
        }
        this.m_endRowHeader = this.m_startRowHeader + headerCount - 1;
        this.m_endRowHeaderPixel = this.m_startRowHeaderPixel + totalRowHeight;
    }

    if (endHeaderSet != null)
    {
        headerCount = endHeaderSet.getCount();
        totalRowHeight = 0;
        c = 0;
        className = this.getMappedStyle("endheadercell") + " " + this.getMappedStyle("rowendheadercell");
        renderer = this.m_options.getRenderer("rowEnd");
        while (headerCount - c > 0)
        {
            index = start + c;
            returnVal = this.buildLevelHeaders(endHeaderFragment, index, 0, 0, this.m_startRowPixel + totalRowHeight, true, rowInsert, renderer, endHeaderSet, 'rowEnd', className, this.m_rowEndHeaderLevelCount);
            c += returnVal.count;
            totalRowHeight += returnVal.totalHeaderDimension;
        }
        this.m_endRowEndHeader = this.m_startRowEndHeader + headerCount - 1;
        this.m_endRowEndHeaderPixel = this.m_startRowEndHeaderPixel + totalRowHeight;
    }

    // end fetch
    this._signalTaskEnd();
};

/**
 * Handle a successful call to the data source fetchCells after sort.
 * @param {DocumentFragment|Element} newRowHeaderElements a document fragment containing the row headers and the fragment
 * @param {boolean|null} scroll true if we need to pre scroll
 * @param {Object} cellSet a CellSet object which encapsulates the result set of cells
 * @param {Array.<Object>} cellRange [rowRange, columnRange] - [{"axis":,"start":,"count":},{"axis":,"start":,"count":,"databody":,"scroller":}]
 */
DvtDataGrid.prototype.handleCellsFetchSuccessForSort = function(newRowHeaderElements, newRowEndHeaderElements, scroll, cellSet, cellRange)
{
    var rowRange, rowStart, rowCount, columnRange, columnStart, columnCount, newCellElements, oldCellElements,
            oldRowHeaderElements, oldRowEndHeaderElements, duration, returnVal, animate;

    this.m_fetching['cells'] = false;

    duration = DvtDataGrid.SORT_ANIMATION_DURATION;

    // size the grid if fetch is done
    if (this.isFetchComplete())
    {
        this.hideStatusText();
    }

    // obtain params for _addCellsToFragment
    rowRange = cellRange[0];
    rowStart = rowRange['start'];
    rowCount = cellSet.getCount("row");

    columnRange = cellRange[1];
    columnStart = columnRange['start'];
    columnCount = cellSet.getCount("column");

    // the rows AFTER sort should be inside the newCellElements fragment
    newCellElements = document.createDocumentFragment();

    returnVal = this._addCellsToFragment(newCellElements, cellSet, rowStart, this.m_startRowPixel, columnStart, this.m_startColPixel);
    this.m_endRow = this.m_startRow + rowCount - 1;
    this.m_endRowPixel = this.m_startRowPixel + returnVal['totalRowHeight'];

    oldCellElements = this.m_databody['firstChild'];
    oldRowHeaderElements = this.m_rowHeader['firstChild'];
    oldRowEndHeaderElements = this.m_rowEndHeader['firstChild'];

    if (scroll == true)
    {
        // disable animation on virtual scrolling
        animate = this._isHighWatermarkScrolling();

        // scroll the databody
        if (!this.m_utils.isTouchDevice())
        {
            this.m_silentScroll = true;
            this.m_databody['scrollTop'] = this.m_currentScrollTop;
            this._syncScroller();
        }
        else
        {
            // for touch we'll call scrollTo directly instead of relying on scroll event to fire due to performance
            this._disableTouchScrollAnimation();
            this.scrollTo(this.m_currentScrollLeft, this.m_currentScrollTop);
        }
    }

    // if there's only one row we don't need to animate
    // don't animate on multi-level headers
    if (!duration || duration == 0 || !this.m_utils.supportsTransitions() || rowCount === 1 ||
            (this.m_rowHeaderLevelCount > 1 && this.m_rowHeaderLevelCount != null) || animate === false)
    {
        // start task since both animation/non use handle sort end
        this._signalTaskStart();
        this._handleSortEnd(newCellElements, newRowHeaderElements, newRowEndHeaderElements);
    }
    else
    {
        this.processSortAnimationToPosition(duration, 0, "ease-in", oldRowHeaderElements, newRowHeaderElements, oldCellElements, newCellElements, oldRowEndHeaderElements, newRowEndHeaderElements);
    }

    // end fetch
    this._signalTaskEnd();
};

/**
 * Handles a sort complete by replacing the dom with the new headers and cells,
 * restoring active and firing a sort event
 * @param {DocumentFragment} newCellElements
 * @param {DocumentFragment|Element} newRowHeaderElements
 */
DvtDataGrid.prototype._handleSortEnd = function(newCellElements, newRowHeaderElements, newRowEndHeaderElements)
{
    var headerContent, databodyContent;
    if (newRowHeaderElements.childNodes.length > 1)
    {
        headerContent = this.m_rowHeader['firstChild'];
        this.m_utils.empty(headerContent);
        headerContent.appendChild(newRowHeaderElements); //@HTMLUpdateOK
    }

    if (newRowEndHeaderElements.childNodes.length > 1)
    {
        headerContent = this.m_rowEndHeader['firstChild'];
        this.m_utils.empty(headerContent);
        headerContent.appendChild(newRowEndHeaderElements); //@HTMLUpdateOK
    }

    databodyContent = this.m_databody['firstChild'];

    this._emptyDatabody(databodyContent);
    this._populateDatabody(databodyContent, newCellElements);

    // restore active cell
    this._fireSortEvent();
    this._restoreActive();
    this.m_isSorting = false;
    this._doDelayedSort();

    // end animation/sort
    this.m_animating = false;
    this._signalTaskEnd();

    // check event queue for outstanding model events
    this._runModelEventQueue();
};

/**
 * The main method for animation of the DataGrid rows from before-sort to the after-sort potitions
 * @param {number} duration the duration of animation
 * @param {number} delay_offset the initial delay of animation
 * @param {string} timing the easing function
 * @param {Element} oldRowHeaderElements the DOM structure on which the animation will be performed. Initially contains DOM elements in before sorting order
 * @param {DocumentFragment|Element} newRowHeaderElements the element that contains set of sub-elements in "after-sorting" order
 * @param {Object} oldElementSet the DOM structure on which the animation will be performed. Initially contains DOM elements in before sorting order
 * @param {DocumentFragment} newElementSet the element that contains set of sub-elements in "after-sorting" order
 * @private
 */
DvtDataGrid.prototype.processSortAnimationToPosition = function(duration, delay_offset, timing, oldRowHeaderElements, newRowHeaderElements, oldElementSet, newElementSet, oldRowEndHeaderElements, newRowEndHeaderElements)
{
    var self, rowKey, animationInformation, oldTop, newTop, cellsForAppend, rowHeadersForAppend, i, child,
            rowHeaderSupport, oldBottom, newBottom, newElementSetClone, newRowHeaderElementsClone, dir,
            viewportTop, viewportBottom, lastAnimationElement, rowEndHeaderSupport, newRowEndHeaderElementsClone,
            rowEndHeadersForAppend, clone, cloneHeader, left, right, newAnimationInformation, viewportLeft, viewportRight;

    // initialize variables
    self = this;
    // animation start
    this._signalTaskStart();
    cellsForAppend = [];
    rowHeadersForAppend = [];
    rowEndHeadersForAppend = [];
    rowHeaderSupport = newRowHeaderElements.childNodes.length > 1 ? true : false;
    rowEndHeaderSupport = newRowEndHeaderElements.childNodes.length > 1 ? true : false;
    dir = this.getResources().isRTLMode() ? "right" : "left";
    viewportTop = this._getViewportTop();
    viewportBottom = this._getViewportBottom();
    viewportLeft = this._getViewportLeft();
    viewportRight = this._getViewportRight();

    //animation information will be an object of objects as follows {key:{oldTop:val, newTop:val}}
    animationInformation = {};
    newAnimationInformation = {};

    // loop over the old elements and set their old and new tops which should eb to the bottom of the viewport if visible
    for (i = 0; i < oldElementSet.childNodes.length; i++)
    {
        child = oldElementSet.childNodes[i];
        rowKey = this._getKey(child, 'row');
        left = this.getElementDir(child, dir);
        right = left + this.getElementWidth(child);
        oldTop = this.getElementDir(child, 'top');
        oldBottom = oldTop + this.getElementHeight(child);

        if (this._isCellBoundryInViewport(left, right, oldTop, oldBottom))
        {
            if (!animationInformation.hasOwnProperty(rowKey))
            {
                newTop = viewportBottom;
                animationInformation[rowKey] = {cell: [child], oldTop: oldTop, newTop: newTop};
            }
            else
            {
                animationInformation[rowKey].cell.push(child);
            }

            if (newTop - oldTop != 0)
            {
                lastAnimationElement = child;
            }
        }
    }

    for (i = 0; i < oldRowHeaderElements.childNodes.length; i++)
    {
        if (rowHeaderSupport)
        {
            child = oldRowHeaderElements.childNodes[i];
            rowKey = this._getKey(child, 'row');
            oldTop = this.getElementDir(child, 'top');
            oldBottom = oldTop + this.getElementHeight(child);

            if (this._isCellBoundryInViewport(viewportLeft, viewportRight, oldTop, oldBottom))
            {
                newTop = viewportBottom;
                if (!animationInformation.hasOwnProperty(rowKey))
                {
                    animationInformation[rowKey] = {oldTop: oldTop, newTop: newTop};
                }
                animationInformation[rowKey].rowHeader = child;
            }
        }
    }

    for (i = 0; i < oldRowEndHeaderElements.childNodes.length; i++)
    {
        if (rowHeaderSupport)
        {
            child = oldRowEndHeaderElements.childNodes[i];
            rowKey = this._getKey(child, 'row');
            oldTop = this.getElementDir(child, 'top');
            oldBottom = oldTop + this.getElementHeight(child);

            if (this._isCellBoundryInViewport(viewportLeft, viewportRight, oldTop, oldBottom))
            {
                newTop = viewportBottom;
                if (!animationInformation.hasOwnProperty(rowKey))
                {
                    animationInformation[rowKey] = {oldTop: oldTop, newTop: newTop};
                }
                animationInformation[rowKey].rowEndHeader = child;
            }
        }
    }

    // loop over the new elements and set their old and new tops
    for (i = 0; i < newElementSet.childNodes.length; i++)
    {
        child = newElementSet.childNodes[i];
        rowKey = this._getKey(child, 'row');
        newTop = this.getElementDir(child, 'top');

        // if in the old elements, just replace its newTop value
        if (animationInformation.hasOwnProperty(rowKey))
        {
            animationInformation[rowKey].newTop = newTop;
        }
        // if not in the old elements, create an entry for it with oldTop just outside of view
        else
        {
            oldTop = viewportBottom;
            newBottom = newTop + this.getElementHeight(child);
            left = this.getElementDir(child, dir);
            right = left + this.getElementWidth(child);

            // if new element will be in view at all, we will need to add it to the live DOM
            // if the row is not visible at all, it will not need to move
            if (this._isCellBoundryInViewport(left, right, newTop, newBottom))
            {
                clone = child.cloneNode(true);
                clone[this.getResources().getMappedAttribute("context")] = child[this.getResources().getMappedAttribute("context")];
                this.setElementDir(clone, oldTop, 'top');
                oldElementSet.appendChild(clone);

                if (!newAnimationInformation.hasOwnProperty(rowKey))
                {
                    newAnimationInformation[rowKey] = {cell: [clone], oldTop: oldTop, newTop: newTop, add: true};
                }
                else
                {
                    newAnimationInformation[rowKey].cell.push(clone);
                }

                if (newTop - oldTop != 0)
                {
                    lastAnimationElement = clone;
                }
            }
        }
    }

    // loop over the new elements and set their old and new tops
    for (i = 0; i < newRowHeaderElements.childNodes.length; i++)
    {
        if (rowHeaderSupport)
        {
            var rowHeader = newRowHeaderElements.childNodes[i];
            rowKey = this._getKey(rowHeader, 'row');
            if (!animationInformation.hasOwnProperty(rowKey))
            {        
                newTop = this.getElementDir(rowHeader, 'top');
                newBottom = newTop + this.getElementHeight(rowHeader);
                oldTop = viewportBottom;

                if (this._isCellBoundryInViewport(viewportLeft, viewportRight, newTop, newBottom))
                {
                    cloneHeader = rowHeader.cloneNode(true);
                    cloneHeader[this.getResources().getMappedAttribute("context")] = rowHeader[this.getResources().getMappedAttribute("context")];
                    this.setElementDir(cloneHeader, oldTop, 'top');
                    oldRowHeaderElements.appendChild(cloneHeader);
                    if (!newAnimationInformation.hasOwnProperty(rowKey))
                    {
                        newAnimationInformation[rowKey] = {oldTop: oldTop, newTop: newTop, add: true};
                    }
                    newAnimationInformation[rowKey].rowHeader = cloneHeader;
                }
            }
        }
    }

    // loop over the new elements and set their old and new tops
    for (i = 0; i < newElementSet.childNodes.length; i++)
    {
        if (rowEndHeaderSupport)
        {
            var rowEndHeader = newRowHeaderElements.childNodes[i];
            rowKey = this._getKey(rowEndHeader, 'row');
            if (!animationInformation.hasOwnProperty(rowKey))
            {                
                newTop = this.getElementDir(rowEndHeader, 'top');
                newBottom = newTop + this.getElementHeight(rowEndHeader);
                oldTop = viewportBottom;

                if (this._isCellBoundryInViewport(viewportLeft, -1, newTop, newBottom))
                {
                    cloneHeader = rowEndHeader.cloneNode(true);
                    cloneHeader[this.getResources().getMappedAttribute("context")] = rowEndHeader[this.getResources().getMappedAttribute("context")];
                    this.setElementDir(cloneHeader, oldTop, 'top');
                    oldRowEndHeaderElements.appendChild(cloneHeader);
                    newAnimationInformation[rowKey].rowEndHeader = cloneHeader;
                    if (!newAnimationInformation.hasOwnProperty(rowKey))
                    {
                        newAnimationInformation[rowKey] = {oldTop: oldTop, newTop: newTop, add: true};
                    }
                }
            }
        }
    }

    // if nothing is animated which could happen, we can just bail and not animate
    if (lastAnimationElement != null)
    {
        var transitionListener = this._handleSortEnd.bind(this, newElementSet, newRowHeaderElements, newRowEndHeaderElements);
        
        //register transitionend listener on the last row transitioning before applying the transition
        self._onTransitionEnd(lastAnimationElement, transitionListener, duration);

        this.m_animating = true;

        setTimeout(function()
        {
            var deltaY, delay, animationInfo, i, row;

            for (var rowKey in animationInformation)
            {
                row = animationInformation[rowKey];
                delay = delay_offset + "ms";
                deltaY = row.newTop - row.oldTop;
                if (row.cell)
                {
                    for (i = 0; i < row.cell.length; i++)
                    {
                        self.addTransformMoveStyle(row.cell[i], duration / 2 + "ms", delay, timing, 0, deltaY, 0);
                    }
                }
                if (rowHeaderSupport && row.rowHeader)
                {
                    self.addTransformMoveStyle(row.rowHeader, duration / 2 + "ms", delay, timing, 0, deltaY, 0);
                }
                if (rowEndHeaderSupport && row.rowEndHeader)
                {
                    self.addTransformMoveStyle(row.rowEndHeader, duration / 2 + "ms", delay, timing, 0, deltaY, 0);
                }
            }

            for (var rowKey in newAnimationInformation)
            {
                row = newAnimationInformation[rowKey];
                delay = delay_offset * i + "ms";
                deltaY = row.newTop - row.oldTop;
                if (row.cell)
                {
                    for (i = 0; i < row.cell.length; i++)
                    {
                        self.addTransformMoveStyle(row.cell[i], duration / 2 + "ms", delay, timing, 0, deltaY, 0);
                    }
                }
                if (rowHeaderSupport && row.rowHeader)
                {
                    self.addTransformMoveStyle(row.rowHeader, duration / 2 + "ms", delay, timing, 0, deltaY, 0);
                }
                if (rowEndHeaderSupport && row.rowEndHeader)
                {
                    self.addTransformMoveStyle(row.rowEndHeader, duration / 2 + "ms", delay, timing, 0, deltaY, 0);
                }
            }
        }, 0);
    }
    else
    {
        this._handleSortEnd(newElementSet, newRowHeaderElements, newRowEndHeaderElements);
    }
};

/**
 * Restore the active cell after sort.
 * @private
 */
DvtDataGrid.prototype._restoreActive = function()
{
    var cell, rowHeader, cellIndex, axis, root, className, event;
    if (this.m_active != null)
    {
        axis = this.m_active['axis'];
        event = this.m_sortInfo['originalEvent']; // Use the sort event as the close event's originalEvent
        if (this.m_active['type'] == 'cell')
        {
            // see if the cell exists after the sort
            cell = this._getCellByKeys(this.m_active['keys']);
            if (cell != null)
            {
                cellIndex = this.getCellIndexes(cell);

                // select it if selection enabled
                if (this._isSelectionEnabled())
                {
                    // this will clear the selection if there's multiple selection before sort
                    // this is the behavior we want since the ranges in the previous selection
                    // will in most cases be invalid after sort.  The only one we can maintain and
                    // make sense to do is the active cell
                    this.selectAndFocus(cellIndex, event);
                }
                else
                {
                    // make it active
                    this._setActiveByIndex(cellIndex, event);
                }
            }
            else
            {
                this._setActive(null, null, event, true);
                // clear selection it if selection enabled
                if (this._isSelectionEnabled())
                {
                    this._clearSelection(event);
                }
            }
        }
        else if (axis == 'row' || axis == 'rowEnd')
        {
            root = axis == 'row' ? this.m_rowHeader : this.m_rowEndHeader;
            className = axis == 'row' ?  this.getMappedStyle('rowheadercell') : this.getMappedStyle('rowendheadercell');
            rowHeader = this._findHeaderByKey(this.m_active['key'], root, className);
            if (rowHeader != null)
            {
                // make it active
                // Need to add context to scroll databody first before making
                // the row header focus to make it focus properly.
                var rowContext = rowHeader[this.getResources().getMappedAttribute("context")];
                this._setActive(rowHeader, {'type': 'header', 'axis': axis, 'index': rowContext['index'], 'key': rowContext['key'], 'level': rowContext['level']}, event);
            }
            else
            {
                this._setActive(null, null, event);
            }
        }
    }
};

/**
 * Get the default sort icon size
 * @return {number} the default sort icon size
 */
DvtDataGrid.prototype.getSortIconSize = function()
{
    var div, divWidth;
    if (this.m_sortIconSize == null)
    {
        div = document.createElement('div');
        div['className'] = this.getMappedStyle("sortascending") + " " + this.getMappedStyle("icon") + " " + this.getMappedStyle("clickableicon");
        div['style']['visibilty'] = 'hidden';
        div['style']['top'] = '0px';
        div['style']['visibilty'] = '0px';
        this.m_root.appendChild(div); //@HTMLUpdateOK
        divWidth = div['offsetWidth'];
        this.m_root.removeChild(div);
        this.m_sortIconSize = divWidth;
    }
    return this.m_sortIconSize;
};

/**
 * Gets the sort icon froma  header Element
 * @param {Element} header the header to get sort icon for
 * @private
 */
DvtDataGrid.prototype._getSortIcon = function(header)
{
    //presently guaranteed to be the first child of the last child of the parent
    return header['lastChild']['firstChild'];
};

/**
 * Gets the sort container froma  header Element
 * @param {Element} header the header to get sort container for
 * @private
 */
DvtDataGrid.prototype._getSortContainer = function(header)
{
    //presently guaranteed to be thethe last child of the parent
    return header['lastChild'];
};

/**
 * Fire Sort event
 * @private
 */
DvtDataGrid.prototype._fireSortEvent = function()
{
    var details = {
        'event': this.m_sortInfo['event'], 'ui': {
            'header': this.m_sortInfo['key'],
            'direction': this.m_sortInfo['direction']
        }
    };
    this.fireEvent("sort", details);
};

/**
 * Start a delayed sort
 * @private
 */
DvtDataGrid.prototype._doDelayedSort = function()
{
    if (this.m_delayedSort != null)
    {
        this._doHeaderSort(this.m_delayedSort['event'], this.m_delayedSort['header'], this.m_delayedSort['direction']);
    }
    else
    {
        // no pending sort so cleanup
        this.fillViewport();
    }
};
/**
 * @preserve Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
DvtDataGrid.RESIZE_OFFSET = 5;
DvtDataGrid.RESIZE_TOUCH_OFFSET = 8;

/**
 * Handles what to do when the mouse moves. Sets the cursor based on .manageHeaderCursor(),
 * If this.m_isResizing is set to true, treats movement as resizing, calling .handleResizeMouseMove()
 * @param {Event} event - a mousemove event
 */
DvtDataGrid.prototype.handleResize = function(event)
{
    var header;
    //if not resizing, monitor the cursor position, otherwise handle resizing
    if (this.m_isResizing === false)
    {
        header = this.find(event['target'], 'header');
        if (header == null)
        {
            header = this.find(event['target'], 'endheader');
        }
        //only if we are inside our grid's headers, multiple grid case
        if (header != null && (header == this.m_rowHeader || header == this.m_colHeader || header == this.m_rowEndHeader || header == this.m_colEndHeader))
        {
            this.m_cursor = this.manageHeaderCursor(event);
            if (this.m_resizingElement != null)
            {
                if (this.m_cursor == 'default')
                {
                    this.m_resizingElement['style']['cursor'] = '';
                    if (this.m_resizingElementSibling != null)
                    {
                        this.m_resizingElementSibling['style']['cursor'] = '';
                    }
                }
                else
                {
                    this.m_resizingElement['style']['cursor'] = this.m_cursor;
                    if (this.m_resizingElementSibling != null)
                    {
                        this.m_resizingElementSibling['style']['cursor'] = this.m_cursor;
                    }
                }
            }
        }
    }
    else
    {
        this.handleResizeMouseMove(event);
    }
};

/**
 * On mousedown, if the cursor was set to row/col -resize, set the required resize values.
 * @param {Event} event - a mousedown event
 * @return {boolean} true if event processed
 */
DvtDataGrid.prototype.handleResizeMouseDown = function(event)
{
    if (this.m_cursor === 'col-resize' || this.m_cursor === 'row-resize')
    {
        this.m_isResizing = true;
        if (this.m_utils.isTouchDevice())
        {
            this.m_lastMouseX = event.touches[0]['pageX'];
            this.m_lastMouseY = event.touches[0]['pageY'];
        }
        else
        {
            document.addEventListener("mousemove", this.m_docMouseMoveListener, false);
            document.addEventListener("mouseup", this.m_docMouseUpListener, false);
            this.m_lastMouseX = event['pageX'];
            this.m_lastMouseY = event['pageY'];
        }

        this.m_overResizeLeft = 0;
        this.m_overResizeMinLeft = 0;
        this.m_overResizeTop = 0;
        this.m_overResizeMinTop = 0;
        this.m_overResizeRight = 0;
        this.m_overResizeBottom = 0;

        this.m_orginalResizeDimensions = {
            'width': this.getElementWidth(this.m_resizingElement),
            'height': this.getElementHeight(this.m_resizingElement)
        };

        return true;
    }
    return false;
};

/**
 * On mouseup, if we were resizing, handle cursor and callback firing.
 * @param {Event} event - a mouseup event
 */
DvtDataGrid.prototype.handleResizeMouseUp = function(event)
{
    var size, details, newHeight, newWidth;
    if (this.m_isResizing === true)
    {
        newWidth = this.getElementWidth(this.m_resizingElement);
        newHeight = this.getElementHeight(this.m_resizingElement);
        if (newWidth != this.m_orginalResizeDimensions['width'] || newHeight != this.m_orginalResizeDimensions['height'])
        {
            //set the information we want to callback with in the resize event and callback
            size = (this.m_cursor === 'col-resize') ? this.m_resizingElement['style']['width'] : this.m_resizingElement['style']['height'];
            details = {
                'event': event,
                'ui': {
                    'header': this._getKey(this.m_resizingElement),
                    'oldDimensions' : {
                        'width': this.m_orginalResizeDimensions['width'],
                        'height': this.m_orginalResizeDimensions['height']
                    },
                    'newDimensions' : {
                        'width': this.getElementWidth(this.m_resizingElement),
                        'height': this.getElementHeight(this.m_resizingElement)
                    },
                    // deprecating this part in 2.1.0
                    'size': size
                }
            };
            this.fireEvent('resize', details);
        }
        //no longer resizing
        this.m_isResizing = false;
        this.m_cursor = 'default';
        this.m_resizingElement['style']['cursor'] = '';
        if (this.m_resizingElementSibling != null)
        {
            this.m_resizingElementSibling['style']['cursor'] = '';
        }

        this.m_resizingElement = null;
        this.m_resizingElementMin = null;
        this.m_resizingElementSibling = null;
        this.m_orginalResizeDimensions = null;

        document.removeEventListener("mousemove", this.m_docMouseMoveListener, false);
        document.removeEventListener("mouseup", this.m_docMouseUpListener, false);
        // unregister all listeners
    }
};

/**
 * Check if has data-resizable attribute is set to 'true' on a header
 * @param {Element|undefined|null} element - element to check if has data-resizable true
 * @return {boolean} true if data-resizable attribute is 'true'
 */
DvtDataGrid.prototype._isDOMElementResizable = function(element)
{
    if (element == null)
    {
        return false;
    }
    return element.getAttribute(this.getResources().getMappedAttribute('resizable')) === "true";
};

/**
 * Determine what the document cursor should be for header cells.
 * @param {Event} event - a mousemove event
 * @return {string} the cursor type for a given mouse location
 */
DvtDataGrid.prototype.manageHeaderCursor = function(event)
{
    var elem = event['target'], resizeHeaderMode, edges, cursorX, cursorY, offsetPixel,
            widthResizable, heightResizable, siblingResizable, rtl, index, level, sibling, parent,
            leftEdgeCheck, rightEdgeCheck, topEdgeCheck, bottomEdgeCheck, end;
    //determine the element/header type that should be used for resizing if applicable
    elem = this.find(event['target'], 'headercell');
    if (!elem)
    {
        elem = this.find(event['target'], 'endheadercell');
    }

    if (!elem)
    {
        return 'default';
    }

    resizeHeaderMode = this.getHeaderCellAxis(elem);
    index = this.getHeaderCellIndex(elem);
    level = this.getHeaderCellLevel(elem);

    if (resizeHeaderMode === 'column')
    {
        heightResizable = this.m_options.isResizable(resizeHeaderMode, 'height') === "enable" ? true : false;
        widthResizable = this._isDOMElementResizable(elem);
        // previous is the previous index same level
        sibling = this._getHeaderByIndex(index - 1, level, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
        siblingResizable = this._isDOMElementResizable(sibling);
        // parent is the previous level the same index
        parent = this._getHeaderByIndex(index, level - 1, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
    }
    else if (resizeHeaderMode === 'row')
    {
        widthResizable = this.m_options.isResizable(resizeHeaderMode, 'width') === "enable" ? true : false;
        heightResizable = this._isDOMElementResizable(elem);
        // previous is the previous index same level
        sibling = this._getHeaderByIndex(index - 1, level, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
        siblingResizable = this._isDOMElementResizable(sibling);
        // parent is the previous level the same index
        parent = this._getHeaderByIndex(index, level - 1, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
    }
    else if (resizeHeaderMode === 'columnEnd')
    {
        heightResizable = this.m_options.isResizable(resizeHeaderMode)['height'] === "enable" ? true : false;
        widthResizable = this._isDOMElementResizable(elem);
        // previous is the previous index same level
        sibling = this._getHeaderByIndex(index - 1, level, this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
        siblingResizable = this._isDOMElementResizable(sibling);
        // parent is the previous level the same index
        parent = this._getHeaderByIndex(index, level - 1, this.m_colEndHeader, this.m_columnEndHeaderLevelCount, this.m_startColEndHeader);
    }
    else if (resizeHeaderMode === 'rowEnd')
    {
        widthResizable = this.m_options.isResizable(resizeHeaderMode)['width'] === "enable" ? true : false;
        heightResizable = this._isDOMElementResizable(elem);
        // previous is the previous index same level
        sibling = this._getHeaderByIndex(index - 1, level, this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
        siblingResizable = this._isDOMElementResizable(sibling);
        // parent is the previous level the same index
        parent = this._getHeaderByIndex(index, level - 1, this.m_rowEndHeader, this.m_rowEndHeaderLevelCount, this.m_startRowEndHeader);
    }

    //touch requires an area 24px for touch gestures
    if (this.m_utils.isTouchDevice())
    {
        cursorX = event.touches[0]['pageX'];
        cursorY = event.touches[0]['pageY'];
        offsetPixel = DvtDataGrid.RESIZE_TOUCH_OFFSET;
    }
    else
    {
        cursorX = event['pageX'];
        cursorY = event['pageY'];
        offsetPixel = DvtDataGrid.RESIZE_OFFSET;
    }
    edges = this.getHeaderEdgePixels(elem);
    rtl = this.getResources().isRTLMode();
    end = (resizeHeaderMode === 'columnEnd' || resizeHeaderMode === 'rowEnd');

    leftEdgeCheck = cursorX < edges[0] + offsetPixel;
    topEdgeCheck = cursorY < edges[1] + offsetPixel;
    rightEdgeCheck = cursorX > edges[2] - offsetPixel;
    bottomEdgeCheck = cursorY > edges[3] - offsetPixel;

    //check to see if resizable was enabled on the grid and then check the position of the cursor to the element border
    //we always choose the element preceding the border (so for rows the header before the bottom border)
    if (resizeHeaderMode === 'column' || resizeHeaderMode === 'columnEnd')
    {
        // can we resize the width of this header
        if (widthResizable && (rtl ? leftEdgeCheck : rightEdgeCheck))
        {
            this.m_resizingElement = elem;
            return 'col-resize';
        }
        // can we resize the width of the previous header
        else if (siblingResizable && (rtl ? rightEdgeCheck : leftEdgeCheck))
        {
            this.m_resizingElement = sibling;
            this.m_resizingElementSibling = elem;
            if (this.m_resizingElement !== null)
            {
                return 'col-resize';
            }
        }
        else if (heightResizable)
        {
            // can we resize the height of this header
            if ((!end && bottomEdgeCheck) || (end && topEdgeCheck))
            {
                this.m_resizingElement = elem;
                return 'row-resize';
            }
            // can we resize the height of the parent header
            else if ((!end && topEdgeCheck) || (end && bottomEdgeCheck))
            {
                this.m_resizingElement = parent;
                this.m_resizingElementSibling = elem;
                return 'row-resize';
            }
        }
    }
    else if (resizeHeaderMode === 'row' || resizeHeaderMode === 'rowEnd')
    {
        if (heightResizable && bottomEdgeCheck)
        {
            this.m_resizingElement = elem;
            return 'row-resize';
        }
        else if (siblingResizable && topEdgeCheck)
        {
            this.m_resizingElement = sibling;
            this.m_resizingElementSibling = elem;
            if (this.m_resizingElement !== null)
            {
                return 'row-resize';
            }
        }
        if (widthResizable)
        {
            if ((!end && (rtl ? leftEdgeCheck : rightEdgeCheck)) || (end && (rtl ? rightEdgeCheck : leftEdgeCheck)))
            {
                this.m_resizingElement = elem;
                return 'col-resize';
            }
            else if ((!end && (rtl ? rightEdgeCheck : leftEdgeCheck)) || (end && (rtl ? leftEdgeCheck : rightEdgeCheck)))
            {
                this.m_resizingElement = parent;
                this.m_resizingElementSibling = elem;
                if (this.m_resizingElement !== null)
                {
                    return 'col-resize';
                }
            }
        }
    }
    return 'default';
};

/**
 * On mousemove see which type of resizing we are doing and call the appropriate resizer after calculating
 * the new elements width based on current and last X and Y page coordinates.
 * @param {Event} event - a mousemove event
 */
DvtDataGrid.prototype.handleResizeMouseMove = function(event)
{
    var resizeHeaderMode, oldElementWidth, newElementWidth, oldElementHeight, newElementHeight, end;
    //update stored mouse position
    this.m_currentMouseX = event['pageX'];
    this.m_currentMouseY = event['pageY'];

    if (this.m_utils.isTouchDevice())
    {
        this.m_currentMouseX = event.touches[0]['pageX'];
        this.m_currentMouseY = event.touches[0]['pageY'];
    }
    else
    {
        this.m_currentMouseX = event['pageX'];
        this.m_currentMouseY = event['pageY'];
    }

    //check to see if we are resizing a column or row
    if (this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('colheadercell')) ||
        this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('colendheadercell')))
    {
        resizeHeaderMode = 'column';
    }
    else
    {
        resizeHeaderMode = 'row';
    }

    end = this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('endheadercell'));

    //handle width resizing for columns/rows
    if (this.m_cursor === 'col-resize')
    {
        if (resizeHeaderMode === 'column')
        {
            oldElementWidth = this.calculateColumnHeaderWidth(this.m_resizingElement);
            newElementWidth = this.getNewElementWidth('column', oldElementWidth, end);
            this.resizeColWidth(oldElementWidth, newElementWidth);
        }
        else if (resizeHeaderMode === 'row')
        {
            oldElementWidth = this.getElementWidth(this.m_resizingElement);
            newElementWidth = this.getNewElementWidth('row', oldElementWidth, end);
            this.resizeRowWidth(newElementWidth, newElementWidth - oldElementWidth, end);
        }
    }
    //handle height resizing for columns/rows
    else if (this.m_cursor === 'row-resize')
    {
        if (resizeHeaderMode === 'row')
        {
            oldElementHeight = this.calculateRowHeaderHeight(this.m_resizingElement);
            newElementHeight = this.getNewElementHeight('row', oldElementHeight, end);
            this.resizeRowHeight(oldElementHeight, newElementHeight);
        }
        else if (resizeHeaderMode === 'column')
        {
            oldElementHeight = this.getElementHeight(this.m_resizingElement);
            newElementHeight = this.getNewElementHeight('column', oldElementHeight, end);
            this.resizeColHeight(newElementHeight, newElementHeight - oldElementHeight, end);
        }
    }

    //rebuild the corners
    this.buildCorners();

    // re-align touch affordances
    if (this.m_utils.isTouchDevice())
    {
        this._moveTouchSelectionAffordance();
    }

    //update the last mouse X/Y
    this.m_lastMouseX = this.m_currentMouseX;
    this.m_lastMouseY = this.m_currentMouseY;
};

/**
 * Resize the width of column headers, and the column cells. Also resize the
 * scroller and databody accordingly. Set the left(or right) style value on all
 * cells/columns following(preceeding) the resizing element. Update the end
 * column pixel as well.
 * @param {number} oldElementWidth - the elements width prior to resizing
 * @param {number} newElementWidth - the elements width after resizing
 */
DvtDataGrid.prototype.resizeColWidth = function(oldElementWidth, newElementWidth)
{
    var widthChange, oldScrollerWidth, newScrollerWidth;
    widthChange = newElementWidth - oldElementWidth;
    if (widthChange != 0)
    {
        if (this.m_databody['firstChild'] != null)
        {
            oldScrollerWidth = this.getElementWidth(this.m_databody['firstChild']);
            newScrollerWidth = oldScrollerWidth + widthChange;
            this.setElementWidth(this.m_databody['firstChild'], newScrollerWidth);
        }

        //helper to update all elements this effects
        this.resizeColumnWidthAndShift(widthChange);

        this.m_endColPixel += widthChange;
        this.m_endColHeaderPixel += widthChange;
        this.m_endColEndHeaderPixel += widthChange;
        this.m_avgColWidth = newScrollerWidth / this.getDataSource().getCount('column');

        this.manageResizeScrollbars();
    }
};

/**
 * Resize the height of row headers, and the rows cells. Also resize the
 * scroller and databody accordingly. Update the end row pixel as well.
 * @param {number} oldElementHeight - the elements height prior to resizing
 * @param {number} newElementHeight - the elements height after resizing
 */
DvtDataGrid.prototype.resizeRowHeight = function(oldElementHeight, newElementHeight)
{
    var heightChange, oldScrollerHeight, newScrollerHeight;
    heightChange = newElementHeight - oldElementHeight;
    if (heightChange != 0)
    {
        if (this.m_databody['firstChild'] != null)
        {
            oldScrollerHeight = this.getElementHeight(this.m_databody['firstChild']);
            newScrollerHeight = oldScrollerHeight + heightChange;
            this.setElementHeight(this.m_databody['firstChild'], newScrollerHeight);
        }

        //set row height on the appropriate databody row, set the new value in the sizingManager
        this.resizeRowHeightAndShift(heightChange);

        this.m_endRowPixel += heightChange;
        this.m_endRowHeaderPixel += heightChange;
        this.m_endRowEndHeaderPixel += heightChange;
        this.m_avgRowHeight = newScrollerHeight / this.getDataSource().getCount('row');

        this.manageResizeScrollbars();
    }
};

/**
 * Resize the height of column headers. Also resize the scroller and databody
 * accordingly.
 * @param {number} newElementHeight - the column header height after resizing
 * @param {number} heightChange - the change in height
 * @param {boolean} end
 */
DvtDataGrid.prototype.resizeColHeight = function(newElementHeight, heightChange, end)
{
    if (heightChange != 0)
    {
        var level = this.getHeaderCellLevel(this.m_resizingElement) + this.getHeaderCellDepth(this.m_resizingElement) - 1;
        end ? this.m_columnEndHeaderLevelHeights[level] += heightChange : this.m_columnHeaderLevelHeights[level] += heightChange;
        this.resizeColumnHeightsAndShift(heightChange, level, end);

        if (!end)
        {
            this.m_colHeaderHeight += heightChange;
            this.setElementHeight(this.m_colHeader, this.m_colHeaderHeight);
        }
        else
        {
            this.m_colEndHeaderHeight += heightChange;
            this.setElementHeight(this.m_colEndHeader, this.m_colEndHeaderHeight);
        }
        this.manageResizeScrollbars();
    }
};

/**
 * Resize the width of row headers. Also resize the scroller and databody
 * accordingly.
 * @param {number} newElementWidth - the row header width after resizing
 * @param {number} widthChange - the change in width
 * @param {boolean} end
 */
DvtDataGrid.prototype.resizeRowWidth = function(newElementWidth, widthChange, end)
{
    var level;
    if (widthChange != 0)
    {
        level = this.getHeaderCellLevel(this.m_resizingElement) + this.getHeaderCellDepth(this.m_resizingElement) - 1;

        end ? this.m_rowEndHeaderLevelWidths[level] += widthChange : this.m_rowHeaderLevelWidths[level] += widthChange;
        this.resizeRowWidthsAndShift(widthChange, level, end);

        if (!end)
        {
            this.m_rowHeaderWidth += widthChange;
            this.setElementWidth(this.m_rowHeader, this.m_rowHeaderWidth);
        }
        else
        {
            this.m_rowEndHeaderWidth += widthChange;
            this.setElementWidth(this.m_rowEndHeader, this.m_rowEndHeaderWidth);
        }
        this.manageResizeScrollbars();
    }
};

/**
 * Determine what the new element width should be based on minimum values.
 * Accounts for the overshoot potential of passing up the boundries set.
 * @param {string} axis - the axis along which we need a new width
 * @param {number} oldElementWidth - the element width prior to resizing
 * @param {boolean} end
 * @param {number=} deltaWidth
 * @return {number} the element width after resizing
 */
DvtDataGrid.prototype.getNewElementWidth = function(axis, oldElementWidth, end, deltaWidth)
{
    var minWidth, databodyWidth, newElementWidth, halfGridWidth;
    //to account for the 24px resing width
    minWidth = this._getMinValue('width', axis);
    databodyWidth = this.getElementWidth(this.m_databody);
    if (deltaWidth == null)
    {
        deltaWidth = this.getResources().isRTLMode() ? this.m_lastMouseX - this.m_currentMouseX : this.m_currentMouseX - this.m_lastMouseX;
    }

    if (end && axis == 'row')
    {
        deltaWidth = deltaWidth * -1;
    }
    newElementWidth = oldElementWidth + deltaWidth + this.m_overResizeLeft + this.m_overResizeMinLeft + this.m_overResizeRight;
    halfGridWidth = Math.round(this.getWidth() / 2);

    //check to make sure the element exceeds the minimum width
    if (newElementWidth < minWidth)
    {
        this.m_overResizeMinLeft += deltaWidth - minWidth + oldElementWidth;
        newElementWidth = minWidth;
    }
    else
    {
        this.m_overResizeMinLeft = 0;
        this.m_overResizeLeft = 0;
    }
    //check to make sure row header width don't exceed half of the grid width
    if (axis === 'row')
    {
        if (newElementWidth > halfGridWidth)
        {
            this.m_overResizeRight += deltaWidth - halfGridWidth + oldElementWidth;
            newElementWidth = halfGridWidth;
        }
        else
        {
            this.m_overResizeRight = 0;
        }
    }
    return newElementWidth;
};

/**
 * Determine what the new element height should be based on minimum values.
 * Accounts for the overshoot potential of passing up the boundries set.
 * @param {string} axis - the axis along which we need a new width
 * @param {number} oldElementHeight - the element height prior to resizing
 * @param {boolean} end
 * @param {number=} deltaHeight
 * @return {number} the element height after resizing
 */
DvtDataGrid.prototype.getNewElementHeight = function(axis, oldElementHeight, end, deltaHeight)
{
    var databodyHeight, minHeight, newElementHeight, halfGridHeight;
    minHeight = this._getMinValue('height', axis);
    databodyHeight = this.getElementHeight(this.m_databody);
    if (deltaHeight == null)
    {
        deltaHeight = this.m_currentMouseY - this.m_lastMouseY;
    }
    if (end && axis == 'column')
    {
        deltaHeight = deltaHeight * -1;
    }
    newElementHeight = oldElementHeight + deltaHeight + this.m_overResizeTop + this.m_overResizeMinTop + this.m_overResizeBottom;
    halfGridHeight = Math.round(this.getHeight() / 2);

    //Check to make sure the element height exceeds the minimum height
    if (newElementHeight < minHeight)
    {
        this.m_overResizeMinTop += deltaHeight - minHeight + oldElementHeight;
        newElementHeight = minHeight;
    }
    else
    {
        this.m_overResizeMinTop = 0;
        this.m_overResizeTop = 0;
    }
    //check to make sure column header width don't exceed half of the grid height
    if (axis === 'column')
    {
        if (newElementHeight > halfGridHeight)
        {
            this.m_overResizeBottom += deltaHeight - halfGridHeight + oldElementHeight;
            newElementHeight = halfGridHeight;
        }
        else
        {
            this.m_overResizeBottom = 0;
        }
    }
    return newElementHeight;
};

/**
 * Determine what the minimum value for the resizing element is
 * @param {string} dimension - the width or height
 * @param {string} axis - the axis
 * @return {number} the minimum height for the element
 * @private
 */
DvtDataGrid.prototype._getMinValue = function(dimension, axis)
{
    var index, level, elem, minValue, extent, currentDimensionValue, inner, innerDimensionValue, depth, paddingBorder;
    elem = this.m_resizingElement;
    level = this.getHeaderCellLevel(elem);
    depth = this.getHeaderCellDepth(elem);
    paddingBorder = this._getCellPaddingBorder(dimension, elem);
    minValue = Math.max((this.m_utils.isTouchDevice() ? 2 * DvtDataGrid.RESIZE_TOUCH_OFFSET : 2 * DvtDataGrid.RESIZE_OFFSET), paddingBorder);
    if ((axis === 'column' && (this.m_columnHeaderLevelCount === 1 || (dimension === 'width' && this.m_columnHeaderLevelCount === level + 1) || (dimension === 'height' && depth === 1))) ||
            (axis === 'row' && (this.m_rowHeaderLevelCount === 1 || (dimension === 'height' && this.m_rowHeaderLevelCount === level + 1) || (dimension === 'width' && depth === 1))))
    {
        return minValue;
    }
    index = this.getHeaderCellIndex(elem);
    extent = this._getAttribute(this.m_resizingElement['parentNode'], 'extent', true);
    currentDimensionValue = this.getElementDir(elem, dimension);

    if (axis === 'column')
    {
        if (dimension === 'width')
        {
            inner = this._getHeaderByIndex(index + extent - 1, this.m_columnHeaderLevelCount - 1, this.m_colHeader, this.m_columnHeaderLevelCount, this.m_startColHeader);
            innerDimensionValue = this.getElementDir(inner, dimension);
        }
        else
        {
            innerDimensionValue = this._getColumnHeaderLevelHeight(level + depth - 1, elem);
        }
    }
    else if (axis === 'row')
    {
        if (dimension === 'height')
        {
            inner = this._getHeaderByIndex(index + extent - 1, this.m_rowHeaderLevelCount - 1, this.m_rowHeader, this.m_rowHeaderLevelCount, this.m_startRowHeader);
            innerDimensionValue = this.getElementDir(inner, dimension);
        }
        else
        {
            innerDimensionValue = this._getHeaderLevelDimension(level + depth - 1, elem, this.m_rowHeaderLevelWidths, 'width', 1);
        }
    }
    return currentDimensionValue - (innerDimensionValue - minValue);
};

/**
 * Get the cell padding + border size along a certain dimenison
 * @param {string} dimension
 * @param {Element} elem
 * @returns {Number}
 */
DvtDataGrid.prototype._getCellPaddingBorder = function(dimension, elem)
{
    var cssExpand, i, val, style;
    if (this.m_resizingElementMin == null)
    {
        cssExpand = [ "top", "right", "bottom", "left" ];
        i = dimension === "width" ? 1 : 0;
        val = 0;
        style = window.getComputedStyle(elem);
        for ( ; i < 4; i += 2 ) {
            val += parseFloat(style.getPropertyValue("padding-" + cssExpand[ i ]));
            val += parseFloat(style.getPropertyValue("border-" + cssExpand[ i ] + "-width"));
        }
        this.m_resizingElementMin = Math.round(val);
    }
    return this.m_resizingElementMin;
};

/**
 * Manages the databody and scroller sizing when the scrollbars are added and
 * removed scrollbars from the grid. This allows the grid container to maintain
 * size as it renders scrollbars inside rahther than out. Method mimics resizeGrid
 */
DvtDataGrid.prototype.manageResizeScrollbars = function()
{
    var width, height, colHeader, rowHeader, databody, colHeaderHeight, columnHeaderWidth, rowHeaderWidth, rowHeaderHeight,
            databodyContentWidth, databodyWidth, databodyContentHeight, databodyHeight, isDatabodyHorizontalScrollbarRequired,
            isDatabodyVerticalScrollbarRequired, scrollbarSize, dir, rowEndHeaderDir, columnEndHeaderDir, isEmpty,
            deltaX = 0, deltaY = 0, colEndHeader, rowEndHeader, availableHeight, availableWidth, colEndHeaderHeight, rowEndHeaderWidth,
            databodyScroller, empty, emptyWidth, emptyHeight;


    width = this.getWidth();
    height = this.getHeight();
    colHeader = this.m_colHeader;
    colEndHeader = this.m_colEndHeader;
    rowHeader = this.m_rowHeader;
    rowEndHeader = this.m_rowEndHeader;
    databody = this.m_databody;
    databodyScroller = databody['firstChild'];

    // cache these since they will be used in multiple places and we want to minimize reflow
    colHeaderHeight = this.getColumnHeaderHeight();
    colEndHeaderHeight = this.getColumnEndHeaderHeight();
    rowHeaderWidth = this.getRowHeaderWidth();
    rowEndHeaderWidth = this.getRowEndHeaderWidth();

    //adjusted to make the databody wrap the databody content, and the scroller to fill the remaing part of the grid
    //this way our scrollbars are always at the edges of our viewport
    availableHeight = height - colHeaderHeight - colEndHeaderHeight;
    availableWidth = width - rowHeaderWidth - rowEndHeaderWidth;

    scrollbarSize = this.m_utils.getScrollbarSize();
    dir = this.getResources().isRTLMode() ? "right" : "left";
    isEmpty = this._databodyEmpty();

    // check if there's no data
    if (isEmpty)
    {
        //could be getting here in the handle resize of an empty grid
        if (this.m_empty == null)
        {
            empty = this._buildEmptyText();
            this.m_root.appendChild(empty); //@HTMLUpdateOK
        }
        else
        {
            empty = this.m_empty;
        }
        emptyHeight = this.getElementHeight(empty);
        emptyWidth = this.getElementWidth(empty);

        if (emptyHeight > this.getElementHeight(databodyScroller))
        {
            this.setElementHeight(databodyScroller, emptyHeight);
        }
        if (emptyWidth > this.getElementWidth(databodyScroller))
        {
            this.setElementWidth(databodyScroller, emptyWidth);
        }
    }

    databodyContentWidth = this.getElementWidth(databody['firstChild']);
    databodyContentHeight = this.getElementHeight(databody['firstChild']);
    //determine which scrollbars are required, if needing one forces need of the other, allows rendering within the root div
    isDatabodyHorizontalScrollbarRequired = this.isDatabodyHorizontalScrollbarRequired(availableWidth);
    if (isDatabodyHorizontalScrollbarRequired)
    {
        isDatabodyVerticalScrollbarRequired = this.isDatabodyVerticalScrollbarRequired(availableHeight - scrollbarSize);
        databody['style']['overflow'] = "auto";
    }
    else
    {
        isDatabodyVerticalScrollbarRequired = this.isDatabodyVerticalScrollbarRequired(availableHeight);
        if (isDatabodyVerticalScrollbarRequired)
        {
            isDatabodyHorizontalScrollbarRequired = this.isDatabodyHorizontalScrollbarRequired(availableWidth - scrollbarSize);
            databody['style']['overflow'] = "auto";
        }
        else
        {
            // for an issue where same size child causes scrollbars (similar code used in resizing already)
            databody['style']['overflow'] = "hidden";
        }
    }

    this.m_hasHorizontalScroller = isDatabodyHorizontalScrollbarRequired;
    this.m_hasVerticalScroller = isDatabodyVerticalScrollbarRequired;

    if (this.m_endColEndHeader != -1)
    {
        databodyHeight = Math.min(databodyContentHeight + (isDatabodyHorizontalScrollbarRequired ? scrollbarSize : 0), availableHeight);
        rowHeaderHeight = isDatabodyHorizontalScrollbarRequired ? databodyHeight - scrollbarSize : databodyHeight;
    }
    else
    {
        databodyHeight = availableHeight;
        rowHeaderHeight = Math.min(databodyContentHeight, isDatabodyHorizontalScrollbarRequired ? databodyHeight - scrollbarSize : databodyHeight);
    }

    if (this.m_endRowEndHeader != -1)
    {
        databodyWidth = Math.min(databodyContentWidth + (isDatabodyVerticalScrollbarRequired ? scrollbarSize : 0), availableWidth);
        columnHeaderWidth = isDatabodyVerticalScrollbarRequired ? databodyWidth - scrollbarSize : databodyWidth;
    }
    else
    {
        databodyWidth = availableWidth;
        columnHeaderWidth = Math.min(databodyContentWidth, isDatabodyVerticalScrollbarRequired ? databodyWidth - scrollbarSize : databodyWidth);
    }

    rowEndHeaderDir = rowHeaderWidth + columnHeaderWidth + (isDatabodyVerticalScrollbarRequired ? scrollbarSize : 0);
    columnEndHeaderDir = colHeaderHeight + rowHeaderHeight + (isDatabodyHorizontalScrollbarRequired ? scrollbarSize : 0);

    this.setElementDir(rowHeader, 0, dir);
    this.setElementDir(rowHeader, colHeaderHeight, 'top');
    this.setElementHeight(rowHeader, rowHeaderHeight);

    this.setElementDir(rowEndHeader, rowEndHeaderDir, dir);
    this.setElementDir(rowEndHeader, colHeaderHeight, 'top');
    this.setElementHeight(rowEndHeader, rowHeaderHeight);

    this.setElementDir(colHeader, rowHeaderWidth, dir);
    this.setElementWidth(colHeader, columnHeaderWidth);

    this.setElementDir(colEndHeader, rowHeaderWidth, dir);
    this.setElementDir(colEndHeader, columnEndHeaderDir, 'top');
    this.setElementWidth(colEndHeader, columnHeaderWidth);

    this.setElementDir(databody, colHeaderHeight, 'top');
    this.setElementDir(databody, rowHeaderWidth, dir);
    this.setElementWidth(databody, databodyWidth);
    this.setElementHeight(databody, databodyHeight);

    // cache the scroll width and height to minimize reflow
    this.m_scrollWidth = databodyContentWidth - columnHeaderWidth;
    this.m_scrollHeight = databodyContentHeight - rowHeaderHeight;

    this.buildCorners();

    // check if we need to remove border on the last column header/add borders to headers
    this._adjustHeaderBorders();

    // on touch devices the scroller doesn't automatically scroll into view when resizing the last columns or rows to be smaller
    if (this.m_utils.isTouchDevice())
    {
        // if the visible window plus the scrollLeft is bigger than the scrollable region maximum, rescroll the window
        if (this.m_currentScrollLeft > this.m_scrollWidth)
        {
            deltaX = this.m_scrollWidth - this.m_currentScrollLeft;
        }

        if (this.m_currentScrollTop > this.m_scrollHeight)
        {
            deltaY = this.m_scrollHeight - this.m_currentScrollTop;
        }

        if (deltaX != 0 || deltaY != 0)
        {
            // eliminate bounce back for touch scroll
            this._disableTouchScrollAnimation();
            var delta = this.adjustTouchScroll(deltaX, deltaY);
            deltaX = delta[0];
            deltaY = delta[1];


            this.scrollDelta(deltaX, deltaY);
        }
    }
};

/**
 * Resizes all cell in the resizing element's column, and updates the left(right)
 * postion on the cells and column headers that follow(preceed) that column.
 * @param {number} widthChange - the change in width of the resizing element
 */
DvtDataGrid.prototype.resizeColumnWidthAndShift = function(widthChange)
{
    var dir, databodyRows, i, newStart, j, index, cells, cell, colHeaderDisplay, colEndHeaderDisplay;
    dir = this.getResources().isRTLMode() ? "right" : "left";

    colHeaderDisplay = this.m_colHeader['style']['display'];
    colEndHeaderDisplay = this.m_colEndHeader['style']['display'];

    // hide the databody and col header for performance
    this.m_databody['style']['display'] = 'none';
    this.m_colHeader['style']['display'] = 'none';
    this.m_colEndHeader['style']['display'] = 'none';

    //get the index of the header, if it is a nested header make it the last child index
    index = this.getHeaderCellIndex(this.m_resizingElement);
    if (this.m_columnHeaderLevelCount > 1
            && this.m_resizingElement === this.m_resizingElement['parentNode']['firstChild']
            && this.m_resizingElement['nextSibling'] != null) // has children
    {
        index += this._getAttribute(this.m_resizingElement['parentNode'], 'extent', true) - 1;
    }

    // move column headers within the container and adjust the widths appropriately
    this._shiftHeadersAlongAxisInContainer(this.m_colHeader['firstChild'], index, widthChange, dir, this.getMappedStyle('colheadercell'), 'column');

    // move column headers within the container and adjust the widths appropriately
    this._shiftHeadersAlongAxisInContainer(this.m_colEndHeader['firstChild'], index, widthChange, dir, this.getMappedStyle('colendheadercell'), 'column');

    // shift the cells widths and left/right values in the databody
    this._shiftCellsAlongAxis('column', widthChange, index);

    //restore visibility
    this.m_databody['style']['display'] = '';
    this.m_colHeader['style']['display'] = colHeaderDisplay;
    this.m_colEndHeader['style']['display'] = colEndHeaderDisplay;
};

/**
 * Moves cells inside of all rows/columns starting at a certain index, will also resize the given index.
 * @param {string} axis
 * @param {number} dimensionDelta
 * @param {number} startAxisChange
 * @param {boolean|null=} shiftOnly true if only moving cells along an axis by dimensionDelta without resizing
 */
DvtDataGrid.prototype._shiftCellsAlongAxis = function(axis, dimensionDelta, startAxisChange, shiftOnly)
{
    var tempArray, dimension, endAxisChange, startOuterLoop, endOuterLoop, index, cell, cellContext, cellAxisStartIndex, axisExtent, startAdjustment, newStart, dir, i, j, newDimension;

    tempArray = [];

    if (shiftOnly == null)
    {
        shiftOnly = false;
    }

    if (axis == 'row')
    {
        dimension = 'height';
        dir = 'top';
        endAxisChange = this.m_endRow;
        startOuterLoop = this.m_startCol;
        endOuterLoop = this.m_endCol;
    }
    else
    {
        dimension = 'width';
        dir = this.getResources().isRTLMode() ? "right" : "left";
        endAxisChange = this.m_endCol;
        startOuterLoop = this.m_startRow;
        endOuterLoop = this.m_endRow;
    }

    // shift the cells widths and left/right values in the databody
    if (this.m_databody['firstChild'] != null)
    {
        for (i = startOuterLoop; i <= endOuterLoop; i++)
        {
            // set the new width on the appropriate column
            index = axis === 'row' ? this.createIndex(startAxisChange, i) : this.createIndex(i, startAxisChange);
            cell = this._getCellByIndex(index);
            cellContext = cell[this.getResources().getMappedAttribute('context')];
            cellAxisStartIndex = cellContext['indexes'][axis];
            axisExtent = cellContext['extents'][axis];
            var extentWithinCellToIgnore = startAxisChange - cellAxisStartIndex;
            startAdjustment = (axisExtent - extentWithinCellToIgnore);

            if (!(tempArray[i] && tempArray[i][startAxisChange]) && shiftOnly !== true)
            {
                this.setElementDir(cell, this.getElementDir(cell, dimension) + dimensionDelta, dimension);
                if (axis == 'row')
                {
                    this._updateTempArray(tempArray, true, i, startAxisChange, cellContext['extents']['column'], axisExtent - extentWithinCellToIgnore);
                }
                else
                {
                    this._updateTempArray(tempArray, true, i, startAxisChange, cellContext['extents']['row'], axisExtent - extentWithinCellToIgnore);
                }
            }

            // start value ignore adjustment only if shiftOnly flag is true
            var axisStartValue = startAxisChange;
            if (shiftOnly !== true)
            {
                axisStartValue = startAxisChange + startAdjustment;
            }

            // move the columns within the data body to account for width change
            for (j = axisStartValue; j <= endAxisChange; j++)
            {
                if (tempArray[i] && tempArray[i][j])
                {
                    continue;
                }

                index = axis === 'row' ? this.createIndex(j, i) : this.createIndex(i, j);
                cell = this._getCellByIndex(index);
                cellContext = cell[this.getResources().getMappedAttribute('context')];
                axisExtent = cellContext['extents'][axis];

                newStart = this.getElementDir(cell, dir) + dimensionDelta;
                this.setElementDir(cell, newStart, dir);
                if (axis == 'row')
                {
                    this._updateTempArray(tempArray, true, i, j, cellContext['extents']['column'], axisExtent);
                }
                else
                {
                    this._updateTempArray(tempArray, true, i, j, cellContext['extents']['row'], axisExtent);
                }
            }
        }
    }
};

/**
 * Resizes the resizing elements row, and updates the top
 * postion on the rows and row headers that follow that column.
 * @param {number} heightChange - the change in width of the resizing element
 */
DvtDataGrid.prototype.resizeRowHeightAndShift = function(heightChange)
{
    var databodyRows, i, j, newStart, index, cell, newHeight, rowHeaderDisplay, rowEndHeaderDisplay;

    rowHeaderDisplay = this.m_rowHeader['style']['display'];
    rowEndHeaderDisplay = this.m_rowEndHeader['style']['display'];;

    // hide the databody and row header for performance
    this.m_databody['style']['display'] = 'none';
    this.m_rowHeader['style']['display'] = 'none';
    this.m_rowEndHeader['style']['display'] = 'none';

    //get the index of the header, if it is a nested header make it the last child index
    index = this.getHeaderCellIndex(this.m_resizingElement);
    if (this.m_rowHeaderLevelCount > 1
            && this.m_resizingElement === this.m_resizingElement['parentNode']['firstChild']
            && this.m_resizingElement['nextSibling'] != null) // has children)
    {
        index += this._getAttribute(this.m_resizingElement['parentNode'], 'extent', true) - 1;
    }

    // move row headers within the container
    this._shiftHeadersAlongAxisInContainer(this.m_rowHeader['firstChild'], index, heightChange, 'top', this.getMappedStyle('rowheadercell'), 'row');

    // move row headers within the container
    this._shiftHeadersAlongAxisInContainer(this.m_rowEndHeader['firstChild'], index, heightChange, 'top', this.getMappedStyle('rowendheadercell'), 'row');

    // shift the cells hieghts and top values in the databody
    this._shiftCellsAlongAxis('row', heightChange, index);

    this.m_databody['style']['display'] = '';
    this.m_rowHeader['style']['display'] = rowHeaderDisplay;
    this.m_rowEndHeader['style']['display'] = rowEndHeaderDisplay;
};

/**
 * This method recursively shifts a group over until it reaches the index where the group width/height needs to be adjusted
 * @param {Element} headersContainer the header grouping or scroller
 * @param {number|string} index the index that is being adjusted
 * @param {number} dimensionChange the change in width or height
 * @param {string} dir top, left, or right the appropriate value to adjust along the axis
 * @param {string} className the header cell className along that axis
 * @param {string=} axis the axis we are shifting on
 * @private
 */
DvtDataGrid.prototype._shiftHeadersAlongAxisInContainer = function(headersContainer, index, dimensionChange, dir, className, axis)
{
    var element, header, isHeader, groupingContainer, headerStart, headers, i, newStart, newVal = 0;

    // get the last element in the container
    element = headersContainer['lastChild'];
    if (element == null)
    {
        return;
    }

    // is the last element a header or a group
    isHeader = this.m_utils.containsCSSClassName(element, className);
    // what is the index of the container/header
    if (isHeader)
    {
        groupingContainer = element['parentNode'];
        header = element;
        headerStart = this.getHeaderCellIndex(header);
    }
    else
    {
        groupingContainer = element;
        header = element['firstChild'];
        headerStart = this._getAttribute(groupingContainer, 'start', true);
    }

    // if the group is after the specified index move all the dir values under that group
    while (index < headerStart)
    {
        if (isHeader)
        {
            //move this header to the right left up down
            newStart = this.getElementDir(header, dir) + dimensionChange;
            this.setElementDir(header, newStart, dir);

            element = element['previousSibling'];
            isHeader = this.m_utils.containsCSSClassName(element, className);
            groupingContainer = element['parentNode'];
            header = element;
            headerStart = this.getHeaderCellIndex(header);
        }
        else
        {
            //move all children of a group
            headers = groupingContainer.getElementsByClassName(className);
            for (i = 0; i < headers.length; i++)
            {
                newStart = this.getElementDir(headers[i], dir) + dimensionChange;
                this.setElementDir(headers[i], newStart, dir);
            }

            element = element['previousSibling'];
            isHeader = this.m_utils.containsCSSClassName(element, className);
            groupingContainer = element;
            header = element['firstChild'];
            headerStart = this._getAttribute(groupingContainer, 'start', true);
        }
    }

    if (axis == 'column')
    {
        //the last header we moved to should be the one that needs its width updated
        newVal = this.getElementWidth(header) + dimensionChange;
        this.setElementWidth(header, newVal);
    }
    else if (axis == 'row')
    {
        newVal = this.getElementHeight(header) + dimensionChange;
        this.setElementHeight(header, newVal);
    }
    else if (axis == null)
    {
        newStart = this.getElementDir(header, dir) + dimensionChange;
        this.setElementDir(header, newStart, dir);
    }

    // if we aren't innermost then repeat for its children
    if (!isHeader && header['nextSibling'] != null) // has children
    {
        this._shiftHeadersAlongAxisInContainer(element, index, dimensionChange, dir, className, axis);
    }
    else if (axis != null)
    {
        //store the width/height change in the sizing manager, only care about innermost
        this.m_sizingManager.setSize(axis, this._getKey(header), newVal);
    }
};

/**
 * Resizes all cell in the resizing element's column, and updates the left(right)
 * postion on the cells and column headers that follow(preceed) that column.
 * @param {number} heightChange - the change in width of the resizing element
 * @param {number} level - the level we are resizing
 * @param {boolean} end
 */
DvtDataGrid.prototype.resizeColumnHeightsAndShift = function(heightChange, level, end)
{
    var root, className, axis, dir;
    if (!end)
    {
       root = this.m_colHeader;
       className = this.getMappedStyle('colheadercell');
       axis  = 'column';
       dir = 'top';
    }
    else
    {
       root = this.m_colEndHeader;
       className = this.getMappedStyle('colendheadercell');
       axis  = 'columnEnd';
       dir = 'bottom';
    }

    root['style']['display'] = 'none';
    this.m_databody['style']['display'] = 'none';
    // move column headers within the container
    this._shiftHeadersDirInContainer(root['firstChild'], heightChange, level, dir, className, axis);
    root['style']['display'] = '';
    this.m_databody['style']['display'] = '';
};

/**
 * Resizes all cell in the resizing element's column, and updates the left(right)
 * postion on the cells and column headers that follow(preceed) that column.
 * @param {number} widthChange - the change in width of the resizing element
 * @param {number} level - the level we are resizing
 * @param {boolean} end
 */
DvtDataGrid.prototype.resizeRowWidthsAndShift = function(widthChange, level, end)
{
    var root, className, axis, dir;

    if (!end)
    {
       root = this.m_rowHeader;
       className = this.getMappedStyle('rowheadercell');
       axis  = 'row';
       dir = this.getResources().isRTLMode() ? "right" : "left";
    }
    else
    {
       root = this.m_rowEndHeader;
       className = this.getMappedStyle('rowendheadercell');
       axis  = 'rowEnd';
       dir = this.getResources().isRTLMode() ? "left" : "right";
    }

    root['style']['display'] = 'none';
    this.m_databody['style']['display'] = 'none';
    // move column headers within the container
    this._shiftHeadersDirInContainer(root['firstChild'], widthChange, level, dir, className, axis);
    root['style']['display'] = '';
    this.m_databody['style']['display'] = '';

};

/**
 * Shifts the headers after a particular level over and adjusts the dimension of that level across the whole container
 * @param {Element} headersContainer
 * @param {number} dimensionChange
 * @param {number} level
 * @param {string} dir
 * @param {string} className
 * @param {string} axis
 * @private
 */
DvtDataGrid.prototype._shiftHeadersDirInContainer = function(headersContainer, dimensionChange, level, dir, className, axis)
{
    var groupings, i, grouping, isHeader, headerLevel, headers, newDir, j, headerDepth;
    groupings = headersContainer['childNodes'];
    // for all children in the group
    for (i = 0; i < groupings.length; i++)
    {
        grouping = groupings[i];
        isHeader = this.m_utils.containsCSSClassName(grouping, className);
        //if it is a group
        if (!isHeader)
        {
            headerLevel = this._getAttribute(grouping, 'level', true);
            // if before or on the level we need to go deeper into the grouping
            if (headerLevel <= level)
            {
                this._shiftHeadersDirInContainer(grouping, dimensionChange, level, dir, className, axis);
            }
            else
            {
                // if level is higher then we need to adjust the dir of all the headers under that group
                headers = grouping.getElementsByClassName(className);
                for (j = 0; j < headers.length; j++)
                {
                    newDir = this.getElementDir(headers[j], dir) + dimensionChange;
                    this.setElementDir(headers[j], newDir, dir);
                }
            }
        }
        else
        {
            headerLevel = this.getHeaderCellLevel(grouping);
            headerDepth = this.getHeaderCellDepth(grouping);

            // if we have a header at that level adjust it's value
            if (headerLevel <= level && level < headerLevel + headerDepth)
            {
                if (axis === 'column' || axis === 'columnEnd')
                {
                    newDir = this.getElementHeight(grouping) + dimensionChange;
                    this.setElementHeight(grouping, newDir);
                }
                else
                {
                    newDir = this.getElementWidth(grouping) + dimensionChange;
                    this.setElementWidth(grouping, newDir);
                }
            }
            // if we have a header inside the group then adjust its dimension
            else if (headerLevel > level)
            {
                newDir = this.getElementDir(grouping, dir) + dimensionChange;
                this.setElementDir(grouping, newDir, dir);
            }
        }
    }
};

/**
 * Takes the original target of the context menu and maps it to the appropriate
 * column/row header to resize and selects the right resize function.
 * @param {Event} event - the event that spawned context menu
 * @param {string} id - 'width' or 'height'
 * @param {string} val - new width or height to resize to
 * @param {Element|undefined} target - the target element
 */
DvtDataGrid.prototype.handleContextMenuResize = function(event, id, val, target)
{
    var initialWidth, initialHeight, newWidth, newHeight, value, details, end, deltaWidth, deltaHeight;

    this.m_overResizeLeft = 0;
    this.m_overResizeMinLeft = 0;
    this.m_overResizeTop = 0;
    this.m_overResizeMinTop = 0;
    this.m_overResizeRight = 0;
    this.m_overResizeBottom = 0;

    value = parseInt(val, 10);

    deltaWidth = value - this.getElementWidth(target);
    deltaHeight = value - this.getElementHeight(target);

    if (this.m_utils.containsCSSClassName(target, this.getMappedStyle('cell')))
    {
        if (id === this.m_resources.getMappedCommand('resizeHeight'))
        {
            target = this.getHeaderFromCell(target, 'row', true);
        }
        else
        {
            target = this.getHeaderFromCell(target, 'column', true);
        }
    }

    this.m_resizingElement = target;
    initialWidth = this.getElementWidth(target);
    initialHeight = this.getElementHeight(target);
    end = this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('endheadercell'));

    if (id === this.m_resources.getMappedCommand('resizeWidth'))
    {
        if (initialWidth !== value)
        {
            if (this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('colheadercell')) ||
                this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('colendheadercell')))
            {
                if (this._isDOMElementResizable(this.m_resizingElement))
                {
                    this.resizeColWidth(initialWidth, this.getNewElementWidth('column', initialWidth, end, deltaWidth));
                }
            }
            else
            {
                this.resizeRowWidth(value, value - initialWidth, end);
            }
        }
    }
    else if (id === this.m_resources.getMappedCommand('resizeHeight'))
    {
        if (initialHeight !== value)
        {
            if (this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('colheadercell')) ||
                this.m_utils.containsCSSClassName(this.m_resizingElement, this.getMappedStyle('colendheadercell')))
            {
                this.resizeColHeight(value, value - initialHeight, end);
            } else
            {
                if (this._isDOMElementResizable(this.m_resizingElement))
                {
                    this.resizeRowHeight(initialHeight, this.getNewElementHeight('row', initialHeight, end, deltaHeight));
                }
            }
        }
    }

    newWidth = this.getElementWidth(target);
    newHeight = this.getElementHeight(target);
    if (newWidth != initialWidth || newHeight != initialHeight)
    {
        //set the information we want to callback with in the resize event and callback
        details = {
            'event': event,
            'ui': {
                'header': this._getKey(this.m_resizingElement),
                'oldDimensions' : {
                    'width': initialWidth,
                    'height': initialHeight
                },
                'newDimensions' : {
                    'width': newWidth,
                    'height': newHeight
                },
                // deprecating this part in 2.1.0
                'size': value
            }
        };
        this.fireEvent('resize', details);

        this.buildCorners();
        // re-align touch affordances
        if (this.m_utils.isTouchDevice())
        {
            this._moveTouchSelectionAffordance();
        }
    }

    this.m_resizingElement = null;
    this.m_resizingElementMin = null;
};

/**
 * Get the edges (left,right,top,bottom) pixel locations relative to the page
 * @param {Element} elem - the element to find edges of
 * @return {Array.<number>} An array of numbers [leftEdge, topEdge, rightEdge, bottomEdge]
 */
DvtDataGrid.prototype.getHeaderEdgePixels = function(elem)
{
    var elementXY, leftEdge, topEdge, rightEdge, bottomEdge, targetWidth, targetHeight;
    elementXY = this.findPos(elem);
    leftEdge = elementXY[0];
    topEdge = elementXY[1];
    if (this.m_utils.containsCSSClassName(elem, this.getMappedStyle('colheadercell')))
    {
        targetWidth = this.calculateColumnHeaderWidth(elem);
        targetHeight = this.getElementHeight(elem);
    }
    else
    {
        targetWidth = this.getElementWidth(elem);
        targetHeight = this.calculateRowHeaderHeight(elem);
    }
    rightEdge = leftEdge + targetWidth;
    bottomEdge = topEdge + targetHeight;
    return [leftEdge, topEdge, rightEdge, bottomEdge];
};
/**
 * Class used to keep track of whcih elements have been resized, has an object
 * containing two objects 'row' and 'column', which should have objects of
 * index:{size}. this.sizes = {axis:{index:{size}}}
 * @constructor
 * @private
 */
var DvtDataGridSizingManager = function()
{
    this.sizes = {'column': new Map(), 'row': new Map()};
};

/**
 * Set a size in the sizes object in the sizing manager
 * @param {string} axis - column/row
 * @param {*} headerKey - key of the element
 * @param {number} size - the size to put in the object
 */
DvtDataGridSizingManager.prototype.setSize = function(axis, headerKey, size)
{
    this.sizes[axis].set(headerKey, size);
};

/**
 * Get a size from the sizing manager for a given axis and index,
 * @param {string} axis - column/row
 * @param {*} headerKey - key of the element
 * @return {number|null} a size if it exists
 */
DvtDataGridSizingManager.prototype.getSize = function(axis, headerKey)
{
    // get does reference comparison
    var size = this.sizes[axis].get(headerKey);
    if (size != null)
    {
        return size;
    }
    
    // if reference comparison does not work we should check using compare values
    this.sizes[axis].forEach(function(value, key, map){
        if (size == null && oj.Object.compareValues(key, headerKey))
        {
            size = value;
        }
    });

    return size;
};

/**
 * Empty the sizing manager sizes
 */
DvtDataGridSizingManager.prototype.clear = function()
{
    this.sizes['column']["clear"]();
    this.sizes['row']["clear"]();
};
/**
 * @preserve Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * This class contains all utility methods used by the Grid.
 * @param {Object} dataGrid the dataGrid using the utils
 * @constructor
 * @private
 */
var DvtDataGridUtils = function(dataGrid)
{
    this.scrollbarSize = -1;
    this.dataGrid = dataGrid;

    var userAgent = (navigator && navigator.userAgent) ? navigator.userAgent.toLowerCase() : null;

    this.os = this._determineOS(userAgent);
    this.platform = this._determinePlatform(userAgent);
};

// Platform Constants
DvtDataGridUtils.IE_PLATFORM = "ie";
DvtDataGridUtils.GECKO_PLATFORM = "gecko";
DvtDataGridUtils.WEBKIT_PLATFORM = "webkit";
DvtDataGridUtils.UNKNOWN_PLATFORM = "unknown";
DvtDataGridUtils.EDGE_PLATFORM = "edge";

// OS Constants
DvtDataGridUtils.WINDOWS_OS = "Windows";
DvtDataGridUtils.SOLARIS_OS = "Solaris";
DvtDataGridUtils.MAC_OS = "Mac";
DvtDataGridUtils.UNKNOWN_OS = "Unknown";

/**
 * Get the maximum scrollable browser height
 * @returns {Number}
 */
DvtDataGridUtils.prototype._getMaxDivHeightForScrolling = function()
{
    if (this.m_maxDivHeightForScrolling == null)
    {
        this._setMaxValuesForScrolling();
    }
    return this.m_maxDivHeightForScrolling;
};

/**
 * Get the maximum scrollable browser width
 * @returns {Number}
 */
DvtDataGridUtils.prototype._getMaxDivWidthForScrolling = function()
{
    if (this.m_maxDivWidthForScrolling == null)
    {
        this._setMaxValuesForScrolling();
    }
    return this.m_maxDivWidthForScrolling;
};

/**
 * Set the maximum scrollable browser height
 */
DvtDataGridUtils.prototype._setMaxValuesForScrolling = function()
{
    var div;
    // ie lets the value go forever without actual support, so we hard cap it at 1 million pixels
    if (this.platform === DvtDataGridUtils.IE_PLATFORM || this.platform === DvtDataGridUtils.EDGE_PLATFORM)
    {
        this.m_maxDivHeightForScrolling = 1000000;
        this.m_maxDivWidthForScrolling = 1000000;
        return;
    }

    div = document.createElement("div");
    div.style.cssText = "width:1000000000px;height:1000000000px;display:none;";
    document.body.appendChild(div); //@HTMLUpdateOK
    // for some reason chrome stops rendering absolutely positioned content at half the value on osx
    this.m_maxDivHeightForScrolling = parseInt(parseFloat(window.getComputedStyle(div)['height'])/2, 10);
    this.m_maxDivWidthForScrolling =  parseInt(parseFloat(window.getComputedStyle(div)['width'])/2, 10);
    document.body.removeChild(div); //@HTMLUpdateOK
    return;
};

DvtDataGridUtils.prototype.calculateScrollbarSize = function()
{
    // Create the measurement node
    var scrollDiv = document.createElement("div");
    scrollDiv.style.cssText = "width:100px;height:100px;overflow:scroll;position:absolute;top:-9999px;";
    document.body.appendChild(scrollDiv); //@HTMLUpdateOK

    // Get the scrollbar width/height
    this.scrollbarSize = scrollDiv.offsetWidth - scrollDiv.clientWidth;

    // Delete the DIV
    document.body.removeChild(scrollDiv);
};

/**
 * Gets the size of the native scrollbar
 */
DvtDataGridUtils.prototype.getScrollbarSize = function()
{
    if (this.scrollbarSize == -1)
    {
        this.calculateScrollbarSize();
    }
    return this.scrollbarSize;
};

/**
 * Determine if the current agent is touch device
 */
DvtDataGridUtils.prototype.isTouchDevice = function()
{
    if (this.isTouch == undefined)
    {
        var agentName = navigator.userAgent.toLowerCase();
        if (agentName.indexOf("mobile") != -1 || agentName.indexOf("android") != -1)
        {
            this.isTouch = true;
        }
        else
        {
            this.isTouch = false;
        }
    }
    return this.isTouch;
};

/**
 * Copied from AdfDhtmlCommonUtils used in Faces
 * Adds a CSS className to the dom node if it doesn't already exist in the classNames list and
 * returns <code>true</code> if the class name was added.
 * @param {Element|Node|null|undefined} domElement DOM Element to add style class name to
 * @param {string|null} className Name of style class to add
 * @return {boolean|null} <code>true</code> if the style class was added
 */
DvtDataGridUtils.prototype.addCSSClassName = function(domElement, className)
{
    var added, currentClassName, classNameIndex, newClassName;
    added = false;

    if (className != null && domElement != null)
    {
        currentClassName = domElement.className;

        // get the current location of the className to add in the classNames list
        classNameIndex = DvtDataGridUtils._getCSSClassNameIndex(currentClassName, className);

        // the className doesn't exist so add it
        if (classNameIndex == -1)
        {
            newClassName = currentClassName
                    ? className + " " + currentClassName
                    : className;

            domElement.className = newClassName;

            added = true;
        }
    }

    return added;
};

/**
 * Copied from AdfDhtmlCommonUtils used in Faces
 * Removes a CSS className to the dom node if it existd in the classNames list and
 * returns <code>true</code> if the class name was removed.
 * @param {Element|Node|null|undefined} domElement DOM Element to remove style class name from
 * @param {string|null} className Name of style class to remove
 * @return {boolean|null} <code>true</code> if the style class was removed
 */
DvtDataGridUtils.prototype.removeCSSClassName = function(domElement, className)
{
    var removed, currentClassName, classNameIndex, classNameEndIndex, beforestring, afterstring, newClassName;
    removed = false;

    if (className != null && domElement != null)
    {
        currentClassName = domElement.className;

        classNameIndex = DvtDataGridUtils._getCSSClassNameIndex(currentClassName, className);

        // only need to do work if CSS class name is present
        if (classNameIndex != -1)
        {
            classNameEndIndex = classNameIndex + className.length;

            // the new classNames string is the string before our className and leading whitespace plus
            // the string after our className and trailing whitespace
            beforestring = (classNameIndex == 0)
                    ? null
                    : currentClassName.substring(0, classNameIndex);
            afterstring = (classNameEndIndex == currentClassName.length)
                    ? null
                    : currentClassName.substring(classNameEndIndex + 1); // skip space

            if (beforestring == null)
            {
                if (afterstring == null)
                {
                    newClassName = "";
                }
                else
                {
                    newClassName = afterstring;
                }
            }
            else
            {
                if (afterstring == null)
                {
                    newClassName = beforestring;
                }
                else
                {
                    newClassName = beforestring + afterstring;
                }
            }

            domElement.className = newClassName;

            removed = true;
        }
    }

    return removed;
};

/**
 * @param {Element|Node|null|undefined} domElement DOM Element to check for the style <code>className</code>
 * @param {string} className the CSS className to check for
 * @return {boolean} <code>true</code> if the className is in the element's list of classes
 */
DvtDataGridUtils.prototype.containsCSSClassName = function(domElement, className)
{
    if (className != null && domElement != null)
    {
        return DvtDataGridUtils._getCSSClassNameIndex(domElement.className, className) != -1;
    }
    return false;
};

/**
 * Copied from AdfDhtmlCommonUtils used in Faces
 * Returns the index at which <code>className</code> appears within <code>currentClassName</code>
 * with either a space or the beginning or end of <code>currentClassName</code> on either side.
 * This function optimizes the runtime speed by not creating objects in most cases and assuming
 * 1) It is OK to only check for spaces as whitespace characters
 * 2) It is rare for the searched className to be a substring of another className, therefore
 *    if we get a hit on indexOf, we have almost certainly found our className
 * 3) It is even rarer for the searched className to be a substring of more than one className,
 *    therefore, repeating the search from the end of the string should almost always either return
 *    our className or the original search hit from indexOf
 * @param {string} currentClassName Space-separated class name string to search
 * @param {string} className string to search for within currentClassName
 * @return {number} index of className in currentClassName, or -1 if it doesn't exist
 */
DvtDataGridUtils._getCSSClassNameIndex = function(currentClassName, className)
{
    var classNameLength, currentClassNameLength, nameIndex, hasStart, endIndex, hasEnd, lastNameIndex;
    // if no current class
    // SVG element className property is not of type string
    if (!currentClassName || !currentClassName.indexOf)
    {
        return -1;
    }

    // if the strings are equivalent, then the start index is the beginning of the string
    if (className === currentClassName)
    {
        return 0;
    }

    classNameLength = className.length;
    currentClassNameLength = currentClassName.length;

    // check if our substring exists in the string at all
    nameIndex = currentClassName.indexOf(className);

    // if our substring exists then our class exists if either:
    // 1) It is at the beginning of the classNames string and has a following space
    // 2) It is at the end of the classNames string and has a leading space
    // 3) It has a space on either side
    if (nameIndex >= 0)
    {
        hasStart = (nameIndex == 0) || (currentClassName.charAt(nameIndex - 1) == " ");
        endIndex = nameIndex + classNameLength;
        hasEnd = (endIndex == currentClassNameLength) || (currentClassName.charAt(endIndex) == " ");

        //one of the three condition above has been met so our string is in the parent string
        if (hasStart && hasEnd)
        {
            return nameIndex;
        }

        // our substring exists in the parent string but didn't meet the above conditions,  Were
        // going to be lazy and retest, searchig for our substring from the back of the classNames
        // string
        lastNameIndex = currentClassName.lastIndexOf(className);

        // if we got the same index as the search from the beginning then we aren't in here
        if (lastNameIndex != nameIndex)
        {
            // recheck the three match cases
            hasStart = currentClassName.charAt(lastNameIndex - 1);
            endIndex = lastNameIndex + classNameLength;
            hasEnd = (endIndex == currentClassNameLength) || (currentClassName.charAt(endIndex) == " ");

            if (hasStart && hasEnd)
            {
                return lastNameIndex;
            }

            // this should only happen if the searched for className is a substring of more
            // than one className in the classNames list, so it is OK for this to be slow.  We
            // also know at this point that we definitely didn't get a match at either the very
            // beginning or very end of the classNames string, so we can safely append spaces
            // at either end
            return currentClassName.indexOf(" " + className + " ");

        }
    }


    // no match
    return -1;


};

/**
 * Returns either the ctrl key or the command key in Mac OS
 * @param {Event} event
 */
DvtDataGridUtils.prototype.ctrlEquivalent = function(event)
{
    var isMac = (this.os === DvtDataGridUtils.MAC_OS);
    return (isMac ? event.metaKey : event.ctrlKey);
};

/**
 * Gets the scroll left of an element.  This method abstracts the inconsistency of scrollLeft
 * behavior in RTL mode between browsers.
 * @param {Element} element the dom element to retrieve scroll left
 * @private
 */
DvtDataGridUtils.prototype.getElementScrollLeft = function(element)
{
    var width, elemWidth;
    if (this.dataGrid.getResources().isRTLMode())
    {
        // see mozilla Bug 383026 scrollLeft property now returns negative values in rtl environment
        if (this.platform == DvtDataGridUtils.GECKO_PLATFORM || this.platform == DvtDataGridUtils.IE_PLATFORM || this.platform == DvtDataGridUtils.EDGE_PLATFORM)
        {
            return Math.abs(element['scrollLeft']);
        }
        else
        {
            // webkit browser
            width = this.dataGrid.getScrollableWidth();
            elemWidth = element['clientWidth'];
            return Math.max(0, width - elemWidth - element['scrollLeft']);
        }
    }

    return element['scrollLeft'];

};

/**
 * Sets the scroll left of an element.  This method abstracts the inconsistency of scrollLeft
 * behavior in RTL mode between browsers.
 * @param {Element} element the dom element to set scroll left
 * @param {number} scrollLeft the scroll left of the dom element
 * @private
 */
DvtDataGridUtils.prototype.setElementScrollLeft = function(element, scrollLeft)
{
    var width, elemWidth, newPos;
    if (this.dataGrid.getResources().isRTLMode())
    {
        if (this.platform === DvtDataGridUtils.GECKO_PLATFORM)
        {
            // see mozilla Bug 383026 scrollLeft property now returns negative values in rtl environment
            element['scrollLeft'] = -scrollLeft;
        }
        else if (this.platform == DvtDataGridUtils.IE_PLATFORM || this.platform == DvtDataGridUtils.EDGE_PLATFORM)
        {
            element['scrollLeft'] = scrollLeft;
        }
        else
        {
            // webkit based browsers
            width = this.dataGrid.getScrollableWidth();
            elemWidth = element['clientWidth'];
            newPos = width - elemWidth - scrollLeft;
            element['scrollLeft'] = newPos;
        }
    }
    else
    {
        element['scrollLeft'] = scrollLeft;
    }
};

/**
 * Determines the operating system. This value should be cached to prevent costly calculations. This value should be
 * treated as a guess, as this code is copied from AdfAgent.guessOS().
 * @param {string|null} userAgent The lowercase user agent string, if available.
 * @return {string} The DvtDataGridUtils.***_OS constant describing the platform.
 * @private
 */
DvtDataGridUtils.prototype._determineOS = function(userAgent)
{
    if (userAgent)
    {
        if (userAgent.indexOf('win') != -1)
        {
            return DvtDataGridUtils.WINDOWS_OS;
        }
        else if (userAgent.indexOf('mac') != -1)
        {
            return DvtDataGridUtils.MAC_OS;
        }
        else if (userAgent.indexOf('sunos') != -1)
        {
            return DvtDataGridUtils.SOLARIS_OS;
        }
    }

    return DvtDataGridUtils.UNKNOWN_OS;
};

/**
 * Determines the name of the platform. This value should be cached to prevent costly calculations.
 * Copied from DvtAgent (which itself is copied from AdfAgent)
 * @param {string|null} userAgent The lowercase user agent string, if available.
 * @return {string} The DvtDataGridUtils.***_PLATFORM constant describing the platform.
 * @private
 */
DvtDataGridUtils.prototype._determinePlatform = function(userAgent)
{
    if (userAgent)
    {
        if (userAgent.indexOf("opera") != -1) // check opera first, since it mimics other browsers
        {
            return DvtDataGridUtils.UNKNOWN_PLATFORM;
        }
        else if (userAgent.indexOf("trident") != -1 || userAgent.indexOf("msie") != -1)
        {
            return DvtDataGridUtils.IE_PLATFORM;
        }
        else if (userAgent.indexOf("edge") != -1)
        {
            return DvtDataGridUtils.EDGE_PLATFORM;
        }
        else if ((userAgent.indexOf("applewebkit") != -1) || (userAgent.indexOf("safari") != -1))
        {
            return DvtDataGridUtils.WEBKIT_PLATFORM;
        }
        else if (userAgent.indexOf("gecko/") != -1)
        {
            return DvtDataGridUtils.GECKO_PLATFORM;
        }
    }

    return DvtDataGridUtils.UNKNOWN_PLATFORM;
};

/**
 * Determines the what mousewheel event the browser recognizes
 * All modern browsers support wheel event
 * @return {string} The event which the browser uses to track mosuewheel events
 * @private
 */
DvtDataGridUtils.prototype.getMousewheelEvent = function()
{
    return "wheel";
};

/**
 * The standard wheel event and WheelEvent API now uses deltaMode and just deltaX and deltaY as the
 * properties for determining scroll.
 * @param {Event} event the mousewheel scroll event
 * @return {Object} change in X and Y if applicable through a mousewheel event, properties are deltaX, deltaY
 * @private
 */
DvtDataGridUtils.prototype.getMousewheelScrollDelta = function(event)
{
    var deltaX = 0, deltaY = 0, deltaMode, scrollConstant;
    deltaMode = event['deltaMode'];
    if (deltaMode == event['DOM_DELTA_PIXEL'])
    {
         scrollConstant = -1;
    }
    else if (deltaMode == event['DOM_DELTA_LINE'] || deltaMode == event['DOM_DELTA_PAGE'])
    {
        // only on firefox now, we will scroll 40 times the number of lines they
        // they want to scroll
         scrollConstant = -40;
    }
    deltaX = event['deltaX'] * scrollConstant;
    deltaY = event['deltaY'] * scrollConstant;

    return {"deltaX": deltaX, "deltaY": deltaY};
};

/**
 * Empty out (clear all children and attributes) from an element
 * @param {Element} elem the dom element to empty out
 * @private
 */
DvtDataGridUtils.prototype.empty = function(elem)
{
    while (elem.firstChild)
    {
        this.dataGrid._remove(elem.firstChild);
    }
};

/**
 * Does the browser support transitions
 * @private
 */
DvtDataGridUtils.prototype.supportsTransitions = function() {
    var b = document.body || document.documentElement,
            s = b.style,
            p = 'transition',
            ieBefore11 = /MSIE \d/.test(navigator.userAgent) &&
            (document.documentMode == null || document.documentMode < 11);

    if (ieBefore11)
    {
        return false;
    }

    if (typeof s[p] == 'string') {
        return true;
    }

    // Tests for vendor specific prop
    var v = ['Moz', 'webkit', 'Webkit', 'Khtml', 'O', 'ms'];
    p = p.charAt(0).toUpperCase() + p.substr(1);

    for (var i = 0; i < v.length; i++)
    {
        if (typeof s[v[i] + p] == 'string') {
            return true;
        }
    }

    return false;
};

/**
 * Return whether the node is editable or clickable
 * @param {Node|Element} node Node
 * @param {Element} databody Databody
 * @return {boolean} true or false
 * @private
 */
DvtDataGridUtils.prototype._isNodeEditableOrClickable = function(node, databody)
{
    var nodeName, tabIndex, origTabIndex;
    while (null != node && node != databody)
    {
        nodeName = node.nodeName;

        // If the node is a text node, move up the hierarchy to only operate on elements
        // (on at least the mobile platforms, the node may be a text node)
        if (node.nodeType == 3) // 3 is Node.TEXT_NODE
        {
            node = node.parentNode;
            continue;
        }

        tabIndex = parseInt(node.getAttribute('tabIndex'), 10);
        // datagrid overrides the tab index, so we should check if the data-oj-tabindex is populated
        origTabIndex = parseInt(node.getAttribute(this.dataGrid.getResources().getMappedAttribute('tabindex')), 10);

        if (tabIndex != null && tabIndex >= 0)
        {
            if (this.containsCSSClassName(node, this.dataGrid.getResources().getMappedStyle('cell'))
                    || this.containsCSSClassName(node, this.dataGrid.getResources().getMappedStyle('headercell'))
                    || this.containsCSSClassName(node, this.dataGrid.getResources().getMappedStyle('endheadercell')))
            {
                // this is the cell element
                return false;
            }
            else
            {
                return true;
            }
        }
        else if (nodeName.match(/^INPUT|SELECT|OPTION|BUTTON|^A\b|TEXTAREA/))
        {
            // ignore elements with tabIndex == -1
            if (tabIndex != -1 || origTabIndex != -1)
            {
                return true;
            }
        }
        node = node.parentNode;
    }
    return false;
};

/**
 * On certain browser the outline is postioned differently and requires offset. Chrome/Safari on Mac.
 * @return {boolean} true if the outline needs to be offset
 * @private
 */
DvtDataGridUtils.prototype.shouldOffsetOutline = function()
{
    if (this.os == DvtDataGridUtils.MAC_OS && this.platform == DvtDataGridUtils.WEBKIT_PLATFORM)
    {
        return true;
    }
    return false;
};
/**
 * The DvtDataGridOptions object provides convenient methods to access the options passed to the Grid.
 * @param {Object=} options - options set on the grid.
 * @param {Function=} rendererWrapperFunction - callback function used to fix renderer function for custom element.
 * @constructor
 * @private
 */
var DvtDataGridOptions = function(options, rendererWrapperFunction)
{
    this.options = options;
    this.rendererWrapperFunction = rendererWrapperFunction;
};

/**
 * Helper method to extract properties
 * @param {string=} arg1 - key to extract in object
 * @param {string=} arg2 - key to extract in the value of object[arg1]
 * @param {string=} arg3 - key to extract in the value of object[arg1][arg2]
 * @param {string=} arg4 - key to extract in the value of object[arg1][arg2][arg3]
 * @return {string|number|Object|boolean|null}
 */
DvtDataGridOptions.prototype.extract = function(arg1, arg2, arg3, arg4)
{
    var val1, val2, val3;
    if (arg1 != null)
    {
        val1 = this.options[arg1];
        if (arg2 != null && val1 != null)
        {
            val2 = val1[arg2];
            if (arg3 != null && val2 != null)
            {
                val3 = val2[arg3];
                if (arg4 != null && val3 != null)
                {
                    return val3[arg4];
                }
                return val3;
            }
            return val2;
        }
        return val1;
    }
    return null;
};

/**
 * Evaluate the a function, if it is a constant return it
 * @param {string|number|Object|boolean|null} value - function or string of the raw property
 * @param {Object} obj - object to pass into value if it is a function
 * @return {string|number|Object|boolean|null} the evaluated value(obj) if value a function, else return value
 */
DvtDataGridOptions.prototype.evaluate = function(value, obj)
{
    if (typeof value == 'function')
    {
        return value.call(this, obj);
    }
    return value;
};

/**
 * Get the property that was set on the option
 * @param {string} prop - the property to get from options
 * @param {string|undefined} axis - subobject to get row/column/cell/rowEnd/columnEnd
 * @return {string|number|Object|boolean|null} the raw value of the property
 */
DvtDataGridOptions.prototype.getRawProperty = function(prop, axis)
{
    var arg1, arg2, arg3;
    if (axis == "row" || axis == "column" || axis == "rowEnd" || axis == "columnEnd")
    {
        arg1 = "header";
        arg2 = axis;
        arg3 = prop;
    }
    else if (axis == "cell")
    {
        arg1 = "cell";
        arg2 = prop;
    }

    return this.extract(arg1, arg2, arg3);
};

/**
 * Get the evaluated property of the options
 * @param {string} prop - the property to get from options
 * @param {string=} axis - subobject to get row/column/cell
 * @param {Object=} obj - object to pass into property if it is a function
 * @return the evaluated property
 */
DvtDataGridOptions.prototype.getProperty = function(prop, axis, obj)
{
    if (obj === undefined)
    {
        return this.extract(prop, axis);
    }
    else
    {
        return this.evaluate(this.getRawProperty(prop, axis), obj);
    }
};

/**
 * Get the null or default value if no value set
 * @param {string|number|Object|boolean|null} value - the value of the property
 * @param {boolean|null} defValue - the default value of the property
 * @return {string|number|Object|boolean|null} - value if not null, defValue if value is null, null if neither null
 */
DvtDataGridOptions.prototype.nullOrDefault = function(value, defValue)
{
    if (value != null)
    {
        return value;
    }
    if (defValue != null)
    {
        return defValue;
    }
    return null;
};

////////////////////////// Grid options /////////////////////////////////

/**
 * Get the row banding interval from the grid options
 * @return {string|number|Object|boolean} the row banding interval
 */
DvtDataGridOptions.prototype.getRowBandingInterval = function()
{
    var bandingInterval = this.getProperty("bandingInterval", "row");
    if (bandingInterval != null)
    {
        return bandingInterval;
    }
    return 0;
};

/**
 * Get the column banding interval from the grid options
 * @return {string|number|Object|boolean} the column banding interval
 */
DvtDataGridOptions.prototype.getColumnBandingInterval = function()
{
    var bandingInterval = this.getProperty("bandingInterval", "column");
    if (bandingInterval != null)
    {
        return bandingInterval;
    }
    return 0;
};

/**
 * Get the empty text from the grid options
 * @return {string|number|Object|boolean|null} the empty text
 */
DvtDataGridOptions.prototype.getEmptyText = function()
{
    return this.getProperty("emptyText");
};

/**
 * Get the horizontal gridlines from the grid options
 * @return {string|number|Object|boolean|null} horizontal gridlines visible/hidden
 */
DvtDataGridOptions.prototype.getHorizontalGridlines = function()
{
    var horizontalGridlines;
    horizontalGridlines = this.extract("gridlines", "horizontal");
    if (horizontalGridlines != null)
    {
        return horizontalGridlines;
    }
    return "visible";
};

/**
 * Get the vertical gridlines from the grid options
 * @return {string|number|Object|boolean|null} vertical gridlines visible/hidden
 */
DvtDataGridOptions.prototype.getVerticalGridlines = function()
{
    var verticalGridlines;
    verticalGridlines = this.extract("gridlines", "vertical");
    if (verticalGridlines != null)
    {
        return verticalGridlines;
    }
    return "visible";
};

/**
 * Get the scroll position from the grid options
 * @return {string|number|Object|boolean|null} scrollPositionObject
 */
DvtDataGridOptions.prototype.getScrollPosition = function()
{
    var scrollPosition = this.getProperty("scrollPosition");
    if (scrollPosition != null)
    {
        return scrollPosition;
    }
    return null;
};

/**
 * Get the selection mode cardinality (none, single multiple)
 * @return {string|number|Object|boolean|null} none/single/multiple
 */
DvtDataGridOptions.prototype.getSelectionCardinality = function()
{
    var mode, key, val;
    mode = this.getProperty("selectionMode");
    if (mode == undefined)
    {
        return "none";
    }

    key = this.getSelectionMode();
    val = mode[key];
    if (val != null)
    {
        return val;
    }
    return "none";
};

/**
 * Get the selection mode (row,cell)
 * @return {string|number|Object|boolean|null} row/cell
 */
DvtDataGridOptions.prototype.getSelectionMode = function()
{
    var mode, cardinality;
    mode = this.getProperty("selectionMode");
    if (mode == undefined)
    {
        return "cell";
    }

    cardinality = mode['row'];
    if (cardinality != null && cardinality != "none")
    {
        return "row";
    }
    return "cell";
};

/**
 * Get the current selection from the grid options
 * @return {Array|null} the current selection from options
 */
DvtDataGridOptions.prototype.getSelection = function()
{
    return this.getProperty("selection");
};

/**
 * Get the current cell from the grid options
 * @return {Object|null} the current cell from options
 */
DvtDataGridOptions.prototype.getCurrentCell = function()
{
    return this.getProperty("currentCell");
};

/**
 * Get the editMode (none,cell)
 * @return {string|number|Object|boolean|null} default/enter
 */
DvtDataGridOptions.prototype.getEditMode = function()
{
    return this.getProperty("editMode");
};
////////////////////////// Grid header/cell options /////////////////////////////////
/**
 * Is the given header sortable
 * @param {string} axis - axis to check if sort enabled
 * @param {Object} obj - header context
 * @return {string|number|Object|boolean|null} default or null
 */
DvtDataGridOptions.prototype.isSortable = function(axis, obj)
{
    return this.nullOrDefault(this.getProperty("sortable", axis, obj), false);
};

/**
 * Is the given header resizable
 * @param {string} axis - axis to check if resizing enabled
 * @param {string} dir - width/height
 * @return {string|number|Object|boolean|null} enable, disable, or null
 */
DvtDataGridOptions.prototype.isResizable = function(axis, dir, obj)
{
    var v = this.extract('header', axis, 'resizable', dir);
    if (obj != undefined)
    {
        return this.evaluate(v, obj);
    }
    return v;
};

/**
 * Gets the dnd rorderable option
 * @param {string} axis the axis to get the reorder property from
 * @return {string|number|Object|boolean|null} enable, disable, or null
 */
DvtDataGridOptions.prototype.isMoveable = function(axis)
{
    return this.nullOrDefault(this.extract("dnd", "reorder", axis), false);
};

/**
 * Get the inline style property on an object
 * @param {string} axis - axis to get style of
 * @param {Object} obj - context
 * @return {string|number|Object|boolean|null} inline style
 */
DvtDataGridOptions.prototype.getInlineStyle = function(axis, obj)
{
    return this.getProperty("style", axis, obj);
};

/**
 * Get the style class name property on an object
 * @param {string} axis - axis to get class name of
 * @param {Object} obj - context
 * @return {string|number|Object|boolean|null} class name
 */
DvtDataGridOptions.prototype.getStyleClass = function(axis, obj)
{
    return this.getProperty("className", axis, obj);
};

/**
 * Get the renderer of an axis
 * @param {string} axis - axis to get style of
 * @return {string|number|Object|boolean|null} axis renderer
 */
DvtDataGridOptions.prototype.getRenderer = function(axis)
{
    // return type expected to be function, so just return without further
    // evaluation
    if (this.rendererWrapperFunction)
      return this.rendererWrapperFunction(this.getRawProperty("renderer", axis));
    return this.getRawProperty("renderer", axis);
};

/**
 * Get the scroll mode
 * @return {string} the scroll mode, which can be either "scroll", "loadMoreOnScroll", "auto".
 */
DvtDataGridOptions.prototype.getScrollPolicy = function()
{
    var mode = this.getProperty("scrollPolicy");
    if (mode == null)
    {
        mode = "auto";
    }

    return mode;
};


/**
 * Get the scroll policy options
 */
DvtDataGridOptions.prototype.getScrollPolicyOptions = function()
{
    return this.getProperty("scrollPolicyOptions");
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojDataGrid
 * @augments oj.baseComponent
 * @since 0.6
 * @ojstatus preview
 * @ojrole application
 * @ojrole grid
 * @ojshortdesc Displays data in a cell oriented grid.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="datagridOverview-section">
 *   JET DataGrid
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#datagridOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET DataGrid is a themable, WAI-ARIA compliant element that displays data in a cell oriented grid.  Data inside the DataGrid can be associated with row and column headers.  Page authors can customize the content rendered inside cells and headers.</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-data-grid
 *   id="datagrid"
 *   style="width:250px;height:250px"
 *   aria-label="My Data Grid"
 *   data='{{dataSource}}'>
 * &lt;/oj-data-grid></code></pre>
 *
 * <h3 id="data-section">
 *   Data
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#data-section"></a>
 * </h3>
 * <p>The JET DataGrid gets its data from a DataGridDataSource.  There are several types of DataGridDataSource that are provided out of the box:</p>
 * <ul>
 * <li>{@link oj.ArrayDataGridDataSource} - Use this when the underlying data is a static array.  The ArrayDataGridDataSource supports both single array (in which case each item in the array represents a row of data in the DataGrid) and two dimensional array (in which case each item in the array represents a cell in the DataGrid).  See the documentation for oj.ArrayDataGridDataSource for more details on the available options.</li>
 * <li>{@link oj.CollectionDataGridDataSource} - Use this when oj.Collection is the model for the underlying data.  Note that the DataGrid will automatically react to model event from the underlying oj.Collection.  See the documentation for oj.CollectionDataGridDataSource for more details on the available options.</li>
 * <li>{@link oj.CubeDataGridDataSource} - Use this when aggregating data with oj.Cube. See the documentation for oj.CubeDataGridDataSource for more details on the available options.</li>
 * <li>{@link oj.PagingDataGridDataSource} - Use this when the DataGrid is driven by an associating ojPagingControl.  See the documentation for oj.PagingDataGridDataSource for more details on the available options.</li>
 * <li>{@link oj.FlattenedTreeDataGridDataSource} - Use this when hierarchical data is displayed in the DataGrid.  The FlattenedDataGridDataSource takes an oj.TreeDataSource and adapts that to the DataGridDataSource.  The ojRowExpander works with the FlattenedTreeDataGridDataSource to enable expanding/collapsing of rows.</li>
 * </ul>
 *
 * <p>Developers can also create their own DataSource by extending the oj.DataGridDataSource class.  See the cookbook for an example of a custom DataGridDataSource.</p>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>Since <code class="prettyprint">role="application"</code> is used in the DataGrid, application should always apply an <code class="prettyprint">aria-label</code> to the DataGrid element so that it can distinguish from other elements with application role.</p>
 *
 * <h3 id="context-section">
 *   Header Context And Cell Context
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#context-section"></a>
 * </h3>
 *
 * <p>For all header and cell attributes, developers can specify a function as the return value.  The function takes a single argument, which is an object that contains contextual information about the particular header or cell.  This gives developers the flexibility to return different value depending on the context.</p>
 *
 * <p>For header attributes, the context parameter contains the following keys:</p>
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>axis</kbd></td>
 *       <td>The axis of the header.  Possible values are 'row', 'column', 'rowEnd', and 'columnEnd'.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>componentElement</kbd></td>
 *       <td>A reference to the DataGrid element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>datasource</kbd></td>
 *       <td>A reference to the data source object.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>index</kbd></td>
 *       <td>The index of the header, where 0 is the index of the first header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>key</kbd></td>
 *       <td>The key of the header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>data</kbd></td>
 *       <td>The data object for the header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>parentElement</kbd></td>
 *       <td>The header cell element.  The renderer can use this to directly append content to the header cell element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>level</kbd></td>
 *       <td>The level of the header. The outermost header is level zero.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>depth</kbd></td>
 *       <td>The the number of levels the header spans.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>extent</kbd></td>
 *       <td>The number of indexes the header spans.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <p></p>
 * <p>For cell attributes, the context paramter contains the following keys:</p>
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>componentElement</kbd></td>
 *       <td>A reference to the DataGrid element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>datasource</kbd></td>
 *       <td>A reference to the data source object.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>indexes</kbd></td>
 *       <td>The object that contains both the zero based row index and column index in which the cell is bound to.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>keys</kbd></td>
 *       <td>The object that contains both the row key and column key which identifies the cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>cell</kbd></td>
 *       <td>An object containing attribute data which should be used to reference the data in the cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>data</kbd></td>
 *       <td>The plain data for the cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>parentElement</kbd></td>
 *       <td>The data cell element.  The renderer can use this to directly append content to the data cell element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>extents</kbd></td>
 *       <td>The object that contains both the row extent and column extent of the cell.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <p></p>
 * <p>If a FlattenedTreeDataGridDataSource is used, the following additional contextual information are available:</p>
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>depth</kbd></td>
 *       <td>The depth of the row.  The depth of root row is 0.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>index</kbd></td>
 *       <td>The index of the row relative to its parent.  The index of the first child is 0.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>state</kbd></td>
 *       <td>The state of the row.  Possible values are 'expanded', 'collapsed', 'leaf'.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>parentKey</kbd></td>
 *       <td>The key of the parent row.  For root row the parent key is null.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <p></p>
 * <p>Note that a custom DataGridDataSource can return additional header and cell context information.  Consult the documentation of the DataGridDataSource API for details.</p>
 *
 * <h3 id="context-section">
 *   Selection
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#selection-section"></a>
 * </h3>
 *
 * <p>The DataGrid supports both cell based and row based selection mode, which developers can specify using the selectionMode attribute.  For each mode developers can also specify whether single or multiple cells/rows can be selected.</p>
 * <p>Developers can specify or retrieve selection from the DataGrid using the selection attribute.  A selection in DataGrid consists of an array of ranges.  Each range contains the following keys: startIndex, endIndex, startKey, endKey.  Each of the keys contains value for 'row' and 'column'.  If endIndex and endKey are not specified, -1, or null, that means the range is unbounded, i.e. the cells of the entire row/column are selected.</p>
 * 
 *
 * <h3 id="geometry-section">
 *   Geometry Management
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#geometry-section"></a>
 * </h3>
 *
 * <p>If the DataGrid is not styled with a fixed size, then it will responds to a change to the size of its container.  Note that unlike Table the content of the cell does not affect the height of the row.  The height of the rows must be pre-determined and specified by the developer or a default size will be used.</p>
 *
 * <p>The DataGrid does not support % width and height values in the header style or style class.</p>
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>The order of the column headers will be rendered in reverse order in RTL reading direction.  The location of the row header will also be different between RTL and LTR direction.  It is up to the developers to ensure that the content of the header and data cell are rendered correctly according to the reading direction.</p>
 * <p>As with any JET element, in the unusual case that the directionality (LTR or RTL) changes post-init, the DataGrid must be <code class="prettyprint">refresh()</code>ed.
 *
 * <h3 id="rtl-section">
 *   Templating Alignment
 *   <a class="bookmarkable-link" title="Templating Alignment" href="#templating-section"></a>
 * </h3>
 * <p>The DataGrid cells are horizontal flex boxes. To change the alignment use the classes documented in the flex layout section of the cookbook along with header and cell className attributes.
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it's recommended that applications limit the amount of data to display.  Displaying large
 * number of items in DataGrid makes it hard for users to find what they are looking for, but affects the
 * scrolling performance as well.  If displaying large number of items is neccessary, consider use a paging control with DataGrid
 * to limit the number of items to display at a time.  Also consider setting <code class="prettyprint">scrollPolicy</code> to
 * <code class="prettyprint">'scroll'</code> to enable virtual scrolling to reduce the number of elements in the DOM at any given time .</p>
 *
 * <h4>Cell Content</h4>
 * <p>DataGrid allows developers to specify arbitrary content inside its cells. In order to minimize any negative effect on
 * performance, you should avoid putting large numbers of heavy-weight content inside a cell because as you add more complexity
 * to the structure, the effect will be multiplied because there can be many items in the DataGrid.</p>
 *
 * <h4>Templating</h4>
 * <p>When deciding to use a template or renderer, consider the conditionality of the template. Having templates with several if blocks
 * can significantly hinder performance. If you have this case either specify a function for the template to remove conditions from the template itself or
 * use the renderer attribute.</p>
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="contextmenu-section">
 *   Context Menu
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#contextmenu-section"></a>
 * </h3>
 * <p>The DataGrid has a default context menu for accessibly performing operations such as header resize and sort.
 * When defining a context menu, DataGrid allows the app to use the built-in behavior for operations such as header
 * resize and sort by specifying menu list items as follows.</p>
 *
 * <ul><li> &lt;li data-oj-command="oj-datagrid-['commandname']" /&gt;</li></ul>
 *
 * <p>Note that if no &lt;a&gt; element is specified inside of a list item with a command,
 * the translated text from the default menu will be supplied in an anchor tag.</p>
 *
 * <p>The supported commands:</p>
 * <table class="keyboard-table">
 *   <thead>
 *      <tr>
 *       <th>Default Function</th>
 *       <th>data-oj-command value</th>
 *      </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Resize menu</kbd> (contains width and height resize)</td>
 *       <td>oj-datagrid-resize</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Sort Row menu</kbd> (contains ascending and descending sort)</td>
 *       <td>oj-datagrid-sortRow</td>
 *      </tr>
 *     <tr>
 *        <td><kbd>Sort Column menu</kbd> (contains ascending and descending sort)</td>
 *       <td>oj-datagrid-sortCol</td>
 *      </tr>
 *     <tr>
 *        <td><kbd>Resize Width</kbd></td>
 *       <td>oj-datagrid-resizeWidth</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Resize Height</kbd></td>
 *       <td>oj-datagrid-resizeHeight</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Sort Row Ascending</kbd></td>
 *       <td>oj-datagrid-sortRowAsc</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Sort Row Descending</kbd></td>
 *       <td>oj-datagrid-sortRowDsc</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Sort Column Ascending</kbd></td>
 *       <td>oj-datagrid-sortColAsc</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Sort Column Descending</kbd></td>
 *       <td>oj-datagrid-sortColDsc</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Cut</kbd></td>
 *       <td>oj-datagrid-cut</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Paste</kbd></td>
 *       <td>oj-datagrid-paste</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Toggle Non-Contiguous Selection on Touch Device</kbd></td>
 *       <td>oj-datagrid-discontiguousSelection</td>
 *     </tr>
 * </tbody></table>
 */
oj.__registerWidget('oj.ojDataGrid', $['oj']['baseComponent'],
{
    widgetEventPrefix: 'oj',
    options:
            {
                /**
                 * <p>Row banding and column banding intervals within the DataGrid body.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {Object.<string, number>}
                 * @default {"row":0, "column":0}
                 * @property {number} row row banding interval
                 * @property {number} column column banding interval
                 * @ojshortdesc Specifies the data body row and column banding intervals.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">bandingInterval</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid banding-interval.row='1' banding-interval.column='1'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid banding-interval='{"row": 1, "column": 1}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">bandingInterval</code> property after initialization:</caption>
                 * // Get one
                 * var bandingIntervalValue = myDataGrid.bandingInterval.row;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('bandingInterval.row', 1);
                 *
                 * // Get all
                 * var bandingIntervalValues = myDataGrid.bandingInterval;
                 *
                 * // Set all.  Must list every resource key, as those not listed are lost.
                 * myDataGrid.bandingInterval = {
                 *     row: 1,
                 *     column: 1
                 * };
                 */
                bandingInterval: {'row': 0, 'column': 0},
                /**
                 * <p>The data source for the DataGrid must be an extension of oj.DataGridDataSource.
                 * See the <a href="#data-section">data section</a> in the introduction for data sources provided out of the box.
                 * If the data attribute is not specified, an empty DataGrid is displayed.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {oj.DataGridDataSource}
                 * @default null
                 * @ojshortdesc Specifies the grid data source.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">data</code> attribute specified:</caption>
                 * &lt;oj-data-grid data='{{myDataSource}}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">data</code> property after initialization:</caption>
                 * // getter
                 * var dataValue = myDataGrid.data;
                 *
                 * // setter
                 * myDataGrid.data = myDataSource;
                 */
                data: null,
                /**
                 * <p>Display or hide the horizontal or vertical gridlines in the data body. Gridlines are
                 * visible by default and must be set to <code class="prettyprint">'hidden'</code> in order to be hidden.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {Object.<string, string>}
                 * @default {"horizontal": "visible", "vertical": "visible"}
                 * @property {string} horizontal horizontal gridlines, valid values are: 'hidden', 'visible'
                 * @property {string} vertical vertical gridlines, valid values are: 'hidden', 'visible'
                 * @ojshortdesc Specifies the visibility of the grid body vertical and horizontal gridlines.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">gridlines</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid gridlines.horizontal='hidden' gridlines.vertical='hidden'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid gridlines='{"horizontal":"hidden", "vertical":"hidden"}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">gridlines</code> property after initialization:</caption>
                 * // Get one
                 * var gridlinesValue = myDataGrid.gridlines.horizontal;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('gridlines.horizontal', 'hidden');
                 *
                 * // Get all
                 * var gridlinesValues = myDataGrid.gridlines;
                 *
                 * // Set all.  Must list every resource key, as those not listed are lost.
                 * myDataGrid.gridlines = {
                 *     horizontal: 'hidden',
                 *     vertical: 'hidden'
                 * };
                 */
                gridlines: {'horizontal': 'visible', 'vertical': 'visible'},
                /**
                 * <p>The information about the data grid scroll position. Contains the x,y pixel coordinates, as well as the index
                 * and key information of the cell closest to the origin of the grid data body.
                 *
                 * <p>The default value contains just the x,y pixel values specified. Once data is fetched the index and key sub-properties will be added.
                 * If there is no data the index and key sub-properties will not be available.
                 *
                 * <p>When setting the scrollPosition property applications can change any combination of the sub-properties.
                 * If multiple sub-properties are set at once they will be used in key, index, pixel order where the latter can serve as hints.
                 * If upon scrolling to the key the pixel position is within the cell the adjustment will be made to scroll to that pixel location.
                 * If a sparse object is set the other sub-properties will be populated and updated once the grid has scrolled to that position.
                 *
                 * <p>If the scrollPolicy is 'loadMoreOnScroll' and the scrollPosition is set to something outside of the currently rendered region,
                 * the grid will attempt to fetch until the specified scroll position is satisfied. If scrollPosition pixels are set out of the maximum visible
                 * bounds once all cells have been rendered or the indexes are set out of the data source bounds once all data has been fetched, the scroll
                 * position will be set at its maximum value along that axis.
                 *
                 * <p>If only setting scrollPosition using the keys sub-property to something outside of the currently rendered region, the grid will only scroll to that position
                 * if the keys are present in the data source.
                 *
                 * <p>Note that a scrollPosition set on the grid may be changed if the set value cannot be displayed the origin of the grid. For example if you scroll to the last row index,
                 * it will likely be updated to an index before that which corresponds to the row index at the origin, rather than the last row index which is at the bottom
                 * of the visible region.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {Object}
                 * @default <code class="prettyprint">{x: 0, y: 0}</code>
                 * @property {number=} x the horizontal position in pixel
                 * @property {number=} y the vertical position in pixel
                 * @property {number=} rowIndex the zero-based row index of the cell at the origin of the grid.  If <a href="#scrollPolicy">scrollPolicy</a> is set to 'loadMoreOnScroll' 
                 * and the index is greater than maxCount set in <a href="#scrollPolicyOptions">scrollPolicyOptions</a>, then it will scroll and fetch
                 * until the end of the list is reached and there's no more items to fetch.
                 * @property {number=} columnIndex the zero-based column index of the cell at the origin of the grid.  If <a href="#scrollPolicy">scrollPolicy</a> is set to 'loadMoreOnScroll' 
                 * and the index is greater than maxCount set in <a href="#scrollPolicyOptions">scrollPolicyOptions</a>, then it will scroll and fetch
                 * until the end of the list is reached and there's no more items to fetch.
                 * @property {*=} rowKey the row key of the cell at the origin of the grid. If DataGridDataSource is used for <a href="data">data</a> and the key does not exist in the 
                 * DataGridDataSource, then the value is ignored.  If it is unknown in the data source then the grid will fetch and scroll until the item is found
                 * or the end of the axis is reached and there's no more items to fetch.
                 * @property {*=} columnKey the column key of the cell at the origin of the grid. If DataGridDataSource is used for <a href="data">data</a> and the key does not exist in the 
                 * DataGridDataSource, then the value is ignored.  If it is unknown in the data source then the grid will fetch and scroll until the item is found
                 * or the end of the axis is reached and there's no more items to fetch.
                 * @property {number=} offsetX the horizontal offset in pixel relative to the cell identified by key/index.
                 * @property {number=} offsetY the vertical offset in pixel relative to the cell identified by key/index.
                 * @ojshortdesc Gets and sets the scroll position of the data grid.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">scrollPosition</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid scroll-position.rowIndex='1' scroll-position.columnIndex='1'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid scroll-position='{"rowIndex": 1, "columnIndex": 1}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">scrollPosition</code> property after initialization:</caption>
                 * // Get one
                 * var scrollPositionValue = myDataGrid.scrollPosition.rowIndex;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('scrollPosition.rowIndex', 1);
                 *
                 * // Get all
                 * var scrollPositionValues = myDataGrid.scrollPosition;
                 *
                 * // Set all. Those not listed will be lost until the scroll completes and the remaining fields are populated.
                 * myDataGrid.scrollPosition = {
                 *     x: 50,
                 *     y: 45
                 * };
                 */
                scrollPosition: {x: 0, y: 0}, 
                /**
                 * <p>Specifies whether row or cell selection can be made and the cardinality of each (single/multiple/none) selection in the DataGrid.
                 * Only one of the attributes, row or cell, should not be <code class="prettyprint">none</code> at a time. 
                 * If both are not <code class="prettyprint">none</code> then an error will be thrown.
                 * If the selection mode is changed at runtime the selection will be cleared.
                 * 
                 * Selection is initially disabled, but setting both attributes' value to <code class="prettyprint">none</code> will disable selection.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {Object.<string, string>}
                 * @default {cell:'none', row:'none'}
                 * @property {string} row set row selection mode, valid values are: 'none', 'single', 'multiple'
                 * @property {string} cell set cell selection mode, valid values are: 'none', 'single', 'multiple'
                 * @ojshortdesc Specifies row or cell selection.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">selectionMode</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid selection-mode.cell='multiple'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid selection-mode='{"cell": "multiple"}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> property after initialization:</caption>
                 * // Get one
                 * var selectionModeValue = myDataGrid.selectionMode.cell;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('selectionMode.cell', 'multiple');
                 *
                 * // Get all
                 * var selectionModeValues = myDataGrid.selectionMode;
                 *
                 * // Set all.  Must list every resource key, as those not listed are lost.
                 * myDataGrid.selectionMode = {
                 *     cell: 'multiple'
                 *     row: 'none'
                 * };
                 */
                selectionMode: {
                    /**
                     * <p>Specifies whether cell selection can be made and the cardinality (single/multiple/none).
                     * Only one of the <code class="prettyprint">selectionMode</code> attributes, row or cell, should not be <code class="prettyprint">none</code> at a time. 
                     * If both are not <code class="prettyprint">none</code> then an error will be thrown.
                     * 
                     * Cell selection is initially disabled, but setting the attribute's value to <code class="prettyprint">none</code> will disable cell selection. 
                     * If the selection mode is changed at runtime the selection will be cleared.
                     *
                     * @expose
                     * @memberof! oj.ojDataGrid
                     * @name selectionMode.cell
                     * @instance
                     * @type {string}
                     * @default 'none'
                     * @ojvalue {string} "none" no cell selection
                     * @ojvalue {string} "single" at most one cell selected at a time
                     * @ojvalue {string} "multiple" any number of cells selected at a time
                     * @ojshortdesc Specifies cell selection.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">selectionMode.cell</code> attribute specified:</caption>
                     * &lt;!-- Using dot notation -->
                     * &lt;oj-data-grid selection-mode.cell='multiple'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">selectionMode.cell</code> property after initialization:</caption>
                     * // getter
                     * var selectionModecellValue = myDataGrid.selectionMode.cell;
                     *
                     * // setter
                     * myDataGrid.selectionMode.cell = 'single';
                     */
                    cell: 'none',
                    /**
                     * <p>Specifies whether row selection can be made and the cardinality (single/multiple/none).
                     * Only one of the <code class="prettyprint">selectionMode</code> attributes, row or cell, should not be <code class="prettyprint">none</code> at a time. 
                     * If both are not <code class="prettyprint">none</code> then an error will be thrown.
                     * 
                     * Row selection is initially disabled, but setting the attribute's value to <code class="prettyprint">none</code> will row disable selection.
                     * If the selection mode is changed at runtime the selection will be cleared.
                     *
                     * @expose
                     * @memberof! oj.ojDataGrid
                     * @name selectionMode.row
                     * @instance
                     * @type {string}
                     * @default 'none'
                     * @ojvalue {string} "none" no row selection
                     * @ojvalue {string} "single" at most one row selected at a time
                     * @ojvalue {string} "multiple" any number of rows selected at a time
                     * @ojshortdesc Specifies row selection.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">selectionMode.row</code> attribute specified:</caption>
                     * &lt;!-- Using dot notation -->
                     * &lt;oj-data-grid selection-mode.row='multiple'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">selectionMode.row</code> property after initialization:</caption>
                     * // getter
                     * var selectionModeRowValue = myDataGrid.selectionMode.row;
                     *
                     * // setter
                     * myDataGrid.selectionMode.row = 'single';
                     */
                    row: 'none'
                },
                /**
                 * <p>Enables or disables reordering the rows within the same DataGrid using drag and drop.
                 *
                 * <p>Specify an object with a subproperty <code class="prettyprint">reorder</code> set to <code class="prettyprint">{row:'enable'}</code> to enable
                 * reordering.  Setting the <code class="prettyprint">reorder</code> subproperty to <code class="prettyprint">{row:'disable'}</code>,
                 * or setting the <code class="prettyprint">dnd</code> attribute to <code class="prettyprint">null</code> (or omitting
                 * it), disables reordering support. There must be move capability support on the data source to enable this feature.
                 *
                 * @type {Object}
                 * @property {Object} reorder an object with property row
                 * @property {string} reorder.row row reordering within the DataGrid, valid values are: 'enable', 'disable'
                 *
                 * @default {"reorder": {"row" :"disable"}}
                 * @expose
                 * @instance
                 * @memberof oj.ojDataGrid
                 * @ojshortdesc Specifies the drag and drop row reordering.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">dnd</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid dnd.reorder.row='enable'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid dnd='{"reorder": {"row": "enable"}}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">dnd</code> property after initialization:</caption>
                 * // Get one
                 * var dndValue = myDataGrid.dnd.reorder.row;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('dnd.reorder.row', 'enable');
                 *
                 * // Get all
                 * var dndValues = myDataGrid.dnd;
                 *
                 * // Set all.  Must list every resource key, as those not listed are lost.
                 * myDataGrid.dnd = {
                 *     reorder: {
                 *          row: 'enable'
                 *     }
                 * };
                 */
                dnd : {'reorder': {'row' :'disable'}},
                /**
                 * <p>Specifies the mechanism used to scroll the data inside the DataGrid.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {string|null}
                 * @default 'auto'
                 * @ojvalue {string} "auto" the DataGrid will decide the scroll policy
                 * @ojvalue {string} "loadMoreOnScroll" additional data are fetched when the user scrolls to the bottom of the DataGrid and all previous data are kept in the DOM
                 * @ojvalue {string} "scroll" virtual scrolling is used meaning only rows/columns visible in the viewport are fetched and kept in the DOM
                 * @ojshortdesc Specifies the grid scroll mechanism.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">scrollPolicy</code> attribute specified:</caption>
                 * &lt;oj-data-grid scroll-policy='scroll'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">scrollPolicy</code> property after initialization:</caption>
                 * // getter
                 * var scrollPolicyValue = myDataGrid.scrollPolicy;
                 *
                 * // setter
                 * myDataGrid.scrollPolicy = 'scroll';
                 */
                scrollPolicy: 'auto',
                /**
                 * <p> The following options are supported:
                 * <ul>
                 *   <li>maxRowCount: Maximum rows which will be displayed before fetching more will be stopped. Only applies when scrollPolicy is 'loadMoreOnScroll'.</li>
                 *   <li>maxColumnCount: Maximum columns which will be displayed before fetching more will be stopped. Only applies when scrollPolicy is 'loadMoreOnScroll'.</li>
                 * </ul>
                 *
                 * @ojshortdesc Gets and sets the fetch options for scrolling behaviors that triggers data fetches.
                 * @expose
                 * @instance
                 * @memberof! oj.ojDataGrid
                 * @type {Object.<number>|null}
                 * @property {number=} maxRowCount The maximum total number of rows to fetch, -1 is unbounded
                 * @property {number=} maxColumnCount The maximum total number of columns to fetch, -1 is unbounded
                 * @default {'maxRowCount': 500, 'maxColumnCount': 500}
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">scroll-policy-options</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid scroll-policy-options.max-row-count='500'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">scrollPolicyOptions</code> property after initialization:</caption>
                 * // Get one
                 * var maxCountValue = myDataGrid.scrollPolicyOptions.maxRowCount;
                 *
                 * // Get all
                 * var scrollPolicyOptionsValues = myDataGrid.scrollPolicyOptions;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('scrollPolicyOptions.maxRowCount', 250);
                 *
                 * // Set all. Properties not listed will be lost.
                 * myDataGrid.scrollPolicyOptions = {maxRowCount: 1000};
                 */
                scrollPolicyOptions: {
                    /**
                     * The maximum number of rows which will be displayed before fetching more rows will be stopped.
                     * <p>See the <a href="#scrollPolicyOptions">scroll-policy-options</a> attribute for usage examples.</p>
                     *
                     * @expose
                     * @name scrollPolicyOptions.maxRowCount
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @type {number}
                     * @default 500
                     * @ojmin -1
                     */                    
                    'maxRowCount': 500, 
                    /**
                     * The maximum number of columns which will be displayed before fetching more rows will be stopped.
                     * <p>See the <a href="#scrollPolicyOptions">scroll-policy-options</a> attribute for usage examples.</p>
                     *
                     * @expose
                     * @name scrollPolicyOptions.maxColumnCount
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @type {number}
                     * @default 500
                     * @ojmin -1
                     */                    
                    'maxColumnCount': 500
                },
                /**
                 * <p>Specifies the current selections in the DataGrid.
                 * Returns an array of range objects, or an empty array if there's no selection.
                 *
                * <p>Cell Selection Range Objects:
                * The cell selection range object contains subproperties:
                * startIndex and startKey referring to the cell in the selection closest to the origin of the grid.
                * endIndex and endKey referring to the cell in the selection furthest from the origin of the grid.
                * All four subproperties will be objects with row and column subproperties themselves.
                * Single cell selection would always have the same startIndex/endIndex and startKey/endKey.
                * Example: cell selection for cell (1,2) to cell (3,4)
                * [{startIndex: {row: 1, column: 2}, endIndex: {row: 3, column: 4}, startKey: {row: "r1", column: "c2"}, endKey: {row: "r3", column: "c4"}}]
                 *
                * <p>Row Selection Range Objects:
                * The row selection range object contains subproperty objects:
                * startIndex and startKey referring to the row in the selection closest to the origin of the grid.
                * endIndex and endKey referring to the row in the selection furthest from the origin of the grid.
                * All four subproperties will be objects with only row as a subproperty themselves.
                * Single row selection would always have the same startIndex/endIndex and startKey/endKey.
                * Example: row selection for selecting row 1 to row 2
                * [{startIndex: {row: 1}, endIndex: {row: 2}, startKey: {row: "r1"}, endKey: {row: "r2"}}]
                *
                * <p>Entire Row/Column Selection in Cell Selection Mode:
                * In the case of an entire row or column selection:
                * The startIndex subproperty along that axis will be set to 0.
                * The endIndex subproperty along that axis will be set to -1, regardless of if the endIndex is known.
                * The endKey subproperty along that axis will be set to null, regardless of if the endKey is known.
                * Example: Cell selection spanning column 1 to column 2
                * [{startIndex: {row: 0, column: 1}, endIndex: {row: -1, column: 2}, startKey: {row: "r0", column: "c1"}, endKey: {row: null, column: "c2"}}]
                *
                * <p>Select All:
                * The startIndex subproperty values will be set to 0.
                * The endIndex subproperty values will be set to -1, regardless of if the endIndex is known.
                * Keys are set to null to remain consistent with row and column selections and to prevent misinterpretation of the selection range.
                * Select All will cover all the data in the data source including that which is not displayed in the viewport.
                * A select all type selection will not allow for deselections. 
                * Example:
                * [{startIndex: {row: 0, column: 0}, endIndex: {row: -1, column: -1}, startKey: {row: "r0", column: "c0"}, endKey: {row: null, column: null}]
                *
                * <p>Discontiguous Selection Ranges:
                * In the case of multiple discontiguous selections, each individual selection will be one entry in the selection range array. Selection ranges
                * may overlap with each other.
                * Example: (1,2) to (3,4), and (5,6) to (7,8) would look like
                * [{startIndex: {row: 1, column: 2}, endIndex: {row: 3, column: 4}, startKey: {row: "r1", column: "c2"}, endKey: {row: "r3", column: "c4"}},
                * {startIndex: {row: 5, column: 6}, endIndex: {row: 7, column: 8}, startKey: {row: "r5", column: "c6"}, endKey: {row: "r7", column: "c8"}}]
                *
                * <p>Setting the selection:
                * The selection can be set using any combination of startIndex/startKey and endIndex/endKey.
                * If a sparse object is set the DataGrid will fill in the unspecifed portions of the object, if the cells are in the viewport.
                * If they are not the DataGrid will attempt to ask the data source for the missing index or key. In the case of datasource unable to provide
                * a key or index, the key or index will be marked with undefined values. If Datagrid lacks the appropriate data to create the selection, it
                * will not create a selection and remove the invalid selection object.
                * If the selection is set out of the scope of the currently displayed data the DataGrid will not change the selection object.
                * As items within the selected range are scrolled into view they will be displayed as selected.
                *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {Array.<Object>}
                 * @default []
                 * @ojwriteback
                 * @ojshortdesc Specifies the grid's current selection.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">selection</code> attribute specified:</caption>
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid selection='[{"startIndex": {"row": 1, "column": 1}, "endIndex": {"row": 2, "column": 2}}]'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">selection</code> property after initialization:</caption>
                 * // Get one
                 * var selectionValue = myDataGrid.selection[0];
                 *
                 * // Get all
                 * var selectionValues = myDataGrid.selection;
                 *
                * // Set all. Setter may be used with only indices or keys.
                 * myDataGrid.selection = [{startIndex: {row: 1, column: 1}, endIndex: {row: 2, column: 2}}];
                 */
                selection: [],
                /**
                 * <p>The cell that currently have keyboard focus.  Note that if the current cell
                 * is set to an item that is currently not available (not fetched in highwatermark scrolling case or
                 * inside a collapsed parent node) or invalid, then the value is not applied.
                 *
                 * <p>If the currentCell is a databody cell the object will contain the following information:
                 * <code class="prettyprint">{type: 'cell', indexes: {row: rowIndexValue, column: columnIndexValue}, keys: {row: 'rowKeyValue', column: 'columnKeyValue'}}</code>
                 *
                 * <p>If the currentCell is a header cell the object will contain the following information:
                 * <code class="prettyprint">{type: 'header', axis:'axisValue', index: indexValue, key: 'keyValue', level: levelValue}</code>
                 *
                 * <p>If setting the property to a databody cell, either indexes or keys must be specified, if both are specified indexes will be used as a hint.
                 * If setting the property to a header cell, axis and either "index and level" or "key" must be specified, if both are specified "index and level" will be used as a hint.
                 * If level is not specified it will default to 0.
                 *
                 * @expose
                 * @instance
                 * @memberof oj.ojDataGrid
                 * @type {Object}
                 * @default null
                 * @ojwriteback
                 * @property {string} type designates whether a header or databody cell is the current cell, valid values are: 'cell', 'header'
                 * @property {string} axis header axis, available if type is header, valid values are: 'row', 'column', 'rowEnd', 'columnEnd'
                 * @property {number} index header index, available if type is header
                 * @property {number} level header level, available if type is header
                 * @property {number} key header key, available if type is header
                 * @property {Object} indexes cell indexes, available if type is cell
                 * @property {number} indexes.row cell row index
                 * @property {number} indexes.column cell column index
                 * @property {Object} keys cell keys, available if type is cell
                 * @property {string} keys.row cell row key
                 * @property {string} keys.column cell column key
                 * @ojshortdesc Specifies the grid's currently focused cell.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">currentCell</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid current-cell.type='cell' current-cell.indexes.row='1' current-cell.indexes.column='1'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid current-cell='{"type": "cell", "indexes":{"row": 1, "column": 1}}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">currentCell</code> property after initialization:</caption>
                 * // Get one
                 * var currentCellValue = myDataGrid.currentCell.type;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('currentCell.type', 'cell');
                 *
                 * // Get all
                 * var currentCellValues = myDataGrid.currentCell;
                 *
                 * // Set all.  Must list every resource key, as those not listed are lost.
                 * myDataGrid.currentCell = {
                 *    type: 'cell',
                 *    indexes:{
                 *        row: 1,
                 *        column: 1
                 *    }
                 * };
                 */
                currentCell: null,
                /**
                 * <p>Determine if the DataGrid is read only or editable.
                 * <p>Use <code class="prettyprint">none</code> if the DataGrid is strictly read only.
                 *
                 * <p>The DataGrid <code class="prettyprint">editMode</code> is designed to support fast editing and requires that the input controls render fast and near synchronously
                 * to be stamped inside an editable cell. The DataGrid supports overwrite behavior for all of JET's input components. Custom components that render synchronously or
                 * components that only render asynchronously due to Promises that resolve immediately will also be stampable. The modules for the stamped component loaded/required when the DataGrid is  
                 * loaded/required. Components that must go back to a server to render are not supported.
                 *
                 * <p>Use <code class="prettyprint">cellNavigation</code> to allow editable cells, but the DataGrid is currently read only and a single tab stop on the page.
                 * Pressing F2 or double click while in this mode will switch the DataGrid to <code class="prettyprint">cellEdit</code> mode.
                 *
                 * <p>Use <code class="prettyprint">cellEdit</code> to allow editable cells, and tab navigates to the next cell behavior.
                 * Pressing ESC while in this mode will switch the DataGrid to <code class="prettyprint">cellNavigation</code> mode.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {string}
                 * @default 'none'
                 * @ojwriteback
                 * @ojvalue {string} "none" the DataGrid is read only
                 * @ojvalue {string} "cellNavigation" the DataGrid is a single tab stop and editable at the cell level, but currently read only
                 * @ojvalue {string} "cellEdit" the DataGrid cells are individually tabbable and editable
                 * @ojshortdesc Specifies if the grid is read-only or editable.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">editMode</code> attribute specified:</caption>
                 * &lt;oj-data-grid edit-mode='cellEdit'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">editMode</code> property after initialization:</caption>
                 * // getter
                 * var editModeValue = myDataGrid.editMode;
                 *
                 * // setter
                 * myDataGrid.editMode = 'cellEdit';
                 */
                editMode: 'none',
                /**
                 * The header attribute contains a subset of attributes for row and column headers.
                 *
                 * @expose
                 * @alias header
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @ojshortdesc Specifies attributes on the headers.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">header</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid header.some-key='some value'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid header='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">header</code> property after initialization:</caption>
                 * // Get one
                 * var value = myDataGrid.header.someKey;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('header.someKey', 'someValue');
                 *
                 * // Get all
                 * var values = myDataGrid.header;
                 *
                 * // Set all.  Must list every resource key, as those not listed are lost.
                 * myDataGrid.header = {
                 *      someKey: 'some value',
                 *      someOtherKey: 'some other value'
                 * };
                 */
                header: {
                    /**
                     * The header row attribute contains a subset of attributes for row headers.
                     *
                     * @expose
                     * @alias header.row
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @ojshortdesc Specifies attributes on the row headers.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">row</code> attribute specified:</caption>
                     * &lt;!-- Using dot notation -->
                     * &lt;oj-data-grid header.row.some-key='some value'>&lt;/oj-data-grid>
                     *
                     * &lt;!-- Using JSON notation -->
                     * &lt;oj-data-grid header.row='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">row</code> property after initialization:</caption>
                     * // Get one
                     * var value = myDataGrid.header.row.someKey;
                     *
                     * // Set one, leaving the others intact
                     * myDataGrid.setProperty('header.row.someKey', 'someValue');
                     *
                     * // Get all
                     * var values = myDataGrid.header.row;
                     *
                     * // Set all.  Must list every resource key, as those not listed are lost.
                     * myDataGrid.header.row = {
                     *      someKey: 'some value',
                     *      someOtherKey: 'some other value'
                     * };
                     */
                    row: {
                        /**
                         * <p>The CSS style class to apply to row headers in the DataGrid. If a string is specified
                         * the class will be added to all row header cells. If a function is specified it takes a single parameter,
                         * <a href="#context-section">headerContext</a> and must return a string to be set as a className.
                         *
                         * @expose
                         * @alias header.row.className
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc CSS class names to apply to row headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">className</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.row.class-name='myClassName'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">className</code> property after initialization:</caption>
                         * // getter
                         * var classNameValue = myDataGrid.header.row.className;
                         *
                         * // setter
                         * myDataGrid.header.row.className = 'myClassName';
                         */
                        className: null,
                        /**
                         * <p>The renderer function that renders the content of the row header. See <a href="#context-section">headerContext</a>
                         * for information on the object passed into the row header renderer function.
                         * The function should return one of the following: 
                         * <ul>
                         *   <li>An Object with the following property:
                         *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the row header.</li></ul>
                         *   </li>
                         *   <li>undefined: If the developer chooses to append to the row header element directly, the function should return undefined.</li>
                         * </ul>
                         * If no renderer is specified, the DataGrid will treat the header data as a string.
                         * 
                         * @expose
                         * @alias header.row.renderer
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|null}
                         * @default null
                         * @ojshortdesc Specifies what renders is the row headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">renderer</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.row.renderer='{{myRendererFunction}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">renderer</code> property after initialization:</caption>
                         * // getter
                         * var rendererValue = myDataGrid.header.row.renderer;
                         *
                         * // setter
                         * myDataGrid.header.row.renderer = myRendererFunction;
                         */
                        renderer: null,
                        /**
                         * <p>Enable or disable width or height resize along the row headers. Note
                         * that for row header, a function cannot be used with the <code class="prettyprint">width</code> subproperty.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string of
                         * <code class="prettyprint">enable</code> or <code class="prettyprint">disable</code>.
                         *
                         * @expose
                         * @alias header.row.resizable
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {Object.<string, string>|Object.<string, function(Object)>|null}
                         * @default {"width": "disable", "height": "disable"}
                         * @property {string} width row width resizable, valid values are: 'enable', 'disable'
                         * @property {string} height row header height resizable, valid values are: 'enable', 'disable'
                         * @ojshortdesc Specifies resizing width or height on the row headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">resizable</code> attribute specified:</caption>
                         * &lt;!-- Using dot notation -->
                         * &lt;oj-data-grid header.row.resizable.width='enable'>&lt;/oj-data-grid>
                         *
                         * &lt;!-- Using JSON notation -->
                         * &lt;oj-data-grid header='{"row":{"resizable": {"width": "enable"}}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">resizable</code> property after initialization:</caption>
                         * // Get one
                         * var resizableValue = myDataGrid.header.row.resizable.width;
                         *
                         * // Set one, leaving the others intact
                         * myDataGrid.setProperty('header.row.resizable.width', 'enable');
                         *
                         * // Get all
                         * var resizableValues = myDataGrid.header.row.resizable;
                         *
                         * // Set all.  Must list every resource key, as those not listed are lost.
                         * myDataGrid.header = {
                         *    row: {
                         *        resizable: {
                         *            width: 'enable'
                         *         }
                         *    }
                         * };
                         */
                        resizable: {'width': 'disable', 'height': 'disable'},
                        /**
                         * <p>Enable or disable sorting on the field bounded by this header. The
                         * data source associated with the DataGrid must have the sort function defined and the capability supported.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string of
                         * <code class="prettyprint">auto</code>, <code class="prettyprint">enable</code>, or <code class="prettyprint">disable</code>.
                         *
                         * @expose
                         * @alias header.row.sortable
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string}
                         * @default 'auto'
                         * @ojvalue {string} "auto" get the sortable property from the data source
                         * @ojvalue {string} "enable" enable sorting on row headers
                         * @ojvalue {string} "disable" disable sorting on row headers
                         * @ojshortdesc Specifies whether sorting is enabled on the row headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">sortable</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.row.class-name='disable'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">sortable</code> property after initialization:</caption>
                         * // getter
                         * var sortableValue = myDataGrid.header.row.sortable;
                         *
                         * // setter
                         * myDataGrid.header.row.sortable = 'disable';
                         */
                        sortable: 'auto',
                        /**
                         * <p>The inline style to apply to row headers in the DataGrid. If a string is specified
                         * the class will be added to all row header cells. Note that percentage (%) width and height values are not supported.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string.
                         *
                         * @expose
                         * @alias header.row.style
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc Inline style to add to the row headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">style</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.row.style='myStyle'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">style</code> property after initialization:</caption>
                         * // getter
                         * var styleValue = myDataGrid.header.row.style;
                         *
                         * // setter
                         * myDataGrid.header.row.style = 'myStyle';
                         */
                        style: null
                    },
                    /**
                     * The header column attribute contains a subset of attributes for column headers.
                     *
                     * @expose
                     * @alias header.column
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @ojshortdesc Specifies attributes on the column headers.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">column</code> attribute specified:</caption>
                     * &lt;!-- Using dot notation -->
                     * &lt;oj-data-grid header.column.some-key='some value'>&lt;/oj-data-grid>
                     *
                     * &lt;!-- Using JSON notation -->
                     * &lt;oj-data-grid header.column='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">column</code> property after initialization:</caption>
                     * // Get one
                     * var value = myDataGrid.header.column.someKey;
                     *
                     * // Set one, leaving the others intact
                     * myDataGrid.setProperty('header.column.someKey', 'someValue');
                     *
                     * // Get all
                     * var values = myDataGrid.header.column;
                     *
                     * // Set all.  Must list every resource key, as those not listed are lost.
                     * myDataGrid.header.column = {
                     *      someKey: 'some value',
                     *      someOtherKey: 'some other value'
                     * };
                     */
                    column: {
                        /**
                         * <p>The CSS style class to apply to column headers in the DataGrid. If a string is specified
                         * the class will be added to all column header cells. If a function is specified it takes a single parameter,
                         * <a href="#context-section">headerContext</a> and must return a string to be set as a className.
                         *
                         * @expose
                         * @alias header.column.className
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc CSS class names to apply to column headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">className</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.column.class-name='myClassName'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">className</code> property after initialization:</caption>
                         * // getter
                         * var classNameValue = myDataGrid.header.column.className;
                         *
                         * // setter
                         * myDataGrid.header.column.className = 'myClassName';
                         */
                        className: null,
                        /**
                         * <p>The renderer function that renders the content of the column header. See <a href="#context-section">headerContext</a>
                         * for information on the object passed into the column header renderer function.
                         * The function should return one of the following: 
                         * <ul>
                         *   <li>An Object with the following property:
                         *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the column header.</li></ul>
                         *   </li>
                         *   <li>undefined: If the developer chooses to append to the column header element directly, the function should return undefined.</li>
                         * </ul>
                         * If no renderer is specified, the DataGrid will treat the header data as a string.
                         * 
                         * @expose
                         * @alias header.column.renderer
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|null}
                         * @default null
                         * @ojshortdesc Specifies what renders is the column headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">renderer</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.column.renderer='{{myRendererFunction}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">renderer</code> property after initialization:</caption>
                         * // getter
                         * var rendererValue = myDataGrid.header.column.renderer;
                         *
                         * // setter
                         * myDataGrid.header.column.renderer = myRendererFunction;
                         */
                        renderer: null,
                        /**
                         * <p>Enable or disable width or height resize along the column headers. Note
                         * that for column header, a function cannot be used with the <code class="prettyprint">height</code> subproperty.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string of
                         * <code class="prettyprint">enable</code> or <code class="prettyprint">disable</code>.
                         *
                         * @expose
                         * @alias header.column.resizable
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {Object.<string, string>|Object.<string, function(Object)>|null}
                         * @default {"width": "disable", "height": "disable"}
                         * @property {string} width column width resizable, valid values are: 'enable', 'disable'
                         * @property {string} height column header height resizable, valid values are: 'enable', 'disable'
                         * @ojshortdesc Specifies resizing width or height on the column headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">resizable</code> attribute specified:</caption>
                         * &lt;!-- Using dot notation -->
                         * &lt;oj-data-grid header.column.resizable.width='enable'>&lt;/oj-data-grid>
                         *
                         * &lt;!-- Using JSON notation -->
                         * &lt;oj-data-grid header='{"column":{"resizable": {"width": "enable"}}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">resizable</code> property after initialization:</caption>
                         * // Get one
                         * var resizableValue = myDataGrid.header.column.resizable.width;
                         *
                         * // Set one, leaving the others intact
                         * myDataGrid.setProperty('header.column.resizable.width', 'enable');
                         *
                         * // Get all
                         * var resizableValues = myDataGrid.header.column.resizable;
                         *
                         * // Set all.  Must list every resource key, as those not listed are lost.
                         * myDataGrid.header = {
                         *    column: {
                         *        resizable: {
                         *            width: 'enable'
                         *         }
                         *    }
                         * };
                         */
                        resizable: {'width': 'disable', 'height': 'disable'},
                        /**
                         * <p>Enable or disable sorting on the field bounded by this header. The
                         * data source associated with the DataGrid must have the sort function defined and the capability supported.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string of
                         * <code class="prettyprint">auto</code>, <code class="prettyprint">enable</code>, or <code class="prettyprint">disable</code>.
                         *
                         * @expose
                         * @alias header.column.sortable
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string}
                         * @default 'auto'
                         * @ojvalue {string} "auto" get the sortable property from the data source
                         * @ojvalue {string} "enable" enable sorting on column headers
                         * @ojvalue {string} "disable" disable sorting on column headers
                         * @ojshortdesc Specifies whether sorting is enabled on the column headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">sortable</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.column.class-name='disable'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">sortable</code> property after initialization:</caption>
                         * // getter
                         * var sortableValue = myDataGrid.header.column.sortable;
                         *
                         * // setter
                         * myDataGrid.header.column.sortable = 'disable';
                         */
                        sortable: 'auto',
                        /**
                         * <p>The inline style to apply to column headers in the DataGrid. If a string is specified
                         * the class will be added to all column header cells. Note that percentage (%) width and height values are not supported.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string.
                         *
                         * @expose
                         * @alias header.column.style
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc Inline style to add to the column headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">style</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.column.style='myStyle'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">style</code> property after initialization:</caption>
                         * // getter
                         * var styleValue = myDataGrid.header.column.style;
                         *
                         * // setter
                         * myDataGrid.header.column.style = 'myStyle';
                         */
                        style: null
                    },
                    /**
                     * The header columnEnd attribute contains a subset of attributes for column end headers.
                     *
                     * @expose
                     * @alias header.columnEnd
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @ojshortdesc Specifies attributes on the column end headers.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">columnEnd</code> attribute specified:</caption>
                     * &lt;!-- Using dot notation -->
                     * &lt;oj-data-grid header.column-end.some-key='some value'>&lt;/oj-data-grid>
                     *
                     * &lt;!-- Using JSON notation -->
                     * &lt;oj-data-grid header.column-end='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">columnEnd</code> property after initialization:</caption>
                     * // Get one
                     * var value = myDataGrid.header.columnEnd.someKey;
                     *
                     * // Set one, leaving the others intact
                     * myDataGrid.setProperty('header.columnEnd.someKey', 'someValue');
                     *
                     * // Get all
                     * var values = myDataGrid.header.columnEnd;
                     *
                     * // Set all.  Must list every resource key, as those not listed are lost.
                     * myDataGrid.header.columnEnd = {
                     *      someKey: 'some value',
                     *      someOtherKey: 'some other value'
                     * };
                     */
                    columnEnd: {
                        /**
                         * <p>The CSS style class to apply to columnEnd headers in the DataGrid. If a string is specified
                         * the class will be added to all columnEnd header cells. If a function is specified it takes a single parameter,
                         * <a href="#context-section">headerContext</a> and must return a string to be set as a className.
                         *
                         * @expose
                         * @alias header.columnEnd.className
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc CSS class names to apply to column end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">className</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.column-end.class-name='myClassName'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">className</code> property after initialization:</caption>
                         * // getter
                         * var classNameValue = myDataGrid.header.columnEnd.className;
                         *
                         * // setter
                         * myDataGrid.header.columnEnd.className = 'myClassName';
                         */
                        className: null,
                        /**
                         * <p>The renderer function that renders the content of the columnEnd header. See <a href="#context-section">headerContext</a>
                         * for information on the object passed into the columnEnd header renderer function.
                         * The function should return one of the following: 
                         * <ul>
                         *   <li>An Object with the following property:
                         *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the columnEnd header.</li></ul>
                         *   </li>
                         *   <li>undefined: If the developer chooses to append to the columnEnd header element directly, the function should return undefined.</li>
                         * </ul>
                         * If no renderer is specified, the DataGrid will treat the header data as a string.
                         *
                         * @expose
                         * @alias header.columnEnd.renderer
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|null}
                         * @default null
                         * @ojshortdesc Specifies what renders is the column end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">renderer</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.column-end.renderer='{{myRendererFunction}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">renderer</code> property after initialization:</caption>
                         * // getter
                         * var rendererValue = myDataGrid.header.columnEnd.renderer;
                         *
                         * // setter
                         * myDataGrid.header.columnEnd.renderer = myRendererFunction;
                         */
                        renderer: null,
                        /**
                         * <p>Enable or disable width or height resize along the columnEnd headers. Note
                         * that for columnEnd header, a function cannot be used with the <code class="prettyprint">height</code> subproperty.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string of
                         * <code class="prettyprint">enable</code> or <code class="prettyprint">disable</code>.
                         *
                         * @expose
                         * @alias header.columnEnd.resizable
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {Object.<string, string>|Object.<string, function(Object)>|null}
                         * @default {"width": "disable", "height": "disable"}
                         * @property {string} width columnEnd width resizable, valid values are: 'enable', 'disable'
                         * @property {string} height columnEnd header height resizable, valid values are: 'enable', 'disable'
                         * @ojshortdesc Specifies resizing width or height on the column end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">resizable</code> attribute specified:</caption>
                         * &lt;!-- Using dot notation -->
                         * &lt;oj-data-grid header.column-end.resizable.width='enable'>&lt;/oj-data-grid>
                         *
                         * &lt;!-- Using JSON notation -->
                         * &lt;oj-data-grid header='{"column-end":{"resizable": {"width": "enable"}}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">resizable</code> property after initialization:</caption>
                         * // Get one
                         * var resizableValue = myDataGrid.header.columnEnd.resizable.width;
                         *
                         * // Set one, leaving the others intact
                         * myDataGrid.setProperty('header.columnEnd.resizable.width', 'enable');
                         *
                         * // Get all
                         * var resizableValues = myDataGrid.header.columnEnd.resizable;
                         *
                         * // Set all.  Must list every resource key, as those not listed are lost.
                         * myDataGrid.header = {
                         *    columnEnd: {
                         *        resizable: {
                         *            width: 'enable'
                         *         }
                         *    }
                         * };
                         */
                        resizable: {'width': 'disable', 'height': 'disable'},
                        /**
                         * <p>The inline style to apply to columnEnd headers in the DataGrid. If a string is specified
                         * the class will be added to all columnEnd header cells. Note that percentage (%) width and height values are not supported.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string.
                         *
                         * @expose
                         * @alias header.columnEnd.style
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc Inline style to add to the column end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">style</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.column-end.style='myStyle'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">style</code> property after initialization:</caption>
                         * // getter
                         * var styleValue = myDataGrid.header.columnEnd.style;
                         *
                         * // setter
                         * myDataGrid.header.columnEnd.style = 'myStyle';
                         */
                        style: null
                    },
                    /**
                     * The header rowEnd attribute contains a subset of attributes for row end headers.
                     *
                     * @expose
                     * @alias header.rowEnd
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @ojshortdesc Specifies attributes on the row end headers.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">rowEnd</code> attribute specified:</caption>
                     * &lt;!-- Using dot notation -->
                     * &lt;oj-data-grid header.row-end.some-key='some value'>&lt;/oj-data-grid>
                     *
                     * &lt;!-- Using JSON notation -->
                     * &lt;oj-data-grid header.row-end='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">rowEnd</code> property after initialization:</caption>
                     * // Get one
                     * var value = myDataGrid.header.rowEnd.someKey;
                     *
                     * // Set one, leaving the others intact
                     * myDataGrid.setProperty('header.rowEnd.someKey', 'someValue');
                     *
                     * // Get all
                     * var values = myDataGrid.header.rowEnd;
                     *
                     * // Set all.  Must list every resource key, as those not listed are lost.
                     * myDataGrid.header.rowEnd = {
                     *      someKey: 'some value',
                     *      someOtherKey: 'some other value'
                     * };
                     */
                    rowEnd: {
                         /**
                         * <p>The CSS style class to apply to rowEnd headers in the DataGrid. If a string is specified
                         * the class will be added to all rowEnd header cells. If a function is specified it takes a single parameter,
                         * <a href="#context-section">headerContext</a> and must return a string to be set as a className.
                         *
                         * @expose
                         * @alias header.rowEnd.className
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc CSS class names to apply to row end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">className</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.row-end.class-name='myClassName'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">className</code> property after initialization:</caption>
                         * // getter
                         * var classNameValue = myDataGrid.header.rowEnd.className;
                         *
                         * // setter
                         * myDataGrid.header.rowEnd.className = 'myClassName';
                         */
                        className: null,
                        /**
                         * <p>The renderer function that renders the content of the rowEnd header. See <a href="#context-section">headerContext</a>
                         * for information on the object passed into the rowEnd header renderer function.
                         * The function should return one of the following: 
                         * <ul>
                         *   <li>An Object with the following property:
                         *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the rowEnd header.</li></ul>
                         *   </li>
                         *   <li>undefined: If the developer chooses to append to the rowEnd header element directly, the function should return undefined.</li>
                         * </ul>
                         * If no renderer is specified, the DataGrid will treat the header data as a string.
                         *
                         * @expose
                         * @alias header.rowEnd.renderer
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|null}
                         * @default null
                         * @ojshortdesc Specifies what renders is the row end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">renderer</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.row-end.renderer='{{myRendererFunction}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">renderer</code> property after initialization:</caption>
                         * // getter
                         * var rendererValue = myDataGrid.header.rowEnd.renderer;
                         *
                         * // setter
                         * myDataGrid.header.rowEnd.renderer = myRendererFunction;
                         */
                        renderer: null,
                        /**
                         * <p>Enable or disable width or height resize along the rowEnd headers. Note
                         * that for rowEnd header, a function cannot be used with the <code class="prettyprint">width</code> subproperty.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string of
                         * <code class="prettyprint">enable</code> or <code class="prettyprint">disable</code>.
                         *
                         * @expose
                         * @alias header.rowEnd.resizable
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {Object.<string, string>|Object.<string, function(Object)>|null}
                         * @default {"width": "disable", "height": "disable"}
                         * @property {string} width rowEnd width resizable, valid values are: 'enable', 'disable'
                         * @property {string} height rowEnd header height resizable, valid values are: 'enable', 'disable'
                         * @ojshortdesc Specifies resizing width or height on the row end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">resizable</code> attribute specified:</caption>
                         * &lt;!-- Using dot notation -->
                         * &lt;oj-data-grid header.row-end.resizable.width='enable'>&lt;/oj-data-grid>
                         *
                         * &lt;!-- Using JSON notation -->
                         * &lt;oj-data-grid header='{"row-end":{"resizable": {"width": "enable"}}}'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">resizable</code> property after initialization:</caption>
                         * // Get one
                         * var resizableValue = myDataGrid.header.row-end.resizable.width;
                         *
                         * // Set one, leaving the others intact
                         * myDataGrid.setProperty('header.row-end.resizable.width', 'enable');
                         *
                         * // Get all
                         * var resizableValues = myDataGrid.header.rowEnd.resizable;
                         *
                         * // Set all.  Must list every resource key, as those not listed are lost.
                         * myDataGrid.header = {
                         *    rowEnd: {
                         *        resizable: {
                         *            width: 'enable'
                         *         }
                         *    }
                         * };
                         */
                        resizable: {'width': 'disable', 'height': 'disable'},
                        /**
                         * <p>The inline style to apply to rowEnd headers in the DataGrid. If a string is specified
                         * the class will be added to all rowEnd header cells. Note that percentage (%) width and height values are not supported.
                         * If a function is specified it takes a single parameter, <a href="#context-section">headerContext</a> and must return a string.
                         *
                         * @expose
                         * @alias header.rowEnd.style
                         * @memberof! oj.ojDataGrid
                         * @instance
                         * @type {function(Object)|string|null}
                         * @default null
                         * @ojshortdesc Inline style to add to the row end headers.
                         *
                         * @example <caption>Initialize the DataGrid with the <code class="prettyprint">style</code> attribute specified:</caption>
                         * &lt;oj-data-grid header.row-end.style='myStyle'>&lt;/oj-data-grid>
                         *
                         * @example <caption>Get or set the <code class="prettyprint">style</code> property after initialization:</caption>
                         * // getter
                         * var styleValue = myDataGrid.header.rowEnd.style;
                         *
                         * // setter
                         * myDataGrid.header.rowEnd.style = 'myStyle';
                         */
                        style: null
                    }
                },
                /**
                 * The cell attribute contains a subset of attributes for databody cells.
                 *
                 * @expose
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @type {Object}
                 * @ojshortdesc Specifies attributes on the cells.
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">cell</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-data-grid cell.some-key='some value'>&lt;/oj-data-grid>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-data-grid cell='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-data-grid>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">cell</code> property after initialization:</caption>
                 * // Get one
                 * var value = myDataGrid.cell.someKey;
                 *
                 * // Set one, leaving the others intact
                 * myDataGrid.setProperty('cell.someKey', 'someValue');
                 *
                 * // Get all
                 * var values = myDataGrid.cell;
                 *
                 * // Set all.  Must list every resource key, as those not listed are lost.
                 * myDataGrid.cell = {
                 *      someKey: 'some value',
                 *      someOtherKey: 'some other value'
                 * };
                 */
                cell: {
                    /**
                     * <p>The CSS style class to apply to data body in the DataGrid. If a string is specified
                     * the class will be added to all data body cells. If a function is specified it takes a single parameter,
                     * <a href="#context-section">cellContext</a> and must return a string to be set as a className.
                     *
                     * @expose
                     * @alias cell.className
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @type {function(Object)|string|null}
                     * @default null
                     * @ojshortdesc CSS class names to apply to cells.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">className</code> attribute specified:</caption>
                     * &lt;oj-data-grid cell.class-name='myClassName'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">className</code> property after initialization:</caption>
                     * // getter
                     * var classNameValue = myDataGrid.cell.className;
                     *
                     * // setter
                     * myDataGrid.cell.className = 'myClassName';
                     */
                    className: null,
                    /**
                     * <p>The renderer function that renders the content of the data body. See <a href="#context-section">cellContext</a>
                     * for information on the object passed into the cell renderer function.
                     * The function should return one of the following: 
                     * <ul>
                     *   <li>An Object with the following property:
                     *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the data body.</li></ul>
                     *   </li>
                     *   <li>undefined: If the developer chooses to append to the data body element directly, the function should return undefined.</li>
                     * </ul>
                     * If no renderer is specified, the DataGrid will treat the cell data as a string.
                     *
                     * @expose
                     * @alias cell.renderer
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @type {function(Object)|null}
                     * @default null
                     * @ojshortdesc Specifies what renders is the cell.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">renderer</code> attribute specified:</caption>
                     * &lt;oj-data-grid cell.renderer='{{myRendererFunction}}'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">renderer</code> property after initialization:</caption>
                     * // getter
                     * var rendererValue = myDataGrid.cell.renderer;
                     *
                     * // setter
                     * myDataGrid.cell.renderer = myRendererFunction;
                     */
                    renderer: null,
                    /**
                     * <p>The inline style to apply to directly to cells in the data body. If a string is specified
                     * the class will be added to all data body cells.
                     * If a function is specified it takes a single parameter, <a href="#context-section">cellContext</a> and must return a string.
                     *
                     * @expose
                     * @alias cell.style
                     * @memberof! oj.ojDataGrid
                     * @instance
                     * @type {function(Object)|string|null}
                     * @default null
                     * @ojshortdesc Inline style to add to the cells.
                     *
                     * @example <caption>Initialize the DataGrid with the <code class="prettyprint">style</code> attribute specified:</caption>
                     * &lt;oj-data-grid cell.style='myStyle'>&lt;/oj-data-grid>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">style</code> property after initialization:</caption>
                     * // getter
                     * var styleValue = myDataGrid.cell.style;
                     *
                     * // setter
                     * myDataGrid.cell.style = 'myStyle';
                     */
                    style: null
                },
                /**
                 * Triggered when a portion of the DataGrid is resized
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @property {string|number} header the key of the header which was resized
                 * @property {Object} oldDimensions the oldDimensions
                 * @property {number} oldDimensions.width the old pixel size (ex: '75px' would be 75)
                 * @property {number} oldDimensions.height the old pixel size (ex: '75px' would be 75)
                 * @property {Object} newDimensions the newDimensions
                 * @property {number} newDimensions.width the new pixel size (ex: '75px' would be 75)
                 * @property {number} newDimensions.height the new pixel size (ex: '75px' would be 75)
                 */
                resize: null,
                /**
                 * Triggered when a sort is performed on the DataGrid
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @property {string|number} header the key of the header which was sorted on
                 * @property {string} direction the direction of the sort ascending/descending
                 */
                sort: null,

                /**
                 * Triggered after all items in the DataGrid has been rendered.
                 * Note that in the highwatermark or virtual scrolling case this
                 * means all items means the items that are fetched so far.
                 *
                 * @ignore
                 * @expose
                 * @event
                 * @deprecated Use the <a href="#whenReady">whenReady</a> method instead.
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @property {Event} event <code class="prettyprint">jQuery</code> event object
                 * @property {Object} ui Parameters
                 *
                 * @example <caption>Initialize the DataGrid with the <code class="prettyprint">ready</code> callback specified:</caption>
                 * $( ".selector" ).ojDataGrid({
                 *     "ready": function( event, ui ) {}
                 * });
                 *
                 * @example <caption>Bind an event listener to the <code class="prettyprint">ojready</code> event:</caption>
                 * $( ".selector" ).on( "ojready", function( event, ui ) {
                 *      // verify that the component firing the event is a component of interest
                 *      if ($(event.target).is(".mySelector"))
                 * });
                 */
                ready: null,

                /**
                 * Triggered after the DataGrid has been scrolled via the UI or the scrollTo method.
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @property {number} scrollX the x position in pixels of the scrollable region calculated from the origin of the DataGrid. In RTL this would be the right of the grid.
                 * @property {number} scrollY the y position in pixels of the scrollable region
                 */
                scroll: null,

                /**
                 * Triggered before the current cell is changed via the <code class="prettyprint">currentCell</code> attribute or via the UI.
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @property {Object} currentCell the new current cell, see <code class="prettyprint">currentCell</code> for the object information
                 * @property {Object} previousCurrentCell the previous current cell, see <code class="prettyprint">currentCell</code> for the object information
                 */
                beforeCurrentCell: null,

                /**
                 * Triggered before the DataGrid is going to enter edit mode. To prevent editing the cell prevent default on the event.
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @property {Object} cellContext the cellContext of the cell that editing is going to be performed on
                 */
                beforeEdit: null,

                /**
                 * Triggered before the DataGrid is going to exit edit mode. To prevent exit editing the prevent default on the event.
                 * There is a provided beforeEditEnd function, oj.DataCollectionEditUtils.basicHandleEditEnd, which can be specified.
                 * This function will handle canceling edits as well as invoking validation on input elements.
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojDataGrid
                 * @instance
                 * @property {Object} cellContext the cellContext of the cell that editing is going to be performed on
                 * @property {boolean} cancelEdit true if the edit should be negated based on actions (i.e. escape key)
                 */
                beforeEditEnd: null
            },
    /**
     * Create the grid
     * @override
     * @memberof oj.ojDataGrid
     * @protected
     */
    _ComponentCreate: function()
    {
        this._super();
        this.root = this.element[0];
        this.rootId = this.root.getAttribute('id');

        this.grid = new DvtDataGrid(this.root);
        //set the visibility state to render until rendering is completed
        this.grid.setVisibility(DvtDataGrid.VISIBILITY_STATE_RENDER);
        //required classes on init, oj-component-initnode is added by this._super
        $(this.root).addClass("oj-datagrid oj-component");
        $(this.root).attr(oj.Components._OJ_CONTAINER_ATTR, this['widgetName']);
        this.redrawSet = {'data':'all', 'header':['className','renderer','style','template']}; //vvc
    },
    /**
     * Initialize the grid after creation
     * @protected
     * @override
     * @memberof oj.ojDataGrid
     */
    _AfterCreate: function ()
    {
        var self = this;

        // unregister existing resize listener before emptying out the root
        this._unregisterResizeListener(this.root);

        // remove everything except for slots
        this._removeAllChildren();

        this._super();
        this.resources = new oj.DataGridResources(this._GetReadingDirection(), this._getTranslation.bind(self), oj.ThemeUtils.parseJSONFromFontFamily("oj-datagrid-option-defaults"));
        this._setDataSource();
        // sets the initial (or default) selection on internal grid
        this._setSelection();

        if (this.datasource != null)
        {
            this.grid.SetDataSource(this.datasource);
            
            // add context menu before _SetupResources super sets it up
            this._addContextMenu();

            this._isCreate = true;
                        
            // data source listeners will be added in SetupResources, but we need them earlier 
            // for handling expanded option on rowexpander, and any other rendered content
            // that could trigger a datasource event.
            this.grid._addDataSourceEventListeners();
        }
        this._checkSelectionMode(this.options);
        this.grid.SetOptions(this.options, this._WrapCustomElementRenderer.bind(self));
        this.grid.SetResources(this.resources);
        this.grid.SetCreateContextCallback(this._modifyContext.bind(self));
        this.grid.SetFixContextCallback(this._FixRendererContext.bind(self));
        this.grid.SetRemoveCallback(this._remove.bind(self));
        this.grid.SetUniqueIdCallback(this._uniqueId.bind(self));
        this.grid.SetCompareValuesCallback(this._compareValues.bind(self));
        this.grid.SetNotReadyCallback(this._NotReady.bind(self));
        this.grid.SetMakeReadyCallback(this._MakeReady.bind(self));
        this.grid.SetOptionCallback(this.option.bind(self));
        this.grid.SetSubtreeAttachedCallback(oj.Components.subtreeAttached);
        this.grid.SetContextCallback(oj.Context.getContext);

        this._focusable({
            'applyHighlight': true,
            'setupHandlers': function(focusInHandler, focusOutHandler) {
                var noJQFocusInHandler = function(element) {
                    return focusInHandler($(element))
                };

                var noJQFocusOutHandler = function(element) {
                    return focusOutHandler($(element))
                };

                self.grid.SetFocusableCallback.call(self.grid, noJQFocusInHandler, noJQFocusOutHandler);
            }
        });

        this._registerEventListeners();

        //attempt to render the grid if visible and attached
        this._possiblyRenderOrRefresh();
    },
    /**
     * Sets up resources after creation
     * @protected
     * @override
     * @memberof oj.ojDataGrid
     */
    _SetupResources: function ()
    {
        // adds context menu gesture support
        this._super();
        
        if (this.datasource != null)
        {
            // sets the data source event listeners
            if (!this._isCreate)
            {
                this.grid._addDataSourceEventListeners();
            }
            
            // sets the root/document DOM listeners
            this.grid._addDomEventListeners();
            
            // register a resize listener
            this._registerResizeListener(this.root);
        }
        
        this._isCreate = false;
    },
    /**
     * Releases resources after creation
     * @protected
     * @override
     * @memberof oj.ojDataGrid
     */
    _ReleaseResources: function()
    {
        // release context menu listeners
        this._super();

        // remove data source event listeners
        this.grid._removeDataSourceEventListeners();

        // remove root/document DOM listeners
        this.grid._removeDomEventListeners();

        // remove the resize DOM listener
        this._unregisterResizeListener(this.root);
    }, 
    
    _registerEventListeners: function()
    {
        var self = this;

        //listen for resizing, selection, sort and trigger relevant events
        this.grid.addListener('resize', function(details)
        {
            self._trigger('resize', details['event'], details['ui']);
        });
        this.grid.addListener('select', function(details)
        {
            self.option("selection", details['ui']['selection'],
                         {'_context': {originalEvent: details['event'], internalSet: true},
                          'changed': true});
        });
        this.grid.addListener('currentCell', function(details)
        {
            self.option("currentCell", details['ui'],
                         {'_context': {originalEvent: details['event'], internalSet: true},
                          'changed': true});
        });
        this.grid.addListener('beforeCurrentCell', function(details)
        {
            return self._trigger('beforeCurrentCell', details['event'], details['ui']);
        });
        this.grid.addListener('sort', function(details)
        {
            var sortEvent = self._trigger2('sort', details['event'], details['ui']);
            self.grid.SetSortOriginalEvent(sortEvent['event']);
        });
        this.grid.addListener('keydown', function(details)
        {
            return self._trigger('keydown', details['event'], details['ui']);
        });
        this.grid.addListener('ready', function(details)
        {
            self._trigger('ready', null, {});
        });
        this.grid.addListener('scroll', function(details)
        {
            self._trigger('scroll', details['event'], details['ui']);
        });
        this.grid.addListener('beforeEdit', function(details)
        {
            return self._trigger('beforeEdit', details['event'], details['ui']);
        });
        this.grid.addListener('beforeEditEnd', function(details)
        {
            return self._trigger('beforeEditEnd', details['event'], details['ui']);
        });
    },
    /**
     * Redraw the entire data grid after having made some external modifications.
     *
     * @expose
     * @memberof oj.ojDataGrid
     * @instance
     *
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myDataGrid.refresh();
     */
    refresh: function()
    {
        this._super();

        // unregister existing listeners
        this._ReleaseResources();

        this._removeAllChildren();

        // if the context menu is internal we should reset the option so that it
        // gets rebuilt on refresh
        if (this._defaultContextMenu)
        {
            this.options['contextMenu'] = null;
            this._defaultContextMenu = null;
        }
        
        this._setDataSource();
        if (this.datasource != null)
        {
            // if it is flattened datasource, we'll need to reinitialize it
            if (oj.FlattenedTreeDataGridDataSource && this.datasource instanceof oj.FlattenedTreeDataGridDataSource)
            {
                this.datasource.Destroy();
                this.datasource.Init();
            }

            this.grid.SetDataSource(this.datasource);
            
            this._addContextMenu();            
        }
        this.grid.SetOptions(this.options, this._WrapCustomElementRenderer.bind(this));
        this.grid.SetResources(this.resources);
        
        // addListeners back in before rendering, including context menu
        this._SetupResources();
        
        // so long as the visibility property is not 'render', overwrite it with
        // a refresh and try to refresh the grid
        if (this.grid.getVisibility() != DvtDataGrid.VISIBILITY_STATE_RENDER)
        {
            this.grid.setVisibility(DvtDataGrid.VISIBILITY_STATE_REFRESH);
        }
        this._possiblyRenderOrRefresh();
    },
    /**
     * Destroy the grid
     * @memberof oj.ojDataGrid
     * @private
     */
    _destroy: function()
    {
        $(this.root).removeAttr(oj.Components._OJ_CONTAINER_ATTR);

        // destroy the datasource if neccessary (FlattenedTreeDataSource)
        if (this.datasource != null && this.datasource.Destroy)
        {
            this.datasource.Destroy();
        }
        this.grid.destroy();
        this._unregisterResizeListener(this.root);
        $(this.root).empty();
        this._MakeReady();
    },


    /**
     * Sets multiple options
     * @param {Object} options the options object
     * @param {Object} flags additional flags for option
     * @override
     * @private
     */
    _setOptions: function( options, flags )
    {
	var handleUpdate;

        if(!this.datasource)
        {
            // not initialized yet, just call super
            this._super(options, flags);

            // if datasource is one of the options specified, then re-render the grid
            if (options['data'] != null)
            {
                this.refresh();
            }
        }
        else
        {
            // check whether a full refresh from the wrapper layer is needed
            handleUpdate = this._handleOptionUpdate(options, flags);
            // update options
            this._super(options, flags);

            if (handleUpdate == 'refresh')
            {
                //redraw whole grid if required
                this.refresh();
            }
            else if (handleUpdate == 'pass')
            {
                // or process updated option(s) through the DvtDataGrid
                this.grid.UpdateOptions(options, flags);
            }
        }
    },

    _checkSelectionMode: function(options)
    {
        var selectionMode = options['selectionMode'];
        if (selectionMode && 
            selectionMode['row'] && 
            selectionMode['cell'] &&
            selectionMode['row'] != 'none' && 
            selectionMode['cell'] != 'none')
        {
            throw new Error("JET DataGrid does not support selectionMode on cell and row simultaneously.");
        }
    },
    
    /**
     * <p>Notifies the component that its subtree has been made visible programmatically after the component has
     * been created.
     *
     * @memberof oj.DataGrid
     * @instance
     * @protected
     */
    _NotifyShown: function()
    {
        this._super();
        // if we are notified the grid is now shown attempt to render or refresh
        this._possiblyRenderOrRefresh();
    },

    /**
     * <p>Notifies the component that its subtree has been made hidden programmatically after the component has
     * been created.
     *
     * @memberof oj.DataGrid
     * @instance
     * @protected
     */
    _NotifyHidden: function()
    {
        this._super();
        if (this.grid.getVisibility() === DvtDataGrid.VISIBILITY_STATE_VISIBLE)
        {
            this.grid.setVisibility(DvtDataGrid.VISIBILITY_STATE_HIDDEN);
        }
    },

    /**
     * <p>Notifies the component that its subtree has been connected to the document programmatically after the component has
     * been created.
     *
     * @memberof oj.DataGrid
     * @instance
     * @protected
     */
    _NotifyAttached: function()
    {
        this._super();
        // if we are notified the grid is now attached attempt to render or refresh
        this._possiblyRenderOrRefresh();
    },

    /**
     * <p>Notifies the component that its subtree has been removed from the document programmatically after the component has
     * been created.
     *
     * @memberof oj.DataGrid
     * @instance
     * @protected
     */
    _NotifyDetached: function()
    {
        this._super();
        if (this.grid.getVisibility() === DvtDataGrid.VISIBILITY_STATE_VISIBLE)
        {
            this.grid.setVisibility(DvtDataGrid.VISIBILITY_STATE_HIDDEN);
        }
    },

    /**
     * Determine if the entire datagrid should refresh based on which options are updated.
     * @param {Object} options the options object
     * @param {Object|null} flags may contain subkey option
     * @return {string} 'refresh'/'pass'/'ignore'
     * @private
     */
    _handleOptionUpdate: function(options, flags)
    {
        var key, returnVal = 'ignore', subKey, subKeyArray;
        for (key in options)
        {
            switch (key)
            {
                // pass to underlying grid to handle
                case "bandingInterval":
                case "currentCell":
                case "editMode":
                case "gridlines":
                case "scrollPosition":
                case "selection":
                    // just set returnVal in case another option is refresh at
                    // the wrapper level
                    returnVal = 'pass';
                    break;

                // do nothing for event listeners
                case "beforeCurrentCell":
                case "beforeEdit":
                case "beforeEditEnd":
                case "create":
                case "disabled":
                case "dnd":
                case "optionChange":
                case "ready":
                case "resize":
                case "rootAttributes":
                case "scroll":
                case "sort":
                    break;
                    
                // check if error for having two selections
                // set to pass to clear selection on update
                case "selectionMode":
                    this._checkSelectionMode(options);
                    returnVal = 'pass'
                    break;

                // refresh at wrapper level
                case "cell":
                case "header":
                    // we will only not do a full refresh if it's just updating one of the following options specified using the
                    // dot notation like $('#datagrid').ojDataGrid('option', 'header.column.resizable', {'height': 'enable'}); if they
                    // use something like $('#datagrid').ojDataGrid('option', 'header', {'column': {'resizable': {'height': 'enable'}}});
                    // we are going to do a hard refresh because tracing down the changes across axis isn't trivial
                    subKey = flags['subkey'];
                    if (subKey != null)
                    {
                        subKeyArray = subKey.split('.');
                        if (subKeyArray[1] == 'resizable' || subKeyArray[1] == 'sortable')
                        {
                            returnVal = 'pass';
                            break;
                        }
                    }
                case "contextMenu":
                case "data":
                case "scrollPolicy":
                default:
                    return 'refresh';
                    break;
            }
        }

        return returnVal;
    },

    /**
     * Checks if resize is enabled along a given axis width/height
     * @private
     * @param {string} axis column/row/columnEnd/rowEnd
     * @param {string} direction width/height
     * @return {boolean} true if resize is not set to 'disable'
     */
    _isResizeEnabled: function(axis, direction)
    {
        if (this.options['header'][axis] && this.options['header'][axis]['resizable'])
        {
            return this.options['header'][axis]['resizable'][direction] !== 'disable';
        }
        return false;
    },

    /**
     * Checks if sorting is enabled along a given axis
     * @private
     * @param {string} axis column/row/columnEnd/rowEnd
     * @return {boolean} true if sorting is not set to 'disable'
     */
    _isSortEnabled: function(axis)
    {
        if (this.options['header'][axis])
        {
            return this.options['header'][axis]['sortable'] !== 'disable';
        }
        return false;
    },

    /**
     * Remove all children except context menu slot
     * @private
     */
    _removeAllChildren: function()
    {
        $(this.root).children("[slot!='contextMenu']").remove();
    },

    //** @inheritdoc */
    _GetDefaultContextMenu: function()
    {
        return this._defaultContextMenu;
    },

    /**
     * Add a default context menu to the grid if there is none. If there is
     * a context menu set on the grid options we use that one. Add listeners
     * for context menu before show and select.
     * @private
     */
    _addContextMenu: function()
    {
        var self, menuContainer, rootId, resizeMenu = null, sortMenu = null, selectMenu = null,
                moveMenu = null, listItems, sortCapability, menuItemsSetByGrid, isCustomElement;
        self = this;
        isCustomElement = self._IsCustomElement();
        menuContainer = this._GetContextMenu();
        if (!menuContainer)
        {
            if (this.datasource != null)
            {
                menuContainer = this._createDefaultContextMenu();
                if (menuContainer)
                {
                    this._defaultContextMenu = menuContainer[0];
                }
            }
        }
        else
        {
            // this keeps track of which menu items were generated by the data grid dynamically
            // this way on a refresh we know to recreate them in case there was a locale or
            // translations change
            if (this._menuItemsSetByGrid == null)
            {
                this._menuItemsSetByGrid = [];
            }

            menuContainer = $(menuContainer);
            listItems = menuContainer.find('[data-oj-command]');
            menuItemsSetByGrid = [];
            listItems.each(function() {
                var command, newListItem, newLabel, returnVal;
                returnVal = self._shouldUpdateListItem(this, isCustomElement);
                
                if (returnVal.shouldUpdate)
                {
                    command = $(this).attr('data-oj-command').split("-");
                    if (isCustomElement)
                    {
                        // we do not support nested menus yet so just get the label
                        newLabel = self._buildContextMenuLabel(command[command.length - 1], isCustomElement)[0];

                        // replace a label we previously provided with a new one
                        if (returnVal.itemToUpdate)
                        {
                            returnVal.itemToUpdate.textContent = newLabel.textContent;
                            menuItemsSetByGrid.push(returnVal.itemToUpdate);
                        }
                        // the menu has been upgraded and contains a icon or something that is not in the default slot
                        else if (returnVal.toFirstChild)
                        {
                            this.firstChild.appendChild(newLabel);
                            menuItemsSetByGrid.push(newLabel);
                        }
                        // the menu has not been upgraded and has no default slot so add our label
                        else
                        {
                            this.appendChild(newLabel);
                            menuItemsSetByGrid.push(newLabel); 
                        }
                    }
                    else
                    {
                        newListItem = self._buildContextMenuItem(command[command.length - 1], isCustomElement);
                        $(this).replaceWith(newListItem); //@HTMLUpdateOK
                        menuItemsSetByGrid.push(newListItem.children('a').get(0));
                    }
                }
            });
            this._menuItemsSetByGrid = menuItemsSetByGrid;
            if (menuContainer.data('oj-ojMenu'))
            {
                if (isCustomElement)
                {
                    menuContainer[0].refresh();
                }
                else
                {
                    menuContainer.ojMenu('refresh');
                }
            }
        }

        if (menuContainer)
        {
            this._addMenuSelectListener(menuContainer);
        }
    },
    
    /**
     * Should the list item be updated
     * In widget case this is if there is no anchor or we modified it previously, and we will just blow it out with a new one
     * In Custom element we will replace the text that we previously provided or provide text if there was none previously
     * @param {Element} item the oj-option or li that has a data-oj-command on it
     * @param {boolean} isCustomElement true if custom element world
     * @private
     */
    _shouldUpdateListItem: function(item, isCustomElement)
    {    
        var anchor, slots, appendToFirstChild = false;
        if (isCustomElement)
        {
            // if the menu is upgraded we need the slots of the anchor underneath
            // ojmenu currently has a bug where they do not upgrade if there is no text and just an icon
            // need a way to know if the upgrade has happened was hoping oj-complete would be on the oj-option but it wasn't
            if (item.firstElementChild != null && item.firstElementChild.getAttribute('ojmenu') === 'opt')
            {
                slots = oj.BaseCustomElementBridge.getSlotMap(item.firstElementChild);
                appendToFirstChild = true;
            }
            else
            {
                slots = oj.BaseCustomElementBridge.getSlotMap(item);
            }
            // either there are no default slots or default slot length is zero so we should add
            if (!slots[""] || slots[""].length === 0)
            {
                return {shouldUpdate: true, toFirstChild: appendToFirstChild};
            }
            // or there are default slots and we provided something
            else if (slots[""])
            {
                var itemIndex = this._menuItemsSetByGrid.indexOf(slots[""][0]);
                if (itemIndex != -1)
                {
                    return {shouldUpdate: true, itemToUpdate: this._menuItemsSetByGrid[itemIndex]}
                }
            }
            return {shouldUpdate: false};
        }
        anchor = $(item).children('a');
        return {shouldUpdate: (anchor.length === 0 || this._menuItemsSetByGrid.indexOf(anchor.get(0)) != -1)};
    },    
    
    /**
     * @param {jQuery} menu
     * @private
     */
    _addMenuSelectListener: function(menu)
    {
        var listener = this._handleContextMenuSelect.bind(this);
        if (this._IsCustomElement() && !this._defaultContextMenu)
        {
            menu[0].addEventListener("ojAction", listener);
        }
        else
        {
            menu.on("ojselect", listener);
        }
    },

    /**
     * Builds the data grid context menu
     * @private
     */    
    _createDefaultContextMenu: function()
    {
        var self = this, menuContainer, rootId, resizeMenu = null, sortMenu = null, selectMenu = null,
                moveMenu = null, listItems, sortCapability, menuItemsSetByGrid;

        menuContainer = $('<ul>');
        menuContainer.css('display', 'none').attr('id', this.rootId + 'contextmenu');
        $(this.root).append(menuContainer); //@HTMLUpdateOK
        if (this._isResizeEnabled('column', 'width') || this._isResizeEnabled('column', 'height') ||
                this._isResizeEnabled('row', 'width') || this._isResizeEnabled('row', 'height'))
        {
            resizeMenu = this._buildContextMenuItem('resize');
        }

        sortCapability = this.datasource.getCapability('sort');
        if (this._isSortEnabled('column'))
        {
            if (sortCapability === 'column' || sortCapability === 'full')
            {
                sortMenu = this._buildContextMenuItem('sortCol');
            }
        }
        if (this._isSortEnabled('row'))
        {
            if (sortCapability === 'row' || sortCapability === 'full')
            {
                if (sortMenu != null)
                {
                    sortMenu = sortMenu.add(this._buildContextMenuItem('sortRow'));
                }
                else
                {
                    sortMenu = this._buildContextMenuItem('sortRow');
                }
            }
        }

        if (this.options['dnd']['reorder']['row'] === 'enable')
        {
            switch (this.datasource.getCapability('move'))
            {
                case 'none':
                    break;
                default:
                    moveMenu = this._buildContextMenuListItem('cut').add(this._buildContextMenuListItem('paste'));
            }
        }

        // add the discontiguous selection menu if multiple selection and touch device
        if (this._isMultipleSelection() && oj.DomUtils.isTouchSupported())
        {
            this._discontiguousSelection = false;
            selectMenu = this._buildContextMenuListItem('discontiguousSelection');
        }

        if (resizeMenu != null || sortMenu != null || moveMenu != null || selectMenu != null)
        {
            menuContainer.append(resizeMenu).append(sortMenu).append(moveMenu).append(selectMenu); //@HTMLUpdateOK
            menuContainer.ojMenu();

            // keep track of the fact that we're using our own context menu
            // this way on refresh we know that we should remove the contextMenu option
            // so that this block is executed since we are removing the old menu by emptying
            // see this._defaultContextMenu
            this.option("contextMenu", menuContainer.get(0), {'_context': {internalSet: true}});

            return menuContainer;
        }

        return null;
    },

    /**
     * Builds a menu for a command, takes care of submenus where appropriate
     * @param {string} command the command that the datagrid should build a menu item for
     * @param {boolean=} useOjOption whether or not oj-option tag should be use which is the case when the menu is a custom element
     * @private
     */
    _buildContextMenuItem: function(command, useOjOption)
    {
        if (command === 'resize')
        {
            return this._buildContextMenuListItem('resize', useOjOption).append($('<ul></ul>').append(this._buildContextMenuListItem('resizeWidth', useOjOption)).append(this._buildContextMenuListItem('resizeHeight', useOjOption))); //@HTMLUpdateOK
        }
        else if(command === 'sortCol')
        {
            return this._buildContextMenuListItem('sortCol', useOjOption).append($('<ul></ul>').append(this._buildContextMenuListItem('sortColAsc', useOjOption)).append(this._buildContextMenuListItem('sortColDsc', useOjOption))); //@HTMLUpdateOK
        }
        else if(command === 'sortRow')
        {
            return this._buildContextMenuListItem('sortRow', useOjOption).append($('<ul></ul>').append(this._buildContextMenuListItem('sortRowAsc', useOjOption)).append(this._buildContextMenuListItem('sortRowDsc', useOjOption))); //@HTMLUpdateOK
        }
        else if (Object.keys(this.resources.commands).indexOf(command) != -1)
        {
            return $(this._buildContextMenuListItem(command, useOjOption));
        }
    },

    /**
     * Builds a context menu list item from a command
     * @param {string} command the string to look up command value for as well as translation
     * @param {boolean=} useOjOption whether or not oj-option tag should be use which is the case when the menu is a custom element
     * @return {Object} a jQuery object with HTML containing a list item
     * @private
     */
    _buildContextMenuListItem: function(command, useOjOption)
    {
        var listItem = useOjOption ? $('<oj-option></oj-option>') : $('<li></li>');
        listItem.attr('data-oj-command', this._getMappedCommand(command));
        listItem.append(this._buildContextMenuLabel(command, useOjOption)); //@HTMLUpdateOK
        return listItem;
    },
    /**
     * Builds a context menu label by looking up command translation
     * @param {string} command the string to look up translation for
     * @param {boolean=} useOjOption whether or not oj-option tag should be use which is the case when the menu is a custom element
     * @return {jQuery|string} a jQuery object with HTML containing a label
     * @private
     */
    _buildContextMenuLabel: function(command, useOjOption)
    {
        // convert to the translation key convention
        var key = 'label' + command.charAt(0).toUpperCase() + command.slice(1);
        if (command === 'discontiguousSelection')
        {
            // always initialize to enable
            key = 'labelEnableNonContiguous';
        }
        var text = this._getTranslation(key);
        return useOjOption ? $('<span>').text(text) : $('<a href="#"></a>').text(text);
    },

    /**
     * Get the context menu from the grid
     * @return {Array.<Element>|Element} the context menu element that is set in the options
     * @private
     */
    _getDataGridContextMenu: function() // named to avoid overriding baseComponent._getContextMenu()
    {
        var contextMenu = this._GetContextMenu();
        return contextMenu ? contextMenu : this._GetDefaultContextMenu();
    },
    /**
     * Get a translation from the translation resources or one the user set
     * @param {string} key the key of the translation to look up
     * @param {Array|Object|null} args the arguments to pass into the translated string
     * @return {string} the string returned from the resources
     * @private
     */
    _getTranslation: function(key, args)
    {
        return this.getTranslatedString(key, args);
    },
    /**
     * Callback from the resize dialog box, which sends the results to the grid
     * @param {Event} event the event that triggered the dialog button press
     * @private
     */
    _handleResizeDialog: function(event)
    {
        var value = $('#' + this.rootId + 'spinner').ojInputNumber("option", "value");
        $('#' + this.rootId + 'dialog').ojDialog("close");
        this.grid.handleContextMenuReturn(this.contextMenuEvent, this.menuItemFunction, value);
        this.contextMenuEvent['target'].focus();
    },
    /**
     * Build the html for the resize dialog and add it to the root node
     * @param {string} title the header title for the dialog
     * @param {number} initialSize the initial size to put in the spinner
     * @private
     */
    _buildResizeDialog: function(title, initialSize)
    {
        var dialog, dialogBody, spinner, dialogFooter, dialogOKButton;
        //create the base dialog
        dialog =  $('#' + this.rootId + 'dialog');
        spinner = $('#' + this.rootId + 'spinner');
        if (dialog.length === 0 || spinner.length === 0)
        {
            dialog = $('<div>');
            dialog.attr('id', this.rootId + 'dialog');
            dialogBody = $('<div class="oj-dialog-body"></div>');
            dialogFooter = $('<div class="oj-dialog-footer"></div>');
            dialog.append(dialogBody).append(dialogFooter); //@HTMLUpdateOK

            //create the dialog content
            spinner = $('<input id="' + this.rootId + 'spinner"/>');
            dialogOKButton = $('<button id="' + this.rootId + 'dialogsubmit"/>');

            dialogBody.append(spinner); //@HTMLUpdateOK
            dialogFooter.append(dialogOKButton); //@HTMLUpdateOK
            $(this.root).append(dialog); //@HTMLUpdateOK

            dialogOKButton.ojButton({component: 'ojButton', label: this._getTranslation('labelResizeDialogSubmit')});
            dialogOKButton.on('click', this._handleResizeDialog.bind(this));
            spinner.ojInputNumber({component: 'ojInputNumber', max:1000, min:20, step:1, value:initialSize});
            dialog.ojDialog({initialVisibility:'show', title: title, position:{my: "center center", at: "center center", collision:"none", of:$(this.root)}});
        }
        else
        {
            dialog = $('#' + this.rootId + 'dialog');
            spinner.ojInputNumber('option', 'value', initialSize);
            dialog.ojDialog('option', 'title', title);
            dialog.ojDialog("open");
        }
    },
    /**
     * Handle an ojselect event on a menu item, if sort call the handler on the core.
     * If resize prompt the user with a dialog box
     * @param {Event|CustomEvent} event event triggering context menu
     * @param {Object=} ui an object containing the menu item that was selected
     * @private
     */
    _handleContextMenuSelect: function(event, ui)
    {
        var item, initialSize, parent;

        if (ui)
        {
            item = ui.item;
        }
        else
        {
            item = $(event.target);
        }

        this.menuItemFunction = item.attr('data-oj-command');
        if (this.menuItemFunction === this._getMappedCommand('sortColAsc') || this.menuItemFunction === this._getMappedCommand('sortColDsc')
            || this.menuItemFunction === this._getMappedCommand('cut') || this.menuItemFunction === this._getMappedCommand('paste'))
        {
            this.grid.handleContextMenuReturn(this.contextMenuEvent, this.menuItemFunction, null);
            //this.contextMenuEvent['target'].focus();
        }
        else if (this.menuItemFunction === this._getMappedCommand('resizeWidth') || this.menuItemFunction === this._getMappedCommand('resizeHeight'))
        {
            parent = $(this.contextMenuEvent['target']).closest('.' + this._getMappedStyle('cell'));
            if (parent.length == 0)
            {
                parent = $(this.contextMenuEvent['target']).closest('.' + this._getMappedStyle('headercell'));
            }
            if (parent.length == 0)
            {
                parent = $(this.contextMenuEvent['target']).closest('.' + this._getMappedStyle('endheadercell'));
            }
            if (parent.length > 0)
            {
                initialSize = this.menuItemFunction === this._getMappedCommand('resizeWidth') ? parent.outerWidth() : parent.outerHeight();
                this._buildResizeDialog(ui.item.text(), initialSize);
            }
        }
        else if (this.menuItemFunction === this._getMappedCommand('discontiguousSelection'))
        {
            this._discontiguousSelection = !this._discontiguousSelection;
            this.grid.handleContextMenuReturn(this.contextMenuEvent, this.menuItemFunction, this._discontiguousSelection);

            // toggle discontiguous context menu label text
            var key = this._discontiguousSelection ? 'labelDisableNonContiguous' : 'labelEnableNonContiguous';
            item.children().first().text(this._getTranslation(key)); //@HTMLUpdateOK
        }
    },

   /**
    * @param {Object} menu The JET Menu to open as a context menu
    * @param {Event} event What triggered the menu launch
    * @param {string} eventType "mouse", "touch", "keyboard"
    * @private
    */
    _NotifyContextMenuGesture: function(menu, event, eventType)
    {
        this.grid.handleContextMenuGesture(event, eventType, this._contextMenuGestureCallback.bind(this));
    },

   /**
    * Callback for NotifyContextMenuGesture
    * @param {Object} returnVal Object containing capabilities and launcher
    * @param {Event} event What triggered the menu launch
    * @param {string} eventType "mouse", "touch", "keyboard"
    * @private
    */
    _contextMenuGestureCallback: function(returnVal, event, eventType)
    {
        var launcher, capabilities, openOptions;

        if (returnVal == null)
        {
            return;
        }

        this.contextMenuEvent = event['originalEvent'];
        launcher = returnVal['launcher'];
        capabilities = returnVal['capabilities'];
        this._manageContextMenu(capabilities);

        //setting position relative to the cell/header in the SHIFT+F10 case
        //set here to avoid conflicting with user override in before open event
        openOptions = (eventType === "keyboard")
                ? {"position": {"of": launcher}, "launcher": $(launcher)}
        : {"launcher": $(launcher)};

        this._OpenContextMenu(event, eventType, openOptions);
    },
    /**
     * Add the disabled class to the menu item with a given command
     * @param {string} command the command to add the disabled attribute to
     * @private
     */
    _addContextMenuCapability: function(command)
    {
        var contextMenu;
        contextMenu = $(this._getDataGridContextMenu());
        if (!contextMenu.find("[data-oj-command=" + command + "]").hasClass('oj-disabled'))
        {
            contextMenu.find("[data-oj-command=" + command + "]").addClass('oj-disabled');
        }
    },
    /**
     * Remove the disabled class to the menu item with a given command
     * @param {string} command the command to remove the disabled attribute to
     * @private
     */
    _removeContextMenuCapability: function(command)
    {
        $(this._getDataGridContextMenu()).find("[data-oj-command=" + command + "]").removeClass('oj-disabled');
    },
    /**
     * Based on an object containing the capabilities, add or remove the disable attribute
     * @param {Object} capabilities an object with keys of resizable, sortable
     * @private
     */
    _manageContextMenu: function(capabilities)
    {
        var property, command;
        for (property in capabilities)
        {
            if (capabilities.hasOwnProperty(property))
            {
                command = this.resources.getMappedCommand(property);
                if (capabilities[property] === 'disable')
                {
                    this._addContextMenuCapability(command);
                }
                else
                {
                    this._removeContextMenuCapability(command);
                }
            }
        }
    },

    /**
     * Find the index of a cell
     * @param {Object} element the cell to find the index of
     * @return {Object} an object containing row and column
     * @private
     */
    _findCellIndex: function(element)
    {
        return element[this._getMappedAttribute('context')]['indexes'];
    },

    /**
     * Get the root grid as a jquery object
     * @private
     */
    _getGrid: function()
    {
        return $(this.root);
    },

    /**
     * Get the column header container as a jquery object
     * @private
     */
    _getColumnHeader: function()
    {
        return $('#' + this.rootId + '\\:columnHeader');
    },

    /**
     * Get the row header container as a jquery object
     * @private
     */
    _getRowHeader: function()
    {
        return $('#' + this.rootId + '\\:rowHeader');
    },

    /**
     * @private
     */
    _setDataSource: function()
    {
        if (this.options['data'] != null)
        {
            this.datasource = this.options['data'];
        }
        else
        {
            this.datasource = null;
        }
    },

    /**
     * Sets selection on internal grid from options
     * @private
     */
    _setSelection: function()
    {
        var selection = this.options['selection'];
        if (selection != null)
        {
            this.grid.SetSelection(selection);
        }
    },

    /**
     * Modify the header and cell context before passing to the renderer.
     * @param {Object} context the header or cell context.
     * @private
     */
    _modifyContext: function(context)
    {
        context['component'] = oj.Components.__GetWidgetConstructor(this.element, 'ojDataGrid');
    },

    /**
     * Sets accessible context information about the current active cell.
     * Invoked by row expander to set accessible context info on the datagrid (and
     * the info is then read by the screen reader)
     * @param {Object} context
     * @private
     */
    _setAccessibleContext: function(context)
    {
        this.grid.SetAccessibleContext(context);
    },

    /**
     * Unregister event listeners for resize the container DOM element.
     * @param {Element} element  DOM element
     * @private
     */
    _unregisterResizeListener: function(element)
    {
        if (element && this._resizeHandler)
        {
            // remove existing listener
            oj.DomUtils.removeResizeListener(element, this._resizeHandler);
        }
    },

    /**
     * Register event listeners for resize the container DOM element.
     * @param {Element} element  DOM element
     * @private
     */
    _registerResizeListener: function(element)
    {
        if (element)
        {
            if (this._resizeHandler == null)
            {
                this._resizeHandler = this._handleResize.bind(this);
            }

            oj.DomUtils.addResizeListener(element, this._resizeHandler);
        }
    },

    /**
     * The resize handler.
     * @param {number} width the new width
     * @param {number} height the new height
     * @private
     */
    _handleResize: function(width, height)
    {
        if (width > 0 && height > 0)
        {
            //if we get a resize event make sure there aren't pending refresh or render calls
            this._possiblyRenderOrRefresh();
            this.grid.HandleResize(width, height);
        }
    },

    // @inheritdoc
    getNodeBySubId: function(locator)
    {
        var subId, header, rowIndex,columnIndex, index, axis, level, returnElement;
        if (locator == null)
        {
          return this.element ? this.element[0] : null;
        }

        subId = locator['subId'];
        if (subId === 'oj-datagrid-cell')
        {
            rowIndex = locator['rowIndex'];
            columnIndex = locator['columnIndex'];
            returnElement = $(this.grid._getCellByIndex({'row':rowIndex, 'column':columnIndex}));
        }
        else if (subId === 'oj-datagrid-sort-icon' || subId === 'oj-datagrid-sort-ascending' || subId === 'oj-datagrid-sort-descending' || subId === 'oj-datagrid-header')
        {
            axis = locator['axis'];
            index = locator['index'];
            level = locator['level'] == null ? 0 : locator['level'];
            if (axis === 'column')
            {
                header = this._getHeaderByIndex(index, level, $('#' + this.rootId + '\\:columnHeader'), this.grid.getStartColumnHeader());
            }
            else if (axis === 'row')
            {
                header = this._getHeaderByIndex(index, level, $('#' + this.rootId + '\\:rowHeader'), this.grid.getStartRowHeader());
            }
            else if (axis === 'columnEnd')
            {
                header = this._getHeaderByIndex(index, level, $('#' + this.rootId + '\\:columnEndHeader'), this.grid.getStartColumnEndHeader());
            }
            else if (axis === 'rowEnd')
            {
                header = this._getHeaderByIndex(index, level, $('#' + this.rootId + '\\:rowEndHeader'), this.grid.getStartRowEndHeader());
            }
            if (header == null)
            {
                return null;
            }

            // deprecated in 1.2 look to remove in the future
            if (subId === 'oj-datagrid-sort-icon')
            {
                return header.children('.' + this._getMappedStyle('sortcontainer')).children().get(0);
            }
            else if (subId === 'oj-datagrid-sort-ascending')
            {
                returnElement = header.find('.' + this._getMappedStyle('sortascending'));
            }
            else if (subId === 'oj-datagrid-sort-descending')
            {
                returnElement = header.find('.' + this._getMappedStyle('sortdescending'));
            }
            else
            {
                returnElement = header;
            }
        }

        if (returnElement != null && returnElement.length > 0)
        {
            return returnElement[0];
        }
       // Non-null locators have to be handled by the component subclasses
        return null;
    },

    // @inheritdoc
    getSubIdByNode: function(node)
    {
        var cell, context, header, subId, indexes;
        cell = $(node).closest('.' + this._getMappedStyle('cell'));
        if (cell.length > 0)
        {
            indexes = this._findCellIndex(cell[0]);
            return {
                'subId': 'oj-datagrid-cell',
                'rowIndex': indexes['row'],
                'columnIndex': indexes['column']
            };
        }

        header = $(node).closest('.' + this._getMappedStyle('headercell'));
        if (header.length == 0)
        {
            header = $(node).closest('.' + this._getMappedStyle('endheadercell'));
        }

        if (header.length > 0)
        {
            context = header[0][this._getMappedAttribute('context')];
            if ($(node).hasClass(this._getMappedStyle('sortascending')))
            {
                subId = 'oj-datagrid-sort-ascending';
            }
            else if ($(node).hasClass(this._getMappedStyle('sortdescending')))
            {
                subId = 'oj-datagrid-sort-descending';
            }
            else
            {
                subId = 'oj-datagrid-header';
            }
            return {
                'subId': subId,
                'axis': context['axis'],
                'index': this._getHeaderIndex(header),
                'level': context['level']
            };
        }

        return null;
    },

    /**
     * <p>Returns an object with context for the given child DOM node.
     * <p>This will always contain the subid for the node, defined as the 'subId' property on the context object. Additional element specific information may also be included.
     * <p>For more details on returned objects, see <a href="#contextobjects-section">context objects</a>.
     *
     * @param {!Element} node the child DOM node
     * @returns {Object|null} the context for the DOM node, or null when none is found.
     *
     * @example <caption>Invoke the <code class="prettyprint">getContextByNode</code> method:</caption>
     * myDataGrid.getContextByNode(nodeInsideElement);
     *
     * @expose
     * @instance
     * @memberof oj.ojDataGrid
     */
    getContextByNode: function(node)
    {
        var cell, header, context, returnObj;

        cell = $(node).closest('.' + this._getMappedStyle('cell'));
        if (cell.length > 0)
        {
            context = cell[0][this._getMappedAttribute('context')];
            returnObj = oj.CollectionUtils.copyInto({}, context, undefined, true);
            returnObj['subId'] = 'oj-datagrid-cell';
            return returnObj;
        }

        header = $(node).closest('.' + this._getMappedStyle('headercell'));
        if (header.length == 0)
        {
            header = $(node).closest('.' + this._getMappedStyle('endheadercell'));
        }

        if (header.length > 0)
        {
            context = header[0][this._getMappedAttribute('context')];
            returnObj = oj.CollectionUtils.copyInto({}, context, undefined, true);
            returnObj['subId'] = 'oj-datagrid-header';
            // index not updated on header context yet for CRUD operations
            returnObj['index'] = this._getHeaderIndex(header);
            return returnObj;
        }
        
        return null;
    },

    /**
     * Get the mapped style from the resources
     * @param {string} key style mapping key
     * @private
     */
    _getMappedStyle: function(key)
    {
        return this.resources.getMappedStyle(key);
    },

    /**
     * Get the mapped attribute from the resources
     * @param {string} key attribute mapping key
     * @private
     */
    _getMappedAttribute: function(key)
    {
        return this.resources.getMappedAttribute(key);
    },

    /**
     * Get the mapped command from the resources
     * @param {string} key command mapping key
     * @private
     */
    _getMappedCommand: function(key)
    {
        return this.resources.getMappedCommand(key);
    },

    /**
     * Checks if sizing is available for the grid
     * @return {boolean} true if the root element is visible and attached to the DOM
     * @private
     */
    _isDataGridSizingAvailable: function()
    {
        if(this.root.offsetParent != null)
        {
            return true;
        }
        return false;
    },

    /**
     * Render or refresh the datagrid depending on the internal visibility state of the DataGrid
     * If the data grid has sizing information available and is in render or refresh state call
     * the appropriate action and update the visibility property. If sizing is not available leave
     * the visibility property alone
     * @private
     */
    _possiblyRenderOrRefresh: function()
    {
        var visibility = this.grid.getVisibility();
        // If sizing not available yet do not change any flags or render
        if (this._isDataGridSizingAvailable())
        {
            if (visibility === DvtDataGrid.VISIBILITY_STATE_RENDER)
            {
                this.grid.render(this.root);
            }
            else if (visibility === DvtDataGrid.VISIBILITY_STATE_REFRESH)
            {
                this.grid.refresh(this.root);
            }
            //if sizing is available we are visible
            this.grid.setVisibility(DvtDataGrid.VISIBILITY_STATE_VISIBLE);
        }
        else
        {
            // Sizing isn't available, make sure we know the datagrid is hidden
            // should the app developer fail to call notifyHide, handle it here
            if (visibility === DvtDataGrid.VISIBILITY_STATE_VISIBLE)
            {
                this.grid.setVisibility(DvtDataGrid.VISIBILITY_STATE_HIDDEN);
            }
        }
    },

    /**
     * Get a header at a particular index and level
     * @param {number} index the absolute index of the header to get
     * @param {number} level the absolute level
     * @param {Object} headerContainer jQuery object of the row or column header
     * @param {number} start the start index of the headers in the data grid
     * @return {Object|null} jQuery object of the header or null if not found
     * @private
     */
    _getHeaderByIndex: function(index, level, headerContainer, start)
    {
        var lastHeader, grouping, relativeIndex;
        if (level < 0)
        {
            return null;
        }

        //get the last header and make sure it's a grouping or a cell
        lastHeader = headerContainer.children().first().children().last();
        if (lastHeader.hasClass(this._getMappedStyle('headercell')))
        {
            //if the first header is just a cell there is only one level, get the cell by just index
            return headerContainer.children().first().children('.' + this._getMappedStyle('headercell')).eq(index - start);
        }

        if (lastHeader.hasClass(this._getMappedStyle('endheadercell')))
        {
            //if the first header is just a cell there is only one level, get the cell by just index
            return headerContainer.children().first().children('.' + this._getMappedStyle('endheadercell')).eq(index - start);
        }

        //otherwise get the grouping container
        grouping = this._getGroupingContainer(index, level, 0, headerContainer.children().first().children());
        if (grouping == null)
        {
            return null;
        }

        if (level <= (parseInt(grouping.attr(this._getMappedAttribute('level')), 10) + parseInt(grouping.children().eq(0).attr(this._getMappedAttribute('depth')), 10) - 1))
        {
            //otherwise first child of the group is the header
            return grouping.children().eq(0);
        }

        // if the level we want is not the level of the group we wanted the innermost level
        start = parseInt(grouping.attr(this._getMappedAttribute('start')), 10);
        relativeIndex = index - start + 1;
        return grouping.children().eq(relativeIndex);
    },

    /**
     * Get a header container for nested headers at a particular index and level
     * @param {number} index the absolute index of the header to get
     * @param {number} level the absolute level
     * @param {number} currentLevel the level we are looking on
     * @param {Object} headers a jquery object of headers and groupings at the currentLevel
     * @return {Object|null} jQuery object of the header grouping or null if not found
     * @private
     */
     _getGroupingContainer: function (index, level, currentLevel, headers)
    {
        var headerIndex, headerExtent, i, headerRoot, headerDepth;
        //if the second child is the header cell or there is no second child we have the grouping
        if(headers.eq(1) == null || headers.eq(1).hasClass(this._getMappedStyle('headercell')) || headers.eq(1).hasClass(this._getMappedStyle('endheadercell')))
        {
            // if we are on the innermost level
            if (level === currentLevel)
            {
                return headers.eq(0).parent();
            }
            return null;
        }
        headerRoot = headers.parent().parent();
        // avoids skipping the first group on headers
        if (headerRoot.hasClass(this._getMappedStyle('header')) || headerRoot.hasClass(this._getMappedStyle('endheader')))
        {
            i = 0;
        }
        else
        {
            i = 1;
        }

        // loop over all headers skipping firstChild of groups
        for (;i < headers.length; i++)
        {
            // if the index is between that header start and start+extent dig deeper
            headerIndex = parseInt(headers.eq(i).attr(this._getMappedAttribute('start')), 10);
            headerExtent = parseInt(headers.eq(i).attr(this._getMappedAttribute('extent')), 10);
            headerDepth = parseInt(headers.eq(i).children().eq(0).attr(this._getMappedAttribute('depth')), 10);
            if (index >= headerIndex && index < headerIndex + headerExtent)
            {
                if (level < currentLevel + headerDepth)
                {
                    return headers.eq(i);
                }
                return this._getGroupingContainer(index, level, currentLevel + headerDepth, headers.eq(i).children());
            }
        };
        return null;
    },

    /**
     * Get the absolute index of a header
     * @param {Object} header the header
     * @return {number} the absolute index of the header
     * @private
     */
    _getHeaderIndex: function (header)
    {
        var index;

        // if there are multiple levels on the row header
        if (header.parent().hasClass(this._getMappedStyle('groupingcontainer')))
        {
            // get the groupingContainer's start value and set that to the index
            index = parseInt(header.parent().attr(this._getMappedAttribute('start')), 10);
            //if this is the groupingContainer's first child return that value
            if (header.get(0) === header.parent().children(":first").get(0))
            {
                return index;
            }
            //decrement the index by one for the first header element at the level above it
            index--;
        }
        else if (header.hasClass(this._getMappedStyle('rowheadercell')))
        {
            index = this.grid.getStartRowHeader();
        }
        else if (header.hasClass(this._getMappedStyle('colheadercell')))
        {
            index = this.grid.getStartColumnHeader();
        }
        else if (header.hasClass(this._getMappedStyle('colendheadercell')))
        {
            index = this.grid.getStartColumnEndHeader();
        }
        else
        {
            index = this.grid.getStartRowEndHeader();
        }

        index += header.index();
        return index;
    },
    /**
     * Get the level of a header
     * @param {Object} header the header
     * @return {number} the level of the header
     * @private
     */
    _getHeaderLevel: function (header)
    {
        var level;
        // if there are multiple levels on the row header
        if (header.parent().hasClass(this._getMappedStyle('groupingcontainer')))
        {
            level = parseInt(header.parent().attr(this._getMappedAttribute('level')), 10);
            if (header.get(0) === header.parent().children(":first").get(0))
            {
                return level;
            }
            // plus one case is if we are on the innermost level the headers do not have their own
            // grouping containers so if it is the first child it is the level of the grouping container
            // but all subsequent children are the next level in
            return level + 1;
        }
        return 0;
    },
    /**
     * Is multiple selection enabled
     * @return {boolean} true if multiple selection
     * @private
     */
    _isMultipleSelection: function ()
    {
        if (this.options['selectionMode'] != null)
        {
            if (this.options['selectionMode']['row'] === 'multiple' ||
                this.options['selectionMode']['cell'] === 'multiple')
            {
                return true;
            }
        }
        return false;
    },
    /**
     * Callback for datagrid to remove an element and call destroy on anything that is removed
     * @param {Element} element to remove
     * @private
     */
    _remove: function (element)
    {
        $(element).remove();
    },

    /**
     * Callback for datagrid to add or remove unique ID
     * @param {Element} element
     * @param {boolean} remove should it remove ID otherwise will add
     * @private
     */
    _uniqueId: function (element, remove)
    {
        if (remove)
        {
            $(element).removeUniqueId();
        }
        else
        {
            $(element).uniqueId();
        }
    },

    /**
     * Callback for datagrid to compare values
     * @param {*} value1
     * @param {*} value2
     * @returns {boolean}
     * @private
     */
    _compareValues: function (value1, value2)
    {
        return oj.Object.compareValues(value1, value2);
    },

    /**
     * <p>Scroll the datagrid to an x,y pixel location.
     * <p>If the x,y point is outside the range of the viewport the grid will scroll to the nearest location.
     * <p>If high-watermark scrolling is used, the grid will scroll within the currently fetched range.
     *
     * @expose
     * @memberof oj.ojDataGrid
     * @instance
     * @ignore
     *
     * @param {Object} options an object containing the scrollTo information
     * @param {Object} options.position scroll to an x,y pixel location which is relative to the origin of the grid
     * @param {Object} options.position.scrollX the x position in pixels of the scrollable region relative to the origin of the grid, this should always be positive. If RTL the value is the scroll position from the right of the grid.
     * @param {Object} options.position.scrollY the Y position in pixels of the scrollable region, this should always be positive.
     *
     * @example <caption>Invoke the <code class="prettyprint">scrollTo</code> method:</caption>
     * myDataGrid.scrollTo({"position": {"scrollX": 50, "scrollY":100}});
     */
    scrollTo: function(options)
    {
        this.grid.scroll(options);
    },

    /**
     * Returns a Promise that resolves when the element is ready, i.e. after data fetching, rendering, and animations complete.
     * Note that in the highwatermark scrolling case, element is ready after data fetching, rendering, and associated animations of items fetched so far are complete.
     *
     * <p>This method does not accept any arguments.
     *
     * @ignore
     * @expose
     * @memberof oj.ojDataGrid
     * @instance
     * @return {Promise} A Promise that resolves when the element is ready.
     */
    whenReady: function()
    {
        if (this._ready)
        {
            return Promise.resolve(true);
        }
        if (!this._readyPromise)
        {
            var self = this;
            this._readyPromise = new Promise(function(resolve)
            {
                self._readyPromiseResolve = resolve;
            });
        }
        return this._readyPromise;
    },

    /**
    * Called by component to declare rendering is not finished. This method currently handles the ready state
    * for the component whenReady API, the page level BusyContext, and the static whenReady API for the custom element
    * version of this component.
    * @private
    * @instance
    * @memberof oj.ojDataGrid
    */
    _NotReady : function() {
        // For component whenReady API
        this._ready = false;

        // For page level BusyContext
        // If we've already registered a busy state with the page's busy context, don't need to do anything further
        if (!this._readyResolveFunc)
        {
            var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
            var options = {'description' : "The component identified by '" + this.element.attr('id') + "' is being loaded."};
            this._readyResolveFunc = busyContext.addBusyState(options);
        }
    },

    /**
    * Called by component to declare rendering is finished. This method currently handles the ready state
    * for the component whenReady API, the page level BusyContext, and the static whenReady API for the custom element
    * version of this component.
    * @private
    * @instance
    * @memberof oj.ojDataGrid
    */
    _MakeReady : function() {
        // For component whenReady API
        if (this._readyPromiseResolve) {
            this._readyPromiseResolve(true);
            this._readyPromiseResolve = null;
        }
        this._ready = true;
        this._promise = null;

        // For page level BusyContext
        if (this._readyResolveFunc) {
            this._readyResolveFunc();
            this._readyResolveFunc = null;
        }
    },

    //** @inheritdoc */
    _CompareOptionValues: function(option, value1, value2)
    {
        switch(option)
        {
            case 'currentCell':
            case 'selection':
                return oj.Object.compareValues(value1, value2);
            default:
                return this._super(option, value1, value2);
        }
    }

    //////////////////     FRAGMENTS    //////////////////
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan="2">Cell</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Focus on the cell.  If <code class="prettyprint">selectionMode</code> for cells is enabled, selects the cell as well.
     *       If multiple selection is enabled the selection handles will appear. Tapping a different cell will deselect the previous selection.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Press & Hold</kbd></td>
     *       <td>Display context menu</td>
     *     </tr>
     *
     *     <tr>
     *       <td rowspan="3">Row</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>If <code class="prettyprint">selectionMode</code> for rows is enabled, selects the row as well.
     *       If multiple selection is enabled the selection handles will appear. Tapping a different row will deselect the previous selection.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Drag</kbd></td>
     *       <td>If the row that is dragged contains the active cell and <code class="prettyprint">dnd reorder row</code> is enabled the row will be moved within the DataGrid.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Press & Hold</kbd></td>
     *       <td>Display context menu</td>
     *     </tr>
     *
     *     <tr>
     *       <td rowspan="2">Header</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>If Multiple Selection is enabled or in row <code>selectionMode</code>, the row or column will be selected. Selection handles will appear for multiple selection. Otherwise, header cell will be focused.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Press & Hold</kbd></td>
     *       <td>Display context menu</td>
     *     </tr>
     *     <tr>
     *       <td>Header Gridline</td>
     *       <td><kbd>Drag</kbd></td>
     *       <td>Resizes the header if <code class="prettyprint">resizable</code> enabled along the axis.</td>
     *     </tr>
     *     <tr>
     *       <td>Corner</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Tapping on the corner will perform a select all operation on the datagrid if multiple selection is enabled.</td>
     *     </tr>
     *
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch section of classdesc, and standalone gesture doc
     * @memberof oj.ojDataGrid
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan="24">Cell</td>
     *       <td><kbd>Tab</kbd></td>
     *       <td>The first Tab into the DataGrid moves focus to the first cell of the first row.  The second Tab moves focus to the next focusable element outside of the DataGrid.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Tab</kbd></td>
     *       <td>The first Shift + Tab into the DataGrid moves focus to the first cell of the first row.  The second Shift + Tab moves focus to the previous focusable element outside of the DataGrid.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>LeftArrow</kbd></td>
     *       <td>Moves focus to the cell of the previous column within the current row.  There is no wrapping at the beginning or end of the columns.  If a row header is present, then the row header next to the first column of the current row will gain focus.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>RightArrow</kbd></td>
     *       <td>Moves focus to the cell of the next column within the current row.  There is no wrapping at the beginning or end of the columns.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>Moves focus to the cell of the previous row within the current column.  There is no wrapping at the beginning or end of the rows.  If a column header is present, then the column header above the first row of the current column will gain focus.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Moves focus to the cell of the next row within the current column.  There is no wrapping at the beginning or end of the rows.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Home</kbd></td>
     *       <td>Moves focus to the first (available) cell of the current row.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>End</kbd></td>
     *       <td>Moves focus to the last (available) cell of the current row.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>PageUp</kbd></td>
     *       <td>Moves focus to the first (available) cell in the current column.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>PageDown</kbd></td>
     *       <td>Moves focus to the last (available) cell in the current column.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + Space</kbd></td>
     *       <td>Selects all the cells of the current column.  This is only available if multiple cell selection mode is enabled.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Space</kbd></td>
     *       <td>Selects all the cells of the current row.  This is only available if multiple cell selection mode is enabled.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Arrow</kbd></td>
     *       <td>Extends the current selection.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + Arrow</kbd></td>
     *       <td>Move focus to level 0 of the active index of the header in the arrow direction if it exists.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + F8</kbd></td>
     *       <td>Freezes the current selection, therefore allowing user to move focus to another location to add additional cells to the current selection.  This is used to accomplish non-contiguous selection.  Use the Esc key or press Shift+F8 again to exit this mode.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + F10</kbd></td>
     *       <td>Brings up the context menu.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + X</kbd></td>
     *       <td>Marks the current row to move if dnd is enabled and the datasource supports move operation.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + V</kbd></td>
     *       <td>Move the row that is marked to directly under the current row.  If the row with the focused cell is the last row, then it will be move to the row above the current row.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + A</kbd></td>
     *       <td>If multiple selection is enabled, performs a select all on the datagrid.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + Alt + 5</kbd></td>
     *       <td>Read the context and content of the current cell to the screen reader.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>F2</kbd></td>
     *       <td>Makes the content of the cell actionable, such as a link.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Enter</kbd></td>
     *       <td>Makes the content of the cell actionable and acts on the content, such as going to a link.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Alt + Enter</kbd></td>
     *       <td>Makes the content of the cell actionable, such as a link.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Esc</kbd></td>
     *       <td>If the cell is actionable it exits actionable mode.</td>
     *     </tr>
     *     <tr>
     *       <td rowspan="13">Column Header Cell</td>
     *       <td><kbd>LeftArrow</kbd></td>
     *       <td>Moves focus to the previous column header.  There is no wrapping at the beginning or end of the column headers.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>RightArrow</kbd></td>
     *       <td>Moves focus to the next column header.  There is no wrapping at the beginning or end of the column headers.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Moves focus to the cell of the first row directly below the column header. If using nested headers will move focus up a level.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>If using nested headers will move focus down a level.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + UpArrow</kbd></td>
     *       <td>If in the column end header, move focus to level 0 of the active index in the column header if it exists.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + DownArrow</kbd></td>
     *       <td>If in the column header, move focus to level 0 of the active index in the column end header if it exists.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Enter</kbd></td>
     *       <td>Toggle the sort order of the column if the column is sortable.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + F10</kbd></td>
     *       <td>Brings up the context menu.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Space</kbd></td>
     *       <td>If multiple selection is enabled and not in <code>selectionMode</code> row, the column(s) underneath the header will be selected.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Right Arrow</kbd></td>
     *       <td>If multiple selection is enabled and not in <code>selectionMode</code> row, the column selection will extend to the right by the number of columns covered by the header to the right of the current selection frontier header.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Left Arrow</kbd></td>
     *       <td>If multiple selection is enabled and not in <code>selectionMode</code> row, the column selection will extend to the right left the number of columns covered by the header to the left of the current selection frontier header.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Up Arrow</kbd></td>
     *       <td>If multiple selection is enabled and not in <code>selectionMode</code> row and the current selection frontier header has a parent nested header, the column selection will extend to cover the columns beneath the parent header. 
     *           <br/> Extending the selection with arrow keys will use the parent level. If the parent header is directly above the anchor header, the anchor will shift to the parent header and future selections will be based on the parent header. 
     *           <br/> If we are already at the highest level, nothing will happen.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Down Arrow</kbd></td>
     *       <td>If multiple selection is enabled and not in <code>selectionMode</code> row and the current selection frontier header has a child nested header, the column selection will extend to cover the columns beneath the child header. 
     *           <br/> Extending the selection with arrow keys will use the child level. If the child header is directly below the anchor header, the anchor will shift to the child header and future selections will be based on the child header. 
     *           <br/> If we are already at the lowest level, it will simply move into the databody and select the first cell underneath the header.</td>
     *     </tr>
     *     <tr>
     *       <td rowspan="12">Row Header Cell</td>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>Moves focus to the previous row header.  There is no wrapping at the beginning or end of the row headers.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Moves focus to the next row header.  There is no wrapping at the beginning or end of the row headers.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>RightArrow</kbd></td>
     *       <td>Moves focus to the cell of the first column directly next to the row header. If using nested headers will move focus up a level.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>LeftArrow</kbd></td>
     *       <td>Moves focus to the cell of the first column directly next to the row header in RTL direction. If using nested headers will move focus down a level.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + LeftArrow</kbd></td>
     *       <td>If in the row end header, move focus to level 0 of the active index in the row header if it exists.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + RightArrow</kbd></td>
     *       <td>If in the row header, move focus to level 0 of the active index in the row end header if it exists.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + F10</kbd></td>
     *       <td>Brings up the context menu.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Space</kbd></td>
     *       <td>If multiple selection is enabled, the row(s) underneath the header will be selected.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Up Arrow</kbd></td>
     *       <td>If multiple selection is enabled, the row selection will extend up by the number of rows covered by the header above the current selection frontier header.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Down Arrow</kbd></td>
     *       <td>If multiple selection is enabled, the row selection will extend down by the number of rows covered by the header below the current selection frontier header.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Left Arrow</kbd></td>
     *       <td>If multiple selection is enabled and the current selection frontier header has a parent nested header, the row selection will extend to cover the rows beneath the parent header. 
     *           <br/> Extending the selection with arrow keys will use the parent level. If the parent header is directly above the anchor header, the anchor will shift to the parent header and future selections will be based on the parent header. 
     *           <br/> If we are already at the highest level, nothing will happen.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Right Arrow</kbd></td>
     *       <td>If multiple selection is enabled  and the current selection frontier header has a child nested header, the row selection will extend to cover the rows beneath the child header. 
     *           <br/> Extending the selection with arrow keys will use the child level. If the child header is directly below the anchor header, the anchor will shift to the child header and future selections will be based on the child header. 
     *           <br/> If we are already at the lowest level, it will simply move into the databody and select the first cell underneath the header.</td>
     *     </tr>


     *   </tbody>
     * </table>
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojDataGrid
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     *
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-helper-justify-content-[direction]</td>
     *       <td>Use this class on cells' and headers' className property to align your content horizontally. By default the alignment is flex-end on cells and varies on headers, see other possibilities in the flex layout justify section.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-helper-align-items-[direction]</td>
     *       <td>Use this class on cells' and headers' className property to align your content vertically. By default the alignment is center on cells and headers, see other possibilities in the flex layout align section.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-datagrid-cell-no-padding</td>
     *       <td>Used to style a datagrid cell so that it has no padding.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-datagrid-cell-padding</td>
     *       <td>Used to style a datagrid cell so that it has the default padding.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-datagrid-depth-[number]</td>
     *       <td>Used to style the default header widths and heights. By default the datagrid supports up to depth 7. If you have headers width depth greater than 7 specify
     *       the defaults using the class name description or use apply custom style rules to the headers.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojDataGrid
     */

    //////////////////     SUB-IDS     //////////////////

    /**
     * <p>Sub-ID for the DataGrid element's cells.</p>
     *
     * To lookup a cell the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-datagrid-cell'</li>
     * <li><b>rowIndex</b>: the zero based absolute row index</li>
     * <li><b>columnIndex</b>: the zero based absolute column index</li>
     * </ul>
     *
     * @ojsubid oj-datagrid-cell
     * @memberof oj.ojDataGrid
     * @example <caption>Get the cell at the specified location:</caption>
     * var node = myDataGrid.getNodeBySubId({subId: 'oj-datagrid-cell', rowIndex: rowIndexValue, columnIndex: columnIndexValue});
     */

    /**
     * <p>Sub-ID for the DataGrid element's headers.</p>
     *
     * To lookup a header the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-datagrid-header'</li>
     * <li><b>axis</b>: 'column'/'row'/'columnEnd'/'rowEnd'</li>
     * <li><b>index</b>: the zero based absolute row/column index.</li>
     * <li><b>level</b>: the zero based header level, 0 is the outer edge, if not specified will default to 0</li>
     * </ul>
     *
     * @ojsubid oj-datagrid-header
     * @memberof oj.ojDataGrid
     *
     * @example <caption>Get the header at the specified location:</caption>
     * var node = myDataGrid.getNodeBySubId({subId: 'oj-datagrid-header', axis: 'axisValue', index: indexValue, level: levelValue});
     */

    /**
     * <p>Sub-ID for the DataGrid element's sort ascending icon in column headers.</p>
     *
     * To lookup a sort icon the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-datagrid-sort-ascending'</li>
     * <li><b>axis</b>: 'column'</li>
     * <li><b>index</b>: the zero based absolute column index</li>
     * <li><b>level</b>: the zero based header level, 0 is the outer edge, if not specified will default to 0</li>
     * </ul>
     *
     * @ojsubid oj-datagrid-sort-ascending
     * @memberof oj.ojDataGrid
     *
     * @example <caption>Get the sort icon from the header at the specified location:</caption>
     * var node = myDataGrid.getNodeBySubId({subId: 'oj-datagrid-sort-ascending', axis: 'axisValue', index: indexValue, level: levelValue});
     */

    /**
     * <p>Sub-ID for the DataGrid element's sort descending icon in column headers.</p>
     *
     * To lookup a sort icon the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-datagrid-sort-descending'</li>
     * <li><b>axis</b>: 'column'</li>
     * <li><b>index</b>: the zero based absolute column index</li>
     * <li><b>level</b>: the zero based header level, 0 is the outer edge, if not specified will default to 0</li>
     * </ul>
     *
     * @ojsubid oj-datagrid-sort-descending
     * @memberof oj.ojDataGrid
     *
     * @example <caption>Get the descending sort icon from the header at the specified location:</caption>
     * var node = myDataGrid.getNodeBySubId({subId: 'oj-datagrid-sort-descending', axis: 'axisValue', index: indexValue, level: levelValue});
     */

    //////////////////     CONTEXTS     //////////////////
    /**
     * <p>Context for the ojDataGrid element's cells.</p>
     *
     * @property {function} component a reference to the DataGrid widgetConstructor
     * @property {Object} cell the container data object for the header
     * @property {Object} data the data object for the header
     * @property {Object} datasource a reference to the data source object
     * @property {Object} indexes the object that contains both the zero based row index and column index in which the cell is bound to
     * @property {number} indexes.row the zero based absolute row index
     * @property {number} indexes.column the zero based absolute column index
     * @property {Object} keys the object that contains both the row key and column key which identifies the cell
     * @property {number|string} keys.row the row key
     * @property {number|string} keys.column the column key
     * @property {Object} extents the object that contains both the row extent and column extent of the cell
     * @property {number} extents.row the row extent
     * @property {number} extents.column the column extent     
     * @property {string} mode the mode the cell is rendered in
     * @property {string} subId the subId of the cell
     *
     * @ojnodecontext oj-datagrid-cell
     * @memberof oj.ojDataGrid
     */
    /**
     * <p>Context for the ojDataGrid element's headers.</p>
     *
     * @property {number} axis the axis of the header, possible values are 'row'/'column'/'columnEnd'/'rowEnd'
     * @property {function} component a reference to the DataGrid widgetConstructor
     * @property {Object} data the data object for the header
     * @property {Object} datasource a reference to the data source object
     * @property {number} depth the the number of levels the header spans
     * @property {number} extent the number of indexes the header spans
     * @property {number} index the index of the header, where 0 is the index of the first header
     * @property {number|string} key the key of the header
     * @property {number} level the level of the header. The outermost header is level zero
     * @property {string} subId the subId of the header
     *
     * @ojnodecontext oj-datagrid-header
     * @memberof oj.ojDataGrid
     */
});

/**
 * Build the actions object which maps actions to the methods to invoke because of them */
DvtDataGrid.prototype._setupActions = function()
{
    this.actions = {
        'ACTIONABLE': this._handleActionable,
        'EXIT_ACTIONABLE': this._handleExitActionable,
        'TAB_NEXT_IN_CELL': this._handleActionableTab,
        'TAB_PREV_IN_CELL': this._handleActionablePrevTab,
        'EDITABLE': this._handleEditable, // if editable go into edit, if not go into actionable
        'EXIT_EDITABLE': this._handleExitEditable,
        'DATA_ENTRY': this._handleDataEntry,
        'EXIT_DATA_ENTRY': this._handleExitDataEntry,
        'EDIT': this._handleEdit,
        'EXIT_EDIT': this._handleExitEdit,
        'CANCEL_EDIT': this._handleCancelEdit,
        'NO_OP': this._handleNoOp,
        'EAT': this._handleEat, // eat event
        'FOCUS_LEFT': this._handleFocusLeft,
        'FOCUS_RIGHT': this._handleFocusRight,
        'FOCUS_UP': this._handleFocusUp,
        'FOCUS_DOWN': this._handleFocusDown,
        'FOCUS_ROW_FIRST': this._handleFocusRowFirst,
        'FOCUS_ROW_LAST': this._handleFocusRowLast,
        'FOCUS_COLUMN_FIRST': this._handleFocusColumnFirst,
        'FOCUS_COLUMN_LAST': this._handleFocusColumnLast,
        'FOCUS_COLUMN_HEADER': this._handleFocusColumnHeader,
        'FOCUS_COLUMN_END_HEADER': this._handleFocusColumnEndHeader,
        'FOCUS_ROW_HEADER': this._handleFocusRowHeader,
        'FOCUS_ROW_END_HEADER': this._handleFocusRowEndHeader,
        'READ_CELL': this.readCurrentContent,
        'SORT': this._handleSortKey,
        'SELECT_DISCONTIGUOUS': this._handleSelectDiscontiguous,
        'SELECT_EXTEND_LEFT': this._handleExtendSelectionLeft,
        'SELECT_EXTEND_RIGHT': this._handleExtendSelectionRight,
        'SELECT_EXTEND_UP': this._handleExtendSelectionUp,
        'SELECT_EXTEND_DOWN': this._handleExtendSelectionDown,
        'SELECT_ROW': this._handleSelectRow,
        'SELECT_COLUMN': this._handleSelectColumn,
        'SELECT_ALL': this._handleSelectAll,
        'CUT': this._handleCut,
        'CANCEL_REORDER': this._handleCancelReorder,
        'PASTE': this._handlePaste
    };
};

/**
 * Get the function for a given keydown event.
 * @param {Event} event
 * @param {string} cellOrHeader 'cell'/'header'
 * @returns {Function|undefined} the function to invoke due to the keydown
 */
DvtDataGrid.prototype._getActionFromKeyDown = function(event, cellOrHeader)
{
    var capabilities = {
        'cellOrHeader': cellOrHeader,
        'readOnly': !this._isCellEditable(),
        'currentMode': this._getCurrentMode(),
        'activeMove': (this.m_cutCells != null),
        'rowMove': this._isMoveEnabled('row'),
        'columnSort': cellOrHeader == 'column' ? this._isDOMElementSortable(this._getActiveElement()) : false,
        'selection': this._isSelectionEnabled(),
        'selectionMode': this.m_options.getSelectionMode(),
        'multipleSelection': this.isMultipleSelection()
    };
    return this.actions[this.m_keyboardHandler.getAction(event, capabilities)];
};

//////////////////////////////////// ACTIONABLE METHODS/////////////////////////
/**
 * Determine if the data grid is in actionable mode.
 * @return returns true if the data grid is in actionable mode, false otherwise.
 * @protected
 */
DvtDataGrid.prototype.isActionableMode = function()
{
    return (this.m_currentMode === "actionable");
};

/**
 * Sets whether the data grid is in actionable mode or reverts it to navigation mode
 * @param {boolean} flag true to set grid to actionable mode, false otherwise
 * @protected
 */
DvtDataGrid.prototype.setActionableMode = function(flag)
{
    if (flag)
    {
        this.m_currentMode = "actionable";
    }
    else
    {
        this.m_currentMode = "navigation";
    }

    // update screen reader alert
    this._setAccInfoText(this.isActionableMode() ? 'accessibleActionableMode' : 'accessibleNavigationMode');
};

/**
 * Enter actionable mode
 * @param {Event} event the event triggering actionable mode
 * @param {Element|undefined|null} element to set actionable
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleActionable = function(event, element)
{
    this._enterActionableMode(element);
    return false;
};

/**
 * Exit actionable mode on the active cell if in actionable mode
 * @param {Event} event the event exiting actionable mode
 * @param {Element|undefined|null} element to unset actionable
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleExitActionable = function(event, element)
{
    this._exitActionableMode();
    this._highlightActive();
    return false;
};

/**
 * Handle a tab that is pressed when in actionable mode
 * @param {Event} event the event causing the actionable tab
 * @param {Element|undefined|null} element to unset actionable
 * @returns {boolean} true if we have shifted focus within the actionable cell
 */
DvtDataGrid.prototype._handleActionableTab = function(event, element)
{
    var focusElems = this.getFocusableElementsInNode(element);
    if (focusElems.length > 0 && event.target == focusElems[focusElems.length - 1])
    {
        // recycle to first focusable element in the cell
        focusElems[0].focus();
        return true;
    }
    // let the tab go to the next item in the cell on its own
    return false;
};

/**
 * Handle a tab that is pressed when in actionable mode
 * @param {Event} event the event causing the actionable tab
 * @param {Element|undefined|null} element to unset actionable
 * @returns {boolean} true if we have shifted focus within the actionable cell
 */
DvtDataGrid.prototype._handleActionablePrevTab = function(event, element)
{
    var focusElems = this.getFocusableElementsInNode(element);
    if (focusElems.length > 0 && event.target == focusElems[0])
    {
        // recycle to last focusable element in the cell
        focusElems[focusElems.length - 1].focus();
        return true;
    }
    // let the tab go to the previous item in the cell on its own
    return false;
};

//////////////////////////////////// EDITING METHODS/////////////////////////
/**
 * Get the edit mode values can be none/cell
 * @returns {string}  none or cell
 * @private
 */
DvtDataGrid.prototype._getEditMode = function()
{
    if (this.m_editMode == null)
    {
         this.m_editMode = this.m_options.getEditMode();
    }
    return this.m_editMode;
};

/**
 * Get the current mode of the datagrid
 * @returns {string} navigation/actionable/enter/edit
 * @private
 */
DvtDataGrid.prototype._getCurrentMode = function()
{
    if (this.m_currentMode == null)
    {
         this.m_currentMode = 'navigation';
    }
    return this.m_currentMode;
};

/**
 * Is the current mode edit or enter
 * @returns {boolean} true if edit or enter
 * @private
 */
DvtDataGrid.prototype._isEditOrEnter = function()
{
    var c = this._getCurrentMode();
    return c === 'edit';
};

/**
 * Can the grid as a whole be editable
 * @returns {boolean} true if the edit mode is cell
 * @private
 */
DvtDataGrid.prototype._isGridEditable = function()
{
    var editMode = this._getEditMode();
    if (editMode === 'cellNavigation' || editMode === 'cellEdit')
    {
        return true;
    }
    return false;
};

/**
 * Is the grid in ediable or readOnly mode
 * @returns {boolean} true if the edit mode is cell and editable enabled
 * @private
 */
DvtDataGrid.prototype._isCellEditable = function()
{
    var editMode = this._getEditMode();
    if (editMode === 'cellEdit')
    {
        return true;
    }
    return false;
};


/**
 * Enter editable mode
 * @param {Event} event the event triggering editable mode
 * @param {Element|undefined|null} element
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleEditable = function(event, element)
{
    if (this._isGridEditable())
    {
        this.m_editMode = null;
        this.m_setOptionCallback("editMode", 'cellEdit', {'_context': {writeback: true, internalSet: true}});
        this.m_utils.removeCSSClassName(this.m_root, this.getMappedStyle('readOnly'));
        this.m_utils.addCSSClassName(this.m_root, this.getMappedStyle('editable'));
        this._updateEdgeCellBorders('');
        this._setAccInfoText('accessibleEditableMode');
        this._setEditableClone(element);
    }
    else
    {
        this._handleActionable(event, element);
    }
    return false;
};

/**
 * Exit editable mode
 * @param {Event} event the event triggering editable mode
 * @param {Element|undefined|null} element
 */
DvtDataGrid.prototype._handleExitEditable = function(event, element)
{
    this.m_editMode = null;
    this.m_setOptionCallback("editMode", 'cellNavigation', {'_context': {writeback: true, internalSet: true}});
    this.m_utils.addCSSClassName(this.m_root, this.getMappedStyle('readOnly'));
    this.m_utils.removeCSSClassName(this.m_root, this.getMappedStyle('editable'));
    this._updateEdgeCellBorders('none');
    this._setAccInfoText('accessibleNavigationMode');
    delete this.m_editableClone;
};

/**
 * Enter enter mode
 * @param {Event} event the event triggering enter mode
 * @param {Element|undefined|null} element
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleDataEntry = function(event, element)
{
    var details, rerender;
    details = {'event': event, 'ui': {'cell': element, 'cellContext': element[this.getResources().getMappedAttribute("context")]}};
    rerender = this.fireEvent('beforeEdit', details);
    if (rerender)
    {
        this._reRenderCell(element, 'edit', this.getMappedStyle('cellEdit'), this.m_editableClone);
        this._enableAllFocusableElements(element);
        // focus on first focusable item in the cell
        this._overwriteFlag = true;
        if (this._setFocusToFirstFocusableElement(element))
        {
            this.m_currentMode = 'edit';
        }
        else
        {
            // if there was nothing to edit remove the edit class
            this.m_utils.removeCSSClassName(element, this.getMappedStyle('cellEdit'));
        }
        this._overwriteFlag = false;
   }
   return false;
};

/**
 * Exit enter mode
 * @param {Event} event the event triggering exit enter mode
 * @param {Element|undefined|null} element
 * @returns {boolean} true if enter is left successfully
 */
DvtDataGrid.prototype._handleExitDataEntry = function(event, element)
{
    return this._leaveEditing(event, element, false);
};

/**
 * Enter edit mode
 * @param {Event} event the event triggering edit mode
 * @param {Element|undefined|null} element
 * @returns {boolean} true if edit mode entered
 */
DvtDataGrid.prototype._handleEdit = function(event, element)
{
    var details, rerender;

    details = {'event': event, 'ui': {'cell': element, 'cellContext': element[this.getResources().getMappedAttribute("context")]}};
    rerender = this.fireEvent('beforeEdit', details);
    if (rerender)
    {
        this._reRenderCell(element, 'edit', this.getMappedStyle('cellEdit'), this.m_editableClone);
        
        var self = this;
        var busyContext = oj.Context.getContext(element).getBusyContext();
        busyContext.whenReady().then(function () 
        {
            // enable all focusable elements
            self._enableAllFocusableElements(element);

            // focus on first focusable item in the cell
            if (self._setFocusToFirstFocusableElement(element))
            {
                self.m_currentMode = 'edit';
            }
            else
            {
                // if there was nothing to edit remove the edit class
                self.m_utils.removeCSSClassName(element, self.getMappedStyle('cellEdit'));
            }
        });
    }
    else
    {
        rerender = false;
        this._enterActionableMode(element);
    }
    return rerender;
};

/**
 * Exit edit mode
 * @param {Event} event
 * @param {Element|undefined|null} element
 * @returns {boolean} true if editing left successully
 */
DvtDataGrid.prototype._handleExitEdit = function(event, element)
{
    return this._leaveEditing(event, element, false);
};

/**
 * Cancel an edit
 * @param {Event} event
 * @param {Element|undefined|null} element
 * @returns {boolean} true if editing cancelled successully
 */
DvtDataGrid.prototype._handleCancelEdit = function(event, element)
{
    if (this._leaveEditing(event, element, true))
    {
        this._setEditableClone(element);
        return true;
    }
    return false;
};

/**
 * Leave editing
 * @param {Event} event the event triggering actionable mode
 * @param {Element|undefined|null} element
 * @returns {boolean} false
 */
DvtDataGrid.prototype._leaveEditing = function(event, element, cancel)
{
    var details, rerender;
    details = {'event': event, 'ui': {'cell': element, 'cellContext': element[this.getResources().getMappedAttribute("context")], 'cancelEdit': cancel}};
    if (!cancel)
    {
        this._disableAllFocusableElements(element);
        this._highlightActive();
    }
    rerender = this.fireEvent('beforeEditEnd', details);
    if (rerender)
    {
        this.m_currentMode = 'navigation';
        this._disableAllFocusableElements(element);
        this._highlightActive();
        this._reRenderCell(element, 'navigation', this.getMappedStyle('cellEdit'), this.m_editableClone);
    }
    else
    {
        rerender = false;
        this._scrollToActive(this.m_active);
        this._enableAllFocusableElements(element);
        // focus on first focusable item in the cell
        this._setFocusToFirstFocusableElement(element);
    }
    return rerender;
};

//////////////////////////////////// FOCUS METHODS/////////////////////////
/**
 * Handles all the various focus changes by keystroke
 * @param {Event} event
 * @param {Element} element
 * @param {number} keyCode
 * @param {boolean} isExtend
 * @param {boolean} jumpToHeaders
 * @returns {boolean} true if event processed
 */
DvtDataGrid.prototype._handleFocusKey = function(event, element, keyCode, isExtend, jumpToHeaders)
{
    var changeFocus = true, changeRegions = true, editing, returnVal, oldActive;
    if (this.m_active != null)
    {
        if (this.m_active['type'] == 'cell')
        {
            if (this._isEditOrEnter())
            {
                editing = true;
                changeFocus = this._leaveEditing(event, element, false);
                changeRegions = false;
            }
            else if (this.isActionableMode())
            {
                this._exitActionableMode();
            }

            if (changeFocus)
            {
                oldActive = this.m_active;
                returnVal = this.handleFocusChange(keyCode, isExtend, event, changeRegions, jumpToHeaders);
                if (this._isGridEditable() && oldActive != this.m_active && editing && this.m_utils.isTouchDevice())
                {
                    return this._handleDataEntry(event, this._getActiveElement());
                }
                return returnVal;
            }
            return true;
        }
        else
        {
            return this.handleHeaderFocusChange(keyCode, event, isExtend, jumpToHeaders);
        }
    }
    return false;
};

/**
 * Handle a focus to the left element
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusLeft = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.LEFT_KEY, false, false);
};

/**
 * Handle a focus to the right element
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusRight = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.RIGHT_KEY, false, false);
};

/**
 * Handle a focus to the up element
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusUp = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.UP_KEY, false, false);
};

/**
 * Handle a focus to the down element
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusDown = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.DOWN_KEY, false, false);
};

/**
 * Handle a focus to the first row same column
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusRowFirst = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.PAGEUP_KEY, false, false);
};

/**
 * Handle a focus to the last row same column
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusRowLast = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.PAGEDOWN_KEY, false, false);
};

/**
 * Handle a focus to the first column same row
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusColumnFirst = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.HOME_KEY, false, false);
};

/**
 * Handle a focus to the last column same row
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusColumnLast = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.END_KEY, false, false);
};

/**
 * Handle a focus to the row header
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusRowHeader = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.LEFT_KEY, false, true);
};

/**
 * Handle a focus to the row end header
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusRowEndHeader = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.RIGHT_KEY, false, true);
};

/**
 * Handle a focus to the column header
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusColumnHeader = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.UP_KEY, false, true);
};

/**
 * Handle a focus to the column end header
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleFocusColumnEndHeader = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.DOWN_KEY, false, true);
};
/////////////////////// SELECTION METHODS ////////////////////////////
/**
 * Handle a selection of the whole row
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleSelectRow = function(event, element)
{
    var axis, start, end, elem, level, index;
    if  (!this._isSelectionEnabled() || (!this.isMultipleSelection() && this.m_options.getSelectionMode() != 'row'))
    {
        return false;
    }

    if (this.m_utils.containsCSSClassName(element, this.getMappedStyle('cell')))
    {
        index = this.m_active['indexes']['row'];
        start = index;
        end  = index;
        level = this.m_rowHeaderLevelCount - 1;
    }
    else
    {
        if (this.m_active == null || this.m_active['type'] != 'header' || this.m_active['axis'].indexOf('row') == -1)
        {
            return false;
        }
        axis = this.m_active['axis'];
        index = this.m_active['index'];
        level = this.m_active['level'];
        if (this.m_rowHeaderLevelCount - 1 === level)
        {
            start = index;
            end = index;
        }
        else
        {
            elem = this._getActiveElement();
            start = this._getAttribute(elem['parentNode'], 'start', true);
            end = start + this._getAttribute(elem['parentNode'], 'extent', true) - 1;
        }
    }

    if (this.m_selectionFrontier)
    {
        if (this.m_selectionFrontier['axis'] && this.m_selectionFrontier['axis'].indexOf('column') != -1)
        {
            this._handleSelectAll(event);
            return true;        
        } 
    }

    // set the frontier as it would be with header selection
    this.setHeaderSelectionFrontier('row', end, index, level, element, true);

    //handle the space key in headers
    this._selectEntireRow(start, end, event);

    // announce to screen reader, no need to include context info
    this._setAccInfoText('accessibleRowSelected', {'row': index + 1});
    return true;
};

/**
 * Handle a selection of the whole column
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleSelectColumn = function(event, element)
{
    var axis, start, end, elem, level, index;
    if  (!this._isSelectionEnabled() || !this.isMultipleSelection() || this.m_options.getSelectionMode() != 'cell')
    {
        return false;
    }

    if (this.m_utils.containsCSSClassName(element, this.getMappedStyle('cell')))
    {
        index = this.m_active['indexes']['column'];
        start = index;
        end  = index;
        level = this.m_columnHeaderLevelCount - 1;
    }
    else
    {
        if (this.m_active == null || this.m_active['type'] != 'header' || this.m_active['axis'].indexOf('column') == -1)
        {
            return false;
        }
        axis = this.m_active['axis'];
        index = this.m_active['index'];
        level = this.m_active['level'];
        if (this.m_columnHeaderLevelCount - 1 === level)
        {
            start = index;
            end = index;
        }
        else
        {
            elem = this._getActiveElement();
            start = this._getAttribute(elem['parentNode'], 'start', true);
            end = start + this._getAttribute(elem['parentNode'], 'extent', true) - 1;
        }
    }

    if (this.m_selectionFrontier)
    {
        if (this.m_selectionFrontier['axis'] && this.m_selectionFrontier['axis'].indexOf('row') != -1)
        {
            this._handleSelectAll(event);
            return true;        
        } 
    }

    // set the frontier as it would be with header selection
    this.setHeaderSelectionFrontier('column', end, index, level, element, true);

    //handle the space key in headers
    this._selectEntireColumn(start, end, event);

    // announce to screen reader, no need to include context info
    this._setAccInfoText('accessibleColumnSelected', {'column': index + 1});
    return true;
};

/**
 * Handle entering discontiguous selection
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleSelectDiscontiguous = function(event, element)
{
    this.setDiscontiguousSelectionMode(!this.m_discontiguousSelection);
    return true;
};

/**
 * Handle extend selection left
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleExtendSelectionLeft = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.LEFT_KEY, true, false);
};

/**
 * Handle extend selection right
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleExtendSelectionRight = function(event, element)
{
    return this._handleFocusKey(event, element,DvtDataGrid.keyCodes.RIGHT_KEY, true, false);
};

/**
 * Handle extend selection up
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleExtendSelectionUp = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.UP_KEY, true, false);
};

/**
 * Handle extend selection down
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleExtendSelectionDown = function(event, element)
{
    return this._handleFocusKey(event, element, DvtDataGrid.keyCodes.DOWN_KEY, true, false);
};

/**
 * Handle sort key
 * @param {Event} event the event causing the action
 * @param {Element} element target cell or header of the event
 * @returns {boolean} true if processed
 */
DvtDataGrid.prototype._handleSortKey = function(event, element)
{
    // sort, first check if the column is sortable
    if (element.getAttribute(this.getResources().getMappedAttribute("sortable")) == "true")
    {
        this._handleKeyboardSort(element, event);
        return true;
    }
    else
    {
        // enter actionable mode but don't prevent default so the action is taken
        return this._handleActionable(event, element);
    }
};

/**
 * Enter editable mode
 * @param {Event} event the event triggering actionable mode
 * @param {Element} element to set actionable
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleNoOp = function(event, element)
{
    return false;
};

/**
 * Enter editable mode
 * @param {Event} event the event triggering actionable mode
 * @param {Element} element to set actionable
 * @returns {boolean} false
 */
DvtDataGrid.prototype._handleEat = function(event, element)
{
    return true;
};
/**
 * This class contains all utility methods used by the Grid.
 * @param {Object} dataGrid the dataGrid using the utils
 * @constructor
 * @private
 */
var DvtDataGridKeyboardHandler = function(dataGrid)
{
    this.grid = dataGrid;
};

/**
 * Get the action of a particular keydown event given information about the state of the frid
 * @param {Event} event
 * @param {Object} capabilities
 * @returns {string}
 */
DvtDataGridKeyboardHandler.prototype.getAction = function(event, capabilities)
{
    var keyCode, keyCodes, ctrlKey, shiftKey, altKey, cellOrHeader, readOnly,
            currentMode, activeMove, rowMove, columnSort, selection, selectionMode, multipleSelection;
    keyCode = event.keyCode;
    ctrlKey = this.grid.m_utils.ctrlEquivalent(event);
    shiftKey = event.shiftKey;
    altKey = event.altKey;
    keyCodes = DvtDataGrid.keyCodes;

    cellOrHeader = capabilities['cellOrHeader'];
    readOnly = capabilities['readOnly'];
    currentMode = capabilities['currentMode'];
    activeMove = capabilities['activeMove'];
    rowMove = capabilities['rowMove'];
    columnSort = capabilities['columnSort'];
    selection = capabilities['selection'];
    selectionMode = capabilities['selectionMode'];
    multipleSelection = capabilities['multipleSelection'];

    switch (keyCode)
    {
        case keyCodes.TAB_KEY:
            if (currentMode === 'actionable')
            {
                if (shiftKey)
                {
                    return 'TAB_PREV_IN_CELL';
                }
                return 'TAB_NEXT_IN_CELL';
            }
            else if (!readOnly)
            {
                if (shiftKey)
                {
                    return 'FOCUS_LEFT';
                }
                return 'FOCUS_RIGHT';
            }
            break;
        case keyCodes.ENTER_KEY:
            if (cellOrHeader === 'column' && columnSort)
            {
                return 'SORT';
            }
            // enter actionable mode on headers since they cannot be edited
            else if ((!altKey && readOnly && currentMode === 'navigation') || cellOrHeader != 'cell')
            {
                return 'ACTIONABLE';
            }
            else if (!readOnly && !altKey)
            {
                if (shiftKey)
                {
                    return 'FOCUS_UP';
                }
                return 'FOCUS_DOWN';
            }
            else if (altKey && readOnly && currentMode === 'navigation')
            {
                return 'EDITABLE';
            }
            else if (!readOnly)
            {
                if (currentMode == 'navigation' || currentMode === 'edit')
                {
                    return 'EDIT';
                }
            }
            break;
        case keyCodes.ESC_KEY:
            if (currentMode === 'actionable')
            {
                return 'EXIT_ACTIONABLE';
            }
            else if (activeMove)
            {
                return 'CANCEL_REORDER';
            }
            else if (!readOnly)
            {
                if (currentMode === 'navigation')
                {
                    return 'EXIT_EDITABLE';
                }
                else if (currentMode === 'edit')
                {
                    return 'CANCEL_EDIT';
                }
            }
            else if (this.grid.m_discontiguousSelection)
            {
                return 'SELECT_DISCONTIGUOUS';
            }
            break;
        case keyCodes.SPACE_KEY:
            if (cellOrHeader.indexOf('row') != -1 && selection && ((selectionMode == 'cell' && multipleSelection) || selectionMode == 'row'))
            {
                return 'SELECT_ROW';
            }
            else if (cellOrHeader.indexOf('column') != -1 && selection && selectionMode == 'cell' && multipleSelection)
            {
                return 'SELECT_COLUMN';
            }
            else if (cellOrHeader === 'cell')
            {
                if (readOnly && currentMode === 'navigation')
                {
                    if (ctrlKey && selection && selectionMode == 'cell' && multipleSelection)
                    {
                        return 'SELECT_COLUMN';
                    }
                    else if (shiftKey && selection && ((selectionMode == 'cell' && multipleSelection) || selectionMode == 'row'))
                    {
                        return 'SELECT_ROW';
                    }
                }
                else if (currentMode === 'navigation')
                {
                    return 'DATA_ENTRY';
                }
            }
            break;
        case keyCodes.PAGEUP_KEY:
            if (currentMode !== 'edit')
            {
                return 'FOCUS_ROW_FIRST';
            }
            break;
        case keyCodes.PAGEDOWN_KEY:
            if (currentMode !== 'edit')
            {
                return 'FOCUS_ROW_LAST';
            }
            break;
        case keyCodes.END_KEY:
            if (currentMode !== 'edit')
            {
                return 'FOCUS_COLUMN_LAST';
            }
            break;
        case keyCodes.HOME_KEY:
            if (currentMode !== 'edit')
            {
                return 'FOCUS_COLUMN_FIRST';
            }
            break;
        case keyCodes.LEFT_KEY:
            if (currentMode === 'actionable')
            {
                return 'NO_OP';
            }
            else if (currentMode !== 'edit')
            {
                if (shiftKey && selection && selectionMode == 'cell' && multipleSelection)
                {
                    return 'SELECT_EXTEND_LEFT';
                }
                if (ctrlKey && cellOrHeader != 'column')
                {
                    return 'FOCUS_ROW_HEADER';
                }
                return 'FOCUS_LEFT';
            }
            break;
        case keyCodes.UP_KEY:
            if (currentMode === 'actionable')
            {
                return 'NO_OP';
            }
            else if (currentMode !== 'edit')
            {
                if (shiftKey && selection && multipleSelection)
                {
                    return 'SELECT_EXTEND_UP';
                }
                if (ctrlKey && cellOrHeader != 'row')
                {
                    return 'FOCUS_COLUMN_HEADER';
                }
                return 'FOCUS_UP';
            }
            break;
        case keyCodes.RIGHT_KEY:
            if (currentMode === 'actionable')
            {
                return 'NO_OP';
            }
            else if (currentMode !== 'edit')
            {
                if (shiftKey && selection && selectionMode == 'cell' && multipleSelection)
                {
                    return 'SELECT_EXTEND_RIGHT';
                }
                if (ctrlKey && cellOrHeader != 'column')
                {
                    return 'FOCUS_ROW_END_HEADER';
                }
                return 'FOCUS_RIGHT';
            }
            break;
        case keyCodes.DOWN_KEY:
            if (currentMode === 'actionable')
            {
                return 'NO_OP';
            }
            else if (currentMode !== 'edit')
            {
                if (shiftKey && selection && multipleSelection)
                {
                    return 'SELECT_EXTEND_DOWN';
                }
                if (ctrlKey && cellOrHeader != 'row')
                {
                    return 'FOCUS_COLUMN_END_HEADER';
                }
                return 'FOCUS_DOWN';
            }
            break;
        case keyCodes.F2_KEY:
            if (cellOrHeader != 'cell')
            {
                return 'ACTIONABLE';
            }
            else if (readOnly && currentMode === 'navigation')
            {
                return 'EDITABLE';
            }
            else if (!readOnly && currentMode == 'navigation')
            {
                return 'EDIT';
            }
            break;
        case keyCodes.F8_KEY:
            if (shiftKey && selection && multipleSelection)
            {
                return 'SELECT_DISCONTIGUOUS';
            }
            break;
        case keyCodes.F10_KEY:
            if (shiftKey)
            {
                return 'NO_OP';
            }
            break;
        case keyCodes.V_KEY:
            if (currentMode === 'navigation' && ctrlKey  && rowMove)
            {
                return 'PASTE';
            }
            if (!readOnly && currentMode === 'navigation')
            {
                return 'DATA_ENTRY';
            }
            break;
        case keyCodes.X_KEY:
            if (currentMode === 'navigation' && ctrlKey && rowMove)
            {
                return 'CUT';
            }
            if (!readOnly && currentMode === 'navigation')
            {
                return 'DATA_ENTRY';
            }
            break;

        case keyCodes.SHIFT_KEY:
        case keyCodes.CTRL_KEY:
        case keyCodes.ALT_KEY:
            break;
        case keyCodes.A_KEY:
            if (ctrlKey && selection && multipleSelection)
            {
                return 'SELECT_ALL';
            }
        case keyCodes.NUM5_KEY:
            if (ctrlKey && altKey)
            {
                return 'READ_CELL';
            }
        default:
            if ((keyCode < keyCodes.F1_KEY || keyCode > keyCodes.F15_KEY) && !readOnly && currentMode === 'navigation' && cellOrHeader == 'cell')
            {
                return 'DATA_ENTRY';
            }
            break;
    }
    return 'NO_OP';
};
(function() {
var ojDataGridMeta = {
  "properties": {
    "bandingInterval": {
      "type": "object",
      "properties": {
        "row": {
          "type": "number"
        },
        "column": {
          "type": "number"
        }
      }
    },
    "cell": {
      "type": "object",
      "properties": {
        "className": {
          "type": "string|function"
        },
        "renderer": {},
        "style": {
          "type": "string|function"
        }
      }
    },
    "currentCell": {
      "type": "object",
      "writeback": true
    },
    "data": {},
    "dnd": {
      "type": "object",
      "properties": {
        "reorder": {
          "type": "object",
          "properties": {
            "row": {
              "type": "string",
              "enumValues": ["disable", "enable"]
            }
          }
        }
      }
    },
    "editMode": {
        "type": "string",
        "enumValues": ["none", "cellNavigation", "cellEdit"]
    },
    "gridlines": {
      "type": "object",
      "properties": {
        "horizontal": {
          "type": "string",
          "enumValues": ["visible", "hidden"]
        },
        "vertical": {
          "type": "string",
          "enumValues": ["visible", "hidden"]
        }
      }
    },
    "header": {
      "type": "object",
      "properties": {
        "column": {
          "type": "object",
          "properties": {
            "className": {
              "type": "string|function"
            },
            "renderer": {},
            "resizable": {
              "type": "object",
              "properties": {
                "width": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                },
                "height": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                }
              }
            },
            "sortable": {
              "type": "string|function",
              "enumValues": ["auto", "disable", "enable"]
            },
            "style": {
              "type": "string|function"
            }
          }
        },
        "columnEnd": {
          "type": "object",
          "properties": {
            "className": {
              "type": "string|function"
            },
            "renderer": {},
            "resizable": {
              "type": "object",
              "properties": {
                "width": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                },
                "height": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                }
              }
            },
            "sortable": {
              "type": "string|function",
              "enumValues": ["auto", "disable", "enable"]
            },
            "style": {
              "type": "string|function"
            }
          }
        },
        "row": {
          "type": "object",
          "properties": {
            "className": {
              "type": "string|function"
            },
            "renderer": {},
            "resizable": {
              "type": "object",
              "properties": {
                "width": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                },
                "height": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                }
              }
            },
            "sortable": {
              "type": "string|function",
              "enumValues": ["auto", "disable", "enable"]
            },
            "style": {
              "type": "string|function"
            }
          }
        },
        "rowEnd": {
          "type": "object",
          "properties": {
            "className": {
              "type": "string|function"
            },
            "renderer": {},
            "resizable": {
              "type": "object",
              "properties": {
                "width": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                },
                "height": {
                  "type": "string|function",
                  "enumValues": ["disable", "enable"]
                }
              }
            },
            "sortable": {
              "type": "string|function",
              "enumValues": ["auto", "disable", "enable"]
            },
            "style": {
              "type": "string|function"
            }
          }
        }
      }
    },
    "scrollPolicy": {
      "type": "string",
      "enumValues": ["auto", "loadMoreOnScroll", "scroll"]
    },
    "scrollPolicyOptions": {
      "type": "object",
      "properties": {
        "maxRowCount": {
          "type": "number"
        },
        "maxColumnCount": {
          "type": "number"
        }          
      }
    },
    "scrollPosition": {
      "writeback": true,        
      "type": "object",
      "properties": {
        "rowIndex": {
          "type": "number"
        },
        "columnIndex": {
          "type": "number"
        },
        "rowKey": {
          "type": "*"
        },
        "columnKey": {
          "type": "*"
        },
        "x": {
          "type": "number"
        },
        "y": {
          "type": "number"
        },
        "offsetX": {
          "type": "number"
        },
        "offsetY": {
          "type": "number"
        }
      }
    },
    "selection": {
      "type": "Array<object>",
      "writeback": true
    },
    "selectionMode": {
      "type": "object",
      "properties": {
        "row": {
          "type": "string",
          "enumValues": ["none", "single", "multiple"]
        },
        "cell": {
          "type": "string",
          "enumValues": ["none", "single", "multiple"]
        }
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "accessibleActionableMode": {
          "type": "string"
        },
        "accessibleColumnContext": {
          "type": "string"
        },
        "accessibleColumnEndHeaderContext": {
          "type": "string"
        },
        "accessibleColumnHeaderContext": {
          "type": "string"
        },
        "accessibleColumnSelected": {
          "type": "string"
        },
        "accessibleColumnSpanContext": {
          "type": "string"
        },
        "accessibleFirstColumn": {
          "type": "string"
        },
        "accessibleFirstRow": {
          "type": "string"
        },
        "accessibleLastColumn": {
          "type": "string"
        },
        "accessibleLastRow": {
          "type": "string"
        },
        "accessibleLevelContext": {
          "type": "string"
        },
        "accessibleMultiCellSelected": {
          "type": "string"
        },
        "accessibleNavigationMode": {
          "type": "string"
        },
        "accessibleRangeSelectModeOff": {
          "type": "string"
        },
        "accessibleRangeSelectModeOn": {
          "type": "string"
        },
        "accessibleRowCollapsed": {
          "type": "string"
        },
        "accessibleRowContext": {
          "type": "string"
        },
        "accessibleRowEndHeaderContext": {
          "type": "string"
        },
        "accessibleRowExpanded": {
          "type": "string"
        },
        "accessibleRowHeaderContext": {
          "type": "string"
        },
        "accessibleRowSelected": {
          "type": "string"
        },
        "accessibleRowSpanContext": {
          "type": "string"
        },
        "accessibleSelectionAffordanceBottom": {
          "type": "string"
        },
        "accessibleSelectionAffordanceTop": {
          "type": "string"
        },
        "accessibleSortAscending": {
          "type": "string"
        },
        "accessibleSortDescending": {
          "type": "string"
        },
        "accessibleStateSelected": {
          "type": "string"
        },
        "accessibleSummaryEstimate": {
          "type": "string"
        },
        "accessibleSummaryExact": {
          "type": "string"
        },
        "accessibleSummaryExpanded": {
          "type": "string"
        },
        "labelCut": {
          "type": "string"
        },
        "labelDisableNonContiguous": {
          "type": "string"
        },
        "labelEnableNonContiguous": {
          "type": "string"
        },
        "labelPaste": {
          "type": "string"
        },
        "labelResize": {
          "type": "string"
        },
        "labelResizeDialogSubmit": {
          "type": "string"
        },
        "labelResizeHeight": {
          "type": "string"
        },
        "labelResizeWidth": {
          "type": "string"
        },
        "labelSortCol": {
          "type": "string"
        },
        "labelSortColAsc": {
          "type": "string"
        },
        "labelSortColDsc": {
          "type": "string"
        },
        "labelSortRow": {
          "type": "string"
        },
        "labelSortRowAsc": {
          "type": "string"
        },
        "labelSortRowDsc": {
          "type": "string"
        },
        "msgFetchingData": {
          "type": "string"
        },
        "msgNoData": {
          "type": "string"
        }
      }
    }
  },
  "methods": {
    "getContextByNode": {}
  },
  "events": {
    "beforeCurrentCell": {},
    "beforeEdit": {},
    "beforeEditEnd": {},
    "resize": {},
    "scroll": {},
    "sort": {}
  },
  "extension": {
    _WIDGET_NAME: "ojDataGrid"
  }
};
oj.CustomElementBridge.registerMetadata('oj-data-grid', 'baseComponent', ojDataGridMeta);
oj.CustomElementBridge.register('oj-data-grid', {'metadata': oj.CustomElementBridge.getMetadata('oj-data-grid')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2012 Igor Vaynberg
 *
 * This software is licensed under the Apache License, Version 2.0 (the "Apache License") or the GNU
 * General Public License version 2 (the "GPL License"). You may choose either license to govern your
 * use of this software only upon the condition that you accept all of the terms of either the Apache
 * License or the GPL License.
 *
 * You may obtain a copy of the Apache License and the GPL License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 * http://www.gnu.org/licenses/gpl-2.0.html
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * Apache License or the GPL Licesnse is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the Apache License and the GPL License for
 * the specific language governing permissions and limitations under the Apache License and the GPL License.
 */

/**
 * @private
 */
  var _ComboUtils = {
    // native renderMode: marker class for generated options list
    GENERATED_OPTIONS_SELECTOR : "oj-select-options-generated",

    KEY:
    {
      TAB : 9,
      ENTER : 13,
      ESC : 27,
      SPACE : 32,
      LEFT : 37,
      UP : 38,
      RIGHT : 39,
      DOWN : 40,
      SHIFT : 16,
      CTRL : 17,
      ALT : 18,
      PAGE_UP : 33,
      PAGE_DOWN : 34,
      HOME : 36,
      END : 35,
      BACKSPACE : 8,
      DELETE : 46,

      isControl : function (e)
      {
        var k = e.which;
        switch (k)
        {
        case _ComboUtils.KEY.SHIFT:
        case _ComboUtils.KEY.CTRL:
        case _ComboUtils.KEY.ALT:
          return true;
        }
        if (e.metaKey || e.ctrlKey)
          return true;
        return false;
      },

      isFunctionKey : function (k)
      {
        k = k.which ? k.which : k;
        return k >= 112 && k <= 123;
      }
    },

    /*
     * The default fetch size from the data provider
     */
    DEFAULT_FETCH_SIZE: 15,

    /*
     * The default delay in milliseconds between when a keystroke occurs
     * and when a search is performed to get the filtered options.
     */
    DEFAULT_QUERY_DELAY: 70,

    ValueChangeTriggerTypes:
    {
      ENTER_PRESSED: "enter_pressed",
      OPTION_SELECTED: "option_selected",
      BLUR: "blur",
      SEARCH_ICON_CLICKED: "search_icon_clicked"
    },

    lastMousePosition: {x : 0, y : 0},
    nextUid: (function () {var counter = 1; return function () { return counter++; };}()),

    //TODO:
    scrollBarDimensions: null,

    //_ComboUtils
    /*
     * 4-10 times faster .each replacement
     * it overrides jQuery context of element on each iteration
     */
    each2: function(list, c)
    {
      var j = $.isFunction(list[0]) ? $(list[0]()): $(list[0]),
      i = -1,
      l = list.length;
      while (
        ++i < l
          && (j.context = j[0] = ($.isFunction(list[0]) ? list[i]():list[i]))
          && c.call(j[0], i, j) !== false //i=index, j=jQuery object
      )
      {};
      return list;
    },

    //_ComboUtils
    measureScrollbar: function()
    {
      var $template = $("<div class='oj-listbox-measure-scrollbar'></div>");
      $template.appendTo('body'); // @HTMLUpdateOK
      var dim =
        {
          width : $template.width() - $template[0].clientWidth,
          height : $template.height() - $template[0].clientHeight
        };
      $template.remove();
      return dim;
    },

    //_ComboUtils
    /*
     * Splits the string into an array of values, trimming each value.
     * An empty array is returned for nulls or empty
     */
    splitVal: function(string, separator)
    {
      var val,
      i,
      l;
      if (string === null || string.length < 1)
        return [];
      val = string.split(separator);
      for (i = 0, l = val.length; i < l; i = i + 1)
        val[i] = $.trim(val[i]);
      return val;
    },

    //_ComboUtils
    getSideBorderPadding: function(element)
    {
      return element.outerWidth(false) - element.width();
    },

    //_ComboUtils
    installKeyUpChangeEvent: function(element)
    {
      var key = "keyup-change-value";
      element.on("keydown", function ()
        {
          if ($.data(element, key) === undefined)
          {
            $.data(element, key, element.val());
          }
        });

      element.on("keyup", function (e)
        {
          if (e.which === _ComboUtils.KEY.ENTER)
          {
            // - select and combobox stop keyboard event propegation
            e.preventDefault();
            return;
          }
          var val = $.data(element, key);
          if (val !== undefined && element.val() !== val)
          {
            $.removeData(element, key);
            element.trigger("keyup-change");
          }
        });
    },

    //_ComboUtils
    getSearchText: function(event)
    {
      var searchText;
      // - start typing 1 letter on select box, but 2 letters displayed on searchbox
      //In case of chrome/IE, typed key is added on search element as we move focus from select element to search
      //But this is not happening on firefox and hence we need to set it as part of select element's event
      //and kill the event to avoid duplicate charecters on search field later in IE/chrome.
      //Dropdown popup will be opened on up/down/left/right arrows so excluding those as search text.

      var keycode = event.which || event.keyCode;
      if (event && event.type == "keydown" &&
        (keycode == 32 || // spacebar
          (keycode > 47 && keycode < 58) || // number keys
          (keycode > 64 && keycode < 91) || // letter keys
          (keycode > 95 && keycode < 112) || // numpad keys
          (keycode > 185 && keycode < 193) || // ;=,-./` (in order)
          (keycode > 218 && keycode < 223))) // [\]' (in order)
      {
        // Numpad keys return different keyCodes for the numbers
        // String.fromCharCode would return 'a' for '1' and so forth
        // Need to convert those keyCodes to regular number keyCodes
        if (keycode >= 96 && keycode <= 105) 
          keycode -= 48;
          
        searchText = String.fromCharCode(keycode);
        //keydown event always return uppercase letter
        if (!event.shiftKey)
          searchText = searchText.toLowerCase();
        // - select and combobox stop keyboard event propegation
        event.preventDefault();
      }
      return searchText;
    },

    //_ComboUtils
    /*
     * filters mouse events so an event is fired only if the mouse moved.
     *
     * filters out mouse events that occur when mouse is stationary but
     * the elements under the pointer are scrolled.
     */
    installFilteredMouseMove: function(element)
    {
      element.on("mousemove", function (e)
        {
          var lastpos = _ComboUtils.lastMousePosition;
          if (lastpos === undefined || lastpos.x !== e.pageX || lastpos.y !== e.pageY)
          {
            $(e.target).trigger("mousemove-filtered", e);
            _ComboUtils.lastMousePosition.x = e.pageX;
            _ComboUtils.lastMousePosition.y = e.pageY;
          }
        });
    },

    //_ComboUtils
    thunk: function(formula)
    {
      var evaluated = false,
      value;
      return function ()
        {
          if (evaluated === false)
          {
            value = formula();
            evaluated = true;
          }
          return value;
        };
    },

    //_ComboUtils
    _focus: function(widget, $el)
    {
      if ($el[0] === document.activeElement)
        return;

      // add busy state
      var resolveBusyState = _ComboUtils._addBusyState(widget.container, "setting focus");

      /* set the focus in a timeout - that way the focus is set after the processing
         of the current event has finished - which seems like the only reliable way
         to set focus */
      var timer = oj.TimerUtils.getTimer(40);
      timer.getPromise().then(function ()
        {
          var el = $el[0],
          pos = $el.val().length,
          range;
          $el.focus();

          /* make sure el received focus so we do not error out when trying to manipulate the caret.
             sometimes modals or others listeners may steal it after its set */
          if ($el.is(":visible") && el === document.activeElement)
          {
            /* after the focus is set move the caret to the end, necessary when we val()
               just before setting focus */
            if (el.setSelectionRange)
              el.setSelectionRange(pos, pos);
            else if (el.createTextRange)
            {
              range = el.createTextRange();
              range.collapse(false);
              range.select();
            }
          }

          resolveBusyState();
        // Set a 40 timeout. In voiceover mode, previous partial value was read. See 
        // This happens on ios Safari only, not Chrome. Setting a 40 timeout fixes the issue
        // on Safari in voiceover.
        });
    },

    //_ComboUtils
    getCursorInfo: function (el)
    {
      el = $(el)[0];
      var offset = 0;
      var length = 0;
      if ('selectionStart' in el)
      {
        offset = el.selectionStart;
        length = el.selectionEnd - offset;
      }
      else if ('selection' in document)
      {
        el.focus(); //Fixed???
        var sel = document.selection.createRange();
        length = document.selection.createRange().text.length;
        sel.moveStart('character', -el.value.length);
        offset = sel.text.length - length;
      }
      return {offset : offset, length : length};
    },

    //_ComboUtils
    killEvent: function(event)
    {
      event.preventDefault();
    },

    //_ComboUtils
    killEventImmediately: function(event)
    {
      event.preventDefault();
      event.stopImmediatePropagation();
    },

    //_ComboUtils
    defaultEscapeMarkup: function(markup)
    {
      var replace_map =
      {
        '\\' : '&#92;',
        '&' : '&amp;',
        '<' : '&lt;',
        '>' : '&gt;',
        '"' : '&quot;',
        "'" : '&#39;'
      };

      return String(markup).replace(/[&<>"'\\]/g, //@HTMLUpdateOK
        function (match) {
          return replace_map[match];
        });
    },

    //_ComboUtils
    /*
     * Produces a query function that works with a local array
     */
    local: function(options, optKeys)
    {
      var data = options, // data elements
          dataText,
          tmp,
          // function used to retrieve the text portion of a data item that is matched against the search
          text = function (item)
            {
              return "" + item['label'];
            };

      if ($.isArray(data))
      {
        tmp = data;
        data =
          {
            results : tmp
          };
      }
      if ($.isFunction(data) === false)
      {
        tmp = data;
        data = function ()
          {
            return tmp;
          };
      }
      var dataItem = data();
      //select with no options
      if (dataItem && dataItem.text)
      {
        text = dataItem.text;
        // if text is not a function we assume it to be a key name
        if (!$.isFunction(text))
        {
          // we need to store this in a separate variable because in the next step data gets reset
          // and data.text is no longer available
          dataText = dataItem.text;
          text = function (item)
            {
              return item[dataText];
            };
        }
      }
      return function (query)
        {
          var t = query.term,
          filtered =
            {
              results : []
            };

          // if optionsKeys is set, we need to do the key mapping, don't return
          if (t === "" && !optKeys)
          {
            query.callback(data());
            return;
          }

          if(data()){
            _ComboUtils.each2($(data().results), function (i, datum)
              {
                _ComboUtils._processData(query, datum, filtered.results,
                  optKeys, true, text);
              });
          }
          query.callback(filtered);
        };
    },

    // native renderMode
    createOptionTag : function (depth, value, label, formatFunc)
    {
      var node = $("<option>");
      node.addClass("oj-listbox-result oj-listbox-result-selectable oj-listbox-results-depth-" + depth);

      //option label
      node.attr("role", "option");
      node.attr("id", "oj-listbox-result-label-" + _ComboUtils.nextUid());

      node.text(formatFunc(label));
      node.attr("value", value);

      return node;
    },

    // native renderMode
    createOptgroupTag : function (container, label, formatFunc)
    {
      var node = $("<optgroup>");
      node.addClass("oj-listbox-results-sub");
      node.attr("label", formatFunc(label));

      container.addClass("oj-listbox-result-with-children");
      return node;
    },

    // native renderMode
    // This method turns a list of <ul>s and <li>s into <optgroup>s and <option>s
    listPopulateResults : function (container, list, formatFunc)
    {
      var populate = function (container, list, depth)
      {
        var node, li, label, ul;

        list.each(function(index) {
          li = $(this);

          if (li.is("li"))
          {
            // process <li> with children
            if (li.children("ul").length > 0)
            {
              //get the <li> text only dont include its descendants
              label = li.contents().filter(
                function() {
                  return this.nodeType !== 1 || this.tagName.toLowerCase() !== 'ul'
                })
                .text();

              node = _ComboUtils.createOptgroupTag(container, label, formatFunc);

              ul = li.children("ul");
              populate(node, ul.children(), depth + 1);
            }
            // process <li> without children
            else
            {
              node = _ComboUtils.createOptionTag(depth, li.attr("oj-data-value"), li.text(), formatFunc);
            }
            node.appendTo(container); // @HTMLUpdateOK
          }
        });
      }

      populate(container, list, 0);
    },

    // native renderMode
    // This method turns <oj-optgroup>s and <oj-option>s into <optgroup>s and <option>s
    ojOptionPopulateResults : function (container, ojOptions, formatFunc)
    {
      var populate = function (container, ojOptions, depth)
      {
        var node, ojOption, label;

        ojOptions.each(function(index) {
          ojOption = $(this);

          if (ojOption.is("oj-option"))
          {
            label = ojOption.text() || ojOption.attr("label");
            node = _ComboUtils.createOptionTag(depth, ojOption.prop("value"), label , formatFunc);
          }
          else if (ojOption.is("oj-optgroup"))
          {
            label = ojOption.text() || ojOption.attr("label");
            node = _ComboUtils.createOptgroupTag(container, label, formatFunc);
            populate(node, ojOption.children(), depth + 1);
          }
          // the option element created for the placeholder in native mode
          else if (ojOption.is("option"))
          {
            node = ojOption;
          }
          node.appendTo(container); // @HTMLUpdateOK
        });
      }

      populate(container, ojOptions, 0);
    },

    // native renderMode
    lookupOptionKeys : function (result, optionsKeys, key)
    {
      key = optionsKeys[key] || key;
      return result[key];
    },

    // native renderMode
    // This method turns a JSON object into <optgroup>s and <option>s
    arrayPopulateResults : function (container, arrlist, formatFunc, optionsKeys)
    {
      var populate = function (container, arrlist, depth, optionsKeys)
      {
        var item, node, children, label, value;
        for (var i = 0, l = arrlist.length; i < l; i++)
        {
          item = arrlist[i];

          // process children
          children = _ComboUtils.lookupOptionKeys(item, optionsKeys, "children");
          label = _ComboUtils.lookupOptionKeys(item, optionsKeys, "label");

          if (children && children.length > 0)
          {
            node = _ComboUtils.createOptgroupTag(container, label, formatFunc);
            populate(node, children, depth + 1, optionsKeys['childKeys'] || {});
          }
          // without children
          else
          {
            value = _ComboUtils.lookupOptionKeys(item, optionsKeys, "value");
            node = _ComboUtils.createOptionTag(depth, value, label, formatFunc);
          }

          node.appendTo(container); // @HTMLUpdateOK
        }
      };

      populate(container, arrlist, 0, optionsKeys || {});
    },

    // native renderMode
    cleanupResults : function (container)
    {
      container.children().not('oj-option, oj-optgroup').remove();
      container.removeClass("oj-listbox-result-with-children");
    },

    //_ComboUtils
    _addBusyState: function (element, description)
    {
      var desc = "The component identified by '" + element.attr("id") + "' " + description;
      var busyStateOptions = {"description": desc};
      var busyContext = oj.Context.getContext(element[0]).getBusyContext();

      return busyContext.addBusyState(busyStateOptions);
    },

    _clearBusyState: function (resolveFunc)
    {
      if (resolveFunc)
        resolveFunc();
    },

    isDataProvider: function(data)
    {
      return (data && oj.DataProviderFeatureChecker)?
        oj.DataProviderFeatureChecker.isIteratingDataProvider(data) : false;
    },

    getDataProvider: function(options)
    {
      var dataProvider = options._dataProvider || options.options;
      return _ComboUtils.isDataProvider(dataProvider)? dataProvider : null;
    },

    clearDataProviderWrapper: function(widget)
    {
      widget.options._dataProvider = null;
    },

    // - need to be able to specify the initial value of select components bound to dprv
    //traversal using depth first search
    //return an oj.Option object if found otherwise return null
    findOption: function (arOpts, value)
    {
      for (var i = 0, len = arOpts.length; i < len; i++)
      {
        var ojOption = arOpts[i];
        if (ojOption['children'])
        {
          var result = _ComboUtils.findOption(ojOption['children'], value);
          if (result)
            return result;
        }
        else if (oj.Object.compareValues(value, ojOption['value']))
        {
          return ojOption;
        }
      }
      //not found
      return null;
    },

    findOptions: function (ojOptgroup, values)
    {
      var ojOptionArr = [];
      for (var i = 0; i < values.length; i++)
      {
        var option = _ComboUtils.findOption(ojOptgroup, values[i]);
        if (option)
          ojOptionArr.push(option);
      }

      return ojOptionArr;
    },

    // - oj.tests.input.combobox.testcombobox display value mismatch automation failure
    findOptionFromResult: function (context, val, data)
    {
      var queryResult = _ComboUtils.getLastQueryResult(context);
      var match;
      if (queryResult) {
        match = _ComboUtils.findOption(queryResult, val);
      }
      if (match) {
        return {'value': val, 'label': match['label']};
      }
      return data;
    },

    //merge value and valueOption, value wins if both are specified
    //return true if the value is specified and it's not contained in valueOptions
    mergeValueAndValueOptions: function (ojContext)
    {
      var value = ojContext.options.value;
      var resolveLater = false;

      //multiple
      if (ojContext.multiple) {
        var valueOptions = ojContext.options.valueOptions;
        //value specified
        if (value && value.length > 0) {
          //both value and valueOptions specified, find the option for the value
          var ojoptionArr;
          if (valueOptions && valueOptions.length) {
            ojoptionArr = _ComboUtils.findOptions(valueOptions, value);
          }
          //update valueOptions if more than one
          if (! ojoptionArr || ojoptionArr.length !== valueOptions.length) {
            //need to find out the label and setValueOptions later
            resolveLater = true;
          }
        }
        //value not specified
        else if (valueOptions) {
          _ComboUtils.syncValueWithValueOptions(ojContext, valueOptions, value);
        }
      }
      //single
      else {
        var valueOption = ojContext.options.valueOption;
        //value specified
        if (value !== null && value !== undefined) {
          //both value and valueOption specified, find the option for the value
          var ojoption;
          if (valueOption) {
            ojoption = _ComboUtils.findOption(valueOption, value);
          }
          //update valueOption
          if (! ojoption) {
            //need to find out the label and setValueOption later
            resolveLater = true;
          }
        }
        //value not specified
        else if (valueOption) {    
          _ComboUtils.syncValueWithValueOption(ojContext, valueOption, value);
        }
      }

      return resolveLater;
    },

    //single selection: keep value and valueOption in sync
    syncValueWithValueOption: function (ojContext, valueOption, value)
    {
      var newVal = valueOption ? valueOption['value'] : null;
      if (! oj.Object.compareValues(newVal, value)) {
        ojContext._SetValue(newVal, null, {doValueChangeCheck: false,
                                           '_context': {internalSet: true,
                                                        writeback: true}});
      }
    },

    //multiple selection: keep value in sync with valueOptions
    syncValueWithValueOptions: function (ojContext, valueOptions, value)
    {
      if (valueOptions) {
        var newVal = [];
        for (var i = 0; i < valueOptions.length; i++) {
          newVal.push(valueOptions[i]['value']);
        }

        if (! oj.Object.compareValues(newVal, value)) {
          ojContext._SetValue(newVal, null, {doValueChangeCheck: false,
                                          '_context': {internalSet: true,
                                                       writeback: true}});
        }
      }
    },

    // - need to be able to specify the initial value of select components bound to dprv
    setValueOptions : function (ojContext, valueOptions)
    {
      var context = {
        internalSet: true,
        changed: true,
        writeback: true
      };

      var newValueOptions;
      if (ojContext.multiple) {
        if (valueOptions && valueOptions.length) {
          newValueOptions = [];
          for (var i = 0; i < valueOptions.length; i++) {
            newValueOptions.push({"value": valueOptions[i]['value'], 
                                  "label": valueOptions[i]['label']});            
          }
        }
        else {
          newValueOptions = valueOptions;
        }
        ojContext.option('valueOptions', newValueOptions, {'_context': context});

        // - placeholder is not displayed after removing selections from select many
        //update internal valueOptions
        if (ojContext.combobox && ojContext.combobox.opts) {
          ojContext.combobox.opts.valueOptions = newValueOptions;
        }
        else if (ojContext.select && ojContext.select.opts) {
          ojContext.select.opts.valueOptions = newValueOptions;
        }
      }
      else {
        if (Array.isArray(valueOptions))
          valueOptions = valueOptions[0];

        if (valueOptions) {
          newValueOptions = {"value": valueOptions['value'], "label": valueOptions['label']};
          ojContext.option('valueOption', 
                           {"value": valueOptions['value'], "label": valueOptions['label']},
                           {'_context': context});
        }
        else {
          newValueOptions = valueOptions;
        }
        ojContext.option('valueOption', newValueOptions, {'_context': context});

        // - placeholder is not displayed after removing selections from select many
        //update internal valueOption
        if (ojContext.combobox && ojContext.combobox.opts) {
          ojContext.combobox.opts.valueOption = newValueOptions;
        }
        else if (ojContext.select && ojContext.select.opts) {
          ojContext.select.opts.valueOption = newValueOptions;
        }
      }
    },

    //update display label(s) and valueOption(s) after value was set
    updateValueOptions: function (context)
    {
      // - qunit: failure across browsers in select tests
      if (! context)
        return;

      var element = context.datalist ? context.datalist : context.opts.element;
      context.opts.initSelection.call(null, element, function(selected) {
        var multiple = context.ojContext.multiple;

        //in combobox, values may be new entries 
        if (selected === undefined && context._classNm === "oj-combobox") {
          var value = context.ojContext.options.value;
          if (multiple) {
            selected = [];
            for (var i = 0; i < value.length; i++) {
              selected.push(context.opts["manageNewEntry"](value[i]));
            }
          }
          else {
            selected = context.opts["manageNewEntry"](value);
          }
        }
        if (selected) {
          context.setValOpts(selected);
          if (multiple) {
            context._updateSelection(selected);
          }
          else {
            context._updateSelectedOption(selected);
          }
        }
      });
    },

    /*  - need to be able to specify the initial value of select components bound to dprv
     * If both dataProvider and valueOption[s] are specified, use valueOption[s[ for display values.
     * If valueOption[s] is not specified or a selected value is missing then we will fetch the real data 
     * from the dataProvider like before.
     * return true if valueOption[s] is applied, false otherwise
     */
    applyValueOptions: function(context, options)
    {
      if (context && ! context.ojContext._resolveValueOptionsLater &&
          (context._classNm === "oj-combobox" || context._classNm === "oj-select")) {

        var valueOptions;
        if (context.ojContext.multiple) {
          valueOptions = options.valueOptions;

          // - placeholder is not displayed after removing selections from select many
          if (valueOptions && valueOptions.length == 0 && options.placeholder)
            return false;
        }
        else {
          valueOptions = options.valueOption;
        }
        if (valueOptions) {
          context._updateSelection(valueOptions);
          return true;
        }
      }
      return false;
    },

    /*
     * When dataProvider is used, wrap a ListViewDataProviderView around it
     * if optionsKeys is specified or the fetchByKeys method is not available
     * save the new wrapper or the original data provider
     */
    wrapDataProviderIfNeeded: function(widget, opts)
    {
      var data = widget.options.options;

      if (_ComboUtils.isDataProvider(data)) {
        var wrapper;
        var optionsKeys = widget.options.optionsKeys;

        if (optionsKeys && (optionsKeys.label != null || optionsKeys.value != null)) {
          if (!(data instanceof oj.ListDataProviderView)) {
            var mapFields = function(item) {
              var data = item['data'];
              var mappedItem = {};
              mappedItem['data'] = {};
              
              // copy all the fields
              for (var key in data) {
                mappedItem['data'][key] = data[key];
              }
              
              // map label field
              if (optionsKeys['label'] != null)
                mappedItem['data']['label'] = data[optionsKeys['label']];
              // map value field
              if (optionsKeys['value'] != null)
                mappedItem['data']['value'] = data[optionsKeys['value']];

              mappedItem['metadata'] = {'key': data[optionsKeys['value']]};
              
              return mappedItem;
            }; 
            //create ListDataProviderView with dataMapping
            wrapper = new oj.ListDataProviderView(data, {'dataMapping': {'mapFields': mapFields}});
          }
        }
        else if (! oj.DataProviderFeatureChecker.isFetchByKeys(data)) {
          if (!(data instanceof oj.ListDataProviderView)) {
            wrapper = new oj.ListDataProviderView(data);
          }
        }
        //save the data provider or wrapper
        if (wrapper) {
          widget.options._dataProvider = wrapper;
          //update internal wrapper
          if (opts) {
            opts._dataProvider = wrapper;
          }
        }
      }
    },

    /**
     * get display label. If label is missing, String(value) will be returned
     * @private
     */
    getLabel: function(item)
    {
      // - number converter with comboboxes fails on zero value entry
      //if label is null or undefined use value
      return item['label'] != null ? item['label'] : String(item['value']);
    },

    /**
     * data provider event handler
     * @private
     */
    _handleDataProviderEvents: function(widget, event)
    {
      if (event.type === "mutate") {
        if (event.detail.remove != null) {
          var data = event.detail.remove.data;
          var changed = false;

          var newVal = [].concat(widget.options.value);
          for (var i = 0; i < data.length; i++) {
            var index = newVal.indexOf(data[i].value);
            if (index >= 0) {
              newVal.splice(index, 1);
              changed = true;
            }
          }
          if (changed) {
            widget._setOption("value", newVal);
          }
        }
      }
      widget._setOption("options", widget.options.options);
    },

    /*
     * Add data provider event listeners
     */
    addDataProviderEventListeners: function(widget)
    {
      var dataProvider = _ComboUtils.getDataProvider(widget.options);
      if (dataProvider) {
        _ComboUtils.removeDataProviderEventListeners(widget);

        var dataProviderEventHandler = _ComboUtils._handleDataProviderEvents.bind(null, widget);
        widget._saveDataProviderEH = dataProviderEventHandler;

        dataProvider.addEventListener("mutate", dataProviderEventHandler);
        dataProvider.addEventListener("refresh", dataProviderEventHandler);

      }
    },

    /*
     * Remove data provider event listeners
     */
    removeDataProviderEventListeners: function(widget)
    {
      var dataProvider = _ComboUtils.getDataProvider(widget.options);
      var dataProviderEventHandler = widget._saveDataProviderEH;

      if (dataProvider != null && dataProviderEventHandler)
      {
        dataProvider.removeEventListener("mutate", dataProviderEventHandler);
        dataProvider.removeEventListener("refresh", dataProviderEventHandler);

        widget._saveDataProviderEH = undefined;
      }
    },

    //_ComboUtils
    //Add a loading indicator to the select box
    addLoadingIndicator: function(container)
    {
      // check if it's already added
      if (container._saveLoadingIndicator != null)
        return;

      //TODO: center icon
      var item = $(document.createElement("div"));
      item.uniqueId()
          .attr("role", "presentation")
          .addClass("oj-listbox-loading-icon-container");

      var icon = $(document.createElement("div"));
      icon.addClass("oj-icon oj-listbox-loading-icon");
      item.append(icon); // @HtmlUpdateOK

      container.prepend(item); // @HtmlUpdateOK

      container._saveLoadingIndicator = item;
    },

    //_ComboUtils
    //Remove the loading indicator 
    removeLoadingIndicator: function(container)
    {
      if (container._saveLoadingIndicator != null)
        container._saveLoadingIndicator.remove();

      container._saveLoadingIndicator = undefined;
    },

    //_ComboUtils
    //Add dropdown message such as filter further or no matches found
    addDropdownMessage: function(container, context, messageText)
    {
      // check if it's already added
      if (container._saveDropdownMessage)
        return;

      var msgParent = $(document.createElement("div"));
      msgParent.addClass("oj-listbox-filter-message-box");

      var message = $(document.createElement("div"));
      message.addClass("oj-listbox-filter-message-text oj-listbox-liveregion");
      message.attr({
        'role': 'region',
        'aria-live': 'polite',
      });

      var separator = $(document.createElement("div"));
      separator.addClass("oj-listbox-filter-message-separator");

      msgParent.append(message); // @HtmlUpdateOK
      msgParent.append(separator); // @HtmlUpdateOK
      container.prepend(msgParent); // @HtmlUpdateOK

      message.text(messageText);
      container._saveDropdownMessage = msgParent;

    },

    //_ComboUtils
    //Remove dropdown message
    removeDropdownMessage: function(container)
    {
      if (container._saveDropdownMessage) {
        container._saveDropdownMessage.remove();
        container._saveDropdownMessage = undefined;
      }
    },

    //_ComboUtils
    // Fetch from the data provider and filter the data locally until 
    // the end of data or fetch size has reached
    fetchFilteredData: function (dataProvider, fetchSize, context, query, dropdown) {
      var results = [];
      var fetchListParms = {
        'size': fetchSize
      };

      //check if data provider support filtering?
      var filters = dataProvider.getCapability("filter");
      var $co = oj.AttributeFilterOperator.AttributeOperator.$co;
      var filterThruDataProvider = false;

      //only filter thru data provider if it supports contains($co) operator
      if (filters && filters.length > 0) {
        for (var f = 0; f < filters.length; f++) {
          if (filters[f] == $co) {
            filterThruDataProvider = true;
            break;
          }
        }
      }

      //use dataprovider filtering if supported
      if (filterThruDataProvider && query && query.term) {
        //TODO test in the data mapping case to see if "label" works
        //Note: revisit when tree dataProvider is supported
        //for now if optionsKey is used, 'label' must specify in optionsKeys
        var optKeys = context.options.optionsKeys;
        var attrName;
        if (optKeys && optKeys['label']) {
          attrName = optKeys['label'];
        }
        else {
          attrName = 'label';
        }

        fetchListParms['filterCriterion'] = 
          {'op': $co, 'attribute': attrName, 'value': query.term};
      }

      var asyncIterable = dataProvider.fetchFirst(fetchListParms)[Symbol.asyncIterator]();

      function filterData(fetchResults) {
        if (fetchResults) {
          var data = fetchResults.value.data;
          if (data) {
            //skip filter locally if it is already done thru data provider
            if (filterThruDataProvider) {
              results = data;
            }
            else {
              for (var item, i = 0; i < data.length; i++) {
                item = data[i];
                if (! query || ! query.matcher || query.matcher(query.term, item.label, item))
                  results.push(item);
              }
            }
          }
          // clear message
          if (dropdown)
            _ComboUtils.removeDropdownMessage(dropdown);

          // not all results are fetched, display mesage for filtering further 
          if (dropdown && ! fetchResults.done && results.length >= fetchListParms['size']) {
            _ComboUtils.addDropdownMessage(dropdown, context, 
                                           context.getTranslatedString("filterFurther"));
            context._hasMore = true;
          }
          else {
            // fetch more
            if (! fetchResults.done && results.length < fetchListParms['size']) {
              return asyncIterable.next().then(filterData);
            }
            context._hasMore = false;
          }
        }
        return Promise.resolve(results);
      }

      return asyncIterable.next().then(filterData);
    },

    // used as rejected error in the fetchFromDataProvider method
    rejectedError: new Object(),

    //_ComboUtils
    // add busy state
    // display an animated gif if it is fetch initially
    // fetch from the data provider
    // display message for furthur filtering if not all results are fetched
    // if multiple queries are in progress, discard all but the last query
    fetchFromDataProvider : function (widget, options, query, fetchSize)
    {
      var context = widget.ojContext;
      var spinnerContainer = widget.selection;

      //add busy context
      if (! context._fetchResolveFunc) {
        context._fetchResolveFunc = _ComboUtils._addBusyState(widget.container, "fetching data");
      }

      //reject the previous promise to avoid out of order request
      if (context._saveRejectFunc)
        context._saveRejectFunc(_ComboUtils.rejectedError);

      //save the current reject function
      var remotePromise = new Promise(function(resolve, reject) {
        context._saveRejectFunc = reject;
      });

      //display spinning icon only for the initial fetch
      if (spinnerContainer && options.fetchType == "init") {
        _ComboUtils.addLoadingIndicator(spinnerContainer);
        options.fetchType = null;
      }

      //fetch data from dataProvider
      var fetchPromise = 
        _ComboUtils.fetchFilteredData(_ComboUtils.getDataProvider(options),
                                      (fetchSize || options.fetchSize || _ComboUtils.DEFAULT_FETCH_SIZE),
                                      context, query, widget.dropdown
                                     ).then(
          function (fetchResults) {
            if (spinnerContainer) {
              _ComboUtils.removeLoadingIndicator(spinnerContainer);
            }
            // - search not shown before typing a character
            context._resultCount = fetchResults? fetchResults.length : 0;
            return fetchResults;
          });

      //if multiple queries are in progress, discard all but the last query
      Promise.race([remotePromise, fetchPromise]).then(function(fetchResults) {
        //clear the reject function
        context._saveRejectFunc = null;

        // - search not shown before typing a character
        if (context._resolveSearchBoxLater) {
          widget._showSearchBox("");
        }

        query.callback({
          results: fetchResults
        });

        if (context._fetchResolveFunc) {
          //clear busy context
          context._fetchResolveFunc();
          context._fetchResolveFunc = null;
        }
      }, 
      function(error){
        //don't remove busy state if the reject coming from Promise.race
        if (error !== _ComboUtils.rejectedError && context._fetchResolveFunc) {
          query.callback();
          context._fetchResolveFunc();
          context._fetchResolveFunc = null;
        }
      });
    },

    //_ComboUtils
    //fetch first block of data from the data provider
    fetchFirstBlockFromDataProvider : function (container, dataProvider, fetchSize)
    {
      //add busy context
      var fetchResolveFunc = _ComboUtils._addBusyState(container, "fetching selected data");

      var fetchListParms = {
        'size': (fetchSize || _ComboUtils.DEFAULT_FETCH_SIZE)
      };
      var asyncIterable = dataProvider.fetchFirst(fetchListParms)[Symbol.asyncIterator]();
      return asyncIterable.next().then(
        function (fetchResults) {
          _ComboUtils._clearBusyState(fetchResolveFunc);
          return fetchResults.value.data;
        }, 
        function(error){
          _ComboUtils._clearBusyState(fetchResolveFunc);
          return null;
        });
    },

    //_ComboUtils
    //fetch the data row by its key("value")
    fetchByKeyFromDataProvider : function (container, dataProvider, query)
    {
      if (! oj.DataProviderFeatureChecker.isFetchByKeys(dataProvider))
        return;

      //add busy context
      var fetchResolveFunc = _ComboUtils._addBusyState(container, "fetching selected data");

      //fetch the data row by its key("value")
      dataProvider.fetchByKeys({keys:query.value}).then(function(fetchResults) {
        var values = [];
        fetchResults.results.forEach(function(val, key, map) {
          values.push(val.data);
        });

        query.callback({
          results: values
        });
        _ComboUtils._clearBusyState(fetchResolveFunc);
      }, 
      function(error) {
        query.callback();
        _ComboUtils._clearBusyState(fetchResolveFunc);
      });
    },

    //_ComboUtils
    //check if the specified value is in the dataProvider
    //return a promise with the following results
    //1) null: value is invalid or reject by fetchByKeyFromDataProvider
    //2) array of valid values
    validateFromDataProvider : function (container, dataProvider, value)
    {
      return new Promise(function(resolve, reject) {
        _ComboUtils.fetchByKeyFromDataProvider(container, dataProvider,
          {value: Array.isArray(value)? value : [value],
           callback: function(data) {
             var results = null;
             // - need to be able to specify the initial value of select components bound to dprv
             if (data && data.results.length) {
               results = {};
               results.value = [];
               results.valueOptions = [];
               for (var i = 0; i < data.results.length; i++) {
                 results.valueOptions.push(data.results[i]);
                 results.value.push(data.results[i].value);
               }
             }
             resolve(results);
         }});
      });
    },

    //_ComboUtils
    /*
     * Produces a query function that works with a remote data
     */
    remote: function(options, optKeys)
    {
      return function(query)
      {
        var context = {
          "component": this.ojContext
        };
        if (query.value) {
          context["value"] = query.value;
        } else {
          context["term"] = query.term || "";
        }

        options(context).then(function(data) {
          var filtered = {
            results: []
          };

          if (data) {
            _ComboUtils.each2($(data), function(i, datum)
            {
              _ComboUtils._processData(query, datum, filtered.results, optKeys,
                false);
            });
          }

          query.callback(filtered);
        });
      };
    },

    //_ComboUtils
    /*
     * Maps the optionsKeys and options array and creates the array of
     * Label-Value objects. If options array is local data then
     * it filters the result array based on the term entered in the search field.
     */
    _processData: function(query, datum, collection, keys, isLocal, text)
    {
      var group,
        attr;

      datum = datum[0];

      // key mappings
      if (!datum['label'] && (keys && keys['label'])) {
        datum['label'] = datum[keys['label']];
      }
      if (!datum['value'] && (keys && keys['value'])) {
        datum['value'] = datum[keys['value']];
      }
      if (!datum['children'] && (keys && keys['children'])) {
        datum['children'] = datum[keys['children']];
        delete datum[keys['children']];
      }

      if (datum['children'])
      {
        group = {};
        for (attr in datum)
        {
          if (datum.hasOwnProperty(attr))
            group[attr] = datum[attr];
        }
        group.children = [];
        _ComboUtils.each2($(datum['children']), function(i, childDatum)
        {
          _ComboUtils._processData(query, childDatum, group.children,
            ((keys && keys['childKeys']) ? keys['childKeys'] : null),
            isLocal, text);
        });

        // - group labels participate in the filtering
        //Reverted. In the nested data case, group may be selectable. Without putting the
        //group data in the collection, we will find no match and new entry may be created for combobox.
        //For group that has matching children, we need to call the matcher on the group label
        //to update the matches array. This needs to be done before checking the group.children.length.
        if (!isLocal || query.matcher(query.term, text(group), datum) || group.children.length)
        {
          collection.push(group);
        }
      }
      else
      {
        if (!isLocal || query.matcher(query.term, text(datum), datum))
        {
          collection.push(datum);
        }
      }
    },

    //_ComboUtils
    checkFormatter: function (ojContext, formatter, formatterName)
    {
      if ($.isFunction(formatter))
        return true;
      if (!formatter)
        return false;

      throw new Error(formatterName + " must be a function or a false value");
    },

    //_ComboUtils
    /*
     * Creates a new class
     */
    clazz: function(SuperClass, methods)
    {
      var constructor = function ()  {};
      oj.Object.createSubclass(constructor, SuperClass, '');
      constructor.prototype = $.extend(constructor.prototype, methods);
      return constructor;
    },

    //_ComboUtils
    LAST_QUERY_RESULT: "last-query-result",
    getLastQueryResult: function(context)
    {
      var queryResult = $.data(context.container,
        context._classNm + "-" + _ComboUtils.LAST_QUERY_RESULT);

      return queryResult;
    },
    saveLastQueryResult: function(context, queryResult)
    {
      $.data(context.container,
        context._classNm + "-" + _ComboUtils.LAST_QUERY_RESULT, queryResult);
    }
  };


  var _AbstractOjChoice = _ComboUtils.clazz(Object,
    {
      //_AbstractOjChoice
      _bind : function (func)
      {
        var self = this;
        return function ()
        {
          func.apply(self, arguments);
        };
      },

      //_AbstractOjChoice
      _customOptionRenderer : function (elems)
      {
        var elem,
            self = this;

        elems.each(function(index) {
          elem = $(this);

          if (elem.is("oj-option"))
          {
            elem.wrap("<li></li>"); // @HTMLUpdateOK
          }
          else if (elem.is("oj-optgroup"))
          {
            elem.wrap("<li></li>"); // @HTMLUpdateOK
            self._customOptionRenderer(elem.children());
            elem.children().wrapAll("<ul class='oj-listbox-result-sub'></ul>"); // @HTMLUpdateOK

          }
        })
      },

      //_AbstractOjChoice
      _init : function (opts)
      {
        var results,
        search,
        className = this._classNm,
        elemName = this._elemNm,
        resultsSelector = ".oj-listbox-results";

        this.ojContext = opts.ojContext;
        this.opts = opts = this._prepareOpts(opts);
        this.id = opts.id;
        this.headerInitialized = false;
        this.isOjOption = this.ojContext._IsCustomElement() && !opts['options'] &&
                          opts.element.find("oj-option").length > 0;

        // destroy if called on an existing component
        if (opts.element.data(elemName) !== undefined &&
          opts.element.data(elemName) !== null)
        {
          opts.element.data(elemName)._destroy();
        }
        this.container = this._createContainer();

        // - ojselect - rootAttributes are not propagated to generated jet component
        var rootAttr = opts.rootAttributes;
        this.containerId = (rootAttr && rootAttr.id) ?
          rootAttr.id :
          "ojChoiceId_" + (this._getAttribute("id") || "autogen" + _ComboUtils.nextUid());

        this.containerSelector = "#" + this.containerId.replace(/([;&,\.\+\*\~':"\!\^#$%@\[\]\(\)=>\|])/g, '\\$1'); //@HTMLUpdateOK
        this.container.attr("id", this.containerId);
        // cache the body so future lookups are cheap
        this.body = _ComboUtils.thunk(function ()
          {
            return opts.element.closest("body");
          }
          );
        this.container.attr("style", this._getAttribute("style"));
        this.elementTabIndex = this._getAttribute("tabindex");

        // 'opts.element' is initialized in _setup() menthod in component files
        // ojcombobox.js, ojselect.js and ojInputSearch.js.

        // swap container for the element
        this.opts.element
        .data(elemName, this)
        .attr("tabindex", "-1")
        .before(this.container);  // @HtmlUpdateOk
        this.container.data(elemName, this);
        this.dropdown = this.container.find(".oj-listbox-drop");
        this.dropdown.data('ojlistbox', this);

        // - let the ojselect popup accept the custom css class name from the component
        this._setPickerAttributes(opts.pickerAttributes);

        // link the shared dropdown dom to the target component instance
        var containerId = this.containerId;
        this.dropdown.attr("data-oj-containerid", containerId);

        this.results = results = this.container.find(resultsSelector);
        this.results.on("click", _ComboUtils.killEvent);

        // - oghag missing label for ojselect and ojcombobox
        var label = this.ojContext._GetLabelElement();
        if (label) {
          var labelId = label.attr("id");
          if (labelId)
            results.attr("aria-labelledby", labelId);
          else
            results.attr("aria-label", label.text());
        }
        else {
          var alabel = this.ojContext.element.attr("aria-label");
          if (alabel)
            results.attr("aria-label", alabel);
        }

        // if html ul element is provided, use it instead
        if (opts['list'] && $('#' + opts['list']).is("ul"))
        {
          var dropdownList = $('#' + opts['list']);
          this.dropdownListParent = dropdownList.parent();
          dropdownList.addClass("oj-listbox-results").attr("role", "listbox");
          this.results.replaceWith(dropdownList); //@HTMLUpdateOK
          this.results = results = this.container.find(resultsSelector);
          this.results.css("display", '');
        }

        // custom element syntax with oj-option elements
        if (this.isOjOption)
        {
          var elems = opts.element.children();
          this._customOptionRenderer(elems);
          this.results.append(opts.element.children()); //@HTMLUpdateOK
          this.datalist = this.results;
        }

        if (className == "oj-select")
          search = this.container.find("input.oj-listbox-input");
        else
          search = this.container.find("input." + className + "-input");
        this.search = search;

        this.queryCount = 0;
        this.resultsPage = 0;
        this.context = null;

        // initialize the container
        this._initContainer();
        this.container.on("click", _ComboUtils.killEvent);
        _ComboUtils.installFilteredMouseMove(this.results);
        this.dropdown.on("mousemove-filtered touchstart touchmove touchend", resultsSelector, this._bind(this._highlightUnderEvent));
        // do not propagate change event from the search field out of the component
        $(this.container).on("change", "." + className + "-input", function (e)
        {
          e.stopPropagation();
        }
        );
        $(this.dropdown).on("change", "." + className + "-input", function (e)
        {
          e.stopPropagation();
        }
        );

        var self = this;
        _ComboUtils.installKeyUpChangeEvent(search);
        search.on("keyup-change input paste", this._bind(this._updateResults));
        search.on("focus", function ()
        {
          search.addClass(className + "-focused");

          if (className !== "oj-select")
            self.container.addClass("oj-focus");
        }
        );
        search.on("blur", function ()
        {
          search.removeClass(className + "-focused");

          if (className !== "oj-select")
            self.container.removeClass("oj-focus");
        }
        );
        this.dropdown.on("mouseup", resultsSelector, this._bind(function (e)
          {
            if ($(e.target).closest(".oj-listbox-result-selectable").length > 0)
            {
              this._highlightUnderEvent(e);
              this._selectHighlighted(null, e);
            }
          }
          ));
        // trap all mouse events from leaving the dropdown. sometimes there may be a modal that is listening
        // for mouse events outside of itself so it can close itself. since the dropdown is now outside the combobox's
        // dom it will trigger the popup close, which is not what we want
        this.dropdown.on("click mouseup mousedown", function (e)
        {
          e.stopPropagation();
        }
        );
        if ($.isFunction(this.opts.initSelection))
        {
          ///support ko options-binding
          //init dataProvider fetchType
          this.opts.fetchType = "init";
          var valOpts = this.getValOpts();
          if ((this.ojContext.multiple && (! valOpts || valOpts.length == 0)) ||
              (! this.ojContext.multiple && ! valOpts)) {
            valOpts = null;
          }
          this._initSelection(valOpts);
        }
        var disabled = opts.element.prop("disabled");
        if (disabled === undefined)
          disabled = false;
        this._enable(!disabled);
        var readonly = opts.element.prop("readonly");
        if (readonly === undefined)
          readonly = false;
        this._readonly(readonly);
        // Calculate size of scrollbar
        _ComboUtils.scrollBarDimensions = _ComboUtils.scrollBarDimensions ||
                                          _ComboUtils.measureScrollbar();
        this.autofocus = opts.element.prop("autofocus");
        opts.element.prop("autofocus", false);
        if (this.autofocus)
          this._focus();

        _ComboUtils.addDataProviderEventListeners(opts.ojContext);
      },

      //_AbstractOjChoice
      _clickAwayHandler: function(event)
      {

        var dropdown = this.dropdown,
            self;

        if ($(event.target).closest(dropdown).length ||
            $(event.target).closest("#"+dropdown.attr("data-oj-containerid")).length)
                    return;

                if (dropdown.length > 0) {
                    self = dropdown.data('ojlistbox');
                    self.close(event);
                }
      },

      //_AbstractOjChoice
      _surrogateRemoveHandler: function()
      {
        if (this.dropdown)
          this.dropdown.remove();
      },

      //_AbstractOjChoice
      _destroy : function ()
      {
        var closeDelayTimer = this._closeDelayTimer;
        if (!isNaN(closeDelayTimer))
        {
          delete this._closeDelayTimer;
          window.clearTimeout(closeDelayTimer);
        }

        var element = this.opts.element,
            ojcomp = element.data(this._elemNm);

        this.close();
        if (this.propertyObserver)
        {
          delete this.propertyObserver;
          this.propertyObserver = null;
        }

        // 'results' is initialized in _init() method and it can not be changed by an external developer.

        // clean up the ul list
        if (this.opts['list'] && this.results) {
          this._cleanupList(this.results);
          // Move to original parent
          if (this.dropdownListParent)
            this.dropdownListParent.append(this.results); // @HtmlUpdateOk
        }
        else if (this.isOjOption && this.results)
        {
          this._unwrapOjOptions(this.results);
          this.opts.element.append(this.results.children()); // @HtmlUpdateOk
        }

        if (ojcomp !== undefined)
        {
          // Move the element outside the container to its original place
          ojcomp.container.after(element); //@HTMLUpdateOK
          ojcomp.container.remove();
          ojcomp.dropdown.remove();
          element
          .removeAttr("aria-hidden")
          .removeData(this._elemNm)
          .off("." + this._classNm)
          .prop("autofocus", this.autofocus || false);
          if (this.elementTabIndex)
          {
            element.attr(
            {
              tabindex : this.elementTabIndex
            }
            );
          }
          else
          {
            element.removeAttr("tabindex");
          }
          element.show();
        }
      },

      //_AbstractOjChoice
      /*
       * unwrap oj-option and oj-optgroup
       */
      _unwrapOjOptions : function (list)
      {
        var children = list.children(),
            elem,
            self = this;

        children.each(function(index) {
          elem = $(this);

          if (elem.is("li"))
          {
            // oj-optgroup
            if (elem.hasClass("oj-listbox-result-with-children") || elem.children("oj-optgroup").length > 0)
            {
              // if the oj-optgroup went through the populateResults call
              if (elem.hasClass("oj-listbox-result-with-children"))
                elem.children(".oj-listbox-result-label").remove();
              // unwrap children
              self._unwrapOjOptions(elem.find("oj-optgroup").children());
              // unwrap the <ul>
              elem.find("oj-optgroup").children().children().unwrap();
              // unwrap the <li>
              elem.find("oj-optgroup").unwrap();
            }
            else
            {
              // unwrap the <li>
              // if the oj-option went through the populateResults call
              if (elem.hasClass("oj-listbox-result"))
                elem.find("oj-option").unwrap().unwrap();
              else
                elem.find("oj-option").unwrap();
            }
          }
        });
      },

      //_AbstractOjChoice
      /*
       * Clean up the html list provided by app
       */
      _cleanupList : function (list)
      {
        if (list && list.is("ul"))
        {
          list.removeClass("oj-listbox-results oj-listbox-result-sub");
          list.removeAttr("role");
          for (var i = list.children().length - 1; i>=0; i--) {
            this._cleanupList($(list.children()[i]));
          }
        } else if (list.is("li"))
        {
          if (list.hasClass("oj-listbox-placeholder") ||
              list.hasClass("oj-listbox-no-results"))
            list.remove();

          // remove added li classes starts with oj-listbox-
          if (list.attr('class'))
            list.attr('class',list.attr('class').replace(/\oj-listbox-\S+/g, ''));

          // remove wrapping div
          var wrappingDiv = list.children(".oj-listbox-result-label");
          if (wrappingDiv)
            wrappingDiv.contents().unwrap();

          if (list.css('display') == 'none')
            list.css('display', '');

          this._cleanupList(list.children("ul"));
        }
      },

      //_AbstractOjChoice
      /*
       * Processes option/optgroup/li element and return data object
       */
      _optionToData : function (element)
      {
        if (element.is("option") || element.is("oj-option"))
        {
          return {
            value : element.prop("value") || element.attr("value"),
            label : element.text() || element.attr("label"),
            element : element.get(),
            css : element.attr("class"),
            disabled : element.prop("disabled"),
            locked : (element.attr("locked") === "locked") || (element.data("locked") === true)
          };
        }
        else if (element.is("optgroup") || element.is("oj-optgroup"))
        {
          return {
            label : element.prop("label") || element.attr("label"),
            disabled : element.prop("disabled"),
            children : [],
            element : element.get(),
            css : element.attr("class")
          };
        }
        else if (element.is("li"))
        {
          var itemLabel,
              itemValue,
              disabled,
              groupData = null,
              elemChildren = element.children();
          if (elemChildren && elemChildren.length > 0 && elemChildren.is("ul"))
          {
            itemLabel = element.attr("oj-data-label") ? element.attr("oj-data-label") :
                        element.clone().children().remove().end().text().trim();
            itemValue = element.attr("oj-data-value");
            groupData = [];
          }
          else if (elemChildren && elemChildren.length > 0 && elemChildren.is("oj-optgroup"))
          {
            itemLabel = elemChildren.prop("label");
            disabled = elemChildren.prop("disabled");
            groupData = [];
          }
          else
          {
            var ojOptionItem = element.find("oj-option");
            itemLabel = element.attr("oj-data-label") ? element.attr("oj-data-label") :
                        element.text().trim();
            itemValue = ojOptionItem.length > 0 ? ojOptionItem.prop("value") :
                                                  element.attr("oj-data-value");
            disabled =  ojOptionItem.length > 0 ? ojOptionItem.prop("disabled") :
                                                  undefined;
          }

          return {
            value: itemValue,
            label: itemLabel,
            disabled: disabled,
            element: element.get(),
            css: element.attr("class"),
            children: groupData
          };
        }
      },

      //_AbstractOjChoice
      /*
       * Prepares the option items to display in the drop down
       */
      _prepareOpts : function (opts)
      {
        var element,
        datalist,
        self = this;

        // clone the options array if optionsKeys is specified
        if (opts['options'] && Array.isArray(opts['options']) && opts['optionsKeys'])
          opts['options'] = $.extend(true,[],opts['options']);

        element = opts.element;
        var tagName = element.get(0).tagName.toLowerCase();

        if (opts.ojContext._IsCustomElement())
        {
          if (!opts['options'] && opts.element.children().length > 0)
            this.datalist = datalist = $(element);
        }
        else if (tagName === "input" && element.attr("list"))
        {
          this.datalist = datalist = $('#' + element.attr("list"));
        }
        ///ojselect
        else if (tagName === "select" && element.children().length > 0)
        {
          this.datalist = datalist = element;
        }
        // if html ul list is provided
        else if (opts['list'])
        {
          this.datalist = datalist = $('#' + opts['list']);
        }

        opts = $.extend({},
          {
            populateResults : function (container, results, query, showPlaceholder)
            {
              var populate,
              id = this.opts.id;

              var optionRenderer = this.opts.optionRenderer;
              if (typeof optionRenderer !== "function")
              {
                // cannot be non-function
                optionRenderer = null;
              }
              if (this.opts.ojContext._WrapCustomElementRenderer)
              {
                //update renderer for custom element format
                optionRenderer = this.opts.ojContext._WrapCustomElementRenderer(optionRenderer);
              }

              populate = function (resultsParent, results, container, depth, showPlaceholder)
              {
                var i,
                l,
                result,
                selectable,
                disabled,
                node,
                label,
                innerContainer,
                formatted;

                var processChildren = function (node, result)
                {
                  if (result.children && result.children.length > 0)
                  {
                    var nestedList = result.element && $(result.element[0]).is("li") &&
                        $(result.element[0]).children("ul");

                    var innerContainer = nestedList ? $(result.element[0]).children("ul") :
                        $("<ul></ul>");

                    if (!innerContainer.hasClass("oj-listbox-result-sub"))
                      innerContainer.addClass("oj-listbox-result-sub");

                    populate(result, result.children, innerContainer, depth + 1, false);

                    if (!nestedList)
                      node.append(innerContainer); // @HTMLUpdateOK
                  }
                };

                var termHighlight = function(highlighterSection,
                                      highlighterClass, pattern) {
                  function innerHighlight(node, pat) {
                    var skip = 0;
                    if (node.nodeType === 3) {
                      var pos = node.data.toUpperCase().indexOf(pat);
                      if (pos >= 0) {
                        var spannode = document.createElement("span");
                        spannode.className = highlighterClass;
                        // - issue with ojselect component when filtering certain lower case char
                        //This is a temporary fix to avoid exceptions.
                        //For real fix, we need to know the beginning and ending offset of the match.
                        var middlebit = node.splitText(Math.min(pos, node.data.length));
                        var endbit = middlebit.splitText(Math.min(pat.length, middlebit.data.length));
                        var middleclone = middlebit.cloneNode(true);

                        spannode.appendChild(middleclone); // @HTMLUpdateOK
                        middlebit.parentNode.replaceChild(spannode, middlebit); // @HtmlUpdateOk

                        skip = 1;
                      }
                    } else if (node.nodeType === 1 && node.childNodes
                      && !/(script|style)/i.test(node.tagName)) {

                      // This function is to highlight the text appeared in the passed-in node.
                      // So recursively it checks for child nodes also.
                      // But need not to highlight the text appeared in <script> and <style> tags, so skipping them.
                      for (var i = 0; i < node.childNodes.length; ++i) {
                        i += innerHighlight(node.childNodes[i], pat);
                      }
                    }
                    return skip;
                  }

                  if (highlighterSection.length && pattern && pattern.length) {
                    highlighterSection.each(function() {
                      innerHighlight(this, pattern.toUpperCase());
                    });
                  }
                };

                var highlightLabelNode = function(labelNode)
                {
                  if (opts.highlightTermInOptions(query))
                  {
                    var highlighterSection = labelNode
                      .find(".oj-listbox-highlighter-section");
                    if (!highlighterSection.length)
                    {
                      highlighterSection = labelNode;
                    }

                    termHighlight(highlighterSection, "oj-listbox-highlighter",
                      query.term);
                  }
                };

                var createLabelContent = function(labelNode, result)
                {
                  if (optionRenderer)
                  {
                    var context = {
                      "index": i,
                      "depth": depth,
                      "leaf": !result.children,
                      "parent": resultsParent,
                      "data": result,
                      "component": opts.ojContext,
                      "componentElement": self.container[0],
                      "parentElement": labelNode.get(0)
                    };

                    // if an element is returned from the renderer and
                    // the parent of that element is null, we will append
                    // the returned element to the parentElement.
                    // If non-null, we won't do anything, assuming that the
                    // rendered content has already added into the DOM somewhere.
                    var content = optionRenderer.call(opts.ojContext, context);
                    if (content !== null)
                    {
                      // allow return of document fragment from jquery create or
                      // js document.createDocumentFragment
                      if (content["parentNode"] === null
                        || content["parentNode"] instanceof DocumentFragment)
                      {
                        labelNode.get(0).appendChild(content); // @HTMLUpdateOK
                      }
                    }
                  }
                  else
                  {
                    formatted = opts.formatResult(result);
                    if (formatted !== undefined)
                    {
                      labelNode.text(formatted);
                      labelNode.attr("aria-label", formatted);
                    }
                  }

                  highlightLabelNode(labelNode);
                };

                // - ojselect does not show placeholder text when data option is specified
                // - placeholder text is a selectable item that results in an error for ojcomponent
                ///ojselect only add placeholder to dropdown if there is no search filter
                ///and not required
                var placeholder = self._getPlaceholder();
                if (showPlaceholder && placeholder !== null && ! query.term &&
                    container.find(".oj-listbox-placeholder").length <= 0 &&
                    (tagName !== "select" || ! self.ojContext._IsRequired()))
                {
                  //create placeholder item
                  result = {
                    value: '',
                    label: placeholder
                  };

                  node = $("<li></li>");
                  node.addClass("oj-listbox-placeholder oj-listbox-results-depth-0 oj-listbox-result oj-listbox-result-selectable");
                  node.attr("role", "presentation");

                  label = $(document.createElement("div"));
                  label.addClass("oj-listbox-result-label");
                  label.attr("id", "oj-listbox-result-label-" + _ComboUtils.nextUid());
                  label.attr("role", "option");

                  formatted = opts.formatResult(result);
                  if (formatted !== undefined)
                    label.text(formatted);

                  label.attr("aria-label", formatted);
                  node.append(label); // @HTMLUpdateOK

                  node.data(self._elemNm, result);
                  container.prepend(node); // @HTMLUpdateOK
                }

                for (i = 0, l = results.length; i < l; i = i + 1)
                {
                  result = results[i];
                  disabled = (result.disabled === true);
                  selectable = (!disabled) && (id(result) !== undefined);

                  var isList = result.element && $(result.element[0]).is("li");
                  node = isList ? $(result.element[0]) : $("<li></li>");

                  if (node.hasClass("oj-listbox-result"))
                  {
                    if (result.children && result.children.length > 0)
                      processChildren(node, result);

                    highlightLabelNode($(result.element[0]).children("div"));
                    $(result.element[0]).css('display', '');
                  }
                  else
                  {
                    node.addClass("oj-listbox-results-depth-" + depth);
                    node.addClass("oj-listbox-result");
                    node.addClass(selectable ? "oj-listbox-result-selectable" : "oj-listbox-result-unselectable");
                    if (disabled)
                    {
                      node.addClass("oj-disabled");
                    }
                    if (result.children)
                    {
                      node.addClass("oj-listbox-result-with-children");
                    }
                    node.attr("role", "presentation");

                    label = $(document.createElement("div"));
                    label.addClass("oj-listbox-result-label");
                    label.attr("id", "oj-listbox-result-label-" + _ComboUtils.nextUid());
                    label.attr("role", "option");
                    if (disabled)
                      label.attr("aria-disabled", "true");

                    // append label to node
                    if (!isList)
                    {
                      createLabelContent(label, result);
                      node.append(label); // @HTMLUpdateOK
                    }

                    // process children
                    if (result.children && result.children.length > 0)
                      processChildren(node, result);

                    node.data(self._elemNm, result);
                    if (!isList)
                    {
                      container.append(node); // @HTMLUpdateOK
                    }
                    else
                    {
                      var elem = $(result.element[0]);
                      // oj-optgroup
                      if (elem.children("oj-optgroup").length > 0)
                      {
                        var groupLabel = elem.children("oj-optgroup").prop("label")+"";
                        elem.prepend(label.text(groupLabel)); // @HTMLUpdateOK
                      }
                      else if (elem.children("oj-option").length > 0)
                         elem.contents().wrapAll(label);  // @HTMLUpdateOK
                      else
                        // wrap the li contents except the nested ul with wrapping div
                        elem.contents().filter(function(){return this.tagName !== "UL";}).wrapAll(label); // @HTMLUpdateOK

                      highlightLabelNode(elem.children("div"));
                      elem.css('display', '');
                    }
                  }
                }
              };

              ///ojselect placehholder
              populate(null, results, container, 0, showPlaceholder);
            },

            highlightTermInOptions : function(query)
            {
              return (!(query.initial === true));
            }
          }, _ojChoice_defaults, opts);

        opts.id = function (e)
        {
          return e['value'];
        };

        opts.formatResult = function (result)
        {
          var label = _ComboUtils.getLabel(result);
          return (!isNaN(label) ? this.ojContext._formatValue(label) : label);
        };

        opts.formatSelection = function (data)
        {
          var label = _ComboUtils.getLabel(data);
          return (data && label) ? (!isNaN(label) ? this.ojContext._formatValue(label) : label) : undefined;

        };

        if (tagName !== "select" && opts["manageNewEntry"] !== null)
        {
          opts["manageNewEntry"] = function (term)
          {
            var entry = {};
            entry['value'] = entry['label'] = $.trim(term);
            return entry;
          }
        }

        if (this.datalist)
        {
          opts.query = this._bind(function (query)
            {
              var data =
              {
                results : [],
                more : false
              },
              term = query.term,
              children,
              process;

              process = function (element, collection)
              {
                var group, elems;
                var nestedLi = element.children() && element.children().length > 0
                      && (element.children().is("ul") || element.children().is("oj-optgroup"));
                if (element.is("option") || element.is("oj-option") || (element.is("li") && !nestedLi))
                {
                  if (query.matcher(term, element.text() || element.attr("label"), element))
                  {
                    collection.push(self._optionToData(element));
                  }
                }
                else if (element.is("optgroup") || element.is("oj-optgroup") || (element.is("li") && nestedLi))
                {
                  group = self._optionToData(element);

                  if (element.is("optgroup") || element.is("oj-optgroup"))
                    elems = element.children();
                  else if (element.children("oj-optgroup"))
                    elems = element.children().children("ul").children();
                  else
                    element.children("ul").children();

                  _ComboUtils.each2(elems, function (i, elm)
                  {
                    process(elm, group.children);
                  }
                  );
                  if (group.children.length > 0)
                  {
                    collection.push(group);
                  }
                }
              };

              children = this.datalist.children();

              ///ojselect remove existing placeholder item
              if (this._getPlaceholder() !== undefined && children.length > 0 &&
                  children.first().attr("value") == "")
              {
                children = children.slice(1);
              }

              _ComboUtils.each2(children, function (i, elm)
              {
                process(elm, data.results);
              }
              );
              query.callback(data);
            }
            );
        }
        else if ("options" in opts)
        {
          var dataOptions = opts.options;
          var dataProvider;

          if (_ComboUtils.isDataProvider(dataOptions)) {
            dataProvider = _ComboUtils.getDataProvider(opts);
          }

          if (dataProvider)
          {
            opts.query = function(query) {
              if (query.value) {
                _ComboUtils.fetchByKeyFromDataProvider(self.container, 
                                                       dataProvider, query);
              }
              else {
                _ComboUtils.fetchFromDataProvider(self, opts, query);
              }
            };
          }
          else if ($.isFunction(dataOptions))
          {
            opts.query = _ComboUtils.remote(dataOptions, opts.optionsKeys ? opts.optionsKeys : null);
          }
          else
          {
            opts.query = _ComboUtils.local(dataOptions, opts.optionsKeys ? opts.optionsKeys : null);
          }
        }

        return opts;
      },

      //_AbstractOjChoice
      _createHeader: function()
      {
        var headerElem = this.opts.element.find(".oj-listbox-header");
        if (headerElem.length)
        {
          this.header = $("<li>", {
            "class": "oj-listbox-result-header oj-listbox-result-unselectable",
            "role": "presentation"
          });

          this.header.append(headerElem.children()); // @HTMLUpdateOK
          this._initializeHeaderItems();

          var resultsNHeaderContainer = $("<ul>",
            {"class": "oj-listbox-results-with-header",
              "role": "listbox"});

          resultsNHeaderContainer.append(this.header); // @HTMLUpdateOK
          resultsNHeaderContainer.appendTo(this.results.parent()); // @HTMLUpdateOK

          var resultsWrapper = $("<li>", {"role": "presentation"});
          resultsNHeaderContainer.append(resultsWrapper); // @HTMLUpdateOK

          this.results.attr("role", "presentation");
          this.results.appendTo(resultsWrapper); // @HTMLUpdateOK
        }

        this.headerInitialized = true;
      },

      _initializeHeaderItems: function()
      {
        this.headerItems = this.header.find("li[role='option'], li:not([role])");
        this.headerItems.uniqueId();

        this.header.find("ul").attr("role", "presentation");
        this.header.find("li:not([role])").attr("role", "option");

        var selector = "a, input, select, textarea, button, object, .oj-component-initnode";
        this.header.find(selector).each(function()
        {
          $(this).attr("tabIndex", -1);
        });
      },

      _isHeaderItem: function(item)
      {
        var isHeaderItem = false;

        this.headerItems.each(function()
        {
          if ($(this).attr("id") === item)
          {
            isHeaderItem = true;
            return false;
          }
        });

        return isHeaderItem;
      },

      _getNextHeaderItem: function(currentItem)
      {
        if (!this.headerItems) {
          return null;
        }

        if (!currentItem) {
          return this.headerItems.first();
        }

        var foundCurrentItem = false;
        var nextItem = null;
        this.headerItems.each(function() {
          if (foundCurrentItem) {
            nextItem = $(this);
            return false;
          }

          foundCurrentItem = ($(this).attr("id") === currentItem);
        });

        return nextItem;
      },

      _getPreviousHeaderItem: function(currentItem) {
        if (!this.headerItems) {
          return null;
        }

        var previousItem = null;
        this.headerItems.each(function() {
          if ($(this).attr("id") === currentItem)
            return false;

          previousItem = $(this);
        });

        return previousItem;
      },

      _setFocusOnHeaderItem: function(item) {
        var focusable = item.find(".oj-component .oj-enabled").first();
        if (focusable.length === 0) {
          var selector = "a, input, select, textarea, button, object, .oj-component-initnode";
          focusable = item.find(selector).first();
          if (focusable.length === 0) {
            focusable = item.children().first();
          }
        }
        if (focusable) {
          focusable.addClass("oj-focus oj-focus-highlight oj-focus-only");
        }
      },

      _removeHighlightFromHeaderItems: function() {
        if (this.headerItems) {
          this.headerItems.find(".oj-focus")
                  .removeClass("oj-focus oj-focus-highlight oj-focus-only");
        }
      },

      //_AbstractOjChoice
      _triggerSelect : function (data)
      {
        var evt = $.Event(this._elemNm + "-selecting",
          {
            val : this.id(data),
            object : data
          }
          );
        this.opts.element.trigger(evt);
        return !evt.isDefaultPrevented();
      },

      //_AbstractOjChoice
      _isInterfaceEnabled : function ()
      {
        return this.enabledInterface === true;
      },

      //_AbstractOjChoice
      _enableInterface : function ()
      {
        var enabled = this._enabled && !this._readonly,
        disabled = !enabled;

        if (enabled === this.enabledInterface)
          return false;

        this.container.toggleClass("oj-disabled", disabled);
        this.close();
        this.enabledInterface = enabled;

        return true;
      },

      //_AbstractOjChoice
      _enable : function (enabled)
      {
        if (enabled === undefined)
          enabled = true;
        if (this._enabled === enabled)
          return;
        this._enabled = enabled;

        this.opts.element.prop("disabled", !enabled);
        this.container.toggleClass("oj-enabled", enabled);

        this._enableInterface();
      },

      //_AbstractOjChoice
      _disable : function ()
      {
        this._enable(false);
      },

      //_AbstractOjChoice
      _readonly : function (enabled)
      {
        if (enabled === undefined)
          enabled = false;
        if (this._readonly === enabled)
          return false;
        this._readonly = enabled;

        this.opts.element.prop("readonly", enabled);
        this._enableInterface();
        return true;
      },

      //_AbstractOjChoice
      _opened : function ()
      {
        return this.container.hasClass("oj-listbox-dropdown-open");
      },

      // return the element one which we want to position the listbox-dropdown. We don't
      // want it to be the container because we add the inline messages to the container
      // and if we line up to the container when it has inline messages, the dropdown
      // appears after the inline messages. We want it to always appear next to the input,
      // which is the first child of the container.
      //_AbstractOjChoice
      _getDropdownPositionElement : function ()
      {
        return this.container.children().first();
      },

      //_AbstractOjChoice
      _usingHandler : function (pos, props)
      {

        // if the input part of the component is clipped in overflow, implicitly close the dropdown popup.
        if (oj.PositionUtils.isAligningPositionClipped(props))
        {
          // add busy state
          var resolveBusyState = _ComboUtils._addBusyState(this.container, "closing popup");
          this._closeDelayTimer = window.setTimeout(function () {
            $.proxy(this.close, this);
            resolveBusyState();
          }, 1);

          return;
        }

        var container = this.container;
        var dropdown = props["element"]["element"];
        dropdown.css(pos);

        if ("bottom" === props["vertical"])
        {
          container.addClass("oj-listbox-drop-above");
          dropdown.addClass("oj-listbox-drop-above");
          }
        else
        {
          container.removeClass("oj-listbox-drop-above");
          dropdown.removeClass("oj-listbox-drop-above");
        }
      },

      //_AbstractOjChoice
      _getDropdownPosition: function ()
        {
        var position = {
                        'my': 'start top',
                        'at': 'start bottom',
                        'of': this._getDropdownPositionElement(),
                        'collision': 'flip',
                        'using': $.proxy(this._usingHandler, this)
        };
        var isRtl = oj.DomUtils.getReadingDirection() === "rtl";
        return oj.PositionUtils.normalizeHorizontalAlignment(position, isRtl);
      },

      //_AbstractOjChoice
      _positionDropdown : function ()
        {
        var dropdown = this.dropdown;
        var position = this._getDropdownPosition();
        var container = this.container;

        dropdown.css('width', container.outerWidth());
        dropdown.position(position);
      },

      //_AbstractOjChoice
      // beforeExpand
      _shouldOpen : function (e)
      {
        if (this._opened())
          return false;
        if (this._enabled === false || this._readonly === true)
          return false;

        var eventData = {
          'component' : this.opts.element
        };

        return this.ojContext._trigger("beforeExpand", e, eventData);
      },

      //_AbstractOjChoice
      _clearDropdownAlignmentPreference : function ()
      {
        // clear the classes used to figure out the preference of where the dropdown should be opened
        this.container.removeClass("oj-listbox-drop-above");
        this.dropdown.removeClass("oj-listbox-drop-above");
      },

      //_AbstractOjChoice
      /**
       * Opens the dropdown
       *
       * @return {boolean} whether or not dropdown was opened. This method will return false if, for example,
       * the dropdown is already open, or if the 'open' event listener on the element called preventDefault().
       * @ignore
       */
      open : function (e, dontUpdateResults)
      {
        if (!this._shouldOpen(e))
          return false;
        this._opening(e, dontUpdateResults);
        return true;
      },

      //_AbstractOjChoice
      _opening : function ()
      {
        if (!this.headerInitialized)
          this._createHeader();

        //this._clearPlaceholder();
        this.container.addClass("oj-listbox-dropdown-open");
      },

      //_AbstractOjChoice
      _showDropDown: function ()
      {
        if (!this._opened())
        {
          // Just to make sure that _opening() method is called before calling
          // the _showDropDown().
          return;
        }

        var alreadyExpanded = this._getActiveContainer().attr("aria-expanded");
        if (alreadyExpanded === "true")
        {
          return;
        }

        this._clearDropdownAlignmentPreference();

        // - picking ojselect value using filter and keyboard may cause dropdown close error
        // For signle select only -
        // if popup exists, refresh its content, otherwise create a popup

        var popupRoot = this.dropdown.parent();
        if (this._classNm === "oj-select" && this.opts['multiple'] !== true &&
            popupRoot && popupRoot.hasClass("oj-listbox-drop-layer"))
        {
          oj.PopupService.getInstance().triggerOnDescendents(popupRoot, oj.PopupService.EVENT.POPUP_REFRESH);
        }
        else
        {
          if (this.dropdown[0] !== this.body().children().last()[0])
          {
            this.dropdown.detach().appendTo(this.body()); // @HTMLUpdateOK
          }

          this.dropdown.appendTo(this.body()); // @HTMLUpdateOK

          if (this.header)
          {
            this.dropdown.find(".oj-listbox-results-with-header").prepend(this.header); // @HTMLUpdateOK
            this.header.show();
          }

          var psEvents = {};
          psEvents[oj.PopupService.EVENT.POPUP_CLOSE] = $.proxy(this.close, this);
          psEvents[oj.PopupService.EVENT.POPUP_REMOVE] = $.proxy(this._surrogateRemoveHandler, this);
          psEvents[oj.PopupService.EVENT.POPUP_AUTODISMISS] = $.proxy(this._clickAwayHandler, this);
          psEvents[oj.PopupService.EVENT.POPUP_REFRESH] = $.proxy(this._positionDropdown, this);

          /** @type {!Object.<oj.PopupService.OPTION, ?>} */
          var psOptions = {};
          psOptions[oj.PopupService.OPTION.POPUP] = this.dropdown;
          psOptions[oj.PopupService.OPTION.LAUNCHER] = this.opts.element;
          psOptions[oj.PopupService.OPTION.EVENTS] = psEvents;
          psOptions[oj.PopupService.OPTION.POSITION] = this._getDropdownPosition();
          psOptions[oj.PopupService.OPTION.LAYER_SELECTORS] = "oj-listbox-drop-layer";
          psOptions[oj.PopupService.OPTION.CUSTOM_ELEMENT] = this.ojContext._IsCustomElement();
          oj.PopupService.getInstance().open(psOptions);

          // move the global id to the correct dropdown
          $("#oj-listbox-drop").removeAttr("id");
          this.dropdown.attr("id", "oj-listbox-drop");

          var containerId = this.containerId;
          this.dropdown.attr("data-oj-containerid", containerId);
        }

        // show the elements
        this._positionDropdown();

        ///select: accessibility
        this._getActiveContainer().attr("aria-expanded", true);
      },

      //_AbstractOjChoice
      close : function (event)
      {
        if (!this._opened())
          return;

        this.container.removeClass("oj-listbox-dropdown-open");

        var dropDownVisible = this._getActiveContainer().attr("aria-expanded");
        delete this.ojContext._resolveSearchBoxLater;

        if (!dropDownVisible || dropDownVisible === "false")
        {
          return;
        }

        var cid = this.containerId,
        scroll = "scroll." + cid,
        resize = "resize." + cid,
        orient = "orientationchange." + cid;

        // unbind event listeners
        this.container.parents().add(window).each(function ()
        {
          $(this).off(scroll).off(resize).off(orient);
        }
        );

        this._clearDropdownAlignmentPreference();

        /** @type {!Object.<oj.PopupService.OPTION, ?>} */
        var psOptions = {};
        psOptions[oj.PopupService.OPTION.POPUP] = this.dropdown;
        oj.PopupService.getInstance().close(psOptions);

        if (this.header)
        {
          // When popup opened header will be shown in the popup.
          // But once it is closed contents of the popup will be removed,
          // but the header should not be detached from the DOM,
          // because knockout binding will be lost. That is why header will be
          // moved under the component container. And when again popup opened
          // it will be added back to the popup.
          this.header.hide();
          this.header.appendTo(this.container); // @HTMLUpdateOK
        }

        this.dropdown.removeAttr("data-oj-containerid");
        this.dropdown.removeAttr("id");

        if (this.opts["list"] || this.isOjOption)
        {
          this._removeHighlight();
        }
        else
        {
          this.dropdown.detach();
          this.results.empty();
        }

        ///select: accessibility
        this._getActiveContainer().attr("aria-expanded", false);

        if (this._elemNm === "ojcombobox")
          this._getActiveContainer().removeAttr("aria-activedescendant");

        // - press escape after search in select causes select to become unresponsive
        $.removeData(this.container, this._classNm + "-last-term");
      },


      //_AbstractOjChoice
      _setPickerAttributes: function (pickerAttributes)
      {
        oj.EditableValueUtils.setPickerAttributes(this.dropdown, pickerAttributes);
      },

      //_AbstractOjChoice
      _clearSearch : function ()  {},

      //_AbstractOjChoice
      _ensureHighlightVisible : function ()
      {
        var results = this.results,
        children,
        index,
        child,
        hb,
        rb,
        y,
        more;

        index = this._highlight();

        if (index < 0)
          return;

        children = this._findHighlightableChoices();
        child = $(children[index]);
        hb = child.offset().top + child.outerHeight(true);

        // if this is the last child lets also make sure oj-combobox-more-results is visible
        if (index === children.length - 1)
        {
          more = results.find("li.oj-listbox-more-results");
          if (more.length > 0)
          {
            hb = more.offset().top + more.outerHeight(true);
          }
        }

        rb = results.offset().top + results.outerHeight(true);
        if (hb > rb)
        {
          results.scrollTop(results.scrollTop() + (hb - rb));
        }
        y = child.offset().top - results.offset().top;

        // make sure the top of the element is visible
        if (y < 0 && child.css('display') != 'none')
        {
          results.scrollTop(results.scrollTop() + y); // y is negative
        }
      },

      //_AbstractOjChoice
      _findHighlightableChoices : function ()
      {
        return this.results.find(".oj-listbox-result-selectable:not(.oj-disabled, .oj-selected)")
                           .filter(function(){
                              return $(this).css('display') != 'none';
                            });
      },

      //_AbstractOjChoice
      _moveHighlight : function (delta)
      {
        var choices = this._findHighlightableChoices(),
        index = this._highlight();

        if (this.header && (index <= 0 || index === (choices.length - 1)))
        {
          var activeDescendant = this._getActiveContainer().attr("aria-activedescendant");
          var isHeaderItem = this._isHeaderItem(activeDescendant);
          if (!isHeaderItem)
            activeDescendant = null;

          var headerItem = null;
          if (delta > 0 && (index < 0 || index === (choices.length - 1)))
          {
            headerItem = this._getNextHeaderItem(activeDescendant);
          }
          else if (delta < 0 && ((isHeaderItem && index < 0) || index === 0))
          {
            headerItem = this._getPreviousHeaderItem(activeDescendant);
          }

          if (headerItem)
          {
            this._removeHighlight();
            this._setFocusOnHeaderItem(headerItem);

            this._getActiveContainer().attr("aria-activedescendant",
              headerItem.attr("id"));

            return;
          }
          else if (isHeaderItem && delta < 0)
          {
            index = 0;
          }
        }

        while (index >= -1 && index < choices.length)
        {
          index += delta;

          // support cycling through the first/last item
          if (index == choices.length)
            index = 0;
          else if (index == -1)
            index = choices.length - 1;

          var choice = $(choices[index]);
          if (choice.hasClass("oj-listbox-result-selectable") && !choice.hasClass("oj-disabled") &&
              !choice.hasClass("oj-selected"))
          {
            this._highlight(index);
            break;
          }
        }
      },

      //_AbstractOjChoice
      _highlight : function (index)
      {
        var choices = this._findHighlightableChoices(),
        choice,
        data;

        var searchBox = this.dropdown.find(".oj-listbox-search");
        var searchVisible = (searchBox && ! searchBox.attr("aria-hidden"));

        if (arguments.length === 0)
        {
          // If no argumnets passed then currently highlighted
          // option will be returned.
          var curSelected = choices.filter(".oj-hover");
          if (!curSelected.length)
          {
            curSelected = choices.children(".oj-hover")
              .closest(".oj-listbox-result");
          }
          // - acc: screenreader not reading ojselect items
          if (searchVisible) {
            this._updateMatchesCount(curSelected.text());
          }
          return choices.get().indexOf(curSelected[0]);
        }

        if (index >= choices.length)
          index = choices.length - 1;
        if (index < 0)
          index = 0;

        this._removeHighlight();

        choice = $(choices[index]);

        if (choice.hasClass("oj-listbox-result-with-children"))
        {
          var sels = choice.children(".oj-listbox-result-label");
          sels.addClass("oj-hover");
          // - acc: screenreader not reading ojselect items
          if (searchVisible) {
            this._updateMatchesCount(sels.text());
          }
        }
        else
        {
          choice.addClass("oj-hover");
          // - acc: screenreader not reading ojselect items
          if (searchVisible) {
            this._updateMatchesCount(choice.text());
          }
        }

        // ensure assistive technology can determine the active choice
        ///select: accessibility
        this._getActiveContainer().attr("aria-activedescendant",
                                        choice.find(".oj-listbox-result-label").attr("id"));

        this._ensureHighlightVisible();
      },

      //_AbstractOjChoice
      _removeHighlight : function ()
      {
        this.results.find(".oj-hover").removeClass("oj-hover");
        this._removeHighlightFromHeaderItems();

        if (this._elemNm === "ojcombobox")
          this._getActiveContainer().removeAttr("aria-activedescendant");
      },

      //_AbstractOjChoice
      _highlightUnderEvent : function (event)
      {
        var el = $(event.target).closest(".oj-listbox-result-selectable");
        if (el.length > 0 && !el.is(".oj-hover"))
        {
          var choices = this._findHighlightableChoices();
          this._highlight(choices.index(el));
        }
        else if (el.length == 0)
        {
          // if we are over an unselectable item remove all highlights
          this._removeHighlight();
        }
      },

      //_AbstractOjChoice
      _updateMatchesCount : function (translatedString)
      {
        //write to this liveRegion only when the dropdown message box doesn't exist
        if (! this.dropdown.find(".oj-listbox-filter-message-text").length) {
          var liveRegion = this.container.find(".oj-listbox-liveregion");
          if (liveRegion.length) {
            liveRegion.text(translatedString);
          }
        }
      },

      //_AbstractOjChoice
      _updateResults : function (initial)
      {
        var search = this.search,
        self = this,
        term = search.val(),
        lastTerm = $.data(this.container, this._classNm + "-last-term");

        // prevent duplicate queries against the same term
        // not applying to multi select since user can search the same term after making selection
        // it's ok for single select since the last term will be updated after selection
        if (initial !== true && lastTerm && (term === lastTerm) && this.opts['multiple'] !== true)
          return;

        // In IE even for chnage of placeholder fires 'input' event,
        // so in such cases we don't need to query for results.
        if (!lastTerm && !term && initial && initial.type === "input")
          return;

        $.data(this.container, this._classNm + "-last-term", term);

        var minLength = this.opts.minLength || 0;
        if (term.length >= minLength)
        {
          if (this._queryTimer) {
              this._queryTimer.clear();
          }

          if (!initial || initial === true)
          {
            this._queryResults(initial);
          }
          else
          {
            this._queryResolveBusyState = _ComboUtils._addBusyState(this.container, "query results");
            this._queryTimer = oj.TimerUtils.getTimer(_ComboUtils.DEFAULT_QUERY_DELAY);
            this._queryTimer.getPromise()
            .then(function(completed) {
                completed && self._queryResults(initial);
            })
            .then(this._queryResolveBusyState);
          }
        }
        else
        {
          this.close();
        }
      },

      //_AbstractOjChoice
      _queryResults : function (initial)
      {
        var search = this.search,
        results = this.results,
        opts = this.opts,
        self = this,
        input,
        term = search.val(),
        // sequence number used to drop out-of-order responses
        queryNumber;

        var minLength = opts.minLength || 0;
        if (minLength > term.length)
        {
          this.close();
          return;
        }
        else
        {
          this.open(null, true);
        }

        function postRender()
        {
          self._positionDropdown();

          if (self.header && self.headerItems.length)
          {
            var highlightableChoices = self._findHighlightableChoices();
            var totalOptions = self.headerItems.length
              + highlightableChoices.length;

            self.headerItems.attr("aria-setsize", totalOptions);
            if (highlightableChoices.length)
            {
              var highlightableOptions = highlightableChoices.children("[role='option']");
              highlightableOptions.attr("aria-setsize", totalOptions);
              highlightableOptions.first().attr("aria-posinset", self.headerItems.length + 1);
            }
          }

        }

        queryNumber = ++this.queryCount;

        this._removeHighlight();
        input = this.search.val();
        if (input !== undefined && input !== null
          && (initial !== true || opts.inputSearch || opts.filterOnOpen === "rawValue" || opts.minLength > 0))
        {
          term = input;
        }
        else
        {
          term = "";
        }

        this.resultsPage = 1;

        opts.query(
        {
          element : opts.element,
          term : term,
          page : this.resultsPage,
          context : null,
          matcher : opts.matcher,
          callback : this._bind(function (data)
          {
            // ignore old responses
            if (queryNumber !== this.queryCount)
            {
              return;
            }

            // ignore a response if the oj-combobox has been closed before it was received
            if (!this._opened())
            {
              return;
            }

            //clear dropdown message if any
            if (! opts.ojContext._hasMore)
              _ComboUtils.removeDropdownMessage(self.dropdown);

            // save context, if any
            this.context = (!data || data.context === undefined) ? null : data.context;
            // create a default choice and prepend it to the list

            if ((!data || (data.results && data.results.length === 0) || this._isDataSelected(data)) 
                && _ComboUtils.checkFormatter(self.ojContext, opts.formatNoMatches, "formatNoMatches"))
            {
              var transtr = opts.formatNoMatches(self.ojContext, search.val());
              if (this._classNm === "oj-select" || this.header)
              {
                this._showDropDown();
                this._preprocessResults(results);

                postRender();
                _ComboUtils.addDropdownMessage(self.dropdown, self.ojContext, transtr);

                //if no search box, don't need a separator 
                if (! this._hasSearchBox()) {
                  var separator = self.dropdown.find(".oj-listbox-filter-message-separator");
                  if (separator.length)
                    separator.css("display", "none");
                }
              }
              else
              {
                this.close();
              }

              this._updateMatchesCount(transtr);
              return;
            }

            _ComboUtils.saveLastQueryResult(this, data.results);

            this._showDropDown();
            this._preprocessResults(results);

            self.opts.populateResults.call(this, results, data.results,
              {
                term : search.val(),
                page : this.resultsPage,
                context : null,
                initial : initial
              },
              this._showPlaceholder()
            );
            this._postprocessResults(data, initial);
            postRender();

            this._updateMatchesCount(opts.formatMoreMatches(self.ojContext,
                                      this._findHighlightableChoices().length));
          })
        }
        );
      },

      //_AbstractOjChoice
      _preprocessResults: function (results)
      {
        if (this.opts["list"] || this.isOjOption)
        {
          var resultList = results.children();
          var highlighters = resultList.find(".oj-listbox-highlighter");
          
          // hide the list items
          this._hideResultList(resultList);
        }
        else 
        {
          results.empty();
        }
      },
      
      //_AbstractOjChoice
      _normalizeHighlighterLabel : function (item)
      {
        var highlighterSection,
            labelNode;
        if (item.children("div").children("oj-option").length > 0)
        {
          highlighterSection = item.children("div").children("oj-option").children(".oj-listbox-highlighter");
          labelNode = item.children("div").children("oj-option")[0];
        }
        else
        {
          highlighterSection = item.children("div").children(".oj-listbox-highlighter");
          labelNode = item.children("div")[0];
        }
        if (highlighterSection.length > 0)
        {
          for (var i = 0; i < highlighterSection.length; i++)
          {
            // remove spans
            $(highlighterSection[i].childNodes).unwrap();
          }
          // merge back text nodes
          labelNode.normalize();
        }
      },
      
      //_AbstractOjChoice
      _hideResultList : function (resultList)
      {
        for (var i = 0; i<resultList.length; i++) {
          var item = $(resultList[i]);
          if (item.is("LI")) {
            if (item.hasClass("oj-listbox-no-results") ||
                item.hasClass("oj-listbox-placeholder"))
              item.remove();
              
            item.css('display', 'none');
            if (item.hasClass("oj-selected"))
              item.removeClass("oj-selected")
            
            // remove highlighter section and merge back text nodes
            this._normalizeHighlighterLabel(item);
          }
          
          var nested;
          if (item.children("oj-optgroup"))
            nested = item.children("oj-optgroup").children("ul");
          else
            nested = item.children("ul");
          
          if (nested && nested.children())
            this._hideResultList(nested.children());   
        }
      },

      //_AbstractOjChoice
      _cancel : function (event)
      {
        this.close(event);
      },

      //_AbstractOjChoice
      _focusSearch : function ()
      {
        _ComboUtils._focus(this, this.search);
      },

      //_AbstractOjChoice
      _selectHighlighted : function (options, event)
      {
        if (this.header)
        {
          var activeDescendant = this._getActiveContainer().attr("aria-activedescendant");
          if (this._isHeaderItem(activeDescendant))
          {
            // There can be clickable elements in the custom header and
            // which can also be selected through UP/DOWN arrow keys.
            // When such header elements selected through keyboard
            // they should work as if they have clicked.
            // That is why simulating the click on header options.

            var activeItem = $("#" + activeDescendant);
            var selector = "a, input, select, textarea, button, object";
            var clickable = activeItem.find(selector).first();
            if (clickable.length === 0) {
              clickable = activeItem.children();
            }
            if (clickable.length) {
              clickable[0].click();
            }

            this.close(event);
            return;
          }
        }

        var index = this._highlight(),
        highlighted = this.results.find(".oj-hover"),
        data = highlighted.closest(".oj-listbox-result").data(this._elemNm);

        if (data)
        {
          this._highlight(index);
          
          var previousValue = this.getVal();
          options = options || {};
          options.trigger = _ComboUtils.ValueChangeTriggerTypes.OPTION_SELECTED;

          this._onSelect(data, options, event);
          this._triggerUpdateEvent(data, options, event);
          this._triggerValueUpdatedEvent(data, previousValue);

          if (event && event.type === "keydown")
          {
            // This flag will be used in "keyup" event handler to avoid 
            // the re-process of the event.
            this.enterKeyEventHandled = true;
          }
        }
        else if (options && options.noFocus)
        {
          this.close(event);
        }
      },

      //_AbstractOjChoice
      _getPlaceholder : function ()
      {
        return this.opts.element.attr("placeholder") ||
        this.opts.element.attr("data-placeholder") || // jquery 1.4 compat
        this.opts.element.data("placeholder") ||
        this.opts.placeholder;
      },

      //_AbstractOjChoice
      _setPlaceholder : function ()
      {
        var placeholder = this._getPlaceholder();

        if (!placeholder)
          return;
        this.search.attr("placeholder", placeholder);
        this.container.removeClass(this._classNm + "-allowclear");
      },

      //_AbstractOjChoice
      _initContainerWidth : function ()
      {
        function resolveContainerWidth()
        {
          var style,
          attrs,
          matches,
          i,
          l,
          attr;

          // check if there is inline style on the element that contains width
          style = this._getAttribute('style');
          if (style !== undefined && style !== null)
          {
            attrs = style.split(';');
            for (i = 0, l = attrs.length; i < l; i = i + 1)
            {
              attr = attrs[i].replace(/\s/g, ''); //@HTMLUpdateOK
              matches = attr.match(/^width:(([-+]?([0-9]*\.)?[0-9]+)(px|em|ex|%|in|cm|mm|pt|pc))/i);
              if (matches !== null && matches.length >= 1)
                return matches[1];
            }
          }
        };

        var width = resolveContainerWidth.call(this);
        if (width !== null)
        {
          this.container.css("width", width);
        }
      },

      //_AbstractOjChoice
      getVal: function () {
        return this.ojContext.option("value");
      },


      //_AbstractOjChoice
      ///pass original event
      setVal: function (val, event, context) {
        
        var options = { doValueChangeCheck: false };
        if (context)
          options["_context"] = context;

        // - need to be asble to specify the initial value of select components bound to dprv
        if (! this._skipSetValueOptions) {
          var queryResult = _ComboUtils.getLastQueryResult(this);
          var match;


          if (queryResult) {
            if (this.ojContext.multiple) {
              match = _ComboUtils.findOptions(queryResult, val);
            }
            else {
              match = _ComboUtils.findOption(queryResult, val);
            }
          }
          //set valueOption
          if (match) {
            //clone valueOption otherwise it will not trigger change event
            this.setValOpts(match);
          }
          else {
            //new entry?
            if (this._classNm === "oj-combobox") {            
              this.ojContext._resolveValueOptionsLater = 
                _ComboUtils.findOption(this.getValOpts(), val) == null ? false : true;
            }
            else {
              this.ojContext._resolveValueOptionsLater = true;
            }
          }
        }
        if (!Array.isArray(val) && !this.ojContext._IsCustomElement())
        {
          // - select needs implementation fixes...
          // 1. _SetValue() compares the value passed in to the last saved display value. This is
          // from ADF and is useful for comapring display values for inputs but since combo sets an
          // array, this check is not needed.
          // 2. To bypass the check call this method when the value has changed and with the
          // additional parameter.
          this.ojContext._SetValue([val], event, options);
        }
        else
        {
          // - select needs implementation fixes...
          this.ojContext._SetValue(val, event, options);
        }
        // also set on the input element
        this.opts.element.val(val);
      },

      getValOpts: function () {
        var ojContext = this.ojContext;
        return ojContext.multiple ? 
          ojContext.option("valueOptions") : ojContext.option("valueOption"); 
      },

      setValOpts: function (valOpts) {
        var ojContext = this.ojContext;
        _ComboUtils.setValueOptions(ojContext, valOpts);

        if (ojContext.multiple) {
          this.opts.valueOptions = valOpts;
        }
        else {
          this.opts.valueOption = valOpts;
        }
      },

      //_AbstractOjChoice
      _triggerUpdateEvent: function(val, context, event) {
        // This method is overridden in OjInputSeachContainer to fire the
        // "update" event. As this event is relevant for only ojInputSearch,
        // there is no default implementation.
      },
      
      //_AbstractOjChoice
      _triggerValueUpdatedEvent: function(data, previousValue) {
        // This method is overridden in OjSingleCombobox to fire the
        // "ojValueUpdated" custom event. As this event is relevant for only oj-combobox-one,
        // there is no default implementation.
      },

      //_AbstractOjChoice
      ///ojselect placeholder
      _showPlaceholder : function ()
      {
        return false;
      },

      //_AbstractOjChoice
      ///select: accessibility
      _getActiveContainer : function ()
      {
        return this.search;
      },
      
      //_AbstractOjChoice
      _getAttribute : function (id)
      {
        return this.opts.ojContext._IsCustomElement() ? 
               this.opts.ojContext.OuterWrapper.getAttribute(id) :
               this.opts.element.attr(id);
      },

      //_AbstractOjChoice
      _showSearchBox : function (searchText)
      {
        var focusOnSearchBox = false;
        var searchBox = this.dropdown.find(".oj-listbox-search");
        if (searchBox)
        {
          //hide and show the search box
          if (this._hasSearchBox())
          {
            this.dropdown.find(".oj-listbox-search-wrapper")
              .removeClass("oj-helper-hidden-accessible");

            $(searchBox).removeAttr("aria-hidden");
            this.search.val(searchText);
            focusOnSearchBox = true;

          }
          else
          {
            this.dropdown.find(".oj-listbox-search-wrapper")
              .addClass("oj-helper-hidden-accessible");

            $(searchBox).attr("aria-hidden", "true");

          }
        }

        //if search box is being displayed, focus on the search box otherwise focus on the select box
        _ComboUtils._focus(this, focusOnSearchBox ? this.search : this.selection);

        ///disable "click" on spyglass
        if (focusOnSearchBox)
        {
          var self = this;
          searchBox.find(".oj-listbox-spyglass-box").on("mouseup click", function (e)
          {
            self.search.focus();
            // - select and combobox stop keyboard event propegation
            e.preventDefault();
          });
        }

      },

      //_AbstractOjChoice
      _hasSearchBox : function ()
      {
        if (this._userTyping)
          return true;

        var threshold = this.opts.minimumResultsForSearch;
        var len;
        if (this.opts['list']) {
          len = $("#"+this.opts['list']).find("li").length;
        }
        // - search not shown before typing a character
        //in case of dataProvider and if data is not available, 
        //return true temporary, but resolve later when data is ready
        else if (_ComboUtils.isDataProvider(this.opts.options)) {
          if (this.ojContext._resultCount === undefined) {
            len = threshold + 1;
            this.ojContext._resolveSearchBoxLater = true;
          }
          else {
            len = this.ojContext._resultCount;
            delete this.ojContext._resolveSearchBoxLater;
          }
        } 
        else if (this.datalist) {
          if (this.ojContext._IsCustomElement())
            // get the count of oj-options
            len = this.datalist.children().find("oj-option").length;
          else
            // get the length from the select options
            len = this.datalist[0].length;
        } else if (this.opts.options) {
          len = this.opts.options.length;
        }
        return len > threshold;
      },
      
      _isDataSelected : function (data)
      {
        return false;
      },

      _findItem: function(list, value)
      {
        for (var i=0; i<list.length; i++)
        {
          if ($(list[i]).data(this._elemNm)["value"] === value)
            return list[i];
        }
        return null;
      }

    }
    );

  var _ojChoice_defaults =
  {
    closeOnSelect : true,
    openOnEnter : true,
    formatNoMatches : function (ojContext, val)
    {
      return ojContext.getTranslatedString("noMatchesFound");
    },
    formatMoreMatches : function (ojContext, num)
    {
      if (num === 1) {
        return ojContext.getTranslatedString("oneMatchesFound");
      }
      else {
        return ojContext.getTranslatedString("moreMatchesFound",
                                             {"num":("" + num)});
      }
    },
    id : function (e)
    {
      return e.id;
    },
    matcher : function (term, text)
    {
      return ('' + text).toUpperCase().indexOf(('' + term).toUpperCase()) >= 0;
    },

    separator : ","
  };

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2012 Igor Vaynberg
 *
 * This software is licensed under the Apache License, Version 2.0 (the "Apache License") or the GNU
 * General Public License version 2 (the "GPL License"). You may choose either license to govern your
 * use of this software only upon the condition that you accept all of the terms of either the Apache
 * License or the GPL License.
 *
 * You may obtain a copy of the Apache License and the GPL License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 * http://www.gnu.org/licenses/gpl-2.0.html
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * Apache License or the GPL Licesnse is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the Apache License and the GPL License for
 * the specific language governing permissions and limitations under the Apache License and the GPL License.
 */
  /**
   * @private
   */
  var _AbstractSingleChoice = _ComboUtils.clazz(_AbstractOjChoice,
    {
      //_AbstractSingleChoice
      _enableInterface : function ()
      {
        if (_AbstractSingleChoice.superclass._enableInterface.apply(this, arguments))
        {
          this.search.prop("disabled", !this._isInterfaceEnabled());
        }
      },

      //_AbstractSingleChoice
      _focus : function ()
      {
        if (this._opened())
        {
          this.close();
        }
      },

      //_AbstractSingleChoice
      _destroy : function ()
      {
        $("label[for='" + this.search.attr('id') + "']")
        .attr('for', this.opts.element.attr("id"));
        _AbstractSingleChoice.superclass._destroy.apply(this, arguments);
      },

      //_AbstractSingleChoice
      _clear : function (event)
      {
        var data = this.selection.data(this._elemNm);
        if (data)
        {
        // guard against queued quick consecutive clicks

          //This method will be invoked with or without event but 'data' will be null when it is invoked without event.
          //This logic is intended to clear the existing value when user manualy removes text in placeholder input box(which only happens for Combobox).
          //Ideally we should pass event, when we invoke _SetValue() if it is invoked on a UI action. So adding a warning message if event is null.
          if(!event)
             oj.Logger.warn("Event should not be null when user modified the value in UI");

          // - error 'value is required' is shown even though it has a value
          //only clear value when it's not "select" and "required"
          var emptyVal = this.ojContext._IsCustomElement() ? "" : [];
          
          if (this._classNm !== "oj-select" || ! this.ojContext._IsRequired())
            this.setVal(emptyVal, event);

          this.search.val("");
          this.selection.removeData(this._elemNm);
        }
        this._setPlaceholder();
      },

      //_AbstractSingleChoice
      _initSelection : function ()
      {
        //  - need to be able to specify the initial value of select components bound to dprv
        if (! _ComboUtils.applyValueOptions(this, this.opts)) 
        {
          var element = this.datalist ? this.datalist : this.opts.element;
          this.opts.initSelection.call(null, element, this._bind(this._updateSelectedOption));
        }
      },

      //_AbstractSingleChoice
      _containerKeydownHandler : function (e)
      {
        if (!this._isInterfaceEnabled())
          return;

        if (e.which === _ComboUtils.KEY.PAGE_UP || e.which === _ComboUtils.KEY.PAGE_DOWN)
        {
          // prevent the page from scrolling
          e.preventDefault();
          return;
        }

        switch (e.which)
        {
        case _ComboUtils.KEY.UP:
        case _ComboUtils.KEY.DOWN:
          if (this._opened())
          {
            this._moveHighlight((e.which === _ComboUtils.KEY.UP) ? -1 : 1);
          }
          else
          {
            this.open(e);
          }
          // - select and combobox stop keyboard event propegation
          e.preventDefault();
          return;

        case _ComboUtils.KEY.ENTER:
          this._selectHighlighted(null, e);
          // - select and combobox stop keyboard event propegation
          e.preventDefault();
          if (!this._opened())
          {
            this._userTyping = false;
          }
          return;

        case _ComboUtils.KEY.TAB:
          this.close(e);
          this._userTyping = false;
          return;

        case _ComboUtils.KEY.ESC:
          if (this._opened())
          {
            this._cancel(e);
            // prevent the page from scrolling
            e.preventDefault();
          }
          this._userTyping = false;
          return;
        }

        ///ojselect: used by select
        this._userTyping = true;
      },

      //_AbstractSingleChoice
      _containerKeyupHandler : function (e)
      {
        if (this._isInterfaceEnabled())
        {
          if (!this._opened())
            this.open(e);
        }
      },

      //_AbstractSingleChoice
      _initContainer : function ()
      {
        var selection,
        container = this.container,
        dropdown = this.dropdown,
        elementLabel;

        // - ojselect id attribute on oj-select-choice div is not meaningful
        var rootAttr = this.opts.rootAttributes;
        var idSuffix = (rootAttr && rootAttr.id) ?
          rootAttr.id :
          (this._getAttribute("id") || _ComboUtils.nextUid());

        this.selection = selection = container.find("." + this._classNm + "-choice");
        // - ojselect missing id attribute on oj-select-choice div
        selection.attr("id", this._classNm + "-choice-" + idSuffix);

        elementLabel = $("label[for='" + this._getAttribute("id") + "']");
        if (!elementLabel.attr("id"))
          elementLabel.attr('id', this._classNm + "-label-" + idSuffix);

        // add aria associations
        selection.find("." + this._classNm + "-input").attr("id", this._classNm + "-input-" + idSuffix);
        if (!this.results.attr("id"))
          this.results.attr("id", "oj-listbox-results-" + idSuffix);
          
        var liveRegion = container.find(".oj-listbox-liveregion");
        if (liveRegion.length) {
          liveRegion.attr("id", "oj-listbox-live-" + idSuffix);
        }
        // - Accessibility : JAWS does not read aria-controls attribute set on ojselect 
        if (this._classNm !== "oj-select")
          this.search.attr("aria-owns", this.results.attr("id"));

        this.search.attr("aria-labelledby", elementLabel.attr("id")); 
        this.opts.element.attr("aria-labelledby", elementLabel.attr("id"));
        
        if (this.search.attr('id'))
          elementLabel.attr('for', this.search.attr('id'));
          
        if (this._getAttribute("aria-label"))
          this.search.attr("aria-label", this._getAttribute("aria-label"));
          
        if (this._getAttribute("aria-controls"))
          this.search.attr("aria-controls", this._getAttribute("aria-controls"));

        selection.on("keydown", this._bind(this._containerKeydownHandler));
        //selection.on("keyup-change input", this._bind(this._containerKeyupHandler));

        selection.on("mousedown", "abbr", this._bind(function (e)
          {
            if (!this._isInterfaceEnabled())
              return;
            this._clear(e);
            _ComboUtils.killEventImmediately(e);
            this.close(e);
            this.selection.focus();
          }
          ));

        selection.on("mousedown", this._bind(function (e)
          {
            // if the mousedown target is the end slot, do nothing
            if (e.target.getAttribute("slot") === "end" || $(this._endSlot).find(e.target).length > 0)
              return;
            ///prevent user from focusing on disabled select
            if (this.opts.element.prop("disabled"))
              _ComboUtils.killEvent(e);

            //if select box gets focus ring via keyboard event previously, clear it now
            selection.removeClass("oj-focus-highlight");

            if (this._opened())
            {
              this.close(e);
            }
            else if (this._isInterfaceEnabled())
            {
              this.open(e);
            }

            // - keyboard flashes briefly on ios.
            var hidden = this.search.parent().attr("aria-hidden");
            if (hidden && hidden == "true")
              this.selection.focus();
            else
              this.search.focus();
          
            // prevent focus move back
            if ($(e.target).hasClass("oj-combobox-open-icon"))
              _ComboUtils.killEvent(e);

            this.container.addClass("oj-active");
          }
          ));
        
        selection.on("mouseup", this._bind(function (e)
          {
            this.container.removeClass("oj-active");
          }
          ));

        selection.on("focus", this._bind(function (e)
          {
            _ComboUtils.killEvent(e);
          }
          ));
		  
        this.search.on("input", this._bind(function (e) 
          {
            this.ojContext._SetRawValue(this.search.val(), e);
          }
          ));
          
        this.search.on("focus", this._bind(function () 
          {
            this._previousDisplayValue = this.search.val();
          }
          ));

        this.search.on("blur keyup", this._bind(function (e)
          {
            if(e.type === 'keyup' && e.keyCode !== 10 && e.keyCode !== 13) return;
            
            if (this.search.val() !== undefined && this.results.find(".oj-hover").length <= 0
                  && (e.type !== 'keyup' || !this.enterKeyEventHandled)) {
              // Call _onSelect if no previous data and there is typed in text
              // or the previous data is different from typed in text
              if (this.opts["manageNewEntry"]) {
                
                var value = this.search.val();
                var data = this.opts["manageNewEntry"](value);

                var trigger = e.type === "blur" 
                  ? _ComboUtils.ValueChangeTriggerTypes.BLUR
                  : _ComboUtils.ValueChangeTriggerTypes.ENTER_PRESSED; 
                var options = {
                  trigger: trigger
                };

                var selectionData = this.selection.data(this._elemNm);
                var previousValue = this.getVal();
                
                if ((!selectionData && value !== "")
                    || (selectionData && (selectionData.label !== value))
                    || (!this.ojContext.isValid() && value !== this._previousDisplayValue)) 
                {                  
                  this._onSelect(data, options, e);

                  if (e.type !== "blur")
                  {
                    this._triggerUpdateEvent(data, options, e);
                    this._triggerValueUpdatedEvent(data, previousValue);
                  }

                } else if (e.type === 'keyup') {
                  // if the value stays the same, we still want to fire valueUpdated event to support search use cases
                  if (selectionData && selectionData.label === value)
                    data = selectionData;

                  this._triggerUpdateEvent(data, options, e);
                  this._triggerValueUpdatedEvent(data, previousValue);
                }
                
              } else if (this.opts["manageNewEntry"] == null) {
                var data = this.selection.data(this._elemNm);
                if (this.search.val() == "")
                {
                  if (this._classNm !== "oj-select")
                    this._clear(e);
                }
                else if (!data && this.search.val() !== "") 
                  this._clearSearch();
                // - typing in search text & pressing enter, changes user entered search text
                else if (this._classNm !== "oj-select")
                {
                  var formatted = this.opts.formatSelection(data);
                  if (formatted !== undefined) 
                  {
                    this.search.val(formatted);
                  }
                }
              }
            }
            this.search.removeClass(this._classNm + "-focused");
            this.container.removeClass("oj-focus");

            // Clearing the flag which is set while processing the keydown event
            // in _selectHighlighted() method.
            this.enterKeyEventHandled = false;
          }
          ));

        this._initContainerWidth();

        this.opts.element.hide().attr("aria-hidden", true);
        this.container.append(this.opts.element); //@HTMLUpdateOK

        this._setPlaceholder();

      },

      //_AbstractSingleChoice
      _prepareOpts : function ()
      {
        var opts = _AbstractSingleChoice.superclass._prepareOpts.apply(this, arguments),
        self = this;

        ///ojselect set initial selected value
        var tagName = opts.element.get(0).tagName.toLowerCase();
        if ((tagName === "input" && opts.element.attr("list")) ||
          (tagName === "select" && opts.element.children().length > 0) ||
          (opts.ojContext._IsCustomElement() && !opts['options']) ||
          opts['list'])
        {
          var eleName = opts['list'] ? "li" : "option";
          
          if (opts.ojContext._IsCustomElement())
            eleName = "oj-option";
          
          // install the selection initializer
          opts.initSelection = function (element, callback)
          {
            var selected;
            var value = self.getVal();
            if (Array.isArray(value) && !opts.ojContext._IsCustomElement())
              value = value[0];

            if (value !== undefined && value !== null)
            {
              selected = self._optionToData(element.find(eleName).filter(function ()
                {
                  var elemValue;
                  if (eleName == "li")
                    elemValue = this.getAttribute("oj-data-value");
                  else if (eleName == "option" || eleName == "oj-option")
                    elemValue = this.value;
                  
                  return oj.Object.compareValues(elemValue, value);
                }
                ));

              // - select list behaves differently when using options attribute vs options tag
              if (tagName === "select" && selected === undefined)
                value = null;
            }
            if (value === undefined || value === null)
            {
              selected = self._optionToData(element.find(eleName).filter(function ()
                {
                  if (eleName == "li")
                    return this.getAttribute("oj-data-selected") === true;
                  else if (eleName == "option")
                    return this.selected;
                }
                ));
                // set first oj-option for oj-select
                if (self._classNm === "oj-select" && selected === undefined 
                    && opts.ojContext._IsCustomElement())
                  selected = self._optionToData($(element.find(eleName)[0]));
            }
            callback(selected);
          };

          // - ojselect should ignore the invalid value set programmatically
          opts.validate = function (element, value)
          {
            var selected;

            if (value !== undefined && value !== null)
            {
              selected = self._optionToData(element.find(eleName).filter(function ()
                  {
                    var elemValue;
                    if (eleName == "li")
                      elemValue = this.getAttribute("oj-data-value");
                    else if (eleName == "option" || eleName == "oj-option")
                      elemValue = this.value;
                    
                    return oj.Object.compareValues(elemValue, value);
                  }
                  ));
            }

            return !! selected;
          };
        }
        else if ("options" in opts || (this.getVal() && this.getVal().length > 0))
        {
          if (_ComboUtils.isDataProvider(opts.options) ||
              $.isFunction(opts.options))
          {
            // install default initSelection when applied to hidden input
            // and getting data from remote
            opts.initSelection = function (element, callback)
            {
              var findOption = function(results, optionValue)
              {
                for (var i = 0, l = results.length; i < l; i++)
                {
                  var result = results[i];
                  if (oj.Object.compareValues(optionValue, opts.id(result))) {
                    return result;
                  }
                  
                  if (result.children) {
                    var found = findOption(result.children, optionValue);
                    if (found)
                      return found;
                  }
                }

                return null;
              };

              var id = self._getValueItem();
              var match = null;
              if (!id)
              {
                callback(match);
                return;
              }

              // This data will be saved after querying the options.
              var queryResult = _ComboUtils.getLastQueryResult(self);
              if (queryResult)
              {
                match = findOption(queryResult, id);
              }

              if (!match)
              {
                // currentItem will hold the selected object with value and label.
                // Which updated everytime value is changed.
                var currentItem = self.currentItem;
                if (currentItem && currentItem.length 
                  && oj.Object.compareValues(id, opts.id(currentItem[0])))
                {
                  match = currentItem[0];
                }
              }

              // valueChangeTrigger will have one of the values from 
              // _ComboUtils._ValueChangeTriggerTypes, which represents the
              // what triggered the value change. But if value is programmatically
              // updated this will be null. So if valueChangeTrigger is null 
              // querying for the options again as component will not have list
              // of options in case value is updated programmatically.
              if (!match && !self.valueChangeTrigger)
              {
                opts.query(
                {
                  value: [id],
                  callback: !$.isFunction(callback) ? $.noop : function(queryResult)
                  {
                    if (queryResult && queryResult.results)
                    {
                      match = findOption(queryResult.results, id);
                    }
                    callback(match);
                  } 
                });
              }
              else
              {
                callback(match);
              }
            };
          }
          else
          {
            // install default initSelection when applied to hidden input and data is local
            // - ojselect does not display selected value
            opts.initSelection = function (element, callback)
            {
              var id = self._getValueItem();
              //search in data by id, storing the actual matching item
              //var first = null;
              // - ojselect - validator error message is not shown
              //initialize first = placeholder if we have a placeholder and select value is not required
              var usePlaceholder = (tagName == "select") && 
                                   self.ojContext._HasPlaceholderSet() && 
                                   ! self.ojContext._IsRequired();
              var first = usePlaceholder ? self._getPlaceholder() : null;

              var match = null;
              opts.query(
              {
                matcher : function (term, text, el)
                {
                  var is_match = oj.Object.compareValues(id, opts.id(el));
                  if (is_match)
                  {
                    match = el;
                  }
                  ///ojselect save the 1st option
                  if (first == null)
                  {
                    first = el;
                  }
                  return is_match;
                },
                callback : !$.isFunction(callback) ? $.noop : function ()
                {
                  // ojselect if no match, pick the 1st option
                  // If the option data is pending, don't send the placeholder to the callback
                  // so the value won't be nulled out
                  if (!match && tagName === "select" && !self.ojContext._isOptionDataPending())
                  {
                    match = first;
                  }
                  callback(match);
                }
              }
              );
            };

            // - ojselect should ignore the invalid value set programmatically
            opts.validate = function (element, value)
            {
              var id = value;

              //search in data by id, storing the actual matching item
              var match = null;
              opts.query(
              {
                matcher : function (term, text, el)
                {
                  var is_match = oj.Object.compareValues(id, opts.id(el));
                  if (is_match)
                  {
                    match = el;
                  }
                  return is_match;
                },
                callback : $.noop
              }
              );

              return !! match;
            };
          }
        }
        return opts;
      },

      //_AbstractSingleChoice
      _postprocessResults : function (data, initial, noHighlightUpdate)
      {
        var selected = -1,
        self = this,
        highlightableChoices;

        highlightableChoices = this._findHighlightableChoices();
        _ComboUtils.each2(highlightableChoices, function (i, elm)
        {
          var valueItem = self._getValueItem();
          
          if (valueItem && oj.Object.compareValues(valueItem, self.id(elm.data(self._elemNm))))
          {
            selected = i;
            return false;
          }
        }
        );

        // and highlight it
        if (noHighlightUpdate !== false)
        {
          if (initial === true && selected >= 0)
          {
            this._highlight(selected);
          }
        }
      },

      //_AbstractSingleChoice
      ///pass original event
      _onSelect : function (data, options, event)
      {
        if (!this._triggerSelect(data))
          return;

        var context;
        if (options && options.trigger) {
          context = {
            optionMetadata: {
              "trigger": options.trigger
            }
          };
        }

        // var old = this.getVal()? this.getVal()[0] : null;
        //selection will be updated after _SetValue is called
        //this._updateSelection(data);
        this.close(event);
        // When there is validation error, the value option may retain the previous value
        // although the display value is different. In that case, user should be able to still
        // select the previous valid value to get rid off the invalid style and message.
        /*if (!(old === this.id(data)))*/
        var val;
        if (this.id(data).length === 0) {
          val = this.ojContext._IsCustomElement() ? "" : [];
          data = [];
        }
        else {
          val = this.id(data);
        }
        //setValueOptions before setVal so that new entry already in valueOption.
        //this is to avoid looking for new entry on the dataProvider
        if (this._classNm === "oj-combobox") {
          this._skipSetValueOptions = true;
          // - oj.tests.input.combobox.testcombobox display value mismatch automation failure
          this.setValOpts(_ComboUtils.findOptionFromResult(this, val, data));
        }
        this.setVal(val, event, context);
        this._skipSetValueOptions = false;
        if (event.type !== "blur")
          this._focusSearch();
      },

      //_AbstractSingleChoice
      _clearSearch : function ()
      {
        this.search.val("");
      },
      
      //_AbstractSingleChoice
      _getValueItem : function ()
      {
        var valueItem = null;
        var val = this.getVal();
        
        if (val) {
          if (!this.ojContext._IsCustomElement() && val.length)
            valueItem = val[0];
          else
            valueItem = val;
        }
        return valueItem;
      }

    }
    );


/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2012 Igor Vaynberg
 *
 * This software is licensed under the Apache License, Version 2.0 (the "Apache License") or the GNU
 * General Public License version 2 (the "GPL License"). You may choose either license to govern your
 * use of this software only upon the condition that you accept all of the terms of either the Apache
 * License or the GPL License.
 * 
 * You may obtain a copy of the Apache License and the GPL License at:
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * http://www.gnu.org/licenses/gpl-2.0.html
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * Apache License or the GPL Licesnse is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the Apache License and the GPL License for
 * the specific language governing permissions and limitations under the Apache License and the GPL License.
 */
/**
 * @private
 */
  var _OjSingleCombobox = _ComboUtils.clazz(_AbstractSingleChoice,
    {
      _elemNm : "ojcombobox",
      _classNm : "oj-combobox",

      _createContainer : function ()
      {                                      
        var container = $(document.createElement("div")).attr(
          {
            "class" : "oj-combobox oj-component"
          }
          ).html([ //@HTMLUpdateOK
              "<div class='oj-combobox-choice' tabindex='-1' role='presentation'>",
              "   <input type='text' autocomplete='off' autocorrect='off' autocapitalize='off'",
              "       spellcheck='false' class='oj-combobox-input' role='combobox' aria-expanded='false' aria-autocomplete='list' />",
              "   <abbr class='oj-combobox-clear-entry' role='presentation'></abbr>",
              "   <span class='oj-combobox-divider' role='presentation'></span>",
              "   <a class='oj-combobox-arrow oj-combobox-icon oj-component-icon oj-clickable-icon-nocontext oj-combobox-open-icon'",
              "       role='button' aria-label='expand'></a>",
              "</div>",
              "<div class='oj-listbox-drop' style='display:none' role='presentation'>",
              "   <ul class='oj-listbox-results' role='listbox'>",
              "   </ul>",
              "</div>"].join("")); //@HTMLUpdateOK
        
        // if end slot is provided, use the slot instead
        if (this.ojContext._IsCustomElement())
        {
          var slotMap = oj.BaseCustomElementBridge.getSlotMap(this.ojContext.OuterWrapper);
          var endSlot = slotMap['end'];
          
          if (endSlot)
          {
            // remove the divider
            container.find(".oj-combobox-divider").remove();
            // remove the default arrow anchor
            container.find(".oj-combobox-arrow").remove();
            // append the slot at the end
            container.find(".oj-combobox-choice").append(endSlot);
            
            this._endSlot = endSlot;
          }  
        }
        
        return container;
      },
      
      _triggerValueUpdatedEvent: function(data, previousValue) 
      {
        if (!this.ojContext._IsCustomElement())
          return;
      
        var value = this.id(data);
        if (value === undefined || value === null)
        {
          // If the value is entered by user (not by selecting an option) then 
          // only 'label' will be present in the data object.
          value = data['label'] ? data['label'] : '';
        }
        
        // validate the value
        var parsed = this.ojContext._Validate(value, null, null);
        if (parsed === undefined || !this.ojContext.isValid())
          return;

        var detail = {};
        var element = this.ojContext.OuterWrapper;
        
        detail['value'] = value;
        detail['previousValue'] = previousValue;
        
        var eventName = "ojValueUpdated";
        var valueUpdatedEvent = new CustomEvent(eventName, {'detail': detail});
        element.dispatchEvent(valueUpdatedEvent);
      },
      
      _enable : function (enabled)
      {
        _OjSingleCombobox.superclass._enable.apply(this, arguments);

        if (this._enabled) 
          this.container.find(".oj-combobox-arrow").removeClass("oj-disabled");
        else 
          this.container.find(".oj-combobox-arrow").addClass("oj-disabled");
      },
      
      close : function (event)
      {
        if (!this._opened())
          return;
        _OjSingleCombobox.superclass.close.apply(this, arguments);
      },

      _opening : function (event, dontUpdateResults)
      {
        //if beforeExpand is not cancelled
        _OjSingleCombobox.superclass._opening.apply(this, arguments);

        if (!dontUpdateResults)
          this._updateResults(true);
      },
      
      _containerKeydownHandler : function (e)
      {
        ///ignore control key and function key
        if (_ComboUtils.KEY.isControl(e) || _ComboUtils.KEY.isFunctionKey(e))
          return;
          
        _OjSingleCombobox.superclass._containerKeydownHandler.apply(this, arguments);
      },
      
      _updateSelection : function (data)
      {
        var formatted;
        
        var item = [];
        this.selection.data(this._elemNm, data);
        if (data !== null && data.length != 0)
        {
          formatted = this.opts.formatSelection(data);
          if (formatted !== undefined && this.search.val() !== formatted) 
          {
              this.search.val(formatted);
          }
          this.search.removeClass(this._classNm + "-default");

          item.push(data);
        }else{
          //data will be null only when user set it programmatically.
          this.search.val("");
          this._setPlaceholder();
        }

        // Storing this data so that it will be used when setting the display value.
        this.currentItem = item;

        if (this.opts.allowClear)
        {
          this.container.addClass(this._classNm + "-allowclear");
        }
      },
      
      _updateSelectedOption : function(selected)
      {
        if (selected !== undefined && selected !== null) {
          this._updateSelection(selected);
        }else{
          // if we found no match, update the selection with the value
          var value = this.getVal();
          var data = !value ? null : !Array.isArray(value) ? {'label':value}: (Array.isArray(value) && value.length ) ? { 'label': value[0] } : null ;
          this._updateSelection(data);
        } 

        // - need to be able to specify the initial value of select components bound to dprv
        this.setValOpts(selected);
      }
    }
    );


/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2012 Igor Vaynberg
 *
 * This software is licensed under the Apache License, Version 2.0 (the "Apache License") or the GNU
 * General Public License version 2 (the "GPL License"). You may choose either license to govern your
 * use of this software only upon the condition that you accept all of the terms of either the Apache
 * License or the GPL License.
 *
 * You may obtain a copy of the Apache License and the GPL License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 * http://www.gnu.org/licenses/gpl-2.0.html
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * Apache License or the GPL Licesnse is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the Apache License and the GPL License for
 * the specific language governing permissions and limitations under the Apache License and the GPL License.
 */
  /**
 * @private
 */
var _OjSingleSelect = _ComboUtils.clazz(_AbstractSingleChoice,
    {
      _elemNm : "ojselect",
      _classNm : "oj-select",
      _userTyping : false,


      //_OjSingleSelect
      _createContainer : function ()
      {
        var container = $(document.createElement("div")).attr(
          {
            "class" : "oj-select oj-component"
          }
          ).html([  //@HTMLUpdateOK
              "<div class='oj-select-choice' tabindex='0' role='combobox' ",
              "     aria-autocomplete='none' aria-expanded='false'>",
              "  <span class='oj-select-chosen'></span>",
              "  <abbr class='oj-select-search-choice-close' role='presentation'></abbr>",
              "  <a class='oj-select-arrow oj-component-icon oj-clickable-icon-nocontext oj-select-open-icon' role='presentation'>",
              "</a></div>",

              "<div class='oj-listbox-drop' style='display:none' role='dialog'>",

              "  <div class='oj-listbox-search-wrapper'>",

              "  <div class='oj-listbox-search'>",
              "    <input type='text' autocomplete='off' autocorrect='off' autocapitalize='off'",
              "           spellcheck='false' class='oj-listbox-input' title='Search field' ",
              "           role='combobox' aria-expanded='false' aria-autocomplete='list' />",

              "    <span class='oj-listbox-spyglass-box'>",
              "      <span class='oj-component-icon oj-clickable-icon-nocontext oj-listbox-search-icon' role='presentation'>",
              "       <b role='presentation'></b></span>",
              "    </span>",
              "  </div>",

              "  </div>",

              "   <ul class='oj-listbox-results' role='listbox'>",
              "   </ul>",
              "</div>",

              "<div role='region' class='oj-helper-hidden-accessible oj-listbox-liveregion' aria-live='polite'></div>"

            ].join("")); //@HTMLUpdateOK

        return container;
      },

      //_OjSingleSelect
      _enable : function (enabled)
      {
        _OjSingleSelect.superclass._enable.apply(this, arguments);

        // - dropdown icon is in disabled state after enabling ojselect
        if (this._enabled) {
          this.selection.attr("tabindex", "0");
          this.container.find(".oj-select-arrow").removeClass("oj-disabled");
        }
        else {
          //Don't allow focus on a disabled "select"
          this.selection.attr("tabindex", "-1");
          // - disabled select icon hover still shows changes
          this.container.find(".oj-select-arrow").addClass("oj-disabled");
        }
      },

      //_OjSingleSelect
      close : function (event)
      {
        if (!this._opened())
          return;
        _OjSingleSelect.superclass.close.apply(this, arguments);

        this.selection
          .attr("aria-expanded", false)
          .removeAttr("aria-haspopup")
          .removeAttr("aria-owns");

        this.search
          .attr("aria-expanded", false)
          .removeAttr("aria-controls");

        // - required validation err is not displayed when user tabs out
        //always clear search text when dropdown close
        if (! this._testClear(event))
          this._clearSearch();

        // - ojselect input field grabs focus on paste
        //don't set focus on the select box if event target is not select element
        if (! (event instanceof MouseEvent) ||
            event.target == this.selection || event.target == this.search) {
          _ComboUtils._focus(this, this.selection);
        }

        ///remove "mouse click" listeners on spyglass
        this.container.find(".oj-listbox-spyglass-box").off("mouseup click");
      },

      //_OjSingleSelect
      _opening : function (event, dontUpdateResults)
      {
        _OjSingleSelect.superclass._opening.apply(this, arguments);

        var searchText = _ComboUtils.getSearchText(event);

        //select: focus still stay on the selectBox if open dropdown by mouse click
        this._showSearchBox(searchText);

        if (!dontUpdateResults)
        {
          if(searchText)
            this._updateResults();
          else
            this._updateResults(true);          
        }
      },

      //_OjSingleSelect
      _showDropDown: function()
      {
        if (!this._opened())
        {
          // Just to make sure that _opening() method is called before calling
          // the _showDropDown().
          return;
        }

        var expanding = (this.selection.attr("aria-expanded") !== "true");
        _OjSingleSelect.superclass._showDropDown.apply(this, arguments);

        if (expanding) {
          this.selection
            .attr("aria-expanded", true)
            .attr("aria-haspopup", "dialog")
            .attr("aria-owns", this.results.attr("id"));

          this.search
            .attr("aria-expanded", true)
            .attr("aria-controls", 
                  this.results.attr("id") + " " + this.container.find(".oj-listbox-liveregion").attr("id"));
        }

        // - search moves cursor to end, difficult to edit search
      },

      //_OjSingleSelect
      _initContainer : function ()
      {
        ///ojselect placeholder
        var selectedId = this.containerId + "_selected";
        this.text = this.container.find(".oj-select-chosen").attr("id", selectedId);

        _OjSingleSelect.superclass._initContainer.apply(this, arguments);

        ///select: accessibility
        this.selection
          .attr({
            "aria-labelledby": this.search.attr("aria-labelledby"),
            "aria-describedby": selectedId
          });

        // - missing select label
        var label = this._getAttribute("aria-label");
        if (label)
          this.selection.attr("aria-label", label);

        this.search.on("keydown", this._bind(this._containerKeydownHandler));
        this.search.on("keyup-change input", this._bind(this._containerKeyupHandler));

        // - nls: hardcoded string 'search field' in select component
        this.search.attr("title", this.ojContext.getTranslatedString("searchField"));

        // - required validation err is not displayed when user tabs out
        var self = this;
        this.selection.on("blur", function(e) {
          self._testClear(e);
        });

      },

      //_OjSingleSelect
      _initSelection : function ()
      {
        if (this._isPlaceholderOptionSelected())
        {
          this._updateSelection(null);
          this.close();
          this._setPlaceholder();
        }
        else
        {
          _OjSingleSelect.superclass._initSelection.apply(this, arguments);
        }

      },

      //_OjSingleSelect
      _updateSelectedOption : function(selected)
      {
        if (selected !== undefined && selected !== null)
        {
          //ojSelect by default use first option if user set a value which is not listed in original option items.
          //So need to update options to reflect the correct value in component state.
          var selectedVal,
          value = this.getVal();
          
          if (Array.isArray(value) && !this.ojContext._IsCustomElement())
            value = value[0];
          
          selectedVal = this.opts.id(selected);
          
          // editableValue doesn't accept undefined value
          if (selectedVal === undefined)
            selectedVal = null;

          // - the selected option of the ojselect not reflected in the value variable
          if (!oj.Object.compareValues(value, selectedVal)) {
            this.ojContext._setInitialSelectedValue(selectedVal);
          }

          this.setValOpts(selected);
          this._updateSelection(selected);

          this.close();
        }
        else {
          this.setValOpts(null);
        }
      },

      //_OjSingleSelect
      _updateSelection : function (data)
      {
        this.selection.data(this._elemNm, data);
        // - ojet select displaying values incorrectly
        if (data !== null)
        {
          this.text.text((typeof data === "string") ? 
                         data : 
                         _ComboUtils.getLabel(data));
        }

        //make sure placeholder text has "oj-select-default" class
        if (data && data.id != "")
          this.text.removeClass(this._classNm + "-default");

        if (this.opts.allowClear)
        {
          this.container.addClass(this._classNm + "-allowclear");
        }
      },

      //_OjSingleSelect
      _getActiveContainer : function ()
      {
        var expanded = this.search.attr("aria-expanded");
        return (expanded && this._hasSearchBox()) ? this.search : this.selection;
      },

      //_OjSingleSelect
      _isPlaceholderOptionSelected : function ()
      {
        var placeholderOption;
        ///ojselect allow placeholder to be an empty string
        if (this._getPlaceholder() === null)
          return false; // no placeholder specified so no option should be considered

        var cval = this.getVal();
        cval = Array.isArray(cval)? cval[0]: cval;
        //This method is used to check whether placeholder text need to be displayed in ui or not and hence checking current value should be fine.
        return (cval === "")
         || (cval === undefined)
         || (cval === null);
      },

      //_OjSingleSelect
      ///ojselect placeholder this method should be in AbstractOjChoice
      _getPlaceholder : function ()
      {
        return this.opts.placeholder;
      },

      //_OjSingleSelect
      _showPlaceholder : function ()
      {
        return true;
      },

      //_OjSingleSelect
      _setPlaceholder : function ()
      {
        var placeholder = this._getPlaceholder();

        if (this._isPlaceholderOptionSelected() && placeholder !== undefined)
        {
          if (placeholder === undefined)
            placeholder = "";
          this.text.text(placeholder).addClass(this._classNm + "-default");
          this.container.removeClass(this._classNm + "-allowclear");
        }
      },

      //_OjSingleSelect
      setVal: function (val, event, context)
      {
        ///pass original event
        _OjSingleSelect.superclass.setVal.call(this, val, event, context);
        this.selection.data("selectVal", val);
      },

      //_OjSingleSelect
      _containerKeydownHandler : function (e)
      {
        // - strange text show up after type in "<" in the select component
        // - keyboard handling issues
        if ((_ComboUtils.KEY.isControl(e) && e.which != _ComboUtils.KEY.SHIFT) || 
            (e.which == _ComboUtils.KEY.SHIFT) || 
            _ComboUtils.KEY.isFunctionKey(e))
        {
          return;
        }

        switch (e.which)
        {
        case _ComboUtils.KEY.TAB:
          this.close(e);
          //James: tab out of an expanded poplist, focus is going all the way to the top of the page.
          this.selection.focus();

          // - required validation err is not displayed when user tabs out
          this._testClear(e);
          return;

        // open dropdown on Enter
        case _ComboUtils.KEY.ENTER:
          if (e.target === this.selection[0] && ! this._opened())
          {
            this.open(e);
            // - select and combobox stop keyboard event propegation
            e.preventDefault();
            return;
          }
          break;
        }

        _OjSingleSelect.superclass._containerKeydownHandler.apply(this, arguments);

        
        if (this._userTyping) {
          // - ojselect search box does not appear when i start typing
          if (this._opened()) {

            var searchBox = this.dropdown.find(".oj-listbox-search");
            if ($(searchBox).attr("aria-hidden") === "true") {

              var searchText = _ComboUtils.getSearchText(e);                
              if (searchText) {
                this._showSearchBox(searchText);
                this._updateResults();
              }
            }
          }
          else {
            this.open(e);
          }
        }

      },

      //_OjSingleSelect
      // - required validation err is not displayed when user tabs out
      _testClear : function (event)
      {
        if (this.text.text() == "")
        {
          this._clear(event);
          return true;
        }
        return false;
      }

    }
    );


/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2012 Igor Vaynberg
 *
 * This software is licensed under the Apache License, Version 2.0 (the "Apache License") or the GNU
 * General Public License version 2 (the "GPL License"). You may choose either license to govern your
 * use of this software only upon the condition that you accept all of the terms of either the Apache
 * License or the GPL License.
 *
 * You may obtain a copy of the Apache License and the GPL License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 * http://www.gnu.org/licenses/gpl-2.0.html
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * Apache License or the GPL Licesnse is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the Apache License and the GPL License for
 * the specific language governing permissions and limitations under the Apache License and the GPL License.
 */
/**
 * @private
 */
var _AbstractMultiChoice = _ComboUtils.clazz(_AbstractOjChoice,
  {
    _prepareOpts : function ()
    {
      var opts = _AbstractMultiChoice.superclass._prepareOpts.apply(this, arguments),
      self = this;

      var tagName = opts.element.get(0).tagName.toLowerCase();
      if ((tagName === "input" && opts.element.attr("list")) ||
        (tagName === "select" && opts.element.children().length > 0) ||
        (opts.ojContext._IsCustomElement() && !opts['options']) ||
        opts["list"])
      {
        var eleName = opts['list'] ? "li" : "option";
        
        if (opts.ojContext._IsCustomElement())
            eleName = "oj-option";

        // install the selection initializer
        opts.initSelection = function (element, callback)
        {
          var data = [];
          if (self.getVal())
          {
            var selected;
            var ids = self.getVal();
            for (var i = 0; i < ids.length; i++)
            {
              var id = ids[i];
              selected = element.find(eleName).filter(function ()
                {
                  var elemValue;
                  if (eleName == "li")
                    elemValue = this.getAttribute("oj-data-value");
                  else if (eleName == "option" || eleName == "oj-option")
                    elemValue = this.value;
                  
                  return oj.Object.compareValues(elemValue, id);
                }
                );
              if (selected && selected.length)
              {
                data.push(self._optionToData(selected));
              }
              else if (self._elemNm === "ojcombobox")
              {
                // If user entered value which is not listed in predefiend options
                data.push({'value' : id, 'label' : id});
              }

            }
          }
          // don't do this for select since it returns the first option as selected by default
          else if (tagName !== "select")
          {
            var selected;
            selected = element.find(eleName).filter(function ()
              {
                if (eleName === "option")
                  return this.selected;
                else if (eleName === "li")
                  return this.getAttribute("oj-data-selected") === true;
                else if (eleName === "oj-option")
                    return this.getAttribute("selected") === true;
              }
              );
            _ComboUtils.each2(selected, function (i, elm)
            {
              data.push(self._optionToData(elm));
            }
            );
          }
          callback(data);
        };
      }
      else if ("options" in opts)
      {
        if (_ComboUtils.isDataProvider(opts.options) ||
            $.isFunction(opts.options))
        {
          // install default initSelection when applied to hidden input and data is remote
          opts.initSelection = function (element, callback)
          {
            var findOptions = function (results, optionValues)
            {
              var foundOptions = [];
              for (var i = 0, l = results.length; i < l; i++)
              {
                var result = results[i];
                var idx = optionValues.indexOf(opts.id(result));
                if (idx >= 0)
                {
                  foundOptions.push(result);
                }

                if (result.children)
                {
                  var childOptions = findOptions(result.children, optionValues);
                  if (childOptions && childOptions.length)
                    $.merge(foundOptions, childOptions);
                }
              }

              return foundOptions;
            };

            var ids = self.getVal();
            //search in data by array of ids, storing matching items in a list
            var matches = [];

            // This data will be saved after querying the options.
            var queryResult = _ComboUtils.getLastQueryResult(self);
            if (queryResult)
            {
              matches = findOptions(queryResult, ids);
            }

            var reorderOptions = function ()
            {
              // Reorder matches based on the order they appear in the ids array because right now
              // they are in the order in which they appear in data array.
              // If not found in the current result, then will check in the saved current item.
              var ordered = [];
              for (var i = 0; i < ids.length; i++)
              {
                var id = ids[i];
                var found = false;
                for (var j = 0; j < matches.length; j++)
                {
                  var match = matches[j];
                  if (oj.Object.compareValues(id, opts.id(match)))
                  {
                    ordered.push(match);
                    matches.splice(j, 1);
                    found = true;
                    break;
                  }
                }
                if (!found)
                {
                  // currentItem will hold the selected object with value and label.
                  // Which updated everytime value is changed.
                  var currentItem = self.currentItem;
                  if (currentItem && currentItem.length)
                  {
                    for (var k = 0; k < currentItem.length; k++)
                    {
                      if (oj.Object.compareValues(id, opts.id(currentItem[k])))
                      {
                        ordered.push(currentItem[k]);
                        found = true;
                        break;
                      }
                    }
                  }

                  if (!found && self._elemNm === "ojcombobox")
                  {
                    //If user entered value which is not listed in predefiend options
                    ordered.push({'value' : id, 'label' : id});
                  }
                }
              }

              callback(ordered);
            };

            // valueChangeTrigger will have one of the values from
            // _ComboUtils._ValueChangeTriggerTypes, which represents the
            // what triggered the value change. But if value is programmatically
            // updated this will be null. So if valueChangeTrigger is null
            // querying for the options again as component will not have list
            // of options in case value is updated programmatically.
            if (!self.valueChangeTrigger)
            {
              opts.query(
              {
                value : ids,
                callback : function (queryResult)
                {
                  if (queryResult && queryResult.results)
                  {
                    var results = findOptions(queryResult.results, ids);
                    if (results && results.length)
                      $.merge(matches, results);
                  }
                  reorderOptions();
                }
              }
              );
            }
            else
            {
              reorderOptions();
            }
          };
        }
        else
        {
          // install default initSelection when applied to hidden input and data is local
          opts.initSelection = function (element, callback)
          {
            var ids = self.getVal();
            
            if (!ids || ids.length === 0)
              return;
            
            //search in data by array of ids, storing matching items in a list
            var matches = [];
            opts.query(
            {
              matcher : function (term, text, el)
              {
                var is_match = $.grep(ids, function (id)
                  {
                    return oj.Object.compareValues(id, opts.id(el));
                  }
                  ).length;
                if (is_match)
                {
                  matches.push(el);
                }
                return is_match;
              },
              callback : !$.isFunction(callback) ? $.noop : function ()
              {
                // reorder matches based on the order they appear in the ids array because right now
                // they are in the order in which they appear in data array
                var ordered = [];
                for (var i = 0; i < ids.length; i++)
                {
                  var id = ids[i];
                  var found = false;
                  for (var j = 0; j < matches.length; j++)
                  {
                    var match = matches[j];
                    if (oj.Object.compareValues(id, opts.id(match)))
                    {
                      ordered.push(match);
                      matches.splice(j, 1);
                      found = true;
                      break;
                    }
                  }
                  if (!found && self._elemNm === "ojcombobox")
                  {
                    //If user entered value which is not listed in predefiend options
                    ordered.push({'value' : id, 'label' : id});
                  }
                }
                callback(ordered);
              }
            }
            );
          };
        }
      }
      return opts;
    },

    _selectChoice : function (choice)
    {
      var selected = this.container.find("." + this._classNm + "-selected-choice.oj-focus");
      if (selected.length && choice && choice[0] == selected[0])
      {}
      else
      {
        if (selected.length)
        {
          this.opts.element.trigger("choice-deselected", selected);
        }
        selected.removeClass("oj-focus");
        if (choice && choice.length)
        {
          if (this._elemNm === "ojcombobox")
            this.close();
          choice.addClass("oj-focus");
          this.container.find("." + this._classNm + "-description").text(choice.attr("valueText") + ". Press back space to delete.")
          .attr("aria-live", "assertive");
          this.opts.element.trigger("choice-selected", choice);
        }
      }
    },

    _destroy : function ()
    {
      $("label[for='" + this.search.attr('id') + "']")
      .attr('for', this.opts.element.attr("id"));
      _AbstractMultiChoice.superclass._destroy.apply(this, arguments);
    },

    _initContainer : function ()
    {
      var selector = "." + this._classNm + "-choices",
      selection,
      idSuffix = _ComboUtils.nextUid(),
      elementLabel;

      this.searchContainer = this.container.find("." + this._classNm + "-search-field");
      this.selection = selection = this.container.find(selector);

      var _this = this;
      this.selection.on("click", "." + this._classNm + "-selected-choice:not(." + this._classNm + "-locked)", function (e)
      {
        if (this._elemNm === "ojcombobox")
          _this.search[0].focus(); //Fixed??
        _this._selectChoice($(this));
      }
      );

      elementLabel = $("label[for='" + this._getAttribute("id") + "']");
      if (!elementLabel.attr("id"))
        elementLabel.attr('id', this._classNm + "-label-" + idSuffix);
      
      this._contentElement = (this._elemNm === "ojcombobox") ? this.search : this.selection;

      // add aria associations
      selection.find("." + this._classNm + "-input").attr("id", this._classNm + "-input-" + idSuffix);
      if (!this.results.attr("id"))
        this.results.attr("id", "oj-listbox-results-" + idSuffix);
      this._contentElement.attr("aria-owns", this.results.attr("id"));
      this._contentElement.attr("aria-labelledby", elementLabel.attr("id"));
      this.opts.element.attr("aria-labelledby", elementLabel.attr("id"));

      if (this.search.attr('id'))
        elementLabel.attr('for', this.search.attr('id'));

      if (this._getAttribute("aria-label"))
        this._contentElement.attr("aria-label", this._getAttribute("aria-label"));

      if (this._getAttribute("aria-controls"))
        this._contentElement.attr("aria-controls", this._getAttribute("aria-controls"));

      if (this.elementTabIndex)
        this._contentElement.attr("tabindex", this.elementTabIndex);
      
      this.keydowns = 0;
      // Add keydown keyup handler on the select box for ojselect
      if (this._elemNm === "ojselect") 
      {
        this.selection.on("keydown", this._bind(this._containerKeydownHandler));
        this.selection.on("keyup", this._bind(function (e)
        {
          this.keydowns = 0;
        }
        ));
      };
      this.search.on("keydown", this._bind(this._containerKeydownHandler));

      this.search.on("keyup", this._bind(function (e)
        {
          this.keydowns = 0;
        }
        ));

      this.search.on("input", this._bind(function (e)
        {
          this.ojContext._SetRawValue(this.search.val(), e);
        }
        ));

      this.search.on("blur keyup", this._bind(function (e)
        {
          if (e.type === 'keyup' && e.keyCode !== 10 && e.keyCode !== 13)
            return;

          if (this.opts["manageNewEntry"] && this.search.val() && this.results.find(".oj-hover").length <= 0)
          {
            var data = this.opts["manageNewEntry"](this.search.val());

            var trigger = e.type === "blur"
               ? _ComboUtils.ValueChangeTriggerTypes.BLUR
               : _ComboUtils.ValueChangeTriggerTypes.ENTER_PRESSED;
            var options =
            {
              trigger : trigger
            };

            this._onSelect(data, options, e);
          }
          this.search.removeClass(this._classNm + "-focused");
          this.container.removeClass("oj-focus");
          this._selectChoice(null);
          if (!this._opened() && this._classNm !== "oj-select")
            this._clearSearch();
          e.stopImmediatePropagation();
        }
        ));

      this.container.on("click touchstart", selector, this._bind(function (e)
        {
          if (!this._isInterfaceEnabled())
            return;
          if ($(e.target).closest("." + this._classNm + "-selected-choice").length > 0)
          {
            // clicked inside a selected choice, do not open
            return;
          }
          this._selectChoice(null);
          if (this._opened())
          {
            this.close(e);
          }
          else
          {
            this.open(e);
            if (this._elemNm === "ojcombobox" || this._hasSearchBox())
              this._focusSearch();
          }
          e.preventDefault();
        }
        ));

      this.container.on("focus", selector, this._bind(function ()
        {
          if (!this._isInterfaceEnabled())
            return;
        }
        ));

      this._initContainerWidth();
      this.opts.element.hide().attr("aria-hidden", true);
      this.container.append(this.opts.element); //@HTMLUpdateOK

      // set the placeholder if necessary
      this._clearSearch();
    },

    _containerKeydownHandler : function (e)
    {
      if (!this._isInterfaceEnabled())
        return;

      ++this.keydowns;
      var selected = this.selection.find("." + this._classNm + "-selected-choice.oj-focus");
      var prev = selected.prev("." + this._classNm + "-selected-choice:not(." + this._classNm + "-locked)");
      var next = selected.next("." + this._classNm + "-selected-choice:not(." + this._classNm + "-locked)");
      var pos = (this._elemNm === "ojselect" && !this._userTyping) ? _ComboUtils.getCursorInfo(this.selection) : _ComboUtils.getCursorInfo(this.search);

      if (selected.length &&
        (e.which == _ComboUtils.KEY.LEFT || e.which == _ComboUtils.KEY.RIGHT || e.which == _ComboUtils.KEY.BACKSPACE || e.which == _ComboUtils.KEY.DELETE || e.which == _ComboUtils.KEY.ENTER))
      {
        var selectedChoice = selected;
        if (e.which == _ComboUtils.KEY.LEFT && prev.length)
        {
          selectedChoice = prev;
        }
        else if (e.which == _ComboUtils.KEY.RIGHT)
        {
          selectedChoice = next.length ? next : null;
        }
        else if (e.which === _ComboUtils.KEY.BACKSPACE)
        {
          this._unselect(selected.first(), e);
          this._resetSearchWidth();
          selectedChoice = prev.length ? prev : next;
        }
        else if (e.which == _ComboUtils.KEY.DELETE)
        {
          this._unselect(selected.first(), e);
          this._resetSearchWidth();
          selectedChoice = next.length ? next : null;
        }
        else if (e.which == _ComboUtils.KEY.ENTER)
        {
          selectedChoice = null;
        }

        this._selectChoice(selectedChoice);
        e.preventDefault();
        if (!selectedChoice || !selectedChoice.length)
        {
          this.open(e);
        }
        return;
      }
      else if (((e.which === _ComboUtils.KEY.BACKSPACE && this.keydowns == 1)
           || e.which == _ComboUtils.KEY.LEFT) && (pos.offset == 0 && !pos.length))
      {
        this._selectChoice(this.selection.find("." + this._classNm + "-selected-choice:not(." + this._classNm + "-locked)").last());
        e.preventDefault();
        return;
      }
      else
      {
        this._selectChoice(null);
      }

      if (this._opened())
      {
        switch (e.which)
        {
        case _ComboUtils.KEY.UP:
        case _ComboUtils.KEY.DOWN:
          this._moveHighlight((e.which === _ComboUtils.KEY.UP) ? -1 : 1);
          e.preventDefault();
          return;
        case _ComboUtils.KEY.ENTER:
          this._selectHighlighted(null, e);
          e.preventDefault();
          return;
        case _ComboUtils.KEY.TAB:
          this.close(e);
          return;
        case _ComboUtils.KEY.ESC:
          this._cancel(e);
          e.preventDefault();
          return;
        }
        // bring up the search box if user started typing after the drop down is already opened
        // do not close the drop down if user pressed control keys or function keys
        if (this._userTyping === false &&
            !(_ComboUtils.KEY.isControl(e) || _ComboUtils.KEY.isFunctionKey(e))) 
        {
          this._userTyping = true;
          this.close();
        }
      }

      if (e.which === _ComboUtils.KEY.TAB || _ComboUtils.KEY.isControl(e) || _ComboUtils.KEY.isFunctionKey(e)
         || e.which === _ComboUtils.KEY.ESC)
      {
        return;
      }

      // when user typed in text and hit enter, we don't want to open drop down
      if (e.which === _ComboUtils.KEY.ENTER && this.search.val() && this._elemNm === "ojcombobox")
        return;

      switch (e.which)
      {
      case _ComboUtils.KEY.UP:
      case _ComboUtils.KEY.DOWN:
        this.open(e);
        e.preventDefault();
        return;
      case _ComboUtils.KEY.PAGE_UP:
      case _ComboUtils.KEY.PAGE_DOWN:
        // prevent the page from scrolling
        e.preventDefault();
        return;
      case _ComboUtils.KEY.ENTER:
        // prevent form from being submitted
        e.preventDefault();
        return;
      }
      
      // ojselect: used by select
      this._userTyping = true;
    },

    _enableInterface : function ()
    {
      if (_AbstractMultiChoice.superclass._enableInterface.apply(this, arguments))
      {
        this.search.prop("disabled", !this._isInterfaceEnabled());
      }
    },

    _initSelection : function (valueOptions)
    {
      var data;
      if ((this.getVal() === null || this.getVal().length === 0) && (this._classNm === "oj-select" || this.opts.element.text().trim() === ""))
      {
        this._updateSelection(valueOptions? valueOptions : []);
        this.close();
        // set the placeholder if necessary
        this._clearSearch();
      }
      if (this.datalist || (this.getVal() !== null && this.getVal().length))
      {
        var self = this,
        element;
        if (this.datalist)
          element = this.datalist;
        else
          element = this.opts.element;

        //  - need to be able to specify the initial value of select components bound to dprv
        if (! _ComboUtils.applyValueOptions(this, this.opts)) {
          this.opts.initSelection.call(null, element, function (data)
          {
            if (data !== undefined && data !== null && data.length !== 0)
            {
              self._updateSelection(data);
              self.close();
              // set the placeholder if necessary
              self._clearSearch();
            }
          });                                      
        }
      }
    },

    close : function (event)
    {
      if (!this._opened())
        return;
      _AbstractMultiChoice.superclass.close.apply(this, arguments);
    },

    _focus : function ()
    {
      this.close();
      this.search.focus();
    },

    _updateSelection : function (data)
    {
      var ids = [],
      filtered = [],
      self = this;

      // filter out duplicates
      $(data).each(function ()
      {
        if (ids.indexOf(self.id(this)) < 0)
        {
          ids.push(self.id(this));
          filtered.push(this);
        }
      }
      );
      data = filtered;

      if (this.opts.fetchType == "init" && data && data.length > 0 &&
          (this._classNm == "oj-combobox" || this._classNm == "oj-select")) {
        this.setValOpts(data);
      }

      this.selection.find("." + this._classNm + "-selected-choice").remove();
      this.selection.find(".oj-select-default").remove();
      $(data).each(function ()
      {
        self._addSelectedChoice(this);
      }
      );

      // Storing this data so that it will be used when setting the display value.
      this.currentItem = data;

      this.opts.element.val(ids.length === 0 ? "" : ids.join(this.opts.separator));


      self._postprocessResults();
    },

    _onSelect : function (data, options, event)
    {
      if (!this._triggerSelect(data))
      {
        return;
      }

      var context;
      if (options && options.trigger)
      {
        context =
        {
          optionMetadata :
          {
            "trigger" : options.trigger
          }
        };
      }

      //selection will be added when _SetValue is called
      //this._addSelectedChoice(data);
      var id = this.id(data);
      //Clone the value before invoking setVal(), otherwise it will not trigger change event.
      var val = this.getVal() ? this.getVal().slice(0) : [];
      var valOpts = this.getValOpts() ? this.getValOpts().slice(0) : [];
      var isSelectCombobox = (this._classNm === "oj-combobox" || this._classNm === "oj-select");

      // If the component is invalid, we will not get all the values matching the displayed value
      if (!this.ojContext.isValid()) {
        val = _ComboUtils.splitVal(this.opts.element.val(), this.opts.separator);
      }

      var self = this;
      $(data).each(function (index)
      {
        if (val.indexOf(id) < 0 && id !== "")
        {
          val.push(id);
          if (isSelectCombobox) {
            // - oj.tests.input.combobox.testcombobox display value mismatch automation failure
            if (Array.isArray(data)) {
              valOpts.push(_ComboUtils.findOptionFromResult(self, id, data[index]));
            }
            else {
              valOpts.push(_ComboUtils.findOptionFromResult(self, id, data));
            }
          }
        }
      });

      //setValueOptions before setVal so that new entry already in valueOptions
      //this is to avoid looking for new entry on the dataProvider
      if (isSelectCombobox) {
        this._skipSetValueOptions = true;
        this.setValOpts(valOpts);
      }
      this.setVal(val, event, context);
      this._skipSetValueOptions = false;
      
      if (this.select || !this.opts.closeOnSelect)
        this._postprocessResults(data, false, this.opts.closeOnSelect === true);
      if (this.opts.closeOnSelect)
      {
        this.close(event);
        this._resetSearchWidth();
      }

      if ((!options || !options.noFocus) && this._elemNm === "ojcombobox")
        this._focusSearch();
    },

    _cancel : function (event)
    {
      this.close(event);
      if (this._elemNm === "ojcombobox")
        this._focusSearch();
    },

    _addSelectedChoice : function (data)
    {
      var enableChoice = !data.locked,
      enabledItem = $(
          "<li class='" + this._classNm + "-selected-choice'>" +
          "    <div></div>" +
          "    <a href='#' onclick='return false;' role='button' aria-label='remove' class='" + this._classNm + "-clear-entry " +
          "      oj-component-icon oj-clickable-icon-nocontext " + this._classNm + "-clear-entry-icon' tabindex='-1'>" +
          "    </a>" +
          "</li>"),
      disabledItem = $(
          //"<li class='oj-combobox-selected-choice oj-combobox-locked'>" +
          "<li class='" + this._classNm + "-selected-choice " + this._classNm + "-locked'>" +
          "<div></div>" +
          "</li>");
      var choice = enableChoice ? enabledItem : disabledItem,
      id = this.id(data),
      formatted;

      formatted = this.opts.formatSelection(data);
      if (formatted !== undefined)
      {
        choice.find("div").addClass(this._classNm + "-selected-choice-label").text(formatted);
        choice.find("." + this._classNm + "-clear-entry").attr("aria-label", formatted + " remove");
        choice.attr("valueText", formatted);
      }
      if (enableChoice)
      {
        choice.find("." + this._classNm + "-clear-entry")
        .on("mousedown", _ComboUtils.killEvent)
        .on("click dblclick", this._bind(function (e)
          {
            if (!this._isInterfaceEnabled())
              return;

            $(e.target).closest("." + this._classNm + "-selected-choice").fadeOut('fast', this._bind(function ()
              {
                this._unselect($(e.target), e);
                this.selection.find("." + this._classNm + "-selected-choice.oj-focus").removeClass("oj-focus");
                this.close(e);
                if (this._elemNm === "ojcombobox")
                  this._focusSearch();
              }
              )).dequeue();
            _ComboUtils.killEvent(e);
          }
          ));
      }
      choice.data(this._elemNm, data);

      // searchContainer is initialized in _initContainer() method.
      // And this can not be changed by an external developer. It is constructed by component only.
      if (this._elemNm === "ojcombobox")
        choice.insertBefore(this.searchContainer); // @HtmlUpdateOk
      else
        this.selection.append(choice); // @HtmlUpdateOk

    },

    //keep valueOptions in sync with value
    _syncValueOptions: function (ojContext, value, valueOptions)
    {
      var newValOpts = [];
      if (value && valueOptions) {
        for (var i = 0; i < value.length; i++) {
          for (var j = 0; j < valueOptions.length; j++) {
            var newOpt = valueOptions[j];
            if (oj.Object.compareValues(newOpt['value'], value[i])) {
              newValOpts.push(newOpt);
            }
          }
        }
        _ComboUtils.setValueOptions(ojContext, newValOpts);
      }
    },

    _unselect : function (selected, event)
    {
      var val = this.getVal() ? this.getVal().slice(0) : [],
      data,
      index;
      selected = selected.closest("." + this._classNm + "-selected-choice");

      if (selected.length === 0)
      {
        //TODO: translation string
        throw "Invalid argument: " + selected + ". Must be ." + this._classNm + "-selected-choice";
      }
      data = selected.data(this._elemNm);
      if (!data)
      {
        // prevent a race condition when the 'x' is clicked really fast repeatedly the event can be queued
        // and invoked on an element already removed
        return;
      }

      // If the component is invalid, we will not get all the values matching the displayed value
      if (!this.ojContext.isValid())
        val = _ComboUtils.splitVal(this.opts.element.val(), this.opts.separator);

      while ((index = val.indexOf(this.id(data))) >= 0)
      {
        val.splice(index, 1);
        this._syncValueOptions(this.ojContext, val, this.getValOpts());
        this._skipSetValueOptions = true;
        this.setVal(val, event);
        this._skipSetValueOptions = false;
        if (this.select)
          this._postprocessResults();
      }

      selected.remove();
    },

    _postprocessResults : function (data, initial, noHighlightUpdate)
    {
      var val = (this.getVal() && (this.opts.element.val() || this.ojContext.isValid())) ? this.getVal() : [],
      choices = this.results.find(".oj-listbox-result"),
      compound = this.results.find(".oj-listbox-result-with-children"),
      self = this;

      _ComboUtils.each2(choices, function (i, choice)
      {
        var id = self.id(choice.data(self._elemNm));
        if (val && val.indexOf(id) >= 0)
        {
          choice.addClass("oj-selected");
          // mark all children of the selected parent as selected
          choice.find(".oj-listbox-result-selectable").addClass("oj-selected");
        }
      }
      );
      _ComboUtils.each2(compound, function (i, choice)
      {
        // hide an optgroup if it doesnt have any selectable children
        if (!choice.is(".oj-listbox-result-selectable")
           && choice.find(".oj-listbox-result-selectable:not(.oj-selected)").length === 0)
        {
          choice.addClass("oj-selected");
        }
      }
      );

      if (!choices.filter('.oj-listbox-result:not(.oj-selected)').length > 0 && this._classNm !== "oj-select")
        this.close();
    },
    
    _isDataSelected : function (data)
    {
      var val = this.getVal();
      
      if (!val || val.length == 0)
        return false;
      
      var results = data.results;
      for (var i=0; i < results.length; i++) 
      {
        if (val.indexOf(this.id(results[i])) == -1)
          return false;
      }
      
      return true;
    },
    
    _resetSearchWidth : function ()
    {
      //do nothing. subclass override
    },

    setVal : function (val, event, context)
    {
      var unique;
      unique = [];

      if (typeof val === "string")
        val = _ComboUtils.splitVal(val, this.opts.separator);
      // filter out duplicates
      for (var i = 0; i < val.length; i++)
      {
        if (unique.indexOf(val[i]) < 0)
          unique.push(val[i]);
      }

      var options =
      {
        doValueChangeCheck : false
      };
      if (context)
        options["_context"] = context;

      // - need to be able to specify the initial value of select components bound to dprv
      //set valueOption
      if (! this._skipSetValueOptions) {
        var queryResult = _ComboUtils.getLastQueryResult(this);
        var match;
        if (queryResult) {
          match = _ComboUtils.findOptions(queryResult, val);
        }
        if (match && match.length) {
          _ComboUtils.setValueOptions(this.ojContext, match);
        }
      }
      this.ojContext._SetValue(unique, event, options);
      if (this.ojContext.isValid() || unique.length === 0)
        this.opts.element.val(unique.length === 0 ? "" : unique.join(this.opts.separator));

      this.search.attr("aria-activedescendant", this.opts.element.attr("id"));
    }
  }
  );

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2012 Igor Vaynberg
 *
 * This software is licensed under the Apache License, Version 2.0 (the "Apache License") or the GNU
 * General Public License version 2 (the "GPL License"). You may choose either license to govern your
 * use of this software only upon the condition that you accept all of the terms of either the Apache
 * License or the GPL License.
 *
 * You may obtain a copy of the Apache License and the GPL License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 * http://www.gnu.org/licenses/gpl-2.0.html
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * Apache License or the GPL Licesnse is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the Apache License and the GPL License for
 * the specific language governing permissions and limitations under the Apache License and the GPL License.
 */
/**
 * @private
 */
var _OjMultiCombobox = _ComboUtils.clazz(_AbstractMultiChoice,
  {
    _elemNm : "ojcombobox",
    _classNm : "oj-combobox",

    _createContainer : function ()
    {
      var container = $(document.createElement("div")).attr(
        {
          "class" : "oj-combobox oj-combobox-multi oj-component"
        }
        ).html([  //@HTMLUpdateOK
            "<ul class='oj-combobox-choices'>",
            "  <li class='oj-combobox-search-field'><span style='display:none'>&nbsp;</span>",
            "    <input type='text' role='combobox' aria-expanded='false' aria-autocomplete='list' autocomplete='off' autocorrect='off' autocapitalize='off' spellcheck='false' class='oj-combobox-input'>",
            "  </li>",
            "</ul>",
            "<div class='oj-combobox-description oj-helper-hidden-accessible'/>",
            "<div class='oj-listbox-drop oj-listbox-drop-multi' style='display:none'>",
            "   <ul class='oj-listbox-results' role='listbox'>",
            "   </ul>",
            "</div>"].join("")); //@HTMLUpdateOK
      return container;
    },

    _opening : function (event, dontUpdateResults)
    {
      //if beforeExpand is not cancelled
      // beforeExpand event will be triggered in base class _shouldOpen method
      this._resizeSearch();
      _OjMultiCombobox.superclass._opening.apply(this, arguments);
      this._focusSearch();

      if (!dontUpdateResults)
        this._updateResults(true);

      this.search.focus();
    },

    _clearSearch : function ()
    {
      var placeholder = this._getPlaceholder(),
      maxWidth = this._getMaxSearchWidth();

      // - need to be able to specify the initial value of select components bound to dprv
      if (placeholder !== undefined && 
          ((!this.getVal() || this.getVal().length === 0) && 
           !(this.getValOpts() && this.getValOpts().length)))
      {
        this.search.attr("placeholder", placeholder);
        // stretch the search box to full width of the container so as much of the placeholder is visible as possible
        // we could call this._resizeSearch(), but we do not because that requires a sizer and we do not want to create one so early because of a firefox bug, see #944
        this.search.val("").width(maxWidth > 0 ? maxWidth : this.container.css("width"));

        //  when the component is pre-created, the input box would get the default size
        this.searchContainer.width("100%");
      }
      else
      {
        this.search.attr("placeholder", "");
        this.search.val("").width(10);

        // reset the search container, so the input doesn't go to the next line if there is still room
        this.searchContainer.width("auto");
      }
    },
    
    _resetSearchWidth : function ()
    {
      this.search.width(10);
    },

    _getMaxSearchWidth : function ()
    {
      return this.selection.width() - _ComboUtils.getSideBorderPadding(this.search);
    },

    _textWidth : function (text)
    {
      var textSpan = document.createElement("span"),
      textNode = document.createTextNode(text);

      textSpan.style.display = "none";
      textSpan.appendChild(textNode); //@HTMLUpdateOK
      $('body').append(textSpan); //@HTMLUpdateOK
      var width = $('body').find('span:last').width();
      $('body').find('span:last').remove();
      return width;
    },

    _resizeSearch : function ()
    {
      var minimumWidth,
      left,
      maxWidth,
      containerLeft,
      searchWidth,
      sideBorderPadding = _ComboUtils.getSideBorderPadding(this.search);

      minimumWidth = this._textWidth(this.search.val()) + 10;
      left = this.search.offset().left;
      maxWidth = this.selection.width();
      containerLeft = this.selection.offset().left;
      searchWidth = maxWidth - (left - containerLeft) - sideBorderPadding;
      if (searchWidth < minimumWidth)
      {
        searchWidth = maxWidth - sideBorderPadding;
      }
      if (searchWidth < 40)
      {
        searchWidth = maxWidth - sideBorderPadding;
      }
      if (searchWidth <= 0)
      {
        searchWidth = minimumWidth;
      }
      this.search.width(Math.floor(searchWidth));
    }

  }
  );

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2012 Igor Vaynberg
 *
 * This software is licensed under the Apache License, Version 2.0 (the "Apache License") or the GNU
 * General Public License version 2 (the "GPL License"). You may choose either license to govern your
 * use of this software only upon the condition that you accept all of the terms of either the Apache
 * License or the GPL License.
 *
 * You may obtain a copy of the Apache License and the GPL License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 * http://www.gnu.org/licenses/gpl-2.0.html
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * Apache License or the GPL Licesnse is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the Apache License and the GPL License for
 * the specific language governing permissions and limitations under the Apache License and the GPL License.
 */
/**
 * @private
 */
var _OjMultiSelect = _ComboUtils.clazz(_AbstractMultiChoice,
  {
    _elemNm : "ojselect",
    _classNm : "oj-select",
    _userTyping : false,

    _createContainer : function ()
    {
      var container = $(document.createElement("div")).attr(
        {
          "class" : "oj-select oj-select-multi oj-component"
        }
        ).html([//@HTMLUpdateOK
            "<ul class='oj-select-choices' tabindex='0' role='combobox' ",
            "  aria-autocomplete='none' aria-expanded='false'>",
            "</ul>",
            "<div class='oj-listbox-drop' style='display:none' role='dialog'>",

            "  <div class='oj-listbox-search-wrapper'>",

            "  <div class='oj-listbox-search'>",
            "    <input type='text' autocomplete='off' autocorrect='off' autocapitalize='off'",
            "           spellcheck='false' class='oj-listbox-input' title='Search field' ",
            "           role='combobox' aria-expanded='false' aria-autocomplete='list' />",

            "    <span class='oj-listbox-spyglass-box'>",
            "      <span class='oj-component-icon oj-clickable-icon-nocontext oj-listbox-search-icon' role='presentation'>",
            "       <b role='presentation'></b></span>",
            "    </span>",
            "  </div>",

            "  </div>",

            "   <ul class='oj-listbox-results' role='listbox'>",
            "   </ul>",
            "</div>",

            "<div role='region' class='oj-helper-hidden-accessible oj-listbox-liveregion' aria-live='polite'></div>"
          ].join("")); //@HTMLUpdateOK
      return container;
    },
    
    _containerKeydownHandler : function (e)
    {
      _OjMultiSelect.superclass._containerKeydownHandler.apply(this, arguments);

      if (this._userTyping && !this._opened())
        this.open(e);
    },

    _opening : function (event, dontUpdateResults)
    {
      //if beforeExpand is not cancelled
      // beforeExpand event will be triggered in base class _shouldOpen method
      _OjMultiSelect.superclass._opening.apply(this, arguments);
      
      var searchText = _ComboUtils.getSearchText(event);

      //select: focus still stay on the selectBox if open dropdown by mouse click
      this._showSearchBox(searchText);

      if (!dontUpdateResults)
      {
        if (searchText)
          this._updateResults();
        else
          this._updateResults(true);
      }
    },
    
    close : function (event)
    {
      // reset _userTyping after drop down is closed
      if (this._userTyping === true)
        this._userTyping = false;
      
      _OjMultiSelect.superclass.close.apply(this, arguments);
      
      if (event && (! (event instanceof MouseEvent) ||
          event.target == this.selection || event.target == this.search))
        _ComboUtils._focus(this, this.selection);
    },

    _clearSearch : function ()
    {
      var placeholder = this._getPlaceholder();

      if (placeholder && (!this.getVal() || this.getVal().length === 0))
      {
        var node = $("<li></li>");
        node.addClass("oj-select-default");
        node.text(placeholder);
        this.selection.append(node); //@HTMLUpdateOK
      }
    }

  }
  );

/**
 * @private
 */
var _OjInputSeachContainer = _ComboUtils.clazz(_OjSingleCombobox,
  {
    _elemNm: "ojinputsearch",
    _classNm: "oj-inputsearch",
    
    // _OjInputSeachContainer
    _createContainer: function()
    {
      var container = $(document.createElement("div")).attr(
        {
          "class": "oj-inputsearch oj-component"
        }
      ).html([//@HTMLUpdateOK
        "<div class='oj-inputsearch-choice' tabindex='-1' role='presentation'>",
        "   <input type='text' autocomplete='off' autocorrect='off' autocapitalize='off'",
        "       spellcheck='false' class='oj-inputsearch-input' role='combobox' aria-expanded='false' aria-autocomplete='list' />",
        "   <a class='oj-inputsearch-search-button oj-inputsearch-search-icon oj-component-icon oj-clickable-icon-nocontext'",
        "       role='button' aria-label='search'></a>",
        "</div>",
        "<div class='oj-listbox-drop' style='display:none' role='presentation'>",
        "   <ul class='oj-listbox-results' role='listbox'>",
        "   </ul>",
        "</div>",

        "<div role='region' class='oj-helper-hidden-accessible oj-listbox-liveregion' aria-live='polite'></div>"
      ].join(""));

      var trigger = container.find(".oj-inputsearch-search-button");
      this._attachSearchIconClickHandler(trigger);
      return container;
    },

    _attachSearchIconClickHandler: function(trigger)
    {
      var self = this;
      trigger.on("click", function(event, ui) {
        if (!self._isInterfaceEnabled())
          return;

        if (self.opts["manageNewEntry"]) {

          var value = self.search.val();
          var data = self.opts["manageNewEntry"](value);
          var options = {
            trigger: _ComboUtils.ValueChangeTriggerTypes.SEARCH_ICON_CLICKED
          };

          var selectionData = self.selection.data(self._elemNm);
          if (((!selectionData && value !== "")
              || (selectionData && (selectionData.label !== value))
              || (!self.ojContext.isValid()
                    && value !== self._previousDisplayValue))) {

            self._onSelect(data, options, event);
            self._triggerUpdateEvent(data, options, event);

          } else {
            if (selectionData && selectionData.label === value)
              data = selectionData;

            self._triggerUpdateEvent(data, options, event);
          }
        }
        return false;
      }).on("mousedown", function(event) {
        event.stopPropagation();
        return false;
      });
    },

    // _OjInputSeachContainer
    _enable: function(enabled)
    {
      _OjInputSeachContainer.superclass._enable.apply(this, arguments);

      if (this._enabled)
        this.container.find(".oj-inputsearch-search-button").removeClass("oj-disabled");
      else
        this.container.find(".oj-inputsearch-search-button").addClass("oj-disabled");
    },

    // _OjInputSeachContainer
    // Overriding this method to fire the "update" event which is relevant to
    // only InputSearch
    _triggerUpdateEvent: function(data, context, event)
    {
      var trigger;
      if (context) {
        trigger = context.trigger;
      }

      var options = {
        "_context" : {
            optionMetadata: {
              "trigger": trigger
            }
          }
      };

      var value = this.id(data);
      if (!value || value.length === 0)
      {
        // If the value is entered by user (not by selecting an option) then 
        // only 'label' will be present in the data object.
        value = data['label'] ? data['label'] : [];
      }

      var parsed = this.ojContext._Validate(value, event, options);
      if (parsed === undefined || !this.ojContext.isValid())
      {
        return;
      }

      if(typeof value === "string") {
        value = [value];
      }

      var eventData = {
        "value": value,
        "optionMetadata": {
          "trigger": trigger
        }
      };

      this.ojContext._trigger("update", event, eventData);
    },
    
    //_OjInputSeachContainer
    _prepareOpts : function (opts)
    {
      opts = _OjInputSeachContainer.superclass._prepareOpts.apply(this, arguments);

      opts.highlightTermInOptions = function(query)
      {
        return true;
      };

      return opts;
    }
  }
);
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true,devel:true*/
/**
 * The oj.Optgroup interface defines API for a group option of the JET Select and Combobox.
 * Use <code class="prettyprint">oj.Option</code> for a leaf option and <code class="prettyprint">oj.Optgroup</code> for a group option.
 * <ul>
 * <li><code class="prettyprint">children</code> is a required attribute to group the Option children.</li>
 * <li><code class="prettyprint">label</code> is a required attribute</li>
 * </ul>
 * <p>See <a href="oj.Option.html">oj.Option</a> for a leaf option</p>
 * <p>See {@link oj.ojComboboxOne#options}</p>
 * <p>See {@link oj.ojComboboxMany#options}</p>
 * <p>See {@link oj.ojSelectOne#options}</p>
 * <p>See {@link oj.ojSelectMany#options}</p>
 * @since 4.1.0
 * @export
 * @interface oj.Optgroup
 */

/**
 * Label is a required attribute. It is used to display the group label.
 * 
 * @export
 * @expose
 * @memberof oj.Optgroup
 * @instance
 * @name label
 * @type {string}
 */

/**
 * disabled is an optional attribute. When specified, all options in the group is not selectable.
 * 
 * @export
 * @expose
 * @memberof oj.Optgroup
 * @instance
 * @name disabled
 * @type {boolean}
 * @ojsignature { target: "Type",
 *                value: "?"}
 */

/**
 * Children is a required attribute. It is used to group its Option or Optgroup children.
 * 
 * @export
 * @expose
 * @memberof oj.Optgroup
 * @instance
 * @name children
 * @type {Array.<oj.Option|oj.Optgroup>}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true,devel:true*/
/**
 * The oj.Option interface defines API for a leaf option of the JET select and combobox.
 * Use <code class="prettyprint">oj.Option</code> for a leaf option and <code class="prettyprint">oj.Optgroup</code> for a group option.
 * <ul>
 * <li><code class="prettyprint">value</code> is a required attribute. It must be a row key in the data provider.</li>
 * <li><code class="prettyprint">label</code> is an optional attribute. If missing, String(value) is used.</li>
 * </ul>
 * <p>See <a href="oj.Optgroup.html">oj.Optgroup</a> for group option</p>
 * <p>See {@link oj.ojComboboxOne#options}</p>
 * <p>See {@link oj.ojComboboxMany#options}</p>
 * <p>See {@link oj.ojSelectOne#options}</p>
 * <p>See {@link oj.ojSelectMany#options}</p>
 * @since 4.1.0
 * @export
 * @interface oj.Option
 */

/**
 * value is required attribute. It must be the row key in the data provider because JET Select and Combobox use "value" to fetch the displayed "label".
 * 
 * @export
 * @expose
 * @memberof oj.Option
 * @instance
 * @name value
 * @type {Object}
 */

/**
 * label is an optional attribute. It is the display label for the option item. If it's missing, String(value) will be used.
 * 
 * @export
 * @expose
 * @memberof oj.Option
 * @instance
 * @name label
 * @type {string}
 * @ojsignature { target: "Type",
 *                value: "?"}
 */

/**
 * disabled is an optional attribute. When disabled is true, the option item is not selectable.
 * 
 * @export
 * @expose
 * @memberof oj.Option
 * @instance
 * @name disabled
 * @type {boolean}
 * @ojsignature { target: "Type",
 *                value: "?"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
  
  /**
   * @ojcomponent oj.ojComboboxOne
   * @augments oj.ojCombobox
   * @since 0.6
   * @ojdisplayname Single-select Combobox
   * @ojshortdesc A dropdown list that supports single selection, text input, and search filtering.
   * @ojrole combobox
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojComboboxOne extends ojCombobox<any, ojComboboxOneSettableProperties, any, string>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojComboboxOneSettableProperties extends ojComboboxSettableProperties<any, any, string>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojstatus preview
   *
   * @classdesc
   * <h3 id="comboboxOneOverview-section">
   *   JET Combobox One
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#comboboxOneOverview-section"></a>
   * </h3>
   * <p>Description: JET Combobox One provides support for single-select, text input, and search filtering.</p>
   *
   * <p>A JET Combobox One can be created with the following markup.</p>
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-combobox-one>
   *   &lt;oj-option value="option 1">option 1&lt;/oj-option>
   *   &lt;oj-option value="option 2">option 2&lt;/oj-option>
   *   &lt;oj-option value="option 3">option 3&lt;/oj-option>
   *   &lt;oj-option value="option 4">option 4&lt;/oj-option>
   * &lt;/oj-combobox-one> 
   * </code></pre>
   *
   * {@ojinclude "name":"validationAndMessagingDoc"}
   * 
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDocOne"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDocOne"}
   *
   *
   * {@ojinclude "name":"comboboxCommon"}
   */
   
   /**
   * @ojcomponent oj.ojComboboxMany
   * @augments oj.ojCombobox
   * @since 0.6
   * @ojdisplayname Multi-select Combobox
   * @ojshortdesc A dropdown list that supports multiple selections, text input, and search filtering.
   * @ojrole combobox
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojComboboxMany extends ojCombobox<Array<any>, ojComboboxManySettableProperties, Array<any>, string>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojComboboxManySettableProperties extends ojComboboxSettableProperties<Array<any>, Array<any>, string>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojstatus preview
   * 
   * @classdesc
   * <h3 id="comboboxManyOverview-section">
   *   JET Combobox Many
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#comboboxManyOverview-section"></a>
   * </h3>
   * <p>Description: JET Combobox Many provides support for multi-select, text input, and search filtering.</p>
   *
   * <p>A JET Combobox Many can be created with the following markup.</p>
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-combobox-many>
   *   &lt;oj-option value="option 1">option 1&lt;/oj-option>
   *   &lt;oj-option value="option 2">option 2&lt;/oj-option>
   *   &lt;oj-option value="option 3">option 3&lt;/oj-option>
   *   &lt;oj-option value="option 4">option 4&lt;/oj-option>
   * &lt;/oj-combobox-many> 
   * </code></pre>
   *
   * {@ojinclude "name":"validationAndMessagingDoc"}
   * 
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDocMany"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDocMany"}
   *
   *
   * {@ojinclude "name":"comboboxCommon"}
   */
   
  /**
   * @ojcomponent oj.ojCombobox
   * @augments oj.editableValue
   * @since 0.6
   * @abstract
   * @ojsignature [{
   *                target: "Type",
   *                value: "abstract class ojCombobox<V, SP extends ojComboboxSettableProperties<V>, SV=V, RV=V> extends editableValue<V, SP, SV, RV>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojComboboxSettableProperties<V, SV= V, RV= V> extends editableValueSettableProperties<V, SV, RV>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojstatus preview
   *
   * @classdesc
   */

  /**
   *
   * <h3 id="rtl-section">
   *   Reading direction
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
   * </h3>
   *
   * <p>As with any JET element, in the unusual case that the directionality (LTR or RTL) changes post-init, the Combobox must be <code class="prettyprint">refresh()</code>ed.</p>
   *
   *
   * <h3 id="a11y-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
   * </h3>
   * <p>
   * It is up to the application developer to associate an oj-label to the combobox element.
   * You should put an <code>id</code> on the combobox element, and then set
   * the <code>for</code> attribute on the oj-label to be the combobox element's id.
   * </p>
   * <p>
   * The element will decorate its associated label with required and help
   * information, if the <code>required</code> and <code>help</code> attributes are set.
   * </p>
   *
   * @ojfragment comboboxCommon
   * @memberof oj.ojCombobox
   */     
  oj.__registerWidget("oj.ojCombobox", $['oj']['editableValue'],
  {
    defaultElement : "<input>",
    widgetEventPrefix : "oj",
    options :
    {
      /**
       * A converter instance that duck types {@link oj.Converter}. Or an object literal containing
       * the following properties.
       * <p>
       * When <code class="prettyprint">converter</code> property changes due to programmatic
       * intervention, the element performs various tasks based on the current state it is in. </br>
       *
       * <h4>Steps Performed Always</h4>
       * <ul>
       * <li>Any cached converter instance is cleared and new converter created. The converter hint is
       * pushed to messaging. E.g., notewindow displays the new hint(s).
       * </li>
       * </ul>
       *
       * <h4>Running Validation</h4>
       * <ul>
       * <li>if element is valid when <code class="prettyprint">converter</code> property changes, the
       * display value is refreshed.</li>
       * <li>if element is invalid and is showing messages when
       * <code class="prettyprint">converter</code> property changes, then all messages generated by the
       * element are cleared and full validation run using its current display value.
       * <ul>
       *   <li>if there are validation errors, then <code class="prettyprint">value</code>
       *   property is not updated, and the errors are shown. The display value is not refreshed in this case. </li>
       *   <li>if no errors result from the validation, <code class="prettyprint">value</code>
       *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code>
       *   event on the <code class="prettyprint">value</code> property to clear custom errors. The
       *   display value is refreshed with the formatted value provided by converter.</li>
       * </ul>
       * </li>
       * <li>if element is invalid and has deferred messages when
       * <code class="prettyprint">converter</code> property changes, then the display value is
       * refreshed with the formatted value provided by converter.</li>
       * </ul>
       * </p>
       *
       * <h4>Clearing Messages</h4>
       * <ul>
       * <li>When element messages are cleared in the cases described above, messages created by
       * the element are cleared.</li>
       * <li><code class="prettyprint">messages-custom</code> property is not cleared. Page authors can
       * choose to clear it explicitly when setting the converter property.</li>
       * </ul>
       * </p>
       *
       * @property {string} type - the conveter type registered with the oj.ConverterFactory.
       * Supported type is 'number'. See {@link oj.ConverterFactory} for details. <br/>
       * E.g., <code class="prettyprint">{converter: {type: 'number'}</code>
       * @property {Object=} options - optional Object literal of options that the converter expects.
       * See {@link oj.IntlNumberConverter} for options supported by the number converter.
       * E.g., <code class="prettyprint">{converter: {type: 'number', options: {style: 'decimal'}}</code>
       *
       * @ojshortdesc An object to convert value.
       * @expose
       * @access public
       * @instance
       * @memberof! oj.ojCombobox
       * @type {Object|undefined}
       */
      converter: undefined,
      
      /**
       * Whether to filter the list with the current display value on opening the drop down. This can be used to support search use cases. 
       * This only applies to the initial opening of the drop down. When the user starts typing, the dropdown filters as usual.
       *
       * @example <caption>Initialize the combobox with the <code class="prettyprint">filter-on-open</code> attribute specified:</caption>
       * &lt;oj-combobox-one filter-on-open="rawValue">&lt;/oj-combobox-one>
       *
       * @example <caption>Get or set the <code class="prettyprint">filter-on-open</code> property after initialization:</caption>
       * // getter
       * var filterOnOpenValue = myCombobox.filterOnOpen;
       *
       * // setter
       * myCombobox.filterOnOpen = "rawValue";
       *
       * @ojshortdesc Whether to filter the drop down list on open.
       * @expose
       * @instance
       * @memberof oj.ojComboboxOne
       * @since 4.2.0
       * @ojstatus preview
       * @ojvalue {string} "none"  Show all available options without filtering on open.
       * @ojvalue {string} "rawValue" Filter the drop down list on open with the rawValue (current display value).
       * @default "none"
       */
      filterOnOpen: "none",

      /**
       * The placeholder text to set on the element. The placeholder specifies a short hint that can be displayed before user
       * selects or enters a value.
       *
       * @example <caption>Initialize the combobox with the <code class="prettyprint">placeholder</code> attribute specified:</caption>
       * &lt;oj-combobox-one placeholder="Please select ...">&lt;/oj-combobox-one>
       *
       * @example <caption>Get or set the <code class="prettyprint">placeholder</code> property after initialization:</caption>
       * // getter
       * var placeholderValue = myCombobox.placeholder;
       *
       * // setter
       * myCombobox.placeholder = "Select a value";
       *
       * @name placeholder
       * @ojshortdesc A short hint that can be displayed before user selects or enters a value.
       * @expose
       * @access public
       * @instance
       * @memberof oj.ojComboboxOne
       * @type {string|null|undefined}
       * @default null
       */
      /**
       * The placeholder text to set on the element. The placeholder specifies a short hint that can be displayed before user
       * selects or enters a value.
       *
       * @example <caption>Initialize the combobox with the <code class="prettyprint">placeholder</code> attribute specified:</caption>
       * &lt;oj-combobox-many placeholder="Please select ...">&lt;/oj-combobox-many>
       *
       * @example <caption>Get or set the <code class="prettyprint">placeholder</code> property after initialization:</caption>
       * // getter
       * var placeholderValue = myCombobox.placeholder;
       *
       * // setter
       * myCombobox.placeholder = "Select values";
       *
       * @name placeholder
       * @ojshortdesc A short hint that can be displayed before user selects or enters a value.
       * @expose
       * @access public
       * @instance
       * @memberof oj.ojComboboxMany
       * @type {string|null|undefined}
       * @default null
       */
      placeholder: null,

      
      /**
       * {@ojinclude "name":"comboboxCommonOptions"}
       * @name options
       * @ojshortdesc The option items for the Combobox.
       * @expose
       * @access public
       * @instance
       * @type {null | Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>}
       * @default null
       * @ojsignature { target: "Type",
       *                value: "Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>",
       *                jsdocOverride: true}
       * @memberof oj.ojComboboxOne
       *
       * @example <caption>Initialize the Combobox with the <code class="prettyprint">options</code> specified:</caption>
       * &lt;oj-combobox-one options="[[dataArray]]">&lt;/oj-combobox-one>
       *
       * @example <caption>The options array should contain objects with value and label properties:</caption>
       * var dataArray = [{value: 'option1', label: 'Option 1'}, 
       *                  {value: 'option2', label: 'Option 2', disabled: true}, 
       *                  {value: 'option3', label: 'Option 3'}];
       *
       * @example <caption>Initialize the Combobox with a data provider and data mapping:</caption>
       * &lt;oj-combobox-one options="[[dataProvider]]">&lt;/oj-combobox-one>
       *
       * @example <caption>Data mapping can be used if data doesn't have value and label properties.</caption>
       * // actual field names are "id" and "name"
       * var dataArray = [
       *            {id: 'Id 1', name: 'Name 1'},
       *            {id: 'Id 2', name: 'Name 2'},
       *            {id: 'Id 3', name: 'Name 3'}];
       *
       * // In mapfields, map "name" to "label" and "id" to "value"
       * var mapFields = function(item) {
       *   var data = item['data'];
       *   var mappedItem = {};
       *   mappedItem['data'] = {};
       *   mappedItem['data']['label'] = data['name'];
       *   mappedItem['data']['value'] = data['id'];
       *   mappedItem['metadata'] = {'key': data['id']};
       *   return mappedItem;
       * }; 
       * var dataMapping = {'mapFields': mapFields};
       *
       * var arrayDataProvider = new oj.ArrayDataProvider(dataArray, {idAttribute: 'id'});
       * var dataProvider = new oj.ListDataProviderView(arrayDataProvider, {'dataMapping': dataMapping});
       */
      /**
       * {@ojinclude "name":"comboboxCommonOptions"}
       * @name options
       * @ojshortdesc The option items for the Combobox.
       * @expose
       * @access public
       * @instance
       * @type {null | Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>}
       * @default null
       * @ojsignature { target: "Type",
       *                value: "Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>",
       *                jsdocOverride: true}
       * @memberof oj.ojComboboxMany
       *
       * @example <caption>Initialize the Combobox with the <code class="prettyprint">options</code> specified:</caption>
       * &lt;oj-combobox-many options="[[dataArray]]">&lt;/oj-combobox-many>
       *
       * @example <caption>The options array should contain objects with value and label properties:</caption>
       * var dataArray = [{value: 'option1', label: 'Option 1'}, 
       *                  {value: 'option2', label: 'Option 2', disabled: true}, 
       *                  {value: 'option3', label: 'Option 3'}];
       *
       * @example <caption>Initialize the Combobox with a data provider and data mapping:</caption>
       * &lt;oj-combobox-many options="[[dataProvider]]">&lt;/oj-combobox-many>
       *
       * @example <caption>Data mapping can be used if data doesn't have value and label properties.</caption>
       * // actual field names are "id" and "name"
       * var dataArray = [
       *            {id: 'Id 1', name: 'Name 1'},
       *            {id: 'Id 2', name: 'Name 2'},
       *            {id: 'Id 3', name: 'Name 3'}];
       *
       * // In mapfields, map "name" to "label" and "id" to "value"
       * var mapFields = function(item) {
       *   var data = item['data'];
       *   var mappedItem = {};
       *   mappedItem['data'] = {};
       *   mappedItem['data']['label'] = data['name'];
       *   mappedItem['data']['value'] = data['id'];
       *   mappedItem['metadata'] = {'key': data['id']};
       *   return mappedItem;
       * }; 
       * var dataMapping = {'mapFields': mapFields};
       *
       * var arrayDataProvider = new oj.ArrayDataProvider(dataArray, {idAttribute: 'id'});
       * var dataProvider = new oj.ListDataProviderView(arrayDataProvider, {'dataMapping': dataMapping});
       */
      /**
       * The option items for the Combobox. This attribute can be used instead of providing a list of <code class="prettyprint">oj-option</code> or <code class="prettyprint">oj-optgroup</code> child elements of the Combobox element.
       * This attribute accepts:
       * <ol>
       * <li>an array of <a href="oj.Option.html">oj.Option</a>s and/or <a href="oj.Optgroup.html">oj.Optgroup</a>s.
       *   <ul>
       *   <li>Use <code class="prettyprint">oj.Option</code> for a leaf option.</li>
       *   <li>Use <code class="prettyprint">oj.Optgroup</code> for a group option.</li>
       *   </ul>
       * </li>
       * <li>a data provider of <a href="oj.Option.html">oj.Option</a>s. This data provider must implement <a href="oj.DataProvider.html">oj.DataProvider</a>.
       *   <ul>
       *   <li><code class="prettyprint">value</code> in <code class="prettyprint">oj.Option</code> must be the row key in the data provider.</li>
       *   <li>A maximum of 15 rows will be displayed in the dropdown. If more than 15 results are available then users need to filter further.</li>
       *   <li>If the data provider supports the filter criteria capability including the contains (<code class="prettyprint">$co</code>) operator, JET Combobox will request the data provider to do filtering. Otherwise it will filter internally.</li>
       *   </ul>
       * </li>
       * </ol>
       *
       * @expose
       * @memberof oj.ojCombobox
       * @instance
       * @ojfragment comboboxCommonOptions
       */
      options : null,

      /**
       * Specify the key names to use in the options array. 
       * <p>Depending on options-keys means that the signature of the data does not match what is supported by the options attribute. When using Typescript, this would result in a compilation error.</p>
       * <p>Best practice is to use a <a href="oj.ListDataProviderView.html">oj.ListDataProviderView</a> with data mapping as a replacement.</p>
       * <p>However, for the app that must fetch data from a REST endpoint where the data fields do not match those that are supported by the options attribute, you may use the options-keys with any dataProvider that implements <a href="oj.DataProvider.html">oj.DataProvider</a> interface.</p>
       *
       * @expose
       * @access public
       * @instance
       * @type {?Object}
       * @default null
       * @memberof! oj.ojCombobox
       *
       * @example <caption>Initialize the Combobox with <code class="prettyprint">options-keys</code> specified. This allows the key names to be redefined in the options array.</caption>
       * &lt;oj-combobox-xxx options-keys="[[optionsKeys]]">&lt;/oj-combobox-xxx>
       * @example var optionsKeys = {value : "state_abbr", label : "state_name"};
       * @example <caption>Redefine keys for data with subgroups.</caption>
       * var optionsKeys = {label : "regions", children : "states", 
       *                    childKeys : {value : "state_abbr", label : "state_name"}};
       */
      optionsKeys : {
        /**
         * The key name for the label.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojCombobox
         * @alias optionsKeys.label
         * @type {?string}
         * @default null
         */
        label: null,
        /**
         * The key name for the value.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojCombobox
         * @alias optionsKeys.value
         * @type {?string}
         * @default null
         */
        value: null,
        /**
         * The key name for the children.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojCombobox
         * @alias optionsKeys.children
         * @type {?string}
         * @default null
         */
        children: null,
        /**
         * The object for the child keys.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojCombobox
         * @alias optionsKeys.childKeys
         * @type {?Object}
         * @default null
         * @property {?string} label The key name for the label.
         * @property {?string} value The key name for the value.
         * @property {?string} children The key name for the children.
         * @property {?Object} childKeys The object for the child keys.
         */
         childKeys: null
      },
      /**
       * <p>Attributes specified here will be set on the picker DOM element when it's launched.
       * <p>The supported attributes are <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended to the picker's class and style, if any.
       * Note: 1) picker-attributes is not applied in the native theme.
       * 2) setting this attribute after element creation has no effect.
       *
       * @example <caption>Initialize the combobox specifying a set of attributes to be set on the picker DOM element:</caption>
       * &lt;oj-combobox-one picker-attributes="[[pickerAttributes]]">&lt;/oj-combobox-one>
       * @example var pickerAttributes = {
       *   "style": "color:blue;",
       *   "class": "my-class"
       * };
       *
       * @name pickerAttributes
       * @ojshortdesc The style attributes for the drop down.
       * @expose
       * @memberof oj.ojComboboxOne
       * @instance
       * @type {?Object}
       * @default null
       */
      /**
       * <p>Attributes specified here will be set on the picker DOM element when it's launched.
       * <p>The supported attributes are <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended to the picker's class and style, if any.
       * Note: 1) picker-attributes is not applied in the native theme.
       * 2) setting this attribute after element creation has no effect.
       *
       * @example <caption>Initialize the combobox specifying a set of attributes to be set on the picker DOM element:</caption>
       * &lt;oj-combobox-many picker-attributes="[[pickerAttributes]]">&lt;/oj-combobox-many>
       * @example var pickerAttributes = {
       *   "style": "color:blue;",
       *   "class": "my-class"
       * };
       *
       * @name pickerAttributes
       * @ojshortdesc The style attributes for the drop down.
       * @expose
       * @memberof oj.ojComboboxMany
       * @instance
       * @type {?Object}
       * @default null
       */
      pickerAttributes: null,

      /**
       * {@ojinclude "name":"comboboxCommonOptionRenderer"}
       * @name optionRenderer
       * @ojshortdesc The renderer function that renders the content of each option.
       * @expose
       * @memberof oj.ojComboboxOne
       * @instance
       * @type {function(Object)|null}
       * @default null
       * @example <caption>Initialize the Combobox with a renderer:</caption>
       * &lt;oj-combobox-one option-renderer="[[optionRenderer]]">&lt;/oj-combobox-one>
       * @example var optionRenderer = function(optionContext) {
       *            return optionContext['data']['FIRST_NAME'];
       *          };
       */
      /**
       * {@ojinclude "name":"comboboxCommonOptionRenderer"}
       * @name optionRenderer
       * @ojshortdesc The renderer function that renders the content of each option.
       * @expose
       * @memberof oj.ojComboboxMany
       * @instance
       * @type {function(Object)|null}
       * @default null
       * @example <caption>Initialize the Combobox with a renderer:</caption>
       * &lt;oj-combobox-many option-renderer="[[optionRenderer]]">&lt;/oj-combobox-many>
       * @example var optionRenderer = function(optionContext) {
       *            return optionContext['data']['FIRST_NAME'];
       *          };
       */
      /**      
       * The renderer function that renders the content of an each option.
       * The function should return one of the following: 
       * <ul>
       *   <li>An Object with the following property:
       *     <ul><li>insert: HTMLElement - A DOM element representing the content of the option.</li></ul>
       *   </li>
       *   <li>undefined: If the developer chooses to manipulate the option content directly, the function should return undefined.</li>
       * </ul>
       * 
       * <p>The <code class="prettyprint">option-renderer</code> decides only 
       * how the options' content has to be rendered in the drop down.
       * Once an option is selected from the drop down, 
       * what value has to be displayed in the in input field is decided by the 
       * label field in the data object. See <a href="#options">options</a> 
       * and <a href="#optionsKeys">options-keys</a> for configuring option label and value.
       * </p>
       *
       * <p>The context parameter passed to the renderer contains the following keys:</p>
       * <table class="keyboard-table">
       *   <thead>
       *     <tr>
       *       <th>Key</th>
       *       <th>Description</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td><kbd>component</kbd></td>
       *       <td>A reference to the Combobox element.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>parent</kbd></td>
       *       <td>The parent of the data item. The parent is null for root node.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>index</kbd></td>
       *       <td>The index of the option, where 0 is the index of the first option. In the hierarchical case the index is relative to its parent.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>depth</kbd></td>
       *       <td>The depth of the option. The depth of the first level children under the invisible root is 0.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>leaf</kbd></td>
       *       <td>Whether the option is a leaf or a group.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>data</kbd></td>
       *       <td>The data object for the option.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>parentElement</kbd></td>
       *       <td>The option label element.  The renderer can use this to directly append content.</td>
       *     </tr>
       *   </tbody>
       * </table>
       * 
       * @expose
       * @memberof oj.ojCombobox
       * @instance
       * @ojfragment comboboxCommonOptionRenderer
       */
      optionRenderer: null,

      /**
       * The minimum number of characters a user must type before a options 
       * filtering is performed. Zero is useful for local data with just a few items, 
       * but a higher value should be used when a single character search could match a few thousand items.
       *
       * @expose
       * @memberof oj.ojComboboxOne
       * @name minLength
       * @ojshortdesc The minimum number of characters a user must type before search filtering is performed.
       * @instance
       * @type {?number}
       * @default 0
       * @ojmin 0
       *
       * @example <caption>Initialize the Combobox with the <code class="prettyprint">min-length</code> attribute specified:</caption>
       * &lt;oj-combobox-one min-length="2">&lt;/oj-combobox-one>
       *
       * @example <caption>Get or set the <code class="prettyprint">minLength</code> property after initialization:</caption>
       * // getter
       * var minLengthValue = myCombobox.minLength;
       *
       * // setter
       * myCombobox.minLength = 3;
       *
       */
      /**
       * The minimum number of characters a user must type before a options 
       * filtering is performed. Zero is useful for local data with just a few items, 
       * but a higher value should be used when a single character search could match a few thousand items.
       *
       * @expose
       * @memberof oj.ojComboboxMany
       * @name minLength
       * @ojshortdesc The minimum number of characters a user must type before search filtering is performed.
       * @instance
       * @type {?number}
       * @default 0
       * @ojmin 0
       *
       * @example <caption>Initialize the Combobox with the <code class="prettyprint">min-length</code> attribute specified:</caption>
       * &lt;oj-combobox-many min-length="2">&lt;/oj-combobox-many>
       *
       * @example <caption>Get or set the <code class="prettyprint">minLength</code> property after initialization:</caption>
       * // getter
       * var minLengthValue = myCombobox.minLength;
       *
       * // setter
       * myCombobox.minLength = 3;
       *
       */
      minLength : 0,

      /**
       * <p>The  <code class="prettyprint">rawValue</code> is the read-only property for retrieving 
       * the current value from the input field in string form. The main consumer of
       * <code class="prettyprint">rawValue</code> is a converter.</p>
       * <p>
       * The <code class="prettyprint">rawValue</code> updates on the 'input' javascript event, 
       * so the <code class="prettyprint">rawValue</code> changes as the value of the input is changed. 
       * If the user types in '1,200' into the field, the rawValue will be '1', then '1,', then '1,2', 
       * ..., and finally '1,200'. Then when the user blurs or presses 
       * Enter the <code class="prettyprint">value</code> property gets converted and validated
       * (if there is a converter or validators) and then gets updated if valid.
       * </p>
       * <p>This is a read-only attribute so page authors cannot set or change it directly.</p>
       *
       * @example <caption>Get the <code class="prettyprint">rawValue</code> property after initialization:</caption>
       * // getter
       * var rawValue = myCombobox.rawValue;
       *
       * @ojshortdesc The currently displayed text retrieved from the input field.
       * @expose 
       * @access public
       * @instance
       * @memberof! oj.ojCombobox
       * @type {string|undefined}
       * @since 1.2.1
       * @readonly
       * @ojwriteback
       */
      rawValue: undefined,
      
     /**
      * {@ojinclude "name":"comboboxCommonRequired"}
      *
      * @example <caption>Initialize the Combobox with the <code class="prettyprint">required</code> attribute:</caption>
      * &lt;oj-combobox-one required="[[isRequired]]">&lt;/oj-combobox-one>
      *
      * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
      * // getter
      * var requiredValue = myCombobox.required;
      *
      * // setter
      * myCombobox.required = true;
      *
      * @name required
      * @ojshortdesc Specifies whether a value is required.
      * @expose 
      * @access public
      * @instance
      * @memberof oj.ojComboboxOne
      * @type {boolean}
      * @default false
      * @since 0.7
      * @see #translations
      */
     /**
      * {@ojinclude "name":"comboboxCommonRequired"}
      *
      * @example <caption>Initialize the Combobox with the <code class="prettyprint">required</code> attribute:</caption>
      * &lt;oj-combobox-many required="[[isRequired]]">&lt;/oj-combobox-many>
      *
      * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
      * // getter
      * var requiredValue = myCombobox.required;
      *
      * // setter
      * myCombobox.required = true;
      *
      * @name required
      * @ojshortdesc Specifies whether a value is required.
      * @expose 
      * @access public
      * @instance
      * @memberof oj.ojComboboxMany
      * @type {boolean}
      * @default false
      * @since 0.7
      * @see #translations
      */
     /** 
      * Whether the Combobox is required or optional. When required is set to true, an implicit 
      * required validator is created using the validator factory - 
      * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED).createValidator()</code>.
      * 
      * Translations specified using the <code class="prettyprint">translations.required</code> attribute 
      * and the label associated with the Combobox, are passed through to the options parameter of the 
      * createValidator method. 
      * 
      * <p>
      * When <code class="prettyprint">required</code> property changes due to programmatic intervention, 
      * the Combobox may clear messages and run validation, based on the current state it's in. </br>
      *  
      * <h4>Running Validation</h4>
      * <ul>
      * <li>if element is valid when required is set to true, then it runs deferred validation on 
      * the value. This is to ensure errors are not flagged unnecessarily.
      * </li>
      * <li>if element is invalid and has deferred messages when required is set to false, then 
      * element messages are cleared but no deferred validation is run.
      * </li>
      * <li>if element is invalid and currently showing invalid messages when required is set, then 
      * element messages are cleared and normal validation is run using the current display value. 
      * <ul>
      *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
      *   property is not updated and the error is shown. 
      *   </li>
      *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
      *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code> 
      *   event on the <code class="prettyprint">value</code> property to clear custom errors.</li>
      * </ul>
      * </li>
      * </ul>
      * 
      * <h4>Clearing Messages</h4>
      * <ul>
      * <li>Only messages created by the element are cleared.</li>
      * <li><code class="prettyprint">messages-custom</code> attribute is not cleared.</li>
      * </ul>
      * 
      * </p>
      * 
      * @ojvalue {boolean} false - implies a value is not required to be provided by the user. 
      * This is the default.
      * @ojvalue {boolean} true - implies a value is required to be provided by user and the 
      * input's label will render a required icon. Additionally a required validator - 
      * {@link oj.RequiredValidator} - is implicitly used if no explicit required validator is set. 
      * An explicit required validator can be set by page authors using the validators attribute. 
      * 
      * @expose 
      * @access public
      * @instance
      * @memberof oj.ojCombobox
      * @since 0.7
      * @see #translations
      * @ojfragment comboboxCommonRequired
      */
      required: false,
      
     /**
      * {@ojinclude "name":"comboboxCommonValidators"}
      *
      * @example <caption>Initialize the Combobox with validator object literal:</caption>
      * &lt;oj-combobox-one validators="[[validators]]">&lt;/oj-combobox-one>
      * @example var validators = [{
      *     type: 'regExp', 
      *     options : {
      *       pattern: '[a-zA-Z0-9]{3,}'
      *     }
      *   }];
      * 
      * NOTE: oj.Validation.validatorFactory('numberRange') returns the validator factory that is used 
      * to instantiate a range validator for numbers.
      * 
      * @example <caption>Initialize the Combobox with multiple validator instances:</caption>
      * var validator1 = new MyCustomValidator({'foo': 'A'}); 
      * var validator2 = new MyCustomValidator({'foo': 'B'});
      * var validators = [validator1, validator2];
      * &lt;oj-combobox-one validators="[[validators]]">&lt;/oj-combobox-one>
      *
      * @example <caption>Get or set the <code class="prettyprint">validators</code> property after initialization:</caption>
      * // get one
      * var validator = myCombobox.validators[0];
      *
      * // get all
      * var validators = myCombobox.validators;
      *
      * // set all
      * myCombobox.validators = myValidators;
      *
      * @name validators
      * @ojshortdesc Specifies a list of validators for performing validation by the element.
      * @expose 
      * @access public
      * @instance
      * @memberof oj.ojComboboxOne
      * @type {Array|undefined}
      */
     /**
      * {@ojinclude "name":"comboboxCommonValidators"}
      *
      * @example <caption>Initialize the Combobox with validator object literal:</caption>
      * &lt;oj-combobox-many validators="[[validators]]">&lt;/oj-combobox-many>
      * @example var validators = [{
      *     type: 'regExp', 
      *     options : {
      *       pattern: '[a-zA-Z0-9]{3,}'
      *     }
      *   }];
      * 
      * NOTE: oj.Validation.validatorFactory('numberRange') returns the validator factory that is used 
      * to instantiate a range validator for numbers.
      * 
      * @example <caption>Initialize the Combobox with multiple validator instances:</caption>
      * var validator1 = new MyCustomValidator({'foo': 'A'}); 
      * var validator2 = new MyCustomValidator({'foo': 'B'});
      * var validators = [validator1, validator2];
      * &lt;oj-combobox-many validators="[[validators]]">&lt;/oj-combobox-many>
      * 
      * @example <caption>Get or set the <code class="prettyprint">validators</code> property after initialization:</caption>
      * // get one
      * var validator = myCombobox.validators[0];
      *
      * // get all
      * var validators = myCombobox.validators;
      *
      * // set all
      * myCombobox.validators = myValidators;
      *
      * @name validators
      * @ojshortdesc Specifies a list of validators for performing validation by the element.
      * @expose 
      * @access public
      * @instance
      * @memberof oj.ojComboboxMany
      * @type {Array|undefined}
      */
      /** 
       * List of validators used by element when performing validation. Each item is either an 
       * instance that duck types {@link oj.Validator}, or is an Object literal containing the 
       * properties listed below. Implicit validators created by a element when certain options 
       * are present (e.g. <code class="prettyprint">required</code> attribute), are separate from 
       * validators specified through this attribute. At runtime when the element runs validation, it 
       * combines the implicit validators with the list specified through this attribute. 
       * <p>
       * Hints exposed by validators are shown in the notewindow by default, or as determined by the 
       * 'validatorHint' property set on the <code class="prettyprint">display-options</code> 
       * attribute. 
       * </p>
       * 
       * <p>
       * When <code class="prettyprint">validators</code> property changes due to programmatic 
       * intervention, the element may decide to clear messages and run validation, based on the 
       * current state it is in. </br>
       * 
       * <h4>Steps Performed Always</h4>
       * <ul>
       * <li>The cached list of validator instances are cleared and new validator hints is pushed to 
       * messaging. E.g., notewindow displays the new hint(s).
       * </li>
       * </ul>
       *  
       * <h4>Running Validation</h4>
       * <ul>
       * <li>if element is valid when validators changes, element does nothing other than the 
       * steps it always performs.</li>
       * <li>if element is invalid and is showing messages -
       * <code class="prettyprint">messages-shown</code> property is non-empty, when 
       * <code class="prettyprint">validators</code> changes then all element messages are cleared 
       * and full validation run using the display value on the element. 
       * <ul>
       *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
       *   property is not updated and the error is shown. 
       *   </li>
       *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
       *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code> 
       *   event on the <code class="prettyprint">value</code> property to clear custom errors.</li>
       * </ul>
       * </li>
       * <li>if element is invalid and has deferred messages when validators changes, it does 
       * nothing other than the steps it performs always.</li>
       * </ul>
       * </p>
       * 
       * <h4>Clearing Messages</h4>
       * <ul>
       * <li>Only messages created by the element are cleared.</li>
       * <li><code class="prettyprint">messages-custom</code> property is not cleared.</li>
       * </ul>
       * </p>
       * 
       * @property {string} type - the validator type that has a {@link oj.ValidatorFactory} that can 
       * be retrieved using the {@link oj.Validation} module. For a list of supported validators refer 
       * to {@link oj.ValidatorFactory}. <br/>
       * E.g., <code class="prettyprint">{validators: [{type: 'regExp'}]}</code>
       * @property {Object=} options - optional Object literal of options that the validator expects. 
       * <br/>
       * E.g., <code class="prettyprint">{validators: [{type: 'regExp', options: {pattern: '[a-zA-Z0-9]{3,}'}}]}</code>
       *
       * 
       * @expose 
       * @access public
       * @instance
       * @memberof oj.ojCombobox
       * @ojfragment comboboxCommonValidators
       */    
      validators: undefined,

      /**
       * The <code class="prettyprint">valueOption</code> is similar to the <code class="prettyprint">value</code>, but is an 
       * Object which contains both a value and display label.
       * The <code class="prettyprint">value</code> and <code class="prettyprint">valueOption</code> are kept in sync. 
       * If initially both are set, the selected value in the <code class="prettyprint">value</code> attribute has precedence.
       * <p>Note: There may be a performance gain if the <code class="prettyprint">options</code> attribute is a data provider and
       * the <code class="prettyprint">valueOption</code> is initially set then its label is used before the real data is fetched. 
       * If <code class="prettyprint">valueOption</code> is not specified or the selected value is missing, then the label will 
       * be fetched from the data provider.
       *
       * @name valueOption
       * @ojshortdesc The current value of the element and its associated display label.
       * @expose
       * @instance
       * @type {null | Object}
       * @default null
       *
       * @property {any} value current value of JET Combobox
       * @property {string} [label] display label of value above. If missing, String(value) is used. 
       * @memberof oj.ojComboboxOne
       *
       * @example <caption>Initialize the Combobox with the <code class="prettyprint">value-option</code> attribute specified:</caption>
       * &lt;oj-combobox-one value-option="[[valueOption]]">&lt;/oj-combobox-one>
       *
       * @example <caption>Object with value and label properties:</caption>
       * var valueOption = {'value': 'val1', 'label': 'Label 1'};
       *
       * @example <caption>Get or set the <code class="prettyprint">valueOption</code> property after initialization:</caption>
       * // getter
       * var valueOption = myCombobox.valueOption;
       *
       * // setter
       * myCombobox.valueOption = valueOption;
       */
      valueOption: null,

      /**
       * The <code class="prettyprint">valueOptions</code> is similar to the <code class="prettyprint">value</code>, but is an array 
       * of Objects and each Object contains both a value and display label.
       * The <code class="prettyprint">value</code> and <code class="prettyprint">valueOptions</code> are kept in sync. 
       * If initially both are set, the selected values in the <code class="prettyprint">value</code> attribute has precedence.
       * <p>Note: There may be a performance gain if the <code class="prettyprint">options</code> attribute is a data provider and the 
       * <code class="prettyprint">valueOptions</code> is initially set then its labels are used before the real data is fetched. 
       * If <code class="prettyprint">valueOptions</code> is not specified or one of the selected value is missing, then the labels
       * will be fetched from the data provider.
       *
       * @name valueOptions
       * @ojshortdesc The current values of the element and theirs associated display labels.
       * @expose
       * @instance
       * @type {null | Array.<Object>}
       * @default null
       *
       * @property {any} value a current value of JET Combobox
       * @property {string} [label] display label of value above. If missing, String(value) is used. 
       * @ojsignature { target: "Type",
       *                value: "Array<{value: any, label?: string}> | null",
       *                jsdocOverride: true}
       *
       * @memberof oj.ojComboboxMany
       *
       * @example <caption>Initialize the Combobox with the <code class="prettyprint">value-options</code> attribute specified:</caption>
       * &lt;oj-combobox-many value-options="[[optionsArray]]">&lt;/oj-combobox-many>
       *
       * @example <caption>Array of Objects with value and label properties:</caption>
       * var optionsArray = [{'value': 'val1', 'label': 'Label 1'}, 
       *                     {'value': 'val2', 'label': 'Label 2'}];
       *
       * @example <caption>Get or set the <code class="prettyprint">valueOptions</code> property after initialization:</caption>
       * // getter
       * var valueOptions = myCombobox.valueOptions;
       *
       * // setter
       * myCombobox.valueOptions = optionsArray;
       */
      valueOptions: null,

      /**
       * The value of the element. It supports any type.
       *
       * @example <caption>Initialize the combobox with the <code class="prettyprint">value</code> attribute specified:</caption>
       * &lt;oj-combobox-one value="option1">&lt;/oj-combobox-one>
       *
       * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
       * // getter
       * var value = myCombobox.value;
       *
       * // setter
       * myCombobox.value = "option1";
       *
       * @member
       * @name value
       * @ojshortdesc The value of the element.
       * @access public
       * @instance
       * @memberof oj.ojComboboxOne
       * @type {*}
       * @ojwriteback
       */
      /**
       * The value of the element. The value is an array with any type of items.
       *
       * @example <caption>Initialize the combobox with the <code class="prettyprint">value</code> attribute specified:</caption>
       * &lt;oj-combobox-many value="{{val}}">&lt;/oj-combobox-many>
       * @example var val = ['option1', 'option2'];
       *
       * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
       * // getter
       * var value = myCombobox.value;
       *
       * // setter
       * myCombobox.value = ["option1", "option2"];
       *
       * @member
       * @name value
       * @ojshortdesc The value of the element.
       * @access public
       * @instance
       * @memberof oj.ojComboboxMany
       * @type {Array.<*>}
       */
       
      /**
       * Triggered when the value is submitted by the user through pressing the enter key or selecting from the drop down list.
       * This is to support search use cases.
       * The event will be fired even if the value is the same. This will help the application to re-submit the search query for the same value.
       * The event is not triggered if the submitted value fails validation. The event is not triggered on blur or tab out. 
       *
       * @example <caption>Define an event listener.</caption>
       * var listener = function( event )
       * {
       *   // Get the search term
       *   var term = event['detail']['value'];
       * 
       *   // search with the term
       * };
       * @ojshortdesc Event handler for when the value is submitted by the user.
       * @expose
       * @event
       * @memberof oj.ojComboboxOne
       * @since 4.2.0
       * @ojstatus preview
       * @instance
       * @property {*} value the current value
       * @property {*} previousValue the previous value
       */
      valueUpdated: null
    },

    /**
     * Returns a jQuery object containing the element visually representing the combobox.
     *
     * <p>This method does not accept any arguments.
     *
     * @memberof! oj.ojCombobox
     * @instance
     * @public
     * @ignore
     * @return {jQuery} the combobox
     */
    widget : function ()
    {
      return this.combobox.container;
    },


    /**
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     * @override
     */
    _ComponentCreate : function ()
    {
      this._super();
      _ComboUtils.wrapDataProviderIfNeeded(this, null);
      this._setup();
    },
    
    /**
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     * @override
     */
    _AfterCreate: function () 
    {     
      this._super();
      
      // For custom element syntax, oj-label puts for=forId+"|input" on its label element.
      // So we need to change the input element's id to match that by calling
      // setSubIdForCustomLabelFor on EditableValueUtils.
      if (this._IsCustomElement()) 
      {
        var inputId = this.OuterWrapper.id;
        if (inputId)
          oj.EditableValueUtils.setSubIdForCustomLabelFor(this._GetContentElement()[0], inputId);
      }
    },
    
    /**
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     * @override
     */
    _InitOptions : function (originalDefaults, constructorOptions)
    {
      var props = [{attribute: "disabled", validateOption: true},
                   {attribute: "placeholder"},
                   {attribute: "required",
                    coerceDomValue: true, validateOption: true},
                   {attribute: "title"}
                   // {attribute: "value"}
                 ];

      this._super(originalDefaults, constructorOptions);
      oj.EditableValueUtils.initializeOptionsFromDom(props, constructorOptions, this);
      
      this.multiple = !this._IsCustomElement() ? this.options['multiple'] :
                      this.OuterWrapper.nodeName === "OJ-COMBOBOX-MANY" ?
                      true : false;
      
      if (this.options['value'] === undefined) {
        if (!this._IsCustomElement())
          this.options['value'] = (this.element.attr('value') !== undefined) ? _ComboUtils.splitVal(this.element.val(), ",") : null;
      } else {
        //clone the value, otherwise _setDisplayValue will not be invoked on binding value to ko observableArray.
        //TODO: Need to revisit this once 18724975 is fixed.
        var value = this.options['value'];
        if (Array.isArray(value)) {
          if (!this._IsCustomElement())
            value = value.slice(0);
        } else if(typeof value === "string") {
          if (this.multiple === true)
            value = _ComboUtils.splitVal(value, ",");
          else if (!this._IsCustomElement())
            value = [value];
        }
        this.options['value'] = value;
      }
      
    },
   /**
    * Whether the component is required.
    * 
    * @return {boolean} true if required; false
    * 
    * @memberof! oj.ojCombobox
    * @instance
    * @protected
    * @override
    */
    _IsRequired : function () 
    {
      return this.options['required'];
    },
    /**
     * Performs post processing after required option is set by taking the following steps.
     * 
     * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
     * run full validation with UI value (we don't know if the UI error is from a required validator 
     * or something else);<br/>
     * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
     * updated<br/>
     * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
     * listen to optionChange(value) to clear custom errors.<br/>
     * 
     * - if component is invalid and has messagesHidden -> required: false -> clear component 
     * errors; no deferred validation is run.<br/>
     * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
     * errors unnecessarily)<br/>
     * - messagesCustom is never cleared<br/>
     * 
     * @param {string} option
     * 
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     */
    _AfterSetOptionRequired : oj.EditableValueUtils._AfterSetOptionRequired,
    /**
     * When validators option changes, take the following steps.
     * 
     * - Clear the cached normalized list of all validator instances. push new hints to messaging.<br/>
     * - if component is valid -> validators changes -> no change<br/>
     * - if component is invalid has messagesShown -> validators changes -> clear all component 
     * messages and re-run full validation on displayValue. if there are no errors push value to 
     * model;<br/>
     * - if component is invalid has messagesHidden -> validators changes -> do nothing; doesn't change 
     * the required-ness of component <br/>
     * - messagesCustom is not cleared.<br/>
     * 
     * NOTE: The behavior applies to any option that creates implicit validators - min, max, pattern, 
     * etc. Components can call this method when these options change.
     * 
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     */
    _AfterSetOptionValidators : oj.EditableValueUtils._AfterSetOptionValidators,   
    /**
     * Performs post processing after converter option changes by taking the following steps.
     * 
     * - always push new converter hint to messaging <br/>
     * - if component has no errors -> refresh UI value<br/>
     * - if component is invalid has messagesShown -> clear all component errors and run full 
     * validation using display value. <br/>
     * &nbsp;&nbsp;- if there are validation errors, value is not pushed to model; messagesShown is 
     * updated.<br/>
     * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
     * listen to optionChange(value) to clear custom errors.<br/>
     * - if component is invalid has messagesHidden -> refresh UI value. no need to run deferred 
     * validations. <br/>
     * - messagesCustom is never cleared<br/>
     * 
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     */
    _AfterSetOptionConverter : oj.EditableValueUtils._AfterSetOptionConverter, 
    /**
     * Clears the cached converter stored in _converter and pushes new converter hint to messaging.
     * Called when convterer option changes 
     * @memberof! oj.ojCombobox
     * @instance
     * @protected 
     */
    _ResetConverter : oj.EditableValueUtils._ResetConverter,    
    /**
     * Returns the normalized converter instance.
     * 
     * @return {Object} a converter instance or null
     * @memberof! oj.ojCombobox
     * @instance
     * @protected 
     */
    _GetConverter : oj.EditableValueUtils._GetConverter,    
    /**
     * This returns an array of all validators 
     * normalized from the validators option set on the component. <br/>
     * @return {Array} of validators. 
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     */
    _GetNormalizedValidatorsFromOption : oj.EditableValueUtils._GetNormalizedValidatorsFromOption,
    
    _setup : function ()
    {
      var opts = {},
          self = this;
       
      opts.element = this.element;
      opts.ojContext = this;
      opts = $.extend(this.options, opts);
 
      this.combobox = this.multiple ? new _OjMultiCombobox() : new _OjSingleCombobox();

      this.combobox._init(opts);
      this._refreshRequired(this.options['required']);

      // - need to be able to specify the initial value of select components bound to dprv
      this._resolveValueOptionsLater = _ComboUtils.mergeValueAndValueOptions(this);
    },
    
    /**
     * @memberof! oj.ojCombobox
     * @instance
     * @private 
     */
    _refreshRequired : oj.EditableValueUtils._refreshRequired,    

    /**
     * Called to find out if aria-required is unsupported.
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     */
    _AriaRequiredUnsupported : function()
    {
      return false;
    },
    /**
     * @memberof! oj.ojCombobox
     * @instance
     * @protected
     * @override
     */
    _destroy : function ()
    {
      this.combobox._destroy();
      this._super();
    },

    /**
     * Refreshes the combobox.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof oj.ojCombobox
     * @instance
     * @public
     */
    refresh : function ()
    {
      this._super();

      this.combobox._destroy();
      this._setup();
      this._SetRootAttributes();
    },

    /**
     * Handles options specific to combobox.
     * @override
     * @protected
     * @memberof! oj.ojCombobox
     */
    _setOption : function (key, value, flags)
    {
      if (key === "value") {
       
        if (Array.isArray(value)) {
          if (!this._IsCustomElement())
            value = value.slice(0);
        } 
        else if(typeof value === "string") {
          if (this.multiple === true)
            value = _ComboUtils.splitVal(value, ",");
          else if (!this._IsCustomElement())
            value = [value];
        }
        
        // valueChangeTrigger will be used while setting the display value.
        if (flags && flags["_context"] && flags["_context"].optionMetadata)
        {
          this.combobox.valueChangeTrigger = flags["_context"].optionMetadata["trigger"];
        }
        else
          this.combobox.valueChangeTrigger = null;

        // - placeholder is not displayed after removing selections from select many
        if ((typeof this.options['placeholder'] === 'string') &&
            ((value == null || value && value.length == 0) ||
             (this._IsCustomElement() && value==""))) {
          _ComboUtils.setValueOptions(this, this.multiple? [] : "");
        }
      }

      //if we have a new data provider, remove the old dataProvider event listeners 
      if (key === "options") {
        _ComboUtils.removeDataProviderEventListeners(this);
        _ComboUtils.clearDataProviderWrapper(this);
      }
      this._super(key, value, flags);  

      // - need to be able to specify the initial value of select components bound to dprv
      if (key === "valueOption" && this.multiple !== true) {
        _ComboUtils.syncValueWithValueOption(this, value, null);
      }
      else if (key === "valueOptions" && this.multiple === true && Array.isArray(value)) {
        _ComboUtils.syncValueWithValueOptions(this, value, null);
      }
      //update valueOptions
      else if (key === "value") {
        _ComboUtils.updateValueOptions(this.combobox);
      }

      else if (key === "options") 
      {
        //only add data provider event listeners to new data provider
        if (_ComboUtils.isDataProvider(value)) {
          _ComboUtils.wrapDataProviderIfNeeded(this, this.combobox? this.combobox.opts : null);
          _ComboUtils.addDataProviderEventListeners(this);
        }
        this.combobox.opts.options = value;
        this.combobox.opts = this.combobox._prepareOpts(this.combobox.opts);
      }

      else if (key === "disabled")
      {
        if (value)
          this.combobox._disable();
        else
          this.combobox._enable();
      }
    },
    /**
      * Performs post processing after _SetOption() is called. Different options when changed perform
      * different tasks. See _AfterSetOption[OptionName] method for details.
      *
      * @param {string} option
      * @param {Object|string=} previous
      * @param {Object=} flags
      * @protected
      * @memberof! oj.ojCombobox
      * @instance
      * @override
      */
    _AfterSetOption : function (option, previous, flags)
    {
      this._superApply(arguments);
      switch (option)
      {
        case "required":
          this._AfterSetOptionRequired(option);
          break;
        case "validators":
          this._AfterSetOptionValidators(option);
          break;
        case "converter":
          this._AfterSetOptionConverter(option);
          break;   
        default:
          break;
      }
    },

    //19670748, dropdown popup should be closed on subtreeDetached notification.
    _NotifyDetached : function() {
      this._superApply(arguments);
      this.combobox.close();
    },

    //19670748, dropdown popup should be closed on subtreeHidden notification.
    _NotifyHidden : function() {
      this._superApply(arguments);
      this.combobox.close();
    },

    /**
     * Updates display value of combobox.
     * @override
     * @protected
     * @memberof! oj.ojCombobox
     */
    _SetDisplayValue: function(displayValue)
    {
      //  - need to be able to specify the initial value of select components bound to dprv
      if (_ComboUtils.applyValueOptions(this.combobox, this.options)) {
        if (this.multiple) {
          this.combobox._clearSearch();
        }
      }
      else {
        this.combobox._initSelection();
      }
      this._resolveValueOptionsLater = false;
    },

    /**
     * Set the placeholder.
     * @override
     * @protected
     * @memberof! oj.ojCombobox
     */
    _SetPlaceholder : function(value)
    {
      if (this.combobox)
      {
        this.combobox.opts.placeholder = value;
        // TODO: pavitra - noticed that some combobox tests fail because the _setPlaceholder is
        // undefined, when this method is called from _AfterCreate().
        if (this.combobox._setPlaceholder)
        {
          this.combobox._setPlaceholder();
        }
      }
    },

    /**
     * <ol> 
     * <li>All messages are cleared, including custom messages added by the app. </li>
     * <li>If no converter is present then processing continues to next step. If a converter is 
     * present, the UI value is first converted (i.e., parsed). If there is a parse error then 
     * the messages are shown.</li>
     * <li>If there are no validators setup for the component the <code class="prettyprint">value</code> 
     * option is updated using the display value. Otherwise all 
     * validators are run in sequence using the parsed value from the previous step. The implicit 
     * required validator is run first if the component is marked required. When a validation error is 
     * encountered it is remembered and the next validator in the sequence is run. </li>
     * <li>At the end of validation if there are errors, the messages are shown. 
     * If there were no errors, then the 
     * <code class="prettyprint">value</code> option is updated.</li>
     * </ol>
     *
     * @example <caption>Validate component using its current value.</caption>
     * myComp.validate();
     * 
     * @example <caption>Validate component and use the Promise's resolved state.</caption>
     * myComp.validate().then(
     *  function(result) {
     *    if(result === "valid")
     *    {
     *      submitForm();
     *    }
     *  });
     * @return {Promise} Promise resolves to "valid" if there were no converter parse errors and
     * the component passed all validations. 
     * The Promise resolves to "invalid" if there were converter parse errors or 
     * if there were validation errors.
     * @method
     * @access public
     * @expose
     * @instance
     * @memberof oj.ojCombobox
     * @ojstatus preview
     */
    validate : oj.EditableValueUtils.validate,
    /**
     * Parses the value using the converter set and returns the parsed value. If parsing fails the
     * error is written into the element
     *
     * @override
     * @protected
     * @memberof! oj.ojCombobox
     * @instance
     */
    _parseValue: function(submittedValue)
    {
      var parsedVal;

      if (typeof submittedValue === "string") {
        if (this.multiple === true)
          submittedValue = _ComboUtils.splitVal(submittedValue, ",");
        else if (!this._IsCustomElement())
          submittedValue = [submittedValue];
      }
      if (Array.isArray(submittedValue)) {
        parsedVal = [];
        for (var i = 0; i<submittedValue.length; i++) {
          var parsed = this._super(submittedValue[i]);
          parsedVal.push(parsed);
        }
      } else {
        parsedVal = this._super(submittedValue);
      }
      return parsedVal;
    },

    /**
     * Returns the messaging launcher element  i.e., where user sets focus that triggers the popup.
     * Usually this is the element input or select that will receive focus and on which the popup
     * for messaging is initialized.
     *
     * @override
     * @protected
     * @memberof! oj.ojCombobox
     * @return {Object} jquery element which represents the messaging launcher component
     */
    _GetMessagingLauncherElement : function ()
    {
      return this.combobox.search;
    },

    /**
     * @override
     * @protected
     * @memberof! oj.ojCombobox
     *
     * @param {!Object} menu The JET Menu to open as a context menu.  Always non-<code class="prettyprint">null</code>.
     * @param {!Event} event What triggered the menu launch.  Always non-<code class="prettyprint">null</code>.
     * @param {string} eventType "mouse", "touch", or "keyboard".  Never <code class="prettyprint">null</code>.
     */
    _NotifyContextMenuGesture: function(menu, event, eventType)
    {
      // The default baseComponent behavior in _OpenContextMenu assumes this.element for the
      // launcher. In this case, the original element the component is bound to is
      // hidden (display: none). Pass in an openOption override.
      var launcher = this._GetMessagingLauncherElement();
      this._OpenContextMenu(event, eventType, {"launcher": launcher});
    },

    /**
     * Returns the jquery element that represents the content part of the component.
     * This is usually the component that user sets focus on (tabindex is set 0) and
     * where aria attributes like aria-required, aria-labeledby etc. are set. This is
     * also the element where the new value is updated. Usually this is the same as
     * the _GetMessagingLauncherElement.
     *
     * @override
     * @protected
     * @memberof! oj.ojCombobox
     * @return {jQuery} jquery element which represents the content.
     */
    _GetContentElement : function ()
    {
      return this.combobox.search;
    },

    /**
     * Returns the default styleclass for the component.
     *
     * @return {string}
     * @expose
     * @memberof! oj.ojCombobox
     * @override
     * @protected
     */
    _GetDefaultStyleClass : function ()
    {
    return "oj-combobox";
    },
    
    /**
     * Returns true if validation passes, false otherwise
     *
     * @return {boolean} true if validation passes, false otherwise
     * @memberof! oj.ojCombobox
     * @override
     * @protected
     * @instance
     * @ignore
     */
    _ValidateReturnBoolean: function()
    {
      var displayValue = this.combobox.search.val();
      var newValue = null;

      if (this.multiple !== true) {
        if (!this._IsCustomElement()) {
          if (displayValue === undefined || displayValue === null || displayValue === "")
            newValue = [];
          else
            newValue = [displayValue];
        } else {
          newValue = displayValue;
        }
      } else {
        var existingValue = [];
        if (this.isValid())
            existingValue = this.combobox.getVal();
        
        if (displayValue === undefined || displayValue === null || displayValue === "")
          newValue = existingValue;
        else {
          existingValue.push(displayValue);
          newValue = existingValue;
        }
      }

      return this._SetValue(newValue, null, this._VALIDATE_METHOD_OPTIONS);
    },

    _getDropdown : function ()
    {
      if (this.combobox && this.combobox._opened())
      {
        var dropdown = $(".oj-listbox-drop");
        for (var i=0; i<dropdown.length; i++)
        {
          if ($(dropdown[i]).attr("id") == "oj-listbox-drop" &&
            $(dropdown[i]).attr("data-oj-containerid") == this.combobox.containerId)
          return $(dropdown[i]);
        }
      }
      return null;
    },
    
    _findItem : function (list, value)
    {      
      for (var i=0; i<list.length; i++)
      {
        if ($(list[i]).data("ojcombobox")["value"] === value)
          return list[i];
      }
      return null;
    },

    
    
    //////////////////     SUB-IDS     //////////////////

    /**
     * <p>Sub-ID for the input field</p>
     * @ojsubid oj-combobox-input
     * @memberof oj.ojCombobox
     *
     * @example <caption>Get the input field element</caption>
     * var node = myElement.getNodeBySubId({'subId': 'oj-combobox-input'});
     */

    /**
     * <p>Sub-ID for the drop down arrow of single-select combobox.</p>
     *
     * @ojsubid oj-combobox-arrow
     * @memberof oj.ojComboboxOne
     *
     * @example <caption>Get the drop down arrow of the combobox</caption>
     * var node = myElement.getNodeBySubId({'subId': 'oj-combobox-arrow'});
     */
     
    /**
     * <p>Sub-ID for the list item.</p>
     *
     * @ojsubid oj-listitem
     * @memberof oj.ojCombobox
     *
     * @example <caption>Get the listitem corresponding to value "myVal"</caption>
     * var node = myElement.getNodeBySubId({'subId': 'oj-listitem', 'value': 'myVal'});
     */ 
     
    /**
     * <p>Sub-ID for the remove icon of selected item.</p>
     *
     * @ojsubid oj-combobox-remove
     * @memberof oj.ojComboboxMany
     *
     * @example <caption>Get the element corresponding to the remove icon for the selected item with 
     * value "myVal"</caption>
     * var node = myElement.getNodeBySubId({'subId': 'oj-combobox-remove', 'value': 'myVal'});
     */ 

    /**
     * <p>Sub-ID for the dropdown box.</p>
     *
     * @ojsubid oj-combobox-drop
     * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojCombobox
     * @ignore
     *
     * @example <caption>Get the dropdown box</caption>
     * var node = myElement.getNodeBySubId({'subId': 'oj-combobox-drop'});
     */

    /**
     * <p>Sub-ID for the filtered result list.</p>
     *
     * @ojsubid oj-combobox-results
     * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojCombobox
     * @ignore
     *
     * @example <caption>Get the filtered result list</caption>
     * var node = myElement.getNodeBySubId({'subId': 'oj-combobox-results'});
     */

    /**
     * <p>Sub-ID for the selected items. This returns a
     * list of the selected items.</p>
     *
     * @ojsubid oj-combobox-selection
     * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
     * @memberof oj.ojComboboxMany
     * @ignore
     *
     * @example <caption>Get the list of selected items</caption>
     * var node = myElement.getNodeBySubId({'subId': 'oj-combobox-selection'});
     */

    // @inheritdoc 
    getNodeBySubId: function(locator)
    {
      var node = null, subId;
	    if (locator == null)
	    {
        return this.combobox.container ? this.combobox.container[0] : null;
	    }
      else
      {
        node = this._super(locator);
      }

      if (!node)
      {
        subId = locator['subId'];
        if (subId === "oj-combobox-drop")
          subId = "oj-listbox-drop";

        if (subId === "oj-combobox-results")
          subId = "oj-listbox-results";

        if (subId === "oj-combobox-selection")
          subId = "oj-combobox-selected-choice";

        var dropdown = this._getDropdown();

        switch (subId)
        {
          case "oj-combobox-input":
          case "oj-combobox-arrow":
            node = this.widget().find("." + subId)[0];
            break;
          case "oj-listitem":
            if (dropdown) 
            {
              var list = dropdown.find(".oj-listbox-result");
              node = this._findItem(list, locator['value']);
            }
            break;
          case "oj-combobox-remove":
            var selectedItems = this.widget().find(".oj-combobox-selected-choice");
            var item = this._findItem(selectedItems, locator['value']);
            node = item ? $(item).find(".oj-combobox-clear-entry-icon")[0] : null;
            break;
          case "oj-listbox-drop":
            if (dropdown) {
              node = dropdown[0];
            }
            break;
          case "oj-listbox-results":
            if (dropdown) {
              node = dropdown.find("." + subId)[0];
            }
            break;
          case "oj-combobox-selected-choice":
            node = this.widget().find("." + subId).toArray();
            break;
        }
	    }

	    // Non-null locators have to be handled by the component subclasses
	    return node || null;
    },
    
    /**
     * Returns the subId object for the given child DOM node.  For more details, see
     * <a href="#getNodeBySubId">getNodeBySubId</a>.
     *
     * @expose
     * @override
     * @ignore
     * @memberof oj.ojCombobox
     * @instance
     *
     * @param {!Element} node - child DOM node
     * @return {Object|null} The subId for the DOM node, or <code class="prettyprint">null</code> when none is found.
     *
     * @example <caption>Get the subId for a certain DOM node:</caption>
     * myElement.getSubIdByNode(node);
     */
    getSubIdByNode: function(node)
    {
      var subId = null;
      
      if (node != null)
      {
        var nodeCached = $(node);
                
        if (nodeCached.hasClass('oj-combobox-input'))
          subId = {'subId': 'oj-combobox-input'};
        else if (nodeCached.hasClass('oj-combobox-arrow'))
          subId = {'subId': 'oj-combobox-arrow'};
        else if (nodeCached.hasClass('oj-listbox-result'))
          subId = {'subId': 'oj-listitem', 'value': nodeCached.data('ojcombobox')['value']};
        else if (nodeCached.hasClass('oj-combobox-clear-entry-icon'))
          subId = {'subId': 'oj-combobox-remove', 'value': nodeCached.closest('.oj-combobox-selected-choice').data('ojcombobox')['value']};
        else
          subId = this._super(node);
      }
      
      return subId;      
    }
     
    
  // Fragments:
  
  /**
   * <p>The &lt;oj-combobox-one> element accepts <code class="prettyprint">oj-option</code>s as children. See the [oj-option]{@link oj.ojOption} doc for details about
   * accepted children and slots.</p>
   *
   * @ojchild Default
   * @memberof oj.ojComboboxOne
   *
   * @example <caption>Initialize the Combobox with child content specified:</caption>
   * &lt;oj-combobox-one>
   *   &lt;oj-option value="option1">Option1&lt;/oj-option>
   * &lt;/oj-combobox-one>
   */

  /**
   * <p>The &lt;oj-combobox-many> element accepts <code class="prettyprint">oj-option</code>s as children. See the [oj-option]{@link oj.ojOption} doc for details about
   * accepted children and slots.</p>
   *
   * @ojchild Default
   * @memberof oj.ojComboboxMany
   *
   * @example <caption>Initialize the Combobox with child content specified:</caption>
   * &lt;oj-combobox-many>
   *   &lt;oj-option value="option1">Option1&lt;/oj-option>
   * &lt;/oj-combobox-many>
   */
   
  /**
   * <p>The <code class="prettyprint">end</code> slot is for replacing combobox one's drop down arrow and the divider. 
   * For example, a magnifying glass icon for a search field can be provided in this slot.
   * When the slot is provided with empty content, nothing will be rendered in the slot.
   * When the slot is not provided, the default drop down arrow icon and the divider will be rendered.</p>
   *
   * @ojslot end
   * @since 4.2.0
   * @ojstatus preview
   * @memberof oj.ojComboboxOne
   *
   * @example <caption>Initialize the Combobox one with child content specified for the end slot:</caption>
   * &lt;oj-combobox-one>
   *   &lt;a slot='end' class='mySearchButtonClass'>&lt;/a>
   * &lt;/oj-combobox-one>
   */


	/**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
	 *       <th>Target</th>
	 *       <th>Gesture</th>
	 *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
	 *       <td>Input Field</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td> If the drop down is not open, expand the drop down list. Otherwise, close the drop down list.
   *       If hints, title or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr>
   *     <tr>
	 *       <td>Arrow Button</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td> If the drop down is not open, expand the drop down list. Otherwise, close the drop down list.</td>
   *     </tr>
   *     <tr>
	 *       <td>Option Item</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Tap on an option item in the drop down list to select.</td>
   *     </tr>
   *   </tbody>
   *  </table>
   *
   * <p>Disabled option items receive no highlight and are not selectable.</p>
	 *
	 *
	 * @ojfragment touchDocOne - Used in touch gesture section of classdesc, and standalone gesture doc
	 * @memberof oj.ojComboboxOne
	 */
   
  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
	 *       <th>Target</th>
	 *       <th>Gesture</th>
	 *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
	 *       <td>Input Field</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td> If the drop down is not open, expand the drop down list. Otherwise, close the drop down list.
   *       If hints, title or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr>
   *     <tr>
	 *       <td>Option Item</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Tap on an option item in the drop down list to add to selection.</td>
   *     </tr>
	 *     <tr>
	 *       <td>Selected Item with Clear Entry Button</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Remove item from the selected items list by taping on the clear button next to the data item.</td>
   *     </tr>
   *   </tbody>
   *  </table>
   *
   * <p>Disabled option items receive no highlight and are not selectable.</p>
	 *
	 *
	 * @ojfragment touchDocMany - Used in touch gesture section of classdesc, and standalone gesture doc
	 * @memberof oj.ojComboboxMany
	 */

	/**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *      <td>Option item</td>
   *       <td><kbd>Enter</kbd></td>
   *       <td> Select the highlighted choice from the drop down.</td>
   *     </tr>
   *     <tr>
	 *       <td>Input field</td>
   *       <td><kbd>Enter</kbd></td>
   *       <td>Set the input text as the value.</td>
   *     </tr>
   *     <tr>
   *      <td>Drop down</td>
   *       <td><kbd>UpArrow or DownArrow</kbd></td>
   *       <td> Highlight the option item on the drop down list in the direction of the arrow.
   *         If the drop down is not open, expand the drop down list.</td>
   *     </tr>
   *     <tr>
   *      <td>Drop down</td>
   *       <td><kbd>Esc</kbd></td>
   *       <td> Collapse the drop down list. If the drop down is already closed, do nothing.</td>
   *     </tr>
   *     <tr>
   *      <td>Combobox</td>
   *       <td><kbd>Tab In</kbd></td>
   *       <td>Set focus to the combobox. If hints, title or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr> 
   *   </tbody>
   *  </table>
   *
   * <p>Disabled option items receive no highlight and are not selectable.</p>
	 *
	 * @ojfragment keyboardDocOne - Used in keyboard section of classdesc, and standalone gesture doc
	 * @memberof oj.ojComboboxOne
	 */
   
  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *      <td>Option item</td>
   *       <td><kbd>Enter</kbd></td>
   *       <td> Select the highlighted item from the drop down.</td>
   *     </tr>
   *     <tr>
	 *       <td>Input field</td>
   *       <td><kbd>Enter</kbd></td>
   *       <td>Add the input text to selections.</td>
   *     </tr>
   *     <tr>
   *      <td>Drop down</td>
   *       <td><kbd>UpArrow or DownArrow</kbd></td>
   *       <td> Highlight the option item on the drop down list in the direction of the arrow.
   *         If the drop down is not open, expand the drop down list.</td>
   *     </tr>
   *     <tr>
   *      <td>Combobox</td>
   *       <td><kbd>LeftArrow or RightArrow</kbd></td>
   *       <td> Move focus to the previous or next selected item.</td>
   *     </tr>
   *     <tr>
   *      <td>Drop down</td>
   *       <td><kbd>Esc</kbd></td>
   *       <td> Collapse the drop down list. If the drop down is already closed, do nothing.</td>
   *     </tr>
   *     <tr>
   *      <td>Combobox</td>
   *       <td><kbd>Tab In</kbd></td>
   *       <td>Set focus to the combobox. If hints, title or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr> 
   *   </tbody>
   *  </table>
   *
   * <p>Disabled option items receive no highlight and are not selectable.</p>
	 *
	 * @ojfragment keyboardDocMany - Used in keyboard section of classdesc, and standalone gesture doc
	 * @memberof oj.ojComboboxMany
	 */
  });

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

  /**
   * @ojcomponent oj.ojInputSearch
   * @ignore
   * @augments oj.editableValue
   * @since 1.2.0
   *
   * @classdesc
   * <h3 id="inputSearchOverview-section">
   *   JET InputSearch Component
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputSearchOverview-section"></a>
   * </h3>
   *
   * <p>Description: JET InputSearch enhances a html input into a auto-suggest search input field.</p>
   * 
   * <p>A JET InputSearch can be created with the following markup.</p>
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;input list="items" data-bind="ojComponent: {component: 'ojInputSearch'}"/>
   * &lt;datalist id="items">
   *   &lt;option value="option 1">option 1&lt;/option>
   *   &lt;option value="option 2">option 2&lt;/option>
   *   &lt;option value="option 3">option 3&lt;/option>
   *   &lt;option value="option 4">option 4&lt;/option>
   * &lt;/datalist>
   * </code></pre>
   *
   * <p>Static <code class="prettyprint">options</code> array to provide the option items.</p>
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;input data-bind="ojComponent: {component: 'ojInputSearch', options: 
   *                                     [{value: 'option1', label: 'option1'}, {value: 'option2', label: 'option2'}]}"/>
   * </code></pre>
   * 
   * <p>Options can be provided dynamically based on input instead of statically specifying the options array as shown in the above example</p>
   * 
   * <pre class="prettyprint">
   * <code>
   * &lt;input data-bind="ojComponent: {component: 'ojInputSearch', options: function(optionContext) {
   *                                                                           return new Promise(function(fulfill, reject) {
   *                                                                             var term = context.term;
   *
   *                                                                             // Prepare options based on current 'term'.
   *                                                                             var options = [];
   * 
   *                                                                             fulfill(options);
   *                                                                           }
   *                                                                         }}"/>
   * </code></pre>
   * 
   * <h3 id="styling-section">
   *   Styling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
   * </h3>
   * 
   * {@ojinclude "name":"stylingDoc"}
   *
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   * 
   * {@ojinclude "name":"keyboardDoc"}
   *
   *
   *
   * <h3 id="rtl-section">
   *   Reading direction
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
   * </h3>
   *
   * <p>As with any JET component, in the unusual case that the directionality (LTR or RTL) changes post-init, the InputSearch must be <code class="prettyprint">refresh()</code>ed.</p>
   *
   *
   * <h3 id="pseudos-section">
   *   Pseudo-selectors
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#pseudos-section"></a>
   * </h3>
   *
   * <p>The <code class="prettyprint">:oj-inputsearch</code> pseudo-selector can be used in jQuery expressions to select JET InputSearch.  For example:</p>
   *
   * <pre class="prettyprint">
   * <code>$( ":oj-inputsearch" ) // selects all JET InputSearch on the page
   * $myEventTarget.closest( ":oj-inputsearch" ) // selects the closest ancestor that is a JET InputSearch
   * </code></pre>
   *
   * <h3 id="a11y-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
   * </h3>
   * <p>
   * It is up to the application developer to associate the label to the input component.
   * For InputSearch, you should put an <code>id</code> on the input, and then set 
   * the <code>for</code> attribute on the label to be the input's id.
   * </p>
   * <h3 id="label-section">
   *   Label and InputSearch
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
   * </h3>
   * <p>
   * For accessibility, you should associate a label element with the input
   * by putting an <code>id</code> on the input, and then setting the 
   * <code>for</code> attribute on the label to be the input's id.
   * </p>
   * <p>
   * The component will decorate its associated label with required and help 
   * information, if the <code>required</code> and <code>help</code> options are set. 
   * </p>
   * <h3 id="jqui2jet-section">
   *   JET for jQuery UI developers
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#jqui2jet-section"></a>
   * </h3>
   *
   * <p>Event names for all JET components are prefixed with "oj", instead of component-specific prefixes like "InputSearch".</p>
   *
   * @desc Creates a JET InputSearch.
   * @example <caption>Initialize the InputSearch with no options specified:</caption>
   * $( ".selector" ).ojInputSearch();
   *
   * @example <caption>Initialize the InputSearch with some options:</caption>
   * $( ".selector" ).ojInputSearch( { "minLength": 2, "placeholder": "Search..." } );
   *
   * @example <caption>Initialize the InputSearch via the JET <code class="prettyprint">ojComponent</code> binding:</caption>
   * &lt;div id="search" data-bind="ojComponent: { component: 'ojInputSearch',
   *                                                    minLength: 2}">
   */
  oj.__registerWidget("oj.ojInputSearch", $['oj']['editableValue'],
  {
    defaultElement : "<input>",
    widgetEventPrefix : "oj",
    options :
    {
      /**
       * The placeholder text to set on the element. Though it is possible to set placeholder 
       * attribute on the element itself, the component will only read the value when the component
       * is created. Subsequent changes to the element's placeholder attribute will not be picked up 
       * and page authors should update the option directly.
       * 
       *
       * @example <caption>Initialize the InputSearch with the <code class="prettyprint">placeholder</code> option specified:</caption>
       * $( ".selector" ).ojInputSearch( { "placeholder": "Search ..." } );
       * 
       * @default when the option is not set, the element's placeholder attribute is used if it exists. 
       * 
       * @expose 
       * @access public
       * @instance
       * @memberof! oj.ojInputSearch
       * @type {string|null|undefined}
       */    
      placeholder: undefined,
      /** 
        * Whether the component is required or optional. When required is set to true, an implicit 
        * required validator is created using the validator factory - 
        * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED).createValidator()</code>.
        * 
        * Translations specified using the <code class="prettyprint">translations.required</code> option 
        * and the label associated with the component, are passed through to the options parameter of the 
        * createValidator method. 
        * 
        * <p>
        * When <code class="prettyprint">required</code> option changes due to programmatic intervention, 
        * the component may clear messages and run validation, based on the current state it's in. </br>
        *  
        * <h4>Running Validation</h4>
        * <ul>
        * <li>if component is valid when required is set to true, then it runs deferred validation on 
        * the option value. This is to ensure errors are not flagged unnecessarily.
        * </li>
        * <li>if component is invalid and has deferred messages when required is set to false, then 
        * component messages are cleared but no deferred validation is run.
        * </li>
        * <li>if component is invalid and currently showing invalid messages when required is set, then 
        * component messages are cleared and normal validation is run using the current display value. 
        * <ul>
        *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
        *   option is not updated and the error is shown. 
        *   </li>
        *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
        *   option is updated; page author can listen to the <code class="prettyprint">optionChange</code> 
        *   event on the <code class="prettyprint">value</code> option to clear custom errors.</li>
        * </ul>
        * </li>
        * </ul>
        * 
        * <h4>Clearing Messages</h4>
        * <ul>
        * <li>Only messages created by the component are cleared.</li>
        * <li><code class="prettyprint">messagesCustom</code> option is not cleared.</li>
        * </ul>
        * 
        * </p>
        * 
        * @ojvalue {boolean} false - implies a value is not required to be provided by the user. 
        * This is the default.
        * @ojvalue {boolean} true - implies a value is required to be provided by user and the 
        * input's label will render a required icon. Additionally a required validator - 
        * {@link oj.RequiredValidator} - is implicitly used if no explicit required validator is set. 
        * An explicit required validator can be set by page authors using the validators option. 
        * 
        * @example <caption>Initialize the component with the <code class="prettyprint">required</code> option:</caption>
        * $(".selector").ojInputSearch({required: true});<br/>
        * @example <caption>Initialize <code class="prettyprint">required</code> option from html attribute 'required':</caption>
        * &lt;input list="browsers" required/><br/>
        * // retreiving the required option returns true
        * $(".selector").ojInputSearch("option", "required");<br/>
        * 
        * @example <caption>Customize messages and hints used by implicit required validator when 
        * <code class="prettyprint">required</code> option is set:</caption> 
        * &lt;input list="browsers" required data-bind="ojComponent: {
        *   component: 'ojInputSearch', 
        *   value: password, 
        *   translations: {'required': {
        *                 hint: 'custom: enter at least 3 alphabets',
        *                 messageSummary: 'custom: \'{label}\' is Required', 
        *                 messageDetail: 'custom: please enter a valid value for \'{label}\''}}}"/\>
        * @expose 
        * @access public
        * @instance
        * @default when the option is not set, the element's required property is used as its initial value if it exists.
        * @memberof! oj.ojInputSearch
        * @type {boolean}
        * @default false
        * @since 0.7
        * @see #translations
        */
       required: false,

      /**
       * The id of the html list for the InputSearch.
       *
       * @example <caption>Initialize the InputSearch with the <code class="prettyprint">list</code> option specified:</caption>
       * $( ".selector" ).ojInputSearch( { "list": "list" } );
       *
       * @example <caption>The <code class="prettyprint">list</code> points to a html <code class="prettyprint">ul</code> element.
       * The value for the list item should be specified with <code class="prettyprint">oj-data-value</code> field. By default, we use the first text node for search filtering. An optional <code class="prettyprint">oj-data-label</code> field can be added to the list item, in which case it will take precedence over the text node.</caption>
       * &lt;ul id="list"&gt;
       * &lt;li oj-data-value="li1"&gt;Item 1&lt;/li&gt;
       * &lt;li oj-data-value="li2"&gt;Item 2&lt;/li&gt;
       * &lt;/ul&gt;
       *
       * @expose
       * @memberof! oj.ojInputSearch
       * @instance
       * @type {string|null|undefined}
       */
      list: undefined,

      /**
       * The option items for the InputSearch. Options can be specified as an array of objects containing value and label.
       * The value is used as the value of the option item and label as the label. Both should be of string type. 
       * Group data can be provided with label and a children array containing the option items. 
       * Option item can be set as disabled.
       * 
       * <p>Options can be provided dynamically based on the text typed in the input field 
       * by specifying a function as <code class="prettyprint">options</code> instead of static array. 
       * This function will be invoked when there is a change in input field and it should return a <code class="prettyprint">Promise</code>. 
       * The return type should be an array of objects containing value and label as similar to the static array as mentioned above.</p>
       * 
       * <p>The context paramter passed to the <code class="prettyprint">options</code> function contains the following keys:</p>
       * <table class="keyboard-table">
       *   <thead>
       *     <tr>
       *       <th>Key</th>
       *       <th>Description</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td><kbd>component</kbd></td>
       *       <td>A reference to the InputSearch widget constructor.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>term</kbd></td>
       *       <td>The text based on which options have to be filtered.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>value</kbd></td>
       *       <td>The same options callback function will be used get the value-label object when <code class="prettyprint">value</code> is set programmatically. In such cases context object will have <code class="prettyprint">value</code> instead of <code class="prettyprint">term</code>. Value will be of type Array and when this key is passed in the context, callback function should return array of options for all values.</td>
       *     </tr>
       *   </tbody>
       * </table>
       * 
       * @expose
       * @memberof! oj.ojInputSearch
       * @instance
       * @type {Array|function(Object)}
       *
       * @example <caption>Initialize the InputSearch with the <code class="prettyprint">options</code> specified:</caption>
       * $( ".selector" ).ojInputSearch( { "options": [{value: 'option1', label: 'option1'}, {value: 'option2', label: 'option2', disabled: true}, {value: 'option3', label: 'option3'}] } );
       * 
       * @example <caption>Initialize the InputSearch with group data:</caption>
       * $( ".selector" ).ojInputSearch( { "options": [{label : 'group1', children: [{value: 'option1', label: 'option1'}, {value: 'option2', label: 'option2'}]}, {label: 'group2', children: [{value: 'option3', label: 'option3'}]} ] } );
       * 
       * @example <caption>Initialize the InputSearch <code class="prettyprint">options</code> with a function:</caption>
       * $( ".selector" ).ojInputSearch({ "options": function(optionContext) {
       *                                               return new Promise(function(fulfill, reject) {
       *                                                 var term = context.term;
       *
       *                                                 // Prepare options based on current 'term'.
       *                                                 var options = [];
       * 
       *                                                 fulfill(options);
       *                                               }
       *                                             }});
       */
      options : null,

      /**
       * Specify the key names to use in the options array.
       *
       * @expose
       * @memberof! oj.ojInputSearch
       * @instance
       * @type {Object}
       *
       * @example <caption>Initialize the InputSearch with <code class="prettyprint">optionsKeys</code> specified. This allows the key names to be redefined in the options array.</caption>
       * $( ".selector" ).ojInputSearch( { "optionsKeys": {value : "state_abbr", label : "state_name"} } );
       * 
       * @example <caption>Redefine keys for data with subgroups.</caption>
       * $( ".selector" ).ojInputSearch( { "optionsKeys": {label : "regions", children : "states", childKeys : {value : "state_abbr", label : "state_name"}} } );
       */
      optionsKeys : null,

      /**
       * <p>Attributes specified here will be set on the picker DOM element when it's launched.
       * <p>The supported attributes are <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended to the picker's class and style, if any.
       * Note: 1) pickerAttributes is not applied in the native theme.
       * 2) setting this option after component creation has no effect.
       *
       * @example <caption>Initialize the inputSearch specifying a set of attributes to be set on the picker DOM element:</caption>
       * $( ".selector" ).ojInputSearch({ "pickerAttributes": {
       *   "style": "color:blue;",
       *   "class": "my-class"
       * }});
       *
       * @example <caption>Get the <code class="prettyprint">pickerAttributes</code> option, after initialization:</caption>
       * // getter
       * var inputSearch = $( ".selector" ).ojInputSearch( "option", "pickerAttributes" );
       *
       * @expose
       * @memberof! oj.ojInputSearch
       * @instance
       * @type {?Object}
       * @default <code class="prettyprint">null</code>
       */
      pickerAttributes: null,

      /**      
       * The renderer function that renders the content of an each option.
       * The function must return a DOM element representing the content of the option.
       * If the developer chooses to manipulate the option content directly, 
       * the function should return nothing.
       * 
       * <p>The <code class="prettyprint">optionRenderer</code> decides only 
       * how the options' content has to be rendered in the drop down.
       * Once an option is selected from the drop down, 
       * what value has to be displayed in the in input field is decided by the 
       * label field in the data object. See <a href="#options">options</a> 
       * and <a href="#optionsKeys">optionsKeys</a> for configuring option label and value.
       * </p>
       *
       * <p>The context paramter passed to the renderer contains the following keys:</p>
       * <table class="keyboard-table">
       *   <thead>
       *     <tr>
       *       <th>Key</th>
       *       <th>Description</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td><kbd>component</kbd></td>
       *       <td>A reference to the InputSearch widget constructor.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>parent</kbd></td>
       *       <td>The parent of the data item. The parent is null for root node.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>index</kbd></td>
       *       <td>The index of the option, where 0 is the index of the first option. In the hierarchical case the index is relative to its parent.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>depth</kbd></td>
       *       <td>The depth of the option. The depth of the first level children under the invisible root is 0.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>leaf</kbd></td>
       *       <td>Whether the option is a leaf or a group.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>data</kbd></td>
       *       <td>The data object for the option.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>parentElement</kbd></td>
       *       <td>The option label element.  The renderer can use this to directly append content.</td>
       *     </tr>
       *   </tbody>
       * </table>
       * 
       * @expose
       * @memberof! oj.ojInputSearch
       * @instance
       * @type {function(Object)|null}
       * @default <code class="prettyprint">null</code>
       * 
       * @example <caption>Initialize the InputSearch with a renderer:</caption>
       * $( ".selector" ).ojInputSearch({ "optionRenderer": function(optionContext) {
       *                                                      return optionContext['data']['FIRST_NAME'];}});
       *
       * @example <caption>Get or set the <code class="prettyprint">renderer</code> option, after initialization:</caption>
       * // set the renderer function
       * $( ".selector" ).ojInputSearch( "option", "optionRenderer", myFunction});
       */
      optionRenderer: null,

      /**
       * The knockout template used to render the content of the option in drop down.
       *
       * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a 
       * component option.
       *
       * @ojbindingonly
       * @name optionTemplate
       * @memberof! oj.ojInputSearch
       * @instance
       * @type {string|null}
       * @default <code class="prettyprint">null</code>
       * 
       * @example <caption>Specify the <code class="prettyprint">template</code> when initializing InputSearch:</caption>
       * // set the template
       * &lt;ul id="inputsearch" data-bind="ojComponent: {component: 'ojInputSearch', optionTemplate: 'my_template'}"&gt;&lt;/ul&gt;
       */

      /**
       * The minimum number of characters a user must type before a options 
       * filtering is performed. Zero is useful for local data with just a few items, 
       * but a higher value should be used when a single character search could match a few thousand items.
       *
       * @expose
       * @memberof! oj.ojInputSearch
       * @instance
       * @type {?number}
       * @default <code class="prettyprint">0</code>
       *
       * @example <caption>Initialize the InputSearch with the <code class="prettyprint">minLength</code> option specified:</caption>
       * $( ".selector" ).ojInputSearch( { "minLength": 2 } );
       */
      minLength : 0,

      /**
       * Triggered immediately before the InputSearch drop down is expanded. 
       *
       * @expose
       * @event
       * @memberof! oj.ojInputSearch
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       *
       * @example <caption>Initialize the InputSearch with the <code class="prettyprint">beforeExpand</code> callback specified:</caption>
       * $( ".selector" ).ojInputSearch({
       *     "beforeExpand": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojbeforeexpand</code> event:</caption>
       * $( ".selector" ).on( "ojbeforeexpand", function( event, ui ) {} );
       */
      beforeExpand : null,
      
      /**
	   * <p>The  <code class="prettyprint">rawValue</code> is the read-only option for retrieving 
       * the currently displayed value from the input field in text form.</p>
       * <p>
       * The <code class="prettyprint">rawValue</code> updates on the 'input' javascript event, 
       * so the <code class="prettyprint">rawValue</code> changes as the value of the input is changed. 
       * If the user types in '1,200' into the field, the rawValue will be '1', then '1,', then '1,2', 
       * ..., and finally '1,200'. Then when the user blurs or presses 
       * Enter the <code class="prettyprint">value</code> option gets updated.
       * </p>
       * <p>This is a read-only option so page authors cannot set or change it directly.</p>
       * @expose 
       * @access public
       * @instance
       * @default n/a
       * @memberof! oj.ojInputSearch
       * @type {string|undefined}
       * @since 2.0.2
       * @readonly
       */
      rawValue: undefined, 

      /**
       * Fired whenever a component option changes, whether due to user interaction or programmatic
       * intervention.  If the new value is the same as the previous value, no event will be fired.  The event 
       * listener will receive two parameters described below:
       *
       * @expose
       * @event
       * @memberof! oj.ojInputSearch
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui event payload
       * @property {string} ui.option the name of the option that is changing
       * @property {Object} ui.previousValue - an Object holding the previous value of the option.
       * When previousValue is not a primitive type, i.e., is an Object, it may hold the same value as
       * the value property.
       * @property {Object} ui.value - an Object holding the current value of the option.
       * @property {?Object} ui.subproperty - an Object holding information about the subproperty that changed.
       * @property {string} ui.subproperty.path - the subproperty path that changed.
       * @property {Object} ui.subproperty.previousValue - an Object holding the previous value of the subproperty.
       * @property {Object} ui.subproperty.value - an Object holding the current value of the subproperty.
       * @property {Object} ui.optionMetadata information about the option that is changing
       * @property {string} ui.optionMetadata.writeback <code class="prettyprint">"shouldWrite"</code> or
       *           <code class="prettyprint">"shouldNotWrite"</code>.  For use by the JET writeback mechanism.
       * @property {string} ui.optionMetadata.trigger This property indicates the what triggered the
       *           <code class="prettyprint">value</code> option change. Possible trigger types are: 
       *             <code class="prettyprint">enter_pressed</code>, 
       *             <code class="prettyprint">option_selected</code>, 
       *             <code class="prettyprint">blur</code> and 
       *             <code class="prettyprint">search_icon_clicked</code>
       *
       * @example <caption>Initialize component with the <code class="prettyprint">optionChange</code> callback</caption>
       * $(".selector").ojInputSearch({
       *   'optionChange': function (event, ui) {
       *        if (ui['option'] === 'value') { // handle value change }
       *    }
       * });
       * @example <caption>Bind an event listener to the ojoptionchange event</caption>
       * $(".selector").on({
       *   'ojoptionchange': function (event, ui) {
       *       window.console.log("option that changed is: " + ui['option']);
       *   };
       * });
       */
      optionChange: null,
      
      /**
       * <p>Fired whenever the value is submitted by the user.</p>
       * 
       * <p>This event is similar to the <code class="prettyprint">value</code> 
       * <a href="#event:optionChange">optionChange</a> event. The optionChange 
       * event will be fired only if there is a change in the value,
       * but the <code class="prettyprint">update</code> event will be fired
       * even if there is no change in the value. This will help the
       * application to re-submit the search query for the same value.</p>
       * 
       * <p><code class="prettyprint">update</code> event will be fired after the 
       * 'value' <code class="prettyprint">optionChange</code> event.</p>
       * 
       * @expose
       * @event
       * @memberof! oj.ojInputSearch
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {Object} ui.value - an Object holding the current value.
       * @property {Object} ui.optionMetadata information about the event.
       * @property {string} ui.optionMetadata.trigger This property indicates the what triggered the
       *             <code class="prettyprint">update</code> event. Possible trigger types are: 
       *             <code class="prettyprint">enter_pressed</code>, 
       *             <code class="prettyprint">option_selected</code> and
       *             <code class="prettyprint">search_icon_clicked</code>
       *
       * @example <caption>Initialize component with the <code class="prettyprint">update</code> callback</caption>
       * $(".selector").ojInputSearch({
       *   'update': function (event, data) {
       *        // handle update event
       *    }
       * });
       * @example <caption>Bind an event listener to the ojupdate event</caption>
       * $(".selector").on({
       *   'ojupdate': function (event, data) {
       *       window.console.log("Update event fired");
       *   };
       * });
       */
      update: null,
    /** 
     * List of validators used by component when performing validation. Each item is either an 
     * instance that duck types {@link oj.Validator}, or is an Object literal containing the 
     * properties listed below. Implicit validators created by a component when certain options 
     * are present (e.g. <code class="prettyprint">required</code> option), are separate from 
     * validators specified through this option. At runtime when the component runs validation, it 
     * combines the implicit validators with the list specified through this option. 
     * <p>
     * Hints exposed by validators are shown in the notewindow by default, or as determined by the 
     * 'validatorHint' property set on the <code class="prettyprint">displayOptions</code> 
     * option. 
     * </p>
     * 
     * <p>
     * When <code class="prettyprint">validators</code> option changes due to programmatic 
     * intervention, the component may decide to clear messages and run validation, based on the 
     * current state it is in. </br>
     * 
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>The cached list of validator instances are cleared and new validator hints is pushed to 
     * messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when validators changes, component does nothing other than the 
     * steps it always performs.</li>
     * <li>if component is invalid and is showing messages when 
     * <code class="prettyprint">validators</code> changes then all component messages are cleared 
     * and full validation run using the display value on the component. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   option is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   option is updated; page author can listen to the <code class="prettyprint">optionChange</code> 
     *   event on the <code class="prettyprint">value</code> option to clear custom errors.</li>
     * </ul>
     * </li>
     * <li>if component is invalid and has deferred messages when validators changes, it does 
     * nothing other than the steps it performs always.</li>
     * </ul>
     * </p>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the component are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> option is not cleared.</li>
     * </ul>
     * </p>
     * 
     * @property {string} type - the validator type that has a {@link oj.ValidatorFactory} that can 
     * be retrieved using the {@link oj.Validation} module. For a list of supported validators refer 
     * to {@link oj.ValidatorFactory}. <br/>
     * E.g., <code class="prettyprint">{validators: [{type: 'regExp'}]}</code>
     * @property {Object=} options - optional Object literal of options that the validator expects. 
     * <br/>
     * E.g., <code class="prettyprint">{validators: [{type: 'regExp', options: {pattern: '[a-zA-Z0-9]{3,}'}}]}</code>

     * 
     * @example <caption>Initialize the component with validator object literal:</caption>
     * $(".selector").ojInputSearch({
     *   validators: [{
     *     type: 'regExp', 
     *     options : {
     *       pattern: '[a-zA-Z0-9]{3,}'
     *     }
     *   }],
     * });
     * 
     * NOTE: oj.Validation.validatorFactory('numberRange') returns the validator factory that is used 
     * to instantiate a range validator for numbers.
     * 
     * @example <caption>Initialize the component with multiple validator instances:</caption>
     * var validator1 = new MyCustomValidator({'foo': 'A'}); 
     * var validator2 = new MyCustomValidator({'foo': 'B'});
     * $(".selector").ojInputSearch({
     *   value: 'option', 
     *   validators: [validator1, validator2]
     * });
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof! oj.ojInputSearch
     * @type {Array|undefined}
     */    
      validators: undefined       

      /** 
       * The type of value is an array, and an array will always be returned from the component.
       * The first element of the array will be used as the value.
       * As a convenience we allow a string to be passed into the setter, 
       * but note that the value option can only be bound to a knockout observableArray.
       * 
       * @example <caption>Initialize the InputSearch with the <code class="prettyprint">value</code> option specified:</caption>
       * $(".selector").ojInputSearch({'value': "option"});<br/>
       * 
       * @example <caption>Get or set the <code class="prettyprint">value</code> option, after initialization:</caption>
       * // Getter: returns value
       * $(".selector").ojInputSearch("option", "value");
       * // Setter: sets value with array containing "option1"
       * $(".selector").ojInputSearch("option", "value", ["option1"]);
       * // Setter: sets value with string "option1"
       * $(".selector").ojInputSearch("option", "value", "option1"); 
       * 
       * @member 
       * @name  value
       * @access public
       * @instance
       * @default When the option is not set, the element's value property is used as its initial value if it exists. 
       * @memberof! oj.ojInputSearch
       * @type {string|Array}
       */
    },

    /**
     * Returns a jQuery object containing the element visually representing the InputSearch.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof! oj.ojInputSearch
     * @instance
     * @return {jQuery} the ojInputSearch
     */
    widget : function ()
    {
      return this.inputSearch.container;
    },
    /**
     * @override
     * @private
     */
    _ComponentCreate : function ()
    {
      this._super();
      this._setup();
    },

    _InitOptions : function (originalDefaults, constructorOptions)
    {
      var props = [{attribute: "disabled", validateOption: true},
                   {attribute: "placeholder"},
                   {attribute: "required", coerceDomValue: true, validateOption: true},
                   {attribute: "title"}
                 ]; 
    
      this._super(originalDefaults, constructorOptions);
      oj.EditableValueUtils.initializeOptionsFromDom(props, constructorOptions, this);  

      if (this.options['value'] === undefined) {
          this.options['value'] = (this.element.attr('value') !== undefined)
            ? _ComboUtils.splitVal(this.element.val(), ",") : null;
      } else {
          var value = this.options['value'];
          if (Array.isArray(value)) {
            value = value.slice(0);
          } else if(typeof value === "string") {
            value = [value];
          }
          this.options['value'] = value;
      }
    },
        /**
      * Performs post processing after _SetOption() is called. Different options when changed perform
      * different tasks. See _AfterSetOption[OptionName] method for details.
      *
      * @param {string} option
      * @param {Object|string=} previous
      * @param {Object=} flags
      * @protected
      * @memberof! oj.ojInputSearch
      * @instance
      * @override
      */
    _AfterSetOption : function (option, previous, flags)
    {
      this._superApply(arguments);
      switch (option)
      {
        case "required":
          this._AfterSetOptionRequired(option);
          break;
        case "validators":
          this._AfterSetOptionValidators(option);
          break;
        default:
          break;
      }
    },
    /**
     * Whether the component is required.
     * 
     * @return {boolean} true if required; false
     * 
     * @memberof! oj.ojInputSearch
     * @instance
     * @protected
     * @override
     */
    _IsRequired : function () 
    {
      return this.options['required'];
    },
    /**
     * Performs post processing after required option is set by taking the following steps.
     * 
     * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
     * run full validation with UI value (we don't know if the UI error is from a required validator 
     * or something else);<br/>
     * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
     * updated<br/>
     * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
     * listen to optionChange(value) to clear custom errors.<br/>
     * 
     * - if component is invalid and has messagesHidden -> required: false -> clear component 
     * errors; no deferred validation is run.<br/>
     * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
     * errors unnecessarily)<br/>
     * - messagesCustom is never cleared<br/>
     * 
     * @param {string} option
     * 
     * @memberof! oj.ojInputSearch
     * @instance
     * @protected
     */
    _AfterSetOptionRequired : oj.EditableValueUtils._AfterSetOptionRequired,
    /**
     * When validators option changes, take the following steps.
     * 
     * - Clear the cached normalized list of all validator instances. push new hints to messaging.<br/>
     * - if component is valid -> validators changes -> no change<br/>
     * - if component is invalid has messagesShown -> validators changes -> clear all component 
     * messages and re-run full validation on displayValue. if there are no errors push value to 
     * model;<br/>
     * - if component is invalid has messagesHidden -> validators changes -> do nothing; doesn't change 
     * the required-ness of component <br/>
     * - messagesCustom is not cleared.<br/>
     * 
     * NOTE: The behavior applies to any option that creates implicit validators - min, max, pattern, 
     * etc. Components can call this method when these options change.
     * 
     * @memberof! oj.ojInputSearch
     * @instance
     * @protected
     */
    _AfterSetOptionValidators : oj.EditableValueUtils._AfterSetOptionValidators,    
    /**
     * @memberof! oj.ojInputSearch
     * @instance
     * @private 
     */
    _refreshRequired : oj.EditableValueUtils._refreshRequired,  
    /**
     * This returns an array of all validators 
     * normalized from the validators option set on the component. <br/>
     * @return {Array} of validators. 
     * @memberof! oj.ojInputSearch
     * @instance
     * @protected
     */
    _GetNormalizedValidatorsFromOption : oj.EditableValueUtils._GetNormalizedValidatorsFromOption,

    /**
     * Called to find out if aria-required is unsupported.
     * @memberof! oj.ojInputSearch
     * @instance
     * @protected
     */
    _AriaRequiredUnsupported : function()
    {
      return false;
    },   
    /**
     * @memberof! oj.ojInputSearch
     * @instance
     * @private 
     */
    _setup : function ()
    {
      var opts = {};
      opts.element = this.element;
      opts.ojContext = this;
      opts.inputSearch = true;
      opts = $.extend(this.options, opts);

      this.inputSearch = new _OjInputSeachContainer();
      this.inputSearch._init(opts);
      this._refreshRequired(this.options['required']);
    },

    /**
     * @override
     * @private
     */
    _destroy : function ()
    {
      this.inputSearch._destroy();
      this._super();
    },
    
    /**
     * Refreshes the InputSearch.
     *
     * <p>This method does not accept any arguments.
     * 
     * @expose 
     * @memberof! oj.ojInputSearch
     * @instance
     */
    refresh : function ()
    {
      this._super();

      this.inputSearch._destroy();
      this._setup();
      this._SetRootAttributes();
    },

    /**
     * Handles options specific to InputSearch.
     * @override
     * @protected
     * @memberof! oj.ojInputSearch
     */
    _setOption : function (key, value, flags)
    {
      if (key === "value") {
        //clone the value, otherwise _setDisplayValue will not be invoked on binding value to ko observableArray.
        //TODO: Need to revisit this once 18724975 is fixed.
        
        if (value != null) {
          if (Array.isArray(value)) {
            value = value.slice(0);
          }
          else if (typeof value === "string") {
            value = [value];
          } else {
            oj.Logger.error("ojInputSearch value has to be an array of string or a string.");
          }
        }

        // valueChangeTrigger will be used while setting the display value.
        if (flags && flags["_context"] && flags["_context"].optionMetadata)
        {
          this.inputSearch.valueChangeTrigger = flags["_context"].optionMetadata["trigger"];
        }
        else
          this.inputSearch.valueChangeTrigger = null;
      }

      this._super(key, value, flags);

      if (key === "options") 
      {
        this.inputSearch.opts.options = value;
        this.inputSearch.opts = this.inputSearch._prepareOpts(this.inputSearch.opts);
      }

      if (key === "disabled")
      {
        if (value)
          this.inputSearch._disable();
        else
          this.inputSearch._enable();
      }
    },

    _NotifyDetached : function() {
      this._superApply(arguments);
      this.inputSearch.close();
    },
    
    //19670748, dropdown popup should be closed on subtreeHidden notification.
    _NotifyHidden : function() {
      this._superApply(arguments);
      this.inputSearch.close(); 
    },

    /**
     * Updates display value of InputSearch.
     * @override
     * @protected
     * @memberof! oj.ojInputSearch
     */
    _SetDisplayValue: function(displayValue)
    {
      this.inputSearch._initSelection();
    },
    
    /**
     * Set the placeholder.
     * @override
     * @protected
     * @memberof! oj.ojInputSearch
     */
    _SetPlaceholder : function(value)
    {
      if (this.inputSearch) 
      {
        this.inputSearch.opts.placeholder = value;
        if (this.inputSearch._setPlaceholder) 
        {
          this.inputSearch._setPlaceholder();
        }
      }
    },
    
    /**
     * Validates the component's value using the converter and all validators registered on 
     * the component. 
     * @return {boolean} true if component passed validation, 
     * false if there were validation errors.
     * @example <caption>Validate component using its current value.</caption>
     * // validate display value. 
     * $(.selector).ojInputSearch('validate');
     *
     * @expose 
     * @override
     * @memberof! oj.ojInputSearch
     * @instance
     */
    validate : function ()
    {
      var displayValue = this.inputSearch.search.val();
      var newValue = null;

      var existingValue = [];
      if (this.isValid())
        existingValue = this.inputSearch.getVal();

      if (displayValue === undefined || displayValue === null || displayValue === "")
        newValue = existingValue;
      else
        newValue = [displayValue];

      return this._SetValue(newValue, null, this._VALIDATE_METHOD_OPTIONS);
    },
    
    /**
     * Parses the value using the converter set and returns the parsed value. If parsing fails the 
     * error is written into the element 
     * 
     * @override
     * @protected
     * @memberof! oj.ojInputSearch
     * @instance
     */
    _parseValue: function(submittedValue) 
    {
      var parsedVal = [];

      if (submittedValue == null)
        return parsedVal;

      if (Array.isArray(submittedValue)) {
        for (var i = 0; i<submittedValue.length; i++) {
          var parsed = this._super(submittedValue[i]);
          parsedVal.push(parsed.toString());
        }
      } else if (typeof submittedValue === "string") {
        var parsed = this._super(submittedValue);
        parsedVal.push(parsed.toString());
      } else {
        oj.Logger.error("ojInputSearch value has to be an array of string or a string.");
      }

      return parsedVal;
    },
    
    /**
     * Returns the messaging launcher element  i.e., where user sets focus that triggers the popup. 
     * Usually this is the element input or select that will receive focus and on which the popup 
     * for messaging is initialized. 
     *
     * @override
     * @protected
     * @memberof! oj.ojInputSearch
     * @return {Object} jquery element which represents the messaging launcher component
     */
    _GetMessagingLauncherElement : function ()
    {
      return this.inputSearch.search;
    }, 
    
    /**
     * Returns the jquery element that represents the content part of the component.
     * This is usually the component that user sets focus on (tabindex is set 0) and 
     * where aria attributes like aria-required, aria-labeledby etc. are set. This is
     * also the element where the new value is updated. Usually this is the same as
     * the _GetMessagingLauncherElement.
     *
     * @override
     * @protected
     * @memberof! oj.ojInputSearch
     * @return {jQuery} jquery element which represents the content.
     */
    _GetContentElement : function ()
    {
      return this.inputSearch.search;
    },      
    
    /**
     * Returns the default styleclass for the component.
     * 
     * @return {string}
     * @expose
     * @memberof! oj.ojInputSearch
     * @override
     * @protected
     */
    _GetDefaultStyleClass : function ()
    {
    return "oj-inputsearch";
    },
    
    _getDropdown : function ()
    {
      if (this.inputSearch && this.inputSearch._opened())
      {
        var dropdown = $(".oj-listbox-drop");
        for (var i=0; i<dropdown.length; i++)
        {
          if ($(dropdown[i]).attr("id") == "oj-listbox-drop" &&
            $(dropdown[i]).attr("data-oj-containerid") == this.inputSearch.containerId)
          return $(dropdown[i]);
        }
      }
      return null;
    },

    /**
     * Opens the InputSearch drop-down.
     * 
     * @expose 
     * @memberof oj.ojInputSearch
     * @instance
     */
    expand : function()
    {
      this.inputSearch.open();
    },

    /**
     * Closes the InputSearch drop-down.
     * 
     * @expose 
     * @memberof oj.ojInputSearch
     * @instance
     */
    collapse : function()
    {
      this.inputSearch.close();
    },

    //////////////////     SUB-IDS     //////////////////

    /**
     * <p>Sub-ID for the input field
     * 
     * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
     * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.
     * 
     * @ojsubid
     * @member
     * @name oj-inputsearch-input
     * @memberof oj.ojInputSearch
     * @instance
     * 
     * @example <caption>Get the input field element</caption>
     * var node = $( ".selector" ).ojInputSearch( "getNodeBySubId", {'subId': 'oj-inputsearch-input'} );
     */
 
    /**
     * <p>Sub-ID for the search icon of InputSearch. 
     * 
     * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
     * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.
     * 
     * @ojsubid
     * @member
     * @name oj-inputsearch-search
     * @memberof oj.ojInputSearch
     * @instance
     * 
     * @example <caption>Get the search icon of the InputSearch</caption>
     * var node = $( ".selector" ).ojInputSearch( "getNodeBySubId", {'subId': 'oj-inputsearch-search'} );
     */

    /**
     * <p>Sub-ID for the list item.</p>
     *
     * @ojsubid oj-listitem
     * @memberof oj.ojInputSearch
     *
     * @example <caption>Get the listitem corresponding to value "myVal"</caption>
     * var node = $( ".selector" ).ojInputSearch( "getNodeBySubId", {'subId': 'oj-listitem', 'value': 'myVal'} );
     */

    getNodeBySubId: function(locator)
    {
      var node = null, subId;
      if (locator === null)
      {
        return this.inputSearch.container ? this.inputSearch.container[0] : null;
      }
      else
      {
        node = this._super(locator);
      }
      
      if (!node)
      {
        subId = locator['subId'];
        if (subId === "oj-inputsearch-search")
          subId = "oj-inputsearch-search-button";

        switch (subId)
        {
          case "oj-inputsearch-input":
          case "oj-inputsearch-search-button":
            node = this.widget().find("." + subId)[0];
            break;
          case "oj-listitem":
            var dropdown = this._getDropdown();
            if (dropdown) 
            {
              var list = dropdown.find(".oj-listbox-result");
              node = this.inputSearch._findItem(list, locator['value']);
            }
            break;
        }
      }

      // Non-null locators have to be handled by the component subclasses
      return node || null;
    },

    /**
     * Returns the subId object for the given child DOM node.  For more details, see
     * <a href="#getNodeBySubId">getNodeBySubId</a>.
     *
     * @expose
     * @override
     * @ignore
     * @memberof oj.ojInputSearch
     * @instance
     *
     * @param {!Element} node - child DOM node
     * @return {Object|null} The subId for the DOM node, or <code class="prettyprint">null</code> when none is found.
     *
     * @example <caption>Get the subId for a certain DOM node:</caption>
     * var subId = $( ".selector" ).ojInputSearch( "getSubIdByNode", node );
     */
    getSubIdByNode: function(node)
    {
      var subId = null;
      
      if (node != null)
      {
        var nodeCached = $(node);
                
        if (nodeCached.hasClass('oj-inputsearch-input'))
          subId = {'subId': 'oj-inputsearch-input'};
        else if (nodeCached.hasClass('oj-inputsearch-search-button'))
          subId = {'subId': 'oj-inputsearch-search'};
        else if (nodeCached.hasClass('oj-listbox-result'))
          subId = {'subId': 'oj-listitem', 'value': nodeCached.data('ojinputsearch')['value']};
        else
          subId = this._super(node);
      }
      return subId;      
    }

    // Fragments:

  /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Input Field</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td> If the drop down is not open, expand the drop down list. Otherwise, close the drop down list
     *       If hints, title or messages exist in a notewindow, 
     *        pop up the notewindow.</td>
     *     </tr>
     *     <tr>
     *       <td>Search Button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td> Updates the value with the text entered in the input field and fires the <code class="prettyprint">optionChange</code> event which can be used to perform search.</td>
     *     </tr>
     *     <tr>
     *       <td>Option Item</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Tap on a option item in the drop down list to select item.</td>
     *     </tr>
     *   </tbody>
     *  </table>
     *
     * <p>Disabled option items receive no highlight and are not selectable.</p>
   *
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojInputSearch
   */

  /**
     * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
     *   <tbody>
     *     <tr>
     *      <td>Input</td>
     *       <td><kbd>Enter</kbd></td>
     *       <td> Select the highlighted choice from the drop down.</td>
     *     </tr>
     *     <tr>
     *      <td>Input</td>
     *       <td><kbd>UpArrow or DownArrow</kbd></td>
     *       <td> Highlight the option item on the drop down list in the direction of the arrow.
     *         If the drop down is not open, expand the drop down list.</td>
     *     </tr>
     *     <tr>
     *      <td>Input</td>
     *       <td><kbd>Esc</kbd></td>
     *       <td> Collapse the drop down list. If the drop down is already closed, do nothing.</td>
     *     </tr>
     *     <tr>
     *        <td>Input</td>
     *       <td><kbd>Tab In</kbd></td>
     *       <td>Set focus to the input. If hints, title or messages exist in a notewindow, 
     *        pop up the notewindow.</td>
     *     </tr> 
     *   </tbody>
     *  </table>
     *
     * <p>Disabled option items receive no highlight and are not selectable.</p>
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojInputSearch
   */

  /**
   * {@ojinclude "name":"ojStylingDocIntro"}
   * 
   * <table class="generic-table styling-table">
   *   <thead>
   *     <tr>
   *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
   *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>oj-listbox-header</td>
   *       <td>Optional. Custom header options can be added to the drop down through this styling.</td>
   *     </tr>
   *     <tr>
   *       <td>oj-listbox-highlighter-section</td>
   *       <td>Optional. Styling to control the which part of the option label has to be considered for highlighting.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
   * @memberof oj.ojInputSearch
   */

  });
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

  /**
   * @ojcomponent oj.ojSelectOne
   * @augments oj.ojSelect
   * @since 0.6
   * @ojdisplayname Single Select
   * @ojshortdesc A dropdown list that supports single selection and search filtering.
   * @ojrole combobox
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojSelectOne extends ojSelect<any, ojSelectOneSettableProperties>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojSelectOneSettableProperties extends ojSelectSettableProperties<any>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojstatus preview
   *
   * @classdesc
   * <h3 id="selectOverview-section">
   *   JET Select One
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#selectOverview-section"></a>
   * </h3>
   * <p>Description: JET Select One provides support for single-select and search filtering.</p>
   *
   * <p>A JET Select One can be created with the following markup.</p>
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-select-one>
   *   &lt;oj-option value="option 1">option 1&lt;/oj-option>
   *   &lt;oj-option value="option 2">option 2&lt;/oj-option>
   *   &lt;oj-option value="option 3">option 3&lt;/oj-option>
   *   &lt;oj-option value="option 4">option 4&lt;/oj-option>
   * &lt;/oj-select-one> 
   * </code></pre>
   *
   * {@ojinclude "name":"validationAndMessagingDoc"}
   * 
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDocOne"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDocOne"}
   *
   *
   * {@ojinclude "name":"selectCommon"}
   */
   
  /**
   * @ojcomponent oj.ojSelectMany
   * @augments oj.ojSelect
   * @since 0.6
   * @ojdisplayname Multi Select
   * @ojshortdesc A dropdown list that supports multiple selections and search filtering.
   * @ojrole combobox
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojSelectMany extends ojSelect<Array<any>, ojSelectManySettableProperties>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojSelectManySettableProperties extends ojSelectSettableProperties<Array<any>>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojstatus preview
   *
   * @classdesc
   * <h3 id="selectOverview-section">
   *   JET Select Many
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#selectOverview-section"></a>
   * </h3>
   * <p>Description: JET Select Many provides support for multi-select and search filtering.</p>
   *
   * <p>A JET Select Many can be created with the following markup.</p>
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-select-many>
   *   &lt;oj-option value="option 1">option 1&lt;/oj-option>
   *   &lt;oj-option value="option 2">option 2&lt;/oj-option>
   *   &lt;oj-option value="option 3">option 3&lt;/oj-option>
   *   &lt;oj-option value="option 4">option 4&lt;/oj-option>
   * &lt;/oj-select-many> 
   * </code></pre>
   *
   * {@ojinclude "name":"validationAndMessagingDoc"}
   * 
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDocMany"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDocMany"}
   *
   *
   * {@ojinclude "name":"selectCommon"}
   */
   
  /**
   * @ojcomponent oj.ojSelect
   * @augments oj.editableValue
   * @since 0.6
   * @abstract
   * @ojsignature [{
   *                target: "Type",
   *                value: "abstract class ojSelect<V, SP extends ojSelectSettableProperties<V>, SV=V> extends editableValue<V, SP, SV>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojSelectSettableProperties<V, SV=V> extends editableValueSettableProperties<V, SV>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojstatus preview  
   *
   * @classdesc
   */
  /**
   * <h3 id="rtl-section">
   *   Reading direction
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
   * </h3>
   *
   * <p>As with any JET element, in the unusual case that the directionality (LTR or RTL) changes post-init, the Select must be <code class="prettyprint">refresh()</code>ed.
   *
   * <h3 id="a11y-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
   * </h3>
   * <p>
   * It is up to the application developer to associate an oj-label to the select element.
   * You should put an <code>id</code> on the select element, and then set
   * the <code>for</code> attribute on the oj-label to be the select's id.
   * </p>
   * <p>
   * The element will decorate its associated label with required and help
   * information, if the <code>required</code> and <code>help</code> attributes are set.
   * </p>
   *
   * @ojfragment selectCommon
   * @memberof oj.ojSelect
   */
  oj.__registerWidget("oj.ojSelect", $['oj']['editableValue'],
  {
    defaultElement : "<select>",
    widgetEventPrefix : "oj",
    options :
    {
      /**
       * The threshold for showing the search box in the dropdown when it's expanded.
       * The search box is always displayed when the results size is greater than
       * the threshold, otherwise the search box is initially turned off.
       * However, the search box is displayed as soon as the user starts typing.
       *
       * @example <caption>Get or set the <code class="prettyprint">minimumResultsForSearch</code> property after initialization:</caption>
       * // getter
       * var minimumResultsForSearch = mySelect.minimumResultsForSearch;
       *
       * // setter
       * mySelect.minimumResultsForSearch = 10;
       *
       * @ojshortdesc The threshold for showing the search box in the dropdown.
       * @expose
       * @memberof! oj.ojSelect
       * @instance
       * @type {number}
       * @default 15
       * @ojmin 0
       */
      minimumResultsForSearch : 15,


      /**
       * The placeholder text to set on the element.<p>
       * If the <code class="prettyprint">placeholder</code> attribute is specified to a string, ojselect will adds a placeholder item at the beginning of the dropdown list with
       *  <ul>
       *  <li>displayValue: placeholder text</li>
       *  <li>value: an empty string</li>
       *  </ul>
       * The placeholder item in the dropdown is selectable. However, it's not a valid choice, i.e. validation will fail if the select element is a required field.<p>
       * The placeholder item doesn't participate in the filtering, so it will not appear in the result list with a filter specified.<p>
       * Placeholder text can be an empty string, please see the select placeholder cookbook demo.
       *
       * @example <caption>Initialize the select with the <code class="prettyprint">placeholder</code> attribute specified:</caption>
       * &lt;oj-select-one placeholder="Please select ...">&lt;/oj-select-one>
       *
       * @example <caption>Get or set the <code class="prettyprint">placeholder</code> property after initialization:</caption>
       * // getter
       * var placeholderValue = mySelect.placeholder;
       *
       * // setter
       * mySelect.placeholder = "Select a value";
       *
       * @name placeholder
       * @ojshortdesc A short hint that can be displayed before user selects a value.
       * @expose
       * @access public
       * @instance
       * @memberof oj.ojSelectOne
       * @type {string|null|undefined}
       * @default null
       */
      /**
       * The placeholder text to set on the element. The placeholder specifies a short hint that can be displayed before user
       * selects a value.
       *
       * @example <caption>Initialize the select with the <code class="prettyprint">placeholder</code> attribute specified:</caption>
       * &lt;oj-select-many placeholder="Please select ...">&lt;/oj-select-many>
       *
       * @example <caption>Get or set the <code class="prettyprint">placeholder</code> property after initialization:</caption>
       * // getter
       * var placeholderValue = mySelect.placeholder;
       *
       * // setter
       * mySelect.placeholder = "Select values";
       *
       * @name placeholder
       * @ojshortdesc A short hint that can be displayed before user selects a value.
       * @expose
       * @access public
       * @instance
       * @memberof oj.ojSelectMany
       * @type {string|null|undefined}
       * @default null
       */
      placeholder: null,
      
      /**
       * {@ojinclude "name":"selectCommonOptionRenderer"}
       * @name optionRenderer
       * @ojshortdesc The renderer function that renders the content of each option.
       * @expose
       * @memberof oj.ojSelectOne
       * @instance
       * @type {function(Object)|null}
       * @default null
       * @example <caption>Initialize the select with a renderer:</caption>
       * &lt;oj-select-one option-renderer="[[optionRenderer]]">&lt;/oj-select-one>
       * @example var optionRenderer = function(optionContext) {
       *            return optionContext['data']['FIRST_NAME'];
       *          };
       */
      /**
       * {@ojinclude "name":"selectCommonOptionRenderer"}
       * @name optionRenderer
       * @ojshortdesc The renderer function that renders the content of each option.
       * @expose
       * @memberof oj.ojSelectMany
       * @instance
       * @type {function(Object)|null}
       * @default null
       * @example <caption>Initialize the Select with a renderer:</caption>
       * &lt;oj-select-many option-renderer="[[optionRenderer]]">&lt;/oj-select-many>
       * @example var optionRenderer = function(optionContext) {
       *            return optionContext['data']['FIRST_NAME'];
       *          };
       */
      /**      
       * The renderer function that renders the content of each option.
       * The function should return one of the following: 
       * <ul>
       *   <li>An Object with the following property:
       *     <ul><li>insert: HTMLElement - A DOM element representing the content of the option.</li></ul>
       *   </li>
       *   <li>undefined: If the developer chooses to manipulate the option content directly, the function should return undefined.</li>
       * </ul>
       * 
       * <p>The <code class="prettyprint">option-renderer</code> decides only 
       * how the options' content has to be rendered in the drop down.
       * Once an option is selected from the drop down, 
       * what value has to be displayed in the in input field is decided by the 
       * label field in the data object. See <a href="#options">options</a> 
       * and <a href="#optionsKeys">options-keys</a> for configuring option label and value.
       * </p>
       *
       * <p>The context parameter passed to the renderer contains the following keys:</p>
       * <table class="keyboard-table">
       *   <thead>
       *     <tr>
       *       <th>Key</th>
       *       <th>Description</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td><kbd>component</kbd></td>
       *       <td>A reference to the Select element.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>parent</kbd></td>
       *       <td>The parent of the data item. The parent is null for root node.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>index</kbd></td>
       *       <td>The index of the option, where 0 is the index of the first option. In the hierarchical case the index is relative to its parent.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>depth</kbd></td>
       *       <td>The depth of the option. The depth of the first level children under the invisible root is 0.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>leaf</kbd></td>
       *       <td>Whether the option is a leaf or a group.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>data</kbd></td>
       *       <td>The data object for the option.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>parentElement</kbd></td>
       *       <td>The option label element.  The renderer can use this to directly append content.</td>
       *     </tr>
       *   </tbody>
       * </table>
       * 
       * @expose
       * @memberof oj.ojSelect
       * @instance
       * @ojfragment selectCommonOptionRenderer
       */
      optionRenderer: null,

      /**
       * {@ojinclude "name":"selectCommonOptions"}
       *
       * @name options
       * @ojshortdesc The option items for the Select.
       * @expose
       * @access public
       * @instance
       * @type {null | Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>}
       * @default null
       * @ojsignature { target: "Type",
       *                value: "Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>",
       *                jsdocOverride: true}
       * @memberof oj.ojSelectOne
       *
       * @example <caption>Initialize the Select with the <code class="prettyprint">options</code> attribute specified:</caption>
       * &lt;oj-select-one options="[[dataArray]]">&lt;/oj-select-one>
       *
       * @example <caption>The options array should contain objects with value and label properties:</caption>
       * var dataArray = [{value: 'option1', label: 'Option 1'}, 
       *                  {value: 'option2', label: 'Option 2', disabled: true}, 
       *                  {value: 'option3', label: 'Option 3'}];
       *
       * @example <caption>Initialize the Select with a data provider and data mapping:</caption>
       * &lt;oj-select-one options="[[dataProvider]]">&lt;/oj-select-one>
       *
       * @example <caption>Data mapping can be used if data doesn't have value and label properties.</caption>
       * // actual field names are "id" and "name"
       * var dataArray = [
       *            {id: 'Id 1', name: 'Name 1'},
       *            {id: 'Id 2', name: 'Name 2'},
       *            {id: 'Id 3', name: 'Name 3'}];
       *
       * // In mapfields, map "name" to "label" and "id" to "value"
       * var mapFields = function(item) {
       *   var data = item['data'];
       *   var mappedItem = {};
       *   mappedItem['data'] = {};
       *   mappedItem['data']['label'] = data['name'];
       *   mappedItem['data']['value'] = data['id'];
       *   mappedItem['metadata'] = {'key': data['id']};
       *   return mappedItem;
       * }; 
       * var dataMapping = {'mapFields': mapFields};
       *
       * var arrayDataProvider = new oj.ArrayDataProvider(dataArray, {idAttribute: 'id'});
       * var dataProvider = new oj.ListDataProviderView(arrayDataProvider, {'dataMapping': dataMapping});
       */
      /**
       * {@ojinclude "name":"selectCommonOptions"}
       *
       * @name options
       * @ojshortdesc The option items for the Select.
       * @expose
       * @access public
       * @instance
       * @type {null | Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>}
       * @default null
       * @ojsignature { target: "Type",
       *                value: "Array.<oj.Option|oj.Optgroup> | oj.DataProvider.<oj.Option>",
       *                jsdocOverride: true}
       * @memberof oj.ojSelectMany
       *
       * @example <caption>Initialize the Select with the <code class="prettyprint">options</code> attribute specified:</caption>
       * &lt;oj-select-many options="[[dataArray]]">&lt;/oj-select-many>
       *
       * @example <caption>The options array should contain objects with value and label properties:</caption>
       * var dataArray = [{value: 'option1', label: 'Option 1'}, 
       *                  {value: 'option2', label: 'Option 2', disabled: true}, 
       *                  {value: 'option3', label: 'Option 3'}];
       *
       * @example <caption>Initialize the Select with a data provider and data mapping:</caption>
       * &lt;oj-select-many options="[[dataProvider]]">&lt;/oj-select-many>
       *
       * @example <caption>Data mapping can be used if data doesn't have value and label properties.</caption>
       * // actual field names are "id" and "name"
       * var dataArray = [
       *            {id: 'Id 1', name: 'Name 1'},
       *            {id: 'Id 2', name: 'Name 2'},
       *            {id: 'Id 3', name: 'Name 3'}];
       *
       * // In mapfields, map "name" to "label" and "id" to "value"
       * var mapFields = function(item) {
       *   var data = item['data'];
       *   var mappedItem = {};
       *   mappedItem['data'] = {};
       *   mappedItem['data']['label'] = data['name'];
       *   mappedItem['data']['value'] = data['id'];
       *   mappedItem['metadata'] = {'key': data['id']};
       *   return mappedItem;
       * }; 
       * var dataMapping = {'mapFields': mapFields};
       *
       * var arrayDataProvider = new oj.ArrayDataProvider(dataArray, {idAttribute: 'id'});
       * var dataProvider = new oj.ListDataProviderView(arrayDataProvider, {'dataMapping': dataMapping});
       */
      /**
       * The option items for the Select. This attribute can be used instead of providing a list of <code class="prettyprint">oj-option</code> or <code class="prettyprint">oj-optgroup</code> child elements of the Select element.
       * This attribute accepts:
       * <ol>
       * <li>an array of <a href="oj.Option.html">oj.Option</a>s and/or <a href="oj.Optgroup.html">oj.Optgroup</a>s.
       *   <ul>
       *   <li>Use <code class="prettyprint">oj.Option</code> for a leaf option.</li>
       *   <li>Use <code class="prettyprint">oj.Optgroup</code> for a group option.</li>
       *   </ul>
       * </li>
       * <li>a data provider of <a href="oj.Option.html">oj.Option</a>s. This data provider must implement <a href="oj.DataProvider.html">oj.DataProvider</a>.
       *   <ul>
       *   <li><code class="prettyprint">value</code> in <code class="prettyprint">oj.Option</code> must be the row key in the data provider.</li>
       *   <li>A maximum of 15 rows will be displayed in the dropdown. If more than 15 results are available then users need to filter further. Please note that users can't filter further if render-mode is <code class="prettyprint">native</code>.</li>
       *   <li>If the data provider supports the filter criteria capability including the contains (<code class="prettyprint">$co</code>) operator, JET Select will request the data provider to do filtering. Otherwise it will filter internally.</li>
       *   </ul>
       * </li>
       * </ol>
       *
       * @expose
       * @memberof oj.ojSelect
       * @instance
       * @ojfragment selectCommonOptions
       */
      options : null,

      /**
       * Specify the key names to use in the options array. 
       * <p>Depending on options-keys means that the signature of the data does not match what is supported by the options attribute. When using Typescript, this would result in a compilation error.</p>
       * <p>Best practice is to use a <a href="oj.ListDataProviderView.html">oj.ListDataProviderView</a> with data mapping as a replacement.</p>
       * <p>However, for the app that must fetch data from a REST endpoint where the data fields do not match those that are supported by the options attribute, you may use the options-keys with any dataProvider that implements <a href="oj.DataProvider.html">oj.DataProvider</a> interface.</p>
       * 
       * @name optionsKeys
       * @expose
       * @access public
       * @instance
       * @type {?Object}
       * @default null
       * @memberof oj.ojSelect
       *
       * @example <caption>Initialize the Select with <code class="prettyprint">options-keys</code> specified. This allows the key names to be redefined in the options array.</caption>
       * &lt;oj-select-xxx options-keys="[[optionsKeys]]">&lt;/oj-select-xxx>
       * @example var optionsKeys = {value : "state_abbr", label : "state_name"};
       * @example <caption>Redefine keys for data with subgroups.</caption>
       * var optionsKeys = {label : "regions", children : "states", 
       *                    childKeys : {value : "state_abbr", label : "state_name"}};
       */
      optionsKeys : {
        /**
         * The key name for the label.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojSelect
         * @alias optionsKeys.label
         * @type {?string}
         * @default null
         */
        label: null,
        /**
         * The key name for the value.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojSelect
         * @alias optionsKeys.value
         * @type {?string}
         * @default null
         */
        value: null,
        /**
         * The key name for the children.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojSelect
         * @alias optionsKeys.children
         * @type {?string}
         * @default null
         */
        children: null,
        /**
         * The object for the child keys.
         *
         * @expose
         * @public
         * @instance
         * @memberof! oj.ojSelect
         * @alias optionsKeys.childKeys
         * @type {?Object}
         * @default null
         * @property {?string} label The key name for the label.
         * @property {?string} value The key name for the value.
         * @property {?string} children The key name for the children.
         * @property {?Object} childKeys The object for the child keys.
         */
        childKeys: null
      },
      /**
       * <p>Attributes specified here will be set on the picker DOM element when it's launched.
       * <p>The supported attributes are <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended to the picker's class and style, if any.
       * Note: 1) picker-attributes is not applied in the native renderMode.
       * 2) setting this attribute after element creation has no effect.
       *
       * @example <caption>Initialize the select specifying a set of attributes to be set on the picker DOM element:</caption>
       * &lt;oj-select-one picker-attributes="[[pickerAttributes]]">&lt;/oj-select-one>
       * @example var pickerAttributes = {
       *   "style": "color:blue;",
       *   "class": "my-class"
       * };
       *
       * @name pickerAttributes
       * @ojshortdesc The style attributes for the drop down.
       * @expose
       * @memberof oj.ojSelectOne
       * @instance
       * @type {?Object}
       * @default null
       */
      /**
       * <p>Attributes specified here will be set on the picker DOM element when it's launched.
       * <p>The supported attributes are <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended to the picker's class and style, if any.
       * Note: 1) picker-attributes is not applied in the native renderMode.
       * 2) setting this attribute after element creation has no effect.
       *
       * @example <caption>Initialize the select specifying a set of attributes to be set on the picker DOM element:</caption>
       * &lt;oj-select-many picker-attributes="[[pickerAttributes]]">&lt;/oj-select-many>
       * @example var pickerAttributes = {
       *   "style": "color:blue;",
       *   "class": "my-class"
       * };
       *
       * @name pickerAttributes
       * @ojshortdesc The style attributes for the drop down.
       * @expose
       * @memberof oj.ojSelectMany
       * @instance
       * @type {?Object}
       * @default null
       */
      pickerAttributes: null,
      
      /**
      * {@ojinclude "name":"selectCommonRequired"}
      *
      * @example <caption>Initialize the select with the <code class="prettyprint">required</code> attribute:</caption>
      * &lt;oj-select-one required="[[isRequired]]">&lt;/oj-select-one>
      *
      * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
      * // getter
      * var requiredValue = mySelect.required;
      *
      * // setter
      * mySelect.required = true;
      *
      * @name required
      * @expose 
      * @ojshortdesc Specifies whether a value is required.
      * @access public
      * @instance
      * @memberof oj.ojSelectOne
      * @type {boolean}
      * @default false
      * @since 0.7
      * @see #translations
      */
     /**
      * {@ojinclude "name":"selectCommonRequired"}
      *
      * @example <caption>Initialize the select with the <code class="prettyprint">required</code> attribute:</caption>
      * &lt;oj-select-many required="[[isRequired]]">&lt;/oj-select-many>
      *
      * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
      * // getter
      * var requiredValue = mySelect.required;
      *
      * // setter
      * mySelect.required = true;
      *
      * @name required
      * @expose 
      * @ojshortdesc Specifies whether a value is required.
      * @access public
      * @instance
      * @memberof oj.ojSelectMany
      * @type {boolean}
      * @default false
      * @since 0.7
      * @see #translations
      */
      /** 
       * Whether the element is required or optional. When required is set to true, an implicit 
       * required validator is created using the validator factory - 
       * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED).createValidator()</code>.
       * 
       * Translations specified using the <code class="prettyprint">translations.required</code> attribute 
       * and the label associated with the element, are passed through to the options parameter of the 
       * createValidator method. 
       * 
       * <p>
       * When <code class="prettyprint">required</code> property changes due to programmatic intervention, 
       * the element may clear messages and run validation, based on the current state it's in. </br>
       *  
       * <h4>Running Validation</h4>
       * <ul>
       * <li>if element is valid when required is set to true, then it runs deferred validation on 
       * the value. This is to ensure errors are not flagged unnecessarily.
       * </li>
       * <li>if element is invalid and has deferred messages when required is set to false, then 
       * element messages are cleared but no deferred validation is run.
       * </li>
       * <li>if element is invalid and currently showing invalid messages when required is set, then 
       * element messages are cleared and normal validation is run using the current display value. 
       * <ul>
       *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
       *   property is not updated and the error is shown. 
       *   </li>
       *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
       *   property is updated; page author can listen to the <code class="prettyprint">optionChange</code> 
       *   event on the <code class="prettyprint">value</code> property to clear custom errors.</li>
       * </ul>
       * </li>
       * </ul>
       * 
       * <h4>Clearing Messages</h4>
       * <ul>
       * <li>Only messages created by the element are cleared.</li>
       * <li><code class="prettyprint">messages-custom</code> property is not cleared.</li>
       * </ul>
       * 
       * </p>
       * 
       * @ojvalue {boolean} false - implies a value is not required to be provided by the user. 
       * This is the default.
       * @ojvalue {boolean} true - implies a value is required to be provided by user and the 
       * input's label will render a required icon. Additionally a required validator - 
       * {@link oj.RequiredValidator} - is implicitly used if no explicit required validator is set. 
       * An explicit required validator can be set by page authors using the validators attribute. 
       *
       * @expose 
       * @access public
       * @instance
       * @memberof oj.ojSelect
       * @since 0.7
       * @see #translations
       * @ojfragment selectCommonRequired
       */
      required: false,

      /**
       * {@ojinclude "name":"selectCommonRenderMode"}
       * @example <caption>Set the <code class="prettyprint">render-mode</code> attribute:</caption>
       * &lt;oj-select-one render-mode="native">&lt;/oj-select-one>
       *
       * @example <caption>Get or set the <code class="prettyprint">renderMode</code> property after initialization:</caption>
       * // getter
       * var renderMode = mySelect.renderMode;
       *
       * // setter
       * mySelect.renderMode = "native";
       * 
       * @name renderMode
       * @ojshortdesc Specifies whether to render select in JET or as a HTML Select tag.
       * @expose 
       * @memberof oj.ojSelectOne
       * @instance
       * @type {string}
       */
      /**
       * {@ojinclude "name":"selectCommonRenderMode"}
       * @example <caption>Set the <code class="prettyprint">render-mode</code> attribute:</caption>
       * &lt;oj-select-many render-mode="native">&lt;/oj-select-many>
       *
       * @example <caption>Get or set the <code class="prettyprint">renderMode</code> property after initialization:</caption>
       * // getter
       * var renderMode = mySelect.renderMode;
       *
       * // setter
       * mySelect.renderMode = "native";
       *
       * @name renderMode
       * @ojshortdesc Specifies whether to render select in JET or as a HTML Select tag.
       * @expose 
       * @memberof oj.ojSelectMany
       * @instance
       * @type {string}
       */
      /** 
       * The render-mode attribute allows applications to specify whether to render select in JET or as a HTML Select tag.
       * Valid Values: jet, native
       *
       * <ul>
       *  <li> jet - Applications get full JET functionality.</li>
       *  <li> native - Applications get the HTML Select tag functionality and additional JET features below:
       *    <ul>
       *      <li>validation</li>
       *      <li>placeholder</li>
       *      <li>options (ko.observableArray)</li>
       *      <li>list</li>
       *      <li>optionKeys</li>
       *    </ul>
       *  With native renderMode, the functionality that is sacrificed compared to jet rendermode is:
       *    <ul>
       *      <li>no search box (no filtering)</li>
       *      <li>for multiple select, only number of selected items is displayed in the selectbox, not the selected values</li>
       *      <li>beforeExpand event isn't fired when the popup picker opens.</li>
       *      <li>only one level nesting optgroups in the popup picker due to the HTML optgroup limitation</li>
       *      <li>no image support in the option list</li>
       *      <li>All Sub-IDs are not available in the native renderMode.</li>
       *      <li>pickerAttributes is not applied in the native renderMode.</li>
       *      <li>when using data provider a maximum of 15 rows will be displayed in the dropdown, users can't filter further.</li>
       *    </ul>
       * </ul>
       *
       * The default value depends on the theme. In alta-android, alta-ios and alta-windows themes, the default is "native" and it's "jet" for all other themes.
       *
       * @expose 
       * @memberof oj.ojSelect
       * @instance
       * @ojfragment selectCommonRenderMode
       */
      renderMode : "jet",

      /**
       * The <code class="prettyprint">valueOption</code> is similar to the <code class="prettyprint">value</code>, but is an
       * Object which contains both a value and display label. 
       * The <code class="prettyprint">value</code> and <code class="prettyprint">valueOption</code> are kept in sync. 
       * If initially both are set, the selected value in the <code class="prettyprint">value</code> attribute has precedence.
       * <p>Note: There may be a performance gain if the <code class="prettyprint">options</code> attribute is a data provider and 
       * the <code class="prettyprint">valueOption</code> is initially set then its label is used before the real data is fetched. 
       * If <code class="prettyprint">valueOption</code> is not specified or the selected value is missing, then the label will 
       * be fetched from the data provider.
       *
       * @name valueOption
       * @ojshortdesc The current value of the element and its associated display label.
       * @expose
       * @instance
       * @type {null | Object}
       * @default null
       *
       * @property {any} value current value of JET Select
       * @property {string} [label] display label of value above. If missing, String(value) is used. 
       * @memberof oj.ojSelectOne
       *
       * @example <caption>Initialize the Select with the <code class="prettyprint">value-option</code> attribute specified:</caption>
       * &lt;oj-select-one value-option="[[valueOption]]">&lt;/oj-select-one>
       *
       * @example <caption>Object with value and label properties:</caption>
       * var valueOption = {'value': 'val1', 'label': 'Label 1'};
       *
       * @example <caption>Get or set the <code class="prettyprint">valueOption</code> property after initialization:</caption>
       * // getter
       * var valueOption = mySelect.valueOption;
       *
       * // setter
       * mySelect.valueOption = valueOption;
       */
      valueOption: null,

      /**
       * The <code class="prettyprint">valueOptions</code> is similar to the <code class="prettyprint">value</code>, but is an array
       * of Objects and each Object contains both a value and display label. 
       * The <code class="prettyprint">value</code> and <code class="prettyprint">valueOptions</code> are kept in sync.
       * If initially both are set, the selected values in the <code class="prettyprint">value</code> attribute has precedence.
       * <p>Note: There may be a performance gain if the <code class="prettyprint">options</code> attribute is a data provider and the 
       * <code class="prettyprint">valueOptions</code> is initially set then its labels are used before the real data is fetched. 
       * If <code class="prettyprint">valueOptions</code> is not specified or one of the selected value is missing, then the labels 
       * will be fetched from the data provider.
       *
       * @name valueOptions
       * @ojshortdesc The current values of the element and theirs associated display labels.
       * @expose
       * @instance
       * @type {null | Array.<Object>}
       * @default null
       *
       * @property {any} value a current value of JET Select
       * @property {string} [label] display label of value above. If missing, String(value) is used. 
       * @ojsignature { target: "Type",
       *                value: "Array<{value: any, label?: string}> | null",
       *                jsdocOverride: true}
       *
       * @memberof oj.ojSelectMany
       *
       * @example <caption>Initialize the Select with the <code class="prettyprint">value-options</code> attribute specified:</caption>
       * &lt;oj-select-many value-options="[[optionsArray]]">&lt;/oj-select-many>
       *
       * @example <caption>Array of Objects with value and label properties:</caption>
       * var optionsArray = [{'value': 'val1', 'label': 'Label 1'}, 
       *                     {'value': 'val2', 'label': 'Label 2'}];
       *
       * @example <caption>Get or set the <code class="prettyprint">valueOptions</code> property after initialization:</caption>
       * // getter
       * var valueOptions = mySelect.valueOptions;
       *
       * // setter
       * mySelect.valueOptions = optionsArray;
       */
      valueOptions: null

      /**
       * The value of the element. It supports any type. Select only accepts a value that's in the drop down list. 
       * Trying to set a new value that's not in the drop down list fails validation and the new value is not set. 
       * <p>If a value is specified before the data for the drop down list is available, then that value is set initially.
       * When the data for the drop down list is available, the initially set value is validated. 
       * If it fails validation, the first option will be set as the value instead.</p>
       *
       * @example <caption>Initialize the select with the <code class="prettyprint">value</code> attribute specified:</caption>
       * &lt;oj-select-one value="option1">&lt;/oj-select-one>
       *
       * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
       * // getter
       * var value = mySelect.value;
       *
       * // setter
       * mySelect.value = "option1";
       *
       * @member
       * @name value
       * @ojshortdesc The value of the element.
       * @access public
       * @instance
       * @default When the value attribute is not set, the first option is used as its initial value if it exists.
       * @memberof oj.ojSelectOne
       * @type {*}
       * @ojwriteback
       */
      /**
       * The value of the element. The value is an array with any type of items.
       *
       * @example <caption>Initialize the select with the <code class="prettyprint">value</code> attribute specified:</caption>
       * &lt;oj-select-many value="{{val}}">&lt;/oj-select-many>
       * @example var val = ['option1', 'option2'];
       *
       * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
       * // getter
       * var value = mySelect.value;
       *
       * // setter
       * mySelect.value = ["option1", "option2"];
       *
       * @member
       * @name value
       * @ojshortdesc The value of the element.
       * @access public
       * @instance
       * @memberof oj.ojSelectMany
       * @type {Array.<*>}
       */
    },

    /**
     * Returns a jQuery object containing the element visually representing the select.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof! oj.ojSelect
     * @instance
     * @public
     * @ignore
     * @return {jQuery} the select
     */
    widget : function ()
    {
      // native renderMode
      if (this.select)
        return this.select.container;
      else
        return this.element.parent();
    },

    /**
     * @override
     * @protected
     * @instance
     * @memberof! oj.ojSelect
     */
    _ComponentCreate : function ()
    { 
      this._super();
      _ComboUtils.wrapDataProviderIfNeeded(this, null);
      this._setup();
    },
    
    /**
     * @override
     * @protected
     * @instance
     * @memberof! oj.ojSelect
     */   
    _AfterCreate: function () 
    {
      this._super();
      
      // For custom element syntax, we need to get the label id and
      // set aria-labelledby on the focusable element.
      if (this._IsCustomElement()) 
      {
        var defaultLabelId = this["uuid"] + "_Label";
        var labelId = oj.EditableValueUtils.getOjLabelId($(this.OuterWrapper), defaultLabelId);
        if (labelId)
          this._GetContentElement().attr("aria-labelledby", labelId);
      }
    },
    
    /**
      * Performs post processing after _SetOption() is called. Different options when changed perform
      * different tasks. See _AfterSetOption[OptionName] method for details.
      *
      * @param {string} option
      * @param {Object|string=} previous
      * @param {Object=} flags
      * @protected
      * @memberof! oj.ojSelect
      * @instance
      * @override
      */
    _AfterSetOption : function (option, previous, flags)
    {
      this._superApply(arguments);
      switch (option)
      {
        case "required":
          this._AfterSetOptionRequired(option);
          break;
        default:
          break;
      }
    },
    
    /**
     * Whether the component is required.
     * 
     * @return {boolean} true if required; false
     * 
     * @memberof! oj.ojSelect
     * @instance
     * @protected
     * @override
     */
    _IsRequired : function () 
    {
      return this.options['required'];
    },
    
    /**
     * Performs post processing after required option is set by taking the following steps.
     * 
     * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
     * run full validation with UI value (we don't know if the UI error is from a required validator 
     * or something else);<br/>
     * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
     * updated<br/>
     * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
     * listen to optionChange(value) to clear custom errors.<br/>
     * 
     * - if component is invalid and has messagesHidden -> required: false -> clear component 
     * errors; no deferred validation is run.<br/>
     * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
     * errors unnecessarily)<br/>
     * - messagesCustom is never cleared<br/>
     * 
     * @param {string} option
     * 
     * @memberof! oj.ojSelect
     * @instance
     * @protected
     */
    _AfterSetOptionRequired : oj.EditableValueUtils._AfterSetOptionRequired, 
    
    // native renderMode
    _nativeSetDisabled : function (disabled)
    {
      if (disabled) {
        this.element.attr("disabled", "");
        this.element.parent()
          .addClass("oj-disabled")
          .removeClass("oj-enabled");
      }
      else {
        this.element.removeAttr("disabled");
        this.element.parent()
          .removeClass("oj-disabled")
          .addClass("oj-enabled");
      }
    },

    _nativeChangeHandler : function (event)
    {
      var arr = [];
      var arrValOpts = [];
      var emptyValueAllowed = (! this._IsRequired() && this._HasPlaceholderSet());

      //NOTE: option and optgroup
      $(event.target).find('option').each(function() {
        if (this.selected && (this.value || 
                              (emptyValueAllowed && this.value === ""))) {
          arr.push(this.value);
          arrValOpts.push({"value": this.value, "label": this.text});
        }
      });

      // - ie11 multiple select with keyboard fails
      if (!this._IsCustomElement() || this.multiple === true ) {
        this._SetValue(arr, event, {doValueChangeCheck: false,
                                  '_context': {internalSet: true}});
        _ComboUtils.setValueOptions(this, arrValOpts);
      }
      else {
        this._SetValue(arr[0], event, {doValueChangeCheck: false,
                                  '_context': {internalSet: true}});
        _ComboUtils.setValueOptions(this, arrValOpts[0]);
      }
    },

    _nativeQueryCallback : function (data)
    {
      if (!data)
        return;

      //populate data in dropdown here
      var element = this.element;

      //if options is a dataProvider, it was wrapped with LPDV
      //don't pass in optionsKeys
      _ComboUtils.arrayPopulateResults(element, data,
        this._formatValue.bind(this), 
        _ComboUtils.isDataProvider(this.options.options) ? 
          null : this.options.optionsKeys);
                        
      element.addClass(_ComboUtils.GENERATED_OPTIONS_SELECTOR);
    },

    _nativeSetSelected : function (value)
    {
      var selected = null;
      if (value) {
        selected = value;
      }
      else {
        if (this._HasPlaceholderSet()) {
          if (this.options.required)
            selected = this._nativeFindFirstEnabledOptionValue();
          this._SetPlaceholder(this.options.placeholder);
        }

        // default to the first enabled option
        if (selected === null) {
          selected = this._nativeFindFirstEnabledOptionValue();
        }
      }
      this._setInitialSelectedValue(selected);
    },

    //fetch first block from data provider
    //when data available validate the selected value
    _nativeFetchFromDataProvider : function (dataProvider)
    {
      var self = this;
      _ComboUtils.fetchFirstBlockFromDataProvider(this.element, 
                                                  dataProvider).then( 
        function(data) {
          self._nativeQueryCallback(data);
          if (data.length) {
          //make sure value still valid
            var selected = self.options.value;
            if (selected) {
              _ComboUtils.validateFromDataProvider(self.element,
                                                   dataProvider, 
                                                   selected).then(
              function(results) {
                // - need to be able to specify the initial value of select components bound to dprv
                if (results) {
                  var valueOptions = results.valueOptions;
                  if (Array.isArray(valueOptions) && valueOptions.length) {
                    _ComboUtils.setValueOptions(self, valueOptions);
                  }
                  //only set value if it is valid
                  var values = results.value;
                  if (Array.isArray(values) && values.length) {
                    self._nativeSetSelected(self.multiple? values : values[0]);
                  }
                }
              });
            }
            //no selected value, default to placeholder or 1st item
            else {
              self._nativeSetSelected();
            }
          }
        });
    },

    _nativeSetup : function ()
    {
      var element = this.element;

      //add a <div> around <select> for validation error
      element.wrap("<div>").parent() // @HTMLUpdateOK
        .addClass("oj-select-native oj-component oj-select oj-form-control");
      element.addClass("oj-select-select oj-component-initnode");

      //multiple attr
      if (this.multiple) {
        if (! element[0].multiple)
          element[0].multiple = true;

        element.parent().prepend("<a class='oj-select-arrow oj-component-icon oj-clickable-icon-nocontext oj-select-multiple-open-icon' role='presentation'></a>");  // @HTMLUpdateOK
      }
      else {
        element.parent().prepend("<a class='oj-select-arrow oj-component-icon oj-clickable-icon-nocontext oj-select-open-icon' role='presentation'></a>");  // @HTMLUpdateOK
      }
      //disable attr
      this._nativeSetDisabled(this.options.disabled);

      if (this.options.list)
      {
        _ComboUtils.listPopulateResults(element, 
                                        $("#" + this.options.list).children(),
                                        this._formatValue.bind(this));
        element.addClass(_ComboUtils.GENERATED_OPTIONS_SELECTOR);
      }
      else if (this.options.options)
      {
        var dataProvider = _ComboUtils.getDataProvider(this.options);
        if (dataProvider) {
          this._nativeFetchFromDataProvider(dataProvider);
        }
        else {
          this._nativeQueryCallback(this.options.options);
        }
      }
      else if (this._IsCustomElement()) 
      {
//TODO: TEST if the dropdown only contains a placeholder don't call ojOptionPopulateResults
        var children = element.children();
        if (children.length !== 1 || ! children.hasClass("oj-listbox-placeholder")) {
          // handle custom element with oj-option, oj-optgroup case
          _ComboUtils.ojOptionPopulateResults(element, 
                                              children,
                                              this._formatValue.bind(this));        
          element.addClass(_ComboUtils.GENERATED_OPTIONS_SELECTOR);
        }
      }

      this._focusable({
        'element': element,
        'applyHighlight': true
      });

      // - the selected option of the ojselect not reflected in the value variable
      var dataProvider = _ComboUtils.getDataProvider(this.options);
      if (! this.options.value && ! this._HasPlaceholderSet() &&
          ! dataProvider) {
        this._setInitialSelectedValue(this._nativeFindFirstEnabledOptionValue());
      }

      //add a change listener
      element.change(this._nativeChangeHandler.bind(this));

      _ComboUtils.addDataProviderEventListeners(this);
    },

    // native renderMode
    _jetSetup : function ()
    {
      var opts = {};
      
      opts.element = this.element;
      opts.ojContext = this;
      opts = $.extend(this.options, opts);

      this.select = this.multiple ? new _OjMultiSelect() : new _OjSingleSelect();
      this.select._init(opts);

      this.select.container.addClass("oj-select-jet oj-form-control");

      this._focusable({
        'element': this.select.selection,
        'applyHighlight': true
      });

//      var arrow = this.select.selection.children(".oj-select-arrow");
//      this._AddHoverable(arrow);
//      this._AddActiveable(arrow);

    },

    // - the selected option of the ojselect not reflected in the value variable
    _setInitialSelectedValue : function (selectedVal)
    {
      var selected;
      
      if (!this._IsCustomElement())
        selected = Array.isArray(selectedVal)? selectedVal : [selectedVal];
      else 
        selected = selectedVal;
      
      this._SetValue(selected, 
                     null, {doValueChangeCheck: false,
                            '_context': {internalSet: true, writeback: true},
                            'changed': true,
                           });
    },

    //ojselect
    _setup : function ()
    {
      // - need to be able to specify the initial value of select components bound to dprv
      this._resolveValueOptionsLater = _ComboUtils.mergeValueAndValueOptions(this);
      this._isNative() ? this._nativeSetup() : this._jetSetup();
      this._refreshRequired(this.options['required']);
    },

    /**
     * Refreshes the visual state of the select. JET elements require a <code 
     * class="prettyprint">refresh()</code> after the DOM is programmatically
     * changed.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof oj.ojSelect
     * @instance
     * @public
     */
    refresh : function ()
    {
      this._super();

      // cleanup the old HTML and setup the new HTML markups
      this._cleanup();

      this._setup();
      //TODO: apply value in options for the selected value

      if (this._isNative() && this.options.value)
        this.element.val(this.options.value);
      
      //re apply root attributes settings
      this._SetRootAttributes();

    },
    /**
     * @memberof! oj.ojSelect
     * @instance
     * @private 
     */
    _refreshRequired : oj.EditableValueUtils._refreshRequired,

    /**
     * Called to find out if aria-required is unsupported.
     * @memberof! oj.ojSelect
     * @instance
     * @protected
     */
    _AriaRequiredUnsupported : function()
    {
      return false;
    },
    /**
     * @override
     * @private
     * @memberof! oj.ojSelect
     */
    _destroy : function ()
    {
      // native renderMode
      this._cleanup();
      this._super();
    },

    //19670760, dropdown popup should be closed on subtreeDetached notification.
    _NotifyDetached : function() {
      this._superApply(arguments);
      // native renderMode
      if (this.select)
        this.select.close();
    },

    //19670760, dropdown popup should be closed on subtreeHidden notification.
    _NotifyHidden : function() {
      this._superApply(arguments);
      // native renderMode
      if (this.select)
        this.select.close();
    },

    /**
     * Set the placeholder.
     * @override
     * @protected
     * @memberof! oj.ojSelect
     */
    _SetPlaceholder : function(value)
    {
      /*
       * Commented out the content to fix the side effect from
       * the change made in EditableValue._initComponentMessaging (Revision: 9016).
       * EditableValue called SetPlaceholder with an empty string which made
       * every ojselect to have an empty placeholder in the dropdown.
       * Note: don't remove the method because there is more calls from EditableValue
       * to ojselect before this.select is initialized.
       * ex: _GetContentElement

      if (this.select)
      {
        this.select.opts.placeholder = value;
        this.select._setPlaceholder();
      }
      */

      // native renderMode
      //add a placeholder option if needed
      if (this._isNative() && value != null) {
        var placeholder = $(this.element.children("option:first-child"));

        //if a placeholder option exists, use it
        if (placeholder && placeholder.attr("value") === "")
        {
          placeholder.text(this.options.placeholder);
          placeholder.attr("value", "");
        }
        else {
          placeholder = _ComboUtils.createOptionTag(0, "", value, this._formatValue.bind(this));
          placeholder.addClass("oj-listbox-placeholder");

          // - placeholder text is a selectable item that results in an error for ojcomponent
          this._hidePlaceholder(placeholder, this._IsRequired());
          placeholder.prependTo(this.element); // @HTMLUpdateOK
        }
      }
    },

    /**
     * whether the placeholder option is set
     *
     * @memberof! oj.ojSelect
     * @instance
     * @protected
     */
    _HasPlaceholderSet : function()
    {
      // - an empty placeholder shows up if data changed after first binding
      return typeof this.options['placeholder'] === 'string';
    },

    /**
     * Clear the placeholder option
     *
     * @memberof! oj.ojSelect
     * @instance
     * @protected
     */
    _ClearPlaceholder : function()
    {
      // - an empty placeholder shows up if data changed after first binding
      this._SetPlaceholderOption(null);
      this._SetPlaceholder(null);
    },

    /**
     * @memberof! oj.ojSelect
     * @instance
     * @protected
     * @override
     */
    _InitOptions : function (originalDefaults, constructorOptions)
    {
      var props = [{attribute: "disabled", validateOption: true},
                   {attribute: "placeholder"},
                   {attribute: "required",
                    coerceDomValue: true, validateOption: true},
                   {attribute: "title"}
                   // {attribute: "value"}
                 ];

      this._super(originalDefaults, constructorOptions);
      oj.EditableValueUtils.initializeOptionsFromDom(props, constructorOptions, this);

      this.multiple = !this._IsCustomElement() ? this.options['multiple'] :
                      this.OuterWrapper.nodeName === "OJ-SELECT-MANY" ?
                      true : false;
                      
      // TODO: PAVI - Let's discuss
      if (this.options['value'] === undefined) {
        if (!this._IsCustomElement())
          this.options['value'] = (this.element.attr('value') !== undefined) ? _ComboUtils.splitVal(this.element.val(), ",") : null;
      }
      else {
        //clone the value, otherwise _setDisplayValue will not be invoked on binding value to ko observableArray.
        //TODO: Need to revisit this once 18724975 is fixed.
        var value = this.options['value'];
        if (Array.isArray(value)) {
          if (!this._IsCustomElement())
            value = value.slice(0);
        }

        this.options['value'] = value;
      }
    },

    /**
     * <ol> 
     * <li>All messages are cleared, including custom messages added by the app. </li>
     * <li>The implicit required validator is run first if the component
     * is marked required. </li>
     * <li>At the end of validation if there are errors, the messages are shown. 
     * If there were no errors, then the 
     * <code class="prettyprint">value</code> option is updated.</li>
     * </ol>
     *
     * @example <caption>Validate component using its current value.</caption>
     * myComp.validate();
     * 
     * @example <caption>Validate component and use the Promise's resolved state.</caption>
     * myComp.validate().then(
     *  function(result) {
     *    if(result === "valid")
     *    {
     *      submitForm();
     *    }
     *  });
     * @return {Promise} Promise resolves to "valid" if the component passed validation. 
     * The Promise resolves to "invalid" if there were validation errors.
     * @method
     * @access public
     * @expose
     * @instance
     * @memberof oj.ojSelect
     * @ojstatus preview
     */
    validate : oj.EditableValueUtils.validate,

     /**
     * Updates display value.
     * @override
     * @protected
     * @memberof! oj.ojSelect
     */
    _SetDisplayValue: function(displayValue)
    {
      // native renderMode
      if (this.select)
      {
        //  - need to be able to specify the initial value of select components bound to dprv
        if (! _ComboUtils.applyValueOptions(this.select, this.options)) {
          this.select._initSelection();
        }
        this._resolveValueOptionsLater = false;
      }
      else
      {
        //if displayValue is null, let it default to 1st option or placeholder
        //see demo pages: disable and placeholder
        if (displayValue == null)
        {
          if (this._HasPlaceholderSet()) {
            this.element[0].selectedIndex = 0;
            this.element.addClass("oj-select-default");
          }
        }
        else {
          this.element.val(displayValue);
        }
        //update valueOptions
        if (this._resolveValueOptionsLater) {
          this._resolveValueOptionsLater = false;
          var opts;
          if (displayValue == null) {
            opts = {"value": "", "label": this.options.placeholder};
            if (this.multiple)
              opts = [opts];
            _ComboUtils.setValueOptions(this, opts);
          }
          else {
            if (this.multiple) {
              var options = this.element[0].selectedOptions;
              opts = [];
              for (var i = 0; i < options.length; i++) {
                opts.push({"value": displayValue[i], "label": options[i].text});
              }
            }
            else {
              var label = this.element[0].options[this.element[0].selectedIndex].text;
              opts = {"value": displayValue, "label": label};
            }

            _ComboUtils.setValueOptions(this, opts);
          }
        }
      }
    },
    
    /**
     * Returns true if validation passes, false otherwise
     *
     * @return {boolean} true if validation passes, false otherwise
     * @memberof! oj.ojSelect
     * @override
     * @protected
     * @instance
     * @ignore
     */
    _ValidateReturnBoolean: function()
    {
      if (this.multiple === true)
      {
        var displayValue = this.select.search.val();  
        var existingValue = [];
        var newValue = null;
        if (this.isValid())
            existingValue = this.select.getVal();
        
        if (displayValue === undefined || displayValue === null || displayValue === "")
          newValue = existingValue;
        else {
          existingValue.push(displayValue);
          newValue = existingValue;
        }
        return this._SetValue(newValue, null, this._VALIDATE_METHOD_OPTIONS);
      }
      
      // - select needs implementation fixes...
      if (this.select)
        return this._SetValue(this.select.getVal(), null, this._VALIDATE_METHOD_OPTIONS);

      return true;

    },

    // native renderMode
    _nativeFindFirstEnabledOptionValue : function ()
    {
      var enaOptions = this.element.children("option:not(:disabled)");
      if (enaOptions.length > 0)
        return [$(enaOptions[0]).attr("value")];

      return null;
    },

    _nativeSetOptions : function (options)
    {
      var oSelected = this.options.value;
      var element = this.element;

      //if options list is generated during setup, remove it
      if (element.hasClass(_ComboUtils.GENERATED_OPTIONS_SELECTOR)) {
        _ComboUtils.cleanupResults(element);
      }
      else {
        var children = element.children();
        if (children.length > 0)
          children.remove();
      }

      //if data provider, fetch data
      if (_ComboUtils.isDataProvider(options)) {
        this._nativeFetchFromDataProvider(options);
      }
      else {
        _ComboUtils.arrayPopulateResults(element, 
                                         options,
                                         this._formatValue.bind(this), 
                                         this.options.optionsKeys);
      var defVal = null;
      if (this._HasPlaceholderSet())
      {
        if (this.options.required)
          defVal = this._nativeFindFirstEnabledOptionValue();

        this._SetPlaceholder();
      }

      // default to the first enabled option
      if (defVal === null)
        defVal = this._nativeFindFirstEnabledOptionValue();

      this.options.value = defVal;

      this.option("value", oSelected);
      }
      element.addClass(_ComboUtils.GENERATED_OPTIONS_SELECTOR);
    },

    _removePlaceholderInMultiValues : function (valArr)
    {
      var val, narr = [];
      for (var i = 0; i < valArr.length; i++)
      {
        val = valArr[i];
        if (val != null) {
          if (val.length > 0) {
            //remove placeholder if it was already added to arr
            if (narr.length == 1 && narr[0] === "")
              narr.pop();
            narr.push(val);
          }
          // val is a placeholder
          else if (narr.length == 0)
            narr.push(val);
        }
      }
      return narr;
    },

    /**
     * Handles options specific to select.
     * @override
     * @protected
     * @memberof! oj.ojSelect
     */
    _setOption : function (key, value, flags)
    {
      if (key === "value") {
        //clone the value, otherwise _setDisplayValue will not be invoked on binding value to ko observableArray.
        //TODO: Need to revisit this once 18724975 is fixed.

        if (this._HasPlaceholderSet() && 
            ((value && value.length == 0) ||
             (this._IsCustomElement() && value==""))) {

          // - placeholder is not displayed after removing selections from select many
          _ComboUtils.setValueOptions(this, this.multiple? [] : "");
          this._super(key, value, flags);
          return;
        }

        // native renderMode
        var element;
        if (this.select)
        {
          element = this.select.datalist;
          if (! element)
            element = this.select.opts.element;
        }
        
        var customSelectOne = this._IsCustomElement() && !this.multiple;

        // turn value to an array
        if (! Array.isArray(value) && !customSelectOne)
          value = [value];

        if (this._isNative())
        {
          if (!customSelectOne)
            value = this._removePlaceholderInMultiValues(value);

          //set oj-select-default on the select tag if the selected value is a placeholder and a singleton
          if (value.length == 1 && value[0] == "")
            this.element.addClass("oj-select-default");
          else
            this.element.removeClass("oj-select-default");
        }

        var dataProvider = _ComboUtils.getDataProvider(this.options);
        if (dataProvider && value) {
          var self = this;
          var selfSuper = this._super;
          this.select.opts.options = dataProvider;

          _ComboUtils.validateFromDataProvider(
            this._isNative()? this.element : this.select.container, 
            dataProvider, value).then(
            function(results) {
              // - need to be able to specify the initial value of select components bound to dprv
              if (results) {
                var valueOptions = results.valueOptions;
                if (Array.isArray(valueOptions) && valueOptions.length) {
                  _ComboUtils.setValueOptions(self, valueOptions);
                }
                //only set value if it is valid
                var values = results.value;
                if (Array.isArray(values) && values.length) {
                  selfSuper.call(self, key, self.multiple? values : values[0]);
                }
              }
            });
            // - ojselect allows to set invalid value when using dataprovider
            return;
        }
        // - ojselect should ignore the invalid value set programmatically
        else if (!customSelectOne) {
          var newArr = [];
          for (var i = 0; i < value.length; i++) {
            if (this.select)
            {
              //Note: both multi select and remote data cases, the validate function is not available
              if (! this.select.opts.validate ||
                  this.select.opts.validate(element, value[i]) ||
                  this._isOptionDataPending())
                newArr.push(value[i]);
            }
            else
            {
              if (this.element.find("option[value='" + value[i] + "']").length > 0)
                newArr.push(value[i]);
            }
          }
        
          //only set values that are valid
          // - can't remove last selected value in multi-select ojselect
          //multi select allows empty array
          if (newArr.length > 0 || this.multiple) {
            this._super(key, newArr, flags);
            // - need to be able to specify the initial value of select components bound to dprv
            _ComboUtils.updateValueOptions(this.select);
          }
          return;
        } 
        else {
          if (! (this.select && this.select.opts.validate) ||
              this.select.opts.validate(element, value) ||
              this._isOptionDataPending()) {
            this._super(key, value, flags);

            // - need to be able to specify the initial value of select components bound to dprv
            _ComboUtils.updateValueOptions(this.select);
          }
          return;
        }
      }
      else if (key === "placeholder")
      {
        // native renderMode
        if (this.select) {
          this.select.opts.placeholder = value;
          this.select._setPlaceholder();
        }
        else {
          var selected = this.options.value;
          if (! selected || selected.length === 0 || ! selected[0])
            this.element[0].selectedIndex = 0;
        }
      }

      else if (key === "minimumResultsForSearch")
      {
        // native renderMode
        if (this.select)
          this.select.opts.minimumResultsForSearch = value;
      }

      else if (key === "renderMode")
      {
        this._cleanup();
        this.options.renderMode = value;
        this.refresh();
      }

      //if we have a new data provider, remove the old dataProvider event listeners 
      if (key === "options") {
        _ComboUtils.removeDataProviderEventListeners(this);
        _ComboUtils.clearDataProviderWrapper(this);
      }
      this._super(key, value, flags);

      if (key === "disabled")
      {
        if (this.select)
        {
          if (value)
            this.select._disable();
          else
            this.select._enable();
        }
        else {
          this._nativeSetDisabled(value);
        }
      }
      // - need to be able to specify the initial value of select components bound to dprv
      else if (key === "valueOption" && this.multiple !== true) {
        _ComboUtils.syncValueWithValueOption(this, value, null);
      }
      else if (key === "valueOptions" && this.multiple === true && Array.isArray(value)) {
        _ComboUtils.syncValueWithValueOptions(this, value, null);
      }
      else if (key === "options")
      {
        //if options is a new data provider
        //wrap it with LVDP if it doesn't implement FetchByKeys
        //add event listeners
        if (_ComboUtils.isDataProvider(value)) {
          //update internal dataProviderWrapper
          _ComboUtils.wrapDataProviderIfNeeded(this, this.select ? this.select.opts : null);
          _ComboUtils.addDataProviderEventListeners(this);
        }
        if (this.select)
        {
          //make sure the value still valid
          var selected = this.select.getVal();
          var dataProvider = _ComboUtils.getDataProvider(this.options);

          if (dataProvider && selected) {
            var self = this;
            var selfSuper = this._super;

            //  - need to be able to specify the initial value of select components bound to dprv
            if (_ComboUtils.applyValueOptions(this.select, this.options)) {
              this.select.opts.options = value;
              this.select.opts = self.select._prepareOpts(this.select.opts);
            }
            else {
              _ComboUtils.validateFromDataProvider(this.select.container, dataProvider, 
                                                   this.options.value).then(
              function(results) {
                // - need to be able to specify the initial value of select components bound to dprv
                var values = results? results.value : null;
                if (values) {
                  var valueOptions = results.valueOptions;
                  if (Array.isArray(valueOptions) && valueOptions.length) {
                    _ComboUtils.setValueOptions(self, valueOptions);
                  }
                  //only set value if it is valid
                  if (Array.isArray(values) && values.length) {
                    selfSuper.call(self, "value", self.multiple? values : values[0]);
                  }
                }
                //use placeholder if specified
                else if (self.options.placeholder) {
                  self.select._updateSelectedOption(self.options.placeholder);
                }
                //the selected value is no longer valid, fetch 1st item from data provider
                else {
                  _ComboUtils.fetchFirstBlockFromDataProvider(self.select.container, 
                                                              dataProvider, 1).then(
                    function(data) {
                      //At this point if we still don't have a selected value then default to the first item
                      if (data && data.length > 0 && selected === self.select.getVal()) {
                        self.select._updateSelectedOption(data[0]);
                      }
                      else {
                        selfSuper.call(self, "value", null);
                        //update select box
                        if (! self.multiple)
                          self.select.text.text("");
                      }
                    });
                }
                self.select.opts.options = value;
                self.select.opts = self.select._prepareOpts(self.select.opts);
              });
            }
          }
          else {
            // - an empty placeholder shows up if data changed after first binding
            // - ojselect - validator error message is not shown
            // - ojselect tooltip no longer appears once options and value observables change
            this.select.opts.options = value;
            this.select.opts = this.select._prepareOpts(this.select.opts);

            //make sure the value still valid
            this.select.setValOpts(null);
            this._super("value", selected);
          }
        }
        else
        {
          this._nativeSetOptions(value);
        }
      }
      else if (key === "required" && this._isNative())
      {
        var placeholder = $(this.element.find(".oj-listbox-placeholder"));
        if (placeholder && placeholder.attr("value") === "")
        {
          //hide placeholder when required is true
          this._hidePlaceholder(placeholder, value);
        }
      }
/*
      else if (key === "pickerAttributes")
      {
        if (this.select)
          this.select._setPickerAttributes(value);
      }
*/
    },
    
    _isOptionDataPending : function ()
    {
      var options = this['options']['options'];
      var datalist = this.select.datalist;
      
      if (datalist)
      {
        // check if the data provided via html
        if (datalist.children().length === 0)
          return true;
      } 

      //  - replacing options in select after value is set -> writes over the value
      // skip validation if a value is specified when the data is empty, don't fetch data until user opens dropdown.
      // We validate value and set displayValue when the drop down data is available
      else if (_ComboUtils.isDataProvider(options)) {
        return true;
      }
      else 
      {
        // check the options array
        if (!options || options.length === 0)
          return true;
      }  
      return false;
    },

    _getDropdown : function ()
    {
      // native renderMode
      if (this.select && this.select._opened())
      {
        // - certain subids does not work inside a popup or dialog
        var dropdown = this.select.dropdown;
        if (dropdown &&
            dropdown.attr("data-oj-containerid") === this.select.containerId)
          return dropdown;
      }
      return null;
    },

    _hidePlaceholder: function(placeholder, hide) {
      if (hide) {
        placeholder.attr("disabled", "");
        placeholder.attr("hidden", "");
      }
      else {
        placeholder.removeAttr("disabled");
        placeholder.removeAttr("hidden");
      }
    },

    // native renderMode
    _isNative: function() {
      return this.options.renderMode === "native";
    },

    // native renderMode
    _cleanup: function() {
      var isNative = this._isNative();

      if (isNative && this.element.parent().hasClass("oj-select-native"))
      {
        //remove the change listner
        this.element.off("change");

        //if options list is generated during setup, remove it
        if (this.element.hasClass(_ComboUtils.GENERATED_OPTIONS_SELECTOR))
          _ComboUtils.cleanupResults(this.element);

        //remove wrapper
        if (this.element.parent().hasClass("oj-select-native")) {
          this.element.parent().children(".oj-select-arrow").remove();
          this.element.unwrap();
        }

        this.element.removeClass("oj-select-select oj-component-initnode");
        this.element.attr({
          "aria-labelledby": ""
        });
      }
      else if (! isNative && this.select)
      {
        this.select._destroy();
        this.select = undefined;
      }
    },


//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the selected text in the select box. This is not available in the native renderMode.</p>
 *
 * @ojsubid oj-select-chosen
 * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
 * @memberof oj.ojSelect
 * @ignore
 * @instance
 */

/**
 * <p>Sub-ID for the dropdown box. See the <a href="#minimumResultsForSearch">minimum-results-for-search</a> attribute for details. This is not available in the native renderMode.</p>
 *
 * @ojsubid oj-select-drop
 * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
 * @memberof oj.ojSelect
 * @ignore
 */

/**
 * <p>Sub-ID for the search box. Note:</p>
 * <ul>
 * <li>the search box is not always visible</li>
 * <li>the Sub-Id is not available in the native renderMode</li>
 * </ul>
 * <p>See the <a href="#minimumResultsForSearch">minimum-results-for-search</a> attribute for details.
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.
 *
 * @ojsubid oj-select-search
 * @ignore
 * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
 * @memberof oj.ojSelect
 */

/**
 * <p>Sub-ID for the search input element. Note:</p>
 * <ul>
 * <li>the search input is not always visible</li>
 * <li>the Sub-Id is not available in the native renderMode</li>
 * </ul>
 *
 * <p>See the <a href="#minimumResultsForSearch">minimum-results-for-search</a> attribute for details.</p>
 *
 * @ojsubid oj-listbox-input
 * @deprecated 1.2.0 please see oj-select-input
 * @memberof oj.ojSelect
 * @ignore
 */

/**
 * <p>Sub-ID for the filtered result list. This Sub-Id is not available in the native renderMode.</p>
 *
 * @ojsubid oj-select-results
 * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
 * @memberof oj.ojSelect
 * @ignore
 */

/**
 * <p>Sub-ID for the filtered result item. Note:</p>
 * <ul>
 * <li>To lookup a filtered result item, the dropdown must be open and the locator object should have: 
 *     subId: 'oj-listbox-result-label' and index: number.
 * </li>
 * <li>the Sub-Id is not available in the native renderMode</li>
 * </ul>
 *
 * @ojsubid oj-listbox-result-label
 * @deprecated 1.2.0 This sub-ID is not needed since it is not an interactive element.
 * @memberof oj.ojSelect
 * @ignore
 */

/**
 * <p>Sub-ID for the search input field. It's only visible when the results size is greater than the minimum-results-for-search threshold or when user starts typing into the select box. This Sub-Id is not available in the native renderMode.</p>
 * @ojsubid oj-select-input
 * @memberof oj.ojSelect
 *
 * @example <caption>Get the input field element</caption>
 * var node = myElement.getNodeBySubId({'subId': 'oj-select-input'});
 */

/**
 * <p>Sub-ID for the drop down arrow. This Sub-Id is not available in the native renderMode.</p>
 *
 * @ojsubid oj-select-arrow
 * @memberof oj.ojSelectOne
 *
 * @example <caption>Get the drop down arrow of the select</caption>
 * var node = myElement.getNodeBySubId({'subId': 'oj-select-arrow'});
 */
     
/**
 * <p>Sub-ID for the list item. This Sub-Id is not available in the native renderMode.</p>
 *
 * @ojsubid oj-listitem
 * @memberof oj.ojSelect
 *
 * @example <caption>Get the listitem corresponding to value "myVal"</caption>
 * var node = myElement.getNodeBySubId({'subId': 'oj-listitem', 'value': 'myVal'});
 */ 
    
/**
 * <p>Sub-ID for the remove icon of selected item. This Sub-Id is not available in the native renderMode.</p>
 *
 * @ojsubid oj-select-remove
 * @memberof oj.ojSelectMany
 *
 * @example <caption>Get the element corresponding to the remove icon for the selected item with 
 * value "myVal"</caption>
 * var node = myElement.getNodeBySubId({'subId': 'oj-select-remove', 'value': 'myVal'});
 */ 

    // @inheritdoc 
    getNodeBySubId: function(locator)
    {
      var node = null, subId;
      if (locator == null)
      {
        var container = this.widget();
        return container ? container[0] : null;
      }
      // doesn't support any sub ID in native mode
      else if (this._isNative())
      {
        return null;
      }
      else
      {
        node = this._super(locator);
      }

      if (!node)
      {
        var dropdown = this._getDropdown();

        subId = locator['subId'];
        switch (subId) {
          case "oj-select-drop":
            if (dropdown) {
              node = dropdown[0];
            }
            break;
          case "oj-select-results":
            if (dropdown)
              node = dropdown.find(".oj-listbox-results")[0];
            break;
          case "oj-select-search":
            if (dropdown)
              node = dropdown.find(".oj-listbox-search")[0];
            break;
          case "oj-select-input":
          case "oj-listbox-input":
            if (dropdown)
              node = dropdown.find(".oj-listbox-input")[0];
            break;
          case "oj-select-choice":
          case "oj-select-chosen":
          case "oj-select-arrow":
            node =  this.widget().find("." + subId)[0];
            break;
          case "oj-listitem":
            if (dropdown) 
            {
              var list = dropdown.find(".oj-listbox-result");
              node = this.select._findItem(list, locator['value']);
            }
            break;
          case "oj-select-remove":
            var selectedItems = this.widget().find(".oj-select-selected-choice");
            var item = this.select._findItem(selectedItems, locator['value']);
            node = item ? $(item).find(".oj-select-clear-entry-icon")[0] : null;
            break;

          // - ojselect - not able to attach id for generated jet component
          case "oj-listbox-result-label":
            if (dropdown)
            {
              //list of 'li'
              var ddlist =  $("#" + this.select.results.attr("id")).children();
              var index = locator['index'];

              if (ddlist.length && index < ddlist.length) {
                node = ddlist.eq(index).find("." + subId)[0];
              }
            }
            break;
        }
      }
      // Non-null locators have to be handled by the component subclasses
      return node || null;
    },

    /**
     * Returns the subId object for the given child DOM node.  For more details, see
     * <a href="#getNodeBySubId">getNodeBySubId</a>.
     *
     * @expose
     * @ignore
     * @override
     * @memberof oj.ojSelect
     * @instance
     *
     * @param {!Element} node - child DOM node
     * @return {Object|null} The subId for the DOM node, or <code class="prettyprint">null</code> when none is found.
     *
     * @example <caption>Get the subId for a certain DOM node:</caption>
     * var subId = myElement.getSubIdByNode(node);
     */
    getSubIdByNode: function(node)
    {
      if (this._isNative())
        return this._super(node);
        
      var subId = null;      
      if (node != null)
      {
        var nodeCached = $(node);

        if (nodeCached.hasClass('oj-listbox-input'))
          subId = {'subId': 'oj-select-input'};
        else if (nodeCached.hasClass('oj-select-arrow'))
          subId = {'subId': 'oj-select-arrow'};
        else if (nodeCached.hasClass('oj-listbox-result'))
          subId = {'subId': 'oj-listitem', 'value': nodeCached.data('ojselect')['value']};
        else if (nodeCached.hasClass('oj-select-clear-entry-icon'))
          subId = {'subId': 'oj-select-remove', 'value': nodeCached.closest('.oj-select-selected-choice').data('ojselect')['value']};
        else
          subId = this._super(node);
      }
      
      return subId;      
    },
    
    /**
     * Returns the default styleclass for the component. Currently this is
     * used to pass to the ojLabel component, which will append -label and
     * add the style class onto the label. This way we can style the label
     * specific to the input component. For example, for inline labels, the
     * radioset/checkboxset components need to have margin-top:0, whereas all the
     * other inputs need it to be .5em. So we'll have a special margin-top style
     * for .oj-label-inline.oj-radioset-label
     * All input components must override
     *
     * @return {string}
     * @expose
     * @memberof! oj.ojSelect
     * @override
     * @protected
     */
    _GetDefaultStyleClass : function ()
    {
      return "oj-select";
    },

    /**
     * Returns the messaging launcher element i.e., where user sets focus that triggers the popup.
     * Usually this is the element input or select that will receive focus and on which the popup
     * for messaging is initialized.
     *
     * @override
     * @protected
     * @memberof! oj.ojSelect
     * @return {Object} jquery element which represents the messaging launcher component
     */
    _GetMessagingLauncherElement : function ()
    {
      // native renderMode
      if (this.select)
        return this.select.selection;
      else
        return this.element;
    },

    /**
     * Returns the jquery element that represents the content part of the component.
     * This is usually the component that user sets focus on (tabindex is set 0) and
     * where aria attributes like aria-required, aria-labeledby etc. are set. This is
     * also the element where the new value is updated. Usually this is the same as
     * the _GetMessagingLauncherElement.
     *
     * @override
     * @protected
     * @memberof! oj.ojSelect
     * @return {jQuery} jquery element which represents the content.
     */
    _GetContentElement : function ()
    {
      // native renderMode
      if (this.select)
        return this.select.selection;
      else
        return this.element;
    }

    // Fragments:

    /**
     * <p>The &lt;oj-select-one> element accepts <code class="prettyprint">oj-option</code>s as children. See the [oj-option]{@link oj.ojOption} doc for details about
     * accepted children and slots.</p>
     *
     * @ojchild Default
     * @memberof oj.ojSelectOne
     *
     * @example <caption>Initialize the select with child content specified:</caption>
     * &lt;oj-select-one>
     *   &lt;oj-option value="option1">Option1&lt;/oj-option>
     * &lt;/oj-select-one>
     */

    /**
     * <p>The &lt;oj-select-many> element accepts <code class="prettyprint">oj-option</code>s as children. See the [oj-option]{@link oj.ojOption} doc for details about
     * accepted children and slots.</p>
     *
     * @ojchild Default
     * @memberof oj.ojSelectMany
     *
     * @example <caption>Initialize the select with child content specified:</caption>
     * &lt;oj-select-many>
     *   &lt;oj-option value="option1">Option1&lt;/oj-option>
     * &lt;/oj-select-many>
     */
     
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Select box or Arrow button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>If the drop down is not open, expand the drop down list. Otherwise, close the drop down list.
     *       If hints, title or messages exist in a notewindow, 
     *        pop up the notewindow on tapping on select box.</td>
     *     </tr>
     *     <tr>
     *       <td>Option item</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Tap on a option item in the drop down list to select a new item.</td>
     *     </tr>
     *     <tr>
     *       <td>Drop down</td>
     *       <td><kbd>swipe up/down</kbd></td>
     *       <td>Scroll the drop down list vertically</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDocOne - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojSelectOne
     */
     
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Select box</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>If the drop down is not open, expand the drop down list. Otherwise, close the drop down list.
     *       If hints, title or messages exist in a notewindow, 
     *        pop up the notewindow.</td>
     *     </tr>
     *     <tr>
     *       <td>Option item</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Tap on a option item in the drop down list to add to selection.</td>
     *     </tr>
     *     <tr>
     *       <td>Selected Item with Clear Entry Button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Remove item from the selected items list by taping on the clear button next to the data item.</td>
     *     </tr>
     *     <tr>
     *       <td>Drop down</td>
     *       <td><kbd>swipe up/down</kbd></td>
     *       <td>Scroll the drop down list vertically</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDocMany - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojSelectMany
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Option item</td>
     *       <td><kbd>Enter</kbd></td>
     *       <td>Select the highlighted choice from the drop down list.</tr>
     *     </tr>
     *     <tr>
     *       <td>Drop down</td>
     *       <td><kbd>UpArrow or DownArrow</kbd></td>
     *       <td>Highlight the option item in the direction of the arrow. If the drop down is not open, expand the drop down list.</tr>
     *     </tr>
     *     <tr>
     *       <td>Drop down</td>
     *       <td><kbd>Esc</kbd></td>
     *       <td>Collapse the drop down list. If the drop down is already closed, do nothing.</tr>
     *     </tr>
     *     <tr>
     *       <td>Select box or search box</td>
     *       <td><kbd>any characters for the search term</kbd></td>
     *       <td>filter down the results with the search term.</td>
     *     </tr>
     *     <tr>
     *       <td>Select</td>
     *       <td><kbd>Tab In</kbd></td>
     *       <td>Set focus to the select. If hints, title or messages exist in a notewindow, 
     *        pop up the notewindow.</td>
     *     </tr> 
     *   </tbody>
     * </table>
     *
     * <p>Disabled option items receive no highlight and are not selectable.
     *
     * @ojfragment keyboardDocOne - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojSelectOne
     */
     
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Option item</td>
     *       <td><kbd>Enter</kbd></td>
     *       <td>Select the highlighted choice from the drop down list.</tr>
     *     </tr>
     *     <tr>
     *       <td>Drop down</td>
     *       <td><kbd>UpArrow or DownArrow</kbd></td>
     *       <td>Highlight the option item in the direction of the arrow. If the drop down is not open, expand the drop down list.</tr>
     *     </tr>
     *     <tr>
     *      <td>Select box</td>
     *       <td><kbd>LeftArrow or RightArrow</kbd></td>
     *       <td> Move focus to the previous or next selected item.</td>
     *     </tr>
     *     <tr>
     *       <td>Drop down</td>
     *       <td><kbd>Esc</kbd></td>
     *       <td>Collapse the drop down list. If the drop down is already closed, do nothing.</tr>
     *     </tr>
     *     <tr>
     *       <td>Select box or search box</td>
     *       <td><kbd>any characters for the search term</kbd></td>
     *       <td>filter down the results with the search term.</td>
     *     </tr>
     *     <tr>
     *       <td>Select</td>
     *       <td><kbd>Tab In</kbd></td>
     *       <td>Set focus to the select. If hints, title or messages exist in a notewindow, 
     *        pop up the notewindow.</td>
     *     </tr> 
     *   </tbody>
     * </table>
     *
     * <p>Disabled option items receive no highlight and are not selectable.
     *
     * @ojfragment keyboardDocMany - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojSelectMany
     */


  }
  );

oj.Components.setDefaultOptions(
  {
    // converterHint is defaulted to placeholder and notewindow in EditableValue.
    // For ojselect, we don't want a converterHint.
    'ojSelect': // properties for all ojSelect components
    {
      'displayOptions':
        {
          'converterHint': ['none']
        },

      // native renderMode
      'renderMode': oj.Components.createDynamicPropertyGetter(
        function()
        {
          return (oj.ThemeUtils.parseJSONFromFontFamily('oj-select-option-defaults') || {})["renderMode"];
        })
    }

  }
);

(function() {
var ojComboboxOneMeta = {
  "properties": {
    "converter": {
      "type": "Object"
    },
    "filterOnOpen": {
      "type": "string",
      "enumValues": ["none", "rawValue"]
    },
    "minLength": {
      "type": "number"
    },
    "optionRenderer": {},
    "options": {
    },
    "optionsKeys": {
      "type": "Object",
      "properties": {
        "label": {
          "type": "string"
        },
        "value": {
          "type": "string"
        },
        "children": {
          "type": "string"
        },
        "childKeys": {
          "type": "Object",
          "properties": {
            "label": {
              "type": "string"
            },
            "value": {
              "type": "string"
            },
            "children": {
              "type": "string"
            }
          }
        }
      }
    },
    "pickerAttributes": {
      "type": "Object"
    },
    "placeholder": {
      "type": "string"
    },
    "rawValue": {
      "type": "string",
      "readOnly": true,
      "writeback": true
    },
    "required": {
      "type": "boolean"
    },  
    "translations": {
      "type": "Object",
      "properties": {
        "filterFurther": {
          "type": "string",
          "value": "More results available, please filter further."
        },
        "noMatchesFound": {
          "type": "string",
          "value": "No matches found"
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        }
      }
    },
    "value": {
      "type": "any",
      "writeback": true
    },
    "validators": {
      "type": "Array"
    },
    "valueOption": {
      "type": "Object",
      "writeback": true
    }
  },
  "events": {
    "valueUpdated": {}
  },
  "methods": {
    "validate": {}
  },
  "extension": {
    _INNER_ELEM: 'input',
    _WIDGET_NAME: "ojCombobox"
  }
};
oj.CustomElementBridge.registerMetadata('oj-combobox-one', 'editableValue', ojComboboxOneMeta);
oj.CustomElementBridge.register('oj-combobox-one', {'metadata': oj.CustomElementBridge.getMetadata('oj-combobox-one')});
})();

(function() {
var ojComboboxManyMeta = {
  "properties": {
    "converter": {
      "type": "Object"
    },
    "minLength": {
      "type": "number"
    },
    "optionRenderer": {},
    "options": {
    },
    "optionsKeys": {
      "type": "Object",
      "properties": {
        "label": {
          "type": "string"
        },
        "value": {
          "type": "string"
        },
        "children": {
          "type": "string"
        },
        "childKeys": {
          "type": "Object",
          "properties": {
            "label": {
              "type": "string"
            },
            "value": {
              "type": "string"
            },
            "children": {
              "type": "string"
            }
          }
        }
      }
    },
    "pickerAttributes": {
      "type": "Object"
    },
    "placeholder": {
      "type": "string"
    },
    "rawValue": {
      "type": "string",
      "readOnly": true,
      "writeback": true
    },
    "required": {
      "type": "boolean"
    },  
    "translations": {
      "type": "Object",
      "properties": {
        "filterFurther": {
          "type": "string",
          "value": "More results available, please filter further."
        },
        "noMatchesFound": {
          "type": "string",
          "value": "No matches found"
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        }
      }
    },
    "value": {
      "type": "Array",
      "writeback": true
    },
    "validators": {
      "type": "Array"
    },
    "valueOptions": {
      "type": "Array",
      "writeback": true
    }
  },
  "methods": {
    "validate": {}
  },
  "extension": {
    _INNER_ELEM: 'input',
    _WIDGET_NAME: "ojCombobox"
  }
};
oj.CustomElementBridge.registerMetadata('oj-combobox-many', 'editableValue', ojComboboxManyMeta);
oj.CustomElementBridge.register('oj-combobox-many', {'metadata': oj.CustomElementBridge.getMetadata('oj-combobox-many')});
})();

(function() {
var ojSelectOneMeta = {
  "properties": {
    "minimumResultsForSearch": {
      "type": "number"
    },
    "optionRenderer": {},
    "options": {
    },
    "optionsKeys": {
      "type": "Object",
      "properties": {
        "label": {
          "type": "string"
        },
        "value": {
          "type": "string"
        },
        "children": {
          "type": "string"
        },
        "childKeys": {
          "type": "Object",
          "properties": {
            "label": {
              "type": "string"
            },
            "value": {
              "type": "string"
            },
            "children": {
              "type": "string"
            }
          }
        }
      }
    },
    "pickerAttributes": {
      "type": "Object"
    },
    "placeholder": {
      "type": "string"
    },
    "renderMode": {
      "type": "string"
    },
    "required": {
      "type": "boolean"
    }, 
    "translations": {
      "type": "Object",
      "properties": {
        "filterFurther": {
          "type": "string",
          "value": "More results available, please filter further."
        },
        "moreMatchesFound": {
          "type": "string",
          "value": "num matches found"
        },
        "noMatchesFound": {
          "type": "string",
          "value": "No matches found"
        },
        "oneMatchesFound": {
          "type": "string",
          "value": "One match found"
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "searchField": {
          "type": "string",
          "value": "Search field"
        }
      }
    }, 
    "value": {
      "type": "any",
      "writeback": true
    },
    "valueOption": {
      "type": "Object",
      "writeback": true
    }
  },
  "methods": {
    "validate": {}
  },
  "extension": {
    _INNER_ELEM: 'select',
    _WIDGET_NAME: "ojSelect"
  }
};
oj.CustomElementBridge.registerMetadata('oj-select-one', 'editableValue', ojSelectOneMeta);
oj.CustomElementBridge.register('oj-select-one', {'metadata': oj.CustomElementBridge.getMetadata('oj-select-one')});
})();

(function() {
var ojSelectManyMeta = {
  "properties": {
    "minimumResultsForSearch": {
      "type": "number"
    },
    "optionRenderer": {},
    "options": {
    },
    "optionsKeys": {
      "type": "Object",
      "properties": {
        "label": {
          "type": "string"
        },
        "value": {
          "type": "string"
        },
        "children": {
          "type": "string"
        },
        "childKeys": {
          "type": "Object",
          "properties": {
            "label": {
              "type": "string"
            },
            "value": {
              "type": "string"
            },
            "children": {
              "type": "string"
            }
          }
        }
      }
    },
    "pickerAttributes": {
      "type": "Object"
    },
    "placeholder": {
      "type": "string"
    },
    "renderMode": {
      "type": "string"
    },
    "required": {
      "type": "boolean"
    },  
    "translations": {
      "type": "Object",
      "properties": {
        "filterFurther": {
          "type": "string",
          "value": "More results available, please filter further."
        },
        "moreMatchesFound": {
          "type": "string",
          "value": "num matches found"
        },
        "noMatchesFound": {
          "type": "string",
          "value": "No matches found"
        },
        "oneMatchesFound": {
          "type": "string",
          "value": "One match found"
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "searchField": {
          "type": "string",
          "value": "Search field"
        }
      }
    },
    "value": {
      "type": "Array",
      "writeback": true
    },
    "valueOptions": {
      "type": "Array",
      "writeback": true
    }
  },
  "methods": {
    "validate": {}
  },
  "extension": {
    _INNER_ELEM: 'select',
    _WIDGET_NAME: "ojSelect"
  }
};
oj.CustomElementBridge.registerMetadata('oj-select-many', 'editableValue', ojSelectManyMeta);
oj.CustomElementBridge.register('oj-select-many', {'metadata': oj.CustomElementBridge.getMetadata('oj-select-many')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @private
 */
var _config = oj.ThemeUtils.parseJSONFromFontFamily('oj-datepicker-config') || {};

/**
 * @private
 */
var _matchMedia = (function() {
  var dateTimePickerDropDownThresholdWidth = _config['dateTimePickerDropDownThresholdWidth'];
  var queryString;
  if(dateTimePickerDropDownThresholdWidth)
  {
    queryString = "(min-width: " + dateTimePickerDropDownThresholdWidth + ")";
  }

  return window.matchMedia(queryString || oj.ResponsiveUtils.getFrameworkQuery(oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.MD_UP) || "(min-width: 768px)");
})();

/**
 * @private
 */
var _isLargeScreen = _matchMedia.matches;
var resizeLargeScreenChange = function() {
  _isLargeScreen = _matchMedia.matches;
};

window.addEventListener('resize', resizeLargeScreenChange, false);

/**
 * @private
 */
function _getNativePickerDate(converter, isoString)
{
  isoString = converter.parse(isoString);

  var valueParams = oj.IntlConverterUtils._dateTime(isoString, ["date", "fullYear", "month", "hours", "minutes", "seconds"], true);
  var date = new Date();

  date.setFullYear(valueParams["fullYear"]);
  date.setDate(valueParams["date"]);
  date.setMonth(valueParams["month"]);
  date.setHours(valueParams["hours"]);
  date.setMinutes(valueParams["minutes"]);
  date.setSeconds(valueParams["seconds"]);
  date.setMilliseconds(0);

  return date;
}

/**
 * Placed here to avoid duplicate code for ojdatepicker + ojtimepicker
 *
 * Used for oj.EditableValueUtils.initializeOptionsFromDom
 *
 * @ignore
 */
function coerceIsoString(value)
{
  //reason for coersion is if one refreshes the page; then the input element's value might be the formatted string
  //thought about setting element's value to parsed value on destroy but goes against what destroy is suppose to do
  return this.options["converter"]["parse"](value);
}

/**
 * Placed here to avoid duplicate code for ojdatepicker + ojtimepicker
 *
 * @ignore
 */
function getImplicitDateTimeRangeValidator(options, converter, defaultStyleClass)
{
  var translationKeys = {'oj-inputdatetime': 'datetime', 'oj-inputtime': 'time', 'oj-inputdate': 'date'},
      dateTimeRangeTranslations = options['translations']['dateTimeRange'] || {},
          translations = [{'category': 'hint', 'entries': ['min', 'max', 'inRange']},
                          {'category': 'messageDetail', 'entries': ['rangeUnderflow', 'rangeOverflow']},
                          {'category': 'messageSummary', 'entries': ['rangeUnderflow', 'rangeOverflow']}],
          dateTimeRangeOptions = {'min': options['min'], 'max': options['max'], 'converter': converter,
                                  'translationKey': translationKeys[defaultStyleClass]};

  //note the translations are defined in ojtranslations.js, but it is possible to set it to null, so for sanity
  if(!$.isEmptyObject(dateTimeRangeTranslations))
  {
    for(var i=0, j=translations.length; i < j; i++)
    {
      var category = dateTimeRangeTranslations[translations[i]['category']];

      if(category)
      {
        var translatedContent = {},
            entries = translations[i]['entries'];

        for(var k=0, l=entries.length; k < l; k++)
        {
          translatedContent[entries[k]] = category[entries[k]];
        }

        dateTimeRangeOptions[translations[i]['category']] = translatedContent;
      }
    }
  }

  return oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE).createValidator(dateTimeRangeOptions);
}

/**
 * Shared for ojInputDate + ojInputTime
 *
 * @ignore
 */
function disableEnableSpan(children, val)
{
  var filteredChildren = children.filter("span");
  if (val)
  {
    filteredChildren.addClass("oj-disabled").removeClass("oj-enabled oj-default");
  }
  else
  {
    filteredChildren.removeClass("oj-disabled").addClass("oj-enabled oj-default");
  }
}

/**
 * For dayMetaData
 *
 * @ignore
 */
function _getMetaData(dayMetaData, position, params) {
  if(!dayMetaData || position === params.length) {
    return dayMetaData;
  }

  var nextPos = position + 1;
  return _getMetaData(dayMetaData[params[position]], nextPos, params) || _getMetaData(dayMetaData["*"], nextPos, params);
}

/**
 * Bind hover events for datepicker elements.
 * Done via delegate so the binding only occurs once in the lifetime of the parent div.
 * Global instActive, set by _updateDatepicker allows the handlers to find their way back to the active picker.
 *
 * @ignore
 */
function bindHover(dpDiv)
{
  var selector = ".oj-datepicker-prev-icon, .oj-datepicker-prev-icon .oj-clickable-icon-nocontext.oj-component-icon, .oj-datepicker-next-icon," +
    " .oj-datepicker-next-icon .oj-clickable-icon-nocontext.oj-component-icon, .oj-datepicker-calendar td a";
  return dpDiv.delegate(selector, "mouseout", function ()
  {
    $(this).removeClass("oj-hover");
  }).delegate(selector, "mouseover", function ()
  {
    $(this).addClass("oj-hover");
  }).delegate(selector, "focus", function ()
  {
    $(this).addClass("oj-focus");
  }).delegate(selector, "blur", function ()
  {
    $(this).removeClass("oj-focus");
  });
}

/**
 * Binds active state listener that set appropriate style classes. Used in
 * ojInputDate/ojInputDateTime/ojInputTime
 *
 * @ignore
 */
function bindActive(dateTime)
{
  var triggerRootContainer;
  if(dateTime._isInLine)
  {
    triggerRootContainer = $(dateTime.element[0]).parent();
  }
  else
  {
    triggerRootContainer = $(dateTime.element[0]).parent().parent();
  }

  // There are few issues in mobile using hover and active marker classes (iOS and Android, more
  // evident on iOS). Some fix is needed in _activeable(), tracking .
  dateTime._AddActiveable(triggerRootContainer);
}

/**
 * returns if the native picker is supported - depends on renderMode set to 'native' and this
 * cordova plugin being configured... https://github.com/VitaliiBlagodir/cordova-plugin-datepicker
 *
 * @ignore
 */
function isPickerNative(dateTime)
{
  // use bracket notation to avoid closure compiler renaming the variables
  return (dateTime.options['renderMode'] === "native" && window['cordova'] && window['datePicker']);
}

//to display the suffix for the year
var yearDisplay = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter(
{
  "year" : "numeric"
});


/**
 * @ojcomponent oj.ojDatePicker
 * @augments oj.ojInputDate
 * @since 4.0.0
 * @ojstatus preview
 * @ojshortdesc An inline element for picking a date value.
 * @ojdisplayname Inline Date Picker
 * @ojrole combobox
 *
 * @classdesc
 * <h3 id="inputDateOverview-section">
 *   JET DatePicker (Inline mode)
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputDateOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET DatePicker (Inline mode) provides basic support for datepicker selection. This will render the date picker as an inline element. Other behaviors of this element is similar to JET InputDate</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-date-picker>&lt;/oj-date-picker></code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate the label to the input element.
 * For DatePicker, you should put an <code>id</code> on the element, and then set
 * the <code>for</code> attribute on the label to be the element's id.
 * </p>
 * <h3 id="label-section">
 *   Label and DatePicker
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * For accessibility, you should associate a label element with the input
 * by putting an <code>id</code> on the input, and then setting the
 * <code>for</code> attribute on the label to be the input's id.
 * </p>
 * <p>
 * The DatePicker will decorate its associated label with required and help
 * information, if the <code>required</code> and <code>help</code> attributes are set.
 * </p>
 */

/**
 * @ojcomponent oj.ojInputDate
 * @augments oj.inputBase
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojInputDate<SP extends ojInputDateSettableProperties = ojInputDateSettableProperties> extends inputBase<string, SP>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojInputDateSettableProperties extends inputBaseSettableProperties<string>",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @since 0.6
 * @ojstatus preview
 * @ojshortdesc Provides basic support for specifying a date value.
 * @ojrole combobox
 * @ojtsimport ojvalidation-base
 * @ojtsimport ojvalidation-datetime
 *
 * @classdesc
 * <h3 id="inputDateOverview-section">
 *   JET InputDate
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputDateOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET InputDate provides basic support for datepicker selection.</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-input-date>&lt;/oj-input-date></code></pre>
 *
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate the label to the input element.
 * For InputDate, you should put an <code>id</code> on the element, and then set
 * the <code>for</code> attribute on the label to be the element's id.
 * </p>
 * <h3 id="label-section">
 *   Label and InputDate
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * For accessibility, you should associate a label element with the input
 * by putting an <code>id</code> on the input, and then setting the
 * <code>for</code> attribute on the label to be the input's id.
 * </p>
 * <p>
 * The InputDate will decorate its associated label with required and help
 * information, if the <code>required</code> and <code>help</code> attributes are set.
 * </p>
 */
oj.__registerWidget("oj.ojInputDate", $['oj']['inputBase'],
{
  widgetEventPrefix : "oj",

  //-------------------------------------From base---------------------------------------------------//
  _CLASS_NAMES : "oj-inputdatetime-input",
  _WIDGET_CLASS_NAMES : "oj-inputdatetime-date-only oj-component oj-inputdatetime",
  _ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES : "",
  _INPUT_HELPER_KEY: "inputHelp",
  _ATTR_CHECK : [{"attr": "type", "setMandatory": "text"}],
  _GET_INIT_OPTIONS_PROPS_FOR_WIDGET:  [{attribute: "disabled", validateOption: true},
                             {attribute: 'pattern'},
                             {attribute: "title"},
                             {attribute: "placeholder"},
                             {attribute: "value", coerceDomValue: coerceIsoString},
                             {attribute: "required",
                              coerceDomValue: true, validateOption: true},
                             {attribute: 'readonly', option: 'readOnly',
                             validateOption: true},
                             {attribute: "min", coerceDomValue: coerceIsoString},
                             {attribute: "max", coerceDomValue: coerceIsoString}],
  //-------------------------------------End from base-----------------------------------------------//

  _TRIGGER_CLASS : "oj-inputdatetime-input-trigger",
  _TRIGGER_CALENDAR_CLASS : "oj-inputdatetime-calendar-icon",

  _CURRENT_CLASS : "oj-datepicker-current-day",
  _DAYOVER_CLASS : "oj-datepicker-days-cell-over",
  _UNSELECTABLE_CLASS : "oj-datepicker-unselectable",

  _DATEPICKER_DIALOG_DESCRIPTION_ID : "oj-datepicker-dialog-desc",
  _DATEPICKER_DESCRIPTION_ID : "oj-datepicker-desc",
  _CALENDAR_DESCRIPTION_ID : "oj-datepicker-calendar",
  _MAIN_DIV_ID : "oj-datepicker-div",

  _INLINE_CLASS : "oj-datepicker-inline",
  _INPUT_CONTAINER_CLASS : " oj-inputdatetime-input-container",
  _INLINE_WIDGET_CLASS: " oj-inputdatetime-inline",

  _ON_CLOSE_REASON_SELECTION: "selection",  // A selection was made
  _ON_CLOSE_REASON_CANCELLED: "cancelled",  // Selection not made
  _ON_CLOSE_REASON_TAB: "tab",              // Tab key
  _ON_CLOSE_REASON_CLOSE: "close",          // Disable or other closes

  _KEYBOARD_EDIT_OPTION_ENABLED: "enabled",
  _KEYBOARD_EDIT_OPTION_DISABLED: "disabled",

  options :
  {
    /**
     * <p>
     * Note that Jet framework prohibits setting subset of properties which are object types.<br/><br/>
     * For example myInputDate.datePicker = {footerLayout: "today"}; is prohibited as it will
     * wipe out all other sub-properties for "datePicker" object.<br/><br/> If one wishes to do this [by above syntax or knockout] one
     * will have to get the "datePicker" object, modify the necessary sub-property and pass it to above syntax.<br/><br/>
     * Default values for the datePicker sub-properties can also be overridden with the theming variable
     * <code class="prettyprint">$inputDateTimeDatePickerOptionDefault</code>, which is merged with other defaults.<br/><br/>
     * Note that all of the datePicker sub-properties except showOn are not available when renderMode is 'native'.<br/><br/>
     *
     * @memberof oj.ojInputDate
     * @ojfragment datePickerCommonDatePicker
     */
    /**
     * {@ojinclude "name":"datePickerCommonDatePicker"}
     *
     * @expose
     * @instance
     * @memberof oj.ojDatePicker
     * @name datePicker
     * @type {Object}
     * @ojtsignore
     *
     * @example <caption>Override defaults in the theme (SCSS) :</caption>
     * $inputDateTimeDatePickerOptionDefault: (footerLayout: 'today', weekDisplay: 'number') !default;
     *
     * @example <caption>Initialize the component, overriding some date-picker attributes and leaving the others intact:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-date-picker date-picker.some-key='some value' date-picker.some-other-key='some other value'>&lt;/oj-date-picker>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-date-picker date-picker='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-date-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">datePicker</code> property after initialization:</caption>
     * // Get one
     * var value = myComponent.datePicker.someKey;
     *
     * // Set one, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myComponent.setProperty('datePicker.someKey', 'some value');
     *
     * // Get all
     * var values = myComponent.datePicker;
     *
     * // Set all.  Must list every datePicker key, as those not listed are lost.
     * myComponent.datePicker = {
     *     someKey: 'some value',
     *     someOtherKey: 'some other value'
     * };
     *
     */
    /**
     * {@ojinclude "name":"datePickerCommonDatePicker"}
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputDate
     * @type {Object}
     *
     * @example <caption>Override defaults in the theme (SCSS) :</caption>
     * $inputDateTimeDatePickerOptionDefault: (footerLayout: 'today', weekDisplay: 'number') !default;
     *
     * @example <caption>Initialize the component, overriding some date-picker attributes and leaving the others intact:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-input-date date-picker.some-key='some value' date-picker.some-other-key='some other value'>&lt;/oj-input-date>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-input-date date-picker='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-input-date>
     *
     * @example <caption>Get or set the <code class="prettyprint">datePicker</code> property after initialization:</caption>
     * // Get one
     * var value = myComponent.datePicker.someKey;
     *
     * // Set one, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myComponent.setProperty('datePicker.someKey', 'some value');
     *
     * // Get all
     * var values = myComponent.datePicker;
     *
     * // Set all.  Must list every datePicker key, as those not listed are lost.
     * myComponent.datePicker = {
     *     someKey: 'some value',
     *     someOtherKey: 'some other value'
     * };
     *
     */
    datePicker:
    {
      /**
       * Will dictate what content is shown within the footer of the calendar.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.footerLayout
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string}
       * @ojvalue {string} '' Do not show anything
       * @ojvalue {string} 'today' Show the today button
       * @default "today"
       */
      footerLayout : "",  

      /**
       * Whether the month should be rendered as a button to allow selection instead of text.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.changeMonth
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string}
       * @ojvalue {string} 'select' As a button
       * @ojvalue {string} 'none' As text
       * @default "select"
       */
      changeMonth : "select",

      /**
       * Whether the year should be rendered as a button to allow selection instead of text.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.changeYear
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string}
       * @ojvalue {string} 'select' As a button
       * @ojvalue {string} 'none' As text
       * @default "select"
       */
      changeYear : "select",

      /**
       * The position in multipe months at which to show the current month (starting at 0).
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.currentMonthPos
       * @memberof! oj.ojInputDate
       * @instance
       * @type {number}
       * @default 0
       * @ojmin 0
       * @ojmax 12
       */
      currentMonthPos : 0,

      /**
       * Dictates the behavior of days outside the current viewing month.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.daysOutsideMonth
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string}
       * @ojvalue {string} 'hidden' Days outside the current viewing month will be hidden
       * @ojvalue {string} 'visible' Days outside the current viewing month will be visible
       * @ojvalue {string} 'selectable' Days outside the current viewing month will be visible + selectable
       * @default "hidden"
       */
      daysOutsideMonth : "hidden",

      /**
       * The number of months to show at once. Note that if one is using a 
       * numberOfMonths > 4 then one should define a CSS rule for the width of 
       * each of the months. For example if numberOfMonths is set to 6 then one 
       * should define a CSS rule .oj-datepicker-multi-6 .oj-datepicker-group
       * providing the width each month should take in percentage.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.numberOfMonths
       * @memberof! oj.ojInputDate
       * @instance
       * @type {number}
       * @default 1
       * @ojmin 1
       */
      numberOfMonths : 1,

      /**
       * When the datepicker should be shown.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.showOn
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string}
       * @ojvalue {string} 'focus' when the element receives focus or when the trigger calendar image is
       *   clicked. When the picker is closed, the field regains focus and is editable.
       * @ojvalue {string} 'image' when the trigger calendar image is clicked
       * @default "focus"
       */
      showOn : "focus",

      /**
       * How the prev + next will step back/forward the months. The following are the valid values:
       * 
       * <ul>
       * <li>
       * <code class="prettyprint">"numberOfMonths"</code> - When set to this string, will use numberOfMonths property value as value.
       * </li>
       * <li>
       * <code class="prettyprint"> &lt;number&gt;</code> - Number of months to step back/forward.
       * </li>
       * </ul>
       * 
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.stepMonths
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string|number}
       * @ojsignature {target: "Type", value:"'numberOfMonths'|number"}
       * @default "numberOfMonths"
       */
      stepMonths : "numberOfMonths",

      /**
       * Number of months to step back/forward for the (Alt + Page up) + (Alt + Page down) key strokes.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.stepBigMonths
       * @memberof! oj.ojInputDate
       * @instance
       * @type {number}
       * @default 12
       */
      stepBigMonths : 12,

      /**
       * Whether week of the year will be shown.
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.weekDisplay
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string}
       * @ojvalue {string} 'number' Will show the week of the year as a number
       * @ojvalue {string} 'none' Nothing will be shown
       * @default "none"
       */
      weekDisplay : "none", // "number" to show week of the year, "none" to not show it

      /**
       * The range of years displayed in the year drop-down: either relative to 
       * today's year ("-nn:+nn"), relative to the currently selected year 
       * ("c-nn:c+nn"), absolute ("nnnn:nnnn"), or combinations of these formats 
       * ("nnnn:-nn").
       *
       * <p>See the <a href="#datePicker">date-picker</a> attribute for usage examples.
       *
       * @expose
       * @name datePicker.yearRange
       * @memberof! oj.ojInputDate
       * @instance
       * @type {string}
       * @default "c-10:c+10"
       */
      yearRange : "c-10:c+10" // Range of years to display in drop-down,
      // either relative to today's year (-nn:+nn), relative to currently displayed year
      // (c-nn:c+nn), absolute (nnnn:nnnn), or a combination of the above (nnnn:-n)

    },

    /**
     * A datetime converter instance that duck types {@link oj.DateTimeConverter}. Or an object literal
     * containing the properties listed below.
     *
     * The converter used for InputDate. Page authors can set a custom converter by creating one using the datetime converter factory
     * and providing custom options -
     * oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter(customOptions).
     *
     * {@ojinclude "name":"inputBaseConverterOptionDoc"}
     *
     * @property {string} type - the converter type registered with the oj.ConverterFactory.
     * Usually 'datetime'. See {@link oj.DateTimeConverterFactory} for details. <br/>
     * E.g., <code class="prettyprint">{converter: {type: 'datetime'}</code>
     * @property {Object=} options - optional Object literal of options that the converter expects.
     * See {@link oj.IntlDateTimeConverter} for options supported by the jet datetime converter.
     * E.g., <code class="prettyprint">{converter: {type: 'datetime', options: {formatType: 'date'}}</code>
     *
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputDate
     * @type {Object}
     * @ojsignature  { target: "Type", value: "oj.Converter<string>|oj.Validation.FactoryRegisteredValidatorOrConverter|null"}
     * @default oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({"day":"2-digit","month":"2-digit","year":"2-digit"})
     */
    converter : oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter(
    {
      "day" : "2-digit", "month" : "2-digit", "year" : "2-digit"
    }),

    /**
     * Determines if keyboard entry of the text is allowed.
     * When the datepicker is inline, the only supported value is "disabled".
     *
     * @expose
     * @instance
     * @memberof! oj.ojDatePicker
     * @name keyboardEdit
     * @type {string}
     * @ojvalue {string} "disabled" Changing the date can only be done with the picker.
     * @default "disabled"
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">keyboard-edit</code> attribute specified:</caption>
     * &lt;oj-date-picker keyboard-edit='disabled'>&lt;/oj-date-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">keyboardEdit</code> property after initialization:</caption>
     * // getter
     * var keyboardEdit = myInputDate.keyboardEdit;
     *
     * // setter
     * myInputDate.keyboardEdit = 'disabled';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeKeyboardEditOptionDefault: disabled !default;
     */
    /**
     * Determines if keyboard entry of the text is allowed.
     * When disabled the picker must be used to select a date.
     * 
     * Default value depends on the theme. In alta-android, alta-ios and alta-windows themes, the
     * default is <code class="prettyprint">"disabled"</code>
     * and it's <code class="prettyprint">"enabled"</code> for alta web theme.
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputDate
     * @type {string}
     * @ojvalue {string} "enabled"  Allow keyboard entry of the date.
     * @ojvalue {string} "disabled" Changing the date can only be done with the picker.
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">keyboard-edit</code> attribute specified:</caption>
     * &lt;oj-input-date keyboard-edit='disabled'>&lt;/oj-input-date>
     *
     * @example <caption>Get or set the <code class="prettyprint">keyboardEdit</code> property after initialization:</caption>
     * // getter
     * var keyboardEdit = myInputDate.keyboardEdit;
     *
     * // setter
     * myInputDate.keyboardEdit = 'disabled';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeKeyboardEditOptionDefault: disabled !default;
     */
    keyboardEdit : "enabled",

    /**
     * The maximum selectable date (in ISO string format). When set to null, there is no maximum.
     *
     * @expose
     * @instance
     * @memberof! oj.ojDatePicker
     * @name max
     * @type {string|null}
     * @default null
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">max</code> attribute specified:</caption>
     * &lt;oj-date-picker max='2018-09-25'>&lt;/oj-date-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">max</code> property after initialization:</caption>
     * // getter
     * var maxValue = myInputDate.max;
     *
     * // setter
     * myInputDate.max = '2018-09-25';
     */
    /**
     * The maximum selectable date (ISO string). When set to null, there is no maximum.
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputDate
     * @type {string|null}
     * @default null
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">max</code> attribute specified:</caption>
     * &lt;oj-input-date max='2018-09-25'>&lt;/oj-input-date>
     *
     * @example <caption>Get or set the <code class="prettyprint">max</code> property after initialization:</caption>
     * // getter
     * var maxValue = myInputDate.max;
     *
     * // setter
     * myInputDate.max = '2018-09-25';
     */
    max : undefined,

    /**
     * The minimum selectable date (ISO string format). When set to null, there is no minimum.
     *
     * @expose
     * @instance
     * @memberof! oj.ojDatePicker
     * @name min
     * @type {string|null}
     * @default null
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">min</code> attribute specified:</caption>
     * &lt;oj-date-picker min='2014-08-25'>&lt;/oj-date-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">min</code> property after initialization:</caption>
     * // getter
     * var minValue = myInputDate.min;
     *
     * // setter
     * myInputDate.min = '2014-08-25';
     */
    /**
     * The minimum selectable date (in ISO string format). When set to null, there is no minimum.
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputDate
     * @type {string|null}
     * @default null
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">min</code> attribute specified:</caption>
     * &lt;oj-input-date min='2014-08-25'>&lt;/oj-input-date>
     *
     * @example <caption>Get or set the <code class="prettyprint">min</code> property after initialization:</caption>
     * // getter
     * var minValue = myInputDate.min;
     *
     * // setter
     * myInputDate.min = '2014-08-25';
     */
    min : undefined,

    /**
     * <p>Attributes specified here will be set on the picker DOM element when it's launched.
     * <p>The supported attributes are <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended to the picker's class and style, if any.
     * Note: 1) pickerAttributes is not applied in the native theme.
     * 2) setting this property after element creation has no effect.
     * 
     * @property {string=} style
     * @property {string=} class
     *
     * @example <caption>Initialize the inputDate specifying a set of attributes to be set on the picker DOM element:</caption>
     * myInputDate.pickerAttributes = {
     *   "style": "color:blue;",
     *   "class": "my-class"
     * };
     *
     * @example <caption>Get the <code class="prettyprint">pickerAttributes</code> property, after initialization:</caption>
     * // getter
     * var inputDate = myInputDate.pickerAttributes;
     *
     * @expose
     * @memberof! oj.ojInputDate
     * @instance
     * @type {?Object}
     * @default null
     */
    pickerAttributes: null,

    /**
     * Allows applications to specify whether to render date picker in JET or
     * as a native picker control. In inline mode, the only value supported is "jet"</br>
     *
     * @expose
     * @memberof! oj.ojDatePicker
     * @instance
     * @name renderMode
     * @type {string}
     * @ojvalue {string} 'jet' Applications get full JET functionality.
     * @default "jet"
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">render-mode</code> attribute specified:</caption>
     * &lt;oj-date-picker render-mode='jet'>&lt;/oj-date-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">renderMode</code> property after initialization:</caption>
     * // getter
     * var renderMode = myInputDate.renderMode;
     *
     * // setter
     * myInputDate.renderMode = 'jet';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeRenderModeOptionDefault: native !default;
     */
    /**
     * Allows applications to specify whether to render date picker in JET or
     * as a native picker control.</br>
     * 
     * Default value depends on the theme. In alta-android, alta-ios and alta-windows themes, the
     * default is "native" and it's "jet" for alta web theme.
     *
     * @expose
     * @memberof! oj.ojInputDate
     * @instance
     * @type {string}
     * @ojvalue {string} 'jet' Applications get full JET functionality.
     * @ojvalue {string} 'native' Applications get the functionality of the native picker. Native picker is
     *  not available when the picker is inline, defaults to jet instead.</br></br>
     *  Note that the native picker support is limited to Cordova plugin published
     *  at 'https://github.com/VitaliiBlagodir/cordova-plugin-datepicker'.</br></br>
     *  With native renderMode, the functionality that is sacrificed compared to jet renderMode are:
     *    <ul>
     *      <li>Date picker cannot be themed</li>
     *      <li>Accessibility is limited to what the native picker supports</li>
     *      <li>pickerAttributes is not applied</li>
     *      <li>Sub-IDs are not available</li>
     *      <li>hide() function is no-op</li>
     *      <li>translations sub properties pertaining to the picker is not available</li>
     *      <li>All of the 'datepicker' sub-properties except 'showOn' are not available</li>
     *    </ul>
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">render-mode</code> attribute specified:</caption>
     * &lt;oj-input-date render-mode='native'>&lt;/oj-input-date>
     *
     * @example <caption>Get or set the <code class="prettyprint">renderMode</code> property after initialization:</caption>
     * // getter
     * var renderMode = myInputDate.renderMode;
     *
     * // setter
     * myInputDate.renderMode = 'native';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeRenderModeOptionDefault: native !default;
     */
    renderMode : "jet",

    /**
     * Additional info to be used when rendering the day
     *
     * This should be a JavaScript Function reference which accepts as its argument the following JSON format
     * {fullYear: Date.getFullYear(), month: Date.getMonth()+1, date: Date.getDate()}
     *
     * and returns null or all or partial JSON data of
     * {disabled: true|false, className: "additionalCSS", tooltip: 'Stuff to display'}
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputDate
     * @type {Function}
     * @ojsignature {target: "Type", value: "(param: oj.ojInputDate.DayFormatterInput)=> (null|'all'|oj.ojInputDate.DayFormatterOutput)"}
     * @default null
     */
    dayFormatter : null

    /**
     * Additional info to be used when rendering the day
     *
     * This should be in the following JSON format with the year, month, day based on Date.getFullYear(), Date.getMonth()+1, and Date.getDate():
     * {year: {month: {day: {disabled: true|false, className: "additionalCSS", tooltip: 'Stuff to display'}}}
     *
     * There also exists a special '*' character which represents ALL within that field [i.e. * within year, represents for ALL year].
     *
     * Note that this property will override the value of the dayFormatter property. Setting both dayFormatter and dayMetaData properties is not supported.
     *
     * @expose
     * @name dayMetaData
     * @instance
     * @memberof oj.ojInputDate
     * @type {object}
     * @default null
     * @ojsignature [{target: "Type", 
     *                value: "{[key:string]: {[key:string]: {[key:string]: {disabled?: boolean, className?: string, tooltip?: string}}}}"}
     *              ]
     * @example
     * {2013: {11: {25: {disabled: true, className: 'holiday', tooltip: 'Stuff to display'}, 5: {disabled: true}}}}}
     */

    // DOCLETS
    /**
     * List of validators used by element when performing validation. Each item is either an
     * instance that duck types {@link oj.Validator}, or is an Object literal containing the
     * properties listed below. Implicit validators created by an element when certain attributes
     * are present (e.g. <code class="prettyprint">required</code> property), are separate from
     * validators specified through this property. At runtime when the element runs validation, it
     * combines the implicit validators with the list specified through this property.
     * <p>
     * Hints exposed by validators are shown in the notewindow by default, or as determined by the
     * 'validatorHint' property set on the <code class="prettyprint">displayOptions</code>
     * property.
     * </p>
     *
     * <p>
     * When <code class="prettyprint">validators</code> property changes due to programmatic
     * intervention, the element may decide to clear messages and run validation, based on the
     * current state it is in. </br>
     *
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>The cached list of validator instances are cleared and new validator hints is pushed to
     * messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if element is valid when validators changes, element does nothing other than the
     * steps it always performs.</li>
     * <li>if element is invalid and is showing messages when
     * <code class="prettyprint">validators</code> changes then all element messages are cleared
     * and full validation run using the display value on the element.
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code>
     *   property is not updated and the error is shown.
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code>
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code>
     *   event to clear custom errors.</li>
     * </ul>
     * </li>
     * <li>if element is invalid and has deferred messages when validators changes, it does
     * nothing other than the steps it performs always.</li>
     * </ul>
     * </p>
     *
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the element are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * </p>
     *
     * @property {string} type - the validator type that has a {@link oj.ValidatorFactory} that can
     * be retrieved using the {@link oj.Validation} module. For a list of supported validators refer
     * to {@link oj.ValidatorFactory}. <br/>
     * @property {Object=} options - optional Object literal of options that the validator expects.
     *
     * @example <caption>Initialize the element with validator object literal:</caption>
     * myInputDate.validators = [{
     *     type: 'dateTimeRange',
     *     options : {
     *       max: '2014-09-10',
     *       min: '2014-09-01'
     *     }
     *   }];
     *
     * NOTE: oj.Validation.validatorFactory('dateTimeRange') returns the validator factory that is used
     * to instantiate a range validator for dateTime.
     *
     * @example <caption>Initialize the element with multiple validator instances:</caption>
     * var validator1 = new MyCustomValidator({'foo': 'A'});
     * var validator2 = new MyCustomValidator({'foo': 'B'});
     * // myInputElement is InputText, InputNumber, Select, etc.
     * myInputElement.value = 10;
     * myInputElement.validators = [validator1, validator2];
     *
     * @expose
     * @name validators
     * @instance
     * @memberof oj.ojInputDate
     * @ojsignature  { target: "Type", value: "Array<oj.Validator<string>|oj.Validation.FactoryRegisteredValidatorOrConverter>|null"}
     * @type {Array|undefined}
     */

    /**
     * The value of the DatePicker element which should be an ISOString.
     * 
     * When the attribute is not set, the element's value attribute is used as its initial value
     * if it exists. This value must be an ISOString.
     *
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> attribute:</caption>
     * &lt;oj-date-picker value='2014-09-10' /&gt;
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> property specified programmatically
     * using oj.IntlConverterUtils.dateToLocalIso :</caption>
     * myInputDate.value = oj.IntlConverterUtils.dateToLocalIso(new Date());<br/>
     * @example <caption>Get or set the <code class="prettyprint">value</code> property, after initialization:</caption>
     * // Getter: returns Today's date in ISOString
     * myInputDate.value;
     * // Setter: sets it to a different date
     * myInputDate.value = "2013-12-01";
     *
     * @expose
     * @name value
     * @instance
     * @memberof oj.ojDatePicker
     * @type {string}
     * @ojwriteback
     */
    /**
     * The value of the InputDate element which should be an ISOString.
     * 
     * When the attribute is not set, the element's value attribute is used as its initial value
     * if it exists. This value must be an ISOString.
     *
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> attribute:</caption>
     * &lt;oj-input-date value='2014-09-10' /&gt;
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> property specified programmatically
     * using oj.IntlConverterUtils.dateToLocalIso :</caption>
     * myInputDate.value = oj.IntlConverterUtils.dateToLocalIso(new Date());<br/>
     * @example <caption>Get or set the <code class="prettyprint">value</code> property, after initialization:</caption>
     * // Getter: returns Today's date in ISOString
     * myInputDate.value;
     * // Setter: sets it to a different date
     * myInputDate.value = "2013-12-01";
     *
     * @expose
     * @name value
     * @instance
     * @memberof! oj.ojInputDate
     * @type {string}
     * @ojwriteback
     */

    // Events

  },

  /**
   * @ignore
   * @protected
   * @memberof oj.ojInputDate
   */
  _InitBase : function ()
  {
    this._triggerNode = null;
    this._inputContainer = null;
    this._redirectFocusToInputContainer = false;
    this._isMobile = false;

    //only case is when of showOn of focus and one hides the element [need to avoid showing]
    this._ignoreShow = false;

    // need this flag to keep track of native picker opened, there is no callback on native API
    //  to find out otherwise.
    this._nativePickerShowing = false;
    this._maxRows = 4;

    this._currentDay = 0;
    this._drawMonth = this._currentMonth = 0;
    this._drawYear = this._currentYear = 0;

    this._datePickerDefaultValidators = {};
    this._nativePickerConverter = null;

    var nodeName = this.element[0].nodeName.toLowerCase();
    this._isInLine = (nodeName === "div" || nodeName === "span");
    this._dpDiv = bindHover($('<div class="oj-datepicker-popup" style="display:none"><div id="' + this._GetSubId(this._MAIN_DIV_ID) + '" role="region" aria-describedby="' + 
      this._GetSubId(this._DATEPICKER_DESCRIPTION_ID) + '" class="oj-datepicker-content"></div></div>'));
    $("body").append(this._dpDiv); //@HTMLUpdateOK

    if(this._isInLine)
    {
      //if inline then there is no input element, so reset _CLASS_NAMES
      // TODO:Jmw trying to understand what to do in the case of inline. If it is dateTime inline, then I don't wrap the date part.
      // But if it is just date inline, I should... but the use case is probably not frequent.
      this._WIDGET_CLASS_NAMES += this._INLINE_WIDGET_CLASS;
      this._CLASS_NAMES = "";
    }
    else
    {
      //append input container class to be applied to the root node as well, since not inline
      //[note the special case where input container class will NOT be on the widget node is when
      //ojInputDateTime is of inline and ojInputTime places container around the input element]
      // jmw. this is now different. It's no longer on the widget. I add a new wrapper dom.
      // Ji will need to help me with this probably.
      // One thing I know I'm not doing is wrapping the calendar if only date. hmm...
      this._ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES += this._INPUT_CONTAINER_CLASS;
     
      var self = this;
      //DISABLE FOR NOW, as animation is coming quite clunky (not sure if the css of popup or of animation)
      //var animation = _isLargeScreen ? {"open": null, "close": null} : {"close": null};
      var animation = {"open": null, "close": null};
      this._popUpDpDiv = this._dpDiv.ojPopup({"initialFocus": "none",
                                              "role": "dialog",
                                              "modality": _isLargeScreen ? "modeless" : "modal",
                                              "open": function () {
                                                self._popUpDpDiv.attr("aria-describedby", self._GetSubId(self._DATEPICKER_DIALOG_DESCRIPTION_ID));
                                                if (self.options["datePicker"]["showOn"] === "image")
                                                {
                                                  self._dpDiv.find(".oj-datepicker-calendar").focus();
                                                }
                                              },
                                              "animateStart": function (event, ui)
                                              {
                                                if ('open' === ui["action"])
                                                {
                                                  event.preventDefault();
                                                  oj.AnimationUtils.slideIn(ui.element, {"offsetY": ui.element.offsetHeight + "px"}).then(ui["endCallback"]);
                                                }
                                              },
                                              "animation": animation
                                            }).attr('data-oj-internal', ''); // mark internal component, used in oj.Components.getComponentElementByNode;
      this.element.attr('data-oj-popup-' + this._popUpDpDiv.attr('id') + '-parent', ''); // mark parent of pop up

      var pickerAttrs = this.options.pickerAttributes;
      if (pickerAttrs)
        oj.EditableValueUtils.setPickerAttributes(this._popUpDpDiv.ojPopup("widget"), pickerAttrs);

    }
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDate
   */
  _ComponentCreate : function ()
  {
    this._InitBase();

    var retVal = this._super();

    if(this.options["dayMetaData"])
    {
      this.options["dayFormatter"] = (function(value)
      {
        return function(dateInfo) {
          return _getMetaData(value, 0, [dateInfo["fullYear"], dateInfo["month"], dateInfo["date"]]);
        };
      })(this.options["dayMetaData"]);
    }

    //if isoString has a different timezone then the one provided in the converter, need to perform
    //conversion so pass it through the method
    if(this.options["value"])
    {
      var formatted = this._GetConverter()["format"](this.options["value"]);
      this._SetValue(formatted, {});
    }

    //Need to set the currentDay, currentMonth, currentYear to either the value or the default of today's Date
    //Note that these are days indicator for the datepicker, so it is correct in using today's date even if value
    //hasn't been set
    this._setCurrentDate(this._getDateIso());

    // jmw. Add a wrapper around the element and the trigger. This is needed so that we can
    // add inline messages to the root dom node. We want the input+trigger to be one child and
    // the inline messages to be another child of the root dom node. This way the inline
    // messages can be stacked after the main component, and will grow or shrink in size the same
    // as the main component.
    // doing this in InputBase now.

    if (this._isInLine)
    {
      this.element.append(this._dpDiv); //@HTMLUpdateOK dpDiv is generated internally
      this.element.addClass(this._INLINE_CLASS); //by applying the inline class it places margin bottom, to separate in case ojInputTime exists

      // Set display:block in place of inst._dpDiv.show() which won't work on disconnected elements
      // http://bugs.jqueryui.com/ticket/7552 - A Datepicker created on a detached div has zero height
      this._dpDiv.css("display", "block");
    }
    else
    {
      this._processReadOnlyKeyboardEdit();
      this._attachTrigger();
    }

    this._registerSwipeHandler();

    // attach active state change handlers
    bindActive(this);
    return retVal;
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDate
   */
  _AfterCreate : function ()
  {
    var ret = this._superApply(arguments);

    this._disableEnable(this.options["disabled"]);

    if (!this._IsCustomElement())
    {
      var label = this.$label;
      if(this._inputContainer && label && label.length === 1) {
        var LId = label.attr("id");

        if(!LId) {
          LId = this["uuid"] + "_Label";
          label.attr("id", LId);
        }

        this._inputContainer.attr("aria-labelledby", LId);
      }     
    }
    else
    {
      if(this._inputContainer)
      {
        var defaultLabelId = this["uuid"] + "_Label";
        var LId = oj.EditableValueUtils.getOjLabelId(this.widget(), defaultLabelId);
        if (LId)
          this._inputContainer.attr("aria-labelledby", LId);

      }
      
    }


    return ret;
  },

  /**
   * @ignore
   * @protected
   * @override
   * @memberof oj.ojInputDate
   */
  _setOption : function (key, value, flags)
  {

    var retVal = null;

    //When a null, undefined, or "" value is passed in set to null for consistency
    //note that if they pass in 0 it will also set it null
    if (key === "value")
    {
      if(!value)
      {
        value = null;
      }

      retVal = this._super(key, value, flags);
      this._setCurrentDate(value);

      if(this._datepickerShowing())
      {
        // _setOption is called after user picks a date from picker, we dont want to bring
        //  focus to input element if the picker is showing still for the non-inline case. For the
        //  case of inline date picker, if there is a time field and already focussed (brought in when
        //  the picker was hidden), we want to update the date picker, but not set focus on it.
        var focusOnCalendar = !(this._isInLine && this._timePicker && this._timePicker[0] === document.activeElement);
        this._updateDatepicker(focusOnCalendar);
      }

      return retVal;
    }

    if (key === "dayMetaData")
    {
      //need to invoke w/ dayFormatter and return for the case where user invoke myInputDate.dayMetaData = {};
      //since that doesn't trigger ComponentBinding

      this.option("dayFormatter", function(dateInfo) {
          return _getMetaData(value, 0, [dateInfo["fullYear"], dateInfo["month"], dateInfo["date"]]);
      }, flags);
      return; //avoid setting in this.options and etc
    }

    retVal = this._super(key, value, flags);

    if (key === "disabled")
    {
      this._disableEnable(value);
    }
    else if (key === "max" || key === "min")
    {
      //since validators are immutable, they will contain min + max as local values. B/c of this will need to recreate
      this._datePickerDefaultValidators[oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE] = this._getValidator("min");

      this._AfterSetOptionValidators();
    }
    else if(key === "readOnly")
    {
      this._processReadOnlyKeyboardEdit();

      if (value)
      {
        this._hide(this._ON_CLOSE_REASON_CLOSE);
      }
      this._AfterSetOptionDisabledReadOnly("readOnly", oj.EditableValueUtils.readOnlyOptionOptions);
    }
    else if(key === "keyboardEdit")
    {
      this._processReadOnlyKeyboardEdit();
    }
    else if (key === "dayFormatter")
    {
      //since validators are immutable, they will contain dayFormatter as local values. B/c of this will need to recreate
      this._datePickerDefaultValidators[oj.ValidatorFactory.VALIDATOR_TYPE_DATERESTRICTION] = this._getValidator("dayFormatter");

      this._AfterSetOptionValidators();
    }
    else if(key === "converter")
    {
      this._nativePickerConverter = null;
    }

    if (key === "datePicker" && flags["subkey"] === "currentMonthPos")
    {
      //need to reset up the drawMonth + drawYear
      this._setCurrentDate(this._getDateIso());
    }

    var updateDatePicker = {"max": true, "min": true, "dayFormatter": true, "datePicker": true, "translations": true};

    if(this._datepickerShowing() && key in updateDatePicker)
    {
      this._updateDatepicker();
    }

    return retVal;
  },

  /**
   * @ignore
   */
  _processReadOnlyKeyboardEdit: function()
  {
    var readonly = this.options["readOnly"] ||
            this._isKeyboardEditDisabled();
    this.element.prop("readOnly", !!readonly);
  },

  /**
   * @ignore
   * @return {boolean}
   */
  _isKeyboardEditDisabled: function()
  {
    return this.options["keyboardEdit"] === this._KEYBOARD_EDIT_OPTION_DISABLED;
  },

  /**
   * Need to override due to usage of display: inline-table [as otherwise for webkit the hidden content takes up
   * descent amount of space]
   *
   * @protected
   * @instance
   * @memberof oj.ojInputDate
   */
  _AppendInputHelperParent : function ()
  {
    return this._triggerNode;
  },

  /**
   * @ignore
   * @protected
   * @override
   */
  _destroy : function ()
  {
    this._cleanUpDateResources();
    var retVal = this._super();
    return retVal;
  },

  _datepickerShowing: function()
  {
    return this._isInLine || (oj.Components.isComponentInitialized(this._popUpDpDiv, "ojPopup") && this._popUpDpDiv.ojPopup("isOpen")) || this._nativePickerShowing;
  },
  /**
   * @ignore
   * @protected
   * @override
   */
  _SetupResources : function ()
  {
    if (!this._isInLine)
    {
      this._resizePopupBind = function ()
      {
        if (oj.Components.isComponentInitialized(this._popUpDpDiv, "ojPopup"))
        {
          this._popUpDpDiv.ojPopup("option", "modality", (_isLargeScreen ? "modeless" : "modal"));
        }
      }.bind(this);
      window.addEventListener('resize', this._resizePopupBind, false);
    }
    return this._super();
  },
  /**
   * @ignore
   * @protected
   * @override
   */
  _ReleaseResources : function ()
  {
    //for non custom element, this is done as first step in destroy
    if (this._IsCustomElement())
      this._cleanUpListeners();
    return this._super();
  },
  /**
   * @ignore
   * @private
   */
  _cleanUpDateResources : function ()
  {
    var triggerRootContainer;
    if (this._isInLine)
    {
      triggerRootContainer = $(this.element[0]).parent();
    }
    else 
    {
      triggerRootContainer = $(this.element[0]).parent().parent();
    }
    this._RemoveActiveable(triggerRootContainer);

    this.element.off("focus touchstart");
    this._wrapper.off("touchstart");

    if (this._triggerNode)
    {
      this._triggerNode.remove();
    }

    if (this._isInLine)
    {
      //need to remove disabled + readOnly b/c they are set by super classes and datepicker is special in that this.element
      //can be a div element for inline mode
      this.element.removeProp("disabled");
      this.element.removeProp("readonly");
    }

    this._cleanUpListeners();

    if (this._animationResolve)
    {
      this._animationResolve();
      this._animationResolve = null;
    }
    if (this._popUpDpDiv && oj.Components.isComponentInitialized(this._popUpDpDiv, "ojPopup"))
     this._popUpDpDiv.ojPopup("destroy");
    this._dpDiv.remove();
  },
  /**
   * @ignore
   * @private
   */
  _cleanUpListeners : function ()
  {
    if (this._resizePopupBind)
    {
      window.removeEventListener('resize', this._resizePopupBind);
    }
    window.removeEventListener('resize', resizeLargeScreenChange);
  },
  /**
   * @protected
   * @override
   * @ignore
   */
  _WrapElement: function()
  {
    this._inputContainer = this._superApply(arguments);
    this._inputContainer.attr({"role": "combobox", "aria-haspopup": "true", "tabindex": "-1"});
  },

  /**
   * When input element has focus
   * @private
   */
  _onElementFocus : function()
  {
    var showOn = this.options["datePicker"]["showOn"];

    if(this._redirectFocusToInputContainer)
    {
      this._redirectFocusToInputContainer = false;
      this._inputContainer.focus();
    }
    else
    {
      if (showOn === "focus")
      {
        // pop-up date picker when focus placed on the input box
        this.show();
      }
      else
      {
        if(this._datepickerShowing())
        {
          this._hide(this._ON_CLOSE_REASON_CLOSE);
        }
      }
    }
  },

  /**
   * When input element is touched
   *
   * @ignore
   * @protected
   */
  _OnElementTouchStart : function()
  {
    var showOn = this.options["datePicker"]["showOn"];

    // If the focus is already on the text box and can't edit with keyboard
    // and show on is focus then reopen the picker.
    if(showOn === "focus")
    {
      if (this._datepickerShowing())
      {
        this._ignoreShow = true;
        this._hide(this._ON_CLOSE_REASON_CLOSE);
      }
      else
      {
        var inputActive = this.element[0] === document.activeElement;

        this.show();
        this._redirectFocusToInputContainer = true;

        if(inputActive)
        {
          this._inputContainer.focus();
        }
      }
    }
  },

  /**
   * This function will create the necessary calendar trigger container [i.e. image to launch the calendar]
   * and perform any attachment to events
   *
   * @private
   */
  _attachTrigger : function()
  {
    var showOn = this.options["datePicker"]["showOn"];
    var triggerContainer = $("<span>").addClass(this._TRIGGER_CLASS);

    // pop-up date picker when button clicked
    var triggerCalendar =
      $("<span title='" + this._GetCalendarTitle() + "'/>").addClass(this._TRIGGER_CALENDAR_CLASS + " oj-clickable-icon-nocontext oj-component-icon");

    triggerContainer.append(triggerCalendar); //@HTMLUpdateOK

    this.element.on("focus", $.proxy(this._onElementFocus, this));
    this.element.on("touchstart", $.proxy(this._OnElementTouchStart, this));

    var self = this;

    this._wrapper.on("touchstart", function(e)
    {
      self._isMobile = true;
    });

    if (showOn === "image")
    {
      // we need to show the icon that we hid by display:none in the mobile themes
      triggerCalendar.css("display", "block");

      // In iOS theme, we defaulted to use border radius given that showOn=focus is default and
      //  we will not have trigger icon. For showOn=image case, we will show the icon, so
      //  we need to remove the border radius. iOS is the only case we use border radius, so this
      //  setting for all cases is fine.
      if (this._IsRTL())
      {
        this.element.css("border-top-left-radius", 0);
        this.element.css("border-bottom-left-radius", 0);
      }
      else
      {
        this.element.css("border-top-right-radius", 0);
        this.element.css("border-bottom-right-radius", 0);
      }
    }

    triggerCalendar.on("click", function ()
    {
      if (self._datepickerShowing())
      {
        self._hide(self._ON_CLOSE_REASON_CLOSE);
      }
      else
      {
        self.show();
        self._dpDiv.find(".oj-datepicker-calendar").focus();
      }
      return false;
    });

    this._AddHoverable(triggerCalendar);
    this._AddActiveable(triggerCalendar);

    this._triggerIcon = triggerCalendar;
    this._triggerNode = triggerContainer;
    this.element.after(triggerContainer); //@HTMLUpdateOK
  },

  //This handler is when an user keys down with the calendar having focus
  _doCalendarKeyDown : function (event)
  {
    var sel, handled = false,
        kc = $.ui.keyCode,
        isRTL = this._IsRTL();

    if (this._datepickerShowing())
    {
      switch (event.keyCode)
      {
        case 84: //t character
          if (event.altKey && event.ctrlKey)
          {
            this._dpDiv.find(".oj-datepicker-current").focus();
            handled = true;
          }
          break;
        case kc.TAB:
          // Tab key is used to navigate to different buttons/links in the
          // datepicker to make them accessible.  It shouldn't be used to hide
          // the datepicker.
          break;
        case kc.SPACE:
        case kc.ENTER:
          sel = $("td." + this._DAYOVER_CLASS, this._dpDiv);
          if (sel[0])
          {
            this._selectDay(this._currentMonth, this._currentYear, sel[0], event);
          }
          //need to return false so preventing default + stop propagation here
          event.preventDefault();
          event.stopPropagation();
          return false;
        case kc.ESCAPE:
          this._hide(this._ON_CLOSE_REASON_CANCELLED);
          handled = true;
          break;// hide on escape
        case kc.PAGE_UP:
          if(event.ctrlKey && event.altKey)
          {
            this._adjustDate(- this.options["datePicker"]["stepBigMonths"], "M", true);
          }
          else if (event.altKey)
          {
            this._adjustDate( - 1, "Y", true);
          }
          else
          {
            this._adjustDate(- this._getStepMonths(), "M", true);
          }
          handled = true;
          break;// previous month/year on page up/+ ctrl
        case kc.PAGE_DOWN:
          if(event.ctrlKey && event.altKey)
          {
            this._adjustDate(+ this.options["datePicker"]["stepBigMonths"], "M", true);
          }
          else if (event.altKey)
          {
            this._adjustDate(1, "Y", true);
          }
          else
          {
            this._adjustDate(+ this._getStepMonths(), "M", true);
          }
          handled = true;
          break;// next month/year on page down/+ ctrl
        case kc.END:
          this._currentDay = this._getDaysInMonth(this._currentYear, this._currentMonth);
          this._changeCurrentDay();

          handled = true;
          break;
        case kc.HOME:
          this._currentDay = 1;
          this._changeCurrentDay();

          handled = true;
          break;
        case kc.LEFT:
          
          // next month/year on alt +left on Mac
          if ((event.originalEvent && event.originalEvent.altKey) || event.altKey)
          {
            this._adjustDate((event.ctrlKey ?  - this.options["datePicker"]["stepBigMonths"] :  - this._getStepMonths()), "M", true);
          }
          else 
          {
            this._adjustDate((isRTL ?  + 1 :  - 1), "D", true);
            // -1 day on ctrl or command +left
          }
          
          handled = true;
          break;
        case kc.UP:
          this._adjustDate( - 7, "D", true);
          handled = true;
          break;// -1 week on ctrl or command +up
        case kc.RIGHT:
          
          // next month/year on alt +right
          if ((event.originalEvent && event.originalEvent.altKey) || event.altKey)
          {
            this._adjustDate((event.ctrlKey ?  + this.options["datePicker"]["stepBigMonths"] :  + this._getStepMonths()), "M", true);
          }
          else 
          {
            this._adjustDate((isRTL ?  - 1 :  + 1), "D", true);
            // +1 day on ctrl or command +right
          }
          
          handled = true;
          break;
        case kc.DOWN:
          this._adjustDate( + 7, "D", true);
          handled = true;
          break;// +1 week on ctrl or command +down
        default : ;
      }
    }
    else if (event.keyCode === kc.HOME && event.ctrlKey)
    {
      // display the date picker on ctrl+home
      this.show();
      handled = true;
    }

    if (handled)
    {
      event.preventDefault();
      event.stopPropagation();
    }

  },

  _changeCurrentDay : function() 
  {
    var cOver = $("." + this._DAYOVER_CLASS, this._dpDiv);
    var cDay = this._currentDay + '';

    if(cOver.length === 1)
    {
      cOver.removeClass(this._DAYOVER_CLASS);
    }

    var datePickerCalendar = $('table.oj-datepicker-calendar', this._dpDiv);
    cOver = $('a.oj-enabled:contains(' + this._currentDay + ')', datePickerCalendar)
              .filter(function() { 
                return $(this).text() === cDay; 
              });
    if(cOver.length === 1) {
      var cParent = cOver.parent();
      datePickerCalendar.addClass('oj-focus-highlight');
      datePickerCalendar.attr('aria-activedescendant', cParent.attr("id") + "");
      cParent.addClass(this._DAYOVER_CLASS);
    }
  },

  //This handler is when an user keys down with the Month View having focus
  //TODO, during month/year work apparently CalendarKeyDown was copied + pasted for month/year. try to clean up code
  _doMonthViewKeyDown : function(event)
  {
    var sel, handled = false,
        kc = $.ui.keyCode,
        isRTL = this._IsRTL();

    if (this._datepickerShowing())
    {
      switch (event.keyCode)
      {
        case 84: //t character
          if (event.altKey && event.ctrlKey)
          {
            this._dpDiv.find(".oj-datepicker-current").focus();
            handled = true;
          }
          break;
        case kc.SPACE:
        case kc.ENTER:
          sel = $("td." + this._DAYOVER_CLASS, this._dpDiv);
          if (sel[0])
          {
            this._selectMonthYear(sel[0], 'M');
          }
          //need to return false so preventing default + stop propagation here
          event.preventDefault();
          event.stopPropagation();
          return false;
        case kc.ESCAPE:
          this.hide();
          handled = true;
          break;// hide on escape
        case kc.PAGE_UP:
          if(event.ctrlKey && event.altKey)
          {
            this._adjustDate(- this.options["datePicker"]["stepBigMonths"], "M", true, 'month');
          }
          else if (event.altKey)
          {
            this._adjustDate( - 1, "Y", true, 'month');
          }
          else
          {
            this._adjustDate(- this._getStepMonths(), "M", true, 'month');
          }
          handled = true;
          break;// previous month/year on page up/+ ctrl
        case kc.PAGE_DOWN:
          if(event.ctrlKey && event.altKey)
          {
            this._adjustDate(+ this.options["datePicker"]["stepBigMonths"], "M", true, 'month');
          }
          else if (event.altKey)
          {
            this._adjustDate(1, "Y", true, 'month');
          }
          else
          {
            this._adjustDate(+ this._getStepMonths(), "M", true, 'month');
          }
          handled = true;
          break;// next month/year on page down/+ ctrl
        case kc.END:
          this._currentMonth = 11;
          this._updateDatepicker(true, 'month');
          handled = true;
          break;
        case kc.HOME:
          this._currentMonth = 0;
          this._updateDatepicker(true, 'month');
          handled = true;
          break;
        case kc.LEFT:
          this._adjustDate((isRTL ?  + 1 :  - 1), "M", true, 'month');
          handled = true;
          break;
        case kc.UP:
          this._adjustDate( - 3, "M", true, 'month');
          handled = true;
          break;// -1 week on ctrl or command +up
        case kc.RIGHT:
          this._adjustDate((isRTL ?  - 1 :  + 1), "M", true, 'month');
          handled = true;
          break;
        case kc.DOWN:
          this._adjustDate( + 3, "M", true, 'month');
          handled = true;
          break;// +1 week on ctrl or command +down
        default : ;
      }
    }
    else if (event.keyCode === kc.HOME && event.ctrlKey)
    {
      // display the date picker on ctrl+home
      this.show();
      handled = true;
    }

    if (handled)
    {
      event.preventDefault();
      event.stopPropagation();
    }

  },

  //This handler is when an user keys down with the Year View having focus
  _doYearViewKeyDown : function(event)
  {
    var sel, handled = false,
        kc = $.ui.keyCode,
        isRTL = this._IsRTL();

    if (this._datepickerShowing())
    {
      switch (event.keyCode)
      {
        case 84: //t character
          if (event.altKey && event.ctrlKey)
          {
            this._dpDiv.find(".oj-datepicker-current").focus();
            handled = true;
          }
          break;
        case kc.SPACE:
        case kc.ENTER:
          sel = $("td." + this._DAYOVER_CLASS, this._dpDiv);
          if (sel[0])
          {
            this._selectMonthYear(sel[0], 'Y');
          }
          //need to return false so preventing default + stop propagation here
          event.preventDefault();
          event.stopPropagation();
          return false;
        case kc.ESCAPE:
          this.hide();
          handled = true;
          break;// hide on escape
        case kc.PAGE_UP:
          if (event.altKey)
          {
            this._adjustDate( - 1, "Y", true, 'year');
          }
          handled = true;
          break;// previous month/year on page up/+ ctrl
        case kc.PAGE_DOWN:
          if (event.altKey)
          {
            this._adjustDate(1, "Y", true, 'year');
          }
          handled = true;
          break;// next month/year on page down/+ ctrl
        case kc.END:
          this._currentYear = Math.floor(this._currentYear / 10) * 10 + 9;
          this._updateDatepicker(true, 'year');
          handled = true;
          break;
        case kc.HOME:
          this._currentYear = Math.floor(this._currentYear / 10) * 10;
          this._updateDatepicker(true, 'year');
          handled = true;
          break;
        case kc.LEFT:
          this._adjustDate((isRTL ?  + 1 :  - 1), "Y", true, 'year');
          handled = true;
          break;
        case kc.UP:
          this._adjustDate( - 3, "Y", true, 'year');
          handled = true;
          break;// -1 week on ctrl or command +up
        case kc.RIGHT:
          this._adjustDate((isRTL ?  - 1 :  + 1), "Y", true, 'year');
          handled = true;
          break;
        case kc.DOWN:
          this._adjustDate( + 3, "Y", true, 'year');
          handled = true;
          break;// +1 week on ctrl or command +down
        default : ;
      }
    }
    else if (event.keyCode === kc.HOME && event.ctrlKey)
    {
      // display the date picker on ctrl+home
      this.show();
      handled = true;
    }

    if (handled)
    {
      event.preventDefault();
      event.stopPropagation();
    }

  },

  /**
   * @returns {jQuery} returns the content element of the datepicker
   * 
   * @private
   */
  _getDatepickerContent: function ()
  {
    return $(this._dpDiv.find(".oj-datepicker-content")[0]);
  },

  /**
   * Function to whether it is a datetimepicker with the switcher
   * 
   * @private
   */
  _isDateTimeSwitcher: function()
  {
    return this._dateTimeSwitcherActive;
  },

  /**
   * Thie function will update the calendar display
   *
   * @private
   * @param {boolean=} focusOnCalendar - Whether to put focus in the calendar.
   * @param {string=} view - The view to update to. Default is 'day'.
   * @param {string=} navigation - Type of navigation to animate.
   */
  _updateDatepicker : function (focusOnCalendar, view, navigation)
  {
    this._maxRows = 4;//Reset the max number of rows being displayed (see #7043)
    var generatedHtmlContent;

    if (view === 'year')
    {
      generatedHtmlContent = this._generateViewHTML('Y');
    }
    else if (view === 'month')
    {
      generatedHtmlContent = this._generateViewHTML('M');
    }
    else
    {
      generatedHtmlContent = this._generateViewHTML('D');
    }

    generatedHtmlContent.html = "<div" + (this._isDateTimeSwitcher() ? "" : " class='oj-datepicker-wrapper'") +">" + generatedHtmlContent.html + "</div>";

    this._currentView = view;
    var dpContentDiv = this._getDatepickerContent();

    if (navigation)
    {
      var oldChild = dpContentDiv.children().first();
      oldChild.css({position: 'absolute', left: 0, top: 0});

      dpContentDiv.prepend(generatedHtmlContent.html); //@HTMLUpdateOK
      var newChild = dpContentDiv.children().first();
      var direction = (navigation == 'previous') ? 'end' : 'start';

      if(!this._animationResolve) 
      {
        var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
        this._animationResolve = busyContext.addBusyState({"description" : "The datepicker id='" + 
          this.element.attr('id') + "' is animating."});
      }
      
      oj.AnimationUtils.startAnimation(newChild[0], 'open', {'effect':'slideIn', 'direction':direction});
      var promise = oj.AnimationUtils.startAnimation(oldChild[0], 'close', {'effect':'slideOut', 'direction':direction, 'persist':'all'});
      var self = this;
      promise.then(function() {
        if (oldChild)
        {
          oldChild.remove();
        }

        self._setupNewView(focusOnCalendar, view, generatedHtmlContent.dayOverId);
        self._animationResolve();
        self._animationResolve = null;
      });
    }
    else
    {
      dpContentDiv.empty().append(generatedHtmlContent.html); //@HTMLUpdateOK
      this._setupNewView(focusOnCalendar, view, generatedHtmlContent.dayOverId);
    }
  },

  _setupNewView : function(focusOnCalendar, view, dayOverId)
  {
    this._attachHandlers();

    if (dayOverId)
    {
      this._dpDiv.find(".oj-datepicker-calendar").attr("aria-activedescendant", dayOverId);
    }

    var numMonths = this._getNumberOfMonths(),
        cols = numMonths[1],
        width = 275;

    this._dpDiv.removeClass("oj-datepicker-multi-2 oj-datepicker-multi-3 oj-datepicker-multi-4").width("");
    if (view === 'year' || view === 'month')
    {
      this._dpDiv.removeClass("oj-datepicker-multi");
    }
    else
    {
      numMonths = this._getNumberOfMonths(),
          cols = numMonths[1];

      if (cols > 1)
      {
        // Try to determine the width dynamically
        var calendar = this._dpDiv.find(".oj-datepicker-calendar");
        var daySelectors = calendar.find("tbody a");
        var cellWidth = parseFloat(daySelectors.css("width"));
        var padding = parseFloat(calendar.css("margin-left"));
        if (!isNaN(cellWidth) && !isNaN(padding))
        {
          width = cellWidth * (this.options["datePicker"]["weekDisplay"] === "number" ? 8 : 7) + (padding * 2);
        }

        this._dpDiv.addClass("oj-datepicker-multi-" + cols).css("width", (width * cols + (this._isInLine ? 2 : 0)) + "px");
      }
      this._dpDiv[(numMonths[0] !== 1 || numMonths[1] !== 1 ? "add" : "remove") + "Class"]("oj-datepicker-multi");
    }

    // #6694 - don't focus the input if it's already focused
    // this breaks the change event in IE
    if (this._datepickerShowing() && this.element.is(":visible") && !this.element.is(":disabled"))
    {
      if (!focusOnCalendar)
      {
        if (!this._isInLine && this.element[0] !== document.activeElement)
        {
          this.element.focus();
        }
      }
      else
      {
        this._placeFocusOnCalendar();
      }
    }

  },

  _placeFocusOnCalendar: function() 
  {
    var calendar = this._dpDiv.find(".oj-datepicker-calendar");
    if (calendar[0] !== document.activeElement)
    {
      $(calendar[0]).focus();
    }
  },

  /**
   * Adjust one of the date sub-fields.
   *
   * @private
   * @param {number} offset
   * @param {string} period
   * @param {boolean=} focusOnCalendar - Whether to put focus in the calendar.
   * @param {string=} view - The view to update to. Default is 'day'.
   * @param {string=} navigation - Type of navigation to animate.
   */
  _adjustDate : function (offset, period, focusOnCalendar, view, navigation)
  {
    if (this.options["disabled"])
    {
      return;
    }

    var currMonth = this._currentMonth;
    var currYear = this._currentYear;
    this._adjustInstDate(offset + (period === "M" ? this.options["datePicker"]["currentMonthPos"] : 0), // undo positioning
                          period);

    if(period === 'D' && currMonth === this._currentMonth && currYear === this._currentYear) 
    {
      //just day update so change day over class
      this._changeCurrentDay();
    }
    else
    {
      this._updateDatepicker(focusOnCalendar, view, navigation);
    }
    
  },

  /**
   * Action for current link. Note that this is of today relative to client's locale so this is ok.
   *
   * @private
   */
  _gotoToday : function ()
  {
    var date = new Date();

    this._currentDay = date.getDate();
    this._drawMonth = this._currentMonth = date.getMonth();
    this._drawYear = this._currentYear = date.getFullYear();

    this._adjustDate(null, null, true, 'day');
  },

  /**
   * Action for selecting a new month/year.
   *
   * @private
   * @param {Object} select
   * @param {string} period
   */
  _selectMonthYear : function (select, period)
  {
    var selected;
    var converterUtils = oj.IntlConverterUtils;
    var value = this._getDateIso();
    var yearAttr = select.getAttribute("data-year");

    if (yearAttr)
    {
      selected = parseInt(yearAttr, 10);
      this._currentYear = this._drawYear = selected;
    }

    if (period === "M")
    {
      selected = parseInt(select.getAttribute("data-month"), 10);
      this._currentMonth = this._drawMonth = selected;
      value = converterUtils._dateTime(value, {"fullYear": this._currentYear, "month": this._currentMonth});
    }
    else
    {
      value = converterUtils._dateTime(value, {"fullYear": this._currentYear});
    }

    //Take care of accessibility. Note that this is using an INTERNAL converter to display only the year portion [no timezone]
    //so is okay
    $("#" + this._GetSubId(this._CALENDAR_DESCRIPTION_ID)).html(this._EscapeXSS(this.options["monthWide"][this._drawMonth]) + " " + //@HTMLUpdateOK
      yearDisplay.format(oj.IntlConverterUtils.dateToLocalIso(new Date(this._drawYear, this._drawMonth, 1)))); //@HTMLUpdateOK

    this._adjustDate(0, 0, true, period === "M" ? 'day' : this._toYearFromView);
  },

  //Action for selecting a day.
  _selectDay : function (month, year, td, event)
  {
    if ($(td).hasClass(this._UNSELECTABLE_CLASS) || this.options["disabled"])
    {
      return;
    }

    if(!this._isDateTimeSwitcher()) 
    {
      this._hide(this._ON_CLOSE_REASON_SELECTION);
    }
    
    this._currentDay = $("a", td).html(); //@HtmlUpdateOk
    this._currentMonth = month;
    this._currentYear = year;

    var converterUtils = oj.IntlConverterUtils,
        value = this.options['value'],
        tempDate = new Date(this._currentYear, this._currentMonth, this._currentDay);

    if (value)
    {

      //need to preserve the time portion when of ojInputDateTime, so update only year, month, and date
      value = converterUtils._dateTime(value, {"fullYear": tempDate.getFullYear(), "month": tempDate.getMonth(),
                "date": tempDate.getDate()});
    }
    else
    {
      //per discussion when date doesn't exist use local isostring
      value = converterUtils.dateToLocalIso(tempDate);
    }

    this._setDisplayAndValue(value, {});

    if(this._isDateTimeSwitcher()) 
    {
      this._placeFocusOnCalendar();
    }
  },

  _setDisplayAndValue: function(isoString, event)
  {
    if(!this._isDateTimeSwitcher()) 
    {
      var formatted = this._GetConverter()["format"](isoString);
      this._SetDisplayValue( formatted ); //need to set the display value, since _SetValue doesn't trigger it per discussion
                                          //need to use formatted value as otherwise it doesn't go through framework's cycle
                                          //in updates
      this._SetValue(formatted, event); //TEMP TILL FIXED PASS IN formatted
    }
    else
    {
      this._switcherDateValue = isoString;
      this._setCurrentDate(isoString);

      if(this._datepickerShowing())
      {
        var focusOnCalendar = !(this._isInLine && this._timePicker && this._timePicker[0] === document.activeElement);
        this._updateDatepicker(focusOnCalendar);
      }
    }
    
  },

  /**
   * Get the default isostring date
   *
   * @ignore
   * @private
   */
  _getDefaultIsoDate: function()
  {
    return oj.IntlConverterUtils.dateToLocalIso(this._getTodayDate());
  },

  /**
   * Updates the internal current + draw values
   *
   * @private
   * @param {string} isoDate
   */
  _setCurrentDate : function (isoDate)
  {
    var newDate = oj.IntlConverterUtils._dateTime(isoDate || this._getDefaultIsoDate(), ["fullYear", "month", "date"],
                                              true);

    this._currentDay = newDate["date"];
    this._currentMonth = newDate["month"];
    if(!this._isMultiMonth())
    {
      //request not to change month
      this._drawMonth = newDate["month"];
    }
    
    this._drawYear = this._currentYear = newDate["fullYear"];

    this._adjustInstDate();
  },

  _getStepMonths : function ()
  {
    var stepMonths = this.options["datePicker"]["stepMonths"];
    return $.isNumeric(stepMonths) ? stepMonths : this.options["datePicker"]["numberOfMonths"];
  },

  // Check if an event is a button activation event
  _isButtonActivated : function(evt)
  {
    // We are using <a role='button'> for the buttons.  They fire click event
    // on Enter keydown.  We just need to check for Space/Enter key here.

    return(!this.options["disabled"] &&
           ((evt.type === 'click') ||
            (evt.type === 'keydown' && (evt.keyCode === 32 || evt.keyCode === 13))));
  },

  _gotoPrev : function(stepMonths)
  {
    if (this._currentView === 'year')
    {
      this._adjustDate(-10, "Y", true, 'year', 'previous');
    }
    else if (this._currentView === 'month')
    {
      this._adjustDate(-1, "Y", true, 'month', 'previous');
    }
    else
    {
      this._adjustDate( - stepMonths, "M", true, 'day', 'previous');
    }
  },

  _gotoNext: function(stepMonths)
  {
    if (this._currentView === 'year')
    {
      this._adjustDate(+10, "Y", true, 'year', 'next');
    }
    else if (this._currentView === 'month')
    {
      this._adjustDate(+1, "Y", true, 'month', 'next');
    }
    else
    {
      this._adjustDate( + stepMonths, "M", true, 'day', 'next');
    }
  },

  /**
   * Attach the onxxx handlers.  These are declared statically so
   * they work with static code transformers like Caja.
   *
   * @private
   */
  _attachHandlers : function ()
  {
    var stepMonths = this._getStepMonths(), self = this;
    var keyDownPrevent = function(evt) {
      var allowDefaultEvent = false;
      if(evt.type === "keydown" && evt.keyCode === $.ui.keyCode.TAB) 
      {
        allowDefaultEvent = true;
      }

      return allowDefaultEvent;
    }

    this._dpDiv.find("[data-handler]").map(function ()
    {
      var handler =
      {
        /** @expose */
        prev : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            self._gotoPrev(stepMonths);
          }
          return keyDownPrevent(evt);
        },
        /** @expose */
        next : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            self._gotoNext(stepMonths);
          }
          return keyDownPrevent(evt);
        },
        /** @expose */
        today : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            self._gotoToday();
          }
          return keyDownPrevent(evt);
        },
        /** @expose */
        selectDay : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            self._selectDay( + this.getAttribute("data-month"),  + this.getAttribute("data-year"), this, evt);
          }
          return keyDownPrevent(evt);
        },
        /** @expose */
        selectMonth : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            self._selectMonthYear(this, "M");
          }
          return keyDownPrevent(evt);
        },
        /** @expose */
        selectYear : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            self._selectMonthYear(this, "Y");
          }
          return keyDownPrevent(evt);
        },
        /** @expose */
        calendarKey : function (evt)
        {
          if (self._currentView === 'year')
          {
            self._doYearViewKeyDown(evt);
          }
          else if (self._currentView === 'month')
          {
            self._doMonthViewKeyDown(evt);
          }
          else
          {
            self._doCalendarKeyDown(evt);
          }
        },
        /** @expose */
        selectMonthHeader : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            if (self._currentView === 'month')
            {
              self._updateDatepicker(true, 'day');
            }
            else
            {
              self._updateDatepicker(true, 'month');
            }
          }
          return keyDownPrevent(evt);
        },
        /** @expose */
        selectYearHeader : function (evt)
        {
          if (self._isButtonActivated(evt))
          {
            if (self._currentView === 'year')
            {
              self._updateDatepicker(true, 'day');
            }
            else
            {
              // Remember where we are navigating to the Year view from
              self._toYearFromView = self._currentView;
              self._updateDatepicker(true, 'year');
            }
          }
          return keyDownPrevent(evt);
        }
      };
      $(this).bind(this.getAttribute("data-event"), handler[this.getAttribute("data-handler")]);
    });

    // Only show the day focus if the user starts using keyboard
    this._dpDiv.find(".oj-datepicker-calendar").map(function ()
    {
      oj.DomUtils.makeFocusable({
        'element': $(this),
        'applyHighlight': true
      });
    });

    // Avoid problem with hover/active state on header/footer not going away on touch devices
    var buttons = this._dpDiv.find(".oj-datepicker-header a, .oj-datepicker-buttonpane a");
    this._AddHoverable(buttons);
    this._AddActiveable(buttons);
  },

  _registerSwipeHandler : function()
  {
    if (oj.DomUtils.isTouchSupported())
    {
      var self = this;
      var stepMonths = this._getStepMonths();
      var rtl = this._IsRTL();
      var options = {
        'recognizers': [
          [Hammer.Swipe, {'direction': Hammer['DIRECTION_HORIZONTAL']}]
      ]};

      this._dpDiv.ojHammer(options).on(rtl ? 'swiperight' : 'swipeleft', function(evt)
      {
        self._gotoNext(stepMonths);
        return false;
      })
      .on(rtl ? 'swipeleft' : 'swiperight', function(evt) {
        self._gotoPrev(stepMonths);
        return false;
      });
    }
  },

  /**
   * Generate the HTML for the current state of the date picker.
   *
   * @private
   */
  _getMinMaxDateIso : function(minOrMax)
  {
    var minMaxDateIso = this.options[minOrMax];
    if(minMaxDateIso)
    {
      var dateIso = this._getDateIso();
      minMaxDateIso = oj.IntlConverterUtils._minMaxIsoString(minMaxDateIso, dateIso);
    }

    return minMaxDateIso;
  },

  /**
   * Generate the HTML for the header of the date picker.
   *
   * @private
   */
  _generateHeader : function(drawMonth, drawYear, monthControl, enablePrev, enableNext)
  {
    var header;
    var prevText, prev, nextText, next;
    var isRTL = this._IsRTL();

    prevText = this._EscapeXSS(this.getTranslatedString("prevText"));

    prev = (enablePrev ? "<a role='button' href='#' onclick='return false;' class='oj-datepicker-prev-icon oj-enabled oj-default oj-component-icon oj-clickable-icon-nocontext' data-handler='prev' data-event='click keydown'" + " aria-label='" + prevText + "'></a>" : "<a class='oj-datepicker-prev-icon oj-disabled oj-component-icon oj-clickable-icon-nocontext' title='" + prevText + "'></a>");

    nextText = this._EscapeXSS(this.getTranslatedString("nextText"));

    next = (enableNext ? "<a role='button' href='#' onclick='return false;' class='oj-datepicker-next-icon oj-enabled oj-default oj-component-icon oj-clickable-icon-nocontext' data-handler='next' data-event='click keydown'" + " aria-label='" + nextText + "'></a>" : "<a class='oj-datepicker-next-icon oj-disabled oj-component-icon oj-clickable-icon-nocontext' title='" + nextText + "'></a>");

    header = "<div class='oj-datepicker-header" + (this.options["disabled"] ? " oj-disabled " : " oj-enabled oj-default ") + "'>";

    header += (/all|left/.test(monthControl) ? (isRTL ? next : prev) : "");
    header += (/all|right/.test(monthControl) ? (isRTL ? prev : next) : "");
    header += this._generateMonthYearHeader(drawMonth, drawYear);

    header += "</div>";

    return header;
  },

  /**
   * Generate the HTML for the footer of the date picker.
   *
   * @protected
   * @ignore
   */
  _generateFooter : function(footerLayoutDisplay, gotoDate)
  {
    var footerLayout = "";
    var currentText = this._EscapeXSS(this.getTranslatedString("currentText"));
    var enabledClass = this.options["disabled"] ? "oj-disabled disabled" : "oj-enabled";
    var todayControl = "<a role='button' href='#' onclick='return false;' class='oj-datepicker-current oj-priority-secondary " + (this.options["disabled"] ? "oj-disabled' disabled" : "oj-enabled'") + " data-handler='today' data-event='click keydown'" + ">" + currentText + "</a>";

    if(footerLayoutDisplay.length > 1) //keep the code for future multiple buttons
    {
      var todayIndex = footerLayoutDisplay.indexOf("today"),
          loop = 0,
          footerLayoutButtons = [{index: todayIndex, content: (this._isInRange(gotoDate) ? todayControl : "")}];

      //rather than using several if + else statements, sort the content to add by index of the strings
      footerLayoutButtons.sort(function(a, b)
      {
        return a.index - b.index;
      });

      //continue to loop until the index > -1 [contains the string]
      while(loop < footerLayoutButtons.length && footerLayoutButtons[loop].index < 0) { loop++; }

      while(loop < footerLayoutButtons.length)
      {
        footerLayout += footerLayoutButtons[loop++].content;
      }

      if(footerLayout.length > 0)
      {
        footerLayout = "<div class='oj-datepicker-buttonpane'>" + footerLayout + "</div>";
      }
    }

    return footerLayout;
  },

  _isMultiMonth: function() 
  {
    var numMonths = this._getNumberOfMonths();
    return (numMonths[0] !== 1 || numMonths[1] !== 1);
  },

  /**
   * Generate the HTML for the current state of the date picker. Might be ugly in passing so many parameters, but 
   * during the month+year feature apparently the code was duplicated with copy+paste so though not pretty at least 
   * helps to manage main variables and etc
   * 
   * @param {string} view - 'Y', 'M', or 'D' like in other areas of the code
   * @private
   */
  _generateViewHTML : function(view)
  {
    var maxDraw, enablePrev, enableNext, converterUtils = oj.IntlConverterUtils,
        dateParams = ["date", "month", "fullYear"], converter = this._GetConverter(), footerLayout, weekDisplay, tempDate = new Date(),
        today = new Date(tempDate.getFullYear(), tempDate.getMonth(), tempDate.getDate()), // clear time
        isRTL = this._IsRTL(), footerLayoutDisplay = this.options["datePicker"]["footerLayout"], numMonths = this._getNumberOfMonths(),
        currentMonthPos = this.options["datePicker"]["currentMonthPos"], 
        isMultiMonth = this._isMultiMonth(),
        minDateIso = this._getMinMaxDateIso("min"), minDateParams,
        maxDateIso = this._getMinMaxDateIso("max"), maxDateParams,
        drawMonth = this._drawMonth - currentMonthPos, drawYear = this._drawYear,
        compareDate = new Date(this._currentYear, this._currentMonth, this._currentDay), valueDateIso = this._getDateIso(),
        valueDateParams = converterUtils._dateTime(valueDateIso, dateParams, true),
        selectedYear = valueDateParams["fullYear"],
        selectedDay = valueDateParams["date"], selectedMonth = valueDateParams["month"],
        valueDate = new Date(selectedYear, selectedMonth, selectedDay),
        wDisabled = this.options["disabled"], weekText = this._EscapeXSS(this.getTranslatedString("weekText"));

    if(minDateIso) {
      //convert it to the correct timezone for comparison, since need to display the month, date, year as displayed in isoString
      minDateIso = converter.parse(minDateIso);
      minDateParams = converterUtils._dateTime(minDateIso, dateParams, true);
    }
    if(maxDateIso) {
      maxDateIso = converter.parse(maxDateIso);
      maxDateParams = converterUtils._dateTime(maxDateIso, dateParams, true);
    }

    valueDateIso = converterUtils._clearTime(valueDateIso);

    //So per discussion calendar will display the year, month, date based on how represented in the isoString
    //meaning 2013-12-01T20:00:00-08:00 and 2013-12-01T20:00:00-04:00 will both display the same content as no
    //conversion will take place. In order to achieve this it will rip out the necessary info by string parsing
    //and in regards to isoString date comparison (i.e. whether one is before the other, will need to use converter's
    //compareISODates passing the MODIFIED printDate isoString)
    if (drawMonth < 0)
    {
      drawMonth += 12;
      drawYear--;
    }

    if(minDateParams)
    {
      var minDraw = new Date(minDateParams["fullYear"], minDateParams["month"], minDateParams["date"]);

      //tech shouldn't this error out? [previous existing jquery logic so keep, maybe a reason]
      if(maxDateParams && converter.compareISODates(maxDateIso, minDateIso) < 0)
      {
        minDraw = new Date(maxDateParams["fullYear"], maxDateParams["month"], maxDateParams["date"]);
      }
      while (new Date(drawYear, drawMonth, this._getDaysInMonth(drawYear, drawMonth)) < minDraw)
      {
        drawMonth++;
        if (drawMonth > 11)
        {
          drawMonth = 0;
          drawYear++;
        }
      }
    }

    if (maxDateParams)
    {
      maxDraw = new Date(maxDateParams["fullYear"], maxDateParams["month"] - (numMonths[0] * numMonths[1]) + 1, maxDateParams["date"]);

      //tech shouldn't this error out? [previous existing jquery logic so keep, maybe a reason]
      if(minDateParams && converter.compareISODates(maxDateIso, minDateIso) < 0)
      {
        maxDraw = new Date(minDateParams["fullYear"], minDateParams["month"], minDateParams["date"]);
      }
      while (new Date(drawYear, drawMonth, 1) > maxDraw)
      {
        drawMonth--;
        if (drawMonth < 0)
        {
          drawMonth = 11;
          drawYear--;
        }
      }
    }
    this._drawMonth = drawMonth;
    this._drawYear = drawYear;

    enablePrev = this._canAdjustMonth( - 1, drawYear, drawMonth) && !wDisabled;
    enableNext = this._canAdjustMonth( + 1, drawYear, drawMonth) && !wDisabled;

    footerLayout = this._generateFooter(footerLayoutDisplay, today);

    weekDisplay = this.options["datePicker"]["weekDisplay"];

    var result;

    if(view === 'D') 
    {
      result = this._generateDayHTMLContent(enablePrev, enableNext, converterUtils, converter, footerLayout, weekDisplay, 
                today, isRTL, numMonths, isMultiMonth, minDateParams, maxDateParams, drawMonth, drawYear, compareDate, valueDateParams, 
                selectedYear, selectedDay, selectedMonth, valueDate, wDisabled, weekText);
    } 
    else if(view === 'M') 
    {
      result = this._generateMonthHTMLContent(enablePrev, enableNext, converter, footerLayout, minDateParams,
                maxDateParams, drawMonth, drawYear, valueDate, wDisabled);
    } 
    else 
    {
      result = this._generateYearHTMLContent(enablePrev, enableNext, converterUtils, footerLayout, minDateParams,
                maxDateParams, drawMonth, drawYear, valueDate, wDisabled);
    }

    return result;
  },

  /**
   * @private
   */
  _generateDayHTMLContent : function(enablePrev, enableNext, converterUtils, converter, footerLayout, 
                              weekDisplay, today, isRTL, numMonths, isMultiMonth, minDateParams,
                              maxDateParams, drawMonth, drawYear, compareDate, valueDateParams,
                              selectedYear, selectedDay, selectedMonth, valueDate, wDisabled, weekText)
  {
    var dayNames = this.options["dayWide"], dayNamesMin = this.options["dayNarrow"],
        firstDay = this.options["firstDayOfWeek"], daysOutsideMonth, html, dow, row, group, col, selected, rowCellId,
        dayOverClass, dayOverId = "", dayOverClassStr, calender, thead, day, daysInMonth, leadDays, curRows, numRows,
        printDate, dRow, tbody, daySettings, otherMonth, unselectable, 
        dayFormatter = this.options["dayFormatter"], currMetaData = null, calculatedWeek;

    daysOutsideMonth = this.options["datePicker"]["daysOutsideMonth"];
    html = "";

    var monthControl = "all";
    for (row = 0;row < numMonths[0];row++)
    {
      group = "";
      this._maxRows = 4;
      for (col = 0;col < numMonths[1];col++)
      {
        monthControl = row === 0 ? "all" : "";
        calender = "";
        if (isMultiMonth)
        {
          calender += "<div class='oj-datepicker-group";
          if (numMonths[1] > 1)
          {
            switch (col)
            {
              case 0:
                calender += " oj-datepicker-group-first";
                monthControl = row === 0 ? (isRTL ? "right" : "left") : "";
                break;
              case numMonths[1] - 1:
                calender += " oj-datepicker-group-last";
                monthControl = row === 0 ? (isRTL ? "left" : "right") : "";
                break;
              default :
                calender += " oj-datepicker-group-middle";
                monthControl = "";
                break;
            }
          }
          calender += "'>";
        }

        calender += this._generateHeader(drawMonth, drawYear, monthControl, enablePrev, enableNext);

        calender += "<table class='oj-datepicker-calendar" + (weekDisplay === "number" ? " oj-datepicker-weekdisplay" : "") + (wDisabled ? " oj-disabled " : " oj-enabled oj-default ") + "' tabindex=-1 data-handler='calendarKey' data-event='keydown' aria-readonly='true' role='grid' " + "aria-labelledby='" + this._GetSubId(this._CALENDAR_DESCRIPTION_ID) + "'><thead role='presentation'>" + "<tr role='row'>";
        thead = (weekDisplay === "number" ? "<th class='oj-datepicker-week-col'>" + this._EscapeXSS(this.getTranslatedString("weekHeader")) + "</th>" : "");
        for (dow = 0;dow < 7;dow++)
        {
          // days of the week
          day = (dow + parseInt(firstDay, 10)) % 7;
          thead += "<th role='columnheader' aria-label='" + dayNames[day] + "'" + ((dow + firstDay + 6) % 7 >= 5 ? " class='oj-datepicker-week-end'" : "") + ">" + "<span title='" + dayNames[day] + "'>" + dayNamesMin[day] + "</span></th>";
        }
        calender += thead + "</tr></thead><tbody role='presentation'>";
        daysInMonth = this._getDaysInMonth(drawYear, drawMonth);
        if (drawYear === selectedYear && drawMonth === selectedMonth)
        {
          selectedDay = Math.min(selectedDay, daysInMonth);
        }
        leadDays = (this._getFirstDayOfMonth(drawYear, drawMonth) - firstDay + 7) % 7;
        curRows = Math.ceil((leadDays + daysInMonth) / 7);// calculate the number of rows to generate
        numRows = (isMultiMonth ? this._maxRows > curRows ? this._maxRows : curRows : curRows);//If multiple months, use the higher number of rows (see #7043)
        this._maxRows = numRows;
        printDate = new Date(drawYear, drawMonth, 1 - leadDays);
        for (dRow = 0;dRow < numRows;dRow++)
        {
          // create date picker rows
          calender += "<tr role='row'>";

          calculatedWeek = this._GetConverter().calculateWeek(converterUtils.dateToLocalIso(printDate));
          tbody = (weekDisplay === "none" ? "" : "<td class='oj-datepicker-week-col' role='rowheader' aria-label='" + weekText + " " + calculatedWeek + "'>" + calculatedWeek + "</td>");
          for (dow = 0;dow < 7;dow++)
          {
            // create date picker days
            otherMonth = (printDate.getMonth() !== drawMonth);
            selected = printDate.getTime() === valueDate.getTime();
            rowCellId = "oj-dp-" + this["uuid"] + "-" + dRow + "-" + dow + "-" + row + "-" + col;
            dayOverClass = (printDate.getTime() === compareDate.getTime() && drawMonth === this._currentMonth);
            if (dayOverClass)
            {
              dayOverId = rowCellId;
              dayOverClassStr = " " + this._DAYOVER_CLASS;
            }
            else
            {
              dayOverClassStr = "";
            }

            daySettings = [true, ""];
            var pYear = printDate.getFullYear(),
                pMonth = printDate.getMonth(),
                pDate = printDate.getDate();

            if (dayFormatter)
            {
              currMetaData = dayFormatter({"fullYear": pYear, "month": pMonth+1, "date": pDate}); //request to start from 1 rather than 0
              if (currMetaData)
              {
                //has content
                daySettings = [!currMetaData["disabled"], currMetaData["className"] || ""];
                if (currMetaData["tooltip"])
                {
                  daySettings.push(currMetaData["tooltip"]);
                }
              }
            }
            var selectedDate = printDate.getTime() === valueDate.getTime();

            unselectable = (otherMonth && daysOutsideMonth !== "selectable") || !daySettings[0] || this._outSideMinMaxRange(printDate, minDateParams, maxDateParams);
            tbody += "<td role='gridcell' aria-disabled='" + !!unselectable + "' aria-selected='" + selected + "' id='" + rowCellId + "' " + "class='" + ((dow + firstDay + 6) % 7 >= 5 ? " oj-datepicker-week-end" : "") + // highlight weekends
(otherMonth ? " oj-datepicker-other-month" : "") + // highlight days from other months
(dayOverClassStr) + // highlight selected day
(unselectable || wDisabled ? " " + this._UNSELECTABLE_CLASS + " oj-disabled" : " oj-enabled ") + // highlight unselectable days
(otherMonth && daysOutsideMonth === "hidden" ? "" : " " + daySettings[1] + // highlight custom dates
(selected ? " " + this._CURRENT_CLASS : "") + // highlight selected day
(printDate.getTime() === today.getTime() ? " oj-datepicker-today" : "")) + "'" + // highlight today (if different)
((!otherMonth || daysOutsideMonth !== "hidden") && daySettings[2] ? " title='" + daySettings[2].replace(/'/g, "&#39;") + "'" : "") + // cell title
(unselectable ? "" : " data-handler='selectDay' data-event='click' data-month='" + printDate.getMonth() + "' data-year='" + printDate.getFullYear() + "'") + ">" + // actions
(otherMonth && daysOutsideMonth === "hidden" ? "&#xa0;" : // display for other months
(unselectable || wDisabled ? "<span class='oj-disabled'>" + printDate.getDate() + "</span>" : "<a role='button' class='oj-enabled" + (selectedDate ? " oj-selected" : "") + // highlight selected day
(otherMonth ? " oj-priority-secondary" : "") + // distinguish dates from other months
"' " + (dayOverClass ? "" : "tabindex='-1' ") + " onclick='return false;' href='#'>" + printDate.getDate() + "</a>")) + "</td>";// display selectable date
            printDate.setDate(printDate.getDate() + 1);
          }
          calender += tbody + "</tr>";
        }
        drawMonth++;
        if (drawMonth > 11)
        {
          drawMonth = 0;
          drawYear++;
        }
        calender += "</tbody></table>" + (isMultiMonth ? "</div>" + ((numMonths[0] > 0 && col === numMonths[1] - 1) ? "<div class='oj-datepicker-row-break'></div>" : "") : "");
        group += calender;
      }
      html += group;
    }
    html += footerLayout;
    return {html : html, dayOverId : dayOverId};
  },

  /**
   * Generate the month and year header.
   *
   * @private
   */
  _generateMonthYearHeader : function(drawMonth, drawYear)
  {
    var changeMonth = this.options["datePicker"]["changeMonth"], changeYear = this.options["datePicker"]["changeYear"],
        positionOfMonthToYear = oj.LocaleData.isMonthPriorToYear() ? "before" : "after",
        html = "<div class='oj-datepicker-title'>", monthHtml = "", converterUtils = oj.IntlConverterUtils,
        monthNames = this.options["monthWide"],
        wDisabled = this.options["disabled"];

    // month selection
    if (monthNames)
    {
      if (changeMonth === "none")
      {
        monthHtml += "<span class='oj-datepicker-month'>" + monthNames[drawMonth] + "</span>";
      }
      else
      {
        monthHtml += "<a role='button' onclick='return false;' href='#' data-handler='selectMonthHeader' data-event='click keydown' class='oj-datepicker-month " + (wDisabled ? "oj-disabled' disabled" : "oj-enabled'") + ">";
        monthHtml += monthNames[drawMonth] + "</a>";
      }

      if (positionOfMonthToYear === "before")
      {
        html += monthHtml + (!((changeMonth === "select") && (changeYear === "select")) ? "&#xa0;" : "");
      }
    }

    // year selection
    if (!this.yearshtml)
    {
      this.yearshtml = "";
      if (changeYear === "none")
      {
        html += "<span class='oj-datepicker-year'>" + yearDisplay.format(converterUtils.dateToLocalIso(new Date(drawYear, drawMonth, 1))) + "</span>";
      }
      else
      {
        html += "<a role='button' onclick='return false;' href='#' data-handler='selectYearHeader' data-event='click keydown' class='oj-datepicker-year " + (wDisabled ? "oj-disabled' disabled" : "oj-enabled'") + ">";
        html += yearDisplay.format(converterUtils.dateToLocalIso(new Date(drawYear, drawMonth, 1))) + "</a>";
        this.yearshtml = null;
      }
    }

    if (monthNames)
    {
      if (positionOfMonthToYear === "after")
      {
        html += (!((changeMonth === "select") && (changeYear === "select")) ? "&#xa0;" : "") + monthHtml;
      }
    }

    html += "<span aria-hidden='true' class='oj-helper-hidden-accessible' id='" + this._GetSubId(this._DATEPICKER_DIALOG_DESCRIPTION_ID) + "'>";
    html += this._EscapeXSS(this.getTranslatedString("picker")) + "</span>";
    html += "<span aria-hidden='true' class='oj-helper-hidden-accessible' id='" + this._GetSubId(this._CALENDAR_DESCRIPTION_ID) + "'>";
    html += (monthNames ? (monthNames[drawMonth] + " ") : "") + yearDisplay.format(converterUtils.dateToLocalIso(new Date(drawYear, drawMonth, 1))) + "</span>";

    html += "<span aria-hidden='true' class='oj-helper-hidden-accessible' id='" + this._GetSubId(this._DATEPICKER_DESCRIPTION_ID) + "'>" + this._EscapeXSS(this.getTranslatedString("datePicker")) + "</span>";

    html += "</div>";// Close datepicker_header
    return html;
  },

  /**
   * Adjust one of the date sub-fields.
   *
   * @private
   */
  _adjustInstDate : function (offset, period)
  {
    var year = this._drawYear + (period === "Y" ? offset : 0),
        month = this._drawMonth + (period === "M" ? offset : 0),
        day = Math.min(this._currentDay, this._getDaysInMonth(year, month)) + (period === "D" ? offset : 0),
        date = new Date(year, month, day);

    this._currentDay = date.getDate();
    this._drawMonth = this._currentMonth = date.getMonth();
    this._drawYear = this._currentYear = date.getFullYear();
  },

  /**
   * Generate the HTML for the month view of the date picker.
   *
   * @private
   */
  _generateMonthHTMLContent : function(enablePrev, enableNext, converter, footerLayout, minDateParams,
                                maxDateParams, drawMonth, drawYear, valueDate, wDisabled)
  {
    var monthNames = this.options["monthWide"], monthNamesShort = this.options["monthAbbreviated"],
        html, dow, selected, rowCellId, dayOverClass, dayOverId = "", dayOverClassStr, calender, 
        printDate, dRow, tbody, unselectable;

    enablePrev = this._canAdjustYear( - 1, drawYear) && !wDisabled;
    enableNext = this._canAdjustYear( + 1, drawYear) && !wDisabled;

    html = "";

    this._maxRows = 4;

    calender = "";

    calender += this._generateHeader(drawMonth, drawYear, "all", enablePrev, enableNext);

    calender += "<table class='oj-datepicker-calendar oj-datepicker-monthview" + (wDisabled ? " oj-disabled " : " oj-enabled oj-default ") + "' tabindex=-1 data-handler='calendarKey' data-event='keydown' aria-readonly='true' role='grid' " + "aria-labelledby='" + this._GetSubId(this._CALENDAR_DESCRIPTION_ID) + "'>";

    calender += "<tbody role='presentation'>";
    printDate = new Date(drawYear, 0, 1);
    for (dRow = 0;dRow < 4;dRow++)
    {
      // create date picker rows
      calender += "<tr role='row'>";

      tbody = "";
      for (dow = 0;dow < 3;dow++)
      {
        var month = dRow * 3 + dow;
        // create date picker days
        selected = printDate.getMonth() === valueDate.getMonth();
        rowCellId = "oj-dp-" + this["uuid"] + "-" + dRow + "-" + dow + "-" + 0 + "-" + 0;
        dayOverClass = (month === this._currentMonth);
        if (dayOverClass)
        {
          dayOverId = rowCellId;
          dayOverClassStr = " " + this._DAYOVER_CLASS;
        }
        else
        {
          dayOverClassStr = "";
        }

        var selectedDate = printDate.getMonth() === valueDate.getMonth();
        var inMinYear = (minDateParams && minDateParams["fullYear"] === drawYear);
        var inMaxYear = (maxDateParams && maxDateParams["fullYear"] === drawYear);

        unselectable = !((!inMinYear || month >= minDateParams["month"]) && (!inMaxYear || month <= maxDateParams["month"]));

        tbody += "<td role='gridcell' aria-disabled='" + !!unselectable + "' aria-selected='" + selected + "' id='" + rowCellId + "' " + "class='" +
(dayOverClassStr) + // highlight selected day
(unselectable || wDisabled ? " " + this._UNSELECTABLE_CLASS + " oj-disabled" : " oj-enabled ") + // highlight unselectable days
(selected ? " " + this._CURRENT_CLASS : "") + "'" + // highlight selected day
(unselectable ? "" : " data-handler='selectMonth' data-event='click' data-month='" + printDate.getMonth() + "' data-year='" + printDate.getFullYear() + "'") + ">" + // actions
((unselectable || wDisabled ? "<span class='oj-disabled'>" + monthNamesShort[month] + "</span>" : "<a role='button' class='oj-enabled" + (selectedDate ? " oj-selected" : "") + // highlight selected day
"' " + (dayOverClass ? "" : "tabindex='-1' ") + " onclick='return false;' href='#'>" + monthNamesShort[month] + "</a>")) + "</td>";// display selectable date
        printDate.setMonth(printDate.getMonth() + 1);
      }
      calender += tbody + "</tr>";
    }
    drawMonth++;
    if (drawMonth > 11)
    {
      drawMonth = 0;
      drawYear++;
    }
    calender += "</tbody></table>";

    html += calender;

    html += footerLayout;
    return {html : html, dayOverId : dayOverId};
  },

  /**
   * Generate the HTML for the current state of the date picker.
   *
   * @private
   */
  _generateYearHTMLContent : function(enablePrev, enableNext, converterUtils, footerLayout, minDateParams,
                                maxDateParams, drawMonth, drawYear, valueDate, wDisabled)
  {
    var html, dow, selected, rowCellId, dayOverClass, dayOverId = "", dayOverClassStr, calender, thead, 
        printDate, dRow, tbody, unselectable;

    enablePrev = this._canAdjustDecade( - 1, drawYear) && !wDisabled;
    enableNext = this._canAdjustDecade( + 1, drawYear) && !wDisabled;

    html = "";

    this._maxRows = 4;

    calender = "";

    calender += this._generateHeader(drawMonth, drawYear, "all", enablePrev, enableNext);

    calender += "<table class='oj-datepicker-calendar oj-datepicker-yearview" + (wDisabled ? " oj-disabled " : " oj-enabled oj-default ") + "' tabindex=-1 data-handler='calendarKey' data-event='keydown' aria-readonly='true' role='grid' " + "aria-labelledby='" + this._GetSubId(this._CALENDAR_DESCRIPTION_ID) + "'>";

    calender += "<tbody role='presentation'>";

    var yearRange = this._getYearRange(drawYear, minDateParams, maxDateParams);
    var baseYear = (Math.floor(drawYear / 10) * 10);
    printDate = new Date(baseYear, drawMonth, 1);
    for (dRow = 0;dRow < 4;dRow++)
    {
      // create date picker rows
      calender += "<tr role='row'>";

      tbody = "";
      for (dow = 0;dow < 3;dow++)
      {
        if (dRow == 3 && dow == 1)
          break;

        var year = baseYear + (dRow * 3) + dow;
        // create date picker days
        selected = printDate.getFullYear() === valueDate.getFullYear();
        rowCellId = "oj-dp-" + this["uuid"] + "-" + dRow + "-" + dow + "-" + 0 + "-" + 0;
        dayOverClass = (year === this._currentYear);
        if (dayOverClass)
        {
          dayOverId = rowCellId;
          dayOverClassStr = " " + this._DAYOVER_CLASS;
        }
        else
        {
          dayOverClassStr = "";
        }

        var selectedDate = printDate.getFullYear() === valueDate.getFullYear();
        var yearText = yearDisplay.format(converterUtils.dateToLocalIso(new Date(year, drawMonth, 1)));

        unselectable = (year < yearRange['startYear'] || year > yearRange['endYear']);

        tbody += "<td role='gridcell' aria-disabled='" + !!unselectable + "' aria-selected='" + selected + "' id='" + rowCellId + "' " + "class='" +
(dayOverClassStr) + // highlight selected day
(unselectable || wDisabled ? " " + this._UNSELECTABLE_CLASS + " oj-disabled" : " oj-enabled ") + // highlight unselectable days
(selected ? " " + this._CURRENT_CLASS : "") + "'" + // highlight selected day
(unselectable ? "" : " data-handler='selectYear' data-event='click' data-month='" + printDate.getMonth() + "' data-year='" + printDate.getFullYear() + "'") + ">" + // actions
((unselectable || wDisabled ? "<span class='oj-disabled'>" + yearText + "</span>" : "<a role='button' class='oj-enabled" + (selectedDate ? " oj-selected" : "") + // highlight selected day
"' " + (dayOverClass ? "" : "tabindex='-1' ") + " onclick='return false;' href='#'>" + yearText + "</a>")) + "</td>";// display selectable date
        printDate.setFullYear(printDate.getFullYear() + 1);
      }
      calender += tbody + "</tr>";
    }
    drawMonth++;
    if (drawMonth > 11)
    {
      drawMonth = 0;
      drawYear++;
    }
    calender += "</tbody></table>";

    html += calender;

    html += footerLayout;
    return {html : html, dayOverId : dayOverId};
  },

  /**
   * Determine the selectable years.
   *
   * @private
   */
  _getYearRange : function(drawYear, minDateParams, maxDateParams)
  {
    var years, thisYear, determineYear, year, endYear;

    years = this.options["datePicker"]["yearRange"].split(":");
    thisYear = new Date().getFullYear();
    determineYear = function (value)
    {
      var year = (value.match(/c[+\-].*/) ? drawYear + parseInt(value.substring(1), 10) : (value.match(/[+\-].*/) ? thisYear + parseInt(value, 10) : parseInt(value, 10)));
      return (isNaN(year) ? thisYear : year);
    };
    year = determineYear(years[0]);
    endYear = Math.max(year, determineYear(years[1] || ""));
    year = (minDateParams ? Math.max(year, minDateParams["fullYear"]) : year);
    endYear = (maxDateParams ? Math.min(endYear, maxDateParams["fullYear"]) : endYear);

    return {'startYear': year, 'endYear': endYear};
  },

  /**
   * Determine the number of months to show.
   *
   * @private
   */
  _getNumberOfMonths : function ()
  {
    var numMonths = this.options["datePicker"]["numberOfMonths"];
    numMonths = typeof numMonths === "string" ? parseInt(numMonths, 10) : numMonths;
    return (numMonths == null ? [1, 1] : (typeof numMonths === "number" ? [1, numMonths] : numMonths));
  },

  /**
   * Find the number of days in a given month.
   *
   * @private
   */
  _getDaysInMonth : function (year, month)
  {
    return 32 - new Date(year, month, 32).getDate();
  },

  /**
   * Find the day of the week of the first of a month.
   *
   * @private
   */
  _getFirstDayOfMonth : function (year, month)
  {
    return new Date(year, month, 1).getDay();
  },

  /**
   * Determines if we should allow a "next/prev" month display change.
   *
   * @private
   */
  _canAdjustMonth : function (offset, curYear, curMonth)
  {
    var numMonths = this._getNumberOfMonths(), date = new Date(curYear, curMonth + (offset < 0 ? offset : numMonths[0] * numMonths[1]), 1);

    if (offset < 0)
    {
      date.setDate(this._getDaysInMonth(date.getFullYear(), date.getMonth()));
    }
    return this._isInRange(date);
  },

  /**
   * Determines if we should allow a "next/prev" year display change.
   *
   * @private
   */
  _canAdjustYear : function(offset, curYear)
  {
    var date;

    if (offset < 0)
    {
      date = new Date(curYear + offset, 12, 1);
      date.setDate(this._getDaysInMonth(date.getFullYear(), date.getMonth()));
    }
    else
    {
      date = new Date(curYear + offset, 1, 1);
    }
    return this._isInRange(date);
  },

  /**
   * Determines if we should allow a "next/prev" decade display change.
   *
   * @private
   */
  _canAdjustDecade : function(offset, curYear)
  {
    var date;
    var baseYear = (Math.floor(curYear / 10) * 10);

    if (offset < 0)
    {
      date = new Date(baseYear + 9 + (offset * 10), 12, 1);
      date.setDate(this._getDaysInMonth(date.getFullYear(), date.getMonth()));
    }
    else
    {
      date = new Date(baseYear + (offset * 10), 1, 1);
    }
    return this._isInRange(date);
  },

  /**
   * Returns a boolean of whether the print date is outside the min + max range, ignoring time since if of
   * ojInputDateTime should allow selection of date and restrict based on ojInputTime.
   *
   * @private
   */
  _outSideMinMaxRange : function (printDate, minDateParams, maxDateParams)
  {
    var minDate = minDateParams ? new Date(minDateParams["fullYear"], minDateParams["month"], minDateParams["date"]) : null;
    var maxDate = maxDateParams ? new Date(maxDateParams["fullYear"], maxDateParams["month"], maxDateParams["date"]) : null;

    return (minDate !== null && printDate < minDate) || (maxDate !== null && printDate > maxDate);
  },

  /**
   * Is the given date in the accepted range?
   *
   * @param {Object} date constructed using local; however need to compare with min + max using their timezone
   * @private
   */
  _isInRange : function (date)
  {
    var yearSplit, currentYear, converterUtils = oj.IntlConverterUtils,
        converter = this._GetConverter(),
        minDate, maxDate,
        minDateIso = this._getMinMaxDateIso("min"), minYear = null,
        maxDateIso = this._getMinMaxDateIso("max"), maxYear = null, years = this.options["datePicker"]["yearRange"];

    if (years)
    {
      yearSplit = years.split(":");
      currentYear = new Date().getFullYear();
      minYear = parseInt(yearSplit[0], 10);
      maxYear = parseInt(yearSplit[1], 10);
      if (yearSplit[0].match(/[+\-].*/))
      {
        minYear += currentYear;
      }
      if (yearSplit[1].match(/[+\-].*/))
      {
        maxYear += currentYear;
      }
    }

    if(minDateIso)
    {
      //need to convert it to the same timezone as the value, since the calendar and etc
      //all work by using string manipulation of the isoString
      minDateIso = converter.parse(minDateIso);
      var minDateParams = converterUtils._dateTime(minDateIso, ["fullYear", "month", "date"], true);
      minDate = new Date(minDateParams["fullYear"], minDateParams["month"], minDateParams["date"]);
    }
    if(maxDateIso)
    {
      maxDateIso = converter.parse(maxDateIso);
      var maxDateParams = converterUtils._dateTime(maxDateIso, ["fullYear", "month", "date"], true);
      maxDate = new Date(maxDateParams["fullYear"], maxDateParams["month"], maxDateParams["date"]);
    }

    return ((!minDate || date.getTime() >= minDate.getTime()) && (!maxDate || date.getTime() <= maxDate.getTime())
            && (!minYear || date.getFullYear() >= minYear) && (!maxYear || date.getFullYear() <= maxYear));
  },

  /**
   * @protected
   * @override
   * @instance
   * @ignore
   * @memberof! oj.ojInputDate
   */
  _GetCalendarTitle : function ()
  {
    return this._EscapeXSS(this.getTranslatedString("tooltipCalendar" + (this.options["disabled"] ? "Disabled" : "")));
  },

  /**
   * To disable or enable the widget
   *
   * @private
   * @instance
   */
  _disableEnable : function (val)
  {
    if (this._triggerNode)
    {
      disableEnableSpan(this._triggerNode.children(), val);
      this._triggerNode.find("." + this._TRIGGER_CALENDAR_CLASS).attr("title", this._GetCalendarTitle());
    }

    if(val)
    {
      this._hide(this._ON_CLOSE_REASON_CLOSE);
    }

    //need to update the look, note that if it is displaying the datepicker dropdown it would be hidden in _setOption function
    if(this._isInLine)
    {
      this._updateDatepicker();
    }
  },

  /**
   * Invoke super only if it is not inline
   *
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDate
   */
  _AppendInputHelper : function ()
  {
    if (!this._isInLine)
    {
      this._superApply(arguments);
    }
  },

  /**
   * This handler will set the value from the input text element.
   *
   * @ignore
   * @protected
   * @override
   * @param {Event} event
   * @instance
   * @memberof! oj.ojInputDate
   */
  _onBlurHandler : function (event)
  {
    if(this._isInLine)
    {
      return;
    }

    this._superApply(arguments);
  },

  /**
   * This handler will be invoked when keydown is triggered for this.element. When is of inline ignore the keydowns
   *
   * @ignore
   * @protected
   * @override
   * @param {Event} event
   * @instance
   * @memberof! oj.ojInputDate
   */
  _onKeyDownHandler : function (event)
  {
    if(this._isInLine)
    {
      return;
    }

    this._superApply(arguments);

    var kc = $.ui.keyCode,
        handled = false;

    if (this._datepickerShowing())
    {

      switch (event.keyCode)
      {
        case kc.TAB:
          this._hide(this._ON_CLOSE_REASON_TAB);
          break;
        case kc.ESCAPE:
          this._hide(this._ON_CLOSE_REASON_CANCELLED);
          handled = true;
          break;
        case kc.UP: ;
        case kc.DOWN:
          // when entering the datepicker via the up/down arrows, we should put the focus on the
          // current day, which will be the only tabbable stop.  This also helps an IE11 bug that
          // wasn't allowing tabbing to different elements in the popup ().
          // This also fixes an unreported issue of needing to tab an extra time before it tabs to
          // the next tab stop (all browsers).
          var datePickerCalElem = this._dpDiv.find(".oj-datepicker-calendar")[0]
          var tabStop = oj.FocusUtils.focusFirstTabStop(datePickerCalElem);
          // if we don't find a tab stop, we'll focus on the whole calendar like before.
          if (!tabStop)
            datePickerCalElem.focus();
          
          handled = true;
          break;
        default: ;
      }

    }
    else
    {

      switch (event.keyCode)
      {
        case kc.UP: ;
        case kc.DOWN:
          this._SetValue(this._GetDisplayValue(), event);
          this.show();
          handled = true;
          break;
        default: ;
      }

    }

    if (handled)
    {
      event.preventDefault();
      event.stopPropagation();
      return false;
    }
  },

  /**
   * Ignore for when of inline, since then this.element would be of div and has a funky nature
   *
   * @param {String} displayValue of the new string to be displayed
   *
   * @memberof! oj.ojInputDate
   * @instance
   * @protected
   * @override
  */
  _SetDisplayValue : function (displayValue)
  {
    if(!this._isInLine)
    {
      this._superApply(arguments);
    }

    this._setCurrentDate(this._getDateIso());

    //so this is a change in behavior from original design. Previously it was decided that app developer
    //would have to invoke refresh to render the calendar after setting the new value programatically; however now it is
    //required to hook it in when _SetDisplayValue is invoked [can't use _SetValue b/c that function is not invoked
    //when developer invokes ("option", "value", oj.IntlConverterUtils.dateToLocalIso(new Date()))
    if(this._datepickerShowing())
    {
      // _SetDisplayValue is called after user picks a date from picker, we dont want to bring
      //  focus to input element if the picker is showing still for the non-inline case. For the
      //  case of inline date picker, if the time picker field already had focus (brought in when
      //  the picker was hidden), we want to update the date picker, but not set focus on it.
      var focusOnCalendar = !(this._isInLine && this._timePicker && this._timePicker[0] === document.activeElement);
      this._updateDatepicker(focusOnCalendar);
    }
  },

  /**
   * Need to override since apparently we allow users to set the converter to null, undefined, and etc and when
   * they do we use the default converter
   *
   * @return {Object} a converter instance or null
   *
   * @memberof! oj.ojInputDate
   * @instance
   * @protected
   * @override
   */
  _GetConverter : function ()
  {
    return this.options['converter'] ?
        this._superApply(arguments) :
        $["oj"]["ojInputDate"]["prototype"]["options"]["converter"];
  },

  /**
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDate
   */
  _GetElementValue : function ()
  {
    return this.options['value'] || "";
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDate
   * @return {string}
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-inputdate";
  },

  /**
   * Sets up the default dateTimeRange and dateRestriction validators.
   *
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDate
   */
  _GetImplicitValidators : function ()
  {
    var ret = this._superApply(arguments);

    if(this.options['min'] != null || this.options['max'] != null)
    {
      //need to alter how the default validators work as validators are now immutable
      this._datePickerDefaultValidators[oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE] = this._getValidator("min");
    }

    if(this.options["dayFormatter"] != null)
    {
      this._datePickerDefaultValidators[oj.ValidatorFactory.VALIDATOR_TYPE_DATERESTRICTION] = this._getValidator("dayFormatter");
    }

    return $.extend(this._datePickerDefaultValidators, ret);
  },

  /**
   * Notifies the component that its subtree has been removed from the document programmatically after the component has
   * been created
   * @memberof! oj.ojInputDate
   * @instance
   * @protected
   */
  _NotifyDetached: function()
  {
    this._hide(this._ON_CLOSE_REASON_CLOSE);

    // hide sets focus to the input, so we want to call super after hide. If we didn't, then
    // the messaging popup will reopen and we don't want that.
    this._superApply(arguments);
  },

  /**
   * Notifies the component that its subtree has been made hidden programmatically after the component has
   * been created
   * @memberof! oj.ojInputDate
   * @instance
   * @protected
   */
  _NotifyHidden: function()
  {
    this._hide(this._ON_CLOSE_REASON_CLOSE);

    // hide sets focus to the input, so we want to call super after hide. If we didn't, then
    // the messaging popup will reopen and we don't want that.
    this._superApply(arguments);
  },

  _getValidator : function (key)
  {
    var validator = null;

    if(key === "min" || key === "max")
    {

      validator = getImplicitDateTimeRangeValidator(this.options, this._GetConverter(), this._GetDefaultStyleClass());
    }
    else if(key === "dayFormatter")
    {
      var dateRestrictionOptions = {'dayFormatter': this.options["dayFormatter"],
                                    'converter': this._GetConverter() };

      $.extend(dateRestrictionOptions, this.options['translations']['dateRestriction'] || {});
      validator = oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_DATERESTRICTION).createValidator(dateRestrictionOptions);
    }

    return validator;
  },

  /**
   * Gets today's date w/o time
   *
   * @private
   * @return {Object} date
   */
  _getTodayDate : function ()
  {
    var date = new Date();
    date.setHours(0);
    date.setMinutes(0);
    date.setSeconds(0);
    date.setMilliseconds(0);
    return date;
  },

  /**
   * Retrieve the default date shown on opening.
   *
   * @private
   */
  _getDateIso : function ()
  {
    var value = this.options['value'] || this._getDefaultIsoDate();
    if(this._isDateTimeSwitcher() && this._switcherDateValue) 
    {
      value = this._switcherDateValue;
    }

    return value;
  },

  // @inheritdoc
  getNodeBySubId: function(locator)
  {
    var node = null,
        subId = locator && locator['subId'],
        dpDiv = this._dpDiv,
        dpContentDiv = this._getDatepickerContent();

    if(subId)
    {
      switch(subId)
      {
      case "oj-datepicker-content": node = dpContentDiv[0]; break;
      case "oj-inputdatetime-calendar-icon": node = $(".oj-inputdatetime-calendar-icon", this._triggerNode)[0]; break;
      case "oj-datepicker-prev-icon": node = $(".oj-datepicker-prev-icon", dpDiv)[0]; break;
      case "oj-datepicker-next-icon": node = $(".oj-datepicker-next-icon", dpDiv)[0]; break;
      case "oj-datepicker-month": node = $(".oj-datepicker-month", dpDiv)[0]; break;
      case "oj-datepicker-year": node = $(".oj-datepicker-year", dpDiv)[0]; break;
      case "oj-datepicker-current": node = $(".oj-datepicker-current", dpDiv)[0]; break;
      case "oj-inputdatetime-date-input": node = this._isInLine ? null : this.element[0]; break;
      default: node = null;
      }
    }

    // Non-null locators have to be handled by the component subclasses
    return node || this._superApply(arguments);
  },

  // @inheritdoc
  getSubIdByNode: function(node)
  {
    var dpDiv = this._dpDiv,
        subId = null,
        checks = [{"selector": ".oj-inputdatetime-calendar-icon", "ele": this._triggerNode},
                  {"selector": ".oj-datepicker-prev-icon", "ele": dpDiv},
                  {"selector": ".oj-datepicker-next-icon", "ele": dpDiv},
                  {"selector": ".oj-datepicker-month", "ele": dpDiv},
                  {"selector": ".oj-datepicker-year", "ele": dpDiv},
                  {"selector": ".oj-datepicker-current", "ele": dpDiv}];

    if(node === dpDiv[0])
    {
      return "oj-datepicker-content";
    }
    if(!this._isInLine && node === this.element[0])
    {
      return "oj-inputdatetime-date-input";
    }

    for(var i=0, j=checks.length; i < j; i++)
    {
      var map = checks[i],
          entry = $(map["selector"], map["ele"]);

      if(entry.length === 1 && entry[0] === node)
      {
        subId = map["selector"].substr(1);
        break;
      }
    }

    return subId || this._superApply(arguments);
  },

  /**
   * Hides the datepicker. Note that this function is a no-op when renderMode is 'native'.
   *
   * @expose
   * @memberof oj.ojInputDate
   * @instance
   * @return {void}
   */
  hide : function ()
  {
    return this._hide(this._ON_CLOSE_REASON_CLOSE);
  },

  /**
   * Hides the datepicker
   *
   * @param {string} reason - the reason that the popup is being hidden ("selection", "cancelled", "tab")
   *
   * @protected
   * @ignore
   */
  _hide : function (reason)
  {
    if (!isPickerNative(this) && this._datepickerShowing() && !this._isInLine)
    {
      this._popUpDpDiv.ojPopup("close");
      this._onClose(reason);
    }

    return this;
  },

  /**
   * Sets focus to the right place after the picker is closed
   *
   * @param {string} reason - the reason that the popup is being hidden ("selection", "cancelled", "tab")
   *
   * @protected
   * @ignore
   */
  _onClose : function (reason)
  {
    if(this._isMobile && this.options["datePicker"]["showOn"] === "focus")
    {
      this._inputContainer.focus();
    }
    else
    {
      if(this.options["datePicker"]["showOn"] === "focus")
      {
        this._ignoreShow = true;
      }
      this.element.focus();
    }
  },

  /**
   * Refreshes the element. Usually called after dom changes have been made.
   * @expose
   * @memberof oj.ojInputDate
   * @instance
   * @return {void}
   */
  refresh : function ()
  {
    if(this._triggerNode)
    {
      this._triggerNode.find("." + this._TRIGGER_CALENDAR_CLASS).attr("title", this._GetCalendarTitle());
    }
    return this._superApply(arguments) || this;
  },

  /**
   * Shows the datepicker
   *
   * @expose
   * @memberof oj.ojInputDate
   * @instance
   * @return {void}
   */
  show : function ()
  {
    if (this._datepickerShowing() || this.options["disabled"] || this.options["readOnly"])
    {
      return;
    }

    if (this._ignoreShow)
    {
      //set within hide or elsewhere and focus is placed back on this.element
      this._ignoreShow = false;
      return;
    }

    if (isPickerNative(this))
    {
      // our html picker is inside popup, which will take care of removing focus from input element,
      //  for native case we do it explicitly
      this.element.blur();

      // picker expects the fields like 'date' and 'mode' to retain its names. Use bracket notation
      //  to avoid closure compiler from renaming them
      var pickerOptions = {};
      pickerOptions['date'] = _getNativePickerDate(this._getNativeDatePickerConverter(), this._getDateIso());
      pickerOptions['mode'] = "date";

      return this._ShowNativeDatePicker(pickerOptions);
    }
    else
    {
      return this._ShowHTMLDatePicker();
    }
  },

  /**
   * TODO: Technically i think should be used for the calendar, but later since late in release
   * @ignore
   */
  _getNativeDatePickerConverter: function () {
    if(this._nativePickerConverter === null) {
      var resolvedOptions = this._GetConverter().resolvedOptions();
      var options = {};
      $.extend(options, resolvedOptions, {"isoStrFormat": "offset"});

      this._nativePickerConverter = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter(options);
    }

    return this._nativePickerConverter;
  },

  /**
   * Shows the native datepicker
   *
   * @protected
   * @memberof! oj.ojInputDate
   * @instance
   */
  _ShowNativeDatePicker : function (pickerOptions)
  {
    var minDate = this._getMinMaxDateIso("min");
    var maxDate = this._getMinMaxDateIso("max");
    var conv = this._getNativeDatePickerConverter();

    if (minDate)
    {
      pickerOptions['minDate'] = _getNativePickerDate(conv, minDate).valueOf();
    }

    if (maxDate)
    {
      pickerOptions['maxDate'] = _getNativePickerDate(conv, minDate).valueOf();
    }

    var self = this;

    // onError is called only for Android for cases where picker is cancelled, or if there were
    //  to be any error at the native picker end
    function onError(error)
    {
      self._nativePickerShowing = false;

      // if user cancels the picker dialog, we just bring the focus back
      // closure compiler renames 'startsWith', using bracket notation hence
      if (error["startsWith"]('cancel'))
      {
        self._onClose(self._ON_CLOSE_REASON_CANCELLED);
      }
      else
      {
        oj.Logger.log('Error: native date or time picker failed: ' + error);
      }
    }

    self._nativePickerShowing = true;

    // datePicker is variable at the top level available when the cordova date picker plugin is
    //  included.
    window['datePicker'].show(pickerOptions, $.proxy(this._OnDatePicked, this), onError);
  },

  /**
   * callback upon picking date from native picker
   *
   * @protected
   * @memberof! oj.ojInputDate
   * @instance
   */
  _OnDatePicked : function (date)
  {
    this._nativePickerShowing = false;

    // for iOS and windows, from the current implementation of the native datepicker plugin,
    //  for case when the picker is cancelled, this callback gets called without the parameter
    if (date)
    {
      // Explicitly setting timezone is supported only in iOS, and we do not have a need to do
      //  so at this time, so not exposing this feature for now.
      // The native date picker will preserve the timezone set on the supplied date upon returning,
      // however the returned value has its time part reset to 00:00 when in 'date' mode
      //  - need to copy time over hence
      var isoString = oj.IntlConverterUtils._dateTime(this._getDateIso(), {"month": date.getMonth(), "date": date.getDate(), "fullYear": date.getFullYear()});
      var formattedTime = this._GetConverter().format(isoString);

      // _SetValue will inturn call _SetDisplayValue
      this._SetValue(formattedTime, {});
    }

    this._onClose(this._ON_CLOSE_REASON_SELECTION);
  },

  /**
   * Shows the HTML datepicker
   *
   * @ignore
   * @protected
   * @memberof oj.ojInputDate
   */
  _ShowHTMLDatePicker : function ()
  {
    var rtl = this._IsRTL();

    this._switcherPrevValue = this._getDateIso();
    this._switcherPrevDay = this._currentDay;
    this._switcherPrevMonth = this._currentMonth;
    this._switcherPrevYear = this._currentYear;

    //to avoid flashes on Firefox
    this._getDatepickerContent().empty();
    this._updateDatepicker();

    if(!_isLargeScreen)
    {
      this._popUpDpDiv.ojPopup("open", this.element.parent(), {
        "my" : {"horizontal":"center", "vertical": "bottom"},
        "at" : {"horizontal":"center", "vertical": "bottom"},
        "of" : window,
        "collision" : "flipfit flipfit"
      });
      
      // if we don't have a large screen, the popup will be modal so
      // we need to give it the focus
      this._dpDiv.find(".oj-datepicker-calendar").focus();
    }
    else
    {
      var position = oj.PositionUtils.normalizeHorizontalAlignment({"my" : "start top", "at" : "start bottom", "of" : this.element, "collision" : "flipfit flipfit"}, rtl);
      this._popUpDpDiv.ojPopup("open", this.element.parent(), position);
    }
    
    return this;
  }
});

// Add custom getters for properties
oj.Components.setDefaultOptions(
{
  'ojInputDate':
  {
    'firstDayOfWeek': oj.Components.createDynamicPropertyGetter(
      function()
      {
         return oj.LocaleData.getFirstDayOfWeek();
      }),

    'dayWide': oj.Components.createDynamicPropertyGetter(
      function()
      {
         return oj.LocaleData.getDayNames("wide");
      }),

    'dayNarrow': oj.Components.createDynamicPropertyGetter(
      function()
      {
          return oj.LocaleData.getDayNames("narrow");
      }),

    'monthWide': oj.Components.createDynamicPropertyGetter(
      function()
      {
         return oj.LocaleData.getMonthNames("wide");
      }),

    'monthAbbreviated': oj.Components.createDynamicPropertyGetter(
      function()
      {
         return oj.LocaleData.getMonthNames("abbreviated");
      }),

    'datePicker': oj.Components.createDynamicPropertyGetter(
      function()
      {
        return (oj.ThemeUtils.parseJSONFromFontFamily('oj-inputdatetime-option-defaults') || {})["datePicker"];
      }),

    'renderMode': oj.Components.createDynamicPropertyGetter(
      function()
      {
        return (oj.ThemeUtils.parseJSONFromFontFamily('oj-inputdatetime-option-defaults') || {})["renderMode"];
      }),

    'keyboardEdit': oj.Components.createDynamicPropertyGetter(
      function()
      {
        return (oj.ThemeUtils.parseJSONFromFontFamily('oj-inputdatetime-option-defaults') || {})["keyboardEdit"];
      })
    }
  }
);

// Fragments:

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Input element and calendar trigger icon</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>When not inline, shows the grid and moves the focus into the expanded date grid</td>
 *     </tr>
 *     <tr>
 *       <td>Input element with picker open</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Set focus to the input. If hints, title or messages exist in a notewindow,
 *        pop up the notewindow.</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Left</kbd></td>
 *       <td>Switch to next month (or previous month on RTL page).</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Right</kbd></td>
 *       <td>Switch to previous month (or next month on RTL page).</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputDate
 */

 /**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Left</kbd></td>
 *       <td>Switch to next month (or previous month on RTL page).</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Right</kbd></td>
 *       <td>Switch to previous month (or next month on RTL page).</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojDatePicker
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>DownArrow or UpArrow</kbd></td>
 *       <td>Shows the calender grid and moves the focus into the expanded grid</td>
 *     </tr>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Close the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>Tab In</kbd></td>
 *       <td>Set focus to the input. If hints, title or messages exist in a notewindow,
 *        pop up the notewindow.</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Select the currently focused day</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move up in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move down in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move right in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move left in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Close the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Home</kbd></td>
 *       <td>Move focus to first day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>End</kbd></td>
 *       <td>Move focus to last day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageUp</kbd></td>
 *       <td>Switch to previous month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageDown</kbd></td>
 *       <td>Switch to next month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageUp</kbd></td>
 *       <td>Switch to previous year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageDown</kbd></td>
 *       <td>Switch to next year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageUp</kbd></td>
 *       <td>Switch to previous by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageDown</kbd></td>
 *       <td>Switch to next by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + T</kbd></td>
 *       <td>Places focus on Today button if it exists.</tr>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputDate
 */

 /**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Select the currently focused day</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move up in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move down in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move right in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move left in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Close the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Home</kbd></td>
 *       <td>Move focus to first day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>End</kbd></td>
 *       <td>Move focus to last day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageUp</kbd></td>
 *       <td>Switch to previous month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageDown</kbd></td>
 *       <td>Switch to next month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageUp</kbd></td>
 *       <td>Switch to previous year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageDown</kbd></td>
 *       <td>Switch to next year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageUp</kbd></td>
 *       <td>Switch to previous by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageDown</kbd></td>
 *       <td>Switch to next by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + T</kbd></td>
 *       <td>Places focus on Today button if it exists.</tr>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojDatePicker
 */


//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the InputDate and InputDateTime input element. Note that if element is inline for
 * InputDate it would return null whereas InputDateTime would return the input element of the internally created
 * InputTime element.
 *
 * @ojsubid oj-inputdatetime-date-input
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the node for the input element:</caption>
 * var node = myInputDate.getNodeBySubId( {'subId': 'oj-inputdatetime-date-input'} );
 */

/**
 * <p>Sub-ID for the calendar drop down node.
 *
 * @ojsubid oj-datepicker-content
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the calendar drop down node:</caption>
 * // myInputElement is InputDate or InputDateTime.
 * var node = myInputElement.getNodeBySubId( {'subId': 'oj-datepicker-content'} );
 */

/**
 * <p>Sub-ID for the calendar icon that triggers the calendar drop down.
 *
 * @ojsubid oj-inputdatetime-calendar-icon
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the calendar icon that triggers the calendar drop down:</caption>
 * // myInputElement is InputDate or InputDateTime.
 * var node = myInputElement.getNodeBySubId( {'subId': 'oj-inputdatetime-calendar-icon'} );
 */

/**
 * <p>Sub-ID for the previous month icon.
 *
 * @ojsubid oj-datepicker-prev-icon
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the previous month icon:</caption>
 * // myInputElement is InputDate or InputDateTime.
 * var node = myInputElement.getNodeBySubId( {'subId': 'oj-datepicker-prev-icon'} );
 */

/**
 * <p>Sub-ID for the next month icon.
 *
 * @ojsubid oj-datepicker-next-icon
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the next month icon:</caption>
 * // myInputElement is InputDate or InputDateTime.
 * var node = myInputElement.getNodeBySubId( {'subId': 'oj-datepicker-next-icon'} );
 */

/**
 * <p>Sub-ID for the month span or select element.
 *
 * @ojsubid oj-datepicker-month
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the month span or select element:</caption>
 * // myInputElement is InputDate or InputDateTime.
 * var node = myInputElement.getNodeBySubId( {'subId': 'oj-datepicker-month'} );
 */

/**
 * <p>Sub-ID for the year span or select element.
 *
 * @ojsubid oj-datepicker-year
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the year span or select element:</caption>
 * // myInputElement is InputDate or InputDateTime.
 * var node = myInputElement.getNodeBySubId( {'subId': 'oj-datepicker-year'} );
 */

/**
 * <p>Sub-ID for the current/today button for button bar.
 *
 * @ojsubid oj-datepicker-current
 * @memberof oj.ojInputDate
 *
 * @example <caption>Get the current/today button for button bar:</caption>
 * // myInputElement is InputDate or InputDateTime.
 * var node = myInputElement.getNodeBySubId( {'subId': 'oj-datepicker-current'} );
 */

/**
 * Input type for the dayFormatter callback function
 * @typedef {object} oj.ojInputDate.DayFormatterInput
 * @property {number} fullYear
 * @property {number} month
 * @property {number} date
 */
/**
 * Output type for the dayFormatter callback function
 * @typedef {object} oj.ojInputDate.DayFormatterOutput
 * @property {boolean=} disabled
 * @property {string=} className
 * @property {string=} tooltip
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Helper function to split the timeIncrement into its constituents and returns the split object.
 * Used in ojInputTime and ojInputDateTime
 *
 * @ignore
 */
function splitTimeIncrement(timeIncrement)
{
  var splitIncrement = timeIncrement.split(":");

  if (splitIncrement.length !== 4)
  {
    throw new Error("timeIncrement value should be in the format of hh:mm:ss:SSS");
  }

  var increments =
  {
    hourIncr : parseInt(splitIncrement[0].substring(0), 10),
    minuteIncr : parseInt(splitIncrement[1], 10),
    secondIncr : parseInt(splitIncrement[2], 10),
    millisecondIncr : parseInt(splitIncrement[3], 10)
  };

  var sum = 0;
  for(var key in increments) {
    sum += increments[key];
  }

  if(sum === 0) {
    throw new Error("timeIncrement must have a non 00:00:00:000 value");
  }

  return increments;
}

/**
 * Helper function to create a timepicker converter
 *
 * @ignore
 * @param {Object} converter
 * @param {Object=} addOpts
 * @return {Object}
 */
function _getTimePickerConverter(converter, addOpts) {
  var resolvedOptions = converter.resolvedOptions();
  var options = { };
  var params = ["hour", "hour12", "minute", "second", "millisecond", "timeFormat",
      "timeZone", "timeZoneName", "isoStrFormat", "dst"], i, j;

  for (i = 0, j = params.length;i < j;i++)
  {
    if (params[i] in resolvedOptions)
    {
      if(params[i] === "timeFormat") {
        //special case for timeFormat, formatType of time must be added
        options["formatType"] = "time";
      }
      options[params[i]] = resolvedOptions[params[i]];
    }
  }

  var testEmpty = {};
  $.extend(testEmpty, options);
  delete testEmpty["isoStrFormat"]; //since below two are provided even with only date portion
  delete testEmpty["dst"];
  
  if ($.isEmptyObject(testEmpty))
  {
    throw new Error("Empty object for creating a time picker converter");
  }

  $.extend(options, addOpts || {});
  return oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter(options);;
}

/**
 * @ojcomponent oj.ojInputTime
 * @augments oj.inputBase
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojInputTime extends inputBase<string, ojInputTimeSettableProperties>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojInputTimeSettableProperties extends inputBaseSettableProperties<string>",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @since 0.6
 * @ojstatus preview
 * @ojshortdesc Provides basic support for specifying a time value.
 * @ojrole combobox
 *
 * @classdesc
 * <h3 id="inputTimeOverview-section">
 *   JET InputTime
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputTimeOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET InputTime provides a simple time selection drop down.</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-input-time>&lt;/oj-input-time></code></pre>
 *
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 */
oj.__registerWidget("oj.ojInputTime", $['oj']['inputBase'],
{
  widgetEventPrefix : "oj",

  //-------------------------------------From base---------------------------------------------------//
  _CLASS_NAMES : "oj-inputdatetime-input",
  _WIDGET_CLASS_NAMES : "oj-inputdatetime-time-only oj-component oj-inputdatetime",
  _INPUT_CONTAINER_CLASS : "oj-inputdatetime-input-container",
  _ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES : "",
  _INPUT_HELPER_KEY: "inputHelp",
  _ATTR_CHECK : [{"attr": "type", "setMandatory": "text"}],
  _GET_INIT_OPTIONS_PROPS_FOR_WIDGET:  [{attribute: "disabled", validateOption: true},
                             {attribute: 'pattern'},
                             {attribute: "title"},
                             {attribute: "placeholder"},
                             {attribute: "value", coerceDomValue: coerceIsoString},
                             {attribute: "required",
                              coerceDomValue: true, validateOption: true},
                             {attribute: 'readonly', option: 'readOnly',
                             validateOption: true},
                             {attribute: "min", coerceDomValue: coerceIsoString},
                             {attribute: "max", coerceDomValue: coerceIsoString}],
  //-------------------------------------End from base-----------------------------------------------//

  _TIME_PICKER_ID : "ojInputTime",
  _TRIGGER_CLASS : "oj-inputdatetime-input-trigger",
  _TRIGGER_TIME_CLASS : "oj-inputdatetime-time-icon",

  _ON_CLOSE_REASON_SELECTION: "selection",  // A selection was made
  _ON_CLOSE_REASON_CANCELLED: "cancelled",  // Selection not made
  _ON_CLOSE_REASON_TAB: "tab",              // Tab key
  _ON_CLOSE_REASON_CLOSE: "close",          // Disable or other closes

  _KEYBOARD_EDIT_OPTION_ENABLED: "enabled",
  _KEYBOARD_EDIT_OPTION_DISABLED: "disabled",

  options :
  {
    /**
     * Default converter for InputTime
     *
     * If one wishes to provide a custom converter for the InputTime override the factory returned for
     * oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME)
     * {@ojinclude "name":"inputBaseConverterOptionDoc"}
     * @expose
     * @memberof! oj.ojInputTime
     * @instance
     * @type {Object}
     * @ojsignature  { target: "Type", value: "oj.Converter<string>|oj.Validation.FactoryRegisteredValidatorOrConverter|null"}
     * @default oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({"hour": "2-digit", "minute": "2-digit"})
     */
    converter : oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter(
    {
      "hour" : "2-digit", "minute" : "2-digit"
    }),

    /**
     * Determines if keyboard entry of the text is allowed.
     * When disabled the picker must be used to select a time.
     * 
     * Default value depends on the theme. In alta-android, alta-ios and alta-windows themes, the
     * default is <code class="prettyprint">"disabled"</code> and
     * it's <code class="prettyprint">"enabled"</code> for alta web theme.
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputTime
     * @type {string}
     * @ojvalue {string} "enabled"  Allow keyboard entry of the time.
     * @ojvalue {string} "disabled" Changing the time can only be done with the picker.
     *
     * @example <caption>Initialize the InputTime with the <code class="prettyprint">keyboard-edit</code> attribute specified:</caption>
     * &lt;oj-input-time keyboard-edit='disabled'>&lt;/oj-input-time>
     *
     * @example <caption>Get or set the <code class="prettyprint">keyboardEdit</code> property after initialization:</caption>
     * // getter
     * var keyboardEdit = myInputTime.keyboardEdit;
     *
     * // setter
     * myInputTime.keyboardEdit = 'disabled';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeKeyboardEditOptionDefault: disabled !default;
     */
    keyboardEdit : "enabled",

    /**
     * The maximum selectable time. When set to null, there is no maximum.
     *
     * <ul>
     *  <li> type string - ISOString
     *  <li> null - no limit
     * </ul>
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputTime
     * @type {string|null}
     * @default null
     *
     * @example <caption>Initialize the InputTime with the <code class="prettyprint">max</code> attribute specified:</caption>
     * &lt;oj-input-time max='T13:30:00.000-08:00'>&lt;/oj-input-time>
     *
     * @example <caption>Get or set the <code class="prettyprint">max</code> property after initialization:</caption>
     * // getter
     * var maxValue = myInputTime.max;
     *
     * // setter
     * myInputTime.max = 'T13:30:00.000-08:00';
     */
    max : undefined,

    /**
     * The minimum selectable time. When set to null, there is no minimum.
     *
     * <ul>
     *  <li> type string - ISOString
     *  <li> null - no limit
     * </ul>
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputTime
     * @type {string|null}
     * @default null
     *
     * @example <caption>Initialize the InputTime with the <code class="prettyprint">min</code> attribute specified:</caption>
     * &lt;oj-input-time min='T08:00:00.000-08:00'>&lt;/oj-input-time>
     *
     * @example <caption>Get or set the <code class="prettyprint">min</code> property after initialization:</caption>
     * // getter
     * var minValue = myInputTime.min;
     *
     * // setter
     * myInputTime.min = 'T08:00:00.000-08:00';
     */
    min : undefined,

    /**
     * JSON data passed when the widget is of ojInputDateTime
     *
     * {
     *  widget : dateTimePickerInstance,
     *  inline: true|false
     * }
     *
     * @expose
     * @memberof! oj.ojInputTime
     * @instance
     * @private
     */
    datePickerComp : null,

    /**
     * <p>Attributes specified here will be set on the picker DOM element when it's launched.
     * <p>The supported attributes are <code class="prettyprint">class</code> and <code class="prettyprint">style</code>, which are appended to the picker's class and style, if any.
     * Note: 1) pickerAttributes is not applied in the native theme.
     * 2) setting this property after element creation has no effect.
     * 
     * @property {string=} style
     * @property {string=} class
     *
     * @example <caption>Initialize the inputTime specifying a set of attributes to be set on the picker DOM element:</caption>
     * myInputTime.pickerAttributes = {
     *   "style": "color:blue;",
     *   "class": "my-class"
     * };
     *
     * @example <caption>Get the <code class="prettyprint">pickerAttributes</code> property, after initialization:</caption>
     * // getter
     * var pickerAttrs = myInputTime.pickerAttributes;
     *
     * @expose
     * @memberof! oj.ojInputTime
     * @instance
     * @type {?Object}
     * @default null
     */
    pickerAttributes: null,

    /**
     * Allows applications to specify whether to render time picker in JET or
     * as a native picker control.</br>
     * 
     * Default value depends on the theme. In alta-android, alta-ios and alta-windows themes, the
     * default is "native" and it's "jet" for alta web theme.
     *
     * @expose
     * @memberof! oj.ojInputTime
     * @instance
     * @type {string}
     * @ojvalue {string} 'jet' Applications get full JET functionality.
     * @ojvalue {string} 'native' Applications get the functionality of the native picker.</br></br>
     *  Note that the native picker support is limited to Cordova plugin published
     *  at 'https://github.com/VitaliiBlagodir/cordova-plugin-datepicker'.</br></br>
     *  With native renderMode, the functionality that is sacrificed compared to jet renderMode are:
     *    <ul>
     *      <li>Time picker cannot be themed</li>
     *      <li>Accessibility is limited to what the native picker supports</li>
     *      <li>pickerAttributes is not applied</li>
     *      <li>Sub-IDs are not available</li>
     *      <li>hide() function is no-op</li>
     *      <li>translations sub properties pertaining to the picker is not available</li>
     *      <li>'timePicker.timeIncrement' property is limited to iOS and will only take a precision of minutes</li>
     *    </ul>
     *
     * @example <caption>Initialize the InputTime with the <code class="prettyprint">render-mode</code> attribute specified:</caption>
     * &lt;oj-input-time render-mode='native'>&lt;/oj-input-time>
     *
     * @example <caption>Get or set the <code class="prettyprint">renderMode</code> property after initialization:</caption>
     * // getter
     * var renderMode = myInputTime.renderMode;
     *
     * // setter
     * myInputTime.renderMode = 'native';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeRenderModeOptionDefault: native !default;
     */
    renderMode : "jet",

    /**
     * Note that Jet framework prohibits setting subset of properties which are object types.<br/><br/>
     * For example myInputTime.timePicker = {timeIncrement: "00:30:00:00"}; is prohibited as it will
     * wipe out all other sub-properties for "timePicker" object.<br/><br/> If one wishes to do this [by above syntax or knockout] one
     * will have to get the "timePicker" object, modify the necessary sub-property and pass it to above syntax.<br/><br/>
     * Note that when renderMode is 'native', the only timePicker sub-properties available are showOn and, to a limited extent, timeIncrement.<br/><br/>
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputTime
     * @type {Object}
     *
     * @example <caption>Initialize the component, overriding some time-picker attributes and leaving the others intact:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-input-time time-picker.time-increment='00:10:00:00'>&lt;/oj-input-time>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-input-time time-picker='{"timeIncrement":"00:10:00:00"}'>&lt;/oj-input-time>
     *
     * @example <caption>Get or set the <code class="prettyprint">timePicker</code> property after initialization:</caption>
     * // Get one
     * var value = myComponent.timePicker.showOn;
     *
     * // Set one, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myComponent.setProperty('timePicker.showOn', 'image');
     *
     * // Get all
     * var values = myComponent.timePicker;
     *
     * // Set all.  Must list every timePicker key, as those not listed are lost.
     * myComponent.timePicker = {
     *     timeIncrement: '00:10:00:00',
     *     showOn: 'image',
     *     footerLayout: 'now'
     * };
     */
    timePicker:
    {
      /**
       * Will dictate what content is shown within the footer of the wheel timepicker.
       *
       * <p>See the <a href="#timePicker">time-picker</a> attribute for usage examples.
       *
       * @expose
       * @name timePicker.footerLayout
       * @memberof! oj.ojInputTime
       * @instance
       * @type {string}
       * @ojvalue {string} '' Do not show anything
       * @ojvalue {string} 'now' Show the now button
       * @default ""
       */
      footerLayout : "",

      /**
       * Time increment to be used for InputTime, the format is hh:mm:ss:SS. <br/><br/>
       * Note that when renderMode is 'native', timeIncrement property is limited to iOS and will only take a precision of minutes.<br/><br/> 
       *
       * <p>See the <a href="#timePicker">time-picker</a> attribute for usage examples.
       *
       * @expose
       * @name timePicker.timeIncrement
       * @memberof! oj.ojInputTime
       * @instance
       * @type {string}
       * @default "00:05:00:00"
       */
      timeIncrement : "00:05:00:00",

      /**
       * When the timepicker should be shown.
       *
       * <p>See the <a href="#timePicker">time-picker</a> attribute for usage examples.
       *
       * @expose
       * @name timePicker.showOn
       * @memberof! oj.ojInputTime
       * @instance
       * @type {string}
       * @ojvalue {string} 'focus' when the element receives focus or when the trigger clock image is clicked. When the picker is closed, the field regains focus and is editable.
       * @ojvalue {string} 'image' when the trigger clock image is clicked
       * @default "focus"
       */
      showOn : "focus"
    }

    // DOCLETS

    /**
     * List of validators used by element when performing validation. Each item is either an
     * instance that duck types {@link oj.Validator}, or is an Object literal containing the
     * properties listed below. Implicit validators created by an element when certain attributes
     * are present (e.g. <code class="prettyprint">required</code> attribute), are separate from
     * validators specified through this attribute. At runtime when the element runs validation, it
     * combines the implicit validators with the list specified through this attribute.
     * <p>
     * Hints exposed by validators are shown in the notewindow by default, or as determined by the
     * 'validatorHint' property set on the <code class="prettyprint">displayOptions</code>
     * property.
     * </p>
     *
     * <p>
     * When <code class="prettyprint">validators</code> property changes due to programmatic
     * intervention, the element may decide to clear messages and run validation, based on the
     * current state it is in. </br>
     *
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>The cached list of validator instances are cleared and new validator hints is pushed to
     * messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if element is valid when validators changes, element does nothing other than the
     * steps it always performs.</li>
     * <li>if element is invalid and is showing messages when
     * <code class="prettyprint">validators</code> changes then all element messages are cleared
     * and full validation run using the display value on the element.
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code>
     *   property is not updated and the error is shown.
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code>
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code>
     *   event to clear custom errors.</li>
     * </ul>
     * </li>
     * <li>if element is invalid and has deferred messages when validators changes, it does
     * nothing other than the steps it performs always.</li>
     * </ul>
     * </p>
     *
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the element are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * </p>
     *
     * @property {string} type - the validator type that has a {@link oj.ValidatorFactory} that can
     * be retrieved using the {@link oj.Validation} module. For a list of supported validators refer
     * to {@link oj.ValidatorFactory}. <br/>
     * @property {Object=} options - optional Object literal of options that the validator expects.
     *
     * @example <caption>Initialize the element with validator object literal:</caption>
     * myInputTime.validators = [{
     *     type: 'dateTimeRange',
     *     options : {
     *       max: 'T14:30:00',
     *       min: 'T02:30:00'
     *     }
     *   }];
     *
     * NOTE: oj.Validation.validatorFactory('dateTimeRange') returns the validator factory that is used
     * to instantiate a range validator for dateTime.
     *
     * @example <caption>Initialize the element with multiple validator instances:</caption>
     * var validator1 = new MyCustomValidator({'foo': 'A'});
     * var validator2 = new MyCustomValidator({'foo': 'B'});
     * myInputTime.value = 10;
     * myInputTime.validators = [validator1, validator2];
     *
     * @expose
     * @name validators
     * @instance
     * @memberof oj.ojInputTime
     * @ojsignature  { target: "Type", value: "Array<oj.Validator<string>|oj.Validation.FactoryRegisteredValidatorOrConverter>|null"}
     * @type {Array|undefined}
     */

    /**
     * The value of the InputTime which should be an ISOString.
     *
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> attribute:</caption>
     * &lt;oj-input-time value='T10:30:00.000'&gt;&lt;/oj-input-time&gt;
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> property specified programmatically
     * using oj.IntlConverterUtils.dateToLocalIso :</caption>
     * myInputTime.value = oj.IntlConverterUtils.dateToLocalIso(new Date());<br/>
     * @example <caption>Get or set the <code class="prettyprint">value</code> property, after initialization:</caption>
     * // Getter: returns Today's date in ISOString
     * myInputTime.value;
     * // Setter: sets it to a different date
     * myInputTime.value = "T20:00:00-08:00";
     *
     * @expose
     * @name value
     * @type {string}
     * @instance
     * @ojwriteback
     * @memberof! oj.ojInputTime
     */

    // Events

  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputTime
   */
  _InitOptions: function(originalDefaults, constructorOptions)
  {
    this._super(originalDefaults, constructorOptions);
    //when it is of ojInputDateTime component, do not initialize values from dom node since it's an empty input node if inline or
    //if not inline the values should be taken care of by ojInputDateTime. Note that option values would have been passed by
    //ojInputDateTime
    if(this.options["datePickerComp"] === null && !this._IsCustomElement())
    {
      oj.EditableValueUtils.initializeOptionsFromDom(this._GET_INIT_OPTIONS_PROPS_FOR_WIDGET, constructorOptions, this);
    }
  },

  _getPrependNode: function() 
  {
    return this._isIndependentInput() ? $("body") : $(".oj-popup-content", this._datePickerComp.widget._popUpDpDiv.ojPopup("widget"));
  },

  /**
   * @ignore
   */
  _InitBase : function()
  {
    this._timePickerDefaultValidators = {};
    this._datePickerComp = this.options["datePickerComp"];
    this._inputContainer = null;
    this._redirectFocusToInputContainer = false;
    this._isMobile = false;

    //only case is when of showOn of focus and one hides the element [need to avoid showing]
    this._ignoreShow = false;

    // need this flag to keep track of native picker opened, there is no callback on native API
    //  to find out otherwise.
    this._nativePickerShowing = false;

    var pickerAttrs = this.options.pickerAttributes;

    var self = this;
    this._wheelPicker = $('<div class="oj-timepicker-popup" style="display:none"><div id="' + this._GetSubId(this._TIME_PICKER_ID) +
                          '" class="oj-timepicker-content"></div></div>');
    this._getPrependNode().prepend(this._wheelPicker); //@HTMLUpdateOK

    if(this._isIndependentInput()) 
    {
      //DISABLE FOR NOW, as animation is coming quite clunky (not sure if the css of popup or of animation)
      //var animation = _isLargeScreen ? {"open": null, "close": null} : {"close": null};
      var animation = {"open": null, "close": null};

      this._popUpWheelPicker = this._wheelPicker.ojPopup(
      {
        "initialFocus": "none",
        "role": "dialog",
        "chrome": "default",
        "modality": _isLargeScreen ? "modeless" : "modal",
        "open": function () {
        },
        "beforeClose": function () {
        },
        "animateStart": function (event, ui)
        {
          if ('open' === ui.action)
          {
            event.preventDefault();
            oj.AnimationUtils.slideIn(ui.element, {"offsetY": ui.element.offsetHeight + "px"}).then(ui.endCallback);
          }
        },
        "animation": animation
      }).attr('data-oj-internal', ''); // mark internal component, used in oj.Components.getComponentElementByNode;
      this.element.attr('data-oj-popup-' + this._popUpWheelPicker.attr('id') + '-parent', ''); // mark parent of pop up

      if (pickerAttrs)
        oj.EditableValueUtils.setPickerAttributes(this._popUpWheelPicker.ojPopup("widget"), pickerAttrs);
    }
    // I want to wrap the inputTime if it is all by itself, or if it is
    // part of the inline inputDateTime component which is the inline date stacked on top of an
    // inputTime. The inline error messages will go under the inputTime part. TODO: how?
    // right now the destroy fails because I am whacking away something.. the dom.
    if (this._isIndependentInput())
      this._ELEMENT_TRIGGER_WRAPPER_CLASS_NAMES += this._INPUT_CONTAINER_CLASS;
  },

  _timepickerShowing: function ()
  {
    if(this._isIndependentInput())
    {
      var picker = this._popUpWheelPicker;
      return oj.Components.isComponentInitialized(picker, "ojPopup") && picker.ojPopup("isOpen") || this._nativePickerShowing;
    }
    else
    {
      var widget = this._datePickerComp.widget;
      var picker = widget._popUpDpDiv;
      return (widget._isShowingDatePickerSwitcher() && oj.Components.isComponentInitialized(picker, "ojPopup") && picker.ojPopup("isOpen")) || this._nativePickerShowing;
    }
    
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputTime
   */
  _ComponentCreate : function()
  {
    this._InitBase();

    var ret = this._superApply(arguments);

    if (this._isContainedInDateTimePicker() && !this._isDatePickerInline())
    {
      //set to nothing since then of not inline and don't want to place two component classes to
      //the same input element
      this._CLASS_NAMES = "";
    }
    else
    {
      //if isoString has a different timezone then the one provided in the converter, need to perform
      //conversion so pass it through the method
      if(this.options["value"])
      {
        var formatted = this._GetConverter()["format"](this.options["value"]);
        this._SetValue(formatted, {});
      }

      // active state handler, only in case time picker is independent
      bindActive(this);
    }

    this._processReadOnlyKeyboardEdit();

    if(this._isIndependentInput()) {
      this._attachTrigger();
    }
    
    return ret;
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputTime
   */
  _AfterCreate : function ()
  {
    var ret = this._superApply(arguments);

    if(this._isIndependentInput()) {
      disableEnableSpan(this._triggerNode.children(), this.options["disabled"]);

      if (!this._IsCustomElement()) {
        var label = this.$label;
        if (this._inputContainer && label && label.length === 1) {
          var icId = this._inputContainer.attr("id");
          var LId = label.attr("id");

          if (!LId) {
            LId = this["uuid"] + "_Label";
            label.attr("id", LId);
          }

          this._inputContainer.attr("aria-labelledby", LId);
        }
      }
      else {
        if (this._inputContainer) {
          var defaultLabelId = this["uuid"] + "_Label";
          var LId = oj.EditableValueUtils.getOjLabelId(this.widget(), defaultLabelId);
          if (LId)
            this._inputContainer.attr("aria-labelledby", LId);
        }

      }
    }

    return ret;
  },

  /**
   * @ignore
   * @protected
   * @override
   */
  _setOption : function (key, value, flags)
  {
    var retVal = null;
    var footer;

    //When a null, undefined, or "" value is passed in set to null for consistency
    //note that if they pass in 0 it will also set it null
    if (key === "value")
    {
      if(!value)
      {
        value = null;
      }

      retVal = this._super(key, value, flags);

      this._createWheelPickerDom(true);
      
      return retVal;
    }

    //set options on the timePicker wipe out the footerLayout settings
    //save footerLayout and restore after superApply
    else if (key === "timePicker" && value["footerLayout"] === undefined)
    {
      footer = this.options["timePicker"]["footerLayout"];
    }

    retVal = this._superApply(arguments);

    //restore footerLayout
    if (footer)
      this.options["timePicker"]["footerLayout"] = footer;

    if(key === "disabled" && this._isIndependentInput())
    {
      if(value)
      {
        this._hide(this._ON_CLOSE_REASON_CLOSE);
      }
      this._triggerNode.find("." + this._TRIGGER_TIME_CLASS).attr("title", this._getTimeTitle());
      disableEnableSpan(this._triggerNode.children(), value);
    }
    else if ((key === "max" || key === "min") && !this._isContainedInDateTimePicker())
    {
      //since validators are immutable, they will contain min + max as local values. B/c of this will need to recreate

      this._timePickerDefaultValidators[oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE] =
        getImplicitDateTimeRangeValidator(this.options, this._GetConverter(), this._GetDefaultStyleClass());
      this._AfterSetOptionValidators();
    }
    else if(key === "readOnly")
    {
      this._processReadOnlyKeyboardEdit();

      if (value)
      {
        this._hide(this._ON_CLOSE_REASON_CLOSE);
      }
      this._AfterSetOptionDisabledReadOnly("readOnly", oj.EditableValueUtils.readOnlyOptionOptions);
    }
    else if(key === "keyboardEdit")
    {
      this._processReadOnlyKeyboardEdit();
    }

    var redrawTimePicker = {"max": true, "min": true, "converter": true, "timePicker": true};
    if(key in redrawTimePicker)
    {
      this._createWheelPickerDom();
    }

    return retVal;
  },

  /**
   * @ignore
   * @protected
   * @override
   */
  _destroy : function ()
  {
    this._cleanUpTimeResources();
    var retVal = this._super();
    return retVal;
  },
  /**
   * @ignore
   * @private
   * @override
   */
  _clearWheelModels : function ()
  {
    var timePickerModel = this._timePickerModel;
	if (!timePickerModel) return;
	var model = timePickerModel.getWheelModel("hour");
	if (model) model.wheel = undefined;
	model = timePickerModel.getWheelModel("minute");
	if (model) model.wheel = undefined;
	model = timePickerModel.getWheelModel("ampm");
	if (model) model.wheel = undefined;
  },
  /**
   * @ignore
   * @protected
   * @override
   */
  _destroyHammer : function ()
 
  {
    if (this._wheelGroup && this._wheelGroup.length) {
      var group = this._wheelGroup[0];
      var wheel;
      var children = group.children;
      for (var index = 0; index < children.length; index++)
      {
        wheel = children[index];
        wheel.ojDestroy();
      }
    }
  },
  /**
   * @ignore
   * @protected
   * @override
   */
  _SetupResources : function ()
  {
    this._resizePopupBind = function ()
    {
      $('.oj-timepicker-content', this._wheelPicker)[_isLargeScreen ? 'removeClass' : 'addClass']('oj-timepicker-fixedheight');
      if (this._isIndependentInput() && oj.Components.isComponentInitialized(this._popUpWheelPicker, "ojPopup"))
      {
        this._popUpWheelPicker.ojPopup("option", "modality", (_isLargeScreen ? "modeless" : "modal"));
      }
    }
.bind(this);
    window.addEventListener('resize', this._resizePopupBind, false);
    return this._super();
  },
  /**
   * @ignore
   * @protected
   * @override
   */
  _ReleaseResources : function ()
  {
    //for non custom element, this is done as first step in destroy
    if (this._IsCustomElement())
      this._cleanUpListeners();
    return this._super();
  },
  /**
   * @ignore
   * @private
   */
  _cleanUpTimeResources : function ()
  {
    if (this._isIndependentInput())
    {
      this.element.off("focus touchstart");
      this._wrapper.off("touchstart");
    }

    if (this._triggerNode)
    {
      this._triggerNode.remove();
    }

    this._cleanUpListeners();
    this._destroyHammer();
    if (this._wheelPicker)
    {
      this._wheelPicker.remove();
      this._clearWheelModels();
    }

    if (this._isIndependentInput() && oj.Components.isComponentInitialized(this._popUpWheelPicker, "ojPopup"))
    {
      this._popUpWheelPicker.ojPopup("destroy");
    }
  },
  /**
   * @ignore
   * @private
   */
  _cleanUpListeners : function ()
  {
    if (this._resizePopupBind)
    {
      window.removeEventListener('resize', this._resizePopupBind);
    }
    window.removeEventListener('resize', resizeLargeScreenChange);
  },
  /**
   * @ignore
   */
  _processReadOnlyKeyboardEdit: function()
  {
    if(this._isIndependentInput()) 
    {
      var readonly = this.options["readOnly"] ||
            this._isKeyboardEditDisabled();
      this.element.prop("readOnly", !!readonly);
    }
  },

  /**
   * @ignore
   * @return {boolean}
   */
  _isKeyboardEditDisabled: function()
  {
    return this.options["keyboardEdit"] === this._KEYBOARD_EDIT_OPTION_DISABLED;
  },

  /**
   * Invoke super only if it is standlone or if it is part of ojInputDateTime and ojInputDateTime is inline
   *
   * @ignore
   * @protected
   * @override
   */
  _AppendInputHelper : function ()
  {
    if (this._isIndependentInput())
    {
      this._superApply(arguments);
    }
  },

  /**
   * Only time to have ojInputTime handle the display of timepicker by keyDown is when datePickerComp reference is null or
   * when it is not null and is inline
   *
   * @ignore
   * @protected
   * @override
   * @param {Event} event
   */
  _onKeyDownHandler : function (event)
  {
    if(this._isIndependentInput())
    {
      this._superApply(arguments);

      var kc = $.ui.keyCode;
      var handled = false;

      if (this._timepickerShowing())
      {
        switch (event.keyCode)
        {
          case kc.TAB: ;
            this._hide(this._ON_CLOSE_REASON_TAB);
            break;
          case kc.ESCAPE:
            this._hide(this._ON_CLOSE_REASON_CANCELLED);
            handled = true;
            break;
          case kc.UP: ;
          case kc.DOWN:
            this._wheelGroup.children().first().focus();
            handled = true;
            break;
          default:;
        }
      }
      else
      {
        switch (event.keyCode)
        {
          case kc.UP: ;
          case kc.DOWN:
            this._SetValue(this._GetDisplayValue(), event);
            this.show();
            handled = true;
            break;
          default:;
        }
      }

      if (handled || event.keyCode === kc.ENTER)
      {
        event.preventDefault();
        event.stopPropagation();
        return false;
      }
    }
  },

  _getTimeTitle: function ()
  {
    return this._EscapeXSS(this.getTranslatedString("tooltipTime" + (this.options["disabled"] ? "Disabled" : "")));
  },

  /**
   * @protected
   * @override
   * @ignore
   */
  _WrapElement: function()
  {
    this._inputContainer = this._superApply(arguments);
    this._inputContainer.attr({"role": "combobox", "aria-haspopup": "true", "tabindex": "-1"});
  },

  /**
   * When input element has focus
   * @private
   */
  _onElementFocus : function()
  {
    var showOn = this.options["timePicker"]["showOn"];

    if(this._redirectFocusToInputContainer)
    {
      this._redirectFocusToInputContainer = false;
      if (!isPickerNative(this))
      {
        this._wheelGroup.children().first().focus();
      }
      else
      {
        this._inputContainer.focus();
      }
    }
    else
    {
      if (showOn === "focus")
      {
        // pop-up date picker when focus placed on the input box
        this.show();
      }
      else
      {
        if(this._timepickerShowing())
        {
          this._hide(this._ON_CLOSE_REASON_CLOSE);
        }
      }
    }
  },

  /**
   * When input element is touched
   *
   * @ignore
   * @protected
   */
  _OnElementTouchStart : function()
  {
    var showOn = this.options["timePicker"]["showOn"];

    // If the focus is already on the text box and can't edit with keyboard
    // and show on is focus then reopen the picker.
    if(showOn === "focus")
    {
      if (this._timepickerShowing())
      {
        this._ignoreShow = true;
        this._hide(this._ON_CLOSE_REASON_CLOSE);
      }
      else
      {
        var inputActive = this.element[0] === document.activeElement;

        this.show();
        this._redirectFocusToInputContainer = true;

        // Don't change focus on wheel picker since it should have acquired focus
        if (isPickerNative(this))
        {
          if(inputActive)
          {
            this._inputContainer.focus();
          }
        }
      }
    }
  },

  /**
   * This function will create the necessary time trigger container [i.e. image to launch the time drop down]
   * and perform any attachment to events
   *
   * @private
   */
  _attachTrigger : function ()
  {
    var showOn = this.options["timePicker"]["showOn"];
    var triggerContainer = $("<span>").addClass(this._TRIGGER_CLASS);
    var triggerTime = $("<span title='" + this._getTimeTitle() + "'/>").addClass(this._TRIGGER_TIME_CLASS + " oj-clickable-icon-nocontext oj-component-icon");

    var self = this;

    this.element.on("focus", $.proxy(this._onElementFocus, this));
    this.element.on("touchstart", $.proxy(this._OnElementTouchStart, this));
    
    var wrapper = this._wrapper;
    wrapper.on("touchstart", function(e)
    {
      self._isMobile = true;
    });

    if (showOn === "image")
    {
      // we need to show the icon that we hid by display:none in the mobile themes
      triggerTime.css("display", "block");

      // In iOS theme, we defaulted to use border radius given that showOn=focus is default and
      //  we will not have trigger icon. For showOn=image case, we will show the icon, so
      //  we need to remove the border radius. iOS is the only case we use border radius, so this
      //  setting for all cases is fine.
      if (this._IsRTL())
      {
        this.element.css("border-top-left-radius", 0);
        this.element.css("border-bottom-left-radius", 0);
      }
      else
      {
        this.element.css("border-top-right-radius", 0);
        this.element.css("border-bottom-right-radius", 0);
      }
    }

    // do not attach time picker icon if not independent input mode and native picker is in use
    //  - this is because the native pickers let pick both date and time, showing one icon is
    //  sufficient and less clutter hence
    if (!isPickerNative(this))
    {
      triggerContainer.append(triggerTime); //@HTMLUpdateOK

      triggerTime.on("click", function ()
      {
        if (self._timepickerShowing())
        {
          self._hide(self._ON_CLOSE_REASON_CLOSE);
        }
        else
        {
          self.show();
          self._wheelGroup.children().first().focus();        }
      });

      this._AddHoverable(triggerTime);
      this._AddActiveable(triggerTime);

      this._triggerIcon = triggerTime;
    }

    this._triggerNode = triggerContainer;

    this.element.after(triggerContainer); //@HTMLUpdateOK
  },

  /**
   * Returns a boolean of whether the date is in the min + max range
   *
   * @private
   */
  _notInMinMaxRange : function (dateIso, minDateIso, maxDateIso)
  {
    var converter = this._GetConverter();

    return ((minDateIso && converter.compareISODates(dateIso, minDateIso) < 0)
          || (maxDateIso && converter.compareISODates(dateIso, maxDateIso) > 0));
  },

  _getValue : function ()
  {
    //need to use ojInputDateTime's value when created internally [i.e. for min + max and etc].
    return this._isContainedInDateTimePicker() ? this._datePickerComp["widget"].getValueForInputTime() : this.options["value"];
  },

  /**
   * Invoked when blur is triggered of the this.element
   *
   * @ignore
   * @protected
   * @param {Event} event
   */
  _onBlurHandler : function (event)
  {
    if(this._isIndependentInput())
    {
      this._superApply(arguments);
    }
  },

  /**
   * Shows the timepicker
   *
   * @expose
   * @instance
   * @return {void}
   * @memberof! oj.ojInputTime
   */
  show : function ()
  {
    if (this._timepickerShowing() || this.options["disabled"] || this.options["readOnly"])
    {
      return;
    }

    if (this._ignoreShow)
    {
      //set within hide or elsewhere and focus is placed back on this.element
      this._ignoreShow = false;
      return;
    }

    if (isPickerNative(this))
    {
      // our html picker is inside popup, which will take care of removing focus from input element,
      //  for native case we do it explicitly
      this.element.blur();
      return this._showNativeTimePicker();
    }
    else
    {
      return this._showWheelPicker();
    }
  },

  /**
   * Shows the native time picker
   *
   * @private
   */
  _showNativeTimePicker : function ()
  {
    // picker expects the fields like 'date' and 'mode' to retain its names. Use bracket notation
    //  to avoid closure compiler from renaming them
    var pickerOptions = {};
    var converter = _getTimePickerConverter(this._GetConverter(), {"isoStrFormat": "offset"});
    var date = _getNativePickerDate(converter, this._getIsoDateValue(converter));

    pickerOptions['date'] = date;
    pickerOptions['mode'] = 'time';

    var splitIncrements = splitTimeIncrement(this.options["timePicker"]["timeIncrement"]);

    // native picker supports only minute interval and only on iOS, we consider
    //  minute interval only when hours is not specified
    pickerOptions['minuteInterval'] = (splitIncrements.hourIncr === 0) ? splitIncrements.minuteIncr : 1;

    // if part of datetime, then get the min/max from the date component
    var minDate = this._isContainedInDateTimePicker() ?
      this._datePickerComp["widget"].options["min"] : this.options["min"];
    var maxDate = this._isContainedInDateTimePicker() ?
      this._datePickerComp["widget"].options["max"] : this.options["max"];

    if (minDate)
    {
      // get a correctly formatted ISO date string
      var minDateProcessed = _getNativePickerDate(converter, oj.IntlConverterUtils._minMaxIsoString(minDate, this._getIsoDateValue(converter)));
      pickerOptions['minDate'] = minDateProcessed.valueOf();
    }

    if (maxDate)
    {
    // get a correctly formatted ISO date string
      var maxDateProcessed = _getNativePickerDate(converter, oj.IntlConverterUtils._minMaxIsoString(maxDate, this._getIsoDateValue(converter)));
      pickerOptions['maxDate'] = maxDateProcessed.valueOf();
    }

    var self = this;

    function onTimePicked(date)
    {
      self._nativePickerShowing = false;

      // for iOS and windows, from the current implementation of the native datepicker plugin,
      //  for case when the picker is cancelled, this callback gets called without the parameter
      if (date)
      {
        var converter = _getTimePickerConverter(self._GetConverter(), {"isoStrFormat": "offset"});
        // The time picker displays the time portion as is supplied, regardless of device timezone.
        //  Explicitly setting timezone is supported only in iOS, and we do not have a need to do
        //  so at this time, so not exposing this feature for now.
        //  The value returned after pick will have the supplied timezone preserved, however, the
        //  date portion will be reset to current date when in 'time' mode. This will not impact us
        //  because we extract only the time portion to be set on the component.
        var isoString = oj.IntlConverterUtils._dateTime(self._getIsoDateValue(converter), {"hours": date.getHours(), "minutes": date.getMinutes(), "seconds": date.getSeconds()});
        var formattedTime = self._GetConverter().format(isoString);

        // _SetValue will inturn call _SetDisplayValue
        self._SetValue(formattedTime, {});
      }

      self._onClose(self._ON_CLOSE_REASON_SELECTION);
    }

    // onError is called only for Android for cases where picker is cancelled, or if there were
    //  to be any error at the native picker end
    function onError(error)
    {
      self._nativePickerShowing = false;

      // if user cancels the picker dialog, we just bring the focus back
      // closure compiler renames 'startsWith', using bracket notation hence
      if (error["startsWith"]('cancel'))
      {
        self._onClose(self._ON_CLOSE_REASON_CANCELLED);
      }
      else
      {
        oj.Logger.log('Error: native time picker failed: ' + error);
      }
    }

    this._nativePickerShowing = true;

    // datePicker is variable at the top level available when the cordova date picker plugin is
    //  included
    window['datePicker'].show(pickerOptions, onTimePicked, onError);
  },

  /**
   * Hides the timepicker. Note that this function is a no-op when renderMode is 'native'.
   *
   * @expose
   * @instance
   * @memberof! oj.ojInputTime
   * @return {void}
   */
  hide : function ()
  {
    return this._hide(this._ON_CLOSE_REASON_CLOSE);
  },

  /**
   * Hides the timepicker
   *
   * @param {string} reason - the reason that the popup is being hidden ("selection", "cancelled", "tab")
   *
   * @ignore
   * @expose
   * @memberof! oj.ojInputTime
   * @instance
   */
  _hide : function (reason)
  {
    if (!isPickerNative(this) && this._timepickerShowing())
    {
      this._popUpWheelPicker.ojPopup("close");
      
      this._onClose(reason);
    }

    return this;
  },

  /**
   * Sets focus to the right place after the picker is closed
   *
   * @param {string} reason - the reason that the popup is being hidden ("selection", "cancelled", "tab", "close")
   * @ignore
   */
  _onClose : function (reason)
  {
    if(this._isMobile && this.options["timePicker"]["showOn"] === "focus")
    {
      var inputContainer = this._isIndependentInput() ? this._inputContainer : this._datePickerComp["widget"]._inputContainer;
      inputContainer.focus();
    }
    else
    {

      if(this.options["timePicker"]["showOn"] === "focus")
      {
        if (!this._isIndependentInput())
        {
          this._datePickerComp["widget"]._ignoreShow = true;
        }
        else
        {
          this._ignoreShow = true;
        }
      }
      this.element.focus();
    }
  },

  /**
   * Refreshes the element. Usually called after dom changes have been made.
   * @expose
   * @instance
   * @memberof oj.ojInputTime
   * @return {void}
   */
  refresh : function ()
  {
    if(this._triggerNode) {
      this._triggerNode.find("." + this._TRIGGER_TIME_CLASS).attr("title", this._getTimeTitle());
    }
    return this._superApply(arguments) || this;
  },

  /**
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputTime
   */
  _SetDisplayValue : function (displayValue)
  {
    //When not part of datePickerComp or of inline should update input element
    if (this._isIndependentInput())
    {
      this._superApply(arguments);
    }

    //so this is a change in behavior from original design. Previously it was decided that app developer
    //would have to invoke refresh to render the calendar after setting the new value programatically; however now it is
    //required to hook it in when _SetDisplayValue is invoked [can't use _SetValue b/c that function is not invoked
    //when developer invokes ("option", "value", "..")
    if(this._timepickerShowing())
    {
      this._createWheelPickerDom(true);
    }
  },

  /**
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputTime
   */
  _SetValue : function (newValue, event, options)
  {
    var ret = false;

    if(this._isContainedInDateTimePicker())
    {
      //never update the model if part of ojInputDateTime. Have ojInputDateTime update the model's value [otherwise 2 updates]
      //this is mainly for check of whether the format is correct [i.e when ojInputDateTime is inline], since the value
      //is always picked from the ojInputDateTime component
      
      try{
        //originally this._super call was invoked above, but the the timepicker wheel is now redrawing
        //before the below timeselected is propagated to the datetimepicker (which has the complete value that must be 
        //used and because of that kicked it down)

        //if the operation is an enter key in the input field (not in the picker), the newValue is not just time, it is date+time. So we
        //need to deal differently
        var converter, datePickerCompWidget = this._datePickerComp["widget"];
        if(event && event.target && event.target === this.element[0] && !this._isDatePickerInline())
        {
          converter = datePickerCompWidget._GetConverter();
        }
        else
        {
          converter = this._GetConverter();
        }

        var parsedNewValue = converter["parse"](newValue),
            converterUtils = oj.IntlConverterUtils,
            dateTimeValue = datePickerCompWidget.getValueForInputTime() || converterUtils.dateToLocalIso(new Date());

        if(parsedNewValue && converter.compareISODates(dateTimeValue, parsedNewValue) === 0)
        {
          //need to kick out if _SetValue happened due to Blur w/o changing of value
          return false;
        }

        var isoString = converterUtils._copyTimeOver(parsedNewValue || converterUtils.dateToLocalIso(new Date()),
            dateTimeValue);

        datePickerCompWidget.timeSelected(isoString, event);
        if(this._isDatePickerInline())
        {
          //need to update the input element, reason one can't when of the same input is 
          //b/c the wheelpicker apparently doesn't update minute for some odd reason
          ret = this._super(newValue, null, options);
        }

      }catch(e)
      {
        ret = this._super(newValue, null, options);
      }
    }
    else
    {
      ret = this._superApply(arguments);
    }

    return ret;
  },

  /**
   * Whether the this.element should be wrapped. Function so that additional conditions can be placed
   *
   * @ignore
   * @protected
   * @override
   * @return {boolean}
   */
  _DoWrapElement : function ()
  {
    return this._isIndependentInput();
  },

  /**
   * Whether the input element of ojInputTime is shared or not [i.e. not part of ojInputDateTime or if it has
   * been created by ojInputDateTime that is inline
   *
   * @ignore
   * @return {boolean}
   */
  _isIndependentInput : function ()
  {
    return !this._isContainedInDateTimePicker() || this._isDatePickerInline();
  },

  /**
   * @protected
   * @override
   * @return {string}
   * @instance
   * @memberof! oj.ojInputTime
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-inputtime";
  },

  /**
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputTime
   */
  _GetElementValue : function ()
  {
    return this.options['value'] || "";
  },

  /**
   * Sets up the default dateTimeRange and dateRestriction validators.
   *
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputTime
   */
  _GetImplicitValidators : function ()
  {
    var ret = this._superApply(arguments);

    if((this.options['min'] != null || this.options['max'] != null) && !this._isContainedInDateTimePicker())
    {
      //need to alter how the default validators work as validators are now immutable and to create the implicit validator only
      //if independent input [i.e. otherwise have ojInputDateTime take care of it]
      this._timePickerDefaultValidators[oj.ValidatorFactory.VALIDATOR_TYPE_DATETIMERANGE] =
        getImplicitDateTimeRangeValidator(this.options, this._GetConverter(), this._GetDefaultStyleClass());
    }

    return $.extend(this._timePickerDefaultValidators, ret);
  },

  /**
   * Need to override since apparently we allow users to set the converter to null, undefined, and etc and when
   * they do we use the default converter
   *
   * @return {Object} a converter instance or null
   *
   * @memberof! oj.ojInputTime
   * @instance
   * @protected
   * @override
   */
  _GetConverter : function ()
  {
    return this.options['converter'] ?
            this._superApply(arguments) :
            $["oj"]["ojInputTime"]["prototype"]["options"]["converter"];
  },

  /**
   * Whether ojInputTime has been created by ojInputDateTime
   *
   * @private
   */
  _isContainedInDateTimePicker : function ()
  {
    return this._datePickerComp !== null;
  },

  /**
   * Helper function to determine whether the provided datePickerComp is inline or not
   *
   * @private
   */
  _isDatePickerInline : function ()
  {
    return this._datePickerComp["inline"];
  },

  /**
   * Notifies the component that its subtree has been removed from the document programmatically after the component has
   * been created
   * @memberof! oj.ojInputTime
   * @instance
   * @protected
   */
  _NotifyDetached: function()
  {
    this._hide(this._ON_CLOSE_REASON_CLOSE);
    // hide sets focus to the input, so we want to call super after hide. If we didn't, then
    // the messaging popup will reopen and we don't want that.
    this._superApply(arguments);
  },

  /**
   * Notifies the component that its subtree has been made hidden programmatically after the component has
   * been created
   * @memberof! oj.ojInputTime
   * @instance
   * @protected
   */
  _NotifyHidden: function()
  {
    this._hide(this._ON_CLOSE_REASON_CLOSE);
    // hide sets focus to the input, so we want to call super after hide. If we didn't, then
    // the messaging popup will reopen and we don't want that.
    this._superApply(arguments);
  },

  /**
   * Return true if the element is in the timepicker popup
   *
   * @param {?Element} element
   * @return {boolean}
   *
   * @memberof! oj.ojInputTime
   * @instance
   * @protected
   */
  _InPopup: function(element)
  {
    if (!element)
      return false;

    var picker = this._wheelPicker;
    
    return $.contains(picker[0], element);
  },

  /**
   * Generate the HTML for the header of the time picker.
   *
   * @private
   */
  _generateHeader : function()
  {
    var isRTL = this._IsRTL();

    var cancelText = this._EscapeXSS(this.getTranslatedString("cancelText"));
    var cancelButton = "<a role='button' onclick='return false;' href='#' class='oj-enabled oj-default oj-timepicker-cancel-button'" + " aria-label='" + cancelText + "'>" + cancelText + "</a>";

    var okText = this._EscapeXSS(this.getTranslatedString("okText"));
    var okButton = "<a role='button' onclick='return false;' href='#' class='oj-enabled oj-default oj-timepicker-ok-button'" + " aria-label='" + okText + "'>" + okText + "</a>";

    var header = "<div class='oj-timepicker-header" + (this.options["disabled"] ? " oj-disabled " : " oj-enabled oj-default ") + "'>";

    header += cancelButton;
    header += okButton;

    header += "</div>";
    return header;
  },

  /**
   * Generate the HTML for the footer of the time picker.
   *
   * @private
   */
  _generateFooter : function (footerLayoutDisplay, gotoTime)
  {
    var footerLayout = "";
    var currentText = this._EscapeXSS(this.getTranslatedString("currentTimeText"));
    var nowControl = "<a role='button' onclick='return false;' href='#' class='oj-timepicker-now oj-priority-secondary oj-enabled'" + ">" + currentText + "</a>";

    if(footerLayoutDisplay && footerLayoutDisplay.length > 1) //keep the code for future multiple buttons
    {
      var nowIndex = footerLayoutDisplay.indexOf("now"),
          loop = 0,
          footerLayoutButtons = [{index: nowIndex, content: (gotoTime ? nowControl : "")}];

      //rather than using several if + else statements, sort the content to add by index of the strings
      footerLayoutButtons.sort(function(a, b)
      {
        return a.index - b.index;
      });

      //continue to loop until the index > -1 [contains the string]
      while(loop < footerLayoutButtons.length && footerLayoutButtons[loop].index < 0) { loop++; }

      while(loop < footerLayoutButtons.length)
      {
        footerLayout += footerLayoutButtons[loop++].content;
      }

      if(footerLayout.length > 0)
      {
        footerLayout = "<div class='oj-timepicker-footer'>" + footerLayout + "</div>";
      }
    }
    return footerLayout;
  },

  /**
   * Get the ISO string for the min or max date limit if applicable
   *
   * @private
   */
  _getIsoDateLimit : function(converter, optionName, valueDate)
  {
    // Fetch option from correct picker
    var dateIso = this._isContainedInDateTimePicker() ? this._datePickerComp["widget"].options[optionName] : this.options[optionName];
    // Fill in date from value
    dateIso = dateIso ? oj.IntlConverterUtils._minMaxIsoString(dateIso, this._getValue()) : dateIso;
    // Move to the converter's timezone
    dateIso = dateIso ? converter.parse(dateIso) : dateIso;
    // If the dates don't match, then min or max doesn't apply to this value
    var date = dateIso ? oj.IntlConverterUtils._clearTime(dateIso) : null;
    if (valueDate && date && valueDate.substring(0, valueDate.indexOf("T")) !== date.substring(0, date.indexOf("T")))
    {
      dateIso = null;
    }

    return dateIso;
  },

  /**
   * Get the ISO string for the current value
   *
   * @private
   */
  _getIsoDateValue : function(converter)
  {
    var processDateIso = this._getValue();
    if(!processDateIso)
    {
      processDateIso = new Date();
      processDateIso.setHours(0);
      processDateIso.setMinutes(0);
      processDateIso.setSeconds(0);
      processDateIso.setMilliseconds(0);
      processDateIso = oj.IntlConverterUtils.dateToLocalIso(processDateIso);
    }
    processDateIso = converter.parse(processDateIso);

    return processDateIso;
  },

  /**
   * Get the local strings for AM/PM representation
   *
   * @private
   */
  _getAmPmStrings : function()
  {
    var factory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME);
    var converter = factory.createConverter({"pattern": "a"});

    return [converter.format("2016-01-01T01:00:00Z"),
            converter.format("2016-01-01T13:00:00Z")];
  },

  /**
   * @private
   * @param {jQuery} wheelPicker popup root node
   * @return {jQuery} returns the timepicker content node relative to the root node
   */
  _getWheelPickerContent: function (wheelPicker)
  {
    if (!wheelPicker)
      return null;

    var wheelPickerContent = $(wheelPicker.find(".oj-timepicker-content")[0]);
    return wheelPickerContent;
  },

  /**
   * Create the wheel timepicker DOM
   *
   * @private
   */
  _createWheelPickerDom : function (keepHeaderFooter)
  {
    if(!keepHeaderFooter) 
    {
      this._detachWheelHandler();
    }
    
    var wheelPicker = this._wheelPicker;
    var wheelPickerContent = this._getWheelPickerContent(wheelPicker);
    this._destroyHammer();
    if(keepHeaderFooter) 
    {
      $(".oj-timepicker-wheel-group", wheelPickerContent).remove();
    }
    else
    {
      wheelPickerContent.empty();
    }

    var converter = this._GetConverter();
    var options = $.extend({}, converter.resolvedOptions());
    if (options.isoStrFormat === "zulu")
    {
      options.isoStrFormat = "offset";
      var factory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME);
      converter = factory.createConverter(options);
      options = converter.resolvedOptions();
    }

    var converterUtils = oj.IntlConverterUtils;
    var value = this._getValue();
    var date = new Date();
    if (!value)
    {
      value = oj.IntlConverterUtils.dateToLocalIso(date);
    }
    value = converter.parse(value);  // Convert to proper timezone
    var valueDate = converterUtils._clearTime(value);

    var minDateIso = this._getIsoDateLimit(converter, "min", valueDate);
    var maxDateIso = this._getIsoDateLimit(converter, "max", valueDate);

    var footerLayoutDisplay = this.options["timePicker"]["footerLayout"];

    var timePickerModel = this._timePickerModel = new TimePickerModel(null);
    var wheelPos = ["", "", "", ""];

    var timePositions = converter._getTimePositioning();
    for(var position in timePositions)
    {
      wheelPos[timePositions[position]] = position;
    }
    timePickerModel["wheelOrder"] = wheelPos.filter(function(val) { return !!val; }).join("");

    var pattern = options["pattern"] || options["patternFromOptions"];
    if (pattern)
    {
      timePickerModel["format"] = pattern;
    }

    timePickerModel["ampmStrings"] = this._getAmPmStrings();

    //set increment before setting value so that positions will be calculated correctly
    var timeIncrement = this.options["timePicker"]["timeIncrement"];
    if (timeIncrement)
    {
      var splitIncrements = splitTimeIncrement(timeIncrement);
      timePickerModel["increment"] = splitIncrements.hourIncr * 60 + splitIncrements.minuteIncr;
    }

    //set value
    timePickerModel["isoValue"] = this._getIsoDateValue(converter);

    //set min and max values
    if (minDateIso)
      timePickerModel["isoMin"] = minDateIso;

    if (maxDateIso)
      timePickerModel["isoMax"] = maxDateIso;

    //add header, wheel group and footer
    if(this._isIndependentInput() && (!keepHeaderFooter || $(".oj-timepicker-header", wheelPickerContent).length === 0))
    {
      wheelPickerContent.append($(this._generateHeader(wheelPicker))); //@HTMLUpdateOK; header is generated internally so ok
    }
    this._wheelGroup = $(createWheelGroup(timePickerModel));

    if(this._isIndependentInput())
    {
      this._wheelGroup.insertAfter($(".oj-timepicker-header", wheelPickerContent)); //@HTMLUpdateOK; wheelGroup is created internally using numbers so ok
    }
    else
    {
      wheelPickerContent.append(this._wheelGroup); //@HTMLUpdateOK; wheelGroup is created internally using numbers so ok
    }
    
    if(!keepHeaderFooter || $(".oj-timepicker-footer", wheelPickerContent).length === 0) 
    {
      wheelPickerContent.append($(this._generateFooter(footerLayoutDisplay, date))); //@HTMLUpdateOK; footer is generated internally so ok
    }

    //add aria labels to wheels for accessibility
    wheelPickerContent.find(".oj-timepicker-hour").attr("aria-label", this.getTranslatedString("hourWheelLabel"));
    wheelPickerContent.find(".oj-timepicker-minute").attr("aria-label", this.getTranslatedString("minuteWheelLabel"));
    wheelPickerContent.find(".oj-timepicker-meridian").attr("aria-label", this.getTranslatedString("ampmWheelLabel"));

    if(!keepHeaderFooter)
    {
      this._attachWheelHandler();
    }
  },

  /**
   * Detach wheel handlers
   *
   * @private
   */
  _detachWheelHandler: function() 
  {
    this._wheelPicker.find(".oj-timepicker-now").off('click');

    if(this._isIndependentInput())
    {
      this._wheelPicker.find(".oj-timepicker-cancel-button").off('click');
      this._wheelPicker.find(".oj-timepicker-ok-button").off('click');
    }
    else
    {
      this._wheelPicker.find(".oj-timepicker-wheel").off('blur');
      return;
    }

    this._wheelPicker.off('keydown');
  },

  /**
   * Attaches wheel handlers
   *
   * NOTE: Timepicker wheel was written to redraw the whole content on basically every action, b/c of this it causes random 
   * behavior. In order to stabilize it, will minimize the creation of header + footer in the _createWheelPickerDom
   * function and to off before on the handler for the handlers (before would fire multiple times).
   * @private
   */
  _attachWheelHandler: function() 
  {
    var self = this;
    //now control
    
    this._wheelPicker.find(".oj-timepicker-now").on('click',
      function(event)
      {
        var value = oj.IntlConverterUtils.dateToLocalIso(new Date());
        value = self._GetConverter().parse(value);  // Convert to proper timezone
        self._timePickerModel["isoValue"] = value;
        if (!self._isIndependentInput())
        {
          //when is not an independent component (i.e. part of switcher)
          //need to set the value to have datetimepicker handle the value
          self._SetValue(self._timePickerModel["isoValue"], event);
        }

        event.preventDefault();
      });

    if(this._isIndependentInput())
    {
      this._wheelPicker.find(".oj-timepicker-cancel-button").on('click',
        function(event)
        {
          event.preventDefault();
          self._hide(self._ON_CLOSE_REASON_CANCELLED);
        });

      this._wheelPicker.find(".oj-timepicker-ok-button").on('click',
        function(event)
        {
          self._hide(self._ON_CLOSE_REASON_SELECTION);
          self._SetValue(self._timePickerModel["isoValue"], event);
          event.preventDefault();
        });
    }
    else
    {
      this._wheelPicker.find(".oj-timepicker-wheel").on('blur',
        function(event)
        {
          self._SetValue(self._timePickerModel["isoValue"], event);
        });

      //need to hide the datePickerComp prior to showing timepicker
      this._datePickerComp["widget"]._togglePicker();
      //set focus on the 1st child
      this._wheelGroup.children().first().focus();
      return;
    }

    this._wheelPicker.on('keydown',
      function(event)
      {
        if (event.keyCode == $.ui.keyCode.ESCAPE)
        {
          event.preventDefault();
          self._hide(self._ON_CLOSE_REASON_CANCELLED);
        }
      });
  },

  /**
   * Shows the wheel timepicker
   *
   * @private
   */
  _showWheelPicker : function ()
  {
    
    this._createWheelPickerDom();

    var self = this;
    var popUpWheelPicker = this._popUpWheelPicker;
    if(!popUpWheelPicker) return;

    // ojPopup open must be passed the input container (this.element.parent)
    // as the launcher so that any event within the container won't auto-dismiss
    // the dialog.
    if(!_isLargeScreen)
    {
      popUpWheelPicker.ojPopup("open", this.element.parent(), {
        "my" : {"horizontal":"center", "vertical": "bottom"},
        "at" : {"horizontal":"center", "vertical": "bottom"},
        "of" : window,
        "collision" : "flipfit flipfit"
      });
      
      // if we don't have a large screen, the popup will be modal so
      // we need to give it the focus
      this._wheelGroup.children().first().focus();
    }
    else
    {
      var position = oj.PositionUtils.normalizeHorizontalAlignment({"my" : "start top", "at" : "start bottom", "of" : this.element, "collision" : "flipfit flipfit"}, this._IsRTL());
      popUpWheelPicker.ojPopup("open", this.element.parent(), position);
    }
    
  },

  // @inheritdoc
  getNodeBySubId: function(locator)
  {
    var node = null,
        subId = locator && locator['subId'],
        wheelPicker = this._wheelPicker,
        wheelPickerContent = this._getWheelPickerContent(wheelPicker)

    if(subId) {
      switch(subId)
      {
      case "oj-inputdatetime-time-icon": node = $(".oj-inputdatetime-time-icon", this._triggerNode)[0]; break;
      case "oj-inputdatetime-time-input": node = this.element[0]; break;

      case "oj-timepicker-content": node = wheelPickerContent ? wheelPickerContent[0] : null; break;
      case "oj-timepicker-cancel-button": node = $(".oj-timepicker-cancel-button", wheelPicker)[0]; break;
      case "oj-timepicker-ok-button": node = $(".oj-timepicker-ok-button", wheelPicker)[0]; break;
      case "oj-timepicker-hour": node = $(".oj-timepicker-hour", wheelPicker)[0]; break;
      case "oj-timepicker-minute": node = $(".oj-timepicker-minute", wheelPicker)[0]; break;
      case "oj-timepicker-meridian": node = $(".oj-timepicker-meridian", wheelPicker)[0]; break;
      case "oj-timepicker-now": node = $(".oj-timepicker-now", wheelPicker)[0]; break;

      default: node = null;
      }
    }

    return node || this._superApply(arguments);
  },

  // @inheritdoc
  getSubIdByNode: function(node)
  {
    var timeIcon = $(".oj-inputdatetime-time-icon", this._triggerNode),
        subId = null,
        wheelPicker = this._wheelPicker,
        wheelPickerContent = this._getWheelPickerContent(wheelPicker),
        checks = [{"selector": ".oj-timepicker-cancel-button", "ele": wheelPicker},
                  {"selector": ".oj-timepicker-ok-button", "ele": wheelPicker},
                  {"selector": ".oj-timepicker-hour", "ele": wheelPicker},
                  {"selector": ".oj-timepicker-minute", "ele": wheelPicker},
                  {"selector": ".oj-timepicker-meridian", "ele": wheelPicker},
                  {"selector": ".oj-timepicker-now", "ele": wheelPicker}];


    if(node === timeIcon[0])
    {
      subId = "oj-inputdatetime-time-icon";
    }
    else if(node === this.element[0])
    {
      subId = "oj-inputdatetime-time-input";
    }
    else if(wheelPickerContent && node === wheelPickerContent[0])
    {
      subId = "oj-timepicker-content";
    }
    else
    {
      for(var i=0; i < checks.length; i++)
      {
        var map = checks[i],
            entry = $(map["selector"], map["ele"]);

        if(entry.length === 1 && entry[0] === node)
        {
          subId = map["selector"].substr(1);
          break;
        }
      }
    }

    return subId || this._superApply(arguments);
  },

  /**
   * Returns the root node
   *
   * @expose
   * @instance
   * @memberof! oj.ojInputTime
   * @ignore
   */
  widget : function ()
  {
    return this._isIndependentInput() ? this._super() : this._datePickerComp["widget"].widget();
  }

});

// Add custom getters for properties
oj.Components.setDefaultOptions(
{
  'ojInputTime':
  {
    'renderMode': oj.Components.createDynamicPropertyGetter(
      function()
      {
        return (oj.ThemeUtils.parseJSONFromFontFamily('oj-inputdatetime-option-defaults') || {})["renderMode"];
      }),
    'keyboardEdit': oj.Components.createDynamicPropertyGetter(
      function()
      {
        return (oj.ThemeUtils.parseJSONFromFontFamily('oj-inputdatetime-option-defaults') || {})["keyboardEdit"];
      }),

    'timePicker': oj.Components.createDynamicPropertyGetter(
      function()
      {
        return (oj.ThemeUtils.parseJSONFromFontFamily('oj-inputdatetime-option-defaults') || {})["timePicker"];
      })
    }
  }
);

// Fragments:

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Input element and time trigger icon</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Shows the time picker and moves the focus into the expanded time picker</td>
 *     </tr>
 *     <tr>
 *       <td>Input element with picker open</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Set focus to the input. If hints, title or messages exist in a notewindow,
 *        pop up the notewindow.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputTime
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>DownArrow or UpArrow</kbd></td>
 *       <td>Shows the time picker and moves the focus into the expanded time picker</td>
 *     </tr>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>Tab In</kbd></td>
 *       <td>Set focus to the input. If hints, title or messages exist in a notewindow,
 *        pop up the notewindow.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputTime
 */

//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the InputTime component's input element.</p>
 *
 * @ojsubid oj-inputdatetime-time-input
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the node for the input element:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-inputdatetime-time-input'} );
 */

/**
 * <p>Sub-ID for the time icon that triggers the time drop down display.</p>
 *
 * @ojsubid oj-inputdatetime-time-icon
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the time icon that triggers the time drop down display:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-inputdatetime-time-icon'} );
 */

/**
 * <p>Sub-ID for the time wheel picker drop down node.
 *
 * @ojsubid oj-timepicker-content
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the time wheel picker drop down node:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-timepicker-content'} );
 */

/**
 * <p>Sub-ID for the cancel button.
 *
 * @ojsubid oj-timepicker-cancel-button
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the cancel button:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-timepicker-cancel-button'} );
 */

/**
 * <p>Sub-ID for the OK button.
 *
 * @ojsubid oj-timepicker-ok-button
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the OK button:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-timepicker-ok-button'} );
 */

/**
 * <p>Sub-ID for the hour picker.
 *
 * @ojsubid oj-timepicker-hour
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the hour picker:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-timepicker-hour'} );
 */

/**
 * <p>Sub-ID for the minute picker.
 *
 * @ojsubid oj-timepicker-minute
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the minute picker:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-timepicker-minute'} );
 */

/**
 * <p>Sub-ID for the meridian picker.
 *
 * @ojsubid oj-timepicker-meridian
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the meridian picker:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-timepicker-meridian'} );
 */

/**
 * <p>Sub-ID for the now button for button bar.
 *
 * @ojsubid oj-timepicker-now
 * @memberof oj.ojInputTime
 * @instance
 *
 * @example <caption>Get the now/now button for button bar:</caption>
 * var node = myInputTime.getNodeBySubId( {'subId': 'oj-timepicker-now'} );
 */

/**
 * @ignore
 * @protected
 * @constructor
 */
function TimePickerModel(properties)
{
  /** @const */
  var ISO_TIME = /^.*T(\d{2})(?::?(\d{2}).*$)/;
  /** @const */
  var MIN_TIME = 0;
  /** @const */
  var MAX_TIME = 60 * 24;
  /** @const */
  var HOURS12 = 12 * 60;
  
  var FORMAT_MAP = {
    'h': hFormat,
    'hh': hhFormat,
    'H': HFormat,
    'HH': HHFormat,
    'k': kFormat,
    'kk': kkFormat,
    'K': KFormat,
    'KK': KKFormat,
    'mm': mmFormat
  };
  var PARSER_MAP = {
    'h': hour12Parser,
    'hh': hour12Parser,
    'H': numberParser,
    'HH': numberParser,
    'k': hour24Parser,
    'kk': hour24Parser,
    'K': numberParser,
    'KK': numberParser,
    'mm': numberParser
  };
  
  var _value = 0;

  var _increment;
  var _min = MIN_TIME;
  var _minValue = MIN_TIME;
  var _max = MAX_TIME;
  var _maxValue = MAX_TIME;

  var _format;
  var _12Hour;
  var _grouped = "auto";
  var _wheelOrder = "";
  var _ampmStrings = ["AM", "PM"];

  var _minuteModel = new WheelModel(this, {
    'valueRange': 60,
    'displayRange': 60,
    'valueMultiplier': 1,
    'displayMultiplier': 1  
  });
  var _hourModel = new WheelModel(this, {
    'valueMultiplier': 60,
    'displayMultiplier': 1
  });
  var _ampmModel = new WheelModel(this, {
    'valueRange': 2,
    'displayRange': 2,
    'formatter': ampmFormat,
    'parser': ampmParser,
    'valueMultiplier': HOURS12, 
    'displayMultiplier': 1 
  });
  var _models = [_minuteModel, _hourModel, _ampmModel];

  var _settingProps = false;
  
//  /** type {TimePickerModel} */
  var self = this;

  defineProperties();
  defineMethods();
  setProperties(properties);

  function defineProperties()
  {
    _settingProps = true;

    Object.defineProperty(self, "increment", {
      'enumerable': true,
      'get': function()
      {
        return _increment;
      },
      'set': function(increment)
      {
        _increment = Math.floor(increment);
        refreshSettings();
      }
    });
    self["increment"] = 1;

    Object.defineProperty(self, "grouped", {
      'enumerable': true,
      'get': function()
      {
        return _grouped;
      },
      'set': function(grouped)
      {
        switch (grouped)
        {
        case "auto":
        case "all":
        case "none":
        case "hoursMinutes":
        case "hoursMeridiem":
          _grouped = grouped;
          break;
        default:
          throw new Error("invalid grouped value: " + grouped);
        }

        refreshSettings();
      }
    });

    Object.defineProperty(self, "min", {
      'enumerable': true,
      'get': function()
      {
        return _min;
      },
      'set': function(min)
      {
        _min = Math.floor(min);
        refreshSettings();
      }
    });

    Object.defineProperty(self, "isoMin", {
      'enumerable': true,
      'get': function()
      {
        return minutesToIso(_min);
      },
      'set': function(iso)
      {
        var newMin = isoToMinutes(iso);
        if (!isNaN(newMin))
        {
          self["min"] = newMin;
        }
        else
        {
          console.log("Invalid ISO min time: " + iso);
        }
      }
    });

    Object.defineProperty(self, "max", {
      'enumerable': true,
      'get': function()
      {
        return _max - 1;
      },
      'set': function(max)
      {
        _max = Math.floor(max) + 1;
        refreshSettings();
      }
    });

    Object.defineProperty(self, "isoMax", {
      'enumerable': true,
      'get': function()
      {
        return minutesToIso(_max);
      },
      'set': function(iso)
      {
        var newMax = isoToMinutes(iso);
        if (!isNaN(newMax))
        {
          self["max"] = newMax;
        }
        else
        {
          console.log("Invalid ISO max time: " + iso);
        }
      }
    });

    Object.defineProperty(self, "value", {
      'enumerable': true,
      'get': function()
      {
        return _value;
      },
      'set': function(value)
      {
        _value = Math.floor(value);
        refreshSettings();

      }
    });

    Object.defineProperty(self, "isoValue", {
      'enumerable': true,
      'get': function()
      {
        return minutesToIso(_value);
      },
      'set': function(iso)
      {
        var newValue = isoToMinutes(iso);
        if (!isNaN(newValue))
        {
          self["value"] = newValue;
        }
        else
        {
          console.log("Invalid ISO value time: " + iso);
        }
      }
    });

    Object.defineProperty(self, "format", {
      'enumerable': true,
      'get': function()
      {
        return _format;
      },
      'set': function(format)
      {
        _format = format;
        // remove stuff between quotes.
        // remove anything that is not h, H, k, K, m or a
        format = format.replace(/\'[^']*\'/g, "").replace(/[^hHkKma]*/g, "");
        
        var matches = format.match(/([hHkK]+)/);
        var hourCode = matches[1];
        _hourModel["formatter"] = FORMAT_MAP[hourCode];
        _hourModel["parser"] = PARSER_MAP[hourCode];
        
        matches = format.match(/(m+)/);
        var minuteCode = matches[1];
        _minuteModel["formatter"] = FORMAT_MAP[minuteCode];
        _minuteModel["parser"] = PARSER_MAP[minuteCode];

        _12Hour = _wheelOrder.indexOf("a") >= 0;
        refreshSettings();
      }
    });

    Object.defineProperty(self, "wheelOrder", {
      'enumerable': true,
      'get': function()
      {
        return _wheelOrder;
      },
      'set': function(wheelOrder)
      {
        _wheelOrder = wheelOrder;
      }
    });

    Object.defineProperty(self, "ampmStrings", {
      'enumerable': true,
      'get': function()
      {
        return _ampmStrings;
      },
      'set': function(ampmStrings)
      {
        _ampmStrings = ampmStrings;
      }
    });

    _settingProps = false;
  }

  function defineMethods()
  {
    self.setProperties = setProperties;
  }

  function setProperties(properties)
  {
    try
    {
      _settingProps = true;

      if (properties)
      {
        for (var key in properties)
        {
          self[key] = properties[key];
        }
      }
    }
    finally
    {
      _settingProps = false;
      refreshSettings();
    }
  }

  /*
   * recalculates dependent values after settings change
   */
  function refreshSettings()
  {
    if (!_settingProps)
    {
      _minValue = Math.ceil(_min / _increment) * _increment;
      _maxValue = Math.ceil((_max - 1)/ _increment) * _increment;

      if (_maxValue < _minValue)
      {
        throw new Error("Invalid min and max settings with current increment: " +
                _min + " " + _max + + _increment);
      }

      if (_value < _minValue)
      {
        _value = _minValue;
      }

      if (_value > _maxValue)
      {
        _value = _maxValue;
      }

      _value = Math.round(_value / _increment) * _increment;

      if (_12Hour)
      {
        _hourModel["displayRange"] = 12;

        var grouped = _grouped;
        if (grouped === "auto")
        {
          // If the increment is a divisor or multiple of 60
          // Then use hoursMeridiem grouping
          var divisor = gcd(_increment, 60);
          if (divisor === _increment || divisor === 60)
          {
            grouped = "hoursMeridiem";
          }
          else
          {
            divisor = gcd(_increment, HOURS12);
            if (divisor === _increment)
            {
              grouped = "all";
            }
            else  
            {
              grouped = "all";
            }
          }
        }

        switch (grouped)
        {
        case "all":
          _minuteModel["displayMultiplier"] = 1;
          _minuteModel["valueMultiplier"] = 1;
          _minuteModel["valueRange"] = 24 * 60;
          _minuteModel["linked"] = true;

          _hourModel["displayMultiplier"] = 60;
          _hourModel["valueMultiplier"] = 1;
          _hourModel["valueRange"] = 24 * 60;
          _hourModel["linked"] = true;

          _ampmModel["displayMultiplier"] = 1;
          _ampmModel["valueMultiplier"] = 60 * 12;
          _ampmModel["valueRange"] = 2;
          _ampmModel["linked"] = true;
          break;

        //case "none":
        //  _minuteModel["displayMultiplier"] = 1;
        //  _minuteModel["valueMultiplier"] = 1;
        //  _minuteModel["valueRange"] = 60;
        //  _minuteModel["linked"] = false;

        //  _hourModel["displayMultiplier"] = 1;
        //  _hourModel["valueMultiplier"] = 60;
        //  _hourModel["valueRange"] = 12;
        //  _hourModel["linked"] = false;

        //  _ampmModel["displayMultiplier"] = 1;
        //  _ampmModel["valueMultiplier"] = 60 * 12;
        //  _ampmModel["valueRange"] = 2;
        //  _ampmModel["linked"] = false;
        //  break;

        //case "hoursMinutes":
        //  _minuteModel["displayMultiplier"] = 1;
        //  _minuteModel["valueMultiplier"] = 1;
        //  _minuteModel["valueRange"] = HOURS12;
        //  _minuteModel["linked"] = true;

        //  _hourModel["displayMultiplier"] = 60;
        //  _hourModel["valueMultiplier"] = 1;
        //  _hourModel["valueRange"] = HOURS12;
        //  _hourModel["linked"] = true;

        //  _ampmModel["displayMultiplier"] = 1;
        //  _ampmModel["valueMultiplier"] = 60 * 12;
        //  _ampmModel["valueRange"] = 2;
        //  _ampmModel["linked"] = false;
        //  break;

        case "hoursMeridiem":
          _minuteModel["displayMultiplier"] = 1;
          _minuteModel["valueMultiplier"] = 1;
          _minuteModel["valueRange"] = 60;
          _minuteModel["linked"] = false;

          _hourModel["displayMultiplier"] = 1;
          _hourModel["valueMultiplier"] = 60;
          _hourModel["valueRange"] = 24;
          _hourModel["linked"] = true;

          _ampmModel["displayMultiplier"] = 1;
          _ampmModel["valueMultiplier"] = 60 * 12;
          _ampmModel["valueRange"] = 2;
          _ampmModel["linked"] = true;
          break;
        }
      }
      else // 24 hour
      {
        _hourModel["displayRange"] = 24;
        
        var grouped = _grouped;
        if (grouped === "auto")
        {
          // If the increment is a divisor or multiple of 60
          // Then no grouping
          var divisor = gcd(_increment, 60);
          if (divisor === _increment || divisor === 60)
          {
            grouped = "none";
          }
          else
          {
            grouped = "all";
          }
        }

        switch (grouped)
        {
        case "none":
        case "hoursMeridiem":
          _minuteModel["displayMultiplier"] = 1;
          _minuteModel["valueMultiplier"] = 1;
          _minuteModel["valueRange"] = 60;
          _minuteModel["linked"] = false;

          _hourModel["displayMultiplier"] = 1;
          _hourModel["valueMultiplier"] = 60;
          _hourModel["valueRange"] = 24;
          _hourModel["linked"] = false;
          break;

        case "all":
        case "hoursMinutes":
          _minuteModel["displayMultiplier"] = 1;
          _minuteModel["valueMultiplier"] = 1;
          _minuteModel["valueRange"] = 24 * 60;
          _minuteModel["linked"] = true;

          _hourModel["displayMultiplier"] = 60;
          _hourModel["valueMultiplier"] = 1;
          _hourModel["valueRange"] = 24 * 60;
          _hourModel["linked"] = true;
          break;
        }
      }

      for (var i = 0; i < _models.length; i++)
      {
        _models[i].refresh();
      }
    }
  }

  function isoToMinutes(isoString)
  {
    var matches = ISO_TIME.exec(isoString);
    var hours = parseInt(matches[1], 10);
    var minutes = parseInt(matches[2], 10);
    return hours * 60 + minutes;
  }
  function minutesToIso(minutes)
  {
    minutes = Math.floor(minutes);
    var hours = Math.floor(minutes / 60);
    minutes = minutes % 60;
    return "T" + HHFormat(hours) + ":" + mmFormat(minutes);
  }

  function hFormat(value)
  {
    if (value === 0)
      return "12";
    
    return "" + value;
  }

  function hhFormat(value)
  {
    if (value === 0)
      return "12";
    
    value = "0" + value;
    return value.slice(-2);
  }

  function HFormat(value)
  {
    return "" + value;
  }

  function HHFormat(value)
  {
    value = "0" + value;
    return value.slice(-2);
  }

  function kFormat(value)
  {
    if (value === 0)
      return "24";
    return "" + value;
  }

  function kkFormat(value)
  {
    if (value === 0)
      return "24";
    value = "0" + value;
    return value.slice(-2);
  }

  function KFormat(value)
  {
    return "" + value;
  }

  function KKFormat(value)
  {
    value = "0" + value;
    return value.slice(-2);
  }

  function mmFormat(value)
  {
    value = "0" + value;
    return value.slice(-2);
  }

  function ampmFormat(value)
  {
    return _ampmStrings[value];
  }

  function numberParser(value)
  {
    if (value.match(/^\d+$/))
    {
      return parseInt(value, 10);
    }
    return -1;
  }

  function hour12Parser(value)
  {
    if (value.match(/^\d+$/))
    {
      var hour =  parseInt(value, 10);
      if (hour === 0)
      {
        hour = -1;
      }
      if (hour === 24)
      {
        hour = 0;
      }

      return hour;
    }
    return -1;
  }

  function hour24Parser(value)
  {
    if (value.match(/^\d+$/))
    {
      var hour =  parseInt(value, 10);
      if (hour === 0)
      {
        hour = -1;
      }
      if (hour === 24)
      {
        hour = 0;
      }

      return hour;
    }
    return -1;
  }

  function ampmParser(value)
  {
    value = value.toLowerCase().charAt();
    if (value === "a")
      return 0;
    else if (value === "p")
      return 1;
    else
      return -1;
  }

  self.getWheelModel = function(type)
  {
    switch (type)
    {
      case "hour":
        return _hourModel;

      case "minute":
        return _minuteModel;

      case "ampm":        
        return (_12Hour ? _ampmModel : null);
    };
  };

  function mod(val1, val2)
  {
    // Make modulus out of remainder.  The fancy stuff deals with neg values.
    return ((val1 % val2) + val2) % val2;
  }

  function gcd(a, b) {
    if (b === 0)
    {
        return a;
    }
    return gcd(b, a % b);
  }
}


/* global Promise, Hammer */

function createWheel(model, isNumber, classList)
{
  var CURRENT_POSITION = 7;
  var PAN_SPIN_THRESHOLD = 2;
  var TAP_THRESHOLDS = [0.152, 0.362, 0.638, 0.848];
  var MOMENTUM_FACTOR = 0.007;

  var KEYCODE_BACKSPACE = 8;
  var KEYCODE_UP = 38;
  var KEYCODE_DOWN = 40;

  var _wheel;
  var _items = [];
  var _panStartY;
  var _panLastSpinY;
  var _panLastZone;
  var _momentum;
  var _isMeridian = !isNumber && classList === "oj-timepicker-meridian";

  createDom(classList);
  var $wheel = $(_wheel);

  defineMethods();
  defineEvents();
  refresh();

  return _wheel;

  function createDom(classList)
  {
    _wheel = document.createElement("div");
    _wheel.classList.add("oj-timepicker-wheel");
    if (classList)
      _wheel.classList.add(classList);
    _wheel.setAttribute("id", "_ojWheel" + createWheel.counter++);
    _wheel.setAttribute("tabIndex", "0");
    _wheel.setAttribute("role", "spinbutton");
    model.wheel = _wheel;
  }

  function defineMethods()
  {
    _wheel.ojSpinUp = spinUp;
    _wheel.ojSpinDown = spinDown;
    _wheel.ojRefresh = refresh;
    _wheel.ojDestroy = destroy;
    _wheel.ojLinked = function()
    {
      return model["linked"];
    };
  }

  function defineEvents()
  {
    var hammerOptions = {
      "recognizers": [
        [Hammer.Pan, {"direction": Hammer["DIRECTION_VERTICAL"]}],
        [Hammer.Tap],
        [Hammer.Swipe, {"direction": Hammer["DIRECTION_VERTICAL"]}]
      ]
    };

    $wheel.ojHammer(hammerOptions)
    .on("tap", tapHander)
    .on("swipeup", swipeUpHandler)
    .on("swipedown", swipeDownHandler)
    .on("panstart", panStartHandler)
    .on("panend pancancel", panEndHandler)
    .on("panup pandown", panHandler);

    _wheel.addEventListener('wheel', wheelHandler, false);
    _wheel.addEventListener('keydown', keydownHandler, false);
    _wheel.addEventListener('focus', focusHandler, false);
    _wheel.addEventListener('blur', blurHandler, false);
  }

  function destroy()
  {
    $wheel.ojHammer()
    .off("tap", tapHander)
    .off("swipeup", swipeUpHandler)
    .off("swipedown", swipeDownHandler)
    .off("panstart", panStartHandler)
    .off("panend pancancel", panEndHandler)
    .off("panup pandown", panHandler);
    $wheel.ojHammer("destroy");
  }

  function spinUp()
  {
    var next = _items[CURRENT_POSITION + 1];
    if (next)
    {
      spin(1);
      var oldItem = _items.shift();
      if (oldItem)
      {
        _wheel.removeChild(oldItem);
      }
      var newItem = createWheelItem(model, CURRENT_POSITION, _isMeridian);
      if (newItem)
      {
        _wheel.appendChild(newItem); //@HTMLUpdateOK newItem is generated internally using number so ok
      }
      _items.push(newItem);
    }
  }

  function spinDown()
  {
    var prev = _items[CURRENT_POSITION - 1];
    if (prev)
    {
      spin(-1);
      var oldItem = _items.pop();
      if (oldItem)
      {
        _wheel.removeChild(oldItem);
      }
      var newItem = createWheelItem(model, -CURRENT_POSITION, _isMeridian);
      if (newItem)
      {
        _wheel.insertBefore(newItem, _items[0]); //@HTMLUpdateOK newItem is generated internally using number so ok
      }
      _items.unshift(newItem);
    }
  }

  function spin(direction)
  {
    for (var i = 0; i < _items.length; i++)
    {
      var item = _items[i];
      if (item)
      {
        item.ojUpdatePosition(i - CURRENT_POSITION - direction);            
      }
    }

    _wheel.setAttribute("aria-valuenow", model.getText(0));
  }

  function refresh()
  {
    _items.forEach(function(item)
    {
      if (item)
      {
        _wheel.removeChild(item);
      }
    });
    _items = [];

    for (var offset = -CURRENT_POSITION; offset <= CURRENT_POSITION; offset++)
    {
      var item = createWheelItem(model, offset, _isMeridian);
      if (item)
      {
        _wheel.appendChild(item); //@HTMLUpdateOK item is generated internally using number so ok
      }

      _items.push(item);
    };

    // Set the current value on the wheel for accessibility
    _wheel.setAttribute("aria-valuenow", model.getText(0));
  }

  function keydownHandler(event)
  {
    var keyCode = event.keyCode;

    switch (keyCode) {
    case KEYCODE_UP:
      model["position"]++;
      event.preventDefault();
      break;
    case KEYCODE_DOWN:
      model["position"]--;
      event.preventDefault();
      break;
    case KEYCODE_BACKSPACE:
      model.keyboardValue = model.keyboardValue.slice(0,-1);
      event.preventDefault();
      break;

    default:
      if ((keyCode > 47 && keyCode < 58) || // number keys
          (keyCode > 95 && keyCode < 112) || // numpad keys
          (!isNumber &&(keyCode > 64 && keyCode < 91))) // letter keys
      {
        model["keyboardValue"] += event.key;
      }
      break;
    }
  }

  function tapHander(event)
  {
    _wheel.focus();
    var tapY = event["gesture"]["center"].y;
    var wheelTop = $wheel.offset().top;
    var wheelHeight = $wheel.height();
    var tapFraction = (tapY - wheelTop) / wheelHeight;
    var tapZone = 0;

    while ((tapZone < 4) && (tapFraction > TAP_THRESHOLDS[tapZone]))
    {
      tapZone++;
    }

    if (tapZone !== 2)
    {
      model["position"] += (tapZone - 2);
    }
  }
  
  function swipeUpHandler(event)
  {
    _wheel.focus();
    var velocity = event["gesture"].velocityY;
    var extraPixels = velocity * velocity / MOMENTUM_FACTOR;
    _momentum = Math.floor(extraPixels / $wheel.height() * 5);
    event.preventDefault();
  }

  function swipeDownHandler(event)
  {
    _wheel.focus();
    var velocity = event["gesture"].velocityY;
    var extraPixels = velocity * velocity / MOMENTUM_FACTOR;
    _momentum = -Math.floor(extraPixels / $wheel.height() * 5);
    event.preventDefault();
  }

  function panStartHandler(event)
  {
    _wheel.focus();
    _panStartY = _panLastSpinY = event["gesture"]["center"].y;
    _panLastZone = 0;
    _momentum = 0;
  }
 
  function panEndHandler(event)
  {
    _wheel.focus();
    if (_momentum)
    {
      model["position"] += _momentum;
    }
      
    _panStartY = _panLastSpinY = null;
    _panLastZone = null;
  }
 
  function panHandler(event)
  {
    _wheel.focus();
    var panY = event["gesture"]["center"].y;
    var newZone = Math.round(((_panStartY - panY) / $wheel.height()) * 5);
    if (newZone !== _panLastZone && Math.abs(_panLastSpinY - panY) > PAN_SPIN_THRESHOLD)
    {
      _panLastSpinY = panY;
      model["position"] += newZone - _panLastZone;
      _panLastZone = newZone;
    }
    event.preventDefault();
  }

  function wheelHandler(event)
  {
    if (event["deltaY"])
    {
      event.currentTarget.focus();
      event.preventDefault();
    }
    if (event["deltaY"] < 0)
    {
      model["position"]++;
    }
    if (event["deltaY"] > 0)
    {
      model["position"]--;
    }
  }

  function focusHandler()
  {
    model.keyboardValue = "";
    _wheel.classList.add("oj-focus");
  };

  function blurHandler()
  {
    _wheel.classList.remove("oj-focus");
    model.update();
  };
}

createWheel.counter = 0;
/**
 * @protected
 * @ignore
 */
function createWheelGroup(timePickerModel)
{
  var KEYCODE_LEFT = 37;
  var KEYCODE_RIGHT = 39;


  var _group;
  var _wheels = [];

  createDom();
  defineMethods();
  defineEvents();
  refresh();

  return _group;

/**
 * @protected
 * @ignore
 */
  function createDom()
  {
    _group = document.createElement("div");
    _group.classList.add("oj-timepicker-wheel-group");
  }
  
/**
 * @protected
 * @ignore
 */
  function defineMethods()
  {
    _group.ojRefresh = refresh;
  }

/**
 * @protected
 * @ignore
 */
  function defineEvents()
  {
    _group.addEventListener("keydown", keydownHandler, false);
  }

/**
 * @protected
 * @ignore
 */
  function keydownHandler(event)
  {
    var wheel = event.target;
    var keyCode = event.keyCode;

    switch (keyCode) {
    case KEYCODE_LEFT:
      $(wheel).prev().focus();
      break;
    case KEYCODE_RIGHT:
      $(wheel).next().focus();
      break;

    default:
      break;
    }
  }

/**
 * @protected
 * @ignore
 */
  function focusHandler(event)
  {
    var wheel = event.target;
    if (wheel.ojLinked())
    {
      for (var i = 0; i < _wheels.length; i++)
      {
        if (_wheels[i].ojLinked())
        {
          _wheels[i].classList.add("oj-active");
        }
      }
    }
    else
    {
      wheel.classList.add("oj-active");
    }
  }

/**
 * @protected
 * @ignore
 */
  function blurHandler()
  {
    for (var i = 0; i < _wheels.length; i++)
    {
      _wheels[i].classList.remove("oj-active");
    }
  }

/**
 * @protected
 * @ignore
 */
  function refresh()
  {
    var wheel;
    while (wheel = _group.firstChild)
    {
      wheel.ojDestroy();
      _group.removeChild(wheel);
    }
    _wheels = [];

    createWheels();
  }

/**
 * @protected
 * @ignore
 */
  function createWheels()
  {
    var wheelOrder = timePickerModel["wheelOrder"];

    var hourModel = timePickerModel.getWheelModel("hour");
    var hourWheel = createWheel(hourModel, true, "oj-timepicker-hour");
    var minuteModel = timePickerModel.getWheelModel("minute");
    var minuteWheel = createWheel(minuteModel, true, "oj-timepicker-minute");

    var ampmModel;
    var ampmWheel;
    if (wheelOrder.indexOf("a") >= 0)
    {
      ampmModel = timePickerModel.getWheelModel("ampm");
      ampmWheel = createWheel(ampmModel, false, "oj-timepicker-meridian");
    }

    var codes = wheelOrder.split("");
    var i;
    for (i = 0; i < codes.length; i++)
    {
      switch (codes[i])
      {
        case "h":
          _wheels.push(hourWheel);
          break;
        case "m":
          _wheels.push(minuteWheel);
          break;
        case "a":
          _wheels.push(ampmWheel);
          break;
        default:
          console.log("Unknown wheelOrder code: " + codes[i]);
      }
    }
    for (i = 0; i < _wheels.length; i++)
    {
      var wheel = _wheels[i];
      wheel.addEventListener("focus", focusHandler, false);
      wheel.addEventListener("blur", blurHandler, false);
      _group.appendChild(wheel); //@HTMLUpdateOK wheel is generated internally using number so ok
    }
  }
}

function createWheelItem(model, position, isMeridian)
{
  var _item;
  var _position;
  var _disabled = false;

  function updatePosition(newPosition)
  {
    _item.classList.remove("oj-timepicker-wheel-item-position" + _position);
    _item.classList.add("oj-timepicker-wheel-item-position" + newPosition);
    _position = newPosition;
    return Promise.resolve();
  }

  function itemFocusHandler()
  {
    _item.parentNode.focus();
  }

  var text = model.getText(position);
  if (text)
  {
    _item = document.createElement("div");
    _item.classList.add("oj-timepicker-wheel-item");
    _item.classList.add("oj-timepicker-wheel-item-position" + position);
    _position = position;
    if (model.isDisabled(position))
    {
      _disabled = true;
      _item.classList.add("oj-disabled");
    }

    _item.ojUpdatePosition = updatePosition;
    
    var content = document.createElement("div");
    content.textContent = text;
    content.classList.add("oj-timepicker-wheel-item-content");

    if(isMeridian) 
    {
      //need to add title attr for hover for long am/pm translations
      content.setAttribute("title", text);
    }

    _item.appendChild(content); //@HTMLUpdateOK content is generated internally with content being nunmber so ok
 
    Object.defineProperty(_item, "ojDisabled", {
      'enumerable': true,
      'get': function()
      {
        return _disabled;
      },
      'set': function(disabled)
      {
        if (disabled !== _disabled)
        {
          $(_item).toggleClass("oj-disabled");
          _disabled = disabled;
        }
      }
    });
    
    // The item div and its content div can get focus from mouse click on IE,
    // which doesn't happen on other browsers.  Add a focus handler so that
    // focus can be redirected to the wheel.
    _item.addEventListener("focus", itemFocusHandler, false);
    content.addEventListener("focus", itemFocusHandler, false);
  }

  return _item;
}

/**
 * @ignore
 * @constructor
 * @protected
 * @param {TimePickerModel} parentModel parent picker model
 * @param {Object} properties initial property values
 */
function WheelModel(parentModel, properties)
{
  var SPIN_TIMES = [150, 100, 50, 25, 16];  // Note: No transitions for faster spins

  var _value = 0;

  var _position = 0;
  var _currentPosition = 0;
  
  var _displayRange = 1;
  var _valueRange = 1;
  var _increment = 1;
  var _wheelSize;
  var _keyboardValue = "";
  var _min;
  var _max;
  var _wrapped;
  
  var _valueMultiplier;
  var _valueUpperMultiplier;
  var _displayMultiplier = 1;
  var _displayUpperMultiplier;
  
  var _spinning = false;
  var _settingProps = false;
  
  var self = this;

  defineProperties();
  defineMethods();
  setProperties(properties);

  function defineProperties()
  {
    Object.defineProperty(self, "position", {
      'enumerable': true,
      'get': function()
      {
        return _position;
      },
      'set': function(position)
      {
        var val = mod(position, _wheelSize) * _increment;
        if ((val >= _min) && (val < _max) &&
            (position !== _position))
        {
          self["value"] += (position - _position) * _increment;
        }
     }
    });

    Object.defineProperty(self, "value", {
      'enumerable': true,
      'get': function()
      {
        return _value;
      },
      'set': function(value)
      {
        value = Math.round(value / _increment) * _increment;

        if (validValue(value) && _value !== value)
        {
          _value = mod(value, _valueRange);
          setPosition();
          if (self["linked"])
          {
            if (_valueRange == 2)
            {
              // Don't spin the linked wheel when changing am/pm.  Just set the
              // model value and update the DOM structure since we don't want the
              // hour wheel to spin back to the same number.
              parentModel["disableSpin"] = true;
            }
            parentModel["value"] = wheelValueToParentValue(parentModel["value"], _value);
            parentModel["disableSpin"] = false;
          }
        }
      }
    });

    Object.defineProperty(self, "increment", {
      'enumerable': true,
      'get': function()
      {
        return _increment;
      },
      'set': function(increment)
      {
        if (_increment !== increment)
        {
          _increment = increment;
          refreshSettings();
        }
      }
    });

    
    Object.defineProperty(self, "valueMultiplier", {
      'enumerable': true,
      'get': function()
      {
        return _valueMultiplier;
      },
      'set': function(valueMultiplier)
      {
        if (_valueMultiplier !== valueMultiplier)
        {
          _valueMultiplier = valueMultiplier;
          refreshSettings();
        }
      }
    });

    Object.defineProperty(self, "valueRange", {
      'enumerable': true,
      'get': function()
      {
        return _valueRange;
      },
      'set': function(valueRange)
      {
        if (_valueRange !== valueRange)
        {
          _valueRange = valueRange;
          refreshSettings();
        }
      }
    });

    Object.defineProperty(self, "displayMultiplier", {
      'enumerable': true,
      'get': function()
      {
        return _displayMultiplier;
      },
      'set': function(displayMultiplier)
      {
        if (_displayMultiplier !== displayMultiplier)
        {
          _displayMultiplier = displayMultiplier;
          refreshSettings();
        }
      }
    });

    Object.defineProperty(self, "displayRange", {
      'enumerable': true,
      'get': function()
      {
        return _displayRange;
      },
      'set': function(displayRange)
      {
        if (_displayRange !== displayRange)
        {
          _displayRange = displayRange;
          refreshSettings();
        }
      }
    });

    Object.defineProperty(self, "keyboardValue", {
      'enumerable': true,
      'get': function()
      {
        return _keyboardValue;
      },
      'set': function(keyboardValue)
      {
        _keyboardValue = keyboardValue;
        if (self["parser"] && _keyboardValue)
        {
          var value = self["parser"](_keyboardValue);
          if (value >= 0)
          {
            value = displayValueToWheelValue(_value, value);
            if (_min <= value && value < _max)
            {
              self["value"] = value;
            }
          }
          
          // Clear _keyboardValue when the user has typed in 2 digits or after
          // 1 second, so that a new value can be accepted.
          if (_keyboardValue.length >= 2)
          {
            _keyboardValue = "";
          }
          else
          {
            setTimeout(function() {
              _keyboardValue = "";
            }, 1000);
          }
        }
      }
    });
  };

  function defineMethods()
  {
    self.getText = function(position)
    {
      var text;
      var pos = mod(_currentPosition, _wheelSize) + position;
      var haveText = _wrapped || (pos >= 0 && pos < _wheelSize);
      if (self["formatter"] && haveText)
      {
        var val = positionToDisplayValue(_currentPosition + position);
        text = self["formatter"](val);
      }
      return text;
    };

    self.isDisabled = function(position)
    {      
      var value = mod(_currentPosition + position, _wheelSize) * _increment;
      if (_min !== 0 && value < _min)
      {
        return true;
      }
      if (_max !== _valueRange && value >= _max)
      {
        return true;
      }
      return false;
    };

    /*
     * 
     * called by wheel on blur
     */
    self.update = function()
    {
      parentModel["value"] = wheelValueToParentValue(parentModel["value"], _value);
    };

    self.refresh = refresh;
    self.setProperties = setProperties;
  }

  function setProperties(properties)
  {
    if (properties)
    {
      for (var key in properties)
      {
        self[key] = properties[key];
      }
    }
  }

  function refresh()
  {
    var needRefresh = false;
    var parentValue = parentModel.value;
    self["value"] = parentValueToWheelValue(parentValue);

    var parentMax = parentModel["max"];
    var newMax;
    if (parentValueUpperPart(parentValue) === parentValueUpperPart(parentMax))
    {
      newMax = parentValueToWheelValue(parentMax) + 1;
    }
    else
    {
      newMax = _valueRange;
    }
    if (_max !== newMax)
    {
      needRefresh = true;
      _max = newMax;
    }

    var parentMin = parentModel["min"];
    var newMin;
    if (parentValueUpperPart(parentValue) === parentValueUpperPart(parentMin))
    {
      newMin = parentValueToWheelValue(parentMin);
    }
    else
    {
      newMin = 0;
    }
    if (_min !== newMin)
    {
      needRefresh = true;
      _min = newMin;
    }

    var parentIncrement = parentModel["increment"];
    var inc = gcd(parentIncrement, _valueUpperMultiplier);  // For example 60 for minutes

    // If increment is a multiple of 60 then min and max are 0;
    if (inc === _valueUpperMultiplier)
    {
      _min = 0;
      _max = 1;
      needRefresh = true;
    }
    else if (self["linked"] && parentIncrement > _valueMultiplier)
    {
      inc = parentIncrement;
    }
    else if (mod(inc, _valueMultiplier) === 0)
    {
      inc = inc / _valueMultiplier;
    }
    else
    {
      inc = 1;
    }
    if (_increment !== inc)
    {
      _increment = inc;
      needRefresh = true;
    }

    _wheelSize = Math.floor(_valueRange / _increment);
    _wrapped = _wheelSize > 4;
    
    if (self.wheel && needRefresh)
    {
      self.wheel.ojRefresh();
    }
  }

  function setPosition()
  {
    var newPos = mod(wheelValueToPosition(_value), _wheelSize);
    var oldPos = mod(self.position, _wheelSize);
    var diff = newPos - oldPos;
    if (_wrapped)
    {
      if (newPos > oldPos)
      {
        if (oldPos + _wheelSize - newPos < Math.abs(diff))
        {
          diff = newPos - oldPos - _wheelSize;
        }
      }
      else 
      {
        if (newPos + _wheelSize - oldPos < Math.abs(diff))
        {
          diff = newPos + _wheelSize - oldPos;
        }
      }
    }
    if (diff !== 0)
    {
      _position += diff;

      if (self.wheel)
      {
        if (!_spinning)
        {
          _spinning = true;
          spinWheel.call(self);
        }
      }
      else
      {
        _currentPosition = _position;
      }
    }
  }

  /*
   * recalculates dependent values after settings change
   */
  function refreshSettings()
  {
    _valueUpperMultiplier = _valueMultiplier * _valueRange;
    _displayUpperMultiplier = _displayMultiplier * _displayRange;
  }

  function spinWheel()
  {
    if(!self.wheel)
    { //wheel destroyed already
	  return;
	}	
	SPIN_TIMES.forEach(function(time)
    {
      self.wheel.classList.remove("oj-timepicker-wheel-spin-" + time);
    });

    var dist = Math.abs(_position - _currentPosition);
    if (dist === 0)
    {
      _spinning = false;
      return;
    }

    var delay;

    dist--;
    dist = Math.min(dist, SPIN_TIMES.length - 1);
    delay = SPIN_TIMES[dist];
    self.wheel.classList.add("oj-timepicker-wheel-spin-" + SPIN_TIMES[dist]);

    if (_position > _currentPosition)
    {
      _currentPosition++;
      self.wheel.ojSpinUp();
    }
    if (_position < _currentPosition)
    {
      _currentPosition--;
      self.wheel.ojSpinDown();
    }
    if (delay)
    {
      if (parentModel["disableSpin"])
      {
        // Calling spinWheel without delay will update the DOM structure
        // without visually spinning the wheel.
        spinWheel.call(self);
      }
      else
      {
        setTimeout(spinWheel.bind(self), delay);
      }
    }
  }

  function validValue(value)
  {
    if (_wrapped)
      return true;
    
    return (_min <= value && value < _max);
  }

  function mod(val1, val2)
  {
    // Make modulus out of remainder.  The fancy stuff deals with neg values.
    return ((val1 % val2) + val2) % val2;
  }

  function gcd(a, b) {
    if (b === 0)
    {
        return a;
    }
    return gcd(b, a % b);
  }

  function positionToDisplayValue(position)
  {
    return wheelValueToDisplayValue(positionToWheelValue(position));
  }

  function wheelValueToPosition(value)
  {
    return Math.floor(value / _increment);
  }

  function wheelValueToParentValue(parentValue, value)
  {
    return Math.floor(parentValue / _valueUpperMultiplier ) * _valueUpperMultiplier +
                  mod(value, _valueRange) * _valueMultiplier +
                  mod(parentValue, _valueMultiplier);
  }

  function parentValueToWheelValue(value)
  {
    return mod(Math.floor(value / _valueMultiplier), _valueRange);
  }

  function parentValueUpperPart(value)
  {
    return Math.floor(value / _valueUpperMultiplier );
  }

  function positionToWheelValue(position)
  {
    return mod(mod(position, _wheelSize) * _increment, _valueRange);
  }

  function wheelValueToDisplayValue(value)
  {
    return mod(Math.floor(value / _displayMultiplier), _displayRange);
  }

  function displayValueToWheelValue(wheelValue, displayValue)
  {
    return Math.floor(wheelValue / _displayUpperMultiplier ) * _displayUpperMultiplier +
                mod(displayValue, _displayRange) * _displayMultiplier +
                mod(wheelValue, _displayMultiplier);
  }
}

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @private
 */
var dateSwitcherConverter = $["oj"]["ojInputDate"]["prototype"]["options"]["converter"];

/**
 * @private
 */
var timeSwitcherConverter =  $["oj"]["ojInputTime"]["prototype"]["options"]["converter"];

/**
 * @ojcomponent oj.ojDateTimePicker
 * @augments oj.ojInputDateTime
 * @since 4.0.0
 * @ojstatus preview
 * @ojshortdesc An inline element for picking a date-time value.
 * @ojdisplayname Inline Date Time Picker
 * @ojrole combobox
 * 
 * @classdesc
 * <h3 id="inputDateTimeOverview-section">
 *   JET DateTimePicker (Inline mode)
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputDateTimeOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET DateTimePicker extends from DatePicker providing additionally time selection drop down. This behaves similar to JET InputDateTime element
 *    except for the fact that the date picker is rendered inline here.</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-date-time-picker>&lt;/oj-date-time-picker></code></pre>
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate the label to the element.
 * For DateTimePicker, you should put an <code>id</code> on the element, and then set 
 * the <code>for</code> attribute on the label to be the element's id.
 * </p>
 * <h3 id="label-section">
 *   Label and DateTimePicker
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * For accessibility, you should associate a label element with the input
 * by putting an <code>id</code> on the input, and then setting the 
 * <code>for</code> attribute on the label to be the input's id.
 * </p>
 * <p>
 * The element will decorate its associated label with required and help 
 * information, if the <code>required</code> and <code>help</code> attributes are set. 
 * </p> 
 */
/**
 * @ojcomponent oj.ojInputDateTime
 * @augments oj.ojInputDate
 * @since 0.6
 * @ojstatus preview
 * @ojshortdesc Provides basic support for specifying a date-time value.
 * @ojrole combobox
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojInputDateTime<SP extends ojInputDateTimeSettableProperties = ojInputDateTimeSettableProperties> extends ojInputDate<SP>"
 *               }
 *              ]
 * 
 * @classdesc
 * <h3 id="inputDateTimeOverview-section">
 *   JET InputDateTime
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputDateTimeOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET InputDateTime extends from InputDate providing additionally time selection drop down.</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-input-date-time>&lt;/oj-input-date-time></code></pre>
 * 
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate the label to the element.
 * For InputDateTime, you should put an <code>id</code> on the element, and then set 
 * the <code>for</code> attribute on the label to be the element's id.
 * </p>
 * <h3 id="label-section">
 *   Label and InputDateTime
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * For accessibility, you should associate a label element with the input
 * by putting an <code>id</code> on the input, and then setting the 
 * <code>for</code> attribute on the label to be the input's id.
 * </p>
 * <p>
 * The element will decorate its associated label with required and help 
 * information, if the <code>required</code> and <code>help</code> attributes are set. 
 * </p> 
 */
oj.__registerWidget("oj.ojInputDateTime", $['oj']['ojInputDate'], 
{
  widgetEventPrefix : "oj",
  
  //-------------------------------------From base---------------------------------------------------//
  _WIDGET_CLASS_NAMES : "oj-inputdatetime-date-time oj-component oj-inputdatetime",
  _INPUT_HELPER_KEY: "inputHelpBoth",
  //-------------------------------------End from base-----------------------------------------------//
  _TRIGGER_CALENDAR_CLASS : "oj-inputdatetime-calendar-clock-icon",
  
  options : 
  {
    /**
     * Default converter for InputDateTime
     *
     * If one wishes to provide a custom converter for the InputDateTime override the factory returned for
     * oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME)
     * {@ojinclude "name":"inputBaseConverterOptionDoc"}
     * @expose
     * @instance
     * @memberof! oj.ojInputDateTime
     * @type {Object}
     * @ojsignature  { target: "Type", value: "oj.Converter<string>|oj.Validation.FactoryRegisteredValidatorOrConverter|null"}
     * @default oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({"day": "2-digit", "month": "2-digit", "year": "2-digit", "hour": "2-digit", "minute": "2-digit"})
     */
    converter : oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter(
    {
      "day" : "2-digit", "month" : "2-digit", "year" : "2-digit", "hour" : "2-digit", "minute" : "2-digit"
    }),

    /**
     * Allows applications to specify whether to render date picker in JET or
     * as a native picker control. In inline mode, the only value supported is "jet"</br>
     *
     * @expose
     * @memberof! oj.ojDateTimePicker
     * @instance
     * @name renderMode
     * @type {string}
     * @ojvalue {string} 'jet' Applications get full JET functionality.
     * @default "jet"
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">render-mode</code> attribute specified:</caption>
     * &lt;oj-date-time-picker render-mode='jet'>&lt;/oj-date-time-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">renderMode</code> property after initialization:</caption>
     * // getter
     * var renderMode = myInputDate.renderMode;
     *
     * // setter
     * myInputDate.renderMode = 'jet';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeRenderModeOptionDefault: native !default;
     */
    /**
     * Allows applications to specify whether to render date and time pickers 
     * in JET or as a native picker control.</br>
     * 
     * Valid values: jet, native
     * 
     * Default value depends on the theme. In alta-android, alta-ios and alta-windows themes, the 
     * default is "native" and it's "jet" for alta web theme.
     *
     * <ul>
     *  <li> jet - Applications get full JET functionality.</li>
     *  <li> native - Applications get the functionality of the native picker.</li></br>
     *  Note that the native picker support is limited to Cordova plugin published 
     *  at 'https://github.com/VitaliiBlagodir/cordova-plugin-datepicker'.</br>
     *  With native renderMode, the functionality that is sacrificed compared to jet renderMode are:
     *    <ul>
     *      <li>Date and time pickers cannot be themed</li>
     *      <li>Accessibility is limited to what the native picker supports</li>
     *      <li>pickerAttributes is not applied</li>
     *      <li>Sub-IDs are not available</li>
     *      <li>hide() and hideTimePicker() functions are no-op</li>
     *      <li>translations sub properties pertaining to the picker is not available</li>
     *      <li>All of the 'datepicker' sub-properties except 'showOn' are not available</li>
     *      <li>'timePicker.timeIncrement' property is limited to iOS and will only take a precision of minutes</li>
     *    </ul>
     * </ul>
     *
     * @expose 
     * @memberof! oj.ojInputDateTime
     * @instance
     * @type {string}
     * @ojvalue {string} 'jet' Applications get full JET functionality.
     * @ojvalue {string} 'native' Applications get the functionality of the native picker. Native picker is
     *  not available when the picker is inline, defaults to jet instead.</br></br>
     *  Note that the native picker support is limited to Cordova plugin published
     *  at 'https://github.com/VitaliiBlagodir/cordova-plugin-datepicker'.</br></br>
     *  With native renderMode, the functionality that is sacrificed compared to jet renderMode are:
     *    <ul>
     *      <li>Date Time picker cannot be themed</li>
     *      <li>Accessibility is limited to what the native picker supports</li>
     *      <li>pickerAttributes is not applied</li>
     *      <li>Sub-IDs are not available</li>
     *      <li>hide() function is no-op</li>
     *      <li>translations sub properties pertaining to the picker is not available</li>
     *      <li>All of the 'datepicker' sub-properties except 'showOn' are not available</li>
     *      <li>'timePicker.timeIncrement' property is limited to iOS and will only take a precision of minutes</li>
     *    </ul>
     *
     * @example <caption>Get or set the <code class="prettyprint">renderMode</code> property for
     * an InputDateTime after initialization:</caption>
     * // getter
     * var renderMode = myInputDateTime.renderMode;
     * // setter
     * myInputDateTime.renderMode = "native";
     * // Example to set the default in the theme (SCSS)
     * $inputDateTimeRenderModeOptionDefault: native !default;
     */
    renderMode : "jet",
    
    /**
     * <p>
     * Note that Jet framework prohibits setting subset of properties which are object types.<br/><br/>
     * For example myInputDateTime.timePicker = {timeIncrement: "00:30:00:00"}; is prohibited as it will
     * wipe out all other sub-properties for "timePicker" object.<br/><br/> If one wishes to do this [by above syntax or knockout] one
     * will have to get the "timePicker" object, modify the necessary sub-property and pass it to above syntax.<br/><br/>
     * </p>
     *
     * @memberof oj.ojInputDateTime
     * @ojfragment dateTimeCommonTimePicker
     */
    /**
     * {@ojinclude "name":"datePickerCommonDatePicker"}
     *
     * @expose
     * @instance
     * @memberof! oj.ojDateTimePicker
     * @name timePicker
     * @type {Object}
     * @ojtsignore
     *
     * @example <caption>Initialize the component, overriding some time-picker attributes and leaving the others intact:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-date-time-picker time-picker.time-increment='00:10:00:00'>&lt;/oj-date-time-picker>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-date-time-picker time-picker='{"timeIncrement":"00:10:00:00"}'>&lt;/oj-date-time-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">timePicker</code> property after initialization:</caption>
     * // Get one
     * var value = myComponent.timePicker.showOn;
     *
     * // Set one, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myComponent.setProperty('timePicker.showOn', 'image');
     *
     * // Get all
     * var values = myComponent.timePicker;
     *
     * // Set all.  Must list every timePicker key, as those not listed are lost.
     * myComponent.timePicker = {
     *     timeIncrement: '00:10:00:00',
     *     showOn: 'image'
     * };
     */
    /**
     * <p>
     * Note that Jet framework prohibits setting subset of properties which are object types.<br/><br/>
     * For example myInputDateTime.timePicker = {timeIncrement: "00:30:00:00"}; is prohibited as it will
     * wipe out all other sub-properties for "timePicker" object.<br/><br/> If one wishes to do this [by above syntax or knockout] one
     * will have to get the "timePicker" object, modify the necessary sub-property and pass it to above syntax.<br/><br/>
     * </p>
     *
     * @expose
     * @instance
     * @memberof! oj.ojInputDateTime
     * @type {Object}
     *
     * @example <caption>Initialize the component, overriding some time-picker attributes and leaving the others intact:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-input-date-time time-picker.time-increment='00:10:00:00'>&lt;/oj-input-date-time>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-input-date-time time-picker='{"timeIncrement":"00:10:00:00"}'>&lt;/oj-input-date-time>
     *
     * @example <caption>Get or set the <code class="prettyprint">timePicker</code> property after initialization:</caption>
     * // Get one
     * var value = myComponent.timePicker.showOn;
     *
     * // Set one, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myComponent.setProperty('timePicker.showOn', 'image');
     *
     * // Get all
     * var values = myComponent.timePicker;
     *
     * // Set all.  Must list every timePicker key, as those not listed are lost.
     * myComponent.timePicker = {
     *     timeIncrement: '00:10:00:00',
     *     showOn: 'image'
     * };
     */
    timePicker:
    {
      /**
       * Time increment to be used for InputDateTime, the format is hh:mm:ss:SS. <br/><br/>
       * Note that when renderMode is 'native', timeIncrement property is limited to iOS and will only take a precision of minutes.<br/><br/> 
       *
       * <p>See the <a href="#timePicker">time-picker</a> attribute for usage examples.
       *
       * @expose
       * @name timePicker.timeIncrement
       * @memberof! oj.ojInputDateTime
       * @instance
       * @type {string}
       * @default "00:05:00:00"
       */
      timeIncrement : "00:05:00:00",

      /**
       * When the timepicker should be shown.
       *
       * <p>See the <a href="#timePicker">time-picker</a> attribute for usage examples.
       *
       * @expose
       * @name timePicker.showOn
       * @memberof! oj.ojInputDateTime
       * @instance
       * @type {string}
       * @ojvalue {string} 'focus' when the element receives focus or when the trigger clock image is clicked. When the picker is closed, the field regains focus and is editable.
       * @ojvalue {string} 'image' when the trigger clock image is clicked
       * @default "focus"
       */
      showOn : "focus"
    }
    
    /**
     * The maximum selectable datetime. When set to null, there is no maximum.
     *
     * <ul>
     *  <li> type string - ISOString
     *  <li> null - no limit
     * </ul>
     * 
     * @example <caption>Initialize the element with the <code class="prettyprint">max</code> attribute:</caption>
     * &lt;oj-date-time-picker max='2014-09-25T13:30:00.000-08:00'&gt;&lt;/oj-date-time-picker&gt;
     * 
     * @expose
     * @name max
     * @instance
     * @memberof! oj.ojDateTimePicker
     * @type {string|null}
     * @default null
     */
    /**
     * The maximum selectable datetime. When set to null, there is no maximum.
     *
     * <ul>
     *  <li> type string - ISOString
     *  <li> null - no limit
     * </ul>
     * 
     * @example <caption>Initialize the element with the <code class="prettyprint">max</code> attribute:</caption>
     * &lt;oj-input-date-time max='2014-09-25T13:30:00.000-08:00'&gt;&lt;/oj-input-date-time&gt;
     * 
     * @expose
     * @name max
     * @instance
     * @memberof! oj.ojInputDateTime
     * @type {string|null}
     * @default null
     */
    
    /**
     * The minimum selectable date. When set to null, there is no minimum.
     *
     * <ul>
     *  <li> type string - ISOString
     *  <li> null - no limit
     * </ul>
     * 
     * @example <caption>Initialize the element with the <code class="prettyprint">min</code> attribute:</caption>
     * &lt;oj-date-time-picker min='2014-08-25T08:00:00.000-08:00'&gt;&lt;/oj-date-time-picker&gt;
     * 
     * @expose
     * @name min
     * @instance
     * @memberof! oj.ojDateTimePicker
     * @type {string|null}
     * @default null
     */
    /**
     * The minimum selectable date. When set to null, there is no minimum.
     *
     * <ul>
     *  <li> type string - ISOString
     *  <li> null - no limit
     * </ul>
     * 
     * @example <caption>Initialize the element with the <code class="prettyprint">min</code> attribute:</caption>
     * &lt;oj-input-date-time min='2014-08-25T08:00:00.000-08:00'&gt;&lt;/oj-input-date-time&gt;
     * 
     * @expose
     * @name min
     * @instance
     * @memberof! oj.ojInputDateTime
     * @type {string|null}
     * @default null
     */
    
    /** 
     * List of validators used by element when performing validation. Each item is either an 
     * instance that duck types {@link oj.Validator}, or is an Object literal containing the 
     * properties listed below. Implicit validators created by an element when certain properties 
     * are present (e.g. <code class="prettyprint">required</code> property), are separate from 
     * validators specified through this property. At runtime when the element runs validation, it 
     * combines the implicit validators with the list specified through this property. 
     * <p>
     * Hints exposed by validators are shown in the notewindow by default, or as determined by the 
     * 'validatorHint' property set on the <code class="prettyprint">displayOptions</code> 
     * property. 
     * </p>
     * 
     * <p>
     * When <code class="prettyprint">validators</code> property changes due to programmatic 
     * intervention, the element may decide to clear messages and run validation, based on the 
     * current state it is in. </br>
     * 
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>The cached list of validator instances are cleared and new validator hints is pushed to 
     * messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if element is valid when validators changes, element does nothing other than the 
     * steps it always performs.</li>
     * <li>if element is invalid and is showing messages when 
     * <code class="prettyprint">validators</code> changes then all element messages are cleared 
     * and full validation run using the display value on the element. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   property is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code> 
     *   event to clear custom errors.</li>
     * </ul>
     * </li>
     * <li>if element is invalid and has deferred messages when validators changes, it does 
     * nothing other than the steps it performs always.</li>
     * </ul>
     * </p>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the element are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * </p>
     * 
     * @property {string} type - the validator type that has a {@link oj.ValidatorFactory} that can 
     * be retrieved using the {@link oj.Validation} module. For a list of supported validators refer 
     * to {@link oj.ValidatorFactory}. <br/>
     * @property {Object=} options - optional Object literal of options that the validator expects. 
     * 
     * @example <caption>Initialize the element with validator object literal:</caption>
     * myInputDateTime.validators = [{
     *     type: 'dateTimeRange', 
     *     options : {
     *       max: '2014-09-10T13:30:00.000',
     *       min: '2014-09-01T00:00:00.000'
     *     }
     *   }];
     * 
     * NOTE: oj.Validation.validatorFactory('dateTimeRange') returns the validator factory that is used 
     * to instantiate a range validator for dateTime.
     * 
     * @example <caption>Initialize the element with multiple validator instances:</caption>
     * var validator1 = new MyCustomValidator({'foo': 'A'}); 
     * var validator2 = new MyCustomValidator({'foo': 'B'});
     * myInputDateTime.value = 10;
     * myInputDateTime.validators = [validator1, validator2];
     * 
     * @expose 
     * @name validators
     * @instance
     * @memberof oj.ojInputDateTime
     * @ojsignature  { target: "Type", value: "Array<oj.Validator<string>|oj.Validation.FactoryRegisteredValidatorOrConverter>|null"}
     * @type {Array|undefined}
     */
    
    /**
     * Determines if keyboard entry of the text is allowed.
     * When the datepicker is inline, the only supported value is "disabled".
     *
     * @expose
     * @instance
     * @memberof! oj.ojDateTimePicker
     * @name keyboardEdit
     * @type {string}
     * @ojvalue {string} "disabled" Changing the date can only be done with the picker.
     * @default "disabled"
     *
     * @example <caption>Initialize the InputDate with the <code class="prettyprint">keyboard-edit</code> attribute specified:</caption>
     * &lt;oj-date-time-picker keyboard-edit='disabled'>&lt;/oj-date-time-picker>
     *
     * @example <caption>Get or set the <code class="prettyprint">keyboardEdit</code> property after initialization:</caption>
     * // getter
     * var keyboardEdit = myInputDate.keyboardEdit;
     *
     * // setter
     * myInputDate.keyboardEdit = 'disabled';
     *
     * @example <caption>Set the default in the theme (SCSS)</caption>
     * $inputDateTimeKeyboardEditOptionDefault: disabled !default;
     */
    /** 
     * The value of the DateTimePicker element which should be an ISOString
     * 
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> attribute:</caption>
     * &lt;oj-date-time-picker value='2014-09-10T13:30:00.000'&gt;&lt;/oj-date-time-picker&gt;
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> property specified programmatically 
     * using oj.IntlConverterUtils.dateToLocalIso :</caption>
     * myInputDateTime.value = oj.IntlConverterUtils.dateToLocalIso(new Date());<br/>
     * @example <caption>Get or set the <code class="prettyprint">value</code> property, after initialization:</caption>
     * // Getter: returns Today's date in ISOString
     * myInputDateTime.value;
     * // Setter: sets it to a different date time
     * myInputDateTime.value = "2013-12-01T20:00:00-08:00";
     * 
     * @expose 
     * @name value
     * @instance
     * @memberof! oj.ojDateTimePicker
     * @type {string}
     */
    /** 
     * The value of the InputDateTime element which should be an ISOString
     * 
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> attribute:</caption>
     * &lt;oj-input-date-time value='2014-09-10T13:30:00.000'&gt;&lt;/oj-input-date-time&gt;
     * @example <caption>Initialize the element with the <code class="prettyprint">value</code> property specified programmatically 
     * using oj.IntlConverterUtils.dateToLocalIso :</caption>
     * myInputDateTime.value = oj.IntlConverterUtils.dateToLocalIso(new Date());<br/>
     * @example <caption>Get or set the <code class="prettyprint">value</code> property, after initialization:</caption>
     * // Getter: returns Today's date in ISOString
     * myInputDateTime.value;
     * // Setter: sets it to a different date time
     * myInputDateTime.value = "2013-12-01T20:00:00-08:00";
     * 
     * @expose 
     * @name value
     * @instance
     * @memberof! oj.ojInputDateTime
     * @type {string}
     */

    // Events

  },
  
  /**
   * @protected
   * @override
   * @instance
   * @ignore
   * @memberof! oj.ojInputDateTime
   */
  _InitBase : function () 
  {
    this._super();
    
    this._timePickerElement = this.element; //if the ojInputDateTime is inline, then this ref will change to a NEW input element
    this._timePicker = null;
    this._timeConverter = null;

    //need to remember the last _SetValue for the case of timepicker [i.e. select a date that is not in range due to 
    //time; however since we don't push invalid values to this.options["value"] the timepicker would pick up the wrong 
    //selected date
    this._previousValue = null;

    //below is when the switcher is active (i.e. datetimepicker and when one clicks on Cancel need to reset it)
    this._switcherDiv = null;
    this._switcherPrevValue = null;
    this._switcherPrevDay = null;
    this._switcherPrevMonth = null;
    this._switcherPrevYear = null;

    this._switcherDateValue = null;
    this._switcherTimeValue = null;
    this._dateTimeSwitcherActive = null;
    
    if(!this._isInLine) 
    {
      this._switcherDiv = $(this._generateSwitcher());

      $(".oj-popup-content", this._popUpDpDiv.ojPopup("widget")).append(this._switcherDiv); //@HTMLUpdateOK

      var self = this;
      var cancelHandler = function() {
        //only time when value is _Set in switcher is by clicking Done where both of the fields would be set to null
        if(self._switcherDateValue !== null || self._switcherTimeValue !== null) 
        {
          self._currentDay = self._switcherPrevDay;
          self._drawMonth = self._currentMonth = self._switcherPrevMonth;
          self._drawYear = self._currentYear = self._switcherPrevYear;
          self._SetValue(self._switcherPrevValue);

          self._switcherTimeValue = self._switcherDateValue = self._switcherPrevDay = self._switcherPrevMonth = 
                                    self._switcherPrevYear = self._switcherPrevValue = null;
        }
      };
      this._popUpDpDiv.on("ojclose", function( event, ui ) {
        cancelHandler();
      });
      this._switcherDiv.find("[data-handler]").map(function ()
        {
          var handler =
          {
            /** @expose */
            switchMe : function (evt)
            {
              var keyCode = evt.keyCode;
              if((evt.type === "keydown" && ($.ui.keyCode.ENTER === keyCode || $.ui.keyCode.SPACE === keyCode)) || evt.type === "click") 
              {
                if(!self._isShowingDatePickerSwitcher())
                {
                  self._timePicker.ojInputTime("show");
                }
                else
                {
                  self.show();
                  self._placeFocusOnCalendar();
                }

                return false;
              }
            },
            /** @expose */
            switchDone : function (evt)
            {
              var keyCode = evt.keyCode;
              if((evt.type === "keydown" && ($.ui.keyCode.ENTER === keyCode || $.ui.keyCode.SPACE === keyCode)) || evt.type === "click") 
              {
                var newVal = self._getDateIso();
                if(self._switcherDateValue) 
                {
                  newVal = self._switcherDateValue;
                }

                if(self._switcherTimeValue) 
                {
                  newVal = oj.IntlConverterUtils._copyTimeOver(self._switcherTimeValue, newVal);
                }

                var formatted = self._GetConverter()["format"](newVal);
                self._SetDisplayValue( formatted );
                self._SetValue(formatted, {});

                self._switcherTimeValue = self._switcherDateValue = self._switcherPrevDay = self._switcherPrevMonth = 
                  self._switcherPrevYear = self._switcherPrevValue = null;
                self._hide(self._ON_CLOSE_REASON_SELECTION);
                return false;
              }
              
            },
            /** @expose */
            switchCancel : function (evt)
            {
              var keyCode = evt.keyCode;
              if((evt.type === "keydown" && ($.ui.keyCode.ENTER === keyCode || $.ui.keyCode.SPACE === keyCode)) || evt.type === "click") 
              {
                self._hide(self._ON_CLOSE_REASON_CANCELLED);
                return false;
              }
            }
          };

          $(this).bind(this.getAttribute("data-event"), handler[this.getAttribute("data-handler")]);
        });
    }
    
  },
  
  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDateTime
   */
  _ComponentCreate : function ()
  {
    var ret = this._super();
    var timeConverter = this._getTimePickerConverter(this._GetConverter());
    
    if (timeConverter === null)
    {
      throw new Error("Please use ojInputDate if you do not have time portion");
    }
    
    if (this._isInLine)
    {
      //Since DatePicker never intended to have timepicker associated to it
      //need to have an input element that is tied to the time selector
      
      var input = $("<input type='text'>");
      input.insertAfter(this.element); //@HTMLUpdateOK
      
      //Now need to reset this._timePickerElement to the newly created input element
      this._timePickerElement = input;
    }
    else
    {
      var switcherButtons = this._switcherDiv.find("a");
      this._AddHoverable(switcherButtons);
      this._AddActiveable(switcherButtons);
    }
    
    var passOptions = ["title", "placeholder", "disabled", "required", "readOnly", 
                        "keyboardEdit", "pickerAttributes"];
    var passObject = {};
        
    for(var i=0, j=passOptions.length; i < j; i++) 
    {
      passObject[passOptions[i]] = this.options[passOptions[i]];
    }
    
    var messagesDisplayOption = this.options['displayOptions']['messages'];
    
    //create time instance for the time portion
    // jmw Seems to be a bug where messages are always in notewindow. So I think I should
    // carry the displayOptions over to the timePicker.

    var timePickerOptions = this.options["timePicker"];

    if(!this._isInLine) 
    {
      $.extend(timePickerOptions, {"footerLayout": "now"});
    }

    this._timePicker = this._timePickerElement.ojInputTime(
    $.extend(passObject, {
      "converter" : timeConverter,
      "displayOptions" : {"converterHint": "none", "title": "none", "messages": messagesDisplayOption},
      //need to pass the value down as otherwise if the value is null then it might pickup this.element.val() from 
      //our frameworks generic if options.value is not defined then pick up from element; however that would be a formatted 
      //value from ojInputDateTime
      "value": this.options["value"], 
      "timePicker" : timePickerOptions, 
      "datePickerComp" : {"widget": this, "inline": this._isInLine} 
    }));
    
    return ret;
  },

  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDateTime
   */
  _AfterCreate : function ()
  {
    var ret = this._superApply(arguments);

    this._dateTimeSwitcherActive = !isPickerNative(this) && this._timePicker && !this._isInLine;
    return ret;
  },
  
  _setOption : function (key, value, flags)
  {
    var retVal = this._superApply(arguments);
    
    if (key === "value") 
    {
      //if goes through model does it needs to update or should be only used by selection + keydown
      this._previousValue = this.options["value"]; //get from this.options["value"] as would be cleaned up by editablevalue
    }
    
    if(this._timePicker) 
    {
      //note that min + max are not passed through since it should be taken care of by ojInputDateTime and not ojInputTime
      //as it needs to use the fulle datetime
      var timeInvoker = {"disabled": true, "readOnly": true, "keyboardEdit": true};
      
      if (key in timeInvoker)
      {
        this._timePicker.ojInputTime("option", key, value);
      }
      else if(key === "timePicker")
      {
        this._timePicker.ojInputTime("option", "timePicker.timeIncrement", value["timeIncrement"]);
      }
      else if (key === "converter")
      {
        this._timeConverter = null;
        this._timePicker.ojInputTime("option", key, this._getTimePickerConverter(this._GetConverter())); //need to invoke _GetConverter for the case when null and etc sent in
      }
    }
    
    return retVal;
  },
  
  /**
   * @ignore
   * @protected
   * @override
   * @memberof oj.ojInputDateTime
   */
  _destroy : function ()
  {
    this._cleanUpDateTimeResources();
    return this._super();
  },
  /**
   * @ignore
   * @private
   * @memberof oj.ojInputDateTime
   */
  _cleanUpDateTimeResources : function ()
  {
    this._timePicker.ojInputTime("destroy");
    if (this._isInLine)
    {
      //note that this.element below would be of the TimePicker's input element
      this._timePickerElement.remove();
    }
    else 
    {
      var switcherButtons = this._switcherDiv.find("a");
      this._RemoveActiveable(switcherButtons);
      this._RemoveHoverable(switcherButtons);

      if (this._switcherDiv)
      {
        this._switcherDiv.remove();
      }
    }
  },
  /**
   * @protected
   * @override
   * @instance
   * @ignore
   * @memberof! oj.ojInputDateTime
   */
  _GetCalendarTitle : function ()
  {
    return this._EscapeXSS(this.getTranslatedString("tooltipCalendarTime" + (this.options["disabled"] ? "Disabled" : "")));
  },

  /**
   * Generate the HTML for the footer of the date picker.
   *
   * @ignore
   */
  _generateSwitcher : function()
  {
    var ret = "<div class='oj-datetimepicker-switcher'>" +
                "<div data-handler='switchMe' data-event='click keydown'>" +
                  "<div class='oj-inputdatetime-time-icon oj-clickable-icon-nocontext oj-component-icon oj-enabled oj-default'></div>" +
                  "<a onclick='return false;' href='#' class='oj-enabled oj-datetimepicker-switcher-text' role='button'></a>" +
                "</div>" +
                "<div class='oj-datetimepicker-switcher-buttons'>" +
                  "<a onclick='return false;' href='#' class='oj-enabled' data-handler='switchDone' data-event='click keydown' role='button'>" + this._EscapeXSS(this.getTranslatedString("done")) + "</a>" +
                  "<a onclick='return false;' href='#' class='oj-enabled' data-handler='switchCancel' data-event='click keydown' role='button'>" + this._EscapeXSS(this.getTranslatedString("cancel")) + "</a>" +
                "</div>" +
            "</div>";

    return ret;
  },

  /**
   * @ignore
   */
  _updateSwitcherText: function()
  {
    var switcherText = "";
    if(this._isShowingDatePickerSwitcher())
    {
      switcherText = dateSwitcherConverter.format(this._switcherDateValue || this._getDateIso());
    }
    else
    {
      switcherText = timeSwitcherConverter.format(this._switcherTimeValue || this._getDateIso());
    }

    $(".oj-datetimepicker-switcher-text", this._switcherDiv).text(switcherText);
  },

  /**
   * @ignore
   */
  _isShowingDatePickerSwitcher: function() 
  {
    return !this._switcherDiv || $(".oj-inputdatetime-time-icon", this._switcherDiv).length === 0;
  },

  _togglePicker: function()
  {
    var datepickerNode;
    var timepickerNode;

    var removeCss;
    var addCss;
    var newText;
    var switcher = this._switcherDiv;
    var timePickerShown = !this._isShowingDatePickerSwitcher();

    if(timePickerShown)
    {
      addCss = 'oj-inputdatetime-calendar-icon';
      removeCss = 'oj-inputdatetime-time-icon';
      newText = 'Set Date';
    }
    else
    {
      addCss = 'oj-inputdatetime-time-icon';
      removeCss = 'oj-inputdatetime-calendar-icon';
      newText = 'Set Time';
    }

    var children = $(switcher.children()[0]).children();
    $(children[0]).removeClass(removeCss).addClass(addCss);
    $(children[1]).text(newText);

    children = $(".oj-popup-content", this._popUpDpDiv.ojPopup("widget")).children();

    if(children[0].classList.contains("oj-datepicker-popup")) 
    {
      datepickerNode = $(children[0]);
      timepickerNode = $(children[1]);
    }
    else 
    {
      datepickerNode = $(children[1]);
      timepickerNode = $(children[0]);
    }

    if(!timePickerShown)
    {
      timepickerNode.css("display", "none");
      datepickerNode.css("display", "block");
    }
    else
    {
      datepickerNode.css("display", "none");
      timepickerNode.css("display", "block");
    }
    
    this._updateSwitcherText();
  },

  /*
   * Will provide the timePicker converter based on the actual converter
   */
  _getTimePickerConverter : function (converter) 
  {
    if(this._timeConverter !== null) 
    {
      return this._timeConverter;
    }

    this._timeConverter = _getTimePickerConverter(converter);
    return this._timeConverter;
  },
  
  /**
   * Handler for when the time is selected. Should be invoked ONLY by the ojInputTime component
   * 
   * @ignore
   * @param {string} newValue
   * @param {Event} event
   */
  timeSelected : function (newValue, event)
  {
    //TEMP TILL FIXED pass in formatted for _SetValue (should be newValue)
    if(!this._dateTimeSwitcherActive)
    {
      var formatted = this._GetConverter()["format"](newValue);
      this._SetDisplayValue( formatted );
      this._SetValue(formatted, event);
    }
    else
    {
      this._switcherTimeValue = newValue;
    }
  },
  
  /**
   * Provides the current displayed selected value for ojInputTime component [i.e. when is invalid return this._previousValue]
   * The complication occurs b/c we do not push invalid values to the model and b/c of that reason this.options["value"] 
   * might contain outdated isoString for ojInputTime. For instance let's say the min date is 02/01/14 2PM then 
   * when an user selects 02/01/14 the component would be invalid [as 12AM] and the value would not be pushed. However one needs 
   * to give opportunity for ojInputTime to allow user in selecting the valid datetime in full so the _previousValue 
   * must be passed through.
   * 
   * @ignore
   */
  getValueForInputTime : function ()
  {
    var value = null;

    if(this.isValid()) 
    {
      value = this.options["value"];
    }
    else 
    {
      if(this._previousValue) 
      {
        try 
        {
          //might have been that the user typed in an incorrect format, so try to parse it
          value = this._GetConverter()["parse"](this._previousValue);
        }
        catch(e) 
        {
          value = this.options["value"];
        }
      }
      else 
      {
        value = null;
      }
    }

    if(this._isDateTimeSwitcher()) 
    {
      value = this._switcherDateValue || value;

      if(this._switcherTimeValue && value)
      {
        value = oj.IntlConverterUtils._copyTimeOver(this._switcherTimeValue, value);
      }
    }

    return value;
  },
  
  /**
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDateTime
   */
  _SetValue : function (newValue, event, options)
  {
    var ret = this._superApply(arguments);
    
    this._previousValue = newValue;

    return ret;
  },

  /**
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDateTime
   */
  _SetDisplayValue : function (displayValue)
  {
    if(!this._isInLine)
    {
      this._superApply(arguments);
      this._updateSwitcherText();
    }
    else
    {
      // When the picker is inline, we should update the picker when the value changes so that
      // the selected day has the correct CSS class set.
      var focusOnCalendar = !(this._isInLine && this._timePicker && this._timePicker[0] === document.activeElement);
      this._updateDatepicker(focusOnCalendar);
    }
    this._switcherTimeValue = null;
  },
  
  /**
   * Just for the case of launching timepicker with Shift + Up or Shift + Down
   * 
   * @ignore
   * @protected
   * @override
   * @param {Event} event
   * @instance
   * @memberof! oj.ojInputDateTime
   */
  _onKeyDownHandler : function (event) 
  {
    var kc = $.ui.keyCode;
    var handled = false;
    
    switch (event.keyCode)
    {
      case kc.UP: ;
      case kc.DOWN:
        if(event.shiftKey)
        {
          this._SetValue(this._GetDisplayValue(), event);

          this._timePicker.ojInputTime("show");
          handled = true;
        }
        break;
      default: ;
    }

    if (handled)
    {
      event.preventDefault();
      event.stopPropagation();
      return false;
    }
    
    return this._superApply(arguments);
  },
  
  /**
   * @instance
   * @memberof oj.ojInputDateTime
   * @return {void}
   */
  show : function ()
  {
    if(this._isShowingDatePickerSwitcher()) 
    {
      this._togglePicker();
    }
    
    return this._super();
  },

  /**
   * Shows the native datepicker
   *
   * @protected
   * @override
   * @memberof! oj.ojInputDateTime
   * @instance
   */
  _ShowNativeDatePicker : function (pickerOptions)
  {
    // override the mode set by base class
    pickerOptions['mode'] = "datetime";
    
    var splitIncrements = splitTimeIncrement(this.options["timePicker"]["timeIncrement"]);
    
    // native picker supports only minute interval and only on iOS, we consider minute interval 
    //  only when hours is not specified
    pickerOptions.minuteInterval = (splitIncrements.hourIncr === 0) ? splitIncrements.minuteIncr : 1;
    
    return this._super(pickerOptions);
  },

  /**
   * callback upon picking date from native picker
   *
   * @protected
   * @override
   * @memberof! oj.ojInputDateTime
   * @instance
   */
  _OnDatePicked : function (date)
  {
    this._nativePickerShowing = false;
    
    // for iOS and windows, from the current implementation of the native datepicker plugin,
    //  for case when the picker is cancelled, this callback gets called without the parameter
    if (date)
    {
      var isoString = oj.IntlConverterUtils._dateTime(this._getDateIso(), {"month": date.getMonth(), "date": date.getDate(), "fullYear": date.getFullYear(), 
                                            "hours": date.getHours(), "minutes": date.getMinutes(), "seconds": date.getSeconds()});
      var formattedTime = this._GetConverter().format(isoString);

      // _SetValue will inturn call _SetDisplayValue
      this._SetValue(formattedTime, {});
    }
    
    this._onClose(this._ON_CLOSE_REASON_SELECTION);
  },

  /**
   * Shows the HTML datepicker
   *
   * @override
   * @ignore
   */
  _ShowHTMLDatePicker : function ()
  {
    var retVal = this._superApply(arguments)

    // if there is a value, save the Iso date version as the default value. Otherwise, set to ""
    // so that we correctly restore the empty string if there is no value.
    this._switcherPrevValue = this.options["value"] ? this._getDateIso() : "";
    this._switcherPrevDay = this._currentDay;
    this._switcherPrevMonth = this._currentMonth;
    this._switcherPrevYear = this._currentYear;

    this._updateSwitcherText();
    return retVal;
  },
  
  /**
   * Method to show the internally created InputTime
   * 
   * @expose
   * @memberof oj.ojInputDateTime
   * @instance
   * @return {void}
   */
  showTimePicker : function ()
  {
    if(!this._datepickerShowing())
    {
      this.show();
    }

    if(!this._isShowingDatePickerSwitcher()) 
    {
      this._timePicker.ojInputTime("show");
    }
  },
  
  /**
   * @expose
   * @memberof oj.ojInputDateTime
   * @instance
   * @return {void}
   */
  hideTimePicker : function ()
  {
    return this.hide();
  },
  
  /** 
   * Refreshes the element. Usually called after dom changes have been made.
   * @expose
   * @override
   * @instance
   * @memberof oj.ojInputDateTime
   * @return {void}
   */
  refresh : function ()
  {
    var retVal = this._superApply(arguments) || this;
    
    this._timePicker.ojInputTime("refresh");

    if(this._switcherDiv)
    {
      $("a[data-handler='switchDone']", this._switcherDiv).text(this.getTranslatedString("done"));
      $("a[data-handler='switchCancel']", this._switcherDiv).text(this.getTranslatedString("cancel"));
    }
    
    return retVal;
  },
  
  // @inheritdoc
  getNodeBySubId: function(locator)
  {
    var subId = locator && locator['subId'];
    var node = null;
    
    if(subId) 
    {
      if(subId === "oj-inputdatetime-date-input") 
      {
        node = this._isInLine ? this._timePickerElement[0] : this.element[0];
      }
      else if(subId === "oj-inputdatetime-calendar-clock-icon") 
      {
        node = $(".oj-inputdatetime-calendar-clock-icon", this._triggerNode)[0];
      }
      else
      {
        node = this._timePicker.ojInputTime("getNodeBySubId", locator);
      }
    }
    
    return node || this._superApply(arguments);
  },
  
  // @inheritdoc
  getSubIdByNode: function(node)
  {
    var dateTimeSpecific = null;
    
    if(this._isInLine) 
    {
      if(node === this._timePickerElement[0]) 
      {
        dateTimeSpecific = "oj-inputdatetime-date-input";
      }
    }
    else 
    {
      if(node === this.element[0]) 
      {
        dateTimeSpecific = "oj-inputdatetime-date-input";
      }
      else if(node === $(".oj-inputdatetime-calendar-clock-icon", this._triggerNode)[0])
      {
        dateTimeSpecific = "oj-inputdatetime-calendar-clock-icon";
      }
    }
    
    return dateTimeSpecific || this._timePicker.ojInputTime("getSubIdByNode", node) || this._superApply(arguments);
  },
  
  /**
   * Need to override since apparently we allow users to set the converter to null, undefined, and etc and when 
   * they do we use the default converter
   * 
   * @return {Object} a converter instance or null
   * 
   * @memberof! oj.ojInputDateTime
   * @instance
   * @protected
   * @override
   */
  _GetConverter : function () 
  {
    return this.options['converter'] ? 
            this._superApply(arguments) :
            $["oj"]["ojInputDateTime"]["prototype"]["options"]["converter"];
  },
  
  /**
   * Notifies the component that its subtree has been removed from the document programmatically after the component has
   * been created
   * @memberof! oj.ojInputDateTime
   * @instance
   * @protected 
   */
  _NotifyDetached: function()
  {
    if(this._timePicker) 
    {
      this.hideTimePicker();
    }

    // hide sets focus to the input, so we want to call super after hide. If we didn't, then
    // the messaging popup will reopen and we don't want that.
    this._superApply(arguments);
  },

  /**
   * Notifies the component that its subtree has been made hidden programmatically after the component has
   * been created
   * @memberof! oj.ojInputDateTime
   * @instance
   * @protected 
   */
  _NotifyHidden: function()
  {
    if(this._timePicker) 
    {
      this.hideTimePicker();
    }

    // hide sets focus to the input, so we want to call super after hide. If we didn't, then
    // the messaging popup will reopen and we don't want that.
    this._superApply(arguments);
  },
  
  /**
   * 
   * @return {Object} jquery object 
   * 
   * 
   * @expose
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDateTime
   */
  _GetMessagingLauncherElement : function ()
  {
    return !this._isInLine ? this._super() : this._timePickerElement;
  },
  
  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDateTime
   * @return {string}
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-inputdatetime";
  },
  
  /**
   * @protected
   * @override
   * @instance
   * @memberof! oj.ojInputDateTime
   */
  _GetTranslationsSectionName: function()
  {
    return "oj-ojInputDate"; 
  }
});
  
// Fragments:

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Input element and calendar trigger icon</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>When not inline, shows the grid and moves the focus into the expanded date grid.</td>
 *     </tr>
 *     <tr>
 *       <td>Time trigger icon</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Shows the time picker and moves the focus into the expanded time picker</td>
 *     </tr>
 *     <tr>
 *       <td>Input element with picker open</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Set focus to the input. If hints, title or messages exist in a notewindow, 
 *        pop up the notewindow.</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Left</kbd></td>
 *       <td>Switch to next month (or previous month on RTL page).</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Right</kbd></td>
 *       <td>Switch to previous month (or next month on RTL page).</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputDateTime
 */
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Time trigger icon</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Shows the time picker and moves the focus into the expanded time picker</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Left</kbd></td>
 *       <td>Switch to next month (or previous month on RTL page).</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Swipe Right</kbd></td>
 *       <td>Switch to previous month (or next month on RTL page).</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojDateTimePicker
 */
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>DownArrow or UpArrow</kbd></td>
 *       <td>When not in inline mode, shows the calendar grid and moves the focus into the 
 *       expanded grid. When in inline mode, shows the time picker and moves the focus into the 
 *       expanded time picker</td>
 *     </tr>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>Shift + DownArrow or UpArrow</kbd></td>
 *       <td>Shows the time picker and moves the focus into the expanded time picker</td>
 *     </tr>
 *     <tr>
 *       <td>Input element</td>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Close the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Input Element</td>
 *       <td><kbd>Tab In</kbd></td>
 *       <td>Set focus to the input. If hints, title or messages exist in a notewindow, 
 *        pop up the notewindow.</td>
 *     </tr> 
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Select the currently focused day</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move up in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move down in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move right in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move left in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Close the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Home</kbd></td>
 *       <td>Move focus to first day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>End</kbd></td>
 *       <td>Move focus to last day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageUp</kbd></td>
 *       <td>Switch to previous month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageDown</kbd></td>
 *       <td>Switch to next month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageUp</kbd></td>
 *       <td>Switch to previous year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageDown</kbd></td>
 *       <td>Switch to next year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageUp</kbd></td>
 *       <td>Switch to previous by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageDown</kbd></td>
 *       <td>Switch to next by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + T</kbd></td>
 *       <td>Places focus on Today button if it exists.</tr>
 *     </tr>   
 *   </tbody>
 * </table>
 * 
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojInputDateTime
 */
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Select the currently focused day</td>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move up in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move down in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move right in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move left in the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Close the grid.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Home</kbd></td>
 *       <td>Move focus to first day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>End</kbd></td>
 *       <td>Move focus to last day of the month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageUp</kbd></td>
 *       <td>Switch to previous month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>PageDown</kbd></td>
 *       <td>Switch to next month.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageUp</kbd></td>
 *       <td>Switch to previous year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Alt + PageDown</kbd></td>
 *       <td>Switch to next year.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageUp</kbd></td>
 *       <td>Switch to previous by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + PageDown</kbd></td>
 *       <td>Switch to next by stepBigMonths.</tr>
 *     </tr>
 *     <tr>
 *       <td>Picker</td>
 *       <td><kbd>Ctrl + Alt + T</kbd></td>
 *       <td>Places focus on Today button if it exists.</tr>
 *     </tr>   
 *   </tbody>
 * </table>
 * 
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojDateTimePicker
 */

//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the InputDateTime element's input element when not inline.</p>
 *
 * @ojsubid oj-inputdatetime-time-input
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the node for the input element:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-inputdatetime-time-input'} );
 */

/**
 * <p>Sub-ID for the icon that triggers the calendar display.</p>
 *
 * @ojsubid oj-inputdatetime-calendar-clock-icon
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the icon that triggers the calendar display:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-inputdatetime-calendar-clock-icon'} );
 */

/**
 * <p>Sub-ID for the time wheel picker drop down node.
 *
 * @ojsubid oj-timepicker-content
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the time wheel picker drop down node:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-timepicker-content'} );
 */

/**
 * <p>Sub-ID for the cancel button.
 *
 * @ojsubid oj-timepicker-cancel-button
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the cancel button:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-timepicker-cancel-button'} );
 */

/**
 * <p>Sub-ID for the OK button.
 *
 * @ojsubid oj-timepicker-ok-button
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the OK button:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-timepicker-ok-button'} );
 */

/**
 * <p>Sub-ID for the hour picker.
 *
 * @ojsubid oj-timepicker-hour
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the hour picker:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-timepicker-hour'} );
 */

/**
 * <p>Sub-ID for the minute picker.
 *
 * @ojsubid oj-timepicker-minute
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the minute picker:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-timepicker-minute'} );
 */

/**
 * <p>Sub-ID for the meridian picker.
 *
 * @ojsubid oj-timepicker-meridian
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the meridian picker:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-timepicker-meridian'} );
 */

/**
 * <p>Sub-ID for the now button for button bar.
 *
 * @ojsubid oj-timepicker-now
 * @memberof oj.ojInputDateTime
 * @instance
 *
 * @example <caption>Get the now/now button for button bar:</caption>
 * var node = myInputDateTime.getNodeBySubId( {'subId': 'oj-timepicker-now'} );
 */

(function() {
var ojInputTimeMeta = {
  "properties": {
    "converter": {
      "type": "Object"
    },
    "keyboardEdit": {
      "type": "string"
    },
    "pickerAttributes": {
      "type": "Object"
    },
    "renderMode": {
      "type": "string"
    },
    "min": {
      "type": "string"
    },
    "max": {
      "type": "string"
    },
    "timePicker": {
      "type": "Object",
      "properties": {
        "footerLayout": {
          "type": "string"
        },
        "timeIncrement": {
          "type": "string"
        },
        "showOn": {
          "type": "string"
        }
      }
    },
    "value": {
      "type": "string",
      "writeback": true
    },
    "translations": {
      "type": "Object",
      "properties": {
        "ampmWheelLabel": {
          "type": "string",
          "value": "AMPM"
        },
        "cancelText": {
          "type": "string",
          "value": "Prev"
        },
        "currentTimeText": {
          "type": "string",
          "value": "Now"
        },
        "dateTimeRange": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "Object",
              "properties": {
                "inRange": {
                  "type": "string"
                },
                "max": {
                  "type": "string"
                },
                "min": {
                  "type": "string"
                }
              }
            },
            "messageDetail": {
              "type": "Object",
              "properties": {
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            },
            "messageSummary": {
              "type": "Object",
              "properties": {
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            }
          }
        },
        "hourWheelLabel": {
          "type": "string",
          "value": "Hour"
        },
        "minuteWheelLabel": {
          "type": "string",
          "value": "Minute"
        },
        "okText": {
          "type": "string",
          "value": "OK"
        },
        "regexp": {
          "type": "Object",
          "properties": {
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "tooltipTime": {
          "type": "string",
          "value": "Select Time"
        },
        "tooltipTimeDisabled": {
          "type": "string",
          "value": "Select Time Disabled"
        }
      }
    }
  },
  "methods": {
    "hide": {},
    "show": {}
  },
  "extension": {
    _WIDGET_NAME: "ojInputTime"
  }
};

oj.CustomElementBridge.registerMetadata('oj-input-time', 'inputBase', ojInputTimeMeta);
oj.CustomElementBridge.register('oj-input-time', {'metadata': oj.CustomElementBridge.getMetadata('oj-input-time')});
})();

(function() {
var ojInputDateMeta = {
  "properties": {
    "converter": {
      "type": "Object"
    },
    "datePicker": {
      "type": "Object",
      "properties": {
        "footerLayout": {
          "type": "string"
        },
        "changeMonth": {
          "type": "string"
        },
        "changeYear": {
          "type": "string"
        },
        "currentMonthPos": {
          "type": "number"
        },
        "daysOutsideMonth": {
          "type": "string"
        },
        "numberOfMonths": {
          "type": "number"
        },
        "showOn": {
          "type": "string"
        },
        "stepMonths": {
          "type": "string"
        },
        "stepBigMonths": {
          "type": "number"
        },
        "weekDisplay": {
          "type": "string"
        },
        "yearRange": {
          "type": "string"
        }
      }
    },
    "min": {
      "type": "string"
    },
    "max": {
      "type": "string"
    },
    "dayFormatter": {
      "type": "function"
    },
    "dayMetaData": {
      "type": "Object"
    },
    "keyboardEdit": {
      "type": "string"
    },
    "pickerAttributes": {
      "type": "Object"
    },
    "renderMode": {
      "type": "string"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "currentText": {
          "type": "string",
          "value": "Today"
        },
        "dateRestriction": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "dateTimeRange": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "Object",
              "properties": {
                "inRange": {
                  "type": "string"
                },
                "max": {
                  "type": "string"
                },
                "min": {
                  "type": "string"
                }
              }
            },
            "messageDetail": {
              "type": "Object",
              "properties": {
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            },
            "messageSummary": {
              "type": "Object",
              "properties": {
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            }
          }
        },
        "nextText": {
          "type": "string",
          "value": "Next"
        },
        "prevText": {
          "type": "string",
          "value": "Prev"
        },
        "regexp": {
          "type": "Object",
          "properties": {
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "tooltipCalendar": {
          "type": "string",
          "value": "Select Date"
        },
        "tooltipCalendarDisabled": {
          "type": "string",
          "value": "Select Date Disabled"
        },
        "tooltipCalendarTime": {
          "type": "string",
          "value": "Select Date Time"
        },
        "tooltipCalendarTimeDisabled": {
          "type": "string",
          "value": "Select Date Time Disabled"
        },
        "weekHeader": {
          "type": "string",
          "value": "Wk"
        }
      }
    },
    "value": {
      "type": "string",
      "writeback": true
    }
  },
  "methods": {
    "hide": {},
    "show": {}
  },
  "extension": {

    _WIDGET_NAME: "ojInputDate"
  }
};

oj.CustomElementBridge.registerMetadata('oj-input-date', 'inputBase', ojInputDateMeta);
oj.CustomElementBridge.register('oj-input-date', {'metadata': oj.CustomElementBridge.getMetadata('oj-input-date')});

var ojDatePickerMeta = oj.CollectionUtils.copyInto({}, oj.CustomElementBridge.getMetadata('oj-input-date'), undefined, true);
ojDatePickerMeta['extension']._INNER_ELEM = 'div';
oj.CustomElementBridge.register('oj-date-picker', {'metadata': ojDatePickerMeta});
})();

(function() {
var ojInputDateTimeMeta = {
  "properties": {
    "timePicker": {
      "type": "Object",
      "properties": {
        "footerLayout": {
          "type": "string"
        },
        "timeIncrement": {
          "type": "string"
        },
        "showOn": {
          "type": "string"
        }
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "cancel": {
          "type": "string",
          "value": "Cancel"
        },
        "currentText": {
          "type": "string",
          "value": "Today"
        },
        "dateRestriction": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "dateTimeRange": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "Object",
              "properties": {
                "inRange": {
                  "type": "string"
                },
                "max": {
                  "type": "string"
                },
                "min": {
                  "type": "string"
                }
              }
            },
            "messageDetail": {
              "type": "Object",
              "properties": {
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            },
            "messageSummary": {
              "type": "Object",
              "properties": {
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            }
          }
        },
        "done": {
          "type": "string",
          "value": "Done"
        },
        "nextText": {
          "type": "string",
          "value": "Next"
        },
        "prevText": {
          "type": "string",
          "value": "Prev"
        },
        "regexp": {
          "type": "Object",
          "properties": {
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "tooltipCalendar": {
          "type": "string",
          "value": "Select Date"
        },
        "tooltipCalendarDisabled": {
          "type": "string",
          "value": "Select Date Disabled"
        },
        "tooltipCalendarTime": {
          "type": "string",
          "value": "Select Date Time"
        },
        "tooltipCalendarTimeDisabled": {
          "type": "string",
          "value": "Select Date Time Disabled"
        },
        "weekHeader": {
          "type": "string",
          "value": "Wk"
        }
      }
    }
  },
  "methods": {
    "hideTimePicker": {},
    "showTimePicker": {}
  },
  "extension": {

    _WIDGET_NAME: "ojInputDateTime"
  }
};

oj.CustomElementBridge.registerMetadata('oj-input-date-time', 'oj-input-date', ojInputDateTimeMeta);
oj.CustomElementBridge.register('oj-input-date-time', {'metadata': oj.CustomElementBridge.getMetadata('oj-input-date-time')});

var ojDateTimePickerMeta = oj.CollectionUtils.copyInto({}, oj.CustomElementBridge.getMetadata('oj-input-date-time'), undefined, true);
ojDateTimePickerMeta['extension']._INNER_ELEM = 'div';
oj.CustomElementBridge.register('oj-date-time-picker', {'metadata': ojDateTimePickerMeta});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojTable
 * @augments oj.baseComponent
 * @ojstatus preview
 * @since 0.6
 * @ojtsimport ojdataprovider
 * @ojshortdesc Displays data items in a tabular format with highly interactive features.
 * @ojrole grid
 * @ojrole gridcell
 * @ojrole rowheader
 * @ojrole columnheader
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojTable<K, D> extends baseComponent<ojTableSettableProperties<K,D>>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojTableSettableProperties<K,D> extends baseComponentSettableProperties",
 *                for: "SettableProperties"
 *               }
 *              ]
 *
 * @classdesc
 * <h3 id="tableOverview-section">
 *   JET Table
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#tableOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET Table enhances a HTML table element into one that supports all
 * the features in JET Table.</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-table
 *   aria-label="Departments Table" 
 *   data='{{datasource}}'
 *   columns='[{"headerText": "Department Id", "field": "DepartmentId"},
 *             {"headerText": "Department Name", "field": "DepartmentName"},
 *             {"headerText": "Location Id", "field": "LocationId"},
 *             {"headerText": "Manager Id", "field": "ManagerId"}]' &gt;
 * &lt;/oj-table>
 * </code>
 * </pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>Developers should always either specify the <code class="prettyprint">aria-label</code> attribute or use other alternatives for the table element to conform to accessibility guidelines.</p>
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 * 
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it's recommended that applications limit the amount of data to display.  Displaying large
 * number of items in Table makes it hard for users to find what they are looking for, but affects the
 * rendering performance as well.  If displaying large number of items is neccessary, consider use a paging control with Table
 * to limit the number of items to display at a time.  Also consider setting <code class="prettyprint">scrollPolicy</code> to
 * 'loadMoreOnScroll' to enable infinite scrolling to reduce the number of elements in the DOM at any given time .</p>
 *
 * <h4>Cell Content</h4>
 * <p>Table allows developers to specify arbitrary content inside its cells. In order to minimize any negative effect on
 * performance, you should avoid putting a large number of heavy-weight components inside a cell because as you add more complexity
 * to the structure, the effect will be multiplied because there can be many items in the Table.</p>
 *
 * <h3 id="animation-section">
 *   Animation
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#animation-section"></a>
 * </h3>
 * 
 * {@ojinclude "name":"animationDoc"}
 *
 */
(function() {
  oj.__registerWidget("oj.ojTable", $['oj']['baseComponent'],
    {
      version: '1.0.0',
      defaultElement: '<table>',
      widgetEventPrefix: 'oj',
      options:
        {
          /**
           * Accessibility attributes.
           *
           * @expose
           * @name accessibility
           * @public
           * @instance
           * @memberof oj.ojTable
           * @type {Object.<string, string>|null}
           * @default null
           *
           * @example <caption>Initialize the Table, overriding accessibility value:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-table accessibility.row-header='headerColumnId'>&lt;/oj-table>
           * 
           * &lt;!-- Using JSON notation -->
           * &lt;oj-table accessibility='{"rowHeader":"headerColumnId"}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">accessibility</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.accessibility.rowHeader;
           *
           * // Set one. Always use the setProperty API for 
           * // subproperties rather than setting a subproperty directly.
           * myTable.setProperty('accessibility.rowHeader', 'headerColumnId');
           *
           * // Get all
           * var values = myTable.accessibility;
           *
           * // Set all.  Must list every accessibility key, as those not listed are lost.
           * myTable.accessibility = {
           *     rowHeader: 'headerColumnId'
           * };
           */
          accessibility: null,
          /**
           * The column id to be used as the row header by screen readers.
           *
           * <p>The td cells in the column specified by this
           * attribute will be assigned an id and then referenced by the
           * headers attribute in the rest of the cells in the row.
           * This is required by screen readers. By default the first column
           * will be taken as the row header.</p>
           * <p>See the <a href="#accessibility">accessibility</a> attribute for usage examples.</p>
           *
           * @expose
           * @name accessibility.rowHeader
           * @memberof! oj.ojTable
           * @instance
           * @public
           * @type {string}
           */
          /**
           * Triggered when the default animation of a particular action is about to start.  The default animation can be cancelled by calling <code class="prettyprint">event.preventDefault</code>.
           * If the event listener calls <code class="prettyprint">event.preventDefault</code> to cancel the default animation, it must call the <code class="prettyprint">event.detail.endCallback</code> function when it finishes its own animation handling.
           * Row animations will only be triggered for rows in the current viewport and an event will be triggered for each cell in the animated row.
           * 
           * @expose
           * @event
           * @memberof oj.ojTable
           * @instance
           * @ojbubbles
           * @ojcancelable
           * @property {"add"|"remove"|"update"} action the action that starts the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
           * @property {Element} element the target of animation. For row animations this will be the cell contents wrapped in a div.
           * @property {function():void} endCallback if the event listener calls event.preventDefault to cancel the default animation, it must call the endCallback function when it finishes its own animation handling and when any custom animation ends.
           */
          animateStart: null,
          /**
           * Triggered when the default animation of a particular action has ended.  Note this event will not be triggered if application cancelled the default animation on animateStart.
           * Row animations will only be triggered for rows in the current viewport and an event will be triggered for each cell in the animated row.
           * 
           * @expose
           * @event
           * @memberof oj.ojTable
           * @instance
           * @ojbubbles
           * @ojcancelable
           * @property {"add"|"remove"|"update"} action the action that started the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
           * @property {Element} element the target of animation. For row animations this will be the cell contents wrapped in a div.
           */
          animateEnd: null,

          /**
           * Table's current row can be on index and/or key value, when both are
           * specified, the index is used as a hint.
           * @typedef {Object} oj.ojTable.CurrentRow<K>
           * @ojsignature {target:"Type", value:"{rowIndex: number, rowKey?: K}|{rowIndex?: number, rowKey: K}"}
           */

          /**
           * An alias for the current context when referenced inside the cell template. This can be especially useful
           * if oj-bind-for-each element is used inside the cell template since it has its own scope of data access.
           *
           * @ojshortdesc Gets and sets the alias for the current context when referenced inside the cell template.
           * @ojstatus preview
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {string}
           * @default ''
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">as</code> attribute specified:</caption>
           * &lt;oj-table as='cell' columns='[{"headerText": "Department Id",
           *                                   "field": "DepartmentId",
           *                                   "template": "cellTemplate"}]'>
           *   &lt;template slot='cellTemplate'>
           *     &lt;p>&lt;oj-bind-text value='[[cell.data.name]]'>&lt;/oj-bind-text>&lt;/p>
           *   &lt;/template>
           * &lt;/oj-table>
           */ 
           as: '',
          /**
           * The row that currently have keyboard focus.  Can be an index and/or key value.
           * When both are specified, the index is used as a hint.
           * Returns the current row or null if there is none.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {Object}
           * @ojsignature {target: "Type", value: "oj.ojTable.CurrentRow<K> | null"}
           * @default null
           * @ojwriteback
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">current-row</code> attribute specified:</caption>
           * &lt;oj-table current-row='{"rowIndex": 1}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">currentRow</code> property after initialization:</caption>
           * // getter
           * var value = myTable.currentRow;
           *
           * // setter
           * myTable.currentRow = {rowKey: '123'};
           */
          currentRow: null,
          /**
           * The data to bind to the element.
           * <p>
           * Must be of type oj.DataProvider {@link oj.DataProvider}
           * or type oj.TableDataSource {@link oj.TableDataSource}
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {oj.DataProvider|oj.TableDataSource|null}
           * @ojsignature {target: "Accessor", value: {GetterType: "oj.DataProvider<K, D>", SetterType: "oj.DataProvider<K, D>|null"}}
           * @default null
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">data</code> attribute specified:</caption>
           * &lt;oj-table data='{{dataProvider}}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">data</code> property after initialization:</caption>
           * // getter
           * var dataProvider = myTable.data;
           *
           * // setter
           * myTable.data = dataProvider;
           */
          data: null,
          /**
           * Whether to display table in list or grid mode. Setting a value of grid
           * will cause the table to display in grid mode. The default value of this
           * attribute is set through the theme.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {string}
           * @ojvalue {string} "list" Display table in list mode.
           * @ojvalue {string} "grid" Display table in grid mode. This is a more compact look than list mode.
           * @default "list"
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">display</code> attribute specified:</caption>
           * &lt;oj-table display='grid'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">display</code> property after initialization:</caption>
           * // getter
           * var displayValue = myTable.display;
           *
           * // setter
           * myTable.display = 'grid';
           */
           display: 'list',
          /**
           * Enable drag and drop functionality.<br><br>
           * JET provides support for HTML5 Drag and Drop events.  Please refer to {@link https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/Drag_and_drop third party documentation} 
           * on HTML5 Drag and Drop to learn how to use it.
           * 
           * @type {Object}
           * @default {'drag': null, 'drop': null, 'reorder': {'columns' :'disabled'}}
           * @expose
           * @instance
           * @memberof! oj.ojTable
           * 
           * @example <caption>Initialize the table with some dnd functionality:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-table dnd.reorder.columns='enabled' dnd.drag.rows.dataTypes='application/ojtablerows+json'>&lt;/oj-table>
           * 
           * &lt;!-- Using JSON notation -->
           * &lt;oj-table dnd='{"reorder":{"columns":"enabled"}, "drag":{"rows":{"dataTypes":"application/ojtablerows+json"}}}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">dnd</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.dnd.reorder;
           *
           * // Set one, leaving the others intact. Use the setProperty API for 
           * // subproperties so that a property change event is fired.
           * myTable.setProperty('dnd.reorder', {"columns":"enabled"});
           *
           * // Get all
           * var values = myTable.dnd;
           *
           * // Set all.  Must list every dnd functionality, as those not listed are lost.
           * myTable.dnd = {
           *     reorder: {"columns":"enabled"},
           *     drag: {"rows":{"dataTypes":"application/ojtablerows+json"}}
           * };
           */
          dnd : {
            /**
             * An object that describes drag functionlity.
             *
             * <p>See the <a href="#dnd">dnd</a> attribute for usage examples.
             *
             * @expose
             * @name dnd.drag
             * @memberof! oj.ojTable
             * @instance
             * @type {Object}
             * @default null
             */
            'drag': null,
            /**
             * @typedef {object} oj.ojTable.DragRowContext Context for table DnD on Rows
             * @property {Array<Object>} rows An array of objects, with each object representing the data of one selected row in the structure below
             * @property {D} rows.data The raw row data
             * @property {number} rows.index The index for the row
             * @property {K} rows.key The key value for the row
             * @ojsignature {target:"Type", value:"<K,D>", for:"genericTypeParameters"}
             */
            /**
             * If this object is specified, the table will initiate drag operation when the user drags on selected rows.
             *
             * <p>See the <a href="#dnd">dnd</a> attribute for usage examples.
             *
             * @property {string | Array.<string>} [dataTypes]  (optional) The MIME types to use for the dragged data in the dataTransfer object.  This can be a string if there is only one
             * type, or an array of strings if multiple types are needed.<br><br>
             * For example, if selected rows of employee data are being dragged, dataTypes could be "application/employees+json". Drop targets can examine the data types and decide
             * whether to accept the data. A text input may only accept "text" data type, while a chart for displaying employee data may be configured to accept the "application/employees+json" type.<br><br>
             * For each type in the array, dataTransfer.setData will be called with the specified type and the JSON version of the selected rows data as the value. The selected rows data 
             * is an array of objects, with each object representing one selected row in the format returned by TableDataSource.get().<br><br>
             * This property is required unless the application calls setData itself in a dragStart callback function.
             * @property {function(DragEvent, oj.ojTable.DragRowContext<K,D>):void} [dragStart]  (optional) A callback function that receives the "dragstart" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code><br><br>
             * Parameters:<br><br>
             * <code class="prettyprint">event</code>: The DOM event object<br><br>
             * <code class="prettyprint">context</code>: {@link oj.ojTable.DragRowContext} object with the following properties:<br>
             * <ul>
             *   <li><code class="prettyprint">rows</code>: An array of objects, with each object representing the data of one selected row in the structure below:
             *     <table>
             *     <tbody>
             *     <tr><td><b>data</b></td><td>The raw row data</td></tr>
             *     <tr><td><b>index</b></td><td>The index for the row</td></tr>
             *     <tr><td><b>key</b></td><td>The key value for the row</td></tr>
             *     </tbody>
             *     </table>
             *   </li>
             * </ul><br><br>
             * This function can set its own data and drag image as needed. If dataTypes is specified, event.dataTransfer is already populated with the default data when this function is invoked.
             * If dataTypes is not specified, this function must call event.dataTransfer.setData to set the data or else the drag operation will be cancelled.  In either case, the drag image is  
             * set to an image of the selected rows visible on the table.
             * @property {function(DragEvent):void} [drag]  (optional) A callback function that receives the "drag" event as argument:<br><br>
             * <code class="prettyprint">function(event)</code><br><br>
             * Parameters:<br><br>
             * <code class="prettyprint">event</code>: The DOM event object
             * @property {function(DragEvent):void} [dragEnd]  (optional) A callback function that receives the "dragend" event as argument:<br><br>
             * <code class="prettyprint">function(event)</code><br><br>
             * Parameters:<br><br>
             * <code class="prettyprint">event</code>: The DOM event object<br>
             *
             * @expose
             * @name dnd.drag.rows
             * @memberof! oj.ojTable
             * @instance
             * @type {Object}
             */
            /**
             * An object that describes drop functionlity.
             *
             * <p>See the <a href="#dnd">dnd</a> attribute for usage examples.
             *
             * @expose
             * @name dnd.drop
             * @memberof! oj.ojTable
             * @instance
             * @type {Object}
             * @default null
             */
            'drop': null,
            /**
             * @typedef {object} oj.ojTable.DropColumnContext
             * @property {number} columnIndex The index for the column
             */
            /**
             * An object that specifies callback functions to handle dropping columns<br><br>
             * For all callback functions, the following arguments will be passed:<br><br>
             * <code class="prettyprint">event</code>: The DOM event object<br><br>
             * <code class="prettyprint">context</code>: Context object with the following properties:
             * <ul>
             *   <li><code class="prettyprint">columnIndex</code>: The index of the column being dropped on</li>
             * </ul>
             *
             * <p>See the <a href="#dnd">dnd</a> attribute for usage examples.
             *
             * @property {string | Array.<string>} dataTypes  A data type or an array of data types this element can accept.<br><br>
             * This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events.
             * @property {function(DragEvent, oj.ojTable.DropColumnContext):void} [dragEnter]  (optional) A callback function that receives the "dragenter" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code><br><br>
             * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
             * Calling <code class="prettyprint">event.preventDefault()</code> is required by HTML5 Drag and Drop to indicate acceptance of data.<br><br>
             * If dataTypes is specified, it will be matched against the drag data types to determine if the data is acceptable.  If there is a match, JET will call 
             * <code class="prettyprint">event.preventDefault()</code> to indicate that the data can be accepted.
             * @property {function(DragEvent, oj.ojTable.DropColumnContext):void} [dragOver]  (optional) A callback function that receives the "dragover" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code><br><br>
             * Similar to dragEnter, this function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.  If dataTypes is specified,
             * it will be matched against the drag data types to determine if the data is acceptable.       
             * @property {function(DragEvent, oj.ojTable.DropColumnContext):void} [dragLeave]  (optional) A callback function that receives the "dragleave" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code>
             * @property {function(DragEvent, oj.ojTable.DropColumnContext):void} drop  (required) A callback function that receives the "drop" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code><br><br>
             * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data is accepted.           
             *
             * @expose
             * @name dnd.drop.columns
             * @memberof! oj.ojTable
             * @instance
             * @type {Object}
             */
            /**
             * @typedef {object} oj.ojTable.DropRowContext
             * @property {number} rowIndex The index for the row
             */
            /**
             * An object that specifies callback functions to handle dropping rows<br><br>
             * For all callback functions, the following arguments will be passed:<br><br>
             * <code class="prettyprint">event</code>: The DOM event object<br><br>
             * <code class="prettyprint">context</code>: Context object with the following properties:
             * <ul>
             *   <li><code class="prettyprint">rowIndex</code>: The index of the row being dropped on</li>
             * </ul>
             *
             * <p>See the <a href="#dnd">dnd</a> attribute for usage examples.
             *
             * @property {string | Array.<string>} dataTypes  A data type or an array of data types this element can accept.<br><br>
             * This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events.
             * @property {function(DragEvent, oj.ojTable.DropRowContext):void} [dragEnter]  (optional) A callback function that receives the "dragenter" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code><br><br>
             * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
             * Calling <code class="prettyprint">event.preventDefault()</code> is required by HTML5 Drag and Drop to indicate acceptance of data.<br><br>
             * If dataTypes is specified, it will be matched against the drag data types to determine if the data is acceptable.  If there is a match, JET will call
             * <code class="prettyprint">event.preventDefault()</code> to indicate that the data can be accepted.
             * @property {function(DragEvent, oj.ojTable.DropRowContext):void} [dragOver]  (optional) A callback function that receives the "dragover" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code><br><br>
             * Similar to dragEnter, this function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.  If dataTypes is specified,
             * it will be matched against the drag data types to determine if the data is acceptable.       
             * @property {function(DragEvent, oj.ojTable.DropRowContext):void} [dragLeave]  (optional) A callback function that receives the "dragleave" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code>
             * @property {function(DragEvent, oj.ojTable.DropRowContext):void} drop  (required) A callback function that receives the "drop" event and context information as arguments:<br><br>
             * <code class="prettyprint">function(event, context)</code><br><br>
             * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data is accepted.<br><br>
             * If the application needs to look at the data for the row being dropped on, it can use the getDataForVisibleRow method.
             *
             * @expose
             * @name dnd.drop.rows
             * @memberof! oj.ojTable
             * @instance
             * @type {Object}
             */
            /**
             * An object that describes reorder functionlity.
             *
             * <p>See the <a href="#dnd">dnd</a> attribute for usage examples.
             *
             * @expose
             * @name dnd.reorder
             * @memberof! oj.ojTable
             * @instance
             * @type {Object}
             * @default {'columns' :'disabled'}
             */
            'reorder': {
              /**
               * Enable or disable reordering the columns within the same table using drag and drop.<br><br>
               * Re-ordering is supported one column at a time. In addition, re-ordering will not re-order
               * any cells which have the colspan attribute with value > 1. Such cells will need to be re-ordered manually by listening to
               * the property change event on the columns property.
               * 
               * <p>See the <a href="#dnd">dnd</a> attribute for usage examples.
               *
               * @expose
               * @name dnd.reorder.columns
               * @memberof! oj.ojTable
               * @instance
               * @type {string}
               * @ojvalue {string} 'enabled' Enable column reordering
               * @ojvalue {string} 'disabled' Disable column reordering
               * @default "disabled"
               * @ojwriteback
               */
              'columns' :'disabled'
            }
          },
          /**
           * Determine if the table is read only or editable. Use 'none' if the table is strictly read only and will be a single Tab stop on the page.
           * Use 'rowEdit' if you want single row at a time editability. The table will initially render with all rows in read only mode. Pressing Enter/F2 or double click will make the current row editable and pressing Tab navigates to the next cell. Pressing ESC/F2 while in this mode will switch the table back to all rows in read only mode and will be a single Tab stop the page.
           *
           * @expose
           * @memberof! oj.ojTable
           * @instance
           * @type {string}
           * @ojvalue {string} "none" The table is read only and is a single Tab stop.
           * @ojvalue {string} "rowEdit" The table has single row at a time editability and the cells within the editable row are tabbable.
           * @default "none"
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">edit-mode</code> attribute specified:</caption>
           * &lt;oj-table edit-mode='rowEdit'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">editMode</code> property after initialization:</caption>
           * // getter
           * var value = myTable.editMode;
           *
           * // setter
           * myTable.editMode = 'rowEdit';
           */
          editMode: 'none',
          /**
           * The text to display when there are no data in the Table. If it is not defined,
           * then a default empty text is extracted from the resource bundle.
           * 
           * @expose
           * @memberof! oj.ojTable
           * @instance
           * @type {string|null}
           * @default null
           * @example <caption>Initialize the table with the <code class="prettyprint">empty-text</code> attribute specified:</caption>
           * &lt;oj-table 
           *   aria-label="Departments Table" 
           *   data='{{datasource}}' 
           *   empty-text='No data'
           *   columns='[{"headerText": "Department Id", "field": "DepartmentId"},
           *             {"headerText": "Department Name", "field": "DepartmentName"}]'&gt;
           * &lt;/oj-table>
           *
           * @ignore
           */
          emptyText: null,
          /**
           * Whether the horizontal gridlines are to be drawn. Can be enabled or disabled.
           * The default value of auto means it's determined by the display attribute.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {string}
           * @ojvalue {string} "auto" Determined by display attribute.
           * @ojvalue {string} "enabled" Enabled.
           * @ojvalue {string} "disabled" Disabled.
           * @default "auto"
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">horizontal-grid-visible</code> attribute specified:</caption>
           * &lt;oj-table horizontal-grid-visible='disabled'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">horizontalGridVisible</code> property after initialization:</caption>
           * // getter
           * var value = myTable.horizontalGridVisible;
           *
           * // setter
           * myTable.horizontalGridVisible = 'disabled';
           */
          horizontalGridVisible: 'auto',
          /**
           * The row renderer function to use.
           * <p>
           * The renderer function will be passed in an Object which contains the fields:
           * <ul>
           *   <li>componentElement: A reference to the Table root element</li>
           *   <li>data: Key/value pairs of the row</li>
           *   <li>parentElement: Empty rendered TR element</li>
           *   <li>rowContext.datasource: The "data" attribute of the Table</li>
           *   <li>rowContext.mode: The mode of the row.  It can be "edit" or "navigation".</li>
           *   <li>rowContext.status: Contains the rowIndex, rowKey, and currentRow</li>
           * </ul>
           * The function returns  either a String or
           * a DOM element of the content inside the row. If the developer chooses
           * to manipulate the row element directly, the function should return
           * nothing.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {Function|null}
           * @ojsignature {target: "Type", value: "((context: oj.ojTable.RowRendererContext<K,D>) => string | HTMLElement | void) | null"}
           * @default null
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">row-renderer</code> attribute specified:</caption>
           * &lt;oj-table row-renderer='{{myRowRenderer}}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">rowRenderer</code> property after initialization:</caption>
           * // getter
           * var value = myTable.rowRenderer;
           *
           * // setter
           * myTable.rowRenderer = myRowRenderer;
           */
          rowRenderer: null,
          
          /**
           * @typedef {Object} oj.ojTable.RowRendererContext context object used by rowRenderer callback.
           * @property {Element} componentElement A reference to the Table root element
           * @property {*} data Key/value pairs of the row
           * @property {Element} parentElement Empty rendered TR element
           * @property {Object} rowContext contex of the row
           * @property {oj.DataProvider<K, D>|null} rowContext.datasource The "data" attribute of the Table
           * @property {"edit"|"navigation"} rowContext.mode The mode of the row.  It can be "edit" or "navigation".
           * @property {oj.ojTable.ContextStatus<K>} rowContext.status Contains the rowIndex, rowKey, and currentRow
           * @ojsignature [{target:"Type", value:"{[name: string]: any}", for:"data"},
           *               {target:"Type", value:"<K,D>", for:"genericTypeParameters"}]
           */

          /**
           * @typedef {Object} oj.ojTable.ContextStatus<K>
           * @property {number} rowIndex index of the row
           * @property {K} rowKey key of the row
           * @property {oj.ojTable.CurrentRow<K>} currentRow current row.
           */

          /**
           * Specifies the mechanism used to scroll the data inside the table. Possible values are: auto and loadMoreOnScroll.
           * When loadMoreOnScroll is specified, additional data are fetched when the user scrolls to the bottom of the table.
           *
           * @expose
           * @memberof! oj.ojTable
           * @instance
           * @type {string}
           * @ojvalue {string} "auto" Determined by element. The default is to display all data.
           * @ojvalue {string} "loadMoreOnScroll" Additional data are fetched when the user scrolls to the bottom of the table.
           * @default "auto"
           *
           * @example <caption>Initialize the Table with the <code class="prettyprint">scroll-policy</code> attribute specified:</caption>
           * &lt;oj-table scroll-policy='loadMoreOnScroll'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">scrollPolicy</code> property after initialization:</caption>
           * // getter
           * var value = myTable.scrollPolicy;
           *
           * // setter
           * myTable.scrollPolicy = 'loadMoreOnScroll';
           */
          scrollPolicy: "auto",
          /**
           * scrollPolicy options.
           * <p>
           * When scrollPolicy is loadMoreOnScroll, the next block of rows is fetched
           * when the user scrolls to the end of the table. The fetchSize property
           * determines how many rows are fetched in each block.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {Object.<string, string>|null}
           * @default {'fetchSize': 25, 'maxCount': 500}
           *
           * @example <caption>Initialize the component, overriding some scroll-policy-options values and leaving the others intact:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-table scroll-policy-options.some-key='some value' scroll-policy-options.some-other-key='some other value'>&lt;/oj-table>
           * 
           * &lt;!-- Using JSON notation -->
           * &lt;oj-table scroll-policy-options='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">scrollPolicyOptions</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.scrollPolicyOptions.someKey;
           *
           * // Set one, leaving the others intact. Always use the setProperty API for 
           * // subproperties rather than setting a subproperty directly.
           * myTable.setProperty('scrollPolicyOptions.someKey', 'some value');
           *
           * // Get all
           * var values = myTable.scrollPolicyOptions;
           *
           * // Set all.  Must list every scrollPolicyOptions key, as those not listed are lost.
           * myTable.scrollPolicyOptions = {
           *     someKey: 'some value',
           *     someOtherKey: 'some other value'
           * };
           */
          scrollPolicyOptions: {
            /**
             * The number of rows to fetch in each block of rows.
             * <p>See the <a href="#scrollPolicyOptions">scroll-policy-options</a> attribute for usage examples.</p>
             *
             * @expose
             * @name scrollPolicyOptions.fetchSize
             * @memberof! oj.ojTable
             * @instance
             * @type {number}
             * @default 25
             * @ojmin 1
             */
            'fetchSize': 25, 
            /**
             * The maximum number of rows which will be displayed before fetching more rows will be stopped.
             * <p>See the <a href="#scrollPolicyOptions">scroll-policy-options</a> attribute for usage examples.</p>
             *
             * @expose
             * @name scrollPolicyOptions.maxCount
             * @memberof! oj.ojTable
             * @instance
             * @type {number}
             * @default 500
             * @ojmin 0
             */
            'maxCount': 500
          },

          /**
           * @typedef {Object} oj.ojTable.RowSelectionStart start of one row selection, can be on index or key or both.
           * @ojsignature [{target:"Type", value:"{startIndex: {row: number}, startKey?:{row: K}}|{startIndex?: {row: number}, startKey:{row: K}}"},
           *               {target:"Type", value:"<K>", for:"genericTypeParameters"}]
           */

          /**
           * @typedef {Object} oj.ojTable.RowSelectionEnd end of one row selection, can be on index or key or both.
           * @ojsignature [{target:"Type", value:"{endIndex: {row: number}, endKey?: {row: K}}|{endIndex?:{row: number}, endKey:{row: K}}"},
           *               {target:"Type", value:"<K>", for:"genericTypeParameters"}]
           */

          /**
           * @typedef {Object} oj.ojTable.ColumnSelectionStart start of one column selection, can be on index or key or both.
           * @ojsignature [{target:"Type", value:"{startIndex:{column:number}, startKey?:{column:K}}|{startIndex?:{column:number}, startKey:{column:K}}"}, 
           *               {target:"Type", value:"<K>", for:"genericTypeParameters"}]
           */

          /**
           * @typedef {Object} oj.ojTable.ColumnSelectionEnd end of one column selection, can be on index or key or both.
           * @ojsignature [{target:"Type", value:"{endIndex:{column:number},endKey?:{column:K}}|{endIndex?:{column:number}, endKey:{column:K}}"},
           *               {target:"Type", value:"<K>", for:"genericTypeParameters"}]
           */

          /**
           * Gets the key and data of the first selected row.  The first selected row is defined as the first
           * key returned by the <a href="#selection">selection</a> property.  The value of this property contains:
           * <ul>
           * <li>key - the key of the first selected row.</li>
           * <li>data - the data of the first selected row.  If the selected row is not locally available, this will
           *        be null.</li>
           * </ul>
           * If no rows are selected then this property will return an object with both key and data properties set to null.
           *
           * @expose
           * @memberof! oj.ojTable
           * @instance
           * @default {'key': null, 'data': null}
           * @type {Object}
           *
           * @ojwriteback
           * @readonly
           *
           * @example <caption>Get the data of the first selected row:</caption>
           * // getter
           * var firstSelectedRowValue = myTable.firstSelectedRow;
           */
          firstSelectedRow: {'key': null, 'data': null},
          /**
           * Specifies the current selections in the table. Can be either an index or key value.
           * When both are specified, the index is used as a hint.
           * Returns an array of range objects, or an empty array if there's no selection.
           *
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {Array.<Object>}
           * @ojsignature {target: "Type", value: "Array<oj.ojTable.RowSelectionStart<K> & oj.ojTable.RowSelectionEnd<K>> | Array<oj.ojTable.ColumnSelectionStart<K> & oj.ojTable.ColumnSelectionEnd<K>>"}
           * @default []
           * @ojwriteback
           *
           * @example <caption>Initialize the table with the <code class="prettyprint">selection</code> attribute specified:</caption>
           * &lt;oj-table selection='[{"startIndex": {"row":1}, "endIndex":{"row":3}}]'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">selection</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.selection[0];
           *
           * // Get all
           * var values = myTable.selection;
           *
           * // Set all (There is no permissible "set one" syntax.)
           * myTable.selection = [{startIndex: {"row":1}, endIndex:{"row":3}},
           *                      {startIndex: {"row":7}, endIndex:{"row":9}}];
           *
           * @example <caption>Set a row selection using index:</caption>
           * myTable.selection = [{startIndex: {"row":1}, endIndex:{"row":3}}];
           *
           * @example <caption>Set a column selection using index:</caption>
           * myTable.selection = [{startIndex: {"column":2}, endIndex: {"column":4}}];
           *
           * @example <caption>Set a row selection using key value:</caption>
           * myTable.selection = [{startKey: {"row":"10"}, endKey:{"row":"30"}}];
           *
           * @example <caption>Set a column selection using key value:</caption>
           * myTable.selection = [{startKey: {"column": "column1"}, endKey: {"column": "column2"}}];
           */
          selection: [],
          /**
           * The row and column selection modes. Both can be either single or multiple.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {Object.<string, string>|null}
           * @default null
           *
           * @example <caption>Initialize the Table, setting selection modes:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-table selection-mode.row='single' selection-mode.column='multiple'>&lt;/oj-table>
           * 
           * &lt;!-- Using JSON notation -->
           * &lt;oj-table selection-mode='{"row":"single", "column":"multiple"}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.selectionMode.row;
           *
           * // Set one. Use the setProperty API for 
           * // subproperties so that a property change event is fired.
           * myTable.setProperty('selectionMode.row', 'single');
           *
           * // Get all
           * var values = myTable.selectionMode;
           *
           * // Set all
           * myTable.selectionMode = {
           *     row: 'single',
           *     column: 'multiple'
           * };
           */
          selectionMode: null,
          /**
           * Whether selection is required. If true, then at least one row will always
           * remain selected.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {boolean}
           * @default false
           *
           * @example <caption>Initialize the Table, setting selection required:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-table selection-required='true'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">selectionRequired</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.selectionRequired;
           *
           * // Set one. Use the setProperty API for 
           * // subproperties so that a property change event is fired.
           * myTable.setProperty('selectionRequired', 'true');
           */
          selectionRequired: false,
          /**
           * The selection mode for rows.
           *
           * <p>See the <a href="#selectionMode">selection-mode</a> attribute for usage examples.
           * By default, this element does not allow any selection.
           *
           * @expose
           * @name selectionMode.row
           * @memberof! oj.ojTable
           * @instance
           * @type {string}
           * @ojvalue {string} 'single' Allow single selection
           * @ojvalue {string} 'multiple' Allow multiple selections
           */
          /**
           * The selection mode for columns.
           *
           * <p>See the <a href="#selectionMode">selection-mode</a> attribute for usage examples.
           * By default, this element does not allow any selection.
           *
           * @expose
           * @name selectionMode.column
           * @memberof! oj.ojTable
           * @instance
           * @type {string}
           * @ojvalue {string} 'single' Allow single selection
           * @ojvalue {string} 'multiple' Allow multiple selections
           */
          /**
           * Whether the vertical gridlines are to be drawn. Can be enabled or disabled.
           * The default value of auto means it's determined by the display attribute.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {string}
           * @ojvalue {string} "auto" Determined by display attribute.
           * @ojvalue {string} "enabled" Enabled.
           * @ojvalue {string} "disabled" Disabled.
           * @default "auto"
           */
          verticalGridVisible: 'auto',
          /**
           * An array of column definitions.
           * <p>If the application change the column definitions after the Table is loaded, it must call the
           * <a href="#refresh"><code class="prettyprint">refresh()</code></a> method to update the Table display.
           * 
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {Array.<Object>|null}
           * @default null
           * @example <caption>Initialize the table with the <code class="prettyprint">columns</code> attribute specified:</caption>
           * &lt;oj-table 
           *   columns='[{"headerText": "Department Id", "field": "DepartmentId"},
           *             {"headerText": "Department Name", "field": "DepartmentName"}]'&gt;
           * &lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">columns</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.columns[0];
           *
           * // Get all
           * var values = myTable.columns;
           *
           * // Set all (There is no permissible "set one" syntax.)
           * myTable.columns = [{"headerText": "Department Id", "field": "DepartmentId"},
           *                    {"headerText": "Department Name", "field": "DepartmentName"}];
           */
          columns: [{
              /**
               * @typedef {Object} oj.ojTable.ColumnsRendererContext context for columns.renderer callback function.
               * @property {Object} cellContext context of the cell containing properties
               * @property {oj.DataProvider<K, D>|null} cellContext.datasource The "data" attribute of the Table
               * @property {"edit"|"navigation"} cellContext.mode The mode of the row.  It can be "edit" or "navigation".
               * @property {oj.ojTable.ContextStatus<K>} cellContext.status Contains the rowIndex, rowKey, and currentRow
               * @property {number} columnIndex The column index
               * @property {Element} componentElement A reference to the Table root element
               * @property {D} data The cell data. 
               * @property {Element} parentElement Empty rendered <td> element
               * @property {Object} row Key/value pairs of the row
               * @ojsignature [{target:"Type", value:"{[name: string]: any}", for:"row"},
               *               {target:"Type", value:"<K,D>", for:"genericTypeParameters"}]
               */

              /**
               * The renderer function that renders the content of the cell.
               * The function will be passed a context object which contains
               * the following objects:
               * <ul>
               *   <li>cellContext.datasource: The "data" attribute of the Table</li>
               *   <li>cellContext.mode: The mode of the row.  It can be "edit" or "navigation".</li>
               *   <li>cellContext.status: Contains the rowIndex, rowKey, and currentRow</li>
               *   <li>columnIndex: The column index</li>
               *   <li>componentElement: A reference to the Table root element</li>
               *   <li>data: The cell data</li>
               *   <li>parentElement: Empty rendered <td> element</li>
               *   <li>row: Key/value pairs of the row</li>
               * </ul>
               * The function should return one of the following: 
               * <ul>
               *   <li>An Object with the following property:
               *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the header.</li></ul>
               *   </li>
               *   <li>undefined: If the developer chooses to manipulate the cell element directly, the function should return undefined.</li>
               * </ul>
               * If no renderer is specified, the Table will treat the cell data as a String.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].renderer
               * @type {Function|null}
               * @ojsignature {target: "Type", value: "?((context: oj.ojTable.ColumnsRendererContext<K,D>) => {insert: HTMLElement | string} | void) | null"}
               * @default null
               */
              renderer: null,
              /**
               * The CSS class to apply to the column cells
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].className
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              className: null,
              /**
               * The data field this column refers to.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].field
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              field: null,
              /**
               * The CSS class to apply to the footer cell.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].footerClassName
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              footerClassName: null,
              /**
               * @typedef {Object} oj.ojTable.FooterRendererContext
               * @property {number} columnIndex The column index
               * @property {Element} componentElement A reference to the Table root element
               * @property {Object} footerContext Context of the footer
               * @property {oj.DataProvider<K, D>|null} footerContext.datasource The "data" attribute of the Table
               * @property {Element} parentElement Empty rendered <td> element
               * @ojsignature {target:"Type", value:"<K,D>", for:"genericTypeParameters"}
               */
              /**
               * The renderer function that renders the content of the footer.
               * The function will be passed a context object which contains
               * the following objects:
               * <ul>
               *   <li>columnIndex: The column index</li>
               *   <li>componentElement: A reference to the Table root element</li>
               *   <li>footerContext.datasource: The "data" attribute of the Table</li>
               *   <li>parentElement: Empty rendered <td> element</li>
               * </ul>
               * The function should return one of the following: 
               * <ul>
               *   <li>An Object with the following property:
               *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the footer.</li></ul>
               *   </li>
               *   <li>undefined: If the developer chooses to manipulate the footer element directly, the function should return undefined.</li>
               * </ul>
               * If no renderer is specified, the Table will treat the footer data as a String.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].footerRenderer
               * @type {Function|null}
               * @ojsignature {target: "Type", value: "?((context: oj.ojTable.FooterRendererContext<K,D>) => {insert: HTMLElement | string} | void) | null"}
               * @default null
               */
              footerRenderer: null,
              /**
               * The CSS styling to apply to the footer cell.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].footerStyle
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              footerStyle: null,
              /**
               * The name of the slot used to specify the template for rendering the footer cell. The slot must be a &lt;template> element.</p>  
               *
               * <p>When the template is executed each footer, it will have access to the binding context containing the following properties:</p>
               * <ul>
               * <li>$current - an object that contains information for the current footer being rendered
               *   <ul>
               *     <li>componentElement - the &lt;oj-table> custom element</li>
               *   </ul>
               * </li>
               * <li>alias - if as attribute was specified, the value will be used to provide an application-named alias for $current.
               * </li>
               * </ul>
               *
               * <p>The content of the template should not include the &lt;td> element, only what's inside it.</p>
               * <p>When both footerTemplate and footerRenderer are specified, the footerRenderer takes precedence.</p>
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].footerTemplate
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              footerTemplate: null,
              /**
               * The CSS class to apply to the column header text.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].headerClassName
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              headerClassName: null,
              /**
               * @typedef {Object} oj.ojTable.HeaderRendererContext
               * @property {number} columnIndex The column index
               * @property {function(Object, function(object):void):void} columnHeaderDefaultRenderer
               *           If the column is not sortable then this function will be included in the context.
               *           The options parameter specifies the options (future use) for the renderer while the
               *           delegateRenderer parameter specifies the function which the developer would
               *           like to be called during rendering of the column header.
               * @property {function(Object, function(Object):void):void} columnHeaderSortableIconRenderer
               *           If the column is sortable then this function will be included in the context.
               *           The options parameter specifies the options (future use) for the renderer while the
               *           delegateRenderer parameter specifies the function which the developer would
               *           like to be called during rendering of the sortable column header. Calling the
               *           columnHeaderSortableIconRenderer function enables rendering custom header content
               *           while also preserving the sort icons.
               * @property {Element} componentElement A reference to the Table root element
               * @property {string} data The header text for the column
               * @property {Object} headerContext context for the header
               * @property {oj.DataProvider<K, D>|null} headerContext.datasource
               *           The "data" attribute of the Table
               * @property {Element} parentElement Empty rendered TH element
               * @ojsignature {target:"Type", value:"<K,D>", for:"genericTypeParameters"}
               */   
              /**
               * The renderer function that renders the content of the header.
               * The function will be passed a context object which contains
               * the following objects:
               * <ul>
               *   <li>columnIndex: The column index</li>
               *   <li>columnHeaderDefaultRenderer(options, delegateRenderer): If the column
               *   is not sortable then this function will be included in the context.
               *   The options parameter specifies the options (future use) for the renderer while the
               *   delegateRenderer parameter specifies the function which the developer would
               *   like to be called during rendering of the column header.</li>
               *   <li>columnHeaderSortableIconRenderer(options, delegateRenderer): If the column
               *   is sortable then this function will be included in the context.
               *   The options parameter specifies the options (future use) for the renderer while the
               *   delegateRenderer parameter specifies the function which the developer would
               *   like to be called during rendering of the sortable column header. Calling the
               *   columnHeaderSortableIconRenderer function enables rendering custom header content
               *   while also preserving the sort icons.</li>
               *   <li>componentElement: A reference to the Table root element</li>
               *   <li>data: The header text for the column</li>
               *   <li>headerContext.datasource: The "data" attribute of the Table</li>
               *   <li>parentElement: Empty rendered TH element</li>
               * </ul>
               * The function should return one of the following: 
               * <ul>
               *   <li>An Object with the following property:
               *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the header.</li></ul>
               *   </li>
               *   <li>undefined: If the developer chooses to manipulate the header element directly, the function should return undefined.</li>
               * </ul>
               * If no renderer is specified, the Table will treat the header data as a String.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].headerRenderer
               * @type {Function|null}
               * @ojsignature {target: "Type", value: "?((context: oj.ojTable.HeaderRendererContext<K,D>) => {insert: HTMLElement | string} | void) | null"}
               * @default null
               */
              headerRenderer: null,
              /**
               * The CSS styling to apply to the column header text.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].headerStyle
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              headerStyle: null,
              /**
               * The name of the slot used to specify the template for rendering the header cell. The slot must be a &lt;template> element.</p>  
               *
               * <p>When the template is executed for the header, it will have access to the binding context containing the following properties:</p>
               * <ul>
               * <li>$current - an object that contains information for the current header being rendered
               *   <ul>
               *     <li>componentElement - the &lt;oj-table> custom element</li>
               *     <li>data - the data for the current header being rendered</li>
               *   </ul>
               * </li>
               * <li>alias - if as attribute was specified, the value will be used to provide an application-named alias for $current.
               * </li>
               * </ul>
               *
               * <p>The content of the template should not include the &lt;th> element, only what's inside it.</p>
               * <p>When both headerTemplate and headerRenderer are specified, the headerRenderer takes precedence.</p>
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].headerTemplate
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              headerTemplate: null,
              /**
               * Text to display in the header of the column.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].headerText
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              headerText: null,
              /**
               * The identifier for the column
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].id
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              id: null,
              /**
               * Enable or disable width resize along the column end headers.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].resizable
               * @type {string}
               * @ojvalue {string} 'enabled'
               * @ojvalue {string} 'disabled'
               * @default "disabled"
               * @ojsignature {target:"Type", value:"?"}
               */
              resizable: 'disabled',
              /**
               * Whether or not the column is sortable.
               * <p>
               * A sortable column has a clickable header that (when clicked)
               * sorts the table by that column's property. Note that
               * in order for a column to be sortable, this attribute
               * must be set to "enabled" and the underlying model must
               * support sorting by this column's property. If this attribute
               * is set to "auto" then the column will be sortable if the
               * underlying model supports sorting. A value of "disabled" will
               * disable sorting on the column.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].sortable
               * @type {string}
               * @ojvalue {string} "auto" Column will be sortable if the underlying model supports sorting.
               * @ojvalue {string} "enabled" Enabled.
               * @ojvalue {string} "disabled" Disabled.
               * @default "auto"
               * @ojsignature {target:"Type", value:"?"}
               */
              sortable: 'auto',
              /**
               * Indicates the row attribute used for sorting when sort is invoked on this
               * column. Useful for concatenated columns, where the sort is done by only a subset
               * of the concatenated items.
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].sortProperty
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              sortProperty: null,
              /**
               * The CSS styling to apply to the column cells
               *
               * <p>See the <a href="#columns">columns</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].style
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              style: null,
              /**
               * The name of the slot used to specify the template for rendering the cell. The slot must be a &lt;template> element.</p>  
               *
               * <p>When the template is executed for the cell, it will have access to the binding context containing the following properties:</p>
               * <ul>
               * <li>$current - an object that contains information for the current cell being rendered
               *   <ul>
               *     <li>componentElement - the &lt;oj-table> custom element</li>
               *     <li>data - the data for the current cell being rendered</li>
               *     <li>index - the zero-based index of the current row being rendered</li>
               *     <li>key - the key of the current cell being rendered</li>
               *     <li>row - the data for the current row</li>
               *   </ul>
               * </li>
               * <li>alias - if as attribute was specified, the value will be used to provide an application-named alias for $current.
               * </li>
               * </ul>
               *
               * <p>The content of the template should not include the &lt;td> element, only what's inside it.</p>
               * <p>When both cell template and cell renderer are specified, the cell renderer takes precedence.</p>
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].template
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              template: null,
              /**
               * The width in px of the column
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columns[].width
               * @type {number|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              width: null
              /**
               * The knockout template used to render the content of the column header.
               *
               * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a
               * component option. The following
               *   variables are also passed into the template
               *     <ul>
               *       <li>$columnIndex: The column index</li>
               *       <li>$data: The header text</li>
               *       <li>$headerContext: The header context</li>
               *     </ul>
               *   </li>
               *
               * @ojbindingonly
               * @name columns[].headerTemplate
               * @memberof! oj.ojTable
               * @instance
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               *
               * @example <caption>Specify the column header <code class="prettyprint">template</code> when initializing Table:</caption>
               * // set the template
               * &lt;oj-table aria-label="Departments Table" 
               *      data='{{dataProvider}}' 
               *      columns='[{"headerText": "Department Id", "field": "DepartmentId"},
               *                {"headerText": "Department Name", "field": "DepartmentName"},
               *                {"headerText": "Location Id", "field": "LocationId"},
               *                {"headerText": "Manager Id", "field": "ManagerId"},
               *                {"headerTemplate": "oracle_link_hdr"}]'&gt;
               * &lt;/oj-table&gt;
               *
               * @ignore
               */
              /**
               * The knockout template used to render the content of the column footer.
               *
               * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a
               * component option. The following
               *   variables are also passed into the template
               *     <ul>
               *       <li>$columnIndex: The column index</li>
               *       <li>$footerContext: The header context</li>
               *     </ul>
               *   </li>
               *
               * @ojbindingonly
               * @name columns[].footerTemplate
               * @memberof! oj.ojTable
               * @instance
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               *
               * @example <caption>Specify the column footer <code class="prettyprint">template</code> when initializing Table:</caption>
               * // set the template
               * &lt;oj-table aria-label="Departments Table" 
               *      data='{{dataProvider}}' 
               *      columns-default='[{"headerText": "Department Id", "field": "DepartmentId"},
               *                       {"headerText": "Department Name", "field": "DepartmentName"},
               *                       {"headerText": "Location Id", "field": "LocationId"},
               *                       {"headerText": "Manager Id", "field": "ManagerId"},
               *                       {"footerTemplate": "oracle_link_ftr"}]'&gt;
               * &lt;/oj-table&gt;
               *
               * @ignore
               */
            }],
          /**
           * Default values to apply to all columns objects.
           * @expose
           * @public
           * @instance
           * @memberof! oj.ojTable
           * @type {Object.<string, string|null>}
           * @default null
           *
           * @example <caption>Initialize the component, overriding some columns defaults and leaving the others intact:</caption>
           * &lt;!-- Using dot notation -->
           * &lt;oj-table columns-default.sortable='disabled' columns-default.header-style='text-align: left; white-space:nowrap;'>&lt;/oj-table>
           * 
           * &lt;!-- Using JSON notation -->
           * &lt;oj-table columns-default='{"sortable":"disabled", "headerStyle":"text-align: left; white-space:nowrap;"}'>&lt;/oj-table>
           *
           * @example <caption>Get or set the <code class="prettyprint">columnsDefault</code> property after initialization:</caption>
           * // Get one
           * var value = myTable.columnsDefault.headerStyle;
           *
           * // Set one, leaving the others intact. Use the setProperty API for 
           * // subproperties so that a property change event is fired.
           * myTable.setProperty('columnsDefault.headerStyle', 'text-align: left; white-space:nowrap;');
           *
           * // Get all
           * var values = myTable.columnsDefault;
           *
           * // Set all.  Must list every default, as those not listed are lost.
           * myTable.columnsDefault = {
           *     sortable: 'disabled',
           *     headerStyle: 'text-align: left; white-space:nowrap;'
           * };
           */
          columnsDefault: {
              /**
               * The renderer function that renders the content of the cell.
               * The function will be passed a context object which contains
               * the following objects:
               * <ul>
               *   <li>cellContext.datasource: The "data" attribute of the Table</li>
               *   <li>cellContext.mode: The mode of the row.  It can be "edit" or "navigation".</li>
               *   <li>cellContext.status: Contains the rowIndex, rowKey, and currentRow</li>
               *   <li>columnIndex: The column index</li>
               *   <li>componentElement: A reference to the Table root element</li>
               *   <li>data: The cell data</li>
               *   <li>parentElement: Empty rendered <td> element</li>
               *   <li>row: Key/value pairs of the row</li>
               * </ul>
               * The function should return one of the following: 
               * <ul>
               *   <li>An Object with the following property:
               *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the header.</li></ul>
               *   </li>
               *   <li>undefined: If the developer chooses to manipulate the cell element directly, the function should return undefined.</li>
               * </ul>
               * If no renderer is specified, the Table will treat the cell data as a String.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.renderer
               * @type {Function|null}
               * @ojsignature {target: "Type", value: "?((context: oj.ojTable.ColumnsRendererContext<K,D>) => {insert: HTMLElement | string} | void) | null"}
               * @default null
               */
              renderer: null,
              /**
               * The default CSS class for column cells
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.className
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              className: null,
              /**
               * The default data field for column.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.field
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              field: null,
              /**
               * The CSS class to apply to the footer cell.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.footerClassName
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              footerClassName: null,
              /**
               * The renderer function that renders the content of the footer.
               * The function will be passed a context object which contains
               * the following objects:
               * <ul>
               *   <li>columnIndex: The column index</li>
               *   <li>componentElement: A reference to the Table root element</li>
               *   <li>footerContext.datasource: The "data" attribute of the Table</li>
               *   <li>parentElement: Empty rendered <td> element</li>
               * </ul>
               * The function should return one of the following: 
               * <ul>
               *   <li>An Object with the following property:
               *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the footer.</li></ul>
               *   </li>
               *   <li>undefined: If the developer chooses to manipulate the footer element directly, the function should return undefined.</li>
               * </ul>
               * If no renderer is specified, the Table will treat the footer data as a String.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.footerRenderer
               * @type {Function|null}
               * @ojsignature {target: "Type", value: "?((context: oj.ojTable.FooterRendererContext<K,D>) => {insert: HTMLElement | string} | void) | null"}
               * @default null
               */
              footerRenderer: null,
              /**
               * The CSS styling to apply to the footer cell.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.footerStyle
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              footerStyle: null,
              /**
               * The name of the slot used to specify the template for rendering the footer cell. The slot must be a &lt;template> element.</p>  
               *
               * <p>When the template is executed each footer, it will have access to the binding context containing the following properties:</p>
               * <ul>
               * <li>$current - an object that contains information for the current footer being rendered
               *   <ul>
               *     <li>componentElement - the &lt;oj-table> custom element</li>
               *   </ul>
               * </li>
               * <li>alias - if as attribute was specified, the value will be used to provide an application-named alias for $current.
               * </li>
               * </ul>
               *
               * <p>The content of the template should not include the &lt;td> element, only what's inside it.</p>
               * <p>When both footerTemplate and footerRenderer are specified, the footerRenderer takes precedence.</p>
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.footerTemplate
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              footerTemplate: null,
              /**
               * The default CSS class to apply to the column header.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.headerClassName
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              headerClassName: null,
              /**
               * The renderer function that renders the content of the header.
               * The function will be passed a context object which contains
               * the following objects:
               * <ul>
               *   <li>columnIndex: The column index</li>
               *   <li>columnHeaderDefaultRenderer(options, delegateRenderer): If the column
               *   is not sortable then this function will be included in the context.
               *   The options parameter specifies the options (future use) for the renderer while the
               *   delegateRenderer parameter specifies the function which the developer would
               *   like to be called during rendering of the column header.</li>
               *   <li>columnHeaderSortableIconRenderer(options, delegateRenderer): If the column
               *   is sortable then this function will be included in the context.
               *   The options parameter specifies the options (future use) for the renderer while the
               *   delegateRenderer parameter specifies the function which the developer would
               *   like to be called during rendering of the sortable column header. Calling the
               *   columnHeaderSortableIconRenderer function enables rendering custom header content
               *   while also preserving the sort icons.</li>
               *   <li>componentElement: A reference to the Table root element</li>
               *   <li>data: The header text for the column</li>
               *   <li>headerContext.datasource: The "data" attribute of the Table</li>
               *   <li>parentElement: Empty rendered TH element</li>
               * </ul>
               * The function should return one of the following: 
               * <ul>
               *   <li>An Object with the following property:
               *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the header.</li></ul>
               *   </li>
               *   <li>undefined: If the developer chooses to manipulate the header element directly, the function should return undefined.</li>
               * </ul>
               * If no renderer is specified, the Table will treat the header data as a String.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.headerRenderer
               * @type {Function|null}
               * @ojsignature {target: "Type", value: "?((context: oj.ojTable.HeaderRendererContext<K,D>) => {insert: HTMLElement | string} | void) | null"}
               * @default null
               */
              headerRenderer: null,
              /**
               * The default CSS styling to apply to the column header.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.headerStyle
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              headerStyle: null,
              /**
               * The name of the slot used to specify the template for rendering the header cell. The slot must be a &lt;template> element.</p>  
               *
               * <p>When the template is executed for the header, it will have access to the binding context containing the following properties:</p>
               * <ul>
               * <li>$current - an object that contains information for the current header being rendered
               *   <ul>
               *     <li>componentElement - the &lt;oj-table> custom element</li>
               *     <li>data - the data for the current header being rendered</li>
               *   </ul>
               * </li>
               * <li>alias - if as attribute was specified, the value will be used to provide an application-named alias for $current.
               * </li>
               * </ul>
               *
               * <p>The content of the template should not include the &lt;th> element, only what's inside it.</p>
               * <p>When both headerTemplate and headerRenderer are specified, the headerRenderer takes precedence.</p>
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.headerTemplate
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              headerTemplate: null,
              /**
               * Default text to display in the header of the column.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.headerText
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              headerText: null,
              /**
               * Enable or disable width resize along the column end headers.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.resizable
               * @type {string|null}
               * @ojvalue {string} 'enabled'
               * @ojvalue {string} 'disabled'
               * @default "disabled"
               * @ojsignature {target:"Type", value:"?"}
               */
              resizable: 'disabled',
              /**
               * Whether or not the column is sortable.
               * <p>
               * A sortable column has a clickable header that (when clicked)
               * sorts the table by that column's property. Note that
               * in order for a column to be sortable, this attribute
               * must be set to "enabled" and the underlying model must
               * support sorting by this column's property. If this attribute
               * is set to "auto" then the column will be sortable if the
               * underlying model supports sorting. A value of "disabled" will
               * disable sorting on the column.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.sortable
               * @type {string|null}
               * @ojvalue {string} "auto" Column will be sortable if the underlying model supports sorting.
               * @ojvalue {string} "enabled" Enabled.
               * @ojvalue {string} "disabled" Disabled.
               * @default "auto"
               * @ojsignature {target:"Type", value:"?"}
               */
              sortable: 'auto',
              /**
               * Indicates the row attribute used for sorting when sort is invoked on this
               * column. Useful for concatenated columns, where the sort is done by only a subset
               * of the concatenated items.
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.sortProperty
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              sortProperty: null,
              /**
               * Default CSS styling to apply to the column cells
               *
               * <p>See the <a href="#columnsDefault">columns-default</a> attribute for usage examples.
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.style
               * @type {string|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              style: null,
              /**
               * The name of the slot used to specify the template for rendering the cell. The slot must be a &lt;template> element.</p>  
               *
               * <p>When the template is executed for the cell, it will have access to the binding context containing the following properties:</p>
               * <ul>
               * <li>$current - an object that contains information for the current cell being rendered
               *   <ul>
               *     <li>componentElement - the &lt;oj-table> custom element</li>
               *     <li>data - the data for the current cell being rendered</li>
               *     <li>index - the zero-based index of the current row being rendered</li>
               *     <li>key - the key of the current cell being rendered</li>
               *   </ul>
               * </li>
               * <li>alias - if as attribute was specified, the value will be used to provide an application-named alias for $current.
               * </li>
               * </ul>
               *
               * <p>The content of the template should not include the &lt;td> element, only what's inside it.</p>
               * <p>When both cell template and cell renderer are specified, the cell renderer takes precedence.</p>
               *
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.template
               * @type {string|null}
               * @ojsignature {target:"Type", value:"?"}
               * @default null
               */
              template: null,
              /**
               * Default CSS width to apply to the column
               * @expose
               * @public
               * @instance
               * @memberof! oj.ojTable
               * @alias columnsDefault.width
               * @type {number|null}
               * @default null
               * @ojsignature {target:"Type", value:"?"}
               */
              width: null
               /**
                * The default knockout template used to render the content of the column header.
                *
                * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a
                * component option. The following
                *   variables are also passed into the template
                *     <ul>
                *       <li>$columnIndex: The column index</li>
                *       <li>$data: The header text</li>
                *       <li>$headerContext: The header context</li>
                *     </ul>
                *   </li>
                *
                * @ojbindingonly
                * @name columnsDefault.headerTemplate
                * @memberof! oj.ojTable
                * @instance
                * @type {string|null}
                * @ojsignature {target:"Type", value:"?"}
                * @default null
                *
                * @example <caption>Specify the column header <code class="prettyprint">template</code> when initializing Table:</caption>
                * // set the template
                * &lt;oj-table aria-label="Departments Table" 
                *      data='{{dataProvider}}'
                *      columns-default='[{"headerText": "Department Id", "field": "DepartmentId"},
                *                        {"headerText": "Department Name", "field": "DepartmentName"},
                *                        {"headerText": "Location Id", "field": "LocationId"},
                *                        {"headerText": "Manager Id", "field": "ManagerId"},
                *                        {"headerTemplate": "oracle_link_hdr"}]'&gt;
                * &lt;/oj-table&gt;
                *
                * @ignore
                */
               /**
                * The default knockout template used to render the content of the column footer.
                *
                * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a
                * component option. The following
                *   variables are also passed into the template
                *     <ul>
                *       <li>$columnIndex: The column index</li>
                *       <li>$footerContext: The header context</li>
                *     </ul>
                *   </li>
                *
                * @ojbindingonly
                * @name columnsDefault.footerTemplate
                * @memberof! oj.ojTable
                * @instance
                * @type {string|null}
                * @ojsignature {target:"Type", value:"?"}
                * @default null
                *
                * @example <caption>Specify the column footer <code class="prettyprint">template</code> when initializing Table:</caption>
                * // set the template
                * &lt;oj-table aria-label="Departments Table" 
                *      data='{{dataProvider}}' 
                *      columns-default='[{"headerText": "Department Id", "field": "DepartmentId"},
                *                        {"headerText": "Department Name", "field": "DepartmentName"},
                *                        {"headerText": "Location Id", "field": "LocationId"},
                *                        {"headerText": "Manager Id", "field": "ManagerId"},
                *                        {"footerTemplate": "oracle_link_ftr"}]'&gt;
                * &lt;/oj-table&gt;
                *
                * @ignore
                */
            },
            /**
              * Triggered before the current row is changed via the <code class="prettyprint">currentRow</code> property or via the UI.
              *
              * @expose
              * @event
              * @ojbubbles
              * @ojcancelable
              * @memberof oj.ojTable
              * @instance
              * @property {oj.ojTable.CurrentRow<K>} currentRow the new current row
              * @property {oj.ojTable.CurrentRow<K>} previousCurrentRow the previous current row
              * @ojsignature {target:"Type", value:"<K>", for:"genericTypeParameters"}
              */
            beforeCurrentRow: null,
            /**
              * Triggered before the table is going to enter edit mode. To prevent editing the row, call <code class="prettyprint">event.preventDefault()</code> in the listener.
              *
              * @expose
              * @event
              * @ojbubbles
              * @ojcancelable
              * @memberof oj.ojTable
              * @instance
              * @property {Object} rowContext the rowContext of the row that editing is going to be performed on.
              */
            beforeRowEdit: null,
             /**
              * Triggered before the table is going to exit edit mode. To prevent exit editing, call <code class="prettyprint">event.preventDefault()</code> in the listener. 
              * There is a provided beforeRowEditEnd function, oj.DataCollectionEditUtils.basicHandleRowEditEnd, which can be specified. 
              * This function will handle canceling edits as well as invoking validation on input elements.
              *
              * @expose
              * @event
              * @ojbubbles
              * @ojcancelable
              * @memberof oj.ojTable
              * @instance
              * @property {Object} rowContext the rowContext of the row that editing is going to be performed on.
              * @property {Object} cancelEdit true if the edit should be negated based on actions (i.e. escape key).
              */
            beforeRowEditEnd: null,
          /**
            * Triggered when the table has finished rendering
            *
            * @expose
            * @event
            * @memberof! oj.ojTable
            * @instance
            *
            * @ignore
            */
          ready: null,
            /**
              * Triggered when a sort is performed on the table
              *
              * @expose
              * @event
              * @ojbubbles
              * @ojcancelable
              * @memberof oj.ojTable
              * @instance
              * @property {Element} header the key of the header which was sorted on
              * @property {'ascending'|'descending'} direction the direction of the sort
              */
            sort: null
             /**
              * The knockout template used to render the content of the row.
              *
              * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a
              * component option. The following
              *   variables are also passed into the template
              *     <ul>
              *       <li>$rowContext: The row context</li>
              *     </ul>
              *   </li>
              *
              * @ojbindingonly
              * @name rowTemplate
              * @memberof! oj.ojTable
              * @instance
              * @type {string|null}
              * @default null
              *
              * @example <caption>Specify the row <code class="prettyprint">template</code> when initializing Table:</caption>
              * // set the template
              * &lt;oj-table aria-label="Departments Table" 
              *   data='{{dataProvider}}' row-template='row_tmpl'&gt;&lt;/oj-table&gt;
              *
              * @ignore
              */
        },
      /**
       * @private
       */
      _BUNDLE_KEY:
        {
          _MSG_FETCHING_DATA:                             'msgFetchingData',
          _MSG_NO_DATA:                                   'msgNoData',
          _MSG_INITIALIZING:                              'msgInitializing',
          _MSG_STATUS_SORT_ASC:                           'msgStatusSortAscending',
          _MSG_STATUS_SORT_DSC:                           'msgStatusSortDescending',
          _LABEL_SELECT_COLUMN:                           'labelSelectColumn',
          _LABEL_SELECT_ROW:                              'labelSelectRow',
          _LABEL_EDIT_ROW:                                'labelEditRow',
          _LABEL_SELECT_AND_EDIT_ROW:                     'labelSelectAndEditRow'
        },
      /**
       * @private
       */
      _LOGGER_MSG:
        {
          _ERR_PRECURRENTROW_ERROR_SUMMARY:                'Did not change current row due to error.',
          _ERR_PRECURRENTROW_ERROR_DETAIL:                 'Error detail: {error}.',
          _ERR_CURRENTROW_UNAVAILABLE_INDEX_SUMMARY:       'Did not change current row due to unavailable row index.',
          _ERR_CURRENTROW_UNAVAILABLE_INDEX_DETAIL:        'Unavailable row index: {rowIdx}.',
          _ERR_REFRESHROW_INVALID_INDEX_SUMMARY:          'Invalid row index value.',
          _ERR_REFRESHROW_INVALID_INDEX_DETAIL:           'Row index: {rowIdx}.',
          _ERR_DATA_INVALID_TYPE_SUMMARY:                 'Invalid data type.',
          _ERR_DATA_INVALID_TYPE_DETAIL:                  'Please specify the appropriate data type.',
          _ERR_ELEMENT_INVALID_TYPE_SUMMARY:               'Invalid element type.',
          _ERR_ELEMENT_INVALID_TYPE_DETAIL:                'Only a <table> element can be specified for ojTable.'
        },
      /**
       * @private
       * @type {string}
       */
      _FIELD_ID:                                          'id',
      /**
       * @private
       * @type {string}
       */
      _CONST_ATTRIBUTE:                                   'attribute',
      /**
       * @private
       * @type {string}
       */
      _CONST_DATA:                                        'data',
      /**
       * @private
       * @type {string}
       */
      _CONST_METADATA:                                    'metadata',
      /**
       * @private
       * @type {string}
       */
      _CONST_INDEXES:                                     'indexes',
      /**
       * @private
       * @type {string}
       */
      _CONST_INDEX:                                       'index',
      /**
       * @private
       * @type {string}
       */
      _CONST_KEY:                                         'key',
      /**
       * @private
       * @type {string}
       */
      _CONST_KEYS:                                        'keys',
      /**
       * @private
       * @type {string}
       */
      _CONST_AFTERKEYS:                                   'afterKeys',
      /**
       * @private
       * @type {string}
       */
      _CONST_STARTINDEX:                                  'startIndex',
      /**
       * @private
       * @type {string}
       */
      _CONST_ENDINDEX:                                    'endIndex',
      /**
       * @private
       * @type {string}
       */
      _CONST_PAGESIZE:                                     'size',
      /**
       * @private
       * @type {string}
       */
      _CONST_OFFSET:                                        'offset',
      /**
       * @private
       * @type {string}
       */
      _CONST_COLUMN:                                      'column',
      /**
       * @private
       * @type {string}
       */
      _CONST_ROW:                                         'row',
      /**
       * @private
       * @type {string}
       */
      _CONST_VALUE:                                       'value',
      /**
       * @private
       * @type {string}
       */
      _CONST_SORTCRITERIA:                                'sortCriteria',
      /**
       * @private
       * @type {string}
       */
      _COLUMN_HEADER_ID:                                  '_headerColumn',
      /**
       * @private
       * @type {string}
       */
      _COLUMN_HEADER_TEXT_ID:                             '_headerColumnText',
      /**
       * @private
       * @type {string}
       */
      _COLUMN_HEADER_ASC_ID:                              '_headerColumnAsc',
      /**
       * @private
       * @type {string}
       */
      _COLUMN_HEADER_DSC_ID:                              '_headerColumnDsc',
      /**
       * @private
       * @type {string}
       */
      _COLUMN_HEADER_ID_PREFIX:                           '_hdrCol',
      /**
       * @private
       * @type {string}
       */
      _FOCUS_CALLED:                                      '_focusedCalled',
      /**
       * @private
       * @type {string}
       */
      _ROW_TEMPLATE:                                      'rowTemplate',
      /**
       * @private
       * @type {string}
       */
      _CELL_TEMPLATE:                                      'template',
      /**
       * @private
       * @type {string}
       */
      _HEADER_TEMPLATE:                                    'headerTemplate',
      /**
       * @private
       * @type {string}
       */
      _FOOTER_TEMPLATE:                                     'footerTemplate',
      /**
       * @private
       * @type {string}
       */
      _OPTION_AUTO:                                       'auto',
      /**
       * @private
       * @type {string}
       */
      _OPTION_ENABLED:                                    'enabled',
      /**
       * @private
       * @type {string}
       */
      _OPTION_DISABLED:                                   'disabled',
      /**
       * @private
       */
      _OPTION_EDIT_MODE:
        {
          _NONE:                                          'none',
          _ROW_EDIT:                                      'rowEdit'
        },
      /**
       * @private
       */
      _OPTION_SELECTION_MODES:
        {
          _SINGLE:                                        'single',
          _MULTIPLE:                                      'multiple'
        },
      /**
       * @private
       */
        _OPTION_SCROLL_POLICY:
        {
          _AUTO:                                          'auto',
          _LOADMORE_ON_SCROLL:                            'loadMoreOnScroll'
        },
      /**
       * @private
       */
      _COLUMN_SORT_ORDER:
        {
          _ASCENDING:                                     'ascending',
          _DESCENDING:                                    'descending'
        },
      /**
       * @private
       */
      _DND_REORDER_TABLE_ID_DATA_KEY:                       'oj-table-dnd-reorder-table-id',
      /**
       * @private
       */
      _KEYBOARD_CODES:
        {
          _KEYBOARD_CODE_SPACEBAR:                        32,
          _KEYBOARD_CODE_ENTER:                           13,
          _KEYBOARD_CODE_UP:                              38,
          _KEYBOARD_CODE_DOWN:                            40,
          _KEYBOARD_CODE_LEFT:                            37,
          _KEYBOARD_CODE_RIGHT:                           39,
          _KEYBOARD_CODE_HOME:                            36,
          _KEYBOARD_CODE_END:                             35,
          _KEYBOARD_CODE_TAB:                             9,
          _KEYBOARD_CODE_ESC:                             27,
          _KEYBOARD_CODE_F2:                              113,
          _KEYBOARD_MODIFIER_SHIFT:                       'shiftKey'
        },
      /**** start Public APIs ****/
      /**
       * {@ojinclude "name":"nodeContextDoc"}
       * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
       * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
       * @ojsignature {target:"Type", value:"{subId: 'oj-table-cell', rowIndex: number, columnIndex: number, key:string} | {subId: 'oj-table-footer'|'oj-table-header',index: number}",for:"returns"}
       *
       * @example {@ojinclude "name":"nodeContextExample"}
       *
       * @expose
       * @instance
       * @memberof oj.ojTable
       */
      getContextByNode: function(node)
      {
        // context objects are documented with @ojnodecontext
        var context = this.getSubIdByNode(node, true);
        
        if (context) 
        {
          if (context['subId'] == 'oj-table-cell')
          {
            var rowIdx = context['rowIndex'];
            var rowKey = this._getRowKeyForRowIdx(rowIdx);
            context['key'] = rowKey;
          }
        }
        
        return context;
      },
      /**
       * Return the row data for a rendered row in the table.
       * @param {number} rowIndex row index
       * @returns {Object | null} a compound object which has the structure below. If the row has not been rendered, returns null.<p>
       * <table>
       * <tbody>
       * <tr><td><b>data</b></td><td>The raw row data</td></tr>
       * <tr><td><b>index</b></td><td>The index for the row</td></tr>
       * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
       * </tbody>
       * </table>
       * @ojsignature {target: "Type", value: "(rowIndex: number) : {data: D, index: number, key: K} | null"}
       * @export
       * @expose
       * @memberof oj.ojTable
       * @instance
       * @example <caption>Invoke the <code class="prettyprint">getDataForVisibleRow</code> method:</caption>
       * myTable.getDataForVisibleRow(2);
       */
      'getDataForVisibleRow': function(rowIndex)
      {
        var tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIndex);
        
        if (tableBodyRow != null)
        {
          return {'key': $(tableBodyRow).data('rowKey'),
                  'data': $(tableBodyRow).data('rowData'),
                  'index': rowIndex};
        }
        
        return null;
      },
      // @inheritdoc
      'getNodeBySubId': function(locator)
      {
        if (locator == null)
        {
          return this.element ? this.element[0] : null;
        }

        var subId = locator['subId'];

        if (subId === 'oj-table-cell')
        {
          var rowIdx = parseInt(locator['rowIndex'], 10);
          var columnIdx = parseInt(locator['columnIndex'], 10);
          return this._getTableDomUtils().getTableBodyLogicalCells(rowIdx)[columnIdx];
        }
        else if (subId === 'oj-table-header' ||
                 subId === 'oj-table-sort-ascending' ||
                 subId === 'oj-table-sort-descending')
        {
          var columnIdx = locator['index'];
          var tableHeaderColumn = this._getTableDomUtils().getTableHeaderLogicalColumns()[columnIdx];
          if (tableHeaderColumn != null)
          {
            if (subId === 'oj-table-header')
            {
              return tableHeaderColumn;
            }
            else if (subId === 'oj-table-sort-ascending')
            {
              var tableHeaderColumnSortAsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_LINK_CLASS);
              if (tableHeaderColumnSortAsc.length > 0)
              {
                return tableHeaderColumnSortAsc[0];
              }
            }
            else
            {
              var tableHeaderColumnSortDsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_LINK_CLASS);
              if (tableHeaderColumnSortDsc.length > 0)
              {
                return tableHeaderColumnSortDsc[0];
              }
            }
          }
        }
        else if (subId === 'oj-table-footer')
        {
          var columnIdx = locator['index'];
          var tableFooterCell = this._getTableDomUtils().getTableFooterLogicalCells()[columnIdx];

          if (tableFooterCell != null)
          {
            return tableFooterCell;
          }
        }

        // Non-null locators have to be handled by the component subclasses
        return null;
      },
      // @inheritdoc
      'getSubIdByNode': function(node, ignoreSortIcons)
      {
        var jqCell = $(node).closest('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_CLASS);
        
        if (jqCell.length > 0)
        {
          return {'subId': 'oj-table-cell',
                  'rowIndex': this._getTableDomUtils().getElementRowIdx(jqCell[0]),
                  'columnIndex': this._getTableDomUtils().getElementColumnIdx(jqCell[0])};
        }
        
        var jqHeaderSortAsc = $(node).closest('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_LINK_CLASS);
        
        if (jqHeaderSortAsc.length > 0)
        {
          return {'subId': ignoreSortIcons ? 'oj-table-header' : 'oj-table-sort-ascending',
                  'index': this._getTableDomUtils().getElementColumnIdx(jqHeaderSortAsc[0])};
        }
        
        var jqHeaderSortDsc = $(node).closest('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_LINK_CLASS);
        
        if (jqHeaderSortDsc.length > 0)
        {
          return {'subId': ignoreSortIcons ? 'oj-table-header' : 'oj-table-sort-descending',
                  'index': this._getTableDomUtils().getElementColumnIdx(jqHeaderSortDsc[0])};
        }
        
        var jqHeader = $(node).closest('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS);
        
        if (jqHeader.length > 0)
        {
          return {'subId': 'oj-table-header',
                  'index': this._getTableDomUtils().getElementColumnIdx(jqHeader[0])};
        }
        
        var jqFooter = $(node).closest('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CELL_CLASS);
        
        if (jqFooter.length > 0)
        {
          return {'subId': 'oj-table-footer',
                  'index': this._getTableDomUtils().getElementColumnIdx(jqFooter[0])};
        }
        
        return null;
      },
      /**
       * Refresh the table.
       * @export
       * @expose
       * @memberof oj.ojTable
       * @instance
       * @return {void}
       * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
       * myTable.refresh();
       */
      'refresh': function()
      {
        this._super();
        this._refresh();
      },
      /**
       * Refresh a row in the table.
       * @param {number} rowIdx  Index of the row to refresh.
       * @return {Promise.<boolean>} Promise resolves when done to true if refreshed, false if not
       * @throws {Error}
       * @export
       * @expose
       * @memberof oj.ojTable
       * @instance
       * @example <caption>Invoke the <code class="prettyprint">refreshRow</code> method:</caption>
       * myTable.refreshRow(1);
       */
      'refreshRow': function(rowIdx)
      {
        var dataprovider = this._getData();
        // if no data then bail
        if (!dataprovider)
        {
          return Promise.resolve(false);
        }

        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();

        if (isNaN(rowIdx) || 
          rowIdx < 0 || 
          (tableBodyRows != null && rowIdx >= tableBodyRows.length) || 
          tableBodyRows == null || 
          tableBodyRows.length == 0)
        {
          // validate rowIdx value
          var errSummary = this._LOGGER_MSG._ERR_REFRESHROW_INVALID_INDEX_SUMMARY;
          var errDetail = oj.Translations.applyParameters(this._LOGGER_MSG._ERR_REFRESHROW_INVALID_INDEX_DETAIL, {'rowIdx': rowIdx.toString()});
          throw new RangeError(errSummary + '\n' + errDetail);
        }

        // get row at rowIdx        
        var rowKey = this._getRowKeyForRowIdx(rowIdx);
        var self = this;
        return self._queueTask(function()
        {
          return dataprovider.fetchByKeys({'keys':[rowKey]}).then(function(keyResult)
          {
            if (keyResult == null || 
              keyResult.results == null ||
              keyResult.results.size == 0)
            {
              return Promise.resolve(false);
            }
            self._refreshTableBodyRow(rowIdx, {'data': keyResult.results.get(rowKey)['data'], 'index': rowIdx, 'key': rowKey});
            var tableBodyRow = self._getTableDomUtils().getTableBodyRow(rowIdx);
            var busyContextArray = self._getAllBusyContextsForElement(tableBodyRow);

            if (busyContextArray != null)
            {
              var i, busyContextPromiseArray = [];
              for (i = 0; i < busyContextArray.length; i++)
              {
                busyContextPromiseArray.push(busyContextArray[i].whenReady());
              }

              // If there are components busy then wait first
              return Promise.all(busyContextPromiseArray);
            }
            return Promise.resolve();
          });
        });
      },
      /**
       * Returns a jQuery object containing the root dom element of the table
       *
       * <p>This method does not accept any arguments.
       *
       * @expose
       * @override
       * @memberof oj.ojTable
       * @instance
       * @ignore
       * @return {jQuery} the root DOM element of table
       */
      'widget' : function ()
      {
        var tableContainer = this._getTableDomUtils().getTableContainer();

        if (tableContainer != null)
        {
          return $(tableContainer);
        }
        return this.element;
      },
      /**** end Public APIs ****/

      /**** start internal widget functions ****/

      /**
       * @override
       * @protected
       * @instance
       * @memberof! oj.ojTable
       */
      _ComponentCreate : function ()
      {
        this._super();
        this._initTemplateEngine();
        this._draw();
        this._registerCustomEvents();
        this._on(this._events);
        if (this._isTouchDevice())
        {
          var tableContainer = this._getTableDomUtils().getTableContainer();
          this._on($(tableContainer), this._eventsContainer);
        }
        this._registerDomEventListeners();
        // cache the options
        this._cachedOptions = $.extend(true, {}, this.options);
        this._setEditableRowIdx(null);
      },
      /**
       * Initialize the table after creation
       * @protected
       * @override
       * @memberof! oj.ojTable
       */
      _AfterCreate: function ()
      {
        this._super();
        // create the context menu
        this._getTableDomUtils().createContextMenu(this._handleContextMenuSelect.bind(this));
        this._isInitFetch = true;
      },
      /**
       * Sets up needed resources for table, for example, add
       * listeners.
       * @protected
       * @override
       * @memberof! oj.ojTable
       */
      _SetupResources: function ()
      {
        this._super();
        this._registerResizeListener();
        // register event listeners for table on the datasource so that the table
        // component is notified when rows are added, deleted, etc from the datasource.
        this._registerDataSourceEventListeners();
        if (this._isInitFetch)
        {
          this._initFetch();
          this._isInitFetch = false;
        }
        else
        {
          this._invokeDataFetchRows(null, false);
        }
      },
      /**
       * Releases resources for table.
       * @protected
       * @override
       * @memberof! oj.ojTable
       */
      _ReleaseResources: function ()
      {
        this._super();
        // unregister the listeners on the datasource
        this._unregisterDataSourceEventListeners();
        this._unregisterResizeListener();
      },
      /**
       * <p>Notifies the component that its subtree has been connected to the document programmatically after the component has
       * been created.
       *
       * @memberof! oj.ojTable
       * @instance
       * @protected
       * @override
       */
      _NotifyAttached: function()
      {
        this._super();
        this._getTableDomUtils().refreshTableDimensions();
      },
      /**
       * <p>Notifies the component that its subtree has been made visible programmatically after the component has
       * been created.
       *
       * @memberof! oj.ojTable
       * @instance
       * @protected
       * @override
       */
      _NotifyShown: function()
      {
        this._super();
        this._getTableDomUtils().refreshTableDimensions();
      },
      //** @inheritdoc */
      _GetDefaultContextMenu: function()
      {
        return this._defaultContextMenu;
      },
      /**
       * @param {Object} contextMenu The JET Menu to open as a context menu
       * @param {Event} event What triggered the menu launch
       * @param {string} eventType "mouse", "touch", "keyboard"
       * @private
       */
      _NotifyContextMenuGesture: function(contextMenu, event, eventType)
      {
        var openOptions = {};
        this._contextMenuEvent = event['originalEvent'];

        // first check if we are invoking on an editable or clickable element, or draggable element on touch event. If so bail
        if (this._isNodeEditable(this._contextMenuEvent['target']) || this._isNodeClickable(this._contextMenuEvent['target']) ||
            (eventType == 'touch' && this._isNodeDraggable(this._contextMenuEvent['target'])))
        {
          return;
        }

        var headerColumn = this._getTableDomUtils().getFirstAncestor(this._contextMenuEvent['target'], 'oj-table-column-header-cell');
        headerColumn = headerColumn == null ? this._getTableDomUtils().getTableHeaderColumn(this._activeColumnIndex) : headerColumn;
        this._contextMenuEventHeaderColumn = headerColumn;
        var tableBodyCell = this._getTableDomUtils().getFirstAncestor(this._contextMenuEvent['target'], 'oj-table-data-cell');

        if (tableBodyCell != null)
        {
          var columnIdx = this._getTableDomUtils().getElementColumnIdx(tableBodyCell);
          headerColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);
        }

        if (this._contextMenuEvent['type'] === 'keydown')
        {
          var table = this._tableDomUtils.getTable();

          if (this._contextMenuEvent['target'] == table)
          {
            if (headerColumn != null)
            {
              openOptions['position'] = {"my": "start top", "at": "start bottom", "of": headerColumn};
            }
            else
            {
              var focusedRowIdx = this._getFocusedRowIdx();
              if (focusedRowIdx >= 0)
              {
                var tableBodyRow = this._getTableDomUtils().getTableBodyRow(focusedRowIdx);
                openOptions['position'] = {"my": "start top", "at": "start bottom", "of": tableBodyRow};
              }
              else
              {
                openOptions['position'] = {"my": "start top", "at": "start bottom", "of": this._contextMenuEvent['target']};
              }
            }
          }
          else
          {
            openOptions['position'] = {"my": "start top", "at": "start bottom", "of": this._contextMenuEvent['target']};
          }
        }

        // for jQuery it's the menu instance, for custom element it's the element node
        var contextMenuNode = contextMenu['element'] ? contextMenu['element'][0] : contextMenu;

        if (headerColumn && headerColumn.getAttribute('data-oj-sortable') == this._OPTION_ENABLED)
        {
          var contextMenuItemAsc = contextMenuNode.querySelectorAll('[data-oj-command=oj-table-sortAsc]');
          if (contextMenuItemAsc.length > 0)
          {
            contextMenuItemAsc = contextMenuItemAsc[0];
            contextMenuItemAsc.classList.remove('oj-disabled');
          }
          var contextMenuItemDsc = contextMenuNode.querySelectorAll('[data-oj-command=oj-table-sortDsc]');
          if (contextMenuItemDsc.length > 0)
          {
            contextMenuItemDsc = contextMenuItemDsc[0];
            contextMenuItemDsc.classList.remove('oj-disabled');
          }
        }
        else
        {
          var contextMenuItemAsc = contextMenuNode.querySelectorAll('[data-oj-command=oj-table-sortAsc]');
          if (contextMenuItemAsc.length > 0)
          {
            contextMenuItemAsc = contextMenuItemAsc[0];
            contextMenuItemAsc.classList.add('oj-disabled');
          }
          var contextMenuItemDsc = contextMenuNode.querySelectorAll('[data-oj-command=oj-table-sortDsc]');
          if (contextMenuItemDsc.length > 0)
          {
            contextMenuItemDsc = contextMenuItemDsc[0];
            contextMenuItemDsc.classList.remove('oj-disabled');
          }
        }
        if (headerColumn && headerColumn.getAttribute('data-oj-resizable') == this._OPTION_ENABLED)
        {
          var contextMenuItemResize = contextMenuNode.querySelectorAll('[data-oj-command=oj-table-resize]');
          if (contextMenuItemResize.length > 0)
          {
            contextMenuItemResize = contextMenuItemResize[0];
            contextMenuItemResize.classList.remove('oj-disabled');
          }
        }
        else
        {
          var contextMenuItemResize = contextMenuNode.querySelectorAll('[data-oj-command=oj-table-resize]');
          if (contextMenuItemResize.length > 0)
          {
            contextMenuItemResize = contextMenuItemResize[0];
            contextMenuItemResize.classList.add('oj-disabled');
          }
        }
        this._OpenContextMenu(event, eventType, openOptions);
      },
      /**
       * @override
       * @private
       */
      _destroy: function()
      {
        $(this._getTableDomUtils().getTableBody()).removeAttr(oj.Components._OJ_CONTAINER_ATTR);

        this.element.children().remove('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_HEADER_CLASS);
        this.element.children().remove('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_CLASS);
        this.element.children().remove('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CLASS);
        this.element.children().remove('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_STATUS_MESSAGE_CLASS);
        this.element.children().remove('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_MESSAGE_ROW_CLASS);

        //  - DomUtils.unwrap() will avoid unwrapping if the node is being destroyed by Knockout
        oj.DomUtils.unwrap(this.element, $(this._getTableDomUtils().getTableContainer()));

        this.element[0].classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_CLASS);
        
        // Remove any pending busy states
        if (this._readyResolveFunc) 
        {
          this._readyResolveFunc();
          this._readyResolveFunc = null;
        }
        if (this._dataResolveFunc)
        {
          this._dataResolveFunc();
          this._dataResolveFunc = null;
        }

        this._componentDestroyed = true;
      },
      /**
       * @override
       * @private
       */
      _draw: function()
      {
        var options = this.options;

        this._setFinalTask(function()
        {
          this._getTableDomUtils().refreshTableDimensions();
          this._setSelection(this.options['selection']);
          this._setSelectionDefault();
          
          if (this._hasEditableRow())
          {
            // set the editable row class if editable
            var editableRowIdx = this._getEditableRowIdx();
            var editableTableBodyRow = this._getTableDomUtils().getTableBodyRow(editableRowIdx);

            if (editableTableBodyRow != null)
            {
              editableTableBodyRow.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_EDIT_CLASS);
            }
          }
          // if loadMoreOnScroll then check if we have underflow and do a
          // fetch if we do
          if (this._isLoadMoreOnScroll() && !this._dataFetching && this._domScroller)
          {
            this._domScroller.checkViewport().then(this._domScrollerSuccessFunc, null);
          }
        });

        if (!this.element.is('table'))
        {
          var errSummary = this._LOGGER_MSG._ERR_ELEMENT_INVALID_TYPE_SUMMARY;
          var errDetail = this._LOGGER_MSG._ERR_ELEMENT_INVALID_TYPE_DETAIL;
          throw new RangeError(errSummary + '\n' + errDetail);
        }

        // add css class to element
        this.element[0].classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_ELEMENT_CLASS);

        // create the initial table structure
        this._getTableDomUtils().createInitialTable(this._isTableHeaderless(),
                                                    this._isTableFooterless());
        // style the initial table structure
        this._getTableDomUtils().styleInitialTable();

        var self = this;
        this._queueTask(function()
        {
          // populate the table header DOM with header content
          self._refreshTableHeader();
          // populate the table footer DOM with footer content
          self._refreshTableFooter();
          self._refreshTableBody();
          self._processSlottedChildren();
          self._getTableDomUtils().refreshTableDimensions();
        });

        if (this.options.disabled)
        {
          this.disable();
        }
      },
      /**
       * @override
       * @private
       */
      _events:
        {
          /*
           * Reset the keyboard state on focusout and set the inactive
           * selected rows
           */
          'focusout': function(event)
          {
            // make sure the focusout isn't for a focus to an element within
            // the table
            var table = this._getTableDomUtils().getTable();
            var focusElement = null;
            
            if (event.relatedTarget != null)
            {
              // In Chrome we can check relatedTarget
              focusElement = event.relatedTarget;
            }
            else if (event.originalEvent != null && 
                     event.originalEvent.explicitOriginalTarget != null)
            {
              // In FF we check explicitOriginalTarget
              focusElement = event.originalEvent.explicitOriginalTarget;
            }
            else if (this._getTableDomUtils()._isIE() && 
                     document.activeElement != null)
            {
              // In IE we check document.activeElement
              focusElement = document.activeElement;
            }
            
            if (focusElement != null &&
                (focusElement == table || 
                 $(table).has(focusElement).length > 0))
            {
              return;
            }

            // Components that open popups (such as ojSelect, ojCombobox, 
            // ojInputDate, etc.) will trigger focusout, but we don't want to
            // change mode in those cases since the user is still editing.
            var popups = oj.ZOrderUtils.findOpenPopups();
            for (var i = 0; i < popups.length; i++)
            {
              // Get the launcher of the popup.
              // popups[i] is just a wrapper with the real popup as its child.
              var launcher = oj.DomUtils.getLogicalParent($(popups[i]).children());
              
              // Check if the table contains the launcher
              if ($(table).has(launcher.get(0)).length > 0)
              {
                return;
              }
            }

            this._focusOutHandler($(table));
            this._focusOutHandler($(this._getTableDomUtils().getTableContainer()));
            this._clearKeyboardKeys();
            this._clearFocusedHeaderColumn();
            this._clearFocusedRow(false);
            this._getTableResizeUtils().clearTableHeaderColumnsResize();
            this._setTableEditable(false, false, 0, true, event);
            this._setTableActionableMode(false);
          },
          /*
           * Remove the cell edit class on cell focus when row is editable.
           */
          'blur .oj-table-data-cell': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            if (eventTarget.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_EDIT_CLASS))
            {
              eventTarget.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_EDIT_CLASS);
            }
          },
          /*
           * Check the keyboard state on focus
           */
          'focus': function(event)
          {
            this._checkRowOrHeaderColumnFocus(event);
          },
          /*
           * Handle focus on child row elements
           */
          'focusin': function(event)
          {
            this._focusInHandler($(this._getTableDomUtils().getTable()));
            this._focusInHandler($(this._getTableDomUtils().getTableContainer()));
            var tableBody = this._getTableDomUtils().getTableBody();
            if ($(tableBody).has(event.target).length > 0)
            {
              var focusedRowIdx = this._getFocusedRowIdx();
              var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();
 
              if (!this._isNodeEditable(event.target) &&
                  !this._isNodeClickable(event.target) &&
                  !this._isTableActionableMode() &&
                  !this._hasEditableRow() &&
                  !this._temporaryTableChildElementFocus &&
                  focusedRowIdx == null && 
                  focusedHeaderColumnIdx == null)
              {
                // when table is not in editable/actionable mode there shouldn't be
                // focus to child row elements. Delay the focus to prevent a race
                // condition with menu launcher re-focus
                var self = this;
                setTimeout(function () 
                {
                  self._getTableDomUtils().getTable().focus();
                }, 0);
              }
            }
          },
          /*
           * Check the keyboard state on focus
           */
          'focus .oj-table-column-header-acc-asc-link': function(event)
          {
            var self = this;
            setTimeout(function()
            {
              // delay the column/row focus just in case a column/row is clicked.
              self._checkRowOrHeaderColumnFocus(event);
              self = null;
            }, 0);
          },
          /*
           * Set the cell edit class on cell focus when row is editable.
           */
          'focus .oj-table-data-cell': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            var rowIdx = this._getTableDomUtils().getElementRowIdx(eventTarget);
            
            if (rowIdx == this._getEditableRowIdx())
            {
              eventTarget.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_EDIT_CLASS);
            }
          },
          /*
           * Capture acc selected column event
           */
          'click .oj-table-checkbox-acc-select-column': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(eventTarget);
            var selected = $(eventTarget).is(':checked');
            // if selected then focus on the column
            if (selected)
            {
              this._setHeaderColumnFocus(columnIdx, true, true, null);
            }
            this._setHeaderColumnSelection(columnIdx, selected, eventTarget, event, true);
            event.stopPropagation();
          },
          /*
           * Capture acc selected row event
           */
          'click .oj-table-checkbox-acc-select-row': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            var rowIdx = this._getTableDomUtils().getElementRowIdx(eventTarget);
            var selected = $(eventTarget).is(':checked');

            var focused = false;

            // if selected then focus on the row
            if (selected)
            {
              focused = this._setRowFocus(rowIdx, true, true, null, event, true);
            }
            else
            {
              this._setTableEditable(false, false, 0, true, event);
            }

            if (focused)
            {
              this._setRowSelection(rowIdx, selected, eventTarget, event, true);
              this._setTableEditable(true, false, 0, true, event);
            }
            event.stopPropagation();
          },
          /*
           * Capture keyboard down events
           */
          'keydown': function(event)
          {
            this._addKeyboardKey(event.keyCode);
            
            // ignore key event on the footer or target is editable
            var keyboardCode1 = this._getKeyboardKeys()[0];
            
            if (keyboardCode1 != this._KEYBOARD_CODES._KEYBOARD_CODE_ESC && 
                keyboardCode1 != this._KEYBOARD_CODES._KEYBOARD_CODE_ENTER &&
                keyboardCode1 != this._KEYBOARD_CODES._KEYBOARD_CODE_F2 &&
                keyboardCode1 != this._KEYBOARD_CODES._KEYBOARD_CODE_TAB &&
                (this._isNodeEditable(event.target) ||
                this._getTableDomUtils().getTableFooter() != null &&
                $(this._getTableDomUtils().getTableFooter()).has(event.target).length > 0))
            {
              return;
            }

            // process single or two key events
            if (this._getKeyboardKeys().length == 1 ||
                (this._getKeyboardKeys().length == 2 && event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT]))
            {
              if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_UP) ||
                this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_DOWN) ||
                this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_SPACEBAR) ||
                this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_HOME) ||
                this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_END))
              {
                // need to do this so that these keys don't act on the page. e.g. pressing Down would cause the
                // page to go down as well as the row to change
                event.preventDefault();
                event.stopPropagation();
              }

              if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_UP) ||
                this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_DOWN))
              {
                this._handleKeydownUpDown(event);
              }
              else if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_LEFT) ||
                this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_RIGHT))
              {
                this._handleKeydownLeftRight(event);
              }
              else if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_TAB))
              {
                this._handleKeydownTab(event);
              }
              else if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_F2)
              {
                this._handleKeydownF2(event);
              } 
              else if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_SPACEBAR)
              {
                this._handleKeydownSpacebar(event);
              }
              else if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_ENTER)
              {
                this._handleKeydownEnter(event);
              }
              else if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_HOME)
              {
                this._handleKeydownHome(event);
              }
              else if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_END)
              {
                this._handleKeydownEnd(event);
              }
              else if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_ESC)
              {
                this._handleKeydownEsc(event);
              }
            }
          },
          /*
           * Capture keyboard up events
           */
          'keyup': function(event)
          {
             // process single or 2 key events
            if (this._getKeyboardKeys().length == 1)
            {
              var keyboardCode1 = this._getKeyboardKeys()[0];
              
              // ignore key event on the footer or target is editable
              if (this._isNodeEditable(event.target) ||
                  this._getTableDomUtils().getTableFooter() != null &&
                  $(this._getTableDomUtils().getTableFooter()).has(event.target).length > 0)
              {
                this._removeKeyboardKey(keyboardCode1);
                return;
              }
              this._removeKeyboardKey(keyboardCode1);
            }
            // remove the keycode from our internal list of pressed keys.
            this._removeKeyboardKey(event.keyCode);
          },
          /*
           * Keep track of mousedown/mouseup for multiple selection
           */
          'mousedown .oj-table-body': function(event)
          {
            // get the row index if the mousedown was on a row
            this._mouseDownRowIdx = this._getTableDomUtils().getElementRowIdx(event.target);
            
            if (this._mouseDownRowIdx == null)
            {
              return;
            }
            
            var tableBodyRow = this._getTableDomUtils().getTableBodyRow(this._mouseDownRowIdx);
            
            if (tableBodyRow != null && tableBodyRow['draggable'])
            {
              // do not do row selection if we are dragging
              this._mouseDownRowIdx = null;
              return;
            }
            
            // Only clear if Shift or Ctrl are not pressed since this
            // could be multiple selection
            // Also only clear for left click
            if (this._mouseDownRowIdx != null &&
                !event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT] && 
                !oj.DomUtils.isMetaKeyPressed(event) &&
                event.which === 1)
            {
              var rowSelected = this._getRowSelection(this._mouseDownRowIdx);
              
              // check if the row which we clicked on is already selected
              // if it is and it's the only selected row, then don't clear
              if (rowSelected &&
                  this._getSelectedRowIdxs().length == 1)
              {
                return;
              }
              
              // only clear if non-contiguous selection is not enabled for touch
              if (!this._nonContiguousSelection)
              {
                this._clearSelectedRows(true);
              }
            }
          },
          /*
           * Keep track of mousedown/mouseup for multiple selection
           */
          'mouseup .oj-table-body': function(event)
          {
            this._mouseDownRowIdx = null;
          },
          /*
           * show the row hover when the mouse enters a table row
           */
          'mouseenter .oj-table-body-row': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            eventTarget.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
            var rowIdx = this._getTableDomUtils().getElementRowIdx(eventTarget);
            this._updateRowStateCellsClass(rowIdx, {hover: true});
            this._handleMouseEnterSelection(event.target); 
          },
          /*
           * hide the row hover when the mouse leaves a table row
           */
          'mouseleave .oj-table-body-row': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            eventTarget.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
            var rowIdx = this._getTableDomUtils().getElementRowIdx(eventTarget);
            this._updateRowStateCellsClass(rowIdx, {hover: false});
          },
          /*
           * set the column header focus.
           */
          'mousedown .oj-table-column-header-cell': function(event)
          {
            if (event.which === 1)
            {
              if (!this._getTableResizeUtils().handleHeaderColumnResizeStart(event))
              {
                // get the column index
                var columnIdx = this._getTableDomUtils().getElementColumnIdx(this._getEventTargetElement(event));
                // set the column focus
                this._setHeaderColumnFocus(columnIdx, true, true, event);
                $(event.target).data(this._FOCUS_CALLED, true);
              }
            }
          },
          /*
           * show the ascending/descending links when the mouse
           * enters a column header
           */
          'mouseenter .oj-table-column-header-cell': function(event)
          {
            if (this._getTableResizeUtils().setResizeCursor(event))
            {
              return;
            }
            var eventTarget = this._getEventTargetElement(event);
            eventTarget.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
            // get the column index of the header element
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(eventTarget);
            // show the asc/dsc links for the header
            this._showTableHeaderColumnSortLink(columnIdx);
          },
          /*
           * show the resize cursor
           */
          'mousemove .oj-table-header': function(event)
          {
            this._getTableResizeUtils().setResizeCursor(event);
          },
          /*
           * remove the hover for resize
           */
          'mousemove .oj-table-column-header-cell': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            if (this._getTableResizeUtils().setResizeCursor(event))
            {
              eventTarget.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
              return;
            }
            eventTarget.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
          },
          /*
           * hide the ascending/descending links when the mouse
           * leaves a column header
           */
          'mouseleave .oj-table-column-header-cell': function(event)
          {
            var eventTarget = this._getEventTargetElement(event);
            eventTarget.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
            // get the column index of the header element
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(eventTarget);
            // hide the asc/dsc links for the header
            this._hideTableHeaderColumnSortLink(columnIdx, true);
            this._hideTableHeaderColumnSortLink(columnIdx, false);
          },
          /*
           * handle column resizing.
           */
          'mouseup .oj-table-column-header-cell': function(event)
          {
            this._getTableResizeUtils().handleHeaderColumnResizeEnd(event);
          },
          /*
           * set the row focus when the mouse clicks on a cell.
           */
          'mousedown .oj-table-data-cell': function(event)
          {
            // get the row index of the cell element
            var eventTarget = this._getEventTargetElement(event);
            var rowIdx = this._getTableDomUtils().getElementRowIdx(eventTarget);
            this._setRowFocus(rowIdx, true, true, eventTarget, event);
            $(event.target).data(this._FOCUS_CALLED, true);
          },
          /*
           * handle column resizing.
           */
          'mouseup .oj-table-data-cell': function(event)
          {
            this._getTableResizeUtils().handleHeaderColumnResizeEnd(event);
          },
          /*
           * invoke a sort on the column data when the mouse clicks the ascending link
           */
          'click .oj-table-column-header-asc-link': function(event)
          {
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(event.target);
            var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);

            if (!tableHeaderColumn)
            {
              return;
            }

            // check if the column is currently sorted
            var sorted = $(tableHeaderColumn).data('sorted');
            if (sorted == this._COLUMN_SORT_ORDER._ASCENDING)
            {
              this._handleSortTableHeaderColumn(columnIdx, false, event);
            }
            else
            {
              this._handleSortTableHeaderColumn(columnIdx, true, event);
            }
            event.preventDefault();
            event.stopPropagation();
          },
          'click .oj-table-column-header-acc-asc-link': function(event)
          {
            if (this._getKeyboardKeys().length > 0)
            {
              var keyboardCode1 = this._getKeyboardKeys()[0];
              if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_ENTER)
              {
                // An Enter key press can cause a click event
                event.preventDefault();
                event.stopPropagation();
                return;
              }
            }
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(event.target);
            var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);

            if (!tableHeaderColumn)
            {
              return;
            }
            var sorted = $(tableHeaderColumn).data('sorted');
            if (sorted == this._COLUMN_SORT_ORDER._DESCENDING)
            {
              this._handleSortTableHeaderColumn(columnIdx, true, event);
            }
            else
            {
              this._handleSortTableHeaderColumn(columnIdx, false, event);
            }
            event.preventDefault();
            event.stopPropagation();
          },
          /*
           * invoke a sort on the column data when the mouse clicks the descending link
           */
          'click .oj-table-column-header-dsc-link': function(event)
          {
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(event.target);
            var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);

            if (!tableHeaderColumn)
            {
              return;
            }

            // check if the column is currently sorted
            var sorted = $(tableHeaderColumn).data('sorted');
            if (sorted == this._COLUMN_SORT_ORDER._DESCENDING)
            {
              this._handleSortTableHeaderColumn(columnIdx, true, event);
            }
            else
            {
              this._handleSortTableHeaderColumn(columnIdx, false, event);
            }
            event.preventDefault();
            event.stopPropagation();
          },
          'click .oj-table-column-header-acc-dsc-link': function(event)
          {
            if (this._getKeyboardKeys().length > 0)
            {
              var keyboardCode1 = this._getKeyboardKeys()[0];
              if (keyboardCode1 == this._KEYBOARD_CODES._KEYBOARD_CODE_ENTER)
              {
                // An Enter key press can cause a click event
                event.preventDefault();
                event.stopPropagation();
                return;
              }
            }
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(event.target);
            var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);

            if (!tableHeaderColumn)
            {
              return;
            }
            var sorted = $(tableHeaderColumn).data('sorted');
            if (sorted == this._COLUMN_SORT_ORDER._DESCENDING)
            {
              this._handleSortTableHeaderColumn(columnIdx, true, event);
            }
            else
            {
              this._handleSortTableHeaderColumn(columnIdx, false, event);
            }
            event.preventDefault();
            event.stopPropagation();
          },
          /*
           * set the row focus or selection when the mouse clicks on a cell.
           * Ctrl + click results in selection and focus. Plain click results in focus.
           * Plain click on a selected row removes the selection.
           */
          'click .oj-table-data-cell': function(event)
          {
            // get the row index of the cell element
            var eventTarget = this._getEventTargetElement(event);
            var rowIdx = this._getTableDomUtils().getElementRowIdx(eventTarget);
            var focusCalled = $(event.target).data(this._FOCUS_CALLED);
 
            if (!focusCalled)
            {
              var focused = this._setRowFocus(rowIdx, true, true, eventTarget, event);
              $(event.target).data(this._FOCUS_CALLED, false);
              
              if (!focused)
              {
                return;
              }
            }
            
            // check if we are selecting
            if (event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              var lastSelectedRowIdx = this._getLastRowSelection();
              if (lastSelectedRowIdx != null)
              {
                // remove the selection highlight
                window.getSelection().removeAllRanges()

                // shift selection is always from the last selected row
                if (rowIdx < lastSelectedRowIdx)
                {
                  var i;
                  for (i = rowIdx; i <= lastSelectedRowIdx; i++)
                  {
                    this._setRowSelection(i, true, eventTarget, event, true);
                  }
                }
                else
                {
                  var i;
                  for (i = lastSelectedRowIdx; i <= rowIdx; i++)
                    this._setRowSelection(i, true, eventTarget, event, true);
                }
              }
            }
            else if (oj.DomUtils.isMetaKeyPressed(event))
            {
              this._setRowSelection(rowIdx, !this._getRowSelection(rowIdx), eventTarget, event, true);
            }
            else if (this._getKeyboardKeys().length == 0)
            {
              var rowSelected = this._getRowSelection(rowIdx);
              this._setRowSelection(rowIdx, !rowSelected, eventTarget, event, true);
              
              if (this._isTouchDevice() && 
                  this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._MULTIPLE &&
                  !rowSelected)
              {
                this._getTableDomUtils().createTableBodyRowTouchSelectionAffordance(rowIdx);
              }
            }
          },
          /*
           * Set row to editable.
           */
          'dblclick .oj-table-data-cell': function(event)
          {
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(event.target);
            this._setTableEditable(true, false, columnIdx, true, event);
          },
          /*
           * set current row when the mouse right clicks on a cell.
           */
          'contextmenu .oj-table-data-cell': function(event)
          {
            // get the row index of the cell element
            var rowIdx = this._getTableDomUtils().getElementRowIdx(this._getEventTargetElement(event));
            var rowKey = this._getRowKeyForRowIdx(rowIdx);
            this._setCurrentRow({'rowKey': rowKey}, event, false);
          },
          /*
           * set the column header selection and focus. Plain click results in
           * focus and selection. If Ctrl is not pressed then we have single column selection.
           */
          'click .oj-table-column-header-cell': function(event)
          {
            // get the column index
            var eventTarget = this._getEventTargetElement(event);
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(eventTarget);
            // check if we need to focus
            var focusCalled = $(event.target).data(this._FOCUS_CALLED);
 
            if (!focusCalled)
            {
              // set the column focus
              this._setHeaderColumnFocus(columnIdx, true, true, event);
              $(event.target).data(this._FOCUS_CALLED, false);
            }
            
            // check if we are selecting
            if (event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              var lastSelectedColumnIdx = this._getLastHeaderColumnSelection();
              if (lastSelectedColumnIdx != null)
              {
                // shift selection is always from the last selected column
                if (columnIdx < lastSelectedColumnIdx)
                {
                  var i;
                  for (i = columnIdx; i <= lastSelectedColumnIdx; i++)
                  {
                    this._setHeaderColumnSelection(i, true, eventTarget, event, true);
                  }
                }
                else
                {
                  var i;
                  for (i = lastSelectedColumnIdx; i <= columnIdx; i++)
                    this._setHeaderColumnSelection(i, true, eventTarget, event, true);
                }
              }
            }
            else if (oj.DomUtils.isMetaKeyPressed(event))
            {
              this._setHeaderColumnSelection(columnIdx, !this._getHeaderColumnSelection(columnIdx), eventTarget, event, true);
            }
            else if (this._getKeyboardKeys().length == 0)
            {
              var selectedColumnIdxs = this._getSelectedHeaderColumnIdxs();
              
              if (selectedColumnIdxs != null &&
                selectedColumnIdxs.length > 0) 
              {
                if (selectedColumnIdxs.indexOf(columnIdx) == -1 ||
                  selectedColumnIdxs.length > 1)
                {
                  this._clearSelectedHeaderColumns();
                }
              }
              this._setHeaderColumnSelection(columnIdx, !this._getHeaderColumnSelection(columnIdx), eventTarget, event, true);
              this._getTableDndContext().setTableHeaderColumnDraggable(columnIdx, true);
            }
          },
          /*
           * Set dragstart handler for column DnD.
           */
          'dragstart .oj-table-column-header-cell': function(event)
          {
            return this._getTableDndContext().handleColumnDragStart(event);
          },
          /*
           * Set dragenter handler for column DnD.
           */
          'dragenter .oj-table-column-header-cell': function(event)
          {
            return this._getTableDndContext().handleColumnDragEnter(event);
          },
          /*
           * Set dragover handler for column DnD.
           */
          'dragover .oj-table-column-header-cell': function(event)
          {
            return this._getTableDndContext().handleColumnDragOver(event);
          },
          /*
           * Set dragleave handler for column DnD.
           */
          'dragleave .oj-table-column-header-cell': function(event)
          {
            return this._getTableDndContext().handleColumnDragLeave(event);
          },
          /*
           * Set drop handler for column DnD.
           */
          'drop .oj-table-column-header-cell': function(event)
          {
            return this._getTableDndContext().handleColumnDrop(event);
          },
          /*
           * Set dragend handler for column DnD.
           */
          'dragend .oj-table-column-header-cell': function(event)
          {
            return this._getTableDndContext().handleColumnDragEnd(event);
          },
          /*
           * handle the dragstart event on rows and invoke event callback.
           */
          'dragstart .oj-table-body-row': function(event)
          {
            return this._getTableDndContext().handleRowDragStart(event);
          },
          /*
           * handle the drag event on rows and invoke event callback.
           */
          'drag .oj-table-body-row': function(event)
          {
            return this._getTableDndContext().handleRowDrag(event);
          },
          /*
           * handle the dragend event on rows and invoke event callback.
           */
          'dragend .oj-table-body-row': function(event)
          {
            return this._getTableDndContext().handleRowDragEnd(event);
          },
          /*
           * handle the dragenter event and invoke event callback.
           */
          'dragenter .oj-table-body': function(event)
          {
            return this._getTableDndContext().handleRowDragEnter(event);
          },
          /*
           * handle the dragover event and invoke event callback.
           */
          'dragover .oj-table-body': function(event)
          {
            return this._getTableDndContext().handleRowDragOver(event);
          },
          /*
           * handle the dragleave event and invoke event callback.
           */
          'dragleave .oj-table-body': function(event)
          {
            return this._getTableDndContext().handleRowDragLeave(event);
          },
          /*
           * handle the drop event and invoke event callback.
           */
          'drop .oj-table-body': function(event)
          {
            return this._getTableDndContext().handleRowDrop(event);
          }
        },
      /**
       * @override
       * @private
       */
      _eventsContainer:
      {
          /*
           * Keep track of touchstart on selection affordance
           */
          'touchstart .oj-table-body-row-touch-selection-affordance-touch-area': function(event)
          {
            var fingerCount = event.originalEvent.touches.length;
    
            if (fingerCount == 1 && this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._MULTIPLE)
            {
              if (event.target.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOP_CLASS) ||
                  event.target.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOP_ICON_CLASS))
              {
                event.preventDefault();
                this._mouseDownRowIdx = $(this._getTableDomUtils().getTableBodyRowTouchSelectionAffordanceBottom()).data('rowIdx');
              }
              else if (event.target.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_BOTTOM_CLASS) ||
                       event.target.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_BOTTOM_ICON_CLASS))
              {
                event.preventDefault();
                this._mouseDownRowIdx = $(this._getTableDomUtils().getTableBodyRowTouchSelectionAffordanceTop()).data('rowIdx');
              }
            }
          },
          /*
           * column resizing
           */
          'touchstart .oj-table-column-header-cell': function(event)
          {
            var fingerCount = event.originalEvent.touches.length;
            
            if (fingerCount == 1)
            {
              if (this._isTableColumnsResizable() &&
                this._getTableResizeUtils().handleHeaderColumnResizeStart(event))
              {
                event.preventDefault();
              }
            }
          },
          /*
           * Keep track of touchmove for column resize
           */
          'touchmove .oj-table-header': function(event)
          {
            if (this._isTableColumnsResizable())
            {
              this._getTableResizeUtils().setResizeCursor(event);
            }
          },
          /*
           * Keep track of touchmove for multiple selection
           */
          'touchmove .oj-table-body-row-touch-selection-affordance-touch-area': function(event)
          {
            if (this._mouseDownRowIdx != null)
            {
              event.preventDefault();
              var eventTarget = this._getEventTargetElement(event);
              this._handleMouseEnterSelection(eventTarget); 
            }
          },
          /*
           * Keep track of touchend for multiple selection
           */
          'touchend': function(event)
          {
            if (this._mouseDownRowIdx != null)
            {
              var eventTarget = this._getEventTargetElement(event);
              this._handleMouseEnterSelection(eventTarget); 
            }
            this._mouseDownRowIdx = null;
            if (this._getTableResizeUtils().handleHeaderColumnResizeEnd(event))
            {
              event.preventDefault();
            }
          },
          /*
           * Keep track of touchend for edit
           */
          'touchend .oj-table-body': function(event)
          {
            var self = this;
            this._touchEventDoubleTapFunction(event, function(event) {
              return function()
              {
                var eventTarget = self._getEventTargetElement(event);
                var columnIdx = self._getTableDomUtils().getElementColumnIdx(eventTarget);
                self._setTableEditable(true, false, columnIdx, true, event);
                event.preventDefault();
              };
            }(event));
          },
          /*
           * Keep track of touchcancel for multiple selection
           */
          'touchcancel': function(event)
          {
            this._mouseDownRowIdx = null;
            this._getTableResizeUtils().clearTableHeaderColumnsResize();
          }
      },
      /**
       * @private
       */
      _refresh: function()
      {
        var startIndex = null;
        var initFetch = false;

        if (this._dataOption != this.options[this._CONST_DATA])
        {
          this._clearCachedDataMetadata();

          if (this._data == null)
          {
            // need to do an initial fetch
            initFetch = true
          }
          else
          {
            startIndex = 0;
          }
        }
        var contextMenu = this._GetContextMenu();
        if (contextMenu != null && contextMenu != this._getTableDomUtils().getContextMenu())
        {
          this._getTableDomUtils().createContextMenu(this._handleContextMenuSelect.bind(this));
        }
        this._getTableDomUtils().clearCachedDom();
        this._getTableDomUtils().refreshContextMenu();
        this._refreshTableStatusMessage();

        if (initFetch)
        {
          return this._initFetch();
        }
        else
        {
          var self = this;
          return this._queueTask(function()
          {
            var result = self._invokeDataFetchRows(null, false);
            startIndex = null;
            self = null;
            
            return result;
          });
        }
      },
      /**
       * @override
       * @private
       */
      _setOption: function(key, value)
      {
        if (key == 'selection')
        {
          this._clearSelectedRows();
          this._clearSelectedHeaderColumns();
          this._setSelection(value);
          this._setSelectionDefault();
          this._superApply(arguments);
        }
        else if (key == 'currentRow')
        {
          var updated = this._setCurrentRow(value, null, true);

          if (updated)
          {
            this._superApply(arguments);
          }
        }
        else
        {
          this._superApply(arguments);
        var shouldRefresh = this._isTableRefreshNeeded(key, value);

        if (shouldRefresh)
        {
          if (key == 'columns')
          {
            this._clearCachedMetadata();
            this._refreshTableHeader();
          }
          this._refresh();
        }
          }
      },
      /**** end internal widget functions ****/

      /**** start internal functions ****/

      /**
       * Add a keyCode to internally track pressed keys. keyCodes should be added on
       * mouse down and then later removed on mouse up.
       * @param {number} keyCode  KeyCode of the keyboard key.
       * @private
       */
      _addKeyboardKey: function(keyCode)
      {
        var foundCode = false;
        for (var prop in this._KEYBOARD_CODES)
        {
          if (this._KEYBOARD_CODES.hasOwnProperty(prop))
          {
            if (this._KEYBOARD_CODES[prop] == keyCode)
            {
              foundCode = true;
            }
          }
        }

        if (!foundCode)
        {
          // only add keys we are interested in
          return;
        }

        var keyboardKeys = this._getKeyboardKeys();
        var found = false;
        var i, keyboardKeysCount = keyboardKeys.length;
        for (i = 0; i < keyboardKeysCount; i++)
        {
          if (keyboardKeys[i] == keyCode)
          {
            found = true;
            break;
          }
        }
        if (!found)
        {
          keyboardKeys.push(keyCode);
        }
      },
      /**
	     * Called by component to add a busy state and return the resolve function
       * to call when the busy state can be removed.
	     * @private
       */
      _addComponentBusyState: function(msg) 
      {
        var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
        var options = {'description' : "The component identified by '" + this._getTableDomUtils().getTableId() + "' " + msg};
        return busyContext.addBusyState(options);
      },
      /**
       * Add a new tr and refresh the DOM at the row index and refresh the table
       * dimensions to accomodate the new row
       * @param {number} rowIdx  row index relative to the start of the table
       * @param {Object} row row
       * @param {Object} docFrag  document fragment
       * @param {number} docFragStartIdx  document fragment row start index
       * @return {Element} Returns the added tr
       *
       * @private
       */
      _addSingleTableBodyRow: function(rowIdx, row, docFrag, docFragStartIdx)
      {
        var tableBodyRow = this._getTableDomUtils().createTableBodyRow(rowIdx, row[this._CONST_KEY]);
        this._getTableDomUtils().styleTableBodyRow(tableBodyRow, true);
        // insert the <tr> element in to the table body DOM
        this._getTableDomUtils().insertTableBodyRow(rowIdx, tableBodyRow, row, docFrag);
        tableBodyRow = this._refreshTableBodyRow(rowIdx, row, tableBodyRow, docFrag, docFragStartIdx, true);
        if (!docFrag) {
          // call subtreeAttached on individual row if we are not batching in documentFragment
          oj.Components.subtreeAttached(tableBodyRow);
        }
        return tableBodyRow;
      },
      /**
       * Animate only the visibles in the provided row array
       * @param tableBodyRows Array of tr DOM elements
       * @param rowIdxArray Array of row indexes for the tr elements
       * @param action Animation action
       * @return Promise Return a Promise which resolves when the animation is complete
       * @private
       */
      _animateVisibleRows: function(tableBodyRows, rowIdxArray, action)
      {
        action = this._animationActionOverride == null ? action : this._animationActionOverride;
            
        if (!this._hasPendingTasks())
        {
          this._animationActionOverride = null;
          var visibleRowIdxArray = this._getVisibleRowIdxs();
          var nonVisibleRowIdxArray = [];
          var i;
          for (i = 0; i < rowIdxArray.length; i++)
          {
            if (visibleRowIdxArray.indexOf(rowIdxArray[i]) == -1)
            {
              nonVisibleRowIdxArray.push(rowIdxArray[i]);
            }
          }
          nonVisibleRowIdxArray.sort(function (a, b)
            {  
              return a - b;  
            });
          for (i = nonVisibleRowIdxArray.length - 1; i >= 0; i--)
          {
            tableBodyRows.splice(nonVisibleRowIdxArray[i], 1);
          }
          // remove resize listeners to prevent triggering resize notifications
          this._unregisterResizeListener();
          var self = this;
          return this._getTableAnimationUtils().animateTableBodyRows(tableBodyRows, action).then(function()
          {
            self._registerResizeListener();
          });
        }
        else
        {
          this._animationActionOverride = 'update';
          return Promise.resolve();
        }
      },
      /**
       * Check and set the row or header column focus
       * @private
       */
      _checkRowOrHeaderColumnFocus: function(event)
      {
        var focusedRowIdx = this._getFocusedRowIdx();
        var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();

        if (focusedRowIdx == null && focusedHeaderColumnIdx == null)
        {
          var focusRowIdx = null;
          var currentRow = this._getCurrentRow();
          var currentRowKey = currentRow != null ? currentRow['rowKey'] : null;
          if (currentRowKey != null)
          {
            focusRowIdx = this._getRowIdxForRowKey(currentRowKey);
          }

          // if no row or column is focused 
          // and currentRow is null then set the focus on the first column or row
          if (focusRowIdx == null)
          {
            if (this._isTableHeaderless())
            {
              this._setRowFocus(0, true, true, null, event);
            }
            else
            {
              this._setHeaderColumnFocus(0, true, false, event);
            }
          }
          else
          {
            this._setRowFocus(focusRowIdx, true, true, null, event);
          }
        }
      },
      /**
       * Clear any cached metadata
       * @private
       */
      _clearCachedMetadata: function()
      {
        this._columnDefArray = null;
        this._setTableActionableMode(false);
      },
      /**
       * Clear any cached data metadata
       * @private
       */
      _clearCachedDataMetadata: function()
      {
        if (this._data != null)
        {
          this._unregisterDataSourceEventListeners();
        }
        this._data = null;
      },
      /**
       * Clear waiting state and hide the Fetching Data... status message.
       * @private
       */
      _clearDataWaitingState: function()
      {
        this._hideInlineMessage();
        this._hideStatusMessage();
        this._dataFetching = false;
        if (this._dataResolveFunc)
        {
          this._dataResolveFunc();
          this._dataResolveFunc = null;
        }
      },
      /**
       * Clear any keyboard keys
       * @private
       */
      _clearKeyboardKeys: function()
      {
        this._keyboardKeys = [];
      },
      /**
       * Clear the focused column header
       * @private
       */
      _clearFocusedHeaderColumn: function()
      {
        var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();
        if (focusedHeaderColumnIdx != null)
        {
          this._setHeaderColumnFocus(focusedHeaderColumnIdx, false, false, null);
        }
        this._activeColumnIndex = -1;
      },
      /**
       * Clear the focused row
       * @param {boolean} updateCurrentRow  whether to update the currentRow
       * @private
       */
      _clearFocusedRow: function(updateCurrentRow)
      {
        var focusedRowIdx = this._getFocusedRowIdx();

        if (focusedRowIdx != null)
        {
          this._setRowFocus(-1, true, updateCurrentRow, null, null);
        }
      },
      /**
       * Clear the selected column headers
       * @private
       */
      _clearSelectedHeaderColumns: function()
      {
        var selectedHeaderColumnIdxs = this._getSelectedHeaderColumnIdxs();

        var i, selectedHeaderColumnIdxsCount = selectedHeaderColumnIdxs.length;
        for (i = 0; i < selectedHeaderColumnIdxsCount; i++)
        {
          this._setHeaderColumnSelection(selectedHeaderColumnIdxs[i], false, null, null, false, true);
        }
      },
      /**
       * Clear the selected rows
       * @private
       */
      _clearSelectedRows: function()
      {
        var selectedRowIdxs = this._getSelectedRowIdxs();

        var i, selectedRowIdxsCount = selectedRowIdxs.length;
        for (i = 0; i < selectedRowIdxsCount; i++)
        {
          this._setRowSelection(selectedRowIdxs[i], false, null, null, false, true);
        }
        if (this._isTouchDevice() && this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._MULTIPLE)
        {
          this._getTableDomUtils().removeTableBodyRowTouchSelectionAffordance();
        }
      },
      /**
       * Clear the sorted column header indicator. Note this does not affect the order
       * of the data. This is just to clear the UI indication.
       * @param {number|null} columnIdx  column index
       * @private
       */
      _clearSortedHeaderColumn: function(columnIdx)
      {
        var sortedTableHeaderColumnIdx = this._getSortedTableHeaderColumnIdx();
        if (sortedTableHeaderColumnIdx != null)
        {
          var sortedTableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(sortedTableHeaderColumnIdx);
          var sorted = $(sortedTableHeaderColumn).data('sorted');
          $(sortedTableHeaderColumn).data('sorted', null);

          if (columnIdx == null || 
            sortedTableHeaderColumnIdx !== columnIdx)
          {
            this._hideTableHeaderColumnSortLink(sortedTableHeaderColumnIdx, sorted == this._COLUMN_SORT_ORDER._ASCENDING);
          }
          else
          {
            var sortedTableHeaderColumnAscLink = sortedTableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_LINK_CLASS);
            if (sortedTableHeaderColumnAscLink.length > 0)
            {
              sortedTableHeaderColumnAscLink = sortedTableHeaderColumnAscLink[0];
              sortedTableHeaderColumnAscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
            }
            var sortedTableHeaderColumnDscLink = sortedTableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_LINK_CLASS);
            if (sortedTableHeaderColumnDscLink.length > 0)
            {
              sortedTableHeaderColumnDscLink = sortedTableHeaderColumnDscLink[0];
              sortedTableHeaderColumnDscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
            }
          }
        }
      },
      /**
       * Add all the rows contained in the input array.
       * @param {Array} rows Array of row contexts to add
       * @private
       */
      _executeTableBodyRowsAdd: function(rows)
      {
        var self = this;
        this._queueTask(function()
        {
          // see if we should batch add
          // only batch if we are adding a block of contiguous rows
          var batchAdd = false;
          var addedTableBodyRows = [];
          var rowIdxArray = [];
          if (rows.length > 1)
          {
            var i, rowsCount = rows.length;
            var isContiguous = true;
            for (i = 0; i < rowsCount; i++)
            {
              if (i != 0)
              {
                if (rows[i - 1].rowIdx != rows[i].rowIdx - 1)
                {
                  isContiguous = false;
                }
              }
              rowIdxArray.push(rows[i].rowIdx);
            }

            if (isContiguous)
            {
              var tableBody = self._getTableDomUtils().getTableBody();
              var tableBodyDocFrag = document.createDocumentFragment();
              rowsCount = rows.length;
              for (i = 0; i < rowsCount; i++)
              {
                addedTableBodyRows.push(self._addSingleTableBodyRow(rows[i].rowIdx, rows[i].row, tableBodyDocFrag, rows[0].rowIdx));
              }
              if (rows[0].rowIdx == 0)
              {
                tableBody.insertBefore(tableBodyDocFrag, tableBody.firstChild); //@HTMLUpdateOK
              }
              else
              {
                var tableBodyRowBefore = self._getTableDomUtils().getTableBodyRow(rows[0].rowIdx);
                if (tableBodyRowBefore != null)
                {
                  tableBody.insertBefore(tableBodyDocFrag, tableBodyRowBefore);//@HTMLUpdateOK
                }
                else
                {
                  tableBody.appendChild(tableBodyDocFrag, null);//@HTMLUpdateOK
                }
              }
              self._getTableDomUtils().clearCachedDomRowData();
              oj.Components.subtreeAttached(tableBody);
              batchAdd = true;
            }
          }

          if (!batchAdd)
          {
            rowsCount = rows.length;
            for (i = 0; i < rowsCount; i++)
            {
              addedTableBodyRows.push(self._addSingleTableBodyRow(rows[i].rowIdx, rows[i].row));
            }
          }
          self._getTableDomUtils().clearCachedDomRowData();
          // row values may have changed so refresh the footer
          self._refreshTableFooter();
          if (self._IsCustomElement())
          {
            return self._animateVisibleRows(addedTableBodyRows, rowIdxArray, 'add');
          }
          else
          {
            return Promise.resolve();
          }
        });
      },
      /**
       * Change all the rows contained in the input array.
       * @param {Array} rows Array of row contexts to change
       * @param {number} startIndex index of first row in the table in the data source
       * @private
       */
      _executeTableBodyRowsChange: function(rows, startIndex)
      {
        var self = this;
        var i, tableBodyRow, rowsCount = rows.length;
        var updatedTableBodyRows = [];
        var rowIdxArray = [];

        this._queueTask(function()
        {
          for (i = 0; i < rowsCount; i++)
          {
            tableBodyRow = self._refreshTableBodyRow(rows[i].rowIdx, rows[i].row);
            updatedTableBodyRows.push(tableBodyRow);
            rowIdxArray.push(rows[i].rowIdx);
          }
          // row values may have changed so refresh the footer
          self._refreshTableFooter();
          if (self._IsCustomElement())
          {
            return self._animateVisibleRows(updatedTableBodyRows, rowIdxArray, 'update');
          }
          else
          {
            return Promise.resolve();
          }
        });
      },
      /**
       * Change all the rows contained in the input array.
       * @param {Array} rows Array of row contexts to change
       * @private
       */
      _executeTableBodyRowsRemove: function(rows)
      {
        // sort array
        rows.sort(function(a, b) {
          return b.rowIdx - a.rowIdx;
        });
        var self = this;
        var currentRow = this._getCurrentRow();
        var currentRowIndex = currentRow != null ? currentRow['rowIndex'] : 0;
        var currentRowKey = currentRow != null ? currentRow['rowKey'] : null;
        this._queueTask(function()
        {
          // first check if we are removing all rows. If so, we can do a removeAll
          var remainingRowIdxArray = [];
          var rowIdxArray = [];
          var removedTableBodyRows = [];
          var i, j, rowKey, rowIdx, rowsCount = rows.length;
          var removeAll = false;
          var tableBodyRow;
          var tableBodyRows = self._getTableDomUtils().getTableBodyRows();
          
          if (tableBodyRows != null && tableBodyRows.length > 0)
          {
            for (i = 0; i < tableBodyRows.length; i++)
            {
              remainingRowIdxArray.push(i);
            }
            for (i = 0; i < rowsCount; i++)
            {
              rowIdx = rows[i].rowIdx;
              rowIdxArray.push(rowIdx);
              for (j = 0; j < remainingRowIdxArray.length; j++)
              {
                if (remainingRowIdxArray[j] == rowIdx)
                {
                  tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);
                  removedTableBodyRows.push(tableBodyRow);
                  remainingRowIdxArray.splice(j, 1);
                  break;
                }
              }
            }
            
            if (remainingRowIdxArray.length == 0)
            {
              removeAll = true;
            }
          }
                              
          var tableBody = this._getTableDomUtils().getTableBody();
          var checkFocus = $.contains(tableBody, document.activeElement);
          var resetFocus = false;
          
          if (removeAll)
          {
            if (checkFocus)
            {
              resetFocus = true;
            }
            
            // Clear out all the existing selection state
            self._setSelection(null);
            self.option('selection', null, {'_context': {writeback: true, internalSet: true}});
          }
          else
          {
            for (i = 0; i < rowsCount; i++)
            {
              rowIdx = rows[i].rowIdx;
              tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);
              if (checkFocus)
              {
                if (tableBodyRow != null &&
                    $.contains(tableBodyRow, document.activeElement))
                {
                  resetFocus = true;
                  checkFocus = false;
                  break;
                }
              }
            }
          }
          
          var _removeRows = function()
          {
            for (i = 0; i < rowsCount; i++)
            {
              rowIdx = rows[i].rowIdx;
              self._getTableDomUtils().removeTableBodyRow(rowIdx);
            }
          }
          var _afterRemoveRows = function()
          {
            // row values may have changed so refresh the footer
            self._refreshTableFooter();
            tableBodyRows = self._getTableDomUtils().getTableBodyRows();
            if (tableBodyRows == null || tableBodyRows.length == 0)
            {
              self._showNoDataMessage();
            }
            
            // set the currentRow to the next row if needed or clear
            if (currentRowKey != null)
            {
              if (tableBodyRows == null || 
                removeAll || 
                currentRowIndex >= tableBodyRows.length)
              {
                self._setCurrentRow(null, null, false);
              }
              else
              {
                self._setCurrentRow({'rowIndex': currentRowIndex, 'rowKey': null}, null, false);
              }
            }

            if (resetFocus)
            {
              self._getTableDomUtils().getTable().focus();
            }
          }
          if (removeAll)
          {
            self._getTableDomUtils().removeAllTableBodyRows();
            _afterRemoveRows();
          }
          else
          {
            return new Promise(function(resolve, reject)
            {
              if (self._IsCustomElement())
              {
                self._animateVisibleRows(removedTableBodyRows, rowIdxArray, 'remove').then(function()
                {      
                  _removeRows();
                  _afterRemoveRows();
                  resolve(true);
                });
              }
              else
              {
                _removeRows();
                _afterRemoveRows();
                resolve(true);
              }
            });
          }
        });
      },
      /**
       * Return all the busy contexts for the element
       * @param {Element} element DOM element
       * @return {Array|null} Array of Busy Contexts
       * @private
       */
      _getAllBusyContextsForElement: function(element)
      {
        if (!element)
        {
          return null;
        }
        var contextNodeList = element.querySelectorAll('[data-oj-context]');
        if (contextNodeList.length > 0)
        {
          var i, busyContextArray = [];
          for (i = 0; i < contextNodeList.length; i++) 
          {
            busyContextArray.push(oj.Context.getContext(contextNodeList[i]).getBusyContext());
          }
          return busyContextArray;
        }
        return null;
      },
      /**
       * Return the column definitions
       * @return {Array} array of column metadata Objects.
       * @private
       */
      _getColumnDefs: function()
      {
        // cache the columns array in this._columnDefArray
        if (!this._columnDefArray)
        {
          this._columnDefArray = this._getColumnMetadata();
        }
        return this._columnDefArray;
      },
      /**
       * Return the column metadata in sorted oder.
       * @return {Array} array of column metadata Objects.
       * @private
       */
      _getColumnMetadata: function()
      {
        // get the columns metadata
        var columns = this.options['columns'];
        var columnsDefault = this.options['columnsDefault'];

        if ((columns.length == 0 ||
            (columns.length == 1 &&
            columns[0].id == null &&
            columns[0].headerText == null &&
            columns[0].field == null)) &&
            (columnsDefault.headerText == null &&
            columnsDefault.field == null))
        {
          return [];
        }

        var defaultedColumns = [];
        var i, columnsCount = columns.length;
        for (i = 0; i < columnsCount; i++)
        {
          defaultedColumns[i] = $.extend({}, columnsDefault, columns[i]);
        }

        var columnsSortedArray = [];
        // add the rest of the columns in the array
        var defaultedColumnsCount = defaultedColumns.length;
        for (i = 0; i < defaultedColumnsCount; i++)
        {
          columnsSortedArray.push(defaultedColumns[i]);
        }

        var data = this._getData();
        var sortSupportedData = false;
        if (data != null && data.getCapability('sort') != null)
        {
          sortSupportedData = true;
        }

        for (i = 0; i < defaultedColumnsCount; i++)
        {
          // generate ids for columns which don't have it specified
          if (columnsSortedArray[i][this._FIELD_ID] == null)
          {
            columnsSortedArray[i][this._FIELD_ID] = this._COLUMN_HEADER_ID_PREFIX + i;
          }
          // for the columns which have sortable = 'auto' check the datasource
          // and enable or disable
          if ((columnsSortedArray[i]['sortable'] == null ||
               columnsSortedArray[i]['sortable'] == this._OPTION_AUTO)
               && sortSupportedData)
          {
            columnsSortedArray[i]['sortable'] = this._OPTION_ENABLED;
          }
        }
        return columnsSortedArray;
      },
      /**
       * Return the column index for column key.
       * @param {Object} columnKey column key
       * @return {number|null} column index
       * @private
       */
      _getColumnIdxForColumnKey: function(columnKey)
      {
        var columns = this._getColumnDefs();

        if (columns != null)
        {
          var i, column, columnsCount = columns.length;
          for (i = 0; i < columnsCount; i++)
          {
            column = columns[i];
            if (oj.Object.compareValues(column.id, columnKey))
            {
              return i;
            }
          }
        }
        return null;
      },
      /**
       * Return all the column indexes for elements with a particular style class
       * @param {string} styleClass  style class
       * @return {Array} Array of column indexes
       * @private
       */
      _getColumnIdxsForElementsWithStyleClass: function(styleClass)
      {
        var elements = this._getTableDomUtils().getTable().querySelectorAll(styleClass);
        var columnIdxs = [];
        if (elements && elements.length > 0)
        {
          var i, j, alreadyAdded, columnIdx, columnIdxsCount, elementsCount = elements.length;
          for (i = 0; i < elementsCount; i++)
          {
            columnIdx = this._getTableDomUtils().getElementColumnIdx(elements[i]);

            alreadyAdded = false;
            columnIdxsCount = columnIdxs.length;
            for (j = 0; j < columnIdxsCount; j++)
            {
              if (columnIdxs[j] == columnIdx)
              {
                alreadyAdded = true;
              }
            }
            if (!alreadyAdded)
            {
              columnIdxs.push(columnIdx);
            }
          }
        }

        return columnIdxs;
      },
      /**
       * Return the column key for column index.
       * @param {number} columnIdx column index
       * @return {Object} column key
       * @private
       */
      _getColumnKeyForColumnIdx: function(columnIdx)
      {
        var columns = this._getColumnDefs();

        if (columns != null && columnIdx < columns.length)
        {
          return columns[columnIdx][this._FIELD_ID];
        }
        return null;
      },
      /**
       * Return the column renderer
       * @param {number} columnIdx  column index
       * @param {String} type  renderer type
       * @return {Object} renderer
       * @private
       */
      _getColumnRenderer: function(columnIdx, type)
      {
        var columns = this._getColumnDefs();
        var column = columns[columnIdx];
        var renderer = null;
        if (type == 'cell')
        {
          renderer = column['renderer'];
        }
        else if (type == 'footer')
        {
          renderer = column['footerRenderer'];
        }
        else if (type == 'header')
        {
          renderer = column['headerRenderer'];
        }
        return this._WrapCustomElementRenderer(renderer);
      },
      /**
       * Get the current row.
       * @return {Object|null} current row object or null if none.
       * @throws {Error}
       * @private
       */
      _getCurrentRow: function()
      {
        var dataprovider = this._getData();
        // if no data then bail
        if (!dataprovider)
        {
          return null;
        }
        return this._currentRow;
      },
      /**
       * Return the datasource object defined for this table
       * @return {Object} Datasource object.
       * @throws {Error}
       * @private
       */
      _getData: function()
      {
        if (!this._data && this.options.data != null)
        {
          var dataprovider = this.options.data;
          if ((oj.TableDataSource && dataprovider instanceof oj.TableDataSource) ||
              (oj.PagingTableDataSource && dataprovider instanceof oj.PagingTableDataSource)) 
          {
            this._data = new oj.TableDataSourceAdapter(dataprovider); 
          }
          else if (oj.DataProviderFeatureChecker.isIteratingDataProvider(dataprovider))
          {
            if (!(dataprovider instanceof oj.ListDataProviderView)) 
            {
              this._data = new oj.ListDataProviderView(dataprovider);
            }
            else
            {
              this._data = dataprovider;
            }
          }
          else
          {
            // we only support TableDataSource
            var errSummary = this._LOGGER_MSG._ERR_DATA_INVALID_TYPE_SUMMARY;
            var errDetail = this._LOGGER_MSG._ERR_DATA_INVALID_TYPE_DETAIL;
            throw new Error(errSummary + '\n' + errDetail);
          }
          this._dataOption = this.options.data;
          this._registerDataSourceEventListeners();
        }
        return this._data;
      },
      /**
       * Get the focused column header index
       * @return {number|null} the column index
       * @private
       */
      _getFocusedHeaderColumnIdx: function()
      {
        // focused column headers have the focused style class. There should only be one focused header
        return this._getColumnIdxsForElementsWithStyleClass('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS + '.' + oj.TableDomUtils.MARKER_STYLE_CLASSES._FOCUS_HIGHLIGHT + ',' +
                                                            '.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS + '.' + oj.TableDomUtils.MARKER_STYLE_CLASSES._FOCUS)[0];
      },
      /**
       * Get the focused row index
       * @return {number|null} the row index
       * @private
       */
      _getFocusedRowIdx: function()
      {
        // focused rows have cells with focused style class. There should only be one focused row
        return this._getRowIdxsForElementsWithStyleClass('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS + '.' + oj.TableDomUtils.MARKER_STYLE_CLASSES._FOCUS_HIGHLIGHT + ',' +
                                                         '.' + oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS + '.' + oj.TableDomUtils.MARKER_STYLE_CLASSES._FOCUS)[0];
      },
      /**
       * Return whether the column header at the index is selected
       * @param {number} columnIdx  column index
       * @return {boolean} whether it's selected
       * @private
       */
      _getHeaderColumnSelection: function(columnIdx)
      {
        var headerColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);
        return headerColumn.classList.contains(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
      },
      /**
       * Return the column selection mode
       * @return {string|null} single, multiple, or null
       * @private
       */
      _getColumnSelectionMode: function()
      {
        var columnSelectionMode = this.options.selectionMode == null ? null : this.options.selectionMode[this._CONST_COLUMN];
        return columnSelectionMode;
      },
      /**
       * Return the currently pressed keyboard keys
       * @return {Array} Array of keyCodes
       * @private
       */
      _getKeyboardKeys: function()
      {
        if (!this._keyboardKeys)
        {
          this._keyboardKeys = [];
        }

        // reverse the array since we want the keybaord keys to be a LIFO stack
        return this._keyboardKeys.reverse();
      },
      /**
       * Return the last column which was selected (chronologically)
       * @return {number|null} last selected column
       * @private
       */
      _getLastHeaderColumnSelection: function()
      {
        if (this._lastSelectedColumnIdxArray != null &&
          this._lastSelectedColumnIdxArray.length > 0)
        {
          return this._lastSelectedColumnIdxArray[0];
        }
        return null;
      },
      /**
       * Return the last row which was selected (chronologically)
       * @return {number|null} last selected row
       * @private
       */
      _getLastRowSelection: function()
      {
        if (this._lastSelectedRowIdxArray != null &&
          this._lastSelectedRowIdxArray.length > 0)
        {
          return this._lastSelectedRowIdxArray[0];
        }
        return null;
      },
      /**
       * Return an array with row and row indices relative to the table row indices
       * @param {Object} resultObject Object containing data array, key array, and index array
       * @param {number} startIndex start index
       * @return {Array} Array of rows and row index
       * @private
       */
      _getRowIdxRowArray: function(resultObject, startIndex)
      {
        var rowIdxRowArray = [];
        if (resultObject != null)
        {
          var i, indexesCount = resultObject[this._CONST_INDEXES].length;
          for (i = 0; i < indexesCount; i++)
          {
            rowIdxRowArray.push({row: {'data': resultObject[this._CONST_DATA][i], 'metadata': resultObject[this._CONST_METADATA] ? resultObject[this._CONST_METADATA][i] : null, 'key': resultObject[this._CONST_KEYS][i], 'index': resultObject[this._CONST_INDEXES][i]}, rowIdx: startIndex + i});
          }
        }

        return rowIdxRowArray;
      },
      /**
       * Return the row index for row key. Only loop through displayed rows.
       * @param {Object} rowKey row key
       * @return {number|null} row index
       * @private
       */
      _getRowIdxForRowKey: function(rowKey)
      {
        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();

        if (tableBodyRows != null && tableBodyRows.length > 0)
        {
          var i, tableBodyRowsCount = tableBodyRows.length;
          for (i = 0; i < tableBodyRowsCount; i++)
          {
            if (oj.Object.compareValues($(tableBodyRows[i]).data('rowKey'), rowKey))
            {
              return i;
            }
          }
        }

        return null;
      },
      /**
       * Return the datasource's row index for the row key.
       * @param {Object} rowKey row key
       * @return {number|null} row index
       * @private
       */
      _getDataSourceRowIndexForRowKey: function(rowKey)
      {
        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();

        if (tableBodyRows != null && tableBodyRows.length > 0)
        {
          var i, tableBodyRowsCount = tableBodyRows.length;;
          for (i = 0; i < tableBodyRowsCount; i++)
          {
            if (oj.Object.compareValues($(tableBodyRows[i]).data('rowKey'), rowKey))
            {
              var dataprovider = this._getData();
              var offset = 0;
              if (this._isPagingModelDataProvider()) 
              {
                offset = dataprovider.getStartItemIndex();
              }
              return i + offset;
            }
          }
        }

        return null;
      },
      /**
       * Return the editable row index if there is one
       * @return {number|null} row index
       * @private
       */
      _getEditableRowIdx: function()
      {
        var rowIdx = null;
        
        if (this._getEditableRowKey() != null)
        {
          rowIdx = this._getRowIdxForRowKey(this._getEditableRowKey());
          
          if (rowIdx !== null)
          {
            return rowIdx;
          }
        }
        return this._editableRowIdx;
      },
      /**
       * Return the editable row key if there is one
       * @return {Object|null} row key
       * @private
       */
      _getEditableRowKey: function()
      {
        return this._editableRowKey;
      },
      /**
       * 
       * Get the target element
       * @param {Object} event DOM touch event
       * @return {Element} element  DOM element
       * @private
       */
      _getEventTargetElement: function(event)
      {
        return event.type.indexOf('touch') == 0 ? this._getTouchEventTargetElement(event) : event.currentTarget;
      },
      /**
       * Return all the row indexes for elements with a particular style class
       * @param {string} styleClass  style class
       * @return {Array} Array of row indexes
       * @private
       */
      _getRowIdxsForElementsWithStyleClass: function(styleClass)
      {
        var elements = this._getTableDomUtils().getTable().querySelectorAll(styleClass);
        var rowIdxs = [];
        if (elements && elements.length > 0)
        {
          var i, j, rowIdx, rowIdxsCount, alreadyAdded, elementsCount = elements.length;
          for (i = 0; i < elementsCount; i++)
          {
            rowIdx = this._getTableDomUtils().getElementRowIdx(elements[i]);

            alreadyAdded = false;
            rowIdxsCount = rowIdxs.length;
            for (j = 0; j < rowIdxsCount; j++)
            {
              if (rowIdxs[j] == rowIdx)
              {
                alreadyAdded = true;
              }
            }
            if (!alreadyAdded)
            {
              rowIdxs.push(rowIdx);
            }
          }
        }

        return rowIdxs;
      },
      /**
       * Return the row key for datasource's row index.
       * @param {number} rowIndex row index
       * @return {*} row key
       * @private
       */
      _getRowKeyForDataSourceRowIndex: function(rowIndex)
      {
        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();

        if (tableBodyRows != null && tableBodyRows.length > 0)
        {
          var dataprovider = this._getData();
          var offset = 0;
          if (this._isPagingModelDataProvider()) 
          {
            offset = dataprovider.getStartItemIndex();
          }
          var i, tableBodyRowsCount = tableBodyRows.length;
          for (i = 0; i < tableBodyRowsCount; i++)
          {
            if (offset + i == rowIndex)
            {
              return $(tableBodyRows[i]).data('rowKey');
            }
          }
        }

        return null;
      },
      /**
       * Return the row key for row index.
       * @param {number} rowIdx row index
       * @return {*|null} row key
       * @private
       */
      _getRowKeyForRowIdx: function(rowIdx)
      {
        var tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);

        if (tableBodyRow != null)
        {
          return $(tableBodyRow).data('rowKey');
        }

        return null;
      },
      /**
       * Return the row renderer
       * @return {Object} renderer
       * @private
       */
      _getRowRenderer: function()
      {
        return this._WrapCustomElementRenderer(this.options['rowRenderer']);
      },
      /**
       * Return whether the row is selected
       * @param {number} rowIdx  row index
       * @return {boolean} whether the row is selected
       * @private
       */
      _getRowSelection: function(rowIdx)
      {
        return this._getTableDomUtils().getTableBodyRow(rowIdx).classList.contains(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
      },
      /**
       * Return the row selection mode
       * @return {string|null} single, multiple, or null
       * @private
       */
      _getRowSelectionMode: function()
      {
        var rowSelectionMode = this.options['selectionMode'] == null ? null : this.options['selectionMode'][this._CONST_ROW];
        return rowSelectionMode;
      },
      /**
       * Return the selected column header indexes
       * @return {Array} array of column header indexes
       * @private
       */
      _getSelectedHeaderColumnIdxs: function()
      {
        // selected column headers have the selected css class
        return this._getColumnIdxsForElementsWithStyleClass('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS + '.' + oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
      },
      /**
       * Return the selected row indexes
       * @return {Array} array of row indexes
       * @private
       */
      _getSelectedRowIdxs: function()
      {
        // selected rows have the selected css class
        return this._getRowIdxsForElementsWithStyleClass('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS + '.' + oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
      },
      /**
       * Gets the selection
       * @private
       */
      _getSelection: function()
      {
        var selectedRowIdxs = this._getSelectedRowIdxs();
        var selectedColumnIdxs = this._getSelectedHeaderColumnIdxs();
        var selectionIdxs = null;
        var rowSelection = true;
        if (selectedRowIdxs != null && selectedRowIdxs.length > 0)
        {
          selectionIdxs = selectedRowIdxs;
        }
        else if (selectedColumnIdxs != null && selectedColumnIdxs.length > 0)
        {
          selectionIdxs = selectedColumnIdxs;
          rowSelection = false;
        }
        else
        {
          return [];
        }

        var rangeArray = [];

        // first count the number of ranges we have by seeing how many
        // non-continguous selections we have
        var rangeCount = 0;
        var previousIdx = null;
        var rangeObj, selectionIndex, selectionIdx, selectionKey;
        var i, selectionIdxsCount = selectionIdxs.length;
        for (i = 0; i < selectionIdxsCount; i++)
        {
          selectionIdx = selectionIdxs[i];
          if (i == 0)
          {
            rangeObj = {};
            rangeObj[this._CONST_STARTINDEX] = {};
            rangeObj[this._CONST_ENDINDEX] = {};
            rangeObj['startKey'] = {};
            rangeObj['endKey'] = {};
            if (rowSelection)
            {
              selectionKey = this._getRowKeyForRowIdx(selectionIdx);
              selectionIndex = this._getDataSourceRowIndexForRowKey(selectionKey);
              rangeObj['startKey'][this._CONST_ROW] = selectionKey;
              rangeObj['endKey'][this._CONST_ROW] = selectionKey;
              rangeObj[this._CONST_STARTINDEX][this._CONST_ROW] = selectionIndex;
              rangeObj[this._CONST_ENDINDEX][this._CONST_ROW] = selectionIndex;
            }
            else
            {
              rangeObj[this._CONST_STARTINDEX][this._CONST_COLUMN] = selectionIdx;
              rangeObj[this._CONST_ENDINDEX][this._CONST_COLUMN] = selectionIdx;
              selectionKey = this._getColumnKeyForColumnIdx(selectionIdx);
              rangeObj['startKey'][this._CONST_COLUMN] = selectionKey;
              rangeObj['endKey'][this._CONST_COLUMN] = selectionKey;
            }
            rangeArray[0] = rangeObj;
          }
          else
          {
            rangeObj = rangeArray[rangeCount];
            if (rowSelection)
            {
              selectionKey = this._getRowKeyForRowIdx(selectionIdx);
              selectionIndex = this._getDataSourceRowIndexForRowKey(selectionKey);
              rangeObj['endKey'][this._CONST_ROW] = selectionKey;
              rangeObj[this._CONST_ENDINDEX][this._CONST_ROW] = selectionIndex;
            }
            else
            {
              rangeObj[this._CONST_ENDINDEX][this._CONST_COLUMN] = selectionIdx;
              selectionKey = this._getColumnKeyForColumnIdx(selectionIdx);
              rangeObj['endKey'][this._CONST_COLUMN] = selectionKey;
            }
            if (selectionIdx != previousIdx + 1)
            {
              if (rowSelection)
              {
                selectionKey = this._getRowKeyForRowIdx(previousIdx);
                selectionIndex = this._getDataSourceRowIndexForRowKey(selectionKey);
                rangeObj['endKey'][this._CONST_ROW] = selectionKey;
                rangeObj[this._CONST_ENDINDEX][this._CONST_ROW] = selectionIndex;
                rangeObj = {};
                rangeObj[this._CONST_STARTINDEX] = {};
                rangeObj[this._CONST_ENDINDEX] = {};
                rangeObj['startKey'] = {};
                rangeObj['endKey'] = {};
                selectionKey = this._getRowKeyForRowIdx(selectionIdx);
                selectionIndex = this._getDataSourceRowIndexForRowKey(selectionKey);
                rangeObj['startKey'][this._CONST_ROW] = selectionKey;
                rangeObj['endKey'][this._CONST_ROW] = selectionKey;
                rangeObj[this._CONST_STARTINDEX][this._CONST_ROW] = selectionIndex;
                rangeObj[this._CONST_ENDINDEX][this._CONST_ROW] = selectionIndex;
              }
              else
              {
                rangeObj[this._CONST_ENDINDEX][this._CONST_COLUMN] = previousIdx;
                selectionKey = this._getColumnKeyForColumnIdx(previousIdx);
                rangeObj['endKey'][this._CONST_COLUMN] = selectionKey;
                rangeObj = {};
                rangeObj[this._CONST_STARTINDEX] = {};
                rangeObj[this._CONST_ENDINDEX] = {};
                rangeObj['startKey'] = {};
                rangeObj['endKey'] = {};
                rangeObj[this._CONST_STARTINDEX][this._CONST_COLUMN] = selectionIdx;
                rangeObj[this._CONST_ENDINDEX][this._CONST_COLUMN] = selectionIdx;
                selectionKey = this._getColumnKeyForColumnIdx(selectionIdx);
                rangeObj['startKey'][this._CONST_COLUMN] = selectionKey;
                rangeObj['endKey'][this._CONST_COLUMN] = selectionKey;
              }
              rangeCount++;
              rangeArray[rangeCount] = rangeObj;
            }
          }
          previousIdx = selectionIdx;
        }
        return rangeArray;
      },
      /**
       * Returns the inline template element inside an oj-table slot
       * @param {string} slotName slot name
       * @return {Element|null} the inline template element
       * @private
       */
      _getSlotTemplate: function(slotName)
      {
        var slotMap, slot;

        if (this._IsCustomElement() && slotName)
        {
          slotMap = oj.BaseCustomElementBridge.getSlotMap(this._getRootElement());        
          slot = slotMap[slotName];
          if (slot && slot.length > 0 && slot[0].tagName.toLowerCase() === "template")
          {
            return slot[0];
          }
        }
        return null;
      },
      /**
       * Return the currnetly sorted column index
       * @return {number|null} column index
       * @private
       */
      _getSortedTableHeaderColumnIdx: function()
      {
        var tableHeaderColumns = this._getTableDomUtils().getTableHeaderColumns();

        var i, sorted, tableHeaderColumnsCount = tableHeaderColumns ? tableHeaderColumns.length : 0;
        for (i = 0; i < tableHeaderColumnsCount; i++)
        {
          // sorted column will have the sorted data attr
          sorted = $(tableHeaderColumns[i]).data('sorted');

          if (sorted != null)
          {
            return i;
          }
        }
        return null;
      },
      /**
       * Get tabbable elements within the element
       * @param {Element} element  DOM element
       * @param {boolean=} ignoreHidden ignore accessibility hidden elements
       * @return {Array|null} Array of DOM elements
       * @private
       */
      _getTabbableElements: function(element, ignoreHidden)
      {
        var tabbableElements = null;
        
        if (ignoreHidden)
        {
          tabbableElements = $(element).find(':tabbable');
        }
        else
        {
          tabbableElements = $(element).find(':tabbable').not('.oj-helper-hidden-accessible');
        }

        if (tabbableElements != null && tabbableElements.length > 0)
        {
          return tabbableElements.toArray();
        }
        return null;
      },
      /**
       * Return table animation utils instance
       * @return {Object} instance of table DOM utils
       * @private
       */
      _getTableAnimationUtils: function()
      {
        if (!this._tableAnimationUtils)
        {
          this._tableAnimationUtils = new oj.TableAnimationUtils(this);
        }
        return this._tableAnimationUtils;
      },
      /**
       * Return table DnD utils instance
       * @return {Object} instance of table DnD utils
       * @private
       */
      _getTableDndContext: function()
      {
        if (!this._tableDndContext)
        {
          this._tableDndContext = new oj.TableDndContext(this);
        }
        return this._tableDndContext;
      },
      /**
       * Return table DOM utils instance
       * @return {Object} instance of table DOM utils
       * @private
       */
      _getTableDomUtils: function()
      {
        if (!this._tableDomUtils)
        {
          this._tableDomUtils = new oj.TableDomUtils(this);
        }
        return this._tableDomUtils;
      },
      /**
       * Return table Resize utils instance
       * @return {Object} instance of table Resize utils
       * @private
       */
      _getTableResizeUtils: function()
      {
        if (!this._tableResizeUtils)
        {
          this._tableResizeUtils = new oj.TableResizeUtils(this);
        }
        return this._tableResizeUtils;
      },
      /**
       * Return the template engine or null
       * @return {Object|null} template engine
       * @private
       */
      _getTemplateEngine: function()
      {
        return this._templateEngine;
      },
      /**
       * 
       * Get the target element at the touch event
       * @param {Object} event DOM touch event
       * @return {Element} element  DOM element
       * @private
       */
      _getTouchEventTargetElement: function(event)
      {
        var eventLocation = event.originalEvent.changedTouches[0];
        return document.elementFromPoint(eventLocation.clientX, eventLocation.clientY);
      },
      /**
       * Return the visible row indexes
       * @return {Array} array of row indexes
       * @private
       */
      _getVisibleRowIdxs: function()
      {
        // return the row indexes of all rows in the viewport
        var visibleRowIdxArray = [];
        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
          
        if (tableBodyRows != null && tableBodyRows.length > 0)
        {
          var windowHeight = $(window).height();
          var tableRect = this.element[0].getBoundingClientRect();
          var i;
          
          // If the whole table is visible then all the rows are visible
          if (tableRect.top >= 0 &&
            tableRect.bottom <= windowHeight)
          {
            for (i = 0; i < tableBodyRows.length; i++)
            {
              visibleRowIdxArray.push(i);
            }
          }
          // check if not visible at all
          else if (tableRect.top > windowHeight)
          {
            return visibleRowIdxArray; 
          }
          // get the endpoint rows for the viewport
          else
          {
            var tableElemX = tableRect.left >= 0 ? tableRect.left + 1 : tableRect.right - 1;
            var startRowIdx, lastRowIdx;
            
            // check if the first visible row is at the top of the viewport
            var tableElem = document.elementFromPoint(tableElemX, 0);
            var rowIdx = null;
            if (tableElem != null)
            {
              rowIdx = this._getTableDomUtils().getElementRowIdx(tableElem);
            }
            startRowIdx = rowIdx != null ? rowIdx : 0;
            
            if (tableRect.bottom > windowHeight)
            {
              // the last visible row is at the bottom of the viewport
              tableElem = document.elementFromPoint(tableElemX, windowHeight - 1);
              rowIdx = null;
              if (tableElem != null)
              {
                rowIdx = this._getTableDomUtils().getElementRowIdx(tableElem);
              }
              lastRowIdx = rowIdx != null ? rowIdx : tableBodyRows.length - 1;
            }
            else
            {
              // the last visible row is the last row
              lastRowIdx = tableBodyRows.length - 1;
            }
            
            for (i = startRowIdx; i <= lastRowIdx; i++)
            {
              visibleRowIdxArray.push(i);
            }     
          }
        }
        return visibleRowIdxArray;
      },
      /**
       * Handle an ojselect event on a menu item, if sort call the handler on the core.
       * If resize prompt the user with a popup
       * @private
       */
      _handleContextMenuSelect: function(event, ui)
      {
        var item;
        if (ui)
          item = ui.item;
        else
          item = $(event.target);
        var menuItemCommand = item.attr('data-oj-command');
        var headerColumn = this._getTableDomUtils().getFirstAncestor(this._contextMenuEvent['target'], 'oj-table-column-header-cell');
        headerColumn = headerColumn == null ? this._contextMenuEventHeaderColumn : headerColumn;
        var tableBodyCell = this._getTableDomUtils().getFirstAncestor(this._contextMenuEvent['target'], 'oj-table-data-cell');
        var columnIdx = null;

        if (headerColumn != null)
        {
          columnIdx = this._getTableDomUtils().getElementColumnIdx(headerColumn);
        }
        if (tableBodyCell != null)
        {
          columnIdx = this._getTableDomUtils().getElementColumnIdx(tableBodyCell);
        }
        if (columnIdx === null)
        {
          return;
        }
        if (menuItemCommand == 'oj-table-sortAsc')
        {
          this._handleSortTableHeaderColumn(columnIdx, true, event);
        }
        else if (menuItemCommand == 'oj-table-sortDsc')
        {
          this._handleSortTableHeaderColumn(columnIdx, false, event);
        }
        else if (menuItemCommand == 'oj-table-enableNonContiguousSelection')
        {
          this._nonContiguousSelection = true;
          // update to disable command
          item.attr('data-oj-command', 'oj-table-disableNonContiguousSelection');
          item.children().first().text(this.getTranslatedString('labelDisableNonContiguousSelection')); //@HTMLUpdateOK
        }
        else if (menuItemCommand == 'oj-table-disableNonContiguousSelection')
        {
          this._nonContiguousSelection = false;
          // update to enable command
          item.attr('data-oj-command', 'oj-table-enableNonContiguousSelection');
          item.children().first().text(this.getTranslatedString('labelEnableNonContiguousSelection')); //@HTMLUpdateOK
        }
        else if (menuItemCommand == 'oj-table-resize')
        {
          var popup = this._getTableDomUtils().getContextMenuResizePopup();
          popup.setAttribute('data-oj-columnIdx', columnIdx);
          var columnRect = headerColumn != null ? headerColumn.getBoundingClientRect() : tableBodyCell.getBoundingClientRect();
          var spinner = document.getElementById(this._getTableDomUtils().getTableId() + '_resize_popup_spinner');
          $(spinner).ojInputNumber('option', 'value', Math.round(columnRect.width));
          var launcher = headerColumn != null ? headerColumn : tableBodyCell != null ? tableBodyCell : this.getTable();
          $(popup).ojPopup('open', launcher);
        }
      },
      /**
       * Callback from the resize popup box
       * @param {Event} event the event that triggered the popup button press
       * @private
       */
      _handleContextMenuResizePopup: function(event)
      {
        var spinner = document.getElementById(this._getTableDomUtils().getTableId() + '_resize_popup_spinner');
        var widthValue = $(spinner).ojInputNumber("option", "value");
        var popup = this._getTableDomUtils().getContextMenuResizePopup();
        var columnIdx = parseInt(popup.getAttribute('data-oj-columnIdx'), 10);
        $(popup).ojPopup('close');
        var clonedColumnsOption = [];
        var i, columnsCount = this.options['columns'].length;
        for (i = 0; i < columnsCount; i++)
        {
          clonedColumnsOption[i] = $.extend({}, {}, this.options['columns'][i]);
        }
        clonedColumnsOption[columnIdx]['width'] = widthValue;
        this.option('columns', clonedColumnsOption, {'_context': {writeback: true, internalSet: true}});
        this._clearCachedMetadata();
        var self = this;
        this._refresh().then(function()
        {
          self._getTableDomUtils().getTable().focus();
        });
      },
      /**
       * Callback handler for fetch completed in the datasource. Refresh entire
       * table body DOM and refresh the table dimensions if refresh == true. Hide the Fetching Data...
       * status message.
       * @param {Object} event
       * @private
       */
      _handleDataAppend: function(event)
      {
        try
        {
          var self = this;
          this._queueTask(function()
          {
            var indexArray = [];
            var i, eventDataCount = event[self._CONST_DATA].length;

            for (i = 0; i < eventDataCount; i++)
            {
              // event['startIndex'] contains the offset at which the data should be inserted in the table. In paging mode
              // this is always zero. In loadMore mode it contains an offset.
              // Therefore we have to add both. e.g. in paging mode offset is non-zero while in loadMore event['startIndex'] is non-zero.
              // The indexArray will contain the indexes as contained in the datasource.
              indexArray[i] = i + event[self._CONST_STARTINDEX];
            }

            self._refreshAll({'data': event[self._CONST_DATA], 'keys' : event[self._CONST_KEYS], 'indexes': indexArray}, event[self._CONST_STARTINDEX]);
            self = null;
          });
        }
        catch (e)
        {
          oj.Logger.error(e);
        }
        finally
        {
          this._clearDataWaitingState();
        }
      },
      /**
       * Callback handler for data error.
       * @param {Object} error
       * @private
       */
      _handleDataError: function(error)
      {
        this._clearDataWaitingState();
        oj.Logger.error(error);
      },
      /**
       * Callback handler for fetch start in the datasource.
       * @param {Object} event
       * @private
       */
      _handleDataFetchStart: function(event)
      {
        this._setDataWaitingState();
      },
      /**
       * Callback handler for refresh in the datasource. Refresh entire
       * table body DOM and refresh the table dimensions.
       * @param {Object} event
       * @private
       */
      _handleDataRefresh: function (event)
      {
        try
        {
          if (this._dataFetching) 
          {
            return;
          }
          var self = this;
          this._queueTask(function ()
          {
            var options = {};
            var fetchPromise = self._invokeDataFetchRows(options);
            self = null;
            return fetchPromise;
          });
        }
        catch (e)
        {
          oj.Logger.error(e);
        }
        finally
        {
          this._clearDataWaitingState();
        }
      },
      /**
       * Callback handler for rows mutations.
       * @param {Object} event
       * @private
       */
      _handleDataRowMutate: function(event)
      {
        if (this._dataFetching) 
        {
          return;
        }
        if (event['detail']['remove'] != null)
        {
          this._handleDataRowRemove(event['detail']['remove']);
        }
        if (event['detail']['add'] != null)
        {
          this._handleDataRowAdd(event['detail']['add']);
        }
        if (event['detail']['update'] != null)
        {
          this._handleDataRowChange(event['detail']['update']);
        }
      },
      /**
       * Callback handler for rows added into the datasource. Add a new tr and refresh the DOM
       * at the row index and refresh the table dimensions to accomodate the new
       * row
       * @param {Object} eventDetail event detail
       * @private
       */
      _handleDataRowAdd: function(eventDetail)
      {
        try
        {
          var self = this;
          var dataprovider = this._getData();
          var eventData = eventDetail[this._CONST_DATA];
          var eventIndexes = eventDetail[this._CONST_INDEXES];
          var eventKeys = [];
          eventDetail[this._CONST_KEYS].forEach(function(key) {
            eventKeys.push(key);
          });
          var eventAfterKeys = [];
          if (eventDetail[this._CONST_AFTERKEYS] != null) {
            eventDetail[this._CONST_AFTERKEYS].forEach(function(key) {
              eventAfterKeys.push(key);
            });
          }
          if (!(eventData instanceof Array))
          {
            eventData = [eventData];
          }
          var offset = 0;
          if (this._isPagingModelDataProvider()) 
          {
            offset = dataprovider.getStartItemIndex();
          }
          
          var metadataSource = null;
          
          if (dataprovider instanceof oj.TableDataSourceAdapter)
          {
            if (oj.FlattenedTreeTableDataSource && dataprovider.tableDataSource instanceof oj.FlattenedTreeTableDataSource)
            {
              metadataSource = dataprovider.tableDataSource;
            }
            else if (oj.PagingTableDataSource && 
              dataprovider.tableDataSource instanceof oj.PagingTableDataSource &&
              oj.FlattenedTreeTableDataSource && 
              dataprovider.tableDataSource.dataSource instanceof oj.FlattenedTreeTableDataSource)
            {
              metadataSource = dataprovider.tableDataSource.dataSource;
            }
          }
          
          var rowArray = [];
          var rowIdx;
          var eventIndex;
          var eventAfterKey;
          var metadata = null;
          var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
          var rowCount = tableBodyRows != null ? tableBodyRows.length : 0;
          var i, eventDataCount = eventData.length;
          
          // if specified, afterKeys takes precendence over indexes
          for (i = 0; i < eventDataCount; i++)
          {
            eventAfterKey = eventAfterKeys != null ? eventAfterKeys[i] : '';
            rowIdx = this._getRowIdxForRowKey(eventAfterKey);
            if (rowIdx != null && 
              rowIdx >= 0) {
              // insertion is before the afterKey
              eventIndex = rowIdx + offset;
            } else if (eventIndexes[i] >= 0) {
              // if the afterKey is not in the table
              // then use if index has been specified
              eventIndex = eventIndexes[i];
            } else {
              // afterKey was not found and index was
              // not specified. eventIndex is just set as
              // the previous one + 1. If the first item
              // then just append
              if (i == 0) {
                eventIndex = rowCount;
              } else {
                eventIndex = eventIndexes[i - 1] + 1;
              }
            }
            eventIndexes[i] = eventIndex;
          }
          for (i = 0; i < eventDataCount; i++)
          {
            rowIdx = eventIndexes[i] - offset;
            if (rowIdx !== undefined)
            {
              if (metadataSource)
              {
                metadata = metadataSource._getMetadata(rowIdx);
              }
              var row = {'data': eventData[i], 'metadata': metadata, 'key': eventKeys[i], 'index': eventIndexes[i]};

              rowArray.push({row: row, rowIdx: rowIdx});
            }
          }
          if (rowArray.length > 0)
          {
            this._executeTableBodyRowsAdd(rowArray);
          }
        }
        catch (e)
        {
          oj.Logger.error(e);
        }
        finally
        {
          this._clearDataWaitingState();
        }
      },
      /**
       * Callback handler for row change in the datasource. Refresh the changed
       * row.
       * @param {Object} eventDetail event detail
       * @private
       */
      _handleDataRowChange: function(eventDetail)
      {
        try
        {
          var self = this;
          var eventData = eventDetail[this._CONST_DATA];
          var eventKeys = [];
          eventDetail[this._CONST_KEYS].forEach(function(key) {
            eventKeys.push(key);
          });
          if (!(eventData instanceof Array))
          {
            eventData = [eventData];
          }
          var rowArray = [];
          var i, eventDataCount = eventData.length;
          for (i = 0; i < eventDataCount; i++)
          {
            var rowIdx = this._getRowIdxForRowKey(eventKeys[i]);
            if (rowIdx !== undefined)
            {
              var row = {'data': eventData[i], 'key': eventKeys[i], 'index': rowIdx};
              rowArray.push({row: row, rowIdx: rowIdx});
            }
          }
          
          if (rowArray.length > 0)
          {
            this._executeTableBodyRowsChange(rowArray);
          }
        }
        catch (e)
        {
          oj.Logger.error(e);
        }
        finally
        {
          this._clearDataWaitingState();
        }
      },
      /**
       * Callback handler for row removed in the datasource. Remove the row DOM from the
       * table body by searching for the matching rowKey. New rows will have null rowKey.
       * After removing the row, refresh all the remaining row indexes since
       * they will have shifted. Lastly, refresh the table dimensions
       * @param {Object} eventDetail event detail
       * @private
       */
      _handleDataRowRemove: function(eventDetail)
      {
        try
        {
          var self = this;
          var eventData = eventDetail[this._CONST_DATA];
          var eventKeys = [];
          eventDetail[this._CONST_KEYS].forEach(function(key) {
            eventKeys.push(key);
          });
          if (!(eventData instanceof Array))
          {
            eventData = [eventData];
          }
          var rowArray = [];
          var i, eventDataCount = eventData.length
          for (i = 0; i < eventDataCount; i++)
          {
            var rowIdx = this._getRowIdxForRowKey(eventKeys[i]);
            if (rowIdx !== undefined)
            {
              var row = {'data': eventData[i], 'key': eventKeys[i], 'index': rowIdx};
              rowArray.push({row: row, rowIdx: rowIdx});
            }
          }

          if (rowArray.length > 0)
          {
            this._executeTableBodyRowsRemove(rowArray);
          }
        }
        catch (e)
        {
          oj.Logger.error(e);
        }
        finally
        {
          this._clearDataWaitingState();
        }
      },
      /**
       * Handler for End keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownEnd: function(event)
      {
        if (this._isTableActionableMode())
        {
          // ignore in actionable mode
          return;
        }
        // pressing End focuses on last column
        var focusedColumnIdx = this._getFocusedHeaderColumnIdx();

        if (focusedColumnIdx != null &&
          focusedColumnIdx != this._getColumnDefs().length - 1)
        {
          this._setHeaderColumnFocus(this._getColumnDefs().length - 1, true, false, null);
        }
        else if (!this._hasEditableRow())
        {
          var focusedRowIdx = this._getFocusedRowIdx();
          var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
          var rowCount = tableBodyRows != null ? tableBodyRows.length : 0;

          if (focusedRowIdx != null && focusedRowIdx != rowCount - 1 && rowCount > 0)
          {
            this._setRowFocus(rowCount - 1, true, true, null, event);
          }
        }
      },
      /**
       * Handler for Enter keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownEnter: function(event)
      {
        if (this._isTableActionableMode())
        {
          // ignore in actionable mode
          return;
        }
        // pressing enter does sort on the focused column header
        var focusedColumnIdx = this._getFocusedHeaderColumnIdx();

        if (focusedColumnIdx != null && this._getColumnDefs()[focusedColumnIdx]['sortable'] == this._OPTION_ENABLED)
        {
          var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(focusedColumnIdx);
          var sorted = $(tableHeaderColumn).data('sorted');
          // if not already sorted then sort ascending. If already sorted
          // ascending then do descending sort and vice versa.
          if (sorted == null || sorted == this._COLUMN_SORT_ORDER._DESCENDING)
          {
            this._handleSortTableHeaderColumn(focusedColumnIdx, true, event);
          }
          else
          {
            this._handleSortTableHeaderColumn(focusedColumnIdx, false, event);
          }
        }
        else
        {
          var currentRow = this._getCurrentRow();
          currentRow = currentRow || {};
          var currentRowIdx = currentRow['rowIndex'];
          
          if (currentRowIdx >= 0)
          {
            if (this._isTableEditMode())
            {
            if (!this._hasEditableRow())
            {
              this._setTableEditable(true, false, 0, true, event);
              return;
            }
              var columnIdx = this._getTableDomUtils().getElementColumnIdx(event.target);
            
            if (!event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              this._setNextRowEditable(columnIdx, event);
            }
            else
            {
              this._setPreviousRowEditable(columnIdx, event);
            }
          }
            else
            {
              this._setTableActionableMode(true);
        }
          }
        }
      },
      /**
       * Handler for Esc keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownEsc: function(event)
      {
        // pressing Esc always returns focus back to the table.
        event.preventDefault();
        event.stopPropagation();
        this._setTableEditable(false, true, 0, true, event);
        this._getTableDomUtils().getTable().focus();
        this._setTableActionableMode(false);
        this._getTableResizeUtils().clearTableHeaderColumnsResize();
      },
      /**
       * Handler for F2 keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownF2: function(event)
      {
        event.preventDefault();
        event.stopPropagation();
        if (this._isTableEditMode())
        {
          // pressing F2 toggles between editable modes.
          if (!this._hasEditableRow())
          {
            this._setTableEditable(true, false, 0, true, event);
          }
          else
          {
            this._setTableEditable(false, false, 0, true, event);
          }
        }
        else
        {
          if (this._isTableActionableMode())
          {
            this._setTableActionableMode(false);
          }
          else
          {
            this._setTableActionableMode(true);
          }
        }
      },
      /**
       * Handler for Home keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownHome: function(event)
      {
        if (this._isTableActionableMode())
        {
          // ignore in actionable mode
          return;
        }
        // pressing Home focuses on first column
        var focusedColumnIdx = this._getFocusedHeaderColumnIdx();

        if (focusedColumnIdx != null && focusedColumnIdx != 0)
        {
          this._setHeaderColumnFocus(0, true, false, null);
        }
        else if (!this._hasEditableRow())
        {
          var focusedRowIdx = this._getFocusedRowIdx();

          if (focusedRowIdx != null && focusedRowIdx != 0)
          {
            this._setRowFocus(0, true, true, null, event);
          }
        }
      },
      /**
       * Handler for Left/Right keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownLeftRight: function(event)
      {
        if (this._isTableActionableMode())
        {
          // ignore in actionable mode
          return;
        }
        // pressing left/right navigates the column headers
        var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();
        var columns = this._getColumnDefs();

        if (focusedHeaderColumnIdx != null)
        {
          var newFocusedHeaderColumnIdx = focusedHeaderColumnIdx;

          if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_LEFT))
          {
            newFocusedHeaderColumnIdx = focusedHeaderColumnIdx > 0 ? focusedHeaderColumnIdx - 1 : focusedHeaderColumnIdx;
          }
          else if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_RIGHT))
          {
            newFocusedHeaderColumnIdx = focusedHeaderColumnIdx < columns.length - 1 ? focusedHeaderColumnIdx + 1 : focusedHeaderColumnIdx;
          }

          if (newFocusedHeaderColumnIdx != focusedHeaderColumnIdx)
          {
            this._setHeaderColumnFocus(newFocusedHeaderColumnIdx, true, false, null);

            if (event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              // if shift is also pressed then we need to select too
              var newFocusedHeaderColumnSelection = this._getHeaderColumnSelection(newFocusedHeaderColumnIdx);
              // we may be clearing or setting the selection
              this._setHeaderColumnSelection(newFocusedHeaderColumnIdx, !newFocusedHeaderColumnSelection, null, event, true);
              // if we are clearing the selection, then clear the previous column too.
              if (newFocusedHeaderColumnSelection)
              {
                if (this._getHeaderColumnSelection(focusedHeaderColumnIdx))
                {
                  this._setHeaderColumnSelection(focusedHeaderColumnIdx, false, null, event, true);
                }
              }
            }
          }
        }
      },
      /**
       * Handler for Spacebar keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownSpacebar: function(event)
      {
        if (this._isTableActionableMode())
        {
          // ignore in actionable mode
          return;
        }
        // pressing spacebar selects the focused row/column
        var focusedRowIdx = this._getFocusedRowIdx();

        if (focusedRowIdx != null)
        {
          this._setRowSelection(focusedRowIdx, !this._getRowSelection(focusedRowIdx), null, event, true);
        }
        else if (!this._hasEditableRow())
        {
          var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();
          if (focusedHeaderColumnIdx != null)
          {
            this._clearSelectedRows();
            this._setHeaderColumnSelection(focusedHeaderColumnIdx, !this._getHeaderColumnSelection(focusedHeaderColumnIdx), null, event, true);
          }
        }
      },
      /**
       * Handler for Tab keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownTab: function(event)
      {
        // if Tab is pressed while a row has focus and we are in actionable/editable 
        // mode then want to Tab within that row until Esc or F2 is pressed
        var tabHandled = false;
        var focusedRowIdx = this._getFocusedRowIdx();
        var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();

        if ((focusedRowIdx != null ||
          focusedHeaderColumnIdx != null) &&
          (this._isTableActionableMode() ||
            this._hasEditableRow()))
        {
          var currentFocusElement = document.activeElement;
          var tableBody = this._getTableDomUtils().getTableBody();
          var tableHeader = this._getTableDomUtils().getTableHeader();

          if (this._getEditableRowIdx() === focusedRowIdx)
          {
            // If we are on an editable row and there are no more editable
            // elements to focus to then go to the next row
            var tableBodyRow = this._getTableDomUtils().getTableBodyRow(focusedRowIdx);
            var tabbableElementsInRow = this._getTabbableElements(tableBodyRow);
            var tabbableElementsInRowCount = tabbableElementsInRow != null ? tabbableElementsInRow.length : 0;
            var rowElementTabIndex = $(tabbableElementsInRow).index(currentFocusElement);
            
            if (rowElementTabIndex == tabbableElementsInRowCount - 1
                && !event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              // last tabbable element in row so go to the next row
              this._setNextRowEditable(0, event);
              event.preventDefault();
              event.stopPropagation();
              return;
            }
            else if (rowElementTabIndex == 0 
                     && event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              // first tabbable element in row and Shift+Tab so go to the previous row
              var tableHeaderColumns = this._getTableDomUtils().getTableHeaderColumns();
              if (tableHeaderColumns)
              {
                var tableHeaderColumnsCount = tableHeaderColumns.length;
                this._setPreviousRowEditable(tableHeaderColumnsCount - 1, event);
                event.preventDefault();
                event.stopPropagation();
                return;
              }
            }
            else
            {
              return;
            }
          }
          else if ($.contains(tableBody, currentFocusElement) ||
            $.contains(tableHeader, currentFocusElement))
          {
            var focusedElement = null;
            
            if ($.contains(tableBody, currentFocusElement)) 
            {
              focusedElement = this._getTableDomUtils().getTableBodyRow(focusedRowIdx);
            }
            else if ($.contains(tableHeader, currentFocusElement)) 
            {
              focusedElement = this._getTableDomUtils().getTableHeader();
            }
            var tabbableElementsInFocusedElement = this._getTabbableElements(focusedElement);
            
            if (tabbableElementsInFocusedElement.length > 0)
            {
              // If only one tabbable element then stay on it
              if (tabbableElementsInFocusedElement.length > 1)
              {
                if (!event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
                {
                  // Tabbing on the last tabbable element will wrap back
                  var lastTabbableElementFocusedElement = tabbableElementsInFocusedElement[tabbableElementsInFocusedElement.length - 1];

                  if (currentFocusElement == lastTabbableElementFocusedElement)
                  {
                    $(tabbableElementsInFocusedElement[0]).focus();
                    event.preventDefault();
                    event.stopPropagation();
                  }
                  else
                  {
                    // find which element it is
                    var i;
                    for (i = 0; i < tabbableElementsInFocusedElement.length; i++)
                    {
                      if (currentFocusElement == tabbableElementsInFocusedElement[i])
                      {
                        tabbableElementsInFocusedElement[i + 1].focus();
                        event.preventDefault();
                        event.stopPropagation();
                        break;
                      }
                    }
                  }
                }
                else
                {
                  // Shift+Tabbing on the first tabbable element in a row will wrap back
                  var firstTabbableElementFocusedElement = tabbableElementsInFocusedElement[0];

                  if (currentFocusElement == firstTabbableElementFocusedElement)
                  {
                    $(tabbableElementsInFocusedElement[tabbableElementsInFocusedElement.length - 1]).focus();
                    event.preventDefault();
                    event.stopPropagation();
                  }
                  else
                  {
                    // find which element it is
                    var i;
                    for (i = 0; i < tabbableElementsInFocusedElement.length; i++)
                    {
                      if (currentFocusElement == tabbableElementsInFocusedElement[i])
                      {
                        tabbableElementsInFocusedElement[i - 1].focus();
                        event.preventDefault();
                        event.stopPropagation();
                        break;
                      }
                    }
                  }
                }
              }
              else
              {
                event.preventDefault();
                event.stopPropagation();
              }
              return;
            }
          }

          if (focusedRowIdx != null && !this._hasEditableRow() ||
            focusedHeaderColumnIdx != null)
          {
            if (!event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              tabHandled = true;
              var focusedElement = null;
              if (focusedRowIdx != null) 
              {
                focusedElement = this._getTableDomUtils().getTableBodyRow(focusedRowIdx);
              } 
              else if (focusedHeaderColumnIdx != null) 
              {
                focusedElement = this._getTableDomUtils().getTableHeaderColumn(focusedHeaderColumnIdx);
              }
              var tabbableElementsInFocusedElement = this._getTabbableElements(focusedElement);

              if (tabbableElementsInFocusedElement != null)
              {
                $(tabbableElementsInFocusedElement[0]).focus();
              }
              else
              {
                if (focusedRowIdx != null) 
                {
                  // if there are no tabbable elements
                  // in the row then focus on the first
                  // tabbable element in the body
                  var tabbableElementsInBody = this._getTabbableElements(tableBody);
                  $(tabbableElementsInBody[0]).focus();
                } 
                else if (focusedHeaderColumnIdx != null) 
                {
                  // if there are no tabbable elements
                  // in the column then focus on the first
                  // tabbable element in the thead
                  var tabbableElementsInHeader = this._getTabbableElements(tableHeader);
                  $(tabbableElementsInHeader[0]).focus();
                }
              }
              event.preventDefault();
              event.stopPropagation();
            }
          }
        } 

        if (!tabHandled)
        {
          // tab out of the component to the next tabbable
          // element on the page
          // clear all highlights etc, because we are tabbing out of the table
          this._clearKeyboardKeys();
          this._clearFocusedHeaderColumn();
          this._clearFocusedRow(false);
          this._setTableActionableMode(false);
            
          var table = this._getTableDomUtils().getTable();
          var tabbableElementsInTable = this._getTabbableElements(table, true);
          var tabbableElementsInTableCount = tabbableElementsInTable != null ? tabbableElementsInTable.length : 0;
          
          // Table should not set focus to elements outside of itself because
          // ancestor such as modal dialog may want to confine the focus.  It
          // should just set focus to itself in backward tabbing or to its last
          // tabbable descendant in forward tabbing.  Then allow the ancestors
          // to handle the event and determine the next tabstop.
          var temporaryFocus = table;
          
          if (!event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
          {
            // If we are tabbing forward and the table contains tabbable 
            // elements, focus on the last tabbable element.
            if (tabbableElementsInTableCount > 0)
            {
              if (!this._getTableDomUtils()._isFF()) 
              {
                temporaryFocus = tabbableElementsInTable[tabbableElementsInTableCount - 1];
              }
              else
              {
                // workaround for FF. In FF, the tbody is focusable even though it has
                // tabindex -1. So we have to create a temporary focusable element
                // as the last element in the table container and focus on that one.
                var tableContainer = this._getTableDomUtils().getTableContainer();
                temporaryFocus = document.createElement(oj.TableDomUtils.DOM_ELEMENT._A);
                temporaryFocus.setAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX, '0');
                temporaryFocus.setAttribute(oj.TableDomUtils.DOM_ATTR._HREF, '#');
                temporaryFocus.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
                tableContainer.appendChild(temporaryFocus);
                this._temporaryFocusForFF = temporaryFocus;
              }
            }            
          }

          // Need to set this variable because the focus() call will
          // trigger a focusin which we do not want to handle.
          this._temporaryTableChildElementFocus = true;
          temporaryFocus.focus();
          var self = this;
          setTimeout(function()
          {
            self._temporaryTableChildElementFocus = false;
            if (self._temporaryFocusForFF)
            {
              $(self._temporaryFocusForFF).remove();
              self._temporaryFocusForFF = null;
            }
          }, 0);

          return;
        }
        // we need to remove Tab on keydown because we may not
        // get a keyup for it if focus moves
        // outside of table
        this._removeKeyboardKey(event.keyCode);
      },
      /**
       * Handler for Up/Down keydown.
       * @param {Object} event
       * @private
       */
      _handleKeydownUpDown: function(event)
      {
        if (this._isTableActionableMode())
        {
          // ignore in actionable mode
          return;
        }
        var focusedRowIdx = this._getFocusedRowIdx();
        var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();

        if (focusedRowIdx != null && !this._hasEditableRow())
        {
          // if row is focused then up/down navigates the rows
          var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
          var rowCount = tableBodyRows != null ? tableBodyRows.length : 0;
          var newFocusedRowIdx = focusedRowIdx;

          if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_UP))
          {
            if (focusedRowIdx > 0)
            {
              newFocusedRowIdx = focusedRowIdx - 1;
          }
          else if (!this._hasEditableRow())
          {
              newFocusedRowIdx = focusedRowIdx;
            }
          }
          else if (this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_DOWN))
          {
            newFocusedRowIdx = focusedRowIdx < rowCount - 1 ? focusedRowIdx + 1 : focusedRowIdx;
          }

          if (newFocusedRowIdx != focusedRowIdx)
          {
            var focused = this._setRowFocus(newFocusedRowIdx, true, true, null, event);

            if (!focused)
            {
              return;
            }
            this._getTableDomUtils().getTable().focus();

            if (event[this._KEYBOARD_CODES._KEYBOARD_MODIFIER_SHIFT])
            {
              // if shift is also pressed then we need to select too
              var newFocusedRowSelection = this._getRowSelection(newFocusedRowIdx);
              // we may be clearing or setting the selection
              this._setRowSelection(newFocusedRowIdx, !newFocusedRowSelection, null, event, true);
              // if we are clearing the selection, then clear the previous row too.
              if (newFocusedRowSelection)
              {
                if (this._getRowSelection(focusedRowIdx))
                {
                  this._setRowSelection(focusedRowIdx, false, null, event, true);
                }
              }
            }
          }
          // if user is on the first row and presses up the focus on the first column header
          else if (newFocusedRowIdx == focusedRowIdx &&
            focusedRowIdx == 0 &&
            this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_UP))
          {
            this._setHeaderColumnFocus(0, true, false, null);
          }
        }
        // if user is on a column header and pressed down then focus on the first row
        else if (focusedHeaderColumnIdx != null &&
          this._isKeyboardKeyPressed(this._KEYBOARD_CODES._KEYBOARD_CODE_DOWN))
        {
          this._setRowFocus(0, true, true, null, event);
        }
      },
      /**
       * Callback handler mouse move for selection.
       * @private
       */
      _handleMouseEnterSelection: function(element)
      {
        var rowIdx = this._getTableDomUtils().getElementRowIdx(element);
        
        if (rowIdx != null && 
            this._mouseDownRowIdx != null && 
            this._mouseDownRowIdx != rowIdx)
        {
          var lastSelectedRowIdx = this._mouseDownRowIdx;
          var selectedRowIdxs = this._getSelectedRowIdxs();
          
          if (rowIdx < lastSelectedRowIdx)
          {
            var i;
            for (i = 0; i < selectedRowIdxs.length; i++)
            {
              if (selectedRowIdxs[i] < rowIdx ||
                  selectedRowIdxs[i] > lastSelectedRowIdx)
                {
                  this._setRowSelection(selectedRowIdxs[i], false, element, null, true);
                }
            }
            for (i = lastSelectedRowIdx; i >= rowIdx; i--)
            {
              this._setRowSelection(i, true, element, null, true);
              this._getTableDomUtils().moveTableBodyRowTouchSelectionAffordanceTop(i);
            }
          }
          else
          {
            var i;
            for (i = 0; i < selectedRowIdxs.length; i++)
            {
              if (selectedRowIdxs[i] > rowIdx ||
                  selectedRowIdxs[i] < lastSelectedRowIdx)
                {
                  this._setRowSelection(selectedRowIdxs[i], false, element, null, true);
                }
            }
            for (i = lastSelectedRowIdx; i <= rowIdx; i++)
            {
              this._setRowSelection(i, true, element, null, true);
              this._getTableDomUtils().moveTableBodyRowTouchSelectionAffordanceBottom(i);
            }
          }
        }
      },
      /**
       * Callback handler max fetch count.
       * @private
       */
      _handleScrollerMaxRowCount: function()
      {
        var errSummary = this.getTranslatedString('msgScrollPolicyMaxCountSummary');
        var errDetail = this.getTranslatedString('msgScrollPolicyMaxCountDetail');
        this._showInlineMessage(errSummary, errDetail, oj.Message.SEVERITY_LEVEL['WARNING']);
      },
     /**
       * Handle scrollLeft on scroller
       * @private
       */
      _handleScrollerScrollLeft: function(scrollLeft)
      {
        this._scrollLeft = scrollLeft;
        
          var tableHeaderRow = this._getTableDomUtils().getTableHeaderRow();
          if (tableHeaderRow)
          {
            if (this._GetReadingDirection() === "rtl")
            {
            tableHeaderRow.style[oj.TableDomUtils.CSS_PROP._RIGHT] = '-' + scrollLeft + oj.TableDomUtils.CSS_VAL._PX;
            }
            else
            {
            tableHeaderRow.style[oj.TableDomUtils.CSS_PROP._LEFT] = '-' + scrollLeft + oj.TableDomUtils.CSS_VAL._PX;
            }
          }

          var tableFooterRow = this._getTableDomUtils().getTableFooterRow();
          if (tableFooterRow)
          {
            if (this._GetReadingDirection() === "rtl")
            {
            tableFooterRow.style[oj.TableDomUtils.CSS_PROP._RIGHT] = '-' + scrollLeft + oj.TableDomUtils.CSS_VAL._PX;
            }
            else
            {
            tableFooterRow.style[oj.TableDomUtils.CSS_PROP._LEFT] = '-' + scrollLeft + oj.TableDomUtils.CSS_VAL._PX;
            }
          }
      },
      /**
       * Handle scrollTop on scroller
       * @private
       */
      _handleScrollerScrollTop: function(scrollTop)
      {
        this._scrollTop = scrollTop;
      },
      /**
       * Handler for column sort
       * @param {number} columnIdx  column index
       * @param {boolean} ascending  sort order ascending
       * @param {Object} event
       * @private
       */
      _handleSortTableHeaderColumn: function(columnIdx, ascending, event)
      {
        // clear the sorted indicator on any other column
        this._clearSortedHeaderColumn(columnIdx);
        // get the column metadata
        var column = this._getColumnDefs()[columnIdx];
        // get which field to sort on
        var sortField = column['sortProperty'] == null ? column['field'] : column['sortProperty'];
        // invoke sort on the data
        this._invokeDataSort(sortField, ascending, event);
        this._sortColumn = column;
        this._refreshSortTableHeaderColumn(sortField, ascending);
      },
      /**
       * Return whether the table currently has an editable row
       * @return {boolean} true or false
       * @private
       */
      _hasEditableRow: function()
      {
        if (!this._isTableEditMode())
        {
          return false;
        }
        
        return this._getEditableRowIdx() !== null;
      },
      /**
       * Has row or cell renderer or template
       * @param {number|null} columnIdx  column index
       * @return {boolean} true or false
       * @private
       */
      _hasRowOrCellRendererOrTemplate: function(columnIdx)
      {
        var rowRenderer = this._getRowRenderer();
        
        if (rowRenderer != null)
        {
          return true;
        }
        else
        {
          var cellRenderer = null;
          var cellSlotTemplate = null;
          var columns = this._getColumnDefs();
          
          if (columnIdx != null)
          {
            cellRenderer = this._getColumnRenderer(columnIdx, 'cell');
            cellSlotTemplate = this._getSlotTemplate(columns[columnIdx][this._CELL_TEMPLATE]);
          }
          else
          {
            var i, columnsCount = columns.length;
            
            for (i = 0; i < columnsCount; i++)
            {
              cellRenderer = this._getColumnRenderer(i, 'cell');
              
              if (cellRenderer != null)
              {
                break;
              }
              
              cellSlotTemplate = this._getSlotTemplate(columns[i][this._CELL_TEMPLATE]);
        
              if (cellSlotTemplate != null)
              {
                break;
              }
            }
          }
          if (cellRenderer != null || cellSlotTemplate != null)
          {
            return true;
          }
        }
        return false;
      },
      /**
       * Hide the inline message.
       * @private
       */
      _hideInlineMessage: function()
      {
        if (this._inlineMessageShown)
        {
        var inlineMessage = this._getTableDomUtils().getTableInlineMessage();
          var tableContainer = this._getTableDomUtils().getTableContainer();
          tableContainer.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._WARNING);
          inlineMessage.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._WARNING);
          inlineMessage.style[oj.TableDomUtils.CSS_PROP._BOTTOM] = '';
          inlineMessage.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._NONE;
          this._inlineMessageShown = false;
        }
      },
      /**
       * Hide the 'No data to display.' message.
       * @private
       */
      _hideNoDataMessage: function()
      {
        if (this._noDataMessageShown)
        {
        var tableBodyMessageRow = this._getTableDomUtils().getTableBodyMessageRow();
        if (tableBodyMessageRow != null) {
            $(tableBodyMessageRow).remove();
        }
          this._noDataMessageShown = false;
        }
      },
      /**
       * Hide the Fetching Data... status message.
       * @private
       */
      _hideStatusMessage: function()
      {
        if (this._showStatusTimeout)
        {
          clearTimeout(this._showStatusTimeout);
          this._showStatusTimeout = null;
        }
        if (this._statusMessageShown)
        {
        var statusMessage = this._getTableDomUtils().getTableStatusMessage();
          statusMessage.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._NONE;
          this._statusMessageShown = false;
        }
      },
      /**
       * Hide the column header sort link
       * @param {number} columnIdx  column index
       * @param {boolean} ascending  sort order ascending
       * @private
       */
      _hideTableHeaderColumnSortLink: function(columnIdx, ascending)
      {
        // check if the column is sortable. If not, then there won't be any sort links
        if (this._getColumnDefs()[columnIdx]['sortable'] == this._OPTION_ENABLED)
        {
          var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);
          // check if the column is currently sorted
          var sorted = $(tableHeaderColumn).data('sorted');

          // we should only hide the ascending sort link if the column is not sorted or
          // is sorted by descending order
          if (ascending && (sorted == null || sorted == this._COLUMN_SORT_ORDER._DESCENDING))
          {
            var headerColumnAscLink = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_LINK_CLASS);
            if (headerColumnAscLink.length > 0)
            {
              headerColumnAscLink = headerColumnAscLink[0];
              headerColumnAscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
              headerColumnAscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._ENABLED);
              headerColumnAscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
          }
          }
          // we should only hide the descending sort link if the column is not sorted or
          // is sorted by ascending order
          else if (!ascending && (sorted == null || sorted == this._COLUMN_SORT_ORDER._ASCENDING))
          {
            var headerColumnDscLink = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_LINK_CLASS);
            if (headerColumnDscLink.length > 0)
            {
              headerColumnDscLink = headerColumnDscLink[0];
              headerColumnDscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
              headerColumnDscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._ENABLED);
              headerColumnDscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
          }
        }
        }
      },
      /**
       * Do an initial fetch
       * @param {Object} options options for the fetch
       * @param {boolean} force whether to force fetch
       * @private
       */
      _initFetch: function(options, force)
      {
        options = options || {};
        var self = this;
        var dataprovider = this._getData();
        // do an initial fetch if a TableDataSource
        // paging control should do the fetches for PagingTableDataSource
        if (dataprovider != null && 
            oj.DataProviderFeatureChecker.isIteratingDataProvider(dataprovider) &&
            (!this._isPagingModelDataProvider()
            || force))
        {
          return this._queueTask(function()
          {
            // reset the scrollTop when we do an initial fetch
            self._getTableDomUtils().getScroller().scrollTop = 0;
            if (dataprovider instanceof oj.TableDataSourceAdapter)
            {
              options['fetchType'] = 'init';
              
              if (self._isLoadMoreOnScroll())
              {
                options[self._CONST_OFFSET] = 0;
              }
            }
            return self._invokeDataFetchRows(options, true);
          });
        }
        else if (dataprovider == null)
        {
          return this._queueTask(function()
          {
            return Promise.resolve();
          });
        }
      },
      /**
       * Initialize the template engine
       * @private
       */
      _initTemplateEngine: function()
      {
       // initialize the template engine
        var self = this;
        this._queueTask(function()
        {
          try 
          {
            var enginePromise = oj.Config.__getTemplateEngine().then(
              function(engine)
              {
                self._templateEngine = engine;
              },
              function(err)
              {
                oj.Logger.warn(err);
              }
            );
          } 
          catch (err) 
          {
            oj.Logger.warn(err);
            enginePromise = Promise.resolve(null);
          }
          return enginePromise;
        });
      },
      /**
       * Fetch rows
       * @param {Object} options options for the fetch
       * @param {boolean} init initial fetch
       * @return {Promise} Promise resolves with the result when done.
       * @private
       */
      _invokeDataFetchRows: function(options, init)
      {
        options = options || {};
        if (!options[this._CONST_PAGESIZE] && this._isLoadMoreOnScroll())
        {
          options[this._CONST_PAGESIZE] = this.options['scrollPolicyOptions']['fetchSize'];
        }
        else
        {
          options[this._CONST_PAGESIZE] = -1;
        }
        var dataprovider = this._getData();
        if (dataprovider != null)
        {
          var self = this;
          return new Promise(function(resolve, reject)
          {
            self._setDataWaitingState();
            self._dataProviderAsyncIterator = dataprovider.fetchFirst(options)[Symbol.asyncIterator]();
            self._dataProviderAsyncIterator.next().then(function(result)
            {
              var value = result[self._CONST_VALUE];
              var data = value[self._CONST_DATA];
              var keys = value.metadata.map(function(value) {
                return value[self._CONST_KEY];
              });
              
              var offset = 0;
              if (dataprovider instanceof oj.TableDataSourceAdapter) 
              {
                offset = dataprovider[self._CONST_OFFSET];
              }
              
              var startIndex = 0;
              if (self._isPagingModelDataProvider()) 
              {
                startIndex = dataprovider.getStartItemIndex();
              }
              
              var indexArray = [];
              var i, resultDataCount = data.length;

              for (i = 0; i < resultDataCount; i++)
              {
                indexArray[i] = offset + startIndex + i;
              }

              var metadataArray = [];

              self._refreshAll({'data': data, 'metadata': metadataArray, 'keys' : keys, 'indexes': indexArray}, offset, init, init);
              self._clearDataWaitingState();
              self._processFetchSort(value);
              if (self._isLoadMoreOnScroll())
              {
                self._registerDomScroller();
              }
              self = null;
              resolve(result);
            }, function(error) {
              self._clearDataWaitingState();
              var tableBodyRows = self._getTableDomUtils().getTableBodyRows();
              if (tableBodyRows == null || tableBodyRows.length == 0)
              {
                self._showNoDataMessage();
              }
              resolve(null);
            });
          });
        }
        else
        {
          return Promise.resolve(null);
        }
      },
      /**
       * Invoke sort on a field. This function is called when a user clicks the
       * column header sort links
       * @param {string} sortField  field name
       * @param {boolean} ascending  sort order ascending
       * @param {Object} event
       * @private
       */
      _invokeDataSort: function(sortField, ascending, event)
      {
        var dataprovider = this._getData();
        // if no data then bail
        if (!dataprovider)
        {
          return null;
        }

        // show the Fetching Data... message
        this._showStatusMessage();

        var sortCriteria = [];
        var sortCriterion = {};
        sortCriterion[this._CONST_ATTRIBUTE] = sortField;
        if (ascending)
        {
          sortCriterion['direction'] = this._COLUMN_SORT_ORDER._ASCENDING;
        }
        else
        {
          sortCriterion['direction'] = this._COLUMN_SORT_ORDER._DESCENDING;
        }
        sortCriteria.push(sortCriterion);
        this._trigger('sort', event, {'header': sortCriteria[0][this._CONST_ATTRIBUTE], 'direction': sortCriteria[0]['direction']});
        this._initFetch({'sortCriteria': sortCriteria}, true);
      },
      /**
       * Whether the columns have been updated
       * @return {boolean} true or false
       * @private
       */
      _isColumnMetadataUpdated: function()
      {
        if (this._columnDefArray != null)
        {
          var columnsMetadata = this._getColumnMetadata();
          if (this._columnDefArray.length != columnsMetadata.length)
          {
            return true;
          }
          else
          {
            var i, prop, columnsMetadataCount = columnsMetadata.length;
            for (i = 0; i < columnsMetadataCount; i++)
            {
              for (prop in columnsMetadata[i]) {
                if (columnsMetadata[i].hasOwnProperty(prop)) {
                  if (columnsMetadata[i][prop] != this._columnDefArray[i][prop])
                  {
                    if (prop != 'id' ||
                        columnsMetadata[i][prop] == null ||
                        columnsMetadata[i][prop].indexOf(this._COLUMN_HEADER_ID_PREFIX) != 0 ||
                        this._columnDefArray[i][prop] == null ||
                        this._columnDefArray[i][prop].indexOf(this._COLUMN_HEADER_ID_PREFIX) != 0)
                    {
                      // ignore generated ids
                      return true;
                    }
                  }
                }
              }
            }
          }
          return false;
        }
        return true;
      },
      /**
       * Is keybaord key pressed
       * @param {number} keyCode  KeyCode of the keyboard key.
       * @return {boolean} true or false
       * @private
       */
      _isKeyboardKeyPressed: function(keyCode)
      {
        var keyboardKeys = this._getKeyboardKeys();
        var i, keyboardKeysCount = keyboardKeys.length;
        for (i = 0; i < keyboardKeysCount; i++)
        {
          if (keyboardKeys[i] == keyCode)
          {
            return true;
          }
        }
        return false;
      },
      /**
       * Is loadMoreOnScroll
       * @return {boolean} true or false
       * @private
       */
      _isLoadMoreOnScroll: function()
      {
        return this.options['scrollPolicy'] == this._OPTION_SCROLL_POLICY._LOADMORE_ON_SCROLL ? true: false;
      },
      /**
       * Return whether the node is editable
       * @param {Node} node  DOM Node
       * @return {boolean} true or false
       * @private
       */
      _isNodeEditable: function(node)
      {
        return this._isNodeType(node, /^INPUT|TEXTAREA/);
      },
      /**
       * Return whether the node is clickable
       * @param {Node} node  DOM Node
       * @return {boolean} true or false
       * @private
       */
      _isNodeClickable: function(node)
      {
        return this._isNodeType(node, /SELECT|OPTION|BUTTON|^A\b/);
      },
      /**
       * Return whether the node or any of its ancestors is draggable
       * @param {Node} node  DOM Node
       * @return {boolean} true or false
       * @private
       */
      _isNodeDraggable: function(node)
      {
        return ($(node).closest("[draggable='true']").length > 0);
      },
      /**
       * Return whether the node is editable or clickable
       * @param {Node} node  DOM Node
       * @param {Object} type regex
       * @return {boolean} true or false
       * @private
       */
      _isNodeType: function(node, type)
      {
        var nodeName;
        var table = this._getTableDomUtils().getTable();

        while (null != node && node != table &&
          (nodeName = node['nodeName']) != oj.TableDomUtils.DOM_ELEMENT._TD && nodeName != oj.TableDomUtils.DOM_ELEMENT._TH)
        {
          // If the node is a text node, move up the hierarchy to only operate on elements
          // (on at least the mobile platforms, the node may be a text node)
          if (node.nodeType == 3) // 3 is Node.TEXT_NODE
          {
            node = node.parentNode;
            continue;
          }

          var tabIndex = node.getAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX);

          if (nodeName.match(type))
          {
            // ignore elements with tabIndex == -1
            if (tabIndex != -1)
            {
              return true;
            }
          }
          node = node.parentNode;
        }
        return false;
      },
      /**       
       * Return whether this is a PagingModel dataprovider
       * @returns {boolean}
       * @private
       */
      _isPagingModelDataProvider: function()
      {
        var dataprovider = this._getData();
        if (dataprovider.getStartItemIndex != null &&
          dataprovider.getStartItemIndex() !== null &&
          dataprovider.getStartItemIndex() >= 0)
        {
          return true;
        }
        return false;
      },
      /**
       * Return whether the status message is shown
       * @return {boolean} true or false
       * @private
       */
      _isStatusMessageShown: function()
      {
        return this._statusMessageShown;
      },
      /**
       * Return whether the component is in table actionable mode
       * @return {boolean} true or false
       * @private
       */
      _isTableActionableMode: function()
      {
        return this._tableActionableMode;
      },
      /**
       * Returns whether the table is editabe mode
       * @return {boolean} true or false
       * @private
       */
      _isTableEditMode: function()
      { 
        var editMode = this['options']['editMode'];
        
        if (editMode == this._OPTION_EDIT_MODE._ROW_EDIT)
        {
          return true;
        }
        return false;
      },
      /**
       * Returns whether the table is footerless
       * @return {boolean} true or false
       * @private
       */
      _isTableFooterless: function()
      {
        var columns = this._getColumnDefs();
        var i, footerRenderer, footerSlotTemplate, columnsCount = columns.length;

        for (i = 0; i < columnsCount; i++)
        {
          footerRenderer = this._getColumnRenderer(i, 'footer');
          footerSlotTemplate = this._getSlotTemplate(columns[i][this._FOOTER_TEMPLATE]);
          
          if (footerRenderer != null)
          {
            return false
          }
          if (footerSlotTemplate != null)
          {
            return false
          }
        }
        return true;
      },
      /**
       * Returns whether the table is headerless
       * @return {boolean} true or false
       * @private
       */
      _isTableHeaderless: function()
      {
        var columns = this._getColumnDefs();

        var i, j, headerSlotTemplate, columnsCount = columns.length;
        for (i = 0; i < columnsCount; i++)
        {
          headerSlotTemplate = this._getSlotTemplate(columns[i][this._HEADER_TEMPLATE]);
          
          if (columns[i]['headerText'] != null ||
            columns[i]['headerStyle'] != null ||
            (columns[i]['sortable'] != null &&
              columns[i]['sortable'] != this._OPTION_DISABLED) ||
            columns[i]['sortProperty'] != null ||
            columns[i]['headerRenderer'] != null ||
            headerSlotTemplate != null)
          {
            return false;
          }
        }

        return true;
      },
      /**
       * Returns whether the table header columns were rendered
       * @return {boolean} true or false
       * @private
       */
      _isTableHeaderColumnsRendered: function()
      {
        return this._renderedTableHeaderColumns == true;
      },
      /**
       * Returns whether the table refresh is needed based on option change
       * @param {string} key option key
       * @param {Object} value option value
       * @return {boolean} true or false
       * @private
       */
      _isTableRefreshNeeded: function(key, value)
      {
        var currentOptions = this._cachedOptions;
        var refresh = false;

        if (key == 'contextMenu' && value == '#' + this._getTableDomUtils().getTableId() + '_contextmenu')
        {
          refresh = false;
        }
        else if (key == 'columns' && !this._isColumnMetadataUpdated())
        {
          // optimization for columns. Column re-order can change the columns options but we don't need to refresh
          refresh = false;
        }
        else if (key != 'selection' && key != 'currentRow' && !oj.Object.compareValues(value, currentOptions[key]))
        {
          refresh = true;
        }

        this._cachedOptions = $.extend(true, {}, this.options);

        return refresh;
      },
      /**
       * Returns whether any of the table columns are sortable
       * @return {boolean} true or false
       * @private
       */
      _isTableSortable: function()
      {
        var columns = this._getColumnDefs();

        var i, columnsCount = columns.length;
        for (i = 0; i < columnsCount; i++)
        {
          if (columns[i]['sortable'] == this._OPTION_ENABLED)
          {
            return true;
          }
        }

        return false;
      },
      /**
       * Returns whether any of the table columns are resizable
       * @return {boolean} true or false
       * @private
       */
      _isTableColumnsResizable: function()
      {
        var columns = this._getColumnDefs();

        var i, columnsCount = columns.length;
        for (i = 0; i < columnsCount; i++)
        {
          if (columns[i]['resizable'] == this._OPTION_ENABLED)
          {
            return true;
          }
        }

        return false;
      },
      /**
       * Returns whether any of the table columns have width specified
       * @return {boolean} true or false
       * @private
       */
      _isTableColumnsWidthSet: function()
      {
        var columns = this._getColumnDefs();

        var i, columnsCount = columns.length;
        for (i = 0; i < columnsCount; i++)
        {
          if (columns[i]['width'] !== null)
          {
            return true;
          }
        }

        return false;
      },
      _isTouchDevice: function()
      {
        if (this._isTouch == undefined)
        {
          var agentName = navigator.userAgent.toLowerCase();
          if (agentName.indexOf("mobile") != -1 || agentName.indexOf("android") != -1)
          {
            this._isTouch = true;
          }
          else
          {
            this._isTouch = false;
          }
        }
        return this._isTouch;
      },
      /**
       * Process any sort from the fetch
       * @param {Object} result result of the fetch
       * @private
       */
      _processFetchSort: function(result)
      {
        try
        { 
          var self = this;
          var fetchParameters = result.fetchParameters;
          var columns = this._getColumnDefs();
          var sortCriteria = fetchParameters[this._CONST_SORTCRITERIA];
          if (sortCriteria != null && 
            sortCriteria.length > 0) 
          {
            var sortcriterion = sortCriteria[0];
            this._refreshSortTableHeaderColumn(sortcriterion['attribute'], sortcriterion['direction'] == this._COLUMN_SORT_ORDER._ASCENDING);

            // clear selection if not single selection
            var existingSelection = this['options']['selection'];
            if (existingSelection != null)
            {
              var clearSelection = false;

              if (existingSelection.length > 1)
              {
                clearSelection = true;
              }
              else if (existingSelection[0] != null)
              {
                var startIndex = existingSelection[0][this._CONST_STARTINDEX];
                var endIndex = existingSelection[0][this._CONST_ENDINDEX];

                if (!oj.Object.compareValues(startIndex, endIndex) && endIndex != null)
                {
                  clearSelection = true;
                }
              }
              if (clearSelection)
              {
                this._setSelection(null);
                this.option('selection', [], {'_context': {writeback: true, internalSet: true}});
              }
            }
            // set the current row
            this._setCurrentRow(this.options['currentRow'], null, false);
            var columnIdx = null;
            var i, column, sortField, columnsCount = columns.length;

            for (i = 0; i < columnsCount; i++)
            {
              column = columns[i];
              sortField = column['sortProperty'] == null ? column['field'] : column['sortProperty'];

              if (sortcriterion['attribute'] == sortField)
              {
                columnIdx = i;
                break;
              }
            }
            if (columnIdx != null)
            {
              setTimeout(function()
              {
                self._scrollColumnIntoViewport(columnIdx);
                self = null;
                columnIdx = null;
              }, 0);
            }
          }
          else
          {
            this._clearSortedHeaderColumn();
          }
        }
        catch (e)
        {
          oj.Logger.error(e);
        }
      },
      /**
       * Process any slotted children and move them into the correct location
       * @private
       */
      _processSlottedChildren: function()
      {
        var tableBottomSlot = this._getTableDomUtils().getTableBottomSlot();
        
        if (tableBottomSlot != null)
        {
          // clear the existing bottom slot
          $(tableBottomSlot).remove();
        }
        var slotMap = oj.BaseCustomElementBridge.getSlotMap(this._getRootElement());
        
        for (var slot in slotMap) 
        {
          if (slotMap.hasOwnProperty(slot) && slot.length > 0) 
          {
            if (slot === "bottom")
            {
              var slotElements = slotMap[slot];

              if (slotElements != null)
              {
                tableBottomSlot = this._getTableDomUtils().createTableBottomSlot();
                var i;
                for (i = 0; i < slotElements.length; i++)
                {
                  tableBottomSlot.appendChild(slotElements[i]);
                }
              }
            }
          }
        }
      },
      /**
       * @param {Object} resultObject Object containing data array, key array, and index array
       * @param {number} startIndex start index
       * @param {boolean} resetScrollTop reset the scrollTop
       * @param {boolean} resetScrollLeft reset the scrollLeft
       * @private
       */
      _refreshAll: function(resultObject, startIndex, resetScrollTop, resetScrollLeft)
      {
        if (this._isColumnMetadataUpdated() ||
            (!this._isTableHeaderColumnsRendered() &&
            !this._isTableHeaderless()))
        {
          this._clearCachedMetadata();
          this._refreshTableHeader();

          // see if we need to clear the sort. If the column we sorted on is no
          // longer there then clear it.
          if (this._sortColumn != null)
          {
            var i, column;
            var foundColumn = false;
            var columns = this._getColumnDefs();
            if (columns != null)
            {
              var columnsCount = columns.length;
              for (i = 0; i < columnsCount; i++)
              {
                column = columns[i];
                if (oj.Object.compareValues(column, this._sortColumn))
                {
                  foundColumn = true;
                  break;
                }
              }
              if (!foundColumn)
              {
                this._initFetch(null, true);
              }
            }
          }
        }
        this._refreshTableFooter();
        this._refreshTableBody(resultObject, startIndex, resetScrollTop, resetScrollLeft);
      },
      /**
       * Handler for column sort
       * @param {string} key sort key
       * @param {boolean} ascending  sort order ascending
       * @private
       */
      _refreshSortTableHeaderColumn: function(key, ascending)
      {
        var columns = this._getColumnDefs();
        var i, column, columnIdx = null, columnsCount = columns.length, sortField;

        for (i = 0; i < columnsCount; i++)
        {
          column = columns[i];
          sortField = column['sortProperty'] == null ? column['field'] : column['sortProperty'];

          if (key == sortField)
          {
            columnIdx = i;
            break;
          }
        }
        if (columnIdx == null)
        {
          return;
        }
                  
        // clear the sorted indicator on any other column
        this._clearSortedHeaderColumn(columnIdx);
        // get the column header DOM element
        var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);
        if (tableHeaderColumn == null)
        {
          return;
        }

        var sorted = $(tableHeaderColumn).data('sorted');

        if (ascending && sorted != this._COLUMN_SORT_ORDER._ASCENDING)
        {
          // store sort order on the DOM element
          $(tableHeaderColumn).data('sorted', this._COLUMN_SORT_ORDER._ASCENDING);
          var headerColumnAscLink = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_LINK_CLASS);
          if (headerColumnAscLink.length > 0)
          {
            headerColumnAscLink = headerColumnAscLink[0];
            headerColumnAscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._ENABLED);
            headerColumnAscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
          }
          var headerColumnAsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_CLASS);
          if (headerColumnAsc.length > 0)
          {
            headerColumnAsc = headerColumnAsc[0];
            headerColumnAsc.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
          }
          var headerColumnDsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_CLASS);
          if (headerColumnDsc.length > 0)
          {
            headerColumnDsc = headerColumnDsc[0];
            headerColumnDsc.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
          }
          this._hideTableHeaderColumnSortLink(columnIdx, !ascending);
        }
        else if (!ascending && sorted != this._COLUMN_SORT_ORDER._DESCENDING)
        {
          // store sort order on the DOM element
          $(tableHeaderColumn).data('sorted', this._COLUMN_SORT_ORDER._DESCENDING);
          var headerColumnDscLink = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_LINK_CLASS);
          if (headerColumnDscLink.length > 0)
          {
            headerColumnDscLink = headerColumnDscLink[0];
            headerColumnDscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._ENABLED);
            headerColumnDscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
        }
          var headerColumnDsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_CLASS);
          if (headerColumnDsc.length > 0)
          {
            headerColumnDsc = headerColumnDsc[0];
            headerColumnDsc.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
          }
          var headerColumnAsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_CLASS);
          if (headerColumnAsc.length > 0)
          {
            headerColumnAsc = headerColumnAsc[0];
            headerColumnAsc.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
          }
        }
      },
      /**
       * Refresh the entire table body with data from the datasource
       * @param {Object} resultObject Object containing data array, key array, and index array
       * @param {number} startIndex start index
       * @param {boolean} resetScrollTop reset the scrollTop
       * @param {boolean} resetScrollLeft reset the scrollLeft
       * @private
       */
      _refreshTableBody: function(resultObject, startIndex, resetScrollTop, resetScrollLeft)
      {
        var tableBody = this._getTableDomUtils().getTableBody();

        if (tableBody == null)
        {
          return;
        }

        var rows = this._getRowIdxRowArray(resultObject, startIndex);
        var checkFocus = $.contains(tableBody, document.activeElement);
        var tableBodyRows, tableBodyRow, resetFocus = false;

        if (startIndex == 0)
        {
          if (checkFocus)
          {
            resetFocus = true;
          }
          this._getTableDomUtils().removeAllTableBodyRows();
        }
        else
        {
          tableBodyRows = this._getTableDomUtils().getTableBodyRows();
          if (tableBodyRows != null && tableBodyRows.length > 0)
          {
            var i, tableBodyRowsCount = tableBodyRows.length;
            for (i = tableBodyRowsCount - 1; i >= startIndex; i--)
            {
              if (checkFocus)
              {
                tableBodyRow = this._getTableDomUtils().getTableBodyRow(i);

                if (tableBodyRow != null &&
                    $.contains(tableBodyRow, document.activeElement))
                {
                  resetFocus = true;
                  checkFocus = false;
                }
              }
              this._getTableDomUtils().removeTableBodyRow(i);
            }
          }
        }
        
        if (resetFocus)
        {
          this._getTableDomUtils().getTable().focus();
        }
        this._getTableDomUtils().clearCachedDomRowData();
        this._hideNoDataMessage();
        tableBodyRows = tableBody.children;

        // if no data then bail
        if (rows.length == 0 && (tableBodyRows == null || tableBodyRows.length == 0))
        {
          this._showNoDataMessage();
        }
        else
        {
          var tableBodyDocFrag = document.createDocumentFragment();
          var i, row, rowIdx, rowsCount = rows.length;
          for (i = 0; i < rowsCount; i++)
          {
            row = rows[i].row;
            rowIdx = rows[i].rowIdx;
            if (row != null)
            {
              tableBodyRow = this._getTableDomUtils().createTableBodyRow(rowIdx, row[this._CONST_KEY]);
              this._getTableDomUtils().styleTableBodyRow(tableBodyRow, true);
              this._getTableDomUtils().insertTableBodyRow(rowIdx, tableBodyRow, row, tableBodyDocFrag);
              this._refreshTableBodyRow(rowIdx, row, tableBodyRow, tableBodyDocFrag, startIndex, true);
            }
          }
          tableBody.appendChild(tableBodyDocFrag); //@HTMLUpdateOK
          this._getTableDomUtils().clearCachedDomRowData();
          // only bother calling subtree attached if there are potentially
          // components in our rows
          if (this._hasRowOrCellRendererOrTemplate())
          {
            this._getTableDomUtils().refreshTableDimensions(null, null, resetScrollTop, resetScrollLeft);
            oj.Components.subtreeAttached(tableBody);
          }
        }
      },
      /**
       * Refresh the row at a particular index with the row data
       * @param {number} rowIdx  row index relative to the start of the table
       * @param {Object} row  row and key object
       * @param {Object} tableBodyRow tr element
       * @param {Object} docFrag  document fragment
       * @param {number} docFragStartIdx  document fragment row start index
       * @param {boolean} isNew is new row
       * @return {Element|null} tableBodyRow  DOM element
       * @private
       */
      _refreshTableBodyRow: function(rowIdx, row, tableBodyRow, docFrag, docFragStartIdx, isNew)
      {
        var options = this.options;
        var rowRenderer = this._getRowRenderer();
        var columns = this._getColumnDefs();
        var tableBody = this._getTableDomUtils().getTableBody();

        if (isNaN(rowIdx) || rowIdx < 0)
        {
          // validate rowIdx value
          oj.Logger.error('Error: Invalid rowIdx value: ' + rowIdx);
        }

        var rowHashCode = this._getTableDomUtils().hashCode(row[this._CONST_KEY]);

        if (tableBodyRow == null)
        {
          // check if we already have a <tr> element at that index
          tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);
          if (!tableBodyRow)
          {
            // if not return
            return null;
          }
          else
          {
            $(tableBodyRow).empty();
            isNew = true;
            this._getTableDomUtils().createTableBodyCellAccSelect(rowIdx, row[this._CONST_KEY], rowHashCode, tableBodyRow, isNew);
          }
        }

        this._hideNoDataMessage();
        
        var currentRow = this._getCurrentRow();
        currentRow = currentRow || {};
        var rowContext = oj.TableRendererUtils.getRendererContextObject(this, tableBodyRow, 
                                          {'row': row, 'isCurrentRow': currentRow['rowIndex'] == rowIdx});
        // Copy additional properties to top-level context to work with custom element
        var context = {'rowContext': rowContext,
                       'row': row[this._CONST_DATA],
                       'componentElement': rowContext['componentElement'],
                       'parentElement': rowContext['parentElement'],
                       'data': row[this._CONST_DATA]};
                  
        if (rowRenderer != null)
        {
          var rowContent = rowRenderer(context);
          if (rowContent != null)
          {
            // if the renderer returned a value then we set it as the content
            // for the row
            tableBodyRow.appendChild(rowContent); //@HTMLUpdateOK
          }
          else
          {
            // if the renderer didn't return a value then the existing
            // row was manipulated. So get it and set the required
            // attributes just in case it was replaced or the attributes
            // got removed
            if (docFrag == null)
            {
              tableBodyRow = tableBody.children[rowIdx];
            }
            else
            {
              docFragStartIdx = docFragStartIdx == null ? 0 : docFragStartIdx;
              if (!docFrag.children) 
              {
                // use jquery children() because documentFragments do not have
                // good browser support for .children
                tableBodyRow = $(docFrag).children()[rowIdx - docFragStartIdx];
              }
              else
              {
                tableBodyRow = docFrag.children[rowIdx - docFragStartIdx];
              }
            }
            this._getTableDomUtils().clearCachedDomRowData();
            this._getTableDomUtils().setTableBodyRowAttributes(row, tableBodyRow);
            this._getTableDomUtils().styleTableBodyRow(tableBodyRow, false);
          }
          this._getTableDomUtils().createTableBodyCellAccSelect(rowIdx, row[this._CONST_KEY], rowHashCode, tableBodyRow, false);

          // set the cell attributes and styling. Skip the 1st one
          // because it's the acc row select td
          var tableBodyCells = tableBodyRow.getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._TD);
          var i, tableBodyCellsCount = tableBodyCells.length;
          var tableBodyCell;
          for (i = 1; i < tableBodyCellsCount; i++)
          {
            tableBodyCell = tableBodyCells[i];
            this._getTableDomUtils().setTableBodyCellAttributes(rowIdx, row[this._CONST_KEY], rowHashCode, i - 1, tableBodyCell);
            this._getTableDomUtils().styleTableBodyCell(i - 1, tableBodyCell, false);
          }
          
          // sort the re-ordered columns in place
          if (this._columnsDestMap != null)
          {
            var moveTableBodyCell, swapTableBodyCell;
            for (i = 0; i < this._columnsDestMap.length - 1; i++)
            {
              if (this._columnsDestMap[i] > i)
              {
                moveTableBodyCell = tableBodyCells[this._columnsDestMap[i] + 1];
                swapTableBodyCell = tableBodyCells[i + 1];
                moveTableBodyCell.parentNode.insertBefore(moveTableBodyCell, swapTableBodyCell); //@HTMLUpdateOK
                tableBodyCells = tableBodyRow.getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._TD);
              }
            }
          }
        }
        else
        {
          oj.TableRendererUtils.tableBodyRowDefaultRenderer(this, rowIdx, row, context);
        }
        return tableBodyRow;
      },
      /**
       * Refresh the table footer
       * @private
       */
      _refreshTableFooter: function()
      {
        var columns = this._getColumnDefs();
        var tableFooter = this._getTableDomUtils().getTableFooter();

        if (!tableFooter)
        {
          if (this._isTableFooterless())
          {
            return;
          }
          else
          {
            // metadata could have been updated to add column headers
            tableFooter = this._getTableDomUtils().createTableFooter();
            this._getTableDomUtils().styleTableFooter(tableFooter);
          }
        }

        var tableFooterRow = this._getTableDomUtils().getTableFooterRow();
        // remove all the existing footer cells
        $(tableFooterRow).empty();

        if (columns && columns.length > 0)
        {
          this._getTableDomUtils().createTableFooterAccSelect(tableFooterRow);
          var i, column, footerRenderer, footerCell, footerCellContent, columnsCount = columns.length;
          for (i = 0; i < columnsCount; i++)
          {
            column = columns[i];
            footerRenderer = this._getColumnRenderer(i, 'footer');
            footerCell = this._getTableDomUtils().createTableFooterCell(i, this._getColumnSelectionMode());
            this._getTableDomUtils().styleTableFooterCell(i, footerCell);
            this._getTableDomUtils().insertTableFooterCell(i, footerCell);

            if (footerRenderer)
            {
              // if footerRenderer is defined then call that
              var footerContext = oj.TableRendererUtils.getRendererContextObject(this, footerCell, {});
              // Copy additional properties to top-level context to work with custom element
              var context = {'footerContext': footerContext,
                             'columnIndex': i,
                             'componentElement': footerContext['componentElement'],
                             'parentElement': footerContext['parentElement']}
              footerCellContent = footerRenderer(context);

              if (footerCellContent != null)
              {
                // if the renderer returned a value then we set it as the content
                // for the footer cell
                footerCell.empty();
                footerCell.appendChild(footerCellContent); //@HTMLUpdateOK
              }
              else
              {
                // if the renderer didn't return a value then the existing
                // footer cell was manipulated. So get it and set the required
                // attributes just in case it was replaced or the attributes
                // got removed
                footerCell = $(tableFooterRow).children(':not(.' + oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS + ')')[i];
                this._getTableDomUtils().styleTableFooterCell(i, footerCell, this._getColumnSelectionMode());
              }
            }
            else
            {
              var footerSlotTemplate = this._getSlotTemplate(column[this._FOOTER_TEMPLATE]);
              if (footerSlotTemplate)
              {
                var componentElement = this._getRootElement();
                var templateEngine = this._getTemplateEngine();
                if (templateEngine != null)
                {
                  var slotContext = oj.TableRendererUtils.getSlotTemplateContextObject(this);
                  var footerContent = templateEngine.execute(componentElement, footerSlotTemplate, slotContext);
                  if (!(footerContent instanceof Array)) 
                  {
                    footerContent = [footerContent];
                  }
                  footerContent.map(function(content)
                  {
                    footerCell.appendChild(content);
                  });
                }
              }
            }
          }
        }
      },
      /**
       * Refresh the table header
       * @private
       */
      _refreshTableHeader: function()
      {
        var self = this;
        var columns = this._getColumnDefs();
        var tableHeader = this._getTableDomUtils().getTableHeader();

        if (!tableHeader)
        {
          if (this._isTableHeaderless())
          {
            return;
          }
          else
          {
            // metadata could have been updated to add column headers
            tableHeader = this._getTableDomUtils().createTableHeader();
            this._getTableDomUtils().styleTableHeader(tableHeader);
          }
        }

        var tableHeaderRow = this._getTableDomUtils().getTableHeaderRow();
        this._unregisterChildStateListeners(tableHeaderRow);
        // remove all the existing column headers
        $(tableHeaderRow).empty();

        if (columns && columns.length > 0)
        {
          var tableHeaderAccSelectRowColumn = this._getTableDomUtils().createTableHeaderAccSelectRowColumn();
          tableHeaderRow.appendChild(tableHeaderAccSelectRowColumn); //@HTMLUpdateOK
          
          var i, j, column, headerRenderer, headerColumn, headerColumnContent, headerContext, context, columnsCount = columns.length;
          for (i = 0; i < columnsCount; i++)
          {
            column = columns[i];
            headerRenderer = this._getColumnRenderer(i, 'header');
            headerColumn = this._getTableDomUtils().createTableHeaderColumn(i, this._getColumnSelectionMode());

            if (headerRenderer)
            {
              // if headerRenderer is defined then call that
              headerContext = oj.TableRendererUtils.getRendererContextObject(this, headerColumn, {});
              // Copy additional properties to top-level context to work with custom element
              context = {'headerContext': headerContext,
                         'columnIndex': i,
                         'data': column['headerText'],
                         'componentElement': headerContext['componentElement'],
                         'parentElement': headerContext['parentElement']};

              if (column.sortable == oj.TableDomUtils._OPTION_ENABLED)
              {
                // add the sortable icon renderer
                context['columnHeaderSortableIconRenderer'] = function(options, delegateRenderer)
                {
                  oj.TableRendererUtils.columnHeaderSortableIconRenderer(self, this, options, delegateRenderer);
                }
              }
              else
              {
                context['columnHeaderDefaultRenderer'] = function(options, delegateRenderer)
                {
                  oj.TableRendererUtils.columnHeaderDefaultRenderer(self, this, options, delegateRenderer);
                }
              }
              headerColumnContent = headerRenderer(context);

              if (headerColumnContent != null)
              {
                // if the renderer returned a value then we set it as the content
                // for the headerColumn
                $(headerColumn).empty();
                // Use jquery append() for this as a convenience because 
                // headerColumnContent could be a Node element or arbitrary content and 
                // we don't want to write code to convert everything to Node type and call 
                // appendChild.
                $(headerColumn).append(headerColumnContent); //@HTMLUpdateOK
              }
              else
              {
                // if the renderer didn't return a value then the existing
                // headerColumn was manipulated. So get it and set the required
                // attributes just in case it was replaced or the attributes
                // got removed
                headerColumn = $(tableHeaderRow).children(':not(' + '.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ACC_SELECT_ROW_CLASS + ')')[i];
                this._getTableDomUtils().setTableHeaderColumnAttributes(i, headerColumn);
                this._getTableDomUtils().styleTableHeaderColumn(i, headerColumn, this._getColumnSelectionMode(), false);
              }
            }
            else
            {
              var headerSlotTemplate = this._getSlotTemplate(column[this._HEADER_TEMPLATE]);
              
              if (headerSlotTemplate)
              {
                var componentElement = this._getRootElement();
                var templateEngine = this._getTemplateEngine();
                if (templateEngine != null)
                {
                  if (headerColumn)
                  {
                    $(headerColumn).empty();
                  }
                  var slotContext = oj.TableRendererUtils.getHeaderSlotTemplateContextObject(this, column['headerText']);
                  var headerContent = templateEngine.execute(componentElement, headerSlotTemplate, slotContext);
                  if (!(headerContent instanceof Array)) 
                  {
                    headerContent = [headerContent];
                  }
                  headerContent.map(function(content)
                  {
                    headerColumn.appendChild(content);
                  });
                }
              }
            }
            // set the acc column selection checkbox
            this._getTableDomUtils().createTableHeaderColumnAccSelect(i, this._getColumnSelectionMode());
          }
          this._renderedTableHeaderColumns = true;
        }
      },
      /**
       * Refresh the status message
       * @private
       */
      _refreshTableStatusMessage: function()
      {
        var tableStatusMessage = this._getTableDomUtils().getTableStatusMessage();

        if (tableStatusMessage != null)
        {
          $(tableStatusMessage).remove();
        }
        this._getTableDomUtils().createTableStatusMessage();
      },
      /**
       * Register the events which will be published by the table component.
       * @private
       */
      _registerCustomEvents: function()
      {
        var jqEvent = (/** @type {{special: Object}} */($.event));
        var jqEventSpecial = jqEvent['special'];
        // ojtablebeforecurrentrow handlers will be passed an object which contains the
        // old and new current row
        jqEventSpecial['ojtablebeforecurrentrow'] = {
          /**
           * Handle event
           * @param {{handleObj: {handler: {apply: Function}}}} event
           * @private
           */
          handle: function(event) {
            var handleObj = event['handleObj'];
            return handleObj['handler'].apply(this, [event, arguments[1]]);
          }
        };
        // ojtablesort handlers will be passed an object which contains the
        // header and direction
        jqEventSpecial['ojtablesort'] = {
          /**
           * Handle event
           * @param {{handleObj: {handler: {apply: Function}}}} event
           * @private
           */
          handle: function(event) {
            var handleObj = event['handleObj'];
            return handleObj['handler'].apply(this, [event, arguments[1]]);
          }
        };
      },
      /**
       * Register event listeners which need to be registered datasource.
       * @private
       */
      _registerDataSourceEventListeners: function()
      {
        // register the listeners on the datasource
        var dataprovider = this._getData();
        if (dataprovider != null)
        {
          this._unregisterDataSourceEventListeners();

          this._dataProviderEventHandlers = [];
          this._dataProviderEventHandlers.push({'eventType': 'mutate', 'eventHandler': this._handleDataRowMutate.bind(this)});
          this._dataProviderEventHandlers.push({'eventType': 'refresh', 'eventHandler': this._handleDataRefresh.bind(this)});
          var i;
          var ev, dataProviderEventHandlersCount = this._dataProviderEventHandlers.length;
          for (i = 0; i < dataProviderEventHandlersCount; i++) {
            ev = dataprovider.addEventListener(this._dataProviderEventHandlers[i]['eventType'], this._dataProviderEventHandlers[i]['eventHandler']);
            if (ev) {
                this._dataProviderEventHandlers[i]['eventHandler'] = ev;
            }
          }
        }
      },
      /**
       * Register event listeners which need to be registered directly on
       * the DOM element.
       * @private
       */
      _registerDomEventListeners: function()
      {
        if (this._getTableDomUtils().getScroller() != null)
        {
          // if width or height is defined then we can have scrollbars so register scroll event listeners
          $(this._getTableDomUtils().getScroller()).scroll((function(event) {
            this._handleScrollerScrollLeft(this._getTableDomUtils().getScrollLeft(event.target));
            this._handleScrollerScrollTop(event.target.scrollTop);
          }).bind(this));
        }
      },
      /**
       * @private
       */
      _registerDomScroller: function()
      {
        var self = this;

        if (this._domScroller != null)
        {
          this._domScroller.destroy();
        }

        this._domScrollerSuccessFunc = function(result)
        {
          self._clearDataWaitingState();
          if (result != null)
          {
            if (result['maxCountLimit'])
            {
              self._handleScrollerMaxRowCount();
            }
            else
            {
              var value = result[self._CONST_VALUE];
              var data = value[self._CONST_DATA];
              var keys = value.metadata.map(function(value) {
                return value[self._CONST_KEY];
              });
              keys = keys.filter(function(value, index) {
                if (self._getRowIdxForRowKey(value) !== null)
                {
                  data.splice(index, 1);
                  return false;
                }
                return true;
              });
              
              self._queueTask(function()
              {
                var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
                var rowCount = tableBodyRows != null ? tableBodyRows.length : 0;
                var i, indexArray = [];
                for (i = 0; i < data.length; i++)
                {
                  indexArray[i] = rowCount + i;
                }
                self._refreshAll({'data': data, 'keys' : keys, 'indexes': indexArray}, rowCount);
              });
            }
          }
        };        
        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
        var rowCount = tableBodyRows != null ? tableBodyRows.length : 0;
        this._domScroller = new oj.DomScroller(this._getTableDomUtils().getScroller(),
          this._getData(),
          {'asyncIterator': this._dataProviderAsyncIterator,
            'fetchSize': this.options['scrollPolicyOptions']['fetchSize'],
            'maxCount': this.options['scrollPolicyOptions']['maxCount'],
            'initialRowCount': rowCount,
            'success': this._domScrollerSuccessFunc.bind(this),
            'request': this._handleDataFetchStart.bind(this),
            'fetchTrigger': 1});
      },
      /**
       * Register event listeners for resize the container DOM element.
       * @private
       */
      _registerResizeListener: function()
      {
        var element = this._getTableDomUtils().getTableContainer();
        if (!this._resizeListener)
        {
          var self = this;
          this._resizeListener = function(width, height)
                                 {
                                   if (oj.Context.getContext(self.element[0]).getBusyContext().isReady())
                                   {
                                     var tableContainerHeight = $(self._getTableDomUtils().getTableContainer()).outerHeight();
                                     var tableContainerWidth = $(self._getTableDomUtils().getTableContainer()).outerWidth();
                                   self._getTableDomUtils().refreshTableDimensions(tableContainerWidth, tableContainerHeight);
                                   }
                                 };
        }
        if (!this._isResizeListenerAdded)
        {
          oj.DomUtils.addResizeListener(element, this._resizeListener, 50);
          this._isResizeListenerAdded = true;
        }
      },
      /**
       * Remove a keyCode from our internal list of pressed keys. This is done on keyup.
       * @private
       */
      _removeKeyboardKey: function(keyCode)
      {
        var keyboardKeys = this._getKeyboardKeys();
        var i, keyboardKeysCount = keyboardKeys.length;
        for (i = 0; i < keyboardKeysCount; i++)
        {
          if (keyboardKeys[i] == keyCode)
          {
            keyboardKeys.splice(i, 1);
          }
        }
      },
      /**
       * Scroll column into viewport
       * @param {number} columnIdx  row index
       * @private
       */
      _scrollColumnIntoViewport: function(columnIdx)
      {
        var isRTL = (this._GetReadingDirection() === "rtl");
        var tableBody = this._getTableDomUtils().getTableBody();
        var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);

        if (!tableHeaderColumn)
        {
          return;
        }

        var scrollbarWidth = this._getTableDomUtils().getScrollbarWidth();
        var headerColumnRect = tableHeaderColumn.getBoundingClientRect();
        var tableBodyRect = tableBody.getBoundingClientRect();

        var scrolledLeft = false;

        if (isRTL)
        {
          if (headerColumnRect.left < tableBodyRect.left + scrollbarWidth)
          {
            var scrollLeftDiff = tableBodyRect.left - headerColumnRect.left + scrollbarWidth;
            if (!this._getTableDomUtils()._isIE())
            {
              scrollLeftDiff = -1 * scrollLeftDiff;
            }
            $(tableBody).scrollLeft($(tableBody).scrollLeft() + scrollLeftDiff);
            scrolledLeft = true;
          }

          if (headerColumnRect.right > tableBodyRect.right && !scrolledLeft)
          {
            var scrollLeftDiff = headerColumnRect.right - tableBodyRect.right;
            if (!this._getTableDomUtils()._isIE())
            {
              scrollLeftDiff = -1 * scrollLeftDiff;
            }
            $(tableBody).scrollLeft($(tableBody).scrollLeft() - scrollLeftDiff);
          }
        }
        else
        {
          if (headerColumnRect.left < tableBodyRect.left)
          {
            var scrollLeftDiff = tableBodyRect.left - headerColumnRect.left;
            $(tableBody).scrollLeft($(tableBody).scrollLeft() - scrollLeftDiff);
            scrolledLeft = true;
          }

          if (headerColumnRect.right > tableBodyRect.right - scrollbarWidth && !scrolledLeft)
          {
            var scrollLeftDiff = headerColumnRect.right - tableBodyRect.right + scrollbarWidth;
            $(tableBody).scrollLeft($(tableBody).scrollLeft() + scrollLeftDiff);
          }
        }
      },
      /**
       * Scroll row into viewport
       * @param {number} rowIdx  row index
       * @private
       */
      _scrollRowIntoViewport: function(rowIdx)
      {

        var tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);
        var scrollbarHeight = this._getTableDomUtils().getScrollbarHeight();
        var rowRect = tableBodyRow.getBoundingClientRect();
        var scrollingElement = this._getTableDomUtils().getScroller();
        var scrollingElementRect = scrollingElement.getBoundingClientRect();

        var scrolledDown = false;
        if (rowRect.bottom > scrollingElementRect.bottom - scrollbarHeight)
        {
          var scrollTopDiff = rowRect.bottom - scrollingElementRect.bottom + scrollbarHeight;
          $(scrollingElement).scrollTop($(scrollingElement).scrollTop() + scrollTopDiff);
          scrolledDown = true;
        }

        if (rowRect.top < scrollingElementRect.top && !scrolledDown)
        {
          var scrollTopDiff = scrollingElementRect.top - rowRect.top;
          $(scrollingElement).scrollTop($(scrollingElement).scrollTop() - scrollTopDiff);
        }

      },
      /**
       * Update the current row. If called with null then resets the currentRow.
       * If index/key argument is specified then sets the current row. A beforecurrentrow
       * event is fired before the current row is changed. If that event results in
       * an error then the current row will not be changed.
       * @param {Object} currentRow current row
       * @param {Object} event
       * @param {boolean} optionChange whether it was invoked through an optionChange call
       * @return {boolean} whether setting the current row was successful
       * @throws {Error}
       * @private
       */
      _setCurrentRow: function(currentRow, event, optionChange)
      {
        var existingCurrentRow = this._currentRow;
        var tableBodyRow, existingCurrentRowIndex, existingCurrentRowKey, existingCurrentRowIdx;
        var updateCurrentRow = true;

        if (currentRow == null)
        {
          if (existingCurrentRow != null)
          {
            try
            {
              updateCurrentRow = this._trigger('beforeCurrentRow', event, {'currentRow': null, 'previousCurrentRow': this._currentRow});
            }
            catch (err)
            {
              // caught an error. Do not change current row
              var errSummary = this._LOGGER_MSG._ERR_PRECURRENTROW_ERROR_SUMMARY;
              var errDetail = oj.Translations.applyParameters(this._LOGGER_MSG._ERR_PRECURRENTROW_ERROR_DETAIL, {'error': err.toString()});
              oj.Logger.info(errSummary + '\n' + errDetail);
              // do not update the currentRow to the new value if an exception was caught
              return false;
            }

            if (!updateCurrentRow)
            {
              // do not update the currentRow to the new value if a listener returned false
              return false;
            }

            existingCurrentRowIndex = existingCurrentRow['rowIndex'];
            existingCurrentRowKey = this._getRowKeyForDataSourceRowIndex(existingCurrentRowIndex);
            existingCurrentRowIdx = this._getRowIdxForRowKey(existingCurrentRowKey);
            
            var updateEditable = this._setTableEditable(false, false, 0, true, event);

            if (updateEditable === false)
            {
              this._currentRow = existingCurrentRow;
              var currentFocusElement = document.activeElement;
              var columnIdx = this._getTableDomUtils().getElementColumnIdx(currentFocusElement);
              var self = this;
              this._queueTask(function()
              {
                var focusRowIdx = existingCurrentRowIdx;
                var focusColumnIdx = columnIdx;
                if (focusRowIdx != null &&
                  focusColumnIdx != null)
                {
                setTimeout(function(){self._setCellFocus(focusRowIdx, focusColumnIdx)}, 0);
                }
                else if (focusRowIdx != null)
                {
                  setTimeout(function(){self._setRowFocus(focusRowIdx, true, false, null, null)}, 0);
                }
              });
              
              // do not update the currentRow to the new value if updateEditable returned false
              return false;
            }
            this._currentRow = null;
            this.option('currentRow', null,  {'_context': {writeback: true, originalEvent:event, internalSet: true}});

            if (event == null)
            {
              this._setRowFocus(-1, true, false, null, event);
            }

            tableBodyRow = this._getTableDomUtils().getTableBodyRow(existingCurrentRowIdx);

            if (tableBodyRow != null)
            {
              tableBodyRow.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CURRENT_ROW_CLASS);
            }
          }

          return true;
        }
        var dataprovider = this._getData();
        var rowIndex = currentRow['rowIndex'];
        var rowIdx;
        var rowKey = currentRow['rowKey'];
        if (rowKey == null)
        {
          rowKey = this._getRowKeyForDataSourceRowIndex(rowIndex);
        }
        rowIndex = this._getDataSourceRowIndexForRowKey(rowKey);
        rowIdx = this._getRowIdxForRowKey(rowKey);
        currentRow = {'rowIndex': rowIndex, 'rowKey': rowKey};

        if (rowIdx != -1 &&
            (!dataprovider ||
            rowIdx < -1 ||
            rowIdx === null ||
            rowKey === null))
        {
          var errSummary = this._LOGGER_MSG._ERR_CURRENTROW_UNAVAILABLE_INDEX_SUMMARY;
          var errDetail = oj.Translations.applyParameters(this._LOGGER_MSG._ERR_CURRENTROW_UNAVAILABLE_INDEX_DETAIL, {'rowIdx': rowIdx});
          if (optionChange)
          {
            // Only throw an Error if the current row was set through option change
            // so that the caller can be notified of the invalid row.
            throw new Error(errSummary + '\n' + errDetail);
          }
          else
          {
            oj.Logger.info(errSummary + '\n' + errDetail);
          }
          // do not update the currentRow
          return false;
        }

        var currentFocusedRowIdx = this._getFocusedRowIdx();
        var currentRowChanged = !oj.Object.compareValues(this._currentRow, currentRow);

        if (currentRowChanged)
        {
          try
          {
            updateCurrentRow = this._trigger('beforeCurrentRow', event, {'currentRow': {'rowIndex': rowIndex, 'rowKey': rowKey}, 'previousCurrentRow': this._currentRow});
          }
          catch (err)
          {
            // caught an error. Do not change current row
            var errSummary = this._LOGGER_MSG._ERR_PRECURRENTROW_ERROR_SUMMARY;
            var errDetail = oj.Translations.applyParameters(this._LOGGER_MSG._ERR_PRECURRENTROW_ERROR_DETAIL, {'error': err.toString()});
            oj.Logger.info(errSummary + '\n' + errDetail);
            // do not update the currentRow to the new value if an exception was caught
            return false;
          }

          if (!updateCurrentRow)
          {
            // do not update the currentRow to the new value if a listener returned false
            return false;
          }

          this._currentRow = {'rowIndex': rowIndex, 'rowKey': rowKey};
          
          if (existingCurrentRow != null)
          {
            existingCurrentRowIndex = existingCurrentRow['rowIndex'];
            existingCurrentRowKey = this._getRowKeyForDataSourceRowIndex(existingCurrentRowIndex);
            existingCurrentRowIdx = this._getRowIdxForRowKey(existingCurrentRowKey);
          }
          
          var updateEditable = this._setTableEditable(false, false, 0, true, event);

          if (updateEditable === false)
          {
            this._currentRow = existingCurrentRow;
            var currentFocusElement = document.activeElement;
            var columnIdx = this._getTableDomUtils().getElementColumnIdx(currentFocusElement);
            var self = this;
            this._queueTask(function()
            {
              var focusRowIdx = existingCurrentRowIdx;
              var focusColumnIdx = columnIdx;
              if (focusRowIdx != null &&
                  focusColumnIdx != null)
              {
                setTimeout(function(){self._setCellFocus(focusRowIdx, focusColumnIdx)}, 0);
              }
              else if (focusRowIdx != null)
              {
                setTimeout(function(){self._setRowFocus(focusRowIdx, true, false, null, null)}, 0);
              }
            });

            // do not update the currentRow to the new value if updateEditable returned false
            return false;
          }
            
          this.option('currentRow', this._currentRow, {'_context': {writeback: true, originalEvent:event, internalSet: true}});
          tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);

          if (tableBodyRow != null)
          {
            tableBodyRow.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CURRENT_ROW_CLASS);
          }
          if (existingCurrentRow != null)
          {
            tableBodyRow = this._getTableDomUtils().getTableBodyRow(existingCurrentRowIdx);

            if (tableBodyRow != null)
            {
              tableBodyRow.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CURRENT_ROW_CLASS);
            }
          }
        }

        if (currentRowChanged || currentFocusedRowIdx != currentRow['rowIndex'])
        {
          if (event == null)
          {
            this._setRowFocus(rowIdx, true, false, null, event);
          }
        }
        return true;
      },
      /**
       * Set waiting state and show the Fetching Data... status message.
       * @private
       */
      _setDataWaitingState: function()
      {
        this._showStatusMessage();
        this._hideNoDataMessage();
        this._dataFetching = true;
        if (!this._dataResolveFunc)
        {
	        this._dataResolveFunc = this._addComponentBusyState('is waiting for data.');
        }
      },
      /**
       * Set focus on a cell
       * @param {number} rowIdx  row index
       * @param {number} columnIdx  column index
       * @return {boolean} whether setting the cell focus was successful
       * @private
       */
      _setCellFocus: function(rowIdx, columnIdx)
      {
        var tableBodyCell = this._getTableDomUtils().getTableBodyCell(rowIdx, columnIdx);
        
        if (!tableBodyCell)
        {
          return false;
        }
        var self = this;
        var focused = false;
        var elements = tableBodyCell.querySelectorAll('*');
        var i;
        for (i = 0; i < elements.length; i++)
        {
          if (!focused)
          {
            if ($(elements[i]).is(':focusable'))
            {
              $(elements[i]).focus();
              focused = $(elements[i]).is(':focus') || elements[i] == document.activeElement;
            }
          }
        }
        return focused;
      },
      /**
       * Try setting focus in the cell. If unsuccesful, set focus on the first focusable cell in the row
       * if forwardSearch is true. Else try from the last cell in the row.
       * @param {number} rowIdx  row index
       * @param {number} columnIdx  column index
       * @param {boolean} forwardSearch try setting focus starting at the first column
       * @private
       */
      _setCellInRowFocus: function(rowIdx, columnIdx, forwardSearch)
      {
        if (this._setCellFocus(rowIdx, columnIdx))
        {
          return;
        }
        
        var tableBodyCells = this._getTableDomUtils().getTableBodyCells(rowIdx);
        
        if (!tableBodyCells)
        {
          return;
        }

        var i, tableBodyCellIndex, tableBodyCellsCount = tableBodyCells.length;
        
        for (i = 0; i < tableBodyCellsCount; i++)
        {
          tableBodyCellIndex = i;
          
          if (!forwardSearch)
          {
            tableBodyCellIndex = tableBodyCellsCount - i - 1;
          }
          if (this._setCellFocus(rowIdx, tableBodyCellIndex))
          {
            return;
          }
        }
      },
      /**
	     * Called by component to declare rendering is not finished. This method currently 
       * handles the ready state for the component page level BusyContext
	     * @private
       */
      _setComponentNotReady : function() 
      {
        // For page level BusyContext
	      // If we've already registered a busy state with the page's busy context, don't need to do anything further
        if (!this._readyResolveFunc) 
        {
	        this._readyResolveFunc = this._addComponentBusyState('is being loaded.');
	      }
      },
	    /**
	     * Called by component to declare rendering is finished. This method currently
       * handles the page level BusyContext.
	     * @private
       */
	    _setComponentReady : function() 
      {
        if (this._readyResolveFunc) 
        {
          this._readyResolveFunc();
          this._readyResolveFunc = null;
        }
      },
      /**
       * Set the editable row index
       * @param {number} rowIdx  row index
       * @private
       */
      _setEditableRowIdx: function(rowIdx)
      {
        // store the rowKey so we can restore easily after sort
        this._editableRowKey = this._getRowKeyForRowIdx(rowIdx);
        this._editableRowIdx = rowIdx;
      },
      /**
       * Set the focus in handler
       * @param {Function} focusHandler focus in handler
       * @private
       */
      _setFocusInHandler: function(focusHandler)
      {
        this._focusInHandler = focusHandler;
      },
      /**
       * Set the focus out handler
       * @param {Function} focusHandler focus out handler
       * @private
       */
      _setFocusOutHandler: function(focusHandler)
      {
        this._focusOutHandler = focusHandler;
      },
      /**
       * Set focus on column header
       * @param {number} columnIdx  column index
       * @param {boolean} focused  whether it's focused
       * @param {boolean} clearSelectedRows  whether to clear the selected rows
       * @param {Object} event
       * @private
       */
      _setHeaderColumnFocus: function(columnIdx, focused, clearSelectedRows, event)
      {
        var element = null;

        if (event != null)
        {
          element = this._getEventTargetElement(event);
        }

        if (focused)
        {
          var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();
          if (focusedHeaderColumnIdx != null && focusedHeaderColumnIdx != columnIdx)
          {
            this._setHeaderColumnFocus(focusedHeaderColumnIdx, false, false, event);
          }
          // clear focused row
          this._clearFocusedRow(true);
          // clear selected rows
          if (clearSelectedRows)
          {
            this._clearSelectedRows();
          }
          // scroll column into view
          this._scrollColumnIntoViewport(columnIdx);
          this._activeColumnIndex = columnIdx;

        }
        this._setHeaderColumnState(columnIdx, {focused: focused}, element);
      },
      /**
       * Set selection on column header
       * @param {number} columnIdx  column index
       * @param {boolean} selected  whether it's focused
       * @param {Element} element  DOM element which triggered the column header selection
       * @param {Object} event
       * @param {boolean} updateSelection  whether to update the selection
       * @param {boolean} ignoreSelectionRequired whether to ignore selection required
       * @private
       */
      _setHeaderColumnSelection: function(columnIdx, selected, element, event, updateSelection, ignoreSelectionRequired)
      {
        if (this._getColumnSelectionMode() == this._OPTION_SELECTION_MODES._SINGLE ||
          this._getColumnSelectionMode() == this._OPTION_SELECTION_MODES._MULTIPLE)
        {
          if (isNaN(columnIdx) || columnIdx < 0)
          {
            // validate value
            oj.Logger.error('Error: Invalid column selection value: ' + columnIdx);
          }
          
          // selection-required enabled then don't allow unselection of the last
          // selected column if it's the only one
          if (!ignoreSelectionRequired && ((new String(this.options['selectionRequired'])).toLowerCase() == 'true') && !selected)
          {
            var selectedHeaderColumnIdxs = this._getSelectedHeaderColumnIdxs();

            if (selectedHeaderColumnIdxs != null && selectedHeaderColumnIdxs.length == 1)
            {
              return;
            }
          }

          // if we have single selection then clear any existing selections
          if (this._getColumnSelectionMode() == this._OPTION_SELECTION_MODES._SINGLE && selected)
          {
            this._clearSelectedHeaderColumns();
          }
          this._setHeaderColumnState(columnIdx, {selected: selected}, element, event);
          // save it
          this._setLastHeaderColumnSelection(columnIdx, selected);

          // update the acc checkbox
          var accSelectionColumn = this._getTableDomUtils().getTableHeaderColumnAccSelect(columnIdx);
          var accSelectCheckbox = accSelectionColumn.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._CHECKBOX_ACC_SELECT_COLUMN_CLASS)[0];
          accSelectCheckbox['checked'] = selected;
        }
        if (updateSelection)
        {
          var selection = this._getSelection();
          this.option('selection', selection, {'_context': {writeback: true, internalSet: true}});
          
          if (selected)
          {
            this.option('firstSelectedRow', {key: null, data: null}, {'_context': {writeback: true, internalSet: true}});
          }
        }
      },
      /**
       * Set the state of the column header. e.g., focused, selected, etc.
       * @param {number} columnIdx  column index
       * @param {Object} state  Object which contains whether it's focused or selected
       * @param {Element} element  DOM element which triggered the column header state
       * @param {Object} event
       * @private
       */
      _setHeaderColumnState: function(columnIdx, state, element, event)
      {
        var headerColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);

        if (!headerColumn)
        {
          return;
        }

        var focused = state.focused;
        var selected = state.selected;

        if (selected != null)
        {
          var headerColumnSelected = headerColumn.classList.contains(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);

          if (headerColumnSelected != selected)
          {
            if (!selected)
            {
              headerColumn.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
            }
            else
            {
              headerColumn.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
            }
          }
        }
        if (focused != null)
        {
          if (!focused)
          {
            this._focusOutHandler($(headerColumn));
            this._hideTableHeaderColumnSortLink(columnIdx, true);
            this._hideTableHeaderColumnSortLink(columnIdx, false);
          }
          else
          {
            this._focusInHandler($(headerColumn));
            this._showTableHeaderColumnSortLink(columnIdx);
          }
        }
        this._updateHeaderColumnStateCellsClass(columnIdx);
      },
      /**
       * Set the last column which was selected (chronologically)
       * @param {number} columnIdx  column index
       * @param {boolean} selected  whether it's selected
       * @private
       */
      _setLastHeaderColumnSelection: function(columnIdx, selected)
      {
        if (!this._lastSelectedColumnIdxArray)
        {
          this._lastSelectedColumnIdxArray = [];
        }

        var i, lastSelectedColumnIdxArrayCount = this._lastSelectedColumnIdxArray.length;
        for (i = 0; i < lastSelectedColumnIdxArrayCount; i++)
        {
          if (this._lastSelectedColumnIdxArray[i] == columnIdx)
          {
            this._lastSelectedColumnIdxArray.splice(i, 1);
            break;
          }
        }

        if (selected)
        {
          this._lastSelectedColumnIdxArray.push(columnIdx);
        }
      },
      /**
       * Set the last row which was selected (chronologically)
       * @param {number} rowIdx  row index
       * @param {boolean} selected  whether it's selected
       * @private
       */
      _setLastRowSelection: function(rowIdx, selected)
      {
        if (!this._lastSelectedRowIdxArray)
        {
          this._lastSelectedRowIdxArray = [];
        }

        var i, lastSelectedRowIdxArrayCount = this._lastSelectedRowIdxArray.length;
        for (i = 0; i < lastSelectedRowIdxArrayCount; i++)
        {
          if (this._lastSelectedRowIdxArray[i] == rowIdx)
          {
            this._lastSelectedRowIdxArray.splice(i, 1);
            break;
          }
        }

        if (selected)
        {
          this._lastSelectedRowIdxArray.push(rowIdx);
        }
      },
      /**
       * Set the next row to editable
       * @param {number} columnIdx  column index
       * @param {Object} event
       * @throws {Error}
       * @private
       */
      _setNextRowEditable: function(columnIdx, event)
      {
        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
        var rowCount = tableBodyRows != null ? tableBodyRows.length : 0;
        var editableRowIdx = this._getEditableRowIdx();
            
        if (editableRowIdx >= 0 && editableRowIdx < rowCount - 1)
        {
          this._setTableEditable(false, false, columnIdx, true, event);
          this._setCurrentRow({'rowIndex': editableRowIdx + 1}, event, false);
          this._setTableEditable(true, false, columnIdx, true, event);
        }
        else if (editableRowIdx === rowCount - 1)
        {
          this._setTableEditable(false, false, columnIdx, true, event);
          this._getTableDomUtils().getTable().focus();
        }
      },
      /**
       * Set the previous row to editable
       * @param {number} columnIdx  column index
       * @param {Object} event
       * @throws {Error}
       * @private
       */
      _setPreviousRowEditable: function(columnIdx, event)
      {
        var editableRowIdx = this._getEditableRowIdx();
        
        if (editableRowIdx >= 1)
        {
          this._setTableEditable(false, false, columnIdx, false, event);
          this._setCurrentRow({'rowIndex': editableRowIdx - 1}, event, false);
          this._setTableEditable(true, false, columnIdx, false, event);
        }
        else if (editableRowIdx === 0)
        {
          this._setTableEditable(false, false, columnIdx, false, event);
          this._getTableDomUtils().getTable().focus();
        }
      },
      /**
       * Set focus on row
       * @param {number} rowIdx  row index
       * @param {boolean} focused  whether it's focused
       * @param {boolean} updateCurrentRow  whether to update the currentRow
       * @param {Element} element  DOM element which triggered the row focus
       * @param {Object} event
       * @return {boolean} whether setting the row focus was successful
       * @private
       */
      _setRowFocus: function(rowIdx, focused, updateCurrentRow, element, event)
      {
        if (rowIdx == -1)
        {
          var focusedRowIdx = this._getFocusedRowIdx();
          if (focusedRowIdx != null)
          {
            this._setRowFocus(focusedRowIdx, false, updateCurrentRow, null, null);
          }
          if (updateCurrentRow)
          {
            var updateRowFocus = this._setCurrentRow(null, event, false);

            if (!updateRowFocus)
            {
              return false;
            }
          }
          return true;
        }
        var tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);

        if (!tableBodyRow)
        {
          return false;
        }

        if (focused)
        {
          var focusedRowIdx = this._getFocusedRowIdx();
          if (focusedRowIdx != null && focusedRowIdx != rowIdx)
          {
            this._setRowFocus(focusedRowIdx, false, updateCurrentRow, element, null);
          }

          if (updateCurrentRow)
          {
            var rowKey = this._getRowKeyForRowIdx(rowIdx);
            var updateRowFocus = this._setCurrentRow({'rowKey': rowKey}, event, false);

            if (!updateRowFocus)
            {
              return false;
            }
          }
          this._focusInHandler($(tableBodyRow));
          this._scrollRowIntoViewport(rowIdx);
          // clear any hover on the row
          this._updateRowStateCellsClass(rowIdx, {focused: true, hover: false});
          // clear any focused column header
          this._clearFocusedHeaderColumn();
          // clear any selected column header
          this._clearSelectedHeaderColumns();
          // unset table actionable mode
          this._setTableActionableMode(false);
        }
        else
        {
          this._focusOutHandler($(tableBodyRow));
          // update focus style for the cells
          this._updateRowStateCellsClass(rowIdx, {focused: false});
        }
        
        return true;
      },
      /**
       * Set selection on row
       * @param {number} rowIdx  column index
       * @param {boolean} selected  whether it's selected
       * @param {Element} element  DOM element which triggered the row selection
       * @param {Object} event
       * @param {boolean} updateSelection  whether to update the selection
       * @param {boolean} ignoreSelectionRequired whether to ignore selection required
       * @private
       */
      _setRowSelection: function(rowIdx, selected, element, event, updateSelection, ignoreSelectionRequired)
      {
        if (this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._SINGLE ||
          this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._MULTIPLE)
        {
          if (isNaN(rowIdx) || rowIdx < 0)
          {
            // validate value
            oj.Logger.error('Error: Invalid row selection value: ' + rowIdx);
          }
          
          // selection-required enabled then don't allow unselection of the last
          // selected row if it's the only one
          if (!ignoreSelectionRequired && ((new String(this.options['selectionRequired'])).toLowerCase() == 'true') && !selected)
          {
            var selectedRowIdxs = this._getSelectedRowIdxs();
            
            if (selectedRowIdxs != null && selectedRowIdxs.length == 1)
            {
              return;
            }
          }

          // if we have single selection then clear any existing selections
          if (this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._SINGLE && selected)
          {
            this._clearSelectedRows();
          }
          var tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);

          if (tableBodyRow != null)
          {
            var selectionChanged = false;
            var rowSelected = tableBodyRow.classList.contains(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);

            if (rowSelected != selected)
            {
              if (!selected)
              {
                tableBodyRow.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
              }
              else
              {
                tableBodyRow.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
              }
              selectionChanged = true;
              
              // Set the draggable property on the row element if the dnd.drag.rows option is specified
              var dragOption = this.options['dnd']['drag'];
              if (dragOption && (dragOption === 'rows' || dragOption.rows))
              {
                tableBodyRow["draggable"] = selected;
              }
            }

            if (selectionChanged)
            {
              // if selection was set then we want to override
              // the default style precedence
              if (selected)
              {
                this._updateRowStateCellsClass(rowIdx, {hover: false, focused: false, selected: true});
              }
              else
              {
                this._updateRowStateCellsClass(rowIdx, {selected: false});
              }
            }
            // save it
            this._setLastRowSelection(rowIdx, selected);

            // update the acc checkbox
            var accSelectionCell = this._getTableDomUtils().getTableBodyCellAccSelect(tableBodyRow);
            var accSelectCheckbox = accSelectionCell.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._CHECKBOX_ACC_SELECT_ROW_CLASS)[0];
            accSelectCheckbox['checked'] = selected;

            if (updateSelection)
            {
              var selection = this._getSelection();
              this.option('selection', selection, {'_context': {writeback: true, internalSet: true}});
              
              // set firstSelectedRow
              if (selection != null && selection.length > 0)
              {
                var range = selection[0];
                var startKey = range['startKey'];
                if (startKey != null && 
                  startKey[this._CONST_ROW] != null) 
                {
                  var row = this['getDataForVisibleRow'](this._getRowIdxForRowKey(startKey[this._CONST_ROW]));
                  if (row)
                  {
                    this.option('firstSelectedRow', {key: row[this._CONST_KEY], data: row[this._CONST_DATA]}, {'_context': {writeback: true, internalSet: true}});
                    return;
                  }
                }
              }
              this.option('firstSelectedRow', {key: null, data: null}, {'_context': {writeback: true, internalSet: true}});
            }
          }
        }
      },
      /**
       * Sets selection
       * @param {Object} selection
       * @private
       */
      _setSelection: function(selection)
      {
        if (selection == null)
        {
          this._clearSelectedRows();
          this._clearSelectedHeaderColumns();
          return;
        }

        // we need to set the selection
        var i, j, rangeObj, startRowKey, endRowKey, startRowIndex, endRowIndex, startRowIdx, endRowIdx, startColumnIdx, endColumnIdx, updateSelection, selectionCount = selection.length;
        for (i = 0; i < selectionCount; i++)
        {
          rangeObj = selection[i];

          if ((rangeObj['startKey'] == null && rangeObj[this._CONST_STARTINDEX] == null) ||
              (rangeObj['endKey'] == null && rangeObj[this._CONST_ENDINDEX] == null))
          {
            oj.Logger.error('Error: Invalid range object in selection. Both start and end objects must be specified');
            return null;
          }

          startRowKey = null;
          endRowKey = null;
          startRowIndex = null;
          endRowIndex = null;
          startRowIdx = null;
          endRowIdx = null;
          startColumnIdx = null;
          endColumnIdx = null;
          updateSelection = false;

          // if keys are specified, we get the index from the key
          if (rangeObj['startKey'] != null && rangeObj['startKey'][this._CONST_ROW] != null)
          {
            startRowIndex = this._getDataSourceRowIndexForRowKey(rangeObj['startKey'][this._CONST_ROW]);
            
            if (rangeObj[this._CONST_STARTINDEX] != null && 
                rangeObj[this._CONST_STARTINDEX][this._CONST_ROW] != null &&
                startRowIndex != rangeObj[this._CONST_STARTINDEX][this._CONST_ROW])
            {
              updateSelection = true;
            }
          }
          if (rangeObj['endKey'] != null && rangeObj['endKey'][this._CONST_ROW] != null)
          {
            endRowIndex = this._getDataSourceRowIndexForRowKey(rangeObj['endKey'][this._CONST_ROW]);
            
            if (rangeObj[this._CONST_ENDINDEX] != null && 
                rangeObj[this._CONST_ENDINDEX][this._CONST_ROW] != null &&
                endRowIndex != rangeObj[this._CONST_ENDINDEX][this._CONST_ROW])
            {
              updateSelection = true;
            }
          }
          if (rangeObj['startKey'] != null && rangeObj['startKey'][this._CONST_COLUMN] != null)
          {
            startColumnIdx = this._getColumnIdxForColumnKey(rangeObj['startKey'][this._CONST_COLUMN]);
            
            if (rangeObj[this._CONST_STARTINDEX] != null && 
                rangeObj[this._CONST_STARTINDEX][this._CONST_COLUMN] != null &&
                startColumnIdx != rangeObj[this._CONST_STARTINDEX][this._CONST_COLUMN])
            {
              updateSelection = true;
            }
          }
          if (rangeObj['endKey'] != null && rangeObj['endKey'][this._CONST_COLUMN] != null)
          {
            endColumnIdx = this._getColumnIdxForColumnKey(rangeObj['endKey'][this._CONST_COLUMN]);
            
            if (rangeObj[this._CONST_ENDINDEX] != null && 
                rangeObj[this._CONST_ENDINDEX][this._CONST_COLUMN] != null &&
                endColumnIdx != rangeObj[this._CONST_ENDINDEX][this._CONST_COLUMN])
            {
              updateSelection = true;
            }
          }

          if (startRowIndex == null && rangeObj[this._CONST_STARTINDEX] != null)
          {
            startRowIndex = rangeObj[this._CONST_STARTINDEX][this._CONST_ROW];
          }
          if (endRowIndex == null && rangeObj[this._CONST_ENDINDEX] != null)
          {
            endRowIndex = rangeObj[this._CONST_ENDINDEX][this._CONST_ROW];
          }
          if (startColumnIdx == null && rangeObj[this._CONST_STARTINDEX] != null)
          {
            startColumnIdx = rangeObj[this._CONST_STARTINDEX][this._CONST_COLUMN];
          }
          if (endColumnIdx == null && rangeObj[this._CONST_ENDINDEX] != null)
          {
            endColumnIdx = rangeObj[this._CONST_ENDINDEX][this._CONST_COLUMN];
          }

          if (startRowIndex != null && endRowIndex != null && startColumnIdx != null && endColumnIdx != null)
          {
            oj.Logger.error('Error: Invalid range object in selection - Can only support row or column selection. Not both');
            return null;
          }
          if (startRowIndex != null && startRowIndex >= 0 && endRowIndex != null && endRowIndex >= 0)
          {
            // this is a row based selection
            startRowKey = this._getRowKeyForDataSourceRowIndex(startRowIndex);
            endRowKey = this._getRowKeyForDataSourceRowIndex(endRowIndex);
            startRowIdx = this._getRowIdxForRowKey(startRowKey);
            endRowIdx = this._getRowIdxForRowKey(endRowKey);
            for (j = startRowIdx; j <= endRowIdx; j++)
            {
              try
              {
                this._setRowSelection(j, true, null, null, updateSelection);
              }
              catch (e)
              {
                oj.Logger.error('Error: ' + e);
              }
            }
          }
          else if (startColumnIdx != null && endColumnIdx != null)
          {
            // this is a column based selection
            for (j = startColumnIdx; j <= endColumnIdx; j++)
            {
              try
              {
                this._setHeaderColumnSelection(j, true, null, null, updateSelection);
              }
              catch (e)
              {
                oj.Logger.error('Error: ' + e);
              }
            }
          }
          else
          {
            oj.Logger.error('Error: Invalid range object in selection - \n start row: ' + startRowIndex + '\n' + 'end row: ' + endRowIndex+ '\n' + 'start column: ' + startColumnIdx + '\n' + 'end column: ' + endColumnIdx);
            return null;
          }
        }
      },
      /**
       * Sets default selection when selection-required is enabled
       * @private
       */
      _setSelectionDefault: function()
      {
        if (((new String(this.options['selectionRequired'])).toLowerCase() == 'true'))
        {
          if (this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._MULTIPLE ||
              this._getRowSelectionMode() == this._OPTION_SELECTION_MODES._SINGLE) 
          {
            var selectedRowIdxs = this._getSelectedRowIdxs();

            if (selectedRowIdxs == null ||
              selectedRowIdxs.length == 0)
            {
              this._setRowSelection(0, true, null, null, true);
            }
          }
          else if (this._getColumnSelectionMode() == this._OPTION_SELECTION_MODES._MULTIPLE ||
                   this._getColumnSelectionMode() == this._OPTION_SELECTION_MODES._SINGLE)
          {
            var selectedColumnIdxs = this._getSelectedHeaderColumnIdxs();

            if (selectedColumnIdxs == null ||
              selectedColumnIdxs.length == 0)
            {
              this._setHeaderColumnSelection(0, true, null, null, true);
            }
          }
        }
      },
      /**
       * Set whether the component is in table actionable mode
       * @param {boolean} value true or false
       * @private
       */
      _setTableActionableMode: function(value)
      {
        if (this._isTableEditMode())
        {
          this._tableActionableMode = false;
          return;
        }
        
        // don't do anything if actionable mode was not updated
        if (this._tableActionableMode != null && this._tableActionableMode === value ||
          this._tableActionableMode == null && value === false)
        {
          return;
        }
        
        // only set actionable mode is there are any tabbable elements in the row 
        if (value)
        {
          var focusedRowIdx = this._getFocusedRowIdx();
          var focusedHeaderColumnIdx = this._getFocusedHeaderColumnIdx();

          if (focusedRowIdx != null ||
            focusedHeaderColumnIdx != null)
          {
            var focusedElement = null;
            
            if (focusedRowIdx != null)
            {
              focusedElement = this._getTableDomUtils().getTableBodyRow(focusedRowIdx);
            }
            else
            {
              focusedElement = this._getTableDomUtils().getTableHeaderColumn(focusedHeaderColumnIdx);
            }

            if (focusedElement != null)
            {
              var tabbableElementsInFocusedElement = this._getTabbableElements(focusedElement);

              if (tabbableElementsInFocusedElement != null && tabbableElementsInFocusedElement.length > 0)
              {
                this._tableActionableMode = value;
                // set focus on the first tabbable element
                tabbableElementsInFocusedElement[0].focus();
              }
            }
          }
        }
        else
        {
          this._tableActionableMode = value;
          var tableBody = this._getTableDomUtils().getTableBody();
          var resetFocus = $.contains(tableBody, document.activeElement);
          if (resetFocus)
          {
            this._getTableDomUtils().getTable().focus();
          }
        }
      },
      /**
       * Set the table editable
       * @param {boolean} editable true if editable, false otherwise
       * @param {boolean} cancelled true if edit was cancelled
       * @param {number} columnIdx  column index
       * @param {boolean} forwardSearch try setting focus starting at the first column
       * @param {Object} event
       * @private
       */
      _setTableEditable: function(editable, cancelled, columnIdx, forwardSearch, event)
      { 
        if (!this._isTableEditMode())
        {
          return;
        }
        
        var currentRow = this._getCurrentRow();

        if (currentRow != null)
        {
          var rowKey = currentRow['rowKey'];
          var rowIdx = this._getRowIdxForRowKey(rowKey);

          var updateEditMode;

          try
          {
            if (editable && !this._hasEditableRow())
            {
              // fire the beforeEdit event if there are no existing editable rows
              // and we are starting edit on a row
              var tableBodyRow = this._getTableDomUtils().getTableBodyRow(rowIdx);
              var rowContext = oj.TableRendererUtils.getRendererContextObject(this, tableBodyRow, {'row': {key: rowKey, index: currentRow['rowIndex']}, 'isCurrentRow': true});
              updateEditMode = this._trigger('beforeRowEdit', event, {'rowContext': rowContext});
            }
            else if (!editable && this._hasEditableRow())
            {
              // only trigger the beforeRowEditEnd if we are actually ending an edit
              // fire on the edited row
              var tableBodyRow = this._getTableDomUtils().getTableBodyRow(this._getEditableRowIdx());
              rowKey = this._getRowKeyForRowIdx(this._getEditableRowIdx());
              var rowContext = oj.TableRendererUtils.getRendererContextObject(this, tableBodyRow, {'row': {key: rowKey, index: this._getDataSourceRowIndexForRowKey(rowKey)}, 'isCurrentRow': true});
              updateEditMode = this._trigger('beforeRowEditEnd', event, {'rowContext': rowContext, 'cancelEdit': cancelled});
            }
            else
            {
              // No updates so just exit
              return;
            }
          }
          catch (err)
          {
            return false;
          }

          if (!updateEditMode)
          {
            return false;
          }
          
          // save the old editable row index
          var prevEditableRowIdx = this._getEditableRowIdx();
          var self = this;
          
          if (editable)
          {
            // set the editable row index
            this._setEditableRowIdx(rowIdx);
            // re-render the newly editable row
            this['refreshRow'](rowIdx).then(function()
            {
              self._queueTask(function()
              {
                var focusRowIdx = rowIdx;
                var focusColumnIdx = columnIdx;
                var focusForwardSearch = forwardSearch;
                // set focus on the column in the row
                self._setCellInRowFocus(focusRowIdx, focusColumnIdx, focusForwardSearch);
              });
            });
          }
          else
          {
            this._setEditableRowIdx(null);
            this._getTableDomUtils().getTable().focus();
          }
          
          // clear out the old editable row
          if (prevEditableRowIdx != null)
          {
            // make sure the row still exists before we refresh it
            var prevEditableTableBodyRow = this._getTableDomUtils().getTableBodyRow(prevEditableRowIdx);
            if (prevEditableTableBodyRow != null)
            {
            // re-render the previously editable row, which will be read-only now
            this['refreshRow'](prevEditableRowIdx).then(function(){
              self._queueTask(function()
              {
                var selfPrevEditableRowIdx = prevEditableRowIdx;
                var prevEditableRow = this._getTableDomUtils().getTableBodyRow(selfPrevEditableRowIdx);

                if (prevEditableRow != null)
                {
                    prevEditableRow.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_EDIT_CLASS);
                }
              });
            });
          }
        }
        }
      },
      /**
       * Show the inline message.
       * @param {string} summary
       * @param {string} detail
       * @param {number} severityLevel
       * @private
       */
      _showInlineMessage: function(summary, detail, severityLevel)
      {
        if (!this._inlineMessageShown)
        {
        this._getTableDomUtils().setTableInlineMessage(summary, detail, severityLevel);
        var inlineMessage = this._getTableDomUtils().getTableInlineMessage();
          var tableContainer = this._getTableDomUtils().getTableContainer();
          if (severityLevel == oj.Message.SEVERITY_LEVEL['WARNING'])
          {
            tableContainer.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._WARNING);
            inlineMessage.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._WARNING);
          }
          inlineMessage.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._BLOCK;
          inlineMessage.style[oj.TableDomUtils.CSS_PROP._WIDTH] = '100%';
          inlineMessage.style[oj.TableDomUtils.CSS_PROP._BOX_SIZING] = oj.TableDomUtils.CSS_VAL._BORDER_BOX;
          this._inlineMessageShown = true;
        }
      },
      /**
       * Show the 'No data to display.' or 'Initializing...' message.
       * @private
       */
      _showNoDataMessage: function()
      {
        if (!this._noDataMessageShown)
        {
        var messageRow = this._getTableDomUtils().getTableBodyMessageRow();
        var dataprovider = this._getData();
          var emptyTextMsg = null;
          if (this.options['emptyText'] != null)
          {
            emptyTextMsg = this.options['emptyText'];
          }
          else
          {
            emptyTextMsg = this.getTranslatedString(this._BUNDLE_KEY._MSG_NO_DATA);
          }
        // if data is null then we are initializing
          var messageText = dataprovider != null ? emptyTextMsg : this.getTranslatedString(this._BUNDLE_KEY._MSG_INITIALIZING);
        if (messageRow == null) 
        {
          this._getTableDomUtils().createTableBodyMessageRow(this._getColumnDefs().length, messageText);
        }
        else
        {
          this._getTableDomUtils().setTableBodyMessage(messageText);
        }
          this._noDataMessageShown = true;
        }
      },
      /**
       * Show the Fetching Data... status message.
       * @private
       */
      _showStatusMessage: function()
      {
        if (this._showStatusTimeout)
        {
          return;
        }

        if (!this._statusMessageShown)
        {
          var statusMessage = this._getTableDomUtils().getTableStatusMessage();
          var self = this;
          this._showStatusTimeout = setTimeout(function()
          {
            statusMessage.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._INLINE;
            self._statusMessageShown = true;
            self._getTableDomUtils().refreshTableStatusPosition();
          }, this._getShowStatusDelay());
        }
      },
      /**
       * Retrieve the delay before showing status
       * @return {number} the delay in ms
       * @private
       */
      _getShowStatusDelay: function()
      {
        var delay;

        if (this._defaultOptions == null)
        {
            this._defaultOptions = oj.ThemeUtils.parseJSONFromFontFamily('oj-table-option-defaults');
        }
        delay = parseInt(this._defaultOptions['showIndicatorDelay'], 10);

        return isNaN(delay) ? 0 : delay;
      },
      /**
       * Show the column header sort link
       * @param {number} columnIdx  column index
       * @private
       */
      _showTableHeaderColumnSortLink: function(columnIdx)
      {
        if (this._getColumnDefs()[columnIdx]['sortable'] == this._OPTION_ENABLED)
        {
          var tableHeaderColumn = this._getTableDomUtils().getTableHeaderColumn(columnIdx);

          if (!tableHeaderColumn)
          {
            return;
          }

          // check if the column is currently sorted
          var sorted = $(tableHeaderColumn).data('sorted');

          // we should only show the ascending sort link if the column is not sorted
          if (sorted == null)
          {
            var headerColumnAscLink = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_LINK_CLASS);
            if (headerColumnAscLink.length > 0)
            {
              headerColumnAscLink = headerColumnAscLink[0];
              headerColumnAscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._ENABLED);
              headerColumnAscLink.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
          }
            var headerColumnAsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_CLASS);
            if (headerColumnAsc.length > 0)
            {
              headerColumnAsc = headerColumnAsc[0];
              headerColumnAsc.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
        }
            var headerColumnDsc = tableHeaderColumn.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_CLASS);
            if (headerColumnDsc.length > 0)
            {
              headerColumnDsc = headerColumnDsc[0];
              headerColumnDsc.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
            }
          }
        }
      },
      /**
       * Only process the event handler if it's a double tap
       * @param {Event} event
       * @param {Function} eventHandler
       * @private
       */
      _touchEventDoubleTapFunction: function(event, eventHandler)
      {
        var eventTarget = $(event.target);
            
        if (this._lastTapTime != null &&
          ((new Date()).getTime() - this._lastTapTime) < 250 &&
          this._lastTapTarget[0] == eventTarget[0]) 
        {
          this._lastTapTime = null;
          this._lastTapTarget = null;
          eventHandler();
        } 
        else 
        {
          this._lastTapTarget = eventTarget;
          this._lastTapTime = (new Date()).getTime();
        }
      },
      /**
       * Unregister _focusable(), etc, which were added to the child elements
       * @param {Element} parent div DOM element
       * @private
       */
      _unregisterChildStateListeners: function(parent)
      {
        var self = this;
        var elements = parent.querySelectorAll('*');
        var i;
        for (i = 0; i < elements.length; i++)
        {
          self._UnregisterChildNode(elements[i]);
        }
        self = null;
      },
      /**
       * Unregister event listeners which are registered on datasource.
       * @private
       */
      _unregisterDataSourceEventListeners: function()
      {
        var dataprovider = this._getData();
        // unregister the listeners on the datasource
        if (this._dataProviderEventHandlers != null && dataprovider != null)
        {
          var i, dataProviderEventHandlersCount = this._dataProviderEventHandlers.length;
          for (i = 0; i < dataProviderEventHandlersCount; i++)
            dataprovider.removeEventListener(this._dataProviderEventHandlers[i]['eventType'], this._dataProviderEventHandlers[i]['eventHandler']);
        }
      },
      /**
       * Unregister event listeners for resize the container DOM element.
       * @private
       */
      _unregisterResizeListener: function()
      {
        var element = this._getTableDomUtils().getTableContainer();
        oj.DomUtils.removeResizeListener(element, this._resizeListener);
        this._isResizeListenerAdded = false;
      },
      /**
       * Update the css class from all the cells in a column according to column state
       * @param {number} columnIdx  column index
       * @param {boolean} blur  true or false
       * @private
       */
      _updateHeaderColumnStateCellsClass: function(columnIdx, blur)
      {
        var selected = this._getHeaderColumnSelection(columnIdx);
        var selectedRowIdxs = this._getSelectedRowIdxs();
        var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
        if (tableBodyRows != null && tableBodyRows.length > 0)
        {
          var i, j, tableBodyCell, rowSelected, selectedRowIdxsCount, tableBodyRowsCount = tableBodyRows.length;
          for (i = 0; i < tableBodyRowsCount; i++)
          {
            tableBodyCell = this._getTableDomUtils().getTableBodyCell(i, columnIdx);
            if (!selected)
            {
              rowSelected = false;
              selectedRowIdxsCount = selectedRowIdxs.length;
              for (j = 0; j < selectedRowIdxsCount; j++)
              {
                if (i == selectedRowIdxs[j])
                {
                  rowSelected = true;
                  break;
                }
              }
              if (!rowSelected)
              {
                tableBodyCell.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
              }
            }
            else
            {
              tableBodyCell.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
            }
          }
        }
      },
      /**
       * Update the css class from all the cells in a row according to row state
       * @param {number} rowIdx  row index
       * @param {Object} state  row state
       * @param {boolean} blur  true or false
       * @private
       */
      _updateRowStateCellsClass: function(rowIdx, state, blur)
      {
        var tableBodyCells = this._getTableDomUtils().getTableBodyCells(rowIdx);
        var focused = state.focused;
        var selected = state.selected;
        var hover = state.hover;

        if (!tableBodyCells)
        {
          return;
        }

        if (hover != null)
        {
          var i, tableBodyCellsCount = tableBodyCells.length;
          for (i = 0; i < tableBodyCellsCount; i++)
          {
            if (!hover)
            {
              tableBodyCells[i].classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
            }
            else
            {
              tableBodyCells[i].classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
            }
          }
        }
        if (focused != null)
        {
          var i, tableBodyCellsCount = tableBodyCells.length;
          for (i = 0; i < tableBodyCellsCount; i++)
          {
            if (!focused)
            {
              this._focusOutHandler($(tableBodyCells[i]));
            }
            else
            {
              this._focusInHandler($(tableBodyCells[i]));
            }
          }
        }
        if (selected != null)
        {
          var i, tableBodyCellsCount = tableBodyCells.length;;
          for (i = 0; i < tableBodyCellsCount; i++)
          {
            if (!selected)
            {
              tableBodyCells[i].classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
            }
            else
            {
              tableBodyCells[i].classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._SELECTED);
            }
          }
        }
      },
      _setFinalTask: function(task)
      {
        this._finalTask = (task ? task.bind(this) : undefined);
      },
      _hasPendingTasks: function()
      {
        return this._taskCount > 1 ? true : false;
      },
      _queueTask: function(task)
      {
        var self = this;
        if (!this._pendingTasks)
        {
          this._taskCount = 0;
          this._pendingTasks = Promise.resolve();
          this._setComponentNotReady();
        }
        this._taskCount++;
        this._pendingTasks = this._pendingTasks
          .then(
            function()
            {
		          if (!self._componentDestroyed)
              {
                return task.bind(self)();
              }
            })
        .then(function(value)
        {
          self._taskCount--;
          if (self._taskCount == 0)
          {
            self._pendingTasks = undefined;
            try 
            {
              if (!self._componentDestroyed)
              {
                if (self._finalTask)
                {
                  self._finalTask();
                }
                self._trigger("ready");
              }
            }
            catch(err)
            {
              oj.Logger.error(err);
            }
            // Need to remove busy state even if the component is destroyed
            self._setComponentReady();
            self = null;
          }
          return value;
        },
        function(error)
        {
          self._taskCount--;
          if (self._taskCount == 0)
          {
            self._pendingTasks = undefined;
            oj.Logger.error(error);
            self._setComponentReady();
            self = null;
          }
          
          return Promise.reject(error);
        });

        return this._pendingTasks;
      },

      // @inheritdoc
      _CompareOptionValues: function(option, value1, value2)
      {
        switch(option)
        {
           case 'columns':
           case 'currentRow':
           case 'selection':
              return oj.Object.compareValues(value1, value2);
           default:
              return this._super(option, value1, value2);
      }
      }

      /* Later when needed
      _whenReady: function()
      {
        if (this._pendingTasks)
        {
          return this._pendingTasks;
        }
        return Promise.resolve();
      }
      */
    /**** end internal functions ****/
    }
  )
}());

oj.Components.setDefaultOptions(
  {
    'ojTable':
      {      
        'display': oj.Components.createDynamicPropertyGetter(function()
          {           
            return (oj.ThemeUtils.parseJSONFromFontFamily('oj-table-option-defaults') || {})['display'];
          })
      }
  }
);

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @export
 * @class oj.TableAnimationUtils
 * @classdesc Animation Utils for ojTable
 * @param {jQuery=} component ojTable instance
 * @constructor
 */
oj.TableAnimationUtils = function(component)
{
  this.component = component;
  this.options = component['options'];
  this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.TableAnimationUtils, oj.Object, "oj.TableAnimationUtils");

/**
 * Initializes the instance.
 * @export
 * @memberof oj.TableAnimationUtils
 */
oj.TableAnimationUtils.prototype.Init = function()
{
  oj.TableAnimationUtils.superclass.Init.call(this);
};

/**
 * Animate a tableBodyRow
 * @param {Element} tableBodyRow  DOM element
 * @param {string} action the animation action
 * @return {Promise} Returns a Promise which resolves to true when animation is complete. Will
 * resolve to false if no animations were run.
 * @memberof oj.TableAnimationUtils
 */
oj.TableAnimationUtils.prototype.animateTableBodyRow = function(tableBodyRow, action)
{
  if (this._isAnimationDisabled(action))
  {
    return Promise.resolve(false);
  }
  var self = this;
  return new Promise(function(resolve, reject)
    { 
      self._startAnimation(tableBodyRow, action).then(function()
      {
        resolve(true);
      });
    });
};

/**
 * Animate an array of tableBodyRows
 * @param {Array} tableBodyRowArray Array of tr DOM elements
 * @param {string} action the animation action
 * @return {Promise} Returns a Promise which resolves when animation is complete..
 * @memberof oj.TableAnimationUtils
 */
oj.TableAnimationUtils.prototype.animateTableBodyRows = function(tableBodyRowArray, action)
{
  if (this._isAnimationDisabled(action))
  {
    return Promise.resolve(false);
  }
  var self = this;
  var i, tableBodyRow, animationPromiseArray = [];
  for (i = 0; i < tableBodyRowArray.length; i++)
  {
    tableBodyRow = tableBodyRowArray[i];
    animationPromiseArray.push((function(tableBodyRow)
    {
      return self.animateTableBodyRow(tableBodyRow, action);
    })(tableBodyRow));
  }
  
  if (animationPromiseArray.length > 0)
  {
    return Promise.all(animationPromiseArray);
  }
  return Promise.resolve(false);
};

/**
 * Gets the animation effect for the specific action
 * @param {string} action the action to retrieve the effect
 * @return {Object} the animation effect for the action
 * @memberof oj.TableAnimationUtils
 * @private
 */
oj.TableAnimationUtils.prototype._getAnimationEffect = function(action)
{
  var defaultOptions;

  if (this.defaultAnimations == null)
  {
    defaultOptions = oj.ThemeUtils.parseJSONFromFontFamily('oj-table-option-defaults');
    this.defaultAnimations = defaultOptions['animation'] || {};
  }

  return this.defaultAnimations[action];
};

/**
 * Returns whether animation is disabled
 * @param {string} action the animation action
 * @return {boolean} Returns true or false
 * @private
 * @memberof oj.TableAnimationUtils
 */
oj.TableAnimationUtils.prototype._isAnimationDisabled = function(action)
{
  // check if animation effect is null
  var effect = this._getAnimationEffect(action);
  if (effect == null || 
    effect.length == 0)
  {
    return true;
  }
  return false;
};

/**
 * Utility method to start animation
 * @param {Element} elem element to animate
 * @param {string} action the animation action
 * @param {Object=} effect optional animation effect, if not specified then it will be derived based on action
 * @return {Promise} the promise which will be resolve when animation ends
 * @private
 * @memberof oj.TableAnimationUtils
 */
oj.TableAnimationUtils.prototype._startAnimation = function(elem, action, effect)
{
  if (effect == null)
  {
    effect = this._getAnimationEffect(action);
  }
  return oj.AnimationUtils.startAnimation(elem, action, effect, this.component);
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @export
 * @class oj.TableDndContext
 * @classdesc Drag and Drop Utils for ojTable
 * @param {Object} component ojTable instance
 * @constructor
 */
oj.TableDndContext = function(component)
{
  this.component = component;
  this.options = component['options'];
  this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.TableDndContext, oj.Object, "oj.TableDndContext");

/**
 * Initializes the instance.
 * @memberof oj.TableDndContext
 * @export
 */
oj.TableDndContext.prototype.Init = function()
{
  oj.TableDndContext.superclass.Init.call(this);
};

/**
 * Add oj-drag marker class to cells in a column
 * @param {number} columnIdx  the index of the column to mark
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._addDragMarkerClass = function(columnIdx)
{
  var column = this._getTableDomUtils().getTableHeaderColumn(columnIdx);
  column.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAG);
  this._getTableDomUtils().setTableColumnCellsClass(columnIdx, true, oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAG);
};

/**
 * Remove oj-drag marker class from cells in dragged columns
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._removeDragMarkerClass = function()
{
  var dragColumns = this._getTableDomUtils().getTableHeader().querySelectorAll('.' + oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAG);
  if (dragColumns != null && dragColumns.length > 0)
  {
    var i;
    var dragColumnsCount = dragColumns.length;
    for (i = 0; i < dragColumnsCount; i++)
    {
      dragColumns[i].classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAG);
    }
  }
  this._getTableDomUtils().setTableColumnCellsClass(null, false, oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAG);
};

/**
 * Clone the table container
 * @param {Element} tableContainer  the div DOM object
 * @return {Element} DOM object for the cloned table container
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._cloneTableContainer = function(tableContainer)
{
  var tableContainerClone;
  var tableBody = tableContainer.querySelectorAll(oj.TableDomUtils.DOM_ELEMENT._TBODY)[0];
  
  var scrollLeft = $(tableBody).scrollLeft();
  var scrollTop = $(tableBody).scrollTop();
  
  tableContainerClone = tableContainer.cloneNode(true);
  var tableBodyClone = tableContainerClone.querySelectorAll(oj.TableDomUtils.DOM_ELEMENT._TBODY)[0];
  
  // Use absolute positioning with a large negative top to put it off the screen without
  // affecting the scrollbar.  Fixed positioning does not work on Safari.
  // Set both overflow and overflow-x/y because on IE, setting overflow does not
  // affect overflow-x/y that have been explicitly set.
  tableBodyClone.style[oj.TableDomUtils.CSS_PROP._OVERFLOW] = oj.TableDomUtils.CSS_VAL._HIDDEN;
  tableBodyClone.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_X] = oj.TableDomUtils.CSS_VAL._HIDDEN;
  tableBodyClone.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_Y] = oj.TableDomUtils.CSS_VAL._HIDDEN;
  tableBodyClone.style[oj.TableDomUtils.CSS_PROP._BACKGROUND_COLOR] = oj.TableDomUtils.CSS_VAL._TRANSPARENT;
  tableBodyClone.style[oj.TableDomUtils.CSS_PROP._WIDTH] = $(tableBody).width() + oj.TableDomUtils.CSS_VAL._PX;
  tableBodyClone.style[oj.TableDomUtils.CSS_PROP._HEIGHT] = $(tableBody).height() + oj.TableDomUtils.CSS_VAL._PX;
  tableContainerClone.style[oj.TableDomUtils.CSS_PROP._POSITION] = oj.TableDomUtils.CSS_VAL._ABSOLUTE;
  tableContainerClone.style[oj.TableDomUtils.CSS_PROP._TOP] = -10000 + oj.TableDomUtils.CSS_VAL._PX;
  $(tableBodyClone).scrollLeft(scrollLeft * 1.0);        
  $(tableBodyClone).scrollTop(scrollTop * 1.0);
  
  document.body.appendChild(tableContainerClone); //@HTMLUpdateOK
  
  return tableContainerClone;
};

/**
 * Destroy the drag image
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._destroyDragImage = function()
{
  if (this._dragImage)
  {
    $(this._dragImage).remove();
    this._dragImage = null;
  }
};

/**
 * Get the column index of the header target of an event
 * @param {Event} event  jQuery event object
 * @return {number} the column index of the header target
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._getEventColumnIndex = function(event)
{
  return this._getTableDomUtils().getElementColumnIdx(event.currentTarget);
};

/**
 * Get the index of the row under the pointer
 * @param {Event} event  jQuery event object
 * @return {number} index of the row under the pointer
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._getOverRowIndex = function(event)
{
  var newRowIndex;
  var jqOverRow = $(event.target).closest(oj.TableDomUtils.DOM_ELEMENT._TR);
  var context = this.component.getContextByNode(event.target);
  if (context && context['subId'] == 'oj-table-cell')
  {
    newRowIndex = context['rowIndex'];
  }
  else if (jqOverRow && jqOverRow.hasClass(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_DRAG_INDICATOR_BEFORE_CLASS))
  {
    newRowIndex = this._dropRowIndex;
  }
  else
  {            
    // If we are not in a cell and not in the indicator row, we are in an empty part 
    // of the table body, so add any new row to the end.
    var tableBodyRows = this._getTableDomUtils().getTableBodyRows();
    newRowIndex = tableBodyRows ? tableBodyRows.length : 0;
  }
  
  return newRowIndex;
};
 
/**
 * Get the TableDomUtils object from the ojTable component.
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._getTableDomUtils = function()
{
  return this.component._getTableDomUtils();
};

/**
 * Handle dragstart on column
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragstart, returning false cancel the drag operation.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnDragStart = function(event)
{
  if (this._isColumnReorderEnabled())
  {
    var columnIdx = this._getEventColumnIndex(event);
    
    this._dragStartColumnIdx = columnIdx;
    
    this._setReorderColumnsDataTransfer(event, columnIdx);

    // Remove any row and column selection
    this.component._setSelection(null);
    this.component.option('selection', null, {'_context': {writeback: true, internalSet: true}});

    // Remove text selection
    // Text selection doesn't get cleared unless we put it in a setTimeout
    setTimeout(function()
    {
      window.getSelection().removeAllRanges();
    }, 0);
    
    this._addDragMarkerClass(columnIdx);
    
    return true;
  }
};

/**
 * Handle dragend on column
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragend, returning false has no special implication.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnDragEnd = function(event)
{
  if (this._isColumnReorderEnabled())
  {
    this.setTableHeaderColumnDraggable(null, false);
    this._dragStartColumnIdx = null;
    this._getTableDomUtils().removeDragOverIndicatorColumn();
    this._getTableDomUtils().removeTableHeaderColumnDragImage();
    
    this._removeDragMarkerClass();
  }
};

/**
 * Handle dragenter on column
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragenter, returning false indicates target can accept the data.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnDragEnter = function(event)
{
  if (this._isColumnReordering())
  {
    return;
  }
  
  var columnIdx = this._getEventColumnIndex(event);

  return this._invokeDropCallback('columns', 'dragEnter', event,
                                  {'columnIndex': columnIdx});
};

/**
 * Handle dragover on column reordering
 * @param {Event} event  jQuery event object
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnReorderDragOver = function(event)
{
  var columnIdx = this._getEventColumnIndex(event);
  var dragStartColumnIdx = this._dragStartColumnIdx;
  
  if (dragStartColumnIdx != null && dragStartColumnIdx != columnIdx)
  {
    this._currentDropColumnBefore = this._isDragOverBeforeColumn(event);

    // Check the current before/after column position against indicator position
    // to see if we need to move the indicator
    if (!(this._currentDropColumnBefore && dragStartColumnIdx == columnIdx  - 1) &&
        !(!this._currentDropColumnBefore && dragStartColumnIdx == columnIdx + 1))
    {
      this._getTableDomUtils().displayDragOverIndicatorColumn(columnIdx, this._currentDropColumnBefore);
    }
    event.preventDefault();
  }
};

/**
 * Handle dragover on column
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragover, returning false indicates target can accept the data.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnDragOver = function(event)
{
  if (this._isColumnReordering())
  {
    return this.handleColumnReorderDragOver(event);
  }
  
  var columnIdx = this._getEventColumnIndex(event);
  this._currentDropColumnBefore = this._isDragOverBeforeColumn(event);

  var returnValue = this._invokeDropCallback('columns', 'dragOver', event,
                                             {'columnIndex': columnIdx});  
  
  if (returnValue === false || event.isDefaultPrevented())
  {
    this._getTableDomUtils().displayDragOverIndicatorColumn(columnIdx, this._isDragOverBeforeColumn(event));                
  }
  
  return returnValue;
};

/**
 * Handle dragleave on column
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragleave, returning false has no special implication.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnDragLeave = function(event)
{
  if (this._isColumnReordering())
  {
    return;
  }
  
  this._getTableDomUtils().removeDragOverIndicatorColumn();

  var columnIdx = this._getEventColumnIndex(event);

  return this._invokeDndCallback('drop', 'columns', 'dragLeave', event,
                                 {'columnIndex': columnIdx});
};

/**
 * Handle drop on column reordering
 * @param {Event} event  jQuery event object
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnReorderDrop = function(event)
{
  var columnIdx = this._getEventColumnIndex(event);
  if (!this._currentDropColumnBefore)
  {
    columnIdx++;
  }
  this.component._columnsDestMap = this._getTableDomUtils().moveTableHeaderColumn(this._dragStartColumnIdx, columnIdx, event);
  event.preventDefault();
};

/**
 * Handle drop on column
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of drop, returning false indicates target can accept the data.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleColumnDrop = function(event)
{
  if (this._isColumnReordering())
  {
    return this.handleColumnReorderDrop(event);
  }
  
  this._getTableDomUtils().removeDragOverIndicatorColumn();
  var columnIdx = this._getEventColumnIndex(event);
  if (!this._currentDropColumnBefore)
  {
    columnIdx++;
  }

  return this._invokeDropCallback('columns', 'drop', event,
                                  {'columnIndex': columnIdx});
};

/**
 * Handle dragstart on row
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragstart, returning false cancel the drag operation.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleRowDragStart = function(event)
{
  var dragOption = this.options['dnd']['drag'];
  if (dragOption && dragOption['rows'])
  {
    var ui = this._setDragRowsDataTransfer(event, dragOption['rows']['dataTypes'], this.component._getSelectedRowIdxs());
    if (ui)
    {
      return this._invokeDndCallback('drag', 'rows', 'dragStart', event, ui);
    }

    // Return false to cancel the dragstart event if no data
    return false;
  }
};

/**
 * Handle drag on row
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of drag, returning false has no special implication.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleRowDrag = function(event)
{
  return this._invokeDndCallback('drag', 'rows', 'drag', event);
};

/**
 * Handle dragend on row
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragend, returning false has no special implication.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleRowDragEnd = function(event)
{
  // Perform any cleanup
  this._destroyDragImage();
  
  return this._invokeDndCallback('drag', 'rows', 'dragEnd', event);
};

/**
 * Handle dragenter on row
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragenter, returning false indicates target can accept the data.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleRowDragEnter = function(event)
{
  var newRowIndex = this._getOverRowIndex(event);

  var returnValue = this._invokeDropCallback('rows', 'dragEnter', event,
                                             {'rowIndex': newRowIndex});
  
  if (returnValue === false || event.isDefaultPrevented())
  {
    this._updateDragRowsState(event, newRowIndex);
  }
  
  return returnValue;
};

/**
 * Handle dragover on row
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragover, returning false indicates target can accept the data.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleRowDragOver = function(event)
{
  return this._invokeDropCallback('rows', 'dragOver', event,
                                  {'rowIndex': this._dropRowIndex});
};

/**
 * Handle dragleave on row
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragleave, returning false has no special implication.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleRowDragLeave = function(event)
{
  var returnValue = this._invokeDndCallback('drop', 'rows', 'dragLeave', event, 
                                            {'rowIndex': this._dropRowIndex});

  // Remove the indicator row if we are no longer in table body since
  // this may be the last dnd event we get.
  if (!this._isDndEventInElement(event, event.currentTarget))
  {
    this._getTableDomUtils().removeDragOverIndicatorRow();
    this._dropRowIndex = null;
  }
  
  return returnValue;
};

/**
 * Handle drop on row
 * @param {Event} event  jQuery event object
 * @return {boolean|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of drop, returning false indicates target can accept the data.
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.handleRowDrop = function(event)
{
  var dropRowIndex = this._dropRowIndex;
  
  this._getTableDomUtils().removeDragOverIndicatorRow();
  this._dropRowIndex = null;
  
  return this._invokeDropCallback('rows', 'drop', event,
                                  {'rowIndex': dropRowIndex});
};

/**
 * Hide unselected rows in a tbody element
 * @param {Element} tableContainerClone  the cloned div DOM object
 * @param {Array.<number>} selArray  array of selected row index
 * @return {number|null} row index of the first selected row
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._hideUnselectedRows = function(tableContainerClone, selArray)
{
  var tableBodyRows;
  var firstRow = null;
    
  tableBodyRows = tableContainerClone.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS);
  for (var i = tableBodyRows.length - 1; i >= 0; i--)
  {
    if (selArray.indexOf(i) == -1)
    {
      tableBodyRows[i].style[oj.TableDomUtils.CSS_PROP._VISIBILITY] = oj.TableDomUtils.CSS_VAL._HIDDEN;
    }
    else
    {
      firstRow = i;
    }
  }
  
  // hide thead and tfoot
  var tableHeader = tableContainerClone.querySelectorAll(oj.TableDomUtils.DOM_ELEMENT._THEAD);
  
  if (tableHeader != null && tableHeader.length > 0)
  {
    tableHeader[0].style[oj.TableDomUtils.CSS_PROP._VISIBILITY] = oj.TableDomUtils.CSS_VAL._HIDDEN;
  }
  
  var tableFooter = tableContainerClone.querySelectorAll(oj.TableDomUtils.DOM_ELEMENT._TFOOT);
  
  if (tableFooter != null && tableFooter.length > 0)
  {
    tableFooter[0].style[oj.TableDomUtils.CSS_PROP._VISIBILITY] = oj.TableDomUtils.CSS_VAL._HIDDEN;
  }
  
  return firstRow;
};

/**
 * Invoke user callback function specified in a drag or drop option
 * @param {string} dndType  the dnd option type ('drag' or 'drop')
 * @param {string} itemType  the drag or drop item type such as 'rows'
 * @param {string} callbackType  the callback type such as 'dragStart'
 * @param {Event} event  the jQuery Event object from drag and drop event
 * @param {Object} [ui]  additional properties to pass to callback function
 * @return {boolean} the return value from the callback function
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._invokeDndCallback = function(dndType, itemType, callbackType, event, ui)
{
  var dndOption = this.options['dnd'][dndType];
  var returnValue;
  
  if (dndOption && dndOption[itemType])
  {
    // First let the callback decide if data can be accepted
    var callback = dndOption[itemType][callbackType];          
    if (callback && typeof callback == 'function')
    {
      try
      {
        if (this.component._IsCustomElement())
        {
          // For custom element, pass original DOM event and ignore return value.
          callback(event.originalEvent, ui);
        }
        else
        {
          // Hoist dataTransfer object from DOM event to jQuery event
          event.dataTransfer = event.originalEvent.dataTransfer;
          
          // Invoke callback function
          returnValue = callback(event, ui);
        }
      }
      catch (e)
      {
        oj.Logger.error('Error: ' + e);
      }
    }
  }
  
  return returnValue;
};

/**
 * Invoke user callback function specified in a drop option
 * @param {string} itemType  the drag or drop item type such as 'rows'
 * @param {string} callbackType  the callback type such as 'dragStart'
 * @param {Event} event  the jQuery Event object from drag and drop event
 * @param {Object} [ui]  additional properties to pass to callback function
 * @return {boolean} the return value from the callback function
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._invokeDropCallback = function(itemType, callbackType, event, ui)
{
  var returnValue = this._invokeDndCallback('drop', itemType, callbackType, event, ui);

  if (returnValue === undefined)
  {
    if (this._matchDragDataType(event, itemType))
    {
      event.preventDefault();
    }
  }

  return returnValue;
};

/**
 * Whether the column reorder is enabled
 * @return {boolean} true or false
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._isColumnReorderEnabled = function()
{
  var dndOption = this.options['dnd'];
  
  if (dndOption && dndOption['reorder'] &&
      dndOption['reorder']['columns'] == this.component._OPTION_ENABLED)
  {
    return true;
  }
  
  return false;
};

/**
 * Return true if column reorder is in progress
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._isColumnReordering = function()
{
  return (this._dragStartColumnIdx != null);
};

/**
 * Return true if the mouse/touch point of a dnd event is in an element
 * @param {Event} event  jQuery event object
 * @param {EventTarget} element  DOM element
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._isDndEventInElement = function(event, element)
{
  var rect = element.getBoundingClientRect();
  var domEvent = event.originalEvent;
  
  // clientX and clientY are only available on the original DOM event.
  // IE returns client rect in non-integer values.  Trim it down to make sure
  // the pointer is really still in the element.
  return (domEvent.clientX >= Math.ceil(rect.left) && domEvent.clientX < Math.floor(rect.right) &&
          domEvent.clientY >= Math.ceil(rect.top) && domEvent.clientY < Math.floor(rect.bottom));
};

/**
 * @param {Object} event event
 * @return {boolean} <code>true</code> if the event is considered before the
 *                   column, <code>false</code> otherwise.
 * @private
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype._isDragOverBeforeColumn = function(event)
{
  var columnRect = event.currentTarget.getBoundingClientRect();
  
  if (event.originalEvent.clientX != null)
  {
    var cursorPosX = columnRect.right - event.originalEvent.clientX;

    // First figure out whether the pointer is on the left half or right half    
    var onRightHalf = (cursorPosX < (columnRect.right - columnRect.left) / 2);
    
    // Whether we are before/after the column depends on the reading direction
    return (oj.DomUtils.getReadingDirection() === "rtl") ? onRightHalf : !onRightHalf;
  }
  return false;
};

/**
 * Return true if the data types from dnd event match one of the values in an array
 * @param {Event} event  jQuery event object for a drag and drop event
 * @param {string} itemType  The drop item type such as "rows" or "columns".
 * @return {boolean} true if one of the types in dragDataTypes and allowedTypes matches  
 * @memberof oj.TableDndContext
 * @private     
 */
oj.TableDndContext.prototype._matchDragDataType = function(event, itemType)
{
  var dragDataTypes = event.originalEvent.dataTransfer.types;
  var dndOption = this.options['dnd']['drop'];
  
  if (dndOption && dndOption[itemType] && dndOption[itemType]['dataTypes'])
  {
    var allowedTypes = dndOption[itemType]['dataTypes'];
    var allowedTypeArray = (typeof allowedTypes == 'string') ? [allowedTypes] : allowedTypes;
    
    // dragDataTypes can be either an array of strings (Chrome) or a 
    // DOMStringList (Firefox and IE).  For cross-browser compatibility, use its 
    // length and index to traverse it.
    for (var i = 0; i < dragDataTypes.length; i++)
    {
      if (allowedTypeArray.indexOf(dragDataTypes[i]) >= 0)
        return true;
    }
  }

  return false;
};

/**
 * Set the draggable attribute of a header column
 * @param {Element} headerColumn  the header column DOM element
 * @param {boolean} draggable  true if header column is draggable; false otherwise.
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._setHeaderColumnDraggable = function(headerColumn, draggable)
{
  headerColumn.draggable = draggable;
  
  if (draggable)
  {
    headerColumn.style[oj.TableDomUtils.CSS_PROP._CURSOR] = oj.TableDomUtils.CSS_VAL._MOVE;
    headerColumn.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAGGABLE);
  }
};
 
/**
 * Clear the draggable attribute of a header column
 * @param {Element} headerColumn  the header column DOM element
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._clearHeaderColumnDraggable = function(headerColumn)
{
  headerColumn.draggable = '';
  headerColumn.style[oj.TableDomUtils.CSS_PROP._CURSOR] = oj.TableDomUtils.CSS_VAL._AUTO;
  headerColumn.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAGGABLE);
};

/**
 * Set the data of the selected rows into the dataTransfer object
 * @param {Event} nativeEvent  DOM event object 
 * @param {string | Array.<string>} dataTypes  a data type or array of data types
 * @param {Array.<Object>} rowDataArray  array of row data
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._setDragRowsData = function(nativeEvent, dataTypes, rowDataArray)
{
  if (dataTypes)
  {
    var dataTransfer = nativeEvent.dataTransfer;
    var jsonStr = JSON.stringify(rowDataArray);

    if (typeof dataTypes == 'string')
    {
      dataTransfer.setData(dataTypes, jsonStr);
    }
    else
    {
      for (var i = 0; i < dataTypes.length; i++)
      {
        dataTransfer.setData(dataTypes[i], jsonStr);
      }
    }
  }
};

/**
 * Set the data and drag image of the selected row into the dataTransfer object
 * @param {Event} event  jQuery event object 
 * @param {string | Array.<string>} dataTypes  a data type or array of data types
 * @param {Array.<number>} selArray  array of selected row index
 * @memberof oj.TableDndContext
 * @private
 */      
oj.TableDndContext.prototype._setDragRowsDataTransfer = function(event, dataTypes, selArray)
{
  var rowDataArray = [];

  // Add data for selected rows to an array
  for (var i = 0; i < selArray.length; i++)
  {
    var row = this.component.getDataForVisibleRow(selArray[i]);
    if (row)
    {
      rowDataArray.push(row);
    }
  }

  // Use the row data array as drag data and create drag image
  if (rowDataArray.length)
  {
    this._setDragRowsData(event.originalEvent, dataTypes, rowDataArray);
    
    this._dragImage = this._setDragRowsImage(event.originalEvent, 
                                             $(event.currentTarget).closest(oj.TableDomUtils.DOM_ELEMENT._TABLE).parent()[0],
                                             selArray);
    
    return {'rows': rowDataArray};
  }

  return null;
};

/**
 * Set a drag image of the selected rows into the dataTransfer object
 * @param {Event} nativeEvent  DOM event object 
 * @param {Element} tableContainer  the div DOM object
 * @param {Array.<number>} selArray  array of selected row index
 * @return {Element} DOM object for the cloned table container
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._setDragRowsImage = function(nativeEvent, tableContainer, selArray)
{
  var tableContainerClone = this._cloneTableContainer(tableContainer);
  this._hideUnselectedRows(tableContainerClone, selArray);
  var rect = tableContainer.getBoundingClientRect();
  var tableBody = tableContainer.querySelectorAll(oj.TableDomUtils.DOM_ELEMENT._TBODY)[0];
  var scrollLeft = $(tableBody).scrollLeft();
  var scrollTop = $(tableBody).scrollTop();
  var offsetX = nativeEvent.clientX - rect.left + scrollLeft;
  var offsetY = nativeEvent.clientY - rect.top + scrollTop;
  nativeEvent.dataTransfer.setDragImage(tableContainerClone, offsetX, offsetY);
  
  return tableContainerClone;
};

/**
 * Set the data and drag image for column reorder into the dataTransfer object
 * @param {Event} event  jQuery event object 
 * @param {number} columnIdx  the index of the column being dragged
 * @memberof oj.TableDndContext
 * @private
 */      
oj.TableDndContext.prototype._setReorderColumnsDataTransfer = function(event, columnIdx)
{
  var dataTransfer = event.originalEvent.dataTransfer;
  var tableIdHashCode = this._getTableDomUtils().hashCode(this._getTableDomUtils().getTableUID());
  dataTransfer.setData('Text', this.component._DND_REORDER_TABLE_ID_DATA_KEY + ':' + tableIdHashCode + ':' + columnIdx);
  var dragImage = this._getTableDomUtils().createTableHeaderColumnDragImage(columnIdx);
  try
  {
    dataTransfer.setDragImage(dragImage, 0, 0);
  }
  catch(e)
  {
    // MS Edge doesn't allow calling setDragImage()
  }
};

/**
 * Set the column draggable
 * @param {number|null} columnIdx  column index
 * @param {boolean} draggable true or false
 * @memberof oj.TableDndContext
 */
oj.TableDndContext.prototype.setTableHeaderColumnDraggable = function(columnIdx, draggable)
{
  // set draggable only if DnD is enabled
  if (!this._isColumnReorderEnabled())
  {
    return;
  }
  
  var headerColumns = this._getTableDomUtils().getTableHeaderColumns();

  if (headerColumns != null && headerColumns.length > 0)
  {
    var i;
    
    for (i = 0; i < headerColumns.length; i++)
    {
      var headerColumn = headerColumns[i];
      
      if (columnIdx != null && i == columnIdx)
      {
        this._setHeaderColumnDraggable(headerColumn, draggable);
      }
      else
      {
        this._clearHeaderColumnDraggable(headerColumn);
      }
    }
  }
};

/**
 * Update the state of dragging rows
 * @param {Event} event  jQuery event object
 * @param {number} newRowIndex  index of the row that can receive the drop
 * @memberof oj.TableDndContext
 * @private
 */
oj.TableDndContext.prototype._updateDragRowsState = function(event, newRowIndex)
{
  if (this._dropRowIndex != newRowIndex)
  {
    var overRow = $(event.target).closest(oj.TableDomUtils.DOM_ELEMENT._TR)[0];
    this._dropRowIndex = newRowIndex;
    this._getTableDomUtils().displayDragOverIndicatorRow(this._dropRowIndex, true, overRow);
  }
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @export
 * @class oj.TableDomUtils
 * @classdesc DOM Utils for ojTable
 * @param {Object} component ojTable instance
 * @constructor
 */
oj.TableDomUtils = function(component)
{
  this.component = component;
  this.options = component['options'];
  this.element = component['element'][0];
  this.OuterWrapper = component.OuterWrapper;
  this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.TableDomUtils, oj.Object, "oj.TableDomUtils");

/**
 * Initializes the instance.
 * @export
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.Init = function()
{
  oj.TableDomUtils.superclass.Init.call(this);
};

/**
 * Clear any cached DOM
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.clearCachedDom = function()
{
  this.clearCachedDomRowData();
  this._clearCachedDom(null);
  this._tableDimensions = null;
}

/**
 * Clear any cached DOM rows
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.clearCachedDomRowData = function()
{
  this._cachedDomTableBodyRows = null;
}

/**
 * Create a span element for acc purposes
 * @param {string} text span text
 * @param {string|null} className css class
 * @return {Element} span DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createAccLabelSpan = function(text, className)
{
  var accLabel = document.createElement('span');
  if (className != null)
  {
    accLabel.classList.add(className);
  }
  accLabel.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  accLabel.appendChild(document.createTextNode(text)); //@HTMLUpdateOK

  return accLabel;
};

/**
 * Add a default context menu to the table container if there is none. If there is
 * a context menu set on the table options we use that one. Add listeners
 * for context menu before show and select.
 * @param {function(Object)} handleContextMenuSelect function called for menu select
 * @return {Element} ul DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createContextMenu = function(handleContextMenuSelect)
{
  var menuContainer = this.component._GetContextMenu();
  var self = this;
  var enableNonContigousSelectionMenu = this.component._isTouchDevice() ? this.component._getRowSelectionMode() == this.component._OPTION_SELECTION_MODES._MULTIPLE : false;

  if (menuContainer)
  {
    var listItems = menuContainer.querySelectorAll('[data-oj-command]');
    if (listItems != null && listItems.length > 0)
    {
      var command;
      var i;
      var newItem;
      for (i = 0; i < listItems.length; i++)
      {
        if (listItems[i].tagName === "OJ-OPTION" || listItems[i].getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._A).length === 0)
        {
          command = listItems[i].getAttribute('data-oj-command').split("-");
          newItem = self.createContextMenuItem(command[command.length - 1], this.component._IsCustomElement());
          if (listItems[i].tagName === "OJ-OPTION")
          {
            listItems[i].innerHTML = newItem.innerHTML; //@HTMLUpdateOK
          }
          else
          {
            $(listItems[i]).replaceWith(newItem); //@HTMLUpdateOK
          }
        }
      }
      this._menuContainer = menuContainer;
      if ($(menuContainer).data('oj-ojMenu'))
      {
          if (this.component._IsCustomElement())
              menuContainer.refresh();
          else
              $(menuContainer).ojMenu('refresh');
      }

      if (this.component._IsCustomElement())
          menuContainer.addEventListener("ojAction", handleContextMenuSelect);
      else
          $(menuContainer).on("ojselect", handleContextMenuSelect);
    }
  }
  else if (this.component._isTableSortable() ||
    enableNonContigousSelectionMenu ||
    this.component._isTableColumnsResizable())
  {
    menuContainer = document.createElement(oj.TableDomUtils.DOM_ELEMENT._UL);
    menuContainer.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._NONE;
    menuContainer.setAttribute(oj.TableDomUtils.DOM_ATTR._ID, this.getTableId() + '_contextmenu');
    this.getTableContainer().appendChild(menuContainer); //@HTMLUpdateOK
    if (this.component._isTableSortable())
    {
      var sortMenu = this.createContextMenuItem('sort');
      menuContainer.appendChild(sortMenu); //@HTMLUpdateOK
    }
    if (enableNonContigousSelectionMenu)
    {
      var nonContigousSelectionMenu = this.createContextMenuItem('enableNonContiguousSelection');
      menuContainer.appendChild(nonContigousSelectionMenu); //@HTMLUpdateOK
    }
    if (this.component._isTableColumnsResizable())
    {
      var resizeMenu = this.createContextMenuItem('resize');
      menuContainer.appendChild(resizeMenu); //@HTMLUpdateOK
    }
    $(menuContainer).ojMenu();
    this._menuContainer = menuContainer;
    $(menuContainer).on("ojselect", handleContextMenuSelect);

    var oldMenuOption = this.options["contextMenu"];
    this.component.option("contextMenu", '#' + menuContainer.getAttribute(oj.TableDomUtils.DOM_ATTR._ID), {'_context': {internalSet: true}});

    // fulfill _GetDefaultContextMenu contract
    this.component._defaultContextMenu = menuContainer;
  }
  
  return menuContainer;
};

/**
 * Builds a menu for a command, takes care of submenus where appropriate
 * @param {string} command the string to look up command value for as well as translation
 * @param {boolean=} useOjOption whether oj-option tag should be used, which is the case for custom element (except for default menu)
 * @return {Element} li DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createContextMenuItem = function(command, useOjOption)
{
  if (command === 'sort')
  {
    var sortMenuListItem = this.createContextMenuListItem(command, useOjOption);
    var listElement = document.createElement(oj.TableDomUtils.DOM_ELEMENT._UL);
    sortMenuListItem.appendChild(listElement); //@HTMLUpdateOK
    listElement.appendChild(this.createContextMenuListItem('sortAsc', useOjOption)); //@HTMLUpdateOK
    listElement.appendChild(this.createContextMenuListItem('sortDsc', useOjOption)); //@HTMLUpdateOK 
    return sortMenuListItem;
  }
  else if (command === 'sortAsc')
  {
    return this.createContextMenuListItem(command, useOjOption);
  }
  else if (command === 'sortDsc')
  {
    return this.createContextMenuListItem(command, useOjOption);
  }
  else if (command == 'enableNonContiguousSelection')
  {
    return this.createContextMenuListItem(command, useOjOption);
  }
  else if (command == 'disableNonContiguousSelection')
  {
    return this.createContextMenuListItem(command, useOjOption);
  }
  else if (command === 'resize')
  {
    // create the resize popup too
    this.createContextMenuResizePopup(0);
    return this.createContextMenuListItem(command);
  }
  return null;
};

/**
 * Builds a context menu list item from a command
 * @param {string} command the string to look up command value for as well as translation
 * @param {boolean=} useOjOption whether oj-option tag should be used, which is the case for custom element (except for default menu)
 * @return {Element} DOM element for menu item
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createContextMenuListItem = function(command, useOjOption)
{
  var contextMenuListItem = document.createElement(useOjOption ? 'oj-option' : 'li');
  contextMenuListItem.setAttribute('data-oj-command', 'oj-table-' + command);
  contextMenuListItem.appendChild(this.createContextMenuLabel(command, useOjOption)); //@HTMLUpdateOK

  return contextMenuListItem;
};

/**
 * Builds a context menu label by looking up command translation
 * @param {string} command the string to look up translation for
 * @param {boolean=} useOjOption whether oj-option tag should be used, which is the case for custom element (except for default menu)
 * @return {Node} a DOM node
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createContextMenuLabel = function(command, useOjOption)
{
  var commandString = null;
  if (command == 'sort')
  {
    commandString = this.component.getTranslatedString('labelSort');
  }
  else if (command == 'sortAsc')
  {
    commandString = this.component.getTranslatedString('labelSortAsc');
  }
  else if (command == 'sortDsc')
  {
    commandString = this.component.getTranslatedString('labelSortDsc');
  }
  else if (command == 'enableNonContiguousSelection')
  {
    commandString = this.component.getTranslatedString('labelEnableNonContiguousSelection');
  }
  else if (command == 'disableNonContiguousSelection')
  {
    commandString = this.component.getTranslatedString('labelDisableNonContiguousSelection');
  }
  else if (command == 'resize')
  {
    commandString = this.component.getTranslatedString('labelResize');
  }
  
  var textNode = document.createTextNode(commandString);
  
  // for custom elements, no <a> tag is required
  if (useOjOption)
    return textNode;
  else
  {
    var contextMenuLabel = document.createElement(oj.TableDomUtils.DOM_ELEMENT._A);
    contextMenuLabel.setAttribute(oj.TableDomUtils.DOM_ATTR._HREF, '#');
    contextMenuLabel.appendChild(textNode); //@HTMLUpdateOK
    
    return contextMenuLabel;
  }
};

/**
 * Build the html for the resize popup and add it to the root node
 * @param {number} initialSize the initial size to put in the spinner
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createContextMenuResizePopup = function(initialSize)
{
  initialSize = Math.round(initialSize);
  var popup = this.getContextMenuResizePopup();

  if (popup == null)
  {
   //create the base popup
   var tableContainer = this.getTableContainer();
   popup = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
   popup.setAttribute('id', this.getTableId() + '_resize_popup');
   popup.setAttribute('data-oj-context', '');
   tableContainer.appendChild(popup); //@HTMLUpdateOK
   var popupBody = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
   var popupFooter = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
   popup.appendChild(popupBody); //@HTMLUpdateOK
   popup.appendChild(popupFooter); //@HTMLUpdateOK

   //create the popup content
   var spinner = document.createElement(oj.TableDomUtils.DOM_ELEMENT._INPUT);
   spinner.setAttribute('id', this.getTableId() + '_resize_popup_spinner');
   var spinnerLabel = document.createElement(oj.TableDomUtils.DOM_ELEMENT._LABEL);
   spinnerLabel.setAttribute('for', this.getTableId() + '_resize_popup_spinner');
   spinnerLabel.textContent = this.component.getTranslatedString('labelResizePopupSpinner');
   var popupOKButton = document.createElement(oj.TableDomUtils.DOM_ELEMENT._BUTTON);
   popupOKButton.setAttribute('id', this.getTableId() + '_resize_popup_popupsubmit');

   popupBody.appendChild(spinnerLabel); //@HTMLUpdateOK
   popupBody.appendChild(spinner); //@HTMLUpdateOK
   popupFooter.appendChild(popupOKButton); //@HTMLUpdateOK

   $(popupOKButton).ojButton({component: 'ojButton', label: this.component.getTranslatedString('labelResizePopupSubmit')});
   $(popupOKButton).on('click', this.component._handleContextMenuResizePopup.bind(this.component));
    $(spinner).ojInputNumber({component: 'ojInputNumber', max: 10000, min: 10, step: 1, value: initialSize, validators: [{
          type: 'regExp',
          options: {
            pattern: '^[1-9][0-9]*$',
            messageDetail: this.component.getTranslatedString('msgColumnResizeWidthValidation')
          }
        }]});
   $(popup).ojPopup({modality: 'modal', position: {my: 'start top', at: 'start bottom', collision: 'flipfit'}});
   this._cacheDomElement(this.getTableId() + '_resize_popup', popup);
 } 
 else 
 {
   var spinner = document.getElementById(this.getTableId() + '_resize_popup_spinner');
   $(spinner).ojInputNumber('option', 'value', initialSize);
 }
  return popup;
};

/**
 * Create the initial empty table
 * @param {boolean} isTableHeaderless is table headerless
 * @param {boolean} isTableFooterless is table footerless
 * @return {Element} table DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createInitialTable = function(isTableHeaderless, isTableFooterless)
{
  var table = this.getTable();
  this.createTableContainer();

  if (!isTableHeaderless)
  {
    this.createTableHeader();
  }
  if (!isTableFooterless)
  {
    this.createTableFooter();
  }
  this.createTableBody();
  this.createTableStatusMessage();
  this.createTableInlineMessage();
  this.createTableStatusAccNotification();

  return table;
};

/**
 * Create an empty tbody element with appropriate styling
 * @return {Element} tbody DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBody = function()
{
  var table = this.getTable();
  var tableBody = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TBODY);
  table.appendChild(tableBody); //@HTMLUpdateOK
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_CLASS, tableBody);

  return tableBody;
};

/**
 * Create an empty td element with appropriate styling
 * @param {number} rowIdx  row index
 * @param {number} columnIdx  column index
 * @return {Element} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBodyCell = function(rowIdx, columnIdx)
{
  var tableBodyCell = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TD);

  return tableBodyCell;
};

/**
 * Create a checkbox for accessibility row selection
 * @param {number} rowIdx  row index
 * @param {Object} rowKey  row key
 * @param {Object} rowHashCode  row hash code
 * @param {Element} tableBodyRow  tr DOM element
 * @param {boolean} isNew is new row
 * @return {Element} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBodyCellAccSelect = function(rowIdx, rowKey, rowHashCode, tableBodyRow, isNew)
{
  var accSelectionCell = null;
  
  if (!isNew)
  {
    accSelectionCell = this.getTableBodyCellAccSelect(tableBodyRow);
  }
  
  if (accSelectionCell != null)
  {
    return accSelectionCell;
  }
  
  var rowKeyStr = rowKey != null ? rowKey.toString() : rowIdx.toString();
  var rowKeyStrHashCode = rowHashCode == null ? this.hashCode(rowKeyStr) : rowHashCode;

  accSelectionCell = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TD);
  accSelectionCell.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_ACC_SELECT_CLASS);
  accSelectionCell.style['padding'] = '0';
  accSelectionCell.style['border'] = '0';
  
  var isTableHeaderless = this.component._isTableHeaderless();
  
  if (!isTableHeaderless)
  {
    var accessibility = this.options['accessibility'];
    var rowHeaderColumnId = null;
    
    if (accessibility != null && accessibility['rowHeader'] != null)
    {
      rowHeaderColumnId = accessibility['rowHeader'];
    }
    else
    {
      rowHeaderColumnId = this.component._getColumnDefs()[0].id;
    }
    var cellRowHeaderId = this.getTableId() + ':' + rowHeaderColumnId + '_' + rowKeyStrHashCode;
    accSelectionCell.setAttribute(oj.TableDomUtils.DOM_ATTR._HEADERS, this.getTableId() + ':' + oj.TableDomUtils._COLUMN_HEADER_ROW_SELECT_ID + ' ' + cellRowHeaderId);
  }
  var accSelectCheckbox = document.createElement(oj.TableDomUtils.DOM_ELEMENT._INPUT);
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._ID, this.getTableId() + ':acc_sel_row_' + rowKeyStrHashCode);
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._TYPE, 'checkbox');
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX, '-1');
  var selectRowTitle = null;
  
  var editMode = this['options']['editMode'];
  var isRowSelectionEnabled = this.component._getRowSelectionMode() != null ? true : false;
        
  if (editMode == this.component._OPTION_EDIT_MODE._ROW_EDIT)
  {
    if (!isRowSelectionEnabled)
    {
      selectRowTitle = this.component.getTranslatedString(this.component._BUNDLE_KEY._LABEL_EDIT_ROW);
    }
    else
    {
      selectRowTitle = this.component.getTranslatedString(this.component._BUNDLE_KEY._LABEL_SELECT_AND_EDIT_ROW);
    }
  }
  else
  {
    selectRowTitle = this.component.getTranslatedString(this.component._BUNDLE_KEY._LABEL_SELECT_ROW);
  }
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._TITLE, selectRowTitle);
  accSelectCheckbox.classList.add(oj.TableDomUtils.CSS_CLASSES._CHECKBOX_ACC_SELECT_ROW_CLASS);
  accSelectCheckbox.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  
  if (this.component._getEditableRowIdx() == rowIdx)
  {
    accSelectCheckbox['checked'] = true;
  }
  else
  {
    accSelectCheckbox['checked'] = false;
  }
  accSelectionCell.appendChild(accSelectCheckbox); //@HTMLUpdateOK
  tableBodyRow.insertBefore(accSelectionCell, tableBodyRow.firstChild); //@HTMLUpdateOK

  return accSelectionCell;
};

/**
 * Create the message td element with appropriate styling and message
 * @param {Element} tableBodyMessageRow tr DOM element
 * @param {number} columnCount  number of visible columns
 * @param {string} message message
 * @return {Element} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBodyMessageCell = function(tableBodyMessageRow, columnCount, message)
{
  var messageCell = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TD);
  messageCell.setAttribute(oj.TableDomUtils.DOM_ATTR._COLSPAN, columnCount + 1);
  messageCell.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_MESSAGE_CLASS);
  messageCell.appendChild(document.createTextNode(message)); //@HTMLUpdateOK

  tableBodyMessageRow.appendChild(messageCell); //@HTMLUpdateOK
  return messageCell;
};

/**
 * Create the message row with appropriate styling
 * @param {number} columnCount  number of visible columns
 * @param {string} message message
 * @return {Element} row DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBodyMessageRow = function(columnCount, message)
{
  var tableBody = this.getTableBody();
  var tableBodyMessageRow = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TR);
  tableBodyMessageRow.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_MESSAGE_ROW_CLASS);
  this.createTableBodyMessageCell(tableBodyMessageRow, columnCount, message);

  tableBody.appendChild(tableBodyMessageRow); //@HTMLUpdateOK
  return tableBodyMessageRow;
};

/**
 * Create an empty tr element with appropriate styling
 * @param {number} rowIdx  row index
 * @param {Object} rowKey  row key
 * @return {Element} tr DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBodyRow = function(rowIdx, rowKey)
{
  var tableBodyRow = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TR);
  this.createTableBodyCellAccSelect(rowIdx, rowKey, null, tableBodyRow, true);
  
  return tableBodyRow;
};

/**
 * Add the touch affordance to the table row.
 * @param {number} rowIdx  row index
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBodyRowTouchSelectionAffordance = function(rowIdx)
{
    var topAffordance = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
    topAffordance.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOP_CLASS);
    topAffordance.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOUCH_AREA_CLASS);
    $(topAffordance).data('rowIdx', rowIdx);
    var topIcon = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
    topIcon.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOP_ICON_CLASS);
    topIcon.setAttribute(oj.TableDomUtils.DOM_ATTR._ROLE, 'button');
    topIcon.setAttribute(oj.TableDomUtils.DOM_ATTR._ARIA_LABEL, this.component.getTranslatedString('labelAccSelectionAffordanceTop'));
    $(topIcon).data('rowIdx', rowIdx);
    topAffordance.appendChild(topIcon); //@HTMLUpdateOK

    var bottomAffordance = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
    bottomAffordance.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_BOTTOM_CLASS);
    bottomAffordance.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOUCH_AREA_CLASS);
    $(bottomAffordance).data('rowIdx', rowIdx);
    var bottomIcon = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
    bottomIcon.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_BOTTOM_ICON_CLASS);
    bottomIcon.setAttribute(oj.TableDomUtils.DOM_ATTR._ROLE, 'button');
    bottomIcon.setAttribute(oj.TableDomUtils.DOM_ATTR._ARIA_LABEL, this.component.getTranslatedString('labelAccSelectionAffordanceBottom'));
    $(bottomIcon).data('rowIdx', rowIdx);
    bottomAffordance.appendChild(bottomIcon); //@HTMLUpdateOK
    
    var tableContainer = this.getTableContainer();
    tableContainer.appendChild(topAffordance); //@HTMLUpdateOK
    tableContainer.appendChild(bottomAffordance); //@HTMLUpdateOK
    
    this.moveTableBodyRowTouchSelectionAffordanceTop(rowIdx);
    this.moveTableBodyRowTouchSelectionAffordanceBottom(rowIdx);
};

/**
 * Create the bottom slot element with appropriate styling
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableBottomSlot = function()
{
  var tableContainer = this.getTableContainer();
  var tableBottomSlot = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  tableBottomSlot.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_BOTTOM_SLOT_CLASS);
  tableContainer.appendChild(tableBottomSlot); //@HTMLUpdateOK 
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_BOTTOM_SLOT_CLASS, tableBottomSlot);

  return tableBottomSlot;
};

/**
 * Create an empty div element with appropriate styling
 * @return {Element} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableContainer = function()
{
  var options = this.options;
  // need to enclose the table in a div to provide horizontal scrolling
  var tableContainer;
  if (this.OuterWrapper)
  {
    tableContainer = this.OuterWrapper;
  } 
  else 
  {
    tableContainer = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
    this.element.parentNode.replaceChild(tableContainer, this.element);
    tableContainer.insertBefore(this.element, tableContainer.firstChild); //@HTMLUpdateOK
  }
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_CONTAINER_CLASS, tableContainer);

  return tableContainer;
};

/**
 * Create an empty tfoot with appropriate styling
 * @return {Element} tfoot DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableFooter = function()
{
  var table = this.getTable();
  var tableFooter = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TFOOT);
  var tableFooterRow = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TR);
  this.createTableFooterAccSelect(tableFooterRow);

  tableFooter.appendChild(tableFooterRow); //@HTMLUpdateOK
  
  // check if thead is already there. If so add relative to thead.
  var tableHeader = this.getTableHeader();
  if (tableHeader != null)
  {
    tableHeader.parentNode.insertBefore(tableFooter, tableHeader.nextSibling); //@HTMLUpdateOK
  }
  else
  {
    // check if tbody is already there. If so add relative to tbody.
    var tableBody = this.getTableBody();
    if (tableBody != null)
    {
      tableBody.parentNode.insertBefore(tableFooter, tableBody); //@HTMLUpdateOK
    }
    else
    {
      table.appendChild(tableFooter); //@HTMLUpdateOK
    }
  }
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CLASS, tableFooter);
  
  return tableFooter;
};

/**
 * Create a checkbox for accessibility row selection
 * @param {Element} tableFooterRow  tr DOM element
 * @return {Element} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableFooterAccSelect = function(tableFooterRow)
{
  var accFooterCell = tableFooterRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);

  if (accFooterCell != null && accFooterCell.length > 0)
  {
    return accFooterCell[0];
  }
  accFooterCell = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TD);
  accFooterCell.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  accFooterCell.setAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX, '-1');
  tableFooterRow.insertBefore(accFooterCell, tableFooterRow.firstChild); //@HTMLUpdateOK

  return accFooterCell;
};

/**
 * Create an empty td element with appropriate styling
 * @param {number} columnIdx  column index
 * @return {Element} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableFooterCell = function(columnIdx)
{
  var tableFooterCell = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TD);

  return tableFooterCell;
};

/**
 * Create an empty thead & tr element with appropriate styling
 * @return {Element} thead DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableHeader = function()
{
  var table = this.getTable();
  var tableHeader = document.createElement(oj.TableDomUtils.DOM_ELEMENT._THEAD);
  var tableHeaderRow = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TR);
  tableHeader.appendChild(tableHeaderRow); //@HTMLUpdateOK
  
  // check if tfoot is already there. If so add relative to tfoot.
  var tableFooter = this.getTableFooter();
  if (tableFooter != null)
  {
    tableFooter.parentNode.insertBefore(tableHeader, tableFooter); //@HTMLUpdateOK
  }
  else
  {
    // check if tbody is already there. If so add relative to tbody.
    var tableBody = this.getTableBody();
    if (tableBody != null)
    {
      tableBody.parentNode.insertBefore(tableHeader, tableBody); //@HTMLUpdateOK
    }
    else
    {
      table.appendChild(tableHeader); //@HTMLUpdateOK
    }
  }
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_HEADER_ROW_CLASS, tableHeaderRow);
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_HEADER_CLASS, tableHeader);
  
  return tableHeader;
};

/**
 * Create a th element for accessibility row selection
 * @return {Element} th DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableHeaderAccSelectRowColumn = function()
{
  var headerColumn = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TH);
  headerColumn.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ACC_SELECT_ROW_CLASS);
  headerColumn.setAttribute(oj.TableDomUtils.DOM_ATTR._ID, this.getTableId() + ':' + oj.TableDomUtils._COLUMN_HEADER_ROW_SELECT_ID);
  var selectRowTitle;
  var editMode = this['options']['editMode'];
  var isRowSelectionEnabled = this.component._getRowSelectionMode() != null ? true : false;
        
  if (editMode == this.component._OPTION_EDIT_MODE._ROW_EDIT)
  {
    if (!isRowSelectionEnabled)
    {
      selectRowTitle = this.component.getTranslatedString(this.component._BUNDLE_KEY._LABEL_EDIT_ROW);
    }
    else
    {
      selectRowTitle = this.component.getTranslatedString(this.component._BUNDLE_KEY._LABEL_SELECT_AND_EDIT_ROW);
    }
  }
  else
  {
    selectRowTitle = this.component.getTranslatedString(this.component._BUNDLE_KEY._LABEL_SELECT_ROW);
  }
  headerColumn.setAttribute(oj.TableDomUtils.DOM_ATTR._TITLE, selectRowTitle);
  headerColumn.style['padding'] = '0';
  headerColumn.style['border'] = '0';
  var headerColumnText = document.createElement(oj.TableDomUtils.DOM_ELEMENT._SPAN);
  headerColumnText.appendChild(document.createTextNode(selectRowTitle)); //@HTMLUpdateOK
  headerColumnText.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  headerColumn.appendChild(headerColumnText); //@HTMLUpdateOK

  return headerColumn;
};

/**
 * Create a th element with appropriate styling and column content
 * @param {number} columnIdx  column index
 * @param {string} columnSelectionMode  column selection mode
 * @return {Element} th DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableHeaderColumn = function(columnIdx, columnSelectionMode)
{
  var column = this.component._getColumnDefs()[columnIdx];
  var headerColumnCell = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TH);
  this.styleTableHeaderColumn(columnIdx, headerColumnCell, columnSelectionMode, true);
  // add abbr for acc
  headerColumnCell.setAttribute('abbr', column.headerText);
  // add title for tooltip
  headerColumnCell.setAttribute(oj.TableDomUtils.DOM_ATTR._TITLE, column.headerText);
  this.insertTableHeaderColumn(columnIdx, headerColumnCell);
  var headerContext = {'columnIndex' : columnIdx, 'headerContext': {'component': this.component, 'parentElement': headerColumnCell}};
  if (column.resizable == oj.TableDomUtils._OPTION_ENABLED)
  {
    headerColumnCell.setAttribute('data-oj-resizable', oj.TableDomUtils._OPTION_ENABLED);
  }
  if (column.sortable == oj.TableDomUtils._OPTION_ENABLED)
  {
    headerColumnCell.setAttribute('data-oj-sortable', oj.TableDomUtils._OPTION_ENABLED);
    oj.TableRendererUtils.columnHeaderSortableIconRenderer(this.component, headerContext, null, null);
  }
  else
  {
    oj.TableRendererUtils.columnHeaderDefaultRenderer(this.component, headerContext, null, null);
  }
  
  return headerColumnCell;
};

/**
 * Create the drag image for the column
 * @param {number} columnIdx  column index
 * @return {Element} DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableHeaderColumnDragImage = function(columnIdx)
{
  var headerColumn = this.getTableHeaderColumn(columnIdx);
  var headerColumnDragImage = headerColumn.cloneNode(true);
  headerColumnDragImage.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._DRAG);
  headerColumnDragImage.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._FOCUS);
  headerColumnDragImage.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._FOCUS_HIGHLIGHT);
  headerColumnDragImage.classList.remove(oj.TableDomUtils.MARKER_STYLE_CLASSES._HOVER);
  headerColumnDragImage.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_IMAGE);
  headerColumnDragImage.style[oj.TableDomUtils.CSS_PROP._POSITION] = oj.TableDomUtils.CSS_VAL._ABSOLUTE;
  headerColumnDragImage.style[oj.TableDomUtils.CSS_PROP._TOP] = '0';
  headerColumnDragImage.style[oj.TableDomUtils.CSS_PROP._LEFT] = '-999em';
  headerColumnDragImage.style[oj.TableDomUtils.CSS_PROP._ZINDEX] = '-999';

  // The drag image element must be either visible or a child of document.body 
  // in order to show on Safari.
  document.body.appendChild(headerColumnDragImage); //@HTMLUpdateOK
  
  return headerColumnDragImage;
};

/**
 * Create a checkbox for accessibility column selection
 * @param {number} columnIdx  column index
 * @param {string} columnSelectionMode  column selection mode
 * @return {Element} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableHeaderColumnAccSelect = function(columnIdx, columnSelectionMode)
{
  if (columnSelectionMode != oj.TableDomUtils._OPTION_SELECTION_MODES._SINGLE &&
    columnSelectionMode != oj.TableDomUtils._OPTION_SELECTION_MODES._MULTIPLE)
  {
    return null;
  }
  var headerColumn = this.getTableHeaderColumn(columnIdx);
  var accSelectionHeaderColumn = this.getTableHeaderColumnAccSelect(columnIdx);

  if (accSelectionHeaderColumn != null)
  {
    return accSelectionHeaderColumn;
  }

  accSelectionHeaderColumn = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  accSelectionHeaderColumn.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ACC_SELECT_COLUMN_CLASS);
  accSelectionHeaderColumn.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  var accSelectCheckbox = document.createElement(oj.TableDomUtils.DOM_ELEMENT._INPUT);
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._ID, this.getTableId() + ':acc_sel_col' + columnIdx);
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._TYPE, 'checkbox');
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX, '-1');
  var selectColumnTitle = this.component.getTranslatedString(this.component._BUNDLE_KEY._LABEL_SELECT_COLUMN);
  accSelectCheckbox.setAttribute(oj.TableDomUtils.DOM_ATTR._TITLE, selectColumnTitle);
  accSelectCheckbox.classList.add(oj.TableDomUtils.CSS_CLASSES._CHECKBOX_ACC_SELECT_COLUMN_CLASS);
  accSelectionHeaderColumn.appendChild(accSelectCheckbox); //@HTMLUpdateOK
  headerColumn.insertBefore(accSelectionHeaderColumn, headerColumn.firstChild); //@HTMLUpdateOK

  return accSelectionHeaderColumn;
};

/**
 * Create a div element for resize indicator
 * @return {Element} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableHeaderColumnResizeIndicator = function()
{
  var tableHeaderColumnResizeIndicator = this.getTableHeaderColumnResizeIndicator();
  
  if (!tableHeaderColumnResizeIndicator)
  {
    var tableContainer = this.getTableContainer();
    tableHeaderColumnResizeIndicator = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
    tableContainer.appendChild(tableHeaderColumnResizeIndicator);
    tableHeaderColumnResizeIndicator.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_RESIZE_INDICATOR_CLASS);
    this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_RESIZE_INDICATOR_CLASS, tableHeaderColumnResizeIndicator);
  }
  
  return tableHeaderColumnResizeIndicator;
};

/**
 * Create a div element for inline messages
 * @return {Element} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableInlineMessage = function()
{
  var tableContainer = this.getTableContainer();
  var inlineMessage = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  inlineMessage.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_INLINE_MESSAGE_CLASS);
  inlineMessage.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._NONE;
  tableContainer.appendChild(inlineMessage); //@HTMLUpdateOK
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_INLINE_MESSAGE_CLASS, inlineMessage);
  
  return inlineMessage;
};

/**
 * Create a div element for the accessibility notifications
 * @return {Element} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableStatusAccNotification = function()
{
  var tableContainer = this.getTableContainer();
  var statusNotification = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  statusNotification.setAttribute(oj.TableDomUtils.DOM_ATTR._ROLE, 'status');
  statusNotification.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_STATUS_ACC_NOTIFICATION_CLASS);
  statusNotification.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  tableContainer.appendChild(statusNotification); //@HTMLUpdateOK
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_STATUS_ACC_NOTIFICATION_CLASS, statusNotification);

  return statusNotification;
};

/**
 * Create a div element for the Fetching Data... status message
 * @return {Element} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.createTableStatusMessage = function()
{
  var tableContainer = this.getTableContainer();
  var statusMessage = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  statusMessage.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_STATUS_MESSAGE_CLASS);
  statusMessage.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._NONE;

  var statusMessageLoadingIndicator = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  statusMessageLoadingIndicator.classList.add(oj.TableDomUtils.CSS_CLASSES._ICON_CLASS);
  statusMessageLoadingIndicator.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_LOADING_ICON_CLASS);
  statusMessageLoadingIndicator.setAttribute("aria-label", this.component.getTranslatedString(this.component._BUNDLE_KEY._MSG_FETCHING_DATA));
  statusMessage.appendChild(statusMessageLoadingIndicator); //@HTMLUpdateOK
  tableContainer.appendChild(statusMessage); //@HTMLUpdateOK
  this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_STATUS_MESSAGE_CLASS, statusMessage);

  return statusMessage;
};

/**
 * Display the visual indicator for column drag over
 * @param {number} columnIdx  column index
 * @param {boolean} before before the column
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.displayDragOverIndicatorColumn = function(columnIdx, before)
{
  this.removeDragOverIndicatorColumn();
  var tableHeaderRow = this.getTableHeaderRow();
  var tableHeaderColumn = this.getTableHeaderColumn(columnIdx);

  if (tableHeaderColumn != null)
  {
    if (before)
    {
      tableHeaderColumn.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_BEFORE_CLASS);
    }
    else
    {
      tableHeaderColumn.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_AFTER_CLASS);
    }
  }
  else
  {
    var columns = this.component._getColumnDefs();
      
    if (columns == null || columns.length == 0)
    {
      if (before)
      {
        tableHeaderRow.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_BEFORE_CLASS);
      }
      else
      {
        tableHeaderRow.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_AFTER_CLASS); 
      }
    }
  } 
}

/**
 * Display the visual indicator for row drag over
 * @param {number} rowIdx  row index
 * @param {boolean} before before the row
 * @param {Element} modelRow tr DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.displayDragOverIndicatorRow = function(rowIdx, before, modelRow)
{
  this.removeDragOverIndicatorRow();
  var tableBodyRowDragIndicator = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TR);
  if (before)
  {
    tableBodyRowDragIndicator.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_DRAG_INDICATOR_BEFORE_CLASS);
  }
  else
  {
    tableBodyRowDragIndicator.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_DRAG_INDICATOR_AFTER_CLASS);
  }
  if (modelRow)
  {
    $(tableBodyRowDragIndicator).height(parseInt($(modelRow).height(), 10));
  }
    
  var tableBodyDragIndicatorCell = document.createElement(oj.TableDomUtils.DOM_ELEMENT._TD);
  var columns = this.component._getColumnDefs();
  tableBodyDragIndicatorCell.setAttribute(oj.TableDomUtils.DOM_ATTR._COLSPAN, columns.length + 1);
  tableBodyRowDragIndicator.appendChild(tableBodyDragIndicatorCell); //@HTMLUpdateOK
  var tableBodyRow = this.getTableBodyRow(rowIdx);

  if (tableBodyRow != null)
  {
    if (before)
    {
      tableBodyRow.parentNode.insertBefore(tableBodyRowDragIndicator, tableBodyRow); //@HTMLUpdateOK
    }
    else
    {
      tableBodyRow.parentNode.insertBefore(tableBodyRowDragIndicator, tableBodyRow.nextSibling); //@HTMLUpdateOK
    }
  }
  else
  {
    var tableBodyRows = this.getTableBodyRows();
    
    if (tableBodyRows == null || tableBodyRows.length == 0)
    {
      this.component._hideNoDataMessage();
    }
    this.getTableBody().appendChild(tableBodyRowDragIndicator); //@HTMLUpdateOK
  }
}

/**
 * Get the context menu
 * @return  {Element} menu DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getContextMenu = function()
{
  return this._menuContainer;
};

oj.TableDomUtils.prototype.getContextMenuResizePopup = function()
{
  var popupId = this.getTableId() + '_resize_popup';
  if (!this._isCachedDomElement(popupId))
  {
    var popup = document.getElementById(popupId);
    this._cacheDomElement(popupId, popup);
  }
  
  return this._getCachedDomElement(popupId);
};

/**
 * Get the column index of the DOM element. e.g. pass in the table cell to
 * see which column it's in.
 * @param {Element} element  DOM element
 * @return {number|null} the column index
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getElementColumnIdx = function(element)
{
  var tableBodyCell = this.getFirstAncestor(element, oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_CLASS);
  if (tableBodyCell != null)
  {
    return $(tableBodyCell.parentNode).children('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_CLASS).index(tableBodyCell);
  }
  
  var tableHeaderColumn = this.getFirstAncestor(element, oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS);
  if (tableHeaderColumn != null)
  {
    return $(tableHeaderColumn.parentNode).children('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS).index(tableHeaderColumn);
  }
  
  var tableFooterCell = this.getFirstAncestor(element, oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CELL_CLASS);
  if (tableFooterCell != null)
  {
    return $(tableFooterCell.parentNode).children('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CELL_CLASS).index(tableFooterCell);
  }
  return null;
};

/**
 * Get the row index of the DOM element. e.g. pass in the table cell to
 * see which row it's in.
 * @param {Element} element  DOM element
 * @return {number|null} the row index
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getElementRowIdx = function(element)
{
  var tableBodyRow = this.getFirstAncestor(element, oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS);
  
  if (tableBodyRow != null)
  {
    var tableBodyRows = this.getTableBodyRows();
    return $(tableBodyRows).index(tableBodyRow);
  }
  return null;
};

/**
  * Find the first ancestor of an element with a specific class name
  * @param {Element} element the element to find the nearest class name to
  * @param {string} className the class name to look for
  * @return {Element|null} the element with the className, if there is none returns null 
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype.getFirstAncestor = function(element, className) 
{
  var parents;
  
  if (element == null)
  {
    return null;
  }

  if (element.classList.contains(className))
  {
    return element;
  }
  parents = $(element).parents('.' + className);
  if (parents.length != 0)
  {
    return parents[0];
  }
  return null;
};

/**
 * Return the scrollbar height
 * @return {number} scrolbar height
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getScrollbarHeight = function()
{
  var scroller = this.getScroller();
  if (scroller.clientHeight > 0)
  {
    var scrollbarHeight = scroller.offsetHeight - scroller.clientHeight;

    return scrollbarHeight;
  }
  return 0;
};

/**
 * Return the scrollbar width
 * @return {number} scrolbar width
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getScrollbarWidth = function()
{
  var scroller = this.getScroller();
  if (scroller.clientWidth > 0)
  {
    var scrollbarWidth = scroller.offsetWidth - scroller.clientWidth;

    return scrollbarWidth;
  }
  return 0;
};

/**
 * Return the table scroller
 * @return {Element} scrolbar
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getScroller = function()
{
    return this.getTableBody();
};

/**
 * Get the element scrollLeft
 * @param {Element} element DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getScrollLeft = function(element)
{
  var scrollLeft = element.scrollLeft;
  
  if (this.component._GetReadingDirection() === "rtl")
  {
    scrollLeft = Math.abs(scrollLeft);
    
    if (this._isWebkit())
    {
      var maxScrollLeft = element.scrollWidth - element.clientWidth;
      scrollLeft = maxScrollLeft - scrollLeft;
    }
  }
  return scrollLeft;
};
        
/**
 * Return the table element
 * @return {Element} table DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTable = function()
{
  return this.element;
};

/**
 * Return the table body element
 * @return {Element|null} tbody DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBody = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_CLASS, true);
};

/**
 * Return the cell element
 * @param {number} rowIdx  row index
 * @param {number} columnIdx  column index
 * @param {Element|null} tableBodyRow  tr DOM element
 * @return {Element|null} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyCell = function(rowIdx, columnIdx, tableBodyRow)
{
  var tableBodyCells = this.getTableBodyCells(rowIdx, tableBodyRow);
  if (!tableBodyCells)
  {
    return null;
  }

  if (tableBodyCells.length > columnIdx)
  {
    return tableBodyCells[columnIdx];
  }

  return null;
};

/**
 * Get checkbox cell for accessibility row selection
 * @param {Element} tableBodyRow  tr DOM element
 * @return {Element|null} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyCellAccSelect = function(tableBodyRow)
{
  if (tableBodyRow != null)
  {
    var accSelectionCell = tableBodyRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_ACC_SELECT_CLASS);

    if (accSelectionCell != null && accSelectionCell.length > 0)
    {
      return accSelectionCell[0];
    }
  }
  return null;
};

/**
 * Return all the logical cell elements in a row
 * @param {number} rowIdx  row index
 * @param {Element|null} tableBodyRow  tr DOM element
 * @return {Array|null} array of td DOM elements
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyLogicalCells = function(rowIdx, tableBodyRow)
{
  var tableBodyCells = this.getTableBodyCells(rowIdx, tableBodyRow);
  if (!tableBodyCells)
  {
    return null;
  }

  return this._getColspanLogicalElements(tableBodyCells);
}

/**
 * Return all the cell elements in a row
 * @param {number|null} rowIdx  row index
 * @param {Element|null} tableBodyRow  tr DOM element
 * @return {Array|null} array of td DOM elements
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyCells = function(rowIdx, tableBodyRow)
{
  if (!tableBodyRow)
  {
    tableBodyRow = this.getTableBodyRow(rowIdx);
  }

  if (!tableBodyRow)
  {
    return null;
  }

  var tableBodyCellElements = tableBodyRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_CLASS);

  if (tableBodyCellElements != null && tableBodyCellElements.length > 0)
  {
    return tableBodyCellElements;
  }

  return null;
};

/**
 * Return the table body message cell element
 * @return {Element|null} tr DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyMessageCell = function()
{
  var tableBody = this.getTableBody();
  if (tableBody)
  {
    var messageCell = tableBody.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_MESSAGE_CLASS);
    if (messageCell && messageCell.length > 0)
    {
      return messageCell[0];
    }
  }
  return null;
};

/**
 * Return the table body message row element
 * @return {Element|null} tr DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyMessageRow = function()
{
  var tableBody = this.getTableBody();
  if (tableBody)
  {
    var messageRow = tableBody.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_MESSAGE_ROW_CLASS);
    if (messageRow && messageRow.length > 0)
    {
      return messageRow[0];
    }
  }
  return null;
};

/**
 * Return table row
 * @param {number|null} rowIdx  row index
 * @return {Element|null} tr DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyRow = function(rowIdx)
{
  var tableBodyRows = this.getTableBodyRows();

  if (!tableBodyRows)
  {
    return null;
  }
  
  if (rowIdx == null)
  {
    return null;
  }

  if (tableBodyRows.length > rowIdx)
  {
    return tableBodyRows[rowIdx];
  }
  
  return null;
};

/**
 * Return all the table rows
 * @return {Array|null} array of tr DOM elements
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyRows = function()
{
  if (!this._cachedDomTableBodyRows)
  {
    var tableBody = this.getTableBody();
    
    if (tableBody != null)
    {
      var tableBodyRowElements = tableBody.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS);

      if (tableBodyRowElements != null && tableBodyRowElements.length > 0)
      {
        this._cachedDomTableBodyRows = tableBodyRowElements;
      }
    }
  }

  return this._cachedDomTableBodyRows;
};

/**
 * Get top touch affordance to the table row.
 * @return {Element|null} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyRowTouchSelectionAffordanceTop = function()
{
  var tableContainer = this.getTableContainer();
  var topAffordance = tableContainer.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOP_CLASS);
  
  if (topAffordance != null && topAffordance.length > 0)
  {
    topAffordance = topAffordance[0];
    
    return topAffordance;
  }
  return null;
};

/**
 * Get bottom touch affordance to the table row.
 * @return {Element|null} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBodyRowTouchSelectionAffordanceBottom = function()
{
  var tableContainer = this.getTableContainer();
  var bottomAffordance = tableContainer.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_BOTTOM_CLASS);
  
  if (bottomAffordance != null && bottomAffordance.length > 0)
  {
    bottomAffordance = bottomAffordance[0];
    
    return bottomAffordance;
  }
  return null;
};

/**
 * Return the bottom slot element
 * @return {Element|null} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableBottomSlot = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_BOTTOM_SLOT_CLASS, true);
};

/**
 * Return the table container
 * @return {Element|null} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableContainer = function()
{
  if (!this._isCachedDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_CONTAINER_CLASS))
  {
    this._cacheDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_CONTAINER_CLASS, this.element.parentNode);
  }

  return this._getCachedDomElement(oj.TableDomUtils.CSS_CLASSES._TABLE_CONTAINER_CLASS);
};

/**
 * Return the table footer
 * @return {Element|null} tfoot DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableFooter = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CLASS, true);
};

/**
 * Return the footer cell element
 * @param {number} columnIdx  column index
 * @return {Element|null} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableFooterCell = function(columnIdx)
{
  var tableFooterCells = this.getTableFooterCells();

  if (tableFooterCells != null && tableFooterCells.length > columnIdx)
  {
    return tableFooterCells[columnIdx]; 
  }

  return null;
};

/**
 * Return all footer cells
 * @return {Array|null} array of td DOM elements
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableFooterCells = function()
{
  var tableFooterRow = this.getTableFooterRow();
  var tableFooterCells = tableFooterRow != null ? tableFooterRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CELL_CLASS) : null;

  if (tableFooterCells != null && tableFooterCells.length > 0)
  {
    return tableFooterCells;
  }

  return null;
};

/**
 * Return all logical footer cells
 * @return {Array|null} array of td DOM elements
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableFooterLogicalCells = function()
{
  var tableFooterCells = this.getTableFooterCells();
  if (!tableFooterCells)
  {
    return null;
  }

  return this._getColspanLogicalElements(tableFooterCells);
}

/**
 * Return table footer row
 * @return {Element|null} tr DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableFooterRow = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_ROW_CLASS);
};

/**
 * Return the table header
 * @return {Element|null} thead DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableHeader = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_HEADER_CLASS, true);
};

/**
 * Return table column header
 * @param {number} columnIdx  column index
 * @return {Element|null} th DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableHeaderColumn = function(columnIdx)
{
  var headerColumns = this.getTableHeaderColumns();

  if (!headerColumns)
  {
    return null;
  }

  if (headerColumns.length > columnIdx && columnIdx >= 0)
  {
    return headerColumns[columnIdx];
  }

  return null;
};

/**
 * Get checkbox cell for accessibility column selection
 * @param {number} columnIdx  column index
 * @return {Element|null} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableHeaderColumnAccSelect = function(columnIdx)
{
  var headerColumn = this.getTableHeaderColumn(columnIdx);

  if (headerColumn != null)
  {
    var accSelectionCell = headerColumn.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ACC_SELECT_COLUMN_CLASS);

    if (accSelectionCell != null && accSelectionCell.length > 0)
    {
      return accSelectionCell[0];
    }
  }
  return null;
};

/**
 * Return resize indicator
 * @return {Element} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableHeaderColumnResizeIndicator = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_RESIZE_INDICATOR_CLASS, true);
};

/**
 * Return all table column headers
 * @return {Array|null} array of th DOM elements
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableHeaderColumns = function()
{
  var tableHeaderRow = this.getTableHeaderRow();

  if (tableHeaderRow != null)
  {
    var headerColumnElements = tableHeaderRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS);

    if (headerColumnElements != null && headerColumnElements.length > 0)
    {
      return headerColumnElements;
    }
  }

  return null;
};

/**
 * Return all logical table column headers
 * @return {Array|null} array of th DOM elements
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableHeaderLogicalColumns = function()
{
  var tableHeaderColumns = this.getTableHeaderColumns();
  if (!tableHeaderColumns)
  {
    return null;
  }

  return this._getColspanLogicalElements(tableHeaderColumns);
}

/**
 * Return table header row
 * @return {Element|null} th DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableHeaderRow = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_HEADER_ROW_CLASS);
};

/**
 * Return the table DOM element id.
 * @return {string} Id for table
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableId = function()
{
  if (!this._tableId)
  {
    // Id for custom element is on the container
    var idElem = this.component._IsCustomElement() ? this.getTableContainer() : this.getTable();
    this._tableId = idElem.getAttribute(oj.TableDomUtils.DOM_ATTR._ID);
  }
  return this._tableId;
};

/**
 * Return the table inline message element
 * @return {Element|null} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableInlineMessage = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_INLINE_MESSAGE_CLASS, true);
};

/**
 * Return the table status accessibility notification
 * @return {Element|null} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableStatusAccNotification = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_STATUS_ACC_NOTIFICATION_CLASS, true);
};

/**
 * Return the table status message element
 * @return {Element|null} div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableStatusMessage = function()
{
  return this._getTableElementByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_STATUS_MESSAGE_CLASS, true);
};

/**
 * Return the unique identifier for the table. If the DOM element has an id then
 * return that. If not, generate a random UID.
 * @return {string} UID for table
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.getTableUID = function()
{
  if (!this._tableUID)
  {
    var uid = this.getTableId();
    
    if (uid == null)
    {
      uid = (Math.random()*1e32).toString(36); // used to generate a unique id. Should not be a security issue.
    }
    
    this._tableUID = uid;
  }
  return this._tableUID;
};


/**
 * Get a hash code for a string
 * @param {string} str String
 * @return {number} hashCode
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.hashCode = function(str)
{
  if ($.type(str) != 'string')
  {
    str = str.toString();
  }
  // Same hash algorithm as Java's String.hashCode
  var hash = 0;
  if (str.length == 0)
  {
    return hash;
  }
  
  var charVal, i, strCount = str.length;
  for (i = 0; i < strCount; i++) 
  {
    charVal = str.charCodeAt(i);
    hash = ((hash << 5) - hash) + charVal;
    hash = hash & hash;
  }
  return hash;
};

/**
 * Insert a td element in the appropriate place in the DOM
 * @param {number} rowIdx  row index
 * @param {Object} rowKey  row key
 * @param {Object} rowHashCode  row hash code
 * @param {number} columnIdx  column index
 * @param {Element} tableBodyCell  DOM element
 * @param {Element} tableBodyRow  tr DOM element
 * @param {boolean} isNew is new row
 * @return {Element|null} td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.insertTableBodyCell = function(rowIdx, rowKey, rowHashCode, columnIdx, tableBodyCell, tableBodyRow, isNew)
{
  this.setTableBodyCellAttributes(rowIdx, rowKey, rowHashCode, columnIdx, tableBodyCell);
  
  if (isNew)
  {
    // if it's a new row then the cells are appended in order
    // so don't bother trying to find the position to insert to
    tableBodyRow.appendChild(tableBodyCell); //@HTMLUpdateOK
    return tableBodyCell;
  }

  if (columnIdx == 0)
  {
    // insert right after the acc cell
    var accSelectionCell = tableBodyRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_ACC_SELECT_CLASS);

    if (accSelectionCell != null && accSelectionCell.length > 0)
    {
      accSelectionCell = accSelectionCell[0];
      accSelectionCell.parentNode.insertBefore(tableBodyCell, accSelectionCell.nextSibling); //@HTMLUpdateOK
    }
    else
    {
      // just prepend it
      tableBodyRow.insertBefore(tableBodyCell, tableBodyRow.firstChild); //@HTMLUpdateOK
    }
  }
  else
  {
    var tableBodyCells = tableBodyRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_CLASS);

    if (tableBodyCells.length >= columnIdx)
    {
      var previousCell = tableBodyCells[columnIdx - 1];
      previousCell.parentNode.insertBefore(tableBodyCell, previousCell.nextSibling); //@HTMLUpdateOK
    }
    else
    {
      tableBodyRow.appendChild(tableBodyCell); //@HTMLUpdateOK
    }
  }
    

  return tableBodyCell;
};

/**
 * Insert a tr element in the appropriate place in the DOM
 * @param {number} rowIdx  row index
 * @param {Element} tableBodyRow  DOM element
 * @param {Object} row  row and key object
 * @param {Element} docFrag  document fragment
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.insertTableBodyRow = function(rowIdx, tableBodyRow, row, docFrag)
{
  this.setTableBodyRowAttributes(row, tableBodyRow);

  if (docFrag == null)
  {
    var tableBody = this.getTableBody();
    var tableBodyRows = tableBody.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS);
    
    if (rowIdx == 0)
    {
      tableBody.insertBefore(tableBodyRow, tableBody.firstChild); //@HTMLUpdateOK
    }
    else if (tableBodyRows.length >= rowIdx)
    {
      var previousRow = tableBodyRows[rowIdx - 1];
      previousRow.parentNode.insertBefore(tableBodyRow, previousRow.nextSibling); //@HTMLUpdateOK
    }
    else
    {
      tableBody.appendChild(tableBodyRow); //@HTMLUpdateOK
    }
  }
  else
  {
    docFrag.appendChild(tableBodyRow); //@HTMLUpdateOK
  }
  this.clearCachedDomRowData();
};

/**
 * Insert a td element in the appropriate place in the DOM
 * @param {number} columnIdx  column index
 * @param {Element} tableFooterCell  DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.insertTableFooterCell = function(columnIdx, tableFooterCell)
{
  var tableFooterRow = this.getTableFooterRow();
  var tableFooterCells = tableFooterRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CELL_CLASS);

  if (columnIdx == 0)
  {
    // insert right after the acc cell
    var accFooterCell = tableFooterRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);

    if (accFooterCell != null && accFooterCell.length > 0)
    {
      accFooterCell = accFooterCell[0];
      accFooterCell.parentNode.insertBefore(tableFooterCell, accFooterCell.nextSibling); //@HTMLUpdateOK
    }
    else
    {
      // just prepend it
      tableFooterRow.insertBefore(tableFooterCell, tableFooterRow.firstChild); //@HTMLUpdateOK
    }
  }
  else if (tableFooterRow.length >= columnIdx)
  {
    var previousCell = tableFooterCells[columnIdx - 1];
    previousCell.parentNode.insertBefore(tableFooterCell, previousCell.nextSibling); //@HTMLUpdateOK
  }
  else
  {
    tableFooterRow.appendChild(tableFooterCell); //@HTMLUpdateOK
  }

  return tableFooterCell;
};

/**
 * Insert a th element in the appropriate place in the DOM
 * @param {number} columnIdx  column index
 * @param {Element} tableHeaderColumn  DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.insertTableHeaderColumn = function(columnIdx, tableHeaderColumn)
{
  var tableHeaderRow = this.getTableHeaderRow();
  var tableHeaderColumns = this.getTableHeaderColumns();
  // save the column index on the element
  this.setTableHeaderColumnAttributes(columnIdx, tableHeaderColumn);

  // if there is an existing th at the index then replace it
  var oldTableHeaderColumn = this.getTableHeaderColumn(columnIdx);
  if (oldTableHeaderColumn)
    $(oldTableHeaderColumn).replaceWith($(tableHeaderColumn)); //@HTMLUpdateOK
  else
  {
    if (columnIdx == 0)
    {
      // insert right after the acc column
      var accSelectionColumn = tableHeaderRow.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ACC_SELECT_ROW_CLASS);

      if (accSelectionColumn != null && accSelectionColumn.length > 0)
      {
        accSelectionColumn = accSelectionColumn[0];
        accSelectionColumn.parentNode.insertBefore(tableHeaderColumn, accSelectionColumn.nextSibling); //@HTMLUpdateOK
      }
      else
      {
        // just prepend it
        tableHeaderRow.insertBefore(tableHeaderColumn, tableHeaderRow.firstChild); //@HTMLUpdateOK
      }
    }
    else if (tableHeaderColumns && tableHeaderColumns.length >= columnIdx)
    {
      var previousColumn = tableHeaderColumns[columnIdx - 1];
      previousColumn.parentNode.insertBefore(tableHeaderColumn, previousColumn.nextSibling); //@HTMLUpdateOK
    }
    else
    {
      tableHeaderRow.appendChild(tableHeaderColumn); //@HTMLUpdateOK
    }
  }
};

/**
  * Returns true if scrollHeight > clientHeight for height and width.
  * @return {Array} First element is height boolean, followed by width boolean.
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype.isTableContainerScrollable = function()
{
  var tableContainer = this.getTableContainer();
  var table = this.getTable();
  var isTableStatusMessageShown = this.component._isStatusMessageShown();
  if (isTableStatusMessageShown)
  {
    this.component._hideStatusMessage();
  }
  
  var result = [];
  if (tableContainer.clientHeight > 0)
  {
    // don't use tableContainer[0].scrollHeight as it may not be accurate
    if (table.clientHeight > tableContainer.clientHeight)
    {
      // overflow
      result[0] = 1;
    }
    else if (tableContainer.clientHeight - table.clientHeight > 1)
    {
      // underflow
      result[0] = -1;
    }
    else
    {
      result[0] = 0;
    }
  }
  else
  {
    result[0] = 0;
  }
  
  if (tableContainer.clientWidth > 0)
  {
      if (table.clientWidth > tableContainer.clientWidth)
      {
        // overflow
        result[1] = 1;
      }
      else if (tableContainer.clientWidth - table.clientWidth > 1)
      {
        // underflow
        result[1] = -1;
      }
      else
      {
        result[1] = 0;
      }
  }
  else
  {
    result[1] = 0;
  }
  
  if (isTableStatusMessageShown)
  {
    this.component._showStatusMessage();
  }
  return result;
}

/**
 * Move the top touch affordance to the table row.
 * @param {number} rowIdx  row index
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.moveTableBodyRowTouchSelectionAffordanceTop = function(rowIdx)
{
  var topAffordance = this.getTableBodyRowTouchSelectionAffordanceTop();
  
  if (topAffordance != null)
  {
    $(topAffordance).data('rowIdx', rowIdx);
    $(topAffordance.children[0]).data('rowIdx', rowIdx);
    var tableBody = this.getTableBody();
    var tableBodyRow = this.getTableBodyRow(rowIdx);
    var tableBodyRowRect = tableBodyRow.getBoundingClientRect();
    var tableContainer = this.getTableContainer();
    var tableContainerRect = tableContainer.getBoundingClientRect();
    topAffordance.style[oj.TableDomUtils.CSS_PROP._TOP] = tableBodyRowRect.top - tableContainerRect.top - topAffordance.clientHeight / 2 + oj.TableDomUtils.CSS_VAL._PX;
    topAffordance.style[oj.TableDomUtils.CSS_PROP._LEFT] = tableBody.clientWidth / 2 + oj.TableDomUtils.CSS_VAL._PX;
  }
};

/**
 * Move the bottom touch affordance to the table row.
 * @param {number} rowIdx  row index
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.moveTableBodyRowTouchSelectionAffordanceBottom = function(rowIdx)
{
  var bottomAffordance = this.getTableBodyRowTouchSelectionAffordanceBottom();
  
  if (bottomAffordance != null)
  {
    $(bottomAffordance).data('rowIdx', rowIdx);
    $(bottomAffordance.children[0]).data('rowIdx', rowIdx);
    var tableBody = this.getTableBody();
    var tableBodyRow = this.getTableBodyRow(rowIdx);
    var tableBodyRowRect = tableBodyRow.getBoundingClientRect();
    var tableContainer = this.getTableContainer();
    var tableContainerRect = tableContainer.getBoundingClientRect();
    bottomAffordance.style[oj.TableDomUtils.CSS_PROP._TOP] = tableBodyRowRect.top - tableContainerRect.top + tableBodyRowRect.height - bottomAffordance.clientHeight / 2 + oj.TableDomUtils.CSS_VAL._PX;
    bottomAffordance.style[oj.TableDomUtils.CSS_PROP._LEFT] = tableBody.clientWidth / 2 + oj.TableDomUtils.CSS_VAL._PX;
  }
};

/**
 * Move the column to the destination index. If there is already a column at destIdx,
 * then insert before it.
 * @param {number} columnIdx  column index
 * @param {number} destIdx column index
 * @param {Object} event
 * @return {Array} Array of moved columns map
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.moveTableHeaderColumn = function(columnIdx, destIdx, event)
{
  var columns = this.component._getColumnDefs();
  
  var tableHeaderColumn = this.getTableHeaderColumn(columnIdx);
  var tableFooterCell = this.getTableFooterCell(columnIdx);
  var tableBodyCell = null;
  var destTableHeaderColumn = null;
  var destTableFooterCell = null;
  var destTableBodyCell = null;
  var colSpan = null;
  
  var afterColumn = false;
  
  if (destIdx == columns.length)
  {
    destIdx = destIdx - 1;
    afterColumn = true;
  }

  if (tableHeaderColumn != null)
  {
    colSpan = tableHeaderColumn.getAttribute(oj.TableDomUtils.DOM_ATTR._COLSPAN);
    destTableHeaderColumn = this.getTableHeaderColumn(destIdx);
    
    if (destTableHeaderColumn != null && 
        (colSpan == null || colSpan == 1))
    {
      if (afterColumn)
      {
        destTableHeaderColumn.parentNode.insertBefore(tableHeaderColumn, destTableHeaderColumn.nextSibling); //@HTMLUpdateOK
      }
      else
      {
        destTableHeaderColumn.parentNode.insertBefore(tableHeaderColumn, destTableHeaderColumn); //@HTMLUpdateOK
      }
    }
  }
  
  if (tableFooterCell != null)
  {
    colSpan = tableFooterCell.getAttribute(oj.TableDomUtils.DOM_ATTR._COLSPAN);
    destTableFooterCell = this.getTableFooterCell(destIdx);
    
    if (destTableFooterCell != null && 
        (colSpan == null || colSpan == 1))
    {
      if (afterColumn)
      {
        destTableFooterCell.parentNode.insertBefore(tableFooterCell, destTableFooterCell.nextSibling); //@HTMLUpdateOK
      }
      else
      {
        destTableFooterCell.parentNode.insertBefore(tableFooterCell, destTableFooterCell); //@HTMLUpdateOK
      }
    }
  }

  var tableBodyRows = this.getTableBodyRows();
  var i;
  
  if (tableBodyRows != null)
  {
    for (i = 0; i < tableBodyRows.length; i++)
    {
      tableBodyCell = this.getTableBodyCell(i, columnIdx, null);
      
      if (tableBodyCell != null)
      {
        destTableBodyCell = this.getTableBodyCell(i, destIdx, null);
        colSpan = tableBodyCell.getAttribute(oj.TableDomUtils.DOM_ATTR._COLSPAN);

        if (destTableBodyCell != null && 
            (colSpan == null || colSpan == 1))
        {
          if (afterColumn)
          {
            destTableBodyCell.parentNode.insertBefore(tableBodyCell, destTableBodyCell.nextSibling); //@HTMLUpdateOK
          }
          else
          {
            destTableBodyCell.parentNode.insertBefore(tableBodyCell, destTableBodyCell); //@HTMLUpdateOK
          }
        }
      }
    }
  }
  
  // update options
  var clonedColumnsOption = [];
  var columnsCount = this.options['columns'].length;
  for (i = 0; i < columnsCount; i++)
  {
    clonedColumnsOption[i] = $.extend({}, {}, this.options['columns'][i]);
  }
  var destColIdx = columnIdx < destIdx && !afterColumn ? destIdx - 1 : destIdx;
  var columnOption = clonedColumnsOption.splice(columnIdx, 1);
  clonedColumnsOption.splice(destColIdx, 0, columnOption[0]);
  // re-order the column definition metadata
  var columnDef = this.component._columnDefArray.splice(columnIdx, 1);
  this.component._columnDefArray.splice(destColIdx, 0, columnDef[0]);
  if (!this._columnsDestMap)
  {
    this._columnsDestMap = [];
    for (i = 0; i < clonedColumnsOption.length; i++)
    {
      this._columnsDestMap[i] = i;
    }
  }
  var columnsDestMapItem = this._columnsDestMap.splice(columnIdx, 1);
  this._columnsDestMap.splice(destColIdx, 0, columnsDestMapItem[0]);
  
  // clone the array so we can trigger that it's changed
  this.component.option('columns', clonedColumnsOption, {'_context': {writeback: true, originalEvent: event, internalSet: true}});
  this.component._queueTask(function(){});

  return this._columnsDestMap;
}

/**
 * Refresh any translated strings in the context menu.
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.refreshContextMenu = function()
{
  var self = this;
  var menuContainer = this._menuContainer;

  if (menuContainer)
  {
    var listItems = menuContainer.querySelectorAll('[data-oj-command]');
    var i;
    for (i = 0; i < listItems.length; i++)
    {
      var contextMenuLabel = $(listItems[i]).children(oj.TableDomUtils.DOM_ELEMENT._A);
      if (contextMenuLabel.length > 0)
      {
        var command = listItems[i].getAttribute('data-oj-command').split("-");
        command = command[command.length - 1];

        var commandString;
        if (command == 'sort')
        {
          commandString = self.component.getTranslatedString('labelSort');
        }
        else if (command == 'sortAsc')
        {
          commandString = self.component.getTranslatedString('labelSortAsc');
        }
        else if (command == 'sortDsc')
        {
          commandString = self.component.getTranslatedString('labelSortDsc');
        }
        else if (command == 'resize')
        {
          commandString = self.component.getTranslatedString('labelResize');
        }
        contextMenuLabel.contents().filter(function() {return this.nodeType === 3;})[0].nodeValue = commandString;
      }
    }
  }
};

/**
  * Refresh the table dimensions
  * @param {number|null} width  table container width
  * @param {number|null} height  table container height
  * @param {boolean} resetScrollTop reset the scrollTop
  * @param {boolean} resetScrollLeft reset the scrollLeft
  * @memberof oj.TableDomUtils
  */
 oj.TableDomUtils.prototype.refreshTableDimensions = function(width, height, resetScrollTop, resetScrollLeft)
{
  // remove resize listeners to prevent triggering resize notifications
  this._refreshTableDimensions(width, height, resetScrollTop, resetScrollLeft);
  // set the current row
  this.component._setCurrentRow(this.options['currentRow'], null);
};

/**
 * Remove the visual indicator for column drag over
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.removeDragOverIndicatorColumn = function()
{
  var table = this.getTable();
  var indicatorElements = table.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_BEFORE_CLASS + ',' +
                                     '.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_AFTER_CLASS);

  var i = 0;
  if (indicatorElements && indicatorElements.length > 0)
  {
    var indicatorElementsCount = indicatorElements.length;
    
    for (i = 0; i < indicatorElementsCount; i++)
    {
      indicatorElements[i].classList.remove(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_BEFORE_CLASS);
      indicatorElements[i].classList.remove(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_INDICATOR_AFTER_CLASS);
    }
  }
}

/**
 * Remove the visual indicator for row drag over
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.removeDragOverIndicatorRow = function()
{
  var tableBody = this.getTableBody();
  var indicatorRowBeforeElements = tableBody.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_DRAG_INDICATOR_BEFORE_CLASS);
  var indicatorRowAfterElements = tableBody.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_DRAG_INDICATOR_AFTER_CLASS);
  
  var i = 0;
  if (indicatorRowBeforeElements && indicatorRowBeforeElements.length > 0)
  {
    var indicatorRowBeforeElementsCount = indicatorRowBeforeElements.length;
    
    for (i = 0; i < indicatorRowBeforeElementsCount; i++)
    {
      $(indicatorRowBeforeElements[i]).remove();
    }
  }
  if (indicatorRowAfterElements && indicatorRowAfterElements.length > 0)
  {
    var indicatorRowAfterElementsCount = indicatorRowAfterElements.length;
    
    for (i = 0; i < indicatorRowAfterElementsCount; i++)
    {
      $(indicatorRowAfterElements[i]).remove();
    }
  }
  
  var tableBodyRows = this.getTableBodyRows();
  
  if (tableBodyRows == null || tableBodyRows.length == 0)
  {
    this.component._showNoDataMessage();
  }
}

/**
 * Remove a tr element from the tbody DOM
 * @param {number} rowIdx  row index
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.removeTableBodyRow = function(rowIdx)
{
  var tableBodyRow = this.getTableBodyRow(rowIdx);
  if (tableBodyRow != null)
  {
    oj.Components.subtreeDetached(tableBodyRow);
    $(tableBodyRow).remove();
    this.clearCachedDomRowData();
  }
};

/**
 * Remove all tr elements from the tbody DOM
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.removeAllTableBodyRows = function()
{
  var tableBodyRows = this.getTableBodyRows();
  
  if (tableBodyRows != null && tableBodyRows.length > 0)
  {
    var tableBody = this.getTableBody();
    if (tableBody != null)
    {
      oj.Components.subtreeDetached(tableBody);
      $(tableBody).empty();
    }
    this.clearCachedDomRowData();
  }
};

/**
 * Finds and removes the touch selection icons from the DOM
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.removeTableBodyRowTouchSelectionAffordance = function()
{
  var tableContainer = this.getTableContainer();
  var touchAffordance = tableContainer.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOUCH_AREA_CLASS);
  
  if (touchAffordance != null && touchAffordance.length > 0)
  {
    var i;
    for (i = 0; i < touchAffordance.length; i++)
    {
      $(touchAffordance[i]).remove();
    }
  }
};

/**
 * Remove the drag image for the column
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.removeTableHeaderColumnDragImage = function()
{
  var tableContainer = this.getTableContainer();
  var headerDragImage = tableContainer.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DRAG_IMAGE);
  
  if (headerDragImage && headerDragImage.length > 0)
  {
    var i;
    var headerDragImageCount = headerDragImage.length;
    
    for (i = 0; i < headerDragImageCount; i++)
    {
      $(headerDragImage[i]).remove();
    }
  }
};

/**
 * Remove resize indicator
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.removeTableHeaderColumnResizeIndicator = function()
{
  var tableHeaderColumnResizeIndicator = this.getTableHeaderColumnResizeIndicator();
  if (tableHeaderColumnResizeIndicator)
  {
    $(tableHeaderColumnResizeIndicator).remove();
    this._clearCachedDom(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_RESIZE_INDICATOR_CLASS);
  }
};

/**
 * Set the element scrollLeft
 * @param {Element} element DOM element
 * @param {number} scrollLeft scrollLeft
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setScrollLeft = function(element, scrollLeft)
{
  if (this.component._GetReadingDirection() === "rtl")
  {
    scrollLeft = Math.abs(scrollLeft);
    if (this._isWebkit())
    {
      var maxScrollLeft = element.scrollWidth - element.clientWidth;
      scrollLeft = maxScrollLeft - scrollLeft;
    }
    else if (this._isFF())
    {
      scrollLeft = -1 * scrollLeft;
    }
  }
  
  if (element.scrollLeft != scrollLeft)
  {
    element.scrollLeft = scrollLeft;
  }
}

/**
 * Set the attributes on the cell
 * @param {number} rowIdx  row index
 * @param {Object} rowKey  row key
 * @param {Object} rowHashCode  row hash code
 * @param {number} columnIdx  column index
 * @param {Element} tableBodyCell  td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setTableBodyCellAttributes = function(rowIdx, rowKey, rowHashCode, columnIdx, tableBodyCell)
{
  // if cell already has an id then bail
  var cellId = tableBodyCell.getAttribute(oj.TableDomUtils.DOM_ATTR._ID);
  
  if (cellId != null &&
    cellId.length > 0)
  {
    return;
  }
    
  var accessibility = this.options['accessibility']; 
  var column = this.component._getColumnDefs()[columnIdx];
  
  if (column == null)
  {
    return;
  }
  
  var rowHeaderColumnId = null;
  var isTableHeaderless = this.component._isTableHeaderless();

  if (accessibility != null && accessibility['rowHeader'] != null)
  {
    rowHeaderColumnId = accessibility['rowHeader'];
  }
  else
  {
    rowHeaderColumnId = this.component._getColumnDefs()[0].id;
  }
  
  var rowKeyStr = rowKey != null ? rowKey.toString() : rowIdx.toString();
  
  var rowKeyStrHashCode = rowHashCode == null ? this.hashCode(rowKeyStr) : rowHashCode;

  var cellRowHeaderId = this.getTableId() + ':' + rowHeaderColumnId + '_' + rowKeyStrHashCode;

  var headers = this.getTableId() + ':' + column.id;
  if (rowHeaderColumnId == column.id)
  {
    cellId = cellRowHeaderId;
    
    if (isTableHeaderless)
    {
      headers = '';
    }
  }
  else
  {
    cellId = cellRowHeaderId + '_' + columnIdx;
    
    if (!isTableHeaderless)
    {
      headers = headers + ' ' + cellRowHeaderId;
    }
    else
    {
      headers = cellRowHeaderId;
    }
  }
  tableBodyCell.setAttribute(oj.TableDomUtils.DOM_ATTR._ID, cellId);

  if (!tableBodyCell.getAttribute(oj.TableDomUtils.DOM_ATTR._HEADERS))
  {
    tableBodyCell.setAttribute(oj.TableDomUtils.DOM_ATTR._HEADERS, headers);
  }
};

/**
 * Set the attributes on the row like rowIdx, etc
 * @param {Object} row row
 * @param {Element} tableBodyRow  tr DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setTableBodyRowAttributes = function(row, tableBodyRow)
{
  $(tableBodyRow).data('rowKey', row['key']);
  $(tableBodyRow).data('rowData', row['data']);
};

/**
 * Set the attributes on the header like columndx, etc
 * @param {number} columnIdx  column index
 * @param {Element} tableHeaderColumn  th DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setTableHeaderColumnAttributes = function(columnIdx, tableHeaderColumn)
{
  var column = this.component._getColumnDefs()[columnIdx];

  if (!tableHeaderColumn.getAttribute(oj.TableDomUtils.DOM_ATTR._ID))
  {
    tableHeaderColumn.setAttribute(oj.TableDomUtils.DOM_ATTR._ID, this.getTableId() + ':' + column.id);
  }
};

/**
 * Set the css class from all the cells in a column with the styleClass
 * @param {number|null} columnIdx  column index
 * @param {boolean} add add or remove the class
 * @param {string} styleClass style class
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setTableColumnCellsClass = function(columnIdx, add, styleClass)
{
  var tableBodyRows = this.getTableBodyRows();
  if (tableBodyRows != null && tableBodyRows.length > 0)
  {
    if (columnIdx === null)
    {
      var tableBodyCells = null;

      if (!add)
      {
        tableBodyCells = this.getTableBody().querySelectorAll('.' + styleClass);
      }
      else
      {
        tableBodyCells = this.getTableBody().querySelectorAll('[td]');
      }

      if (tableBodyCells != null && tableBodyCells.length > 0)
      {
        var i, tableBodyCellsCount = tableBodyCells.length;
        for (i = 0; i < tableBodyCellsCount; i++)
        {
          if (!add)
          {
            tableBodyCells[i].classList.remove(styleClass);
          }
          else
          {
            tableBodyCells[i].classList.add(styleClass);
          }
        }
      }
    }
    else
    {
      var i, tableBodyCell, tableBodyRowsCount = tableBodyRows.length;
      for (i = 0; i < tableBodyRowsCount; i++)
      {
        tableBodyCell = this.getTableBodyCell(i, columnIdx, null);

        if (!add)
        {
          tableBodyCell.classList.remove(styleClass);
        }
        else
        {
          tableBodyCell.classList.add(styleClass);
        }
      }
    }
  }
};

/**
 * Set the table body message.
 * @param {string} message
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setTableBodyMessage = function(message)
{
  var tableBodyMessageCell = this.getTableBodyMessageCell();
  $(tableBodyMessageCell).empty();
  tableBodyMessageCell.appendChild(document.createTextNode(message)); //@HTMLUpdateOK
};

/**
 * Set the table inline message.
 * @param {string} summary
 * @param {string} detail
 * @param {number} severityLevel
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setTableInlineMessage = function(summary, detail, severityLevel)
{
  var inlineMessage = this.getTableInlineMessage();
  $(inlineMessage).empty();
  $(inlineMessage).append(oj.PopupMessagingStrategyUtils.buildMessageHtml(document, summary, detail, severityLevel, null)); //@HTMLUpdateOK
};

/**
 * Set the table accessibility notification.
 * @param {string} status
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.setTableStatusAccNotification = function(status)
{
  var statusNotification = this.getTableStatusAccNotification();
  $(statusNotification).empty();
  statusNotification.appendChild(document.createTextNode(status)); //@HTMLUpdateOK
}

/**
 * Style the initial table
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleInitialTable = function()
{
  var table = this.getTable();
  var tableContainer = this.getTableContainer();
  var tableHeader = table.getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._THEAD);
  tableHeader = tableHeader.length > 0 ? tableHeader[0] : null;
  var tableFooter = table.getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._TFOOT);
  tableFooter = tableFooter.length > 0 ? tableFooter[0] : null;
  var tableBody = table.getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._TBODY);
  tableBody = tableBody.length > 0 ? tableBody[0] : null;
  // set the tabindex
  table.setAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX, '0');
  // set focusable
  this.component._focusable({'element': table, 'applyHighlight': true, 'setupHandlers': this._focusSetupHandlers.bind(this)});  
  this.styleTableHeader(tableHeader);
  this.styleTableFooter(tableFooter);
  this.styleTableBody(tableBody);
};

/**
 * Style the tbody element
 * @param {Element} tableBody tbody DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableBody = function(tableBody)
{
  tableBody.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_BODY_CLASS);
  // Add a special marker attribute to tell child components that they are container within table
  tableBody.setAttribute(oj.Components._OJ_CONTAINER_ATTR, this.component['widgetName']);
};

/**
 * Style the td element
 * @param {number} columnIdx  column index
 * @param {Element} tableBodyCell  td DOM element
 * @param {boolean} isNew is new cell
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableBodyCell = function(columnIdx, tableBodyCell, isNew)
{
  var options = this.options;
  var column = this.component._getColumnDefs()[columnIdx];
 
  if (isNew || !tableBodyCell.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_CLASS))
  {
    tableBodyCell.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_CLASS);
    tableBodyCell.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATE_CELL_FORM_CONTROL_CLASS);
    
    // add the column option stuff in here since we only want to do this once when adding
    // the table data cell class. Column option styling or class can be overridden
    // later in a custom renderer
    if (column != null && column.style != null && (isNew || tableBodyCell.getAttribute(oj.TableDomUtils.DOM_ATTR._STYLE) != column.style))
    {
      tableBodyCell.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, column.style);
    }
    // Use jquery hasClass and addClass because column.className can contain multiple classes
    if (column != null && column.className != null && (isNew || !$(tableBodyCell).hasClass(column.className)))
    {
      $(tableBodyCell).addClass(column.className);
    }
  }
  
  if (this._isVerticalGridEnabled())
  {
    if (isNew || !tableBodyCell.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_VGRID_LINES_CLASS))
    {
      tableBodyCell.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_VGRID_LINES_CLASS);
    }
  }
};

/**
 * Style the tr element
 * @param {Element} tableBodyRow  tr DOM element
 * @param {boolean} isNew is new row
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableBodyRow = function(tableBodyRow, isNew)
{
  if (isNew || !tableBodyRow.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS))
  {
    tableBodyRow.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_ROW_CLASS);
  }
  if (this._isHorizontalGridEnabled())
  {
    if (isNew || !tableBodyRow.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_HGRID_LINES_CLASS))
    {
      tableBodyRow.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_HGRID_LINES_CLASS);
    }
  }
};

/**
 * Style the table container
 * @param {Element} tableContainer  div DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableContainer = function(tableContainer)
{
  // add main css class to container
  tableContainer.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_CLASS);
  tableContainer.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_CONTAINER_CLASS);
  tableContainer.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._WIDGET);
  
  if (this.options.display == oj.TableDomUtils._OPTION_DISPLAY._GRID)
  {
    tableContainer.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_COMPACT_CLASS);
  }
  else
  {
    tableContainer.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_COMPACT_CLASS);
  }
  var editMode = this.options['editMode'];
        
  if (editMode != null && editMode != this.component._OPTION_EDIT_MODE._NONE)
  {
    tableContainer.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_EDIT_CLASS);
  }
  else
  {
    tableContainer.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_EDIT_CLASS);
  }
};

/**
 * Style the tfoot element
 * @param {Element} tableFooter tfoot DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableFooter = function(tableFooter)
{
  if (!tableFooter)
  {
    return;
  }
  tableFooter.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CLASS);
  var tableFooterRow = tableFooter.getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._TR)[0];
  tableFooterRow.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_ROW_CLASS);
};

/**
 * Style the td element
 * @param {number} columnIdx  column index
 * @param {Element} tableFooterCell  td DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableFooterCell = function(columnIdx, tableFooterCell)
{
  var options = this.options;
  var lastColumn = columnIdx == this.component._getColumnDefs().length - 1 ? true : false;
  var column = this.component._getColumnDefs()[columnIdx];

  if (column.footerStyle != null)
  {
    tableFooterCell.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, column.footerStyle);
  }
  if (!tableFooterCell.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CELL_CLASS))
  {
    tableFooterCell.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_FOOTER_CELL_CLASS);
  }
  if (this._isVerticalGridEnabled() && !tableFooterCell.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_VGRID_LINES_CLASS))
  {
    tableFooterCell.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_VGRID_LINES_CLASS);
  }
  if (column.footerClassName)
  {
    // Use jquery addClass because column.footerClassName can contain multiple classes
    $(tableFooterCell).addClass(column.footerClassName);
  }
};

/**
 * Style the thead element
 * @param {Element} tableHeader thead DOM element
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableHeader = function(tableHeader)
{
  if (!tableHeader)
  {
    return;
  }
  tableHeader.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_HEADER_CLASS);
  tableHeader.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = 'table-header-group';
  var tableHeaderRow = tableHeader.getElementsByTagName(oj.TableDomUtils.DOM_ELEMENT._TR)[0];
  tableHeaderRow.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_HEADER_ROW_CLASS);
  tableHeaderRow.style[oj.TableDomUtils.CSS_PROP._POSITION] = oj.TableDomUtils.CSS_VAL._RELATIVE;
};

/**
 * Style the th element
 * @param {number} columnIdx  column index
 * @param {Element} tableHeaderColumn  th DOM element
 * @param {string} columnSelectionMode  column selection mode
 * @param {boolean} isNew is new column
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype.styleTableHeaderColumn = function(columnIdx, tableHeaderColumn, columnSelectionMode, isNew)
{
  var column = this.component._getColumnDefs()[columnIdx];
  
  if (isNew || !tableHeaderColumn.classList.contains(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS))
  {
    tableHeaderColumn.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CELL_CLASS);
    if (this._isVerticalGridEnabled())
    {
      if (isNew || !tableHeaderColumn.classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_VGRID_LINES_CLASS))
      {
        tableHeaderColumn.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_VGRID_LINES_CLASS);
      }
    }
    if (column.headerStyle != null && (isNew || tableHeaderColumn.getAttribute(oj.TableDomUtils.DOM_ATTR._STYLE) != column.headerStyle))
    {
      tableHeaderColumn.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, column.headerStyle);
    }
    // Use jquery hasClass and addClass because column.headerClassName can contain multiple classes
    if (column.headerClassName != null && (isNew || !$(tableHeaderColumn).hasClass(column.headerClassName)))
    {
      $(tableHeaderColumn).addClass(column.headerClassName);
    }
  }
};

/**
 * Return all the colspanned logical elements
 * @param {Array} elementArray array of DOM elements
 * @return {Array|null} array of DOM elements
 * @memberof oj.TableDomUtils
 * @private
 */
oj.TableDomUtils.prototype._getColspanLogicalElements = function(elementArray)
{
  var i, j, indexedElementArrayNum = 0, indexedElementArray = [];
  var elementArrayCount = elementArray.length;
  for (i = 0; i < elementArrayCount; i++)
  {
    var colSpan = elementArray[i].getAttribute(oj.TableDomUtils.DOM_ATTR._COLSPAN);
    if (colSpan != null)
    {
      colSpan = parseInt(colSpan, 10);
      
      for (j = 0; j < colSpan; j++)
      {
        indexedElementArray[indexedElementArrayNum + j] = elementArray[i];
      }
      indexedElementArrayNum = indexedElementArrayNum + colSpan;
    }
    else
    {
      indexedElementArray[indexedElementArrayNum++] = elementArray[i];
    }
  }
  
  return indexedElementArray;
}

/**
  * Helper function which returns if the horizontal grid lines are enabled.
  * @private
  * @return {boolean} enabled
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._isHorizontalGridEnabled = function()
{
  if (this.options.horizontalGridVisible == oj.TableDomUtils._OPTION_ENABLED ||
      this.options.horizontalGridVisible == oj.TableDomUtils._OPTION_AUTO)
  {
    return true;
  }
  return false;
};

/**
  * Helper function which returns if the vertical grid lines are enabled.
  * @private
  * @return {boolean} enabled
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._isVerticalGridEnabled = function()
{
  if (this.options.verticalGridVisible == oj.TableDomUtils._OPTION_ENABLED ||
      (this.options.verticalGridVisible == oj.TableDomUtils._OPTION_AUTO &&
       this.options.display == oj.TableDomUtils._OPTION_DISPLAY._GRID))
  {
    return true;
  }
  return false;
};

/**
  * Helper function which returns if the browser is FF
  * @private
  * @memberof oj.TableDomUtils
  * @return {boolean} FF
  */
oj.TableDomUtils.prototype._isFF = function()
{
  if (navigator.userAgent.toLowerCase().indexOf('firefox') > -1)
  {
    return true;
  }
  return false;
};

/**
  * Helper function which returns if the browser is IE and if so the version.
  * @private
  * @memberof oj.TableDomUtils
  * @return {number|null} IE version. null if not IE.
  */
oj.TableDomUtils.prototype._isIE = function()
{
  if (typeof this._browserIsIE == 'undefined')
  {
    var userAgent = navigator.userAgent;
    if (navigator.appName == 'Microsoft Internet Explorer')
    {
      var resultArray = (new RegExp('MSIE ([0-9]{1,}[\.0-9]{0,})')).exec(userAgent);
      if (resultArray != null)
      {
        this._browserIsIE = parseFloat(resultArray[1]);
      }
    }
    else if (userAgent.indexOf('Trident') >= 0)
    {
      this._browserIsIE = 11;
    }
    else
    {
      this._browserIsIE = null;
    }
  }
  return this._browserIsIE;
};

/**
  * Helper function which returns if the browser is webkit.
  * @private
  * @memberof oj.TableDomUtils
  * @return {boolean} webkit
  */
oj.TableDomUtils.prototype._isWebkit = function()
{
  if (typeof this._browserIsWebkit == 'undefined')
  {
    var ua = navigator.userAgent.toLowerCase();
    this._browserIsWebkit = (/webkit/.test(ua)) && !(/edge/.test(ua));
  }
  return this._browserIsWebkit;
};

/**
  * Refresh the table dimensions
  * @param {number|null} width  table container width
  * @param {number|null} height  table container height
  * @param {boolean} resetScrollTop reset the scrollTop
  * @param {boolean} resetScrollLeft reset the scrollLeft
  * @memberof oj.TableDomUtils
  * @private
  */
 oj.TableDomUtils.prototype._refreshTableDimensions = function(width, height, resetScrollTop, resetScrollLeft)
{
  var options = this.options;
  var table = this.getTable();
  var tableContainer = this.getTableContainer();
  var tableBody = this.getTableBody();
    
  // preserve the scrollTop & scrollLeft
  var scrollTop = null;
  var scrollLeft = null;
  
  if (this.component._scrollTop != null && 
    this.component._scrollTop > 0)
  {
    scrollTop = this.component._scrollTop;
  }
  if (this.component._scrollLeft != null && 
    this.component._scrollLeft > 0)
  {
    scrollLeft = this.component._scrollLeft;
  }
  
  
  // first remove any styling so that the browser sizes the table
  this._removeTableDimensionsStyling();
  this.styleTableContainer(tableContainer);
  
  var tableContainerScrollableState = this.isTableContainerScrollable();
  
  if (tableContainerScrollableState[0] === 1)
  {
    this._tableHeightConstrained = true;
  }
  else
  {
    this._tableHeightConstrained = false;
  }
  if (tableContainerScrollableState[1] === 1)
  {
    this._tableWidthConstrained = true;
  }
  else
  {
    this._tableWidthConstrained = false;
  }
  
  
  if (tableBody == null)
  {
    return;
  }
   
  if (this._tableHeightConstrained || this._tableWidthConstrained)
  {
    var tableHeader = this.getTableHeader();
    var tableFooter = this.getTableFooter();
    var tableBottomSlot = this.getTableBottomSlot();
    // Add the oj-table-scroll class because some styling only applies
    // to scrollable table.
    tableContainer.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_SCROLL_CLASS);
    
    if (!this._tableDimensions)
    {
      this._tableDimensions = {};
    }
    
    // use constants for width height
    if (width > 0 || height > 0)
    {
      if (width > 0 && this._tableWidthConstrained)
      {
        this._tableDimensions[oj.TableDomUtils.CSS_PROP._WIDTH] = width;
      }
      if (height > 0 && this._tableHeightConstrained)
      {
        this._tableDimensions[oj.TableDomUtils.CSS_PROP._HEIGHT] = height;
      }
    }
    else
    {
      if (this._tableHeightConstrained)
      {
        this._tableDimensions[oj.TableDomUtils.CSS_PROP._HEIGHT] = tableContainer.offsetHeight;
      }
      else
      {
        this._tableDimensions[oj.TableDomUtils.CSS_PROP._HEIGHT] = 0;
      }
      if (this._tableWidthConstrained)
      {
        this._tableDimensions[oj.TableDomUtils.CSS_PROP._WIDTH] = tableContainer.offsetWidth;
      }
      else
      {
        this._tableDimensions[oj.TableDomUtils.CSS_PROP._WIDTH] = 0;
      }
    }

    if (this._tableBorderWidth == null)
    {
      this._tableBorderWidth = $(tableContainer).outerWidth() - $(tableContainer).innerWidth();
    }
    if (this._tableBorderHeight == null)
    {
      this._tableBorderHeight = $(tableContainer).outerHeight() - $(tableContainer).innerHeight();
    }
    
    // if there is a vertical scrollbar but no horizontal scrollbar then we need
    // to make sure the width sizes to accommodate the scrollbar
    var tableContainerScrollbarWidth = 0;
    if (this._tableHeightConstrained && !this._tableWidthConstrained)
    {
      tableContainer.style[oj.TableDomUtils.CSS_PROP._OVERFLOW] = oj.TableDomUtils.CSS_VAL._AUTO;
      tableContainerScrollbarWidth = tableContainer.offsetWidth - tableContainer.clientWidth;
      
      // if tableContainer has set width then we need to shrink, otherwise we can expand
      if (this._tableDimensions[oj.TableDomUtils.CSS_PROP._WIDTH] > 0)
      {
        table.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableContainer.clientWidth + oj.TableDomUtils.CSS_VAL._PX;
      }
      else
      {
        table.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableContainer.scrollWidth + tableContainerScrollbarWidth + oj.TableDomUtils.CSS_VAL._PX;
        
        // check if we caused overflow
        tableContainerScrollableState = this.isTableContainerScrollable();
        
        if (tableContainerScrollableState[1] === 1)
        {
          // if we caused overflow we need to shrink so that we can set the column widths to the width without
          // the scrollbar
          table.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableContainer.clientWidth + oj.TableDomUtils.CSS_VAL._PX;
        }
      }
      tableContainer.style[oj.TableDomUtils.CSS_PROP._OVERFLOW] = '';
    }
    
    // let the browser layout the column widths
    this._setColumnWidths();
    
    // add in scrolling class for specific vertical or horizontal scrolling
    if (this._tableHeightConstrained)
    {
      tableContainer.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_SCROLL_VERTICAL_CLASS);
    }
    if (this._tableWidthConstrained)
    {
      tableContainer.classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_SCROLL_HORIZONTAL_CLASS);
    }
    var tableWidth = table.clientWidth;
    
    if (tableContainerScrollbarWidth > 0)
    {
      // if the container scrollbar width > 0 that means that we have overflow even though we're not width constrained
      // so we've set table width to table container clientWidth to set the column widths to accommodate the vertical scrollbar
      // so we can set the table back to container offset width
      table.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableContainer.offsetWidth + oj.TableDomUtils.CSS_VAL._PX;
    }

    var captionHeight = 0;
    var caption = table.getElementsByTagName('caption');
    if (caption != null && caption.length > 0)
    {
      caption = caption[0];
      captionHeight = $(caption).outerHeight();
      caption.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._INLINE;
      if (tableHeader != null)
      {
        var tableContainerStyle = window.getComputedStyle(tableContainer);
        tableHeader.style[oj.TableDomUtils.CSS_PROP._BORDER_TOP] = tableContainerStyle[oj.TableDomUtils.CSS_PROP._BORDER_TOP];
      }
    }

    // apply the styling which sets the fixed column headers, etc
    var tableHeaderHeight = 0;
    if (tableHeader != null)
    {
      tableHeaderHeight = $(tableHeader).outerHeight();
      }

    var tableFooterHeight = 0;
    var scrollbarWidth;
    var tableBodyHeight = 0;

      if (this._tableWidthConstrained)
      {
      tableBody.style[oj.TableDomUtils.CSS_PROP._WIDTH] = this._tableDimensions[oj.TableDomUtils.CSS_PROP._WIDTH] - this._tableBorderWidth + oj.TableDomUtils.CSS_VAL._PX;
      }
      else if (tableContainerScrollbarWidth > 0)
      {
      tableBody.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableWidth - this._tableBorderWidth + tableContainerScrollbarWidth + oj.TableDomUtils.CSS_VAL._PX;
      }
      
      if (tableFooter != null)
      {
      tableFooterHeight = $(tableFooter).outerHeight();
      tableBody.style[oj.TableDomUtils.CSS_PROP._TOP] = -1 * tableFooterHeight + oj.TableDomUtils.CSS_VAL._PX;
      }

    var tableInlineMessageHeight = 0;
    var inlineMessage = this.getTableInlineMessage();
    if (inlineMessage != null && inlineMessage.style[oj.TableDomUtils.CSS_PROP._DISPLAY] != oj.TableDomUtils.CSS_VAL._NONE && inlineMessage.clientHeight > 0)
    {
      tableInlineMessageHeight = $(inlineMessage).outerHeight();
    }
    
    var tableBottomSlotHeight = 0;
    
    if (tableBottomSlot != null && tableBottomSlot.style[oj.TableDomUtils.CSS_PROP._DISPLAY] != oj.TableDomUtils.CSS_VAL._NONE && tableBottomSlot.clientHeight > 0)
    {
      tableBottomSlotHeight = $(tableBottomSlot).outerHeight();
    }

    // Size the table body to fit in the height
      if (this._tableHeightConstrained)
      {
      tableBodyHeight = this._tableDimensions[oj.TableDomUtils.CSS_PROP._HEIGHT] - tableHeaderHeight - tableFooterHeight - captionHeight - tableInlineMessageHeight - tableBottomSlotHeight - this._tableBorderHeight;
        if (tableBodyHeight > 0)
        {
        tableBody.style[oj.TableDomUtils.CSS_PROP._HEIGHT] = tableBodyHeight + oj.TableDomUtils.CSS_VAL._PX;
        tableBody.style[oj.TableDomUtils.CSS_PROP._MIN_HEIGHT] = tableBodyHeight + oj.TableDomUtils.CSS_VAL._PX;
        }
      }
      else
      {
      tableBodyHeight = $(tableBody).outerHeight();
      }

      scrollbarWidth = this.getScrollbarWidth();

      if (this._tableWidthConstrained)
      {
        var tableBodyRows = this.getTableBodyRows();

        if (tableBodyRows != null && tableBodyRows.length > 0)
        {
        tableBody.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_X] = oj.TableDomUtils.CSS_VAL._AUTO;
        tableContainer.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_X] = oj.TableDomUtils.CSS_VAL._HIDDEN;
        tableContainer.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_Y] = oj.TableDomUtils.CSS_VAL._HIDDEN;

          if (tableHeader != null)
          {
            var tableHeaderWidth = this._tableDimensions[oj.TableDomUtils.CSS_PROP._WIDTH] - this._tableBorderWidth;

            if (tableHeaderWidth > 0)
            {
              if (scrollbarWidth > 0)
              {
                // if we have scrollbars then size the tableheader 
                // to align with the scrollbars
              tableHeader.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableHeaderWidth - scrollbarWidth + oj.TableDomUtils.CSS_VAL._PX;
              }
              else
              {
              tableHeader.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableHeaderWidth + oj.TableDomUtils.CSS_VAL._PX;
              }
            }
          }
        }
        else
        {
          // if we have no data then use the tableContainer's horizontal scroller
        tableBody.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_X] = oj.TableDomUtils.CSS_VAL._HIDDEN;
        tableContainer.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_X] = oj.TableDomUtils.CSS_VAL._AUTO;
        tableContainer.style[oj.TableDomUtils.CSS_PROP._OVERFLOW_Y] = oj.TableDomUtils.CSS_VAL._HIDDEN;
        }
      }

    if (tableFooter != null)
    {
      tableFooter.style[oj.TableDomUtils.CSS_PROP._TOP] = tableBodyHeight + oj.TableDomUtils.CSS_VAL._PX;
    }
    this._setHeaderColumnLastWidth();
  }
  else
  {
    if (this.component._isTableColumnsWidthSet())
    {
      this._setColumnWidths();
    }
  }
  this._setHeaderColumnOverflowWidths();
  this._removeTableBodyRowBottomBorder(tableContainerScrollableState[0] < 0);
  this.refreshTableStatusPosition();
  
  if (scrollTop != null && !resetScrollTop)
  {
    var maxScrollTop = this.getScroller().scrollHeight - this.getScroller().clientHeight;
    scrollTop = scrollTop > maxScrollTop ? maxScrollTop : scrollTop;
    if (this.component._isLoadMoreOnScroll() && maxScrollTop == scrollTop)
    {
      // Do not set to maxScrollTop or we will cause another fetch
      scrollTop--;
    }
    this.getScroller().scrollTop = scrollTop;
    $(this.getScroller()).scroll();
  }
  else if (resetScrollTop)
  {
    this.getScroller().scrollTop = 0;
    $(this.getScroller()).scroll();
  }
  
  if (scrollLeft != null && !resetScrollLeft)
  {
    this.setScrollLeft(this.getScroller(), scrollLeft);
    $(this.getScroller()).scroll();
  }
  else if (resetScrollLeft)
  {
    this.setScrollLeft(this.getScroller(), 0);
    $(this.getScroller()).scroll();
  }
};

/**
  * Refresh the table status position
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype.refreshTableStatusPosition = function()
{
  // if status message is hidden then return
  if (!this.component._isStatusMessageShown())
    return;

  var tableContainer = this.getTableContainer();
  var tableBody = this.getTableBody();
  var tableStatusMessage = this.getTableStatusMessage();
  
  var overTableElement = tableContainer;
  
  if (tableBody.clientHeight > 0)
  {
    overTableElement = tableBody;
  }

  // size the messaging background
  tableStatusMessage.style[oj.TableDomUtils.CSS_PROP._HEIGHT] = tableBody.clientHeight + oj.TableDomUtils.CSS_VAL._PX;
  tableStatusMessage.style[oj.TableDomUtils.CSS_PROP._WIDTH] = tableBody.clientWidth + oj.TableDomUtils.CSS_VAL._PX;
  var tableStatusMessageText = tableStatusMessage.children[0];
  // refresh the status message position
  var isRTL = (this.component._GetReadingDirection() === "rtl");
  // position status in the center
  var options = {'my': 'center', 'at': 'center', 'collision': 'none', 'of': overTableElement};
  options = oj.PositionUtils.normalizeHorizontalAlignment(options, isRTL);
  (/** @type {Object} */ ($(tableStatusMessage))).position(options);
  var msgTextOptions = {'my': 'center', 'at': 'center', 'collision': 'none', 'of': tableStatusMessage};
  msgTextOptions = oj.PositionUtils.normalizeHorizontalAlignment(msgTextOptions, isRTL);
  (/** @type {Object} */ ($(tableStatusMessageText))).position(msgTextOptions);
};

/**
  * Cache DOM Element
  * @param {string} key key for identifying the element
  * @param {Element} element  DOM element
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._cacheDomElement = function (key, element) 
{
  if (!this._domCache)
  {
    this._domCache = {};
  }
  this._domCache[key] = element;
};

/**
  * Clear the DOM cache
  * @param {string|null} key key for identifying the element
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._clearCachedDom = function (key)
{
  if (key && this._domCache) 
  {
    delete this._domCache[key];
  }
  else
  {
    this._domCache = {};
  }
};

/**
  * Get cached DOM Element
  * @param {string} key key for identifying the element
  * @return {Element|null} returns the cached DOM element
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._getCachedDomElement = function (key) 
{
  if (this._domCache)
  {
    return this._domCache[key];
  }
  return null;
};

/**
  * Check cached DOM Element
  * @param {string} key key for identifying the element
  * @return {boolean} returns whether the key is cached
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._isCachedDomElement = function (key) 
{
  if (!this._domCache)
  {
    this._domCache = {};
  }
  return Object.keys(this._domCache).indexOf(key) !== -1;
};

/**
  * Get the DOM element contained in table
  * @param {string} className css class name for the table element
  * @param {boolean=} onlyChildren check only the direct children if true
  * @return {Element|null} returns the DOM element
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._getTableElementByClassName = function (className, onlyChildren) 
{
  if (!this._isCachedDomElement(className))
  {
    var tableContainer = this.getTableContainer();
    if (tableContainer)
    {
      this._cacheDomElement(className, null);
      
      if (!onlyChildren)
      {
        var domElement = tableContainer.getElementsByClassName(className);
        if (domElement && domElement.length > 0)
        {
          this._cacheDomElement(className, domElement[0]);
        }
      }
      else
      {
        this._cacheDomElement(className, this._getChildElementByClassName(tableContainer, className));
       
        if (this._getCachedDomElement(className) == null)
        {
          var table = this.getTable();
          this._cacheDomElement(className, this._getChildElementByClassName(table, className));
        }
      }
    }
  }

  return this._getCachedDomElement(className);
};

/**
  * Get the child DOM element by class name
  * @param {Element} parentElement parent element
  * @param {string} className css class name for the table element
  * @return {Element|null} returns the DOM element
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._getChildElementByClassName = function (parentElement, className) 
{
  if (parentElement.childNodes != null && parentElement.childNodes.length > 0)
  {
    var i;

    for (i = 0; i < parentElement.childNodes.length; i++) 
    {
      if (parentElement.childNodes[i].classList != null &&
        parentElement.childNodes[i].classList.contains(className)) 
      {
        return parentElement.childNodes[i];
      }
    }
  }
  return null;
};

/**
  * Focus setup handlers
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._focusSetupHandlers = function (focusInHandler, focusOutHandler) 
{
  this.component._setFocusInHandler(focusInHandler);
  this.component._setFocusOutHandler(focusOutHandler);
};

/**
  * Iterate through the columns and remove the widths
  * @private
  * @memberof oj.TableDomUtils
  */
oj.TableDomUtils.prototype._removeHeaderColumnAndCellColumnWidths = function()
{
  var columns = this.component._getColumnDefs();

  var i, headerColumn, footerCell, columnCount = columns.length;
  for (i = 0; i < columnCount; i++)
  {
    headerColumn = this.getTableHeaderColumn(i);
    if (headerColumn != null)
    {
      headerColumn.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = '';
    }
  }

  var tableBodyRows = this.getTableBodyRows();
  if (tableBodyRows != null && tableBodyRows.length > 0)
  {
    var tableBodyCell, j;
    for (i = 0; i < columnCount; i++)
    {

      tableBodyCell = this.getTableBodyCell(0, i, null);
      if (tableBodyCell != null)
      {
        tableBodyCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = '';
     
        if (this._forcedWidthColumns != null && this._forcedWidthColumns[i])
        {
          for (j = 0; j < tableBodyRows.length; j++) 
          {
            tableBodyCell = this.getTableBodyCell(j, i, null);
            tableBodyCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = '';
            tableBodyCell.style[oj.TableDomUtils.CSS_PROP._WIDTH] = '';
            tableBodyCell.style[oj.TableDomUtils.CSS_PROP._MAX_WIDTH] = '';
          }
        }
      }
    }
  }
  
  for (i = 0; i < columnCount; i++)
  {
    footerCell = this.getTableFooterCell(i);
    if (footerCell != null)
    {
      footerCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = '';
    }
  }
};

/**
 * Remove table cell bottom border
 * @param {boolean} underflow  table content underflow
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype._removeTableBodyRowBottomBorder = function(underflow)
{
  if (!this._isHorizontalGridEnabled())
  {
    return;
  }
  
  var tableBodyRows = this.getTableBodyRows();

  if (tableBodyRows != null && tableBodyRows.length > 0)
  {
    // first make sure that all rows have hgrid
    var i, tableBodyRowsCount = tableBodyRows.length;
    for (i = 0; i < tableBodyRowsCount; i++)
    {
      if (!tableBodyRows[i].classList.contains(oj.TableDomUtils.CSS_CLASSES._TABLE_HGRID_LINES_CLASS))
      {
        tableBodyRows[i].classList.add(oj.TableDomUtils.CSS_CLASSES._TABLE_HGRID_LINES_CLASS);
      }
    }
    var lastTableBodyRow = tableBodyRows[tableBodyRows.length - 1];

    if (!underflow)
    {
      lastTableBodyRow.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_HGRID_LINES_CLASS);
    }
  }
};
 
/**
 * Remove table dimensions styling
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype._removeTableDimensionsStyling = function()
{
  var table = this.getTable();
  var tableContainer = this.getTableContainer();
  var tableHeader = this.getTableHeader();        
  var tableHeaderRow = this.getTableHeaderRow();
  var tableFooter = this.getTableFooter();
  var tableFooterRow = this.getTableFooterRow();
  var tableBody = this.getTableBody();

  tableContainer.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_SCROLL_CLASS);
  tableContainer.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_SCROLL_VERTICAL_CLASS);
  tableContainer.classList.remove(oj.TableDomUtils.CSS_CLASSES._TABLE_SCROLL_HORIZONTAL_CLASS);
  tableContainer.style[oj.TableDomUtils.CSS_PROP._OVERFLOW] = oj.TableDomUtils.CSS_VAL._VISIBLE;
  // first remove any styling so that the browser sizes the table
  if (tableHeader != null)
  {
    tableHeader.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, '');
    tableHeaderRow.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, '');
    
    var headerColumnTextDivs = tableHeaderRow.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_TEXT_CLASS);
    var i;
    if (headerColumnTextDivs && headerColumnTextDivs.length > 0)
    {
      var headerColumnTextDivsCount = headerColumnTextDivs.length;
      for (i = 0; i < headerColumnTextDivsCount; i++)
      {
        headerColumnTextDivs[i].style[oj.TableDomUtils.CSS_PROP._WIDTH] = '';
      }
    }
  }
  if (tableFooter != null)
  {
    tableFooter.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, '');
    tableFooterRow.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, '');
  }
  table.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = '';
  table.style[oj.TableDomUtils.CSS_PROP._WIDTH] = '';
  
  if (tableBody != null)
  {
    tableBody.setAttribute(oj.TableDomUtils.DOM_ATTR._STYLE, '');
  }

  this._removeHeaderColumnAndCellColumnWidths();
};

/**
 * Iterate through the columns and get and then set the widths
 * for the columns and first row this is so that when we re-apply the styling
 * the headers and footers will align with the cells
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype._setColumnWidths = function()
{
  var columns = this.component._getColumnDefs();
  var columnWidths = [];
  var columnPaddingWidths = [];
  var headerColumnDivMinHeights = [];
  var tableBodyCellPaddingWidths = [];
  var footerPaddingWidths = [];
  var defaultColumnPaddingWidth, defaultTableBodyCellPaddingWidth, headerRenderer, cellRenderer = null;
  this._forcedWidthColumns = []
  var i, j, headerColumnCell, headerColumnCellStyle, headerColumnDiv, headerColumnTextDiv, footerCell, footerCellStyle, columnWidth, forceWidth, columnsCount = columns.length;
  for (i = 0; i < columnsCount; i++)
  {
    headerColumnCell = this.getTableHeaderColumn(i);
    if (headerColumnCell != null)
    {
      // read in the widths first. Set the widths in a separate loop so setting
      // the widths of early columns does not affect the widths of the rest
      headerColumnCellStyle = window.getComputedStyle(headerColumnCell);
      if (columns[i].width > 0)
      {
        this._forcedWidthColumns[i] = true;
        columnWidth = columns[i].width;
      }
      else
      {
        this._forcedWidthColumns[i] = false;
        columnWidth = parseInt(headerColumnCellStyle[oj.TableDomUtils.CSS_PROP._WIDTH], 10) || 0;
      }
      columnWidths[i] = columnWidth;
      headerRenderer = this.component._getColumnRenderer(i, 'header');
      
      if (!defaultColumnPaddingWidth && headerRenderer == null)
      {
        defaultColumnPaddingWidth = (parseFloat(headerColumnCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_RIGHT]) || 0) + (parseFloat(headerColumnCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_LEFT]) || 0);
        columnPaddingWidths[i] = defaultColumnPaddingWidth;
      }
      else if (headerRenderer != null)
      {
        columnPaddingWidths[i] = (parseFloat(headerColumnCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_RIGHT]) || 0) + (parseFloat(headerColumnCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_LEFT]) || 0);
      }
      else
      {
        columnPaddingWidths[i] = defaultColumnPaddingWidth;
      }
      // also determine the header heights
      headerColumnTextDiv = headerColumnCell.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_TEXT_CLASS);
      if (headerColumnTextDiv && headerColumnTextDiv.length > 0)
      {
        headerColumnDivMinHeights[i] = headerColumnTextDiv[0].clientHeight + oj.TableDomUtils.CSS_VAL._PX;
      }
    }
  }

  var adjustedColumnWidths = [];
  var tableBodyRows = this.getTableBodyRows();
  if (tableBodyRows != null && tableBodyRows.length > 0)
  {
    var tableBodyCell, tableBodyCellStyle, tableBodyCellPaddingWidth, adjustedColumnWidth;
    for (i = 0; i < columnsCount; i++)
    {

      tableBodyCell = this.getTableBodyCell(0, i, null);
      if (tableBodyCell != null)
      {
        cellRenderer = this.component._getColumnRenderer(i, 'cell');
      
        if (this.component._hasRowOrCellRendererOrTemplate(i))
        {
          tableBodyCellStyle = window.getComputedStyle(tableBodyCell);
          tableBodyCellPaddingWidth = (parseFloat(tableBodyCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_RIGHT]) || 0) + (parseFloat(tableBodyCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_LEFT]) || 0);
        }
        else
        {
          if (!defaultTableBodyCellPaddingWidth)
          {
            tableBodyCellStyle = window.getComputedStyle(tableBodyCell);
            defaultTableBodyCellPaddingWidth = (parseFloat(tableBodyCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_RIGHT]) || 0) + (parseFloat(tableBodyCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_LEFT]) || 0);
          }
          tableBodyCellPaddingWidth = defaultTableBodyCellPaddingWidth;
        }
        tableBodyCellPaddingWidths[i] = tableBodyCellPaddingWidth;
        adjustedColumnWidth = null;
        if (tableBodyCellPaddingWidth > columnPaddingWidths[i])
        {
          adjustedColumnWidth = columnWidths[i] - tableBodyCellPaddingWidth + columnPaddingWidths[i];
        }
        else
        {
          adjustedColumnWidth = columnWidths[i] + columnPaddingWidths[i] - tableBodyCellPaddingWidth;
        }
        adjustedColumnWidths[i] = adjustedColumnWidth;
      }
    }
  }
  
  for (i = 0; i < columnsCount; i++)
  {
    footerCell = this.getTableFooterCell(i);
    if (footerCell != null)
    {
      footerCellStyle = window.getComputedStyle(footerCell);
      footerPaddingWidths[i] = (parseFloat(footerCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_RIGHT]) || 0) + (parseFloat(footerCellStyle[oj.TableDomUtils.CSS_PROP._PADDING_LEFT]) || 0);
    }
  }

  for (i = 0; i < columnsCount; i++)
  {
    headerColumnCell = this.getTableHeaderColumn(i);
    if (headerColumnCell != null)
    {
      // also set the header heights
      if (headerColumnDivMinHeights[i])
      {
        headerColumnDiv = headerColumnCell.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CLASS);
        
        if (headerColumnDiv && headerColumnDiv.length > 0)
        {
          headerColumnDiv[0].style[oj.TableDomUtils.CSS_PROP._MIN_HEIGHT] = headerColumnDivMinHeights[i];
        }
      }

      if (!this._forcedWidthColumns[i])
      {
        headerColumnCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = columnWidths[i] + oj.TableDomUtils.CSS_VAL._PX;
      }
      else
      {
        forceWidth = columnWidths[i] - columnPaddingWidths[i] + oj.TableDomUtils.CSS_VAL._PX;
        headerColumnCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = forceWidth;
        headerColumnCell.style[oj.TableDomUtils.CSS_PROP._WIDTH] = forceWidth;
        headerColumnCell.style[oj.TableDomUtils.CSS_PROP._MAX_WIDTH] = forceWidth;
      }
    }
   
    tableBodyCell = this.getTableBodyCell(0, i, null);
    if (tableBodyCell != null)
    {
      if (!this._forcedWidthColumns[i])
      {
        tableBodyCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = adjustedColumnWidths[i] + oj.TableDomUtils.CSS_VAL._PX;
      }
      else
      {
        forceWidth = columnWidths[i] - tableBodyCellPaddingWidths[i] + oj.TableDomUtils.CSS_VAL._PX;
        for (j = 0; j < tableBodyRows.length; j++)
        {
          tableBodyCell = this.getTableBodyCell(j, i, null);
          tableBodyCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = forceWidth;
          tableBodyCell.style[oj.TableDomUtils.CSS_PROP._WIDTH] = forceWidth;
          tableBodyCell.style[oj.TableDomUtils.CSS_PROP._MAX_WIDTH] = forceWidth;
        }
      }
    }
    
    footerCell = this.getTableFooterCell(i);
    if (footerCell != null)
    {      
      if (!this._forcedWidthColumns[i])
      {
        // adjust the padding widths if the footer has more padding
        if ( footerPaddingWidths[i] > columnPaddingWidths[i])
        {
          adjustedColumnWidth = columnWidths[i] -  footerPaddingWidths[i] + columnPaddingWidths[i];
        }
        else
        {
          adjustedColumnWidth = columnWidths[i] + columnPaddingWidths[i] -  footerPaddingWidths[i];
        }

        footerCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = adjustedColumnWidth + oj.TableDomUtils.CSS_VAL._PX;
      }
      else
      {
        forceWidth = columnWidths[i] - footerPaddingWidths[i] + oj.TableDomUtils.CSS_VAL._PX;
        footerCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = forceWidth;
        footerCell.style[oj.TableDomUtils.CSS_PROP._WIDTH] = forceWidth;
        footerCell.style[oj.TableDomUtils.CSS_PROP._MAX_WIDTH] = forceWidth;
      }
    }
  }
};

/**
 * Fix up the last header column width when there is a scrollbar
 * on the tbody
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype._setHeaderColumnLastWidth = function()
{
 var scrollbarWidth = this.getScrollbarWidth();
 
 if (scrollbarWidth > 0)
 {
   var scrollbarHeight = this.getScrollbarHeight();
   
   // only do this if we don't have a horizontal scrollbar. If we have a
   // horizontal scrollbar then the header is constrained by the scroller
   if (scrollbarHeight <= 1)
   {
     // make the last column header take up the space of the table header.
     // we need to do this because the vertical scrollbar takes up space
     // in the tbody, making the last td cells smaller. So the corresponding
     // th ends up being shrunk unnecessarily
     var columns = this.component._getColumnDefs();
     var columnsCount = columns.length;
     var lastHeaderColumnCell = this.getTableHeaderColumn(columnsCount - 1);
     var totalHeaderColumnCellWidth = 0;

     if (lastHeaderColumnCell != null)
     {
       // get the total header column width
       var i, column, headerColumnCell, headerColumnCellStyle, headerColumnCellWidth;
       for (i = 0; i < columnsCount; i++)
       {
         column = columns[i];  
         headerColumnCell = this.getTableHeaderColumn(i);
         headerColumnCellStyle = window.getComputedStyle(headerColumnCell);
         headerColumnCellWidth = parseInt(headerColumnCellStyle[oj.TableDomUtils.CSS_PROP._WIDTH], 10) || 0;
         totalHeaderColumnCellWidth = totalHeaderColumnCellWidth + headerColumnCellWidth;
       }
       var tableHeader = this.getTableHeader();
       var tableHeaderStyle = window.getComputedStyle(tableHeader);
       var tableHeaderWidth = parseInt(tableHeaderStyle[oj.TableDomUtils.CSS_PROP._WIDTH], 10) || 0;
       
       if (tableHeaderWidth > totalHeaderColumnCellWidth)
       {
         var lastHeaderColumnCellStyle = window.getComputedStyle(lastHeaderColumnCell);
         var lastHeaderColumnCellWidth = parseInt(lastHeaderColumnCellStyle[oj.TableDomUtils.CSS_PROP._WIDTH], 10) || 0;
         lastHeaderColumnCell.style[oj.TableDomUtils.CSS_PROP._MIN_WIDTH] = lastHeaderColumnCellWidth + scrollbarWidth + oj.TableDomUtils.CSS_VAL._PX;
       }
     }
   }
 }
};

/**
 * Iterate through the header columns and set widths for those
 * which have overflow so that the text displays an ellipsis
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils.prototype._setHeaderColumnOverflowWidths = function()
{
 var columns = this.component._getColumnDefs();
 var i, column, headerColumnCell, headerColumnDiv, headerColumnTextDiv, headerColumnTextDivWidth, newHeaderColumnTextDivWidth, sortPlaceHolderDivWidth, columnsCount = columns.length;
 for (i = 0; i < columnsCount; i++)
 {
   column = columns[i];  
   headerColumnCell = this.getTableHeaderColumn(i);
   if (headerColumnCell != null)
   {
     // if we have overflow
     headerColumnDiv = headerColumnCell.getElementsByClassName(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CLASS);
     if (headerColumnDiv.length > 0)
     {
       headerColumnDiv = headerColumnDiv[0];
       headerColumnTextDiv = headerColumnCell.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_TEXT_CLASS);
       if (headerColumnTextDiv && headerColumnTextDiv.length > 0)
       {
         sortPlaceHolderDivWidth = 0;
         if (column.sortable == oj.TableDomUtils._OPTION_ENABLED)
         {
           sortPlaceHolderDivWidth = headerColumnCell.querySelectorAll('.' + oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_SORT_PACEHOLDER_CLASS)[0].clientWidth;
         }
         
         if (headerColumnDiv.clientWidth > 0 && headerColumnTextDiv[0].clientWidth + sortPlaceHolderDivWidth > headerColumnDiv.clientWidth)
         {
           headerColumnTextDiv[0].style[oj.TableDomUtils.CSS_PROP._WIDTH] = '';
           headerColumnTextDivWidth = headerColumnTextDiv[0].clientWidth
           newHeaderColumnTextDivWidth = headerColumnCell.clientWidth - sortPlaceHolderDivWidth;
           
           // we only want to constrain the width.
           if (headerColumnTextDivWidth > newHeaderColumnTextDivWidth + 1)
           {
             headerColumnTextDiv[0].style[oj.TableDomUtils.CSS_PROP._WIDTH] = newHeaderColumnTextDivWidth + oj.TableDomUtils.CSS_VAL._PX;
           }
         }
       }
     }
   }
 }
};

/**
 * @private
 */
oj.TableDomUtils.CSS_CLASSES =
  {
    _CHECKBOX_ACC_SELECT_COLUMN_CLASS: 'oj-table-checkbox-acc-select-column',
    _CHECKBOX_ACC_SELECT_ROW_CLASS: 'oj-table-checkbox-acc-select-row',
    _TABLE_CONTAINER_CLASS: 'oj-table-container',
    _TABLE_SCROLLER_CLASS: 'oj-table-scroller',
    _TABLE_CLASS: 'oj-table',
    _TABLE_COMPACT_CLASS: 'oj-table-grid-display',
    _TABLE_EDIT_CLASS: 'oj-table-editable',
    _TABLE_SCROLL_CLASS: 'oj-table-scroll',
    _TABLE_SCROLL_VERTICAL_CLASS: 'oj-table-scroll-vertical',
    _TABLE_SCROLL_HORIZONTAL_CLASS: 'oj-table-scroll-horizontal',
    _TABLE_ELEMENT_CLASS: 'oj-table-element',
    _TABLE_FOOTER_CLASS: 'oj-table-footer',
    _TABLE_FOOTER_ROW_CLASS: 'oj-table-footer-row',
    _TABLE_HEADER_CLASS: 'oj-table-header',
    _TABLE_HEADER_ROW_CLASS: 'oj-table-header-row',
    _TABLE_BOTTOM_SLOT_CLASS: 'oj-table-slot-bottom',
    _COLUMN_HEADER_CELL_CLASS: 'oj-table-column-header-cell',
    _COLUMN_HEADER_DROP_EMPTY_CELL_CLASS: 'oj-table-column-header-drop-empty-cell',
    _COLUMN_HEADER_ACC_SELECT_COLUMN_CLASS: 'oj-table-column-header-acc-select-column',
    _COLUMN_HEADER_ACC_SELECT_ROW_CLASS: 'oj-table-column-header-acc-select-row',
    _COLUMN_HEADER_ACC_SELECT_ROW_TEXT_CLASS: 'oj-table-column-header-acc-select-row-text',
    _COLUMN_HEADER_CLASS: 'oj-table-column-header',
    _COLUMN_HEADER_TEXT_CLASS: 'oj-table-column-header-text',
    _COLUMN_HEADER_ASC_CLASS: 'oj-table-column-header-asc',
    _COLUMN_HEADER_DSC_CLASS: 'oj-table-column-header-dsc',
    _COLUMN_HEADER_SORT_PACEHOLDER_CLASS: 'oj-table-column-header-sort-placeholder',
    _COLUMN_HEADER_ACC_ASC_LINK_CLASS: 'oj-table-column-header-acc-asc-link',
    _COLUMN_HEADER_ACC_DSC_LINK_CLASS: 'oj-table-column-header-acc-dsc-link',
    _COLUMN_HEADER_ASC_LINK_CLASS: 'oj-table-column-header-asc-link',
    _COLUMN_HEADER_DSC_LINK_CLASS: 'oj-table-column-header-dsc-link',
    _COLUMN_HEADER_ASC_ICON_CLASS: 'oj-table-column-header-asc-icon',
    _COLUMN_HEADER_DSC_ICON_CLASS: 'oj-table-column-header-dsc-icon',
    _COLUMN_HEADER_DRAG_INDICATOR_BEFORE_CLASS: 'oj-table-column-header-drag-indicator-before',
    _COLUMN_HEADER_DRAG_INDICATOR_AFTER_CLASS: 'oj-table-column-header-drag-indicator-after',
    _COLUMN_HEADER_DRAG_IMAGE: 'oj-table-column-header-cell-drag-image',
    _COLUMN_HEADER_RESIZE_INDICATOR_CLASS: 'oj-table-column-header-resize-indicator',
    _TABLE_BODY_CLASS: 'oj-table-body',
    _TABLE_DATA_ROW_CLASS: 'oj-table-body-row',
    _TABLE_DATA_ROW_DRAG_INDICATOR_BEFORE_CLASS: 'oj-table-body-row-drag-indicator-before',
    _TABLE_DATA_ROW_DRAG_INDICATOR_AFTER_CLASS: 'oj-table-body-row-drag-indicator-after',
    _TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOP_CLASS: 'oj-table-body-row-touch-selection-affordance-top',
    _TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_BOTTOM_CLASS: 'oj-table-body-row-touch-selection-affordance-bottom',
    _TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOP_ICON_CLASS: 'oj-table-body-row-touch-selection-affordance-top-icon',
    _TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_BOTTOM_ICON_CLASS: 'oj-table-body-row-touch-selection-affordance-bottom-icon',
    _TABLE_DATA_ROW_TOUCH_SELECTIOM_AFFORDANCE_TOUCH_AREA_CLASS: 'oj-table-body-row-touch-selection-affordance-touch-area',
    _TABLE_DATA_ROW_EDIT_CLASS: 'oj-table-body-row-edit',
    _TABLE_DATA_CURRENT_ROW_CLASS: 'oj-table-body-current-row',
    _TABLE_DATA_CELL_CLASS: 'oj-table-data-cell',
    _TABLE_DATA_CELL_ACC_SELECT_CLASS: 'oj-table-data-cell-acc-select',
    _TABLE_DATA_CELL_EDIT_CLASS: 'oj-table-data-cell-edit',
    _TABLE_DATE_CELL_FORM_CONTROL_CLASS: 'oj-form-control-inherit',
    _TABLE_VGRID_LINES_CLASS: 'oj-table-vgrid-lines',
    _TABLE_HGRID_LINES_CLASS: 'oj-table-hgrid-lines',
    _TABLE_FOOTER_CELL_CLASS: 'oj-table-footer-cell',
    _TABLE_FOOTER_DROP_EMPTY_CELL_CLASS: 'oj-table-footer-drop-empty-cell',
    _TABLE_INLINE_MESSAGE_CLASS: 'oj-table-inline-message',
    _TABLE_STATUS_ACC_NOTIFICATION_CLASS: 'oj-table-status-acc-notification',
    _TABLE_STATUS_MESSAGE_CLASS: 'oj-table-status-message',
    _TABLE_STATUS_MESSAGE_TEXT_CLASS: 'oj-table-status-message-text',
    _TABLE_LOADING_ICON_CLASS: 'oj-table-loading-icon',
    _TABLE_BODY_MESSAGE_CLASS: 'oj-table-body-message',
    _TABLE_BODY_MESSAGE_ROW_CLASS: 'oj-table-body-message-row',
    _ICON_CLASS: 'oj-icon',
    _WIDGET_ICON_CLASS: 'oj-component-icon',
    _HIDDEN_CONTENT_ACC_CLASS: 'oj-helper-hidden-accessible'
  };

/**
 * @private
 */
oj.TableDomUtils.CSS_PROP =
  {
    _DISPLAY: 'display',
    _VISIBILITY: 'visibility',
    _POSITION: 'position',
    _HEIGHT: 'height',
    _WIDTH: 'width',
    _TOP: 'top',
    _BOTTOM: 'bottom',
    _LEFT: 'left',
    _RIGHT: 'right',
    _PADDING_TOP: 'padding-top',
    _PADDING_LEFT: 'padding-left',
    _PADDING_RIGHT: 'padding-right',
    _OVERFLOW: 'overflow',
    _OVERFLOW_X: 'overflow-x',
    _OVERFLOW_Y: 'overflow-y',
    _MIN_WIDTH: 'min-width',
    _MAX_WIDTH: 'max-width',
    _MIN_HEIGHT: 'min-height',
    _FLOAT: 'float',
    _BORDER_TOP: 'border-top',
    _BORDER_BOTTOM_WIDTH: 'border-bottom-width',
    _BORDER_LEFT_WIDTH: 'border-left-width',
    _MARGIN_BOTTOM: 'margin-bottom',
    _VERTICAL_ALIGN: 'vertical-align',
    _CURSOR: 'cursor',
    _ZINDEX: 'z-index',
    _BACKGROUND_COLOR: 'background-color',
    _BOX_SIZING: 'box-sizing'
  };
  
/**
 * @private
 */
oj.TableDomUtils.CSS_VAL =
  {
    _NONE: 'none',
    _BLOCK: 'block',
    _INLINE_BLOCK: 'inline-block',
    _RELATIVE: 'relative',
    _ABSOLUTE: 'absolute',
    _INLINE: 'inline',
    _AUTO: 'auto',
    _HIDDEN: 'hidden',
    _VISIBLE: 'visible',
    _LEFT: 'left',
    _PX: 'px',
    _MIDDLE: 'middle',
    _MOVE: 'move',
    _FIXED: 'fixed',
    _TRANSPARENT: 'transparent',
    _BORDER_BOX: 'border-box'
  };
  
/**
 * @private
 */
oj.TableDomUtils.DOM_ATTR =
  {
    _STYLE: 'style',
    _TABINDEX: 'tabindex',
    _TYPE: 'type',
    _ID: 'id',
    _TITLE: 'title',
    _HREF: 'href',
    _HEADERS: 'headers',
    _COLSPAN: 'colspan',
    _ROLE: 'role',
    _ARIA_LABEL: 'aria-label',
    _ARIA_HIDDEN: 'aria-hidden'
  };
  
/**
 * @private
 */
oj.TableDomUtils.DOM_ELEMENT =
  {
    _DIV: 'div',
    _A: 'a',
    _TR: 'tr',
    _TD: 'td',
    _TH: 'th',
    _TABLE: 'table',
    _TBODY: 'tbody',
    _THEAD: 'thead',
    _TFOOT: 'tfoot',
    _INPUT: 'input',
    _UL: 'ul',
    _SPAN: 'span',
    _BUTTON: 'button',
    _LABEL: 'label'
  };
  

/**
 * @private
 */
oj.TableDomUtils.MARKER_STYLE_CLASSES =
  {
    _WIDGET: 'oj-component',
    _ACTIVE: 'oj-active',
    _CLICKABLE_ICON: 'oj-clickable-icon-nocontext',
    _DISABLED: 'oj-disabled',
    _ENABLED: 'oj-enabled',
    _FOCUS: 'oj-focus',
    _FOCUS_HIGHLIGHT: 'oj-focus-highlight',
    _HOVER: 'oj-hover',
    _SELECTED: 'oj-selected',
    _WARNING: 'oj-warning',
    _DRAGGABLE: 'oj-draggable',
    _DRAG: 'oj-drag'
  };
  
/**
 * @private
 * @type {string}
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils._COLUMN_HEADER_ROW_SELECT_ID =   '_hdrColRowSel';
/**
 * @private
 * @type {string}
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils._OPTION_AUTO = 'auto';
/**
 * @private
 * @type {string}
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils._OPTION_ENABLED = 'enabled';
/**
 * @private
 * @type {string}
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils._OPTION_DISABLED = 'disabled';
/**
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils._OPTION_SELECTION_MODES =
  {
    _SINGLE: 'single',
    _MULTIPLE: 'multiple'
  };
/**
 * @private
 * @memberof oj.TableDomUtils
 */
oj.TableDomUtils._OPTION_DISPLAY =
  {
    _LIST: 'list',
    _GRID: 'grid'
  };

/**
 * @preserve Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 */
oj.TableRendererUtils = {};

/**
 * Column Header Renderer
 * @param {Object} component component instance
 * @param {Object} context renderer context
 * @param {Object|null} options renderer options
 * @param {function(Object)|null} delegateRenderer delegate renderer
 * @memberof oj.TableRendererUtils
 */
oj.TableRendererUtils.columnHeaderDefaultRenderer = function(component, context, options, delegateRenderer)
{
  var parentElement = context['headerContext']['parentElement'];
  var headerColumnDiv = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  headerColumnDiv.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CLASS);
  $(parentElement).empty();
  parentElement.appendChild(headerColumnDiv); //@HTMLUpdateOK
  
  // call the delegateRenderer
  var headerContentDiv = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  headerContentDiv.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_TEXT_CLASS);
  headerColumnDiv.insertBefore(headerContentDiv, headerColumnDiv.firstChild); //@HTMLUpdateOK
  
  if (delegateRenderer != null)
  {
    delegateRenderer($(headerContentDiv));
  }
  else
  {
    this.columnHeaderDefaultTextRenderer(component, headerContentDiv, context);
  }
};

/**
 * Column Header with Sort Icons Renderer
 * @param {Object} component component instance
 * @param {Object} context renderer context
 * @param {Object|null} options renderer options
 * @param {function(Object)|null} delegateRenderer delegate renderer
 * @memberof oj.TableRendererUtils
 */
oj.TableRendererUtils.columnHeaderSortableIconRenderer = function(component, context, options, delegateRenderer)
{
  var columnIdx = context['columnIndex'];
  var column = component._getColumnDefs()[columnIdx];
  var parentElement = context['headerContext']['parentElement'];
  var headerColumnDiv = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  headerColumnDiv.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_CLASS);
  $(parentElement).empty();
  parentElement.appendChild(headerColumnDiv); //@HTMLUpdateOK
  
  if (component._GetReadingDirection() === "rtl")
  {
    headerColumnDiv.style[oj.TableDomUtils.CSS_PROP._PADDING_LEFT] = '0px';
  }
  else
  {
    headerColumnDiv.style[oj.TableDomUtils.CSS_PROP._PADDING_RIGHT] = '0px';
  }
  
  var headerColumnAscDiv = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  headerColumnAscDiv.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_CLASS);
  headerColumnDiv.appendChild(headerColumnAscDiv); //@HTMLUpdateOK
  var headerColumnAscLink = document.createElement(oj.TableDomUtils.DOM_ELEMENT._A);
  headerColumnAscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._ARIA_HIDDEN, 'true');
  headerColumnAscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_LINK_CLASS);
  headerColumnAscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._WIDGET_ICON_CLASS);
  headerColumnAscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ASC_ICON_CLASS);
  headerColumnAscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
  headerColumnAscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._CLICKABLE_ICON);
  component._AddHoverable($(headerColumnAscLink));
  headerColumnAscDiv.appendChild(headerColumnAscLink); //@HTMLUpdateOK
  var headerColumnAscWidth;
  var headerColumnAscHeight;
  if (oj.TableRendererUtils._headerColumnAscWidth !== undefined)
  {
    headerColumnAscWidth = oj.TableRendererUtils._headerColumnAscWidth;
    headerColumnAscHeight = oj.TableRendererUtils._headerColumnAscHeight;
  }
  else
  {
    // We only need to get this once per page since all asc div should be the same size
    headerColumnAscWidth = $(headerColumnAscDiv).width();
    headerColumnAscHeight = $(headerColumnAscDiv).height();
    
    // Only save the dimensions if they are > 0 in case the table is hidden
    if (headerColumnAscWidth > 0 && headerColumnAscHeight > 0)
    {
      oj.TableRendererUtils._headerColumnAscWidth = headerColumnAscWidth;
      oj.TableRendererUtils._headerColumnAscHeight = headerColumnAscHeight;
    }
  }
    
  // separate link for acc
  var headerColumnAccAscLink = document.createElement(oj.TableDomUtils.DOM_ELEMENT._A);
  headerColumnAccAscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX, '0');
  headerColumnAccAscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._HREF, '#');
  headerColumnAccAscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ACC_ASC_LINK_CLASS);
  headerColumnAccAscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  headerColumnAccAscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._ARIA_LABEL, component.getTranslatedString('labelSortAsc') + ' ' + column.headerText);
  headerColumnAscDiv.appendChild(headerColumnAccAscLink); //@HTMLUpdateOK
  
  var headerColumnSortPlaceholderDiv = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  headerColumnSortPlaceholderDiv.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_SORT_PACEHOLDER_CLASS);
  headerColumnSortPlaceholderDiv.style[oj.TableDomUtils.CSS_PROP._DISPLAY] = oj.TableDomUtils.CSS_VAL._INLINE_BLOCK;
  headerColumnSortPlaceholderDiv.style[oj.TableDomUtils.CSS_PROP._VERTICAL_ALIGN] = oj.TableDomUtils.CSS_VAL._MIDDLE;
  headerColumnSortPlaceholderDiv.style[oj.TableDomUtils.CSS_PROP._WIDTH] = headerColumnAscWidth + oj.TableDomUtils.CSS_VAL._PX;
  headerColumnSortPlaceholderDiv.style[oj.TableDomUtils.CSS_PROP._HEIGHT] = headerColumnAscHeight + oj.TableDomUtils.CSS_VAL._PX;
  headerColumnDiv.appendChild(headerColumnSortPlaceholderDiv); //@HTMLUpdateOK

  //sort descending link
  var headerColumnDscDiv = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  headerColumnDscDiv.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_CLASS);
  // descending sort is initially not visible
  headerColumnDscDiv.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
  headerColumnDiv.appendChild(headerColumnDscDiv); //@HTMLUpdateOK

  var headerColumnDscLink = document.createElement(oj.TableDomUtils.DOM_ELEMENT._A);
  headerColumnDscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._ARIA_HIDDEN, 'true');
  headerColumnDscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_LINK_CLASS);
  headerColumnDscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._WIDGET_ICON_CLASS);
  headerColumnDscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_DSC_ICON_CLASS);
  headerColumnDscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._DISABLED);
  headerColumnDscLink.classList.add(oj.TableDomUtils.MARKER_STYLE_CLASSES._CLICKABLE_ICON);
  component._AddHoverable($(headerColumnDscLink));
  headerColumnDscDiv.appendChild(headerColumnDscLink); //@HTMLUpdateOK
    
  // separate link for acc
  var headerColumnAccDscLink = document.createElement(oj.TableDomUtils.DOM_ELEMENT._A);
  headerColumnAccDscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._TABINDEX, '0');
  headerColumnAccDscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._HREF, '#');
  headerColumnAccDscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_ACC_DSC_LINK_CLASS);
  headerColumnAccDscLink.classList.add(oj.TableDomUtils.CSS_CLASSES._HIDDEN_CONTENT_ACC_CLASS);
  headerColumnAccDscLink.setAttribute(oj.TableDomUtils.DOM_ATTR._ARIA_LABEL, component.getTranslatedString('labelSortDsc') + ' ' + column.headerText);
  headerColumnDscDiv.appendChild(headerColumnAccDscLink); //@HTMLUpdateOK
  
  // call the delegateRenderer
  var headerContentDiv = document.createElement(oj.TableDomUtils.DOM_ELEMENT._DIV);
  headerContentDiv.classList.add(oj.TableDomUtils.CSS_CLASSES._COLUMN_HEADER_TEXT_CLASS);
  headerColumnDiv.insertBefore(headerContentDiv, headerColumnDiv.firstChild); //@HTMLUpdateOK
  
  if (delegateRenderer != null)
  {
    delegateRenderer($(headerContentDiv));
  }
  else
  {
    this.columnHeaderDefaultTextRenderer(component, headerContentDiv, context);
  }
};

/**
 * Default column header context text renderer
 * @param {Object} component component instance
 * @param {Element} headerContentDiv header content div
 * @param {Object} context context object
 * @memberof oj.TableRendererUtils
 */
oj.TableRendererUtils.columnHeaderDefaultTextRenderer = function(component, headerContentDiv, context)
{
  var columnIdx = context['columnIndex'];
  var column = component._getColumnDefs()[columnIdx];
  var textValue = column.headerText == null ? '' : column.headerText;
  headerContentDiv.appendChild(document.createTextNode(textValue)); //@HTMLUpdateOK
};

/**
 * Default table body row renderer
 * @param {Object} component component instance
 * @param {number} rowIdx  row index
 * @param {Object} row row
 * @param {Object} context context object
 * @memberof oj.TableRendererUtils
 */
oj.TableRendererUtils.tableBodyRowDefaultRenderer = function(component, rowIdx, row, context)
{
  var tableBodyRow = context['rowContext']['parentElement'];
  var rowHashCode = component._getTableDomUtils().hashCode(row['key']);
  var columns = component._getColumnDefs();
  component._getTableDomUtils().setTableBodyRowAttributes(row, tableBodyRow);
  var j, columnsCount = columns.length;
  for (j = 0; j < columnsCount; j++)
  {
    // set the cells in the inserted row with values from the row
    this.tableBodyCellDefaultRenderer(component, rowIdx, j, row, rowHashCode, context);
  }
};

/**
 * Default table body cell renderer
 * @param {Object} component component instance
 * @param {number} rowIdx  row index
 * @param {number} columnIdx  column index
 * @param {Object} row row
 * @param {number} rowHashCode  row hash code
 * @param {Object} context context object
 * @memberof oj.TableRendererUtils
 */
oj.TableRendererUtils.tableBodyCellDefaultRenderer = function(component, rowIdx, columnIdx, row, rowHashCode, context)
{
  var tableBodyRow = context['rowContext']['parentElement'];
  var columns = component._getColumnDefs();
  var column = columns[columnIdx];
  
  var tableBodyCell = component._getTableDomUtils().createTableBodyCell(rowIdx, columnIdx);
  component._getTableDomUtils().styleTableBodyCell(columnIdx, tableBodyCell, true);
  component._getTableDomUtils().insertTableBodyCell(rowIdx, row['key'], rowHashCode, columnIdx, tableBodyCell, tableBodyRow, true);
  
  var data = null;

  if (column.field != null)
  {
    data = oj.TableRendererUtils._getObjectPath(row['data'], column.field);
  }
  
  var cellRenderer = component._getColumnRenderer(columnIdx, 'cell');

  if (cellRenderer)
  {
    var cellContext = this.getRendererContextObject(component, tableBodyCell, {'row': row});
    // Copy additional properties to top-level context to work with custom element
    var rendererContext = {'cellContext': cellContext,
                           'columnIndex': columnIdx,
                           'data': data,
                           'row': $.extend({}, row['data']),
                           'componentElement': cellContext['componentElement'],
                           'parentElement': cellContext['parentElement']};
    var cellColumnContent = cellRenderer(rendererContext);

    if (cellColumnContent != null)
    {
      // if the renderer returned a value then we set it as the content
      // for the cell. Use jquery append() for this as a convenience because 
      // cellColumnContent could be a Node element or arbitrary content and 
      // we don't want to write code to convert everything to Node type and call 
      // appendChild.
      $(tableBodyCell).append(cellColumnContent); //@HTMLUpdateOK
    }
    else
    {
      // if the renderer didn't return a value then the existing
      // cell was manipulated. So get it and set the required
      // attributes just in case it was replaced or the attributes
      // got removed
      tableBodyCell = $(tableBodyRow).children(':not(' + '.' + oj.TableDomUtils.CSS_CLASSES._TABLE_DATA_CELL_ACC_SELECT_CLASS + ')')[columnIdx];
      component._getTableDomUtils().setTableBodyCellAttributes(rowIdx, row['key'], rowHashCode, columnIdx, tableBodyCell);
      component._getTableDomUtils().styleTableBodyCell(columnIdx, tableBodyCell, false);
    }
  }
  else
  {
    var cellSlotTemplate = component._getSlotTemplate(column['template']);
    
    if (cellSlotTemplate)
    {
      var componentElement = component._getRootElement();
      var templateEngine = component._getTemplateEngine();
      if (templateEngine != null)
      {
        var slotContext = oj.TableRendererUtils.getCellSlotTemplateContextObject(component, row, data);
        var cellContent = templateEngine.execute(componentElement, cellSlotTemplate, slotContext);
        if (!(cellContent instanceof Array)) 
        {
          cellContent = [cellContent];
        }
        cellContent.map(function(content)
        {
          tableBodyCell.appendChild(content);
        });
      }
    }
    else
    {
      var textValue = data == null ? '' : data;
      tableBodyCell.appendChild(document.createTextNode(textValue)); //@HTMLUpdateOK
    }
  }
};

/**
  * Get the context object to pass into the renderer
  * @param {Object} component component
  * @param {Object} parentElement element
  * @param {Object} options options
  * @memberof oj.TableRendererUtils
  */
oj.TableRendererUtils.getRendererContextObject = function(component, parentElement, options)
{
  var context = {};
  context['component'] = oj.Components.__GetWidgetConstructor(component.element, 'ojTable');
  var dataSource = component['options']['data'];
  // unwrap the datasource if we have a PagingTableDataSource
  if (oj.PagingTableDataSource && dataSource instanceof oj.PagingTableDataSource)
  {
    dataSource = dataSource.getWrappedDataSource();
  }
  context['datasource'] = dataSource;
  context['parentElement'] = parentElement;
  
  if (options['row'] != null)
  {
    var row = options['row'];
    var rowKey = row['key'];
    context['status'] = this.getRendererStatusObject(component, row);
    
    if (component._hasEditableRow())
    {
      // Check using the editable row key
      var editableRowKey = component._getEditableRowKey();
      
      // only set to edit mode for the editable row
      if (oj.Object.compareValues(rowKey, editableRowKey))
      {
        context['mode'] = 'edit';
      }
      else
      {
        context['mode'] = 'navigation';
      }
    }
    else
    {
      context['mode'] = 'navigation';
    }

    this._copyMetadata(context, row, dataSource);
  }

  // Fix up context to work with custom element
  return component._FixRendererContext(context);
};

/**
 * Get the status object to pass into the renderer
 * @param {Object} component component
 * @param {Object} row row instance
 * @return {Object} status object
 * @memberof oj.TableRendererUtils
 */
oj.TableRendererUtils.getRendererStatusObject = function(component, row)
{
  return {'rowIndex': row['index'],
    'rowKey': row['key'],
    'currentRow': $.extend({}, component._getCurrentRow())};
};

/**
  * Get the context object for inline cell templates
  * @param {Object} component component
  * @param {Object} row row
  * @param {Object} data data
  * @memberof oj.TableRendererUtils
  */
oj.TableRendererUtils.getCellSlotTemplateContextObject = function(component, row, data)
{
  var as = component['options']['as'];
  var currentName = (as != null && as.length > 0) ? as : '$current';
  var slotContext = this.getSlotTemplateContextObject(component);
  slotContext[currentName][component._CONST_DATA] = data;
  slotContext[currentName]['row'] = row['data'];
  slotContext[currentName][component._CONST_INDEX] = row[component._CONST_INDEX];
  slotContext[currentName][component._CONST_KEY] = row[component._CONST_KEY];
  
  return slotContext;
};

/**
  * Get the context object for inline header templates
  * @param {Object} component component
  * @param {Object} data data
  * @memberof oj.TableRendererUtils
  */
oj.TableRendererUtils.getHeaderSlotTemplateContextObject = function(component, data)
{
  var as = component['options']['as'];
  var currentName = (as != null && as.length > 0) ? as : '$current';
  var slotContext = this.getSlotTemplateContextObject(component);
  slotContext[currentName][component._CONST_DATA] = data;
  
  return slotContext;
};

/**
  * Get the context object for inline templates
  * @param {Object} component component
  * @memberof oj.TableRendererUtils
  */
oj.TableRendererUtils.getSlotTemplateContextObject = function(component)
{
  var componentElement = component._getRootElement();
  var as = component['options']['as'];
  var currentName = (as != null && as.length > 0) ? as : '$current';
  var slotContext = {};
  slotContext[currentName] = {'componentElement': componentElement};
  return slotContext;
};

oj.TableRendererUtils._copyMetadata = function(context, row, dataSource)
{
  if (oj.FlattenedTreeTableDataSource && dataSource instanceof oj.FlattenedTreeTableDataSource)
    {
      var metadataContext = row['metadata'];
      if (metadataContext == null)
      {
        metadataContext = dataSource._getMetadata(row['index']);
      }

      var i;
      for (i in metadataContext)
      {
        if (metadataContext.hasOwnProperty(i))
        {
          context[i] = metadataContext[i];
        }
      } 
    }
};

oj.TableRendererUtils._getObjectPath = function(obj, path)
{
  if (obj != null &&
    (typeof path === 'string' || path instanceof String) && path.indexOf('.') != -1)
  {
    var currentObj = obj; 
    var foundPath = false;
    var pathArray = path.split('.');
    pathArray.map(function(pathSegment)
    { 
      if (currentObj != null && currentObj[pathSegment] !== undefined)
      {
        currentObj = currentObj[pathSegment];
        foundPath = true;
      }
    }); 
    if (foundPath)
    {
      return currentObj;
    }
  }
  else if (obj == null)
  {
    return null;
  }
  return obj[path];
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @export
 * @class oj.TableResizeUtils
 * @classdesc Resize Utils for ojTable
 * @param {Object} component ojTable instance
 * @constructor
 */
oj.TableResizeUtils = function(component)
{
  this.component = component;
  this.options = component['options'];
  this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.TableResizeUtils, oj.Object, "oj.TableResizeUtils");

/**
 * Initializes the instance.
 * @export
 * @memberof oj.TableResizeUtils
 */
oj.TableResizeUtils.prototype.Init = function()
{
  oj.TableResizeUtils.superclass.Init.call(this);
};

/**
  * Set resize cursor
  * @param {Event} event Event
  * @memberof oj.TableResizeUtils
  */
oj.TableResizeUtils.prototype.setResizeCursor = function(event)
{
  var eventTarget = this.component._getEventTargetElement(event);
  var columnIdx = this.component._getTableDomUtils().getElementColumnIdx(eventTarget);
  
  if (columnIdx == null)
  {
    return false;
  }
  
  var column = this.component._getColumnDefs()[columnIdx];
  
  if (column['resizable'] == this.component._OPTION_DISABLED && 
    this._resizeStartColumnIdx !== columnIdx - 1 &&
    this._resizeStartColumnIdx !== columnIdx + 1)
  {
    return false;
  }
  
  if (this._resizeStartColumnIdx == null)
  {
    if (this._isHeaderColumnResizeStart(event) !== null)
    {
      eventTarget.style.cursor = 'col-resize';
      return true;
    }
    eventTarget.style.cursor = '';
    return false;
  }
  else
  {
    var headerColumns = this.component._getTableDomUtils().getTableHeaderColumns();
    
    if (!headerColumns)
    {
      return false;
    }
    var headerColumnCount = headerColumns.length;
    
    if (columnIdx == this._resizeStartColumnIdx ||
	      this._resizeColumnStart && columnIdx == this._resizeStartColumnIdx - 1 ||
	      !this._resizeColumnStart && columnIdx == this._resizeStartColumnIdx + 1)
  	{
      // cannot resize the leftmost border of the first column or the rightmost border of the last column
      if (this._resizeStartColumnIdx == 0 && this._resizeColumnStart)
      {
        return false;
      }
      else if (this._resizeStartColumnIdx == headerColumnCount - 1 && !this._resizeColumnStart)
      {
        return false;
      }
      // move the indicator
      var tableHeaderColumnResizeIndicator = this.component._getTableDomUtils().getTableHeaderColumnResizeIndicator();

      if (tableHeaderColumnResizeIndicator != null)
      {
        if (this.component._GetReadingDirection() === "rtl")
        {
          var columnRect = event.target.getBoundingClientRect();
          if (this._resizeColumnStart)
          {
            tableHeaderColumnResizeIndicator.style.left = this._getPageX(event) - columnRect.left + columnRect.width + 'px';
          }
          else
          {
            tableHeaderColumnResizeIndicator.style.left = this._getPageX(event) - columnRect.left + 'px';
          }
        }
        else
        {
          tableHeaderColumnResizeIndicator.style.left = this._getPageX(event) + 'px';
        }
        return true;
      }
    }
  }
  return false;
};

/**
  * Handle header column resize start
  * @param {Event} event Event
  * @return {boolean} Return whether column resize started
  * @memberof oj.TableResizeUtils
  */
oj.TableResizeUtils.prototype.handleHeaderColumnResizeStart = function(event)
{
  var eventTarget = this.component._getEventTargetElement(event);
  var columnIdx = this.component._getTableDomUtils().getElementColumnIdx(eventTarget);
  
  if (columnIdx != null)
  {
    var column = this.component._getColumnDefs()[columnIdx];
    this._resizeColumnStart = this._isHeaderColumnResizeStart(event);
    
    if (column['resizable'] == this.component._OPTION_ENABLED &&
      this._resizeColumnStart !== null)
    {
      this._resizeStartColumnIdx = columnIdx;
      this._resizeStartPageX = this._getPageX(event);
      this._setTableHeaderColumnsResizeStyling();
      this._setTableHeaderColumnResizeIndicator(columnIdx);
      return true;
    }
  }
  this._resizeStartColumnIdx = null;
  this._resizeStartPageX = null;
  return false;
};

/**
  * Handle header column resize end
  * @param {Event} event Event
  * @return {boolean} Return whether column resize ended
  * @memberof oj.TableResizeUtils
  */
oj.TableResizeUtils.prototype.handleHeaderColumnResizeEnd = function(event)
{
  var eventTarget = this.component._getEventTargetElement(event);
  var columnIdx = this.component._getTableDomUtils().getElementColumnIdx(eventTarget);

  // only resize if we end the resize on the same column or adjacent columns
  if ((columnIdx !== null && columnIdx === this._resizeStartColumnIdx) ||
      (this._resizeColumnStart && columnIdx === this._resizeStartColumnIdx - 1) ||
      (!this._resizeColumnStart && columnIdx === this._resizeStartColumnIdx + 1))
  {
    var headerColumn = this.component._getTableDomUtils().getTableHeaderColumn(this._resizeStartColumnIdx);
    if (headerColumn != null)
    {
      var headerColumnRect = headerColumn.getBoundingClientRect();
      var headerColumnWidth = headerColumnRect['width'];
      var updatedWidth, widthChange;

      if (this._resizeColumnStart)
      {
        widthChange = this._resizeStartPageX - this._getPageX(event);
      }
      else
      {
        widthChange = this._getPageX(event) - this._resizeStartPageX;
      }
      if (Math.abs(widthChange) > 2)
      {
        updatedWidth = headerColumnWidth + widthChange;
        var clonedColumnsOption = [];
        var i, columnsCount = this.options['columns'].length;
        for (i = 0; i < columnsCount; i++) {
          clonedColumnsOption[i] = $.extend({}, {}, this.options['columns'][i]);
        }
        clonedColumnsOption[this._resizeStartColumnIdx]['width'] = updatedWidth;
        var headerColumnAdjacent, headerColumnAdjacentIdx;
        if (this._resizeColumnStart)
        {
          // resizing from the left should make the column in the left bigger or smaller
          headerColumnAdjacentIdx = this._resizeStartColumnIdx - 1;
          headerColumnAdjacent = this.component._getTableDomUtils().getTableHeaderColumn(headerColumnAdjacentIdx);
          if (headerColumnAdjacent)
          {
            var headerColumnAdjacentRect = headerColumnAdjacent.getBoundingClientRect();
            var headerColumnAdjacentWidth = headerColumnAdjacentRect['width'];
            clonedColumnsOption[headerColumnAdjacentIdx]['width'] = headerColumnAdjacentWidth - widthChange;
          }
        }
        this.component.option('columns', clonedColumnsOption, {'_context': {writeback: true, internalSet: true}});
        this.component._clearCachedMetadata();
        this.component._refresh();
        this.clearTableHeaderColumnsResize();
        return true;
      }
    }
  }
  this.clearTableHeaderColumnsResize();
  return false;
};

/**
  * Clear any column resize
  * @memberof oj.TableResizeUtils
  */
oj.TableResizeUtils.prototype.clearTableHeaderColumnsResize = function()
{
  this._resizeStartColumnIdx = null;
  this._resizeColumnStart = null;
  this._resizeStartPageX = null;
  this._clearTableHeaderColumnsResizeStyling();
  this.component._getTableDomUtils().removeTableHeaderColumnResizeIndicator();
};

oj.TableResizeUtils.prototype._getPageX = function(event)
{
  if (event.pageX !== undefined)
  {
    // MouseEvent has pageX on event itself
    return event.pageX;
  }
  else if (event.changedTouches !== undefined)
  {
    // TouchEvent has pageX on changedTouches, targetTouches, and touches.
    // For one-finger drag, they contain the same value on all touch events
    // except for touchend, in which only changedTouches has the point at
    // which the finger leaves the touch surface.
    return event.changedTouches[0].pageX;
  }
  
  // We shouldn't get here unless event is neither MouseEvent nor TouchEvent
  return 0;
};

oj.TableResizeUtils.prototype._isHeaderColumnResizeStart = function(event)
{
  var resizeColumnStart = null;
  var columnRect = event.target.getBoundingClientRect();
  var distFromLeft = Math.abs(this._getPageX(event) - columnRect.left);
  var distFromRight = Math.abs(this._getPageX(event) - columnRect.right);
  if (distFromLeft <= oj.TableResizeUtils.RESIZE_OFFSET) 
  {
    if (this.component._GetReadingDirection() === "rtl") 
    {
      resizeColumnStart = false;
    } 
    else 
    {
      resizeColumnStart = true;
    }
  } 
  else if (distFromRight <= oj.TableResizeUtils.RESIZE_OFFSET) 
  {
    if (this.component._GetReadingDirection() === "rtl") 
    {
      resizeColumnStart = true;
    } 
    else 
    {
      resizeColumnStart = false;
    }
  }
  return resizeColumnStart;
};

oj.TableResizeUtils.prototype._setTableHeaderColumnResizeIndicator = function(columnIdx)
{
  var tableHeaderColumnResizeIndicator = this.component._getTableDomUtils().getTableHeaderColumnResizeIndicator();
  
  if (tableHeaderColumnResizeIndicator == null)
  {
    tableHeaderColumnResizeIndicator = this.component._getTableDomUtils().createTableHeaderColumnResizeIndicator();
  }
  var table = this.component._getTableDomUtils().getTable();
  var tableRect = table.getBoundingClientRect();
  var headerColumn = this.component._getTableDomUtils().getTableHeaderColumn(columnIdx);
  var headerColumnRect = headerColumn.getBoundingClientRect();
  var tableHeaderColumnResizeIndicatorRect = tableHeaderColumnResizeIndicator.getBoundingClientRect();
  tableHeaderColumnResizeIndicator.style.height = tableRect.height + 'px';
  
  if (this._resizeColumnStart)
  {
    if (this.component._GetReadingDirection() === "rtl")
	  {
      tableHeaderColumnResizeIndicator.style.left = headerColumnRect.left - tableHeaderColumnResizeIndicatorRect.left + headerColumnRect.width + 'px';
    }
    else
    {
      tableHeaderColumnResizeIndicator.style.left = -1 * tableHeaderColumnResizeIndicatorRect.left + headerColumnRect.left + 'px';
    }
    tableHeaderColumnResizeIndicator.style.borderLeftWidth = '2px';
    tableHeaderColumnResizeIndicator.style.borderRightWidth = '0';
  }
  else
  {
    if (this.component._GetReadingDirection() === "rtl")
	  {
      tableHeaderColumnResizeIndicator.style.left = headerColumnRect.left - tableHeaderColumnResizeIndicatorRect.left + 'px';
    }
    else
    {
      tableHeaderColumnResizeIndicator.style.left = -1 * tableHeaderColumnResizeIndicatorRect.left + headerColumnRect.left + headerColumnRect.width + 'px';
    }
    tableHeaderColumnResizeIndicator.style.borderRightWidth = '2px';
    tableHeaderColumnResizeIndicator.style.borderLeftWidth = '0';
  }
};

oj.TableResizeUtils.prototype._setTableHeaderColumnsResizeStyling = function()
{
  var table = this.component._getTableDomUtils().getTable();
  table.classList.add(oj.TableResizeUtils.CSS_CLASSES._COLUMN_HEADER_RESIZING_CLASS);
};

oj.TableResizeUtils.prototype._clearTableHeaderColumnsResizeStyling = function()
{
  var headerColumns = this.component._getTableDomUtils().getTableHeaderColumns();
  
  if (headerColumns)
  {
    var i;
    for (i = 0; i < headerColumns.length; i++)
    {
      headerColumns[i].style.cursor = '';
    }
  }
  
  var table = this.component._getTableDomUtils().getTable();
  table.classList.remove(oj.TableResizeUtils.CSS_CLASSES._COLUMN_HEADER_RESIZING_CLASS);
};

oj.TableResizeUtils.RESIZE_OFFSET = 10;

/**
 * @private
 */
oj.TableResizeUtils.CSS_CLASSES =
  {
    _COLUMN_HEADER_RESIZING_CLASS: 'oj-table-column-header-resizing'
  };
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="2">Cell</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Focus on the row. If <code class="prettyprint">selectionMode</code> for rows is enabled, selects the row as well.
 *       If multiple selection is enabled the selection handles will appear. Tapping a different row will deselect the previous selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display context menu</td>
 *     </tr>
 *
 *     <tr>
 *       <td rowspan="2">Column Header</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Focus on the header. If <code class="prettyprint">selectionMode</code> for columns is enabled, selects the column as well.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display context menu</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch section of classdesc, and standalone gesture doc
 * @memberof oj.ojTable
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *   <tr>
 *       <td rowspan="13">Cell</td>
 *       <td><kbd>Tab</kbd></td>
 *       <td>The first Tab into the Table moves focus to the first column header. The second Tab moves focus to the next focusable element outside of the Table.
 *           <br>If focus is on a row and the row is actionable then Tab moves focus to the next focusable element within the row. If focus is already on the last focusable element then focus will wrap to the first focusable element in the row.
 *           <br>If <code class="prettyprint">editMode</code> is rowEdit, please see the section 'Cell in EditableRow'.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Tab</kbd></td>
 *       <td>The first Shift+Tab into the Table moves focus to the first column header. The second Shift+Tab moves focus to the previous focusable element outside of the Table.
 *           <br>If focus is on a row and the row is actionable then Shift+Tab moves focus to the previous focusable element within the row. If focus is already on the first focusable element then focus will wrap to the last focusable element in the row.
 *           <br>If <code class="prettyprint">editMode</code> is rowEdit, please see the section 'Cell in EditableRow'.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus to the next row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+DownArrow</kbd></td>
 *       <td>Select and move focus to the next row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus to the previous row. If at the first row then move to the column header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+UpArrow</kbd></td>
 *       <td>Select and move focus to the previous row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Do nothing.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Do nothing.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Home</kbd></td>
 *       <td>Move focus to first row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>End</kbd></td>
 *       <td>Move focus to last row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Space</kbd></td>
 *       <td>Select row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>If the table <code class="prettyprint">editMode</code> is rowEdit then make the current row editable. 
 *           <br>If the table <code class="prettyprint">editMode</code> is none then toggle the current row to actionable mode if there exists a tabbable element in the row. Once toggled to actionable mode, focus will be moved to be first tabbable element in the row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>F2</kbd></td>
 *       <td>If the table <code class="prettyprint">editMode</code> is none then toggle the current row to actionable mode if there exists a tabbable element in the row. Once toggled to actionable mode, focus will be moved to be first tabbable element in the row.
 *           <br>If the table <code class="prettyprint">editMode</code> is rowEdit then toggle the current row between editable and readonly.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td rowspan="15">Cell in Editable Row</td>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next editable cell or focusable element in the row.
 *           <br>If focus is on the last editable cell or focusable element in the row, make the next row editable and move focus to the first editable cell or focusable element in the next row.
 *           <br>If focus is on the last editable cell or focusable element in the last row, move focus to next focusable element on the page (outside table).
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Tab</kbd></td>
 *       <td>Move focus to previous editable cell or focusable element in the row.
 *           <br>If focus is on the first editable cell or focusable element in the row, make the previous row editable and move focus to the last editable cell or focusable element in the previous row.
 *           <br>If focus is on the first editable cell or focusable element in the first row, move focus to previous focusable element on the page (outside table).
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+DownArrow</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+UpArrow</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Handled by the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Home</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>End</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Space</kbd></td>
 *       <td>Handled in the editable cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Make the next row editable and move focus to the editable cell in current column in the next row.
 *          <br>If last row is editable then make it readonly.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Enter</kbd></td>
 *       <td>Make the previous row editable and move focus to the editable cell in current column in the previous row.
 *          <br>If first row is editable then make it readonly.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>F2</kbd></td>
 *       <td>Toggle the current row between editable and readonly.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Make the current row readonly.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="11">Column Header</td>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Navigate to next focusable element on page (outside table).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Tab</kbd></td>
 *       <td>Navigate to previous focusable element on page (outside table).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus to the first row.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Do nothing.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus to previous column header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+LeftArrow</kbd></td>
 *       <td>Select and move focus to previous column header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus to next column header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+RightArrow</kbd></td>
 *       <td>Select and move focus to next column header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Home</kbd></td>
 *       <td>Move focus to first column header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>End</kbd></td>
 *       <td>Move focus to last column header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Space</kbd></td>
 *       <td>Select column.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * 
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojTable
 */

/**
 * {@ojinclude "name":"ojStylingDocIntro"}
 * 
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
 *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-table-data-cell-no-padding</td>
 *       <td><p>Used to style a table cell so that it has no padding. An app developer would likely use
 *       this in the case of editable tables when an editable cell content does not need the default cell padding.
 *
 *           <p>The class is applied as follows:
 *
 *           <ul>
 *             <li>The class must be applied to the table cell.</li>
 *           </ul>
 *       </td>
 *     </tr>
 *     <tr>
 *       <td>oj-table-data-cell-padding</td>
 *       <td><p>Used to style a table cell so that it has the default padding. An app developer would likely use
 *       this in the case of editable tables when an editable cell content needs to maintain default cell padding.
 *
 *           <p>The class is applied as follows:
 *
 *           <ul>
 *             <li>The class must be applied to the table cell.</li>
 *           </ul>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 * 
 * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
 * @memberof oj.ojTable
 */

/**
* <p>Applications can customize animations triggered by actions in Table by either listening for <code class="prettyprint">animateStart/animateEnd</code>
 *    events or overriding action specific style classes on the animated item. To disable animations for specific table, please call event.preventDefault() from your listener.
 *    To disable animations for all tables, the sass variable values can be modified to specify empty effects. See the documentation of <a href="oj.AnimationUtils.html">oj.AnimationUtils</a> 
 *    class for details.</p>
 *    
 * <p>The following are actions and their corresponding sass variables in which applications can use to customize animation effects.  
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Action</th>
 *       <th>Sass Variable</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>add</kbd></td>
 *       <td>$tableAddAnimation</td>
 *       <td>When a new row is added to the oj.TableDataSource associated with Table.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>remove</kbd></td>
 *       <td>$tableRemoveAnimation</td>
 *       <td>When an existing row is removed from the oj.TableDataSource associated with Table.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>update</kbd></td>
 *       <td>$tableUpdateAnimation</td>
 *       <td>When an existing row is updated in the oj.TableDataSource associated with Table.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment animationDoc - Used in animation section of classdesc
 * @memberof oj.ojTable
 */

/**
 * <p>Table supports a bottom custom element slot where applications can add content such as a paging control.</p>
 *    
 *
 * @ojfragment slotDoc - Used in slot section of classdesc
 * @memberof oj.ojTable
 */
/**
 * <p>Named slot for the Table's bottom panel where applications can add content such as a paging control.</p>
 *
 * @ojslot bottom
 * @memberof oj.ojTable
 *
 * @example <caption>Initialize the Table with the <code class="prettyprint">bottom</code> slot specified:</caption>
 * &lt;oj-table>
 *   &lt;div slot='bottom'>&lt;oj-paging-control>&lt;/oj-paging-control>&lt;/div>
 * &lt;/oj-table>
 */
// SubId Locators **************************************************************

/**
 * <p>Sub-ID for the Table element's cells.</p>
 * To lookup a cell the locator object should have the following:
 * <ul>
 * <li><b>subId</b>: 'oj-table-cell'</li>
 * <li><b>rowIndex</b>: the zero based absolute row index</li>
 * <li><b>columnIndex</b>: the zero based absolute column index</li>
 * </ul>
 *
 * @ojsubid oj-table-cell
 * @memberof oj.ojTable
 *
 * @example <caption>Get the cell at row index 1 and column index 2:</caption>
 * var node = myTable.getNodeBySubId( {'subId': 'oj-table-cell', 'rowIndex': 1, 'columnIndex': 2} );
 */

/**
 * <p>Sub-ID for the Table element's headers.</p>
 *
 *  To lookup a header the locator object should have the following:
 * <ul>
 * <li><b>subId</b>: 'oj-table-header'</li>
 * <li><b>index</b>: the zero based absolute column index.</li>
 * </ul>
 *
 * @ojsubid oj-table-header
 * @memberof oj.ojTable
 *
 * @example <caption>Get the header at the specified location:</caption>
 * var node = myTable.getNodeBySubId( {'subId': 'oj-table-header', 'index':0} );
 */

/**
 * <p>Sub-ID for the Table element's sort ascending icon in column headers.</p>
 *
 * To lookup a sort icon the locator object should have the following:
 * <ul>
 * <li><b>subId</b>: 'oj-table-sort-ascending'</li>
 * <li><b>index</b>: the zero based absolute column index</li>
 * </ul>
 *
 * @ojsubid oj-table-sort-ascending
 * @memberof oj.ojTable
 *
 * @example <caption>Get the sort ascending icon from the header at the specified location:</caption>
 * var node = myTable.getNodeBySubId( {'subId': 'oj-table-sort-ascending', 'index':0} );
 */

/**
 * <p>Sub-ID for the Table element's sort descending icon in column headers.</p>
 *
 * To lookup a sort icon the locator object should have the following:
 * <ul>
 * <li><b>subId</b>: 'oj-table-sort-descending'</li>
 * <li><b>index</b>: the zero based absolute column index</li>
 * </ul>
 *
 * @ojsubid oj-table-sort-descending
 * @memberof oj.ojTable
 *
 * @example <caption>Get the sort descending icon from the header at the specified location:</caption>
 * var node = myTable.getNodeBySubId( {'subId': 'oj-table-sort-descending', 'index':0} );
 */

/**
 * <p>Sub-ID for the Table element's footers.</p>
 *
 *  To lookup a footer the locator object should have the following:
 * <ul>
 * <li><b>subId</b>: 'oj-table-footer'</li>
 * <li><b>index</b>: the zero based absolute column index.</li>
 * </ul>
 *
 * @ojsubid oj-table-footer
 * @memberof oj.ojTable
 *
 * @example <caption>Get the header at the specified location:</caption>
 * var node = myTable.getNodeBySubId( {'subId': 'oj-table-footer', 'index':0} );
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for the Table element's cells.</p>
 *
 * @property {number} rowIndex the zero based absolute row index
 * @property {number} columnIndex the zero based absolute column index
 * @property {string} key the row key
 *
 * @ojnodecontext oj-table-cell
 * @memberof oj.ojTable
 */

/**
 * <p>Context for the Table element's headers.</p>
 *
 * @property {number} index the zero based absolute column index
 *
 * @ojnodecontext oj-table-header
 * @memberof oj.ojTable
 */

/**
 * <p>Context for the Table element's footers.</p>
 *
 * @property {number} index the zero based absolute column index
 *
 * @ojnodecontext oj-table-footer
 * @memberof oj.ojTable
 */

(function() {
var ojTableMeta = {
  "properties": {
    "accessibility": {
      "type": "Object",
      "properties": {
        "rowHeader": {
          "type": "string"
        }
      }
    },
    "as": {
      "type": "string"
    },
    "columns": {
      "type": "Array<Object>",
      "writeback": true
    },
    "columnsDefault": {
      "type": "Object",
      "properties": {
        "className": {
          "type": "string"
        },
        "field": {
          "type": "string"
        },
        "footerClassName": {
          "type": "string"
        },
        "footerRenderer": {},
        "footerStyle": {
          "type": "string"
        },
        "footerTemplate": {
          "type": "string"
        },
        "headerClassName": {
          "type": "string"
        },
        "headerRenderer": {},
        "headerStyle": {
          "type": "string"
        },
        "headerTemplate": {
          "type": "string"
        },
        "headerText": {
          "type": "string"
        },
        "renderer": {},
        "resizable": {
          "type": "string"
        },
        "sortable": {
          "type": "string",
          "enumValues": ["auto", "enabled", "disabled", "none"]
        },
        "sortProperty": {
          "type": "string"
        },
        "style": {
          "type": "string"
        },
        "template": {
          "type": "string"
        },
        "width": {
          "type": "number"
        }
      }
    },
    "currentRow": {
      "type": "Object",
      "writeback": true
    },
    "data": {},
    "display": {
      "type": "string",
      "enumValues": ["list", "grid"]
    },
    "dnd": {
      "type": "Object",
      "properties": {
        "drag": {
          "type": "Object",
          "properties": {
            "rows": {
              "type": "Object",
              "properties": {
                "dataTypes": {"type": "string|Array<string>"},
                "dragStart": {},
                "drag": {},
                "dragEnd": {}
              }
            }
          }
        },
        "drop": {
          "type": "Object",
          "properties": {
            "columns": {
              "type": "Object",
              "properties": {
                "dataTypes": {"type": "string|Array<string>"},
                "dragEnter": {},
                "dragOver": {},
                "dragLeave": {},
                "drop": {}
              }
            },
            "rows": {
              "type": "Object",
              "properties": {
                "dataTypes": {"type": "string|Array<string>"},
                "dragEnter": {},
                "dragOver": {},
                "dragLeave": {},
                "drop": {}
              }
            }
          }
        },
        "reorder": {
          "type": "Object",
          "properties": {
            "columns": {
              "type": "string",
              "enumValues": ["disabled", "enabled"]
            }
          }
        }
      }
    },
    "editMode": {
      "type": "string",
      "enumValues": ["none", "rowEdit"]
    },
    "firstSelectedRow": {
      "type": "Object"
    },
    "horizontalGridVisible": {
      "type": "string",
      "enumValues": ["auto", "enabled", "disabled"]
    },
    "rowRenderer": {},
    "scrollPolicy": {
      "type": "string",
      "enumValues": ["auto", "loadMoreOnScroll"]
    },
    "scrollPolicyOptions": {
      "type": "Object",
      "properties": {
        "fetchSize": {
          "type": "number"
        },
        "maxCount": {
          "type": "number"
        }
      }
    },
    "selection": {
      "type": "Array<Object>",
      "writeback": true
    },
    "selectionMode": {
      "type": "Object",
      "properties": {
        "row": {
          "type": "string",
          "enumValues": ["single", "multiple"]
        },
        "column": {
          "type": "string",
          "enumValues": ["single", "multiple"]
        }
      }
    },
    "selectionRequired": {
      "type": "boolean"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "labelAccSelectionAffordanceBottom": {
          "type": "string"
        },
        "labelAccSelectionAffordanceTop": {
          "type": "string"
        },
        "labelDisableNonContiguousSelection": {
          "type": "string"
        },
        "labelEditRow": {
          "type": "string",
          "value": "Edit Row"
        },
        "labelEnableNonContiguousSelection": {
          "type": "string"
        },
        "labelResize": {
          "type": "string"
        },
        "labelResizePopupSpinner": {
          "type": "string"
        },
        "labelResizePopupSubmit": {
          "type": "string"
        },
        "labelSelectAndEditRow": {
          "type": "string",
          "value": "Select And Edit Row"
        },
        "labelSelectColum": {
          "type": "string",
          "value": "Select Column"
        },
        "labelSelectRow": {
          "type": "string",
          "value": "Select Row"
        },
        "labelSort": {
          "type": "string",
          "value": "Sort"
        },
        "labelSortAsc": {
          "type": "string",
          "value": "Sort Ascending"
        },
        "labelSortDsc": {
          "type": "string",
          "value": "Sort Descending"
        },
        "msgColumnResizeWidthValidation": {
          "type": "string",
          "value": "Width value must be an integer."
        },
        "msgFetchingData": {
          "type": "string",
          "value": "Fetching Data..."
        },
        "msgInitializing": {
          "type": "string",
          "value": "Initializing..."
        },
        "msgNoData": {
          "type": "string",
          "value": "No data to display."
        },
        "msgScrollPolicyMaxCountDetail": {
          "type": "string",
          "value": "Please reload with smaller data set."
        },
        "msgScrollPolicyMaxCountSummary": {
          "type": "string",
          "value": "Exceeded maximum rows for table scrolling."
        },
        "msgStatusSortAscending": {
          "type": "string",
          "value": "{0} sorted in ascending order."
        },
        "msgStatusSortDescending": {
          "type": "string",
          "value": "{0} sorted in descending order."
        }
      }
    },
    "verticalGridVisible": {
      "type": "string",
      "enumValues": ["auto", "enabled", "disabled"]
    }
  },
  "methods": {
    "getContextByNode": {},
    "getDataForVisibleRow": {},
    "refreshRow": {},
  },
  "events": {
    "animateEnd": {},
    "animateStart": {},
    "beforeCurrentRow": {},
    "beforeRowEdit": {},
    "beforeRowEditEnd": {},
    "sort": {}
  },
  "extension": {
    _INNER_ELEM: 'table',
    _WIDGET_NAME: "ojTable"
  }
};
oj.CustomElementBridge.registerMetadata('oj-table', 'baseComponent', ojTableMeta);
oj.CustomElementBridge.register('oj-table', {'metadata': oj.CustomElementBridge.getMetadata('oj-table')});
})();

/**
 * Creates an attribute group handler that will generate stylistic attribute values such as colors or shapes based on data set categories.
 * 
 * @param {Object.<string, *>=} [matchRules] A map of key value pairs for categories and the matching attribute value e.g. {"soda" : "square", "water" : "circle", "iced tea" : "triangleUp"}.
 *                            Attribute values listed in the matchRules object will be reserved only for the matching categories when getAttributeValue is called.
 * @ojsignature {target: "Type", value: "{[propName: string]: any}", for: "matchRules"}
 * @export
 * @constructor
 * @since 1.0
 */
oj.AttributeGroupHandler = function(matchRules) {
  this.Init(matchRules);
};

oj.Object.createSubclass(oj.AttributeGroupHandler, oj.Object, "oj.AttributeGroupHandler");

oj.AttributeGroupHandler.prototype.Init = function(matchRules) {
  this._assignments = {};
  this._valueIndex = 0;
  this._matchRules = {};
  for (var category in matchRules) {
    this.addMatchRule(category, matchRules[category]);
  }
  // Delay initializing value ramp by calling subclass getValueRamp impl until needed either for adding match rule or assigning values
};

/**
 * Returns the array of possible attribute values for this attribute group handler.
 * @returns {Array.<*>} The array of attribute values
 * @export
 */
oj.AttributeGroupHandler.prototype.getValueRamp = function() {
  return [];
};

/**
 * Assigns the given category an attribute value.  Will consistently return the same attribute value for equal categories.
 * @param {string} category The category to assign
 * @returns {*} The attribute value for the category
 * @export
 */
oj.AttributeGroupHandler.prototype.getValue = function(category) {
  // When assigning value, first check match rules, then assign to next attribute group value.
  if (this._matchRules[category]) {
    return this._matchRules[category];
  }
  else if (!this._assignments[category]) {
    if (!this._values)
      this._values = this.getValueRamp().slice();

    this._assignments[category] = this._values[this._valueIndex];
    
    this._valueIndex++;
    if (this._valueIndex == this._values.length)
      this._valueIndex = 0;
  }
  return this._assignments[category];
};

/**
 * Returns the current list of assigned categories as an array of objects with "category" and "value" keys. Note that match rules are not
 * reflected in category assignments.
 * @return {Array.<Object.<string, *>>} The current list of category and value pairings
 * @ojsignature {target: "Type", value: "Array<{[propName: string]: any}>", for: "returns"}
 * @export
 */
oj.AttributeGroupHandler.prototype.getCategoryAssignments  = function() {
  var assignments = [];
  for (var i in this._assignments)
    assignments.push({"category": i, "value": this._assignments[i]});
  return assignments;
};

/**
 * Reserves an attribute value for the given category.  All match rules should be added before any category
 * assignments are done with the <a href="#getValue">getValue</a> API.
 * @param {string} category Used for checking inputs to getAttributeValue against when assigning an attribute value
 * @param {*} attributeValue The attribute value to assign for inputs matching the given category e.g. "square" or "circle"
 * @return {void}
 * @export
 */
oj.AttributeGroupHandler.prototype.addMatchRule = function(category, attributeValue) {
  this._matchRules[category] = attributeValue;
};

/**
 * Creates a color attribute group handler that will generate color attribute values.
 * 
 * @param {Object.<string, string>=} [matchRules] A map of key value pairs for categories and the
 * matching attribute value e.g. {"soda" : "#336699", "water" : "#CC3300", "iced tea" : "#F7C808"}.
 * Attribute values listed in the matchRules object will be reserved only for the
 * matching categories when getAttributeValue is called.  Note that not all colors
 * in the default color ramp will meet minimum contrast requirements for text.
 * @ojsignature {target: "Type", value: "{[propName: string]: any}", for: "matchRules"}
 * @export
 * @constructor
 * @since 1.2
 * @extends {oj.AttributeGroupHandler}
 */
oj.ColorAttributeGroupHandler = function(matchRules) {
  // Create the array of colors for this instance.
  this._attributeValues = [];

  if ($(document.body).hasClass('oj-hicontrast'))  {
    // High Contrast: CSS colors get overridden to all white or all black. Use the default colors instead.
    this._attributeValues = oj.ColorAttributeGroupHandler._DEFAULT_COLORS.slice();
  }
  else {
    // Process the colors from the skin if not done already.
    // To improve performance, append the divs for each style class first then process the colors for each div.
    var attrGpsDiv = oj.ColorAttributeGroupHandler.__createAttrDiv();
    if (attrGpsDiv) {
      oj.ColorAttributeGroupHandler.__processAttrDiv(attrGpsDiv);
      attrGpsDiv.remove();
    }

    // Clone and use the processed colors.
    if (oj.ColorAttributeGroupHandler._colors.length > 0)
      this._attributeValues = oj.ColorAttributeGroupHandler._colors.slice();
    else
      this._attributeValues = oj.ColorAttributeGroupHandler._DEFAULT_COLORS.slice();
  }

  this.Init(matchRules);
};

oj.Object.createSubclass(oj.ColorAttributeGroupHandler, oj.AttributeGroupHandler, "oj.ColorAttributeGroupHandler");

/** @private */
oj.ColorAttributeGroupHandler._DEFAULT_COLORS = ['#237bb1', '#68c182', '#fad55c', '#ed6647', 
  '#8561c8', '#6ddbdb', '#ffb54d', '#e371b2', '#47bdef', '#a2bf39', '#a75dba', '#f7f37b'];

/** @private */
oj.ColorAttributeGroupHandler._STYLE_CLASSES = ['oj-dvt-category1', 'oj-dvt-category2', 'oj-dvt-category3', 
  'oj-dvt-category4', 'oj-dvt-category5', 'oj-dvt-category6', 'oj-dvt-category7', 'oj-dvt-category8', 
  'oj-dvt-category9', 'oj-dvt-category10', 'oj-dvt-category11', 'oj-dvt-category12'];

/** @private */
oj.ColorAttributeGroupHandler._colors = null;

/**
 * Returns the array of possible color values for this attribute group handler.
 * @returns {Array.<string>} The array of color values
 * @export
 */
oj.ColorAttributeGroupHandler.prototype.getValueRamp = function() {
  return this._attributeValues;
};

/**
 * Creates an element and appends a div for each style class
 * @returns {jQuery} The jQuery element containing divs for each style class
 * @export
 * @deprecated since version 4.2.0
 */
oj.ColorAttributeGroupHandler.createAttrDiv = function() {
  return oj.ColorAttributeGroupHandler.__createAttrDiv();
};

/**
 * Processes the colors for each div on the given element
 * @param {jQuery} attrGpsDiv The jQuery element containing divs for each style class
 * @return {void}
 * @export
 * @deprecated since version 4.2.0
 */
oj.ColorAttributeGroupHandler.processAttrDiv = function(attrGpsDiv) {
  oj.ColorAttributeGroupHandler.__processAttrDiv(attrGpsDiv);
};

/**
 * Creates an element and appends a div for each style class
 * @returns {jQuery} The jQuery element containing divs for each style class
 * @ignore
 */
oj.ColorAttributeGroupHandler.__createAttrDiv = function() {
  if (oj.ColorAttributeGroupHandler._colors)
    return null;

  var attrGpsDiv = $(document.createElement("div"));
  attrGpsDiv.attr("style", "display:none;");
  attrGpsDiv.attr("id", "attrGps");
  $(document.body).append(attrGpsDiv); // @HTMLUpdateOK
  for (var i = 0; i < oj.ColorAttributeGroupHandler._STYLE_CLASSES.length; i++) {
    var childDiv = $(document.createElement("div"));
    childDiv.addClass(oj.ColorAttributeGroupHandler._STYLE_CLASSES[i]);
    attrGpsDiv.append(childDiv); // @HTMLUpdateOK
  }
  return attrGpsDiv;
};

/**
 * Processes the colors for each div on the given element
 * @param {jQuery} attrGpsDiv The jQuery element containing divs for each style class
 * @return {void}
 * @ignore
 */
oj.ColorAttributeGroupHandler.__processAttrDiv = function(attrGpsDiv) {
  oj.ColorAttributeGroupHandler._colors = [];
  
  var childDivs = attrGpsDiv.children();
  for (var i = 0; i < childDivs.length; i++) {
    var childDiv = $(childDivs[i]);
    var color = childDiv.css('color');
    if (color)
      oj.ColorAttributeGroupHandler._colors.push(color);
  }
};

/**
 * Utility class with functions for parsing common DVT attributes.
 * Currently the following DVT components are using it: ojchart, ojdiagram, ojgauge, ojnbox, ojthematicmap.
 * @ignore
 * @memberof oj.dvtBaseComponent
 * @private
 */
var DvtAttributeUtils = {};
// Consider a string with at least one digit a valid SVG path
DvtAttributeUtils._SHAPE_REGEXP = /\d/;
// Default shape types supported by DvtSimpleMarker
DvtAttributeUtils._SHAPE_ENUMS = {
  "circle": true,
  "ellipse": true,
  "square": true,
  "rectangle": true,
  "diamond": true,
  "triangleUp": true,
  "triangleDown": true,
  "plus": true,
  "human": true,
  "star": true
};
/**
 * @ignore
 * @param {Object} shapeAttrs A map containing the shape attribute name which should be full path using dot notation.
 * @param {Array.<string>} shapeEnums The array of allowed enumerated values for the shape property of this component.
 * @return {function(string, string, Object, Function):string}
 */
DvtAttributeUtils.shapeParseFunction = function(shapeAttrs, shapeEnums) {
  var shapes = shapeEnums || DvtAttributeUtils._SHAPE_ENUMS;
  return function(value, name, meta, defaultParseFunction) {
    if (shapeAttrs[name]) {
      if (DvtAttributeUtils._SHAPE_REGEXP.test(value))
        return value;
      else if (!shapes[value])
        throw new Error("Found: '" + value + "'. Expected one of the following: " + Object.keys(shapeEnums).toString());
      else
        return value;
    }
    return defaultParseFunction(value);
  };
};
/**
 * Class to help set css properties on the component root options object
 * @param {Object} object The root options object from which this path should be resolved
 * @param {string} path The string path within the options object to resolve
 * @protected
 * @constructor
 * @ignore
 */
var DvtJsonPath = function(object, path)
{
  this._path = path;
  this._root = object;
  this._delimiter = '/';
};

/**
 * Resolves the parameter of the leaf object and the leaf object itself
 * @param {Object} root The root object to update
 * @param {string} path The string path within the root object to resolve
 * @param {string} delimiter The string delimiter for the path string
 * @param {boolean} createIfMissing Flag to create the hierarchy of the namespaces if they do not exist
 * @return {Object} The resolved parameter
 * @private
 */
DvtJsonPath.prototype._resolveLeafObjectAndProperty = function(root, path, delimiter, createIfMissing)
{
  var result = {};
  while (root && path.indexOf(delimiter) > -1)
  {
    var subProperty = path.substring(0, path.indexOf(delimiter));
    if (createIfMissing && root[subProperty] === undefined)
    {
      root[subProperty] = {};
    }
    root = root[subProperty];
    path = path.substring(path.indexOf(delimiter) + 1, path.length);
  }

  if (root)
  {
    result['object'] = root;
    result['parameter'] = path;
  }

  return result;
};

/**
 * Resolves path to the leaf object and parameter of this object
 * @param {boolean} createIfMissing Flag to create the hierarchy of the namespaces if they do not exist
 * @private
 */
DvtJsonPath.prototype._resolvePath = function(createIfMissing)
{
  if (this._leaf === undefined)
  {
    var result = this._resolveLeafObjectAndProperty(this._root, this._path, this._delimiter, createIfMissing);

    this._leaf = result['object'];
    this._param = result['parameter'];
  }
};

/**
 * Returns value of the leaf element of the path.
 * @return {Object} value The value of the leaf element or undefined if path structure is not yet created
 */
DvtJsonPath.prototype.getValue = function()
{
  this._resolvePath(false);
  return this._leaf === undefined ? undefined : this._leaf[this._param];
};

/**
 * Sets value of the leaf element of the path.
 * @param {Object} value The value of the leaf element
 * @param {boolean} bOverride Whether to override the original value
 */
DvtJsonPath.prototype.setValue = function(value, bOverride)
{
  this._resolvePath(true);

  if (bOverride || !this._leaf[this._param])
  {
    this._leaf[this._param] = value;
  }
};

var DvtStyleProcessor = {
  'TEXT': function(cssDiv) {
    var ignoreProperties = {};
    if (cssDiv) {
      if (cssDiv.hasClass("oj-gauge-metric-label")) {
        // Ignored because the size and color are fit to shape and based on background color.
        ignoreProperties['font-size'] = true;
        ignoreProperties['color'] = true;
      }
      else if (cssDiv.hasClass("oj-treemap-node-header")) {
        // Ignored because the weight is automatically determined based on the layer of the header.
        ignoreProperties['font-weight'] = true;
      }
    }
    return DvtStyleProcessor._buildTextCssPropertiesObject(cssDiv, ignoreProperties);
  },
  'BACKGROUND': function(cssDiv, styleClass, property, path) {
    return DvtStyleProcessor._buildCssBackgroundPropertiesObject(cssDiv);
  },
  'ANIM_DUR': function(cssDiv) {
    var animDur = cssDiv.css('animation-duration');
    if (animDur) {
      // Convert to milliseconds
      if (animDur.slice(-2) == 'ms')
        animDur = parseInt((animDur.slice(0, -2)), 10);
      else if (animDur.slice(-1) == 's')
        animDur = parseFloat(animDur.slice(0, -1)) * 1000;
      return animDur;
    }
  }
};

DvtStyleProcessor._INHERITED_FONT_COLOR = "rgb(254, 0, 254)";
DvtStyleProcessor._INHERITED_FONT_FAMILY = "Times";
DvtStyleProcessor._INHERITED_FONT_SIZE = "1px";
DvtStyleProcessor._INHERITED_FONT_WEIGHT = "1";
DvtStyleProcessor._INHERITED_FONT_STYLE = "normal";

// Chrome adjusts px font size when zooming, so only set the font size if the inherited size is less than 4px.
DvtStyleProcessor._FONT_SIZE_BUFFER = 4.0;

/**
 * A map of resolved style properties, where each key is a style class and each value is an object with property to
 * value mappings.
 * @private
 */
DvtStyleProcessor._styleCache = {};

DvtStyleProcessor.defaultStyleProcessor = function(cssDiv, property) {
  return cssDiv.css(property);
};

/**
 * Returns the css background properties object.
 * @param {Object} cssDiv The element with style class or with some default style
 * @return {Object} The object of CSS background properties including
 *                  border-color, border-width, and background-color
 * @private
 */
DvtStyleProcessor._buildCssBackgroundPropertiesObject = function(cssDiv) {
  var cssObj = {};
  if (cssDiv.css('border-top-color')) {
    cssObj["borderColor"] = cssDiv.css('border-top-color');
  }
  if (cssDiv.css('border-width') && cssDiv.css('border-style') &&
      cssDiv.css('border-style') != 'none') {
    cssObj["borderWidth"] = cssDiv.css('border-width');
  }
  if (cssDiv.css('background-color')) {
    cssObj["backgroundColor"] = cssDiv.css('background-color');
  }
  return cssObj;
};

/**
 * Returns the css text properties object.
 * @param {Object} cssDiv The div to use for processing CSS style.
 * @param {Object} ignoreProperties The css properties to ignore
 * @return {Object} The CSS text properties object including font-family,
 *                  font-size, font-weight, color, and font-style
 */
DvtStyleProcessor._buildTextCssPropertiesObject = function(cssDiv, ignoreProperties)
{
  var cssObj = {};
  var value = cssDiv.css('font-family');
  if (value && value !== DvtStyleProcessor._INHERITED_FONT_FAMILY) {
    cssObj["fontFamily"] = value.replace(/"/g,"'");
  }
  value = cssDiv.css('font-size');
  if (value && !(value.indexOf("px") > -1 &&
      parseFloat(value) < DvtStyleProcessor._FONT_SIZE_BUFFER) &&
      !ignoreProperties ['font-size']) {
    cssObj["fontSize"] = value;
  }
  value = cssDiv.css('font-weight');
  if (value && value !== DvtStyleProcessor._INHERITED_FONT_WEIGHT &&
      !ignoreProperties['font-weight']) {
    cssObj["fontWeight"] = value;
  }
  value = cssDiv.css('color');
  if (value && value !== DvtStyleProcessor._INHERITED_FONT_COLOR &&
      !ignoreProperties ['color']) {
    cssObj["color"] = value;
  }
  value = cssDiv.css('font-style');
  if (value && value !== DvtStyleProcessor._INHERITED_FONT_STYLE) {
    cssObj["fontStyle"] = value;
  }
  return cssObj;
};

/**
 * Creates dummy divs for each component style class and merges their values with the component options object.
 * @param {Object} element DOM node to add CSS styles to for processing
 * @param {Object} options The options object to merge CSS properties with
 * @param {Array} componentClasses The style classes associated with the component
 * @param {Object} childClasses Style classes associated with a component's children
 */
DvtStyleProcessor.processStyles = function(element, options, componentClasses, childClasses)
{
  var outerDummyDiv = null;
  var innerDummyDiv = null;
  
  var styleClasses = '';
  for (var i=0; i<componentClasses.length; i++)
    styleClasses = styleClasses + componentClasses[i] + " ";
  
  var styleDefinitions = {};
  var styleDivs = {}
  
  for (var styleClass in childClasses) {
    // Retrieve the definitions for the styleClass. Convert to an array if not already.
    var definitions = childClasses[styleClass];
    if(!(definitions instanceof Array))
      definitions = [definitions];
    
    // Performance: Only create these divs if we need to process uncached definitions
    var hasUncachedProperty = DvtStyleProcessor._hasUncachedProperty(styleClass, definitions);
    if (!innerDummyDiv && !outerDummyDiv && hasUncachedProperty) {
      // Add the component style classes to a hidden dummy div
      outerDummyDiv = $(document.createElement("div"));
      outerDummyDiv.attr("style", "display:none;");
      element.append(outerDummyDiv); // @HTMLUpdateOK
      outerDummyDiv.attr("class", styleClasses);
      $(document.body).append(outerDummyDiv); // @HTMLUpdateOK
      
      // Add an inner dummy div to overwrite inherited values and prevent populating options object with them
      innerDummyDiv = $(document.createElement("div"));
      innerDummyDiv.css("font-size", DvtStyleProcessor._INHERITED_FONT_SIZE);
      innerDummyDiv.css("color", DvtStyleProcessor._INHERITED_FONT_COLOR);
      innerDummyDiv.css("font-weight", DvtStyleProcessor._INHERITED_FONT_WEIGHT);
      innerDummyDiv.css("font-style", DvtStyleProcessor._INHERITED_FONT_STYLE);
      outerDummyDiv.append(innerDummyDiv); // @HTMLUpdateOK    
    }
    
    // store style definitions to avoid recalculation in the next loop
    styleDefinitions[styleClass] = definitions;
    
    // create cssDivs for styles to be resolved during processing where necessary 
    if (hasUncachedProperty) 
      styleDivs[styleClass] = DvtStyleProcessor._createStyleDivs(innerDummyDiv, styleClass, definitions);
  }
  
  // Performance: Process color attribute groups for components along with the component styles
  var attrGpsDiv = oj.ColorAttributeGroupHandler.__createAttrDiv(); 
  
  for (var styleClass in childClasses) {
    DvtStyleProcessor._processStyle(styleDivs[styleClass], options, styleClass, styleDefinitions[styleClass]);
  }
  
  // Performance: Process color attribute groups for components along with the component styles
  if (attrGpsDiv) {
    oj.ColorAttributeGroupHandler.__processAttrDiv(attrGpsDiv);
    attrGpsDiv.remove();
  }
  
  // Remove the dummy div when complete
  if (outerDummyDiv)
    outerDummyDiv.remove();
};

/**
 * Creates and returns a cssDiv for the specified style class if there are definitions that need to be evaluated
 * @param {Object} parentElement The parent DOM node of the element that will be created to add CSS styles to for processing.
 * @param {string} styleClass The style class.
 * @param {Array} definitions Array of maps of CSS style attribute and values to process.
 * @return {Object} The created cssDiv
 * @private
 */
DvtStyleProcessor._createStyleDivs= function(parentElement, styleClass, definitions)
{
  // Ensure an entry for the styleClass exists in the cache.
  if(!DvtStyleProcessor._styleCache[styleClass])
    DvtStyleProcessor._styleCache[styleClass] = {};

  // Create a cssDiv for each definition.
  var cssDiv = null;
  for (var i=0;i<definitions.length; i++) {
    var definition = definitions[i];
    var property = definition['property'];
    if(property) {
      // Check if the style cache already has a resolved copy of the value
      var value = DvtStyleProcessor._styleCache[styleClass][property];
      if(typeof value == 'undefined') {
        // Property not found in the cache. Ensure creation of the cssDiv to resolve the property.
        if(!cssDiv) {
          cssDiv = $(document.createElement("div"));
          cssDiv.addClass(styleClass);
          parentElement.append(cssDiv); // @HTMLUpdateOK
        }
      }
    }
  }
  return cssDiv;  
};

/**
 * Resolves the css properties for a specified style class.
 * @param {Object} cssDiv The DOM node with addeed CSS styles to process.
 * @param {Object} options The options object to merge CSS properties with.
 * @param {string} styleClass The style class.
 * @param {Array} definitions Array of maps of CSS style attribute and values to process.
 * @private
 */
DvtStyleProcessor._processStyle = function(cssDiv, options, styleClass, definitions)
{
  for (var i = 0; i < definitions.length; i++) {
    var definition = definitions[i];
    var property = definition['property'];
    if(property) {
      // Check if the style cache already has a resolved copy of the value
      var value = DvtStyleProcessor._styleCache[styleClass][property];
      
      if(typeof value == 'undefined' && cssDiv) {
        // Resolve and store in the cache.
        value = DvtStyleProcessor._resolveStyle(cssDiv, property);
        DvtStyleProcessor._styleCache[styleClass][property] = value;
      }

      // If a value exists, apply to the JSON.
      if (value != null) {
        var path = new DvtJsonPath(options, definition['path']);
        // If the value is an object, combine the value from the options with the value from the CSS.
        var optionsValue = path.getValue()
        var isObj = typeof optionsValue === 'object';
        if (isObj && optionsValue) {
          var mergedValue = {};
          var keys = Object.keys(value);
          for (var i = 0; i < keys.length; i++) {
            var key = keys[i];
            if (optionsValue[key] === undefined)
              optionsValue[key] = value[key];
          }
        } else {
          // If we have already merged the original value with the new value, such as for
          // text or background styles, we do not need to update the path value.
          path.setValue(value, isObj);
        }
      }
    }
  }
};

/**
 * Helper function to resolve the css properties within a dummy div and handle caching.
 * @param {Object} cssDiv The div to use for processing CSS style.
 * @param {string} property The css property to be resolved.
 * @return {string | null} The resolved value.
 * @private
 */
DvtStyleProcessor._resolveStyle = function(cssDiv, property)
{
  var handler = DvtStyleProcessor[property];
  var value = handler ? handler(cssDiv) : DvtStyleProcessor.defaultStyleProcessor(cssDiv, property);

  // Ensure the returned value is not null or empty string.
  return (value != null && !(typeof value == 'string' && value.replace(/^\s+/g, '') == '')) ? value : null;
};

/**
 * Helper function to verify if any style properties are not in the cache.
 * @param {string} styleClass The style class.
 * @param {Array} definitions The array of definitions for the style class.
 * @return {boolean}
 * @private
 */
DvtStyleProcessor._hasUncachedProperty = function(styleClass, definitions)
{
  var styleCache = DvtStyleProcessor._styleCache[styleClass];
  if (!styleCache)
    return true;
    
  for (var i = 0;i < definitions.length; i++) {
    var property = definitions[i]['property'];
    if(property) {
      var value = styleCache[property];    
      if(typeof value == 'undefined')
        return true;
    }
  }
  return false;
};
/**
 * Creates a shape attribute group handler that will generate shape attribute values.
 * 
 * @param {Object.<string, string>=} [matchRules] A map of key value pairs for categories and the matching attribute value e.g. {"soda" : "square", "water" : "circle", "iced tea" : "triangleUp"}.
 *                            Attribute values listed in the matchRules object will be reserved only for the matching categories when getAttributeValue is called.
 * @ojsignature {target: "Type", value: "{[propName: string]: any}", for: "matchRules"}
 * @export
 * @constructor
 * @since 1.0
 * @extends {oj.AttributeGroupHandler}
 */
oj.ShapeAttributeGroupHandler = function(matchRules) {
  this.Init(matchRules);
};

oj.Object.createSubclass(oj.ShapeAttributeGroupHandler, oj.AttributeGroupHandler, "oj.ShapeAttributeGroupHandler");

oj.ShapeAttributeGroupHandler._attributeValues = ['square', 'circle', 'diamond', 'plus', 'triangleDown', 'triangleUp', 'human'];

/**
 * Returns the array of possible shape values for this attribute group handler.
 * @returns {Array.<string>} The array of shape values
 * @export
 */
oj.ShapeAttributeGroupHandler.prototype.getValueRamp = function() {
  return oj.ShapeAttributeGroupHandler._attributeValues;
};
 /**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.dvtBaseComponent
 * @augments oj.baseComponent
 * @since 0.7
 * @abstract
 * @hideconstructor
 */
oj.__registerWidget('oj.dvtBaseComponent', $['oj']['baseComponent'], {
  options: {
    /**
     * Defines whether the element will automatically render in response to
     * changes in size. If set to <code class="prettyprint">off</code>, then the
     * application is responsible for calling <code class="prettyprint">refresh</code>
     * to render the element at the new size.
     * @expose
     * @name trackResize
     * @memberof oj.dvtBaseComponent
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "on"
     * 
     * @example <caption>Initialize the data visualization element with the 
     * <code class="prettyprint">track-resize</code> attribute specified:</caption>
     * &lt;oj-some-dvt track-resize='off'>&lt;/oj-some-dvt>
     * 
     * @example <caption>Get or set the <code class="prettyprint">trackResize</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myComponent.trackResize;
     * 
     * // setter
     * myComponent.trackResize="off";
     */
    trackResize: "on"
  },

  //** @inheritdoc */
  _ComponentCreate : function() {
    this._super();
    this._renderCount = 0;
    this._numDeferredObjs = 0;
    this._optionsCopy = null;
    this._templateMap = {};

    // Append the component style classes to the element
    var componentStyles = this._GetComponentStyleClasses();
    for(var i=0; i<componentStyles.length; i++) {
      this.element.addClass(componentStyles[i]);
    }

    // Create a reference div within the element to be used for computing relative event coords.
    this._referenceDiv = $(document.createElement("div"));
    this._referenceDiv.attr("style", "visibility:hidden;");
    this.element.append(this._referenceDiv); // @HTMLUpdateOK

    // Create the dvt.Context, which creates the svg element and adds it to the DOM.
    var parentElement = this.element[0].parentElement;
    if (parentElement && parentElement._dvtcontext)
      this._context = parentElement._dvtcontext;
    else
      this._context = new dvt.Context(this.element[0], null, this._referenceDiv[0]);

    // Store JET reference on context so toolkit can access things like loggers
    this._context['oj'] = oj;

    // Set the reading direction on the context
    this._context.setReadingDirection(this._GetReadingDirection());

    // Set the tooltip and datatip callbacks and div style classes
    this._context.setTooltipAttachedCallback(oj.Components.subtreeAttached);
    this._context.setOverlayAttachedCallback(oj.Components.subtreeAttached);
    this._context.setTooltipStyleClass('oj-dvt-tooltip');
    this._context.setDatatipStyleClass('oj-dvt-datatip');

    // Pass back method for cleaning up renderer context
    this._context.setFixContextCallback(this._FixRendererContext.bind(this));

    this._context.setCustomElement(this._IsCustomElement());

    // Set high contrast mode if needed
    if ($(document.body).hasClass('oj-hicontrast'))
      dvt.Agent.setHighContrast(true);

    // Create and cache the component instance
    this._component = this._CreateDvtComponent(this._context, this._HandleEvent, this);

    // Add the component to the display tree of the rendering context.
    this._context.getStage().addChild(this._component);

    // If requireJS is not used, can't rely on internationalization modules.
    if(dvt['requireJS'] !== false) {
      // Set the helpers for locale support
      this._setLocaleHelpers();

      // Retrieve and apply the translated strings onto the component bundle
      this._processTranslations();
    }

    // Load component resources
    this._LoadResources();
    
    // Initialize array to store data provider listeners
    this._dataProviderEventListeners = [];

    // Pass the environment and widget constructor through the options for JET specific behavior
    this.options['_environment'] = 'jet';
    this.options['_widgetConstructor'] = oj.Components.__GetWidgetConstructor(this.element);
  },

  //** @inheritdoc */
  _AfterCreate : function ()
  {
    // Allow superclass to process root attributes and context menus
    this._super();

    // Resize Listener Support
    if(this.options['trackResize'] != 'off') {
      this._addResizeListener();
    }

    this._ProcessOptions();

    // Render the component
    this._Render();
  },

  //** @inheritdoc */
  refresh : function() {
    this._super();

    // Update the reading direction on the context
    this._context.setReadingDirection(this._GetReadingDirection());

    // Retrieve and apply the translated strings onto the component bundle
    this._processTranslations();

    // Render the component with any changes
    this._Render();
  },

  //** @inheritdoc */
  getNodeBySubId : function(locator) {
    var automation = (this._component && this._component.getAutomation) ? this._component.getAutomation() : null;
    if (automation) {
      // Convert the locator to the subid string, since the shared JS layer only accepts the subid string syntax.
      var subId = this._ConvertLocatorToSubId(locator);
      return automation.getDomElementForSubId(subId);
    }
    return null;
  },

  //** @inheritdoc */
  getSubIdByNode : function(node) {
    var automation = (this._component && this._component.getAutomation) ? this._component.getAutomation() : null;
    if (automation) {
      // Retrieve the subid string from the shared JS layer, and convert it to the locator object.
      var subId = automation.getSubIdForDomElement(node);
      return subId ? this._ConvertSubIdToLocator(subId) : null;
    }
    return null;
  },

  /**
   * Converts the specified locator object into a subId string.
   * @param {Object} locator
   * @return {string|null}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _ConvertLocatorToSubId : function(locator) {
    // subclasses must override to support getNodeBySubId
    return null;
  },

  /**
   * Converts the specified subId string into a locator object.
   * @param {string} subId
   * @return {Object|null}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _ConvertSubIdToLocator : function(subId) {
    // subclasses must override to support getSubIdByNode
    return null;
  },

  /**
   * Create dummy divs for style classes and merge style class values with json .
   * options object
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _ProcessStyles : function() {
    // Process selectors for this component
    DvtStyleProcessor.processStyles(this.element, this.options,
                                    this._GetComponentStyleClasses(),
                                    this._GetChildStyleClasses());
  },

  /**
   * Returns the style classes associated with the component.
   * @return {Array}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _GetComponentStyleClasses : function() {
    return ['oj-dvtbase'];
  },

  /**
   * Returns a map of the style classes associated with a component's children.
   * @return {Object}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _GetChildStyleClasses : function() {
    var styleClasses = {};
    styleClasses['oj-dvt-no-data-message'] = {'path': '_statusMessageStyle', 'property': 'TEXT'};
    return styleClasses;
  },

  /**
   * Returns an array of supported event types.  Used in conjunction with _setOptions to skip unnecessary rendering when
   * event listeners are bound. Subclasses must override to return supported event types.
   * @return {Array}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _GetEventTypes : function() {
    return [];
  },

  /**
   * Returns a map containing keys corresponding to the string ids in ojtranslations.js and values corresponding to the
   * toolkit constants for the DvtBundle objects.  This map must be guaranteed to be a new instance so that subclasses can
   * add their translations to it.
   * @return {Object}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Create the mapping to return.
    var ret = {
      'DvtUtilBundle.CLEAR_SELECTION': translations['labelClearSelection'],
      'DvtUtilBundle.COLON_SEP_LIST' : translations['labelAndValue'],
      'DvtUtilBundle.INVALID_DATA': translations['labelInvalidData'],
      'DvtUtilBundle.NO_DATA': translations['labelNoData'],

      // Accessibility
      'DvtUtilBundle.DATA_VISUALIZATION': translations['labelDataVisualization'],
      'DvtUtilBundle.STATE_SELECTED': translations['stateSelected'],
      'DvtUtilBundle.STATE_UNSELECTED': translations['stateUnselected'],
      'DvtUtilBundle.STATE_MAXIMIZED': translations['stateMaximized'],
      'DvtUtilBundle.STATE_MINIMIZED': translations['stateMinimized'],
      'DvtUtilBundle.STATE_EXPANDED': translations['stateExpanded'],
      'DvtUtilBundle.STATE_COLLAPSED': translations['stateCollapsed'],
      'DvtUtilBundle.STATE_ISOLATED': translations['stateIsolated'],
      'DvtUtilBundle.STATE_HIDDEN': translations['stateHidden'],
      'DvtUtilBundle.STATE_VISIBLE': translations['stateVisible'],

      'DvtUtilBundle.SCALING_SUFFIX_THOUSAND': translations['labelScalingSuffixThousand'],
      'DvtUtilBundle.SCALING_SUFFIX_MILLION': translations['labelScalingSuffixMillion'],
      'DvtUtilBundle.SCALING_SUFFIX_BILLION': translations['labelScalingSuffixBillion'],
      'DvtUtilBundle.SCALING_SUFFIX_TRILLION': translations['labelScalingSuffixTrillion'],
      'DvtUtilBundle.SCALING_SUFFIX_QUADRILLION': translations['labelScalingSuffixQuadrillion']
    };

    // Add abbreviated month strings
    var monthNames = oj.LocaleData.getMonthNames("abbreviated");
    ret['DvtUtilBundle.MONTH_SHORT_JANUARY'] = monthNames[0];
    ret['DvtUtilBundle.MONTH_SHORT_FEBRUARY'] = monthNames[1];
    ret['DvtUtilBundle.MONTH_SHORT_MARCH'] = monthNames[2];
    ret['DvtUtilBundle.MONTH_SHORT_APRIL'] = monthNames[3];
    ret['DvtUtilBundle.MONTH_SHORT_MAY'] = monthNames[4];
    ret['DvtUtilBundle.MONTH_SHORT_JUNE'] = monthNames[5];
    ret['DvtUtilBundle.MONTH_SHORT_JULY'] = monthNames[6];
    ret['DvtUtilBundle.MONTH_SHORT_AUGUST'] = monthNames[7];
    ret['DvtUtilBundle.MONTH_SHORT_SEPTEMBER'] = monthNames[8];
    ret['DvtUtilBundle.MONTH_SHORT_OCTOBER'] = monthNames[9];
    ret['DvtUtilBundle.MONTH_SHORT_NOVEMBER'] = monthNames[10];
    ret['DvtUtilBundle.MONTH_SHORT_DECEMBER'] = monthNames[11];

    return ret;
  },

  /**
   * Sets up resources needed by the component
   * @instance
   * @override
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _SetupResources: function() {
    this._super();
    /* DataProvider Uptake : Commented out for code coverage. Uncomment for supporting DVTs
    this._addDataProviderEventListeners();
    */
  },
  
  /**
   * Release resources held by the component
   * @instance
   * @override
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _ReleaseResources: function() {
    this._super();
    /* DataProvider Uptake : Commented out for code coverage. Uncomment for supporting DVTs
    this._removeDataProviderEventListeners();
    */
  },
  
  /**
   * Called to process the translated strings for this widget.
   * @private
   */
  _processTranslations: function() {
    // Retrieve the map of translation keys + DvtBundle identifiers
    var translationMap = this._GetTranslationMap();

    // Register with the DvtBundle
    dvt.Bundle.addLocalizedStrings(translationMap);
  },

  /**
   * @private
   * @instance
   * @memberof! oj.dvtBaseComponent
   */
  _setLocaleHelpers : function () {
    var helpers = {};

    // Number converter factory for use in formatting default strings
    helpers['numberConverterFactory'] = oj.Validation.getDefaultConverterFactory('number');

    // Iso to date converter to be called for JS that requires Dates
    helpers['isoToDateConverter'] = function(input) {
      if (typeof(input) == 'string') {
        var dateWithTimeZone = oj.IntlConverterUtils.isoToDate(input);
        var localIsoTime = dateWithTimeZone.toJSON() ? oj.IntlConverterUtils.dateToLocalIso(dateWithTimeZone) : input;
        return oj.IntlConverterUtils.isoToLocalDate(localIsoTime);
      }
      else
        return input;
    }

    // Date to iso converter to be called before passing to the date time converter
    helpers['dateToIsoWithTimeZoneConverter'] = function(input) {
      if (input instanceof Date) {
        var timeZoneOffest = -1*input.getTimezoneOffset();
        var offsetSign = (timeZoneOffest >= 0 ? "+" : "-");
        var offsetHour = Math.floor(Math.abs(timeZoneOffest)/60);
        var offsetMinutes = Math.abs(timeZoneOffest)%60;
        var isoTimeZone =  offsetSign+(offsetHour.toString().length !== 2 ? "0"+offsetHour : offsetHour)+":"+ (offsetMinutes.toString().length !== 2 ? offsetMinutes+"0" : offsetMinutes);
        return oj.IntlConverterUtils.dateToLocalIso(input) + isoTimeZone;
      }
      else
        return input;
    }

    this._context.setLocaleHelpers(helpers);
  },

  //** @inheritdoc */
  _destroy : function() {
    // Hide all component tooltips and remove references to the DvtContext
    this._context.hideTooltips();
    this._context = null;

    var parentElement = this.element[0].parentElement;
    if (parentElement && parentElement._dvtcontext)
      parentElement._dvtcontext = null;

    // Call destroy on the JS component
    if (this._component.destroy)
      this._component.destroy();
    this._component = null;

    // Remove DOM resize listener
    this._removeResizeListener();

    this._CleanAllTemplates();
    
    // Remove children and clean up DOM changes
    this.element.children().remove();
    this.element.removeAttr('role').removeAttr('tabIndex').removeAttr('aria-activedescendant');

    // Remove style classes that were added
    var componentStyles = this._GetComponentStyleClasses();
    for(var i=0; i<componentStyles.length; i++) {
      this.element.removeClass(componentStyles[i]);
    }
    
    // Remove any pending busy states
    this._MakeReady();

    // Call super last for destroy
    this._super();
  },

  //** @inheritdoc */
  _setOptions : function(options, flags) {
    // Call the super to update the property values
    this._superApply(arguments);

    // Add or remove the resize tracking if changed.
    var trackResize = this.options['trackResize'];
    if(trackResize == 'off' && this._resizeListener)
      this._removeResizeListener();
    else if(trackResize != 'off' && !this._resizeListener)
      this._addResizeListener();

    this._ProcessOptions();

    // Render the component with the updated options.
    if(this._bUserDrivenChange) {
      // Option change fired in response to user gesture. Already reflected in UI, so no render needed.
      return;
    }
    else {
      // Event listeners don't require rendering.  Iterate through options to check for non-event options.
      // Also no render is needed if the component has exposed a method to update the option without rerendering.
      var bRenderNeeded = false;
      var eventTypes = this._GetEventTypes();
      var optimizedOptions = ['highlightedCategories', 'selection', 'dataCursorPosition'];
      $.each(options, function(key, value) {
        if(eventTypes.indexOf(key) < 0 && optimizedOptions.indexOf(key) < 0) {
          bRenderNeeded = true;
          return false;
        }
      });

      if(bRenderNeeded)
        this._Render();
      else {
        // Update options without rerendering. Check for undefined to allow nulls.
        if (options['highlightedCategories'] !== undefined)
          this._component.highlight(options['highlightedCategories']);
        if (options['selection'] !== undefined)
          this._component.select(options['selection']);
        if (options['dataCursorPosition'] !== undefined && this._component.positionDataCursor)
          this._component.positionDataCursor(options['dataCursorPosition']);
      }
    }
  },


  /**
   * Called by _create to instantiate the specific DVT component instance.  Subclasses must override.
   * @param {dvt.Context} context
   * @param {Function} callback
   * @param {Object} callbackObj
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    return null; // subclasses must override
  },

  /**
   * Called by the component to process events.  Subclasses should override to delegate DVT component events to their
   * JQuery listeners.
   * @param {Object} event
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _HandleEvent : function(event) {
    // TODO: hiddenCategories and highlightedCategories should use the option change event

    var type = event['type'];
    if(type === 'selection') {
      this._UserOptionChange('selection', event['selection']);
    }
    else if (type === 'categoryHide' || type === 'categoryShow') {
      this._UserOptionChange('hiddenCategories', event['hiddenCategories']);
    }
    else if (type === 'categoryHighlight') {
      this._UserOptionChange('highlightedCategories', event['categories']);
    }
    else if (type === 'optionChange') {
      this._UserOptionChange(event['key'], event['value'], event['optionMetadata']);
    }
    else if (type === 'touchHoldRelease' && this._GetContextMenu()) {
      this._OpenContextMenu($.Event(event['nativeEvent']), 'touch');
    }
    else if (type === 'ready') {
      // Handles case where two option sets occur and the second set
      // containing deferred data. We don't want to prematurely resolve
      // any when ready promises until the deferred data has finished.
      if (this._numDeferredObjs === 0)
        this._MakeReady();
    }
  },

  /**
   * Adds a resize listener for this component.
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _addResizeListener : function() {
    if(!this._resizeListener) {
      this._resizeListener = this._handleResize.bind(this);
      oj.DomUtils.addResizeListener(this.element[0], this._resizeListener, 250);
    }
  },

  /**
   * Removes the resize listener for this component.
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _removeResizeListener : function() {
    if(this._resizeListener) {
      oj.DomUtils.removeResizeListener(this.element[0], this._resizeListener);
      this._resizeListener = null;
    }
  },

  /**
   * Called when the component is resized.
   * @param {number} width
   * @param {number} height
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _handleResize : function(width, height) {
    // Render the component at the new size if it changed enough
    var newWidth = this.element.width();
    var newHeight = this.element.height();
    if(this._width == null || this._height == null || (Math.abs(newWidth - this._width) + Math.abs(newHeight - this._height) >= 5))
      this._Render(true);
  },

  /**
   * Called once during component creation to load resources.
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _LoadResources : function() {
    // subcomponents should override
  },

  /**
   * Called to render the component at the current size.
   * @param {boolean} isResize (optional) Whether it is a resize rerender.
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _Render : function(isResize) {
    // Hide all component tooltips on rerender cases to avoid abandoned tooltips
    this._context.hideTooltips();

    // Starting a new render - no longer ready
    this._NotReady();

    // Fix 18498656: If the component is not attached to a visible subtree of the DOM, rendering will fail because
    // getBBox calls will not return the correct values.
    // Note: Checking offsetParent() does not work here since it returns false for position: fixed.
    if(!this._context.isReadyToRender()) {
      this._renderNeeded = true;
      this._MakeReady();
    }
    else {
      // If flowing layout is supported, don't pass width and height to the component render method
      this._width = this._IsFlowingLayoutSupported() ? null : this.element.width();
      this._height = this._IsFlowingLayoutSupported() ? null : this.element.height();
      
      // Set the root font-family
      this._context.setDefaultFontFamily(this._referenceDiv.css('font-family'));

      // Add the width, height, and locale as private fields in the options for debugging purposes
      this.options['_width'] = this._width;
      this.options['_height'] = this._height;
      this.options['_locale'] = oj.Config.getLocale();

      // Add draggable attribute if DnD is supported
      if (this._IsDraggable())
        this.element.attr('draggable', true);

      // Merge css styles with with json options object
      this._ProcessStyles();

      // Render the component.
      this._renderCount++;
      // Skip the options on resize to suppress animation.
      if (isResize) {
        // Skip the resize render if Promises are not fully resolved because
        // the component will be rerendered with the new width/height when all
        // Promises are fully resolved
        if (this._numDeferredObjs === 0) {
          this._RenderComponent(this._optionsCopy, isResize);
        }
      } else {
        // Component rendering will be done when all Promises are fully resolved
        if (this._resolveDeferredDataItems()) {
          this._RenderComponent(this._optionsCopy);
        }
      }

      this._renderNeeded = false;
    }
  },
  
  /**
   * Check needed for setting draggable attribute on the component.
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _IsDraggable: function()
  {
    return !!this.options['dnd'];
  },

  //** @inheritdoc */
  _NotifyShown: function()
  {
    this._super();
    this._notifyShownAttached();
  },

  //** @inheritdoc */
  _NotifyAttached: function()
  {
    this._super();
    this._notifyShownAttached();
  },

  //** @inheritdoc */
  _NotifyDetached : function() 
  {
    this._super();
    this._notifyHiddenDetached();
  },

  //** @inheritdoc */
  _NotifyHidden : function() 
  {
    this._super();
    this._notifyHiddenDetached();
  },

  /**
   * Helper method to perform common logic for when 
   * a DVT is hidden or detached from the DOM.
   * @memberof oj.dvtBaseComponent
   * @instance
   * @private
   */
  _notifyShownAttached: function()
  {
    if(this._renderNeeded)
      this._Render();
  },

  /**
   * Helper method to perform common logic for when 
   * a DVT is shown or detached from the DOM.
   * @memberof oj.dvtBaseComponent
   * @instance
   * @private
   */
  _notifyHiddenDetached: function() 
  {
    this._context.hideTooltips();

    // Remove any pending busy states
    this._MakeReady();
  },

  /**
   * Sets an option change that was driven by user gesture.  Used in conjunction with _setOption to ensure that the
   * correct optionMetadata flag for writeback is set.
   * @param {string} key The name of the option to set.
   * @param {Object} value The value to set for the option.
   * @param {Object} optionMetadata The optionMetadata for the optionChange event
   * @memberof oj.dvtBaseComponent
   * @instance
   * @protected
   */
  _UserOptionChange : function(key, value, optionMetadata) {
    this._bUserDrivenChange = true;

    this.option(key, value, {'_context': {writeback:true, optionMetadata: optionMetadata, internalSet: true}});

    this._bUserDrivenChange = false;
  },

  //** @inheritdoc */
  _NotifyContextMenuGesture: function(menu, event, eventType) {
    // DVTs support context menus on touch hold release which is detected by the
    // toolkit and handled in _HandleEvent after receiving touch hold release event.
    if (eventType === "touch")
      return;
    // Position context menus relative to the current keyboard focus when keyboard triggered
    else if (eventType === "keyboard") {
      var compElementRect = this.element[0].getBoundingClientRect(),
        focusElementRect = this._component.getKeyboardFocus() ? this._component.getKeyboardFocus().getBoundingClientRect() : null;
      var position = focusElementRect ? 
        "left+" + (focusElementRect.left + focusElementRect.width*.5 - compElementRect.left) + " top+" + (focusElementRect.top + focusElementRect.height*.5 - compElementRect.top) : "center";
      this._OpenContextMenu(event, eventType, {"position": {"at": position}});
    }
    else
      this._super(menu, event, eventType);
  },

  /**
   * Returns a DVT component associated with a DOMElement
   * @param {Element} element The DOMElement to get the DVT component from.
   * @return {Object} The DVT component associated with the DOMElement or null
   * @memberof oj.dvtBaseComponent
   * @instance
   * @protected
   */
  _GetDvtComponent: function(element) {
    var widget = oj.Components.__GetWidgetConstructor(element)("instance");
    if (widget) {
      return widget._component;
    }
    return null;
  },

  /**
   * Converts an indexPath array to a string of the form '[index0][index1]...[indexN]'.
   * @param {Array} indexPath
   * @return {string} The resulting string.
   * @instance
   * @memberof oj.dvtBaseComponent
   * @protected
   */
  _GetStringFromIndexPath : function(indexPath) {
    var ret = '';
    for(var i=0; i<indexPath.length; i++) {
      ret += '[' + indexPath[i] + ']';
    }
    return ret;
  },

  /**
   * Converts a string containing indices in the form '[index0][index1]...[indexN]' to an array of indices.
   * @param {string} subId
   * @return {Array} The resulting array to be used for locator indexPath.
   * @instance
   * @memberof oj.dvtBaseComponent
   * @protected
   */
  _GetIndexPath : function(subId) {
    var indexPath = [];
    var currentIndex = 0;
    while(subId.indexOf('[', currentIndex) > 0) {
      var start = subId.indexOf('[', currentIndex) + 1;
      var end = subId.indexOf(']', currentIndex)
      indexPath.push(Number(subId.substring(start, end)));
      currentIndex = end + 1;
    }
    return indexPath;
  },

  /**
   * Converts a string containing a single index in the form '[index]' into the numerical index.
   * @param {string} subId
   * @return {number}
   * @instance
   * @memberof oj.dvtBaseComponent
   * @protected
   */
  _GetFirstIndex : function(subId) {
    return Number(this._GetFirstBracketedString(subId));
  },

  /**
   * Returns the first bracketed substring in the specified string.
   * @param {string} subId
   * @return {string}
   * @instance
   * @memberof oj.dvtBaseComponent
   * @protected
   */
  _GetFirstBracketedString : function(subId) {
    var start = subId.indexOf('[') + 1;
    var end = subId.indexOf(']');
    return subId.substring(start, end);
  },

  /**
   * Returns an object containing the top level options key and subkeys for
   * deferred data options.  'root' is used for top level keys. For example,
   * {'areaLayers': ['areaDataLayer/areas', 'areaDataLayer/markers'] indicates
   * that we should check this.options['areaLayers'][i]['areaDataLayer']['areas']
   * and this.options['areaLayers'][i]['areaDataLayer']['markers']. To indicate
   * a top level option, use the options key 'root', i.e. {'root': ['items']}.
   * @return {Object}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _GetComponentDeferredDataPaths : function() {
    return {};
  },

  /**
   * Returns an object containing the no clone paths for a component. For example,
   * {'areaLayers': {'areaDataLayer': {'areas': true, 'markers': true}}} indicates
   * that we should check this.options['areaLayers'][i]['areaDataLayer']['areas']
   * and this.options['areaLayers'][i]['areaDataLayer']['markers']. The base implementation
   * will handle the basic case where the deferred data path contains only top level data options.
   * @return {Object}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _GetComponentNoClonePaths : function() {
    if (!this._noClonePaths) {
      this._noClonePaths = {};
      var self = this;
      var rootPaths = this._GetComponentDeferredDataPaths()['root'];
      if (rootPaths) {
        rootPaths.forEach(function(path) {
          self._noClonePaths[path] = true;
        });
      }
    }
    return this._noClonePaths;
  },

  /**
   * Resolves the deferred data for a component and returns whether there is any deferred data to wait on.
   * @return {boolean} True if all data has been resolved
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _resolveDeferredDataItems : function() {
    // Reset stored options copy so we get the updated copy
    // Make a copy of options except for data options with the noClone parameter on DvtJsonUtils.
    // Cloning shouldn't be expensive if we're skipping data.
    this._optionsCopy =  dvt.JsonUtils.clone(this.options, null, this._GetComponentNoClonePaths());
    this._FixCustomRenderers(this._optionsCopy);
    this._numDeferredObjs = 0;

    var self = this;
    var paths = this._GetComponentDeferredDataPaths();
    // Do an initial loop to determine if we need to copy the options object
    for (var path in paths) {
      var subpaths = paths[path];
      subpaths.forEach(function(subpath) {
        if (path === 'root') {
          self._resolveDeferredDataItem.bind(self)(self.options, self._optionsCopy, subpath);
        } else {        
          var suboptions = self.options[path];
          // Deal with arrays suboptions that are arrays like legend's options['sections']
          if (suboptions && suboptions instanceof Array) {
            for (var i = 0; i < suboptions.length; i++)
              self._resolveDeferredDataItem.bind(self)(suboptions[i], self._optionsCopy[path][i], subpath);
          }
          /* DataProvider Uptake : Commented out for code coverage. Uncomment for chart/nbox support
          else if (suboptions && suboptions[subpath]) {  // Deal with arrays suboptions that are keys in an object like chart's options['data']['series']
              self._resolveDeferredDataItem.bind(self)(suboptions, self._optionsCopy[path], subpath);
          }
          */
        }
      });
    }
    return this._numDeferredObjs === 0;
  },

  /**
   * Resolves a deferred data item for a given options object and option path.
   * @param {Object} optionsFrom The option object to copy from
   * @param {Object} optionsTo The option object to copy to
   * @param {string} path  The option path to use
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _resolveDeferredDataItem : function(optionsFrom, optionsTo, path) {
    var optionPath = new DvtJsonPath(optionsFrom, path);
    var optionValue = optionPath.getValue();
    if (optionValue instanceof Function) {
      // Call functions and update the option value, creating a Promise from the result
      optionValue = Promise.resolve(optionValue(this._GetDataContext(optionsFrom)));
    }
    else if (this._IsCustomElement() && Array.isArray(optionValue)) {
      // For custom elements, the getter always should return a Promise for 
      // data options so we just always convert arrays to Promises in the setter
      optionValue = Promise.resolve(optionValue);
      optionPath.setValue(optionValue, true);
    }
    
    /* DataProvider Uptake : Commented out for code coverage. Uncomment for supporting DVTs
    if (optionValue && oj.DataProviderFeatureChecker.isIteratingDataProvider(optionValue)) {
      optionValue = new Promise(function(resolve, reject) 
      {
        optionValue['fetchFirst']({size: -1})[Symbol.asyncIterator]()['next']()
        .then(function(finalResult) {
          resolve(finalResult['value']['data']);
        });      
      });     
    };
    */
    
    if (optionValue instanceof Promise) {
      var renderCount = this._renderCount;
      var self = this;
      optionValue.then(
        function(value) {
          self._renderDeferredData(renderCount, optionsTo, path, value);
        },
        function(reason) {
          self._renderDeferredData(renderCount, optionsTo, path, []);
        }
      );
      this._numDeferredObjs++;
    }
  },

  /**
   * Checks to see if all deferred data promises have been resolved or rejected,
   * updates the options copy with the resolved value and renders the component when ready.
   * @param {number} renderCount The render count when this Promise was evaluated
   * @param {Object} optionsTo The option object to copy to
   * @param {string} path  The option path to use
   * @param {Object} value The value to update the option path with
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _renderDeferredData : function(renderCount, optionsTo, path, value) {
    if (renderCount === this._renderCount) {
      this._numDeferredObjs--;
      (new DvtJsonPath(optionsTo, path)).setValue(value, true);
      if (this._numDeferredObjs === 0) {
        this._RenderComponent(this._optionsCopy);
        this._optionsCopy = null;
      }
    }
  },

  /**
   * Renders the component.
   * @param {Object} options The options to render the component with
   * @param {boolean} isResize True if we are rendering due to a resize event.
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _RenderComponent : function(options, isResize) {
    // We do this check in _Render but since our rendering is asynchronous, the DOM state could have changed
    // in that time, e.g. when used inside an oj-switcher
    // Fix 18498656: If the component is not attached to a visible subtree of the DOM, rendering will fail because
    // getBBox calls will not return the correct values.
    // Note: Checking offsetParent() does not work here since it returns false for position: fixed.
    if (!this._context.isReadyToRender()) {
      this._renderNeeded = true;
      this._MakeReady();
    } else {
      // Cleanup
      this._CleanAllTemplates();

      // If flowing layout is supported, resize may happen during render, but we
      // don't want the resize listener to be triggered as it causes double render.
      // Thus we should remove the resize listener temporarily.
      var bRemoveResizeListener = this._IsFlowingLayoutSupported() && this._resizeListener;
      if (bRemoveResizeListener)
        this._removeResizeListener();

      this._component.render(isResize ? null : options, this._width, this._height);

      if (bRemoveResizeListener)
        this._addResizeListener();

      // Remove the tabindex from the element to disable keyboard handling if the component
      // does not have a role on the parent element like for non-interactive legends
      if (!this.element.attr("role"))
        this.element.attr("tabindex", null);
      else
        this.element.attr("tabindex", 0);
    }
  },
  
  /**
   * Returns a function that will handle events triggered on the component's 
   * dataProvider(s) and update the component
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   * @return {function(Event):void} A function that takes an event and updates a component
   * using the data from the data provider it is binded to.
   */
  /* DataProvider Uptake : Commented out for code coverage. Uncomment for supporting DVTs
  _GetDataProviderEventHandler : function(component, optionPath) {
    // subclasses may override
    return function(event) {
      if (event.type === "refresh" || event.type === "mutate") {
        var dataProvider = this;      
        // create a Promise that will resolve to the newly modified data
        var dataPromise = new Promise(function(resolve, reject) {
          dataProvider['fetchFirst']({size: -1})[Symbol.asyncIterator]()['next']()
          .then(function(finalResult) {
            resolve(finalResult['value']['data']);
          });      
        });     
        
        var newOptions = optionPath[0] == "root" ? {} : component.options[optionPath[0]];
        newOptions[optionPath[1]] = dataPromise;
        
        component._setOptions(newOptions);
      }
    };  
  },
  */
  
  /**
   * Checks for all potential dataProviders on the component and attaches event listeners
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  /* DataProvider Uptake : Commented out for code coverage. Uncomment for supporting DVTs
  _addDataProviderEventListeners: function()
  {
    var component = this;
    var paths = this._GetComponentDeferredDataPaths();
    var pathKeys = Object.keys(paths)

    for (var i = 0; i < pathKeys.length; i++) {
      var path = pathKeys[i]
      var subpathArray = paths[path];
      for (var j = 0; j < subpathArray.length; j++) {
        var subpath = subpathArray[j];
        var optionValue = (path === 'root') ? component.options[subpath] : (component.options[path] ? component.options[path][subpath] : null);
        if (optionValue && oj.DataProviderFeatureChecker.isIteratingDataProvider(optionValue)) {
          var dataProviderEventHandler = component._GetDataProviderEventHandler(component, [path, subpath]);
          optionValue.addEventListener("mutate", dataProviderEventHandler);
          optionValue.addEventListener("refresh", dataProviderEventHandler);
          
          component._dataProviderEventListeners.push({dataProvider: optionValue, listener:  dataProviderEventHandler});
        }
      };
    }
  },
  */
  
  /**
   * Checks for all potential dataProviders on the component and removes event listeners 
   * @private
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  /* DataProvider Uptake : Commented out for code coverage. Uncomment for supporting DVTs 
  _removeDataProviderEventListeners: function()
  {
    for (var i = 0; i < this._dataProviderEventListeners.length; i++) {
      var info = this._dataProviderEventListeners[i];
      var dataProvider = info.dataProvider;
      var dataProviderEventHandler = info.listener;
      dataProvider.removeEventListener("mutate", dataProviderEventHandler);
      dataProvider.removeEventListener("refresh", dataProviderEventHandler);
    }
    this._dataProviderEventListeners = [];
  },
  */
  
  /**
   * Returns the data context passed to data function callbacks.
   * @param {Object} options The options to retrieve the data context for
   * @return {Object}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _GetDataContext : function(options) {
    return {};
  },

  /**
   * Returns whether flowing layout is supported for the component.
   * If this returns true, the component will render at the preferred size of
   * the component if the user doesn't specify the width and height in the div.
   * If this returns false, the component will always render at the browser-
   * computed width and height.
   * @return {boolean}
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _IsFlowingLayoutSupported : function() {
    return false;
  },

  /**
   * Returns a promise that is resolved when the component is finished rendering.
   * This can be used to determine when it is okay to call automation and other APIs on the component.
   * @returns {Promise}
   * @ignore
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  whenReady : function() {
    if (this._ready)
      return Promise.resolve(true);
    if (!this._promise) {
      var self = this;
      this._promise = new Promise(function(resolve) {
        self._promiseResolve = resolve;
      });
    }
    return this._promise;
  },

  /**
   * Called by component to declare rendering is not finished. This method currently handles the ready state
   * for the component whenReady API, the page level BusyContext, and the static whenReady API for the custom element
   * version of this component.
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _NotReady : function() {
    // For component whenReady API
    this._ready = false;

    // For page level BusyContext
    // Only decrement ready state when there's no deferred data. Otherwise rendering will be blocked 
    // until all deferred data are resolved and we will only get one ready state increment.
    if (this._numDeferredObjs === 0) {
      // If we've already registered a busy state with the page's busy context, don't need to do anything further
      if (!this._readyResolveFunc) {
        var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
        var options = {'description' : "The component identified by '" + this.element.attr('id') + "' is being loaded."};
        this._readyResolveFunc = busyContext.addBusyState(options);
      }
    }
  },

  /**
   * Called by component to declare rendering is finished. This method currently handles the ready state
   * for the component whenReady API, the page level BusyContext, and the static whenReady API for the custom element
   * version of this component.
   * @protected
   * @instance
   * @memberof oj.dvtBaseComponent
   */
  _MakeReady : function() {
    // For component whenReady API
    if (this._promiseResolve) {
      this._promiseResolve(true);
      this._promiseResolve = null;
    }
    this._ready = true;
    this._promise = null;

    // For page level BusyContext
    if (this._readyResolveFunc) {
      this._readyResolveFunc();
      this._readyResolveFunc = null;
    }
  },

  /**
   * Sanitize options variables
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _ProcessOptions: function() {
    // Convert the tooltip to an object if the deprecated API structure is passed in
    var tooltip = this.options['tooltip'];
    if (tooltip && tooltip['_renderer']) {
      this.options['tooltip'] = {'renderer' : this._GetTemplateRenderer(tooltip['_renderer'], 'tooltip')};
    }
  },

  /**
   * Returns a wrapper function for custom elements that converts an object
   * returned by a custom renderer into an old format supported by widgets 
   * and toolkit code.
   * @param {Function} origRenderer Renderer function called to create custom content
   * @return {Function} A wrapper function that will used to convert result into toolkit format
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _WrapCustomElementRenderer: function (origRenderer) {
    var self = this;
    var customRenderer = function (context) {
      //add dvt context for the knockout template rendering
      context['_dvtcontext'] = self._context; 
      var obj = origRenderer(context);
      
      // template cleanup section - template name and cleanup function are passed through context
      if (context['_templateName'] && context['_templateCleanup']) {
        self._AddTemplate(context['_templateName'], context['_templateCleanup']);
      }
      
      //tooltip case: don't check 'insert' property, if 'preventDefault' is set to true
      if (obj && obj['preventDefault'] !== true && obj['insert'] ) {
        var insertObj = obj['insert'];
        // handle returns for knockout template renderer
        if (insertObj.classList && insertObj.classList.contains("oj-dvtbase")) {
          return self._GetDvtComponent(insertObj);
        }
        return insertObj;
      }
      else {
        return null;
      }
    }
    return customRenderer;
  },

  /**
   * Iterates through custom renderer options and replaces them with wrapper  
   * functions that supports format accepted by DVT Toolkit components
   * @param {Object} options Options for rendering the component
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _FixCustomRenderers: function (options) {
    if (this._IsCustomElement()) {
      var renderers = this._GetComponentRendererOptions();
      for(var i = 0; i < renderers.length; i++) {
        var optionPath = renderers[i];
        var path = new DvtJsonPath(options, optionPath);
        var value = path.getValue();
        if (value) {
          path.setValue(this._WrapCustomElementRenderer(value), true);
        }
      }
    }
  },

  /**
   * Returns an array of options that contain custom renderer paths 
   * for the given component.
   * @return {Array}
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _GetComponentRendererOptions: function() {
    return ['tooltip/renderer'];
  },

  /**
   * Creates a callback function that will be used as a custom renderer for a template that will be used directly.
   * @param {Function} templateFunction Template function used to render a knockout template
   * @param {string} templateName The name of the template
   * @return {Function} A function that will be used as a custom renderer
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _GetTemplateRenderer: function(templateFunction, templateName) {
    var self = this;
    var templateRenderer = function (context) {
      // Create a dummy div
      var dummyDiv = document.createElement("div");
      dummyDiv.style.display = "none";

      // Call the ko template renderer with the dummy div and context
      templateFunction({'parentElement': dummyDiv, 'context': context});

      var elem = dummyDiv.children[0];
      if (elem) {
        // Save a reference to cleanup function for the dummyDiv 
        // for ko cleanup to prevent memory leaks
        self._AddTemplate(templateName, function(){$(dummyDiv).remove();});

        dummyDiv.removeChild(elem);
        $(dummyDiv).remove();
        return elem;
      }
      return null;
    };
    return templateRenderer;
  },

  /**
   * Creates a callback function that will be used as a custom renderer for data items.
   * For components like ojDiagram and ojThematicMap that supports ko templates for stamping
   * data items, the component will reparent the svg element so we have special dummyDiv handling
   * for these cases.
   * @param {Function} templateFunction Template function used to render a knockout template
   * @param {string} templateName The name of the template
   * @return {Function} A function that will be used as a custom renderer
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _GetTemplateDataRenderer: function(templateFunction, templateName) {
    var self = this;
    var templateRenderer = function (context) {
      // Create a dummy div
      var dummyDiv = document.createElement("div");
      dummyDiv.style.display = "none";
      dummyDiv._dvtcontext = self._context;
      self.element.append(dummyDiv); // @HTMLUpdateOK

      // Call the ko template renderer with the dummy div and context
      templateFunction({'parentElement': dummyDiv, 'data': context['data']});

      var elem = dummyDiv.children[0];
      if (elem) {
        // Save a reference to cleanup function for the dummyDiv 
        // for ko cleanup to prevent memory leaks
        self._AddTemplate(templateName, function(){$(dummyDiv).remove();});

        // The dummy div can be removed for custom svg elements, but need to be
        // kept around for stamped DVTs so the oj components aren't removed.
        if (elem.namespaceURI === 'http://www.w3.org/2000/svg') {
          dummyDiv.removeChild(elem);
          $(dummyDiv).remove();
          return elem;
        } else {
          return self._GetDvtComponent(elem);
        }
      }
      return null;
    };
    return templateRenderer;
  },

  /**
   * Cleans all templates stored by the component
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _CleanAllTemplates: function() {
    for (var templateName in this._templateMap) {
      this._CleanTemplate(templateName);
    }    
    this._templateMap = {};
  },

  /**
   * Cleans a specific template stored by the component
   * @param {string} templateName The name of the template to clean
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _CleanTemplate: function(templateName) {
    if (this._templateMap[templateName]) {
      var length = this._templateMap[templateName].length;
      for (var i = 0; i < length; i++) {
        this._templateMap[templateName][i]();
      }
      this._templateMap[templateName] = [];
    }
  },

  /**
   * Adds a cleanup function that should be used to cleanup a specific template. 
   * @param {string} templateName The name of the template to clean
   * @param {Function} cleanupCallback A callback function to cleanup the template
   * @protected
   * @memberof oj.dvtBaseComponent
   */
  _AddTemplate: function(templateName, cleanupCallback) {
    if (!this._templateMap[templateName]) {
      this._templateMap[templateName] = [];
    }
    this._templateMap[templateName].push(cleanupCallback);    
  },

  //** @inheritdoc */
  _CompareOptionValues: function(option, value1, value2)
  {
    switch(option)
    {
      case 'hiddenCategories':
      case 'highlightedCategories':
      case 'selection':
        return oj.Object.compareValues(value1, value2);
      default:
        return this._super(option, value1, value2);
    }
  }

}, true);

/**
 * <p>The SVG DOM that this component generates should be treated as a black box, as it is subject to change.</p>
 *
 * @ojfragment warning
 * @memberof oj.dvtBaseComponent
 */

/**
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>The application is responsible for populating the shortDesc value in the
 * component properties object with meaningful descriptors when the component does
 * not provide a default descriptor.</p>
 *
 * @ojfragment a11y
 * @memberof oj.dvtBaseComponent
 */

/**
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>The application is responsible for populating the shortDesc value in the
 * component properties object with meaningful descriptors when the component does
 * not provide a default descriptor.  Since component terminology for keyboard
 * and touch shortcuts can conflict with those of the application, it is the
 * application's responsibility to provide these shortcuts, possibly via a help
 * popup.</p>
 *
 * @ojfragment a11yKeyboard
 * @memberof oj.dvtBaseComponent
 */

/**
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>
 *   As with any JET component, in the unusual case that the directionality (LTR or RTL) changes post-init, the
 *   component must be <code class="prettyprint">refresh()</code>ed.
 * </p>
 *
 * @ojfragment rtl
 * @memberof oj.dvtBaseComponent
 */

/**
 * <h4>Tracking Resize</h4>
 * <p>By default, the element will track resizes and render at the new size. This functionality adds a small
 * overhead to the initial render for simple elements like gauges or spark charts, which become noticable when
 * using large numbers of these simple elements. To disable resize tracking, set <code class="prettyprint">trackResize</code>
 * to <code class="prettyprint">off</code>. The application can manually request a re-render at any time by calling
 * the <code class="prettyprint">refresh</code> function.
 * </p>
 *
 * @ojfragment fragment_trackResize
 * @memberof oj.dvtBaseComponent
 */

(function() {

var dvtBaseComponentMeta = {
  "properties": {
    "trackResize": {
      "type": "string"
    },
    "translations": {
      "properties": {
        "labelAndValue": {
          "type": "string"
        },
        "labelClearSelection": {
          "type": "string"
        },
        "labelCountWithTotal": {
          "type": "string"
        },
        "labelDataVisualization": {
          "type": "string"
        },
        "labelInvalidData": {
          "type": "string"
        },
        "labelNoData": {
          "type": "string"
        },
        "stateCollapsed": {
          "type": "string"
        },
        "stateDrillable": {
          "type": "string"
        },
        "stateExpanded": {
          "type": "string"
        },
        "stateHidden": {
          "type": "string"
        },
        "stateIsolated": {
          "type": "string"
        },
        "stateMaximized": {
          "type": "string"
        },
        "stateMinimized": {
          "type": "string"
        },
        "stateSelected": {
          "type": "string"
        },
        "stateUnselected": {
          "type": "string"
        },
        "stateVisible": {
          "type": "string"
        }
      }
    }
  },
  "methods": {},
  "extension": {
    _WIDGET_NAME: "dvtBaseComponent",
    _DVT_PARSE_FUNC: DvtAttributeUtils.shapeParseFunction
  }
};
oj.CustomElementBridge.registerMetadata('dvtBaseComponent', 'baseComponent', dvtBaseComponentMeta);
})();
/**This file is generated. Do not edit directly. Actual file located in 3rdparty/dvt/prebuild.**/
/**
 * Copyright (c) 2016, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.dvtTimeComponent
 * @augments oj.dvtBaseComponent
 * @since 2.1.0
 * @ojtsimport ojvalidation-base
 * @ojtsimport ojvalidation-datetime
 * @ojtsimport ojtimeaxis
 * @abstract
 * @ojtsignore
 */
oj.__registerWidget('oj.dvtTimeComponent', $['oj']['dvtBaseComponent'],
{
  //** @inheritdoc */
  _GetEventTypes : function() 
  {
    return ['optionChange', 'viewportChange'];
  },

  //** @inheritdoc */
  _GetTranslationMap: function() 
  {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    var ret = this._super();
    // Add full month strings
    var monthNames = oj.LocaleData.getMonthNames("wide");
    ret['DvtUtilBundle.MONTH_JANUARY'] = monthNames[0];
    ret['DvtUtilBundle.MONTH_FEBRUARY'] = monthNames[1];
    ret['DvtUtilBundle.MONTH_MARCH'] = monthNames[2];
    ret['DvtUtilBundle.MONTH_APRIL'] = monthNames[3];
    ret['DvtUtilBundle.MONTH_MAY'] = monthNames[4];
    ret['DvtUtilBundle.MONTH_JUNE'] = monthNames[5];
    ret['DvtUtilBundle.MONTH_JULY'] = monthNames[6];
    ret['DvtUtilBundle.MONTH_AUGUST'] = monthNames[7];
    ret['DvtUtilBundle.MONTH_SEPTEMBER'] = monthNames[8];
    ret['DvtUtilBundle.MONTH_OCTOBER'] = monthNames[9];
    ret['DvtUtilBundle.MONTH_NOVEMBER'] = monthNames[10];
    ret['DvtUtilBundle.MONTH_DECEMBER'] = monthNames[11];

    // Add full day strings
    var dayNames = oj.LocaleData.getDayNames("wide");
    ret['DvtUtilBundle.DAY_SUNDAY'] = dayNames[0];
    ret['DvtUtilBundle.DAY_MONDAY'] = dayNames[1];
    ret['DvtUtilBundle.DAY_TUESDAY'] = dayNames[2];
    ret['DvtUtilBundle.DAY_WEDNESDAY'] = dayNames[3];
    ret['DvtUtilBundle.DAY_THURSDAY'] = dayNames[4];
    ret['DvtUtilBundle.DAY_FRIDAY'] = dayNames[5];
    ret['DvtUtilBundle.DAY_SATURDAY'] = dayNames[6];

    // Add abbreviated day strings
    var dayShortNames = oj.LocaleData.getDayNames("abbreviated");
    ret['DvtUtilBundle.DAY_SHORT_SUNDAY'] = dayShortNames[0];
    ret['DvtUtilBundle.DAY_SHORT_MONDAY'] = dayShortNames[1];
    ret['DvtUtilBundle.DAY_SHORT_TUESDAY'] = dayShortNames[2];
    ret['DvtUtilBundle.DAY_SHORT_WEDNESDAY'] = dayShortNames[3];
    ret['DvtUtilBundle.DAY_SHORT_THURSDAY'] = dayShortNames[4];
    ret['DvtUtilBundle.DAY_SHORT_FRIDAY'] = dayShortNames[5];
    ret['DvtUtilBundle.DAY_SHORT_SATURDAY'] = dayShortNames[6];

    return ret;
  },

  //** @inheritdoc */
  _HandleEvent: function(event)
  {
    var type = event['type'];
    if(type === 'viewportChange')
    {
      var viewportStart = new Date(event['viewportStart']).toISOString();
      var viewportEnd = new Date(event['viewportEnd']).toISOString();
      var minorAxisScale = event['minorAxisScale'];
      var viewportChangePayload = {
        'viewportStart': viewportStart,
        'viewportEnd': viewportEnd,
        'minorAxisScale': minorAxisScale
      };

      this._UserOptionChange('viewportStart', viewportStart);
      this._UserOptionChange('viewportEnd', viewportEnd);
      this._UserOptionChange('minorAxis.scale', minorAxisScale);
      this._trigger('viewportChange', null, viewportChangePayload);
    }
    else
    {
      this._super(event);
    }
  },

  //** @inheritdoc */
  _LoadResources: function()
  {
    // Ensure the resources object exists
    if (this.options['_resources'] == null)
      this.options['_resources'] = {};

    var resources = this.options['_resources'];

    // Create default converters
    var converterFactory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME);
    var secondsConverter = converterFactory.createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'});
    var minutesConverter = converterFactory.createConverter({'hour': 'numeric', 'minute': '2-digit'});
    var hoursConverter = converterFactory.createConverter({'hour': 'numeric'});
    var daysConverter = converterFactory.createConverter({'month': 'numeric', 'day': '2-digit'});
    var monthsConverter = converterFactory.createConverter({'month': 'long'});
    var yearsConverter = converterFactory.createConverter({'year': 'numeric'});

    var converter = {
      'seconds': secondsConverter,
      'minutes': minutesConverter,
      'hours': hoursConverter,
      'days': daysConverter,
      'weeks': daysConverter,
      'months': monthsConverter,
      'quarters': monthsConverter,
      'years': yearsConverter
    };

    resources['converter'] = converter;
    resources['converterFactory'] = converterFactory;

    // first day of week; locale specific
    resources['firstDayOfWeek'] = oj.LocaleData.getFirstDayOfWeek();
  }
});

(function() {
var dvtTimeComponentMeta = {
  "properties": {},
  "methods": {
    "getContextByNode": {}
  },
  "extension": {
    _WIDGET_NAME: "dvtTimeComponent"
  }
};
oj.CustomElementBridge.registerMetadata('dvtTimeComponent', 'dvtBaseComponent', dvtTimeComponentMeta);
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.dvtBaseGauge
 * @augments oj.dvtBaseComponent
 * @since 0.7
 * @abstract
 */
oj.__registerWidget('oj.dvtBaseGauge', $['oj']['dvtBaseComponent'],
{
  //** @inheritdoc */
  _ProcessStyles: function() {
    // The superclass evaluates the style classes, including those in _GetChildStyleClasses
    this._super();

    // Transfer the threshold colors to the correct location
    this.options['_thresholdColors'] = [this.options['_threshold1'], this.options['_threshold2'], this.options['_threshold3']];
    this.options['_threshold1'] = null;
    this.options['_threshold2'] = null;
    this.options['_threshold3'] = null;
  },
  //** @override */
  _AfterCreate: function() {
    this._super();
    var flags = {};
    flags['_context'] = {writeback: true, internalSet: true, readOnly: true};
    this.option("rawValue",this.options['value'], flags);
  },

  //** @inheritdoc */
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-gauge-metric-label'] = {'path' : 'metricLabel/style', 'property' : 'TEXT'};
    styleClasses['oj-gauge-tick-label'] = {'path' : 'tickLabel/style', 'property' : 'TEXT'};
    styleClasses['oj-gauge-threshold1'] = {'path' : '_threshold1', 'property' : 'color'};
    styleClasses['oj-gauge-threshold2'] = {'path' : '_threshold2', 'property' : 'color'};
    styleClasses['oj-gauge-threshold3'] = {'path' : '_threshold3', 'property' : 'color'};
    return styleClasses;
  },

  //** @inheritdoc */
  _GetEventTypes : function() {
    return ['input', 'optionChange'];
  },

  //** @inheritdoc */
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtGaugeBundle.EMPTY_TEXT'] = translations['labelNoData'];
    ret['DvtUtilBundle.GAUGE'] = translations['componentName'];
    return ret;
  },

  //** @inheritdoc */
  _HandleEvent : function(event) {
    var type = event['type'];
    if(type === 'valueChange') {
      var newValue = event['newValue'];
      if(event['complete'])
        this._UserOptionChange('value', newValue);
      else {
        // Fired during the value change interaction for each change
        this._trigger('input', null, {'value': newValue});
        this._UserOptionChange('rawValue', newValue);
      }
    }
    else {
      this._super(event);
    }
  },

  /**
   * @override
   * @private
   */
  _setOption : function (key, value, flags)
  {
    if (key === "rawValue") {
      // rawValue is a read-only option
      oj.Logger.error("'rawValue' is a read-only option and cannot be set");
      return;
    }

    if(key === "value") {
      var rawValueFlags = {};
      rawValueFlags['_context'] = {writeback: true, internalSet: true, readOnly: true};
      this.option("rawValue", value, rawValueFlags);
    }

    this._super(key, value, flags);
  },

  //** @inheritdoc */
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-dialgauge-tooltip' || subId == 'oj-ledgauge-tooltip' || subId == 'oj-ratinggauge-tooltip' || subId == 'oj-statusmetergauge-tooltip') {
      subId = 'tooltip';
    }
	if (subId == 'oj-ratinggauge-item' && locator['index'] != null)
	  subId = 'item[' + locator['index'] + ']';

    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  }
}, true);

/**
 * @ojcomponent oj.ojDialGauge
 * @ignore
 * @augments oj.dvtBaseGauge
 * @since 0.7
 *
 * @classdesc
 * <h3 id="dialGaugeOverview-section">
 *   JET Dial Gauge Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#dialGaugeOverview-section"></a>
 * </h3>
 *
 * <p>Dial gauge component for JET.  Dial gauges are used to display a metric value in relation to the minimum and
 * maximum possible values for that metric.</p>
 *
 * {@ojinclude "name":"warning"}
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;div data-bind="ojComponent: {
 *   component: 'ojDialGauge',
 *   value: 63, min: 0, max: 100,
 *   metricLabel: {rendered: 'on'}
 * }"/>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"rtl"}
 *
 * @desc Creates a JET Dial Gauge.
 * @example <caption>Initialize the Dial Gauge with no options specified:</caption>
 * $(".selector").ojDialGauge();
 *
 * @example <caption>Initialize the Dial Gauge with some options:</caption>
 * $(".selector").ojDialGauge({value: 63, min: 0, max: 100, metricLabel: {rendered: 'on'}});
 *
 * @example <caption>Initialize the Dial Gauge via the JET <code class="prettyprint">ojComponent</code> binding:</caption>
 * &lt;div data-bind="ojComponent: {component: 'ojDialGauge'}">
 */
oj.__registerWidget('oj.ojDialGauge', $['oj']['dvtBaseGauge'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * <p>The <code class="prettyprint">rawValue</code> is the read-only option for retrieving
     * the transient value from the dial gauge.</p>
     *
     * <p>This is a read-only option so page authors cannot set or change it directly.</p>
     * @expose
     * @instance
     * @type {?number|undefined}
     * @memberof oj.ojDialGauge
     * @since 1.2
     * @readonly
     */
     rawValue: undefined
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    this._focusable({'element': this.element, 'applyHighlight': true});
    return dvt.DialGauge.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId == 'tooltip') {
      locator['subId'] = 'oj-dialgauge-tooltip';
    }
    return locator;
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-dialgauge');
    return styleClasses;
  },

  //** @inheritdoc */
  _Render : function() {
    // Display the title of the surrounding div as the tooltip. Remove title from div to avoid browser default tooltip.
    if(this.element.attr('title'))
    {
      this.options['shortDesc'] =  this.element.attr('title');
      this.element.data( this.element,'title', this.element.attr('title'));
      this.element.removeAttr('title');
    }
    else if (this.element.data('title'))
      this.options['shortDesc'] =  this.element.data('title');

    // Set images for dial gauge
    this._setImages();

    // Call the super to render
    this._super();
  },

  /**
   * Applies image URLs to the options object passed into the dial gauge.
   * @private
   */
  _setImages: function() {
    // Pass the correct background image information set the default circleAlta and needleAlta.
    var backgroundImages = this.options['background'];
    if(backgroundImages == null) {
      backgroundImages = "circleAlta";
      this.options['background'] = "circleAlta";
    }
    var indicatorImages = this.options['indicator'];
    if(indicatorImages == null) {
      indicatorImages = "needleAlta"
      this.options['indicator'] = "needleAlta";
    }
    if(typeof backgroundImages === 'string') {
      var backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/alta-circle-200x200.png'), width: 200, height: 200},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/alta-circle-400x400.png'), width: 400, height: 400}];
      if(backgroundImages === "rectangleAlta") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/alta-rectangle-200x200.png'), width: 200, height: 154},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/alta-rectangle-400x400.png'), width: 400, height: 309}];
        }

      else if(backgroundImages === "domeAlta") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/alta-dome-200x200.png') , width: 200, height: 154},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/alta-dome-400x400.png') , width: 400, height: 309}];
      }

      else if(backgroundImages === "circleAntique") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/antique-circle-200x200.png'), width: 200, height: 200},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/antique-circle-400x400.png'), width: 400, height: 400}];
      }

      else if(backgroundImages === "rectangleAntique") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/antique-rectangle-200x200.png'), width: 200, height: 168},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/antique-rectangle-400x400.png'), width: 400, height: 335}];
      }

      else if(backgroundImages === "domeAntique") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/antique-dome-200x200.png'), width: 200, height: 176},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/antique-dome-400x400.png'), width: 400, height: 352}];
      }

      else if(backgroundImages === "circleLight") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/light-circle-200x200.png'), width: 200, height: 200},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/light-circle-400x400.png'), width: 400, height: 400}];
      }

      else if(backgroundImages === "rectangleLight") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/light-rectangle-200x200.png'), width: 200, height: 154},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/light-rectangle-400x400.png'), width: 400, height: 307}];
      }

      else if(backgroundImages === "domeLight") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/light-dome-200x200.png'), width: 200, height: 138},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/light-dome-400x400.png'), width: 400, height: 276}];
      }

      else if(backgroundImages === "circleDark") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/dark-circle-200x200.png'), width: 200, height: 200},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/dark-circle-400x400.png'), width: 400, height: 400}];
      }

      else if(backgroundImages === "rectangleDark") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/dark-rectangle-200x200.png'), width: 200, height: 154},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/dark-rectangle-400x400.png'), width: 400, height: 307}];
      }
      else if(backgroundImages === "domeDark") {
        backgroundInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/dark-dome-200x200.png'), width: 200, height: 138},
      {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/dark-dome-400x400.png'), width: 400, height: 276}];
      }
      this.options['_backgroundImages'] = backgroundInfo;
    }
    if(typeof indicatorImages === 'string') {
      var indicatorInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/alta-needle-1600x1600.png'),  width: 374, height: 575}];
      if(indicatorImages === "needleAntique") {
        indicatorInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/antique-needle-1600x1600.png'), width: 81, height: 734}];
        }

      else if(indicatorImages === "needleDark") {
        indicatorInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/dark-needle-1600x1600.png'),  width: 454, height: 652}];
      }

      else if(indicatorImages === "needleLight") {
        indicatorInfo = [{src: oj.Config.getResourceUrl('resources/internal-deps/dvt/gauge/light-needle-1600x1600.png'),  width: 454, height: 652}];
      }
      this.options['_indicatorImages'] = indicatorInfo;
    }
  },

  /**
   * Returns the gauge's metric label.
   * @return {Object} The metric label object
   * @expose
   * @instance
   * @memberof oj.ojDialGauge
   */
  getMetricLabel: function() {
    var auto = this._component.getAutomation();
    return auto.getMetricLabel();
  }
});


/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Drag</kbd></td>
 *       <td>Value change when <code class="prettyprint">readOnly</code> is <code class="prettyprint">false</code>.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojDialGauge
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Submit the current value of the gauge.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next component and submit the current value of the gauge.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous component.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Increase rawValue. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Decrease rawValue. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Decrease rawValue in left-to-right locales. Increase rawValue in right-to-left locales. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Increase rawValue in left-to-right locales. Decrease rawValue in right-to-left locales. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojDialGauge
 */

/**
 * The metric value.
 * @expose
 * @name value
 * @memberof oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The minimum value of the gauge.
 * @expose
 * @name min
 * @memberof oj.ojDialGauge
 * @instance
 * @type {number}
 * @default <code class="prettyprint">0</code>
 */
/**
 * The maximum value of the gauge.
 * @expose
 * @name max
 * @memberof oj.ojDialGauge
 * @instance
 * @type {number}
 * @default <code class="prettyprint">100</code>
 */
/**
 *  An object containing an optional callback function for tooltip customization. 
 * @expose
 * @name tooltip
 * @memberof oj.ojDialGauge
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  A function that returns a custom tooltip. The function takes a dataContext argument, provided by the gauge, with the following properties: <ul> <li>parentElement: The tooltip element. The function can directly modify or append content to this element.</li> <li>label: The computed metric label.</li> <li>component: The widget constructor for the gauge. The 'component' is bound to the associated jQuery element so can be called directly as a function.</li> </ul> The function may return an HTML element, which will be appended to the tooltip, or a tooltip string. 
 * @expose
 * @name tooltip.renderer
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {function(Object)}
 * @default null
 */
/**
 * An object or string defining the background specification for the gauge. Acceptable string options are: circleAlta, domeAlta, rectangleAlta, circleLight, domeLight, rectangleLight, circleDark, domeDark, rectangleDark, circleAntique, domeAntique, rectangleAntique
 * @expose
 * @name background
 * @memberof oj.ojDialGauge
 * @instance
 * @type {object|string}
 * @default null
 */
/**
 * An array of objects with the following properties, used to the define the image for the background. Multiple versions of the same image to be specified for different resolutions and for right to left locales, and the first image with enough detail for the requested resolution will be used.
 * @expose
 * @name background.images
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {Array.<Object>}
 * @default null
 */
/**
 * The URI specifying the location of the image resource.
 * @expose
 * @name background.images[].src
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of this image resource. The size of the first image is considered the reference size, upon which the anchor and other coordinates are based.
 * @expose
 * @name background.images[].width
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The height of this image resource. The size of the first image is considered the reference size, upon which the anchor and other coordinates are based.
 * @expose
 * @name background.images[].height
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * Specifies the text direction for which this image is used.
 * @expose
 * @name background.images[].dir
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "rtl"
 * @ojvalue {string} "ltr"
 * @default <code class="prettyprint">"ltr"</code>
 */
/**
 * The x coordinate of the indicator anchor point. Defaults to the center of the background.
 * @expose
 * @name background.anchorX
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The y coordinate of the indicator anchor point. Defaults to the center of the background.
 * @expose
 * @name background.anchorY
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The start angle of the dial in degrees.
 * @expose
 * @name background.startAngle
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default <code class="prettyprint">180</code>
 */
/**
 * The angular extent of the dial in degrees.
 * @expose
 * @name background.angleExtent
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default <code class="prettyprint">180</code>
 */
/**
 * The distance from the anchor to the center of the tick labels.
 * @expose
 * @name background.radius
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The height bound for the tick labels.
 * @expose
 * @name background.tickLabelHeight
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The width bound for the tick labels.
 * @expose
 * @name background.tickLabelWidth
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The number of ticks that have labels. By default, no ticks are drawn.
 * @expose
 * @name background.majorTickCount
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default <code class="prettyprint">0</code>
 */
/**
 * An object defining the bounds of the metric label. By default, the metric label is centered within the gauge.
 * @expose
 * @name background.metricLabelBounds
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The x coordinate of the bounding box.
 * @expose
 * @name background.metricLabelBounds.x
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The y coordinate of the bounding box.
 * @expose
 * @name background.metricLabelBounds.y
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The width of the bounding box.
 * @expose
 * @name background.metricLabelBounds.width
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The height of the bounding box.
 * @expose
 * @name background.metricLabelBounds.height
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The length of the indicator as a fraction of the background radius. Valid values are between 0 and 1.
 * @expose
 * @name background.indicatorLength
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default <code class="prettyprint">0.7</code>
 */
/**
 * An object or string defining the indicator specification for the gauge. Acceptable string options are: needleAlta, needleLight, needleDark, needleAntique
 * @expose
 * @name indicator
 * @memberof oj.ojDialGauge
 * @instance
 * @type {object|string}
 * @default null
 */
/**
 * An array of objects with the following properties, used to the define the image for the indicator. Multiple versions of the same image to be specified for different resolutions, and the first image with enough detail for the requested resolution will be used.
 * @expose
 * @name indicator.images
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {Array.<Object>}
 * @default null
 */
/**
 * The URI specifying the location of the image resource. The image must be provided with the indicator at 90 degrees (pointing up).
 * @expose
 * @name indicator.images[].src
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of this image resource. The size of the first image is considered the reference size, upon which the anchor and other coordinates are based.
 * @expose
 * @name indicator.images[].width
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The height of this image resource. The size of the first image is considered the reference size, upon which the anchor and other coordinates are based.
 * @expose
 * @name indicator.images[].height
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The x coordinate of the indicator anchor point. Defaults to the center of the indicator.
 * @expose
 * @name indicator.anchorX
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The y coordinate of the indicator anchor point. Defaults to the bottom of the indicator.
 * @expose
 * @name indicator.anchorY
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * An object defining the value label.
 * @expose
 * @name metricLabel
 * @memberof oj.ojDialGauge
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style string defining the style of the label.
 * @expose
 * @name metricLabel.style
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines if the label is rendered.
 * @expose
 * @name metricLabel.rendered
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * The scaling behavior of the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name metricLabel.scaling
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The converter used to format the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name metricLabel.converter
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * An object defining the dial tick labels.
 * @expose
 * @name tickLabel
 * @memberof oj.ojDialGauge
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style string defining the style of the label.
 * @expose
 * @name tickLabel.style
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Define the label to be displayed as number or as a percentage of the total value.
 * @expose
 * @name tickLabel.textType
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "percent"
 * @ojvalue {string} "number"
 * @default <code class="prettyprint">"number"</code>
 */
/**
 * Defines if the label is rendered.
 * @expose
 * @name tickLabel.rendered
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * The scaling behavior of the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name tickLabel.scaling
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The converter used to format the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name tickLabel.converter
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The duration of the animations, in milliseconds. Also accepts CSS strings such as 1s and 1000ms.
 * @expose
 * @ignore
 * @name animationDuration
 * @memberof oj.ojDialGauge
 * @instance
 * @type {number}
 * @default null
 */
/**
 * Defines the animation that is applied on data changes.
 * @expose
 * @name animationOnDataChange
 * @memberof oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "auto"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * Defines the animation that is shown on initial display.
 * @expose
 * @name animationOnDisplay
 * @memberof oj.ojDialGauge
 * @instance
 * @type {string}
 * @ojvalue {string} "auto"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * Defines whether the value of the gauge can be changed by the end user.
 * @expose
 * @name readOnly
 * @memberof oj.ojDialGauge
 * @instance
 * @type {boolean}
 * @default <code class="prettyprint">true</code>
 */
 
// SubId Locators **************************************************************

/**
 * <p>Sub-ID for the the dial guage tooltip.</p>
 * 
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 * 
 * @ojsubid
 * @member
 * @name oj-dialgauge-tooltip
 * @memberof oj.ojDialGauge
 * @instance
 * 
 * @example <caption>Get the tooltip object of the gauge, if displayed:</caption>
 * var nodes = $( ".selector" ).ojDialGauge( "getNodeBySubId", {'subId': 'oj-dialgauge-tooltip'} );
 */

/**
 * The knockout template used to render the content of the tooltip.
 *
 * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a
 * component option. The following variables are also passed into the template:
 *  <ul> 
 *    <li>parentElement: The tooltip element. This can be used to change the tooltip border or background color.</li> 
 *    <li>label: The computed metric label.</li> 
 *    <li>component: The widget constructor for the gauge. The 'component' is bound to the associated jQuery element so can be called directly as a function.</li> 
 *  </ul>
 *
 * @ojbindingonly
 * @name tooltip.template
 * @memberof! oj.ojDialGauge
 * @instance
 * @type {string|null}
 * @default null
 */
/**
 * @ojcomponent oj.ojLedGauge
 * @augments oj.dvtBaseGauge
 * @since 0.7
 * @ojstatus preview
 * @ojshortdesc Displays information graphically, highlighting a specific metric value in relation to its thresholds.
 * @ojrole img
 * @ojrole application
 *
 * @classdesc
 * <h3 id="ledGaugeOverview-section">
 *   JET LED Gauge
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#ledGaugeOverview-section"></a>
 * </h3>
 *
 * <p>LED gauges are used to highlight a specific metric value in relation to its
 * thresholds.<p>
 *
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-led-gauge
 *   value='63' 
 *   min='0' 
 *   max='100'
 *   thresholds='[{"max": 33}, {"max": 67}, {}]'>
 * &lt;/oj-led-gauge>
 * </code>
 * </pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"a11y"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojLedGauge', $['oj']['dvtBaseGauge'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * The border color of the gauge. Only applies when thresholds are not defined. The default value varies based on theme.
     * @expose
     * @name borderColor
     * @memberof oj.ojLedGauge
     * @instance
     * @type {string}
     */
    borderColor: '',

    /**
     * The color of the gauge. Only applies when thresholds are not defined. The default value varies based on theme.
     * @expose
     * @name color
     * @memberof oj.ojLedGauge
     * @instance
     * @type {string}
     */
    color: '#393737',

    /**
     * An object defining the label.
     * @expose
     * @name label
     * @memberof oj.ojLedGauge
     * @instance
     * @type {Object}
     * @default {"style": {}, "text": ""}
     */
    label: {
      /**
       * The CSS style object defining the style of the label.
       * @expose
       * @name label.style
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      style: {},

      /**
       * The text for the label.
       * @expose
       * @name label.text
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       * @ojtranslatable
       */
      text: ''
    },

    /**
     * An object defining the value label.
     * @expose
     * @name metricLabel
     * @memberof oj.ojLedGauge
     * @instance
     * @type {Object}
     * @default {"converter": null, "rendered": "off", "scaling": "auto", "style": {}, "text": "", "textType": "number"}
     */
    metricLabel: {
      /**
       * The converter used to format the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
       * @expose
       * @name metricLabel.converter
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?Converter<string>"}
       * @default null
       */
      converter: null,

      /**
       * Defines if the label is rendered.
       * @expose
       * @name metricLabel.rendered
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "on"
       * @ojvalue {string} "off"
       * @default "off"
       */
      rendered: 'off',

      /**
       * The scaling behavior of the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
       * @expose
       * @name metricLabel.scaling
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "none"
       * @ojvalue {string} "thousand"
       * @ojvalue {string} "million"
       * @ojvalue {string} "billion"
       * @ojvalue {string} "trillion"
       * @ojvalue {string} "quadrillion"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      scaling: 'auto',

      /**
       * The CSS style object defining the style of the label.
       * @expose
       * @name metricLabel.style
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      style: {},

      /**
       * The text for the label. If specified, text will overwrite the numeric value that is displayed by default. The converter, scaling, and textType attributes are ignored when text is specified.
       * @expose
       * @name metricLabel.text
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       * @ojtranslatable
       */
      text: '',

      /**
       * Defines whether the label is a number or a percentage of the total value.
       * @expose
       * @name metricLabel.textType
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "percent"
       * @ojvalue {string} "number"
       * @default "number"
       */
      textType: 'number'
    },

    /**
     * The maximum value of the gauge.
     * @expose
     * @name max
     * @memberof oj.ojLedGauge
     * @instance
     * @type {number}
     * @default 100
     */
    max: 100,

    /**
     * The minimum value of the gauge.
     * @expose
     * @name min
     * @memberof oj.ojLedGauge
     * @instance
     * @type {number}
     * @default 0
     */
    min: 0,

    /**
     * The rotation angle for the gauge. Useful for changing the direction of triangle or arrow gauges.
     * @expose
     * @name rotation
     * @memberof oj.ojLedGauge
     * @instance
     * @type {number}
     * @ojvalue {number} 90
     * @ojvalue {number} 180
     * @ojvalue {number} 270
     * @ojvalue {number} 0
     * @ojunits degrees
     * @default 0
     */
    rotation: 0,

    /**
     * Fraction of area to use. Values range from 0 to 1.
     * @expose
     * @name size
     * @memberof oj.ojLedGauge
     * @instance
     * @type {number}
     * @default 1
     * @ojmin 0
     * @ojmax 1
     */
    size: 1,

    /**
     * The CSS style class to apply to the gauge. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name svgClassName
     * @memberof oj.ojLedGauge
     * @instance
     * @type {string}
     * @default ""
     */
    svgClassName: '',

    /**
     * The inline style to apply to the gauge. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name svgStyle
     * @memberof oj.ojLedGauge
     * @instance
     * @type {Object}
     * @default {}
     */
    svgStyle: {},

    /**
     * An array of objects with the following properties defining the thresholds for the gauge.
     * @expose
     * @name thresholds
     * @memberof oj.ojLedGauge
     * @instance
     * @type {Array.<Object>}
     * @ojsignature {target: "Type", value: "Array<oj.ojLedGauge.Threshold>", jsdocOverride: true}
     * @default []
     */
    thresholds: [],

    /**
     *  An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojLedGauge
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     */
    tooltip: {
      /**
       * A function that returns a custom tooltip. The function takes a tooltip context argument,
       * provided by the gauge, and should return an object that contains only one of the two properties:
       *  <ul>
       *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
       *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
       *  </ul>
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojLedGauge
       * @instance
       * @type {function(Object):Object|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojLedGauge.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))", jsdocOverride: true}
       * @default null
       */
      renderer: null
    },

    /**
     * The shape of the LED gauge. Can take the name of a built-in shape or the svg path commands for a custom shape. 
     * @expose
     * @name type
     * @memberof oj.ojLedGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "arrow"
     * @ojvalue {string} "diamond"
     * @ojvalue {string} "square"
     * @ojvalue {string} "rectangle"
     * @ojvalue {string} "triangle"
     * @ojvalue {string} "star"
     * @ojvalue {string} "human"
     * @ojvalue {string} "circle"
     * @default "circle"
     */
    type: 'circle',

    /**
     * The metric value.
     * @expose
     * @name value
     * @memberof oj.ojLedGauge
     * @instance
     * @type {number|null}
     * @ojwriteback
     */
    value: null,

    /**
     * Defines whether visual effects such as overlays are applied to the gauge.
     * @expose
     * @name visualEffects
     * @memberof oj.ojLedGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "auto"
     * @default "auto"
     */
    visualEffects: 'auto'
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    this._focusable({'element': this.element, 'applyHighlight': true});
    return dvt.LedGauge.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId == 'tooltip') {
      locator['subId'] = 'oj-ledgauge-tooltip';
    }
    return locator;
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-ledgauge');
    return styleClasses;
  },

  //** @inheritdoc */
  _Render : function() {
    // Display the title of the surrounding div as the tooltip. Remove title from div to avoid browser default tooltip.
    if(this.element.attr('title'))
    {
      this.options['shortDesc'] =  this.element.attr('title');
      this.element.data( this.element,'title', this.element.attr('title'));
      this.element.removeAttr('title');
    }
    else if (this.element.data('title'))
      this.options['shortDesc'] =  this.element.data('title');

    // Call the super to render
    this._super();
  },

  /**
   * Returns the gauge's formatted metric label.
   * @return {string} The formatted metric label.
   * @expose
   * @instance
   * @memberof oj.ojLedGauge
   */
  getMetricLabel: function() {
    var auto = this._component.getAutomation();
    return auto.getMetricLabel();
  }
});

/**
 * <p>This element has no touch interaction.  </p>
 *
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojLedGauge
 */

/**
 * <p>This element has no keyboard interaction.  </p>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojLedGauge
 */


// TYPEDEFS

/**
 * @typedef {Object} oj.ojLedGauge.Threshold
 * @property {string} [borderColor] The border color of the threshold.
 * @property {string} [color] The color of the threshold.
 * @property {number} [max] The upper bound of the threshold. This value is ignored for the final threshold, which uses the maximum value of the gauge.
 * @property {string} [shortDesc] Specific description for the threshold and overwrites the shortDesc specified on gauge. This is used for accessibility and also for customizing the tooltip text.
 */
/**
 * @typedef {Object} oj.ojLedGauge.TooltipContext
 * @property {string} color The indicator color of the gauge.
 * @property {Element} componentElement The LED gauge HTML element.
 * @property {string} label The computed metric label.
 * @property {Element} parentElement The tooltip element. The function can directly modify or append content to this element.
 */


// SubId Locators **************************************************************

/**
 * <p>Sub-ID for the the LED gauge tooltip.</p>
 * 
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 * 
 * @ojsubid
 * @member
 * @name oj-ledgauge-tooltip
 * @memberof oj.ojLedGauge
 * @instance
 * 
 * @example <caption>Get the tooltip object of the gauge, if displayed:</caption>
 * var nodes = myLedGauge.getNodeBySubId({'subId': 'oj-ledgauge-tooltip'});
 */
/**
 * @ojcomponent oj.ojRatingGauge
 * @augments oj.dvtBaseGauge
 * @since 0.7
 * @ojstatus preview
 * @ojshortdesc Displays information graphically, typically displaying or accepting user feedback on a product or service.
 * @ojrole img
 * @ojrole application
 *
 * @classdesc
 * <h3 id="ratingGaugeOverview-section">
 *   JET Rating Gauge
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#ratingGaugeOverview-section"></a>
 * </h3>
 *
 * <p>Rating gauges are typically used to display or accept user feedback on a product
 * or service.</p>
 *
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-rating-gauge
 *   value='4' >
 * &lt;/oj-rating-gauge>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojRatingGauge', $['oj']['dvtBaseGauge'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * Whether there has been a value entered by the user.
     * @expose
     * @name changed
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {boolean}
     * @default false
     * @ojwriteback
     */
    changed: false,

    /**
     * The changed shape for the gauge. Displayed after the user has set a value, or when the changed attribute of the data object is set to true.
     * @expose
     * @name changedState
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {Object}
     * @default {"borderColor": "", "color": "#ED2C02", "shape": "star", "source": "", "svgClassName": "", "svgStyle": {}}
     */
    changedState: {
      /**
       * The border color for changed state. Does not apply if a custom image is specified.
       * @expose
       * @name changedState.borderColor
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      borderColor: '',

      /**
       * The color for changed state. Does not apply if a custom image is specified.
       * @expose
       * @name changedState.color
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      color: '#ED2C02',

      /**
       * The shape to be used. Can take the name of a built-in shape or the svg path commands for a custom shape. Does not apply if a custom image is specified.
       * @expose
       * @name changedState.shape
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?'circle'|'diamond'|'human'|'square'|'star'|'triangle'|string", jsdocOverride: true}
       * @default "star"
       */
      shape: 'star',

      /**
       * The URI of the custom image. If specified, it takes precedence over shape. For SVG images, the width and height must be defined on the SVG element as pixels. 
       * @expose
       * @name changedState.source
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      source: '',

      /**
       * The CSS style class to apply to the changed state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name changedState.svgClassName
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      svgClassName: '',

      /**
       * The inline style to apply to the changed state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name changedState.svgStyle
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {}
    },

    /**
     * The shape that displays on hover.
     * @expose
     * @name hoverState
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {Object}
     * @default {"borderColor": "", "color": "#007CC8", "shape": "star", "source": "", "svgClassName": "", "svgStyle": {}}
     */
    hoverState: {
      /**
       * The border color for hover state. Does not apply if a custom image is specified.
       * @expose
       * @name hoverState.borderColor
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      borderColor: '',

      /**
       * The color for hover state. Does not apply if a custom image is specified.
       * @expose
       * @name hoverState.color
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      color: '#007CC8',

      /**
       * The shape to be used. Can take the name of a built-in shape or the svg path commands for a custom shape. Does not apply if a custom image is specified.
       * @expose
       * @name hoverState.shape
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?'circle'|'diamond'|'human'|'square'|'star'|'triangle'|string", jsdocOverride: true}
       * @default "star"
       */
      shape: 'star',

      /**
       * The URI of the custom image. If specified, it takes precedence over shape. For SVG images, the width and height must be defined on the SVG element as pixels. 
       * @expose
       * @name hoverState.source
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      source: '',

      /**
       * The CSS style class to apply to the hover state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name hoverState.svgClassName
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      svgClassName: '',

      /**
       * The inline style to apply to the hover state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name hoverState.svgStyle
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {}
    },

    /**
     * Integer value specifying the maximum value of the gauge, which determines the number of shapes or images that are displayed.
     * @expose
     * @name max
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {number}
     * @default 5
     * @ojmin 0
     */
    max: 5,

    /**
     * The minimum value that can be set on the gauge by the end user. Does not affect the value set on the gauge by API.
     * @expose
     * @name min
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {number}
     * @default 0
     * @ojmin 0
     */
    min: 0,
    
    /**
     * Defines the type of rating gauge to be rendered.
     * @expose
     * @name orientation
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "vertical"
     * @ojvalue {string} "horizontal"
     * @default "horizontal"
     */
    orientation: 'horizontal',

    /**
     * Specifies whether the images provided should show up at their defined aspect ratio. With 'none', the space is allocated evenly, and shapes could be stretched. With 'meet', The aspect ratio of the shape or image is taken into account when space is allocated. When aspect ratios conflict, the aspect ratio of the selectedState will be used.
     * @expose
     * @name preserveAspectRatio
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "meet"
     * @default "meet"
     */
    preserveAspectRatio: 'meet',

    /**
     * <p>The <code class="prettyprint">transientValue</code> is the read-only property for retrieving
     * the transient value from the rating gauge. It is triggered when hovering over the rating gauge.</p>
     *
     * <p>This is a read-only property so page authors cannot set or change it directly.</p>
     * @expose
     * @alias transientValue
     * @instance
     * @type {number|null}
     * @memberof oj.ojRatingGauge
     * @since 4.2.0
     * @ojstatus preview
     * @readonly
     * @ojwriteback
     */
    rawValue: null,

    /**
     * Defines whether the value of the gauge can be changed by the end user.
     * @expose
     * @name readonly
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {boolean}
     * @default false
     */
    readonly: false,

    /**
     * The selected shape for the gauge.
     * @expose
     * @name selectedState
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {Object}
     * @default {"borderColor": "", "color": "#F8C15A", "shape": "star", "source": "", "svgClassName": "", "svgStyle": {}}
     */
    selectedState: {
      /**
       * The border color for selected state. Does not apply if a custom image is specified.
       * @expose
       * @name selectedState.borderColor
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      borderColor: '',

      /**
       * The color for selected state. Does not apply if a custom image is specified.
       * @expose
       * @name selectedState.color
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      color: '#F8C15A',

      /**
       * The shape to be used. Can take the name of a built-in shape or the svg path commands for a custom shape. Does not apply if a custom image is specified.
       * @expose
       * @name selectedState.shape
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?'circle'|'diamond'|'human'|'square'|'star'|'triangle'|string", jsdocOverride: true}
       * @default "star"
       */
      shape: 'star',

      /**
       * The URI of the custom image. If specified, it takes precedence over shape. For SVG images, the width and height must be defined on the SVG element as pixels. 
       * @expose
       * @name selectedState.source
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      source: '',

      /**
       * The CSS style class to apply to the selected state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name selectedState.svgClassName
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      svgClassName: '',

      /**
       * The inline style to apply to the selected state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name selectedState.svgStyle
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {}
    },

    /**
     * Specifies the increment by which values can be specified by the end user.
     * @expose
     * @name step
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {number}
     * @ojvalue {number} 0.5
     * @ojvalue {number} 1
     * @default 1
     */
    step: 1,

    /**
     * An array of objects with the following properties defining the thresholds for the gauge.
     * @expose
     * @name thresholds
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {Array.<Object>}
     * @ojsignature {target: "Type", value: "Array<oj.ojRatingGauge.Threshold>", jsdocOverride: true}
     * @default []
     */
    thresholds: [],

    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     */
    tooltip: {
      /**
       * A function that returns a custom tooltip. The function takes a tooltip context argument,
       * provided by the gauge, and should return an object that contains only one of the two properties:
       *  <ul>
       *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
       *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
       *  </ul>
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {function(Object):Object|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojRatingGauge.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))", jsdocOverride: true}
       * @default null
       */
      renderer: null
    },

    /**
     * The unselected shape for the gauge.
     * @expose
     * @name unselectedState
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {Object}
     * @default {"borderColor": "", "color": "#C4CED7", "shape": "star", "source": "", "svgClassName": "", "svgStyle": {}}
     */
    unselectedState: {
      /**
       * The border color for unselected state. Does not apply if a custom image is specified.
       * @expose
       * @name unselectedState.borderColor
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      borderColor: '',

      /**
       * The color for unselected state. Does not apply if a custom image is specified.
       * @expose
       * @name unselectedState.color
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      color: '#C4CED7',

      /**
       * The shape to be used. Can take the name of a built-in shape or the svg path commands for a custom shape. Does not apply if a custom image is specified.
       * @expose
       * @name unselectedState.shape
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?'circle'|'diamond'|'human'|'square'|'star'|'triangle'|string", jsdocOverride: true}
       * @default "star"
       */
      shape: 'star',

      /**
       * The URI of the custom image. If specified, it takes precedence over shape. For SVG images, the width and height must be defined on the SVG element as pixels. 
       * @expose
       * @name unselectedState.source
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      source: '',

      /**
       * The CSS style class to apply to the unselected state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name unselectedState.svgClassName
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      svgClassName: '',

      /**
       * The inline style to apply to the unselected state. The style class and inline style will override any other styling specified through the properties. Does not apply if custom image is specified.
       * @expose
       * @name unselectedState.svgStyle
       * @memberof! oj.ojRatingGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {}
    },

    /**
     * The value set on the gauge.
     * @expose
     * @name value
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {number|null}
     * @ojwriteback
     * @ojmin 0
     */
    value: null,

    /**
     * Defines whether visual effects such as overlays are applied to the gauge.
     * @expose
     * @name visualEffects
     * @memberof oj.ojRatingGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "auto"
     * @default "auto"
     */
    visualEffects: 'auto'
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    this._focusable({'element': this.element, 'applyHighlight': true});
    return dvt.RatingGauge.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId == 'tooltip') {
      locator['subId'] = 'oj-ratinggauge-tooltip';
    }
	else if (subId.indexOf('item') == 0){
      locator['subId'] = 'oj-ratinggauge-item';
      locator['index'] = this._GetFirstIndex(subId);
    }
    return locator;
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-ratinggauge');
    //TODO  Add style classes for rating gauge selected/hover/unselected/changed
    return styleClasses;
  },

  //** @inheritdoc */
  _Render : function() {
    // Display the title of the surrounding div as the tooltip. Remove title from div to avoid browser default tooltip.
    if(this.element.attr('title'))
    {
      this.options['shortDesc'] =  this.element.attr('title');
      this.element.data( this.element,'title', this.element.attr('title'));
      this.element.removeAttr('title');
    }
    else if (this.element.data('title'))
      this.options['shortDesc'] =  this.element.data('title');

    // Call the super to render
    this._super();
  },

  //** @inheritdoc */
  _UserOptionChange : function(key, value) {
    this._superApply(arguments);

    // If this was a value change, also update the changed value
    if(key == 'value')
      this._UserOptionChange('changed', true);
  }
});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Drag</kbd></td>
 *       <td>Value change when <code class="prettyprint">readonly</code> is <code class="prettyprint">false</code>.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojRatingGauge
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Submit the current value of the gauge.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element and submit the current value of the gauge.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Increase the gauge's transient value. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Decrease the gauge's transient value. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Decrease the gauge's transient value in left-to-right locales. Increase the gauge's transient value in right-to-left locales. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Increase the gauge's transient value in left-to-right locales. Decrease the gauge's transient value in right-to-left locales. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojRatingGauge
 */


// TYPEDEFS

/**
 * @typedef {Object} oj.ojRatingGauge.Threshold
 * @property {string} [borderColor] The border color of the threshold.
 * @property {string} [color] The color of the threshold.
 * @property {number} [max] The upper bound of the threshold. This value is ignored for the final threshold, which uses the maximum value of the gauge.
 * @property {string} [shortDesc] Specific description for the threshold and overwrites the shortDesc specified on gauge. This is used for accessibility and also for customizing the tooltip text.
 */
/**
 * @typedef {Object} oj.ojRatingGauge.TooltipContext
 * @property {string} color The indicator color of the gauge.
 * @property {Element} componentElement The rating gauge HTML element.
 * @property {string} label The computed metric label.
 * @property {Element} parentElement The tooltip element. The function can directly modify or append content to this element.
 */


// DEPRECATED OPTIONS FOR WIDGET SYNTAX

/**
 * Defines whether the value of the gauge can be changed by the end user.
 * @ignore
 * @expose
 * @name readOnly
 * @memberof oj.ojRatingGauge
 * @instance
 * @type {boolean}
 * @default true
 */


// SubId Locators **************************************************************

/**
 * <p>Sub-ID for the the rating guage tooltip.</p>
 * 
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 * 
 * @ojsubid
 * @member
 * @name oj-ratinggauge-tooltip
 * @memberof oj.ojRatingGauge
 * @instance
 * 
 * @example <caption>Get the tooltip object of the gauge, if displayed:</caption>
 * var nodes = myRatingGauge.getNodeBySubId('subId': 'oj-ratinggauge-tooltip'});
 */
 
/**
 * <p>Sub-ID for a rating gauge item indexed by its position.</p>
 *
 * @property {number} index The index of the item within the gauge.
 *
 * @ojsubid oj-ratinggauge-item
 * @memberof oj.ojRatingGauge
 *
 * @example <caption>Get the first item from the rating gauge:</caption>
 * var nodes = myRatingGauge.getNodeBySubId('subId': 'oj-ratinggauge-item', index: 0});
 */
/**
 * @ojcomponent oj.ojStatusMeterGauge
 * @augments oj.dvtBaseGauge
 * @since 0.7
 * @ojstatus preview
 * @ojshortdesc Displays information graphically, highlighting a specific metric value's progress in relation to its thresholds.  Horizontal, vertical, and circular formats are supported.
 * @ojrole img
 * @ojrole application
 *
 * @classdesc
 * <h3 id="statusMeterGaugeOverview-section">
 *   JET Status Meter Gauge
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#statusMeterGaugeOverview-section"></a>
 * </h3>
 *
 * <p>Status meter gauges support horizontal and circular status meters.</p>
 *
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-status-meter-gauge
 *   value='63' 
 *   min='0' 
 *   max='100'
 *   thresholds='[{"max": 33}, {"max": 67}, {}]'>
 * &lt;/oj-status-meter-gauge>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojStatusMeterGauge', $['oj']['dvtBaseGauge'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * Specifies the angle extent of a gauge with circular orientation. Value should be provided in degrees.
     * @expose
     * @name angleExtent
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number}
     * @default 360
     * @ojunits degrees
     * @ojmin 0
     * @ojmax 360
     */
    angleExtent: 360,

    /**
     * Defines the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDataChange: 'none',

    /**
     * Defines the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDisplay: 'none',

    /**
     * The duration of the animations, in milliseconds. The default value comes from the CSS and varies based on theme.
     * @expose
     * @name animationDuration
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number}
     * @ojsignature {target: "Type", value: "?"}
     * @ojunits milliseconds
     * @ojmin 0
     */
    animationDuration: undefined,

    /**
     * The border color of the gauge. Only applies when thresholds are not defined. The default value varies based on theme.
     * @expose
     * @name borderColor
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     */
    borderColor: '',

    /**
     * Defines the border radius of the indicator and plot area. When set to "auto", the border radius is set to a built-in default. Acceptable input follows CSS border-radius attribute specifications. The plot area border radius can be overwritten with the plotArea borderRadius atribute.
     * @expose
     * @name borderRadius
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     * @default "auto"
     */
    borderRadius: 'auto',

    /**
     * An object defining the center content of a status meter with circular orientation. 
     * @expose
     * @name center
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     */
    center: {
      /**
       * A function that returns custom center content. The function takes a center context argument,
       * provided by the gauge, and should return an object with the following property: 
       * <ul>
       *   <li>insert: HTMLElement - HTML element, which will be overlaid on top of the gauge. 
       *   This HTML element will block interactivity of the gauge by default, but the CSS pointer-events property 
       *   can be set to 'none' on this element if the gauge's interactivity is desired. 
       *   </li>
       * </ul>
       * @expose
       * @name center.renderer
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {function(Object):Object|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojStatusMeterGauge.CenterContext) => ({insert: Element|string}|{preventDefault: boolean}))", jsdocOverride: true}
       * @default null
       */
      renderer: null
    },

    /**
     * The color of the gauge. Only applies when thresholds are not defined. The default value varies based on theme.
     * @expose
     * @name color
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     */
    color: '#393737',

    /**
     * Defines the ratio of relative thickness of the indicator to the plot area.
     * @expose
     * @name indicatorSize
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number}
     * @default 1
     * @ojmin 0
     */
    indicatorSize: 1,

    /**
     * Specifies the inner radius of a gauge with circular orientation, defined by the distance from the center of the gauge to the innermost edge of the indicator and plot area. Valid values are a percent or ratio from 0 to 1.
     * @expose
     * @name innerRadius
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number}
     * @default .7
     * @ojmin 0
     * @ojmax 1
     */
    innerRadius: .7,

    /**
     * An object defining the label.
     * @expose
     * @name label
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Object}
     * @default {"position": "auto", "style": {}, "text": ""}
     */
    label: {
      /**
       * Defines the position of the label for horizontal and vertical gauges. The default position for horizontal gauges is 'start' and for vertical gauges is 'center'.
       * @expose
       * @name label.position
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "center"
       * @ojvalue {string} "start"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      position: 'auto',

      /**
       * The CSS style object defining the style of the label.
       * @expose
       * @name label.style
       * @ojsignature {target: "Type", value: "?"}
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {Object}
       * @default {}
       */
      style: {},

      /**
       * The text for the label.
       * @expose
       * @name label.text
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       * @ojtranslatable
       */
      text: ''
    },

    /**
     * The maximum value of the gauge.
     * @expose
     * @name max
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number}
     * @default 100
     */
    max: 100,

    /**
     * An object defining the value label.
     * @expose
     * @name metricLabel
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Object}
     * @default {"position": "auto", "rendered": "auto", "scaling": "auto", "style": {}, "text": "", "textType": "number"}
     */
    metricLabel: {
      /**
       * The converter used to format the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
       * @expose
       * @name metricLabel.converter
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?Converter<string>"}
       * @default null
       */
      converter: null,

      /**
       * Defines the position of the metric label for horizontal and vertical gauges. The default position of the metric label is outside of the plot area. If the label is not rendered, then 'withLabel' will render the metric label outside the plot area. When the label is rendered, all positions are treated as 'withLabel' except 'auto' and 'outsidePlotArea' which render the metric label outside the plot area. When the metric label is rendered 'withLabel', the metric label is displayed with the same style as the label. The position in the 'withLabel' case is specified by the label position attribute.
       * @expose
       * @name metricLabel.position
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "center"
       * @ojvalue {string} "insideIndicatorEdge"
       * @ojvalue {string} "outsideIndicatorEdge"
       * @ojvalue {string} "outsidePlotArea"
       * @ojvalue {string} "withLabel"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      position: 'auto',

      /**
       * Defines if the label is rendered. If set to auto, the label is rendered if the orientation is circular.
       * @expose
       * @name metricLabel.rendered
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "on"
       * @ojvalue {string} "off"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      rendered: 'auto',

      /**
       * The scaling behavior of the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
       * @expose
       * @name metricLabel.scaling
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "none"
       * @ojvalue {string} "thousand"
       * @ojvalue {string} "million"
       * @ojvalue {string} "billion"
       * @ojvalue {string} "trillion"
       * @ojvalue {string} "quadrillion"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      scaling: 'auto',

      /**
       * The CSS style object defining the style of the label.
       * @expose
       * @name metricLabel.style
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      style: {},

      /**
       * The text for the label. If specified, text will overwrite the numeric value that is displayed by default. The converter, scaling, and textType attributes are ignored when text is specified.
       * @expose
       * @name metricLabel.text
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       * @ojtranslatable
       */
      text: '',

      /**
       * Defines whether the label is a number or a percentage of the total value.
       * @expose
       * @name metricLabel.textType
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "percent"
       * @ojvalue {string} "number"
       * @default "number"
       */
      textType: 'number'
    },

    /**
     * The minimum value of the gauge.
     * @expose
     * @name min
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number}
     * @default 0
     */
    min: 0,

    /**
     * Defines the type of status meter to be rendered.
     * @expose
     * @name orientation
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "circular"
     * @ojvalue {string} "vertical"
     * @ojvalue {string} "horizontal"
     * @default "horizontal"
     */
    orientation: 'horizontal',

    /**
     * Plot Area for Status Meter Guage
     * @expose
     * @name plotArea
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Object}
     * @default {"borderRadius": "auto", "rendered": "auto", "svgClassName": "", "svgStyle": {}}
     */
    plotArea: {
      /**
       * The border color of the plot area.
       * @expose
       * @name plotArea.borderColor
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      borderColor: undefined,

      /**
       * Defines the border radius of the plot area shape. When set to "auto", the border radius is the same as the top level border radius. Acceptable input follows CSS border-radius attribute specifications.
       * @expose
       * @name plotArea.borderRadius
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default "auto"
       */
      borderRadius: 'auto',

      /**
       * The color of the plot area. Only applies when useThresholdFillColor is off.
       * @expose
       * @name plotArea.color
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      color: undefined,

      /**
       * Defines if the plot area is to be rendered. If set to auto, the plot area is rendered if the orientation is circular or if the thresholdDisplay is not onIndicator.
       * @expose
       * @name plotArea.rendered
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "on"
       * @ojvalue {string} "off"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      rendered: 'auto',

      /**
       * The CSS style class to apply to the plot area. The style class and inline style will override any other styling specified through the properties.
       * @expose
       * @name plotArea.svgClassName
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      svgClassName: '',

      /**
       * The inline style to apply to the plot area. The style class and inline style will override any other styling specified through the properties.
       * @expose
       * @name plotArea.svgStyle
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {}
    },

    /**
     * <p>The <code class="prettyprint">transientValue</code> is the read-only property for retrieving
     * the transient value from the status meter gauge. It is triggered when dragging over the status meter gauge.</p>
     *
     * <p>This is a read-only property so page authors cannot set or change it directly.</p>
     * @expose
     * @alias transientValue
     * @instance
     * @type {number|null}
     * @memberof oj.ojStatusMeterGauge
     * @since 4.2.0
     * @ojstatus preview
     * @readonly
     * @ojwriteback
     */
    rawValue: null,

    /**
     * Defines whether the value of the gauge can be changed by the end user.
     * @expose
     * @name readonly
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {boolean}
     * @default false
     */
    readonly: false,

    /**
     * An array of objects with the following properties defining the reference lines for the gauge.
     * @expose
     * @name referenceLines
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Array.<Object>}
     * @ojsignature {target: "Type", value: "Array<oj.ojStatusMeterGauge.ReferenceLine>", jsdocOverride: true}
     * @default []
     */
    referenceLines: [],

    /**
     * Specifies the start angle of a gauge with circular orientation. Value should be provided in degrees.
     * @expose
     * @name startAngle
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number}
     * @default 90
     * @ojunits degrees
     * @ojmin 0
     * @ojmax 360
     */
    startAngle: 90,

    /**
     * Specifies the increment by which values can be changed by the end user when readonly is false. The step must be a positive value that is smaller than the difference between the min and max. If not specified, the default step is 1/100 if the difference between the min and max.
     * @expose
     * @name step
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number|null}
     * @ojexclusivemin 0
     */
    step: null,

    /**
     * The CSS style class to apply to the gauge indicator. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name svgClassName
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     * @default ""
     */
    svgClassName: '',

    /**
     * The inline style to apply to the gauge indicator. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name svgStyle
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Object}
     * @default {}
     */
    svgStyle: {},

    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     */
    tooltip: {
      /**
       * A function that returns a custom tooltip. The function takes a tooltip context argument,
       * provided by the gauge, and should return an object that contains only one of the two properties:
       *  <ul>
       *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
       *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
       *  </ul>
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojStatusMeterGauge
       * @instance
       * @type {function(Object):Object|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojStatusMeterGauge.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))", jsdocOverride: true}
       * @default null
       */
      renderer: null
    },

    /**
     * Controls whether the current threshold is displayed on the indicator, in the plotArea, or if all the thresholds are diplayed in the plot area
     * @expose
     * @name thresholdDisplay
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "currentOnly"
     * @ojvalue {string} "all"
     * @ojvalue {string} "onIndicator"
     * @default "onIndicator"
     */
    thresholdDisplay: 'onIndicator',

    /**
     * An array of objects with the following properties defining the thresholds for the gauge.
     * @expose
     * @name thresholds
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {Array.<Object>}
     * @ojsignature {target: "Type", value: "Array<oj.ojStatusMeterGauge.Threshold>", jsdocOverride: true}
     * @default []
     */
    thresholds: [],

    /**
     * The metric value.
     * @expose
     * @name value
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {number|null}
     * @ojwriteback
     */
    value: null,

    /**
     * Defines whether visual effects such as overlays are applied to the gauge.
     * @expose
     * @name visualEffects
     * @memberof oj.ojStatusMeterGauge
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "auto"
     * @default "auto"
     */
    visualEffects: 'auto'
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    this._focusable({'element': this.element, 'applyHighlight': true});
    return dvt.StatusMeterGauge.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId == 'tooltip') {
      locator['subId'] = 'oj-statusmetergauge-tooltip';
    }
    return locator;
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-statusmetergauge');
    return styleClasses;
  },
  
   //** @inheritdoc */
  _GetComponentRendererOptions: function() {
    return ['tooltip/renderer', 'center/renderer'];
  },

  //** @inheritdoc */
  _ProcessOptions: function() {
    this._super();
    var center = this.options['center'];
    if (center && center['_renderer'])
      center['renderer'] = this._GetTemplateRenderer(center['_renderer'], 'center');
  },

  // @inheritdoc
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-dvtbase oj-statusmetergauge'] = {'path': 'animationDuration', 'property': 'ANIM_DUR'};
    return styleClasses;
  },

  //** @inheritdoc */
  _Render: function() {
    // Display the title of the surrounding div as the tooltip. Remove title from div to avoid browser default tooltip.
    if(this.element.attr('title'))
    {
      this.options['shortDesc'] =  this.element.attr('title');
      this.element.data( this.element,'title', this.element.attr('title'));
      this.element.removeAttr('title');
    }
    else if (this.element.data('title'))
      this.options['shortDesc'] =  this.element.data('title');

    // Call the super to render
    this._super();
  },

  /**
   * Returns the gauge's formatted metric label.
   * @return {string} The formatted metric label.
   * @expose
   * @instance
   * @memberof oj.ojStatusMeterGauge
   */
  getMetricLabel: function() {
    var auto = this._component.getAutomation();
    return auto.getMetricLabel();
  }
});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Drag</kbd></td>
 *       <td>Value change when <code class="prettyprint">readonly</code> is <code class="prettyprint">false</code>.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojStatusMeterGauge
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Submit the current value of the gauge.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element and submit the current value of the gauge.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Increase the gauge's transient value. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Decrease the gauge's transient value. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Decrease the gauge's transient value in left-to-right locales. Increase the gauge's transient value in right-to-left locales. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Increase the gauge's transient value in left-to-right locales. Decrease the gauge's transient value in right-to-left locales. Value is set after using <kbd>Enter</kbd> or <kbd>Tab</kbd> to submit.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojStatusMeterGauge
 */


// TYPEDEFS

/**
 * @typedef {Object} oj.ojStatusMeterGauge.ReferenceLine
 * @property {string} [color]  The color of the reference line.
 * @property {number} [value] The value of the reference line.
 */
/**
 * @typedef {Object} oj.ojStatusMeterGauge.Threshold
 * @property {string} [borderColor] The border color of the threshold.
 * @property {string} [color] The color of the threshold.
 * @property {number} [max] The upper bound of the threshold. This value is ignored for the final threshold, which uses the maximum value of the gauge.
 * @property {string} [shortDesc] Specific description for the threshold and overwrites the shortDesc specified on gauge. This is used for accessibility and also for customizing the tooltip text.
 */
/**
 * @typedef {Object} oj.ojStatusMeterGauge.TooltipContext
 * @property {string} color The indicator color of the gauge.
 * @property {Element} componentElement The status meter gauge HTML element.
 * @property {string} label The computed metric label.
 * @property {Element} parentElement The tooltip element. The function can directly modify or append content to this element.
 */
/**
 * @typedef {Object} oj.ojStatusMeterGauge.CenterContext
 * @property {Element} componentElement The status meter gauge HTML element.
 * @property {oj.ojStatusMeterGauge.Bounds} innerBounds Object containing (x, y, width, height) of the rectangle inscribed in the center area. The inner bounds are useful for inserting content that is guaranteed to fit within the center area. If the angleExtent isn't 360 then we will provide the clipped square.
 * @property {string} metricLabel The computed metric label.
 * @property {oj.ojStatusMeterGauge.Bounds} outerBounds Object containing (x, y, width, height) of the rectangle circumscribing the center area. The outer bounds are useful for creating background for the entire center area when used with a CSS border-radius. If the angleExtent isn't 360 then we will provide the clipped square.
 */
/**
 * @typedef {Object} oj.ojStatusMeterGauge.Bounds
 * @property {number} x The x position of the bounding rectangle.
 * @property {number} y The y position of the bounding rectangle.
 * @property {number} width The width of the bounding rectangle.
 * @property {number} height The height of the bounding rectangle.
 */


// DEPRECATED OPTIONS FOR WIDGET SYNTAX

/**
 * Defines whether the value of the gauge can be changed by the end user.
 * @ignore
 * @expose
 * @name readOnly
 * @memberof oj.ojStatusMeterGauge
 * @instance
 * @type {boolean}
 * @default true
 */


// SubId Locators **************************************************************

/**
 * <p>Sub-ID for the the status meter guage tooltip.</p>
 * 
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 * 
 * @ojsubid
 * @member
 * @name oj-statusmetergauge-tooltip
 * @memberof oj.ojSatusMeterGauge
 * @instance
 * 
 * @example <caption>Get the tooltip object of the gauge, if displayed:</caption>
 * var nodes = myStatusMeterGauge.getNodeBySubId({'subId': 'oj-statusmetergauge-tooltip'});
 */
/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var dvtBaseGaugeMeta = {
  "properties": {
    "translations": {
      "properties": {
        "componentName": {
          "type": "string"
        }
      }
    }
  },
  "methods": {},
  "extension": {
    _WIDGET_NAME: "dvtBaseGauge"
  }
};
oj.CustomElementBridge.registerMetadata('dvtBaseGauge', 'dvtBaseComponent', dvtBaseGaugeMeta);
})();

(function() {
var ojLedGaugeMeta = {
  "properties": {
    "borderColor": {
      "type": "string"
    },
    "svgClassName": {
      "type": "string"
    },
    "color": {
      "type": "string"
    },
    "max": {
      "type": "number"
    },
    "metricLabel": {
      "type": "object",
      "properties": {
        "converter": {
          "type": "object"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["on", "off"]
        },
        "scaling": {
          "type": "string",
          "enumValues": ["auto", "none", "thousand", "million", "billion", "trillion", "quadrillion"]
        },
        "style": {
          "type": "object"
        },
        "text": {
          "type": "string"
        },
        "textType": {
          "type": "string",
          "enumValues": ["percent", "number"]
        }
      }
    },
    "min": {
      "type": "number"
    },
    "rotation": {
      "type": "number",
      "enumValues": ["0", "90", "180", "270"]
    },
    "size": {
      "type": "number"
    },
    "svgStyle": {
      "type": "object"
    },
    "thresholds": {
      "type": "Array<object>"
    },
    "label": {
      "type": "object",
      "properties": {
        "style": {
          "type": "object"
        },
        "text": {
          "type": "string"
        }
      }
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "type": {
      "type": "string"
    },
    "value": {
      "type": "number",
      "writeback": true
    },
    "visualEffects": {
      "type": "string",
      "enumValues": ["none", "auto"]
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Gauge"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    }
  },
  "methods": {
    "getMetricLabel": {}
  },
  "extension": {
    _WIDGET_NAME: "ojLedGauge"
  }
};
oj.CustomElementBridge.registerMetadata('oj-led-gauge', 'dvtBaseGauge', ojLedGaugeMeta);
// Supported marker shapes for gauges
var _LED_GAUGE_SHAPE_ENUMS = {
  "arrow": true,
  "square": true,
  "rectangle": true,
  "circle": true,
  "diamond": true,
  "triangle": true,
  "human": true,
  "star": true
};
// Get the combined meta of superclass which contains a shape parse function generator
var dvtMeta = oj.CustomElementBridge.getMetadata('oj-led-gauge');
oj.CustomElementBridge.register('oj-led-gauge', {
  'metadata': dvtMeta,
  'parseFunction': dvtMeta['extension']._DVT_PARSE_FUNC({'type': true}, _LED_GAUGE_SHAPE_ENUMS)
});
})();

(function() {
var ojRatingGaugeMeta = {
  "properties": {
    "changed": {
      "type": "boolean",
      "writeback": true
    },
    "changedState": {
      "type": "object",
      "properties": {
        "borderColor": {
          "type": "string"
        },
        "svgClassName": {
          "type": "string"
        },
        "color": {
          "type": "string"
        },
        "shape": {
          "type": "string"
        },
        "source": {
          "type": "string"
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "hoverState": {
      "type": "object",
      "properties": {
        "borderColor": {
          "type": "string"
        },
        "svgClassName": {
          "type": "string"
        },
        "color": {
          "type": "string"
        },
        "shape": {
          "type": "string"
        },
        "source": {
          "type": "string"
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "max": {
      "type": "number"
    },
    "min": {
      "type": "number"
    },
    "orientation": {
      "type": "string",
      "orientation": ["vertical", "horizontal"]
    },
    "preserveAspectRatio": {
      "type": "string"
    },
    "readonly": {
      "type": "boolean"
    },
    "selectedState": {
      "type": "object",
      "properties": {
        "borderColor": {
          "type": "string"
        },
        "svgClassName": {
          "type": "string"
        },
        "color": {
          "type": "string"
        },
        "shape": {
          "type": "string"
        },
        "source": {
          "type": "string"
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "step": {
      "type": "number",
      "enumValues": ["1", ".5"]
    },
    "transientValue": {
      "type": "number",
      "writeback": true,
      "readOnly": true
    },
    "thresholds": {
      "type": "Array<object>"
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Gauge"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    },
    "unselectedState": {
      "type": "object",
      "properties": {
        "borderColor": {
          "type": "string"
        },
        "svgClassName": {
          "type": "string"
        },
        "color": {
          "type": "string"
        },
        "shape": {
          "type": "string"
        },
        "source": {
          "type": "string"
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "value": {
      "type": "number",
      "writeback": true
    },
    "visualEffects": {
      "type": "string",
      "enumValues": ["none", "auto"]
    }
  },
  "methods": {},
  "extension": {
    _ALIASED_PROPS: {"transientValue": "rawValue"},
    _WIDGET_NAME: "ojRatingGauge"
  }
};
oj.CustomElementBridge.registerMetadata('oj-rating-gauge', 'dvtBaseGauge', ojRatingGaugeMeta);

// Consider a string with at least one digit a valid SVG path
var _SHAPE_REGEXP = /\d/;
var _RATING_GAUGE_SHAPE_ENUMS = {
  "circle": true,
  "square": true,
  "diamond": true,
  "triangle": true,
  "human": true,
  "star": true
};
var _UNSELECTED_RATING_GAUGE_SHAPE_ENUMS = {
  "circle": true,
  "square": true,
  "diamond": true,
  "triangle": true,
  "human": true,
  "star": true,
  "dot": true,
  "none": true
};
var _RATING_GAUGE_SHAPE_PROPS = {
  'changed-state.shape': true, 
  'hover-state.shape': true,
  'selected-state.shape': true,
  'unselected-state.shape': true
};
function shapePropertyParser(value, name, meta, defaultParseFunction) {
  if (_RATING_GAUGE_SHAPE_PROPS[name] || name === 'unselected-state.shape') {
    if (_SHAPE_REGEXP.test(value))
      return value;
    else if (_RATING_GAUGE_SHAPE_PROPS[name] && !_RATING_GAUGE_SHAPE_ENUMS[name])
      throw "Found: " + value + ". Expected: " + _RATING_GAUGE_SHAPE_ENUMS.toString();
    else if (name === 'unselected-state.shape' && !_UNSELECTED_RATING_GAUGE_SHAPE_ENUMS[name])
      throw "Found: " + value + ". Expected: " + _UNSELECTED_RATING_GAUGE_SHAPE_ENUMS.toString();
    else
      return value;
  }
  return defaultParseFunction(value);
};
oj.CustomElementBridge.register('oj-rating-gauge', {
  'metadata': oj.CustomElementBridge.getMetadata('oj-rating-gauge'),
  'parseFunction': shapePropertyParser
});
})();

(function() {
var ojStatusMeterGaugeMeta = {
  "properties": {
    "angleExtent": {
      "type": "number"
    },
    "animationDuration": {
      "type": "number"
    },
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "borderColor": {
      "type": "string"
    },
    "borderRadius": {
      "type": "string"
    },
    "center": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "svgClassName": {
      "type": "string"
    },
    "color": {
      "type": "string"
    },
    "indicatorSize": {
      "type": "number"
    },
    "innerRadius": {
      "type": "number"
    },
    "max": {
      "type": "number"
    },
    "metricLabel": {
      "type": "object",
      "properties": {
        "converter": {
          "type": "object"
        },
        "position": {
          "type": "string",
          "enumValues": ["auto", "center", "insideIndicatorEdge", "outsideIndicatorEdge", "outsidePlotArea", "withTitle"]
        },
        "rendered": {
          "type": "string",
          "enumValues": ["auto", "on", "off"]
        },
        "scaling": {
          "type": "string",
          "enumValues": ["auto", "none", "thousand", "million", "billion", "trillion", "quadrillion"]
        },
        "style": {
          "type": "object"
        },
        "text": {
          "type": "string"
        },
        "textType": {
          "type": "string",
          "enumValues": ["percent", "number"]
        }
      }
    },
    "min": {
      "type": "number"
    },
    "orientation": {
      "type": "string",
      "enumValues": ["circular", "horizontal", "vertical"]
    },
    "plotArea": {
      "type": "object",
      "properties": {
        "borderColor": {
          "type": "string"
        },
        "borderRadius": {
          "type": "string"
        },
        "svgClassName": {
          "type": "string"
        },
        "color": {
          "type": "string"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["auto", "on", "off"]
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "readonly": {
      "type": "boolean"
    },
    "referenceLines": {
      "type": "Array<object>"
    },
    "startAngle": {
      "type": "number"
    },
    "step": {
      "type": "number"
    },
    "svgStyle": {
      "type": "object"
    },
    "transientValue": {
      "type": "number",
      "writeback": true,
      "readOnly": true
    },
    "thresholdDisplay": {
      "type": "string",
      "enumValues": ["currentOnly", "all", "onIndicator"]
    },
    "thresholds": {
      "type": "Array<object>"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Gauge"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    },
    "label": {
      "type": "object",
      "properties": {
        "position": {
          "type": "string",
          "enumValues": ["auto", "center", "start"]
        },
        "style": {
          "type": "object"
        },
        "text": {
          "type": "string"
        }
      }
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "value": {
      "type": "number",
      "writeback": true
    },
    "visualEffects": {
      "type": "string",
      "enumValues": ["none", "auto"]
    }
  },
  "methods": {
    "getMetricLabel": {}
  },
  "extension": {
    _ALIASED_PROPS: {"transientValue": "rawValue"},
    _WIDGET_NAME: "ojStatusMeterGauge"
  }
};
oj.CustomElementBridge.registerMetadata('oj-status-meter-gauge', 'dvtBaseGauge', ojStatusMeterGaugeMeta);
oj.CustomElementBridge.register('oj-status-meter-gauge', {'metadata': oj.CustomElementBridge.getMetadata('oj-status-meter-gauge')});
})();

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/
/*global oj, ko, Promise, signals */

/**
 * The ojRouter module.
 */
// Wrap in a IIFE to prevents the possiblity of collision in a non-AMD scenario.
(function() {
"use strict";
/**
 * Hold the value of the oj.Router.defaults.baseUrl property.
 * @private
 * @type {!string}
 */
var _baseUrlProp = '/';

/**
 * Hold the title before being modified by router
 * @private
 * @type {?string}
 */
var _originalTitle;

/**
 * Hold the url adapter to be used.
 * @private
 * @type {oj.Router.urlPathAdapter|oj.Router.urlParamAdapter}
 */
var _urlAdapter;

/**
 * The default name for the root instance.
 * @private
 * @const
 * @type {string}
 */
var _DEFAULT_ROOT_NAME = 'root';

/**
 * The name of the request param for bookmarkable data.
 * @private
 * @const
 * @type {string}
 */
var _ROUTER_PARAM = 'oj_Router';
/**
 * The separator used to build the title from router labels.
 * @private
 * @const
 * @type {string}
 */
var _TITLE_SEP = ' | ';

/**
 * Maximum size of URL
 * @private
 * @const
 * @type {number}
 */
var _MAX_URL_LENGTH = 1024;

/**
 * Name of the window event used to listen to the browser history changes
 * @private
 * @const
 * @type {string}
 */
var _POPSTATE = 'popstate';

/**
 * Name of the property of the object in the Promise returned by go() or sync()
 * @private
 * @const
 * @type {string}
 */
var _HAS_CHANGED = 'hasChanged';

/**
 * Object commonly used as return value for go() or sync()
 * @private
 * @const
 * @type {{hasChanged:boolean}}
 */
var _NO_CHANGE_OBJECT = { 'hasChanged': false };

/**
 * Flag set to true when oj.Router is initialized
 * @private
 * @type {boolean}
 */
var _initialized = false;

/**
 * Hold the leftover path when using ojModule and the child router
 * doesn't exist yet.
 * @private
 * @type {string|undefined}
 */
var _deferredPath;

/**
 * A queue to hold unprocessed transitions
 * @private
 * @type {Array.<Object>}
 */
var _transitionQueue = [];

/**
 * A promise that resolve when all transition in the queue is resolved.
 * @private
 */
var _queuePromise;

/**
 * A shortcut to access window.location
 * @private
 */
var _location = window.location;

/**
* The name of the object containing state parameters.
* @private
*/
var _parametersValue = 'parameters';

/**
 * The instance of the root router.
 * @private
 * @const
 * @type {!oj.Router}
 */
var rootRouter;

/**
 * Return key/value object of query parameters.
 * @private
 * @return {!Object.<string, string>}
 */
function parseQueryParam(queryString) {
   var params = {},
       keyValPairs;

   // Remove starting '?'
   queryString = queryString.split('?')[1];

   if (queryString) {
      keyValPairs = queryString.split('&');
      keyValPairs.forEach(function(pair) {
         var parts = pair.split(/\=(.+)?/);
         var key = parts[0];
         var value;

         if (key.length) {
            value = parts[1] && decodeURIComponent(parts[1]);
            params[key] = value;
         }
      });
   }

   return params;
}

/**
 * Takes a path of segment separated by / and returns an array of segments
 * @param  {string=} path a path of segment separated by /
 * @return {!Array} an array of segment
 */
function _getSegments(path) {
   var array = path ? path.split('/') : [];
   return array;
}

/**
 * Returns the first segment of a path separated by /
 * @param  {string=} id
 * @return {string}
 */
function _getShortId(id) {
   var segment = _getSegments(id)[0];
   return _decodeSlash(segment);
}

/**
 * Build an URL by replacing portion of the existing URL. Portion that can be replaces are
 * pathname and search field. Use the extraState to build the new state param.
 * @private
 * @param  {!Object} pieces
 * @param  {!Object.<string, Object>} extraState
 * @return {!string}
 */
function _buildUrl(pieces, extraState) {
   var parser = document.createElement('a');
   parser.href = _location.href;

   if (pieces.search !== undefined) {
      parser.search = pieces.search;
   }

   if (pieces.pathname !== undefined) {
      parser.pathname = pieces.pathname;
   }

   // Add or replace the existing state param
   parser.search = putStateParam(parser.search, extraState);

   return parser.href.replace(/\?$/, ''); // Remove trailing ? for IE
}

/**
 * Return the {@link oj.RouterState} object for a specific router given a state id.
 * @private
 * @param {oj.Router} router The router object.
 * @param {string} stateId The state id.
 * @return {oj.RouterState | undefined} The state object.
 */
function getStateFromId(router, stateId) {
   var state;

   if (stateId && router._states) {
      oj.Assert.assertString(stateId);

      router._states.every(function(stateAt) {
         if (stateAt._id === stateId) {
            state = stateAt;
            return false;
         }
         return true;
      });
   }
   return state;
}

/**
 * Retrieve a router full name. A path of all rooter name from root.
 * @private
 * @param  {!oj.Router} router
 * @return {!string}
 */
function getRouterFullName(router) {
   if (router._parentRouter) {
      return getRouterFullName(router._parentRouter) + '.' + router._name;
   }

   return router._name;
}

/**
 * Retrieves the absolute path to the current state. If one of the parent router current state is
 * not defined, the path is meaningless so returns undefined.
 * @private
 * @param {oj.Router|undefined} router
 * @return {string|undefined} path
 */
function getCurrentPath(router) {
   var path, sId;

   if (router) {
      path = getCurrentPath(router._parentRouter);

      if (path) {
         sId = router._stateId();

         if (sId) {
            path += sId + '/';
         }
         else {
            path = undefined;
         }
      }
   }
   else {
      path = '/';
   }

   return path;
}

/**
 * Return the child router for a specific parent state value
 * @private
 * @param {oj.Router} router
 * @param {string|undefined} value
 * @return {oj.Router|undefined}
 */
function _getChildRouter(router, value) {
   var subRouter;

   router._childRouters.every(function(sr) {
      // If child router doesn't have _parentState, then it was created before
      // the parent router navigated to a state, and so it's used as the default
      // child router.  There can only be one child router per parent state (even
      // if the parent state is "undefined"), so we should only ever encounter,
      // at most, one sub-router whose parent state is undefined.
      if (!sr._parentState) {
        subRouter = sr;
      }
      // Otherwise, if a specific subrouter matches the queried value, then use
      // it instead of the default.
      else if (sr._parentState === value) {
         subRouter = sr;
         // Once found, exit the loop right away.  There can be only one child
         // router per parent state, so the first one we find will be the only
         // one.
         return false;
      }
      return true;
   });

   return subRouter;
}

/**
 * Only keep changes where the value doesn't match the current router state
 * @private
 * @param {!Array.<_StateChange>} changes
 * @return {!Array.<_StateChange>}
 */
function _filterNewState(changes) {
   var newChanges = changes.filter(function(change) {
      return (change.value !== change.router._stateId());
   });

   if (oj.Logger.option('level') === oj.Logger.LEVEL_INFO) {
      oj.Logger.info('Potential changes are: ');
      newChanges.forEach(function(change) {
         oj.Logger.info('   { router: %s, value: %s }',
                        change.router && getRouterFullName(change.router), change.value);
      });
   }

   return newChanges;
}

/**
 * Update the bookmarkable data
 * @private
 * @this {!Object.<string, Object>}
 * @param {?Object} change
 */
function _updateBookmarkableData(change) {
   var ex = this[change.router._name];
   if (ex !== undefined) {
      change.router._extra = ex;
   }
}

/**
 * Return true if the current transition is cancelled.
 * See queuing of transaction in _queueTransaction
 * @private
 * @return {boolean}
 */
function isTransitionCancelled() {
   return (_transitionQueue[0] && _transitionQueue[0].cancel);
}

/**
 * Encode and compress the a state object. This is used for bookmarkable data.
 * @private
 * @param {!Object.<string, Object>} extraState
 * @return {!string}
 * @throws An error if bookmarkable state is too big.
 */
function encodeStateParam(extraState) {
   var jsonState = JSON.stringify(extraState),
       encodedState = encodeURIComponent(jsonState),
       compressedState = oj.LZString.compressToEncodedURIComponent(jsonState),
       useCompressed = false,
       param = _ROUTER_PARAM + '=';

   if (compressedState.length <= encodedState.length) {
      useCompressed = true;
   }

   if (useCompressed) {
      param += '1' + compressedState;
   }
   else {
      param += '0' + encodedState;
   }

   if (param.length > _MAX_URL_LENGTH) {
      throw new Error('Size of bookmarkable data is too big.');
   }

   return param;
}

/**
 * Decompress and decode the state param from the URL.  This is used for bookmarkable data.
 * @private
 * @param {!string} param
 * @return {!Object.<string, Object>}
 * @throws An error if parsing fails or format is invalid.
 */
function decodeStateParam(param) {
   var extraState;
   // First character is the compression type. Right now only 0 and 1 are supported.
   // 0 for no compression, 1 for LZW
   var compressionType = param.charAt(0);

   param = param.slice(1);

   if (compressionType === '0') {
      param = decodeURIComponent(param);
   }
   else if (compressionType === '1') {
      param = oj.LZString.decompressFromEncodedURIComponent(param);
   }
   else {
      throw new Error('Error retrieving bookmarkable data. Format is invalid');
   }

   extraState = /** @type {!Object.<string, Object>} */ (JSON.parse(param));

   if (oj.Logger.option('level') === oj.Logger.LEVEL_INFO) {
      var name;
      oj.Logger.info('Bookmarkable data: ');
      for (name in extraState) {
         oj.Logger.info('   { router: %s, value: %s }', name, extraState[name]);
      }
   }

   return extraState;
}

/**
 * Replace the state param in the URL.
 * @private
 * @param {!string} url the url to which the param will be added
 * @param {!Object.<string, Object>} extraState the object to be stored in the param
 * @return {!string} the URL with the new state param
 */
function putStateParam(url, extraState) {
   var stateParam = '';
   var startSegment, endSegment;
   var start = url.indexOf(_ROUTER_PARAM);

   if (start !== -1) {
      var end = url.indexOf('&', start);
      if (end === -1) {
         end = url.length;
      }

      startSegment = url.substring(0, start);
      endSegment = url.substr(end);
   }
   else {
      startSegment = url + ((url.indexOf('?') === -1) ? '?' : '&');
      endSegment = '';
   }

   if (extraState && Object.getOwnPropertyNames(extraState).length > 0) {
      stateParam = encodeStateParam(extraState);
   }
   else {
      // Remove the '?' or '&'
      startSegment = startSegment.substring(0, startSegment.length -1);
   }

   return startSegment + stateParam + endSegment;
}

/**
 * Traverse all the child routers in order to find a router that has the
 * state id given as an argument.
 * @private
 * @param  {!oj.Router} router
 * @param  {!string} sId
 * @param  {string=} parentStateId
 * @return {oj.Router|undefined}
 */
function _findRouterForStateId(router, sId, parentStateId) {
   var result;

   router._childRouters.every(function(child) {
      if ((!child._parentState || child._parentState === parentStateId) &&
          child.stateFromIdCallback(sId)) {
         result = child;
         return false;
      }
      return true;
   });

   return result;
}

/**
 * Traverse the tree of routers and build an array of states made of the router and an
 * undefined value.
 * The first item of the array is the root and the last is the leaf.
 * @private
 * @param  {oj.Router} router
 */
function _buildAllUndefinedState(router) {
   var states = [];

   if (router._currentState()) {
      // Push a state change with undefined value (2nd argument in constructor missing)
      states.push(new _StateChange(router));

      router._childRouters.forEach(function (child) {
         states = states.concat(_buildAllUndefinedState(child));
      });
   }

   return states;
}

/**
 * Build a page title using the label of the current child routers state
 * @private
 * @param  {oj.Router|undefined} router
 * @return {!{segment: string, title: string}}
 */
function _buildTitle(router) {
   if (!router) {
      return { title: '', segment: '' };
   }

   // Recurse leaf first
   var titleInfo = _buildTitle(_getChildRouter(router, router._stateId()));

   // If we don't have a title yet, build one.
   if (titleInfo.title === '') {
      var state = router._currentState();
      if (state) {
         // If a title property is present, it has precedence.
         var title = state._title;
         if (title !== undefined) {
            if (typeof title === 'function') {
               title = title();
            }
            titleInfo.title = String(title);
         }
         else {
            // Otherwise, compose the title with the label
            title = state._label;
            if (title !== undefined) {
               title =  String(title);
               // Append existing segment
               if (titleInfo.segment !== '') {
                  title += _TITLE_SEP + titleInfo.segment;
               }
               titleInfo.segment = title;
            }
         }
      }
   }

   return titleInfo;
}

/**
 * Dispatch the transitionedToState signal
 * @private
 * @param {Object} param
 * @param {boolean} param.haschanged
 * @param {oj.Router=} param.router
 * @param {oj.RouterState=} param.oldState
 * @param {oj.RouterState=} param.newState
 */
function dispatchTransitionedToState(param) {
   oj.Router._transitionedToState.dispatch(param);
}

/**
 * An object use to represent a change in a RouterState
 * @constructor
 * @param {!oj.Router} router
 * @param {string=} value
 * @private
 * @ojtsignore
 */
function _StateChange(router, value) {
   this.router = router;
   this.value = value;     // the value is also the stateId
}

/**
 * Returns the RouterState object for the state id in this change object
 * @return {oj.RouterState} the RouterState matching the value
 * @private
 */
_StateChange.prototype.getState = function () {
   if (!this.state) {
      if (this.value) {
         this.state = this.router.stateFromIdCallback(_getShortId(this.value));
      }
   }

   return this.state;
};

/**
 * Store a state parameter value by appending it to the value as a path
 * @param {string=} value
 * @private
 */
_StateChange.prototype.addParameter = function (value) {
   if (value) {
      this.value += '/' + value;
   }
};

/**
 * Takes an array of changes from parsing and appends other changes needed to be done.
 * 1) All cascading default state
 * 2) All the state that need to become undefined
 * @private
 * @param {!Array.<_StateChange>} states
 * @return {!Array.<_StateChange>}
 */
function _appendOtherChanges(states) {
   var lastItem = states[states.length -1];
   var router, value;

   // If there is a state, starts with it
   if (lastItem) {
      router = lastItem.router;
      value = _getShortId(lastItem.value);
   }
   // Otherwise, starts at the root router
   else {
      router = rootRouter;
      value = rootRouter._defaultStateId;
      if (value) {
         states.push(new _StateChange(router, value));
      }
   }

   // Append all the default states all the way to the leaf router
   while (!!(router = _getChildRouter(router, value))) {
      value = router._defaultStateId;
      if (value) {
         states.push(new _StateChange(router, value));
      }
   }

   // Build an array of all the state to become undefined due to the parent state changing. The
   // order of execution is leaf first.

   var undefStates = _buildAllUndefinedState(rootRouter); // First build an array of undefined state
   var undef = [];

   undefStates.forEach(function (select, i) {
      var change = states[i];

      // Only insert change for a different router since the undef change will already happen when
      // a router transition to a different state.
      if  (!change || select.router !== change.router) {
         undef.unshift(select);
      }
   });

   // The order of execution is exit(undef) from leaf to root followed by enter from root to leaf
   states = undef.concat(states);

   return states;
}

/**
 * Build an array of objects by visiting the parent hierarchy.
 * Each element of the array represent the state of a router.
 * @private
 * @param {!oj.Router} router
 * @param {!string} path
 * @return {!Array.<_StateChange>}
 */
function _buildStateFromPath(router, path) {
   var newStates = [],
       routers = [],
       rt = router,
       parts = _getSegments(path),
       sId, parent, parentStateId, state, pName, stateChange, pi = 0;

   // Since path is absolute, it always starts with '/', so remove the first
   // element (empty string)
   parts.splice(0, 1);

   // Build an array of routers, from the root to the current one.
   while (rt) {
      routers.unshift(rt);
      rt = rt._parentRouter;
   }

   // Traverse path and routers simultaneously.
   while (!!(sId = parts.shift())) {

      if (state) {
         pName = state._paramOrder[pi];

         // If state has parameters, save the state Id as the parameter value
         if (pName) {
            stateChange.addParameter(sId);
            pi++;
            continue;
         }
         // Otherwise, reset parameter index
         else {
            pi = 0;
         }
      }

      rt = routers.shift();

      if (!rt) {
         rt = _findRouterForStateId(parent, sId, parentStateId);

         // Router doesn't exist, save deferredPath and stop
         if (!rt) {
            _deferredPath = path;
            return newStates;
         }
      }

      stateChange = new _StateChange(rt, sId);
      state = stateChange.getState();
      if (!state) {
         throw new Error('Invalid path "' + path +
               '". State id "' + sId + '" does not exist on router "' +
                  rt._name + '".');
      }

      newStates.push(stateChange);
      parent = rt;
      parentStateId = sId;
   }

   return newStates;
}

/**
 * Execute a callback and log if returns false.
 * @private
 * @param  {(function (): (?)|undefined)} callback
 * @param  {!string}   type
 * @param  {string}   stateId
 * @return {boolean}
 */
function _executeCallback(callback, type, stateId) {
   var result = callback();
   if (!result) {
      oj.Logger.info('%s is false for state: %s', type, stateId);
   }

   return result;
}

/**
 * Chain callback into a sequence if it's a function.
 * @private
 * @param  {(function (): (?)|undefined)} callback
 * @param  {IThenable.<?>|null}           sequence
 * @param  {!string}                      type
 * @param  {string}                       stateId
 * @return {IThenable.<?>|null}
 */
function _chainCallback(callback, sequence, type, stateId) {
   // Check if we can enter this new state by executing the callback.
   // If the callback is a function, chain it.
   if (typeof callback === 'function') {
      // Check if this is the start of the chain
      if (!sequence) {
         sequence = new Promise(function(resolve) {
            resolve(_executeCallback(callback, type, stateId));
         });
      }
      else {
         sequence = sequence.then(function(result) {
            // Only test the next state if the previous promise return true
            if (result) {
               result = _executeCallback(callback, type, stateId);
            }

            return result;
         });
      }
   }

   return sequence;
}

/**
 * Traverse the child router and build a chain of promise for each canExit callback.
 * @private
 * @param  {oj.Router} router
 * @param  {IThenable.<?>|null}  sequence
 * @return {IThenable.<?>|null}  chain of promises executing the canExit on the current states
 */
function _buildCanExitSequence(router, sequence) {
   var currentState = router._currentState(),
       canExitCallback,
       i;

   if (currentState) {
      // Traverse each child router and ask for canExit
      for (i = 0; i < router._childRouters.length; i++) {
         sequence = _buildCanExitSequence(router._childRouters[i], sequence);
      }

      // A callback defined on bound viewModel has precedence.
      if (currentState.viewModel && currentState.viewModel['canExit']) {
         canExitCallback = currentState.viewModel['canExit'];
      } else {
         canExitCallback = currentState._canExit;
      }

      sequence = _chainCallback(canExitCallback, sequence, 'canExit', currentState._id);
   }

   return sequence;
}

/**
 * Invoke canExit callbacks in a deferred way.
 * @private
 * @param  {oj.Router} router
 * @return {IThenable.<?>|null} a promise returning true if all of the current state can exit.
 */
function _canExit(router) {
   var sequence;

   if (isTransitionCancelled()) {
      return Promise.resolve(false);
   }

   oj.Logger.info('Start _canExit.');

   if (router) {
      sequence = _buildCanExitSequence(router, null);
      if (sequence === null) {
         sequence = Promise.resolve(true);
      }
      else {
         sequence = sequence.then(function(result) {
            return (result && !isTransitionCancelled());
         });
      }
   }
   else {
      sequence = Promise.resolve(true);
   }

   return sequence;
}

/**
 * Return a promise returning an object with an array of all the changes and the origin if all of
 * the new state in the allChanges array can enter.
 * @private
 * @param {!Array.<_StateChange>} allChanges
 * @param {string=} origin a string specifying the origin of the transition ('sync', 'popState')
 * @return {!Promise} a promise returning an object with an array of all the changes and the origin.
 */
function _canEnter(allChanges, origin) {
   if (isTransitionCancelled()) {
      return Promise.resolve();
   }

   oj.Logger.info('Start _canEnter.');

   var sequence = null;

   // Build a chain of canEnter promise for each state in the array of changes
   allChanges.forEach(function(change) {
      var newState = change.getState();

      // It is allowed to transition to an undefined state, but no state
      // callback need to be executed.
      if (newState) {
         sequence = _chainCallback(newState._canEnter, sequence, 'canEnter', newState._id);
      }
   });

   if (sequence === null) {
      sequence = Promise.resolve({ allChanges: allChanges, origin: origin });
   }
   else {
      sequence = sequence.then(function(result) {
         var returnObj;

         if (result && !isTransitionCancelled()) {
            returnObj = { allChanges: allChanges, origin: origin };
         }

         return returnObj;
      });
   }

   return sequence;
}

/**
 * Update the state of a router with the new value.
 * @private
 * @param {{value:string, router:!oj.Router}} change
 * @param {string | undefined} origin
 */
function _update(change, origin) {
   var oldState = change.router.stateFromIdCallback(_getShortId(change.router._stateId())),
       newState = change.getState();

   return Promise.resolve().
      then(function() {
         if (oj.Logger.option('level') === oj.Logger.LEVEL_INFO) {
            oj.Logger.info('Updating state of %s to %s.',
               getRouterFullName(change.router), change.value);
         }
      }).
      // Execute exit on the current state
      then(oldState ? oldState._exit : undefined).
      then(function() {
         var rt = change.router,
             i, length, goingBackward;

         // Are we going back to the previous state?
         if (origin === 'popState') {
            length = rt._navHistory.length;
            // Are we going back to the previous state?
            for (i = length-1; i >= 0; i--) {
               if (rt._navHistory[i] === change.value) {
                  goingBackward = true;
                  // Delete all elements up the one matching
                  rt._navHistory.splice(i, length - i);
                  break;
               }
            }

            // Back only if going back 1
            if ((length - i) === 1) {
               rt._navigationType = 'back';
            }
         }

         if (!goingBackward) {
            delete rt._navigationType;
            rt._navHistory.push(_getShortId(rt._stateId()));
         }

         // Update the parameters
         if (change.value && newState) {

            var segments = _getSegments(change.value);
            newState._paramOrder.forEach(function(name, i) {
               var newValue = _decodeSlash(segments[i+1]);
               var oldValue = newState._parameters[name];

               // Update the parameter value
               if (newValue !== oldValue) {
                  newState._parameters[name] = newValue;
               }
            });

            //TODO: Should we execute a callback for case where state doesn't change
            //      and using a configure with function? or disable for function configure?
         }

         // Change the value of the stateId
         rt._stateId(change.value);
      }).
      // Execute enter on the new state
      then(newState ? newState._enter : undefined);
}

/**
 * Update the state of all routers in the change array.
 * @private
 * @param {Object} updateObj
 * @return {!Promise}
 */
function _updateAll(updateObj) {
   if (!updateObj) {
      return Promise.resolve(_NO_CHANGE_OBJECT);
   }

   var sequence = Promise.resolve().then(function() {
      oj.Logger.info('Entering _updateAll.');
      oj.Router._updating = true;
   });

   var oldState;
   var allChanges = updateObj.allChanges;
   allChanges.forEach(function(change) {
      oldState = change.router['currentState'].peek();
      sequence = sequence.then(function() {
         if (!isTransitionCancelled()) {
            return _update(change, updateObj.origin);
         }
      });
   });

   return sequence.then(function() {
      var hasChanged = false;
      var router;
      var newState;
      if (allChanges.length) {
        hasChanged = !isTransitionCancelled();
        /*
         * Pass the last state of a multi-state transition as the new state to
         * which we're transitioning.
         */
        var change = allChanges[allChanges.length-1];
        router = change.router;
        newState = change.state;
      }
      oj.Router._updating = false;
      oj.Logger.info('_updateAll returns %s.', String(hasChanged));
      return {
        'hasChanged': hasChanged,
        'router': router,
        'oldState': oldState,
        'newState': newState
      };
   }, function(error) {
      oj.Router._updating = false;
      return Promise.reject(error);
   });
}

/**
 * Update the state using the current URL.
 * @private
 * @param {string=} origin the transition origin
 * @return a Promise that resolves when the routers state are updated
 */
function parseAndUpdate(origin) {
   var allChanges;

   try {
      allChanges = _urlAdapter.parse();

      // Only keep changes where the value doesn't match the router state
      allChanges = _filterNewState(allChanges);
   }
   catch (error) {
      return Promise.reject(error);
   }

   return _canEnter(allChanges, origin).then(_updateAll);
}

/**
 * Log the action and transaction as Logger.info.
 * @param       {string} action     The action performed
 * @param       {Object} transition The router transition.  The transition object
 * contains the following properties:
 * - {string} path The path to where the transition is occurring
 * - {string} origin The origin of the transition.  Will be one of, "sync", "go",
 *   "popState"
 * - {oj.Router} router The instance of the router performing the transition
 * - {string} historyUpdate A string indicating how the router should update the
 *   browser history.  Possible values are, "skip" or "replace"
 *   updated
 * @private
 */
function _logTransition(action, transition) {
   if (oj.Logger.option('level') === oj.Logger.LEVEL_INFO) {
      var path = transition.path ? 'path=' + transition.path : '';
      var deferString = transition.deferredHandling ? 'deferredHandling=true' : '';
      var router = transition.router ? getRouterFullName(transition.router) : 'null';
      oj.Logger.info('>> %s: origin=%s router=%s %s %s',
         action, transition.origin, router, path, deferString);
   }
}

/**
 * Execute a transition. There are 3 types of transitions depending if they are
 * called from go, sync or handlePopState.
 * @private
 * @param  {Object} transition An object with properties describing the transition.
 * @return A Promise that resolves when the router is done with the state transition.
 */
function _executeTransition(transition) {

   _logTransition('Executing', transition);

   if (!transition.deferredHandling) {
      // if the transition originate from a sync call, don't call canExit
      if (transition.origin === 'sync') {
         return parseAndUpdate();
      }
      else if (transition.origin === 'popState') {
         return _canExit(transition.router).then(function (canExit) {
            if (canExit) {
               return parseAndUpdate(transition.origin);
            }
            return Promise.resolve(_NO_CHANGE_OBJECT);
         });
      }
   }
   return transition.router._go(transition);
}

/**
 * Executes first transition on the queue then unqueue then recurse.
 * @private
 */
function _resolveTransition() {
   var transition = _transitionQueue[0],
       promise;

   _logTransition('Resolving', transition);

   if (transition.cancel) {
      _logTransition('Cancelled', transition);
      promise = Promise.resolve(_NO_CHANGE_OBJECT);
   }
   else {
      promise = _executeTransition(transition);
   }

   return promise.then(function(params) {
      var done = _transitionQueue.shift();
      _logTransition('Done with', done);
      if (params[_HAS_CHANGED] === true) {
         // Build the window title that will appear in the browser history
         var titleInfo = _buildTitle(rootRouter);
         var title;

         if (titleInfo.title !== '') {
            title = titleInfo.title;
         }
         else {
            if (_originalTitle && _originalTitle.length > 0) {
               title = _originalTitle;
               if (titleInfo.segment !== '') {
                  title += _TITLE_SEP + titleInfo.segment;
               }
            }
            else {
               title = titleInfo.segment;
            }
         }

         if (title !== window.document.title) {
            window.document.title = title;
         }
      }
      dispatchTransitionedToState(params);
      return params;
   }, function(error) {
      _transitionQueue = [];
      oj.Logger.error('Error when executing transition: %o', (error || 'Unknown'));
      dispatchTransitionedToState(_NO_CHANGE_OBJECT);
      return Promise.reject(error);
   });
}

/**
 * Queue a transition. It will execute as soon as previous transitions in the
 * queue are done.
 * @private
 * @param  {Object} transition An object with properties describing the transition
 * @return A Promise that resolves when the router is done with the given state transition.
 */
function _queueTransition(transition) {
   var lastTransition, length;

   _logTransition('Queuing  ', transition);

   var path = transition['path'];
   var bc = oj.Context.getPageContext().getBusyContext();
   // Disabling the state due to  - OJ.TESTS.ROUTER.SAMPLEOJMODULETEST FAILS
   // var removeBusyState = bc.addBusyState({description:'router transitioning to new state "'+path+'"'});

   // Push new transition at the end. Current transition is always at index 0
   length = _transitionQueue.push(transition);

   // Simple case when the transition is the only one in the queue.
   if (length === 1) {
      _queuePromise = _resolveTransition();
   }
   // Cancel transition in queue and chain it
   else {
      lastTransition = _transitionQueue[length-2];
      // Don't cancel transitions from popstate event or for deferred path
      if (!lastTransition.deferredHandling) {
         _logTransition('Cancelling', lastTransition);
         lastTransition.cancel = true;
      }
      _queuePromise = _queuePromise.then(_resolveTransition);
   }

   return _queuePromise;
   // .then(function(result) {
   //     removeBusyState();
   //     return result;
   // }, function(error) {
   //     removeBusyState();
   //     return _queuePromise;
   // });
}

/**
 * @class
 * @requires ojs/ojcore
 * @requires knockout
 * @since 1.1.0
 * @classdesc
 * <h3>JET Router</h3>
 * <p>The router is designed to simplify writing navigation for Single Page Applications.
 * The approach taken is to think of navigation in terms of states and transitions instead
 * of URLs and hashes. A router is always in one in a number of possible states and when
 * a UI action is taken in the application, a transition between states is executed. The
 * router is responsible to properly format the URL to reflect the current state and to
 * restore the application to the matching state when the URL changes.
 * <p>Building navigation is done in three steps:
 * <h6>Define the states that can be taken by the router:</h6>
 * <pre class="prettyprint"><code>
 * var router = oj.Router.rootInstance;
 * // Add three states to the router with id 'home', 'book' and 'tables
 * router.configure({
 *    'home':   { label: 'Home',   value: 'homeContent', isDefault: true },
 *    'book':   { label: 'Book',   value: 'bookContent' },
 *    'tables': { label: 'Tables', value: 'tablesContent' }
 * });
 *
 * var viewModel = {
 *    router: router
 * };
 *
 * oj.Router.sync().then(
 *    function() {
 *       ko.applyBindings(viewModel);
 *    },
 *    function(error) {
 *       oj.Logger.error('Error when starting router: ' + error.message);
 *    });
 * </code></pre>
 *
 * <h6>Trigger a state transition when user ask to navigate:</h6>
 * <pre class="prettyprint"><code>
 * &lt;div id="routing-container">
 *    &lt;div id='buttons-container' data-bind="foreach: router.states">
 *      &lt;!-- Use the go function of the state as the handler for a click binding -->
 *      &lt;input type="button"
 *             data-bind="click: go,  attr: {id: id},
 *             ojComponent: {component: 'ojButton', label: label}"/>
 *    &lt;/div>
 * &lt;/div>
 * </code></pre>
 *
 * <h6>Listen to the state change and updates the dependent parts:</h6>
 * <pre class="prettyprint"><code>
 * &lt;!-- Display the content of the current state -->
 * &lt;h2 id="pageContent" data-bind="text: router.currentValue"/>
 * </code></pre>
 *
 * @desc
 * A Router cannot be instantiated. A static Router is created when the module is loaded and can be
 * accessed using the method {@link oj.Router.rootInstance|rootInstance}.
 * A child router can be created using the method {@link oj.Router#createChildRouter|createChildRouter}.
 * @see oj.Router.rootInstance
 * @see oj.Router#createChildRouter
 * @constructor
 * @hideconstructor
 * @export
 * @ojtsimport knockout
 */
oj.Router = function(key, parentRouter, parentState) {
   var router = this;

   /**
    * A string identifier of the router. It is required the name is unique within all the
    * sibling routers.
    * @name oj.Router#name
    * @member
    * @readonly
    * @type {!string}
    * @see oj.Router#createChildRouter
    */
   this._name = key;

   /**
    * The state of the parent router when this router is current.
    * @private
    * @type {!string | undefined}
    */
   this._parentState = parentState || (parentRouter ? _getShortId(parentRouter._stateId()) : undefined);

   /**
    * The parent router. Root router doesn't have one.
    * @private
    * @type {oj.Router | undefined}
    */
   this._parentRouter = parentRouter;

   /**
    * Array of child router.
    * @private
    * @type {Array.<oj.Router>}
    */
   this._childRouters = [];

   /**
    * Used to store the bookmarkable data.
    * @private
    * @type {Object|undefined}
    */
   this._extra = undefined;

   /**
    * A Knockout observable for the id of the current state of the router.
    * @private
    */
   this._stateId = ko.observable();

   /**
    * A Knockout observable for the id of the current state.<br>
    * <code class="prettyprint">stateId()</code> returns the string id.<br>
    * <code class="prettyprint">stateId('book')</code> transitions the router to
    * the state with id 'book'.<br>
    * It is convenient to use the stateId observable when working with component
    * with 2-way binding like {@link oj.ojButtonset#checked|checked} for
    * <code class="prettyprint">ojButtonset</code> or
    * {@link oj.ojNavigationList#selection|selection} for
    * <code class="prettyprint">ojNavigationList</code> because it does not
    * require a click on optionChange handler (See example below).
    * @name oj.Router#stateId
    * @type {function(string=): string}
    * @readonly
    *
    * @example <caption>A buttonSet using the router stateId for 2-way binding:</caption>
    * &lt;div data-bind="ojComponent: { component: 'ojButtonset',
    *                                checked: router.stateId,
    *                                focusManagement:'none' }"&gt;
    *    &lt;!-- ko foreach: router.states --&gt;
    *       &lt;label data-bind="attr: {for: id}"&gt;&lt;/label&gt;
    *       &lt;input type="radio" name="chapter"
    *              data-bind="value: id, attr: {id: id},
    *                         ojComponent: { component: 'ojButton',
    *                                        label: label}"/&gt;
    *    &lt;!-- /ko --&gt;
    * &lt;/div&gt;
    *
    */
   this._stateIdComp = ko.pureComputed({
      'read': function() {
         // Return only the significant part of the id, the part without the state parameters.
         return _getShortId(this._stateId());
      },
      'write': function(value) {
         this.go(value).then(
            null,
            function(error) {
               throw error;
            });
      },
      'owner': router
   });

   /**
    * An array of all the possible states of the router. This array is null if the router is configured
    * using a callback.
    * @name oj.Router#states
    * @type {Array.<oj.RouterState>|null}
    * @readonly
    * @see oj.RouterState
    */
   this._states = null;

   /**
    * The state id of the default state for this router. The value is set when
    * {@link oj.Router#configure|configure} is called on the router and the state isDefault property is true.
    * If it is undefined, the router will start without a state selected.
    * This property is writable and can be used to set the default state id when
    * the router is configured using a callback.
    * @name oj.Router#defaultStateId
    * @type {string|undefined}
    */
   this._defaultStateId = undefined;

   /**
    * A Knockout observable on the current {@link oj.RouterState|RouterState} object.
    * @name oj.Router#currentState
    * @type {function():(oj.RouterState|undefined)}
    * @readonly
    *
    * @example <caption>Hide a panel when the state of the router is not yet defined:</caption>
    *    &lt;div data-bind="if: router.currentState()"&gt;
    *       &lt;!-- content of the panel --&gt;
    *    &lt;/div&gt;
    */
   this._currentState = ko.pureComputed(function() {
         var shortId = _getShortId(router._stateId());
         return ko.ignoreDependencies(router.stateFromIdCallback, router, [shortId]);
      });


   /**
    * A Knockout observable on the value property of the current state.<br>
    * The state value property is the part of the state object that will be used in the application.
    * It is a shortcut for <code class="prettyprint">router.currentState().value;</code>
    * @name oj.Router#currentValue
    * @type {function():(string|undefined)}
    * @readonly
    *
    * @example <caption>Display the content of the current state:</caption>
    * &lt;h2 id="pageContent" data-bind="text: router.currentValue"/&gt;
    */
   this._currentValue = ko.pureComputed(function() {
         var shortId = _getShortId(router._stateId());
         var retValue;
         var currentState = ko.ignoreDependencies(router.stateFromIdCallback, router, [shortId]);
         if (currentState) {
            retValue = currentState.value;
         }
         return retValue;
      });

   /**
    * The current navigation direction of the router, used for module animations.
    * The value will either be undefined if navigating forward, or "back" if
    * navigating backwards.
    * This is the same value available as part of {@link oj.Router#moduleConfig|moduleConfig}.
    * @name oj.Router#direction
    * @type {string|undefined}
    * @readonly
    * @since 5.0.0
    * @ojstatus preview
    */
   this._navigationType = undefined;

   /**
    * Keep track for history to managed the animation direction
    * @private
    * @type {Array}
    */
   this._navHistory = [];

   /**
    * An object used to pass Router information to ojModule in the moduleConfig
    * object.
    * @ignore
    * @constructor
    * @private
    */
   function _RouterParams() {
      Object.defineProperties(this, {
         'parentRouter': {
            value: router,
            enumerable: true
         },
         'direction': {
            get: function () {
               return router._navigationType;
            },
            enumerable: true
         }
      });
   }

   /**
    * An object to simplify integration between ojRouter and ojModule.
    * Use this object to configure an ojModule where the module name is the router
    * state. When the router changes state, ojModule will automatically load and
    * render the content of a new module based on the name specified in the
    * {@link oj.RouterState#value|value} or {@link oj.RouterState#id|id} property
    * of the current {@link oj.RouterState|RouterState}.<br>
    * The object moduleConfig provides the following functionality to the ojModule binding:
    * <ol>
    *   <li>it defines the name of the module by setting the <code class="prettyprint">name</code>
    * option to the value of the current router state.</li>
    *   <li>it makes the parent router accessible to the module using the
    * <code class="prettyprint">params['ojRouter']['parentRouter']</code> property.</li>
    *   <li>it defines a direction hint that can be use for the module animation.</li>
    *   <li>it makes the callback <code class="prettyprint">canExit</code> invokable on the
    * viewModel. If <code class="prettyprint">canExit</code> is not defined on the viewModel,
    * it will be invoked on the {@link oj.RouterState|RouterState}.</li>
    * </ol>
    * The moduleConfig object has the following properties:
    * <ul>
    *   <li><code class="prettyprint">name</code>: is set to the {@link oj.RouterState#value|value}
    * property of the current state of the router.  If a current state is not defined, the default
    * state is used and if the default state is not defined, the first state is used. On the
    * state object, if the <code class="prettyprint">value</code> is not defined or if it is
    * not a string, the {@link oj.RouterState#id|id} property is used.</li>
    *   <li><code class="prettyprint">params</code>: an object with a property named
    * <code class="prettyprint">ojRouter</code> which value is a object with two properties
    * <code class="prettyprint">parentRouter</code> which value is the parent router and
    * <code class="prettyprint">direction</code> which value is undefined or the string 'back'. In
    * JET version 1.1, the parent router was the entire params object making it impossible to pass
    * any other parameters. Application built using version 1.1 of JET now need to retrieve
    * the parent router from the <code class="prettyprint">ojRouter.parentRouter</code> property of
    * <code class="prettyprint">params</code>.</li>
    *   <li><code class="prettyprint">lifecycleListener</code>: an object implementing the
    * <code class="prettyprint">attached</code> callback to bind canExit to the router if it is
    * defined on the viewModel. If you override <code class="prettyprint">attached</code>, this
    * functionality will be lost unless you set the viewModel to the current state of the parent
    * router in your custom <code class="prettyprint">attached</code> implementation.</li>
    * </ul>
    *
    * The router calculate the direction of the navigation and make it available to the child module
    * using the parameter <code class="prettyprint">ojRouter.direction</code>.<br> This can be used
    * to specify a different module animation when going 'back'. The value of direction is either
    * undefined or 'back'. It is 'back' when the state transition is caused by a back button on the
    * browser and the new state is equal to the previous state.

    * To customize the behavior of the moduleduleConfig object, it is possible to create your own
    * moduleConfig and merge other properties or modifies the value of existing properties.
    * It is recommended to use $.extend as described in the third example below.
    *
    * @name oj.Router#moduleConfig
    * @type {Object}
    * @readonly
    * @property {KnockoutObservable<string>} name
    * @property {Object} params
    * @property {Object} params.ojRouter
    * @property {oj.Router} params.ojRouter.parentRouter
    * @property {string} params.ojRouter.direction
    * @property {Object} lifecycleListener
    * @property {function(string):void} lifecycleListener.attached
    *
    * @example <caption>Configure an ojModule binding with a router</caption>
    * &lt;!-- This is where your main page content will be loaded --&gt;
    * &lt;div id="mainContainer" data-bind="ojModule: router.moduleConfig"&gt;&lt;/div&gt;
    *
    * @example <caption>Creates a child router in the viewModel of a module</caption>
    * var viewModel = {
    *    initialize: function(params) {
    *       // Retrieve the parent router from the parameters
    *       var parentRouter = params.valueAccessor().params['ojRouter']['parentRouter'];
    *       // Create a child router for this viewModel
    *       this.router = parentRouter.createChildRouter('chapter')
    *          .configure({
    *             'preface':  { label: 'Preface',   value: storage['preface']  },
    *             'chapter1': { label: 'Chapter 1', value: storage['chapter1'] },
    *             'chapter2': { label: 'Chapter 2', value: storage['chapter2'] },
    *             'chapter3': { label: 'Chapter 3', value: storage['chapter3'] }
    *          });
    *       oj.Router.sync();
    *    },
    *
    *    // canExit callback will be called here
    *    canExit: function() {
    *       return (okToExit) ? true: false;
    *    }
    * };
    *
    * @example <caption>Creates a custom moduleConfig replacing the name property</caption>
    * dynamicConfig = ko.pureComputed(function () {
    *    if (smallOnly()) {
    *       // Add the prefix 'phone/' to change the viewModel location
    *       return $.extend({},
    *                       router.moduleConfig,
    *                       {name: 'phone/' + router.moduleConfig.name()});
    *    }
    *    return router.moduleConfig;
    * });
    */
   this._moduleConfig = Object.create(null, {
      'name': {
         value: ko.pureComputed(function() {
            var retValue;

            // ojModule name cannot afford to be null
            var stateId = _getShortId(this._stateId()) || this._defaultStateId || this._states[0];
            var currentState = this.stateFromIdCallback(stateId);
            if (currentState) {
               retValue = currentState.value;
               if (!retValue || (typeof retValue !== 'string')) {
                  retValue = currentState._id;
               }
            }
            return retValue;
         }, router),
         enumerable: true
      },
      'params': {
         value: Object.create(null, {
            'ojRouter': {
               value: new _RouterParams(),
               enumerable: true
            }
         }),
         enumerable: true
      },
      'lifecycleListener': {
         value: Object.create(null, {
            'attached': {
               value: function(params) {
                  // Retrieve router passed as a parameter to ojModule using params defined above.
                  var paramRouter = ko.unwrap(params['valueAccessor']())['params']['ojRouter']['parentRouter'],
                      state = paramRouter._currentState();
                  if (state) {
                     state.viewModel = params['viewModel'];
                  }
               },
               // User can change attached
               writable: true, enumerable: true
            }
         }),
         enumerable: true
      }
   });

    /**
     * @typedef {object} oj.Router.ModuleConfigType
     * @property {KnockoutObservable<string>} name The value of the current state.  See {@link oj.Router#currentValue}
     * @property {Object} params
     * @property {Object} params.ojRouter
     * @property {oj.Router} params.ojRouter.parentRouter A reference to the current router
     * @property {string} params.ojRouter.direction The animation direction of the module, if defined.
     * @property {Object} params.ojRouter.parameters An object containing the observable parameters of the current router state.
     * @property {Object} lifecycleListener
     * @property {function(Object):void} lifecycleListener.attached The 'attached' listener for the module
     */

    /**
     * Similar to {@link oj.Router#moduleConfig}, this object is meant to be used to configure
     * a module binding which reacts to changes in the router state or its parameters.
     * This configuration also creates observables around all of the
     * {@link oj.RouterState#parameters} found in the router state.
     <p>
     This observable is dyanmically created only when it's first requested.  To
     use observableModuleConfig and its observable parameters, first reference
     the property in your own code, which
     will setup the observable and add observable state parameters.  To use
     the observable parameters, retrieve the instance from the parameters passed
     to your view model.
     </p>
     @example <caption>Use observable parameters</caption>
     var root = oj.Router.rootInstance;
     root.configure(...);
     var mc = root.observableModuleConfig;
     ...
     function ViewModel(viewParams) {
        var router = viewParams['ojRouter']['parentRouter'];
        var currentState = router.currentState();
        // Retrieve object containing observable parameters
        var params = viewParams['ojRouter']['parameters'];
        var employeeId = params['employeeId'];
        // Get an observable parameter's value
        var idValue = employeeId();

        // Optionally subscribe to a parameter's value change
        employeeId.subscribe(function(newId) {
        });
     }
     * @name oj.Router#observableModuleConfig
     * @type {Object}
     * @since 4.2.0
     * @readonly
     * @ojstatus preview
     * @ojsignature {target: "Type", value: "KnockoutObservable<oj.Router.ModuleConfigType>", jsdocOverride: true}
     */
    this._getObservableModuleConfig = function() {
        if (!this._observableModuleConfig) {
            var currentStateObservable = router['currentState'];

            var currentStateVal = currentStateObservable.peek();

            var config2 = ko.observable(_getConfigObject(currentStateVal, router));

            // Subscribe to the changes and update config2 appropriately
            currentStateObservable['subscribe'](function (state) {
                var oldConfig = config2.peek();
                var oldName = oldConfig['name'];
                var newName = _getNameFromState(state);
                if (oldName != newName) {
                    config2(_getConfigObject(state, router));// all new config value when the name changes
                }
                else {
                    // update parameters only
                    var params = oldConfig['params']['ojRouter'][_parametersValue];
                    var value;
                    for (var key in params) {
                        value = state[_parametersValue][key];
                        params[key](value);
                    }
                }
            });
            this._observableModuleConfig = config2;
        }
        return this._observableModuleConfig;
    }
    function _getConfigObject(currentState, router) {
        var obj = {};

        obj['name'] = router['moduleConfig']['name'];

        if (currentState) {
            var params = {};
            obj['params'] = params;
            var paramsFromRouter = params['ojRouter'] = {};
            paramsFromRouter['parentRouter'] = router;

            Object.defineProperty(paramsFromRouter, 'direction',
            {
                get : function () {
                    return router._navigationType;
                },
                enumerable : true
            });
            var observableParams = paramsFromRouter[_parametersValue] = {};

            var routerParams = currentState[_parametersValue];
            var keys = Object.keys(routerParams);
            var value;
            keys.forEach(function (key) {
                value = routerParams[key];
                observableParams[key] = ko.observable(value);
            });

            obj['lifecycleListener'] = function (prms) {
                var state = router.currentState();
                if (state) {
                    state.viewModel = prms['viewModel'];
                }
            };

        }

        return obj;
    }

    function _getNameFromState(currentState) {
        var name = 'oj-blank';// deal with empty state
        if (currentState) {
            name = currentState['value'];
            if (!name || (typeof name !== 'string')) {
                name = currentState._id;
            }
            return name;
        }
    }

    Object.defineProperties(this, {
       'parent': { value:
          /**
           * The parent router if it exits.
           * Only the 'root' router does not have a parent router.
           * @name oj.Router#parent
           * @member
           * @type {oj.Router|undefined}
           * @readonly
           */
          (this._parentRouter), enumerable: true
       }
    });

};

Object.defineProperties(oj.Router.prototype, {
   'name': { get: function () { return this._name; }, enumerable: true },
   'states': { get: function () { return this._states; }, enumerable: true },
   'stateId': { get: function () { return this._stateIdComp; }, enumerable: true },
   'currentState': { get: function () { return this._currentState; }, enumerable: true },
   'currentValue': { get: function () { return this._currentValue; }, enumerable: true },
   'direction': { get: function () { return this._navigationType; }, enumerable: true },
   'defaultStateId': { get: function () { return this._defaultStateId; },
                       set: function(newValue) { this._defaultStateId = newValue; },
                       enumerable: true },
   'moduleConfig': { get: function () { return this._moduleConfig; }, enumerable: true },
   'observableModuleConfig': { get: function () { return this._getObservableModuleConfig(); }, enumerable: true }
});

/**
 * Create the instance of the root router.
 */
rootRouter = new oj.Router(_DEFAULT_ROOT_NAME, undefined, undefined);

/**
 * Function use to handle the popstate event.
 */
function handlePopState() {
   var i, sr;
   var sId = _getShortId(rootRouter._stateId());
   var subRouter = null;

   oj.Logger.info('Handling popState event with URL: %s', _location.href);

   // First retrieve the sub-router associated with the current state, if there is one.
   if (sId) {
      for (i = 0; i < rootRouter._childRouters.length; i++) {
         sr = rootRouter._childRouters[i];
         if (sId === sr._parentState) {
            subRouter = sr;
            break;
         }
      }
   }

   _queueTransition({ router: subRouter, origin: 'popState' });
}

/**
 * Return a child router by name.  The name is the value given to
 * {@link oj.Router#createChildRouter|createChildRouter}.
 * @param  {!string} name The name of of the child router to find
 * @return {oj.Router|undefined} The child router
 * @since 1.2.0
 * @export
 */
oj.Router.prototype.getChildRouter = function(name) {
   var childRouter;

   if (name && (typeof name === 'string')) {
      name = name.trim();
      if (name.length > 0) {
         this._childRouters.every(function(sr) {
            if (sr._name === name) {
               childRouter = sr;
               return false;
            }
            return true;
         });
      }
   }

   return childRouter;
};

/**
 * Get the child router associated with the parent's current state.  See
 * {@link oj.Router#createChildRouter|createChildRouter} for details on how child routers are associated
 * with parent states.
 * @return {oj.Router|undefined} The child router for the current state, if defined.
 * @since 5.0.0
 * @export
 * @ojstatus preview
 */
oj.Router.prototype.getCurrentChildRouter = function() {
  var sId = _getShortId(this._stateId() || this._defaultStateId);
  var child = _getChildRouter(this, sId);

  return child;
}

/**
 * Create a child router with the given name. A router can either have one child
 * router that handles all possible states, or one child router per state.
 * See the examples below.
 * @param {!string} name The unique name representing the router.  The name is
 * used by the function {@link oj.Router#getChildRouter|getChildRouter} to retrieve
 * the child router.
 * @param {string=} parentStateId The state Id of the parent router for determining
 * when this child router is used.
 * If not defined, the child router is created for the current state of the router.
 * @return {oj.Router} the child router
 * @throws An error if a child router exist with the same name or if the current
 * state already has a child router.
 * @export
 * @example <caption>
 * <b>Create a default child router for the parent</b>
 * In this example, the parent router is assumed to have no current state (it
 * has not yet navigated to any particular state). Since we are not specifying a
 * value for parentStateId, the newly created router will be the default child
 * router.
 * </caption>
 * // Parent router has no current state
 * router = oj.Router.rootInstance;
 * // This child router is the default child router for all parent router states
 * childRouter = router.createChildRouter('chapter');
 * @example <caption>
 * <b>Create a child router for the root's current state</b>
 * In this example, the parent has navigated to a state before the child router
 * is created.  Even though no value is given for parentStateId, the child router
 * is only used when the parent is in the particular state.
 * </caption>
 * // Parent router navigates to a given state
 * router = oj.Router.rootInstance;
 * router.go('book').then(function(result) {
 *   if (result.hasChanged) {
 *     // Child router is only used when parent router's state is 'book'
 *     // because parent now has a current state
 *     var childRouter = router.createChildRouter('chapter');
 *   }
 * });
 * @example <caption>
 * <b>Create a child router for parent state id 'book'</b>
 * In this example, the parent router hasn't yet navigated to a particular state
 * but the child specifies 'book' as its parentStateId, therefore, the child will
 * only be used when the parent is in that particular state.
 * </caption>
 * // Parent router has no current state
 * router = oj.Router.rootInstance;
 * // Child router is only used when parent router's state is 'book'
 * childRouter = router.createChildRouter('chapter', 'book');
 */
oj.Router.prototype.createChildRouter = function(name, parentStateId) {
   var i,
       sr,
       childRouter;

   oj.Assert.assertString(name);

   parentStateId = parentStateId||_getShortId(this._stateId());

   name = encodeURIComponent(name.trim());
   // Make sure it doesn't already exist.
   for (i = 0; i < this._childRouters.length; i++) {
      sr = this._childRouters[i];
      if (sr._name === name) {
         throw new Error('Invalid router name "' + name + '", it already exists.');
      }
      else if (sr._parentState === parentStateId) {
         throw new Error('Cannot create more than one child router for parent state id "' + sr._parentState + '".');
      }
   }

   childRouter = new oj.Router(name, this, parentStateId);

   this._childRouters.push(childRouter);

   return childRouter;
};

/**
 * @private
 * @param {string} stateId The state id.
 * @return {oj.RouterState | undefined} The state object.
 */
oj.Router.prototype.stateFromIdCallback = function(stateId) {
   return getStateFromId(this, stateId);
};

/**
 * Configure the states of the router. The router can be configured in two ways:
 * <ul>
 *  <li>By describing all of the possible states that can be taken by this router.</li>
 *  <li>By providing a callback returning a {@link oj.RouterState|RouterState}
 *      object given a string state id.</li>
 * </ul>
 * This operation resets any previous configuration, and is chainable.<br>
 * Configuring {@link oj.RouterState#parameteters router state parameters} should
 * be done here.  See the example below.
 *
 * @param {!(Object.<string, {oj.RouterState.ConfigOptions}> |
 *         function(string): (oj.RouterState | undefined)) } option
 * Either a callback or a dictionary of states.
 * <h6>A callback:</h6>
 * <h4 id="stateFromIdCallback" class="name">
 *    stateFromIdCallback
 *    <span class="signature">(stateId)</span>
 *    <span class="type-signature">
 *       → {<a href="oj.RouterState.html">oj.RouterState</a>|undefined}
 *    </span>
 * </h4>
 * A function returning a {@link oj.RouterState|RouterState} given a string state id.<br>
 * When using a callback, the {@link oj.Router.states|states} property will always be null since
 * states are defined on the fly.<br>See second example below.
 * <h6>A dictionary of states:</h6>
 * It is a dictionary in which the keys are state {@link oj.Router#id|id}s and values are objects
 * defining the state.  Note that the forward slash character '/' is not allowed
 * in the state Id.  See {@link oj.RouterState~ConfigOptions|ConfigOptions}.<br>See first example below.
 * <h6>Key</h6>
 * <table class="params">
 *   <thead><tr>
 *     <th>Type</th>
 *     <th class="last">Description</th>
 *   </tr></thead>
 *   <tbody>
 *     <tr>
 *       <td class="type">
 *         <span class="param-type">string</span>
 *       </td>
 *       <td class="description last">the state id.
 *       See the RouterState <a href="oj.RouterState.html#id">id</a> property.</td>
 *    </tr>
 *    <tr>
 *      <td class="type">
 *        <span class="param-type">oj.RouterState.ConfigOptions</span>
 *      </td>
 *      <td class="value">
 *        <span class="param-type">See {@link oj.RouterState~ConfigOptions|ConfigOptions}
 *        for the options available for configuring router states</span>
 *      </td>
 *   </tbody>
 * </table>
 * @export
 * @see oj.RouterState
 * @ojsignature { target:'Type',
 *              value: '{[key:string]: oj.RouterState.ConfigOptions}|((id:string)=> oj.RouterState)',
 *              for: 'option'}
 * @example <caption>Add three states with id 'home', 'book' and 'tables':</caption>
 * router.configure({
 *    'home':   { label: 'Home',   value: 'homeContent', isDefault: true },
 *    'book':   { label: 'Book',   value: 'bookContent' },
 *    'tables': { label: 'Tables', value: 'tablesContent' }
 * });
 * @example <caption>Configure dynamic states via callback function:</caption>
 * router.configure(function(stateId) {
 *    var state;
 *
 *    if (stateId) {
 *       state = new oj.RouterState(stateId, { value: data[stateId] }, router);
 *    }
 *    return state;
 * });
 * @example <caption>Configuring {@link oj.RouterState#parameters state parameters}:</caption>
 * router.configure({
 *    'home':   { label: 'Home',   value: 'homeContent', isDefault: true },
 *    'book/{chapter}/{paragraph}':   { label: 'Book',   value: 'bookContent' },
 * });
 */
oj.Router.prototype.configure = function(option) {
   this._stateId(undefined);
   delete this._defaultStateId;
   // StateId are changing so erase history.
   this._navigationType = undefined;
   this._navHistory = [];

   if (typeof option === 'function') {
      this._states = null;
      // Override prototype
      this.stateFromIdCallback = option;
   }
   else {
      this._states = [];
      // Restore prototype
      delete this.stateFromIdCallback;

      Object.keys(option).forEach(function(key) {
         var rsOptions = option[key];
         this._states.push(new oj.RouterState(key, rsOptions, this));
         // Set the defaultStateId of the router from the isDefault property
         if ((typeof(rsOptions['isDefault']) === 'boolean') && rsOptions['isDefault']) {
            this._defaultStateId = key;
         }
      }, this);
   }

   return this;
};

/**
 * Return the {@link oj.RouterState} object which state id matches one of the possible states of the router.
 * @param {string} stateId - the id of the requested {@link oj.RouterState} object.
 * @return {oj.RouterState|undefined} the state object matching the id.
 * @export
 * @example <caption>Retrieve the RouterState for id 'home':</caption>
 * var homeState = router.getState('home');
 * var homeStateValue = homeState.value;
 */
oj.Router.prototype.getState = function(stateId) {
   return this.stateFromIdCallback(stateId);
};

/**
 * Initialize the default for oj.Router if needed. Dispose on the root router
 * will de-initialize.
 * @private
 */
function _initialize() {
   if (!_initialized) {
      if (!_urlAdapter) {
         _urlAdapter = new oj.Router.urlPathAdapter();
      }

      _urlAdapter.init(_baseUrlProp);

      _originalTitle = window.document.title;

      /**
       * Listen to URL changes caused by back/forward button
       * using the popstate event. Call handlePopState to dispatch the change of URL.
       */
      window.addEventListener(_POPSTATE, handlePopState, false);

      oj.Logger.info('Initializing rootInstance.');
      oj.Logger.info('Base URL is %s', _baseUrlProp);
      oj.Logger.info('Current URL is %s', _location.href);

      _initialized = true;
   }
}

/**
 * Convert forward slash characters used for paths to a character which doesn't
 * require URL encoding (to save space in the URL) and is unambiguous from the
 * path separator.
 * @param       {string} value The string whose slash characters are to be encoded.
 * @return      {string}      An encoded form of the string
 * @private
 */
function _encodeSlash(value) {
  var enc = value;
  if (enc && enc.replace) {
    enc = enc.replace(/~/g, '~0');
    enc = enc.replace(/\//g, '~1');
  }
  return enc;
}

/**
 * Given an encoded string from _encodeSlash, decode the characters to restore
 * the value to its original form.
 * @param       {string} value The string whose encoded slash values will be decoded.
 * @return      {string}      The decoded form of the string
 * @private
 */
function _decodeSlash(value) {
  var dec = value;
  if (dec && dec.replace) {
    dec = dec.replace(/~1/g, '/');
    dec = dec.replace(/~0/g, '~');
  }
  return dec;
}

/**
 * Go is used to transition to a new state using a path made of state ids separated by a slash.  The
 * path can be absolute or relative.<br>
 * <br>
 * Example of valid path:
 * <ul>
 *   <li><code class="prettyprint">router.go('home')</code>: transition router
 *    to state id 'home'</li>
 *   <li><code class="prettyprint">router.go('/book/chapt2')</code>: transition
 *    the root instance to state id 'book' and the child router to state id
 *    'chapt2'</li>
 *   <li><code class="prettyprint">router.go('chapt2/edit')</code>: transition
 *   router to state id 'chapt2' and child router to state id 'edit'</li>
 *   <li><code class="prettyprint">router.go(['chapt2','edit'])</code>: equivalent
 *   to the previous transition, but using an array of path strings in place of
 *   forward slashes.
 * </ul>
 * <br>
 * If the stateIdPath argument is <code class="prettyprint">undefined</code> or an empty string, go
 * transition to the default state of the router.<br>
 * A {@link oj.Router.transitionedToState|transitionedToState} signal is
 * dispatched when the state transition has completed.
 * @param {(string|Array.<string>)=} stateIdPath A path of ids representing the state to
 * which to transition, separated by forward slashes (/).  This can also be an Array
 * of strings, each segment representing individual states.  An array is typically used
 * if the forward slash should be part of the state Id and needs to be distinguished
 * from the path separators.
 * @param {Object=} options - an object with additional information on how to execute the transition.
 * @param {string} options.historyUpdate Specify how the transition should act on the browser
 * history. If this property is not specified, a new URL is added to the history.<br>
 * <b><i>Supported Values:</i></b><br>
 * <code class="prettyprint">'skip'</code>: does not update the history with the new URL<br>
 * <code class="prettyprint">'replace'</code>: modifies the current history with the new URL
 * @return {!Promise.<{hasChanged: boolean}>} A Promise that resolves when the
 * router is done with the state transition.<br>
 * When the promise is fullfilled, the parameter value is an object with the property
 * <code class="prettyprint">hasChanged</code>.<br>
 * The value of <code class="prettyprint">hasChanged</code> is:
 * <ul>
 *   <li>true: If the router state changed.</li>
 * </ul>
 * When the Promise is rejected, the parameter value is:
 * <ul>
 *   <li>An Error object stipulating the reason for the rejection during the
 *   resolution. Possible errors are:
 *   <ul>
 *     <li>If stateIdPath is defined but is not of type string.</li>
 *     <li>If stateIdPath is undefined but the router has no default state.</li>
 *     <li>If a state id part of the path cannot be found in a router.</li>
 *   </ul>
 *   </li>
 * </ul>
 * @export
 * @example <caption>Transition a router to the state id 'home':</caption>
 * router.go('home');
 * @example <caption>Transition a router to its default state and handle errors:</caption>
 * router.go().then(
 *    function(result) {
 *       if (result.hasChanged) {
 *          oj.Logger.info('Router transitioned to default state.');
 *       }
 *       else {
 *          oj.Logger.info('No transition, Router was already in default state.');
 *       }
 *    },
 *    function(error) {
 *       oj.Logger.error('Transition to default state failed: ' + error.message);
 *    }
 * );
 * @example <caption>Transition a router to state id 'stepB' without updating the URL:</caption>
 * wizardRouter.go('stepB', { historyUpdate: 'skip' });
 */
oj.Router.prototype.go = function(stateIdPath, options) {
   _initialize();

   options = options || [];

   if (Array.isArray(stateIdPath)) {
     stateIdPath = stateIdPath.map(_encodeSlash).join('/');
   }

   return _queueTransition({ router: this, path: stateIdPath, origin: 'go',
                             historyUpdate: options['historyUpdate'] });
};

/**
 * Internal go used by _executeTransition
 * @private
 * @param  {Object} transition An object with properties describing the transition
 * @return {*} A Promise that resolves when the routing is done
 */
oj.Router.prototype._go = function(transition) {
   var path,
       newStates,
       useDefault = true,
       stateIdPath = transition.path,
       replace = false,
       skip = false;

   switch (transition.historyUpdate) {
      case 'skip':
         skip = true;
         break;
      case 'replace':
         replace = true;
   }

   if (stateIdPath) {
      if (typeof stateIdPath === 'string') {
         useDefault = false;
      }
      else {
         return Promise.reject(new Error('Invalid object type for state id.'));
      }
   }

   if (useDefault) {
      stateIdPath = this._defaultStateId;
      if (!stateIdPath) {
         // No default defined, so nowhere to go.
        oj.Logger.info(function() {
            return 'Undefined state id with no default id on router '+
                getRouterFullName(this);
        });
        return Promise.resolve(_NO_CHANGE_OBJECT);
      }
   }

   // Absolute or relative?
   if ('/' === stateIdPath.charAt(0)) {
      path = stateIdPath;
   }
   else {
      path = getCurrentPath(this._parentRouter);
      if (!path) {
         return Promise.reject(new Error('Invalid path "' + stateIdPath +
                               '". The parent router does not have a current state.'));
      }
      path += stateIdPath;
   }

   oj.Logger.info('Destination path: %s', path);

   try {
      newStates = _buildStateFromPath(this, path);
      newStates = _appendOtherChanges(newStates);
   }
   catch (err) {
      return Promise.reject(err);
   }

   // It is important that we do not call canEnter on state that we not going to enter so
   // only keep changes where the value doesn't match the current router state.
   // reducedState is an array of object with 2 properties, value and router.
   var reducedState = _filterNewState(newStates);

   // Only transition if replace is true or if the new state is different.
   // When replace is true, it is possible the states are the same (by example when going to the
   // default state of a child router) but the transition still need to be executed.
   if (replace || reducedState.length > 0) {
      oj.Logger.info('Deferred mode or new state is different.');
      return _canExit(this).then(function (canExit) {
         if (canExit) {
            // Only calls canEnter callback on state that are changing
            return _canEnter(reducedState).then(_updateAll).then(function(params) {
               if (params[_HAS_CHANGED]) {
                  if (skip) {
                     oj.Logger.info('Skip history update.');
                  }
                  else {
                     var url = _urlAdapter.buildUrlFromStates(newStates);
                     oj.Logger.info('%s URL to %s', replace ? 'Replacing' : 'Pushing', url);
                     window.history[replace ? 'replaceState' : 'pushState'](null, '', url);
                  }
               }
               return params;
            });
         }
         return Promise.resolve(_NO_CHANGE_OBJECT);
      });
   }

   return Promise.resolve(_NO_CHANGE_OBJECT);
};

/**
 * Store additional data for this router that will be added in a compressed form to the URL
 * so it can be bookmarked. When calling this method, the URL is immediately modified.
 * @param {!Object} data the data to store with this state.
 * @throws An error if the bookmarkable state is too big.
 * @return {undefined}
 * @export
 * @example <caption>Store a color in the URL:</caption>
 * try {
 *    var color = '#99CCFF';
 *    router.store(color);
 *    $('#chapter').css('background', color);
 * }
 * catch (error) {
 *    oj.Logger.error('Error while storing data: ' + error.message);
 * }
 */
oj.Router.prototype.store = function(data) {
   this._extra = data;

   var extraState = {},
       router = this;

   // Walk the parent routers
   while (router) {
      if (router._extra !== undefined) {
         extraState[router._name] = router._extra;
      }
      router = router._parentRouter;
   }

   // and the children routers
   router = this;
   var nextLevel, i, sr, shortId;
   while (router) {
      for (i = 0; i < router._childRouters.length; i++) {
         sr = router._childRouters[i];
         shortId = _getShortId(router._stateId());
         if (shortId && shortId === sr._parentState) {
            if (sr._extra !== undefined) {
               extraState[sr._name] = sr._extra;
            }
            nextLevel = sr;
            break;
         }
      }
      router = nextLevel;
      nextLevel = undefined;
   }

   window.history.replaceState(null, '', _buildUrl({}, extraState));
};

/**
 * Retrieve the additional data stored in the URL.
 * @return {*} the content stored in the URL
 * @export
 * @example <caption>Retrieve the value of the background color stored in the URL:</caption>
 *  oj.Router.sync().then(
 *     function() {
 *        var color = viewModel.router.retrieve();
 *        if (color) {
 *           $('#chapter').css('background', color);
 *        }
 *     },
 *     function(error) {
 *        oj.Logger.error('Error during sync: ' + error.message);
 *     }
 *  );
 */
oj.Router.prototype.retrieve = function() {
   return this._extra;
};

/**
 * Dispose the router.<br>
 * Erase all states of this router and its children.
 * Remove itself from parent router child list.<br>
 * When this method is invoked on the {@link oj.Router.rootInstance|rootInstance}, it
 * also remove internal event listeners and re-initialize the
 * {@link oj.Router.defaults|defaults}.
 * @return {undefined}
 * @export
 */
oj.Router.prototype.dispose = function() {
   var parentChildren, i;

   // Depth first
   while (this._childRouters.length > 0) {
      this._childRouters[0].dispose();
   }

   // If this is the root, clean up statics
   if (!this._parentRouter) {
      _baseUrlProp = '/'; // Restore the default value
      _urlAdapter = null;
      this._name = _DEFAULT_ROOT_NAME;
      // Restore title
      window.document.title = _originalTitle;

      window.removeEventListener(_POPSTATE, handlePopState);
      oj.Router._transitionedToState.removeAll();
      _initialized = false;
   }
   else {
      // Remove itself from parent children array.
      parentChildren = this._parentRouter._childRouters;
      for (i = 0; i < parentChildren.length; i++) {
        if (parentChildren[i]._name === this._name) {
           parentChildren.splice(i, 1);
           break;
        }
      }

      delete this._parentState;
   }

   delete this._navigationType;
   this._navHistory = [];
   this._states = null;
   delete this._defaultStateId;
   delete this._extra;

};

/**
 * Alias for property oj.Router.transitionedToState
 * @private
 */
oj.Router._transitionedToState = new signals.Signal();

/**
 * Flag set to true when the router is in the process of updating the states
 * involved in the current transition
 * @private
 * @type {boolean}
 */
oj.Router._updating = false;

Object.defineProperties(oj.Router, {
  'rootInstance': {
    /**
     * The static instance of {@link oj.Router} representing the unique root router.
     * This instance is created at the time the module is loaded.<br>
     * All other routers will be children of this object.<br>
     * The name of this router is 'root'. To change this name use the
     * {@link oj.Router.defaults|rootInstanceName} property.
     * @name oj.Router.rootInstance
     * @type oj.Router
     * @readonly
     * @example <caption>Retrieve the root router and configure it:</caption>
     * var router = oj.Router.rootInstance;
     * router.configure({
     *    'home':   { label: 'Home',   value: 'homeContent', isDefault: true },
     *    'book':   { label: 'Book',   value: 'bookContent' },
     *    'tables': { label: 'Tables', value: 'tablesContent' }
     * });
     */
    value: rootRouter,
    enumerable: true
  },
  'transitionedToState': {
    /**
     * A {@link http://millermedeiros.github.io/js-signals/|signal} dispatched when the state transition
     * has completed either by successfully changing the state or cancelling.<br>
     * The parameter of the event handler is a boolean true when the state has changed.<br>
     * This is usefull when some post processing is needed or to test the result after a state change.
     * @name oj.Router.transitionedToState
     * @type {Object}
     * @readonly
     * @example <caption>Creates promise that resolve when the state transition is complete.</caption>
     * var promise = new Promise(function(resolve, reject) {
     *       oj.Router.transitionedToState.add(function(result) {
     *          if (result.hasChanged) {
     *             oj.Logger.info('The state has changed');
     *          }
     *          resolve();
     *       });
     */
    value: oj.Router._transitionedToState,
    enumerable: true
  }
});

/**
 * A set of Router defaults properties.<br>
 * <h6>Warning: </h6>Defaults can not be changed after the first call to {@link oj.Router.sync|sync()}
 * has been made. To re-initialize the router, you need to call {@link oj.Router#dispose|dispose()} on
 * the {@link oj.Router.rootInstance|rootInstance} first then change the defaults.
 * @property {oj.Router.urlPathAdapter|oj.Router.urlParamAdapter} [urlAdapter] an instance of the url adapter to use. If not specified, the router
 * will be using the path url adapter. Possible values are an instance of
 * {@link oj.Router.urlPathAdapter} or {@link oj.Router.urlParamAdapter}.
 * @property {string} [baseUrl] the base URL to be used for relative URL addresses. The value can be
 * absolute or relative.  If not specified, the default value is '/'.<br>
 * <b>Warning</b>: When using the {@link oj.Router.urlPathAdapter|path URL adapter} it is necessary
 * to set the base URL if your application is not using <code class="prettyprint">index.html</code>
 * or is not starting at the root folder. Using the base URL is the only way the router can retrieve
 * the part of the URL representing the state.<br>
 * @property {string} [rootInstanceName] the name used for the root router. If not defined,
 * the name is 'root'. This is used by the {@link oj.Router.urlParamAdapter|urlParamAdapter} to build
 * the URL in the form of <code class="prettyprint">/index.html?root=book</code>.
 * @export
 * @example <caption>Change the default URL adapter to the urlParamAdapter</caption>
 * oj.Router.defaults['urlAdapter'] = new oj.Router.urlParamAdapter();
 * @example <caption>Set the base URL for an application located at the root and a starting page
 * named <code class="prettyprint">index.html</code>. This is the default.</caption>
 * oj.Router.defaults['baseUrl'] = '/';
 * @example <caption>Set the base URL for an application with a page named
 * <code class="prettyprint">main.html</code> and located in the
 * <code class="prettyprint">/myApp</code> folder.</caption>
 * oj.Router.defaults['baseUrl'] = '/myApp/main.html';
 * @example <caption>Change the default root router name to 'id'</caption>
 * oj.Router.defaults['rootInstanceName'] = 'id';
 */
oj.Router.defaults = {};

Object.defineProperties(oj.Router.defaults, {
   'urlAdapter': {
      get: function() {
         if (!_urlAdapter) {
            _urlAdapter = new oj.Router.urlPathAdapter();
         }
         return _urlAdapter;
      },
      set: function(urlAdapter) {
         if (_initialized) {
            throw new Error('Incorrect operation. Cannot change URL adapter after calling sync() or go().');
         }
         _urlAdapter = urlAdapter;
      },
      enumerable: true,
      readonly: false
   },
   'baseUrl': {
      get: function() {
         return _baseUrlProp;
      },
      set: function(baseUrl) {
         if (_initialized) {
            throw new Error('Incorrect operation. Cannot change base URL after calling sync() or go().');
         }
         if (!baseUrl) {
            _baseUrlProp = '/';
         }
         else {
            // Remove anything after ? or #
            _baseUrlProp = baseUrl.match(/[^?#]+/)[0];
         }
      },
      enumerable: true,
      readonly: false
   },
   'rootInstanceName': {
      get: function() {
         return rootRouter._name;
      },
      set: function(rootName) {
         if (_initialized) {
            throw new Error('Incorrect operation. Cannot change the name of the root instance after calling sync() or go().');
         }

         oj.Assert.assertString(rootName);
         rootRouter._name = encodeURIComponent(rootName.trim());
      },
      enumerable: true,
      readonly: false
   }
});

/**
 * Synchronise the router with the current URL. The process parse the URL and
 * <ol>
 *   <li>transition the router to a new state matching the URL.</li>
 *   <li>initialize the bookmarkable storage.</li>
 *   <li>dispatch a {@link oj.Router.transitionedToState|transitionedToState} signal.</li>
 * </ol>
 * It has to be called after a router is configured, to synchronise the URL with the
 * router state.<br>
 * If a default state is defined, the router will transition to it, otherwise no transition will
 * occur and the router will be in an undefined state.<br>
 * Because the process of transitioning between two states invokes callbacks (canExit, canEnter)
 * that are promises, this function also returns a promise.
 * @return {!Promise.<{hasChanged: boolean}>} A Promise that resolves when the router is done with
 * the state transition.<br>
 * When the Promise is fullfilled, the parameter value is an object with the property
 * <code class="prettyprint">hasChanged</code>.<br>
 * The value of <code class="prettyprint">hasChanged</code> is:
 * <ul>
 *   <li>true: If the router state changed.</li>
 * </ul>
 * When the Promise is rejected, the parameter value is:
 * <ul>
 *   <li>An Error object stipulating the reason for the rejection when an error
 * occurred during the resolution.</li>
 * </ul>
 * @export
 * @example <caption>Start the root instance</caption>
 * var router = oj.Router.rootInstance;
 * // Add three states to the router with id 'home', 'book' and 'tables
 * router.configure({
 *    'home':   { label: 'Home',   value: 'homeContent', isDefault: true },
 *    'book':   { label: 'Book',   value: 'bookContent' },
 *    'tables': { label: 'Tables', value: 'tablesContent' }
 * });
 *
 * var viewModel = {
 *    router: router
 * };
 *
 * oj.Router.sync().then(
 *    function() {
 *       ko.applyBindings(viewModel);
 *    },
 *    function(error) {
 *       oj.Logger.error('Error when starting the router: ' + error.message);
 *    }
 * );
 * @example <caption>Synchronise a newly created child Router and retrieve the bookmarkable state</caption>
 *  oj.Router.sync().then(
 *     function() {
 *        var color = viewModel.router.retrieve();
 *        if (color) {
 *           $('#chapter').css('background', color);
 *        }
 *     },
 *     function(error) {
 *        oj.Logger.error('Error during sync: ' + error.message);
 *     }
 *  );
 *
 */
oj.Router.sync = function() {
   var transition = { router: rootRouter, origin: 'sync' };

   _initialize();

   oj.Logger.info('Entering sync with URL: %s', _location.href);

   if (_deferredPath) {
      transition.path = _deferredPath;
      transition.deferredHandling = true;
      transition.historyUpdate = 'replace';
      _deferredPath = undefined;
      return _queueTransition(transition);
   }

   if (oj.Router._updating) {
      oj.Logger.info('Sync called while updating, waiting for updates to end.');
      // Returms a promise that resolve as soon as the current transition is complete.
      return new Promise(function(resolve) {
         oj.Router._transitionedToState.addOnce(function(result) {
            oj.Logger.info('Sync updates done.');
            resolve(result);
         });
      });
   }

   return _queueTransition(transition);
};

/*------------------------------------------------------------------------------
  URL Apdaters section
  ------------------------------------------------------------------------------*/

/**
 *
 * @class
 * @since 1.1.0
 * @classdesc Url adapter used by the {@link oj.Router} to manage URL in the form of
 * <code class="prettyprint">/book/chapter2</code>.<br>The UrlPathAdapter is the default
 * adapter used by the {@link oj.Router|router} as it makes more human-readable URLs,
 * is user-friendly, and less likely to exceed the maximum charaacter limit in the
 * browser URL.
 * <br>Since this adapter generates path URLs, it's advisable that your application
 * be able to restore the page should the user bookmark or reload the page.  For
 * instance, given the URL <code class="prettyprint">/book/chapter2</code>, your
 * application server ought to serve up content for "chapter2" if the user should
 * bookmark or reload the page.  If that's not possible, then consider using the
 * {@link oj.Router.urlParamAdapter|urlParamAdapter}.
 * <br>There are two available URL adapters,
 * this one and the {@link oj.Router.urlParamAdapter|urlParamAdapter}.<br>To change
 * the URL adapter, use the {@link oj.Router.defaults|urlAdapter} property.
 * @see oj.Router.urlParamAdapter
 * @see oj.Router.defaults
 * @constructor
 * @export
 * @memberof oj.Router
 */
oj.Router.urlPathAdapter = function () {

   /**
    * Variable to store the base path. This is used to retrieve the portion of the path
    * representing the routers state.
    * @ignore
    * @type {!string}
    */
   var _basePath = '';

   /**
    * Initialize the adapter given the baseUrlProp.
    * For the urlPathAdapter, retrieve the potential file name to handle application with
    * index.html in their URL.
    * @ignore
    * @param {!string} baseUrlProp the value of oj.Router.defaults.baseUrl
    */
   this.init = function(baseUrlProp) {
      // Use the browser parser to get the pathname. It works with absolute or relative URL.
      var parser = document.createElement('a');
      parser.href = baseUrlProp;

      var path = parser.pathname;
      path = path.replace(/^([^\/])/,'/$1');  // Should always start with slash (for IE)

      // Normalize the base path. Always ends with a '/'
      if (path.slice(-1) !== '/') {
         path = path + '/';
      }

      _basePath = path;
   };

   /**
    * Construct an array of states where each item is an object made of a router and
    * the new state for it.
    * @ignore
    * @return {!Array.<_StateChange>}
    * @throws Error when parsing of router param fails.
    */
   this.parse = function() {
      var router = rootRouter,
          // To retrieve the portion of the path representing the routers state,
          // remove the base portion of the path.
          path =  _location.pathname.replace(_basePath, ''),
          segments = _getSegments(decodeURIComponent(path)).map(_decodeSlash),
          changes = [],
          stateStr, value,
          state,
          stateChange;

      oj.Logger.info('Parsing: %s', path);

      while (router && (value = segments.shift())) {

         stateChange = new _StateChange(router, value);

         state = stateChange.getState();

         // If this state has parameters, the following segments are parameter values
         state && state._paramOrder.forEach(function (param) {
            // Retrieve the next segment and use it for the parameter value
            stateChange.addParameter(segments.shift());
         });

         changes.push(stateChange);
         router = _getChildRouter(router, value);
      }

      changes = _appendOtherChanges(changes);

      // Retrieve the extra state from request param oj_Router
      stateStr = _location.search.split(_ROUTER_PARAM + '=')[1];
      if (stateStr) {
         stateStr = stateStr.split('&')[0];
         if (stateStr) {
            changes.forEach(_updateBookmarkableData, decodeStateParam(stateStr));
         }
      }

      return changes;
   };

   /**
    * Given an ordered array of states, build the URL representing all
    * the states.
    * Never starts with a '/': "book"  "book/chapter2"
    * @ignore
    * @param {!Array.<_StateChange>} newStates
    * @return {!string} the URL representing the states
    */
   this.buildUrlFromStates = function(newStates) {
      var ns,
          canDefault = false,
          pathname = '',
          extraState = {}; // Compound object of all extra states


      // Build the new URL by walking the array of states backward in order to eliminate
      // the default state from the URL. As soon as a value is not the default, stops the removal.
      while (!!(ns = newStates.pop())) {
         if (ns.value) {
            if (canDefault || (ns.value !== ns.router._defaultStateId)) {
               pathname = pathname ? ns.value + '/' + pathname : ns.value;
               canDefault = true;
            }
         }

         // Build an object made of the extra data of each router
         if (ns.router._extra !== undefined) {
            extraState[ns.router._name] = ns.router._extra;
         }
      }

      return _buildUrl({ pathname: _basePath + pathname }, extraState);
   };
};

/**
 * @class
 * @since 1.1.0
 * @classdesc Url adapter used by the {@link oj.Router} to manage URL in the form of
 * <code class="prettyprint">/index.html?root=book&book=chapter2</code>.  This adapter
 * can be used if the {@link oj.Router.urlPathAdapter|urlPathAdapter} doesn't meet
 * the application's needs.
 * <br>This adapter is well-suited for single-page applications whose entry point
 * is always a single document, i.e., "index.html" which restores its router state
 * from additional parameters.  The router state is encoded as URL parameters and
 * then restored after the page is loaded.  This is ideal for applications which
 * cannot handle multiple entry points, as recommended by {@link oj.Router.urlPathAdapter|urlPathAdapter}.
 * <br>There are two available
 * URL adapters, this one and the {@link oj.Router.urlPathAdapter|urlPathAdapter}.<br>To change
 * the URL adapter, use the {@link oj.Router.defaults|urlAdapter} property.
 * @see oj.Router.urlPathAdapter
 * @see oj.Router.defaults
 * @constructor
 * @export
 * @example <caption>Change the default URL adapter to urlParamAdapter instead of urlPathAdapter:</caption>
 * oj.Router.defaults['urlAdapter'] = new oj.Router.urlParamAdapter();
 */
oj.Router.urlParamAdapter = function () {
   /**
    * Initialize the adapter.
    * @ignore
    */
    this.init = function() {
      // No-op
    };

   /**
    * Construct an array of states where each item is an object made of a router and
    * the new state for it.
    * @ignore
    * @return {!Array.<_StateChange>}
    * @throws Error when parsing of router param fails.
    */
   this.parse = function() {
      var search = _location.search,
          params = parseQueryParam(search),
          router = rootRouter,
          changes = [],
          stateStr, value, segments, state, stateChange;

      oj.Logger.info('Parsing: %s', search);

      while (router) {
         value = params[router._name] || router._defaultStateId;

         // Retrieve all value separated by '/'
         segments = _getSegments(value);
         value = segments.shift();

         stateChange = new _StateChange(router, value);

         if (value) {
            state = stateChange.getState();

            // If this state has parameters, retrieve their values from the segments
            state && state._paramOrder.forEach(function (param) {
               stateChange.addParameter(segments.shift());
            });

            changes.push(stateChange);
         }
         router = _getChildRouter(router, value);
      }

      changes = _appendOtherChanges(changes);

      // Retrieve the extra state from oj_Router param
      stateStr = params[_ROUTER_PARAM];
      if (stateStr) {
         changes.forEach(_updateBookmarkableData, decodeStateParam(stateStr));
      }

      return changes;
   };

   /**
    * Given an ordered array of states, build the URL representing all
    * the states.
    * Never starts with a '/': "index.html", "book/chapter2"
    * @ignore
    * @param {!Array.<_StateChange>} newStates
    * @return {!string} the URL representing the states
    * @throws An error if bookmarkable state is too big.
    */
   this.buildUrlFromStates = function(newStates) {
      var ns,
          canDefault = false,
          search = '',
          paramName, paramValue,
          extraState = {}; // Compound object of all extra states

      // Build the new URL by walking the array of states backward in order to eliminate
      // the default state from the URL. As soon as a value is not the default, stops the removal.
      while (!!(ns = newStates.pop())) {
         if (ns.value) {
            if (canDefault || (ns.value !== ns.router._defaultStateId)) {
               // _name is already encoded
               paramName = '&' + ns.router._name + '=';
               paramValue = ns.value;

               // Because we are traversing the array backward, insert instead of append
               search = paramName + encodeURIComponent(paramValue) + search;
               canDefault = true;
            }
         }

         // Build an object made of the extra data of each router
         if (ns.router._extra !== undefined) {
            extraState[ns.router._name] = ns.router._extra;
         }
      }

      // Replace first parameter separator from '&' to '?'
      if (search) {
         search = '?' + search.substr(1);
      }

      return _buildUrl({ search: search }, extraState);
   };

};

return rootRouter;

}());

/**
 * Utility to compress JSON to store on the URL.
 */
(function() {

oj.LZString = {

   /**
    * Compress into a string that is URI encoded
    * @ignore
    * @param {?string} input
    * @return {string}
    */
   compressToEncodedURIComponent: function(input) {
      if (input === null) {
         return '';
      }
      return _compress(input, 6, function(a) {
         return _keyStrUriSafe.charAt(a);
      });
   },

   /**
    * Decompress from an output of compressToEncodedURIComponent
    * @ignore
    * @param {?string} input
    * @return {?string}
    */
   decompressFromEncodedURIComponent: function(input) {
      if (input === null) {
         return '';
      }
      if (input === '') {
         return null;
      }
      return _decompress(input.length, 32, function(index) {
         return _getBaseValue(_keyStrUriSafe, input.charAt(index));
      });
   }
};

var _fcc = String.fromCharCode;
var _keyStrUriSafe = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-$';
var _baseReverseDic;

/**
 * @param  {string} alphabet
 * @param  {string} character
 * @return {string}
 */
function _getBaseValue(alphabet, character) {
   var i;

   if (!_baseReverseDic) {
      _baseReverseDic = {};
   }
   if (!_baseReverseDic[alphabet]) {
      _baseReverseDic[alphabet] = {};
      for (i = 0; i < alphabet.length; i++) {
         _baseReverseDic[alphabet][alphabet[i]] = i;
      }
   }
   return _baseReverseDic[alphabet][character];
}

/**
 * @param  {?string} uncompressed
 * @param  {number} bitsPerChar
 * @param  {function(number):string} getCharFromInt
 * @return {string}
 */
function _compress(uncompressed, bitsPerChar, getCharFromInt) {
   if (uncompressed === null) {
      return '';
   }
   var i, value,
      context_dictionary = {},
      context_dictionaryToCreate = {},
      context_w = '',
      context_enlargeIn = 2, // Compensate for the first entry which should not count
      context_dictSize = 3,
      context_numBits = 2,
      context_data_string = '',
      context_data_val = 0,
      context_data_position = 0,
      context_c,
      context_wc,
      ii,
      len = uncompressed.length;

   for (ii = 0; ii < len; ii++) {
      context_c = uncompressed[ii];
      if (!Object.prototype.hasOwnProperty.call(context_dictionary, context_c)) {
         context_dictionary[context_c] = context_dictSize++;
         context_dictionaryToCreate[context_c] = true;
      }

      context_wc = context_w + context_c;
      if (Object.prototype.hasOwnProperty.call(context_dictionary, context_wc)) {
         context_w = context_wc;
      } else {
         if (Object.prototype.hasOwnProperty.call(context_dictionaryToCreate, context_w)) {
            if (context_w.charCodeAt(0) < 256) {
               for (i = context_numBits; i--;) {
                  context_data_val = (context_data_val << 1);
                  if (context_data_position == bitsPerChar - 1) {
                     context_data_position = 0;
                     context_data_string += getCharFromInt(context_data_val);
                     context_data_val = 0;
                  } else {
                     context_data_position++;
                  }
               }
               value = context_w.charCodeAt(0);
               for (i = 8; i--; ) {
                  context_data_val = (context_data_val << 1) | (value & 1);
                  if (context_data_position == bitsPerChar - 1) {
                     context_data_position = 0;
                     context_data_string += getCharFromInt(context_data_val);
                     context_data_val = 0;
                  } else {
                     context_data_position++;
                  }
                  value = value >> 1;
               }
            } else {
               value = 1;
               for (i = context_numBits; i--;) {
                  context_data_val = (context_data_val << 1) | value;
                  if (context_data_position == bitsPerChar - 1) {
                     context_data_position = 0;
                     context_data_string += getCharFromInt(context_data_val);
                     context_data_val = 0;
                  } else {
                     context_data_position++;
                  }
                  value = 0;
               }
               value = context_w.charCodeAt(0);
               for (i = 16; i--;) {
                  context_data_val = (context_data_val << 1) | (value & 1);
                  if (context_data_position == bitsPerChar - 1) {
                     context_data_position = 0;
                     context_data_string += getCharFromInt(context_data_val);
                     context_data_val = 0;
                  } else {
                     context_data_position++;
                  }
                  value = value >> 1;
               }
            }
            context_enlargeIn--;
            if (context_enlargeIn == 0) {
               context_enlargeIn = Math.pow(2, context_numBits);
               context_numBits++;
            }
            delete context_dictionaryToCreate[context_w];
         } else {
            value = context_dictionary[context_w];
            for (i = context_numBits; i--;) {
               context_data_val = (context_data_val << 1) | (value & 1);
               if (context_data_position == bitsPerChar - 1) {
                  context_data_position = 0;
                  context_data_string += getCharFromInt(context_data_val);
                  context_data_val = 0;
               } else {
                  context_data_position++;
               }
               value = value >> 1;
            }


         }
         context_enlargeIn--;
         if (context_enlargeIn == 0) {
            context_enlargeIn = Math.pow(2, context_numBits);
            context_numBits++;
         }
         // Add wc to the dictionary.
         context_dictionary[context_wc] = context_dictSize++;
         context_w = String(context_c);
      }
   }

   // Output the code for w.
   if (context_w !== "") {
      if (Object.prototype.hasOwnProperty.call(context_dictionaryToCreate, context_w)) {
         if (context_w.charCodeAt(0) < 256) {
            for (i = context_numBits; i--;) {
               context_data_val = (context_data_val << 1);
               if (context_data_position == bitsPerChar - 1) {
                  context_data_position = 0;
                  context_data_string += getCharFromInt(context_data_val);
                  context_data_val = 0;
               } else {
                  context_data_position++;
               }
            }
            value = context_w.charCodeAt(0);
            for (i = 8; i--;) {
               context_data_val = (context_data_val << 1) | (value & 1);
               if (context_data_position == bitsPerChar - 1) {
                  context_data_position = 0;
                  context_data_string += getCharFromInt(context_data_val);
                  context_data_val = 0;
               } else {
                  context_data_position++;
               }
               value = value >> 1;
            }
         } else {
            value = 1;
            for (i = context_numBits; i--;) {
               context_data_val = (context_data_val << 1) | value;
               if (context_data_position == bitsPerChar - 1) {
                  context_data_position = 0;
                  context_data_string += getCharFromInt(context_data_val);
                  context_data_val = 0;
               } else {
                  context_data_position++;
               }
               value = 0;
            }
            value = context_w.charCodeAt(0);
            for (i = 16; i--;) {
               context_data_val = (context_data_val << 1) | (value & 1);
               if (context_data_position == bitsPerChar - 1) {
                  context_data_position = 0;
                  context_data_string += getCharFromInt(context_data_val);
                  context_data_val = 0;
               } else {
                  context_data_position++;
               }
               value = value >> 1;
            }
         }
         context_enlargeIn--;
         if (context_enlargeIn == 0) {
            context_enlargeIn = Math.pow(2, context_numBits);
            context_numBits++;
         }
         delete context_dictionaryToCreate[context_w];
      } else {
         value = context_dictionary[context_w];
         for (i = context_numBits; i--;) {
            context_data_val = (context_data_val << 1) | (value & 1);
            if (context_data_position == bitsPerChar - 1) {
               context_data_position = 0;
               context_data_string += getCharFromInt(context_data_val);
               context_data_val = 0;
            } else {
               context_data_position++;
            }
            value = value >> 1;
         }


      }
      context_enlargeIn--;
      if (context_enlargeIn == 0) {
         context_enlargeIn = Math.pow(2, context_numBits);
         context_numBits++;
      }
   }

   // Mark the end of the stream
   value = 2;
   for (i = context_numBits; i--;) {
      context_data_val = (context_data_val << 1) | (value & 1);
      if (context_data_position == bitsPerChar - 1) {
         context_data_position = 0;
         context_data_string += getCharFromInt(context_data_val);
         context_data_val = 0;
      } else {
         context_data_position++;
      }
      value = value >> 1;
   }

   // Flush the last char
   while (true) {
      context_data_val = (context_data_val << 1);
      if (context_data_position == bitsPerChar - 1) {
         context_data_string += getCharFromInt(context_data_val);
         break;
      } else context_data_position++;
   }
   return context_data_string;
}

/**
 * @param  {number} length
 * @param  {number} resetValue
 * @param  {function(number):string} getNextValue
 * @return {?string}
 */
function _decompress(length, resetValue, getNextValue) {
   var dictionary = [],
      next,
      enlargeIn = 4,
      dictSize = 4,
      numBits = 3,
      entry = '',
      result = '',
      i,
      w,
      bits, resb, maxpower, power,
      c,
      data = {
         val: getNextValue(0),
         position: resetValue,
         index: 1
      };

   for (i = 0; i < 3; i += 1) {
      dictionary[i] = i;
   }

   bits = 0;
   maxpower = Math.pow(2, 2);
   power = 1;
   while (power != maxpower) {
      resb = data.val & data.position;
      data.position >>= 1;
      if (data.position == 0) {
         data.position = resetValue;
         data.val = getNextValue(data.index++);
      }
      bits |= (resb > 0 ? 1 : 0) * power;
      power <<= 1;
   }

   switch (next = bits) {
      case 0:
         bits = 0;
         maxpower = Math.pow(2, 8);
         power = 1;
         while (power != maxpower) {
            resb = data.val & data.position;
            data.position >>= 1;
            if (data.position == 0) {
               data.position = resetValue;
               data.val = getNextValue(data.index++);
            }
            bits |= (resb > 0 ? 1 : 0) * power;
            power <<= 1;
         }
         c = _fcc(bits);
         break;
      case 1:
         bits = 0;
         maxpower = Math.pow(2, 16);
         power = 1;
         while (power != maxpower) {
            resb = data.val & data.position;
            data.position >>= 1;
            if (data.position == 0) {
               data.position = resetValue;
               data.val = getNextValue(data.index++);
            }
            bits |= (resb > 0 ? 1 : 0) * power;
            power <<= 1;
         }
         c = _fcc(bits);
         break;
      case 2:
         return "";
   }
   dictionary[3] = c;
   w = result = c;
   while (true) {
      if (data.index > length) {
         return "";
      }

      bits = 0;
      maxpower = Math.pow(2, numBits);
      power = 1;
      while (power != maxpower) {
         resb = data.val & data.position;
         data.position >>= 1;
         if (data.position == 0) {
            data.position = resetValue;
            data.val = getNextValue(data.index++);
         }
         bits |= (resb > 0 ? 1 : 0) * power;
         power <<= 1;
      }

      switch (c = bits) {
         case 0:
            bits = 0;
            maxpower = Math.pow(2, 8);
            power = 1;
            while (power != maxpower) {
               resb = data.val & data.position;
               data.position >>= 1;
               if (data.position == 0) {
                  data.position = resetValue;
                  data.val = getNextValue(data.index++);
               }
               bits |= (resb > 0 ? 1 : 0) * power;
               power <<= 1;
            }

            dictionary[dictSize++] = _fcc(bits);
            c = dictSize - 1;
            enlargeIn--;
            break;
         case 1:
            bits = 0;
            maxpower = Math.pow(2, 16);
            power = 1;
            while (power != maxpower) {
               resb = data.val & data.position;
               data.position >>= 1;
               if (data.position == 0) {
                  data.position = resetValue;
                  data.val = getNextValue(data.index++);
               }
               bits |= (resb > 0 ? 1 : 0) * power;
               power <<= 1;
            }
            dictionary[dictSize++] = _fcc(bits);
            c = dictSize - 1;
            enlargeIn--;
            break;
         case 2:
            return result;
      }

      if (enlargeIn == 0) {
         enlargeIn = Math.pow(2, numBits);
         numBits++;
      }

      if (dictionary[c]) {
         entry = dictionary[c];
      } else {
         if (c === dictSize) {
            entry = w + w[0];
         } else {
            return null;
         }
      }
      result += entry;

      // Add w+entry[0] to the dictionary.
      dictionary[dictSize++] = w + entry[0];
      enlargeIn--;

      w = entry;

      if (enlargeIn == 0) {
         enlargeIn = Math.pow(2, numBits);
         numBits++;
      }

   }
}

})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/
/*global oj, Promise */

/**
 * The RouterState module.
 */

// Wrap in a IIFE to prevents the possiblity of collision in a non-AMD scenario.
(function() {
"use strict";

var stateParamExp = /^{(\w+)}$/;
  /**
   * @typedef {object} oj.RouterState.ConfigOptions
   * @property {string=} label the string to be displayed in the command component.
   * See the {@link oj.RouterState#label} property.
   * @property {*=} value the object associated with this state.
   * See the {@link oj.RouterState#value} property.
   * @property {boolean=} isDefault true if this state is the default.
   * See the {@link oj.Router#defaultStateId|defaultStateId} property.
   * @property {(function(): boolean|function(): Promise.<boolean>)=} canEnter A callback that
   * either returns a boolean or the Promise of a boolean. If the boolean is true the transition
   * will continue.
   * The default value is a method that always returns true.
   * See the {@link oj.RouterState#canEnter} property.
   * @property {(function():void|function(): Promise.<void>)=} enter A callback or
   * the promise of a callback which execute when entering this state.
   * See the {@link oj.RouterState#enter} property.
   * @property {(function(): boolean|function(): Promise.<boolean>)=} canExit  A callback that
   * either returns a boolean or the Promise of a boolean. If the boolean is true the transition
   * will continue.
   * The default value is a method that always returns true.
   * See the {@link oj.RouterState#canExit} property.
   * @property {(function():void|function(): Promise.<void>)=} exit A callback or
   * the promise of a callback which execute when exiting this state.
   * See the {@link oj.RouterState#canExit} property.
   */

   /**
    * @class
    * @since 1.1.0
    * @classdesc
    * Object representing a state of the router.
    * @desc
    * It is the type of the {@link oj.Router#currentState|currentState} property and type of
    * the value returned by {@link oj.Router#getState|getState(String)}.
    * @param {!string} id the state id.
    * See the {@link oj.RouterState#id} property.
    * @param {!Object=} options an object defining the state.
    * @param {string=} options.label the string to be displayed in the command component.
    * See the {@link oj.RouterState#label} property.
    * @param {*=} options.value the object associated with this state.
    * See the {@link oj.RouterState#value} property.
    * @param {boolean=} options.isDefault true if this state is the default.
    * See the {@link oj.Router#defaultStateId|defaultStateId} property.
    * @param {(function(): boolean|function(): Promise.<boolean>)=} options.canEnter A callback that
    * either returns a boolean or the Promise of a boolean. If the boolean is true the transition
    * will continue.
    * The default value is a method that always returns true.
    * See the {@link oj.RouterState#canEnter} property.
    * @param {(function():void|function(): Promise.<void>)=} options.enter A callback or
    * the promise of a callback which execute when entering this state.
    * See the {@link oj.RouterState#enter} property.
    * @param {(function(): boolean|function(): Promise.<boolean>)=} options.canExit  A callback that
    * either returns a boolean or the Promise of a boolean. If the boolean is true the transition
    * will continue.
    * The default value is a method that always returns true.
    * See the {@link oj.RouterState#canExit} property.
    * @param {(function():void|function(): Promise.<void>)=} options.exit A callback or
    * the promise of a callback which execute when exiting this state.
    * See the {@link oj.RouterState#canExit} property.
    * @param {oj.Router=} router The router this state belongs to. If undefined, the method
    * {@link oj.RouterState#go|go} and {@link oj.RouterState#isCurrent|isCurrent} will not work.
    * @constructor
    * @export
    * @ojsignature {
    *              target: 'Type',
    *              for: 'options',
    *              value: 'oj.RouterState.ConfigOptions',
    *              jsdocOverride: true }
    */
   oj.RouterState = function (id, options, router) {
      options = options || {};
      oj.Assert.assertString(id);
      id = id.trim();

      // The id is in the form /aaa/{p1}/{p2}
      var path = id.split('/');
      // We cannot have duplicate because the format of the object parameter
      // doesn't allow it.
      this._id = path.shift();

      /**
       * Router parameters allow passing of name/value pairs easily from one state
       * to another.
       *
       * Configure the router with the parameter names embedded in
       * the key of the router state, enclosed by curly-braces.  The parameter
       * keys will be used to retrieve their values.
       * @name oj.RouterState#parameters
       * @example
       * <caption>Configure router with a parameterized state</caption>
       * oj.Router.rootInstance.configure({
       *    'list': { label: 'List' },
       *    'detail/{empId}/{edit}'
       * });
       * @description Parameter values are passed in the URL, and are position-
       * dependent.
       * @example
       * <caption>Parameter values are passed in the URL, and are position-
       * dependent.</caption>
       * /app/detail/e100/true
       * @description The newly-activated state is passed the 'parameters' Object
       * alongside the {@link oj.Router#instance} property to the View Model of
       * the state (if {@link oj.Router#moduleConfig} is used).
       * To reference the values, use the key name by which the router states
       * were configured.
       * @example
       * <caption>Retrieve parameter values from router</caption>
       * function ViewModel(config) {
       *    var router = config['ojRouter']['instance'];
       *    var params = config['ojRouter']['parameters'];
       *    var empId = params['empId'];
       *    var edit = params['edit'];
       * }
       * @type {Object}
       * @ojstatus preview
       * @since 4.2.0
       */
      this._parameters = {};

      /**
       * Internal array to track the order of configured param names.
       * @type {Array}
       * @private
       */
      this._paramOrder = new Array(path.length);

      path.forEach(function(token, i) {
         /*
          * Match pattern "{token}"
          */
         var match = token.match(stateParamExp);
         if (match) {
            token = match[1];
            this._parameters[token] = null;
            this._paramOrder[i] = token;
         }
      }, this);

      /**
       * A callback that either returns a boolean or the Promise of a boolean.
       * When defined, this callback is executed before entering this state. If the boolean
       * is true the transition will continue, otherwise the state is not entered and the
       * current state of the router does not change.
       * <br>This allows the application to veto the transition into the
       * state if there are known conditions that must be met before the state
       * should be entered into.
       * <br>If the operation (to check if the state can be entered) is synchronous,
       * return a boolean directly; if asynchronous, a Promise of a boolean should
       * be returned.  The state transition will not occur until either the boolean
       * returned is true, or the Promise is resolved and returns true.
       * <br>The default value is a method that always returns true.
       * @name oj.RouterState#canEnter
       * @type {(function():boolean|function():Promise.<boolean>)}
       */
      this._canEnter = options['canEnter'];
      if (this._canEnter) {
         oj.Assert.assertFunctionOrNull(this._canEnter);
      }

      /**
       * A callback or the promise of a callback which executes when entering this
       * state.
       * This callback executes after the router stateId changes.
       * @name oj.RouterState#enter
       * @type {(function():void|function():Promise.<void>)}
       */
      this._enter = options['enter'];
      if (this._enter) {
         oj.Assert.assertFunctionOrNull(this._enter);
      }

      /**
       * A callback that either returns a boolean or the Promise of a boolean.
       * When defined, this callback is executed before exiting this state. If the boolean
       * is true the transition will continue, otherwise the state is not exited and the
       * current state of the router does not change.
       * <br>This is typically used by states which need to verify user input is
       * correct before moving on.  Synchronous checks can return a boolean directly,
       * and asynchronous checks can return a Promise of a boolean.
       * The state transition will not occur until either the boolean returned is
       * true, or the Promise returned is resolved and returns true.
       * <br>The default value is a method that always returns true.
       * @name oj.RouterState#canExit
       * @type {(function():boolean|function():Promise.<boolean>)}
       */
      this._canExit = options['canExit'];
      if (this._canExit) {
         oj.Assert.assertFunctionOrNull(this._canExit);
      }

      /**
       * A callback or the promise of a callback which execute when exiting this
       * state.
       * This callback executes before the router stateId changes.
       * @name oj.RouterState#exit
       * @type {(function():void|function():Promise.<void>)}
       */
      this._exit = options['exit'];
      if (this._exit ) {
         oj.Assert.assertFunctionOrNull(this._exit );
      }
      /**
       * The value associated with this state. When this state is the current state of the
       * router, it is the value returned by the observable {@link oj.Router#currentValue}.
       * @name oj.RouterState#value
       * @type {*}
       */
      this._value = options['value'];

      /**
       * The string to be used for the navigation component that will transition to this state.
       * This is also used to build the title of the page when the {@link oj.RouterState#title}
       * property is not defined. The title will be composed of the labels of all current
       * states in the router hierarchy like "My Page | label lvl1 | label lvl2".
       * @name oj.RouterState#label
       * @type {string|undefined}
       * @example <caption>Use the label property for the text of anchor tags in a list:</caption>
       * &lt;ul id="foreachMenu" data-bind="foreach: router.states">
       *   &lt;li>
       *     &lt;a data-bind="text: label, css: {'active': isCurrent()},
       *                   attr: {id: id}, click: go"> &lt;/a>
       *   &lt;/li>
       * &lt;/ul>
       */
      this._label = options['label'];

      /**
       * The string to be used for the page title. This can either be a string or a function
       * returning a string. When more than one level of child router is defined, the title of
       * the current state of the router nested the deepest has precedence. If the leaf router
       * current state does not have a title property defined, the title of the current state of
       * the parent router is used. If no title property is defined in the router hierarchy, a
       * title is built using the {@link oj.RouterState#label} property.
       * @name oj.RouterState#title
       * @type {string|function():string|undefined}
       */
      this._title = options['title'];

      /**
       * @private
       * @type {oj.Router|undefined}
       *
       */
      this._router = router;

      /**
       * This property is used when the router states are ojModule names.
       * When the router transition to this state, it will attempt to execute the callbacks canExit, canEnter,
       * enter and exit on the viewModel object first. If the callback is not defined on the viewModel, the
       * router will attempt to execute the callback on the {@link oj.RouterState|RouterState}.
       * @private
       * @type {Object|undefined}
       */
      this.viewModel = undefined;

      Object.defineProperties(this, {
         'id': { value:
            /**
             * The id of this state.<br>
             * It uniquely identify a state object in a router. The id property can be used as the
             * attribute id of a navigation component like link or button.
             * @name oj.RouterState#id
             * @readonly
             * @type {!string}
             * @example <caption>Use the state id property for the attribute id of anchor tags in a list:</caption>
             * &lt;ul id="foreachMenu" data-bind="foreach: router.states">
             *   &lt;li>
             *     &lt;a data-bind="text: label, css: {'active': isCurrent()},
             *                   attr: {id: id}, click: go"> &lt;/a>
             *   &lt;/li>
             * &lt;/ul>
             */
            (this._id), enumerable: true
         },
         'value': {
            get: function () { return this._value; },
            set: function(newValue) { this._value = newValue; },
            enumerable: true
         },
         'label': {
            get: function () { return this._label; },
            set: function(newValue) { this._label = newValue; },
            enumerable: true
         },
         'title': {
            get: function () { return this._title; },
            set: function(newValue) { this._title = newValue; },
            enumerable: true
         },
         'canEnter': {
            get: function () { return this._canEnter; },
            set: function(newValue) { this._canEnter = newValue; },
            enumerable: true
         },
         'enter': {
            get: function () { return this._enter; },
            set: function(newValue) { this._enter = newValue; },
            enumerable: true
         },
         'canExit': {
            get: function () { return this._canExit; },
            set: function(newValue) { this._canExit = newValue; },
            enumerable: true
         },
         'exit': {
            get: function () { return this._exit; },
            set: function(newValue) { this._exit = newValue; },
            enumerable: true
         },
         'parameters': {
            get: function () { return this._parameters; },
            enumerable: true
         }
      });
   };

   /**
    * Transition the router to this state.
    * This is a convenience method used as the event handler for a Knockout click binding on
    * a button or <code class="prettyprint">a</code> tag.<br>
    * A {@link oj.Router.transitionedToState|transitionedToState} signal is dispatched when the
    * state transition has completed.
    * @return {!Promise.<{hasChanged: boolean}>} A Promise that resolves when the
    * router is done with the state transition.<br>
    * When the promise is fullfilled, the parameter value is an object with the property
    * <code class="prettyprint">hasChanged</code>.<br>
    * The value of <code class="prettyprint">hasChanged</code> is:
    * <ul>
    *   <li>true: If the router state changed.</li>
    * </ul>
    * When the Promise is rejected, the parameter value is:
    * <ul>
    *   <li>An Error object stipulating the reason for the rejection when an error
    * occurred during the resolution.</li>
    * </ul>
    * @export
    * @example <caption>Use the go function as the handler for a click binding:</caption>
    * &lt;ul id="foreachMenu" data-bind="foreach: router.states">
    *   &lt;li>
    *     &lt;a data-bind="text: label, css: {'active': isCurrent()},
    *                   attr: {id: id}, click: go"> &lt;/a>
    *   &lt;/li>
    * &lt;/ul>
    */
   oj.RouterState.prototype.go = function() {
      if (!this._router) {
         oj.Router._transitionedToState.dispatch({ 'hasChanged': false });
         return Promise.reject(new Error('Router is not defined for this RouterState object.'));
      }
      return this._router.go(this._id);
   };

   /**
    * Determine if the router current state is this state.
    * This method is typically used by elements in the markup to show the appropriate selection value.
    * @return {boolean} true if this state is the current router state.
    * @throws An error if an owning router was not specified when the state was
    * created.
    * @export
    * @example <caption>Use the is function to change the css of the state links:</caption>
    * &lt;ul id="foreachMenu" data-bind="foreach: router.states">
    *   &lt;li>
    *     &lt;a data-bind="text: label, css: {'active': isCurrent()},
    *                   attr: {id: id}, click: go"> &lt;/a>
    *   &lt;/li>
    * &lt;/ul>
    */
   oj.RouterState.prototype.isCurrent = function() {
      if (!this._router) {
         throw new Error('Router is not defined for this RouterState object.');
      }
      return this._router._stateId() === this._id;
   };

}());

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/


/**
 * @class oj.Cube 
 * @classdesc Functions implemented by oj.DataColumnCube and oj.DataValueAttributeCube
 * @constructor
 * @export
 * @hideconstructor
 * @see oj.DataColumnCube
 * @see oj.DataValueAttributeCube
 * @since 1.1.0
 * @ojtsignore
 */
 oj.Cube = function(rowset, layout)
{
    this.Init();
    
    // Wrap the incoming data in the appropriate interface
    this._rows = rowset;
    this._layout = /** @type {Array.<{axis,levels}>} */ (layout);

    this.BuildCube();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.Cube, oj.Object, "oj.Cube");

/**
 * Initializes instance with the set options
 * @private
 */
oj.Cube.prototype.Init = function() 
{
  oj.Cube.superclass.Init.call(this);
};

/**
 * Get the oj.CubeAxis objects in this cube
 * @return {Array.<oj.CubeAxis>} an array of oj.CubeAxis objects
 * @memberof oj.Cube
 * @export
 */
oj.Cube.prototype.getAxes = function() {
    // Generate axes list as union of subcube axes for particular page + page axes
    var cube = this._getPinnedCube();
    var axes = [];
    Array.prototype.push.apply(axes, cube ? cube.getAxes() : this._axes);
    // Add on the page axes
    for (var i = 2; i < this._axes.length; i++) {
        axes.push(this._axes[i]);
    }
    return axes;
};

/**
 * Get oj.CubeDataValues from this cube.  These represent the values of the data in the "body" of the cube
 *
 * @param {Array.<Object|number>} indices an axis-ordered array of Objects or numbers.  If Objects, each should contain a 'start' property
 * (the zero based start index for the axis) and a 'count' representing the number of data values beginning at 'start' to return on this axis. 
 * This format allows the retrieval of a block of data.  Passing an array of numbers alone is equivalent to passing {start:<index>, count:1} and getting a single
 * oj.CubeDataValue
 * @return {Array.<oj.CubeDataValue>|oj.CubeDataValue} either an array of arrays of oj.CubeDataValue, depending on the number of values requested in indices, or a single oj.CubeDataValue
 *                        The first subscript represents the 0th axis' values, and so on.
 * @ojsignature  {target: "Type",
 *                value: "Array<oj.Cube.Indices|number>",
 *                for: "indices"}
 * @memberof oj.Cube
 * @export
 */
oj.Cube.prototype.getValues = function(indices) {
    // Get pinned cube
    var cube = this._getPinnedCube();
    
    // Normalize the argument to be objects with start/count properties
    var ind = cube._normalizeIndices(indices);

    var origRet = cube._walkIndex(ind, 0, [], []);
    var val = origRet;
    // Unpack single values tucked in arrays
    while (Array.isArray(val) && val.length === 1) {
        val = val[0];
        if (!Array.isArray(val)) {
            // Found an object in the middle--return it
            return val;
        }
    }
    return origRet;
};

/**
 * Object in getValues indices parameter
 * @typedef oj.Cube.Indices
 * @property {number} start
 * @property {number} count
 */

/**
 * Set a pinned index for all axes above axis 1 ("pages")
 * @param {Array.<Object>|Object} pin an array of objects (or a single object) containing an integer 'axis' attribute and its corresponding 'index' value (to which to pin the cube)
 * @return {undefined}
 * @ojsignature  {target: "Type",
 *                value: "Array<oj.Cube.Pin>|oj.Cube.Pin",
 *                for: "pin"}
 * @memberof oj.Cube
 * @export
 */
oj.Cube.prototype.setPage = function(pin) {
    // Make sure it's an array
    if (pin instanceof Array) {
        this._pin = pin;
    }
    else {
        this._pin = [pin];
    }
};
    
/**
 * Object in setPage pin parameter
 * @typedef oj.Cube.Pin
 * @property {number} axis
 * @property {number} index
 */

/**
 * 
 * @param {number} axisFrom the axis from which to move a level
 * @param {number} levelFrom the level within axisFrom to move to axisTo/levelTo (zero is slowest/outermost)
 * @param {number} axisTo the axis to which to move levelFrom
 * @param {number} levelTo the level within axisTo to move the levelFrom level (zero is slowest/outermost)
 * @param {oj.Cube.PivotType} type the type of pivot to perform
 * @returns {boolean} true if successful
 * @memberof oj.Cube
 * @export
 */
oj.Cube.prototype.pivot = function(axisFrom, levelFrom, axisTo, levelTo, type) {
    var layout = this._layout;
    var axisFromObj = this._findAxisInLayout(axisFrom);
    // No from axis found: can't pivot
    if (!axisFromObj) {
        return false;
    }
    var axisToObj = null;
    if (axisTo < layout.length) {
        axisToObj = this._findAxisInLayout(axisTo);
    }
    else {
        // Add a new axis
        axisToObj = {'axis':axisTo, 'levels':[]};
        layout.push(axisToObj);
    }
    var levelsFrom = axisFromObj['levels'];
    var levelsTo = axisToObj['levels'];
    var levelFromObj = levelFrom < levelsFrom.length ? levelsFrom[levelFrom] : null;
    // No from level here, can't pivot
    if (!levelFromObj) {
        return false;
    }
    
    // Find where to move 
    if (levelTo >= levelsTo.length) {
        // beyond end--just add it.  after, before, swap doesn't matter
        levelsTo.push(levelFromObj);
        // Remove it from the old location
        levelsFrom.splice(levelFrom, 1);
    }
    else {
        if (type === oj.Cube.PivotType['SWAP']) {
            // Swap the elements
            levelsFrom[levelFrom] = levelsTo[levelTo];
            levelsTo[levelTo] = levelFromObj;
        }
        else {
            if (type === oj.Cube.PivotType['AFTER']) {
                // Splice operates as a "before" loc
                levelTo++;
            }
            // Insert the from obj into the to
            levelsTo.splice(levelTo, 0, levelFromObj);
            if (levelsTo === levelsFrom && levelTo < levelFrom) {
                // Relocate the from location, because the to splice shifted it (the from was in the same axis as the to, and was after the to)
                levelFrom++;
            }
            
            // On a move, remove the from object from the old axis
            levelsFrom.splice(levelFrom, 1);
        }
    }
    
    // Rebuild using the altered layout
    this.BuildCube();
    return true;
};

/**
 * Return the entry in the layout arg that represents axis #
 * @private
 */
oj.Cube.prototype._findAxisInLayout = function(axis) {
    for (var i = 0; i < this._layout.length; i++) {
        if (this._layout[i]['axis'] === axis) {
            return this._layout[i];
        }
    }
    return null;
};
    
/**
 * Return the current layout used to build the cube
 * 
 * @returns {Array.<Object>} current layout
 * @see oj.Cube
 * @memberof oj.Cube
 * @export
 */
oj.Cube.prototype.getLayout = function() {
    return this._layout;
};

/**
 * Valid pivot types
 * @enum {string}
 * @memberof oj.Cube
 * @export
 */
oj.Cube.PivotType = {
    /**
     * Move the from location before the to location
     */
    'BEFORE' : "before",
    /**
     * Move the from location after the to location
     */
    'AFTER' : "after",
    /**
     * Exchange the from location with the to location
     */
    'SWAP' : "swap"    
};

/**
 * @private
 */
oj.Cube.prototype._walkIndex = function(indices, depth, location, returnValue) {
   if (indices.length === 0) {
       // Get the current value at location
       var loc = location.slice(0);
       //loc.reverse();
       return this._getValue(loc);
   }
   else {
       var remaining = indices.slice(1);
       var start = indices[0].start;
       var count = indices[0].count;
       for (location[depth] = start; location[depth] < start+count; location[depth]++) {
           returnValue.push(this._walkIndex(remaining, depth+1, location, []));
       }
   }
   return returnValue;
};

/**
 * Make sure all start/count data is filled in and within range
 * @private
 */
oj.Cube.prototype._normalizeIndices = function(indices) {
    var ind = [];
    if (!indices) {
        return ind;
    }
    var numAxes = Math.min(indices.length, this._axes.length);
    for (var a = 0; a < numAxes; a++) {        
        var index = indices[a];
        if (index instanceof Object && (index.hasOwnProperty('start') || index.hasOwnProperty('count'))) {
            if (index.hasOwnProperty('start')) {
                if (index.hasOwnProperty('count')) {
                    ind.push(this._generateIndex(index.start, index.count, a));
                }
                else {
                    ind.push(this._generateIndex(index.start, 1, a));
                }
            }
            else {
                // Must have count
                ind.push(this._generateIndex(0, index.count, a));
            }
        }
        else {
            // Convert the number to an object
            ind.push(this._generateIndex(index, 1, a));
        }
    }
    return ind;
};
    
/**
 * @private
 */
oj.Cube.prototype._generateIndex = function(start, count, axis) {
    // Get true count
    var trueCount = this.getAxes()[axis].getExtent();
    
    if (start >= trueCount || start < 0) {
        start = 0;
    }
    count = Math.min(count,trueCount-start);
            
    return {start: start, index: start, count: count};
};

/**
 * @private
 */
oj.Cube.prototype._getValue = function(indices) {
    // indices is just an axis-ordered array of locations for which to get the value
    // Should only be one--only one data value here
    var key = this._createCubeKeys(indices);
    if (key) {
        var hash = key.GetHashCodes();
        if (hash.length > 0) {
            var obj = this._data[hash[0].key];
            if (obj) {
                return new oj.CubeDataValue(obj.value, indices, obj.aggType, obj.rows, obj.square);
            }
        }
    }
    return new oj.CubeDataValue(null, indices, undefined, []);
};
    
/**
 * @protected
 */
oj.Cube.prototype.GenerateAxes = function() {    
    var pageLayout = this._getPageLayout();
    // If we have pages--only generate those for this cube, and set up the default 0th pin    
    this._pin = [];
    for (var i = 0; i < pageLayout.length; i++) {
        this._getAxis(pageLayout[i]['axis'], pageLayout[i]['levels']);
        this._pin.push({'axis':pageLayout[i]['axis'], 'index':0});
    }
    if (pageLayout.length === 0) {
        // Do remaining axes if any (row and column)
        var nonPageLayout = this._getNonPageLayout();
        for (var i = 0; i < nonPageLayout.length; i++) {
            this._getAxis(nonPageLayout[i]['axis'], nonPageLayout[i]['levels']);
        }
    }
};

/**
 * @private
 */
oj.Cube.prototype._getPageLayout = function() {
    var pageOnlyLayout = [];
    for (var i = 0; i < this._layout.length; i++) {
        var axis = this._layout[i]['axis'];
        if (axis > 1) {
            pageOnlyLayout.push(this._layout[i]);
        }
    }
    return pageOnlyLayout;
};

/**
 * @private
 */
oj.Cube.prototype._getNonPageLayout = function() {
    var nonPageLayout = [];
    for (var i = 0; i < this._layout.length; i++) {
        var axis = this._layout[i]['axis'];
        if (axis < 2) {
            nonPageLayout.push(this._layout[i]);
        }
    }
    return nonPageLayout;
};

/**
 * @protected
 */
oj.Cube.prototype.BuildCube = function() {
    this._axes = [];
    
    // Associative array for data
    this._data = [];
    
    // Associative array for cubes
    this._cubes = [];
    
    this.GenerateAxes();
    
    // Can start a cube with no data--add on from "parent"
    if (this._rows === null) {
        return;
    }
        
    // Walk each line of the rowset, and distribute its attributes to the axes and the data store
    for (var row = 0; row < this._rows.length; row++) {
        // Generate a "page key"
        var pageKey = new oj.CubeKeys();
        for (var axis = 2; axis < this._axes.length; axis++) {
            pageKey = this._axes[axis].ProcessRow(this._rows[row], pageKey);
        }
        // Grab the data value from the page if it's there
        var pageHashObj = pageKey.GetHashCodes();
        // See if we have a cubes for these pageKeys
        for (var ph = 0; ph < pageHashObj.length; ph++) {
            var pageHash = pageHashObj[ph].key;
            var cube = this._cubes[pageHash];
            if (!cube) {
                // Generate a new one
                cube = this._cubes[pageHash] = this.GenerateCube(this._getNonPageLayout());
            }
            // Send the row through the lower axes
            var keys = new oj.CubeKeys();
            var maxAxes = cube._axes.length;
            for (var axis = 0; axis < maxAxes; axis++) {
                keys = cube._axes[axis].ProcessRow(this._rows[row], keys);
            }
            // Convert the retrieved keys->data into data storage
            var hash = keys.GetHashCodes();
            // Must account for data value coming from page, possibly...pass in the hash that has the data
            var dataHash = hash;
            if (pageHashObj[ph].dataValue !== undefined) {
                dataHash = [];
                // Must be in an array
                for (var h = 0; h < hash.length; h++) {
                    dataHash.push(pageHashObj[ph]);
                }
            }
            cube._storeData(hash, dataHash, this._rows[row]);
        }
    }
};

/**
 * @private
 */
oj.Cube.prototype._storeData = function(hash, dataHash, row) {
    for (var i = 0; i < hash.length; i++) {
        this._data[hash[i].key] = this._aggregate(dataHash[i], this._data[hash[i].key], row);
    }
};


/**
 * @private
 */
oj.Cube.prototype._getPinnedCube = function() {
    return this._cubes[this._getHashFromPin(this._pin)];
};
        
/**
 * @private
 */
oj.Cube.prototype._getHashFromPin = function(pin) {
    var keys = new oj.CubeKeys();
    if (pin && pin.length > 0) {
        // Sort by axis attribute
        pin.sort(function(a,b) {
            return a['axis'] - b['axis'];
        });
        // Translate the axis/index pin to hash key
        var axes = this._axes;
        for (var i = 0; i < pin.length; i++) {
            keys = axes[pin[i]['axis']].GetCubeKeys(pin[i]['index'], keys);
        }
    }
    return keys.GetHashCodes()[0].key;
};
    
/**
 * @private
 */
oj.Cube._isValid = function(value) {
    if (value) {
        return value.value !== undefined && value.value !== null;
    }
    return false;
};

/**
 * @private
 */
oj.Cube.prototype._createAggValue = function(value, aggType, rows, row, props) {
    rows.push(row);
    var retObj = {};
    for (var p in props) {
        if (props.hasOwnProperty(p)) {
            retObj[p] = props[p];
        }
    }
    retObj.value = value;
    retObj.aggType = aggType;
    retObj.rows = rows;
    return retObj;
};

/**
 * Don't treat strings as numbers, ever
 * @private
 */
oj.Cube._isNumber = function(value) {
    if (oj.StringUtils.isString(value.value)) {
        return false;
    }
    // String versions of numbers return false for isNaN--we want to treat them as non numbers always
    return !isNaN(value.value);
};

/**
 * @private
 */
oj.Cube.prototype._aggregate = function(hash, currValue, row) {
    var aggObj = this.GetAggType(hash.dataValue);
    var aggType = /** @type {Object} */(aggObj.aggregation);
    var validCurr = oj.Cube._isValid(currValue);
    var validHash = oj.Cube._isValid(hash);
    var isNumCurr = validCurr && oj.Cube._isNumber(currValue);
    var isNumHash = validHash && oj.Cube._isNumber(hash);
    switch (aggType) {
        case oj.CubeAggType['SUM']:            
            if (validCurr && validHash) {
                if (isNumCurr && isNumHash) {
                    return this._createAggValue(currValue.value + hash.value, aggType, currValue.rows, row, {});
                }
                return this._createAggValue(NaN, aggType, currValue.rows, row, {});
            }
            if (validHash && !validCurr) {
                if (isNumHash) {
                    return this._createAggValue(hash.value, aggType, [], row, {});
                }
                return this._createAggValue(NaN, aggType, [], row, {});
            }
            return currValue;
        case oj.CubeAggType['AVERAGE']:
            if (validCurr && validHash) {
                if (isNumCurr && isNumHash) {
                    return this._createAggValue((currValue.sum + hash.value) / (currValue.rows.length+1), aggType, currValue.rows, row, {sum:(currValue.sum + hash.value)});
                }
                return this._createAggValue(NaN, aggType, currValue.rows, row, {sum:currValue.sum});
            }
            if (validHash && !validCurr) {
                if (isNumHash) {
                    return this._createAggValue(hash.value, aggType, [], row, {sum:hash.value});
                }
                return this._createAggValue(NaN, aggType, [], row, {sum:NaN});
            }
            return currValue;
        case oj.CubeAggType['VARIANCE']:
        case oj.CubeAggType['STDDEV']:
            if (validCurr && validHash) {
                if (isNumCurr && isNumHash) {
                    var newCount = currValue.rows.length+1;
                    var avg = currValue.value + (hash.value - currValue.value) / newCount;
                    return this._createAggValue(avg, aggType, currValue.rows, row, {square:(currValue.square + (hash.value - currValue.value) * (hash.value - avg))});
                }
                return this._createAggValue(NaN, aggType, currValue.rows, row, {square:NaN});
            }
            if (validHash && !validCurr) {
                if (isNumHash) {
                    return this._createAggValue(hash.value, aggType, [], row, {square:0});
                }
                return this._createAggValue(NaN, aggType, [], row, {square:NaN});
            }
            return currValue;
        case oj.CubeAggType['NONE']:
            return this._createAggValue(null, aggType, validCurr ? currValue.rows : [], row, {});
        case oj.CubeAggType['FIRST']:
            if (validCurr) {
                return this._createAggValue(currValue.value, aggType, currValue.rows, row, {});
            }
            if (validHash) {
                return this._createAggValue(hash.value, aggType, [], row, {});
            }
            return currValue;
        case oj.CubeAggType['MIN']:
            if (validCurr && validHash) {
                if (isNumCurr && isNumHash) {
                    return this._createAggValue(Math.min(currValue.value, hash.value), aggType, currValue.rows, row, {});
                }
                return this._createAggValue(NaN, aggType, currValue.rows, row, {});
            }
            if (validHash && !validCurr) {       
                if (isNumHash) {
                    return this._createAggValue(hash.value, aggType, [], row, {});
                }
                return this._createAggValue(NaN, aggType, [], row, {});
            }
            return currValue;
        case oj.CubeAggType['MAX']:
            if (validCurr && validHash) {
                if (isNumCurr && isNumHash) {
                    return this._createAggValue(Math.max(currValue.value, hash.value), aggType, currValue.rows, row, {});
                }
                return this._createAggValue(NaN, aggType, currValue.rows, row, {});
            }
            if (validHash && !validCurr) {       
                if (isNumHash) {
                    return this._createAggValue(hash.value, aggType, [], row, {});
                }
                return this._createAggValue(NaN, aggType, [], row, {});
            }
            return currValue;
        case oj.CubeAggType['COUNT']:
            if (validCurr && validHash) {
                return this._createAggValue(currValue.value+1, aggType, currValue.rows, row, {});
            }
            if (validHash && !validCurr) {
                return this._createAggValue(1, aggType, [], row, {});
            }
            return currValue;
        case oj.CubeAggType['CUSTOM']:
            var callback = aggObj.callback;
            var val = callback.call(this, validCurr ? currValue.value : undefined, validHash ? hash.value : undefined);
            return this._createAggValue(val, aggType, validCurr ? currValue.rows : [], row, {});
    }
};

/**
 * Get or create the given axis
 * @private
 */
oj.Cube.prototype._getAxis = function(axis, levels) {
    if (axis >= this._axes.length) {
        // Must add on enough to cover
        var newElems = new Array(axis-this._axes.length+1);
        Array.prototype.push.apply(this._axes, newElems);
    }

    // Initialize it if blank
    if (!this._axes[axis]) {
        this._axes[axis] = new oj.CubeAxis(levels, axis, this);
    }
    
    return this._axes[axis];
};

/**
 * Create a key object from an axis-ordered array of slowest-to-fastest arrays of CubeAxisValues
 * @private
 */
oj.Cube.prototype._createCubeKeys = function(indices) {
    var axes = this.getAxes();
    var keys = new oj.CubeKeys();
    for (var a = 0; a < indices.length; a++) {
        keys = axes[a].GetCubeKeys(indices[a], keys);
    }
    return keys;
};
    
    
/**
 * @protected
 */
oj.Cube.prototype.ProcessLevel = function(levels, levelNum, currNode, row, keys, addKeys) {
    oj.Assert.failedInAbstractFunction();
};

/**
 * @protected
 */
oj.Cube.prototype.GenerateCube = function(layout) {
    oj.Assert.failedInAbstractFunction();
};

/**
 * @protected
 */
oj.Cube.prototype.GenerateLevel = function(level, axis) {    
    oj.Assert.failedInAbstractFunction();    
};

/**
 * @param {Object} dataValue
 * @returns {Object}
 * @private
 */
oj.Cube.prototype.GetAggType = function(dataValue) {
    oj.Assert.failedInAbstractFunction();
    return {};
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/

/**
 * @class oj.CubeAggType
 * @classdesc Valid aggregation types
 * @hideconstructor
 * @since 1.1.0
 * @export
 * @ojtsignore
 */
oj.CubeAggType = {
    /**
     * Sum the values
     * @type {string}
     */
    'SUM' : "sum",
    /**
     * Average the values
     * @type {string}
     */
    'AVERAGE' : "avg",
    /**
     * Calculate the standard deviation of the values
     * @type {string}
     */
    'STDDEV' : "stddev",
    /**
     * Calculate the variance of the values
     * @type {string}
     */
    'VARIANCE' : "variance",
    /**
     * Substitute a null for the value
     * @type {string}
     */
    'NONE' : "none",
    /**
     * Substitute the first value encountered
     * @type {string}
     */
    'FIRST' : "first",
    /**
     * Calculate the minimum of the values
     * @type {string}
     */
    'MIN' : "min",
    /**
     * Calculate the maximum of the values
     * @type {string}
     */
    'MAX' : "max",
    /**
     * Count the values
     * @type {string}
     */
    'COUNT' : "count",
    /**
     * Specify a custom callback function to do the aggregation
     * @type {string}
     */
    'CUSTOM': "custom"
};
 
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/

/**
 * @class oj.CubeAxisValue 
 * @classdesc Represents one value within an axis header
 * @see oj.CubeAxis
 * @since 1.1.0
 * @constructor
 * @export
 * @hideconstructor
 * @ojtsignore
 */
oj.CubeAxisValue = function(value, label, level, parent) {
    this.Init();
    this._children = [];
    this._parent = parent;
    this._cubeLevel = level;
    this._data = {};    
    this._data.value = value;
    this._data.label = label;
};
 
// Subclass from oj.Object 
oj.Object.createSubclass(oj.CubeAxisValue, oj.Object, "oj.CubeAxisValue");

/**
 * Initializes instance with the set options
 * @private
 */
oj.CubeAxisValue.prototype.Init = function() 
{
  oj.CubeAxisValue.superclass.Init.call(this);
}; 

/**
 * Get the {@link oj.CubeLevel} of which this value is a member
 * @return {oj.CubeLevel} the level to which this header value belongs
 * @memberof oj.CubeAxisValue
 * @export
 */
oj.CubeAxisValue.prototype.getLevel = function() {
    return this._cubeLevel;
};
 
/**
 * Get the number of levels this header value spans
 * @return {number} the number of levels this header value spans.  Typically one, but in the future something like a grand total may span all the available levels at a particular
 *                  axis value location
 * @memberof oj.CubeAxisValue
 * @export
 */
oj.CubeAxisValue.prototype.getDepth = function() {
    return 1;
};
 
/**
 * Get the parent oj.CubeAxisValues of this value
 * @return {Array.<oj.CubeAxisValue>} an array of oj.CubeAxisValues--all the slower-varying level axis values under which this value is grouped
 * @memberof oj.CubeAxisValue
 * @export
 */
oj.CubeAxisValue.prototype.getParents = function() {
    var parents = [];
    var parent = this._parent;
    // Add parents, don't add root
    while (parent && parent._parent) {
        parents.unshift(parent);
        parent = parent._parent;
    }
    return parents;
};
 
/**
 * Get an ordered list of the child oj.CubeAxisValues of this value, at the next level
 * @return {Array.<oj.CubeAxisValue>} an array of the children of this oj.CubeAxisValue
 * @memberof oj.CubeAxisValue
 * @export
 */
oj.CubeAxisValue.prototype.getChildren = function() {
    return this._children;
};

/**
 * Get the number of indices spanned by this value
 * @return {number} the number of indices spanned by this value
 * @memberof oj.CubeAxisValue
 * @export
 */
oj.CubeAxisValue.prototype.getExtent = function() {
    if (this._extent > -1) {
        return this._extent;
    }
    // Calculate it: add up all my children's extents--if there are none, my extent is one as I'm the innermost level
    if (!this._children || this._children.length === 0) {
        this._extent = 1;
    }
    else {
        this._extent = 0;
        for (var c = 0; c < this.getChildren().length; c++) {
            this._extent += this.getChildren()[c].getExtent();
        }
    }
    return this._extent;
};

/**
 * Get the starting index of this value within the axis
 * 
 * @return {number} the starting index of this value within the axis
 * @memberof oj.CubeAxisValue
 * @export
 */
oj.CubeAxisValue.prototype.getStart = function() {
    if (this._start > -1) {
        return this._start;
    }
    if (!this._parent) {
        // We're the root
        return 0;
    }
    // Add up all my earlier siblings' extents plus my parent's start--that's my start
    var start = this._parent.getStart();
    var currChild = this._parent._getPrevChild(this);    
    while (currChild) {
        start += currChild.getExtent();
        currChild = this._parent._getPrevChild(currChild);
    }
    this._start = start;
    return start;
};

/**
 * Get the actual value of this axis header location
 * @return {Object|string} the value at this location in the header
 * @memberof oj.CubeAxisValue
 * @export
 */
oj.CubeAxisValue.prototype.getValue = function() {
    return this._data.value;
};

/**
 * Get the label for this axis header location, if any.  If none, falls back to {@link getValue}
 * @returns {string|null} the label for this axis header value
 * @memberof oj.CubeAxisValue
 * @since 1.1.0
 * @ojstatus preview
 * @export
 */
oj.CubeAxisValue.prototype.getLabel = function() {
    if (this._data.label) {
        return this._data.label;
    }
    var value = this.getValue();
    if (value) {
      return value.toString();
    }
    return null;
};

/**
 * Return the axis value child with the level/value pair found in key
 * @protected
 */
oj.CubeAxisValue.prototype.GetChild = function(key) {
    var childLevel = this._getChildLevel();
    if (childLevel === null) {
        return null;
    }
    
    if (childLevel.isDataValue()) {
        // Special case: the key is the value
        return this._getDataValueChild(key);
    }
    var levelAttr = childLevel['attribute'];
    // Do we have a key value for this?
    var val = key[levelAttr];
    if (val) {
        // Yes, see if any of the children have this value
        for (var c = 0; c < this._children.length; c++) {
            if (this._children[c].getValue() === val) {
                return this._children[c];
            }
        }
    }
    return null;
};

/**
 * @private
 */
oj.CubeAxisValue.prototype._getDataValueChild = function(key) {
    for (var c = 0; c < this._children.length; c++) {
        var val = this._children[c].getValue();
        if (key.hasOwnProperty(val) && key[val] === val) {
            // found it
            return this._children[c];
        }
    }
    return null;
};

/**
 * @private
 */
oj.CubeAxisValue.prototype._getChildLevel = function() {
    if (this._children && this._children.length > 0) {
        return this._children[0].getLevel();
    }
    return null;
};

/**
 * Return the axis value child at the given index
 * @protected
 */
oj.CubeAxisValue.prototype.GetChildAt = function(index) {
  return this._findChild(index, 0, this._children.length - 1);
};

/**
 * @private
 */
oj.CubeAxisValue.prototype._findChild = function(index, start, end) {
    if (start > end) {
            return null;
    }

    var mid = Math.floor((start + end) / 2);
    var value = this._children[mid];

    var valStart = value.getStart();
    if (valStart > index) {
          return this._findChild(index, start, mid-1);
    }
    if (valStart + value.getExtent() - 1 < index) {
        return this._findChild(index, mid+1, end);
    }
    return value;
}
    
/**
 * Add this value to the node if not already there; return either the new cube axis value or the one where it was found
 * @protected
 */
oj.CubeAxisValue.prototype.AddNode = function(value, label, level) {
    // Check if this value is already in the child list
    for (var c = 0; c < this._children.length; c++) {
        if (this._children[c].getValue() === value) {
            // We already have this node: just return it so any other members below its level get put in the right place
            return this._children[c];
        }
    }
    // Not found: add a new child
    var newValue = new oj.CubeAxisValue(value, label, level, this);
    this._children.push(newValue);
    return newValue;
};

/**
 * Get a "hash value" for this axis value.  Format is levelAttr:value
 * @protected
 */
oj.CubeAxisValue.prototype.GetHashCode = function() {
    var obj = {};
    obj[this.getLevel()['attribute']] = this.getValue();
    return obj;
};
    
/**
 * Return the child immediately preceding this one
 * @protected
 */
oj.CubeAxisValue.prototype._getPrevChild = function(currChild) {
    for (var c = 0; c < this._children.length; c++) {
        if (this._children[c] === currChild) {
            // Found this child: see if there's a previous one
            if (c > 0) {
                return this._children[c-1];
            }
            // First child
            return null;
        }
    }
    // Not found
    return null;
};


/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/

/**
 * @class oj.CubeAxis
 * @since 1.1.0
 * @classdesc Represents an axis (a collection of levels) in a cube.  There is a set of oj.CubeAxisValues at each index within an axis,
 * up to one per oj.CubeLevel on the axis
 * @param {Array.<Object>} levels array of levels, slowest to fastest, to put in this axis
 * @param {number} axis the number of this axis
 * @param {oj.Cube} cube the cube on which to generate levels
 * 
 * @constructor
 * @export
 * @hideconstructor
 * @ojtsignore
 */
oj.CubeAxis = function(levels, axis, cube) {
    this.Init();
    this['axis'] = axis;
    this._levels = [];
    for (var i = 0; i < levels.length; i++) {
        this._levels.push(cube.GenerateLevel(levels[i], this));
    }
    this._cube = cube;
    // Set up a root for the tree of axis values
    this._values = new oj.CubeAxisValue(null, null, null, null);
};
 
// Subclass from oj.Object 
oj.Object.createSubclass(oj.CubeAxis, oj.Object, "oj.CubeAxis");

/**
 * Initializes instance with the set options
 * @private
 */
oj.CubeAxis.prototype.Init = function() 
{
  oj.CubeAxis.superclass.Init.call(this);
};

/**
 * Gets an ordered list (slowest to fastest varying) of all the levels represented within the axis.  Some indices along the axis may not be represented by
 * every level
 * @return {Array.<oj.CubeLevel>} an ordered array of {@link oj.CubeLevel}
 * @memberof oj.CubeAxis
 * @export
 */
oj.CubeAxis.prototype.getLevels = function() {
    return this._levels;
};
 
/**
 * Gets the total number of indices (data value locations) along this axis
 * @return {number} the total number of indices along the axis
 * @memberof oj.CubeAxis
 * @export
 */
oj.CubeAxis.prototype.getExtent = function() {
    return this._values.getExtent();
};
 
/**
 * Get the {@link oj.CubeAxisValue} objects at the given index, one for each level represented at this index within the axis.  Note that the number of oj.CubeAxisValues
 * returned may not match the overall oj.CubeLevel count on the axis
 * @param index {number} index for which to get the oj.CubeAxisValues
 * @return {Array.<oj.CubeAxisValue>} the ordered list of oj.CubeAxisValues at this index (slowest to fastest varying).
 * @memberof oj.CubeAxis
 * @export
 */
oj.CubeAxis.prototype.getValues = function(index) {
    var values = [];
    var node = this._values;
    while (node) {
        node = node.GetChildAt(index);
        if (node) {
            values.push(node);
        }
    }
    return values;
};

/**
 * Return the index of the given key within the axis.  Return -1 if not found.
 * 
 * @param {string} key
 * @returns {number} index of value found using key, -1 if not found
 * @memberof oj.CubeAxis
 * @export
 */
oj.CubeAxis.prototype.getIndex = function(key) {
    var keyVal = key ? JSON.parse(key) : {};
        
    var node = this._values;
    var lastNode = null;
    while (node) {
        lastNode = node;
        node = node.GetChild(keyVal);
    }
    return lastNode ? lastNode.getStart() : -1;
};

/**
 * @desc This axis' number
 * 
 * @type number
 * @memberof oj.CubeAxis
 * @export
 */
oj.CubeAxis.prototype.axis;


/**
 * Retrieve the hash code for the values at a given index
 * @protected
 */
oj.CubeAxis.prototype.GetCubeKeys = function(index, keys) {
    return this.GetPartialCubeKeys(index, this.getLevels().length-1, keys);
};

/**
 * Get the cube keys down to a particular level
 * @protected
 */
oj.CubeAxis.prototype.GetPartialCubeKeys = function(index, level, keys) {
    var values = this.getValues(index);
    var stopLevel = this.getLevels()[level];
    for (var v = 0; v < values.length; v++) {
        var val = values[v];
        if (val.getLevel().isDataValue()) {
            keys.AddDataValue(val.getValue());
        }
        else {
            keys.UpdateKeys(val);
        }
        // Have we hit the stop level?
        if (val.getLevel() === stopLevel) {
            break;
        }
    }    
    return keys;
};


/**
 * Distribute contents of row to this axis according to the layout, etc.
 * @protected
 */
oj.CubeAxis.prototype.ProcessRow = function(row, keys) {
    // Go over the levels--obtain a set of keys representing cell locations and data values of those intersections
    return this._cube.ProcessLevel(this, 0, this._values, row, keys, true);
};




/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

 /**
 * @class oj.CubeCellSet
 * @classdesc A CellSet represents a collection of cells.  The CellSet is an object returned by the success callback
 * of the fetchCells method on DataGridDataSource.  This implementation of CellSet is used by the
 * cube DataGridDataSource. 
 * @since 1.1.0
 * @constructor
 * @export
 * @hideconstructor
 * @ojtsignore
 * @see oj.CubeDataGridDataSource
 */
oj.CubeCellSet = function(cube, cellRange)
{
    // assert startRow/startColumn are number
    var startRow = cellRange.row ? cellRange.row.start : 0;
    var rowCount = cellRange.row ? cellRange.row.count : 0;
    var startColumn = cellRange.column ? cellRange.column.start : 0;
    var colCount = cellRange.column ? cellRange.column.count : 0;
    oj.Assert.assertNumber(startRow, null);
    oj.Assert.assertNumber(rowCount, null);
    oj.Assert.assertNumber(startColumn, null);
    oj.Assert.assertNumber(colCount, null);
    
    this._cube = cube;

    this._starts = {'row':startRow,'column':startColumn};
    
    // Get the data
    this._values = this._cube.getValues([{start:startColumn, count:colCount},{start:startRow,count:rowCount}]);

    var valArray = Array.isArray(this._values);
    
    colCount = valArray ? this._values.length : 1;
    if (colCount > 0) {
        rowCount = valArray ? this._values[0].length : 1;
    }
    this._counts = {'row':rowCount, 'column':colCount};
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available 
 * 2) the index specified is out of bounds. The indices are absolute within the entire data set, not just this cell set
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} the data object for the specified index.
 * @memberof oj.CubeCellSet
 * @export
 */
oj.CubeCellSet.prototype.getData = function(indexes)
{   
    var row = indexes['row'];
    var col = indexes['column'];
    var cell = Array.isArray(this._values) ? this._values[col-this._starts['column']][row-this._starts['row']] : this._values;
    if (cell) {
        return cell.getValue();
    }
    return null;
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available 
 * 2) the index specified is out of bounds. The indices are absolute across the entire data set not just within the cell set
 * @param {Object} indexes the index of each axis in which we want to retrieve the metadata from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} the metadata object for the specific index.  The metadata that the DataGrid supports are: 
 *         1) keys - the key (of each axis) of the cell.
 * @memberof oj.CubeCellSet
 * @export
 */
oj.CubeCellSet.prototype.getMetadata = function(indexes)
{
    // Get each axis' key
    var obj = {};
    obj['keys'] = {};
    obj['keys']['row'] = this._getAxisMetadata(indexes, 'row', 2);
    obj['keys']['column'] = this._getAxisMetadata(indexes, 'column', 1);
    
    return obj;
};

/**
 * @private
 */
oj.CubeCellSet.prototype._getAxisMetadata = function(indexes, axis, len) {
    var axes = this._cube.getAxes();
    
    if (indexes[axis] !== undefined && axes.length >= len) {
        var keys = new oj.CubeKeys();
        var axisIndex = oj.CubeDataGridDataSource._convertAxes(axis);
        keys = axes[axisIndex].GetCubeKeys(indexes[axis], keys);
        return keys.GetHashCodes()[0].key;
    }
    return null;
};

/**
 * Gets the start index of the result set for the specified axis.  Valid values are "row" and "column".
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 * @return {number} the start index of the result set for the specified axis.  
 * @memberof oj.CubeCellSet
 * @export
 */
oj.CubeCellSet.prototype.getStart = function(axis)
{
    return this._starts[axis];
};

/**
 * Gets the actual count of the result set for the specified axis.  Valid values are "row" and "column".
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 * @return {number} the actual count of the result set for the specified axis.  
 * @memberof oj.CubeCellSet
 * @export
 */
oj.CubeCellSet.prototype.getCount = function(axis)
{
    return this._counts[axis];
};

/**
 * Gets the extent of a particular row/column index within the context of the cellSet.
 * Extent is defined as the number of indexes along the appropriate axis spanned by the cell.
 * If the extent extends beyond the start and end of the requested cell range the extent should be trimmed to the edge of the requested cell range and the object for {'more': {'before', 'after'}} should have the value appropriate boolean set.
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from. 
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} an object containing two properties row and column. Each of those properties has two sub properties:
 *              extent: the number of absolute indexes spanned by the cell at this index
 *                      bounded by the edges of the result set for the specified axis.
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is available in the cellSet
 * @example <caption>In this example the cell spans 5 row indexes and 2 column indexes and there are more column indexes spanned by the cell that
 *              aren't included in this cellSet:</caption>
 * {
 *  'row': {'extent':5, 'more': {'before':false, 'after':false}},
 *  'column': {'extent':2, 'more': {'before':false, 'after':true}}
 * }
 * @export
 * @expose
 * @instance
 * @memberof oj.CubeCellSet
 */
oj.CubeCellSet.prototype.getExtent = function(indexes)
{
    return {'row': {'extent':1, 'more': {'before':false, 'after':false}}, 
            'column': {'extent':1, 'more': {'before':false, 'after':false}}};
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.CubeDataGridDataSource
 * @classdesc  A cubic/aggregating DataGridDataSource based on the oj.Cube
 * @see oj.Cube
 * @ojtsignore
 * @param {oj.Cube} cube the cube that will underpin the data source
 * @export
 * @extends oj.DataGridDataSource
 * @constructor
 * @since 1.1.0
 */
oj.CubeDataGridDataSource = function(cube)
{
    oj.CubeDataGridDataSource.superclass.constructor.call(this, cube);
};

// Subclass CubeDataGridDataSource to DataGridDataSource
oj.Object.createSubclass(oj.CubeDataGridDataSource, oj.DataGridDataSource, "oj.CubeDataGridDataSource");

/**
 * Set a new cube on the data source
 * @param {oj.Cube} cube
 * @return {undefined}
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.setCube = function(cube) {
    this.data = cube;
    this._fireRefresh();
};

/**
 * Pin any axes beyond the row and column to specific index values (to allow the idea of "paging" through a cube)
 * @param {Array.<Object>} indices an array of objects each containing an 'axis' attribute and a zero-based 'index' attribute giving the index to "pin" the axis to.
 * @return {undefined}
 * @memberof oj.CubeDataGridDataSource
 * @ojsignature  {target: "Type",
 *                value: "Array<oj.CubeDataGridDataSource.Indices>",
 *                for: "indices"}
 * @export
 */
oj.CubeDataGridDataSource.prototype.setPage = function(indices) {
    this.data.setPage(indices);
    this._fireRefresh();
};

/**
 * Object in setPage indices parameter
 * @typedef {Object} oj.CubeDataGridDataSource.Indices
 * @property {number} axis
 * @property {number} index
 */

/**
 * @private
 */
oj.CubeDataGridDataSource.prototype._fireRefresh = function() {
    var event = {};
    event['source'] = this;
    event['operation'] = "refresh";
    this.handleEvent("change", event);
};
    
/**
 * Returns the total number of rows or columns.  If the value return is not >= 0 then it is automatically assumed
 * that the total count is unknown.
 * @param {string} axis the axis in which we inquire for the total count.  Valid values are "row" and "column".
 * @return {number} the total number of rows/columns.
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.getCount = function(axis)
{
    var axisObj = this._getAxis(axis);
    
    return axisObj ? axisObj.getExtent() : 0;
};


/**
 * Returns whether the total count returned in getCount function is an actual or an estimate.
 * @param {string} axis the axis in which we inquire whether the total count is an estimate.  Valid values are 
 *        "row" and "column".
 * @return {string} "exact" if the count returned in getCount function is the actual count, "estimate" if the 
 *         count returned in getCount function is an estimate.  The default value is "exact".
 * @method
 * @name getCountPrecision
 * @memberof oj.CubeDataGridDataSource
 * @instance
 * @export
 */
oj.CubeDataGridDataSource.prototype.getCountPrecision = function(axis)
{
    return "exact";
};


/**
 * Fetch a range of headers from the data source.
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column"
 * @property {number} headerRange.start the start index of the range in which the header data are fetched
 * @property {number} headerRange.count the size of the range in which the header data are fetched
 * @param {Object} callbacks the optional callbacks to be invoke when fetch headers operation is completed.  The valid callback
 *        types are "success" and "error"
 * @property {function({startHeaderSet: Object}, {headerRange: Object}, {endHeaderSet: Object}):void=} callbacks.success the callback to invoke when fetch headers completed successfully.
 *        The function takes three parameters: HeaderSet object representing start headers, headerRange object passed into the original fetchHeaders call,
 *        and a HeaderSet object representing the end headers along the axis.
 * @property {function({status: Object}):void=} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.  
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @return {void}
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.fetchHeaders = function(headerRange, callbacks, callbackObjects)
{
    
    var cubeheaders = new oj.CubeHeaderSet(this._getAxis(headerRange['axis']), this.data, headerRange['start'], headerRange['count']);
    callbacks['success'].call(callbackObjects ? callbackObjects['success'] : undefined, cubeheaders, headerRange);
};

/**
  * @param {Array.<Object>} cellRange Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.<p>
 *        axis: the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".<br>
 *        start: the start index of the range for this axis in which the cells are fetched.<br>
 *        count: count the size of the range for this axis in which the cells are fetched. <br>
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.  The valid callback
 *        types are "success" and "error"
 * @property {function({cellSet: Object}, {cellRange: Object}):void=} callbacks.success the callback to invoke when fetch cells completed successfully.  The passed arguments are CellSet and the cellRange
 * @property {function({status: Object}):void=} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.  
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @return {void}
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.fetchCells = function(cellRange, callbacks, callbackObjects)
{
    var obj = {};
    for (var i = 0; i < cellRange.length; i++) {
        var start = cellRange[i]['start'] === undefined ? 0 : cellRange[i]['start'];
        if (cellRange[i].axis === 'row') {
            var count = cellRange[i]['count'] === undefined ? this.data.getAxes()[1].getExtent() : cellRange[i]['count'];
            obj.row = {start: start, count: count};
        }
        if (cellRange[i].axis === 'column') {
            var count = cellRange[i]['count'] === undefined ? this.data.getAxes()[0].getExtent() : cellRange[i]['count'];
            obj.column = {start: start, count: count};
        }
    }
    var cubecells = new oj.CubeCellSet(this.data, obj);
    callbacks['success'].call(callbackObjects ? callbackObjects['success'] : undefined, cubecells, cellRange);
};

/**
 * Returns the row and column index based on the keys.
 * @param {Object} keys the key for each axis
 * @property {number} keys.row the key for the row axis
 * @property {number} keys.column the key for the column axis
 * @return {Promise.<Object>} a Promise object which when resolved returns an object containing the key for each axis
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.keys = function(keys)
{
    var retObj = {};
    retObj = this._getKey(keys, 'row', retObj);
    retObj = this._getKey(keys, 'column', retObj);
    return Promise.resolve(retObj);
};

oj.CubeDataGridDataSource.prototype._getKey = function(indexes, axis, retObj) {
    var axisObj = this._getAxis(axis);
    var item = indexes[axis];
    var keys = new oj.CubeKeys();
    keys = axisObj ? axisObj.GetCubeKeys(item, keys) : "";
    retObj[axis] = keys.GetHashCodes()[0].key;
    return retObj;
};


/**
 * Returns the keys based on the indexes. 
 * @param {Object} indexes the index for each axis
 * @property {number} indexes.row the index for the row axis
 * @property {number} indexes.column the index for the column axis
 * @return {Promise.<Object>} a Promise object which when resolved returns an object containing the index for each axis
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.indexes = function(indexes)
{
    var retObj = {};
    retObj = this._getIndex(indexes, 'row', retObj);
    retObj = this._getIndex(indexes, 'column', retObj);

    return Promise.resolve(retObj);
};

/**
 * @private
 */
oj.CubeDataGridDataSource.prototype._getIndex = function(keys, axis, retObj) {
    retObj[axis] = this._getAxis(axis).getIndex(keys[axis]);
    return retObj;
};

/**
 * @ignore
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.sort = function(criteria, callbacks, callbackObjects)
{
    oj.Assert.failedInAbstractFunction();
};

/**
 * @ignore
 * @memberof oj.CubeDataGridDataSource
 * @export
 */ 
oj.CubeDataGridDataSource.prototype.move = function(rowToMove, referenceRow, position, callbacks, callbackObjects)
{
    oj.Assert.failedInAbstractFunction();
};

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {string} position The position of the moved row relative to the reference row. 
 *        Valid values are: "before", "after".
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @memberof oj.CubeDataGridDataSource
 * @export
 */ 
oj.CubeDataGridDataSource.prototype.moveOK = function(rowToMove, referenceRow, position)
{
    return "invalid";
};

/**
 * Determines whether this DataGridDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none", "row", "column".  
 *         For "move", the valid return values are: "row", "none".  
 *         Returns null if the feature is not recognized.
 * @memberof oj.CubeDataGridDataSource
 * @export
 */
oj.CubeDataGridDataSource.prototype.getCapability = function(feature)
{
    switch (feature) {
        case "sort":
            return "none";
        case "move":
            return "none";
    }
    return null;
};

/**
 * @private
 */
oj.CubeDataGridDataSource._convertAxes = function(axis) {
    return axis === "row" ? 1 : 0;
};

/**
 * @private
 */
oj.CubeDataGridDataSource.prototype._getAxis = function(axis) {
    var axisNum = oj.CubeDataGridDataSource._convertAxes(axis);
    var axes = this.data.getAxes();
    if (axes.length > axisNum) {
        return axes[axisNum];
    }
    return null;
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/

/**
 * @class oj.CubeDataValue 
 * @see oj.Cube#getValues
 * @since 1.1.0
 * @classdesc Objects returned by oj.Cube.getValues() and representing a single cell within the cube's data body.
 * Note that more functions or properties can be added to this object in the future as required
 * @param {Object} value
 * @param {Array.<number>} indices
 * @param {string=} aggType
 * @param {Array.<Object>=} rows
 * @param {number=} square
 * @constructor
 * @export
 * @hideconstructor
 * @ojtsignore
 */
oj.CubeDataValue = function(value, indices, aggType, rows, square) {
    this.Init();
    this._data = {};
    this._data.value = value;
    this._data.indices = indices;
    this._data.aggType = aggType;
    this._data.rows = rows;
    this._data.square = square;
};
 
// Subclass from oj.Object 
oj.Object.createSubclass(oj.CubeDataValue, oj.Object, "oj.CubeDataValue");

/**
 * Initializes instance with the set options
 * @private
 */
oj.CubeDataValue.prototype.Init = function() 
{
  oj.CubeDataValue.superclass.Init.call(this);
};

/**
 * Get the actual data value in this cell
 * @return {Object|null|number|string} the data value
 * @memberof oj.CubeDataValue
 * @export
 */
oj.CubeDataValue.prototype.getValue = function() {
    // Need a special case for variance/stddev--they need a final computation
    switch (this._data.aggType) {
        case oj.CubeAggType['STDDEV']:
            return Math.sqrt(this._getVariance());
        case oj.CubeAggType['VARIANCE']:
            return this._getVariance();
        default:
            return this._data.value;
    }
};
 
/**
 * Get the indices in axis order of the location of this data cell within the cube data body
 * @return {Array.<number>} an array of numeric indices in axis order
 * @memberof oj.CubeDataValue
 * @export
 */
oj.CubeDataValue.prototype.getIndices = function() {
    return this._data.indices;
};
 
/**
 * Get the attribute/value pairs of the rows that were aggregated to create this data value.  The result may be a single row's set of attribute/value pairs if
 * no aggregation took place
 * @return {Array.<Object>|null|undefined} the array of attribute/value pairs representing rows aggregated to create this data value
 * @memberof oj.CubeDataValue
 * @export
 */
oj.CubeDataValue.prototype.getRows = function() {
    return this._data.rows;
};
 
/**
 * Get the aggregation type used to create this data value, if any
 * @return {string|undefined} the aggregation type used for this data value
 * @memberof oj.CubeDataValue
 * @export
 */
oj.CubeDataValue.prototype.getAggregation = function() {
    return this._data.aggType;
};

/**
 * @private
 */
oj.CubeDataValue.prototype._getVariance = function() {
    if (isNaN(this._data.square)) {
        return NaN;
    }
    var count = this._data.rows.length;
    return (count > 1) ? this._data.square / (count-1) : 0.0;
};

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * @class oj.CubeHeaderSet
 * @classdesc A HeaderSet represents a collection of headers.  The HeaderSet is an object returned by the success callback
 * of the fetchHeaders method on CubeDataGridDataSource.   
 * @since 1.1.0
 * @constructor
 * @export
 * @hideconstructor
 * @ojtsignore
 * @see oj.CubeDataGridDataSource
 */
oj.CubeHeaderSet = function(axis, cube, start, count)
{
    // The oj.Cube
    this._cube = cube;
    // The oj.CubeAxis
    this._axis = axis;    
    this._start = start === undefined ? 0 : start;
    this._count = count === undefined ? this._axis.getExtent() : Math.min(count, this._axis.getExtent()-start);
    this._end = start + count - 1;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the index of the header in which we want to retrieve the header from.  This is an absolute index across the entire axis
 * @param {number} level the level of the header
 * @return {*} the data object for the specific index.
 * @memberof oj.CubeHeaderSet
 * @export
 */
oj.CubeHeaderSet.prototype.getData = function(index, level)
{
    var val = this._getValue(index, level);
    return val ? val.getLabel() : null;
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source can optionally return are:
 *  1) sortDirection - the initial sort direction of the header.  Valid values are "ascending" and "descending".
 *  2) key - the key of the row/column header.
 * @param {number} index the index of the header in which we want to retrieve the metadata from.   This is an absolute index across the entire axis
 * @param {number} level the level of the header
 * @return {Object} the metadata object for the specific index.
 * @memberof oj.CubeHeaderSet
 * @export
 */
oj.CubeHeaderSet.prototype.getMetadata = function(index, level)
{    
    var keys = new oj.CubeKeys();
    keys = this._axis.GetPartialCubeKeys(index, level, keys);
    var hash = keys.GetHashCodes();
    
    return hash && hash.length > 0 ? {'key':hash[0].key} : null;
};

/**
 * Gets the actual number of levels of the result set for the specified axis.
 * @return {number} the number of levels of the result set for the specified axis. 
 * @memberof oj.CubeHeaderSet
 * @export
 */
oj.CubeHeaderSet.prototype.getLevelCount = function()
{
    return this._axis.getLevels().length;
};
 
/**
 * Gets the extent of an index on a particular level.  This is the extent such as it is within the defined header set, not across the entire
 * axis.  If the header falls across the beginning of the headerset or the end, then its extent should be only that part lying within the header set
 * @param {number} index the index of the header (absolute within the axis)
 * @param {number} level the level of the header
 * @return {Object} an object containing two values. example: {extent:5, more: {'before':false, 'after':true}}
 *              extent: the number of slices of the result set for the specified axis. 
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there may be more rows/columns before or after the headerSet
 * @memberof oj.CubeHeaderSet
 * @export
 */
oj.CubeHeaderSet.prototype.getExtent = function(index, level)
{
    var val = this._getValue(index, level);
    var extent = val.getExtent();
    var start = val.getStart();
    var end = start + extent - 1;
    var before = index > start;
    var after = index < start + extent - 1;
    
    if (start < this._start) {
        // Need to subtract this overage from the extent
        extent -= (this._start-start);
    }
    if (end > this._end) {
        // true extent overruns the header set--adjust it down by that much
        extent -= (end-this._end);
    }
    return {'extent':extent, 'more': {'before':before, 'after':after}};
};

/**
 * Gets the depth of an index starting at a particular level.
 * @param {number} index the index of the header (absolute within the axis)
 * @param {number} level the level of the header
 * @return {number} - the number of levels of the result set for the specified axis. 
 * @memberof oj.CubeHeaderSet
 * @export
 */
oj.CubeHeaderSet.prototype.getDepth = function(index, level)
{
    var val = this._getValue(index, level);
    return val.getDepth();
};

/**
 * Gets the actual count of the result set.
 *
 * @return {number} the actual count of the result set.  
 * @memberof oj.CubeHeaderSet
 * @export
 */
oj.CubeHeaderSet.prototype.getCount = function()
{
    return this._count;
};

/**
 * Gets the start index of the result set.
 * @return {number} the start index of the result set.
 * @memberof oj.CubeHeaderSet
 * @export
 */
oj.CubeHeaderSet.prototype.getStart = function()
{
    return this._start;
};

/**
 * @private
 */
oj.CubeHeaderSet.prototype._getValue = function(index, level) {
    if (level === undefined) {
        level = 0;
    }
    var values = this._axis.getValues(index);
    if (values && values.length > level) {
        return values[level];
    }
    return null;    
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/

/**
 * @constructor
 * @private
 */
oj.CubeKeys = function() {
    // Nodes referencing the data values
    this._key = [];
    
    // List of dataValue name/data value pairs
    this._data = [];
};


/**
 * @protected
 */
oj.CubeKeys.prototype.UpdateKeys = function(node) {
    this._key.push(node);
};

/**
 * @param {string} name
 * @param {Object=} value
 * @private
 */
oj.CubeKeys.prototype.AddDataValue = function(name, value) {
    this._data.push({name:name,value:value});
};

/**
 * @protected
 */
oj.CubeKeys.prototype.GetHashCodes = function() {
    var codes = [];
    
    var keyHash = this._buildKeyHash();
    
    if (this._data.length === 0) {
        codes.push({key:JSON.stringify(keyHash)});
    }
    else {
        // One pairing for each data value
        for (var d = 0; d < this._data.length; d++) {
            var copy = $.extend(true, {}, keyHash);
            // This is done for easier lookups when finding child
            copy[this._data[d].name] = this._data[d].name;
            codes.push({key:JSON.stringify(copy),dataValue:this._data[d].name,value:this._data[d].value});
        }
    }
    return codes;
};

/**
 * @private
 */
oj.CubeKeys.prototype._buildKeyHash = function() {
    var keyHash = {};
    for (var k = 0; k < this._key.length; k++) {
        var hc = this._key[k].GetHashCode();
        for (var p in hc) {
            if (hc.hasOwnProperty(p)) {
                keyHash[p] = hc[p];
            }
        }
    }
    return keyHash;
};


/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/

/**
 * @class oj.CubeLevel
 * @classdesc  Represents a level within an axis.  The level is tied to an attribute within the original rowset.  Each level contains a number of [oj.CubeAxisValues]{@link oj.CubeAxisValue}, all from the same attribute
 * @see oj.CubeAxis
 * @since 1.1.0
 * @constructor
 * @export
 * @hideconstructor
 * @ojtsignore
 */
 oj.CubeLevel = function(attribute, axis, dataValue) {
    this.Init();
    this['attribute'] = attribute;
    this._axisObj = axis;
    this['axis'] = axis['axis'];
    this._dataValue = dataValue;
 };
 
 // Subclass from oj.Object 
oj.Object.createSubclass(oj.CubeLevel, oj.Object, "oj.CubeLevel");

/**
 * Initializes instance with the set options
 * @private
 */
oj.CubeLevel.prototype.Init = function() 
{
  oj.CubeLevel.superclass.Init.call(this);
};

/**
 * The rowset attribute this level represents
 * @type {string}
 * @memberof oj.CubeLevel
 * @export
 */
oj.CubeLevel.prototype.attribute;
 
/**
 * The axis this level is a member of
 * @type {number}
 * @memberof oj.CubeLevel
 * @export
 */
oj.CubeLevel.prototype.axis;
 

/**
 * Return the oj.CubeAxisValue at the given index within this level
 * @param index {number} the axis index for which to get the value
 * @return {oj.CubeAxisValue} the value at the given location within the level
 * @memberof oj.CubeLevel
 * @export
 */
oj.CubeLevel.prototype.getValue = function(index) {
    var values = this._axisObj.getValues(index);
    if (values) {
        // Find this level within the returned list
        for (var v = 0; v < values.length; v++) {
            if (values[v].getLevel() === this) {
                return values[v];
            }
        }
    }
    return null;
};
 
/**
 * Does this level represent the data values within the cube?
 * @return {boolean} true if this level does represent the data values within the cube; false otherwise
 * @memberof oj.CubeLevel
 * @export
 */
oj.CubeLevel.prototype.isDataValue = function() {
    return this._dataValue;
};

/**
 * @type {boolean}
 * @private
 */
oj.CubeLevel.prototype._dataValue = false;

/**
 * @type {oj.CubeAxis}
 * @private
 */
oj.CubeLevel.prototype._axisObj = null;
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/


/**
 * @class oj.DataColumnCube
 * 
 * @classdesc Creates an object used to convert rowset data into grouped "cubic" data, where the data values are specified by a single attribute within the rowset
 * (dataValues.labelAttr) and their header values designated by another rowset attribute (dataValues.valueAttr).
 *
 * @param {Array.<Object>} rowset An array of objects containing attribute/value pairs.  The entire array or collection
 *                       will be read to group its attributes according to information given by layout and dataValues
 * @param {Array.<Object>} layout An array of objects containing two properties: axis - a number representing the number of the axis of the levels;
 *                       levels - a slowest-to-fastest varying ordered array of objects containing:
 *                       attribute - an attribute of the rowset objects to assign to this axis and level.  If the attribute
 *                       is the same as that specified by labelAttr, then this level is the data value level
 * @param {Object} dataValues an object containing the following properties: labelAttr - the rowset attribute used to group the data values in the header
 *                            valueAttr - the rowset attribute used for the actual data values; (optional) defaultAggregation - the default type of
 *                            oj.CubeAggType to use to aggregate data values where necessary.  If the type is 'CUSTOM' then this should be an object with a 'type' property of oj.CubeAggType['CUSTOM'] and a 'callback' property specifying a function to call with each value.  The function takes two arguments, the first being the running value for the cell being calculated, the second being the new value to be aggregated with that running value;
 *                            (optional) aggregation: an array of objects containing:
 *                            value - the value of labelAttr for which this aggregation should apply; aggregation - the oj.CubeAggType for that value; if aggregation is 'CUSTOM', then a 'callback' property should be added specifying a function (for arguments see above) to call with each value
 *                            (defaults to sum)
 * @see oj.Cube
 * @constructor
 * @since 1.1.0
 * @augments oj.Cube
 * @ojsignature  [{target: "Type", value: "Array<oj.DataColumnCube.Layout>", for: "layout"}, {target: "Type",
 *                value: "Array<oj.DataColumnCube.DataValue>",
 *                for: "dataValues"}]
 * @export
 * @ojtsignore
 */
oj.DataColumnCube = function(rowset, layout, dataValues) {
    this.Init();
    this._dataValues = dataValues;
    this._valueAttr = dataValues['valueAttr'];
    this._labelAttr = dataValues['labelAttr'];
    var defAgg = dataValues['defaultAggregation'];
    this._defaultAggregation = defAgg ? oj.DataColumnCube._getDefaultAgg(defAgg) : {aggregation:oj.CubeAggType['SUM']};
    this._aggregation = dataValues['aggregation'];
    this._buildAggTypeLookup();
    
    oj.DataColumnCube.superclass.constructor.call(this, rowset, layout);
};

/**
 * @typedef {Object} oj.DataColumnCube.Aggregation
 * @property {string} value
 * @property {oj.CubeAggType} aggregation
 */

/**
 * @typedef {Object} oj.DataColumnCube.LayoutLevels
 * @property {string} attribute
 */
 
/**
 * @typedef {Object} oj.DataColumnCube.Layout
 * @property {number} axis
 * @property {Array<oj.DataColumnCube.LayoutLevels>} levels
 */
 
/**
 * @typedef {Object} oj.DataColumnCube.DataValue
 * @property {string} valueAttr
 * @property {string} labelAttr
 * @property {oj.CubeAggType=} defaultAggregation
 * @property {Array<oj.DataColumnCube.Aggregation>=} aggregation
 */

// Subclass from oj.Cube 
oj.Object.createSubclass(oj.DataColumnCube, oj.Cube, "oj.DataColumnCube");

/**
 * Initializes instance with the set options
 * @private
 */
oj.DataColumnCube.prototype.Init = function() 
{
  oj.DataColumnCube.superclass.Init.call(this);
};

/**
 * @protected
 */
oj.DataColumnCube.prototype.BuildCube = function() {
    oj.DataColumnCube.superclass.BuildCube.call(this);
};

/**
 * @protected
 */
oj.DataColumnCube.prototype.GetAggType = function(dataValue) {
    if (this._dataValueAggType[dataValue]) {
        return this._dataValueAggType[dataValue];
    }
    return this._defaultAggregation;
};

/**
 * @protected
 */
oj.DataColumnCube.prototype.GenerateCube = function(layout) {
    return new oj.DataColumnCube(null, layout, this._dataValues);
};

/**
 * @protected
 */
oj.DataColumnCube.prototype.GenerateLevel = function(level, axis) {    
    if (level['attribute'] === this._labelAttr) {
        // Data value level
        return new oj.CubeLevel(level['attribute'], axis, true);
    }
    // Regular level
    return new oj.CubeLevel(level['attribute'], axis, false);
};

/**
 * @protected
 */
oj.DataColumnCube.prototype.ProcessLevel = function(axis, levelNum, currNode, row, keys, addKeys) {
    if (levelNum >= axis.getLevels().length) {
        return keys;
    }
    var level = axis.getLevels()[levelNum];
    
    // Not the data value: process this and call for the next
    var value = row[level['attribute']];
    var node = currNode.AddNode(value, null, level);
    
    if (level.isDataValue()) {
        keys.AddDataValue(value, row[this._valueAttr]);
    }
    else {
        keys.UpdateKeys(node);
    }    
    
    return this.ProcessLevel(axis, levelNum+1, node, row, keys, addKeys);
};

/**
 * @private
 */
oj.DataColumnCube._getDefaultAgg = function(agg) {
    if (oj.StringUtils.isString(agg)) {
        return {aggregation:agg};
    }
    return {aggregation:agg['aggregation'],callback:agg['callback']};
};
    
/**
 * @private
 */
oj.DataColumnCube.prototype._buildAggTypeLookup = function() {
    this._dataValueAggType = [];
    if (this._aggregation) {
        for (var i = 0; i < this._aggregation.length; i++) {
            var dv = this._aggregation[i];
            var agg = dv['aggregation'];
            this._dataValueAggType[dv['value']] = agg ? {aggregation:agg, callback:dv['callback']} : this._defaultAggregation;
        }
    }
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*jslint browser: true*/



/**
 * @class oj.DataValueAttributeCube
 * @classdesc Creates an object used to convert rowset data into grouped "cubic" data
 *
 * @param {Array.<Object>} rowset An array of objects containing attribute/value pairs.  The entire array or collection
 *                       will be read to group its attributes according to information given by layout and dataValues
 * @param {Array.<Object>} layout An array of objects containing two properties: axis - a number representing the number of the axis for of the levels;
 *                       levels - a slowest-to-fastest varying ordered array of objects containing one or two properties:
 *                       attribute - an attribute of the rowset objects or oj.Models to assign to this axis and level, or dataValue:true to indicate that this
 *                       axis and level is the position of the data values
 * @param {Array.<Object>} dataValues an array of objects containing these properties: attribute - the name of an attribute in the rowset representing data; (optional) label - a display label to use for this data attribute;
 *                           (optional) aggregation - the oj.CubeAggType for this data value attribute (defaults to sum).  If the type is 'CUSTOM' then a 'callback' property should be specified which is a function to call with each value.  The function takes two arguments, the first being the running value for the cell being calculated, the second being the new value to be aggregated with that running value
 * @see oj.Cube
 * @constructor
 * @since 1.1.0
 * @augments oj.Cube
 * @ojsignature  [{target: "Type", value: "Array<oj.DataValueAttributeCube.Layout>", for: "layout"}, {target: "Type",
 *                value: "Array<oj.DataValueAttributeCube.DataValue>",
 *                for: "dataValues"}]
 * @export
 * @ojtsignore
 */
oj.DataValueAttributeCube = function(rowset, layout, dataValues)
{
    this.Init();
    this._dataValues = dataValues;    
    this._aggTypeLookup = this._buildAggTypeLookup();
    oj.DataValueAttributeCube.superclass.constructor.call(this, rowset, layout);
};

/**
 * @typedef {Object} oj.DataValueAttributeCube.Aggregation
 * @property {string} value
 * @property {oj.CubeAggType} aggregation
 */

/**
 * @typedef {Object} oj.DataValueAttributeCube.LayoutLevels
 * @property {string=} attribute
 * @property {boolean=} dataValue
 */
 
/**
 * @typedef {Object} oj.DataValueAttributeCube.Layout
 * @property {number} axis
 * @property {Array<oj.DataValueAttributeCube.LayoutLevels>} levels
 */
 
/**
 * @typedef {Object} oj.DataValueAttributeCube.DataValue
 * @property {string} attribute
 * @property {string=} label
 * @property {oj.CubeAggType=} aggregation
 */

// Subclass from oj.Cube 
oj.Object.createSubclass(oj.DataValueAttributeCube, oj.Cube, "oj.DataValueAttributeCube");

/**
 * 
 * @type {Array.<Object>}
 * @private
 */
oj.DataValueAttributeCube.prototype._rows = null;

/**
 * Initializes instance with the set options
 * @private
 */
oj.DataValueAttributeCube.prototype.Init = function() 
{
  oj.DataValueAttributeCube.superclass.Init.call(this);
};


/**
 * Construct a cube given the information from the constructor.  Walk each row, assign it to its axis, aggregate values as 
 * need be
 * @protected
 */
oj.DataValueAttributeCube.prototype.BuildCube = function() {
    oj.DataValueAttributeCube.superclass.BuildCube.call(this);
    
};

/**
 * @protected
 */
oj.DataValueAttributeCube.prototype.GetAggType = function(dataValue) {
    return this._dataValueAggType[dataValue];
};

/**
 * @protected
 */
oj.DataValueAttributeCube.prototype.GenerateLevel = function(level, axis) {    
    if (level['dataValue']) {
        // Data value level
        return new oj.CubeLevel(null, axis, true);
    }
    // Regular level
    return new oj.CubeLevel(level['attribute'], axis, false);
};

/**
 * @protected
 */
oj.DataValueAttributeCube.prototype.GenerateCube = function(layout) {
    return new oj.DataValueAttributeCube(null, layout, this._dataValues);
};
    
/**
 * @protected
 */
oj.DataValueAttributeCube.prototype.ProcessLevel = function(axis, levelNum, currNode, row, keys, addKeys) {
    if (levelNum >= axis.getLevels().length) {
        return keys;
    }
    var level = axis.getLevels()[levelNum];
    
    if (level.isDataValue()) {
        // For each defined data attribute, add a node and then process the rest of the levels under them
        // This needs to be handled by the cube, because different types of cube data setups require different processing        
        return this._processDataValue(axis, currNode, row, levelNum, keys);
    }
    // Not the data value: process this and call for the next
    var value = row[level['attribute']];
    var node = currNode.AddNode(value, null, level);
    
    // Gives us control to avoid adding duplicate key sets below each measure coming from one detail row
    if (addKeys) {
        keys.UpdateKeys(node);
    }
    return this.ProcessLevel(axis, levelNum+1, node, row, keys, addKeys);
};

/**
 * Insert the data values into the given node and then process the remaining levels
 * @private
 */
oj.DataValueAttributeCube.prototype._processDataValue = function(axis, node, row, levelNum, keys) {
    // Roll over each data value, place an entry for each, and follow up the rest of the levels under that entry
    var dsKeysAdded = true;
    for (var d = 0; d < this._dataValues.length; d++) {
        var dvAttr = this._dataValues[d]['attribute'];
        var dvLabel = this._dataValues[d]['label'];
        // Only add a node if this row has an entry for the dataValue
        if (row.hasOwnProperty(dvAttr)) {
            var currNode = node.AddNode(dvAttr, dvLabel, axis.getLevels()[levelNum]);
            keys.AddDataValue(dvAttr, row[dvAttr]);

            // Process the rest of the level.  Only have new nodes added to data storage keys for the first data value we process--don't want to duplicate them
            this.ProcessLevel(axis, levelNum+1, currNode, row, keys, dsKeysAdded);
            // Shut off the key adding for future passes
            dsKeysAdded = false;
        }
    }
    return keys;
};

/**
 * @private
 */
oj.DataValueAttributeCube.prototype._buildAggTypeLookup = function() {
    this._dataValueAggType = [];
    for (var i = 0; i < this._dataValues.length; i++) {
        var dv = this._dataValues[i];
        var agg = dv['aggregation'];            
        this._dataValueAggType[dv['attribute']] = agg ? {aggregation:dv['aggregation'], callback:dv['callback']} : {aggregation:oj.CubeAggType['SUM'], callback:dv['callback']};
    }
};
/*jslint browser: true,devel:true*/
/**
 * The interface for oj.IndexerModel which should be implemented by all object instances
 * bound to the data parameter for ojIndexer. 
 * @export
 * @since 1.2
 * @interface
 */
oj.IndexerModel = function()
{
};

// without the at-name tag, JSDoc tool prepends a "." to the field name for some reason, which messes up the QuickNav.
/**
 * Constant for the section that represents all non-letters including numbers and symbols.
 * @export
 * @expose
 * @type {Object}
 * @property {string} id The id of this section
 * @property {string} label The label of this section
 * @name SECTION_OTHERS
 * @memberof oj.IndexerModel
 */
oj.IndexerModel.SECTION_OTHERS = {'id': '__others__', 'label': oj.Translations.getTranslatedString('oj-ojIndexer.indexerOthers')};

/**
 * Make a section current in the Indexer.  The implementation should scroll the associated ListView so that the section becomes visible.
 * @param {string|Object} section the current section
 * @return {Promise.<string>|Promise.<Object>} a Promise which when resolved will return the section that the associated ListView actually scrolls to.  
 *                   For example, the implementation could choose to scroll to the next available section in ListView if no data 
 *                   exists for that section.  
 * @method
 * @name setSection
 * @memberof oj.IndexerModel
 * @instance
 */

/**
 * Returns an array of objects each representing a section in the associated ListView.  The section object could either be
 * a String or an object containing at least a 'label' field.  For example, the implementation may return an array of Strings 
 * representing letters of the alphabet. Or it may return an array of objects each containing a 'label' field for the section 
 * titles. 
 * @return {Array.<string>|Array.<Object>} an array of all indexable sections 
 * @method
 * @name getIndexableSections
 * @memberof oj.IndexerModel
 * @instance
 */

/**
 * Returns an array of objects each representing a section that does not have a corresponding section in the associated ListView.  
 * It must be a subset of the return value of <code>getIndexableSections</code>.  Return null or undefined if there's nothing missing.
 * @return {Array.<string>|Array.<Object>} an array of missing sections
 * @method
 * @name getMissingSections
 * @memberof oj.IndexerModel
 * @instance
 */
/**
 * Implementation of the IndexerModel used by ListView.  This implementation groups the data based on the first letter of the 
 * group header text and the alphabet of the current locale.
 * @export
 * @param {Object} listview the internal ListView instance
 * @class oj.ListViewIndexerModel
 * @implements oj.IndexerModel
 * @classdesc Implementation of IndexerModel used by ListView. 
 * @extends oj.EventSource
 * @constructor
 * @ignore
 */
oj.ListViewIndexerModel = function(listview)
{
    this.listview = listview;
    this.Init();
};

// Subclass from oj.EventSource 
oj.Object.createSubclass(oj.ListViewIndexerModel, oj.EventSource, "oj.ListViewIndexerModel");

/**
 * @export
 * Returns the sections displayed by the Indexer.
 * @returns {Array.<Object>} an array of sections
 * @memberof! oj.ListViewIndexerModel
 */
oj.ListViewIndexerModel.prototype.getIndexableSections = function()
{
    var sections = this.listview.ojContext.getTranslatedString("indexerCharacters");
    return sections.split("|");
};

/**
 * @export
 * Returns the sections that are missing in the associated ListView.    
 * @return {Array.<Object>} an array of sections that are missing.
 * @memberof! oj.ListViewIndexerModel
 */
oj.ListViewIndexerModel.prototype.getMissingSections = function()
{
    if (this.missingSections == null)
    {
        this.missingSections = this._getMissingSections();
    }

    return this.missingSections;
};

/**
 * Returns the sections that are currently missing.
 * @private
 */
oj.ListViewIndexerModel.prototype._getMissingSections = function()
{
    var results = [], groupItems, sections, i, section, found, content;

    groupItems = this.listview._getGroupItemsCache();

    sections = this.getIndexableSections();
    for (i=0; i<sections.length; i++)
    {
        section = sections[i];
        found = false;
        groupItems.each(function(index) 
        {
            content = $(this).text();
            if (content.length > 0 && content.charAt(0) == section)
            {
                found = true;
                return false;
            }            
        });

        if (!found)
        {
            results.push(section);
        }
    }

    return results;
};

/**
 * @export
 * Sets the current section.  When associated with a ListView, this will scroll the ListView to the corresponding group header.
 * @param {Object} section the current section
 * @return {Promise} a Promise object which when resolve will return the section that the IndexerModel actually sets as current.
 * @memberof! oj.ListViewIndexerModel
 */
oj.ListViewIndexerModel.prototype.setSection = function(section)
{
    if (section == oj.IndexerModel.SECTION_OTHERS)
    {
        return this._setOtherSection();
    }
    else
    {
        return this._setSection(section);
    }
};

/**
 * Sets the 'Other' section as current
 * @private
 */
oj.ListViewIndexerModel.prototype._setOtherSection = function()
{
    var sections, self = this, match, content, i, section;

    sections = this.getIndexableSections();

    return new Promise(function(resolve, reject) 
    {
        match = null;

        // find the group header that DOES NOT match ANY of the sections
        self.listview._getGroupItemsCache().each(function(index) 
        {
            content = $(this).text();
            for (i=0; i<sections.length; i++)
            {
                section = sections[i];
                if (content.indexOf(section) == 0)
                {
                    // skip and check next group header
                    return true;
                }            
            }

            match = this;
            return false;
        });

        if (match)
        {
            self.listview._scrollToGroupHeader(match);
            resolve(oj.IndexerModel.SECTION_OTHERS);
        }
        else
        {
            resolve(null);
        }
    });
};

/**
 * Sets the specified section as current.
 * @private
 */
oj.ListViewIndexerModel.prototype._setSection = function(section)
{
    var sections, index, self = this, match = null, groupHeader;

    sections = this.getIndexableSections();
    index = sections.indexOf(section);

    return new Promise(function(resolve, reject) 
    {
        if (index == -1)
        {
            // if it's not even in the indexable sections, then we don't need to process anymore
            resolve(null);
        }
        else
        {
            // try to find the group header, use the next section as needed
            while (index < sections.length)
            {
                section = sections[index];
                groupHeader = self._findGroupHeader(section);
                if (groupHeader != null)
                {
                    self.listview._scrollToGroupHeader(groupHeader);
                    match = section;
                    break;
                }

                index++;
            }

            resolve(match);
        }
    });
};

/**
 * Finds the group header with the specified section
 * @private
 */
oj.ListViewIndexerModel.prototype._findGroupHeader = function(section)
{
    var match, groupItems, content;

    this.listview._getGroupItemsCache().each(function(index) 
    {
        content = $(this).text();
        if (content.indexOf(section) == 0)
        {
            match = this;
            return false;
        }            
    });

    return match;
};
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

(function() {

/*!
 * JET Indexer @VERSION
 *
 *
 * Depends:
 *  jquery.ui.widget.js
 */

/**
 * @ojcomponent oj.ojIndexer
 * @augments oj.baseComponent
 * @since 1.2.0
 * @ojstatus preview
 * @ojtsignore
 * 
 * @ojshortdesc Displays a list of sections that corresponds to group headers of a list.
 * @ojrole slider
 * @classdesc
 * <h3 id="indexerOverview-section">
 *   JET Indexer Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#indexerOverview-section"></a>
 * </h3>
 * <p>Description: The JET Indexer is usually associated with a scrollable JET ListView.  It provides a list of sections that 
 *                 corresponds to group headers in ListView.  When a section is selected the corresponding group header will be 
 *                 scroll to the top of the ListView.
 * </p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-indexer
 *   aria-controls='listview1'
 *   data='{{data}}'>
 * &lt;/oj-indexer>
 * </code>
 * </pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 * <p>
 * The JET Indexer is accessible - it sets and maintains the appropriate aria- attributes, 
 * including <code class="prettyprint">aria-valuenow</code>, <code class="prettyprint">aria-valuemax</code>,
 * <code class="prettyprint">aria-valuemin</code> and <code class="prettyprint">aria-orientation</code>.
 * <p>
 * Application developer should associate a ListView with the Indexer by specifying the id of the ListView in the aria-controls attribute in the Indexer.
 * </p>
 */
oj.__registerWidget('oj.ojIndexer', $['oj']['baseComponent'],
{
    defaultElement: "<ul>",
    version: "1.2",
    widgetEventPrefix : "oj", 

    options: 
    {            
      /**
        * The data model for the Indexer which must be a oj.IndexerModel.  Currently this option can only be 
        * set to the return value of the getIndexerModel method on the ListView instance.
        * The data attribute should always be specified.  If not specified, then an empty indexer is rendered.
        *
        * @ojshortdesc The data provider for the Indexer.
        * @expose
        * @memberof! oj.ojIndexer
        * @instance
        * @type {oj.IndexerModel}
        * @default null
        *
        * @example <caption>Initialize the Indexer with an IndexModel:</caption>
        * &lt;oj-indexer data='{{myIndexerModel}}'>&lt;/oj-indexer>
        *
        * @example <caption>Get or set the <code class="prettyprint">data</code> property after initialization:</caption>
        * // getter
        * var dataValue = myIndexer.data;
        *
        * // setter
        * myIndexer.data = myIndexerModel;
        */
        data: null
    },

    /**
     * Creates the indexer
     * @override
     * @memberof! oj.ojIndexer
     * @protected
     */
    _ComponentCreate : function () 
    {
        this._super();
        this._setup();
    },

    /**
     * Initialize the indexer after creation
     * @protected
     * @override
     * @memberof! oj.ojIndexer
     */
    _AfterCreate : function ()
    {
        this._super();
        this._createIndexerContent();
        this._setAriaProperties();
        this._createInstructionText();
    },

    /**
     * Destroy the indexer
     * @memberof! oj.ojIndexer
     * @override
     * @private
     */
    _destroy: function()
    {
        this._super();
        this._unsetAriaProperties();
        this.element.removeClass("oj-component-initnode");

        oj.DomUtils.unwrap(this.element, this._getIndexerContainer());
    },

    /**
     * Sets a single option
     * @memberof! oj.ojIndexer
     * @override
     * @private
     */
    _setOption: function(key, value)
    {
        this._superApply(arguments);
        if (key == "data")
        {
            this.refresh();
        }
    },

    /**
     * Sets up resources needed by indexer
     * @memberof! oj.ojIndexer
     * @instance
     * @override
     * @protected
     */
    _SetupResources: function()
    {
        this._super();

        // register a resize listener and swipe handler        
        var container = this._getIndexerContainer()[0];
        this._registerResizeListener(container);
        this._registerTouchHandler(container);
    },

    /**
     * Release resources held by indexer
     * @memberof! oj.ojIndexer
     * @instance
     * @override
     * @protected
     */
    _ReleaseResources: function()
    {
        this._super();

        var container = this._getIndexerContainer()[0];
        this._unregisterResizeListener(container);
        this._unregisterTouchHandler(container);

        // if there's outstanding busy state, release it now
        this._resolveBusyState();
    },

    /**
     * @private
     */
    _resolveBusyState: function()
    {
        if (this.busyStateResolve)
        {
            this.busyStateResolve(null);
            this.busyStateResolve = null;
        }
    },

    /**
     * Returns a jQuery object containing the root dom element of the indexer
     * @ignore
     * @expose
     * @override
     * @memberof! oj.ojIndexer
     * @instance
     * @return {jQuery} the root DOM element of the indexer
     */
    widget: function() 
    {
        return this._getIndexerContainer();
    },

    /**
     * Redraw the entire indexer after having made some external modification.
     *
     * <p>This method does not accept any arguments.
     *
     * @ojshortdesc Redraw the entire indexer.
     * @expose
     * @memberof! oj.ojIndexer
     * @instance
     * @return {void}
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * $( ".selector" ).ojIndexer( "refresh" );
     */
    refresh: function()
    {
        this._super();

        this.element.empty();
        this._createIndexerContent();
        this._setAriaProperties();
        this.m_current = null;
    },

    //** @inheritdoc */
    getNodeBySubId: function(locator)
    {
        var subId, section, sections, i, j, node, data, includes;
 
        if (locator == null)
        {
            return this.element ? this.element[0] : null;
        }

        subId = locator['subId'];
        if (subId === 'oj-indexer-section')
        {
            section = locator['section'];
            sections = this.element.children("li");
            for (i=0; i<sections.length; i++)
            {
                node = sections.get(i);
                data = $(node).data("data-range");
                if (data == section)
                {
                    return node;
                }
                else
                {
                    // it's a separator, check the sections included in the range
                    includes = $(node).data("data-includes");
                    if (includes != null)
                    {
                        for (j=0; j<includes.length; j++)
                        {
                            if (includes[j] == section)
                            {
                                return node;
                            }
                        }
                    }
                }
            }
        }
        // Non-null locators have to be handled by the component subclasses
        return null;
    },

    //** @inheritdoc */
    getSubIdByNode: function(node)
    {
        var section;

        if (node != null)
        {   
            section = $(node).data("data-range");
            if (section != null)
            {
                return {'subId': 'oj-indexer-section', 'section': section};
            }
        }

        return null;
    },

    /****************************** core rendering **********************************/
    /**
     * Sets wai-aria properties on root element
     * @private
     */
    _setAriaProperties: function()
    {
        this.element.attr("role", "slider")
                    .attr("aria-orientation", "vertical")
                    .attr("aria-describedby", this.element.prop("id") + ":desc")
                    .attr("aria-valuemin", 0)
                    .attr("aria-valuemax", Math.max(0, this.element.children().length-1));
    },

    /**
     * Removes wai-aria properties on root element
     * @private
     */
    _unsetAriaProperties: function()
    {
        this.element.removeAttr("role")
                    .removeAttr("aria-orientation")
                    .removeAttr("aria-describedby")
                    .removeAttr("aria-valuemin")
                    .removeAttr("aria-valuemax")
                    .removeAttr("aria-valuetext");
    },

    /**
     * Create instruction text for screen reader
     * @private
     */
    _createInstructionText: function()
    {
        var key, text;

        if (oj.DomUtils.isTouchSupported())
        {
            key = "ariaTouchInstructionText"
        }
        else
        {
            key = "ariaKeyboardInstructionText"
        }

        text = $(document.createElement("div"));
        text.prop("id", this.element.prop("id") + ":desc")
        text.addClass("oj-helper-hidden-accessible")
            .text(this.getTranslatedString(key));

        this._getIndexerContainer().append(text); // @HtmlUpdateOK
    },

    /**
     * Retrieves the div around the root element, create one if needed.
     * @return {jQuery} the div around the root element 
     * @private
     */    
    _getIndexerContainer: function()
    {
        if (this.m_container == null)
        {
            this.m_container = this._createIndexerContainer();
        }
        return this.m_container;
    },

    /**
     * Creates the div around the root element.
     * @return {jQuery} the div around the root element
     * @private
     */    
    _createIndexerContainer: function()
    {
        var container;
        if (this.OuterWrapper) 
        {
            container = $(this.OuterWrapper);
        } 
        else 
        {
            container = $(document.createElement('div'));
            this.element.parent()[0].replaceChild(container[0], this.element[0]); //@HTMLUpdateOK
        }
        container.addClass("oj-indexer oj-component");
        container.prepend(this.element); //@HTMLUpdateOK

        return container;
    },

    /**
     * @private
     */
    _createIndexerContent: function()
    {
        var model, root, sections, missingSections, sectionOthers, height, first, itemHeight, max, skip, i, section, item, last, others;

        model = this._getIndexerModel();
        if (model == null)
        {
            return;
        }

        root = this.element;
        sections = model.getIndexableSections();
        if (model.getMissingSections)
        {
            missingSections = model.getMissingSections();
        }
        sectionOthers = this.getTranslatedString("indexerOthers");
        height = this.widget().outerHeight();
 
        // the first character is always present, use it to test height
        first = this._createItem(sections[0], missingSections);
        root.append(first); // @HtmlUpdateOK

        itemHeight = first.outerHeight();
        max = Math.floor(height / itemHeight);

        // first +1 is to include the '#', second +1 is to include rendering of the symbol between letters
        this._getIndexerContainer().removeClass("oj-indexer-abbr");
        skip = Math.floor((sections.length+1) / max) + 1;
        if (skip > 1)
        {
            // the height of item is a little different
            this._getIndexerContainer().addClass("oj-indexer-abbr");
        }

        for (i=1+skip; i<sections.length; i=i+skip+1)
        {
            if (skip > 1)
            {
                item = this._createSeparator(sections, i-skip, i-1);
                root.append(item); // @HtmlUpdateOK
            }
            else
            {
                i--;
            }

            section = sections[i];
            item = this._createItem(section, missingSections);
            root.append(item); // @HtmlUpdateOK
        }

        // the last character is always present
        last = this._createItem(sections[sections.length-1], missingSections);
        root.append(last); // @HtmlUpdateOK

        // the special others character is always present
        others = this._createItem(sectionOthers);
        others.attr("data-others", "true");
        root.append(others); // @HtmlUpdateOK
    },

    /**
     * @private
     */
    _createItem: function(section, missingSections)
    {
        var label, item;

        label = section['label'] ? section['label'] : section;
     
        item = $(document.createElement("li"));
        item.data("data-range", section)
            .text(label);

        if (missingSections != null && missingSections.indexOf(section) > -1)
        {
            item.addClass("oj-disabled");
        }            

        return item;
    },

    /**
     * @private
     */
    _createSeparator: function(sections, from, to)
    {
        var item, i, includes = [];

        item = $(document.createElement("li"));
        item.addClass("oj-indexer-ellipsis")
            .data("data-range", sections[from + Math.round((to-from)/2)]);
        for (i=from; i<=to; i++)
        {
            includes.push(sections[i]);
        }
        item.data("data-includes", includes);
        return item;
    },
    /**************************** end core rendering **********************************/

    /******************************** event handler **********************************/
    /**
     * Initialize the indexer
     * @private
     */
    _setup: function()
    {
        var self = this, model;

        this.element
            .uniqueId()
            .addClass("oj-component-initnode")
            .attr("tabIndex", 0);

        this._on(this.element, {
            "click": function(event) 
            {
                self._handleClick(event);
            },
            "keydown": function(event)
            {
                self._handleKeyDown(event);
            },
            "focus": function(event)
            {
                self._handleFocus(event);
            },
            "blur": function(event)
            {
                self._handleBlur(event);
            }
        });

        this._focusable({
            'applyHighlight': true,
            'setupHandlers': function( focusInHandler, focusOutHandler) {
                self._focusInHandler = focusInHandler;
                self._focusOutHandler = focusOutHandler;
            }
        });
    },

    /**
     * Event handler for when mouse click anywhere in the indexer
     * @param {Event} event mouseclick event
     * @private
     */
    _handleClick: function(event)
    {
        var target;

        //only perform events on left mouse, (right in rtl culture)
        if (event.button === 0)
        {
            target = $(event.target);
            this._setCurrent(target);
        }
    },

    /**
     * Handler for focus event
     * @param {Event} event the focus event
     * @private
     */
    _handleFocus: function(event)
    {
        this._getIndexerContainer().addClass("oj-focus-ancestor");
        if (this.m_current == null)
        {
            this._setFocus(this.element.children("li").first());
        }
        else
        {
            this._setFocus(this.m_current);
        }
    },

    /**
     * Handler for blur event
     * @param {Event} event the blur event
     * @private
     */
    _handleBlur: function(event)
    {
        this._getIndexerContainer().removeClass("oj-focus-ancestor");
    },

    /**
     * Event handler for when user press down a key
     * @param {Event} event keydown event
     * @private
     */
    _handleKeyDown: function(event)
    {
        var next, processed = false;

        switch (event.keyCode)
        {
            // UP key
            case 38:
                next = this.m_current.prev();
                break;
            // DOWN key
            case 40:
                next = this.m_current.next();
                break;
            // ENTER key
            case 13:
                this._setCurrent(this.m_current);
                processed = true;
                break;
        };

        if (next != null && next.length > 0)
        {
            processed = true;
            this._setFocus(next);
        }   

        if (processed)
        {
            event.preventDefault();
        }
    },

    _setFocus: function(item)
    {
        if (this.m_current != null)
        {
            this._focusOutHandler(this.m_current);
        }
        this._focusInHandler(item);

        this._updateAriaProperties(item);        
        this.m_current = item;
    },

    /**
     * Retrieves the indexer model.
     * @private
     */
    _getIndexerModel: function()
    {
        var model = this.option("data");
        if (model != null && (model.setSection == undefined || model.getIndexableSections == undefined))
        {
            throw "Invalid IndexerModel";
        }
        return model;
    },

    /**
     * Sets the character item as current
     * @param {jQuery} item
     * @private
     */
    _setCurrent: function(item)
    {
        var section = item.data("data-range");
        if (item.attr("data-others"))
        {
            section = oj.IndexerModel.SECTION_OTHERS;
        }    

        this._setCurrentSection(section);
    },

    /**
     * Sets the section as current
     * @param {Object} section
     * @private
     */
    _setCurrentSection: function(section)
    {
        var self = this, busyContext, promise, item, val;

        busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
        this.busyStateResolve = busyContext.addBusyState({'description': 'setCurrentSection'});

        // sets on the IndexerModel
        promise = /** @type {Promise} */ (this._getIndexerModel().setSection(section));
        promise.then(function(val)
        {
            // the resolve value is the section that actually scrolls to
            if (val != null)
            {               
                item = self._findItem(val);
                if (item != null)
                {
                    self._setFocus(item);
                }
            }

            self._resolveBusyState();
        }, function(e)
        {
            self._resolveBusyState();
        });
    },

    /**
     * Update wai-aria properties 
     * @param {jQuery} item the item 
     * @private
     */
    _updateAriaProperties: function(item)
    {
        var includes, val, first, second, valueText = "";

        includes = item.data("data-includes");

        if (includes != null)
        {
            // length should always be > 0
            if (includes.length > 0)
            {
                first = includes[0]['label'] ? includes[0]['label'] : includes[0];
                second = includes[includes.length-1]['label'] ? includes[includes.length-1]['label'] : includes[includes.length-1];
                valueText = this.getTranslatedString("ariaInBetweenText", {"first": first, "second": second});
            }
        }
        else
        {
            val = item.data("data-range");
            // checks if it's the special others section
            if (val === oj.IndexerModel.SECTION_OTHERS)
            {
                valueText = this.getTranslatedString("ariaOthersLabel");
            }
            else
            {
                valueText = val;
            }
        }

        // convey to screen reader that it's disabled
        if (item.hasClass("oj-disabled"))
        {
            valueText = valueText + ". " + this.getTranslatedString("ariaDisabledLabel");
        }

        this.element.attr("aria-valuetext", valueText);
        this.element.attr("aria-valuenow", item.index());
    },

    /**
     * Finds the item with the specified section
     * @param {Object} section
     * @return {jQuery} the item, null if not found
     * @private
     */
    _findItem: function(section)
    {
        var children, i, item, value, includes;

        children = this.element.children();
        for (i=0; i<children.length; i++)
        {
            item = children.get(i);
            value = $(item).data("data-range");
            includes = $(item).data("data-includes");

            if ((value != null && value == section) || (includes != null && includes.indexOf(section) > -1))
            {
                return $(item);
            }
        }

        return null;
    },

    /**
     * Unregister event listeners for resize the container DOM element.
     * @param {Element} element the container DOM element
     * @private
     */
    _unregisterResizeListener: function(element)
    {
        if (element && this._resizeHandler)
        {
            // remove existing listener
            oj.DomUtils.removeResizeListener(element, this._resizeHandler);    
        }
    },

    /**
     * Register event listeners for resize the container DOM element.
     * @param {Element} element  DOM element
     * @private
     */
    _registerResizeListener: function(element)
    {         
        if (element)
        {
            if (this._resizeHandler == null)
            {
                this._resizeHandler = this._handleResize.bind(this);
            }

            oj.DomUtils.addResizeListener(element, this._resizeHandler);
        }
    },

    /**
     * Unregister panning handler
     * @param {Element} element  DOM element
     * @private
     */
    _unregisterTouchHandler: function(element)
    {
        if (this.hammer)
        {
            this.hammer.off("panstart panmove panend");
            $(element).ojHammer("destroy");
        }
        this.hammer = null;
    },

    /**
     * Register panning handler
     * @param {Element} element  DOM element
     * @private
     */
    _registerTouchHandler: function(element)
    {    
        var self = this, options, target, x, y, currentTarget, currentSection, currentY, previousY, delta, range, index, section;
        
        options = {
        "recognizers": [
          [Hammer.Pan, {"direction": Hammer["DIRECTION_VERTICAL"]}
          ]
        ]};

        this.hammer = $(element)
        .ojHammer(options)
        .on("panstart", function(event)
        {
            target = event['gesture']['target'];
            // for x, don't use the target, use x relative to the indexer to ensure it reflects item in Indexer 
            // even if the finger is off the Indexer
            x = self.element[0].getBoundingClientRect().left + 5;
            y = target.getBoundingClientRect().top;
            self._setCurrent($(target));

            currentTarget = target;
            currentSection = $(target).data("data-range");
            currentY = y;
        })
        .on("panmove", function(event)
        {
            // calculate point instead of using screenX/Y from touch is better since
            // 1) x stays constant
            // 2) in voiceover user could have pan anywhere on the screen
            previousY = currentY;
            currentY = y + event['gesture']['deltaY'];

            target = document.elementFromPoint(x, currentY);

            // should not happen
            if (target == null)
            {
                return;
            }

            delta = currentY - previousY;

            if (currentTarget == target)
            {
                range = $(target).data("data-includes");
                // if the section is a range (dot), then try to set the next section inside the range current
                // for example, if move on * which represents range BCD, if current is C then move up should go to B and move down should go to D
                if (range != null)
                {
                    index = range.indexOf(currentSection);
                    section = null;
                    if (delta > 0 && index < range.length-1)
                    {
                        section = range[index+1];
                    }
                    else if (delta < 0 && index > 0)
                    {
                        section = range[index-1];
                    }

                    if (section != null)
                    {
                        currentSection = section;
                        self._setCurrentSection(section);                           
                    }
                }
            }
            else
            {
                if ($(target).data("data-range"))
                {                 
                    range = $(target).data("data-includes");
                    section = null;
                    // if the target is a range (dot), check to see if we should set the section to the beginning of
                    // range or end of range.  For example, if you have A * E with * represents BCD, coming from A should go to B
                    // where as coming from E should go to D.
                    if (range != null)
                    {
                        if (delta > 0 && target == currentTarget.nextElementSibling)
                        {
                            section = range[0];
                        }
                        else if (delta < 0 && target == currentTarget.previousElementSibling)
                        {
                            section = range[range.length-1];
                        }
                    }         

                    if (section == null)
                    {
                        section = $(target).data("data-range");
                    }

                    currentTarget = target;
                    currentSection = section;

                    self._setCurrentSection(currentSection);                           
                }
            }
        })
        .on("panend", function(event)
        {
            currentTarget = null;
            currentSection = null;
            currentY = null;
            section = null;
        });
    },

    /**
     * The resize handler.
     * @param {number} width the new width
     * @param {number} height the new height
     * @private
     */
    _handleResize: function(width, height)
    {
        if (width > 0 && height > 0)
        {
            this.refresh();
        }
    }
        
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Characters</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Selects the character in the indexer, which scrolls to the corresponding group header in the associated ListView.  
 *           When tap on the ellipsis character, the character in between will be selected.</td>
 *     </tr>
 *     <tr>
 *       <td>Characters</td>
 *       <td><kbd>Pan</kbd></td>
 *       <td>Selects the character in the indexer, which scrolls to the corresponding group header in the associated ListView.  
 *           When pan up and down the ellipsis character, the indexer will select the range of characters represented by the ellipsis.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojIndexer
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan = "3" nowrap>Section</td>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus to the section below.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus to the section above.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Selects the current section.  No op if the section is already selected.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojIndexer
 */

//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the sections within the Indexer.  See the <a href="#getNodeBySubId">getNodeBySubId</a>
 * method for details.</p>
 *
 * @ojsubid oj-indexer-section
 * @memberof oj.ojIndexer
 *
 * @example <caption>Get the node that represents the specified prefix 'A' in the indexer:</caption>
 * var node = myIndexer.getNodeBySubId({'subId': 'oj-indexer-section', 'section': 'A'});
 */

    });

}() );
(function() {
var ojIndexerMeta = {
  "properties": {
    "data": {},
    "translations": {
      "type": "Object",
      "properties": {
        "ariaDisabledLabel": {
          "type": "string"
        },
        "ariaInBetweenText": {
          "type": "string"
        },
        "ariaKeyboardInstructionText": {
          "type": "string"
        },
        "ariaOthersLabel": {
          "type": "string"
        },
        "ariaTouchInstructionText": {
          "type": "string"
        },
        "indexerCharacters": {
          "type": "string"
        },
        "indexerOthers": {
          "type": "string"
        }
      }
    }
  },
  "extension": {
    _INNER_ELEM: 'ul',
    _WIDGET_NAME: "ojIndexer"
  }
};
oj.CustomElementBridge.registerMetadata('oj-indexer', 'baseComponent', ojIndexerMeta);
oj.CustomElementBridge.register('oj-indexer', {'metadata': oj.CustomElementBridge.getMetadata('oj-indexer')});
})();
/**
 * Base class for TableDataSourceContentHandler and TreeDataSourceContentHandler
 * Handler for DataSource generated content
 * @constructor
 * @ignore
 */
oj.DataSourceContentHandler = function(widget, root, data)
{
    this.m_root = root;
    this.m_widget = widget;

    this.m_fetching = false;

    this.setDataSource(data);
    this.Init();
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.DataSourceContentHandler, oj.Object, "oj.DataSourceContentHandler");

/**
 * Initializes the instance.
 * @export
 */
oj.DataSourceContentHandler.prototype.Init = function()
{
    oj.DataSourceContentHandler.superclass.Init.call(this);
};

/**
 * Handles when the container has resize
 * @param {number} width the new width
 * @param {number} height the new height
 * @protected
 */
oj.DataSourceContentHandler.prototype.HandleResize = function(width, height)
{
    // by default do nothing, to be override by subclass
};

/**
 * Handles when the listview is shown due to for example CSS changes (inside a dialog)
 */
oj.DataSourceContentHandler.prototype.notifyShown = function()
{
    // by default do nothing, to be override by subclass
};

/**
 * Handles when the listview is re-attached to the DOM (ex: when a children of CCA gets re-attached from slotting)
 */
oj.DataSourceContentHandler.prototype.notifyAttached = function()
{
    // by default do nothing, to be override by subclass
};

/**
 * Destroy the content handler
 * @protected
 */
oj.DataSourceContentHandler.prototype.Destroy = function()
{
    // this.m_root was changed in RenderContent
    if (this.m_superRoot != null)
    {
        this.m_root = this.m_superRoot;
    }

    $(this.m_root).empty(); // @HTMLUpdateOK
    this.m_widget = null;
    this.m_root = null;
    this.m_superRoot = null;
};

/**
 * Determines whether the content handler is in a ready state
 * @return {boolean} true if there's no outstanding fetch, false otherwise.
 * @protected
 */
oj.DataSourceContentHandler.prototype.IsReady = function()
{
    return !this.m_fetching;
};

/**
 * @private
 */
oj.DataSourceContentHandler.prototype.setRootAriaProperties = function()
{
    if (this.shouldUseGridRole())
    {
        this.m_root.setAttribute("role", "grid");
    }
    else
    {
        if (this.IsHierarchical())
        {
            this.m_root.setAttribute("role", "tree");
        }
        else
        {
            this.m_root.setAttribute("role", "listbox");
        }
    }
};

/**
 * Renders the content inside the list
 * @protected
 */
oj.DataSourceContentHandler.prototype.RenderContent = function()
{
    var presentation, row;

    this.signalTaskStart("rendering content"); // signal method task start

    this.setRootAriaProperties();
    if (this.shouldUseGridRole() && this.isCardLayout() && !this.IsHierarchical())
    {
        // in card layout, this is going to be a single row, N columns grid
        // so we'll need to wrap all <li> within a row
        presentation = document.createElement("li");
        row = document.createElement("ul");
        presentation.appendChild(row); //@HTMLUpdateOK 
        $(presentation).attr("role", "presentation")
                       .css("width", "100%");
        $(row).attr("role", "row")
              .addClass(this.m_widget.getGroupStyleClass());

        this.m_root.appendChild(presentation); //@HTMLUpdateOK
        this.m_superRoot = this.m_root;
        this.m_root = row;
    }
    this.fetchRows(false);
    this.signalTaskEnd(); // signal method task end
};

/**
 * Retrieve the key given the item element
 * @param {Element} element
 * @return {Object|null}
 * @protected
 */
oj.DataSourceContentHandler.prototype.GetKey = function(element)
{
    // should be in the element
    return element.key;
};

oj.DataSourceContentHandler.prototype.FindElementByKey = function(key)
{
    var children, i, elem;

    children = $(this.m_root).find("."+this.m_widget.getItemElementStyleClass());
    for (i=0; i<children.length; i++)
    {
        elem = children[i];
        // use == for the string number compare case
        if (key == this.GetKey(elem) || oj.Object.compareValues(key, this.GetKey(elem)))
        {
            return elem;
        }
    }

    return null;
};

oj.DataSourceContentHandler.prototype.getDataSource = function()
{
    return this.m_dataSource;
};

/**
 * @protected
 */
oj.DataSourceContentHandler.prototype.setDataSource = function(dataSource)
{
    this.m_dataSource = dataSource;
};

/**
 * Initiate loading of the template engine.  An error is thrown if the template engine failed to load.
 * @return {Promise} resolves to the template engine, or null if:
 *                   1) there's no need because no item template is specified 
 *                   2) a renderer is present which takes precedence
 * @protected
 */
oj.DataSourceContentHandler.prototype.loadTemplateEngine = function()
{
    var self = this;

    if (this.m_widget.getItemTemplate() != null && this.m_widget._getItemRenderer() == null)
    {
        return new Promise(function(resolve, reject)
        {
            oj.Config.__getTemplateEngine().then(
                function(engine)
                {
                    self.m_engine = engine;
                    resolve(engine);
                },
                function(reason)
                {
                    throw "Error loading template engine: "+reason;
                }
            );
        });
    }

    return Promise.resolve(null);
};

/**
 * Retrieve the template engine, returns null if it has not been loaded yet
 */
oj.DataSourceContentHandler.prototype.getTemplateEngine = function()
{
    return this.m_engine;
};

oj.DataSourceContentHandler.prototype.fetchRows = function(forceFetch)
{
    this.m_widget.showStatusText();
};

/**
 * Create a list item and add it to the list
 * @param {Element|DocumentFragment} parentElement the element to add the list items to
 * @param {number} index the index of the item
 * @param {Object|null} data the data for the item
 * @param {Object} metadata the set of metadata for the item
 * @param {Object} templateEngine the template engine to process inline template
 * @param {function(Element, Object)=} callback optional callback function to invoke after item is added 
 * @return {Object} contains the list item and the context object
 * @protected
 */
oj.DataSourceContentHandler.prototype.addItem = function(parentElement, index, data, metadata, templateEngine, callback)
{
    var item, referenceNode, childElements, position;

    item = document.createElement("li");
    $(item).uniqueId();
    if (index === -1)
    {
        referenceNode = null;
    }
    else
    {
        childElements = $(parentElement).children('.'+this.m_widget.getItemElementStyleClass()+', .'+this.m_widget.getEmptyTextStyleClass()+', .oj-listview-temp-item');
        referenceNode = (index === childElements.length) ? null : childElements[index];
    }
    this.m_widget.BeforeInsertItem();
    parentElement.insertBefore(item, referenceNode); // @HTMLUpdateOK
    position = $(parentElement).children().index(item);
    return this._addOrReplaceItem(item, position, parentElement, index, data, metadata, templateEngine, callback);
};

/**
 * Replace an existing list item in the list
 * @param {Element} item the list item to change
 * @param {number} index the index of the item
 * @param {Object|null} data the data for the item
 * @param {Object} metadata the set of metadata for the item
 * @param {Object} templateEngine the template engine to process inline template
 * @param {function(Element, Object)=} callback optional callback function to invoke after item is added 
 * @protected
 */
oj.DataSourceContentHandler.prototype.replaceItem = function(item, index, data, metadata, templateEngine, callback)
{
    var parentElement, position, newItem;

    // animate hiding of existing item first
    this.signalTaskStart("replace item"); // signal replace item animation start. Ends in _handleReplaceTransitionEnd() defined in TableDataSourceContentHandler

    // now actually replace the item
    parentElement = item.parentNode;
    position = $(parentElement).children().index(item);
    newItem = document.createElement("li");
    // this should trigger ko.cleanNode if applicable
    $(item).replaceWith(newItem); //@HTMLUpdateOK; newItem is constructed by the component and not yet manipulated by the application

    this._addOrReplaceItem(newItem, position, parentElement, index, data, metadata, templateEngine, callback);    
};

/**
 * Handles both add and replace item
 * @private
 */
oj.DataSourceContentHandler.prototype._addOrReplaceItem = function(item, position, parentElement, index, data, metadata, templateEngine, callback)
{
    var contentContainer, context, inlineStyle, styleClass, renderer, templateElement, content, textWrapper, componentElement, bindingContext, nodes;

    if (callback == undefined)
    {
        callback = this.afterRenderItem.bind(this);
    }

    context = this.createContext(index, data, metadata, item);
    renderer = this.m_widget._getItemRenderer();
    templateElement = this.m_widget.getItemTemplate();
    if (renderer != null)
    {
        // if an element is returned from the renderer and the parent of that element is null, we will append 
        // the returned element to the parentElement.  If non-null, we won't do anything, assuming that the 
        // rendered content has already added into the DOM somewhere.
        content = renderer.call(this, context);
        if (content != null)
        {
            // allow return of document fragment from jquery create/js document.createDocumentFragment
            if (content['parentNode'] === null || content['parentNode'] instanceof DocumentFragment)
            {
                item.appendChild(content); // @HTMLUpdateOK
            }
            else if (content['parentNode'] != null)
            {
                // parent node exists, do nothing
            }                
            else if (content.toString)
            {
                textWrapper = document.createElement("span");
                textWrapper.appendChild(document.createTextNode(content.toString())); // @HTMLUpdateOK
                item.appendChild(textWrapper); // @HTMLUpdateOK
            }
        }
    }
    else if (templateElement != null && templateEngine != null)
    {
        componentElement = this.m_widget.getRootElement()[0];
        bindingContext = this.GetBindingContext(context);
        nodes = templateEngine.execute(componentElement, templateElement, bindingContext);
        nodes.forEach(
            function(node)
            {
                item.appendChild(node);
            }
        );
    }
    else
    {
        textWrapper = document.createElement("span");
        textWrapper.appendChild(document.createTextNode(data == null ? "" : data.toString())); // @HTMLUpdateOK
        item.appendChild(textWrapper); // @HTMLUpdateOK
    }

    // get the item from root again as template replaces the item element
    item = parentElement.children ? parentElement.children[position] : this._getItemFromDocumentFragment(parentElement, position);
    context['parentElement'] = item;

    // cache data in item element, this is needed for getDataForVisibleItem.
    $.data(item, "data", data);

    // do any post processing
    callback(item, context);

    return {item: item, context: context};
};

/**
 * In IE/Safari, DocumentFragment does not support children property
 * @private
 */
oj.DataSourceContentHandler.prototype._getItemFromDocumentFragment = function(fragment, index)
{
    var nodes, node, i = 0, nodeIndex = 0;

    nodes = fragment.childNodes;
    while (node = nodes[i++]) 
    {
        if (node.nodeType === 1) 
        {
            if (nodeIndex === index)
            {
        	return node;
            }
            nodeIndex++;
        }
    }	
    return null;
}

/**
 * Creates a binding context based on context object
 * To be override by different ContentHandler
 * @protected
 */
oj.DataSourceContentHandler.prototype.GetBindingContext = function(context)
{
    var current, as, ret;

    current = {};
    current['data'] = context['data'];
    current['index'] = context['index'];
    current['key'] = context['key'];
    current['componentElement'] = context['componentElement'];

    ret = {'$current': current};

    as = this.m_widget.GetOption("as");
    if (as != null && as.length > 0)
    {
        ret[as] = current;
    }

    return ret;
};

oj.DataSourceContentHandler.prototype.afterRenderItem = function(item, context)
{
    var elem;

    // save the key in the element (cannot use data- here since it could be a non-string)
    item.key = context['key'];

    item = $(item);
    item.uniqueId();

    // if there's only one element inside the item and it is focusable, set
    // the role on it instead
    elem = this.m_widget.getSingleFocusableElement(item);

    if (this.shouldUseGridRole())
    {
        if (context['leaf'] != undefined && !context['leaf'])
        {
            // it's a group item
            item.attr("role", "presentation");
        }
        else
        {
            if (this.isCardLayout())
            {
                elem.attr("role", "gridcell");
            }
            else
            {
                item.attr("role", "row");
                if (elem != item)
                {
                    elem.attr("role", "gridcell");
                }
                else 
                {
                    // we'll need to wrap content with a gridcell role
                    elem.children().wrapAll("<div role='gridcell' class='oj-listview-cell-element'></div>"); //@HTMLUpdateOK
                }
            }
        }
    }
    else
    {
        elem.attr("role", this.IsHierarchical() ? "treeitem" : "option");
        if (elem != item)
        {
            item.attr("role", "presentation");
        }
    }

    elem.addClass(this.m_widget.getFocusedElementStyleClass());

    // tag it if item is not focusable
    if (!this.isFocusable(context))
    {
        item.addClass("oj-skipfocus");
    }

    item.addClass(this.m_widget.getItemElementStyleClass());
};

oj.DataSourceContentHandler.prototype.createContext = function(index, data, metadata, elem)
{
    var context, prop;
    
    context = {
    };
    context['parentElement'] = elem;
    context['index'] = index;
    context['data'] = data;
    context['component'] = this.m_widget.getWidgetConstructor();
    context['datasource'] = this.getDataSource();
    context = this.m_widget._FixRendererContext(context);

    // merge properties from metadata into cell context
    // the properties in metadata would have precedence
    for (prop in metadata) 
    {
        if (metadata.hasOwnProperty(prop)) 
        {
            context[prop] = metadata[prop];   
        }
    }

    return context;
};

oj.DataSourceContentHandler.prototype.isFocusable = function(context)
{
    return this.m_widget.getItemFocusable(context);
};

oj.DataSourceContentHandler.prototype.isSelectable = function(context)
{
    return this.m_widget.getItemSelectable(context);
};

oj.DataSourceContentHandler.prototype.isCardLayout = function()
{
    return this.m_widget.isCardLayout();
};

oj.DataSourceContentHandler.prototype.shouldUseGridRole = function()
{
    return this.m_widget.ShouldUseGridRole();
};

oj.DataSourceContentHandler.prototype.isAsyncRendering = function()
{
    return false;
//    return this.m_widget.ojContext._IsCustomElement() && this.shouldUseGridRole();
};

oj.DataSourceContentHandler.prototype.signalTaskStart = function(description)
{
    if (this.m_widget) // check that widget exists (e.g. not destroyed)
    {
        this.m_widget.signalTaskStart("DataSource ContentHandler "+description);
    }
};

oj.DataSourceContentHandler.prototype.signalTaskEnd = function()
{
    if (this.m_widget) // check that widget exists (e.g. not destroyed)
    {
        this.m_widget.signalTaskEnd();
    }
};
/**
 * Handler for TreeDataSource generated content
 * @constructor
 * @extends oj.DataSourceContentHandler
 * @ignore
 */
oj.TreeDataSourceContentHandler = function(widget, root, data)
{
    oj.TreeDataSourceContentHandler.superclass.constructor.call(this, widget, root, data);
};

// Subclass from oj.DataSourceContentHandler 
oj.Object.createSubclass(oj.TreeDataSourceContentHandler, oj.DataSourceContentHandler, "oj.TreeDataSourceContentHandler");

/**
 * Initializes the instance.
 * @protected
 */
oj.TreeDataSourceContentHandler.prototype.Init = function()
{
  oj.TreeDataSourceContentHandler.superclass.Init.call(this);
};

/**
 * Determines whether the conent is hierarchical.
 * @return {boolean} returns true if content is hierarhical, false otherwise.
 * @protected
 */
oj.TreeDataSourceContentHandler.prototype.IsHierarchical = function()
{
    return true;
};

/**
 * @protected
 */
oj.TreeDataSourceContentHandler.prototype.fetchRows = function(forceFetch)
{
    this.signalTaskStart("fetching rows"); // signal method task start

    oj.TreeDataSourceContentHandler.superclass.fetchRows.call(this, forceFetch);

    this.fetchChildren(0, null, this.m_root, null);

    this.signalTaskEnd(); // signal method task end
};

oj.TreeDataSourceContentHandler.prototype.fetchChildren = function(start, parent, parentElem, successCallback)
{
    var promise, anchor, collapseClass, expandingClass, range, self = this;

    this.signalTaskStart("fetching children from index: "+start); // signal method task start

    // initiate loading of template engine, note it will not load it unless a template has been specified
    promise = this.loadTemplateEngine();

    // root node would not have expand/collapse icon
    if (parent != null)
    {
        anchor = parentElem.parentNode.firstElementChild.firstElementChild.firstElementChild;
        if (anchor)
        {
            anchor = $(anchor);
            collapseClass = this.m_widget.getCollapseIconStyleClass();
            // switch to loading icon
            if (anchor.hasClass(collapseClass))
            {
                expandingClass = this.m_widget.getExpandingIconStyleClass();
                anchor.removeClass(collapseClass)
                      .addClass(expandingClass);
            }
        }
    } 

    // no need to check ready since multiple fetch from different parents can occur at the same time
    this.m_fetching = true;

    range = {"start": start, "count": this.m_dataSource.getChildCount(parent)};

    this.signalTaskStart("first fetch"); 

    this.m_dataSource.fetchChildren(parent, range, {"success": function(nodeSet){
        promise.then(function(templateEngine)
        {
            self._handleFetchSuccess(nodeSet, parent, parentElem, successCallback, templateEngine);
            self.signalTaskEnd(); // first fetch
        })
    }, "error": function(status)
    {
        self._handleFetchError(status);
        self.signalTaskEnd(); // first fetch
    }});

    this.signalTaskEnd(); // signal method task end
};

oj.TreeDataSourceContentHandler.prototype._handleFetchSuccess = function(nodeSet, parent, parentElem, successCallback, templateEngine)
{
    var self = this, start, count, endIndex, fragment, i, data, metadata, gridcell, postProcessing, busyContext;

    // listview might have been destroyed before fetch success is handled
    if (this.m_widget == null)
    {
        return;
    }

    this.signalTaskStart("handling successful fetch"); // signal method task start

    start = nodeSet.getStart();
    count = nodeSet.getCount();
    endIndex = start+count;

    // walk the node set
    fragment = document.createDocumentFragment();
    for (i=0; i<count; i++)
    {
        data = nodeSet.getData(start+i);
        metadata = nodeSet.getMetadata(start+i);
        // pass -1 for opt since we know it will be inserted at the end of its parent
        this.addItem(fragment, -1, data, metadata, templateEngine);
    }
    parentElem.appendChild(fragment);

    // update aria-colspan on the gridcell representing the group header
    if (this.shouldUseGridRole() && this.isCardLayout() && parent != null && count > 1)
    {
        gridcell = parentElem.parentNode.firstElementChild.firstElementChild;
        $(gridcell).attr("aria-colspan", count);
    }

    // fetch is done
    this.m_fetching = false;

    postProcessing = function()
    {
        if (self.m_widget)
        {
            // if a callback is specified (as it is in the expand case), then invoke it
            if (successCallback != null)
            {
                successCallback.call(null, parentElem);
            }

            self.m_widget.renderComplete();            
        }
    };

    if (this.isAsyncRendering())
    {
        // custom elements renders async so this is needed.  
        // Also, since the root for non-custom element is <ul>, when application do a whenReady within the context
        // of <ul>, the postProcessing might be called after application's whenReady handler.
        busyContext = oj.Context.getContext(this.m_root).getBusyContext();
        busyContext.whenReady().then(function()
        {
            postProcessing();
        });        
    }
    else
    {
        postProcessing();
    }

    this.m_initialized = true;

    this.signalTaskEnd(); // signal method task end
};

/**
 * Creates a binding context based on context object
 * To be override by different ContentHandler
 * @protected
 */
oj.TreeDataSourceContentHandler.prototype.GetBindingContext = function(context)
{
    var bindingContext, current;

    bindingContext = oj.TreeDataSourceContentHandler.superclass.GetBindingContext.call(this, context);
    current = bindingContext['$current'];

    current['depth'] = context['depth'];
    current['leaf'] = context['leaf'];
    current['parentKey'] = context['parentKey'];

    return bindingContext;
};

oj.TreeDataSourceContentHandler.prototype.afterRenderItem = function(item, context)
{
    var groupStyleClass, itemStyleClass, groupItemStyleClass, groupCollapseStyleClass, focusedStyleClass,
        collapseClass, content, icon, groupItem;

    this.signalTaskStart("after rendering an item"); // signal method task start

    oj.TreeDataSourceContentHandler.superclass.afterRenderItem.call(this, item, context);

    groupStyleClass = this.m_widget.getGroupStyleClass();
    itemStyleClass = this.m_widget.getItemStyleClass();
    groupItemStyleClass = this.m_widget.getGroupItemStyleClass(true);
    groupCollapseStyleClass = this.m_widget.getGroupCollapseStyleClass();
    collapseClass = this.m_widget.getCollapseIconStyleClass();
    focusedStyleClass = this.m_widget.getFocusedElementStyleClass();

    item = $(item);

    if (context['leaf'] == false)
    {
        item.children().wrapAll("<div></div>"); //@HTMLUpdateOK

        // collapsed by default
        if(item.hasClass(focusedStyleClass))
        {
            item.removeClass(focusedStyleClass)
                .children().first()
                .addClass(focusedStyleClass)
                .attr("aria-expanded", "false");
        }
        else{
            item.children().first()
                    .attr("role","presentation")
                    .find("." + focusedStyleClass)
                    .attr("aria-expanded", "false");
        }

        content = item.children().first();
        content.uniqueId()
               .addClass(groupItemStyleClass);

        // add the expand icon
        if (this.m_widget.isExpandable())
        {
            item.addClass("oj-collapsed")

            icon = document.createElement("a");
            $(icon).attr("href", "#")
                   .attr("aria-labelledby", content.get(0).id)
                   .addClass("oj-component-icon oj-clickable-icon-nocontext")
                   .addClass(collapseClass);
        
            content.prepend(icon); //@HTMLUpdateOK               
        }

        if (this.shouldUseGridRole())
        {
            content.get(0).removeAttribute("aria-expanded");
            content.removeClass(focusedStyleClass);
            content.attr("role", "row");
            content.children().wrapAll("<div role='gridcell' aria-expanded='false' class='oj-listview-cell-element "+focusedStyleClass+"'></div>"); //@HTMLUpdateOK
        }

        // the yet to be expand group element
        groupItem = document.createElement("ul");
        $(groupItem).addClass(groupStyleClass)
                    .addClass(groupCollapseStyleClass)
                    .attr("role", this.shouldUseGridRole() ? (this.isCardLayout() ? "row" : "presentation") : "group");
        item.append(groupItem); //@HTMLUpdateOK
    }            
    else if (context['leaf'] == true)
    {
        item.addClass(itemStyleClass);
    }

    if (this.m_widget._isSelectionEnabled() && this.isSelectable(context))
    {
        this.m_widget.getFocusItem(item).attr("aria-selected", false);
    }

    // callback to widget
    this.m_widget.itemRenderComplete(item[0], context);

    this.signalTaskEnd(); // signal method task end
};

oj.TreeDataSourceContentHandler.prototype._handleFetchError = function(status)
{
    // listview might have been destroyed before fetch error is handled
    if (this.m_widget == null)
    {
        oj.Logger.info("handleFetchError: widget has already been destroyed");
        return;
    }

    this.signalTaskStart("handling fetch error: "+status); // signal method task start

    // TableDataSource aren't giving me any error message
    oj.Logger.error(status);

    this.m_widget.renderComplete();

    this.signalTaskEnd(); // signal method task end
};

oj.TreeDataSourceContentHandler.prototype.Expand = function(item, successCallback)
{
    var parentKey, parentElem;

    this.signalTaskStart("expanding an item"); // signal method task start

    parentKey = this.GetKey(item[0]);
    parentElem = item.children("ul")[0];
    this.fetchChildren(0, parentKey, parentElem, successCallback);

    this.signalTaskEnd(); // signal method task end
};

oj.TreeDataSourceContentHandler.prototype.Collapse = function(item)
{
    // remove all children nodes
    item.empty(); //@HTMLUpdateOK
};

/**
 * Handler for static HTML content
 * @constructor
 * @ignore
 */
oj.StaticContentHandler = function(widget, root)
{
    this.m_widget = widget;
    this.m_root = root;
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.StaticContentHandler, oj.Object, "oj.StaticContentHandler");

/**
 * Initializes the instance.
 * @protected
 */
oj.StaticContentHandler.prototype.Init = function()
{
  oj.StaticContentHandler.superclass.Init.call(this);
};

/**
 * Destroy the content handler
 * @protected
 */
oj.StaticContentHandler.prototype.Destroy = function()
{
    // check if it's been destroyed or in process
    if (!this.m_root.hasAttribute("role"))
    {
        return;
    }

    this.restoreContent(this.m_root, 0);
    this.unsetRootAriaProperties();

    if (this.shouldUseGridRole() && this.isCardLayout() && !this.IsHierarchical())
    {
        $(this.m_root).children().first().children().unwrap()
                      .children().unwrap();
    }    
};

/**
 * Determine whether the content handler is ready
 * @return {boolean} returns true there's no outstanding request, false otherwise.
 * @protected
 */
oj.StaticContentHandler.prototype.IsReady = function()
{
    // static content does not fetch
    return true;
};

oj.StaticContentHandler.prototype.HandleResize = function(width, height)
{
    // do nothing since all items are present
};

oj.StaticContentHandler.prototype.notifyShown = function()
{
    // do nothing since all items are present
};

oj.StaticContentHandler.prototype.notifyAttached = function()
{
    // do nothing since all items are present
};

oj.StaticContentHandler.prototype.RenderContent = function()
{
    var root = this.m_root;
    if (this.shouldUseGridRole() && this.isCardLayout() && !this.IsHierarchical() && $(root).children("li").length > 0)
    {
        // in card layout, this is going to be a single row, N columns grid
        // so we'll need to wrap all <li> within a row
        $(this.m_root).children().wrapAll("<li role='presentation' style='width:100%'><ul role='row' class='"+this.m_widget.getGroupStyleClass()+"'></ul></li>"); //@HTMLUpdateOK 
        root = $(this.m_root).children("li").first().children("ul").first().get(0);
    }
    this.modifyContent(root, 0);
    this.setRootAriaProperties();
    this.m_widget.renderComplete();
};

oj.StaticContentHandler.prototype.Expand = function(item, successCallback)
{
    var selector, groupItem;

    selector = "."+this.m_widget.getGroupStyleClass();
    groupItem = $(item).children(selector)[0];
    $(groupItem).css("display", "");

    successCallback.call(null, groupItem);
};

oj.StaticContentHandler.prototype.Collapse = function(item)
{
    // nothing to do
};

oj.StaticContentHandler.prototype.IsHierarchical = function()
{
    if (this.m_hier == null)
    {
        this.m_hier = $(this.m_root).children("li").children("ul").length > 0;
    }
    return this.m_hier;
};

/**
 * Restore the static content into its original format by removing all ListView specific style classes and attributes.
 * @param {Element} elem the element it is currently restoring
 * @param {number} depth the depth of the element it is currently restoring
 * @private
 */
oj.StaticContentHandler.prototype.restoreContent = function(elem, depth)
{
    var groupStyleClass, groupCollapseStyleClass, groupExpandStyleClass, groupItemStyleClass, itemStyleClass, itemElementStyleClass,
        items, i, item, groupItems, groupItem;

    groupStyleClass = this.m_widget.getGroupStyleClass();
    groupCollapseStyleClass = this.m_widget.getGroupCollapseStyleClass();
    groupExpandStyleClass = this.m_widget.getGroupExpandStyleClass();
    groupItemStyleClass = this.m_widget.getGroupItemStyleClass(true);
    itemStyleClass = this.m_widget.getItemStyleClass();
    itemElementStyleClass = this.m_widget.getItemElementStyleClass();

    items = elem.children;
    for (i=0; i<items.length; i++)
    {
        item = $(items[i]);
        // skip children that are not modified, this could happen if ko:foreach backed by an observable array is used to generate
        // the content, and the observable array has changed
        if (!item.hasClass(itemElementStyleClass))
        {
            continue;
        }

        this.unsetAriaProperties(item.get(0));
        item.removeClass(itemElementStyleClass)
            .removeClass(itemStyleClass)
            .removeClass(this.m_widget.getDepthStyleClass(depth))
            .removeClass("oj-skipfocus")
            .removeClass("oj-focus")
            .removeClass("oj-hover")
            .removeClass("oj-expanded")
            .removeClass("oj-collapsed")
            .removeClass("oj-selected");

        groupItems = item.children("ul");
        if (groupItems.length > 0)
        {
            item.children("."+groupItemStyleClass).children().unwrap();      
            if (this.shouldUseGridRole())
            {
                this.unsetGroupAriaProperties(item);
            }
            item.children(".oj-component-icon").remove();

            groupItem = $(groupItems[0]);
            groupItem.removeClass(groupStyleClass)
                     .removeClass(groupExpandStyleClass)
                     .removeClass(groupCollapseStyleClass)
                     .removeAttr("role");
            this.restoreContent(groupItem[0], depth+1);
        }
    }
};

/**
 * Modify the static content to include ListView specific style classes and attributes.
 * @param {Element} elem the element it is currently modifying
 * @param {number} depth the depth of the element it is currently modifying
 * @private
 */
oj.StaticContentHandler.prototype.modifyContent = function(elem, depth)
{
    var itemStyleClass, itemElementStyleClass, groupStyleClass, groupItemStyleClass, groupCollapseStyleClass, collapseClass, focusedElementStyleClass,
        items, expandable, i, item, context, groupItems, content, icon, groupItem, count;

    itemStyleClass = this.m_widget.getItemStyleClass();
    itemElementStyleClass = this.m_widget.getItemElementStyleClass();
    groupStyleClass = this.m_widget.getGroupStyleClass();
    groupItemStyleClass = this.m_widget.getGroupItemStyleClass(true);
    groupCollapseStyleClass = this.m_widget.getGroupCollapseStyleClass();
    collapseClass = this.m_widget.getCollapseIconStyleClass();
    focusedElementStyleClass = this.m_widget.getFocusedElementStyleClass();

    items = elem.children;
    expandable = this.m_widget.isExpandable();

    for (i=0; i<items.length; i++)
    {
        item = $(items[i]);
        context = this.createContext(item);

        this.setAriaProperties(item, context);

        item.uniqueId()
            .addClass(itemElementStyleClass);
        if (depth > 0)
        {
            item.addClass(this.m_widget.getDepthStyleClass(depth));
        }

        // tag it if item is not focusable
        if (!this.isFocusable(context))
        {
            item.addClass("oj-skipfocus");
        }

        groupItems = item.children("ul");
        if (groupItems.length > 0)
        {
            this.m_hier = true;

            item.children(":not(ul)")
                .wrapAll("<div></div>"); //@HTMLUpdateOK

            content = item.children().first();
            content.addClass(groupItemStyleClass);

            count = this.getItemsCount(groupItems[0]);
            if (count > 0)
            {
                if(item.hasClass(focusedElementStyleClass)) 
                {
                    item.removeClass(focusedElementStyleClass);
                    content.addClass(focusedElementStyleClass).attr("aria-expanded", "false");
                } 
                else 
                {
                    content.attr("role","presentation");
                    content.find("."+focusedElementStyleClass).attr("aria-expanded", "false");
                }

                // add the expand icon
                if (expandable)
                {
                    item.addClass("oj-collapsed");

                    content.uniqueId();

                    // add the expand icon
                    icon = document.createElement("a");
                    $(icon).attr("href", "#")
                           .attr("role", "button")
                           .attr("aria-labelledby", content.get(0).id)
                           .addClass("oj-component-icon oj-clickable-icon-nocontext")
                           .addClass(collapseClass);
                                
                    content.prepend(icon); //@HTMLUpdateOK               
                }
            }            

            if (this.shouldUseGridRole())
            {
                this.setGroupAriaProperties(content, count);
            }

            groupItem = $(groupItems[0]);
            groupItem.addClass(groupStyleClass)
                     .addClass(groupCollapseStyleClass)
                     .attr("role", this.shouldUseGridRole() ? (this.isCardLayout() ? "row" : "presentation") : "group")
                     .css("display", "none");
            this.modifyContent(groupItem[0], depth+1);
        }
        else
        {
            item.addClass(itemStyleClass);
        }

        if (this.m_widget._isSelectionEnabled() && this.isSelectable(context))
        {
            this.m_widget.getFocusItem(item).attr("aria-selected", false);
        }

        this.m_widget.itemRenderComplete(item[0], context);
    }    
};

/**
 * @private
 */
oj.StaticContentHandler.prototype.setRootAriaProperties = function()
{
    if (this.shouldUseGridRole())
    {
        this.m_root.setAttribute("role", "grid");
    }
    else
    {
        if (this.IsHierarchical())
        {
            this.m_root.setAttribute("role", "tree");
        }
        else
        {
            this.m_root.setAttribute("role", "listbox");
        }
    }
};

/**
 * @private
 */
oj.StaticContentHandler.prototype.unsetRootAriaProperties = function()
{
    this.m_root.removeAttribute("role");
};

/**
 * @private
 */
oj.StaticContentHandler.prototype.getItemsCount = function(item)
{
    return $(item).children("li").length;
};

/**
 * Creates the object with context information for the specified item
 * @param {jQuery} item the item to create context info object for
 * @return {Object} the context object
 * @private
 */
oj.StaticContentHandler.prototype.createContext = function(item)
{
    var context, parents;

    context = {};
    context['key'] = item.attr('id');
    context['parentElement'] = item.children().first()[0];
    context['index'] = item.index();
    context['data'] = item[0];
    context['component'] = this.m_widget.getWidgetConstructor();
    context = this.m_widget._FixRendererContext(context);

    // additional context info for hierarhical data
    if (this.IsHierarchical())
    {
        context['leaf'] = item.children("ul").length == 0;        
        parents = item.parents("li."+this.m_widget.getItemElementStyleClass());
        context['depth'] = parents.length;
        if (parents.length == 0)
        {
            context['parentKey'] = null;
        }
        else
        {
            context['parentKey'] = parents.first().attr('id');
        }     
    }

    return context;
};

/**
 * @private
 */
oj.StaticContentHandler.prototype.setAriaProperties = function(item, context)
{
    // if there's only one element inside the item and it is focusable, set
    // the role on it instead
    var elem = this.m_widget.getSingleFocusableElement(item);
    if (this.shouldUseGridRole())
    {
        if (context['leaf'] != undefined && !context['leaf'])
        {
            // it's a group item
            item.attr("role", "presentation");
            if (elem != item)
            {
                elem.attr("role", "gridcell");
            }
        }
        else
        {
            if (this.isCardLayout())
            {
                elem.attr("role", "gridcell");
            }
            else
            {
                item.attr("role", "row");
                if (elem != item)
                {
                    elem.attr("role", "gridcell");
                }
                else
                {
                    // we'll need to wrap content with a gridcell role
                    elem.children().wrapAll("<div role='gridcell' class='oj-listview-cell-element'></div>"); //@HTMLUpdateOK
                }     
            }
        }
    }
    else
    {
        elem.attr("role", this.IsHierarchical() ? "treeitem" : "option");
        if (elem != item)
        {
            item.attr("role", "presentation");
        }
    }

    elem.addClass(this.m_widget.getFocusedElementStyleClass());
};

/**
 * @private
 */
oj.StaticContentHandler.prototype.setGroupAriaProperties = function(group, count)
{
    var focusedElementStyleClass = this.m_widget.getFocusedElementStyleClass();

    // aria-expanded should be in the cell
    group.get(0).removeAttribute("aria-expanded");
    group.removeClass(focusedElementStyleClass);
    group.attr("role", "row");
    group.children().wrapAll("<div role='gridcell' aria-expanded='false' class='oj-listview-cell-element "+focusedElementStyleClass+"'></div>"); //@HTMLUpdateOK

    if (this.isCardLayout() && count > 1)
    {
        group.children().first().attr("aria-colspan", count);
    }
};

/**
 * @private
 */
oj.StaticContentHandler.prototype.unsetGroupAriaProperties = function(item)
{
    item.children("div").first().children().unwrap();
};

/**
 * @private
 */
oj.StaticContentHandler.prototype.unsetAriaProperties = function(item)
{
    var elem = this.m_widget.getSingleFocusableElement($(item));
    elem.removeAttr("role");
    elem.removeAttr("aria-selected");
    elem.removeAttr("aria-expanded");
    elem.removeClass(this.m_widget.getFocusedElementStyleClass());

    // need to unwrap since in non-card layout we wrap the content with a div for gridcell role
    if (this.shouldUseGridRole() && !this.isCardLayout())
    {
        if (elem != item)
        {
            $(item).removeAttr("role");
            elem.children().first().children().unwrap();
        }
        else
        {
            elem.children().first().children().unwrap()
                .children().unwrap();
        }
    }
};

oj.StaticContentHandler.prototype.GetKey = function(element)
{
    return $(element).attr("id");
};

oj.StaticContentHandler.prototype.FindElementByKey = function(key)
{
    return document.getElementById(key);
};

oj.StaticContentHandler.prototype.isFocusable = function(context)
{
    return this.m_widget.getItemFocusable(context);
};

oj.StaticContentHandler.prototype.isSelectable = function(context)
{
    return this.m_widget.getItemSelectable(context);
};

oj.StaticContentHandler.prototype.isCardLayout = function()
{
    return this.m_widget.isCardLayout();
};

oj.StaticContentHandler.prototype.shouldUseGridRole = function()
{
    return this.m_widget.ShouldUseGridRole();
};
/**
 * Partial Map impl, replace with ES6 Map when possible.
 * @constructor
 * @ignore
 */
oj.KeyMap = function()
{
};

oj.KeyMap.prototype.set = function(key, value)
{
    var index;

    if (this._mapKeys === undefined && this._mapValues === undefined)
    {
        this._mapKeys = [];
        this._mapValues = [];
    }

    index = this._mapKeys.indexOf(key);
    if (index > -1)
    {
        this._mapValues.splice(index, 1, value);
    }
    else
    {
        this._mapKeys.push(key);
        this._mapValues.push(value);
    }
};

oj.KeyMap.prototype.get = function(key)
{
    var index;
    if (this._mapKeys && this._mapValues)
    {
        index = this._mapKeys.indexOf(key);
        if (index > -1 && this._mapValues.length > index)
        {
            return this._mapValues[index];
        }
    }

    return null;
};

oj.KeyMap.prototype.deleteValue = function(value)
{
    var current = this._mapValues.indexOf(value);
    while (current > -1)
    {
        this._mapValues.splice(current, 1);
        this._mapKeys.splice(current, 1);
        current = this._mapValues.indexOf(value, current);
    }
};

/**
 * Default ExpandedKeySet class
 * Need a way to distinguish ExpandedKeySet set by application vs default one created by ListView
 * @extends {ExpandedKeySet}
 * @constructor
 * @ignore
 */
oj._ojListViewExpandedKeySet = function()
{
    oj._ojListViewExpandedKeySet.superclass.constructor.call(this);
};

// Subclass from KeySet
oj.Object.createSubclass(oj._ojListViewExpandedKeySet, oj.ExpandedKeySet, "ListViewExpandedKeySet");

/**
 * todo: create common utility class between combobox and listview
 * @private
 */
var _ListViewUtils = {
    clazz: function(SuperClass, methods)
    {
        var constructor = function() {};
        oj.Object.createSubclass(constructor, SuperClass, '');
        constructor.prototype = $.extend(constructor.prototype, methods);
        return constructor;
    }
};

/**
 * @export
 * @class oj._ojListView
 * @classdesc Listview
 * @constructor
 * @ignore
 * @private
 */ 
oj._ojListView = _ListViewUtils.clazz(Object, 
/** @lends oj._ojListView.prototype */
{
    // constants for key codes, todo: move to ListViewUtils
    LEFT_KEY: 37,
    RIGHT_KEY: 39,
    DOWN_KEY: 40,
    UP_KEY: 38,
    TAB_KEY: 9,
    ENTER_KEY: 13,
    ESC_KEY: 27,
    F2_KEY: 113,
    SPACE_KEY: 32,

    // constants for disclosure state
    /** @protected **/
    STATE_EXPANDED: 0,
    /** @protected **/
    STATE_COLLAPSED: 1,
    /** @protected **/
    STATE_NONE: 2,
    
    // minimum height of an item
    MINIMUM_ITEM_HEIGHT: 20,

    /**
     * Initialize the listview at creation
     * Invoked by widget
     */        
    init: function(opts)
    {
        var self = this, dndContext;

        this.readinessStack = [];
        this.element = opts.element;
        this.ojContext = opts.ojContext;
        this.OuterWrapper = opts.OuterWrapper;
        this.options = opts;

        this.element
            .uniqueId()
            .addClass(this.GetStyleClass()+" oj-component-initnode");

        if (this.OuterWrapper)
        {
            this.element[0].setAttribute("data-oj-context", "");
        }
        this.signalTaskStart("Initializing"); // Move component out of ready state; component is initializing. End in afterCreate()

        this.SetRootElementTabIndex();
        dndContext = this.GetDnDContext();
        // listens for dnd events if ListViewDndContext is defined
        if (dndContext)
        {
            this.m_dndContext = dndContext;

            this.ojContext._on(this.element, {
                "dragstart": function(event) 
                {
                    return dndContext._handleDragStart(event);
                },
                "dragenter": function(event) 
                {
                    return dndContext._handleDragEnter(event);
                },
                "dragover": function(event) 
                {
                    return dndContext._handleDragOver(event);
                },
                "dragleave": function(event) 
                {
                    return dndContext._handleDragLeave(event);
                },
                "dragend": function(event) 
                {
                   //mouseup will not be invoked on drag so resetting it to false.
                    self.m_preActive = false;
                    return dndContext._handleDragEnd(event);
                },
                "drag": function(event) 
                {
                    return dndContext._handleDrag(event);
                },
                "drop": function(event) 
                {
                    //mouseup will not be invoked on drag so resetting it to false.  
                    self.m_preActive = false;
                    return dndContext._handleDrop(event);
                    
                }
            });
        }

        this.ojContext._on(this.element, {
            "click": function(event) 
            {
                self.HandleMouseClick(event);
            },
            "touchstart": function(event)
            {
                self.HandleMouseDownOrTouchStart(event);
            },
            "touchend": function(event)
            {
                self.HandleTouchEndOrCancel(event);
            },
            "touchcancel": function(event)
            {
                self.HandleTouchEndOrCancel(event);
            },
            "mousedown": function(event)
            {
                if (event.button === 0)
                {
                    if (!self._recentTouch())
                    {
                        self.HandleMouseDownOrTouchStart(event);
                    }
                }
                else
                {
                    // on right click we should prevent focus from shifting to first item
                    self.m_preActive = true;
                }
            },
            "mouseup": function(event)
            {
                self._handleMouseUpOrPanMove(event);
                self.m_preActive = false;
            },
            "mouseout": function(event) 
            {
                self._handleMouseOut(event);
            },
            "mouseover": function(event) 
            {
                self._handleMouseOver(event);
            },
            "keydown": function(event) 
            {
                self.HandleKeyDown(event);
            },
            "keyup": function(event) 
            {
                self.HandleKeyUp(event);
            },
            "ojpanmove": function(event)
            {
                self._handleMouseUpOrPanMove(event);
            }
        });
        this.ojContext._on(this.ojContext.element, {
          "focus": function(event)
          {
              self.HandleFocus(event);
          },
          "blur": function(event)
          {
              self.HandleBlur(event);
          },
        });
        
        // in Firefox, need to explicitly make list container not focusable otherwise first tab will focus on the list container
        if (oj.AgentUtils.getAgentInfo()['browser'] === oj.AgentUtils.BROWSER.FIREFOX)
        {
            this._getListContainer().attr("tabIndex", -1);
        }

        // for item focus mode (aka roving focus), we'll need to use focusout handler instead
        // of blur because blur doesn't bubble
        this.ojContext._on(this.ojContext.element, {
            "focusin": function(event)
            {
                self.HandleFocus(event);
            },
            "focusout": function(event)
            {
                self.HandleFocusOut(event);
            }
        });

        this._registerScrollHandler();

        this.ojContext._focusable({
            'applyHighlight': self.ShouldApplyHighlight(),
            'recentPointer' : self.RecentPointerCallback(),
            'setupHandlers': function( focusInHandler, focusOutHandler) {
                self._focusInHandler = focusInHandler;
                self._focusOutHandler = focusOutHandler;
            }
        });
    },
    /**
     * Setup resources on listview after connect
     * Invoked by widget
     */        
    setupResources: function()
    {
        this.ojContext.document.bind("touchend.ojlistview touchcancel.ojlistview", this.HandleTouchEndOrCancel.bind(this));

        this._initContentHandler();
        // register a resize listener        
        this._registerResizeListener(this._getListContainer()[0]);
    },
    /**
     * Release resources held by listview after disconnect
     * Invoked by widget
     */        
    releaseResources: function()
    {
        this.ojContext.document.off(".ojlistview");

        this._destroyContentHandler();
        this._unregisterResizeListener(this._getListContainer());
    },
    /**
     * Initialize the listview after creation
     * Invoked by widget
     */        
    afterCreate: function () 
    {     
        this._buildList();
        this.signalTaskEnd(); // resolve component initializing task. Started in init()
    },
    /**
     * Redraw the entire list view after having made some external modifications.
     * Invoked by widget
     */
    refresh: function()
    {
        // reset content, wai aria properties, and ready state
        this._resetInternal();

        this.signalTaskStart("Refresh"); // signal method task start

        // set the wai aria properties 
        this.SetAriaProperties();

        // recreate the content handler
        this._initContentHandler();

        // reattach scroll listener if neccessary
        this._registerScrollHandler();

        this.signalTaskEnd(); // signal method task end
    },
    /**
     * Returns a Promise that resolves when the component is ready, i.e. after data fetching, rendering, and animations complete.
     * Invoked by widget
     * @return {Promise} A Promise that resolves when the component is ready.
     */
    whenReady: function()
    {
        return this.readyPromise;
    },
    /**
     * Destroy the list view
     * Invoked by widget
     */
    destroy: function()
    {
        this.element.removeClass(this.GetStyleClass()+" oj-component-initnode");

        this._unregisterResizeListener(this._getListContainer());
        this._resetInternal();

        //  - DomUtils.unwrap() will avoid unwrapping if the node is being destroyed by Knockout
        oj.DomUtils.unwrap(this.element, this._getListContainer()); 
    },

    /**
     * Force busy state to be resolve and flush the readiness stack
     * @private
     */
    _clearBusyState: function()
    {
        if (this.readinessStack && this.readinessStack.length > 0)
        {
            oj.Logger.warn("ListView did not end with a clean state, this could happen if ListView is detached before fetch is complete.  State: "+this.readinessStack);

            // this should resolve all the Promises, safe to assume the Promises should already been resolve if readinessStack is empty
            while (this.readinessStack.length > 0)
            {
                this.signalTaskEnd();
            }
        }
    },

    /**
     * Remove any wai-aria properties and listview specific attributes.
     * Reset anything done by the content handler.
     * @private
     */
    _resetInternal: function()
    {
        var self = this;

        this.UnsetAriaProperties();
        this._cleanupTabbableElementProperties(this.element);
        this._destroyContentHandler();

        this.m_active = null;
        this.m_isExpandAll = null;
        this.m_disclosing = null;
        this.m_itemHeight = null;

        this.readyPromise = new Promise(function(resolve, reject)
        {
            self.readyResolve = resolve;
        });

        this.ClearCache();

        // give dnd context a chance to clear internals
        if (this.m_dndContext != null)
        {
            this.m_dndContext.reset();
        }
    },

    /**
     * Called when listview root element is re-attached to DOM tree.
     * Invoke by widget
     */
    notifyAttached: function()
    {
        // call ContentHandler in case action is neccessarily
        if (this.m_contentHandler != null)
        {
            // restore scroll position as needed since some browsers reset scroll position 
            this.syncScrollPosition();

            this.m_contentHandler.notifyAttached();
        }
    },

    /**
     * In browsers [Chrome v35, Firefox v24.5, IE9, Safari v6.1.4], blur and mouseleave events are generated for hidden content but not detached content,
     * so for detached content only, we must use this hook to remove the focus and hover classes.
     * Invoke by widget.
     */
    notifyDetached: function() 
    {
        // Remove focus/hover/active style classes when listview element got detached from document.
        // For details see related button .
        this._getListContainer().removeClass("oj-focus-ancestor");

        if (this.m_active != null)
        {
            $(this.m_active['elem']).removeClass("oj-focus oj-focus-highlight");
        }

        if (this.m_hoverItem != null)
        {
            this._unhighlightElem(this.m_hoverItem, "oj-hover");
        }
    },

    /**
     * Called when application programmatically change the css style so that the ListView becomes visible
     */
    notifyShown: function()
    {
        // call ContentHandler in case action is neccessarily
        if (this.m_contentHandler != null)
        {
            // restore scroll position as needed since some browsers reset scroll position 
            this.syncScrollPosition();

            this.m_contentHandler.notifyShown();
        }
    },

    /**
     * Return the subcomponent node represented by the documented locator attribute values.
     * Invoked by widget
     * @param {Object} locator An Object containing at minimum a subId property 
     *        whose value is a string, documented by the component, that allows 
     *         the component to look up the subcomponent associated with that 
     *        string.  It contains:<p>
     *        component: optional - in the future there may be more than one 
     *        component contained within a page element<p>
     *        subId: the string, documented by the component, that the component 
     *        expects in getNodeBySubId to locate a particular subcomponent
     * @returns {Array.<(Element|null)>|Element|null} the subcomponent located by the subId string passed
     *          in locator, if found.<p>
     */
    getNodeBySubId: function(locator)
    {
        var subId, key, item, anchor;
 
        if (locator == null)
        {
            return this.element[0];
        }

        subId = locator['subId'];
        if (subId === 'oj-listview-disclosure' || subId === 'oj-listview-icon')
        {
            key = locator['key'];
            if (key != null)
            {
                item = this.FindElementByKey(key);
                if (item != null)
                {
                    // this should be the anchor
                    anchor = $(item).find('.oj-clickable-icon-nocontext').first();
                    if (this._isExpandCollapseIcon(anchor))
                    {
                        return anchor.get(0);
                    }
                }
            }     
        }
        else if (subId === 'oj-listview-item')
        {
            key = locator['key'];
            if (key != null)
            {
                return this.FindElementByKey(key);
            }
        }

        // Non-null locators have to be handled by the component subclasses
        return null;
    },
                
    /**
     * Returns the subId locator for the given child DOM node. 
     * Invoked by widget
     * @param {!Element} node - child DOM node
     * @return {Object|null} The subId for the DOM node, or null when none is found.
     */
    getSubIdByNode: function(node)
    {
        var item, key;

        // check to see if it's expand/collapse icon
        if (node != null && this._isExpandCollapseIcon(node))
        {
            item = this.FindItem(node);
            if (item != null && item.length > 0)
            {
                key = this.GetKey(item[0]);            
                if (key != null)
                {
                    return {'subId': 'oj-listview-disclosure', 'key': key};
                }
            }           
        }

        return null;
    },

    /**
     * Returns an object with context for the given child DOM node. 
     * This will always contain the subid for the node, defined as the 'subId' property on the context object. 
     * Additional component specific information may also be included. For more details on returned objects, see context objects.
     * Invoked by widget
     *
     * @param {!Element} node the child DOM node
     * @returns {Object|null} the context for the DOM node, or null when none is found.
     */
    getContextByNode: function(node)
    {
        var item, key, parent, index, context;

        item = this.FindItem(node);
        if (item != null && item.length > 0)
        {
            key = this.GetKey(item[0]);            
            if (key != null)
            {
                parent = item.parent();
                index = parent.children("."+this.getItemElementStyleClass()).index(item);
                context = {'subId': 'oj-listview-item', 'key': key, 'index': index};

                // group item should return the li
                if (parent.get(0) != this.element.get(0))
                {
                    context['parent'] = parent.parent().get(0);
                }

                // check if it's a group item
                if (item.children().first().hasClass(this.getGroupItemStyleClass()))
                {
                    context['group'] = true;
                }
                else
                {
                    context['group'] = false;
                }

                return context;
            }
        }           

        return null;
    },

    /**
     * Return the raw data for an item in ListView.
     * Invoked by widget
     *
     * @param {Object} context the context of the item to retrieve raw data.
     * @param {*=} context.key The key of the item.  If both index and key are specified, then key takes precedence.
     * @param {number=} context.index the index of the item relative to its parent.
     * @param {Element=} context.parent the parent node, not required if parent is the root.
     * @returns {*} data for the item.  Returns null if the item is not available locally.  Returns the item element if static HTML is used as data.
     */
    getDataForVisibleItem: function(context)
    {
        var key, index, parent, item;

        key = context['key'];
        // key takes precedence
        if (key != null)
        {
            item = this.FindElementByKey(key);
        }

        // if we can't find the item with key, try to use index, if specified
        if (item == null)
        {
            index = context['index'];
            parent = context['parent'];

            if (parent == null)
            {
                // use the root element
                parent = this.element.get(0);
            }
            else
            {
                // find the appropriate group element
                parent = $(parent).children("ul."+this.getGroupStyleClass()).first();
            }

            item = $(parent).children("li").get(index);
        }

        if (item != null && $(item).hasClass(this.getItemStyleClass()))
        {
            return this._getDataForItem(item);
        }

        return null;
    },

    /**
     * Retrieve data stored in dom
     * @param {Element} item
     * @return {*} data for item
     * @private
     */
    _getDataForItem: function(item)
    {
        // if static HTML, returns the item's dom element
        if (this.GetOption("data") == null)
        {
            return item;
        }
        else
        {
            return $.data(item, "data");
        }
    },

    /**
     * Unregister event listeners for resize the container DOM element.
     * @param {Element} element  DOM element
     * @private
     */
    _unregisterResizeListener: function(element)
    {
        if (element && this._resizeHandler)
        {
            // remove existing listener
            oj.DomUtils.removeResizeListener(element, this._resizeHandler);    
        }
    },

    /**
     * Register event listeners for resize the container DOM element.
     * @param {Element} element  DOM element
     * @private
     */
    _registerResizeListener: function(element)
    {         
        if (element)
        {
            if (this._resizeHandler == null)
            {
                this._resizeHandler = this.HandleResize.bind(this);
            }

            oj.DomUtils.addResizeListener(element, this._resizeHandler);
        }
    },
    
    /**
     * Returns DnD Context, needed to override in navigationlist
     * @protected
     */
    GetDnDContext: function()
    {
        if (typeof oj.ListViewDndContext != "undefined")
        {
            return new oj.ListViewDndContext(this);
        }
    },
    
    /**
     * The resize handler.
     * @param {number} width the new width
     * @param {number} height the new height
     * @private
     */
    HandleResize: function(width, height)
    {
        if (width > 0 && height > 0 && this.m_contentHandler != null)
        {
            this.m_contentHandler.HandleResize(width, height);
        }
    },   

    /**
     * Whether focus highlight should be applied
     * @return {boolean} true if should apply focus highlight, false otherwise
     * @protected
     */
    ShouldApplyHighlight: function()
    {
        return true;
    },
    /**
     * check Whether recent pointer acivity happened or not. 
     * Only used for sliding navlist to avoid focus ring on new focusable item
     * after completing expand/collapse animation.
     * @protected
     */
    RecentPointerCallback: function()
    {
      return function(){return false};
    },
    /**
     * Whether ListView should refresh if certain option is updated
     * @param {Object} options the options to check
     * @return {boolean} true if should refresh, false otherwise
     * @protected
     */
    ShouldRefresh: function(options)
    {
        return (options['data'] != null || 
                options['drillMode'] != null || 
                options['groupHeaderPosition'] != null || 
                options['item'] != null || 
                options['scrollPolicy'] != null ||
                options['scrollPolicyOptions'] != null);
    },

    /**
     * Returns true if value is a KeySet, false otherwise
     * @private
     */
    _isKeySet: function(value)
    {
        return (value.values && value.isAddAll);
    },

    /**
     * Sets multiple options 
     * Invoke by widget
     * @param {Object} options the options object
     * @param {Object} flags additional flags for option
     * @return {boolean} true to refresh, false otherwise
     */
    setOptions: function(options, flags)
    { 
        var self = this, elem, expanded, keys, active, scroller, pos;

        if (this.ShouldRefresh(options))
        {
            // data updated, need to refresh
            return true;
        }

        if (options['expanded'] != null)
        {            
            // should only apply if data is hierarchical 
            // q: could expanded be change if drillMode is 'expanded'?
            if (this.m_contentHandler.IsHierarchical())
            {
                // clear collapsed items var
                this._collapsedKeys = undefined;

                expanded = options['expanded'];

                if (Array.isArray(expanded) || expanded.values)
                {
                    this.signalTaskStart("Set expanded option"); // signal task start

                    // collapse all expanded items first
                    this._collapseAll();

                    // itemRenderComplete would check expanded option to expand nodes as needed
                    // however, since options has not been updated yet this will cause previous
                    // expanded nodes to expand
                    // an option would be to clear the expanded option to null when doing collapseAll
                    // but the issue is that optionChange would be fired AND even if we can suppress 
                    // the optionChange event, when the actual optionChange event is fired, the
                    // previousValue param would be wrong (it would be null)
                    // so instead we'll use to flag so that itemRenderComplete would detect and ignore
                    // the expanded option
                    this._ignoreExpanded = true;

                    try
                    {
                        // expand each key, both Set and Array supports forEach
                        keys = this._isKeySet(expanded) ? expanded.values() : expanded;
                        keys.forEach(function(key)
                        {
                            self.expandKey(key, true, true, true);
                        });
                    }
                    finally
                    {
                        this._ignoreExpanded = undefined;

                        this.signalTaskEnd(); // signal task end
                    }
                }
                else if (expanded === "all" || expanded.deletedValues) 
                {
                    // if need to return all, just refresh the view
                    return true;
                }
            }
        }
        
        if (options['currentItem'] != null)
        {            
            elem = this.FindElementByKey(options["currentItem"]);
            if (elem != null)
            {
                elem = $(elem);
                if (!this.SkipFocus(elem))
                {
                    active= document.activeElement;
                    // update tab index and focus only if listview currently has focus
                    if (active && this.element.get(0).contains(active))
                    {
                        this.ActiveAndFocus(elem, null);
                    }
                    else
                    {
                        // update internal state only
                        this._setActive(elem, null, true);
                    }
                }
            }
        }
        else if (options['currentItem'] === null)
        {
            // currentItem is deliberately set to null if this case is entered; deliberately clear active element and its focus
            this.UnhighlightActive();
            this.m_active = null;
            this.SetRootElementTabIndex();
        }
        this.HandleSelectionOption(options);
        
        if (options['selectionMode'] != null)
        {            
            // clear existing selection if selection mode changes
            this._clearSelection(true);

            // reset wai aria properties 
            this.SetAriaProperties();

            // update aria-selected on item
            this.UpdateItemAriaProperties(options['selectionMode']);
        }

        if (options['scrollTop'] != null)
        {
            scroller = this._getScroller();
            pos = options['scrollTop'];
            if (pos != null && !isNaN(pos))
            {
                scroller.scrollTop = pos;
            }
        }

        if (options['scrollPosition'] != null)
        {
            this._setCurrentScrollPosition(options['scrollPosition']);
            // remove it so it doesn't trigger an option change 
            delete options['scrollPosition'];
        }

        // if reorder switch to enabled/disabled, we'll need to make sure any reorder styling classes are added/removed from focused item
        if (this._shouldDragSelectedItems() && this.m_active != null && options['dnd'] != null && options['dnd']['reorder'] != null)
        {
            if (options['dnd']['reorder']['items'] === 'enabled')
            {
                this.m_dndContext._setItemDraggable(this.m_active['elem']);
            }
            else if (options['dnd']['reorder']['items'] === 'disabled')
            {
                this.m_dndContext._unsetItemDraggable(this.m_active['elem']);
            }
        }

        return false;
    },

    /**
     * Set Selection option. Overriden by Navlist.
     * @param {Object} options the options object
     * @protected
     */
    HandleSelectionOption: function(options) 
    {
      var elem, selection, i;
      if (options['selection'] != null)
      {            
          if (this._isSelectionEnabled())
          {
              // remove any non-selectable items, this will effectively clone the selection as well,
              // which we'll need to do
              options['selection'] = this._filterSelection(options['selection']);
              selection = options['selection'];

              // clear selection first
              this._clearSelection(false);

              // selects each key
              for (i=0; i<selection.length; i++)            
              {
                  elem = this.FindElementByKey(selection[i]);
                  if (elem != null)
                  {
                      this._applySelection(elem, selection[i]);
                  }
              }
          }
      }
    },

    /**
     * Trigger an event to fire.
     * @param {string} type the type of event
     * @param {Object} event the jQuery event to fire
     * @param {Object} ui the ui param
     * @protected
     */
    Trigger: function(type, event, ui)
    {
        return this.ojContext._trigger(type, event, ui);
    },

    /**
     * Sets an option on the widget
     * @param {string} key the option key
     * @param {Object} value the option value
     * @param {Object=} flags any optional parameters 
     * @protected
     */
    SetOption: function(key, value, flags)
    {
        this.ojContext.option(key, value, flags);
    },

    /**
     * Gets the value of an option from the widget
     * @param {string} key the option key
     * @return {Object} the value of the option
     * @protected
     */
    GetOption: function(key)
    {
        return this.ojContext.option(key);
    },

    /**
     * Compose a description for busy state
     * @param {string} description the description
     * @return {string} the busy state description
     * @private
     */
    _getBusyDescription: function(description)
    {
        var id = this.ojContext._IsCustomElement() ? this.getRootElement().attr("id") : this.element.attr("id");
        return "The component identified by '" + id + "', " + description;
    },

    /**
     * Invoke whenever a task is started. Moves the component out of the ready state if necessary.
     * @param {string=} description the description of the task
     */
    signalTaskStart: function(description)
    {
        var self = this, busyContext, options;
        if (this.readinessStack)
        {
            if (this.readinessStack.length == 0)
            {
                this.readyPromise = new Promise(function(resolve, reject)
                {
                    self.readyResolve = resolve;
                });

                // whenReady is deprecated in favor of page busystate (but we still need to support old syntax)
                busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
                options = description != null ? {'description' : this._getBusyDescription(description)} : {};
                self.busyStateResolve = busyContext.addBusyState(options);
            }
            this.readinessStack.push(description != null ? description : "unknown task");
        }
    },

    /**
     * Invoke whenever a task finishes. Resolves the readyPromise if component is ready to move into ready state.
     */
    signalTaskEnd: function()
    {
        if (this.readinessStack && this.readinessStack.length > 0)
        {
            this.readinessStack.pop();
            if (this.readinessStack.length == 0)
            {
                this.readyResolve(null);

                // whenReady is deprecated in favor of page busystate (but we still need to support old syntax)
                this.busyStateResolve(null);
                this.busyStateResolve = null;
            }
        }
    },

    /**
     * Checks whether ListView is in ready state.  Called by the ContentHandler.
     * @return {boolean} true if ListView is in ready state, false otherwise
     */     
    isReady: function()
    {
        return (this.busyStateResolve == null);
    },

    /**
     * Throw an error.  Do any neccessary cleanup.
     */
    throwError: function(err)
    {
        if (this.readinessStack)
        {
            while (this.readinessStack.length > 0)
            {
                this.signalTaskEnd();
            }
        }

        throw err;
    },

    /**
     * Gets an array of items based on specified ids.
     * @param {Array} ids an array of item ids.
     * @return {Array} an array of elements matching the item ids.
     */
    getItems: function(ids)
    {
        var elem, self = this, items = [];

        $.each(ids, function(index, value)
        {
            elem = self.FindElementByKey(value);
            if (elem != null)
            {
                items.push(elem);
            }
        });

        return items;
    },

    /************************************** Core rendering ********************************/
    /**
     * Whether the listview is in card layout mode
     * @return {boolean} true if it is in card layout mode, false otherwise
     */
    isCardLayout: function()
    {
        var elem = this.ojContext._IsCustomElement() ? this.getRootElement() : this.element;
        return elem.hasClass("oj-listview-card-layout");
    },

    /**
     * Whether to use grid role for ListView, to be override by NavList
     * @protected
     */
    ShouldUseGridRole: function()
    {
        return true;
    },

    /**
     * Whether to scrollPosition is supported, NavList for example do not support this
     * @protected
     */
    ShouldUpdateScrollPosition: function()
    {
        return this.ShouldUseGridRole() && this.ojContext._IsCustomElement();
    },

    /**
     * Destroy the content handler
     * @private
     */
    _destroyContentHandler: function()
    {
        if (this.m_contentHandler != null)
        {
            this.m_contentHandler.Destroy();
            delete this.m_contentHandler;
            this.m_contentHandler = null;
        }

        // ensure all busy states caused by outstanding fetch are resolved
        this._clearBusyState();
    },

    /**
     * Initialize the content handler based on data type
     * @private
     */
    _initContentHandler: function()
    {
        var data, fetchSize;

        this.signalTaskStart("Initialize ContentHandler"); // signal method task start

        this.showStatusText();

        data = this.GetOption("data");
        if (data != null)
        {
            if (typeof oj.TableDataSource !== "undefined" && data instanceof oj.TableDataSource)
            {
                // TODO: load the adapter as needed
                this.m_contentHandler = new oj.IteratingDataProviderContentHandler(this, this.element[0], new oj.TableDataSourceAdapter(data));
            }
            else if (typeof oj.TreeDataSource !== "undefined" && data instanceof oj.TreeDataSource)
            {
                this.m_contentHandler = new oj.TreeDataSourceContentHandler(this, this.element[0], data);
            }
            else if (oj.DataProviderFeatureChecker.isIteratingDataProvider(data))
            {
                this.m_contentHandler = new oj.IteratingDataProviderContentHandler(this, this.element[0], data);
            }
            else
            {
                this.throwError("Invalid data or missing module");
            }
        }
        else
        {
            // StaticContentHandler will handle cases where children are invalid or empty
            this.m_contentHandler = new oj.StaticContentHandler(this, this.element[0]);
        }

        // kick start rendering
        this.m_contentHandler.RenderContent();

        this.signalTaskEnd(); // signal method task end
    },

    /**
     * Update active descendant attribute
     * @param {jQuery} elem the active item element
     * @protected
     */
    UpdateActiveDescendant: function(elem)
    {
        this.element.attr("aria-activedescendant", elem.attr("id"));
    },

    /**
     * Sets wai-aria properties on root element
     * @protected
     */
    SetAriaProperties: function()
    {
        if (this._isMultipleSelection())
        {
            this.element.attr("aria-multiselectable", true);
        }
        else if (this._isSelectionEnabled())
        {
            this.element.attr("aria-multiselectable", false);
        }
    },

    /**
     * Removes wai-aria properties on root element
     * @protected
     */
    UnsetAriaProperties: function()
    {
        this.element.removeAttr("aria-activedescendant")
                    .removeAttr("aria-multiselectable");
    },

    /**
     * When selectionMode option is updated, the aria-selected
     * attribute must be remove or updated
     * @param {string} selectionMode the new selection mode
     * @protected
     */
    UpdateItemAriaProperties: function(selectionMode)
    {
        var self = this, func, items, i;

        if (selectionMode == "none")
        {
            this.element.removeAttr("aria-multiselectable");

            func = function(item)
            {
                self.getFocusItem(item).removeAttr("aria-selected");
            };
        }
        else
        {
            if (selectionMode == "single")
            {
                this.element.attr("aria-multiselectable", false);
            }
            else
            {
                this.element.attr("aria-multiselectable", true);
            }

            func = function(item)
            {
                self.getFocusItem(item).attr("aria-selected", 'false');
            };
        }

        items = this._getItemsCache();
        for (i=0; i<items.length; i++)
        {
            func($(items[i]));
        }
    },

    /**
     * Build the elements inside and around the root
     * @param {Element} root the root element
     * @private
     */
    _buildList: function(root)
    {
        var container, status, accInfo;

        container = this._getListContainer();
        this.SetAriaProperties();

        this.m_elementOffset = this.element.get(0).offsetTop;

        status = this._buildStatus();
        container.append(status); // @HTMLUpdateOK
        this.m_status = status;

        accInfo = this._buildAccInfo();
        container.append(accInfo); // @HTMLUpdateOK
        this.m_accInfo = accInfo;

        // touch specific instruction text for screen reader for reordering
        if (this._isTouchSupport() && this.m_dndContext != null)
        {
            container.append(this._buildAccInstructionText()); // @HTMLUpdateOK
        }
    },

    /**
     * Build a status bar div
     * @return {jQuery} the root of the status bar
     * @private
     */
    _buildStatus: function()
    {
        var icon, root;

        icon = $(document.createElement("div"));
        icon.addClass("oj-icon oj-listview-loading-icon");

        root = $(document.createElement("div"));        
        root.addClass("oj-listview-status-message oj-listview-status")
            .attr({"id": this._createSubId("status"), 
                   "role": "status"});
        root.append(icon); // @HTMLUpdateOK

        return root;
    },

    /**
     * Build the accessible text info div
     * @return {jQuery} the root of the acc info div
     * @private
     */
    _buildAccInfo: function()
    {
        var root = $(document.createElement("div"));
        root.addClass("oj-helper-hidden-accessible")
            .attr({"id": this._createSubId("info"), 
                   "role": "status"});

        return root;
    },

    /**
     * Build the accessible instruction text for touch devices
     * @return {jQuery} the root of the acc info div
     * @private
     */
    _buildAccInstructionText: function()
    {
        var root = $(document.createElement("div"));
        root.addClass("oj-helper-hidden-accessible")
            .attr({"id": this._createSubId("instr")});
        root.text(this.ojContext.getTranslatedString("accessibleReorderTouchInstructionText"));        

        return root;
    },

    /**
     * Sets the accessible text info
     * @param {string} text the text to set on accessible info div
     * @private
     */
    _setAccInfoText: function(text)
    {
        if (text != "" && this.m_accInfo.text() != text)
        {
            this.m_accInfo.text(text);
        }
    },

    /**
     * Displays the 'fetching' status message
     * @private
     */
    showStatusText: function()
    {
        var msg, self = this, statusHeight, container, containerHeight;

        // it's already shown
        if (this.m_showStatusTimeout)
        {
            return;
        }

        this.m_showStatusTimeout = setTimeout(function()
        {
            // remove any empty text div
            $(document.getElementById(self._createSubId('empty'))).remove();        

            msg = self.ojContext.getTranslatedString("msgFetchingData");

            container = self._getListContainer();
            self.m_status.attr("aria-label", msg)
                         .css("left", Math.max(0, container.outerWidth()/2 - self.m_status.outerWidth()/2))
                         .css("top", Math.max(0, container.outerHeight()/2 - self.m_status.outerHeight()/2))
                         .show();        

            // make sure the container is tall enough to show the indicator
            if (self.m_minHeightSet === undefined)
            {
                statusHeight = self.m_status.get(0).offsetHeight;
                containerHeight = container.get(0).offsetHeight;
                container.css('minHeight', Math.max(containerHeight, statusHeight + self._getListContainerBorderWidth()));
                self.m_minHeightSet = true;
            }

            self.m_showStatusTimeout = null;
        }, this._getShowStatusDelay());
    },

    /**
     * Retrieve the delay before showing status
     * @return {number} the delay in ms
     * @private
     */
    _getShowStatusDelay: function()
    {
        var delay;

        if (this.defaultOptions == null)
        {
            this.defaultOptions = oj.ThemeUtils.parseJSONFromFontFamily(this.getOptionDefaultsStyleClass());
        }
        delay = parseInt(this.defaultOptions['showIndicatorDelay'], 10);

        return isNaN(delay) ? 0 : delay;
    },

    /**
     * Hide the 'fetching' status message
     * @private
     */
    hideStatusText: function()
    {
        if (this.m_showStatusTimeout)
        {
            clearTimeout(this.m_showStatusTimeout);
            this.m_showStatusTimeout = null;
        }
        this.m_status.hide();
    },

    /**
     * Retrieves the root element
     * Invoke by widget
     */
    getRootElement: function()
    {
        return this._getListContainer();
    },

    /**
     * Retrieves the div around the root element, create one if needed.
     * @return {jQuery} the div around the root element 
     * @private
     */    
    _getListContainer: function()
    {
        if (this.m_container == null)
        {
            this.m_container = this._createListContainer();
        }
        return this.m_container;
    },

    /**
     * Creates the div around the root element.
     * @return {jQuery} the div around the root element
     * @private
     */    
    _createListContainer: function()
    {
        var listContainer;
        if (this.OuterWrapper) 
        {
            listContainer = $(this.OuterWrapper);
        } 
        else 
        {
            listContainer = $(document.createElement('div'));
            this.element.parent()[0].replaceChild(listContainer[0], this.element[0]); //@HTMLUpdateOK
        }

        listContainer.addClass(this.GetContainerStyleClass()).addClass("oj-component");
        listContainer.prepend(this.element); //@HTMLUpdateOK
        return listContainer;
    },

    /**
     * If the empty text option is 'default' return default empty translated text, 
     * otherwise return the emptyText set in the options 
     * @return {string} the empty text
     * @private
     */
    _getEmptyText: function()
    {
        return this.ojContext.getTranslatedString("msgNoData");
    },

    /**
     * Build an empty text div and populate it with empty text
     * @return {Element} the empty text element
     * @private 
     */
    _buildEmptyText: function()
    {
        var emptyText, empty;

        emptyText = this._getEmptyText();
        empty = document.createElement("li");
        empty['id'] = this._createSubId("empty");
        empty['className'] = this.getEmptyTextStyleClass() + " "+ this.getEmptyTextMarkerClass();
        empty.textContent = emptyText;

        return empty;
    },

    /**
     * Determines whether the specified item is expanded
     * @param {jQuery} item the item element
     * @return {number} 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
     * @protected
     */
    GetState: function(item)
    {
        var expanded = this.getFocusItem(item).attr("aria-expanded");
        if (expanded == "true")
        {
            return this.STATE_EXPANDED;
        }
        else if (expanded == "false")
        {
            return this.STATE_COLLAPSED;
        }
        else
        {
            return this.STATE_NONE;
        }
    },

    /**
     * Sets the disclosed state of the item
     * @param {jQuery} item the item element
     * @param {number} state 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
     * @protected
     */
    SetState: function(item, state)
    {
        var expandable = this.isExpandable();

        if (state == this.STATE_EXPANDED)
        {
            this.getFocusItem(item).attr("aria-expanded", "true");
            if (expandable)
            {
                item.removeClass("oj-collapsed")
                    .addClass("oj-expanded");
            }
        }
        else if (state == this.STATE_COLLAPSED)
        {
            this.getFocusItem(item).attr("aria-expanded", "false");
            if (expandable)
            {
                item.removeClass("oj-expanded")
                    .addClass("oj-collapsed");
            }
        } 
    },

    /**
     * Gets the item option
     * @param {string} name the name of the option
     * @param {Object} context the context object
     * @param {boolean} resolve true evaluate if return value is a function, false otherwise
     * @return {function(Object)|Object|null} returns the item option
     * @private
     */
    _getItemOption: function(name, context, resolve)
    {
        var option, value;

        option = this.GetOption("item");
        value = option[name];

        if (typeof value == 'function' && resolve)
        {
            return value.call(this, context);
        }

        return value;
    },

    /**
     * Gets the item.focusable option.
     * @param {Object} context the context object
     * @return {boolean} true if item.focusable option is derieved to be true, false otherwise
     */
    getItemFocusable: function(context)
    {
        return this._getItemOption("focusable", context, true);
    },

    /**
     * Gets the item.selectable option.
     * @param {Object} context the context object
     * @return {boolean} true if item.selectable option is derieved to be true, false otherwise
     */
    getItemSelectable: function(context)
    {
        // if it's not focusable, it's not selectable also
        return this.getItemFocusable(context) && this._getItemOption("selectable", context, true);
    },

    /**
     * Gets the item renderer
     * @return {function(Object)|null} returns the item renderer
     * @private
     */
    _getItemRenderer: function()
    {
        var renderer = this._getItemOption("renderer", null, false);
        if (typeof renderer != 'function')
        {
            // cannot be non-function
            return null;
        }
        return this._WrapCustomElementRenderer(renderer);
    },

    /**
     * Returns the inline template element inside oj-list-view
     * @return {Element|null} the inline tmeplate element
     */
    getItemTemplate: function()
    {
        var slotMap, slot;

        if (this.m_template === undefined)
        {
            // cache the template, assuming replacing template will require refresh
            this.m_template = null;
            if (this.ojContext._IsCustomElement())
            {
                slotMap = oj.BaseCustomElementBridge.getSlotMap(this.getRootElement()[0]);        
                slot = slotMap['itemTemplate'];
                if (slot && slot.length > 0 && slot[0].tagName.toLowerCase() === "template")
                {
                    this.m_template = slot[0];
                }
            }
        }
        return this.m_template;
    },

    /**
     * Called by content handler once the content of an item is rendered triggered by an insert event
     * @param {Element} elem the item element
     * @param {Object} context the context object used for the item
     */
    itemInsertComplete: function(elem, context)
    {
        // hook for NavList
    },
    
    /**
     * Called by content handler once the content of an item is rendered triggered by an insert event
     */
    BeforeInsertItem: function()
    {
      // hook for NavList
    },

    /**
     * Called by content handler once the content of an item is removed triggered by an remove event
     * @param {Element} elem the item element
     */
    itemRemoveComplete: function(elem)
    {
        // if it's the current focus item, try to focus on the next/prev item.  If there are none, then focus on the root element
        if (this.m_active != null && this.m_active['elem'] && $(this.m_active['elem']).get(0) == elem)
        {
            var next = elem.nextElementSibling;
            if (next == null || !$(next).hasClass(this.getItemElementStyleClass()))
            {
                next = elem.previousElementSibling;
                if (next == null || !$(next).hasClass(this.getItemElementStyleClass()))
                {
                    this.SetOption('currentItem', null);
                }                
            }

            if (next != null && $(next).hasClass(this.getItemElementStyleClass()))
            {
                this.SetOption('currentItem', this.GetKey(next));
            }
        }

        // disassociate element from key map
        if (elem != null && elem.id && this.m_keyElemMap != null)
        {
            this.m_keyElemMap.deleteValue(elem.id);
        }
    },

    /**
     * Called by content handler once the content of an item is rendered
     * @param {Element} elem the item element
     * @param {Object} context the context object used for the item
     */
    itemRenderComplete: function(elem, context)
    {
        var key, selection, selectedItems, i, index, clone, expanded, self, current;

        // dnd
        if (this.m_dndContext != null)
        {
            this.m_dndContext.itemRenderComplete(elem);
        }

        // make all tabbable elements non-tabbable, since we want to manage tab stops
        this._disableAllTabbableElements(elem);

        key = context['key'];

        // update as selected if it is in selection, check if something already selected in single selection
        if (this._isSelectionEnabled())
        {
            selection = this.GetOption("selection");
            if (this.IsSelectable(elem))
            {
                for (i=0; i<selection.length; i++)
                {
                    if (selection[i] == key || oj.Object.compareValues(selection[i], key))
                    {
                        this._applySelection(elem, selection[i]);
                    
                        // if it's single selection, then bail
                        if (!this._isMultipleSelection())
                        {
                            if (selection.length > 1)
                            {
                                // we'll have to modify the value
                                selectedItems = this.FindElementByKey(key);
                                this._setSelectionOption([key], null, selectedItems, context['data']);
                            }
                            break;
                        }
                    }
                }

                // if selectionRequired is set to true and selection is empty, selects the first selectable item
                // this should be run once since selection won't be empty afterwards
                if (selection.length == 0 && this._isSelectionRequired())
                {
                    this._applySelection(elem, key);
                    // need to pass data since 1) to avoid unneccesary lookup 2) since item is not in live dom, getDataForVisibleItem would not work
                    this._setSelectionOption([key], null, [elem], context['data']);
                }
            }
            else
            {
                // the selection is invalid, remove it from selection
                index = this.GetIndexOf(selection, key);
                if (index > -1)
                {
                    // clone before modifying
                    clone = selection.slice(0);
                    clone.splice(index, 1);

                    selectedItems = new Array(clone.length);
                    for (i = 0; i < clone.length; i++)
                    {
                        selectedItems[i] = this.FindElementByKey(clone[i]);
                    }
                    this._setSelectionOption(clone, null, selectedItems);
                }                
            }
        }

        // update if it is in expanded, ensure data is hierarchical
        if (this.m_contentHandler.IsHierarchical() && this._ignoreExpanded == undefined)
        {
            // checks if it is expandable && is collapsed
            if (this.GetState($(elem)) == this.STATE_COLLAPSED)
            {
                expanded = this.GetOption("expanded");
                // checks if expand all
                if (this._isExpandAll())
                {
                    // for legacy syntax, expanded is not real-time in certain cases, i.e. you 
                    // can have expanded='all' but some itms are collapsed
                    // for custom element we don't care about collapsedKeys since KeySet is keeping track of it
                    if (this.ojContext._IsCustomElement() || this._collapsedKeys == undefined)
                    {
                        // don't animate
                        this.ExpandItem($(elem), null, false, null, false, false, false);
                    }
                }
                else if (!this.ojContext._IsCustomElement() && Array.isArray(expanded))
                {
                    // legacy syntax array of expanded keys
                    // checks if specified expanded
                    self = this;
                    $.each(expanded, function(index, value)
                    {
                        // if it was explicitly collapsed
                        if (value == key && (self._collapsedKeys == undefined || self._collapsedKeys.indexOf(value) == -1))
                        {
                            // don't animate
                            self.ExpandItem($(elem), null, false, null, false, false, false);
                        }
                    });
                }
                else if (expanded.has)
                {
                    // KeySet case
                    if (expanded.has(key))
                    {
                        // don't animate
                        this.ExpandItem($(elem), null, false, null, false, false, false);                        
                    }
                }
            }
        }

        // checks if the active element has changed, this could happen in TreeDataSource, where the element gets remove when collapsed
        if (this.m_active != null && key == this.m_active['key'] && this.m_active['elem'] != null && elem != this.m_active['elem'].get(0))
        {
            this.m_active['elem'] = $(elem);
        }
    },

    /**
     * Called by content handler once content of all items are rendered
     */
    renderComplete: function()
    {
        var empty, firstItem, current, elem;

        this.hideStatusText();

        // remove any empty text div
        $(document.getElementById(this._createSubId('empty'))).remove();        

        // clear items cache
        this.m_items = null;
        this.m_groupItems = null;

        // check if it's empty
        if (this.element[0].childElementCount == 0)
        {
            empty = this._buildEmptyText();
            this.element.append(empty); // @HTMLUpdateOK

            // fire ready event
            this.Trigger("ready", null, {});
            
            return;
        }
        else
        {
            // figure out the average item height, we'll need that in scroll listener
            if (this.ShouldUpdateScrollPosition() && this.m_itemHeight == null)
            {
                if (this.m_contentHandler.IsHierarchical())
                {
                    // could be a group
                    firstItem = this.element.children("li."+this.getItemElementStyleClass()).first();
                    if (firstItem.length > 0)
                    {
                        this.m_itemHeight = firstItem.get(0).firstElementChild.offsetHeight;
                    }
                }
                else
                {
                    firstItem = $(this._getRootNodeForItems()).children("li."+this.getItemStyleClass()).first();
                    if (firstItem.length > 0)
                    {
                        this.m_itemHeight = firstItem.get(0).offsetHeight;
                    }                    
                }
            }
        }

        // check if current is specified
        current = this.GetOption("currentItem");
        if (current != null)
        {
            elem = this.FindElementByKey(current);
            if (elem == null)
            {
                // it's not valid anymore, reset current
                this.SetOption('currentItem', null);
            }
            else
            {    
                // make sure item is focusable also
                if (this.m_active == null && !this.SkipFocus($(elem)))
                {
                    this.ActiveAndFocus($(elem), null);
                }
            }
        }

        // if listview has focus but there's no active element, then set focusable item
        // this could happen after refresh from context menu
        if (this._getListContainer().hasClass("oj-focus-ancestor") && this.m_active == null && current == null)
        {
            this._initFocus();
        }

        // clear the scroll and fetch flag before calling syncScrollPosition
        this.m_scrollAndFetch = undefined;

        // update scroll position if it's not in sync, make sure we are not in the middle of scrolling
        if (!this.m_ticking || this.m_scrollPosition != null)
        {
            this.syncScrollPosition();
        }

        // fire ready event
        this.Trigger("ready", null, {});
    },

    /**
     * @private
     */
    _setScrollY: function(scroller, y)
    {
        if (!this._skipScrollUpdate)
        {
            this.signalTaskStart("waiting for scroll handler");
        }

        // flag it so that handleScroll won't do anything
        this._skipScrollUpdate = true;
        scroller.scrollTop = y;

        // update sticky header as needed
        this._handlePinGroupHeader();
    },

    /**
     * Sets the bidi independent position of the horizontal scroll position that
     * is consistent across all browsers.
     * @private
     */
    _setScrollX: function(scroller, x)
    {
        if (!this._skipScrollUpdate)
        {
            this.signalTaskStart("waiting for scroll handler");
        }

        // flag it so that handleScroll won't do anything
        this._skipScrollUpdate = true;

        oj.DomUtils.setScrollLeft(scroller, x);
    },

    /**
     * Retrieve the bidi independent position of the horizontal scroll position that
     * is consistent across all browsers.
     * @private
     */
    _getScrollX: function(scroller)
    {
        return oj.DomUtils.getScrollLeft(scroller);
    },

    /**
     * Synchronize the scroll position
     * @protected
     */
    syncScrollPosition: function(position)
    {
        var scroller, coord, x, y, scrollTop, scrollPosition, newScrollPosition;

        scroller = this._getScroller();
        // check if it's even scrollable
        if (scroller.scrollHeight == scroller.clientHeight)
        {
            return;
        }

        if (this.ShouldUpdateScrollPosition())
        {
            if (this.m_scrollPosition != null)
            {
                position = this.m_scrollPosition;
            }
            else if (position === undefined)
            {
                position = this.GetOption("scrollPosition");
            }

            // figure out what the final y should be
            coord = this._getScrollCoordinates(position);
            x = coord.x;
            y = coord.y;

            if (isNaN(x) && isNaN(y))
            {
                // invalid scroll position
                if (this.m_scrollPosition != null)
                {
                    // we'll still need to report current scroll position, which could have changed because of scroll and fetch
                    this.SetOption('scrollPosition', this._getCurrentScrollPosition(scrollTop), {'_context': {originalEvent: null, internalSet: true}});

                    // free signalTaskStart from earlier when m_scrollPosition is saved
                    this.signalTaskEnd(); 
                    this.m_scrollPosition = null;
                }
                return;
            }
        }

        if (coord === undefined)
        {
            // legacy scrollTop attribute
            y = this.GetOption("scrollTop");
        }

        scrollTop = scroller.scrollTop;
        // check if only x updated
        if ((!isNaN(x) && isNaN(y)) || (!isNaN(x) && y == scrollTop && x != this._getScrollX(scroller, x)))
        {
            this._setScrollX(scroller, x);
            scrollPosition = this.GetOption("scrollPosition");
            x = this._getScrollX(scroller);
            newScrollPosition = {'x': x, 'y': scrollPosition['y'], 'index': scrollPosition['index'], 'key': scrollPosition['key'], 'offsetX': x, 'offsetY': scrollPosition['offsetY']};
            if (scrollPosition['parent'])
            {
                newScrollPosition['parent'] = scrollPosition['parent'];
            }
            this.SetOption('scrollPosition', newScrollPosition, {'_context': {originalEvent: null, internalSet: true}});
        }
        else if (y != scrollTop)
        {
            // flag it so that handleScroll won't do anything
            this._setScrollY(scroller, y);
            if (!isNaN(x) && x != this._getScrollX(scroller, x))
            {
                this._setScrollX(scroller, x);                
            }

            // checks if further scrolling is needed
            scrollTop = scroller.scrollTop;
            // cannot use scrollTop == y, as browser sub-pixel could be off by < 1px
            if (Math.abs(scrollTop - y) >= 1 && this.m_contentHandler.hasMoreToFetch && this.m_contentHandler.hasMoreToFetch())
            {
                if (this.m_scrollPosition == null)
                {
                    // we don't need to signalTaskStart again if we are already in one
                    this.signalTaskStart("Scroll position needs to resolve further");
                }
                this.m_scrollAndFetch = true;

                // yes, save the scrollPosition to set and bail
                this.m_scrollPosition = position;
                return;
            }
            else
            {
                // ok to update scrollPosition option
                this.SetOption('scrollPosition', this._getCurrentScrollPosition(scrollTop), {'_context': {originalEvent: null, internalSet: true}});
            }
        }
        else
        {
            // if x and y is present, but position value is not complete, get it
            if (position && (position['key'] == null || isNaN(position['index'])))
            {
                // ok to update scrollPosition option
                this.SetOption('scrollPosition', this._getCurrentScrollPosition(scrollTop), {'_context': {originalEvent: null, internalSet: true}});
            }
        }

        if (this.m_scrollPosition != null)
        {
            // free signalTaskStart from earlier when m_scrollPosition is saved
            this.signalTaskEnd(); 
            this.m_scrollPosition = null;
        }
    },

    /**
     * When an item is updated, if the item happens to be the current item, the active element needs to be updated.
     * @protected
     */
    restoreCurrentItem: function()
    {
        var current = this.GetOption("currentItem");
        if (current != null)
        {
            var elem = this.FindElementByKey(current);
            // make sure item is focusable also
            if (elem != null && !this.SkipFocus($(elem)))
            {                    
                this.ActiveAndFocus($(elem), null);
            }
        }
    },

    /**
     * Called by content handler to reset the state of ListView
     * @private
     */
    ClearCache: function()
    {
        // clear any element dependent cache
        this.m_items = null;
        this.m_groupItems = null;
    },

    /**
     * Utility method to start animation
     * @param {Element} elem element to animate
     * @param {string} action the animation action
     * @param {Object=} effect optional animation effect, if not specified then it will be derived based on action
     * @return {Promise} the promise which will be resolve when animation ends
     * @protected
     */
    StartAnimation: function(elem, action, effect)
    {
        if (effect == null)
        {
            effect = this.getAnimationEffect(action);
        }
        return oj.AnimationUtils.startAnimation(elem, action, effect, this.ojContext);
    },

    /********************* context menu methods *****************/
    /**
     * @param {!Object} menu The JET Menu to open as a context menu.  Always non-<code class="prettyprint">null</code>.
     * @param {!Event} event What triggered the menu launch.  Always non-<code class="prettyprint">null</code>.
     * @param {string} eventType "mouse", "touch", or "keyboard".  Never <code class="prettyprint">null</code>.
     * Invoked by widget
     */
    notifyContextMenuGesture: function(menu, event, eventType)
    {
        var parent, openOptions, launcher;

        // first check if we are invoking on an editable or clickable element If so bail
        if (this.IsNodeEditableOrClickable($(event['target'])))
        {
            return false;
        }

        // set the item right click on active
        parent = $(event['target']).closest("."+this.getItemElementStyleClass());       
        
        // prepare the context menu that have listview specific menu items
        this.PrepareContextMenu(parent);
        
        if (parent.length > 0 && !this.SkipFocus($(parent[0])))
        {
            this.SetCurrentItem($(parent[0]), null);
        }

        launcher = this.element;
        
        //When user right click on disabled item(non-focusable), this.m_active 
        //will not be updated to disabled item so explicitly setting the launcher from event.target.
        if(event.button === 2) 
        { 
            launcher = this._findItem($(event.target));
        } 
        else if (this.m_active != null)
        {
            launcher = this.m_active['elem'];
        }
        
        openOptions = {"launcher": launcher, "initialFocus": "menu"};

        if (eventType === "keyboard")
        {
            openOptions["position"] = {"my": "start top", "at": "start bottom", "of": this.m_active['elem']};
        }

        this.ojContext._OpenContextMenu(event, eventType, openOptions);
    },
    
    /**
     * Decorates or prepares context menu items. Navlist need overrides this to decorate remove menu item.
     * @param {jQuery} item  Item
     * @protected
     * @ignore
     */
    PrepareContextMenu: function(item)
    {
        if (this.m_dndContext != null)
        {
            this.m_dndContext.prepareContextMenu(this.ojContext._GetContextMenu());
        }
    },
    /**
     * Override helper for NavList to override checks for whether a node is editable or clickable.
     * @param {jQuery} node  Node
     * @return {boolean} true or false
     * @protected
     */
    IsElementEditableOrClickable: function(node)
    {
        var nodeName = node.prop("nodeName");
        return (nodeName.match(/^INPUT|SELECT|OPTION|BUTTON|^A\b|TEXTAREA/) != null);
    },

    /**
     * Return whether the node is editable or clickable.  Go up the parent node as needed.
     * @param {jQuery} node  Node
     * @return {boolean} true or false
     * @protected	 
     */
    IsNodeEditableOrClickable: function(node)
    {
        var nodeName;

        while (null != node && node[0] != this.element[0] && (nodeName = node.prop("nodeName")) != "LI")
        {
            // If the node is a text node, move up the hierarchy to only operate on elements
            // (on at least the mobile platforms, the node may be a text node)
            if (node[0].nodeType == 3) // 3 is Node.TEXT_NODE
            {
                node = node.parent();
                continue;
            }

            var tabIndex = node.attr('tabIndex');
            // listview overrides the tab index, so we should check if the data-oj-tabindex is populated
            var origTabIndex = node.attr('data-oj-tabindex');

            if (tabIndex != null && tabIndex >= 0  && !node.hasClass(this.getFocusedElementStyleClass()) && !node.hasClass("oj-listview-cell-element"))
            {
                return true;
            }
            else if (this.IsElementEditableOrClickable(node))
            {
                // ignore elements with tabIndex == -1
                if (tabIndex != -1 || origTabIndex != -1)
                {
                    return true;
                }
            }
            node = node.parent();
        }
        return false;
    },

    /********************* focusable/editable element related methods *****************/
    /**
     * Make all tabbable elements within the specified cell un-tabbable
     * @param {Element} element
     * @param {boolean=} excludeActiveElement see inline comment for details
     * @private
     */
    _disableAllTabbableElements: function(element, excludeActiveElement)
    {
        var elem, selector, tabIndex;

        elem = $(element);

        // if it's a group item, inspect the direct div only so it will skip all children
        if (!elem.hasClass(this.getItemStyleClass()))
        {
            elem = $(elem.get(0).firstElementChild);            
        }

        // a group cell could have contained a cell element, which should be skip also
        if (elem.children().first().hasClass("oj-listview-cell-element"))
        {
            elem = $(elem.get(0).firstElementChild);            
        }

        // should exclude non-visible elements, but doing a visible check here causes re-layout and since it's done 
        // on every item on render, it becomes expensive.  Do the filter later in enableTabbableElements, which is only 
        // triggered by entering actionable mode.
        selector = "a, input, select, textarea, button, object, .oj-component-initnode, [tabIndex]";
        elem.find(selector).each(function(index)
        {
            $(this).removeAttr("data-first").removeAttr("data-last");

            tabIndex = parseInt($(this).attr("tabIndex"), 10);
            // when disabling elements before/after current item, exclude the activeElement
            // otherwise tab out of component won't work correctly
            // setting tab index to -1 will be taken care of by the component on blur anyways
            if ((isNaN(tabIndex) || tabIndex >= 0) && (excludeActiveElement === undefined || this != document.activeElement))
            {
                $(this).attr("tabIndex", -1)
                       .attr("data-tabmod", isNaN(tabIndex) ? -1 : tabIndex);
            }
        });
    },

    /**
     * Make all tabbable elements before and include current item un-tabbable
     * @param {Element} item
     * @private
     */
    _disableAllTabbableElementsBeforeItem: function(item)
    {
        var items, index, i;

        items = this._getItemsCache();
        index = items.index(item);
        // if -1 it will just bail
        for (i=0; i<=index; i++)
        {
            this._disableAllTabbableElements(items[i], true);
        }
    },

    /**
     * Make all tabbable elements after and include current item un-tabbable
     * @param {Element} item
     * @private
     */
    _disableAllTabbableElementsAfterItem: function(item)
    {
        var items, index, i;

        items = this._getItemsCache();
        index = items.index(item);
        if (index == -1)
        {
            return;
        }

        for (i=index; i<=items.length-1; i++)
        {
            this._disableAllTabbableElements(items[i], true);
        }
    },

    /**
     * Make all previously tabbable elements within the element tabbable again
     * @param {Element} elem
     * @private
     */
    _enableAllTabbableElements: function(elem)
    {
        var elems, tabIndex;

        elems = $(elem).find("[data-tabmod]").each(function(index)
                {
                    tabIndex = parseInt($(this).attr("data-tabmod"), 10);    
                    $(this).removeAttr("data-tabmod");
                    // restore tabIndex as needed
                    if (tabIndex == -1)
                    {
                        $(this).removeAttr("tabIndex");
                    }
                    else
                    {
                        $(this).attr("tabIndex", tabIndex);
                    }
                });

        // mark first and last tabbable element for fast retrieval later
        elems = elems.filter(":visible");
        elems.first().attr("data-first", "true");
        elems.last().attr("data-last", "true");
    },

    /**
     * Cleanup any attributes added by tabbing logic
     * @param {Element} elem the element to cleanup
     * @private
     */
    _cleanupTabbableElementProperties: function(elem)
    {
        $(elem).find("[data-tabmod]")
               .removeAttr("tabIndex")
               .removeAttr("data-tabmod")
               .removeAttr("data-first")
               .removeAttr("data-last");
    },

    /**
     * Checks whether the element is focusable
     * @param {jQuery} item the item to check
     * @return {boolean} true if the item should not be focusable, false otherwise
     * @protected
     */
    SkipFocus: function(item)
    {
        return item.hasClass("oj-skipfocus");
    },

    /*************************************** Event handlers *****************************/
    /**
     * Returns the focus element, or the root element if nothing inside ListView has focus
     * @return {Element} the focus element inside ListView or the root element
     * @protected
     */
    GetFocusElement : function ()
    {
        var emptyText;

        if (this._getListContainer().hasClass("oj-focus-ancestor"))
        {
            // find the focus item
            if (this.m_active)
            {
                return this.getFocusItem(this.m_active['elem'])[0];          
            }
            else
            {
                // empty text could have focus                
                emptyText = this.element.children("."+this.getEmptyTextStyleClass()).first();
                if (emptyText.length > 0 && emptyText.attr("tabIndex") == 0)
                {
                    return emptyText[0];
                }
            }
        }

        return this.element[0];
    },

    /**
     * Determine whether the event is triggered by interaction with element inside ListView
     * Note that Firefox 48 does not support relatedTarget on blur event, see
     * _supportRelatedTargetOnBlur method
     * @param {Event} event the focus or blur event
     * @return {boolean} true if focus/blur is triggered by interaction with element within listview, false otherwise.
     * @private
     */
    _isFocusBlurTriggeredByDescendent: function(event)
    {
        if (event.relatedTarget === undefined)
            return true;

        if (event.relatedTarget == null || !$.contains(this.ojContext.element.get(0), /** @type {Element} */ (event.relatedTarget)))
            return false;
        else
            return true;
    },

    /**
     * Handler for focus event
     * @param {Event} event the focus event
     * @protected
     */
    HandleFocus: function(event)
    {
        var item;

        this._getListContainer().addClass("oj-focus-ancestor");

        // first time tab into listview, focus on first item
        if (this.m_active == null)
        {
            // checks whether there's pending click to active, and the focus target is not inside any item (if it is the focus will shift to that item) or it's a unfocusable item
            item = this._findItem($(event.target));
            if (!this.m_preActive && !this._isFocusBlurTriggeredByDescendent(event) && (item == null || this.SkipFocus(item)))
            {
                this._initFocus(event);
            }
        }
        else
        {
            // focus could be caused by pending click to active
            // do not do this on iOS or Android, otherwise VO/talkback will not work correctly
            // Only one exception is when ever root node gets focus we should highlight active element
            // otherwise vo doesn't follow the focus. ex: when offcanvas is opened,
            // focus will be moved to root node
            if (!this.m_preActive && event.target === this.ojContext.element[0]
                 && !this._isFocusBlurTriggeredByDescendent(event))
            {
                this.HighlightActive();
                this._focusItem(this.m_active['elem']);
            }

            // remove tab index from root and restore tab index on focus item
            this.RemoveRootElementTabIndex();
            this._setTabIndex(this.m_active['elem']);
        }
    },

    /**
     * Initialize focus by finding the first focusable item and set focus on it
     * @private
     */
    _initFocus: function(event)
    {
        var items, i, item, emptyText;

        items = this._getItemsCache();
        for (i=0; i<items.length; i++)
        {
            item = $(items[i]);
            // make sure item can receive focus
            if (!this.SkipFocus(item))
            {
                this.SetCurrentItem(item, event);
                break;
            }
        }

        if (items.length == 0)
        {
            // we need to focus on empty text
            emptyText = this.element.children("."+this.getEmptyTextStyleClass()).first();
            if (emptyText.length > 0)
            {
                emptyText.attr("tabIndex", 0);
                emptyText.focus();

                this.RemoveRootElementTabIndex();
            }
        }
    },

    /**
     * Handler for focus out event
     * @param {Event} event the focusout event
     * @protected
     */
    HandleFocusOut: function(event)
    {
        this.HandleBlur(event);
    },

    /**
     * Checks whether the browser supports relatedTarget field for blur event
     * @return {boolean} true if supported, false otherwise
     * @private
     */
    _supportRelatedTargetOnBlur: function()
    {
        var agent = oj.AgentUtils.getAgentInfo();
        if (agent['browser'] == oj.AgentUtils.BROWSER.FIREFOX && parseInt(agent['browserVersion'], 10) < 48)
        {
            return false;
        }

        return true;
    },

    /**
     * Detects whether this is a double blur event fired by IE11
     * @param {Event} event the blur event
     * @private
     */
    _isExtraBlurEvent: function(event)
    {
        var agent = oj.AgentUtils.getAgentInfo();
        if (event.relatedTarget == null && agent['browser'] == oj.AgentUtils.BROWSER.IE && event.target == this.ojContext.element.get(0))
        {
            return true;
        }

        return false;
    },

    /**
     * Handler for blur event
     * @param {Event} event the blur event
     * @protected
     */
    HandleBlur: function(event)
    {
        var emptyText;

        // NOTE that prior to a fix in Firefox 48, relatedTarget is always null
        // just bail in case if it wasn't supported
        // NOTE also IE11 fires an extra blur event with null relatedTarget, should bail as well
        if (!this._supportRelatedTargetOnBlur() || this._isExtraBlurEvent(event))
        {
            return;
        }

        // event.relatedTarget would be null if focus out of page
        // the other check is to make sure the blur is not caused by shifting focus within listview
        if (!this._isFocusBlurTriggeredByDescendent(event) && !this.m_preActive)
        {
            this._doBlur();
        }

        // remove focus class on blur of expand/collapse icon
        if (this._isExpandCollapseIcon(event.target))
        {
            this._focusOutHandler($(event.target));
        }
    },

    /**
     * @private
     */
    _doBlur: function()
    {
        var emptyText;

        this._getListContainer().removeClass("oj-focus-ancestor");
        this.UnhighlightActive();

        // remove tab index from focus item and restore tab index on list
        // and remove any aria-labelled by set by card navigation
        if (this.m_active != null)
        {
            this._resetTabIndex(this.m_active['elem']);
            this._removeSkipItemAriaLabel(this.m_active['elem']);
        }
        else
        {
            emptyText = this.element.children("."+this.getEmptyTextStyleClass()).first();
            if (emptyText.length > 0)
            {
                emptyText.removeAttr("tabIndex");
            }
        }
        this.SetRootElementTabIndex();
    },

    /**
     * Event handler for when user exits an item in list
     * @param {Event} event mouseout event 
     * @private
     */
    _handleMouseOut: function(event)
    {
        var item = this.FindItem(event.target);
        if (item != null)
        {
            this.m_hoverItem = null;
            this._unhighlightElem(item, "oj-hover");
        }
    },

    /**
     * Event handler for when user hovers over list
     * @param {Event} event mouseover event 
     * @private
     */
    _handleMouseOver: function(event)
    {
        // do this for real mouse enters, but not 300ms after a tap
        if (this._recentTouch())
        {
            return;
        }

        var item = this.FindItem(event.target);        
        if (item != null && !this.SkipFocus(item))
        {
            // have to remember it so we can clear it when listview is detached from DOM
            this.m_hoverItem = item;
            this._highlightElem(item, "oj-hover");         
        }
    },

    _recentTouch: function()
    {
        return Date.now() - this._lastTouch < 500; // must be at least 300 for the "300ms" delay
    },

    /**
     * Event handler for when user press down a key
     * @param {Event} event keydown event
     * @protected
     */
    HandleKeyDown: function(event)
    {
        var keyCode, current, processed;

        if (this.isExpandable())
        {
            keyCode = event.keyCode;
            if (keyCode === this.LEFT_KEY || keyCode === this.RIGHT_KEY)
            {
                current = this.m_active['elem'];
                if (keyCode === this.LEFT_KEY)
                {
                    if (this.GetState(current) == this.STATE_EXPANDED)
                    {
                        this.CollapseItem(current, event, true, this.m_active['key'], true, true);
                        return;
                    }
                }
                else
                {
                    if (this.GetState(current) == this.STATE_COLLAPSED)
                    {
                        this.ExpandItem(current, event, true, this.m_active['key'], true, true, true);
                        return;
                    }
                }
            }
        }

        processed = this.HandleSelectionOrActiveKeyDown(event);

        // dnd
        if (this.m_dndContext != null)
        {
            processed = processed || this.m_dndContext.HandleKeyDown(event);
        }

        if (processed === true)
        {
            event.preventDefault();
        }

        this.m_keyProcessed = processed;
    },

    /**
     * Event handler for whenever press up occurred
     * @param {Event} event keyup event
     * @protected
     */
    HandleKeyUp: function(event)
    {
        // popup process esc key on keyup so we have to stop keyup from bubbling
        if (event.keyCode == this.ESC_KEY && this.m_keyProcessed)
        {
            event.stopPropagation();
        }
        this.m_keyProcessed = undefined;
    },

    /**
     * @private
     */
    _handleMouseUpOrPanMove: function(event)
    {
        // unhighlight item that got focus in mousedown
        if (this.m_preActiveItem)
        {
            this._unhighlightElem(this.m_preActiveItem, "oj-focus");
        }

        // dnd
        if (this.m_dndContext != null)
        {
            this.m_dndContext._unsetDraggable($(event.target));
        }
    },

    /**
     * Event handler for when mouse down or touch start anywhere in the list
     * @param {Event} event mousedown or touchstart event
     * @protected
     */
    HandleMouseDownOrTouchStart: function(event)
    {
        var item, target;

        // click on item
        target = $(event.target);

        // dnd
        if (this.m_dndContext != null)
        {
            this.m_dndContext._setDraggable(target);
        }

        // click on item, explicitly pass true on findItem
        // so that it will return non-null value if clickthrough disabled element is encountered
        item = this._findItem(target, true);
        // we'll still need to set the flag so that the focus do not shift
        if (item != null && this._isClickthroughDisabled(item))
        {
            this.m_preActive = true;
            item = null;
        }

        if (item == null || item.length == 0 || this.SkipFocus(item) || target.hasClass("oj-listview-drag-handle"))
        {
            // one of the following happened:
            // 1) can't find item 
            // 2) item cannot be focus
            // 3) target is an oj-component
            // 4) target or one of its ancestors has the oj-clickthrough-disabled marker class
            // 5) target is the drag handle
            return;
        }

        this.m_preActive = true;

        // make sure listview has focus
        if (!this._getListContainer().hasClass("oj-focus-ancestor"))
        {
            this._getListContainer().addClass("oj-focus-ancestor");
        }

        // we'll need to remove focus in case the actual focus item is different
        this.m_preActiveItem = item;

        // apply focus
        this._highlightElem(item, "oj-focus");

        // focus on item, we need to do it on mousedown instead of click otherwise click handler will
        // steal focus from popup and causes it to close prematurely
        this._makeFocusable(item);
        // checks whether focus is already inside some in item, if it is don't try to steal focus away from it (combobox)
        if (!item.get(0).contains(document.activeElement))
        {
            this._focusItem(item);
        }

        // make sure ul is not tabbable
        this.RemoveRootElementTabIndex();
        // reset tab index must be done after focusing another item
        if (this.m_active != null && this.m_active['elem'].get(0) != item.get(0))
        {
            this._resetTabIndex(this.m_active['elem']);
        }

        // need this on touchend
        if (event.originalEvent.touches && event.originalEvent.touches.length > 0)
        {
            this.m_touchPos = {x: event.originalEvent.changedTouches[0].pageX, y: event.originalEvent.changedTouches[0].pageY};
        }
    },

    /**
     * Event handler for when touch end/cancel happened
     * @param {Event} event touchend or touchcancel event
     * @protected
     */
    HandleTouchEndOrCancel: function(event)
    {
        var offset, action = "pointerUp", effect, elem, groupItem;

        // unhighlight item that got focus in touchstart
        if (this.m_preActiveItem != null)
        {
            this._unhighlightElem(this.m_preActiveItem, "oj-focus");

            // start ripple effect
            if (this.m_touchPos != null)
            {
                offset = this.m_preActiveItem.offset();

                // find where to start the ripple effect based on touch location
                effect = this.getAnimationEffect(action);
                effect['offsetX'] = ((this.m_touchPos.x) - offset['left']) + 'px';
                effect['offsetY'] = ((this.m_touchPos.y) - offset['top']) + 'px';
                groupItem = this.m_preActiveItem.children('.' + this.getGroupItemStyleClass());
                if (groupItem.length > 0)
                {
                  elem = /** @type {Element} */ (groupItem.get(0));
                } 
                else 
                {
                  elem = /** @type {Element} */ (this.getFocusItem(this.m_preActiveItem).get(0));
                }
                // we don't really care when animation ends
                this.StartAnimation(elem, action, effect);

                this.m_touchPos = null;
            }
        }

        // need this so that on mouse over handler would not apply the styles if the last touch was within the last n ms
        this._lastTouch = Date.now();
        this._handleMouseOut(event);
    },

    /**
     * Enters actionable mode
     * @private
     */
    _enterActionableMode: function()
    {
        var current, first;

        current = this.m_active['elem'];

        // in case content has been updated under the cover
        this._disableAllTabbableElements(current);

        // re-enable all tabbable elements
        this._enableAllTabbableElements(current);                    

        // only go into actionable mode if there is something to focus
        first = current.find("[data-first]");
        if (first.length > 0)
        {
            this._setActionableMode(true);     
        }
    },

    /**
     * Exits actionable mode
     * @private
     */
    _exitActionableMode: function()
    {
        this._setActionableMode(false);

        // disable all tabbable elements in the item again
        this._disableAllTabbableElements(this.m_active['elem']);
    },

    /**
     * Event handler for when mouse click anywhere in the list
     * @param {Event} event mouseclick event
     * @protected
     */
    HandleMouseClick: function(event)
    {
        var collapseIconClass, expandIconClass, groupItemClass, target, item;

        //only perform events on left mouse, (right in rtl culture)
        if (event.button === 0)
        {
            collapseIconClass = this.getCollapseIconStyleClass();
            expandIconClass = this.getExpandIconStyleClass();
            groupItemClass = this.getGroupItemStyleClass();
            target = $(event.target);
            if (target.hasClass(expandIconClass))
            {
                this._collapse(event);
                event.preventDefault();
            }
            else if (target.hasClass(collapseIconClass))
            {
                this._expand(event);
                event.preventDefault();
            }
            else
            {
                // click on item
                item = this._findItem(target);
                if (item == null || item.length == 0 || this.SkipFocus(item))
                {
                    // one of the following happened:
                    // 1) can't find item 
                    // 2) item cannot be focus
                    // 3) target is an oj-component
                    // 4) target or one of its ancestors has the oj-clickthrough-disabled marker class
                    return;
                }

                if (this._isActionableMode() && this.m_active != null && this.m_active['elem'].get(0) != item.get(0))
                {
                    // click on item other than current focus item should exit actionable mode
                    this._exitActionableMode();
                }

                // make sure listview has focus
                if (!this._getListContainer().hasClass("oj-focus-ancestor"))
                {
                    this._getListContainer().addClass("oj-focus-ancestor");
                }

                // check if selection is enabled
                if (this._isSelectionEnabled() && this.IsSelectable(item[0]))
                {        
                    if (this._isTouchSupport())
                    {
                        this._handleTouchSelection(item, event);
                    }
                    else
                    {
                        this.HandleClickSelection(item, event);
                    }

                    // if user hits the padding part of item, since LI does not have tabindex anymore, item will not get focus
                    if (this.ShouldUseGridRole() && event.target == item.get(0))
                    {
                        this._focusItem(item);
                    }

                    // need to make sure every item in the selection have the draggable cursor
                    if (this._shouldDragSelectedItems())
                    {
                        this.m_dndContext.setSelectionDraggable();
                    }
                }
                else
                {
                    // if selection is disable, we'll still need to highlight the active item
                    this.HandleClickActive(item, event);
                }

                // click on input element inside item should trigger actionable mode
                if (this._isInputElement(target.get(0)))
                {
                    this._enterActionableMode();
                    return;
                }

                // clicking on header will expand/collapse item
                if (this.isExpandable() && target.closest("."+groupItemClass))
                {
                    if (this.GetState(item) == this.STATE_COLLAPSED)
                    {
                        this._expand(event);
                    }
                    else if (this.GetState(item) == this.STATE_EXPANDED)
                    {
                        this._collapse(event);
                    }              
                }
            }
        }
    },
    /**
     * Return true if Dnd is supported on selected items only.
     * @private
     * @returns {boolean}
     */
    _shouldDragSelectedItems: function()
    {
        return this.m_dndContext != null && !this.m_dndContext.shouldDragCurrentItem()
    },
    /*********************************** end Event handlers *****************************/

    /*************************************** helper methods *****************************/
    /**
     * Whether touch is supported
     * @return {boolean} true if touch is supported, false otherwise
     * @private
     */
    _isTouchSupport: function()
    {
        return oj.DomUtils.isTouchSupported();
    },

    /**
     * Whether it is non-window touch device (iOS or Android)
     * @return {boolean} true if it is a non-window touch device
     * @private
     */
    _isNonWindowTouch: function()
    {
        return this._isTouchSupport() && oj.AgentUtils.getAgentInfo()['os'] != oj.AgentUtils.OS.WINDOWS;
    },

    /**
     * Returns either the ctrl key or the command key in Mac OS
     * @param {!Object} event
     * @private
     */
    _ctrlEquivalent: function(event)
    {
        return oj.DomUtils.isMetaKeyPressed(event);
    },

    /**
     * Helper method to create subid based on the root element's id
     * @param {string} subId the id to append to the root element id
     * @return {string} the subId to append to the root element id
     * @private
     */
    _createSubId: function(subId)
    {
        var id = this.element.attr("id");
        return [id, subId].join(":");
    },

    /**
     * Find the item element
     * @param {jQuery} elem
     * @return {jQuery|null}
     * @protected
     */
    FindItem: function(elem)
    {
        return $(elem).closest("."+this.getItemElementStyleClass());
    },

    /** 
     * Determine if click should be processed based on the element.
     * @param {jQuery} elem the element to check
     * @return {boolean} returns true if the element contains the special marker class, false otherwise.
     * @private
     */
    _isClickthroughDisabled: function(elem)
    {
        return elem.hasClass("oj-clickthrough-disabled");
    },

    /**
     * Find the item element from target, if target contains the oj-clickthrough-disabled class then returns null.
     * @param {jQuery} target the element to check
     * @param {boolean=} retElemOnClickthroughDisabled optional, set to true to force non-null value to return when 
     *                   clickthrough-disabled is encountered
     * @return {jQuery|null} the item element or null if click through is disabled for this element or one of its ancestors.
     * @private
     */
    _findItem: function(target, retElemOnClickthroughDisabled)
    {
        var current = target;
        while (current.length > 0)
        {
            if (this._isClickthroughDisabled(current))
            {
                if (retElemOnClickthroughDisabled)
                {
                    return current;
                }
                return null;
            }

            if (current.hasClass(this.getItemElementStyleClass()))
            {
                return current;
            }

            current = current.parent();
        }

        return null;
    },

    /**
     * Compute the total top and bottom border width of the list container
     * @return {number} the sum of top and bottom border width of the list container
     * @private
     */
    _getListContainerBorderWidth: function()
    {
        if (this.m_borderWidth == null)
        {
            this.m_borderWidth = parseInt(this._getListContainer().css("border-top-width"), 10) + parseInt(this._getListContainer().css("border-bottom-width"), 10);
        }

        return this.m_borderWidth;
    },

    /**
     * Scroll as needed to make the specified item visible
     * @param {Object} item the item to make visible
     */
    scrollToItem: function(item)
    {
        var key, elem, group;

        key = item['key'];
        if (key == null)
        {
            return;
        }

        elem = this.FindElementByKey(key);
        if (elem == null)
        {
            return;
        }

        if ($(elem).hasClass(this.getItemStyleClass()))
        {
            this._scrollToVisible(elem);
        }
        else
        { 
            // group item
            group = $(elem).children("."+this.getGroupItemStyleClass()).first();
            this._scrollToGroupHeader(group.get(0));
        }
    },

    /**
     * Scroll as needed to make an element visible in the viewport
     * @param {Element} elem the element to make visible
     * @private
     */
    _scrollToVisible: function(elem)
    {
        var top, height, container, containerScrollTop, containerHeight, headerTop, headerHeight, offset = 0, scrollTop;

        top = elem.offsetTop;
        height = elem.offsetHeight;
        container = this._getListContainer()[0];
        containerScrollTop = container.scrollTop;
        containerHeight = container.offsetHeight;

        // if there's sticky header, make sure the elem is not behind it
        if (this.m_groupItemToPin != null)
        {
            headerTop = parseInt(this.m_groupItemToPin.style.top, 10);
            headerHeight = $(this.m_groupItemToPin).outerHeight();
            if ((top <= headerTop && headerTop < top + height))
            {
                offset = ((height + top) - headerTop) / 2;    
            }
            else if ((top >= headerTop && top < headerTop + headerHeight))
            {
                offset = ((headerTop + headerHeight) - top) / 2;  
            }                
        }

        // if it's within viewport do nothing
        if (top >= containerScrollTop && top+height <= containerScrollTop+containerHeight)
        {
            if (offset > 0)
            {
                container.scrollTop = containerScrollTop - offset;
            }
            return;
        }

        // how much need to scroll to see the entire element, and to make sure the element top is always visible
        scrollTop = Math.max(0, Math.min(top - offset, Math.abs((top + height) - containerHeight)));
        if (scrollTop > containerScrollTop)
        {
            scrollTop = scrollTop + this._getListContainerBorderWidth();
        }
        container.scrollTop = scrollTop;        
    },

    /**
     * Get the key associated with an item element
     * @param {Element} elem the item element to retrieve the key
     * @return {Object|null} the key associated with the item element
     * @protected
     */
    GetKey: function(elem)
    {
        return this.m_contentHandler.GetKey(elem);
    },

    /**
     * Get the element associated with a key
     * @param {Object} key the key to retrieve the item element
     * @return {Element|null} the item element associated with the key 
     * @protected
     */
    FindElementByKey: function(key)
    {
        var id;

        if (this.m_keyElemMap != null)
        {
            id = this.m_keyElemMap.get(key);
            if (id != null)
            {
                return document.getElementById(id);
            }
        }

        // ask the content handler
        return this.m_contentHandler.FindElementByKey(key);
    },

    /**
     * Special version of array indexOf to take care of object comparison cases
     * @param {Array} arr the array
     * @param {Object} key the key to find the index for in the array
     * @return {number} the index of the key in the array, or -1 if the key does not exists in the array
     */
    GetIndexOf: function(arr, key)
    {
        var i;
        for (i=0; i<arr.length; i++)
        {
            if (key == arr[i] || oj.Object.compareValues(key, arr[i]))
            {
                return i;
            }
        }

        return -1;
    },

    /**
     * Checks whether element is an expand/collapse icon
     * @param {Element|jQuery} elem the element to check
     * @return {boolean} true if it's an expand/collapse icon, false otherwise
     */
    _isExpandCollapseIcon: function(elem)
    {
        return $(elem).hasClass(this.getExpandIconStyleClass()) || $(elem).hasClass(this.getCollapseIconStyleClass());
    },
    /************************************ end helper methods *****************************/

    /*************************************** Navigation Common **************************/
    /**
     * Determine whether the key code is an arrow key
     * @param {number} keyCode
     * @return {boolean} true if it's an arrow key, false otherwise
     * @protected
     */
    IsArrowKey: function(keyCode)
    {
        if (this.isCardLayout())
        {
            return (keyCode == this.UP_KEY || keyCode == this.DOWN_KEY || keyCode == this.LEFT_KEY || keyCode == this.RIGHT_KEY);
        }
        else
        {
            return (keyCode == this.UP_KEY || keyCode == this.DOWN_KEY);
        }
    },

    /**
     * Retrieve the visible (flattened) items cache, create one if it is null.
     * @return {jQuery} a list of items
     * @private 
     */
    _getItemsCache: function()
    {
        var disclosureStyleClass, selector, isGroup;

        if (this.m_items == null)
        {
            disclosureStyleClass = this.getGroupCollapseStyleClass();
            selector = "."+this.getItemElementStyleClass() + ":visible";
            this.m_items = this.element.find(selector).filter(function(index)
            {
                isGroup = $(this).parent().hasClass(disclosureStyleClass);
                if (isGroup)
                {
                    return !($(this).parent().parent().hasClass("oj-collapsed"));
                } 
                return true;
            });
        }
        return this.m_items;
    },

    /**
     * Handles when navigate to the last item 
     * @param {jQuery} item the item element
     */
    _handleLastItemKeyboardFocus: function(item)
    {
        var next = item.get(0).nextElementSibling;
        if (next == null || !$(next).hasClass(this.getItemElementStyleClass()))
        {
           // it could be the last element of the group, if so, make sure it's the last group
           if (this.m_contentHandler.IsHierarchical() && item.parent().hasClass(this.getGroupStyleClass()))
           {
               if (item.parent().parent().get(0).nextElementSibling != null)
               {
                   // bail if it's not the last group
                   return;
               }
           }

           // it's the last element, check scroll bar to make sure it scrolls all the way to the bottom
           var scroller = this._getScroller();
           if (scroller.scrollTop < scroller.scrollHeight)
           {
               scroller.scrollTop = scroller.scrollHeight;
           }
        }
    },

    /**
     * Handles arrow keys navigation on item
     * @param {number} keyCode description
     * @param {boolean} isExtend
     * @param {Event} event the DOM event causing the arrow keys
     * @protected
     */
    HandleArrowKeys: function(keyCode, isExtend, event)
    {
        var current, processed;

        // ensure that there's no outstanding fetch requests
        if (!this.m_contentHandler.IsReady())
        {
            //act as if processed to prevent page scrolling before fetch done
            return true;
        }

        if (!isExtend || this.m_isNavigate)
        {
            current = this.m_active['elem'];
        }
        else
        {
            current = this.m_selectionFrontier;
        }

        // invoke different function for handling focusing on active item depending on whether selection is enabled
        processed = false;

        switch (keyCode)
        {
            case this.UP_KEY:
                if (this.isCardLayout() && $(current).hasClass(this.getItemStyleClass()))
                {
                    this._gotoItemAbove(current, isExtend, event);
                }
                else
                {
                    this._gotoPrevItem(current, isExtend, event);
                }

                // according to James we should still consume the event even if list view did not perform any action
                processed = true;
                break;
            case this.DOWN_KEY:
                if (this.isCardLayout() && $(current).hasClass(this.getItemStyleClass()))
                {
                    this._gotoItemBelow(current, isExtend, event);
                }
                else
                {
                    this._gotoNextItem(current, isExtend, event);
                }

                // according to James we should still consume the event even if list view did not perform any action
                processed = true;
                break;
            case this.LEFT_KEY:
            case this.RIGHT_KEY:
                if (this.isCardLayout())
                {
                    if (this.ojContext._GetReadingDirection() == "rtl")
                    {
                        keyCode = keyCode == this.LEFT_KEY ? this.RIGHT_KEY : this.LEFT_KEY;
                    }

                    if (keyCode == this.LEFT_KEY) 
                    {
                        this._gotoPrevItem(current, isExtend, event) 
                    }
                    else
                    {
                        this._gotoNextItem(current, isExtend, event);
                    }
                    processed = true;
                }
                break;
	}

        return processed;
    },

    /**
     * Go to the next item in the list
     * @private
     */
    _gotoNextItem: function(current, isExtend, event)
    {
        var items, currentIndex, next;

        items = this._getItemsCache();
        currentIndex = items.index(current);

        currentIndex++;
        if (currentIndex < items.length)
        {
            next = $(items[currentIndex]);
            // make sure it's focusable, otherwise find the next focusable item
            while (this.SkipFocus(next))
            {
                currentIndex++;
                if (currentIndex == items.length)
                {
                    return;
                }
                next = $(items[currentIndex]);
            }

            if (isExtend)
            {
                this._extendSelection(next, event);
                this.m_isNavigate = false;
            }
            else
            {
                this.SetCurrentItem(next, event);
                this.m_isNavigate = true;
            }

            this._handleLastItemKeyboardFocus(next);
        }
    },

    /**
     * Go to the previous item in the list
     * @private
     */
    _gotoPrevItem: function(current, isExtend, event)
    {
        var items, currentIndex, prev;

        items = this._getItemsCache();
        currentIndex = items.index(current);

        currentIndex--;
        if (currentIndex >= 0)
        {
            prev = $(items[currentIndex]);
            // make sure it's focusable, otherwise find the next focusable item
            while (this.SkipFocus(prev))
            {
                currentIndex--;
                if (currentIndex < 0)
                {
                    return;
                }
                prev = $(items[currentIndex]);
            }

            if (isExtend)
            {
                this._extendSelection(prev, event);
                this.m_isNavigate = false;
            }
            else
            {
                this.SetCurrentItem(prev, event);
                this.m_isNavigate = true;
            }
        }
    },

    /**
     * Calculate the number of the columns in this group
     * @param {jQuery} children the children iterator
     * @return {number} the number of columns
     * @private
     */
    _getColumnCount: function(children)
    {
        var count, offsetTop, currentOffsetTop;

        children.each(function(index)
        {
            offsetTop = this.offsetTop;

            if (currentOffsetTop === undefined)
            {
                currentOffsetTop = offsetTop;
            }
            else if (currentOffsetTop != offsetTop)
            {
                // on a different row, return immediately
                return false;
            }

            count = index;
        });

        return count + 1;
    },

    /**
     * Inform screen reader that X number of items have been skipped during up/down arrow key navigation
     * @param {jQuery} next the item to navigate to
     * @param {number} count number of items skipped
     * @private
     */
    _updateSkipItemAriaLabel: function(next, count)
    {
        var id, root, focusElem;

        id = this._createSubId("extra_info");
        if (this.m_skipAriaLabelText == null)
        {
            root = $(document.createElement("div"));
            root.addClass("oj-helper-hidden-accessible")
                .attr("id", id); 
            this._getListContainer().append(root); // @HTMLUpdateOK
            this.m_skipAriaLabelText = root;
        }
        this.m_skipAriaLabelText.text(this.ojContext.getTranslatedString("accessibleNavigateSkipItems", {'numSkip': count}));        

        focusElem = this.getFocusItem(next);
        // make sure it has an id for aria-labelledby
        focusElem.uniqueId()
                 .attr("aria-labelledby", id+" "+focusElem.prop("id"));
    },

    /**
     * Undo what _updateSkipItemAriaLabel did to active element
     * @param {jQuery} current the item to remove aria property from
     * @private
     */
    _removeSkipItemAriaLabel: function(current)
    {
        var focusElem = this.getFocusItem(current);
        if (focusElem.length > 0)
        {
            focusElem.get(0).removeAttribute("aria-labelledby");
        }
    },

    /**
     * Go to the item above the current item in the list
     * @private
     */
    _gotoItemAbove: function(current, isExtend, event)
    {
        var parent, items, numOfCols, index, aboveIndex, above, numOfItemSkip;

        // if it's a group, just go to the previous item (which would be the last focusable item in the previous group)
        if (!current.hasClass(this.getItemStyleClass()))
        {
            this._gotoPrevItem(current);
            return;
        }

        parent = current.parent();
        items = parent.children("li."+this.getItemStyleClass())
        numOfCols = this._getColumnCount(items);
        index = items.index(current);
        aboveIndex = index - numOfCols;
        if (aboveIndex < 0)
        {
            // go to header, or stop if there's no header
            if (parent.hasClass(this.getGroupStyleClass()))
            {
                above = parent.parent();
                numOfItemSkip = index;
            }
        }
        else
        {
            above = $(items.get(aboveIndex));
            numOfItemSkip = numOfCols - 1;
        }

        if (above.length > 0)
        {
            // make sure it's focusable, otherwise no-op
            if (this.SkipFocus(above))
            {
                return;
            }

            if (isExtend)
            {
                this._extendSelection(above, event);
                this.m_isNavigate = false;
            }
            else
            {
                if (numOfItemSkip != undefined && numOfItemSkip > 0)
                {
                    this._updateSkipItemAriaLabel(above, numOfItemSkip);
                }
                this.SetCurrentItem(above, event);
                this.m_isNavigate = true;
            }
        }
    },

    /**
     * Go to the item below the current item in the list
     * @private
     */
    _gotoItemBelow: function(current, isExtend, event)
    {
        var parent, items, numOfCols, index, belowIndex, numOfRows, below, numOfItemSkip;

        // if it's a group, just go to the next item (which would be the first focusable item in the next group)
        if (!current.hasClass(this.getItemStyleClass()))
        {
            this._gotoNextItem(current);
            return;
        }

        parent = current.parent();
        items = parent.children("li."+this.getItemStyleClass())
        numOfCols = this._getColumnCount(items);
        index = items.index(current);
        belowIndex = index + numOfCols;
        if (belowIndex >= items.length)
        {
            numOfRows = Math.ceil(items.length/numOfCols);
            numOfItemSkip = (items.length - 1) - index;
            // if the current item is not on the last row, then go to the last item within group
            if (index < Math.max(0, (numOfRows - 1)*numOfCols))
            {
                below = items.last();
                // skip one less since going to an item within the same group
                numOfItemSkip--;
            }
            else
            {
                // go to header, or stop if there's no header
                if (parent.hasClass(this.getGroupStyleClass()))
                {
                    below = parent.parent().next("li."+this.getItemElementStyleClass());
                }
            }
        }
        else
        {
            below = $(items.get(belowIndex));
            numOfItemSkip = numOfCols - 1;
        }

        if (below.length > 0)
        {
            // make sure it's focusable, otherwise no-op
            if (this.SkipFocus(below))
            {
                return;
            }

            if (isExtend)
            {
                this._extendSelection(below, event);
                this.m_isNavigate = false;
            }
            else
            {
                if (numOfItemSkip != undefined && numOfItemSkip > 0)
                {
                    this._updateSkipItemAriaLabel(below, numOfItemSkip);
                }
                this.SetCurrentItem(below, event);
                this.m_isNavigate = true;
            }
        }
    },

    /**
     * Determine if the data grid is in actionable mode.
     * @return {boolean} true if the data grid is in actionable mode, false otherwise.
     * @private
     */
    _isActionableMode: function()
    {
        return (this.m_keyMode == "actionable");
    },

    /**
     * Sets whether the data grid is in actionable mode
     * @param {boolean} flag true to set grid to actionable mode, false otherwise
     * @private
     */
    _setActionableMode: function(flag)
    {
        this.m_keyMode = flag ? "actionable" : "navigation";

        if (!flag)
        {
            // focus should be shift back to active descendant container
            this.element[0].focus();
        }
    },
    /************************************ end Navigation Common **************************/

    /************************************ Active item ******************************/
    /**
     * Retrieve the focus element
     * @param {jQuery} item the list item
     * @return {jQuery} the focus element
     * @private
     */
    getFocusItem: function(item)
    {
        var cell;

        if (!item.hasClass(this.getFocusedElementStyleClass()))
        {
            return $(item.find('.'+this.getFocusedElementStyleClass()).first());
        }
        else
        {
            if (this.ShouldUseGridRole() && item.attr("role") == "row")
            {
                cell = item.children(".oj-listview-cell-element").first();
                return cell.length == 0 ? item.children().first() : cell;
            }
            else
            {
                return item;
            }
        }
    },

    /**
     * Sets the tab index attribute of the root element
     * To be change by NavList
     * @protected
     */
    SetRootElementTabIndex: function()
    {
        this.element.attr("tabIndex", 0);
    },

    /**
     * Removes the tab index attribute of the root element
     * To be change by NavList
     * @protected
     */
    RemoveRootElementTabIndex: function()
    {
        this.element.removeAttr("tabIndex");
    },

    /**
     * Sets the tab index on focus item
     * @param {jQuery} item the focus item
     * @private
     */
    _setTabIndex: function(item)
    {
        // note that page author should not set any tabindex on the item
        this.getFocusItem(item).attr("tabIndex", 0);
    },

    /**
     * Resets the tab index set on focus item
     * @param {jQuery} item the focus item
     * @private
     */
    _resetTabIndex: function(item)
    {
        var removeAttr, isGroupItem;

        removeAttr = true;
        if (item.attr("role") === "presentation")
        {
            removeAttr = false;
        }

        item = this.getFocusItem(item);
        if (removeAttr)
        {
            item.removeAttr("tabIndex");
        }
        else
        {
            item.attr("tabIndex", -1);
        }

    },

    /**
     * Make an item focusable
     * @param {jQuery} item the item to focus
     * @private
     */
    _makeFocusable: function(item)
    {
        this._setTabIndex(item);
    },

    /**
     * Determine the only focusable element inside an item, if the item does not have any or have 
     * more than one focusable element, then just return the item.
     * @param {jQuery} item the list item
     * @return {jQuery} see above for what's get returned
     * @private
     */
    getSingleFocusableElement: function(item)
    {
        var selector, childElements;

        selector = "a, input, select, textarea, button";
        childElements = item.children(selector);

        if(childElements.length === 1 && // check for only one focusbale child
           childElements.first().find(selector).length === 0)
        {
            //check to ensure no nested focusable elements.
            return childElements.first();
        }
        return item;
    },

    /**
     * Sets the selection option with a new value
     * @param {Object} newValue the new value for selection option
     * @param {Event|null} event the DOM event
     * @param {Element} currentElem the current item DOM element
     * @private
     */
    _setCurrentItemOption: function(newValue, event, currentElem)
    {
        var extra = {'item': this.ojContext._IsCustomElement() ? currentElem : $(currentElem)}; 
        this.SetOption("currentItem", newValue, {'_context': {originalEvent: event, internalSet: true, extraData: extra}, 'changed': true});
    },

    /**
     * Sets the active item
     * @param {jQuery} item the item to set as active
     * @param {Event} event the event that triggers set active
     * @param {boolean=} skipFocus true if to skip focusing the item
     * @return {boolean} true if item becomes active, false for all other cases
     * @private
     */
    _setActive: function(item, event, skipFocus)
    {
        var elem, key, ui, cancelled, active;

        // set key info
        if (item != null)
        {
            elem = item[0];
            key = this.GetKey(elem);

            if ((this.m_active == null) || key != this.m_active['key'])
            {
                // fire beforecurrentItem
                ui = {'key': key, 'item': item};
                if (this.m_active != null)
                {
                    ui['previousKey'] = this.m_active['key'];
                    ui['previousItem'] = this.m_active['elem'];
                    // for touch, remove draggable when active item changed
                    if (this._shouldDragSelectedItems() && this._isTouchSupport())
                    {
                        this.m_dndContext._unsetDraggable(ui['previousItem']);                    
                    }

                    // remove aria-labelledby set by arrow key navigation
                    this._removeSkipItemAriaLabel(ui['previousItem']);
                }

                cancelled = !this.Trigger("beforeCurrentItem", event, ui);
                if (cancelled) 
                {
                    return false;
                }

                active = {'key': key, 'elem': item};
                this.m_active = active;

                // for touch, set draggable when item becomes active
                if (this._shouldDragSelectedItems() && this._isTouchSupport())
                {
                    this.m_dndContext._setDraggable(item);
                }

                // update tab index
                if (skipFocus === undefined || !skipFocus)
                {
                    // make item focusable
                    this._makeFocusable(item);

                    // style should be updated before currentItem change event is fired
                    this.HighlightActive();

                    // focus on it only for non-click events or when done programmatically
                    // the issue is we can't shift focus on click since it will steal focus from the popups
                    // the focusing on item is done already on mousedown
                    if (event == null || (event.originalEvent && event.originalEvent.type != "click"))
                    {
                        this._focusItem(item);
                    }

                    // reset tabindex of previous focus item, note this has to be done after focusing on a new item
                    // because in Chrome 57, resetting tabindex on a focus item will cause a blur event
                    // note also it seems this only happens when removing tabindex, if you set it to -1 this does not happen
                    this.RemoveRootElementTabIndex();
                    if (ui['previousItem'])
                    {
                        this._resetTabIndex(ui['previousItem']);
                    }
                }

                return true;
            }
            else if (key == this.m_active['key'])
            {
                active = {'key': key, 'elem': item};
                this.m_active = active;

                // update tab index
                if (skipFocus === undefined || !skipFocus)
                {
                    this._makeFocusable(item);
                    // make sure ul is not tabbable
                    this.RemoveRootElementTabIndex();
                }
            }
        }
        else
        {
            // item is null, just clears the current values
            this.m_active = null;
        }

        return false;
    },

    /**
     * Put browser focus on item (or children of item)
     * @private
     */
    _focusItem: function(item)
    {
        this.getFocusItem(item).focus();
    },

    /**
     * Highlight active element
     * @protected
     */
    HighlightActive: function()
    {
        var item, target;

        // don't highlight and focus item if ancestor does not have focus
        if (this.m_active != null && this._getListContainer().hasClass("oj-focus-ancestor"))
        {   
            item = this.m_active['elem'];
            this._highlightElem(item, "oj-focus");
        }
    },

    /**
     * Unhighlight the active index, and turn the active index to selected instead if selectActive is true.
     * @protected
     */
    UnhighlightActive: function()
    {
        if (this.m_active != null)
        {
            this._unhighlightElem(this.m_active['elem'], "oj-focus");
        }
    },

    HandleClickActive: function(item, event)
    {
        // if click is triggered by target inside the active item, then do nothing
        var active = this.m_active != null ? this.m_active['elem'].get(0) : null;
        // if it's a group, use the group header div instead otherwise active.contains check will not be valid
        if (active != null && !$(active).hasClass(this.getItemStyleClass()))
        {
            active = active.firstElementChild;
        }

        if (event != null && active != null && active != event.target && active.contains(event.target))
        {
            return;
        }

        this.SetCurrentItem(item, event, this.ShouldUseGridRole() ? event.target != item.get(0) : true);
    },

    /**
     * Sets the active item and bring focus to it.  Update the currentItem option.
     * @protected
     */
    SetCurrentItem: function(item, event, skipFocus)
    {
        var proceed = this.ActiveAndFocus(item, event, skipFocus);
        if (proceed)
        {
            this._setCurrentItemOption(this.GetKey(item[0]), event, item.get(0));
        }
    },

    /**
     * Sets the active item and bring focus to it.
     * @protected
     */
    ActiveAndFocus: function(item, event, skipFocus)
    {
        var proceed = false;

        // make sure that it is visible
        this._scrollToVisible(item[0]);

        // unhighlight any previous active item
        this.UnhighlightActive();

        // update active and frontier
        proceed = this._setActive(item, event, skipFocus);

        // highlight active
        this.HighlightActive();

        return proceed;
    },
    /************************************ end Active item ******************************/

    /************************************* Selection ***********************************************/
    /**
     * If selection is enabled and at least one item has to be selected, then
     * make sure the first selectable item is selected in the list.
     */
    enforceSelectionRequired: function()
    {
        var selection;

        if (this._isSelectionEnabled() && this._isSelectionRequired())
        {
            selection = this.GetOption("selection");
            if (selection == null || selection.length == 0)
            {
                this._selectFirstSelectableItem();
            }
        }            
    },

    /**
     * Returns the first selectable item in the list.
     * @private
     */
    _getFirstSelectableItem: function()
    {
        // currently, an item will have the aria-selected attribute defined only if it's selectable
        // so we are using that here to find all selectable items.
        var elem = this.element[0].querySelector(".oj-listview-cell-element[aria-selected]");
        return elem == null ? null : elem.parentNode;
    },

    /**
     * Selects the first selectable item
     * @private
     */
    _selectFirstSelectableItem: function()
    {
        var first, item, selection, key, selectableElems;

        item = this._getFirstSelectableItem();

        // select the item if found
        if (item)
        {
            key = this.m_contentHandler.GetKey(item);
            if (key != null)
            {
                this._applySelection(item, key);
                this._setSelectionOption([key], null, [item]);
            }
        }
    },

    /**
     * Check if selection enabled by options on the list
     * @return {boolean} true if selection enabled
     * @private
     */
    _isSelectionEnabled: function()
    {
        return (this.GetOption("selectionMode") != "none");
    },

    /**
     * Check if there should be at least one item selected in the list
     * @return {boolean} true if selection is required
     * @private
     */
    _isSelectionRequired: function()
    {
        return this.GetOption("selectionRequired");
    },

    /**
     * Check whether multiple selection is allowed by options on the list
     * @return {boolean} true if multiple selection enabled
     * @private
     */
    _isMultipleSelection: function()
    {
        return (this.GetOption("selectionMode") == "multiple");
    },

    /**
     * Check whether the item is selectable
     * @param {Element} item the item element
     * @return {boolean} true if item is selectable
     * @protected
     */
    IsSelectable: function(item)
    {
        item = this.getFocusItem($(item)).get(0);
        return item.hasAttribute("aria-selected");
    },

    /**
     * Filter the array of selection.  Remove any items that are not selectable, or if there are more than one items
     * when selectionMode is single.
     * @param {Array} selection array of selected items
     * @return {Array} filtered array of items
     * @private
     */
    _filterSelection: function(selection)
    {
        var filtered, i, elem;

        filtered = [];
        for (i=0; i<selection.length; i++)
        {
            elem = this.FindElementByKey(selection[i]);
            if (elem == null || (elem != null && this.IsSelectable(elem)))
            {
                filtered.push(selection[i]);

                // if single selection mode, we can stop
                if (!this._isMultipleSelection())
                {
                    break;
                }
            }
        }

        return filtered;
    },

    /**
     * Sets the selection option with a new value
     * @param {Object} newValue the new value for selection option
     * @param {Event|null} event the DOM event
     * @param {Array.<Element>=} selectedElems an array of DOM Elements 
     * @param {*=} firstSelectedItemData the data for first selected item
     * @private
     */
    _setSelectionOption: function(newValue, event, selectedElems, firstSelectedItemData)
    {
        var firstSelectedItem, value = {'key': null, 'data': null}, extra;

        // check if the value has actually changed, based on key
        // firstSelectedItem should never be null and should always have 'key'
        firstSelectedItem = this.GetOption("firstSelectedItem");

        // NavList firstSelectedItem would be undefined
        if (firstSelectedItem != null)
        {
            // first condition is if new value is empty (should never be null) and existing item is non null
            // second condition is if new value is not empty and does not match the existing item
            if (((newValue == null || newValue.length == 0) && firstSelectedItem['key'] != null) || 
                 (!(newValue[0] == firstSelectedItem['key'] || oj.Object.compareValues(newValue[0], firstSelectedItem['key']))))
            {
                // update firstSelectedItem also
                if (newValue != null && newValue.length > 0)
                {
                    value = {'key': newValue[0], 'data': firstSelectedItemData != undefined ? firstSelectedItemData : this.getDataForVisibleItem({'key': newValue[0]})};
                }

                this.SetOption("firstSelectedItem", value, {'_context': {originalEvent: event, internalSet: true}, 'changed': true});
            }
        }

        extra = {'items': this.ojContext._IsCustomElement() ? selectedElems : (selectedElems == null ? $({}) : $(selectedElems))}; 
        this.SetOption("selection", newValue, {'_context': {originalEvent: event, internalSet: true, extraData: extra}, 'changed': true});
    },

    /**
     * Unhighlights the selection.  Does not change selection, focus, anchor, or frontier
     * @private
     */
    _unhighlightSelection: function()
    {
        var self, elem;

        if (this.m_keyElemMap == null)
        {  
            return;
        }

        self = this;
        $.each(this.GetOption("selection"), function(index, value)
        {
            elem = self.FindElementByKey(value);
            if (elem != null)
            {
                self._unhighlightElem(elem, "oj-selected");
            }
        });
    },

    _highlightElem: function(elem, style)
    {
        this.HighlightUnhighlightElem(elem, style, true);
    },

    _unhighlightElem: function(elem, style)
    {
        this.HighlightUnhighlightElem(elem, style, false);
    },

    /**
     * Highlight or unhighlight an element
     * @param {jQuery|Element} elem the element the highlight or unhighlight
     * @param {string} style the style to add or remove
     * @param {boolean} highlight true if it's to highlight, false if it's to unhighlight
     * @protected
     */
    HighlightUnhighlightElem: function(elem, style, highlight)
    {
        var group;

        elem = $(elem);

        if (style == "oj-selected")
        {
            this.getFocusItem(elem).attr("aria-selected", highlight ? "true" : "false");
        }

        // if item is a group, the highlight should be apply to the group item element
        group = elem.children("."+this.getGroupItemStyleClass());
        if (group.length > 0)
        {
            elem = $(group[0]);
        }

        if (style === "oj-focus")
        {
            if (highlight)
            {
                // don't apply focus ring on item if we are in actionable mode
                if (this.m_keyMode != "actionable")
                {
                    this._focusInHandler(elem);
                }
            }
            else
            {
                this._focusOutHandler(elem);
            }
        }
        else
        {
            if (highlight)
            {
                elem.addClass(style);
            }
            else
            {
                elem.removeClass(style);
            }
        }    
    },

    /**
     * Handles click to select multiple cells/rows
     * @param {jQuery} item the item clicked on
     * @param {Event} event the click event
     * @protected
     */
    HandleClickSelection: function(item, event)
    {
        var ctrlKey, shiftKey;

        // make sure that it is visible
        this._scrollToVisible(item[0]);

        ctrlKey = this._ctrlEquivalent(event);
        shiftKey = event.shiftKey;
        if (this._isMultipleSelection())
        {
            if (!ctrlKey && !shiftKey)
            {
                this.SelectAndFocus(item, event);
            }
            else if (!ctrlKey && shiftKey)
            {
                // active item doesn't change in this case
                this._extendSelection(item, event);
            }
            else
            {
                this._augmentSelectionAndFocus(item, event);
            }
        }
        else
        {
            this.SelectAndFocus(item, event);
        }
    },

    /**
     * Handles tap to select multiple cells/rows
     * @param {jQuery} item the item clicked on
     * @param {Event} event the click event
     * @private
     */
    _handleTouchSelection: function(item, event)
    {
        if (this._isMultipleSelection())
        {
            if (event.shiftKey)
            {
                // for touch device with keyboard support
                this._extendSelection(item, event);
            }
            else
            {
                // treat this as like ctrl+click
                this._augmentSelectionAndFocus(item, event);
            }
        }
        else
        {
            this.SelectAndFocus(item, event);
        }
    },

    /**
     * Clear the current selection.
     * @param {boolean} updateOption true if the underlying selection option should be updated, false otherwise.
     * @param {jQuery=} newSelectionFrontier new value to set the selection frontier. If none specified, set to null
     * @private
     */
    _clearSelection: function(updateOption, newSelectionFrontier)
    {
        // unhighlight previous selection
        this._unhighlightSelection();

        if (updateOption)
        {
            // if the intend is to empty selection option, we have to make sure if
            // selectionRequired is set to true that something is selected
            if (this._isSelectionRequired())
            {
                this._selectFirstSelectableItem();
            }
            else
            {
                this._setSelectionOption([], null, null);
            }
        }

        // clear selection frontier also
        this.m_selectionFrontier = (newSelectionFrontier === undefined) ? null : newSelectionFrontier;
    },

    /**
     * Selects the focus on the specified element
     * Select and focus is an asynchronus call
     * @param {jQuery} item the item clicked on
     * @param {Event} event the click event
     * @protected
     */
    SelectAndFocus: function(item, event)
    {
        var key, selection, index;

        key = this.GetKey(item[0]);
        selection = this.GetOption("selection");
        index = this.GetIndexOf(selection, key);

        // if it's already selected, deselect it and update options
        this._clearSelection(index > -1);

        if (index == -1)
        {
            // add the elem to selection
            this._augmentSelectionAndFocus(item, event, []);            
        }
    },

    /**
     * Shift+click to extend the selection
     * @param {jQuery} item the item to extend selection to
     * @param {Event} event the key event
     * @private
     */
    _extendSelection: function(item, event)
    {
        var current;

        if (this.m_active == null)
        {
            return;
        }
    
        // checks if selection has changed
        current = this.m_selectionFrontier;
        if (current == item)
        {
            return;
        }

        // remove focus style on the item click on
        this._unhighlightElem(item, "oj-focus");

        this._extendSelectionRange(this.m_active['elem'], item, event);
    },

    /**
     * Extend the selection
     * @param {jQuery} from the item to extend selection from
     * @param {jQuery} to the item to extend selection to
     * @param {Event} event the event that triggers extend
     * @param {boolean=} keepSelectionFrontier true if we don't want to modify the selectionFrontier
     * @private
     */
    _extendSelectionRange: function(from, to, event, keepSelectionFrontier)
    {
        if (keepSelectionFrontier === true)
        {
        // clear selection as we'll be just re-highlight the entire range
            this._clearSelection(false, this.m_selectionFrontier);
        }
        else 
        {
            this._clearSelection(false, to);
        }

        // highlights the items between active item and new item
        this._highlightRange(from, to, event);
        this.HighlightActive();

        // make sure that it is visible
        this._scrollToVisible(to[0]);
    },

    /**
     * Highlight the specified range
     * @param {jQuery} start the start of the range
     * @param {jQuery} end the end of the range
     * @param {Event} event the event that triggers the highlight
     * @private
     */
    _highlightRange: function(start, end, event)
    {
        var selection, selectedItems, items, startIndex, endIndex, from, to, i, item, key;

        selection = [];
        selectedItems = [];
        items = this._getItemsCache();
        startIndex = items.index(start); 
        endIndex = items.index(end);

        if (startIndex > endIndex)
        {
            from = endIndex;
            to = startIndex;
        }
        else
        {
            from = startIndex;
            to = endIndex;
        }

        // exclude start and include end
        for (i=from; i<=to; i++)
        {
            item = items[i];
            if (this.IsSelectable(item))
            {
                key = this.m_contentHandler.GetKey(item);

                this._applySelection(item, key);
                selection.push(key);
                selectedItems.push(item);
            }
        }

        // trigger the optionChange event
        this._setSelectionOption(selection, event, selectedItems);
    },

    /**
     * Apply selection to the element
     * @param {jQuery|Element} element the item to apply selection
     * @param {Object} key the key of the item
     * @private
     */
    _applySelection: function(element, key)
    {
        // update map that keeps track of key->element
        if (this.m_keyElemMap == null)
        {
            this.m_keyElemMap = new oj.KeyMap();
        }
        this.m_keyElemMap.set(key, $(element).attr("id"));

        // highlight selection
        this._highlightElem(element, 'oj-selected');
    },

    /**
     * Ctrl+click to add item to the current selection
     * @param {jQuery} item the item to augment selection to
     * @param {Event} event the event that triggers the selection
     * @param {Array=} selection the optional selection to augment, if not specified, use current selection
     * @private
     */
    _augmentSelectionAndFocus: function(item, event, selection)
    {
        var active, key, currentActive, proceed = false, index, selectedItems, i;

        active = item;
        key = this.GetKey(item[0]);

        if (selection == undefined)
        {
            selection = this.GetOption("selection").slice(0);
        }

        // update active only if target is not inside the active item
        currentActive = this.m_active != null ? this.m_active['elem'].get(0) : null;
        // if it's a group, use the group header div instead otherwise currentActive.contains check will not be valid
        if (currentActive != null && !$(currentActive).hasClass(this.getItemStyleClass()))
        {
            currentActive = currentActive.firstElementChild;
        }

        if (event != null && (currentActive == null || currentActive == event.target || !currentActive.contains(event.target)))
        {
            this.UnhighlightActive();

            // update active cell and frontier
            proceed = this._setActive(active, event);
            if (proceed)
            {
                // update current option
                this._setCurrentItemOption(key, event, active.get(0));
            }

            // highlight index
            this.HighlightActive();
        }

        // checks if setActive was successful
        currentActive = this.m_active != null ? this.m_active['elem'].get(0) : null;
        if (currentActive == null || currentActive != active.get(0))
        {
            // update selection if it was cleared
            if (selection != null && selection.length == 0)
            {
                this._setSelectionOption(selection, event, []);
            }
            return;
        }

        index = this.GetIndexOf(selection, key);
        if (index > -1)
        {
            // it was selected, deselect it
            this._unhighlightElem(item, "oj-selected");
            selection.splice(index, 1);
        }
        else
        {
            this.m_selectionFrontier = item;
            this._applySelection(item, key);
            selection.push(key);
        }

        selectedItems = new Array(selection.length);
        for (i = 0; i < selection.length; i++)
        {
            selectedItems[i] = this.FindElementByKey(selection[i]);
        }

        // trigger option change
        this._setSelectionOption(selection, event, selectedItems);
    },

    /**
     * Toggle selection of an item.  If an item was selected, it deselects it.  If an item was not selected, it selects it.
     * @param {Event} event the event that triggers the selection
     * @param {boolean} keepCurrentSelection true if selecting an item would not deselect other selected items, false otherwise
     * @param {boolean} skipIfNotSelected true if an selected item should not be deselected, false otherwise
     * @protected
     */
    ToggleSelection: function(event, keepCurrentSelection, skipIfNotSelected)
    {
        // if it's currently selected, deselect it
        var selection, selectedItems, item, key, index, i;

        selection = this.GetOption("selection").slice(0);
        item = this.m_active['elem'];
        key = this.m_active['key'];        

        index = this.GetIndexOf(selection, key);
        if (index > -1)
        {
            // do not deselect the item if it's the last selected item and selection is required
            if (skipIfNotSelected || (selection.length == 1 && this._isSelectionRequired()))
            {
                return;
            }

            // it was selected, deselect it
            this._unhighlightElem(item, "oj-selected");
            selection.splice(index, 1);

            if (selection.length == 0)
            {
                this.m_selectionFrontier = null;
            }
        }
        else
        {
            if (this.IsSelectable(item[0]))
            {
                // deselect any selected items
                if (!keepCurrentSelection)
                {
                    this._clearSelection(false);
                    selection.length = 0;
                }

                this.m_selectionFrontier = item;

                // select current item
                this._applySelection(item, key);
                selection.push(key);
            }
        }

        selectedItems = new Array(selection.length);
        for (i = 0; i < selection.length; i++)
        {
            selectedItems[i] = this.FindElementByKey(selection[i]);
        }

        // trigger option change
        this._setSelectionOption(selection, event, selectedItems);
    },

    /**
     * Checks whether the element is an input element. 
     * @param {Element} elem the element to check
     * @return {boolean} true if it's input, false otherwise
     * @private
     */
    _isInputElement: function(elem)
    {
        var inputRegExp = /^INPUT|SELECT|OPTION|TEXTAREA/; 
        return elem.nodeName.match(inputRegExp) != null && !elem.readOnly;  
    },

    /**
     * Handles key event for selection or active
     * @param {Event} event
     * @return {boolean} true if the event is processed
     * @protected
     */
    HandleSelectionOrActiveKeyDown: function(event)
    {
        var keyCode, current, ctrlKey, shiftKey, processed = false, first, last, ui;

        // this could happen if nothing in the list is focusable
        if (this.m_active == null)
        {
            return false;
        }

        keyCode = event.keyCode;
        current = this.m_active['elem'];

        if (this._isActionableMode())
        {
            // Esc key goes to navigation mode
            if (keyCode == this.ESC_KEY)
            {
                this._exitActionableMode();

                // force focus back on the active cell
                this.HighlightActive();
                this._focusItem(current);

                // make sure active item has tabindex set
                this._setTabIndex(current);
                processed = true;
            }
            else if (keyCode === this.TAB_KEY)
            {
                // check if it's the last element in the item
                first = current.find("[data-first]");  
                last = current.find("[data-last]");
                if (event.shiftKey)
                {
                    if (first.length > 0 && last.length > 0 && first != last && event.target == first[0])
                    {
                        // recycle to last focusable element in the cell
                        last[0].focus();
                        processed = true;
                    }
                }
                else
                {
                    if (first.length > 0 && last.length > 0 && first != last && event.target == last[0])
                    {
                        // recycle to first focusable element in the cell
                        first[0].focus();
                        processed = true;
                    }
                } 
                // otherwise don't process and let browser handles tab
            }
        }
        else
        {        
            // F2 key goes to actionable mode
            if (keyCode == this.F2_KEY)
            {
                this._enterActionableMode();

                // focus on first focusable item in the cell
                first = current.find("[data-first]");
                if (first.length > 0)
                {
                    first[0].focus();

                    if (this._isExpandCollapseIcon(first))
                    {
                        this._focusInHandler(first);
                    }

                    // check if it's group item
                    if (!current.hasClass(this.getItemStyleClass()))
                    { 
                        current = current.children("."+this.getGroupItemStyleClass()).first();
                    }
                    current.removeClass("oj-focus-highlight");
                }
            }
            else if (keyCode == this.SPACE_KEY && this._isSelectionEnabled())
            {
                ctrlKey = this._ctrlEquivalent(event);
                shiftKey = event.shiftKey;
                if (shiftKey && !ctrlKey && this.m_selectionFrontier != null && this._isMultipleSelection())
                {
                    // selects contiguous items from last selected item to current item
                    this._extendSelectionRange(this.m_selectionFrontier, this.m_active['elem'], event, true);
                }
                else
                {
                    // toggle selection, deselect previous selected items
                    this.ToggleSelection(event, ctrlKey && !shiftKey && this._isMultipleSelection(), false);
                }
                processed = true;
            }
            else if (keyCode == this.ENTER_KEY && this._isSelectionEnabled())
            {
                // selects it if it's not selected, do nothing if it's already selected
                this.ToggleSelection(event, false, true);
            }
            else if (this.IsArrowKey(keyCode))
            {
                ctrlKey = this._ctrlEquivalent(event);
                shiftKey = event.shiftKey;
                if (!ctrlKey)
                {
                    processed = this.HandleArrowKeys(keyCode, (shiftKey && this._isSelectionEnabled() && this._isMultipleSelection()), event);
                }
            }
            else if (keyCode === this.TAB_KEY)
            {
                // content could have changed, disable all elements in items before or after the active item
                if (event.shiftKey)
                {
                    this._disableAllTabbableElementsBeforeItem(current);
                }
                else
                {
                    this._disableAllTabbableElementsAfterItem(current);
                }
            }
        }

        return processed;
    },
    /********************************** End Selection **********************************************/

    /********************************** Disclosure **********************************************/
    /**
     * Whether the group item is currently in the middle of expanding/collapsing
     * @param {Object} key the key of the group item
     * @return {boolean} true if it's expanding/collapsing, false otherwise
     * @private
     */
    _isDisclosing: function(key)
    {
        if (key && this.m_disclosing)
        {
            return (this.m_disclosing.indexOf(key) > -1);
        }

        return false;
    },

    /**
     * Marks a group item as currently in the middle of expanding/collapsing
     * @param {Object} key the key of the group item
     * @param {boolean} flag true or false
     * @private
     */
    _setDisclosing: function(key, flag)
    {
        var index;

        if (key == null)
        {
            return;
        }

        if (this.m_disclosing == null)
        {
            this.m_disclosing = [];
        }

        if (flag)
        {
            this.m_disclosing.push(key);
        }
        else
        {
            // there should be at most one entry, but just in case remove all occurrences
            index = this.m_disclosing.indexOf(key);
            while (index > -1)
            {
                this.m_disclosing.splice(index, 1);
                index = this.m_disclosing.indexOf(key);
            }
        }
    },         

    /**
     * Gets the animation effect for the specific action
     * @param {string} action the action to retrieve the effect
     * @return {Object} the animation effect for the action
     */
    getAnimationEffect: function(action)
    {
        var defaultAnimations;

        if (this.defaultOptions == null)
        {
            this.defaultOptions = oj.ThemeUtils.parseJSONFromFontFamily(this.getOptionDefaultsStyleClass());
        }

        defaultAnimations = this.defaultOptions['animation'];
        return defaultAnimations == null ? null : defaultAnimations[action];
    },

    /**
     * Whether group items can be expand/collapse.
     * @return {boolean} true if group items can be expand/collapse, false otherwise.
     */
    isExpandable: function()
    {
        return this.GetOption("drillMode") != "none";
    },
    /**
     * Whether ListView should expand all expandable items.
     * @return {boolean} true if expand all, false otherwise
     * @private
     */
    _isExpandAll: function()
    {
        var expanded = this.GetOption("expanded");
        // for legacy syntax, which supports 'auto' and 'all'
        // TODO: should add check for custom element, but can't do it until NavList also move to KeySet
        if (expanded === 'auto')
        {
            // if drillMode is none and no expanded state is specified, expand all
            if (!this.isExpandable())
            {
                 return true;
            }
        }
        else if (expanded === 'all')
        {
            return true;
        }

        // for custom element and also legacy syntax that uses the new KeySet"expand
        if (expanded.isAddAll)
        {
            // if drillMode is none and no expanded state is specified, expand all
            return (!this.isExpandable() && expanded instanceof oj._ojListViewExpandedKeySet) ? true : expanded.isAddAll();
        }

        return false;
    },

    /**
     * Expand an item with specified key.
     * Invoked by widget
     * @param {Object} key the key of the group item to expand
     * @param {boolean} beforeVetoable true if beforeExpand event can be veto, false otherwise
     * @param {boolean} fireBefore true if this should trigger a beforeExpand event
     * @param {boolean} fireAfter true if this should trigger an expand event
     * @param {boolean} animate true if animate the expand operation, false otherwise
     */
    expandKey: function(key, beforeVetoable, fireBefore, fireAfter, animate)
    {
        var item = this.FindElementByKey(key);
        if (item != null)
        {
            this.ExpandItem($(item), null, animate, key, beforeVetoable, fireAfter, fireBefore);    
        }
    },

    /**
     * Handle expand operation
     * @param {Event} event the event that triggers the expand
     * @private
     */
    _expand: function(event)
    {
        var item = this.FindItem(event.target);
        if (item != null && item.length > 0)
        {
            this.ExpandItem(item, event, true, null, true, true, true);
        }
    },

    /**
     * Expand an item
     * @param {jQuery} item the item to expand
     * @param {Event} event the event that triggers expand.  Note that event could be null in the case where this is programmatically done by the widget
     * @param {boolean} animate true if animate the expand operation, false otherwise
     * @param {Object|null} key the key of the item, if not specified, the logic will figure it out from the item
     * @param {boolean} beforeVetoable true if beforeExpand event can be veto, false otherwise
     * @param {boolean} fireEvent true if fire expand event, false otherwise
     * @param {boolean} fireBeforeEvent true if fire beforeexpand event, false otherwise
     * @protected
     */
    ExpandItem: function(item, event, animate, key, beforeVetoable, fireEvent, fireBeforeEvent)
    {
        var ui, cancelled, index;

        // checks if it's already collapsed or not collapsible at all
        if (this.GetState(item) != this.STATE_COLLAPSED)
        {
            return;
        }

        // if key wasn't specified, find it
        if (key == null)
        {
            key = this.GetKey(item[0]);
        }

        // bail if it's in the middle of expanding/collapsing
        if (animate && this._isDisclosing(key))
        {
            return;
        }

        ui = {"item": item, "key": key};

        if (fireBeforeEvent)
        {
            cancelled = !this.Trigger("beforeExpand", event, ui);
            if (cancelled && beforeVetoable) 
            {
                return;
            }
        }

        this.signalTaskStart("Expand item: "+key); // signal method task start

        if (animate)
        {
            this._setDisclosing(key, true);
        }
        this.m_contentHandler.Expand(item, function(groupItem){this._expandSuccess(groupItem, animate, event, ui, fireEvent)}.bind(this));

        // clear items cache
        this.m_items = null;

        // prevent item click handler to trigger
        if (event != null)
        {
            event.stopPropagation();
        }

        // update var that keeps track of collapsed items
        if (!this.ojContext._IsCustomElement() && this._collapsedKeys != null)
        {
            index = this._collapsedKeys.indexOf(key);
            if (index != -1)
            {
                this._collapsedKeys.splice(index, 1);
            }
        }

        this.signalTaskEnd(); // signal method task end
    },

    /**
     * @param {Element} groupItem
     * @param {boolean} animate
     * @param {Event} event
     * @param {Object} ui
     * @param {boolean} fireEvent
     * @private
     */
    _expandSuccess: function(groupItem, animate, event, ui, fireEvent)
    {
        var animationPromise, item, collapseClass, expandClass, expandingClass, groupItemStyleClass, currValue, newValue, self = this;

        this.signalTaskStart("Handle results from successful expand"); // signal method task start

        // save the key for use when expand complete
        groupItem.key = ui['key'];

        animationPromise = this.AnimateExpand($(groupItem), animate, event);

        item = groupItem.parentNode;

        item = $(item);
        // update aria expanded
        this.SetState(item, this.STATE_EXPANDED);
        // update icon
        collapseClass = this.getCollapseIconStyleClass();
        expandClass = this.getExpandIconStyleClass();
        expandingClass = this.getExpandingIconStyleClass();
        groupItemStyleClass = this.getGroupItemStyleClass();
        item.children("."+groupItemStyleClass).find("."+collapseClass+", ."+expandingClass)
                                              .removeClass(collapseClass)
                                              .removeClass(expandingClass)
                                              .addClass(expandClass);

        // fire expand event after expand animation completes
        if (fireEvent)
        {
            animationPromise.then(function() {
                // update option.  As an optimization do it only when fireEvent is true since this is the
                // only time when it's not triggered by API, in which case the value is already current
                if (self.ojContext._IsCustomElement())
                {
                    currValue = self.GetOption("expanded");
                    if (self._isKeySet(currValue) && !currValue.has(groupItem.key))
                    {
                        newValue = currValue.add([groupItem.key]);
                        self.SetOption("expanded", newValue, {'_context': {originalEvent: event, internalSet: true}, 'changed': true});
                    }
                }

                self.Trigger("expand", event, ui);
            });
        }

        animationPromise.then(function() {
          self.signalTaskEnd(); // signal method task end
        });
    },

    /**
     * Adjust the max height of ancestors of a group items.
     * @param {jQuery} groupItem the group item where we want to adjust its ancestors max height
     * @param {number} delta the height to increase
     * @private
     */
    _adjustAncestorsMaxHeight: function(groupItem, delta)
    {
        var maxHeight;

        groupItem.parentsUntil("ul.oj-component-initnode", "ul."+this.getGroupStyleClass()).each(function(index)
        {
            maxHeight = parseInt($(this).css("maxHeight"), 10);
            if (maxHeight > 0)
            {
                $(this).css("maxHeight", (maxHeight + delta) +"px");
            }
        });
    },

    /**
     * Animate expand operation
     * @param {jQuery} groupItem the group item that is expand (todo: not consistent with animateCollapse)
     * @param {boolean} animate true if animate expand, false otherwise
     * @param {Event} event the event that triggers expand.  Note that event could be null in the case where this is programmatically done by the widget
     * @return {Promise} A Promise that resolves when expand animation completes
     * @protected
     */
    AnimateExpand: function(groupItem, animate, event)
    {
        var totalHeight = 0, animationPromise, animationResolve, self = this, elem, action = "expand", promise;
        
        animationPromise = new Promise(function(resolve, reject) 
        {
            animationResolve = resolve;
        });

        if (animate)
        {
            this.signalTaskStart("Animate expand of group item"); // signal task start

            groupItem.children().each(function()
            {
                totalHeight = totalHeight + $(this).outerHeight(true);
            });

            // for touch we'll need to re-adjust the max height of parent nodes since max height doesn't get remove
            if (this._isTouchSupport())
            {
                this._adjustAncestorsMaxHeight(groupItem, totalHeight);
            }

            groupItem.css("maxHeight", totalHeight+"px");

            this.signalTaskStart("Kick off expand animation"); // signal expand animation started. Ends in _handleExpandTransitionEnd()

            // now show it
            elem = /** @type {Element} */ (groupItem.get(0));
            promise = this.StartAnimation(elem, action);
            promise.then(function(val)
            {
                self._handleExpandTransitionEnd(groupItem, animationResolve);
            });

            this.signalTaskEnd(); // signal task end
        }
        else
        {
            // for touch it will need max height set initially in order for it to animate correctly
            if (this._isTouchSupport())
            {
                groupItem.children().each(function()
                {
                    totalHeight = totalHeight + $(this).outerHeight(true);
                });
                groupItem.css("maxHeight", totalHeight+"px");

                this._adjustAncestorsMaxHeight(groupItem, totalHeight);
            }
            else
            {
                groupItem.css("maxHeight", "");
            }

            this.AnimateExpandComplete(groupItem);
            animationResolve(null); // resolve animationPromise
        }
        return animationPromise;
    },

    _handleExpandTransitionEnd: function(groupItem, animationResolve)
    {
        var item;

        // on ios removing max-height will cause double animation
        if (!this._isTouchSupport())
        {
            groupItem.css("maxHeight", "");
        }

        this.AnimateExpandComplete(groupItem);
        animationResolve(null); // resolve animationPromise

        this.signalTaskEnd(); // signal expand animation ended. Started in this.AnimateExpand()
    },

    /**
     * Invoked when expand animation is completed.  Class who overrides AnimateExpand
     * must call this method upon finish animation.
     * @param {jQuery} groupItem the item to collapse
     * @protected
     */
    AnimateExpandComplete: function(groupItem)
    {
        groupItem.removeClass(this.getGroupCollapseStyleClass()).addClass(this.getGroupExpandStyleClass());
        this._setDisclosing(groupItem[0].key, false);
    },

    /**
     * Collapse an item with specified key.
     * Invoked by widget
     * @param {Object} key the key of the group item to collapse
     * @param {boolean} fireBefore true if this should trigger a beforeCollapse event
     * @param {boolean} fireAfter true if this should trigger a collapse event
     * @param {boolean} animate true if animate the collapse operation, false otherwise
     */
    collapseKey: function(key, fireBefore, fireAfter, animate)
    {
        var item = this.FindElementByKey(key);
        if (item != null)
        {
            this.CollapseItem($(item), null, animate, key, fireBefore, fireAfter);    
        }
    },

    _collapse: function(event)
    {
        var item = this.FindItem(event.target);
        if (item != null && item.length > 0)
        {
            this.CollapseItem(item, event, true, null, true, true);
        }        
    },

    /**
     * Collapse an item
     * @param {jQuery} item the item to expand
     * @param {Event} event the event that triggers collapse.  Note that event could be null in the case where this is programmatically done by the widget
     * @param {boolean} animate true if animate the collapse operation, false otherwise
     * @param {Object|null} key the key of the item, if not specified, the logic will figure it out from the item
     * @param {boolean} beforeVetoable true if beforeCollapse event can be veto, false otherwise
     * @param {boolean} fireEvent true if fire collapse event, false otherwise
     * @protected
     */
    CollapseItem: function(item, event, animate, key, beforeVetoable, fireEvent)
    {
        var ui, cancelled, animationPromise, collapseClass, expandClass, currValue, newValue, self = this;

        // checks if it's already collapsed or not collapsible at all
        if (this.GetState(item) != this.STATE_EXPANDED)
        {
            return;
        }

        // fire beforeCollapse event
        if (key == null)
        {
            key = this.GetKey(item[0]);
        }

        // bail if it is in the middle of expanding/collapsing
        if (animate && this._isDisclosing(key))
        {
            return;
        }

        ui = {"item": item, "key": key};

        cancelled = !this.Trigger("beforeCollapse", event, ui);
        if (cancelled && beforeVetoable) 
        {
            return;
        }

        this.signalTaskStart("Collapse item: "+key); // signal method task start

        if (animate)
        {
            this._setDisclosing(key, true);
        }

        // animate collapse
        animationPromise = this.AnimateCollapse(item, key, animate, event);

        // update aria expanded
        this.SetState(item, this.STATE_COLLAPSED);
        // update icon
        collapseClass = this.getCollapseIconStyleClass();
        expandClass = this.getExpandIconStyleClass();
        item.find("."+expandClass).first().removeClass(expandClass).addClass(collapseClass);

        // clear items cache
        this.m_items = null;

        // prevent item click handler to trigger
        if (event != null)
        {
            event.stopPropagation();
        }

        // fire collapse event after collapse animation completes
        if (fireEvent)
        {
            animationPromise.then(function()
            {
                // update option.  As an optimization do it only when fireEvent is true since this is the
                // only time when it's not triggered by API, in which case the value is already current
                if (self.ojContext._IsCustomElement())
                {
                    currValue = self.GetOption("expanded");
                    if (self._isKeySet(currValue))
                    {
                        newValue = currValue.delete([key]);
                        self.SetOption("expanded", newValue, {'_context': {originalEvent: event, internalSet: true}, 'changed': true});
                    }
                }

                self.Trigger("collapse", event, ui);
            })
        }

        // _collapsedKeys should only be used in the legacy syntax case
        if (!this.ojContext._IsCustomElement())
        {
            // keep track of collapsed item
            if (this._collapsedKeys == undefined)
            {
                this._collapsedKeys = [];
            }
            
            if (this._collapsedKeys.indexOf(key) == -1)
            {
                this._collapsedKeys.push(key);
            }
        }

        animationPromise.then(function()
        {
          self.signalTaskEnd(); // signal method task end
        });
    },

    /**
     * Animate collapse operation
     * To be change by NavList
     * @param {jQuery} item the item to collapse
     * @param {Object} key the key of the group item
     * @param {boolean} animate true if animate the collapse operation, false otherwise
     * @param {Event} event the event that triggers collapse.  Note that event could be null in the case where this is programmatically done by the widget
     * @return {Promise} A Promise that resolves when collapse animation completes
     * @protected
     */
    AnimateCollapse: function(item, key, animate, event)
    {
        var totalHeight = 0, groupItem, animationPromise, animationResolve, self = this, effect, elem, action = "collapse", promise;
        
        animationPromise = new Promise(function(resolve, reject) 
        {
            animationResolve = resolve;
        });

        groupItem = item.children("ul").first();
        // save the key for collapse animation complete
        groupItem[0].key = key;

        if (animate)
        {
            this.signalTaskStart("Animate collapse"); // signal task start

            groupItem.children().each(function()
            {
                totalHeight = totalHeight + $(this).outerHeight();
            });

            groupItem.css("maxHeight", totalHeight+"px");

            effect = this.getAnimationEffect(action);
            // max-height = 0 needs to stick around, especially needed for static content
            effect['persist'] = 'all';

            this.signalTaskStart("Kick off collapse animation"); // signal collapse animation started. Ends in _handleCollapseTransitionEnd()

            // now hide it
            elem = /** @type {Element} */ (groupItem.get(0));
            promise = this.StartAnimation(elem, action, effect);
            promise.then(function(val)
            {
                self._handleCollapseTransitionEnd(groupItem, animationResolve);
            });

            this.signalTaskEnd(); // signal task end
        }
        else
        {
            groupItem.css("maxHeight", "0px");

            this.AnimateCollapseComplete(groupItem);
            animationResolve(null); // resolve animationPromise
        }
        return animationPromise
    },

    _handleCollapseTransitionEnd: function(groupItem, animationResolve)
    {
        var item;

        this.AnimateCollapseComplete(groupItem);

        animationResolve(null); // resolve animationPromise

        this.signalTaskEnd(); // signal collapse animation ended. Started in AnimateCollapse()
    },

    /**
     * Invoked when collapse animation is completed.  Class who overrides AnimateCollapse
     * must call this method upon finish animation.
     * @param {jQuery} groupItem the item to collapse
     * @private
     */
    AnimateCollapseComplete: function(groupItem)
    {
        groupItem.removeClass(this.getGroupExpandStyleClass()).addClass(this.getGroupCollapseStyleClass());

        // ask the content handler to do the collapse operation
        // content handler might have been destroyed if animation ended after destroy is called
        if (this.m_contentHandler != null)
        {
            this.m_contentHandler.Collapse(groupItem);
        }

        this._setDisclosing(groupItem[0].key, false);
    },

    /**
     * Collapse all currently expanded items
     * @private
     */
    _collapseAll: function()
    {
        var self, items;

        this.signalTaskStart("Collapse all"); // signal method task start

        self = this;
        items = this._getItemsCache();
        items.each(function() 
        {
            // collapseItem checks whether item is expanded
            self.CollapseItem($(this), null, false, null, false, false);                
        });

        this.signalTaskEnd(); // signal method task end
    },

    /**
     * Gets the keys of currently expanded items.
     * Invoke by widget
     * @return {Array} array of keys of currently expanded items.
     */
    getExpanded: function()
    {
        var expanded, self, items, item;

        expanded = [];

        self = this;
        items = this._getItemsCache();
        items.each(function() 
        {
            item = $(this);
            if (self.GetState(item) == self.STATE_EXPANDED)
            {
                expanded.push(self.GetKey(item[0]));
            }
        });

        return expanded;
    },
    /********************************* End Disclosure *******************************************/
    
    /**
     * Returns widget constructor.  Used by ContentHandler
     */
    getWidgetConstructor: function () 
    {
        return oj.Components.__GetWidgetConstructor(this.element);
    },
    
    /*********************************** Style Classes *********************************************/
    /**
     * To be change by NavList
     * @return {string} the container style class
     * @protected
     */
    GetContainerStyleClass: function()
    {
        // do not set overflow to scroll for windows touch enabled devices 
        if (this._isNonWindowTouch())
        {
            return "oj-listview oj-listview-container-touch";
        }
        else
        {
            return "oj-listview oj-listview-container";
        }
    },

    /**
     * To be change by NavList
     * @return {string} the main element style class
     * @protected
     */
    GetStyleClass: function()
    {
        return "oj-listview-element";
    },

    /**
     * To be change by NavList.  Access by ContentHandler.
     * @return {string} the list item style class
     */
    getItemStyleClass: function()
    {
        return "oj-listview-item";
    },

    /**
     * To be change by NavList.  Access by ContentHandler.
     * @return {string} the focused element style class
     */
    getFocusedElementStyleClass: function()
    {
        return "oj-listview-focused-element";
    },

    /**
     * To be change by NavList.  Access by ContentHandler.
     * @return {string} the list item element style class
     */
    getItemElementStyleClass: function()
    {
        return "oj-listview-item-element";
    },

    /**
     * To be change by NavList.  Access by ContentHandler.
     * @param {boolean=} includeSupplement optional flag to indicate whether to include any supplement classes
     * @return {string} the group item element style class
     */
    getGroupItemStyleClass: function(includeSupplement)
    {
        if (includeSupplement && this._isPinGroupHeader() && this._isPositionStickySupported())
        {
            return "oj-listview-group-item oj-sticky";
        }
        else
        {
            return "oj-listview-group-item";
        }
    },

    /**
     * To be change by NavList.  Access by ContentHandler.
     * @return {string} the group element style class
     */
    getGroupStyleClass: function()
    {
        return "oj-listview-group";
    },

    /**
     * To be change by NavList.  Access by ContentHandler.
     * @return {string} the group expand style class
     */
    getGroupExpandStyleClass: function()
    {
        return "oj-listview-collapsible-transition";
    },

    /**
     * To be change by NavList.  Access by ContentHandler.
     * @return {string} the group collapse style class
     */
    getGroupCollapseStyleClass: function()
    {
        return this.getGroupExpandStyleClass();
    },

    /**
     * To be change by NavList
     * @return {string} the collapse icon style class
     */
    getCollapseIconStyleClass: function()
    {
        return "oj-listview-collapse-icon";
    },

    /**
     * To be change by NavList
     * @return {string} the expand icon style class
     */
    getExpandIconStyleClass: function()
    {
        return "oj-listview-expand-icon";
    },

    /**
     * To be change by NavList
     * @return {string} the expanding icon style class
     */
    getExpandingIconStyleClass: function()
    {
        return "oj-listview-expanding-icon";
    },
    
    /**
     * To be change by NavList
     * @return {string} the empty text style class
     */
    getEmptyTextStyleClass: function()
    {
       return "oj-listview-no-data-message";    
    },
    
    /**
     * To be change by NavList
     * @return {string} the empty text marker class
     */
    getEmptyTextMarkerClass: function()
    {
        return "oj-listview-empty-text";
    },
    
    /**
     * To be change by NavList
     * @return {string} the depth style class
     */
    getDepthStyleClass: function(depth)
    {
        return "";
    },

    /**
     * To be change by NavList
     * @return {string} the option defaults style class
     */
    getOptionDefaultsStyleClass: function()
    {
        return "oj-listview-option-defaults";
    },

    /********************************* End Style Classes *******************************************/

    /*********************************** Pin Header *********************************************/
    _isPositionStickySupported: function()
    {
        // use native position sticky support for iOS (Safari and Chrome), see 
        return (this._isTouchSupport() && oj.AgentUtils.getAgentInfo()['os'] === oj.AgentUtils.OS.IOS);
    },

    /**
     * Helper method to prevent scroll by mouse wheel causes the page to scroll because it has reached the start/end of the list
     * @param {Element} scroller the scroller
     * @param {Event} event the mouse wheel event
     * @private
     */
    _preventMouseWheelOverscroll: function(scroller, event)
    {
        var delta = event.originalEvent.wheelDelta;
        // should only be applicable to TableDataSourceContentHandler for now
        if (isNaN(delta) || this.m_contentHandler.hasMoreToFetch === undefined)
        {
            return;
        }

        if (delta < 0)
        {
            // scroll down
            if (this.m_contentHandler.hasMoreToFetch() && (scroller.scrollTop + scroller.clientHeight + Math.abs(delta)) >= scroller.scrollHeight)
            { 
                scroller.scrollTop = scroller.scrollHeight;
                event.preventDefault();
            }
        }
        else
        {
            // scroll up
            if (scroller.scrollTop > 0 && (scroller.scrollTop - delta) <= 0)
            {
                scroller.scrollTop = 0;
                event.preventDefault();
            }
        }
    },

   /**
    * Retrieve the element where the scroll listener is registered on.
    * @private
    */
    _getScrollEventElement: function()
    {
        var scroller = this._getScroller();

        // if scroller is the body, listen for window scroll event.  This is the only way that works consistently across all browsers.
        if (scroller == document.body || scroller == document.documentElement)
        {
            return window;
        }
        else
        {
            return scroller;
        }
    },

    /**
     * Sets the current scrollPosition.  Invoked when programmatically sets
     * the scrollPosition.
     * @private
     */
    _setCurrentScrollPosition: function(scrollPosition)
    {
        var found = true, key, dataProvider, set, self = this;

        key = scrollPosition['key'];

        // if y is not specified.  First search local dom
        if (key && this.FindElementByKey(key) == null)
        {
            // need to verify key if we have a DataProvider that supports FetchByKeys
            if (this.m_contentHandler instanceof oj.IteratingDataProviderContentHandler)
            {
                dataProvider = this.m_contentHandler.getDataSource();
                if (dataProvider.containsKeys)
                {
                    // IE 11 does not support specifying value in constructor
                    set = new Set();
                    set.add(key);
                    dataProvider.containsKeys({'keys': set}).then(function(value)
                    {
                        // if not found, try to find by index or y
                        if (value['results'].size == 0)
                        {
                            delete scrollPosition['key'];
                        }
                        self.syncScrollPosition(scrollPosition);
                    })

                    // syncScrollPosition will be done later
                    found = false;
                }
            }
            // else we can't verify, so just let syncScrollPosition tries to fetch
            // and find the item
        }

        if (found)
        {
            this.syncScrollPosition(scrollPosition);
        }
    },

    /**
     * Find the element closest to the top of the viewport
     * @param {Array.<Element>} items an array of item elements to search for
     * @param {number} index the index relative to the parent to start the search
     * @param {number} scrollTop the current scrolltop
     * @private
     */
    _findClosestElementToTop: function(items, index, scrollTop)
    {
        var elem, offsetTop, diff, firstInGroup, forward, found = false;

        if (items.length == 0 || index >= items.length)
        {
            return;
        }

        elem = items[index];
        offsetTop = elem.offsetTop;
        diff = scrollTop - offsetTop;
        firstInGroup = {'index': index, 'elem': elem, 'offsetTop': offsetTop, 'offset': diff};

        // scroll position perfectly line up with the top of item (take sub-pixels into account), we are done
        if (Math.abs(diff) < 1)
        {
            return firstInGroup;
        }

        // go forward or backward to find the item, keep that fix to avoid
        // potentially going back and forth (shouldn't happen)
        forward = diff > 0;

        forward ? index++ : index--;
        while (!found && index >= 0 && index < items.length)
        {
            elem = items[index];
            offsetTop = elem.offsetTop;
            diff = Math.abs(scrollTop - offsetTop);

            found = (diff < 1) || (forward ? scrollTop <= offsetTop : scrollTop >= offsetTop);
            if (found)
            {
                // the one closer to the top wins
                if (diff < 1 || !forward)
                {
                    // the current one is closer
                    firstInGroup = {'index': index, 'elem': elem, 'offsetTop': offsetTop, 'offset': diff};
                }
                break;
            }

            // for card layout, we want to return the first item among items that have the same scrollTop (same row)
            // note when scrolling backward, you'll always want the last one encountered
            if (!forward || firstInGroup['offsetTop'] != offsetTop)
            {
                firstInGroup = {'index': index, 'elem': elem, 'offsetTop': offsetTop, 'offset': diff};
            }

            forward ? index++ : index--;
        }

        if (!found)
        {
            // then it's the first/last item in the group/root
            index = forward ? items.length-1 : 0;
            firstInGroup['index'] = index;
            firstInGroup['elem'] = items[index];
        }

        return firstInGroup;
    },

    /**
     * Returns the element which is the direct parent of all item elements.  Only for non-hier data.
     * @private
     */
    _getRootNodeForItems: function()
    {
        return this.isCardLayout() ? this.element.get(0).firstElementChild.firstElementChild : this.element.get(0);
    },

    /**
     * Returns the scroll position object containing info about current scroll position.
     * @private
     */
    _getCurrentScrollPosition: function(scrollTop)
    {
        var scroller, scrollPosition = {}, isHierData, items, prevScrollPosition, diff, index, result, elem, parent;

        scroller = this._getScroller();
        if (scrollTop === undefined)
        {
            scrollTop = scroller.scrollTop;
        }

        scrollPosition['x'] = this._getScrollX(scroller);  
        scrollPosition['y'] = scrollTop;

        isHierData = this.m_contentHandler.IsHierarchical();

        // we used the item height to approximate where to begin the search
        // for the top most item.  This var should be populated in renderComplete
        // if there's no data then we should skip
        if (!isNaN(this.m_itemHeight) && this.m_itemHeight > 0)
        {
            // getItemsCache returns a flat view of all expanded items
            items = isHierData ? this._getItemsCache() : $(this._getRootNodeForItems()).children("li."+this.getItemElementStyleClass());

            // if the previous scroll position is relatively close to the current one
            // we'll use the previous index as the starting point
            prevScrollPosition = this.GetOption("scrollPosition");
            diff = Math.abs(prevScrollPosition['y'] - scrollTop);
            if (diff < this.MINIMUM_ITEM_HEIGHT && prevScrollPosition['key'] != null && !isNaN(prevScrollPosition['index']))
            {
                if (isHierData)
                {
                    elem = this.FindElementByKey(prevScrollPosition['key']);
                    if (elem != null)
                    {
                        index = items.index(elem);
                    }
                }
                else
                {
                    index = prevScrollPosition['index'];
                }
            }

            // we'll need to approximate the index
            if (isNaN(index))
            {
                index = Math.floor(scrollTop / this.m_itemHeight);
            }

            result = this._findClosestElementToTop(items, index, scrollTop);
            if (result != null)
            {
                elem = result['elem'];
                if (isHierData)
                {
                    parent = elem.parentNode;
                    if (parent != this.element.get(0))
                    {
                        scrollPosition['parent'] = this.GetKey(parent.parentNode);
                    }
                    scrollPosition['key'] = this.GetKey(result['elem']);
                    scrollPosition['index'] = $(parent).children().index(elem);
                }
                else
                {
                    scrollPosition['index'] = result['index'];
                    scrollPosition['key'] = this.GetKey(result['elem']);                    
                }
                scrollPosition['offsetY'] = result['offset'];
                // offsetX is the same as x, even when card layout is used
                // since listview wraps card on space available, there will never be a listview
                // having 2 or more columns with a horizontal scrollbar
                scrollPosition['offsetX'] = scrollPosition['x'];
            }
        }

        return scrollPosition;      
    },

    /**
     * @private
     */
    _getOffsetTop: function(elem)
    {
        var offsetTop = this.element.get(0).offsetTop;
        if (!isNaN(this.m_elementOffset) && this.m_elementOffset != offsetTop)
        {
            return Math.max(0, elem.offsetTop - offsetTop);
        }
        else
        {
            return elem.offsetTop;
        }
    },

    /**
     * Retrieve the scroll top value based on item index (optionally with parent key)
     * @private
     */
    _getScrollTopByIndex: function(index, parent)
    {
        var parentElem, elem;

        if (parent != null)
        {
            parentElem = this.FindElementByKey(parent);
            if (parentElem != null)
            {
                // find the ul element
                parentElem = $(parentElem).children("ul").first();
            }
        }
        else
        {
            parentElem = this.element.get(0);
            if (this.isCardLayout())
            {
                parentElem = parentElem.firstElementChild.firstElementChild;
            }
        }

        if (parentElem != null)
        {
            elem = $(parentElem).children("."+this.getItemElementStyleClass())[index];
            if (elem != null)
            {
                return this._getOffsetTop(elem);
            }
        }

        // we got here because one of the following happened:
        // 1) item has not been fetched yet
        // 2) index is large than the number of items, including reaching maxCount
        // 3) parent key specified does not exists or has not been fetched yet
        if (this.m_contentHandler.hasMoreToFetch && this.m_contentHandler.hasMoreToFetch())
        {
            return this._getScroller().scrollHeight;
        }
        return;
    },

    /**
     * Retrieve the scroll top value based on item key
     * @private
     */
    _getScrollTopByKey: function(key)
    {
        var elem = this.FindElementByKey(key);
        if (elem != null)
        {
            return this._getOffsetTop(elem);
        }

        // we got here because one of the following happened:
        // 1) item has not been fetched yet
        // 2) key does not exists or invalid
        if (this.m_contentHandler.hasMoreToFetch && this.m_contentHandler.hasMoreToFetch())
        {
            return this._getScroller().scrollHeight;
        }
        return;
    },

    /**
     * Gets the scroll coordinate based on value of scrollPosition.
     * @return {Object} the coordinate to scroll to, see syncScrollPosition
     * @private
     */
    _getScrollCoordinates: function(scrollPosition)
    {
        var scroller, x, y, parent, index, key, offsetX, offsetY;

        scroller = this._getScroller();

        x = scrollPosition['x'];
        offsetX = scrollPosition['offsetX'];
        if (!isNaN(x) && !isNaN(offsetX))
        {
            x = x + offsetX;
        }

        // key first
        key = scrollPosition['key'];
        if (isNaN(y) && key != null)
        {
            y = this._getScrollTopByKey(key);
        }

        // then index
        parent = scrollPosition['parent'];
        index = scrollPosition['index'];
        if (isNaN(y) && !isNaN(index))
        {
            y = this._getScrollTopByIndex(index, parent);
        }

        offsetY = scrollPosition['offsetY'];
        if (!isNaN(y) && !isNaN(offsetY))
        {
            y = y + offsetY;
        }
        
        // then pixel position last
        if (isNaN(y) && !isNaN(scrollPosition['y']))
        {
            y = scrollPosition['y'];            
        }

        return {x: x, y: y};
    },

    /**
     * Scroll handler
     * @private
     */
    _handleScroll: function(event)
    {
        var scrollTop;

        // since we are calling it from requestAnimationFrame, ListView could have been destroyed already
        if (this.m_contentHandler == null)
        {
            return;
        }

        // update scrollPosition
        scrollTop = this._getScroller().scrollTop;
        if (!this.ojContext._IsCustomElement())
        {
            this.SetOption('scrollTop', scrollTop, {'_context': {originalEvent: event, internalSet: true}});
        }

        if (this.ShouldUpdateScrollPosition())
        {
            this.SetOption('scrollPosition', this._getCurrentScrollPosition(scrollTop), {'_context': {originalEvent: event, internalSet: true}});
        }

        // handle pinning group header, does not need if position sticky is supported 
        this._handlePinGroupHeader();
    },

    /**
     * Register scroll listener
     * @private
     */
    _registerScrollHandler: function()
    {
        var self = this, scrollElem;

        scrollElem = $(this._getScrollEventElement());

        this.ojContext._off(scrollElem, "scroll mousewheel");

        this.ojContext._on(scrollElem, {
            "scroll": function(event)
            {
                // throttle the event using requestAnimationFrame for performance reason
                // don't update if scroll is triggered by listview internally setting scrollLeft/scrollTop
                if (!self._skipScrollUpdate && !self.m_ticking)
                {
                    window.requestAnimationFrame(function()
                    {
                        self._handleScroll(event);
                        self.m_ticking = false;
                    });

                    self.m_ticking = true;
                }

                if (self._skipScrollUpdate)
                {
                    self.signalTaskEnd();
                }

                self._skipScrollUpdate = false;
            }
        });

        // only do this for highwatermark scrolling, other cases we have (and should not care) no knowledge about the scroller
        if (this.options['scrollPolicy'] == "loadMoreOnScroll")
        {
            this.ojContext._on(scrollElem, {
                "mousewheel": function(event)
                {
                    self._preventMouseWheelOverscroll(self._getScroller(), event);
                }
            });
        }
    },

    /**
     * Whether group header should be pin
     * @return {boolean} true if group header should be pin or false otherwise
     * @private
     */
    _isPinGroupHeader: function()
    {
        return (this.GetOption("groupHeaderPosition") != "static" && this.m_contentHandler.IsHierarchical());
    },

    /**
     * Retrieve the visible (flattened) group items cache, create one if it is null.
     * @return {jQuery} a list of group items
     * @private
     */
    _getGroupItemsCache: function()
    {
        var selector;

        if (this.m_groupItems == null)
        {
            selector = "."+this.getGroupItemStyleClass() + ":visible";
            this.m_groupItems = this.element.find(selector).filter(function(index)
            {
                // if it's expanded and it has children
                if (!$(this).parent().hasClass("oj-collapsed"))
                {
                    if ($(this).next().children().length > 0)
                    {
                        return true;
                    }
                }

                return false;
            });
        }
        return this.m_groupItems;
    },

    /**
     * Unpin a pinned group header
     * @param {Element} groupItem the group header element to unpin
     * @private
     */
    _unpinGroupItem: function(groupItem)
    {
        $(groupItem).removeClass("oj-pinned");
        groupItem.style.top = "auto";
        groupItem.style.width = "auto";
    },

    /**
     * Gets the next group item.  This could be a group item from a different parent.
     * @param {Element} groupItem the reference group item.
     * @return {Element|null} the next group item or null if one cannot be found
     * @private
     */
    _getNextGroupItem: function(groupItem)
    {
        var groupItems, index;

        groupItems = this._getGroupItemsCache();
        index = groupItems.index(groupItem);
        if (index > -1 && index < groupItems.length-1)
        {
            return groupItems[index+1];
        }

        return null;
    },

    /**
     * Pin a group header
     * @param {Element} groupItem the group header element to pin
     * @param {number} scrollTop the scrolltop position of the listview container
     * @private
     */
    _pinGroupItem: function(groupItem, scrollTop)
    {
        var width, height, next;

        width = groupItem.offsetWidth;
        height = groupItem.offsetHeight;

        next = this._getNextGroupItem(groupItem);
        // todo: get rid of 5
        if (next != null && next.offsetTop <= scrollTop + height + 5)
        {
            scrollTop = scrollTop - height;
        }

        $(groupItem).addClass("oj-pinned");
        groupItem.style.top = scrollTop + 'px';
        groupItem.style.width = width + 'px';
    },

    /**
     * Pin the header as neccessary when user scrolls.
     * @private
     */
    _handlePinGroupHeader: function()
    {
        var scroller, scrollTop, groupItemToPin, groupItems, pinHeaderHeight, i, groupItem, top, bottom, next;

        // if groupHeaderPosition is not sticky or if position:sticky is supported natively in the browser
        if (!this._isPinGroupHeader() || this._isPositionStickySupported())
        {
            return;
        }

        scroller = this._getScroller();
        scrollTop = scroller.scrollTop;

        // see if we are at the top
        if (this.m_groupItemToPin != null && scrollTop == 0)
        {
            this._unpinGroupItem(this.m_groupItemToPin);
            this.m_groupItemToPin = null;
            return;
        }

        // find the group item to pin
        groupItems = this._getGroupItemsCache();
        pinHeaderHeight = 0;
        if (this.m_groupItemToPin != null)
        {
            pinHeaderHeight = this.m_groupItemToPin.offsetHeight;
        }

        for (i=0; i<groupItems.length; i++)
        {
            groupItem = groupItems[i];
            if (this.m_groupItemToPin == groupItem)
            {
                continue;
            }

            top = groupItems[i].offsetTop;
            bottom = top + groupItem.parentNode.offsetHeight;

            // if bottom half is in view but not the top half
            if (top < scrollTop && bottom > scrollTop + pinHeaderHeight)
            {
                groupItemToPin = groupItem;
                break;
            }
        }

        // found the group item to pin
        if (groupItemToPin != null && groupItemToPin != this.m_groupItemToPin)
        {
            // unpin the previous item
            if (this.m_groupItemToPin != null)
            {
                this._unpinGroupItem(this.m_groupItemToPin);
            }

            this._pinGroupItem(groupItemToPin, scrollTop);
            this.m_groupItemToPin = groupItemToPin;
        }
        else if (this.m_groupItemToPin != null)
        {
            // is the current pin header touching the next item
            next = this._getNextGroupItem(this.m_groupItemToPin);
            if (next != null && next.offsetTop <= scrollTop + pinHeaderHeight)
            {
                // make sure they really touches
                this.m_groupItemToPin.style.top = (next.offsetTop - pinHeaderHeight) + 'px';
                return;
            }

            this.m_groupItemToPin.style.top = scrollTop + 'px';
        }
    },

    /**
     * Gets the scroller element, which is either the listview container or the scroller element
     * specified in scrollPolicyOptions
     * @return {Element} the scroller element
     * @private
     */
    _getScroller: function()
    {
        var options, scroller;

        options = this.GetOption("scrollPolicyOptions");
        if (options != null)
        {
            scroller = options['scroller'];
            if (scroller != null)
            {
                return scroller;
            }
        }

        return this._getListContainer().get(0);
    },

    /**
     * Scroll to the specified group header
     * @param {Element} groupHeader the group header div element
     * @private
     */
    _scrollToGroupHeader: function(groupHeader)
    {
        var scroller, currentScrollTop, newScrollTop;

        scroller = this._getScroller();
        currentScrollTop = scroller.scrollTop;

        // unpin any pinned group header first before scroll to header
        if (this.m_groupItemToPin != null)
        {
            this._unpinGroupItem(this.m_groupItemToPin);
            this.m_groupItemToPin = null;
        }

        newScrollTop = groupHeader.offsetTop;
        // when scrolling backwards, the offsetTop is going to take position sticky into account, so it will
        // scroll to the minimum where the header is visible, and as a result all children items would not be visible
        if (this._isPinGroupHeader() && this._isPositionStickySupported() && newScrollTop < currentScrollTop)
        {
            newScrollTop = Math.max(0, newScrollTop - groupHeader.parentNode.offsetHeight + groupHeader.offsetHeight);
        }
        scroller.scrollTop = newScrollTop;

        // if it wasn't scroll (ex: already at the end), we'll have to explicitly try to see if we need to pin again
        if (currentScrollTop == scroller.scrollTop)
        {
            this._handlePinGroupHeader();
        }

        // set the first item in group current
        this._setFirstFocusableItemInGroupCurrent(groupHeader);
    },

    /**
     * Find the first focusable item within the group and make it current
     * @param {Element} groupHeader the group header 
     * @private
     */
    _setFirstFocusableItemInGroupCurrent: function(groupHeader)
    {
        var self = this, items, item;

        items = $(groupHeader).next().children();
        items.each(function() 
        {
            item = $(this);

            // make sure item can receive focus
            if (!self.SkipFocus(item))
            {
                self.SetOption('currentItem', this.key);
                return false;
            }
        });
    }
    /********************************* End Pin Header *******************************************/
});

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * @ojcomponent oj.ojListView
 * @augments oj.baseComponent
 * @ojtsimport ojkeyset
 * @since 1.1.0
 * @ojstatus preview
 * @ojtsimport ojdataprovider
 * @ojtsimport ojkeyset
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojListView<K, D> extends baseComponent<ojListViewSettableProperties<K,D>>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojListViewSettableProperties<K,D> extends baseComponentSettableProperties",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @ojshortdesc Displays data items as a list or a grid with highly interactive features.
 * @ojrole grid
 * @classdesc
 * <h3 id="listViewOverview-section">
 *   JET ListView Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#listViewOverview-section"></a>
 * </h3>
 *
 * <p>Description: The JET ListView enhances a HTML list element into a themable, WAI-ARIA compliant, mobile friendly component with advance interactive features.</p>
 *
 * <h3 id="data-section">
 *   Data
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#data-section"></a>
 * </h3>
 * <p>The JET ListView gets its data in three different ways.  The first way is from a DataProvider/TableDataSource.  There are several types of DataProvider/TableDataSource 
 * that are available out of the box:</p>
 * <ul>
 * <li>oj.ArrayDataProvider</li>
 * <li>oj.CollectionTableDataSource</li>
 * <li>oj.PagingTableDataSource</li>
 * </ul>
 *
 * <p><b>oj.ArrayDataProvider</b> - Use this when the underlying data is an array object or an observableArray.  In the observableArray case, ListView will automatically react
 * when items are added or removed from the array.  See the documentation for oj.ArrayDataProvider for more details on the available options.</p>
 *
 * <p><b>oj.CollectionTableDataSource</b> - Use this when oj.Collection is the model for the underlying data.  Note that the ListView will automatically react to model event from
 * the underlying oj.Collection.  See the documentation for oj.CollectionTableDataSource for more details on the available options.</p>
 *
 * <p><b>oj.PagingTableDataSource</b> - Use this when the ListView is driven by an associating ojPagingControl.  See the documentation for oj.PagingTableDataSource for more
 * details on the available options.</p>
 *
 * <p>The second way is from a TreeDataSource.  This is typically used to display data that are logically categorized in groups.  There are several types
 * of TreeDataSource that are available out of the box:</p>
 * <ul>
 * <li>oj.JsonTreeDataSource</li>
 * <li>oj.CollectionTreeDataSource</li>
 * </ul>
 *
 * <p><b>oj.JsonTreeDataSource</b> - Use this when the underlying data is a JSON object.  See the documentation for oj.JsonTreeDataSource for more details on the available options.</p>
 *
 * <p><b>oj.CollectionTreeDataSource</b> - Use this when oj.Collection is the model for each group of data.  See the documentation for oj.CollectionTableDataSource
 * for more details on the available options.</p>
 *
 * <p>Finally, ListView also supports static HTML content as data.  The structure of the content can be either flat or hierarhical.</p>
 *
 * <p>Example of flat static content</p>
 * <pre class="prettyprint">
 * <code>&lt;oj-list-view id="listView">
 *   &lt;ul>
 *     &lt;li>&lt;a id="item1" href="#">Item 1&lt;/a>&lt;/li>
 *     &lt;li>&lt;a id="item2" href="#">Item 2&lt;/a>&lt;/li>
 *     &lt;li>&lt;a id="item3" href="#">Item 3&lt;/a>&lt;/li>
 *   &lt;/ul>
 * &lt;/oj-list-view>
 * </code></pre>
 *
 * <p>Example of hierarchical static content</p>
 * <pre class="prettyprint">
 * <code>&lt;oj-list-view id="listView">
 *   &lt;ul>
 *     &lt;li>&lt;a id="group1" href="#">Group 1&lt;/a>
 *       &lt;ul>
 *         &lt;li>&lt;a id="item1-1" href="#">Item 1-1&lt;/a>&lt;/li>
 *         &lt;li>&lt;a id="item1-2" href="#">Item 1-2&lt;/a>&lt;/li>
 *       &lt;/ul>
 *     &lt;/li>
 *     &lt;li>&lt;a id="group2" href="#">Group 2&lt;/a>
 *       &lt;ul>
 *         &lt;li>&lt;a id="item2-1" href="#">Item 2-1&lt;/a>&lt;/li>
 *         &lt;li>&lt;a id="item2-2" href="#">Item 2-2&lt;/a>&lt;/li>
 *       &lt;/ul>
 *     &lt;/li>
 *   &lt;/ul>
 * &lt;/oj-list-view>
 * </code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="context-section">
 *   Item Context
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#context-section"></a>
 * </h3>
 *
 * <p>For all item options, developers can specify a function as the return value.  The function takes a single argument, which is an object that contains contextual information about the particular item.  This gives developers the flexibility to return different value depending on the context.</p>
 *
 * <p>The context paramter contains the following keys:</p>
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>componentElement</kbd></td>
 *       <td>A reference to the root element of ListView.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>datasource</kbd></td>
 *       <td>A reference to the data source object. (Not available for static content)</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>index</kbd></td>
 *       <td>The index of the item, where 0 is the index of the first item.  In the hierarchical case the index is relative to its parent.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>key</kbd></td>
 *       <td>The key of the item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>data</kbd></td>
 *       <td>The data object for the item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>parentElement</kbd></td>
 *       <td>The list item element.  The renderer can use this to directly append content.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <p></p>
 * <p>If the data is hierarchical, the following additional contextual information are available:</p>
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>depth</kbd></td>
 *       <td>The depth of the item.  The depth of the first level children under the invisible root is 1.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>parentKey</kbd></td>
 *       <td>The key of the parent item.  The parent key is null for root node.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>leaf</kbd></td>
 *       <td>Whether the item is a leaf or a group item.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>Application should specify a value for the aria-label attribute with a meaningful description of the purpose of this list.
 *
 * <p>Application must ensure that the context menu is available and setup with the
 * appropriate clipboard menu items so that keyboard-only users are able to reorder items
 * just by using the keyboard.  
 *
 * <p>Note that ListView uses the grid role and follows the <a href="https://www.w3.org/TR/wai-aria-practices/examples/grid/LayoutGrids.html">Layout Grid</a> design as outlined in the <a href="https://www.w3.org/TR/wai-aria-practices/#grid">grid design pattern</a>. 
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 * 
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it's recommended that applications limit the amount of data to display.  Displaying large
 * number of items in ListView makes it hard for user to find what they are looking for, but affects the load time and
 * scrolling performance as well.  If displaying large number of items is neccessary, use a paging control with ListView
 * to limit the number of items to display at a time.  Setting <code class="prettyprint">scrollPolicy</code> to
 * 'loadMoreOnScroll' will also reduce the number of items to display initially.</p>
 *
 * <h4>Item Content</h4>
 * <p>ListView allows developers to specify arbitrary content inside its item. In order to minimize any negative effect on
 * performance, you should avoid putting a large number of heavy-weight components inside because as you add more complexity
 * to the structure, the effect will be multiplied because there can be many items in the ListView.</p>
 *
 * <h4>Expand All</h4>
 * <p>While ListView provides a convenient way to initially expand all group items in the ListView, it might have an impact
 * on the initial rendering performance since expanding each group item might cause a fetch from the server depending on
 * the TreeDataSource.  Other factors that could impact performance includes the depth of the tree, and the number of children
 * in each level.</p>
 *
 * <h3 id="animation-section">
 *   Animation
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#animation-section"></a>
 * </h3>
 *
 * <p>Applications can customize animations triggered by actions in ListView by either listening for <code class="prettyprint">animateStart/animateEnd</code>
 *    events or overriding action specific style classes on the animated item.  See the documentation of <a href="oj.AnimationUtils.html">oj.AnimationUtils</a> 
 *    class for details.</p>
 *    
 * <p>The following are actions and their corresponding sass variables in which applications can use to customize animation effects.  
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Action</th>
 *       <th>Sass Variable</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>add</kbd></td>
 *       <td>$listViewAddAnimation</td>
 *       <td>When a new item is added to the oj.TableDataSource associated with ListView.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>remove</kbd></td>
 *       <td>$listViewRemoveAnimation</td>
 *       <td>When an existing item is removed from the oj.TableDataSource associated with ListView.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>update</kbd></td>
 *       <td>$listViewUpdateAnimation</td>
 *       <td>When an existing item is updated in the oj.TableDataSource associated with ListView.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>expand</kbd></td>
 *       <td>$listViewExpandAnimation</td>
 *       <td>When user expands a group item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>collapse</kbd></td>
 *       <td>$listViewCollapseAnimation</td>
 *       <td>When user collapses a group item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>pointerUp</kbd></td>
 *       <td>$listViewPointerUpAnimation</td>
 *       <td>When user finish pressing an item (on touch).</td>
 *     </tr>
 *   </tbody>
 * </table>
 */
oj.__registerWidget('oj.ojListView', $['oj']['baseComponent'],
{
    widgetEventPrefix: 'oj',
    options:
            {
               /**
                * An alias for the current item when referenced inside the item template. This can be especially useful
                * if oj-bind-for-each element is used inside the item template since it has its own scope of data access.
                *
                * @ojshortdesc Gets and sets the alias for the current item when referenced inside the item template.
                * @ojstatus preview
                * @expose
                * @public
                * @instance
                * @memberof! oj.ojListView
                * @type {string}
                * @default ''
                *
                * @example <caption>Initialize the ListView with the <code class="prettyprint">as</code> attribute specified:</caption>
                * &lt;oj-list-view as='item'>
                *   &lt;template slot='itemTemplate'>
                *     &lt;p>&lt;oj-bind-text value='[[item.data.name]]'>&lt;/oj-bind-text>&lt;/p>
                *   &lt;/template>
                * &lt;/oj-list-view>
                */ 
                as: '',
               /**
                * The item that currently have keyboard focus.  Note that if current item
                * is set to an item that is not available in the view (either not fetched in highwater mark scrolling case or
                * hidden inside a collapsed parent node), then the value is not applied.
                *
                * @ojshortdesc Gets and sets the key of the item that should have keyboard focus.
                * @expose
                * @public
                * @instance
                * @memberof! oj.ojListView
                * @type {*}
                * @ojsignature {target:"Type", value:"K"}
                * @default null
                * @ojwriteback
                *
                * @example <caption>Initialize the ListView with the <code class="prettyprint">current-item</code> attribute specified:</caption>
                * &lt;oj-list-view current-item='{{myCurrentItem}}'>&lt;/oj-list-view>
                *
                * @example <caption>Get or set the <code class="prettyprint">currentItem</code> property after initialization:</caption>
                * // getter
                * var currentItemValue = myListView.currentItem;
                *
                * // setter
                * myListView.currentItem = "item2";
                */
                currentItem: null,
                /**
                 * The data source for ListView.  Must be of type oj.TableDataSource, oj.TreeDataSource, oj.DataProvider
                 * See the data source section in the introduction for out of the box data source types.
                 * If the data attribute is not specified, the child elements are used as content.  If there's no
                 * content specified, then an empty list is rendered.
                 *
                 * @ojshortdesc Gets and sets the data provider for this list.
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {oj.TableDataSource|oj.TreeDataSource|oj.DataProvider}
                 * @default null
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">data</code> attribute specified:</caption>
                 * &lt;oj-list-view data='{{myDataSource}}'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">data</code> property after initialization:</caption>
                 * // getter
                 * var dataValue = myListView.data;
                 *
                 * // setter
                 * myListView.data = myDataSource;
				         * @ojsignature {target: "Type", value: "oj.DataProvider<K, D>"}
                 */
                data: null,
                /**
                 * Enable drag and drop functionality.<br><br>
                 * JET provides support for HTML5 Drag and Drop events.  Please refer to {@link https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/Drag_and_drop third party documentation} 
                 * on HTML5 Drag and Drop to learn how to use it.
                 *
                 * @ojshortdesc Customize the drag and drop features.
                 * @expose
                 * @memberof! oj.ojListView
                 * @type {Object}
                 * @default {"drag": null, "drop": null, "reorder": {"items": "disabled"}}
                 * @instance
                 */
                dnd: {
                   /**
                    * Enables and customize the drag functionalities.
                    * 
                    * @ojshortdesc Customize the drag functionalities.
                    * @expose
                    * @alias dnd.drag
                    * @memberof! oj.ojListView
                    * @instance
                    * @type {Object}
                    * @ojsignature { target: "Type",
                    *                value: "?"}
                    * @default null
                    */
                   drag: null,
                   /**
                    * @ojshortdesc Customize the drop functionalities.
                    * @expose
                    * @alias dnd.drop
                    * @memberof! oj.ojListView
                    * @instance
                    * @type {Object}
                    * @ojsignature { target: "Type", value: "?"}                   
                    * @default null
                    */
                   drop: null,
                   /**
                    * The reorder option contains a subset of options for reordering items.
                    *
                    * @ojshortdesc Customize the item reordering functionalities.
                    * @expose
                    * @alias dnd.reorder
                    * @memberof! oj.ojListView
                    * @instance
                    * @type {Object}
                    */
                   reorder: {
                          /**
                           * Enable or disable reordering the items within the same listview using drag and drop.<br><br>
                           * Specify 'enabled' to enable reordering.  Setting the value 'disabled' or setting the <code class="prettyprint">"dnd"</code> property 
                           * to <code class="prettyprint">null</code> (or omitting it), disables reordering support. 
                           *
                           * @ojshortdesc Enable or disable the item reordering functionalities.
                           * @expose
                           * @alias dnd.reorder.items
                           * @memberof! oj.ojListView
                           * @instance
                           * @type {string}
                           * @ojvalue {string} "enabled" Item reordering is enabled.
                           * @ojvalue {string} "disabled" Item reordering is disabled.
                           * @default "disabled"
                           * 
                           * @example <caption>Initialize the ListView with the <code class="prettyprint">reorder</code> attribute specified:</caption>
                           * &lt;oj-list-view dnd.reorder.items='enabled'>&lt;/oj-list-view>
                           *
                           * @example <caption>Get or set the <code class="prettyprint">reorder</code> property after initialization:</caption>
                           * // getter
                           * var reorderValue = myListView.dnd.reorder.items;
                           *
                           * // setter
                           * myListView.dnd.reorder.items = 'enabled';
                           */
                          items:'disabled'
                   }
                },
                /**
                 * Changes the expand and collapse operations on ListView.  If "none" is specified, then
                 * the current expanded state is fixed and user cannot expand or collapse an item.
                 *
                 * @ojshortdesc Gets and sets whether expand or collapse operations are allowed.
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {string}
                 * @default "collapsible"
                 * @ojvalue {string} "collapsible" Group item can be expanded or collapsed by user.
                 * @ojvalue {string} "none" The expand state of a group item cannot be changed by user.
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">drill-mode</code> attribute specified:</caption>
                 * &lt;oj-list-view drill-mode='none'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">drillMode</code> property after initialization:</caption>
                 * // getter
                 * var drillModeValue = myListView.drillMode;
                 *
                 * // setter
                 * myListView.drillMode = 'none';
                 */
                drillMode: "collapsible",
                /**
                 * Specifies the key set containing the keys of the items that should be expanded. 
                 *
                 * Use the <a href="ExpandedKeySet.html">ExpandedKeySet</a> class to specify items to expand.
                 * Use the <a href="ExpandAllKeySet.html">ExpandAllKeySet</a> class to expand all items.                 
                 *
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @default new ExpandedKeySet();
                 * @type {KeySet}
                 * @ojsignature {target:"Type", value:"KeySet<K>"}
                 *
                 * @ojwriteback
                 *
                 * @example <caption>Initialize the ListView with specific items expanded:</caption>
                 * myListView.expanded = new ExpandedKeySet(['item1', 'item2']);
                 *
                 * @example <caption>Initialize the ListView with all items expanded:</caption>
                 * myListView.expanded = new ExpandAllKeySet();
                 */
                expanded: new oj._ojListViewExpandedKeySet(),
                /**
                 * Gets the key and data of the first selected item.  The first selected item is defined as the first
                 * key returned by the <a href="#selection">selection</a> property.  The value of this property contains:
                 * <ul>
                 * <li>key - the key of the first selected item.</li>
                 * <li>data - the data of the first selected item.  If the selected item is not locally available, this will
                 *        be null.  If the <a href="#data">data</a> property is not set and that static HTML element is used
                 *        as data, then this will be the item element.</li>
                 * </ul>
                 * If no items are selected then this property will return an object with both key and data properties set to null.
                 *
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @default {'key': null, 'data': null}
                 * @property {*} key The key of the first selected item
                 * @property {*} data The data of the first selected item
                 * @type {Object}
                 * @ojsignature {target:"Type", value:"{key: K, data: D}"}
                 *
                 * @ojwriteback
                 * @readonly
                 *
                 * @example <caption>Get the data of the first selected item:</caption>
                 * // getter
                 * var firstSelectedItemValue = myListView.firstSelectedItem;
                 */
                firstSelectedItem: {'key': null, 'data': null},
                /**
                 * Specifies how the group header should be positioned.  If "sticky" is specified, then the group header 
                 * is fixed at the top of the ListView as the user scrolls.
                 *
                 * @ojshortdesc Gets and sets whether group header should stick to the top as user scrolls.
                 * @expose 
                 * @memberof! oj.ojListView
                 * @instance
                 * @default "sticky"
                 * @ojvalue {string} "static" The group header position updates as user scrolls.
                 * @ojvalue {string} "sticky" The group header is fixed at the top when user scrolls.
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">group-header-position</code> attribute specified:</caption>
                 * &lt;oj-list-view group-header-position='static'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">groupHeaderPosition</code> property after initialization:</caption>
                 * // getter
                 * var groupHeaderPositionValue = myListView.groupHeaderPosition;
                 *
                 * // setter
                 * myListView.groupHeaderPosition = 'static';
                 */
                groupHeaderPosition: "sticky",
                /**
                 * The item option contains a subset of options for items.
                 *
                 * @ojshortdesc Customize the functionalities of each item on the list.  
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 */
                item: {
                    /**
                     * @typedef {Object} oj.ojListView.ItemContext
                     * @property {oj.DataProvider<K, D>} datasource the data source/data provider 
                     * @property {number} index the zero based index of the item, relative to its parent
                     * @property {K} key the key of the item
                     * @property {D} data the data object of the item
                     * @property {Element} parentElement the item DOM element
                     * @property {number=} depth the depth of the item
                     * @property {K=} parentKey the key of the parent item
                     * @property {boolean=} leaf whether the item is a leaf
                     * @ojsignature [{target:"Type", value:"<K,D>", for:"genericTypeParameters"}]
                     */
                    /**
                     * Whether the item is focusable.  An item that is not focusable cannot be clicked on or navigated to.
                     * See <a href="#context-section">itemContext</a> in the introduction to see the object passed into the focusable function.
                     *
                     * @ojshortdesc Gets and sets whether item can receive keyboard focus.
                     * @expose
                     * @alias item.focusable
                     * @memberof! oj.ojListView
                     * @instance
                     * @type {boolean|function(Object):boolean}
                     * @ojsignature { target: "Type",
                     *                value: "?((param0: oj.ojListView.ItemContext<K,D>) => boolean)|boolean",
                     *                jsdocOverride: true}
                     * @default true
                     *
                     * @example <caption>Initialize the ListView with the <code class="prettyprint">focusable</code> attribute specified:</caption>
                     * &lt;oj-list-view item.focusable='{{myFocusableFunc}}'>&lt;/oj-list-view>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">focusable</code> property after initialization:</caption>
                     * // getter
                     * var focusable = myListView.item.focusable;
                     *
                     * // setter
                     * myListView.item.focusable = myFocusableFunc;
                     */
                    focusable: true,
                    /**
                     * The renderer function that renders the content of the item. See <a href="#context-section">itemContext</a>
                     * in the introduction to see the object passed into the renderer function.
                     * The function should return one of the following: 
                     * <ul>
                     *   <li>An Object with the following property:
                     *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the item.</li></ul>
                     *   </li>
                     *   <li>undefined: If the developer chooses to manipulate the list element directly, the function should return undefined.</li>
                     * </ul>
                     * If no renderer is specified, ListView will treat the data as a string.
                     *
                     * @ojshortdesc Gets and sets the renderer for the item.
                     * @expose
                     * @alias item.renderer
                     * @memberof! oj.ojListView
                     * @instance
                     * @type {null|function(Object):Object}  
                     * @ojsignature { target: "Type",
                     *                value: "?((param0: oj.ojListView.ItemContext<K,D>) => {insert: Element|string}|undefined)|null",
                     *                jsdocOverride: true}                 
                     * @default null
                     *
                     * @example <caption>Initialize the ListView with the <code class="prettyprint">renderer</code> attribute specified:</caption>
                     * &lt;oj-list-view item.renderer='{{myRendererFunc}}'>&lt;/oj-list-view>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">renderer</code> property after initialization:</caption>
                     * // getter
                     * var renderer = myListView.item.renderer;
                     *
                     * // setter
                     * myListView.item.renderer = myRendererFunc;
                     */
                    renderer: null,
                    /**
                     * Whether the item is selectable.  Note that if selectionMode is set to "none" this option is ignored.  In addition,
                     * if focusable is set to false, then the selectable option is automatically overridden and set to false also.
                     * See <a href="#context-section">itemContext</a> in the introduction to see the object passed into the selectable function.
                     *
                     * @ojshortdesc Gets and sets whether the item can be selected.
                     * @expose
                     * @alias item.selectable
                     * @memberof! oj.ojListView
                     * @instance
                     * @type {boolean|function(Object):boolean}
                     * @ojsignature { target: "Type",
                     *                value: "?((param0: oj.ojListView.ItemContext<K,D>) => boolean)|boolean",
                     *                jsdocOverride: true}
                     * @default true
                     *
                     * @example <caption>Initialize the ListView with the <code class="prettyprint">selectable</code> attribute specified:</caption>
                     * &lt;oj-list-view item.selectable='{{mySelectableFunc}}'>&lt;/oj-list-view>
                     *
                     * @example <caption>Get or set the <code class="prettyprint">selectable</code> property after initialization:</caption>
                     * // getter
                     * var selectable = myListView.item.selectable;
                     *
                     * // setter
                     * myListView.item.selectable = mySelectableFunc;
                     */
                    selectable: true
                    /**
                     * The knockout template used to render the content of the item.
                     *
                     * This attribute is only exposed via the <code class="prettyprint">ojComponent</code> binding, and is not a
                     * component option.
                     *
                     * @ojbindingonly
                     * @name item.template
                     * @memberof! oj.ojListView
                     * @instance
                     * @type {string|null}
                     * @default null
                     *
                     * @example <caption>Specify the <code class="prettyprint">template</code> when initializing ListView:</caption>
                     * // set the template
                     * &lt;ul id="listview" data-bind="ojComponent: {component: 'ojListView', data: dataSource, item: {template: 'my_template'}}"&gt;&lt;/ul&gt;
                     */
                },
                /**
                 * Specifies the mechanism used to scroll the data inside the list view. Possible values are: auto and loadMoreOnScroll.
                 * When loadMoreOnScroll is specified, additional data is fetched when the user scrolls to the bottom of the ListView.
                 * Note that currently this option is only available when TableDataSource is used.
                 *
                 * @ojshortdesc Gets and sets how data are fetch as user scrolls down the list.
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {string|null}
                 * @default "auto"
                 * @ojvalue {string} "auto" The behavior is determined by the component.
                 * @ojvalue {string} "loadMoreOnScroll" Additional data is fetched when the user scrolls to the bottom of the ListView.
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">scroll-policy</code> attribute specified:</caption>
                 * &lt;oj-list-view scroll-policy='loadMoreOnScroll'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">scrollPolicy</code> property after initialization:</caption>
                 * // getter
                 * var scrollPolicyValue = myListView.scrollPolicy;
                 *
                 * // setter
                 * myListView.scrollPolicy = 'loadMoreOnScroll';
                 */
                scrollPolicy: "auto",
                /**
                 * scrollPolicy options.
                 * <p>
                 * The following options are supported:
                 * <ul>
                 *   <li>fetchSize: The number of items fetched each time when scroll to the end.</li>
                 *   <li>maxCount: Maximum rows which will be displayed before fetching more rows will be stopped.</li>
                 *   <li>scroller: The element which listview uses to determine the scroll position as well as the maximum scroll position where scroll to the end will trigger a fetch.  If not specified then the widget element of listview is used.</li>
                 * </ul>
                 * When scrollPolicy is loadMoreOnScroll, the next block of rows is fetched
                 * when the user scrolls to the end of the list/scroller. The fetchSize option
                 * determines how many rows are fetched in each block.
                 * Note that currently this option is only available when non-hierarchical DataProvider or TableDataSource is used.
                 *
                 * @ojshortdesc Gets and sets the fetch options as user scrolls down the list that triggers fetch of data.
                 * @expose
                 * @instance
                 * @memberof! oj.ojListView
                 * @type {Object.<string, number>|null}
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">scroll-policy-options</code> attribute specified:</caption>
                 * &lt;oj-list-view scroll-policy-options.fetch-size='30'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">scroll-policy-options</code> attribute after initialization:</caption>
                 * // getter
                 * var fetchSizeValue = myListView.scrollPolicyOptions.fetchSize;
                 *
                 * // setter
                 * myListView.scrollPolicyOptions.fetchSize = 30;

                 * @example <caption>Initialize the ListView with the <code class="prettyprint">scroll-policy-options</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-list-view scroll-policy-options.fetch-size='30' scroll-policy-options.max-count='1000'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">scrollPolicyOptions</code> property after initialization:</caption>
                 * // Get one
                 * var fetchSizeValue = myListView.scrollPolicyOptions.fetchSize;
                 *
                 * // Get all
                 * var scrollPolicyOptionsValues = myListView.scrollPolicyOptions;
                 *
                 * // Set one, leaving the others intact
                 * myListView.setProperty('scrollPolicyOptions.fetchSize', 30);
                 *
                 * // Set all.
                 * myListView.scrollPolicyOptions = {fetchSize: 30, maxCount: 1000};
                 */
                scrollPolicyOptions: {'fetchSize': 25, 'maxCount': 500},
                /**
                 * The current scroll position of ListView. The scroll position is updated when either the vertical or horizontal scroll position
                 * (or its scroller, as specified in scrollPolicyOptions.scroller) has changed.  The value contains the x and y scroll position, 
                 * the index and key information of the item closest to the top of the viewport, as well as horizontal and vertical offset from the
                 * position of the item to the actual scroll position.
                 * <p>
                 * The default value contains just the scroll position.  Once data is fetched the 'index' and 'key' sub-properties will be added.
                 * If there is no data then the 'index' and 'key' sub-properties will not be available.
                 * </p>
                 * <p>
                 * When setting the scrollPosition property, applications can change any combination of the sub-properties.
                 * If multiple sub-properties are set at once they will be used in key, index, pixel order where the latter serves as hints.  
                 * If offsetX or offsetY are specified, they will be used to adjust the scroll position from the position where the key or index 
                 * of the item is located.
                 * </p>
                 * <p>
                 * If a sparse object is set the other sub-properties will be populated and updated once ListView has scrolled to that position.
                 * </p>
                 * <p>
                 * Also, if <a href="#scrollPolicy">scrollPolicy</a> is set to 'loadMoreOnScroll' and the scrollPosition is set to a value outside 
                 * of the currently rendered region, then ListView will attempt to fetch until the specified scrollPosition is satisfied or the end
                 * is reached (either at max count or there's no more items to fetch), in which case the scroll position will remain at the end.  
                 * The only exception to this is when the key specified does not exists and a DataProvider is specified for <a href="data">data</a>,
                 * then the scroll position will not change (unless other sub-properties like index or x/y are specified as well).
                 * </p>
                 *
                 * @ojshortdesc Gets and sets the scroll position of list view.
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {Object}
                 * @default {"x": 0, "y": 0}
                 * @property {number=} x the horizontal position in pixel
                 * @property {number=} y the vertical position in pixel
                 * @property {number=} index the zero-based index of the item.  If <a href="#scrollPolicy">scrollPolicy</a> is set to 'loadMoreOnScroll' 
                 * and the index is greater than maxCount set in <a href="#scrollPolicyOptions">scrollPolicyOptions</a>, then it will scroll and fetch
                 * until the end of the list is reached and there's no more items to fetch.
                 * @property {*=} parent the key of the parent where the index is relative to.  If not specified, then the root is assumed
                 * @property {*=} key the key of the item.  If DataProvider is used for <a href="data">data</a> and the key does not exists in the 
                 * DataProvider, then the value is ignored.  If DataProvider is not used then ListView will fetch and scroll until the item is found
                 * or the end of the list is reached and there's no more items to fetch.
                 * @property {number=} offsetX the horizontal offset in pixel relative to the item identified by key/index.
                 * @property {number=} offsetY the vertical offset in pixel relative to the item identified by key/index.
                 *
                 * @ojsignature [{target:"Type", value:"K", for:"parent"},
                 *               {target:"type", value:"K", for:"key"}]
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">scroll-position</code> attribute specified:</caption>
                 * &lt;!-- Using dot notation -->
                 * &lt;oj-list-view scroll-position.index='10'>&lt;/oj-list-view>
                 *
                 * &lt;!-- Using JSON notation -->
                 * &lt;oj-list-view scroll-position='{"index": 10}'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">scrollPosition</code> property after initialization:</caption>
                 * // Get one
                 * var scrollPositionValue = myListView.scrollPosition.index;
                 *
                 * // Set one, leaving the others intact
                 * myListView.setProperty('scrollPosition.index', 10);
                 *
                 * // Get all
                 * var scrollPositionValues = myListView.scrollPosition;
                 *
                 * // Set all.  Those not listed will be lost until the scroll completes and the remaining fields are populated.
                 * myListView.scrollPosition = {x: 0, y: 150};
                 */
                scrollPosition: {'x': 0, 'y': 0},
                /**
                 * The vertical scroll position of ListView.
                 *
                 * @ignore
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {number}
                 * @default 0
                 *
                 * @example <caption>Initialize the list view to a specific scroll position:</caption>
                 * $( ".selector" ).ojListView({ "scrollTop": 100 });
                 */
                scrollTop: 0,
                /**
                 * The current selections in the ListView. An empty array indicates nothing is selected.
                 *
                 * @ojshortdesc Gets and sets the keys of the selected items.
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {Array.<*>}
                 * @ojsignature {target:"Type", value:"Array<K>"}
                 * @default []
                 * @ojwriteback
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">selection</code> attribute specified:</caption>
                 * &lt;oj-list-view selection='{{mySelection}}'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">selection</code> property after initialization:</caption>
                 * // getter
                 * var selectionValue = myListView.selection;
                 *
                 * // setter
                 * myListView.selection = ['item1', 'item2', 'item3'];
                 */
                selection: [],
                /**
                 * Specifies whether selection can be made and the cardinality of selection in the ListView.
                 * Selection is initially disabled, but setting the value to null will disable selection.
                 *
                 * @ojshortdesc Gets and sets whether selection can be made and the cardinality of selection.
                 * @expose
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {string}
                 * @default "none"
                 * @ojvalue {string} "none" Selection is disabled.
                 * @ojvalue {string} "single" Only one item can be selected at a time.
                 * @ojvalue {string} "multiple" Multiple items can be selected at the same time.
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">selection-mode</code> attribute specified:</caption>
                 * &lt;oj-list-view selection-mode='multiple'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> property after initialization:</caption>
                 * // getter
                 * var selectionModeValue = myListView.selectionMode;
                 *
                 * // setter
                 * myListView.selectionMode = 'multiple';
                 */
                selectionMode: "none",
                /**
                 * Specifies whether ListView should enforce that there will always be an item selected when selection is enabled and there are one or 
                 * more selectable items.  Specifically, when this is enabled, then the first selectable item in ListView will be selected. 
                 * Furthermore, users will not be able to toggle selection such that it will result in no selected items.
                 * See <a href="#selectionMode">selectionMode</a> on how to enable/disable selection.  
                 * See <a href="#item.selectable">item.selectable</a> on how to enable/disable selection for individual item.  
                 * 
                 * @ojshortdesc Gets and sets whether there should be at least one item selected when selection is enabled.
                 * @expose
                 * @ojstatus preview
                 * @memberof! oj.ojListView
                 * @instance
                 * @type {boolean}
                 * @default false
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">selection-required</code> attribute specified:</caption>
                 * &lt;oj-list-view selection-required='true'>&lt;/oj-list-view>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">selectionRequired</code> property after initialization:</caption>
                 * // getter
                 * var selectionRequiredValue = myListView.selectionRequired;
                 *
                 * // setter
                 * myListView.selectionRequired = true;
                 */
                selectionRequired: false,
                /**
                 * Triggered when the default animation of a particular action is about to start.  The default animation can be cancelled by calling event.preventDefault.
                 *
                 * @ojshortdesc Event handler for when the default animation of a particular action is about to start.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {string} action the action that starts the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
                 * @property {Element} element the target of animation.  
                 * @property {function():void} endCallback if the event listener calls event.preventDefault to cancel the default animation, it must call the endCallback function when it finishes its own animation handling and when any custom animation ends.
                 */
                animateStart: null,
                /**
                 * Triggered when the default animation of a particular action has ended.  Note this event will not be triggered if application cancelled the default animation on animateStart.
                 *
                 * @ojshortdesc Event handler for when the default animation of a particular action has ended.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {string} action the action that started the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
                 * @property {Element} element the target of animation.  
                 */
                animateEnd: null,
                /**
                 * Triggered before the current item is changed via the <code class="prettyprint">current</code> option or via the UI.
                 *
                 * @ojshortdesc Event handler for when before the current item is changed.
                 * @ojcancelable
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {*} previousKey the key of the previous item
                 * @property {Element} previousItem the previous item
                 * @property {*} key the key of the new current item
                 * @property {Element} item the new current item
                 * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"},
                 *               {target:"Type", value:"K", for:"previousKey"},
                 *               {target:"Type", value:"K", for:"key"}]
                 */
                beforeCurrentItem: null,
                /**
                 * Triggered before an item is expanded via the <code class="prettyprint">expanded</code> option,
                 * the <code class="prettyprint">expand</code> method, or via the UI.
                 *
                 * @ojshortdesc Event handler for when an item is about to expand.
                 * @ojcancelable
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {*} key the key of the item to be expanded
                 * @property {Element} item the item to be expanded
                 * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"},
                 *               {target:"Type", value:"K", for:"key"}]
                 */
                beforeExpand: null,
                /**
                 * Triggered before an item is collapsed via the <code class="prettyprint">expanded</code> option,
                 * the <code class="prettyprint">collapse</code> method, or via the UI.
                 *
                 * @ojshortdesc Event handler for when an item is about to collapse.
                 * @ojcancelable
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {*} key the key of the item to be collapsed
                 * @property {Element} item the item to be collapsed
                 * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"}, 
                 *               {target:"Type", value:"K", for:"key"}]
                 */
                beforeCollapse: null,
                /**
                 * Triggered after an item has been collapsed via the <code class="prettyprint">expanded</code> option,
                 * the <code class="prettyprint">collapse</code> method, or via the UI.
                 *
                 * @ojshortdesc Event handler for after an item has collapsed.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {*} key The key of the item that was just collapsed.
                 * @property {Element} item The list item that was just collapsed.
                 * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"},
                 *               {target:"Type", value:"K", for:"key"}]
                 */
                collapse: null,
                /**
                 * Triggered when the copy action is performed on an item via context menu or keyboard shortcut.
                 *
                 * @ojshortdesc Event handler for after the copy action is performed on an item.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {Element[]} items an array of items in which the copy action is performed on
                 */
                copy: null,
                /**
                 * Triggered when the cut action is performed on an item via context menu or keyboard shortcut.
                 *
                 * @ojshortdesc Event handler for after the cut action is performed on an item.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {Element[]} items an array of items in which the cut action is performed on
                 */
                cut: null,
                /**
                 * Triggered after an item has been expanded via the <code class="prettyprint">expanded</code> option,
                 * the <code class="prettyprint">expand</code> method, or via the UI.
                 *
                 * @ojshortdesc Event handler for after an item has expanded.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {*} key The key of the item that was just expanded.
                 * @property {Element} item The list item that was just expanded.
                 * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"},
                 *               {target:"Type", value:"K", for:"key"}]
                 */
                expand: null,
                /**
                 * Triggered when the paste action is performed on an item via context menu or keyboard shortcut.
                 *
                 * @ojshortdesc Event handler for after the paste action is performed on an item.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {Element} item the element in which the paste action is performed on
                 */
                paste: null,
                /**
                 * Triggered after all items in the ListView has been rendered.  Note that in the highwatermark scrolling case,
                 * all items means the items that are fetched so far.
                 *
                 * @ignore
                 * @event
                 * @deprecated 2.0.0 Use the <a href="#whenReady">whenReady</a> method instead. 
                 * @memberof oj.ojListView
                 * @instance
                 * @property {Event} event <code class="prettyprint">jQuery</code> event object
                 * @property {Object} ui Parameters
                 *
                 * @example <caption>Initialize the ListView with the <code class="prettyprint">ready</code> callback specified:</caption>
                 * $( ".selector" ).ojListView({
                 *     "ready": function( event, ui ) {}
                 * });
                 *
                 * @example <caption>Bind an event listener to the <code class="prettyprint">ojready</code> event:</caption>
                 * $( ".selector" ).on( "ojready", function( event, ui ) {
                 *     // verify that the component firing the event is a component of interest 
                 *     if ($(event.target).is(".mySelector")) {} 
                 * });
                 */
                ready: null,
                /**
                 * Triggered after items are reorder within listview via drag and drop or cut and paste.
                 *
                 * @ojshortdesc Event handler for after the order of the item has changed through drag and drop or cut and paste action.
                 * @expose
                 * @event
                 * @memberof oj.ojListView
                 * @instance
                 * @property {Element[]} items an array of items that are moved
                 * @property {string} position the drop position relative to the reference item.  Possible values are "before", "after", "inside"
                 * @property {Element} reference the item where the moved items are drop on
                 */
                reorder: null
            },

    /**
     * Create the listview
     * @override
     * @memberof! oj.ojListView
     * @protected
     */
    _ComponentCreate: function()
    {
        this._super();
        this._setup();
    },

    /**
     * Initialize the listview
     * @private
     */
    _setup: function()
    {
        var opts = {}, expanded;

        opts.element = this.element;
        opts.OuterWrapper = this.OuterWrapper;
        opts.ojContext = this;

        // for backward compatibility, the default for expanded for legacy syntax
        // should be the same as before
        if (!this._IsCustomElement())
        {
            expanded = this.options["expanded"];
            // check if it's been updated by applications
            if (expanded instanceof oj._ojListViewExpandedKeySet)
            {
                this.options["expanded"] = "auto";
            }
        }

        opts = $.extend(this.options, opts);

        this.listview = new oj._ojListView();
        this.listview.init(opts);
    },

    /**
     * Initialize the listview after creation
     * @protected
     * @override
     * @memberof! oj.ojListView
     */
    _AfterCreate: function ()
    {
        this._super();

        // inject helper function for ContentHandler and custom renderer to use
        var self = this;
        this.listview._FixRendererContext = function(context) { return self._FixRendererContext(context) };
        this.listview._WrapCustomElementRenderer = function(renderer) { return self._WrapCustomElementRenderer(renderer) };

        this.listview.afterCreate();
    },

    /**
     * Sets up resources needed by listview
     * @memberof! oj.ojListView
     * @instance
     * @override
     * @protected
     */
    _SetupResources: function()
    {
        this._super();
        this.listview.setupResources();
    },

    /**
     * Release resources held by listview
     * @memberof! oj.ojListView
     * @instance
     * @override
     * @protected
     */
    _ReleaseResources: function()
    {
        this._super();
        this.listview.releaseResources();
    },

    /**
     * Gets the focus element
     * @override
     * @memberof! oj.ojListView
     * @instance
     * @protected
     * @since 5.0.0
     */
    GetFocusElement: function ()
    {
        return this.listview != null ? this.listview.GetFocusElement() : this._super();
    },

    /**
     * Destroy the list view
     * @memberof! oj.ojListView
     * @private
     */
    _destroy: function()
    {
        this.listview.destroy();
        this._super();
    },

    /**
     * When the <a href="#contextMenu">contextMenu</a> option is set, this method is called when the user invokes the context menu via
     * the default gestures: right-click, pressHold, and <kbd>Shift-F10</kbd>.  Components should not call this method directly.
     *
     * @param {!Object} menu The JET Menu to open as a context menu.  Always non-<code class="prettyprint">null</code>.
     * @param {!Event} event What triggered the menu launch.  Always non-<code class="prettyprint">null</code>.
     * @param {string} eventType "mouse", "touch", or "keyboard".  Never <code class="prettyprint">null</code>.
     * @private
     */
    _NotifyContextMenuGesture: function(menu, event, eventType)
    {
        this.listview.notifyContextMenuGesture(menu, event, eventType);
    },

    /**
     * Sets multiple options
     * @param {Object} options the options object
     * @param {Object} flags additional flags for option
     * @override
     * @private
     */
    _setOptions: function(options, flags)
    {
        var needRefresh = this.listview.setOptions(options, flags);

        // updates the options last
        this._super(options, flags);

        if (needRefresh)
        {
            this.listview.refresh();
        }
        else if (options['selectionRequired'] || options['selection'] || options['selectionMode'])
        {
            // if listview is not refresh, we'll need to ensure selectionRequired is enforced if set to true
            this.listview.enforceSelectionRequired();
        }
    },

   /**
    * Sets a single option
    * @param {Object} key the key for the option
    * @param {Object} value the value for the option
    * @param {Object} flags any flags specified for the option
    * @override
    * @private
    */
    _setOption: function(key, value, flags)
    {
        // checks whether value is valid for the key
        var valid = true, extraData;

        if (key == "selectionMode")
        {
            valid = (value == "none" || value == "single" || value == "multiple");
        }
        else if (key == "drillMode")
        {
            valid = (value == "collapsible" || value == "none");
        }
        else if (key == "scrollPolicy")
        {
            valid = (value == "auto" || value == "loadMoreOnScroll");
        }
        else if (key == "groupHeaderPosition")
        {
            valid = (value == "static" || value == "sticky");
        }
        else if (key == "firstSelectedItem")
        {
            // read only
            valid = false;
        }

        // update option if it's valid otherwise throw an error
        if (valid)
        {
            // inject additional metadata for selection
            if (key == "selection")
            {
                extraData = this.listview.getItems(value);
                flags = {'_context': {extraData: {'items': this._IsCustomElement() ? extraData : $(extraData)}}};
            }
            else if (key == "currentItem")
            {
                extraData = this.listview.getItems([value])[0];
                flags = {'_context': {extraData: {'items': this._IsCustomElement() ? extraData : $(extraData)}}};
            }

            this._super(key, value, flags);
        }
        else
        {
            throw "Invalid value: "+value+" for key: "+key;
        }
    },

   /**
    * Invoked when application calls oj.Components.subtreeAttached.
    * @override
    * @private
    */
    _NotifyAttached: function()
    {
        this.listview.notifyAttached();
    },

   /**
    * In browsers [Chrome v35, Firefox v24.5, IE9, Safari v6.1.4], blur and mouseleave events are generated for hidden content but not detached content,
    * so for detached content only, we must use this hook to remove the focus and hover classes.
    * @override
    * @private
    */
    _NotifyDetached: function()
    {
        this.listview.notifyDetached();
    },

   /**
    * Invoked when application calls oj.Components.subtreeShown.
    * @override
    * @private
    */
    _NotifyShown: function()
    {
        this.listview.notifyShown();
    },

    /********************************* public methods **************************************/
    /**
     * Returns a jQuery object containing the root dom element of the listview.
     *
     * <p>This method does not accept any arguments.
     *
     * @ignore
     * @expose
     * @override
     * @memberof oj.ojListView
     * @instance
     * @return {jQuery} the root DOM element of list
     */
    'widget' : function ()
    {
        return this.listview.getRootElement();
    },

    /**
     * Redraw the entire list view after having made some external modifications.
     *
     * <p>This method does not accept any arguments.
     *
     * @ojshortdesc Redraw the entire list.
     * @expose
     * @memberof oj.ojListView
     * @return {void}
     * @instance
     *
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myListView.refresh();
     */
    refresh: function()
    {
        this._super();
        this.listview.refresh();
    },

    /**
     * Returns a Promise that resolves when the component is ready, i.e. after data fetching, rendering, and animations complete. 
     * Note that in the highwatermark scrolling case, component is ready after data fetching, rendering, and associated animations of items fetched so far are complete.
     *
     * <p>This method does not accept any arguments.
     * 
     * @ignore
     * @ojshortdesc Returns a Promise that resolves when this component becomes ready.
     * @expose
     * @memberof oj.ojListView
     * @instance
     * @return {Promise} A Promise that resolves when the component is ready.
     */
    whenReady: function()
    {
        return this.listview.whenReady();
    },

    // @inheritdoc
    getNodeBySubId: function(locator)
    {
        return this.listview.getNodeBySubId(locator);
    },

    // @inheritdoc
    getSubIdByNode: function(node)
    {
        return this.listview.getSubIdByNode(node);
    },

    /**
     * @typedef {Object} oj.ojListView.ContextByNode
     * @property {string} subId the sub id that represents the element
     * @property {K} key the key of the item
     * @property {number} index the zero based index of the item, relative to its parent
     * @property {Element=} parent the parent group DOM element
     * @property {boolean=} group whether the item is a group item
     * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"}]
     */
    /**
     * {@ojinclude "name":"nodeContextDoc"}
     * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
     * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
     * @ojsignature { target: "Type",
     *                value: "oj.ojListView.ContextByNode<K>|null",
     *                jsdocOverride: true,
     *                for: "returns"}
     *
     * @example {@ojinclude "name":"nodeContextExample"}
     *
     * @expose
     * @instance
     * @memberof oj.ojListView
     */
    getContextByNode: function(node)
    {
        return this.listview.getContextByNode(node);
    },

    /**
     * Return the raw data for an item in ListView.  The item must have been already fetched.
     * @param {Object} context The context of the item to retrieve raw data.
     * @param {*=} context.key The key of the item.  If both index and key are specified, then key takes precedence.
     * @param {number=} context.index The index of the item relative to its parent.
     * @param {Element=} context.parent The parent node, not required if parent is the root.
     * @returns {*} data of the item.  If the item is not found or not yet fetched, returns null.  Also,
     * if static HTML is used as data (data attribute is not specified), then the element for the item is returned.
     * @ojshortdesc Gets the raw data of an item.
     * @export
     * @expose
     * @memberof oj.ojListView
     * @instance
     * @example <caption>Invoke the <code class="prettyprint">getDataForVisibleItem</code> method:</caption>
     * var data = myListView.getDataForVisibleItem( {'index': 2} );
     * @ojsignature [{target:"Type", value:"K", for:"context.key"},
     *               {target:"Type", value:"D", for:"returns"}]
     */
    getDataForVisibleItem: function(context)
    {
        return this.listview.getDataForVisibleItem(context);
    },

    /**
     * Expand an item.<p>
     * Note when vetoable is set to false, beforeExpand event will still be fired but the event cannot be veto.<p>
     *
     * @ignore
     * @expose
     * @memberof oj.ojListView
     * @instance
     * @param {Object} key the key of the item to expand
     * @param {boolean} vetoable whether the event should be vetoable
     */
    expand: function (key, vetoable)
    {
        this.listview.expandKey(key, vetoable, true, true);
    },

    /**
     * Collapse an item.<p>
     * Note when vetoable is set to false, beforeCollapse event will still be fired but the event cannot be veto.<p>
     *
     * @ignore
     * @expose
     * @memberof oj.ojListView
     * @instance
     * @param {Object} key the key of the item to collapse
     * @param {boolean} vetoable whether the event should be vetoable
     */
    collapse: function (key, vetoable)
    {
        this.listview.collapseKey(key, vetoable, true);
    },

    /**
     * Gets the key of currently expanded items.
     *
     * @ignore
     * @expose
     * @memberof oj.ojListView
     * @instance
     * @return {Array} array of keys of currently expanded items
     */
    getExpanded: function ()
    {
        return this.listview.getExpanded();
    },

    /**
     * Gets the IndexerModel which can be used with the ojIndexer.  The IndexerModel provided by ListView
     * by defaults returns a list of locale dependent characters.  See translations for the key used to return
     * all characters.  When a user selects a character in the ojIndexer ListView will scroll to the group 
     * header (or the closest one) with the character as its prefix.
     *
     * @expose
     * @memberof oj.ojListView
     * @ojdeprecated {since:"3.0.0", description:'Implements your own IndexerModel or use the <a href="oj.IndexerModelTreeDataSource.html">IndexerModelTreeDataSource</a> class instead.'} 
     * @instance
     * @return {Object} ListView's IndexerModel to be used with the ojIndexer
     */
    getIndexerModel: function ()
    {
        if (this.indexerModel == null && oj.ListViewIndexerModel)
        {
            this.indexerModel = new oj.ListViewIndexerModel(this.listview);
        }
        return this.indexerModel;
    },

    /**
     * Scrolls the list until the specified item is visible.  If the item is not yet loaded (if scrollPolicy is set to 'loadMoreOnScroll'), then no action is taken.
     *
     * @ojshortdesc Scrolls a loaded item to visible.
     * @param {Object} item An object with a 'key' property that identifies the item to scroll to.
     * @property {K} item.key the key of the item to scroll to.
     * @expose
     * @memberof oj.ojListView
     * @return {void}
     * @instance
     */
    scrollToItem: function (item)
    {
        this.listview.scrollToItem(item);
    },

    //** @inheritdoc */
    _CompareOptionValues: function(option, value1, value2)
    {
        switch(option)
        {
            case 'currentItem':
            case 'selection':
                return oj.Object.compareValues(value1, value2);
            default:
                return this._super(option, value1, value2);
        }
    }

    // Slots

    /**
     * <p>The <code class="prettyprint">itemTemplate</code> slot is used to specify the template for rendering each item in the list. The slot must be a &lt;template> element.</p>  
     *
     * <p>When the template is executed for each item, it will have access to the binding context containing the following properties:</p>
     * <ul>
     * <li>$current - an object that contains information for the current item being rendered
     *   <ul>
     *     <li>componentElement - the &lt;oj-list-view> custom element</li>
     *     <li>data - the data for the current item being rendered</li>
     *     <li>index - the zero-based index of the current item being rendered</li>
     *     <li>key - the key of the current item being rendered</li>
     *     <li>depth (available when hierarchical data is provided) - the depth of the current item being rendered.  The depth of the first level children under the invisible root is 1</li>
     *     <li>leaf (available when hierarchical data is provided) - whether the current item is a leaf node or not</li>
     *     <li>parentKey (available when hierarchical data is provided) - the key of the parent item.  The parent key is null for root nodes</li>
     *   </ul>
     * </li>
     * <li>alias - if as attribute was specified, the value will be used to provide an application-named alias for $current.
     * </li>
     * </ul>
     *
     * <p>The content of the template should not include the &lt;li> element, only what's inside it.</p>
     * <p>When both template and item.renderer are specified, the item.renderer takes precedence.</p>
     *
     * @ojstatus preview
     * @ojslot itemTemplate
     * @memberof oj.ojListView
     *
     * @example <caption>Initialize the ListView with an inline item template specified:</caption>
     * &lt;oj-list-view>
     *   &lt;template slot='itemTemplate'>
     *     &lt;span>&lt;oj-bind-text value='[[$current.data.name]]'>&lt;/span>
     *   &lt;template>
     * &lt;/oj-list-view>
     */

    // Fragments
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="2">List Item</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Focus on the item.  If <code class="prettyprint">selectionMode</code> is enabled, selects the item as well.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display context menu</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2">Group Item</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Expand or collapse the group item if <code class="prettyprint">drillMode</code> is set to collapsible.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display context menu</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojListView
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan = "20" nowrap>List Item</td>
 *       <td><kbd>F2</kbd></td>
 *       <td>Enters Actionable mode.  This enables keyboard action on elements inside the item, including navigate between focusable elements inside the item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Exits Actionable mode.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>When in Actionable Mode, navigates to next focusable element within the item.  If the last focusable element is reached, shift focus back to the first focusable element.
 *           When not in Actionable Mode, navigates to next focusable element on page (outside ListView).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Tab</kbd></td>
 *       <td>When in Actionable Mode, navigates to previous focusable element within the item.  If the first focusable element is reached, shift focus back to the last focusable element.
 *           When not in Actionable Mode, navigates to previous focusable element on page (outside ListView).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus to the item below.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus to the item above.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>When display in card layout, move focus to the item on the left.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>When display in card layout, move focus to the item on the right.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+DownArrow</kbd></td>
 *       <td>Extend the selection to the item below.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+UpArrow</kbd></td>
 *       <td>Extend the selection to the item above.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+LeftArrow</kbd></td>
 *       <td>When display in card layout, extend the selection to the item on the left.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+RightArrow</kbd></td>
 *       <td>When display in card layout, extend the selection to the item on the right.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+F10</kbd></td>
 *       <td>Launch the context menu if there is one associated with the current item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Selects the current item.  No op if the item is already selected.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Space</kbd></td>
 *       <td>Toggles to select and deselect the current item.  If previous items have been selected, deselects them and selects the current item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Space</kbd></td>
 *       <td>Selects contiguous items from the last selected item to the current item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl+Space</kbd></td>
 *       <td>Toggles to select and deselect the current item while maintaining previous selected items.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl+X</kbd></td>
 *       <td>Marks the selected items to move if dnd.reorder is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl+C</kbd></td>
 *       <td>Marks the selected items to copy if dnd.reorder is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl+V</kbd></td>
 *       <td>Paste the items that are marked to directly before the current item (or as the last item if the current item is a folder).</td>
 *     </tr>
 *     <tr>
 *       <td rowspan = "2" nowrap>Group Item</td>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Collapse the current item if it is expanded and is collapsible.  For non-hierarchical data, do nothing.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Expand the current item if it has children and is expandable.  For non-hierarchical data, do nothing.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojListView
 */

/**
 * {@ojinclude "name":"ojStylingDocIntro"}
 * 
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
 *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-clickthrough-disabled</td>
 *       <td>Use on any element inside an item where you do not want ListView to process the click event.</td>
 *       </td>
 *     </tr>
 *     <tr>
 *       <td>oj-focus-highlight</td>
 *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
 *     </tr>
 *     <tr>
 *       <td>oj-full-width</td>
 *       <td>Use when ListView occupies the entire width of the page.</td>
 *       </td>
 *     </tr>
 *     <tr>
 *       <td>oj-listview-card-layout</td>
 *       <td>Shows items as cards and lay them out in a grid.</td>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
 * @memberof oj.ojListView
 */
});

//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for ListView's disclosure icon in group items.  See the <a href="#getNodeBySubId">getNodeBySubId</a>
 * method for details.</p>
 *
 * @ojdeprecated {since: "2.0.0", description: 'Use the <a href="#oj-listview-disclosure">oj-listview-disclosure</a> option instead.'} 
 * @ojsubid oj-listview-icon
 * @memberof oj.ojListView
 *

 * @example <caption>Get the disclosure icon for the group item with key 'foo':</caption>
 * var node = myListView.getNodeBySubId( {'subId': 'oj-listview-icon', 'key': 'foo'} );
 */

/**
 * <p>Sub-ID for ListView's disclosure icon in group items.  See the <a href="#getNodeBySubId">getNodeBySubId</a>
 * method for details.</p>
 *
 * @ojsubid oj-listview-disclosure
 * @memberof oj.ojListView
 *
 * @example <caption>Get the disclosure icon for the group item with key 'foo':</caption>
 * var node = myListView.getNodeBySubId( {'subId': 'oj-listview-disclosure', 'key': 'foo'} );
 */

/**
 * <p>Context for items within ListView.</p>
 *
 * @property {number} index the zero based item index relative to its parent
 * @property {Object} key the key of the item
 * @property {Element} parent the parent group item.  Only available if item has a parent.
 * @property {boolean} group whether the item is a group.
 *
 * @ojnodecontext oj-listview-item
 * @memberof oj.ojListView
 */

//////////////////     SUB-PROPERTIES (dnd, scrollPolicyOptions)     //////////////////

/**
 * If this object is specified, listview will initiate drag operation when the user drags on either a drag handle, which is an element with oj-listview-drag-handle class, or
 * selected items if no drag handle is set on the item.
 * @expose
 * @name dnd.drag.items
 * @memberof! oj.ojListView
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The MIME types to use for the dragged data in the dataTransfer object. This can be a string if there is only one 
 * type, or an array of strings if multiple types are needed.<br><br>
 * For example, if selected items of employee data are being dragged, dataTypes could be "application/employees+json". Drop targets can examine the data types and decide
 * whether to accept the data. A text input may only accept "text" data type, while a chart for displaying employee data may be configured to accept the "application/employees+json" type.<br><br>
 * For each type in the array, dataTransfer.setData will be called with the specified type and the JSON version of the selected item data as the value. The selected item data 
 * is an array of objects, with each object representing a model object from the underlying data source.  For example, if the underlying data is an oj.Collection, then this
 * would be a oj.Model object.  Note that when static HTML is used, then the value would be the html string of the selected item.<br><br>
 * This property is required unless the application calls setData itself in a dragStart callback function.
 * @expose
 * @name dnd.drag.items.dataTypes
 * @memberof! oj.ojListView
 * @instance
 * @type {string|Array.<string>}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * A callback function that receives the "dragstart" event and context information as arguments.  The ontext information has the following properties:<br>
 * <ul>
 *   <li><code class="prettyprint">items</code>: An array of items being dragged
 *   </li>
 * </ul><br><br>
 * This function can set its own data and drag image as needed. If dataTypes is specified, event.dataTransfer is already populated with the default data when this function is invoked.
 * If dataTypes is not specified, this function must call event.dataTransfer.setData to set the data or else the drag operation will be cancelled.  In either case, the drag image is 
 * set to an image of the dragged rows on the listview.
 * @expose
 * @name dnd.drag.items.dragStart
 * @memberof! oj.ojListView
 * @instance
 * @type {function(Event, {items: Array.<Element>}):void}
 * @default null
 * @ojsignature { target: "Type",
  *                value: "?"}
 */
/**
 * An optional callback function that receives the "drag" event as argument.
 * @expose
 * @name dnd.drag.items.drag
 * @memberof! oj.ojListView
 * @instance
 * @type {function(Event)}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * An optional callback function that receives the "dragend" event as argument.         
 * @expose
 * @name dnd.drag.items.dragEnd
 * @memberof! oj.ojListView
 * @instance
 * @type {function(Event)}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * @typedef {Object} oj.ojListView.ItemsDropContext
 * @property {Element} item the item being dropped on
 * @property {'before'|'after'|'inside'} position the drop position relative to the item being dropped on
 * @property {boolean} reorder true if the drop was a reorder in the same listview, false otherwise
 */
/**
 * An object that specifies callback functions to handle dropping items.
 * @expose
 * @name dnd.drop.items
 * @memberof! oj.ojListView
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * A data type or an array of data types this component can accept.<br><br>
 * This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events.
 * @expose
 * @name dnd.drop.items.dataTypes
 * @memberof! oj.ojListView
 * @instance
 * @type {string | Array.<string>}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"} 
 */
/**
 * An optional callback function that receives the "dragenter" event and context information as arguments.  The context information has the following properties:<br>
 * <ul>
 *   <li><code class="prettyprint">item</code>: the item being entered
 *   </li>
 * </ul><br><br>
 * This function should call <code class="prettyprint">event.preventDefault</code> to indicate the dragged data can be accepted.<br><br>
 * @expose
 * @name dnd.drop.items.dragEnter
 * @memberof! oj.ojListView
 * @instance
 * @type {function(Event, {item: Element}):void}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * An optional callback function that receives the "dragover" event and context information as arguments.  The context information has the following properties:<br>
 * <ul>
 *   <li><code class="prettyprint">item</code>: the item being dragged over
 *   </li>
 * </ul><br><br>
 * Similar to dragEnter, this function should call <code class="prettyprint">event.preventDefault</code> to indicate the dragged data can be accepted.  
 * @expose
 * @name dnd.drop.items.dragOver
 * @memberof! oj.ojListView
 * @instance
 * @type {function(Event, {item: Element}):void}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * An optional callback function that receives the "dragleave" event and context information as arguments.  The context information has the following properties:<br>
 * <ul>
 *   <li><code class="prettyprint">item</code>: the item that was last entered
 *   </li>
 * </ul><br><br>
 * @expose
 * @name dnd.drop.items.dragLeave
 * @memberof! oj.ojListView
 * @instance
 * @type {function(Event, {item: Element}):void}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * A callback function that receives the "drop" event and context information as arguments.  The context information has the following properties:<br>
 * <ul>
 *   <li><code class="prettyprint">item</code>: the item being dropped on
 *   <li><code class="prettyprint">position</code>: the drop position relative to the item being dropped on
 *   <li><code class="prettyprint">reorder</code>: true if the drop was a reorder in the same listview, false otherwise
 *   </li>
 * </ul><br><br>
 * This function should call <code class="prettyprint">event.preventDefault</code> to indicate the dragged data is accepted.<br><br>
 * If the application needs to look at the data for the item being dropped on, it can use the getDataForVisibleItem method.
 * @expose
 * @name dnd.drop.items.drop
 * @memberof! oj.ojListView
 * @instance
 * @type {function(Event, Object):void}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?((param0: Event, param1: oj.ojListView.ItemsDropContext)=> void)",
 *                jsdocOverride: true}
 */
/**
 * The number of items to fetch in each block
 * @expose
 * @name scrollPolicyOptions.fetchSize
 * @memberof! oj.ojListView
 * @instance
 * @type {number}
 * @default 25
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * The maximum total number of items to fetch
 * @expose
 * @name scrollPolicyOptions.maxCount
 * @memberof! oj.ojListView
 * @instance
 * @type {number}
 * @default 500
 * @ojsignature { target: "Type",
 *                value: "?"}
 */
/**
 * The element which listview uses to determine the scroll position as well as the maximum scroll position.  For example in a lot of mobile use cases where ListView occupies the entire screen, developers should set the scroller option to document.documentElement.
 * @expose
 * @name scrollPolicyOptions.scroller
 * @memberof! oj.ojListView
 * @instance
 * @type {Element}
 * @default null
 * @ojsignature { target: "Type",
 *                value: "?"}
 */

/**
 * Handler for IteratingDataProvider generated content
 * @constructor
 * @extends oj.DataSourceContentHandler
 * @ignore
 */
oj.IteratingDataProviderContentHandler = function(widget, root, data)
{
    oj.IteratingDataProviderContentHandler.superclass.constructor.call(this, widget, root, data);
};

// Subclass from oj.DataSourceContentHandler 
oj.Object.createSubclass(oj.IteratingDataProviderContentHandler, oj.DataSourceContentHandler, "oj.IteratingDataProviderContentHandler");

/**
 * Initializes the instance.
 * @protected
 */
oj.IteratingDataProviderContentHandler.prototype.Init = function()
{
  oj.IteratingDataProviderContentHandler.superclass.Init.call(this);
};

oj.IteratingDataProviderContentHandler.prototype.IsHierarchical = function()
{
    return false;
};

/**
 * Destroy the internal DomScroller if there is one.  Called when this ContentHandler is destroyed or on refresh.
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._destroyDomScroller = function()
{
    if (this.m_domScroller != null)
    {
        this.m_domScroller.destroy();

        this.m_domScroller = null;
        this.m_domScrollerMaxCountFunc = null;
    }    
}

/**
 * Destroy the content handler
 * @protected
 */
oj.IteratingDataProviderContentHandler.prototype.Destroy = function()
{
    oj.IteratingDataProviderContentHandler.superclass.Destroy.call(this);
    this._removeDataSourceEventListeners();
    this._destroyDomScroller();

    this.m_loadingIndicator = null;
};

/**
 * @override
 */
oj.IteratingDataProviderContentHandler.prototype.HandleResize = function(width, height)
{
    // we only care about the highwatermark scrolling case, and if height changes
    if (!this._isLoadMoreOnScroll() || this.m_height == undefined || this.m_height == height)
    {
        return;
    }

    this.m_height = height;

    // check viewport
    this.checkViewport();
};

/**
 * @override
 */
oj.IteratingDataProviderContentHandler.prototype.notifyShown = function()
{
    // we only care about the highwatermark scrolling case
    if (!this._isLoadMoreOnScroll())
    {
        return;
    }

    // for loadMoreOnScroll case, we will have to make sure the viewport is satisfied
    this.checkViewport();
};

/**
 * @override
 */
oj.IteratingDataProviderContentHandler.prototype.notifyAttached = function()
{
    var currentFetchTrigger, scroller, fetchTrigger;

    // this should only be populated in highwatermark scrolling case with scroller specified
    currentFetchTrigger = this._getFetchTrigger();
    if (currentFetchTrigger != undefined && this.m_domScroller != null)
    {
        // this should force the fetch trigger to recalculate
        scroller = this._getScroller();
        fetchTrigger = this._getFetchTrigger();
        if (currentFetchTrigger != fetchTrigger)
        {
            // update fetch trigger
            this.m_domScroller.setFetchTrigger(fetchTrigger);
        }

        // check again whether the viewport is satisfied
        this.checkViewport();
    }

};

/**
 * Is loadMoreOnScroll
 * @return {boolean} true or false
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._isLoadMoreOnScroll = function()
{
    return this.m_widget.options['scrollPolicy'] == "loadMoreOnScroll" ? true: false;
};

/**
 * Gets the number of items to return in each fetch
 * @return {number} the fetch size
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._getFetchSize = function()
{
    return Math.max(0, this.m_widget.options['scrollPolicyOptions']['fetchSize']);
};

/**
 * Gets the total size from the DataProvider, if the funcationality is supported
 * @return {number} the total size from the DataProvider, or -1 if DataProvider does not support total size or if the size is unknown.
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._getTotalSize = function()
{
    return -1;
};

/**
 * Gets the scroller element used in DomScroller
 * @return {Element} the scroller element
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._getScroller = function()
{
    var scroller = this.m_widget.options['scrollPolicyOptions']['scroller'];
    if (scroller != null)
    {
       // make sure it's an ancestor
       if ($.contains(scroller, this.m_root))
       {
           // might as well calculate offset here
           if (this._fetchTrigger === undefined)
           {
               this._fetchTrigger = oj.DomScroller.calculateOffsetTop(scroller, this.m_root) + this._getLoadingIndicatorHeight();
           }
           return scroller;
       }
    }

    // if not specified or not an ancestor, use the listview root element
    return this.m_widget.getRootElement();
};

/**
 * Gets the distance from maximum scroll position that triggers a fetch
 * @return {number|undefined} the distance in pixel or undefined if no scroller is specified
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._getFetchTrigger = function()
{
    if (this._fetchTrigger === undefined)
    {
        this._fetchTrigger = this._getLoadingIndicatorHeight();
    }
    return this._fetchTrigger;
};

/**
 * Calculates the height of the loading indicator
 * @return {number} the height of the loading indicator
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._getLoadingIndicatorHeight = function()
{
    var container, icon, height;

    container = $(document.createElement("div"));
    container.addClass(this.m_widget.getItemStyleClass())
             .css({"visibility":"hidden", "overflow":"hidden", "position":"absolute"});
    icon = $(document.createElement("div"));
    icon.addClass("oj-icon oj-listview-loading-icon")
    container.append(icon); // @HTMLUpdateOK

    $(this.m_widget.getRootElement()).append(container); // @HTMLUpdateOK
    height = container.get(0).offsetHeight;
    container.remove();

    return height;
};

/**
 * Gets the maximum number of items that can be retrieved from data source
 * @return {number} the maximum fetch count
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._getMaxCount = function()
{
    return this.m_widget.options['scrollPolicyOptions']['maxCount'];
};

/**
 * @override
 */
oj.IteratingDataProviderContentHandler.prototype.setDataSource = function(dataProvider)
{
    var self;

    this._removeDataSourceEventListeners();
        
    if (dataProvider != null)
    {
        this.m_handleModelMutateEventListener = this.handleModelMutateEvent.bind(this);
        this.m_handleModelRefreshEventListener = this.handleModelRefreshEvent.bind(this);
        
        dataProvider.addEventListener("mutate", this.m_handleModelMutateEventListener);
        dataProvider.addEventListener("refresh", this.m_handleModelRefreshEventListener);
    }

    oj.IteratingDataProviderContentHandler.superclass.setDataSource.call(this, dataProvider);
};

/**
 * Add a loading indicator to the list for high watermark scrolling scenario
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._appendLoadingIndicator = function()
{
    var item, icon;

    // check if it's already added
    if (this.m_loadingIndicator != null)
    {
        return;
    }

    item = $(document.createElement("li"));
    item.uniqueId()
        .attr("role", "presentation")
        .addClass(this.m_widget.getItemStyleClass())
        .addClass("oj-listview-loading-icon-container");

    icon = $(document.createElement("div"));
    icon.addClass("oj-icon oj-listview-loading-icon");
    item.append(icon); // @HtmlUpdateOK

    $(this.m_root).append(item); // @HtmlUpdateOK

    this.m_loadingIndicator = item;
};

/**
 * Remove the loading indicator 
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._removeLoadingIndicator = function()
{
    if (this.m_loadingIndicator != null)
    {
        this.m_loadingIndicator.remove();
    }
    this.m_loadingIndicator = null;
};

/**
 * Whether there are more items to fetch when scroll policy loadMoreOnScroll is used.
 * @return {boolean} true if there are more items to fetch, false otherwise.
 * @protected
 */
oj.IteratingDataProviderContentHandler.prototype.hasMoreToFetch = function()
{
    return (this.m_loadingIndicator != null);
};

/**
 * Add required attributes to item after it is rendered by the renderer
 * @param {Element} item the item element to modify
 * @param {Object} context the item context 
 * @protected
 */
oj.IteratingDataProviderContentHandler.prototype.afterRenderItem = function(item, context)
{
    var size;

    oj.IteratingDataProviderContentHandler.superclass.afterRenderItem.call(this, item, context);

    $(item).addClass(this.m_widget.getItemStyleClass());

    if (this.m_widget._isSelectionEnabled() && this.isSelectable(context))
    {
        this.m_widget.getFocusItem($(item)).attr("aria-selected", false);
    }

    // for highwatermark scrolling, we'll need to add additional wai-aria attribute since not
    // all items are in the DOM
    if (this._isLoadMoreOnScroll())
    {
        size = Math.min(this._getTotalSize(), this._getMaxCount());
        if (size === -1)
        {
            // if count is unknown, then use max count, and re-adjust later as necessary
            size = this._getMaxCount();
        }

        if (size > 0)
        {
            $(item).attr("aria-setsize", size)
                   .attr("aria-posinset", context['index']+1);
        }
    }

    this.m_widget.itemRenderComplete(item, context);
};

/**
 * Add required attributes to item after it is rendered by the renderer, and perform
 * animation for insert
 * @param {Element} item the item element to modify
 * @param {Object} context the item context 
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.afterRenderItemForInsertEvent = function(item, context)
{
    var elem, itemStyleClass, content, self, busyContext, action = "add", promise;

    this.signalTaskStart("after render item from model insert event"); // signal post rendering processing start. Ends at the end of the method.

    item.setAttribute("data-oj-context", "");

    this.afterRenderItem(item, context);

    // hide it before starting animation to show added item
    elem = $(item);

    itemStyleClass = item.className;
    item.className = "oj-listview-temp-item oj-listview-item-add-remove-transition";
    if (!this.shouldUseGridRole())
    {
        elem.children().wrapAll("<div></div>"); //@HTMLUpdateOK
    }

    content = elem.children().first();
    content[0].className = itemStyleClass; 
    // transfer key and role for FindElementByKey lookup that might happen while animating (navlist)
    content[0].key = item.key;

    // transfer aria-selected for selectable checks that might happen while animating (navlist)
    if (!this.shouldUseGridRole())
    {
        content.attr("role", item.getAttribute("role"));
        if (elem[0].hasAttribute("aria-selected"))
        {
            content.attr("aria-selected", item.getAttribute("aria-selected"));
        }
    }

    self = this;
    busyContext = oj.Context.getContext(item).getBusyContext();    
    busyContext.whenReady().then(function()
    {
        if (self.m_widget == null)
        {
            return;
        }

        self.signalTaskStart("kick off animation for insert item"); // signal add animation start. Ends in _handleAddTransitionEnd().

        promise = self.m_widget.StartAnimation(item, action);

        // now show it
        promise.then(function(val)
        {
            item.removeAttribute("data-oj-context");
            self._handleAddTransitionEnd(context, item);
        });

        self.signalTaskEnd(); // signal post rendering processing end. Started at the beginning of the method.
    });
};

/**
 * Callback handler max fetch count.
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._handleScrollerMaxRowCount = function()
{
    // TODO: use resource bundle
    oj.Logger.error("max count reached");
};

/**
 * Remove data source event listeners
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._removeDataSourceEventListeners = function()
{
    var dataProvider = this.getDataSource();
    if (dataProvider != null)
    {
        dataProvider.removeEventListener("mutate", this.m_handleModelMutateEventListener);
        dataProvider.removeEventListener("refresh", this.m_handleModelRefreshEventListener);
    }
};

/**
 * @param {boolean} forceFetch
 * @override
 */
oj.IteratingDataProviderContentHandler.prototype.fetchRows = function(forceFetch)
{
    var options, self, enginePromise, promise, value, templateEngine, dataSource, offset = 0;

    this.signalTaskStart("fetching rows"); // signal method task start

    // checks if we are already fetching cells
    if (this.IsReady())
    {
        this.m_fetching = true;

        oj.IteratingDataProviderContentHandler.superclass.fetchRows.call(this, forceFetch);

        // initiate loading of template engine, note it will not load it unless a template has been specified
        enginePromise = this.loadTemplateEngine();

        // signal fetch started. Ends in fetchEnd() if successful. Otherwise, ends in the reject block of promise below right after _handleFetchError().
        // Cannot end in _handleFetchError() to be consistent with pagingTableDataSource behavior (see comment above)
        this.signalTaskStart("first fetch"); 

        options = {};
        // use fetch size if loadMoreOnScroll, otherwise specify -1 to fetch all rows
        options['size'] = this._isLoadMoreOnScroll() ? this._getFetchSize() : -1;

        self = this;
        this.m_dataProviderAsyncIterator = this.getDataSource().fetchFirst(options)[Symbol.asyncIterator]();
        promise = this.m_dataProviderAsyncIterator.next();
        Promise.all([promise, enginePromise]).then(function(values){
                         // check if content handler has been destroyed already
                         if (self.m_widget == null)
                         {
                             return;
                         }

                         value = values[0];
                         templateEngine = values[1];

                         dataSource = self.getDataSource();
                         if (dataSource instanceof oj.TableDataSourceAdapter)
                         {
                             // paging control loadMore mode, offset will not be 0 after first fetch
                             offset = dataSource['offset'];
                         } 

                         if (offset == 0)
                         {
                             // empty content now that we have data
                             $(self.m_root).empty();
                         }

                         // append loading indicator at the end as needed
                         self._handleFetchedData(value, templateEngine);
                     }, 
                     function(reason){ 
                        self._handleFetchError(reason); 
                        self.signalTaskEnd(); // signal fetch stopped. Started above.
                    });
        this.signalTaskEnd(); // signal method task end
        return;
    }
    this.signalTaskEnd(); // signal method task end
};

oj.IteratingDataProviderContentHandler.prototype._handleFetchError = function(msg)
{
    // TableDataSource aren't giving me any error message
    oj.Logger.error(msg);

    // listview might have been destroyed before fetch error is handled
    if (this.m_widget == null)
    {
        oj.Logger.info("handleFetchError: widget has already been destroyed");
        return;
    }

    if (this._isLoadMoreOnScroll())
    {
        this._removeLoadingIndicator();
    }

    this.m_widget.renderComplete();
};

/**
 * Callback for handling fetch success
 * @param {Array} data the array of data
 * @param {Array} keys the array of keys
 * @param {Object} templateEngine the template engine to process inline template
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._handleFetchSuccess = function(data, keys, templateEngine)
{
    var index, parent, i, row;

    // listview might have been destroyed before fetch success is handled
    if (this.m_widget == null)
    {
        return;
    }

    index = this.m_root.childElementCount;

    parent = document.createDocumentFragment();
    for (i=0; i<data.length; i++)
    {
        row = data[i];
        // passing -1 for opt since we know it will be inserted at the end of the parent
        this.addItem(parent, -1, row, this.getMetadata(index, keys[i], row), templateEngine);

        index = index + 1;
    }    
    this.m_root.appendChild(parent);
};

/**
 * Register the DomScroller
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._registerDomScroller = function()
{
  var self = this, options;
  this.m_domScrollerMaxCountFunc = function (result) 
  {
    if (result != null) 
    {
      self.signalTaskStart("handle results when maxCountLimit reached"); // signal task start

      // remove any loading indicator, which is always added to the end after fetch
      self._removeLoadingIndicator();

      if (self.IsReady()) {
        self.signalTaskStart("dummy task"); // start a dummy task to be paired with the fetchEnd() call below if no new data were fetched.
      }
      self._handleFetchedData(result, self.getTemplateEngine()); // will call fetchEnd(), which signals a task end. Started either in fetchRows() or in a dummy task not involving data fetch.

      self.signalTaskEnd(); // signal domscroller fetch end. Started in this.m_domScroller._handleScrollerScrollTop monkey patched below
      self.signalTaskEnd(); // signal task end
    } else 
    {
      // when there's no more data or any other unexpected cases
      self._removeLoadingIndicator();
      self.signalTaskEnd(); // signal domscroller fetch end. Started in this.m_domScroller._handleScrollerScrollTop monkey patched below
    }
  };

  options = {'fetchSize': this._getFetchSize(), 
             'fetchTrigger': this._getFetchTrigger(), 
             'maxCount': this._getMaxCount(), 
             'asyncIterator': this.m_dataProviderAsyncIterator,
             'initialRowCount': this.m_root.childElementCount, 
             'success': this.m_domScrollerMaxCountFunc, 
             'error': this.signalTaskEnd.bind(this)};
  this.m_domScroller = new oj.DomScroller(this._getScroller(), this.getDataSource(), options);

  // Monkey patch this.m_domScroller's _handleScrollerScrollTop() to signal a task start before starting data fetch
  this.m_domScroller._handleScrollerScrollTop = function (scrollTop, maxScrollTop) 
  {
    if (maxScrollTop - scrollTop <= 1 && self.hasMoreToFetch())
      self.signalTaskStart("starts highwatermark scrolling"); // signal domscroller data fetching. Ends either in success call (m_domScrollerMaxCountFunc) or in error call (self.signalTaskEnd)
    oj.DomScroller.prototype._handleScrollerScrollTop.call(this, scrollTop, maxScrollTop);
  }
}

oj.IteratingDataProviderContentHandler.prototype._pushToEventQueue = function(event)
{
    if (this.m_eventQueue == undefined)
    {
        this.m_eventQueue = [];
    };

    this.m_eventQueue.push(event);
};

oj.IteratingDataProviderContentHandler.prototype._clearEventQueue = function()
{
    if (this.m_eventQueue != null)
    {
        this.m_eventQueue.length = 0;
    }
};

oj.IteratingDataProviderContentHandler.prototype._processEventQueue = function()
{
    var i, event;

    if (this.m_eventQueue != null && this.m_eventQueue.length > 0)
    {
        // see if we can find a refresh event
        for (i=0; i<this.m_eventQueue.length; i++)
        {
            event = this.m_eventQueue[i];
            if (event.type == 'refresh')
            {
                this.handleModelRefreshEvent(event.event);
                // we are done
                return;
            }
        }

        // we'll just need to handle one event at a time since processEventQueue will be triggered whenever an event is done processing
        event = this.m_eventQueue.shift();
        if (event.type == 'mutate')
        {
            this.handleModelMutateEvent(event.event);
        }
    }
};

/**
 * Model mutate event handler.  Called on rows mutation.
 * @param {Object} event the mutate model event
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.handleModelMutateEvent = function(event)
{
    if ( event['detail']['remove'] != null)
    {
        this.handleModelRemoveEvent(event);
    }
    if ( event['detail']['add'] != null)
    {
        this.handleModelAddEvent(event);
    }
    if ( event['detail']['update'] != null)
    {
        this.handleModelChangeEvent(event);
    }
};

/**
 * Retrieve the index of the item with the specified key
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._getIndex = function(keys, index)
{
    var key, elem;

    if (keys == null || keys.length == 0 || index >= keys.length)
    {
        return -1;
    }

    key = keys[index];
    elem = this.FindElementByKey(key);
    return (elem != null) ? $(this.m_root).children().index(elem) : -1;
};

/**
 * Model add event handler.  Called when either a new row of data is added to the data source, or a set of rows are added as a result of
 * highwater mark scrolling.
 * @param {Object} event the add model event
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.handleModelAddEvent = function(event)
{
    var addEvent, data, keys, afterKeys, templateEngine, indexes, i, index;

    if (this.m_root == null)
    {
        return;
    }

    // if listview is busy, queue it for processing later
    if (!this.IsReady())
    {
        this._pushToEventQueue({type: event.type, event: event});
        return;
    }

    this.signalTaskStart("handling model add event"); // signal method task start

    addEvent = event['detail']['add'];
    data = addEvent['data'];
    keys = [];
    addEvent['keys'].forEach(function(key) {
        keys.push(key);
    });
    if (addEvent['afterKeys'])
    {
        afterKeys = [];
        addEvent['afterKeys'].forEach(function(key) {
            afterKeys.push(key);
        });
    }

    // template engine should have already been loaded
    templateEngine = this.getTemplateEngine();

    // indexes could be undefined if not supported by DataProvider
    indexes = addEvent['indexes'];

    if (data != null && keys != null && keys.length > 0 && data.length > 0 && keys.length == data.length && (indexes == null || indexes.length == data.length))
    {
        for (i=0; i<data.length; i++)
        {
            this.signalTaskStart("handling model add event for item: "+keys[i]); // signal add item start
            // indexes takes precedence
            index = (indexes == null) ? this._getIndex(afterKeys, i) + 1 : indexes[i];
            this.addItem(this.m_root, index, data[i], this.getMetadata(index, keys[i], data[i]), templateEngine, this.afterRenderItemForInsertEvent.bind(this));
            this.signalTaskEnd(); // signal add item end
        }

        if (this.IsReady())
        {
            this.signalTaskStart("dummy task"); // start a dummy task to be paired with the fetchEnd() call below if no new data were fetched.
        }
        // do whatever post fetch processing
        this.fetchEnd(); // signals a task end. Started either in fetchRows() or in a dummy task not involving data fetch.
    }
    this.signalTaskEnd(); // signal method task end
};

/**
 * Handles when add item animation transition ends
 * @param {Object} context
 * @param {Element} elem
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._handleAddTransitionEnd = function(context, elem)
{
    var itemStyleClass;

    // this could have been called after listview is destroyed
    if (this.m_widget == null)
    {
        this.signalTaskEnd(); 
        return;
    }

    // cleanup
    itemStyleClass = $(elem).children().first().attr("class");
    elem.className = itemStyleClass;

    if (this.shouldUseGridRole())
    {
        $(elem).children().first()[0].className = "oj-listview-cell-element";
    }
    else
    {
        $(elem).children().children().unwrap(); //@HTMLUpdateOK
    }

    this.m_widget.itemInsertComplete(elem, context);

    this.signalTaskEnd(); // signal add animation end. Started in afterRenderItemForInsertEvent();
};

/**
 * Model remove event handler.  Called when a row has been removed from the underlying data.
 * @param {Object} event the model remove event
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.handleModelRemoveEvent = function(event)
{
    var keys, elem, itemStyleClass, self = this, selection, newSelection, index, i, selectedItems;

    keys = event['detail']['remove']['keys'];
    if (this.m_root == null || keys == null || keys.size == 0)
    {
        return;
    }

    // if listview is busy, hold that off until later
    if (!this.IsReady())
    {
        this._pushToEventQueue({type: event.type, event: event});
        return;
    }

    this.signalTaskStart("handling model remove event"); // signal method task start 

    selection = this.m_widget.options['selection'];
    newSelection = selection.slice(0);

    keys.forEach(function(key)
    {
        elem = self.FindElementByKey(key);
        if (elem != null)
        {
            self.signalTaskStart("handling model remove event for item: "+key); // signal removeItem start
            self._removeItem(elem);
            self.signalTaskEnd(); // signal removeItem end
        }

        // checks whether the removed item is selected, and adjust the value as needed
        index = self.m_widget.GetIndexOf(selection, key);
        if (index > -1)
        {
            newSelection.splice(index, 1);
        }
    });

    // update selection option if it did changed
    if (selection.length != newSelection.length)
    {
        selectedItems = new Array(newSelection.length);
        for (i = 0; i < newSelection.length; i++)
        {
            selectedItems[i] = this.FindElementByKey(newSelection[i]);
        }
        this.m_widget._setSelectionOption(newSelection, null, selectedItems);        
    }

    // since the items are removed, need to clear cache
    this.m_widget.ClearCache();

    this.signalTaskEnd(); // signal method task end
};

/**
 * Remove a single item element
 * @param {jQuery|Element} elem the element to remove
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._removeItem = function(elem)
{
    var active, restoreFocus, itemStyleClass, self = this, action = "remove", item, promise;

    this.signalTaskStart("removing an item"); // signal method task start

    // got to do this before wrapAll since that changes activeElement
    active = document.activeElement;
    restoreFocus = elem.contains(active);

    elem = $(elem);
    itemStyleClass = elem.get(0).className;
    elem.children().wrapAll("<div class='"+itemStyleClass+"'></div>"); // @HtmlUpdateOK
    elem.get(0).className = "oj-listview-item-add-remove-transition";

    this.signalTaskStart("kick off animation to remove an item"); // signal remove item animation start. Ends in _handleRemoveTransitionEnd()

    item = /** @type {Element} */ (elem.get(0));
    promise = this.m_widget.StartAnimation(item, action);

    // now hide it
    promise.then(function(val)
    {
        self._handleRemoveTransitionEnd(elem, restoreFocus);
    });

    this.signalTaskEnd(); // signal method task end
};

/**
 * Handles when remove item animation transition ends
 * @param {Element|jQuery} elem
 * @param {boolean} restoreFocus
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._handleRemoveTransitionEnd = function(elem, restoreFocus)
{
    // this could have been called after listview is destroyed
    if (this.m_widget == null)
    {
        this.signalTaskEnd(); 
        return;
    }

    elem = $(elem);
    var parent = elem.parent();
    // could happen if there is a reset right after model update, the content has already been cleared out
    if (parent.length == 0)
    {
        this.signalTaskEnd(); 
        return;
    }

    // invoke hook before actually removing the item
    this.m_widget.itemRemoveComplete(elem.get(0));

    elem.remove();

    // if it's the last item, show empty text
    if (parent.get(0).childElementCount == 0)
    {
        this.m_widget.renderComplete();
    }

    // ensure something is selected if the removed item is the last selected item
    // need to complete after the DOM element is removed 
    this.m_widget.enforceSelectionRequired();

    // this should focus on the current item, set by itemRemoveComplete
    if (restoreFocus)
    {
        this.m_root.focus();
    }

    this.signalTaskEnd(); // signal remove item animation end. Started in _removeItem()
};

/**
 * Model change event handler.  Called when a row has been changed from the underlying data.
 * @param {Object} event the model change event
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.handleModelChangeEvent = function(event)
{
    var changeEvent, keys, data, templateEngine, indexes, i, elem, index;

    if (this.m_root == null)
    {
        return;
    }

    this.signalTaskStart("handling model update event"); // signal method task start

    changeEvent = event['detail']['update'];
    data = changeEvent['data'];
    keys = [];
    changeEvent['keys'].forEach(function(key) {
        keys.push(key);
    });

    // template engine should have already been loaded
    templateEngine = this.getTemplateEngine();

    // indexes could be undefined if not supported by DataProvider
    indexes = changeEvent['indexes'];
    for (i=0; i<keys.length; i++)
    {
        elem = this.FindElementByKey(keys[i]);
        if (elem != null)
        {
            this.signalTaskStart("handling model update event for item: "+keys[i]); // signal replace item start
            index = (indexes == null) ? -1 : indexes[i];
            this.replaceItem(elem, index, data[i], this.getMetadata(index, keys[i], data[i]), templateEngine, this.afterRenderItemForChangeEvent.bind(this));
            this.signalTaskEnd(); // signal replace item end
        }
    }

    // since the item element will change, need to clear cache
    this.m_widget.ClearCache();

    this.signalTaskEnd(); // signal method task end
};

/**
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.afterRenderItemForChangeEvent = function(item, context)
{
    var self = this, action = "update", promise;

    this.signalTaskStart("after render item for model change event"); // signal method task start

    // adds all neccessary wai aria role and classes
    this.afterRenderItem(item, context);

    promise = this.m_widget.StartAnimation(item, action);

    // now hide it
    promise.then(function(val)
    {
        self._handleReplaceTransitionEnd(item);
    });

    this.signalTaskEnd(); // signal method task end
};

/**
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._handleReplaceTransitionEnd = function(item)
{
    // this could have been called after listview is destroyed
    if (this.m_widget == null)
    {
        this.signalTaskEnd(); 
        return;
    }

    $(item).removeClass("oj-listview-item-add-remove-transition");

    this.m_widget.restoreCurrentItem();

    this.signalTaskEnd(); // signal replace item animation end. Started in replaceItem() from handleModelChangeEvent() (see base class DataSourceContentHandler)
};

/**
 * Model refresh event handler.  Called when all rows has been removed from the underlying data.
 * @param {Object} event the model refresh event
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.handleModelRefreshEvent = function(event)
{
    if (this.m_root == null)
    {
        return;
    }

    // if listview is busy, hold that off until later, the refresh must be handled in order    
    // since we don't know when the results are coming back in
    if (!this.IsReady())
    {
        this._pushToEventQueue({type: event.type, event: event});
        return;
    }

    this.signalTaskStart("handling model reset event"); // signal method task start

    // since we are refetching everything, we should just clear out any outstanding model events
    this._clearEventQueue();

    // empty everything (later) and clear cache
    this.m_widget.ClearCache();

    // it will be recreated with a new asyncIterator
    this._destroyDomScroller();

    // fetch data
    this.fetchRows(true);

    this.signalTaskEnd(); // signal method task end
};

/**
 * Handle fetched data, either from a fetch call or from a sync event
 * @param {Object} dataObj the fetched data object
 * @return {boolean} true if a loading indicator should be appended, false otherwise
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype._handleFetchedData = function(dataObj, templateEngine)
{
    var result = false, data, keys;

    // this could happen if destroy comes before fetch completes (note a refresh also causes destroy)
    if (this.m_root == null || dataObj.value == null)
    {
        return result;
    }

    data = dataObj.value['data'];
    keys = dataObj.value['metadata'].map(function(value) {
      return value['key'];
    });

    if (data.length == keys.length)
    {
        this._handleFetchSuccess(data, keys, templateEngine);

        if (this._isLoadMoreOnScroll())
        {
            result = !dataObj['done'] && this._isLoadMoreOnScroll();
            if (result)
            {
                // if number of items returned is zero but result indicates it's not done
                // log it
                if (keys != null && keys.length == 0)
                {
                    oj.Logger.info("handleFetchedData: zero data returned while done flag is false");
                }

                if (this.m_domScroller == null)
                {
                    this._registerDomScroller();
                }

                // always append the loading indicator at the end except the case when max limit has been reached
                if (!dataObj['maxCountLimit']) 
                {
                    this._appendLoadingIndicator();
                }
            }

            if (dataObj['maxCountLimit']) 
            {
                this._handleScrollerMaxRowCount();
            } 
        }

        this.fetchEnd();
    }

    return result;
};

/**
 * Do any logic needed after results from fetch are processed
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.fetchEnd = function()
{
    var self = this, postProcessing, scrollAndFetch, busyContext;

    // fetch is done
    this.m_fetching = false;

    postProcessing = function()
    {
    	if (self.m_widget)
    	{
	        self.m_widget.renderComplete();

	        // process any outstanding events
	        self._processEventQueue();

	        // check viewport 
	        self.checkViewport();                    		
    	}
    };

    if (this.isAsyncRendering())
    {
        // release busyContext added by scroll and fetch
        scrollAndFetch = this.m_widget.m_scrollAndFetch;
        if (scrollAndFetch)
        {
            self.signalTaskEnd();
        }

        // busyContext for async rendering
        busyContext = oj.Context.getContext(this.m_root).getBusyContext();    
        busyContext.whenReady().then(function()
        {
            // put it back to busy state, will be free up by syncScrollPosition
            if (scrollAndFetch)
            {
                self.signalTaskStart("scroll and fetch");
            }        

            postProcessing();
        });
    }
    else
    {
        postProcessing();
    }

    self.signalTaskEnd(); // signal fetch end. Started in either fetchRows() or started as a dummy task whenever this method is called without fetching rows first (e.g. see m_domScrollerMaxCountFunc).
};

/**
 * Checks the viewport to see if additional fetch is needed
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.checkViewport = function()
{
    var self = this, fetchPromise;

    this.signalTaskStart("checking viewport"); // signal method task start

    // if loadMoreOnScroll then check if we have underflow and do a fetch if we do
    if (this.m_domScroller != null && this.IsReady())
    {
        fetchPromise = this.m_domScroller.checkViewport();
        if (fetchPromise != null)
        {
            this.signalTaskStart("got promise from checking viewport"); // signal fetchPromise started. Ends in promise resolution below
            fetchPromise.then(function(result)
            {
                if (result != null)
                {
                    self.m_domScrollerMaxCountFunc(result);
                }
                self.signalTaskEnd(); // signal checkViewport task end. Started above before fetchPromise resolves here;
            }, null);
        }
    }

    this.signalTaskEnd(); // signal method task end
};

/**
 * Creates the context object containing metadata
 * @param {number} index the index
 * @param {Object} key the key
 * @param {Object} data the data
 * @return {Object} the context object
 * @private
 */
oj.IteratingDataProviderContentHandler.prototype.getMetadata = function(index, key, data)
{
    var context = data['context'];
    if (context == null)
    {
        context = {};
    }

    if (context['index'] == null)
    {
        context['index'] = index;
    }

    if (context['key'] == null)
    {
        context['key'] = key;
    }

    return context;
};
(function() {
var ojListViewMeta = {
  "properties": {
    "as": {
      "type": "string"
    },
    "currentItem": {
      "type": "any",
      "writeback": true
    },
    "data": {},
    "dnd": {
      "type": "object",
      "properties": {
        "drag": {
          "type": "object",
          "properties":{
            "items":{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "drag": {},
                "dragEnd": {},
                "dragStart": {}
              }
            }
          }
        },
        "drop": {
          "type": "object",
          "properties":{
            "items" :{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            }
          }          
        },
        "reorder": {
          "type": "object",
          "properties":{
            "items" :{
              "type":"string",
              "enumValues": ["disabled", "enabled"]
            }
          }
        }
      }
    },
    "drillMode": {
      "type": "string",
      "enumValues": ["collapsible", "none"]
    },
    "expanded": {
      "type": "object",
      "writeback": true
    },
    "firstSelectedItem": {
      "type": "object",
      "writeback": true,
      "readOnly": true
    },
    "groupHeaderPosition": {
      "type": "string",
      "enumValues": ["sticky", "static"]
    },
    "item": {
      "type": "object",
      "properties": {
        "focusable": {},
        "renderer": {},
        "selectable": {}
      }
    },
    "scrollPolicy": {
      "type": "string"
    },
    "scrollPolicyOptions": {
      "type": "object",
      "properties": {
        "fetchSize": {
          "type": "number"
        },
        "maxCount": {
          "type": "number"
        },
        "scroller": {}
      }
    },
    "scrollPosition": {
      "type": "object",
      "writeback": true,
      "properties": {
        "x": {
          "type": "number"
        },
        "y": {
          "type": "number"
        },
        "offsetX": {
          "type": "number"
        },
        "offsetY": {
          "type": "number"
        },
        "key": {
          "type": "any"
        },
        "index": {
          "type": "number"
        },
        "parent": {
          "type": "any"
        }
      }
    },
    "selection": {
      "type": "Array<any>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["none", "single", "multiple"]
    },
    "selectionRequired": {
      "type": "boolean"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "accessibleNavigateSkipItems": {
          "type": "string"
        },
        "accessibleReorderAfterItem": {
          "type": "string"
        },
        "accessibleReorderBeforeItem": {
          "type": "string"
        },
        "accessibleReorderInsideItem": {
          "type": "string"
        },
        "accessibleReorderTouchInstructionText": {
          "type": "string"
        },
        "indexerCharacters": {
          "type": "string"
        },
        "labelCopy": {
          "type": "string"
        },
        "labelCut": {
          "type": "string"
        },
        "labelPaste": {
          "type": "string"
        },
        "labelPasteAfter": {
          "type": "string"
        },
        "labelPasteBefore": {
          "type": "string"
        },
        "msgFetchingData": {
          "type": "string"
        },
        "msgNoData": {
          "type": "string"
        }
      }
    }
  },
  "events": {
    "animateEnd": {},
    "animateStart": {},
    "beforeCollapse": {},
    "beforeCurrentItem": {},
    "beforeExpand": {},
    "collapse": {},
    "copy": {},
    "cut": {},
    "expand": {},
    "paste": {},
    "reorder": {}
  },
  "methods": {
    "getContextByNode": {},
    "getDataForVisibleItem": {},
    "getIndexerModel": {},
    "scrollToItem": {}
  },
  "extension": {
    _INNER_ELEM: 'ul',
    _GLOBAL_TRANSFER_ATTRS: ["aria-label"],
    _WIDGET_NAME: "ojListView"
  }
};
oj.CustomElementBridge.registerMetadata('oj-list-view', 'baseComponent', ojListViewMeta);
oj.CustomElementBridge.register('oj-list-view', {'metadata': oj.CustomElementBridge.getMetadata('oj-list-view')});
})();
/**
 * Handler for Expanded Navigation List
 * @constructor
 * @ignore
 */
oj.DefaultNavListHandler = function (widget, root, component) {
  this.m_widget = widget;
  this.m_root = root;
  this.m_component = component;
  if(this.m_widget.GetOption(this.m_widget.OPTION_EDGE) === this.m_widget.OPTION_EDGE_END){
    this.m_root.addClass(this.m_widget.getNavListEndEdgeStyleClass());
  }
};

/**
 * Restore Listview element
 * @private
 */
oj.DefaultNavListHandler.prototype.Destroy = function () {
  this.m_root.removeClass(this.m_widget.getNavListExpandedStyleClass())
              .removeClass(this.m_widget.getNavListVerticalStyleClass())
              .removeClass(this.m_widget.getNavListEndEdgeStyleClass());
};

/**
 * Expand group Item
 * @param {jQuery} groupItem  Group Item
 * @param {boolean} animate true if animate the expand operation, false otherwise
 * @param {Event} event the event that triggers expand.  Note that event could be null in the case where this is programmatically done by the widget
 * @return {Promise} A Promise that resolves when expand animation completes
 * @private
 */
oj.DefaultNavListHandler.prototype.Expand = function (groupItem, animate, event) {
  return Promise.resolve(null);
};

/**
 * Collapse group Item
 * @param {jQuery} item  Group Item
 * @param {Object} key key of Group item
 * @param {boolean} animate true if animate the collapse operation, false otherwise
 * @param {Event} event the event that triggers collapse.  Note that event could be null in the case where this is programmatically done by the widget
 * @return {Promise} A Promise that resolves when collapse animation completes
 * @private
 */
oj.DefaultNavListHandler.prototype.Collapse = function (item, key, animate, event) {
  return Promise.resolve(null);
};

/**
 * Handle keyboard events for expand and collapse
 * @private
 */
oj.DefaultNavListHandler.prototype.HandleExpandAndCollapseKeys = function (event, keyCode, current, currentItemKey) {
  return false;
};

/**
 * Mofify List item
 * @private
 */
oj.DefaultNavListHandler.prototype.ModifyListItem = function ($item, itemContent) {
  if (this.m_widget.isTabBar()) {
    var focusableElement = this.m_widget.getSingleFocusableElement($item);
    focusableElement.attr('role', 'tab');
  }
};

/**
 * Update aria properties when an item is selected.
 * @private
 */
oj.DefaultNavListHandler.prototype.UpdateAriaPropertiesOnSelectedItem = function (elem, highlight) {
  elem.attr("aria-selected", highlight ? 'true' : 'false');
};

/**
 * Placeholder to override or update generated dom after completing rendering
 * @private
 */
oj.DefaultNavListHandler.prototype.BeforeRenderComplete = function () {
  var role = this.m_widget.element.attr("role");
  if (role && role != "presentation") {
    this.m_widget.element.attr('role', 'presentation');
    if (!this.m_widget.isTabBar()) {
      this.m_root.attr("role", role);
    } else {
      this.m_root.attr("role", 'tablist');
    }
  }
};

/**
 * Called by content handler once the content of an item is rendered triggered by an insert event
 * @param {Element} elem the item element
 * @param {Object} context the context object used for the item
 * @private
 */
oj.DefaultNavListHandler.prototype.ItemInsertComplete = function (elem, context) {

};

/**
 * Called by content handler once the content of an item is removed triggered by an remove event
 * @param {Element} elem the item element
 * @private
 */
oj.DefaultNavListHandler.prototype.ItemRemoveComplete = function (elem) {

};

/**
 * Handles arrow keys navigation on item
 * @param {number} keyCode description
 * @param {boolean} isExtend
 * @param {Event} event the DOM event causing the arrow keys
 * @private
 */
oj.DefaultNavListHandler.prototype.HandleArrowKeys = function (keyCode, isExtend, event) {
  return _ojNavigationListView.superclass.HandleArrowKeys.apply(this.m_widget, arguments);
};

/**
 * Determine whether the key code is an arrow key
 * @param {number} keyCode
 * @return {boolean} true if it's an arrow key, false otherwise
 * @private
 */
oj.DefaultNavListHandler.prototype.IsArrowKey = function (keyCode) {
  return _ojNavigationListView.superclass.IsArrowKey.apply(this.m_widget, arguments);
};

/**
 * Determines whether the specified item is expanded
 * @param {jQuery} item the item element
 * @return {number} 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
 * @private
 */
oj.DefaultNavListHandler.prototype.GetState = function (item) {
  return _ojNavigationListView.superclass.GetState.apply(this.m_widget, arguments);
};

/**
 * Sets the disclosed state of the item
 * @param {jQuery} item the item element
 * @param {number} state 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
 * @private
 */
oj.DefaultNavListHandler.prototype.SetState = function (item, state) {
  _ojNavigationListView.superclass.SetState.apply(this.m_widget, arguments);
};

oj.DefaultNavListHandler.prototype.Init = function (opts) {
  this.m_root.addClass(this.m_widget.getNavListExpandedStyleClass())
              .addClass(this.m_widget.getNavListVerticalStyleClass());
};

/**
 * Check whether the item is selectable
 * @param {Element} item the item element
 * @return {boolean} true if item is selectable
 * @private
 */
oj.DefaultNavListHandler.prototype.IsSelectable = function (item) {
  return _ojNavigationListView.superclass.IsSelectable.apply(this.m_widget, arguments);
};

/**
 * Restore Item
 * @param {jQuery} item, List item
 * @param {jQuery} itemContent, Item contentelement
 * @param {jQuery} sublist, sub list when item is a group item.
 * @private
 */
oj.DefaultNavListHandler.prototype.RestoreItem = function (item, itemContent, sublist) {

};

/**
 * Invoked when an option was updated after component initialization.
 * Note: This will be invoked only for specifc options. see setOptions method in NavigationList.js.
 * @private
 */
oj.DefaultNavListHandler.prototype.SetOptions = function (options) {
  //DO Nothing
};

/**
 * Handler for blur event
 * @param {Event} event the blur event
 * @protected
 */
oj.DefaultNavListHandler.prototype.HandleBlur = function(event)
{
  return _ojNavigationListView.superclass.HandleBlur.apply(this.m_widget, arguments);
};

/**
 * Handler for focus event
 * @param {Event} event the focus event
 * @protected
 */
oj.DefaultNavListHandler.prototype.HandleFocus = function(event)
{
  return _ojNavigationListView.superclass.HandleFocus.apply(this.m_widget, arguments);
};

/**
 * Return node for the given subid
 * @param {Object} locator
 * @protected
 */
oj.DefaultNavListHandler.prototype.GetNodeBySubId = function (locator) {
  return null;
};

/**
 * Return subid for the given node
 * @param {Element} node
 * @protected
 */
oj.DefaultNavListHandler.prototype.GetSubIdByNode = function (node) {
  return null;
};

/**
 * Handle resize event
 * @param {number} width
 * @param {number} height
 * @protected
 */
oj.DefaultNavListHandler.prototype.HandleResize = function (width, height) {

};

/**
 * Invoked when root node is attached
 * @protected
 */
oj.DefaultNavListHandler.prototype.NotifyAttached = function () {

};

/**
 * Return subid for the given node
 * @param {string} action
 * @protected
 */
oj.DefaultNavListHandler.prototype.GetAnimationEffect = function (action) {
  return _ojNavigationListView.superclass.getAnimationEffect.apply(this.m_widget, arguments);
};

/**
 * Return true if option update allowed.
 * @protected
 */
oj.DefaultNavListHandler.prototype.IsOptionUpdateAllowed = function (key, value, flags) {
  return true;
};

/**
 * Triggered when an option is updated
 * @protected
 */
oj.DefaultNavListHandler.prototype.OptionUpdated = function(key, value, flags) {
  
};
/**
 * Invoked before inserting an item in to nav list
 * @protected
 */
oj.DefaultNavListHandler.prototype.BeforeInsertItem = function () {

};

/**
 * Invoked when click event fired
 * @protected
 */
oj.DefaultNavListHandler.prototype.HandleClick = function(event) {
  
};

/**
 * Invoked when click event fired
 * @protected
 */
oj.DefaultNavListHandler.prototype.HandleKeydown = function(event) {
  
};
/**
 * Handler for Horizontal Navigation List
 * @constructor
 * @extends oj.DefaultNavListHandler
 * @ignore
 */

oj.HorizontalNavListHandler = function (widget, root, component) {
  this.m_duringInit = true;
  oj.HorizontalNavListHandler.superclass.constructor.call(this, widget, root, component);
  if(this.m_widget.GetOption(this.m_widget.OPTION_EDGE) === this.m_widget.OPTION_EDGE_BOTTOM){
    this.m_root.addClass(this.m_widget.getNavListBottomEdgeStyleClass());
  }
};

// Subclass from oj.Object
oj.Object.createSubclass(oj.HorizontalNavListHandler, oj.DefaultNavListHandler, "oj.HorizontalNavListHandler");

oj.HorizontalNavListHandler.prototype.Destroy = function () {

  this.m_root.removeClass(this.m_widget.getNavListExpandedStyleClass())
            .removeClass(this.m_widget.getHorizontalNavListStyleClass())
            .removeClass(this.m_widget.getNavListBottomEdgeStyleClass());
  this.m_root.find('.' + this.m_widget.getDividerStyleClass()).remove();
  
  this._destroyOverflowMenu();
  
  if(this.m_overflowMenuItem != null) {
    this.m_overflowMenuItem.remove();
    this.m_overflowMenuItem = null;  
  }
  this.m_overflowMenuItems = [];
  this.m_duringInit = true;
};

oj.HorizontalNavListHandler.prototype.UpdateAriaPropertiesOnSelectedItem = function (elem, highlight) {
  elem.attr(this._isTabBar() ? 'aria-selected' : 'aria-pressed', highlight ? 'true' : 'false');
};

oj.HorizontalNavListHandler.prototype._isTabBar = function () {
  return this.m_widget.isTabBar();
};

/**
 * Handles arrow keys navigation on item
 * @param {number} keyCode description
 * @param {boolean} isExtend
 * @param {Event} event the DOM event causing the arrow keys
 * @private
 */
oj.HorizontalNavListHandler.prototype.HandleArrowKeys = function (keyCode, isExtend, event) {
  //Change keyCode to reverse left/right arrow keys for rtl case.
  if (keyCode === $.ui.keyCode.LEFT) {
    keyCode = this.m_widget.isRtl() ? $.ui.keyCode.DOWN : $.ui.keyCode.UP;
  } else if (keyCode === $.ui.keyCode.RIGHT) {
    keyCode = this.m_widget.isRtl() ? $.ui.keyCode.UP : $.ui.keyCode.DOWN;
  }

  var processed = _ojNavigationListView.superclass.HandleArrowKeys.call(this.m_widget, keyCode, isExtend, event);
    var current = this.m_widget.m_active['elem'];
    current[0].scrollIntoView(false);
  return processed;
};

/**
 *
 * @param {number} keyCode
 * @return {boolean}
 * @private
 */
oj.HorizontalNavListHandler.prototype.IsArrowKey = function (keyCode) {
  return (keyCode === this.m_widget.UP_KEY || 
          keyCode === this.m_widget.DOWN_KEY || 
          keyCode === this.m_widget.LEFT_KEY || 
          keyCode === this.m_widget.RIGHT_KEY);
};

oj.HorizontalNavListHandler.prototype.ModifyListItem = function ($item, itemContent) {
  var focusableElement = this.m_widget.getSingleFocusableElement($item);
  focusableElement.attr('role', this._isTabBar() ? 'tab' : 'button');
  if (!this._isTabBar() && focusableElement[0].hasAttribute('aria-selected')) {
    focusableElement.attr('aria-pressed', 'false');
    focusableElement.removeAttr('aria-selected');
  }
};

oj.HorizontalNavListHandler.prototype.BeforeRenderComplete = function () {
  var ele, visibleItems, self = this;
  
  this._isTabBar() ? this.m_root.attr('role', 'tablist') : this.m_root.attr('role', 'toolbar');
  this.m_widget.element.attr('role','presentation');
  visibleItems = this.m_widget.element.find('.' + this.m_widget.getItemElementStyleClass() + ':visible');
  visibleItems.each(function (index) {
    ele = $(this);
    if (index > 0) {
      self._addSeparator(this, index);
    }
    
    if(index === visibleItems.length - 1) {
      ele.addClass(self.m_widget.getLastItemStyleClass());
    } else {
      ele.removeClass(self.m_widget.getLastItemStyleClass());
    }
  });
  
  if (this.m_duringInit) {
    this.m_duringInit = false;
    this._handleOverflow();
  }
};

oj.HorizontalNavListHandler.prototype._addSeparator = function (elem, index) {
  var $elem = $(elem);
  var previousElement = $elem.prev();
  if (index > 0 && previousElement.length && !previousElement.is("li." + this.m_widget.getDividerStyleClass())) {
    $elem.before('<li role="separator" class="' + this.m_widget.getDividerStyleClass() + '"></li>'); //@HTMLUpdateOK
  }
};

oj.HorizontalNavListHandler.prototype.ItemInsertComplete = function (elem, context) {
  this._addSeparator(elem, context.index);
  this._handleOverflow();
};

oj.HorizontalNavListHandler.prototype.ItemRemoveComplete = function (elem) {
  var $elem = $(elem),
    previousElement = $elem.prev();
  if (previousElement.length && previousElement.is("li." + this.m_widget.getDividerStyleClass())) {
    previousElement.remove();
  } else {
    // Means,this could be the first element,
    //so ensure that divider is not present on next element which is going to be the first one.
    //This is needed because when complete observable array is updated with new set of records, first new records will be added before removing old ones.
    var nextElement = $elem.next();
    if (nextElement.is("li." + this.m_widget.getDividerStyleClass())) {
      nextElement.remove();
    }
  }
  this._handleOverflow();
};

oj.HorizontalNavListHandler.prototype.IsSelectable = function (item) {
  var itemSelectionMarkerAttr = "aria-selected";

  if (!this._isTabBar()) {
    itemSelectionMarkerAttr = "aria-pressed";
  }
  return (!(this.m_overflowMenuItem && this.m_overflowMenuItem[0] === $(item)[0]) && 
            this.m_widget.getFocusItem($(item))[0].hasAttribute(itemSelectionMarkerAttr));
};

oj.HorizontalNavListHandler.prototype.Init = function(opts) {
  this.m_root.addClass(this.m_widget.getNavListExpandedStyleClass())
            .addClass(this.m_widget.getHorizontalNavListStyleClass());
  this.m_overflowMenuItems = [];
};

oj.HorizontalNavListHandler.prototype.HandleClick = function (event) {
  if ($(event.target).closest("." + this.m_widget.getOverflowItemStyleClass() + " a." + this.m_widget.getItemContentStyleClass()).length > 0) {
    this._launchOverflowMenu(event);
  }
};

oj.HorizontalNavListHandler.prototype.HandleKeydown = function (event) {
  if ($(event.target).closest("." + this.m_widget.getOverflowItemStyleClass() + " a." + this.m_widget.getItemContentStyleClass()).length > 0) {
    if (event.keyCode === $.ui.keyCode.SPACE) {
      this._launchOverflowMenu(event);
    }
  }
};

oj.HorizontalNavListHandler.prototype.NotifyAttached = function() {
  this._handleOverflow();
};

oj.HorizontalNavListHandler.prototype.HandleResize = function(width, height) {
  if (this.m_ignoreNextResize) {
    this.m_ignoreNextResize = false;
    return;
  }
  this._handleOverflow();
};

oj.HorizontalNavListHandler.prototype.SetOptions = function(options) {
  var overflow,
          truncation,
          currentOverflow = this.m_widget.GetOption('overflow'),
          currentTruncation = this.m_widget.GetOption('truncation');
  if (options['overflow'] && currentOverflow !== options['overflow']) {
    overflow = options['overflow'];
  }
  if (options['truncation'] && currentTruncation !== options['truncation']) {
    truncation = options['truncation'];
  }
  if (overflow || truncation)
    this._handleOverflow(overflow, truncation);
};

oj.HorizontalNavListHandler.prototype.GetAnimationEffect = function(action) {
  if (action === 'add' ||
    action === 'remove') {
    return _ojNavigationListView.superclass.getAnimationEffect.call(this.m_widget, action + 'HorizontalItem');
  }
  return _ojNavigationListView.superclass.getAnimationEffect.apply(this.m_widget, arguments);
};

oj.HorizontalNavListHandler.prototype.GetNodeBySubId = function(locator) {
  var key,
    item = null,
    subId = locator['subId'],
    overflowMenu,
    menuItems,
    index;

  if (subId === this.m_widget.getItemSubIdKey() && this.m_overflowMenuItems.length > 0) {
    overflowMenu = this._getOverflowMenu();
    key = locator['key'];
    menuItems = overflowMenu.find('.oj-menu-item');

    for (index = 0; index < menuItems.length; index++) {
      if (this.m_widget.compareValues($(menuItems[index]).data('key'),key)) {
        item = menuItems[index];
        this._launchOverflowMenu(null);
        break;
      }
    }
  }
  return item;
};

oj.HorizontalNavListHandler.prototype.GetSubIdByNode = function(node) {
  var subId = null, key, item;
  item = $(node).closest(".oj-menu-item");

  if (this._getOverflowMenu() !== null && item.closest('.' + this.m_widget.getOverflowMenuStyleClass())[0] !== this._getOverflowMenu()[0]) {
    return null;
  }

  if (item != null && item.length > 0) {
    key = item.data('key');
    if (key != null) {
      subId = {
        'subId': this.m_widget.getItemSubIdKey(),
        'key': key
      };
    }
  }
  return subId;
};

oj.HorizontalNavListHandler.prototype.OptionUpdated = function(key, value, flags) {
  if (key === "selection") {
    this._toggleOverflowBtnSelection(value);
  }
};

oj.HorizontalNavListHandler.prototype.IsOptionUpdateAllowed = function(key, value, flags) {
    if (key === 'currentItem') {
      if (this.m_overflowMenuItem && this.m_overflowMenuItem.attr('id') === value) {
        //ignore as this happens only for navigationlist more item and 
        //we should not change currentItem
        return false;
      }
    }
    return true;
};

/**
 * Handle overflow.
 * @param {string=} overflow optional overflow behaviour, if not specified component overflow option will be used.
 * @param {string=} truncation optional truncation behaviour, if not specified component truncation option will be used.
 */
oj.HorizontalNavListHandler.prototype._handleOverflow = function (overflow, truncation) {
  var items,
          overflowHandler,
          threshold = -1;
  items = this._getItems();

  //return if there are no items
  if (items.length === 0) {
    return;
  }

  if (!overflow) {
    overflow = this.m_widget.GetOption('overflow');
  }
  if (!truncation)
    truncation = this.m_widget.GetOption('truncation');

  overflowHandler = new oj._HorizontalNavListOverflowHandler(items, overflow, truncation, this);

  //Return if there is no truncation or overflow handling needed
  if (!overflowHandler.shouldHandleOverflow()) {
    overflowHandler.unApplyTruncation();
    this._releaseContainerWidth();
    return;
  }

  this.showAllItems(items, overflow);
  overflowHandler.unApplyTruncation(); //unpply truncation if anything applied previously
  if (overflowHandler.checkForOverflow() &&
          overflowHandler.applyTruncation()) {
    threshold = overflowHandler.getOverflowThreshold();//caclulate threhsold only when it still overflow after truncation.
  }
  this._applyThreshold(threshold);
  this._releaseContainerWidth();
};

oj.HorizontalNavListHandler.prototype.BeforeInsertItem = function () {
  this._fixContainerWidth();
};

oj.HorizontalNavListHandler.prototype._fixContainerWidth = function () {
  var width, root;
  root = this.m_widget.getRootElement();
  width = root[0].getBoundingClientRect()['width'];
  if (width !== 0) //it will be 0 when there are no elements in the list.
  {
    this.m_freezedCotnainerWidth = width;
    root.css('maxWidth', width);
  }
};

oj.HorizontalNavListHandler.prototype._releaseContainerWidth = function () {
  var root, width;
    root = this.m_widget.getRootElement();
    root.css('maxWidth', 'none');
    width = root[0].getBoundingClientRect()['width'];
    if (this.m_freezedCotnainerWidth) {
      //Already applied truncation logic but still overflows 
      //then no need to run overflow logic again, in case of conveyor belt this helps to show left/right buttons
      this.m_ignoreNextResize = true;
      this.m_freezedCotnainerWidth = null;
    }
};

oj.HorizontalNavListHandler.prototype.showAllItems = function (items, overflow) {
  var self = this;
  //show all the items including overflow button
  if (overflow === "popup") {
    this._showOrHideItem(this._getOverflowMenuButton(), true);
  }

  items.each(function (index, item) {
    self._showOrHideItem($(item), true);
  });
};

oj.HorizontalNavListHandler.prototype._getItems = function() {
  return this.m_root.find("." + this.m_widget.getItemElementStyleClass() + ":not(." + this.m_widget.getOverflowItemStyleClass() + ")");
};

oj.HorizontalNavListHandler.prototype._hasSeparators = function() {
  return this.m_root.hasClass(this.m_widget.getNavListItemsDividerStyleClass());
};

oj.HorizontalNavListHandler.prototype._showOrHideItem = function(item, show) {
  var separator;
  if (this._hasSeparators()) {
    separator = item.prev('.' + this.m_widget.getDividerStyleClass());
  }
  if (show) {
    item.show();
    if (separator) {
      separator.show(); //show separator when hiding overflow item
    }
  } else {
    item.hide();
    if (separator) {
      separator.hide(); //hide separator when hiding item
    }

  }
};

oj.HorizontalNavListHandler.prototype._applyThreshold = function(threshold) {
  var self = this,
    items = this._getItems(),
    overflowMenuData,
    overflowMenuDataArray = [],
    lastVisibleChild,
    isOverflowItemVisible,
    isOverflowMenuVisible;

  if (items.length === 0) {
    return;
  }

  this.m_root.find('.' + this.m_widget.getLastItemStyleClass())
    .removeClass(this.m_widget.getLastItemStyleClass());

  if (threshold === -1 || threshold >= items.length) {
    if(this.m_overflowMenuItem) { //if overflow item is already added need to hide it
      this._showOrHideItem(this.m_overflowMenuItem, false);
    }
  } else {
    // As listview always inserts item using index means it will insert new item before index+1 item
    // some time it is possible that new item is inserted between overflow item  and it's separator, 
    // so ensuring that separator is added if it's not already there.
    this._addSeparator(this._getOverflowMenuButton(), items.length);  

    this._showOrHideItem(this._getOverflowMenuButton(), true);
    isOverflowItemVisible = true;
  }

  items.each(function(index, item) {
    var $item = $(item);
    if (threshold !== -1 && threshold <= index) {
      if ($item.hasClass('oj-focus')) {
        self.m_widget.ActiveAndFocus(self._getOverflowMenuButton(), null);
      }

      self._showOrHideItem($item, false);
      overflowMenuData = {};
      overflowMenuDataArray.push(overflowMenuData);
      overflowMenuData["key"] = self.m_widget.GetKey(item);
      overflowMenuData["label"] = self._getItemLabel($item);
      if ($item.hasClass("oj-disabled")) {
        overflowMenuData["disabled"] = true;
      }
    } else {
      lastVisibleChild = $item;
      self._showOrHideItem($item, true);
      if (self.m_overflowMenuItem && self.m_overflowMenuItem.hasClass('oj-focus') &&
        self.m_widget.compareValues(self.m_widget.GetKey(item), self.m_widget.GetOption('currentItem'))) {
        self.m_widget.ActiveAndFocus(lastVisibleChild, null);
      }
    }
  });

  if (isOverflowItemVisible) {
    this._getOverflowMenuButton().addClass(this.m_widget.getLastItemStyleClass());
  } else {
    lastVisibleChild.addClass(this.m_widget.getLastItemStyleClass());
  }



  this.m_overflowMenuItems = overflowMenuDataArray;
  this.m_widget.ClearCache();

  isOverflowMenuVisible = this.m_overflowMenu && this.m_overflowMenu.is(':visible');
  this._destroyOverflowMenu();
  
  //launching sheetmenu some times make scrollbars visible which is triggering resize event.
  // So relaunching menu if it is already launched. 
  if (isOverflowMenuVisible) {
    this._launchOverflowMenu(null);
  }

};

oj.HorizontalNavListHandler.prototype._destroyOverflowMenu = function() {
  if (this.m_overflowMenu) {
    this.m_overflowMenu.ojMenu('destroy');
    this.m_overflowMenu.remove();
    this.m_overflowMenu = null;
  }
};

oj.HorizontalNavListHandler.prototype._getOverflowMenuButton = function() {
  var anchorElement,
    labelElement,
    iconElement,
    overflowMenuItem,
    items,
    lastItem;
  if (!this.m_overflowMenuItem) {
    overflowMenuItem = $(document.createElement('li'));
    anchorElement = $(document.createElement('a'));
    labelElement = $(document.createElement('span'));
    iconElement = $(document.createElement('span'));
    items = this._getItems();
    overflowMenuItem.uniqueId()
      .attr('role', 'presentation')
      .addClass(this.m_widget.getItemElementStyleClass())
      .addClass(this.m_widget.getItemStyleClass())
      .addClass(this.m_widget.getOverflowItemStyleClass())
      .addClass('oj-default')
      .append(anchorElement);// @HTMLUpdateOK constructed by component and not using string passed through any API
    anchorElement.addClass(this.m_widget.getFocusedElementStyleClass())
      .addClass(this.m_widget.getItemContentStyleClass());
    anchorElement.attr('role', 'button')
      .attr('aria-haspopup', 'true')
      .attr('aria-pressed', 'false')
      .attr('tabindex', '-1')
      .attr('data-tabmod', '-1')
      .attr('href', '#')
      .append(iconElement)// @HTMLUpdateOK constructed by component and not using string passed through any API
      .append(labelElement);// @HTMLUpdateOK label text is read from resource bundle and it was properly escaped

    if (!this.m_root.find('ul:first').hasClass(this.m_widget.getHasIconsStyleClass())) {
      anchorElement.addClass(this.m_widget.getHasNoIconStyleClass());
    }

    labelElement.text(this.m_widget.ojContext.getTranslatedString('overflowItemLabel')) // @HTMLUpdateOK
      .addClass(this.m_widget.getItemLabelStyleClass());

    iconElement.addClass(this.m_widget.getItemIconStyleClass())
      .addClass("oj-fwk-icon")
      .addClass(this.m_widget.getOverflowItemIconStyleClass());

    overflowMenuItem[0].key = overflowMenuItem.attr('id');
    this.m_root.find('.'+this.m_widget.GetStyleClass()).append(overflowMenuItem);// @HTMLUpdateOK constructed by component and not using string passed through any API
    this._addSeparator(overflowMenuItem, items.length);
    this.m_overflowMenuItem = overflowMenuItem;
  }
  return this.m_overflowMenuItem;
};

oj.HorizontalNavListHandler.prototype._getOverflowMenu = function() {
  var self = this,
    data = this.m_overflowMenuItems,
    menuListItem,
    menuItem,
    i,
    overflowMenu;

  if (data.length === 0) {
    return null;
  }

  if (!this.m_overflowMenu) {
    overflowMenu = $(document.createElement('ul'));
    overflowMenu.addClass(this.m_widget.getOverflowMenuStyleClass()).hide()
    this.m_root.append(overflowMenu);// @HTMLUpdateOK constructed by component and not using string passed through any API

    // create menu markup
    for (i = 0; i < data.length; i++) {
      menuListItem = $(document.createElement('li'));
      menuItem = $(document.createElement('a'));

      menuItem.attr('href', '#')
        .text(data[i]["label"]);

      menuListItem.data('key',
        data[i]["key"]);

      if (data[i]["disabled"]) {
        menuListItem.addClass('oj-disabled');
      }

      menuListItem.append(menuItem);// @HTMLUpdateOK menuItem's label text is read from dataSource and it was properly escaped 
      overflowMenu.append(menuListItem);// @HTMLUpdateOK constructed by component and not using string passed through any API
    }

    //initialize ojmenu
    overflowMenu.ojMenu({
      openOptions: {
        display: 'auto'
      },
      open: self.__handleOverflowMenuOpen.bind(self),
      close: self.__handleOverflowMenuClose.bind(self),
      select: self.__handleOverflowMenuSelection.bind(self)
    });
    this.m_overflowMenu = overflowMenu;
  }

  return this.m_overflowMenu;
};

oj.HorizontalNavListHandler.prototype._launchOverflowMenu = function(event) {
  var self = this;
  
  if(this._getOverflowMenu().is(":visible")) {
    return; //ignore if overflow menu already launched
  }
  this._getOverflowMenuButton().addClass('oj-selected').removeClass('oj-default');
  this._getOverflowMenu().ojMenu('open', event, {
    launcher: self._getOverflowMenuButton(),
    initialFocus: "firstItem",
    position: {
      "my": "end bottom",
      "at": "end top",
      "collision": "flipfit"
    }
  });
};

oj.HorizontalNavListHandler.prototype._toggleOverflowBtnSelection = function(selectedItem) {
  var overflowItemSelected, i;
  if (this.m_overflowMenuItem && this.m_overflowMenuItems.length > 0) {
    if(selectedItem) {
      for (i = 0; i < this.m_overflowMenuItems.length; i++) {
        if (this.m_widget.compareValues(this.m_overflowMenuItems[i]["key"], selectedItem)) {
          overflowItemSelected = true;
          break;
        }
      }
    }

    if (!overflowItemSelected) {
      this._getOverflowMenuButton().addClass('oj-default').removeClass('oj-selected');
    }
  }
};

oj.HorizontalNavListHandler.prototype.__handleOverflowMenuOpen = function(event, ui) {
  this._getOverflowMenuButton().find('.' + this.m_widget.getItemContentStyleClass()).attr('tabindex', '0');
};

oj.HorizontalNavListHandler.prototype.__handleOverflowMenuClose = function(event, ui) {
  var selectedItem = this.m_widget.ojContext.option('selection');
  this._toggleOverflowBtnSelection(selectedItem);
  this._getOverflowMenuButton().find('.' + this.m_widget.getItemContentStyleClass()).focus();
};

oj.HorizontalNavListHandler.prototype.__handleOverflowMenuSelection = function(event, ui) {
  var key = ui.item.data('key'),
    selectedItemKey,
    options = {
      "selection": key
    },
    item = this.m_widget.FindElementByKey(key);

  this.m_widget.HandleSelectionOption(options);
  selectedItemKey = options["selection"];
  if (selectedItemKey) {
    // trigger the optionChange event
    this.m_widget.SetOption("selection", selectedItemKey, {
      '_context': {
        originalEvent: event,
        internalSet: true,
        extraData: {
          'items': $(item)
        }
      },
      'changed': true
    });
  }
};


oj.HorizontalNavListHandler.prototype._getItemLabel = function(item) {
  var titleElement = item.find('.' + this.m_widget.getItemTitleStyleClass() + ':first');
  if (titleElement.length > 0) {
    return titleElement.text();
  } else {
    return item.find('.' + this.m_widget.getItemLabelStyleClass() + ':first').text();
  }
};

/**
 * Overflow handler for horizontal navigation list
 * @constructor
 * @ignore
 * @private
 */
oj._HorizontalNavListOverflowHandler = function (items, overflow, truncation, navlistHandler) {
  this._overflow = overflow;
  this._truncation = truncation;
  this._items = items;
  this._navlistHandler = navlistHandler;
};

oj._HorizontalNavListOverflowHandler.prototype._getOverflowMenuButton = function () {
  return this._navlistHandler._getOverflowMenuButton();
};

oj._HorizontalNavListOverflowHandler.prototype._getWidget = function () {
  return this._navlistHandler.m_widget;
};

oj._HorizontalNavListOverflowHandler.prototype.shouldHandleOverflow = function () {
  return this._overflow === "popup" || this._truncation === "progressive";
};

oj._HorizontalNavListOverflowHandler.prototype.unApplyTruncation = function () {
  var self = this;
  this._items.each(function (index, item) {
    $(item).find('.' + self._getWidget().getItemLabelStyleClass()).css("max-width", "");
  });
};

oj._HorizontalNavListOverflowHandler.prototype.checkForOverflow = function () {
  if (!this.shouldHandleOverflow())
    return false;

  //Get overflow data together for performance 
  if (!this._overflowData) {
    this._overflowData = this._collectOverflowData();
  }
  return this._isOverflowed(this._overflowData.containerEdgePos, this._overflowData.itemEdgePos, 0);
};

oj._HorizontalNavListOverflowHandler.prototype.applyTruncation = function () {

  //apply truncation
  if (this._truncation === "progressive") {
    this._applyLabelMaxWidth(this._overflowData);
  }
  var item = this._items.last();
  var itemEdge = item[0].getBoundingClientRect()[oj.DomUtils.getReadingDirection() === 'ltr' ? 'right' : 'left'];
  return this._isOverflowed(this._overflowData.containerEdgePos, itemEdge, 0);
};

oj._HorizontalNavListOverflowHandler.prototype.getOverflowThreshold = function () {
  var threshold = -1;
  if (this._overflow === "popup") {
    threshold = this._calculateThreshold(this._overflowData);
  }
  return threshold;
};

oj._HorizontalNavListOverflowHandler.prototype._collectOverflowData = function () {
  var edge,
          item,
          itemEdge,
          itemNonTextWidth,
          overflowItemWidth,
          containerWidth,
          edgePosition,
          container;
  edge = oj.DomUtils.getReadingDirection() === 'ltr' ? 'right' : 'left';
  container = this._getWidget().ojContext.element;

  item = this._items.last();
  if (this._truncation === 'progressive') {
    itemNonTextWidth = this._calculateItemNonTextWidth();
  }
  
  edgePosition = container[0].getBoundingClientRect()[edge];
  containerWidth = container[0].getBoundingClientRect()['width'];
  
  if (this._overflow === "popup") {
    overflowItemWidth = this._getOverflowMenuButton()[0].getBoundingClientRect()['width'];
  }
  itemEdge = item[0].getBoundingClientRect()[edge];

  return {
    containerEdgePos: edgePosition,
    containerWidth: containerWidth,
    overflowItemWidth: overflowItemWidth,
    itemEdgePos: itemEdge,
    itemNonTextWidth: itemNonTextWidth
  };
};

oj._HorizontalNavListOverflowHandler.prototype._calculateItemNonTextWidth = function () {
  var itemsWithRemovable,
          textWidth,
          pickItem = null,
          items = this._items,
          self = this;
  if (this._getWidget().getRootElement().hasClass('.' + this._getWidget().getHasIconsStyleClass())) {
    items = items.filter(function (index, item) {
      return $(item).find('.' + self._getWidget().getItemIconStyleClass()).length > 0;
    });
  }
  itemsWithRemovable = items.filter(function (index, item) {
    return $(item).find('.' + self._getWidget().getRemovableStyleClass()).length > 0;
  });
  pickItem = itemsWithRemovable && itemsWithRemovable.length > 0 ? itemsWithRemovable.last() : items.last();
  textWidth = pickItem.find('.' + this._getWidget().getItemLabelStyleClass())[0].getBoundingClientRect()['width'];
  return pickItem[0].getBoundingClientRect()['width'] - textWidth;
};

oj._HorizontalNavListOverflowHandler.prototype._isOverflowed = function (containerEdge, itemEdge, overflowItemWidth) {
  var isLtr = oj.DomUtils.getReadingDirection() === 'ltr';
  overflowItemWidth = overflowItemWidth ? overflowItemWidth : 0;
  if (isLtr) {
    //Even after using edge position, some times there is a fraction of pixel
    //difference which is being considered as overflow, so having 0.1 as buffer.
    return (itemEdge - containerEdge + overflowItemWidth) > 0.1;
  } else {
    return (containerEdge - itemEdge + overflowItemWidth) > 0.1;
  }
};

oj._HorizontalNavListOverflowHandler.prototype._calculateThreshold = function (overflowData) {
  var edge, items, itemEdge, index, threshold;
  items = this._items;
  edge = oj.DomUtils.getReadingDirection() === 'ltr' ? 'right' : 'left';
  index = items.length - 1;
  itemEdge = items[index].getBoundingClientRect()[edge];
  //Need to check for overflow including overflow item to avoid overflow item truncation.
  while (this._isOverflowed(overflowData.containerEdgePos, itemEdge, overflowData.overflowItemWidth) && index > 0) {
    index = index - 1;
    itemEdge = items[index].getBoundingClientRect()[edge];
  }

  threshold = index + 1;
  return threshold;
};

oj._HorizontalNavListOverflowHandler.prototype._applyLabelMaxWidth = function (overflowData)
{
  var items = this._items,
          self = this,
          minLabelWidth = this._getMinLabelWidth(this._items);
  var labelMaxWidth = (overflowData.containerWidth / items.length) - overflowData.itemNonTextWidth;
  
  if (labelMaxWidth < minLabelWidth) {
    labelMaxWidth = minLabelWidth; // use min width when it reaches threshold minimum
//      return; //Restore to original width when maxwidth reaches threshold minimum.
  }
  
  items.each(function (index, item) {
    $(item).find('.' + self._getWidget().getItemLabelStyleClass()).css({'max-width': labelMaxWidth + 'px'});
  });
};

oj._HorizontalNavListOverflowHandler.prototype._getMinLabelWidth = function (items) {
  var computedMinWidth = window.getComputedStyle(items.first().find('.' + this._getWidget().getItemLabelStyleClass())[0], null).getPropertyValue("min-width");
  if (computedMinWidth.indexOf('px') > 0) {
    return parseInt(computedMinWidth.substring(0, computedMinWidth.length - 2), 10);
  }
  //This should never happen unless it was overriden intentionally
  return 0;
};

/**
 * Handler for Collapsible Navigation List
 * @constructor
 * @extends oj.DefaultNavListHandler
 * @ignore
 */
oj.CollapsibleNavListHandler = function (widget, root, component) {
  oj.CollapsibleNavListHandler.superclass.constructor.call(this, widget, root, component);
};
// Subclass from oj.Object 
oj.Object.createSubclass(oj.CollapsibleNavListHandler, oj.DefaultNavListHandler, "oj.CollapsibleNavListHandler");

oj.CollapsibleNavListHandler.prototype.Destroy = function () {
  this.m_root.removeClass(this.m_widget.getNavListCollapsibleStyleClass())
              .removeClass(this.m_widget.getNavListVerticalStyleClass());
};

oj.CollapsibleNavListHandler.prototype.Init = function (opts) {
  this.m_root.addClass(this.m_widget.getNavListCollapsibleStyleClass())
              .addClass(this.m_widget.getNavListVerticalStyleClass());
};

oj.CollapsibleNavListHandler.prototype.Expand = function (groupItem, animate, event) {
  return _ojNavigationListView.superclass.AnimateExpand.apply(this.m_widget, arguments);
};

oj.CollapsibleNavListHandler.prototype.Collapse = function (item, key, animate, event) {
  return _ojNavigationListView.superclass.AnimateCollapse.apply(this.m_widget, arguments);
};

oj.CollapsibleNavListHandler.prototype.HandleExpandAndCollapseKeys = function (event, keyCode, current, currentItemKey) {
  var isGroupItem = current.children('.' + this.m_widget.getGroupStyleClass()).length > 0;
  if (keyCode === this.m_widget.LEFT_KEY || keyCode === this.m_widget.RIGHT_KEY) {
    //Need to reverse left/right functionality  in rtl case
    if ((keyCode === this.m_widget.LEFT_KEY && !this.m_widget.isRtl()) || (keyCode === this.m_widget.RIGHT_KEY && this.m_widget.isRtl())) {
      if (this.m_widget.GetState(current) === this.m_widget.STATE_EXPANDED) {
        this.m_widget.CollapseItem(current, event, true, currentItemKey, true, true);
      }
    } else {
      if (this.m_widget.GetState(current) === this.m_widget.STATE_COLLAPSED) {
        this.m_widget.ExpandItem(current, event, true, currentItemKey, true, true, true);
      }
    }
    return true;
  }

  if (isGroupItem && (keyCode === $.ui.keyCode.ENTER || keyCode === $.ui.keyCode.SPACE)) {
    if (current.length <= 0) {
      return;
    }
    if (this.m_widget.GetState(current) === this.m_widget.STATE_COLLAPSED) {
      this.m_widget.ExpandItem(current, null, true, currentItemKey, true, true, true);
    } else if (this.m_widget.GetState(current) === this.m_widget.STATE_EXPANDED) {
      this.m_widget.CollapseItem(current, null, true, currentItemKey, true, true);
    }
    return true;
  }
  return false;
};

/**
 * Handler for Sliding Navigation List
 * @constructor
 * @extends oj.CollapsibleNavListHandler
 * @ignore
 */
oj.SlidingNavListHandler = function (widget, root, component) {
  oj.SlidingNavListHandler.superclass.constructor.call(this, widget, root, component);
  this.m_expanded = [];
};

// Subclass from oj.Object
oj.Object.createSubclass(oj.SlidingNavListHandler, oj.CollapsibleNavListHandler, "oj.SlidingNavListHandler");

oj.SlidingNavListHandler.prototype.Destroy = function () {
  this.m_root.removeClass('oj-navigationlist-slider')
              .removeClass(this.m_widget.getNavListVerticalStyleClass());
  this._toolbar.remove();
};
/**
 * Intiate Expand/collapse animation.
 *
 * @param {Object} item, item which is going to be expanded.
 * @param {boolean} isMovingNext true for expand animation and false for collapse animation.
 * @param {Object} focusableElement, focusable element after animation.
 * @param {Event} event the event that triggers sliding.  Note that event could be null in the case where this is programmatically done by the widget
 * @param {Function} animationResolve the resolve function that resolves the animation Promise
 */
oj.SlidingNavListHandler.prototype._slideAnimation = function (item, isMovingNext, focusableElement, event, animationResolve) {
  var self = this;
  var list_root = this.m_widget.getRootElement();
  var isRtl = this.m_widget.isRtl();
  var hasFocusAncestor = this.m_widget.getRootElement().hasClass("oj-focus-ancestor");
  var action, promise;

  action = isMovingNext ? "sliderExpand" : "sliderCollapse";
  promise = this.m_widget.StartAnimation(/** @type {Element} */ (list_root.get(0)), action, this.m_widget.getAnimationEffect(action));
  promise.then(function () {
    self._slideAnimationComplete(item, isMovingNext, focusableElement, event, hasFocusAncestor);
    animationResolve(null);
  });
};

/**
 * Update ui after animation.
 *
 * @param {Object} item, item which is going to be expanded.
 * @param {boolean} isMovingNext true for expand animation and false for collapse animation
 * @param {Object|null} focusableElement, focusable element after animation.
 * @param {Event} event the event that triggers sliding.  Note that event could be null in the case where this is programmatically done by the widget
 * @param {boolean} needFocusAncestor, true if oj-focus-ancestor needs to be added after slide animation.
 * @private
 */
oj.SlidingNavListHandler.prototype._slideAnimationComplete = function (item, isMovingNext, focusableElement, event, needFocusAncestor) {
  if (this.m_widget.m_contentHandler === null) {
    return;
  }

  if (focusableElement) {
    if (needFocusAncestor) {
      this.m_widget.getRootElement().addClass("oj-focus-ancestor"); // during animation oj-focus-ancestor is being removed due triggering of focusout event. probably this is due to hiding of element user clicked on.
    }
    if (event && event.button === 0) {
      this.m_widget.AvoidFocusHighLight(true);
    }
    this.m_widget.SetCurrentItem(focusableElement, event);
    this.m_widget.AvoidFocusHighLight(false);
  }
  if (!isMovingNext) {
    this.m_widget.AnimateCollapseComplete(item.children('.' + this.m_widget.getGroupStyleClass()));
  } else {
    this.m_widget.AnimateExpandComplete(item.children('.' + this.m_widget.getGroupStyleClass()));
  }
};

/**
 * Expand item in sliding navigation list
 * @param {Object} groupItem group item to be expanded
 * @param {boolean} animate flag to trigger animation or not
 * @param {Event} event the event that triggers expand.  Note that event could be null in the case where this is programmatically done by the widget
 * @return {Promise} A Promise that resolves when expand animation completes
 * @private
 */
oj.SlidingNavListHandler.prototype.Expand = function (groupItem, animate, event) {
  var animationResolve;
  var animationPromise = new Promise(function (resolve, reject) {
    animationResolve = resolve;
  });

  var target = $(groupItem).parents('.oj-navigationlist-item-element:first');
  var sublist = target.children('.' + this.m_widget.getGroupStyleClass()),
    parentLabel,
    nextFocusableItem = null;

  var currentListRoot = target.closest('.' + this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS);
  if (currentListRoot.length > 0) {
    //Move marker style class to expanded node only when it is child of an already expanded node.
    //It is possible that child is expanded before parent, in which we need should not remove marker class from child.
    //Ref , During initialization, listview expands child before parent which is causing the issue.
    currentListRoot.removeClass(this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS);
    target.addClass(this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS);
    nextFocusableItem = sublist.find('.' + this.m_widget.getItemElementStyleClass() + ':eq(0)');
    var listOfParents = target.parentsUntil(currentListRoot, 'ul');
    if (currentListRoot.is(this.m_widget.element)) {
      listOfParents = listOfParents.get().concat(currentListRoot.get());
    } else {
      listOfParents = listOfParents.get();
    }

    var self = this,
      itemNode;
    listOfParents = $(listOfParents.reverse());
    listOfParents.each(function (i, parentList) {
      if ($(parentList).is(self.m_widget.element)) {
        parentLabel = self.m_widget.getRootLabel();
      } else {
        parentLabel = self.m_widget.getItemLabel($(parentList).parent());
      }
      if (i === listOfParents.length - 1) {
        itemNode = target;
      } else {
        itemNode = $(listOfParents.get(i + 1)).parent();
      }
      self._addItemToHviewMenu(self.m_widget.GetKey(itemNode[0]), self.m_widget.getItemLabel(itemNode), parentLabel);
      // this var keeps a stack of expanded items
      self.m_expanded.push(itemNode);
    });
  }

  if (animate) {
    this._slideAnimation(target, true, nextFocusableItem, event, animationResolve);
  } else {
    this._slideAnimationComplete(target, true, nextFocusableItem, event, false);
    animationResolve(null);
  }

  //For ios, it is needed to set aria-hidden.
  target.siblings().attr('aria-hidden', 'true');
  target.children('.' + this.m_widget.getGroupItemStyleClass())
    .children('.' + this.m_widget.getItemContentStyleClass())
    .attr('aria-hidden', 'true');
  sublist.removeAttr('aria-hidden');

  // undo any display set by ListView
  groupItem.css("display", "");
  //Skip focus for group element
  target.addClass('oj-skipfocus');

  return animationPromise;
};

/**
 * Collapse item in sliding navigation list
 * @param {Object} target target list root
 * @param {Object} key the key of the group item
 * @param {boolean} animate flag to trigger animation or not
 * @param {Event} event the event that triggers collapse.  Note that event could be null in the case where this is programmatically done by the widget
 * @return {Promise} A Promise that resolves when collapse animation completes
 * @private
 */
oj.SlidingNavListHandler.prototype.Collapse = function (target, key, animate, event) {
  var animationResolve;
  var animationPromise = new Promise(function (resolve, reject) {
    animationResolve = resolve;
  });

  var currentList = target.children('.' + this.m_widget.getGroupStyleClass()),
    parentlist = target.parent();

  target.children('.' + this.m_widget.getGroupItemStyleClass())
    .children('.' + this.m_widget.getItemContentStyleClass())
    .removeAttr('aria-hidden');
  currentList.attr('aria-hidden', 'true');
  target.siblings().removeAttr('aria-hidden');

  //Enable focus for group element
  target.removeClass('oj-skipfocus');

  target.removeClass(this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS);
  //While collapsing node, if any of parent is already having current style class then don't add again
  if (target.closest('.' + this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS).length === 0) {
    parentlist.closest('.' + this.m_widget.getItemElementStyleClass()).addClass(this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS);
    if (this.m_widget.element.is(parentlist)) {
      this.m_widget.element.addClass(this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS);
    }
  }

  if (animate) {
    this._slideAnimation(target, false, currentList.parent(), event, animationResolve);
  } else {
    this._slideAnimationComplete(target, false, currentList.parent(), event, false);
    animationResolve(null);
  }

  this._removeItemFromHviewMenu();

  return animationPromise;
};

oj.SlidingNavListHandler.prototype.UpdateAriaPropertiesOnSelectedItem = function (elem, highlight) {
  if (highlight) {
    elem.attr('aria-describedby', this._selectedLabelId);
  } else {
    elem.removeAttr('aria-describedby');
  }
};
/**
 * Determines whether the specified item is expanded
 * @param {jQuery} item the item element
 * @return {number} 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
 * @private
 */
oj.SlidingNavListHandler.prototype.GetState = function (item) {
  var expanded = this.m_widget.getFocusItem(item).attr("aria-expanded");
  if (expanded === "true") {
    return this.m_widget.STATE_EXPANDED;
  }

  if (expanded === "false") {
    return this.m_widget.STATE_COLLAPSED;
  }
  return this.m_widget.STATE_NONE;
};
/**
 * Sets the disclosed state of the item
 * @param {jQuery} item the item element
 * @param {number} state 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
 * @private
 */
oj.SlidingNavListHandler.prototype.SetState = function (item, state) {
  if (state === this.m_widget.STATE_EXPANDED) {
    this.m_widget.getFocusItem(item)
      .attr("aria-expanded", "true");
    item.removeClass(this.m_widget.COLLAPSED_STYLE_CLASS)
      .addClass(this.m_widget.EXPANDED_STYLE_CLASS);

  } else if (state === this.m_widget.STATE_COLLAPSED) {
    this.m_widget.getFocusItem(item)
      .attr("aria-expanded", "false");
    item.removeClass(this.m_widget.EXPANDED_STYLE_CLASS)
      .addClass(this.m_widget.COLLAPSED_STYLE_CLASS);
  }
};

oj.SlidingNavListHandler.prototype.ModifyListItem = function ($item, itemContent) {
  var focusableElement = this.m_widget.getFocusItem($item);
  $item.attr('role', 'presentation');
  focusableElement.attr('role', 'menuitem');
  if (!itemContent.attr('id')) {
    itemContent.uniqueId();
  }
  // Ensure that it collapses all nodes.
  var groupItems = $item.children('.' + this.m_widget.getGroupStyleClass());
  if (groupItems.length > 0) {
    focusableElement.attr('aria-haspopup', 'true');
    groupItems.attr('role', 'menu');
    groupItems.css("display", "");
    $item.removeAttr('aria-expanded');
    focusableElement.attr('aria-expanded', 'false');
  }
};

oj.SlidingNavListHandler.prototype.BeforeRenderComplete = function () {
  this.m_root.attr('role', 'menu');
  this.m_widget.element.attr('role','presentation');
};

oj.SlidingNavListHandler.prototype.Init = function (opts) {
  this.m_root.addClass('oj-navigationlist-slider')
              .addClass(this.m_widget.getNavListVerticalStyleClass());
  opts.element.addClass('oj-navigationlist-current');
  this._buildSlidingNavListHeader(opts);
  this._initializeHierarchicalView();
};

oj.SlidingNavListHandler.prototype.HandleClick = function (event) {
  if ($(event.target).closest('.oj-navigationlist-previous-link, .oj-navigationlist-previous-button').length > 0) {
    this.CollapseCurrentList(event);
  }
};

oj.SlidingNavListHandler.prototype.HandleKeydown = function (event) {
  if ($(event.target).closest('.oj-navigationlist-previous-link, .oj-navigationlist-previous-button').length > 0) {
    if (event.keyCode === $.ui.keyCode.ENTER) {
      this.CollapseCurrentList(event);
    }
  }

  if ($(event.target).closest('.' + this.m_widget.GetStyleClass()).length > 0) {
    //check for default prevented as it might have already processed for quiting f2 mode
    if (event.keyCode === $.ui.keyCode.ESCAPE && !event.isDefaultPrevented()) {
      this.CollapseCurrentList(event);
    }
  }
};
/**
 * Initialize sliding navigation list header .
 * @private
 * @memberof! oj.ojNavigationList
 */
oj.SlidingNavListHandler.prototype._buildSlidingNavListHeader = function (opts) {
  var selectedLabel;
  this._toolbar = $(document.createElement('div'));
  this._toolbar.addClass('oj-navigationlist-toolbar');
  this._previousLink = $(document.createElement('a'));
  this._prevButton = $(document.createElement('a'));
  this._prevButton.addClass('oj-navigationlist-previous-button');
  this._prevButton.css('visibility', 'hidden')
                  .attr('tabindex', '-1'); // @HTMLUpdateOK
  this._previousLink.addClass('oj-navigationlist-previous-link')
    .attr('tabindex', '-1'); // @HTMLUpdateOK
  this._headerLabel = $(document.createElement('label'));
  this._headerLabel.addClass('oj-navigationlist-current-header')
    .text(this.m_widget.getRootLabel());
  this._vSeparator = $(document.createElement('span'));
  this._vSeparator.attr('role', 'separator')// @HTMLUpdateOK
    .attr('aria-orientation', 'vertical')// @HTMLUpdateOK
    .addClass('oj-navigationlist-toolbar-separator');
  this._hviewBtn = $(document.createElement('button'));
  this._hviewBtn.addClass('oj-navigationlist-hierarchical-button')
                .attr('tabindex', '-1');// @HTMLUpdateOK
  this._hviewMenu = $(document.createElement('ul'));
  this._hviewMenu.addClass('oj-navigationlist-hierarchical-menu').hide();
  selectedLabel = $(document.createElement('label'));
  selectedLabel.uniqueId().addClass('oj-helper-hidden-accessible')
    .attr('aria-hidden', 'true');// @HTMLUpdateOK
  this._selectedLabelId = selectedLabel.attr('id');
  selectedLabel.text(this.m_component.getTranslatedString('selectedLabel'));
  this._previousLink.append(this._headerLabel); // @HTMLUpdateOK
  this._toolbar.append(this._prevButton); // @HTMLUpdateOK
  this._toolbar.append(this._previousLink) // @HTMLUpdateOK
    .append(this._vSeparator) // @HTMLUpdateOK
    .append(this._hviewBtn) // @HTMLUpdateOK
    .append(this._hviewMenu) // @HTMLUpdateOK
    .append(selectedLabel); // @HTMLUpdateOK
  this.m_root.prepend(this._toolbar); // @HTMLUpdateOK
  this._showOrHideHierarchyMenu(opts.hierarchyMenuDisplayThresholdLevel);
};

/**
 * Collapses the current visible list if it is not the top level list
 * @private
 */
oj.SlidingNavListHandler.prototype.CollapseCurrentList = function (event) {
  // pop the expanded item stack
  var current = this.m_expanded.pop();
  if (current) {
    this.m_widget.CollapseItem(current, event, true, null, true, true);
  }
};



/**
 * Initialize hierarchical view for sliding navlist.
 * @private
 * @memberof! oj.ojNavigationList
 */
oj.SlidingNavListHandler.prototype._initializeHierarchicalView = function () {
  var self = this,
    menuid = this._hviewMenu.uniqueId().attr('id');
  this._hviewMenu.ojMenu({
    openOptions: {
      position: {
        "my": "end top",
        "at": "end bottom"
      }
    },
    select: function (event, ui) {
      var itemsToRemove = ui.item.nextAll(),
        currentKey,
        item,
        expandedItems = self.m_expanded,
        targetItemKey = ui.item.data('key');
		self.m_widget.signalTaskStart(); // signal method task start
      //collapse all child lists untill target list is visible
      while (expandedItems.length > 0) {
        item = expandedItems.pop(); // remove item from list
        currentKey = self.m_widget.GetKey(item[0]);
        self.m_widget.CollapseItem($(item), event, true, currentKey, true, true);
        if (self.m_widget.compareValues(targetItemKey, currentKey)) {
          break;
        }
      }
      itemsToRemove.remove();
      ui.item.remove();
      self._hviewMenu.ojMenu('refresh');
	  self.m_widget.signalTaskEnd(); // signal method task end
    }
  });
  this._hviewBtn.ojButton({
    'label': this.m_component.getTranslatedString('hierMenuBtnLabel'),
    'display': 'icons',
    'icons': {
      start: 'oj-fwk-icon oj-hier-icon'
    },
    'menu': '#' + menuid,
    'disabled': true,
    'chroming': 'half'
  });
  this._prevButton.ojButton({
    'label': this.m_component.getTranslatedString('previousIcon'),
    'display': 'icons',
    'icons': {
      start: 'oj-navigationlist-previous-icon oj-component-icon oj-clickable-icon-nocontext'
    },
    'chroming': 'half'
  });
};

/**
 * Add current visible group item to hierarchical menu and update header text with new list's label.
 * @param {string} itemKey key of the group item whose list will be displayed
 * @param {string} label current list's label
 * @param {string} parentLabel parent list's label
 * @private
 */
oj.SlidingNavListHandler.prototype._addItemToHviewMenu = function (itemKey, label, parentLabel) {
  var i;
  if (this._hviewBtn) {
    var itemsinTree = this._hviewMenu.find('li').length;
    var menuListItem = $(document.createElement('li'));
    var menuItem = $(document.createElement('a'));
    menuItem.attr('href', '#');// @HTMLUpdateOK
    menuListItem.append(menuItem); // @HTMLUpdateOK
    if (itemsinTree > 0) {
      for (i = 0; i < itemsinTree; i++) {
        if (i > 0) {
          menuItem.append($(document.createElement('span')) // @HTMLUpdateOK
            .addClass('oj-navigationlist-hvitem-space'));
        }
      }
      menuItem.append($(document.createElement('span')) // @HTMLUpdateOK
        .addClass('oj-menu-item-icon oj-icon oj-navigationlist-level-indicator'));
    }
    var menuItemLabel = $(document.createElement('span'))
      .addClass('oj-navigationlist-hierarchical-menu-label');
    menuItemLabel.text(parentLabel);
    menuItem.append(menuItemLabel); // @HTMLUpdateOK

    //Store item key reference to refresh the list when user clicks on  hiearchical menu item.
    menuListItem.data('key', itemKey);
    this._hviewMenu.append(menuListItem); // @HTMLUpdateOK

    this._hviewMenu.ojMenu('refresh');
    this._showOrHideHierarchyMenu(this.m_widget.GetOption('hierarchyMenuDisplayThresholdLevel'));
    this._hviewBtn.ojButton("option", "disabled", false);
    this._prevButton.css('visibility', 'visible');
    if (this.m_widget.getRootElement().hasClass('oj-focus-ancestor')) {
      // this._previousLink.attr('tabindex', '0');
      this._prevButton.attr('tabindex', '0');// @HTMLUpdateOK
    }
    this._headerLabel.text(label);
  }
};

oj.SlidingNavListHandler.prototype.SetOptions = function (options) {
  
  if(options['hierarchyMenuDisplayThresholdLevel'] === undefined)
    return;

  var currentHMenuDisplayThreshold = this.m_widget.GetOption('hierarchyMenuDisplayThresholdLevel');
  if (currentHMenuDisplayThreshold !== options['hierarchyMenuDisplayThresholdLevel']) {
    this._showOrHideHierarchyMenu(options.hierarchyMenuDisplayThresholdLevel);
  }
};

oj.SlidingNavListHandler.prototype._showOrHideHierarchyMenu = function (hierarchyMenuDisplayThresholdLevel) {
  var itemsinTree = this._hviewMenu.find('li').length;
  if (hierarchyMenuDisplayThresholdLevel === -1 || itemsinTree < hierarchyMenuDisplayThresholdLevel) {
    this._vSeparator.css('visibility', 'hidden');
    if (this._hviewBtn[0] === document.activeElement) { // is(:focus) failing during test cases so using document.activeElement.
      //tried by moving focus to <ul> using  this.m_widget.focus() but listview listen for focusin event.
      this.m_root.focusin();
    }
    this._hviewBtn.css('visibility', 'hidden');
  } else if (itemsinTree >= hierarchyMenuDisplayThresholdLevel) {
    this._vSeparator.css('visibility', 'visible');
    this._hviewBtn.css('visibility', 'visible');
  }
};

/**
 * Removes parent item of current list from hierarchical menu
 * otherwise return the emptyText set in the options
 * @private
 */
oj.SlidingNavListHandler.prototype._removeItemFromHviewMenu = function () {
  if (this._hviewBtn) {
    var removed = this._hviewMenu.find('li').last().remove();
    this._hviewMenu.ojMenu('refresh');
    this._showOrHideHierarchyMenu(this.m_widget.GetOption('hierarchyMenuDisplayThresholdLevel'));
    if (this._hviewMenu.children('li').length === 0) {
      this._hviewBtn.ojButton("option", "disabled", true);
      this._prevButton.css('visibility', 'hidden');
      // this._previousLink.attr('tabindex', '-1');
      this._prevButton.attr('tabindex', '-1');// @HTMLUpdateOK
      this._headerLabel.text(this.m_widget.getRootLabel());
    } else {
      this._headerLabel.text(removed.children('a').text());
    }
  }
};

oj.SlidingNavListHandler.prototype.RestoreItem = function (item, itemContent, sublist) {
  item.removeClass(this.m_widget.SLIDING_NAVLIST_CURRENT_STYLE_CLASS);
  itemContent.removeAttr('role')
    .removeAttr('aria-expanded')
    .removeAttr('aria-describedby')
    .removeAttr('aria-hidden');
};

oj.SlidingNavListHandler.prototype._makeToolbarItemsFocusable = function (enable) {
  var itemsinTree;
  if (enable) {
    itemsinTree = this._hviewMenu.find('li').length;
    if (itemsinTree) {
      // this._previousLink.attr('tabindex', '0');
      this._prevButton.attr('tabindex', '0');
    }
    this._hviewBtn.attr('tabindex', '0');
  } else {
    // this._previousLink.attr('tabindex', '-1');
    this._prevButton.attr('tabindex', '-1');
    this._hviewBtn.attr('tabindex', '-1');
  }
};

/**
 * Handler for focus event
 * @param {Event} event the focus event
 * @protected
 * @override
 */
oj.SlidingNavListHandler.prototype.HandleFocus = function (event) {
  //If focus is on navlist or items but not on toolbar
  if (!($.contains(this._toolbar.get(0), /** @type {Element} */ (event.target)) ||
        this._hviewMenu.get(0) === /** @type {Element} */ (event.relatedTarget))) {
    //make toolbar items focusable
    this._makeToolbarItemsFocusable(true);
    _ojNavigationListView.superclass.HandleFocus.apply(this.m_widget, arguments);
  }
};

/**
 * Handler for blur event
 * @param {Event} event the blur event
 * @protected
 * @override
 */
oj.SlidingNavListHandler.prototype.HandleBlur = function (event) {
  //If focus going to toolbar remove focus ancestor and unhighlight the active one
  if ($.contains(this._toolbar.get(0), /** @type {Element} */ (event.relatedTarget)) ||
      this._hviewMenu.get(0) === /** @type {Element} */ (event.relatedTarget)) {
    this.m_widget.UnhighlightActive();
  } else {
    // if focus moves out side navlist make toolbar items not focusable
    if (event.relatedTarget == null || !$.contains(this.m_root.get(0), /** @type {Element} */ (event.relatedTarget))) {
      this._makeToolbarItemsFocusable(false);
    }
    _ojNavigationListView.superclass.HandleBlur.apply(this.m_widget, arguments);
  }
};

oj.SlidingNavListHandler.prototype.GetNodeBySubId = function (locator) {
  if (locator["subId"] === 'oj-navigationlist-previous-link') {
    return this._prevButton ? this._prevButton[0] : null;
  }

  if (locator["subId"] === 'oj-navigationlist-hierarchical-button') {
    return this._hviewBtn ? this._hviewBtn[0] : null;
  }

  if (locator["subId"] === 'oj-navigationlist-hierarchical-menu') {
    return this._hviewMenu ? this._hviewMenu[0] : null;
  }

  return null;
};

oj.SlidingNavListHandler.prototype.GetSubIdByNode = function (node) {
  if ($(node).closest(this._prevButton).length > 0) {
    return {
      'subId': 'oj-navigationlist-previous-link'
    };
  }

  if ($(node).closest(this._hviewBtn).length > 0) {
    return {
      'subId': 'oj-navigationlist-hierarchical-button'
    };
  }

  if ($(node).closest(this._hviewMenu).length > 0) {
    return {
      'subId': 'oj-navigationlist-hierarchical-menu'
    };
  }

  return null;
};


/**
 * todo: create common utility class between combobox, listview and navlist
 * @private
 */
var _NavigationListUtils = {
  clazz: function (SuperClass, methods) {
    var constructor = function () {};
    oj.Object.createSubclass(constructor, SuperClass, '');
    $.extend(constructor.prototype, methods);
    return constructor;
  }
};
/**
 * @export
 * @class _ojNavigationListView
 * @classdesc NavigationList
 * @constructor
 * @ignore
 */
var _ojNavigationListView = _NavigationListUtils.clazz(oj._ojListView,
  /** @lends _ojNavigationListView.prototype */
  {
    OPTION_DRILL_MODE: 'drillMode',
    OPTION_DRILL_MODE_NONE: 'none',
    OPTION_DRILL_MODE_COLLAPSIBLE: 'collapsible',
    OPTION_DRILL_MODE_SLIDING: 'sliding',
    OPTION_DISPLAY: 'display',
    OPTION_DISPLAY_ICONS: 'icons',
    OPTION_DISPLAY_ALL: 'all',
    OPTION_EDGE: 'edge',
    OPTION_EDGE_TOP: 'top',
    OPTION_EDGE_END: 'end',
    OPTION_EDGE_START: 'start',
    OPTION_EDGE_BOTTOM: 'bottom',
    OPTION_SELECTION: 'selection',
    OPTION_CURRENT_ITEM: 'currentItem',
    TAG_NAME_TAB_BAR:'oj-tab-bar',
    NAVLIST_ITEM_SUBID_KEY: {
      'navlist': 'oj-navigationlist-item',
      'tabbar': 'oj-tabbar-item'
    },
    NAVLIST_EXPANDED_STYLE_CLASS: {
      'navlist': 'oj-navigationlist-expanded',
      'tabbar': 'oj-tabbar-expanded'
    },
    NAVLIST_COLLAPSIBLE_STYLE_CLASS: {
      'navlist': 'oj-navigationlist-collapsible',
      'tabbar': 'oj-tabbar-collapsible'
    },
    NAVLIST_VERTICAL_STYLE_CLASS: {
      'navlist': 'oj-navigationlist-vertical',
      'tabbar': 'oj-tabbar-vertical'
    },
    
    ITEM_CONTENT_STYLE_CLASS: {
      'navlist':'oj-navigationlist-item-content',
      'tabbar':'oj-tabbar-item-content'
    },
    NAVLIST_DIVIDERS_STYLE_CLASS: {
      'navlist':'oj-navigationlist-item-dividers',
      'tabbar':'oj-tabbar-item-dividers'
    },
    LAST_ITEM_STYLE_CLASS: {
      'navlist':'oj-navigationlist-item-last-child',
      'tabbar':'oj-tabbar-item-last-child'
    },
    EXPANDED_STYLE_CLASS: 'oj-expanded',
    COLLAPSED_STYLE_CLASS: 'oj-collapsed',
    SLIDING_NAVLIST_CURRENT_STYLE_CLASS: 'oj-navigationlist-current',
    DIVIDER_STYLE_CLASS: {
      'navlist':'oj-navigationlist-divider',
      'tabbar':'oj-tabbar-divider'
    },
    NAVLIST_OVERFLOW_MENU_ITEM: {
      'navlist':'oj-navigationlist-overflow-menu-item',
      'tabbar':'oj-tabbar-overflow-menu-item'
    },
    NAVLIST_OVERFLOW_MENU: {
      'navlist':'oj-navigationlist-overflow-menu',
      'tabbar':'oj-tabbar-overflow-menu'
    },
    NAVLIST_OVERFLOW_ITEM_ICON: {
      'navlist':'oj-navigationlist-overflow-item-icon',
      'tabbar':'oj-tabbar-overflow-item-icon'
    },
    _CATEGORY_DIVIDER_STYLE_CLASS: {
      'navlist':'oj-navigationlist-category-divider',
      'tabbar':'oj-tabbar-category-divider'
    },
    _ITEM_LABEL_STYLE_CLASS: {
      'navlist':'oj-navigationlist-item-label',
      'tabbar':'oj-tabbar-item-label'
    },
    _ICON_ONLY_STYLE_CLASS: {
      'navlist':'oj-navigationlist-icon-only',
      'tabbar':'oj-tabbar-icon-only'
    },
    _ITEM_ICON_STYLE_CLASS: {
      'navlist':'oj-navigationlist-item-icon',
      'tabbar':'oj-tabbar-item-icon'
    },
    _HORIZONTAL_NAVLIST_STYLE_CLASS: {
      'navlist':'oj-navigationlist-horizontal',
      'tabbar':'oj-tabbar-horizontal'
    },
    _NAVLIST_END_EDGE_STYLE_CLASS: {
      'navlist':'oj-navigationlist-vertical-end',
      'tabbar':'oj-tabbar-vertical-end'
    },
    _NAVLIST_BOTTOM_EDGE_STYLE_CLASS: {
      'navlist':'oj-navigationlist-horizontal-bottom',
      'tabbar':'oj-tabbar-horizontal-bottom'
    },
    _NAVLIST_HAS_ICONS: {
      'navlist': 'oj-navigationlist-has-icons',
      'tabbar': 'oj-tabbar-has-icons'
    },
    _NAVLIST_ITEM_HAS_NO_ICON: {
      'navlist': 'oj-navigationlist-item-no-icon',
      'tabbar': 'oj-tabbar-item-no-icon'
    },
    _NAVLIST_ITEM_TITLE: {
      'navlist':'oj-navigationlist-item-title',
      'tabbar':'oj-tabbar-item-title'
    },
    _NAVLIST_STYLE_CLASS: {
      'navlist': 'oj-navigationlist',
      'tabbar': 'oj-tabbar'
    },
    _NAVLIST_TOUCH_STYLE_CLASS: {
      'navlist': 'oj-navigationlist-touch',
      'tabbar':'oj-tabbar-touch'
    },
    _NAVLIST_LISTVIEW_CONTAINER_STYLE_CLASS: {
      'navlist':'oj-navigationlist-listview-container',
      'tabbar':'oj-tabbar-listview-container'
    },
    _IS_TITLE_ATTR_ADDED_DUE_TO_TRUNCATION: 'IsTitleAttrDueToTruncation',
    _APPLICATION_LEVEL_NAV_STYLE_CLASS: 'oj-navigationlist-app-level',
    _PAGE_LEVEL_NAV_STYLE_CLASS: 'oj-navigationlist-page-level',
    _NAVLIST_ITEM_ICON_HAS_TITLE: 'navigationListItemIconHastitle',
    _NAVLIST_NO_FOLLOW_LINK_CLASS: {
      'navlist':'oj-navigationlist-nofollow-link',
      'tabbar':'oj-tabbar-nofollow-link'
    },
    _CONTAINER_STYLE_CLASS : {
      'navlist': 'oj-navigationlist-listview',
      'tabbar': 'oj-tabbar-listview'
    },
    _ELEMENT_STYLE_CLASS : {
      'navlist': 'oj-navigationlist-element',
      'tabbar': 'oj-tabbar-element'
    },
    _ELEMENT_EMPTY_TEXT_STYLE_CLASS : {
      'navlist': 'oj-navigationlist-empty-text',
      'tabbar': 'oj-tabbar-empty-text'
    },
    _ELEMENT_NO_DATA_MSG_STYLE_CLASS : {
      'navlist': 'oj-navigationlist-no-data-message',
      'tabbar': 'oj-tabbar-no-data-message'
    },
    _FOCUSED_ELEMENT_STYLE_CLASS : {
      'navlist': 'oj-navigationlist-focused-element',
      'tabbar': 'oj-tabbar-focused-element'
    },
    _ITEM_ELEMENT_STYLE_CLASS : {
      'navlist': 'oj-navigationlist-item-element',
      'tabbar': 'oj-tabbar-item-element'
    },
    _ITEM_STYLE_CLASS : {
      'navlist': 'oj-navigationlist-item',
      'tabbar': 'oj-tabbar-item'
    },
    _OPTION_DEFAULT_STYLE_CLASS: {
      'navlist': 'oj-navigationlist-option-defaults',
      'tabbar': 'oj-tabbar-option-defaults'
    },
    /**
     * Returns Item label text
     * @returns {string|null} Item label
     */
    getItemLabel: function (target) {
      var itemContent = this.getItemContentElement(target);
      if (itemContent.is('a')) { // When using simple <a> markup
        return $.trim(itemContent.text());
      }
      //When using arbitrary content, extract item title from element having marker class .oj-navigationlist-item-title.
      return $.trim(itemContent.find('.' + this.getItemTitleStyleClass()).text());
    },
    /**
     * Returns Item content root element.
     * @returns {Object} Returns jquery object for item content root
     */
    getItemContentElement: function (item) {
      var itemContent = item.children('.' + this.getItemContentStyleClass());
      if (itemContent.length === 0) {
        itemContent = item.children('.' + this.getGroupItemStyleClass()).children('.' + this.getItemContentStyleClass());
        if (itemContent.length === 0) { //This case is when item label is invoked before complete initialization of listview
          itemContent = item.children('.' + this.getGroupItemStyleClass()).children(':not(.' + this.getExpandIconStyleClass() + '):not(.' + this.getCollapseIconStyleClass() + ')');
        }
      }
      return itemContent;
    },
    /**
     * Return true when page is in rtl mode otherwise false.
     * @returns {boolean} true if page is in rtl mode
     */
    isRtl: function () {
      return this.ojContext._GetReadingDirection() === "rtl";
    },

    _initNavigationMode: function(rootElement) {
      var tagName = rootElement[0].tagName.toLowerCase();
      
      if(tagName === this.TAG_NAME_TAB_BAR) {
        this._navigationMode = 'tabbar';
      } else {
          this._navigationMode = 'navlist'; //if user uses div jquery widget continue with application level
      }
    },

    _getNavigationMode: function() {
      return this._navigationMode;
    },
    /**
     * Called by content handler once content of all items are rendered
     * @override
     */
    renderComplete: function () {
      this.m_listHandler.BeforeRenderComplete();
      _ojNavigationListView.superclass.renderComplete.apply(this, arguments);
    },
    
    /**
     * Called by content handler once the content of an item is rendered triggered by an insert event
     * @param {Element} elem the item element
     * @param {Object} context the context object used for the item
     */
    itemInsertComplete: function (elem, context) {
      this.m_listHandler.ItemInsertComplete(elem, context);
    },

    /**
     * Called by content handler once the content of an item is removed triggered by an remove event
     * @param {Element} elem the item element
     */
    itemRemoveComplete: function (elem) {
      this.m_listHandler.ItemRemoveComplete(elem);
      _ojNavigationListView.superclass.itemRemoveComplete.apply(this, arguments);
    },
    /**
     * Restore content while restroying list
     * @private
     */
    _restoreContent: function (list) {
      var i, item, itemContent, itemIcon, focusableItem, sublist, items = list.children();
      list.removeAttr('style')
        .removeClass(this.getHasIconsStyleClass())
        .removeAttr('aria-hidden');
      for (i = 0; i < items.length; i++) {
        item = items[i];
        item = $(item);
        if (item.hasClass(this.getDividerStyleClass())) {
          item.remove();
          continue;
        }
        itemContent = this.getItemContentElement(item);
        itemContent.removeClass(this.getItemContentStyleClass())
          .removeClass(this.getHasNoIconStyleClass())
          .removeAttr('aria-haspopup');
        itemIcon = itemContent.children('.' + this.getItemIconStyleClass());
        if (itemIcon && itemIcon.length > 0) {
          itemIcon.removeAttr('role');
          itemIcon.removeAttr('aria-label');
          this._removeToolTipOnIcon(itemIcon);
        }

        item.removeClass('oj-default')
          .removeAttr('role')
          .removeAttr('aria-disabled')
          .removeAttr('aria-expanded')
          .removeAttr('aria-hidden')
          .removeAttr('aria-pressed');
        focusableItem = this.getFocusItem(item);
        focusableItem.removeAttr('role')
          .removeAttr('aria-hidden')
          .removeAttr('aria-pressed');
        itemContent.children('.' + this.getItemLabelStyleClass())
          .contents().unwrap();
        if (focusableItem.data(this._IS_TITLE_ATTR_ADDED_DUE_TO_TRUNCATION)) {
          focusableItem.removeAttr('title');
        }
        sublist = item.children('ul');
        if (sublist.length > 0) {
          this.m_listHandler.RestoreItem(item, itemContent, sublist);
          this._restoreContent(sublist);
        } else {
          this.m_listHandler.RestoreItem(item, itemContent);
        }
      }
    },
    
    /**
     * Returns DnD Context
     * @private
     * @override
     */
    GetDnDContext: function(){
      if (typeof oj.NavigationListDndContext != "undefined" && 
              this.ojContext.element[0].tagName.toLowerCase() === this.TAG_NAME_TAB_BAR)
      {
        return new oj.NavigationListDndContext(this);
      }
    },
    
    /**
     * Overriding to exclude oj-navigationlist-item-content from clickable components.
     * @override
     * @protected
     */
    IsNodeEditableOrClickable: function (node) {
      if (node.hasClass(this.getItemContentStyleClass()) || node.hasClass(this.getExpandIconStyleClass()) || node.hasClass(this.getCollapseIconStyleClass())) {
        return false;
      }
      return _ojNavigationListView.superclass.IsNodeEditableOrClickable.apply(this, arguments);
    },
    /**
     * checks for whether a node is editable or clickable.
     * @param {jQuery} node  Node
     * @return {boolean} true or false
     * @override
     * @protected
     */
    IsElementEditableOrClickable: function (node) {
      var nodeName = node.prop("nodeName");
      //Exclude navlist item content <a> tag from below list and also checking for oj-component class as it is needed for toggle button.
      return (nodeName.match(/^INPUT|SELECT|OPTION|BUTTON|^A\b|TEXTAREA/) && !node.hasClass(this.getItemContentStyleClass())) || node.hasClass('oj-component');
    },
    
    _focusable: function(context) {
      if ($(context['data']).is('li')) {
        return !$(context['data']).hasClass('oj-disabled');
      }
      return !$(context['parentElement']).hasClass('oj-disabled');
    },
    
    /**
     * Prepare options to initialize ListView base class
     * @private
     */
    _prepareListViewOptions: function (navlistOptions) {

      var self = this,
        opts = $.extend({}, navlistOptions);

      opts["drillMode"] = navlistOptions['drillMode'] !== 'none' ? 'collapsible' : 'none';
      opts["selection"] = navlistOptions['selection'] !== null ? [navlistOptions['selection']] : [];
      opts["selectionMode"] = 'single';
      opts["item"] = $.extend({
        "focusable": this._focusable
      }, navlistOptions['item']);

      opts.element = this._list;
      return opts;
    },
    /**
     * Returns root label for navigation list.
     * otherwise return the emptyText set in the options
     * @return {string} Label for navigation list
     */
    getRootLabel: function () {
      if (this.ojContext.options['rootLabel']) {
        return this.ojContext.options['rootLabel'];
      }

      return this.ojContext.getTranslatedString('defaultRootLabel');
    },

    /**
     * Sets wai-aria properties on root element
     * @override
     * @protected
     */
     SetAriaProperties: function () {
       this.ojContext.element.attr("aria-multiselectable", false);
     },
    /**
     * Handler for focus event
     * @param {Event} event the focus event
     * @protected
     * @override
     */
    HandleFocus: function(event)
    {
      return this.m_listHandler.HandleFocus(event);
    },
    /**
     * Handler for blur event
     * @param {Event} event the blur event
     * @protected
     * @override
     */
    HandleBlur: function(event)
    {
      return this.m_listHandler.HandleBlur(event);
    },
    /**
     * Sets the tab index attribute of the root element
     * @override
     */
    SetRootElementTabIndex: function()
    {
        this.ojContext.element.attr("tabIndex", 0);
    },

    /**
     * Removes the tab index attribute of the root element
     * @override
     */
    RemoveRootElementTabIndex: function()
    {
        this.ojContext.element.removeAttr("tabIndex");
    },
    /**
     * Removes wai-aria properties on root element
     * @override
     * @protected
     */
     UnsetAriaProperties: function () {
       this.ojContext.element.removeAttr("aria-activedescendant")
                   .removeAttr("aria-multiselectable");
     },
    /**
     * Whether to use grid role
     * @override
     */
    ShouldUseGridRole: function()
    {
        return false;
    },
    /**
     * Overriding init to initialize list with respective List handler.
     * @override
     */
    init: function (navlistopts) {
      var opts, 
        self = this,
        element = navlistopts.ojContext.element;
      this._initNavigationMode(element);
      
      element.addClass(this.getNavListStyleClass());

      if (oj.DomUtils.isTouchSupported()) {
        element.addClass(this.getNavListTouchStyleClass());
      }

      this._list = element.children('ul:first');
      if (this._list.length === 0) {
        this._list = $(document.createElement("ul"));
        element.append(this._list); //@HTMLUpdateOK
      }
      
      var listViewContainer = $(document.createElement("div"));
      listViewContainer.addClass(this.getNavListContainerStyleClass());
      listViewContainer.attr('role', 'presentation');

      this._list.wrap(listViewContainer); // @HTMLUpdateOK
      opts = this._prepareListViewOptions(navlistopts);
     
      _ojNavigationListView.superclass.init.call(this, opts);
      this.getRootElement().attr('role', 'presentation');
      this.element.removeClass('oj-component-initnode');
      this.ojContext._on(this.ojContext.element, 
      {
        "click": function (event) {
          self.m_listHandler.HandleClick(event);
        },
        
        "keydown": function (event) {
          self.m_listHandler.HandleKeydown(event);
        },
        
        "mouseup": function (event) {
          self._clearActiveState(event);
        },
        
        "mouseover": function (event) {
          if($(event.target).closest("a." + self.getItemContentStyleClass()).length > 0) {
            var $itemLink = $(event.target).closest("a." + self.getItemContentStyleClass());
            var $label = $itemLink.find('.' + self.getItemLabelStyleClass());
            //Add title attribute only when the text is truncated.
            if ($label[0].offsetWidth < $label[0].scrollWidth && !$itemLink.attr('title')) {
              $itemLink.attr('title', $label.text().trim());
              $itemLink.data(self._IS_TITLE_ATTR_ADDED_DUE_TO_TRUNCATION, 'true');
            }
          }
        },
        
        "mouseout": function (event) {
          if($(event.target).closest("a." + self.getItemContentStyleClass()).length > 0) {
            //Remove title attribute on mouseleave.
            var $itemLink = $(event.target).closest("a." + self.getItemContentStyleClass());
            if ($itemLink.data(self._IS_TITLE_ATTR_ADDED_DUE_TO_TRUNCATION)) {
              $itemLink.removeData(self._IS_TITLE_ATTR_ADDED_DUE_TO_TRUNCATION);
              $itemLink.removeAttr('title');
            }
            self._clearActiveState(event);
          }
        }
      });
    },

    _initListHandler: function () {
      var edge, drillMode; 
      drillMode = this.ojContext.options['drillMode'];
      edge = this.ojContext.options['edge'];
      
      if (drillMode === this.OPTION_DRILL_MODE_SLIDING) {
        this.m_listHandler = new oj.SlidingNavListHandler(this, this.ojContext.element, this.ojContext);
      } else if (drillMode === this.OPTION_DRILL_MODE_COLLAPSIBLE) {
        this.m_listHandler = new oj.CollapsibleNavListHandler(this, this.ojContext.element, this.ojContext);
      } else if (drillMode === this.OPTION_DRILL_MODE_NONE && (edge === this.OPTION_EDGE_TOP || edge === this.OPTION_EDGE_BOTTOM)) {
        this.m_listHandler = new oj.HorizontalNavListHandler(this, this.ojContext.element, this.ojContext);
      } else {
        this.m_listHandler = new oj.DefaultNavListHandler(this, this.ojContext.element, this.ojContext);
      }

      this.m_listHandler.Init(this.options);
      var navigationLevel = this.ojContext.options['navigationLevel'];
      this._setNavigationLevel(navigationLevel);
    },

    _setNavigationLevel: function (navigationLevel) {
      var drillMode = this.ojContext.options['drillMode'];
      if (drillMode === 'none') {
        if (navigationLevel === 'application' || this.ojContext.element[0].tagName.toLowerCase() === 'oj-navigation-list') {
          this.ojContext.element.addClass(this._APPLICATION_LEVEL_NAV_STYLE_CLASS);
          this.ojContext.element.removeClass(this._PAGE_LEVEL_NAV_STYLE_CLASS);
        } else if (navigationLevel === 'page') {
          this.ojContext.element.addClass(this._PAGE_LEVEL_NAV_STYLE_CLASS);
          this.ojContext.element.removeClass(this._APPLICATION_LEVEL_NAV_STYLE_CLASS);
        }
      }
    },
    _clearActiveState: function (event) {
      var item = this.FindItem($(event.target));
      if (item != null) {
        this.HighlightUnhighlightElem(item, "oj-active", false);
      }
    },
    /**
     * Initialize the listview after creation
     * @override
     */
    afterCreate: function () {
      this._initListHandler();
      _ojNavigationListView.superclass.afterCreate.apply(this, arguments);
    },

    /**
     * Gets the animation effect for the specific action
     * @param {string} action the action to retrieve the effect
     * @return {Object} the animation effect for the action
     */
    getAnimationEffect: function(action)
    {
      return this.m_listHandler.GetAnimationEffect(action);
    },
    
    notifyAttached: function()
    {
        _ojNavigationListView.superclass.notifyAttached.apply(this, arguments);
        this.m_listHandler.NotifyAttached();   
    },

    HandleResize: function(width, height)
    {
      var self = this;
      _ojNavigationListView.superclass.HandleResize.apply(this, arguments);
      if (width > 0 && height > 0 && this.m_listHandler != null)
        {
          self.m_listHandler.HandleResize(width, height);   
        }
    },
    
    BeforeInsertItem: function(){
      this.m_listHandler.BeforeInsertItem();
    },
    
    /**
     * Event handler for when mouse down or touch start anywhere in the list
     * @param {Event} event mousedown or touchstart event
     * @protected
     */
    HandleMouseDownOrTouchStart: function (event) {
      _ojNavigationListView.superclass.HandleMouseDownOrTouchStart.apply(this, arguments);
      var item = this.FindItem($(event.target));
      if (item && !item.hasClass("oj-disabled")) {
        this.HighlightUnhighlightElem(item, "oj-active", true);
      }
    },

    /**
     * Event handler for when touch end/cancel happened
     * @param {Event} event touchend or touchcancel event
     * @protected
     */
    HandleTouchEndOrCancel: function (event) {
      this._clearActiveState(event);
      _ojNavigationListView.superclass.HandleTouchEndOrCancel.apply(this, arguments);
    },
    getItemContentStyleClass: function() {
      return this.ITEM_CONTENT_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListExpandedStyleClass: function() {
      return this.NAVLIST_EXPANDED_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListCollapsibleStyleClass: function() {
      return this.NAVLIST_COLLAPSIBLE_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListVerticalStyleClass: function() {
      return this.NAVLIST_VERTICAL_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListItemsDividerStyleClass: function() {
      return this.NAVLIST_DIVIDERS_STYLE_CLASS[this._getNavigationMode()];
    },
    getLastItemStyleClass: function() {
      return this.LAST_ITEM_STYLE_CLASS[this._getNavigationMode()];
    },
    getDividerStyleClass: function() {
      return this.DIVIDER_STYLE_CLASS[this._getNavigationMode()];
    },
    getOverflowItemStyleClass: function() {
      return this.NAVLIST_OVERFLOW_MENU_ITEM[this._getNavigationMode()];
    },
    getOverflowMenuStyleClass: function() {
      return this.NAVLIST_OVERFLOW_MENU[this._getNavigationMode()];
    },
    getOverflowItemIconStyleClass: function() {
      return this.NAVLIST_OVERFLOW_ITEM_ICON[this._getNavigationMode()];
    },
    getCategoryDividerStyleClass: function() {
      return this._CATEGORY_DIVIDER_STYLE_CLASS[this._getNavigationMode()];
    },
    getItemLabelStyleClass: function() {
      return this._ITEM_LABEL_STYLE_CLASS[this._getNavigationMode()];
    },
    getIconOnlyStyleClass: function() {
      return this._ICON_ONLY_STYLE_CLASS[this._getNavigationMode()];
    },
    getItemIconStyleClass: function() {
      return this._ITEM_ICON_STYLE_CLASS[this._getNavigationMode()];
    },
    getHorizontalNavListStyleClass: function() {
      return this._HORIZONTAL_NAVLIST_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListEndEdgeStyleClass: function(){
      return this._NAVLIST_END_EDGE_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListBottomEdgeStyleClass: function(){
      return this._NAVLIST_BOTTOM_EDGE_STYLE_CLASS[this._getNavigationMode()];
    },
    getHasIconsStyleClass: function() {
      return this._NAVLIST_HAS_ICONS[this._getNavigationMode()];
    },
    getHasNoIconStyleClass: function() {
      return this._NAVLIST_ITEM_HAS_NO_ICON[this._getNavigationMode()];
    },
    getItemTitleStyleClass: function() {
      return this._NAVLIST_ITEM_TITLE[this._getNavigationMode()];
    },
    getNavListStyleClass: function() {
      return this._NAVLIST_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListTouchStyleClass: function() {
      return this._NAVLIST_TOUCH_STYLE_CLASS[this._getNavigationMode()];
    },
    getNavListContainerStyleClass: function() {
      return this._NAVLIST_LISTVIEW_CONTAINER_STYLE_CLASS[this._getNavigationMode()];
    },
    getNoFollowLinkStyleClass: function() {
      return this._NAVLIST_NO_FOLLOW_LINK_CLASS[this._getNavigationMode()];
    },
    getItemSubIdKey: function(){
      if (this.isTabBar()) {
        return this.NAVLIST_ITEM_SUBID_KEY['tabbar'] 
      } 
      return this.NAVLIST_ITEM_SUBID_KEY['navlist'];
    },

    /**
     * Returns Navlist specific container style class
     * @override
     * @protected
     */
    GetContainerStyleClass: function () {
      return this._CONTAINER_STYLE_CLASS[this._getNavigationMode()];
    },

    /**
     * Returns Navlist specific root element style class
     * @override
     * @protected
     */
    GetStyleClass: function () {
      return this._ELEMENT_STYLE_CLASS[this._getNavigationMode()];
    },
    
    /**
     * Returns Navlist specific Item style class
     * @override
     */
    getEmptyTextMarkerClass: function () {
      return this._ELEMENT_EMPTY_TEXT_STYLE_CLASS[this._getNavigationMode()];
    },

    /**
     * Returns Navlist specific Item style class
     * @override
     */
    getEmptyTextStyleClass: function () {
      return this._ELEMENT_NO_DATA_MSG_STYLE_CLASS[this._getNavigationMode()];
    },

    getNavListRemoveIcon: function () {
      return "oj-tabbar-remove-icon";
    },

    getNavListRemoveCommand: function () {
      return "oj-tabbar-remove";
    },

    getRemovableStyleClass: function () {
      return 'oj-removable';
    },

    /**
     * Returns Navlist specific Item style class
     * @override
     */
    getItemStyleClass: function () {
      return this._ITEM_STYLE_CLASS[this._getNavigationMode()];
    },

    getFocusedElementStyleClass: function () {
      return this._FOCUSED_ELEMENT_STYLE_CLASS[this._getNavigationMode()];
    },
    /**
     * Returns Navlist specific Item element style class
     * @override
     */
    getItemElementStyleClass: function () {
      return this._ITEM_ELEMENT_STYLE_CLASS[this._getNavigationMode()];
    },
    /**
     * Returns Navlist specific collapse icon style class
     * @override
     */
    getCollapseIconStyleClass: function () {
      return "oj-navigationlist-collapse-icon";
    },
    /**
     * Returns Navlist specific expand icon style class
     * @override
     */
    getExpandIconStyleClass: function () {
      return "oj-navigationlist-expand-icon";
    },
    /**
     * Returns Navlist specific depth style class
     * @override
     */
    getDepthStyleClass: function (depth) {
      return "oj-navigationlist-depth-" + depth;
    },
    /**
     * Returns Navlist specific Group Item style class
     * @override
     */
    getGroupItemStyleClass: function () {
      return "oj-navigationlist-group-item";
    },
    /**
     * Returns Navlist specific Group style class
     * @override
     */
    getGroupStyleClass: function () {
      return "oj-navigationlist-group";
    },
    /**
     * Returns Navlist specific group expand style class
     * @override
     */
    getGroupExpandStyleClass: function () {
      return "oj-navigationlist-collapsible-transition";
    },
    /**
     * Returns Navlist specific group collapse style class
     * @override
     */
    getGroupCollapseStyleClass: function () {
      return this.getGroupExpandStyleClass();
    },
    
    /**
     * Override options default style class
     * @return {string} the option defaults style class
     */
    getOptionDefaultsStyleClass: function() {
        return this._OPTION_DEFAULT_STYLE_CLASS[this._getNavigationMode()];
    },    
    /**
     * Override expand animation behaviour to support differnet types of navlists.
     * Returns a Promise that resolves when animation completes
     * @override
     * @protected
     */
    AnimateExpand: function (groupItem, animate, event) {
      return this.m_listHandler.Expand(groupItem, animate, event);
    },
    /**
     * Override collapse animation behaviour to support differnet types of navlists.
     * Returns a Promise that resolves when animation completes
     * @override
     * @protected
     */
    AnimateCollapse: function (item, key, animate, event) {
      return this.m_listHandler.Collapse(item, key, animate, event);
    },
    /**
     * Handles arrow keys navigation on item
     * @param {number} keyCode description
     * @param {boolean} isExtend
     * @param {Event} event the DOM event causing the arrow keys
     * @override
     * @protected
     */
    HandleArrowKeys: function (keyCode, isExtend, event) {
      return this.m_listHandler.HandleArrowKeys(keyCode, isExtend, event);
    },
    /**
     * Determine whether the key code is an arrow key
     * @param {number} keyCode
     * @return {boolean} true if it's an arrow key, false otherwise
     * @override
     * @protected
     */
    IsArrowKey: function (keyCode) {
      return this.m_listHandler.IsArrowKey(keyCode);
    },
    /**
     * Determines whether the specified item is expanded
     * @param {jQuery} item the item element
     * @return {number} 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
     * @override
     * @protected
     */
    GetState: function (item) {
      return this.m_listHandler.GetState(item);
    },
    /**
     * Sets the disclosed state of the item
     * @param {jQuery} item the item element
     * @param {number} state 0 if item is expanded, 1 if item is collapsed, 2 if item cannot be expand or collapse.
     * @override
     * @protected
     */
    SetState: function (item, state) {
      this.m_listHandler.SetState(item, state);
    },

    /**
     * Override to expand and collapse events on clicking item instead of icon. Also avoids expand/collapse when user click on arbitrary clickable elements.
     * @override
     * @protected
     */
    HandleMouseClick: function (event) {
      var $target = $(event.target);
      var item = this.FindItem($target);
      if (item == null || item.length === 0) {
        // can't find item or if item cannot be focus
        return;
      }
      if ($target.hasClass(this.getNavListRemoveIcon()) && this.isTabBar()) {
        this._handleRemove(event, item);
        return;
      }
      if (this.SkipFocus(item)) {
        event.preventDefault();
        return;
      }

      //TODO , once it is fixd in listview probably don't need to override HandleMouseClick

      if (this.IsNodeEditableOrClickable($target)) {
        return;
      }

      var itemContent = this.getItemContentElement(item);
      var url = itemContent.attr('href');
      if (url && url !== "#") {
        if (event.button === 0 && (event.shiftKey || event.ctrlKey)) {
          return; // Do nothing, as Browser will launch  target it in new tab or browser
        }
      }

      _ojNavigationListView.superclass.HandleMouseClick.apply(this, arguments);
      event.preventDefault();
    },
    /**
     * Override to support flip left/right arrows for expand and collapse actions.
     * @override
     * @protected
     */
    HandleKeyDown: function (event) {
      var keyCode, current, currentItemKey;
      if (!this.m_active) {
        //If there is no active item, generally happes when user click on disabled item 
        //and immediately press a key.
        return;
      }
      keyCode = event.keyCode;
      current = this.m_active['elem'];
      currentItemKey = this.m_active['key'];

      if (keyCode === $.ui.keyCode.HOME || keyCode === $.ui.keyCode.END) {
        var item;
        if (keyCode === $.ui.keyCode.HOME) {
          item = this.element.find('.' + this.getItemContentStyleClass() + ':visible').first().closest('.' + this.getItemElementStyleClass());
        } else {
          item = this.element.find('.' + this.getItemContentStyleClass() + ':visible').last().closest('.' + this.getItemElementStyleClass());
        }
        this.SetCurrentItem(item, event);
        event.preventDefault();
      } else if (keyCode === $.ui.keyCode.DELETE && this.isTabBar()) {
        this._handleRemove(event, current);
      } else {
        var processed = this.HandleSelectionOrActiveKeyDown(event);
        var processExpansion = this.m_listHandler.HandleExpandAndCollapseKeys(event, keyCode, current, currentItemKey);
        processed = processed || processExpansion || (this.m_dndContext != null && this.m_dndContext.HandleKeyDown(event));
        if (processed) {
          event.preventDefault();
        }
      }
    },

    AvoidFocusHighLight: function (flag) {
      this._avoidFocusHighLight = flag;
    },
    
    /**
     * check Whether recent pointer acivity happened or not. 
     * Only used for sliding navlist to avoid focus ring on new focusable item
     * after completing expand/collapse animation.
     * @protected
     * @override
     */
    RecentPointerCallback: function () {
      var self = this;
      return function () {
        return !!(self._avoidFocusHighLight);
      };
    },
    /**
     * Whether focus highlight should be applied
     * @return {boolean} true if should apply focus highlight, false otherwise
     * @protected
     */
    ShouldApplyHighlight: function () {
      return true;
    },
    /**
     * Override to trigger beforeSelect event when user selecets an item.
     * @override
     * @protected
     */
    ToggleSelection: function (event, keepCurrentSelection, skipIfNotSelected) {
      var item = this.m_active['elem'];

      if (!this.IsSelectable(item[0])) {
        return;
      }
      if (this._isSelected(item)) {
        return;
      }
      var shouldSelect = this._fireBeforeSelectEvent(event, item);
      if (shouldSelect) {
        _ojNavigationListView.superclass.ToggleSelection.apply(this, arguments);
        this._initiateNavigation(item);
      }
    },
    /**
     * Highlight or unhighlight an element
     * @param {jQuery|Element} elem the element the highlight or unhighlight
     * @param {string} style the style to add or remove
     * @param {boolean} highlight true if it's to highlight, false if it's to unhighlight
     * @override
     * @protected
     */
    HighlightUnhighlightElem: function (elem, style, highlight) {
      var group;

      elem = $(elem);

      if (style === "oj-selected") {
        this.m_listHandler.UpdateAriaPropertiesOnSelectedItem(this.getFocusItem(elem), highlight);
      }

      // if item is a group, the highlight should be apply to the group item element
      group = elem.children("." + this.getGroupItemStyleClass());
      if (group.length > 0) {
        elem = $(group[0]);
      }

      if (style === "oj-focus") {
        if (highlight) {
          this._focusInHandler(elem);
        } else {
          this._focusOutHandler(elem);
        }
      } else {
        if (highlight) {
          elem.addClass(style);
        } else {
          elem.removeClass(style);
        }
      }

      if (elem.hasClass('oj-selected') || elem.hasClass('oj-hover') || elem.hasClass('oj-active') || elem.hasClass('oj-disabled')) {
        elem.removeClass('oj-default');
      } else {
        elem.addClass('oj-default');
      }
    },
    /**
     * In Navlist,it is not gaurenteed that item will have aria-selected. So overriding it to make all items selectable.
     * @override
     * @protected
     */
    IsSelectable: function (item) {
      return !$(item).hasClass('oj-disabled') && this.m_listHandler.IsSelectable($(item));
    },
    /**
     * Override to skip firing events like ready which are not needed for navlist.
     * @override
     * @protected
     */
    Trigger: function (type, event, ui) {
      if (type !== 'ready') {
        return this.ojContext._trigger(type, event, ui);
      }
      return true;
    },
    /**
     * Override to convert selection option value from array to string before setting on component.
     * @override
     * @protected
     */
    SetOption: function (key, value, flags) {
      if (this.m_listHandler.IsOptionUpdateAllowed(key, value, flags)) {
        var modifiedValue = null;
        if (key === 'selection') { // selection Data type is different for listview and navlist
          //change items to item as navlist does't allow multiple selection.
          var context = flags['_context'],
            selectedItem = context && context.extraData && context.extraData['items'];
          if (selectedItem) {
            //extraData['items'] is a jquery object(not an array) and in case of navlist it will always have single item.
            context.extraData['item'] = this.ojContext._IsCustomElement() ? (selectedItem.length === 0 ? null: selectedItem[0]):selectedItem;
            context.extraData['items'] = undefined;
          }
          
          if (value && value.length > 0) {
            modifiedValue = value[0];
          }
          this._fireDeselectEvent(context.originalEvent, selectedItem, value);
        } else {
          modifiedValue = value;
        }
        
        // Common options
        this.ojContext.option(key, modifiedValue, flags);
        this.options[key] = value;
        this.m_listHandler.OptionUpdated(key, modifiedValue, flags);
      }
    },
    /**
     * Whether ListView should refresh if certain option is updated
     * @param {Object} options the options to check
     * @return {boolean} true if should refresh, false otherwise
     * @override
     * @protected
     */
    ShouldRefresh: function (options) {
      return (options['data'] != null ||
        options['drillMode'] != null ||
        options['item'] != null ||
        options['display'] != null ||
        options['edge'] != null);
    },


    /**
     * Sets multiple options
     * @param {Object} options the options object
     * @param {Object} flags additional flags for option
     * @return {Object} result result object holding skipUpdate and needRefresh flags.
     * @override
     */
    setOptions: function (options, flags) {
      var result = {
          skipOptions: []
        },
        newSelectionValue;

      this.m_listHandler.SetOptions(options);

      if (options['navigationLevel'] !== undefined) {
        this._setNavigationLevel(options['navigationLevel']);
      }
      if (options['selection'] !== undefined) {
        newSelectionValue = options['selection'];
      }
      result.needRefresh = _ojNavigationListView.superclass.setOptions.call(this, options, flags);
      if (newSelectionValue !==  undefined && options['selection'] === undefined) {
        result.skipOptions.push('selection');
      }
      //Restore original user provided selection value
      if (newSelectionValue !== undefined) {
        options['selection'] = newSelectionValue;
      }
      return result;
    },
    compareValues: function(value1, value2){
      return value1 == value2 || oj.Object.compareValues(value1, value2);
    },
    /**
     * Set Selection option.
     * @param {Object} options the options object
     * @override
     */
    HandleSelectionOption: function (options) {
      var newSelectionValue;
      if (options['selection'] !== undefined) {
        newSelectionValue = options['selection'];
        if (newSelectionValue) {
          var selection = this.GetOption('selection');
          if (!selection || selection.length === 0 || !this.compareValues(selection[0],newSelectionValue)) {
            var item = this.FindElementByKey(newSelectionValue);
            if (!item || this.IsSelectable(item)) {
              var shouldSelect = this._fireBeforeSelectEvent(null, $(item), newSelectionValue);
              if (shouldSelect) {
                this._fireDeselectEvent(null, item, newSelectionValue);
                options['selection'] = [newSelectionValue];
                if (item)
                  this._initiateNavigation($(item));
              } else {
                delete options['selection'];
              }
            } else {
              delete options['selection'];
            }
          } else {
            //When new value is same as existing one, no need to update it on listview. 
            delete options['selection'];
          }
        } else {
          options['selection'] = [];
        }
      }
      _ojNavigationListView.superclass.HandleSelectionOption.call(this, options);
    },
    /**
     * Override to convert selection option value from string to array before setting on component.
     * @override
     * @protected
     */
    GetOption: function (key) {
      var optionValue = this.ojContext.option(key);
      if (key === 'selection') { // selection Data type is different for listview and navlist
        if (optionValue) {
          return [optionValue];
        }
        return [];
      }

      if(key === 'item' && !optionValue["focusable"]) {
        optionValue["focusable"] = this._focusable;
      }

      if (optionValue === null) { // if the option is only applicable to listview
        return this.options[key];
      }
      return optionValue;
    },
    /**
     * Override to return widget constructor
     * @override
     */
    getWidgetConstructor: function () {
      return oj.Components.__GetWidgetConstructor(this.ojContext.element);
    },
    /**
     * Override to skip intermediate nodes(only in case of collapsible) and disabled items from selection.
     * @override
     * @protected
     */
    SelectAndFocus: function (item, event) {
      //Disabled items are not selectable and focuable.
      if (item.hasClass('oj-disabled')) {
        return;
      }

      if (!this.IsSelectable(item[0])) {
        return;
      }

      if (!this._isSelected(item) && this._fireBeforeSelectEvent(event, item)) {
        _ojNavigationListView.superclass.SelectAndFocus.apply(this, arguments);
        this._initiateNavigation(item);
      } else {
        //clicking on selected item/item whose selection is prevented should also get focus.
        this.HandleClickActive(item, event);

      }
    },
    
    isTabBar: function() {
      return this.ojContext.element[0].tagName.toLowerCase() === this.TAG_NAME_TAB_BAR;
    },
    
    _fireBeforeDeselectEvent: function (event, item, key) {
      if (!key)
        key = this.GetKey(item[0]);
      
      var fromKey = this.GetOption('selection');
      var fromItem = this.FindElementByKey(fromKey);
      return this.Trigger("beforeDeselect", event, {
        'toItem': item,
        'toKey': key,
        'fromItem': fromItem,
        'fromKey': fromKey,
      });
    },
    
    _fireBeforeSelectEvent: function (event, item, key) {
      var beforeDeselect = true;
      
      if (!key)
        key = this.GetKey(item[0]);
      
      if (this.isTabBar()) {
        beforeDeselect = this._fireBeforeDeselectEvent(event, item, key);
      }
      return beforeDeselect && this.Trigger("beforeSelect", event, {
        'item': item,
        'key': key
      });
    },
    
    _fireRemoveEvent: function (event, item) {
      var key = this.GetKey(item[0]);
      var beforeRemove = this.Trigger("beforeRemove", event, {
        'item': item,
        'key': key
      });
      return beforeRemove && this.Trigger("remove", event, {
        'item': item,
        'key': key
      });
    },
    
    _fireDeselectEvent: function(event, item, key) {
      var fromKey = this.GetOption('selection');
      var fromItem = this.FindElementByKey(fromKey);
      this.Trigger("deselect", event, {
        'toItem': item,
        'toKey': key,
        'fromItem': fromItem,
        'fromKey': fromKey
      });
    },
    
    _initiateNavigation: function (item) {

      if (this.ojContext.element.hasClass(this.getNoFollowLinkStyleClass())) {
        return false;
      }

      var itemContent = this.getItemContentElement(item);
      var url = itemContent.attr('href');
      var target = itemContent.attr('target');;
      if (url && url !== "#") {
        //In case of javascript uri, javascript will get executed on assigning it to href.
        //Ideally user can use beforeSelect/optionChange events to do this,Will there be any issue in supporting this?
         if(!target || target === '_self') {
            window.location.href = url;
         } else {
            window.open(url, target);
         }
         
        return true;
      }
      return false;
    },
    _isSelected: function (item) {
      var selection, key;

      selection = this.GetOption("selection");
      key = this.GetKey(item[0]);
      if(selection && selection.length === 1 && this.compareValues(selection[0], key)) {
        return true;
      }
      return false;
    },
    _setToolTipOnIcon: function (icon, itemLabel) {
      if (!icon.attr('title')) {
        icon.attr('title', itemLabel);
      }
    },
    _removeToolTipOnIcon: function (icon) {
      if (!icon.data(this._NAVLIST_ITEM_ICON_HAS_TITLE)) {
        icon.removeAttr('title');
      } else {
        icon.removeData(this._NAVLIST_ITEM_ICON_HAS_TITLE);
      }
    },
    
    /**
     * Override to ignore remove icon link.
     * @param {jQuery} item  Item
     * @override
     * @private
     * @ignore
     */
    getSingleFocusableElement: function(item) {
        var selector, childElements, self = this;

        selector = "a, input, select, textarea, button";
        childElements = item.children(selector).filter(function(index){
          return !$(this).hasClass(self.getNavListRemoveIcon());
        });;

        if(childElements.length === 1 && // check for only one focusbale child
           childElements.first().find(selector).length === 0)
        {
            //check to ensure no nested focusable elements.
            return childElements.first();
        }
        return item;
    },
    
    /**
     * Override to decorate remove menu item
     * @param {jQuery} item  Item
     * @override
     * @ignore
     */
    PrepareContextMenu: function (item) {
      if (this.m_dndContext != null)
      {
        this.m_dndContext.prepareContextMenu(this.ojContext._GetContextMenu());
      }
      var contextMenu = this.ojContext._GetContextMenu();
      if (item.hasClass(this.getRemovableStyleClass()) && contextMenu) {
        if (this.m_contextMenu != contextMenu)
        {
          this.m_contextMenu = contextMenu;
          contextMenu.addEventListener("ojBeforeOpen", this._handleContextMenuBeforeOpen.bind(this));
          contextMenu.addEventListener("ojAction", this._handleContextMenuSelect.bind(this));
        }
        var removeItem = $(contextMenu).find('[data-oj-command=' + this.getNavListRemoveCommand() +']');
        var textNode = this.ojContext.getTranslatedString('labelRemove');
        if (!this.ojContext._IsCustomElement())
            removeItem.empty().append($('<a href="#"></a>').text(textNode)); //@HTMLUpdateOK
        else
            removeItem.empty().append(document.createTextNode(textNode)); //@HTMLUpdateOK
        contextMenu.refresh();
      }
    },
    
    _handleContextMenuSelect: function (event, ui) {
      var item = $(event.target);
      if (item.attr("data-oj-command") === this.getNavListRemoveCommand()) {
        this._handleRemove(event, this.m_contextMenuItem);
      }
    },
    
    _handleContextMenuBeforeOpen: function (event) {
      this.m_contextMenuItem = event['detail']['openOptions']['launcher'];
    },
    
    _handleRemove: function (event, item) {
      if(item.hasClass(this.getRemovableStyleClass()))
        this._fireRemoveEvent(event, item);
    },
    /**
     * Called by content handler once the content of an item is rendered
     * @param {Element} elem the item element
     * @param {Object} context the context object used for the item
     */
    itemRenderComplete: function (elem, context) {
      var $item = $(elem);
      if ($item.hasClass(this.getCategoryDividerStyleClass())) {
        $item.removeClass(this.getItemElementStyleClass());
        $item.removeClass(this.getFocusedElementStyleClass());
        $item.removeClass(this.getItemStyleClass());
        $item.removeAttr('aria-selected');
        $item.children().remove();
        $item.attr('role', 'separator');
        return;
      }
      var groupItemClass = this.getGroupItemStyleClass();
      var collapseIconClass = this.getCollapseIconStyleClass();
      var expandIconClass = this.getExpandIconStyleClass();
      var itemIconClass = this.getItemIconStyleClass();
      var groupItem = $item.children('.' + groupItemClass);
      var itemContent;
      if (groupItem.length > 0) {
        //Adding oj-navigationlist-item class on group node. Listview does't add this any more.
        groupItem.addClass(this.getItemStyleClass());
        itemContent = groupItem.children(':not(.' + expandIconClass + '):not(.' + collapseIconClass + ')');
        var groupIcon = groupItem.children('.' + expandIconClass);
        if (groupIcon.length === 0) {
          groupIcon = groupItem.children('.' + collapseIconClass);
        }
        groupIcon.attr('role', 'presentation');
        groupIcon.removeAttr('aria-labelledby');
        if ($item.hasClass('oj-disabled')) {
          groupItem.addClass('oj-disabled');
        }
      } else {
        itemContent = $item.children().first();
      }
      itemContent.addClass(this.getItemContentStyleClass());
      var icon = itemContent.wrapInner('<span class="' + this.getItemLabelStyleClass() + '"></span>').find('.' + itemIconClass); //@HTMLUpdateOK
      if (icon.length > 0) {
        icon.insertBefore(icon.parent()); // @HTMLUpdateOK

        if (icon.attr('title')) { // preserve the title to restore it after destroy.
          icon.data(this._NAVLIST_ITEM_ICON_HAS_TITLE, icon.attr('title'));
        }

        if (this.ojContext.options['display'] === 'icons') {
          this.ojContext.element.addClass(this.getIconOnlyStyleClass());
          var itemLabel = this.getItemLabel($item);
          icon.attr('aria-label', itemLabel);
          icon.attr('role', 'img');
          this._setToolTipOnIcon(icon, itemLabel);
        }

        $item.closest('ul').addClass(this.getHasIconsStyleClass());
      } else {
        itemContent.addClass(this.getHasNoIconStyleClass());
      }

      if ($item.hasClass('oj-disabled')) {
        this.getFocusItem($item).attr('aria-disabled', 'true');
      } else {
        if (groupItem.length > 0) {
          groupItem.addClass('oj-default');
        } else {
          $item.addClass('oj-default');
        }
      }
      
      if ($item.hasClass(this.getRemovableStyleClass()) && this.isTabBar()) {
        var removableLink = $('<a>');
        removableLink.addClass(this.getNavListRemoveIcon()).addClass('oj-clickable-icon-nocontext oj-component-icon')
                    .attr('aria-label',this.ojContext.getTranslatedString('removeCueText'))
                    .attr('role','presentation')
                    .attr('aria-hidden','true');
        $item.append(removableLink); // @HTMLUpdateOK
        itemContent.attr('aria-describedby', removableLink.uniqueId().attr('id'));
      }
      this.m_listHandler.ModifyListItem($item, itemContent);
      _ojNavigationListView.superclass.itemRenderComplete.apply(this, arguments);

    },
    
    /**
     * Return node for given locator
     *
     * @override
     */
    getNodeBySubId: function (locator) {
      var subId, key, item = null;

      if (locator == null) {
        return this.ojContext.element ? this.ojContext.element[0] : null;
      }
      
      item = this.m_listHandler.GetNodeBySubId(locator);
      
      if(!item) {
        subId = locator['subId'];
        if (subId === this.getItemSubIdKey()) {
          key = locator['key'];
          item = this.FindElementByKey(key);
        }
      }

      return item;
    },
    /**
     * Returns the subId locator for the given child DOM node.
     * Invoked by widget
     * @param {!Element} node - child DOM node
     * @return {Object|null} The locator for the DOM node, or <code class="prettyprint">null</code> when none is found.
     */
    getSubIdByNode: function (node) {
      
      var item, key, subId = null;
	  
      if(node != null){
		  subId = this.m_listHandler.GetSubIdByNode(node);

		  if(!subId) {
			item = this.FindItem(node);
			if (item != null && item.length > 0) {
			  key = this.GetKey(item[0]);
			  if (key != null) {
				subId = {
				  'subId': this.getItemSubIdKey(),
				  'key': key
				};
			  }
			}
		  }
	  }
      return subId;
    },
    /**
     * Returns an object with context for the given child DOM node. 
     * This will always contain the subid for the node, defined as the 'subId' property on the context object. 
     * Additional component specific information may also be included. For more details on returned objects, see context objects.
     * Invoked by widget
     *
     * @param {!Element} node the child DOM node
     * @returns {Object|null} the context for the DOM node, or null when none is found.
     */
    getContextByNode: function (node) {
      var context = _ojNavigationListView.superclass.getContextByNode.call(this, node);
      if (context && context['subId'] == 'oj-listview-item') {
        context['subId'] = this.getItemSubIdKey();
        return context;
      }
      return null;
    },
    /**
     * Refresh navlist
     * @override
     */
    refresh: function () {
      this._resetNavlist();
      this._initListHandler();
      _ojNavigationListView.superclass.refresh.apply(this, arguments);
    },
    /**
     * Uninitialize dom.
     * @private
     */
    _resetNavlist: function () {
      this.ojContext.element.removeClass(this.getIconOnlyStyleClass());
      this.ojContext.element.removeClass(this._APPLICATION_LEVEL_NAV_STYLE_CLASS);
      this.ojContext.element.removeClass(this._PAGE_LEVEL_NAV_STYLE_CLASS);
      this._restoreContent(this.element);
      this.m_listHandler.Destroy();
    },
    /**
     * Destroy the Navigation List
     * @override
     */
    destroy: function () {
      this._resetNavlist();
      _ojNavigationListView.superclass.destroy.apply(this, arguments);
      this._list.unwrap();
      //for data source, this will be empty.
      if (this._list.is(':empty')) {
        this._list.remove();
      }
      this.ojContext.element.removeClass(this.getNavListStyleClass() + ' ' + this.getNavListTouchStyleClass());
      this.ojContext._off(this.element, "click");
      this.ojContext._off(this.element, "focus");
      this.ojContext._off(this.element, "blur");
      this.ojContext._off(this.element, "mouseover");
      this.ojContext._off(this.element, "mousein");
      this.ojContext._off(this.element, "mouseout");
      this.ojContext._off(this.element, "keydown");
    }

  });

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
(function () {
  /**
   * @ojcomponent oj.ojNavigationList
   * @augments oj.baseComponent
   * @since 1.1.0
   * @ojstatus preview
   * @ojrole menu
   * @ojrole tree
   * @ojrole listbox
   * @ojrole toolbar
   * @ojtsimport ojdataprovider
   * @ojtsimport ojkeyset
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojNavigationList<K, D> extends baseComponent<ojNavigationListSettableProperties<K,D>>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojNavigationListSettableProperties<K,D> extends baseComponentSettableProperties",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojshortdesc Displays items as a collapsible or sliding navigation list with highly interactive features.
   * @classdesc
   * <h3 id="navlistOverview-section">
   *   JET Navigation List
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#navlistOverview-section"></a>
   * </h3>
   *
   * <p>Description: JET Navigation List enhances a HTML list element into a themable, WAI-ARIA compliant, mobile friendly component with advance interactive features.
   * <h3 id="data-section">
   *   Data
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#data-section"></a>
   * </h3>
   * <p>The JET NavigationList gets its data in three different ways.  The first way is from a DataProvider/TableDataSource.  There are several types of DataProvider/TableDataSource 
   * that are available out of the box:</p>
   * <ul>
   * <li>oj.ArrayDataProvider</li>
   * <li>oj.CollectionTableDataSource</li>
   * </ul>
   *
   * <p><b>oj.ArrayDataProvider</b> - Use this when the underlying data is an array object or an observableArray.  In the observableArray case, Navigation List will automatically react
   * when items are added or removed from the array.  See the documentation for oj.ArrayDataProvider for more details on the available options.</p>
   *
   * <p><b>oj.CollectionTableDataSource</b> - Use this when oj.Collection is the model for the underlying data.  Note that the Navigation List will automatically react to model event from
   * the underlying oj.Collection.  See the documentation for oj.CollectionTableDataSource for more details on the available options.</p>
   *
   * <p> NOTE: oj.PagingTableDataSource is not supported by Navigation List.
   *
   * <p>The second way is from a TreeDataSource.  This is typically used to display data that are logically categorized in groups.  There are several types
   * of TreeDataSource that are available out of the box:</p>
   * <ul>
   * <li>oj.JsonTreeDataSource</li>
   * <li>oj.CollectionTreeDataSource</li>
   * </ul>
   *
   * <p><b>oj.JsonTreeDataSource</b> - Use this when the underlying data is a JSON object.  See the documentation for oj.JsonTreeDataSource for more details on the available options.</p>
   *
   * <p><b>oj.CollectionTreeDataSource</b> - Use this when oj.Collection is the model for each group of data.  See the documentation for oj.CollectionTableDataSource
   * for more details on the available options.</p>
   *
   * <p>Finally, Navigation List also supports static HTML content as data.  The structure of the content can be either flat or hierarhical.</p>
   *
   * <p>Example of flat static content</p>
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-navigation-list>
   *  &lt;ul>
   *   &lt;li>&lt;a href="#">Item 1&lt;/a>&lt;/li>
   *   &lt;li>&lt;a href="#">Item 2&lt;/a>&lt;/li>
   *   &lt;li>&lt;a href="#">Item 3&lt;/a>&lt;/li>
   *  &lt;/ul>
   * &lt;/oj-navigation-list>
   * </code></pre>
   *
   * <p>Example of hierarchical static content</p>
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-navigation-list>
   *  &lt;ul >
   *   &lt;li>&lt;a href="#">Item 1&lt;/a>&lt;/li>
   *   &lt;li>&lt;a href="#">Item 2&lt;/a>&lt;/li>
   *   &lt;li>&lt;a href="#">Item 3&lt;/a>
   *     &lt;ul>
   *       &lt;li>&lt;a href="#">Item 3-1&lt;/a>&lt;/li>
   *       &lt;li>&lt;a href="#">Item 3-2&lt;/a>&lt;/li>
   *       &lt;li>&lt;a href="#">Item 3-3&lt;/a>&lt;/li>
   *       &lt;li>&lt;a href="#">Item 3-4&lt;/a>&lt;/li>
   *       &lt;li>&lt;a href="#">Item 3-5&lt;/a>&lt;/li>
   *     &lt;/ul>
   *   &lt;/li>
   *   &lt;li>&lt;a href="#">Item 4&lt;/a>&lt;/li>
   *   &lt;li>&lt;a href="#">Item 5&lt;/a>&lt;/li>
   *  &lt;/ul>
   * &lt;/oj-navigation-list>
   * </code></pre>
   * <h4 id="treeJSON-section"> JSON Tree Node Format</h4>
   * </br>
   * Each node object typically have an
   * <code class="prettyprint">attr</code> property. This attr value will be provided as <code class="prettyprint">itemContext.data</code> to renderer function. See <a href="#itemContext">itemContext</a> section. Any node can be defined as a parent by supplying
   * a <code class="prettyprint">children</code> property, which is an array of one or more node definitions.
   * (Note: lazy-loading is not supported in navigation list so <code class="prettyprint">children</code> property should be a non empty array)
   * <p>Example: Basic JSON Tree definition
   * <pre class="prettyprint">
   * <code>
   *[
   *   {
   *     "attr": {
   *               "id": "home",
   *               "title": "Home"
   *             },
   *   },
   *   {
   *     "attr": {
   *               "id": "news",
   *               "title": "News"
   *             }
   *   },
   *   {
   *      "attr": {
   *                "id": "blogs",
   *                "title": "Blogs"
   *              },
   *      "children": [ {
   *
   *                       "attr": {
   *                                   "id": "today",
   *                                   "title": "Today"
   *                               }
   *                    },
   *                    {
   *                       "attr": {
   *                                   "id": "yesterday",
   *                                   "title": "Yesterday"
   *                               }
   *                    }
   *                  ]
   *   }
   *]
   *</code></pre>
   *</p></br>
   * <p>Any list item can be disabled by adding the <code class="prettyprint">oj-disabled</code> class to that element.  As with any DOM change, doing so post-init
   * requires a <code class="prettyprint">refresh()</code> of the element.
   * 
   * <h3 id="key-section">
   *   Key
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#key-section"></a>
   * </h3>
   * <p>Key is an identifier which uniquely identifies an item in Navigation List. 
   *  <ul>
   *  <li>When static html is used, it will be the <code class="prettyprint">id</code> attribute of <code class="prettyprint">&lt;li></code>. If no <code class="prettyprint">id</code> is specified then component will generate an id and will use it as key.</li>
   *  <li>When data source is used, it will be the <code class="prettyprint">id</code> attribute of item's data object.
   *  </ul>
   * <h3 id="icons-section">
   *   Icons
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#icons-section"></a>
   * </h3>
   *
   * <p>Sublist icons are inserted automatically.  To add other icons to list items, include them in the markup and include the <code class="prettyprint">oj-navigationlist-item-icon</code> class, as follows:
   *
   * <pre class="prettyprint">
   * <code>&lt;oj-navigation-list>
   *   &lt;ul>    
   *     &lt;li id="foo">&lt;a href="#">&lt;span class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>Foo&lt;/a>&lt;/li>
   *   &lt;/ul>
   *  &lt;/oj-navigation-list>
   * </code></pre>
   *
   * <h3 id="styling-section">
   *   Styling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"stylingDoc"}
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   * {@ojinclude "name":"keyboardDoc"}
   *
   * <p>Disabled items will not receive keyboard focus and do not allow any interaction.
   *
   *<h3 id="context-section">
   *   Item Context
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#context-section"></a>
   * </h3>
   *
   * <p>For all item options, developers can specify a function as the return value.  The function takes a single argument, which is an object that contains contextual information about the particular item.  This gives developers the flexibility to return different value depending on the context.</p>
   *
   * <p>The context paramter contains the following keys:</p>
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Key</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td><kbd>componentElement</kbd></td>
   *       <td>oj-navigation-list element.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>datasource</kbd></td>
   *       <td>A reference to the data source object. (Not available for static content)</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>index</kbd></td>
   *       <td>The index of the item, where 0 is the index of the first item.  In the hierarchical case the index is relative to its parent.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>key</kbd></td>
   *       <td>The <a href="#key-section">Key</a> of the item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>data</kbd></td>
   *       <td>The data object for the item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>parentElement</kbd></td>
   *       <td>The list item element.  The renderer can use this to directly append content.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * <p></p>
   * <p>If the data is hierarchical, the following additional contextual information are available:</p>
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Key</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td><kbd>depth</kbd></td>
   *       <td>The depth of the item.  The depth of the first level children under the invisible root is 1.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>parentKey</kbd></td>
   *       <td>The <a href="#key-section">Key</a> of the parent item.  The parent key is null for root node.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>leaf</kbd></td>
   *       <td>Whether the item is a leaf or a group item.</td>
   *     </tr>
   *   </tbody>
   * </table>
   * <h3 id="perf-section">
   *   Performance
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
   * </h3>
   *
   * <h4>Data Set Size</h4>
   * <p>As a rule of thumb, it's recommended that applications limit the amount of data to display.  Displaying large
   * number of items in Navigation List makes it hard for user to find what they are looking for, but affects the load time.
   * If displaying large number of items is neccessary, it is recommended to group the items in to hierarchial navigation list.
   *
   * <h4>Item Content</h4>
   * <p>Navigation List allows developers to specify arbitrary content inside its item. In order to minimize any negative effect on
   * performance, you should avoid putting a large number of heavy-weight components inside because as you add more complexity
   * to the structure, the effect will be multiplied because there can be many items in the Navigation List.</p>
   *
   * <h4>Expand All</h4>
   * <p>While Navigation List provides a convenient way to initially expand all group items, it might have an impact
   * on the initial rendering performance since expanding each group item might cause a fetch from the server depending on
   * the TreeDataSource.  Other factors that could impact performance includes the depth of the tree, and the number of children
   * in each level.</p>
   *
   * <h3 id="a11y-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
   * </h3>
   * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.
   * <p>Disabled content: JET supports an accessible luminosity contrast ratio,
   * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>,
   * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which
   * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user
   * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio
   * required of enabled content, it cannot be used to convey meaningful information.<p>
   *
   *
   * <h3 id="rtl-section">
   *   Reading direction
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
   * </h3>
   *
   * <p>The only supported way to set the reading direction (LTR or RTL) is to set the <code class="prettyprint">"dir"</code> attribute on the
   * <code class="prettyprint">&lt;html></code> element of the page.  As with any JET custom element, in the unusual case that the reading direction
   * is changed post-init, the navigation list must be <code class="prettyprint">refresh()</code>ed, or the page must be reloaded.
   *
   *
   * <h3 id="animation-section">
   *   Animation
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#animation-section"></a>
   * </h3>
   *
   * <p>Applications can customize animations triggered by actions in Navigation List by either listening for <code class="prettyprint">animateStart/animateEnd</code>
   *    events or overriding action specific style classes on the animated item.  See the documentation of <a href="oj.AnimationUtils.html">oj.AnimationUtils</a> 
   *    class for details.</p>
   *
   * <p>The following are actions and their corresponding sass variables in which applications can use to customize animation effects.  
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Action</th>
   *       <th>Sass Variable</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td rowspan="2"><kbd>add</kbd></td>
   *       <td>$navigationListHorizontalAddAnimation</td>
   *       <td>When a new item is added to the oj.TableDataSource associated with Horizontal Navigation List.</td>
   *     </tr>
   *     <tr>
   *       <td>$navigationListAddAnimation</td>
   *       <td>When a new item is added to the oj.TableDataSource associated with Vertical Navigation List.</td>
   *     </tr>
   *     <tr>
   *       <td rowspan="2"><kbd>remove</kbd></td>
   *       <td>$navigationListHorizontalRemoveAnimation</td>
   *       <td>When an existing item is removed from the oj.TableDataSource associated with Horizontal Navigation List.</td>
   *     </tr>
   *     <tr>
   *       <td>$navigationListRemoveAnimation</td>
   *       <td>When an existing item is removed from the oj.TableDataSource associated with Vertical Navigation List.</td>
   *     </tr>   
   *     <tr>
   *       <td><kbd>update</kbd></td>
   *       <td>$navigationListUpdateAnimation</td>
   *       <td>When an existing item is updated in the oj.TableDataSource associated with Navigation List.</td>
   *     </tr>
   *     <tr>
   *       <td rowspan="2"><kbd>expand</kbd></td>
   *       <td>$navigationListExpandAnimation</td>
   *       <td>When user expands a group item.</td>
   *     </tr>
   *       <td>$navigationListSliderExpandAnimation</td>
   *       <td>When user expands a group item in sliding navigation list.</td>
   *     </tr>   
   *     <tr>
   *       <td rowspan="2"><kbd>collapse</kbd></td>
   *       <td>$navigationListCollapseAnimation</td>
   *       <td>When user collapses a group item.</td>
   *     </tr>
   *     <tr>
   *       <td>$navigationListSliderCollapseAnimation</td>
   *       <td>When user collapses a group item in sliding navigation list.</td>
   *     </tr>   
   *     <tr>
   *       <td><kbd>pointerUp</kbd></td>
   *       <td>$navigationListPointerUpAnimation</td>
   *       <td>When user finish pressing an item (on touch).</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   */
  oj.__registerWidget("oj.ojNavigationList", $['oj']['baseComponent'], {
    widgetEventPrefix: 'oj',
    options: {
      /**
       * <a href="#key-section">Key</a> of the current item.  Current item is the list item which is having active focus.  Note that if currentItem
       * is set to an item that is currently not available (not fetched or
       * inside a collapsed parent node), then the value is ignored.
       * <p>
       * When the current item is changed, the <code class="prettyprint">event.detail</code> of the <code class="prettyprint">currentItemChanged</code> event will contain the following additional properties:<br><br>
       * <table class="props">
       *   <thead>
       *     <tr>
       *       <th>Name</th>
       *       <th>Type</th>
       *       <th>Description</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td class="name"><code>item</code></td>
       *       <td class="type">Element</td>
       *       <td class="description">Current Item element</td>
       *     </tr>
       *   </tbody>
       * </table>  
       *
       * @expose
       * @ojshortdesc Gets and sets the key of the item that should have keyboard focus.
       * @public
       * @instance
       * @memberof! oj.ojNavigationList
       * @type {*}
       * @ojsignature {target:"Type", value:"K"}
       * @default null
       * @ojwriteback
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">current-item</code> attribute specified:</caption>
       *  &lt;oj-navigation-list current-item='settings'> ... &lt;/oj-navigation-list>
       * @example <caption>Get the current item:</caption>
       * var currentItem = myNavList.currentItem;
       *
       * @example <caption>Set the current item on the NavigationList during initialization:</caption>
       * myNavList.currentItem = "newItem";
       */
      currentItem: null,
      /**
       * Defines style of navigation used to drill down hierarchical list items.
       * <ul>
       * <li>none</li>
       * <li>collapsible</li>
       * <li>sliding</li>
       * </ul>
       * 
       * @ojshortdesc Gets and sets whether expand/collapse or sliding operations are allowed.
       * @expose
       * @memberof oj.ojNavigationList
       * @instance
       * @type {string}
       * @default none
       * @ojvalue {string} "none" All group items are expanded by default and user not allowed to collapse them.
       * @ojvalue {string} "collapsible" Allows user to expand and collapse group items. If there are more than two levels in hierarchy, <code class="prettyprint">sliding</code> is preferered drill mode.
       * @ojvalue {string} "sliding" This is typically used for hierarchical lists. This allows user to view one level at a time.
       *
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">drill-mode</code> attribute specified:</caption>
       *  &lt;oj-navigation-list drill-mode='collapsible'> ... &lt;/oj-navigation-list>
       *  
       * @example <caption>Set <code class="prettyprint">drillMode</code> property:</caption>
       * myNavList.drillMode = "collapsible";
       *
       * @example <caption>Get the <code class="prettyprint">drillMode</code> property:</caption>
       * // getter
       * var drillMode = myNavList.drillMode;
       */
      drillMode: "none",
      reorderable: "disabled",
      truncation: "none",
      /**
       * The position of the Navigation List. Valid Values: top and start.
       * <p> NOTE: when value is <code class="prettyprint">top</code>,<code class="prettyprint">"none"</code> is the only supported drillMode and it also does't support hierarchical items.
       * @ojshortdesc Gets and sets the edge position of the Navigation List.
       * @expose
       * @name edge
       * @memberof oj.ojNavigationList
       * @instance
       * @type {string}
       * @ojvalue {string} "top" This renders list items horizontally.
       * @ojvalue {string} "start" This renders list items vertically.
       * @default start
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">edge</code> attribute specified:</caption>
       *  &lt;oj-navigation-list edge='top'> ... &lt;/oj-navigation-list>
       *  
       * @example <caption>Get the edge:</caption>
       * var edge = myNavList.edge;
       *
       * @example <caption>Set the edge on the Navigation List:</caption>
       * myNavList.edge = "top";
       */
      edge: "start",
      /**
       * Specifies the level at which user can see hiearchical menu button.This is only valid when <code class="prettyprint">drillMode</code> set to <code class="prettyprint">sliding</code>.
       * Default value is 0, shows hiearchical menu always. If value is -1, then it will never be shown.
       *
       * <p>The default value for hierarchyMenuThreshold varies by theme.  Each theme can set its default by setting
       * <code class="prettyprint">$navigationListHierarchyMenuDisplayThresholdLevelOptionDefault</code> as seen in the example below.
       *
       * @ojshortdesc Gets and sets the level at which user can see hierarchical menu button (defaults to 0 if not specified in theme).
       * @expose
       * @name hierarchyMenuThreshold
       * @memberof oj.ojNavigationList
       * @instance
       * @type {number}
       * @default 0
       * 
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">hierarchy-menu-threshold</code> attribute specified:</caption>
       *  &lt;oj-navigation-list hierarchy-menu-threshold='4'> ... &lt;/oj-navigation-list>
       *  
       * @example <caption>Get or set the <code class="prettyprint">hierarchyMenuThreshold</code> property:</caption>
       * // getter
       * var hierarchyMenuThreshold = myNavList.hierarchyMenuThreshold;
       *
       * // setter
       * myNavList.hierarchyMenuThreshold = "4";
       *
       * @example <caption>Set the default in the theme (SCSS) :</caption>
       * $navigationListHierarchyMenuDisplayThresholdLevelOptionDefault: -1 !default;
       */
      hierarchyMenuDisplayThresholdLevel: 0,
      /**
       * Label for top level list items.
       * <p>NOTE: This is needed only for sliding navigation list where
       * this will be used as title for the top level list elements.
       * 
       * @ojshortdesc Gets and sets the label for top level list items.
       * @type {?string}
       * @default Navigation List
       * @expose
       * @instance
       * @memberof oj.ojNavigationList
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">root-label</code> attribute specified:</caption>
       *  &lt;oj-navigation-list root-label='Global Navigation'> ... &lt;/oj-navigation-list>
       *  
       * @example <caption>Get or set the <code class="prettyprint">rootLabel</code> property:</caption>
       * // getter
       * var rootLabel = myNavList.rootLabel;
       *
       * // setter
       * myNavList.rootLabel = "4";
       */
      rootLabel: null,
      /**
       * Item <a href="#key-section">Key</a> of currently selected list item. If the value is modified
       * by an application, navigation list UI is modified to match the new value and the <code class="prettyprint">event.detail</code> of the <code class="prettyprint">selectionChanged</code> event will contain the following additional properties:<br><br>
       * <table class="props">
       *   <thead>
       *     <tr>
       *       <th>Name</th>
       *       <th>Type</th>
       *       <th>Description</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td class="name"><code>item</code></td>
       *       <td class="type">Element</td>
       *       <td class="description">Selected Item element</td>
       *     </tr>
       *   </tbody>
       * </table>  
       * @ojshortdesc Gets and sets the key of the selected item.
       * @type {*}
       * @ojsignature {target:"Type", value:"K"}
       * @default null
       * @expose
       * @instance
       * @memberof oj.ojNavigationList
       * @ojwriteback
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">selection</code> attribute specified:</caption>
       *  &lt;oj-navigation-list selection='settings'> ... &lt;/oj-navigation-list>
       *  
       * @example <caption>Get the selection:</caption>
       * var selection = myNavList.selection;
       *
       * @example <caption>Set the selection on the Navigation List:</caption>
       * myNavList.selection = "settings";
       */
      selection: null,
      
     /**
      * Specifies the key set containing the keys of the items that should be expanded. 
      *
      * Use the <a href="ExpandedKeySet.html">ExpandedKeySet</a> class to specify items to expand.
      * Use the <a href="ExpandAllKeySet.html">ExpandAllKeySet</a> class to expand all items.                 
      *
      * @expose
      * @memberof! oj.ojNavigationList
      * @instance
      * @default new ExpandedKeySet();
      * @type {KeySet}
      * @ojsignature {target:"Type", value:"KeySet<K>"}
      * @ojwriteback
      *
      * @example <caption>Initialize the NavigationList with specific items expanded:</caption>
      * myNavList.expanded = new ExpandedKeySet(['item1', 'item2']);
      *
      * @example <caption>Initialize the NavigationList with all items expanded:</caption>
      * myNavList.expanded = new ExpandAllKeySet();
      */
      expanded: new oj._ojListViewExpandedKeySet(),
      /**
       * The data source for the NavigationList accepts either a oj.TableDataSource or oj.TreeDataSource.
       * See the data source section in the introduction for out of the box data source types.
       * If the data attribute is not specified, the child elements are used as content.  If there's no
       * content specified, then an empty list is rendered.
       *
       * @ojshortdesc Gets and sets the data provider for Navigation List.
       * @expose
       * @memberof! oj.ojNavigationList
       * @instance
       * @type {oj.TableDataSource|oj.TreeDataSource|oj.DataProvider|null}
       * @default null
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">data</code> attribute specified:</caption>
       *  &lt;oj-navigation-list data='[[tableDataSource]]'> ... &lt;/oj-navigation-list>
       * @example <caption>Get the data:</caption>
       * var dataSource = myNavList.data;
       *
       * @example <caption>Set the data attribute using one-dimensional array:</caption>
       * myNavList.data = new oj.ArrayTableDataSource([1,2,3]);
       * 
       * @example <caption>Set the data attribute using oj.Collection:</caption>
       * myNavList.data = new oj.CollectionTableDataSource(collection);
	     * @ojsignature {target: "Type", value: "oj.DataProvider<K, D>|null"}
       */
      data: null,
      /**
       * Whether to display both the label and icons (<code class="prettyprint">"all"</code>) or just the icons (<code class="prettyprint">"icons"</code>).
       * In the latter case, the label is displayed in a tooltip instead, unless a tooltip was already supplied at create time.
       * Note: <code class="prettyprint">display="icons"</code> is valid only when <code class="prettyprint">drillMode=none</code> and navigation list is a flat list.
       * It is also mandatory to provide icons for each item as stated in <a href="#icons-section">icons section</a>.
       *
       * @ojshortdesc Gets and sets whether all or only icons need to be displayed.
       * @expose
       * @memberof oj.ojNavigationList
       * @instance
       * @type {string}
       * @ojvalue {string} "all" Display both the label and icons.
       * @ojvalue {string} "icons" Display only the icons.
       * @default all
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">display</code> attribute specified:</caption>
       *  &lt;oj-navigation-list display='icons'> ... &lt;/oj-navigation-list>
       * 
       * @example <caption>Get or set the <code class="prettyprint">display</code> property:</caption>
       * // getter
       * var display = myNavList.display;
       *
       * // setter
       * myNavList.display = "icons";
       */
      display: "all",
      /**
       * Navigation List supports two different look-and-feel depends on its usage level, one is <code class="prettyprint">application</code> level and other is <code class="prettyprint">page</code> level.
       *
       * @expose
       * @ignore
       * @deprecated 4.0.0 Use oj-tab-bar custom element for page level navigation list. In future ojNavigationList will only support application level by default.
       * @memberof oj.ojNavigationList
       * @instance
       * @type {string}
       * @ojvalue {string} "application" Render Navigation List for application level navigation.
       * @ojvalue {string} "page" Render Navigation List for page level navigation.
       * @default page
       *
       * @example <caption>Initialize the NavigationList with the <code class="prettyprint">navigationLevel</code> option specified:</caption>
       *  &lt;oj-navigation-list navigation-level='application'> ... &lt;/oj-navigation-list>
       *
       * @example <caption>Get or set the <code class="prettyprint">navigationLevel</code> option, after initialization:</caption>
       * // getter
       * var display = navList.navigationLevel;
       *
       * // setter
       * navList.navigationLevel = "application";
       */
      navigationLevel: "page",
       /**
       * Specifies the overflow behaviour. 
       * NOTE: This is only applicable when <code class="prettyprint">edge</code> attribute set to <code class="prettyprint">top</code>
       * 
       * @ojshortdesc Gets and sets overflow behaviour for Navigation List.
       * @expose
       * @memberof oj.ojNavigationList
       * @instance
       * @type {string}
       * @ojvalue {string} "popup" popup menu will be shown with overflowed items.<p> NOTE: Setting <code class="prettyprint">overflow</code> to <code class="prettyprint">popup</code> can trigger browser reflow, so only set it when it is actually required.
       * @ojvalue {string} "hidden" overflow is clipped, and the rest of the content will be invisible.
       * @default hidden
       * @since 3.0.0
       * @example <caption>Initialize the Navigation List with the <code class="prettyprint">overflow</code> attribute specified:</caption>
       *  &lt;oj-navigation-list overflow='popup'> ... &lt;/oj-navigation-list>
       * @example <caption>Get or set the <code class="prettyprint">overflow</code> property:</caption>
       * // getter
       * var overflow = myNavList.overflow;
       *
       * // setter
       * myNavList.overflow = "hidden";
       */
      overflow: "hidden",
      
      /**
       * @typedef {Object} oj.ojNavigationList.ItemContext
       * @property {Element} componentElement oj-navigation-list element
       * @property {oj.DataProvider<K, D>} [datasource] A reference to the data source object. (Not available for static content)
       * @property {number} index The index of the item, where 0 is the index of the first item.
       * @property {any} key The Key of the item.
       * @property {any} data The data object for the item.
       * @property {Element} parentElement The list item element. The renderer can use this to directly append content.
       * @ojsignature [{target:"Type", value:"<K,D>", for:"genericTypeParameters"}]
       */
      
      /**
       * The item property contains a subset of properties for items.
       *
       * @ojshortdesc Customize the functionalities of each item in Navigation List.  
       * @expose
       * @memberof oj.ojNavigationList
       * @instance
       * @name item
       * @type {Object}
       */
      item: {
        /**
         * The renderer function that renders the content of the item. See <a href="#context-section">itemContext</a>
         * in the introduction to see the object passed into the renderer function.
         * The function should return one of the following: 
         * <ul>
         *   <li>An Object with the following property:
         *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the item.</li></ul>
         *   </li>
         *   <li>undefined: If the developer chooses to manipulate the item element directly, the function should return undefined.</li>
         * </ul>
         * If no renderer is specified, Navigation List will treat the data as a String.
         *
         * @ojshortdesc Gets and sets the renderer for the item.
         * @expose
         * @alias item.renderer
         * @memberof! oj.ojNavigationList
         * @instance
         * @type {function(Object)|null}
         * @ojsignature { target: "Type",
         *                value: "?(((context: oj.ojNavigationList.ItemContext<K,D>) => void)|null)",
         *                jsdocOverride: true}
         * @default null
         * @example <caption>Initialize the Navigation List with the <code class="prettyprint">item.renderer</code> attribute specified:</caption>
         *  &lt;oj-navigation-list item.renderer="{{oj.KnockoutTemplateUtils.getRenderer('template', true)}}"> ... &lt;/oj-navigation-list>
         *  
         * @example <caption>Get or set the <code class="prettyprint">renderer</code> property:</caption>
         * // set the renderer function
         * myNavList.item.renderer = rendererFn;
         * 
         * // get the renderer function
         * var rendererFn = myNavList.item.renderer;
         */
        renderer: null,
        /**
         * Whether the item is selectable.
         * See <a href="#context-section">itemContext</a> in the introduction to see the object passed into the selectable function.
         *
         * @ojshortdesc Gets and sets whether the item can be selected.
         * @expose
         * @alias item.selectable
         * @memberof! oj.ojNavigationList
         * @instance
         * @ojsignature { target: "Type",
         *                value: "?(((context: oj.ojNavigationList.ItemContext<K,D>) => boolean)|boolean)",
         *                jsdocOverride: true}
         * @type {function(Object)|boolean}
         * @default true
         *
         * @example <caption>Initialize the NavigationList such that the first 3 items are not selectable:</caption>
         *  <oj-navigation-list item.selectable="[[skipFirstThreeElementsFn]]"></oj-navigation-list>
         *  
         *  var skipFirstThreeElementsFn = function(itemContext) {
         *                                      return itemContext['index'] > 3; 
         *                                 }
         */
        selectable: true

      },
      //Events
      /**
       * Triggered when the default animation of a particular action is about to start.  The default animation can be cancelled by calling event.preventDefault.
       *
       * @ojshortdesc Event handler for when the default animation of a particular action is about to start.
       * @expose
       * @event
       * @memberof oj.ojNavigationList
       * @instance
       * @property {string} action the action that starts the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
       * @property {Element} element the target of animation.  
       * @property {function():void} endCallback if the event listener calls event.preventDefault to cancel the default animation, it must call the endCallback function when it finishes its own animation handling and when any custom animation ends.
       */
      animateStart: null,
      /**
       * Triggered when the default animation of a particular action has ended.  Note this event will not be triggered if application cancelled the default animation on animateStart.
       *
       * @ojshortdesc Event handler for when the default animation of a particular action has ended.
       * @expose
       * @event
       * @memberof oj.ojNavigationList
       * @instance
       * @property {string} action the action that started the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
       * @property {Element} element the target of animation.  
       */
      animateEnd: null,
      /**
       * <p>Triggered before this list item is selected.
       * To prevent the item selection, invoke <code class="prettyprint">event.preventDefault()</code>.
       *
       * @ojshortdesc Event handler for when before the selection is changed.
       * @expose
       * @event
       * @ojcancelable
       * @memberof oj.ojNavigationList
       * @instance
       * @property {*} key Selected list item <a href="#key-section">Key</a>.
       * @property {Element} item Selected list item.
       */
      beforeSelect: null,
      /**
       * Triggered before an item is collapse via the <code class="prettyprint">expanded</code> property,
       * the <code class="prettyprint">collapse</code> method, or via the UI.
       * To prevent the item being collapsed, invoke <code class="prettyprint">event.preventDefault()</code>.
       *
       * @ojshortdesc Event handler for when an item is about to collapse.
       * @expose
       * @event
       * @memberof oj.ojNavigationList
       * @instance
       * @ojcancelable
       * @property {*} key the <a href="#key-section">Key</a> of the item to be collapse
       * @property {Element} item the item to be collapse
       */
      beforeCollapse: null,
      /**
       * Triggered before the current item is changed via the <code class="prettyprint">currentItem</code> property or via the UI.
       * To prevent the item being focused, invoke <code class="prettyprint">event.preventDefault()</code>.
       * 
       * @ojshortdesc Event handler for when before the current item is changed.
       * @expose
       * @event
       * @memberof oj.ojNavigationList
       * @instance
       * @ojcancelable
       * @property {*} previousKey the <a href="#key-section">Key</a> of the previous item
       * @property {Element} previousItem the previous item
       * @property {*} key the <a href="#key-section">Key</a> of the new current item
       * @property {Element} item the new current item
       */
      beforeCurrentItem: null,
      beforeDeselect: null,
      /**
       * Triggered after an item has been collapsed via the <code class="prettyprint">expanded</code> property,
       * the <code class="prettyprint">collapse</code> method, or via the UI.
       * 
       * @ojshortdesc Event handler for after an item has collapsed.
       * @expose
       * @event
       * @memberof oj.ojNavigationList
       * @instance
       * @property {*} key The <a href="#key-section">Key</a> of the item that was just collapsed.
       * @property {Element} item The list item that was just collapsed.
       */
      collapse: null,
      deselect: null,
      /**
       * Triggered before an item is expand via the <code class="prettyprint">expanded</code> property,
       * the <code class="prettyprint">expand</code> method, or via the UI.
       * To prevent the item being expanded, invoke <code class="prettyprint">event.preventDefault()</code>.
       * 
       * @ojshortdesc Event handler for when an item is about to expand.
       * @expose
       * @event
       * @memberof oj.ojNavigationList
       * @instance
       * @ojcancelable
       * @property {*} key the <a href="#key-section">Key</a> of the item to be expand
       * @property {Element} item the item to be expand
       */
      beforeExpand: null,
      /**
       * Triggered after an item has been expanded via the <code class="prettyprint">expanded</code> property,
       * the <code class="prettyprint">expand</code> method, or via the UI.
       *
       * @ojshortdesc Event handler for after an item has expanded.
       * @expose
       * @event
       * @memberof oj.ojNavigationList
       * @instance
       * @property {*} key The <a href="#key-section">Key</a> of the item that was just expanded.
       * @property {Element} item The list item that was just expanded.
       */
      expand: null,
      reorder: null
    },
    /**
     * Create the Navigation List
     * @override
     * @memberof! oj.ojNavigationList
     * @protected
     */
    _ComponentCreate: function () {
      this._super();
      this._setup();
    },
    /**
     * Initialize the NavigationList after creation
     * @protected
     * @override
     * @memberof! oj.ojNavigationList
     */
    _AfterCreate: function () {
      this._super();
      
      // inject helper functions for ContentHandler and custom renderer to use
      var self = this;
      this.navlist._FixRendererContext = function(context) { return self._FixRendererContext(context); };
      this.navlist._WrapCustomElementRenderer = function(renderer) { return self._WrapCustomElementRenderer(renderer) };
      this.navlist.afterCreate();
    },
    /**
     * When the <a href="#contextMenu">contextMenu</a> option is set, this method is called when the user invokes the context menu via
     * the default gestures: right-click, pressHold, and <kbd>Shift-F10</kbd>.  Components should not call this method directly.
     *
     * @param {!Object} menu The JET Menu to open as a context menu.  Always non-<code class="prettyprint">null</code>.
     * @param {!Event} event What triggered the menu launch.  Always non-<code class="prettyprint">null</code>.
     * @param {string} eventType "mouse", "touch", or "keyboard".  Never <code class="prettyprint">null</code>.
     * @private
     */
    _NotifyContextMenuGesture: function (menu, event, eventType) {
      this.navlist.notifyContextMenuGesture(menu, event, eventType);
    },
    /**
     * Initialize DOM structure for navigation list.
     * @private
     * @memberof! oj.ojNavigationList
     */
    _setup: function () {
      var key, opts;
      this.navlist = new _ojNavigationListView();

      for (key in this.options) {
        this._validateOptionValues(key, this.options[key]);
      }
      this._validateOptionsForIconsOnlyAndHorizontalList(this.options[this.navlist.OPTION_DRILL_MODE], this.options[this.navlist.OPTION_DISPLAY], this.options[this.navlist.OPTION_EDGE]);
      
      opts = {};
      opts.ojContext = this;
      opts = $.extend(this.options, opts);
      this.navlist.init(opts);
    },

    getNodeBySubId: function (locator) {
      return this.navlist.getNodeBySubId(locator);
    },
    getSubIdByNode: function (node) {
      return this.navlist.getSubIdByNode(node);
    },
    
    /**
     * @typedef {Object} oj.ojNavigationList.NodeContext
     * @property {string} subId Sub-id string to identify a particular dom node.
     * @property {number} index The index of the item, where 0 is the index of the first item.
     * @property {K} key The Key of the item.
     * @property {boolean} group whether the item is a group.
     * @property {Element} [parent] the parent group item. Only available if item has a parent.
     * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"}]
     */
    
    /**
     * {@ojinclude "name":"nodeContextDoc"}
     * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
     * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
     * @ojsignature { target: "Type", for:"returns",
     *                value: "oj.ojNavigationList.NodeContext<K>|null",
     *                jsdocOverride: true}
     * @example {@ojinclude "name":"nodeContextExample"}
     *
     * @expose
     * @instance
     * @memberof oj.ojNavigationList
     */
    getContextByNode: function (node) {
      return this.navlist.getContextByNode(node);
    },
    /**
     * Expand an item.<p>
     * Note when vetoable is set to false, beforeExpand event will still be fired but the event cannot be veto.<p>
     *
     * @ignore
     * @expose
     * @memberof oj.ojNavigationList
     * @instance
     * @param {Object|string} key the <a href="#key-section">Key</a> of the item to expand
     * @param {boolean} vetoable if event is vetoable
     * @param {boolean} animate true if animate the expand operation, false otherwise
     * @export
     */
    expand: function (key, vetoable, animate) {
      this.navlist.expandKey(key, vetoable, true, true, animate);
    },
    /**
     * Collapse an item.<p>
     * Note when vetoable is set to false, beforeCollapse event will still be fired but the event cannot be veto.<p>
     *
     * @ignore
     * @expose
     * @memberof oj.ojNavigationList
     * @instance
     * @param {Object|string} key the <a href="#key-section">Key</a> of the item to collapse
     * @param {boolean} vetoable if event is vetoable
     * @param {boolean} animate true if animate the collapse operation, false otherwise
     * @export
     */
    collapse: function (key, vetoable, animate) {
      this.navlist.collapseKey(key, vetoable, true, animate);
    },
    /**
     * Gets the key of currently expanded items.
     *
     * @ignore
     * @expose
     * @memberof oj.ojNavigationList
     * @instance
     * @return {Array} array of keys of currently expanded items
     * @export
     */
    getExpanded: function () {
      return this.navlist.getExpanded();
    },
    _validateOptionsForIconsOnlyAndHorizontalList: function (drillMode, display, edge) {
      if (drillMode !== this.navlist.OPTION_DRILL_MODE_NONE) {
        if (display === this.navlist.OPTION_DISPLAY_ICONS) {
          throw "Icon only navigation list should have drillMode set to 'none'.";
        }
        if (edge === this.navlist.OPTION_EDGE_TOP) {
          throw "Horizontal navigation list should have drillMode set to 'none'.";
        }
      }
    },
    _validateOptionValues: function (key, value) {
      var valid = true;
      if (key === this.navlist.OPTION_DRILL_MODE) {
        valid = (value === this.navlist.OPTION_DRILL_MODE_NONE || value === this.navlist.OPTION_DRILL_MODE_COLLAPSIBLE || value === this.navlist.OPTION_DRILL_MODE_SLIDING);
      } else if (key === this.navlist.OPTION_DISPLAY) {
        valid = (value === this.navlist.OPTION_DISPLAY_ALL || value === this.navlist.OPTION_DISPLAY_ICONS);
      } else if (key === this.navlist.OPTION_EDGE) {
        if(this.element[0].tagName.toLowerCase() === this.navlist.TAG_NAME_TAB_BAR){
          valid = (value === this.navlist.OPTION_EDGE_TOP || value === this.navlist.OPTION_EDGE_START ||
                  value === this.navlist.OPTION_EDGE_END || value === this.navlist.OPTION_EDGE_BOTTOM);
        } else {
          valid = (value === this.navlist.OPTION_EDGE_TOP || value === this.navlist.OPTION_EDGE_START);
        }
      }
      if (!valid) {
        throw "Invalid value: " + value + " for key: " + key;
      }
    },
    /**
     * Set single option
     * @param {string} key the option key
     * @param {Object} value value for option
     * @override
     * @memberof oj.ojNavigationList
     * @instance
     * @private
     */
    _setOption: function (key, value) {
      var flags, extraData;
      this._validateOptionValues(key, value);
      switch (key) {
      case this.navlist.OPTION_DRILL_MODE:
        this._validateOptionsForIconsOnlyAndHorizontalList(value, this.options[this.navlist.OPTION_DISPLAY], this.options[this.navlist.OPTION_EDGE]);
        break;
      case this.navlist.OPTION_DISPLAY:
        this._validateOptionsForIconsOnlyAndHorizontalList(this.options[this.navlist.OPTION_DRILL_MODE], value, this.options[this.navlist.OPTION_EDGE]);
        break;
      case this.navlist.OPTION_EDGE:
        this._validateOptionsForIconsOnlyAndHorizontalList(this.options[this.navlist.OPTION_DRILL_MODE], this.options[this.navlist.OPTION_DISPLAY], value);
        break;
      case this.navlist.OPTION_SELECTION:
        extraData = this.navlist.getItems([value])[0];
        flags = {'_context': {extraData: {'item': this._IsCustomElement() ? extraData : $(extraData)}}};
        break;
      case this.navlist.OPTION_CURRENT_ITEM:
        extraData = this.navlist.getItems([value])[0];
        flags = {'_context': {extraData: {'item': this._IsCustomElement() ? extraData : $(extraData)}}};
        break;  
      }
      if (flags)
      {
        return this._super(key, value, flags);
      }
      return this._super(key, value);
    },
    /**
     * Sets multiple options
     * @param {Object} options the options object
     * @param {Object} flags additional flags for option
     * @override
     * @private
     */
    _setOptions: function (options, flags) {
      var key, result = this.navlist.setOptions(options, flags);

      var nonSkippedOptions = {};
      for (key in options) {
        if (result.skipOptions.indexOf(key) < 0) {
            nonSkippedOptions[key] = options[key];
        }
      }
      // Should call _super as that's where the old and new option value checking
      // logic lives. _setOption should not be called directly.
      this._super(nonSkippedOptions, flags);

      if (result.needRefresh) {
        this.navlist.refresh();
      }
      return this;
    },
    
    /**
    * Invoked when application calls oj.Components.subtreeAttached.
    * @override
    * @private
    */
    _NotifyAttached: function()
    {
        this.navlist.notifyAttached();
    },

    /**
     * In browsers [Chrome v35, Firefox v24.5, IE9, Safari v6.1.4], blur and mouseleave events are generated for hidden content but not detached content,
     * so for detached content only, we must use this hook to remove the focus and hover classes.
     * @override
     * @private
     */
    _NotifyDetached: function () {
      this.navlist.notifyDetached();
    },
    /**
     * Invoked when application calls oj.Components.subtreeShown.
     * @override
     * @private
     */
    _NotifyShown: function () {
      this.navlist.notifyShown();
    },
    /**
     * Refreshes the visual state of the Navigation List. JET components require a <code class="prettyprint">refresh()</code> after the DOM is
     * programmatically changed underneath the component.
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof! oj.ojNavigationList
     * @instance
     * @return {void}
     *
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     *  myNavList.refresh();
     */
    refresh: function () {
      this._super();
      this.navlist.refresh();

    },
    /**
     * Returns a Promise that resolves when the component is ready, i.e. after data fetching, rendering, and animations complete. 
     * Note that in the highwatermark scrolling case, component is ready after data fetching, rendering, and associated animations of items fetched so far are complete.
     *
     * <p>This method does not accept any arguments.
     * 
     * @expose
     * @ignore
     * @memberof oj.ojNavigationList
     * @instance
     * @return {Promise} A Promise that resolves when the component is ready.
     */
    whenReady: function () {
      return this.navlist.whenReady();
    },
     /**
     * Sets up resources needed by navlist
     * @memberof! oj.ojNavigationList
     * @instance
     * @override
     * @protected
     */
    _SetupResources: function()
    {
        this._super();
        this.navlist.setupResources();
    },
    /**
     * Release resources held by navlist
     * @memberof! oj.ojNavigationList
     * @instance
     * @override
     * @protected
     */
    _ReleaseResources: function()
    {
        this._super();
        this.navlist.releaseResources();
    },
    /**
     * Destroy the Navigation List
     * @memberof! oj.ojNavigationList
     * @private
     */
    _destroy: function () {
      this.navlist.destroy();
      this._super();
    },

    //** @inheritdoc */
    _CompareOptionValues: function(option, value1, value2)
    {
      switch(option)
      {
         case 'currentItem':
         case 'selection':
            return oj.Object.compareValues(value1, value2);
         default:
            return this._super(option, value1, value2);
      }
    }

    // Fragments:

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan="2">List Item</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Selects the item.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Press & Hold</kbd></td>
     *       <td>Display context menu</td>
     *     </tr>
     *     <tr>
     *       <td rowspan="2">Group Item</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Expand or collapse the group item if <code class="prettyprint">drillMode</code> is set to collapsible.
     *       When <code class="prettyprint">drillMode</code> is set to sliding, sublist will silde in.
     *       When <code class="prettyprint">drillMode</code> is set to none, group item will be selecetd.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Press & Hold</kbd></td>
     *       <td>Display context menu</td>
     *     </tr>
     *     <tr>
     *       <td>Hierarchical Menu button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Open menu. Refer <a href="oj.ojButton.html#touch-section">menu button</a> touch documentation. Note: This is applicable only for Sliding Navigation List. </td>
     *     </tr>
     *     <tr>
     *       <td>Overflow Menu button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Open menu. Refer <a href="oj.ojButton.html#touch-section">menu button</a> touch documentation. Note: This is applicable only for Horizontal Navigation List when <code class="prettyprint">overflow</code> is set to <code class="prettyprint">popup</code>. </td>
     *     </tr>
     *     <tr>
     *       <td>Previous Icon or List Header</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Collapses the sublist and slides to parent list. Note: This is applicable only for Sliding Navigation List. </td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojNavigationList
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan="10">List Item</td>
     *       <td><kbd>Enter</kbd> or <kbd>Space</kbd></td>
     *       <td>Selects list item.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>Moves focus to the previous visible list item.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Moves focus to the next  visible list item</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>RightArrow</kbd> (<kbd>LeftArrow</kbd> in RTL)</td>
     *       <td>For horizontal navigation list,focus will be moved to next visible item.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td><kbd>LeftArrow</kbd> (<kbd>RightArrow</kbd> in RTL)</td>
     *       <td>For horizontal navigation list,focus will be moved to previous visible item.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Home</kbd></td>
     *       <td>Moves focus to the first visible list item.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>End</kbd></td>
     *       <td>Moves focus to the last visible list item.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>F2</kbd></td>
     *       <td>If focus is on a list item, pressing F2 will make its contents accessible using TAB.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Esc</kbd></td>
     *       <td>When F2 mode is enabled, press Esc to exit F2 mode.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift+Tab</kbd></td>
     *       <td>Move fous to hierarchical menu button.Only applicable for sliding navigation list and when hierarchial menu button is enabled.</td>
     *     </tr>
     *     <tr>
     *       <td rowspan="2">Group Item</td>
     *       <td><kbd>RightArrow</kbd> (<kbd>LeftArrow</kbd> in RTL)</td>
     *       <td>If focus is on collapsed node, expands the sub list.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>LeftArrow</kbd> (<kbd>RightArrow</kbd> in RTL)</td>
     *       <td>If focus is on expanded node, collapses the sub list.</td>
     *     </tr>
     *     <tr>
     *       <td>List Item in sublist</td>
     *       <td><kbd>Esc</kbd></td>
     *       <td>Applicable only for sliding navigation list. If focus is in a sub list, closes the sublist and moves focus to the parent list item.</td>
     *     </tr>
     *     <tr>
     *       <td rowspan="3">Hierarchical Menu button</td>
     *       <td><kbd>Enter</kbd></td>
     *       <td>Open menu. Refer <a href="oj.ojButton.html#keyboard-section">menu button</a> keyboard documentation. Note: This target is visible only for Sliding Navigation List.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Tab</kbd></td>
     *       <td>Moves focus to current list item. Note: This target is visible only for Sliding Navigation List. </td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Shift + Tab</kbd></td>
     *       <td>Moves focus to Previous Icon. Note: This target is visible only for Sliding Navigation List.</td>
     *     </tr>
     *     <tr>
     *       <td>Overflow Menu button</td>
     *       <td><kbd>Enter or Space</kbd></td>
     *       <td>Open menu. Refer <a href="oj.ojButton.html#touch-section">menu button</a> touch documentation. Note: This is applicable only for Horizontal Navigation List when <code class="prettyprint">overflow</code> is set to <code class="prettyprint">popup</code>. </td>
     *     </tr>     
     *     <tr>
     *       <td rowspan="2">Previous Icon or List Header</td>
     *       <td><kbd>Enter</kbd></td>
     *       <td>Collapses the sublist and slides to parent list.Note: This target is visible only for Sliding Navigation List. </td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Tab</kbd></td>
     *       <td>Moves focus to Hierarchical Menu button. Note: This target is visible only for Sliding Navigation List. </td>
     *     </tr>
     *
     *   </tbody>
     * </table>
     *
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojNavigationList
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocExampleHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-navigationlist-stack-icon-label</td>
     *       <td>Displays horizontal Navigation List with icon and label stacked. Applicable only when <code class="prettyprint">edge</code> is <code class="prettyprint">top</code>.</td>
     *       <td>
     *          <pre class="prettyprint">
     *<code>&lt;oj-navigation-list class="oj-navigationlist-stack-icon-label" >
     *  &lt;ul>    
     *    &lt;li id="foo">
     *      &lt;a href="#">
     *        &lt;span class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">
     *        &lt;/span>
     *        Foo
     *      &lt;/a>
     *    &lt;/li>
     *  &lt;/ul>
     *&lt;/oj-navigation-list>
     *</code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-navigationlist-category-divider</td>
     *       <td>Use this class to add horizontal divider line between two categories of items.</td>
     *       <td>
     *          <pre class="prettyprint">
     *<code>&lt;oj-navigation-list>
     *  &lt;ul>
     *    &lt;li ...> &lt;/li>
     *    &lt;li class="oj-navigationlist-category-divider"> &lt;/li>    
     *    &lt;li id="foo">
     *      &lt;a href="#">
     *        &lt;span class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">
     *        &lt;/span>
     *        Foo
     *      &lt;/a>
     *    &lt;/li>
     *    &lt;li ...> &lt;/li>
     *  &lt;/ul>
     *&lt;/oj-navigation-list>
     *</code></pre>
     *       </td>     
     *     </tr>
     *     <tr>
     *       <td>oj-navigationlist-item-icon</td>
     *       <td>Use this class to add icon to list item.</td>
     *       <td>
     *          <pre class="prettyprint">
     *<code>&lt;oj-navigation-list>
     *  &lt;ul>    
     *    &lt;li id="foo">
     *      &lt;a href="#">
     *        &lt;span class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">
     *        &lt;/span>
     *        Foo
     *      &lt;/a>
     *    &lt;/li>
     *  &lt;/ul>
     *&lt;/oj-navigation-list></code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-navigationlist-item-title</td>
     *       <td>When arbitrary content is placed inside item's content area, it's title text can be marked using this style class. This helps component in identifying the Item's label.
     *       </td>
     *       <td>
     * <pre class="prettyprint">
     * <code>&lt;li>
     *   &lt;div>
     *     &lt;span class="oj-navigationlist-item-title">Play&lt;/span>
     *     &lt;button>Button&lt;/button>
     *   &lt;/div>
     * &lt;/li>
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-navigationlist-item-text-wrap</td>
     *       <td>Use this class to wrap item label text. Note: On IE11, this is not supported when <code class="prettyprint">overflow</code> attribute is set to <code class="prettyprint">popup</code>.</td>
     *       <td>
     *          <pre class="prettyprint">
     *<code>&lt;oj-navigation-list class="oj-navigationlist-item-text-wrap" >
     *  &lt;ul>    
     *    &lt;li id="foo">
     *      &lt;a href="#">
     *        &lt;span class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">
     *        &lt;/span>
     *        Foo
     *      &lt;/a>
     *    &lt;/li>
     *  &lt;/ul>
     *&lt;/oj-navigation-list>
     *</code></pre>
     *       </td>     
     *     </tr>
     *     <tr>
     *       <td>oj-navigationlist-item-dividers</td>
     *       <td>Use this class to show dividers between horizontal navigation list items.</td>
     *       <td>
     *          <pre class="prettyprint">
     *<code>&lt;oj-navigation-list class="oj-navigationlist-item-dividers" >
     *  &lt;ul>    
     *    &lt;li id="foo">
     *       &lt;a href="#">
     *         &lt;span class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
     *         Foo
     *       &lt;/a>
     *    &lt;/li>
     *  &lt;/ul>
     *&lt;/oj-navigation-list>
     *</code></pre>
     *       </td>     
     *     </tr>     
     *     </tr>
     *     <tr>
     *       <td>oj-sm-condense</td>
     *       <td>Use this class to condense horizontal navigation list items on small screens and larger.
     *       </td>
     *       <td rowspan="4">
     *          <pre class="prettyprint">
     *<code>&lt;oj-navigation-list class="oj-sm-condense" >
     *  &lt;ul>    
     *    &lt;li id="foo">
     *      &lt;a href="#">
     *        &lt;span class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
     *        Foo
     *      &lt;/a>
     *    &lt;/li>
     *  &lt;/ul>
     *&lt;/oj-navigation-list>
     *</code></pre>
     *       </td>      
     *     </tr>
     *     <tr>
     *       <td>oj-md-condense</td>
     *       <td>Use this class to condense horizontal navigation list items on medium screens and larger.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-lg-condense</td>
     *       <td>Use this class to condense horizontal navigation list items on large screens and larger.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-xl-condense</td>
     *       <td>Use this class to condense horizontal navigation list items on extra large screens and larger.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-navigationlist-nofollow-link</td>
     *       <td> Use this class to prevent automatic navigation to the url specified on <code class="prettyprint">&lt;a></code> tag's <code class="prettyprint">href</code> attribute. In this case, navigation can be handled programmatically by using <code class="prettyprint">selectionChanged</code> event. This is useful to execute some custom logic before browser triggers navigation.
     *       </td>
     *       <td>
     *          <pre class="prettyprint">
     *<code>&lt;oj-navigation-list class="oj-navigationlist-nofollow-link" >
     *  &lt;ul>    
     *    &lt;li id="foo">&lt;a href="folder/foo.html">
     *      &lt;span 
     *        class="oj-navigationlist-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
     *      Foo&lt;/a>
     *    &lt;/li>
     *  &lt;/ul>
     *&lt;/oj-navigation-list>
     *</code></pre>
     *       </td>     
     *     </tr>
     *     <tr>
     *       <td>oj-focus-highlight</td>
     *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
     *       <td></td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojNavigationList
     */

    // SubId Locators *****************************************************

    /**
     * <p>Sub-ID for the oj-navigtion-list component's list item element.</p>
     *
     * <p>
     * To lookup the list items the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-navigationlist-item'</li>
     * <li><b>key</b>: the <a href="#key-section">Key</a> of the item</li>
     * </ul>
     *
     * @ojsubid oj-navigationlist-item
     * @memberof oj.ojNavigationList
     *
     * @example <caption>Get the list item element with key 'foo':</caption>
     * var node = myNavList.getNodeBySubId({'subId': 'oj-navigationlist-item', 'key': 'foo'} );
     */
    /**
     * <p>Sub-ID for the oj-navigtion-list component's previous link element.</p>
     *
     * To lookup the previous link the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-navigationlist-previous-link'</li>
     * </ul>
     * @ojsubid oj-navigationlist-previous-link
     * @memberof oj.ojNavigationList
     *
     * @example <caption>Get the previous link element:</caption>
     * var node = myNavList.getNodeBySubId({'subId': 'oj-navigationlist-previous-link'} );
     */
    /**
     * <p>Sub-ID for the oj-navigtion-list component's hierarchical menu button element.
     * See the <a href="#getNodeBySubId">getNodeBySubId</a> method for details.</p>
     *
     * To lookup the hierarchical menu button the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-navigationlist-hierarchical-button'</li>
     * </ul>
     * @ojsubid
     * @name oj-navigationlist-hierarchical-button
     * @memberof oj.ojNavigationList
     *
     * @example <caption>Get the hierarchical menu button element:</caption>
     * var node = myNavList.getNodeBySubId({'subId': 'oj-navigationlist-hierarchical-button'} );
     */
    /**
     * <p>Sub-ID for the oj-navigation-list component's hierarchical menu element.</p>
     *
     * To lookup the hierarchical menu the locator object should have the following:
     * <ul>
     * <li><b>subId</b>: 'oj-navigationlist-hierarchical-menu'</li>
     * </ul>
     * @ojsubid
     * @name oj-navigationlist-hierarchical-menu
     * @memberof oj.ojNavigationList
     *
     * @example <caption>Get the hierarchical menu element:</caption>
     * var node = myNavList.getNodeBySubId({'subId': 'oj-navigationlist-hierarchical-menu'} );
     */

    // Node Context Objects *********************************************
    /**
     * <p>Context for the oj-navigation-list component's items.</p>
     *
     * @property {number} index the zero based item index relative to its parent
     * @property {Object|string} key the <a href="#key-section">Key</a> of the item
     * @property {Element} parent the parent group item.  Only available if item has a parent.
     * @property {boolean} group whether the item is a group.
     *
     * @ojnodecontext oj-navigationlist-item
     * @memberof oj.ojNavigationList
     */
  });

  // Add custom getters for properties from theming file
  oj.Components.setDefaultOptions({
    'ojNavigationList': {
      'hierarchyMenuDisplayThresholdLevel': oj.Components.createDynamicPropertyGetter(
        function () {
          return (oj.ThemeUtils.parseJSONFromFontFamily('oj-navigationlist-option-defaults') || {})["hierarchyMenuDisplayThresholdLevel"];
        }
      )
    }
  });
}());

// DOCLETS
  /**
   * @ojcomponent oj.ojTabBar
   * @augments oj.baseComponent
   * @ojstatus preview
   * @since 4.0.0
   * @ojrole tablist
   * @ojtsimport ojdataprovider
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojTabBar<K, D> extends baseComponent<ojTabBarSettableProperties<K,D>>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojTabBarSettableProperties<K,D> extends baseComponentSettableProperties",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @ojshortdesc Displays tab bar with advanced interactive features.
   * @classdesc
   * <h3 id="navlistOverview-section">
   *   JET Tab Bar
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#navlistOverview-section"></a>
   * </h3>
   *
   * <p>Description: JET Tab Bar enhances a HTML list element into a themable, WAI-ARIA compliant, mobile friendly component with advance interactive features.
   *
   * <h3 id="data-section">
   *   Data
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#data-section"></a>
   * </h3>
   * <p>The JET Tab Bar gets its data in two different ways.  The first way is from a DataProvider/TableDataSource.  There are several types of DataProvider/TableDataSource 
   * that are available out of the box:</p>
   * <ul>
   * <li>oj.ArrayDataProvider</li>
   * <li>oj.CollectionTableDataSource</li>
   * </ul>
   *
   * <p><b>oj.ArrayDataProvider</b> - Use this when the underlying data is an array object or an observableArray.  In the observableArray case, Tab Bar will automatically react
   * when items are added or removed from the array.  See the documentation for oj.ArrayDataProvider for more details on the available options.</p>
   *
   * <p><b>oj.CollectionTableDataSource</b> - Use this when oj.Collection is the model for the underlying data.  Note that the Tab Bar will automatically react to model event from
   * the underlying oj.Collection.  See the documentation for oj.CollectionTableDataSource for more details on the available options.</p>
   *
   * <p> NOTE: oj.PagingTableDataSource is not supported by Tab Bar.
   * <p>Second way is using static HTML content as data.</p>
   *
   * <p>Example of static content</p>
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-tab-bar>
   *  &lt;ul>
   *   &lt;li>&lt;a href="#">Item 1&lt;/a>&lt;/li>
   *   &lt;li>&lt;a href="#">Item 2&lt;/a>&lt;/li>
   *   &lt;li>&lt;a href="#">Item 3&lt;/a>&lt;/li>
   *  &lt;/ul>
   * &lt;/oj-tab-bar>
   * </code></pre>
   * <p>Any list item can be disabled by adding the <code class="prettyprint">oj-disabled</code> class to that element.  As with any DOM change, doing so post-init
   * requires a <code class="prettyprint">refresh()</code> of the element.
   *
   * <h3 id="key-section">
   *   Key
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#key-section"></a>
   * </h3>
   * <p>Key is an identifier which uniquely identifies an item in tabbar. 
   *  <ul>
   *  <li>When static html is used, it will be the <code class="prettyprint">id</code> attribute of <code class="prettyprint">&lt;li></code>. If no <code class="prettyprint">id</code> is specified then component will generate an id and will use it as key.</li>
   *  <li>When data source is used, it will be the <code class="prettyprint">id</code> attribute of item's data object.
   *  </ul>
   * <h3 id="icons-section">
   *   Icons
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#icons-section"></a>
   * </h3>
   *
   * <p>Sublist icons are inserted automatically.  To add other icons to list items, include them in the markup and include the <code class="prettyprint">oj-tabbar-item-icon</code> class, as follows:
   *
   * <pre class="prettyprint">
   * <code>&lt;oj-tab-bar>
   *   &lt;ul>    
   *     &lt;li id="foo">&lt;a href="#">&lt;span class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>Foo&lt;/a>&lt;/li>
   *   &lt;/ul>
   *  &lt;/oj-tab-bar>
   * </code></pre>
   *
   * <h3 id="styling-section">
   *   Styling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"stylingDoc"}
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   * {@ojinclude "name":"keyboardDoc"}
   *
   * <p>Disabled items will not receive keyboard focus and do not allow any interaction.
   *
   *<h3 id="context-section">
   *   Item Context
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#context-section"></a>
   * </h3>
   *
   * <p>For all item options, developers can specify a function as the return value.  The function takes a single argument, which is an object that contains contextual information about the particular item.  This gives developers the flexibility to return different value depending on the context.</p>
   *
   * <p>The context paramter contains the following keys:</p>
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Key</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td><kbd>componentElement</kbd></td>
   *       <td>oj-tab-bar element.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>datasource</kbd></td>
   *       <td>A reference to the data source object. (Not available for static content)</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>index</kbd></td>
   *       <td>The index of the item, where 0 is the index of the first item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>key</kbd></td>
   *       <td>The <a href="#key-section">Key</a> of the item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>data</kbd></td>
   *       <td>The data object for the item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>parentElement</kbd></td>
   *       <td>The list item element.  The renderer can use this to directly append content.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * <p></p>
   * <h3 id="perf-section">
   *   Performance
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
   * </h3>
   *
   * <h4>Data Set Size</h4>
   * <p>As a rule of thumb, it's recommended that applications limit the amount of data to display.  Displaying large
   * number of items in Tab Bar makes it hard for user to find what they are looking for, but affects the load time.
   *
   * <h4>Item Content</h4>
   * <p>Tab Bar allows developers to specify arbitrary content inside its item. In order to minimize any negative effect on
   * performance, you should avoid putting a large number of heavy-weight components inside because as you add more complexity
   * to the structure, the effect will be multiplied because there can be many items in the Tab Bar.</p>
   *
   * <h3 id="a11y-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
   * </h3>
   * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.
   * <p>Disabled content: JET supports an accessible luminosity contrast ratio,
   * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>,
   * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which
   * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user
   * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio
   * required of enabled content, it cannot be used to convey meaningful information.<p>
   *
   *
   * <h3 id="rtl-section">
   *   Reading direction
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
   * </h3>
   *
   * <p>The only supported way to set the reading direction (LTR or RTL) is to set the <code class="prettyprint">"dir"</code> attribute on the
   * <code class="prettyprint">&lt;html></code> element of the page.  As with any JET custom element, in the unusual case that the reading direction
   * is changed post-init, the tab bar must be <code class="prettyprint">refresh()</code>ed, or the page must be reloaded.
   *
   *
   * <h3 id="animation-section">
   *   Animation
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#animation-section"></a>
   * </h3>
   *
   * <p>Applications can customize animations triggered by actions in Tab Bar by either listening for <code class="prettyprint">animateStart/animateEnd</code>
   *    events or overriding action specific style classes on the animated item.  See the documentation of <a href="oj.AnimationUtils.html">oj.AnimationUtils</a> 
   *    class for details.</p>
   *
   * <p>The following are actions and their corresponding sass variables in which applications can use to customize animation effects.  
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Action</th>
   *       <th>Sass Variable</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td rowspan="2"><kbd>add</kbd></td>
   *       <td>$tabBarHorizontalAddAnimation</td>
   *       <td>When a new item is added to the oj.TableDataSource associated with Horizontal Tab Bar.</td>
   *     </tr>
   *     <tr>
   *       <td>$tabBarAddAnimation</td>
   *       <td>When a new item is added to the oj.TableDataSource associated with Vertical Tab Bar.</td>
   *     </tr>
   *     <tr>
   *       <td rowspan="2"><kbd>remove</kbd></td>
   *       <td>$tabBarHorizontalRemoveAnimation</td>
   *       <td>When an existing item is removed from the oj.TableDataSource associated with Horizontal Tab Bar.</td>
   *     </tr>
   *     <tr>
   *       <td>$tabBarRemoveAnimation</td>
   *       <td>When an existing item is removed from the oj.TableDataSource associated with Vertical Tab Bar.</td>
   *     </tr>   
   *     <tr>
   *       <td><kbd>update</kbd></td>
   *       <td>$tabBarUpdateAnimation</td>
   *       <td>When an existing item is updated in the oj.TableDataSource associated with TabBar.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>pointerUp</kbd></td>
   *       <td>$tabBarPointerUpAnimation</td>
   *       <td>When user finish pressing an item (on touch).</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   */
  /**
   * <a href="#key-section">Key</a> of the current item.  Current item is the list item which is having active focus.  Note that if currentItem
   * is set to an item that is currently not available (not fetched or
   * inside a collapsed parent node), then the value is ignored.
   * <p>
   * When the current item is changed, the <code class="prettyprint">event.detail</code> of the <code class="prettyprint">currentItemChanged</code> event will contain the following additional properties:<br><br>
   * <table class="props">
   *   <thead>
   *     <tr>
   *       <th>Name</th>
   *       <th>Type</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td class="name"><code>item</code></td>
   *       <td class="type">Element</td>
   *       <td class="description">Current Item element</td>
   *     </tr>
   *   </tbody>
   * </table>  
   * @expose
   * @public
   * @instance
   * @name currentItem
   * @memberof oj.ojTabBar
   * @type {*}
   * @default null
   * @ojshortdesc Gets and sets the key of the item that should have keyboard focus.
   * @ojwriteback
   *
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">current-item</code> attribute specified:</caption>
   *  &lt;oj-tab-bar current-item='settings'> ... &lt;/oj-tab-bar>
   * @example <caption>Get the current item:</caption>
   * var currentItem = myTabBar.currentItem;
   *
   * @example <caption>Set the current item on the tabbar:</caption>
   * myTabBar.currentItem = "newItem";
   */

  /**
   * The position of the Tab Bar. Valid Values: top, bottom, start and end.
   * @expose
   * @name edge
   * @memberof oj.ojTabBar
   * @instance
   * @type {string}
   * @ojvalue {string} "top" This renders list items horizontally. Generally used when tab bar placed on top of content section.
   * @ojvalue {string} "bottom" This renders list items horizontally. Generally used when tab bar placed on bottom of content section.
   * @ojvalue {string} "start" This renders list items vertically. Generally used when tab bar placed on left/start of content section.
   * @ojvalue {string} "end" This renders list items vertically. Generally used when tab bar placed on right/end of content section.
   * @default start
   * @ojshortdesc Gets and sets the edge position of the Tab Bar.
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">edge</code> attribute specified:</caption>
   *  &lt;oj-tab-bar edge='top'> ... &lt;/oj-tab-bar>
   * @example <caption>Get the edge:</caption>
   * var edge = myTabBar.edge;
   *
   * @example <caption>Set the edge on the tabbar:</caption>
   * myTabBar.edge = "top";
   */
  /**
   * Specifies if the tabs can be reordered within the tab bar by drag-and-drop.
   * @expose
   * @name reorderable
   * @memberof oj.ojTabBar
   * @instance
   * @ojshortdesc Enable or disable the item reordering functionalities.
   * @ojstatus preview
   * @since 4.1.0
   * @type {string}
   * @ojvalue {string} "enabled" Enables reordering of items in tabbar.
   * @ojvalue {string} "disabled" Disables reordering of items in tabbar.
   * @default disabled
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">reorderable</code> attribute specified:</caption>
   *  &lt;oj-tab-bar reorderable='enabled'> ... &lt;/oj-tab-bar>
   * @example <caption>Get the reorderable:</caption>
   * var reorderable = myTabBar.reorderable;
   *
   * @example <caption>Set the edge on the tabbar:</caption>
   * myTabBar.reorderable = "enabled";
   */
  /**
   * Truncation applies to the tab titles when there is not enough room to display all tabs.
   * @expose
   * @name truncation
   * @memberof oj.ojTabBar
   * @instance
   * @ojstatus preview
   * @ojshortdesc Gets and sets whether truncation needs to be applied.
   * @since 4.1.0
   * @type {string}
   * @ojvalue {string} "none" tabs always take up the space needed by the title texts.
   * @ojvalue {string} "progressive"  If not enough space is available to display all of the tabs, then the width of each tab title is restricted just enough to allow all tabs to fit. All tab titles that are truncated are displayed with ellipses. However the width of each tab title will not be truncated below $tabBarTruncatedLabelMinWidth.
   * @default none
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">truncation</code> attribute specified:</caption>
   *  &lt;oj-tab-bar truncation='progressive'> ... &lt;/oj-tab-bar>
   * @example <caption>Get the truncation property:</caption>
   * var truncation = myTabBar.truncation;
   *
   * @example <caption>Set the truncation on the tabbar:</caption>
   * myTabBar.truncation = "progressive";
   */

  /**
   * Item <a href="#key-section">Key</a> of currently selected list item. If the value is modified
   * by an application, tab bar UI is modified to match the new value and the <code class="prettyprint">event.detail</code> of the <code class="prettyprint">selectionChanged</code> event will contain the following additional properties:<br><br>
   * <table class="props">
   *   <thead>
   *     <tr>
   *       <th>Name</th>
   *       <th>Type</th>
   *       <th>Description</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td class="name"><code>item</code></td>
   *       <td class="type">Element</td>
   *       <td class="description">Selected Item element</td>
   *     </tr>
   *   </tbody>
   * </table>
   * @type {*}
   * @name selection
   * @ojshortdesc Gets and sets the key of the selected item.
   * @default null
   * @ojwriteback
   * @expose
   * @instance
   * @memberof oj.ojTabBar
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">selection</code> attribute specified:</caption>
   *  &lt;oj-tab-bar selection='settings'> ... &lt;/oj-tab-bar>
   *  
   * @example <caption>Get the selection:</caption>
   * var selection = myTabBar.selection;
   *
   * @example <caption>Set the selection on the tabbar:</caption>
   * myTabBar.selection = "settings";
   */

  /**
   * The data source for the Tab Bar accepts either a oj.TableDataSource or oj.DataProvider.
   * See the data source section in the introduction for out of the box data source types.
   * If the data attribute is not specified, the child elements are used as content.  If there's no
   * content specified, then an empty list is rendered.
   *
   * @expose
   * @name data
   * @memberof oj.ojTabBar
   * @instance
   * @ojshortdesc Gets and sets the data provider for tabbar.
   * @type {oj.TableDataSource|oj.DataProvider|null}
   * @ojsignature {target: "Type", value: "oj.DataProvider<K, D>|null"}
   * @default null
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">data</code> attribute specified:</caption>
   *  &lt;oj-tab-bar data='[[tableDataSource]]'> ... &lt;/oj-tab-bar>
   * @example <caption>Get the data:</caption>
   * var dataSource = myTabBar.data;
   *
   * @example <caption>Set the data attribute using one-dimensional array:</caption>
   * myTabBar.data = new oj.ArrayTableDataSource([1,2,3]);
   * 
   * @example <caption>Set the data attribute using oj.Collection:</caption>
   * myTabBar.data = new oj.CollectionTableDataSource(collection);
   */

  /**
   * Whether to display both the label and icons (<code class="prettyprint">"all"</code>) or just the icons (<code class="prettyprint">"icons"</code>).
   * In the latter case, the label is displayed in a tooltip instead, unless a tooltip was already supplied at create time.
   * Note: <code class="prettyprint">display="icons"</code> is valid only when <code class="prettyprint">drillMode=none</code> and tab bar is a flat list.
   * It is also mandatory to provide icons for each item as stated in <a href="#icons-section">icons section</a>.
   *
   * @expose
   * @memberof oj.ojTabBar
   * @instance
   * @name display
   * @type {string}
   * @ojvalue {string} "all" Display both the label and icons.
   * @ojvalue {string} "icons" Display only the icons.
   * @default all
   * @ojshortdesc Gets and sets whether all or only icons need to be displayed.
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">display</code> attribute specified:</caption>
   *  &lt;oj-tab-bar display='icons'> ... &lt;/oj-tab-bar>
   *  
   * @example <caption>Get or set the <code class="prettyprint">display</code> property:</caption>
   * // getter
   * var display = myTabBar.display;
   *
   * // setter
   * myTabBar.display = "icons";
   */

   /**
   * Specifies the overflow behaviour. 
   * NOTE: This is only applicable when <code class="prettyprint">edge</code> attribute set to <code class="prettyprint">top</code>
   * @expose
   * @memberof oj.ojTabBar
   * @name overflow
   * @instance
   * @type {string}
   * @ojvalue {string} "popup" popup menu will be shown with overflowed items. <p> NOTE: Setting <code class="prettyprint">overflow</code> to <code class="prettyprint">popup</code> can trigger browser reflow, so only set it when it is actually required.
   * @ojvalue {string} "hidden" overflow is clipped, and the rest of the content will be invisible.
   * @default hidden
   * @ojshortdesc Gets and sets overflow behaviour for Tab bar.
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">overflow</code> attribute specified:</caption>
   *  &lt;oj-tab-bar overflow='popup'> ... &lt;/oj-tab-bar>
   * @example <caption>Get or set the <code class="prettyprint">overflow</code> property:</caption>
   * // getter
   * var overflow = myTabBar.overflow;
   *
   * // setter
   * myTabBar.overflow = "popup";
   */

/**
   * @typedef {Object} oj.ojTabBar.ItemContext
   * @property {Element} componentElement oj-tab-bar element
   * @property {oj.DataProvider<K, D>} [datasource] A reference to the data source object. (Not available for static content)
   * @property {number} index The index of the item, where 0 is the index of the first item.
   * @property {K} key The Key of the item.
   * @property {D} data The data object for the item.
   * @property {Element} parentElement The list item element. The renderer can use this to directly append content.
   * @ojsignature [{target:"Type", value:"<K, D>", for:"genericTypeParameters"}]
   */

  /**
   * The item property contains a subset of properties for items.
   * @ojshortdesc Customize the functionalities of each tab on Tab bar.  
   * @expose
   * @memberof oj.ojTabBar
   * @instance
   * @name item
   * @type {Object}
   */
  /**
   * The renderer function that renders the content of the item. See <a href="#context-section">itemContext</a>
   * in the introduction to see the object passed into the renderer function.
   * The function should return one of the following: 
   * <ul>
   *   <li>An Object with the following property:
   *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the item.</li></ul>
   *   </li>
   *   <li>undefined: If the developer chooses to manipulate the item element directly, the function should return undefined.</li>
   * </ul>
   * If no renderer is specified, Tab Bar will treat the data as a String.
   *
   * @expose
   * @ojshortdesc Gets and sets the renderer for the tab item.
   * @name item.renderer
   * @memberof! oj.ojTabBar
   * @instance
   * @ojsignature { target: "Type",
   *                value: "?(((context: oj.ojTabBar.ItemContext<K, D>) => void)|null)",
   *                jsdocOverride: true}
   * @type {function(Object)|null}
   * @default null
   * @example <caption>Initialize the Tab Bar with the <code class="prettyprint">item.renderer</code> attribute specified:</caption>
   *  &lt;oj-tab-bar item.renderer="{{oj.KnockoutTemplateUtils.getRenderer('template', true)}}"> ... &lt;/oj-tab-bar>
   *  
   * @example <caption>Get or set the <code class="prettyprint">renderer</code> property:</caption>
   * // set the renderer function
   * myTabBar.item.renderer = rendererFn;
   * 
   * // get the renderer function
   * var rendererFn = myTabBar.item.renderer;
   */
  /**
   * Whether the item is selectable.
   * See <a href="#context-section">itemContext</a> in the introduction to see the object passed into the selectable function.
   *
   * @expose
   * @ojshortdesc Gets and sets whether the tab can be selected.
   * @name item.selectable
   * @memberof! oj.ojTabBar
   * @instance
   * @ojsignature { target: "Type",
   *                value: "?(((context: oj.ojTabBar.ItemContext<K, D>) => boolean)|boolean)",
   *                jsdocOverride: true}
   * @type {function(Object)|boolean}
   * @default true
   *
   * @example <caption>Initialize the Tab bar such that the first 3 items are not selectable:</caption>
   *  &lt;oj-tab-bar item.selectable="[[skipFirstThreeElementsFn]]">&lt;/oj-tab-bar>
   *  
   *  var skipFirstThreeElementsFn = function(itemContext) {
   *                                      return itemContext['index'] > 3; 
   *                                 }
   */
  //Events
  /**
   * Triggered when the default animation of a particular action is about to start.  The default animation can be cancelled by calling event.preventDefault.
   *
   * @expose
   * @ojshortdesc Event handler for when the default animation of a particular action is about to start.
   * @event
   * @name animateStart
   * @memberof oj.ojTabBar
   * @instance
   * @property {string} action the action that starts the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
   * @property {Element} element the target of animation.  
   * @property {function():void} endCallback if the event listener calls event.preventDefault to cancel the default animation, it must call the endCallback function when it finishes its own animation handling and when any custom animation ends.
   */
  /**
   * Triggered when the default animation of a particular action has ended.  Note this event will not be triggered if application cancelled the default animation on animateStart.
   *
   * @expose
   * @ojshortdesc Event handler for when the default animation of a particular action has ended.
   * @event
   * @name animateEnd
   * @memberof oj.ojTabBar
   * @instance
   * @property {string} action the action that started the animation.  See <a href="#animation-section">animation</a> section for a list of actions.
   * @property {Element} element the target of animation.  
   */

  /**
   * <p>Triggered before this list item is selected.
   * To prevent the item selection, invoke <code class="prettyprint">event.preventDefault()</code>.
   * @ojshortdesc Event handler for when before the selection is changed.
   * @expose
   * @event
   * @ojcancelable
   * @name beforeSelect
   * @memberof oj.ojTabBar
   * @instance
   * @property {*} key Selected list item <a href="#key-section">Key</a>.
   * @property {Element} item Selected list item.
   */

 
  /**
   * Triggered before the current item is changed via the <code class="prettyprint">currentItem</code> property or via the UI.
   * To prevent the item being focused, invoke <code class="prettyprint">event.preventDefault()</code>.
   * @ojshortdesc Event handler for when before the current item is changed.
   * @expose
   * @event
   * @memberof oj.ojTabBar
   * @ojcancelable
   * @name beforeCurrentItem
   * @instance
   * @property {*} previousKey the <a href="#key-section">Key</a> of the previous item
   * @property {Element} previousItem the previous item
   * @property {*} key the <a href="#key-section">Key</a> of the new current item
   * @property {Element} item the new current item
   */
  /**
   * Triggered immediately before a tab is deselected.
   * To prevent the item being deselected, invoke <code class="prettyprint">event.preventDefault()</code>.
   * @ojshortdesc Event handler for when before tab is deselected.
   * @expose
   * @event
   * @ojcancelable
   * @memberof oj.ojTabBar
   * @name beforeDeselect
   * @ojstatus preview
   * @since 4.1.0
   * @instance
   * @property {*} fromKey the <a href="#key-section">Key</a> of the tab item being navigated from
   * @property {Element} fromItem the tab item being navigated from
   * @property {*} toKey the <a href="#key-section">Key</a> of the tab item being navigated to
   * @property {Element} toItem the tab item being navigated to
   */
  /**
   * Triggered after a tab has been deselected.
   * @ojshortdesc Event handler for when a tab is deselected.
   * @expose
   * @event
   * @memberof oj.ojTabBar
   * @name deselect
   * @ojstatus preview
   * @since 4.1.0
   * @instance
   * @property {*} fromKey the <a href="#key-section">Key</a> of the tab item being navigated from
   * @property {Element} fromItem the tab item being navigated from
   * @property {*} toKey the <a href="#key-section">Key</a> of the tab item being navigated to
   * @property {Element} toItem the tab item being navigated to
   */
  /**
   * Triggered before the item is removed via the UI.
   * To prevent the item being removed, invoke <code class="prettyprint">event.preventDefault()</code>.
   * @ojshortdesc Event handler for when a tab is about to be removed.
   * @expose
   * @event
   * @ojcancelable
   * @memberof oj.ojTabBar
   * @name beforeRemove
   * @ojstatus preview
   * @since 4.1.0
   * @instance
   * @property {Element} item Item being removed
   * @property {*} key <a href="#key-section">Key</a> of the item being removed
   */
  /**
   * Triggered immediately after a tab is removed. This should be used to remove item from dom or from data source.
   * @ojshortdesc Event handler for when a tab is removed.
   * @expose
   * @event
   * @memberof oj.ojTabBar
   * @name remove
   * @ojstatus preview
   * @since 4.1.0
   * @instance
   * @property {Element} item Item removed
   * @property {*} key <a href="#key-section">Key</a> of the item removed
   */
  /**
   * Triggered after reordering items within tabbar via drag and drop or cut and paste. This should be used to reorder item in dom or data source.
   * @ojshortdesc Event handler for when a tab is reordered.
   * @expose
   * @event
   * @memberof oj.ojTabBar
   * @name reorder
   * @ojstatus preview
   * @since 4.1.0
   * @instance
   * @property {Element} item Item to be moved
   * @property {('before'|'after')} position the drop position relative to the reference item. Possible values are "before" and "after".
   * @property {Element} reference the item where the moved items are drop on.
   */
  /**
   * @typedef {Object} oj.ojTabBar.NodeContext
   * @property {string} subId Sub-id string to identify a particular dom node.
   * @property {number} index The index of the item, where 0 is the index of the first item.
   * @property {K} key The Key of the item.
   * @ojsignature [{target:"Type", value:"<K>", for:"genericTypeParameters"}]
   */
  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   * @ojsignature { target: "Type", for: "returns",
   *                value: "oj.ojTabBar.NodeContext<K>|null",
   *                jsdocOverride: true}
   * @example {@ojinclude "name":"nodeContextExample"}
   * @method
   * @expose
   * @name getContextByNode
   * @instance
   * @memberof oj.ojTabBar
   */

  /**
   * Refreshes the visual state of the Tab Bar. JET components require a <code class="prettyprint">refresh()</code> after the DOM is
   * programmatically changed underneath the component.
   * <p>This method does not accept any arguments.
   *
   * @ojshortdesc Refreshes the visual state of the Tab Bar.
   * @expose
   * @memberof oj.ojTabBar
   * @instance
   * @name refresh
   * @return {void}
   * @method
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   *  myTabBar.refresh();
   */
  // Fragments:

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Gesture</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td rowspan="2">List Item</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Selects the item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>Press & Hold</kbd></td>
   *       <td>Display context menu</td>
   *     </tr>
   *     <tr>
   *       <td>Overflow Menu button</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Open menu. Refer <a href="oj.ojButton.html#touch-section">menu button</a> touch documentation. Note: This is applicable only for Horizontal Tab Bar when <code class="prettyprint">overflow</code> is set to <code class="prettyprint">popup</code>. </td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojTabBar
   */

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td rowspan="12">List Item</td>
   *       <td><kbd>Enter</kbd> or <kbd>Space</kbd></td>
   *       <td>Selects list item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>UpArrow</kbd></td>
   *       <td>Moves focus to the previous visible list item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>DownArrow</kbd></td>
   *       <td>Moves focus to the next  visible list item</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>RightArrow</kbd> (<kbd>LeftArrow</kbd> in RTL)</td>
   *       <td>For horizontal tab bar,focus will be moved to next visible item.
   *       </td>
   *     </tr>
   *     <tr>
   *       <td><kbd>LeftArrow</kbd> (<kbd>RightArrow</kbd> in RTL)</td>
   *       <td>For horizontal tab bar,focus will be moved to previous visible item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>Home</kbd></td>
   *       <td>Moves focus to the first visible list item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>End</kbd></td>
   *       <td>Moves focus to the last visible list item.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>F2</kbd></td>
   *       <td>If focus is on a list item, pressing F2 will make its contents accessible using TAB.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>Esc</kbd></td>
   *       <td>When F2 mode is enabled, press Esc to exit F2 mode.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>Ctrl+X</kbd></td>
   *       <td>Marks the current item to move if reorderable is enabled.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>Ctrl+V</kbd></td>
   *       <td>Paste the item that are marked to directly before the current item</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>DELETE</kbd></td>
   *       <td>Delete the current item. </td>
   *     </tr>
   *     <tr>
   *       <td>Overflow Menu button</td>
   *       <td><kbd>Enter or Space</kbd></td>
   *       <td>Open menu. Refer <a href="oj.ojButton.html#touch-section">menu button</a> touch documentation. Note: This is applicable only for Horizontal Tab Bar when <code class="prettyprint">overflow</code> is set to <code class="prettyprint">popup</code>. </td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojTabBar
   */

  /**
   * {@ojinclude "name":"ojStylingDocIntro"}
   * <table class="generic-table styling-table">
   *   <thead>
   *     <tr>
   *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
   *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
   *       <th>{@ojinclude "name":"ojStylingDocExampleHeader"}</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>oj-tabbar-stack-icon-label</td>
   *       <td>Displays horizontal Tab Bar with icon and label stacked. Applicable only when <code class="prettyprint">edge</code> is <code class="prettyprint">top</code>.</td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar class="oj-tabbar-stack-icon-label" >
   *  &lt;ul>    
   *    &lt;li id="foo">
   *      &lt;a href="#">
   *        &lt;span class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">
   *        &lt;/span>
   *        Foo
   *      &lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-tabbar-category-divider</td>
   *       <td>Use this class to add horizontal divider line between two categories of items.</td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar>
   *  &lt;ul>
   *    &lt;li ...> &lt;/li>
   *    &lt;li class="oj-tabbar-category-divider"> &lt;/li>    
   *    &lt;li id="foo">
   *      &lt;a href="#">
   *        &lt;span class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">
   *        &lt;/span>
   *        Foo
   *      &lt;/a>
   *    &lt;/li>
   *    &lt;li ...> &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>     
   *     </tr>
   *     <tr>
   *       <td>oj-tabbar-item-icon</td>
   *       <td>Use this class to add icon to list item.</td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar>
   *  &lt;ul>    
   *    &lt;li id="foo">
   *      &lt;a href="#">
   *        &lt;span class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">
   *        &lt;/span>
   *        Foo
   *      &lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar></code></pre>
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-tabbar-item-title</td>
   *       <td>When arbitrary content is placed inside item's content area, it's title text can be marked using this style class. This helps component in identifying the Item's label.
   *       </td>
   *       <td>
   * <pre class="prettyprint">
   * <code>&lt;li>
   *   &lt;div>
   *     &lt;span class="oj-tabbar-item-title">Play&lt;/span>
   *     &lt;button>Button&lt;/button>
   *   &lt;/div>
   * &lt;/li>
   * </code></pre>
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-tabbar-item-text-wrap</td>
   *       <td>Use this class to wrap item label text. Note: On IE11, this is not supported when <code class="prettyprint">overflow</code> attribute is set to <code class="prettyprint">popup</code>.</td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar class="oj-tabbar-item-text-wrap" >
   *  &lt;ul>    
   *    &lt;li id="foo">
   *      &lt;a href="#">
   *        &lt;span class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">
   *        &lt;/span>
   *        Foo
   *      &lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>     
   *     </tr>
   *     <tr>
   *       <td>oj-tabbar-item-dividers</td>
   *       <td>Use this class to show dividers between horizontal tab bar items.</td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar class="oj-tabbar-item-dividers" >
   *  &lt;ul>    
   *    &lt;li id="foo">
   *       &lt;a href="#">
   *         &lt;span class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
   *         Foo
   *       &lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>     
   *     </tr>     
   *     </tr>
   *     <tr>
   *       <td>oj-sm-condense</td>
   *       <td>Use this class to condense horizontal tab bar items on small screens and larger.
   *       </td>
   *       <td rowspan="4">
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar class="oj-sm-condense" >
   *  &lt;ul>    
   *    &lt;li id="foo">
   *      &lt;a href="#">
   *        &lt;span class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
   *        Foo
   *      &lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>      
   *     </tr>
   *     <tr>
   *       <td>oj-md-condense</td>
   *       <td>Use this class to condense horizontal tab bar items on medium screens and larger.
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-lg-condense</td>
   *       <td>Use this class to condense horizontal tab bar items on large screens and larger.
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-xl-condense</td>
   *       <td>Use this class to condense horizontal tab bar items on extra large screens and larger.
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-tabbar-nofollow-link</td>
   *       <td> Use this class to prevent automatic navigation to the url specified on <code class="prettyprint">&lt;a></code> tag's <code class="prettyprint">href</code> attribute. In this case, navigation can be handled programmatically by using <code class="prettyprint">selectionChanged</code> event. This is useful to execute some custom tasks before browser triggers navigation.
   *       </td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar class="oj-tabbar-nofollow-link" >
   *  &lt;ul>    
   *    &lt;li id="foo">&lt;a href="folder/foo.html">
   *      &lt;span 
   *        class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
   *      Foo&lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>     
   *     </tr>
   *     <tr>
   *       <td>oj-removable</td>
   *       <td> Use this class to make an item removable.
   *       </td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar >
   *  &lt;ul>    
   *    &lt;li id="foo" class="oj-removable" >&lt;a href="folder/foo.html">
   *      &lt;span 
   *        class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
   *      Foo&lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>     
   *     </tr>
   *     <tr>
   *       <td>oj-tabbar-hide-remove-icon</td>
   *       <td> Use this class to hide the remove icon. In this case, Item can be removed using context menu.
   *       </td>
   *       <td>
   *          <pre class="prettyprint">
   *<code>&lt;oj-tab-bar class="oj-tabbar-hide-remove-icon">
   *  &lt;ul>    
   *    &lt;li id="foo" class="oj-removable" >&lt;a href="folder/foo.html">
   *      &lt;span 
   *        class="oj-tabbar-item-icon demo-icon-font-24 demo-palette-icon-24">&lt;/span>
   *      Foo&lt;/a>
   *    &lt;/li>
   *  &lt;/ul>
   *&lt;/oj-tab-bar>
   *</code></pre>
   *       </td>     
   *     </tr>
   *     <tr>
   *       <td>oj-focus-highlight</td>
   *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
   *       <td></td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
   * @memberof oj.ojTabBar
   */

  // SubId Locators *****************************************************

  /**
   * <p>Sub-ID for the oj-tab-bar component's list item element.</p>
   *
   * <p>
   * To lookup the list items the locator object should have the following:
   * <ul>
   * <li><b>subId</b>: 'oj-tabbar-item'</li>
   * <li><b>key</b>: the <a href="#key-section">Key</a> of the item</li>
   * </ul>
   *
   * @ojsubid oj-tabbar-item
   * @memberof oj.ojTabBar
   *
   * @example <caption>Get the list item element with key 'foo':</caption>
   * var node = myTabBar.getNodeBySubId({'subId': 'oj-tabbar-item', 'key': 'foo'} );
   */
  /**
   * <p>Sub-ID for the oj-tab-bar items's removable icon.</p>
   *
   * <p>
   * To lookup the list items the locator object should have the following:
   * <ul>
   * <li><b>subId</b>: 'oj-tabbar-removable-icon'</li>
   * <li><b>key</b>: the key of the item</li>
   * </ul>
   *
   * @ojsubid oj-tabbar-item
   * @memberof oj.ojTabBar
   *
   * @example <caption>Get the remove icon of item with key 'foo':</caption>
   * var node = myTabBar.getNodeBySubId({'subId': 'oj-tabbar-removable-icon', 'key': 'foo'} );
   */


  // Node Context Objects *********************************************
  /**
   * <p>Context for the oj-tab-bar component's items.</p>
   *
   * @property {number} index the zero based item index
   * @property {Object|string} key the <a href="#key-section">Key</a> of the item
   *
   * @ojnodecontext oj-tabbar-item
   * @memberof oj.ojTabBar
   */
(function() {
var ojNavigationListMeta = {
  "properties": {
    "currentItem": {
      "type": "any",
      "writeback": true
    },
    "data": {},
    "display": {
      "type": "string",
      "enumValues": ["all", "icons"]
    },
    "drillMode": {
      "type": "string",
      "enumValues": ["sliding", "collapsible", "none"]
    },
    "expanded": {
      "type": "object",
      "writeback": true
    },
    "edge": {
      "type": "string",
      "enumValues": ["top", "start"]
    },
    "hierarchyMenuThreshold": {
      "type": "number"
    },
    "item": {
      "type": "object",
      "properties": {
        "renderer": {},
        "selectable": {}
      }
    },
    "overflow": {
      "type": "string",
      "enumValues": ["popup", "hidden"]
    },
    "rootLabel": {
      "type": "string"
    },
    "selection": {
      "type": "any",
      "writeback": true
    },
    "translations": {
      "type": "Object",
      "properties": {
        "defaultRootLabel": {
          "type": "string",
          "value": "Navigation List"
        },
        "hierMenuBtnLabel": {
          "type": "string",
          "value": "Hierarchical Menu button"
        },
        "previousIcon": {
          "type": "string",
          "value": "Previous"
        }
      }
    }
  },
  "events": {
    "animateEnd": {},
    "animateStart": {},
    "beforeCollapse": {},
    "beforeCurrentItem": {},
    "beforeExpand": {},
    "beforeSelect": {},
    "collapse": {},
    "expand": {}
  },
  "methods": {
    "getContextByNode": {}
  },
  "extension": {
    _ALIASED_PROPS: {"hierarchyMenuThreshold":"hierarchyMenuDisplayThresholdLevel"},
    _WIDGET_NAME: "ojNavigationList"
  }
};
oj.CustomElementBridge.registerMetadata('oj-navigation-list', 'baseComponent', ojNavigationListMeta);
oj.CustomElementBridge.register('oj-navigation-list', {'metadata': oj.CustomElementBridge.getMetadata('oj-navigation-list')});
  
var ojTabBarMeta = {
  "properties": {
    "currentItem": {
      "type": "any",
      "writeback": true
    },
    "data": {},
    "display": {
      "type": "string",
      "enumValues": ["all", "icons"]
    },
    "reorderable":{
      "type": "string",
      "enumValues": ["enabled","disabled"]
    },
    "truncation":{
      "type": "string",
      "enumValues": ["none","progressive"]
    },
    "edge": {
      "type": "string",
      "enumValues": ["top", "start", "end", "bottom"]
    },
    "item": {
      "type": "object",
      "properties": {
        "renderer": {},
        "selectable": {}
      }
    },
    "overflow": {
      "type": "string",
      "enumValues": ["popup", "hidden"]
    },
    "selection": {
      "type": "any",
      "writeback": true
    },
    "translations": {
      "type": "Object",
      "properties": {
        "accessibleReorderAfterItem": {
          "type": "string",
          "value": "After {item}"
        },
        "accessibleReorderBeforeItem": {
          "type": "string",
          "value": "Before {item}"
        },
        "accessibleReorderTouchInstructionText": {
          "type": "string",
          "value": "Double tap and hold.  Wait for the sound then drag to rearrange."
        },
        "labelCut": {
          "type": "string",
          "value": "Cut"
        },
        "labelPasteAfter": {
          "type": "string",
          "value": "Paste After"
        },
        "labelPasteBefore": {
          "type": "string",
          "value": "Paste Before"
        },
        "labelRemove": {
          "type": "string",
          "value": "Remove"
        },
        "msgFetchingData": {
          "type": "string",
          "value": "Fetching Data..."
        },
        "msgNoData": {
          "type": "string",
          "value": "No items to display."
        },
        "overflowItemLabel": {
          "type": "string",
          "value": "More"
        },
        "removeCueText": {
          "type": "string",
          "value": "Removable"
        },
        "selectedLabel": {
          "type": "string",
          "value": "selected"
        }
      }
    }
  },
  "events": {
    "animateEnd": {},
    "animateStart": {},
    "beforeCurrentItem": {},
    "beforeSelect": {},
    "beforeRemove": {},
    "remove": {},
    "beforeDeselect": {},
    "deselect": {},
    "reorder": {}
  },
  "methods": {
    "getContextByNode": {}
  },
  "extension": {
    _WIDGET_NAME: "ojNavigationList"
  }
};

oj.CustomElementBridge.registerMetadata('oj-tab-bar', 'baseComponent', ojTabBarMeta);
oj.CustomElementBridge.register('oj-tab-bar', {
    'metadata': oj.CustomElementBridge.getMetadata('oj-tab-bar')});

})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * Convenient class that represents an empty node set
 * @param {*} parent the parent key
 * @param {number} start the start index
 * @constructor
 * @since 1.0
 * @ojtsignore
 * @export
 */
oj.EmptyNodeSet = function(parent, start)
{
    this.m_parent = parent;
    this.m_start = start;
};

/**
 * Gets the parent
 * @return {*} the key of the parent.
 * @export
 */
oj.EmptyNodeSet.prototype.getParent = function()
{
    return this.m_parent;
};

/**
 * Gets the start index of the result set.  
 * @return {number} the start index of the result set.  
 * @export
 */
oj.EmptyNodeSet.prototype.getStart = function()
{
    return this.m_start;
};

/**
 * Gets the actual count of the result set.  
 * @return {number} the actual count of the result set.  
 * @export
 */
oj.EmptyNodeSet.prototype.getCount = function()
{
    return 0;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the index of the node/row in which we want to retrieve the data from.  
 * @return {null} the data for the specified index.
 * @export
 */
oj.EmptyNodeSet.prototype.getData = function(index)
{
    return null;
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source must return are:
 *  1) key - Object, the key of the node/row.
 *  2) state - state of the node, valid values are 'expanded', 'collapsed', 'leaf'. 
 *  3) depth - number, the depth of the node/row. 
 * @param {number} index the index of the node/row in which we want to retrieve the metadata from.  
 * @return {null} the metadata object for the specific index.
 * @export
 */
oj.EmptyNodeSet.prototype.getMetadata = function(index)
{
    return null;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * Flattens a hierarchical node set, which can happen in node set returned from
 * fetchDescendants call.
 * @param {Object} nodeSet the node set to flatten
 * @param {number=} actualStart in the fetch descendant case the result set would
 *        be a subset of the node set.  This param gives the exact start index in the
 *        wrapped node set where the result should start.
 * @constructor
 * @since 1.0
 * @ojtsignore
 * @export
 */
oj.FlattenedNodeSet = function(nodeSet, actualStart)
{
    this.m_nodeSet = nodeSet;
    this.m_start = actualStart;
};

/**
 * Gets the parent
 * @return {Object} the key of the parent.
 * @export
 */
oj.FlattenedNodeSet.prototype.getParent = function()
{
    return this.m_nodeSet.getParent();
};

/**
 * Gets the start index of the result set.  
 * @return {number} the start index of the result set.  
 * @export
 */
oj.FlattenedNodeSet.prototype.getStart = function()
{
    // if explicit start index is specified, use it, otherwise
    // delegate to wrapped node set
    if (this.m_start != undefined)
    {
        return this.m_start;
    }
    else
    {
        return this.m_nodeSet.getStart();
    }
};

/**
 * Gets the actual count of the result set.  
 * @return {number} the actual count of the result set.  
 * @export
 */
oj.FlattenedNodeSet.prototype.getCount = function()
{
    // see if it's calculated already
    if (this.m_count === undefined)
    {
        this.m_count = this._getCount(this.m_nodeSet, 0);

        // if explicit start is specified (subset), need to take that into
        // account when calculating total count
        if (this.m_start != undefined)
        {
            this.m_count = this.m_count - this.m_start;
        }
    }

    return this.m_count;
};

/**
 * Recursive function to calculate the total number of nodes in the node set.
 * @param {Object} nodeSet the node set to calculate count
 * @param {number} total the total number of nodes so far 
 * @return {number} the total number of nodes
 * @private
 */
oj.FlattenedNodeSet.prototype._getCount = function(nodeSet, total)
{
    var start, count, i, child;

    start = nodeSet.getStart();
    count = nodeSet.getCount();
    total = total + count;

    // if there's child node set
    if (nodeSet.getChildNodeSet)
    {
        for (i=0; i<count; i++)
        {
            child = nodeSet.getChildNodeSet(i+start);
            if (child != null)
                total = this._getCount(child, total); 
        }
    }

    return total;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the index of the node/row in which we want to retrieve the data from.  
 * @return {Object} the data for the specified index.
 * @export
 */
oj.FlattenedNodeSet.prototype.getData = function(index)
{
    return this._getDataOrMetadata(this.m_nodeSet, index, {'index': this.m_nodeSet.getStart()}, this._getData);
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source must return are:
 *  1) key - Object, the key of the node/row.
 *  2) state - state of the node, valid values are 'expanded', 'collapsed', 'leaf'. 
 *  3) depth - number, the depth of the node/row. 
 * @param {number} index the index of the node/row in which we want to retrieve the metadata from.  
 * @return {Object} the metadata object for the specific index.
 * @export
 */
oj.FlattenedNodeSet.prototype.getMetadata = function(index)
{
    return this._getDataOrMetadata(this.m_nodeSet, index, {'index': this.m_nodeSet.getStart()}, this._getMetadata);
};

/**
 * Callback function to retrieve metadata of specified index in node set
 * @param {Object} nodeSet the node set to retrieve metadata from
 * @param {number} index the index to retrieve metadata from
 * @return {Object} the metadata
 * @private
 */
oj.FlattenedNodeSet.prototype._getMetadata = function(nodeSet, index)
{
    return nodeSet.getMetadata(index);
};

/**
 * Callback function to retrieve data of specified index in node set
 * @param {Object} nodeSet the node set to retrieve data from
 * @param {number} index the index to retrieve data from
 * @return {Object} the data
 * @private
 */
oj.FlattenedNodeSet.prototype._getData = function(nodeSet, index)
{
    return nodeSet.getData(index);
};

/**
 * Retrieve data or metadata (depending on callback function) from the node set
 * @param {Object} nodeSet the node set to retrieve data from
 * @param {number} index the index to retrieve data from
 * @param {Object} current contains the current index keep track by the method
 * @param {function(Object, number)} func the callback function to retrieve data or metadata
 * @return {Object} the data or metadata
 * @private
 */
oj.FlattenedNodeSet.prototype._getDataOrMetadata = function(nodeSet, index, current, func)
{
    var start, count, i, currIndex, child, result;

    // walk the node set recursively until we found the index
    start = nodeSet.getStart();
    count = nodeSet.getCount();
    for (i=0; i<count; i++)
    {
        currIndex = current['index'];
        // found the index
        if (currIndex === index)
            return func.call(this, nodeSet, i+start);

        current['index'] = currIndex+1;
        // if there's child node set
        if (nodeSet.getChildNodeSet)
        {
            child = nodeSet.getChildNodeSet(i+start);
            if (child != null)
            {
                result = this._getDataOrMetadata(child, index, current, func); 
                if (result != null)
                    return result;
            }
        }
    }       

    return null;
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * Combines two NodeSets together into one.
 * @param {Object} nodeSet1 the first node set
 * @param {Object} nodeSet2 the second node set
 * @param {*} mergeAt the row key on the first node set where the second node set is merge to 
 * @constructor
 * @since 1.0
 * @ojtsignore
 * @export
 */
oj.MergedNodeSet = function(nodeSet1, nodeSet2, mergeAt)
{
    this.m_nodeSet1 = nodeSet1;
    this.m_nodeSet2 = nodeSet2;
    this.m_mergeAt = this._findIndex(mergeAt);
};

/**
 * Retrieve the index of the key within the first node set
 * Which is going to be the index where the two node set merge
 * @param {*} key the key to find the index
 * @return {number} the index of the key within the first node set, if index is not found, then the last index of the first node set is returned.
 * @private
 */
oj.MergedNodeSet.prototype._findIndex = function(key)
{
    var start, end, i, rowKey;

    start = this.m_nodeSet1.getStart();
    end = start + this.m_nodeSet1.getCount();
    for (i=start; i<end; i++)
    {
        rowKey = this.m_nodeSet1.getMetadata(i)['key'];
        if (key === rowKey)
        {
            return i;
        }
    }

    // if the point cannot be found, the merge happens at the end
    return (end-1);
};

/**
 * Gets the parent
 * @return {*} the key of the parent.
 * @export
 */
oj.MergedNodeSet.prototype.getParent = function()
{
    // returns the parent of the top node set
    return this.m_nodeSet1.getParent();
};

/**
 * Gets the start index of the result set.  
 * @return {number} the start index of the result set.  
 * @export
 */
oj.MergedNodeSet.prototype.getStart = function()
{
    // returns the start of the top node set
    return this.m_nodeSet1.getStart();
};

/**
 * Gets the actual count of the result set.  
 * @return {number} the actual count of the result set.  
 * @export
 */
oj.MergedNodeSet.prototype.getCount = function()
{
    // return the total count of both node sets
    return this.m_nodeSet1.getCount() + this.m_nodeSet2.getCount();
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the index of the node/row in which we want to retrieve the data from.  
 * @return {Object} the data for the specified index.
 * @export
 */
oj.MergedNodeSet.prototype.getData = function(index)
{
    var result, set, relIndex;

    result = this._getRelativeIndex(index);
    set = result['set'];
    relIndex = result['index'];

    return set.getData(relIndex);
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source must return are:
 *  1) key - Object, the key of the node/row.
 *  2) state - state of the node, valid values are 'expanded', 'collapsed', 'leaf'. 
 *  3) depth - number, the depth of the node/row. 
 * @param {number} index the index of the node/row in which we want to retrieve the metadata from.  
 * @return {Object} the metadata object for the specific index.
 * @export
 */
oj.MergedNodeSet.prototype.getMetadata = function(index)
{
    var result, set, relIndex;

    result = this._getRelativeIndex(index);
    set = result['set'];
    relIndex = result['index'];

    return set.getMetadata(relIndex);
};

/**
 * Calculate the relative index to the appropriate node set based on where the
 * merge point is.
 * @private
 */
oj.MergedNodeSet.prototype._getRelativeIndex = function(index)
{
    if (index <= this.m_mergeAt)
    {
        return {'set': this.m_nodeSet1, 'index': index};
    }
    else
    {
        var count = this.m_nodeSet2.getCount();
        var end = this.m_mergeAt + count;
        if (index > end)
        {
            // first set
            return {'set': this.m_nodeSet1, 'index': index - count};
        }   
        else
        {
            // second set, do not assume the second node set is zero indexed
            return {'set': this.m_nodeSet2, 'index': index - this.m_mergeAt - 1 + this.m_nodeSet2.getStart()};
        }
    }
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * Wraps around the NodeSet to provide additional metadata
 * @param {Object} nodeSet the node set to wrap
 * @param {function(Object, Object)} metadataCallback callback to inject additional metadata information
 * @param {Object=} range the requested range
 * @param {Array=} collapsedKeys an array of the collapsedKeys that should be passed in the expandAll case only
 * @constructor
 * @since 1.0
 * @ojtsignore
 * @export
 */
oj.NodeSetWrapper = function(nodeSet, metadataCallback, range, collapsedKeys)
{
    this.m_nodeSet = nodeSet;
    this.m_callback = metadataCallback;
    this.m_range = range;
    this.m_collapsedKeys = collapsedKeys;    
};

/**
 * Gets the parent
 * @return {Object} the key of the parent.
 * @export
 */
oj.NodeSetWrapper.prototype.getParent = function()
{
    return this.m_nodeSet.getParent();
};

/**
 * Gets the start index of the result set.  
 * @return {number} the start index of the result set.  
 * @export
 */
oj.NodeSetWrapper.prototype.getStart = function()
{
    // if the requested start is a subset of the result set, adjust
    // accordingly
    if (this.m_range != null)
    {
        return this.m_range['start'];
    }
    else
    {
        return this.m_nodeSet.getStart();
    }
};

/**
 * Gets the actual count of the result set.  
 * @return {number} the actual count of the result set.  
 * @export
 */
oj.NodeSetWrapper.prototype.getCount = function()
{
    var nodeStart, nodeCount;

    nodeStart = this.m_nodeSet.getStart();
    nodeCount = this.m_nodeSet.getCount();

    // if the requested start is a subset of the NodeSet, adjust
    // accordingly
    if (this.m_range != null)
    {
        // if the count was provided it is to limit what may be beneath it
        nodeCount = Math.min(this.m_range['count'], nodeCount);
        if (this.m_range['start'] < nodeStart)
        {
            // this is an invalid NodeSet, so just return 0
            nodeCount = 0;
        }
    }

    return nodeCount;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the index of the node/row in which we want to retrieve the data from.  
 * @return {Object} the data for the specified index.
 * @export
 */
oj.NodeSetWrapper.prototype.getData = function(index)
{
    return this.m_nodeSet.getData(this._getRelativeIndex(index));
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source must return are:
 *  1) key - Object, the key of the node/row.
 *  2) state - state of the node, valid values are 'expanded', 'collapsed', 'leaf'. 
 *  3) depth - number, the depth of the node/row. 
 * @param {number} index the index of the node/row in which we want to retrieve the metadata from.  
 * @return {Object} the metadata object for the specific index.
 * @export
 */
oj.NodeSetWrapper.prototype.getMetadata = function(index)
{
    var metadata, rowKey;

    metadata = this.m_nodeSet.getMetadata(this._getRelativeIndex(index));
    metadata['index'] = index;
    metadata['parentKey'] = this.getParent();
    rowKey = metadata['key'];

    // inject additional metadata
    this.m_callback.call(null, rowKey, metadata);

    return metadata;
};

/**
 * Gets the node set child of the specified index.
 * @param {number} index the index of the node/row in which we want to retrieve the child node set
 * @return {Object|null} the child node set representing the child tree data.
 * @export
 */
oj.NodeSetWrapper.prototype.getChildNodeSet = function(index) 
{
    var result;
    if (this.m_collapsedKeys == null || this.m_collapsedKeys.indexOf(this.m_nodeSet.getMetadata(index)['key']) == -1)
    {
        if (this.m_nodeSet.getChildNodeSet)
        {
            result = this.m_nodeSet.getChildNodeSet(index);
            if (result != null)
            {
                // wraps the child nodeset too
                return new oj.NodeSetWrapper(result, this.m_callback, null, this.m_collapsedKeys);
            }
        }
    }
    return null;
};

/**
 * Gets the relative index to the underlying nodeSet. Since the nodeSetWrapper basically adjusts the indexes relative the flattened model
 * where the wrapped node set indexes are relative to the parent node.
 * @param {number} index the index of the node/row in which we want to retrieve the index of in child node set
 * @return {number} the index of the node in the wrapped node set
 * @private
 */
oj.NodeSetWrapper.prototype._getRelativeIndex = function(index) 
{
    if (this.m_range == null)
    {
        return index;
    }
    return index - this.m_range['start'] + this.m_nodeSet.getStart();
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
 /**
 * Base class for FlattenedTreeDataGridDataSource and FlattenedTreeTableDataSource
 * @param {Object} treeDataSource the instance of TreeDataSource to flattened
 * @param {Object=} options the options set on the FlattenedDataSource
 * @constructor
 * @since 1.0
 * @export
 * @ojtsignore
 * @extends oj.DataSource
 */
oj.FlattenedTreeDataSource = function(treeDataSource, options)
{
    this.m_wrapped = treeDataSource;
    this.m_options = options || {};
    oj.FlattenedTreeDataSource.superclass.constructor.call(this);
};

/**
 * Subclass FlattenedTreeDataSource to TreeDataSource
 * @private
 */
oj.Object.createSubclass(oj.FlattenedTreeDataSource, oj.DataSource, "oj.FlattenedTreeDataSource");

/**
 * Initializes the data source.
 * @return {undefined}
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.Init = function()
{
    var expanded;

    // super
    oj.FlattenedTreeDataSource.superclass.Init.call(this);

    // we have to react if the underlying TreeDataSource has changed
    this.m_wrapped.on('change', this._handleModelEvent.bind(this));

    this.m_busy = false;

    // retrieves the fetch size against the underlying data source
    this.m_fetchSize = parseInt(this.m_options['fetchSize'], 10);
    if (isNaN(this.m_fetchSize))
    {
        this.m_fetchSize = 25;
    }
    // retrieves the maximum number of rows to fetch from the underlying data source
    // once the maximum count has been reached, this data source should stop fetching
    // until either a collapse occurs or a delete model change event.
    this.m_maxCount = parseInt(this.m_options['maxCount'], 10);
    if (isNaN(this.m_maxCount))
    {
        this.m_maxCount = 500;
    }

    // retrieves the initial expanded row keys.  If the expanded is specified to 'all',
    // then mark that all rows should be expanded initially.
    expanded = this.m_options['expanded'];
    if (Array.isArray(expanded))
    {
        this.m_expandedKeys = expanded;
    }
    else
    {
        if (expanded === 'all')
        {
            // if expand all, we'll need to keep track of collapsed keys instead
            this.m_collapsedKeys = [];
        }
        // keep track of expanded row keys
        this.m_expandedKeys = [];
    }

    // cache to keep track of indexes
    // PRIVATE.  Subclass should never need to access this directly.
    // and implementation might change to use different form of caching strategy.
    this.m_cache = [];
};

/**
 * Handle the event
 * @param {string} eventType  event type
 * @param {Object} event  event
 * @return {boolean} Returns false if event is canceled
 * @export
 */
oj.FlattenedTreeDataSource.prototype.handleEvent = function(eventType, event)
{
    return oj.FlattenedTreeDataSource.superclass.handleEvent.call(this, eventType, event);
};

/**
 * Destroy the data source.
 * @export
 * @return {undefined}
 */
oj.FlattenedTreeDataSource.prototype.Destroy = function()
{
    // free internal cache
    delete this.m_cache;
    delete this.m_expandedKeys;
    delete this.m_collapsedKeys;
    if (this.m_queue)
    {
        delete this.m_queue;
    }

    // unload listener
    this.m_wrapped.off('change');

    // delegate to underlying data source
    if (this.m_wrapped.Destroy)
    {
        this.m_wrapped.Destroy();
    }
};

/**
 * Retrieves the fetch size
 * @return {number} the fetch size
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.getFetchSize = function()
{
    return this.m_fetchSize;
};

/**
 * Retrieves the max count
 * @return {number} the max count
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.getMaxCount = function()
{
    return this.m_maxCount;
};

/**
 * Retrieves the expanded row keys
 * @return {Array.<Object>|string} an array of expanded row keys or 'all' if 
 *         all rows are expanded.
 * @export
 */
oj.FlattenedTreeDataSource.prototype.getExpandedKeys = function()
{
    return this.m_expandedKeys;
};

/**
 * Retrieves the value of the specified option.
 * @param {string} option the option to retrieve the value.
 * @return {Object} the value of the specified option.  Returns null if the
 *         value is null or if the option is not recognized.
 * @export
 */
oj.FlattenedTreeDataSource.prototype.getOption = function(option)
{
    if (this.m_options != null)
    {
        return this.m_options[option];
    }

    // unrecognized option or no options set
    return null;
};

/**
 * Retrieves the underlying TreeDataSource.
 * @return {Object} the underlying oj.TreeDataSource.
 * @export
 */
oj.FlattenedTreeDataSource.prototype.getWrappedDataSource = function()
{
    return this.m_wrapped;
};

/**
 * Determine the actual fetch size to use.
 * @param {number} count the child count of the parent node to fetch on.
 * @private
 */
oj.FlattenedTreeDataSource.prototype._getFetchSizeToUse = function(count)
{
    var fetchSize, maxCount;

    fetchSize = this.getFetchSize();
    maxCount = this.getMaxCount();

    if (fetchSize === -1)
    {
        if (count === -1)
        {
            return maxCount;
        }
        return count;
    }
    else
    {
        if (count === -1)
        {
            return Math.min(fetchSize, maxCount);
        }
        return fetchSize;
    } 
};

/**
 * Fetch a range of rows from the underlying data source.  This is a convenient method that
 * the subclasses should use to fetch from the underlying TreeDataSource.  This method will take
 * care of the index mapping between a flattened range to tree indexes.
 * @param {Object} range the range of rows to fetch.  This is the range in a flattened view.
 * @property {number} range.start the start of the range in a flattened view
 * @property {number} range.count the number of rows to fetch
 * @protected
 */ 
oj.FlattenedTreeDataSource.prototype.fetchRows = function(range, callbacks)
{
    this.m_busy = true;

    // check if we should fetch rows from descendants result set or walk the tree
    // to retrieve children
    if (this._isExpandAll())
    {
        this._fetchRowsFromDescendants(range, callbacks);
    }
    else
    {
        this._fetchRowsFromChildren(range, callbacks);
    }
};

/**
 * Fetch a range of rows from the underlying TreeDataSource.  
 * @param {Object} range the range of rows to fetch.  This is the range in a flattened view.
 * @property {number} range.start the start of the range in a flattened view
 * @property {number} range.count the number of rows to fetch
 * @protected
 */ 
oj.FlattenedTreeDataSource.prototype._fetchRowsFromChildren = function(range, callbacks)
{
    var maxFetchSize, lastEntry, parent, count, index, depth, processed, nodeSet, fetchSize, fetchRange, lastEntryKey, lastEntryCount;

    
    // this condition should always be true since in high watermark scrolling we are
    // always asking for rows after the current last row
    if (range['start'] > this._getLastIndex())
    {
        maxFetchSize = this._getMaxFetchSize();
        // initial fetch
        if (this._getLastIndex() < 0)
        {
            fetchRange = {};
            fetchRange['start'] = range['start'];
            // adjust fetch count if neccessary
            fetchRange['count'] = Math.min(maxFetchSize, range['count']);
            this.m_wrapped.fetchChildren(null, fetchRange, {"success": function(nodeSet){this._handleFetchSuccess(nodeSet, null, 0, range, fetchRange, 0, callbacks);}.bind(this), "error": function(status){this._handleFetchError(status, callbacks);}.bind(this)});

            return;
        }
        else if (maxFetchSize > 0)
        {
            lastEntry = this._getLastEntry();
            parent = lastEntry['parent'];
            count = this.m_wrapped.getChildCount(parent);
            index = lastEntry['index'];
            depth = lastEntry['depth'];

            // see if we are fetching within the parent
            if (count === -1 || index < count-1)
            {
                fetchSize = this._getFetchSizeToUse(count);
                fetchRange = {};
                fetchRange['start'] = index+1;
                if (count === -1)
                {
                    fetchRange['count'] = Math.min(fetchSize, range['count']);
                }
                else
                {
                    fetchRange['count'] = Math.min(maxFetchSize, Math.min(Math.min(fetchSize, range['count']), count - fetchRange['start']));
                }
                this.m_wrapped.fetchChildren(parent, fetchRange, {"success": function(nodeSet){this._handleFetchSuccess(nodeSet, parent, depth, range, fetchRange, count, callbacks);}.bind(this), "error": function(status){this._handleFetchError(status, callbacks);}.bind(this)});
            }
            // if this is the last child within the parent, then we still need to see if there are ancestors available to fetch
            else
            {
                lastEntryKey = lastEntry['key'];
                lastEntryCount = this.m_wrapped.getChildCount(lastEntryKey);
                if (this._isExpanded(lastEntryKey) && (lastEntryCount === -1 || lastEntryCount > 0))
                {
                    // if the last entry was expanded and has children fetch its children
                    processed = this._fetchFromAncestors(lastEntry, depth + 1, range, callbacks, maxFetchSize);
                }                       
                else
                {
                    // fetch size is greater than the number of children remaining to fetch
                    // so we'll need to go up the path (recursively if necessary) and see if
                    // if we need to fetch from ancestors.
                    processed = this._fetchFromAncestors(parent, depth, range, callbacks, maxFetchSize);
                }
                if (!processed)
                {
                    // nothing is used from node set, just return a empty node set
                    nodeSet = new oj.EmptyNodeSet(null, range['start']);
                    // invoke original success callback
                    if (callbacks != null && callbacks['success'] != null)
                    {
                        callbacks['success'].call(null, nodeSet);
                    }
                    // busy not handled by the _fetchFromAncestors in this case
                    this.m_busy = false; 
                }
            }
            return;
        }
    }

    // the only case we'll ended up here is if the max count has been reached or
    // for some reason the caller is asking for count = 0
    this.handleMaxCountReached(range, callbacks);
    // busy can't be set by max count reached because it can be overwritten by wrapped datasource
    this.m_busy = false;
};

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference row's current children.
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @export
 */ 
oj.FlattenedTreeDataSource.prototype.moveOK = function(rowToMove, referenceRow, position)
{
    return this.m_wrapped.moveOK(rowToMove, referenceRow, position);
};

/**
 * Moves a row from one location to another (different position within the same parent or a completely different parent)
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference row's current children.
 * @param {Object=} callbacks the callback to invoke when the move completed.
 * @property {function()} callbacks.success the callback to invoke when the move completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when move failed.
 * @return {undefined}
 * @export
 */
oj.FlattenedTreeDataSource.prototype.move = function(rowToMove, referenceRow, position, callbacks)
{
    this.m_wrapped.move(rowToMove, referenceRow, position, callbacks);
};

/**
 * Determine the maximum possible fetch size.
 * @return {number} the maximum fetch size
 * @private
 */
oj.FlattenedTreeDataSource.prototype._getMaxFetchSize = function()
{
    return this.getMaxCount() - (this._getLastIndex()+1);
};

/**
 * Process error callback for fetchChildren operation before handing it back to original caller.
 * @param {Object} status the error status
 * @param {Object} callbacks the original callbacks passed to the fetch operation
 * @private
 */
oj.FlattenedTreeDataSource.prototype._handleFetchError = function(status, callbacks)
{
    if (callbacks != null && callbacks['error'] != null)
    {
        callbacks['error'].call(null, status);
    }

    this.m_busy = false;
};

/**
 * Process success callback for fetchChildren operation before handing it back to original caller.
 * @param {Object} nodeSet the set of fetched nodes
 * @param {Object} parent the parent key of the fetch operation
 * @param {number} depth the depth of the nodes
 * @param {Object} originalRange the request range for the fetch operation
 * @param {Object} requestedRange the request range for the fetch operation
 * @param {number} count the child count of the parent, -1 if count is unknown
 * @param {Object} callbacks the original callbacks passed to the fetch operation
 * @param {Object=} options the original callbacks passed to the fetch operation
 * @private
 */
oj.FlattenedTreeDataSource.prototype._handleFetchSuccess = function(nodeSet, parent, depth, originalRange, requestedRange, count, callbacks, options)
{
    var toExpand, flattenedRange, queue, prevNodeSetInfo;
    toExpand = [];

    // first process the node set to get the rows to expand with the indexes        
    this._processNodeSet(nodeSet, parent, depth, toExpand);

    flattenedRange = {start: originalRange['start'], count: nodeSet.getCount()};
    nodeSet = new oj.NodeSetWrapper(nodeSet, this.insertMetadata.bind(this), flattenedRange);
    
    // expand the fetched nodes to get the node set as full as possible from the original fetch
    if (toExpand.length !== 0)
    {                     
        // there are rows to expand, so we'll need to combine the nodeset after
        // we got the expanded nodeset
        queue = [];
        queue.push(toExpand);

        // we'll reuse the syncExpandRows method, which is used to combine nested expanding
        // nodeset, the only difference is we'll include the callbacks here, see handleExpandSuccess method
        prevNodeSetInfo = {};
        prevNodeSetInfo['callbacks'] = 
        {
            "success": function(newNodeSet){this._verifyFetchResults(newNodeSet, parent, depth, originalRange, requestedRange, count, callbacks, options);}.bind(this), 
            "error": function(status){this._handleFetchError(status, callbacks);}.bind(this)
        };
        prevNodeSetInfo['nodeSet'] = nodeSet;
        prevNodeSetInfo['keys'] = [];

        this._syncExpandRows(queue, prevNodeSetInfo);
    }
    else
    {
        this._verifyFetchResults(nodeSet, parent, depth, originalRange, requestedRange, count, callbacks, options)
    }    
};

/**
 * Process success callback from fetchChildren, fetchFromAncestors or expand after a fetch children
 * @param {Object} nodeSet the set of fetched nodes
 * @param {Object} parent the parent key of the fetch operation
 * @param {number} depth the depth of the nodes
 * @param {Object} originalRange the request range for the fetch operation
 * @param {Object} requestedRange the request range for the fetch operation
 * @param {number} count the child count of the parent, -1 if count is unknown
 * @param {Object} callbacks the original callbacks passed to the fetch operation
 * @param {Object=} options the original callbacks passed to the fetch operation
 * @private
 */
oj.FlattenedTreeDataSource.prototype._verifyFetchResults = function(nodeSet, parent, depth, originalRange, requestedRange, count, callbacks, options)
{
    var prevNodeSet, lastNodeIndex, lastNodeKey, mergedNodeSet, remainingRange, remainingOptions, processed;
;
    if (options != undefined)
    {
        // if there is a previousNodeSet merge it with the new one to try and fulfill the original fetch size
        prevNodeSet = options['prevNodeSet'];
        if (prevNodeSet != undefined)
        {
            lastNodeIndex = prevNodeSet.getStart() + prevNodeSet.getCount() - 1;
            lastNodeKey = prevNodeSet.getMetadata(lastNodeIndex)['key'];
            mergedNodeSet = new oj.MergedNodeSet(prevNodeSet, nodeSet, lastNodeKey);
        }
    }

    // if the nodeSet contains less nodes than originally requested, attempt to fetch the remainder from the ancestors
    if (nodeSet.getCount() < originalRange['count'] && parent != null && depth > 0)
    {
        remainingRange = {};
        remainingRange['start'] = originalRange['start'] + nodeSet.getCount();
        remainingRange['count'] = originalRange['count'] - nodeSet.getCount();
        
        remainingOptions = {};
        remainingOptions['prevNodeSet'] = mergedNodeSet == null ? nodeSet : mergedNodeSet;
        
        processed = this._fetchFromAncestors(parent, depth, remainingRange, callbacks, undefined, remainingOptions);        
    }
    // if we overfetched because of expanded children then trim the nodeSet by wrapping the nodeSet in the correct range
    else if (nodeSet.getCount() > originalRange['count'])
    {
        var difference = nodeSet.getCount() - originalRange['count'];
        if (mergedNodeSet != null)
        {
            mergedNodeSet = new oj.NodeSetWrapper(mergedNodeSet, this.insertMetadata.bind(this), {'start': mergedNodeSet.getStart(), 'count': (mergedNodeSet.getCount() - difference)});
            // remove entries that will not be passed back via the wrapped range
            this._removeEntry(mergedNodeSet.getStart() + mergedNodeSet.getCount(), difference);
        }
        else
        {
            nodeSet = new oj.NodeSetWrapper(nodeSet, this.insertMetadata.bind(this), {'start': nodeSet.getStart(), 'count': (nodeSet.getCount() - difference)});
            // remove entries that will not be passed back via the wrapped range
            this._removeEntry(nodeSet.getStart() + nodeSet.getCount(), difference);            
        }
    }

    if (!processed)
    {
        if (callbacks != null && callbacks['success'] != null)
        {
            callbacks['success'].call(null, mergedNodeSet == null ? nodeSet : mergedNodeSet);
        }
    }

    this.m_busy = false;
};

/**
 * Returns the number of children for a specified parent.  If the value returned is not >= 0 then it is automatically assumed
 * that the child count is unknown.
 * @param {*} parent the parent key.  Specify null if inquiring child count of the root.
 * @return {number} the number of children for the specified parent.
 * @export
 */
oj.FlattenedTreeDataSource.prototype.getChildCount = function(parent)
{
    return this.m_wrapped.getChildCount(parent);
};

/**
 * Go up ancestors and fetch build up fetch requests (if possible) until the fetch size is met.
 * @param {Object} parent the parent key of the fetch operation
 * @param {number} depth the depth of the nodes
 * @param {Object} callbacks the original callbacks passed to the fetch operation
 * @param {number=} maxFetchSize maximum fetch size, optional
 * @param {Object=} options 
 * @return {boolean} true if results are fetched, false if nothing is fetched
 * @private
 */
oj.FlattenedTreeDataSource.prototype._fetchFromAncestors = function(parent, depth, range, callbacks, maxFetchSize, options)
{
    var remainToFetch, current, i, currEntry, currDepth, count, index, countUnknown, fetchSize, batchFetchOptions, fetchRange, fetchedChildren;

    fetchedChildren = false;
    
    if (maxFetchSize === undefined)
    {
        maxFetchSize = this._getMaxFetchSize();
    }

    // fetch size is greater than the number of children remaining to fetch
    // so we'll need to go up the path (recursively if necessary) and see if
    // if we need to fetch from ancestors.
    if (this._isBatchFetching())
    {
        batchFetchOptions = {'queueOnly': true};
    }

    fetchSize = this._getFetchSizeToUse(-1);
    current = this._getLastIndex();
    // adjusted for loop i = current-1 to i = current.
    // this._getLastIndex automatically performs a -1
    // so no need to set that again. 
    for (i=current; i>=0; i--)
    {
        currEntry = this._getEntry(i);
        currDepth = currEntry['depth'];
        if (currDepth < depth)
        {
            parent = currEntry['parent'];
            count = this.m_wrapped.getChildCount(parent);
            index = currEntry['index'];

            countUnknown = (count === -1);
            if (countUnknown || index < count-1)
            {
                fetchRange = {};
                fetchRange['start'] = index+1;
                if (countUnknown)
                {
                    fetchRange['count'] = Math.min(maxFetchSize, Math.max(0, fetchSize));
                    // if count is unknown, we cannot do batch fetch
                    batchFetchOptions = undefined;
                    // stop going up parents
                }
                else
                {
                    fetchRange['count'] = Math.min(maxFetchSize, Math.min(fetchSize, count - fetchRange['start']));
                }

                // if there's nothing to fetch, quit
                if (fetchRange['count'] == 0)
                {
                    break;
                }
                
                // it's always attached at the end
                this.m_wrapped.fetchChildren(parent, fetchRange, {"success": function(nodeSet){this._handleFetchSuccess(nodeSet, parent, currDepth, range, fetchRange, count, callbacks, options);}.bind(this), "error": function(status){this._handleFetchError(status, callbacks);}.bind(this)}, batchFetchOptions);
                
                fetchedChildren = true;
                
                break;
            }
            else
            {
                depth -= 1;
            }
        }
    }            

    // if batching is used, fire a final fetch children call to flush the queue
    if (batchFetchOptions != null)
    {
        this.m_wrapped.fetchChildren(parent, {'start': range['count'], 'count': 0}, {"success": function(nodeSet){this._handleFetchSuccess(nodeSet, parent, currDepth, range, fetchRange, count, callbacks, options);}.bind(this), "error": function(status){this._handleFetchError(status, callbacks);}.bind(this)});
        fetchedChildren = true;
    }

    // return false if no results are fetched
    return fetchedChildren;
};

/**
 * Walk the node set and do whatever processing is neccessary.
 * @param {Object} nodeSet the node set to process
 * @param {Object} parent the parent key of the nodes
 * @param {number} depth the depth of the nodes
 * @param {Array.<Object>=} toExpand the set of keys to be expand.  It is populated by this method.
 * @private
 */
oj.FlattenedTreeDataSource.prototype._processNodeSet = function(nodeSet, parent, depth, toExpand)
{
    var nodeStart, nodeCount, i, metadata, key;

    nodeStart = nodeSet.getStart();
    nodeCount = nodeSet.getCount();

    // walk the node set and populate the internal cache
    for (i=0; i<nodeCount; i++)
    {
        metadata = nodeSet.getMetadata(nodeStart+i);
        key = metadata['key'];

        this._addEntry(key, depth, nodeStart+i, parent);

        if (this._isExpanded(key))
        {
            // keep track of rows that needs to expand later
            toExpand.push(key);
        }
    }
};

/**
 * A hook for FlattenedTreeDataSource to inject additional metadata into the NodeSet
 * @param {Object} key the row key identifying the row
 * @param {Object} metadata the existing metadata to inject into
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.insertMetadata = function(key, metadata)
{
    if (this._isExpanded(key) && !metadata['leaf'])
    {
        // also update metadata with state info
        metadata['state'] = 'expanded';
    }
    else
    {
        // include state metadata for row expander to render correct icon
        if (metadata['leaf'])
        {
            metadata['state'] = 'leaf';
        }
        else
        {
            metadata['state'] = 'collapsed';
        }
    }
};

/**
 * Fetch a range of rows from the underlying TreeDataSource using the fetchDescendants method.
 * @param {Object} range the range of rows to fetch.  This is the range in a flattened view.
 * @property {number} range.start the start of the range in a flattened view
 * @property {number} range.count the number of rows to fetch
 * @protected
 */ 
oj.FlattenedTreeDataSource.prototype._fetchRowsFromDescendants = function(range, callbacks)
{
    // give implementation a hint of maximum to fetch, implementation can choose to ignore it
    var options = {'maxCount': this.getMaxCount()};

    // give implementation a hint of where to start, implementation can choose to ignore it
    if (this._getLastIndex() >= 0)
    {
        options['start'] = this._getEntry(this._getLastIndex())['key'];
    }

    // invoke method on TreeDataSource
    this.m_wrapped.fetchDescendants(null, {"success": function(nodeSet){this._handleFetchDescendantsSuccess(nodeSet, range, callbacks);}.bind(this), "error": function(status){this._handleFetchError(status, callbacks);}.bind(this)}, options);
};

/**
 * Returns the current sort criteria of the tree data.
 * @return {Object} the current sort criteria.  It should contain the following properties: key, direction where
 *         criteria.key the key identifying the attribute (column) to sort on.  Value is null if it's not sorted.
 *         criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @export
 */
oj.FlattenedTreeDataSource.prototype.getSortCriteria = function()
{
    return this.m_wrapped.getSortCriteria();
};
    
/**
 * Process success callback for fetchDescendants operation before handing it back to original caller.
 * @param {Object} nodeSet the set of fetched nodes
 * @param {Object} range the request range for the fetch operation
 * @param {Object} callbacks the original callbacks passed to the fetch operation
 * @private
 */
oj.FlattenedTreeDataSource.prototype._handleFetchDescendantsSuccess = function(nodeSet, range, callbacks)
{
    var maxFetchSize, count, lastEntry, options, actualStart;

    // this condition should always be true since in high watermark scrolling we are
    // always asking for rows after the current last row
    if (range['start'] > this._getLastIndex())
    {
        maxFetchSize = this._getMaxFetchSize();
        count = Math.min(maxFetchSize, range['count']);

        // wrap it to inject additional metadata
        nodeSet = new oj.NodeSetWrapper(nodeSet, this.insertMetadata.bind(this), null, this.m_collapsedKeys);

        if (this._getLastIndex() >= 0)
        {
            // in fetchDescendants case, the result node set would probably contains more than what
            // we would return.  The issue is we can't really use range to filter the set since the 
            // range in the current view does not map one-to-one to the expand all node set as some 
            // node might have been collapsed before the fetch.
            // the solution is to use the last cached entry to find where new data starts in the
            // result node set, and use range count to limit what to return
            lastEntry = this._getLastEntry();
            options = {'index': 0, 'found': false, 'count': 0};
            this._processDescendantsNodeSet(nodeSet, null, 0, lastEntry, count, options);
            actualStart = options['index'] + 1;
        }
        else
        {
            // initial fetch case, just specify the count to limit result
            options = {'count': 0};
            this._processDescendantsNodeSet(nodeSet, null, 0, null, count, options);
            actualStart = 0;
        }

        if (callbacks != null && callbacks['success'] != null)
        {
            if (options != null)
            {
                if (options['count'] === 0)
                {
                    // nothing is used from node set, just return a empty node set
                    nodeSet = new oj.EmptyNodeSet(null, range['start']);
                }
                else 
                {
                    // wraps node set with a filter that only returns nodes that
                    // have not been fetched already
                    nodeSet = new oj.FlattenedNodeSet(nodeSet, actualStart);
                }
            }
            else
            {
                nodeSet = new oj.FlattenedNodeSet(nodeSet);
            }
            callbacks['success'].call(null, nodeSet);
        }
    }
    else
    {
        // the only case we'll ended up here is if the max count has been reached or
        // for some reason the caller is asking for count = 0
        this.handleMaxCountReached(range, callbacks);
    }

    this.m_busy = false;

    // process any outstanding operations
    this._processQueue();
};

/**
 * Walk the node set and do whatever processing is neccessary.
 * @param {Object} nodeSet the node set to process
 * @param {Object} parent the parent key of the nodes
 * @param {number} depth the depth of the nodes
 * @param {Object} lastEntry the last fetched entry
 * @param {number} maxCount the maximum number of rows to process
 * @param {Object=} options this object carries information collected in this method
 * @private
 */
oj.FlattenedTreeDataSource.prototype._processDescendantsNodeSet = function(nodeSet, parent, depth, lastEntry, maxCount, options)
{
    var nodeStart, nodeCount, i, metadata, key, childNodeSet;

    nodeStart = nodeSet.getStart();
    nodeCount = nodeSet.getCount();

    // walk the node set and populate the internal cache
    for (i=0; i<nodeCount; i++)
    {
        // see if we have enough results
        if (options['count'] == maxCount)
        {
            return;
        }

        metadata = nodeSet.getMetadata(nodeStart+i);
        key = metadata['key'];

        // see if we need to check depth
        if (options['checkDepth'])
        {
            if (lastEntry['depth'] === depth)
            {
                options['found'] = true;
                options['checkDepth'] = false;
            }
        }

        if (lastEntry == null || options['found'])
        {
            this._addEntry(key, depth, nodeStart+i, parent);

            options['count'] = options['count'] + 1;

            // include state metadata for row expander
            // in the fetchDescendants case the state is always 'expanded'
            if (metadata['leaf'])
            {
                metadata['state'] = 'leaf';
            }
            else
            {
                metadata['state'] = 'expanded';
            }
        }

        // mark we found the entry in node set that matches the last key
        // the rest of node set we can start pushing to cache
        if (lastEntry != null && !options['found'])
        {
            // we'll need to also check whether the last entry is expanded (or not leaf)
            // if it is collapsed, then we can't add any nodes from the node set until
            // we found child in the node set that has the same depth
            if (key === lastEntry['key'])
            {
                if (metadata['leaf'] || this._isExpanded(key))
                {
                    options['found'] = true;
                }
                else
                {
                    // collapsed.  Mark to check the depth of the next node before
                    // setting found to true.
                    options['checkDepth'] = true;
                }
            }
            else
            {
                options['index'] = options['index'] + 1;
            }
        }

        // process child node set, if any
        if (nodeSet.getChildNodeSet && this._isExpanded(key))
        {
            childNodeSet = nodeSet.getChildNodeSet(i);
            if (childNodeSet != null)
            {
                this._processDescendantsNodeSet(childNodeSet, key, depth+1, lastEntry, maxCount, options);
            }
        }
    }
};

/**
 * Expand the specified row.
 * @param {*} rowKey the key of the row to expand
 * @export
 * @return {undefined}
 */
oj.FlattenedTreeDataSource.prototype.expand = function(rowKey)
{
    this._expand(rowKey);
};

/**
 * Expand the specified row with options
 * @param {*} rowKey the key of the row to expand
 * @param {Object=} options additional options to pass to fetchChildren method
 * @private
 */
oj.FlattenedTreeDataSource.prototype._expand = function(rowKey, options)
{
    var count, fetchSize, maxCount, refIndex, prevNodeSetInfo;

    this.m_busy = true;

    count = this.m_wrapped.getChildCount(rowKey);
    fetchSize = this._getFetchSizeToUse(count);
    maxCount = this.getMaxCount();

    // if cache is full, check if the rowKey is the last row, if it's
    // the last row do nothing
    if (this._getLastIndex()+1 === maxCount)
    {
       refIndex = this.getIndex(rowKey);
       if (refIndex == maxCount-1)
       {
           // we'll still have to return an empty nodeset to trigger done to occur in handleExpandSuccess
           this.handleExpandSuccess(rowKey, new oj.EmptyNodeSet(rowKey, 0), 0, options);
           return;
       }
    }

    // nothing to do
    if (fetchSize == 0)
    {
        // we'll still have to return an empty nodeset to trigger done to occur in handleExpandSuccess
        this.handleExpandSuccess(rowKey, new oj.EmptyNodeSet(rowKey, 0), 0, options);
        return;
    }

    this.m_wrapped.fetchChildren(rowKey, {"start": 0, "count": fetchSize}, {"success": function(nodeSet){this.handleExpandSuccess(rowKey, nodeSet, count, options);}.bind(this), "error": function(status){this.handleExpandError(rowKey, status);}.bind(this)});
};

/**
 * Process any outstanding operation
 * @private
 */
oj.FlattenedTreeDataSource.prototype._processQueue = function()
{
    var i, op;

    if (this.m_queue && this.m_queue.length > 0)
    {
        for (i=this.m_queue.length-1; i>=0; i--)
        {
            op = this.m_queue[i];
            this.collapse(op["key"]);            
        }

        // clear the queue
        this.m_queue.length = 0;
    }
};

/**
 * Queue an operation to be process later
 * @param {string} op the operation
 * @param {*} rowKey the row key where the operation is to be apply
 * @private
 */
oj.FlattenedTreeDataSource.prototype._queueOp = function(op, rowKey)
{
    if (this.m_queue == null)
    {
        this.m_queue = [];
    }

    this.m_queue.push({"op": op, "key": rowKey});
};

/**
 * Collapse the specified row.
 * @param {*} rowKey the key of the row to collapse
 * @return {undefined}
 * @export
 */
oj.FlattenedTreeDataSource.prototype.collapse = function(rowKey)
{
    var rowIndex, parent, count, depth, lastIndex, i, j, keys;

    // if in the middle of fetch, queue up the operation
    if (this.m_busy)
    {
        this._queueOp("collapse", rowKey);
        return;
    }

    rowIndex = this.getIndex(rowKey) + 1;
    parent = this._getEntry(rowIndex-1);
    // could happen if parent node has already collapsed
    if (parent == null)
    {
        return;
    }

    // keeping track of how many rows to remove
    count = 0;

    depth = parent['depth'];
    lastIndex = this._getLastIndex();
    for (j=rowIndex; j<lastIndex+1; j++)
    {
        var rowData = this._getEntry(j);
        var rowDepth = rowData['depth'];
        if (rowDepth > depth)
        {
            count = count + 1;
        }
        else if (rowDepth == depth)
        {
            break;
        }
    }

    // nothing to do
    if (count == 0)
    {
        // still should fire an event to get the icon state updated properly
        this.handleEvent("collapse", {'rowKey':rowKey});
        return;
    }

    // remove from expanded keys or add to collapsed keys
    if (this._isExpandAll())
    {
        this.m_collapsedKeys.push(rowKey);
    }
    else
    {
        this._removeExpanded(rowKey);
    }

    // remove rows from view
    keys = [];
    for (i=0; i<count; i++)
    {
        keys.push({"key": this._getEntry(rowIndex+i)['key'], "index":rowIndex+i});
    }

    // remove from cache.  Note this has to be done before firing row remove event
    // since it could cause a fetch which relies on the internal cache being up to date.
    this._removeEntry(rowIndex, count);    

    // (firing of event to view)
    this.removeRows(keys);

    // fire datasource event
    this.handleEvent("collapse", {'rowKey':rowKey});
};

/**
 * Checks whether the row key is expanded.
 * @param {*} rowKey the key of the row to inquire the state
 * @return {boolean} true if the row is/should be expanded.  False otherwise.
 * @private
 */
oj.FlattenedTreeDataSource.prototype._isExpanded = function(rowKey)
{
    if (this._isExpandAll())
    {
        if (this.m_collapsedKeys && this.m_collapsedKeys.length > 0)
        {
            // call helper method to check collapsed keys
            return (this._getCollapsedKeyIndex(rowKey) === -1);
        }
        else
        {
            // everything expanded
            return true;
        }
    }
    else
    {
        if (this.m_expandedKeys && this.m_expandedKeys.length > 0)
        {
            // call helper method to check expanded keys
            return (this._getExpandedKeyIndex(rowKey) > -1);
        }
        else
        {
            // nothing expanded
            return false;
        }
    }
};

/**
 * Helper method to retrieve the index of the row key in the set of collapsed row keys
 * @param {*} rowKey the key of the row
 * @return {number} the index of the key in the collapsed key array
 * @private
 */ 
oj.FlattenedTreeDataSource.prototype._getCollapsedKeyIndex = function(rowKey)
{
    return this._getKeyIndex(this.m_collapsedKeys, rowKey);
};

/**
 * Helper method to retrieve the index of the row key in the set of expanded row keys
 * @param {*} rowKey the key of the row
 * @return {number} the index of the key in the expanded key array
 * @private
 */ 
oj.FlattenedTreeDataSource.prototype._getExpandedKeyIndex = function(rowKey)
{
    return this._getKeyIndex(this.m_expandedKeys, rowKey);
};

/**
 * Helper method to retrieve the index of the row key in a specified array
 * @param {*} rowKey the key of the row
 * @return {number} the index of the key in the array
 * @private
 */ 
oj.FlattenedTreeDataSource.prototype._getKeyIndex = function(arr, rowKey)
{
    var i, index;

    index = -1;
    for (i=0; i<arr.length; i++)
    {
        if (arr[i] === rowKey)
        {
            index = i;
        }
    }

    return index;
}


/**
 * Remove the row key from the expanded cache
 * @param {*} rowKey the key to remove
 * @private
 */
oj.FlattenedTreeDataSource.prototype._removeExpanded = function(rowKey)
{
    var index = this._getExpandedKeyIndex(rowKey);

    // index found, remove from array
    if (index > -1)
    {
        this.m_expandedKeys.splice(index, 1);    
    }
};

/**
 * Remove the row key from the collapsed cache
 * @param {*} rowKey the key to remove
 * @private
 */
oj.FlattenedTreeDataSource.prototype._removeCollapsed = function(rowKey)
{
    var index = this._getCollapsedKeyIndex(rowKey);

    // index found, remove from array
    if (index > -1)
    {
        this.m_collapsedKeys.splice(index, 1);    
    }
};

/**
 * Callback method to handle fetch error on expand operation.
 * @param {*} rowKey the key of the expanded row
 * @param {Object} status the error status
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.handleExpandError = function(rowKey, status)
{
    // fire event, todo: should include error msg?
    this.handleEvent("expand", {'rowKey':rowKey});
};

/**
 * Callback method to handle fetch success on expand operation.
 * @param {*} rowKey the key of the expanded row
 * @param {Object} nodeSet the node set that describes the children of the expanded row
 * @param {number} childCount the total number of children the expanded row has
 * @param {Object=} options optional parameters to the method
 * @property {Object=} options.queue a queue of expanded rows remaining to process (depth first traversal)
 * @property {Object=} options.prevNodeSetInfo.nodeSet the node set from a previous expand call
 * @property {number=} options.prevNodeSetInfo.firstIndex the ref index for the FIRST expand call, this is needed when firing the insert event, where the insertion point is the first index
 * @property {Object=} options.prevNodeSetInfo.firstKey the ref row key for the FIRST expand call, this is needed when firing the insert event, where the insertion point is the first row key
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.handleExpandSuccess = function(rowKey, nodeSet, childCount, options)
{
    var refIndex, rangeStart, rowStart, rowCount, parent, depth, metadata, key, toExpand, i, j, queue, prevNodeSetInfo, done, maxCount, callbacks;

    // wrap it to inject additional metadata
    nodeSet = new oj.NodeSetWrapper(nodeSet, this.insertMetadata.bind(this));

    refIndex = this.getIndex(rowKey) + 1;
    rangeStart = refIndex;

    rowStart = nodeSet.getStart();
    rowCount = nodeSet.getCount();

    parent = this._getEntry(refIndex-1);
    depth = parent['depth']+1;

    toExpand = [];

    // go through the node set and insert an entry with info about the row into internal cache
    for (i=rowStart; i<rowCount; i++)
    {
        metadata = nodeSet.getMetadata(i);
        key = metadata['key'];
        if (this._isExpanded(key))
        {
            // expand it if the user specified it to be expand (or the
            // parent was previously collapsed before and now expanded again, 
            // the expanded child would need to be expanded also)
            toExpand.push(key);
        }

        // add to cache
        this._insertRow(refIndex, metadata, parent['key'], i, depth);

        refIndex++;
    }

    // keep track of expanded row or collapsed row for expand all case
    if (this._isExpandAll())
    {
        this._removeCollapsed(rowKey);
    }
    else
    {
        // check whether it's already in expanded keys, which is the case
        // if it is expanded by initial expansion
        if (this.m_expandedKeys.indexOf(rowKey) === -1)
        {
            this.m_expandedKeys.push(rowKey);
        }
    }

    // extract optional params
    if (options != undefined)
    {
        queue = options['queue'];
        prevNodeSetInfo = options['prevNodeSetInfo'];
    }

    // see if a previous nodeset has been set and merge with current one 
    // so that we have one nodeset that includes expanded children, a single row insert event
    // is fired and the nodeset will be in the proper order
    if (prevNodeSetInfo != undefined)
    {
        nodeSet = new oj.MergedNodeSet(prevNodeSetInfo['nodeSet'], nodeSet, rowKey);
    }

    // check if there's nothing else to expand and process queue is empty
    done = (toExpand.length == 0 && (queue === undefined || queue.length == 0));
    if (done)
    {
        // fire event to insert the expanded rows
        if (prevNodeSetInfo != undefined)
        {
            // check if this is part of a fetchRows call
            callbacks = prevNodeSetInfo['callbacks'];
            if (callbacks != null)
            {
                // invoke fetch success callback
                callbacks['success'].call(null, nodeSet);
                
                // we are done at this point, we don't fire insert events
                this.m_busy = false;
                return;
            }
            else
            {
                // use the reference insertion point from prevNodeSetInfo instead
                this.insertRows(prevNodeSetInfo['firstIndex'], prevNodeSetInfo['firstKey'], nodeSet);
            }
        }
        else
        {
            this.insertRows(rangeStart, rowKey, nodeSet);
        }

        // if child count is > fetched or child count is unknown and requested fetch count is the same as result set size, 
        // then delete all rows that comes after the reference row so that we can trigger a fetch when user scroll to the end
        // ALSO delete all rows that comes after reference row if the reference row is the last row (according to max row count)
        maxCount = this.getMaxCount();
        if ((childCount === -1 && rowCount === this.getFetchSize()) || childCount > rowCount || refIndex == maxCount)
        {
            this._deleteAllRowsBelow(refIndex);
        }
        else if (this._getLastIndex() >= maxCount)
        {
            // also clean up rows that goes beyond max row count after expand
            this._deleteAllRowsBelow(maxCount);
        }

        if (prevNodeSetInfo != undefined)
        {
            // fire expand event for each row key cached in prevNodeSetInfo
            for (j=0; j<prevNodeSetInfo['keys'].length; j++)
            {
                this.handleEvent("expand", {'rowKey':prevNodeSetInfo['keys'][j]});
            }
        }

        this.m_busy = false;

        // fire event
        this.handleEvent("expand", {'rowKey':rowKey});
    }
    else
    {
        // there are still child rows to expand
        // create queue if not yet created
        if (queue === undefined)
        {
            queue = [];
        }

        // push expanded rows to the queue
        if (toExpand.length > 0)
        {
            queue.push(toExpand);
        }

        // create prevNodeSetInfo if not yet created
        if (prevNodeSetInfo === undefined)
        {
            prevNodeSetInfo = {};
            // populate the initial insertion index and key, this is needed when we are actually firing
            // the insert event
            prevNodeSetInfo['firstIndex'] = rangeStart;
            prevNodeSetInfo['firstKey'] = rowKey;
            // cache of row keys for firing expand event when everything is done
            prevNodeSetInfo['keys'] = [];
        }

        // update the previous node set 
        prevNodeSetInfo['nodeSet'] = nodeSet;
        // update keys array for fire expand events later
        prevNodeSetInfo['keys'].push(rowKey);

        // expand any child rows that should be expanded
        this._syncExpandRows(queue, prevNodeSetInfo);
    }

    this.m_busy = false;

    if (queue && queue.length == 0)
    {
        this._processQueue();
    }
};

/**
 * Expands the specified array of rows synchronously, i.e. one will not start until the previous one is finished.
 * @param {Object} queue the queue of a set of expanded row keys remaining to process 
 * @param {Object} prevNodeSetInfo node set from the previous expand call
 * @private 
 */ 
oj.FlattenedTreeDataSource.prototype._syncExpandRows = function(queue, prevNodeSetInfo)
{
    var last, key, options;

    // peek the last set of expanded rows from queue (since we are doing depth first traversal)
    last = queue[queue.length-1];    
    // then take the first row key from the set
    key = last.shift();
    // if this is the last item in the set, we can remove the set from queue
    if (last.length === 0)
    {
        queue.pop();
    }

    this._expand(key, {'prevNodeSetInfo': prevNodeSetInfo, 'queue': queue});    
};

/**
 * Expands the specified array of rows.  Use batch fetching if supported.
 * @param {Array.<Object>} keys an array of row keys.
 * @private
 */
oj.FlattenedTreeDataSource.prototype._expandRows = function(keys)
{
    var options, i;

    // use batch fetching if supported so we'll have less trip to server.
    if (this._isBatchFetching())
    {
        options = {'queueOnly': true};
    }

    // expand each of the rows
    for (i=0; i<keys.length; i++)
    {
        // last expand should not have any options set to flush to batch queue
        if (i == keys.length-1)
        {
            this._expand(keys[i]);
        }            
        else
        {
            this._expand(keys[i], options);
        }            
    }
};

/**
 * Insert a single row of data into the cache
 * @param {number} index the index (based on flattened view) where this is inserted
 * @param {Object} metadata the metadata of the inserted node
 * @param {Object} parentKey the key of the parent node
 * @param {number} childIndex the index relative to its parent where this is inserted
 * @param {number} depth the depth of the node
 * @private
 */
oj.FlattenedTreeDataSource.prototype._insertRow = function(index, metadata, parentKey, childIndex, depth)
{
    var key, rowData;

    key = metadata['key'];

    if (index <= this._getLastIndex())    
    {
        this._addEntry(key, depth, childIndex, parentKey, index);
    }
    else
    {
        this._addEntry(key, depth, childIndex, parentKey);
    }
};

/**
 * Remove all rows below the row of the specified index including this row.
 * @param {number} index the index from which we start to delete rows 
 * @param {number=} count the number of rows to delete.  If not specified, then delete until the end.
 * @private
 */
oj.FlattenedTreeDataSource.prototype._deleteAllRowsBelow = function(index, count)
{
    var keys, event;

    if (count == undefined)
    {
        count = this._getLastIndex()+1 - index;
    }

    keys = [];
    for (var i=0; i<count; i++)
    {
        keys.push({"key": this._getEntry(index+i)['key'], "index":index+i});
    }

    // update internal cache
    this._removeEntry(index, count);    

    // fire event to remove rows from view
    this.removeRows(keys);
};

/**
 * Handles model event from underlying TreeDataSource.
 * @param {Event} event the model change event
 * @private
 */
oj.FlattenedTreeDataSource.prototype._handleModelEvent = function(event)
{
    var operation, ancestors, parentKey, index;

    operation = event['operation'];
    ancestors = event['parent'];
    if (Array.isArray(ancestors))
    {
        // take the direct key of direct parent
        parentKey = ancestors[ancestors.length-1];
    }
    else
    {
        // single element or null, value is the parent key
        parentKey = ancestors;
    }
    index = event['index'];

    if (operation === 'insert')
    {
        this._handleInsertEvent(parentKey, index, event['data']);
    }
    else if (operation === 'delete')
    {
        this._handleDeleteEvent(parentKey, index);
    }
    else if (operation === 'refresh')
    {
        this._handleRefreshEvent(parentKey);
    }
};

/**
 * Handles insert event from TreeDataSource.
 * @param {Object} parentKey the key of the parent where the node is inserted
 * @param {number} index the index relative to its parent where the node is inserted
 * @param {Object} nodeSet the node set containing the single insert data
 * @private
 */
oj.FlattenedTreeDataSource.prototype._handleInsertEvent = function(parentKey, index, nodeSet)
{
    var parentIndex, parent, depth, insertIndex, metadata;

    parentIndex = this.getIndex(parentKey);
    parent = this._getEntry(parentIndex);
    depth = parent['depth']+1;
    insertIndex = parentIndex + index + 1;

    // there should only be one row in the set
    metadata = nodeSet.getMetadata(nodeSet.getStart());

    // insert into cache
    this._insertRow(insertIndex, metadata, parentKey, index, depth)    
};

/**
 * Handles delete event from TreeDataSource.
 * @param {Object} parentKey the key of the parent where the node is inserted
 * @param {number} index the index relative to its parent where the node is inserted
 * @private
 */
oj.FlattenedTreeDataSource.prototype._handleDeleteEvent = function(parentKey, index)
{
    var parentIndex, parent, startIndex, start, count, currentIndex, lastIndex, current;

    parentIndex = this.getIndex(parentKey);
    parent = this._getEntry(parentIndex);

    startIndex = parentIndex + index;
    start = this._getEntry(startIndex);

    // make sure the child data is valid
    oj.Assert.assert(start['parent'] === parent && start['depth'] === parent['depth']+1); 

    // remove the entry and all of its children from cache
    count = 1;
    currentIndex = startIndex + 1;
    lastIndex = this._getLastIndex();
    while (currentIndex <= lastIndex)
    {
        current = this._getEntry(currentIndex);
        // check if we have reached the last child of the deleted node
        if (current['depth'] != start['depth'])
        {
            break;
        }      
        currentIndex++;
    }

    // remove rows
    this._deleteAllRowsBelow(startIndex, count);
};

/**
 * Handles refresh event from TreeDataSource.
 * @param {Object} parentKey the key of the parent where the node is inserted
 * @private
 */
oj.FlattenedTreeDataSource.prototype._handleRefreshEvent = function(parentKey)
{
    if (parentKey == null)
    {
        // the entire tree is refreshed
        // clean up internal cache
        this.refresh();
    }
};

/**
 * Checks whether all rows should be expanded.
 * @return {boolean} true if expand all rows, false otherwise.
 * @private
 */
oj.FlattenedTreeDataSource.prototype._isExpandAll = function()
{
    var capability = this.m_wrapped.getCapability('fetchDescendants');
    return (this.m_collapsedKeys != undefined && capability != null && capability != 'disable');
};

/**
 * Checks whether batch fetching is supported.
 * @return {boolean} true if batch fetching is supported, false otherwise.
 * @private
 */
oj.FlattenedTreeDataSource.prototype._isBatchFetching = function()
{
    var capability = this.m_wrapped.getCapability('batchFetch');
    return (capability === 'enable');
};

/////////////////////////////// helper methods subclass should find useful //////////////////////////////////////////////
/**
 * Refresh the data source.  Clear out any state.
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.refresh = function()
{
    // clear the cache
    this._clearAll();

    // todo: more work here to force fetch (remove then insert)
};

/**
 * Determine the flattened index for the specified key
 * @param {*} rowKey the key to find the index
 * @return {number} the index representing the specified key.  Returns -1 if the index
 *         cannot be found.
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.getIndex = function(rowKey)
{
    var lastIndex, i, rowData;

    lastIndex = this._getLastIndex();
    for (i=0; i<=lastIndex; i++)
    {
        rowData = this._getEntry(i);
        if (rowData['key'] == rowKey)
        {
            return i;
        }
    }

    // can't find it, return -1
    return -1;
};

/**
 * Determines the key for the specified flattened index
 * @param {number} index the index in flattened view
 * @return {Object|null} the key for the specified index.  Returns null if the index has not been
 *         fetched yet or is invalid.
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.getKey = function(index)
{
    // ensure the index is valid and in range
    if (index < 0 || index > this._getLastIndex())
    {
        return null;
    }

    // just return from internal cache
    return this._getEntry(index)['key'];
};

/**
 * Returns the currently fetched range.
 * @return {Object} the fetched range (start, end).
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.getFetchedRange = function()
{
    return {'start': 0, 'end': this._getLastIndex()+1};
};

/**
 * Returns the key of the ancestors.
 * @param {*} rowKey the row key to find the ancestors.
 * @return {Array} an array of the key of the ancestors from root to the row with specified row key.
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.getAncestors = function(rowKey)
{
    var ancestors, index, parent;

    ancestors = [];
    index = this.getIndex(rowKey);
    parent = this._getParent(index);
    while (parent != null)
    {
        ancestors.push(parent);
        index = this.getIndex(parent);
        parent = this._getParent(index);
    }

    // reverse since we want to return from the root
    return ancestors.reverse();
};
///////////////////////////////////// methods subclass must override ////////////////////////////////////////////////////////
/**
 * Handles what happened when the maximum row count has been reached.
 * @param {Object} range the range of the fetch request which caused the max count to be reached.
 * @param {Object} callbacks the callbacks of the fetch request which caused the max count to be reached.
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.handleMaxCountReached = function(range, callbacks)
{
    // send an error by default
    if (callbacks != null && callbacks['error'] != null)
    {
        callbacks['error'].call(null);
    }
};

/**
 * Abstract method to insert a set of rows into the DataGrid/Table
 * @param {number} insertAtIndex the flattened index of the node where the rows are inserted.
 * @param {*} insertAtKey the key of the node where the rows are inserted (the parent key)
 * @param {Object} nodeSet the node set containing data/metadata of inserted rows
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.insertRows = function(insertAtIndex, insertAtKey, nodeSet)
{
    oj.Assert.failedInAbstractFunction();
};

/**
 * Abstract method to remove the specified rows in the DataGrid/Table
 * @param {Array.<*>} rowKeys an array of keys of the rows to be remove.
 * @protected
 */
oj.FlattenedTreeDataSource.prototype.removeRows = function(rowKeys)
{
    oj.Assert.failedInAbstractFunction();
};

///////////////////////////////// methods that manipulates the internal cache ///////////////////////////////////
/**
 * Retrieve the flattened index of the last entry fetched so far
 * @return {number} the flattened index of the last entry
 * @private
 */
oj.FlattenedTreeDataSource.prototype._getLastIndex = function()
{
    return this.m_cache.length-1;
};

/**
 * Retrieve the metadata for the last entry fetched so far
 * @return {Object} the metadata for the last entry
 * @private
 */
oj.FlattenedTreeDataSource.prototype._getLastEntry = function()
{
    return this.m_cache[this._getLastIndex()];
};

/**
 * Retrieve metadata info for the specified index.
 * @param {number} index the flattened index 
 * @return {Object} the metadata info
 * @private
 */
oj.FlattenedTreeDataSource.prototype._getEntry = function(index)
{
    return this.m_cache[index];
};

/**
 * Retrieve the parent key for the specified index.
 * @param {number} index the flattened index 
 * @return {Object} the parent key
 * @private
 */
oj.FlattenedTreeDataSource.prototype._getParent = function(index)
{
    var entry = this.m_cache[index];
    if (entry != null)
    {
        return entry['parent'];
    }
    return null;
};

/**
 * Add or insert entry to the cache
 * @param {Object} key the key
 * @param {number} depth the depth 
 * @param {number} index the index relative to its parent
 * @param {Object} parent the parent
 * @param {number=} insertAt insert the metadata entry at this flattened index
 * @private
 */
oj.FlattenedTreeDataSource.prototype._addEntry = function(key, depth, index, parent, insertAt)
{
    var rowData = new Object();
    rowData['key'] = key;
    rowData['depth'] = depth;
    rowData['index'] = index;
    rowData['parent'] = parent;

    if (insertAt === undefined)
    {
        this.m_cache.push(rowData);
    }
    else
    {
        this.m_cache.splice(insertAt, 0, rowData);
    }
};

/**
 * Remove entry from cache
 * @param {number} index the flattened index to start remove entry
 * @param {number} count how many entries to remove starting from the flattened index
 * @private
 */
oj.FlattenedTreeDataSource.prototype._removeEntry = function(index, count)
{
    this.m_cache.splice(index, count);
};

/**
 * Clears the internal cache
 * @private
 */
oj.FlattenedTreeDataSource.prototype._clearAll = function()
{
    this.m_cache.length = 0;
};

/**
 * Determines whether this data source supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none".  
 *         Returns null if the feature is not recognized.
 * @export
 * @expose
 * @memberof! oj.FlattenedTreeDataSource
 * @instance
 */
oj.FlattenedTreeDataSource.prototype.getCapability = function(feature)
{
    return this.m_wrapped.getCapability(feature);
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojRowExpander
 * @augments oj.baseComponent
 * @since 1.0.0
 * @ojstatus preview
 * @ojrole button
 * @ojshortdesc Enable hierarchical data to be displayed in a JET Table and JET DataGrid.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="rowexpanderOverview-section">
 *   JET RowExpander
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rowexpanderOverview-section"></a>
 * </h3>
 * <p>Description: A JET RowExpander is primarily used inside the JET Table and JET DataGrid.  It enables hierarchical data to be displayed in a JET Table and JET DataGrid.</p>
 *
 * <p>To enable expand and collapse of rows, developers must specify oj.FlattenedTreeTableDataSource as data when used within JET Table and oj.FlattenedTreeDataGridDataSource as data when used within JET DataGrid.</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-row-expander
 *   context='{{$context}}'>
 * &lt;/oj-row-expander></code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>The location of the RowExpander will be reversed in RTL reading direction.</p>
 * <p>In the unusual case that the directionality (LTR or RTL) changes post-init, the component containing the RowExpander (JET Table or JET DataGrid) must be <code class="prettyprint">refresh()</code>ed.
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Initial expansion</h4>
 * <p>To specify initial expanded rows with RowExpander, it is recommended that applications do this through the initial options in oj.FlattenedTreeDataSource, especially for expanding all rows initially.</p>
 */
oj.__registerWidget('oj.ojRowExpander', $['oj']['baseComponent'],
{
    version: "1.0.0",
    widgetEventPrefix: 'oj',
    options:
            {
                /**
                 * The context object obtained from the column renderer (Table) or cell renderer (DataGrid)
                 *
                 * @expose
                 * @memberof oj.ojRowExpander
                 * @instance
                 * @type {Object}
                 * @default null
                 *
                 */
                context: null,
                /**
                 * Specifies if the RowExpander is expanded.  The default value is determined by the <code class="prettyprint">context</code> obtained from the column renderer (Table) or cell renderer (DataGrid), or null if no context is specified.
                 * See <a href="#perf-section">performance</a> for recommended usage regarding initial expansion state.
                 *
                 * @expose
                 * @memberof oj.ojRowExpander
                 * @instance
                 * @type {boolean|null}
                 * @default null
                 * @ojwriteback
                 * @ojshortdesc Specifies if the RowExpander is expanded.
                 *
                 * @example <caption>Initialize the RowExpander with the <code class="prettyprint">expanded</code> attribute specified:</caption>
                 * &lt;oj-row-expander expanded='true'>&lt;/oj-row-expander>
                 *
                 * @example <caption>Get or set the <code class="prettyprint">expanded</code> property after initialization:</caption>
                 * // getter
                 * var expandedValue = myRowExpander.expanded;
                 *
                 * // setter
                 * myRowExpander.expanded = false;
                 */
                 expanded: null,
                /**
                 * Triggered when a expand is performed on the RowExpander
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojRowExpander
                 * @instance
                 * @property {string} rowKey the key of the expanded row
                 */
                expand: null,
                /**
                 * Triggered when a collapse is performed on the RowExpander
                 *
                 * @expose
                 * @event
                 * @memberof oj.ojRowExpander
                 * @instance
                 * @property {string} rowKey the key of the collapsed row
                 */
                collapse: null
            },
    classNames:
            {
                'root': 'oj-rowexpander',
                'icon': 'oj-component-icon',
                'clickable': 'oj-clickable-icon-nocontext',
                'expand': 'oj-rowexpander-expand-icon',
                'collapse': 'oj-rowexpander-collapse-icon',
                'leaf': 'oj-rowexpander-leaf-icon',
                'lazyload': 'oj-rowexpander-lazyload-icon',
                'toucharea': 'oj-rowexpander-touch-area',
                'indent': 'oj-rowexpander-indent',
                'iconspacer': 'oj-rowexpander-icon-spacer',
                'depth0': 'oj-rowexpander-depth-0',
                'depth1': 'oj-rowexpander-depth-1',
                'depth2': 'oj-rowexpander-depth-2',
                'depth3': 'oj-rowexpander-depth-3',
                'depth4': 'oj-rowexpander-depth-4',
                'depth5': 'oj-rowexpander-depth-5',
                'depth6': 'oj-rowexpander-depth-6',
                'depth7': 'oj-rowexpander-depth-7'
    },
    constants :{
        MAX_STYLE_DEPTH: 7,
        NUM5_KEY: 53
    },
    /**
     * Create the row expander
     * @override
     * @memberof oj.ojRowExpander
     * @protected
     */
    _ComponentCreate: function()
    {
        this._super();
        this.element.addClass(this.classNames['root']);
        this._initContent();
    },
    /**
     * Initialize the row expander after creation
     * @private
     */
    _initContent : function ()
    {
        var self = this, context;

        context = this.options['context'];
        //component now widget constructor or non existent
        if (context['component'] != null)
        {
            this.component = typeof context['component'] === 'function' ? context['component']('instance') : context['component'];
        }
        else if (context['componentElement'])
        {
            var widgetElem = context['componentElement'];
            widgetElem = $(widgetElem).hasClass('oj-component-initnode') ? widgetElem : $(widgetElem).find('.oj-component-initnode')[0];
            this.component = oj.Components.__GetWidgetConstructor(widgetElem)('instance');
        }
        this.datasource = context['datasource'];

        //root hidden so subtract 1
        this.depth = context['depth'];
        this.iconState = context['state'];
        this.rowKey = context['key'];
        this.index = context['index'];
        this.parentKey = context['parentKey'];

        this._addIndentation();
        this._addIcon();
        this._setIconStateClass();

        if (this.iconState === 'expanded' || this.iconState === 'collapsed')
        {
            $(this.toucharea).on('touchend', function(event) {
                //prevent scroll to top and # append, also prevents the following click
                event.preventDefault();
                self._fireExpandCollapse();
            });

            $(this.toucharea).on('click', function(event) {
                //prevent scroll to top and # append
                event.preventDefault();
                self._fireExpandCollapse();
            });
            $(this.element).on('keypress', function(event) {
                var code = event.keyCode || event.which;
                if (code === $.ui.keyCode.ENTER || code === $.ui.keyCode.SPACE)
                {
                    // do expand or collapse
                    self._fireExpandCollapse();
                    // stop browser from for example scrolling the page
                    event.preventDefault();
                    // ensure focus stays
                    event.target.focus();
                }
            });

            // listen for key down event from host component
            this.handleKeyDownCallback = this._handleKeyDownEvent.bind(this);
            this.component.element.get(0).addEventListener('keydown', this.handleKeyDownCallback, true);

            // listens for expand and collapse event from flattened datasource
            // this could be due to user clicks, keyboard shortcuts or programmatically
            this.handleExpandCallback = this._handleExpandEvent.bind(this);
            this.handleCollapseCallback = this._handleCollapseEvent.bind(this);

            this.datasource.on("expand", this.handleExpandCallback, this);
            this.datasource.on("collapse", this.handleCollapseCallback, this);

            // if expanded option is explicitly specified, make sure it's in sync with current state
            this._initExpanded();
        }
        else if (this.iconState === 'leaf')
        {
            // we'll still need to handle ctrl+alt+5 for leaf node
            // listen for key down event from host component
            this.handleKeyDownCallback = this._handleKeyDownEvent.bind(this);
            this.component.element.get(0).addEventListener('keydown', this.handleKeyDownCallback, true);
            $(this.icon).attr('tabindex', -1);
        }

        // listen for active key change event from host component
        this.handleActiveKeyChangeCallback = this._handleActiveKeyChangeEvent.bind(this);
        if (this.component._IsCustomElement())
        {
            $(this.component.element).on('ojBeforeCurrentCell', this.handleActiveKeyChangeCallback);
        }
        else
        {
            $(this.component.element).on('ojbeforecurrentcell', this.handleActiveKeyChangeCallback);
        }
    },
    /**
     * Sync initial state of expanded with context/FlattenedTreeModel
     * @private
     */
    _initExpanded: function()
    {
        var expanded = this.options['expanded'];
        if (expanded != null)
        {
            if (expanded && this.iconState === 'collapsed')
            {
                this._expand();
            }
            else if (!expanded && this.iconState === 'expanded')
            {
                this._collapse();
            }
        }
        else
        {
            // make sure expanded value reflect the current state
            // we don't want to trigger option change event in this case
            this.options['expanded'] = this.iconState === 'collapsed' ? false : true;
        }
    },

    /**
     * Redraw the RowExpander element.
     *
     * @expose
     * @memberof oj.ojRowExpander
     * @instance
     *
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myRowExpander.refresh();
     */
    refresh: function()
    {
        this.element.empty();
        this._initContent();
    },
    /**
     * destroy the row expander
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof oj.ojRowExpander
     * @instance
     * @private
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * $( ".selector" ).ojRowExpander( "destroy" );
     */
    _destroy: function()
    {
        // unregister keydown and active key change handlers
        this.component.element.get(0).removeEventListener('ojkeydown', this.handleKeyDownCallback, true);
        $(this.component.element).off('ojbeforecurrentcell', this.handleActiveKeyChangeCallback);

        // unregister expand/collapse events
        this.datasource.off("expand", this.handleExpandCallback, this);
        this.datasource.off("collapse", this.handleCollapseCallback, this);

        this.element.removeClass(this.classNames['root']);
        this.element.empty();
    },
    /**
     * Expand the current row expander
     * @return {boolean} true if the expand is processed, false if it's a no op.
     * @private
     */
    _expand: function()
    {
        if (this.iconState === 'collapsed')
        {
            this._loading();
            this.datasource.expand(this.rowKey);
            return true;
        }
        return false;
    },
    /**
     * Collapse the current row expander
     * @return {boolean} true if the collapse is processed, false if it's a no op.
     * @private
     */
    _collapse: function()
    {
        if (this.iconState === 'expanded')
        {
            this._loading();
            this.datasource.collapse(this.rowKey);
            return true;
        }
        return false;
    },
    /**
     * Sets a single option value
     * @param {Object} key the option key
     * @param {Object} value the option value
     * @param {Object} flags additional flags for option*
     * @override
     * @private
     */
    _setOption: function(key, value, flags)
    {
        if (key == 'expanded' && (flags['_context'] == null || flags['_context']['internalSet'] != true))
        {
            value ? this._expand() : this._collapse();
            // don't update option, it will be update when the operation completed via expand/collapse event
            return;
        }

        this._super(key, value, flags);

        // refresh if context is updated
        if (key == 'context' && flags['_context'] != null && flags['_context']['internalSet'] != true)
        {
            this.refresh();
        }
    },
    /**
     * Add athe indentation spacers to the row
     * @private
     */
    _addIndentation: function()
    {
        var remainder, i, depth;
        //0 index the depth for style purposes
        depth = this.depth-1;
        if (depth < this.constants.MAX_STYLE_DEPTH)
        {
            this._appendSpacer(depth);
        }
        else
        {
            for (i=1; i <= (depth/(this.constants.MAX_STYLE_DEPTH)); i++)
            {
                this._appendSpacer(this.constants.MAX_STYLE_DEPTH);
            }
            remainder = (depth % this.constants.MAX_STYLE_DEPTH);
            if (remainder < this.constants.MAX_STYLE_DEPTH)
            {
                this._appendSpacer(remainder);
            }
        }
    },
    /**
     * Append appropriate spacer based on depth to the row expander
     * @param {number} depth the depth
     * @private
     */
    _appendSpacer: function(depth)
    {
        var spacer = $(document.createElement('span')).addClass(this.classNames['indent']).addClass(this.classNames['depth'+depth]);
        this.element.append(spacer); //@HTMLUpdateOK
    },
    /**
     * Add an icon to the row expander with appropriate class names for a clickable icon.
     * @private
     */
    _addIcon: function()
    {
        var self, iconSpacer;
        iconSpacer = $(document.createElement('div')).addClass(this.classNames['iconspacer']);
        this.toucharea = $(document.createElement('div')).addClass(this.classNames['toucharea']);
        this.icon = $(document.createElement('a')).attr('href', '#').attr('aria-labelledby', this._getLabelledBy()).addClass(this.classNames['icon']).addClass(this.classNames['clickable']);
        this.element.append(iconSpacer.append(this.toucharea.append(this.icon))); //@HTMLUpdateOK

        self = this;
        this._focusable({
            'element': self.icon,
            'applyHighlight': true
        });
    },
    /**
     * Add a class name on the icon
     * @private
     * @param {string} classKey the key of the appropriate icon class expand/collapse/leaf
     */
    _addIconClass: function(classKey)
    {
        this.icon.addClass(this.classNames[classKey]);
    },
    /**
     * Remove a class name on the icon
     * @private
     * @param {string} classKey the key of the appropriate icon class expand/collapse/leaf
     */
    _removeIconClass: function(classKey)
    {
        this.icon.removeClass(this.classNames[classKey]);
    },
    /**
     * Set the icon class to the the iconState property
     * @private
     */
    _setIconStateClass: function()
    {
        switch (this.iconState)
        {
            case 'leaf':
                this._removeIconClass('icon');
                this._removeIconClass('clickable');
                this._addIconClass('leaf');
                break;
            case 'collapsed':
                this._addIconClass('expand');
                this._ariaExpanded(false);
                break;
            case 'expanded':
                this._addIconClass('collapse');
                this._ariaExpanded(true);
                break;
            case 'loading':
                this._removeIconClass('clickable');
                this._addIconClass('lazyload');
                break;
        }

    },
    /**
     * Removes the icon class of the iconState property
     * @private
     */
    _removeIconStateClass: function()
    {
        switch (this.iconState)
        {
            case 'leaf':
                this._removeIconClass('leaf');
                this._addIconClass('icon');
                this._addIconClass('clickable');
                break;
            case 'collapsed':
                this._removeIconClass('expand');
                break;
            case 'expanded':
                this._removeIconClass('collapse');
                break;
            case 'loading':
                this._removeIconClass('lazyload');
                this._addIconClass('clickable');
                break;
        }

    },
    /**
     * Handles active key change event from host component (ojDataGrid or ojTable)
     * @param {Event} event
     * @param {Object|null|number=} ui
     * @private
     */
    _handleActiveKeyChangeEvent: function(event, ui)
    {
        var rowKey, previousRowKey, context, state;
        if (ui == null)
        {
            ui = event.detail;
        }
        if (ui['currentCell'] != null)
        {
            rowKey = ui['currentCell']['type'] == 'cell' ? ui['currentCell']['keys']['row'] :  ui['currentCell']['key'];
            if (ui['previousValue'] != null)
            {
                previousRowKey = ui['previousCurrentCell']['type'] == 'cell' ? ui['previousCurrentCell']['keys']['row'] :  ui['previousCurrentCell']['key'];
            }
            // if the event is for this row and the active key change event is triggered
            // by row change and not column change
            if (this.rowKey === rowKey && previousRowKey != rowKey)
            {
                // if the component allows AccessibleContext to be set
                if (this.component._setAccessibleContext)
                {
                    // row context of row expander for screen reader
                    // todo: get index from TreeDataSource as well since that could change
                    context = this.getTranslatedString('accessibleRowDescription', {'level': this.depth, 'num': this.index+1, 'total': this.datasource.getWrappedDataSource().getChildCount(this.parentKey)});
                    // state of row expander for screen reader
                    if (this.iconState === 'collapsed')
                    {
                        state = this.getTranslatedString('accessibleStateCollapsed');
                    }
                    else if (this.iconState === 'expanded')
                    {
                        state = this.getTranslatedString('accessibleStateExpanded');
                    }
                    else
                    {
                        // for leaf node don't read anything
                        state = '';
                    }

                    this.component._setAccessibleContext({'context': context, 'state': state});
                }
            }
        }
    },
    /**
     * Handles keydown event from host component (ojDataGrid or ojTable)
     * No longer returns a boolean but rather prevents default if appropriate
     * @param {Event} event
     * @private
     */
    _handleKeyDownEvent: function(event)
    {
        var rowKey, code, context, ancestorInfo, ancestors, i;
        var targetContext = oj.Components.__GetWidgetConstructor(this.component.element.get(0))('getContextByNode', event.target);

        if (targetContext == null)
        {
            return;
        }

        rowKey = targetContext['key'];
        if (rowKey == null)
        {
            rowKey = targetContext['keys']['row'];
        }
        if (this.rowKey === rowKey)
        {
            code = event.keyCode || event.which;
            // ctrl (or equivalent) is pressed
            if (oj.DomUtils.isMetaKeyPressed(event))
            {
                // Ctrl+Right expands, Ctrl+Left collapse in accordance with WAI-ARIA best practice
                // consume the event as it's processed
                if (code == $.ui.keyCode.RIGHT)
                {
                    if (this._expand())
                    {
                        event.preventDefault();
                        return;
                    }
                }
                else if (code == $.ui.keyCode.LEFT)
                {
                    if (this._collapse())
                    {
                        event.preventDefault();
                        return;
                    }
                }
                else if (event.altKey && code == this.constants.NUM5_KEY)
                {
                    // read current cell context
                    if (this.component._setAccessibleContext)
                    {
                        ancestors = this.datasource.getAncestors(this.rowKey);
                        if (ancestors != null && ancestors.length > 0)
                        {
                            ancestorInfo = [];
                            for (i=0; i<ancestors.length; i++)
                            {
                                ancestorInfo.push({'key': ancestors[i], 'label': this.getTranslatedString('accessibleLevelDescription', {'level': i+1})});
                            }
                        }

                        context = this.getTranslatedString('accessibleRowDescription', {'level': this.depth, 'num': this.index+1, 'total': this.datasource.getWrappedDataSource().getChildCount(this.parentKey)});
                        this.component._setAccessibleContext({'context': context, 'state': '', 'ancestors': ancestorInfo});
                    }
                }
            }
        }
        return;
    },
    /**
     * Put row expander in a loading state.  This is called during expand/collapse.
     * @private
     */
    _loading: function()
    {
        this._removeIconStateClass();
        this.iconState = 'loading';
        this._setIconStateClass();
    },

    /**
     * Handle an expand event coming from the datasource,
     * update the icon and the aria-expand property
     * @param {Object} event the expand event from the data source, should contain rowKey
     * @private
     */
    _handleExpandEvent: function(event)
    {
        var rowKey = event['rowKey'];
        if (rowKey === this.rowKey)
        {
            this._removeIconStateClass();
            this.iconState = 'expanded';
            this._setIconStateClass();
            this._ariaExpanded(true);
            this._updateContextState('expanded');

            // if the event is triggered by initial setting of expanded, we should not
            // fire expand or option change event
            var expanded = this.options['expanded'];
            if (expanded == null || (expanded != null && !expanded))
            {
                this._trigger('expand', null, {'rowKey': rowKey});
                this._updateExpandedState(true);
            }
        }
    },
    /**
     * Handle a collapse event coming from the datasource,
     * update the icon and the aria-expand property
     * @param {Object} event the collapse event from the data source, should contain rowKey
     * @private
     */
    _handleCollapseEvent: function(event)
    {
        var rowKey = event['rowKey'];
        if (rowKey === this.rowKey)
        {
            this._removeIconStateClass();
            this.iconState = 'collapsed';
            this._setIconStateClass();
            this._ariaExpanded(false);
            this._updateContextState('collapsed');

            // if the event is triggered by initial setting of expanded, we should not
            // fire expand or option change event
            var expanded = this.options['expanded'];
            if (expanded == null || (expanded != null && expanded))
            {
                this._trigger('collapse', null, {'rowKey': rowKey});
                this._updateExpandedState(false);
            }
        }
    },
    /**
     * Update the expanded option
     * @param {boolean} expanded
     * @private
     */
    _updateExpandedState: function(expanded)
    {
        this.option('expanded', expanded, {'changed':true, '_context': {'internalSet': true, 'writeback':true}});
    },
    /**
     * Update context state
     * @param {string} newState
     * @private
     */
    _updateContextState: function(newState)
    {
        var context = this.options['context'];
        context['state'] = newState;
        // need to reuse the same object so mark it as changed ourselves
        this.option('context', context, {'changed':true, '_context': {'internalSet': true}});
    },
    /**
     * Fire the expand or collapse on the datasource and the oj event on the widget
     * @private
     */
    _fireExpandCollapse: function()
    {
        var state = this.iconState;

        // show loading icon, note this changes the icon state to 'loading'
        this._loading();

        // invoke expand/collapse on datasource
        if (state === 'collapsed')
        {
            this.datasource.expand(this.rowKey);
        }
        else if (state === 'expanded')
        {
            this.datasource.collapse(this.rowKey);
        }
    },
    /**
     * Sets the icon's aria-expanded property to the boolean passed in
     * @param {boolean|null} bool true if expanded false if not
     * @private
     */
    _ariaExpanded: function(bool)
    {
        this.icon.attr('aria-expanded', bool);
    },

    // @inheritdoc
    getNodeBySubId: function(locator)
    {
        var subId;

        if (locator == null)
        {
            return this.element ? this.element[0] : null;
        }

        subId = locator['subId'];
        if ((subId === 'oj-rowexpander-disclosure' || subId === 'oj-rowexpander-icon') && this.icon != null)
        {
            return this.icon.get(0);
        }
        // Non-null locators have to be handled by the component subclasses
        return null;
    },

    // @inheritdoc
    getSubIdByNode: function(node)
    {
        if (node === this.icon.get(0))
        {
            return {'subId': 'oj-rowexpander-disclosure'};
        }
        return null;
    },
    _NotifyAttached: function()
    {
        this._super();
        this.icon.attr('aria-labelledby', this._getLabelledBy());
    },
    /**
     * Get the aria label of the rowexpander from the closest row expander
     * @return {string} the closest id set to the rowexpander
     * @private
     */
    _getLabelledBy: function()
    {
        return this.element.parent().closest('[id]').attr('id');
    }

    //////////////////     FRAGMENTS    //////////////////
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Icon</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Expand or collapse the row with the icon in it.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch section of classdesc, and standalone gesture doc
     * @memberof oj.ojRowExpander
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan="2">Focused Row or Cell with RowExpander</td>
     *       <td><kbd>Ctrl + RightArrow</kbd></td>
     *       <td>Expand</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Ctrl + LeftArrow</kbd></td>
     *       <td>Collapse</td>
     *     </tr>
     *     <tr>
     *       <td rowspan="1">Icon</td>
     *       <td><kbd>Enter</kbd></td>
     *       <td>Expand or Collapse</td>
     *     </tr>
     *   </tbody>
     * </table>
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojRowExpander
     */

    //////////////////     SUB-IDS     //////////////////

    /**
     * <p>Sub-ID for the ojRowExpander's icon.</p>
     *
     * @ojsubid oj-rowexpander-disclosure
     * @memberof oj.ojRowExpander
     *
     * @example <caption>Get the icon from the RowExpander:</caption>
     * var node = myRowExpander.getNodeBySubId({subId: 'oj-rowexpander-disclosure'});
     */
});

(function() {
var ojRowExpanderMeta = {
  "properties": {
    "context": {
      "type": "object"
    },
    "expanded": {
        "type": "boolean",
        "writeback": true
    },
    "translations": {
      "type": "Object",
      "properties": {
        "accessibleLevelDescription": {
          "type": "string"
        },
        "accessibleRowCollapsed": {
          "type": "string"
        },
        "accessibleRowDescription": {
          "type": "string"
        },
        "accessibleRowExpanded": {
          "type": "string"
        },
        "accessibleStateCollapsed": {
          "type": "string"
        },
        "accessibleStateExpanded": {
          "type": "string"
        }
      }
    }
  },
  "events": {
    "collapse": {},
    "expand": {}
  },
  "extension": {
    _WIDGET_NAME: "ojRowExpander"
  }
};
oj.CustomElementBridge.registerMetadata('oj-row-expander', 'baseComponent', ojRowExpanderMeta);
oj.CustomElementBridge.register('oj-row-expander', {'metadata': oj.CustomElementBridge.getMetadata('oj-row-expander')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve jQuery UI Touch Punch 0.2.3
 *
 * Copyright 2011-2014, Dave Furfero
 * Dual licensed under the MIT or GPL Version 2 licenses.
 */

/**
 * Utility class for proxying touch events for a given element and mapping them to mouse events
 * @constructor
 * @ignore
 * @private
 */
oj._TouchProxy = function (elem) {
  this._init(elem);
};

/**
 * Initializes the TouchProxy instance
 *
 * @param {Object} elem
 * @private
 */
oj._TouchProxy.prototype._init = function (elem) {
  this._elem = elem;

  this._touchHandled = false;
  this._touchMoved = false;

  // add touchListeners
  this._touchStartHandler = $.proxy(this._touchStart, this);
  this._touchEndHandler = $.proxy(this._touchEnd, this);
  this._touchMoveHandler = $.proxy(this._touchMove, this);

  this._elem.on({
    touchstart: this._touchStartHandler,
    touchend: this._touchEndHandler,
    touchmove: this._touchMoveHandler,
    touchcancel: this._touchEndHandler
  });
};

oj._TouchProxy.prototype._destroy = function () {
  if (this._elem && this._touchStartHandler) {
    this._elem.off({
      touchstart: this._touchStartHandler,
      touchmove: this._touchMoveHandler,
      touchend: this._touchEndHandler,
      touchcancel: this._touchEndHandler
    });

    this._touchStartHandler = undefined;
    this._touchEndHandler = undefined;
    this._touchMoveHandler = undefined;
  }
};

/**
 * Simulate a mouse event based on a corresponding touch event
 * @param {Object} event A touch event
 * @param {string} simulatedType The corresponding mouse event
 *
 * @private
 */
oj._TouchProxy.prototype._touchHandler = function (event, simulatedType) {
  // Ignore multi-touch events
  if (event.originalEvent.touches.length > 1) {
    return;
  }

  //  - contextmenu issues: presshold should launch the contextmenu on touch devices
  if (event.type !== 'touchstart' && event.type !== 'touchend') {
    event.preventDefault();
  }

  var touch = event.originalEvent.changedTouches[0];
  var simulatedEvent = document.createEvent('MouseEvent');

  // Initialize the simulated mouse event using the touch event's coordinates
  // initMouseEvent(type, canBubble, cancelable, view, clickCount,
  //                screenX, screenY, clientX, clientY, ctrlKey,
  //                altKey, shiftKey, metaKey, button, relatedTarget);
  simulatedEvent.initMouseEvent(simulatedType, true, true, window, 1,
                                touch.screenX, touch.screenY,
                                touch.clientX, touch.clientY, false,
                                false, false, false, 0/* left*/, null);

  touch.target.dispatchEvent(simulatedEvent);
};

/**
 * Handle touchstart events
 * @param {Object} event The element's touchstart event
 *
 * @private
 */
oj._TouchProxy.prototype._touchStart = function (event) {
  // Ignore the event if already being handled
  if (this._touchHandled) {
    return;
  }

  // set the touchHandled flag
  this._touchHandled = true;

  // Track movement to determine if interaction was a click
  this._touchMoved = false;

  // Simulate the mouseover, mousemove and mousedown events
  this._touchHandler(event, 'mouseover');
  this._touchHandler(event, 'mousemove');
  this._touchHandler(event, 'mousedown');
};

/**
 * Handle the touchmove events
 * @param {Object} event The element's touchmove event
 *
 * @private
 */
oj._TouchProxy.prototype._touchMove = function (event) {
  // Ignore event if not handled
  if (!this._touchHandled) {
    return;
  }

  // Interaction was not a click
  this._touchMoved = true;

  // Simulate the mousemove event
  this._touchHandler(event, 'mousemove');
};

/**
 * Handle the touchend events
 * @param {Object} event The element's touchend event
 *
 * @private
 */
oj._TouchProxy.prototype._touchEnd = function (event) {
  // Ignore event if not handled
  if (!this._touchHandled) {
    return;
  }

  // Simulate the mouseup and mouseout events
  this._touchHandler(event, 'mouseup');
  this._touchHandler(event, 'mouseout');

  // If the touch interaction did not move, it should trigger a click
  if (!this._touchMoved && event.type === 'touchend') {
    // Simulate the click event
    this._touchHandler(event, 'click');
  }

  // Unset the flag
  this._touchHandled = false;
};

oj._TouchProxy._TOUCH_PROXY_KEY = '_ojTouchProxy';

oj._TouchProxy.prototype.touchMoved = function () {
  return this._touchMoved;
};

/**
 * Adds touch event listeners
 * @param {Object} elem
 * @ignore
 */
oj._TouchProxy.addTouchListeners = function (elem) {
  var jelem = $(elem);
  var proxy = jelem.data(oj._TouchProxy._TOUCH_PROXY_KEY);
  if (!proxy) {
    proxy = new oj._TouchProxy(jelem);
    jelem.data(oj._TouchProxy._TOUCH_PROXY_KEY, proxy);
  }

  return proxy;
};

/**
 * Removes touch event listeners
 * @param {Object} elem
 * @ignore
 */
oj._TouchProxy.removeTouchListeners = function (elem) {
  var jelem = $(elem);
  var proxy = jelem.data(oj._TouchProxy._TOUCH_PROXY_KEY);
  if (proxy) {
    proxy._destroy();
    jelem.removeData(oj._TouchProxy._TOUCH_PROXY_KEY);
  }
};


/**
 * <p>
 * A composite component is a reusable piece of UI that can be embedded as a custom HTML element and
 * can be composed of other composites, JET components, HTML, JavaScript, or CSS.
 * </p>
 *
 * <h2 id="registration">Packaging and Registration
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#registration"></a>
 * </h2>
 * <p>
 * Composite components should be packaged as a standalone module in a folder matching the tag name it will be registered with, e.g. 'my-chart'.
 * An application would use a composite by requiring it as a module, e.g. 'jet-composites/my-chart/loader'. The composite module could be 
 * stored locally in the app which is the recommended approach, but could also be stored on a different server, or a CDN.  Note that there are
 * XHR restrictions when using the RequireJS text plugin which may need additional RequireJS config settings.  Please see the 
 * <a href="https://github.com/requirejs/text#xhr-restrictions">text plugin documentation</a> for the full set of limitations and options.
 * By using RequireJS path mappings, the application can control where individual composites are loaded from. 
 * See below for a sample RequireJS composite path configuration.
 *
 * Note that the 'jet-composites/my-chart' mapping is only required if the 'my-chart' composite module maps to a folder other than 
 * 'someSubFolder/jet-composites/my-chart' using the configuration below.
 * <pre class="prettyprint"><code>
 * requirejs.config(
 * {
 *   baseUrl: 'js',
 *   paths:
 *   {
 *     'jet-composites': 'someSubFolder/jet-composites',
 *     'jet-composites/my-chart': 'https://someCDNurl',
 *     'jet-composites/my-table': 'https://someServerUrl'
 *   }
 * }
 * </code></pre>
 * </p>
 * 
 * <p>
 * All composite modules should contain a loader.js file which will handle registering and specifying the dependencies for a composite component.
 * We recommend using RequireJS to define your composite module with relative file dependencies.  
 * Registration is done via the <a href="#register">oj.Composite.register</a> API.
 * By registering a composite component, an application links an HTML tag with provided
 * Metadata, View, ViewModel and CSS which will be used to render the composite. These optional
 * pieces can be provided via a descriptor object passed into the register API. See below for sample loader.js file configurations.
 * </p>
 *
 * Note that in this example we are using require-css, a RequireJS plugin for loading css which will load the styles within our page
 * so we do not need to pass any css into the register call. This is the recommended way to load CSS, especially for cases
 * where the composite styles contain references to any external resources.
 * <pre class="prettyprint"><code>
 * define(['text!./my-chart.html', './my-chart', 'text!./my-chart.json', 'css!./my-chart'],
 *   function(view, viewModel, metadata) {
 *     oj.Composite.register('my-chart',
 *     {
 *       metadata: JSON.parse(metadata),
 *       view: view,
 *       viewModel: viewModel
 *     });
 *   }
 * );
 * </code></pre>
 *
 * This example shows how to register a custom parse function which will be called to parse attribute values defined in the metadata.
 * <pre class="prettyprint"><code>
 * define(['text!./my-chart.html', './my-chart', 'text!./my-chart.json'],
 *   function(view, viewModel, metadata) {
 *     var myChartParseFunction = function(value, name, meta, defaultParseFunction) {
 *       // Custom parsing logic goes here which can also return defaultParseFunction(value) for
 *       // values the composite wants to default to the default parsing logic for.
 *       // This function is only called for non bound attributes.
 *     }
 * 
 *     oj.Composite.register('my-chart',
 *     {
 *       metadata: JSON.parse(metadata),
 *       view: view,
 *       viewModel: viewModel,
 *       parseFunction: myChartParseFunction
 *     });
 *   }
 * );
 * </code></pre>
 *
 * <h2 id="usage">Usage
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#usage"></a>
 * </h2>
 * <p>
 *   Once registered within a page, a composite component can be used in the DOM as a custom HTML element like in the example below.
 *   A composite element will be recognized by the framework only after its module is loaded by the application. Once the element is 
 *   recognized, the framework will register a busy state for the element and will begin the process of 'upgrading' the element. 
 *   The element will not be ready for interaction (e.g. retrieving properties or calling methods) until the upgrade process is 
 *   complete with the exception of property setters and the setProperty and setProperties methods.
 *   The application should listen to either the page-level or an element-scoped BusyContext before attempting to interact with 
 *   any JET custom elements. See the <a href="oj.BusyContext.html">BusyContext</a> documentation on how BusyContexts can be scoped.
 * </p>
 * <p>
 *   The upgrade of JET composite elements relies on any data binding resolving, the management of which is done by a binding provider. 
 *   The binding provider is responsible for setting and updating attribute expressions and any custom elements within its managed 
 *   subtree will not finish upgrading until it applies bindings on that subtree. By default, there is a single binding provider for a page, 
 *   but subtree specific binding providers can be added by using the <code>data-oj-binding-provider</code> attribute with values of 
 *   "none" and "knockout". The default binding provider is knockout, but if a page or DOM subtree does not use any expression syntax or 
 *   knockout, the application can set <code>data-oj-binding-provider="none"</code> on that element so its dependent JET composite custom 
 *   elements do not need to wait for bindings to be applied to finish upgrading.
 * </p>
 * <pre class="prettyprint"><code>
 * &lt;my-chart type="bubble" data="{{dataModel}}">&lt;/my-chart>
 * </code></pre>
 *
 * <h2 id="metadata">Metadata
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#metadata"></a>
 * </h2>
 * <p>The composite Metadata is a JSON formatted object which defines the properties, methods, slots, and events fired by
 *  the composite. <b>The name of the composite component properties, event listeners, and methods should avoid collision 
 *  with the existing <a href="https://developer.mozilla.org/en-US/docs/Web/API/HTMLElement">HTMLElement</a> properties, 
 *  event listeners, and methods. Additionally, the composite should not re-define any
 *  <a href="https://developer.mozilla.org/en-US/docs/Web/HTML/Global_attributes">global attributes</a> or events.</b></p>
 *  
 * <p>The Metadata JSON object should have the following required properties: "name", "version", "jetVersion" and 
 *  the following optional properties: "description", "dependencies", "icon", "displayName", "properties", "methods", "events", or "slots". 
 *  See the tables below for descriptions of these properties. The Metadata can be extended by appending any extra information in an "extension" 
 *  field except at the first level of the "properties", "methods", "events" or "slots" objects. Any metadata in an extension field will be ignored.</p>
 *  
 * <p>Keys defined in the "properties" top level object should map to the composite component's properties following
 *  the same naming convention described in the <a href="#attr-to-prop-mapping">Property-to-Attribute mapping</a> section.
 *  Non expression bound composite component attributes will be correctly evaluated only if they are a primitive JavaScript type (boolean, number, string)
 *  or a JSON object. Note that JSON syntax requires that strings use double quotes. Attributes evaluating to any other types must be bound via expression syntax.  
 *  Boolean attributes are considered true if set to the case-insensitive attribute name, the empty string or have no value assignment. 
 *  JET composite components will also evalute boolean attributes set explicitly to 'true' or 'false' to their respective boolean values. All other values are invalid.</p>
 *  
 * <p>Not all information provided in the Metadata is required at run time and those not indicated to be required at run time in the tables
 *  below can be omitted to reduce the Metadata download size. Any non run time information can be used for design time tools and property editors and could
 *  be kept in a separate JSON file which applications can use directly or merge with the run time metadata as needed, but would not be required by the loader.js
 *  file. For methods and events, only the method/event name is required at run time.</p>
 *
 * <h3>Metadata Properties</h3>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Used at Runtime</th>
 *       <th>Required</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="rt">name</td>
 *       <td>yes</td>
 *       <td>yes</td>
 *       <td>{string}</td>
 *       <td>The component tag name.
 *           The component name must meet the following requirements (based upon the <a href="https://www.w3.org/TR/custom-elements/#custom-elements-core-concepts">W3C Custom Element spec</a>):
 *           <ul>
 *             <li>The name can include only letters, digits, '-', and '_'.</li>
 *             <li>The letters in the name should be all lowercase.
 *             <li>At least one hyphen is required.
 *             <li>The first segment (up to the first hyphen) is supposed to be a namespace prefix. The prefix 'oj' is reserved for native JET components.
 *             <li>The first hyphen must be followed by at least one character.
 *             <li>The name cannot be one of the following reserved names:
 *             <ul>
 *               <li>annotation-xml
 *               <li>color-profile
 *               <li>font-face
 *               <li>font-face-src
 *               <li>font-face-uri
 *               <li>font-face-format
 *               <li>font-face-name
 *               <li>missing-glyph
 *             <ul>
 *           <ul> 
 *       </td>
 *     </tr>
 *     <tr>
 *       <td class="rt">version</td>
 *       <td>yes</td>
 *       <td>yes</td>
 *       <td>{string}</td>
 *       <td>The component version. Note that changes to the metadata even for minor updates like updating the
 *         jetVersion should result in at least a minor composite version change, e.g. 1.0.0 -> 1.0.1.</td>
 *     </tr>
 *     <tr>
 *       <td class="rt">jetVersion</td>
 *       <td>yes</td>
 *       <td>yes</td>
 *       <td>{string}</td>
 *       <td>The <a href="http://semver.org/">semantic version</a> of the supported JET version(s). 
 *         Composite authors should not specify a semantic version range that includes unreleased JET major versions
 *         as major releases may contain non backwards compatible changes. Authors should instead recertify composites 
 *         with each major release and update the composite metadata or release a new version that is compatible with the new
 *         release changes.</td>
 *     </tr>
 *     <tr>
 *       <td class="rt">properties</td>
 *       <td>yes</td> 
 *       <td>no</td>
 *       <td>{Object<string,<string>}</td>
 *       <td>See Properties table below for details.</td>
 *     </tr> 
*     <tr>
 *       <td class="rt">methods</td>
 *       <td>yes</td>
 *       <td>no</td>
 *       <td>{Object<string,<string>}</td>
 *       <td>See Methods table below for details.</td>
 *     </tr>
 *     <tr>
 *       <td class="rt">events</td>
 *       <td>yes</td> 
 *       <td>no</td>
 *       <td>{Object<string,<string>}</td>
 *       <td>See Events table below for details.</td>
 *     </tr>
 *     <tr>
 *       <td class="rt">slots</td>
 *       <td>yes</td>
 *       <td>no</td>
 *       <td>{Object<string,<string>}</td>
 *       <td>See Slots table below for details.</td>
 *     </tr> 
 *     <tr>
 *       <td class="name"><code>compositeDependencies</code></td>
 *       <td>no</td>
 *       <td>no</td>
 *       <td>{Object<string, string>}</td>
 *       <td>Dependency to semantic version mapping for composite dependencies. 
 *         3rd party libraries should not be included in this mapping.
 *         <code>{"composite1": "1.2.0", "composite2": ">=2.1.0"}</code>
 *         <p><b><i>This metadata property is deprecated as of JET 5.0.0.</i></b>  Please use the "dependencies" metadata property instead.</p>
 *       </td> 
 *     </tr> 
 *     <tr>
 *       <td class="name"><code>dependencies</code></td>
 *       <td>no</td>
 *       <td>no</td>
 *       <td>{Object<string, string>}</td>
 *       <td>Dependency to semantic version mapping for composite dependencies. 
 *         3rd party libraries should not be included in this mapping.
 *         <code>{"composite1": "1.2.0", "composite2": ">=2.1.0"}</code></td> 
 *     </tr> 
 *     <tr>
 *       <td class="name"><code>description</code></td>
 *       <td>no</td>
 *       <td>no</td>
 *       <td>{string}</td>
 *       <td>A high-level description for the component.</td> 
 *     </tr> 
 *     <tr>
 *       <td class="name"><code>displayName</code></td>
 *       <td>no</td>
 *       <td>no</td>
 *       <td>{string}</td>
 *       <td>A user friendly, translatable name of the component.</td> 
 *     </tr> 
 *     <tr>
 *       <td class="name"><code>extension</code></td>
 *       <td>no</td>
 *       <td>no</td>
 *       <td>{Object}</td>
 *       <td>Placeholder for Extension metadata.  Each section is identified by a key that specifies the downstream tool that will process this metadata.
 *         <h6>For example:</h6> 
 *         <table class="params">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody>
 *             <tr>
 *               <td class="name"><code>vbcs</code></td>
 *               <td>{string}</td>
 *               <td>Indentifies an object with VBCS-specific metadata</td>
 *             </tr> 
 *           </tbody>
 *         </table>
 *         </br>
 *         Please consult the documentation for the downstream tool to determine what (if any) extension metadata is supported.
 *      </td>
 *     </tr> 
 *     <tr>
 *       <td class="name"><code>icon</code></td>
 *       <td>no</td>
 *       <td>no</td>
 *       <td>{Object}</td>
 *       <td>One or more optional images for representing the component within a design time environment's component palette. The object has the following properties:
 *         <h6>Properties</h6>
 *         <table class="params">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody>
 *             <tr>
 *               <td class="name"><code>iconPath</code></td>
 *               <td>{string}</td>
 *               <td>A relative path to the default (enabled) icon.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>selectedIconPath</code></td>
 *               <td>{string}</td>
 *               <td>A relative path to the icon that represents the selected state of the component.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>hoverIconPath</code></td>
 *               <td>{string}</td>
 *               <td>A relative path to the icon that represents the hover state of the component.</td>
 *             </tr>
 *           </tbody>
 *         </table> 
 *      </td>
 *     </tr> 
 *     <tr>
 *       <td class="name"><code>styleClasses</code></td>
 *       <td>no</td>
 *       <td>no</td>
 *       <td>{Array.<{Object}>}</td>
 *       <td>Optional array of groupings of style class names that are applicable to this component.  Each grouping object has the following properties:
 *         <h6>Properties</h6>
 *         <table class="params">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody>
 *             <tr>
 *               <td class="name"><code>styleGroup</code></td>
 *               <td>{Array.<{string}>}</td>
 *               <td>Array of mutually exclusive style class names that belong to this group.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>description</code></td>
 *               <td>{string}</td>
 *               <td>A translatable high-level description for this group of styleClasses.</td>
 *             </tr> 
 *           </tbody>
 *         </table>
 *      </td>
 *     </tr> 
 *   </tbody>
 * </table>
 *
 * <h3>Properties</h3>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Value</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>[property name]</code></td>
 *       <td>Object containing the following properties:
 *         <h6>Properties</h6>
 *         <table class="params">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Used at Runtime</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody> 
 *             <tr>
 *               <td class="rt">enumValues</td>
 *               <td>yes</td>
 *               <td>{Array<string>}</td>
 *               <td>An optional list of valid enum values for a string property. An error is thrown if a property value does not
 *                 match one of the provided enumValues.</td>
 *             </tr> 
 *             <tr>
 *               <td class="rt">properties</td>
 *               <td>yes</td>
 *               <td>{Object}</td>
 *               <td>A nested properties object for complex properties. Subproperties exposed using nested properties objects in the metadata can
 *                 be set using dot notation in the attribute. See the <a href="#subproperties">Subproperties</a> section for more details on 
 *                 working with subproperties.</td>
 *             </tr>
 *             <tr>
 *               <td class="rt">readOnly</td>
 *               <td>yes</td>
 *               <td>{boolean}</td>
 *               <td>Determines whether a property can be updated outside of the ViewModel.
 *                 False by default. If readOnly is true, the property can only be updated by the ViewModel or by the
 *                 components within the composite. This property only needs to be defined for the top level property, 
 *                 with subproperties inheriting that value.</td>
 *             </tr> 
 *             <tr>
 *               <td class="rt">type</td>
 *               <td>yes</td>
 *               <td>{string}</td>
 *               <td>The type of the property, following 
 *                 <a href="https://developers.google.com/closure/compiler/docs/js-for-compiler#types">Google's Closure Compiler</a> syntax.
 *                 We will parse string, number, boolean, array and object types for non data-bound attributes, but will not provide 
 *                 type checking for array and object elements. However, for documentation purposes, it may still be beneficial to provide
 *                 array and object element details using the Closure Compiler syntax.</td>
 *             </tr>  
 *             <tr>
 *               <td class="rt">value</td>
 *               <td>yes</td>
 *               <td>{object}</td>
 *               <td>An optional default value for a property. Only supported on the top level property and will not be evaluated if set on nested subproperties.</td>
 *             </tr>
 *             <tr>
 *               <td class="rt">writeback</td>
 *               <td>yes</td>
 *               <td>{boolean}</td>
 *               <td>Determines whether an expression bound to this property should be written back to. False by default.
 *                 If writeback is true, any updates to the property will result in an update to the expression. This property only needs to be defined
 *                 for the top level property, with subproperties inheriting that value.</td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>description</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A description for the property.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>displayName</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A user friendly, translatable name of the property.</td>
 *             </tr> 
 
 *             <tr>
 *               <td class="name"><code>exclusiveMaximum</code></td>
 *               <td>no</td>
 *               <td>{number}|{string}</td>
 *               <td>Validation metadata - specifies the <i>exclusive</i> high end of a possible range of values (e.g., "exclusiveMaximum": 1.0 → valid property value is <1.0). If the value is a string, then it is assumed to represent a dateTime value in the ISO 8601 extended date/time format.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>exclusiveMinimum</code></td>
 *               <td>no</td>
 *               <td>{number}|{string}</td>
 *               <td>Validation metadata - specifies the <i>exclusive</i> low end of a possible range of values (e.g., "exclusiveMinimum": 0.0 → valid property value is >0.0). If the value is a string, then it is assumed to represent a dateTime value in the ISO 8601 extended date/time format.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>extension</code></td> 
 *               <td>no</td>
 *               <td>{Object}</td>
 *               <td>Placeholder for Extension metadata.  Each section is identified by a key that specifies the downstream tool that will process this metadata.
 *                 <h6>For example:</h6> 
 *                 <table class="params">
 *                   <thead>
 *                     <tr>
 *                       <th>Name</th>
 *                       <th>Type</th>
 *                       <th>Description</th>
 *                     </tr>
 *                   </thead>
 *                   <tbody>
 *                     <tr>
 *                       <td class="name"><code>vbcs</code></td>
 *                       <td>{string}</td>
 *                       <td>Indentifies an object with VBCS-specific metadata</td>
 *                     </tr> 
 *                   </tbody>
 *                 </table>
 *                 </br>
 *                 Please consult the documentation for the downstream tool to determine what (if any) extension metadata is supported.
 *               </td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>maximum</code></td>
 *               <td>no</td>
 *               <td>{number}|{string}</td>
 *               <td>Validation metadata - specifies the <i>inclusive</i> high end of a possible range of values (e.g., "maximum": 1.0 → valid property value is <=1.0). 
 *               If the value is a string, then it is assumed to represent a dateTime value in the ISO 8601 extended date/time format.
 *               NOTE:  Use this in place of the deprecated max metadata property</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>minimum</code></td>
 *               <td>no</td>
 *               <td>{number}|{string}</td>
 *               <td>Validation metadata - specifies the <i>inclusive</i> low end of a possible range of values (e.g., "minimum": 0.0 → valid property value is >=0.0). 
 *               If the value is a string, then it is assumed to represent a dateTime value in the ISO 8601 extended date/time format.
 *               NOTE:  Use this in place of the deprecated min metadata property</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>placeholder</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>User-friendly, translatable hint text that appears in an empty input field at design time.</td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>propertyEditorValues</code></td>
 *               <td>no</td>
 *               <td>{Object}</td>
 *               <td>Additional design time metadata that enhances the <code>enumValues</code> run time metadata. 
 *                   The value is an Object with properties matching the values in the <code>enumValues</code> array. 
 *                   The corresponding value for each key is an Object with the following properties:
 *                   <h6>Properties</h6>
 *                   <table class="params">
 *                     <thead>
 *                       <tr>
 *                         <th>Name</th>
 *                         <th>Type</th>
 *                         <th>Description</th>
 *                       </tr>
 *                     </thead>
 *                     <tbody>
 *                       <tr>
 *                         <td class="name"><code>description</code></td>
 *                         <td>{string}</td>
 *                         <td>A translatable description for the value.</td>
 *                       </tr>
 *                       <tr>
 *                         <td class="name"><code>displayName</code></td>
 *                         <td>{string}</td>
 *                         <td>A displayable, translatable label for the value.</td>
 *                       </tr>
 *                       <tr>
 *                         <td class="name"><code>icon</code></td>
 *                         <td>{Object}</td>
 *                         <td>One or more optional images for representing the value. The object has the following properties:
 *                           <h6>Properties</h6>
 *                           <table class="params">
 *                             <thead>
 *                               <tr>
 *                                 <th>Name</th>
 *                                 <th>Type</th>
 *                                 <th>Description</th>
 *                               </tr>
 *                             </thead>
 *                             <tbody>
 *                               <tr>
 *                                 <td class="name"><code>iconPath</code></td>
 *                                 <td>{string}</td>
 *                                 <td>A relative path to the icon that represents the value.</td>
 *                               </tr>
 *                               <tr>
 *                                 <td class="name"><code>selectedIconPath</code></td>
 *                                 <td>{string}</td>
 *                                 <td>A relative path to the icon that represents the selected state of the value.</td>
 *                               </tr>
 *                               <tr>
 *                                 <td class="name"><code>hoverIconPath</code></td>
 *                                 <td>{string}</td>
 *                                 <td>A relative path to the icon that represents the hover state of the value.</td>
 *                               </tr>
 *                             </tbody>
 *                           </table>
 *                         </td> 
 *                       </tr>
 *                     </tbody>
 *                   </table>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>required</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>Specifies whether the property must have a valid value at run time. False by default.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>translatable</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>True if the <em>value</em> of this property (or its sub-properties, unless explicitly overridden) 
 *                   is eligible to be included when application resources are translated for Internationalization. False by default.
 *               </td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>units</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>User-friendly, translatable text string specifying what units are represented by a property value -- e.g., "pixels".</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>visible</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>Specifies whether the property should be visible at design time. True by default.</td>
 *             </tr> 
 *           </tbody>
 *         </table>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h3>Methods</h3>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Value</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>[method name]</code></td>
 *       <td>Object containing the following properties:
 *         <h6>Properties</h6>
 *         <table class="params">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Used at Runtime</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody> 
 *             <tr>
 *               <td class="rt">internalName</td>
 *               <td>yes</td>
 *               <td>{string}</td>
 *               <td>An optional ViewModel method name that is different from, but maps to this method.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>description</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A description for the method.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>displayName</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A user friendly, translatable name of the method.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>extension</code></td>
 *               <td>no</td> 
 *               <td>{Object}</td>
 *               <td>Placeholder for Extension metadata.  Each section is identified by a key that specifies the downstream tool that will process this metadata.
 *                 <h6>For example:</h6> 
 *                 <table class="params">
 *                   <thead>
 *                     <tr>
 *                       <th>Name</th>
 *                       <th>Type</th>
 *                       <th>Description</th>
 *                     </tr>
 *                   </thead>
 *                   <tbody>
 *                     <tr>
 *                       <td class="name"><code>vbcs</code></td>
 *                       <td>{string}</td>
 *                       <td>Indentifies an object with VBCS-specific metadata</td>
 *                     </tr> 
 *                   </tbody>
 *                 </table>
 *                 </br>
 *                 Please consult the documentation for the downstream tool to determine what (if any) extension metadata is supported.
 *               </td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>params</code></td>
 *               <td>no</td>
 *               <td>{Array<{Object}>}</td>
 *               <td>An array of objects describing the method parameter.  Each parameter object has the following properties:
 *                 <h6>Properties</h6>
 *                 <table class="params">
 *                  <thead>
 *                   <tr>
 *                    <th>Name</th>
 *                    <th>Type</th>
 *                    <th>Description</th>
 *                   </tr>
 *                  </thead>
 *                  <tbody>
 *                   <tr>
 *                    <td class="name"><code>description</code></td>
 *                    <td>{string}</td>
 *                    <td>A translatable description of the parameter</td>
 *                   </tr>
 *                   <tr>
 *                    <td class="name"><code>name</code></td>
 *                    <td>{string}</td>
 *                    <td>The name of the parameter.</td>
 *                   </tr> 
 *                   <tr>
 *                    <td class="name"><code>type</code></td>
 *                    <td>{string}</td>
 *                    <td>The type of the property, typically following <a href="https://github.com/google/closure-compiler/wiki/Annotating-JavaScript-for-the-Closure-Compiler">Google's Closure Compiler</a> syntax. The metadata also supports Typescript data types.</td>
 *                   </tr> 
 *                 </tbody>
 *                </table>
 *              </td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>return</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>The return type of the method, following Closure Compiler syntax.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>visible</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>Specifies whether the method should be visible at design time. True by default.</td>
 *             </tr> 
 *           </tbody>
 *         </table>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h3>Events</h3>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Value</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>[event name]</code></td>
 *       <td>Object containing the following properties:
 *         <h6>Properties</h6>
 *         <table class="params">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Used at Runtime</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody>
 *             <tr>
 *               <td class="name"><code>bubbles</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>Indicates whether the event bubbles up through the DOM or not. Defaults to false.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>cancelable</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>Indicates whether the event is cancelable or not. Defaults to false.</td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>description</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A description for the event.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>displayName</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A user friendly, translatable name of the event.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>detail</code></td>
 *               <td>no</td>
 *               <td>{object}</td>
 *               <td>Describes the properties available on the event's detail property which contains data passed when initializing the event.</p>
 *                 <h6>Properties</h6>
 *                 <table class="params">
 *                   <thead>
 *                     <tr>
 *                       <th>Name</th>
 *                       <th>Type</th>
 *                     </tr>
 *                   </thead>
 *                   <tbody>
 *                     <tr>
 *                       <td class="name"><code>[field name]</code></td>
 *                       <td>{description: string, type: string}</td>
 *                     </tr>
 *                   </tbody>
 *                 </table>
 *               </td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>extension</code></td>
 *               <td>no</td>
 *               <td>{Object}</td>
 *               <td>Placeholder for Extension metadata.  Each section is identified by a key that specifies the downstream tool that will process this metadata.
 *                 <h6>For example:</h6> 
 *                 <table class="params">
 *                   <thead>
 *                     <tr>
 *                       <th>Name</th>
 *                       <th>Type</th>
 *                       <th>Description</th>
 *                     </tr>
 *                   </thead>
 *                   <tbody>
 *                     <tr>
 *                       <td class="name"><code>vbcs</code></td>
 *                       <td>{string}</td>
 *                       <td>Indentifies an object with VBCS-specific metadata</td>
 *                     </tr> 
 *                   </tbody>
 *                 </table>
 *                 </br>
 *                 Please consult the documentation for the downstream tool to determine what (if any) extension metadata is supported. 
 *               </td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>visible</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>Specifies whether the event should be visible at design time. True by default.</td>
 *             </tr> 
 *           </tbody>
 *         </table>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h3>Slots</h3>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Value</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>[slot name]</code></td>
 *       <td>Object containing the following properties:
 *         <h6>Properties</h6>
 *         <table class="params">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Used at Runtime</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody>  
 *             <tr>
 *               <td class="name"><code>description</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A description for the slot.</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>displayName</code></td>
 *               <td>no</td>
 *               <td>{string}</td>
 *               <td>A user friendly, translatable name of the slot.</td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>extension</code></td>
 *               <td>no</td>
 *               <td>{Object}</td>
 *               <td>Placeholder for Extension metadata.  Each section is identified by a key that specifies the downstream tool that will process this metadata.
 *                 <h6>For example:</h6> 
 *                 <table class="params">
 *                   <thead>
 *                     <tr>
 *                       <th>Name</th>
 *                       <th>Type</th>
 *                       <th>Description</th>
 *                     </tr>
 *                   </thead>
 *                   <tbody>
 *                     <tr>
 *                       <td class="name"><code>vbcs</code></td>
 *                       <td>{string}</td>
 *                       <td>Indentifies an object with VBCS-specific metadata</td>
 *                     </tr> 
 *                   </tbody>
 *                 </table>
 *                 </br>
 *                 Please consult the documentation for the downstream tool to determine what (if any) extension metadata is supported.
 *               </td>
 *             </tr> 
 *             <tr>
 *               <td class="name"><code>visible</code></td>
 *               <td>no</td>
 *               <td>{boolean}</td>
 *               <td>Specifies whether the slot should be visible at design time. True by default.</td>
 *             </tr>
 *           </tbody>
 *         </table>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h3>Example of Run Time Metadata</h3>
 * <p>The JET framework will ignore "extension" fields. Extension fields cannot be defined at 
 *   the first level of the "properties", "methods", "events", or "slots" objects.</p>
 * <pre class="prettyprint"><code>
 * {
 *  "name": "demo-card",
 *  "version": "1.0.2",
 *  "jetVersion": ">=3.0.0 <5.0.0",
 *  "properties": {
 *    "currentImage" : {
 *      "type": "string",
 *      "readOnly": true
 *    },
 *    "images": {
 *      "type": "Array<string>"
 *    },
 *    "isShown": {
 *      "type": "boolean",
 *      "value": true
 *    }
 *  },
 *  "methods": {
 *    "nextImage": {
 *      "internalName": "_nextImg"
 *      "extension": "This is where a composite can store additional data."
 *    },
 *    "prevImage": {}
 *   },
 *   "events": {
 *     "cardclick": {}
 *   }
 * }
 * </code></pre>
 *
 * <h2 id="properties">Properties
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#properties"></a>
 * </h2>
 * <p>
 * Properties defined in provided Metadata will be made available through the <code>$properties</code> property of the View binding context
 * and through the <code>properties</code> property on the context passed to the provided ViewModel constructor function or lifecycle listeners. 
 * The application can access the composite component properties by accessing them directly from the 
 * DOM element. Using the DOM setAttribute and removeAttribute APIs will also result in property updates. Changes made to properties will
 * result in a [property]Changed event being fired for that property if the property was modified internally by the composite or externally
 * by the application if the composite has been upgraded and its busy state resolved. Early property sets before the composite has been upgraded
 * while allowed, will not result in [property]Changed events and will be passed to the component as part of its initial state.
 * </p>
 *
 * <h3 id="attr-to-prop-mapping">Property-to-Attribute Mapping
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#attr-to-prop-mapping"></a>
 * </h3>
 * <p>
 * The following rules apply when mapping property to attribute names:
 * 
 * <ul>
 *  <li>Attribute names are case insensitive. CamelCased properties are mapped to 
 *    kebab-cased attribute names by inserting a dash before the uppercase letter and converting that letter to lower case, 
 *    e.g. a "chartType" property will be mapped to a "chart-type" attribute.</li>
 *  <li> The reverse occurs when mapping a property name from an attribute name.</li>
 * </ul>
 * </p>
 *
 * <h3 id="subproperties">Subproperties
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#subproperties"></a>
 * </h3>
 * <p>
 * Subproperties can be exposed and documented by adding nested properties objects in the composite <a href="#metadata">metadata</a>. Event listeners
 * should be added for the top level property, with applications checking the event's <a href="#events">subproperty</a> field to access information about the
 * subproperty changes. Subproperties can be set declaratively using dot notation, e.g. person.first-name="{{name}}".  Setting overlapping attributes, 
 * e.g. person="{{personInfo}}" person.first-name="{{name}}" will cause an error to be thrown.
 * </p> 
 * 
 * <p>Subproperties can also be set programmatically using the <code>set/getProperty</code> methods. 
 * Note that while setting the subproperty using dot notation via the element's top level property is allowed, the setProperty method must be used in order
 * to trigger a property change event.
 * <pre class="prettyprint"><code>
 * element.setProperty("person.firstName", Fred);
 * var firstName = element.getProperty("person.firstName");
 * </code></pre>
 * </p>
 * 
 * <h2 id="styling">Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling"></a>
 * </h2>
 * <p>
 * Composite component styling can be done via provided css. The JET framework will add the <code>oj-complete</code> class to the composite DOM 
 * element after metadata properties have been resolved. To prevent a flash of unstyled content before the composite properties have been setup, 
 * the composite css can include the following rule to hide the composite until the <code>oj-complete</code> class is set on the element.
 * <pre class="prettyprint"><code>
 * my-chart:not(.oj-complete) {
 *   visibility: hidden;
 * }
 * </code></pre>
 * </p>
 *
 * <p>
 * Composite CSS will not be scoped to the composite component and selectors will need to be appropriately selective. We recommend scoping CSS classes 
 * and prefixing class names with the composite name as seen in the example below. <b>Note that we do not recommend overriding JET component CSS.
 * Composites should only update JET component styling via SASS variables.</b>
 * <pre class="prettyprint"><code>
 * my-chart .my-chart-text {
 *   color: white;
 * }
 * </code></pre>
 * </p>
 *
 * <h2 id="events">Events
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#events"></a>
 * </h2>
 * <p>
 * Composite components fire the following events. Any custom composite events should be created and fired by the composite's ViewModel and documented in the metadata
 * for design and run time environments as needed.  In addition to standard add/removeEventListener syntax, declarative event listeners will also be supported for
 * custom composite events and [property]Changed events.  As a result, if a composite declares an event of type "customType", applications can listen by setting the
 * "onCustomType" property or the "on-custom-type" attribute.  Similarly, property change listeners can be set via the appropriate "on[property]Changed" property or
 * "on-[property-name]-changed" attribute. Expression syntax can be used in the on-[event-name] attributes, but we do not support executing arbitrary JavaScript like those for
 * native event attributes like onclick.
 * <ul>
 *  <li><b><i>[property]Changed</i></b> - Fired when a property is modified and contains the following fields in its event detail object:
 *    <ul>
 *      <li>previousValue - The previous property value</li>
 *      <li>value - The new property value</li>
 *      <li>updatedFrom - Where the property was updated from. Supported values are:
 *        <ul>
 *          <li>internal - The View or ViewModel</li>
 *          <li>external - The DOM Element either by its property setter, setAttribute, or external data binding.</li>
 *        </ul>
 *      </li>
 *      <li>subproperty - An object containing information about the subproperty that changed with the following fields:
 *        <ul>
 *          <li>path - The subproperty path that changed</li>
 *          <li>previousValue - The previous subproperty value</li>
 *          <li>value - The new subproperty value</li>
 *        </ul>
 *      </li>
 *    </ul>
 *  </li>
 * </ul>
 * </p>
 *  
 * <h2 id="lifecycle">Lifecycle
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#lifecycle"></a>
 * </h2>
 * <p>
 * If a ViewModel is provided for a composite component, the following optional
 * callback methods can be defined on its ViewModel and will be called at each stage of the composite
 * component's lifecycle. The ViewModel provided at registration can either be a function which will be 
 * treated as a constructor that will be invoked to create the ViewModel instance or an object which will
 * be treated as a singleton instance. <b>The Object instance return type for the ViewModel
 * is deprecated in 5.0.0.</b>
 *
 * <h4 class="name">initialize<span class="signature">(context)</span></h4>
 * <p>
 * <b>Deprecated since 5.0.0.</b> This optional method may be implemented on the ViewModel to perform initialization tasks.
 * This method will be invoked only if the ViewModel specified during registration is an object instance as opposed to a constructor function.
 * If the registered ViewModel is a constructor function, the same context object will be passed to the constructor function instead.
 * This method can return 1) nothing in which case the original model instance will be used, 2) a new model instance which will replace the original, 
 * or 3) a Promise which resolves to a new model instance which will replace the original and delay additional lifecycle phases until it is resolved.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>context</code></td>
 *       <td>Object</td>
 *       <td>An object with the following key-value pairs:
 *          <h6>Properties:</h6>
 *          <table class="params">
 *            <thead>
 *              <tr>
 *                <th>Name</th>
 *                <th>Type</th>
 *                <th>Description</th>
 *              </tr>
 *            </thead>
 *            <tbody>
 *              <tr>
 *                <td class="name"><code>element</code></td>
 *                <td>Node</td>
 *                <td>DOM element where the View is attached.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>properties</code></td>
 *                <td>Object</td>
 *                <td>A map of the composite component's current properties and values.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>props</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use properties instead.</b> A Promise evaluating to the composite component's properties.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotNodeCounts</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use slotCounts instead.</b>A Promise evaluating to a map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotCounts</code></td>
 *                <td>Object</td>
 *                <td>A map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>unique</code></td>
 *                <td>string</td>
 *                <td>A unique string that can be used for unique id generation.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>uniqueId</code></td>
 *                <td>string</td>
 *                <td>The ID of the composite component if specified. Otherwise, it is the same as <code>unique</code>.</td>
 *              </tr>
 *            </tbody>
 *          </table>
 *        </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h4 class="name">activated<span class="signature">(context)</span></h4>
 * <p>
 * This optional method may be implemented on the ViewModel and will be invoked after the ViewModel is initialized.
 * This method can return a Promise which will delay additional lifecycle phases until it is resolved and can be used 
 * as a hook for data fetching.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>context</code></td>
 *       <td>Object</td>
 *       <td>An object with the following key-value pairs:
 *          <h6>Properties:</h6>
 *          <table class="params">
 *            <thead>
 *              <tr>
 *                <th>Name</th>
 *                <th>Type</th>
 *                <th>Description</th>
 *              </tr>
 *            </thead>
 *            <tbody>
 *              <tr>
 *                <td class="name"><code>element</code></td>
 *                <td>Node</td>
 *                <td>DOM element where the View is attached.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>properties</code></td>
 *                <td>Object</td>
 *                <td>A map of the composite component's current properties and values.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>props</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use properties instead.</b> A Promise evaluating to the composite component's properties.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotNodeCounts</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use slotCounts instead.</b>A Promise evaluating to a map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotCounts</code></td>
 *                <td>Object</td>
 *                <td>A map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>unique</code></td>
 *                <td>string</td>
 *                <td>A unique string that can be used for unique id generation.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>uniqueId</code></td>
 *                <td>string</td>
 *                <td>The ID of the composite component if specified. Otherwise, it is the same as <code>unique</code>.</td>
 *              </tr>
 *            </tbody>
 *          </table>
 *        </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h4 class="name">attached<span class="signature">(context)</span></h4>
 * <p>
 * This optional method is deprecated in 4.2.0 in favor of the connected method. This method is invoked after the
 * View is inserted into the DOM and will only be called once. Note that if the composite needs to add/remove event listeners,
 * we recommend using the connected/disconnected methods.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>context</code></td>
 *       <td>Object</td>
 *       <td>An object with the following key-value pairs:
 *          <h6>Properties:</h6>
 *          <table class="params">
 *            <thead>
 *              <tr>
 *                <th>Name</th>
 *                <th>Type</th>
 *                <th>Description</th>
 *              </tr>
 *            </thead>
 *            <tbody>
 *              <tr>
 *                <td class="name"><code>element</code></td>
 *                <td>Node</td>
 *                <td>DOM element where the View is attached.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>properties</code></td>
 *                <td>Object</td>
 *                <td>A map of the composite component's current properties and values.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>props</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use properties instead.</b> A Promise evaluating to the composite component's properties.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotNodeCounts</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use slotCounts instead.</b>A Promise evaluating to a map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotCounts</code></td>
 *                <td>Object</td>
 *                <td>A map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>unique</code></td>
 *                <td>string</td>
 *                <td>A unique string that can be used for unique id generation.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>uniqueId</code></td>
 *                <td>string</td>
 *                <td>The ID of the composite component if specified. Otherwise, it is the same as <code>unique</code>.</td>
 *              </tr>
 *            </tbody>
 *          </table>
 *        </td>
 *     </tr>
 *   </tbody>
 * </table>
 * 
 * <h4 class="name">connected<span class="signature">(context)</span></h4>
 * <p>
 * This optional method may be implemented on the ViewModel and will be invoked after the View is first inserted into the DOM and then
 * each time the composite is reconnected to the DOM after being disconnected. Note that if the composite needs to add/remove event listeners,
 * we recommend using this and the disconnected methods.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>context</code></td>
 *       <td>Object</td>
 *       <td>An object with the following key-value pairs:
 *          <h6>Properties:</h6>
 *          <table class="params">
 *            <thead>
 *              <tr>
 *                <th>Name</th>
 *                <th>Type</th>
 *                <th>Description</th>
 *              </tr>
 *            </thead>
 *            <tbody>
 *              <tr>
 *                <td class="name"><code>element</code></td>
 *                <td>Node</td>
 *                <td>DOM element where the View is attached.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>properties</code></td>
 *                <td>Object</td>
 *                <td>A map of the composite component's current properties and values.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>props</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use properties instead.</b> A Promise evaluating to the composite component's properties.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotNodeCounts</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use slotCounts instead.</b>A Promise evaluating to a map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotCounts</code></td>
 *                <td>Object</td>
 *                <td>A map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>unique</code></td>
 *                <td>string</td>
 *                <td>A unique string that can be used for unique id generation.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>uniqueId</code></td>
 *                <td>string</td>
 *                <td>The ID of the composite component if specified. Otherwise, it is the same as <code>unique</code>.</td>
 *              </tr>
 *            </tbody>
 *          </table>
 *        </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h4 class="name">bindingsApplied<span class="signature">(context)</span></h4>
 * <p>
 * This optional method may be implemented on the ViewModel and will be invoked after the bindings are applied on this View.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>context</code></td>
 *       <td>Object</td>
 *       <td>An object with the following key-value pairs:
 *          <h6>Properties:</h6>
 *          <table class="params">
 *            <thead>
 *              <tr>
 *                <th>Name</th>
 *                <th>Type</th>
 *                <th>Description</th>
 *              </tr>
 *            </thead>
 *            <tbody>
 *              <tr>
 *                <td class="name"><code>properties</code></td>
 *                <td>Object</td>
 *                <td>A map of the composite component's current properties and values.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>props</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use properties instead.</b> A Promise evaluating to the composite component's properties.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotNodeCounts</code></td>
 *                <td>Promise</td>
 *                <td><b>Deprecated: use slotCounts instead.</b>A Promise evaluating to a map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>slotCounts</code></td>
 *                <td>Object</td>
 *                <td>A map of slot name to assigned nodes count for the View.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>unique</code></td>
 *                <td>string</td>
 *                <td>A unique string that can be used for unique id generation.</td>
 *              </tr>
 *              <tr>
 *                <td class="name"><code>uniqueId</code></td>
 *                <td>string</td>
 *                <td>The ID of the composite component if specified. Otherwise, it is the same as <code>unique</code>.</td>
 *              </tr>
 *            </tbody>
 *          </table>
 *        </td>
 *     </tr>
 *   </tbody>
 * </table>
 * 
 * <h4 class="name">propertyChanged<span class="signature">(context)</span></h4>
 * <p>
 * This optional method may be implemented on the ViewModel and will be invoked when properties are updated before the [property]Changed event is fired.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>property</td>
 *       <td>string</td>
 *       <td>The property that changed.</td>
 *     </tr>
 *     <tr>
 *       <td>value</td>
 *       <td>*</td>
 *       <td>The current value of the property that changed.</td>
 *     </tr>
 *     <tr>
 *       <td>previousValue</td>
 *       <td>*</td>
 *       <td>The previous value of the property that changed.</td>
 *     </tr>
 *     <tr>
 *       <td>updatedFrom</td>
 *       <td>string</td>
 *       <td>
 *         Where the property was updated from. Supported values are:
 *         <ul>
 *           <li>external - By the application, using either the element's property setter, setAttribute, or external data binding.</li>
 *           <li>internal - By the component, e.g. after user interaction with a text field or selection.</li>
 *         </ul>  
 *       </td>
 *     </tr>
 *     <tr>
 *       <td>subproperty</td>
 *       <td>Object</td>
 *       <td>An object holding information about the subproperty that changed.
 *         <table class="props">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody>
 *             <tr>
 *               <td>path</td>
 *               <td>string</td>
 *               <td>
 *                 The subproperty path that changed, starting from the top level 
 *                 property with subproperties delimited by '.'.
 *               </td>
 *             </tr>
 *             <tr>
 *               <td>value</td>
 *               <td>*</td>
 *               <td>The current value of the subproperty that changed.</td>
 *             </tr>
 *             <tr>
 *               <td>previousValue</td>
 *               <td>*</td>
 *               <td>The previous value of the subproperty that changed.</td>
 *             </tr>
 *           </tbody>
 *         </table>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 * 
 * <h4 class="name">detached<span class="signature">(element)</span></h4>
 * <p>
 * This method is deprecated in 4.2.0 to the renamed disconnected method with the same behavior.
 * This optional method may be implemented on the ViewModel and will be invoked when this composite component is detached from the DOM.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>element</code></td>
 *       <td>Node</td>
 *       <td>The composite component DOM element.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * 
 * <h4 class="name">disconnected<span class="signature">(element)</span></h4>
 * <p>
 * This optional method may be implemented on the ViewModel and will be invoked when this composite component is disconnected from the DOM.
 * </p>
 * <h5>Parameters:</h5>
 * <table class="params">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>element</code></td>
 *       <td>Node</td>
 *       <td>The composite component DOM element.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * </p>
 *
 * <h2 id="writeback">Data Binding and Expression Writeback
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#writeback"></a>
 * </h2>
 * <p>Besides string literals, composite attributes can be set using expression syntax which is currently compatible with knockout expression syntax.
 * Applications can control expression writeback in the composite component by using {{}} syntax for two-way writable binding expressions
 * or [[]] for one-way only expressions. In the example below, the salesData expression will not be written back to if the 'data' property
 * is updated by the composite component's ViewModel. The 'data' property will contain the current value, but the salesData expression will not
 * be updated. Please note that this will cause the expression and property values to be out of sync until the salesData expression is updated in the
 * application's ViewModel. Alternatively, if the 'axisLabels' property is updated by the ViewModel, both the 'axisLabel' property and the 
 * showAxisLabels expression will contain the updated value.
 * <pre class="prettyprint"><code>
 * &lt;my-chart data="[[salesData]]" axis-labels={{showAxisLabels}} ... >
 * &lt;/my-chart>
 * </code></pre>
 * </p>
 * <h3>readOnly</h3>
 * <p>The composite component Metadata also defines properties to control expression writeback and property updates.
 * If a property's readOnly option is omitted, the value is false by default, meaning the property can be updated outside of the
 * composite component.  If readOnly is true, the property can only be updated inside of the composite component by the ViewModel or View.
 * </p>
 * <h3>writeback</h3>
 * <p>If a property's Metadata defines its writeback property as true (false by default), any bound attribute expressions will
 * be updated when the property is updated unless the attribute binding was done with a one-way "[[]]"" binding syntax.</p>
 *
 * <h2 id="slotting">Slotting
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#slotting"></a>
 * </h2>
 * <p>
 * Complex composite components which can contain additional composites and/or content for child facets defined in its associated View can be constructed via slotting.
 * See the <a href="oj.ojBindSlot.html">oj-bind-slot</a> API doc for more information.
 * </p>
 *
 * <h2 id="bindorder-section">
 *   Binding Order
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#bindorder-section"></a>
 * </h2>
 * <p>The following steps will occur when processing the binding for a composite component:
 * <ol>
 *  <li>Apply bindings to children using the composite component's binding context.</li>
 *  <li>Create a slot map assigning component child nodes to View slot elements.
 *    <ol>
 *      <li>At this point the component child nodes are removed from the DOM and live in the slot map.</li>
 *    </ol>
 *  </li>
 *  <li>Insert the View and apply bindings to it with the ViewModel's binding context.
 *    <ol>
 *      <li>The composite's children will be 'slotted' into their assigned View slots.</li>
 *      <li>The oj-bind-slot's slot attribute, which is "" by default, will override its assigned node's slot attribute.</li>
 *    </ol>
 *  </li>
 * </ol>
 * </p>
 *
 * @namespace
 * @since 2.0
 */
oj.Composite = {};

/**
 * Returns a Promise resolving with the composite metadata with the given name or null if the composite has not been registered.
 * @param {string} name The component name, which should contain a dash '-' and not be a reserved tag name.
 * @return {Promise|null}
 * @ojdeprecated {since: '5.0.0', description: 'Use oj.Composite.getComponentMetadata instead.'}
 * 
 * @export
 * @memberof oj.Composite
 *
 */
oj.Composite.getMetadata = function(name)
{
  var metadata = oj.Composite.getComponentMetadata(name);
  return metadata ? Promise.resolve(metadata) : null;
};

/**
 * Returns the composite metadata with the given name or null if the composite has not been registered.
 * @param {string} name The component name, which should contain a dash '-' and not be a reserved tag name.
 * @return {Object|null}
 * 
 * @export
 * @memberof oj.Composite
 * @since 5.0.0
 * @ojstatus preview
 *
 */
oj.Composite.getComponentMetadata = function(name)
{
  // We have one registry where custom elements, definitional elements, and composites are all stored with
  // the JET framework so we need to check to see if the element is a composite before returning its metadata
  var info = oj.BaseCustomElementBridge.getRegistered(name);
  if (info && info.composite)
  {
    // Descriptor is guaranteed to be there for registered elements because we throw an error at registration
    // time if none is given
    var descriptor = oj.BaseCustomElementBridge.__GetDescriptor(name);
    return descriptor[oj.BaseCustomElementBridge.DESC_KEY_META];
  }
  return null;
};

/**
 * Registers a composite component
 * @param {string} name The component name, which should contain a dash '-' and not be a reserved tag name.
 * @param {Object} descriptor The registration descriptor. The descriptor will contain keys for Metadata, View, ViewModel
 * and CSS that are detailed below. At a minimum a composite must register Metadata and View files, but all others are optional.
 * The composite resources should be mapped directly to each descriptor key. The support for an object with an 'inline' key mapped
 * to the resource has been deprecated in 5.0.0.
 * See the <a href="#registration">registration section</a> above for a sample usage.
 * @param {Object} descriptor.metadata A JSON formatted object describing the composite APIs. See the <a href="#metadata">metadata documentation</a> for more info.
 * @param {string} descriptor.view A string, array of DOM nodes, or document fragment representing the HTML that will be used for the composite.
 *                                 <b>The array of DOM nodes and document fragment types are deprecated in 5.0.0.</b>
 * @param {string} descriptor.css (Deprecated) A string containing the composite CSS. <b>Note that this key should not be used if the composite
 *                                styles contain references to any external resources and is deprecated in 4.1.0. require-css, a RequireJS CSS plugin, is the
 *                                current recommendation for CSS loading.</b>
 * @param {Object} descriptor.viewModel This option is only applicable to composites hosting a Knockout template
 *                                      with a ViewModel and ultimately resolves to a constructor function or object instance. 
 *                                      <b>The Object instance return type for the ViewModel is deprecated in 5.0.0.</b>
 *                                      If the initial ViewModel resolves to an object instance, the initialize lifecycle listener 
 *                                      will be called. See the <a href="#lifecycle">initialize documentation</a> for more information.
 * @param {function(string, string, Object, function(string))} descriptor.parseFunction The function that will be called to parse attribute values.
 *                                                              Note that this function is only called for non bound attributes. The parseFunction will take the following parameters:
 * <ul>
 *  <li>{string} value: The value to parse.</li>
 *  <li>{string} name: The name of the property.</li>
 *  <li>{Object} meta: The metadata object for the property which can include its type, default value, 
 *      and any extensions that the composite has provided on top of the required metadata.</li>
 *  <li>{function(string)} defaultParseFunction: The default parse function for the given attribute 
 *      type which is used when a custom parse function isn't provided and takes as its parameters 
 *      the value to parse.</li>
 * </ul>
 *
 * @export
 * @memberof oj.Composite
 *
 */
oj.Composite.register = function(name, descriptor)
{
  oj.CompositeElementBridge.register(name, descriptor);
};

/**
 * Finds the containing composite component for a given node. If the immediate enclosing
 * composite component is contained by another composite, the method will keep
 * walking up the composite hierarchy until the top-level composite
 * or the optional 'stopBelow' element is reached
 * 
 * @param {Node} node the DOM node whose containing composite should be returned
 * @param {Element=} stopBelow the element where search should stop
 * @return {Element|null} the containing composite
 * 
 * This method is currently intended for internal use only, and it is not exported
 * @ignore
 */
oj.Composite.getContainingComposite = function(node, stopBelow)
{
  var composite = null;
  
  while(node)
  {
    node = oj.CompositeTemplateRenderer.getEnclosingComposite(node);
    if (node)
    {
      if (stopBelow && !(node.compareDocumentPosition(stopBelow) & 16/*contained by*/))
      {
        break;
      }
      composite = node;
    }
  }
  
  return composite;
};


/**
 * @ignore
 */
oj.Composite.getBindingProviderName = function(elem)
{
  return (elem ? elem[oj.Composite.__BINDING_PROVIDER] : null);
}

/**
 * @ignore
 */
oj.Composite.__COMPOSITE_PROP = '__oj_composite';

/**
 * @ignore
 */
oj.Composite.__BINDING_PROVIDER = '__oj_binding_prvdr';
/**
 * JET component custom element bridge.
 * 
 * Composite connnected callbacks occur asynchronously so we cannot
 * guarantee that child composite properties can be accessed before the
 * child busy state resolves.
 * 
 * Composite code and applications should always wait on the element or page level 
 * busy context before accessing properties or methods.
 * 
 * @class
 * @ignore
 */
oj.CompositeElementBridge = {};

/**
 * Prototype for the JET component custom element bridge instance
 */
oj.CompositeElementBridge.proto = Object.create(oj.BaseCustomElementBridge.proto);

oj.CollectionUtils.copyInto(oj.CompositeElementBridge.proto,
{
  beforePropertyChangedEvent: function(element, property, detail)
  {
    var vmContext = {'property': property};
    oj.CollectionUtils.copyInto(vmContext, detail, undefined, true);
    oj.CompositeTemplateRenderer.invokeViewModelMethod(this._VIEW_MODEL, 'propertyChanged', [vmContext]);
  },

  AddComponentMethods: function(proto) 
  {
    // Add subproperty getter/setter
    proto['setProperty'] = function(prop, value) {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      if (!bridge.SaveEarlyPropertySet(prop, value))
        bridge.SetProperty(this, prop, value, this, true);
    };
    proto['getProperty'] = function(prop) {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      return bridge.GetProperty(this, prop, this);
    };
    proto['_propsProto']['setProperty'] =  function(prop, value) {
      // 'this' is the property object we pass to the ViewModel to track internal property changes
      this._BRIDGE.SetProperty(this._ELEMENT, prop, value, this, false);
    };
    proto['_propsProto']['getProperty'] = function(prop) {
      // 'this' is the property object we pass to the ViewModel to track internal property changes
      return this._BRIDGE.GetProperty(this, prop, this);
    };
    // Always add automation methods, but if the ViewModel defines overrides, wrap the overrides
    // and pass the default implementation in as the last parameter to the ViewModel's method.
    proto['getNodeBySubId'] = function(locator) {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      var viewModel = bridge._getViewModel(this);
      if (viewModel.getNodeBySubId)
        return viewModel.getNodeBySubId(locator, bridge._getNodeBySubId.bind(this));
      else
        return bridge._getNodeBySubId.bind(this)(locator);
    };
    proto['getSubIdByNode'] = function(node) {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      var viewModel = bridge._getViewModel(this);
      if (viewModel.getSubIdByNode)
        return viewModel.getSubIdByNode(node, bridge._getSubIdByNode.bind(this));
      else
        return bridge._getSubIdByNode.bind(this)(node);
    };
  },

  CreateComponent: function(element) 
  {
    // Setup the ViewModel context to pass to lifecycle listeners
    var slotNodeCounts = {};
    // Generate slot map before we update DOM with view nodes
    var slotMap = oj.BaseCustomElementBridge.getSlotMap(element, true);
    for (var slot in slotMap)
      slotNodeCounts[slot] = slotMap[slot].length;
    this._SLOT_MAP = slotMap;
    var vmContext = {
      'element': element,
      'props': Promise.resolve(this._PROPS),
      'properties': this._PROPS,
      'slotNodeCounts': Promise.resolve(slotNodeCounts),
      'slotCounts': slotNodeCounts,
      'unique': oj.BaseCustomElementBridge.__GetUnique()
    };
    vmContext['uniqueId'] = element.id ? element.id : vmContext['unique'];
    this._VM_CONTEXT = vmContext;

    var model = oj.BaseCustomElementBridge.__GetDescriptor(element.tagName)[oj.BaseCustomElementBridge.DESC_KEY_VIEW_MODEL];
    if (typeof model === 'function')
      model = new model(vmContext);
    else // The initialize callback is deprecated in 5.0.0. If the function returns a value, use it as the new model instance.
      model = oj.CompositeTemplateRenderer.invokeViewModelMethod(model, 'initialize', [vmContext]) || model;
    this._VIEW_MODEL = model;

    // This method can return a Promise which will delay additional lifecycle phases until it is resolved.
    var activatedPromise = oj.CompositeTemplateRenderer.invokeViewModelMethod(model, 'activated', [vmContext]) || Promise.resolve(true);

    var bridge = this;
    activatedPromise.then(function(values)
    {
      var params = {
        props: bridge._PROPS,
        slotMap: bridge._SLOT_MAP,
        slotNodeCounts: slotNodeCounts,
        unique: bridge._VM_CONTEXT['unique'],
        uniqueId: bridge._VM_CONTEXT['uniqueId'],
        viewModel: bridge._VIEW_MODEL,
        viewModelContext: bridge._VM_CONTEXT
      };
      
      // Store the name of the binding provider on the element when we are about 
      // to insert the view. This will allow custom elements within the view to look
      // up the binding provider used by the composite (currently only KO).
      Object.defineProperty(element, oj.Composite.__BINDING_PROVIDER, {value: 'knockout'});

      if (oj.Components)
      {
        oj.Components.unmarkPendingSubtreeHidden(element);
      }

      var cache = oj.BaseCustomElementBridge.__GetCache(element.tagName);
      // Need to clone nodes first
      var view = oj.CompositeElementBridge._getDomNodes(cache.view, element);
      oj.CompositeTemplateRenderer.renderTemplate(params, element, view);

      // Set flag when we can fire property change events
      bridge.__READY_TO_FIRE = true;

      // Resolve the component busy state 
      bridge.resolveDelayedReadyPromise();
    });
  },

  DefineMethodCallback: function (proto, method, methodMeta) 
  {
    proto[method] = function()
    {
      var methodName = methodMeta['internalName'] || method;
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      var viewModel = bridge._getViewModel(this);
      return viewModel[methodName].apply(viewModel, arguments);
    };
  },
  
  DefinePropertyCallback: function (proto, property, propertyMeta) 
  {
    var set = function(value, bOuterSet)
    {
      // Properties can be set before the component is created. These early
      // sets are actually saved until after component creation and played back.
      if (!this._BRIDGE.SaveEarlyPropertySet(property, value))
      {
        // Property trackers are observables are referenced when the property is set or retrieved,
        // which allows us to automatically update the View when the property is mutated.
        var propertyTracker = oj.CompositeElementBridge._getPropertyTracker(this._BRIDGE, property);
        var previousValue = propertyTracker.peek();
        if (!oj.BaseCustomElementBridge.__CompareOptionValues(property, propertyMeta, value, previousValue)) // We should consider supporting custom comparators
        {
          // Skip validation for inner sets so we don't throw an error when updating readOnly writeable properties
          if (bOuterSet)
            value = this._BRIDGE.ValidatePropertySet(this._ELEMENT, property, value)

          if (propertyMeta._eventListener)
          {
            this._BRIDGE.SetEventListenerProperty(this._ELEMENT, property, value);
          }
          propertyTracker(value);

          if (!propertyMeta._derived)
          {
            var updatedFrom = bOuterSet ? 'external' : 'internal';
            oj.BaseCustomElementBridge.__FirePropertyChangeEvent(this._ELEMENT, property, value, previousValue, updatedFrom);
          }
        }
      }
    }

    // Called on the ViewModel props object
    var innerSet = function(value)
    {
      set.bind(this)(value, false);
    }

    // Called on the custom element
    var outerSet = function(value)
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      set.bind(bridge._PROPS)(value, true);
    }

    var get = function(bOuterSet)
    {
      var propertyTracker = oj.CompositeElementBridge._getPropertyTracker(this._BRIDGE, property);
      // If the attribute has not been set, return the default value
      // Calling .peek() lets us check the propertyTracker value without creating a dependency
      var value = bOuterSet ? propertyTracker.peek() : propertyTracker();
      if (value === undefined)
      {
        value = propertyMeta['value'];
        // Make a copy if the default value is an Object or Array to prevent modification
        // of the metadata copy and store in the propertyTracker so we have a copy
        // to modify in place for the object case
        if (Array.isArray(value))
          value = value.slice();
        else if (typeof value === 'object')
          value = oj.CollectionUtils.copyInto({}, value, undefined, true);
        propertyTracker(value);
      }
      return value;
    }

    // Called on the ViewModel props object
    var innerGet = function()
    {
      return get.bind(this, false)();
    }

    // Called on the custom element
    var outerGet = function()
    {
      var bridge = oj.BaseCustomElementBridge.getInstance(this);
      return get.bind(bridge._PROPS, true)();
    }

    // Don't add event listener properties for inner props
    if (!propertyMeta._derived)
      oj.BaseCustomElementBridge.__DefineDynamicObjectProperty(proto['_propsProto'], property, innerGet, innerSet);
    oj.BaseCustomElementBridge.__DefineDynamicObjectProperty(proto, property, outerGet, outerSet);
  },

  GetMetadata: function(descriptor)
  {
    // Composites have a public getMetadata API so we cannot directly modify the
    // original metadata object when we add additional info for on[PropertyName] properties
    return descriptor['_metadata'];
  },

  HandleDetached: function(element) 
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.HandleDetached.call(this, element);
    
    // Detached is deprecated in 4.2.0 for disconnected
    oj.CompositeTemplateRenderer.invokeViewModelMethod(this._VIEW_MODEL, 'detached', [element]);
    oj.CompositeTemplateRenderer.invokeViewModelMethod(this._VIEW_MODEL, 'disconnected', [element]);
  },  

  HandleReattached: function(element) {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.HandleReattached.call(this, element);

    oj.CompositeTemplateRenderer.invokeViewModelMethod(this._VIEW_MODEL, 'connected', [this._VM_CONTEXT]);
  },

  InitializeElement: function(element)
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializeElement.call(this, element);

    if (oj.Components)
      oj.Components.markPendingSubtreeHidden(element);

    // Cache the View
    var cache = oj.BaseCustomElementBridge.__GetCache(element.tagName);
    if (!cache.view)
    {
      var view = oj.BaseCustomElementBridge.__GetDescriptor(element.tagName)[oj.BaseCustomElementBridge.DESC_KEY_VIEW];
      // when multiple instances of the same CCA are on the same page, because of the async 
      // nature, we could end up with multiple promises created on the same view. The first
      // resolved promise will set up cache.view, all others should just use the cached 
      // view instead of parsing it again. So here we check existence of cache in the resolve
      // callback to avoid parsing the view multiple times.
      if (!cache.view) 
      {
        if (typeof(view) === 'string')
          cache.view = oj.CompositeElementBridge._getDomNodes(view, element);
        cache.view = view;
      }
    }
    
    // Cache the CSS
    if (!cache.css)
    {
      // The CSS Promise will be null if loaded by the require-css plugin
      var css = oj.BaseCustomElementBridge.__GetDescriptor(element.tagName)[oj.BaseCustomElementBridge.DESC_KEY_CSS];
      // CSS is optional so we need to check if it was provided
      if (css)
      {
        var style = document.createElement('style');
        style.type = 'text/css';
        if (style.styleSheet) // for IE
          style.styleSheet.cssText = css;
        else
          style.appendChild(document.createTextNode(css)); // @HTMLUpdateOK
        document.head.appendChild(style); // @HTMLUpdateOK
        // Set a flag that we've already processed and appended the style to the 
        // document head so we only do this once for all composite instances
        cache.css = true;
      }
    }

    // Loop through all element attributes to get initial properties
    oj.BaseCustomElementBridge.__InitProperties(element, this._PROPS);
  },

  InitializePrototype: function(proto)
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializePrototype.call(this, proto);
    
    Object.defineProperty(proto, '_propsProto', {value: {}});
  },

  InitializeBridge: function(element)
  {
    // Invoke callback on the superclass
    oj.BaseCustomElementBridge.proto.InitializeBridge.call(this, element);
    
    if (element['_propsProto'])
    {
      this._PROPS = Object.create(element['_propsProto']);
      this._PROPS._BRIDGE = this;
      this._PROPS._ELEMENT = element;
    }
  },

  _getNodeBySubId: function(locator) 
  {
    // The locator subId can fall into one of 3 categories below:
    // 1) The target node belongs to a JET component or composite with a subId map
    // 2) The target node maps directly to a subId
    // 3) The composite does not have a match for the subId 
    
    // The returned subId map the following key/value pairs:
    // {
    //    [subId]: {
    //      alias: [alias or null for non JET components], 
    //      node: [node]
    //    }
    // }
    var map = oj.CompositeElementBridge.__GetSubIdMap(this);
    var match = map[locator['subId']];
    if (match)
    {
      if (match['alias']) // Case #1
      {
        var clone = oj.CollectionUtils.copyInto({}, locator, undefined, true)
        clone['subId'] = match['alias'];
        var component = match['node'];
        // Check to see if we should call the method on the element or widget
        if (component.getNodeBySubId) 
        {
          return component.getNodeBySubId(clone);
        } 
        else 
        {
          return oj.Components.__GetWidgetConstructor(component)('getNodeBySubId', clone);
        }
      }
      else
      {
        return match['node']; // Case #2
      }
    }

    return null; // Case #3
  },

  _getSubIdByNode: function(node)
  {
    // The node can fall into one of 3 categories below:
    // 1) The node is not a child of this composite.
    // 2) The node is a child of an inner composite and we need to convert its aliased subId
    // 3) The node is a child of this composite
    // 3a) The node is mapped directly to a subId
    // 3b) The node is owned by an element that has a getSubIdByNode method and we need to convert its aliased subId
  
    // Case #1
    if (!this.contains(node))
      return null;

    // The returned node map has the following key/value pairs where nodeKey is 
    // the value of the node's data-oj-subid[-map] attribute:
    // [nodeKey]: { map: [subIdMap], node: [node] }
    var nodeMap = oj.CompositeElementBridge.__GetNodeMap(this);

    // Case #2
    var composite = oj.Composite.getContainingComposite(node, this);
    if (composite != null)
    {
      var nodeKey = composite.node.getAttribute('data-oj-subid-map');
      var match = nodeMap[nodeKey];
      if (match)
      {
        if (composite.getSubIdByNode)
        {
          var locator = composite.getSubIdByNode(node);
          if (locator)
          {
            var alias = match['map'][locator['subId']];
            locator['subId'] = alias;
            return locator;
          }
        }
      }
      // Return null if we did not expose the node even though the inner composite does
      return null;
    }
    
    // Case #3
    // Walk up DOM tree until we find the containing node with the subId mapping
    var curNode = node;
    while (curNode !== this)
    {
      // We do not support an element having both attributes. If both are specified, -map takes precedence.
      var nodeKey = curNode.getAttribute('data-oj-subid-map') || curNode.getAttribute('data-oj-subid');
      if (nodeKey)
        break;
      curNode = curNode.parentNode;
    }

    var match = nodeMap[nodeKey];
    if (match) 
    {
      var map = match['map'];
      if (!map) // Case #3a
      {
        return {'subId': nodeKey};
      }
      else // Case #3b
      {
        var component = match['node'];
        var locator;
        // Check to see if we should call the method on the element or widget
        if (component.getSubIdByNode) 
        {
          locator = component.getSubIdByNode(node);
        } 
        else 
        {
          locator = oj.Components.__GetWidgetConstructor(component)('getSubIdByNode', node);
        }

        if (locator)
        {
          locator['subId'] = match['map'][locator['subId']];
          return locator;
        }
      }
    }

    return null;
  },

  _getViewModel: function(element) 
  {
    if (!this._VIEW_MODEL)
    {
      this._throwError(element, "Cannot access methods before element is upgraded.");
    }
    return this._VIEW_MODEL;
  }
  
});

/*************************/
/* PUBLIC STATIC METHODS */
/*************************/

/**
 * See oj.Composite.register doc for details
 * @ignore
 *
 */
oj.CompositeElementBridge.register = function(tagName, descriptor)
{

  // Convert any descriptor objects using the deprecated inline keys to the new API
  var descrip = {};
  descrip[oj.BaseCustomElementBridge.DESC_KEY_META] = oj.CompositeElementBridge._getResource(descriptor, oj.BaseCustomElementBridge.DESC_KEY_META);
  descrip[oj.BaseCustomElementBridge.DESC_KEY_VIEW] = oj.CompositeElementBridge._getResource(descriptor, oj.BaseCustomElementBridge.DESC_KEY_VIEW);
  descrip[oj.BaseCustomElementBridge.DESC_KEY_CSS] = oj.CompositeElementBridge._getResource(descriptor, oj.BaseCustomElementBridge.DESC_KEY_CSS);
  descrip[oj.BaseCustomElementBridge.DESC_KEY_VIEW_MODEL] = oj.CompositeElementBridge._getResource(descriptor, oj.BaseCustomElementBridge.DESC_KEY_VIEW_MODEL);
  descrip[oj.BaseCustomElementBridge.DESC_KEY_PARSE_FUN] = descriptor[oj.BaseCustomElementBridge.DESC_KEY_PARSE_FUN];

  if (oj.BaseCustomElementBridge.__Register(tagName, descrip, oj.CompositeElementBridge.proto, true))
  {
    var metadata = descrip[oj.BaseCustomElementBridge.DESC_KEY_META];
    if (!metadata)
    {
      // Metadata is required starting in 3.0.0, but to be backwards compatible, just log a warning.
      oj.Logger.warn("Composite registered'" + tagName.toLowerCase() + "' without Metadata.");
      metadata = {};
    }
    else
    {
      // Check that the component name, version, and JET versions are defined in the metadata
      var name = metadata['name'];
      if (!name)
        oj.Logger.warn('Warning registering composite %s. Required property "name" missing from metadata.', tagName);
      else if (tagName != name)
        oj.Logger.warn('Warning registering composite %s. Registered name: %s does not match name: %s provided in metadata.', tagName, tagName, name);
      if (!metadata['version'])
        oj.Logger.warn('Warning registering composite %s. Required property "version" missing from metadata.', tagName);
      if (!metadata['jetVersion'])
        oj.Logger.warn('Warning registering composite %s. Required composite "jetVersion" missing from metadata.', tagName);
    }
    var view = descrip[oj.BaseCustomElementBridge.DESC_KEY_VIEW];
    if (view == null)
      throw new Error("Cannot register composite '" + tagName.toLowerCase() + "' without a View.");

    // __ProcessEventListeners returns a copy of the metadata so we're not updating the original here.
    descrip['_metadata'] = oj.BaseCustomElementBridge.__ProcessEventListeners(metadata, false);
    customElements.define(tagName.toLowerCase(), oj.CompositeElementBridge.proto.getClass(descrip));
  }
};

/*****************************/
/* NON PUBLIC STATIC METHODS */
/*****************************/

/**
 * @ignore
 */
oj.CompositeElementBridge._getDomNodes = function(content, element)
{
  if (typeof content === 'string')
  {
    return oj.__HtmlUtils.stringToNodeArray(content);
  }
  else if (oj.CompositeElementBridge._isDocumentFragment(content))
  {
    var clonedContent = content.cloneNode(true);
    var nodes = [];
    for (var i = 0; i < clonedContent.childNodes.length; i++)
      nodes.push(clonedContent.childNodes[i]);
    return nodes;
  }
  else if (Array.isArray(content))
  {
    var clonedContent = [];
    for (var i = 0; i < content.length; i++)
      clonedContent.push(content[i].cloneNode(true));
    return clonedContent;
  }
  else
  {
    var bridge = oj.BaseCustomElementBridge.getInstance(element);
    // TODO update this error message once we remove support for Array of DOM nodes and DocumentFragment
    bridge._throwError(element, "The composite View is not one of the following supported types: string, Array of DOM nodes, DocumentFragment");
  }
};

/**
 * Creates the subId and node maps needed for automation
 * @ignore
 */
oj.CompositeElementBridge._generateSubIdMap = function(bridge, element)
{
  if (!bridge._SUBID_MAP)
  {
    // The format of the map will be { [composite subId] : {alias: [alias], node: [node] } }
    var subIdMap = {};
    var nodeMap = {};

    // data-oj-subid or data-oj-subid-map attributes can be defined on nested objects so we need
    // to walk the composite tree skipping over slots
    var children = element.children;
    for (var i = 0; i < children.length; i++) {
      oj.CompositeElementBridge._walkSubtree(subIdMap, nodeMap, children[i]);
    }

    bridge._NODE_MAP = nodeMap;
    bridge._SUBID_MAP = subIdMap;
  }
};

/**
 * Walks a composite subtree, parsing and generating subId mappings.
 * @ignore
 */
oj.CompositeElementBridge._walkSubtree = function(subIdMap, nodeMap, node)
{
  if (!node.hasAttribute('slot'))
  {
    oj.CompositeElementBridge._addNodeToSubIdMap(subIdMap, nodeMap, node);
    if (!oj.BaseCustomElementBridge.getRegistered(node.tagName) && !oj.Components.__GetWidgetConstructor(node))
    {
      var children = node.children;
      for (var i = 0; i < children.length; i++)
      {
        oj.CompositeElementBridge._walkSubtree(subIdMap, nodeMap, children[i]);
      }
    }
  }
};

/**
 * Checks to see if a node has defined subIds and adds them to the composite's
 * cached subId -> node and node -> subId maps for automation.
 * @ignore
 */
oj.CompositeElementBridge._addNodeToSubIdMap = function(subIdMap, nodeMap, node)
{
  var nodeSubId = node.getAttribute('data-oj-subid');
  var nodeSubIdMapStr = node.getAttribute('data-oj-subid-map');
  // We do not support an element having both attributes. If both are specified, -map takes precedence.
  if (nodeSubIdMapStr)
  {
    var parsedValue = JSON.parse(nodeSubIdMapStr);
    if (typeof parsedValue === 'object' && !(parsedValue instanceof Array))
    { 
      // Due to closure compiler issues with passing in result of JSON.parse which has type * into Object.keys 
      // which requires an Object, use for loop here instead of iterating over key Array.
      var nodeSubIdMap = parsedValue;
      var reverseMap = {};
      for (var key in nodeSubIdMap)
      {
        subIdMap[key] = {'alias': nodeSubIdMap[key], 'node': node};
        reverseMap[nodeSubIdMap[key]] = key;
      }
      nodeMap[nodeSubIdMapStr] = {'map': reverseMap, 'node': node};
    }
  }
  else if (nodeSubId)
  {
    subIdMap[nodeSubId] = {'node': node};
    nodeMap[nodeSubId] = {'node': node};
  }
};

/**
 * Returns the subId to node mapping for the composite's View.
 * @ignore
 */
oj.CompositeElementBridge.__GetSubIdMap = function(element)
{
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  oj.CompositeElementBridge._generateSubIdMap(bridge, element);
  return bridge._SUBID_MAP;
};

/**
 * Returns the node to subId mapping for the composite's View. The returned map has the 
 * following key/value pairs where nodeKey is value of the node's data-oj-subid[-map] attribute:
 * {
 *   [nodeKey]: {
 *     map: [subIdMap],
 *     node: [node]
 *   }
 * }
 * @return {Map}
 * @ignore
 */
oj.CompositeElementBridge.__GetNodeMap = function(element)
{
  var bridge = oj.BaseCustomElementBridge.getInstance(element);
  oj.CompositeElementBridge._generateSubIdMap(bridge, element);
  return bridge._NODE_MAP;
};

/**
 * @ignore
 */
oj.CompositeElementBridge._getPropertyTracker = function(bridge, property)
{
  if (!bridge._TRACKERS)
    bridge._TRACKERS = {};
  if (!bridge._TRACKERS[property])
    bridge._TRACKERS[property] = oj.CompositeTemplateRenderer.createTracker();
  return bridge._TRACKERS[property];
};

/**
 * @ignore
 */
oj.CompositeElementBridge._getResource = function(descriptor, key)
{
  var resource = descriptor[key];
  if (resource != null)
  {
    if (resource.hasOwnProperty('inline'))
      return resource['inline'];
    else if (resource.hasOwnProperty('promise'))
      throw new Error("The 'promise' resource type for descriptor key '" + key + "' is no longer supported." +
        " The resource should be passed directly as the value instead.");
    else
      return resource;
  }
};

/**
 * @ignore
 */
oj.CompositeElementBridge._isDocumentFragment = function(content)
{
  if (window['DocumentFragment'])
  {
    return content instanceof DocumentFragment;
  }
  else
  {
    return content && content.nodeType === 11;
  }
};


/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */
 /**
 * @ojcomponent oj.ojAvatar
 * @ojsignature {target: "Type", value: "class ojAvatar extends JetElement<ojAvatarSettableProperties>"}
 * @since 4.0.0
 * @ojstatus preview
 * @ojshortdesc An icon capable of displaying a custom image, initials, or a placeholder image.
 *
 * @classdesc
 * <h3 id="avatarOverview-section">
 *   JET Avatar
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#avatarOverview-section"></a>
 * </h3>
 * <p>Description: Themeable, WAI-ARIA-compliant element that often represents a person.</p>
 * <p>A JET avatar is an icon capable of displaying
 * a custom image, or initials, or a placeholder image.  The order of precedence for
 * what is displayed, in order from highest to lowest, is:</p>
 * <ol>
 *  <li>Custom image specified through the "src" attribute</li>
 *  <li>Initials specified through the "initials" attribute</li>
 *  <li>Default placeholder image</li>
 * </ol>
 * <pre class="prettyprint">
 * <code>//Avatar with initials
 *&lt;oj-avatar initials="AB">
 * &lt;/oj-avatar>
 *</code></pre>
 *  <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 *  </h3>
 *
 * <p>The application <b>must</b> set role and aria-label for accessibility purposes.  JET Avatar does not have
 * any interaction with the application, therefore it is not keyboard navigable by default.
 * The aria-label will be picked up by the tabbable/focusable parent unless it is
 * overriden by the application.
 * The application can set a tooltip by setting the title attribute of the avatar.
 *
 *
 * <p>In order to meet accessibility requirements for text, color contrast ratio
 * between the background color and text must be
 * greater than 4.5 for the two smallest avatars and 3.1 for the five larger avatars.
 * Avatar's default background satisfies the 3.1 color contrast ratio.  The background
 * will automatically be darkened for the two smallest sizes to satisfy the more
 * stringent 4.5 contrast ratio.  If colors are customized through theming, the
 * application is responsible for specifying colors that satisfy the 3.1 contrast ratio.
 * The custom background color will be automatically darkened as well for the two smallest avatars.
 *
 * <p>The src attribute will display the image as a background-image.  These images do
 * not appear in high contrast mode.  For this reason, initials must be specified and
 * and will be shown instead.
 *
 *
 * <h3 id="image-section">
 *   Image
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#image-section"></a>
 *  </h3>
 *
 * <p>The avatar will display the image provided from the src attribute if the src
 * attribute is populated.  If the src attribute is not provided and the initials have been,
 * the initials will be displayed.  If neither src nor initials attributes are populated,
 * a single person placeholder image is shown.
 * Use the class oj-avatar-group-image to use the group placeholder image.
 * Examples displaying each type of avatar:
 * <pre class="prettyprint">
 * <code>//Individual Placeholder
 * &lt;oj-avatar>
 *  &lt;/oj-avatar>
 * //Group Placeholder
 * &lt;oj-avatar class="oj-avatar-group-image">
 *  &lt;/oj-avatar>
 * //Initials
 *&lt;oj-avatar initials="AB">
 * &lt;/oj-avatar>
 * //Image
 *&lt;oj-avatar initials="AB" src="image.jpg">
 * &lt;/oj-avatar>
 *</code></pre>
 *
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 * <p> The following CSS classes can be applied by the page author as needed.</p>
 * {@ojinclude "name":"stylingDoc"}
 */
 /**
 * Specifies the size of the avatar.
 * @expose
 * @name size
 * @memberof oj.ojAvatar
 * @ojshortdesc Specifies the size of the avatar.
 * @instance
 * @type {string}
 * @ojvalue {string} "xxs" extra, extra small avatar
 * @ojvalue {string} "xs" extra small avatar
 * @ojvalue {string} "sm" small avatar
 * @ojvalue {string} "md" medium avatar, default value
 * @ojvalue {string} "lg" large avatar
 * @ojvalue {string} "xl" extra large avatar
 * @ojvalue {string} "xxl" extra, extra large avatar
 * @default "md"
 * @example <caption>Renders avatar displaying default placeholder image with <code class="prettyprint">size</code>
 * attribute set to large</caption>
 * &lt;oj-avatar size='lg'>&lt;/oj-avatar>
 * @example <caption>Get or set the <code class="prettyprint">size</code> property after initialization</caption>
 * //Get avatar size
 * var size = myAvatar.size;
 *
 * //Set size property to xs
 * myAvatar.size = "xs";
 */
 /**
 * Specifies the initials of the avatar.  Will only be displayed if the src attribute is null.
 * Required if src attribute is provided for accessibility purposes.  Will be displayed
 * if the src attribute is not specified, or in high contrast mode for accessibility
 * purposes.
 * @expose
 * @name initials
 * @ojtranslatable
 * @ojshortdesc Specifies the initials of the avatar.
 * @memberof oj.ojAvatar
 * @instance
 * @type {string}
 * @default null
 * @example <caption>Renders a default medium avatar with initials</caption>
 * &lt;oj-avatar initials='AB'>&lt;/oj-avatar>
 * @example <caption>Get or set the <code class="prettyprint">initials</code> property after initialization</caption>
 * //Get avatar initials
 * var initials = myAvatar.initials;
 *
 * //Set initials property to 'NT'
 * myAvatar.initials = "NT";
 */
 /**
 * Specifies the src for the image of the avatar.  Image will be rendered as a background image.
 * In high contrast mode, initials will be displayed instead since background images
 * will not be rendered.
 * @ojshortdesc Specifies the src for the image of the avatar.
 * @expose
 * @name src
 * @memberof oj.ojAvatar
 * @instance
 * @type {string}
 * @default null
 * @example <caption>Renders a default medium avatar with a image</caption>
 * &lt;oj-avatar src='image.jpg'>&lt;/oj-avatar>
 * @example <caption>Get or set the <code class="prettyprint">src</code> property after initialization</caption>
 * //Get avatar src
 * var src = myAvatar.src;
 *
 * //Set src property to 'image2.jpg'
 * myAvatar.src = "image2.jpg";
 */
 /**
  *
  * <table class="generic-table styling-table">
  *   <thead>
  *     <tr>
  *       <th>Class</th>
  *       <th>Description</th>
  *     </tr>
  *   </thead>
  *   <tbody>
  *     <tr>
  *       <td>oj-avatar-group-image</td>
  *       <td>Use to diplay avatar group placeholder image instead of single person
  *           placeholder image.</td>
  *     </tr>
  *   </tbody>
  * </table>
  *
  * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
  * @memberof oj.ojAvatar
  */
var view =
  "<div class=\"oj-avatar-outer\" data-bind=\"css: !$properties.initials || $properties.src ? 'oj-avatar-'" +
  "  + $properties.size : 'oj-avatar-has-initials oj-avatar-'+ $properties.size\"aria-hidden=\"true\">" +
  "  <div class=\"oj-avatar-inner\">" +
  "    <!-- ko if: $properties.src -->" +
  "    <div class=\"oj-avatar-background-image\"" +
  "         data-bind=\"style:{'background-image':'url(' + $properties.src + ')'}\">" +
  "      <div class=\"oj-avatar-initials\" data-bind=\"text: $properties.initials\"></div>" +
  "    </div>" +
  "    <!-- /ko -->" +
  "    <!-- ko if: $properties.initials && !$properties.src -->" +
  "    <div class=\"oj-avatar-initials\" data-bind=\"text: $properties.initials\"></div>" +
  "    <!-- /ko -->" +
  "    <!-- ko if: !$properties.initials && !$properties.src -->" +
  "    <div class=\"oj-avatar-placeholder\"></div>" +
  "    <!-- /ko -->" +
  "  </div>" +
  "</div>"



var metadata =
{
  "properties": {
    "initials": {
      "description": "Initials for avatar, used if no image",
      "type": "string"
    },
    "src": {
      "description": "Source for background image",
      "type": "string"
    },
    "size": {
      "description": "Size of the avatar",
      "type": "string",
      "enumValues": ["xxs", "xs", "sm", "md", "lg", "xl", "xxl"],
      "value" : "md"
    }
  }
}

oj.Composite.register('oj-avatar',
{
  "view": view,
  "metadata": metadata
});

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 *
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 *
 * @expose
 * @memberof oj.ojAvatar
 * @instance
 * @return {void}
 *
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 *
 * @expose
 * @memberof oj.ojAvatar
 * @instance
 *
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 *
 * @expose
 * @memberof oj.ojAvatar
 * @instance
 *
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */

/**
 * An immutable set of keys.
 * @class KeySet
 * @ojstatus preview
 * @classdesc The base class for ExpandedKeySet and ExpandAllKeySet.  Represents an immutable set of keys.
 * @constructor
 * @hideconstructor
 * @abstract
 * @since 4.1.0
 * @ojsignature {target: "Type", value: "abstract class KeySet<K>"}
 */
var KeySet = function() {};

// Subclass from oj.Object 
oj.Object.createSubclass(KeySet, oj.Object, "KeySet");

// make it available internally
oj.KeySet = KeySet;

/**
 * Initialize the internal Set with a set of keys.
 *
 * @param {Set|Array|null|undefined} keys the initial keys to create the internal Set with.
 * @protected
 */
KeySet.prototype.InitializeWithKeys = function(keys)
{
    this._keys = this._populate(keys);
};

/**
 * Sets the internal Set object.
 *
 * @param {Set} set the internal Set object to replace with.
 * @protected
 */
KeySet.prototype.SetInternal = function(set)
{
    this._keys = set;
};

/**
 * Returns a new KeySet based on this set with the specified keys included.
 *
 * @param {Set|Array} keys a set of keys to add to this KeySet.
 * @return {KeySet} a new KeySet with the specified keys included.
 * @method
 * @name add
 * @memberof KeySet
 * @instance
 * @abstract
 * @ojsignature {target: "Type", value: "<KS extends KeySet<K>>(keys: Set<K>|Array<K>): KS"}
 */

/**
 * Returns a new KeySet that signals all keys are added to this set. 
 *
 * @return {KeySet} a new KeySet that signals all keys are added to this set.
 * @method
 * @name addAll
 * @memberof KeySet
 * @instance
 * @abstract
 * @ojsignature {target: "Type", value: "<KS extends KeySet<K>>(): KS"}
 */

/**
 * Returns a new KeySet based on this set with the specified keys excluded. 
 *
 * @param {Set|Array} keys a set of keys to remove from this KeySet.
 * @return {KeySet} a new KeySet with the specified keys excluded.
 * @method
 * @name delete
 * @memberof KeySet
 * @instance
 * @abstract
 * @ojsignature {target: "Type", value: "<KS extends KeySet<K>>(keys: Set<K>|Array<K>): KS"}
 */

/**
 * Returns whether this set should include all keys.
 *
 * @return {boolean} true if this set should include all keys, false otherwise.
 * @method
 * @name isAddAll
 * @memberof KeySet
 * @instance
 * @abstract
 */

/**
 * Returns whether the specified key is contained in this set.
 *
 * @param {*} key the key to check whether it is contained in this set.
 * @return {boolean} true if the specified key is contained in this set, false otherwise.
 * @method
 * @name has
 * @memberof KeySet
 * @instance
 * @abstract
 * @ojsignature {target: "Type", value: "K", for: "key"}
 */

/**
 * Returns a new KeySet containing no keys.
 *
 * @return {KeySet} a new KeySet with no keys.
 * @method
 * @name clear
 * @memberof KeySet
 * @instance
 * @abstract
 * @ojsignature {target: "Type", value: "<KS extends KeySet<K>>(): KS"}
 */

/**
 * Adds or deletes a set of keys from the internal Set object.
 * @param {boolean} isAdd true if add operation, false if delete operation
 * @param {Set|Array} keys keys to add or delete
 * @return {KeySet} returns current KeySet if add or delete is not performed, or a new KeySet with the
 *                     specified keys included (add) or excluded (delete).
 * @protected
 */
KeySet.prototype.AddOrDeleteInternal = function(isAdd, keys)
{
    var newSet, keySet;

    newSet = isAdd ? this._add(keys) : this._remove(keys);
    if (newSet == null)
    {
        return this;
    }
    else
    {
        keySet = /** @type {KeySet} */ (Object.create(Object.getPrototypeOf(this)));
        keySet.SetInternal(newSet);
        return keySet;
    }
};

/**
 * Adds the specified keys to the internal Set object.
 * @param {Set|Array} keys a set of keys to add
 * @return {Set} a new Set based on this internal Set with the specified keys appended to the end, or null if nothing was added.
 * @private
 */
KeySet.prototype._add = function(keys)
{
    var self = this, newSet = null, key;

    keys.forEach(function(key)
    {
        // checks if it's already contained in the Set, can't use has() since it does a reference comparison
        if (self.FindEquals(key) == null)
        {
            if (newSet == null)
            {
                newSet = self.Clone();
            }
            newSet.add(key);
        }
    });

    return newSet;
};

/**
 * Helper method to remove the specified keys from its set
 * @param {Set|Array} keys an interable set of keys to remove
 * @return {Set|null} a new Set based on this internal Set with the keys removed, or null if nothing is removed.
 * @private
 */
KeySet.prototype._remove = function(keys)
{
    var self = this, newSet = null, key, keyToDelete;

    // first check if there's anything to remove
    if (this._keys.size == 0)
    {
        return null;
    }

    keys.forEach(function(key)
    {
        // see if we can find a equivalent key in this Set since delete does a reference comparison to find the item to delete
        keyToDelete = self.FindEquals(key);
        if (keyToDelete != null)
        {
            if (newSet == null)
            {
                newSet = self.Clone();
            }
            newSet.delete(keyToDelete);
        }
    });

    return newSet;
};

/**
 * Finds the equavalent key of the specified key within this KeySet.
 * @param {*} keyToFind the key to find
 * @return {*} the key in the key that is equivalent to keyToFind, or null if nothing equivalent can be found.
 * @protected
 */
KeySet.prototype.FindEquals = function(keyToFind)
{
    var iterator, key, found = null;

    if (this._keys.has(keyToFind))
    {
        return keyToFind;
    }

    // using iterator if it's supported since we could break at any time
    if (typeof Symbol === "function" && typeof Set.prototype[Symbol.iterator] === "function")
    {
        iterator = this._keys[Symbol.iterator]();
        key = iterator.next();
        while (!key.done)
        {
            if (oj.Object.compareValues(key.value, keyToFind) || key.value == keyToFind)
            {
                return key.value;
            }
            key = iterator.next();
        }
    }
    else
    {
        // IE11 supports forEach
        this._keys.forEach(function(key)
        {
            if (found == null && oj.Object.compareValues(key, keyToFind) || key == keyToFind)
            {
                found = key;
            }
        });
    }

    return found;
};

/**
 * Returns the size of this Set.
 * @return {number} the size of this Set.
 * @protected
 */
KeySet.prototype.GetInternalSize = function()
{
    return this._keys.size;
};

/**
 * Return a clone of the internal Set
 * @return {Set} the clone of the internal Set
 * @protected
 */
KeySet.prototype.Clone = function()
{
    return this._populate(this._keys);
};

/**
 * IE11 does not support constructor with arguments
 * TODO: move to CollectionUtils
 * @private
 */
KeySet.prototype._populate = function(keys)
{
    var set = new Set(keys);
    if (keys != null && set.size == 0)
    {
        keys.forEach(function(key)
        {
            set.add(key);
        });             
    }
    return set;
};
/**
 * Create a new immutable KeySet containing the keys of the expanded items.
 * Use this KeySet when specifying individual keys to expand.
 *
 * @param {(Set|Array)=} keys A set of keys to initialize this KeySet with.
 * @ojstatus preview
 * @class ExpandedKeySet
 * @classdesc The ExpandedKeySet class contains a set of keys of the expanded items.
 * @extends {KeySet}
 * @constructor
 * @since 4.1.0
 * @ojsignature [{target: "Type", value: "class ExpandedKeySet<K> extends KeySet<K>"},
 *               {target: "Type", value: "Set<K>|Array<K>", for:"keys"}]
 * @example <caption>Creates a new ExpandedKeySet with an initial set of keys to expand:</caption>
 * require(['ojs/ojkeyset'], 
 *   function(keySet) {
 *     var expandedKeySet = new keySet.ExpandedKeySet(['group1', 'group3']);
 *   }
 * ); 
 */
var ExpandedKeySet = function(keys)
{
    this.InitializeWithKeys(keys);
};

// Subclass from KeySet
oj.Object.createSubclass(ExpandedKeySet, KeySet, "ExpandedKeySet");

// make it available internally
oj.ExpandedKeySet = ExpandedKeySet;

/**
 * Returns a new KeySet based on this set with the specified keys included.
 * If none of the keys specified are added then this KeySet is returned.
 *
 * @param {Set|Array} keys a set of keys to add to this KeySet.
 * @return {ExpandedKeySet} a new KeySet with the specified keys included.
 * @expose
 * @instance
 * @alias add
 * @memberof ExpandedKeySet 
 * @ojsignature {target: "Type", value: "<ExpandedKeySet>(keys: Set<K>|Array<K>): ExpandedKeySet"}
 */
ExpandedKeySet.prototype.add = function(keys)
{
    return /** @type {!ExpandedKeySet} */ (this.AddOrDeleteInternal(true, keys));
};

/**
 * Returns a new KeySet that signals all keys are added to this set. 
 *
 * @return {ExpandAllKeySet} a new KeySet that signals all keys are added to this set.
 * @expose
 * @instance
 * @alias addAll
 * @memberof ExpandedKeySet 
 * @ojsignature {target: "Type", value: "<ExpandAllKeySet>(): ExpandAllKeySet"}
 */
ExpandedKeySet.prototype.addAll = function()
{
    return new ExpandAllKeySet();
};

/**
 * Returns whether this set should include all keys.
 *
 * @return {boolean} true if this set includes all keys, false otherwise.
 * @expose
 * @instance
 * @alias isAddAll
 * @memberof ExpandedKeySet 
 */
ExpandedKeySet.prototype.isAddAll = function()
{
    return false;
};

/**
 * Returns a new KeySet based on this set with the specified keys excluded.  
 * If none of the keys specified are deleted then this KeySet is returned.
 *
 * @param {Set|Array} keys a set of keys to remove from this KeySet.
 * @return {ExpandedKeySet} a new KeySet with the specified keys excluded.
 * @expose
 * @instance
 * @alias delete
 * @memberof ExpandedKeySet 
 * @ojsignature {target: "Type", value: "<ExpandedKeySet>(keys: Set<K>|Array<K>): ExpandedKeySet"}
 */
ExpandedKeySet.prototype.delete = function(keys)
{
    return /** @type {!ExpandedKeySet} */ (this.AddOrDeleteInternal(false, keys));
};

/**
 * Returns a new KeySet containing no keys.  If this KeySet already contains no keys then 
 * the current KeySet is returned.
 *
 * @return {ExpandedKeySet} a new KeySet with no keys.
 * @expose
 * @instance
 * @alias clear
 * @memberof ExpandedKeySet 
 * @ojsignature {target: "Type", value: "<ExpandedKeySet>(): ExpandedKeySet"}
 */
ExpandedKeySet.prototype.clear = function()
{
    return this.GetInternalSize() == 0 ? this : new ExpandedKeySet();
};

/**
 * Returns whether the specified key is contained in this set.
 *
 * @param {*} key the key to check whether it is contained in this set.
 * @return {boolean} true if the specified key is contained in this set, false otherwise.
 * @expose
 * @instance
 * @alias has
 * @memberof ExpandedKeySet 
 * @ojsignature {target: "Type", value: "K", for:"key"}
 */
ExpandedKeySet.prototype.has = function(key)
{
    return (this.FindEquals(key) != null);
};

/**
 * Returns the keys in this KeySet in the order they are added.
 *
 * @return {Set} the keys in this KeySet in the order they are added.
 * @expose
 * @instance
 * @alias values
 * @memberof ExpandedKeySet 
 * @ojsignature {target: "Type", value: "Set<K>", for:"returns"}
 */
ExpandedKeySet.prototype.values = function()
{
    return this.Clone();
};
/**
 * Create a new immutable KeySet containing the keys of the collapsed items.
 * Use this KeySet when expanding all keys.
 * 
 * @ojstatus preview
 * @class ExpandAllKeySet
 * @classdesc The ExpandAllKeySet class represents a set with all keys expanded.
 * @extends {KeySet}
 * @constructor
 * @since 4.1.0
 * @ojsignature {target: "Type", value: "class ExpandAllKeySet<K> extends KeySet<K>"}
 * @example <caption>Creates a new ExpandAllKeySet to expand all keys</caption>
 * require(['ojs/ojkeyset'], 
 *   function(keySet) {
 *     var expandAllKeySet = new keySet.ExpandAllKeySet();
 *   }
 * ); 
 */
var ExpandAllKeySet = function()
{
    this.InitializeWithKeys(null);
};

// Subclass from KeySet
oj.Object.createSubclass(ExpandAllKeySet, KeySet, "ExpandAllKeySet");

// make it available internally
oj.ExpandAllKeySet = ExpandAllKeySet;

/**
 * Returns a new KeySet with the specified keys excluded from a set of collapsed keys.
 * If the keys specified are already added then this KeySet is returned.
 *
 * @param {Set|Array} keys a set of keys to add to this KeySet.
 * @return {ExpandAllKeySet} a new KeySet with the specified keys included.
 * @expose
 * @instance
 * @alias add
 * @memberof! ExpandAllKeySet 
 * @ojsignature {target: "Type", value: "<ExpandAllKeySet>(keys: Set<K>|Array<K>): ExpandAllKeySet"}
 */
ExpandAllKeySet.prototype.add = function(keys)
{
    // add keys on expand all = remove collapsed keys
    return /** @type {!ExpandAllKeySet} */ (this.AddOrDeleteInternal(false, keys));
};

/**
 * Returns a new KeySet that signals all keys are added to this set.  If this KeySet already
 * has all keys added, then this KeySet is returned.
 *
 * @return {ExpandAllKeySet} a new KeySet that signals all keys are added to this set.
 * @expose
 * @instance
 * @alias addAll
 * @memberof! ExpandAllKeySet 
 * @ojsignature {target: "Type", value: "<ExpandAllKeySet>(): ExpandAllKeySet"}
 */
ExpandAllKeySet.prototype.addAll = function()
{
    return this.GetInternalSize() == 0 ? this : new ExpandAllKeySet();
};

/**
 * Returns whether this set should include all keys.
 *
 * @return {boolean} true if this set includes all keys, false otherwise.
 * @expose
 * @instance
 * @alias isAddAll
 * @memberof! ExpandAllKeySet 
 */
ExpandAllKeySet.prototype.isAddAll = function()
{
    return true;
};

/**
 * Returns a new KeySet based on this set with the specified keys included in a set of collapsed keys.
 * If the keys specified are already deleted then this KeySet is returned.
 *
 * @param {Set|Array} keys a set of keys to remove from this KeySet.
 * @return {ExpandAllKeySet} a new KeySet with the specified keys excluded.
 * @expose
 * @instance
 * @alias delete
 * @memberof! ExpandAllKeySet 
 * @ojsignature {target: "Type", value: "<ExpandAllKeySet>(keys: Set<K>|Array<K>): ExpandAllKeySet"}
 */
ExpandAllKeySet.prototype.delete = function(keys)
{
    // remove keys on expand all = add collapsed keys
    return /** @type {!ExpandAllKeySet} */ (this.AddOrDeleteInternal(true, keys));
};

/**
 * Returns a new KeySet containing no keys.
 *
 * @return {ExpandedKeySet} a new KeySet with no keys.
 * @expose
 * @instance
 * @alias clear
 * @memberof! ExpandAllKeySet 
 * @ojsignature {target: "Type", value: "<ExpandedKeySet>(): ExpandedKeySet"}
 */
ExpandAllKeySet.prototype.clear = function()
{
    return new ExpandedKeySet();
};

/**
 * Returns whether the specified key is contained in this set.
 *
 * @param {*} key the key to check whether it is contained in this set.
 * @return {boolean} true if the specified key is contained in this set, false otherwise.
 * @expose
 * @instance
 * @alias has
 * @memberof! ExpandAllKeySet 
 * @ojsignature {target: "Type", value: "<ExpandAllKeySet>(): ExpandAllKeySet"}
 */
ExpandAllKeySet.prototype.has = function(key)
{
    return (this.FindEquals(key) == null);
};

/**
 * Returns a set of keys of the collapsed items.
 * 
 * @return {Set} the keys of the collapsed items.
 * @expose
 * @instance
 * @alias deletedValues
 * @memberof! ExpandAllKeySet 
 * @ojsignature {target: "Type", value: "Set<K>", for: "returns"}
 */
ExpandAllKeySet.prototype.deletedValues = function()
{
    return this.Clone();
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * @ojcomponent oj.ojAccordion
 * @augments oj.baseComponent
 * @ojtsimport ojcollapsible
 * @since 0.6
 * @ojstatus preview
 * @ojshortdesc Displays a set of Collapsible child elements.
 * @ojrole group
 * @class oj.ojAccordion
 *
 * @classdesc
 * <h3 id="accordionOverview-section">
 *   JET Accordion
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accordionOverview-section"></a>
 * </h3>
 *
 * <p>Description: A JET Accordion contains one or more {@link oj.ojCollapsible} child elements.
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-accordion>
 *   &lt;oj-collapsible>
 *     &lt;h3>
 *       &lt;img src="images/default.png"/>
 *       &lt;span>Header 1&lt;/span>
 *     &lt;/h3>
 *     &lt;p>Content 1.&lt;/p>
 *   &lt;/oj-collapsible>
 *   &lt;oj-collapsible expanded="true">
 *     &lt;h3>Header 3&lt;/h3>
 *     &lt;p>Content 3&lt;/p>
 *   &lt;/oj-collapsible>
 * &lt;/oj-accordion>
 * </code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>In the unusual case that the directionality (LTR or RTL) changes post-init, the accordion must be <code class="prettyprint">refresh()</code>ed.
 *
 */
(function () {
  oj.__registerWidget('oj.ojAccordion', $.oj.baseComponent,
    {
      widgetEventPrefix: 'oj',
      options:
      {
      /**
       * Allow multiple collapsibles to be open at the same time.
       * Note: if multiple is true, the beforeCollapse/beforeExpand/collapse/expand events will not be fired by the accordion. They are however fired by the collapsibles.
       *
       * @expose
       * @memberof oj.ojAccordion
       * @instance
       * @ojshortdesc Specifies whether multiple collapsibles can be open at the same time.
       * @type {boolean}
       * @default false
       * @example <caption>Initialize the accordion with the <code class="prettyprint">multiple</code> attribute specified:</caption>
       * &lt;oj-accordion multiple='true'>&lt;/oj-accordion>
       *
       * @example <caption>Get or set the <code class="prettyprint">multiple</code> property after initialization:</caption>
       * // getter
       * var multipleValue = myAccordion.multiple;
       *
       * // setter
       * myAccordion.multiple=true;
       */
        multiple: false,

      /**
       * Array contains either string ids or numeric zero-based indices or objects containing string id and/or numeric index of the collapsibles that should be expanded.<p>
       * Setter value: array of either string ids or numeric indices or objects containing either string id or numeric index or both.  If the object contains both id and index, numeric index takes precedence.<p>
       * Getter value: array of objects containing numeric index and string id, if available. If an expanded collapsible has a page author provided id, that id is returned.<p>
       *
       * Note: The default value of null means that accordion doesn't modify the expanded state of its child collapsibles.
       * When the value is specified, it overrides the expanded state of its child collapsibles.<p>
       *
       * @expose
       * @memberof oj.ojAccordion
       * @instance
       * @ojshortdesc A list of expanded child collapsibles.
       * @type {Array.<string>|Array.<number>|Array.<Object>|null}
       * @ojsignature {target: "Accessor", value: {GetterType: "Array<{id?: string, index?: number}>|null", SetterType: "Array<string>|Array<number>|Array<{id?: string, index?: number}>|null"}, jsdocOverride: true}
       * @ojwriteback
       * @default null
       *
       * @example <caption>Initialize the accordion with the
       * <code class="prettyprint">expanded</code> attribute specified:</caption>
       * &lt;oj-accordion expanded="['collapsible2']">&lt;/oj-accordion>
       *
       * @example <caption>Get or set the <code class="prettyprint">expanded</code> property
       * after initialization:</caption>
       * // getter
       * var expanded = myAccordion.expanded;
       *
       * // setter
       * myAccordion.expanded=['collapsible1'];
       */
        expanded: null,

      // callbacks
      /**
       * Triggered immediately before any collapsible in the accordion is expanded.
       * Call <code class="prettyprint">event.preventDefault()</code> in the event listener to veto the event, which prevents the content from expanding.
       * If multiple is true, the beforeExpand event will not be fired by the accordion.
       *
       * @expose
       * @event
       * @memberof oj.ojAccordion
       * @instance
       * @ojshortdesc Triggered immediately before any collapsible in the accordion is expanded.
       * @ojcancelable
       * @property {Element} toCollapsible The collapsible being expanded.
       * @property {Element} fromCollapsible The collapsible being collapsed.
       */
        beforeExpand: null,

      /**
       * Triggered after the accordion has been expanded (after animation completes).
       * The expand can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>.
       * If multiple is true, the expand event will not be fired by the accordion.
       *
       * @expose
       * @event
       * @memberof oj.ojAccordion
       * @instance
       * @ojshortdesc Triggered after any collapsible in the accordion is expanded.
       * @property {Element} toCollapsible The collapsible being expanded.
       * @property {Element} fromCollapsible The collapsible being collapsed.
       */
        expand: null,

      /**
       * Triggered immediately before any collapsible in the accordion is collapsed.
       * Call <code class="prettyprint">event.preventDefault()</code> in the event listener to veto the event, which prevents the content from collapsing.
       * If multiple is true, the beforeCollapse event will not be fired by the accordion.
       *
       * @expose
       * @event
       * @memberof oj.ojAccordion
       * @instance
       * @ojshortdesc Triggered immediately before any collapsible in the accordion is collapsed.
       * @ojcancelable
       * @property {Element} toCollapsible The collapsible being expanded.
       * @property {Element} fromCollapsible The collapsible being collapsed.
       */
        beforeCollapse: null,

      /**
       * Triggered after any collapsible in the accordion has been collapsed (after animation completes).
       * The collapse can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>.
       * If multiple is true, the collapse event will not be fired by the accordion.
       *
       * @expose
       * @event
       * @memberof oj.ojAccordion
       * @instance
       * @ojshortdesc Triggered after any collapsible in the accordion has been collapsed.
       * @property {Element} toCollapsible The collapsible being expanded.
       * @property {Element} fromCollapsible The collapsible being collapsed.
       */
        collapse: null

      },

    /**
     * @memberof oj.ojAccordion
     * @instance
     * @protected
     * @override
     */
      _ComponentCreate: function () {
        this._super();
      //  - Stop using ui-helper-reset in the layout widgets.
        this.element
        .addClass('oj-accordion oj-component')
        // aria
        .attr('role', 'group');

      //  - accordion expanded opt. val doesn't overrides its collaspsible expanded opt.
        this.options.expanded = this._expandedIndexToId(this.options.expanded);
        this._refresh();
      },

    /**
     * @memberof oj.ojAccordion
     * @param {Object} menu The JET Menu to open as a context menu
     * @param {Event} event What triggered the menu launch
     * @param {string} eventType "mouse", "touch", "keyboard"
     * @private
     */
      _NotifyContextMenuGesture: function (menu, event, eventType) {
      // Setting the launcher to the "twisty" icon of the first collapsible in the accordion, since those twisties seem to be
      // the only tabbable things in the accordion, and they seem to remain tabbable even if the collapsible is disabled.
      // Component owner should feel free to specify a different launcher if appropriate, e.g. could specify the "current"
      // twisty rather than the first if desired.  See the superclass JSDoc for _OpenContextMenu for tips on choosing a launcher.
        this._OpenContextMenu(event, eventType, {
          launcher: this.element.find('.oj-collapsible-header-icon').first()
        });
      },

    /**
     * @memberof oj.ojAccordion
     * @override
     * @private
     */
      _destroy: function () {
      // clean up main element
        this.element
        .removeClass('oj-accordion oj-component')
        .removeAttr('role');

        this.element.children()
        .removeClass('oj-accordion-collapsible');

      // remove collapsibles created by accordion
        this.element.children('.oj-accordion-created')
        .removeClass('oj-accordion-created')
        .ojCollapsible('destroy');
      },

    /**
     * @memberof oj.ojAccordion
     * @override
     * @private
     */
      _setOption: function (key, value, flags) {
        if (key === 'multiple') {
        // Transition multiple to single.
        // Keep the first expanded one expanded and collapse the rest.
          if (!value && this.options.multiple) {
          //  - when "multiple" option value is changed to false, exception is displayed
            this.element.children('.oj-expanded')
            .first().siblings('.oj-collapsible')
            .ojCollapsible('collapse', false);
          }
        } else if (key === 'expanded') {
          this._setExpandedOption(value);
          return;
        }

        this._super(key, value, flags);
      },

    /**
     * Refreshes the visual state of the accordion.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof oj.ojAccordion
     * @instance
     * @return {void}
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myAccordion.refresh();
     */
      refresh: function () {
        this._super();
        this._refresh();
      },

    /**
     * @memberof oj.ojAccordion
     * @private
     */
      _refresh: function () {
        this._makeCollapsible();

      // TODO: ignore disabled until disabled propagation is supported
      // need to propagate the option to the collapsible children
      // this.option("disabled", this.options.disabled);

        this._internalSetExpanded = true;
        this._setExpandedOption(this.options.expanded);
        this._internalSetExpanded = false;

        this._setupEvents();
      },

    /**
     * @memberof oj.ojAccordion
     * @private
     */
      _makeCollapsible: function () {
        this.collapsibles = this.element.children();

      // Since oj-collapsible elements may not be upgraded until after oj-accordion elements,
      // just set the expand-area property for custom elements
        if (this._IsCustomElement()) {
          this.element.children('oj-collapsible')
          .each(function (index, collapsible) {
            collapsible.setAttribute('expand-area', 'header');
          });

          this.collapsibles.not('oj-collapsible')
          .ojCollapsible(
            {
              expandArea: 'header'
            })
        .addClass('oj-accordion-created')
        .attr('data-oj-internal', ''); // mark internal component, used in oj.Components.getComponentElementByNode
        } else {
          this.element.children(':oj-collapsible')
          .each(function () {
            $(this).ojCollapsible('option', 'expandArea', 'header');
          });

          this.collapsibles.not(':oj-ojCollapsible')
          .ojCollapsible(
            {
              expandArea: 'header'
            })
        .addClass('oj-accordion-created')
        .attr('data-oj-internal', ''); // mark internal component, used in oj.Components.getComponentElementByNode
        }

        this.collapsibles.addClass('oj-accordion-collapsible');
      },

    /**
     * @memberof oj.ojAccordion
     * @private
     */
      _setupEvents: function () {
        var events;
        if (this._IsCustomElement()) {
          events = {
            keydown: this._keydown,
            ojBeforeExpand: this._beforeExpandHandler,
            ojExpand: this._expandHandler,
            ojBeforeCollapse: this._beforeCollapseHandler,
            ojCollapse: this._collapseHandler
          };
        } else {
          events = {
            keydown: this._keydown,
            ojbeforeexpand: this._beforeExpandHandler,
            ojexpand: this._expandHandler,
            ojbeforecollapse: this._beforeCollapseHandler,
            ojcollapse: this._collapseHandler
          };
        }

        this._off(this.collapsibles);
        this._on(this.collapsibles, events);
      },

    /**
     * @memberof oj.ojAccordion
     * @private
     */
      _keydown: function (event) {
        if (event.altKey || event.ctrlKey) {
          return;
        }

      // ignore event if target is not a header
        if (!($(event.target).hasClass('oj-collapsible-header')) &&
            !($(event.target).hasClass('oj-collapsible-header-icon'))) {
          return;
        }

        var keyCode = $.ui.keyCode;
        var enabledCollapsibles = this.collapsibles.not('.oj-disabled');
        var length = enabledCollapsibles.length;
        var target = $(event.target).closest('.oj-collapsible');
        var currentIndex = enabledCollapsibles.index(target);
        var toFocus = false;

        if (currentIndex >= 0) {
          switch (event.keyCode) {
            case keyCode.RIGHT:
            case keyCode.DOWN:
              toFocus = enabledCollapsibles[(currentIndex + 1) % length];
              break;
            case keyCode.LEFT:
            case keyCode.UP:
              toFocus = enabledCollapsibles[((currentIndex - 1) + length) % length];
              break;
            case keyCode.HOME:
              toFocus = enabledCollapsibles[0];
              break;
            case keyCode.END:
              toFocus = enabledCollapsibles[length - 1];
              break;
            default:
          }
        }

        if (toFocus) {
          if (target) {
            $(target).trigger('ojfocusout');
          }
          $(toFocus).trigger('ojfocus');

          event.preventDefault();
        }
      },

    /**
     * For single expansion
     *   returns a list of expanded collapsible widgets that are sibling
     *   of the current event target
     * For multiple expansion
     *   returns an empty set.
     * @memberof oj.ojAccordion
     * @private
     */
      _findTargetSiblings: function (event) {
        if (!this.options.multiple) {
          var closestCollapsible = $(event.target).closest('.oj-collapsible');
          if (closestCollapsible.parent().is(':oj-ojAccordion')) {
            return closestCollapsible
            .siblings('.oj-collapsible.oj-expanded')
            .map(function () {
              return $(this).data('oj-ojCollapsible');
            });
          }
        }
        return $();
      },

    /**
     * Trigger "beforeCollapse" on all expanded siblings in
     * the before expand handler
     * @memberof oj.ojAccordion
     * @private
     */
      _beforeExpandHandler: function (event, eventData) {
        if (!this._isTargetMyCollapsible(event)) {
          return true;
        }

        this._expandTarget = $(event.target);
        var result;
        var collapsible = null;

        this._findTargetSiblings(event).each(function () {
          collapsible = this.element;

          var beforeCollapsedData = {
          /** @expose */
            header: collapsible.find('.oj-collapsible-header'),

          /** @expose */
            content: collapsible.find('.oj-collapsible-content')
          };

          result = this._trigger('beforeCollapse', event, beforeCollapsedData);

          if (!result) {
            this._expandTarget = null;
          }

          return result;
        });

        if (!this.options.multiple) {
          var newData = this._initEventData(collapsible, this._expandTarget);
          result = this._trigger('beforeExpand', event, newData);
        }

      // make sure collapse all expanded collapsibles before expand others
        var self = this;
        if (result) {
          this._findTargetSiblings(event).each(function () {
            this.collapse(false, event, eventData);
            self._collapsedCollapsible = this.widget();
          });
        }

        return result;
      },

      /**
       * Collapse all expanded siblings and don't allow cancel
       * @memberof oj.ojAccordion
       * @private
       */
      _expandHandler: function (event, eventData) {
      // clear the collapsedCollapsible
        var fromCollapsible = null;
        if (this._collapsedCollapsible) {
          fromCollapsible = this._collapsedCollapsible;
          this._collapsedCollapsible = null;
        }

      //  - accordion expanded opt. val doesn't overrides its collaspsible expanded opt.
      // don't handle event during setExpandedOption
        if (!this._isTargetMyCollapsible(event) || this._duringSetExpandedOption) {
          return;
        }

        var newData;
        var self = this;

        this._findTargetSiblings(event).each(function () {
          this.collapse(false, event, eventData);
          newData = self._initEventData(this.element, $(event.target));
        });

        if (!newData) {
          newData = self._initEventData(fromCollapsible, $(event.target));
        }

        if (!this.options.multiple) {
          this._trigger('expand', event, newData);
        }

        this._updateExpanded();

        this._expandTarget = null;
      },

      /**
       * Trigger "beforecollapse" on all collapsed siblings in
       * the before collapse handler
       * @memberof oj.ojAccordion
       * @private
       */
      _beforeCollapseHandler: function (event, _eventData) {
        var eventData = _eventData;
        if (this._isTargetMyCollapsible(event) &&
          !this.options.multiple) {
          if (!eventData && event.originalEvent instanceof CustomEvent) {
            eventData = event.originalEvent.detail;
          }

          return this._trigger('beforeCollapse', event,
                             this._initCollapseEventData(event, eventData));
        }
        return true;
      },

      /**
       * Collapse all collapsed siblings and don't allow cancel
       * @memberof oj.ojAccordion
       * @private
       */
      _collapseHandler: function (event, _eventData) {
        //  - accordion expanded opt. val doesn't overrides its collaspsible expanded opt.
        // don't handle event during setExpandedOption
        var eventData = _eventData;
        if (!this._duringSetExpandedOption && this._isTargetMyCollapsible(event)) {
          if (!eventData && event.originalEvent instanceof CustomEvent) {
            eventData = event.originalEvent.detail;
          }

          var newData = this._initCollapseEventData(event, eventData);

          if (!this.options.multiple) {
            this._trigger('collapse', event, newData);
          }

          this._updateExpanded();
        }
      },

      /**
       * @memberof oj.ojAccordion
       * @private
       */
      _initEventData: function (fromC, toC) {
        var eventData =
          {
        /** @expose */
            fromCollapsible: fromC, // the collapsible being collapsed.
        /** @expose */
            toCollapsible: toC      // the collapsible being expanded.
          };

        return eventData;
      },

    /**
     * @memberof oj.ojAccordion
     * @private
     */
      _initCollapseEventData: function (event, eventData) {
        var newData;
        if (eventData.toCollapsible) {
          newData = eventData;
        } else if (event.originalEvent && event.originalEvent.target) {
          newData = this._initEventData($(event.target), this._expandTarget);
        }

        if (!newData && this._expandTarget) {
          newData = this._initEventData($(event.target), this._expandTarget);
        }

        return newData;
      },

    /**
     * To filter out events from the nested accordion
     * @memberof oj.ojAccordion
     * @private
     */
      _isTargetMyCollapsible: function (event) {
        return $(event.target).is(this.collapsibles);
      },

    /**
     * @memberof oj.ojAccordion
     * @private
     */
      _updateExpanded: function () {
        var cid;
        var result = [];
        var newexp;

        this.collapsibles.each(function (index) {
          var bExpanded = false;
          if (this.tagName.toLowerCase() === 'oj-collapsible') {
            bExpanded = this.getAttribute('expanded') === 'true';
          } else {
            bExpanded = $(this).ojCollapsible('option', 'expanded');
          }
          if (bExpanded) {
            newexp = {};
            cid = $(this).attr('id');
          // add id property if provided
            if (cid) {
              newexp.id = cid;
            }
          // always add index property
            newexp.index = index;

            result.push(newexp);
          }
        });

      //  - accordion expanded opt. val doesn't overrides its collaspsible expanded opt.
        if (!oj.Object._compareArrayIdIndexObject(result, this.options.expanded)) {
          this.option('expanded', result, { _context: { internalSet: true, writeback: true } });
        }
      },

    /**
     * Return a new sorted expanded array contains IDs if they are available
     * @memberof oj.ojAccordion
     * @private
     */
      _expandedIndexToId: function (expanded) {
      //  - accordion expanded opt. val doesn't overrides its collaspsible expanded opt.
        if (Array.isArray(expanded)) {
          var id;
          var newExp = [];
          var expArr = [];

        // convert expanded from Array{Object} to Array{string or number}
          for (var i = 0; i < expanded.length; i++) {
            var exp = expanded[i];
            if (typeof exp === 'number' || typeof exp === 'string') {
              expArr.push(exp);
            } else if (typeof exp.index === 'number') {
              expArr.push(exp.index);
            } else if (typeof exp.id === 'string') {
              expArr.push(exp.id);
            }
          }

          this.element.children().each(function (index) {
            id = $(this).attr('id');
            if (id) {
              if (expArr.indexOf(id) !== -1 || expArr.indexOf(index) !== -1) {
                newExp.push({ id: id, index: index });
              }
            } else if (expArr.indexOf(index) !== -1) {
              // use index if ID not available
              newExp.push({ index: index });
            }
          });

        // if more than one expanded, keep the last collapsible
          if (!this.options.multiple && newExp.length > 1) {
            newExp = [newExp[newExp.length - 1]];
          }

          return newExp;
        }

        return null;
      },

      /**
       * @memberof oj.ojAccordion
       * @private
       */
      _setExpandedOption: function (_expanded) {
        // sort expanded array if it's from external setOption
        var expanded = _expanded;
        if (!this._internalSetExpanded) {
          expanded = this._expandedIndexToId(expanded); // eslint-disable-line no-param-reassign
        }

      // parent override children expanded setting
        if (expanded) {
        // loop thru collapsibles to collapse/expand based on expandedList
          var self = this;
          var child;
          var childId;
          var parentExp;
          var exp;
          var iexp = 0;

          this.collapsibles.each(function (index) {
            child = $(this);
            childId = child.attr('id');

            parentExp = false;
            exp = expanded[iexp];
            if (exp) {
              if (childId) {
                if (childId === exp.id) {
                  parentExp = true;
                }
              } else if (index === exp.index) {
                parentExp = true;
              }
              if (parentExp) {
                iexp += 1;
              }
            }

            //  - accordion expanded opt. val doesn't overrides its collaspsible expanded opt.
            // when the parent override the child setting:
            // 1) log a warning
            // 2) set child's expanded option with the writeback flag.
            var isCustomElement = this.tagName.toLowerCase() === 'oj-collapsible';
            var childExp = isCustomElement ?
            (this.getAttribute('expanded') === 'true' || this.getAttribute('expanded') === '') :
            child.ojCollapsible('option', 'expanded');

            if (childExp !== parentExp) {
              oj.Logger.warn('JET Accordion: override collapsible ' + index + ' expanded setting');

            // don't fire accordion "exanded" optionChange event here
              self._duringSetExpandedOption = true;
              if (isCustomElement) {
                this.setAttribute('expanded', parentExp);
              } else {
                child.ojCollapsible('option', 'expanded', parentExp);
              }
              self._duringSetExpandedOption = false;
            }
          });
        }

        this._updateExpanded();
      },

    //* * @inheritdoc */
      getNodeBySubId: function (locator) {
        if (locator == null) {
          return this.element ? this.element[0] : null;
        }

        var subId = locator.subId;
        var index = locator.index;

        if ((typeof index !== 'number') ||
            index < 0 || index >= this.collapsibles.length) {
          return null;
        }

        var collapsible = this.collapsibles[index];

        switch (subId) {
          case 'oj-accordion-content':
            subId = 'oj-collapsible-content';
            break;

          case 'oj-accordion-header':
            subId = 'oj-collapsible-header';
            break;

          case 'oj-accordion-disclosure':
          case 'oj-accordion-header-icon':
            subId = 'oj-collapsible-disclosure';
            break;

          case 'oj-accordion-collapsible':
            return collapsible;

          default:
            // Non-null locators have to be handled by the component subclasses
            return null;
        }
        return $(collapsible).ojCollapsible('getNodeBySubId', { subId: subId });
      },

    //* * @inheritdoc */
      getSubIdByNode: function (node) {
      // First, find which collapsible the node is a descendant of
        var collapsibleIndex = -1;
        var currentNode = node;
        while (currentNode) {
          collapsibleIndex = Array.prototype.indexOf.call(this.collapsibles, currentNode);
          if (collapsibleIndex !== -1) {
            break;
          }
          currentNode = currentNode.parentElement;
        }
        var subId = null;
      // Then, find the subId from the collapsible
        if (collapsibleIndex !== -1) {
          var collapsibleSubId = $(this.collapsibles[collapsibleIndex]).ojCollapsible('getSubIdByNode', node);
          collapsibleSubId = collapsibleSubId || {};
          switch (collapsibleSubId.subId) {
            case 'oj-collapsible-content':
              subId = 'oj-accordion-content';
              break;
            case 'oj-collapsible-header':
              subId = 'oj-accordion-header';
              break;
            case 'oj-collapsible-disclosure':
            case 'oj-collapsible-header-icon':
              subId = 'oj-accordion-disclosure';
              break;
            default:
              subId = 'oj-accordion-collapsible';
          }
        }
        if (subId) {
          return { subId: subId, index: collapsibleIndex };
        }
        return null;
      },

    //* * @inheritdoc */
      _CompareOptionValues: function (option, value1, value2) {
        if (option === 'expanded') {
          return oj.Object.compareValues(value1, value2);
        }
        return this._super(option, value1, value2);
      }

    // Fragments:

    /**
     * <p>The <code class="prettyprint">&lt;oj-accordion></code> element accepts one or more <code class="prettyprint">&lt;oj-collapsible></code> elements as children.
     *
     * @ojchild Default
     * @memberof oj.ojAccordion
     *
     * @example <caption>Initialize the Accordion with two Collapsible children specified:</caption>
     * &lt;oj-accordion>
     *   &lt;oj-collapsible>
     *     &lt;h3>Header 1&lt;/h3>
     *     &lt;p>Content 1&lt;/p>
     *   &lt;/oj-collapsible>
     *   &lt;oj-collapsible expanded="true">
     *     &lt;h3>Header 2&lt;/h3>
     *     &lt;p>Content 2&lt;/p>
     *   &lt;/oj-collapsible>
     * &lt;/oj-accordion>
     */

    /**
     * <p>Sub-ID for the specified disclosure icon within an Accordion.</p>
     *
     * @property {number} index The zero-based index of the disclosure icon.
     *
     * @ojsubid oj-accordion-disclosure
     * @memberof oj.ojAccordion
     *
     * @example <caption>Get the second disclosure icon:</caption>
     * var node = myAccordion.getNodeBySubId({"subId": "oj-accordion-disclosure", 'index': 1});
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Toggle disclosure state.</tr>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojAccordion
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>Space or Enter</kbd></td>
     *       <td>Toggle disclosure state.</tr>
     *     </tr>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>Tab</kbd></td>
     *       <td>Navigate to next collapsible header and if none then the next element on page.</td>
     *     </tr>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>Shift+Tab</kbd></td>
     *       <td>Navigate to previous collapsible header and if none then the previous element on page.</td>
     *     </tr>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>UpArrow or LeftArrow</kbd> (<kbd>RightArrow</kbd> in RTL)</td>
     *       <td>Move focus to the previous collapsible header with wrap around.</tr>
     *     </tr>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>DownArrow or RightArrow</kbd> (<kbd>LeftArrow</kbd> in RTL)</td>
     *       <td>Move focus to the next collapsible header with wrap around.</tr>
     *     </tr>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>Home</kbd></td>
     *       <td>Move focus to the first collapsible header.</tr>
     *     </tr>
     *     <tr>
     *       <td>Collapsible header</td>
     *       <td><kbd>End</kbd></td>
     *       <td>Move focus to the last collapsible header.</tr>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>Disabled items can receive keyboard focus, but do not allow any other interaction.
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojAccordion
     */

    });
}());

(function () {
  var ojAccordionMeta = {
    properties: {
      expanded: {
        type: 'Array',
        writeback: true
      },
      multiple: {
        type: 'boolean'
      }
    },
    events: {
      beforeCollapse: {},
      beforeExpand: {},
      collapse: {},
      expand: {}
    },
    extension: {
      _WIDGET_NAME: 'ojAccordion'
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-accordion', 'baseComponent', ojAccordionMeta);
  oj.CustomElementBridge.register('oj-accordion', { metadata: oj.CustomElementBridge.getMetadata('oj-accordion') });
}());

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */


/**
 * Utility methods for animating elements.
 *
 * <h3 id="custom-animation-section">
 *   Customizing and Disabling Animation
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#custom-animation-section"></a>
 * </h3>
 * 
 * Default animations can be customized or disabled at several levels with varying degrees of control:
 * <ol>
 *   <li>For all JET components.</li>
 *   <li>For all instances of a JET component.</li>
 *   <li>For one instance of a JET component.</li>
 * </ol>
 * 
 * <h4>1. For all JET components</h4>
 * 
 * <p>There are several theme variables that control the speed of animations in JET.  Applications can change their values 
 * to speed up or slow down animations for all components, or to disable animations altogether by setting them to 0:</p>
 * <pre class="prettyprint"><code>$animationDurationShort: .25s !default;
 * $animationDurationMedium: .4s !default;
 * $animationDurationLong: .5s !default;
 * </code></pre>
 * <p>Note that setting them to 0 does not make the actions that invoke animations synchronous.  For example, opening a dialog is an asynchronous
 * action.  By setting the animation duration to 0 simply makes the dialog appear to open immediately.  Events related to the dialog opening
 * are still fired asynchronously, though with a much shorter delay.</p>
 * 
 * <h4>2. For All Instances of a JET Component.</h4>
 * 
 * <p>Default animations for JET components are defined by component-specific theme variables.  Changing the values of the theme variables for a particular component
 * will affect the default animations for all instances of that component.  These theme variables are listed in the API documentation
 * for the <code class="prettyprint">ojAnimateStart</code> event of each component.</p>
 * 
 * <h4>3. For One Instance of a JET Component</h4> 
 *
 * <p>Applications can customize animations triggered by actions in some components by listening for <code class="prettyprint">ojAnimateStart/ojAnimateEnd</code>
 *    events and override action specific animations.  See the documentation of individual components for support details of <code
 *    class="prettyprint">ojAnimateStart/ojAnimateEnd</code> events and the associated actions.</p>
 * <p>To customize an animation, applications first listen to <code class="prettyprint">ojAnimateStart</code> event and cancel the default animation.  Then
 *    specify the new animation in one of several ways:</p>
 * <ul>
 *   <li>Call one of the animation effect methods in oj.AnimationUtils.</li>
 *   <li>Call a 3rd-party animation function with a Javascript API, such as jQuery, GreenSock, Velocity.js, etc.</li>
 *   <li>Define action-specific CSS style classes on the animated item.  When an action triggers animation, a marker class of the form "oj-animate-&lt;action&gt;" 
 *       (e.g. "oj-animate-open") is added to the animated element.  After a slight delay, a second marker class of the form "oj-animate-&lt;action&gt;-active"
 *       (e.g. "oj-animate-open-active") is added.  This allows application to define CSS transition on the element.</li>
 * </ul>
 *
 * <h4>Examples</h4>
 * <br>
 * <i>Disable a default "open" animation:</i>
 * <pre class="prettyprint"><code>
 * myComponent.addEventListener( "ojAnimateStart", function( event ) {
 *   if (event.detail.action == "open") {
 *     event.preventDefault();
 *     event.detail.endCallback();
 *   }
 * });
 * </code></pre>
 * <br>
 * <i>Customize a default "open" animation with oj.AnimationUtils method:</i>
 * <pre class="prettyprint"><code>
 * myComponent.addEventListener( "ojAnimateStart", function( event ) {
 *   if (event.detail.action == "open") {
 *     event.preventDefault();
 *     oj.AnimationUtils.slideIn(event.detail.element).then(event.detail.endCallback);
 *   }
 * });
 * </code></pre>
 * <br>
 * <i>Customize a default "close" animation with jQuery method:</i>
 * <pre class="prettyprint"><code>
 * myComponent.addEventListener( "ojAnimateStart", function( event ) {
 *   if (event.detail.action == "close") {
 *     event.preventDefault();
 *     $(event.detail.element).fadeOut(event.detail.endCallback);
 *   }
 * });
 * </code></pre>
 * <br>
 * <i>Customize a default "update" animation with CSS style classes:</i>
 * <pre class="prettyprint"><code>
 * // Cancel the default animation in the event listener
 * myComponent.addEventListener( "ojAnimateStart", function( event ) {
 *   if (event.detail.action == "update") {
 *     event.preventDefault();
 *     event.detail.endCallback();
 *   }
 * });
 *
 * </code></pre>
 *
 * <pre class="prettyprint"><code>
 * /* Define new animation in CSS
 *    Different selectors may be needed to target the CSS correctly *&#47;
 *
 * /* State to animate from is marked by oj-animate-&lt;action&gt; class *&#47; 
 * .selector .oj-animate-update {
 *   color: red;
 * }
 *
 * /* State to animate to is marked by oj-animate-&lt;action&gt; and oj-animate-&lt;action&gt;-active classes *&#47; 
 * .selector .oj-animate-update.oj-animate-update-active {
 *   transition: color 1s;
 *   color: black;
 * }
 * </code></pre>
 *
 * <h3 id="busy-state-section">
 *   Adding Busy State
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#busy-state-section"></a>
 * </h3>
 *
 * <p>Animations are asynchronous by nature.  Sometimes applications may need to wait for an animation to end before
 *    proceeding with other operations.  All the effect methods in oj.AnimationUtils return promises that are
 *    resolved when the animations end.</p>
 * <p>In cases where applications use the {@link oj.BusyContext} class to track the busy state of components or pages, it is
 *    up to the callers of the effect methods to add busy state to the appropriate context, which may or may not
 *    be the context that contains the element being animated.</p>
 *
 * <h4>Examples</h4>
 * <br>
 * <i>Add a busy state while an animation is in progress:</i>
 * <pre class="prettyprint"><code>
 * // Context node is usually the animated element but can also be a node for any
 * // context that wants to wait for the animation to end.
 * var contextNode = element;
 * var busyContext = oj.Context.getContext(contextNode).getBusyContext();
 * var resolveFunc = busyContext.addBusyState({"description": "Animation in progress"});
 * oj.AnimationUtils.slideOut(element).then(resolveFunc);
 * </code></pre>
 *
 * @namespace
 * @since 2.1
 * @export
 */
oj.AnimationUtils = {};

// Return a platform-dependent property or event name from a base name
oj.AnimationUtils._getName = function(element, baseName)
{
  if (!oj.AnimationUtils._nameMap)
  {
    oj.AnimationUtils._nameMap = {};
    var nameMap = oj.AnimationUtils._nameMap;
    var style = element.style;
    
    // Property names
    nameMap['backfaceVisibility'] = (style['webkitBackfaceVisibility'] !== undefined) ? 'webkitBackfaceVisibility' : 'backfaceVisibility';
    nameMap['transform'] = (style['webkitTransform'] !== undefined) ? 'webkitTransform' : 'transform';
    nameMap['transformOrigin'] = (style['webkitTransformOrigin'] !== undefined) ? 'webkitTransformOrigin' : 'transformOrigin';
    nameMap['transition'] = (style['webkitTransition'] !== undefined) ? 'webkitTransition' : 'transition';
    
    // Event names
    nameMap['transitionend'] = (style['webkitTransition'] !== undefined) ? 'webkitTransitionEnd' : 'transitionend';
  }
  
  var mappedName = oj.AnimationUtils._nameMap[baseName];
  
  return mappedName ? mappedName : baseName;
};

oj.AnimationUtils._getElementStyle = function(element, baseName)
{
  return element.style[oj.AnimationUtils._getName(element, baseName)];
};

oj.AnimationUtils._setElementStyle = function(element, baseName, value)
{
  element.style[oj.AnimationUtils._getName(element, baseName)] = value;
};

/**
 * Main utility function for starting a css transition on an element.<br>
 * Currently this function assumes the following:<br>
 * 1. If multiple properties are animated, they all use the same delay and duration.<br>
 * All css property names should be specified in camel case.
 * @param {Element} element  the HTML element to animate
 * @param {Object} fromState  the css class and properties for setting up the starting state
 * @param {Object} toState  the css class and properties for setting up the ending state
 * @param {Object} options  the common options for the css transition or animation.  This
 *                          include 'delay', 'duration', 'timingFunction', and 'persist'.
 * @param {Array} transProps  an array of css properties being transitioned.
 * @param {Array=} persistProps  an array of css properties to persist if the persist option is set.
 *                               If this is omitted, it will be set to transProps.
 * @return {Promise|IThenable} a promise that will be resolved when the animation ends
 * @private
 */
oj.AnimationUtils._animate = function(element, fromState, toState, options, transProps, persistProps)
{
  var propArray = [].concat(transProps);

  var doAnimate = function(resolve, reject)
  {
    var endListener = function(event) {
      // event.propertyName is the hyphenated name.  Entries in propArray is the 
      // camel-case name without prefix.  So we drop any prefix and convert 
      // event.propertyName to camel-case before finding it in propArray. 
      var basePropName = event.propertyName.indexOf('-webkit-') == 0 ? event.propertyName.substr(8) : event.propertyName;
      basePropName = oj.AnimationUtils._getCamelCasePropName(basePropName);
      var idx = propArray.indexOf(basePropName);
      if (idx > -1)
      {
        if (propArray.length > 1)
        {
          propArray.splice(idx, 1);
        }
        else
        {
          resolvePromise();
        }
      }
    };
    var requestId = 0;
    var promiseResolved = false;
    var resolvePromise = function() {
      if (!promiseResolved)
      {
        if (requestId)
        {
          window.cancelAnimationFrame(requestId);
          requestId = 0;
        }
        
        element.removeEventListener(oj.AnimationUtils._getName(element, 'transitionend'), endListener);

        if (resolve)
        {
          resolve(true);
        }
        promiseResolved = true;          
      }
    };

    // For css transition, specify the transition value when applying the toState
    // since we don't want to trigger the transition prematurely.
    if (toState == null)
    {
      toState = {};
    }
    if (toState['css'] == null)
    {
      toState['css'] = {};
    }
    toState['css']['transition'] = oj.AnimationUtils._createTransitionValue(element, transProps, options);

    // Save the orignal style so that we can restore it later if needed
    var effectCount = oj.AnimationUtils._saveStyle(element, fromState, toState, options, persistProps || transProps);

    oj.AnimationUtils._applyState(element, fromState, effectCount > 1);

    element.addEventListener(oj.AnimationUtils._getName(element, 'transitionend'), endListener);

    var duration = options['duration'];
    var delay = options['delay'];
    var skipPromise = options['_skipPromise'];

    function transitionFunc() {
      requestId = 0;
      oj.AnimationUtils._applyState(element, toState, effectCount > 1);
    }

    if (fromState == null)
    {
      // If there is no fromState, assume that we are transitioning from the
      // current state to a new state, so just apply toState immediately.
      transitionFunc();
    }
    else
    {
      if (!options['_noReflow'])
      {
        // If the final state is the same as current state,
        // requestAnimationFrame may not trigger a transition.  Need to force
        // a reflow after applying the initial state by getting one of several 
        // properties that cause reflow.
        //
        // Assign it to an export object so that Closure compiler will not remove
        // this as dead code
        oj.AnimationUtils._x = element.offsetWidth;
      }

      // Add the toState after a delay.  This is necessary to trigger css 
      // transition.
      requestId = window.requestAnimationFrame(transitionFunc);
    }
    
    var totalMs = oj.AnimationUtils._getTotalTiming(duration, delay);
    if (!skipPromise)
    {
      // There are situations in which the transitionend event is never
      // fired (e.g. removing the transiton property or setting the display
      // property to none.)
      // Add a timeout to avoid having unresolved promise.
      setTimeout(resolvePromise, totalMs + 100);
    }
  };

  // Special option to skip promise to reduce overhead.  A side effect is that
  // no clean up will be done on the element, which is fine if the element is
  // temporary
  if (options['_skipPromise'])
  {
    doAnimate(null, null);
    return null;
  }
  else
  {
    var promise = new Promise(doAnimate);
    return promise.then(function() {
      // Remove any temporary effect class when the promise is fulfilled.
      // Do not remove them in the endListener, since the promise fulfillment
      // callback is not in the same animation frame and occurs later than
      // the endListener.  Because any caller cleanup is done on promise
      // fulfillment, the element may flash if we remove the class too early.

      if (fromState && fromState['addClass'])
      {
        $(element).removeClass(fromState['addClass']);
      }
      
      if (toState && toState['addClass'])
      {
        $(element).removeClass(toState['addClass']);
      }

      oj.AnimationUtils._restoreStyle(element);
    });
  }
};

// Save the element style from a property set
oj.AnimationUtils._saveCssValues = function(element, css, savedStyle, persistProps)
{
  for (var cssProp in css)
  {
    if (!savedStyle.hasOwnProperty(cssProp) && (!persistProps || persistProps.indexOf(cssProp) == -1))
    {
      savedStyle[cssProp] = oj.AnimationUtils._getElementStyle(element, cssProp);
    }
  }
};

// Save the original element style before animating it
oj.AnimationUtils._saveStyle = function(element, fromState, toState, options, persistProps)
{
  var savedStyle = element._ojSavedStyle || {};
  var fromStateCss = (fromState && fromState['css']) ? fromState['css'] : {};
  var toStateCss = (toState && toState['css']) ? toState['css'] : {};

  if (!(options && options['persist'] === 'all'))
  {
    persistProps = null;
  }

  oj.AnimationUtils._saveCssValues(element, fromStateCss, savedStyle, persistProps);
  oj.AnimationUtils._saveCssValues(element, toStateCss, savedStyle, persistProps);

  element._ojSavedStyle = savedStyle;

  // Remember how many times this is called to allow composite animation
  var effectCount = element._ojEffectCount || 0;
  element._ojEffectCount = ++effectCount;
  
  return effectCount;
};

// Restore the original element style
oj.AnimationUtils._restoreStyle = function(element)
{
  var effectCount = element._ojEffectCount;

  // In case of composite animation, restore style when the last effect has ended
  if (effectCount > 1)
  {
    element._ojEffectCount = effectCount - 1;
  }
  else
  {
    var savedStyle = element._ojSavedStyle;
    if (savedStyle)
    {
      for (var prop in savedStyle)
      {
        oj.AnimationUtils._setElementStyle(element, prop, savedStyle[prop]);
      }

      delete element._ojSavedStyle;
      delete element._ojEffectCount;
    }
  }
};

// Get the corresponding camel-cased property name
oj.AnimationUtils._getCamelCasePropName = function(propName)
{
  if (propName.indexOf('-') >= 0)
  {
    var newName = '';

    var strArray = propName.split('-');
    for (var i = 0; i < strArray.length; i++)
    {
      var subStr = strArray[i];

      if (subStr)
      {
        if (newName)
        {
          newName += subStr.charAt(0).toUpperCase() + subStr.slice(1);
        }
        else
        {
          // Keep the first segment in lower case
          newName = subStr;
        }
      }
    }
    
    return newName;
  }
  
  return propName;
};

// Get the corresponding hyphenated property name
oj.AnimationUtils._getHyphenatedPropName = function(propName)
{
  var newName = propName.replace(/([a-z])([A-Z])/g, '$1-$2').toLowerCase();
  if (newName.indexOf('webkit') == 0)
  {
    newName = '-' + newName;
  }
  return newName;
};

// Concatenate value for style property that allows multiple values 
oj.AnimationUtils._concatMultiValue = function(element, state, propName, defaultPrefix, separator)
{
  if (state['css'][propName])
  {
    var currPropValue = oj.AnimationUtils._getElementStyle(element, propName);
    if (currPropValue && currPropValue.indexOf(defaultPrefix) != 0)
    {
      state['css'][propName] = currPropValue + separator + state['css'][propName];
    }
  }
};

oj.AnimationUtils._splitTransform = function(transform)
{
  var array = [];

  if (transform && transform != 'none')
  {
    var index;
    
    while ((index = transform.indexOf(')')) > 0)
    {
      var funcExpr = transform.substr(0, index + 1);
      array.push(funcExpr.trim());
      transform = transform.slice(index + 1);
    }
  }
  
  return array;
};

oj.AnimationUtils._getTransformFuncName = function(funcExpr)
{
  var index = funcExpr.indexOf('(');
  if (index >= 1)
  {
    return funcExpr.substr(0, index);
  }
  
  return funcExpr;
};

// Apply the transform style
oj.AnimationUtils._applyTransform = function(element, newTransform)
{
  var oldTransform = oj.AnimationUtils._getElementStyle(element, 'transform');
  var oldTransformArray = oj.AnimationUtils._splitTransform(oldTransform);
  var newTransformArray = oj.AnimationUtils._splitTransform(newTransform);
  var extraTransformArray = [];
  
  for (var i = 0; i < newTransformArray.length; i++)
  {
    var funcName = oj.AnimationUtils._getTransformFuncName(newTransformArray[i]);
    var match = false;

    if (funcName)
    {
      for (var j = 0; j < oldTransformArray.length; j++)
      {
        // Search for funcName + '(' since some transform functions can be the
        // prefix of other functions such as translate and translateZ.
        if (oldTransformArray[j].indexOf(funcName + '(') == 0)
        {
          // Replace any matching old transform function with the new one
          oldTransformArray[j] = newTransformArray[i];
          match = true;
        }
      }
    }

    // Keep track of any new transform function that wasn't specified    
    if (!match)
    {
      extraTransformArray.push(newTransformArray[i]);
    }
  }
  
  // Concatenate the updated transform list with newly added list
  oldTransformArray = oldTransformArray.concat(extraTransformArray);

  // Return the transform list as a string  
  return oldTransformArray.join(' ');
};

oj.AnimationUtils._applyState = function(element, state, isComposite)
{
  if (state)
  {
    if (state['css'])
    {
      var transitionPropName = 'transition';
      var transformPropName = 'transform';
      
      // For composite animation, we need to concatenate certain property values
      // instead of replacing them
      if (isComposite)
      {
        oj.AnimationUtils._concatMultiValue(element, state, transitionPropName, 'all', ', ');
      }

      if (state['css'][transformPropName])
      {
        state['css'][transformPropName] = oj.AnimationUtils._applyTransform(element, state['css'][transformPropName]);
      }

      var newStyle = state['css'];
      for (var prop in newStyle)
      {
        oj.AnimationUtils._setElementStyle(element, prop, newStyle[prop]);
      }
    }
    
    if (state['addClass'])
    {
      $(element).addClass(state['addClass']);
    }
    
    if (state['removeClass'])
    {
      $(element).removeClass(state['removeClass']);
    }
  }
};

// Get a timing value in millisecond from a string such as duration and delay
oj.AnimationUtils._getTimingValue = function(timingStr)
{
  var timingValue = parseFloat(timingStr);
  if (isNaN(timingValue))
  {
    return 0;
  }
  
  return (timingStr.indexOf('ms') > -1) ? timingValue : timingValue * 1000;
};

oj.AnimationUtils._getTotalTiming = function(duration, delay)
{
  var durationMs = oj.AnimationUtils._getTimingValue(duration);
  if (durationMs > 0)
  {
    var delayMs = delay ? oj.AnimationUtils._getTimingValue(delay) : 0;
    return (durationMs + delayMs);
  }
  
  return 0;
};

oj.AnimationUtils._calcCssTime = function(propertyStr, delayStr, durationStr)
{
  var propertyArray = propertyStr.split(',');
  var delayArray = delayStr.split(',');
  var durationArray = durationStr.split(',');
  var propertyLen = propertyArray.length;
  var delayLen = delayArray.length;
  var durationLen = durationArray.length;
  var maxTime = 0;
  
  for (var i = 0; i < propertyLen; i++)
  {
    var duration = durationArray[i % durationLen];
    var delay = delayArray[i % delayLen];
    var totalMs = oj.AnimationUtils._getTotalTiming(duration, delay);

    maxTime = Math.max(maxTime, totalMs);
  }

  return maxTime;  
};

oj.AnimationUtils._calcEffectTime = function(element)
{
  var style = window.getComputedStyle(element);
  var propertyStr, delayStr, durationStr;
  
  propertyStr = style['animationName'] || style['webkitAnimationName'];
  delayStr = style['animationDelay'] || style['webkitAnimationDelay'];
  durationStr = style['animationDuration'] || style['webkitAnimationDuration'];
  var animationTime = oj.AnimationUtils._calcCssTime(propertyStr, delayStr, durationStr);

  propertyStr = style['transitionProperty'] || style['webkitTransitionProperty'];
  delayStr = style['transitionDelay'] || style['webkitTransitionDelay'];
  durationStr = style['transitionDuration'] || style['webkitTransitionDuration'];
  var transitionTime = oj.AnimationUtils._calcCssTime(propertyStr, delayStr, durationStr);
  
  return Math.max(animationTime, transitionTime);
};

// Fill in empty timing options from a set of source options
oj.AnimationUtils._fillEmptyOptions = function(targetOptions, sourceOptions)
{
  targetOptions['delay'] = targetOptions['delay'] || sourceOptions['delay'];
  targetOptions['duration'] = targetOptions['duration'] || sourceOptions['duration'];
  targetOptions['timingFunction'] = targetOptions['timingFunction'] || sourceOptions['timingFunction'];
  targetOptions['persist'] = targetOptions['persist'] || sourceOptions['persist'];
};

// Trigger oj custom event
oj.AnimationUtils._triggerEvent = function(element, eventName, ui, component)
{
  var defaultPrevented;

  if (component && component._trigger)
  {
    // _trigger() returns false if preventDefault has been called
    defaultPrevented = !component._trigger(eventName, null, ui);
  }
  else
  {
    var ojEventType = 'oj' + eventName.substr(0,1).toUpperCase() + eventName.substr(1);
    var customEvent = new CustomEvent(ojEventType, {'detail':ui, 'bubbles':true, 'cancelable':true});
    var eventTarget = component || element;
    if (eventTarget.dispatchEvent)
    {
      eventTarget.dispatchEvent(customEvent);
    }
    defaultPrevented = customEvent.defaultPrevented;
  }
  
  return defaultPrevented;
}  

/**
 * Internal method for starting an animation.
 * @param {Element} element  the HTML element to animate
 * @param {string} action  a component-defined string that identifies the action starting
 *                         the animation, such as "open", "close", "add", "remove", etc.
 * @param {string|Object|Array} effects  The name of one of the effect methods
 *                        in oj.AnimationUtils, or an object that specifies the
 *                        effect method and its options, such as:
 *                        {'effect': 'fadeOut', 'endOpacity': 0.5}, or an array of the above.
 * @param {Object=} component  the component that owns the HTML element
 *                             to animate.  If this is specified and it has a
 *                             _trigger method (jQuery UI widget), animation events will
 *                             be triggered on the component via jQuery UI _trigger(),
 *                             so that listeners specified as event options will work.
 *                             If this is specified but it doesn't have a _trigger
 *                             method (HTML element), animation events will be triggered
 *                             on the component via dispatchEvent.
 *                             If this is not specified, animation events will be triggered
 *                             on the animated HTML element via dispatchEvent.
 * @return {Promise} a promise that will be resolved when the animation ends
 * @export
 * @ignore
 */
oj.AnimationUtils.startAnimation = function(element, action, effects, component)
{
  // Temporary fix for callers that are passing in jQuery object
  element = $(element)[0];

  var promise = new Promise(
    function(resolve, reject)
    {    
      var jelem = $(element);
      var fromMarker = 'oj-animate-' + action;
      var toMarker = fromMarker + '-active';
      var eventProcessed = false;
      var markerProcessed = false;
      var resolvePromise = function() {
        if (eventProcessed && markerProcessed)
        {
          jelem.removeClass(fromMarker);
          jelem.removeClass(toMarker);
          oj.AnimationUtils._restoreStyle(element);
          resolve(true);
          
          var ui = {'action': action, 'element': element};
          oj.AnimationUtils._triggerEvent(element, 'animateEnd', ui, component);
        }
      };
      var eventCallback = function() {
        eventProcessed = true;
        resolvePromise();
      };
      var markerCallback = function() {
        markerProcessed = true;
        resolvePromise();
      };

      // This will add a ref count so that the style is not restored until
      // all effects and user-defined css transitions have ended.  Otherwise
      // there may be screen flash if 'persist' != 'all'.
      oj.AnimationUtils._saveStyle(element, null, null, null, null);

      // Trigger ojanimatestart event so that app can prevent default animation 
      // and define custom effect in JS
      var ui = {'action': action, 'element': element, 'endCallback': eventCallback};
      var defaultPrevented = oj.AnimationUtils._triggerEvent(element, 'animateStart', ui, component);

      // Continue animation handling if app didn't preventDefault
      if (!defaultPrevented)
      {
        var effectArray = [].concat(effects);
        var promiseArray = [];
        var lastOptions = {};

        for (var i = 0; i < effectArray.length; i++)
        {
          var animationEffect = effectArray[i];
          var effectName = '';
          var effectOptions;
          
          // Start any explicit animation effect 
          if (animationEffect != null && animationEffect != 'none')
          {
            if (typeof animationEffect == 'string')
            {
              effectName = animationEffect;
              effectOptions = {};
            }
            else if (typeof animationEffect == 'object')
            {
              effectName = animationEffect['effect'];
              effectOptions = $.extend({}, animationEffect);
            }

            // Fill in empty timing options with what was specified last
            oj.AnimationUtils._fillEmptyOptions(effectOptions, lastOptions);
            
            // Remember the last set of options
            lastOptions = $.extend({}, effectOptions);
          }
            
          if (effectName && oj.AnimationUtils[effectName])
          {
            promiseArray.push(oj.AnimationUtils[effectName](element, effectOptions));
          }
        }
        
        if (promiseArray.length)
        {
          Promise.all(promiseArray).then(eventCallback);
        }
        else
        {
          eventCallback();
        }
      }

      // Add marker class so that app can define custom effect in CSS      
      jelem.addClass(fromMarker);
      var requestId = window.requestAnimationFrame(function() {
        requestId = 0;
        
        jelem.addClass(toMarker);
        
        var totalMs = oj.AnimationUtils._calcEffectTime(element);
        if (totalMs > 0)
        {
          // Set a timeout to resolve the promise.  We can't rely on
          // transitionend event since there can be multiple transition
          // properties, or the transition is never triggered, or the transition
          // is cancelled.
          setTimeout(markerCallback, totalMs + 100);
        }
        else
        {
          markerCallback();
        }
      });
      
      // In case we are in the background and requestAnimationFrame is not
      // called, have a timeout that cancel the request and resolve promise
      setTimeout(function() {
        if (requestId)
        {
          window.cancelAnimationFrame(requestId);
          requestId = 0;
          markerCallback();
        }
      }, 1000);
    }
  );
  
  return promise;
};

oj.AnimationUtils._mergeOptions = function(effect, options)
{
  if (oj.AnimationUtils._defaultOptions == null)
  {
    oj.AnimationUtils._defaultOptions = oj.ThemeUtils.parseJSONFromFontFamily('oj-animation-effect-default-options');
  }
 
  // At the minimum, we should have a duration.  Merge any theming defaults
  // and then any user options to it.  
  return $.extend({'duration': '400ms'},
                  oj.AnimationUtils._defaultOptions ? oj.AnimationUtils._defaultOptions[effect] : null,
                  options);
};

oj.AnimationUtils._createTransitionValue = function(element, transProps, options)
{
  var transValue = '';

  if (transProps)
  {
    for (var i = 0; i < transProps.length; i++)
    {
      var propName = oj.AnimationUtils._getName(element, transProps[i]);
      var hyphenatedName = oj.AnimationUtils._getHyphenatedPropName(propName);

      transValue += (i > 0 ? ', ' : '') + hyphenatedName + ' ' + options['duration'];
      
      if (options['timingFunction'])
        transValue += ' ' + options['timingFunction'];
      
      if (options['delay'])
        transValue += ' ' + options['delay'];
    }
  }
  
  return transValue;
};
  
oj.AnimationUtils._fade = function(element, options, effect, startOpacity, endOpacity)
{
  options = oj.AnimationUtils._mergeOptions(effect, options);
  
  var fromState = {'css': {'opacity': startOpacity}};
  var toState = {'css': {'opacity': endOpacity}};

  if (options)
  {    
    if (options['startOpacity'])
      fromState['css']['opacity'] = options['startOpacity'];
   
    if (options['endOpacity'])
      toState['css']['opacity'] = options['endOpacity'];
  }

  return oj.AnimationUtils._animate(element, fromState, toState, options, ['opacity']);
};

/**
 * Animaton effect method for fading in a HTML element.
 * 
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {number=} options.startOpacity starting opacity. Default is 0.
 * @param {number=} options.endOpacity  ending opacity. Default is 1.
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.fadeIn = function(element, options)
{
  return oj.AnimationUtils._fade(element, options, 'fadeIn', 0, 1);
};

/**
 * Animaton effect method for fading out a HTML element.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {number=} options.startOpacity starting opacity. Default is 1.
 * @param {number=} options.endOpacity  ending opacity. Default is 0.
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.fadeOut = function(element, options)
{
  return oj.AnimationUtils._fade(element, options, 'fadeOut', 1, 0);
};

/**
 * Animaton effect method for expanding a HTML element.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.direction direction to expand. Valid values are "height", "width", or "both". Default is "height".
 * @param {string=} options.startMaxHeight starting max-height value to expand from.  Default is "0".
 * @param {string=} options.endMaxHeight ending max-height value to expand to.  Default is natural element height.
 * @param {string=} options.startMaxWidth starting max-width value to expand from.  Default is "0".
 * @param {string=} options.endMaxWidth starting max-width value to expand to.  Default is natural element width.
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.expand = function(element, options)
{
  return oj.AnimationUtils._expandCollapse(element, options, true);
};

/**
 * Animaton effect method for collapsing a HTML element.
 * <p>When using this method to hide an element, the element should not have any border
 * or padding, because border and padding are visible even if the element's height
 * is set to 0. The use of "box-sizing: border-box" style doesn't change this behavior.
 * If the element needs border and padding, create a wrapper element around it and
 * call this method on the wrapper element instead.</p>
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.direction direction to collapse. Valid values are "height", "width", or "both". Default is "height".
 * @param {string=} options.startMaxHeight starting max-height value to collapse from.  Default is natural element height.
 * @param {string=} options.endMaxHeight ending max-height value to collapse to.  Default is "0".
 * @param {string=} options.startMaxWidth starting max-width value to collapse from.  Default is natural element width.
 * @param {string=} options.endMaxWidth starting max-width value to collapse to.  Default is "0".
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.collapse = function(element, options)
{
  return oj.AnimationUtils._expandCollapse(element, options, false);
};

// Wrap table row content and return an array of wrapper elements to animate
oj.AnimationUtils._wrapRowContent = function(row, rowHeight)
{
  var wrappers = [];
  var cells = row.children;
  var cellsPadding = [];
  var cellsTextAlign = [];
  
  // Collect all the needed style before modifying the DOM.  Otherwise it
  // causes additional reflow and takes more time.

  row._ojSavedHeight = row.style['height'];

  for (var i = 0; i < cells.length; i++)
  {
    var cell = cells[i];
    var cellStyle = window.getComputedStyle(cell);
    cellsPadding.push(cellStyle['padding']);
    cellsTextAlign.push(cellStyle['textAlign']);

    // Remember the inline padding style (not computed style)
    cell._ojSavedPadding = cell.style['padding'];
  }

  // Start modifying the DOM
  
  for (var i = 0; i < cells.length; i++)
  {
    var cell = cells[i];

    // Create the outer wrapper
    var outerWrapper = document.createElement('div');
    outerWrapper.style['overflow'] = 'hidden;';
    
    // Create the inner wrapper    
    var innerWrapper = document.createElement('div');
    innerWrapper.style['display'] = 'table-cell';
    innerWrapper.style['verticalAlign'] = 'middle';
    innerWrapper.style['boxSizing'] = 'border-box';
    innerWrapper.style['height'] = rowHeight;
    innerWrapper.style['padding'] = cellsPadding[i];
    innerWrapper.style['textAlign'] = cellsTextAlign[i];
    
    // Append inner wrapper to outer wrapper
    outerWrapper.appendChild(innerWrapper); //@HTMLUpdateOK; innerWrapper is constructed by component code and is not using string passed in through any APIs.

    // Transfer children of cell to inner wrapper
    while (cell.firstChild)
    {
      innerWrapper.appendChild(cell.firstChild); //@HTMLUpdateOK; cell.firstChild is constructed by component code and is not using string passed in through any APIs.
    }

    // Finally append the outer wrapper back to the cell
    cell.appendChild(outerWrapper); //@HTMLUpdateOK; outerWrapper is constructed by component code and is not using string passed in through any APIs.
    
    // Set the cell padding to 0 so that it can be completely collapsed
    cell.style['padding'] = '0';
    
    wrappers.push(outerWrapper);
  };
  
  // Set the row height to 0 so that it can be completely collapsed
  row.style['height'] = '0';
  
  return wrappers;
};

// Unwrap table row content
oj.AnimationUtils._unwrapRowContent = function(row)
{
  var cells = row.children;
  
  for (var i = 0; i < cells.length; i++)
  {
    var cell = cells[i];

    var outerWrapper = cell.children[0];
    if (outerWrapper)
    {
      var innerWrapper = outerWrapper.children[0];
      if (innerWrapper)
      {
        while (innerWrapper.firstChild)
        {
          cell.appendChild(innerWrapper.firstChild); //@HTMLUpdateOK; innerWrapper.firstChild is constructed by component code and is not using string passed in through any APIs.
        }
      }
      
      // Remove the outer wrapper will also remove the inner wrapper
      cell.removeChild(outerWrapper);
    }

    // Restore any inline padding style to the cell
    cell.style['padding'] = cell._ojSavedPadding;
    delete cell._ojSavedPadding;        
  }

  // Restore any inline height style to the row
  row.style['height'] = row._ojSavedHeight;
  delete row._ojSavedHeight;
};

// Expand or collapse a table row
oj.AnimationUtils._expandCollapseRow = function(element, options, isExpand)
{
  var promise;
  var rowHeight = element.offsetHeight + 'px';
  var wrappers = oj.AnimationUtils._wrapRowContent(element, rowHeight);
  var wrapperOptions = $.extend({}, options);

  // Set the known max height into the options so that each cell doesn't
  // have to recalculate it
  if (isExpand)
  {
    if (!wrapperOptions['endMaxHeight'])
    {
      wrapperOptions['endMaxHeight'] = rowHeight;
    }
  }
  else
  {
    if (!wrapperOptions['startMaxHeight'])
    {
      wrapperOptions['startMaxHeight'] = rowHeight;
    }
  }
  
  // Set persist option so that we don't need to restore the style after 
  // animation.  The wrapper will be removed anyway.
  wrapperOptions['persist'] = 'all';
  
  // Set internal _noReflow option so that we don't force reflow on Firefox.
  // New elements will naturally cause reflow.
  wrapperOptions['_noReflow'] = true;
  
  if (wrappers.length)
  {
    // Animate all the cells
    for (var i = 0; i < wrappers.length; i++)
    {
      // We only need one promise/row.  Skip promise for all cells except first.
      if (i == 0)
      {
        wrapperOptions['_skipPromise'] = false;
        promise = oj.AnimationUtils._expandCollapse(wrappers[i], wrapperOptions, isExpand);
      }
      else
      {
        wrapperOptions['_skipPromise'] = true;
        oj.AnimationUtils._expandCollapse(wrappers[i], wrapperOptions, isExpand);
      }
    }
  }
  else
  {
    promise = Promise.resolve();
  }

  return promise.then(function() {
    if (options == null || options['persist'] != 'all')
    {
      oj.AnimationUtils._unwrapRowContent(element);
    }
  });
};

// Determine the min/max width/height used for animation
oj.AnimationUtils._getSizeLimit = function(element, style, optionValue, isLower, isWidth)
{
  var limitValue = optionValue;

  if (!limitValue)
  {
    if (isLower)
    {
      limitValue = '0';
    }
    else
    {
      var cssValue = isWidth ? style['maxWidth'] : style['maxHeight'];
      if (cssValue !== "none")
      {
        limitValue = cssValue;
      }
      else
      {
        limitValue = (isWidth ? element.offsetWidth : element.offsetHeight) + 'px';
      }
    }
  }
  
  return limitValue;
};

oj.AnimationUtils._expandCollapse = function(element, options, isExpand)
{
  // Handle the case where the element is a <tr> element.  We need to wrap the
  // content of every child <td> and animate the wrappers because the min height
  // of <tr> and <td> are limited by their contents.
  if (element && element.tagName == 'TR')
  {
    return oj.AnimationUtils._expandCollapseRow(element, options, isExpand);
  }
  
  options = oj.AnimationUtils._mergeOptions(isExpand ? 'expand' : 'collapse', options);

  var fromState = {};
  var toState = {};

  var direction = options['direction'] || "height";

  var fromCSS = fromState['css'] = {};
  var toStateCSS = toState['css'] = {};

  var style = window.getComputedStyle(element);
  var transProps = [];
  if(direction === "both" || direction === "height") 
  {
    var startMaxHeight = oj.AnimationUtils._getSizeLimit(element, style, options['startMaxHeight'], isExpand, false);
    var endMaxHeight = oj.AnimationUtils._getSizeLimit(element, style, options['endMaxHeight'], !isExpand, false);

    fromCSS['maxHeight'] = startMaxHeight;
    toStateCSS['maxHeight'] = endMaxHeight;
    transProps.push('maxHeight');
  }

  if(direction === "both" || direction === "width") 
  {
    var startMaxWidth = oj.AnimationUtils._getSizeLimit(element, style, options['startMaxWidth'], isExpand, true);
    var endMaxWidth = oj.AnimationUtils._getSizeLimit(element, style, options['endMaxWidth'], !isExpand, true);
    
    fromCSS['maxWidth'] = startMaxWidth;
    toStateCSS['maxWidth'] = endMaxWidth;
    transProps.push('maxWidth');
  }
  
  // expand and collapse needs overflow hidden to hide the content
  fromCSS['overflow'] = options['overflow'] ? options['overflow'] : 'hidden';
  
  var persistProps = [].concat(transProps);
  persistProps.push('overflow');

  return oj.AnimationUtils._animate(element, fromState, toState, options, transProps, persistProps);
};

/**
 * Animaton effect method for zooming in a HTML element.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.axis the axis along which to scale the element. Valid values are "x", "y", or "both". Default is "both".
 * @param {string=} options.transformOrigin set the CSS transform-origin property, which controls the anchor point for the zoom. Default is "center".
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.zoomIn = function(element, options)
{
  return oj.AnimationUtils._zoom(element, options, true);
};

/**
 * Animaton effect method for zooming out a HTML element.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.axis the axis along which to scale the element. Valid values are "x", "y", or "both". Default is "both".
 * @param {string=} options.transformOrigin set the CSS transform-origin property, which controls the anchor point for the zoom. Default is "center".
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.zoomOut = function(element, options)
{
  return oj.AnimationUtils._zoom(element, options, false);
};

oj.AnimationUtils._zoom = function(element, options, isIn)
{
  options = oj.AnimationUtils._mergeOptions(isIn ? 'zoomIn' : 'zoomOut', options);

  var fromState = {};
  var toState = {};

  var axis = options['axis'] || "both";
    
  var scale = axis === "both" ? "scale" : (axis === "x" ? "scaleX" : "scaleY");
  var fromCSS = fromState['css'] = {};
  var toStateCSS = toState['css'] = {};
  var transformPropName = 'transform';
  var transformOriginPropName = 'transformOrigin';

  fromCSS[transformPropName] = scale + "(" + (isIn ? 0 : 1) + ") translateZ(0)";
  toStateCSS[transformPropName] = scale + "(" + (isIn ? 1 : 0) + ") translateZ(0)";

  fromCSS[transformOriginPropName] = options['transformOrigin'] || "center";

  return oj.AnimationUtils._animate(element, fromState, toState, options, [transformPropName]);
};

/**
 * Animaton effect method for sliding in a HTML element.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.direction Direction of the slide. Valid values are "left", "top", "right", "bottom", "start", and "end". Default is "start".
 *                                    This option is ignored if either offsetX or offsetY is specified.
 * @param {string=} options.offsetX The offset on the x-axis to translate from. This value must be a number followed by a unit such as "px", "em", etc.
 *                                  If moving in a horizontal direction, default to element width. Otherwise, default to "0px".
 * @param {string=} options.offsetY The offset on the y-axis to translate from. This value must be a number followed by a unit such as "px", "em", etc.
 *                                  If moving in a vertical direction, default to element height. Otherwise, default to "0px".
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.slideIn = function(element, options)
{
  return oj.AnimationUtils._slide(element, options, true);
};

/**
 * Animaton effect method for sliding out a HTML element.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.direction Direction of the slide. Valid values are "left", "top", "right", "bottom", "start", and "end". Default is "start".
 *                                    This option is ignored if either offsetX or offsetY is specified.
 * @param {string=} options.offsetX The offset on the x-axis to translate to. This value must be a number followed by a unit such as "px", "em", etc.
 *                                  If moving in a horizontal direction, default to element width. Otherwise, default to "0px".
 * @param {string=} options.offsetY The offset on the y-axis to translate to. This value must be a number followed by a unit such as "px", "em", etc.
 *                                  If moving in a vertical direction, default to element height. Otherwise, default to "0px".
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.slideOut = function(element, options)
{
  return oj.AnimationUtils._slide(element, options, false);
};

oj.AnimationUtils._slide = function(element, options, isIn)
{
  options = oj.AnimationUtils._mergeOptions(isIn ? 'slideIn' : 'slideOut', options);

  var fromState = {};
  var toState = {};

  var direction = options['direction'] || "start";

  var offsetX = '0';
  var offsetY = '0';
  var fromCSS = fromState['css'] = {};
  var toStateCSS = toState['css'] = {};

  if (options['offsetX'] || options['offsetY'])
  {
    if (options['offsetX'])
      offsetX = options['offsetX'];
    
    if (options['offsetY'])
      offsetY = options['offsetY'];
  }
  else
  {
    var isRTL = oj.DomUtils.getReadingDirection() === "rtl";

    switch (direction)
    {
      case 'left':
        offsetX = (isIn ? element.offsetWidth : -element.offsetWidth) + 'px';
        break;
      case 'right':
        offsetX = (isIn ? -element.offsetWidth : element.offsetWidth) + 'px';
        break;
      case 'top':
        offsetY = (isIn ? element.offsetHeight : -element.offsetHeight) + 'px';
        break;
      case 'bottom':
        offsetY = (isIn ? -element.offsetHeight : element.offsetHeight) + 'px';
        break;
      case 'end':
        offsetX = ((isIn ? -element.offsetWidth : element.offsetWidth) * (isRTL ? -1 : 1)) + 'px';
        break;
      default: // 'start'
        offsetX = ((isIn ? element.offsetWidth : -element.offsetWidth) * (isRTL ? -1 : 1)) + 'px';
        break;
    }
  }
  
  var transformPropName = 'transform';
  if (isIn)
  {
    fromCSS[transformPropName] = "translate(" + offsetX + "," + offsetY + ") translateZ(0)";
    toStateCSS[transformPropName] = "translate(0,0) translateZ(0)";
  }
  else
  {
    fromCSS[transformPropName] = "translate(0,0) translateZ(0)";
    toStateCSS[transformPropName] = "translate(" + offsetX + "," + offsetY + ") translateZ(0)";
  }

  return oj.AnimationUtils._animate(element, fromState, toState, options, [transformPropName]);
};

/**
 * Animaton effect method for rippling a HTML element.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.offsetX Horizontal offset of the ripple center, with a unit of either "px" or "%".
 *                                  If the unit is "px", it specifies the offset in pixels.
 *                                  If the unit is "%", it specifies the offset as a percentage of the element's width.
 * @param {string=} options.offsetY Vertical offset of the ripple center, with a unit of either "px" or "%".
 *                                  If the unit is "px", it specifies the offset in pixels.
 *                                  If the unit is "%", it specifies the offset as a percentage of the element's height.
 * @param {string=} options.color Color of the ripple. Default is specified in the "oj-animation-effect-ripple" CSS class.
 * @param {string=} options.diameter Diameter of the ripple, with a unit of either "px" or "%".
 *                                   If the unit is "px", it specifies the diameter in pixels.
 *                                   If the unit is "%", it specifies the diameter as a percentage of either the element's width or height, whichever is less.
 *                                   Default is specified in the "oj-animation-effect-ripple" CSS class.
 * @param {number=} options.startOpacity start opacity of the ripple. Default is specified in the "oj-animation-effect-ripple" CSS class.
 * @param {number=} options.endOpacity end opacity of the ripple. Default is 0.
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.ripple = function(element, options)
{
  options = oj.AnimationUtils._mergeOptions('ripple', options);

  var fromState = {};
  var toState = {};

  var width = element.offsetWidth;
  var height = element.offsetHeight;

  // The rippler need its own container since setting overflow on the target
  // element may not work if the element has no explicit height, which can be
  // the case on buton, listitem, etc.
  var container = $("<div style='position:absolute; overflow:hidden'>");
  var rippler = $("<div class='oj-animation-effect-ripple oj-animation-rippler'>");
  
  // prepend the rippler instead of append so that it doesn't obscure other children
  var style = window.getComputedStyle(element);
  var position = (style['position'] == 'static') ? {'left':element.offsetLeft, 'top':element.offsetTop} : {'left':0, 'top':0};
  element.insertBefore(container[0], element.firstChild); //@HTMLUpdateOK; container is constructed by component code and is not using string passed in through any APIs.
  container.css({'left': position['left'] + 'px', 
                 'top': position['top'] + 'px',
                 'width': width + 'px',
                 'height': height + 'px'});
  
  container.prepend(rippler); //@HTMLUpdateOK; rippler is constructed by component code and is not using string passed in through any APIs.
  
  var fromCSS = fromState['css'] = {};
  var toStateCSS = toState['css'] = {};
  var transformPropName = 'transform';

  oj.AnimationUtils._setRippleOptions(fromCSS, rippler, container, options);
  
  fromCSS[transformPropName] = "scale(0) translateZ(0)";
  fromCSS['opacity'] = options["startOpacity"] || rippler.css('opacity');
  
  toStateCSS[transformPropName] = "scale(1) translateZ(0)";
  toStateCSS['opacity'] = options["endOpacity"] || 0;
  
  // Always persist the ripple state so that it remains invisible until removed.
  // Otherwise it may re-appear briefly on mobile Safari.
  options['persist'] = 'all';

  return oj.AnimationUtils._animate(rippler[0], fromState, toState, options, [transformPropName, 'opacity']).
          then(function() 
          {
            container.remove();
          });
};

oj.AnimationUtils._setRippleOptions = function(css, rippler, parent, options)
{
  var diameter = rippler.width();
  var parentWidth = parent.width();
  var parentHeight = parent.height();

  if (options['diameter'])
  {
    var diameterStr = options['diameter'];
    var value = parseInt(diameterStr, 10);
    if (!isNaN(value))
    {
      if (diameterStr.charAt(diameterStr.length - 1) == '%')
      {
        diameter = Math.floor(Math.min(parentWidth, parentHeight) * (value / 100));
      }
      else
      {
        diameter = value;
      }

      css['width'] = diameter + 'px';
      css['height'] = diameter + 'px';
    }
  }

  var position = (parent.css('position') == 'static') ? parent.position() : {'left':0, 'top':0};
  var offset;
  
  offset = oj.AnimationUtils._calcRippleOffset(options['offsetX'], diameter, parentWidth, position['left']);
  if (offset != null)
  {
    css['left'] = offset + 'px';
  }
  
  offset = oj.AnimationUtils._calcRippleOffset(options['offsetY'], diameter, parentHeight, position['top']);
  if (offset != null)
  {
    css['top'] = offset + 'px';
  }
  
  if (options['color'])
  {
    css['backgroundColor'] = options['color'];
  }
};

oj.AnimationUtils._calcRippleOffset = function(offsetOption, diameter, parentSize, parentOffset)
{
  var offset;
  
  offsetOption = offsetOption || '50%';

  var offsetInt = parseInt(offsetOption, 10);
  if (!isNaN(offsetInt))
  {
    if (offsetOption.charAt(offsetOption.length - 1) == '%')
    {
      offset = (parentSize * (offsetInt / 100)) - (diameter / 2);
    }
    else
    {
      offset = offsetInt - (diameter / 2);
    }

    // offset should be relative to the rippler's offsetParent, which is not
    // the parent element if the parent element has static position.      
    offset = Math.floor(offset + parentOffset);
  }
  
  return offset;
};

oj.AnimationUtils._removeRipple = function(element, options)
{
  options = options || {};

  var possibleEffects = {"fadeOut": 1, "collapse": 1, "zoomOut": 1, "slideOut": 1};
  var removeEffect = options["removeEffect"] || "fadeOut";
  var rippler = $(".oj-animation-rippler", element);

  if(rippler.length === 0) 
  {
    console.warn("No rippler so returning");
    return
  }

  if(!(removeEffect in possibleEffects)) 
  {
    return rippler.remove();
  }
  
  return oj.AnimationUtils[removeEffect](rippler, options).
          then(function() 
          {
            rippler.remove();
          });
};

oj.AnimationUtils._calcBackfaceAngle = function(angle)
{
  var backfaceAngle;
  var expr = /^([\+\-]?\d*\.?\d*)(.*)$/;
  var matchArray = angle.match(expr);
  var amount = parseFloat(matchArray[1]);
  var unit = matchArray[2];

  switch (unit)
  {
    case 'deg': backfaceAngle = (amount - 180) + unit; break;
    case 'grad': backfaceAngle = (amount - 200) + unit; break;
    case 'rad': backfaceAngle = (amount - 3.1416) + unit; break;
    case 'turn': backfaceAngle = (amount - 0.5) + unit; break;
    default: console.log('Unknown angle unit in flip animation: ' + unit); break;
  }
  
  return backfaceAngle;
};

oj.AnimationUtils._flip = function(element, options, effect, startAngle, endAngle)
{
  // Handle the case where the element has children to represent front and back
  // faces.  We need to flip the children instead of the parent since IE doesn't
  // support preserve-3d style, which works on other browsers.
  if (options && options['flipTarget'] == 'children')
  {
    var promises = [];
    var children = $(element).children();
    var childOptions;

    var frontOptions = $.extend({}, options);
    delete frontOptions['flipTarget'];

    var backOptions = $.extend({}, frontOptions);
    backOptions['startAngle'] = oj.AnimationUtils._calcBackfaceAngle(options['startAngle'] || startAngle);
    backOptions['endAngle'] = oj.AnimationUtils._calcBackfaceAngle(options['endAngle'] || endAngle);
    
    for (var i = 0; i < children.length; i++)
    {
      childOptions = $(children[i]).hasClass('oj-animation-backface') ? backOptions : frontOptions;
      promises.push(oj.AnimationUtils._flip(children[i], childOptions, effect, startAngle, endAngle));
    }

    return Promise.all(promises);
  }

  options = oj.AnimationUtils._mergeOptions(effect, options);

  var fromCss = {};
  var toCss = {};
  var fromState = {'css': fromCss};
  var toState = {'css': toCss};
  var rotateFunc = 'rotateY(';
  var perspective = '2000px';
  var backfaceVisibility = 'hidden';
  var transformOrigin = 'center';
  var transform;
  
  if (options)
  {    
    if (options['axis'] === 'x')
      rotateFunc = 'rotateX(';
    
    if (options['startAngle'])
      startAngle = options['startAngle'];

    if (options['endAngle'])
      endAngle = options['endAngle'];
   
    if (options['perspective'])
      perspective = options['perspective'];
    
    if (options['backfaceVisibility'])
      backfaceVisibility = options['backfaceVisibility'];
    
    if (options['transformOrigin'])
      transformOrigin = options['transformOrigin'];
  }

  // perspective() must precede rotate() in the transform value in order for it to work
  transform = 'perspective(' + perspective + ') ' + rotateFunc;

  // Safari still requires webkit prefix for backfaceVisibility property
  var backfaceVisPropName = 'backfaceVisibility';
  var transformPropName = 'transform';
  var transformOriginPropName = 'transformOrigin';
  
  fromCss[transformPropName] = transform + startAngle + ')';
  fromCss[backfaceVisPropName] = backfaceVisibility;
  fromCss[transformOriginPropName] = transformOrigin;  

  toCss[transformPropName] = transform + endAngle + ')';

  // backfaceVisibility and transformOrigin affects the final look of the element,
  // so they should be persisted if the persist option is set.
  return oj.AnimationUtils._animate(element, fromState, toState, options, [transformPropName], [transformPropName, backfaceVisPropName, transformOriginPropName]);
};

/**
 * Animaton effect method for rotating a HTML element into view.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.axis  The axis of the rotation. Valid values are "x" and "y". Default is "y".
 * @param {string=} options.startAngle  The starting angle of the rotation. Refer to CSS rotate() transform for valid values. Default is "-180deg", which shows the back face of the element.
 * @param {string=} options.endAngle  The ending angle of the rotation. Refer to CSS rotate() transform for valid values. Default is "0deg", which shows the front face of the element.
 * @param {string=} options.backfaceVisibility  The visibility of the back face when facing the user. Valid values are "visible" and "hidden". If set to "visible", the back face shows a 
 *                                              mirrored image of the front face. If set to "hidden", the back face is invisible.  Default is "hidden".
 * @param {string=} options.perspective  The 3D perspective for the element. Default is "2000px". A smaller value makes the 3D effect more pronounced during rotation.
 * @param {string=} options.transformOrigin  The axis location for the rotation. Refer to CSS transform-origin for valid values. Default is "center".
 * @param {string=} options.flipTarget  The target for flipping.  Valid values are "element" and "children".  Default is "element".
 *                                      <p>Set to "element" to flip the element itself.</p>
 *                                      <p>Set to "children" to flip the children of the element.  This is used when the element is a card-like structure that
 *                                         has children to represent the front and back faces of a card.  The child that represents the back face must have
 *                                         the "oj-animation-backface" marker class.  Use this option instead of the "transform-style: preserve-3d" CSS style because
 *                                         some browsers do not support "transform-style".  See the cookbook for a Card Flip example of using this option.</p>
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.flipIn = function(element, options)
{
  return oj.AnimationUtils._flip(element, options, 'flipIn', '-180deg', '0deg');
};

/**
 * Animaton effect method for rotating a HTML element out of view.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object=} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {string=} options.persist  Valid values are "none" and "all".  Set to "none" to remove the inline style being animated at the end of animation.  
 *                                    Set to "all" to persist the inline style.  Default is "none".
 * @param {string=} options.axis  The axis of the rotation. Valid values are "x" and "y". Default is "y".
 * @param {string=} options.startAngle  The starting angle of the rotation. Refer to CSS rotate() transform for valid values. Default is "0deg", which shows the front face of the element.
 * @param {string=} options.endAngle  The ending angle of the rotation. Refer to CSS rotate() transform for valid values. Default is "180deg", which shows the back face of the element.
 * @param {string=} options.backfaceVisibility  The visibility of the back face when facing the user. Valid values are "visible" and "hidden". If set to "visible", the back face shows a 
 *                                              mirrored image of the front face. If set to "hidden", the back face is invisible.  Default is "hidden".
 * @param {string=} options.perspective  The 3D perspective for the element. Default is "2000px". A smaller value makes the 3D effect more pronounced during rotation.
 * @param {string=} options.transformOrigin  The axis location for the rotation. Refer to CSS transform-origin for valid values. Default is "center".
 * @param {string=} options.flipTarget  The target for flipping.  Valid values are "element" and "children".  Default is "element".
 *                                      <p>Set to "element" to flip the element itself.</p>
 *                                      <p>Set to "children" to flip the children of the element.  This is used when the element is a card-like structure that
 *                                         has children to represent the front and back faces of a card.  The child that represents the back face must have
 *                                         the "oj-animation-backface" marker class.  Use this option instead of the "transform-style: preserve-3d" CSS style because
 *                                         some browsers do not support "transform-style".  See the cookbook for a Card Flip example of using this option.</p>
 * @return {Promise.<boolean>} a promise that will be resolved when the animation ends
 *
 * @export
 * @memberof oj.AnimationUtils
 */
oj.AnimationUtils.flipOut = function(element, options)
{
  return oj.AnimationUtils._flip(element, options, 'flipOut', '0deg', '180deg');
};

/**
 * Animaton effect method for adding transition to a HTML element.  Caller should
 * set the new style immediately before calling this method.  This is for internal
 * use only.
 *
 * @param {Element} element  the HTML element to animate
 * @param {Object} options Options applicable to the specific animation effect.
 * @param {string=} options.delay  The delay from the time the animation is applied to time the 
 * animation should begin. This may be specified in either seconds (by specifying s as the unit) or milliseconds 
 * (by specifying ms as the unit). Default is "0s".
 * @param {string=} options.duration The duration that an animation should take to complete. This may be 
 * specified in either seconds (by specifying s as the unit) or milliseconds (by specifying ms as the unit). 
 * Default is "400ms".
 * @param {string=} options.timingFunction  One of the valid values for either CSS transition-timing-function or CSS 
 * animation-timing-function. Default is "ease".
 * @param {Array} options.transitionProperties  An array of properties to transition.
 * @return {Promise|IThenable} a promise that will be resolved when the animation ends
 *
 * @export
 * @ignore
 */
oj.AnimationUtils.addTransition = function(element, options)
{
  options = oj.AnimationUtils._mergeOptions('addTransition', options);
   
  return oj.AnimationUtils._animate(element, null, null, options, options['transitionProperties']);
};
/**
 * All the available animation methods supported in oj.AnimationUtils
 * @typedef {Object} oj.AnimationUtils.AnimationMethods
 * @ojvalue {string} "collapse"
 * @ojvalue {string} "expand"
 * @ojvalue {string} "fadeIn"
 * @ojvalue {string} "fadeOut"
 * @ojvalue {string} "flipIn"
 * @ojvalue {string} "flipOut"
 * @ojvalue {string} "ripple"
 * @ojvalue {string} "slideIn"
 * @ojvalue {string} "slideOut"
 * @ojvalue {string} "zoomIn"
 * @ojvalue {string} "zoomOut"
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
 /**
 * @class oj.ArrayCellSet
 * @classdesc A CellSet represents a collection of cells.  The CellSet is an object returned by the success callback
 * of the fetchCells method on DataGridDataSource.  This implementation of CellSet is used by the
 * array DataGridDataSource.
 * @param {number} startRow the start row index of the cell set
 * @param {number} endRow the end row index of the cell set
 * @param {number} startColumn the start column index of the cell set
 * @param {number} endColumn the end column index of the cell set
 * @param {Object} callback the callback to invoke on to retrieve data and metadata.
 * @constructor
 * @since 1.0
 * @export
 * @hideconstructor
 * @ojtsignore
 * @see oj.ArrayDataGridDataSource 
 */
oj.ArrayCellSet = function(startRow, endRow, startColumn, endColumn, callback)
{
    this.m_startRow = startRow;
    this.m_endRow = endRow;
    this.m_startColumn = startColumn;
    this.m_endColumn = endColumn;
    this.m_callback = callback;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available
 * 2) the index specified is out of bounds.
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @param {number} indexes.row the index of the row axis.
 * @param {number} indexes.column the index of the column axis.
 * @return {Object} an object with property data with the data for the specified index.
 * @export
 * @expose
 * @memberof oj.ArrayCellSet
 * @method
 * @instance
 * @memberof oj.ArrayCellSet 
 */
oj.ArrayCellSet.prototype.getData = function(indexes)
{
    var self, returnObj, rowIndex, columnIndex;
    self = this;
    rowIndex = indexes['row'];
    columnIndex = indexes['column'];
    returnObj = {};
    Object.defineProperty(returnObj, 'data', {
        enumerable: true,
        get: function(){return self.m_callback._getCellData(rowIndex, columnIndex)},
        set: function (newValue) { self.m_callback._setCellData(rowIndex, columnIndex, newValue)}
    });
    return returnObj;
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available 
 * 2) the index specified is out of bounds. 
 * @param {Object} indexes the index of each axis in which we want to retrieve the metadata from.  
 * @param {number} indexes.row the index of the row axis.
 * @param {number} indexes.column the index of the column axis.
 * @return {Object} the metadata object for the specific index.  The metadata that the DataGrid supports are: 
 *         1) keys - the key (of each axis) of the cell.
 * @export
 * @expose
 * @memberof oj.ArrayCellSet
 * @method
 * @instance
 * @memberof oj.ArrayCellSet 
 */
oj.ArrayCellSet.prototype.getMetadata = function(indexes)
{
    return this.m_callback._getCellMetadata(indexes['row'], indexes['column']);
};

/**
 * Gets the start index of the result set for the specified axis.  Valid values are "row" and "column".
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 * @return {number} the start index of the result set for the specified axis.
 * @export
 * @expose
 * @memberof oj.ArrayCellSet
 * @method
 * @instance
 * @memberof oj.ArrayCellSet 
 */
oj.ArrayCellSet.prototype.getStart = function(axis)
{
    if (axis == "row")
    {
        return this.m_startRow;
    }
    else if (axis == "column")
    {
        return this.m_startColumn;
    }

    return -1;
};

/**
 * Gets the actual count of the result set for the specified axis.  Valid values are "row" and "column".
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 * @return {number} the actual count of the result set for the specified axis.
 * @export
 * @expose
 * @memberof oj.ArrayCellSet
 * @method
 * @instance
 * @memberof oj.ArrayCellSet 
 */
oj.ArrayCellSet.prototype.getCount = function(axis)
{
    if (axis === "row")
    {
        return Math.max(0, this.m_endRow - this.m_startRow);
    }

    if (axis === "column")
    {
        return Math.max(0, this.m_endColumn - this.m_startColumn);
    }

    return 0;
};

/**
 * Gets the extent of a particular row/column index within the context of the cellSet.
 * Extent is defined as the number of indexes along the appropriate axis spanned by the cell.
 * If the extent extends beyond the start and end of the requested cell range the extent should be trimmed to the edge of the requested cell range and the object for {'more': {'before', 'after'}} should have the value appropriate boolean set.
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from. 
 * @param {number} indexes.row the index of the row axis.
 * @param {number} indexes.column the index of the column axis.
 * @return {Object} an object containing two properties row and column. Each of those properties has two sub properties:
 *              extent: the number of absolute indexes spanned by the cell at this index
 *                      bounded by the edges of the result set for the specified axis.
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is available in the cellSet
 * @example <caption>In this example the cell spans 5 row indexes and 2 column indexes and there are more column indexes spanned by the cell that
 *              aren't included in this cellSet:</caption>
 * {
 *  'row': {'extent':5, 'more': {'before':false, 'after':false}},
 *  'column': {'extent':2, 'more': {'before':false, 'after':true}}
 * }
 * @memberof oj.ArrayCellSet
 * @since 4.0.0
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.ArrayCellSet
 * @since 4.0.0
 */
oj.ArrayCellSet.prototype.getExtent = function(indexes)
{
    return {'row': {'extent':1, 'more': {'before':false, 'after':false}},
            'column': {'extent':1, 'more': {'before':false, 'after':false}}};
};

////// testing methods to get properties //////
/**
 * Gets the start row property for testing
 * @return {number} the start row
 * @export
 * @ignore
 * @method
 * @instance
 * @memberof oj.ArrayCellSet
 */
oj.ArrayCellSet.prototype.getStartRow = function()
{
    return this.m_startRow;
};

/**
 * Gets the start column property for testing
 * @return {number} the start column
 * @export
 * @ignore
 * @method
 * @instance
 * @memberof oj.ArrayCellSet
 */
oj.ArrayCellSet.prototype.getStartColumn = function()
{
    return this.m_startColumn;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.ArrayDataGridDataSource
 * @classdesc An array based implementation of the DataGridDataSource.
 * @param {Array|Object} data the data in the form of array or observable array.
 * @param {Object=} options the options specific to this DataGridDataSource.
 * @property {Array<*>=} options.columns an array of columns to return as column headers.
 * @property {*=} options.rowHeader an object representing the default row header,
 *      if a string is provided it will be used as the key of the header, if a number is provided
 *      it will be used as the index of the header.
 * @property {string=} options.rowHeader.default disable default row headers or provide them as indexes 'none' or 'index'
 * @property {Object=} options.initialSort the information about the array if it is pre-sorted
 * @property {string=} options.initialSort.axis the axis that the array is sorted on valid values are 'column', 'row'
 * @property {*=} options.initialSort.key the key that the array is sorted on
 * @property {string=} options.initialSort.direction valid values are ascending or descending
 * @export
 * @constructor
 * @since 1.0
 * @extends oj.DataGridDataSource
 * @ojtsignore
 */
oj.ArrayDataGridDataSource = function(data, options)
{
    var errSummary, errDetail;
    if (!(data instanceof Array) &&
            (typeof (data) != 'function' &&
                    typeof (data.subscribe) != 'function'))
    {
        // we only support Array or ko.observableArray. To
        // check for observableArray, we can't do instanceof check because it's
        // a function. So we just check if it contains a subscribe function.
        errSummary = '_ERR_DATA_INVALID_TYPE_SUMMARY';
        errDetail = '_ERR_DATA_INVALID_TYPE_DETAIL';
        throw new Error(errSummary + '\n' + errDetail);
    }

    this.rowHeaderKey = this._getRowHeaderFromOptions(options);

    if (options != null)
    {
        // undefined if no row header, 'm_defaultIndex' if indexed, other strings keys, numbers index of array
        this.columns = options['columns'];
        this['sortCriteria'] = options['initialSort'];
    }
    oj.ArrayDataGridDataSource.superclass.constructor.call(this, data);
};

// Subclass from oj.DataGridDataSource
oj.Object.createSubclass(oj.ArrayDataGridDataSource, oj.DataGridDataSource, "oj.ArrayDataGridDataSource");

/**
 * Initial the array based data source.
 * @memberof oj.ArrayDataGridDataSource
 * @instance
 * @override
 * @protected
 */
oj.ArrayDataGridDataSource.prototype.Init = function()
{
    // suck out the column definition from data
    if (this.columns == null)
    {
        this.columns = this._getColumnsForScaffolding(this.getDataArray());
    }
    this._initializeRowKeys();

    //if the data is an observable array subscribe to array change notifications
    if (typeof (this.data) == 'function')
    {
        this.data['subscribe'](this._subscribe.bind(this), null, 'arrayChange');
    }

    // call super
    oj.ArrayDataGridDataSource.superclass.Init.call(this);
};

/**
 * @export
 * @memberof oj.ArrayDataGridDataSource
 * @type {Function}
 * @desc If set to a function(row1, row2), then this function is called comparing raw row data (see the
 * JavaScript array.sort() for details)
 */
oj.ArrayDataGridDataSource.prototype.comparator = null;

/**
 * @export
 * 
 * @type {Object}
 * @property {string} axis the sort axis valid values are "column", "row"
 * @property {string|number} key The key that identifies which field to sort
 * @property {string} direction the sort direction, valid values are "ascending", "descending", "none" (default)
 *
 * @desc The sort criteria. Whenever sort() is called with the criteria parameter, that value is copied to this
 * property. If sort() is called with empty sort criteria then the criteria set in this property is used.
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype.sortCriteria = null;

/**
 * Extract the row header from the options
 * @param {Object|null=} options the options passed into the data source
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getRowHeaderFromOptions = function(options)
{
    if (options != null && options['rowHeader'] != null)
    {
        var option = options['rowHeader'];
        if (typeof option === 'object')
        {
            if (option['default'] != null)
            {
                if (option['default'] == 'none')
                {
                    return undefined;
                }
                else if (option['default'] == 'index')
                {
                    return 'm_defaultIndex';
                }
            }
        }
        else if (option != null)
        {
            return option;
        }
    }

    // nothing set means indexed headers
    return 'm_defaultIndex';
};

/**
 * Initialize the generated row keys.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._initializeRowKeys = function()
{
    var data;
    data = this.getDataArray();
    for (this.lastKey = 0; this.lastKey < data.length; this.lastKey += 1)
    {
        //inject the row key into the object
        data[this.lastKey]['ojKey'] = this.lastKey.toString();
    }
};

/**
 * Get the column headers from the data, if it is an array of arrays with no row header key set,
 * gets the column number as the column header.
 * @param {Object} data the data to extract the column information.
 * @return {Array} the columns extracted from the data.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getColumnsForScaffolding = function(data)
{
    var propertyName, columns;
    if ((typeof data.length !== 'number') || data.length === 0)
    {
        return [];
    }

    columns = [];
    for (propertyName in data[0])
    {
        if (data[0].hasOwnProperty(propertyName))
        {
            if (!(this.rowHeaderKey != undefined && propertyName == this.rowHeaderKey))
            {
                columns.push(propertyName);
            }
        }
    }

    return columns;
};

/**
 * Returns the total number of rows or columns.  If the value return is not >= 0 then it is automatically assumed
 * that the total count is unknown.
 * @param {string} axis the axis in which we inquire for the total count.  Valid values are "row" and "column".
 * @return {number} the total number of rows/columns.
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource 
 */
oj.ArrayDataGridDataSource.prototype.getCount = function(axis)
{
    if (axis === "row")
    {
        return this._size();
    }

    if (axis === "column")
    {
        return this.columns.length;
    }

    return 0;
};

/**
 * Returns whether the total count returned in getCount function is an actual or an estimate.
 * @param {string} axis the axis in which we inquire whether the total count is an estimate.  Valid values are 
 *        "row" and "column".
 * @return {string} "exact" if the count returned in getCount function is the actual count, "estimate" if the
 *         count returned in getCount function is an estimate. The default value is "exact".
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource 
 */
oj.ArrayDataGridDataSource.prototype.getCountPrecision = function(axis)
{
    return "exact";
};

/**
 * Retrieve the data for the header of a specified index.
 * @param {string} axis the axis of the header.  Valid values are "row" and "column".
 * @param {number} index the index in which to get the data.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getHeaderData = function(axis, index)
{
    var data;
    if (axis === 'row')
    {
        if (this.rowHeaderKey == undefined)
        {
            return null;
        }
        else if (this.rowHeaderKey == 'm_defaultIndex')
        {
            // generate data by index
            return this._getRowKeyByIndex(index);
        }
        else
        {
            data = this.getDataArray();
            return data[index][this.rowHeaderKey];
        }
    }
    else if (axis === 'column')
    {
        return this.columns[index];
    }
};

/**
 * Retrieve the metadata for the header of a specified index.
 * @param {string} axis the axis of the header.  Valid values are "row" and "column".
 * @param {number} index the index in which to get the metadata.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getHeaderMetadata = function(axis, index)
{
    var key;
    if (axis === 'row')
    {
        return {'key': this._getRowKeyByIndex(index)};
    }
    else if (axis === 'column')
    {
        key = this._getHeaderData(axis, index);
        if (this['sortCriteria'] != null && this['sortCriteria']['key'] === key)
        {
            return {'key': this._getHeaderData(axis, index), 'sortDirection': this['sortCriteria']['direction']};
        }

        return {'key': key};
    }
};

/**
 * Fetch a range of headers from the data source.
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count.
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column".
 * @property {number} headerRange.start the start index of the range in which the header data are fetched.
 * @property {number} headerRange.count the size of the range in which the header data are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch headers operation is completed.
 * @property {function({startHeaderSet: Object}, {headerRange: Object}, {endHeaderSet: Object}):undefined} callbacks.success the callback to invoke when fetch headers completed successfully.
 *        The function takes three parameters: HeaderSet object representing start headers, headerRange object passed into the original fetchHeaders call,
 *        and a HeaderSet object representing the end headers along the axis.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource 
 * @return {undefined}
 */
oj.ArrayDataGridDataSource.prototype.fetchHeaders = function(headerRange, callbacks, callbackObjects)
{
    var axis, start, count, end, headerSet, data;

    axis = headerRange.axis;
    start = headerRange.start;
    count = headerRange.count;

    start = Math.max(0, start);
    if (axis === "column")
    {
        end = Math.min(this.columns.length, start + count);
    }
    else
    {
        data = this.getDataArray();
        // check if no row header is available
        if (this.rowHeaderKey === undefined)
        {
            // header count = 0
            end = start;
        }
        else
        {
            end = Math.min(data.length, start + count);
        }
    }
    headerSet = new oj.ArrayHeaderSet(start, end, axis, this);

    if (callbacks != null && callbacks['success'] != null)
    {
        // make sure callbackObjects is not null
        if (callbackObjects == null)
        {
            callbackObjects = {};
        }
        callbacks['success'].call(callbackObjects['success'], headerSet, headerRange, null);
    }
};

/**
 * Retrieve the data for the cell of a specified indexes.
 * @param {number} row the row index in which to get the data.
 * @param {number} column the column index in which to get the data.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getCellData = function(row, column)
{
    var col = this.columns[column];
    return this.getDataArray()[row][col];
};

/**
 * Retrieve the data for the cell of a specified indexes.
 * @param {number} row the row index in which to get the data.
 * @param {number} column the column index in which to get the data.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._setCellData = function(row, column, newValue)
{
    this.getDataArray()[row][this.columns[column]] = newValue;
};

/**
 * Retrieve the metadata for the cell of a specified indexes.
 * @param {number} row the row index in which to get the data.
 * @param {number} column the column index in which to get the data.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getCellMetadata = function(row, column)
{
    var keys = {"row": this._getRowKeyByIndex(row), "column": this.columns[column]};
    return {"keys": keys};
};

/**
 * Fetch a range of cells from the data source.
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched. 
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.
 * @property {function({cellSet: Object}, {cellRanges: Array.<Object>}):undefined} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype.fetchCells = function(cellRanges, callbacks, callbackObjects)
{
    var i, cellRange, rowStart, rowEnd, cellSet, colStart, colEnd;

    // extract the start and end row/column info from cellRanges (there should only be two, one for each axis)
    for (i = 0; i < cellRanges.length; i += 1)
    {
        cellRange = cellRanges[i];
        if (cellRange['axis'] === "row")
        {
            rowStart = cellRange['start'];
            rowEnd = Math.min(this._size(), rowStart + cellRange['count']);
        }
        else if (cellRange['axis'] === "column")
        {
            colStart = cellRange['start'];
            colEnd = Math.min(this.columns.length, colStart + cellRange['count']);
        }
    }

    // check for errors
    if (rowEnd === undefined || colEnd === undefined)
    {
        if (callbacks != null && callbacks['error'] != null)
        {
            // make sure callbackObjects is not null
            if (callbackObjects == null)
            {
                callbackObjects = {};
            }
            callbacks['error'].call(callbackObjects['error']);
        }
        return;
    }

    cellSet = new oj.ArrayCellSet(rowStart, rowEnd, colStart, colEnd, this);

    if (callbacks != null && callbacks['success'] != null)
    {
        // make sure callbackObjects is not null
        if (callbackObjects == null)
        {
            callbackObjects = {};
        }
        callbacks['success'].call(callbackObjects['success'], cellSet, cellRanges);
    }
};

/**
 * Returns the keys based on the indexes.
 * @param {Object} indexes the index for each axis
 * @property {number|null} indexes.row the index for the row axis
 * @property {number|null} indexes.column the index for the column axis
 * @return {Promise.<Object>} a Promise object which upon resolution will pass in an object containing the keys for each axis
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource 
 */
oj.ArrayDataGridDataSource.prototype.keys = function(indexes)
{
    var rowIndex = indexes['row'], columnIndex = indexes['column'];
    return new Promise(function(resolve, reject) {
        resolve({"row": this._getRowKeyByIndex(rowIndex), "column": this.columns[columnIndex]});
    }.bind(this));
};

/**
 * Returns the row and column index based on the keys. In a paging case returns the 
 * index on the page, not the absolute index in the array.
 * @param {Object} keys the key for each axis
 * @property {*} keys.row the key for the row axis
 * @property {*} keys.column the key for the column axis
 * @return {Promise.<Object>} a promise object containing the index for each axis
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource 
 */
oj.ArrayDataGridDataSource.prototype.indexes = function(keys)
{
    var rowKey = keys['row'], columnKey = keys['column'];
    return new Promise(function(resolve, reject) {
        resolve({"row": this._getRowIndexByKey(rowKey), "column": this.columns.indexOf(columnKey)});
    }.bind(this));
};

/**
 * Performs a sort on the data source.
 * @param {Object|null} criteria the sort criteria. 
 * @property {string} criteria.axis The axis in which the sort is performed, valid values are "row", "column"
 * @property {*} criteria.key The key that identifies which header to sort
 * @property {string} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the sort operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.  
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource
 * @return {undefined}
 */
oj.ArrayDataGridDataSource.prototype.sort = function(criteria, callbacks, callbackObjects)
{
    var sortArray = [], newColumns = [], i, headerIndex, axis, headerKey, direction;

    // make sure callbackObjects is non null
    if (callbacks != null && callbackObjects == null)
    {
        callbackObjects = {};
    }

    // keep a copy of the original unsorted array.  Both array and observable array have slice method.
    if (this.origData == undefined)
    {
        this._origSortCriteria = this['sortCriteria'];
        this.origData = this.data.slice();
    }

    if (criteria == null)
    {
      criteria = this['sortCriteria'];
    }
    else
    {
        this['sortCriteria'] = criteria;
    }

    // reset sort order if no criteria is set in the call or as a property
    if (criteria == null)
    {
        this._resetSortOrder(callbacks, callbackObjects);
        return;
    }

    axis = criteria['axis'];
    headerKey = criteria['key'];
    direction = criteria['direction'];

    if (axis === 'column')
    {
        this.getDataArray().sort(this._getComparator());

        if (callbacks != null && callbacks['success'] != null)
        {
            callbacks['success'].call(callbackObjects['success']);
        }
    }
    else if (axis === 'row')
    {
        headerIndex = this._getRowIndexByKey(headerKey);
        //rebuild the array to sort on
        for (i = 0; i < this.columns.length; i += 1)
        {
            sortArray[i] = this.getDataArray()[headerIndex][this.columns[i]];
        }

        //sort the given array with no headerKey specified
        sortArray.sort(this._getComparator());

        //reorder the columns property
        for (i = 0; i < this.columns.length; i += 1)
        {
            newColumns[i] = this.columns[sortArray.indexOf(this.getDataArray()[headerIndex][this.columns[i]])];
        }

        // keep a copy of the original column order.
        this.origColumns = this.columns;
        this.columns = newColumns;
        if (callbacks != null && callbacks['success'] != null)
        {
            callbacks['success'].call(callbackObjects['success']);
        }
    }
    else
    {
        if (callbacks !== null && callbacks['error'] != null)
        {
            callbacks['error'].call(callbackObjects['error'], "Invalid axis value");
        }
    }
};

/**
 * Reset the sort order of the data.
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the sort operation.  The callback
 *        properties are "success" and "error".
 * @property {function()} callbacks.success the callback to invoke when the sort completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.
 *        You can specify the callback object for each callbacks using the "success" and "error" properties.
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._resetSortOrder = function(callbacks, callbackObjects)
{
    // reset data to the unsorted array
    if (this.origData != null)
    {
        this.data = this.origData;
        this['sortCriteria'] = this._origSortCriteria;
    }

    // reset column order if row header was sorted before
    if (this.origColumns != null)
    {
        this.columns = this.origColumns;
    }

    if (callbacks != null && callbacks['success'] != null)
    {
        callbacks['success'].call(callbackObjects['success']);
    }
};

/**
 * Determines whether this ArrayDataGridDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort" and "move"
 * @return {string|null} the name of the feature.  For sort, the valid return values are: "full", "none".  Returns null if the
 *         feature is not recognized.
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource
*/
oj.ArrayDataGridDataSource.prototype.getCapability = function(feature)
{
    if (feature === 'sort')
    {
        // array based data source supports column sorting only
        return 'column';
    }
    if (feature === 'move')
    {
        return 'row';
    }
    return null;
};

/**
 * Get the sort comparator either from the property or the internal one in natural sort
 * @returns {function(Object, Object)|undefined} a comparator function, dependent on direction
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getComparator = function()
{
    var comparator, key, direction, axis;
    comparator = this['comparator'];
    if (comparator == null)
    {
      key = this['sortCriteria']['key'];
      direction = this['sortCriteria']['direction'];
      axis = this['sortCriteria']['axis'];
      return this._naturalSort(direction, key, axis);
    }
    return comparator;
};

/**
 * Get a comparator function for natural sorting of objects
 * @param {string} direction ascending, descending
 * @param {string|number} key the key or index to perform the sort on
 * @param {string} axis
 * @returns {function(Object, Object)|undefined} a comparator function, dependent on direction
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._naturalSort = function(direction, key, axis)
{
    if (direction === 'ascending')
    {
        return function(a, b)
        {
            var as, bs;
            //Get the values the array we're sorting
            if (key != undefined && axis == 'column')
            {
                //if the sorting item is an array it will be indexed with strings of ints and needs
                //to be accessed using ints not strings
                if (a instanceof Array)
                {
                    a = a[parseInt(key, 10)];
                    b = b[parseInt(key, 10)];
                }
                else
                {
                    a = a[key];
                    b = b[key];
                }
            }
            //Strings of numbers return false, so we can compare strings of numbers with numbers
            as = isNaN(a);
            bs = isNaN(b);
            //If they are strings, check to see if they are dates, if they are, turn the string to a sortable date formatted string
            if (a instanceof Date) {
                a = a.toISOString();
                as = true;
            }
            if (b instanceof Date) {
                b = b.toISOString();
                bs = true;
            }
            //both are string
            if (as && bs)
            {
                return a < b ? -1 : a === b ? 0 : 1;
            }
            //only a is a string
            if (as)
            {
                return 1;
            }
            //only b is a string
            if (bs)
            {
                return -1;
            }
            //both are numbers
            return a - b;
        };
    }
    if (direction === 'descending')
    {
        return function(a, b)
        {
            var as, bs;
            if (key != undefined && axis == 'column')
            {
                //if the sorting item is an array it will be indexed with strings of ints and needs
                //to be accessed using ints not strings
                if (a instanceof Array)
                {
                    a = a[parseInt(key, 10)];
                    b = b[parseInt(key, 10)];
                }
                else
                {
                    a = a[key];
                    b = b[key];
                }
            }
            as = isNaN(a);
            bs = isNaN(b);
            if (a instanceof Date) {
                a = a.toISOString();
                as = true;
            }
            if (b instanceof Date) {
                b = b.toISOString();
                bs = true;
            }
            if (as && bs)
            {
                return a > b ? -1 : a === b ? 0 : 1;
            }
            if (as)
            {
                return -1;
            }
            if (bs)
            {
                return 1;
            }
            return b - a;
        };
    }

    // only if direction is not recognized
    return;
};

/**
 * Moves a row from one location to another.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {string} position The position of the moved row relative to the reference row.
 *        Valid values are: "before", "after" 
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the move operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource
 * @return {undefined}
 */
oj.ArrayDataGridDataSource.prototype.move = function(rowToMove, referenceRow, position, callbacks, callbackObjects)
{
    var moveKeyIndex, moveData, atKeyIndex, event, data;

    //remove the data from the array, but hold on to it
    moveKeyIndex = this._getRowIndexByKey(rowToMove);
    moveData = this.data.splice(moveKeyIndex, 1)[0];

    //fire the delete event to the datagrid
    if (this.data instanceof Array)
    {
        event = this._getModelEvent('delete', rowToMove, null, moveKeyIndex, -1, true);
        this.handleEvent("change", event);
    }

    //add the stored data back into the array
    if (referenceRow === null)
    {
        this.data.push(moveData);
        atKeyIndex = this.data.length - 1;
    }
    else
    {
        atKeyIndex = this._getRowIndexByKey(referenceRow);
        this.data.splice(atKeyIndex, 0, moveData);
    }

    //fire the insert event to the datagrid
    if (this.data instanceof Array)
    {
        event = this._getModelEvent('insert', rowToMove, null, atKeyIndex, -1);
        this.handleEvent("change", event);
    }

    // if we keep track of original data, we'll need to update it
    if (this.origData != null)
    {
       // note that once a row is moved then the current sort order is the new unsorted order
       this.origData = this.data.slice();
    }
};

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine
 *        the destination of where the row should moved to.
 * @param {string} position The position of the moved row relative to the reference row.  
 *        Valid values are: "before", "after".
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource 
*/
oj.ArrayDataGridDataSource.prototype.moveOK = function(rowToMove, referenceRow, position)
{
    return "valid";
};

/**
 * Gets the data array, if the data property is a function call it, else return data
 * @return {Object|Array} the array of the data
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype.getDataArray = function()
{
    if (typeof (this.data) === 'function')
    {
        return this.data();
    }
    return this.data;
};

/**
 * Gets the row index of a given row key
 * @param {*} key the key to get row index of
 * @return {number} the index with a certain key, -1 if the key doesn't exist
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getRowIndexByKey = function(key)
{
    var i, data = this.getDataArray();
    for (i = 0; i < data.length; i++)
    {
        if (data[i]['ojKey'] === key)
        {
            return i;
        }
    }
    return -1;
};

/**
 * Gets the row key stored at a given index
 * @param {number} index the index to get row key of
 * @return {string|number|null} the key at index, null if the index doesn't exist
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getRowKeyByIndex = function(index)
{
    var data = this.getDataArray();
    if (data[index])
    {
        return data[index]['ojKey'];
    }
    return null;
};

/**
 * Returns an Object for an event
 * @param {string} operation the operation done on the model
 * @param {*} rowKey the key for the row axis
 * @param {*} columnKey the key for the column axis
 * @param {number=} rowIndex the index for the row axis
 * @param {number=} columnIndex the index for the column axis
 * @param {boolean=} silent should the event be silent
 * @return {Object} an object containing the the source, operation, and keys of the event
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._getModelEvent = function(operation, rowKey, columnKey, rowIndex, columnIndex, silent)
{
    var event = {};
    event['source'] = this;
    event['operation'] = operation;
    event['keys'] = {'row': rowKey, 'column': columnKey};
    event['indexes'] = {'row': rowIndex, 'column': columnIndex};
    event['silent'] = silent;
    return event;
};

/**
 * Subscribe to knockout events
 * @param {Array} changes an array of change objects fired by an observable array
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._subscribe = function(changes)
{
    var i, rowData, rowKey, rowIndex, added = false, move = false, keys = [], indexes = [], event, beforeDelCount = 0, change;

    // first loop though the changes,
    for (i = 0; i < changes.length; i++)
    {
        change = changes[i];
        // if a model was moved using a reverseAll or a sort, just refresh the grid
        if (change['moved'] !== undefined)
        {
            move = true;
            event = this._getModelEvent('refresh', null, null);
            this.handleEvent("change", event);
            break;
        }

        // check if there were any adds, this way the delete will know to be fired silently
        if (change['status'] === 'added')
        {
            added = true;
        }
    }

    //if we moved a model we just refreshed
    if (!move)
    {
        //loop through changes looking for deletes
        for (i = 0; i < changes.length; i++)
        {
            change = changes[i];
            if (change['status'] === 'deleted')
            {
                rowData = change['value'];
                rowIndex = change['index'];
                rowKey = rowData['ojKey'];

                //collect the deletes to do in one batch delete
                keys.push({'row': rowKey, 'column': -1});
                indexes.push({'row': rowIndex, 'column': -1});
            }
        }

        // batch delete all deletes
        if (keys.length > 0)
        {
            event = {'source': this, 'operation': 'delete', 'keys': keys, 'indexes': indexes, 'silent': added};
            this.handleEvent("change", event);
        }

        //loop through changes looking for adds
        for (i = 0; i < changes.length; i++)
        {
            change = changes[i];
            if (change['status'] === 'added')
            {
                rowData = change['value'];
                rowIndex = change['index'];
                //if no key add inject one into the add object based on the last assigned key
                if (rowData['ojKey'] == null)
                {
                    rowData['ojKey'] = this.lastKey.toString();
                    this.lastKey++;
                }
                //add at the given index and remove from the end of the page silently
                rowKey = rowData['ojKey'];
                event = this._getModelEvent('insert', rowKey, null, rowIndex, -1);
                this.handleEvent("change", event);
            }
        }
    }

    // if we keep track of original data, we'll need to update it
    if (this.origData != null)
    {
       // note that once the observable array is updated then the current sort order is the new unsorted order
       this.origData = this.data.slice();
    }
};

/**
 * Get the length of the collection. -1 if an initial fetch has not been
 * done yet. Default to the size of the collection. If pageSize is set then
 * limit it.
 * @returns {number} length of the collection
 * @private
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype._size = function()
{
    return this.getDataArray()['length'];
};

//////// testing methods to get properties /////////
/**
 * Gets the rowHeaderKey property.  This is an internal method for testing and should not be used by application.
 * @return {string|null} the row header key
 * @export
 * @ignore
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype.getRowHeaderKey = function()
{
    return this.rowHeaderKey;
};

/**
 * Gets the columns property.  This is an internal method for testing and should not be used by application.
 * @return {Array|null} the keys of the column headers
 * @export
 * @ignore
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype.getColumns = function()
{
    return this.columns;
};

/**
 * Gets the data property.  This is an internal method for testing and should not be used by application.
 * @return {Array|Object|null} the underlying array data.
 * @export
 * @ignore
 * @memberof oj.ArrayDataGridDataSource
 */
oj.ArrayDataGridDataSource.prototype.getData = function()
{
    return this.data;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.ArrayHeaderSet
 * @classdesc A HeaderSet represents a collection of headers.  The HeaderSet is an object returned by the success callback
 * of the fetchHeaders method on DataGridDataSource.  This implementation of HeaderSet is used by the
 * ArrayDataGridDataSource.
 * @param {number} start the absolute start index of the header set.
 * @param {number} end the absolute end index of the header set.
 * @param {string} axis the axis of the header, value is either 'row' or 'column'.
 * @param {Object} callback the callback to invoke on to retrieve data and metadata.
 * @constructor
 * @since 1.0
 * @export
 * @hideconstructor
 * @ojtsignore
 * @see oj.ArrayDataGridDataSource 
 */
oj.ArrayHeaderSet = function(start, end, axis, callback)
{
    this.m_start = start;
    this.m_end = end;
    this.m_axis = axis;
    this.m_callback = callback;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds.
 * @param {number} index the absolute index of the header in which we want to retrieve the header from.
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {*} the data for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.ArrayHeaderSet 
 */
oj.ArrayHeaderSet.prototype.getData = function(index, level)
{
    if (this.m_callback == null)
    {
        return null;
    }

    // make sure index/level are valid
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds');
    oj.Assert.assert(level == null || level == 0, 'level out of bounds');

    return this.m_callback._getHeaderData(this.m_axis, index);
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds.
 * The metadata that the data source can optionally return are:
 *  1) sortDirection - the initial sort direction of the header.  Valid values are "ascending" and "descending".
 *  2) key - the key of the row/column header.
 * @param {number} index the absolute index of the header in which we want to retrieve the metadata from.
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} the metadata object for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.ArrayHeaderSet 
 */
oj.ArrayHeaderSet.prototype.getMetadata = function(index, level)
{
    if (this.m_callback == null)
    {
        return null;
    }

    // make sure index/level are valid
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds');
    oj.Assert.assert(level == null || level == 0, 'level out of bounds');

    return this.m_callback._getHeaderMetadata(this.m_axis, index);
};

/**
 * Gets the actual number of levels of the result set for the specified axis. The levels
 * are the counted from the outermost header indexed at 0, and moving inwards toward the
 * databody would increment the level by 1. The Array case only supports level count of 1.
 * @return {number} the number of levels of the result set
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.ArrayHeaderSet 
 */
oj.ArrayHeaderSet.prototype.getLevelCount = function()
{
    if (this.getCount() > 0)
    {
        return 1;
    }
    return 0;
};

/**
 * Gets the extent of an index on a particular level within the context of the headerSet. Extent is defined as the
 * number of indexes covered by the header. If the extent extends beyond the start and end of the requested
 * range the extent should be trimmed to the edge of the requested range and the object for {'more': {before, after}}
 * should have the value appropriate boolean set. For innermost headers the extent will always be 1.
 * @param {number} index the absolute index along the innermost header of the extent to get, 0 is the first header in the data source
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} an object containing two values
 *              extent: the number of absolute indexes spanned by the header at this index
 *                      bounded by the edges of the result set for the specified axis.
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is in the headerSet
 * @example <caption>In this example the header spans 5 indexes and there are more indexes to cover after the request that
 *              aren't included in this headerSet:</caption>
 * {'extent':5, 'more': {'before':false, 'after':true}}
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.ArrayHeaderSet 
 */
oj.ArrayHeaderSet.prototype.getExtent = function(index, level)
{
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds');
    oj.Assert.assert(level == null || level == 0, 'level out of bounds');
    return {'extent': 1, 'more':{'before': false, 'after':false}};
};

/**
 * Gets the depth of an index starting at a particular level. The depth is the number
 * of levels spanned by the header. The Array case only supports depth of 1.
 * @param {number} index the absolute index of the depth to get
 * @param {number=} level the level of the header, 0 is the outermost header
 * @return {number} the number of levels spanned by the header at the specified position
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.ArrayHeaderSet 
 */
oj.ArrayHeaderSet.prototype.getDepth = function(index, level)
{
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds');
    oj.Assert.assert(level == null || level == 0, 'level out of bounds');
    return 1;
};

/**
 * Gets the actual count of the result set, the total indexes spanned by the headerSet
 * along the innermost header.
 * @return {number} the actual count of the result set.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.ArrayHeaderSet 
 */
oj.ArrayHeaderSet.prototype.getCount = function()
{
    if (this.m_callback == null)
    {
        return 0;
    }

    return Math.max(0, this.m_end - this.m_start);
};

/**
 * Gets the start index of the result set.
 * @return {number} the start index of the result set.
 * @export
 * @ignore
 * @memberof oj.ArrayHeaderSet
 * @method
 * @instance
 */
oj.ArrayHeaderSet.prototype.getStart = function()
{
    return this.m_start;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * Implementation of PagingModel backed by an array of data.  ArrayPagingDataSource
 * provides a window into the array in both standard JavaScript array and Knockout observable array formats.
 * It is designed to feed a Knockout-bound HTML control, for example.
 * @export
 * @class oj.ArrayPagingDataSource
 * @since 1.0
 * @classdesc Implementation of PagingModel using array data
 * @extends oj.DataSource
 * @implements oj.PagingModel
 * @param {Array.<Object>} data
 * @constructor
 * @ojtsignore
 * @ojtsimport knockout
 */
oj.ArrayPagingDataSource = function(data)
{
  this.data = data;
  // Start at the beginning
  this.current = 0;
  this.Init(); 
  // Default the page size to 10
  this._setPageSize(10);
};

// Subclass from oj.DataSource 
oj.Object.createSubclass(oj.ArrayPagingDataSource, oj.DataSource, "oj.ArrayPagingDataSource");

/**
 * Initializes the instance.
 * @export
 * @memberof oj.ArrayPagingDataSource
 * @ojtsignore
 */
oj.ArrayPagingDataSource.prototype.Init = function()
{
  oj.ArrayPagingDataSource.superclass.Init.call(this);
};


oj.ArrayPagingDataSource.prototype._getSize = function() {
    if (!this._hasMore()) {
        // Return a size representing only what's left if we'd go over the bounds
        return this.totalSize() - this.current;
    }
    // Otherwise window size should match the page size
    return this.currentPageSize;
};

oj.ArrayPagingDataSource.prototype._refreshDataWindow = function(options) {
    options = options || {};
    // Reinit the array
    this.dataWindow = new Array(this._getSize());
    // Copy from the source data
    for (var i = 0; i < this.dataWindow.length; i++) {
        this.dataWindow[i] = this.data[this.current + i];
    }
    
    // Update the observable array
    this._refreshObservableDataWindow();
    this._endIndex = this._startIndex + this.dataWindow.length - 1;
    
    if (!options['silent'])
    {
      this.handleEvent("sync", {'data': this.dataWindow, 'startIndex': this.current});    
    }
};

oj.ArrayPagingDataSource.prototype._refreshObservableDataWindow = function() {
    if (this.observableDataWindow !== undefined) {
        // Manage the observable window array
        this.observableDataWindow.removeAll();
        for (var i = 0; i < this.dataWindow.length; i++) {
            this.observableDataWindow.push(this.dataWindow[i]);
        }
    }
};

    
oj.ArrayPagingDataSource.prototype.handleEvent = function(eventType, event)
{
    return oj.ArrayPagingDataSource.superclass.handleEvent.call(this, eventType, event);
};

/**
 * @export
 * Return the current set of data in the paging window
 * 
 * @returns {Array.<Object>} the current set of data in the paging window
 * @memberof oj.ArrayPagingDataSource
 */
oj.ArrayPagingDataSource.prototype.getWindow = function() {
    return this.dataWindow;
};

/**
 * @export
 * Get the observable array representing the current set of data in the paging window
 * 
 * @returns {Object} an observable array representing the current data in the paging window
 * @ojsignature {target: "Type",
 *               value: "KnockoutObservableArray<object>",
 *               for: "returns"}
 * @memberof oj.ArrayPagingDataSource
 */
oj.ArrayPagingDataSource.prototype.getWindowObservable = function() {
    if (this.observableDataWindow === undefined) {
        this.observableDataWindow = ko.observableArray();
        this._refreshObservableDataWindow();
    }
    return this.observableDataWindow;
};

/**
 * Get the current page
 * @return {number} The current page
 * @export
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance
 */
oj.ArrayPagingDataSource.prototype.getPage = function()
{
  return this._page;
};

/**
 * Set the current page
 * @param {number} value The current page
 * @param {Object=} options Options<p>
 * @param {number} [options.pageSize] The page size.<p>
 * @return {Promise.<null>} promise object triggering done when complete..
 * @export
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance
 */
oj.ArrayPagingDataSource.prototype.setPage = function(value, options)
{
  options = options || {};
  value = parseInt(value, 10);
  try 
  {
    oj.ArrayPagingDataSource.superclass.handleEvent.call(this, oj.PagingModel.EventType['BEFOREPAGE'], {'page' : value, 'previousPage' : this._page});
  }
  catch (err)
  {
    return Promise.reject(null);
  }
  this.pageSize = options['pageSize'] != null ? options['pageSize'] : this.pageSize;
  options['startIndex'] = value * this.pageSize;
  var previousPage = this._page;
  this._page = value;
  this._startIndex = options['startIndex'];
  var self = this;
  
  return new Promise(function(resolve, reject)
  {
    self._fetchInternal(options).then(function(result)
    {
      oj.ArrayPagingDataSource.superclass.handleEvent.call(self, oj.PagingModel.EventType['PAGE'], {'page' : self._page, 'previousPage' : previousPage});
      resolve(null);
    }, function (error)
    {
      // restore old page
      self._page = previousPage;
      self._startIndex = self._page * self.pageSize;
      reject(null);  
    });
  });
};

/**
 * Get the current page start index
 * @return {number} The current page start index
 * @export
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance
 */
oj.ArrayPagingDataSource.prototype.getStartItemIndex = function()
{
  return this._startIndex;
};

/**
 * Get the current page end index
 * @return {number} The current page end index
 * @export
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance
 */
oj.ArrayPagingDataSource.prototype.getEndItemIndex = function()
{
  return this._endIndex;
};

/**
 * Get the page count
 * @return {number} The total number of pages
 * @export
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance
 */
oj.ArrayPagingDataSource.prototype.getPageCount = function()
{
  var totalSize = this.totalSize();
  return totalSize == -1 ? -1 : Math.ceil(totalSize / this.pageSize);
};

/**
 * Fetch the row data.
 * @param {Object=} options Options to control fetch
 * @param {number} [options.startIndex] The index at which to start fetching records.
 * @param {boolean} [options.silent] If set, do not fire a sync event.
 * @return {Promise.<void>} Promise object resolves to void when done
 * @export
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance
 */
oj.ArrayPagingDataSource.prototype.fetch = function(options)
{
    var opts = options || {};
    
    if (opts['pageSize'] !== undefined && opts['startIndex'] !== undefined) {
        if (!this._hasMore()) 
        {
            // At the limit
            return Promise.resolve();
        }
        this.currentPageSize = opts['startIndex'] + opts['pageSize'];
    }
    this._refreshDataWindow(null);
    return Promise.resolve();
};

/**
 * Fetch the row data.
 * @private
 * @memberof oj.ArrayPagingDataSource
 */
oj.ArrayPagingDataSource.prototype._fetchInternal = function(options)
{
    // No fetching, but set up window according to options
    var opts = options || {};
    if (opts['startIndex'] !== undefined) {
        this.current = opts['startIndex'];
    }
    if (opts['pageSize'] !== undefined) {
        this.pageSize = opts['pageSize'];
        this.currentPageSize = opts['pageSize'];
    }
    this._refreshDataWindow(options);
    
    return Promise.resolve({'data': this.dataWindow, 'startIndex': this.current});
};

/**
 * Return whether there is more data which can be fetched.
 * @private
 * @memberof oj.ArrayPagingDataSource
 */
oj.ArrayPagingDataSource.prototype._hasMore = function()
{
  return this.current + this.currentPageSize < this.totalSize();
};

/**
 * Set or change the number of models in a page
 * 
 * @private
 * @memberof oj.ArrayPagingDataSource
 */
oj.ArrayPagingDataSource.prototype._setPageSize = function(n)
{
  this.pageSize = n;
  this.currentPageSize = n;
  this._refreshDataWindow(null);
};

/**
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance
 */
oj.ArrayPagingDataSource.prototype.totalSize = function()
{
    return this.data.length;
};

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @ojsignature {target:"Type",
 *               value: "'actual'|'estimate'|'atLeast'|'unknown'",
 *               for: "returns"}
 * @export
 * @expose
 * @memberof oj.ArrayPagingDataSource
 * @instance 
 */
oj.ArrayPagingDataSource.prototype.totalSizeConfidence = function()
{ 
  return "actual";
};

/**
 * Determines whether this data source supports a certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For sort, the valid return values are: "full", "none".  Returns null if the
 *         feature is not recognized.
 * @export
 * @memberof oj.ArrayPagingDataSource
 */
oj.ArrayPagingDataSource.prototype.getCapability = function(feature)
{
    return null;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @export
 * @class oj.ArrayTableDataSource
 * @since 1.0
 * @ojtsignore
 * @extends oj.TableDataSource
 * @classdesc Object representing data available from an array.  This data source can be used by [ListView]{@link oj.ojListView}, [NavigationList]{@link oj.ojNavigationList},
 *            [TabBar]{@link oj.ojTabBar}, and [Table]{@link oj.ojTable}.<br><br>
 *            See the <a href="../jetCookbook.html?component=table&demo=basicTable">Table - Base Table</a> demo for an example.<br><br>
 *            Refer to {@link oj.TableDataSource} for other data sources that represent tabular data.
 * @param {Array|Object|function():Array} data data supported by the components
 *                                      <p>This can be either an Array, or a Knockout observableArray.</p>
 *                                      <p>Each array element should be an object representing one row of data, with the property names and values corresponding to column names and values.
 *                                         Array of primitive values such as ["Apple", "Orange"] is not currently supported.</p>
 * @param {Object|null} [options] Options for the TableDataSource
 * @param {string} [options.idAttribute] The column that contains the row key.
 *                 If this is not specified, all the values in a row are used as the key.
 * @param {"enabled"|"disabled"} [options.startFetch] Control whether to start initial fetch when the TableDataSource is bound to a component.  Valid values are:<br><br>
 *                                    <b>"enabled"</b> (default) - Start initial fetch automatically when the TableDataSource is bound to a component.<br>
 *                                    <b>"disabled"</b> - Do not start initial fetch automatically.  Application will call the <a href="#fetch">fetch()</a> method to
 *                                                        start the first fetch.
 * @ojsignature {target: "Type",
 *               value: "KnockoutObservableArray<object>|Array<object>",
 *               for: "data"}
 * @constructor
 * @ojtsignore
 * @example
 * // First initialize an array
 * var deptArray = [{DepartmentId: 10, DepartmentName: 'Administration', LocationId: 200},
 *                  {DepartmentId: 20, DepartmentName: 'Marketing', LocationId: 200},
 *                  {DepartmentId: 30, DepartmentName: 'Purchasing', LocationId: 200}];
 *
 * // Then create an ArrayTableDataSource object with the array
 * var dataSource = new oj.ArrayTableDataSource(deptArray, {idAttribute: 'DepartmentId'});
 */
oj.ArrayTableDataSource = function(data, options)
{
  // Initialize
  this.data = data || {};   // This was put in to keep closure happy...
  if (!(data instanceof Array) &&
      (typeof (data) != 'function' &&
       typeof (data['subscribe']) != 'function'))
  {
    // we only support Array or ko.observableArray. To
    // check for observableArray, we can't do instanceof check because it's
    // a function. So we just check if it contains a subscribe function.
    var errSummary = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_SUMMARY'];
    var errDetail = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_DETAIL'];
    throw new Error(errSummary + '\n' + errDetail);
  }
  
  if (options == null || options['idAttribute'] == null)
  {
    oj.Logger.info(oj.ArrayTableDataSource._LOGGER_MSG['_INFO_ARRAY_TABLE_DATASOURCE_IDATTR']);
  }

  oj.ArrayTableDataSource.superclass.constructor.call(this, data, options);

  this._eventHandlers = [];
  this._rows = {};

  if (data != null && data !== undefined)
  {
    this._idAttribute = null;
    if (options != null && options['idAttribute'] != null)
    {
      this._idAttribute = options['idAttribute'];
    }
    this._data = (data instanceof Array) ? data : (/** @type {Function} */(data))();
    this._totalSize = this._data.length;
    this._subscribeObservableArray(data);
  }

  if ((options != null && (options['startFetch'] == 'enabled' || options['startFetch'] == null))
    || options == null)
  {
    this._startFetchEnabled = true;
  }
};

// Subclass from oj.DataSource 
oj.Object.createSubclass(oj.ArrayTableDataSource, oj.TableDataSource, "oj.ArrayTableDataSource");

/**
 * @export
 * @desc If set to a function(row1, row2), then this function is called comparing raw row data (see the
 * JavaScript array.sort() for details)
 * @memberof oj.ArrayTableDataSource
 * @type {null|string|Function}
 * @ojsignature {target: "Type",
 *               value: "null|string|((param0: object, param1?: object)=> number|string|object)"}
 */
oj.ArrayTableDataSource.prototype.comparator = null;

/**
 * @export
 * 
 * @type {Object} 
 * @property {*} criteria.key The key that identifies which field to sort
 * @property {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 *
 * @desc The sort criteria. Whenever sort() is called with the criteria parameter, that value is copied to this
 * property. If sort() is called with empty sort criteria then the criteria set in this property is used.
 * @memberof oj.ArrayTableDataSource
 */
oj.ArrayTableDataSource.prototype.sortCriteria = null;

/**
 * Initializes the instance.
 * @memberof oj.ArrayTableDataSource
 * @instance
 * @override
 * @protected
 */
oj.ArrayTableDataSource.prototype.Init = function()
{
  oj.ArrayTableDataSource.superclass.Init.call(this);
};

/**
 * Add a row (or array of rows) to the end
 * 
 * @param {Object|Array.<Object>} m Row object data (or array of rows) to add. These should be sets of attribute/values.
 * @param {Object=} options 
 * @param {boolean} [options.silent] if set, do not fire an add event<p>
 * @param {number|Array.<number>} [options.at] splice the new row at the value given (at:index). If an array of rows then this should be an array of indexes <p>
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects, an array of keys, and an array of indexes which were added triggering done when complete.<p>
 *         The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>indexes</b></td><td>An array of index values for the rows</td></tr>
 * </tbody>
 * </table> 
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.ArrayTableDataSource.RowDatas>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.add = function(m, options)
{
  options = options || {};
  this._checkDataLoaded();
  var index = options['at'];

  return this._addToRowSet(m, index, options);
};

/**
 * Return the row data found at the given index.
 * 
 * @param {number} index Index for which to return the row data. 
 * @param {Object=} options Options to control the at.
 * @return {Promise} Promise resolves to a compound object which has the structure below. If the index is out of range, Promise resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.TableDataSource.RowData>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.at = function(index, options)
{
  this._checkDataLoaded();
  var row;

  if (index < 0 || index >= this._rows['data'].length)
  {
    row = null;
  }
  else
  {
    row = {'data': this._rows['data'][index], 'index': index, 'key': this._getId(this._rows['data'][index])};
  }

  return new Promise(function(resolve, reject) {
    resolve(row);
  });
};

/**
 * Change a row (or array of rows), if found.
 * @param {Object|Array.<Object>} m Row object data (or array of rows) to change. These should be sets of attribute/values.
 * @param {Object=} options 
 * @param {boolean} [options.silent] if set, do not fire an add event<p>
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects, an array of keys, and an array of indexes which were changed triggering done when complete.<p>
 *          The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>indexes</b></td><td>An array of index values for the rows</td></tr>
 * </tbody>
 * </table>
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.ArrayTableDataSource.RowDatas>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.change = function(m, options)
{
  options = options || {};
  this._checkDataLoaded();
  var silent = options['silent'];
  var i, row, key, changedRow, rowArray = {};
  rowArray['data'] = [];
  rowArray['keys'] = [];
  rowArray['indexes'] = [];
  
  if (!(m instanceof Array))
  {
    m = [m];
  }
  
  for (i = 0; i < m.length; i++)
  {
    row = m[i];
    
    if (row != null)
    {
      key = this._getId(row);
      changedRow = this._getInternal(key, false);
      rowArray['data'].push(this._wrapWritableValue(row));
      rowArray['keys'].push(key);
      rowArray['indexes'].push(changedRow['index']);
      this._rows['data'][changedRow['index']] = row;
    }
  }
    
  if (!silent && rowArray['data'].length > 0)
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['CHANGE'], rowArray);
  }
    
  return Promise.resolve(rowArray);
};

/**
 * Fetch the row data.
 * @param {Object=} options Options to control fetch
 * @param {number} [options.startIndex] The index at which to start fetching records.
 * @param {boolean} [options.silent] If set, do not fire a sync event.
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects, an array of ids, and the startIndex triggering done when complete.<p>
 *      The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>startIndex</b></td><td>The startIndex for the returned set of rows</td></tr>
 * </tbody>
 * </table> 
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.TableDataSource.RowDatas>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.fetch = function(options)
{
  options = options || {};
  var fetchType = options['fetchType'];
  
  if (fetchType == 'init' && !this._startFetchEnabled)
  {
    return Promise.resolve();
  }
  
  return this._fetchInternal(options);
};

/**
 * Return the first row data whose id value is the given id
 * @param {string} id ID for which to return the row data, if found. 
 * @param {Object=} options Options to control the get.
 * @return {Promise} Promise which resolves to a compound object which has the structure below where the id matches the given id. If none are found, resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.TableDataSource.RowData>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.get = function(id, options)
{
  options = options || {};
  this._checkDataLoaded();
  return Promise.resolve(this._getInternal(id, true));
};

/**
 * Determines whether this TableDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none".  
 *         Returns null if the feature is not recognized.
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.getCapability = function(feature)
{
  return 'full';
};

/**
 * Remove a row (or array of rows), if found.
 * @param {Object|Array.<Object>}  m Row object data (or array of rows) to remove. These should be sets of attribute/values.
 * @param {Object=} options 
 * @param {boolean} [options.silent] if set, do not fire a remove event 
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects, an array of keys, and an array of indexes which were removed triggering done when complete.<p>
 *      The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>indexes</b></td><td>An array of index values for the rows</td></tr>
 * </tbody>
 * </table> 
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.ArrayTableDataSource.RowDatas>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.remove = function(m, options)
{
  options = options || {};
  this._checkDataLoaded();
  return this._removeInternal(m, options);
};

/**
 * Remove and replace the entire list of rows with a new set of rows, if provided. Otherwise, empty the datasource. The next fetch
 * call will re-populate the datasource with the original array data. To empty out the data, call reset with an empty array.
 * @param {Array.<Object>=} data Array of row objects with which to replace the data. 
 * @param {Object=} options user options, passed to event
 * @return {Promise.<void>} promise object triggering done when complete.
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.reset = function(data, options)
{
  var i;
  options = options || {};
  options['previousRows'] = this._rows;
  var silent = options['silent'];

  if (data != null)
  {
    this._data = data;
    this._subscribeObservableArray(data);
    this.data = data;
  }
  this._rows = {};
  this._totalSize = 0;

  if (!silent)
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['RESET'], null);
  }
  return Promise.resolve();
};

/**
 * Performs a sort on the data source.
 * @param {Object|null} [criteria] the sort criteria.
 * @param {*} criteria.key The key that identifies which field to sort
 * @param {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @return {Promise.<null>} promise object triggering done when complete.
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.sort = function(criteria)
{
  if (criteria == null)
  {
    criteria = this['sortCriteria'];
  }
  else
  {
    this['sortCriteria'] = criteria;
  }
  
  this._checkDataLoaded();
  var self = this;
  return new Promise(function(resolve, reject) {
    criteria = criteria || {};
    var comparator = self._getComparator();
    
    self._rows['data'].sort(function(a, b)
    {
      return oj.ArrayTableDataSource._sortFunc(a, b, comparator, self);
    });
    self._sorted = true;
    var result = {'header': criteria['key'], 'direction': criteria['direction']};
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['SORT'], result);
    resolve(result);
  });
};

/**
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @export
 * @expose
 * @memberof oj.ArrayTableDataSource
 * @instance
 */
oj.ArrayTableDataSource.prototype.totalSize = function()
{
  return this._totalSize;
};

oj.ArrayTableDataSource.prototype._addToRowSet = function(m, index, options)
{
  var i, j, row, key;
  options = options || {};
  var silent = options['silent'];
  var rowArray = {};
  rowArray['data'] = [];
  rowArray['keys'] = [];
  rowArray['indexes'] = [];
  
  if (!(m instanceof Array))
  {
    m = [m];
  }
  if (index != null && !(index instanceof Array))
  {
    index = [index];
  }
  
  for (i = 0; i < m.length; i++)
  {
    row = m[i];
    
    if (row != null)
    {
      key = this._getId(row);

      rowArray['data'].push(this._wrapWritableValue(row));
      rowArray['keys'].push(key);

      if (this._sorted == true && this._rows['data'].length > 0)
      {
        var self = this;
        for (j = 0; j < this._rows['data'].length; j++)
        {
          if (oj.ArrayTableDataSource._sortFunc(row, this._rows['data'][j], self._getComparator(), self) < 0)
          {
            this._rows['data'].splice(j, 0, row);
            rowArray['indexes'].push(j);
            break;
          }
          else if (j == this._rows['data'].length - 1)
          {
            this._rows['data'].push(row);
            rowArray['indexes'].push(j + 1);
            break;
          }
        }
      }
      else
      {
        if (index == null)
        { 
          this._rows['data'].push(row);
          rowArray['indexes'].push(this._rows['data'].length - 1);
        }
        else
        {
          this._rows['data'].splice(index[i], 0, row);
          rowArray['indexes'].push(index[i]);
        }
      }
      this._totalSize++;
      this._realignRowIndices();
    }
  }
  
  if (!silent && rowArray['data'].length > 0)
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['ADD'], rowArray);
  }
  
  return Promise.resolve(rowArray);
}

oj.ArrayTableDataSource.prototype._checkDataLoaded = function()
{
  if (!this._isDataLoaded())
  {
    if (!(this.data instanceof Array) &&
        typeof (this.data) == 'function' &&
        typeof (this.data['subscribe']) == 'function')
    {
      // if we have an observableArray, reload just in case it was changed
      this._data = (/** @type {Function} */(this.data))();
    }
    this._rows = this._getRowArray(this._data);
    this._totalSize = this._data.length;
  }
}

oj.ArrayTableDataSource.prototype._isDataLoaded = function()
{
  if (this._rows == null || this._rows['data'] == null)
  {
    return false;
  }
  
  return true;
}

oj.ArrayTableDataSource.prototype._fetchInternal = function(options)
{
  options = options || {};
  this._startFetch(options);
  this._checkDataLoaded();
  var pageSize;
  try
  {
    pageSize = options['pageSize'] > 0 ? options['pageSize'] : -1;
    if (!this._startIndex)
    {
      this._startIndex = 0;
    }
    
    this._startIndex = options['startIndex'] == null ? this._startIndex : options['startIndex'];
    var endIndex = oj.ArrayTableDataSource._getEndIndex(this._rows, this._startIndex, pageSize);
    var rowArray = [];
    var keyArray = [];
    var i, key, wrappedRow;
    for (i = this._startIndex; i <= endIndex; i++)
    {
      key = this._getId(this._rows['data'][i]);
      wrappedRow = this._wrapWritableValue(this._rows['data'][i]);
      rowArray[i - this._startIndex] = wrappedRow;
      keyArray[i - this._startIndex] = key;
    }
  }
  catch (err)
  {
    this._endFetch(options, null, err);
    return Promise.reject(err);
  }
  
  if (endIndex < this._startIndex)
  {
    // this means we have no more rows at the startIndex. So adjust our
    // startIndex down to indicate the last row
    this._startIndex = endIndex + 1;
  }
  options['pageSize'] = pageSize;
  options['startIndex'] = this._startIndex;
  options['refresh'] = true;
  var result = {'data': rowArray, 'keys': keyArray, 'startIndex': this._startIndex};
  this._endFetch(options, result, null);
  
  return Promise.resolve(result);
}

oj.ArrayTableDataSource.prototype._getInternal = function(id, wrap)
{
  var i, j, row, wrappedRow, key;
  var result = null;
  for (i = 0; i < this._rows['data'].length; i++)
  {
    row = this._rows['data'][i];
    if (row !== undefined)
    {
      key = this._getId(row);
      if ($.isArray(key) && $.isArray(id))
      {
        if (key.length == id.length)
        {
          var equal = true;
          for (j = 0; j < id.length; j++)
          {
            if (key[j] != id[j])
            {
              equal = false;
              break;
            }
          }
          if (equal)
          {
            if (wrap)
            {
              wrappedRow = this._wrapWritableValue(row);
              result = {'data': wrappedRow, 'key': key, 'index': this._rows['indexes'][i]};
            }
            else
            {
              result = {'data': row, 'key': key, 'index': this._rows['indexes'][i]};
            }
            break;
          }
        }
      }
      else if (key == id)
      {
        if (wrap)
        {
          wrappedRow = this._wrapWritableValue(row);
          result = {'data': wrappedRow, 'key': key, 'index': this._rows['indexes'][i]};
        }
        else
        {
          result = {'data': row, 'key': key, 'index': this._rows['indexes'][i]};
        }
        break;
      }
    }
  }
  return result;
};

oj.ArrayTableDataSource.prototype._getComparator = function()
{
  var comparator = this['comparator'];
  
  if (comparator == null)
  {
    var key = this['sortCriteria']['key']; 
    var direction = this['sortCriteria']['direction'];
  
    if (direction == 'ascending')
    {
      comparator = function(row) 
      {
        if ($.isFunction(row[key]))
        {
          return row[key]();
        }
        return row[key];
      };
    }
    else if (direction == 'descending')
    {
      comparator = function(rowA, rowB) {
        var a, b;
        if ($.isFunction(rowA[key]))
        {
          a = rowA[key]();
          b = rowB[key]();
        }
        else
        {
          a = rowA[key];
          b = rowB[key];
        }

        if (a === b)
        {
          return 0;
        }
        return a > b ? -1 : 1;
      };
    }
  }
  return comparator;
};

// Realign all the indices of the rows (after sort for example)
oj.ArrayTableDataSource.prototype._realignRowIndices = function()
{
  for (var i = 0; i < this._rows['data'].length; i++)
  {
    this._rows['indexes'][i] = i;
  }
};

oj.ArrayTableDataSource.prototype._removeInternal = function(m, options)
{
  var i, j, row, key, deletedRow;
  var self = this;
  options = options || {};
  var silent = options['silent'];
  var rowArray = {};
  rowArray['data'] = [];
  rowArray['keys'] = [];
  rowArray['indexes'] = [];
  
  if (!(m instanceof Array))
  {
    m = [m];
  }
  
  var sortedRowArray = [];
  for (i = 0; i < m.length; i++)
  {
    row = m[i];
    
    if (row != null)
    {
      key = this._getId(row);
      deletedRow = this._getInternal(key, false);
      
      if (deletedRow != null)
      {
        sortedRowArray.push({'data': deletedRow['data'], 'key': deletedRow['key'], 'index': deletedRow['index']});
      }
    }
  }
  sortedRowArray.sort(function(a, b) 
  {
    return a['index'] - b['index'];
  });
  
  for (i = 0; i < sortedRowArray.length; i++)
  {
    rowArray['data'].push(sortedRowArray[i]['data']);
    rowArray['keys'].push(sortedRowArray[i]['key']);
    rowArray['indexes'].push(sortedRowArray[i]['index']);
  }
  
  for (i = rowArray['indexes'].length - 1; i >=0; i--)
  {
    this._rows['data'].splice(rowArray['indexes'][i], 1);
    this._rows['indexes'].splice(rowArray['indexes'][i], 1);
    this._totalSize--;
  }
  this._realignRowIndices();
  
  if (!silent && rowArray['data'].length > 0)
  {
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['REMOVE'], rowArray);
  }
      
  return Promise.resolve(rowArray);
};

oj.ArrayTableDataSource.prototype._setRow = function(index, row)
{
  this._rows[index] = row;
  row['index'] = index;
};

/**
 * Indicate starting fetch
 * @param {Object} options
 * @private
 * @memberof oj.ArrayTableDataSource
 */
oj.ArrayTableDataSource.prototype._startFetch = function(options)
{
  if (!options['silent'])
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['REQUEST'], {'startIndex' : options['startIndex']});
  }
};

/**
 * Indicate ending fetch
 * @param {Object} options
 * @param {Object} result Result object
 * @param {Object} error Error
 * @private
 * @memberof oj.ArrayTableDataSource
 */
oj.ArrayTableDataSource.prototype._endFetch = function(options, result, error)
{
  if (error != null)
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['ERROR'], error);
  }
  else
  {
    if (!options['silent'])
    {
      oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['SYNC'],  result);
    }
  }
};

oj.ArrayTableDataSource.prototype._handleRowChange = function(event)
{
  event['startIndex'] = this._startIndex;
  oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['CHANGE'], event);
};

oj.ArrayTableDataSource._compareKeys = function(keyA, keyB, direction)
{
  if (direction == 'descending')
  {
    if (keyA < keyB)
    {
      return 1;
    }
    if (keyB < keyA)
    {
      return -1;
    }
  }
  else
  {
    if (keyA > keyB)
    {
      return 1;
    }
    if (keyB > keyA)
    {
      return -1;
    }
  }
  return 0;
};

oj.ArrayTableDataSource._getEndIndex = function(rows, startIndex, pageSize)
{
  var endIndex = rows['data'].length - 1;
  
  if (pageSize > 0)
  {
    endIndex = startIndex + pageSize - 1;
    endIndex = endIndex > rows['data'].length - 1 ? rows['data'].length - 1 : endIndex;
  }
  
  return endIndex;
};

oj.ArrayTableDataSource._getKey = function(val, attr) {
  if (typeof (val[attr]) == 'function') {
    return val[attr]();
  }
  return val[attr];
};

oj.ArrayTableDataSource.prototype._getRowArray = function(values)
{
  var endIndex = values.length - 1;

  var rowArray = {}, i, prop;
  rowArray['data'] = [];
  rowArray['indexes'] = [];
  this._attributes = null;
  
  for (i = 0; i <= endIndex; i++)
  {
    var clonedRowValues = {};
    var rowValues = values[i];

    for (prop in rowValues)
    {
      if (rowValues.hasOwnProperty(prop))
      {
        clonedRowValues[prop] = rowValues[prop];
        if (i == 0)
        {
          if (this._attributes == null)
          {
            this._attributes = [];
          }
          this._attributes.push(prop);
        }
      }
    }
    rowArray['data'][i] = clonedRowValues;
    rowArray['indexes'][i] = i;
  }

  return rowArray;
};

oj.ArrayTableDataSource.prototype._getId = function(row)
{
  var id;
  var idAttribute = this._getIdAttr(row);
  
  if ($.isArray(idAttribute))
  {
    var i;
    id = [];
    for (i = 0; i < idAttribute.length; i++)
    {
      if (idAttribute[i] in row)
      {
        id[i] =  oj.ArrayTableDataSource._getKey(row, idAttribute[i]);
      }
      else
      {
        var errDetail = oj.Translations.applyParameters(oj.ArrayTableDataSource._LOGGER_MSG['_ERR_ARRAY_TABLE_DATASOURCE_IDATTR_NOT_IN_ROW'], [idAttribute[i]]);
        throw new Error(errDetail);
      }
    }
  }
  else
  {
    if (idAttribute in row)
    {
      id = oj.ArrayTableDataSource._getKey(row, idAttribute);
    }
    else
    {
        var errDetail = oj.Translations.applyParameters(oj.ArrayTableDataSource._LOGGER_MSG['_ERR_ARRAY_TABLE_DATASOURCE_IDATTR_NOT_IN_ROW'], [idAttribute]);
        throw new Error(errDetail);
    }
  }
  
  return id;
};

oj.ArrayTableDataSource.prototype._getIdAttr = function(row)
{
  if (this._idAttribute != null)
  {
    return this._idAttribute;
  }
  else
  {
    if (this._attributes == null)
    {
      this._attributes = [];
      var prop;
      for (prop in row)
      {
        if (row.hasOwnProperty(prop))
        {
          this._attributes.push(prop);
        }
      }
    }

    if (this._attributes.hasOwnProperty('id'))
    {
      return 'id';
    }
    else
    {
      return this._attributes;
    }
  }
};

oj.ArrayTableDataSource._sortFunc = function(a, b, comparator, self)
{
  var keyA, keyB, i, retVal;
  var direction = self['sortCriteria']['direction'];

  if ($.isFunction(comparator))
  {
    // How many args?
    if (comparator.length === 1)
    {
      // "sortBy" comparator option
      keyA = comparator.call(self, a);
      keyB = comparator.call(self, b);
      var attrs1 = oj.StringUtils.isString(keyA) ? keyA.split(",") : [keyA];
      var attrs2 = oj.StringUtils.isString(keyB) ? keyB.split(",") : [keyB];
      for (i = 0; i < attrs1.length; i++)
      {
        retVal = oj.ArrayTableDataSource._compareKeys(attrs1[i], attrs2[i], direction);
        if (retVal !== 0)
        {
          return retVal;
        }
      }
      return 0;
    }
    // "sort" comparator option
    return comparator.call(self, a, b);
  }
  // String option
  if (oj.StringUtils.isString(comparator))
  {
    var attrs = comparator.split(",");

    for (i = 0; i < attrs.length; i++)
    {
      keyA = oj.ArrayTableDataSource._getKey(a, attrs[i]);
      keyB = oj.ArrayTableDataSource._getKey(b, attrs[i]);
      retVal = oj.ArrayTableDataSource._compareKeys(keyA, keyB, direction);
      if (retVal !== 0)
      {
        return retVal;
      }
    }
  }
  return 0;
};

oj.ArrayTableDataSource.prototype._subscribeObservableArray = function(data)
{
  if (!(data instanceof Array))
  {
    var self = this;
    // subscribe to observableArray arrayChange event to get individual updates
    (/** @type {{subscribe: Function}} */(data))['subscribe']
      (
        function(changes) 
        {
          if (!self._isDataLoaded())
          {
            // don't bother with data change notifications
            // if the data hasn't even been loaded yet (e.g. initial 
            // fetch hasn't happened
            return;
          }
          var i, j;

          // do two passes, first for deletes and the second for adds
          var rowArray = [];
          var indexArray = [];
          for (i = 0; i < changes.length; i++)
          {
            if (changes[i]['status'] === 'deleted')
            {
              rowArray.push(changes[i].value);
            }
          }
          self.remove(rowArray, null);

          rowArray = []; 
          indexArray = [];
          for (i = 0; i < changes.length; i++)
          {
            if (changes[i]['status'] === 'added')
            {
              rowArray.push(changes[i].value);
              indexArray.push(changes[i].index);
            }
          }
          self.add(rowArray, {'at': indexArray});
        }, null, 'arrayChange');
  }
};

oj.ArrayTableDataSource.prototype._wrapWritableValue = function(m)
{
  var returnObj = {};
  var i, props = Object.keys(m);
  
  for (i = 0; i < props.length; i++)
  {
    oj.ArrayTableDataSource._defineProperty(returnObj, m, props[i]);
  }
  
  return returnObj;
};

oj.ArrayTableDataSource._defineProperty = function(row, m, prop)
{
  Object.defineProperty(row, prop,
    {
      get: function()
      {
        return m[prop];
      },
      set: function(newValue)
      {
        m[prop] = newValue;
      },
      enumerable: true
    });
};

oj.ArrayTableDataSource._LOGGER_MSG =
  {
    '_INFO_ARRAY_TABLE_DATASOURCE_IDATTR': "idAttribute option has not been specified. Will default to using 'id' if the field exists. If not, will use all the fields.",
    '_ERR_ARRAY_TABLE_DATASOURCE_IDATTR_NOT_IN_ROW': "Specified idAttribute {0} not in row data. Please ensure all specified idAttribute fields are in the row data or do not specify idAttribute and all fields will be used as id."
  };

/**
 * Shape of Data Info returned by methods like, add/change/remove.
 * @typedef {Object} oj.ArrayTableDataSource.RowDatas
 * @property {Array.<Object>} data An array of raw row data. 
 * @property {Array.<any>} keys An array of key values for the rows.
 * @property {Array.<number>} indexes An array of index values for the rows.
 */
var ArrayTreeDataProvider = /** @class */ (function () {
    function ArrayTreeDataProvider(treeData, options, _rootDataProvider) {
        this.treeData = treeData;
        this.options = options;
        this._rootDataProvider = _rootDataProvider;
        this.TreeAsyncIterator = /** @class */ (function () {
            function class_1(_parent, _baseIterable) {
                this._parent = _parent;
                this._baseIterable = _baseIterable;
            }
            class_1.prototype['next'] = function () {
                var self = this;
                return this._baseIterable[Symbol.asyncIterator]().next().then(function (result) {
                    var metadata = result.value.metadata;
                    for (var i = 0; i < metadata.length; i++) {
                        // Replace flat array metadata with tree array metadata 
                        metadata[i] = self._parent._getNodeMetadata(result.value.data[i]);
                    }
                    return result;
                });
            };
            return class_1;
        }());
        this.TreeAsyncIterable = /** @class */ (function () {
            function class_2(_parent, _asyncIterator) {
                this._parent = _parent;
                this._asyncIterator = _asyncIterator;
                this[Symbol.asyncIterator] = function () {
                    return this._asyncIterator;
                };
            }
            return class_2;
        }());
        this._baseDataProvider = new oj['ArrayDataProvider'](treeData, options);
        this._mapKeyToNode = new Map();
        this._mapNodeToKey = new Map();
        this._nodeSequenceNum = 0;
        // Subscribe to all children observableArray at the top-level
        if (_rootDataProvider == null) {
            this._subscribeObservableArray(treeData, []);
        }
    }
    ArrayTreeDataProvider.prototype.containsKeys = function (params) {
        var self = this;
        return this.fetchByKeys(params).then(function (fetchByKeysResult) {
            var results = new Set();
            params['keys'].forEach(function (key) {
                if (fetchByKeysResult['results'].get(key) != null) {
                    results.add(key);
                }
            });
            return Promise.resolve({ 'containsParameters': params, 'results': results });
        });
    };
    ArrayTreeDataProvider.prototype.getCapability = function (capabilityName) {
        return this._baseDataProvider.getCapability(capabilityName);
    };
    ArrayTreeDataProvider.prototype.getTotalSize = function () {
        return this._baseDataProvider.getTotalSize();
    };
    ArrayTreeDataProvider.prototype.isEmpty = function () {
        return this._baseDataProvider.isEmpty();
    };
    ArrayTreeDataProvider.prototype.getChildDataProvider = function (parentKey, options) {
        var node = this._getNodeForKey(parentKey);
        if (node) {
            var children = this._getChildren(node);
            if (children) {
                var childDataProvider = new ArrayTreeDataProvider(children, this.options, this._getRootDataProvider());
                return childDataProvider;
            }
        }
        return null;
    };
    ArrayTreeDataProvider.prototype.fetchFirst = function (params) {
        var baseIterable = this._baseDataProvider.fetchFirst(params);
        return new this.TreeAsyncIterable(this, new this.TreeAsyncIterator(this, baseIterable));
    };
    ArrayTreeDataProvider.prototype.fetchByOffset = function (params) {
        var basePromise = this._baseDataProvider.fetchByOffset(params);
        var self = this;
        return basePromise.then(function (result) {
            // Repackage the results with tree node metadata
            var results = result.results;
            var newResults = [];
            for (var i = 0; i < results.length; i++) {
                var metadata = results[i]['metadata'];
                var data = results[i]['data'];
                metadata = self._getNodeMetadata(data);
                newResults.push({ 'data': data, 'metadata': metadata });
            }
            return { 'done': result['done'],
                'fetchParameters': result['fetchParameters'],
                'results': newResults };
        });
    };
    ArrayTreeDataProvider.prototype.fetchByKeys = function (params) {
        var self = this;
        var results = new Map();
        params['keys'].forEach(function (key) {
            var node = self._getNodeForKey(key);
            if (node) {
                results.set(key, { 'metadata': { 'key': key },
                    'data': node });
            }
        });
        return Promise.resolve({ 'fetchParameters': params,
            'results': results });
    };
    ArrayTreeDataProvider.prototype._getChildren = function (node) {
        var childrenAttr = this.options && this.options['childrenAttribute'] ? this.options['childrenAttribute'] : 'children';
        return this._getVal(node, childrenAttr);
    };
    ArrayTreeDataProvider.prototype._getRootDataProvider = function () {
        if (this._rootDataProvider) {
            return this._rootDataProvider;
        }
        else {
            return this;
        }
    };
    ArrayTreeDataProvider.prototype._handleArrayChange = function (changes, treeData, parentKeyPath) {
        var self = this;
        var i, id, dataArray = [], keyArray = [], indexArray = [], metadataArray = [];
        // first see if we have deletes and adds. If we do then just do a refresh
        var foundDelete = false;
        var foundAdd = false;
        var dispatchRefreshEvent = false;
        for (i = 0; i < changes.length; i++) {
            if (changes[i]['status'] === 'deleted') {
                foundDelete = true;
                break;
            }
        }
        for (i = 0; i < changes.length; i++) {
            if (changes[i]['status'] === 'added') {
                foundAdd = true;
                break;
            }
        }
        if (foundAdd && foundDelete) {
            dispatchRefreshEvent = true;
        }
        if (foundDelete) {
            for (i = 0; i < changes.length; i++) {
                if (changes[i]['status'] === 'deleted') {
                    var node = changes[i].value;
                    var key = self._getKeyForNode(node);
                    keyArray.push(key);
                    dataArray.push(node);
                    indexArray.push(changes[i].index);
                    self._deleteMapEntry(key, node);
                }
            }
            if (keyArray.length > 0 && !dispatchRefreshEvent) {
                metadataArray = keyArray.map(function (value) {
                    return { 'key': value };
                });
                var keySet_1 = new Set();
                keyArray.map(function (key) {
                    keySet_1.add(key);
                });
                var operationEventDetail = { data: dataArray, indexes: indexArray, keys: keySet_1, metadata: metadataArray };
                var mutationEventDetail = { remove: operationEventDetail };
                self.dispatchEvent(new oj.DataProviderMutationEvent(mutationEventDetail));
            }
        }
        if (foundAdd) {
            var nodeArray_1 = treeData();
            for (i = 0; i < changes.length; i++) {
                if (changes[i]['status'] === 'added') {
                    var node = changes[i].value;
                    var keyObj = self._createKeyObj(node, parentKeyPath, -1);
                    self._setMapEntry(keyObj.key, node);
                    keyArray.push(keyObj.key);
                    dataArray.push(node);
                    indexArray.push(changes[i].index);
                }
            }
            if (keyArray.length > 0 && !dispatchRefreshEvent) {
                metadataArray = keyArray.map(function (value) {
                    return { key: value };
                });
                var keySet_2 = new Set();
                keyArray.map(function (key) {
                    keySet_2.add(key);
                });
                var afterKeySet_1 = new Set();
                indexArray.map(function (addIndex) {
                    var afterKey;
                    if (addIndex >= nodeArray_1.length - 1) {
                        afterKey = '';
                    }
                    else {
                        afterKey = self._getKeyForNode(nodeArray_1[addIndex + 1]);
                    }
                    afterKeySet_1.add(afterKey);
                });
                var operationEventDetail = { afterKeys: afterKeySet_1, data: dataArray, indexes: indexArray, keys: keySet_2, metadata: metadataArray };
                var mutationEventDetail = { add: operationEventDetail };
                self.dispatchEvent(new oj.DataProviderMutationEvent(mutationEventDetail));
            }
            if (dispatchRefreshEvent) {
                self.dispatchEvent(new oj.DataProviderRefreshEvent());
            }
        }
    };
    /**
     * If observableArray, then subscribe to it
     */
    ArrayTreeDataProvider.prototype._subscribeObservableArray = function (treeData, parentKeyPath) {
        var self = this;
        var dataArray;
        if (treeData instanceof Array) {
            dataArray = treeData;
        }
        else {
            if (!(ko.isObservable(treeData) && !(treeData['destroyAll'] === undefined))) {
                // we only support Array or ko.observableArray
                throw new Error('Invalid data type. ArrayTreeDataProvider only supports Array or observableArray.');
            }
            // subscribe to observableArray arrayChange event to get individual updates
            treeData['subscribe'](function (changes) {
                self._handleArrayChange(changes, treeData, parentKeyPath);
            }, null, 'arrayChange');
            dataArray = treeData();
        }
        dataArray.forEach(function (node, i) {
            var keyObj = self._createKeyObj(node, parentKeyPath, i);
            self._setMapEntry(keyObj.key, node);
            // Keep track of a sequence number so if any node get added later on,
            // we won't reuse existing key if there is no keyAttributes and we generate
            // the id.
            ++self._nodeSequenceNum;
            var children = self._getChildren(node);
            if (children) {
                self._subscribeObservableArray(children, keyObj.keyPath);
            }
        });
    };
    ArrayTreeDataProvider.prototype._createKeyObj = function (node, parentKeyPath, idx) {
        var key = this._getId(node);
        var keyPath = parentKeyPath ? parentKeyPath.slice() : [];
        if (key == null) {
            // _getId returns null if keyAttributes is not specified.  In this case we 
            // use the index path of the node as the key.
            keyPath.push(idx >= 0 ? idx : this._nodeSequenceNum++);
            key = keyPath;
        }
        else {
            keyPath.push(key);
            if (this.options && this.options['keyAttributesScope'] == 'siblings') {
                // If the id is only unique among siblings, we use the id path of the 
                // node as the key.
                key = keyPath;
            }
        }
        return { 'key': key, 'keyPath': keyPath };
    };
    /**
     * Get id value for row
     */
    ArrayTreeDataProvider.prototype._getId = function (row) {
        var id;
        var keyAttributes = this.options != null ? this.options['keyAttributes'] : null;
        if (keyAttributes != null) {
            if (Array.isArray(keyAttributes)) {
                var i;
                id = [];
                for (i = 0; i < keyAttributes.length; i++) {
                    id[i] = this._getVal(row, keyAttributes[i]);
                }
            }
            else if (keyAttributes == '@value') {
                id = this._getAllVals(row);
            }
            else {
                id = this._getVal(row, keyAttributes);
            }
            return id;
        }
        else {
            return null;
        }
    };
    ;
    /**
     * Get value for attribute
     */
    ArrayTreeDataProvider.prototype._getVal = function (val, attr) {
        if (typeof attr == 'string') {
            var dotIndex = attr.indexOf('.');
            if (dotIndex > 0) {
                var startAttr = attr.substring(0, dotIndex);
                var endAttr = attr.substring(dotIndex + 1);
                var subObj = val[startAttr];
                if (subObj) {
                    return this._getVal(subObj, endAttr);
                }
            }
        }
        if (typeof (val[attr]) == 'function') {
            return val[attr]();
        }
        return val[attr];
    };
    ;
    /**
     * Get all values in a row
     */
    ArrayTreeDataProvider.prototype._getAllVals = function (val) {
        var self = this;
        return Object.keys(val).map(function (key) {
            return self._getVal(val, key);
        });
    };
    ;
    ArrayTreeDataProvider.prototype._getNodeMetadata = function (node) {
        return { 'key': this._getKeyForNode(node) };
    };
    ArrayTreeDataProvider.prototype._getNodeForKey = function (key) {
        var rootDataProvider = this._getRootDataProvider();
        return rootDataProvider._mapKeyToNode.get(key);
    };
    ArrayTreeDataProvider.prototype._getKeyForNode = function (node) {
        var rootDataProvider = this._getRootDataProvider();
        return rootDataProvider._mapNodeToKey.get(node);
    };
    ArrayTreeDataProvider.prototype._setMapEntry = function (key, node) {
        var rootDataProvider = this._getRootDataProvider();
        rootDataProvider._mapKeyToNode.set(key, node);
        rootDataProvider._mapNodeToKey.set(node, key);
    };
    ArrayTreeDataProvider.prototype._deleteMapEntry = function (key, node) {
        var rootDataProvider = this._getRootDataProvider();
        rootDataProvider._mapKeyToNode.delete(key);
        rootDataProvider._mapNodeToKey.delete(node);
    };
    return ArrayTreeDataProvider;
}());
oj.ArrayTreeDataProvider = ArrayTreeDataProvider;
oj['ArrayTreeDataProvider'] = ArrayTreeDataProvider;
oj.EventTargetMixin.applyMixin(ArrayTreeDataProvider);

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @ojstatus preview
 * @since 5.1.0
 * @export
 * @class oj.ArrayTreeDataProvider
 * @implements oj.TreeDataProvider
 * @classdesc This class implements {@link oj.TreeDataProvider} and is used to represent hierachical data available from an array.<br><br>
 *            Each array element represents a tree node, which can contain nested child object array for its subtree.  
 *            Array elements can be in any shape and form, but is usually an object with a "children" property.  The name of the "children" property
 *            can optionaly be specified with the "childrenAttribute" option.<br><br>
 *            For nodes that cannot have children, the "children" property should not be set.
 *            For nodes that can but don't have children, the "children" property should be set to an empty array.<br><br>
 *            Data can be passed as a regular array or a Knockout observableArray.  If a Knockout observableArray is
 *            used, any mutation must be performed with observableArray methods.  The events described below will be dispatched to the ArrayTreeDataProvider
 *            with the appropriate event payload.  
 *
 * <h3 id="events-section">
 *   Events
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#events-section"></a>
 * </h3>
 * Consumers can add event listeners to listen for the following event types and respond to data change.
 * Event listeners should be added to the root-level ArrayTreeDataProvider created by the application. The root-level ArrayTreeDataProvider receives events for the entire tree.
 * Child-level ArrayTreeDataProvider returned by getChildDataProvider does not receive events.
 * <h4 id="event:mutate" class="name">
 *   mutate
 * </h4>
 * This event is fired when items have been added or removed from the data.
 * <p>
 * Event payload is found under <code class="prettyprint">event.detail</code>, which implements the {@link oj.DataProviderMutationEventDetail} interface.
 * </p>
 *
 * <h4 id="event:refresh" class="name">
 *   refresh
 * </h4>
 * This event is fired when the data has been refreshed and components need to re-fetch the data.
 * <p>
 * This event contains no additional event payload.
 * </p>
 *
 * <i>Example of consumer listening for the "mutate" event type:</i>
 * <pre class="prettyprint"><code>var listener = function(event) {
 *   if (event.detail.remove) {
 *     var removeDetail = event.detail.remove;
 *     // Handle removed items
 *   }
 * };
 * dataProvider.addEventListener("mutate", listener);
 * </code></pre>
 *
 * @param {(Array|function():Array)} data data supported by the components
 *                                      <p>This can be either an Array, or a Knockout observableArray.</p>
 * @param {Object=} options Options for the ArrayTreeDataProvider
 * @param {oj.SortComparators=} options.sortComparators Optional {@link oj.sortComparator} to use for sort.
 * @param {Array.<oj.SortCriterion>=} options.implicitSort Optional array of {@link oj.sortCriterion} used to specify sort information when the data loaded into the dataprovider is already sorted.
 * @param {(string | Array.<string>)=} options.keyAttributes Optional attribute name(s) which stores the key in the data. Can be a string denoting a single key attribute or an array
 *                                                         of strings for multiple key attributes.  Dot notation can be used to specify nested attribute (e.g. 'attr.id').<br><br>
 *                      If specified, caller must ensure that the keyAttributes contains values that are either unique within the entire tree,
 *                        or unique among the siblings of each node.  In the latter case, Caller must also set the keyAttributesScope option to 'siblings'.<br>
 *                      If keyAttributes is specified and keyAttributesScope is 'global', the attribute value will be used as the key.<br>
 *                      If keyAttributes is specified and keyAttributesScope is 'siblings', a path array of the attribute values, starting from the root node, will be used as the key.<br>
 *                      If keyAttributes is not specified, a path array of node index, starting from the root node, will be used as the key.
 * @param {('global'|'siblings')=} options.keyAttributesScope Optional scope of the key values in the fields specified by keyAttributes.  Supported values:<br>
 *                                           <ul>
 *                                             <li>'global': the key values are unique within the entire tree.
 *                                             <li>'siblings': the key values are unique among the siblings of each node.
 *                                           </ul>
 *                                           Default is 'global'.
 * @param {string=} options.childrenAttribute Optional field name which stores the children of nodes in the data.  Dot notation can be used to specify nested attribute.
 *                                                  If this is not specified, the default is "children".
 * @ojsignature [{target: "Type",
 *               value: "class ArrayTreeDataProvider<K, D> implements TreeDataProvider<K, D>"},
 *               {target: "Type",
 *               value: "SortCriterion<D>[]", 
 *               for: "options.implicitSort"},
 *               {target: "Type",
 *               value: "SortComparators<D>", 
 *               for: "options.sortComparators"}]
 * @example
 * // First initialize the tree data.  This can be defined locally or read from file.
 * var treeData = [
 *                  {"attr": {"id": "dir1", "title": "Directory 1"},
 *                   "children": [
 *                     {"attr": {"id": "subdir1", "title": "Subdirectory 1"},
 *                      "children": [
 *                        {"attr": {"id": "file1", "title": "File 1"}},
 *                        {"attr": {"id": "file2", "title": "File 2"}},
 *                        {"attr": {"id": "file3", "title": "File 3"}}
 *                      ]},
 *                     {"attr": {"id": "subdir2", "title": "Subdirectory 2"},
 *                      "children": []}
 *                   ]},
 *                  {"attr": {"id": "dir2", "title": "Directory 2"},
 *                   "children": [
 *                     {"attr": {"id": "file4", "title": "File 4"}},
 *                     {"attr": {"id": "file5", "title": "File 5"}},
 *                   ]}
 *                ];
 *
 * // Then create an ArrayTreeDataProvider object with the array
 * var dataprovider = new oj.ArrayTreeDataProvider(treeData, {keyAttributes: 'attr.id'});
 */

/**
 * Fetch rows by keys in the entire tree.
 *
 * @ojstatus preview
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.FetchByKeysResults>} Promise which resolves to {@link oj.FetchByKeysResults}
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name fetchByKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<FetchByKeysResults<K, D>>"}
 */

/**
 * Get a data provider for the children of the row identified by parentKey.
 * 
 * @ojstatus preview
 * @param {any} parentKey key of the row to get child data provider for.
 * @return {ArrayTreeDataProvider | null} An ArrayTreeDataProvider if the row can (but doesn't have to) have children; or null if the row cannot have children.
 *   Use the <code class="prettyprint">isEmpty</code> method on the returned ArrayTreeDataProvider to determine if it currently has children.
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name getChildDataProvider
 * @ojsignature {target: "Type",
 *               value: "(any): ArrayTreeDataProvider<K, D>"}
 */

/**
 * Check if there are rows containing the specified keys in the entire tree.
 *
 * @ojstatus preview
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.ContainsKeysResults>} Promise which resolves to {@link oj.ContainsKeysResults}
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name containsKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<ContainsKeysResults<K>>"}
 */

/**
 * Fetch rows by offset at the top level.
 *
 * @ojstatus preview
 * @param {oj.FetchByOffsetParameters} params Fetch by offset parameters
 * @return {Promise.<oj.FetchByOffsetResults>} Promise which resolves to {@link oj.FetchByOffsetResults}
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name fetchByOffset
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByOffsetParameters<D>): Promise<FetchByOffsetResults<K, D>>"}
 */

/**
 * Fetch the first set of rows at the current level.
 * <p>
 * If <code class="prettyprint">params.sortCriteria</code> is specified, the default sorting algorithm used is natural sort.
 * </p>
 * 
 * @ojstatus preview
 * @param {oj.FetchListParameters=} params Fetch parameters
 * @return {AsyncIterable.<oj.FetchListResult>} AsyncIterable with {@link oj.FetchListResult}
 * @see {@link https://github.com/tc39/proposal-async-iteration} for further information on AsyncIterable.
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name fetchFirst
 * @ojsignature {target: "Type",
 *               value: "<F extends FetchListResult<K, D>>(params?: FetchListParameters<D>): AsyncIterable<F>"}
 */

/**
 * Determines whether this data provider supports certain feature.
 * 
 * @ojstatus preview
 * @param {string} capabilityName capability name. Supported capability names are:
 *                  "fetchByKeys", "fetchByOffset", and "sort".
 * @return {Object} capability information or null if unsupported
 * <ul>
 *   <li>If capabilityName is "fetchByKeys", returns a {@link oj.FetchByKeysCapability} object.</li>
 *   <li>If capabilityName is "fetchByOffset", returns a {@link oj.FetchByOffsetCapability} object.</li>
 *   <li>If capabilityName is "sort", returns a {@link oj.SortCapability} object.</li>
 * </ul>
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name getCapability
 * @ojsignature {target: "Type",
 *               value: "(capabilityName?: string): any"}
 */

/**
 * Return the total number of rows at the top level.
 * 
 * @ojstatus preview
 * @return {Promise.<number>} Returns a Promise which resolves to the total number of rows. -1 is unknown row count.
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name getTotalSize
 */

/**
 * Return a string that indicates if this data provider is empty.
 * 
 * @ojstatus preview
 * @return {string} a string that indicates if this data provider is empty. Valid values are:
 *                  "yes": this data provider is empty.
 *                  "no": this data provider is not empty.
 *                  "unknown": it is not known if this data provider is empty until a fetch is made.
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name isEmpty
 */

/**
 * Add a callback function to listen for a specific event type.
 *
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name addEventListener
 * @param {string} eventType The event type to listen for.
 * @param {EventListener} listener The callback function that receives the event notification.
 * @ojsignature {target: "Type",
 *               value: "(eventType: string, listener: EventListener): void"}
 */

/**
 * Remove a listener previously registered with addEventListener.
 *
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name removeEventListener
 * @param {string} eventType The event type that the listener was registered for.
 * @param {EventListener} listener The callback function that was registered.
 * @ojsignature {target: "Type",
 *               value: "(eventType: string, listener: EventListener): void"}
 */

/**
 * Dispatch an event and invoke any registered listeners.
 *
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ArrayTreeDataProvider
 * @instance
 * @method
 * @name dispatchEvent
 * @param {Event} event The event object to dispatch.
 * @return {boolean} Return false if a registered listener has cancelled the event. Return true otherwise.
 * @ojsignature {target: "Type",
 *               value: "(evt: Event): boolean"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

// jim retrieved from https://raw.github.com/jquery/jquery-ui/1-10-stable/ui/jquery.ui.button.js on 6/29/2013, and then modified

// -------------------------------------------------------------------------------------------------------
// This file contains both the Button and Buttonset components,
// so that they can share the same wrapper function / closure containing shared "private static members".
// -------------------------------------------------------------------------------------------------------

(function() // Button / Buttonset wrapper function, to keep "private static members" private
{

/**
 * @ojcomponent oj.ojButton
 * @augments oj.baseComponent
 * @since 0.6
 * @ojstatus preview
 * @ojshortdesc A clickable button.
 * @ojrole button
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojButton<SP extends ojButtonSettableProperties = ojButtonSettableProperties> extends baseComponent<SP>"
 *               }
 *              ]
 *
 * @classdesc
 * <h3 id="buttonOverview-section">
 *   JET Button
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#buttonOverview-section"></a>
 * </h3>
 *
 * <p>Description: Themeable, WAI-ARIA-compliant push buttons, with appropriate styles for hover, active, checked, and disabled.
 *
 * <p>To create toggle buttons, see the [JET Buttonset]{@link oj.ojButtonset}.
 *
 * <pre class="prettyprint"><code>&lt;oj-button id="myButton">
 *     &lt;span>My Button&lt;/span>
 * &lt;/oj-button>
 * </code></pre>
 *
 * <h3 id="pushButtons-section">
 *   Push Buttons
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#pushButtons-section"></a>
 * </h3>
 *
 * <p>Push buttons are ordinary buttons that do not stay pressed in when clicked.
 * Push buttons are created from <code class="prettyprint">oj-button</code> elements.
 *
 * <p>Anchor-based buttons are recommended only if native anchor functionality such as href navigation is desired.  If only a click listener is needed,
 * button-based buttons are recommended.
 *
 * <h3 id="buttonsetToolbar-section">
 *   Buttonsets and Toolbars
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#buttonsetToolbar-section"></a>
 * </h3>
 *
 * <p>The [JET Buttonset]{@link oj.ojButtonset} component can be used to create toggle buttons or group related buttons, such as a group of radios or checkboxes.  Buttonset provides
 * visual and semantic grouping and WAI-ARIA-compliant focus management.  See the Buttonset API doc for more information.
 *
 * <p>Also, buttons and buttonsets can be placed in a [JET Toolbar]{@link oj.ojToolbar}.  Like Buttonset, Toolbar is themable and provides WAI-ARIA-compliant
 * focus management.  See the Toolbar API doc for more information.
 *
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>For accessibility, a JET Button must always have its default slot filled, even if it is <a href="#display">icon-only</a>.
 *
 * {@ojinclude "name":"accessibilityCommon"}
 * 
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 *
 * <h3 id="eventHandling-section">
 *   Event Handling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#eventHandling-section"></a>
 * </h3>
 *
 * <p>It's most reliable to register click handlers directly on the button, rather than on an ancestor.  For example, if any of the button's DOM is swapped out in a click
 * handler (e.g. alternating the text and icon between "Play" and "Pause"), the bubbling will stop if the click target was part of the content that was removed in the swap.
 *
 * <p>In lieu of a shared listener on an ancestor, syntax like <code class="prettyprint">$( "#ancestor :oj-button" ).click( myFunc );</code> can be used to set a handler on many
 * buttons at once.
 *
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <p>In lieu of stamping a button in a table, dataGrid, or other container, consider placing a single Button outside the
 * container that acts on the currently selected row or cell.
 *
 * <h3 id="state-section">
 *   Setting Component State
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#state-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stateCommon"}
 */
oj.__registerWidget("oj.ojButton", $['oj']['baseComponent'],
{
    defaultElement: "<button>", // added to externs.js, since this is an override of a superclass member.  (That's the rule for public methods, what about protected fields?)  Would @override do the job and be better than externing?
    widgetEventPrefix : "oj",
    options: // options is in externs.js, so no need for quotes
    {
        /**
         * {@ojinclude "name":"buttonCommonChroming"}
         * 
         * @name chroming
         * @instance
         * @memberof oj.ojButton
         * @type {string}
         * @ojvalue {string} "full" In typical themes, full-chrome buttons always have chrome.
         * @ojvalue {string} "half" In typical themes, half-chrome buttons acquire chrome only in their hover, active, and selected states. Half-chroming is recommended for buttons in a toolbar.
         *     (This is the toolbar default in most themes.)
         * @ojvalue {string} "outlined" In typical themes, outlined buttons are similar to half-chrome buttons, but have a border in the default state.
         * @ojshortdesc Indicates in what states the button has chrome (background and border).
         *
         * @example <caption>Initialize the Button with the <code class="prettyprint">chroming</code> attribute specified:</caption>
         * &lt;oj-button chroming='half'>&lt;/oj-button>
         *
         * @example <caption>Get or set the <code class="prettyprint">chroming</code> property after initialization:</caption>
         * // getter
         * var chromingValue = myButton.chroming;
         *
         * // setter
         * myButton.chroming = 'half';
         *
         * @example <caption>Set the default in the theme (SCSS) :</caption>
         * $buttonChromingOptionDefault: half !default;
         */
        
        /**
         * {@ojinclude "name":"buttonCommonChroming"}
         * 
         * @name chroming
         * @instance
         * @memberof oj.ojMenuButton
         * @type {string}
         * @ojvalue {string} "full" In typical themes, full-chrome buttons always have chrome.
         * @ojvalue {string} "half" In typical themes, half-chrome buttons acquire chrome only in their hover, active, and selected states. Half-chroming is recommended for buttons in a toolbar.
         * (This is the toolbar default in most themes.)
         * @ojvalue {string} "outlined" In typical themes, outlined buttons are similar to half-chrome buttons, but have a border in the default state.
         * @ojshortdesc Indicates in what states the button has chrome (background and border).
         *
         * @example <caption>Initialize the Menu Button with the <code class="prettyprint">chroming</code> attribute specified:</caption>
         * &lt;oj-menu-button chroming='half'>&lt;/oj-menu-button>
         *
         * @example <caption>Get or set the <code class="prettyprint">chroming</code> property after initialization:</caption>
         * // getter
         * var chromingValue = myMenuButton.chroming;
         *
         * // setter
         * myMenuButton.chroming = 'half';
         *
         * @example <caption>Set the default in the theme (SCSS) :</caption>
         * $buttonChromingOptionDefault: half !default;
         */
        
        /**
         * <p>Indicates in what states the button has chrome (background and border).
         *
         * <p>The default chroming varies by theme and containership as follows:
         * <ul>
         *   <li>If the button is in a buttonset or toolbar, then the default chroming is the current <code class="prettyprint">chroming</code> value of the nearest such container.</li>
         *   <li>Else, if <code class="prettyprint">$buttonChromingOptionDefault</code> is set in the current theme as seen in the example below, then that value is the chroming default.</li>
         *   <li>Else, the default chroming is <code class="prettyprint">"full"</code>.</li>
         * </ul>
         *
         * <p>Once a value has been set on this button attribute, that value applies regardless of theme and containership.
         *
         * @expose
         * @memberof oj.ojButton
         * @instance
         * @since 1.2.0
         * @ojfragment buttonCommonChroming
         */
        chroming: "full",

        /**
         * {@ojinclude "name":"buttonCommonDisabled"}
         * 
         * @name disabled
         * @memberof oj.ojButton
         * @instance
         * @type {boolean}
         * @default false
         * @ojshortdesc Specifies that the button element should be disabled.
         *
         * @example <caption>Initialize the Button with the <code class="prettyprint">disabled</code> attribute specified:</caption>
         * &lt;oj-button disabled='true'>&lt;/oj-button>
         *
         * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
         * // getter
         * var disabledValue = myButton.disabled;
         *
         * // setter
         * myButton.disabled = true;
         */
        
        /**
         * {@ojinclude "name":"buttonCommonDisabled"}
         * 
         * @name disabled
         * @memberof oj.ojMenuButton
         * @instance
         * @type {boolean}
         * @default false
         * @ojshortdesc Specifies that the button element should be disabled.
         *
         * @example <caption>Initialize the Menu Button with the <code class="prettyprint">disabled</code> attribute specified:</caption>
         * &lt;oj-menu-button disabled='true'>&lt;/oj-menu-button>
         *
         * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
         * // getter
         * var disabledValue = myMenuButton.disabled;
         *
         * // setter
         * myMenuButton.disabled = true;
         */
                
        /**
         * <p>Disables the button if set to <code class="prettyprint">true</code>.
         *
         * <p>After create time, the <code class="prettyprint">disabled</code> state should be set via this API, not by setting the underlying DOM attribute.
         *
         * @expose
         * @memberof oj.ojButton
         * @instance
         * @ojfragment buttonCommonDisabled
         */
        disabled: false,

        /**
         * {@ojinclude "name":"buttonCommonDisplay"}
         * 
         * <p>For accessibility, a JET Button must always have a label set via the default slot, even if it is icon-only.
         * 
         * @name display
         * @memberof oj.ojButton
         * @instance
         * @type {string}
         * @ojvalue {string} "all" Display both the label and icons.
         * @ojvalue {string} "icons" Display only the icons.
         * @default "all"
         * @ojshortdesc Specifies whether the button displays label and icons, or just icons.
         *
         * @example <caption>Initialize the Button with the <code class="prettyprint">display</code> attribute specified:</caption>
         * &lt;oj-button display='icons'>&lt;/oj-button>
         *
         * @example <caption>Get or set the <code class="prettyprint">display</code> property after initialization:</caption>
         * // getter
         * var displayValue = myButton.display;
         *
         * // setter
         * myButton.display = 'icons';
         */
        
        /**
         * {@ojinclude "name":"buttonCommonDisplay"}
         * 
         * <p>For accessibility, a JET Menu Button must always have a label set via the default slot, even if it is icon-only.
         * 
         * @name display
         * @instance
         * @memberof oj.ojMenuButton
         * @type {string}
         * @ojvalue {string} "all" Display both the label and icons.
         * @ojvalue {string} "icons" Display only the icons.
         * @default "all"
         * @ojshortdesc Specifies whether the button displays label and icons, or just icons.
         *
         * @example <caption>Initialize the Menu Button with the <code class="prettyprint">display</code> attribute specified:</caption>
         * &lt;oj-menu-button display='icons'>&lt;/oj-menu-button>
         *
         * @example <caption>Get or set the <code class="prettyprint">display</code> property after initialization:</caption>
         * // getter
         * var displayValue = myMenuButton.display;
         *
         * // setter
         * myMenuButton.display = 'icons';
         */
        
        /**
         * <p>Whether to display both the label and icons (<code class="prettyprint">"all"</code>)
         * or just the icons (<code class="prettyprint">"icons"</code>).  In the latter case, the label is displayed in a tooltip instead, unless a
         * tooltip was already supplied at create time.
         *
         * @expose
         * @memberof oj.ojButton
         * @instance
         * @ojfragment buttonCommonDisplay
         */
        display: "all",

        /**
         * <p>Text to show in the button. The default is from the label in the DOM.
         *
         * <p>When not specified at create time, the element's HTML content is used, or its
         * <code class="prettyprint">value</code> attribute if the element is an input element of type button, submit, or reset, or
         * the HTML content of the associated label element if the element is an input of type radio or checkbox.
         *
         * <p>After create time, the label should be set via this API, not by modifying the underlying DOM.
         *
         * <p>The 2-way <code class="prettyprint">label</code> binding offered by the <code class="prettyprint">ojComponent</code>
         * binding should be used instead of Knockout's built-in <code class="prettyprint">text</code> binding, as the former
         * sets the API, while the latter sets the underlying DOM attribute.
         *
         * <p>Values set on this option, at create time or later, are treated as plain text, not HTML.  If the label is specified via
         * DOM at create time as described above, that HTML content is kept.
         *
         * <p>For accessibility, a JET Button must always have a label, even if it is <a href="#display">icon-only</a>.
         *
         * @expose
         * @memberof oj.ojButton
         * @instance
         * @type {?string}
         * @ignore
         * @example <caption>Initialize the button with the <code class="prettyprint">label</code> option specified:</caption>
         * $( ".selector" ).ojButton( { "label": "custom label" } );
         *
         * @example <caption>Get or set the <code class="prettyprint">label</code> option, after initialization:</caption>
         * // getter
         * var label = $( ".selector" ).ojButton( "option", "label" );
         *
         * // setter
         * $( ".selector" ).ojButton( "option", "label", "custom label" );
         */
        label: null,

        /**
         * <p>Icons to display in the button.  Support is as follows:
         * <ul>
         *   <li>Any combination of start and end icons can be specified, with or without the label (see <code class="prettyprint">display</code> option).</li>
         *   <li>Icons are supported for push buttons created from buttons and anchors, and for toggle buttons (radios and checkboxes).</li>
         *   <li>Icons are not supported for push buttons created from inputs of type button, submit, and reset.</li>
         * </ul>
         *
         * <p>The start icon is displayed before the label text (on the left in LTR), and the end icon is displayed after the
         * label (on the right in LTR).  In RTL, the positions are reversed.
         *
         * <p>The <code class="prettyprint">start</code> and <code class="prettyprint">end</code> properties accept one or more
         * style class names (as seen in the examples), or <code class="prettyprint">null</code>, indicating "no icon."
         * <p>The default start value is null and the default end values is
         * <code class="prettyprint">"oj-component-icon oj-button-menu-dropdown-icon"</code> if this is a menu button, and
         * <code class="prettyprint">null</code> otherwise.  See the <code class="prettyprint">menu</code> option.
         *
         * @ignore
         * @expose
         * @memberof oj.ojButton
         * @instance
         * @type {Object}
         *
         * @example <caption>Initialize the button, specifying both icons:</caption>
         * $( ".selector" ).ojButton({ "icons": { start: "demo-icon-font demo-bookmark-icon-16",
         *                                        end: "demo-icon-font demo-grid-icon-16" } });
         *
         * @example <caption>Get or set the <code class="prettyprint">icons</code> option, after initialization:</caption>
         * // getter
         * var icons = $( ".selector" ).ojButton( "option", "icons" );
         *
         * // setter, specifying both icons:
         * $( ".selector" ).ojButton( "option", "icons", { start: "oj-fwk-icon-caret-start oj-fwk-icon",
         *                                                 end: "demo-icon-font demo-grid-icon-16" } );
         */
        icons:
        {
            /**
             * <p>The start icon of the button.  See the top-level <code class="prettyprint">icons</code> option for details.
             *
             * @ignore
             * @expose
             * @alias icons.start
             * @memberof! oj.ojButton
             * @instance
             * @type {?string}
             * @default null
             *
             * @example <caption>Get or set the <code class="prettyprint">icons.start</code> sub-option, after initialization:</caption>
             * // getter
             * var startIcon = $( ".selector" ).ojButton( "option", "icons.start" );
             *
             * // setter:
             * $( ".selector" ).ojButton( "option", "icons.start", "oj-fwk-icon-caret-start oj-fwk-icon" );
             */
            start: null,
            /**
             * <p>The end icon of the button.  See the top-level <code class="prettyprint">icons</code> option for details.
             * The default is <code class="prettyprint">"oj-component-icon oj-button-menu-dropdown-icon"</code> if this is a menu button, and
             * <code class="prettyprint">null</code> otherwise.  See the <code class="prettyprint">menu</code> option.
             * @ignore
             * @expose
             * @alias icons.end
             * @memberof! oj.ojButton
             * @instance
             * @type {?string}
             *
             * @example <caption>Get or set the <code class="prettyprint">icons.end</code> sub-option, after initialization:</caption>
             * // getter
             * var startIcon = $( ".selector" ).ojButton( "option", "icons.end" );
             *
             * // setter:
             * $( ".selector" ).ojButton( "option", "icons.end", "demo-icon-font demo-grid-icon-16" );
             */
            end: null
        },

        /**
         * <p>Identifies the [JET Menu]{@link oj.ojMenu} that the button should launch. If specified, the button is a menu button.
         *
         * <p>The value can be an HTML element, JQ selector, JQ object, NodeList, or array of elements. In all cases, the first indicated element is used.
         *
         * <p>By default, menu buttons have a downward pointing "dropdown" arrow for their end icon.  See the <code class="prettyprint">icons</code> option for details.
         *
         * <p>Menu button functionality is supported for Buttons based on button or anchor tags.  (Buttons based on input tags either do not support the dropdown icon,
         * or do not make sense for use as a menu button, or both.)  Buttons are recommended over anchors, as anchor-based buttons are intended for use when native
         * anchor functionality such as href navigation is needed.
         *
         * <p>See [Menu's]{@link oj.ojMenu} Accessibility section for a discussion of how <code class="prettyprint">aria-label</code> and
         * <code class="prettyprint">aria-labelledby</code> are handled for menu buttons and other menu launchers.
         *
         * @expose
         * @memberof oj.ojButton
         * @instance
         * @type {Element|Array.<Element>|string|jQuery|NodeList}
         * @default null
         * @ignore
         * @ojtsignore
         * @example <caption>Initialize a menu button:</caption>
         * $( ".selector" ).ojButton({ "menu": "#myMenu" });
         *
         * @example <caption>Get or set the <code class="prettyprint">menu</code> option, after initialization:</caption>
         * // getter
         * var menu = $( ".selector" ).ojButton( "option", "menu" );
         *
         * // setter
         * $( ".selector" ).ojButton( "option", "menu", ".my-marker-class" );
         */
        menu: null,

        // Events
        /**
         * <p>Triggered when a button is clicked. This will be triggered by keyboard events as well as mouse/touch events.
         *
         * <p>To ensure keyboard accessibility, the only correct, supported way to react to the click of a button is to listen
         * for this event. Click listeners and <code class="prettyprint">href</code> navigation should not be used.
         *
         * @expose
         * @event
         * @memberof oj.ojButton
         * @instance
         * @ojcancelable
         * @ojbubbles
         * @since 5.0.0
         */
        action: null
    },

    _InitOptions: function(originalDefaults, constructorOptions) {
        this._super(originalDefaults, constructorOptions);

        // if custom element we are supporting slots
        if (this._IsCustomElement())
        {
            this._processSlots();
        }

        this._initButtonTypes(); // init this.type and this.buttonElement, used below

        // At create time, disabled and label can be set via either option or DOM.
        // If app set the option, then that wins over the DOM, in which case _ComponentCreate() will later set that value on the DOM.
        // Else DOM wins, in which case we set the option from the DOM here, with any remaining tasks done later in _ComponentCreate().

        if (!('disabled' in constructorOptions)) // if app didn't set option, then set the option from the DOM
        {   // For anchors, this line always sets disabled option to false.  (Neither JQUI nor JET look for the .oj-disabled class for anchors,
            // so the only way to disable an anchor button at create time is via the API.  At refresh time, JQUI did look
            // for the .oj-disabled class, but our refresh doesn't handle disabled.)
            this.option('disabled',
                        !!this.element.prop( "disabled" ),
                        {'_context': {internalSet: true}}); // writeback not needed since "not in constructorOptions" means "not bound"
        }

        if (!('label' in constructorOptions)) // if app didn't set option, then set the option from the DOM
        {
            this.keepDomLabel = true;
            this.option('label',
                        this.type === "inputPush" ? this.buttonElement.val() : this.buttonElement.html(), // @HTMLUpdateOK getter not setter
                        {'_context': {internalSet: true}}); // writeback not needed since "not in constructorOptions" means "not bound"
        }

        // if this is a menuButton and app didn't set icons.end to its own icon or to null to suppress the icon, then default to built-in menuButton dropdown icon
        if (!this._IsCustomElement())
        {
            if (this.options.menu && (!constructorOptions.icons || constructorOptions.icons.end === undefined))
            {
                this.option('icons.end',
                            "oj-component-icon oj-button-menu-dropdown-icon",
                            {'_context': {writeback: true, internalSet: true}});
            }
        }
    },

    _ComponentCreate: function() {
        this._super();

        // facilitate removing menuButton handlers separately, if app sets/clears the "menu" option
        this.menuEventNamespace = this.eventNamespace + "menu";

        this._initButtonTypes2();
        this.hasTitle = !!this.rootElement.attr( "title" );

        var self = this,
            options = this.options,
            toggleButton = this._isToggle,
            activeClass = !toggleButton ? "oj-active" : "";

        this.rootElement.addClass( BASE_CLASSES );
        _setChromingClass(this.rootElement, this.options.chroming);

        // Called for touchend/cancel on both button and document.  Listening only on button isn't completely reliable
        // on at least iOS and Android since the touchend can happen slightly off of the button.  Listening only on the
        // document runs the risk that we won't hear it because someone eats it.  Could use capture listener to dodge
        // that risk, but just listening on both seems to work great.
        var endHandler = function() {
            self.rootElement.removeClass( activeClass );
            self.rootElement.removeClass( "oj-hover" );
            self._toggleDefaultClasses();
        };

        this.buttonElement
            .bind( "touchstart" + this.eventNamespace, function() {
                if ( self._IsEffectivelyDisabled() )
                    return;

                self.rootElement.addClass( activeClass );
                self._toggleDefaultClasses();

                // don't pass "touchend touchcancel", due to semantics of one() : it's called once per event type.
                // It's almost always touchend, not touchcancel, that is fired, so the touchend listeners would pile up.
                // The likelihood is very small that the double edge case would occur where both endHandler is needed,
                // AND the touch ends with touchcancel rather than touchend, and the result would only be that the hover
                // style sticks to the button.
                self.document.one("touchend", endHandler);
            })
            .bind( "touchend" + this.eventNamespace + " " + "touchcancel" + this.eventNamespace, endHandler)
            .bind( "mouseenter" + this.eventNamespace, function() {
                if ( self._IsEffectivelyDisabled() )
                    return;

                // do this for real mouse enters, but not 300ms after a tap
                if (!oj.DomUtils.recentTouchEnd()) {
                    if ( this === _lastActive )
                        self.rootElement.addClass( "oj-active" );

                    self.rootElement.addClass( "oj-hover" )
                                    .removeClass( "oj-default oj-focus-only" );
               }
            })
            .bind( "mouseleave" + this.eventNamespace, function() {
                self.rootElement.removeClass( "oj-hover" );

                if ( self._IsEffectivelyDisabled() )
                    return;
                self.rootElement.removeClass( activeClass );
                self._toggleDefaultClasses();
            });

        this._disabledClickHandler = function( event ) {
            if ( self._IsEffectivelyDisabled() )
            {
                event.preventDefault();
                event.stopImmediatePropagation();
            }
        };
        
        this._ojActionClickHandler = function(event)
        {
            self._trigger('action', event, {});
        };

        // Must do this in capture phase to avoid race condition where app's click
        // handlers on anchor buttons can be called if their listeners get registered
        // before ours, e.g. if their KO click binding is before the ojComponent binding.
        if (this._IsCustomElement())
        {
            this.rootElement[0].addEventListener("click", this._disabledClickHandler, true);
            this.rootElement[0].addEventListener("click", this._ojActionClickHandler, false);
        }
        else
        {
            this.buttonElement[0].addEventListener("click", this._disabledClickHandler, true);
        }
        
        this._focusable({
            'element': this.rootElement,
            'applyHighlight': true,
            'afterToggle' : function() {
                self._toggleDefaultClasses();
            }
        });

        if ( toggleButton )
        {
            this.element.bind( "change" + this.eventNamespace, function( event ) {
                self._applyCheckedStateFromDom(true); // we just get one change event for entire radio group, even though up to 2 changed, so must refresh entire radio group, not just this button

                // if in a buttonset that tracks checked state (i.e. checkbox set or single
                // radio group), then set that option and fire optionChange event
                var buttonset = self._getEnclosingContainerComponent("buttonset");
                var checkedState = buttonset && buttonset._getCheckedFromDom(buttonset.$buttons);
                if (buttonset && checkedState!==undefined)
                {
                    buttonset.option("checked", checkedState,  {'_context': {writeback: true, originalEvent: event, internalSet: true}});
                }
            });

            // Required for FF, where click-drag on checkbox/radio btn's label(JET decorates label as button for toggle buttons )
            // will not yield in to click event and also trasfers focus out of the <input> element and hence breaks arrow navigation.
            // To fix, If mouse moves between mouseDown/mouseUp (drag) with in the boundaries of button then focus should be set
            // on the button which will ensure proper arrow key navigation(see  for more details).
            this.buttonElement
                .bind("mousedown" + this.eventNamespace, function(event) {
                    if (self._IsEffectivelyDisabled())
                        return;
                    _lastToggleActive = this;
                    self.document.one("mouseup", function() {
                        _lastToggleActive = null;
                    });
                })
                .bind("mouseup" + this.eventNamespace, function(event) {
                    if (self._IsEffectivelyDisabled())
                        return;
                    if (this === _lastToggleActive) {
                        self.element.focus();
                    }
                });
        }

        if ( this.type === "checkbox" )
        {
            this.buttonElement.bind( "click" + this.eventNamespace, function(event) {
                if ( self._IsEffectivelyDisabled() )
                    return false;
            });

            // role="button" requires both Space and Enter support, but radios and checkboxes natively support only Space, so add Enter.
            // Update: now that we're not using role=button for checkboxes, we don't strictly need this.
            // This fire-click-on-Enter logic gives Enter the same behavior as Space for checkboxes in Chrome, FF, and IE9:
            // In Chrome28 and IE9, for Space and Enter on Checkboxes, first the "checked" value updates, then change event, then click event.
            // In FF22, for Space and Enter on Checkboxes, first the "checked" value updates, then click event, then change event.
            // Unlike the radio Enter handler, we get this good behavior by only firing "click".
            this.element.bind( "keyup" + this.eventNamespace, function(event) {
                if ( event.keyCode === $.ui.keyCode.ENTER ) {
                    if ( !self._IsEffectivelyDisabled() ) {
                        //console.log("checkbox Enter handler firing click event");
                        self.element.click();
                    }
                }
            });
        } else if ( this.type === "radio" )
        {
            this.buttonElement.bind( "click" + this.eventNamespace, function() {
                if ( self._IsEffectivelyDisabled() )
                {
                    return false;
                }
            });

            // role="button" requires both Space and Enter support, but radios and checkboxes natively support only Space, so add Enter.
            // Update: now that we're not using role=button for radios, we don't strictly need this.
            // For radios, this handler gives Enter in Chrome28/ IE9/ FF22 the same behavior as Space on Chrome.  Since Space in IE/FF is
            // different than Space in Chrome, this means that Enter and Space are not quite the same within those browsers.
            // For Space and Enter on unchecked radios in Chrome28 and IE9:
            //   - first the "Checked" value updates, then change event, then click event.
            // For already-checked radios, where activation would just check it again:
            //   - Space and Enter in Chrome28, and Enter in IE9:  do nothing.
            //   - Space in IE9:  first the "Checked" value updates, then click event.  (No change event since nothing changed.)
            //     (This is the only Chrome/IE9 Space/Enter difference.)
            // In FF22:
            //   - Space on unchecked radios: first the "Checked" value updates, then click event, then change event.
            //   - Enter on unchecked radios is same as Chrome.  (Different event ordering than FF.)
            //   - Space on checked radios is like IE9: first the "Checked" value updates, then click event.  (No change event since nothing changed.)
            //   - Enter on checked radios is same as Chrome.  (Do-nothing.)
            this.element.bind( "keyup" + this.eventNamespace, function(event) {
                if ( event.keyCode === $.ui.keyCode.ENTER )
                {
                    if ( !self.element[0].checked && !self._IsEffectivelyDisabled() ) {
                        //console.log("radio Enter handler found radio unchecked, so checking it and firing click event");

                        // Unlike the checkbox Enter handler, we set checked=true and fire changed before firing click, without which
                        // we didn't get the exact behavior described above.  Without setting checked, the click handlers in all 3 desktop
                        // browsers saw the "old" "checked" value when Enter was pressed on radios, which broke Buttonset's focus mgmt logic.
                        // If we set checked, then change no longer fires automatically, in at least some browsers, so we have to fire it
                        // manually, both for apps that rely on it, and because our code relies on our change listener to update the "checked" styling.
                        self.element[0].checked = true;
                        self.element.change();
                        self.element.click();
                    }
                }
            });
        } else // neither checkbox nor radio, so not a toggle button, so element, buttonElement (and rootElement in JQUI) are all the same node
        {
            this.buttonElement
                .bind( "mousedown" + this.eventNamespace, function( event ) {
                    if ( self._IsEffectivelyDisabled() ) {
                        return false;
                    }

                    // don't show active/pressed-down state unless left mouse button, since only that button will click the button after mouseup
                    // do this for real mousedowns, but not 300ms after a tap
                    if ( event.which === 1 && !oj.DomUtils.recentTouchEnd() )
                    {
                        self.rootElement.addClass( "oj-active" )
                                        .removeClass( "oj-default oj-focus-only" );
                        _lastActive = this;
                        self.document.one( "mouseup", function() { // TODO: prob need capture listener like Menu for reliability
                            _lastActive = null;
                        });
                    }
                })
                .bind( "mouseup" + this.eventNamespace, function() {
                    if ( self._IsEffectivelyDisabled() )
                        return false;
                    self.rootElement.removeClass( "oj-active" );
                    self._toggleDefaultClasses();
                })
                .bind( "keydown" + this.eventNamespace, function(event) {
                    if ( self._IsEffectivelyDisabled() )
                        // ...then bail out always, also eating event unless key is Tab or left/right arrow, since:
                        // - Must allow Tab so KB user can't get stuck here.
                        // - Nice to allow Buttonset/Toolbar's left/right arrow handling too, but not strictly essential as long as user
                        //   can Tab out and back in, since (if app refreshed Buttonset/Toolbar after disabling button as required), the
                        //   tab-back-in will go to an enabled button of the Buttonset/Toolbar, or skip Buttonset/Toolbar if all buttons disabled.
                        // - Must eat Enter/Space/DownArrow to prevent that functionality from occurring. (For non-anchor buttons, the native
                        //   disabled status prevents some of those on at least some platforms.)
                        // Since anchor buttons don't have a native disabled status, they remain focusable when disabled, and thus are most
                        // susceptible to having key events while disabled. (We ensure they are not tabbable, but they remain focusable)
                        return event.keyCode === $.ui.keyCode.TAB || event.keyCode === $.ui.keyCode.LEFT || event.keyCode === $.ui.keyCode.RIGHT;

                    var isSpace = event.keyCode === $.ui.keyCode.SPACE;
                    var isAnchor = self.type === "anchor";

                    // now that anchor doesn't support Space, still keep this line, in case users try to click via Space
                    if ( isAnchor && isSpace ) {
                        event.preventDefault(); // prevent scrolling down one page when clicking anchor button via Spacebar.  Only prevent for anchor!
                    }
                    if ( (isSpace && !isAnchor) || event.keyCode === $.ui.keyCode.ENTER ) {
                        self.rootElement.addClass( "oj-active" )
                                        .removeClass( "oj-default oj-focus-only" );
                    }
                })
                // see #8559, we bind to blur here in case the button element loses
                // focus between keydown and keyup, it would be left in an "active" state
                .bind( "keyup" + this.eventNamespace + " blur" + this.eventNamespace, function() {
                    self.rootElement.removeClass( "oj-active" );
                    self._toggleDefaultClasses();
                });

            if ( this.type === "anchor" )
            {
                //Chrome is not updating document.activeElement on click of <a> which is needed for ojPopup and
                //setting tabIndex to a non-negative value will fix this. Refer 
                // isNaN(..) returns true when tabIndex is undefined (or 'undefined', which the original impl checked for...)
                var tabIndex = this.buttonElement.attr( "tabindex" );
                if( tabIndex === null || isNaN(tabIndex) ) // Don't override if user already set a tabIndex.
                {
                    this.buttonElement.attr( "tabindex", "0" );
                }
            }
        } // end of: if (checkbox) {} else if (radio) {} else {}

        // at create time, we want only the "if disabled" part of callee, not the "if enabled" part, so only call if disabled
        if (this.options.disabled)
            this._manageAnchorTabIndex(false, true);

        this._updateEffectivelyDisabled();

        var buttonset = self._getEnclosingContainerComponent("buttonset");
        if (this._IsCustomElement() || (buttonset && buttonset._IsCustomElement()))
        {
            this._setDisplayOptionOnDom();
        }
        else
        {
            this._handleLabelAndIconsAtCreateTime();
        }

        this._setupMenuButton(null);

        // call this at the *end* of _ComponentCreate, since it needs to know whether any state classes like oj-active, oj-disabled, oj-selected, oj-hover, .oj-focus
        // have been applied.
        this._toggleDefaultClasses();
    },

    _NotifyContextMenuGesture: function(menu, event, eventType)
    {
        // For toggle buttons, launcher must be the hidden focusable input, but for Shift-F10 we want the CM aligned to the root element, not that
        // launcher.  rootElement works for push buttons too.
        this._OpenContextMenu(event, eventType, {
            "position": {"of": eventType==="keyboard" ? this.rootElement : event}
        });
    },

    _processSlots: function() {
        var self = this;
        var elem = this.element[0];
        var rootElem = elem.parentNode;
        var isMenuButton = rootElem.tagName === 'OJ-MENU-BUTTON';

        // we are responsible for reparenting any slots because of use of inner elem
        var rootSlots = oj.BaseCustomElementBridge.getSlotMap(rootElem);

        // don't add context menu here since we do not ever need to move it
        if (isMenuButton)
            var supportedSlots = ["startIcon", "", "endIcon", "menu"];
        else
            supportedSlots = ["startIcon", "", "endIcon"];

        // move all desired slots down one level
        $.each(supportedSlots, function(i, slotName) {
            if (rootSlots[slotName] && slotName != "") {
                $.each(rootSlots[slotName], function(i, node) {
                    elem.appendChild(node);
                })
            }
        });

        // be sure to not remove the context menu slot
        $(rootElem).children().not(elem).not("[slot='contextMenu']").remove();
        
        // all slots are now within the inner button element
        var slots = oj.BaseCustomElementBridge.getSlotMap(elem);

        // rearrange slots
        $.each(supportedSlots, function(i, slotName) {
            if (slots[slotName]) {
                $.each(slots[slotName], function(i, node) {
                    elem.appendChild(node);
                    if (slotName === "")
                    {
                        var currentText = node;
                        var wrapperSpan = currentText;
                        if (currentText.nodeType === 3)
                        {
                            wrapperSpan = document.createElement('span');
                            $(currentText).replaceWith(wrapperSpan);  // @HTMLUpdateOK
                            wrapperSpan.appendChild(currentText);
                        }
                        $(wrapperSpan).addClass('oj-button-text');
                        self._setTextSpanIdAndLabelledBy($(wrapperSpan));
                    }
                    else if (slotName === "startIcon")
                    {
                        $(node).addClass('oj-button-icon');
                        $(node).addClass('oj-start');
                    }
                    else if (slotName === "endIcon")
                    {
                        $(node).addClass('oj-button-icon');
                        $(node).addClass('oj-end');
                    }
                    else if (slotName === "menu")
                    {
                        $(node).uniqueId();
                        self.menuSlot = '#' + node.id;
                        if (slots["endIcon"] === undefined)
                        {
                            var dropDownElem = document.createElement('span');
                            dropDownElem.className = 'oj-button-icon oj-end oj-component-icon oj-button-menu-dropdown-icon';
                            dropDownElem['slot'] = 'endIcon';
                            elem.insertBefore(dropDownElem, node);  // @HTMLUpdateOK
                        }
                    }
                })
            }
        });

        // wrap button content in a label for consistent DOM/theming structure
        var wrapperDiv = $('<div>');
        wrapperDiv.addClass('oj-button-label');
        $(elem).children().wrapAll(wrapperDiv);  // @HTMLUpdateOK
    },
    
    // Helper function to return the correct menu reference between custom element and non custom element components.
    _getMenuNode: function() // Private, not an override (not in base class). Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if (this._IsCustomElement())
            return this.menuSlot;
        else
            return this.options.menu;
    },

    // Part 1 of type-specific component init.  Called from _InitOptions, so very limited component state is available!
    _initButtonTypes: function() // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        // for toggle buttons (radio/checkbox), element is <input>, buttonElement is <label>,
        // and rootElement is a new wrapper element we create.  This is true in JQUI; in CustomElements not exposing toggle buttons.
        if ( this.element.is("input[type=checkbox]") ) {
            this.type = "checkbox";
            this._isToggle = true;
        } else if ( this.element.is("input[type=radio]") ) {
            this.type = "radio";
            this._isToggle = true;
        }

        // for push buttons (next 3 cases), element, buttonElement, and rootElement are all the same elem in JQUI.
        // In CustomElements, element and buttonElement are the same elem, and rootElement is the parent custom element.
        // We ignore <label> if present.
        else if ( this.element.is("input[type=button],input[type=submit],input[type=reset]") )
            this.type = "inputPush";
        else if ( this.element.is("button"))
            this.type = "button";
        else if ( this.element.is("a"))
            this.type = "anchor";
        else
            throw new Error("JET Button not supported on this element type");

        if ( this._isToggle )
        {
            // TBD: rather than requiring the label to be supplied, should we just create one for them if it's not there?
            var labelSelector = "label[for='" + this.element.attr("id") + "']";
            this.buttonElement = this.element.siblings().filter( labelSelector );
        } else {
            this.buttonElement = this.element;
        }
    },

    // Part 2 of type-specific component init, called from _ComponentCreate().
    _initButtonTypes2: function() // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if ( this._isToggle )
        {
            this.buttonElement.addClass( "oj-button-label" );

            this.element
                .addClass( "oj-button-input oj-helper-hidden-accessible" )
                .attr("data-oj-internal", "") // automation support
                .add(this.buttonElement) // doesn't mutate this.element
                .wrapAll("<span></span>"); // add root node around label/input.  @HTMLUpdateOK trusted string

            this.rootElement = this.element.parent(); // the new root
            this.rootElement.addClass( "oj-button-jqui oj-button-toggle" );

            var checked = this.element[0].checked;
            if ( checked ) {
                this.rootElement.addClass( "oj-selected" )
                                .removeClass( "oj-default oj-focus-only" );
            }
            // else no need to removeClass since this code runs only at create time

        } else
        {
            if (this._IsCustomElement()) {
                this.rootElement = this.element.parent();
                this.element.addClass("oj-button-button");
            } else {
                this.rootElement = this.element;
                this.element.addClass("oj-button-jqui");
            }
        }
    },

    /**
     * Returns a <code class="prettyprint">jQuery</code> object containing the root element of the Button component.
     * @ignore
     * @expose
     * @memberof oj.ojButton
     * @instance
     * @return {jQuery} the root element of the component
     *
     * @example <caption>Invoke the <code class="prettyprint">widget</code> method:</caption>
     * var widget = $( ".selector" ).ojButton( "widget" );
     */
    widget: function() // Override of public base class method.  Method name needn't be quoted since is in externs.js.
    {
        return this.rootElement;
    },

    _destroy: function() // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
    {
        this._removeMenuBehavior(this._getMenuNode());
        this.buttonElement[0].removeEventListener("click", this._disabledClickHandler, true);
        this.buttonElement[0].removeEventListener("click", this._ojActionClickHandler, false);

        // TBD: won't need this after the restore-attrs feature is in place.
        this.element
            .removeClass( "oj-helper-hidden-accessible" )
            .removeAttr("aria-labelledby")
            .removeUniqueId();

        // If disabled, we want to run the "changing from disabled to enabled" part of callee, to restore original tabindex.
        // If enabled, don't want to run any part of callee.
        if (this.options.disabled)
            this._manageAnchorTabIndex(true, false);

        var isToggle = this._isToggle;

        // TBD: won't need this after the restore-attrs feature is in place.
        if ( !isToggle )
            this.rootElement.removeClass( BASE_CLASSES + " oj-button-jqui " + STATE_CLASSES + " " + TYPE_CLASSES + " " + CHROMING_CLASSES );

        this.buttonElement.html( this.buttonElement.find(".oj-button-text").html() ); // @HTMLUpdateOK reparent existing DOM

        if ( !isToggle )
        {
            // TBD: won't need this after the restore-attrs feature is in place.
            if ( !this.hasTitle )
                this.rootElement.removeAttr( "title" );
        } else
        {
            this.buttonElement.removeClass( "oj-button-label" ); // TBD: won't need this after the restore-attrs feature is in place.

            // : If the button is stamped out by a KO foreach (with or without a containing buttonset), and the foreach
            // observable is updated to no longer include the button, then _destroy() is called.  Due to the ordering of tasks, if
            // JQ's unwrap() is called directly, the nodes winds up back into the DOM.  To avoid this, we previously successfully
            // put the JQ unwrap() call in a setTimeout(0), and now we do the following:
            //
            //  - DomUtils.unwrap() will avoid unwrapping if the node is being destroyed by Knockout
            oj.DomUtils.unwrap(this.element);
        }

        if( _lastToggleActive === this.buttonElement[0] )
        {
            _lastToggleActive = null; //clear _lastToggleActive flag, while destroying the button.
        }
    },

    _NotifyDetached: function() {
        // In browsers [Chrome v35, Firefox v24.5, IE9, Safari v6.1.4], blur and mouseleave events are generated for hidden content but not detached content,
        // so when the component is detached from the document, we must use this hook to remove the oj-focus, oj-focus-highlight, oj-hover, and oj-active classes, to
        // ensure that button is displayed without those classes when it is re-attached to the DOM. Refer .
        // _super() now removes those 3 classes, so just need to call _tDC() afterwards.
        this._super();
        this._toggleDefaultClasses();
    },

    __setAncestorComponentDisabled: function( disabled )
    {
        this._super( disabled ); // sets this._ancestorDisabled
        this._updateEffectivelyDisabled();
    },

    // Prereq:  this.options.disabled and this._ancestorDisabled must already be set to their updated values before calling this method.
    //
    // When called by __setAncestorComponentDisabled(), it has already set this._ancestorDisabled.
    //
    // When called by _setOption("disabled"), its _super() has already done the following things:
    // - Set this.options.disabled.
    // - Applied .oj-disabled and aria-disabled to rootElement, often incorrectly per comments below.  The below code fixes it up.  The _super() code has tbd to fix this.
    // - If option is being set to true, it's removed .oj-hover/focus/focus-highlight/active.  (See comment below.)
    _updateEffectivelyDisabled: function()
    {
        var effectivelyDisabled = this._IsEffectivelyDisabled();

        // Unlike JQUI, root element should have exactly one of .oj-enabled and .oj-disabled at any point in time, for all flavors of Button.
        // _setOption._super() sets .oj-disabled to potentially wrong value since it doesn't know about "effectively disabled".  This fixes it up.
        this.rootElement.toggleClass( "oj-disabled", effectivelyDisabled );
        this.rootElement.toggleClass( "oj-enabled", !effectivelyDisabled ); // _setOption._super() doesn't try to set this

        if (this.type !== "anchor") // i.e. <button> or <input> (including type=radio|checkbox|other)
        {
            // <button> and <input> (including type=radio|checkbox|other) have this property, but <a> doesn't
            this.element.prop( "disabled", effectivelyDisabled ); // JQUI's _setOption sets this for <a>'s too, which seems harmless but is incorrect.

            // _setOption._super() puts aria-disabled on the rootElement.  Per A11y team, don't put aria-disabled on element already having disabled
            // attr.  (And if we DID apply aria-disabled, for radios/checkboxes it should go on the element / input, not the buttonElement / label or rootElement,
            // so the _setOption._super() behavior used by JQUI button is doubly wrong.)  Further, _setOption._super() can set it wrong since it doesn't know
            // about "effectively disabled".  This fixes it up.
            this.rootElement.removeAttr( "aria-disabled" );
        } else { // else is <a>
            // _setOption._super() puts aria-disabled on the rootElement. For <a>'s, element and rootElement are both the <a> in JQUI,
            // but are different in CustomElements.  aria-disabled belongs on element.  Also, the value it sets is potentially wrong since it
            // doesn't know about "effectively disabled".
            this.element.attr( "aria-disabled", effectivelyDisabled ); // set element attr to correct value in both JQUI and Custom Elements.
            if (this._IsCustomElement()) this.rootElement.removeAttr( "aria-disabled" );
        }

        if (effectivelyDisabled)
        {
            // TBD: when the handling of oj-active in baseComponent._setOption("disabled") is finalized, review whether this should be handled there instead.
            // baseComponent._setOption("disabled") removes oj-active, oj-hover, oj-focus, and oj-focus-highlight, but _updateEffectivelyDisabled is called in a number of other
            // cases too, so do it here too to be safe.
            this.widget().removeClass("oj-active oj-default oj-focus-only oj-hover oj-focus oj-focus-highlight");
            _lastActive = null; // avoid (very slight) possibility that first mouseIn after button is subsequently re-enabled will set oj-active

            // when disabling a menu button, dismiss the menu if open
            this._dismissMenu(this._getMenuNode());
        }
        else
        {
            this._toggleDefaultClasses();
        }
    },

    _setOption: function( key, value, flags ) // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
    {
        var oldValue = this.options[key];
        this._super( key, value, flags );
        // TBD: Currently the below code relies on super already having been called.  Consider removing that dependency
        // and calling super at end instead, so that optionChange (fired at end of super) is fired at very end.

        switch (key)
        {
            case "chroming":
                _setChromingClass(this.rootElement, value);
                break;
            case "disabled":
                // call this from _setOption, not _updateEffectivelyDisabled, as discussed in callee
                this._manageAnchorTabIndex(oldValue, value);

                // must call this *after* _super(), as discussed in callee
                this._updateEffectivelyDisabled();
                break;
            case "label":
                this._setLabelOption();
                break;
            case "display":
                if ( this.type !== "inputPush" ) // <input type=button|submit|reset> doesn't support child nodes, thus no icons, icon-only buttons, etc.
                    this._setDisplayOptionOnDom();
                break;
            case "icons":
                this._setIconsOption(true);
                break;
            case "menu": // setting/clearing the menu sets whether this is a menuButton
                this._setupMenuButton(oldValue);
                break;
        }
    },
    
    // @inheritdoc
    refresh: function() // Override of public base class method (unlike JQUI).  Method name needn't be quoted since is in externs.js.
    {
        this._super();

        // TODO:
        // - JSDoc says to call refresh() when Button reparented.  Instead, app should call oj.Components.subtreeAttached(), and our
        //   _NotifyAttached override should do the handling.
        // - There are other things we should do in the "no longer in buttonset" case, like removing buttonset listeners.  Anything else too?
        // - The reason the jsdoc doesn't mention the "moved into Buttonset (possibly from another buttonset)" case is that in that case
        //   Bset.refresh() must be called.  However, not sure that it's doing all the necessary things for the "from another Bset" case.

        // Handle the rare case where we just got reparented out of a disabled Buttonset
        if ( this._ancestorDisabled && !this._getEnclosingContainerElement("buttonset").length )
            this.__setAncestorComponentDisabled(false);

        // re-fetch the chroming option, so that if it's still set to the default dynamic getter, which takes its value from the containing
        // buttonset or toolbar if present, refresh() will update the visible chroming.  This is needed for cases like the following, all of
        // which call button.refresh() :
        // - The button previously wasn't in a buttonset or toolbar, but now is, or vice versa.  This is common at init time, when the bset/
        //   toolbar is created after their buttons, in which case the bset/toolbar refreshes their buttons.  Can also happen due to reparenting
        //   the button into or out of the bset/toolbar. "into": app must refresh bset/toolbar, which refreshes their buttons.  "out of": app
        //   must refresh the button.
        // - The app sets the chroming option of the containing buttonset or toolbar, which refreshes their buttons.
        // Do this after the _super() call, which updates the list of containers (buttonset/toolbar) that the component is in.
        _setChromingClass(this.rootElement, this.options.chroming);
    },

    // If this button is radio/checkbox, then this method gets the checked state from DOM's
    // "checked" prop and toggles oj-selected accordingly.
    // If wholeGroup param is true, and this button is a radio, then do that for all buttons in the group.
    _applyCheckedStateFromDom: function(wholeGroup) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if ( this.type === "radio" )
        {
            (wholeGroup ? _radioGroup( this.element[0] ) : this.element ).each(function() {
                var $radioWidget = $( this ).data( "oj-ojButton" ); // definitely exists because _radioGroup() checks for :oj-button

                if ( this.checked )
                {
                    $radioWidget.rootElement.addClass( "oj-selected" )
                                            .removeClass( "oj-default oj-focus-only" );
                } else
                {
                    $radioWidget.rootElement.removeClass( "oj-selected" );
                    $radioWidget._toggleDefaultClasses();
                }
            });
        } else if ( this.type === "checkbox" )
        {
            if ( this.element[0].checked )
            {
                this.rootElement.addClass( "oj-selected" )
                                .removeClass( "oj-default oj-focus-only" );
            } else
            {
                this.rootElement.removeClass( "oj-selected" );
                this._toggleDefaultClasses();
            }
        }
    },

    /*
     * Method name sums it up.  Should only be called at create time.
     */
    _handleLabelAndIconsAtCreateTime: function() // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if ( this.type === "inputPush" ) { // <input type=button|submit|reset> doesn't support child nodes, thus no icons, etc.
            this._setLabelOnDomOfSpanlessButton();
        } else { // <button>, <a>, checkboxes, and radios
            var textSpan = this._setLabelOnDomAtCreateTime();
            var hasStartIcon = this._setIconOnDom(true);
            var hasEndIcon = this._setIconOnDom(false);

            this._setDisplayOptionOnDom(textSpan, hasStartIcon, hasEndIcon);
        }
    },

    /*
     * Replace the button contents with a span containing the label:
     * - If app didn't set the label option, then reparent their DOM label to the new span.
     * - If app set the label option, then set escaped version onto label span.
     * Either way, button contents are completely replaced with the new span.
     *
     * Should only be called at create time.
     *
     * return the new textSpan
     */
    _setLabelOnDomAtCreateTime: function() // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        var buttonElement = this.buttonElement;
        var textSpan = $( "<span class='oj-button-text'></span>", this.document[0] );

        if (this.keepDomLabel) {
            textSpan.append(buttonElement.contents()); // @HTMLUpdateOK reparent existing DOM
        } else {
            buttonElement.empty();
            textSpan.text(this.options.label); // performs escaping; e.g. if label is <a>foo</a>, then text() replaces the span's contents with a text node containing that literal string (rather than setting innerHtml).
        }

        // Due to FF bug (see Bugzilla's Bug 984869), <button> with flex/inline-flex display doesn't work. Workaround by wrapping <button> contents with a <div> and setting the latter display flex/inline-flex
        if (this.type === "button") {
            var contentContainer = $("<div></div>").addClass("oj-button-label");
            contentContainer.append(textSpan); // @HTMLUpdateOK append span containing trusted content and existing DOM, per above comments on lines referencing textSpan.
            this.element.append(contentContainer); // @HTMLUpdateOK attach detached wrapped DOM created from trusted string and existing DOM
        } else {
            buttonElement.append(textSpan); // @HTMLUpdateOK attach detached DOM created from trusted string and existing DOM
        }

        // Need to set "aria-labelledby" attribute of (button/anchor) element to point to label span as fix for  (accessibility: icon-only button label is read twice by screen reader)
        // This is only a problem for <button> and <a> at the time of writing, so the fix is only applied to these two button types.
        if ((this.type === "button" || this.type === "anchor") && !this.element[0].hasAttribute('aria-label')  && !this.element[0].hasAttribute('aria-labelledby') ) {
            this._setTextSpanIdAndLabelledBy(textSpan);
        }
        return textSpan;
    },

    _setTextSpanIdAndLabelledBy: function(textSpan)
    {
        textSpan.uniqueId(); // assign id so that this.element can have "aria-labelledby" attribute pointing to the textspan
        this.element.attr("aria-labelledby", textSpan.attr("id"));
    },

    /*
     * Following is an old TBD from before the refactoring.  Now that we've sharply reduced the amount of unnecessary DOM replacement when
     * options like label and icons are set, is this still an issue?  If so, can we further minimize the churn to fix it?
     * TBD: Per http://www.quirksmode.org/blog/archives/2007/01/a_note_about_ev.html, if a click handler blows away the thing that was
     * clicked, it stops the click event from bubbling any further.  For antonym buttons like Play/Pause where the app puts a click handler
     * on the button that toggles its label and/or icons, this method replaces the clicked element, e.g. a <span> inside the <button>,
     * stopping the bubble. [UPDATE: we keep the span now but change its contents.  Does that fix it?].  This makes bubbling brittle.
     * Consider improving and/or doc'ing the issue.
     */

    /*
     * This method takes care of everything that needs to happen when the "label" option is set *after* create time.
     */
    _setLabelOption: function() // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if ( this.type === "inputPush" ) { // <input type=button|submit|reset> doesn't support child nodes, thus no icons, etc.
            this._setLabelOnDomOfSpanlessButton();
        } else {
            var textSpan = this.buttonElement.find( '.oj-button-text' );
            textSpan.text(this.options.label); // performs escaping; e.g. if label is <a>foo</a>, then text() replaces the span's contents with a text node containing that literal string (rather than setting innerHtml).
            this._setDisplayOptionOnDom(textSpan);
        }
    },

    /*
     * This method should only be called when the button is a "spanless" button, i.e. <input type=button|submit|reset>. It is called when the label is set,
     * both at create time and when it is set later.
     */
    _setLabelOnDomOfSpanlessButton: function() // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        // TBD: The "if label" check is from JQUI.  Is there some reason that setting "" should be ignored?
        // Probably not too harmful since they should set "display" option to "icons" instead, or set " " if they really want to.
        if ( this.options.label )
            this.element.val( this.options.label ); // escaping is automatic; e.g. if label is <span>foo</span>, then val() sets that literal string on the input's "value" attr.
    },

    _setIconsOption: function() // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if ( this.type === "inputPush" ) // <input type=button|submit|reset> doesn't support child nodes, thus no icons, etc.
            return;

        var hasStartIcon = this._setIconOnDom(true);
        var hasEndIcon = this._setIconOnDom(false);
        this._setDisplayOptionOnDom(undefined, hasStartIcon, hasEndIcon);
    },

    /*
     * This method sets either the start or end icon on the DOM, depending on the param, adding or removing the
     * icon span as needed.  It is called both at create time and when the icon changes later.
     *
     * param isStart boolean, whether is start or end icon
     * return boolean indicating whether an icon was set on DOM
     */
    _setIconOnDom: function(isStart) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        var contentContainer = this.buttonElement;
        if (this.type === "button")
            contentContainer = this.element.children("div.oj-button-label");

        if (isStart) {
            var iconSpanSelector = '.oj-button-icon.oj-start';
            var standardIconClasses = 'oj-button-icon oj-start';
            var appIconClass = this.options.icons.start;
            var lastIvar = "_lastStartIcon";
            var pendTo = "prependTo";
        } else {
            iconSpanSelector = '.oj-button-icon.oj-end';
            standardIconClasses = 'oj-button-icon oj-end';
            appIconClass = this.options.icons.end;
            lastIvar = "_lastEndIcon";
            pendTo = "appendTo";
        }

        var iconSpan = contentContainer.find( iconSpanSelector );

        if ( appIconClass ) {
            if ( iconSpan.length ) {
                // remove the app icon class we set last time
                var oldAppIconClass = this[lastIvar];
                iconSpan.removeClass(oldAppIconClass);
            } else {
                iconSpan = $( "<span></span>" )
                    .addClass( standardIconClasses )
                    [pendTo]( contentContainer ); // @HTMLUpdateOK append or prepend trusted new DOM to button elem
            }
            iconSpan.addClass( appIconClass );
        } else {
            iconSpan.remove();
        }

        // remember the app icon class we set, so we can remove it next time
        this[lastIvar] = appIconClass;

        return !!appIconClass;
    },

    /*
     * Hides / shows the label, and adds / removes the rootElement tooltip, depending on whether
     * display is "icons" (and whether there are in fact icons).  Doesn't add/remove the
     * tooltip if app provided one initially.
     *
     * Sets the corresponding root style class, e.g. "oj-button-text-only" or "oj-button-text-icon-start".
     *
     * Note: if rootAttributes ever supports "title", then need to call this *after* _SetRootAttributes so
     * app has a chance to set their own tooltip for checkbox/radio.
     *
     * Must be called when display, icons, or label options are set.
     * (If icons.end ever supports "auto" and is set to "auto", then must be called when menu option is set.)
     *
     * Caller must check that it's not a spanless button before calling this method.
     *
     * All params optional.
     */
    _setDisplayOptionOnDom: function(textSpan, hasStartIcon, hasEndIcon) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        // The case for menu buttons and top level buttons
        var buttonLabelElem = this.buttonElement.children( '.oj-button-label' );
        
        // The case for buttonset buttons
        if (!buttonLabelElem.length)
        {
            var buttonset = this._getEnclosingContainerComponent("buttonset");
            if (this._IsCustomElement() || (buttonset && buttonset._IsCustomElement()))
                buttonLabelElem = this.buttonElement.children( 'oj-option' );
            else
                buttonLabelElem = this.buttonElement;
        }
        
        if (textSpan === undefined)
            textSpan = buttonLabelElem.children( '.oj-button-text' );
        if (hasStartIcon === undefined)
            hasStartIcon = (!!this.options.icons.start) || buttonLabelElem.children("[slot='startIcon']").length;
        if (hasEndIcon === undefined)
            hasEndIcon = (!!this.options.icons.end) || buttonLabelElem.children("[slot='endIcon']").length;

        var multipleIcons = hasStartIcon && hasEndIcon;
        var atLeastOneIcon = hasStartIcon || hasEndIcon;
        var displayIsIcons = this.options.display === "icons";

        if ( atLeastOneIcon && displayIsIcons )
        {
            textSpan.addClass( "oj-helper-hidden-accessible" );

            if ( !this.hasTitle )
            {
                var buttonText = /** @type {string} */
                                 (textSpan.text());
                this.rootElement.attr( "title", $.trim( buttonText ) ); // use buttonText, which is escaped, not options.label, which isn't!
            }
        } else
        {
            textSpan.removeClass( "oj-helper-hidden-accessible" );

            if ( !this.hasTitle )
                this.rootElement.removeAttr( "title" );
        }

        var buttonClass =
            !atLeastOneIcon
                ? "oj-button-text-only"
                : displayIsIcons
                    ? multipleIcons ? "oj-button-icons-only" : "oj-button-icon-only"
                    : multipleIcons
                        ? "oj-button-text-icons"
                        : hasStartIcon
                            ? "oj-button-text-icon-start"
                            : "oj-button-text-icon-end";

        this.rootElement.removeClass( TYPE_CLASSES )
                        .addClass( buttonClass );
    },

    // Anchors are the only Button type lacking a native disabled API, so this tabindex logic is needed to prevent
    // them from being tabbable when disabled.  We only set the tabindex for anchors, and only when the Button is
    // standalone (not in a Buttonset or Toolbar), since those components already manage the tabindex.
    //
    // Each time a standalone anchor Button is set to disabled, including create time, we set its tabindex to -1, and
    // stash its old tabindex in an ivar, so that we can restore it when it is later enabled, and at destroy time.
    // This handles the common case where the button's "is standalone" status never changes.
    //
    // For the rare case that a Button is reparented into one of those components, those components will take
    // over the tabindex, so Button needn't do anything special at that time.
    //
    // Likewise, for the routine case that a Button is inited before its Buttonset/Toolbar, those components will take
    // over the tabindex when they're inited, so it's harmless for Button to have already set a tabindex that those
    // components will overwrite, and to have already set an ivar that those components will ignore.
    //
    // For the rare case that a Button is reparented out of one of those components, so as to become standalone, we
    // take no action, and in fact should NOT call this method, for the following reasons:
    // - Already, for all button types (not just anchor), the app must fix up the tabindex of a button reparented out of
    //   these containers, since it might be -1. There's little reason to handle anchors specially.
    // - We prefer not to introduce special handling just for this rare edge case.
    // - We don't want to guess whether the tabindex was set by the former container (which we try to fix up), vs. being
    //   set a moment ago by the app (which we hope not to clobber).
    // - If the button is disabled, we don't want to call the regular "set tabindex and stash old tabindex" logic, since
    //   the old tabindex is often -1 set by the old container.  Stashing -1 would mean that the next enable or destroy
    //   incorrectly sets -1 on the tabindex. And we don't want to special-case that logic for this rare reparenting case.
    //   Also, disabled buttons from those containers already have the desired -1 value (unless app changed it, which is
    //   their decision), so no action needed anyway.
    //
    // Since this logic never runs when in a Buttonset, callers don't need to worry about "effectively disabled", and can
    // just pass the old/new values of this.options.disabled.
    //
    // This method should NOT be called by refresh(), since there's no need, and since refresh() is called when reparented
    // out of a buttonset/toolbar.  Per above, this method should NOT be called at that time.
    //
    // This method should be called by _setOption("disabled") and at create time, NOT by _updateEffectivelyDisabled()
    // (which is called in both of those cases), since _updateEffectivelyDisabled() can be called
    // indirectly by refresh() when the button was just reparented out of a disabled Buttonset.  Since this logic never
    // runs when in a Buttonset, calling from _updateEffectivelyDisabled() is not needed.
    _manageAnchorTabIndex: function( oldDisabled, disabled )
    {
        // bail if:
        // - truthiness of disabled option is same as before, e.g. changed from "a" to "b"
        // - not a standalone anchor button
        if (!oldDisabled == !disabled || this.type !== "anchor"
                || this._getEnclosingContainerElement("buttonset").length
                || this._getEnclosingContainerElement("toolbar").length)
            return;

        if (disabled) { // enabled button becoming disabled, at create time or later. (Not destroy time, since that logic only passes disabled=false.)
            // If the existing tabindex is unset (attr() returns undefined) or invalid (not a (stringified) integer), set the
            // ivar to null, in which case when we later restore the old value, we just clear the attr.  Obviously correct
            // in unset case. For invalid case, we prefer not to set anything invalid on the dom, and per MDN
            // unset and invalid tabindexes are handled the same.
            var attr = this.element.attr("tabindex");
            this._oldAnchorTabIndex = this._isInteger(Number(attr)) ? attr: null;

            this.element.attr("tabindex", -1);
        } else { // disabled button becoming enabled after create time, incl. destroy time.  (Not create time, since that logic only passes disabled=true.)
            this._oldAnchorTabIndex == null
                ? this.element.removeAttr("tabindex")
                : this.element.attr("tabindex", this._oldAnchorTabIndex);
            // no need to clear ivar
        };
    },

    // IE11 and several modern platforms don't support Number.isInteger(), so use MDN's polyfill:
    _isInteger: function( value )
    {
        return typeof value === "number" &&
            isFinite(value) &&
            Math.floor(value) === value;
    },

    _selectorMap: {
        "buttonset": ".oj-buttonset",
        "toolbar":   ".oj-toolbar"
    },

    _constructorMap: {
        "buttonset": "ojButtonset",
        "toolbar":   "ojToolbar"
    },

    // component param is "buttonset" or "toolbar".
    // Returns non-null JQ object that's length 1 iff this button is contained in a container of the specified type
    _getEnclosingContainerElement: function(component)
    {
        return this.rootElement.closest(this._selectorMap[component]);
    },

    // component param is "buttonset" or "toolbar"
    // Returns buttonset/toolbar component, or null if none.
    _getEnclosingContainerComponent: function(component)
    {
        var elem = this._getEnclosingContainerElement(component)[0];
        var constructor = oj.Components.__GetWidgetConstructor(elem, this._constructorMap[component]);
        return constructor && constructor("instance");
    },

    /*
     * Call this method at create time and whenever the "menu" option is set by the app.
     *
     * - This method first removes menuButton stuff from the element and from any Menu previously set on this Button.
     * - Then, if the Button's "menu" option is set, then it sets the needed listeners on the button.
     *
     * We don't set listeners on the menu until _getMenu() is called on the first launch,
     * so that the menu needn't be inited before the button.
     */
    _setupMenuButton: function(oldMenuOption) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if ( this._getMenuNode() && this.element.is("input")) // both push and toggle buttons based on <input>
            throw new Error("Menu Button functionality is not supported on input elements.");

        this._removeMenuBehavior(oldMenuOption);

        if ( this._getMenuNode() )
        {
            var self = this;
            this.element
                .attr("aria-haspopup", true)
                .on( "keydown" + this.menuEventNamespace, function( event ) {
                    if (event.which === $.ui.keyCode.DOWN)
                    {
                        self._showMenu(event, "firstItem");
                        event.preventDefault();
                        return true;
                    } else if (event.which === $.ui.keyCode.ESCAPE)
                    {
                        var bubbleEscUp = !self.rootElement.hasClass('oj-selected');
						self._dismissMenu(self._getMenuNode(), event);
                        return bubbleEscUp;
                    }

                    return true;
                })
                .on( "click" + this.menuEventNamespace, function( event ) {
                    //console.log("mb click handler");
                    var menu = self._getMenu();
                    if (!menu.__spaceEnterDownInMenu)
                    {
                        // console.log("mb click handler showing menu");
                        // Ideally a click (Enter/Space) would toggle (open/close) the menu without moving focus to it, per WAI-ARIA.
                        // But on IE, JAWS is not recognizing the menu on click/Enter/Space.
                        // Workaround for this, cleared with A11y team, is to move focus to menu like DownArrow. Ref .
                        self._showMenu(event, "firstItem");
                    }
                    menu.__spaceEnterDownInMenu = false;
                    event.preventDefault();
                    return true;
                });
        }
    },

    /*
     * This method removes menuButton functionality from the button and specified menu
     *
     * param menuOption - a current or previous value of the "menu" option
     */
    _removeMenuBehavior: function(menuOption) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        this.element
            .removeAttr( "aria-haspopup" )
            .off( this.menuEventNamespace );

        this._dismissMenu(menuOption);

        // access menu elem directly, rather than using _getMenuOnly(menuOption).widget(), so listener is cleared even if component no longer exists.
        $(menuOption).off( this.menuEventNamespace );
        this._menuListenerSet = false;
    },

    /*
     * Lazy getter for the menu.  It finds the menu instance pointed to by the "menu" option, registers menuButton listeners on it, and returns the menu.
     *
     * This method should be called only by the "user is launching the menu" listeners, which should only be registered if the "menu" option is set.  Do not call at create time.
     *
     * We wait until menu-launch time to lazily get and configure the menu, to avoid an init-order dependency.  It should be OK to init the button before the menu.
     *
     * This method throws if no Menu found, which is app error since Menu should be inited by the time a user launches the menu.
     *
     * (Do NOT return null just because button is disabled, since that would mean disabled menuButtons lose their dropdown arrow, and possibly other problems.)
     */
    _getMenu: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        // The JET Menu of the first element found.
        // Per architect discussion, get it every time rather than caching the menu.
        var menu = this._getMenuOnly(this._getMenuNode());

        // if no element found, or if element has no JET Menu
        if (!menu)
            throw new Error('JET Button: "menu" option specified, but does not reference a valid JET Menu.');

        if (!this._menuListenerSet) {
            var self = this;

            if (!this._IsCustomElement())
             var closeEvent = "ojclose";
            else
                closeEvent = "ojClose";
            
            // must use "on" syntax rather than clobbering whatever "close" handler the app may have set via the menu's "option" syntax
            menu.widget().on( closeEvent + this.menuEventNamespace, function( event, ui ) {
                self._menuDismissHandler(event);
            });
            this._menuListenerSet = true;
        }

        return menu;
    },

    /*
     * Returns the JET Menu of the (first) element specified by the menuOption param, which should be a (current or old)
     * value of the menu option.  Returns null if no element found, or if element has no JET Menu.
     *
     * Most callers should call _getMenu instead, as _getMenuOnly() performs no init, and doesn't throw if no menu found.
     * This method is suitable as a helper for _getMenu, or (say) to close an open menu (which isn't needed if menu is already gone).
     *
     * If you need the menu's *element* (not component), then it's better to call $(foo) than
     * menu=this._getMenuOnly(foo); elem=menu && menu.widget(),
     * since $(foo) works when the element is found but its component is not (and is more efficient).
     *
     * If you need the menu's element *and* component, and should throw if they're missing, then just call
     * menu=this._getMenu(); elem=menu.widget(),
     */
    _getMenuOnly: function(menuOption) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        // Call instance() since need to access non-public Menu api's like __dismiss()
        var constructor = oj.Components.__GetWidgetConstructor($(menuOption)[0], "ojMenu");
        return constructor && constructor("instance");
    },

    /*
     * Shows the menuButton menu if there is one and we're not disabled.
     *
     * MenuButton types:
     * - We support menuButtons on push buttons based on <button> and <a>.
     *     - Architects require that we support menuButtons on at least some types of push buttons.
     *       I.e. if checkbox-based menuButtons were supported, it would not be OK for that to be the only type that's supported.
     * - <input type=button|submit|reset> work fine using the same code path as <button> and <a>, but architects approved proposal
     *   not to support them for following reasons:
     *     - submit/reset menubuttons are semantically / behaviorally absurd.
     *     - these types don't support icons, i.e. the dropdown icon.
     *     - reduces exposure to issues like we had for checkboxes, without sacrificing anything since these flavors were undesirable
     *       for menuButton use anyway.
     *  - Radios are inappropriate for menuButtons.  Alternate clicks should open and close the menu, with corresponding visual
     *    feedback from button, but radios are idempotent: they stay pressed on repeated clicks.
     *  - Architects approved proposal not to support checkboxes for following reasons:
     *      - Checkbox-based menuButtons were considered a bit odd, and we had zero requirements for it.
     *      - The only thing they buy anyone is the checked state corresponding to the menu-open state, but there are several ways for
     *        apps to find out if the menu is open, e.g. check its "hidden" status, check whether the oj-selected class is present, etc.
     *      - They proved to be somewhat brittle, with *different* event-ordering problems in each browser.  Even if we spent the time
     *        to fix it, they could still be a source of problems later, which didn't seem worth it for a flavor no one wanted.
     *
     * param event required
     * param focus required
     */
    _showMenu: function(event, focus) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if (this._IsEffectivelyDisabled())
            return;

        // No need to fire click event, since not appropriate for DownArrow, and already fired for user click.

        var menu = this._getMenu();
        var menuElem = menu.widget();
        menu.open(event, {"launcher": this.element, "initialFocus": focus});

        // bail if launch was cancelled by a beforeOpen listener
        if (!menuElem.is(":visible"))
            return;

        this._menuVisible = true;

        // If menu has neither aria-label nor aria-labelledby after menu.open() calls the beforeOpen listeners, then set aria-labelledby
        // referencing the menu button, and remove it when the menu is closed.  This approach provides a useful default while allowing
        // the menu to be shared by several launchers.
        if (!menuElem.attr("aria-label") && !menuElem.attr("aria-labelledby")) {
            this.element.uniqueId(); // add id if not already there
            this._setAriaLabelledBy = true;
            menuElem.attr("aria-labelledby", this.element.attr("id"));
        }

        // Per UX requirements, menuButtons should look pressed/checked iff the menu is open:
        // - For push buttons, per architectural review, just add/remove oj-selected even though it's a push button.
        //     - Per a11y review, that's fine, but do NOT apply aria-pressed to push buttons, which would turn it into a toggle button for AT users.  He said that
        //       would just confuse things, and that the visual pressed-in look was just eye candy in this case, not semantics that we need to show to AT users.
        // - If checkbox menuButtons were supported, obviously we'd toggle the pressed look by checking/unchecking the button, which in turn would toggle oj-selected.
        //   In that case, we'd fire DOM checked event (right?) and if wrapped in Buttonset, update its checked option and fire optionChange event.
        //ER 19167450
        //calling open on a already open menu now first dismisses it.  the button
        //dismissal handler removes the "oj-select" style.  move the logic that sets
        //the oj-select to after the menu is open.
        this.rootElement.addClass( "oj-selected" )
                        .removeClass( "oj-default oj-focus-only" );
    },

    /*
     * Dismisses the menuButton menu if *we* launched it
     *
     * param menuOption required.  An old or current value of the menu option, indicating which menu to close.
     * param event optional.  Pass iff dismissing due to UI event.
     */
    _dismissMenu: function(menuOption, event) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        // this._menuVisible is set iff *we* launched the menu.  If *something else* launched it, don't dismiss it.
        if (this._menuVisible)
        {
            // Doesn't throw if menu not found. Something is likely wrong in that case, but don't sweat it unless they try to *launch* an MIA menu.
            var menu = this._getMenuOnly(menuOption);
            if (menu) {
                // TODO: this should be called by __dismiss(), rather than the caller having to call this too.
                menu.__collapseAll( event, true ); // close open submenus before hiding the popup so that submenus will not be shown on reopen

                menu.__dismiss(event); // causes _menuDismissHandler(event) to be called
            }
        }
    },

    /*
     * Handles menu dismissals, whether or not we dismissed it ourselves.
     * See comments on similar code in _showMenu().
     *
     * Also called by the beforeOpen listener we put on the menu, *if* the launch was by something else,
     * including our own context menu.  So if something steals our menu, we deselect the button.
     *
     * param event must remain optional, since some callers of _dismissMenu have no event
     */
    _menuDismissHandler: function(event) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        if (this._setAriaLabelledBy) {
            $(this._getMenuNode()).removeAttr("aria-labelledby");
            this._setAriaLabelledBy = false;
        }

        //console.log(this.options.label + ": button._menuDismissHandler called");
        // Since only push buttons are supported for menu buttons, the only reason for .oj-selected to be present is if it's an open menu button,
        // so we remove the class since the menu is being dismissed.
        this.rootElement.removeClass( "oj-selected" );
        this._toggleDefaultClasses();

        this._menuVisible = false;
    },

    /*
     * Let the "state classes" be oj-active, oj-disabled, oj-selected, oj-hover, and oj-focus.
     * This method ensures that the root element has:
     *   - the oj-default class iff it has none of the state classes, and
     *   - the oj-focus-only class iff it has oj-focus but no other state classes.
     * Note that oj-focus-highlight should never be present without oj-focus, so there's no need to check for that separately.
     */
    _toggleDefaultClasses: function()
    {
        var otherStates = this.rootElement.is( ".oj-hover, .oj-active, .oj-selected, .oj-disabled" );
        var defaultState, focusedOnly;

        if (otherStates) {
            defaultState = false;
            focusedOnly = false;
        } else {
            var focused = this.rootElement.is( ".oj-focus" );
            defaultState = !focused;
            focusedOnly = focused;
        }

        this.rootElement.toggleClass( "oj-default", defaultState );
        this.rootElement.toggleClass( "oj-focus-only", focusedOnly );
    }

    // API doc for inherited methods with no JS in this file:

    /**
     * <p>The default slot is the button's text label. The <code class="prettyprint">&lt;oj-button></code> element accepts plain text or DOM nodes as children for the default slot.
     * A default slot label is required for all buttons for accessibility purposes. The label can be hidden using the display attribute.</p>
     *
     * <p>If a text node is provided it will be wrapped in a span.</p>
     *
     * @ojchild Default
     * @memberof oj.ojButton
     *
     * @example <caption>Initialize the Button with child content specified:</caption>
     * &lt;oj-button>
     *   &lt;span>myValue&lt;/span>
     * &lt;/oj-button>
     *
     * @example <caption>Initialize the Button with data-bound child content specified in a span:</caption>
     * &lt;oj-button>
     *   &lt;span data-bind='text: myText'>&lt;/span>
     * &lt;/oj-button>
     *
     * @example <caption>Initialize the Button with data-bound child content specified without a container element:</caption>
     * &lt;oj-button>
     *   &lt;!-- ko text: myText -->&lt;!--/ko-->
     * &lt;/oj-button>
     */
    
    /**
     * <p>The default slot is the menu button's text label. The <code class="prettyprint">&lt;oj-menu-button></code> element accepts plain text or DOM nodes as children for the default slot.
     * A default slot label is required for all menu buttons for accessibility purposes. The label can be hidden using the display attribute.</p>
     *
     * <p>If a text node is provided it will be wrapped in a span.</p>
     *
     * @ojchild Default
     * @memberof oj.ojMenuButton
     *
     * @example <caption>Initialize the Menu Button with child content specified:</caption>
     * &lt;oj-menu-button>
     *   &lt;span>myValue&lt;/span>
     * &lt;/oj-menu-button>
     *
     * @example <caption>Initialize the Menu Button with data-bound child content specified in a span:</caption>
     * &lt;oj-menu-button>
     *   &lt;span data-bind='text: myText'>&lt;/span>
     * &lt;/oj-menu-button>
     *
     * @example <caption>Initialize the Menu Button with data-bound child content specified without a container element:</caption>
     * &lt;oj-menu-button>
     *   &lt;!-- ko text: myText -->&lt;!--/ko-->
     * &lt;/oj-menu-button>
     */

    /**
     * <p>The <code class="prettyprint">startIcon</code> slot is the button's start icon. The  <code class="prettyprint">&lt;oj-button></code> element accepts DOM nodes as children with the startIcon slot.</p>
     *
     * @ojslot startIcon
     * @memberof oj.ojButton
     *
     * @example <caption>Initialize the Button with child content specified for the startIcon:</caption>
     * &lt;oj-button>
     *   &lt;span slot='startIcon' class='myIconClass'>&lt;/span>
     *   &lt;span>myValue&lt;/span>
     * &lt;/oj-button>
     */
    
    /**
     * <p>The <code class="prettyprint">startIcon</code> slot is the menu button's start icon. The  <code class="prettyprint">&lt;oj-menu-button></code> element accepts DOM nodes as children with the startIcon slot.</p>
     *
     * @ojslot startIcon
     * @memberof oj.ojMenuButton
     *
     * @example <caption>Initialize the Menu Button with child content specified for the startIcon:</caption>
     * &lt;oj-menu-button>
     *   &lt;span slot='startIcon' class='myIconClass'>&lt;/span>
     *   &lt;span>myValue&lt;/span>
     * &lt;/oj-menu-button>
     */
     
    /**
     * <p>The <code class="prettyprint">endIcon</code> slot is the button's end icon. The  <code class="prettyprint">&lt;oj-button></code> element accepts DOM nodes as children with the endIcon slot.</p>
     *
     * @ojslot endIcon
     * @memberof oj.ojButton
     *
     * @example <caption>Initialize the Button with child content specified for the endIcon:</caption>
     * &lt;oj-button>
     *   &lt;span>myValue&lt;/span>
     *   &lt;span slot='endIcon' class='myIconClass'>&lt;/span>
     * &lt;/oj-button>
     */
     
    /**
     * <p>The <code class="prettyprint">endIcon</code> slot is the menu button's end icon. The  <code class="prettyprint">&lt;oj-menu-button></code> element accepts DOM nodes as children with the endIcon slot. If no end icon is provided, a default end icon is used.</p>
     *
     * @ojslot endIcon
     * @memberof oj.ojMenuButton
     *
     * @example <caption>Initialize the Menu Button with child content specified for the endIcon:</caption>
     * &lt;oj-menu-button>
     *   &lt;span>myValue&lt;/span>
     *   &lt;span slot='endIcon' class='myIconClass'>&lt;/span>
     * &lt;/oj-menu-button>
     */
    
    /**
     * <p>The <code class="prettyprint">menu</code> menu associatied with the menu button. The <code class="prettyprint">oj-menu-button</code> element accepts a single <code class="prettyprint">oj-menu</code> element as a child with the menu slot. See the [JET Menu]{@link oj.ojMenu} for more information on setting up a menu.</p>
     *
     * @ojslot menu
     * @memberof oj.ojMenuButton
     *
     * @example <caption>Initialize the Menu Button with child content specified for the menu:</caption>
     * &lt;oj-menu-button>
     *   &lt;span>myValue&lt;/span>
     *   &lt;oj-menu slot="menu" style="display:none" aria-label="This menu button's menu">
     *   ...
     *   &lt;/oj-menu>
     * &lt;/oj-menu-button>
     */
     
    // Fragments:

    /**
     * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.
     *
     * <p>Disabled content: JET supports an accessible luminosity contrast ratio,
     * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>,
     * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which
     * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user
     * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio
     * required of enabled content, it cannot be used to convey meaningful information.<p>
     * 
     * @ojfragment accessibilityCommon
     * @memberof oj.ojButton
     * @instance
     */
    
    /**
     * <p>Built-in KO bindings, like KO's <code class="prettyprint">disable</code> binding, should not be used for state with a JS API, since that is tatamount to
     * updating the DOM directly.  The component attribute should be bound instead.
     *
     * <p>State with no JS API should be set by manipulating the DOM directly in an allowable way, and then calling <code class="prettyprint">refresh()</code>
     * on the affected component(s).  E.g. the reading direction (LTR / RTL) is changed by by setting the <code class="prettyprint">"dir"</code> attribute on the
     * <code class="prettyprint">&lt;html></code> node and calling <code class="prettyprint">refresh()</code>.
     *
     * <p>When using a built-in Knockout binding, keep in mind that those bindings do not
     * execute the necessary <code class="prettyprint">refresh()</code> call after updating the DOM.  Updates that flow from the component to the observable,
     * as a result of user interaction, are not problematic.  But updates in the other direction, that programmatically update the DOM because the observable changed,
     * will not be picked up until the next <code class="prettyprint">refresh()</code>.
     * 
     * @ojfragment stateCommon
     * @memberof oj.ojButton
     * @instance
     */
    
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Push Button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Push the button.</td>
     *     </tr>
     *     <tr>
     *       <td>Toggle Button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Toggle the button.</td>
     *     </tr>
     *     <tr>
     *       <td>Menu Button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Open the menu.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>See also the [Menu]{@link oj.ojMenu} touch gesture doc.
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojButton
     * @instance
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Push Button</td>
     *       <td><kbd>Enter</kbd> or <kbd>Space</kbd>*</td>
     *       <td>Push the button.</td>
     *     </tr>
     *     <tr>
     *       <td>Toggle Button</td>
     *       <td><kbd>Enter</kbd> or <kbd>Space</kbd></td>
     *       <td>Toggle the button.</td>
     *     </tr>
     *     <tr>
     *       <td rowspan="2">Menu Button</td>
     *       <td><kbd>Enter</kbd>, <kbd>Space</kbd>*, or <kbd>DownArrow</kbd></td>
     *       <td>Open the menu.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Esc</kbd></td>
     *       <td>Close the menu.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>* Some types of Push and Menu Buttons support <kbd>Enter</kbd>, not <kbd>Space</kbd>.
     *
     * <p>See the [Menu]{@link oj.ojMenu} keyboard doc for keystrokes that apply when focus is on the menu.
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojButton
     * @instance
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     *
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-button-sm<br>
     *           oj-button-lg<br>
     *           oj-button-xl</td>
     *       <td>Makes the button small, large, or extra large. Is applied to the Button's root element.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-button-primary</td>
     *       <td>Draws attention to the button, often identifying the primary action in a set of buttons.
     *           Designed for use with a push button. In some themes, this class does nothing. Is applied to the Button's root element.</td>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-button-confirm</td>
     *       <td>Identifies an action to confirm. Designed for use with a push button. Is applied to the Button's root element.</td>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-focus-highlight</td>
     *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojButton
     * @instance
     */
});

/**
 * @ojcomponent oj.ojMenuButton
 * @ojdisplayname Menu Button
 * @since 4.0
 * @ojstatus preview
 * @augments oj.ojButton
 * @ojshortdesc A button that launches a menu when clicked.
 * @ojrole button
 *
 * @classdesc
 * <h3 id="menubuttonOverview-section">
 *   JET Menu Button
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#menubuttonOverview-section"></a>
 * </h3>
 *
 * <pre class="prettyprint"><code>&lt;oj-menu-button id="menuButton">
 *   &lt;oj-menu id="menu" slot="menu" style="display:none">
 *       &lt;oj-option>Item 1&lt;/oj-option>
 *       &lt;oj-option>Item 2&lt;/oj-option>
 *       &lt;oj-option>Item 3&lt;/oj-option>
 *   &lt;/oj-menu>
 * &lt;/oj-menu-button></code></pre>
 * 
 * <h3 id="buttonsetToolbar-section">
 *   Buttonsets and Toolbars
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#buttonsetToolbar-section"></a>
 * </h3>
 * 
 * <p>The [JET Buttonset]{@link oj.ojButtonset} component can be used to create toggle buttons or group related buttons.  It cannot be used
 * to create menu buttons or regular push buttons. Buttonset provides visual and semantic grouping and WAI-ARIA-compliant focus management.  See the Buttonset API doc for more information.
 *
 * <p>Menu buttons, push buttons, and buttonsets can be placed in a [JET Toolbar]{@link oj.ojToolbar}.  Like Buttonset, Toolbar is themable and provides WAI-ARIA-compliant
 * focus management.  See the Toolbar API doc for more information.
 * 
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>For accessibility, a JET Menu Button must always have its default slot filled, even if it is <a href="#display">icon-only</a>.
 * 
 * {@ojinclude "name":"accessibilityCommon"}
 *
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <p>In lieu of stamping a menu button in a table, dataGrid, or other container, consider placing a single Menu Button outside of the
 * container that acts on the currently selected row or cell.
 *
 * <h3 id="state-section">
 *   Setting Component State
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#state-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stateCommon"}
 */

/**
 * @ojcomponent oj.ojButtonsetOne
 * @since 0.6
 * @ojstatus preview
 * @augments oj.ojButtonset
 * @ojshortdesc A grouping of related buttons where only one button may be selected.
 * @ojrole button
 * @ojrole radiogroup
 *
 * @classdesc
 * <h3 id="buttonsetOverview-section">
 *   JET Buttonset One
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#buttonsetOverview-section"></a>
 * </h3>
 *
 * <p>Description: Themeable, WAI-ARIA-compliant visual and semantic grouping container for [JET Buttons]{@link oj.ojButton}.
 *
 * <p>The JET ButtonsetOne can be used to group related buttons, where only one button may be selected.  Buttonset provides
 * visual and semantic grouping and WAI-ARIA-compliant focus management.
 *
 * <p>When a Buttonset is created or refreshed, it creates JET Buttons out of all contained <code class="prettyprint">oj-option</code>
 * DOM elements by wrapping them and calling <code class="prettyprint">.ojButton()</code> on the generated content. The Buttonset will remove all
 * non <code class="prettyprint">oj-option</code> DOM elements from the Buttonset.
 *
 * <p><code class="prettyprint">oj-option</code>s inside of the buttonset should specify the <code class="prettyprint">oj-option</code>
 * <code class="prettyprint">value</code> attribute, since the <code class="prettyprint">oj-buttonset-one</code> <code class="prettyprint">value</code> attribute refers to that attribute.
 *
 * <pre class="prettyprint"><code>&lt;oj-buttonset-one id="myButtonset">
 *   &lt;oj-option value="myValue0">Value0&lt;/oj-option>
 *   &lt;oj-option value="myValue1">Value1&lt;/oj-option>
 * &lt;/oj-buttonset-one></code></pre>
 *
 * {@ojinclude "name":"buttonsetCommon"}
 *
 * <pre class="prettyprint">
 * <code>&lt;oj-buttonset-one id="drinkset">
 *       &lt;!-- ko foreach: drinkValues -->
 *       &lt;oj-option value="[[id]]">
 *           &lt;span data-bind="text: label">&lt;/span>
 *       &lt;/oj-option>
 *       &lt;!-- /ko -->
 *   &lt;/oj-buttonset-one>
 * </code></pre>
 */

/**
 * @ojcomponent oj.ojButtonsetMany
 * @since 0.6
 * @ojstatus preview
 * @augments oj.ojButtonset
 * @ojshortdesc A grouping of related buttons where any number of buttons may be selected.
 * @ojrole button
 * @ojrole group
 *
 * @classdesc
 * <h3 id="buttonsetOverview-section">
 *   JET Buttonset Many
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#buttonsetOverview-section"></a>
 * </h3>
 *
 * <p>Description: Themeable, WAI-ARIA-compliant visual and semantic grouping container for [JET Buttons]{@link oj.ojButton}.
 *
 * <p>The JET ButtonsetMany can be used to group related buttons, where any number of the buttons can be selected.  Buttonset provides
 * visual and semantic grouping and WAI-ARIA-compliant focus management.
 *
 * <p>When a Buttonset is created or refreshed, it creates JET Buttons out of all contained <code class="prettyprint">oj-option</code>
 * DOM elements by wrapping them and calling <code class="prettyprint">.ojButton()</code> on the generated content. The Buttonset will remove all
 * non <code class="prettyprint">oj-option</code> DOM elements from the Buttonset.
 *
 * <p><code class="prettyprint">oj-option</code>s in the buttonset should specify the <code class="prettyprint">oj-option</code>
 * <code class="prettyprint">value</code> attribute, since the <code class="prettyprint">oj-buttonset-mnay</code> <code class="prettyprint">value</code> attribute refers to that attribute.
 *
 * <pre class="prettyprint"><code>&lt;oj-buttonset-many id="myButtonset">
 *   &lt;oj-option value="myValue0">Value0&lt;/oj-option>
 *   &lt;oj-option value="myValue1">Value1&lt;/oj-option>
 * &lt;/oj-buttonset-many></code></pre>
 *
 * {@ojinclude "name":"buttonsetCommon"}
 *
 * <pre class="prettyprint">
 * <code>&lt;oj-buttonset-many id="drinkset">
 *       &lt;!-- ko foreach: drinkValues -->
 *       &lt;oj-option value="[[id]]">
 *           &lt;span data-bind="text: label">&lt;/span>
 *       &lt;/oj-option>
 *       &lt;!-- /ko -->
 *   &lt;/oj-buttonset-many>
 * </code></pre>
 */

/**
 * @ojcomponent oj.ojButtonset
 * @augments oj.baseComponent
 * @since 0.6
 * @abstract
 * @classdesc
 * @hideconstructor
 */

/**
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="keyboard-appdev-section">
 *   Keyboard Application Developer Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-appdev-section"></a>
 * </h3>
 *
 * <p>The application should not do anything to interfere with the Buttonset's focus management, such as setting the <code class="prettyprint">tabindex</code>
 * of the buttons.  Also, enabled buttons should remain user-visible, without which arrow-key navigation to the button would cause the focus to seemingly disappear.
 *
 * <p>The buttonset's focus management should be turned off when placing the buttonset in a [JET Toolbar]{@link oj.ojToolbar}.  See the <code class="prettyprint">focusManagement</code> attribute.
 * In this case, the "Keyboard End User Information" documented above is superseded by the Toolbar's documented keyboard behavior.
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>As shown in the online demos, the application is responsible for applying <code class="prettyprint">aria-label</code> and/or
 * <code class="prettyprint">aria-controls</code> attributes like the following to the buttonset element, if applicable per the instructions that follow:
 *
 * <pre class="prettyprint">
 * <code>aria-label="Choose only one beverage."
 * aria-controls="myTextEditor"
 * </code></pre>
 *
 * <p>An <code class="prettyprint">aria-label</code> conveying the "choose only one" semantics should be included for a buttonset-one.
 *
 * <p>The <code class="prettyprint">aria-controls</code> attribute should be included if the buttonset is controlling something else on the page, e.g.
 * bold / italic / underline buttons controlling a rich text editor.  If the buttonset is contained in a toolbar, <code class="prettyprint">aria-controls</code>
 * should be placed on the toolbar, not on the buttonsets within the toolbar.
 *
 * <p>Disabled content: JET supports an accessible luminosity contrast ratio,
 * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>,
 * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which
 * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user
 * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio
 * required of enabled content, it cannot be used to convey meaningful information.<p>
 *
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>The only supported way to set the reading direction (LTR or RTL) is to set the <code class="prettyprint">"dir"</code> attribute on the
 * <code class="prettyprint">&lt;html></code> element of the page.  As with any JET component, in the unusual case that the reading direction
 * is changed post-create, the buttonset must be <code class="prettyprint">refresh()</code>ed, or the page must be reloaded.
 *
 * <h3 id="binding-section">
 *   Declarative Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#binding-section"></a>
 * </h3>
 *
 * <p>For components like Buttonset and Menu that contain a number of like items, applications may wish to use a <code class="prettyprint">foreach</code> Knockout binding
 * to stamp out the contents.  This binding cannot live on the same node as the <code class="prettyprint">ojComponent</code> binding, and must instead live on a nested
 * virtual element as follows:
 *
 * @ojfragment buttonsetCommon
 * @memberof oj.ojButtonset
 * @instance
 */

oj.__registerWidget("oj.ojButtonset", $['oj']['baseComponent'],
{
    // private.  Was an undocumented JQUI option, which we removed, so I moved from options to here and added underscore.  Leave unquoted so gets renamed by GCC as desired.
    _items: "button, input[type=button], input[type=submit], input[type=reset], input[type=checkbox], input[type=radio], a",

    widgetEventPrefix : "oj",

    options: // options is in externs.js.  TODO: same as other prototype fields.
    {
        /**
         * The <code class="prettyprint">value</code> attribute indicates which <code class="prettyprint">oj-option</code> in the Buttonset is selected.
         * It corresponds to the <code class="prettyprint">value</code> attribute of the <code class="prettyprint">oj-option</code>, which should always be set.
         *
         * <code class="prettyprint">value</code> is any type equal to the <code class="prettyprint">value</code> attribute of the selected <code class="prettyprint">oj-option</code>.  The attribute is
         * <code class="prettyprint">null</code> if and only if no <code class="prettyprint">oj-option</code> is selected.  Thus, an n-<code class="prettyprint">oj-option</code> group has n+1 valid
         * <code class="prettyprint">value</code> values: the n <code class="prettyprint">oj-option</code> values, and <code class="prettyprint">null</code> .
         *
         * {@ojinclude "name":"buttonsetCommonValue"}
         *
         * @name value
         * @memberof oj.ojButtonsetOne
         * @instance
         * @type {*}
         * @default null
         * @ojwriteback
         * @ojshortdesc Indicates which oj-option in the buttonset is selected.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">value</code> attribute specified:</caption>
         * &lt;oj-buttonset-one value='bold'>&lt;/oj-buttonset-one>
         *
         * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
         * // Get one
         * var value = myButtonset.value;
         *
         * // Set one.  (No property change event will be fired.)
         * myButtonset.value = 'bold';
         */
        /**
         * The <code class="prettyprint">value</code> attribute indicates which <code class="prettyprint">oj-option</code>s in the Buttonset are selected.
         * It corresponds to the <code class="prettyprint">value</code> attribute of those elements, which should always be set.
         *
         * <code class="prettyprint">value</code> is a possibly empty, non-<code class="prettyprint">null</code> string array containing the <code class="prettyprint">value</code>
         * attributes of the selected <code class="prettyprint">oj-option</code>s.  This array has "set", not "list", semantics; i.e. order is neither important nor guaranteed.
         * Thus, an n-<code class="prettyprint">oj-option</code> set has 2^n valid <code class="prettyprint">value</code> values: the 2^n possible subsets of n <code class="prettyprint">oj-option</code>s.
         *
         * {@ojinclude "name":"buttonsetCommonValue"}
         *
         * @name value
         * @memberof oj.ojButtonsetMany
         * @instance
         * @type {Array.<*>|null}
         * @default null
         * @ojwriteback
         * @ojshortdesc Indicates which oj-options in the buttonset are selected.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">value</code> attribute specified:</caption>
         * &lt;oj-buttonset-many value='{{["bold", "italic"]}}'>&lt;/oj-buttonset-many>
         *
         * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
         * // Get one (if many)
         * var value = myButtonset.value[0];
         *
         * // Get all
         * var values = myButtonset.value;
         *
         * // Set one.  (If many. No property change event will be fired.)
         * myButtonset.value[1] = 'bold';
         *
         * // Set all.
         * myButtonset.value = ["bold", "italic"];
         */
        /**
         * <p>In all other cases, <code class="prettyprint">value</code> is <code class="prettyprint">null</code>.
         *
         * <p>It's still possible for the <code class="prettyprint">value</code> attribute and DOM to get out of sync by other means.
         * In this case, the app is responsible for updating the <code class="prettyprint">value</code> attribute.  A typical case is
         * when the set of Buttons contained in the Buttonset changes, possibly due to a Knockout binding, in which case the app must first call
         * <code class="prettyprint">refresh</code> (as in all cases when the DOM changes underneath a component), and then
         * update the <code class="prettyprint">value</code> attribute to the desired value.
         *
         * <p>Often there is no need to listen for this event, since the
         * <code class="prettyprint">value</code> binding, discussed above, will update the bound observable whenever the
         * <code class="prettyprint">value</code> state changes.  The declarative binding is often preferable to an explicit listener.
         *
         * <p>A click listener should not be used to detect changes to the <code class="prettyprint">value</code> state.
         * The attribute <code class="prettyprint">value</code> binding and/or
         * the <code class="prettyprint">valueChange</code> event should be used instead.
         *
         * @alias value
         * @expose
         * @memberof oj.ojButtonset         
         * @instance
         * @ojfragment buttonsetCommonValue
         */
        checked: null,

        /**
         * {@ojinclude "name":"buttonsetCommonChroming"}
         *
         * @name chroming
         * @memberof oj.ojButtonsetOne
         * @instance
         * @type {string|undefined}
         * @ojvalue {string} "full" In typical themes, full-chrome buttons always have chrome.
         * @ojvalue {string} "half" In typical themes, half-chrome buttons acquire chrome only in their hover, active, and selected states. Half-chroming is recommended for buttons in a toolbar.
         *     (This is the toolbar default in most themes.)
         * @ojvalue {string} "outlined" In typical themes, outlined buttons are similar to half-chrome buttons, but have a border in the default state.
         * @ojshortdesc Indicates in what states the buttonset has chrome (background and border).
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">chroming</code> attribute specified:</caption>
         * &lt;oj-buttonset-one chroming='half'>&lt;/oj-buttonset-one>
         *
         * @example <caption>Get or set the <code class="prettyprint">chroming</code> property after initialization:</caption>
         * // getter
         * var chromingValue = myButtonset.chroming;
         *
         * // setter
         * myButtonset.chroming = 'half';
         *
         * @example <caption>Set the default in the theme (SCSS) :</caption>
         * $buttonsetChromingOptionDefault: half !default;
         */
        /**
         * {@ojinclude "name":"buttonsetCommonChroming"}
         *
         * @name chroming
         * @memberof oj.ojButtonsetMany
         * @instance
         * @type {string}
         * @ojvalue {string} "full" In typical themes, full-chrome buttons always have chrome.
         * @ojvalue {string} "half" In typical themes, half-chrome buttons acquire chrome only in their hover, active, and selected states. Half-chroming is recommended for buttons in a toolbar.
         *     (This is the toolbar default in most themes.)
         * @ojvalue {string} "outlined" In typical themes, outlined buttons are similar to half-chrome buttons, but have a border in the default state.
         * @ojshortdesc Indicates in what states the buttonset has chrome (background and border).
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">chroming</code> attribute specified:</caption>
         * &lt;oj-buttonset-many chroming='half'>&lt;/oj-buttonset-many>
         *
         * @example <caption>Get or set the <code class="prettyprint">chroming</code> property after initialization:</caption>
         * // getter
         * var chromingValue = myButtonset.chroming;
         *
         * // setter
         * myButtonset.chroming = 'half';
         *
         * @example <caption>Set the default in the theme (SCSS) :</caption>
         * $buttonsetChromingOptionDefault: half !default;
         */

        /**
         * <p>Indicates in what states the buttonset's buttons have chrome (background and border).
         *
         * <p>A buttonset's chroming must be set by setting this buttonset attribute (or setting the [chroming]{@link oj.ojToolbar#chroming} attribute
         * of a containing toolbar).
         *
         * <p>The default chroming varies by theme and containership as follows:
         * <ul>
         *   <li>If the buttonset is in a toolbar, then the default chroming is the current value of the toolbar's [chroming]{@link oj.ojToolbar#chroming} attribute.</li>
         *   <li>Else, if <code class="prettyprint">$buttonsetChromingOptionDefault</code> is set in the current theme as seen in the example below, then that value is the chroming default.</li>
         *   <li>Else, the default chroming is <code class="prettyprint">"full"</code>.</li>
         * </ul>
         *
         * <p>Once a value has been set on this buttonset attribute, that value applies regardless of theme and containership.
         *
         * @expose
         * @memberof oj.ojButtonset
         * @instance
         * @since 1.2.0
         * @ojfragment buttonsetCommonChroming
         */
        chroming: "full",

        /**
         * {@ojinclude "name":"buttonsetCommonDisplay"}
         *
         * @name display
         * @memberof oj.ojButtonsetOne
         * @instance
         * @type {string}
         * @ojvalue {string} "all" Display both the label and icons.
         * @ojvalue {string} "icons" Display only the icons.
         * @default "all"
         * @ojshortdesc Specifies whether the buttonset displays label and icons, or just icons.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">display</code> attribute specified:</caption>
         * &lt;oj-buttonset-one display='icons'>&lt;/oj-buttonset-one>
         *
         * @example <caption>Get or set the <code class="prettyprint">display</code> property after initialization:</caption>
         * // getter
         * var displayValue = myButtonset.display;
         *
         * // setter
         * myButtonset.display = 'icons';
         */
        /**
         * {@ojinclude "name":"buttonsetCommonDisplay"}
         *
         * @name display
         * @memberof oj.ojButtonsetMany
         * @instance
         * @type {string}
         * @ojvalue {string} "all" Display both the label and icons.
         * @ojvalue {string} "icons" Display only the icons.
         * @default "all"
         * @ojshortdesc Specifies whether the buttonset displays label and icons, or just icons.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">display</code> attribute specified:</caption>
         * &lt;oj-buttonset-many display='icons'>&lt;/oj-buttonset-many>
         *
         * @example <caption>Get or set the <code class="prettyprint">display</code> property after initialization:</caption>
         * // getter
         * var displayValue = myButtonset.display;
         *
         * // setter
         * myButtonset.display = 'icons';
         */
        /**
         * <p>Whether to display both the label and icons (<code class="prettyprint">"all"</code>)
         * or just the icons (<code class="prettyprint">"icons"</code>) of the buttons.  In the latter case, the label is displayed in a tooltip instead.
         *
         * <p>The <code class="prettyprint">display</code> attribute will be ignored if no icons exist in the button.
         *
         * @expose
         * @memberof oj.ojButtonset
         * @instance
         * @ojfragment buttonsetCommonDisplay
         */
        display: "all",

        /**
         * {@ojinclude "name":"buttonsetCommonDisabled"}
         *
         * @member
         * @name disabled
         * @memberof oj.ojButtonsetOne
         * @instance
         * @type {boolean}
         * @default false
         * @ojshortdesc Specifies that the buttonset element should be disabled.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">disabled</code> attribute specified:</caption>
         * &lt;oj-buttonset-one disabled='true'>&lt;/oj-buttonset-one>
         *
         * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
         * // getter
         * var disabledValue = myButtonset.disabled;
         *
         * // setter
         * myButtonset.disabled = true;
         */
        /**
         * {@ojinclude "name":"buttonsetCommonDisabled"}
         *
         * @member
         * @name disabled
         * @memberof oj.ojButtonsetMany
         * @instance
         * @type {boolean}
         * @default false
         * @ojshortdesc Specifies that the buttonset element should be disabled.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">disabled</code> attribute specified:</caption>
         * &lt;oj-buttonset-many disabled='true'>&lt;/oj-buttonset-many>
         *
         * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
         * // getter
         * var disabledValue = myButtonset.disabled;
         *
         * // setter
         * myButtonset.disabled = true;
         */
        /**
         * <p>Setting the Buttonset's <code class="prettyprint">disabled</code> attribute effectively disables all its Buttons, without affecting
         * their <code class="prettyprint">disabled</code> attributes.  Thus, a Button is effectively disabled if either its own
         * <code class="prettyprint">disabled</code> attribute is set, or the Buttonset's <code class="prettyprint">disabled</code> attribute is set.
         *
         * @member
         * @name disabled
         * @memberof oj.ojButtonset
         * @instance
         * @ojfragment buttonsetCommonDisabled
         */
        // disabled option declared in superclass, but we still want the above API doc

        // Consider getting rid of this option in the future.  Under his proposal (whose particulars we agreed needed to be ironed out),
        // rather than using this option to turn off Bset's handling, Bset would always handle arrow keys and preventDefault or stopPropagation.  Arrow
        // keys would no longer wrap around when reach end of Bset, in which case Bset would let the event bubble up to Toolbar (or to whoever if
        // not in a TB).  Any edge cases?  e.g. with tabstops, TB contents with special arrow-key behavior like inputTexts, etc.?
        /**
         * {@ojinclude "name":"buttonsetCommonFocusManagement"}
         *
         * @name focusManagement
         * @memberof oj.ojButtonsetOne
         * @instance
         * @type {string}
         * @ojvalue {string} "oneTabstop" Focus management is enabled.  The Buttonset is a single tabstop with arrow-key navigation.
         * @ojvalue {string} "none" Focus management is disabled, to avoid interfering with the focus management of a containing component.
         * @default "oneTabstop"
         * @ojshortdesc Should be set to none when the buttonset is placed within a toolbar.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">focusManagement</code> attribute specified:</caption>
         * &lt;oj-buttonset-one focus-management='none'>&lt;/oj-buttonset-one>
         *
         * @example <caption>Get or set the <code class="prettyprint">focusManagement</code> property after initialization:</caption>
         * // getter
         * var focusManagementValue = myButtonset.focusManagement;
         *
         * // setter
         * myButtonset.focusManagement = 'none';
         */
        /**
         * {@ojinclude "name":"buttonsetCommonFocusManagement"}
         *
         * @name focusManagement
         * @memberof oj.ojButtonsetMany
         * @instance
         * @type {string}
         * @ojvalue {string} "oneTabstop" Focus management is enabled.  The Buttonset is a single tabstop with arrow-key navigation.
         * @ojvalue {string} "none" Focus management is disabled, to avoid interfering with the focus management of a containing component.
         * @default "oneTabstop"
         * @ojshortdesc Should be set to none when the buttonset is placed within a toolbar.
         *
         * @example <caption>Initialize the Buttonset with the <code class="prettyprint">focusManagement</code> attribute specified:</caption>
         * &lt;oj-buttonset-many focus-management='none'>&lt;/oj-buttonset-many>
         *
         * @example <caption>Get or set the <code class="prettyprint">focusManagement</code> property after initialization:</caption>
         * // getter
         * var focusManagementValue = myButtonset.focusManagement;
         *
         * // setter
         * myButtonset.focusManagement = 'none';
         */
        /**
         * The <code class="prettyprint">focusManagement</code> attribute should be set to <code class="prettyprint">"none"</code> when the
         * Buttonset is placed in a [JET Toolbar]{@link oj.ojToolbar}.  This allows the Toolbar to manage the focus with no interference from the Buttonset,
         * so that arrow keys move within the entire Toolbar, not just within the Buttonset.
         *
         * @expose
         * @memberof oj.ojButtonset
         * @instance
         * @ojfragment buttonsetCommonFocusManagement
         */
        focusManagement: "oneTabstop"

        // Events
    },

    // If this is a radio or checkbox buttonset, and the specified checked value is valid,
    //   then sets it on the buttons' checked properties in the DOM,
    // Else if it's the non-radio, non-checkbox case and null was passed, which is the one and only valid value for that case,
    //   then doesn't change DOM.
    // Else it's an invalid value,
    //   so throws.
    // Does not set oj-selected, so must call _applyCheckedStateFromDom() after this.
    //   TBD: optionally, this method could do that work, perhaps controlled by a param
    // This method is called by _setOption("checked", ...) and _ComponentCreate > _setup.
    // Does NOT require the buttons to already be JET Buttons (useful for _setup caller).
    _setCheckedOnDom: function(checked, $buttons) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        var type = $.type(checked);
        var valid, allCheckboxes;
                
        // whether buttonset contains exclusively checkboxes, therefore is buttonset many
        allCheckboxes = (this.element[0].tagName === 'OJ-BUTTONSET-MANY' || ($buttons.length > 0 && $buttons.filter("input[type=checkbox]").length === $buttons.length));
        
        // requires an array in buttonset many case
        if (allCheckboxes && type != "array")
        {
            throw new Error("Invalid 'checked' value set on JET Buttonset: " + checked);
        }
        
        // cannot have different names in radio groups
        if (!allCheckboxes && $buttons.length > 0)
        {
            var name = $buttons[0].name;
            var validRadios = (name || $buttons.length<=1) // if name is "" and there's >1 radio, then they're in separate radio groups
                && $buttons.filter("input[type=radio][name=" + name + "]").length === $buttons.length;
        }

        valid = this._setCheckedOnButtons(checked, $buttons, type, allCheckboxes);
        
        valid = valid || checked===null;

        if (!valid && (!this._IsCustomElement() || $buttons.length === this.element.children().length))
        {
            throw new Error("Invalid 'checked' value set on JET Buttonset: " + checked);
        }
    },
    // does comparsion and sets checked on inputs
    // for custom elements it will accept objects as values on inputs and will get the value from oj option rather than the input
    // compares references first
    // if any references match it will not do object comparison
    // if there are no reference matches and it is custom element and it is an object or array containing at least one object do deep comparison on the objects
    _setCheckedOnButtons: function(checked, $buttons, type, isMany)
    {
        // loop once for reference comparison
        var checkedFoundCount = this._shallowCompare(checked, $buttons, isMany);

        // loop second for deep object comparison, supported in custom elements only, and only if no references were found in prior check
        if (this._shouldDeepCompare(checked, isMany, checkedFoundCount))
        {
            checkedFoundCount = this._deepCompare(checked, $buttons, isMany, checkedFoundCount);
        }

        return isMany ? checkedFoundCount === checked.length : checkedFoundCount === 1;
    },

    // compares references and increments the match counter
    _shallowCompare: function(checked, $buttons, isMany)
    {
        var self = this;
        var checkedFoundCount = 0;
        $buttons.each(function() {
            // gets value off ojOption in custom element case
            var value = self._getInputValue(this);
            
            // if a buttonset many, checked is an array so do indexOf for reference comparison
            // if a buttonset one, check reference
            if (isMany ? (checked.indexOf(value) > -1) : value === checked)
            {
                this.checked = true;
                checkedFoundCount++;
            }
            else
            {
                this.checked = false;
            }
        });
        return checkedFoundCount;
    },
    // checks if deep comparison needed
    // if buttonset many - deep compare if we have not found all of our values in the value array
    // if buttonset one -  deep compare if we did not find the value
    _shouldDeepCompare: function(checked, isMany, checkedFoundCount)
    {
       return isMany ? checkedFoundCount !== checked.length : !checkedFoundCount;
    },

    // compares objects deeply using
    _deepCompare: function(checked, $buttons, isMany, checkedFoundCount)
    {
        var self = this;
        $buttons.each(function() {
            // gets value off ojOption in custom element case
            var value = self._getInputValue(this);

            // if array loop over values and only compare objects
            if (isMany)
            {
                for (var i = 0; i < checked.length; i++)
                {
                    // need to compare all types within many due to parsing
                    if (self._deepCompareValues(value, checked[i]))
                    {
                        this.checked = true;
                        checkedFoundCount++;
                    }
                }
            }
            else if (self._deepCompareValues(value, checked))
            {
                this.checked = true;
                checkedFoundCount++;
            }
        });
        return checkedFoundCount;
    },

    // performs deep comparison using public method if available otherwise internal compareValues
    _deepCompareValues: function(value, checked){
        // ojCompareValues is a custom comparator that returns 0 if the values are equal
        return ($.type(value) == 'object' && value['ojCompareValues']) ? value['ojCompareValues'](value, checked) === 0 : oj.Object.compareValues(value, checked);
    },

    // gets value from ojOption in custom element case
    _getInputValue: function(input)
    {
        return this._IsCustomElement() ? this._getOjOptionFromInput(input).value : input.value;
    },

    // gets the oj option associated with an input
    _getOjOptionFromInput: function(input)
    {
        // oj option is first child of preceeding label
        return $(input).prev().children()[0];
    },

    // if all buttons are radios with same group, returns value attr of selected radio (string), or null if none selected
    // else if all buttons are checkboxes, returns non-null, possibly empty string array containing values of selected checkboxes
    // else returns undefined.  In the API we use null, not undefined, for this case, so callers should map undefined to null before setting on DOM.
    // NOTE: Called from _InitOptions, so very limited component state is available!
    _getCheckedFromDom: function($buttons) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        var checked = undefined;
        var isRadio = null;
        var name = null;
        var self = this;

        $buttons.each(function(index) {
            // at this point, all previous buttons, if any, were all radios in same group, or were all checkboxes
            var tagName = this.tagName.toLowerCase();
            if (tagName !== "input")
            {
                checked = undefined;
                return false; // stop iterating
            }

            var currentType = this.type.toLowerCase();
            var currentIsRadio;
            var currentName;

            if (currentType === "radio"){
                currentIsRadio = true;
                currentName = this.name.toLowerCase();
            } else if (currentType === "checkbox")
            {
                currentIsRadio = false;
                currentName = null;
            } else
            {
                checked = undefined;
                return false; // stop iterating
            }

            // at this point, current button is input of type radio or checkbox

            // If this is not the first button, then bail if:
            // - this is a radio and previous were checkboxes or vice versa
            // - this is a radio in a different group than previous ones, which can happen
            //   if the group names are different, or if the names are all "", in which
            //   case each radio is in a separate radio group.
            if ( (checked !== undefined) // must be !== not !=
                 && ( currentIsRadio !== isRadio
                      || currentName !== name
                      || (isRadio && !name) ) )
            {
                checked = undefined;
                return false; // stop iterating
            }

            // at this point, all buttons so far including this one are either all radios in same group, or are all checkboxes

            var value = self._getInputValue(this);
            if (checked === undefined) // this is first button
            {
                checked = currentIsRadio
                    ? this.checked ? value : null
                    : this.checked ? [value] : [];
                isRadio = currentIsRadio;
                name = currentName;
            } else if (this.checked)
            {
                if (isRadio)
                    checked = value;
                else
                    checked.push(value);
            } // else not first button and not checked, so leave "checked" at whatever value we set on previous iteration
        });

        return checked;
    },

    _CompareOptionValues: function(option, value1, value2) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
      if (option === 'checked')
      {
        // handle null, string, and (rare) identically equal arrays
        if (value1 === value2)
            return true;
        else // handle arrays.  order needn't be same
            return $.type(value1) === "array" && $.type(value2) === "array"
                   && this._compareArraysAsSets(value1, value2);
      }

      return this._superApply(arguments);
    },

    // Params must be arrays.  Returns true iff the arrays have the same set of elements regardless of order.
    _compareArraysAsSets: function(first, second) {
        return !first.some(function (elem) {
            return second.indexOf(elem)<0;
        }) && !second.some(function (elem) {
            return first.indexOf(elem)<0;
        });
    },

    // remove non oj-options from DOM
    _removeNonOjOptions: function ()
    {
        this.element.children(":not(oj-option)").remove();
    },

    // Add the custom option renderer to all oj options
    _processOjOptions: function ()
    {
        var self = this;
        var ojOptions = this.element.find("oj-option");

        $.each(ojOptions, function(i, option) {
            option["customOptionRenderer"] = self._customOptionRenderer.bind(self);
        });
    },

    // Call render on all the oj options directly
    _customOptionRenderer: function(option)
    {
        this._removeOptionDecoration(option);
        this._addOptionDecoration(option);
    },

    // Wrap the oj option in a label and add an input as a sibling
    _addOptionDecoration: function(option) {
        var ojOption = $(option);

        // add classes to slots and wrap text if applicable
        this._addOptionClasses(option);

        // is the buttonset one or many
        var isOne = this.element[0].tagName === 'OJ-BUTTONSET-ONE';
        var type = isOne ? 'radio': 'checkbox';

        // get values from the oj option to set on the input
        var value = option.value;
        var disabled = option.disabled;

        // create label/input to decorate oj option
        var input = $('<input>');
        var label = $('<label>');

        // input needs an id so label can set the for attribute
        input.uniqueId();
        input.prop('value', value);
        input.prop('type', type);
        input.prop('disabled', disabled);

        // set the label for attribute
        var id = input[0].id;
        label.prop('for', id);

        // if a radio buttonset need a uniform name, create one using the first input id
        if (isOne)
        {
            if (!this._name)
            {
                this._name = '_n_' + id;
            }
            input.prop('name', this._name);
        }

        // rearrange the dom
        ojOption.before(label);  // @HTMLUpdateOK
        label.append(ojOption);  // @HTMLUpdateOK
        label.after(input);  // @HTMLUpdateOK

        // reset buttonset group properties
        this._setup(false);
        this._setCheckedOnDom(this.options.checked, this.$buttons); // throws if checked option invalid
    },

    // remove the dom that we generated excluding the wrapped span
    // destroy the button and leave just what was originally there
    _removeOptionDecoration: function(option) {
        var ojOption = $(option);
        var parentLabel = ojOption.parent();

        if (parentLabel.prop('tagName') === 'LABEL')
        {
            this._removeOptionClasses(option);

            var input = parentLabel.siblings(".oj-button-input")

            input.ojButton('destroy');

            input.removeUniqueId();

            input.remove();
            parentLabel.replaceWith(option);  // @HTMLUpdateOK
        }
    },

    // add needed CSS Classes to slots and wrap the text spans
    _addOptionClasses: function(option) {
        var slotMap = oj.BaseCustomElementBridge.getSlotMap(option);
        var text = slotMap[""] ? slotMap[""] : null;
        var startIcon = slotMap["startIcon"] ? slotMap["startIcon"][0] : null;
        var endIcon = slotMap["endIcon"] ? slotMap["endIcon"][0] : null;

        if (startIcon)
        {
            $(startIcon).addClass('oj-button-icon');
            $(startIcon).addClass('oj-start');
        }

        if (text)
        {
            for (var i = 0; i < text.length; i++)
            {
                var currentText = text[i];
                var wrapperSpan = currentText;
                if (currentText.nodeType === 3)
                {
                    wrapperSpan = $('<span>');
                    $(currentText).replaceWith(wrapperSpan);  // @HTMLUpdateOK
                    wrapperSpan.append(currentText);  // @HTMLUpdateOK
                }

                $(wrapperSpan).addClass('oj-button-text');
            }
        }

        if (endIcon)
        {
            $(endIcon).addClass('oj-button-icon');
            $(endIcon).addClass('oj-end');
        }
    },

    // remove needed CSS Classes to slots
    _removeOptionClasses: function(option) {
        var slotMap = oj.BaseCustomElementBridge.getSlotMap(option);
        var text = slotMap[""] ? slotMap[""] : null;
        var startIcon = slotMap["startIcon"] ? slotMap["startIcon"][0] : null;
        var endIcon = slotMap["endIcon"] ? slotMap["endIcon"][0] : null;

        if (startIcon)
        {
            $(startIcon).removeClass('oj-button-icon');
            $(startIcon).removeClass('oj-start');
        }

        if (text)
        {
            for (var i = 0; i < text.length; i++)
            {
                $(text[i]).removeClass('oj-button-text');
            }
        }

        if (endIcon)
        {
            $(endIcon).removeClass('oj-button-icon');
            $(endIcon).removeClass('oj-end');
        }
    },

    _InitOptions: function(originalDefaults, constructorOptions) {
        this._super(originalDefaults, constructorOptions);

        this.$buttons = this.element.find( this._items );

        // At create time, checked can be set via either option or DOM, i.e. the "checked" properties of the buttons.
        // If app set the option, then that wins over the DOM, in which case _ComponentCreate() will later set that value on the DOM.
        // Else DOM wins, in which case we set the option from the DOM here, with any remaining tasks done later in _ComponentCreate().

        if (!('checked' in constructorOptions)) // if app didn't set option, then set the option from the DOM
        {
            this.initCheckedFromDom = true;
            var checked = this._getCheckedFromDom(this.$buttons); // doesn't rely on any component state, e.g. this.foo

            // 3 cases:  (1) checkbox set, (2) radio group, and (3) none of the above.
            // Conceptually, the default value is [] for (1), and null for (2) and (3).
            // However, the value to which it's actually inited, in the declaration, is null in all 3 cases.
            //
            // Per discussion with architect, we want to update the option value in all cases where it's different than the
            // init value of null, but only fire the event if the new value is different than the conceptual default.
            //
            // The only problem case is when the new value is [], "no checkboxes checked".  To avoid firing the event in this
            // case, while still handling all other cases correctly, we update the ivar to [], the conceptual default,
            // for all checkbox / array cases, before calling option().
            //
            // Writeback is not relevant here, since this code block handles the case where no option value was supplied, in
            // which case there must not be an observable to write to.
            if ($.type(checked) === "array" || this.element[0].tagName === 'OJ-BUTTONSET-MANY')
                this.options.checked = [];

            if (checked !== undefined)
                this.option("checked", checked,  {'_context': {internalSet: true}}); // writeback not needed since "not in constructorOptions" means "not bound"
        }
    },

    _ComponentCreate: function()
    {
        this._super();

        this.element
            .attr(oj.Components._OJ_CONTAINER_ATTR, this['widgetName'])
            .addClass( "oj-buttonset oj-component" )
        this._setRole(this.options.focusManagement);

        if (this._IsCustomElement())
        {
            this._removeNonOjOptions();
            this._processOjOptions();
        }
        
        this._setup(true);
    },

    _NotifyContextMenuGesture: function(menu, event, eventType)
    {
        // Set launcher to the current tabbable button
        // For toggle buttons, launcher must be the hidden focusable input, but for Shift-F10 we want the CM aligned to the Button's root element, not that
        // hidden input.  This is no change from the default for push buttons, since in that case the root element and launcher (input) are the same.
        var currentButton = this.element.find(":oj-button[tabindex=0]");
        this._OpenContextMenu(event, eventType, {
            "launcher": currentButton,
            "position": {"of": eventType==="keyboard" ? currentButton.ojButton("widget") : event}
        });
},

    _propagateDisabled: function( disabled ) {
        disabled = !!disabled;
        this.$buttons.each(function() {
            $( this ).data( "oj-ojButton" ).__setAncestorComponentDisabled(disabled);
        });
    },

    _setRole: function(focusManagement)
    {
        if (focusManagement === "oneTabstop")
            this.element.attr( "role", "toolbar" );
        else
            this.element.removeAttr( "role" );
    },

    _setOption: function( key, value, flags ) // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
    {
        // previously called super at end, so that optionChange (fired at end of super) is fired at very end, but now must call at start, so that
        // when the chroming case calls Button.refreshrefresh(), callee sees the new value of the option.
        this._superApply( arguments );

        if ( key === "disabled" )
            this._propagateDisabled( value );
        else if ( key === "checked" )
        {
            // This "checked" block should run only if app called option(), but not if called because user clicked button,
            // since in the latter case, we know we passed a valid non-undefined value, and DOM is already up to date.
            // Fortunately, this is guaranteed, since _setOption is no longer called in the latter case.
            this._setCheckedOnDom(value, this.$buttons); // throws if checked option invalid

            // Set oj-selected on all buttons' DOM:
            this.$buttons.each(function() {
                $( this ).data( "oj-ojButton" )._applyCheckedStateFromDom(false);
            });
        } else if ( key === "focusManagement" ) {
            this._setRole(value);
        } else if ( key === "chroming" ) {
            _setChromingClass(this.element, value);

            // refresh the buttons to make them re-fetch their chroming option, in case it's still set to the default dynamic getter,
            // which takes its value from the containing buttonset or toolbar if present.
            // TBD: Consider only calling refresh() on children that haven't had their chroming option set, i.e. those still using the dynamic getter.
            this.$buttons.ojButton( "refresh" );
        } else if ( key === "display" ) {
            this.$buttons.ojButton( "option", key, value );
        }
    },

    // TODO: JSDoc says: "refresh() is required ... after a change to the disabled status of any of the buttons in the buttonset."  Instead, shouldn't
    // Button._setOption("disabled") look for a containing Buttonset and do the necessary housekeeping?
    // @inheritdoc
    refresh: function() // Override of public base class method (unlike JQUI).  Method name needn't be quoted since is in externs.js.
    {
        this._super();

        // Should not need to call refresh on oj-options. If the content of an oj-option is modified then
        // app should call refresh on the option not the container and that should trigger a re-render
        // only need to add custom renderer for case where something is added that the buttonset did not 
        // know about before
        if (this._IsCustomElement())
        {
            this._processOjOptions();
        }
        
        // Call this after _super(), which updates the list of containers (toolbar) that the buttonset is in, which must be updated
        // when _setup calls the chroming option getter.
        this._setup(false);
    },

    _setup: function(isCreate) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        var self = this;
        this.isRtl = this._GetReadingDirection() === "rtl";
        _setChromingClass(this.element, this.options.chroming);

        if ((isCreate && !this.initCheckedFromDom && !this._IsCustomElement()))
        {
            // if app provided a "checked" option, it wins over whatever's in the DOM.
            this._setCheckedOnDom(this.options.checked, this.$buttons); // throws if checked option invalid

            // At create time, whether set from DOM or option, checked option and checked props are now in synch, so we just need to
            // set .oj-selected on each button.  This is done below by either the _applyCheckedStateFromDom()
            // call (for existing buttons) or the initializer call (for new buttons).
        }

        if (!isCreate)
        { // it's refresh time.
            // If the DOM's checked state is out of synch with the checked option, it's either because
            // the app directly set the "checked" attr of an existing Button in the Buttonset, which we don't support (they should
            // have used the component API instead), or the set of buttons in the set has changed (possibly because a KO foreach
            // binding added/removed buttons).  In the latter case, we require that the app update the "checked" option, by setting
            // the bound observable (if present) or calling option() (otherwise).

            this.$buttons = this.element.find( this._items ); // only needed at refresh time, since at create time this was already done in _InitOptions()
        }

        // buttonset styling should only be applied if this is a multi-button buttonset.  When a single button is wrapped in a buttonset, that's an implementation
        // detail to get the "checked" option; users still see it as a standalone button, and it should be themed as such.
        this.element.toggleClass("oj-buttonset-multi", this.$buttons.length>1);

        this.$buttons
            // refresh any buttons underneath us that already exist, like JQUI does
            // TBD:  Now that Bset has a checked option, the recursive refreshing of the Bset's buttons is necessary in more cases than before.
            //   Review whether it's still desirable to add a refresh() param that would allow turning off this recursive
            //   refreshing.  That was previously approved, but changing the default (compared to JQUI) was not approved.
            //   The refresh param wouldn't help for creates; for that we'd need to add a Buttonset option.
            //   See also _destroy() code comment.
            .filter( ":oj-button" )
                .ojButton( "refresh" )
                .each(function() {
                    $( this ).data( "oj-ojButton" )._applyCheckedStateFromDom(false); // set .oj-selected
                })
            .end()

            // Create buttons underneath us
            .not( ":oj-button" )
                .ojButton({'display': this.options.display}) // sets .oj-selected
            .end()

            // Update rounded corners, etc.
            .map(function() {
                return $( this ).ojButton( "widget" )[ 0 ];
            })
                .removeClass( "oj-buttonset-first oj-buttonset-last" )
                .filter( ":first" )
                    .addClass( "oj-buttonset-first" )
                .end()
                .filter( ":last" )
                    .addClass( "oj-buttonset-last" )
                .end()
            .end();

        // Must do this after creating the buttons above since callee calls Button API.
        // Must do this before the focus mgmt code, which needs to know which buttons are effectively disabled.
        // Must do this at refresh time, not just create time, in case new buttons were added to the Bset (whether
        // reparented or created e.g. by KO foreach).
        this._propagateDisabled( this.options.disabled );

        if (this.options.focusManagement==="oneTabstop")
        {
            // When buttonset is binding listeners to buttons, use the Buttonset's eventNamespace, not the Button's
            // eventNamespace, to facilitate later unbinding only the Buttonset listeners.

            // For checkbox/radio, we're binding to inputs, not labels.

            // Put listeners on every button, b/c it's too unreliable to put them on the buttonset node and rely on event bubbling.
            // - E.g. bubbling doesn't work for antonym buttons (is this still true after the refactoring?) -- see comment on Button._setLabelOption().
            // - Likewise, focus mgmt can't just break if app listener stops propagation.
            // - Both of these problems still happen when using the delegation / selector overload of .on(); there is no special JQ bubbling magic.

            this.$buttons
                .unbind( "keydown" + this.eventNamespace )
                .bind( "keydown" + this.eventNamespace, function(event) {
                    self._handleKeyDown(event, $(this));
                })

                .unbind( "click" + this.eventNamespace )
                .bind( "click" + this.eventNamespace, function(event) {
                    if ( !$( this ).data( "oj-ojButton" )._IsEffectivelyDisabled() )
                    {
                        // Normally the button will be tabbable after the click, since (a) if we reach here, the clicked button is enabled, and
                        // (b) an unchecked radio before the click will normally be checked after the click.  But just in case it's unchecked
                        // (e.g. due to app listener), we let callee run it thru _mapToTabbable() before using, as usual.
                        self._setTabStop( $(this) );
                    }
                })
                .unbind( "focus" + this.eventNamespace )
                .bind( "focus" + this.eventNamespace, function(event) {
                    self._setTabStop( $(this) );
                });

            // the subset of Buttonset buttons that are enabled.  Effectively disabled buttons are not tabbable.
            this.$enabledButtons = this.$buttons.filter(function(index) {
                return !$( this ).data( "oj-ojButton" )._IsEffectivelyDisabled();
            });

            this._initTabindexes(isCreate);
        }
    },

    // For create, make only the first enabled button tabbable.  (We decided to have Shift-Tab go to first, not last, button.)
    // For refreshes, keep the existing tabstop if we can, otherwise proceed as with create.
    // Either way, if that button is a radio and some radio in its group is checked, make that one tabbable instead.
    // If there are no enabled buttons, makes them all untabbable.
    // No return value.
    _initTabindexes: function(isCreate) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        // even for refreshes where we'll wind up keeping the same tabstop, must make everything untabbable first, to ensure any new buttons become untabbable.
        var $last = $(this._lastTabStop);
        this._lastTabStop = undefined;
        this.$buttons.attr( "tabindex", "-1" );
        var $newTabStop; // callee might map this to radio groupmate

        // TBD: for refreshes when $last is an effectively disabled radio with a checked enabled groupmate and they are in the buttonset, the groupmate would be
        // a valid tabstop, but this defaults to the first.
        if (isCreate || !$last.is(this.$enabledButtons))
        {
            // is create, or is refresh but must treat like create b/c $last is empty, or not enabled, or no longer in the buttonset
            $newTabStop = this.$enabledButtons.first(); // if empty (none enabled), no tabstop will be set
        } else
        {
            // is a refresh, and $last is non-empty and is an enabled button still in the buttonset.  May be a radio whose groupmate
            // has become checked, in which case callee will map it to that groupmate.
            $newTabStop = $last;
        }
        this._setTabStop( $newTabStop );
    },

    // For each button in $button (in our usage always 0-1 button hence $button singular), if that button is an unchecked radio
    // with a checked groupmate (which means it's not tabbable), then map it to the checked one (checked enabled radios are
    // tabbable, and we know it's enabled per the argument below).
    //
    // $button contains 0 or more buttons to map.  Must be enabled since effectively disabled buttons aren't tabbable.
    // Returns the mapped JQ object (which the caller will make the tabstop).
    //
    // We know that this.$enabledButtons contains all buttons in $button, and all of their potentially checked radio-groupmates, since:
    // - The above "enabled" requirement guarantees that $button's contents are all in $enabledButtons.
    // - The prohibition against radio groupmates that are not in the buttonset, and the prohibition against checked disabled groupmates
    //   of enabled radios, guarantee that if $button is a radio, then all of its potentially checked groupmates are enabled and thus in
    //   $enabledButtons.
    //
    // Firefox browser issue:   (TODO: should we doc this?  File FF and/or JET bug?)
    //
    // When this method maps an unchecked radio to its checked groupmate, the caller ensures that the former still has focus, but the latter
    // is the tabstop for when the user tabs out and back in.  When tabbing / Shift-Tabbing from the unchecked radio in the direction of the
    // checked one, Chrome and IE9 are smart enough to tab out of the radio group to the adjacent tabstop as desired.  But in FF, focus goes to
    // the checked one, which is not what we want.
    //
    // Reason: Chrome and IE9 are smart enough never to tab within a radio group.  If focus is in the radio group, Tab and Shift-Tab exit the radio group.
    // But in FF, [a radio is reachable via Tab/Shift-Tab from within the group] if [it's enabled, tabindex != -1, and either checked or has no
    // checked groupmates], i.e. [it would be reachable via Tab/Shift-Tab from outside the group if all its groupmates happened to be untabbable
    // (e.g. disabled)].  In other words, FF is the only one that fails to distinguish between radios that could be valid tabstops from outside, and
    // those that should be valid tabstops from inside.
    //
    // This impl improves on the native behavior.  In FF, in an unchecked radio group, every single radio is a tabstop.  Our use of tabindex=-1
    // guarantees that we never tab within the group in that case.  It's only a checked groupmate that can be tabbed to from within. (So at
    // most one unwanted tabstop.)
    //
    // After much time and effort, the latter issue seems to be infeasible to fix in any robust, non-brittle way.  E.g.:
    // - Clearing all tabstops and restoring on tab-out of buttonset:  FF provides no reliable way to find out whether a blur is exiting the entire
    //   buttonset.  The obvious setTimeout workaround needed to be unacceptably long (e.g. 250ms) and even then was unreliable.  If we ever fail
    //   to restore the tabstop, the buttonset becomes untabbable and inaccessible.
    // - Every other approach had similar robustness issues.
    _mapToTabbable: function( $button ) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        var $enabledButtons = this.$enabledButtons;
        return $button.map(function(index, elem) {
            // Buttons other than radios, and checked radios, are always tabbable if they're enabled, which this method requires.
            // Radios w/ name="" (incl name omitted) are not in a radio group, not even with other radios with w/ name="".  Radios
            // with no groupmates are always tabbable, since either they're checked, or they're unchecked with no checked groupmate.
            if (elem.type != "radio" || elem.checked || elem.name == "")
                return elem;
            else
            {
                // elem is unchecked radio in real (not "") group, which is tabbable iff no groupmate is checked.  Per above doc, we know that
                // all of its potentially checked groupmates are in $enabledButtons.
                var $checkedRadio = _radioGroup(elem, $enabledButtons).filter(":checked");
                return ($checkedRadio.length ? $checkedRadio[0] : elem);
            }
        });
    },

    // Set which button is in the tab sequence.
    // $button should contain 0 or 1 button to be made tabbable (since at most one should be tabbable at a time).
    //   If 0 (i.e. no enabled buttons), all will become untabbable.  If 1, it must be tabbable in every way (e.g. enabled) except possibly
    //   being an unchecked radio with a checked groupmate, which this method will map to its checked groupmate, which
    //   we know is enabled thus tabbable since we require that checked radios with enabled groupmates not be disabled.
    // No return value.
    _setTabStop: function( $button ) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
//        if (!window.setTabStopCounter) window.setTabStopCounter=1; // REMOVE, is only for console.log's
//        console.log("in _setTabStop: " + window.setTabStopCounter++ + ".  Orig (premap) button checked: " + $button[0].checked); // + " and is:");
//        console.log($button[0]);

        $button = this._mapToTabbable( $button );
        var button = $button[0]; // button is undefined iff $button is empty iff we need to clear all tabstops b/c there are no enabled buttons to make tabbable
        var last = this._lastTabStop; // last is undefined iff $(last) is empty iff there are no existing tabstops to clear (b/c _initTabindexes just ran
                                      // or previously there were no enabled buttons to make tabbable)

//        console.log("mapped button and last button are:");  console.log(button);  console.log(last);  console.log(".");

        // Cases: both are undefined: have no tabstops; want to keep it that way (b/c none enabled), so do nothing
        //        both are node X: X is the tabstop; want to keep it that way, so do nothing
        //        last is node X; button is undefined: X is the tabstop; want to clear it w/o replacing it (b/c none enabled).  This logic does that.
        //        last is undefined; button is node X: no existing tabstop; want to make X the tabstop.  This logic does that.
        //        last is node X; button is node Y: X is the tabstop; want to clear it and make Y the tabstop.  This logic does that.
        if ( button !== last) {
            //console.log("setting tab stop to " + $button.attr("id"));  console.log("$(last).length:");  console.log($(last).length);

            $(last).attr( "tabindex", "-1" ); // no-op iff $(last) is empty iff (see comment above)
            $button.attr( "tabindex", "0" ); // no-op iff $button is empty iff (see comment above)
            this._lastTabStop = button;
        }
    },

    // No return value.
    _handleKeyDown: function(event, $button) // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
    {
        switch (event.which) {
            case $.ui.keyCode.UP:   // up arrow
            case $.ui.keyCode.DOWN: // down arrow
                if ( $button.attr("type")!="radio" )
                    break;
                // fall thru for radio only.  See comments below.
            case $.ui.keyCode.LEFT:  // left arrow
            case $.ui.keyCode.RIGHT: // right arrow
                event.preventDefault();

                var $enabledButtons = this.$enabledButtons;
                var length = $enabledButtons.length;
                if (length<2) // nowhere to navigate to; currently focused button is the only enabled one in buttonset
                    break;

                var oldIndex = $enabledButtons.index($button);
                var increment = (event.which == $.ui.keyCode.DOWN || ((event.which == $.ui.keyCode.RIGHT) ^ this.isRtl)) ? 1 : -1;
                var newIndex = (oldIndex+increment+length)%length; // wrap around if at start/end of buttonset

                // When radios are inside an element with role=toolbar, WAI-ARIA doesn't specify how to reconcile its recommended
                // Toolbar behavior (left/right arrows move focus w/o selecting) and radio behavior (all 4 arrow keys both move focus
                // and check/select that radio).  A11y office recommended treating radios in a Buttonset or Toolbar like other buttons:
                // Arrow moves focus without selecting, Spacebar selects, which we prefer too.
                // Previously we did that for only left/right arrows, and disabled up/down arrows, but since both native and WAI-ARIA-
                // compliant radios support up/down arrows, and since JAWS automatically instructs the user to use up/down arrows even
                // when the radio group is inside a role=toolbar, we now support up/down arrows for radios via the fall-thru above
                // (but still focus only, not select).
                $enabledButtons.eq(newIndex).focus();
                break;

            // Don't need Space/Enter handlers.  For all buttons except already-checked radios in some browsers, Space/Enter fire a click event
            // (natively or manually), which already calls _setTabStop.  For checked radios (which are focused if they're getting
            // this key event), _setTabStop has already been called for whichever happened 2nd:  focus (an already checked radio) or
            // check (an already focused radio) via click/Space/Enter.  We don't support programmatically checking the button; it must
            // be done via the "checked" option.
        }
    },

    _destroy: function() // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
    {
        this.element
            .removeClass( "oj-buttonset oj-component " + CHROMING_CLASSES )
            .removeAttr(oj.Components._OJ_CONTAINER_ATTR)
            .removeAttr( "role" );

        if (this.options.focusManagement==="oneTabstop")
            this.$buttons.attr( "tabindex", "0" );

        this.$buttons
            .map(function() {
                return $( this ).ojButton( "widget" )[ 0 ];
            })
                // do .removeClass outside the filter in case button has been destroyed but still has these Buttonset styles on it.
                // TBD: if this has definitely been taken care of already for destroyed buttons, then move inside filter.
                .removeClass( "oj-buttonset-first oj-buttonset-last" )
            .end()

            // Recursively destroy Bset's buttons like JQUI.
            // TBD: The recursive destroy makes it impossible to ungroup the buttons if desired, i.e. destroy the Buttonset without destroying its buttons.
            //   As discussed in _setup() code comment, it was approved to add refresh() and/or destroy() params that would allow turning off
            //   the recursive behavior, but changing the default (compared to JQUI) was not approved.
            //   When not destroying the buttons, must instead restore the buttons to a not-in-buttonset state, i.e. remove Bset stuff, restore any
            //   Button stuff we removed, etc.
            .ojButton( "destroy" );
    }

    // API doc for inherited methods with no JS in this file:

    // Fragments:

    /**
     * <p>The &lt;oj-buttonset-one> element accepts <code class="prettyprint">oj-option</code>s as children. See the <code class="prettyprint">oj-option</code> doc for details about
     * accepted children and slots.</p>
     *
     * @ojchild Default
     * @memberof oj.ojButtonsetOne
     *
     * @example <caption>Initialize the Buttonset with child content specified:</caption>
     * &lt;oj-buttonset-one>
     *   &lt;oj-option value="myValue">myValue&lt;/oj-option>
     * &lt;/oj-buttonset-one>
     */

    /**
     * <p>The &lt;oj-buttonset-many> element accepts <code class="prettyprint">oj-option</code>s as children. See the <code class="prettyprint">oj-option</code> doc for details about
     * accepted children and slots.</p>
     *
     * @ojchild Default
     * @memberof oj.ojButtonsetMany
     *
     * @example <caption>Initialize the Buttonset with child content specified:</caption>
     * &lt;oj-buttonset-many>
     *   &lt;oj-option value="myValue">myValue&lt;/span>
     * &lt;/oj-buttonset-many>
     */

    /**
     * <p>All Buttonset touch interaction is with the individual buttons.  See the [JET Button]{@link oj.ojButton} touch gesture doc.
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojButtonset
     * @instance
     */

    /**
     * <p>JET Buttonset is a single tabstop, with arrow-key navigation within the buttonset, as follows:
     *
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td><kbd>LeftArrow</kbd></td>
     *       <td>Navigate to the previous enabled button on the left, wrapping around at the end.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>RightArrow</kbd></td>
     *       <td>Navigate to the next enabled button on the right, wrapping around at the end.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>See also the [JET Button]{@link oj.ojButton} keyboard doc, for details on interacting with
     * the individual buttons.
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojButtonset
     * @instance
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     *
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-buttonset-width-auto</td>

     *       <td>Forces Buttonset Buttons' widths to be determined by the total width of their icons and label contents,
     *           overriding any theming defaults.
     *           <p>Optionally, specify the overall width of the Buttonset for further width control.</p>
     *           <p>Can be applied to Buttonset's root element, or on an ancestor such as Toolbar or document.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-buttonset-width-equal</td>
     *       <td>Forces Buttonset Buttons' widths to be equal, overriding any theming defaults.
     *           <p>Note that the overall width of the Buttonset defaults to 100%; set the
     *              <code class="prettyprint">max-width</code> (recommended) or
     *              <code class="prettyprint">width</code> of the Buttonset for further width control.
     *           <p>Can be applied to Buttonset's root element, or on an ancestor such as Toolbar or document.
     *       </td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojButtonset
     * @instance
     */
});

// -----------------------------------------------------------------------------
// "private static members" shared by all buttons and buttonsets
// -----------------------------------------------------------------------------

var _lastActive;
var _lastToggleActive;

// "static" namespace for events associated with all buttons on page/form/etc., not just one.  E.g. see form reset handler at top of _ComponentCreate().
// this.eventNamespace, used for individual button instances, is .ojButtonX, whereX = 0, 1, 2, etc.
var BUTTON_EVENT_NAMESPACE = ".ojButton",

    BASE_CLASSES = "oj-button oj-component oj-enabled oj-default", // oj-enabled is a state class, but convenient to include in this var instead
    STATE_CLASSES = "oj-hover oj-active oj-selected",
    TYPE_CLASSES = "oj-button-icons-only oj-button-icon-only oj-button-text-icons oj-button-text-icon-start oj-button-text-icon-end oj-button-text-only",
    CHROMING_CLASSES = "oj-button-full-chrome oj-button-half-chrome oj-button-outlined-chrome",

    _chromingMap = {
        "full": "oj-button-full-chrome",
        "half": "oj-button-half-chrome",
        "outlined": "oj-button-outlined-chrome"
    },

    // SECURITY NOTE: To avoid injection attacks, do NOT compute the class via string concatenation, i.e. don't do "oj-button-" + chroming + "-chrome"
    _setChromingClass = function( $elem, chroming )
    {
        $elem.removeClass(CHROMING_CLASSES)
             .addClass(_chromingMap[chroming]);
    },

    /**
     * In all cases, the return value includes only radios that are an :oj-button, i.e. radios that have been buttonized.
     *
     * Where this method looks for radio groupmates:
     *
     * - If $elems is present (even if empty),
     *     - This method will only look in that set, and will not attempt to weed out any false positives as defined below.
     *       (So in this case, return value includes the specified radio iff it's an :oj-button in $elems.)
     * - Else this method looks in exactly the places where groupmates (including the original radio) would live, i.e. not in
     *   the places false positives would live.  (So in both of the following cases, return value includes the specified radio
     *   iff it's an :oj-button.)  Specifically:
     *     - If radio is in a form, this method will only look in that form.
     *     - Else, this method will look in the radio's document, but not in any forms.
     *
     * Radios w/ name="" (incl name omitted) are not in a radio group (i.e. no SelectOne semantics), not even with other radios with
     * w/ name="".  So if radio is nameless, the return value will include only radio (or nothing at all if it isn't an :oj-button, or
     * if $elems is passed and it doesn't include radio).
     *
     * False positives: radios with nonempty names that match radio's name, but are actually not groupmates (i.e. no SelectOne
     * relationship), e.g. because they're from a different form.
     *
     *
     * @param {!Element} radio  a radio button.  Not a JQ object, other button or element type, or null.
     * @param {jQuery=} $elems  optional JQ object, containing 0 or more elems that aren't necessarily radios or buttons, in which to look for groupmates.
     *                          E.g. the elements in a buttonset or toolbar.  Must not contain any false positives as defined above.
     * @private
     */
    _radioGroup = function( radio, $elems )
    {
        var name = radio.name,
            form = radio.form,
            $radios;
        if ( name )
        {
            name = name.replace( /'/g, "\\'" ); // escape single quotes
            var selector = ":radio[name='" + name + "']:oj-button";
            if ( $elems )
            {
                $radios = $elems.filter( selector );
            } else if ( form )
            {
                $radios = $( form ).find( selector );
            } else
            {
                $radios = $( selector, radio.ownerDocument )
                    .filter(function() {
                        return !this.form;
                    });
            }
        } else
        {
            $radios = ($elems ? $elems.filter( radio ) : $( radio )).filter(":oj-button");
        }
        return $radios;
    },

    // searches actualContainers array for each elem of interestingContainers in order, until one is found,
    // walks up the tree to find that container, and returns its widget constructor.  Returns null if no containers found.
    _findContainer = function( element, actualContainers, interestingContainers )
    {
        for (var i=0; i<interestingContainers.length; ++i) {
            var containerName = interestingContainers[i];
            if (actualContainers.indexOf(containerName) >= 0) {
                // walk up parents until find the container
                for (; ; element = element.parentNode) {
                    var func = oj.Components.__GetWidgetConstructor(element, containerName);
                    if (func) {
                        return func;
                    }
                }
            }
        }
        return null;
    },

    _interestingContainers = {
        "button":    ["ojButtonset", "ojToolbar"],
        "buttonset": ["ojToolbar"]
    },

    _getChromingDefault = function( componentName, element, actualContainers )
    {
        var containerConstructor = _findContainer(element, actualContainers, _interestingContainers[componentName]);

        // If the component is in an interesting container (buttonset or toolbar), then the default chroming is the current value of the chroming option of the nearest such container.
        if (containerConstructor) {
            return containerConstructor("option", "chroming");
        } else {
            // Else, if $___ChromingOptionDefault is set in the current theme, then this expr returns that value for use as the chroming default.
            // Else, returns undefined, so that the prototype default is used.
            return (oj.ThemeUtils.parseJSONFromFontFamily('oj-' + componentName + '-option-defaults') || {})["chroming"];
        }
    };

// Set theme-based defaults
oj.Components.setDefaultOptions({
    'ojButton': {
        'chroming': oj.Components.createDynamicPropertyGetter( function(context) {
            return _getChromingDefault("button", context["element"], context["containers"]);
         })
    },
    'ojButtonset': {
        'chroming': oj.Components.createDynamicPropertyGetter( function(context) {
            return _getChromingDefault("buttonset", context["element"], context["containers"]);
        })
    }
});

}() ); // end of Button / Buttonset wrapper function
(function() {
var ojButtonMeta = {
  "properties": {
    "chroming": {
      "type": "string",
      "enumValues": ["full", "half", "outlined"]
    },
    "disabled": {
      "type": "boolean"
    },
    "display": {
      "type": "string",
      "enumValues": ["all", "icons"]
    }
  },
  "events": {
    "action" : {}
  },
  "extension": {
    _WIDGET_NAME: "ojButton"
  }
};
oj.CustomElementBridge.registerMetadata('oj-button', 'baseComponent', ojButtonMeta);
oj.CustomElementBridge.register('oj-button', {
  'metadata': oj.CustomElementBridge.getMetadata('oj-button'),
  'innerDomFunction': function(element) {
    return element.getAttribute("href") ? 'a' : 'button';
  }
});
})();

(function() {
var ojMenuButtonMeta = {
  "properties": {
    "chroming": {
      "type": "string",
      "enumValues": ["full", "half", "outlined"]
    },
    "disabled": {
      "type": "boolean"
    },
    "display": {
      "type": "string",
      "enumValues": ["all", "icons"]
    }
  },
  "extension": {
    _WIDGET_NAME: "ojButton"
  }
};
oj.CustomElementBridge.registerMetadata('oj-menu-button', 'baseComponent', ojMenuButtonMeta);
oj.CustomElementBridge.register('oj-menu-button', {
  'metadata': oj.CustomElementBridge.getMetadata('oj-menu-button'),
  'innerDomFunction': function(element) {
    return element.getAttribute("href") ? 'a' : 'button';
  }
});
})();

(function() {
var ojButtonSetOneMeta = {
  "properties": {
    "chroming": {
      "type": "string",
      "enumValues": ["full", "half", "outlined"]
    },
    "disabled": {
      "type": "boolean"
    },
    "display": {
      "type": "string",
      "enumValues": ["all", "icons"]
    },
    "focusManagement": {
      "type": "string",
      "enumValues": ["oneTabstop", "none"]
    },
    "value": {
      "type": "any",
      "writeback": true
      }
  },
  "extension": {
    _WIDGET_NAME: "ojButtonset",
    _ALIASED_PROPS: {"value": "checked"}
  }
};
oj.CustomElementBridge.registerMetadata('oj-buttonset-one', 'baseComponent', ojButtonSetOneMeta);
oj.CustomElementBridge.register('oj-buttonset-one', {
  'metadata': oj.CustomElementBridge.getMetadata('oj-buttonset-one')
});
})();

(function() {
var ojButtonSetManyMeta = {
  "properties": {
    "chroming": {
      "type": "string",
      "enumValues": ["full", "half", "outlined"]
    },
    "disabled": {
      "type": "boolean"
    },
    "display": {
      "type": "string",
      "enumValues": ["all", "icons"]
    },
    "focusManagement": {
      "type": "string",
      "enumValues": ["oneTabstop", "none"]
    },
    "value": {
      "type": "Array",
      "writeback": true
      }
  },
  "extension": {
    _WIDGET_NAME: "ojButtonset",
    _ALIASED_PROPS: {"value": "checked"}
  }
};
oj.CustomElementBridge.registerMetadata('oj-buttonset-many', 'baseComponent', ojButtonSetManyMeta);
oj.CustomElementBridge.register('oj-buttonset-many', {
  'metadata': oj.CustomElementBridge.getMetadata('oj-buttonset-many')
});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojChart
 * @augments oj.dvtBaseComponent
 * @since 0.7
 * @ojstatus preview
 * @ojtsignore
 * @ojshortdesc Displays information graphically, making relationships among the data easier to understand.
 * @ojrole application
 *
 * @classdesc
 * <h3 id="chartOverview-section">
 *   JET Chart
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#chartOverview-section"></a>
 * </h3>
 *
 * <p>JET Chart with support for bar, line, area, combination, pie, scatter, bubble, funnel, box plot, and stock
 * chart types.</p>
 *
 * {@ojinclude "name":"warning"}
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-chart
 *   type='bar'
 *   series='[{"name": "Q1 Sales", "items": [50, 60, 20]}]'
 *   groups='["Phone", "Tablets", "Laptops"]'
 * >
 * &lt;/oj-chart>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets. When animating changes to larger
 *    data sets or when animating between data sets, it's recommended to turn off the
 *    <code class="prettyprint">styleDefaults.animationIndicators</code>, since they effectively double the amount of
 *    work required for the animation.
 * </p>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it's recommended that applications only set usable data densities on the chart. For example,
 *    it's not recommended to show more than 500 bars on a 500 pixel wide chart, since the  bars will be unusably thin.
 *    While there are several optimizations within the chart to deal with large data sets, it's always more efficient to
 *    reduce the data set size as early as possible. Future optimizations will focus on improving end user experience as
 *    well as developer productivity for common use cases.
 * </p>
 *
 * <h4>Styling</h4>
 * <p>Use the highest level property available. For example, consider setting styling properties on
 *    <code class="prettyprint">styleDefaults</code> or <code class="prettyprint">series</code>, instead of styling properties
 *    on the individual data items. The chart can take advantage of these higher level properties to apply the style properties on
 *    containers, saving expensive DOM calls.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojChart', $['oj']['dvtBaseComponent'],
  {
    widgetEventPrefix: "oj",
    options: {
      /**
       * An object with the following properties, used to define the series and groups when using a DataProvider to provide data to the chart. Also accepts a Promise for deferred data rendering.
       * @ignore
       * @name data
       * @memberof oj.ojChart
       * @instance
       * @type {Object|Promise|null}
       * @ojsignature {target: "Accessor", value: {GetterType: "Promise<oj.ojChart.Data>|null", SetterType: "oj.ojChart.Data|Promise<oj.ojChart.Data>|null"}, jsdocOverride: true}
       * @default null
       */
       data: null,
       
      /**
       * Triggered during a selection gesture, such as a change in the marquee selection rectangle.
       *
       * @property {Array} items an array containing the string ids of the selected data items
       * @property {Array} selectionData an array containing objects describing the selected data items
       * @property {object} selectionData.data the data of the item, if one was specified
       * @property {Array} selectionData.groupData the group data of the item
       * @property {object} selectionData.seriesData the series data of the item
       * @property {string} endGroup the end group of a marquee selection on a chart with categorical axis
       * @property {string} startGroup the start group of a marquee selection on a chart with categorical axis
       * @property {number} xMax the maximum x value of a marquee selection
       * @property {number} xMin the minimum x value of a marquee selection
       * @property {number} yMax the maximum y value of a marquee selection
       * @property {number} yMin the minimum y value of a marquee selection
       *
       * @expose
       * @event
       * @memberof oj.ojChart
       * @instance
       * @ojbubbles
       */
      selectInput: null,

      /**
       * Triggered after the viewport is changed due to a zoom or scroll operation.
       *
       * @property {string} endGroup the end group of the new viewport on a chart with categorical axis
       * @property {string} startGroup the start group of the new viewport on a chart with categorical axis
       * @property {number} xMax the maximum x value of the new viewport
       * @property {number} xMin the minimum x value of the new viewport
       * @property {number} yMax the maximum y value of the new viewport
       * @property {number} yMin the minimum y value of the new viewport
       *
       * @expose
       * @event
       * @memberof oj.ojChart
       * @instance
       * @ojbubbles
       */
      viewportChange: null,

      /**
       * Triggered during a viewport change gesture, such as a drag operation on the overview window. Note: There are
       * situations where the chart cannot determine whether the viewport change gesture is still in progress, such
       * as with mouse wheel zoom interactions. Standard viewportChange events are fired in these cases.
       *
       * @property {string} endGroup the end group of the new viewport on a chart with categorical axis
       * @property {string} startGroup the start group of the new viewport on a chart with categorical axis
       * @property {number} xMax the maximum x value of the new viewport
       * @property {number} xMin the minimum x value of the new viewport
       * @property {number} yMax the maximum y value of the new viewport
       * @property {number} yMin the minimum y value of the new viewport
       *
       * @expose
       * @event
       * @memberof oj.ojChart
       * @instance
       * @ojbubbles
       */
      viewportChangeInput: null,

      /**
       * Triggered during a drill gesture (double click if selection is enabled, single click otherwise).
       *
       * @property {string} id the id of the drilled object
       * @property {string} series the series id of the drilled object, if applicable
       * @property {string} group the group id of the drilled object, if applicable
       * @property {Object} data  the data object of the drilled item
       * @property {Object} seriesData the data for the series of the drilled object
       * @property {Array} groupData an array of data for the group the drilled object belongs to. For hierarchical groups, it will be an array of outermost to innermost group data related to the drilled object
       *
       * @expose
       * @event
       * @memberof oj.ojChart
       * @instance
       * @ojbubbles
       */
      drill: null
    },

    //** @inheritdoc */
    _CreateDvtComponent: function(context, callback, callbackObj) {
      return dvt.Chart.newInstance(context, callback, callbackObj);
    },

    //** @inheritdoc */
    _ConvertLocatorToSubId : function(locator) {
      var subId = locator['subId'];

      // Convert the supported locators
      if(subId == 'oj-chart-item') {
        // dataItem[seriesIndex][itemIndex]
        subId = 'dataItem[' + locator['seriesIndex'] + '][' + locator['itemIndex'] + ']';
      }
      else if(subId == 'oj-chart-group') {
        // group[index1][index2]...[indexN]
        subId = 'group' + this._GetStringFromIndexPath(locator['indexPath']);
      }
      else if(subId == 'oj-chart-series') {
        // series[index]
        subId = 'series[' + locator['index'] + ']';
      }
      else if(subId == 'oj-chart-axis-title') {
        // xAxis/yAxis/y2Axis:title
        subId = locator['axis'] + ':title';
      }
      else if(subId == 'oj-chart-reference-object') {
        // xAxis/yAxis/y2Axis:referenceObject[index]
        subId = locator['axis'] + ':referenceObject[' + locator['index'] + ']';
      }
      else if(subId == 'oj-legend-item') {
        // legend:section[sectionIndex0][sectionIndex1]...[sectionIndexN]:item[itemIndex]
        subId = 'legend:section' + this._GetStringFromIndexPath(locator['sectionIndexPath']);
        subId += ':item[' + locator['itemIndex'] + ']';
      }
      else if(subId == 'oj-chart-tooltip') {
        subId = 'tooltip';
      }
      else if(subId == 'oj-chart-pie-center-label') {
        subId = 'pieCenterLabel';
      }

      // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
      // support for the old subId syntax in 1.2.0.
      return subId;
    },


    //** @inheritdoc */
    _ProcessOptions: function() {
      this._super();
      var center = this.options['pieCenter'];
      if (center && center['_renderer'])
        center['renderer'] = this._GetTemplateRenderer(center['_renderer'], 'center');
      
      var selection = this.options['selection'];
      // The array<object> type is deprecated since 2.1.0 so for custom elements 
      // we are breaking selection in order to push developers to use the correct syntax
      if(this._IsCustomElement() && selection && typeof selection[0] == 'object')
        this.options['selection'] = null;
    },

    //** @inheritdoc */
    _ConvertSubIdToLocator : function(subId) {
      var locator = {};

      if(subId.indexOf('dataItem') == 0) {
        // dataItem[seriesIndex][itemIndex]
        var indexPath = this._GetIndexPath(subId);

        locator['subId'] = 'oj-chart-item';
        locator['seriesIndex'] = indexPath[0];
        locator['itemIndex'] = indexPath[1];
      }
      else if(subId.indexOf('group') == 0) {
        // group[index1][index2]...[indexN]
        locator['subId'] = 'oj-chart-group';
        locator['indexPath'] = this._GetIndexPath(subId);
      }
      else if(subId.indexOf('series') == 0) {
        // series[index]
        locator['subId'] = 'oj-chart-series';
        locator['index'] = this._GetFirstIndex(subId);
      }
      else if(subId.indexOf('Axis:title') > 0) {
        // xAxis/yAxis/y2Axis:title
        locator['subId'] = 'oj-chart-axis-title';
        locator['axis'] = subId.substring(0, subId.indexOf(':'));
      }
      else if(subId.indexOf('Axis:referenceObject') > 0) {
        // xAxis/yAxis/y2Axis:referenceObject[index]
        locator['subId'] = 'oj-chart-reference-object';
        locator['axis'] = subId.substring(0, subId.indexOf(':'));
        locator['index'] = this._GetFirstIndex(subId);
      }
      else if(subId.indexOf('legend') == 0) {
        if(subId.indexOf(':item') > 0) {
          // legend:section[sectionIndex0][sectionIndex1]...[sectionIndexN]:item[itemIndex]
          var itemStartIndex = subId.indexOf(':item');
          var sectionSubstr = subId.substring(0, itemStartIndex);
          var itemSubstr = subId.substring(itemStartIndex);

          locator['subId'] = 'oj-legend-item';
          locator['sectionIndexPath'] = this._GetIndexPath(sectionSubstr);
          locator['itemIndex'] = this._GetFirstIndex(itemSubstr);
        }
      }
      else if(subId == 'tooltip') {
        locator['subId'] = 'oj-chart-tooltip';
      }
      else if(subId == 'pieCenterLabel') {
        locator['subId'] = 'oj-chart-pie-center-label';
      }

      return locator;
    },

    //** @inheritdoc */
    _ProcessStyles: function() {
      this._super();
      if (!this.options['styleDefaults'])
        this.options['styleDefaults'] = {};
      if (!this.options['styleDefaults']['colors']) {
        var handler = new oj.ColorAttributeGroupHandler();
        // override default colors with css attribute group colors
        this.options['styleDefaults']['colors'] = handler.getValueRamp();
      }
    },
    
    //** @inheritdoc */
    _GetComponentRendererOptions: function() {
      return ['tooltip/renderer', 'pieCenter/renderer'];
    },
    
    //** @inheritdoc */
    _GetComponentStyleClasses: function() {
      var styleClasses = this._super();
      styleClasses.push('oj-chart');
      return styleClasses;
    },

    //** @inheritdoc */
    _GetChildStyleClasses: function() {
      var styleClasses = this._super();
      styleClasses['oj-chart-data-label'] = {'path': 'styleDefaults/dataLabelStyle', 'property': 'TEXT'};
      styleClasses['oj-chart-data-cursor-line'] = {'path': 'styleDefaults/dataCursor/lineColor', 'property': 'color'};
      styleClasses['oj-chart-stack-label'] = {'path': 'styleDefaults/stackLabelStyle', 'property': 'TEXT'};
      styleClasses['oj-chart-pie-center-label'] = {'path': 'pieCenter/labelStyle', 'property': 'TEXT'};
      styleClasses['oj-chart-slice-label'] = {'path': 'styleDefaults/sliceLabelStyle', 'property': 'TEXT'};
      styleClasses['oj-chart-stock-falling'] = {'path': 'styleDefaults/stockFallingColor', 'property': 'background-color'};
      styleClasses['oj-chart-stock-range'] = {'path': 'styleDefaults/stockRangeColor', 'property': 'background-color'};
      styleClasses['oj-chart-stock-rising'] = {'path': 'styleDefaults/stockRisingColor', 'property': 'background-color'};
      styleClasses['oj-chart-tooltip-label'] = {'path': 'styleDefaults/tooltipLabelStyle', 'property': 'TEXT'};
      styleClasses['oj-chart-xaxis-tick-label'] = {'path': 'xAxis/tickLabel/style', 'property': 'TEXT'};
      styleClasses['oj-chart-xaxis-title'] = {'path': 'xAxis/titleStyle', 'property': 'TEXT'};
      styleClasses['oj-chart-yaxis-tick-label'] = {'path': 'yAxis/tickLabel/style', 'property': 'TEXT'};
      styleClasses['oj-chart-yaxis-title'] = {'path': 'yAxis/titleStyle', 'property': 'TEXT'};
      styleClasses['oj-chart-y2axis-tick-label'] = {'path': 'y2Axis/tickLabel/style', 'property': 'TEXT'};
      styleClasses['oj-chart-y2axis-title'] = {'path': 'y2Axis/titleStyle', 'property': 'TEXT'};

      // Legend
      styleClasses['oj-legend'] = {'path': 'legend/textStyle', 'property': 'TEXT'};
      styleClasses['oj-legend-title'] = {'path': 'legend/titleStyle', 'property': 'TEXT'};

      return styleClasses;
    },

    //** @inheritdoc */
    _GetEventTypes : function() {
      return ['drill', 'optionChange', 'selectInput', 'viewportChange', 'viewportChangeInput'];
    },

    //** @inheritdoc */
    _GetTranslationMap: function() {
      // The translations are stored on the options object.
      var translations = this.options['translations'];

      // Safe to modify super's map because function guarentees a new map is returned
      var ret = this._super();
      ret['DvtChartBundle.DEFAULT_GROUP_NAME'] = translations['labelDefaultGroupName'];
      ret['DvtChartBundle.LABEL_SERIES'] = translations['labelSeries'];
      ret['DvtChartBundle.LABEL_GROUP'] = translations['labelGroup'];
      ret['DvtChartBundle.LABEL_VALUE'] = translations['labelValue'];
      ret['DvtChartBundle.LABEL_TARGET_VALUE'] = translations['labelTargetValue'];
      ret['DvtChartBundle.LABEL_X'] = translations['labelX'];
      ret['DvtChartBundle.LABEL_Y'] = translations['labelY'];
      ret['DvtChartBundle.LABEL_Z'] = translations['labelZ'];
      ret['DvtChartBundle.LABEL_PERCENTAGE'] = translations['labelPercentage'];
      ret['DvtChartBundle.LABEL_LOW'] = translations['labelLow'];
      ret['DvtChartBundle.LABEL_HIGH'] = translations['labelHigh'];
      ret['DvtChartBundle.LABEL_OPEN'] = translations['labelOpen'];
      ret['DvtChartBundle.LABEL_CLOSE'] = translations['labelClose'];
      ret['DvtChartBundle.LABEL_VOLUME'] = translations['labelVolume'];
      ret['DvtChartBundle.LABEL_Q1'] = translations['labelQ1'];
      ret['DvtChartBundle.LABEL_Q2'] = translations['labelQ2'];
      ret['DvtChartBundle.LABEL_Q3'] = translations['labelQ3'];
      ret['DvtChartBundle.LABEL_MIN'] = translations['labelMin'];
      ret['DvtChartBundle.LABEL_MAX'] = translations['labelMax'];
      ret['DvtChartBundle.LABEL_OTHER'] = translations['labelOther'];
      ret['DvtChartBundle.PAN'] = translations['tooltipPan'];
      ret['DvtChartBundle.MARQUEE_SELECT'] = translations['tooltipSelect'];
      ret['DvtChartBundle.MARQUEE_ZOOM'] = translations['tooltipZoom'];
      ret['DvtUtilBundle.CHART'] = translations['componentName'];
      return ret;
    },

    //** @inheritdoc */
    _HandleEvent: function(event) {
      var type = event['type'];
      if (type === 'selection') {
        var selection = event['selection'];
        if (selection) {
          // Convert the graph selection context into the JET context
          var selectedItems = [];
          var selectionData = [];
          for (var i = 0; i < selection.length; i++) {
            var selectedItem;
            if (this._IsCustomElement())
              selectedItem = selection[i]['id'];
            else {
              selectedItem = {'id': selection[i]['id'],
                'series': selection[i]['series'],
                'group': selection[i]['group']};
            }
            var selectedItemData = {'data': selection[i]['data'],
              'seriesData': selection[i]['seriesData'],
              'groupData': selection[i]['groupData']};
            selectedItems.push(selectedItem);
            selectionData.push(selectedItemData);
          }

          var selectPayload = {
            'endGroup': event['endGroup'], 'startGroup': event['startGroup'],
            'xMax': event['xMax'], 'xMin': event['xMin'],
            'yMax': event['yMax'], 'yMin': event['yMin'],
            'y2Max': event['y2Max'], 'y2Min': event['y2Min'],
            'selectionData': selectionData
          };

          if (!this._IsCustomElement())
            selectPayload['component'] = event['component'];

          // Update the options selection state if the user interaction is complete
          if(event['complete'])
            this._UserOptionChange('selection', selectedItems, selectPayload);
          else {
            selectPayload['items'] = selectedItems;
            this._trigger('selectInput', null, selectPayload);
          }
        }
      }
      else if (type === 'viewportChange') {
        var viewportChangePayload = {'endGroup': event['endGroup'], 'startGroup': event['startGroup'],
                                     'xMax': event['xMax'], 'xMin': event['xMin'],
                                     'yMax': event['yMax'], 'yMin': event['yMin']};

        // Maintain the viewport state
        // TODO we should be firing option change event for this, but it doesn't support nested props yet.
        if(event['complete']) {
          // Ensure the axis options both exist
          if(!this.options['xAxis'])
            this.options['xAxis'] = {};

          if(!this.options['yAxis'])
            this.options['yAxis'] = {};

          // X-Axis: Clear the start and end group because min/max more accurate.
          this.options['xAxis']['viewportStartGroup'] = null;
          this.options['xAxis']['viewportEndGroup'] = null;
          if(event['xMin'] != null && event['xMax'] != null) {
            this.options['xAxis']['viewportMin'] = event['xMin'];
            this.options['xAxis']['viewportMax'] = event['xMax'];
          }

          // Y-Axis
          if(event['yMin'] != null && event['yMax'] != null) {
            this.options['yAxis']['viewportMin'] = event['yMin'];
            this.options['yAxis']['viewportMax'] = event['yMax'];
          }
        }

        this._trigger(event['complete'] ? 'viewportChange' : 'viewportChangeInput', null, viewportChangePayload);
      }
      else if (type === 'drill') {
        this._trigger('drill', null, {'id': event['id'], 'series': event['series'], 'group': event['group'], 'data': event['data'], 'seriesData': event['seriesData'],'groupData': event['groupData'],'component': event['component']});
      }
      else {
        this._super(event);
      }
    },

    //** @inheritdoc */
    _LoadResources: function() {
      // Ensure the resources object exists
      if (this.options['_resources'] == null)
        this.options['_resources'] = {};

      var resources = this.options['_resources'];

      // Add images
      // TODO these should be defined in the skin instead
      resources['overviewGrippy'] = oj.Config.getResourceUrl('resources/internal-deps/dvt/chart/drag_horizontal.png');

      // Add cursors
      resources['panCursorDown'] = oj.Config.getResourceUrl('resources/internal-deps/dvt/chart/hand-closed.cur');
      resources['panCursorUp'] = oj.Config.getResourceUrl('resources/internal-deps/dvt/chart/hand-open.cur');

      // Drag button images
      resources['panUp'] = 'oj-chart-pan-icon';
      resources['panUpHover'] = 'oj-chart-pan-icon oj-hover';
      resources['panDown'] = 'oj-chart-pan-icon oj-active';
      resources['panDownHover'] = 'oj-chart-pan-icon oj-hover oj-active';
      resources['selectUp'] = 'oj-chart-select-icon';
      resources['selectUpHover'] = 'oj-chart-select-icon oj-hover';
      resources['selectDown'] = 'oj-chart-select-icon oj-active';
      resources['selectDownHover'] = 'oj-chart-select-icon oj-hover oj-active';
      resources['zoomUp'] = 'oj-chart-zoom-icon';
      resources['zoomUpHover'] = 'oj-chart-zoom-icon oj-hover';
      resources['zoomDown'] = 'oj-chart-zoom-icon oj-active';
      resources['zoomDownHover'] = 'oj-chart-zoom-icon oj-hover oj-active';
    },
     
    /**
     * Returns the chart title.
     * @ignore
     * @return {string} The chart title
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getTitle: function() {
      var auto = this._component.getAutomation();
      return auto.getTitle();
    },

    /**
     * Returns the group corresponding to the given index
     * @param {string} groupIndex the group index
     * @return {string} The group name corresponding to the given group index
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getGroup: function(groupIndex) {
      var auto = this._component.getAutomation();
      return auto.getGroup(groupIndex);
    },

    /**
     * Returns the series corresponding to the given index
     * @param {string} seriesIndex the series index
     * @return {string} The series name corresponding to the given series index
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getSeries: function(seriesIndex) {
      var auto = this._component.getAutomation();
      return auto.getSeries(seriesIndex);
    },

    /**
     * Returns number of groups in the chart data
     * @return {number} The number of groups
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getGroupCount: function() {
      var auto = this._component.getAutomation();
      return auto.getGroupCount();
    },

    /**
     * Returns number of series in the chart data
     * @return {number} The number of series
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getSeriesCount: function() {
      var auto = this._component.getAutomation();
      return auto.getSeriesCount();
    },

    /**
     * Returns an object with the following properties for automation testing verification of the data item with
     * the specified series and group indices.
     *
     * @param {number} seriesIndex
     * @param {number} groupIndex
     * @property {string} borderColor
     * @property {string} color
     * @property {number} close The closing value for a stock item
     * @property {string} group The group id.
     * @property {number} high The high value for a range or stock item
     * @property {string} label
     * @property {number} low  The low value for a range or stock item
     * @property {number} open The opening value for a stock item
     * @property {boolean} selected
     * @property {string} series The series id.
     * @property {number} targetValue The target value for a funnel slice.
     * @property {string} tooltip
     * @property {number} value
     * @property {number} volume  The volume of a stock item
     * @property {number} x
     * @property {number} y
     * @property {number} z
     * @return {Object|null} An object containing properties for the data item, or null if none exists.
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getDataItem: function(seriesIndex, groupIndex) {
      return this._component.getAutomation().getDataItem(seriesIndex, groupIndex);
    },

    /**
     * Returns an object with the following properties for automation testing verification of the chart legend.
     *
     * @property {Object} bounds An object containing the bounds of the legend.
     * @property {number} bounds.x
     * @property {number} bounds.y
     * @property {number} bounds.width
     * @property {number} bounds.height
     * @property {string} title
     * @return {Object} An object containing properties for the chart legend.
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getLegend: function() {
      return this._component.getAutomation().getLegend();
    },

    /**
     * Returns an object with the following properties for automation testing verification of the chart plot area.
     *
     * @property {Object} bounds An object containing the bounds of the plot area.
     * @property {number} bounds.x
     * @property {number} bounds.y
     * @property {number} bounds.width
     * @property {number} bounds.height
     * @return {Object} An object containing properties for the chart plot area.
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getPlotArea: function() {
      return this._component.getAutomation().getPlotArea();
    },

    /**
     * Returns an object with the following properties for automation testing verification of the x axis.
     *
     * @property {Object} bounds An object containing the bounds of the legend.
     * @property {number} bounds.x
     * @property {number} bounds.y
     * @property {number} bounds.width
     * @property {number} bounds.height
     * @property {string} title
     * @property {Function(number, number)} getPreferredSize Returns the preferred size of the axis, given the available
     *   width and height. This value can be passed into the <code class="prettyprint">size</code> and
     *   <code class="prettyprint">maxSize</code> options of the axis. A re-render must be triggered by calling
     *   <code class="prettyprint">refresh</code> after invoking this function.
     * @property {number} getPreferredSize.width
     * @property {number} getPreferredSize.height
     * @return {Object} An object containing properties for the x axis.
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getXAxis: function() {
      return this._component.getAutomation().getXAxis();
    },

    /**
     * Returns an object with the following properties for automation testing verification of the y axis.
     *
     * @property {Object} bounds An object containing the bounds of the legend.
     * @property {number} bounds.x
     * @property {number} bounds.y
     * @property {number} bounds.width
     * @property {number} bounds.height
     * @property {string} title
     * @property {Function(number, number)} getPreferredSize Returns the preferred size of the axis, given the available
     *   width and height. This value can be passed into the <code class="prettyprint">size</code> and
     *   <code class="prettyprint">maxSize</code> options of the axis. A re-render must be triggered by calling
     *   <code class="prettyprint">refresh</code> after invoking this function.
     * @property {number} getPreferredSize.width
     * @property {number} getPreferredSize.height
     * @return {Object} An object containing properties for the y axis.
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getYAxis: function() {
      return this._component.getAutomation().getYAxis();
    },

    /**
     * Returns an object with the following properties for automation testing verification of the y2 axis.
     *
     * @property {Object} bounds An object containing the bounds of the legend.
     * @property {number} bounds.x
     * @property {number} bounds.y
     * @property {number} bounds.width
     * @property {number} bounds.height
     * @property {string} title
     * @property {Function(number, number)} getPreferredSize Returns the preferred size of the axis, given the available
     *   width and height. This value can be passed into the <code class="prettyprint">size</code> and
     *   <code class="prettyprint">maxSize</code> options of the axis. A re-render must be triggered by calling
     *   <code class="prettyprint">refresh</code> after invoking this function.
     * @property {number} getPreferredSize.width
     * @property {number} getPreferredSize.height
     * @return {Object} An object containing properties for the y2 axis.
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getY2Axis: function() {
      return this._component.getAutomation().getY2Axis();
    },

    /**
     * Returns the x, y, and y2 axis values at the specified X and Y coordinate.
     * @param {number} x The X coordinate relative to the component.
     * @param {number} y The Y coordinate relative to the component.
     * @return {Object} An object containing the "x", "y", and "y2" axis values.
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getValuesAt: function(x, y) {
      return this._component.getValuesAt(x, y);
    },

    /**
     * {@ojinclude "name":"nodeContextDoc"}
     * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
     * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
     *
     * @example {@ojinclude "name":"nodeContextExample"}
     *
     * @expose
     * @instance
     * @memberof oj.ojChart
     */
    getContextByNode: function(node)
    {
      // context objects are documented with @ojnodecontext
      var context = this.getSubIdByNode(node);
      if (context && context['subId'] !== 'oj-chart-tooltip')
        return context;

      return null;
    },

    //** @inheritdoc */
    _GetComponentDeferredDataPaths : function() {
      return {'root': ['groups', 'series'], 'data': ['groups', 'series']};
    },

    //** @inheritdoc */
    _CompareOptionValues: function(option, value1, value2)
    {
      if (option == 'dataCursorPosition')
        return oj.Object.compareValues(value1, value2);
       else
        return this._super(option, value1, value2);
    }
  });

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="5">Data Item</td>
 *       <td rowspan="2"><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td>Drill when <code class="prettyprint">drilling</code> is enabled and <code class="prettyprint">selectionMode</code> is <code class="prettyprint">none</code>.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Double Tap</kbd></td>
 *       <td>Drill when <code class="prettyprint">drilling</code> is enabled and <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *     <tr>
 *       <td>Categorical Axis Item</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Drill when <code class="prettyprint">drilling</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2">Legend Item</td>
 *       <td rowspan="2"><kbd>Tap</kbd></td>
 *       <td>Drill when <code class="prettyprint">drilling</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td>Filter when <code class="prettyprint">hideAndShowBehavior</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="4">Plot Area</td>
 *       <td rowspan="2"><kbd>Drag</kbd></td>
 *       <td>Pan when panning is enabled and toggled into that mode.</td>
 *     </tr>
 *     <tr>
 *       <td>Marquee select when <code class="prettyprint">selectionMode</code> is <code class="prettyprint">multiple</code> and toggled into that mode.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Pinch-close</kbd></td>
 *       <td>Zoom out when zooming is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Spread-open</kbd></td>
 *       <td>Zoom in when zooming is enabled.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojChart
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus and selection to previous data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus and selection to next data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus and selection to previous data item (on left).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus and selection to next data item (on right).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + UpArrow</kbd></td>
 *       <td>Move focus and multi-select previous data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + DownArrow</kbd></td>
 *       <td>Move focus and multi-select next data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + LeftArrow</kbd></td>
 *       <td>Move focus and multi-select previous data item (on left).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + RightArrow</kbd></td>
 *       <td>Move focus and multi-select next data item (on right).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + UpArrow</kbd></td>
 *       <td>Move focus to previous data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + DownArrow</kbd></td>
 *       <td>Move focus to next data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + LeftArrow</kbd></td>
 *       <td>Move focus to previous data item (on left), without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + RightArrow</kbd></td>
 *       <td>Move focus to next data item (on right), without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Spacebar</kbd></td>
 *       <td>Multi-select data item with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>=</kbd> or <kbd>+</kbd></td>
 *       <td>Zoom in one level if zooming is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>-</kbd> or <kbd>_</kbd></td>
 *       <td>Zoom out one level if zooming is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>PageUp</kbd></td>
 *       <td>Pan up if scrolling is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>PageDown</kbd></td>
 *       <td>Pan down if scrolling is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + PageUp</kbd></td>
 *       <td>Pan left in left-to-right locales. Pan right in right-to-left locales.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + PageDown</kbd></td>
 *       <td>Pan right in left-to-right locales. Pan left in right-to-left locales.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Drill on data item, categorical axis label, or legend item when <code class="prettyprint">drilling</code> is enabled.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojChart
 */
  
/**
 * An array of objects with the following properties, used to define series labels and override series styles. Only a single series is supported for stock charts. Also accepts a Promise for deferred data rendering.
 * @expose
 * @ojtsignore
 * @name series
 * @memberof oj.ojChart
 * @instance
 * @type {Array.<Object>|Promise}
 * @default null
 */
/**
 * The name of the series, displayed in the legend and tooltips.
 * @expose
 * @ojtsignore
 * @name series[].name
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An array of values or an array of objects with the following properties that defines the data items for the series.
 * @expose
 * @ojtsignore
 * @name series[].items
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>|Array.<number>}
 * @default null
 */
/**
 * An array of nested data items to be used for defining the markers for outliers or additional data items of a box plot. 
 * @expose
 * @ojtsignore
 * @name series[].items[].items
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>|Array.<number>}
 * @default null
 */
/**
 * (Optional) The id of the data item. This id will be provided as part of the context for events on the chart.
 * @expose
 * @ojtsignore
 * @name series[].items[].id
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The x value. Mainly used for scatter and bubble chart and to specify the date for mixed-frequency time axis.
 * For categorical axis, if the x value is not specified, it will default to the item index.
 * For regular time axis, if the x value is not specified, it will default to the group name of the item.
 * @expose
 * @ojtsignore
 * @name series[].items[].x
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The y value. Also the primary value for charts without a y-Axis, such as pie charts.
 * @expose
 * @ojtsignore
 * @name series[].items[].y
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The z value. Defines the bubble radius for a bubble chart, as well as the width of a bar or a box plot item.
 * @expose
 * @ojtsignore
 * @name series[].items[].z
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The low value for range bar/area, stock candlestick, or box plot item. Define 'low' and 'high' instead of 'value' or 'y' to create a range bar/area chart.
 * @expose
 * @ojtsignore
 * @name series[].items[].low
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The high value for range bar/area, stock candlestick, or box plot item. Define 'low' and 'high' instead of 'value' or 'y' to create a range bar/area chart.
 * @expose
 * @ojtsignore
 * @name series[].items[].high
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The open value for stock candlestick.
 * @expose
 * @ojtsignore
 * @name series[].items[].open
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The close value for stock candlestick. When bar, line, or area series type are used on a stock chart, this value is displayed.
 * @expose
 * @ojtsignore
 * @name series[].items[].close
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The value for stock volume bar. When this value is provided, the volume bar is displayed on the y2 axis.
 * @expose
 * @ojtsignore
 * @name series[].items[].volume
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The first quartile value for box plot.
 * @expose
 * @ojtsignore
 * @name series[].items[].q1
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The second quartile (median) value for box plot.
 * @expose
 * @ojtsignore
 * @name series[].items[].q2
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The third quartile value for box plot.
 * @expose
 * @ojtsignore
 * @name series[].items[].q3
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The description of this object. This is used for accessibility and also for customizing the tooltip text.
 * @expose
 * @ojtsignore
 * @name series[].items[].shortDesc
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The color of the data item.
 * @expose
 * @ojtsignore
 * @name series[].items[].color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border color of the data item. For funnel and pyramid charts, it is used for the slice border.
 * @expose
 * @ojtsignore
 * @name series[].items[].borderColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border width of the data item. For funnel and pyramid charts, it is used for the slice border.
 * @expose
 * @ojtsignore
 * @name series[].items[].borderWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The pattern used to fill the data item. A solid fill is used by default, unless the seriesEffect is 'pattern'.
 * @expose
 * @ojtsignore
 * @name series[].items[].pattern
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The CSS style class to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @expose
 * @ojtsignore
 * @name series[].items[].svgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @expose
 * @ojtsignore
 * @name series[].items[].svgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * Defines whether the data marker is displayed. Only applies to line, area, scatter, and bubble series. If auto, the markers will be displayed whenever the data points are not connected by a line.
 * @expose
 * @ojtsignore
 * @name series[].items[].markerDisplayed
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The shape of the data markers. In addition to the built-in shapes, it may also take SVG path commands to specify a custom shape. The chart will style the custom shapes the same way as built-in shapes, supporting properties like color and borderColor and applying hover and selection effects. Only 'auto' is supported for range series. 
 * @expose
 * @ojtsignore
 * @name series[].items[].markerShape
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "square"
 * @ojvalue {string} "circle"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "human"
 * @ojvalue {string} "star"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The size of the data markers. Does not apply to bubble charts, which calculate marker size based on the z values.
 * @expose
 * @ojtsignore
 * @name series[].items[].markerSize
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 *  The URI of the custom image. If specified, it takes precedence over shape. 
 * @expose
 * @ojtsignore
 * @name series[].items[].source
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The optional URI for the hover state. If not specified, the source image will be used. 
 * @expose
 * @ojtsignore
 * @name series[].items[].sourceHover
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The optional URI for the selected state. If not specified, the source image will be used. 
 * @expose
 * @ojtsignore
 * @name series[].items[].sourceSelected
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The optional URI for the hover selected state. If not specified, the source image will be used. 
 * @expose
 * @ojtsignore
 * @name series[].items[].sourceHoverSelected
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The label for the data item. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively. Not supported for box plot or candlestick.
 * @expose
 * @ojtsignore
 * @name series[].items[].label
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 * @ojtranslatable
 */
/**
 * The position of the data label. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively. The 'outsideSlice' value only applies to pie charts. The 'aboveMarker', 'belowMarker', 'beforeMarker', and 'afterMarker' values only apply to line, area, scatter, and bubble series. The 'insideBarEdge' and 'outsideBarEdge' values only apply to non-polar bar series. Stacked bars do not support 'outsideBarEdge'. The chart does not currently adjust layout to fit labels within the plot area or deal with any overlaps between labels. 
 * @expose
 * @ojtsignore
 * @name series[].items[].labelPosition
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @ojvalue {string} "center"
 * @ojvalue {string} "outsideSlice"
 * @ojvalue {string} "aboveMarker"
 * @ojvalue {string} "belowMarker"
 * @ojvalue {string} "beforeMarker"
 * @ojvalue {string} "afterMarker"
 * @ojvalue {string} "insideBarEdge"
 * @ojvalue {string} "outsideBarEdge"
 * @ojvalue {string} "none"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The CSS style object defining the style of the data label. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively.
 * @expose
 * @ojtsignore
 * @name series[].items[].labelStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {object|Array.<Object>}
 * @default null
 */
/**
 * An optional array of category strings corresponding to this data item. This enables highlighting and filtering of individual data items through interactions with the legend or other visualization elements. If not defined, series categories are used.
 * @expose
 * @ojtsignore
 * @name series[].items[].categories
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The value for this data item. Corresponding to the y value for bar, line, area, and combo charts and the slice values for pie, funnel and pyramid charts. Null can be specified to skip a data point.
 * @expose
 * @ojtsignore
 * @name series[].items[].value
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The target value for a funnel chart. When this is set, the value attribute defines the filled area within the slice and this represents the value of the whole slice.
 * @expose
 * @ojtsignore
 * @name series[].items[].targetValue
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 *  Whether drilling is enabled for the data item. Drillable objects will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). To enable drilling for all data items at once, use the drilling attribute in the top level. 
 * @expose
 * @ojtsignore
 * @name series[].items[].drilling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default "inherit"
 */
/**
 * An object containing the style properties of the box plot item.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  The color of the Q2 segment of the box.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.q2Color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The CSS style class to apply to the Q2 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q2Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.q2SvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The CSS inline style to apply to the Q2 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q2Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.q2SvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  The CSS inline style to apply to the Q3 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q3Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.q3SvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  The CSS style class to apply to the Q3 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q3Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.q3SvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the whisker stems.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.whiskerSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS inline style to apply to the whisker stems.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.whiskerSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class to apply to the whisker ends.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.whiskerEndSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS inline style to apply to the whisker ends.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.whiskerEndSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  Specifies the length of the whisker ends in pixels (e.g. '9px') or as a percentage of the box width (e.g. '50%'). The specified length will be rounded down to an odd number of pixels to ensure symmetry. 
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.whiskerEndLength
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the median line.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.medianSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS inline style to apply to the median line.
 * @expose
 * @ojtsignore
 * @name series[].items[].boxPlot.medianSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The id of the series. Defaults to the name or the series index if not specified.
 * @expose
 * @ojtsignore
 * @name series[].id
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The type of data objects to display for this series. Only applies to bar, line, area, stock, box plot, and combo charts.
 * @expose
 * @ojtsignore
 * @name series[].type
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "bar"
 * @ojvalue {string} "line"
 * @ojvalue {string} "area"
 * @ojvalue {string} "lineWithArea"
 * @ojvalue {string} "candlestick"
 * @ojvalue {string} "boxPlot"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The color of the series.
 * @expose
 * @ojtsignore
 * @name series[].color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border color of the series.
 * @expose
 * @ojtsignore
 * @name series[].borderColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border width of the series.
 * @expose
 * @ojtsignore
 * @name series[].borderWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The area color of the series. Only applies if series type is area or lineWithArea.
 * @expose
 * @ojtsignore
 * @name series[].areaColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply if series type is area or lineWithArea. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @expose
 * @ojtsignore
 * @name series[].areaSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply if series type is area or lineWithArea. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @expose
 * @ojtsignore
 * @name series[].areaSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class to apply to the series. For series of type lineWithArea, this style will only be applied to the line if areaClassName is also specified. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @expose
 * @ojtsignore
 * @name series[].svgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the series. For series of type lineWithArea, this style will only be applied to the line if areaStyle is also specified.The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @expose
 * @ojtsignore
 * @name series[].svgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class to apply to the data markers.The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the marker color attribute.
 * @expose
 * @ojtsignore
 * @name series[].markerSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the data markers. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the marker color attribute.
 * @expose
 * @ojtsignore
 * @name series[].markerSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The pattern used to fill the series. A solid fill is used by default, unless the seriesEffect is 'pattern'.
 * @expose
 * @ojtsignore
 * @name series[].pattern
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The shape of the data markers. In addition to the built-in shapes, it may also take SVG path commands to specify a custom shape. The chart will style the custom shapes the same way as built-in shapes, supporting properties like color and borderColor and applying hover and selection effects. Only 'auto' is supported for range series. 
 * @expose
 * @ojtsignore
 * @name series[].markerShape
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "square"
 * @ojvalue {string} "circle"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "human"
 * @ojvalue {string} "star"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The color of the data markers, if different from the series color.
 * @expose
 * @ojtsignore
 * @name series[].markerColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the data markers should be displayed. Only applies to line, area, scatter, and bubble series. If auto, the markers will be displayed whenever the data points are not connected by a line.
 * @expose
 * @ojtsignore
 * @name series[].markerDisplayed
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The size of the data markers.
 * @expose
 * @ojtsignore
 * @name series[].markerSize
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The width of the data line. Only applies to line, lineWithArea, scatter, and bubble series.
 * @expose
 * @ojtsignore
 * @name series[].lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The line style of the data line. Only applies to line, lineWithArea, scatter, and bubble series.
 * @expose
 * @ojtsignore
 * @name series[].lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The line type of the data line or area. Only applies to line, area, scatter, and bubble series. centeredStepped and centeredSegmented are not supported for polar, scatter, and bubble charts. 
 * @expose
 * @ojtsignore
 * @name series[].lineType
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "straight"
 * @ojvalue {string} "curved"
 * @ojvalue {string} "stepped"
 * @ojvalue {string} "centeredStepped"
 * @ojvalue {string} "segmented"
 * @ojvalue {string} "centeredSegmented"
 * @ojvalue {string} "none"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 *  The URI of the custom image. If specified, it takes precedence over shape. 
 * @expose
 * @ojtsignore
 * @name series[].source
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The optional URI for the hover state. If not specified, the source image will be used. 
 * @expose
 * @ojtsignore
 * @name series[].sourceHover
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The optional URI for the selected state. If not specified, the source image will be used. 
 * @expose
 * @ojtsignore
 * @name series[].sourceSelected
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The optional URI for the hover selected state. If not specified, the source image will be used. 
 * @expose
 * @ojtsignore
 * @name series[].sourceHoverSelected
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * A number from 0 to 1 indicating the amount to explode the pie slice. Only applies to pie charts.
 * @expose
 * @ojtsignore
 * @name series[].pieSliceExplode
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default 0
 * @ojmin 0
 * @ojmax 1
 */
/**
 * Defines whether the series is associated with the y2 axis. Only applies to Cartesian bar, line, area, and combo charts.
 * @expose
 * @ojtsignore
 * @name series[].assignedToY2
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * In stacked charts, groups series together for stacking. All series without a stackCategory will be assigned to the same stack.
 * @expose
 * @ojtsignore
 * @name series[].stackCategory
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the series should be shown in the legend. When set to 'auto', the series will not be displayed in the legend if it has null data or if it is a stock, funnel, or pyramid series.
 * @expose
 * @ojtsignore
 * @name series[].displayInLegend
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 *  Whether drilling is enabled on the series item. Drillable objects will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click. To enable drilling for all series items at once, use the drilling attribute in the top level. 
 * @expose
 * @ojtsignore
 * @name series[].drilling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default "inherit"
 */
/**
 * An optional array of category strings corresponding to this series. This allows highlighting and filtering of a series through interactions with legend sections. If not defined, the series id is used.
 * @expose
 * @ojtsignore
 * @name series[].categories
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The description of this series. This is used for accessibility and for customizing the tooltip text on the corressponding legend item for the series.
 * @expose
 * @ojtsignore
 * @name series[].shortDesc
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * An object containing the style properties of the box plot series.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  The color of the Q2 segment of the box.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.q2Color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The CSS style class to apply to the Q2 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q2Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.q2SvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The CSS inline style to apply to the Q2 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q2Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.q2SvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  The color of the Q3 segment of the box.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.q3Color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The CSS style class to apply to the Q3 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q3Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.q3SvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The CSS inline style to apply to the Q3 segment of the box. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the q3Color attribute. 
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.q3SvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class to apply to the whisker stems.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.whiskerSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS inline style to apply to the whisker stems.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.whiskerSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class to apply to the whisker ends.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.whiskerEndSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS inline style to apply to the whisker ends.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.whiskerEndSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  Specifies the length of the whisker ends in pixels (e.g. '9px') or as a percentage of the box width (e.g. '50%'). 
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.whiskerEndLength
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the median line.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.medianSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS inline style to apply to the median line.
 * @expose
 * @ojtsignore
 * @name series[].boxPlot.medianSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * An array of strings identifying the group labels, or an array of objects with the following properties. Also accepts a Promise for deferred data rendering.
 * @expose
 * @name groups
 * @memberof oj.ojChart
 * @instance
 * @type {Array.<Object>|Array.<string>|Promise}
 * @default null
 */
/**
 * The name of the group.
 * @expose
 * @name groups[].name
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The id of the group. Defaults to the name if not specified.
 * @expose
 * @name groups[].id
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object defining the style of the group label text. Supports color, 
 * fontFamily, fontSize, fontStyle, fontWeight, textDecoration, cursor,
 * backgroundColor, borderColor, borderRadius, and borderWidth properties. 
 * Only applies to a categorical axis.
 * @name groups[].labelStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The description of the group. This is used for customizing the tooltip text and only applies to a categorical axis.
 * @expose
 * @name groups[].shortDesc
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 *  Whether drilling is enabled on the group label. Drillable objects will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click. To enable drilling for all group labels at once, use the drilling attribute in the top level. 
 * @expose
 * @name groups[].drilling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default "inherit"
 */
/**
 * An array of nested group objects.
 * @expose
 * @name groups[].groups
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>|Array.<Object>}
 * @default null
 */
/**
 * An array used to define the ids of the initially selected objects.
 * When the selection is changed, the <code class="prettyprint">event.detail</code> of the <code class="prettyprint">selectionChanged</code> event will contain the following additional properties:<br><br>
 * <table class="props">
 *   <thead>
 *     <tr>
 *       <th>Name</th>
 *       <th>Type</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td class="name"><code>selectionData</code></td>
 *       <td class="type">Object</td>
 *       <td class="description">an array containing objects describing the selected data items
 *         <h6>Properties</h6>
 *         <table class="props">
 *           <thead>
 *             <tr>
 *               <th>Name</th>
 *               <th>Type</th>
 *               <th>Description</th>
 *             </tr>
 *           </thead>
 *           <tbody>
 *             <tr>
 *               <td class="name"><code>data</code></td>
 *               <td class="type">object</td>
 *               <td class="description">the data of the item, if one was specified</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>groupData</code></td>
 *               <td class="type">Array</td>
 *               <td class="description">the group data of the item</td>
 *             </tr>
 *             <tr>
 *               <td class="name"><code>seriesData</code></td>
 *               <td class="type">object</td>
 *               <td class="description">the series data of the item</td>
 *             </tr>
 *           </tbody>
 *         </table>
 *       </td>
 *     </tr>
 *     <tr>
 *       <td class="name"><code>endGroup</code></td>
 *       <td class="type">string</td>
 *       <td class="description">the end group of a marquee selection on a chart with categorical axis</td>
 *     </tr>
 *     <tr>
 *       <td class="name"><code>startGroup</code></td>
 *       <td class="type">string</td>
 *       <td class="description">the start group of a marquee selection on a chart with categorical axis</td>
 *     </tr>
 *     <tr>
 *       <td class="name"><code>xMax</code></td>
 *       <td class="type">number</td>
 *       <td class="description">the maximum x value of a marquee selection</td>
 *     </tr>
 *     <tr>
 *       <td class="name"><code>xMin</code></td>
 *       <td class="type">number</td>
 *       <td class="description">the minimum x value of a marquee selection</td>
 *     </tr>
 *     <tr>
 *       <td class="name"><code>yMax</code></td>
 *       <td class="type">number</td>
 *       <td class="description">the maximum y value of a marquee selection</td>
 *     </tr>
 *     <tr>
 *       <td class="name"><code>yMin</code></td>
 *       <td class="type">number</td>
 *       <td class="description">the minimum y value of a marquee selection</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @expose
 * @name selection
 * @memberof oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default []
 * @ojwriteback
 */
/**
 * Defines whether the plot area is split into two sections, so that sets of data assigned to the different Y-axes appear in different parts of the plot area. Stock charts do not support "off".
 * @expose
 * @name splitDualY
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * In a split dual-Y chart, specifies the fraction of the space that is given to the Y-axis subchart. Valid values are numbers from 0 to 1.
 * @expose
 * @name splitterPosition
 * @memberof oj.ojChart
 * @instance
 * @type {number}
 * @default 0.5
 * @ojmin 0
 * @ojmax 1
 */
/**
 * The type of time axis to display in the chart. Time axis is only supported for Cartesian bar, line, area, stock, box plot, and combo charts. If the type is "enabled" or "skipGaps", the time values must be provided through the "groups" attribute and stacking is supported. If the type is "skipGaps", the groups will be rendered at a regular interval regardless of any time gaps that may exist in the data. If the type is "mixedFrequency", the time values must be provided through the "x" attribute of the the data items and stacking is not supported.
 * @expose
 * @name timeAxisType
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "enabled"
 * @ojvalue {string} "mixedFrequency"
 * @ojvalue {string} "skipGaps"
 * @ojvalue {string} "disabled"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The type of selection behavior that is enabled on the chart.
 * @expose
 * @name selectionMode
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "single"
 * @ojvalue {string} "multiple"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The action that is performed when a drag occurs on the chart. Pan and marquee zoom are only available if zoom and scroll is turned on. Marquee select is only available if multiple selection is turned on. If the value is set to "user" and multiple actions are available, buttons will be displayed on the plot area to let users switch between modes.
 * @expose
 * @name dragMode
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "pan"
 * @ojvalue {string} "zoom"
 * @ojvalue {string} "select"
 * @ojvalue {string} "off"
 * @ojvalue {string} "user"
 * @default "user"
 */
/**
 * The chart type.
 * @expose
 * @name type
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "line"
 * @ojvalue {string} "area"
 * @ojvalue {string} "lineWithArea"
 * @ojvalue {string} "stock"
 * @ojvalue {string} "boxPlot"
 * @ojvalue {string} "combo"
 * @ojvalue {string} "pie"
 * @ojvalue {string} "scatter"
 * @ojvalue {string} "bubble"
 * @ojvalue {string} "funnel"
 * @ojvalue {string} "pyramid"
 * @ojvalue {string} "bar"
 * @default "bar"
 */
/**
 * Defines whether the data items are stacked. Only applies to bar, line, area, and combo charts. Does not apply to range series.
 * @expose
 * @name stack
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * Defines whether or not the total values of stacked data items should be displayed. Only applies to bar charts. It can be formatted by the valueFormat of the type 'label'.
 * @expose
 * @name stackLabel
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * The chart orientation. Only applies to bar, line, area, combo, box plot, and funnel charts.
 * @expose
 * @name orientation
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "horizontal"
 * @ojvalue {string} "vertical"
 * @default "vertical"
 */
/**
 * Defines whether the grid shape of the polar chart is circle or polygon. Only applies to polar line and area charts.
 * @expose
 * @name polarGridShape
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "polygon"
 * @ojvalue {string} "circle"
 * @default "circle"
 */
/**
 * The coordinate system of the chart. Only applies to bar, line, area, combo, scatter, and bubble charts.
 * @expose
 * @name coordinateSystem
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "polar"
 * @ojvalue {string} "cartesian"
 * @default "cartesian"
 */
/**
 * Defines the hide and show behavior that is performed when clicking on a legend item. When data items are hidden, the y axes can be optionally rescaled to fit to the remaining data.
 * @expose
 * @name hideAndShowBehavior
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "withRescale"
 * @ojvalue {string} "withoutRescale"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * An array of category strings used for filtering. Series or data items with any category matching an item in this array will be filtered.
 * @expose
 * @name hiddenCategories
 * @memberof oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default []
 * @ojwriteback
 */
/**
 * Defines the behavior applied when hovering over data items.
 * @expose
 * @name hoverBehavior
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dim"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * An array of category strings used for highlighting. Series or data items matching all categories in this array will be highlighted.
 * @expose
 * @name highlightedCategories
 * @memberof oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default []
 * @ojwriteback
 */
/**
 * The matching condition for the highlightedCategories property. By default, highlightMatch is 'all' and only items whose categories match all of the values specified in the highlightedCategories array will be highlighted. If highlightMatch is 'any', then items that match at least one of the highlightedCategories values will be highlighted.
 * @expose
 * @name highlightMatch
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "any"
 * @ojvalue {string} "all"
 * @default "all"
 */
/**
 * Defines the animation that is applied on data changes. Animation is automatically disabled when there are a large number of data items.
 * @expose
 * @name animationOnDataChange
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "auto"
 * @ojvalue {string} "slideToLeft"
 * @ojvalue {string} "slideToRight"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * Defines the animation that is shown on initial display. Animation is automatically disabled when there are a large number of data items.
 * @expose
 * @name animationOnDisplay
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "auto"
 * @ojvalue {string} "alphaFade"
 * @ojvalue {string} "zoom"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * Defines whether the data cursor is enabled. If set to "auto", the data cursor is shown only for line or area charts on touch devices. The data cursor is not shown when the tooltip is null and it is not supported on polar charts. 
 * @expose
 * @name dataCursor
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Defines the behavior of the data cursor when moving between data items.
 * @expose
 * @name dataCursorBehavior
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "smooth"
 * @ojvalue {string} "snap"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Speficies the position of the data cursor. Used for synchronizing data cursors across multiple charts. Null if the data cursor is not displayed.
 * @expose
 * @name dataCursorPosition
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 * @ojwriteback
 */
/**
 * The x value of the data cursor.
 * @expose
 * @name dataCursorPosition.x
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The y value of the data cursor. If both y and y2 are defined, y will take precedence.
 * @expose
 * @name dataCursorPosition.y
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The y2 value of the data cursor. If both y and y2 are defined, y will take precedence.
 * @expose
 * @name dataCursorPosition.y2
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * Specifies the sorting of the data. It should only be used for pie charts, bar/line/area charts with one series, or stacked bar/area charts. Sorting will not apply when using a hierarchical group axis.
 * @expose
 * @name sorting
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "ascending"
 * @ojvalue {string} "descending"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * Specifies the fraction of the whole pie under which a slice would be aggregated into an "Other" slice. Valid values range from 0 (default) to 1. For example, a value of 0.1 would cause all slices which are less than 10% of the pie to be aggregated into the "Other" slice. Only applies to pie chart.
 * @expose
 * @name otherThreshold
 * @memberof oj.ojChart
 * @instance
 * @type {number}
 * @default 0
 * @ojmin 0
 * @ojmax 1
 */
/**
 * An object defining the center content of a pie chart. Either a label can be displayed at the center of the pie chart or custom HTML content. 
 * @expose
 * @name pieCenter
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) to format the label if it is numeric. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name pieCenter.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the label if it is numeric. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name pieCenter.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the text for the label. When a innerRadius is specified, the label will automatically be scaled to fit within the inner circle. If the innerRadius is 0, the default font size will be used.
 * @expose
 * @name pieCenter.label
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The CSS style object defining the style of the label.
 * @expose
 * @name pieCenter.labelStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * A function that returns custom center content. The function takes a dataContext argument, 
 * provided by the chart, with the following properties:
 * <ul>
 *   <li>outerBounds: Object containing (x, y, width, height) of the rectangle 
 *   circumscribing the center area. The x and y coordinates are relative to the top, 
 *   left corner of the element.</li>
 *   <li>innerBounds: Object containing (x, y, width, height) of the rectangle 
 *   inscribed in the center area. The x and y coordinates are relative to the top, 
 *   left corner of the element.</li>
 *   <li>label: The pieCenter label.</li>
 *   <li>componentElement: The chart element.</li> 
 * </ul>
 * The function should return an Object with the following property: 
 * <ul>
 *   <li>insert: HTMLElement - An HTML element, which will be overlaid on top of the pie chart. 
 *   This HTML element will block interactivity of the chart  by default, but the CSS pointer-events property 
 *   can be set to 'none' on this element if the chart's interactivity is desired.
 *   </li>
 * </ul>
 * @expose
 * @name pieCenter.renderer
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Object)}
 * @default null
 */
/**
 * An object defining properties for the axis, tick marks, tick labels, and axis titles.
 * @expose
 * @name xAxis
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether the axis is rendered.
 * @expose
 * @name xAxis.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * Defines the size of the axis in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name xAxis.size
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines the maximum size of the axis in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name xAxis.maxSize
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The axis title. Does not apply to polar charts.
 * @expose
 * @name xAxis.title
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The array of reference objects associated with the axis.
 * @expose
 * @name xAxis.referenceObjects
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>}
 * @default []
 */
/**
 * The id of the reference object.
 * @expose
 * @name xAxis.referenceObjects[].id
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional array of category strings corresponding to this reference object. This allows highlighting and filtering of a reference object through interactions with legend sections. If not defined, the reference object id is used.
 * @expose
 * @name xAxis.referenceObjects[].categories
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The text displayed in the legend for the reference object.
 * @expose
 * @name xAxis.referenceObjects[].text
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The type of reference object being shown.
 * @expose
 * @name xAxis.referenceObjects[].type
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "area"
 * @ojvalue {string} "line"
 * @default "line"
 */
/**
 * The location of the reference object relative to the data items.
 * @expose
 * @name xAxis.referenceObjects[].location
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "front"
 * @ojvalue {string} "back"
 * @default "back"
 */
/**
 * The color of the reference object.
 * @expose
 * @name xAxis.referenceObjects[].color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the reference line.
 * @expose
 * @name xAxis.referenceObjects[].lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The line style of the reference line.
 * @expose
 * @name xAxis.referenceObjects[].lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The line type of the reference line. Only applies if the line value is not constant. centeredStepped and centeredSegmented are not supported for polar, scatter, and bubble charts.
 * @expose
 * @name xAxis.referenceObjects[].lineType
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "curved"
 * @ojvalue {string} "stepped"
 * @ojvalue {string} "centeredStepped"
 * @ojvalue {string} "segmented"
 * @ojvalue {string} "centeredSegmented"
 * @ojvalue {string} "straight"
 * @default "straight"
 */
/**
 * The CSS style class to apply to the reference object. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the reference object color attribute.
 * @expose
 * @name xAxis.referenceObjects[].svgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the reference object. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the reference object color attribute.
 * @expose
 * @name xAxis.referenceObjects[].svgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The value of a reference line. For categorical axes, the value represents the group index. For example,   0 is the position of the first group, and 1.5 is the position half way between the second and the third group.
 * @expose
 * @name xAxis.referenceObjects[].value
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The low value of a reference area. For categorical axes, the value represents the group index. For example,   0 is the position of the first group, and 1.5 is the position half way between the second and the third group.
 * @expose
 * @name xAxis.referenceObjects[].low
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The high value of a reference area. For categorical axes, the value represents the group index. For example,    0 is the position of the first group, and 1.5 is the position half way between the second and the third group.
 * @expose
 * @name xAxis.referenceObjects[].high
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The description of this object. This is used for accessibility and also for customizing the tooltip text.
 * @expose
 * @name xAxis.referenceObjects[].shortDesc
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the reference object should be shown in the legend.
 * @expose
 * @name xAxis.referenceObjects[].displayInLegend
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * The minimum value of the axis. Defaults to null for automatic calculation based on the data. For categorical axes, the value represents the group index. For example, 0 is the position of the first group, and 1.5 is the position half way between the second and the third group.
 * @expose
 * @name xAxis.min
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The maximum value of the axis. Defaults to null for automatic calculation based on the data. For categorical axes, the value represents the group index. For example, 0 is the position of the first group, and 1.5 is the position half way between the second and the third group.
 * @expose
 * @name xAxis.max
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The minimum data value corresponding to an axis. If specified, the automatic axis extent calculation will use this value. Only applies to numerical axes.
 * @expose
 * @name xAxis.dataMin
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The maximum data value corresponding to an axis. If specified, the automatic axis extent calculation will use this value. Only applies to numerical axes.
 * @expose
 * @name xAxis.dataMax
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The increment between major tick marks. Defaults to null for automatic calculation based on the data. Only applies to time and numerical axes. For log axis, the step is a multiplier, so for example, if the step is 2, the major tick marks will be rendered at 1, 2, 4, 8, and so on.
 * @expose
 * @name xAxis.step
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * The minimum increment between major tick marks. This is typically used to prevent fractional axis values for discrete measures. Only applies to numerical axes.
 * @expose
 * @name xAxis.minStep
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * The increment between minor tick marks. Defaults to null for automatic calculation based on the data. Only applies to numerical axes. For log axis, the step is a multiplier, so for example, if the minorStep is 2, the minor tick marks will be rendered at 1, 2, 4, 8, and so on.
 * @expose
 * @name xAxis.minorStep
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * Defines the axis scale. Only applies to numerical axes.
 * @expose
 * @name xAxis.scale
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "log"
 * @ojvalue {string} "linear"
 * @default "linear"
 */
/**
 * The CSS style object defining the style of the axis title. The CSS white-space property can be defined with value "nowrap" to disable default text wrapping of the title.
 * @expose
 * @name xAxis.titleStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object defining the properties of the tick labels.
 * @expose
 * @name xAxis.tickLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether the tick labels are rendered.
 * @expose
 * @name xAxis.tickLabel.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * The CSS style object defining the style of the labels. The CSS white-space property can be defined with value "nowrap" to disable default text wrapping of categorical labels.
 * @expose
 * @name xAxis.tickLabel.style
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The scaling behavior of the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name xAxis.tickLabel.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Defines whether the chart will automatically rotate the labels by 90 degrees in order to fit more labels on the axis. The rotation will only be applied to categorical labels for a horizontal axis.
 * @expose
 * @name xAxis.tickLabel.rotation
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) to format the labels. When using a time axis, this attribute also takes an array of two converters, which apply respectively to the first and second label levels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name xAxis.tickLabel.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * An object defining properties for the axis line.
 * @expose
 * @name xAxis.axisLine
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the axis line.
 * @expose
 * @name xAxis.axisLine.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the axis line.
 * @expose
 * @name xAxis.axisLine.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the axis line is rendered.
 * @expose
 * @name xAxis.axisLine.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * An object defining properties for the major tick marks.
 * @expose
 * @name xAxis.majorTick
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the major tick marks.
 * @expose
 * @name xAxis.majorTick.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The line style of the major tick marks.
 * @expose
 * @name xAxis.majorTick.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the major tick marks.
 * @expose
 * @name xAxis.majorTick.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The color of the major tick mark at the baseline (x = 0). Valid values are auto, inherit, or a custom color. If set to inherit, it will follow the lineColor attribute.
 * @expose
 * @name xAxis.majorTick.baselineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "inherit"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The line style of the major tick mark at the baseline (x = 0). If not specified, it will follow the lineStyle attribute.
 * @expose
 * @name xAxis.majorTick.baselineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the major tick mark at the baseline (x = 0) If not specified, it will follow the lineWidth attribute.
 * @expose
 * @name xAxis.majorTick.baselineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the major tick marks are rendered.
 * @expose
 * @name xAxis.majorTick.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * An object defining properties for the minor tick marks.
 * @expose
 * @name xAxis.minorTick
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the minor tick marks.
 * @expose
 * @name xAxis.minorTick.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The line style of the minor tick marks.
 * @expose
 * @name xAxis.minorTick.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the minor tick marks.
 * @expose
 * @name xAxis.minorTick.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the minor tick marks are rendered.
 * @expose
 * @name xAxis.minorTick.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Defines whether the axis baseline starts at the minimum value of the data or at zero. Only applies to numerical data axes.
 * @expose
 * @name xAxis.baselineScaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "min"
 * @ojvalue {string} "zero"
 * @default "zero"
 */
/**
 * Specifies the minimum x coordinate of the current viewport for zoom and scroll. For group axis, the group index will be treated as the axis coordinate. If both viewportStartGroup and viewportMin are specified, then viewportMin takes precedence. If not specified, this value will be the axis min.
 * @expose
 * @name xAxis.viewportMin
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * Specifies the maximum x coordinate of the current viewport for zoom and scroll. For group axis, the group index will be treated as the axis coordinate. If both viewportEndGroup and viewportMax are specified, then viewportMax takes precedence. If not specified, this value will be the axis max.
 * @expose
 * @name xAxis.viewportMax
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * Specifies the start group of the current viewport. Only applies to charts with group or time axis. If not specified, the default start group is the first group in the data set.
 * @expose
 * @name xAxis.viewportStartGroup
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * Specifies the end group of the current viewport. Only applies to charts with group or time axis. If not specified, the default end group is the last group in the data set.
 * @expose
 * @name xAxis.viewportEndGroup
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * An object defining properties for the axis, tick marks, tick labels, and axis titles.
 * @expose
 * @name yAxis
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether the axis is rendered.
 * @expose
 * @name yAxis.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * Defines the size of the axis in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name yAxis.size
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines the maximum size of the axis in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name yAxis.maxSize
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The axis title. Does not apply to polar charts.
 * @expose
 * @name yAxis.title
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The position of the axis relative to its content. For vertical charts, only start and end apply. For horizontal charts, only top and bottom apply.
 * @expose
 * @name yAxis.position
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "start"
 * @ojvalue {string} "end"
 * @ojvalue {string} "top"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The array of reference objects associated with the axis.
 * @expose
 * @name yAxis.referenceObjects
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>}
 * @default []
 */
/**
 * The id of the reference object.
 * @expose
 * @name yAxis.referenceObjects[].id
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional array of category strings corresponding to this reference object. This allows highlighting and filtering of a reference object through interactions with legend sections. If not defined, the reference object id is used.
 * @expose
 * @name yAxis.referenceObjects[].categories
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The text displayed in the legend for the reference object.
 * @expose
 * @name yAxis.referenceObjects[].text
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The type of reference object being shown.
 * @expose
 * @name yAxis.referenceObjects[].type
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "area"
 * @ojvalue {string} "line"
 * @default "line"
 */
/**
 * The location of the reference object relative to the data items.
 * @expose
 * @name yAxis.referenceObjects[].location
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "front"
 * @ojvalue {string} "back"
 * @default "back"
 */
/**
 * The color of the reference object.
 * @expose
 * @name yAxis.referenceObjects[].color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the reference line.
 * @expose
 * @name yAxis.referenceObjects[].lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The line style of the reference line.
 * @expose
 * @name yAxis.referenceObjects[].lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The line type of the reference line. Only applies if the line value is not constant. centeredStepped and centeredSegmented are not supported for polar, scatter, and bubble charts.
 * @expose
 * @name yAxis.referenceObjects[].lineType
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "curved"
 * @ojvalue {string} "stepped"
 * @ojvalue {string} "centeredStepped"
 * @ojvalue {string} "segmented"
 * @ojvalue {string} "centeredSegmented"
 * @ojvalue {string} "straight"
 * @default "straight"
 */
/**
 * The CSS style class to apply to the reference object. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the reference object color attribute.
 * @expose
 * @name yAxis.referenceObjects[].svgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the reference object. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the reference object color attribute.
 * @expose
 * @name yAxis.referenceObjects[].svgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The value of a reference line. This property defines a constant value across the entire reference line and is ignored if the items array is specified.
 * @expose
 * @name yAxis.referenceObjects[].value
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The low value of a reference area. This property defines a constant value across the entire reference area and is ignored if the items array is specified.
 * @expose
 * @name yAxis.referenceObjects[].low
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The high value of a reference area. This property defines a constant value across the entire reference area and is ignored if the items array is specified.
 * @expose
 * @name yAxis.referenceObjects[].high
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The description of this object. This is used for accessibility and also for customizing the tooltip text.
 * @expose
 * @name yAxis.referenceObjects[].shortDesc
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the reference object should be shown in the legend.
 * @expose
 * @name yAxis.referenceObjects[].displayInLegend
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * An array of values or an array of objects with the following properties that defines the data for a varying reference object. Only supported for y1 and y2 axes for all chart types.
 * @expose
 * @name yAxis.referenceObjects[].items
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>|Array.<number>}
 * @default null
 */
/**
 * The low value of this point of a reference area.
 * @expose
 * @name yAxis.referenceObjects[].items[].low
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The high value of this point of a reference area.
 * @expose
 * @name yAxis.referenceObjects[].items[].high
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The value of this point of a line object. Null can be specified to skip a data point.
 * @expose
 * @name yAxis.referenceObjects[].items[].value
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The x value of this point. Mainly used for scatter and bubble chart and to specify the date for mixed-frequency time axis.
 * For categorical axis, if the x value is not specified, it will default to the item index.
 * For regular time axis, if the x value is not specified, it will default to the group name of the item.
 * @expose
 * @name yAxis.referenceObjects[].items[].x
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The minimum value of the axis. Defaults to null for automatic calculation based on the data.
 * @expose
 * @name yAxis.min
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The maximum value of the axis. Defaults to null for automatic calculation based on the data.
 * @expose
 * @name yAxis.max
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The minimum data value corresponding to an axis. If specified, the automatic axis extent calculation will use this value.
 * @expose
 * @name yAxis.dataMin
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The maximum data value corresponding to an axis. If specified, the automatic axis extent calculation will use this value.
 * @expose
 * @name yAxis.dataMax
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The increment between major tick marks. Defaults to null for automatic calculation based on the data. For log axis, the step is a multiplier, so for example, if the step is 2, the major tick marks will be rendered at 1, 2, 4, 8, and so on.
 * @expose
 * @name yAxis.step
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * The minimum increment between major tick marks. This is typically used to prevent fractional axis values for discrete measures.
 * @expose
 * @name yAxis.minStep
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * The increment between minor tick marks. Defaults to null for automatic calculation based on the data. For log axis, the step is a multiplier, so for example, if the minorStep is 2, the minor tick marks will be rendered at 1, 2, 4, 8, and so on.
 * @expose
 * @name yAxis.minorStep
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * Defines the axis scale. Only applies to numerical axes.
 * @expose
 * @name yAxis.scale
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "log"
 * @ojvalue {string} "linear"
 * @default "linear"
 */
/**
 * The CSS style object defining the style of the axis title.
 * @expose
 * @name yAxis.titleStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object defining the properties of the tick labels.
 * @expose
 * @name yAxis.tickLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether the tick labels are rendered.
 * @expose
 * @name yAxis.tickLabel.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * Defines the position of the tick labels relative to the plot area. Inside position is not supported for scatter and bubble charts.
 * @expose
 * @name yAxis.tickLabel.position
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "inside"
 * @ojvalue {string} "outside"
 * @default "outside"
 */
/**
 * The CSS style object defining the style of the labels.
 * @expose
 * @name yAxis.tickLabel.style
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The scaling behavior of the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name yAxis.tickLabel.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) to format the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name yAxis.tickLabel.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * An object defining properties for the axis line.
 * @expose
 * @name yAxis.axisLine
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the axis line.
 * @expose
 * @name yAxis.axisLine.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the axis line.
 * @expose
 * @name yAxis.axisLine.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the axis line is rendered.
 * @expose
 * @name yAxis.axisLine.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * An object defining properties for the major tick marks.
 * @expose
 * @name yAxis.majorTick
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the major tick marks.
 * @expose
 * @name yAxis.majorTick.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the major tick marks.
 * @expose
 * @name yAxis.majorTick.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The line style of the major tick marks.
 * @expose
 * @name yAxis.majorTick.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The color of the major tick mark at the baseline (y = 0). Valid values are auto, inherit, or a custom color. If set to inherit, it will follow the lineColor attribute.
 * @expose
 * @name yAxis.majorTick.baselineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "inherit"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The line style of the major tick mark at the baseline (y = 0). If not specified, it will follow the lineStyle attribute.
 * @expose
 * @name yAxis.majorTick.baselineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the major tick mark at the baseline (y = 0) If not specified, it will follow the lineWidth attribute.
 * @expose
 * @name yAxis.majorTick.baselineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the major tick marks are rendered.
 * @expose
 * @name yAxis.majorTick.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * An object defining properties for the minor tick marks.
 * @expose
 * @name yAxis.minorTick
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the minor tick marks.
 * @expose
 * @name yAxis.minorTick.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The line style of the minor tick marks.
 * @expose
 * @name yAxis.minorTick.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the minor tick marks.
 * @expose
 * @name yAxis.minorTick.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the minor tick marks are rendered.
 * @expose
 * @name yAxis.minorTick.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Defines whether the axis baseline starts at the minimum value of the data or at zero. Only applies to numerical data axes.
 * @expose
 * @name yAxis.baselineScaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "min"
 * @ojvalue {string} "zero"
 * @default "zero"
 */
/**
 * Specifies the minimum y coordinate of the current viewport for zoom and scroll. Only applies to bubble and scatter charts. If not specified, this value will be the axis min.
 * @expose
 * @name yAxis.viewportMin
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * Specifies the maximum y coordinate of the current viewport for zoom and scroll. Only applies to bubble and scatter charts. If not specified, this value will be the axis max.
 * @expose
 * @name yAxis.viewportMax
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * An object defining properties for the axis, tick marks, tick labels, and axis titles. Y2 axis is only supported for Cartesian bar, line, area, and combo charts.
 * @expose
 * @name y2Axis
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether the axis is rendered.
 * @expose
 * @name y2Axis.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * Defines the size of the axis in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name y2Axis.size
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines the maximum size of the axis in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name y2Axis.maxSize
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The axis title.
 * @expose
 * @name y2Axis.title
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The position of the axis relative to its content. For vertical charts, only start and end apply. For horizontal charts, only top and bottom apply.
 * @expose
 * @name y2Axis.position
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "start"
 * @ojvalue {string} "end"
 * @ojvalue {string} "top"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The array of reference objects associated with the axis.
 * @expose
 * @name y2Axis.referenceObjects
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>}
 * @default []
 */
/**
 * The id of the reference object.
 * @expose
 * @name y2Axis.referenceObjects[].id
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional array of category strings corresponding to this reference object. This allows highlighting and filtering of a reference object through interactions with legend sections. If not defined, the reference object id is used.
 * @expose
 * @name y2Axis.referenceObjects[].categories
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The text displayed in the legend for the reference object.
 * @expose
 * @name y2Axis.referenceObjects[].text
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The type of reference object being shown.
 * @expose
 * @name y2Axis.referenceObjects[].type
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "area"
 * @ojvalue {string} "line"
 * @default "line"
 */
/**
 * The location of the reference object relative to the data items.
 * @expose
 * @name y2Axis.referenceObjects[].location
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "front"
 * @ojvalue {string} "back"
 * @default "back"
 */
/**
 * The color of the reference object.
 * @expose
 * @name y2Axis.referenceObjects[].color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the reference line.
 * @expose
 * @name y2Axis.referenceObjects[].lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The line style of the reference line.
 * @expose
 * @name y2Axis.referenceObjects[].lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The line type of the reference line. Only applies if the line value is not constant. centeredStepped and centeredSegmented are not supported for polar, scatter, and bubble charts.
 * @expose
 * @name y2Axis.referenceObjects[].lineType
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "curved"
 * @ojvalue {string} "stepped"
 * @ojvalue {string} "centeredStepped"
 * @ojvalue {string} "segmented"
 * @ojvalue {string} "centeredSegmented"
 * @ojvalue {string} "straight"
 * @default "straight"
 */
/**
 * The CSS style class to apply to the reference object. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the reference object color attribute.
 * @expose
 * @name y2Axis.referenceObjects[].svgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the reference object. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the reference object color attribute.
 * @expose
 * @name y2Axis.referenceObjects[].svgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The value of a reference line. This property defines a constant value across the entire reference line and is ignored if the items array is specified.
 * @expose
 * @name y2Axis.referenceObjects[].value
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The low value of a reference area. This property defines a constant value across the entire reference area and is ignored if the items array is specified.
 * @expose
 * @name y2Axis.referenceObjects[].low
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The high value of a reference area. This property defines a constant value across the entire reference area and is ignored if the items array is specified.
 * @expose
 * @name y2Axis.referenceObjects[].high
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The description of this object. This is used for accessibility and also for customizing the tooltip text.
 * @expose
 * @name y2Axis.referenceObjects[].shortDesc
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the reference object should be shown in the legend.
 * @expose
 * @name y2Axis.referenceObjects[].displayInLegend
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * An array of values or an array of objects with the following properties that defines the data for a varying reference object. Only supported for y1 and y2 axes for all chart types.
 * @expose
 * @name y2Axis.referenceObjects[].items
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>|Array.<number>}
 * @default null
 */
/**
 * The low value of this point of a reference area.
 * @expose
 * @name y2Axis.referenceObjects[].items[].low
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The high value of this point of a reference area.
 * @expose
 * @name y2Axis.referenceObjects[].items[].high
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The value of this point of a line object. Null can be specified to skip a data point.
 * @expose
 * @name y2Axis.referenceObjects[].items[].value
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The x value of this point. Mainly used for scatter and bubble chart and to specify the date for mixed-frequency time axis.
 * For categorical axis, if the x value is not specified, it will default to the item index.
 * For regular time axis, if the x value is not specified, it will default to the group name of the item.
 * @expose
 * @name y2Axis.referenceObjects[].items[].x
 * @memberof! oj.ojChart
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The minimum value of the axis. Defaults to null for automatic calculation based on the data.
 * @expose
 * @name y2Axis.min
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The maximum value of the axis. Defaults to null for automatic calculation based on the data.
 * @expose
 * @name y2Axis.max
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The minimum data value corresponding to an axis. If specified, the automatic axis extent calculation will use this value.
 * @expose
 * @name y2Axis.dataMin
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The maximum data value corresponding to an axis. If specified, the automatic axis extent calculation will use this value.
 * @expose
 * @name y2Axis.dataMax
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The increment between major tick marks. Defaults to null for automatic calculation based on the data. For log axis, the step is a multiplier, so for example, if the step is 2, the major tick marks will be rendered at 1, 2, 4, 8, and so on.
 * @expose
 * @name y2Axis.step
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * The minimum increment between major tick marks. This is typically used to prevent fractional axis values for discrete measures.
 * @expose
 * @name y2Axis.minStep
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * The increment between minor tick marks. Defaults to null for automatic calculation based on the data. For log axis, the step is a multiplier, so for example, if the minorStep is 2, the minor tick marks will be rendered at 1, 2, 4, 8, and so on.
 * @expose
 * @name y2Axis.minorStep
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojexclusivemin 0
 */
/**
 * Defines the axis scale. Only applies to numerical axes.
 * @expose
 * @name y2Axis.scale
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "log"
 * @ojvalue {string} "linear"
 * @default "linear"
 */
/**
 * The CSS style object defining the style of the axis title.
 * @expose
 * @name y2Axis.titleStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object defining the properties of the tick labels.
 * @expose
 * @name y2Axis.tickLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether the tick labels are rendered.
 * @expose
 * @name y2Axis.tickLabel.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * Defines the position of the tick labels relative to the plot area.
 * @expose
 * @name y2Axis.tickLabel.position
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "inside"
 * @ojvalue {string} "outside"
 * @default "outside"
 */
/**
 * The CSS style object defining the style of the labels.
 * @expose
 * @name y2Axis.tickLabel.style
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The scaling behavior of the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name y2Axis.tickLabel.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) to format the labels. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name y2Axis.tickLabel.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * An object defining properties for the axis line.
 * @expose
 * @name y2Axis.axisLine
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the axis line.
 * @expose
 * @name y2Axis.axisLine.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the axis line.
 * @expose
 * @name y2Axis.axisLine.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the axis line is rendered.
 * @expose
 * @name y2Axis.axisLine.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * An object defining properties for the major tick marks.
 * @expose
 * @name y2Axis.majorTick
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the major tick marks.
 * @expose
 * @name y2Axis.majorTick.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The line style of the major tick marks.
 * @expose
 * @name y2Axis.majorTick.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the major tick marks.
 * @expose
 * @name y2Axis.majorTick.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The color of the major tick mark at the baseline (y = 0). Valid values are auto, inherit, or a custom color. If set to inherit, it will follow the lineColor attribute.
 * @expose
 * @name y2Axis.majorTick.baselineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "inherit"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The line style of the major tick mark at the baseline (y = 0). If not specified, it will follow the lineStyle attribute.
 * @expose
 * @name y2Axis.majorTick.baselineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the major tick mark at the baseline (y = 0) If not specified, it will follow the lineWidth attribute.
 * @expose
 * @name y2Axis.majorTick.baselineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the major tick marks are rendered.
 * @expose
 * @name y2Axis.majorTick.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * An object defining properties for the minor tick marks.
 * @expose
 * @name y2Axis.minorTick
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The color of the minor tick marks.
 * @expose
 * @name y2Axis.minorTick.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The line style of the minor tick marks.
 * @expose
 * @name y2Axis.minorTick.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The width of the minor tick marks.
 * @expose
 * @name y2Axis.minorTick.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * Defines whether the minor tick marks are rendered.
 * @expose
 * @name y2Axis.minorTick.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Defines whether the axis baseline starts at the minimum value of the data or at zero. Only applies to numerical data axes.
 * @expose
 * @name y2Axis.baselineScaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "min"
 * @ojvalue {string} "zero"
 * @default "zero"
 */
/**
 * Defines whether the tick marks of the y1 and y2 axes are aligned. Not supported for logarithmic axes.
 * @expose
 * @name y2Axis.alignTickMarks
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * An object defining the overview scrollbar. Only applies if zoomAndScroll is not off. Currently only supported for vertical bar, line, area, stock, and combo charts.
 * @expose
 * @name overview
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Specifies whether the overview scrollbar is rendered. If not, simple scrollbar will be used.
 * @expose
 * @name overview.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * Specifies the height of the overview scrollbar in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name overview.height
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An object containing the property override for the overview chart. The API is the same as the chart property API, and the property provided here will be merged on top of the default property of the overview chart. This can be used to customize the style or the type of the overview chart.
 * @expose
 * @name overview.content
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object defining the style of the plot area.
 * @expose
 * @name plotArea
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The border color to be set on the chart's plot area.
 * @expose
 * @name plotArea.borderColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border width to be set on the chart's plot area.
 * @expose
 * @name plotArea.borderWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The color of the plot area background.
 * @expose
 * @name plotArea.backgroundColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Specifies whether the plot area is rendered.
 * @expose
 * @name plotArea.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * An object defining the style, positioning, and behavior of the legend.
 * @expose
 * @name legend
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The legend title.
 * @expose
 * @name legend.title
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * An array of objects with the following properties defining the additional legend sections, other than the default series and reference object sections.
 * @expose
 * @name legend.sections
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>}
 * @default []
 */
/**
 * The title of the legend section.
 * @expose
 * @name legend.sections[].title
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The horizontal alignment of the section title. If the section is collapsible or nested, only start alignment is supported.
 * @expose
 * @name legend.sections[].titleHalign
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "center"
 * @ojvalue {string} "end"
 * @ojvalue {string} "start"
 * @default "start"
 */
/**
 * The CSS style object defining the style of the section title.
 * @expose
 * @name legend.sections[].titleStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * An array of nested legend sections.
 * @expose
 * @name legend.sections[].sections
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>}
 * @default null
 */
/**
 * An array of objects with the following properties defining the legend items. Also accepts a Promise for deferred data rendering.
 * @expose
 * @name legend.sections[].items
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<Object>|Promise}
 * @default null
 */
/**
 * The id of the legend item, which is provided as part of the context for events fired by this chart. If not specified, the id defaults to the text of the legend item.
 * @expose
 * @name legend.sections[].items[].id
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The legend item text.
 * @expose
 * @name legend.sections[].items[].text
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * An array of categories for the legend item. Legend items currently only support a single category. If no category is specified, this defaults to the id or text of the legend item.
 * @expose
 * @name legend.sections[].items[].categories
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The type of legend symbol to display.
 * @expose
 * @name legend.sections[].items[].symbolType
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "line"
 * @ojvalue {string} "lineWithMarker"
 * @ojvalue {string} "image"
 * @ojvalue {string} "marker"
 * @default "marker"
 */
/**
 * The URI of the image of the legend symbol.
 * @expose
 * @name legend.sections[].items[].source
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The color of the legend symbol (line or marker). When symbolType is "lineWithMarker", this attribute defines the line color and the markerColor attribute defines the marker color.
 * @expose
 * @name legend.sections[].items[].color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border color of the marker. Only applies if symbolType is "marker" or "lineWithMarker".
 * @expose
 * @name legend.sections[].items[].borderColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The pattern used to fill the marker. Only applies if symbolType is "marker" or "lineWithMarker".
 * @expose
 * @name legend.sections[].items[].pattern
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The line style. Only applies when the symbolType is "line" or "lineWithMarker".
 * @expose
 * @name legend.sections[].items[].lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The line width in pixels. Only applies when the symbolType is "line" or "lineWithMarker".
 * @expose
 * @name legend.sections[].items[].lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The shape of the marker. In addition to the built-in shapes, it may also take SVG path commands to specify a custom shape. The legend will style the custom shapes the same way as built-in shapes, supporting properties like color and borderColor and applying hover and filter effects. Only applies if symbolType is "marker" or "lineWithMarker". 
 * @expose
 * @name legend.sections[].items[].markerShape
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "circle"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "human"
 * @ojvalue {string} "rectangle"
 * @ojvalue {string} "star"
 * @ojvalue {string} "square"
 * @default "square"
 */
/**
 * The color of the marker, if different than the line color. Only applies if the symbolType is "lineWithMarker".
 * @expose
 * @name legend.sections[].items[].markerColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the legend item corresponds to visible data items. A hollow symbol is shown if the value is "hidden".
 * @expose
 * @name legend.sections[].items[].categoryVisibility
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "hidden"
 * @ojvalue {string} "visible"
 * @default "visible"
 */
/**
 * The description of this legend item. This is used for accessibility and for customizing the tooltip text.
 * @expose
 * @name legend.sections[].items[].shortDesc
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An object with the following properties for the series section in the legend.
 * @expose
 * @name legend.seriesSection
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The title of the series section.
 * @expose
 * @name legend.seriesSection.title
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The horizontal alignment of the section title. If the section is collapsible or nested, only start alignment is supported.
 * @expose
 * @name legend.seriesSection.titleHalign
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "center"
 * @ojvalue {string} "end"
 * @ojvalue {string} "start"
 * @default "start"
 */
/**
 * The CSS style object defining the style of the section title.
 * @expose
 * @name legend.seriesSection.titleStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object with the following properties for the reference object section in the legend.
 * @expose
 * @name legend.referenceObjectSection
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The title of the reference object section.
 * @expose
 * @name legend.referenceObjectSection.title
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * The horizontal alignment of the section title. If the section is collapsible or nested, only start alignment is supported.
 * @expose
 * @name legend.referenceObjectSection.titleHalign
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "center"
 * @ojvalue {string} "end"
 * @ojvalue {string} "start"
 * @default "start"
 */
/**
 * The CSS style object defining the style of the section title.
 * @expose
 * @name legend.referenceObjectSection.titleStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The position of the legend within the chart. By default, the legend will be placed on the side or bottom, based on the size of the chart and the legend contents.
 * @expose
 * @name legend.position
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "start"
 * @ojvalue {string} "end"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "top"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Defines whether the legend is displayed. If set to auto, the legend will be hidden for charts with a large number of series. To ensure that the legend is always displayed, set this attribute to 'on'. To turn on legend for stock, funnel and pyramid charts, set the displayInLegend property for the series items to 'on'.
 * @expose
 * @name legend.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Defines the size of the legend in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name legend.size
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines the maximum size of the legend in pixels (e.g. '50px') or percent (e.g. '15%').
 * @expose
 * @name legend.maxSize
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The color of the legend background.
 * @expose
 * @name legend.backgroundColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border color of the legend.
 * @expose
 * @name legend.borderColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object defining the style of the legend text.
 * @expose
 * @name legend.textStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The width of the legend symbol (line or marker) in pixels.
 * @expose
 * @name legend.symbolWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The height of the legend symbol (line or marker) in pixels.
 * @expose
 * @name legend.symbolHeight
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The horizontal alignment of the title.
 * @expose
 * @name legend.titleHalign
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "center"
 * @ojvalue {string} "end"
 * @ojvalue {string} "start"
 * @default "start"
 */
/**
 * The CSS style object defining the style of the title.
 * @expose
 * @name legend.titleStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether scrolling is enabled for the legend.
 * @expose
 * @name legend.scrolling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "asNeeded"
 * @default "asNeeded"
 */
/**
 * An object defining the default styles for series colors, marker shapes, and other style attributes. Properties specified on this object may be overridden by specifications on the data object.
 * @expose
 * @name styleDefaults
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines the fill effect for the data items.
 * @expose
 * @name styleDefaults.seriesEffect
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "color"
 * @ojvalue {string} "pattern"
 * @ojvalue {string} "gradient"
 * @default "gradient"
 */
/**
 * The array defining the default color ramp for the series.
 * @expose
 * @name styleDefaults.colors
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The array defining the default pattern ramp for the series. This is used only when seriesEffect is 'pattern'.
 * @expose
 * @name styleDefaults.patterns
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * Specifies the color of the "Other" slice. Only applies to pie chart.
 * @expose
 * @name styleDefaults.otherColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The array defining the default shape ramp for the series. Valid values are defined in the markerShape attribute.
 * @expose
 * @name styleDefaults.shapes
 * @memberof! oj.ojChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The default border color for the data items. For funnel and pyramid charts, it is used for the slice border.
 * @expose
 * @name styleDefaults.borderColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The default border width for the data items. For funnel and pyramid charts, it is used for the slice border.
 * @expose
 * @name styleDefaults.borderWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The default background color of funnel slices that show actual/target values.
 * @expose
 * @name styleDefaults.funnelBackgroundColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the chart is displayed with a 3D effect. Only applies to pie, funnel and pyramid charts.
 * @expose
 * @name styleDefaults.threeDEffect
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * The selection effect that is applied to selected items. The values explode and highlightAndExplode only apply to pie charts.
 * @expose
 * @name styleDefaults.selectionEffect
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "explode"
 * @ojvalue {string} "highlightAndExplode"
 * @ojvalue {string} "highlight"
 * @default "highlight"
 */
/**
 * The duration of the animations, in milliseconds.
 * @expose
 * @name styleDefaults.animationDuration
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits milliseconds
 * @ojmin 0
 */
/**
 * Defines whether data change indicators are displayed during animation.
 * @expose
 * @name styleDefaults.animationIndicators
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "all"
 * @default "all"
 */
/**
 * The color of the indicator shown for an increasing data change animation.
 * @expose
 * @name styleDefaults.animationUpColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The color of the indicator shown for a decreasing data change animation.
 * @expose
 * @name styleDefaults.animationDownColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The fill color of the marquee. Applies to marquee selection and marquee zoom.
 * @expose
 * @name styleDefaults.marqueeColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border color of the marquee. Applies to marquee selection and marquee zoom.
 * @expose
 * @name styleDefaults.marqueeBorderColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  Specifies the radius of the inner circle that can be used to create a donut chart. Valid values range from 0 (default) to 1. Not supported if 3D effect is on.
 * @expose
 * @name styleDefaults.pieInnerRadius
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default 0
 * @ojmin 0
 * @ojmax 1
 */
/**
 * The width of the data line. Only applies to line, lineWithArea, scatter, and bubble series.
 * @expose
 * @name styleDefaults.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The line type of the data line or area. Only applies to line, area, scatter, and bubble series. centeredStepped and centeredSegmented are not supported for polar, scatter, and bubble charts. 
 * @expose
 * @name styleDefaults.lineType
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "straight"
 * @ojvalue {string} "curved"
 * @ojvalue {string} "stepped"
 * @ojvalue {string} "centeredStepped"
 * @ojvalue {string} "segmented"
 * @ojvalue {string} "centeredSegmented"
 * @ojvalue {string} "none"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The line style of the data line. Only applies to line, lineWithArea, scatter, and bubble series.
 * @expose
 * @name styleDefaults.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The color of the data markers, if different from the series color.
 * @expose
 * @name styleDefaults.markerColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the data markers should be displayed. Only applies to line, area, scatter, and bubble series. If auto, the markers will be displayed whenever the data points are not connected by a line.
 * @expose
 * @name styleDefaults.markerDisplayed
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The shape of the data markers. In addition to the built-in shapes, it may also take SVG path commands to specify a custom shape. The chart will style the custom shapes the same way as built-in shapes, supporting properties like color and borderColor and applying hover and selection effects. Only 'auto' is supported for range series. 
 * @expose
 * @name styleDefaults.markerShape
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "square"
 * @ojvalue {string} "circle"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "human"
 * @ojvalue {string} "star"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The size of the data markers in pixels.
 * @expose
 * @name styleDefaults.markerSize
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * The color of the line extending from the pie slice to the slice label.
 * @expose
 * @name styleDefaults.pieFeelerColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Specifies the presence and size of the gaps between data items, such as bars, markers, and areas. Valid values are a percentage string from 0% to 100%, where 100% produces the maximum supported gaps.
 * @expose
 * @name styleDefaults.dataItemGaps
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * In stock charts, the color of the candlestick when the 'close' value is greater than the 'open' value.
 * @expose
 * @name styleDefaults.stockRisingColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * In stock charts, the color of the candlestick when the 'open' value is greater than the 'close' value.
 * @expose
 * @name styleDefaults.stockFallingColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * In stock charts, the color of the range bars for candlestick.
 * @expose
 * @name styleDefaults.stockRangeColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * In stock charts, the color of the volume bars. If specified, overrides the default rising and falling colors used by the volume bars.
 * @expose
 * @name styleDefaults.stockVolumeColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The position of the data label. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively. The 'outsideSlice' value only applies to pie charts. The 'aboveMarker', 'belowMarker', 'beforeMarker', and 'afterMarker' values only apply to line, area, scatter, and bubble series. The 'insideBarEdge' and 'outsideBarEdge' values only apply to non-polar bar series. Stacked bars do not support 'outsideBarEdge'. The chart does not currently adjust layout to fit labels within the plot area or deal with any overlaps between labels. 
 * @expose
 * @name styleDefaults.dataLabelPosition
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @ojvalue {string} "center"
 * @ojvalue {string} "outsideSlice"
 * @ojvalue {string} "aboveMarker"
 * @ojvalue {string} "belowMarker"
 * @ojvalue {string} "beforeMarker"
 * @ojvalue {string} "afterMarker"
 * @ojvalue {string} "insideBarEdge"
 * @ojvalue {string} "outsideBarEdge"
 * @ojvalue {string} "none"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The CSS style object defining the style of the data label text. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively.
 * @expose
 * @name styleDefaults.dataLabelStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {object|Array.<Object>}
 * @default null
 */
/**
 * The CSS style object defining the style of the stack label. Only applies to stacked bar charts.
 * @expose
 * @name styleDefaults.stackLabelStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Specifies the width of the bar group gap as a ratio of the group width. The valid value is a number from 0 to 1.
 * @expose
 * @name styleDefaults.barGapRatio
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojmin 0
 * @ojmax 1
 */
/**
 * Specifies the maximum width of each bar in pixels.
 * @expose
 * @name styleDefaults.maxBarWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * Specifies initial hover delay in ms for highlighting items in chart.
 * @expose
 * @name styleDefaults.hoverBehaviorDelay
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits milliseconds
 * @ojmin 0
 */
/**
 * The CSS style object defining the style of the labels in the tooltip.
 * @expose
 * @name styleDefaults.tooltipLabelStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The CSS style object defining the style of the values in the tooltip.
 * @expose
 * @name styleDefaults.tooltipValueStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object defining the data cursor style.
 * @expose
 * @name styleDefaults.dataCursor
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The width of the data cursor line in pixels.
 * @expose
 * @name styleDefaults.dataCursor.lineWidth
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 * @ojmin 0
 */
/**
 * The color of the data cursor line.
 * @expose
 * @name styleDefaults.dataCursor.lineColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The line style of the data cursor line.
 * @expose
 * @name styleDefaults.dataCursor.lineStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default "solid"
 */
/**
 * The color of the data cursor marker. Defaults to the data series color.
 * @expose
 * @name styleDefaults.dataCursor.markerColor
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The size of the data cursor marker in pixels.
 * @expose
 * @name styleDefaults.dataCursor.markerSize
 * @memberof! oj.ojChart
 * @instance
 * @type {number}
 * @default null
 * @ojunits pixels
 */
/**
 * Whether the data cursor marker is displayed. Marker should only be hidden if the data cursor is displaying information for the entire group.
 * @expose
 * @name styleDefaults.dataCursor.markerDisplayed
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * An object defining the style for hierarchical label separators.
 * @expose
 * @name styleDefaults.groupSeparators
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Defines whether the group separators are displayed.
 * @expose
 * @name styleDefaults.groupSeparators.rendered
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * The color of the separators lines.
 * @expose
 * @name styleDefaults.groupSeparators.color
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An object containing the style properties of the box plot items.
 * @expose
 * @name styleDefaults.boxPlot
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The CSS style class to apply to the whisker stems.
 * @expose
 * @name styleDefaults.boxPlot.whiskerSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default ""
 */
/**
 * The CSS inline style to apply to the whisker stems.
 * @expose
 * @name styleDefaults.boxPlot.whiskerSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The CSS style class to apply to the whisker ends.
 * @expose
 * @name styleDefaults.boxPlot.whiskerEndSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default ""
 */
/**
 * The CSS inline style to apply to the whisker ends.
 * @expose
 * @name styleDefaults.boxPlot.whiskerEndSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 *  Specifies the length of the whisker ends in pixels (e.g. '9px') or as a percentage of the box width (e.g. '50%'). 
 * @expose
 * @name styleDefaults.boxPlot.whiskerEndLength
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the median line.
 * @expose
 * @name styleDefaults.boxPlot.medianSvgClassName
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default ""
 */
/**
 * The CSS inline style to apply to the median line.
 * @expose
 * @name styleDefaults.boxPlot.medianSvgStyle
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object specifying value formatting and tooltip behavior, whose keys generally correspond to the attribute names on the data items.
 * @name valueFormats
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Specifies tooltip behavior for the series.
 * @expose
 * @name valueFormats.series
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.series.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.series.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies tooltip behavior for the groups.
 * @expose
 * @name valueFormats.group
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * A string representing the label that is displayed before the value in the tooltip. This value can also take an array of strings to be applied to hierarchical group names, from outermost to innermost.
 * @expose
 * @name valueFormats.group.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.group.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the x values.
 * @expose
 * @name valueFormats.x
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.x.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.x.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.x.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.x.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the y values.
 * @expose
 * @name valueFormats.y
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.y.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.y.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.y.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.y.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the y2 values.
 * @expose
 * @name valueFormats.y2
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.y2.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.y2.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.y2.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.y2.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the z values.
 * @expose
 * @name valueFormats.z
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.z.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.z.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.z.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.z.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the values.
 * @expose
 * @name valueFormats.value
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.value.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.value.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.value.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.value.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the target values of a funnel chart.
 * @expose
 * @name valueFormats.targetValue
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.targetValue.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.targetValue.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.targetValue.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.targetValue.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the low values.
 * @expose
 * @name valueFormats.low
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.low.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.low.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.low.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.low.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the high values.
 * @expose
 * @name valueFormats.high
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.high.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.high.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.high.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.high.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the open values of a stock chart.
 * @expose
 * @name valueFormats.open
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.open.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.open.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.open.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.open.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the close values of a stock chart.
 * @expose
 * @name valueFormats.close
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.close.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.close.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.close.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.close.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the volume values of a stock chart.
 * @expose
 * @name valueFormats.volume
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.volume.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.volume.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.volume.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.volume.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the q1 values of a box plot.
 * @expose
 * @name valueFormats.q1
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.q1.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.q1.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.q1.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.q1.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the q2 values of a box plot.
 * @expose
 * @name valueFormats.q2
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.q2.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.q2.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.q2.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.q2.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting and tooltip behavior for the q3 values of a box plot.
 * @expose
 * @name valueFormats.q3
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.q3.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.q3.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * A string representing the label that is displayed before the value in the tooltip.
 * @expose
 * @name valueFormats.q3.tooltipLabel
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @default null
 * @ojtranslatable
 */
/**
 * Whether the value is displayed in the tooltip.
 * @expose
 * @name valueFormats.q3.tooltipDisplay
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the value formatting for the data item labels.
 * @expose
 * @name valueFormats.label
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The converter (an object literal or instance that duck types <a href="oj.Converter.html">oj.Converter</a>) used to format the label. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.label.converter
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The scaling behavior of the value. When using a converter, scaling should be set to none, as the formatted result may not be compatible with the scaling suffixes.
 * @expose
 * @name valueFormats.label.scaling
 * @memberof! oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "none"
 * @ojvalue {string} "thousand"
 * @ojvalue {string} "million"
 * @ojvalue {string} "billion"
 * @ojvalue {string} "trillion"
 * @ojvalue {string} "quadrillion"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies the zoom and scroll behavior of the chart. "Live" behavior means that the chart will be updated continuously as it is being manipulated, while "delayed" means that the update will wait until the zoom/scroll action is done. While "live" zoom and scroll provides the best end user experience, no guarantess are made about the rendering performance or usability for large data sets or slow client environments. If performance is an issue, "delayed" zoom and scroll should be used instead.
 * @expose
 * @name zoomAndScroll
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "delayedScrollOnly"
 * @ojvalue {string} "liveScrollOnly"
 * @ojvalue {string} "delayed"
 * @ojvalue {string} "live"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * Specifies the zoom direction of bubble and scatter charts. "Auto" zooms in both x and y direction. Use "x" or "y" for single direction zooming. 
 * @expose
 * @name zoomDirection
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "x"
 * @ojvalue {string} "y"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Whether automatic initial zooming is enabled. The valid values are "first" to initially zoom to the first data points (after the viewportMin) that can fit in the plot area, "last" to initially zoom to the last data points (before the viewportMax), and "none" to disable initial zooming. Only applies to bar, line, area, and combo charts with zoomAndScroll turned on.
 * @expose
 * @name initialZooming
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "first"
 * @ojvalue {string} "last"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 *  Whether drilling is enabled. Drillable objects will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). Use "on" to enable drilling for all series objects (legend items), group objects (x-axis labels), and data items. Use "seriesOnly" or "groupsOnly" to enable drilling for series objects or group objects only. To enable or disable drilling on individual series, group, or data item, use the drilling attribute in each series, group, or data item. 
 * @expose
 * @name drilling
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "seriesOnly"
 * @ojvalue {string} "groupsOnly"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * An object containing an optional callback function for tooltip customization. 
 * @expose
 * @name tooltip
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 *  A function that returns a custom tooltip. The function takes a dataContext argument, 
 *  provided by the chart, with the following properties: 
 *  <ul>
 *    <li>parentElement: The tooltip element. This can be used to change the tooltip border or background color.</li>
 *    <li>id: The id of the hovered item.</li>
 *    <li>series: The id of the series the hovered item belongs to.</li>
 *    <li>group: The ids or an array of ids of the group(s) the hovered item belongs to. 
 *        For hierarchical groups, it will be an array of outermost to innermost group ids.</li>
 *    <li>value, targetValue, x, y, z, low, high, open, close, volume: The values of the hovered item.</li>
 *    <li>label: The data label of the hovered item.</li>
 *    <li>data: The data object of the hovered item. For nested items, it will be an array containing 
 *     the parent item data and nested item data.</li>
 *    <li>seriesData: The data for the series the hovered item belongs to.</li>
 *    <li>groupData: An array of data for the group the hovered item belongs to. For hierarchical groups,
 *     it will be an array of outermost to innermost group data related to the hovered item.</li> 
 *    <li>componentElement: The chart element.</li>
 *    <li>color: The color of the hovered item.</li>
 *  </ul> 
 *  For reference objects, this tooltip function is only called if the reference object has an id. 
 *  Note: In future releases, tooltips may be fired for objects other than data items and reference objects. 
 *  To know whether a hovered object is a data item, please check that the "series" and "group" properties are not null. 
 *  The function should return an Object that contains only one of the two properties:
 *  <ul>
 *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
 *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
 *  </ul>
 * @expose
 * @name tooltip.renderer
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Object)}
 * @default null
 */
/**
 * Data visualizations require a press and hold delay before triggering tooltips, marquee selection, and rollover effects on mobile devices to avoid interfering with page panning, but these hold delays can make applications seem slower and less responsive. For a better user experience, the application can remove the touch and hold delay when data visualizations are used within a non scrolling container or if there is sufficient space outside of the visualization for panning. If touchResponse is touchStart the element will instantly trigger the touch gesture and consume the page pan events if the element does not require an internal feature that requires a touch start gesture like panning, zooming, or when marquee selection is initiated. If touchResponse is auto, the element will behave like touchStart if it determines that it is not rendered within scrolling content and if panning is not available for those elements that support the feature. 
 * @expose
 * @name touchResponse
 * @memberof oj.ojChart
 * @instance
 * @type {string}
 * @ojvalue {string} "touchStart"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 *  A function that returns a custom data label. The function takes a dataContext argument, provided by the chart, with the following properties: <ul> <li>id: The id of the data item.</li> <li>series: The id of the series the data item belongs to.</li> <li>group: The id or an array of ids of the group(s) the data item belongs to. For hierarchical groups, it will be an array of outermost to innermost group ids.</li> <li>value, targetValue, x, y, z, low, high, open, close, volume: The values of the data item.</li> <li>label: The label for the data item if the dataLabel callback is ignored. The dataLabel callback can concatenate this with another string to easily enhance the default label.</li> <li>totalValue: The total of all values in the chart. This will only be included for pie charts.</li> <li>data: The data object of the data item. For nested items, it will be an array containing the parent item data and nested item data.</li> <li>seriesData: The data for the series the data item belongs to.</li> <li>groupData: An array of data for the group the data item belongs to. For hierarchical groups, it will be an array of outermost to innermost group data related to the data item.</li> <li>componentElement: The chart element.</li> </ul> The function may return a number or a string or in the case of range charts, an array of numbers or strings. If any label is a number, it will be formatted by the valueFormat of the type 'label' before being used as labels. 
 * @expose
 * @name dataLabel
 * @memberof oj.ojChart
 * @instance
 * @type {function(Object)}
 * @default null
 */
/**
 * Provides support for HTML5 Drag and Drop events. Please refer to <a href="https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/Drag_and_drop">third party documentation</a> on HTML5 Drag and Drop to learn how to use it. 
 * @expose
 * @name dnd
 * @memberof oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An object that describes drag functionality.
 * @expose
 * @name dnd.drag
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Allows dragging of chart data items, including bars, line/area/scatter markers, bubbles, and pie/funnel/pyramid slices. 
 * @expose
 * @name dnd.drag.items
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The MIME types to use for the dragged data in the dataTransfer object. This can be a string if there is only one type, or an array of strings if multiple types are needed. For example, if selected employee data items are being dragged, dataTypes could be "application/employees+json". Drop targets can examine the data types and decide whether to accept the data. For each type in the array, dataTransfer.setData will be called with the specified type and the data. The data is an array of the dataContexts of the selected data items. The dataContext is the JSON version of the dataContext that we use for "tooltip" and "dataLabels" properties, excluding componentElement and parentElement. This property is required unless the application calls setData itself in a dragStart callback function. 
 * @expose
 * @name dnd.drag.items.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "drag" event as argument. 
 * @expose
 * @name dnd.drag.items.drag
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragend" event as argument. 
 * @expose
 * @name dnd.drag.items.dragEnd
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragstart" event and context information as arguments. The context information is as follows: <ul> <li> item {Array.(object)}: An array of dataContexts of the dragged data items. The dataContext is the same as what we use for "tooltip" and "dataLabels" properties. </li> </ul> This function can set its own data and drag image as needed. When this function is called, event.dataTransfer is already populated with the default data and drag image. 
 * @expose
 * @name dnd.drag.items.dragStart
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * Allows dragging of chart series from the legend items.
 * @expose
 * @name dnd.drag.series
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The MIME types to use for the dragged data in the dataTransfer object. This can be a string if there is only one type, or an array of strings if multiple types are needed. For example, if selected employee data items are being dragged, dataTypes could be "application/employees+json". Drop targets can examine the data types and decide whether to accept the data. For each type in the array, dataTransfer.setData will be called with the specified type and the data. The data is an array of the dataContexts of the selected series. The dataContext is the JSON version of the dataContext that we use for "tooltip" and "dataLabels" properties, excluding componentElement and parentElement. This property is required unless the application calls setData itself in a dragStart callback function.
 * @expose
 * @name dnd.drag.series.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "drag" event as argument. 
 * @expose
 * @name dnd.drag.series.drag
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragend" event as argument. 
 * @expose
 * @name dnd.drag.series.dragEnd
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragstart" event and context information as arguments. The context information is as follows: <ul> <li> series {Array.(object)}: An array of dataContexts of the dragged series. The dataContext is the same as what we use for "tooltip" and "dataLabels" properties. </li> </ul> This function can set its own data and drag image as needed. When this function is called, event.dataTransfer is already populated with the default data and drag image. 
 * @expose
 * @name dnd.drag.series.dragStart
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * Allows dragging of chart groups from the categorical axis labels.
 * @expose
 * @name dnd.drag.groups
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * The MIME types to use for the dragged data in the dataTransfer object. This can be a string if there is only one type, or an array of strings if multiple types are needed. For example, if selected employee data items are being dragged, dataTypes could be "application/employees+json". Drop targets can examine the data types and decide whether to accept the data. For each type in the array, dataTransfer.setData will be called with the specified type and the data. The data is an array of the dataContexts of the selected groups. The dataContext is the JSON version of the dataContext that we use for "tooltip" and "dataLabels" properties, excluding componentElement and parentElement. This property is required unless the application calls setData itself in a dragStart callback function.
 * @expose
 * @name dnd.drag.groups.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "drag" event as argument. 
 * @expose
 * @name dnd.drag.groups.drag
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragend" event as argument. 
 * @expose
 * @name dnd.drag.groups.dragEnd
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragstart" event and context information as arguments. The context information is as follows: <ul> <li> groups {Array.(object)}: An array of dataContexts of the dragged groups. The dataContext is the same as what we use for "tooltip" and "dataLabels" properties. </li> </ul> This function can set its own data and drag image as needed. When this function is called, event.dataTransfer is already populated with the default data and drag image. 
 * @expose
 * @name dnd.drag.groups.dragStart
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An object that describes drop functionality.
 * @expose
 * @name dnd.drop
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * Allows dropping on the plot area.
 * @expose
 * @name dnd.drop.plotArea
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An array of MIME data types this element can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events.
 * @expose
 * @name dnd.drop.plotArea.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "dragenter" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> <li>y {number}: The Y axis value at the event position.</li> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable.
 * @expose
 * @name dnd.drop.plotArea.dragEnter
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragover" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> <li>y {number}: The Y axis value at the event position.</li> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
 * @expose
 * @name dnd.drop.plotArea.dragOver
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragleave" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> <li>y {number}: The Y axis value at the event position.</li> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> 
 * @expose
 * @name dnd.drop.plotArea.dragLeave
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "drop" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> <li>y {number}: The Y axis value at the event position.</li> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. 
 * @expose
 * @name dnd.drop.plotArea.drop
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * Allows dropping on the X axis.
 * @expose
 * @name dnd.drop.xAxis
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An array of MIME data types this element can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events.
 * @expose
 * @name dnd.drop.xAxis.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "dragenter" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable.
 * @expose
 * @name dnd.drop.xAxis.dragEnter
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragover" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
 * @expose
 * @name dnd.drop.xAxis.dragOver
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragleave" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> </ul> 
 * @expose
 * @name dnd.drop.xAxis.dragLeave
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "drop" event and context information as arguments. The context information is as follows: <ul> <li>x {number}: The X axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. 
 * @expose
 * @name dnd.drop.xAxis.drop
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * Allows dropping on the Y axis.
 * @expose
 * @name dnd.drop.yAxis
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An array of MIME data types this element can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events.
 * @expose
 * @name dnd.drop.yAxis.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "dragenter" event and context information as arguments. The context information is as follows: <ul> <li>y {number}: The Y axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable.
 * @expose
 * @name dnd.drop.yAxis.dragEnter
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragover" event and context information as arguments. The context information is as follows: <ul> <li>y {number}: The Y axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
 * @expose
 * @name dnd.drop.yAxis.dragOver
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragleave" event and context information as arguments. The context information is as follows: <ul> <li>y {number}: The Y axis value at the event position.</li> </ul> 
 * @expose
 * @name dnd.drop.yAxis.dragLeave
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "drop" event and context information as arguments. The context information is as follows: <ul> <li>y {number}: The Y axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. 
 * @expose
 * @name dnd.drop.yAxis.drop
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * Allows dropping on the Y2 axis.
 * @expose
 * @name dnd.drop.y2Axis
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An array of MIME data types this element can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events. 
 * @expose
 * @name dnd.drop.y2Axis.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "dragenter" event and context information as arguments. The context information is as follows: <ul> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
 * @expose
 * @name dnd.drop.y2Axis.dragEnter
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragover" event and context information as arguments. The context information is as follows: <ul> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
 * @expose
 * @name dnd.drop.y2Axis.dragOver
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragleave" event and context information as arguments. The context information is as follows: <ul> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> 
 * @expose
 * @name dnd.drop.y2Axis.dragLeave
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "drop" event and context information as arguments. The context information is as follows: <ul> <li>y2 {number}: The Y2 axis value at the event position.</li> </ul> This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. 
 * @expose
 * @name dnd.drop.y2Axis.drop
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * Allows dropping on the legend.
 * @expose
 * @name dnd.drop.legend
 * @memberof! oj.ojChart
 * @instance
 * @type {Object}
 * @default {}
 */
/**
 * An array of MIME data types this element can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events. 
 * @expose
 * @name dnd.drop.legend.dataTypes
 * @memberof! oj.ojChart
 * @instance
 * @type {string|Array.<string>}
 * @default null
 */
/**
 * An optional callback function that receives the "dragenter" event and empty context information as arguments. This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
 * @expose
 * @name dnd.drop.legend.dragEnter
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragover" event and empty context information as arguments. This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
 * @expose
 * @name dnd.drop.legend.dragOver
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "dragleave" event and empty context information as arguments. 
 * @expose
 * @name dnd.drop.legend.dragLeave
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
/**
 * An optional callback function that receives the "drop" event and emtpy context information as arguments. This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. 
 * @expose
 * @name dnd.drop.legend.drop
 * @memberof! oj.ojChart
 * @instance
 * @type {function(Event, object)}
 * @default null
 */
      
/**
 * Object type that defines a chart series.
 * @ignore
 * @typedef {Object} oj.ojChart.Series
 * @property {string} id The id of the series. Defaults to the name or the series index if not specified.
 * @property {string} name The name of the series, displayed in the legend and tooltips.
 * @property {"bar"|"line"|"area"|"lineWithArea"|"candlestick"|"boxPlot"|"auto"} type The type of data objects to display for this series. Only applies to bar, line, area, stock, box plot, and combo charts.
 * @property {string} color The color of the series.
 * @property {string} borderColor The border color of the series.
 * @property {number} borderWidth The border width of the series.
 * @property {string} areaColor The area color of the series. Only applies if series type is area or lineWithArea.
 * @property {string} areaSvgClassNamer The CSS style class to apply if series type is area or lineWithArea. The style class and inline style will override any other styling specified through the properties. 
 *            For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @property {object} areaSvgStyle The inline style to apply if series type is area or lineWithArea. The style class and inline style will override any other styling specified through the properties. 
 *            For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @property {string} svgClassName The CSS style class to apply to the series. For series of type lineWithArea, this style will only be applied to the line if areaClassName is also specified. 
 *            The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @property {object} svgStyle The inline style to apply to the series. For series of type lineWithArea, this style will only be applied to the line if areaStyle is also specified.
 *            The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @property {string} markerSvgClassName The CSS style class to apply to the data markers.The style class and inline style will override any other styling specified through the properties. 
 *            For tooltips and hover interactivity, it's recommended to also pass a representative color to the marker color attribute.
 * @property {object} markerSvgStyle The inline style to apply to the data markers. The style class and inline style will override any other styling specified through the properties. 
 *            For tooltips and hover interactivity, it's recommended to also pass a representative color to the marker color attribute.
 * @property {"smallChecker"|"smallCrosshatch"|"smallDiagonalLeft"|"smallDiagonalRight"|"smallDiamond"|"smallTriangle"|"largeChecker"|"largeCrosshatch"|"largeDiagonalLeft"|"largeDiagonalRight"|"largeDiamond"|"largeTriangle"} pattern
 *           The pattern used to fill the series. A solid fill is used by default, unless the seriesEffect is 'pattern'.
 * @property {"square"|"circle"|"diamond"|"plus"|"triangleDown"|"triangleUp"|"human"|"star"|"auto"} markerShape The shape of the data markers. In addition to the built-in shapes, it may also take SVG path commands to specify a custom shape. 
 *            The chart will style the custom shapes the same way as built-in shapes, supporting properties like color and borderColor and applying hover and selection effects. Only 'auto' is supported for range series. 
 * @property {string} markerColor The color of the data markers, if different from the series color.
 * @property {"on"|"off"|"auto"} markerDisplayed Defines whether the data marker is displayed. Only applies to line, area, scatter, and bubble series. If auto, the markers will be displayed whenever the data points are not connected by a line.
 * @property {number} markerSize The size of the data markers.
 * @property {number} lineWidth The width of the data line. Only applies to line, lineWithArea, scatter, and bubble series.
 * @property {"dotted"|"dashed"|"solid"} lineStyle
 * @property {"straight"|"curved"|"stepped"|"centeredStepped"|"segmented"|"centeredSegmented"|"none"|"auto"} lineType
 *           The line type of the data line or area. Only applies to line, area, scatter, and bubble series. centeredStepped and centeredSegmented are not supported for polar, scatter, and bubble charts. 
 * @property {string} source The URI of the custom image. If specified, it takes precedence over shape. 
 * @property {string} sourceHover The optional URI for the hover state. If not specified, the source image will be used. 
 * @property {string} sourceSelected The optional URI for the selected state. If not specified, the source image will be used. 
 * @property {string} sourceHoverSelected The optional URI for the hover selected state. If not specified, the source image will be used. 
 * @property {number} pieSliceExploded A number from 0 to 1 indicating the amount to explode the pie slice. Only applies to pie charts.
 * @property {"on"|"off"} assignedToY2 Defines whether the series is associated with the y2 axis. Only applies to Cartesian bar, line, area, and combo charts.
 * @property {string} stackCategory In stacked charts, groups series together for stacking. All series without a stackCategory will be assigned to the same stack.
 * @property {"on"|"off"|"auto"} displayInLegend Defines whether the series should be shown in the legend. When set to 'auto', the series will not be displayed in the legend if 
 *            it has null data or if it is a stock, funnel, or pyramid series.
 * @property {"on"|"off"|"inherit"} drilling Whether drilling is enabled on the series item. Drillable objects will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click. 
 *           To enable drilling for all series items at once, use the drilling attribute in the top level. 
 * @property {Array.<string>} categories An optional array of category strings corresponding to this series. 
 *           This allows highlighting and filtering of a series through interactions with legend sections. If not defined, the series id is used.
 * @property {string} shortDesc The description of this series. This is used for accessibility and for customizing the tooltip text on the corressponding legend item for the series.
 * @property {oj.ojChart.BoxPlotStyle} boxPlot An object containing the style properties of the box plot series.
 */    

/**
 * Object type that defines a chart group.
 * @ignore
 * @typedef {Object} oj.ojChart.Group 
 * @property {string} id The id of the group. Defaults to the name if not specified.
 * @property {string} name The name of the group
 * @property {object} labelStyle The CSS style object defining the style of the group label text. Supports color, 
 *                    fontFamily, fontSize, fontStyle, fontWeight, textDecoration, cursor,
 *                    backgroundColor, borderColor, borderRadius, and borderWidth properties. 
 *                    Only applies to a categorical axis.
 * @property {string} shortDesc The description of the group. This is used for customizing the tooltip text and only applies to a categorical axis.
 * @property {"on"|"off"|"auto"} drilling Whether drilling is enabled on the group label. Drillable objects will show a pointer cursor on hover and fire an 
 *                    <code class="prettyprint">ojDrill</code> event on click. To enable drilling for all group labels at once, use the drilling attribute in the top level. 
 * @property {Array.<oj.ojChart.Group>} groups An array of nested group obects.
 * @property {Array.<oj.ojChart.DataItem>} items An array of chart data items.
 */

/**
 * Object type that defines a chart data item.
 * @ignore
 * @typedef {Object} oj.ojChart.DataItem
 * @property {Array.<oj.ojChart.DataItem>} items An array of nested data items to be used for defining the markers for outliers or additional data items of a box plot. 
 * @property {string} id The id of the data item. This id will be provided as part of the context for events on the chart.
 * @property {number|string} x The x value for a scatter or bubble chart or the date on a time axis.
 * @property {number} y The y value. Also the primary value for charts without a y-Axis, such as pie charts.
 * @property {number} z The z value. Defines the bubble radius for a bubble chart, as well as the width of a bar or a box plot item.
 * @property {number} low The low value for range bar/area, stock candlestick, or box plot item. Define 'low' and 'high' instead of 'value' or 'y' to create a range bar/area chart.
 * @property {number} high The high value for range bar/area, stock candlestick, or box plot item. Define 'low' and 'high' instead of 'value' or 'y' to create a range bar/area chart.
 * @property {number} open The open value for stock candlestick.
 * @property {number} close The close value for stock candlestick. When bar, line, or area series type are used on a stock chart, this value is displayed.
 * @property {number} volume The value for stock volume bar. When this value is provided, the volume bar is displayed on the y2 axis.
 * @property {number} q1 The first quartile value for box plot.
 * @property {number} q2 The second quartile (median) value for box plot.
 * @property {number} q3 The third quartile value for box plot.
 * @property {string} shortDesc The description of this object. This is used for accessibility and also for customizing the tooltip text.
 * @property {string} color The color of the data item.
 * @property {string} borderColor The border color of the data item. For funnel and pyramid charts, it is used for the slice border.
 * @property {number} borderWidth The border width of the data item. For funnel and pyramid charts, it is used for the slice border.
 * @property {"smallChecker"|"smallCrosshatch"|"smallDiagonalLeft"|"smallDiagonalRight"|"smallDiamond"|"smallTriangle"|"largeChecker"|"largeCrosshatch"|"largeDiagonalLeft"|"largeDiagonalRight"|"largeDiamond"|"largeTriangle"} pattern
 *           The pattern used to fill the data item. A solid fill is used by default, unless the seriesEffect is 'pattern'.
 * @property {string} svgClassName The CSS style class to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @property {object} svgStyle The inline style to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @property {"on"|"off"|"auto"} markerDisplayed Defines whether the data marker is displayed. Only applies to line, area, scatter, and bubble series. If auto, the markers will be displayed whenever the data points are not connected by a line.
 * @property {"square"|"circle"|"diamond"|"plus"|"triangleDown"|"triangleUp"|"human"|"star"|"auto"} markerShape The shape of the data markers. In addition to the built-in shapes, it may also take SVG path commands to specify a custom shape. 
 *            The chart will style the custom shapes the same way as built-in shapes, supporting properties like color and borderColor and applying hover and selection effects. Only 'auto' is supported for range series. 
 * @property {number} markerSize The size of the data markers. Does not apply to bubble charts, which calculate marker size based on the z values.
 * @property {string} source The URI of the custom image. If specified, it takes precedence over shape. 
 * @property {string} sourceHover The optional URI for the hover state. If not specified, the source image will be used.
 * @property {string} sourceSelected The optional URI for the selected state. If not specified, the source image will be used.
 * @property {string} sourceHoverSelected The optional URI for the hover selected state. If not specified, the source image will be used.
 * @property {string|Array.<string>} label The label for the data item. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively. Not supported for box plot or candlestick.
 * @property {string|Array.<string>} labelPosition The position of the data label. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively. The 'outsideSlice' value only applies to pie charts. 
 *            The 'aboveMarker', 'belowMarker', 'beforeMarker', and 'afterMarker' values only apply to line, area, scatter, and bubble series. The 'insideBarEdge' and 'outsideBarEdge' values only apply to non-polar bar series. Stacked bars do not support 'outsideBarEdge'. 
 *            The chart does not currently adjust layout to fit labels within the plot area or deal with any overlaps between labels. 
 * @property {string|Array.<object>} labelStyle The CSS style object defining the style of the data label. For range series, if an array of two values are provided, the first and second value will apply to the low and high point respectively.
 * @property {Array.<string>} categories An optional array of category strings corresponding to this data item. This enables highlighting and filtering of individual data items through interactions with the legend or other visualization elements. If not defined, series categories are used.
 * @property {number} value The value for this data item. Corresponding to the y value for bar, line, area, and combo charts and the slice values for pie, funnel and pyramid charts. Null can be specified to skip a data point.
 * @property {number} targetValue The target value for a funnel chart. When this is set, the value attribute defines the filled area within the slice and this represents the value of the whole slice.
 * @property {"on"|"off"|"drilling"} drilling Whether drilling is enabled for the data item. Drillable objects will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). To enable drilling for all data items at once, use the drilling attribute in the top level.  
 * @property {oj.ojChart.BoxPlotStyle} boxPlot An object containing the style properties of the box plot item.
 */
      
/**
 * Object type that defines box plot style properties.
 * @ignore
 * @typedef {Object} oj.ojChart.BoxPlotStyle
 * @property {string} q2Color The color of the Q2 segment of the box.
 * @property {string} q2SvgClassName The CSS style class to apply to the Q2 segment of the box. The style class and inline style will override any other styling specified through the properties. 
 *            For tooltips and hover interactivity, it's recommended to also pass a representative color to the q2Color attribute. 
 * @property {object} q2SvgStyle The CSS inline style to apply to the Q2 segment of the box. The style class and inline style will override any other styling specified through the properties. 
 *            For tooltips and hover interactivity, it's recommended to also pass a representative color to the q2Color attribute. 
 * @property {string} q3Color The color of the Q3 segment of the box.
 * @property {string} q3SvgClassName The CSS style class to apply to the Q3 segment of the box. The style class and inline style will override any other styling specified through the properties.            For tooltips and hover interactivity, it's recommended to also pass a representative color to the q3Color attribute. 
 * @property {string} q3SvgStyle The CSS inline style to apply to the Q3 segment of the box. The style class and inline style will override any other styling specified through the properties. 
 *            For tooltips and hover interactivity, it's recommended to also pass a representative color to the q3Color attribute. 
 * @property {string} whiskerSvgClassName The CSS style class to apply to the whisker stems.
 * @property {string} whiskerSvgStyle The CSS inline style to apply to the whisker stems.
 * @property {string} whiskerEndSvgClassName The CSS style class to apply to the whisker ends.
 * @property {string} whiskerEndSvgStyle The CSS inline style to apply to the whisker ends.
 * @property {string} whiskerEndLength Specifies the length of the whisker ends in pixels (e.g. '9px') or as a percentage of the box width (e.g. '50%'). 
 * @property {string} medianSvgClassName The CSS style class to apply to the median line.
 * @property {string} medianSvgStyle The CSS inline style to apply to the median line.
 */  

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for chart data items indexed by series and group indices. The group index is not required for pie and
 * funnel charts.</p>
 *
 * @property {number} seriesIndex
 * @property {number} itemIndex
 *
 * @ojsubid oj-chart-item
 * @memberof oj.ojChart
 *
 * @example <caption>Get the data item from the first series and second group:</caption>
 * var nodes = myChart.getNodeBySubId({'subId': 'oj-chart-item', 'seriesIndex': 0, 'itemIndex': 1});
 */

/**
 * <p>Sub-ID for a legend item that represents the series with the specified index.</p>
 *
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 *
 * @property {number} index
 *
 * @ojsubid oj-chart-series
 * @memberof oj.ojChart
 *
 * @example <caption>Get the legend item that represents the first series:</caption>
 * var nodes = myChart.getNodeBySubId({'subId': 'oj-chart-series', 'index': 0});
 */

/**
 * <p>Sub-ID for a categorical axis label that represents the group with the specified index.</p>
 *
 * @property {Array} indexPath The array of indices corresponding to the position of the group in the properties array.
 *
 * @ojsubid oj-chart-group
 * @memberof oj.ojChart
 * @instance
 *
 * @example <caption>Get the categorical axis label that represents the first group:</caption>
 * var nodes = myChart.getNodeBySubId({'subId': 'oj-chart-group', 'indexPath': [0]});
 */

/**
 * <p>Sub-ID for the title of the specified axis.</p>
 *
 * @property {string} axis <code class="prettyprint">xAxis</code>, <code class="prettyprint">yAxis</code>, or <code class="prettyprint">y2Axis</code>
 *
 * @ojsubid oj-chart-axis-title
 * @memberof oj.ojChart
 *
 * @example <caption>Get the title for the x-axis:</caption>
 * var nodes = myChart.getNodeBySubId({'subId': 'oj-chart-axis-title', 'axis': 'xAxis'});
 */

/**
 * <p>Sub-ID for the reference object of the specified axis with the given index.</p>
 *
 * @property {string} axis <code class="prettyprint">xAxis</code>, <code class="prettyprint">yAxis</code>, or <code class="prettyprint">y2Axis</code>
 * @property {number} index The index of the reference object for the specified axis.
 *
 * @ojsubid oj-chart-reference-object
 * @memberof oj.ojChart
 *
 * @example <caption>Get the first reference object of the y-axis:</caption>
 * var nodes = myChart.getNodeBySubId({'subId': 'oj-chart-reference-object', 'axis': 'yAxis', 'index': 0});
 */

/**
 * <p>Sub-ID for the the chart tooltip.</p>
 *
 * @ojsubid oj-chart-tooltip
 * @memberof oj.ojChart
 *
 * @example <caption>Get the tooltip object of the chart, if displayed:</caption>
 * var nodes = myChart.getNodeBySubId({'subId': 'oj-chart-tooltip'});
 */

/**
 * <p>Sub-ID for a legend item indexed by its position in its parent section's
 * item array and its parent's sectionIndex.</p>
 *
 * @property {Array} sectionIndexPath The array of numerical indices for the section.
 * @property {number} itemIndex The index of the item within the specified section.
 *
 * @ojsubid oj-legend-item
 * @memberof oj.ojChart
 *
 * @example <caption>Get the first legend item from the first legend section:</caption>
 * var nodes = myChart.getNodeBySubId({'subId': 'oj-legend-item', sectionIndexPath: [0], itemIndex: 1});
 */

// Node Context Objects ********************************************************

/**
* <p>Context for chart data items indexed by series and group indices.</p>
*
* @property {number} seriesIndex
* @property {number} itemIndex
*
* @ojnodecontext oj-chart-item
* @memberof oj.ojChart
*/

/**
 * <p>Context for a legend item that represents the series with the specified index.</p>
 *
 * @property {number} index
 *
 * @ojnodecontext oj-chart-series
 * @memberof oj.ojChart
 */

/**
 * <p>Context for a categorical axis label that represents the group with the specified index.</p>
 *
 * @property {Array} indexPath The array of indices corresponding to the position of the group in the properties array.
 *
 * @ojnodecontext oj-chart-group
 * @memberof oj.ojChart
 */

/**
 * <p>Context for the title of the specified axis.</p>
 *
 * @property {string} axis <code class="prettyprint">xAxis</code>, <code class="prettyprint">yAxis</code>, or <code class="prettyprint">y2Axis</code>
 *
 * @ojnodecontext oj-chart-axis-title
 * @memberof oj.ojChart
 */
 
/**
 * <p>Context for the center label of a pie chart.</p>
 *
 * @ojnodecontext oj-chart-pie-center-label
 * @memberof oj.ojChart
 */

/**
 * <p>Context for the reference object of the specified axis with the given index.</p>
 *
 * @property {string} axis <code class="prettyprint">xAxis</code>, <code class="prettyprint">yAxis</code>, or <code class="prettyprint">y2Axis</code>
 * @property {number} index The index of the reference object for the specified axis.
 *
 * @ojnodecontext oj-chart-reference-object
 * @memberof oj.ojChart
 */

/**
 * <p>Context for a legend item indexed by its position in its parent section's
 * item array and its parent's sectionIndex.</p>
 *
 * @property {Array} sectionIndexPath The array of numerical indices for the section.
 * @property {number} itemIndex The index of the item within the specified section.
 *
 * @ojnodecontext oj-legend-item
 * @memberof oj.ojChart
 */

/**
 * @ojcomponent oj.ojSparkChart
 * @augments oj.dvtBaseComponent
 * @ojtsignore
 * @since 0.7
 * @ojshortdesc Displays information graphically, typically highlighting the trend of a data set in a compact form factor.
 * @ojstatus preview
 * @ojrole application
 *
 * @classdesc
 * <h3 id="sparkChartOverview-section">
 *   JET Spark Chart
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#sparkChartOverview-section"></a>
 * </h3>
 *
 * <p>Spark Chart component for JET with support for bar, line, area, and floating bar subtypes.  Spark Charts are
 * designed to visualize the trend of a data set in a compact form factor.</p>
 *
 * {@ojinclude "name":"warning"}
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-spark-chart
 *   type='line'
 *   items='[5, 8, 2, 7, 0, 9]'
 * >
 * &lt;/oj-spark-chart>
 * </code>
 * </pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"a11y"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojSparkChart', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * An array of <a href="/jsdocs/oj.ojSparkChart.html#Item">oj.ojSparkChart.Item</a> typed objects. Also accepts a Promise for deferred data rendering.
     * @ignore
     * @name data
     * @memberof oj.ojSparkChart
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @ojsignature {target: "Accessor", value: {GetterType: "oj.DataProvider.<oj.ojSparkChart.Item>|null", SetterType: "oj.DataProvider.<oj.ojSparkChart.Item>|null"}, jsdocOverride: true}
     * @default null
     */  
     data: null,
     
    /**
     * An array of objects with the following properties that defines the data for the spark chart. Also accepts a Promise for deferred data rendering.</ul>
     * @name items
     * @memberof oj.ojSparkChart
     * @instance
     * @type {Array.<Object>|Array.<number>|Promise|null}
     * @default null
     * 
     * @example <caption>Initialize the spark chart with the 
     * <code class="prettyprint">items</code> attribute specified:</caption>
     * &lt;oj-spark-chart type="bar" items='[{"low": 4, "high": 20, "color": "red"}, 
     *                                    {"low": 9, "high": 20, "color": "yellow"}, 
     *                                    {"low": 0, "high": 7, "color": "green"}]'>
     * &lt;/oj-spark-chart>
     * 
     * &lt;oj-spark-chart items='[[itemsPromise]]'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">items</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = mySparkChart.items[0];
     * 
     * // Get all (The items getter always returns a Promise so there is no "get one" syntax)
     * var values = mySparkChart.items;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * mySparkChart.items = [{"low": 4, "high": 20, "color": "red"}, 
     *                       {"low": 9, "high": 20, "color": "yellow"}, 
     *                       {"low": 0, "high": 7, "color": "green"}];
     */
    items: null,
    
    /**
     * An array of objects with the following properties defining the reference objects associated with the y axis of the spark chart.
     * @expose
     * @name referenceObjects
     * @memberof oj.ojSparkChart
     * @instance
     * @type {Array.<Object>}
     * @ojsignature {target: "Accessor", value: {GetterType: "Array<oj.ojSparkChart.ReferenceObject>", SetterType: "Array<oj.ojSparkChart.ReferenceObject>"}, jsdocOverride: true}
     * @default []
     * 
     * @example <caption>Initialize the spark chart with the 
     * <code class="prettyprint">reference-objects</code> attribute specified:</caption>
     * &lt;oj-spark-chart reference-objects='[{"type": "area", "high": 10, "low": 2, "location": "front", "color": "red"},
     *                                     {"type": "line", "value": 9, "location": "front", "color": "yellow"}, 
     *                                     {"type": "area", "high": 10, "low": 0, "location": "back", "color": "green"}]'>&lt;/oj-spark-chart>
     *
     * 
     * &lt;oj-spark-chart reference-objects='[[referencePromise]]'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">referenceObjects</code>
     * property after initialization:</caption>
     * // Get one
     * var value = mySparkChart.referenceObjects[0];
     *
     * // Get all (The items getter always returns a Promise so there is no "get one" syntax)
     * var values = mySparkChart.referenceObjects;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * mySparkChart.referenceObjects=[{"type": "area", "high": 10, "low": 2, "location": "front", "color": "red"}, 
     *                                {"type": "line", "value": 9, "location": "front", "color": "yellow"}, 
     *                                {"type": "area", "high": 10, "low": 0, "location": "back", "color": "green"}];
     */
    referenceObjects: [],

    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojSparkChart
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     * 
     * @example <caption>Initialize the spark chart with the 
     * <code class="prettyprint">tooltip</code> attribute specified:</caption>
     * <!-- Using dot notation -->
     * &lt;oj-spark-chart tooltip.renderer='[[tooltipFun]]'>&lt;/oj-spark-chart>
     * 
     * &lt;oj-spark-chart tooltip='[[{"renderer": tooltipFun}]]'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">tooltip</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = mySparkChart.tooltip.renderer;
     * 
     * // Get all
     * var values = mySparkChart.tooltip;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * mySparkChart.setProperty('tooltip.renderer', tooltipFun);
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * mySparkChart.tooltip={'renderer': tooltipFun};
     */
    tooltip: {
      /**
       *  A function that returns a custom tooltip. The function takes a dataContext argument, 
       *  provided by the chart, with the following properties: 
       *  <ul>
       *   <li>parentElement: The tooltip element. The function can directly modify or append content to this element.</li>
       *   <li>color: The color of the chart.</li>
       *   <li>componentElement: The spark chart element.</li>
       *  </ul>
       *  The function should return an Object that contains only one of the two properties:
       *  <ul>
       *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
       *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
       *  </ul>
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojSparkChart
       * @instance
       * @type {function(Object):Object|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojSparkChart.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))|null", jsdocOverride: true}
       * @example <caption>See the <a href="#tooltip">tooltip</a> attribute for usage examples.</caption>
       */
      renderer: null
    },
  
    /**
     * The chart type.
     * @expose
     * @name type
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "area"
     * @ojvalue {string} "lineWithArea"
     * @ojvalue {string} "bar"
     * @ojvalue {string} "line"
     * @default "line"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">type</code> attribute specified:</caption>
     * &lt;oj-spark-chart type='area'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">type</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.type;
     * 
     * // setter
     * mySparkChart.type="area";
     */
    type: "line",
    
    /**
     * The color of the data items. The default value varies based on theme.
     * @expose
     * @name color
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">color</code> attribute specified:</caption>
     * &lt;oj-spark-chart color='rgb(35, 123, 177)'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">color</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.color;
     * 
     * // setter
     * mySparkChart.color="rgb(35, 123, 177)";
     */
    color: "#267DB3",
    
    /**
     * The color of the area in area or lineWithArea spark chart.
     * @expose
     * @name areaColor
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">area-color</code> attribute specified:</caption>
     * &lt;oj-spark-chart type='area' area-color='red'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">areaColor</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.areaColor;
     * 
     * // setter
     * mySparkChart.areaColor="red";
     */
    areaColor: "", 
  
    /**
     * The CSS style class to apply if the type is area or lineWithArea. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name areaSvgClassName
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">area-svg-class-name</code> attribute specified:</caption>
     * &lt;oj-spark-chart type='lineWithArea' area-svg-class-name='svgClassName'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">areaSvgClassName</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.areaSvgClassName;
     * 
     * // setter
     * mySparkChart.areaSvgClassName = "svgClassName";
     */
    areaSvgClassName: "",
    
    /**
     * The inline style to apply if the type is area or lineWithArea. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name areaSvgStyle
     * @memberof oj.ojSparkChart
     * @instance
     * @type {Object}
     * @default {}
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">area-svg-style</code> attribute specified:</caption>
     * &lt;oj-spark-chart type='lineWithArea' area-svg-style='{"fill":"url(someURL#filterId)"}'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">areaSvgStyle</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.areaSvgStyle;
     * 
     * // setter
     * mySparkChart.areaSvgStyle = {"fill":"url(someURL#filterId)"};
     */
    areaSvgStyle: {},

    /**
     * The CSS style class to apply to the data items. For type lineWithArea, this style will only be applied to the line if areaClassName is also specified. The style class and inline style will override any other styling specified through the properties. For tooltips, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name svgClassName
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">svg-class-name</code> attribute specified:</caption>
     * &lt;oj-spark-chart svg-class-name='className'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">svgClassName</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.svgClassName;
     * 
     * // setter
     * mySparkChart.svgClassName = "className";
     */
    svgClassName: "",
    
    /**
     * The inline style to apply to the data items. For type lineWithArea, this style will only be applied to the line if areaStyle is also specified. The style class and inline style will override any other styling specified through the properties. For tooltips, it's recommended to also pass a representative color to the color attribute.
     * @expose
     * @name svgStyle
     * @memberof oj.ojSparkChart
     * @instance
     * @type {Object}
     * @default {}
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">svg-style</code> attribute specified:</caption>
     * &lt;oj-spark-chart svg-style='{"fill":"url(someURL#filterId)"}'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">svgStyle</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.svgStyle;
     * 
     * // setter
     * mySparkChart.svgStyle = {"fill":"url(someURL#filterId)"};
     */
    svgStyle: {},
    
    /**
     * The color of the first data item.
     * @expose
     * @name firstColor
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">first-color</code> attribute specified:</caption>
     * &lt;oj-spark-chart first-color='yellow'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">firstColor</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.firstColor;
     * 
     * // setter
     * mySparkChart.firstColor = "yellow";
     */
    firstColor: "",
    
    /**
     * The color of the last data item.
     * @expose
     * @name lastColor
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">last-color</code> attribute specified:</caption>
     * &lt;oj-spark-chart last-color='red'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">lastColor</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.lastColor;
     * 
     * // setter
     * mySparkChart.lastColor = "red";
     */
    lastColor: "",  
    
    /**
     * The color of the data item with the greatest value.
     * @expose
     * @name highColor
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">high-color</code> attribute specified:</caption>
     * &lt;oj-spark-chart high-color='blue'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">highColor</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.highColor;
     * 
     * // setter
     * mySparkChart.highColor = "blue";
     */
    highColor: "",
    
    /**
     * The color of the data item with the lowest value.
     * @expose
     * @name lowColor
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">low-color</code> attribute specified:</caption>
     * &lt;oj-spark-chart low-color='blue'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">lowColor</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.lowColor;
     * 
     * // setter
     * mySparkChart.lowColor = "blue";
     */
    lowColor: "",
    
    /**
     * The duration of the animations in milliseconds. The default value comes from the CSS and varies based on theme.
     * @expose
     * @name animationDuration
     * @memberof oj.ojSparkChart
     * @instance
     * @type {?number}
     * @ojunits milliseconds
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">animation-duration</code> attribute specified:</caption>
     * &lt;oj-spark-chart animation-duration='50'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">animationDuration</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.animationDuration;
     * 
     * // setter
     * mySparkChart.animationDuration = 50;
     */
    animationDuration: undefined,
     
    /**
     * Defines the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">animation-on-data-change</code> attribute specified:</caption>
     * &lt;oj-spark-chart animation-on-data-change='auto'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">animationOnDataChange</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.animationOnDataChange;
     * 
     * // setter
     * mySparkChart.animationOnDataChange = "auto";
     */
    animationOnDataChange: "none",
    
    /**
     * Defines the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">animation-on-display</code> attribute specified:</caption>
     * &lt;oj-spark-chart animation-on-display='auto'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">animationOnDisplay</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.animationOnDisplay;
     * 
     * // setter
     * mySparkChart.animationOnDisplay = "auto";
     */
    animationOnDisplay: "none",
     
    /**
     * Defines whether visual effects such as overlays are applied to the spark chart.
     * @expose
     * @name visualEffects
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "auto"
     * @default "auto"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">visual-effects</code> attribute specified:</caption>
     * &lt;oj-spark-chart visual-effects='none'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">visualEffects</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.visualEffects;
     * 
     * // setter
     * mySparkChart.visualEffects = "none";
     */
    visualEffects: "auto",
    
    /**
     * Defines whether the axis baseline starts at the minimum value of the data or at zero.
     * @expose
     * @name baselineScaling
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "zero"
     * @ojvalue {string} "min"
     * @default "min"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">baseline-scaling</code> attribute specified:</caption>
     * &lt;oj-spark-chart baseline-scaling='zero'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">baselineScaling</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.baselineScaling;
     * 
     * // setter
     * mySparkChart.baselineScaling = "zero";
     */
    baselineScaling: "min",
     
    /**
     * The width of the data line. Only applies to line spark charts.
     * @expose
     * @name lineWidth
     * @memberof oj.ojSparkChart
     * @instance
     * @type {number}
     * @default 1
     * @ojunits pixels
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">line-width</code> attribute specified:</caption>
     * &lt;oj-spark-chart line-width='4'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">lineWidth</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.lineWidth;
     * 
     * // setter
     * mySparkChart.lineWidth = 4;
     */
    lineWidth: 1,
     
    /**
     * The line style of the data line. Only applies to line spark charts.
     * @expose
     * @name lineStyle
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "dotted"
     * @ojvalue {string} "dashed"
     * @ojvalue {string} "solid"
     * @default "solid"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">line-style</code> attribute specified:</caption>
     * &lt;oj-spark-chart type="line" line-style='dotted'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">lineStyle</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.lineStyle;
     * 
     * // setter
     * mySparkChart.lineStyle = "dotted";
     */
    lineStyle: "solid",
    
    /**
     * The line type of the data line or area. Only applies to line and area spark charts.
     * @expose
     * @name lineType
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "curved"
     * @ojvalue {string} "stepped"
     * @ojvalue {string} "centeredStepped"
     * @ojvalue {string} "segmented"
     * @ojvalue {string} "centeredSegmented"
     * @ojvalue {string} "none"
     * @ojvalue {string} "straight"
     * @default "straight"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">line-type</code> attribute specified:</caption>
     * &lt;oj-spark-chart type="line" line-type='curved'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">lineType</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.lineType;
     * 
     * // setter
     * mySparkChart.lineType = "curved";
     */
    lineType: "straight",
     
    /**
     * The shape of the data markers. Can take the name of a built-in shape or the svg path commands for a custom shape. Only applies to line and area spark charts.
     * @expose
     * @name markerShape
     * @memberof oj.ojSparkChart
     * @instance
     * @type {string}
     * @ojvalue {string} "square"
     * @ojvalue {string} "circle"
     * @ojvalue {string} "diamond"
     * @ojvalue {string} "plus"
     * @ojvalue {string} "triangleDown"
     * @ojvalue {string} "triangleUp"
     * @ojvalue {string} "human"
     * @ojvalue {string} "star"
     * @ojvalue {string} "auto"
     * @default "auto"
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">marker-shape</code> attribute specified:</caption>
     * &lt;oj-spark-chart type="area" marker-shape="triangleUp">&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">markerShape</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.markerShape;
     * 
     * // setter
     * mySparkChart.markerShape = "triangleUp";
     */
    markerShape: "auto",
     
    /**
     * The size of the data markers in pixels. Only applies to line and area spark charts.
     * @expose
     * @name markerSize
     * @memberof oj.ojSparkChart
     * @instance
     * @type {number}
     * @default 5
     * @ojunits pixels
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">marker-size</code> attribute specified:</caption>
     * &lt;oj-spark-chart type="area" marker-size='15'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">markerSize</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.markerSize;
     * 
     * // setter
     * mySparkChart.markerSize = 15;
     */
    markerSize: 5,
     
    /**
     * Specifies the width of the bar gap as a ratio of the item width. The valid value is a number from 0 to 1.
     * @expose
     * @name barGapRatio
     * @memberof oj.ojSparkChart
     * @instance
     * @type {number}
     * @default 0.25
     * 
     * @example <caption>Initialize the spark chart with the <code class="prettyprint">bar-gap-ratio</code> attribute specified:</caption>
     * &lt;oj-spark-chart type="bar" bar-gap-ratio='0.12'>&lt;/oj-spark-chart>
     * 
     * @example <caption>Get or set the <code class="prettyprint">barGapRatio</code> 
     * property after initialization:</caption>
     * // getter
     * var value = mySparkChart.barGapRatio;
     * 
     * // setter
     * mySparkChart.barGapRatio = 0.12;
     */
    barGapRatio: 0.25
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    this._focusable({'element': this.element, 'applyHighlight': true});
    return dvt.SparkChart.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-sparkchart');
    return styleClasses;
  },
  
  //** @inheritdoc */
  _GetChildStyleClasses: function() {
    var styleClasses = this._super();
    styleClasses['oj-dvtbase oj-sparkchart'] = {'path': 'animationDuration', 'property': 'ANIM_DUR'};
    return styleClasses;
  },

  //** @inheritdoc */
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtUtilBundle.CHART'] = translations['componentName'];
    return ret;
  },

  //** @inheritdoc */
  _Render : function() {
    // Display the title of the surrounding div as the tooltip. Remove title from div to avoid browser default tooltip.
    if(this.element.attr('title'))
    {
      this.options['shortDesc'] =  this.element.attr('title');
      this.element.data( this.element,'title', this.element.attr('title'));
      this.element.removeAttr('title');
    }
    else if (this.element.data('title'))
      this.options['shortDesc'] =  this.element.data('title');

    // Call the super to render
    this._super();
  },

  /**
   * Returns an object with the following properties for automation testing verification of the data item with
   * the specified item index.
   * @param {number} itemIndex The item index
   * @property {string} borderColor The border color of the item	
   * @property {string} color The color of the item	
   * @property {Date} date The date (x value) of the item	
   * @property {number} high The high value for a range item	
   * @property {number} low  The low value for a range item	
   * @property {number} value The value of the item
   * @return {Object|null} An object containing properties for the data item, or null if none exists.
   * @ojsignature {target: "Type", value: "oj.ojSparkChart.ItemContext|null", jsdocOverride: true, for: "returns"}
   * @expose
   * @instance
   * @memberof oj.ojSparkChart
   */
  getDataItem: function(itemIndex) {
    var ret = this._component.getAutomation().getDataItem(itemIndex);

    // : Provide backwards compatibility for getters until 1.2.0.
    this._AddAutomationGetters(ret);
    if (ret)
      ret['getFloatValue'] = ret['getLow'];

    return ret;
  },

  //** @inheritdoc */
  _GetComponentDeferredDataPaths : function() {
    return {'root': ['items', 'data']};
  },

  /**
   * Adds getters for the properties on the specified map. 
   * @param {Object|null} map
   * @memberof oj.ojSparkChart
   * @instance
   * @protected
   */
  _AddAutomationGetters: function(map) {
    if(!map)
      return;

    // These getters are deprecated in 3.0.0
    var props = {};
    for (var key in map) {
      this._addGetter(map, key, props);
    }
    Object.defineProperties(map, props);
  },

  /**
   * Adds getter for the specified property on the specified properties map.
   * @param {Object} map
   * @param {string} key
   * @param {Object} props The properties map onto which the getter will be added.
   * @memberof oj.ojSparkChart
   * @instance
   * @private
   */
  _addGetter: function(map, key, props) {
    var prefix = (key == 'selected') ? 'is' : 'get';
    var getterName = prefix + key.charAt(0).toUpperCase() + key.slice(1);
    props[getterName] = {'value': function() { return map[key] }};
  },
});

/**
 * <p>This element has no touch interaction.  </p>
 *
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojSparkChart
 */

/**
 * <p>This element has no keyboard interaction.  </p>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojSparkChart
 */
 
/**
 * The value of the data item.
 * @expose
 * @name items[].value
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The low value for range bar/area. Define 'low' and 'high' instead of 'value' to create a range bar/area spark chart.
 * @expose
 * @name items[].low
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The high value for range bar/area. Define 'low' and 'high' instead of 'value' to create a range bar/area spark chart.
 * @expose
 * @name items[].high
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The date for the data item. The date should only be specified if the interval between data items is irregular.
 * @expose
 * @name items[].date
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The color of the bar or marker for the data item. This override can be used to highlight important values or thresholds.
 * @expose
 * @name items[].color
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @expose
 * @name items[].svgClassName
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @expose
 * @name items[].svgStyle
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {object}
 * @default null
 */
/**
 * Defines whether a marker should be displayed for the data item. Only applies to line and area spark charts.
 * @expose
 * @name items[].markerDisplayed
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * The default border color for the data items.
 * @expose
 * @name items[].borderColor
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The shape of the data markers. Can take the name of a built-in shape or the svg path commands for a custom shape. Only applies to line and area spark charts.
 * @expose
 * @name items[].markerShape
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {string}
 * @ojvalue {string} "square"
 * @ojvalue {string} "circle"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "human"
 * @ojvalue {string} "star"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The size of the data markers in pixels. Only applies to line and area spark charts.
 * @expose
 * @name items[].markerSize
 * @memberof! oj.ojSparkChart
 * @instance
 * @type {number}
 * @default null
 */
 /**
  * The type of reference object being shown.
  * @expose
  * @name referenceObjects[].type
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {string}
  * @ojvalue {string} "area"
  * @ojvalue {string} "line"
  * @default "line"
  */
 /**
  * The location of the reference object relative to the data items.
  * @expose
  * @name referenceObjects[].location
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {string}
  * @ojvalue {string} "front"
  * @ojvalue {string} "back"
  * @default "back"
  */
 /**
  * The color of the reference object.
  * @expose
  * @name referenceObjects[].color
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {string}
  * @default null
  */
 /**
  * The width of a reference line.
  * @expose
  * @name referenceObjects[].lineWidth
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {number}
  * @default null
  */
 /**
  * The line style of a reference line.
  * @expose
  * @name referenceObjects[].lineStyle
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {string}
  * @ojvalue {string} "dotted"
  * @ojvalue {string} "dashed"
  * @default null
  */
 /**
  * The CSS style class to apply to the reference object. The style class and inline style will override any other styling specified through the properties.
  * @expose
  * @name referenceObjects[].svgClassName
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {string}
  * @default null
  */
 /**
  * The inline style to apply to the reference object. The style class and inline style will override any other styling specified through the properties.
  * @expose
  * @name referenceObjects[].svgStyle
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {object}
  * @default null
  */
 /**
  * The value of a reference line.
  * @expose
  * @name referenceObjects[].value
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {number}
  * @default null
  */
 /**
  * The low value of a reference area.
  * @expose
  * @name referenceObjects[].low
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {number}
  * @default null
  */
 /**
  * The high value of a reference area.
  * @expose
  * @name referenceObjects[].high
  * @memberof! oj.ojSparkChart
  * @instance
  * @type {number}
  * @default null
  */

// PROPERTY TYPEDEFS

/**
 * @typedef {Object} oj.ojSparkChart.Item
 * @ignore
 * @property {string} [borderColor] The default border color for the data items.
 * @property {string} [color] The color of the bar or marker for the data item. This override can be used to highlight important values or thresholds.
 * @property {Date} [date] The date for the data item. The date should only be specified if the interval between data items is irregular.
 * @property {number} [high] The high value for range bar/area. Define 'low' and 'high' instead of 'value' to create a range bar/area spark chart.
 * @property {number} [low] The low value for range bar/area. Define 'low' and 'high' instead of 'value' to create a range bar/area spark chart.
 * @property {"on"|"off"} [markerDisplayed="off"] Defines whether a marker should be displayed for the data item. Only applies to line and area spark charts.
 * @property {"square"|"circle"|"diamond"|"plus"|"triangleDown"|"triangleUp"|"human"|"star"|"auto"|string} [markerShape="auto"] The shape of the data markers. Can take the name of a built-in shape or the svg path commands for a custom shape. Only applies to line and area spark charts.
 * @property {number} [markerSize] The size of the data markers in pixels. Only applies to line and area spark charts.
 * @property {string} [svgClassName] The CSS style class to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @property {Object} [svgStyle] The inline style to apply to the data item. The style class and inline style will override any other styling specified through the properties. For tooltips and hover interactivity, it's recommended to also pass a representative color to the item color attribute.
 * @property {number} [value] The value of the data item.
 */
 
/**
 * @typedef {Object} oj.ojSparkChart.ReferenceObject
 * @ignore
 * @property {string} [color] The color of the reference object.
 * @property {number} [high] The high value of a reference area.
 * @property {number} [lineWidth] The width of a reference line.
 * @property {"dotted"|"dashed"|"solid"} [lineStyle="solid"] The line style of a reference line.
 * @property {"front"|"back"} [location="back"]  The location of the reference object relative to the data items.
 * @property {number} [low] The low value of a reference area.
 * @property {string} [svgClassName] The CSS style class to apply to the reference object. The style class and inline style will override any other styling specified through the properties.
 * @property {Object} [svgStyle] The inline style to apply to the reference object. The style class and inline style will override any other styling specified through the properties.
 * @property {"area"|"line"} [type="line"] The type of reference object being shown.
 * @property {number} [value] The value of a reference line.
 */

/**
 * @typedef {Object} oj.ojSparkChart.TooltipContext
 * @ignore
 * @property {string} color The color of the chart.
 * @property {Element} componentElement The spark chart element.
 * @property {Element} parentElement The tooltip element. The function can directly modify or append content to this element.
 */
 
// METHOD TYPEDEFS

/**
 * @typedef {Object} oj.ojSparkChart.ItemContext
 * @ignore
 * @property {string} borderColor The border color of the item
 * @property {string} color The color of the item
 * @property {Date} date The date (x value) of the item
 * @property {number} high The high value for a range item
 * @property {number} low  The low value for a range item
 * @property {number} value The value of the item
 */
 
/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojChartMeta = {
  "properties": {
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "slideToLeft", "slideToRight", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "alphaFade", "zoom", "none"]
    },
    "coordinateSystem": {
      "type": "string",
      "enumValues": ["polar", "cartesian"]
    },
    "dataCursor": {
      "type": "string",
      "enumValues": ["on", "off", "auto"]
    },
    "dataCursorBehavior": {
      "type": "string",
      "enumValues": ["smooth", "snap", "auto"]
    },
    "dataCursorPosition": {
      "type": "object",
      "properties": {
        "x": {
          "type": "number|string"
        },
        "y": {
          "type": "number"
        },
        "y2": {
          "type": "number"
        }
      },
      "writeback": true
    },
    "dataLabel": {},
    "dnd": {
      "type": "object",
      "properties": {
        "drag": {
          "type": "object",
          "properties": {
            "groups": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "drag": {},
                "dragEnd": {},
                "dragStart": {}
              }
            },
            "items": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "drag": {},
                "dragEnd": {},
                "dragStart": {}
              }
            },
            "series": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "drag": {},
                "dragEnd": {},
                "dragStart": {}
              }
            }
          }
        },
        "drop": {
          "type": "object",
          "properties": {
            "legend": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            },
            "plotArea": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            },
            "xAxis": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            },
            "y2Axis": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            },
            "yAxis": {
              "type": "object",
              "properties": {
                "dataTypes": {
                  type: "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            }
          }
        }
      }
    },
    "dragMode": {
      "type": "string",
      "enumValues": ["pan", "zoom", "select", "off", "user"]
    },
    "drilling": {
      "type": "string",
      "enumValues": ["on", "seriesOnly", "groupsOnly", "off"]
    },
    "groups": {
      "type": "Array<any>|Promise"
    },
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "hideAndShowBehavior": {
      "type": "string",
      "enumValues": ["withRescale", "withoutRescale", "none"]
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string",
      "enumValues": ["any", "all"]
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "initialZooming": {
      "type": "string",
      "enumValues": ["first", "last", "none"]
    },
    "legend": {
      "type": "object",
      "properties": {
        "backgroundColor": {
          "type": "string"
        },
        "borderColor": {
          "type": "string"
        },
        "maxSize": {
          "type": "string"
        },
        "position": {
          "type": "string",
          "enumValues": ["start", "end", "bottom", "top", "auto"]
        },
        "referenceObjectSection": {
          "type": "object",
          "properties": {
            "title": {
              "type": "string"
            },
            "titleHalign": {
              "type": "string",
              "enumValues": ["center", "end", "start"]
            },
            "titleStyle": {
              "type": "object"
            }
          }
        },
        "rendered": {
          "type": "string",
          "enumValues": ["on", "off", "auto"]
        },
        "scrolling": {
          "type": "string",
          "enumValues": ["off", "asNeeded"]
        },
        "sections": {
          "type": "Array<object>"
        },
        "seriesSection": {
          "type": "object",
          "properties": {
            "title": {
              "type": "string"
            },
            "titleHalign": {
              "type": "string",
              "enumValues": ["center", "end", "start"]
            },
            "titleStyle": {
              "type": "object"
            }
          }
        },
        "size": {
          "type": "string"
        },
        "symbolHeight": {
          "type": "number"
        },
        "symbolWidth": {
          "type": "number"
        },
        "textStyle": {
          "type": "object"
        },
        "title": {
          "type": "string"
        },
        "titleHalign": {
          "type": "string"
        },
        "titleStyle": {
          "type": "object"
        }
      }
    },
    "orientation": {
      "type": "string",
      "enumValues": ["horizontal", "vertical"]
    },
    "otherThreshold": {
      "type": "number"
    },
    "overview": {
      "type": "object",
      "properties": {
        "content": {
          "type": "object"
        },
        "height": {
          "type": "string"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["on", "off"]
        }
      }
    },
    "pieCenter": {
      "type": "object",
      "properties": {
        "converter": {
          "type": "object"
        },
        "label": {
          "type": "string"
        },
        "labelStyle": {
          "type": "object"
        },
        "renderer": {},
        "scaling": {
          "type": "string",
          "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
        }
      }
    },
    "pieCenterLabel": {
      "type": "object",
      "properties": {
        "style": {
          "type": "object"
        },
        "text": {
          "type": "string"
        }
      }
    },
    "plotArea": {
      "type": "object",
      "properties": {
        "backgroundColor": {
          "type": "string"
        },
        "borderColor": {
          "type": "string"
        },
        "borderWidth": {
          "type": "number"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["on", "off"]
        }
      }
    },
    "polarGridShape": {
      "type": "string",
      "enumValues": ["polygon", "circle"]
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["single", "multiple", "none"]
    },
    "series": {
      "type": "Array<object>|Promise"
    },
    "sorting": {
      "type": "string",
      "enumValues": ["ascending", "descending", "off"]
    },
    "splitDualY": {
      "type": "string",
      "enumValues": ["on", "off", "auto"]
    },
    "splitterPosition": {
      "type": "number"
    },
    "stack": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "stackLabel": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "styleDefaults": {
      "type": "object",
      "properties": {
        "animationDownColor": {
          "type": "string"
        },
        "animationDuration": {
          "type": "number"
        },
        "animationIndicators": {
          "type": "string",
          "enumValues": ["none", "all"]
        },
        "animationUpColor": {
          "type": "string"
        },
        "barGapRatio": {
          "type": "number"
        },
        "borderColor": {
          "type": "string"
        },
        "borderWidth": {
          "type": "number"
        },
        "boxPlot": {
          "type": "object",
          "properties": {
            "medianSvgClassName": {
              "type": "string"
            },
            "medianSvgStyle": {
              "type": "object"
            },
            "whiskerSvgClassName": {
              "type": "string"
            },
            "whiskerEndSvgClassName": {
              "type": "string"
            },
            "whiskerEndLength": {
              "type": "string"
            },
            "whiskerEndSvgStyle": {
              "type": "object"
            },
            "whiskerSvgStyle": {
              "type": "object"
            }
          }
        },
        "colors": {
          "type": "Array<string>"
        },
        "dataCursor": {
          "type": "object",
          "properties": {
            "lineColor": {
              "type": "string"
            },
            "lineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "lineWidth": {
              "type": "number"
            },
            "markerColor": {
              "type": "string"
            },
            "markerDisplayed": {
              "type": "string",
              "enumValues": ["on", "off"]
            },
            "markerSize": {
              "type": "number"
            }
          }
        },
        "dataItemGaps": {
          "type": "string"
        },
        "dataLabelPosition": {
          "type": "string|Array<string>",
          "enumValues": ["center", "outsideSlice", "aboveMarker", "belowMarker", "beforeMarker",
                         "afterMarker", "insideBarEdge", "outsideBarEdge", "none", "auto"]
        },
        "dataLabelStyle": {
          "type": "object|Array<object>"
        },
        "funnelBackgroundColor": {
          "type": "string"
        },
        "groupSeparators": {
          "type": "object",
          "properties": {
            "color": {
              "type": "string"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["auto", "off"]
            }
          }
        },
        "hoverBehaviorDelay": {
          "type": "number"
        },
        "lineStyle": {
          "type": "string",
          "enumValues": ["dotted", "dashed", "solid"]
        },
        "lineType": {
          "type": "string",
          "enumValues": ["straight", "curved", "stepped", "centeredStepped", "segmented", "centeredSegmented", "none", "auto"]
        },
        "lineWidth": {
          "type": "number"
        },
        "markerColor": {
          "type": "string"
        },
        "markerDisplayed": {
          "type": "string",
          "enumValues": ["on", "off", "auto"]
        },
        "markerShape": {
          "type": "string"
        },
        "markerSize": {
          "type": "number"
        },
        "marqueeBorderColor": {
          "type": "string"
        },
        "marqueeColor": {
          "type": "string"
        },
        "maxBarWidth": {
          "type": "number"
        },
        "otherColor": {
          "type": "string"
        },
        "patterns": {
          "type": "Array<string>"
        },
        "pieFeelerColor": {
          "type": "string"
        },
        "pieInnerRadius": {
          "type": "number"
        },
        "selectionEffect": {
          "type": "string",
          "enumValues": ["explode", "highlightAndExplode", "highlight"]
        },
        "seriesEffect": {
          "type": "string",
          "enumValues": ["color", "pattern", "gradient"]
        },
        "shapes": {
          "type": "Array<string>"
        },
        "stackLabelStyle": {
          "type": "object"
        },
        "stockFallingColor": {
          "type": "string"
        },
        "stockRangeColor": {
          "type": "string"
        },
        "stockRisingColor": {
          "type": "string"
        },
        "stockVolumeColor": {
          "type": "string"
        },
        "threeDEffect": {
          "type": "string",
          "enumValues": ["on", "off"]
        },
        "tooltipLabelStyle": {
          "type": "object"
        },
        "tooltipValueStyle": {
          "type": "object"
        }
      }
    },
    "timeAxisType": {
      "type": "string",
      "enumValues": ["enabled", "mixedFrequency", "skipGaps", "disabled", "auto"]
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "touchResponse": {
      "type": "string"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Chart"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelClose": {
          "type": "string",
          "value": "Close"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelDate": {
          "type": "string",
          "value": "Value"
        },
        "labelDefaultGroupName": {
          "type": "string",
          "value": "Group {0}"
        },
        "labelGroup": {
          "type": "string",
          "value": "Group"
        },
        "labelHigh": {
          "type": "string",
          "value": "High"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelLow": {
          "type": "string",
          "value": "Low"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "labelOpen": {
          "type": "string",
          "value": "Open"
        },
        "labelOther": {
          "type": "string",
          "value": "Other"
        },
        "labelPercentage": {
          "type": "string",
          "value": "Percentage"
        },
        "labelQ1": {
          "type": "string",
          "value": "Q1"
        },
        "labelQ2": {
          "type": "string",
          "value": "Q2"
        },
        "labelQ3": {
          "type": "string",
          "value": "Q3"
        },
        "labelSeries": {
          "type": "string",
          "value": "Series"
        },
        "labelTargetValue": {
          "type": "string",
          "value": "Target"
        },
        "labelValue": {
          "type": "string",
          "value": "Value"
        },
        "labelVolume": {
          "type": "string",
          "value": "Volume"
        },
        "labelX": {
          "type": "string",
          "value": "X"
        },
        "labelY": {
          "type": "string",
          "value": "Y"
        },
        "labelZ": {
          "type": "string",
          "value": "Z"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        },
        "tooltipPan": {
          "type": "string",
          "value": "Pan"
        },
        "tooltipSelect": {
          "type": "string",
          "value": "Marquee select"
        },
        "tooltipZoom": {
          "type": "string",
          "value": "Marquee zoom"
        }
      }
    },
    "type": {
      "type": "string",
      "enumValues": ["line", "area", "lineWithArea", "stock", "boxPlot", "combo", "pie", "scatter", "bubble", "funnel", "pyramid", "bar"]
    },
    "valueFormats": {
      "type": "object",
      "properties": {
        "series": {
          "type": "object",
          "properties": {
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "group": {
          "type": "object",
          "properties": {
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "x": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "y": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "y2": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "z": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "value": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "targetValue": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "low": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "high": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "open": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "close": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "volume": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "q1": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "q2": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "q3": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "label": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        }
      }
    },
    "xAxis": {
      "type": "object",
      "properties": {
        "axisLine": {
          "type": "object",
          "properties": {
            "lineColor": {
              "type": "string"
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["on", "off"]
            }
          }
        },
        "baselineScaling": {
          "type": "string",
          "enumValues": ["min", "zero"]
        },
        "dataMax": {
          "type": "number"
        },
        "dataMin": {
          "type": "number"
        },
        "majorTick": {
          "type": "object",
          "properties": {
            "baselineColor": {
              "type": "string",
              "enumValues": ["inherit", "auto"]
            },
            "baselineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "baselineWidth": {
              "type": "number"
            },
            "lineColor": {
              "type": "string"
            },
            "lineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["on", "off", "auto"]
            }
          }
        },
        "max": {
          "type": "number|string"
        },
        "maxSize": {
          "type": "string"
        },
        "min": {
          "type": "number|string"
        },
        "minorStep": {
          "type": "number"
        },
        "minorTick": {
          "type": "object",
          "properties": {
            "lineColor": {
              "type": "string"
            },
            "lineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["on", "off", "auto"]
            }
          }
        },
        "minStep": {
          "type": "number"
        },
        "referenceObjects": {
          "type": "Array<object>"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["off", "on"]
        },
        "scale": {
          "type": "string",
          "enumValues": ["log", "linear"]
        },
        "size": {
          "type": "string"
        },
        "step": {
          "type": "number"
        },
        "tickLabel": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["off", "on"]
            },
            "rotation": {
              "type": "string",
              "enumValues": ["none", "auto"]
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "style": {
              "type": "object"
            }
          }
        },
        "title": {
          "type": "string"
        },
        "titleStyle": {
          "type": "object"
        },
        "viewportEndGroup": {
          "type": "number|string"
        },
        "viewportMax": {
          "type": "number|string"
        },
        "viewportMin": {
          "type": "number|string"
        },
        "viewportStartGroup": {
          "type": "number|string"
        }
      }
    },
    "y2Axis": {
      "type": "object",
      "properties": {
        "alignTickMarks": {
          "type": "string",
          "enumValues": ["off", "on"]
        },
        "axisLine": {
          "type": "object",
          "properties": {
            "lineColor": {
              "type": "string"
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["on", "off", "auto"]
            }
          }
        },
        "baselineScaling": {
          "type": "string",
          "enumValues": ["min", "zero"]
        },
        "dataMax": {
          "type": "number"
        },
        "dataMin": {
          "type": "number"
        },
        "majorTick": {
          "type": "object",
          "properties": {
            "baselineColor": {
              "type": "string",
              "enumValues": ["inherit", "auto"]
            },
            "baselineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "baselineWidth": {
              "type": "number"
            },
            "lineColor": {
              "type": "string"
            },
            "lineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["off", "on", "auto"]
            }
          }
        },
        "max": {
          "type": "number"
        },
        "maxSize": {
          "type": "string"
        },
        "min": {
          "type": "number"
        },
        "minorStep": {
          "type": "number"
        },
        "minorTick": {
          "type": "object",
          "properties": {
            "lineColor": {
              "type": "string"
            },
            "lineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["on", "auto", "off"]
            }
          }
        },
        "minStep": {
          "type": "number"
        },
        "position": {
          "type": "string",
          "enumValues": ["start", "end", "top", "bottom", "auto"]
        },
        "referenceObjects": {
          "type": "Array<object>"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["off", "on"]
        },
        "scale": {
          "type": "string",
          "enumValues": ["log", "linear"]
        },
        "size": {
          "type": "string"
        },
        "step": {
          "type": "number"
        },
        "tickLabel": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "position": {
              "type": "string",
              "enumValues": ["inside", "outisde"]
            },
            "rendered": {
              "type": "string",
              "enumValues": ["off", "on"]
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "style": {
              "type": "object"
            }
          }
        },
        "title": {
          "type": "string"
        },
        "titleStyle": {
          "type": "object"
        }
      }
    },
    "yAxis": {
      "type": "object",
      "properties": {
        "axisLine": {
          "type": "object",
          "properties": {
            "lineColor": {
              "type": "string"
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["off", "on", "auto"]
            }
          }
        },
        "baselineScaling": {
          "type": "string",
          "enumValues": ["min", "zero"]
        },
        "dataMax": {
          "type": "number"
        },
        "dataMin": {
          "type": "number"
        },
        "majorTick": {
          "type": "object",
          "properties": {
            "baselineColor": {
              "type": "string",
              "enumValues": ["inherit", "auto"]
            },
            "baselineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "baselineWidth": {
              "type": "number"
            },
            "lineColor": {
              "type": "string"
            },
            "lineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["on", "off", "auto"]
            }
          }
        },
        "max": {
          "type": "number"
        },
        "maxSize": {
          "type": "string"
        },
        "min": {
          "type": "number"
        },
        "minorStep": {
          "type": "number"
        },
        "minorTick": {
          "type": "object",
          "properties": {
            "lineColor": {
              "type": "string"
            },
            "lineStyle": {
              "type": "string",
              "enumValues": ["dotted", "dashed", "solid"]
            },
            "lineWidth": {
              "type": "number"
            },
            "rendered": {
              "type": "string",
              "enumValues": ["on", "off", "auto"]
            }
          }
        },
        "minStep": {
          "type": "number"
        },
        "position": {
          "type": "string",
          "enumValues": ["start", "end", "top", "bottom", "auto"]
        },
        "referenceObjects": {
          "type": "Array<object>"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["off", "on"]
        },
        "scale": {
          "type": "string",
          "enumValues": ["log", "linear"]
        },
        "size": {
          "type": "string"
        },
        "step": {
          "type": "number"
        },
        "tickLabel": {
          "type": "object",
          "properties": {
            "converter": {
              "type": "object"
            },
            "position": {
              "type": "string",
              "enumValues": ["inside", "outside"]
            },
            "rendered": {
              "type": "string",
              "enumValues": ["off", "on"]
            },
            "scaling": {
              "type": "string",
              "enumValues": ["none", "auto", "thousand", "million", "billion", "trillion", "quadrillion"]
            },
            "style": {
              "type": "object"
            }
          }
        },
        "title": {
          "type": "string"
        },
        "titleStyle": {
          "type": "object"
        },
        "viewportMax": {
          "type": "number"
        },
        "viewportMin": {
          "type": "number"
        }
      }
    },
    "zoomAndScroll": {
      "type": "string"
    },
    "zoomDirection": {
      "type": "string"
    }
  },
  "methods": {
    "getContextByNode": {},
    "getDataItem": {},
    "getGroup": {},
    "getGroupCount": {},
    "getLegend": {},
    "getPlotArea": {},
    "getSeries": {},
    "getSeriesCount": {},
    "getValuesAt": {},
    "getXAxis": {},
    "getY2Axis": {},
    "getYAxis": {}
  },
  "events": {
    "drill": {},
    "viewportChange": {},
    "selectInput": {},
    "viewportChangeInput": {}
  },
  "extension": {
    _WIDGET_NAME: "ojChart"
  }
};
oj.CustomElementBridge.registerMetadata('oj-chart', 'dvtBaseComponent', ojChartMeta);
// Get the combined meta of superclass which contains a shape parse function generator
var dvtMeta = oj.CustomElementBridge.getMetadata('oj-chart');
var _CHART_SHAPE_ENUMS = {
  "square": true,
  "circle": true,
  "diamond": true,
  "plus": true,
  "triangleDown": true,
  "triangleUp": true,
  "human": true,
  "star": true,
  "auto": true
};
var chartShapeParser = dvtMeta['extension']._DVT_PARSE_FUNC({'style-defaults.marker-shape': true}, _CHART_SHAPE_ENUMS)
var chartParseFunction = function(value, name, metadata, defaultParseFunction) {
  if (metadata['type'] === 'number|string')
    return isNaN(value) ? value : Number(value);
  else if (name === 'style-defaults.data-label-style')
    return JSON.parse(value);
  else // shape parser will handle default cases as well
    return chartShapeParser(value, name, metadata, defaultParseFunction);
};
oj.CustomElementBridge.register('oj-chart', {
  'metadata': dvtMeta,
  'parseFunction': chartParseFunction
});
})();

(function() {
var ojSparkChartMeta = {
  "properties": {
    "animationDuration": {
      "type": "number"
    },
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "areaSvgClassName": {
      "type": "string"
    },
    "areaColor": {
      "type": "string"
    },
    "areaSvgStyle": {
      "type": "object"
    },
    "barGapRatio": {
      "type": "number"
    },
    "baselineScaling": {
      "type": "string",
      "enumValues": ["zero", "min"]
    },
    "svgClassName": {
      "type": "string"
    },
    "color": {
      "type": "string"
    },
    "firstColor": {
      "type": "string"
    },
    "highColor": {
      "type": "string"
    },
    "items": {
      "type": "Array<number>|Promise"
    },
    "lastColor": {
      "type": "string"
    },
    "lineStyle": {
      "type": "string",
      "enumValues": ["dotted", "dashed", "solid"]
    },
    "lineType": {
      "type": "string",
      "enumValues": ["curved", "stepped", "centeredStepped", "segmented", "centeredSegmented", "none", "straight"]
    },
    "lineWidth": {
      "type": "number"
    },
    "lowColor": {
      "type": "string"
    },
    "markerShape": {
      "type": "string"
    },
    "markerSize": {
      "type": "number"
    },
    "referenceObjects": {
      "type": "Array<object>"
    },
    "svgStyle": {
      "type": "object"
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Chart"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    },
    "type": {
      "type": "string",
      "enumValues": ["area", "lineWithArea", "bar", "line"]
    },
    "visualEffects": {
      "type": "string",
      "enumValues": ["none", "auto"]
    }
  },
  "methods": {
    "getDataItem": {}
  },
  "extension": {
    _WIDGET_NAME: "ojSparkChart"
  }
};
oj.CustomElementBridge.registerMetadata('oj-spark-chart', 'dvtBaseComponent', ojSparkChartMeta);
// Get the combined meta of superclass which contains a shape parse function generator
var dvtMeta = oj.CustomElementBridge.getMetadata('oj-spark-chart');
var _SPARK_SHAPE_ENUMS = {
  "square": true,
  "circle": true,
  "diamond": true,
  "plus": true,
  "triangleDown": true,
  "triangleUp": true,
  "human": true,
  "star": true,
  "auto":true
};
oj.CustomElementBridge.register('oj-spark-chart', {
  'metadata': dvtMeta,
  'parseFunction': dvtMeta['extension']._DVT_PARSE_FUNC({'marker-shape': true}, _SPARK_SHAPE_ENUMS)
});
})();

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
(function() {

/*!
 * JET Checkboxset @VERSION
 */
/**
 * @ojcomponent oj.ojCheckboxset
 * @augments oj.editableValue
 * @since 0.6
 * @ojshortdesc A grouping of related checkboxes where any number of boxes may be checked.
 * @ojrole checkbox
 * @ojrole checkboxgroup
 * @ojrole option
 * @ojdisplayname Checkbox Set
 * @ojstatus preview
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojCheckboxset extends editableValue<Array<any>, ojCheckboxsetSettableProperties>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojCheckboxsetSettableProperties extends editableValueSettableProperties<Array<any>>",
 *                for: "SettableProperties"
 *               }
 *              ]
 *
 * @classdesc
 * <h3 id="checkboxsetOverview-section">
 *   JET Checkboxset
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#checkboxsetOverview-section"></a>
 * </h3> 
 * <p>
 * The JET oj-checkboxset component manages a set of
 * <code class="prettyprint">oj-option</code> child elements and creates the necessary dom elements for
 * the actual checkboxes.
 * </p>
 * <p>To use an oj-oheckboxset, add one or more oj-option child elements for each checkbox desired. 
 * Note, if you add or remove an oj-option after the oj-checkboxset is rendered, you should call
 * refresh() on the oj-checkboxset.
 * </p>
 * <p>
 *  You can enable and disable an oj-checkboxset,
 *  which will enable and disable all contained checkboxes.
 * </p>
 * <p>
 *  oj-checkboxset does not have a readonly attribute since HTML does not support
 *  readonly on radios and checkboxes.
 * </p>
 * <p>Use <code class="prettyprint">labelled-by</code> to associate an oj-label with
 * the checkboxset component. Doing this also makes the checkboxset accessible.
 * </p>
 * <p>
 * The oj-label element is not required. If you don't use an oj-label element,
 * then you need to set <code class="prettyprint">aria-label</code> on the input for accessibility.
 * </p>
 * 
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>JET Checkboxset takes care of setting <code class="prettyprint">role="group"</code>
 * on the oj-checkboxset element.
 *
 * <p>As shown in the online demos, the application is responsible for applying
 * <code class="prettyprint">labelled-by</code>
 * to point to an <code class="prettyprint">oj-label</code> element for the oj-checkboxset.
 * <p>Disabled content: JET supports an accessible luminosity contrast ratio,
 * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>,
 * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which
 * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user
 * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio
 * required of enabled content, it cannot be used to convey meaningful information.<p>
 * <h3 id="label-section">
 *   Label and Checkboxset
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * For accessibility, you should associate an oj-label element with the checkboxset
 * by putting an <code class="prettyprint">id</code> on the oj-label, and then setting the
 * <code class="prettyprint">labelled-by</code> attribute on the oj-checkboxset to be the
 * oj-label's <code class="prettyprint">id</code>.
 * </p>
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 * {@ojinclude "name":"stylingDoc"}
 *
 * @example <caption>Initialize the checkboxset with no attributes specified:</caption>
 * &lt;oj-checkboxset id="colorCheckbox" value="{{currentColor}}">
 *   &lt;oj-option value="blue">Blue&lt;/oj-option>
 *   &lt;oj-option value="green">Green&lt;/oj-option>
 * &lt;/oj-checkboxset>
 *
 * @example <caption>Initialize component and an associated oj-label component</caption>
 * &lt;oj-label id="grouplabel">Greetings&lt;/oj-label>
 * &lt;oj-checkboxset id="checkboxset" labelled-by="grouplabel" value="{{currentGreeting}}">
 *   &lt;oj-option id="helloid" value="hello">Hello&lt;/oj-option>
 *   &lt;oj-option id="bonjourid" value="bonjour"/>Bonjour&lt;/oj-option>
 *   &lt;oj-option id="ciaoid" value="ciao"/>Ciao&lt;/oj-option>
 * &lt;/oj-checkboxset>
 * <br/>
 * // set the value to "ciao". (The 'ciao' checkbox will be checked)
 * myComp.value = ["ciao"];
 *
 * @example <caption>Initialize an oj-checkboxset via the JET component binding:</caption>
 * &lt;oj-label id="grouplabel">Time&lt;/oj-label>
 * &lt;oj-checkboxset id="checkboxset" value="{{checkboxsetValue}}" 
 *                    labelled-by="grouplabel">
 *   &lt;oj-option id="morningid" value="morning">Morning&lt;/oj-option>
 *   &lt;oj-option id="nightid" value="night">Night&lt;/oj-option>
 * &lt;/oj-checkboxset>
 */
oj.__registerWidget("oj.ojCheckboxset", $['oj']['editableValue'],
{
  version : "1.0.0",
  defaultElement : "<div>",
  widgetEventPrefix : "oj",
  options :
  {
     /**
     * <p>
     * Disabled <code class="prettyprint">true</code> disables the component and disables
     * all the inputs/labels.
     * Disabled <code class="prettyprint">false</code> enables the component, and leaves the inputs
     * disabled state as it is in the dom.
     * <p>
     *
     * @example <caption>Initialize component with <code class="prettyprint">disabled</code> attribute:</caption>
     * &lt;oj-checkboxset disabled>
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-checkboxset>
     *
     * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
     * // getter
     * var disabled = myComp.disabled;
     *
     * // setter
     * myComp.disabled = false;
     *
     * @expose
     * @type {boolean}
     * @default false
     * @public
     * @instance
     * @memberof oj.ojCheckboxset
     */
    disabled: false,
     /**
     * labelled-by is used to establish a relationship between this component and another element.
     * A common use is to tie the oj-label and the oj-checkboxset together.
     * The oj-label custom element has an id, and you use the labelled-by attribute
     * to tie the two components together.
     *
     * @example <caption>Initialize component with <code class="prettyprint">labelled-by</code> attribute:</caption>
     * &lt;oj-label id="labelId">Name:&lt;/oj-label>
     * &lt;oj-checkboxset labelled-by="labelId">
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-checkboxset>
     *
     * @example <caption>Get or set the <code class="prettyprint">labelledBy</code> property after initialization:</caption>
     * // getter
     * var disabled = myComp.labelledBy;
     *
     * // setter
     * myComp.labelledBy = "labelId";
     *
     * @expose
     * @type {string|null}
     * @public
     * @instance
     * @memberof oj.ojCheckboxset
     */
    labelledBy: null,  
    /** 
     * Whether the component is required or optional. When required is set to true, an implicit 
     * required validator is created using the validator factory - 
     * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED).createValidator()</code>.
     * 
     * Translations specified using the <code class="prettyprint">translations.required</code> attribute 
     * and the label associated with the component, are passed through to the options parameter of the 
     * createValidator method. 
     * 
     * <p>
     * When <code class="prettyprint">required</code> property changes due to programmatic intervention, 
     * the component may clears message and run validation, based on the current state it's in. </br>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when required is set to true, then it runs deferred validation on 
     * the value property. This is to ensure errors are not flagged unnecessarily.
     * </li>
     * <li>if component is invalid and has deferred messages when required is set to false, then 
     * component messages are cleared but no deferred validation is run.
     * </li>
     * <li>if component is invalid and currently showing invalid messages when required is set, then 
     * component messages are cleared and normal validation is run using the current display value. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   property is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   property is updated; page author can listen to the <code class="prettyprint">onValueChanged</code> 
     *   event to clear custom errors.</li>
     * </ul>
     * </li>
     * </ul>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the component are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * 
     * </p>
     * 
     * @ojvalue {boolean} false - implies a value is not required to be provided by the user. 
     * This is the default.
     * @ojvalue {boolean} true - implies a value is required to be provided by user and the 
     * input's label will render a required icon. Additionally a required validator - 
     * {@link oj.RequiredValidator} - is implicitly used if no explicit required validator is set. 
     * An explicit required validator can be set by page authors using the validators option. 
     * 
     * when the option is not set, the element's required property is used as its initial 
     * value if it exists.
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">required</code> attribute:</caption>
     * &lt;oj-checkboxset required>
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-checkboxset>
     * 
     * @example <caption>Customize messages and hints used by implicit required validator when 
     * <code class="prettyprint">required</code> option is set:</caption> 
     * &lt;oj-checkboxset required value="{{colors}}"
     *                    translations='{"required":
     *                                    {"hint": "custom: check at least one value",
     *                                     "messageSummary": "custom: \'{label}\' is Required",
     *                                     "messageDetail", "custom: please check at least one value for \'{label}\'"}}'>
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-checkboxset>
     *
     * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
     * // getter
     * var required = myComp.required;
     *
     * // setter
     * myComp.required = false;
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof oj.ojCheckboxset
     * @type {boolean}
     * @default false
     * @since 0.7
     * @see #translations
     */
    required: false,    
    /**
     * The value of the component.
     *
     * <p>
     * When <code class="prettyprint">value</code> property changes due to programmatic
     * intervention, the component always clears all messages and runs deferred validation, and
     * always refreshes UI display value.</br>
     * 
     * When the option is not set, value is set to an array built from the checked checkboxes, if
     * any are checked.
     *
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>All messages are cleared. This includes
     * the <code class="prettyprint">messagesCustom</code> property.</li>
     * </ul>
     *
     *
     * <h4>Running Validation</h4>
     * <ul>
     * <li>component always runs deferred validation</li>
     * </ul>
     * </p>
     *
     * @example <caption>Initialize the component with the <code class="prettyprint">value</code> attribute specified:</caption>
     * &lt;oj-checkboxset required value='["coffee"]'>
     *   &lt;oj-option value="coffee">Coffee&lt;/oj-option>
     *   &lt;oj-option value="tea">Tea&lt;/oj-option>
     * &lt;/oj-checkboxset>
     * @example <caption>Get or set <code class="prettyprint">value</code> option, after initialization:</caption>
     * // Getter: returns ["coffee"]
     * var val = myComp.value;
     * // Setter: sets ["coffee", "tea"]
     * myComp.value = ["coffee", "tea"];
     *
     * @expose
     * @access public
     * @instance
     * @memberof oj.ojCheckboxset
     * @type {Array.<*>}
     */
    value: []
  },
  /**** start Public APIs ****/

   /**
   * Refreshes the checkboxset
   * <p>A <code class="prettyprint">refresh()</code> or re-init is required
   * when a checkboxset is programatically changed, like in the following circumstances:
   * <ul>
   *   <li>After oj-options are added or removed.</li>
   * </ul>
   * 
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   * myComp.refresh();
   * 
   * @expose
   * @public
   * @return {void}
   * @memberof oj.ojCheckboxset
   * @instance
   */
  refresh: function()
  {
    this._super();
    this._setup();
  },
  /**
   * Returns a jQuery object containing the element visually representing the checkboxset.
   *
   * <p>This method does not accept any arguments.
   *
   * @expose
   * @memberof oj.ojCheckboxset
   * @instance
   * @public
   * @return {jQuery} the checkbox
   * @ignore
  */
  widget : function ()
  {
    return this.uiCheckboxset;
  },
/**
   * Validates the component's display value using all validators registered on 
   * the component and updates the <code class="prettyprint">value</code> option by performing the 
   * following steps. 
   * 
   * <p>
   * <ol> 
   * <li>All messages are cleared, including custom messages added by the app. </li>
   * <li>The implicit 
   * required validator is run if the component is marked required.</li>
   * <li>At the end of validation if there are errors, the messages are shown. 
   * If there were no errors, then the 
   * <code class="prettyprint">value</code> property is updated.</li>
   * </ol>
   * 
   * @example <caption>Validate component using its current value.</caption>
   * myComp.validate();
   * @example <caption>Validate component and use the Promise's resolved state.</caption>
   * myComp.validate().then(
   *  function(result) {
   *    if(result === "valid")
   *    {
   *      submitForm();
   *    }
   *  });
   * @return {Promise.<string>} Promise resolves to "valid" if 
   * the component passed all validations. 
   * The Promise resolves to "invalid" if there were validation errors.
   * 
   * @method
   * @access public
   * @expose
   * @instance
   * @memberof oj.ojCheckboxset
   * @since 4.0.0
   * @ojstatus preview
   */
  validate : oj.EditableValueUtils.validate,

   /**** end Public APIs ****/

  /**** start internal widget functions ****/

  /**
   * Overridden to set the options.value. When constructorOptions value is undefined,
   * we read the CHECKED options on the checkboxes and build the value array from that.
   *
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   */
  _InitOptions : function (originalDefaults, constructorOptions)
  {
    var checkedValues = new Array(), selectedCheckbox, domValue,
        props = [{attribute: "disabled", validateOption: true},
                 {attribute: "title"},
                 //{attribute: "value", "defaultValue": null},  // code below sets value
                 {attribute: "required", coerceDomValue: true, validateOption: true
                 }];

    this._super(originalDefaults, constructorOptions);
    
    
    if (!this._IsCustomElement())
    {
      oj.EditableValueUtils.initializeOptionsFromDom(props, constructorOptions, this);
    }

    // component, app, and constructor are merged into this.options.option by the time _InitOptions
    // is called. Let's take this example:
    // component (widget) default - 'foo'<br/>
    // app default - 'bar'<br/>
    // dom  - 'lucy'<br/>
    // constructorOptions['value'] - undefined<br/>
    // this.options.option is set to 'bar' initially. We don't want to just take this value, because
    // we want DOM value to win over the app and components default if DOM is set.
    // Therefore, the component needs to check if the constructorOptions['value'] is
    // undefined and if so, set value option to 'lucy' (the DOM value in this example). <br/>
    //
    // use DOM value if constructorOptions is undefined. if DOM value is undefined, then
    // leave this.options["value"] alone since it's the merged app/widget default at this point.
    if (!this._IsCustomElement())
    {
      if (constructorOptions['value'] === undefined)
      {
        // constructor option for value is undefined. Then we check the dom.
        this.$checkboxes = this._findCheckboxesWithMatchingName();
        selectedCheckbox = this.$checkboxes.filter(":checked");
        if (selectedCheckbox.length > 0)
        {
          selectedCheckbox.each(function(){
            checkedValues.push($(this).val());
          });
          domValue = checkedValues;

          // when defaulting from DOM we want to trigger optionChange to writeback new value

          this.option("value", domValue, {'_context':{writeback: true, internalSet: true}});
        }
        // if nothing is checked, we leave this.options["value"] as it is if not undefined, else
          // the widget's default is [].
          if (this.options["value"] === undefined)
            this.options["value"] = [];

      }
      else
      {
        this._checkValueType(this.options["value"]);
      }
    }
    else
      this._checkValueType(this.options["value"]);
  },
  /**
   * After _ComponentCreate and _AfterCreate,
   * the widget should be 100% set up. this._super should be called first.
   * @override
   * @protected
   * @memberof oj.ojCheckboxset
   * @instance
   */
  _ComponentCreate : function ()
  {
    var element = this.element;
    this._super();
    // first check to see if element is NOT a fieldset. If fieldset, throw error.
    if (element.is("fieldset"))
    {
      throw new Error("ojCheckboxset cannot be bound to a fieldset. Use a div instead.");
    }
    
    // since the oj-option renderer uses the oj-checkboxset ID for the name
    // attribute of the rendered chekcboxes, let's make sure the checkboxset
    // has an ID
    element.uniqueId();
    this._processOjOptions();
    // The processOjOptions renders input/label, so we need to go through and 
    // get the this.$checkboxes after this is called.
    this.$checkboxes = this._findCheckboxesWithMatchingName();
    // Turn each checkbox into ojCheckbox. Do this first, since we need it
    // in calls from 'create'. Also, since ojCheckboxSet delegates to the _ojRadioCheckbox
    // component, and we need to mark this as an internal node so that oj.Components.getComponentElementByNode
    // knows it is an internal component in this case, not a stand-alone component
    this.$checkboxes._ojRadioCheckbox().attr('data-oj-internal', '');
 
    // keep the root dom element and slots as is, and add a wrapper dom underneath it. This way we can
    // have one div around all the inputs and labels, and for inline messaging we can have another
    // div around the inline messaging content. And we can style the borders of the two boxes differently.
    this.uiCheckboxset = element.addClass("oj-checkboxset oj-component").attr( "role", "group" );
    // need to grab all elements using childNodes first and then do filter because jquery children will automatically 
    // exclude all comment and text nodes
    $(element.get(0).childNodes).filter(function(){
        return (this.getAttribute && this.getAttribute("slot") === 'contextMenu') ? false : true;
    }).wrapAll("<div class='oj-checkboxset-wrapper'></div>"); //@HTMLUpdateOK

    this._on(this._events);
    this._setup();
    
  },
    /**
   * Resets this.checkboxes. This is called at the beginning of a refresh in EditableValue
   * @override
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   */
  _ResetComponentState : function ()
  {
    // we could have added, removed, or modified radios, so we need to re-find all the
    // oj-options, inputs on refresh and turn the ones that aren't already 
    // ojRadioCheckboxes into them.
    this._processOjOptions();
    this.$checkboxes = this._findCheckboxesWithMatchingName();

     // we have a rule for refresh: if we have a public API for it, then the app dev has to use the
    // option, and not expect changing the dom will update the state with refresh.
    // However, ojCheckboxset does not expose a public API for the individual checkbox's disabled state
    // to the app developer. Our private ojRadioCheckbox component has a disabled option that
    // our code has access to.
    // For each checkbox, we need to look at the disabled attribute dom and update the
    // ojradiocheckbox's disabled option.

    // !! ensures it is a boolean
    // update the private ojradiocheckbox component's disabled option to keep it in sync with the dom
    this.$checkboxes.filter( ".oj-checkbox" ).each(function()
    {
        var disabledValue = $( this ).attr("disabled") !== undefined ?
          !!$( this ).prop("disabled") : false;
        $( this )._ojRadioCheckbox("option", "disabled", disabledValue);
    });

    // no need to refresh the ojRadioCheckbox's that exist since we have options for everything.
    // of the type=radio inputs that are not yet ojRadioCheckboxs, make them ojRadioCheckboxes.

    // create ojRadioCheckboxes on any new ones.
    this.$checkboxes.not(".oj-checkbox")._ojRadioCheckbox();
    
  },

  /**
   * @override
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   * @since 5.0.0
   */
  GetFocusElement : function ()
  {
    // We need :focusable here so that we don't try to focus on an element that isn't focusable.
    return this._GetContentElement().filter(":focusable").first()[0];
  },
  /**
   * Sets the disabled option onto the dom.
   * This is a no-op for checkboxset since its root dom element is a div, and disabled is
   * invalid on a div. If we did try to set disabled on the div, then restore attributes doesn't
   * work correctly since it wasn't saved correctly.
   * @param {Object} node - dom node
   *
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   * @since 1.0.0
   */
  _SetDisabledDom : function(node)
  {
    // no-op
  },
  /**
   * Whether the component is required.
   * 
   * @return {boolean} true if required; false
   * 
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   * @override
   */
  _IsRequired : function () 
  {
    return this.options['required'];
  },
  /**
   * @memberof oj.ojCheckboxset
   * @instance
   * @private
   */
  _refreshRequired : oj.EditableValueUtils._refreshRequired,
  
  /**
   * the validate method from v3.x that returns a boolean
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   * @ignore
   */
  _ValidateReturnBoolean: oj.EditableValueUtils._ValidateReturnBoolean,

  /** 
   * This function processes the oj-option children, sets the custom renderer, and 
   * creates input type=checkbox and label dom from them.
   * 
   * We don't want to rely on the framework calling the customOptionRenderer 
   * as a result of setting options[i]["customOptionRenderer"] = renderer; in this function. 
   * This could lead to timing bugs when data-oj-binding-provider="none". (when not "none",
   * we know the oj-options are created before the oj-checkboxset gets created, so no timing issue)
   * Therefore we create the input/label not in the customOptionRenderer, 
   * but in a separate function that we call.
   * @private
   * @instance
   */
  _processOjOptions : function ()
  {
    if (this._IsCustomElement())
    {
      // set the custom renderer on oj-option
      var i, len;
      var renderer = this._customOptionRenderer.bind(this)
      var options = this.element.children("oj-option");
      if (options.length > 0) 
      {
        for (i = 0, len = options.length; i < len; i++) {
          options[i]["customOptionRenderer"] = renderer;
          this._initInputLabelFromOjOption(options[i]);
        }
      }

      // for oj-option that are inside the wrapper (after refresh)
      options = this.element.children(".oj-checkboxset-wrapper").find("oj-option");
      if (options.length > 0) 
      {
        for (i = 0, len = options.length; i < len; i++) {
          options[i]["customOptionRenderer"] = renderer;
          this._initInputLabelFromOjOption(options[i]);
        }
      }
    }        
  },
  
   /**
   * Create the input type='checkbox'/label dom from attributes on oj-option element.
   * oj-checkboxset is made up of input/labels.
   * This gets called during the oj-checkboxset _CreateComponent and refresh
   * @param {Element} elem the oj-option element
   * @private
   * @instance
   */
  _initInputLabelFromOjOption : function (elem)
  {
    var span;
    var label;
    var ojoption = elem;

    // let's make sure that each oj-option has an ID so the
    // label element can reference the input element via the 'for' attribute
    // we have tests in place where oj-option doesn't have id and where it does
    // both for the databound case and non-databound case. In the databound case, the 
    // bindings are resolved before we get here, so we will be fine.
    $(ojoption).uniqueId();
    
    var id = ojoption.getAttribute("id");
    var checkboxId = id+"|cb";
    var checkbox = document.getElementById(checkboxId);
    var alreadyProcessed = (checkbox !== null); // Was the oj-option already processed
    
    // if the oj-option is already processed, we don't need to add the additional dom
    // in the code below, we use setAttribute() for everything as we want to be
    // setting the initial value for these elements.
    if (!alreadyProcessed)
    {
      span = document.createElement("span");
      checkbox = document.createElement("input");
      label = document.createElement("label");
      span.setAttribute("class", "oj-choice-item");
      checkbox.setAttribute("type", "checkbox");
      checkbox.setAttribute("id", checkboxId);
      label.setAttribute("for", checkboxId);
      
      // if the ojoption doesn't have any textContent, hide the label element.  The use case for
      // this is the case where you have an oj-checkboxset with one checkbox with no label (i.e. you 
      // specify one oj-option with no textContent).  We want to hide the generated label element so
      // that the checkbox is easier to center in the parent container (such as a table cell).
      // Note: this is not an issue for oj-radioset, as there is no use case that uses a single
      // radio button in an oj-radioset.
      if (!ojoption.textContent || ojoption.textContent === "")
        label.classList.add("oj-helper-hidden");
      
      ojoption.parentElement.insertBefore(span, ojoption);//@HTMLUpdateOK
      label.appendChild(ojoption); // append the oj-option as a child of label
      span.appendChild(checkbox);
      span.appendChild(label);
    }
    else
    {
      // find the parent label element.  This is the element we need to hide if there is no text
      // content in the oj-option
      label = ojoption;
      
      while ((label= label.parentElement) && !(label.tagName === "LABEL"));
      // if the ojoption doesn't have any textContent, hide the label element by adding the
      // class oj-helper-hidden
      // if for some reason, a label element is not found, don't do anything
      if (label)
      {
        if (!ojoption.textContent || ojoption.textContent === "")
          label.classList.add("oj-helper-hidden");
        else
          label.classList.remove("oj-helper-hidden");
      }
    }

    var name = this.element[0].id; // Use the id of the ojcheckboxset as the name of the oj-options.
    var ariaLabel = ojoption.getAttribute("aria-label");
    var ariaLabelledBy = ojoption.getAttribute("aria-labelledby");

    // The value attribute of the checkbox only supports text, so we need to be
    // able to access the oj-option's value property instead.  This attribute
    // is a link back to the oj-option so that we don't need to use dom traversal
    // to get to the oj-option to get its value.
    checkbox.setAttribute("data-oj-option-id", id);

    if (name && name != "")
      checkbox.setAttribute("name", name);
    else
      checkbox.removeAttribute("name");

    if (ariaLabel && ariaLabel != "")
      checkbox.setAttribute("aria-label", ariaLabel);
    else
      checkbox.removeAttribute("aria-label");

    if (ariaLabelledBy && ariaLabelledBy != "")
      checkbox.setAttribute("aria-labelledby", ariaLabelledBy);
    else
      checkbox.removeAttribute("aria-labelledby");

    if (ojoption.disabled)
      checkbox.setAttribute("disabled", true);
    else
      checkbox.removeAttribute("disabled");
  },  
  
  // custom oj-option renderer
  // Because we can't rely on this being called when we set the customOptionRenderer property
  // in _processOjOptions we shouldn't do the input.label creation from the oj-option 
  // in this function. (If we did, the _ComponentCreate code that relies on the
  // inputs/labels being created already would not work.)
  // The correct thing to do is to create the input/label in _initInputLabelFromOjOption
  // Then rely on this function being called after the oj-option has been created and we are
  // changing properties on it.
  _customOptionRenderer : function (elem)
  {
    var ojoption = elem;
    var id = ojoption.getAttribute("id");
    var checkboxId = id+"|cb";
    var checkbox = document.getElementById(checkboxId);
    // Was the oj-option already rendered into an _ojRadioCheckbox() in _CreateComponent?
    var checkboxExists = (checkbox !== null); 
    var hasOjCheckboxClass = checkboxExists && checkbox.classList.contains('oj-checkbox');

    
    // When an oj-option child is disabled (by setting the disabled attribute to
    // true) and it re-renders, the component should refresh automatically rather than requiring the
    // user to call refresh. See .   
    if (hasOjCheckboxClass)
    {
      $(checkbox)._ojRadioCheckbox("option", "disabled", ojoption.disabled);
    }
  },
  /**
   * If custom element, get the labelledBy option, and set this 
   * onto the root dom element as aria-labelledby. We append "|label" so it matches the id that
   * is on the oj-label's label element.
   * @memberof oj.ojCheckboxset
   * @instance
   * @private
   */
  _updateLabelledBy : oj.EditableValueUtils._updateLabelledBy,
  /**
   * Returns a jquery object that is a set of elements that are input type checkbox
   * and have the name of the first checkbox found.
   *
   * @return {jQuery} jquery object of all the checkboxes within the root dom element
   * that have the same 'name' attribute as the first checkbox found.
   * @private
   * @memberof oj.ojCheckboxset
   */
  _findCheckboxesWithMatchingName : function ()

  {
    var allcheckboxes;
    var $first = this.element.find("input[type=checkbox]:first");
    var name;
    var selector;
    
    if ($first.length === 0)
    {
      oj.Logger.warn("Could not find any input type=checkbox within this element");
    }
    // get the name attribute of the first input checkbox
    name = $first.attr("name");
    // find all input checkboxes with matching name
    if (name === undefined)
    {
      // search for all checkboxes with no name
      allcheckboxes = this.element.find("input[type=checkbox]");
      // now loop and find the ones without 'name' attribute
      return allcheckboxes.not("[name]");
    }
    else
    {
      // search for all checkboxes with the name
      selector = "input[type=checkbox][name=" + name + "]";
      return this.element.find(selector);

    }
  },

  // Override to set custom launcher
  _NotifyContextMenuGesture: function(menu, event, eventType)
  {
    // Setting the launcher to the first tabbable checkbox in the set.
    // Component owner should feel free to specify a different launcher if appropriate,
    // e.g. could specify the "current" checkbox rather than the first if desired.
    // See the superclass JSDoc for _OpenContextMenu for tips on choosing a launcher.
    var launcher = this.element.find("input[type=checkbox]:tabbable").first();
    this._OpenContextMenu(event, eventType, {"launcher": launcher});
  },
  // Override to set launcher to widget
  _GetMessagingLauncherElement : function ()
  {
    // focus events only get triggered on input, but they do bubble up and we will capture them
    // on the widget.
    // mouseenter events get called once once if the user hovers over for the entire widget. if
    // we put it on the inputs, it gets called every time you leave and enter a new input. Plus,
    // this doesn't work when we hide the input like we do in the native themes.
     return this.widget();                          
  },
  /**
   * _setup is called on create and refresh. Use the disabled option to
   * update the component. If the component's option is disabled, then
   * leave it alone.
   * @memberof oj.ojCheckboxset
   * @instance
   * @private
   */
  _setup: function()
  {
     // at this point we already have this.$checkboxes set to a list of checkboxes for this
     // checkboxset
    this._propagateDisabled(this.options.disabled );

    if (this.$checkboxes !== null)
    {
      if (this.$checkboxes.length === 1)
      {
        this.element.addClass("oj-checkboxset-single");
      }
      else
      {
        this.element.removeClass("oj-checkboxset-single");
      }
    }
    
    // add to the root dom the style class 'oj-choice-direction-column' 
    // if there isn't already a 'oj-choice-direction-row' or 'oj-choice-direction-column' there.
    if (!(this.element.hasClass("oj-choice-direction-column")) && 
        !(this.element.hasClass("oj-choice-direction-row")))
    {
      this.element.addClass("oj-choice-direction-column");
    }
    this._refreshRequired(this.options['required']);
    
    this._updateLabelledBy(null, this.options['labelledBy'], this.widget());
  },
  _events :
  {
    'change' : function (event)
    {

      this._HandleChangeEvent(event);
    }
  },
   /**
   * If value is undefined, set it to the default value, which is [].
   * Else, confirm it is an Array and throw an error if it isn't.
   * @memberof oj.ojCheckboxset
   * @instance
   * @private
   * @param {Object|null|undefined} value the value to check
   * @throws new Error if value (undefined is converted to an []) isn't an Array 
   */
  _checkValueType: function(value)
  {
    // if value is undefined, set it to its default value [].
    if (value === undefined)
    {

      this.option("value", [], {'_context':{writeback: true, internalSet: true}});
    }
    else
    {
      this._confirmValueIsArray(value);
    }

  }, 
  /**
   * Confirm value is an Array.
   * @memberof oj.ojCheckboxset
   * @instance
   * @private
   * @param {Object|null|undefined} value the value to check
   * @throws new Error if value isn't an Array
   */
  _confirmValueIsArray: function(value)
  {
    if (!Array.isArray(value))
    {
      throw new Error("Invalid 'value' set on JET Checkboxset: " + value + ".It must be an Array. ");
    }
  },
  /**
   * @param {Event} event DOM event
   * @override
   * @protected
   * @memberof oj.ojCheckboxset
   */
  _HandleChangeEvent: function(event)
  {
    var submittedValue;
    var checkboxes;
    
    // keep oj-selected in sync with the input element's checked state
    checkboxes = this.$checkboxes;
    if (checkboxes.length > 0)
    {
      checkboxes.each(function(){
        if (this === event.target) {
          // the target is one of the checkboxes. Update the oj-selected class to keep it 
          // in sync with the input's HTML checked attribute
          $(this)._ojRadioCheckbox("setSelectedClass", event.target.checked);
        }
      });
    }
    // run full validation. There is no need to check if values have changed
    // since for checkboxset/radioset if we get into this function we know value has changed.
    // passing in doValueChangeCheck: false will skip the new-old value comparison
    submittedValue = this._GetDisplayValue();
    this._SetValue(submittedValue, event, _sValueChangeCheckFalse);
  },

  /**
   * Returns the display value that is ready to be passed to the converter.
   *
   * @param {Object} value the stored value if available that needs to be formatted for display
   * @override
   * @protected
   * @memberof oj.ojCheckboxset
   */
  _GetDisplayValue : function (value)
  {
    // return the value of the 'checked' checkboxes
    return this._GetElementValue();
  },
  /**
   * Called when the display value on the element needs to be updated
   * as a result of a value change.
   * ojCheckboxset stores an Array value, and this value matches the values
   * of the currently checked checkboxes. So, if we need to set the display value,
   * what this means is we need to 'check' the checkboxes whose values match the
   * displayValue and 'uncheck' those that don't.
   *
   * @param {Array} displayValueArray an Array of values that need to be checked, e.g., ["red","blue"].
   *  Any checkbox in the checkboxset that doesn't match one of the checkBoxes needs to be unchecked
   *  if it isn't already.
   * @override
   * @protected
   * @throws new Error if displayValueArray is not an Array
   * @memberof oj.ojCheckboxset
  */
 _SetDisplayValue : function (displayValueArray)
  {
    var i;
    var length = this.$checkboxes.length;
    var optionValue;
    var checkbox;
    var $checkbox;
    
    this._checkValueType(displayValueArray);
      
    // If it is empty or not an array, then set all the _ojRadioCheckboxes's checked option to false.
    // _GetDisplayValue gets the checked options and creates an Array from it.
    if (displayValueArray === null || displayValueArray === undefined || 
        displayValueArray.length === 0)
    {
      this.$checkboxes._ojRadioCheckbox("option", "checked", false);
    }
    else
    {
      // go through each _ojRadioCheckbox and see if it needs to be checked or unchecked.
      for (i = 0; i < length; i++)
      {
        checkbox = this.$checkboxes[i];
        $checkbox = $(checkbox);
        optionValue = this._GetOptionValue(checkbox);
        // does the checkbox's value exist in the checkedBoxes array?
        var index = this._GetOptionIndex(displayValueArray, optionValue);
        var checked = $checkbox._ojRadioCheckbox("option", "checked"); 
        
        if (index !== -1)
        {
          // yes. this needs to be checked, if it isn't already
          if (!checked)
          {
            $checkbox._ojRadioCheckbox("option", "checked", true);
          }
        }
        else
        {
          /// no. this needs to be unchecked, if it isn't already
          if (checked)
          {
            $checkbox._ojRadioCheckbox("option", "checked", false);
          }
        }
      }
    }
  },
  /**
   * Returns the element's value. Normally, this is a call to this.element.val(),
   * but in the case of ojCheckboxset, the element's value is really the value
   * of the checked checkboxes in the set.
   * @override
   * @protected
   * @memberof oj.ojCheckboxset
   * @return {Array} An Array<any> of selected values or the empty array [] if nothing selected.
   */
  _GetElementValue : function ()
  {
    // "input:checked" selects checkboxes that are currently checked as
    // reflected in their boolean (true or false) checked property,
    // which is affected when the user clicks the checkbox for example.
    // for checkboxset, there will be zero or more checked;
    var self = this;
    var checkedValues=new Array();
    var selectedCheckboxes = this.$checkboxes.filter(":checked");
    if (selectedCheckboxes.length === 0)
      return [];
    else
    {
      selectedCheckboxes.each(function(){checkedValues.push(self._GetOptionValue(this))});
      return checkedValues;
    }
  },

  /**
   * Returns the index of the array that matches the optionValue. First check
   * equality with ===, then if not found, check with deep compare,
   * oj.Object.compareValues(), because values can be objects or arrays.
   * @override
   * @protected
   * @memberof oj.ojCheckboxset
   */
  _GetOptionIndex : function (optionValueArray, optionValue)
  {
    // Find the matching value via === comparison that indexOf uses
    var matchIndex = optionValueArray.indexOf(optionValue);
    var length;

    // If not found via indexOf(), do a deep equals compare
    if (matchIndex == -1)
    {
      length = optionValueArray.length;
      for (var i = 0; i < length; i++)
      {
        if (oj.Object.compareValues(optionValueArray[i], optionValue))
        {
          matchIndex = i;
          break;
        }
      }
    }
    
    return matchIndex;
  },
  /**
   * For custom element, we get the value from the oj-option element, otherwise
   * we get the value from the checkbox element.
   * @override
   * @protected
   * @memberof oj.ojCheckboxset
   * @return {*} Returns the value property of the associated oj-option, or the value attribute of the checkbox element.
   */
  _GetOptionValue : function (checkboxElem)
  {
    var option;
    var val;
    
    if (this._IsCustomElement())
    {
      option = document.getElementById(checkboxElem.getAttribute("data-oj-option-id"));

      if (option)
      {
        val = option.value;
      }
    }
    else
    {
      val = checkboxElem.value;
    }
    
    return val;
  },

  /**
   * Returns the default styleclass for the component. Currently this is
   * used to pass to the ojLabel component, which will append -label and
   * add the style class onto the label. This way we can style the label
   * specific to the input component. For example, for inline labels, the
   * checkboxset/checkboxset components need to have margin-top:0, whereas all the
   * other inputs need it to be .5em. So we'll have a special margin-top style
   * for .oj-label-inline.oj-checkboxset-label
   * All input components must override
   *
   * @return {string}
   * @memberof oj.ojCheckboxset
   * @override
   * @protected
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-checkboxset";
  },
  /**
   * Returns a jquery object of the elements representing the content nodes (checkboxes/labels).
   * @protected
   * @override
   * @memberof oj.ojCheckboxset
   */
  _GetContentElement : function ()
  {
    if (this.$checkboxes != null)
      return this.$checkboxes;
    else
    {
      this.$checkboxes = this._findCheckboxesWithMatchingName();
      return this.$checkboxes;
    }
  },

  /**
   * Called to find out if aria-required is unsupported. This is needed for the label.
   * It is not legal to have aria-required on radio/checkboxes, nor on
   * radiogroup/group.
   * If aria-required is not supported, then we wrap the required icon as well as the
   * help icons so that JAWS can read required. We don't do this for form controls that use
   * aria-required because if we did JAWS would read required twice.
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   * @return {boolean}
   */
  _AriaRequiredUnsupported : function()
  {
    return true;
  },
 
  /**
   * Performs post processing after required option is set by taking the following steps.
   * 
   * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
   * run full validation with UI value (we don't know if the UI error is from a required validator 
   * or something else);<br/>
   * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
   * updated<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to optionChange(value) to clear custom errors.<br/>
   * 
   * - if component is invalid and has messagesHidden -> required: false -> clear component 
   * errors; no deferred validation is run.<br/>
   * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
   * errors unnecessarily)<br/>
   * - messagesCustom is never cleared<br/>
   * 
   * @param {string} option
   * 
   * @memberof oj.ojCheckboxset
   * @instance
   * @protected
   */
  _AfterSetOptionRequired : oj.EditableValueUtils._AfterSetOptionRequired,

  /**
   * 
   * @param {boolean} disabled
   * @private
   * @memberof oj.ojCheckboxset
   */
  _propagateDisabled: function( disabled ) {
      disabled = !!disabled;
      this.$checkboxes.each(function()
      {
        // this is the technique to use to call package-private functions
        // Calling it like this.$radios.ojRadioCheckbox("__setAncestorComponentDisabled",value)
        // gives an error because jquery prevents you from calling functions with an "_"
        //
        // This is how we handle 'disabled' for the checkboxset. We don't change the radiocheckbox
        // component's disabled option ever since if we do that we've lost what the initial disabled
        // state is (we store the disabled dom value from the radio into its disabled option)
        // and we need that when we refresh. Instead what we do
        // is we mark if its ancestor (the checkboxset) is disabled or not. Then, when we render
        // out the checkboxes 'disabled' state, like oj-disabled, we look to see if it is 'effectively
        // disabled' (see _IsEffectivelyDisabled call in ojRadioCheckbox), that is if its
        // option is disabled OR its ancestor (the checkboxset) is disabled.
        $( this ).data("oj-_ojRadioCheckbox").__setAncestorComponentDisabled(disabled);
      });

      this.$checkboxes._ojRadioCheckbox("refreshDisabled"); // re-render disabled
  },
  /**
   * @override
   * @private
   * @memberof oj.ojCheckboxset
   */
  _setOption : function (key, value, flags)
  {
    var originalValue = this.options["labelledBy"];;
 
    this._super(key, value, flags);
    
    switch (key)
    {
      case "disabled":
        this._propagateDisabled(value);
        break;
      case "labelledBy":
        // remove the old one and add the new one
        this._updateLabelledBy(originalValue, value, this.widget());
        break;
      default:
        break;
    }
  },

  /**
   * Performs post processing after _SetOption() is called. Different options when changed perform
   * different tasks. See _AfterSetOption[OptionName] method for details.
   *
   * @param {string} option
   * @param {Object|string=} previous
   * @param {Object=} flags
   * @protected
   * @memberof oj.ojCheckboxset
   * @instance
   * @override
   */
  _AfterSetOption : function (option, previous, flags)
  {
    this._superApply(arguments);
    switch (option)
    {
      case "required":
        this._AfterSetOptionRequired(option);
        break;
      default:
        break;
    }
  },

  //** @inheritdoc */  
  getNodeBySubId: function(locator)
  {
    var node = this._super(locator),
        checkboxes,
        subId,
        value;

    if (!node)
    {
      checkboxes = this.$checkboxes.get();
      subId = locator['subId'];
      
      switch (subId)
      {
        case 'oj-checkboxset-inputs': // TODO: deprecated for a while now, remove this in 4.0.0
          node = checkboxes;
          break;

        case 'oj-checkboxset-checkbox':
          // We find these by the value attribute on the input element, which is
          // much more stable that using an index.
          value = locator['value'];

          if (typeof value != 'undefined')
          {
            var arrayLength = checkboxes.length;
            var i;
            var matchIndex = -1;
            var checkboxValues = [];

            // Build the values array
            for (i = 0; i < arrayLength; i++)
              checkboxValues[i] = this._GetOptionValue(checkboxes[i]);        
            
            // Find the index of the matching value.
            matchIndex = this._GetOptionIndex(checkboxValues, value);
            
            if (matchIndex !== -1)
              node = checkboxes[matchIndex];
          }
          break;
      }
    }
    // Non-null locators have to be handled by the component subclasses
    return node || null;
  },

  //** @inheritdoc */
  getSubIdByNode: function(node) {
    var topElem = this._GetContentElement()[0].parentElement.parentElement.parentElement;
    var currentNode = node;
    
    while (currentNode  && currentNode != topElem) {
      if (currentNode.nodeName == "LABEL")
        currentNode = document.getElementById(currentNode.for);
      
      if (currentNode.nodeName == "INPUT")
      {
        return {"subId":"oj-checkboxset-checkbox", "value":this._GetOptionValue(currentNode)};
      }

      currentNode = currentNode.parentElement;
    }
    
    return this._super(node);
  },

  /**
   * @ignore
   * @protected
   * @memberof oj.ojCheckboxset
   * @override
   */
  _destroy : function ()
  {
    var ret = this._super();
    var wrapperDom = this.element[0].firstElementChild;
    
    if (this.$checkboxes)
    {
      this.$checkboxes._ojRadioCheckbox( "destroy" );
    }
    
    // remove the dom we added to wrap the children of this.element, but don't remove the children.
    $(wrapperDom).contents().unwrap();
    
    return ret;
  }
  /**** end internal widget functions ****/
   /**
   * Removes the checkboxset functionality completely.
   * This will return the element back to its pre-init state.
   *
   * <p>This method does not accept any arguments.
   *
   * @method
   * @name oj.ojCheckboxset#destroy
   * @memberof oj.ojCheckboxset
   * @instance
   * @ignore
   *
   * @example <caption>Invoke the <code class="prettyprint">destroy</code> method:</caption>
   * $( ".selector" ).ojCheckboxset( "destroy" );
   */

   // Fragments:

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Gesture</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *    <tr>
   *       <td>Checkbox</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td> Select/unselect the input</td>
   *     </tr>
   *     <tr>
   *       <td>Input's Label</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td> Select/unselect the corresponding input</td>
   *    </tr>
   *     <tr>
   *       <td>Input or Label</td>
   *       <td><kbd>Press & Hold</kbd></td>
   *       <td>If hints, help.instruction or messages exist in a notewindow, 
     *           pop up the notewindow.</td>
   *    </tr>
   *   </tbody>
   *  </table>
   *
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojCheckboxset
   */

   /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>Checkboxset</td>
   *       <td><kbd>Tab In</kbd></td>
   *       <td>Set focus to the first item in the checkboxset. 
   *       If hints, helpInstruction or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr> 
   *   </tbody>
   * </table>
   *
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojCheckboxset
   */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     * 
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocExampleHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-choice-direction-column</td>
     *       <td>This is the default. It lays out the checkboxes in a column.
     *       </td>
     *       <td>
     * <pre class="prettyprint">
     * <code>&lt;oj-checkboxset id="checkboxsetId" 
     *  class="oj-checkboxset-direction-column">
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-choice-direction-row</td>
     *       <td>It lays out the checkboxes in a row.
     *       </td>
     *       <td>
     * <pre class="prettyprint">
     * <code>&lt;oj-checkboxset id="checkboxsetId" 
     *   class="oj-choice-direction-row">
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-checkboxset-no-chrome</td>
     *       <td>Use this styleclass if you don't want the chrome around the set.
     *       </td>
     *       <td>
     * <pre class="prettyprint">
     * <code>&lt;oj-checkboxset id="checkboxsetId"
     *  class="oj-checkboxset-no-chrome">
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-checkboxset-input-start</td>
     *       <td>Use this styleclass to order the checkbox at the start and label text at the end 
     *       even if a theme has a different default order.
     *       </td>
     *       <td>
     * <pre class="prettyprint">
     * <code>&lt;oj-checkboxset id="checkboxsetId"
     *  class="oj-checkboxset-input-start">
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-checkboxset-input-end</td>
     *       <td>Use this styleclass to order the checkbox at the end and the label text at the start 
     *       even if a theme has a different default order.
     *       </td>
     *       <td>
     * <pre class="prettyprint">
     * <code>&lt;oj-checkboxsetid="checkboxsetId" 
     *  class="oj-checkboxset-input-end" >
     * </code></pre>
     *       </td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojCheckboxset
     */
});


//////////////////     SUB-IDS     //////////////////


/**
 * <p>Sub-ID for the checkboxset's checkboxes.
 *
 * @ojsubid oj-checkboxset-inputs
 * @deprecated 3.0.0 Since the application supplies the input elements, it can supply a unique ID by which the input elements can be accessed.
 * @ignore
 * @memberof oj.ojCheckboxset
 * @example <caption>Get the nodes for the checkboxes:</caption>
 * var nodes = myComp.getNodeBySubId('oj-checkboxset-inputs');
 */

// -----------------------------------------------------------------------------
// "private static members" shared by all checkboxsets
// -----------------------------------------------------------------------------
/**
 * do not do a value change check in _SetValue
 */
var _sValueChangeCheckFalse = {doValueChangeCheck: false};

}());
(function() {
var ojCheckboxsetMeta = {
  "properties": {
    "disabled": {
      "type": "boolean"
    },
    "labelledBy": {
      "type": "string"
    },
    "required": {
      "type": "boolean"
    },  
    "value": {
      "type": "Array"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        }
      }
    },
  },
  "methods": {
    "validate": {}
  },
  "extension": {
    _WIDGET_NAME: "ojCheckboxset"
  }
};
oj.CustomElementBridge.registerMetadata('oj-checkboxset', 'editableValue', ojCheckboxsetMeta);
oj.CustomElementBridge.register('oj-checkboxset', {'metadata': oj.CustomElementBridge.getMetadata('oj-checkboxset')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * @ojcomponent oj.ojCollapsible
 * @augments oj.baseComponent
 * @since 0.6
 * @ojstatus preview
 * @class oj.ojCollapsible
 * @ojshortdesc Displays a header that can be expanded to show its content.
 *
 * @classdesc
 * <h3 id="collapsibleOverview-section">
 *   JET Collapsible
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#collapsibleOverview-section"></a>
 * </h3>
 *
 * <p>Description: A JET Collapsible displays a header that can be expanded to show additional content beneath it.
The child element of the oj-collapsible in the named <a href="#header">header</a> slot is displayed in the header, while the child element in the <a href="#Default">default</a> slot is displayed as the content
 *
 * <p>Note for performance reasons, if the collapsible content is expensive to render, you should wrap it in an <code class="prettyprint">oj-defer</code> element (API doc {@link oj.ojDefer}) to defer the rendering of that content.<br/>
 * See the <a href="../jetCookbook.html?component=collapsible&demo=deferredRendering">Collapsible - Deferred Rendering</a> demo for an example.</p>
 *
 * <pre class="prettyprint"><code>
 * &lt;oj-collapsible>
 *   &lt;h3 slot='header'>Header 1&lt;/h3>
 *   &lt;p>Content 1&lt;/p>
 * &lt;/oj-collapsible>
 * </code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>In the unusual case that the directionality (LTR or RTL) changes post-init, the collapsible must be <code class="prettyprint">refresh()</code>ed.
 *
 */
(function () {
  var uid = 0;
  var OPEN_ICON = 'oj-collapsible-open-icon';
  var CLOSE_ICON = 'oj-collapsible-close-icon';

  oj.__registerWidget('oj.ojCollapsible', $.oj.baseComponent,
    {
      widgetEventPrefix: 'oj',
      options:
      {
      /**
       * Specifies if the content is expanded.
       *
       * @expose
       * @memberof oj.ojCollapsible
       * @instance
       * @ojshortdesc Specifies if the content is expanded.
       * @type {boolean}
       * @default false
       * @ojwriteback
       *
       * @example <caption>Initialize the collapsible with the <code class="prettyprint">expanded</code> attribute specified:</caption>
       * &lt;oj-collapsible expanded='true'>&lt;/oj-collapsible>
       *
       * @example <caption>Get or set the <code class="prettyprint">expanded</code> property after initialization:</caption>
       * // getter
       * var expandedValue = myCollapsible.expanded;
       *
       * // setter
       * myCollapsible.expanded = false;
       */
        expanded: false,

      /**
       * Disables the collapsible if set to <code class="prettyprint">true</code>.
       * @name disabled
       * @memberof oj.ojCollapsible
       * @instance
       * @ojshortdesc Disables the collapsible if set to true.
       * @type {boolean}
       * @default false
       * @example <caption>Initialize the collapsible with the <code class="prettyprint">disabled</code> attribute specified:</caption>
       * &lt;oj-collapsible disabled='true'>&lt;/oj-collapsible>
       *
       * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
       * // getter
       * var disabledValue = myCollapsible.disabled;
       *
       * // setter
       * myCollapsible.disabled = false;
       */
        disabled: false,

      /**
       * The type of event to expand/collapse the collapsible.
       * To expand the collapsible on hover, use "mouseover".
       *
       * @ignore
       * @ojtsignore
       * @expose
       * @memberof oj.ojCollapsible
       * @instance
       * @type {string}
       * @default "click"
       */
        expandOn: 'click',

      /**
       * Where in the header to click to toggle disclosure.
       *
       * @expose
       * @memberof oj.ojCollapsible
       * @instance
       * @ojshortdesc Where in the header to click to toggle disclosure.
       * @type {string}
       * @ojvalue {string} "header" click any where in the header to toggle disclosure
       * @ojvalue {string} "disclosureIcon" click the disclosureIcon to toggle disclosure
       * @default "header"
       *
       * @example <caption>Initialize the collapsible with the <code class="prettyprint">expand-area</code> attribute specified:</caption>
       * &lt;oj-collapsible expand-area='disclosureIcon'>&lt;/oj-collapsible>
       *
       * @example <caption>Get or set the <code class="prettyprint">expand-area</code> property after initialization:</caption>
       * // getter
       * var expandAreaValue = myCollapsible.expandArea;
       *
       * // setter
       * myCollapsible.expandArea = 'disclosureIcon';
       */
        expandArea: 'header',

      // callbacks
      /**
       * Triggered immediately before the collapsible is expanded.
       * Call <code class="prettyprint">event.preventDefault()</code> in the event listener to veto the event, which prevents the content from expanding.
       *
       * @expose
       * @event
       * @memberof oj.ojCollapsible
       * @instance
       * @ojshortdesc Triggered immediately before the collapsible is expanded.
       * @ojcancelable
       * @property {Element} header The header that is about to be expanded.
       * @property {Element} content The content that is about to be expanded.
       */
        beforeExpand: null,

      /**
       * Triggered after the collapsible has been expanded (after animation completes).
       *
       * @expose
       * @event
       * @memberof oj.ojCollapsible
       * @instance
       * @ojshortdesc Triggered immediately after the collapsible is expanded.
       * @property {Element} header The header that was just expanded.
       * @property {Element} content The content that was just expanded.
       */
        expand: null,

      /**
       * Triggered immediately before the collapsible is collapsed.
       * Call <code class="prettyprint">event.preventDefault()</code> in the event listener to veto the event, which prevents the content from collapsing.
       *
       * @expose
       * @event
       * @memberof oj.ojCollapsible
       * @instance
       * @ojshortdesc Triggered immediately before the collapsible is collapsed.
       * @ojcancelable
       * @property {Element} header The header that is about to be collapsed.
       * @property {Element} content The content that is about to be collapsed.
       */
        beforeCollapse: null,

      /**
       * Triggered after the collapsible has been collapsed.
       *
       * @expose
       * @event
       * @memberof oj.ojCollapsible
       * @instance
       * @ojshortdesc Triggered immediately after the collapsible is collapsed.
       * @property {Element} header The header that was just collapsed.
       * @property {Element} content The content that was just collapsed.
       */
        collapse: null

      },

    /**
     * @memberof oj.ojCollapsible
     * @instance
     * @protected
     * @override
     */
      _ComponentCreate: function () {
        this._super();

      //  - Stop using ui-helper-reset in the layout widgets.
        this.element.addClass('oj-collapsible oj-component');

        this._processPanels();
        this._refresh();

      //  - collapsible shouldn't implement _init()
        this._initialRender = true;

      // don't fire event on initial render
        var elem = this.element[0];
        this._expandCollapseHandler(this._createEventObject(elem,
        this.options.expanded ? 'ojexpand' : 'ojcollapse'));

        this._initialRender = undefined;
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _createEventObject: function (element, type) {
        return {
          type: type,
          target: element,
          currentTarget: element,
          preventDefault: $.noop
        };
      },

    /**
     * @memberof oj.ojCollapsible
     * @param {Object} menu The JET Menu to open as a context menu
     * @param {Event} event What triggered the menu launch
     * @param {string} eventType "mouse", "touch", "keyboard"
     * @private
     */
      _NotifyContextMenuGesture: function (menu, event, eventType) {
      // Setting the launcher to the "twisty" icon, since that seems to be the only tabbable thing in the collapsible,
      // and it seems to remain tabbable even if the collapsible is disabled.  See the superclass JSDoc for _OpenContextMenu
      // for tips on choosing a launcher.
        this._OpenContextMenu(event, eventType, { launcher: this._getCollapsibleIcon().first() });
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _createIcons: function () {
        var options = this.options;
        var icon = (options.expanded ? OPEN_ICON : CLOSE_ICON);
        var iconTag = this._isDisabled() ? $('<span>') : $("<a href='#'>");

        iconTag.addClass('oj-component-icon oj-clickable-icon-nocontext oj-collapsible-header-icon ' + icon)
        .attr('aria-labelledby', this.header.attr('id'))
        .prependTo(this.header); // @HTMLUpdateOK
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _destroyIcons: function () {
        this.header
        .children('.oj-collapsible-header-icon')
        .remove();
      },

    /**
     * @memberof oj.ojCollapsible
     * @override
     * @private
     */
      _destroy: function () {
      //  - ojcollapsible should resolve busy state when it's destroyed
        this._resolveBusyContext();
        this._cleanup();

      // clean up main element
        this.element
        .removeClass('oj-collapsible oj-component oj-expanded oj-collapsed oj-disabled');

      // clean up headers
        if (this._isDisabled()) {
          this._findFocusables(this.header).removeAttr('tabIndex');
        }

        this.header
        .removeClass('oj-collapsible-header')
        .each(function () {
          if (/^oj-collapsible/.test(this.id)) {
            this.removeAttribute('id');
          }
        });

      // aria
        var focusable = this._findFirstFocusableInHeader();
        focusable.removeAttr('role')
        .removeAttr('aria-controls')
        .removeAttr('aria-expanded')
        .removeAttr('aria-disabled');

        this._destroyIcons();

      // clean up content panels
        this.content
        .css('display', '')
        .removeAttr('aria-hidden')
        .removeAttr('tabIndex')
        .removeClass('oj-component-content oj-collapsible-content')
        .each(function () {
          if (/^oj-collapsible/.test(this.id)) {
            this.removeAttribute('id');
          }
        });
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _cleanup: function () {
      // remove listeners
        this._tearDownEvents();

      // remove wrapper
        if (this.content) {
        //  - DomUtils.unwrap() will avoid unwrapping if the node is being destroyed by Knockout
          oj.DomUtils.unwrap(this.content);
          this.wrapper = null;
        }
      // TODO: remove oj-disabled
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _isDisabled: function () {
        return this.element.hasClass('oj-disabled');
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _getExpandAreaSelector: function () {
        if (this.options.expandArea === 'header') {
          return '> .oj-collapsible-header';
        }

        return '> .oj-collapsible-header > .oj-collapsible-header-icon';
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _getCollapsibleIcon: function () {
        return this.header.find('.oj-collapsible-header-icon');
      },

    /**
     * @memberof oj.ojCollapsible
     * @override
     * @private
     */
      _setOption: function (key, value, flags) {
        if (key === 'expanded') {
          if (value === this.options.expanded) {
            return;
          }

          if (value) {
            this.expand(true);
          } else {
            this.collapse(true);
          }
          return;
        }

      // #5332 - opacity doesn't cascade to positioned elements in IE
      // so we need to add the disabled class to the headers and panels
        if (key === 'disabled') {
          this._super(key, value, flags);
          this.refresh();
          return;
        }

        if (key === 'expandOn' || key === 'expandArea') {
          this._tearDownEvents();
          this._super(key, value, flags);
          this._setupEvents();
        } else {
          this._super(key, value, flags);
        }
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _keydown: function (event) {
        if (event.altKey || event.ctrlKey) {
          return;
        }
        var keyCode = $.ui.keyCode;

        switch (event.keyCode) {
          case keyCode.SPACE:
          case keyCode.ENTER:
            this._toggleHandler(event);
            break;
          default:
        }
      },

    /**
     * Refreshes the visual state of the collapsible.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof oj.ojCollapsible
     * @instance
     * @return {void}
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myCollapsible.refresh();
     */
      refresh: function () {
        this._super();
        this._cleanup();
        this._processPanels();
        this._destroyIcons();
        this._refresh();
      },

    /**
     * Make sure the header slot is the first child of the root element
     * If the slot header is not specified, it will create one with empty text
     * @memberof oj.ojCollapsible
     * @private
     */
      _processHeaderSlots: function () {
        var elem = this.element[0];
        var newHeader = false;

      // add the header node if not specified
        var headers = oj.BaseCustomElementBridge.getSlotMap(elem).header;
        var header;

        if (headers && headers.length) {
        // exactly one child element has the "header" named slot
          if (headers.length === 1) {
            header = headers[0];
          } else {
            // multiple child elements have the "header" named slots, combine them
            var $header = $("<span slot='header'></span>");
            header = $header[0];

            for (var i = 0; i < headers.length; i++) {
              header.appendChild(headers[i]); // @HTMLUpdateOK
            }
            $header.children().attr('slot', '');
            newHeader = true;
          }
        } else {
          header = $("<span slot='header'></span>")[0];
          newHeader = true;
        }

      // make the header slot be the first child
      // Note prepend doesn't work in IE11 and Edge, use insertBefore instead
        if (newHeader || this.element.children().index(header) !== 0) {
          elem.insertBefore(header, elem.firstChild); // @HTMLUpdateOK
        }

        return $(header);
      },

    /**
     * Make sure the default slots are the last child of the root element
     * If there are multiple default slots, combine them
     * @memberof oj.ojCollapsible
     * @private
     */
      _processDefaultSlots: function () {
        var elem = this.element[0];

        var contents = oj.BaseCustomElementBridge.getSlotMap(elem)[''];
        var content;

        if (contents && contents.length === 1) {
          content = contents[0];
        } else {
          var $content = $('<div></div>');
          content = $content[0];
          if (contents && contents.length) {
            for (var i = 0; i < contents.length; i++) {
              content.appendChild(contents[i]); // @HTMLUpdateOK
            }
          }
        // add the content slot
        // Note: append doesn't work in IE11 and Edge, use appendChild instead
          elem.appendChild(content); // @HTMLUpdateOK
        }

        return $(content);
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _processPanels: function () {
      // process header
        if (this._IsCustomElement()) {
          this.header = this._processHeaderSlots();
        } else {
        //  - Stop using ui-helper-reset in the layout widgets.
          this.header = this.element.children(':first-child');
        }
        this.header.addClass('oj-collapsible-header');

      // process content
        if (this._IsCustomElement()) {
          this.content = this._processDefaultSlots();
        } else {
          this.content = this.header.next();
        }
        this.content.addClass('oj-collapsible-content oj-component-content');

        this.content.wrap('<div></div>'); // @HTMLUpdateOK
        this.wrapper = this.content.parent()
        .addClass('oj-collapsible-wrapper');

        if (this.options.disabled) {
          this.element.addClass('oj-disabled');
        }

      // Note: must set tabIndex=-1 to focusable elements
      // to avoid tabbing in a disabled header
        if (this._isDisabled()) {
          this._findFocusables(this.header)
          .attr('tabIndex', -1);
        }
      },

      /**
       * Used for explicit cases where the component needs to be refreshed
       * (e.g., when the value option changes or other UI gestures).
       * @memberof oj.ojCollapsible
       * @private
       */
      _refresh: function () {
        var header = this.header;
        var content = this.content;
        var options = this.options;

        var id = this.element.attr('id');
        if (!id) {
          uid += 1;
          id = uid;
        }

        var collapsibleId = 'oj-collapsible-' + id;
        this.collapsibleId = collapsibleId;

        var headerId = header.attr('id');
        var contentId = content.attr('id');

        if (!headerId) {
          headerId = collapsibleId + '-header';
          header.attr('id', headerId);
        }
        if (!contentId) {
          contentId = collapsibleId + '-content';
          content.attr('id', contentId);
        }

      // aria
        this._createIcons();
        var focusable = this._findFirstFocusableInHeader();
        focusable.attr('role', 'button')
        .attr('aria-controls', contentId)
        .attr('aria-expanded', options.expanded);

        if (this._isDisabled()) {
          focusable.attr('aria-disabled', 'true');
        }

      //  - when collapsible is refreshed, it's content displays & disclosure icon collapsed
        if (options.expanded) {
          content.removeAttr('aria-hidden');
        } else {
          this.wrapper.css({
            'max-height': 0,
            'overflow-y': 'hidden',
            display: 'none'
          });
          content.attr('aria-hidden', 'true');
        }

        this._setupEvents();
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _setupEvents: function () {
        var events =
          {
            keydown: this._keydown
          };

        var event = this.options.expandOn;
        if (event) {
          var self = this;
          $.each(event.split(' '), function (index, eventName) {
          // security test
            if (oj.DomUtils.isValidIdentifier(eventName)) {
              events[eventName] = self._toggleHandler;
            }
          });
        }

        var expandArea = this.element.find(this._getExpandAreaSelector());

      // add listeners on expandArea (event expandArea)
        this._on(expandArea, events);

        this._on(this.wrapper,
          {
            transitionend: this._transitionEndHandler,
            webkitTransitionEnd: this._transitionEndHandler
          });

        if (!this._isDisabled()) {
          this._on(this.element,
            {
              ojfocus: this._focusHandler,
              ojfocusout: this._focusHandler
            });

          this._focusable({
            element: this._getCollapsibleIcon(),
            applyHighlight: true
          });

          this._AddHoverable(expandArea);
          this._AddActiveable(expandArea);
        }
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _tearDownEvents: function () {
        var expandArea = this.element.find(this._getExpandAreaSelector());

        this._RemoveHoverable(expandArea);
        this._RemoveActiveable(expandArea);
        this._off(expandArea);

      // remove wrapper listeners
        if (this.wrapper) {
          this._off(this.wrapper);
        }
        this._off(this.element.add(this.content));
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _toggleHandler: function (event) {
        if (this._isDisabled() || event.isDefaultPrevented()) {
          return;
        }

        if (this.options.expanded) {
          this.collapse(true, event);
        } else {
          this.expand(true, event);
        }

        event.preventDefault();
        event.stopPropagation();

      // set focus on the disclosure icon
        this._getCollapsibleIcon().focus();
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _calcEffectTime: function (jelem) {
        var propertyStr = jelem.css('transitionProperty');
        var delayStr = jelem.css('transitionDelay');
        var durationStr = jelem.css('transitionDuration');
        var propertyArray = propertyStr.split(',');
        var delayArray = delayStr.split(',');
        var durationArray = durationStr.split(',');
        var propertyLen = propertyArray.length;
        var delayLen = delayArray.length;
        var durationLen = durationArray.length;
        var maxTime = 0;

        for (var i = 0; i < propertyLen; i++) {
          var duration = durationArray[i % durationLen];
          var durationMs = (duration.indexOf('ms') > -1) ? parseFloat(duration) : parseFloat(duration) * 1000;
          if (durationMs > 0) {
            var delay = delayArray[i % delayLen];
            var delayMs = (delay.indexOf('ms') > -1) ? parseFloat(delay) : parseFloat(delay) * 1000;

            maxTime = Math.max(maxTime, delayMs + durationMs);
          }
        }

        return maxTime + 100;
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _resolveTransition: function (wrapper) {
        var self = this;

        this._transitionTimer = setTimeout(function () {
          self._transitionEndHandler();
        }, self._calcEffectTime(wrapper));
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _expandCollapseHandler: function (event) {
      // only process expand/collapse for a disabled collapsible on initial render
        if (this._isDisabled() && !this._initialRender) {
          return;
        }

        if (event.target !== this.element[0]) {
          return;
        }

        if (this._initialRender || !event.isDefaultPrevented || !event.isDefaultPrevented()) {
          var element = this.element;
          var content = this.content;
          var wrapper = this.wrapper;
          var isExpanded = (event.type === 'ojexpand');

          var self = this;
          event.preventDefault();

        // fire option change event
          if (!this._initialRender) {
            this._changeExpandedOption(isExpanded);
          }

        //  - expansion animation on initial render.
          if (this._initialRender || document.hidden ||
            this.element.hasClass('oj-collapsible-skip-animation')) {
            if (!isExpanded) {
              wrapper.css('max-height', 0);
              wrapper.hide();
            }
            self._afterExpandCollapse(isExpanded, event);
          } else {
            // do animation
            wrapper.contentHeight = wrapper.outerHeight();

          // Add a busy state for the animation.  The busy state resolver will be invoked
          // when the animation is completed
            if (!this._animationResolve) {
              var busyContext = oj.Context.getContext(element[0]).getBusyContext();
              this._animationResolve = busyContext.addBusyState(
                { description: "The collapsible id='" +
                  this.element.attr('id') + "' is animating." });
            }
            this._transitionEnded = false;

          // expanding
            if (isExpanded) {
            // James: set display:none on the wrapper when it is hidden and then
            // remove display:none when its is shown.
            // This should trigger JAWS into refreshing the buffer.
              wrapper.show();

              setTimeout(function () {
              // if closed, add inner height to content height
                wrapper.contentHeight += content.outerHeight();

                wrapper.addClass('oj-collapsible-transition')
                .css({
                  'max-height': wrapper.contentHeight
                });
                self._resolveTransition(wrapper);
              }, 0);
            } else {
              // collapsing
              // disable transitions & set max-height to content height
              wrapper.removeClass('oj-collapsible-transition');
              wrapper.css({
                'max-height': wrapper.contentHeight,
                'overflow-y': 'hidden'
              });

              // no transition when end state is the same
              if (wrapper.contentHeight === 0) {
                self._transitionEndHandler();
              } else {
                setTimeout(function () {
                // enable & start transition
                  wrapper.addClass('oj-collapsible-transition')
                  .css({
                    'max-height': 0   //! important
                  });
                  self._resolveTransition(wrapper);
                }, 20);
              }
            }
          }
        }
      },

      /**
       * @memberof oj.ojCollapsible
       * @private
       */
      _focusHandler: function (event) {
        if (this._isDisabled()) {
          return;
        }

        if (event.type === 'ojfocusout') {
          this._findFirstFocusableInHeader()
          .attr('tabIndex', -1);

          event.preventDefault();
          event.stopPropagation();
        } else if (event.type === 'ojfocus') {
          this._findFirstFocusableInHeader()
          .attr('tabIndex', 0)
          .focus();
          event.preventDefault();
          event.stopPropagation();
        }
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _findFirstFocusableInHeader: function () {
        return this._findFocusables(this.header).first();
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _findFocusables: function (start) {
      // create <span> or <a> depending on if this.isDisabled
        if (this._isDisabled()) {
          return start.find('span');
        }
        return start.find('a,:input');
      },

    /**
     * Expand a collapsible.<p>
     * Note the beforeExpand event will only be fired when vetoable is true.<p>
     * Please use the <code class="prettyprint">expanded</code> option
     * for expanding a collapsible so that it triggers the beforeExpand event:
     * $( ".selector" ).ojCollapsible( "option", "expanded", true );
     *
     * @expose
     * @ignore
     * @ojtsignore
     * @memberof oj.ojCollapsible
     * @instance
     * @param {boolean} vetoable if event is vetoable
     */
      expand: function (vetoable, event) {
        if (this._isDisabled()) {
          return;
        }

        var eventData =
          {
        /** @expose */
            header: this.header,
        /** @expose */
            content: this.content
          };

        if (!vetoable || this._trigger('beforeExpand', event, eventData) !== false) {
          this._expandCollapseHandler(this._createEventObject(this.element[0], 'ojexpand'));
        }
      },

    /**
     * Collapse a collapsible.<p>
     * Note the beforeCollapse event will only be fired when vetoable is true.<p>
     * Please use the <code class="prettyprint">expanded</code> option
     * for collapsing a collapsible so that it triggers the beforeCollapse event:
     * $( ".selector" ).ojCollapsible( "option", "expanded", false );
     *
     * @expose
     * @ignore
     * @ojtsignore
     * @memberof oj.ojCollapsible
     * @instance
     * @param {boolean} vetoable if event is vetoable
     */
      collapse: function (vetoable, event) {
        if (this._isDisabled()) {
          return;
        }

        var eventData =
          {
        /** @expose */
            header: this.header,
        /** @expose */
            content: this.content
          };

        if (!vetoable || this._trigger('beforeCollapse', event, eventData) !== false) {
          this._expandCollapseHandler(this._createEventObject(this.element[0], 'ojcollapse'));
        }
      },

      /**
       * @memberof oj.ojCollapsible
       * @private
       */
      _transitionEndHandler: function (event) {
      // ignore event if not for this collapsible
        if (this._isDisabled() || (event && event.target !== this.element[0])) {
          return;
        }

        var isMaxHeight = false;
        if (event && event.originalEvent) {
          isMaxHeight = (event.originalEvent.propertyName === 'max-height');
        }

      // if transition property is MaxHeight, clear timer if exists
        if (isMaxHeight && this._transitionTimer) {
          clearTimeout(this._transitionTimer);
          this._transitionTimer = undefined;
        }

        if (event) {
          event.preventDefault();
          event.stopImmediatePropagation();
        }

      // transition end already handled
        if (this._transitionEnded) {
          return;
        }

      // always set flag either timer expired or transition end
        if (isMaxHeight || !event) {
          this._transitionEnded = true;
        }

      // just completed a collapse transition
        if (this.options.expanded) {
          this.wrapper.css({
            'max-height': 9999,
            'overflow-y': ''
          });
        } else {
        // James: set display:none on the wrapper when it is hidden and then remove display:none when its is shown.
        // This should trigger JAWS into refreshing the buffer.
          this.wrapper.hide();
        }

        this.wrapper.removeClass('oj-collapsible-transition');
        this._afterExpandCollapse(this.options.expanded, event);
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _resolveBusyContext: function () {
      // resolve/remove the component busy state
        if (this._animationResolve) {
          this._animationResolve();
          this._animationResolve = null;
        }
      },

    /**
     * @memberof oj.ojCollapsible
     * @private
     */
      _afterExpandCollapse: function (isExpanded, event) {
        var element = this.element;
        var wrapper = this.wrapper;

        if (isExpanded) {
          element.removeClass('oj-collapsed');
          element.addClass('oj-expanded');

        //  - ojcollapsible needs to call oj.components.subtreeshown()/subtreehidden()
          oj.Components.subtreeShown(wrapper[0]);
        } else {
          element.removeClass('oj-expanded');
          element.addClass('oj-collapsed');

        //  - ojcollapsible needs to call oj.components.subtreeshown()/subtreehidden()
          oj.Components.subtreeHidden(wrapper[0]);
        }

        this._getCollapsibleIcon().toggleClass(OPEN_ICON, isExpanded)
      // logic or cause same icon for expanded/collapsed state would remove the oj-icon-class
        .toggleClass(CLOSE_ICON, (!isExpanded || OPEN_ICON === CLOSE_ICON))
        .end();

      // aria
        if (isExpanded) {
          this.content.removeAttr('aria-hidden');
        } else {
          this.content.attr('aria-hidden', 'true');
        }

        this._findFirstFocusableInHeader().attr('aria-expanded', isExpanded);

        this._resolveBusyContext();

        var eventData =
          {
        /** @expose */
            header: this.header,
        /** @expose */
            content: this.content
          };

        if (!this._initialRender) {
          if (isExpanded) {
            this._trigger('expand', event, eventData);
          } else {
            this._trigger('collapse', event, eventData);
          }
        }
      },

    /**
     * @param {boolean} value
     * @memberof oj.ojCollapsible
     * @private
     */
      _changeExpandedOption: function (value) {
        this.option('expanded', value,
          { _context: { writeback: true,
            internalSet: true } });
      },

    //* * @inheritdoc */
      getNodeBySubId: function (locator) {
        if (locator == null) {
          return this.element ? this.element[0] : null;
        }

        var subId = locator.subId;

        switch (subId) {
          case 'oj-collapsible-content':
            return this.content[0];

          case 'oj-collapsible-header':
            return this.header[0];

          case 'oj-collapsible-disclosure':
          case 'oj-collapsible-header-icon':
            return this._getCollapsibleIcon()[0];
          default:
        }

      // Non-null locators have to be handled by the component subclasses
        return null;
      },

    //* * @inheritdoc */
      getSubIdByNode: function (node) {
        var headerIcon = this.getNodeBySubId({ subId: 'oj-collapsible-disclosure' });
        var currentNode = node;
        while (currentNode) {
          if (currentNode === this.content[0]) {
            return { subId: 'oj-collapsible-content' };
          } else if (currentNode === this.header[0]) {
            return { subId: 'oj-collapsible-header' };
          } else if (currentNode === headerIcon) {
            return { subId: 'oj-collapsible-disclosure' };
          }

          currentNode = currentNode.parentElement;
        }
        return null;
      }

    // Fragments:

    /**
     * <p>The default slot is the collapsible's content.
     *
     * @ojchild Default
     * @memberof oj.ojCollapsible
     *
     * @example <caption>Initialize the Collapsible with child content specified:</caption>
     * &lt;oj-collapsible>
     *   &lt;h3 slot='header'>Header 1&lt;/h3>
     *   &lt;p>Content 1&lt;/p>
     * &lt;/oj-collapsible>
     */

    /**
     * <p>The <code class="prettyprint">header</code> slot is the collapsible's header. If not specified, the header contains only an open/close icon. Note that the header text is required for JET collapsible for accessibility purposes.</p>
     *
     * @ojslot header
     * @memberof oj.ojCollapsible
     *
     * @example <caption>Initialize the Collapsible with the header slot specified:</caption>
     * &lt;oj-collapsible>
     *   &lt;h3 slot='header'>Header 1&lt;/h3>
     *   &lt;p>Content 1&lt;/p>
     * &lt;/oj-collapsible>
     */

    /**
     * <p>Sub-ID for the disclosure icon of a Collapsible.</p>
     *
     * @ojsubid oj-collapsible-disclosure
     * @memberof oj.ojCollapsible
     * @example <caption>Get the Collapsible disclosure icon:</caption>
     * var node = myCollapsible.getNodeBySubId({"subId": "oj-collapsible-disclosure"});
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Header</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Toggle disclosure state</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojCollapsible
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Header</td>
     *       <td><kbd>Space or Enter</kbd></td>
     *       <td>Toggle disclosure state.</tr>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>Disabled items can receive keyboard focus, but do not allow any other interaction.
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojCollapsible
     */

    });
}());

(function () {
  var ojCollapsibleMeta = {
    properties: {
      disabled: {
        type: 'boolean'
      },
      expandArea: {
        type: 'string',
        enumValues: ['header', 'disclosureIcon']
      },
      expanded: {
        type: 'boolean',
        writeback: true
      }
    },
    events: {
      beforeCollapse: {},
      beforeExpand: {},
      collapse: {},
      expand: {}
    },
    extension: {
      _WIDGET_NAME: 'ojCollapsible',
      _CONTROLS_SUBTREE_HIDDEN: true
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-collapsible', 'baseComponent', ojCollapsibleMeta);
  oj.CustomElementBridge.register('oj-collapsible', { metadata: oj.CustomElementBridge.getMetadata('oj-collapsible') });
}());

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

 /**
 * @export
 * @class oj.CollectionCellSet
 * @classdesc A CellSet represents a collection of cells.  The CellSet is an object returned by the success callback
 * of the fetchCells method on DataGridDataSource.  This implementation of CellSet is used by the
 * CollectionDataGridDataSource.
 * @param {number} startRow the start row index of the cell set
 * @param {number} endRow the end row index of the cell set
 * @param {number} startColumn the start column index of the cell set
 * @param {number} endColumn the end column index of the cell set
 * @param {Array|null} columns the set of column keys
 * @constructor
 * @since 1.0
 * @hideconstructor
 * @ojtsignore
 * @see oj.CollectionDataGridDataSource
 */
oj.CollectionCellSet = function(startRow, endRow, startColumn, endColumn, columns)
{
    // assert startRow/startColumn are number
    oj.Assert.assertArrayOrNull(columns);

    this.m_startRow = startRow;
    this.m_endRow = endRow;
    this.m_startColumn = startColumn;
    this.m_endColumn = endColumn;
    this.m_columns = columns;
};

/**
 * Sets the models used in this cell set.
 * @param {Array} models an array of oj model for the cell set
 * @memberof oj.CollectionCellSet
 * @private
 */
oj.CollectionCellSet.prototype.setModels = function(models)
{
    oj.Assert.assertArray(models);
    // make sure the array size is valid
    if (models != null && models.length === this.getCount("row"))
    {
        this.m_models = models;
    }
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available
 * 2) the index specified is out of bounds.
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} an object with property data for the specified index.
 * @export
 * @expose
 * @instance
 * @memberof oj.CollectionCellSet 
 */
oj.CollectionCellSet.prototype.getData = function(indexes)
{
    var columnKey, model, self, returnObj;

    // indexes are validated in _getModel
    model = this._getModel(indexes);
    if (model == null)
    {
        return null;
    }

    columnKey = this.m_columns[indexes['column']];
    self = this;
    returnObj = {};
    Object.defineProperty(returnObj, 'data', {
        enumerable: true,
        get: function(){ return model.get(columnKey); },
        set: function (newValue) {
            model.set(columnKey, newValue, {'silent':true});
        }
    });
    return returnObj;
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available
 * 2) the index specified is out of bounds.
 * @param {Object} indexes the index of each axis in which we want to retrieve the metadata from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return the metadata object for the specific index.  The metadata that the DataGrid supports are: 
 *         1) keys - the key (of each axis) of the cell.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionCellSet 
 */
oj.CollectionCellSet.prototype.getMetadata = function(indexes)
{
    var column, model, keys;

    // indexes are validated in _getModel
    model = this._getModel(indexes);
    if (model == null)
    {
        return null;
    }

    // extract column index
    column = indexes['column'];

    keys = {"row": oj.CollectionDataGridUtils._getModelKey(model), "column": this.m_columns[column]};
    return {"keys": keys};
};

/**
 * Gets the Model based on indexes.
 * @private
 * @memberof oj.CollectionCellSet
 */
oj.CollectionCellSet.prototype._getModel = function(indexes)
{
    var row, column;

    // make sure models is populated
    if (this.m_models == null)
    {
        return null;
    }

    oj.Assert.assertObject(indexes);

    // extract row and column index
    row = indexes['row'];
    column = indexes['column'];

    // make sure index are valid
    oj.Assert.assert(row >= this.m_startRow && row <= this.m_endRow && column >= this.m_startColumn && column <= this.m_endColumn);

    return this.m_models[row - this.m_startRow];
};

/**
 * Gets the actual count of the result set for the specified axis.  Valid values are "row" and "column".
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 * @return {number} the actual count of the result set for the specified axis.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionCellSet 
 */
oj.CollectionCellSet.prototype.getCount = function(axis)
{
    if (axis === "row")
    {
        return Math.max(0, this.m_endRow - this.m_startRow);
    }

    if (axis === "column")
    {
        return Math.max(0, this.m_endColumn - this.m_startColumn);
    }

    return 0;
};

/**
 * Gets the extent of a particular row/column index within the context of the cellSet.
 * Extent is defined as the number of indexes along the appropriate axis spanned by the cell.
 * If the extent extends beyond the start and end of the requested cell range the extent should be trimmed to the edge of the requested cell range and the object for {'more': {'before', 'after'}} should have the value appropriate boolean set.
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} an object containing two properties row and column. Each of those properties has two sub properties:
 *              extent: the number of absolute indexes spanned by the cell at this index
 *                      bounded by the edges of the result set for the specified axis.
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is available in the cellSet
 * @example <caption>In this example the cell spans 5 row indexes and 2 column indexes and there are more column indexes spanned by the cell that
 *              aren't included in this cellSet:</caption>
 * {
 *  'row': {'extent':5, 'more': {'before':false, 'after':false}},
 *  'column': {'extent':2, 'more': {'before':false, 'after':true}}
 * }
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionCellSet
 * @since 4.0.0
 */
oj.CollectionCellSet.prototype.getExtent = function(indexes)
{
    return {'row': {'extent':1, 'more': {'before':false, 'after':false}},
            'column': {'extent':1, 'more': {'before':false, 'after':false}}};
};

/**
 * Gets the m_startRow property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionCellSet
 */
oj.CollectionCellSet.prototype.getStartRow = function()
{
    return this.m_startRow;
};

/**
 * Gets the m_endRow property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionCellSet
 */
oj.CollectionCellSet.prototype.getEndRow = function()
{
    return this.m_endRow;
};

/**
 * Gets the m_startColumn property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionCellSet
 */
oj.CollectionCellSet.prototype.getStartColumn = function()
{
    return this.m_startColumn;
};

/**
 * Gets the m_endColumn property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionCellSet
 */
oj.CollectionCellSet.prototype.getEndColumn = function()
{
    return this.m_endColumn;
};

/**
 * Gets the m_columns property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionCellSet
 */
oj.CollectionCellSet.prototype.getColumns = function()
{
    return this.m_columns;
};


/**
 * Copyright (c) 2014, 2015 Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.CollectionDataGridDataSource
 * @classdesc An oj.Collection based implementation of the DataGridDataSource.
 * @see oj.Collection
 * @param {Object} collection the oj collection to adapter the DataGridDataSource
 * @param {Object=} options optional settings on this oj collection data source
 * @property {string=} options.rowHeader the key of the attribute designated as the row header
 * @property {Array.<string>=} options.columns explicitly specifies columns to display and in
 *        what order. These columns must be a subset of attributes from Model.
 * @export
 * @constructor
 * @since 1.0
 * @extends oj.DataGridDataSource
 * @ojtsignore
 */
oj.CollectionDataGridDataSource = function(collection, options)
{
    this.collection = collection;
    if (options != null)
    {
        this.rowHeader = options['rowHeader'];
        this.columns = options['columns'];
    }
    this._setSortInfo();

    oj.CollectionDataGridDataSource.superclass.constructor.call(this);
};

//subclass of DataGridDataSource
oj.Object.createSubclass(oj.CollectionDataGridDataSource, oj.DataGridDataSource, "oj.CollectionDataGridDataSource");

/**
 * Initial the OJ collection based data source.
 * @instance
 * @override
 * @protected
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype.Init = function()
{
    // call super
    oj.CollectionDataGridDataSource.superclass.Init.call(this);

    this.pendingHeaderCallback = {};
    this._registerEventListeners();
};

/**
 * Register event handlers on the underlying OJ collection.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._registerEventListeners = function()
{
    this.collection.on("add", this._handleModelAdded.bind(this));
    this.collection.on("remove", this._handleModelDeleted.bind(this));
    this.collection.on("change", this._handleModelChanged.bind(this));
    this.collection.on("refresh", this._handleCollectionRefresh.bind(this));
    this.collection.on("reset", this._handleCollectionReset.bind(this));
};

/**
 * Determines if data has been fetched if virtual
 * @return {boolean} true if data is locally available, false otherwise.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._isDataAvailable = function()
{
    return (this.data != null);
};

/**
 * Returns the total number of rows or columns.  If the value return is not >= 0 then it is automatically assumed
 * that the total count is unknown. In the case of paging returns the total number of rows/colums on the page.
 * @param {string} axis the axis in which we inquire for the total count.  Valid values are "row" and "column".
 * @return {number} the total number of rows/columns.
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource 
 */
oj.CollectionDataGridDataSource.prototype.getCount = function(axis)
{
    var totalSize;

    if (this.precision == undefined)
    {
        this.precision = {};
    }

    if (axis == "row")
    {
        totalSize = this._totalSize();
        // 1. totalResults undefined
        // 2. data hasn't been fetched yet and totalResults set to 0 on collection
        // 3. the collection has size but not totalResults, truly unknown count
        if (totalSize === -1 || (totalSize === 0  && (!this._isDataAvailable() || this._size() > 0)))
        {
            this.precision[axis] = "estimate";
            return -1;
        }
        // otherwise we will return the size of the collection or the pageSize
        this.precision[axis] = "exact";
        return this._size();
    }
    if (axis == "column")
    {
        // check whether column info are available
        if (this.columns != null)
        {
            this.precision[axis] = "exact";
            return this.columns.length;
        }
        else
        {
            this.precision[axis] = "estimate";
            return -1;
        }
    }

    // should not get here
    return 0;
};
/**
 * Returns whether the total count returned in getCount function is an actual or an estimate.
 * @param {string} axis the axis in which we inquire whether the total count is an estimate.  Valid values are
 *        "row" and "column".
 * @return {string} "actual" if the count returned in getCount function is the actual count, "estimate" if the
 *         count returned in getCount function is an estimate.  The default value is "actual".
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource 
 */
oj.CollectionDataGridDataSource.prototype.getCountPrecision = function(axis)
{
    // if precision has not been determine for the axis, invoke getCount
    if (this.precision == null || this.precision[axis] == null)
    {
        this.getCount(axis);
    }
    return this.precision[axis];
};

/**
 * Fetch a range of headers from the data source.
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count.
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column".
 * @property {number} headerRange.start the start index of the range in which the header data are fetched.
 * @property {number} headerRange.count the size of the range in which the header data are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch headers operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function({startHeaderSet: Object}, {headerRange: Object}, {endHeaderSet: Object}):undefined} callbacks.success the callback to invoke when fetch headers completed successfully.
 *        The function takes three parameters: HeaderSet object representing start headers, headerRange object passed into the original fetchHeaders call,
 *        and a HeaderSet object representing the end headers along the axis.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource 
 * @return {undefined}
 */
oj.CollectionDataGridDataSource.prototype.fetchHeaders = function(headerRange, callbacks, callbackObjects)
{
    var axis, callback;
    // still fetching, just store the callback info
    if (callbacks != null)
    {
        axis = headerRange['axis'];
        callback = {};
        callback.headerRange = headerRange;
        callback.callbacks = callbacks;
        callback.callbackObjects = callbackObjects;
        this.pendingHeaderCallback[axis] = callback;
    }
};

/**
 * Handle success fetchHeaders request
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count.
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column".
 * @property {number} headerRange.start the start index of the range in which the header data are fetched.
 * @property {number} headerRange.count the size of the range in which the header data are fetched.
 * @param {Object} callbacks the callbacks to be invoke when fetch headers operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function({startHeaderSet: Object}, {headerRange: Object}, {endHeaderSet: Object}):undefined} callbacks.success the callback to invoke when fetch headers completed successfully.
 *        The function takes three parameters: HeaderSet object representing start headers, headerRange object passed into the original fetchHeaders call,
 *        and a HeaderSet object representing the end headers along the axis.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @param {Object=} actualRange the count and start returned from the server
 * @property {number} actualRange.start the start index of the data the server returned
 * @property {number} actualRange.count the size of the range the server returned
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._handleHeaderFetchSuccess = function(headerRange, callbacks, callbackObjects, actualRange)
{
    var axis, start, count, end, headerSet;

    axis = headerRange.axis;
    start = headerRange.start;
    count = headerRange.count;

    if (axis === "column")
    {
        // column headers, this.columns should be populated by now
        if (this.columns != null)
        {
            end = Math.min(this.columns.length, start+count);
            headerSet = new oj.CollectionHeaderSet(start, end, this.columns, undefined, this._sortInfo);
        }
    }
    else if (axis === "row")
    {
        // row headers, return non-empty header set if row header is specified
        if (this.rowHeader != null)
        {
            //need the actual rows that the server returned to create the header set
            if (actualRange != null)
            {
                count = actualRange.count;
            }
            end = Math.min(this._size(), start+count);

            headerSet = new oj.CollectionHeaderSet(start, end, this.columns, this.rowHeader);
            // resolve any promises before invoking callbacks
            this._resolveModels(start, end, headerSet, headerRange, callbacks, callbackObjects);

            // resolveModels will invoke callbacks
            return;
        }
    }

    // invoke callback
    if (callbacks != null && callbacks['success'])
    {
        callbacks['success'].call(callbackObjects['success'], headerSet, headerRange, null);
    }
};

/**
 * Helper method to extract range information from cellRanges
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched.
 * @return {Object} an object containing rowStart, rowCount, colStart, colCount
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._getRanges = function(cellRanges)
{
    var i, cellRange, rowStart, rowCount, colStart, colCount;

    // extract the start and end row/column info from cellRanges (there should only be two, one for each axis)
    for (i=0; i<cellRanges.length; i+=1)
    {
        cellRange = cellRanges[i];
        if (cellRange['axis'] === "row")
        {
            rowStart = cellRange['start'];
            rowCount = cellRange['count'];
        }
        else if (cellRange['axis'] === "column")
        {
            colStart = cellRange['start'];
            colCount = cellRange['count'];
        }
    }

    // return object containing the ranges
    return {'rowStart': rowStart, 'rowCount': rowCount, 'colStart': colStart, 'colCount': colCount};
};

/**
 * Handle success fetchCells request
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched.
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function({cellSet: Object}, {cellRanges: Array.<Object>}):undefined} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @param {Object=} actualRange the count and start returned from the server
 * @property {number} actualRange.start the start index of the data the server returned
 * @property {number} actualRange.count the size of the range the server returned
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._handleCellFetchSuccess = function(cellRanges, callbacks, callbackObjects, actualRange)
{
    var ranges, rowStart, rowEnd, colStart, colEnd, cellSet;

    // extract the start and end row/column info from cellRanges (there should only be two, one for each axis)
    ranges = this._getRanges(cellRanges);
    rowStart = ranges['rowStart'];
    //use the actual rows returned by the server
    if (actualRange != null)
    {
        rowEnd = Math.min(this._size(), rowStart + actualRange['count']);
    }
    else
    {
        rowEnd = Math.min(this._size(), rowStart + ranges['rowCount']);
    }
    colStart = ranges['colStart'];
    //if there are no columns make sure we return an empty cellSet
    colEnd = Math.min(this.columns == null ? 0 : this.columns.length, colStart + ranges['colCount']);

    // resolves models at row range
    cellSet = new oj.CollectionCellSet(rowStart, rowEnd, colStart, colEnd, this.columns);
    this._resolveModels(rowStart, rowEnd, cellSet, cellRanges, callbacks, callbackObjects);
};

/**
 * Resolves all the promises from Collection before invoking callbacks
 * @param {number} rowStart the start row index in the cell set
 * @param {number} rowEnd the end row index in the cell set
 * @param {Object} set the result HeaderSet or CellSet that is return to callbacks when models are resolved
 * @param {Array.<Object>|Object} ranges Information about the header/cell range.
 * @property {function(Object)} callbacks.success the callback to invoke when fetch headers/cells completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when fetch headers/cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._resolveModels = function(rowStart, rowEnd, set, ranges, callbacks, callbackObjects)
{
    var promises, i;

    promises = [];
    for (i = rowStart; i < rowEnd; i++)
    {
        promises.push(this.collection.at(i, {'deferred':true}));
    }

    // resolves all promises
    Promise.all(promises).then(function(models)
    {
        // all promises resolved, invoke callback with cell/header set
        set.setModels(models);
        callbacks['success'].call(callbackObjects['success'], set, ranges);
    });
};

/**
 * Fetch a range of cells from the data source.
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched. 
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function({cellSet: Object}, {cellRanges: Array.<Object>}):undefined} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype.fetchCells = function(cellRanges, callbacks, callbackObjects)
{
    // still fetching, just store the callback info
    if (callbacks != null)
    {
        this.pendingCellCallback = {};
        this.pendingCellCallback.cellRanges = cellRanges;
        this.pendingCellCallback.callbacks = callbacks;
        this.pendingCellCallback.callbackObjects = callbackObjects;
    }

    // kick start a setRangeLocal call on the collection
    this._fetchCells(cellRanges);
};

/**
 * Processing pending header callbacks.
 * @param {string} axis the axis to check for pending header callbacks.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._processPendingHeaderCallbacks = function(axis)
{
    var pendingCallback, headerRange, callbacks, callbackObjects, actualRange;

    // check if there's callback remaining for the axis
    pendingCallback = this.pendingHeaderCallback[axis];
    if (pendingCallback != null)
    {
        // todo: check whether pending header range matches result
        headerRange = pendingCallback.headerRange;
        callbacks = pendingCallback.callbacks;
        callbackObjects = pendingCallback.callbackObjects;
        if (axis === "row")
        {
            actualRange = pendingCallback.actualRange;
        }
        this._handleHeaderFetchSuccess(headerRange, callbacks, callbackObjects, actualRange);

        // clear any pending callback
        this.pendingHeaderCallback[axis] = null;
    }
};

/**
 * Processing pending cell callbacks.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._processPendingCellCallbacks = function()
{
    var cellRanges, callbacks, callbackObjects, actualRange;

    cellRanges = this.pendingCellCallback.cellRanges;
    callbacks = this.pendingCellCallback.callbacks;
    callbackObjects = this.pendingCellCallback.callbackObjects;
    actualRange = this.pendingCellCallback.actualRange;
    // handles success cell fetch
    this._handleCellFetchSuccess(cellRanges, callbacks, callbackObjects, actualRange);
    
    this.pendingCellCallback = null;
};

/**
 * Internal method to handle fetching of cells for virtualized collection.
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._fetchCells = function(cellRanges)
{
    var ranges, rowStart, rowCount;

    ranges = this._getRanges(cellRanges);
    rowStart = ranges['rowStart'];
    rowCount = ranges['rowCount'];

    // set the range local for the requested range
    this.collection.setRangeLocal(rowStart, rowCount).then(function (actual)
    {
        // keeps track of whether or not the data source has fetched
        this.data = true;

        this._setActualCallbackRanges(actual.start, actual.count);

        // check if we need to poach columns from row
        if (this.columns === undefined)
        {
            this.collection.at(rowStart, {'deferred':true}).then(function(model)
            {
                if (model != null)
                {
                    this._setupColumns(model);
                }
                // now we can complete the fetch
                this._fetchCellsComplete(cellRanges);
            }.bind(this));

            // process must be done after columns are discovered
            return;
        }
        else
        {
            this._fetchCellsComplete(cellRanges);
        }
    }.bind(this), function (e)
    {
        this._fetchCellsError(e);
    }.bind(this));
};

/**
 * Handles error from setRangeLocal call in fetchCells.
 * @param {Object} error the error returned from setRangeLocal error callback.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._fetchCellsError = function(error)
{
    // log the error
    oj.Logger.error(error);

    // invoke any header error callbacks
    if (this.pendingHeaderCallback != null)
    {
        this._processPendingHeaderErrorCallbacks('column', error);
        this._processPendingHeaderErrorCallbacks('row', error);
    }

    // invoke any cell error callbacks
    if (this.pendingCellCallback != null)
    {
        this._processPendingCellErrorCallbacks(error);
    }
};

/**
 * Handles any pending header callbacks in the case of error being thrown.
 * @param {string} axis the header axis
 * @param {Object} error the error returned from setRangeLocal error callback
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._processPendingHeaderErrorCallbacks = function(axis, error)
{
    var pendingCallback, callbacks, callbackObjects, headerRange;

    pendingCallback = this.pendingHeaderCallback[axis];
    if (pendingCallback != null)
    {
        callbacks = pendingCallback.callbacks;
        callbackObjects = pendingCallback.callbackObjects;
        headerRange = pendingCallback.headerRange;

        // invoke error callback
        if (callbacks['error'])
        {
            callbacks['error'].call(callbackObjects['error'], error, headerRange);
        }

        // clear any pending callback
        this.pendingHeaderCallback[axis] = null;
    }
};

/**
 * Handles any pending cell callbacks in the case of error being thrown.
 * @param {Object} error the error returned from setRangeLocal error callback
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._processPendingCellErrorCallbacks = function(error)
{
    var callbacks, callbackObjects, cellRanges;

    callbacks = this.pendingCellCallback.callbacks;
    callbackObjects = this.pendingCellCallback.callbackObjects;
    cellRanges = this.pendingCellCallback.cellRanges;

    // invoke error callback
    if (callbacks['error'])
    {
        callbacks['error'].call(callbackObjects['error'], error, cellRanges);
    }

    // clear any pending callback
    this.pendingCellCallback = null;
};

/**
 * Finish fetch cells operation
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._fetchCellsComplete = function(cellRanges)
{
    // make sure we only callback when we have the most recent request
    if (this.pendingCellCallback.cellRanges != cellRanges)
    {
        return;
    }
    
    // check outstanding header calls
    if (this.pendingHeaderCallback != null)
    {
        this._processPendingHeaderCallbacks('column');
        this._processPendingHeaderCallbacks('row');
    }

    // finally process outstanding cell calls
    if (this.pendingCellCallback != null)
    {
        this._processPendingCellCallbacks();
    }
};

/**
 * Takes the actual result start and count from the server and adds it to the pending callback objects
 * as the attribute actualRange
 * @param {number} start the start index from the server
 * @param {number} count the count of records from the server
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._setActualCallbackRanges = function(start, count)
{
    var actualRange = {'start':start, 'count':count};

    if (this.pendingHeaderCallback['row'] != null)
    {
        this.pendingHeaderCallback['row'].actualRange = actualRange;
    }

    if (this.pendingCellCallback != null)
    {
        this.pendingCellCallback.actualRange = actualRange;
    }
};

/**
 * @param {Object} model
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._setupColumns = function(model)
{
    this.columns = model.keys();
    if (this.columns.indexOf(this.rowHeader) != -1)
    {
        this.columns.splice(this.columns.indexOf(this.rowHeader),1);
    }
};

/**
 * Returns the keys based on the indexes.
 * @param {Object} indexes the index for each axis
 * @property {number|null} indexes.row the index for the row axis
 * @property {number|null} indexes.column the index for the column axis
 * @return {Promise.<Object>} a Promise object which upon resolution will pass in an object containing the keys for each axis
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource 
 */
oj.CollectionDataGridDataSource.prototype.keys = function(indexes)
{
    var rowIndex, columnIndex, rowKey, columnKey, atPromise, self;
    rowIndex = indexes['row'];
    columnIndex = indexes['column'];
    self = this;

    return new Promise(function(resolve, reject) {
        atPromise = self.collection.at(rowIndex, {deferred: true}).then(function(rowModel)
        {
            if (rowModel == null)
            {
                resolve({"row": null,  "column": null});
            }
            else
            {
                rowKey = oj.CollectionDataGridUtils._getModelKey(rowModel);
                if (self.columns == null)
                {
                    self._setupColumns(rowModel);
                }
                columnKey = self.columns[columnIndex];
				resolve({"row": rowKey ,  "column": columnKey});
            }
        }.bind(self));
    });
};

/**
 * Returns the row and column index based on the keys. In a paging case returns the 
 * index on the page, not the absolute index in the array.
 * @param {Object} keys the key for each axis
 * @property {*} keys.row the key for the row axis
 * @property {*} keys.column the key for the column axis
 * @return {Promise.<Object>} a promise object containing the index for each axis
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource 
 */
oj.CollectionDataGridDataSource.prototype.indexes = function(keys)
{
    var rowKey, columnKey, columnIndex, self;
    rowKey = keys['row'];
    columnKey = keys['column'];
    self = this;

    return new Promise(function(resolve, reject)
    {
        self.collection.indexOf(rowKey, {deferred:true}).then(function(rowIndex)
        {
            // if the row exists but columns not yet set
            if (rowIndex != -1 && self.columns == null)
            {
                self.collection.at(rowIndex, {'deferred':true}).then(function(model)
                {
                    self._setupColumns(model);
                    columnIndex = self.columns.indexOf(columnKey);
                    resolve({"row": rowIndex, "column": columnIndex});
                }.bind(self));
            }
            // if the row exists or not and columns set
            else
            {
                columnIndex = self.columns.indexOf(columnKey);
                resolve({"row": rowIndex, "column": columnIndex});
            }
        }.bind(self));
    });
};

/**
 * Determines whether this CollectionDataGridDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort" and "move"
 * @return {string|null} the name of the feature.  For sort, the valid return values are: "full", "none".  Returns null if the
 *         feature is not recognized.
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource
*/
oj.CollectionDataGridDataSource.prototype.getCapability = function(feature)
{
    if (feature === 'sort')
    {
        // OJ collection based data source supports column sorting only
        return 'column';
    }
    else if (feature === 'move')
    {
        // OJ collection based data source supports row moving only
        return 'row';
    }
    return null;
};

/**
 * Performs a sort on the data source.
 * @param {Object|null} criteria the sort criteria. 
 * @property {string} criteria.axis The axis in which the sort is performed, valid values are "row", "column"
 * @property {*} criteria.key The key that identifies which header to sort
 * @property {string} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the sort operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.  
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource
 * @return {undefined}
 */
oj.CollectionDataGridDataSource.prototype.sort = function(criteria, callbacks, callbackObjects)
{
    var comparator, direction, key, axis;

    // make sure callbackObjects is not null
    if (callbackObjects == null)
    {
        callbackObjects = {};
    }

    // reset sort order if null is specified
    if (criteria == null)
    {
        this._resetSortOrder(callbacks, callbackObjects);
        return;
    }

    direction = criteria['direction'];
    key = criteria['key'];
    axis = criteria['axis'];

    if (axis === "column") {
        //check to see if collection is virtual, if so set the comparator and direction
        if (this.collection.IsVirtual())
        {
            this.collection['comparator'] = key;
            if (direction === 'ascending')
            {
                this.collection['sortDirection'] = 1;
            }
            else
            {
                this.collection['sortDirection'] = -1;
            }
        }
        else
        {
            //if the collection is local supply a comparator to allow date sorting
            if (direction === 'ascending') {
                comparator = function(a, b) {
                    var as, bs;
                    //Get the values from the model objects
                    a = a.get(key);
                    b = b.get(key);
                    //Strings of numbers return false, so we can compare strings of numbers with numbers
                    as = isNaN(a);
                    bs = isNaN(b);
                    //If they dates, turn them into sortable strings
                    if (a instanceof Date) {
                        a = a.toISOString();
                        as = true;
                    }
                    if (b instanceof Date) {
                        b = b.toISOString();
                        bs = true;
                    }
                    //both are string
                    if (as && bs)
                    {
                        return a < b ? -1 : a === b ? 0 : 1;
                    }
                    //only a is a string
                    if (as)
                    {
                        return 1;
                    }
                    //only b is a string
                    if (bs)
                    {
                        return -1;
                    }
                    //both are numbers
                    return a - b;
                };
            }
            if (direction === 'descending') {
                comparator = function(a, b) {
                    var as, bs;
                    a = a.get(key);
                    b = b.get(key);
                    as = isNaN(a);
                    bs = isNaN(b);
                    if (a instanceof Date) {
                        a = a.toISOString();
                    }
                    if (b instanceof Date) {
                        b = b.toISOString();
                    }
                    if (as && bs)
                    {
                        return a > b ? -1 : a === b ? 0 : 1;
                    }
                    if (as)
                    {
                        return -1;
                    }
                    if (bs)
                    {
                        return 1;
                    }
                    return b - a;
                };
            }
            this.collection['comparator'] = comparator;
        }

        this.collection.sort();

        this._setSortInfo(key);

        if (callbacks != null && callbacks['success'] != null)
        {
            callbacks['success'].call(callbackObjects['success']);
        }
    }
    else
    {
        if (callbacks != null && callbacks['error'] != null)
        {
            callbacks['error'].call(callbackObjects['error'], "Axis value not supported");
        }
    }
};

/**
 * Reset the sort order of the data.
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the sort operation.  The callback
 *        properties are "success" and "error".
 * @property {function()} callbacks.success the callback to invoke when the sort completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.
 *        You can specify the callback object for each callbacks using the "success" and "error" properties.
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._resetSortOrder = function(callbacks, callbackObjects)
{
    // clear out comparator to reset sort order
    this.collection['comparator'] = null;
    this.collection.reset();

    if (callbacks != null && callbacks['success'] != null)
    {
        callbacks['success'].call(callbackObjects['success']);
    }
};

/**
 * Get sort info from the collection and setup the original sortInfo
 * @param {string|number=} key the key of the sorted column if we have performed a sort
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._setSortInfo = function(key)
{
    var comparator, direction;
    comparator = this.collection['comparator'];
    direction = this.collection['sortDirection'] === -1 ? 'descending' : 'ascending';

    if (key == null && typeof comparator === 'function')
    {
        this._sortInfo = {};
        return;
    }

    this._sortInfo = {};
    this._sortInfo['axis'] = 'column';
    this._sortInfo['direction'] = direction;
    this._sortInfo['key'] = key == null ? comparator : null;
};

/**
 * Move a model to a new index in the collection, if referenceRow is null adds to the end
 * @param {*} rowToMove the key of the model that should be moved
 * @param {*} referenceRow the key of the model that the moved model should be inserted before or after
 * @param {string} position The position of the moved row relative to the reference row.  
 *        Valid values are: "before", "after" 
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the move operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource
 * @return {undefined}
 */
oj.CollectionDataGridDataSource.prototype.move = function(rowToMove, referenceRow, position, callbacks, callbackObjects)
{
    var indexPromise;
    this.collection.get(rowToMove, {deferred: true}).then(
        function(moveModel) {
            if (referenceRow == null)
            {
                this.collection.remove(moveModel);
                this.collection.add(moveModel);
                if (callbacks != null && callbacks['success'] != null)
                {
                    callbacks['success'].call(callbackObjects['success']);
                }
            }
            else
            {
                if (rowToMove === referenceRow)
                {
                    indexPromise = this.collection.indexOf(referenceRow, {deferred: true});
                    this.collection.remove(moveModel);
                }
                else
                {
                    this.collection.remove(moveModel);
                    indexPromise = this.collection.indexOf(referenceRow, {deferred: true});
                }

                indexPromise.then(function(newIndex) {
                    this.collection.add(moveModel, {at: newIndex, force:true});
                    if (callbacks != null && callbacks['success'] != null)
                    {
                        callbacks['success'].call(callbackObjects['success']);
                    }
                }.bind(this));
            }
        }.bind(this));
};

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine
 *        the destination of where the row should moved to.
 * @param {string} position The position of the moved row relative to the reference row.
 *        Valid values are: "before", "after".
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @export
 * @method
 * @instance
 * @memberof oj.CollectionDataGridDataSource 
 */
oj.CollectionDataGridDataSource.prototype.moveOK = function(rowToMove, referenceRow, position)
{
    return "valid";
};


//////////////////////////////////// Event listeners /////////////////////////////////////
/**
 * Returns an Object for an event
 * @param {string} operation the operation done on the model
 * @param {*} rowKey the key for the row axis
 * @param {*} columnKey the key for the column axis
 * @param {number=} rowIndex the index for the row axis
 * @param {number=} columnIndex the index for the column axis
 * @return {Object} an object containing the the source, operation, and keys of the event
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._getModelEvent = function(operation, rowKey, columnKey, rowIndex, columnIndex)
{
    var event = {};
    event['source'] = this;
    event['operation'] = operation;
    event['keys'] = {'row': rowKey, 'column': columnKey};
    event['indexes'] = {'row': rowIndex, 'column': columnIndex};
    return event;
};

/**
 * Handle a model add to the collection
 * @param {Object} model The model being added to the collection
 * @param {Object} collection The collection the model was added to
 * @param {Object} args additional params passed by the event
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._handleModelAdded = function(model, collection, args)
{
    var event, rowKey;
    rowKey = oj.CollectionDataGridUtils._getModelKey(model);
    event = this._getModelEvent('insert', rowKey, null, model['index'], -1);
    this.handleEvent("change", event);
};

/**
 * Handle a model delete from the collection
 * @param {Object} model The model being deleted from the collection
 * @param {Object} collection The collection the model was added to
 * @param {Object} args additional params passed by the event
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._handleModelDeleted = function(model, collection, args)
{
    var event, rowKey;
    rowKey = oj.CollectionDataGridUtils._getModelKey(model);
    event = this._getModelEvent('delete', rowKey, null, args['index'], -1);
    this.handleEvent("change", event);

};

/**
 * Handle a model change in the collection
 * @param {Object} model The model being changed in the collection
 * @param {Object} collection The collection the model was added to
 * @param {Object} args additional params passed by the event
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._handleModelChanged = function(model, collection, args)
{
    var event, rowKey;
    //pass the indexes into the grid on model change
    rowKey = oj.CollectionDataGridUtils._getModelKey(model);
    event = this._getModelEvent('update', rowKey, null, model['index'], -1);
    this.handleEvent("change", event);
};

/**
 * Handle a collection refresh, by firing change event with operation refresh
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._handleCollectionRefresh = function()
{
    var event;
    this.data = null;
    event = this._getModelEvent('refresh', null, null);
    this.handleEvent("change", event);
};

/**
 * Handle a collection reset, by firing change event with operation reset
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._handleCollectionReset = function()
{
    var event;
    this.data = null;
    event = this._getModelEvent('reset', null, null);
    this.handleEvent("change", event);
};

/**
 * Get the length of the collection. -1 if an initial fetch has not been
 * done yet. Default to the size of the collection. If pageSize is set then
 * limit it.
 * @returns {number} length of the collection
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._size = function()
{
    return this.collection.size();
};

/**
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @private
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype._totalSize = function()
{
    return this.collection['totalResults'] === undefined ? -1:this.collection['totalResults'];
};

//////////////////////////////////// Property Getters  /////////////////////////////////////
/**
 * Gets the collection property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype.getCollection = function()
{
    return this.collection;
};

/**
 * Gets the columns property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype.getColumns = function()
{
    return this.columns;
};

/**
 * Gets the rowHeader property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype.getRowHeader = function()
{
    return this.rowHeader;
};

/**
 * Gets the data property for testing
 * @export
 * @ignore
 * @memberof oj.CollectionDataGridDataSource
 */
oj.CollectionDataGridDataSource.prototype.getData = function()
{
    return this.data;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * This class contains all utility methods used by the data grid collection model.
 * @constructor
 * @private
 */
oj.CollectionDataGridUtils = function()
{
};

/**
 * Returns the key of the model. It is the id if one is set otherwise
 * it is the cId
 * @private
 * @param {Object} model the model to check for id and cid
 * @return {string} the id or cid for the model
 */
oj.CollectionDataGridUtils._getModelKey = function(model)
{
    var key;
    key = model.GetId();
    if (key == null)
    {
        key = model.GetCid();
    }
    return key;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * @export
 * @class oj.CollectionHeaderSet
 * @classdesc A HeaderSet represents a collection of headers.  The HeaderSet is an object returned by the success callback
 * of the fetchHeaders method on DataGridDataSource.  This implementation of HeaderSet is used by the
 * CollectionDataGridDataSource.
 * @param {number} start the start index of header set.
 * @param {number} end the end index of the header set.
 * @param {Array} headers the array of headers
 * @param {string=} rowHeader the id of the row header column.  Required for row headers.
 * @param {Object=} sortInfo the information about the sort direction and key
 * @constructor
 * @since 1.0
 * @hideconstructor
 * @ojtsignore
 * @see oj.CollectionDataGridDataSource 
 */
oj.CollectionHeaderSet = function(start, end, headers, rowHeader, sortInfo)
{
    // assert start/end are number
    oj.Assert.assertArrayOrNull(headers);

    this.m_start = start;
    this.m_end = end;
    this.m_headers = headers;
    this.m_rowHeader = rowHeader;
    this.m_sortInfo = sortInfo;    
};

/**
 * Sets the models used in this header set.
 * @param {Array} models an array of oj model for the header set
 * @private
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.setModels = function(models)
{
    oj.Assert.assertArray(models);
    // make sure the array size is valid
    if (models != null && models.length === this.getCount())
    {
        this.m_models = models;
    }
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the absolute index of the header in which we want to retrieve the header from.  
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {*} the data for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getData = function(index, level)
{
    var model;

    // make sure index/level are valid
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds'); 

    // row or column header
    if (this.m_rowHeader != null)
    {
        if (this.m_models == null)
        {
            return null;
        }

        model = this.m_models[index - this.m_start];
        return model.get(this.m_rowHeader);
    }
    else
    {
        return this.m_headers[index];
    }
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source can optionally return are:
 *  1) sortDirection - the initial sort direction of the header.  Valid values are "ascending" and "descending".
 *  2) key - the key of the row/column header.
 * @param {number} index the absolute index of the header in which we want to retrieve the metadata from.  
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} the metadata object for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getMetadata = function(index, level)
{
    var data, model;

    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds'); 
    
    // row header case
    if (this.m_rowHeader != null)
    {
        if (this.m_models == null)
        {
            return null;
        }

        model = this.m_models[index - this.m_start];        
        return {'key':oj.CollectionDataGridUtils._getModelKey(model)};
    }    
    else
    {
        data = this.getData(index, level);
        if (this.m_sortInfo['key'] === data)
        {
            return {'key': data, 'sortDirection':this.m_sortInfo['direction']};
        }
        return {'key': data};
    }    
    
};

/**
 * Gets the actual number of levels of the result set for the specified axis. The levels
 * are the counted from the outermost header indexed at 0, and moving inwards toward the 
 * databody would increment the level by 1.
 * @return {number} the number of levels of the result set
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getLevelCount = function()
{
    if (this.getCount() > 0)
    {
        return 1;
    }
    return 0;
};

/**
 * Gets the extent of an index on a particular level within the context of the headerSet. Extent is defined as the 
 * number of indexes covered by the header. If the extent extends beyond the start and end of the requested 
 * range the extent should be trimmed to the edge of the requested range and the object for {'more': {before, after}}
 * should have the value appropriate boolean set. For innermost headers the extent will always be 1.
 * @param {number} index the absolute index along the innermost header of the extent to get, 0 is the first header in the data source
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} an object containing two values
 *              extent: the number of absolute indexes spanned by the header at this index 
 *                      bounded by the edges of the result set for the specified axis. 
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is in the headerSet
 * @example <caption>In this example the header spans 5 indexes and there are more indexes to cover after the request that
 *              aren't included in this headerSet:</caption>                     
 * {'extent':5, 'more': {'before':false, 'after':true}}
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getExtent = function(index, level)
{ 
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds'); 
    return {'extent': 1, 'more':{'before': false, 'after':false}};
};

/**
 * Gets the depth of an index starting at a particular level. The depth is the number 
 * of levels spanned by the header.
 * @param {number} index the absolute index of the depth to get
 * @param {number=} level the level of the header, 0 is the outermost header
 * @return {number} the number of levels spanned by the header at the specified position
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getDepth = function(index, level)
{
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds');  
    return 1;
}

/**
 * Gets the actual count of the result set, the total indexes spanned by the headerSet
 * along the innermost header.
 * @return {number} the actual count of the result set.  
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getCount = function()
{
    return Math.max(0, this.m_end - this.m_start);
};

/**
 * For internal testing purposes. Gets the m_start property
 * @return {number} the start index of the result set.
 * @export
 * @ignore
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getStart = function()
{
    return this.m_start;
};

/**
 * For internal testing purposes. Gets the m_end property
 * @export
 * @ignore
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getEnd = function()
{
    return this.m_end;
};

/**
 * For internal testing purposes. Gets the m_headers property
 * @export
 * @ignore
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getHeaders = function()
{
    return this.m_headers;
};

/**
 * For internal testing purposes. Gets the m_rowHeader property
 * @export
 * @ignore
 * @memberof oj.CollectionHeaderSet
 */
oj.CollectionHeaderSet.prototype.getRowHeader = function()
{
    return this.m_rowHeader;
};
/*jslint browser: true,devel:true*/
/**
 * Implementation of PagingModel backed by an oj.Collection.  CollectionPagingDataSource
 * provides a window into the collection, presenting both a standard JavaScript array and Knockout observable array for consumption.
 * It is designed to feed a Knockout-bound HTML control, for example.
 * @export
 * @class oj.CollectionPagingDataSource
 * @classdesc Implementation of PagingModel using an oj.Collection
 * @extends oj.DataSource
 * @implements oj.PagingModel
 * @param {oj.Collection} collection the collection to use as source data for this paging control
 * @constructor
 * @since 1.0
 * @ojtsignore
 * @ojtsimport knockout
 */
oj.CollectionPagingDataSource = function(collection)
{
    this.collection = collection;
    // Start at the beginning
    this.current = 0;
    this.Init();  
    // Default the page size to 10
    var self = this;
    this.dataWindow = [];
    this._noInit = true;
    try {
        this._setPageSize(10);
    }
    finally {
        this._noInit = false;
    }
};

// Subclass from oj.DataSource 
oj.Object.createSubclass(oj.CollectionPagingDataSource, oj.DataSource, "oj.CollectionPagingDataSource");

/**
 * Initializes the instance.
 * @export
 * @ojtsignore
 * @memberof oj.CollectionPagingDataSource
 */
oj.CollectionPagingDataSource.prototype.Init = function()
{
  oj.CollectionPagingDataSource.superclass.Init.call(this);
};


oj.CollectionPagingDataSource.prototype._getSize = function() {
    if (!this._hasMore()) {
        // Return a size representing only what's left if we'd go over the bounds
        return this.totalSize() - this.current;
    }
    // Otherwise window size should match the page size
    return this.currentPageSize;
};

oj.CollectionPagingDataSource.prototype._refreshDataWindow = function() {
    // Reinit the array
    this.dataWindow = new Array(this._getSize());
    var self = this;
    return this.collection.IterativeAt(this.current, this.current + this.dataWindow.length).then(function(array) {
        // Copy from the source data
        for (var i = 0; i < array.length; i++) {
            self.dataWindow[i] = array[i];
        }
        // Update the observable array
        self._refreshObservableDataWindow();
        self._endIndex = self._startIndex + self.dataWindow.length - 1;
    });        
};

oj.CollectionPagingDataSource.prototype._refreshObservableDataWindow = function() {
    if (this.observableDataWindow !== undefined) {
        // Manage the observable window array
        this.observableDataWindow.removeAll();
        for (var i = 0; i < this.dataWindow.length; i++) {
            this.observableDataWindow.push(oj.KnockoutUtils.map(this.dataWindow[i]));
        }
    }
};

/**
 * @export
 * Return the current set of data in the paging window
 * 
 * @returns {Array.<Object>} the current set of data in the paging window
 * @memberof oj.CollectionPagingDataSource
 */
oj.CollectionPagingDataSource.prototype.getWindow = function() {
    return this.dataWindow;
};

/**
 * @export
 * Get the observable array representing the current set of data in the paging window
 * 
 * @returns {Object} an observable array representing the current data in the paging window
 * @ojsignature {target: "Type",
 *               value: "KnockoutObservableArray<object>",
 *               for: "returns"}
 * @memberof oj.CollectionPagingDataSource
 */
oj.CollectionPagingDataSource.prototype.getWindowObservable = function() {
    if (this.observableDataWindow === undefined) {
        this.observableDataWindow = ko.observableArray();
        this._refreshObservableDataWindow();
    }
    return this.observableDataWindow;
};

/**
 * Get the current page
 * @return {number} The current page
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.getPage = function()
{
  return this._page;
};

/**
 * Set the current page
 * @param {number} value The current page
 * @param {Object=} options Options<p>
 * @param {number} [options.pageSize] The page size.<p>
 * @return {Promise.<null>} promise object triggering done when complete..
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.setPage = function(value, options)
{
  options = options || {};
  value = parseInt(value, 10);
  try 
  {
    oj.CollectionPagingDataSource.superclass.handleEvent.call(this, oj.PagingModel.EventType['BEFOREPAGE'], {'page' : value, 'previousPage' : this._page});
  }
  catch (err)
  {
    return Promise.reject(null);
  }
  this.pageSize = options['pageSize'] != null ? options['pageSize'] : this.pageSize;
  options['startIndex'] = value * this.pageSize;
  var previousPage = this._page;
  this._page = value;
  this._startIndex = options['startIndex'];
  var self = this;
  
  return new Promise(function(resolve, reject)
  {
    self._fetchInternal(options).then(function(result)
    {
      oj.CollectionPagingDataSource.superclass.handleEvent.call(self, oj.PagingModel.EventType['PAGE'], {'page' : self._page, 'previousPage' : previousPage});
      resolve(null);
    }, function (error)
    {
      // restore old page
      self._page = previousPage;
      self._startIndex = self._page * self.pageSize;
      reject(null);  
    });
  });
};

/**
 * Get the current page start index
 * @return {number} The current page start index
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.getStartItemIndex = function()
{
  return this._startIndex;
};

/**
 * Get the current page end index
 * @return {number} The current page end index
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.getEndItemIndex = function()
{
  return this._endIndex;
};

/**
 * Get the page count
 * @return {number} The total number of pages
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.getPageCount = function()
{
  var totalSize = this.totalSize();
  return totalSize == -1 ? -1 : Math.ceil(totalSize / this.pageSize);
};

/**
 * Fetch the row data.
 * @param {Object=} options Options to control fetch
 * @param {number} options.startIndex The index at which to start fetching records.
 * @param {boolean} options.silent If set, do not fire a sync event.
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects and the startIndex triggering done when complete.<p>
 *         The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>startIndex</b></td><td>The startIndex for the returned set of rows</td></tr>
 * </tbody>
 * </table>  
 * @ojsignature {
 *          target: "Type",
 *          for: "returns",
 *          value: "void|Promise<void|{data: Array<object>, startIndex: number}>"}
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.fetch = function(options)
{
    var opts = options || {};
    
    if (opts['pageSize'] !== undefined && opts['startIndex'] !== undefined) 
    {
        if (!this._hasMore()) 
        {
            // At the limit
            this._processSuccess(null);
            return Promise.resolve();
        }
        this.currentPageSize = opts['startIndex'] + opts['pageSize'];
        var self = this;
        return this._refreshDataWindow().then(function() 
        {
                self._processSuccess(null);
        });
    }
    else
    {
      return this._fetchInternal(options);
    }
};

oj.CollectionPagingDataSource.prototype._fetchInternal = function(options)
{
    // No fetching, but set up window according to options
    var opts = options || {};
    if (opts['startIndex'] !== undefined) {
        this.current = opts['startIndex'];
    }
    if (opts['pageSize'] !== undefined) {
        this.pageSize = opts['pageSize'];
        this.currentPageSize = opts['pageSize'];
    }
    var self = this;
    return new Promise(function(resolve, reject)
    {
      // Call collection fetch, and refresh the window on success
      // Allow for the fact that this collection might not be backed by a rest service
      try {
          self.collection.fetch({success:function() {
              self._refreshDataWindow().then(function() {
                  self._processSuccess(opts);
                  resolve({'data': self.getWindow(), 'startIndex': self.current});});
          }});
      }
      catch (e) {
          self._refreshDataWindow().then(function() {
              self._processSuccess(opts);
              resolve({'data': self.getWindow(), 'startIndex': self.current});});
      }
    });
};

oj.CollectionPagingDataSource.prototype._processSuccess = function(opts) {
    var options = opts || {};
    if (!options['silent'])
    {
      this.handleEvent("sync", {'data': this.getWindow(), 'startIndex': this.current});
    }
    
    if (options['success']) {
        options['success']();
    }
/*    else if (this.fetchCallback) {
        this.fetchCallback();
    }*/
};
    
oj.CollectionPagingDataSource.prototype.handleEvent = function(eventType, event)
{
    return oj.CollectionPagingDataSource.superclass.handleEvent.call(this, eventType, event);
};

/**
 * Return whether there is more data which can be fetched.
 * @private
 * @memberof oj.CollectionPagingDataSource
 */
oj.CollectionPagingDataSource.prototype._hasMore = function()
{
  return this.current + this.currentPageSize < this.totalSize();
};

/**
 * Set or change the number of models in a page
 * 
 * @param {number} n page size
 * @private
 * @memberof oj.CollectionPagingDataSource
 */
oj.CollectionPagingDataSource.prototype._setPageSize = function(n)
{
  this.pageSize = n;
  this.currentPageSize = n;
/*  if (!this._noInit) {
    var self = this;
    return this._refreshDataWindow().then(function() {
                self._processSuccess(null)});
  }*/
};

oj.CollectionPagingDataSource.prototype.setFetchCallback = function(callback) {
    this.fetchCallback = callback;
};

/**
 * @export
 * Return the size of the data locally in the dataSource. -1 if an initial fetch has not been
 * done yet.
 * @returns {number} size of data
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.size = function()
{
  var w = this.getWindow();
  return w ? w.length : 0;
};

/**
 * @export
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.totalSize = function()
{
    return this.collection.length;
};

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @ojsignature {target:"Type",
 *               value: "'actual'|'estimate'|'atLeast'|'unknown'",
 *               for: "returns"}
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance 
 */
oj.CollectionPagingDataSource.prototype.totalSizeConfidence = function()
{ 
  return "actual";
};

/**
 * Determines whether this data source supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none".  
 *         Returns null if the feature is not recognized.
 * @export
 * @expose
 * @memberof oj.CollectionPagingDataSource
 * @instance
 */
oj.CollectionPagingDataSource.prototype.getCapability = function(feature)
{
    return null;
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @export
 * @class oj.CollectionTableDataSource
 * @extends oj.TableDataSource
 * @classdesc Object representing data available from an {@link oj.Collection} object, such as an external data source.  This data source can be used by [ListView]{@link oj.ojListView}, [NavigationList]{@link oj.ojNavigationList}, 
 *            [TabBar]{@link oj.ojTabBar}, and [Table]{@link oj.ojTable}.<br><br>
 *            See the <a href="../jetCookbook.html?component=table&demo=ojCollectionTable">Table - Using oj.Collection</a> demo for an example.<br><br>
 *            Refer to {@link oj.TableDataSource} for other data sources that represent tabular data.
 * @param {oj.Collection} data data supported by the components
 * @param {Object|null} [options] Options for the TableDataSource
 * @param {"enabled"|"disabled"} [options.startFetch] Control whether to start initial fetch when the TableDataSource is bound to a component.  Valid values are:<br><br>
 *                                    <b>"enabled"</b> (default) - Start initial fetch automatically when the TableDataSource is bound to a component.<br>
 *                                    <b>"disabled"</b> - Do not start initial fetch automatically.  Application will call the <a href="#fetch">fetch()</a> method to
 *                                                        start the first fetch.
 * @constructor
 * @ojtsignore
 * @since 1.0
 */
oj.CollectionTableDataSource = function(data, options)
{
  // Initialize
  this.data = {};   // This was put in to keep closure happy...
  if (!(data instanceof oj.Collection))
  {
    // we only support oj.Collection
    var errSummary = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_SUMMARY'];
    var errDetail = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_DETAIL'];
    throw new Error(errSummary + '\n' + errDetail);
  }
  
  oj.CollectionTableDataSource.superclass.constructor.call(this, data, options);

  this._collection = data;
  this._addCollectionEventListeners();
  
  this.Init();

  if ((options != null && (options['startFetch'] == 'enabled' || options['startFetch'] == null))
    || options == null)
  {
    this._startFetchEnabled = true;
  }
};

// Subclass from oj.DataSource 
oj.Object.createSubclass(oj.CollectionTableDataSource, oj.TableDataSource, "oj.CollectionTableDataSource");

/**
 * @export
 * @memberof oj.CollectionTableDataSource
 * @desc If set to a function(row1, row2), then this function is called comparing raw row data (see the
 * JavaScript array.sort() for details)
 * @ojsignature {target: "Type",
 *               value: "null|string|((param0: object, param1?: object)=> number|string|object)"}
 */
oj.CollectionTableDataSource.prototype.comparator = null;

/**
 * Initializes the instance.
 * @memberof oj.CollectionTableDataSource
 * @instance
 * @override
 * @protected
 */
oj.CollectionTableDataSource.prototype.Init = function()
{
  oj.CollectionTableDataSource.superclass.Init.call(this);
};

/**
 * Return the row data found at the given index.
 * 
 * @param {number} index Index for which to return the row data. 
 * @param {Object=} options Options to control the at.
 * @return {Promise} Promise resolves to a compound object which has the structure below. If the index is out of range, Promise resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @export
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.TableDataSource.RowData>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @expose
 * @memberof oj.CollectionTableDataSource
 * @instance
 */
oj.CollectionTableDataSource.prototype.at = function(index, options)
{
  options = options || {};
  options['deferred'] = true;
  var model = this._collection.at(index, options);
  var self = this;
  // the at() call can result in fetches
  self._isFetchingForAt = true;
  var row;
  return new Promise(function(resolve, reject) 
  {
    if (model != null)
    {
      model.then(function(resolvedModel)
      {
        self._isFetchingForAt = false;
        row = {'data': resolvedModel['attributes'], 'index': index, 'key': resolvedModel['id']};
        resolve(row);
      },
      function(e)
      {
        self._isFetchingForAt = false;
        oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['ERROR'], e);
        reject(e);
      });
    }
    else
    {
      resolve(null);
    }
  });
};

/**
 * Fetch the row data.
 * @param {Object=} options Options to control fetch
 * @param {number} [options.startIndex] The index at which to start fetching records.
 * @param {boolean} [options.silent] If set, do not fire a sync event.
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects, an array of ids, and the startIndex triggering done when complete.<p>
 *         The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>startIndex</b></td><td>The startIndex for the returned set of rows</td></tr>
 * </tbody>
 * </table>  
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.TableDataSource.RowDatas>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.CollectionTableDataSource
 * @instance
 */
oj.CollectionTableDataSource.prototype.fetch = function(options)
{
  options = options || {};
  var fetchType = options['fetchType'];
  
  if (fetchType == 'init' && !this._startFetchEnabled)
  {
    return Promise.resolve();
  }
  
  return this._fetchInternal(options);
};

/**
 * Return the first row data whose id value is the given id
 * @param {string} id ID for which to return the row data, if found. 
 * @param {Object=} options Options to control the get.
 * @return {Promise} Promise which resolves to a compound object which has the structure below where the id matches the given id. If none are found, resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @ojsignature {target:"Type",
 *               value: "Promise<null|oj.TableDataSource.RowData>",
 *               for: "returns",
 *               jsdocOverride: true}
 * @export
 * @expose
 * @memberof oj.CollectionTableDataSource
 * @instance
 */
oj.CollectionTableDataSource.prototype.get = function(id, options)
{
  options = options || {};
  options['deferred'] = true;
  var model = this._collection.get(id, options);
  
  var self = this;
  var row, wrappedRow;
  return new Promise(function(resolve, reject) 
  {
    if (model != null)
    {
      model.then(function(resolvedModel)
      {
        wrappedRow = self._wrapWritableValue(resolvedModel, resolvedModel['attributes']);
        row = {'data': wrappedRow, 'index': resolvedModel['index'], 'key': resolvedModel['id']};
        resolve(row);
      },
      function(e)
      {
        oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['ERROR'], e);
        reject(e);
      });
    }
    else
    {
      resolve(null);
    }
  });
};

/**
 * Performs a sort on the data source.
 * @param {Object} [criteria] the sort criteria.
 * @param {*} criteria.key The key that identifies which field to sort
 * @param {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @return {Promise.<null>} promise object triggering done when complete.
 * @export
 * @expose
 * @memberof oj.CollectionTableDataSource
 * @instance
 */
oj.CollectionTableDataSource.prototype.sort = function(criteria)
{   
  if (criteria == null)
  {
    criteria = this['sortCriteria'];
  }
  else
  {
    this['sortCriteria'] = criteria;
  }
  
  var comparator = this['comparator'];
  
  var self = this;
  return new Promise(function(resolve, reject) {
    
    if (comparator == null)
    {
      self._collection['comparator'] = criteria['key'];

      if (criteria['direction'] == 'ascending')
      {
        self._collection['sortDirection'] = 1;
      }
      else
      {
        self._collection['sortDirection'] = -1;
      }
    }
    else
    {
      self._collection['comparator'] = comparator;
    }
    self._collection.sort(null);
    var result = {'header': criteria['key'], 'direction': criteria['direction']};
    resolve(result);
  });    
};

/**
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @export
 * @expose
 * @memberof oj.CollectionTableDataSource
 * @instance
 */
oj.CollectionTableDataSource.prototype.totalSize = function()
{
  var totalSize = this._collection['totalResults'] >= 0 ? this._collection['totalResults'] : -1;
  
  if (totalSize > -1)
  {
    var size = this._collection.size();
    return size > totalSize ? size : totalSize;
  }
  else
  {
    if (this._fetchResultSize > 0)
    {
      totalSize = this._fetchResultSize;
    }
    else if (this.totalSizeConfidence() == "atLeast")
    {
      return this._collection.size();
    }
  }
  
  return totalSize;
};

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @ojsignature {target:"Type",
 *               value: "'actual'|'estimate'|'atLeast'|'unknown'",
 *               for: "returns"}
 * @export
 * @expose
 * @memberof oj.CollectionTableDataSource
 * @instance 
 */
oj.CollectionTableDataSource.prototype.totalSizeConfidence = function()
{ 
  if (this._collection['totalResults'] >= 0)
  {
    return "actual";
  }
  else if (this._collection['hasMore'])
  {
    // if totalResults is unknown but we know we have more then
    // we know we have at least a certain number of rows
    return "atLeast";
  }
  return "unknown";
};

/**
 * Add event listeners to the collection
 * @private
 * @memberof oj.CollectionTableDataSource
 */
oj.CollectionTableDataSource.prototype._addCollectionEventListeners = function()
{
  var self = this;
  this._collection.on(oj.Events.EventType['SYNC'], function(event) 
  {
    if (event instanceof oj.Model)
    {
      oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['CHANGE'], {'data': [event['attributes']], 'keys': [event['id']], 'indexes': [event['index']]});
    }
    else if (event instanceof oj.Collection)
    {
      if (!self._isFetchingForAt && !self._isFetching)
      {
        var startIndex = event['offset'];
        var pageSize = event['lastFetchCount'] || event['lastFetchSize'];

        // Do not call _getRowArray if this datasource is paged,
        // or if the underlying oj.Collection is virtual since _getRowArray
        // assumes collection.at returns Model objects, which is not the case
        // for virtual collection.
        if (pageSize > 0 || self._collection.IsVirtual())
        {
          self._startIndex = startIndex;
          self._pageSize = pageSize;
          var endIndex = 0;
          
          if (self._collection.totalResults > 0 || 
            self._collection.hasMore)
          {
            endIndex = startIndex + pageSize;
          }
          
          // paged fetch
          self._isFetchingForAt = true;
          event.IterativeAt(startIndex, endIndex).then(function(modelArray)
          {
            self._isFetchingForAt = false;
            var rowArray = [];
            var keyArray = [];
            var i, model, wrappedRow;
            for (i = 0; i < modelArray.length; i++)
            {
              if (modelArray[i] != null)
              {
                model = modelArray[i];
                wrappedRow = self._wrapWritableValue(model, model['attributes']);
                rowArray.push(wrappedRow);
                keyArray.push(model['id']);
              }
            }
            var result = {'data': rowArray, 'keys': keyArray, 'startIndex': startIndex};
            self._endFetch.call(self, {'silent': false}, result, null);
          });
        }
        else
        {
          var result = self._getRowArray();
          self._endFetch.call(self, {'silent': false}, result, null);
        }
      }
    }
  });
  this._collection.on(oj.Events.EventType['ALLADDED'], function(event, modelArray) {
    var rowArray = [];
    var keyArray = [];
    var indexArray = [];
    var i, model, wrappedRow;
    for (i = 0; i < modelArray.length; i++)
    {
      model = modelArray[i];
      wrappedRow = self._wrapWritableValue(model, model['attributes']);
      rowArray.push(wrappedRow);
      keyArray.push(model['id']);
      indexArray.push(model['index']);
    }
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['ADD'], {'data': rowArray, 'keys': keyArray, 'indexes': indexArray});
  });
  this._collection.on(oj.Events.EventType['ALLREMOVED'], function(event, modelArray) {
    var rowArray = [];
    var keyArray = [];
    var indexArray = [];
    var i, model;
    for (i = 0; i < modelArray.length; i++)
    {
      model = modelArray[i];
      // no need to wrapWritableValue as we are just deleting
      rowArray.push(model['attributes']);
      keyArray.push(model['id']);
      indexArray.push(model['index']);
    }
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['REMOVE'], {'data': rowArray, 'keys': keyArray, 'indexes': indexArray});
  });
  this._collection.on(oj.Events.EventType['RESET'], function(event) {
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['RESET'], event);
  });
  this._collection.on(oj.Events.EventType['SORT'], function(event, eventOpts) {
    if (eventOpts == null || !eventOpts['add'])
    {
      var sortCriteria = {};
      
      if (event != null && !event['comparator'] != null && !$.isFunction(event['comparator']))
      {
        sortCriteria['header'] = event['comparator'];
        sortCriteria['direction'] = event['sortDirection'] === 1 ? 'ascending' : 'descending';
      }
      oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['SORT'], sortCriteria);
    }
  });
  this._collection.on(oj.Events.EventType['CHANGE'], function(event) {
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['CHANGE'], {'data': [event['attributes']], 'keys': [event['id']], 'indexes': [event['index']]});
  });
  this._collection.on(oj.Events.EventType['DESTROY'], function(event) {
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['DESTROY'], event);
  });
  this._collection.on(oj.Events.EventType['REFRESH'], function(event) {
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['REFRESH'], event);
  });
  this._collection.on(oj.Events.EventType['ERROR'], function(collection, xhr, options) {
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['ERROR'], collection, xhr, options);
  });
  this._collection.on(oj.Events.EventType['REQUEST'], function(event) {
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['REQUEST'], event);
  });
};

oj.CollectionTableDataSource.prototype._fetchInternal = function(options)
{
  this._startFetch(options);
  options = options || {};
  var self = this;
  this._isPaged =  options['pageSize'] > 0 ? true : false;
  this._startIndex = options['startIndex'] == null ? this._startIndex : options['startIndex'];
  this._pageSize = options['pageSize'] > 0 ? options['pageSize'] : -1;
  options['pageSize'] = this._pageSize;
  options['startIndex'] = this._startIndex;
  options['refresh'] = true;

  return new Promise(function (resolve, reject)
  {
    var pageSize = self._pageSize;
    
    if (!self._isPaged)
    {
      // set an arbitrary page size for setRangeLocal. When non-virtual, will fetch everything anyway.
      pageSize = 25;
    }
    
    self._collection.setRangeLocal(self._startIndex, pageSize).then(function(actual) 
      {
        var result;
        
        // Do not call _getRowArray if this datasource is paged,
        // or if the underlying oj.Collection is virtual since _getRowArray
        // assumes collection.at returns Model objects, which is not the case
        // for virtual collection.
        if (self._isPaged || self._collection.IsVirtual())
        {
          var rowArray = [];
          var keyArray = [];
          var i,  model, wrappedRow;
          for (i = 0; i < actual['models'].length; i++)
          {
            model = actual['models'][i];
            wrappedRow = self._wrapWritableValue(model, model['attributes']);
            rowArray[i] = wrappedRow;
            keyArray[i] = model['id'];
          }
          result = {'data': rowArray, 'keys': keyArray, 'startIndex': self._startIndex};
          
          if (actual['models'].length < self._pageSize)
          {
            // it returned less than a page of data so we're at the end
            // set the totalSize if unknown
            if (self.totalSize() < 0)
            {
              self._fetchResultSize = self._startIndex + actual['models'].length;
            }
          }
          else
          {
            self._fetchResultSize = null;
          }
        }
        else
        {
          result = self._getRowArray();
        }

        self._endFetch.call(self, options, result, null);
        resolve(result);
      },
      function(error) 
      {
        self._endFetch.call(self, options, null, error);
        reject(error);
      });
  });
};

/**
 * Indicate starting fetch
 * @param {Object} options
 * @private
 * @memberof oj.CollectionTableDataSource
 */
oj.CollectionTableDataSource.prototype._startFetch = function(options)
{
  this._isFetching = true;
  if (!options['silent'])
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['REQUEST'], {'startIndex' : options['startIndex']});
  }
};

/**
 * Indicate ending fetch
 * @param {Object} options
 * @param {Object} result Result object
 * @param {*} error Error
 * @private
 * @memberof oj.CollectionTableDataSource
 */
oj.CollectionTableDataSource.prototype._endFetch = function(options, result, error)
{
  this._isFetching = false;
  if (error != null)
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['ERROR'], error);
  }
  else
  {
    if (!options['silent'])
    {
      oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['SYNC'],  result);
    }
  }
};

oj.CollectionTableDataSource.prototype._getRowArray = function()
{
  var endIndex = this._collection.size() - 1;
  var rowArray = [];
  var keyArray = [];
  var i, wrappedRow, model;
  for (i = 0; i <= endIndex; i++)
  {
    model = this._collection.at(i);
    wrappedRow = this._wrapWritableValue(model, model['attributes']);
    rowArray[i] = wrappedRow;
    keyArray[i] = model['id'];
  }
  return {'data': rowArray, 'keys': keyArray, 'startIndex': this._startIndex};
};

/**
 * Determines whether this TableDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none".  
 *         Returns null if the feature is not recognized.
 * @export
 * @expose
 * @memberof oj.CollectionTableDataSource
 * @instance
 */
oj.CollectionTableDataSource.prototype.getCapability = function(feature)
{
    return null;
};

oj.CollectionTableDataSource.prototype._wrapWritableValue = function(model, m)
{
  var returnObj = {};
  var prop;
  
  for (prop in m)
  {
    if (m.hasOwnProperty(prop))
    {
      (function()
      {
        var localProp = prop;
        var localModel = model;
        Object.defineProperty(returnObj, prop,
          {
            get: function()
            {
              return localModel.get(localProp);
            },
            set: function(newValue)
            {
              localModel.set(localProp, newValue, {'silent':true});
            },
            enumerable: true
          });
      })();
    }
  }
  
  return returnObj;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * A CollectionNodeSet represents a collection of nodes.  The CollectionNodeSet is an object returned by the success callback
 * of the fetchChildren method on CollectionTreeDataSource.  
 * @implements oj.NodeSet
 * @constructor
 * @since 1.0
 * @param {*} parentKey the key of the parent node
 * @param {oj.Collection} collection the Collection associated with this NodeSet
 * @param {Array.<oj.Model>} models a set of Model objects associated with this NodeSet
 * @param {oj.CollectionTreeDataSource} treeDataSource the TreeDataSource associated with this NodeSet
 * @param {number} start the index of the first node in this NodeSet relative to its parent node
 * @param {number} count the number of nodes in this NodeSet
 * @export
 * @ojtsignore
 */
oj.CollectionNodeSet = function(parentKey, collection, models, treeDataSource, start, count)
{
  this.parentKey = parentKey;
  this.collection = collection;
  this.models = models;
  this.childNodeSet = [];
  this.treeDataSource = treeDataSource;
  // Can't have start exceeding valid indices
  this.start = start < models.length ? start : (models.length === 0) ? 0 : models.length-1;
  // Unknown count: we'll assign it.  Also can't have count exceeding collection size
  this.count = count === -1 ? models.length : Math.min(models.length, count);
};

/**
 * @protected
 */
oj.CollectionNodeSet.prototype.FetchDescendants = function(callbacks) {
  this._fetchDescendants(this, true).then(function () {
    if (callbacks['success']) {
      callbacks['success']();
    }
  });
};

/**
 * @private
 */
oj.CollectionNodeSet.prototype._fetchDescendants = function(nodeSet, topLevel) {  
  return new Promise(function(resolve, reject) {
    var count = nodeSet.getCount();

    // Walk over each node in this node set, and fetch all the descendants of each
    function nextNode(index) {
      if (index < count) {
        nodeSet.FetchChildNodeSet(index, {'success':function(childNodeSet) {
                                            if (childNodeSet !== null) {
                                                nodeSet._fetchDescendants(childNodeSet, false).then(function() {
                                                    nextNode(index+1);
                                                });
                                            }
                                            else {
                                                nextNode(index+1);
                                            }
        }});
      }
      else {
        resolve(undefined);
      }
    }
    nextNode(0);
  });
};

/**
 * @protected
 */
oj.CollectionNodeSet.prototype.FetchChildNodeSet = function(index, callbacks) {
  var model = this.models[index];
  var parse = this.treeDataSource.parseMetadata(model);
  if (parse['leaf']) {
    // We're at the leaf: don't fetch any more
    this.childNodeSet[index] = null;
    callbacks['success'](null);
    return;
  }

  var collection = this.treeDataSource.GetChildCollection(model);
  var parentKey = this.treeDataSource.parseMetadata(model)['key'];
  var self = this;
  this.treeDataSource.FetchCollection(collection, 0, -1, {'success':function(nodeSet) {
    self.childNodeSet[index] = nodeSet;
    callbacks['success'](nodeSet);
  }}, parentKey); 
};

/**
 * @private
 */
oj.CollectionNodeSet.prototype._getCollection = function() {
  return this.collection;
};

/**
* Gets the parent key for this result set.  
* @return {*} the parent key for this result set. 
* @export
* @memberof oj.CollectionNodeSet
*/
oj.CollectionNodeSet.prototype.getParent = function()
{
    return this.parentKey;
};

/**
* Gets the start index of the result set.  
* @return {number} the start index of the result set.
* @export
* @memberof oj.CollectionNodeSet
*/
oj.CollectionNodeSet.prototype.getStart = function()
{
  return this.start;
};

/**
* Gets the actual count of the result set.  
* @return {number} the actual count of the result set.
* @export
* @memberof oj.CollectionNodeSet
*/
oj.CollectionNodeSet.prototype.getCount = function()
{
  return this.count;
};

/**
* Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
* 2) the index specified is out of bounds. 
* @param {number} index the index of the node/row in which we want to retrieve the data from.  
* @return {*} the data for the specified index.  oj.RowData should be returned for data that represents a row
*         with a number of columns.
* @export
* @memberof oj.CollectionNodeSet
*/
oj.CollectionNodeSet.prototype.getData = function(index)
{
  this._checkRange(index);
  return this.models[index].attributes;
};

/**
 * @private
 */
oj.CollectionNodeSet.prototype._checkRange = function(index) {
  if (index < this.start || index > this.start+this.count) {
    // Out of range
    throw "Out of range";
  }
};

/**
* Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
* 2) the index specified is out of bounds. 
* The metadata that the data source must return are:
*  1) key - Object, the key of the node/row.
*  2) leaf - boolean, true if it's a leaf, false otherwise. 
*  3) depth? - number, the depth of the node/row. (or should the caller just calculate it?)
* @param {number} index the index of the node/row in which we want to retrieve the metadata from.  
* @return {{key: *, leaf: boolean, depth: number}} the metadata object for the specific index.
* @export
* @memberof oj.CollectionNodeSet
*/	
oj.CollectionNodeSet.prototype.getMetadata = function(index)
{
  this._checkRange(index);

  var metadata = {'leaf': false, 'depth': -1};

  var model = this.models[index];

  var parse = this.treeDataSource.parseMetadata(model);
  metadata['key'] = parse['key'];
  metadata['leaf'] = parse['leaf'];
  metadata['depth'] = parse['depth'];

  return metadata;
};

/**
 * Gets the node set child of the specified index.
 * @param {number} index the index of the node/row in which we want to retrieve the child node set
 * @return {oj.CollectionNodeSet|null} the child node set representing the child collection.
 * @export
 * @memberof oj.CollectionNodeSet
 */
oj.CollectionNodeSet.prototype.getChildNodeSet = function(index) {
  if (this.treeDataSource._virtual) {
    return null;
  }

  this._checkRange(index);

  return this.childNodeSet[index];
};
/**
 * @class oj.CollectionTreeDataSource
 * @classdesc TreeDataSource implementation that represents hierachical data available from an {@link oj.Collection} object, such as an external data source.  This data source can be used by [ListView]{@link oj.ojListView}, 
 *            [NavigationList]{@link oj.ojNavigationList}, and [TreeView]{@link oj.ojTreeView}.<br><br>
 *            See the <a href="../jetCookbook.html?component=treeView&demo=collection">Tree View - Data Source: Collection</a> demo for an example.<br><br>
 *            Refer to {@link oj.TreeDataSource} for other data sources that represent hierarachical data.
 * @param {Object=} options an object containing optional properties.
 * @property {oj.Collection=} options.root an oj.Collection specifying the root level Collection
 * @property {function(oj.Collection,oj.Model):oj.Collection=} options.childCollectionCallback callback to return a child collection given a root and model representing the parent
 * @property {function(oj.Model):{key: *, leaf: boolean, depth: number}=} options.parseMetadata callback to return key, leaf, depth metadata from a given Model
 * @constructor
 * @since 1.0
 * @export
 * @extends oj.TreeDataSource
 * @ojtsignore
 */
oj.CollectionTreeDataSource = function(options)
{
  options = options || {};
  this.rootCollection = options['root'];
  this.childCollectionCallback = options['childCollectionCallback'];
  this.parseMetadata = options['parseMetadata'];
  this.sortkey = null;
  this.sortdir = "none";
  // This is a parent->collection cache
  this.cache = {};

  // Once we've seen a virtual collection, set this flag to know
  this._virtual = false;

  oj.CollectionTreeDataSource.superclass.constructor.call(this);
};

// Default implementation
/**
 * @private
 */
oj.CollectionTreeDataSource.prototype.parseMetadata = function(model) {    
  return {'key':model['idAttribute']+"="+model['id']};
};

// Subclass from oj.TreeDataSource
oj.Object.createSubclass(oj.CollectionTreeDataSource, oj.TreeDataSource, "oj.CollectionTreeDataSource");

/**
 * Initializes the data source.
 * @memberof oj.CollectionTreeDataSource
 * @return {void}
 * @ojtsignore
 * @export
 */
oj.CollectionTreeDataSource.prototype.Init = function()
{
  // super
  oj.CollectionTreeDataSource.superclass.Init.call(this);
};


/**
 * Returns the number of children for a specified parent.  If the value returned is not >= 0 then it is automatically assumed
 * that the child count is unknown.
 * @param {*} parent the parent key.  Specify null if inquiring child count of the root.
 * @return {number} the number of children for the specified parent.
 * @export
 * @memberof oj.CollectionTreeDataSource
 */
oj.CollectionTreeDataSource.prototype.getChildCount = function(parent)
{
  var childColl = this.__getParentsChildCollectionFromCache(parent);
  if (childColl) {
    if (childColl.length > 0) {
      return childColl.length;
    }
  }
/*  this.getChildCollection(parent, {'success': function(coll) {
    return coll.length;
  }});
*/
  return -1;
};

/**
 * Gets a collection representing the specified parent key's children.
 * @param {*} key the parent key in which to create an OJ collection for.
 * @param {Object} callbacks success and error callbacks.  The success callback will provide the child collection as the first argument
 * @property {function(oj.Collection):void} callbacks.success the callback to invoke when child collection is fetched successfully.
 * @property {function({status: *})=} callbacks.error the callback to invoke when the fetch failed.
 * @return {void}
 * @export
 * @memberof oj.CollectionTreeDataSource
 */
oj.CollectionTreeDataSource.prototype.getChildCollection = function(key, callbacks) {
    this.fetchChildren(key, null, {'success':function(nodeSet) {
            callbacks['success'](nodeSet._getCollection());
    }, 'error':callbacks['error']});
};

/**
 * Fetch the children
 * @param {*} parent the parent key.  Specify null if fetching children from the root.
 * @param {Object} range information about the range, it must contain the following properties: start, count
 * @property {number} range.start the start index of the range in which the children are fetched
 * @property {number} range.count the size of the range in which the children are fetched
 * @param {Object} callbacks the callbacks to be invoke when fetch children operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function(oj.CollectionNodeSet):void} callbacks.success the callback to invoke when fetch completed successfully.
 * @property {function({status: *})=} callbacks.error the callback to invoke when fetch children failed.
 * @param {Object=} options optional parameters for this operation.  Currently this is used for future expansion only.
 * @return {void}
 * @export
 * @memberof oj.CollectionTreeDataSource
 */
oj.CollectionTreeDataSource.prototype.fetchChildren = function(parent, range, callbacks, options)
{
    range = range || {};
    var start = range['start'] ? range['start'] : 0;
    var count = range['count'] ? range['count'] : -1;
    
    if (parent === null) {
      // Do root
      this.FetchCollection(null, start, count, callbacks, null);
      return;
    }
    // Use child collection callback to set up child collection, then fetch it
    var self = this;
    this._getModelForId(this.rootCollection, start, count, parent, 0).then(function(parentModel) {
    if (parentModel) {
      var collection = self.GetChildCollection(parentModel.model);
      try {
        self.FetchCollection(collection, start, count, callbacks, parent);
      }
      catch (error)
      {
        if (callbacks && callbacks['error']) {
          callbacks['error']({'status': error['message']});
        }
      } 
    }
    else {
      // Couldn't find a model for the given Id: call the error if present
      if (callbacks && callbacks['error']) {
        callbacks['error'](parent);
      }
    }
 });
};


/**
 * Called by common model when a model is added to a collection
 * @private
 */
oj.CollectionTreeDataSource.prototype.ModelAdded = function(model, collection, options) {
    var index = 0;
    if (options && options['at']) {
        index = options['at'];
    }
    var parents = this._getParentChain(collection);
    var parent = parents != null && parents.length > 0 ? parents[parents.length-1] : null;
    var event = this._createEvent(this, "insert", index, parents, this._putModelInNodeSet(parent, model));
    this.handleEvent("change", event);
};

/**
 * Called by common model when a model is removed from a collection
 * @private
 */
oj.CollectionTreeDataSource.prototype.ModelRemoved = function(model, collection, options) {
    var index = 0;
    if (options && options['index']) {
        index = options['index'];
    }
    this._removeCollectionFromCache(model);
    var event = this._createEvent(this, "delete", index, this._getParentChain(collection), null);
    this.handleEvent("change", event);
};

/**
 * Called by common model when a model is updated
 * @private
 */
oj.CollectionTreeDataSource.prototype.ModelUpdated = function(model, options) {
    var collectionForModel = model['collection'];
    var index = model.GetIndex();
    var parents = null;
    if (collectionForModel) {
        // If this model's collection is found, get its parent chain from the actual collection
        parents = this._getParentChain(collectionForModel);
    }
    var parent = parents != null && parents.length > 0 ? parents[parents.length-1] : null;
    var event = this._createEvent(this, "update", index, parents, this._putModelInNodeSet(parent, model));
    this.handleEvent("change", event);
};

/**
 * Called if a collection is refreshed
 * @protected
 */
oj.CollectionTreeDataSource.prototype.CollectionRefreshed = function(collection, resp, options) {
    var event = this._createEvent(this, "refresh", null, this._getParentChain(collection), null);
    this.handleEvent("refresh", event);
};

/**
 * @private
 */
oj.CollectionTreeDataSource.prototype._putModelInNodeSet = function(parent, model) {
/*    var rows = [];
    rows.push(model.attributes);
    var options = {};
    options['idAttribute'] = model['idAttribute'];    
    var rowset = new oj.ArrayRowSet(rows, options);
    rowset.fetch();
    return rowset;*/
    var collection = new oj.Collection();
    collection.add(model);
    return this._getNodeSet(collection, parent, 0, 1, [model]);    
};

/**
 * Return an array of parent keys representing the parentage of the given collection.  Relies on cached collection fetches.  Array is from
 * child key of root on down to collection's parent
 * @private
 */
oj.CollectionTreeDataSource.prototype._getParentChain = function(collection) {
    var parents = [];
    var parent = null;
    var coll = collection;
    do {
        parent = this._getParentOfCollection(coll);
        if (parent !== null) {
            if (parent !== oj.CollectionTreeDataSource.ROOT_CACHE_KEY) {
                parents.unshift(parent);
            }
            coll = this._getCollectionOfKey(parent);
        }
    } while (parent != null);
    return parents;
};

// Generate the cache key
oj.CollectionTreeDataSource.ROOT_CACHE_KEY = "%!@ROOT%#@!";
/**
 * @private
 */
oj.CollectionTreeDataSource.prototype._getCacheKey = function(model) {
    // If model (parent) is null, look for the root
    // If model is not an oj.Model, just use it as the key
    var key = model instanceof oj.Model ? this.parseMetadata(model)['key'] : model;
    return model ? key : oj.CollectionTreeDataSource.ROOT_CACHE_KEY;
};

/**
 * Return a cached collection given a parent Model.  Null if not cached
 * @private
 */
oj.CollectionTreeDataSource.prototype.__getParentsChildCollectionFromCache = function(model) {
    return this.cache[this._getCacheKey(model)];
};

/**
 * Put a collection into the cache whose parent is represented by model
 * @private
 */
oj.CollectionTreeDataSource.prototype._setCollectionInCache = function(model, collection) {
  // Tack on listeners
  collection.on(oj.Events.EventType['ADD'], this.ModelAdded, this);
  collection.on(oj.Events.EventType['REMOVE'], this.ModelRemoved, this);
  collection.on(oj.Events.EventType['CHANGE'], this.ModelUpdated, this);
  collection.on(oj.Events.EventType['SYNC'], this.CollectionRefreshed, this);
  var key = this._getCacheKey(model);
  this.cache[key] = collection;
  return this.cache[key];
};

/**
 * Remove any collections with the given model as a parent
 * @private
 */
oj.CollectionTreeDataSource.prototype._removeCollectionFromCache = function(model) {
  var key = this._getCacheKey(model);
  for (var prop in this.cache) {
    if (this.cache.hasOwnProperty(prop)) {
      if (prop === key) {
        // Found it: remove listeners and delete it
        this.cache[key].off(null, null, this);
        delete this.cache[key];
        return;
      }
    }
  }
};

/**
 * Determine if a model representing a given parent key is found in the given collection
 * The expectation is that the model would have to have been locally fetched
 * @private
 */
oj.CollectionTreeDataSource.prototype._keyInCollection = function(key, collection) {
    var count = collection['length'];
    var model = null;
    for (var i = 0; i < count; i++) {
        model = collection['models'][i];
        if (model) {
          var currKey = this._getCacheKey(model);
          if (key === currKey) {
              return true;
          }
        }
    }
    return false;
};

/**
 * Get the collection containing the given key
 * @private
 */
oj.CollectionTreeDataSource.prototype._getCollectionOfKey = function(key) {
  // Search the cache for the parent key
  for (var prop in this.cache) {
    if (this.cache.hasOwnProperty(prop)) {
      var collection = this.cache[prop];
      if (this._keyInCollection(key, collection)) {
        return collection;
      }
    }
  }
  return null;    
};

/**
 * Get the parent key for the given collection from the cache, if found
 * @private
 */
oj.CollectionTreeDataSource.prototype._getParentOfCollection = function(collection) {
  // Search the cache for the parent key
  for (var prop in this.cache) {
    if (this.cache.hasOwnProperty(prop)) {
      if (this.cache[prop] === collection) {
        return prop;
      }
    }
  }
  return null;
};

/**
 * Return an object giving the child collection for the given model along with a boolean indicating whether the collection was found 
 * in cache. 
 * @protected
 */
oj.CollectionTreeDataSource.prototype.GetChildCollection = function(parentModel) {    
  // Is it in the cache?
  var cached = true;
  var collection = this.__getParentsChildCollectionFromCache(parentModel);
  if (!collection) {
    // Nothing found at all yet
    cached = false;
    // Ask the app for the child collection for the given model
    collection = this.childCollectionCallback(this.rootCollection, parentModel);
    if (collection != null) {
      // Make sure we transfer sorting properties...
      this._applySortToCollection(collection);
      // And put it in the cache
      this._setCollectionInCache(parentModel, collection);
    }
  }

  return {collection:collection,cached:cached};
};

/**
 * @private
 */
oj.CollectionTreeDataSource.prototype._createEvent = function(source, operation, index, parent, data) {
    return {'source':source, 'operation':operation, 'index':index, 'parent':parent, 'data':data};
};

/**
 * Call to fetch models for an already-created collection--used only by fetchDescendants
 * @protected
 */
oj.CollectionTreeDataSource.prototype.FetchCollection = function(collectionObj, start, count, callbacks, parent) {
  var self = this;
  if (collectionObj === null) {
    // Fetch the root
    // Check for the root in the cache
    collectionObj = this.__getParentsChildCollectionFromCache(null);
    if (!collectionObj) {
      // Not found, put a representation in there
      collectionObj = {collection:self.rootCollection,cached:false};
      self._setCollectionInCache(null, this.rootCollection);
    }
    else {
      // Root was found precached: set up its information
      collectionObj = {collection:collectionObj, cached:true};
    }
  }
  if (collectionObj) {
    self._fetch(collectionObj, start, count, function (coll, models) {
      // Check for virtual
      if (coll.IsVirtual()) {
        self._virtual = true;
      }
      if (callbacks['success']) {
        // return a nodeset version of this fetched collection via the callback
        callbacks['success'](self._getNodeSet(coll, parent, start, count, models));
      }
    }, callbacks['error']);    
  }
};

/**
 * Turn collection into a CollectionNodeSet
 * @private
 */
oj.CollectionTreeDataSource.prototype._getNodeSet = function(collection, parent, start, count, models) {
  return new oj.CollectionNodeSet(parent, collection, models, this, start, count);
};

/**
 * Do any of the models in this collection or its children have the given key
 * @private
 */
oj.CollectionTreeDataSource.prototype._scanForKey = function(collection, key) {
    // Search collection for key
    var self = this;
    var models = new Array(collection['length']);
 
    return new Promise(function(resolve, reject) {
      function checkNext(index, collection, key) {
        if (index < collection['length']) {
          // Get the model at location index
          collection.at(index, {'deferred':true}).then(function (model) {
            // Store the model
            models[index] = model;
            
            // Does the retrieved model match the key when parsed out?  If so return it
            if (model) {
              var parse = self.parseMetadata(model);
              if (key === parse['key']) {
                  resolve({model: model, models: models});
                  return;
              }
            }
            // No model or no match: go to the next model in the collection
            index++;
            checkNext(index, collection, key);
         });
        }
        else {
          // Hit the end: not found
          resolve({model: null, models: models});
          return;
        }
      };

      // Start checking at the 0th model in the collection
      checkNext(0, collection, key);
    });
};

/**
 * Returns a promise with an object giving the model that corresponds to the given key, and the depth in the hierarchy at which it was found
 * @private 
 */
oj.CollectionTreeDataSource.prototype._getModelForId = function(collection, start, count, key, depth) {
  var self = this;

  return new Promise(function(resolve, reject) {
    // Does the starting collection contain the key?
    self._scanForKey(collection, key).then(function (obj) {
      // Yes, resolve (return it)
      if (obj.model) {
          resolve({model:obj.model, depth: depth});
          return;
      }
      // It does not
      // Wrong collection: search each model's child collections, and check recursively
      // Check each model in collection now for key
      function getNextCollection(index, models, tds) {
        if (index < models.length) {
          // Return an object containing the child collection of the current collection + whether its cached
          var childColl = tds.GetChildCollection(models[index]);
          if (!childColl.collection) {
            // No child collection found: move on to the next model in the current collection
            index++;
            getNextCollection(index, models, tds);
          } else {
            // Fetch the child collection if necessary
            tds._fetch(childColl, start, count, function(fetchColl, nextModels) {
              // Now check the child collection recursively for the key, moving depth down 1
              tds._getModelForId(fetchColl, start, count, key, depth+1).then(function(childModel) {
                // Found the model in this collection: unwind
                if (childModel) {
                  // Found somewhere down this tree
                  resolve(childModel);
                  return;
                } else {
                  // Not found down that tree
                  // Move to the next model
                  index++;
                  getNextCollection(index, models, tds);
                }
              });                                
            }, null);
          }
        }
        else {
          // Hit the end
          resolve(null);
        }
      }
      // Start checking the collection at the 0th model
      getNextCollection(0, obj.models, self);
    });
  });
};

/**
 * @private
 */
oj.CollectionTreeDataSource.prototype._getModelsFromCollection = function(collection) {
  function loop(collection, a, b, models) {
    return new Promise(function(allresolve, allreject) {
      var i;
      var doTask = function(index) {
                     var promise = new Promise(function(resolve, reject) {
                     collection.at(index).then(function(model) {
                       models[index] = model;
                       resolve(index+1);
                     }, reject);
                     });
                     return promise;
                  };

      var currentStep = Promise.resolve(0);

      for (i = a; i < b; i++) {
        currentStep = currentStep.then(doTask);
      }
      return currentStep.then(allresolve, allreject);
    });
  };

  if (collection.IsVirtual()) {
    var models = new Array(collection['length']);
    return new Promise(function(resolve, reject) {
      loop(collection, 0, collection['length'], models).then(function() {
        resolve(models);
      });
    });
  }
  else {
    return new Promise(function(resolve, reject) {
        resolve(collection['models']);
    });      
  }  
};

/**
 * Do a fetch or just return the collection if it came from cache
 * @private
 */
oj.CollectionTreeDataSource.prototype._fetch = function(collectionCacheObj, start, count, success, error) {    
  var self = this;
  var cached = collectionCacheObj.cached;
  if (cached) {
    // If it's cached, it's fetched
    this._getModelsFromCollection(collectionCacheObj.collection).then(function(models) {
      success(collectionCacheObj.collection, models);
    });
  }
  else {
    // apply sorting if necessary
    if (this.sortkey && this.sortkey !== "none") {
      collectionCacheObj.collection.comparator = this.sortkey;
      collectionCacheObj.collection.sortDirection = this.sortdir;
    }
    if (collectionCacheObj.collection['length'] > 0 || !collectionCacheObj.collection.IsUrlBased(null)) {
      // Already fetched: just return        
      success(collectionCacheObj.collection, collectionCacheObj.collection['models']);
      return;
    }
    // Do a real fetch from the collection's url
    if (count === -1) {
      // Don't want any records thrown out if virtual--rely on the fetched models to be there once cached
      collectionCacheObj.collection['modelLimit'] = -1;
      collectionCacheObj.collection.fetch({'success':function(fetchColl) {
        // Tack on handlers
        self._getModelsFromCollection(fetchColl).then(function(models) {
          success(fetchColl, models);
        });
      }, 'error':error});
    }
    else {
      // Ensure a range if we have one
      // Don't want any records thrown out if virtual--rely on the fetched models to be there once cached
      collectionCacheObj.collection['modelLimit'] = -1;      
      collectionCacheObj.collection.setRangeLocal(start, count).then(function(results) {
        collectionCacheObj.models = results['models'];
        success(collectionCacheObj.collection, results['models']);
      });
    }
  }
};

/**
 * Fetch all children and their children recursively from a specified parent.
 * @param {*} parent the parent key.  Specify null to fetch everything from the root (i.e. expand all)
 * @param {Object} callbacks the callbacks to be invoke when fetch children operation is completed.  The valid callback types are "success" and "error".
 * @property {function(oj.CollectionNodeSet):void} callbacks.success the callback to invoke when fetch completed successfully.
 * @property {function({status: *})=} callbacks.error the callback to invoke when fetch children failed.
 * @param {Object=} options optional parameters for this operation
 * @property {number=} options.start the index related to parent in which to begin fetching descendants from.  If this is not specified, then value zero will be used
 * @property {number=} options.maxCount the maximum number of children to fetch.  If a non-positive number is specified, then the value is ignored and
 *        there is no maximum fetch count
 * @return {void}
 * @export
 * @memberof oj.CollectionTreeDataSource
 */
oj.CollectionTreeDataSource.prototype.fetchDescendants = function(parent, callbacks, options)
{
  // No fetchDescendants for virtual
  if (this._virtual) {
    oj.Assert.failedInAbstractFunction();
  }
    var self = this;
    if (parent === null) {
      // Do root
      this.FetchCollection(null, 0, -1, {'success':function(nodeSet) {
                                                  nodeSet.FetchDescendants({'success':function() {
                                                    if (callbacks['success']) {
                                                      callbacks['success'](nodeSet);
                                                    }
                                                  }});
      }}, null);
      return;
    }
    // Use child collection callback to set up child collection, then fetch it
    this._getModelForId(this.rootCollection, 0, -1, parent, 0).then(function(parentModel) {
      if (parentModel) {
        var collection = self.GetChildCollection(parentModel.model);
        self.FetchCollection(collection, 0, -1, {'success':function(nodeSet) {
          nodeSet.FetchDescendants({'success':function() {
            if (callbacks['success']) {
                callbacks['success'](nodeSet);
            }
          }});                                                                    
        }}, parent);
      }
   });
};

/**
 * Performs a sort operation on the tree data.
 * @param {Object} criteria the sort criteria.  It must contain the following properties: key, direction
 * @param {*} criteria.key the key identifying the attribute (column) to sort on
 * @param {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object} callbacks callbacks for the sort operation
 * @property {function():void} callbacks.success the callback to invoke when the sort completed successfully
 * @property {function({status: *})=} callbacks.error the callback to invoke when sort failed.
 * @return {void}
 * @export
 * @memberof oj.CollectionTreeDataSource
 */
oj.CollectionTreeDataSource.prototype.sort = function(criteria, callbacks)
{
  var key = criteria['key'];
  var dir = criteria['direction'];
  var needSort = false;
  if (key !== this.sortkey) {
    this.sortkey = key;
    needSort = true;
  }
  if (dir !== this.sortdir) {
    this.sortdir = dir;
    needSort = true;
  }
  if (needSort) {
    // If clearing, clear the cache
    if (this.sortdir === "none") {
      this.cache = {};
    }
    // Go off and sort each collection, as long as it has "criteria" as one of its attributes
    for (var prop in this.cache) {
      if (this.cache.hasOwnProperty(prop)) {
        var collection = this.cache[prop];
        this._applySortToCollection(collection);
      }
    }
  }
  if (callbacks && callbacks['success']) {
    callbacks['success']();
  }
};

/**
 * @private
 */
oj.CollectionTreeDataSource.prototype._applySortToCollection = function(collection) {
  collection['comparator'] = this.sortkey;
  collection['sortDirection'] = (this.sortdir === "ascending") ? 1 : -1;
  collection.sort();    
};

/**
 * Returns the current sort criteria of the tree data.
 * @return {{Object}} the current sort criteria.  It should contain the following properties: key, direction where
 *         criteria.key the key identifying the attribute (column) to sort on.  Value is null if it's not sorted.
 *         criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @ojsignature {target: "Type",
 *               value: "{key: any, direction: 'ascending'|'descending'|'none'}",
 *               for: "returns"}
 * @export
 * @memberof oj.CollectionTreeDataSource
 */
oj.CollectionTreeDataSource.prototype.getSortCriteria = function()
{
  return {'key': this.sortkey, 'direction': this.sortdir};
};

/**
 * Moves a row from one location to another (different position within the same parent or a completely different parent)
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference row's current children.
 * @param {Object} callbacks the callbacks for the move function
 * @param {function():void} callbacks.success the callback to invoke when the move completed successfully.  
 * @param {function({status: *})=} callbacks.error the callback to invoke when move failed.
 * @return {void}
 * @export
 * @memberof oj.CollectionTreeDataSource
 */ 
oj.CollectionTreeDataSource.prototype.move = function(rowToMove, referenceRow, position, callbacks)
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference row's current children.
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @ojsignature {target: "Type",
 *               value: "'valid'|'invalid'",
 *               for: "returns"}
 * @export
 * @memberof oj.CollectionTreeDataSource
 */ 
oj.CollectionTreeDataSource.prototype.moveOK = function(rowToMove, referenceRow, position)
{
  return "invalid";
};

/**
 * Determines whether this CollectionTreeDataSource supports the specified feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the valid features "sort", 
 *        "move", "fetchDescendants", "batchFetch"
 * @return {string|null} the name of the feature.  Returns null if the feature is not recognized.
 *         For "sort", the valid return values are: "default", "none".  
 *         For "fetchDescendants", the valid return values are: "enable", "disable", "suboptimal".  
 *         For "move", the valid return values are: "default", "none".  
 *         For "batchFetch", the valid return values are: "enable", "disable".  
 * @export
 * @memberof oj.CollectionTreeDataSource
 */
oj.CollectionTreeDataSource.prototype.getCapability = function(feature)
{
  if (feature === "sort") {
    return "default";
  }
  if (feature === "move") {
    return "none";
  }
  if (feature === "batchFetch") {
    return "disable";
  }
  if (feature === "fetchDescendants") {
    return "disable";
  }
  return null;
};

/**
 * Copyright (c) 2016, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*
 * Portions of this component are based on  
 * TinyColor v1.3.0
 * https://github.com/bgrins/TinyColor
 * Brian Grinstead, MIT License
 */

(function()
{
  var trimLeft    = /^\s+/,
      trimRight   = /\s+$/,
      tinyCounter = 0,
      mathRound   = Math.round,
      mathMin     = Math.min,
      mathMax     = Math.max;

/**
 * @typedef {Object} oj.Color.RGBA
 * @property {number} r the red value
 * @property {number} g the green value
 * @property {number} b the blue value
 * @property {number} [a] the optional alpha value
 */
/**
 * @typedef {Object} oj.Color.HSLA
 * @property {number} h the hue value
 * @property {number} s the saturation value
 * @property {number} l the luminosity or lightness value
 * @property {number} [a] the optional alpha value
 */
/**
 * @typedef {Object} oj.Color.HSVA
 * @property {number} h the hue value
 * @property {number} s the saturation value
 * @property {number} v the value
 * @property {number} [a] the optional alpha value
 */
/**
 * @class oj.Color
 * @since 3.0.0
 * @classdesc Immutable object representing a color.
 * @ojshortdesc Object representing a color definition.
 * @desc Creates an object representing a color. The color may be defined using
 * the RGB, HSL, and HSV model values as an object, or as a CSS3 color specification string (refer to 
 * {@link https://developer.mozilla.org/en-US/docs/Web/CSS/color_value}, and also
 * {@link https://www.w3.org/TR/css3-color/#svg-color} for color visualization.  The
 * CSS3 named colors are available as oj.Color constants (e.g. oj.Color.ALICEBLUE)
 *
 * @example <caption>Typical color definitions</caption>
 *  oj.Color.RED;
 *  oj.Color.ALICEBLUE;
 *  new oj.Color({r:30, g:128, b:201});
 *  new oj.Color({r:30, g:128, b:201, a:0.8});
 *  new oj.Color({h:310, s:50, l:80});
 *  new oj.Color({h:310, s:50, l:80, a:0.8});
 *  new oj.Color({h:310, s:50, v:80});
 *  new oj.Color({h:310, s:50, v:80, a:0.8});
 *  new oj.Color('#4bc');
 *  new oj.Color('#44ccbb');
 *  new oj.Color('rgb(27,128,254)');
 *  new oj.Color('rgba(27,128,254,0.8)');
 *  new oj.Color('hsl(87, 100%,50%)');
 *  new oj.Color('hsla(87, 100%,50%, 0.5)');
 *  new oj.Color('hsv(0, 100%, 100%)') ;
 *  oj.Color.TRANSPARENT;
 *
 * @example <caption>Using a converter to obtain a different color format</caption>
 * var cf       = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_COLOR);
 * var convHsl  = cf.createConverter({"format": "hsl"}) ;
 *
 * var c        = new oj.Color('rgb(0, 191, 255)');
 * var s        = convHsl.format(c) ;            // returns "hsl(197, 71%, 73%)"

 * </p></br>
 * @param {string | Object} color 
 * <ul><li> A valid CSS3 color specification string (refer to 
 *  {@link https://developer.mozilla.org/en-US/docs/Web/CSS/color_value} for syntax)</li>
 * <li>An object containing one of the following groups of fields:
 * <ul>
 *  <li>r: &nbsp; the red value</li>
 *  <li>g: &nbsp; the green value</li>
 *  <li>b: &nbsp; the blue value</li>
 *  <li>a: &nbsp; the alpha value (optional)</li>
 * </ul>
 * <br>
 * <ul>
 *  <li>h: &nbsp; the hue value</li>
 *  <li>s: &nbsp; the saturation value</li>
 *  <li>l: &nbsp; the luminosity or lightness value</li>
 *  <li>a: &nbsp; the alpha value (optional)</li>
 * </ul>
 * <br>
 * <ul>
 *  <li>h: &nbsp; the hue value</li>
 *  <li>s: &nbsp; the saturation value</li>
 *  <li>v: &nbsp; the value</li>
 *  <li>a: &nbsp; the alpha value (optional)</li>
 * </ul>
 * </li></ul>
 * @ojsignature {target: "Type", 
 *               value: "string|oj.Color.RGBA|oj.Color.HSLA|oj.Color.HSVA",
 *               for: "color",
 *               jsdocOverride: true}
 * @constructor
 * @throws {Error} if the color specification cannot be parsed correctly.
 * @export
 */
oj.Color = function (color) {

    var rgb, t ;

    color = (color) ? color : '';
    rgb = inputToRGB(color);

    this._r = rgb.r,
    this._g = rgb.g,
    this._b = rgb.b,
    this._a = Math.round(100*rgb.a)/100;
};

/**
 * Returns the red channel value of the color.
 * @param {boolean=} doNotRound  Omit or set to false to return an integer value.
 *                               Set to true to return the possible fractional value.
 *                               (Fractional values for the red, green, or blue channels can
 *                               arise when a color is defined using the "hsl" format.)
 * @return {number} The red channel value in the range [0,255].
 * @memberof oj.Color
 * @instance
 * @export
 */
oj.Color.prototype.getRed = function(doNotRound)
{
   doNotRound = doNotRound || false ;
   return doNotRound? this._r : mathRound(this._r) ;
};


/**
 * Returns the green channel value of the color.
 * @param {boolean=} doNotRound  Omit or set to false to return an integer value.
 *                               Set to true to return the possible fractional value.
 *                               (Fractional values for the red, green, or blue channels can
 *                               arise when a color is defined using the "hsl" format.)
 * @return {number} The green channel value contained in [0,255].
 * @memberof oj.Color
 * @instance
 * @export
 */
oj.Color.prototype.getGreen = function(doNotRound)
{
   doNotRound = doNotRound || false ;
   return doNotRound? this._g : mathRound(this._g) ;
};


/**
 * Returns the blue channel value of the color.
 * @param {boolean=} doNotRound  Omit or set to false to return an integer value.
 *                               Set to true to return the possible fractional value.
 *                               (Fractional values for the red, green, or blue channels can
 *                               arise when a color is defined using the "hsl" format.)
 * @return {number} The blue channel value contained in [0,255].
 * @memberof oj.Color
 * @instance
 * @export
 */
oj.Color.prototype.getBlue = function(doNotRound)
{
   doNotRound = doNotRound || false ;
   return doNotRound? this._b : mathRound(this._b) ;
};


/**
 * Returns the alpha channel of the color.
 * @return {number} The alpha channel value contained in [0,1].
 * @memberof oj.Color
 * @instance
 * @export
 */
oj.Color.prototype.getAlpha = function()
{
   return this._a ;
};


/**
 * Returns the color as an "rgb" or rgba" (if the alpha value is less than 1) CSS3 color
 * specification string.
 * @return {string} The color as an "rgb" or "rgba" CSS3 color specification string.
 * @memberof oj.Color
 * @instance
 * @export
 */
oj.Color.prototype.toString = function()
{
   return toRgbString(this) ;
}

/**
 * Compares this color object with the supplied color specification, and returns true
 * if they represent the same color, else false.
 * @param {oj.Color} color   The color to be compared.
 * @return {boolean} Returns true if the comperand represents the same color.
 * @memberof oj.Color
 * @instance
 * @export
 */
oj.Color.prototype.isEqual =  function(color)
{
   var ret = false ;

   if (color instanceof oj.Color)
   {
      ret = ((this._r === color._r) && (this._g === color._g) &&
             (this._b === color._b) && (this._a === color._a)) ;
   }

   return ret ;
};


/** CSS3 color <code class="prettyprint">aliceblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:aliceblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.ALICEBLUE = null;
/** CSS3 color <code class="prettyprint">antiquewhite</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:antiquewhite;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.ANTIQUEWHITE = null;
/** CSS3 color <code class="prettyprint">aqua</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:aqua;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.AQUA = null;
/** CSS3 color <code class="prettyprint">aquamarine</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:aquamarine;"></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.AQUAMARINE = null;
/** CSS3 color <code class="prettyprint">azure</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:azure;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.AZURE = null;
/** CSS3 color <code class="prettyprint">beige</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:beige;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BEIGE = null;
/** CSS3 color <code class="prettyprint">bisque</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:bisque;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BISQUE = null;
/** CSS3 color <code class="prettyprint">black</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:black;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BLACK = null;
/** CSS3 color <code class="prettyprint">blanchedalmond</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:blanchedalmond;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BLANCHEDALMOND = null;
/** CSS3 color <code class="prettyprint">blue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:blue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BLUE = null;
/** CSS3 color <code class="prettyprint">blueviolet</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:blueviolet;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BLUEVIOLET = null;
/** CSS3 color <code class="prettyprint">brown</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:brown;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BROWN = null;     
/** CSS3 color <code class="prettyprint">burlywood</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:burlywood;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.BURLYWOOD = null; 
/** CSS3 color <code class="prettyprint">cadetblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:cadetblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CADETBLUE = null;
/** CSS3 color <code class="prettyprint">chartreuse</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:chartreuse;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CHARTREUSE = null;
/** CSS3 color <code class="prettyprint">chocolate</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:chocolate;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CHOCOLATE = null;
/** CSS3 color <code class="prettyprint">coral</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:coral;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CORAL = null;
/** CSS3 color <code class="prettyprint">cornflowerblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:cornflowerblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CORNFLOWERBLUE = null;
/** CSS3 color <code class="prettyprint">cornsilk</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:cornsilk;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CORNSILK = null;
/** CSS3 color <code class="prettyprint">crimson</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:crimson;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CRIMSON = null;
/** CSS3 color <code class="prettyprint">cyan</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:cyan;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.CYAN = null;
/** CSS3 color <code class="prettyprint">darkblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKBLUE = null;
/** CSS3 color <code class="prettyprint">darkcyan</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkcyan;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKCYAN = null;
/** CSS3 color <code class="prettyprint">darkgoldenrod</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkgoldenrod;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKGOLDENROD = null;
/** CSS3 color <code class="prettyprint">darkgray</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkgray;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKGRAY = null;
/** CSS3 color <code class="prettyprint">darkgrey</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkgrey;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKGREY = null;
/** CSS3 color <code class="prettyprint">darkgreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkgreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKGREEN = null;
/** CSS3 color <code class="prettyprint">darkkhaki</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkkhaki;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKKHAKI = null;
/** CSS3 color <code class="prettyprint">darkmagenta</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkmagenta;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKMAGENTA = null;
/** CSS3 color <code class="prettyprint">darkolivegreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkolivegreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKOLIVEGREEN = null;
/** CSS3 color <code class="prettyprint">darkorange</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkorange;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKORANGE = null;
/** CSS3 color <code class="prettyprint">darkorchid</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkorchid;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKORCHID = null;
/** CSS3 color <code class="prettyprint">darkred</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkred;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKRED = null;
/** CSS3 color <code class="prettyprint">darksalmon</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darksalmon;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKSALMON = null;
/** CSS3 color <code class="prettyprint">darkseagreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkseagreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKSEAGREEN = null;
/** CSS3 color <code class="prettyprint">darkslateblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkslateblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKSLATEBLUE = null;
/** CSS3 color <code class="prettyprint">darkslategray</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkslategray;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKSLATEGRAY = null;
/** CSS3 color <code class="prettyprint">darkslategrey</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkslategrey;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKSLATEGREY = null;
/** CSS3 color <code class="prettyprint">darkturquoise</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkturquoise;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKTURQUOISE  = null;
/** CSS3 color <code class="prettyprint">darkviolet</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:darkviolet;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DARKVIOLET = null;
/** CSS3 color <code class="prettyprint">deeppink</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:deeppink;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DEEPPINK = null;
/** CSS3 color <code class="prettyprint">deepskyblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:deepskyblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DEEPSKYBLUE = null;
/** CSS3 color <code class="prettyprint">dimgray</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:dimgray;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DIMGRAY = null;
/** CSS3 color <code class="prettyprint">dimgrey</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:dimgrey;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DIMGREY = null;
/** CSS3 color <code class="prettyprint">dodgerblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:dodgerblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.DODGERBLUE = null;
/** CSS3 color <code class="prettyprint">firebrick</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:firebrick;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.FIREBRICK = null;
/** CSS3 color <code class="prettyprint">floralwhite</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:floralwhite;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.FLORALWHITE = null;
/** CSS3 color <code class="prettyprint">forestgreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:forestgreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.FORESTGREEN = null;
/** CSS3 color <code class="prettyprint">fuchsia</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:fuchsia;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.FUCHSIA = null;
/** CSS3 color <code class="prettyprint">gainsboro</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:gainsboro;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GAINSBORO = null;
/** CSS3 color <code class="prettyprint">ghostwhite</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:ghostwhite;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GHOSTWHITE = null;
/** CSS3 color <code class="prettyprint">gold</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:gold;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GOLD = null;
/** CSS3 color <code class="prettyprint">goldenrod</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:goldenrod;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GOLDENROD = null;
/** CSS3 color <code class="prettyprint">gray</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:gray;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GRAY = null;
/** CSS3 color <code class="prettyprint">green</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:green;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GREEN = null;
/** CSS3 color <code class="prettyprint">greenyellow</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:greenyellow;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GREENYELLOW = null;
/** CSS3 color <code class="prettyprint">grey</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:grey;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.GREY = null;
/** CSS3 color <code class="prettyprint">honeydew</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:honeydew;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.HONEYDEW = null;
/** CSS3 color <code class="prettyprint">hotpink</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:hotpink;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.HOTPINK = null;
/** CSS3 color <code class="prettyprint">indianred</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:indianred;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.INDIANRED = null;
/** CSS3 color <code class="prettyprint">indigo</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:indigo;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.INDIGO = null;
/** CSS3 color <code class="prettyprint">ivory</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:ivory;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.IVORY = null;
/** CSS3 color <code class="prettyprint">khaki</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:khaki;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.KHAKI = null;
/** CSS3 color <code class="prettyprint">lavender</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lavender;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LAVENDER = null;
/** CSS3 color <code class="prettyprint">lavenderblush</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lavenderblush;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LAVENDERBLUSH = null;
/** CSS3 color <code class="prettyprint">lawngreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lawngreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LAWNGREEN = null;
/** CSS3 color <code class="prettyprint">lemonchiffon</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lemonchiffon;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LEMONCHIFFON = null;
/** CSS3 color <code class="prettyprint">lightblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTBLUE = null;
/** CSS3 color <code class="prettyprint">lightcoral</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightcoral;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTCORAL = null;
/** CSS3 color <code class="prettyprint">lightcyan</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightcyan;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTCYAN = null;
/** CSS3 color <code class="prettyprint">lightgoldenrodyellow</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightgoldenrodyellow;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTGOLDENRODYELLOW = null;
/** CSS3 color <code class="prettyprint">lightgray</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightgray;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTGRAY = null;
/** CSS3 color <code class="prettyprint">lightgreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightgreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTGREEN = null;
/** CSS3 color <code class="prettyprint">lightgrey</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightgrey;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTGREY = null;
/** CSS3 color <code class="prettyprint">lightpink</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightpink;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTPINK = null;
/** CSS3 color <code class="prettyprint">lightsalmon</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightsalmon;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTSALMON = null;
/** CSS3 color <code class="prettyprint">lightseagreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightseagreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTSEAGREEN = null;
/** CSS3 color <code class="prettyprint">lightskyblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightskyblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTSKYBLUE = null;
/** CSS3 color <code class="prettyprint">lightslategray</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightslategray;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTSLATEGRAY = null;
/** CSS3 color <code class="prettyprint">lightslategrey</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightslategrey;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTSLATEGREY = null;
/** CSS3 color <code class="prettyprint">lightsteelblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightsteelblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTSTEELBLUE = null;
/** CSS3 color <code class="prettyprint">lightyellow</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lightyellow;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIGHTYELLOW = null;
/** CSS3 color <code class="prettyprint">lime</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:lime;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIME = null;
/** CSS3 color <code class="prettyprint">limegreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:limegreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LIMEGREEN = null;
/** CSS3 color <code class="prettyprint">linen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:linen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.LINEN = null;
/** CSS3 color <code class="prettyprint">magenta</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:magenta;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MAGENTA = null;
/** CSS3 color <code class="prettyprint">maroon</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:maroon;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MAROON = null;
/** CSS3 color <code class="prettyprint">mediumaquamarine</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumaquamarine;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMAQUAMARINE = null;
/** CSS3 color <code class="prettyprint">mediumblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMBLUE = null;
/** CSS3 color <code class="prettyprint">mediumorchid</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumorchid;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMORCHID = null;
/** CSS3 color <code class="prettyprint">mediumpurple</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumpurple;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMPURPLE = null;
/** CSS3 color <code class="prettyprint">mediumseagreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumseagreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMSEAGREEN = null;
/** CSS3 color <code class="prettyprint">mediumslateblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumslateblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMSLATEBLUE = null;
/** CSS3 color <code class="prettyprint">mediumspringgreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumspringgreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMSPRINGGREEN = null;
/** CSS3 color <code class="prettyprint">mediumturquoise</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumturquoise;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMTURQUOISE = null;
/** CSS3 color <code class="prettyprint">mediumvioletred</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mediumvioletred;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MEDIUMVIOLETRED = null;
/** CSS3 color <code class="prettyprint">midnightblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:midnightblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MIDNIGHTBLUE = null;
/** CSS3 color <code class="prettyprint">mintcream</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mintcream;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MINTCREAM = null;
/** CSS3 color <code class="prettyprint">mistyrose</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:mistyrose;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MISTYROSE = null;
/** CSS3 color <code class="prettyprint">moccasin</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:moccasin;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.MOCCASIN = null;
/** CSS3 color <code class="prettyprint">navajowhite</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:navajowhite;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.NAVAJOWHITE = null;
/** CSS3 color <code class="prettyprint">navy</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:navy;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.NAVY = null;
/** CSS3 color <code class="prettyprint">oldlace</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:oldlace;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.OLDLACE = null;
/** CSS3 color <code class="prettyprint">olive</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:olive;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.OLIVE = null;
/** CSS3 color <code class="prettyprint">olivedrab</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:olivedrab;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.OLIVEDRAB = null;
/** CSS3 color <code class="prettyprint">orange</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:orange;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.ORANGE = null;
/** CSS3 color <code class="prettyprint">orangered</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:orangered;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.ORANGERED = null;
/** CSS3 color <code class="prettyprint">orchid</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:orchid;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.ORCHID = null;
/** CSS3 color <code class="prettyprint">palegoldenrod</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:palegoldenrod;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PALEGOLDENROD = null;
/** CSS3 color <code class="prettyprint">palegreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:palegreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PALEGREEN = null;
/** CSS3 color <code class="prettyprint">paleturquoise</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:paleturquoise;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PALETURQUOISE = null;
/** CSS3 color <code class="prettyprint">palevioletred</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:palevioletred;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PALEVIOLETRED = null;
/** CSS3 color <code class="prettyprint">papayawhip</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:papayawhip;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PAPAYAWHIP = null;
/** CSS3 color <code class="prettyprint">peachpuff</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:peachpuff;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PEACHPUFF = null;
/** CSS3 color <code class="prettyprint">peru</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:peru;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PERU = null;
/** CSS3 color <code class="prettyprint">pink</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:pink;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PINK = null;
/** CSS3 color <code class="prettyprint">plum</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:plum;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PLUM = null;
/** CSS3 color <code class="prettyprint">powderblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:powderblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.POWDERBLUE = null;
/** CSS3 color <code class="prettyprint">purple</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:purple;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.PURPLE = null;
/** CSS3 color <code class="prettyprint">rebeccapurple</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:rebeccapurple;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.REBECCAPURPLE = null;
/** CSS3 color <code class="prettyprint">red</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:red;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.RED = null;
/** CSS3 color <code class="prettyprint">rosybrown</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:rosybrown;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.ROSYBROWN = null;
/** CSS3 color <code class="prettyprint">royalblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:royalblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.ROYALBLUE = null;
/** CSS3 color <code class="prettyprint">saddlebrown</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:saddlebrown;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SADDLEBROWN = null;
/** CSS3 color <code class="prettyprint">salmon</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:salmon;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SALMON = null;
/** CSS3 color <code class="prettyprint">sandybrown</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:sandybrown;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SANDYBROWN = null;
/** CSS3 color <code class="prettyprint">seagreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:seagreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SEAGREEN = null;
/** CSS3 color <code class="prettyprint">seashell</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:seashell;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SEASHELL = null;
/** CSS3 color <code class="prettyprint">sienna</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:sienna;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SIENNA = null;
/** CSS3 color <code class="prettyprint">silver</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:silver;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SILVER = null;
/** CSS3 color <code class="prettyprint">skyblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:skyblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SKYBLUE = null;
/** CSS3 color <code class="prettyprint">slateblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:slateblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SLATEBLUE = null;
/** CSS3 color <code class="prettyprint">slategray</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:slategray;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SLATEGRAY = null;
/** CSS3 color <code class="prettyprint">slategrey</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:slategrey;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SLATEGREY = null;
/** CSS3 color <code class="prettyprint">snow</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:snow;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SNOW = null;
/** CSS3 color <code class="prettyprint">springgreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:springgreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.SPRINGGREEN = null;
/** CSS3 color <code class="prettyprint">steelblue</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:steelblue;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.STEELBLUE = null;
/** CSS3 color <code class="prettyprint">tan</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:tan;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.TAN = null;
/** CSS3 color <code class="prettyprint">teal</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:teal;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.TEAL = null;
/** CSS3 color <code class="prettyprint">thistle</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:thistle;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.THISTLE = null;
/** CSS3 color <code class="prettyprint">tomato</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:tomato;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.TOMATO = null;
/** CSS3 color <code class="prettyprint">turquoise</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:turquoise;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.TURQUOISE = null;
/** CSS3 color <code class="prettyprint">violet</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:violet;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.VIOLET = null;
/** CSS3 color <code class="prettyprint">wheat</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:wheat;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.WHEAT = null;
/** CSS3 color <code class="prettyprint">white</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:white;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.WHITE = null;
/** CSS3 color <code class="prettyprint">whitesmoke</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:whitesmoke;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.WHITESMOKE = null;
/** CSS3 color <code class="prettyprint">yellow</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:yellow;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.YELLOW = null;
/** CSS3 color <code class="prettyprint">yellowgreen</code><div style="display:inline;padding-left:40px;margin-left:50px;height:15px;width:40px;background:yellowgreen;"/></div>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.YELLOWGREEN = null;
/** CSS3 color <code class="prettyprint">transparent</code>
 * @type {oj.Color}
 * @expose
 * @public
 * @static
 * @memberof oj.Color
 */
oj.Color.TRANSPARENT = null;




/*-----------------------------------------------------------------------*/
/*     Internal supporting functions below this point                    */
/*-----------------------------------------------------------------------*/
/**
 * Converts a valid CSS3 color specification string (exc. named colors), or
 * an object of the forms {r:, g:, b:}, {h:, s:, l:}, or {h", s:, v:} with
 * optional "a" property, to an object with validated r,g,b properties.
 * Invalid syntax causes an error to be thrown.
 *
 * Example input: 
 *  "#f00" or "f00"
 *  "#ff0000" or "ff0000"
 *  "rgb 255 0 0" or "rgb (255, 0, 0)"
 *  "rgb 1.0 0 0" or "rgb (1, 0, 0)"
 *  "rgba (255, 0, 0, 1)" or "rgba 255, 0, 0, 1"
 *  "rgba (1.0, 0, 0, 1)" or "rgba 1.0, 0, 0, 1"
 *  "hsl(0, 100%, 50%)" or "hsl 0 100% 50%"
 *  "hsla(0, 100%, 50%, 1)" or "hsla 0 100% 50%, 1"
 *  "hsv(0, 100%, 100%)" or "hsv 0 100% 100%"
 *
 *  Note: when converting from hsl or hsv, the resulting r,g,b components may be fractional
 *       e.g.  hsla(0, 97%,37%,0.851) will return the following
 *             r = 185.8695, g = 2.8305, b = 2.8305, a = 0.851
 * @memberof oj.Color
 * @returns {Object} color object
 * @private
 */
function inputToRGB(color)
{
   var rgb = { r: 0, g: 0, b: 0 },
       a   = 1,
       s   = null,
       v   = null,
       l   = null,
       ok  = false;

   if (typeof color == "string")
   {
     color = stringInputToObject(color);  // convert to {"r":, "g":, "b":} or {"h":, "s":, l}, or {"h":, "s":, "v":}
   }

   if (typeof color == "object")
   {  
       if (isValidCSSUnit(color["r"]) && isValidCSSUnit(color["g"]) && isValidCSSUnit(color["b"]))
       {
         rgb = rgbToRgb(color["r"], color["g"], color["b"]);
         ok = true;
       }
       else if (isValidCSSUnit(color["h"]) && isValidCSSUnit(color["s"]) && isValidCSSUnit(color["v"]))
       {
         s = convertToPercentage(color["s"]);
         v = convertToPercentage(color["v"]);
         rgb = hsvToRgb(color["h"], s, v);
         ok = true;
       }
       else if (isValidCSSUnit(color["h"]) && isValidCSSUnit(color["s"]) && isValidCSSUnit(color["l"]))
       {
         s = convertToPercentage(color["s"]);
         l = convertToPercentage(color["l"]);
         rgb = hslToRgb(color["h"], s, l);
         ok = true;
       }

       if (typeof color["a"] !== "undefined")
       {
         a = color["a"];
       }
   }

   if (!ok)
   {
     throw new Error("Invalid Color format");
   }

   a = boundAlpha(a);
   return {
            r: mathMin(255, mathMax(rgb.r, 0)),
            g: mathMin(255, mathMax(rgb.g, 0)),
            b: mathMin(255, mathMax(rgb.b, 0)),
            a: a
          };
};

/**
 * @param {oj.Color} co The color object
 * @returns {string} rgb string
 * @memberof oj.Color
 * @private
 */
function  toRgbString(co)
{
   var b = (co._a < 1);
   return "rgb" + (b? "a(" : "(") +
           mathRound(co._r) + ", " + mathRound(co._g) + ", " + mathRound(co._b) + 
           (b? (", " + co._a) : "") + ")" ;
};


/*---------------------------------------------------------------------------*/
/*               Color conversion supporting functions                       */
/*---------------------------------------------------------------------------*/
/*  <http://mjijackson.com/2008/02/rgb-to-hsl-and-rgb-to-hsv-                */
/*                          color-model-conversion-algorithms-in-javascript> */
/*---------------------------------------------------------------------------*/
/**
 * Converts an RGB color value to RGB.
 * Handle bounds/percentage checking to conform to CSS color spec, and returns
 * an object containg the r,g,b values.
 * <http://www.w3.org/TR/css3-color/>
 * Assumes:  r, g, b in [0, 255] or [0, 1]
 * @param {number} r the red value
 * @param {number} g the green value
 * @param {number} b the blue value
 * @returns {Object} Object with properties r, g, and b, in [0, 255].
 * @memberof oj.Color
 * @private
 */
function rgbToRgb(r, g, b)
{
   return {
            r: bound01(r, 255) * 255,
            g: bound01(g, 255) * 255,
            b: bound01(b, 255) * 255
          };
};

/**
 * Converts an RGB color value to HSL.
 * Handle bounds/percentage checking to conform to CSS color spec, and returns
 * an object containg the h,s,l values.
 * <http://www.w3.org/TR/css3-color/>
 * Assumes:  r, g, b in [0, 255] or [0, 1]
 * @param {number} r the red value
 * @param {number} g the green value
 * @param {number} b the blue value
 * @returns {Object} Object with properties h, s, l, in [0, 1].
 * @memberof oj.Color
 * @private
 */
function rgbToHsl(r, g, b)
{
   r = bound01(r, 255);
   g = bound01(g, 255);
   b = bound01(b, 255);

   var max = mathMax(r, g, b), min = mathMin(r, g, b);
   var h, s, l = (max + min) / 2;

   if (max == min)
   {
     h = s = 0; // achromatic
   }
   else
   {
     var d = max - min;
     s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
     switch(max)
     {
        case r: h = (g - b) / d + (g < b ? 6 : 0); break;
        case g: h = (b - r) / d + 2; break;
        case b: h = (r - g) / d + 4; break;
     }

     h /= 6;
   }

   return { h: h, s: s, l: l };
};

/**
 * Converts an HSL color value to RGB.
 * Handle bounds/percentage checking to conform to CSS color spec, and returns
 * an object containg the r,g,b values.
 * <http://www.w3.org/TR/css3-color/>
 * Assumes:  h is contained in [0,1] or [0,360, and s an l are contained in [0, 1] or [0, 100]
 * @param {number} h the hue value
 * @param {number} s the saturation value
 * @param {number} l the lightness value
 * @returns {Object} Object with properties r, g, b, in [0, 255].
 * @memberof oj.Color
 * @private
 */
function hslToRgb(h, s, l)
{
   var r, g, b;

   h = bound01(h, 360);
   s = bound01(s, 100);
   l = bound01(l, 100);

   function hue2rgb(p, q, t)
   {
     if(t < 0) t += 1;
     if(t > 1) t -= 1;
     if(t < 1/6) return p + (q - p) * 6 * t;
     if(t < 1/2) return q;
     if(t < 2/3) return p + (q - p) * (2/3 - t) * 6;
     return p;
   }

   if(s === 0)
   {
      r = g = b = l; // achromatic
   }
   else
   {
     var q = l < 0.5 ? l * (1 + s) : l + s - l * s;
     var p = 2 * l - q;
     r = hue2rgb(p, q, h + 1/3);
     g = hue2rgb(p, q, h);
     b = hue2rgb(p, q, h - 1/3);
   }

   return { r: r * 255, g: g * 255, b: b * 255 };
};

/**
 * Converts an HSV color value to RGB.
 * Assumes: h is contained in [0, 1] or [0, 360] and s and v are contained in [0, 1] or [0, 100]
 * @returns {Object} An object with properties  r, g, b  in the set [0,255]
 * @memberof oj.Color
 * @private
 */
 function hsvToRgb(h, s, v) {

    h = bound01(h, 360) * 6;
    s = bound01(s, 100);
    v = bound01(v, 100);

    var i = Math.floor(h),
        f = h - i,
        p = v * (1 - s),
        q = v * (1 - f * s),
        t = v * (1 - (1 - f) * s),
        mod = i % 6,
        r = [v, q, p, p, t, v][mod],
        g = [t, v, v, q, p, p][mod],
        b = [p, p, t, v, v, q][mod];

    return { r: r * 255, g: g * 255, b: b * 255 };
}

/**
 * @memberof oj.Color
 * @private
 * @return {void}
 */
function _initColorConstants()
{
  Object.defineProperty(oj.Color, "ALICEBLUE", {writable: false, value: new oj.Color("f0f8ff")});
  Object.defineProperty(oj.Color, "ANTIQUEWHITE", {writable: false, value: new oj.Color("faebd7")});
  Object.defineProperty(oj.Color, "AQUA", {writable: false, value: new oj.Color("0ff")});
  Object.defineProperty(oj.Color, "AQUAMARINE", {writable: false, value: new oj.Color("7fffd4")});
  Object.defineProperty(oj.Color, "AZURE", {writable: false, value: new oj.Color("f0ffff")});
  Object.defineProperty(oj.Color, "BEIGE", {writable: false, value: new oj.Color("f5f5dc")});
  Object.defineProperty(oj.Color, "BISQUE", {writable: false, value: new oj.Color("ffe4c4")});
  Object.defineProperty(oj.Color, "BLACK", {writable: false, value: new oj.Color("000")});
  Object.defineProperty(oj.Color, "BLANCHEDALMOND", {writable: false, value: new oj.Color("ffebcd")});
  Object.defineProperty(oj.Color, "BLUE", {writable: false, value: new oj.Color("00f")});
  Object.defineProperty(oj.Color, "BLUEVIOLET", {writable: false, value: new oj.Color("8a2be2")});
  Object.defineProperty(oj.Color, "BROWN", {writable: false, value: new oj.Color("a52a2a")});
  Object.defineProperty(oj.Color, "BURLYWOOD", {writable: false, value: new oj.Color("deb887")});
  Object.defineProperty(oj.Color, "CADETBLUE", {writable: false, value: new oj.Color("5f9ea0")});
  Object.defineProperty(oj.Color, "CHARTREUSE", {writable: false, value: new oj.Color("7fff00")});
  Object.defineProperty(oj.Color, "CHOCOLATE", {writable: false, value: new oj.Color("d2691e")});
  Object.defineProperty(oj.Color, "CORAL", {writable: false, value: new oj.Color("ff7f50")});
  Object.defineProperty(oj.Color, "CORNFLOWERBLUE", {writable: false, value: new oj.Color("6495ed")});
  Object.defineProperty(oj.Color, "CORNSILK", {writable: false, value: new oj.Color("fff8dc")});
  Object.defineProperty(oj.Color, "CRIMSON", {writable: false, value: new oj.Color("dc143c")});
  Object.defineProperty(oj.Color, "CYAN", {writable: false, value: new oj.Color("0ff")});
  Object.defineProperty(oj.Color, "DARKBLUE", {writable: false, value: new oj.Color("00008b")});
  Object.defineProperty(oj.Color, "DARKCYAN", {writable: false, value: new oj.Color("008b8b")});
  Object.defineProperty(oj.Color, "DARKGOLDENROD", {writable: false, value: new oj.Color("b8860b")});
  Object.defineProperty(oj.Color, "DARKGRAY", {writable: false, value: new oj.Color("a9a9a9")});
  Object.defineProperty(oj.Color, "DARKGREY", {writable: false, value: new oj.Color("a9a9a9")});
  Object.defineProperty(oj.Color, "DARKGREEN", {writable: false, value: new oj.Color("006400")});
  Object.defineProperty(oj.Color, "DARKKHAKI", {writable: false, value: new oj.Color("bdb76b")});
  Object.defineProperty(oj.Color, "DARKMAGENTA", {writable: false, value: new oj.Color("8b008b")});
  Object.defineProperty(oj.Color, "DARKOLIVEGREEN", {writable: false, value: new oj.Color("556b2f")});
  Object.defineProperty(oj.Color, "DARKORANGE", {writable: false, value: new oj.Color("ff8c00")});
  Object.defineProperty(oj.Color, "DARKORCHID", {writable: false, value: new oj.Color("9932cc")});
  Object.defineProperty(oj.Color, "DARKRED", {writable: false, value: new oj.Color("8b0000")});
  Object.defineProperty(oj.Color, "DARKSALMON", {writable: false, value: new oj.Color("e9967a")});
  Object.defineProperty(oj.Color, "DARKSEAGREEN", {writable: false, value: new oj.Color("8fbc8f")});
  Object.defineProperty(oj.Color, "DARKSLATEBLUE", {writable: false, value: new oj.Color("483d8b")});
  Object.defineProperty(oj.Color, "DARKSLATEGRAY", {writable: false, value: new oj.Color("2f4f4f")});
  Object.defineProperty(oj.Color, "DARKSLATEGREY", {writable: false, value: new oj.Color("2f4f4f")});
  Object.defineProperty(oj.Color, "DARKTURQUOISE", {writable: false, value: new oj.Color("00ced1")});
  Object.defineProperty(oj.Color, "DARKVIOLET", {writable: false, value: new oj.Color("9400d3")});
  Object.defineProperty(oj.Color, "DEEPPINK", {writable: false, value: new oj.Color("ff1493")});
  Object.defineProperty(oj.Color, "DEEPSKYBLUE", {writable: false, value: new oj.Color("00bfff")});
  Object.defineProperty(oj.Color, "DIMGRAY", {writable: false, value: new oj.Color("696969")});
  Object.defineProperty(oj.Color, "DIMGREY", {writable: false, value: new oj.Color("696969")});
  Object.defineProperty(oj.Color, "DODGERBLUE", {writable: false, value: new oj.Color("1e90ff")});
  Object.defineProperty(oj.Color, "FIREBRICK", {writable: false, value: new oj.Color("b22222")});
  Object.defineProperty(oj.Color, "FLORALWHITE", {writable: false, value: new oj.Color("fffaf0")});
  Object.defineProperty(oj.Color, "FORESTGREEN", {writable: false, value: new oj.Color("228b22")});
  Object.defineProperty(oj.Color, "FUCHSIA", {writable: false, value: new oj.Color("f0f")});
  Object.defineProperty(oj.Color, "GAINSBORO", {writable: false, value: new oj.Color("dcdcdc")});
  Object.defineProperty(oj.Color, "GHOSTWHITE", {writable: false, value: new oj.Color("f8f8ff")});
  Object.defineProperty(oj.Color, "GOLD", {writable: false, value: new oj.Color("ffd700")});
  Object.defineProperty(oj.Color, "GOLDENROD", {writable: false, value: new oj.Color("daa520")});
  Object.defineProperty(oj.Color, "GRAY", {writable: false, value: new oj.Color("808080")});
  Object.defineProperty(oj.Color, "GREY", {writable: false, value: new oj.Color("808080")});
  Object.defineProperty(oj.Color, "GREEN", {writable: false, value: new oj.Color("008000")});
  Object.defineProperty(oj.Color, "GREENYELLOW", {writable: false, value: new oj.Color("adff2f")});
  Object.defineProperty(oj.Color, "HONEYDEW", {writable: false, value: new oj.Color("f0fff0")});
  Object.defineProperty(oj.Color, "HOTPINK", {writable: false, value: new oj.Color("ff69b4")});
  Object.defineProperty(oj.Color, "INDIANRED", {writable: false, value: new oj.Color("cd5c5c")});
  Object.defineProperty(oj.Color, "INDIGO", {writable: false, value: new oj.Color("4b0082")});
  Object.defineProperty(oj.Color, "IVORY", {writable: false, value: new oj.Color("fffff0")});
  Object.defineProperty(oj.Color, "KHAKI", {writable: false, value: new oj.Color("f0e68c")});
  Object.defineProperty(oj.Color, "LAVENDER", {writable: false, value: new oj.Color("e6e6fa")});
  Object.defineProperty(oj.Color, "LAVENDERBLUSH", {writable: false, value: new oj.Color("fff0f5")});
  Object.defineProperty(oj.Color, "LAWNGREEN", {writable: false, value: new oj.Color("7cfc00")});
  Object.defineProperty(oj.Color, "LEMONCHIFFON", {writable: false, value: new oj.Color("fffacd")});
  Object.defineProperty(oj.Color, "LIGHTBLUE", {writable: false, value: new oj.Color("add8e6")});
  Object.defineProperty(oj.Color, "LIGHTCORAL", {writable: false, value: new oj.Color("f08080")});
  Object.defineProperty(oj.Color, "LIGHTCYAN", {writable: false, value: new oj.Color("e0ffff")});
  Object.defineProperty(oj.Color, "LIGHTGOLDENRODYELLOW", {writable: false, value: new oj.Color("fafad2")});
  Object.defineProperty(oj.Color, "LIGHTGRAY", {writable: false, value: new oj.Color("d3d3d3")});
  Object.defineProperty(oj.Color, "LIGHTGREY", {writable: false, value: new oj.Color("d3d3d3")});
  Object.defineProperty(oj.Color, "LIGHTGREEN", {writable: false, value: new oj.Color("90ee90")});
  Object.defineProperty(oj.Color, "LIGHTPINK", {writable: false, value: new oj.Color("ffb6c1")});
  Object.defineProperty(oj.Color, "LIGHTSALMON", {writable: false, value: new oj.Color("ffa07a")});
  Object.defineProperty(oj.Color, "LIGHTSEAGREEN", {writable: false, value: new oj.Color("20b2aa")});
  Object.defineProperty(oj.Color, "LIGHTSKYBLUE", {writable: false, value: new oj.Color("87cefa")});
  Object.defineProperty(oj.Color, "LIGHTSLATEGRAY", {writable: false, value: new oj.Color("789")});
  Object.defineProperty(oj.Color, "LIGHTSLATEGREY", {writable: false, value: new oj.Color("789")});
  Object.defineProperty(oj.Color, "LIGHTSTEELBLUE", {writable: false, value: new oj.Color("b0c4de")});
  Object.defineProperty(oj.Color, "LIGHTYELLOW", {writable: false, value: new oj.Color("ffffe0")});
  Object.defineProperty(oj.Color, "LIME", {writable: false, value: new oj.Color("0f0")});
  Object.defineProperty(oj.Color, "LIMEGREEN", {writable: false, value: new oj.Color("32cd32")});
  Object.defineProperty(oj.Color, "LINEN", {writable: false, value: new oj.Color("faf0e6")});
  Object.defineProperty(oj.Color, "MAGENTA", {writable: false, value: new oj.Color("f0f")});
  Object.defineProperty(oj.Color, "MAROON", {writable: false, value: new oj.Color("800000")});
  Object.defineProperty(oj.Color, "MEDIUMAQUAMARINE", {writable: false, value: new oj.Color("66cdaa")});
  Object.defineProperty(oj.Color, "MEDIUMBLUE", {writable: false, value: new oj.Color("0000cd")});
  Object.defineProperty(oj.Color, "MEDIUMORCHID", {writable: false, value: new oj.Color("ba55d3")});
  Object.defineProperty(oj.Color, "MEDIUMPURPLE", {writable: false, value: new oj.Color("9370db")});
  Object.defineProperty(oj.Color, "MEDIUMSEAGREEN", {writable: false, value: new oj.Color("3cb371")});
  Object.defineProperty(oj.Color, "MEDIUMSLATEBLUE", {writable: false, value: new oj.Color("7b68ee")});
  Object.defineProperty(oj.Color, "MEDIUMSPRINGGREEN", {writable: false, value: new oj.Color("00fa9a")});
  Object.defineProperty(oj.Color, "MEDIUMTURQUOISE", {writable: false, value: new oj.Color("48d1cc")});
  Object.defineProperty(oj.Color, "MEDIUMVIOLETRED", {writable: false, value: new oj.Color("c71585")});
  Object.defineProperty(oj.Color, "MIDNIGHTBLUE", {writable: false, value: new oj.Color("191970")});
  Object.defineProperty(oj.Color, "MINTCREAM", {writable: false, value: new oj.Color("f5fffa")});
  Object.defineProperty(oj.Color, "MISTYROSE", {writable: false, value: new oj.Color("ffe4e1")});
  Object.defineProperty(oj.Color, "MOCCASIN", {writable: false, value: new oj.Color("ffe4b5")});
  Object.defineProperty(oj.Color, "NAVAJOWHITE", {writable: false, value: new oj.Color("ffdead")});
  Object.defineProperty(oj.Color, "NAVY", {writable: false, value: new oj.Color("000080")});
  Object.defineProperty(oj.Color, "OLDLACE", {writable: false, value: new oj.Color("fdf5e6")});
  Object.defineProperty(oj.Color, "OLIVE", {writable: false, value: new oj.Color("808000")});
  Object.defineProperty(oj.Color, "OLIVEDRAB", {writable: false, value: new oj.Color("6b8e23")});
  Object.defineProperty(oj.Color, "ORANGE", {writable: false, value: new oj.Color("ffa500")});
  Object.defineProperty(oj.Color, "ORANGERED", {writable: false, value: new oj.Color("ff4500")});
  Object.defineProperty(oj.Color, "ORCHID", {writable: false, value: new oj.Color("da70d6")});
  Object.defineProperty(oj.Color, "PALEGOLDENROD", {writable: false, value: new oj.Color("eee8aa")});
  Object.defineProperty(oj.Color, "PALEGREEN", {writable: false, value: new oj.Color("98fb98")});
  Object.defineProperty(oj.Color, "PALETURQUOISE", {writable: false, value: new oj.Color("afeeee")});
  Object.defineProperty(oj.Color, "PALEVIOLETRED", {writable: false, value: new oj.Color("db7093")});
  Object.defineProperty(oj.Color, "PAPAYAWHIP", {writable: false, value: new oj.Color("ffefd5")});
  Object.defineProperty(oj.Color, "PEACHPUFF", {writable: false, value: new oj.Color("ffdab9")});
  Object.defineProperty(oj.Color, "PERU", {writable: false, value: new oj.Color("cd853f")});
  Object.defineProperty(oj.Color, "PINK", {writable: false, value: new oj.Color("ffc0cb")});
  Object.defineProperty(oj.Color, "PLUM", {writable: false, value: new oj.Color("dda0dd")});
  Object.defineProperty(oj.Color, "POWDERBLUE", {writable: false, value: new oj.Color("b0e0e6")});
  Object.defineProperty(oj.Color, "PURPLE", {writable: false, value: new oj.Color("800080")});
  Object.defineProperty(oj.Color, "REBECCAPURPLE", {writable: false, value: new oj.Color("663399")});
  Object.defineProperty(oj.Color, "RED", {writable: false, value: new oj.Color("f00")});
  Object.defineProperty(oj.Color, "ROSYBROWN", {writable: false, value: new oj.Color("bc8f8f")});
  Object.defineProperty(oj.Color, "ROYALBLUE", {writable: false, value: new oj.Color("4169e1")});
  Object.defineProperty(oj.Color, "SADDLEBROWN", {writable: false, value: new oj.Color("8b4513")});
  Object.defineProperty(oj.Color, "SALMON", {writable: false, value: new oj.Color("fa8072")});
  Object.defineProperty(oj.Color, "SANDYBROWN", {writable: false, value: new oj.Color("f4a460")});
  Object.defineProperty(oj.Color, "SEAGREEN", {writable: false, value: new oj.Color("2e8b57")});
  Object.defineProperty(oj.Color, "SEASHELL", {writable: false, value: new oj.Color("fff5ee")});
  Object.defineProperty(oj.Color, "SIENNA", {writable: false, value: new oj.Color("a0522d")});
  Object.defineProperty(oj.Color, "SILVER", {writable: false, value: new oj.Color("c0c0c0")});
  Object.defineProperty(oj.Color, "SKYBLUE", {writable: false, value: new oj.Color("87ceeb")});
  Object.defineProperty(oj.Color, "SLATEBLUE", {writable: false, value: new oj.Color("6a5acd")});
  Object.defineProperty(oj.Color, "SLATEGRAY", {writable: false, value: new oj.Color("708090")});
  Object.defineProperty(oj.Color, "SLATEGREY", {writable: false, value: new oj.Color("708090")});
  Object.defineProperty(oj.Color, "SNOW", {writable: false, value: new oj.Color("fffafa")});
  Object.defineProperty(oj.Color, "SPRINGGREEN", {writable: false, value: new oj.Color("00ff7f")});
  Object.defineProperty(oj.Color, "STEELBLUE", {writable: false, value: new oj.Color("4682b4")});
  Object.defineProperty(oj.Color, "TAN", {writable: false, value: new oj.Color("d2b48c")});
  Object.defineProperty(oj.Color, "TEAL", {writable: false, value: new oj.Color("008080")});
  Object.defineProperty(oj.Color, "THISTLE", {writable: false, value: new oj.Color("d8bfd8")});
  Object.defineProperty(oj.Color, "TOMATO", {writable: false, value: new oj.Color("ff6347")});
  Object.defineProperty(oj.Color, "TURQUOISE", {writable: false, value: new oj.Color("40e0d0")});
  Object.defineProperty(oj.Color, "VIOLET", {writable: false, value: new oj.Color("ee82ee")});
  Object.defineProperty(oj.Color, "WHEAT", {writable: false, value: new oj.Color("f5deb3")});
  Object.defineProperty(oj.Color, "WHITE", {writable: false, value: new oj.Color("fff")});
  Object.defineProperty(oj.Color, "WHITESMOKE", {writable: false, value: new oj.Color("f5f5f5")});
  Object.defineProperty(oj.Color, "YELLOW", {writable: false, value: new oj.Color("ff0")});
  Object.defineProperty(oj.Color, "YELLOWGREEN", {writable: false, value: new oj.Color("9acd32")});

  Object.defineProperty(oj.Color, "TRANSPARENT", {writable: false, value: new oj.Color("rgba(0,0,0,0)")});

};

/**
 * Returns a valid alpha value [0);1] with all invalid values being set to 1
 * @param {number} a alpha value
 * @returns {number} alpha value
 * @memberof oj.Color
 * @private
 */
function boundAlpha(a)
{
  a = parseFloat(a);

  if (isNaN(a) || a < 0 || a > 1)
  {
     a = 1;
  }

  return a;
}


/**
 * Take input from [0, n] and return it as [0, 1]
 * @param {string|number} n The input value
 * @param {number} max The max value
 * @returns {number}
 * @memberof oj.Color
 * @private
 */
function bound01(n, max)
{
   if (isOnePointZero(n)) { n = "100%"; }

   var processPercent = isPercentage(n);
   n = mathMin(max, mathMax(0, parseFloat(n)));

   // Automatically convert percentage into number
   if (processPercent)
   {
     n = parseInt(n * max, 10) / 100;
   }

   // Handle floating point rounding errors
   if ((Math.abs(n - max) < 0.000001))
   {
     return 1;
   }

   // Convert into [0, 1] range if it isn't already
   return (n % max) / parseFloat(max);
}

/**
 * Force a number between 0 and 1
 * @param {number} val Number to clamp
 * @returns {number}
 * @memberof oj.Color
 * @private
 */
function clamp01(val)
{
    return mathMin(1, mathMax(0, val));
};

/**
 * Parse a base-16 hex value into a base-10 integer
 * @param {number} val The input value
 * @returns {number}
 * @memberof oj.Color
 * @private
 */
function parseIntFromHex(val)
{
   return parseInt(val, 16);
}

/**
 * Need to handle 1.0 as 100%, since once it is a number, there is no difference between it and 1
 * <http://stackoverflow.com/questions/7422072/javascript-how-to-detect-number-as-a-decimal-including-1-0>
 * @param {string|number} n The input value
 * @returns {boolean}
 * @memberof oj.Color
 * @private
 */
function isOnePointZero(n)
{
   return typeof n == "string" && n.indexOf('.') != -1 && parseFloat(n) === 1;
}

/**
 * Returns true if supplied string is a percentage
 * @param {string|number} n The input value
 * @returns {boolean} true if supplied string is a percentage
 * @memberof oj.Color
 * @private
 */
function isPercentage(n)
{
   return typeof n === "string" && n.indexOf('%') != -1;
}

/**
 * Force a hex value to have 2 characters
 * @param {string} c The hex value
 * @returns {string} hex value with 2 characters
 * @memberof oj.Color
 * @private
 */
function pad2(c)
{
   return c.length == 1 ? '0' + c : '' + c;
}

/**
 * Replace a decimal with it's percentage value
 * @memberof oj.Color
 * @private
 */
function convertToPercentage(n)
{
   if (n <= 1)
   {
     n = (n * 100) + "%";
   }

   return n;
}

/**
 * Converts a decimal to a hex value
 * @param {number} d The decimal value
 * @memberof oj.Color
 * @private
 */
function convertDecimalToHex(d)
{
   return Math.round(parseFloat(d) * 255).toString(16);
};

/**
 * Converts a hex value to a decimal
 * @param {number} h The hex value
 * @memberof oj.Color
 * @private
 */
function convertHexToDecimal(h)
{
   return (parseIntFromHex(h) / 255);
}

var matchers = (function()
 {
    // <http://www.w3.org/TR/css3-values/#integers>
    var CSS_INTEGER = "[-\\+]?\\d+%?";

    // <http://www.w3.org/TR/css3-values/#number-value>
    var CSS_NUMBER = "[-\\+]?\\d*\\.\\d+%?";

    // Allow positive/negative integer/number.  Don't capture the either/or, just the entire outcome.
    var CSS_UNIT = "(?:" + CSS_NUMBER + ")|(?:" + CSS_INTEGER + ")";

    // Actual matching.
    // Parentheses and commas are optional, but not required.
    // Whitespace can take the place of commas or opening paren
    var PERMISSIVE_MATCH3 = "[\\s|\\(]+(" + CSS_UNIT + ")[,|\\s]+(" + CSS_UNIT + ")[,|\\s]+(" + CSS_UNIT + ")\\s*\\)?";
    var PERMISSIVE_MATCH4 = "[\\s|\\(]+(" + CSS_UNIT + ")[,|\\s]+(" + CSS_UNIT + ")[,|\\s]+(" + CSS_UNIT + ")[,|\\s]+(" + CSS_UNIT + ")\\s*\\)?";

    return {
        CSS_UNIT: new RegExp(CSS_UNIT),
        rgb:      new RegExp("rgb" + PERMISSIVE_MATCH3),
        rgba:     new RegExp("rgba" + PERMISSIVE_MATCH4),
        hsl:      new RegExp("hsl" + PERMISSIVE_MATCH3),
        hsla:     new RegExp("hsla" + PERMISSIVE_MATCH4),
        hsv:      new RegExp("hsv" + PERMISSIVE_MATCH3),
        hsva:     new RegExp("hsva" + PERMISSIVE_MATCH4),
        hex3:     /^#?([0-9a-fA-F]{1})([0-9a-fA-F]{1})([0-9a-fA-F]{1})$/,
        hex6:     /^#?([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/,
        hex8:     /^#?([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/
    };
 })();


/**
 * Accepts a single string/number and checks to see if it looks like a CSS unit
 * (see `matchers` above for definition).
 * @param {string|number} color  The value to check.
 * @returns {boolean} true if valid CSS
 * @memberof oj.Color
 * @private
 */
function isValidCSSUnit(color)
{
   return !!matchers.CSS_UNIT.exec(color);
}

/**
 * Permissive color string parsing.  Takes in a number of formats, and outputs an object
 * based on detected format.
 * @param {string}  color  A CSS color specification string.
 * @return {Object | boolean}  Returns an object with properties r, g, b  or  h, s, l  or  h, s, v,
 *                             or false if parsing failed.
 * @memberof oj.Color
 * @private
 */
function stringInputToObject(color)
{
  color = color.replace(trimLeft,'').replace(trimRight, '').toLowerCase();

  if (color == 'transparent')
  {
     return { "r": 0, "g": 0, "b": 0, "a": 0};
  }


  // Strategy:
  // 1) Try to match string input using regular expressions.
  // 2) Keep most of the number bounding out of this function - don't worry about [0,1] or [0,100] or [0,360]
  // 3) Just return an object and let the conversion functions handle that.
  // This way the result will be the same whether the oj.Color is initialized with string or object.
  var match;
  if ((match = matchers.rgb.exec(color)))
  {
    return { "r": match[1], "g": match[2], "b": match[3] };
  }
  if ((match = matchers.rgba.exec(color)))
  {
    return { "r": match[1], "g": match[2], "b": match[3], "a": match[4] };
  }
  if ((match = matchers.hsl.exec(color)))
  {
    return { "h": match[1], "s": match[2], "l": match[3] };
  }
  if ((match = matchers.hsla.exec(color)))
  {
    return { "h": match[1], "s": match[2], "l": match[3], "a": match[4] };
  }
  if ((match = matchers.hsv.exec(color)))
  {
    return { "h": match[1], "s": match[2], "v": match[3] };
  }
  if ((match = matchers.hsva.exec(color)))
  {
    return { "h": match[1], "s": match[2], "v": match[3], "a": match[4] };
  }
  if ((match = matchers.hex6.exec(color)))
  {
     return {
              "r": parseIntFromHex(match[1]),
              "g": parseIntFromHex(match[2]),
              "b": parseIntFromHex(match[3])
            };
  }
  if ((match = matchers.hex3.exec(color)))
  {
    return {
             "r": parseIntFromHex(match[1] + '' + match[1]),
             "g": parseIntFromHex(match[2] + '' + match[2]),
             "b": parseIntFromHex(match[3] + '' + match[3])
           };
  }
/*    Wait for CSS4
  if ((match = matchers.hex8.exec(color)))
  {
    return {
             "a": convertHexToDecimal(match[1]),
             "r": parseIntFromHex(match[2]),
             "g": parseIntFromHex(match[3]),
             "b": parseIntFromHex(match[4])
           };
  }
*/

//console.log("matchers false") ;
  return false;
};

_initColorConstants() ;


})();

/**
 *  Copyright (c) 2015, Oracle and/or its affiliates.
 *  All rights reserved.
 */

/*---------------------------------------------------------
   ojColorPalette    Jet Color Palette element
   Depends:   jquery.ui.core.js
              jquery.ui.widget.js
----------------------------------------------------------*/

(function ()
{

/*
  function debugObj(o)  {
    var s ;
    try { s = JSON.stringify(o) ; }
    catch (e) { s = "ERROR";}
    return s ;
  };
*/

  //  ojColorPalette class names
  var   /** @const */   OJCP_CONTAINER         = "oj-colorpalette-container",
        /** @const */   OJCP_GRID              = "oj-colorpalette-grid",
        /** @const */   OJCP_LIST              = "oj-colorpalette-list",
        /** @const */   OJCP_TEXT              = "oj-colorpalette-swatch-text",
        /** @const */   OJCP_LIST_CONTAINER    = "oj-listview-container",
        /** @const */   OJCP_LIST_ELEMENT      = "oj-listview-element",
        /** @const */   OJCP_LIST_ITEM_ELEMENT = "oj-listview-item-element",
        /** @const */   OJCP_SELECTED          = "oj-selected";

  // Misc translation keys
  var  /** @const */  TRANSKEY_NONE  = "labelNone" ;


 /**
  * @ojcomponent oj.ojColorPalette
  * @augments oj.editableValue
  * @since 3.0.0
  * @ojstatus preview
  * @class oj.ojColorPalette
  * @ojshortdesc Displays a set of pre-defined colors from which a specific color can be selected.
  * @ojsignature [{
  *                target: "Type",
  *                value: "class ojColorPalette extends editableValue<oj.Color, ojColorPaletteSettableProperties>"
  *               },
  *               {
  *                target: "Type",
  *                value: "ojColorPaletteSettableProperties extends editableValueSettableProperties<oj.Color>",
  *                for: "SettableProperties"
  *               }
  *              ]
  *
  * @classdesc
  * <h3 id="#colorPaletteOverview-section">
  *   JET Color Palette
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#colorPaletteOverview-section"></a>
  * </h3>
  * The Jet Color Palette element allows an application to display a set of pre-defined
  * colors from which a specific color can be selected. The palette's content is specified as a list
  * of color objects containing an oj.Color.
  * <pre class="prettyprint">
  * <code>&lt;oj-color-palette palette="[[myPalette]]" value="{{colorValue}}">
  * &lt;/oj-color-palette>
  * </code></pre>
  * {@ojinclude "name":"validationAndMessagingDoc"}
  * 
  * <p>
  * <h3 id="keyboard-section">
  *   Keyboard End User Information
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
  * </h3>
  *
  * {@ojinclude "name":"keyboardDoc"}
  *
  * <p>
  *
  *
  * <h3 id="rtl-section">
  *   Reading direction
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
  * </h3>
  *
  * <p>As with any JET element, in the unusual case that the directionality (LTR or RTL) changes post-init, the color palette must be <code class="prettyprint">refresh()</code>ed.
  *
  */
  oj.__registerWidget("oj.ojColorPalette", $['oj']['editableValue'],
  {
     widgetEventPrefix : "oj",
     defaultElement    : "<input>",

     options: {

         /**
          * Labelled-by is used to establish a relationship between this and another element.
          * A common use is to tie the oj-label and the oj-color-palette together.
          * The oj-label custom element has an id, and you use the labelled-by attribute
          * to tie the two elements together.
          *
          * @ojshortdesc Used to establish a relationship between this element and another element.
          * @example <caption>Initialize the color palette with the <code class="prettyprint">labelled-by</code> attribute specified:</caption>
          * &ltoj-label id="labelId">Name:&lt/oj-label>
          * &ltoj-color-palette labelled-by="labelId">&lt;/oj-color-palette>
          *
          * @example <caption>Get or set the <code class="prettyprint">labelled-by</code> property, after initialization:</caption>
          * // getter
          * var labelledBy = myColorPalette.labelledBy;
          *
          * // setter
          * myColorPalette.labelledBy = "labelId";
          *
          * @expose
          * @type {?string}
          * @default null
          * @instance
          * @memberof oj.ojColorPalette
          */
         labelledBy: null,

          /**
           * Specify an array of objects defining the palette's color set, and optionally, descriptive labels for the colors.
           * Each object has the following structure:
           * @property {oj.Color} color the color definition
           * @property {string} [label] optional descriptive string (refer to attribute <em>label-display</em>).
           *                                         If omitted, <em>label</em> defaults to the color's hex string format.
           *
           * @member
           * @type {Array.<Object>}
           * @ojsignature  {target: "Type", value: "Array<{color: oj.Color, label?: string}>"} 
           * @default null
           * @ojshortdesc Specifies an array of objects defining the palette's color set.
           * @example <caption>Initialize the color palette with the <code class="prettyprint">palette</code> attribute specified:</caption>
           * &ltoj-color-palette palette='[[myPalette]]'>&lt;/oj-color-palette>
           *
           * @example <caption>Get or set the <code class="prettyprint">palette</code> property, after initialization:</caption>
           * // Get one
           * var oneColor = myColorPalette.palette[0];
           *
           * // Get all
           * var myPallete = myColorPalette.palette;
           *
           * // Set all
           * var palette = [{color:new oj.Color('#ffffff'), label: 'White'},
              {color: new oj.Color('#77bb99'), label: 'Std text'},
              . . .];
           * myColorPalette.palette = palette;
           * @expose
           * @instance
           * @memberof oj.ojColorPalette
           * @ojtranslatable
           */
          palette:       null,

          /**
           * The swatch size.  If the size is <em>'sm'</em> or <em>'xs'</em>, the color <em>label</em> property is used as a tooltip.
           * @member
           * @type {string}
           * @default "lg"
           * @ojshortdesc Specifies the swatch size.
           * @ojvalue {string} "xs" extra small swatch
           * @ojvalue {string} "sm" small swatch
           * @ojvalue {string} "lg" large swatch
           *
           * @example <caption>Initialize the color palette with the <code class="prettyprint">swatch-size</code> attribute specified:</caption>
           * &ltoj-color-palette swatch-size="lg">&lt;/oj-color-palette>
           *
           * @example <caption>Get or set the <code class="prettyprint">swatchSize</code> property, after initialization:</caption>
           * // getter
           * var swatchSize = myColorPalette.swatchSize;
           *
           * // setter
           * myColorPalette.swatchSize = "lg";
           * @expose
           * @instance
           * @memberof oj.ojColorPalette
           */
          swatchSize:    'lg',

          /**
           * Specifies whether a text label accompanies the color swatch.
           * @member
           * @type {string}
           * @default "off"
           * @ojshortdesc Specifies whether a text label accompanies the color swatch.
           * @ojvalue {string} "auto" labels are displayed if the <em>layout</em> property is <em>'list'</em> and swatch-size is <em>'sm'</em> or if the <em>layout</em> is <em>'grid'</em> and <em>swatch-size</em> is <em>'lg'</em>
           * @ojvalue {string} "off" labels are not displayed
           *
           * @example <caption>Initialize the color palette with the <code class="prettyprint">label-display</code> attribute specified:</caption>
           * &ltoj-color-palette label-display="auto">&lt;/oj-color-palette>
           *
           * @example <caption>Get or set the <code class="prettyprint">labelDisplay</code> property, after initialization:</caption>
           * // getter
           * var labelDisplay = myColorPalette.labelDisplay;
           *
           * // setter
           * myColorPalette.labelDisplay = "auto";
           * @expose
           * @instance
           * @memberof oj.ojColorPalette
           */
          labelDisplay:  "off",
          /**
           * Specifies the layout of the color swatches.
           * @member
           * @type {string}
           * @default "grid"
           * @ojshortdesc Specifies the layout of the color swatches.
           * @ojvalue {string} "grid" Layout the color swatches in a grid
           * @ojvalue {string} "list" Layout the color swatches in a list
           * @expose
           * @example <caption>Initialize the color palette with the <code class="prettyprint">layout</code> attribute specified:</caption>
           * &ltoj-color-palette layout="grid">&lt;/oj-color-palette>
           *
           * @example <caption>Get or set the <code class="prettyprint">layout</code> property, after initialization:</caption>
           * // getter
           * var layout = myColorPalette.layout;
           *
           * // setter
           * myColorPalette.layout = "grid";
           * @instance
           * @memberof oj.ojColorPalette
           */
          layout:        'grid',
          /**
           * The current value of the palette element.
           * @member
           * @type {oj.Color}
           * @default null
           * @ojshortdesc The current value of the palette element.
           * @ojwriteback
           * @expose
           * @instance
           * @memberof oj.ojColorPalette
           * @example <caption>Initialize the color palette with the <code class="prettyprint">value</code> attribute specified:</caption>
           * &ltoj-color-palette value='{{myColor}}'>&lt;/oj-color-palette>
           *
           * @example <caption>Get or set the <code class="prettyprint">value</code> property, after initialization:</caption>
           * // getter
           * var color = myColorPalette.value;
           *
           * // setter
           * myColorPalette.value = new oj.Color('rgb(0,0,0)');
           */
          value:         null,
     },   // end options


     //** @inheritdoc */
     getNodeBySubId : function(locator)
     {
       if (locator == null) {
         return this.element ? this.element[0] : null;
       }

       var subId = locator['subId'],
           index = locator["index"],
           ret   = this._super(locator),
           elems;

       if (!ret)
       {
        switch (subId)
        {
          case 'oj-palette-entry' :
                                      elems = this._$LV.find(".oj-listview-item") ;
                                      if (elems.length && index < elems.length)
                                      {
                                        ret = elems[index] ;
                                      }
                                      break ;
        }
      }

      return ret ;
     },

     //** @inheritdoc */
     getSubIdByNode : function(elem)
     {
        var $node = $(elem),
            subId = null,
            index = -1,
            ret   = null,
            id, elems, i ;

        if ($node.is("li") && $node.hasClass(OJCP_LIST_ITEM_ELEMENT))
        {
          subId = 'oj-palette-entry' ;
          id    = $node.attr("id") ;
          elems = this._$LV.find(".oj-listview-item") ;
          $.each(elems, function(i, obj)
              {
                if ($(obj).attr("id") === id)
                {
                   index = i ;
                   return false ;
                }
              });
        }

        if (subId)
        {
           ret = {'subId': subId} ;
           if (index >= 0)
           {
             ret['index'] = index ;
           }
        }

        if (ret == null)
          ret = this._super(elem);

        return ret ;
     },

    /**
     * Add an entry to the palette.
     * @param {oj.Color | Object}   newEntry  An oj.Color object specfying the color to be added,
     * or an object of the same format as used for defining a palette - see the <code class="prettyprint">palette</code> option.
     * @return {void}
     * @memberof oj.ojColorPalette
     * @expose
     * @ojdeprecated {since:"4.0.0", description: "This is deprecated for API consistency with the remove function. The same functionality can be achieved by mutating the observable array that is set as the palette."}
     * @ignore
     * @instance
     */
     add : function(newEntry)
     {
        var o = null, c ;

        if (newEntry instanceof oj.Color)
        {
           o = { "color" : newEntry} ;
        }
        else if (typeof newEntry === "object")
        {
          c = newEntry["color"] ;
          if (c instanceof oj.Color)
          {
            o = newEntry ;
          }
        }

        // Perform the add.
        if (o)
        {
           // Add a palette busy state for the add
           // The busy state resolver will be invoked when ListView completes the add.
           this._setPaletteBusyContext("The swatch add is being animated in the palette (id=" +
                                       this.element.attr("id") + ").");

           // Perform the add
           o["id"] = this._getNewSwatchId() ;  // give the new swatch an id
           if (!this._opStack) {
             this._opStack = [] ;              // fifo stack of post-busy operations
           }
           this._opStack.push({op: "a", obj: o}) ;
           this._palDataSource.add(o) ;        // ListView will render the new entry
           this._waitForLV() ;                 // complete the add when LV is complete
        }
     },

     /**
      * Remove an entry from the palette.
      * @param {Object | number | oj.Color}  palEntry  Can be the zero-based index to the entry, or an
      * object containing the color specfication (using the same format as used for defining the
      * palette - see the <code class="prettyprint">palette</code> option, or an oj.Color object.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @expose
      * @ojdeprecated {since:"4.0.0", description: "This is deprecated due to a name collision with HTMLElement. The same functionality can be achieved by mutating the observable array that is set as the palette."}
      * @ignore
      * @instance
      */
     remove : function(palEntry)
     {
        var index = -1,
            id = null, t, c ;

        t = (typeof palEntry) ;
        if (t === "number")              // index
        {
           if ((palEntry >= 0) && (palEntry < this._palette.length))
           {
             id = this._palette[palEntry]["id"] ;
           }
           else
           {
             oj.Logger.error("JET Color Palette (id='" + this.element.attr("id") + "'): Invalid index for remove (" + palEntry + ")");
           }
        }
        else if (t === "object")
        {
          c  = (palEntry instanceof oj.Color)? palEntry : palEntry["color"];
          id = this._findSwatchIdOfColorInPalette(c)
        }

        if (id)
        {
           palEntry = { "id" : id } ;

           // Add a palette busy state for the remove
           // The busy state resolver will be invoked when ListView completes the remove.
           this._setPaletteBusyContext("The removed swatch is being animated in the palette (id=" +
                                       this.element.attr("id") + ").");
           // Perform the remove
           if (! this._opStack) {
              this._opStack = [] ;      // fifo stack of post-busy operations
           }
           this._opStack.push({op: "r", obj:palEntry}) ;
           this._palDataSource.remove(palEntry) ;    // LV will re-render

          this._waitForLV() ;                        // complete the remove when LV is complete
        }
     },


     /**
      * Returns a Promise that resolves when the component is ready and has finished rendering.
      * This method does not accept any arguments.
      * @ojshortdesc Returns a Promise that resolves when the component is ready and has finished rendering.
      * @memberof oj.ojColorPalette
      * @expose
      * @ignore
      * @instance
      * @return {Promise} A Promise that resolves when the component is ready.
      */
     whenReady: function()
     {
        var self = this ;
        var promise = new Promise(function(resolve, reject)  {
             self._$LV.ojListView("whenReady")
                 .then(function(zz) {
                                      resolve(true);
                                    });
        });
        return promise ;
     },

     /**
      * Destroy the Color Palette
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @override
      * @protected
      */
     _destroy: function()
     {
        this._resolvePaletteBusyContext();

        if (this._$LV)
        {
          this._opStack = [] ;
          this._$LV.ojListView("destroy") ;
        }
        this._palDataSource  = null ;
        this._$paletteContainer.remove() ;          // remove our markup from dom
        this._$boundElem.removeClass("oj-colorpalette");
        this._clear() ;
        this._super();
     },

     /**
      * Called the first time the widget is called on an element.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _ComponentCreate: function ()
     {
        this._super() ;
        this._initPalette() ;
     },

    /**
     * @return {void}
     * @memberof oj.ojColorPalette
     * @instance
     * @private
     */
    _AfterCreate: function()
    {
      var label;

      this._super();

      this._updateLabelledBy(null, this.options['labelledBy'], this._$LV);

      // custom element's use oj-label.
      if (!this._IsCustomElement())
        label = this._GetLabelElement();

      // Apply the label to the listview
      if (label)
      {
        // Set the aria-labelledby attribute of the listview to the returned id
        var labelId = label.attr('id');
        if (!labelId)
        {
          oj.Logger.warn("JET Color Palette: The label for this component needs an id in order to be accessible");
        }
        else
          this._$LV.attr('aria-labelledby', labelId);
      }
      else
      {
        // Check if the element has aria-label
        var ariaLabelString = this.element.attr('aria-label');
        if (ariaLabelString)
        {
          // Set the aria-label of the listview to the returned string
          this._$LV.attr('aria-label', ariaLabelString);
        }
      }
    },

     /**
      * Handle an option change.
      * Called by $(selector).ojColorPalette("option", "prop", value)
      * @param {string}   key
      * @param {string | oj.Color | boolean}   newval
      * @param {Object}   flags
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setOption: function (key, newval, flags)
     {
        var originalValue = this.options["labelledBy"];
        var bSkip = false;

        switch (key)
        {
           case "value"  :        // Fix  - COULD NOT CHANGE A PALETTE VALUE FROM JAVASCRIPT
                                  // If the value has changed and if the value is in the palette, 
                                  // ojlistview would trigger selection event for the color in palette (see method _onLVOptionChange).
                                  // The selection handler (_selected) would set new value (in method _SetValue), which would
                                  // internally set the new value in options and fires option change event.
                                  // So in that case, skip calling _super below to avoid duplicate firing of option change event.
                                  bSkip = this._setOptValue(newval) ;
                                  break ;
           case "palette" :       this._setOptPalette(newval) ;
                                  break ;
           case "swatchSize" :    this._setOptSwatchSize(newval) ;
                                  break ;
           case "layout" :        this._setOptLayout(newval) ;
                                  break ;
           case "labelDisplay" :  this._setOptLabelDisplay(newval) ;
                                  break ;
           case "disabled":       this._setOptDisabled(newval, true) ;
                                  break;
           case "labelledBy":     // remove the old one and add the new one
                                  this._updateLabelledBy(originalValue, newval, this._$LV);
                                  break;
        }

        if (!bSkip)
          this._super(key, newval, flags);
     },

    /**
     * If custom element, get the labelledBy option, and set this
     * onto the root dom element as aria-labelledby. We append "|label" so it matches the id that
     * is on the oj-label's label element.
     * @memberof oj.ojColorPalette
     * @instance
     * @private
     */
    _updateLabelledBy : oj.EditableValueUtils._updateLabelledBy,

     /**
      * @param {Event} event the associated Event object.
      * @param {Object} ui the context object.
      * @return {void}
      * @private
      * @memberof oj.ojColorPalette
      * @instance
      */
     _onLVOptionChange: function(event, ui)
     {
       if (ui.option === "selection") {
         this._selected(event, ui) ;
       }
     },


     /**
      * Wait for ListView to be ready, and complete the add/remove
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _waitForLV: function()
     {
        if (this._LVResolve) {
          return ;
        }

        var self = this;
        this._LVResolve = oj.Context.getContext(this._$LV[0]).getBusyContext();
        this._LVResolve.whenReady().then(function ()
        {
           self._LVResolve = null ;

           var stackEntry, thisObj, postOp, newPalette, i, index ;

           // Resolve palette's busy state
           self._resolvePaletteBusyContext() ;

           if (! self._opStack) {
             return;
           }

           postOp = false ;
           for (i = 0; i < self._opStack.length; i++)
           {
             stackEntry = self._opStack[i] ;
             thisObj    = stackEntry.obj ;

             if ((stackEntry.op === "a") || (stackEntry.op === "r"))
             {
               postOp = true ;
               if (! newPalette) {
                 newPalette = self._palette.slice(0) ;
               }

               if (stackEntry.op === "a")          // add swatch
               {
                 newPalette.push(thisObj) ;
               }
               else if (stackEntry.op === "r")     // remove swatch
               {
                 index = self._findIndexOfSwatchById(newPalette, thisObj["id"]) ;
                 newPalette.splice(index, 1) ;
                 stackEntry.index = index ;      // save doing this again later
               }
             }
           }

           // Fire the optionChange event for 'palette' add/remove
           if (postOp)
           {
             // Fire the optionChange event for 'palette'
             self._fireOptionChangeEvent("palette", newPalette, null);

             // internalSet is true in fireOptionChangeEvent, so update _palette now.
             while(self._opStack.length)
             {
               stackEntry = self._opStack.shift() ;
               thisObj    = stackEntry.obj ;

               if (stackEntry.op === "a")          // add swatch
               {
                 self._palette.push(thisObj) ;
               }
               else if (stackEntry.op === "r")     // remove swatch
               {
                 self._palette.splice(stackEntry.index, 1) ;
               }
             }
           }
        }, function() {
             oj.Logger.error("JET Color Palette (id='" + self.element.attr("id") +
                             "'): ListView timed out.");
             self._opStack = [] ;
        });
     },


     /**
      * Compares two color values (oj.Colors)
      * @param {oj.Color}   color1   a color to match
      * @param {oj.Color}   color2   a color to match
      * @returns {boolean}  true if colors match, else false.
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _compareColorValues  :  function(color1, color2)
     {
        var t1  = (color1 instanceof oj.Color),
            t2  = (color2 instanceof oj.Color),
            ret = false ;

        if (t1 && t2)
        {
          ret = color1.isEqual(color2) ;
        }
        return ret ;
     },

     /**
      * Fire optionChange event
      * @param {string}  key             the option key whose property value has been changed.
      * @param {Object | null}  newVal   the new value after the change
      * @param {Event | null} origEvent  false if option change is not due to user interaction.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _fireOptionChangeEvent : function(key, newVal, origEvent)
     {
        if (key === "palette")  {
          this.option(key, newVal, {'_context' : { originalEvent: origEvent,
                                                   internalSet: true
                                                 },
                                    'changed': true     // don't need comparison check
                                   });
        }    // end if "value"
     },

     /**
      * Find the supplied oj.Color in the palette and return the index to it.
      * @param {oj.Color}  color  the color to be found.
      * @returns {number}  the index in the palette array, or -1 if not found.
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _findColorInPalette: function(color)
     {
        var  index = -1,
             a     = this._palette,
             l     = a.length,
             i, co ;

        for (i = 0; i < l; i++)
        {
          co = a[i] ;
          if (color.isEqual(co["color"]))
          {
             index = i ;
             break ;
          }
        }

        return index ;
     },


     /**
      * Returns the swatch ID of the supplied oj.Color in the palette.
      * @param {oj.Color}  color  the color to be found.
      * @returns {string | null}  the ID of the swatch, or null if not found.
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _findSwatchIdOfColorInPalette: function(color)
     {
        var  id = null,
             a     = this._palette,
             l     = a.length,
             i, co ;

        for (i = 0; i < l; i++)
        {
          co = a[i] ;
          if (color.isEqual(co["color"]))
          {
             id = co["id"] ;
             break ;
          }
        }

        return id ;
     },


     /**
      * Returns the index to the swatch with the supplied ID in the array.
      * @param {string}  a   the array to be searched.
      * @param {string}  id   the ID to be found.
      * @returns {number}  the index to the swatch, or -1 if not found.
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _findIndexOfSwatchById: function(a, id)
     {
        var  index = -1,
             l     = a.length,
             i, co ;

        for (i = 0; i < l; i++)
        {
          co = a[i] ;
          if (co["id"] === id)
          {
             index = i ;
             break ;
          }
        }

        return index ;
     },


     /**
      * set Palette BusyContext
      * @param {string} description
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setPaletteBusyContext : function(description)
     {
        // The busy state resolver will be invoked when ListView completes the add.
        if (! this._resolve)
        {
          var busyContext  = oj.Context.getContext(this.element[0]).getBusyContext();
          this._resolve = busyContext.addBusyState( {"description" : description} );
        }
      },

     /**
      * Resolve PaletteBusyContext
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _resolvePaletteBusyContext : function()
     {
        if (this._resolve)
        {
          this._resolve() ;
          this._resolve = null ;
        }
     },

     /**
      * Swatch renderer  called from ojListView.
      * @param {Object}  context the ojListView context.
      * @return {Object}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _renderer : function(context)
     {
       var raw ;

       var color, label, haveLabel, showLabels, none, tooltip;

       color = context["data"]["color"] ;
       if (!(color instanceof oj.Color))
       {
          // If color is undefined, will substitute black.
          color = oj.Color.BLACK;
          oj.Logger.warn("JET Color Palette (id='" + this.element.attr("id") + "'): Substituting oj.Color.BLACK for an object that is not an instance of oj.Color");
       }

       label = haveLabel = context["data"]["label"] ;
       showLabels = ((this._labelDisplay === 'auto' && this._layout === 'list' && this._swatchSize === 'sm') ||
                     (this._labelDisplay === 'auto' && this._layout === 'grid' && this._swatchSize === 'lg'));

       if (color != null)
       {
          tooltip = label? label : this._convHex.format(color) ;

         if (showLabels)
         {
           label = tooltip? tooltip : this._convHex.format(color) ;
           label = haveLabel? label : label.toUpperCase() ;
         }
         else
         {
           label = null ;
         }

         none  = !!(this._isTransparent(color) || (label && label.toLowerCase() === "none"));
       }

       var selectedClass = "" ;
       if (this._initSelection === context['data']['id'])
       {
         selectedClass       = OJCP_SELECTED ;
         this._initSelection = -1 ;
         // The swatch is not rendered yet so will note the parent.  Will find the
         // swatch to remove the selection highlighting on the next selection.
         this._selectedParent = context['parentElement'] ;
       }

       var swatchClass ;
       if (this._layout === 'list')
       {
         swatchClass = "oj-colorpalette-swatchsize-" + this._swatchSize +
           (none ? " oj-colorpalette-swatch-none" : "");
       }
       else
       {
         swatchClass =  this._swatchClass + (none ? " oj-colorpalette-swatch-none" : "") ;
       }

       if (none)
       {
         // transparent color required
         raw =  this._renderNone(showLabels, label, tooltip, swatchClass, selectedClass); // transparent color
       }
       else {
         //  standard swatch
         raw = this._renderStandard(color, showLabels, label, tooltip, swatchClass, selectedClass);
       }

       return raw ;
     },

     /**
      * Render a standard swatch
      * @param {oj.Color}  color
      * @param {boolean}   showLabels
      * @param {string}    label
      * @param {string}    tooltip
      * @param {string}    swatchClass
      * @param {string}    selectedClass
      * @return {Object}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _renderStandard : function(color, showLabels, label, tooltip, swatchClass, selectedClass)
     {
        var raw;

        raw = "<div class='oj-colorpalette-swatch-entry " + swatchClass +
                (showLabels ? " oj-colorpalette-swatch-showlabel" : "") + "'>" +
                "<div class='oj-colorpalette-swatch-container'>" +
                  "<div class='oj-colorpalette-swatch " + selectedClass +
                    "' style='background-color:" + color.toString() + "'" +
                    ((! label)? " title='" + tooltip + "'" : "") + ">" +
                  "</div>" +
                "</div>" ;

        if (label) {
          raw += "<span class='oj-colorpalette-swatch-text'>" + label + "</span>" ;
        }
        raw += "</div>" ;
        return  $(raw)[0] ;
     },

     /**
      * Render a 'none' swatch.
      * @param {boolean}   showLabels
      * @param {string} label
      * @param {string}    tooltip
      * @param {string} swatchClass
      * @param {string} selectedClass
      * @return {Object}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _renderNone: function(showLabels, label, tooltip, swatchClass, selectedClass)
     {
        var raw;

        raw = "<div class='oj-colorpalette-swatch-entry " + swatchClass +
                (showLabels ? " oj-colorpalette-swatch-showlabel" : "") + "'>" +
                "<div class='oj-colorpalette-swatch-container'>" +
                  "<div class='oj-colorpalette-swatch " + selectedClass + "'" +
                    ((!label) ? " title='" + tooltip + "'" : "") + ">" +
                    "<div class='oj-colorpalette-swatch-none-icon'>" +
                    "</div>" +
                  "</div>" +
                "</div>";

        if (label)
        {
          raw += "<span class='oj-colorpalette-swatch-text'>" + label + "</span>";
        }
        raw += "</div>";
        return $(raw)[0];
     },

     /**
      * Handle selection event of a swatch from ListView
      * @param {Event}   event  the associated Event object.
      * @param {Object}  ui the context object.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _selected : function(event, ui)
     {
         var newColor = null, swatch, lastSelected ;

         // Color the internal swatch border (normally transparent) so that it
         // shows as selected.
         swatch = $(ui.items[0]).find(".oj-colorpalette-swatch") ;
         swatch.addClass(OJCP_SELECTED) ;

         // Remove the last selected swatch highlighting
         lastSelected = this._selectedSwatch ;
         this._selectedSwatch = swatch ;
         if (! lastSelected)
         {
            // Do we have an initial selection to remove
            if (this._selectedParent)
            {
               lastSelected = $(this._selectedParent).find(".oj-colorpalette-swatch")
               this._selectedParent = null ;
            }
         }
         if (lastSelected)
         {
           lastSelected.removeClass(OJCP_SELECTED) ;
         }

         // Fire value-changed event
         if (ui.value.length === 1)
         {
           for (var i=0; i < this._palette.length; i++)
           {
             // listview sends the id in the selection event
             // it may not be the same as the index after palette is updated
             // so comparing with the id here instead of index
             if (this._palette[i]["id"] === ui.value[0])
             {
               newColor = this._palette[i]["color"];
               break;
             }
           }
         }
         else
         {
           // No value, probably because its a deselection caused by a changing the pallete
           // option and the current value is not in the new palette.  Will leave the value
           // as is, and not fire a value change event.
           return ;
         }
         this._SetValue(newColor, event);
         this._value = newColor ;
     },

     /**
      * Handle "disabled" option change.
      * @param {boolean}  disabled  the "disabled" option value.
      * @param {boolean} applyOnlyIfDifferent Only apply the new value if it's different from the
      *        current value.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setOptDisabled : function(disabled, applyOnlyIfDifferent)
     {
        var $swatches, $obj, t,
            change = !applyOnlyIfDifferent ||
                     (applyOnlyIfDifferent && (disabled !== this._disabled)) ;

        if (change)
        {
          if (this._$LV)
          {
            this._$LV.ojListView("option", "disabled", disabled) ;
          }

          $swatches = $(".oj-colorpalette-container .oj-colorpalette-swatch");
          t = this;

          // Enable/diable the appearance of the swatches
          if (disabled)
          {
            //  ListView doesn't show any difference to the disabled items, so do it here
            this._disabledBG = [] ;
            $.each($swatches, function(i, obj) {
                t._disabledBG.push(obj.style.backgroundColor) ;
                obj.style.backgroundColor = "#eee" ;
            });
          }
          else
          {
            if (this._disabledBG && this._disabledBG.length)
            {
              $.each($swatches, function(i, obj) {
                  obj.style.backgroundColor = t._disabledBG[i] ;
              });
            }
            this._disabledBG = null ;
          }

          this._disabled = disabled;
        }
     },

     /**
      * Handle "value" option change.
      * @param {oj.Color}  color  the "value" option color.
      * @return {boolean}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setOptValue :  function(color)
     {
        var palIndex = -1;
        var pal = [] ;

        if (this._palette.length > 0)
        {
          if (color instanceof oj.Color)
          {
            // Check if value has changed
            if (! this._compareColorValues(this._value, color))
            {
               // Color is different from current
               palIndex = this._findColorInPalette(color) ;
               
               // Fix  - COULD NOT CHANGE A PALETTE VALUE FROM JAVASCRIPT 
               // use the id instead of the index
               // this is needed because listview is looking up the id for selection
               if (palIndex >= 0 && this._palette[palIndex]['id'] != null)
               {
                 // Found in palette, so select it
                 var palId = this._palette[palIndex]['id'];
                 pal.push(palId) ;
               }
               this._$LV.ojListView("option", "selection", pal) ;   // select, or deselect if not found
               this._value = color ;
            }
          }
        }

        return (pal.length > 0) ;        // return true if at least one palette color is selected
     },

     /**
      * Handle "palette" option change.
      * @param {Array}  palette  the "palette" option array.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setOptPalette :  function(palette)
     {
        if ($.isArray(palette))
        {
          if (! this._isPaletteEqual(palette, this._palette))
          {
            // Palettes are different
            // Add a palette busy state for the rerender of the ListView
            // The busy state resolver will be invoked when ListView completes the add.
            this._setPaletteBusyContext("The palette (id=" + this.element.attr("id") + ") option change in progress.");

            this._opStack = [] ;                  // clear the fifo stack of post-busy operations

            this._palette = palette.slice(0); ;   // make copy in case app is using same array
            this._initSelection = this._findColorInPalette(this._value) ;
            this._setData(palette, this._initSelection, true) ;

            this._waitForLV() ;
          }
        }
     },

     /**
      * Handle "swatchSize" option change.
      * @param {string} swatchSize
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setOptSwatchSize :  function(swatchSize)
     {
        if (typeof swatchSize === "string")
        {
           if (swatchSize !== this._swatchSize)
           {
             this._swatchSize = swatchSize ;
             this._swatchClass = "oj-colorpalette-swatchsize-" + (swatchSize === "lg"? "lg" :
                                                                 ((swatchSize === "sm")? "sm" : "xs")) ;
             this._$LV.ojListView("refresh") ;
           }
        }
     },

     /**
      * Handle "labelDisplay" option change.
      * @param {string}  labelDisplay  the "labelDisplay" option value.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setOptLabelDisplay :  function(labelDisplay)
     {
        if (typeof labelDisplay === "string")
        {
           if (labelDisplay !== this._labelDisplay)
           {
             if (labelDisplay === "auto" || labelDisplay === "off")
             {
               this._labelDisplay = labelDisplay ;
               this._$LV.ojListView("refresh") ;
             }
           }
        }
     },

     /**
      * Handle "layout" option change.
      * @param {string}  layout  the "layout" option value.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setOptLayout :  function(layout)
     {
        if (typeof layout === "string")
        {
           if (layout !== this._layout)
           {
              this._layout = layout ;
              this._setDisplayFormat() ;
              this._$LV.ojListView("refresh") ;
           }
        }
     },

     /**
      * Update the ListView display format
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setDisplayFormat:  function()
     {
        var grid         = (this._layout === 'grid') ;
        var layoutClass  = (grid ? 'oj-colorpalette-grid' : 'oj-colorpalette-list') ;

        this._$LV.removeClass("oj-colorpalette-grid oj-colorpalette-list oj-listview-card-layout") ;
        this._$LV.addClass(layoutClass) ;
        if (grid)
        {
          this._$LV.addClass("oj-listview-card-layout") ;
        }
     },

     /**
      * Update the ListView data option
      * @param {Array} palette
      * @param {number} initSelected
      * @param {boolean} setOption
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setData : function(palette, initSelected, setOption)
     {
       this._addIdsToPalette(palette) ;            // add "id" props
       this._palDataSource  = new oj.ArrayTableDataSource(palette, {"idAttribute": "id"});

       //  If current value property matches a supplied swatch, select it.
       if (initSelected >= 0)
       {
         // use the id instead of the index
         // this is needed because after palette changes, the id and index may not match
         // and listview is looking up the id for selection
         if (this._palette[initSelected]['id'])
           initSelected = this._palette[initSelected]['id'];
         
         if (this._palInitSelected.length === 0)
         {
           this._palInitSelected.push(initSelected) ;
         }
         else
         {
           this._palInitSelected[0] = initSelected ;
         }
       }

       if (setOption)
       {
         this._$LV.ojListView("option", "data", this._palDataSource) ;
       }
     },


     /**
      * Initializes the widget, examines options and sets up
      * internal data structures.
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _initPalette : function()
     {
        this._setPaletteBusyContext("Palette (id=" + this.element.attr("id")+ ") is initializing.");

        this._initData() ;
        this._setup() ;
     },

     /**
      * Perform setup, and init the ListView
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _setup :  function()
     {
        // Add new listview markup as a child of this component's DOM element
        this._$boundElem.append(this._markup) ;    //@HTMLUpdateOK (strings are all code constants)
        this._$boundElem.addClass("oj-colorpalette");
        this._$paletteContainer = this._$boundElem.find(".oj-colorpalette-container");
        this._$LV               = this._$paletteContainer.find(":first");  //Listview UL

        // Will be using a component context for the ListView
        this._$LV.attr("data-oj-context", "") ;

        //  If value property matches a supplied swatch, note for initial selection.
        if (this._value && this._value instanceof oj.Color)
        {
          this._initSelection = this._findColorInPalette(this._value) ;
        }

        this._swatchId = 0 ;                   // use  _getNextSwatchId() to access
        this._setData(this._palette, this._initSelection, false) ;

        //  Instantiate the ListView
        this._$LV.ojListView({
                               "data":           this._palDataSource,
                               "item":           {"renderer": this._renderer.bind(this)},
                               "optionChange":   this._onLVOptionChange.bind(this),
                               "selectionMode":  'single',
                               "selection":      this._palInitSelected,
                               "rootAttributes": {style: "height:100%;width:100%"}
                  }).attr('data-oj-internal', ''); // for use in automation api

        var self = this;
        var LV_resolve = oj.Context.getContext(this._$LV[0]).getBusyContext();
        LV_resolve.whenReady().then(function()
        {
          // Don't want any listview text if palette is empty
          self._$LV.ojListView("option", "translations.msgNoData", "") ;

          self._setOptDisabled(self._disabled);

          //FIX : when there is a vertical scrollbar, add
          //padding so that no horizontal scrollbar is needed and the
          //text doesn't get cut off or truncated
          if (self._$LV[0].scrollWidth > self._$LV[0].clientWidth)
          {
            var scrollbarWidth = self._getScrollbarWidth();
            var rtl = (self._GetReadingDirection() === "rtl");
            self._$LV.css(rtl ? "padding-left" : "padding-right", scrollbarWidth + 1);
          }

          self._resolvePaletteBusyContext();    // component is ready to use
        });
     },

     /**
      * Set up instance data
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _initData : function()
     {
        this._applyOptions() ;     // process the component options

        this._converterFactory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_COLOR);
        this._convHex  =  this._converterFactory.createConverter({"format": "hex"}) ;

        this._labelNone = this.getTranslatedString(TRANSKEY_NONE);

        var layoutClass ;
        if ( this._layout === 'grid')
        {
          layoutClass =  "oj-colorpalette-grid oj-listview-card-layout" ;
        }
        else
        {
          layoutClass =  "oj-colorpalette-list" ;
        }

        //  Create the inner ListView markup
        this._markup = (function () {
           return [
             "<div class='oj-colorpalette-container'>",
                "<ul class='" + layoutClass + "'/>",
             "</div>"
           ].join("") ;
        })();
     },


     /**
      * Process the component options
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _applyOptions : function()
     {
        var  opts = this.options,
             opt ;

        this._doc             = this.element[0].ownerDocument ;
        this._body            = this._doc.body ;
        this._$boundElem      = $(this.element) ;
        this._disabled        = false ;
        this._palInitSelected = [] ;
        this._palDataSource   = null ;

        opt = opts["swatchSize"] ;
        if (typeof opt === "string") {
          opt = opt.toLowerCase() ;
          if (opt !== "lg" && opt !== "sm" && opt !== "xs")
            opt = "lg" ;
        }
        this._swatchSize  = opt;
        this._swatchClass = "oj-colorpalette-swatchsize-" + opt ;

        opt = opts["labelDisplay"] ;
        if (typeof opt === "string") {
          opt = opt.toLowerCase() ;
          if (opt !== "auto" && opt !== "off")
            opt = "auto" ;
        }
        this._labelDisplay = opt ;

        opt = opts["layout"] ;
        if (typeof opt === "string") {
          opt = opt.toLowerCase() ;
          if (opt !== "grid" && opt !== "list")
            opt = "grid" ;
          if (opt !== "grid" && this._swatchSize === "xs")
          {
            opt = "grid" ;
          }
        }
        this._layout = opt ;

        opt = opts["value"] ;
        if (! (opt instanceof oj.Color))
        {
          opt = null ;
        }
        this._value = (opt? opt : oj.Color.BLACK) ;

        opt = opts["palette"] ;
        if (! $.isArray(opt))
        {
          opt = [] ;
        }
        this._palette = opt.slice(0); ;

        opt = opts["disabled"] ;
        if (typeof opt === "boolean")
        {
          this._disabled = opt ;
        }
     },

     /**
      * Returns true if an oj.Color represents the 'transparent'.
      * @param {oj.Color} color the color object to be tested.
      * @returns {boolean}  true if oj.Color is transparent, else false.
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
    _isTransparent : function(color)
    {
       var r = color.getRed(),
           g = color.getGreen(),
           b = color.getBlue(),
           a = color.getAlpha();

       return (r === 0 && g === 0 && b === 0 && a === 0) ;
    },

     /**
      * Compare two palette arrays. (index property is not checked.)
      * @param {Array} pal1
      * @param {Array} pal2
      * @return {boolean} 
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
    _isPaletteEqual : function(pal1, pal2)
    {
      var len1 = pal1.length,
          len2 = pal2.length,
          lab1, lab2, o1, o2, i, ret  ;

      ret = false ;
      if (len1 === len2)
      {
        for (i = 0; i < len1; i++)
        {
          o1 = pal1[i] ;
          o2 = pal2[i] ;
          if (this._compareColorValues(o1["color"], o2["color"]))
          {
            lab1 = o1["label"] ;
            lab2 = o2["label"] ;
            if (lab1 !== lab2)
            {
              break ;
            }
          }
        }

        ret = (i >= len1) ;
      }
      return ret ;
    },

     /**
      * Add "id" property to palette entries
      * @param {Array} palette
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
    _addIdsToPalette : function(palette)
    {
       var i, o, l = palette.length ;
       for (i = 0; i < l; i++)
       {
         o = palette[i] ;
         o["id"] = this._getNewSwatchId() ;
       }
    },


     /**
      * Returns a new (unique) swatch id.
      * @returns {string} swatch id
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
    _getNewSwatchId :function()
    {
      var s =  this._swatchId.toString() ;
      this._swatchId++ ;
      return s ;
    },

     /**
      * Clear resources
      * @return {void}
      * @memberof oj.ojColorPalette
      * @instance
      * @private
      */
     _clear : function()
     {
        this._converterFactory   =
        this._convHex            =
        this._markup             =
        this._$LV                = null ;
     },

    /**
     * Get the width of a vertical scrollbar.
     * @returns {number} Width of a vertical scrollbar
     * @memberof oj.ojColorPalette
     * @instance
     * @private
     */
    _getScrollbarWidth: function()
    {
      var div = $(
        "<div style='overflow: scroll; width: 100px; height: 100px; " +
            "position: absolute; visibility: hidden;'>" +
          "<div style='width: 100%; height: 100%;'></div>" +
        "</div>"
      );
      this.element.append(div);             //@HTMLUpdateOK (strings are all code constants)
      var outerWidth = div[0].offsetWidth;
      var innerWidth = div.children()[0].offsetWidth;
      div.remove();
      return (outerWidth - innerWidth);
    },

    /**
     * Returns a jquery object of the launcher element representing the content nodes.
     * @returns {Object}
     * @protected
     * @override
     * @instance
     * @memberof oj.ojColorPalette
     */
    _GetMessagingLauncherElement: function () {
      return this.element;
    },

    /**
     * Returns a jquery object of the elements representing the content nodes (spectrum thumb).
     * @returns {Object}
     * @protected
     * @override
     * @instance
     * @memberof oj.ojColorPalette
     */
    _GetContentElement: function () {
      return this._$LV;
    },

    /**
     * Returns the element's value. Normally, this is a call to this.element.val(), but for some
     * components, it could be something else. E.g., for ojRadioset the element's value is really the
     * value of the selected radio in the set.
     * @returns {oj.Color} color value
     * @override
     * @memberof oj.ojColorPalette
     * @instance
     * @protected
     */
    _GetElementValue : function ()
    {
      return this._value;
    },

    /**
     * Called when the display value on the element needs to be updated. This method updates the
     * (content) element value.
     *
     * @param {string} displayValue of the new string to be displayed
     * @return {void}
     *
     * @memberof oj.ojColorPalette
     * @instance
     * @protected
     * @override
     */
    _SetDisplayValue: function (displayValue) {
      // If displayValue is null/undefined, will substitute black.
      if (displayValue)
      {
        if (typeof displayValue === "string")
          this._value = new oj.Color(displayValue);
        else
          this._value = displayValue;
      }
      else
      {
        this._value = oj.Color.BLACK;
        oj.Logger.warn("JET Color Palette (id='" + this.element.attr("id") + "'): Substituting oj.Color.BLACK since display value is not defined.");
      }
    },

    /**
     * Returns the display value that is ready to be passed to the converter.
     *
     * @return {string} usually a string display value
     *
     * @memberof oj.ojColorPalette
     * @instance
     * @protected
     * @override
     */
    _GetDisplayValue: function () {
      return this._value.toString();
    },

    /**
     * Returns the default styleclass for the component. All input components must override.
     *
     * @return {string}
     *
     * @memberof oj.ojColorPalette
     * @instance
     * @protected
     * @override
     */
    _GetDefaultStyleClass : function ()
    {
      return "oj-colorpalette";
    }


    // Fragments:

    /**
     * Sets a property or a single subproperty for complex properties and notifies the component
     * of the change, triggering a [property]Changed event.
     * 
     * @function setProperty
     * @param {string} property - The property name to set. Supports dot notation for subproperty access.
     * @param {*} value - The new value to set the property to.
     * 
     * @expose
     * @memberof oj.ojColorPalette
     * @instance
     * 
     * @example <caption>Set a single subproperty of a complex property:</caption>
     * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
     */ 
    /**
     * Retrieves a value for a property or a single subproperty for complex properties.
     * @function getProperty
     * @param {string} property - The property name to set. Supports dot notation for subproperty access.
     * @return {*}
     * 
     * @expose
     * @memberof oj.ojColorPalette
     * @instance
     * 
     * @example <caption>Get a single subproperty of a complex property:</caption>
     * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
     */ 
    /**
     * Performs a batch set of properties.
     * @function setProperties
     * @param {Object} properties - An object containing the property and value pairs to set.
     * 
     * @expose
     * @memberof oj.ojColorPalette
     * @instance
     * 
     * @example <caption>Set a batch of properties:</caption>
     * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
     */ 
    
    /**
     * <p>Sub-ID for a palette swatch item at a specific index.</p>
     *
     * @ojsubid oj-palette-entry
     * @memberof oj.ojColorPalette
     *
     * @example <caption>Get the palette's internal JET ListView entry for the third palette swatch:</caption>
     * var node = myColorPalette.getNodeBySubId({'subId': 'oj-palette-entry', 'index': 2});
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Swatch</td>
     *       <td><kbd>Space or Enter</kbd></td>
     *       <td>Select a color swatch.</tr>
     *     </tr>
     *     <tr>
     *       <td>Swatch</td>
     *       <td><kbd>PgDn</kbd></td>
     *       <td>Navigates down the swatch display to the next page.</td>
     *     </tr>
     *     <tr>
     *       <td>Swatch</td>
     *       <td><kbd>PgUp</kbd></td>
     *       <td>Navigates up the swatch display to the previous page.</td>
     *     </tr>
     *     <tr>
     *       <td>Swatch</td>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Navigate to the next swatch if in list layout, or to the swatch in the same position in the next row if in grid layout.</td>
     *     </tr>
     *     <tr>
     *       <td>Swatch</td>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>Navigate to the previous swatch if in list layout, or to the swatch in the same position in the previous row if in grid layout.</td>
     *     </tr>
     *     <tr>
     *       <td>Swatch</td>
     *       <td><kbd>RightArrow</kbd></td>
     *       <td>Navigate to the next swatch.</td>
     *     </tr>
     *     <tr>
     *       <td>Swatch</td>
     *       <td><kbd>LeftArrow</kbd></td>
     *       <td>Navigate to the previous swatch.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojColorPalette
     */


  }) ;    // end    $.widget("oj.ojColorPalette", ...


})();

(function() {
var ojColorPaletteMeta = {
  "properties": {
    "labelDisplay": {
      "type": "string",
      "enumValues": ["auto", "off"]
    },
    "labelledBy": {
      "type": "string"
    },
    "layout": {
      "type": "string",
      "enumValues": ["grid", "list"]
    },
    "palette": {
      "type": "Array"
    },
    "swatchSize": {
      "type": "string",
      "enumValues": ["lg", "sm", "xs"]
    },
    "translations": {
      "type": "Object",
      "properties": {
        "labelNone": {
          "type": "string",
          "value": "None"
        }
      }
    },
    "value": {
      "writeback": true
    }
  },
  "extension": {
    _WIDGET_NAME: "ojColorPalette"
  }
};
oj.CustomElementBridge.registerMetadata('oj-color-palette', 'editableValue', ojColorPaletteMeta);
oj.CustomElementBridge.register('oj-color-palette', {'metadata': oj.CustomElementBridge.getMetadata('oj-color-palette')});

})();

/*----------------------------------------------------------------
   ojColorSpectrum    JET Color (spectrum) element
   Depends:   jquery.ui.core.js
              jquery.ui.widget.js
------------------------------------------------------------------*/

(function ()
{
  /*
  function debugObj(o)  {
    var s ;
    try { s = JSON.stringify(o) ; }
    catch (e) { s = "ERROR";}
    return s ;
  };
  */

  //  ojColorSpectrum class names
  var  /** @const */  OJES_CONTAINER      = "oj-colorspectrum-container",
       /** @const */  OJES_SPECTRUM       = "oj-colorspectrum-spectrum",
       /** @const */  OJES_THUMB          = "oj-colorspectrum-thumb",
       /** @const */  OJES_HUE_SLIDER     = "oj-colorspectrum-hue",
       /** @const */  OJES_ALPHA_SLIDER   = "oj-colorspectrum-alpha",
       /** @const */  OJES_DISABLED       = "oj-disabled" ;


  // Misc translation keys
  var  /** @const */  TRANSKEY_HUE        = "labelHue",
       /** @const */  TRANSKEY_OPACITY    = "labelOpacity",
       /** @const */  TRANSKEY_SATLUM     = "labelSatLum",
       /** @const */  TRANSKEY_THUMB_DESC = "labelThumbDesc";

/**
 * Returns an array of Hue, Saturation, Luminance, and Alpha from an oj.Color
 * @private
 * @param   {oj.Color}  color    The color to be converted.
 * @return  {Array}   The HSLA representation [h, s, l, a]
 */
function  _getHslFromRgb(color)
{
   var hsl = _rgbToHsl(color.getRed(), color.getGreen(), color.getBlue()) ;

   hsl[0] *= 360 ;    // hsl was returned in [0,1] so transform into
   hsl[1] *= 100 ;    // [0,360] for hue, and [0,100] for
   hsl[2] *= 100 ;    // saturation/luminance
   hsl.push(color.getAlpha());

   return hsl ;
};

/**
 * Converts an RGB color value to HSL. Conversion formula
 * adapted from http://en.wikipedia.org/wiki/HSL_color_space.
 * Assumes r, g, and b are contained in the set [0, 255] and
 * returns h, s, and l in the set [0, 1].
 * @private
 * @param   {number}  r       The red color value
 * @param   {number}  g       The green color value
 * @param   {number}  b       The blue color value
 * @return  {Array}   The HSL representation [h, s, l]
 */
function _rgbToHsl(r, g, b)
{
   r /= 255, g /= 255, b /= 255;
   var max = Math.max(r, g, b), min = Math.min(r, g, b);
   var h, s, l = (max + min) / 2;

   if (max == min)
   {
      h = s = 0; // achromatic
   }
   else
   {
     var d = max - min;
     s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
     switch(max){
         case r: h = (g - b) / d + (g < b ? 6 : 0); break;
         case g: h = (b - r) / d + 2; break;
         case b: h = (r - g) / d + 4; break;
     }
     h /= 6;
   }

   return [h, s, l];
};


 /**
  * @ojcomponent oj.ojColorSpectrum
  * @augments oj.editableValue
  * @ojsignature [{
  *                target: "Type",
  *                value: "class ojColorSpectrum extends editableValue<oj.Color, ojColorSpectrumSettableProperties>"
  *               },
  *               {
  *                target: "Type",
  *                value: "ojColorSpectrumSettableProperties extends editableValueSettableProperties<oj.Color>",
  *                for: "SettableProperties"
  *               }
  *              ]
  * @since 3.0.0
  * @ojstatus preview
  * @class oj.ojColorSpectrum
  * @ojshortdesc Enables a custom color value to be specified from a display containing a saturation/luminosity spectrum, plus hue and opacity sliders.
  * @classdesc

  * <h3 id="colorSpectrumOverview-section">
  *   JET Color Spectrum
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#colorSpectrumOverview-section"></a>
  * </h3>
  *
  * The JET Color Spectrum element allows a custom color value to be retrieved from a display
  * containing a saturation/luminosity spectrum, and hue and opacity sliders.
  * <pre class="prettyprint">
  * <code>&lt;oj-color-spectrum value="{{colorValue}}">
  * &lt;/oj-color-spectrum>
  * </code></pre>
  * </p></br>
  *
  * <h3 id="touch-section">
  *   Touch End User Information
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
  * </h3>
  *
  * {@ojinclude "name":"touchDoc"}
  * </p></br>
  *
  * <h3 id="keyboard-section">
  *   Keyboard End User Information
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
  * </h3>
  *
  * {@ojinclude "name":"keyboardDoc"}
  * </br>
  *
  *
  */
  oj.__registerWidget("oj.ojColorSpectrum", $['oj']['editableValue'],
  {
     widgetEventPrefix : "oj",
     defaultElement    : "<input>",

     options: {

         /**
          * Labelled-by is used to establish a relationship between this and another element.
          * A common use is to tie the oj-label and the oj-color-spectrum together.
          * The oj-label custom element has an id, and you use the labelled-by attribute
          * to tie the two elements together.
          *
          * @ojshortdesc Used to establish a relationship between this element and another element.
          *
          * @example <caption>Initialize the color spectrum with the <code class="prettyprint">labelled-by</code> attribute specified:</caption>
          * &ltoj-label id="labelId">Name:&lt/oj-label>
          * &ltoj-color-spectrum labelled-by="labelId">&lt;/oj-color-spectrum>
          *
          * @example <caption>Get or set the <code class="prettyprint">labelledBy</code> property, after initialization:</caption>
          * // getter
          * var labelledBy = myColorSpectrum.labelledBy;
          *
          * // setter
          * myColorSpectrum.labelledBy = "labelId";
          *
          * @expose
          * @type {?string}
          * @default null
          * @instance
          * @memberof oj.ojColorSpectrum
          */
         labelledBy: null,

         /**
          * The value of the element representing the current color.
          * @member
          * @type {oj.Color}
          * @default null
          * @ojwriteback
          * @expose
          * @instance
          * @memberof oj.ojColorSpectrum
          * @ojshortdesc Specifies the value of the element representing the current color.
          * @example <caption>Initialize the color spectrum with the <code class="prettyprint">value</code> attribute specified:</caption>
          * &ltoj-color-spectrum value='{{myColor}}'>&lt;/oj-color-spectrum>
          *
          * @example <caption>Get or set the <code class="prettyprint">value</code> property, after initialization:</caption>
          * // getter
          * var color = myColorSpectrum.value;
          *
          * // setter
          * myColorSpectrum.value = new oj.Color('rgb(0,0,0)');
          * </code>
          */
         value:         null,

         /**
          * The <code class="prettyprint">transient-value</code> is the read-only attribute for retrieving the
          * transient value of the color spectrum.<p>
          * The <code class="prettyprint">transient-value</code> updates to display the transient
          * changes of the spectrum thumb, and the hue and opacity sliders. The difference in behavior
          * between the <code class="prettyprint">transient-value</code> and <code class="prettyprint">value</code>
          * is that <code class="prettyprint">transient-value</code> will be updated as the spectrum thumb or
          * hue/opacity sliders are moved, whereas the <code class="prettyprint">value</code> attribute is
          * updated only after the respective thumb is released (or after a key press).
          * </p>
          * @ojshortdesc Retrieves the transient value of the component.
          * @alias transientValue
          * @member
          * @type {oj.Color}
          * @default null
          * @expose
          * @instance
          * @memberof oj.ojColorSpectrum
          * @readonly
          * @ojwriteback
          * @since 4.2.0
          * @ojstatus preview
          */
         rawValue: null,

     },   // end options

     //** @inheritdoc */
     getNodeBySubId : function(locator)
     {
       if (locator == null) {
         return this.element ? this.element[0] : null;
       }

       var subId = locator['subId'],
           ret   = this._super(locator);

       if (!ret)
       {
         switch (subId)
         {
           case 'oj-spectrum' :
                                      ret = this._$spectrum[0];
                                      break ;
           case 'oj-spectrum-thumb' :
                                      ret = this._$spectrumThumb[0];
                                      break ;
           case 'oj-hue-slider-bar' :
                                      ret = this._$hueSlider.ojSlider("getNodeBySubId", {'subId': 'oj-slider-bar'}) ;
                                      break ;
           case 'oj-hue-slider-bar-value' :
                                      ret = this._$hueSlider.ojSlider("getNodeBySubId", {'subId': 'oj-slider-bar-value'}) ;
                                      break ;
           case 'oj-hue-slider-thumb' :
                                      ret = this._$hueSlider.ojSlider("getNodeBySubId", {'subId': 'oj-slider-thumb-0'}) ;
                                      break ;
           case 'oj-opacity-slider-bar' :
                                      ret = this._$alphaSlider.ojSlider("getNodeBySubId", {'subId': 'oj-slider-bar'}) ;
                                      break ;
           case 'oj-opacity-slider-bar-value' :
                                      ret = this._$alphaSlider.ojSlider("getNodeBySubId", {'subId': 'oj-slider-bar-value'}) ;
                                      break ;
           case 'oj-opacity-slider-thumb' :
                                      ret = this._$alphaSlider.ojSlider("getNodeBySubId", {'subId': 'oj-slider-thumb-0'}) ;
                                      break ;
         }
       }

       return ret ;
     },

     //** @inheritdoc */
     getSubIdByNode : function(elem)
     {
        if (! elem) {
          return null ;
        }

        var $node = $(elem),
            subId = null,
            $ancestor,
            sliderType;

        if ($node.hasClass(OJES_SPECTRUM)) {
          subId = 'oj-spectrum' ;
        }
        else if ($node.hasClass(OJES_THUMB)) {
          subId = 'oj-spectrum-thumb' ;
        }
        else
        {
          $ancestor  = $node.closest('.oj-slider') ;
          sliderType = $ancestor.hasClass('oj-slider-vertical')? "hue" : "opacity";

          if ($node.hasClass('oj-slider-bar')) {
            // This is no longer correct because of fix for  - unused css class in ojslider #slider-id-barback node
            //sliderType = $node.hasClass('oj-slider-vertical')? "hue" : "opacity" ;
            subId = "oj-" + sliderType+ "-slider-bar" ;
          }
          else if ($node.hasClass('oj-slider-bar-value')) {
            subId = "oj-" + sliderType+ "-slider-bar-value" ;
          }
          else if ($node.hasClass('oj-slider-thumb')) {
            subId = "oj-" + sliderType+ "-slider-thumb" ;
          }
        }

        if (subId != null)
          return {'subId': subId};
        else
          return this._super(elem);
     },

     /**
      * Destroy the Color Spectrum component
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @override
      * @protected
      */
     _destroy: function()
     {
        if (this._$boundElem)
        {
           this._clearListeners() ;
           this._destroySliders() ;
           this._removeMarkup() ;          // remove our markup from dom

           this._$boundElem.removeClass("oj-colorspectrum");

           this._clear() ;
        }

        this._super();
     },

     /**
      * Called the first time the widget is called on an element.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _ComponentCreate: function ()
     {
        this._super() ;
        this._initEditor() ;
     },

    /**
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
    _AfterCreate: function()
    {
      var label;

      this._super();

      this._updateLabelledBy(null, this.options['labelledBy'],this._$spectrumThumb);

      // custom element's use oj-label.
      if (!this._IsCustomElement())
        label = this._GetLabelElement();

      // Apply the label to the thumb
      if (label)
      {
        // Set the aria-labelledby attribute of the thumb to the returned id
        var labelId = label.attr('id');
        if (!labelId)
        {
          oj.Logger.warn("JET Color Spectrum: The label for this component needs an id in order " +
            "to be accessible");
        }
        else
          this._$spectrumThumb.attr('aria-labelledby', labelId);
      }
      else
      {
        // Check if the element has aria-label
        var ariaLabelString = this.element.attr('aria-label');
        if (ariaLabelString)
        {
          // Set the aria-label of the thumb to the returned string
          this._$spectrumThumb.attr('aria-label', ariaLabelString);
        }
      }
    },

     /**
      * Handle an option change.
      * Called by $(selector).ojColorSpectrum("option", "prop", value)
      * @param {string}   key
      * @param {string | oj.Color | boolean}   newval
      * @param {Object}   flags
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setOption: function (key, newval, flags)
     {
        var change;
        var originalValue = this.options["labelledBy"];

        switch (key)
        {
           case "value"  :     change = this._setOptValue(newval) ;
                               break ;
           case "disabled"  :  change = this._setOptDisabled(newval, true) ;
                               break ;
           case "labelledBy":  // remove the old one and add the new one
                               this._updateLabelledBy(originalValue, newval,this._$spectrumThumb);
                               break;
        }

        this._super(key, newval, flags);
     },


     /**
      * Catch the string rawValue/transientValue sent by editableValue, convert to oj.Color and
      * change the rawValue/transientValue option.
      * @param {oj.Color} val
      * @param {Event} event
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @override
      * @private
      */
     _SetRawValue : function (val, event)
     {
       if (typeof val === "string") {
        try {
           val = new oj.Color(val) ;
        }
        catch (e) {
          oj.Logger("ojColorSpectrum (id='" + this.element.attr("id") +
                    "'): invalid "+this._transientValueName+" (" + val + "), defaulting to black") ;
          val = oj.Color.BLACK ;
        }
       }

       var flags = {};
       flags['_context'] = {originalEvent: event, writeback: true, internalSet: true, readOnly: true};

       if (! this._comparedRoundedColor(this.options[this._transientValueName], val))
       {
         this.option(this._transientValueName, val, flags);
       }
     },

   /**
    * If custom element, get the labelledBy option, and set this
    * onto the root dom element as aria-labelledby. We append "|label" so it matches the id that
    * is on the oj-label's label element.
    * @memberof oj.ojColorSpectrum
    * @instance
    * @private
    */
    _updateLabelledBy : oj.EditableValueUtils._updateLabelledBy,

     /**
       * @memberof oj.ojColorSpectrum
       * @param {oj.Color} val1
       * @param {oj.Color} val2
       * @returns {boolean} true if rounded colors match, else false.
       * @private
       * @instance
       */
     _comparedRoundedColor : function(val1, val2)
     {
       var ret = ((val1.getRed()   === val2.getRed())   &&
                  (val1.getGreen() === val2.getGreen()) &&
                  (val1.getBlue()  === val2.getBlue())  &&
                  (val1.getAlpha() === val2.getAlpha())) ;
       return ret;
     },

     /**
      * Compares two color values (oj.Colors)
      * @param {oj.Color}  color1   a color object to compare.
      * @param {oj.Color}  color2   a color object to compare.
      * @returns {boolean}  true    if colors match, else false.
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _compareColorValues  :  function(color1, color2)
     {
        var t1  = (color1 instanceof oj.Color),
            t2  = (color2 instanceof oj.Color);

        return ((t1 && t2) && color1.isEqual(color2)) ;
     },

     /**
      * Handle "disabled" option change.
      * @param {boolean} disabled True if the component is disabled, false if enabled.
      * @param {boolean} applyOnlyIfDifferent Only apply the new value if it's different from the
      *        current value.
      * @returns {boolean}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setOptDisabled: function(disabled, applyOnlyIfDifferent)
     {
        var change  = !applyOnlyIfDifferent ||
                      (applyOnlyIfDifferent && (disabled !== this._disabled));
        if (change)
        {
           this._enableSliders(! disabled) ;

           if (disabled)
           {
             this.element.addClass(OJES_DISABLED) ;
             this._makeThumbDraggable(false) ;
           }
           else
           {
             this.element.removeClass(OJES_DISABLED) ;
             this._makeThumbDraggable(true) ;
           }

           this._disabled = disabled;
        }
        return change ;
     },

     /**
      * Handle "value" option change on the component.
      * @param {oj.Color}  color   the new color to apply.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setOptValue :  function(color)
     {
        if (color instanceof oj.Color)
        {
           // Check if value has changed
           if (! this._compareColorValues(this._value, color))
           {
              // Color is different from current
              this._setColorVals(color) ;   // note the new color
              this._setSliderValue(this._hueVal,   true) ;
              this._setSliderValue(this._alphaVal, false) ;
              this._setSpectrumHue(this._hueVal, this._satVal, this._lumVal, this._alphaVal, true) ;
           }
        }
     },

     /**
      * Handle slider option change.
      * @param {Event} e  the associated event.
      * @param {Object} ui
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _onSliderOptionChange: function(e, ui)
     {
       var  grad, from, to, newColor ;
       var  hue, sat, lum, alpha ;
       var  isHueSlider, isAlphaSlider ;
       var  $targ ;

       if (! this._sliderSetup) {
         return ;
       }
       if (((ui.option != "rawValue") && (ui.option != "value")) || (ui.value === undefined))  {
         return ;
       }

       // Slider  workround - originalEvent is missing on rawValue,
       // but is set on previous "value"
       if (ui.option == "value") {
         this._isOrigEvent = !!e.originalEvent ;    // set for following rawValue event
       }

       $targ = $(e.target) ;
       isHueSlider   = $targ.hasClass("oj-colorspectrum-hue") ;
       isAlphaSlider = $targ.hasClass("oj-colorspectrum-alpha") ;
       if ((! isHueSlider) && (! isAlphaSlider)) { return;}

       // Get existing values
       hue   = this._hueVal ;
       sat   = this._satVal ;
       lum   = this._lumVal ;
       alpha = this._alphaVal;

       if (isHueSlider)
       {
         hue = ui.value ;

         //  Update alpha slider gradient
         this._updateAlphaBG(hue, sat, lum)

         // Update the spectrum
         this._setSpectrumHue(hue, sat, lum, alpha, false) ;
       }
       else
       {
         // Alpha slider
         alpha = ui.value ;
         this._setAriaText(hue, sat, lum, alpha);
       }

       // Fire value optionChange event
       newColor = new oj.Color({"h":hue, "s":sat, "l":lum, "a":alpha}) ;

       // If optionChange is fired because of direct user action on the slider,
       // then need to tell the Spectrum component which will also fire an
       // optionChange for the app listener.
       if (e.originalEvent || this._isOrigEvent)  {
         this._isOrigEvent = false ;                 // TDO - remove when slider  fixed
         if (ui.option === "rawValue")
           this._SetRawValue(newColor, e);
         else if (ui.option === "value")
           this._SetValue(newColor, e);
       }
       this._setColorVals(newColor) ;
     },


     /**
      * Set a slider value
      * @param {number} value  the value to be applied to a slider
      * @param {boolean} isHue true if the hue slider should be updated,
      *                        or false for the alpha aslider.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
      _setSliderValue : function(value, isHue)
      {
         var $targ = isHue? this._$hueSlider : this._$alphaSlider;
         $targ.ojSlider("option", "value", value) ;
      },

     /**
      * Apply the hue to the spectrum, and optionally moves the spectrum thumb.
      * @param {number}  hue   hue value in [0,360].
      * @param {number}  sat   saturation value in [1, 100].
      * @param {number}  lum   luminosity value in [1,100].
      * @param {number}  alpha alpha value
      * @param {boolean=}  moveThumb  if true the spectrum thumb is also moved.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setSpectrumHue : function(hue, sat, lum, alpha, moveThumb)
     {
       var color = "hsl(" + hue + ", 100%, 50%)" ;

       this._$spectrum.css("background-color", color);

       if (moveThumb)
       {
         var pos = this._getSatLumSpectrumPosition(sat, lum) ;
         this._setThumbPosition(pos.x, pos.y) ;
       }

       this._setAriaText(hue, sat, lum, alpha);
     },

     /**
      * Apply a mask to the spectrum so that the current hue is displayed showing combined
      * saturation and luminosity gradients.
      * http://stackoverflow.com/questions/17224383/generate-a-saturation-brightness-mask-using-gradients
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setSpectrumMask : function()
     {
       var grad  = " -___-linear-gradient(top, hsl(0, 0%, 100%) 0%, hsla(0, 0%, 100%, 0) 50%, hsla(0, 0%, 0%, 0) 50%, hsl(0, 0%, 0%) 100%), -___-linear-gradient(left, hsl(0, 0%, 50%) 0%, hsla(0, 0%, 50%, 0) 100%)",
           ai = oj.AgentUtils.getAgentInfo(),
           rep;

       switch (ai["browser"])
       {
         case oj.AgentUtils.BROWSER.FIREFOX :  rep = "moz";
                                               break ;
         case oj.AgentUtils.BROWSER.CHROME :
         case oj.AgentUtils.BROWSER.SAFARI :
         default :                             rep = "webkit";
                                               break ;
         case oj.AgentUtils.BROWSER.IE :       rep = "ms";
                                               break ;
       }

       grad   = " -___-linear-gradient(top, hsl(0, 0%, 100%) 0%, hsla(0, 0%, 100%, 0) 50%, hsla(0, 0%, 0%, 0) 50%, hsl(0, 0%, 0%) 100%), -___-linear-gradient(left, hsl(0, 0%, 50%) 0%, hsla(0, 0%, 50%, 0) 100%)";
       grad = grad.replace(/___/g, rep) ;
       this._$spectrum.css("backgroundImage", grad);
     },

     /**
      * Updates the thumb's current (x,y) position relative to the spectrum top left. Note this does not
      * move the thumb - refer to _moveThumb().
      * @param {number} x  the x displacement into the spectrum.  If NaN the value is ignored.
      * @param {number} y  the y displacement into the spectrum.  If NaN the value is ignored.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setThumbPosition : function(x, y)
     {
        var xPos, yPos ;

        if (! isNaN(x))
        {
          this._xThumb = x ;
        }
        if (! isNaN(y))
        {
          this._yThumb = y ;
        }
        this._moveThumb(0, 0) ;
     },

     /**
      * Returns the thumb center's displacement into the spectrum for the
      * given saturation and luminosity values.
      * @param {number}  sat   saturation value in [1, 100].
      * @param {number}  lum   luminosity value in [1,100].
      * @returns {Object}  Object containing {x: , y:} where x and y are displacements
      *                    into the spectrum div.
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _getSatLumSpectrumPosition : function(sat, lum)
     {
        var satPos = Math.min(((sat/100) * this._spectrumWidth),  this._spectrumWidth) ;
        var lumPos = this._spectrumHeight - Math.min(((lum/100) * this._spectrumHeight), this._spectrumHeight) ;

        return {x: Math.round(satPos), y: Math.round(lumPos)} ;
     },

     //
     //  *  Contrast is currently created by using a drop shadow on the thumb.
     //  *  Set the thumb to a contrasting color so that it does not become hard to
     //  *  see when over very light parts of the spectrum.
     //  *  @param {number} hue  the hue value.
     //  *  @param {number} sat  the saturation value in [1, 100]
     //  *  @param {number} lum  the luminosity value in [1, 100]
     //  *  @private
     /*
     _setContrastingThumb : function(hue, sat, lum)
     {
        var c ;

        if (((sat < 30) && (lum > 70)) || (lum > 70))
        {
          c = "black" ;
        }
        else
        {
          c = "white" ;
        }

        this._$spectrumThumb.css('border-color', c) ;
     },
       */

    /**
     * Move the spectrum thumb relative to its current position (which is
     * set by _setThumbPosition()) by the supplied deltas. Updates
     * (this._xThumb, this._yThumb). The  call is ignored if the move
     * will take the thumb center outside the spectrum.
     * @param {number} xDelta     Pixels to move in the x-axis
     * @param {number} yDelta     Pixels to move in the y-axis
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _moveThumb : function(xDelta, yDelta)
     {
        var x, y ;

        // Ensure that the thumb center hasn't gone outside the spectrum
        x = this._xThumb + xDelta ;
        y = this._yThumb + yDelta ;
        if (x < 0) x = 0 ;
        if (y < 0) y = 0 ;
        if (x >= this._spectrumWidth)  x = this._spectrumWidth -1;
        if (y >= this._spectrumHeight) y = this._spectrumHeight -1;

        this._xThumb = x ;
        this._yThumb = y ;

        x = (x - this._spectrumThumbRadius) + "px";
        y = (y - this._spectrumThumbRadius) + "px";
        this._$spectrumThumb[0].style.left= x;
        this._$spectrumThumb[0].style.top = y;
     },

     /**
      * Handle click on the spectrum
      * @param {Event}  e the associated event.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _spectrumClick : function(e)
     {
        if (this._disabled)
        {
           return ;
        }

        var pos   = this._$spectrum.offset(),    // get click position relative
            xDisp = e.pageX - pos.left,          // relative to spectrum
            yDisp = e.pageY - pos.top,           // top left
            hue, sat, lum,
            o ;

        xDisp = Math.round(xDisp) ;
        yDisp = Math.round(yDisp) ;
        o     = this._getSatLumFromPosition(xDisp, yDisp) ;
        sat = o.s;
        lum = o.l;
        hue = this._hueVal ;

        // Update alpha slider
        this._updateAlphaBG(hue, sat, lum)

        //  Fire value optionChange event.
        var oNewColor = new oj.Color({"h":hue, "s":sat, "l":lum, "a":this._alphaVal}) ;
        this._SetRawValue(oNewColor, e);
        this._SetValue(oNewColor, e);

        this._setAriaText(hue, sat, lum, this._alphaVal);
        this._value = oNewColor ;
        this._setSatLum(sat, lum) ;

        this._setThumbPosition(xDisp, yDisp) ;
        //this._setContrastingThumb(hue, sat, lum) ;
        this._$spectrumThumb.focus();
     },

     /**
      * Handle keydown on the spectrum
      * @param {Event}  e the associated event.
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _keyDown : function(e)
     {
        var key, xDelta, yDelta, elapsed, accel = 1 ;

        this._keyNow   = new Date().getTime();
        if (this._keyStart < 0)
        {                                   // start new typeamatic period
           this._keyStart = this._keyNow ;
           this._keyCount = 0 ;
        }

        elapsed = this._keyNow - this._keyStart ;
        this._keyCount++ ;
        if ((elapsed > 1400) || (this._keyCount > 25))
        {
          accel = 3 ;                 // now move in increments of 3 pixels
        }

        xDelta = yDelta = 0 ;
        key    = e.keyCode ;

        if (key == 40) {                // down
          yDelta = accel ;
        }
        else if (key == 38) {           // up
          yDelta = -accel ;
        }
        else if (key == 39) {           // right
          xDelta = accel ;
        }
        else if (key == 37) {           // left
          xDelta = -accel ;
        }
        else if (key == 36) {           // home
          this._setThumbPosition(0,0) ;
          this._keyStart = -1 ;
        }
        else if (key == 35) {           // end
          this._setThumbPosition(this._spectrumWidth-1, this._spectrumHeight-1) ;
          this._keyStart = -1 ;
        }
        else if (key == 33) {           // PgUp
          this._setThumbPosition(NaN, 0) ;
          this._keyStart = -1 ;
        }
        else if (key == 34) {           // PgDn
          this._setThumbPosition(NaN, this._spectrumHeight-1) ;
          this._keyStart = -1 ;
        }
        else {
          return ;
        }

        this._moveThumb(xDelta, yDelta) ;
        this._handleThumbMoved(e, this._xThumb, this._yThumb) ;
        e.preventDefault();
        return false ;
     },

     /**
      * Handle keyup on the spectrum
      * @param {Event} e  the associated event.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _keyUp : function(e)
     {
       this._keyStart = -1 ;     // end of elapsed time/keystroke counting period

       // aria-valuetext was not set during the key down's to avoid repetitive
       // JAWS talkback - will do it now.
       var satLum = this._getSatLumFromPosition(this._xThumb, this._yThumb)
       this._setAriaText(this._hueVal, satLum.s, satLum.l, this._alphaVal);

       //  Fire the completing "value" event
       var newVal = this.options[this._transientValueName];
       this._SetValue(newVal, e);
     },


     /**
      * Set up the spectrum thumb to be draggable using jQuery Draggable
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _initThumbDraggable : function()
     {
       var dims = this._$spectrum[0].getBoundingClientRect() ;

       if (! this._thumbDragHandler)
       {
         this._thumbDragHandler = this._thumbDrag.bind(this) ;
       }
       this._$spectrumThumb.draggable({
          drag:  this._thumbDragHandler,
          stop:  this._thumbDragHandler
          // Note: cannot use "containment" option currently,
          //       it fails for touch drag on Edge browser.
        });
     },

     /**
      * Enable or disable dragging the spectrum thumb.
      * @param {boolean} enable True to enable dragging, false to disable it.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _makeThumbDraggable : function(enable)
     {
       this._$spectrumThumb.draggable(enable ? 'enable' : 'disable');
     },

     /**
      * Handle drag and end dragstop of spectrum thumb
      * @param {Event} e  the associated event.
      * @param {Object} ui
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _thumbDrag : function(e, ui)
     {
        var  cx, cy, off1, off2 ;

        // Constrain the thumb center to inside the spectrum
        // Left and right sides
        if (ui.position.left < (-this._spectrumThumbRadius))
        {
          ui.position.left = - this._spectrumThumbRadius ;
          cx = 0 ;
        }
        else  if ((ui.position.left + this._spectrumThumbRadius) >= this._spectrumWidth)
        {
          ui.position.left = this._spectrumWidth -1 - this._spectrumThumbRadius;
          cy = this._spectrumWidth -1 ;
        }

        // Top and bottom sides
        if (ui.position.top < (-this._spectrumThumbRadius))
        {
          ui.position.top = - this._spectrumThumbRadius ;
          cy = 0 ;
        }
        else if ((ui.position.top + this._spectrumThumbRadius) >= this._spectrumHeight)
        {
          ui.position.top = this._spectrumHeight -1 - this._spectrumThumbRadius;
          cy = this._spectrumHeight -1 ;
        }

        // Find center of dragged thumb relative to spectrum
        off1 = this._$spectrumThumb.offset() ;
        off2 = this._$spectrum.offset() ;
        if (cx != 0) {
          cx   = off1.left + this._spectrumThumbRadius - off2.left ;
        }
        if (cy != 0) {
          cy   = off1.top  + this._spectrumThumbRadius - off2.top ;
        }

        if (this._xThumb === cx && this._yThumb === cy)
        {
          return ;
        }

        if (e.type === "drag")
        {
          this._xThumb = cx ;
          this._yThumb = cy ;
        }
        else
        {
          cx = this._xThumb ;        // use last drag position
          cy = this._yThumb ;
        }

        this._handleThumbMoved(e, cx, cy) ; // update UI for the new position
     },

     /**
      * Handle previous spectrum thumb reposition by updating the alpha
      * slider background, and changing rawValue and value option values.
      * @param {Event}  e        the associated event.
      * @param {number} xCenter  the x value relative to the left side of the spectrum
      * @param {number} yCenter  the y value relative to the top of the spectrum
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _handleThumbMoved : function(e, xCenter, yCenter)
     {
        var hue, sat, lum, o, color, aria = false ;

        hue = this._hueVal ;
        o   = this._getSatLumFromPosition(xCenter, yCenter) ;
        sat = o.s;
        lum = o.l;

        // Update alpha slider background
        this._updateAlphaBG(hue, sat, lum)

        // Fire value optionChange event.
        color = new oj.Color({"h":hue, "s":sat, "l":lum, "a":this._alphaVal});
        this._SetRawValue(color, e);

        if (e.type === "dragstop")
        {
          this._SetValue(color, e.originalEvent);
          aria = true ;
        }

        this._value = color ;

        this._setSatLum(sat, lum) ;
        //this._setContrastingThumb(hue, sat, lum) ;    //  ensure thumb is easily viewable

        if (aria) {
          this._setAriaText(hue, sat, lum, this._alphaVal) ;
        }
     },


     /**
      * Return the sat/lum from the spectrum thumb reposition
      * @param {number} xCenter  the x value relative to the left side of the spectrum
      * @param {number} yCenter  the y value relative to the top of the spectrum
      * @returns {Object}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _getSatLumFromPosition : function(xCenter, yCenter)
     {
        var hue, sat, lum ;

        hue = this._hueVal ;
        sat = (xCenter/(this._spectrumWidth-1)) * 100 ;
        lum = 100 - ((yCenter / (this._spectrumHeight-1)) * 100) ;

        return {s: sat, l:lum} ;
     },

     /**
      * Update alpha slider background gradient
      * @param {number} hue  the hue value
      * @param {number} sat  the saturation value
      * @param {number} lum  the luminosity value
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _updateAlphaBG : function(hue, sat, lum)
     {
        //  Update alpha slider gradient
        var grad ;
        var from = "hsla(" + hue + "," + sat + "%," + lum + "%, 0)" ;
        var to   = "hsla(" + hue + "," + sat + "%," + lum + "%, 1.0)" ;

        // Reverse the gradient if RTL
        if (this._isRtl)
        {
          grad = to ;  to = from;  from = grad ;
        }
        grad = "linear-gradient(90deg, " +  from + ", " + to + ")" ;

        grad = grad + "," + this._alphaBgUrl;

        this._disabledAlphaBG = grad;
        this._$alphaBarBack.css("background", grad) ;
     },

     /**
      * Returns an hsl string with the saturation and luminosity rounded to 2 dec places.
      * @param {number} hue  the hue value
      * @param {number} sat  the saturation value
      * @param {number} lum  the luminosity value
      * @param {number} alpha  the alpha value
      * @returns {string} hsl string
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _getRoundedHsl : function(hue, sat, lum, alpha)
     {
       var s         = "hsl",
           haveAlpha = ((typeof alpha === "number") && (alpha < 1));

       s += haveAlpha? "a(" : "(" ;
       hue = Math.round(hue) ;
       sat = Math.round(sat*100)/100 ;
       lum = Math.round(lum*100)/100 ;
       s +=  hue + ", " + sat + "%, " + lum + "%" + (haveAlpha? ("," + alpha) : "") + ")" ;

       return s ;
     },

    /**
     * Set the saturation and luminosity values.
     * @param {number} sat  saturation value in [0,100]
     * @param {number} lum  luminosity value in [0,100]
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _setSatLum: function(sat, lum)
     {
       this._satVal = sat;
       this._lumVal = lum;
     },

    /**
     * @param {oj.Color} color
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _setColorVals : function(color)
     {
        if (color)
        {
          this._value = color ;
          var hsl  = _getHslFromRgb(color) ;
          this._hueVal   = hsl[0];
          this._satVal   = hsl[1]
          this._lumVal   = hsl[2]
          this._alphaVal = hsl[3]
        }
      },

     /**
      * Set the spectrum and thumb aria-textvalue.
      * @param {number} hue  the hue value
      * @param {number} sat  the saturation value
      * @param {number} lum  the luminosity value
      * @param {number} alpha the alpha value
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setAriaText : function(hue, sat,lum, alpha)
     {
       var ariaText = this._getRoundedHsl(hue, sat, lum, alpha) ;
       this._$spectrumThumb.attr('aria-valuetext', ariaText) ;
     },

     /**
      * Initialize the widget, examine options and set-up internal data structures.
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _initEditor : function()
     {
        this._initData() ;
        this._setup() ;
     },

     /**
      * Perform setup, and init the sliders
      * @return {void}
      * @memberof oj.ojColorSpectrum
      * @instance
      * @private
      */
     _setup :  function()
     {
        // Add markup as a child of this component's DOM element
        this._$boundElem.append(this._markup) ;      //@HTMLUpdateOk (contained strings are sanitized)
        this._$boundElem.addClass("oj-colorspectrum");

//-------------------------------------------------------------------------------------
// Temp  - ojSlider needs an Id to work properly, but does not create one if not supplied.
// This is currently being fixed by Lory.  REMOVE when fixed in ojSlider
$(".oj-colorspectrum-hue").uniqueId();
$(".oj-colorspectrum-alpha").uniqueId();
//-------------------------------------------------------------------------------------

        // Get working data
        this._$editorContainer = this._$boundElem.find(".oj-colorspectrum-container");
        this._$hueSlider       = this._$boundElem.find(".oj-colorspectrum-hue");
        this._$alphaSlider     = this._$boundElem.find(".oj-colorspectrum-alpha");
        this._$spectrum        = this._$boundElem.find(".oj-colorspectrum-spectrum");
        this._$spectrumThumb   = this._$boundElem.find(".oj-colorspectrum-thumb");

        // Get the thumb radius.  It appears that if the page has other components
        // and the page geometry is changing, offsetWidth is sometimes zero.  Here
        // we check for this and try it in a different way.
        this._spectrumThumbRadius  = this._$spectrumThumb[0].offsetWidth/2 ;
        if (this._spectrumThumbRadius <= 0) {
          this._spectrumThumbRadius  = this._$spectrumThumb.width()/2 ;
        }

        this._isTouch          = oj.DomUtils.isTouchSupported() ;
        this._spectrumWidth    = this._$spectrum.width() ;
        this._spectrumHeight   = this._$spectrum.height() ;

        this._$spectrumThumb.attr("aria-describedby",
        this._$boundElem.find(".oj-colorspectrum-thumb-description").uniqueId().attr("id"));

        var fakeDiv = document.createElement("div");
        fakeDiv.className = "oj-colorspectrum-alpha-bg";
        this._$boundElem[0].appendChild(fakeDiv);//HTMLUpdateOk
        this._alphaBgUrl =  window.getComputedStyle(fakeDiv,null).getPropertyValue("background-image");
        this._$boundElem[0].removeChild(fakeDiv);

        // Prep and create hue and alpha sliders
        var disabled   = this._disabled? true : false ;
        var hsl        = _getHslFromRgb(this._value) ;

        this._hueVal   = hsl[0];
        this._satVal   = hsl[1]
        this._lumVal   = hsl[2]
        this._alphaVal = hsl[3]

        this._sliderSetup = false ;    // ignore slider optionchange events during setup

        this._$hueSlider.ojSlider({
                                    "max":  360,
                                    "min":   0,
                                    "step":  1,
                                    "value": this._hueVal,
                                    "orientation": 'vertical',
                                    "optionChange": this._onSliderOptionChange.bind(this),
                                    "rootAttributes": {"class": 'oj-slider-color-picker'}
                                  }).attr('data-oj-internal', ''); // for use in automation api / oj.Components.getComponentElementByNode
        this._$alphaSlider.ojSlider({
                                      "max":  1,
                                      "min":  0,
                                      "step":  0.01,
                                      "value": this._alphaVal,
                                      "orientation": 'horizontal',
                                      "optionChange": this._onSliderOptionChange.bind(this),
                                      "rootAttributes": {"class": 'oj-slider-color-picker'}
                                    }).attr('data-oj-internal', ''); // for use in automation api / oj.Components.getComponentElementByNode
        this._initSliders() ;

        //  Apply the saturation/luminosity mask
        this._setSpectrumMask() ;

        // Paint the spectrum and move the thumb to its initial position based
        // on the current color value.
        this._setSpectrumHue(this._hueVal, this._satVal, this._lumVal, this._alphaVal, true) ;

        // Add listeners for the spectrum and the spectrum thumb
        this._$spectrum.click(this._spectrumClick.bind(this)) ;
        this._initThumbDraggable();
        this._setOptDisabled(disabled);

        // Add keyboard listeners
        this._$spectrumThumb.keydown(this._keyDown.bind(this)) ;
        this._$spectrumThumb.keyup(this._keyUp.bind(this)) ;

        this._$spectrum.focus(this._nofocus.bind(this)) ;

        // Focus ring
        this._focusable({element: this._$spectrumThumb,
                         applyHighlight: true});

        if (oj.DomUtils.isTouchSupported())
        {
          this._setupTouch(this._$spectrumThumb);
        }
     },

    /**
     * Move focus to the spectrum thumb
     * @param {Event} e  the associated event.
     * @returns {boolean}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _nofocus: function(e)
     {
       this._$spectrumThumb.focus();
       return false ;
     },

    /**
     * Initialize the sliders
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _initSliders: function()
     {
       var subId = {'subId': 'oj-slider-bar'} ;         // for getNodeBySubId()

       this._$hueBarBack    = $(this._$hueSlider.ojSlider('getNodeBySubId', subId));
       this._$alphaBarBack  = $(this._$alphaSlider.ojSlider('getNodeBySubId', subId)) ;
       subId.subId          = 'oj-slider-bar-value';
       this._$hueBarValue   = $(this._$hueSlider.ojSlider('getNodeBySubId', subId)) ;
       this._$alphaBarValue = $(this._$alphaSlider.ojSlider('getNodeBySubId', subId)) ;

       var grad = "linear-gradient(0deg, #ff0000 0%, #ffff00 17%, #00ff00 33%, #00ffff 50%, #0000ff 67%, #ff00ff 83%, #ff0000 100%)";
       this._$hueBarBack.css("background", grad) ;
       this._disabledHueBG   = grad ;
       this._updateAlphaBG(this._hueVal, this._satVal, this._lumVal) ;

       this._sliderSetup = true ;
     },

    /**
     * Set up instance data
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _initData : function()
     {
        this._applyOptions() ;     // process the component options
        this._xThumb = 0;          // initial spectrum
        this._yThumb = 0 ;         // thumb position.

        // Get slider aria-label's
        var hueTitle   = this._EscapeXSS(this.getTranslatedString(TRANSKEY_HUE)),
            alphaTitle = this._EscapeXSS(this.getTranslatedString(TRANSKEY_OPACITY)),
            thumbDesc  = this._EscapeXSS(this.getTranslatedString(TRANSKEY_THUMB_DESC));

        //  Create the markup
        this._markup = (function () {
           return [
             "<div class='oj-colorspectrum-container'>",
                "<div class='oj-colorspectrum-spectrum' tabindex='-1' style='position:relative'>",
                   "<div class='oj-colorspectrum-thumb' role='slider' aria-describedby='' style='position:absolute;top:-9999px;left:-9999px;' tabIndex='0'></div>",
                "</div>",
                "<div class='oj-colorspectrum-thumb-description oj-helper-hidden-accessible'>" + thumbDesc + "</div>",
                "<input class='oj-colorspectrum-hue' aria-label='" + hueTitle + "'></input>",
                "<input class='oj-colorspectrum-alpha' aria-label='" + alphaTitle + "'></input>",
             "</div>"
           ].join("") ;
        })();

        this._keyStart  = -1 ;
        this._isRtl     = this._GetReadingDirection() === "rtl";
     },

    /**
     * Process the component options
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _applyOptions : function()
     {
        var  opts = this.options,
             opt ;
        
        this._doc         = this.element[0].ownerDocument ;
        this._body        = this._doc.body ;
        this._$boundElem  = $(this.element) ;
        this._disabled    = false ;
        this._transientValueName = this._IsCustomElement() ? "transientValue" : "rawValue";

        opt = opts["value"] ;
        if (! (opt instanceof oj.Color))
        {
          opt = null ;
        }
        this._value = (opt? opt : oj.Color.BLACK) ;
        opts[this._transientValueName] = this._value;
        opt = opts["disabled"] ;
        this._disabled = (typeof opt === "boolean")? opt : false ;
     },

    /**
     * Enable/disable the hue and alpha sliders
     * @param {boolean} enable true if sliders are to be enabled, false for disable
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
    _enableSliders : function(enable)
    {
       var disabled = ! enable ;

       //=======================================================================
       //  Because of ojSlider , which basically destroys the slider
       //  when it disables it, we need to save off the background style, and
       //  reapply it, on the next enable.
       //  --- REMOVE WHEN osSlider is fixed! ---
       if (enable)
       {
         this._$hueSlider.ojSlider("option",   "disabled", disabled) ;
         this._$alphaSlider.ojSlider("option", "disabled", disabled) ;

         // Need to restablish JQ pointers since they are now invalid!
         var subId = {'subId': 'oj-slider-bar'} ;
         this._$hueSlider     = this._$boundElem.find(".oj-colorspectrum-hue");
         this._$alphaSlider   = this._$boundElem.find(".oj-colorspectrum-alpha");
         this._$hueBarBack    = $(this._$hueSlider.ojSlider('getNodeBySubId', subId));
         this._$alphaBarBack  = $(this._$alphaSlider.ojSlider('getNodeBySubId', subId)) ;

         if (this._disabledAlphaBG)
           this._$alphaBarBack.css("background", this._disabledAlphaBG) ;
         if (this._disabledHueBG)
           this._$hueBarBack.css("background",   this._disabledHueBG) ;
       }
       else
       {
         var bg = this._$hueBarBack.css("background") ;
         if (bg && bg.length > 0)
          this._disabledHueBG = bg;
         bg = this._$alphaBarBack.css("background") ;
         if (bg && bg.length > 0)
          this._disabledAlphaBG = bg;

         // remove the inline  style background override
         // otherwise the background of the disabled ojSlider would be overridden
         this._$hueBarBack.css("background", '');
         this._$alphaBarBack.css("background", '');

         this._$hueSlider.ojSlider("option",   "disabled", disabled) ;
         this._$alphaSlider.ojSlider("option", "disabled", disabled) ;
       }
       //===================================================
    },


    /**
     * Destroy the hue and alpha sliders
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
    _destroySliders : function()
    {
       this._$hueSlider.ojSlider("destroy") ;
       this._$alphaSlider.ojSlider("destroy") ;
    },

    /**
     * Remove the added markup
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
    _removeMarkup : function()
    {
       this._$boundElem.empty() ;
    },

    /**
     * Unbind all mouse/keyboard event listeners
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
    _clearListeners : function()
    {
       this._$spectrum.off("click") ;
       this._$spectrumThumb.off("mousedown") ;
       this._$spectrumThumb.off("keydown") ;
       this._$spectrumThumb.off("keyup") ;

       if (this._touchProxy)
       {
         this._tearDownTouch(this._$spectrumThumb);
       }
    },

    /**
     * Clear resources
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
     _clear : function()
     {
        this._markup            =
        this._$boundElem        =
        this._$editorContainer  =
        this._$hue              =
        this._$alpha            =
        this._$editorContainer  =
        this._$hueSlider        =
        this._$alphaSlider      =
        this._$spectrum         =
        this._$spectrumThumb    =
        this._spectrumThumbSize =
        this._strings           =
        this._mouseMoveHandler  =
        this._mouseUpHandler    = null ;
     },

    /**
     * Setup touch support
     * @param {Element} elem
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
    _setupTouch: function (elem) {
      this._touchProxy = oj._TouchProxy.addTouchListeners(elem);
    },

    /**
     * End touch support
     * @param {Element} elem
     * @return {void}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @private
     */
    _tearDownTouch: function (elem) {
      oj._TouchProxy.removeTouchListeners(elem);
    },

    /**
     * Returns a jquery object of the launcher element representing the content nodes (spectrum).
     * @returns {Element}
     * @protected
     * @override
     * @instance
     * @memberof oj.ojColorSpectrum
     */
    _GetMessagingLauncherElement: function () {
      return this.element;
    },

    /**
     * Returns a jquery object of the elements representing the content nodes (spectrum thumb).
     * @returns {Element}
     * @protected
     * @override
     * @instance
     * @memberof oj.ojColorSpectrum
     */
    _GetContentElement: function () {
      return this._$spectrumThumb;
    },

    /**
     * Returns the element's value. Normally, this is a call to this.element.val(), but for some
     * components, it could be something else. E.g., for ojRadioset the element's value is really the
     * value of the selected radio in the set.
     * @returns {oj.Color} element's value
     * @override
     * @memberof oj.ojColorSpectrum
     * @instance
     * @protected
     */
    _GetElementValue : function ()
    {
      return this._value;
    },

    /**
     * Called when the display value on the element needs to be updated. This method updates the
     * (content) element value.
     *
     * @param {string} displayValue of the new string to be displayed
     * @return {void}
     *
     * @memberof oj.ojColorSpectrum
     * @instance
     * @protected
     * @override
     */
    _SetDisplayValue: function (displayValue) {
      if (typeof displayValue === "string")
        this._value = new oj.Color(displayValue);
      else
        this._value = displayValue;
    },

    /**
     * Returns the display value that is ready to be passed to the converter.
     *
     * @return {string} usually a string display value
     *
     * @memberof oj.ojColorSpectrum
     * @instance
     * @protected
     * @override
     */
    _GetDisplayValue: function () {
      return this._value.toString();
    },

    /**
     * Returns the default styleclass for the component. All input components must override.
     *
     * @return {string}
     *
     * @memberof oj.ojColorSpectrum
     * @instance
     * @protected
     * @override
     */
    _GetDefaultStyleClass : function ()
    {
      return "oj-colorspectrum";
    },

    /**
     * Helper function to escape Cross site script text
     * @private
     * @param {string} escapeMe
     * @return {jQuery|string}
     * @memberof oj.ojColorSpectrum
     * @instance
     * @ignore
     */
    _EscapeXSS : function (escapeMe)
    {
      return $("<span>" + escapeMe + "</span>").text();
    }



    /// ///////////////     KEYBOARD     //////////////////
     /**
      * <table class="keyboard-table">
      *   <thead>
      *     <tr>
      *       <th>Target</th>
      *       <th>Key</th>
      *       <th>Action</th>
      *     </tr>
      *   </thead>
      *   <tbody>
      *     <tr>
      *       <td rowspan="6">Spectrum Thumb</td>
      *       <td><kbd>LeftArrow</kbd></td>
      *       <td>Moves the thumb to the left (decreasing saturation).  Holding the key down will cause the thumb to accelerate its movement.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>RightArrow</kbd></td>
      *       <td>Moves the thumb to the right (increasing saturation).  Holding the key down will cause the thumb to accelerate its movement.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>PageUp</kbd></td>
      *       <td>Moves the thumb vertically to the top of the spectrum (luminosity 100%).</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>PageDown</kbd></td>
      *       <td>Moves the thumb vertically to the bottom of the spectrum (luminosity 0%).</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>Home</kbd></td>
      *       <td>Moves the thumb to the top left corner of the spectrum (saturation 0%, luminosity 100%).</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>End</kbd></td>
      *       <td>Moves the thumb to the bottom right of the spectrum (saturation 100%, luminosity 0%).</td>
      *     </tr>
      *     <tr>
      *       <td rowspan="8">Slider Thumb</td>
      *       <td><kbd>LeftArrow</kbd></td>
      *       <td>Moves the thumb to the left for the horizontal <code class="prettyprint">Opacity</code> slider, and downwards for the vertical <code class="prettyprint">Hue</code> slider.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>RightArrow</kbd></td>
      *       <td>Moves the thumb to the right for the horizontal <code class="prettyprint">Opacity</code> slider, and upwards for the vertical <code class="prettyprint">Hue</code> slider.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>UpArrow</kbd></td>
      *       <td>Moves the thumb upwards for the vertical <code class="prettyprint">Hue</code> slider, and to the right for the horizontal <code class="prettyprint">Opacity</code> slider.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>DownArrow</kbd></td>
      *       <td>Moves the thumb downwards for the vertical <code class="prettyprint">Hue</code> slider, and to the left for the horizontal <code class="prettyprint">Opacity</code> slider.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>Home</kbd></td>
      *       <td>Moves the thumb to the leftmost (complete opacity) for the horizontal <code class="prettyprint">Opacity</code> slider,
      *           and to the bottom for the vertical <code class="prettyprint">Hue</code> slider.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>End</kbd></td>
      *       <td>Moves the thumb to the rightmost (complete opacity) for the horizontal <code class="prettyprint">Opacity</code> slider, and to the top for the vertical <code class="prettyprint">Hue</code> slider.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>PageUp</kbd></td>
      *       <td>Moves the thumb 20% of the slider range - upwardsfor the vertical <code class="prettyprint">Hue</code> slider, and to the right (lower opacity) for the horizontal <code class="prettyprint">Opacity</code> slider.</td>
      *     </tr>
      *     <tr>
      *       <td><kbd>PageDown</kbd></td>
      *       <td>Moves the thumb 20% of the slider range - downwards for the vertical <code class="prettyprint">Hue</code> slider, and to the left (higher opacity) for the horizontal <code class="prettyprint">Opacity</code> slider.</td>
      *     </tr>
      *   </tbody>
      * </table>
      * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
      * @memberof oj.ojColorSpectrum
      */

    /// ///////////////     TOUCH     //////////////////

      /**
       * <table class="keyboard-table">
       *   <thead>
       *     <tr>
       *       <th>Target</th>
       *       <th>Gesture</th>
       *       <th>Action</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td>Slider Bar</td>
       *       <td><kbd>Tap</kbd></td>
       *       <td>Repositions the thumb to the tapped position.</td>
       *     </tr>
       *     <tr>
       *       <td>Slider Thumb</td>
       *       <td><kbd>Swipe</kbd></td>
       *       <td>Repositions the thumb to the swiped position.</td>
       *     </tr>
       *     <tr>
       *       <td>Spectrum</td>
       *       <td><kbd>Tap</kbd></td>
       *       <td>Repositions the thumb to the tapped position.</td>
       *     </tr>
       *     <tr>
       *       <td>Spectrum Thumb</td>
       *       <td><kbd>Swipe</kbd></td>
       *       <td>Repositions the thumb to the swiped position.</td>
       *     </tr>
       *   </tbody>
       * </table>
       *
       * @ojfragment touchDoc - Used in touch gesture section of classdesc,
       * and standalone gesture doc
       * @memberof oj.ojColorSpectrum
       */

      /**
       * Sets a property or a single subproperty for complex properties and notifies the component
       * of the change, triggering a [property]Changed event.
       *
       * @function setProperty
       * @param {string} property - The property name to set. Supports dot notation for subproperty access.
       * @param {*} value - The new value to set the property to.
       *
       * @expose
       * @memberof oj.ojColorSpectrum
       * @instance
       *
       * @example <caption>Set a single subproperty of a complex property:</caption>
       * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
       */
      /**
       * Retrieves a value for a property or a single subproperty for complex properties.
       * @function getProperty
       * @param {string} property - The property name to get. Supports dot notation for subproperty access.
       * @return {*}
       *
       * @expose
       * @memberof oj.ojColorSpectrum
       * @instance
       *
       * @example <caption>Get a single subproperty of a complex property:</caption>
       * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
       */
      /**
       * Performs a batch set of properties.
       * @function setProperties
       * @param {Object} properties - An object containing the property and value pairs to set.
       *
       * @expose
       * @memberof oj.ojColorSpectrum
       * @instance
       *
       * @example <caption>Set a batch of properties:</caption>
       * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
       */
       /// ///////////////     SUB-IDS     //////////////////
       /**
        * <p>Sub-ID for the vertical hue slider bar.</p>
        *
        * @ojsubid oj-hue-slider-bar
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the hue slider bar:</caption>
        * var node = myColorSpectrum.getNodeBySubId({'subId': 'oj-hue-slider-bar'} );
        */

       /**
        * <p>Sub-ID for the vertical hue slider bar value.</p>
        *
        * @ojsubid oj-hue-slider-bar-value
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the hue slider bar value:</caption>
        * var node = myColorSpectrum.getNodeBySubId( {'subId': 'oj-hue-slider-bar-value'} );
        */

       /**
        * <p>Sub-ID for the vertical hue slider thumb.</p>
        *
        * @ojsubid oj-hue-slider-thumb
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the hue slider thumb:</caption>
        * var node = myColorSpectrum.getNodeBySubId({'subId': 'oj-hue-slider-thumb'} );
        */

       /**
        * <p>Sub-ID for the horizontal opacity slider bar.</p>
        *
        * @ojsubid oj-opacity-slider-bar
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the opacity slider bar:</caption>
        * var node = myColorSpectrum.getNodeBySubId({'subId': 'oj-opacity-slider-bar'} );
        */

       /**
        * <p>Sub-ID for the horizontal opacity slider bar value.</p>
        *
        * @ojsubid oj-opacity-slider-bar-value
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the opacity slider bar value:</caption>
        * var node = myColorSpectrum.getNodeBySubId({'subId': 'oj-opacity-slider-bar-value'} );
        */

       /**
        * <p>Sub-ID for the horizontal opacity slider thumb.</p>
        *
        * @ojsubid oj-opacity-slider-thumb
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the opacity slider thumb:</caption>
        * var node = myColorSpectrum.getNodeBySubId({'subId': 'oj-opacity-slider-thumb'} );
        */

       /**
        * <p>Sub-ID for the saturation/luminosity spectrum. </p>
        *
        * @ojsubid oj-spectrum
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the saturation/luminosity spectrum:</caption>
        * var node = myColorSpectrum.getNodeBySubId( {'subId': 'oj-spectrum'} );
        *
        */

       /**
        * <p>Sub-ID for the saturation/luminosity spectrum thumb. </p>
        *
        * @ojsubid oj-spectrum-thumb
        * @memberof oj.ojColorSpectrum
        *
        * @example <caption>Get the node for the saturation/luminosity spectrum thumb:</caption>
        * var node = myColorSpectrum.getNodeBySubId({'subId': 'oj-spectrum-thumb'} );
        *
        */

  }) ;    // end    $.widget("oj.ojColorSpectrum"

})();

(function() {
var ojColorSpectrumMeta = {
  "properties": {
    "labelledBy": {
      "type": "string"
    },
    "transientValue": {
      "readOnly": true,
      "writeback": true
    },
    "value": {
      "writeback": true
    },
    "translations": {
      "type": "Object",
      "properties": {
        "labelHue": {
          "type": "string",
          "value": "Hue"
        },
        "labelOpacity": {
          "type": "string",
          "value": "Opacity"
        },
        "labelSatLum": {
          "type": "string",
          "value": "Saturation/Luminance"
        },
        "labelThumbDesc": {
          "type": "string",
          "value": "color spectrum four way slider"
        }
      }
    },
  },
  "methods": {},
  "extension": {
    _WIDGET_NAME: "ojColorSpectrum"
  }
};
oj.CustomElementBridge.registerMetadata('oj-color-spectrum', 'editableValue', ojColorSpectrumMeta);
oj.CustomElementBridge.register('oj-color-spectrum', {'metadata': oj.CustomElementBridge.getMetadata('oj-color-spectrum')});

})();

/**
 * @protected
 * @ignore
 */
(function()
{

  oj.__KO_CUSTOM_BINDING_PROVIDER_INSTANCE.addPostprocessor
  (
    {
      'preprocessNode': function(node, wrappedReturn)
      {
        var newNodes;
        if (node.nodeType === 1)
        {
          var nodeName = node.nodeName.toLowerCase();
          if ('oj-slot' === nodeName || 'oj-bind-slot' === nodeName)
          {
            var attrs = ['name', 'slot', 'index'];

            var binding = 'ko _ojBindSlot_:{'
            var valueExpressions = [];

            for (var i=0; i<attrs.length; i++)
            {
              var attr = attrs[i];
              var expr = _getExpression(node.getAttribute(attr));
              if (expr)
              {
                valueExpressions.push(attr + ':' + expr);
              }
            }
            binding += valueExpressions.join(',');
            binding += '}';

            var openComment = document.createComment(binding);

            var closeComment = document.createComment('/ko');

            newNodes = [openComment];

            var parent  = node.parentNode;
            parent.insertBefore(openComment, node); // @HTMLUpdateOK

            // Copy the 'fallback content' children into the comment node
            while (node.childNodes.length > 0)
            {
              var child = node.childNodes[0];
              parent.insertBefore(child, node); // @HTMLUpdateOK
              newNodes.push(child);
            }

            newNodes.push(closeComment);

            parent.replaceChild(closeComment, node);
          }
        }
        return newNodes? newNodes: wrappedReturn;
      }
    }

  );

  function _getExpression(attrValue)
  {
    if (attrValue != null)
    {
      var exp = oj.__AttributeUtils.getExpressionInfo(attrValue).expr;
      if (exp == null)
      {
        exp = "'" + attrValue + "'";
      }
      return exp;
    }

    return null;
  }

}
)();

/**
 * @ignore
 */
oj.CompositeTemplateRenderer = {};


/**
 * @ignore
 */
oj.CompositeTemplateRenderer.renderTemplate = function(params, element, view) 
{
  // Store composite children on a hidden node while slotting to avoid stale knockout bindings
  // when observables are updated while children are disconnected from DOM. The _storeNodes methods
  // also adds the storage node to the view so it's added to the DOM in setDomNodChildren
  var nodeStorage = oj.CompositeTemplateRenderer._storeNodes(element, view);
  ko.virtualElements.setDomNodeChildren(element, view);

  // Attached is deprecated in 4.2.0 for connected which is called when the view is first attached to the DOM
  // and then each time the component is connected to the DOM after a disconnect
  oj.CompositeTemplateRenderer.invokeViewModelMethod(params.viewModel, 'attached', [params.viewModelContext]);
  oj.CompositeTemplateRenderer.invokeViewModelMethod(params.viewModel, 'connected', [params.viewModelContext]);
  
  var bindingContext = oj.CompositeTemplateRenderer._createKoBindingContext(element);

  // Null out the parent references since we don't want the composite View to be able to access the outside context
  var childBindingContext = bindingContext['createChildContext'](params.viewModel, undefined, 
    function (ctx) {
      ctx[oj.Composite.__COMPOSITE_PROP] = element;
      ctx['__oj_slots'] = params.slotMap;
      ctx['__oj_nodestorage'] = nodeStorage;
      ctx['$slotNodeCounts'] = params.slotNodeCounts;
      ctx['$slotCounts'] = params.slotNodeCounts;
      ctx['$props'] = params.props;
      ctx['$properties'] = params.props;
      ctx['$unique'] = params.unique;
      ctx['$uniqueId'] = params.uniqueId;
      ctx['$parent'] = null;
      ctx['$parentContext'] = null;
      ctx['$parents'] = null;
    }
  );
  
  ko.applyBindingsToDescendants(childBindingContext, element);

  oj.CompositeTemplateRenderer.invokeViewModelMethod(params.viewModel, 'bindingsApplied', [params.viewModelContext]);
};

/**
 * @ignore
 */
oj.CompositeTemplateRenderer.getEnclosingComposite = function(node)
{
  var enclosing = null;

  for(var ctx = ko.contextFor(node); ctx && !enclosing; ctx = ctx['$parentContext'])
  {
    enclosing = ctx[oj.Composite.__COMPOSITE_PROP];
  }
  
  return enclosing;
};

/**
 * @ignore
 */
oj.CompositeTemplateRenderer.createTracker = function()
{
  return ko.observable();
};

/**
 * @ignore
 */
oj.CompositeTemplateRenderer.invokeViewModelMethod = function(model, name, args)
{
  if (model == null)
  {
    return;
  }
  var handler = model[name];
  if (typeof handler === 'function')
  {
    return ko.ignoreDependencies(handler, model, args);
  }
};

/**
 * @ignore
 */
oj.CompositeTemplateRenderer._storeNodes = function(element, view)
{
  var nodeStorage;
  var childNodes = element.childNodes;
  if (childNodes)
  {
    nodeStorage = document.createElement('div');
    nodeStorage.setAttribute('data-bind', '_ojNodeStorage_')
    nodeStorage.style.display = 'none';
    view.push(nodeStorage);
    var assignableNodes = [];
    for (var i = 0; i < childNodes.length; i++)
    {
      var node = childNodes[i];
      if (oj.BaseCustomElementBridge.isSlotAssignable(node))
      {
        assignableNodes.push(node);
      }
    }
    assignableNodes.forEach(function (node) {
      nodeStorage.appendChild(node); // @HTMLUpdateOK
    });
    // Notifies JET components inside nodeStorage that they have been hidden
    if (oj.Components)
      oj.Components.subtreeHidden(nodeStorage);
  }
  return nodeStorage;
};

/**
 * @ignore
 */
oj.CompositeTemplateRenderer._createKoBindingContext = function(elem)
{
  var div = document.createElement("div");
  ko.applyBindings(null, div);
  var context = ko.contextFor(div);
  
  ko.cleanNode(div);
  
  return context;
};
ko['bindingHandlers']['_ojNodeStorage_'] =
{
  'init': function()
  {
    return {'controlsDescendantBindings' : true};
  }
};

ko['bindingHandlers']['_ojBindSlot_'] =
{
  'init': function(element, valueAccessor, allBindingsAccessor, viewModel, bindingContext)
  {

    function cleanup(bindingContext)
    {
      var nodeStorage = bindingContext['__oj_nodestorage'];
      // Move all non default slot children to nodeStorage to keep bindings alive for
      // case where knockout if binding cleans up node when toggling state
      if (nodeStorage)
      {
        // Check to see if we've processed this node as an assigned node, skipping it if we haven't
        var node = ko.virtualElements.firstChild(element);
        while (node)
        {
          // Save a reference to the next node before we move it
          var next = ko.virtualElements.nextSibling(node);
          if (node['__oj_slots'] != null)
          {
            nodeStorage.appendChild(node); // @HTMLUpdateOK
            // Notifies JET components in node that they have been hidden
            if (oj.Components)
              oj.Components.subtreeHidden(node);
          }
          node = next;
        }
      }
    }
    ko.utils.domNodeDisposal.addDisposeCallback(element, cleanup.bind(null, bindingContext));

    var slots = bindingContext['__oj_slots'];

    var values = valueAccessor();
    var unwrap = ko.utils.unwrapObservable;
    var slotName =  unwrap(values['name']) || '';
    var index =  unwrap(values['index']);
    var assignedNodes = index != null ? [slots[slotName][index]] : slots[slotName];

    if (assignedNodes)
    {
      for (var i = 0; i < assignedNodes.length; i++)
      {
        // Save references to nodes we need to cleanup ._slot field
        var node = assignedNodes[i];
        // Get the slot value of this oj-bind-slot element so we can assign it to its
        // assigned nodes for downstream slotting
        node['__oj_slots'] = unwrap(values['slot']) || '';
      }
      ko.virtualElements.setDomNodeChildren(element, assignedNodes);
      
      // Notifies JET components in node that they have been shown
      if (oj.Components)
      {
        for (var i = 0; i < assignedNodes.length; i++)
        {
          oj.Components.subtreeShown(assignedNodes[i]);
        }
      }

      // If no assigned nodes, let ko apply bindings to default slot content
      return {'controlsDescendantBindings' : true};
    }
  }
}

// Allow _ojBindSlot_ binding on virtual elements (comment nodes) which is done during knockout's preprocessNode method
ko.virtualElements.allowedBindings['_ojBindSlot_'] = true;

/**
 * @ojstatus preview
 * @ojcomponent oj.ojBindSlot
 * @ojshortdesc A placeholder for child DOM to appear in a specified slot.
 * @ojbindingelement
 * @since 4.1.0
 *
 * @classdesc
 * <h3 id="overview-section">
 *   Slot Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#overview-section"></a>
 * </h3>
 * <p>
 * The oj-bind-slot element is used inside a composite View as a placeholder for child DOM and is also
 * referred to as a slot. Default markup can be defined if no child DOM is assigned to that particular 
 * slot by adding the markup as children of the slot. An oj-bind-slot element with a name attribute
 * whose value is not the empty string is referred to as a named slot. A slot without a name attribute or one
 * whose name value is the empty string is referred to as the default slot where any composite
 * children without a slot attribute will be moved to.
 * </p> 
 * 
 * <h3 id="slotprops-section">
 *   Slot Properties
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#slotprops-section"></a>
 * </h3>
 * <ul>
 *  <li>A default slot is a slot element whose slot name is the empty string or missing.</li>
 *  <li>More than one node can be assigned to the same slot.</li>
 *  <li>A slot can also have a slot attribute and be assigned to another slot.</li>
 *  <li>A slot can have fallback content which are its child nodes that will be used in the DOM in its place if it has no assigned nodes.</li>
 *  <li>A slot can also also have an index attribute to allow the slot's assigned nodes
 *    to be individually slotted (e.g. in conjunction with an oj-bind-for-each element).</li>
 * </ul>
 * 
 * <h3 id="nodeprops-section">
 *   Assignable Node Properties
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#nodeprops-section"></a>
 * </h3>
 * <ul>
 *  <li>Nodes with slot attributes will be assigned to the corresponding named slots (if
 *    present) and all other assignable nodes (Text or Element) will be assigned to
 *    the default slot (if present).</li>
 *  <li>The slot attribute of a node is only applied once. If the View contains a
 *    composite and the node's assigned slot is a child of that composite, the slot
 *    attribute of the assigned slot is inherited for the slotting of that composite.</li>
 *  <li>Nodes with slot attributes that reference slots not present in the View will not appear in the DOM.</li>
 *  <li>If the View does not contain a default slot, nodes assigned to the default slot will not appear in the DOM.</li>
 *  <li>Nodes that are not assigned to a slot will not appear in the DOM.</li>
 * </ul>
 *
 * <h3 id="example1-section">
 *   Example #1: Basic Usage
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#example1-section"></a>
 * </h3>
 * Note that the IDs are provided for sample purposes only.
 * <h4>Initial DOM</h4>
 * <pre class="prettyprint">
 * <code>
 * &lt;component-a>
 *  &lt;div id="A" slot="foo">&lt;/div>
 *  &lt;div id="B" slot="bar">&lt;/div>
 *  &lt;div id="C">&lt;/div>
 *  &lt;div id="D" slot="foo">&lt;/div>
 *  &lt;div id="E" slot="cat">&lt;/div>
 * &lt;/component-a>
 * </code>
 * </pre>
 *
 * <h4>View</h4>
 * <pre class="prettyprint">
 * <code>
 * &lt;!-- component-a View -->
 * &lt;div id="outerFoo">
 *  &lt;oj-bind-slot name="foo">&lt;/oj-bind-slot>
 * &lt;/div>
 * &lt;div id="outerBar">
 *  &lt;oj-bind-slot name="bar">&lt;/oj-bind-slot>
 * &lt;/div>
 * &lt;div id="outerBaz">
 *  &lt;oj-bind-slot name="baz">
 *    &lt;!-- Default Content -->
 *    &lt;img id="F">&lt;/img>
 *    &lt;div id="G">&lt;/div>
 *  &lt;/oj-bind-slot>
 * &lt;/div>
 * &lt;div id="outerDefault">
 *  &lt;oj-bind-slot>
 *    &lt;!-- Default Content -->
 *    &lt;div id="H">&lt;/div>
 *  &lt;/oj-bind-slot>
 * &lt;/div>
 * </code>
 * </pre>
 *
 * <h4>Final DOM</h4>
 * <pre class="prettyprint">
 * <code>
 * &lt;component-a>
 *  &lt;div id="outerFoo">  
 *      &lt;div id="A" slot="foo">&lt;/div>
 *      &lt;div id="D" slot="foo">&lt;/div>
 *  &lt;/div>
 *  &lt;div id="outerBar">
 *      &lt;div id="B" slot="bar">&lt;/div>
 *   &lt;/div>
 *  &lt;div id="outerBaz">
 *      &lt;img id="F">&lt;/img>
 *      &lt;div id="G">&lt;/div>
 *  &lt;/div>
 *  &lt;div id="outerDefault">
 *      &lt;div id="C">&lt;/div>
 *  &lt;/div>
 * &lt;/component-a>
 * </code>
 * </pre>
 *
 * <h3 id="example2-section">
 *   Example #2: Slot Attribute Evaluation
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#example2-section"></a>
 * </h3>
 * <p>When a node is assigned to a slot, its slot value is not used for subsequent
 *  slot assignments when child bindings are applied. Instead that slot's slot attribute,
 *  which by default is "", overrides the assigned node's slot attribute. No actual
 *  DOM changes will be made to the assigned node's slot attribute, but its evaluated
 *  slot value will be managed internally and used for applying subsequent child bindings.</p>
 *
 * <h4>Initial DOM</h4>
 * <pre class="prettyprint">
 * <code>
 * &lt;component-a>
 *  &lt;div id="A" slot="foo">&lt;/div>
 * &lt;/component-a>
 * </code>
 * </pre>
 *
 * <h4>View</h4>
 * <pre class="prettyprint">
 * <code>
 * &lt;!-- component-a View -->
 * &lt;component-b>
 *  &lt;oj-bind-slot name="foo">&lt;/oj-bind-slot>
 * &lt;/component-b>
 *
 * &lt;!-- component-b View -->
 * &lt;div id="outerFoo">
 *  &lt;oj-bind-slot name="foo">&lt;/oj-bind-slot>
 * &lt;/div>
 * &lt;div id="outerDefault">
 *  &lt;oj-bind-slot>&lt;/oj-bind-slot>
 * &lt;/div>
 * </code>
 * </pre>
 *
 * <p>When applying bindings for the component-a View, the oj-bind-slot binding will replace
 * slot foo with div A. Slot foo's slot attribute ("") overrides div A's ("foo")
 * so that the evaluated slot value ("") will be used when applying subsequent child bindings.<p>
 * <pre class="prettyprint">
 * <code>
 * &lt;!-- DOM -->
 * &lt;component-a>
 *  &lt;!-- Start component-a View -->
 *  &lt;component-b>
 *    &lt;!-- Evaluated slot value is "" -->
 *    &lt;div id="A" slot="foo">&lt;/div>
 *  &lt;/component-b>
 *  &lt;!-- End component-a View -->
 * &lt;/component-a>
 * </code>
 * </pre>
 *
 * <p>When applying bindings for the component-b View, the oj-bind-slot binding will replace
 *  component-b's default slot with div A since it's evaluated slot value is "".</p>
 * <pre class="prettyprint">
 * <code>
 * &lt;!-- DOM -->
 * &lt;component-a>
 *  &lt;!-- Start component-a View -->
 *  &lt;component-b>
 *    &lt;!-- Start component-b View -->
 *    &lt;div id="outerFoo">
 *    &lt;/div>
 *    &lt;div id="outerDefault">
 *      &lt;div id="A" slot="foo">&lt;/div>
 *    &lt;/div>
 *    &lt;!-- End component-b View -->
 *  &lt;/component-b>
 *  &lt;!-- End component-a View -->
 * &lt;/component-a>
 * </code>
 * </pre>
 *
 * <h3 id="deferred-section">
 *   Deferred Slots
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#deferred-section"></a>
 * </h3>
 * As a performance enhancement, the composite can participate in deferred slot
 * rendering by conditionally rendering a slot element inside a conditional oj-bind-if element
 * and document that certain slots will be lazily rendered. This gives the application the opportunity
 * to wrap their slot content in an <a href="oj.ojDefer.html">oj-defer</a> element and have the
 * bindings for that deferred content be delayed. oj.Components.subtreeHidden/Shown will automatically
 * be called on the slot contents when they are added or removed from a slot. <b>Note that due to a current
 * limitation, the slot element should be wrapped in an HTML element (e.g. &lt;div> or &lt;span>), when it is a child of an
 * oj-bind-if element or a knockout if binding on a comment node.</b> It is not required to wrap when
 * the slot element is a child of a knockout if binding on an HTML element.
 *
 * <h4>Initial DOM</h4>
 * <pre class="prettyprint">
 * <code>
 * &lt;card-component>
 *  &lt;oj-defer slot="front">
 *    &lt;div>
 *      ...
 *    &lt;/div>
 *  &lt;/oj-defer>
 *  &lt;oj-defer slot="back">
 *    &lt;div>
 *      ...
 *    &lt;/div>
 *  &lt;/oj-defer>
 * &lt;/card-component>
 * </code>
 * </pre>
 *
 * <h4>View</h4>
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-bind-if test="isFront">
 *   &lt;div>
 *     &lt;oj-bind-slot class="card-component-front">
 *       ...
 *     &lt;/oj-bind-slot>
 *   &lt;/div>
 * &lt;/oj-bind-if>
 * &lt;oj-bind-if test="!isFront">
 *   &lt;div>
 *     &lt;oj-bind-slot class="card-component-back">
 *       ...
 *     &lt;/oj-bind-slot>
 *   &lt;/div>
 * &lt;/oj-bind-if>
 * </code>
 * </pre>
 */

/**
 * An index value allowing the slot children to be individually slotted. This is useful
 * when the composite needs to add additional DOM around slotted children.
 * @expose
 * @name index
 * @memberof oj.ojBindSlot
 * @instance
 * @type {number}
 * @example <caption>
 *          Use an oj-bind-for-each element inside the composite View to stamp out
 *          li wrapped oj-bind-slot elements that correspond to the number of slot children.
 *          The oj-bind-slot elements should have the value for the name attribute, but different indices.
 *          </caption>
 * &lt;!-- Composite View -->
 * &lt;ul>
 *   &lt;oj-bind-for-each data="[[new Array($slotCounts.foo)]]">
 *     &lt;template>
 *       &lt;li>
 *         &lt;oj-bind-slot name="foo" index="[[$current.index]]">&lt;/oj-bind-slot>
 *       &lt;/li>
 *     &lt;/template>
 *   &lt;/oj-bind-for-each>
 * &lt;/ul>
 */

/**
 * The name of the slot.
 * @expose
 * @name name
 * @memberof oj.ojBindSlot
 * @instance
 * @type {boolean}
 * @example <caption>Define a slot within a composite View with the name "foo":</caption>
 * &lt;oj-bind-slot name="foo">
 *   &lt;div>My Contents&lt;/div>
 * &lt;/oj-bind-slot>
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*
** Important:
** - This file is designed to be shared verbatim among the ADFui products.
** - Do not add framework-specific dependencies in this file (it must be self-contained).
** - Do not change this file without testing it in other ADFui products (ADF Faces, JET, etc.).
*/

/**
 * Constructor.
 * @param {Object} elem DOM element associated with the conveyor
 * @param {string} orientation Direction of scrolling, either horizontal or vertical
 * @param {Object} contentParent DOM element whose children are the items to scroll
 * @param {boolean} bRtl True if the reading direction is right-to-left, otherwise false
 * @param {Object} buttonInfo Map of properties for the following button information:
 *  - prevButtonId: Id to use for the scroll previous button,
 *  - nextButtonId: Id to use for the scroll next button,
 *  - prevButtonStyleClass: Style class name to use for the scroll previous button,
 *  - nextButtonStyleClass: Style class name to use for the scroll next button,
 *  - prevButtonIcon: Icon element to use for the scroll previous button,
 *  - nextButtonIcon: Icon element to use for the scroll next button,
 * @param {Object} callbackInfo Map of properties for the following callback information:
 *  - scrollFunc: Callback function to animate scrolling a DOM element, 
 *  - firstVisibleItemChangedFunc: Callback function to notify when the first visible item changes, 
 *  - addResizeListener: Callback function to add a resize listener for a DOM element
 *  - removeResizeListener: Callback function to remove a resize listener for a DOM element
 *  - addStyleClassName: Callback function to add a style class name to a DOM element
 *  - removeStyleClassName: Callback functin to remove a style class name from a DOM element
 *  - hasStyleClassName: Callback function to check whether a style class name is applied
 *    to a DOM element,
 *  - filterContentElements: Callback function to filter the array of conveyor content elements,
 *  - subtreeAttached: Callback function to notify when a DOM subtree is attached,
 *  - subtreeDetached: Callback function to notify when a DOM subtree is detached,
 *  - addBusyState: Callback function to add a busy state to the busy context,
 *  - callbackObj: Optional object on which the callback functions are defined
 *    has changed
 * @param {Object} styleInfo Map of properties for the following style information:
 *  - overflowContainerStyleClass: Style class name for the overflow container DOM element,
 *  - contentContainerStyleClass: Style class name for the content container DOM element,
 *  - itemStyleClass: Style class name for the DOM element of an item in the conveyor,
 *  - hiddenStyleClass: Style class name used to hide a DOM element
 * @param {Object} agentInfo Browser user agent information.
 * @class ConveyorBeltCommon
 * @memberof ConveyorBeltCommon
 * @protected
 * @constructor
 * @ignore
 * @ojtsignore
 */
function ConveyorBeltCommon(
  elem, orientation, contentParent, bRtl, buttonInfo, callbackInfo, styleInfo, agentInfo)
{
  this._elem = elem;
  this._orientation = orientation;
  this._contentParent = contentParent;
  this._bRtl = bRtl;
  if (buttonInfo)
  {
    //need to check for existence of properties on buttonInfo before assigning 
    //them to vars because advanced closure compiler will otherwise complain if 
    //they're not defined
    if (buttonInfo.prevButtonId)
      this._prevButtonId = buttonInfo.prevButtonId;
    if (buttonInfo.nextButtonId)
      this._nextButtonId = buttonInfo.nextButtonId;
    if (buttonInfo.prevButtonStyleClass)
      this._prevButtonStyleClass = buttonInfo.prevButtonStyleClass;
    if (buttonInfo.nextButtonStyleClass)
      this._nextButtonStyleClass = buttonInfo.nextButtonStyleClass;
    if (buttonInfo.prevButtonIcon)
      this._prevButtonIcon = buttonInfo.prevButtonIcon;
    if (buttonInfo.nextButtonIcon)
      this._nextButtonIcon = buttonInfo.nextButtonIcon;
  }
  if (callbackInfo)
  {
    //need to check for existence of properties on callbackInfo before assigning 
    //them to vars because advanced closure compiler will otherwise complain if 
    //they're not defined
    if (callbackInfo.scrollFunc)
      this._scrollFunc = callbackInfo.scrollFunc;
    if (callbackInfo.firstVisibleItemChangedFunc)
      this._firstVisibleItemChangedFunc = callbackInfo.firstVisibleItemChangedFunc;
    if (callbackInfo.callbackObj)
      this._callbackObj = callbackInfo.callbackObj;
    if (callbackInfo.addResizeListener)
      this._addResizeListenerFunc = callbackInfo.addResizeListener;
    if (callbackInfo.removeResizeListener)
      this._removeResizeListenerFunc = callbackInfo.removeResizeListener;
    if (callbackInfo.addStyleClassName)
      this._addStyleClassNameFunc = callbackInfo.addStyleClassName;
    if (callbackInfo.removeStyleClassName)
      this._removeStyleClassNameFunc = callbackInfo.removeStyleClassName;
    if (callbackInfo.hasStyleClassName)
      this._hasStyleClassNameFunc = callbackInfo.hasStyleClassName;
    if (callbackInfo.filterContentElements)
      this._filterContentElementsFunc = callbackInfo.filterContentElements;
    if (callbackInfo.subtreeDetached)
      this._subtreeDetachedFunc = callbackInfo.subtreeDetached;
    if (callbackInfo.subtreeAttached)
      this._subtreeAttachedFunc = callbackInfo.subtreeAttached;
    if (callbackInfo.addBusyState)
      this._addBusyStateFunc = callbackInfo.addBusyState;
  }
  if (styleInfo)
  {
    if (styleInfo.overflowContainerStyleClass)
      this._overflowContainerStyleClass = styleInfo.overflowContainerStyleClass;
    if (styleInfo.contentContainerStyleClass)
      this._contentContainerStyleClass = styleInfo.contentContainerStyleClass;
    if (styleInfo.itemStyleClass)
      this._itemStyleClass = styleInfo.itemStyleClass;
    if (styleInfo.hiddenStyleClass)
      this._hiddenStyleClass = styleInfo.hiddenStyleClass;
  }
  
  this._bExternalScroll = true;
  this._firstVisibleItemIndex = 0;
  if (agentInfo)
    this._agentVersion = agentInfo["browserVersion"];
  //copied basic checks from AdfAgent
  var navUserAgent = navigator.userAgent;
  var agentName = navUserAgent.toLowerCase();
  if (agentName.indexOf("gecko/") !== -1)
  {
    this._bAgentGecko = true;
  }
  else if (agentName.indexOf("opera") !== -1)
  {
    this._bAgentOpera = true;
  }
  else if (agentInfo && agentInfo["browser"] === "safari")
  {
    this._bAgentSafari = true;
  }
  else if (agentInfo && agentInfo["browser"] === "edge")
  {
    this._bAgentEdge = true;
  }
  else if (agentName.indexOf("applewebkit") !== -1 ||
           agentName.indexOf("safari") !== -1)
  {
    this._bAgentWebkit = true;
  }
};

/**
 * Setup the conveyor belt.
 * @param {boolean} bInit True for initialization, false for refresh
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 */
ConveyorBeltCommon.prototype.setup = function(bInit)
{
  var self = this;
  var cbcClass = ConveyorBeltCommon;
  if (bInit)
  {
    //create the content and overflow containers
    this._createInnerContainers();
    
    //create the next/prev buttons
    this._createPrevButton(this._prevButtonId, this._prevButtonStyleClass, this._prevButtonIcon);
    this._createNextButton(this._nextButtonId, this._nextButtonStyleClass, this._nextButtonIcon);
    
    var nextButton = this._nextButton;
    this._buttonWidth = nextButton.offsetWidth;
    this._buttonHeight = nextButton.offsetHeight;
    
    //hide the buttons until we know we need them
    this._hidePrevButton();
    this._hideNextButton();
    
    //handle the mouse wheel on the whole conveyor
    this._mouseWheelListener = function (event) {self._handleMouseWheel(event);};
    cbcClass._addBubbleEventListener(this._elem, "mousewheel", this._mouseWheelListener);
    cbcClass._addBubbleEventListener(this._elem, "wheel", this._mouseWheelListener);
    
    //handle swipe gestures on the overflow container, which excludes the next/prev buttons
    this._touchStartListener = function (event) {self._handleTouchStart(event);};
    cbcClass._addBubbleEventListener(this._overflowContainer, "touchstart", this._touchStartListener);
    this._touchMoveListener = function (event) {self._handleTouchMove(event);};
    cbcClass._addBubbleEventListener(this._overflowContainer, "touchmove", this._touchMoveListener);
    this._touchEndListener = function (event) {self._handleTouchEnd(event);};
    cbcClass._addBubbleEventListener(this._overflowContainer, "touchend", this._touchEndListener);
    cbcClass._addBubbleEventListener(this._overflowContainer, "touchcancel", this._touchEndListener);
    
    this._origScroll = 0;
  }
  else
  {
    //need to restore inner DOM to its initial state before we can accurately
    //calculate sizes
    this._reinitializeInnerDom();
  }

  //clear any old sizes so that new sizes will be calculated
  this._clearCachedSizes();
  //adjust overflow size
  this._adjustOverflowSize(bInit);
  //handle an initial resize
  this._handleResize(true);
  
  if (bInit)
  {
    if (this._addResizeListenerFunc)
    {
      this._handleResizeFunc = function(width, height) {self._handleResize(false);};
      //listen for resizes on both the conveyor itself and on its content
      this._addResizeListenerFunc.call(this._callbackObj, this._elem, this._handleResizeFunc);
      this._addResizeListenerFunc.call(this._callbackObj, this._contentContainer, this._handleResizeFunc);
    }
    
    //notify the child that it's being re-attached to the DOM AFTER attaching it
    //(the detached notification happened in _reparentChildrenToContentContainer())
    if (this._subtreeAttachedFunc)
      this._subtreeAttachedFunc(this._contentContainer);
  }
};

/**
 * Destroy the conveyor belt.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 */
ConveyorBeltCommon.prototype.destroy = function()
{
  //FIX : resolve busy state when destroying
  this._resolveBusyState();
  
  var elem = this._elem;
  var cbcClass = ConveyorBeltCommon;
  cbcClass._removeBubbleEventListener(elem, "mousewheel", this._mouseWheelListener);
  cbcClass._removeBubbleEventListener(elem, "wheel", this._mouseWheelListener);
  cbcClass._removeBubbleEventListener(this._overflowContainer, "touchstart", this._touchStartListener);
  cbcClass._removeBubbleEventListener(this._overflowContainer, "touchmove", this._touchMoveListener);
  cbcClass._removeBubbleEventListener(this._overflowContainer, "touchend", this._touchEndListener);
  cbcClass._removeBubbleEventListener(this._overflowContainer, "touchcancel", this._touchEndListener);
  cbcClass._removeBubbleEventListener(this._overflowContainer, "scroll", this._scrollListener);
  this._mouseWheelListener = null;
  this._touchStartListener = null;
  this._touchMoveListener = null;
  this._touchEndListener = null;
  this._scrollListener = null;
  
  //remove listeners before reparenting original children and clearing member 
  //variables
  if (this._removeResizeListenerFunc && this._handleResizeFunc)
  {
    this._removeResizeListenerFunc.call(this._callbackObj, elem, this._handleResizeFunc);
    this._removeResizeListenerFunc.call(this._callbackObj, this._contentContainer, this._handleResizeFunc);
  }
  this._handleResizeFunc = null;
  
  //move the original content children from the _contentContainer back to the 
  //original DOM element
  this._reparentChildrenFromContentContainer(this._contentContainer, elem);
  
  //the content container is a child of the overflow container
  elem.removeChild(this._overflowContainer);
  elem.removeChild(this._nextButton);
  elem.removeChild(this._prevButton);
  
  this._nextButton = null;
  this._prevButton = null;
  this._overflowContainer = null;
  this._contentContainer = null;
  this._clearCachedSizes();
  
  this._elem = null;
  this._scrollFunc = null;
  this._firstVisibleItemChangedFunc = null;
  this._addResizeListenerFunc = null;
  this._removeResizeListenerFunc = null;
  this._addStyleClassNameFunc = null;
  this._removeStyleClassNameFunc = null;
  this._hasStyleClassNameFunc = null;
  this._filterContentElementsFunc = null;
  this._subtreeDetachedFunc = null;
  this._subtreeAttachedFunc = null;
  this._addBusyStateFunc = null;
  this._callbackObj = null;

  this._contentParent = null;

};

/**
 * Set the id of the item to make first visible.
 * @param {string} id id of the item to make first visible
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 */
ConveyorBeltCommon.prototype.setFirstVisibleItem = function(id)
{
  var sizes = this._getSizes();
  for (var i = 0; i < sizes.length; i++)
  {
    var sizeObj = sizes[i];
    if (sizeObj.id === id)
    {
      this._setCurrScroll(sizeObj.start, true);
      break;
    }
  }
};

/**
 * Handle a component resize.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 */
ConveyorBeltCommon.prototype.handleResize = function()
{
  this._handleResize(false);
};

/**
 * Set the scroll position.
 * @param {number} scroll Desired logical scroll position
 * @param {boolean} bImmediate True to make the change immediately, false to animate it
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 */
ConveyorBeltCommon.prototype.setScroll = function(scroll, bImmediate)
{
  this._setCurrScroll(scroll, bImmediate);
};

/**
 * Get the current scroll position.
 * @return {number} Logical scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 */
ConveyorBeltCommon.prototype.getScroll = function()
{
  return this._getCurrScroll();
};

/**
 * Reparent the DOM child nodes from their old parent node to a new parent 
 * content container node.
 * @param {Object} fromNode Old parent DOM node
 * @param {Object} toNode New parent DOM node
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._reparentChildrenToContentContainer = function(
  fromNode, toNode)
{
  var fromNodeChildren = fromNode.childNodes;
  while (fromNodeChildren.length > 0)
  {
    var child = fromNodeChildren[0];
    
    //notify the child that it's being detached from the DOM BEFORE detaching it
    //(the re-attached notification will happen in setup())
    if (this._subtreeDetachedFunc)
      this._subtreeDetachedFunc(child);
    
    toNode.appendChild(child); // @HtmlUpdateOK
    
    if (child.nodeType === 1 && this._itemStyleClass)
      this._addStyleClassNameFunc(child, this._itemStyleClass);
  }
};

/**
 * Reparent the DOM child nodes from the content container to a new 
 * parent node.
 * @param {Object} fromNode Old parent DOM node
 * @param {Object} toNode New parent DOM node
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._reparentChildrenFromContentContainer = function(
  fromNode, toNode)
{
  if (!fromNode)
    return;
  
  var children = fromNode.childNodes;
  while (children.length > 0)
  {
    var child = children[0];
    toNode.appendChild(child); // @HtmlUpdateOK
    
    if (child.nodeType === 1 && this._itemStyleClass)
      this._removeStyleClassNameFunc(child, this._itemStyleClass);
  }
};

/**
 * Get the computed style of the given DOM element.
 * @param {Object} elem DOM element
 * @return {Object} Computed style for the element
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._getComputedStyle = function(elem)
{
  var elemOwnerDoc = elem.ownerDocument;
  var defView = elemOwnerDoc.defaultView;
  var computedStyle = null;
  if (defView)
  {
    //this line copied from AdfAgent.getComputedStyle()
    computedStyle = defView.getComputedStyle(elem, null);
  }
  else
  {
    //this line copied from AdfIEAgent.getComputedStyle()
    computedStyle = elem.currentStyle;
  }
  return computedStyle;
};

/**
 * Get the inner width of the given DOM element (inside borders and padding).
 * @param {Object} elem DOM element
 * @return {number} Width of element
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._getElemInnerWidth = function(elem)
{
  var cbcClass = ConveyorBeltCommon;
  var computedStyle = cbcClass._getComputedStyle(elem);
  //the computedStyle width is the inner width of the elem
  return cbcClass._getCSSLengthAsInt(computedStyle.width);
};

/**
 * Get the inner height of the given DOM element (inside borders and padding).
 * @param {Object} elem DOM element
 * @return {number} Height of element
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._getElemInnerHeight = function(elem)
{
  var cbcClass = ConveyorBeltCommon;
  var computedStyle = cbcClass._getComputedStyle(elem);
  //the computedStyle height is the inner height of the elem
  return cbcClass._getCSSLengthAsInt(computedStyle.height);
};

/**
 * Get the int value of a CSS length.
 * @param {string} cssLength cssLength as a string
 * @return {number} cssLength as an int
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._getCSSLengthAsInt = function(cssLength)
{
  //this function copied from AdfAgent.getCSSLengthAsInt
  if ((cssLength.length) > 0 && (cssLength != 'auto'))
  {
    var intLength = parseInt(cssLength, 10);

    if (isNaN(intLength))
      intLength = 0;

    return intLength;
  }
  return 0;
};

/**
 * Add a bubble event listener to the given DOM node.
 * @param {Object} node DOM node
 * @param {string} type Event type
 * @param {function(Event):void} listener Listener function
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._addBubbleEventListener = function(node, type, listener)
{
  if (node.addEventListener)
  {
    node.addEventListener(type, listener, false);
  }
  else if (node.attachEvent)
  {
    node.attachEvent("on" + type, listener);
  }
};

/**
 * Remove a bubble event listener from the given DOM node.
 * @param {Object} node DOM node
 * @param {string} type Event type
 * @param {function(Event):void} listener Listener function
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._removeBubbleEventListener = function(node, type, listener)
{
  if (node.removeEventListener)
  {
    node.removeEventListener(type, listener, false);
  }
  else if (node.detachEvent)
  {
    node.detachEvent("on" + type, listener);
  }
};

/**
 * Get the wheel delta from a mousewheel event.
 * @param {Event} event Event object
 * @return {number} Wheel delta
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._getWheelDelta = function(event)
{
  var wheelDelta = 0;
  if (event.wheelDelta != null)
  {
    wheelDelta = event.wheelDelta;
  }
  //use bracket notation to avoid compilation warning in JET
  else if (event["deltaY"] != null)
  {
    wheelDelta = -event["deltaY"];
  }
  else
  {
    wheelDelta = -event.detail;
  }
  return wheelDelta;
};

/**
 * Create a div styled like a table.
 * @return {Node} div styled like a table row
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._createTableDiv = function()
{
  var tableDiv = document.createElement("div");
  var style = tableDiv.style;
  style.display = "table";
  return tableDiv;
};

/**
 * Create a div styled like a table row.
 * @return {Node} div styled like a table row
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._createTableRowDiv = function()
{
  var tableRowDiv = document.createElement("div");
  var style = tableRowDiv.style;
  style.display = "table-row";
  return tableRowDiv;
};

/**
 * Create a div styled like a table cell.
 * @return {Node} div styled like a table cell
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._createTableCellDiv = function()
{
  var tableCellDiv = document.createElement("div");
  var style = tableCellDiv.style;
  style.display = "table-cell";
  return tableCellDiv;
};

/**
 * Wrap the given element in an inline-block div and append it to the given 
 * parent element.  Restrict the max size of the wrapper div to be its reported 
 * offset size.  This helps prevent off-by-one pixel errors when the size of the
 * element is really a float instead of an int, as reported by the offset size.  
 * @param {Node} elem DOM element to wrap
 * @param {Node} parentElem Parent DOM element to append child to
 * @param {boolean} bWidth True to restrict the element's width, false otherwise
 * @param {boolean} bHeight True to restrict the element's height, false otherwise
 * @return {Node} The wrapper div element
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._wrapAndRestrictSize = function(
  elem, parentElem, bWidth, bHeight)
{
  var wrapperDiv = document.createElement("div");
  var wrapperDivStyle = wrapperDiv.style;
  //make sure the wrapper div fits its content
  wrapperDivStyle.display = "inline-block";

  //need to add everything to the DOM before getting sizes
  wrapperDiv.appendChild(elem); // @HtmlUpdateOK
  parentElem.appendChild(wrapperDiv); // @HtmlUpdateOK
  
  //restrict the size of the wrapper div so that it's an integer value
  if (bWidth)
  {
    wrapperDivStyle.maxWidth = wrapperDiv.offsetWidth + "px";
  }
  if (bHeight)
  {
    wrapperDivStyle.maxHeight = wrapperDiv.offsetHeight + "px";
  }
  
  return wrapperDiv;
};

/**
 * Determine if this conveyor belt is horizontal or vertical.
 * @return {boolean} True if the conveyor belt is horizontal, false if vertical
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._isHorizontal = function()
{
  return (this._orientation === "horizontal");
};

/**
 * Determine if this conveyor belt is empty.
 * @return {boolean} True if the conveyor belt is empty, false if not
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._isEmpty = function()
{
  var contentParent = this._getContentParent();
  return !contentParent.hasChildNodes();
};

/**
 * Restore inner DOM to its initial state before sizes were calculated.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._reinitializeInnerDom = function()
{
  //restore inner DOM to initial state in order to accurately calculate new sizes

  //save original scroll value for use in _adjustOverflowSize()
  this._origScroll = this._getCurrScroll();
  this._setOverflowScroll(0);

  //hide the buttons until we know we need them
  this._hidePrevButton();
  this._hideNextButton();
};

/**
 * Clear cached sizes.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._clearCachedSizes = function()
{
  this._totalSize = null;
  this._sizes = null;
};

/**
 * Handle a component resize.
 * @param {boolean} bSetup True when called from setup, false otherwise
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._handleResize = function(bSetup)
{
  //if this is not the first call, need to reinitialize the inner DOM before
  //we can accurately calculate new sizes (if this is the first call, DOM
  //is already in initial state)
  if (!bSetup)
  {
    this._reinitializeInnerDom();
  }
  this._clearCachedSizes();
  if (!this._totalSize || !this._sizes)
  {
    //measure content size
    this._totalSize = this._measureContents();
  }
  //if this is not the first call, need to adjust the overflow size (if this 
  //is the first call, the overflow size was already adjusted in setup)
  if (!bSetup)
  {
    this._adjustOverflowSize(false);
  }
  //center buttons orthogonal to conveyor orientation
  this._alignButtons();
};

/**
 * Center the overflow buttons orthogonal to the conveyor orientation.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._alignButtons = function()
{
  var nextButton = this._nextButton;
  var prevButton = this._prevButton;
  var nextButtonStyle = nextButton.style;
  var prevButtonStyle = prevButton.style;
  var contentContainer = this._contentContainer;
  var totalSize = this._totalSize;
  var sizes = this._sizes;
  
  if (this._isHorizontal())
  {
    var vOffset = .5 * (totalSize.h - contentContainer.offsetHeight);
    nextButtonStyle.top = vOffset + "px";
    prevButtonStyle.top = vOffset + "px";
  }
  else
  {
    var hOffset = .5 * (totalSize.w - contentContainer.offsetWidth);
    if (this._bRtl)
      hOffset = -hOffset;
    nextButtonStyle.left = hOffset + "px";
    prevButtonStyle.left = hOffset + "px";
  }
};

/**
 * Adjust the overflow size.
 * @param {boolean} bInit True for initialization, false for refresh
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._adjustOverflowSize = function(bInit)
{
  var contentContainer = this._contentContainer;
  var bHoriz = this._isHorizontal();

  var cbcClass = ConveyorBeltCommon;
  var elemInnerSize = bHoriz ? 
                      cbcClass._getElemInnerWidth(this._elem) : 
                      cbcClass._getElemInnerHeight(this._elem);
  
  this._minScroll = 0;
  //take the button size into account for max scroll position
  this._maxScroll = bHoriz ? 
                    contentContainer.offsetWidth - elemInnerSize + this._buttonWidth : 
                    contentContainer.offsetHeight - elemInnerSize + this._buttonHeight;
  //constrain max scroll
  if (this._maxScroll < 0)
  {
    this._maxScroll = 0;
  }
  
  //hide buttons AFTER calculating sizes above, but BEFORE updating scroll position below
  this._hidePrevButton();
  this._hideNextButton();
  
  //refresh current scroll position AFTER calculating sizes above
  this._setCurrScroll(bInit ? this._minScroll : this._origScroll, true);
  this._origScroll = 0;
};

/**
 * Create the inner overflow and content containers.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._createInnerContainers = function()
{
  //the original children of the conveyor elem will be reparented to the contentContainer;
  //the conveyor elem will contain the overflowContainer, which will contain the contentContainer, 
  //which will contain the original children
  
  var self = this;
  
  var elem = this._elem;
  var cbcClass = ConveyorBeltCommon;
  var overflowContainer = document.createElement("div");
  this._overflowContainer = overflowContainer;
  if (this._overflowContainerStyleClass)
    this._addStyleClassNameFunc(overflowContainer, this._overflowContainerStyleClass);
  var contentContainer = document.createElement("div");
  this._contentContainer = contentContainer;
  if (this._contentContainerStyleClass)
    this._addStyleClassNameFunc(contentContainer, this._contentContainerStyleClass);
  
  //reparent children from elem to contentContainer before adding 
  //content container to elem
  this._reparentChildrenToContentContainer(elem, contentContainer);
  
  elem.appendChild(overflowContainer); // @HtmlUpdateOK
  overflowContainer.appendChild(contentContainer); // @HtmlUpdateOK
  
  //the overflow container listens to DOM scroll events in case the scroll was triggered externally,
  //for example when the user tabs through the child content
  this._scrollListener = function (event) {self._handleScroll(event);};
  cbcClass._addBubbleEventListener(overflowContainer, "scroll", this._scrollListener);
};

/**
 * Get the content elements of the conveyor.
 * @return {Array} array of content elements
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._getContentElements = function()
{
  var arContentElements = [];
  
  //if there is a nested contentParent, then we need to put its children into
  //arContentElements instead of the children of the contentContainer
  var parent = this._contentParent ? this._contentParent : this._contentContainer;
  
  var contentChildren = parent.children;
  var numContentChildren = contentChildren.length;
  for (var i = 0; i < numContentChildren; i++)
  {
    var child = contentChildren[i];
    if (child.nodeType === 1)
    {
      arContentElements.push(child);
    }
  }
  
  if (this._filterContentElementsFunc)
  {
    var filterFunc = this._filterContentElementsFunc;
    arContentElements = filterFunc(arContentElements);
  }
  
  //if the content elements are direct children of the contentContainer, meaning there's no nested
  //contentParent, then make sure they have the itemStyleClass applied to them in case they were
  //added dynamically
  if (parent === this._contentContainer && this._itemStyleClass)
  {
    for (var i = 0; i < arContentElements.length; i++)
    {
      var contentElem = arContentElements[i];
      if (!this._hasStyleClassNameFunc(contentElem, this._itemStyleClass))
        this._addStyleClassNameFunc(contentElem, this._itemStyleClass);
    }
  }
  
  return arContentElements;
};

/**
 * Create the prev button.
 * @param {string} buttonId Id to use for the button
 * @param {string} buttonStyleClass Style class to use for the button
 * @param {Node} icon Button icon element
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._createPrevButton = function(
  buttonId, buttonStyleClass, icon)
{
  var self = this;
  var prevButton = document.createElement("div");
  this._prevButton = prevButton;
  if (buttonId)
  {
    prevButton.setAttribute("id", buttonId);
  }
  prevButton.setAttribute("class", buttonStyleClass);
  //hide the button from screen readers because it is not keyboard accessible
  prevButton.setAttribute("aria-hidden", "true");
  var bHoriz = this._isHorizontal();
  var cbcClass = ConveyorBeltCommon;
  cbcClass._addBubbleEventListener(prevButton, "click", function (event) {self._scrollPrev();});
  
  if (icon)
  {
    prevButton.appendChild(icon); // @HtmlUpdateOK
  }
  
  var elem = this._elem;
  elem.insertBefore(prevButton, this._overflowContainer); // @HtmlUpdateOK
};

/**
 * Create the next button.
 * @param {string} buttonId Id to use for the button
 * @param {string} buttonStyleClass Style class to use for the button
 * @param {Node} icon Button icon element
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._createNextButton = function(
  buttonId, buttonStyleClass, icon)
{
  var self = this;
  var nextButton = document.createElement("div");
  this._nextButton = nextButton;
  if (buttonId)
  {
    nextButton.setAttribute("id", buttonId);
  }
  nextButton.setAttribute("class", buttonStyleClass);
  //hide the button from screen readers because it is not keyboard accessible
  nextButton.setAttribute("aria-hidden", "true");
  var bHoriz = this._isHorizontal();
  var cbcClass = ConveyorBeltCommon;
  cbcClass._addBubbleEventListener(nextButton, "click", function (event) {self._scrollNext();});
  
  if (icon)
  {
    nextButton.appendChild(icon); // @HtmlUpdateOK
  }
  
  var elem = this._elem;
  elem.appendChild(nextButton); // @HtmlUpdateOK
};

/**
 * Get the content parent.
 * @return {Object} parent DOM element of the content
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._getContentParent = function()
{
  //if an explicit content parent was not specified, it will be the _contentContainer
  var contentParent = this._contentParent;
  if (!contentParent)
  {
    contentParent = this._contentContainer;
  }
  return contentParent;
};

/**
 * Measure the contents of the conveyor.
 * @return {Object} Total size of the contents
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._measureContents = function()
{
  var contentParent = this._getContentParent();
  var arContentElements = this._getContentElements();
  var totalSize = {w: 0, h: 0};
  var sizes = [];
  if (contentParent.hasChildNodes() && arContentElements &&
      arContentElements.length > 0)
  {
    var children = arContentElements;
    var bHoriz = this._isHorizontal();
    var contentWidth = 0;

    //get the width of the contentContainer, not the contentParent, because
    //in JET, if the children are in a buttonset, for example, the offsetLeft
    //of the children is relative to the contentContainer, not the 
    //contentParent
    var contentContainer = this._contentContainer;
    contentWidth = contentContainer.offsetWidth;

    var startOffset = 0;
    var prevSizeObj = null;
    for (var i = 0; i < children.length; i++)
    {
      var child = children[i];
      if (child.nodeType === 1)
      {
        var ww = child.offsetWidth;
        var hh = child.offsetHeight;
        var childId = child.id;
        var sizeObj = {w: ww, h: hh, id: childId};
        //calculating the start assumes that the browser has done the appropriate layout;
        //subtract 1 from the end so it's the last pixel of this child, not the start of the next child
        if (bHoriz)
        {
          //FIX : in IE, the conveyor items all report offsetLeft=0,
          //so we need to get the offset from the parent wrapping table cell div
          //instead
          var offLeft = child.offsetLeft;
          if (!this._contentParent && offLeft === 0)
          {
            var childParent = child.parentNode;
            offLeft = childParent.offsetLeft;
          }
          
          //if RTL, still want to save the start coords in logical, ascending order beginning with 0
          if (this._bRtl)
          {
            sizeObj.start = contentWidth - (offLeft + ww);
          }
          else
          {
            sizeObj.start = offLeft;
          }
          
          //Offset each item's start coord by the first item's offset to handle cases like text-align:right,
          //where the items may be right-aligned within the content container.  We still want our logical 
          //coords to start at 0.
          if (i === 0)
            startOffset = sizeObj.start;
          sizeObj.start -= startOffset;
          
          totalSize.w = sizeObj.start + ww;
          totalSize.h = Math.max(totalSize.h, hh);
          sizeObj.end = totalSize.w - 1;
        }
        else
        {
          //FIX : in IE, the conveyor items all report offsetTop=0,
          //so we need to get the offset from the parent wrapping table cell div
          //instead
          var offTop = child.offsetTop;
          if (!this._contentParent && offTop === 0)
          {
            var childParent = child.parentNode;
            offTop = childParent.offsetTop;
          }
          
          sizeObj.start = offTop;
          totalSize.w = Math.max(totalSize.w, ww);
          totalSize.h = sizeObj.start + hh;
          sizeObj.end = totalSize.h - 1;
        }
        
        //if this item overlaps the previous item, adjust the previous item to
        //end just before this item (can happen, for example, with horizontal
        //JET buttonsets)
        if (prevSizeObj)
        {
          if (prevSizeObj.end >= sizeObj.start)
          {
            var overlap = prevSizeObj.end - (sizeObj.start - 1);
            prevSizeObj.end -= overlap;
            if (bHoriz)
            {
              prevSizeObj.w -= overlap;
            }
            else
            {
              prevSizeObj.h -= overlap;
            }
          }
        }
        
        sizes.push(sizeObj);
        prevSizeObj = sizeObj;
      }
    }
  }
  this._sizes = sizes;
  return totalSize;
};

/**
 * Get the array of content sizes.
 * @return {Array} Array of content sizes
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._getSizes = function()
{
  //only calculate sizes if they haven't been calculated already
  if (!this._sizes)
  {
    var totalSize = this._measureContents();
    //only save the total size if it hasn't already been saved
    if (!this._totalSize)
    {
      this._totalSize = totalSize;
    }
  }
  return this._sizes;
};

/**
 * Show the next button.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._showNextButton = function()
{
  this._removeStyleClassNameFunc(this._nextButton, this._hiddenStyleClass);
};

/**
 * Show the prev button.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._showPrevButton = function()
{
  this._removeStyleClassNameFunc(this._prevButton, this._hiddenStyleClass);
};

/**
 * Hide the next button.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._hideNextButton = function()
{
  this._addStyleClassNameFunc(this._nextButton, this._hiddenStyleClass);
};

/**
 * Hide the prev button.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._hidePrevButton = function()
{
  this._addStyleClassNameFunc(this._prevButton, this._hiddenStyleClass);
};

/**
 * Determine if the next button is shown.
 * @return {boolean} true if shown, false if hidden
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._isNextButtonShown = function()
{
  return !this._hasStyleClassNameFunc(this._nextButton, this._hiddenStyleClass);
};

/**
 * Determine if the prev button is shown.
 * @return {boolean} True if shown, false if hidden
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._isPrevButtonShown = function()
{
  return !this._hasStyleClassNameFunc(this._prevButton, this._hiddenStyleClass);
};

/**
 * Get the size of a next/prev button along the direction of conveyor orientation.
 * @return {number} Size of a button
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._getButtonSize = function()
{
  return this._isHorizontal() ? this._buttonWidth : this._buttonHeight;
};

/**
 * Update visibility of the next/prev buttons and adjust scroll position accordingly.
 * @param {number} scroll Desired scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._updateButtonVisibility = function(scroll)
{
  var buttonSize = this._getButtonSize();
  var ovScroll = this._getCurrScroll();
  var ovSize = this._getCurrViewportSize();
  var bNeedsScroll = this._needsScroll();
  //if scrolling to the start, hide the prev button and reclaim its space
  if (scroll <= this._minScroll)
  {
    if (this._isPrevButtonShown())
    {
      ovSize += buttonSize;
      ovScroll -= buttonSize;
    }
    this._hidePrevButton();
  }
  //if not at the start, show the prev button and allocate space for it
  else if (bNeedsScroll)
  {
    if (!this._isPrevButtonShown())
    {
      ovSize -= buttonSize;
      ovScroll += buttonSize;
    }
    this._showPrevButton();
  }

  //if scrolling to the end, hide the next button and reclaim its space
  if (scroll >= this._maxScroll)
  {
    if (this._isNextButtonShown())
    {
      ovSize += buttonSize;
    }
    this._hideNextButton();
  }
  //if not at the end, show the next button and allocate space for it
  else if (bNeedsScroll)
  {
    if (!this._isNextButtonShown())
    {
      ovSize -= buttonSize;
    }
    this._showNextButton();
  }
  //update the overflow container
  this._setOverflowScroll(ovScroll);
};

/**
 * Set the overflow scroll position.
 * @param {number} scroll Overflow logical scroll position
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._setOverflowScroll = function(scroll)
{
  var container = this._overflowContainer;
  if (this._isHorizontal())
  {
    container.scrollLeft = this._convertScrollLogicalToBrowser(scroll);
  }
  else
  {
    container.scrollTop = scroll;
  }
};

/**
 * Get the current overflow viewport size.
 * @return {number} Overflow viewport size
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._getCurrViewportSize = function()
{
  var container = this._overflowContainer;
  return this._isHorizontal() ? container.offsetWidth : container.offsetHeight;
};

/**
 * Set the scroll position.
 * @param {number} scroll Desired logical scroll position
 * @param {boolean} bImmediate True to make the change immediately, false to animate it
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._setCurrScroll = function(scroll, bImmediate)
{
  //don't do anything if we're already in the middle of scrolling
  if (!this._bScrolling)
  {
    //if this function is called, the conveyor internally initiated the scroll, so turn off the 
    //flag for an externally triggered scroll
    this._bExternalScroll = false;
    this._setCurrScrollHelper(scroll, bImmediate);
  }
};

/**
 * Helper function to set scroll position.
 * @param {number} scroll Desired scroll position
 * @param {boolean} bImmediate True to make the change immediately, false to animate it
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._setCurrScrollHelper = function(scroll, bImmediate)
{
  if (this._isEmpty())
    return;
  
  this._bScrolling = true;
  scroll = this._constrainScroll(scroll);
  //update button visibility before scrolling
  this._updateButtonVisibility(scroll);
  var scrollFunc = this._scrollFunc;
  //if making the change immediately, simply call the anim end function
  if (bImmediate || !scrollFunc || scroll === this._getCurrScroll())
  {
    //FIX : if this is an external scroll, set the current scroll
    //value again because it may be different from the passed in scroll value
    //due to showing/hiding scroll buttons (Note that for an external scroll,
    //the scroll has already happened before this function was called)
    this._onScrollAnimEnd(this._bExternalScroll ? this._getCurrScroll() : scroll);
  }
  //if animating the change, call out to the provided callback
  else
  {
    //FIX : add busy state before animating a scroll
    if (this._addBusyStateFunc)
      this._busyStateResolveFunc = this._addBusyStateFunc("scrolling");
    
    var cbcClass = ConveyorBeltCommon;
    // 1.1 px/ms is the desired animation speed, so calculate the duration based on the distance to scroll
    var duration = Math.abs(this._getCurrScroll() - scroll) / cbcClass._SCROLL_SPEED;
    var self = this;
    var onEndFunc = function ()
    {
      self._onScrollAnimEnd(scroll);
      
      //FIX : resolve busy state after animating a scroll
      self._resolveBusyState();
    };
    //need to convert the logical scroll to the browser value for animating
    scrollFunc.call(this._callbackObj, this._overflowContainer, this._convertScrollLogicalToBrowser(scroll), 
                    duration, onEndFunc);
  }
};

/**
 * Resolve an outstanding busy state.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._resolveBusyState = function()
{
  if (this._busyStateResolveFunc)
  {
    this._busyStateResolveFunc();
    this._busyStateResolveFunc = null;
  }
};

/**
 * Get the current scroll position.
 * @return {number} Logical scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._getCurrScroll = function()
{
  var container = this._overflowContainer;
  return this._isHorizontal() ? this._convertScrollBrowserToLogical(container.scrollLeft) : 
                                container.scrollTop;
};

/**
 * Determine if the conveyor needs to show scroll buttons.
 * @return {boolean} True if scrolling is needed, false if not
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._needsScroll = function()
{
  var contentContainer = this._contentContainer;
  var overflowContainer = this._overflowContainer;
  return this._isHorizontal() ? 
    contentContainer.offsetWidth > overflowContainer.offsetWidth : 
    contentContainer.offsetHeight > overflowContainer.offsetHeight;
};

/**
 * Constrain the scroll position.
 * @param {number} scroll Desired scroll position
 * @return {number} Constrained scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._constrainScroll = function(scroll)
{
  if (!this._needsScroll() || scroll < this._minScroll)
  {
    scroll = this._minScroll;
  }
  else if (scroll > this._maxScroll)
  {
    scroll = this._maxScroll;
  }
  return scroll;
};

/**
 * Handle a mousewheel event.
 * @param {Event} event Event object
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._handleMouseWheel = function(event)
{
  //if we're already scrolling, just consume the event
  var bConsumeEvent = this._bScrolling;
  if (this._needsScroll() && !this._bScrolling)
  {
    var cbcClass = ConveyorBeltCommon;
    var wheelDelta = cbcClass._getWheelDelta(event);
    if (wheelDelta < 0 && this._isNextButtonShown())
    {
      bConsumeEvent = true;
      this._scrollNext();
    }
    else if (wheelDelta > 0 && this._isPrevButtonShown())
    {
      bConsumeEvent = true;
      this._scrollPrev();
    }
  }
  if (bConsumeEvent)
  {
    event.preventDefault();
    event.stopPropagation();
  }
};

/**
 * Handle a touchstart event.
 * @param {Event} event Event object
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._handleTouchStart = function(event)
{
  var eventTouches = event.touches;
  if (this._needsScroll() && !this._bScrolling && eventTouches.length === 1)
  {
    this._bTouch = true;
    //save off some initial information at the start of a swipe
    var firstTouch = eventTouches[0];
    this._touchStartCoord = this._isHorizontal() ? firstTouch.pageX : firstTouch.pageY;
    this._touchStartScroll = this._getCurrScroll();
    this._touchStartNextScroll = this._calcNextScroll();
    this._touchStartPrevScroll = this._calcPrevScroll();
    //FIX : save the initial button state
    this._touchStartNextButtonShown = this._isNextButtonShown();
    this._touchStartPrevButtonShown = this._isPrevButtonShown();
  }
};

/**
 * Handle a touchmove event.
 * @param {Event} event Event object
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._handleTouchMove = function(event)
{
  var bHoriz = this._isHorizontal();
  var eventTouches = event.touches;
  var firstTouch = eventTouches[0];
  var touchCoord = bHoriz ? firstTouch.pageX : firstTouch.pageY;
  var diff = touchCoord - this._touchStartCoord;
  //in non-RTL, if swiping left or up, scroll next; otherwise scroll prev
  //in RTL, if swiping right or up, scroll next; otherwise scroll prev
  var bNext = (bHoriz && this._bRtl) ? (diff > 0) : (diff < 0);
  //determine whether the conveyor can be scrolled in the direction of the swipe
  var canScrollInSwipeDirection = (bNext && this._touchStartNextButtonShown) ||
                                  (!bNext && this._touchStartPrevButtonShown);
  //only need to do something if we also received the touchstart and if we can
  //scroll in the swipe direction
  if (this._bTouch && canScrollInSwipeDirection)
  {
    //only scroll next/prev if the swipe is longer than the threshold; if it's
    //less, then just drag the items with the swipe
    var cbcClass = ConveyorBeltCommon;
    var container = this._overflowContainer;
    var threshold = cbcClass._SWIPE_THRESHOLD * 
                    (bHoriz ? container.offsetWidth : container.offsetHeight);
    
    //if swiping under the threshold, just move the conveyor with the swipe
    if (Math.abs(diff) < threshold)
    {
      this._setCurrScroll(this._touchStartScroll - diff, true);
      
      //if we're under the threshold, but we've already scrolled to the end,
      //then we don't need to continue trying to scroll and we don't need to
      //reset the scroll position at the end of the touch
      if ((this._touchStartNextButtonShown && !this._isNextButtonShown()) ||
          (this._touchStartPrevButtonShown && !this._isPrevButtonShown()))
      {
        this._bTouch = false;
      }
    }
    //if swiping beyond the threshold, scroll to the next/prev set of items
    else
    {
      this._setCurrScroll(bNext ? this._touchStartNextScroll : this._touchStartPrevScroll, false);
      //don't scroll again for this same swipe
      this._bTouch = false;
    }
    
    //FIX : set a flag indicating we've scrolled for this touch event
    this._scrolledForThisTouch = true;
  }
  
  //FIX : if we've scrolled for this touch event, consume the event
  //so that the page doesn't also scroll
  if (this._scrolledForThisTouch)
  {
    event.preventDefault();
    event.stopPropagation();
  }
};

/**
 * Handle a touchend event.
 * @param {Event} event Event object
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._handleTouchEnd = function(event)
{
  //if a full page swipe hasn't happened, scroll back to the original position
  if (this._bTouch)
  {
    this._setCurrScroll(this._touchStartScroll, false);
  }
  this._bTouch = false;
  //FIX : reset the flag indicating if we've scrolled for this touch
  //event
  this._scrolledForThisTouch = false;
};

/**
 * Handle a DOM scroll event.
 * @param {Event} event Event object
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._handleScroll = function(event)
{
  //if the scroll was triggered externally, for example by tabbing through 
  //child items, then update the visual state of the conveyor to match the
  //new scroll state
  if (this._bExternalScroll && !this._bScrolling)
  {
    this._setCurrScrollHelper(this._getCurrScroll(), true);
  }
};

/**
 * Function called after a scroll finishes.
 * @param {number} scroll Scroll position
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._onScrollAnimEnd = function(scroll)
{
  //set the desired value after the animation to make sure that the final value is exactly what was intended,
  //in case the animation introduced interpolation errors
  this._setOverflowScroll(scroll);
  this._bExternalScroll = true;
  this._bScrolling = false;
  //if there is a callback to notify about a change of first visible item, call it now
  if (this._firstVisibleItemChangedFunc)
  {
    this._firstVisibleItemIndex = this._calcFirstVisibleItemIndex();
    //if there is more than one item visible, and the first item is only 
    //partially visible, then save the second item as the actual first visible 
    //item for state purposes (if the last navigation had been prev,
    //we would want the last visible item to remain visible)
    var lastVisIndex = this._calcLastVisibleItemIndex();
    var sizes = this._getSizes();
    var sizeObj = sizes[this._firstVisibleItemIndex];
    if (this._firstVisibleItemIndex !== lastVisIndex &&
        this._getCurrScroll() > sizeObj.start &&
        this._firstVisibleItemIndex < sizes.length - 2)
    {
      this._firstVisibleItemIndex++;
      sizeObj = sizes[this._firstVisibleItemIndex];
    }
    this._firstVisibleItemId = sizeObj.id;
    var firstVisibleItemChangedFunc = this._firstVisibleItemChangedFunc;
    firstVisibleItemChangedFunc.call(this._callbackObj, this._firstVisibleItemId);
  }
};

/**
 * Set a timout to reset the externalScroll flag.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._setExternalScrollTimeout = function()
{
  var self = this;
  var f = function () {if (self) self._bExternalScroll = true;};
  window.setTimeout(f, 0); // @HtmlUpdateOK
};

/**
 * Scroll to the next set of items.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._scrollNext = function()
{
  if (!this._bScrolling)
    this._setCurrScroll(this._calcNextScroll(), false);
};

/**
 * Scroll to the previous set of items.
 * @return {void}
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._scrollPrev = function()
{
  if (!this._bScrolling)
    this._setCurrScroll(this._calcPrevScroll(), false);
};

/**
 * Calculate the scroll position for the next set of items.
 * @return {number} Next scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcNextScroll = function()
{
  var nextIndex = this._calcNextVisibleItemIndex();
  var scroll = 0;
  //if single item is bigger than viewport, then scroll by viewport size
  if (nextIndex === this._calcFirstVisibleItemIndex())
  {
    scroll = this._getCurrScroll() + this._getCurrViewportSize();
  }
  else
  {
    scroll = this._calcStartScroll(nextIndex);
  }
  return scroll;
};

/**
 * Calculate the scroll position for the previous set of items.
 * @return {number} Previous scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcPrevScroll = function()
{
  var prevIndex = this._calcPrevVisibleItemIndex();
  var scroll = 0;
  //if single item is bigger than viewport, then scroll by viewport size
  if (prevIndex === this._calcLastVisibleItemIndex())
  {
    scroll = this._getCurrScroll() - this._getCurrViewportSize();
  }
  else
  {
    scroll = this._calcEndScroll(prevIndex);
  }
  //if at the end and scrolling prev, anticipate the next button becoming 
  //visible and adjust the scroll position
  if (!this._isNextButtonShown())
  {
    scroll += this._getButtonSize();
  }
  //if scrolling prev and the scroll position is less than the size of the prev button, just
  //scroll to the very beginning because the prev button should get hidden
  if (scroll < this._getButtonSize())
  {
    scroll = this._minScroll;
  }
  return scroll;
};

/**
 * Calculate the scroll position for the start of the specified item.
 * @param {number} index Index of the item to scroll to
 * @return {number} Scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcStartScroll = function(index)
{
  var sizes = this._getSizes();
  var sizeObj = sizes[index];
  return sizeObj.start;
};

/**
 * Calculate the scroll position for the end of the specified item.
 * @param {number} index Index of the item to scroll to
 * @return {number} Scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcEndScroll = function(index)
{
  var sizes = this._getSizes();
  var sizeObj = sizes[index];
  return sizeObj.end - this._getCurrViewportSize() + 1;
};

/**
 * Calculate the index of the first visible item.
 * @return {number} Index of first visible item
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcFirstVisibleItemIndex = function()
{
  var currScroll = this._getCurrScroll();
  var i = this._calcItemIndex(currScroll);
  return (i < 0) ? 0 : i;
};

/**
 * Calculate the index of the last visible item.
 * @return {number} Index of last visible item
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcLastVisibleItemIndex = function()
{
  var elemSize = this._getCurrViewportSize();
  var currScroll = this._getCurrScroll() + elemSize - 1;
  var i = this._calcItemIndex(currScroll);
  var sizes = this._getSizes();
  return (i < 0) ? sizes.length - 1 : i;
};

/**
 * Calculate the index of the previous visible item.
 * @return {number} Index of previous visible item
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcPrevVisibleItemIndex = function()
{
  var currScroll = this._getCurrScroll() - 1;
  var i = this._calcItemIndex(currScroll);
  return (i < 0) ? 0 : i;
};

/**
 * Calculate the index of the next visible item.
 * @return {number} Index of next visible item
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcNextVisibleItemIndex = function()
{
  var elemSize = this._getCurrViewportSize();
  var currScroll = this._getCurrScroll() + elemSize;
  var i = this._calcItemIndex(currScroll);
  var sizes = this._getSizes();
  return (i < 0) ? sizes.length - 1 : i;
};

/**
 * Calculate the index of the item at the given scroll position.
 * @param {number} scroll Scroll position
 * @return {number} Index of item at given scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._calcItemIndex = function(scroll)
{
  var sizes = this._getSizes();
  for (var i = 0; i < sizes.length; i++)
  {
    var sizeObj = sizes[i];
    if (scroll <= sizeObj.end)
      return i;
  }
  return -1;
};

/**
 * Convert a logical scroll position to its corresponding browser value.
 * @param {number} scroll logical scroll position
 * @return {number} browser scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._convertScrollLogicalToBrowser = function(scroll)
{
  //(comment mostly copied from AdfConveyorBeltSupport)
  //If this is LTR or RTL mode in IE, then we want the default positive new scroll value.
  //If FF in RTL, then get the negative scroll value
  //If Safari version 10+ in RTL, then get the negative scroll value
  //If Webkit in RTL, to scroll to a position, we resolve this equation:
  // contentContainerWidth - browserScroll = overflowContainerWidth + logicalScroll
  // browserScroll = contentContainerWidth = overflowContainerWidth - logicalScroll
  var newScroll = scroll;
  if (this._bRtl && this._isHorizontal())
  {
    //Safari version 10+ has the correct scroll offset in RTL mode
    //So don't resolve the scroll equation for Safari version 10+
    if (this._bAgentGecko || (this._bAgentSafari && this._agentVersion >= 10)) {
      newScroll = -scroll;
    }
    else if (this._bAgentWebkit || this._bAgentOpera || 
              (this._bAgentSafari && this._agentVersion < 10)) {
      var contentContainer = this._contentContainer;
      var overflowContainer = this._overflowContainer;
      newScroll = contentContainer.offsetWidth - overflowContainer.offsetWidth - scroll;
    }
  }
  return newScroll;
};

/**
 * Convert a browser scroll position to its corresponding logical value.
 * @param {number} scroll browser scroll position
 * @return {number} logical scroll position
 * @memberof ConveyorBeltCommon
 * @instance
 * @private
 */
ConveyorBeltCommon.prototype._convertScrollBrowserToLogical = function(scroll)
{
  //(comment mostly copied from AdfConveyorBeltSupport)
  //If this is LTR or RTL mode in IE, then we want the default positive new scroll value.
  //If FF in RTL, then get the negative scroll value
  //If Webkit in RTL, to scroll to a position, we resolve this equation:
  // contentContainerWidth - browserScroll = overflowContainerWidth + logicalScroll
  // browserScroll = contentContainerWidth = overflowContainerWidth - logicalScroll
  
  //because the equations are the same whether converting from browser -> logical or logical -> browser,
  //simply call _convertScrollLogicalToBrowser from here
  //(NOTE: want to leave _convertScrollBrowserToLogical as a separate function so that it's clear from the
  //calling code which conversion direction is used, and in case the conversion impls ever need to be changed)
  return this._convertScrollLogicalToBrowser(scroll);
};

/**
 * Scroll animation speed (px/ms).
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._SCROLL_SPEED = 1.1;
/**
 * Touch swipe threshold (percentage of conveyor size).
 * @memberof ConveyorBeltCommon
 * @private
 */
ConveyorBeltCommon._SWIPE_THRESHOLD = .33;
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojConveyorBelt
 * @augments oj.baseComponent
 * @since 0.6
 * @ojstatus preview
 * @ojshortdesc Container element that manages overflow for its child elements and allows scrolling among them.
 * @class oj.ojConveyorBelt
 *
 * @classdesc
 * <h3 id="conveyorBeltOverview-section">
 *   JET ConveyorBelt
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#conveyorBeltOverview-section"></a>
 * </h3>
 * 
 * <p>Description: Container element that manages overflow for its child
 * elements and allows scrolling among them.
 * 
 * <p>Child elements of the ConveyorBelt must all be siblings at the same level.
 * The size of the ConveyorBelt must somehow be constrained in order for 
 * there to be overflow to manage, for example by specifying CSS 
 * <code class="prettyprint">max-width</code> or
 * <code class="prettyprint">max-height</code>.  
 * <p>If the elements to be scrolled among are direct children of the
 * ConveyorBelt, then ConveyorBelt will ensure that they are laid out
 * appropriately for its orientation.  However, if the elements to be scrolled
 * among are contained by a single nested descendant element, the 
 * <code class="prettyprint">content-parent</code>, then it is up to calling code
 * to ensure that the elements are laid out appropriately.  For example, 
 * elements can be forced horizontal by using CSS
 * <code class="prettyprint">white-space:nowrap</code>, or vertical by using 
 * <code class="prettyprint">display:block</code>.
 * 
 * <pre class="prettyprint"><code>
 * &lt;oj-conveyor-belt>
 *   &lt;oj-button>Alpha&lt;/oj-button>
 *   &lt;oj-button>Beta&lt;/oj-button>
 *   &lt;oj-button>Gamma&lt;/oj-button>
 *   &lt;oj-button>Delta&lt;/oj-button>
 *   &lt;oj-button>Epsilon&lt;/oj-button>
 *   &lt;oj-button>Zeta&lt;/oj-button>
 * &lt;/oj-conveyor-belt>
 * </code></pre>
 *
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 * 
 * 
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 * 
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * 
 * <h3 id="keyboard-appdev-section">
 *   Keyboard Application Developer Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-appdev-section"></a>
 * </h3>
 * 
 * <p>Providing keyboard support for the items in the conveyor belt is the 
 * responsibility of the application developer, if the items do not already
 * support keyboard interaction.  This could be done, for example, by specifying 
 * <code class="prettyprint">tabindex</code> on each item to enable tab
 * navigation.  Alternatively, this could be done by adding a keyboard listener
 * and responding to key events, like pressing the arrow keys.
 * 
 * 
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>ConveyorBelt itself does nothing special for accessibility.  
 * It is the responsibility of the application developer to make the items in 
 * the conveyor accessible.  Sighted keyboard-only users need to be able to 
 * access the items in the conveyor just by using the keyboard.
 * It is up to the child items of the ConveyorBelt to support keyboard 
 * navigation.  If child items support tab navigation, the browser may scroll 
 * them into view when they receive focus.  If child items support other forms 
 * of keyboard navigation, for example by using the arrow keys, it is up to the 
 * child items to scroll themselves into view.  This may be done, for example, 
 * by calling the DOM function <code class="prettyprint">focus()</code> or 
 * <code class="prettyprint">scrollIntoView()</code> on the item.  
 * ConveyorBelt will be aware of tab based or programmatic scrolling and will 
 * honor it, updating itself to toggle visibility of the overflow indicators as 
 * needed.  
 * 
 * 
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 * 
 * <p>As with any JET element, in the unusual case that the directionality 
 * (LTR or RTL) changes post-init, the conveyorBelt must be 
 * <code class="prettyprint">refresh()</code>ed.
 */
/**
 * <p>The &lt;oj-conveyor-belt> element manages overflow for its child elements and allows scrolling among them.
 * Child elements of the &lt;oj-conveyor-belt> must all be siblings at the same level.</p>
 * <p>If the elements to be scrolled among are nested descendants and not direct
 * children of the conveyor belt, the <code class="prettyprint">content-parent</code> 
 * attribute should specify the nested elements direct parent.</p>
 *
 * @ojchild Default
 * @memberof oj.ojConveyorBelt
 *
 * @example <caption>Initialize the conveyor belt with child content specified:</caption>
 * &lt;oj-conveyor-belt>
 *   &lt;oj-button>Alpha&lt;/oj-button>
 *   &lt;oj-button>Beta&lt;/oj-button>
 *   &lt;oj-button>Gamma&lt;/oj-button>
 *   &lt;oj-button>Delta&lt;/oj-button>
 *   &lt;oj-button>Epsilon&lt;/oj-button>
 *   &lt;oj-button>Zeta&lt;/oj-button>
 * &lt;/oj-conveyor-belt>
 *
 * @example <caption>Initialize the conveyor belt with nested child content
 * specified:</caption>
 * &lt;oj-conveyor-belt content-parent='#myContentElem'>
 *   &lt;div id='myContentElem'>
 *     &lt;oj-button>Item 1&lt;/oj-button>
 *     &lt;oj-button>Item 2&lt;/oj-button>
 *     &lt;oj-button>Item 3&lt;/oj-button>
 *     &lt;oj-button>Item 4&lt;/oj-button>
 *     &lt;oj-button>Item 5&lt;/oj-button>
 *   &lt;/div>
 * &lt;/oj-conveyor-belt>
 */
(function()
{
oj.__registerWidget("oj.ojConveyorBelt", $['oj']['baseComponent'],
{
  defaultElement: "<div>",
  widgetEventPrefix: "oj",

  options:
  {
    /** 
     * Specify the orientation of the conveyorBelt.
     *
     * @expose 
     * @memberof oj.ojConveyorBelt
     * @instance
     * @type {string}
     * @ojvalue {string} "horizontal" Orient the conveyorBelt horizontally.
     * @ojvalue {string} "vertical" Orient the conveyorBelt vertically.
     * @default "horizontal"
     * @ojshortdesc Specifies the orientation of the conveyorBelt.
     *
     * @example <caption>Initialize the conveyorBelt with the 
     * <code class="prettyprint">orientation</code> attribute specified:</caption>
     * &lt;oj-conveyor-belt orientation='vertical'>
     * &lt;/oj-conveyor-belt>
     * 
     * @example <caption>Get or set the <code class="prettyprint">orientation</code> 
     * property after initialization:</caption>
     * // getter
     * var orientation = myConveyorBelt.orientation;
     * 
     * // setter
     * myConveyorBelt.orientation = 'vertical';
     */
    orientation: "horizontal",
    /** 
     * Specify the selector of the descendant DOM element in the conveyorBelt
     * that directly contains the items to scroll among.  
     * 
     * <p>This attribute value is <code class="prettyprint">null</code> by default, 
     * meaning that the items to scroll among are direct children of the 
     * oj-conveyor-belt.  In some cases, the items to scroll among
     * are not direct children of the oj-conveyor-belt, but are instead
     * nested in a descendant DOM element.  In such cases, this attribute should be 
     * specified to point to the descendant DOM element whose direct children
     * are the items to scroll among.  For example, if the items to scroll
     * among are buttons in a buttonset, the buttons are direct children of 
     * the DOM element representing the buttonset.  The buttonset would be 
     * the direct child of the conveyorBelt.  If the 
     * <code class="prettyprint">id</code> of the buttonset DOM element were
     * <code class="prettyprint">'myContentElem'</code>, then content-parent
     * would be specified as <code class="prettyprint">'#myContentElem'</code>.
     * 
     * <p><b>WARNING:</b> The selector specified for this attribute should match 
     * only a single descendant DOM element.  If multiple elements are matched, 
     * then only the first one will be used.  Applications should not depend on 
     * this behavior because we reserve the right to change it in the future in 
     * order to allow and use multiple matching elements.  
     *
     * @expose 
     * @memberof oj.ojConveyorBelt
     * @instance
     * @type {?string}
     * @default null
     * @ojshortdesc Specify the selector of the descendant DOM element in the conveyorBelt that directly contains the items to scroll among.
     *
     * @example <caption>Initialize the conveyorBelt with the 
     * <code class="prettyprint">content-parent</code> attribute specified:</caption>
     * &lt;oj-conveyor-belt content-parent='#myContentElem'>
     *   &lt;div id='myContentElem'>
     *     &lt;oj-button>Item 1&lt;/oj-button>
     *     &lt;oj-button>Item 2&lt;/oj-button>
     *     &lt;oj-button>Item 3&lt;/oj-button>
     *     &lt;oj-button>Item 4&lt;/oj-button>
     *     &lt;oj-button>Item 5&lt;/oj-button>
     *   &lt;/div>
     * &lt;/oj-conveyor-belt>
     * 
     * @example <caption>Get or set the <code class="prettyprint">contentParent</code> 
     * property after initialization:</caption>
     * // getter
     * var contentParent = myConveyorBelt.contentParent;
     * 
     * // setter
     * myConveyorBelt.contentParent = '#myContentElem';
     */
    contentParent: null
    
    /**
     * To avoid tight coupling between a ConveyorBelt and its contents, JET 
     * ConveyorBelt does not support the <code class="prettyprint">disabled</code> 
     * attribute.
     * 
     * <p><b>WARNING:</b> Applications should not depend on this behavior 
     * because we reserve the right to change it in the future in order to 
     * support <code class="prettyprint">disabled</code> and propagate it to 
     * child elements of ConveyorBelt.  
     * 
     * @member
     * @name disabled
     * @memberof oj.ojConveyorBelt
     * @instance
     * @type {boolean}
     * @default false
     * @ignore
     */
    // disabled attribute declared in superclass, but we still want the above API doc

    // Events

    /**
     * Triggered when the conveyorBelt is created.
     *
     * @event
     * @name create
     * @memberof oj.ojConveyorBelt
     * @instance
     * @ignore
     */
    // create event declared in superclass, but we still want the above API doc
  },

  /**
   * After _ComponentCreate and _AfterCreate, 
   * the widget should be 100% set up. this._super should be called first.
   * @return {void}
   * @override
   * @protected
   * @instance
   * @memberof oj.ojConveyorBelt
   */
  _ComponentCreate : function () // Override of protected base class method.  
  {
    //call superclass first
    this._super();
    
    var elem = this.element;  
    elem.addClass("oj-conveyorbelt oj-component");
    
    var options = this.options;
    //FIX : log warning message when "disabled" attribute set
    if (options.disabled)
    {
      oj.Logger.warn(_WARNING_DISABLED_OPTION);
    }
    
    //FIX : remove override of _init() and call _setup() from here
    this._setup(true);
  },

  // This method currently runs at create, init, and refresh time (since refresh() is called by _init()).
  /**
   * Refreshes the visual state of the conveyorBelt. JET elements require a 
   * <code class="prettyprint">refresh()</code> after the DOM is 
   * programmatically changed underneath the element.
   * 
   * <p>This method does not accept any arguments.
   * 
   * @return {void}
   * @expose 
   * @memberof oj.ojConveyorBelt
   * @instance
   * @ojshortdesc Refreshes the visual state of the conveyorBelt.
   * 
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   * myConveyorBelt.refresh();
   */
  refresh: function() // Override of public base class method.  
  {
    this._super();
    
    //Check if the reading direction have changed
    var bRTL = (this._GetReadingDirection() === "rtl");
    var bDirectionChanged = (this._bRTL != bRTL);
    
    //save and restore scroll position only if the reading direction is not changed
    var scroll;
    //save the current scroll position
    if (!bDirectionChanged)
      scroll = this._cbCommon.getScroll();
    
    //destroy the cbCommon and setup from scratch in case items were added/removed
    this._destroyCBCommon();
    this._setup(true);
    
    //restore the saved scroll position
    if (!bDirectionChanged)
      this._cbCommon.setScroll(scroll, true);
  },
  
  /**
   * Notifies the component that its subtree has been made visible 
   * programmatically after the component has been created.
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @protected 
   * @override
   */
  _NotifyShown: function()
  {
    this._super();
    //FIX : perform a deferred layout
    if (this._needsSetup)
    {
      this._setup(this._needsSetup[0]);
    }
    //with internal flexbox layout, conveyor doesn't get notified when
    //content resizes while hidden, so explicitly handle a resize when
    //shown again
    else if (this._cbCommon)
    {
      var cbCommon = this._cbCommon;
      cbCommon.handleResize();
    }
  },
  
  /**
   * Notifies the component that its subtree has been connected to the document
   * programmatically after the component has been created.
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @protected 
   * @override
   */
  _NotifyAttached: function()
  {
    this._super();
    //FIX : perform a deferred layout
    if (this._needsSetup)
    {
      this._setup(this._needsSetup[0]);
    }
    //with internal flexbox layout, conveyor doesn't get notified when
    //content resizes while detached, so explicitly handle a resize when
    //attached again
    else if (this._cbCommon)
    {
      var cbCommon = this._cbCommon;
      cbCommon.handleResize();
    }
  },

  // isInit is true for init (create and re-init), false for refresh
  /** 
   * Setup the conveyorBelt.
   * @param {boolean} isInit true if _setup is called from _init(), false
   *        if called from refresh()
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _setup: function(isInit) // Private, not an override (not in base class).  
  {
    var self = this;
    var elem = this.element;
    var options = this.options;
    var orientation = options.orientation;
    if (orientation === "vertical")
      elem.addClass("oj-conveyorbelt-vertical");
    else
      elem.removeClass("oj-conveyorbelt-vertical");
    
    //FIX : if conveyor is detached or hidden, we can't layout
    //correctly, so defer layout until conveyor is attached or shown
    if (!this._canCalculateSizes())
    {
      //want a true value of isInit to take precedence over a false value
      var oldIsInit = false;
      if (this._needsSetup)
      {
        oldIsInit = this._needsSetup[0];
      }
      this._needsSetup = [isInit || oldIsInit];
      return;
    }
    this._needsSetup = null;
    
    this._bRTL = (this._GetReadingDirection() === "rtl");
    if (isInit)
    {
      if (!this._cbCommon)
      {
        var prevStyleClass = null;
        var nextStyleClass = null;
        var prevIcon = null;
        var nextIcon = null;
        var animateScrollFunc = null;
        if (orientation !== "vertical")
        {
          prevStyleClass = "oj-enabled oj-conveyorbelt-overflow-indicator oj-start oj-default";
          nextStyleClass = "oj-enabled oj-conveyorbelt-overflow-indicator oj-end oj-default";
          prevIcon = this._createIcon("oj-conveyorbelt-overflow-icon oj-start");
          nextIcon = this._createIcon("oj-conveyorbelt-overflow-icon oj-end");
          animateScrollFunc = this._animateScrollLeft;
        }
        else
        {
          prevStyleClass = "oj-enabled oj-conveyorbelt-overflow-indicator oj-top oj-default";
          nextStyleClass = "oj-enabled oj-conveyorbelt-overflow-indicator oj-bottom oj-default";
          prevIcon = this._createIcon("oj-conveyorbelt-overflow-icon oj-top");
          nextIcon = this._createIcon("oj-conveyorbelt-overflow-icon oj-bottom");
          animateScrollFunc = this._animateScrollTop;
        }
        var buttonInfo = {};
        buttonInfo.prevButtonStyleClass = prevStyleClass;
        buttonInfo.nextButtonStyleClass = nextStyleClass;
        buttonInfo.prevButtonIcon = prevIcon;
        buttonInfo.nextButtonIcon = nextIcon;
        var styleInfo = {};
        styleInfo.overflowContainerStyleClass = "oj-conveyorbelt-overflow-container";
        styleInfo.contentContainerStyleClass = "oj-conveyorbelt-content-container";
        styleInfo.itemStyleClass = "oj-conveyorbelt-item";
        styleInfo.hiddenStyleClass = "oj-helper-hidden";
        var callbackInfo = {};
        callbackInfo.addResizeListener = function(elem, listener) {
          oj.DomUtils.addResizeListener(elem, listener, _RESIZE_LISTENER_COLLAPSE_EVENT_TIMEOUT);
        };
        callbackInfo.removeResizeListener = oj.DomUtils.removeResizeListener;
        callbackInfo.addStyleClassName = this._addStyleClassName;
        callbackInfo.removeStyleClassName = this._removeStyleClassName;
        callbackInfo.hasStyleClassName = this._hasStyleClassName;
        callbackInfo.filterContentElements = 
          function(arContentElements) { return self._filterContentElements(arContentElements); };
        callbackInfo.subtreeDetached = oj.Components.subtreeDetached;
        callbackInfo.subtreeAttached = oj.Components.subtreeAttached;
        callbackInfo.addBusyState = 
          function(description) { return self._addBusyState(description); };
        //disable scroll animation during testing
        if (oj.Config.getAutomationMode() !== "enabled")
        {
          callbackInfo.scrollFunc = animateScrollFunc;
        }
        var contentParentElem = null;
        if (options.contentParent)
        {
          //only use the first result returned from the contentParent selector
          contentParentElem = $(options.contentParent)[0];
        }
        var agentInfo = oj.AgentUtils.getAgentInfo(navigator.userAgent);
        this._cbCommon = new ConveyorBeltCommon(
            elem[0],
            orientation, 
            contentParentElem,
            this._bRTL, 
            buttonInfo, 
            callbackInfo, 
            styleInfo,
            agentInfo);
      }
    }
    var cbCommon = this._cbCommon;
    cbCommon.setup(isInit);
    if (isInit)
    {
      var children = elem.find(".oj-conveyorbelt-overflow-indicator");
      //need to setup listeners for styles on each button individually so that callbacks for that
      //listener are specific to each button and don't affect both buttons at the same time
      for (var i = 0; i < children.length; i++)
      {
        this._setupButtonMouseStyles($(children[i]));
      }
    }
  },

  /** 
   * Destroy the conveyorBelt.
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @override
   * @protected
   */
  _destroy: function() // Override of protected base class method.  
  {
    this._destroyCBCommon();
    var elem = this.element;
    elem.removeClass("oj-conveyorbelt oj-component oj-conveyorbelt-vertical");
    
    //call superclass last
    this._super();
  },

  /** 
   * Set an option on the conveyorBelt.
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @override
   * @protected
   */
  _setOption: function(key, value, flags) // Override of protected base class method.
                                   // Method name needn't be quoted since is in externs.js.
  {
    var bRecreate = false;
    var options = this.options;
    switch (key) 
    {
      //when changing containerParent or orientation, just destroy and recreate
      //the ConveyorBeltCommon
      case "containerParent":
      case "orientation":
        bRecreate = (options.orientation != value);
        break;
      case "disabled":
        //FIX : log warning message when "disabled" attribute set
        oj.Logger.warn(_WARNING_DISABLED_OPTION);
        break;
    }
    //if recreating, destroy the ConveyorBeltCommon before calling superclass
    //_setOption
    if (bRecreate)
    {
      this._destroyCBCommon();
    }
    this._super(key, value, flags);
    //if recreating, setup the new ConveyorBeltCommon after calling superclass
    //_setOption
    if (bRecreate)
    {
      this._setup(true);
    }
  },

  /** 
   * Destroy the ConveyorBeltCommon.
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _destroyCBCommon: function()
  {
    var cbCommon = this._cbCommon;
    if (cbCommon)
    {
      //FIX : detach mouse listeners from overflow indicators
      //before destroying cbCommon in order to avoid memory leaks
      var elem = this.element;
      var children = elem.find(".oj-conveyorbelt-overflow-indicator");
      children.off(this.eventNamespace);
      
      cbCommon.destroy();
    }
    this._cbCommon = null;
  },

  /** 
   * Determine whether the conveyorBelt can calculate sizes (when it is 
   * attached to the page DOM and not hidden).
   * @returns {boolean} true if sizes can be calculated, false if not
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _canCalculateSizes: function() 
  {
    var div = document.createElement("div");
    var style = div.style;
    style.width = "10px";
    style.height = "10px";
    
    //make sure flexbox layout doesn't grow/shrink the item
    style["-webkit-flex"] = "0 0 auto";
    style["flex"] = "0 0 auto";
    
    var elem = this.element[0];
    elem.appendChild(div); // @HtmlUpdateOK
    var bCanCalcSizes = false;
    try {
      bCanCalcSizes = div.offsetWidth > 0 && div.offsetHeight > 0;
    }
    catch (e) {
      //do nothing
    }
    elem.removeChild(div);
    return bCanCalcSizes;
  },

  /** 
   * Setup mouse listeners to change button styles.
   * @param {Object} element jQuery element to affect
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _setupButtonMouseStyles: function(element)
  {
    //  - conveyorbelt next/previous oj-hover color don't go away
    this._AddHoverable({
      'element': element,
      'afterToggle': function(eventtype) {
        if (eventtype === "mouseenter") {
          element.removeClass("oj-default");
        }
        else if (eventtype === "mouseleave") {
          element.addClass("oj-default");
        }
      }
    });

    this._AddActiveable({
      'element': element,
      'afterToggle': function(eventtype) {
        if (eventtype === "mousedown" || eventtype === "touchstart" || eventtype === "mouseenter") {
          element.removeClass("oj-default");
        }
        else if (eventtype === "mouseup" || eventtype === "touchend" || 
                 eventtype === "touchcancel" || eventtype === "mouseleave") {
          element.addClass("oj-default");
        }
      }
    });
  },

  /** 
   * Create a DOM element for an icon.
   * @param {string} iconStyleClass Style class for the icon
   * @returns {Element} Icon DOM element
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _createIcon: function(iconStyleClass)
  {
    var span = document.createElement("span");
    span.setAttribute("class", "oj-component-icon " + iconStyleClass);
    return span;
  },

  /** 
   * Animate setting the scrollLeft DOM property.
   * @param {Element} elem DOM element to scroll
   * @param {number} value Scroll value
   * @param {number} duration Duration of animation, in ms
   * @param {function():void} onEndFunc Function to call when the animation ends
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _animateScrollLeft: function(elem, value, duration, onEndFunc)
  {
    var props = {};
    props["scrollLeft"] = value;
    //use swing instead of easeInOutCubic because easeInOutCubic isn't found
    //when running in the cookbook
    $(elem).animate(props, duration, "swing", onEndFunc);
  },

  /** 
   * Animate setting the scrollTop DOM property.
   * @param {Element} elem DOM element to scroll
   * @param {number} value Scroll value
   * @param {number} duration Duration of animation, in ms
   * @param {function():void} onEndFunc Function to call when the animation ends
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _animateScrollTop: function(elem, value, duration, onEndFunc)
  {
    var props = {};
    props["scrollTop"] = value;
    //use swing instead of easeInOutCubic because easeInOutCubic isn't found
    //when running in the cookbook
    $(elem).animate(props, duration, "swing", onEndFunc);
  },

  /** 
   * Add a style class name to an element.
   * @param {Object} elem Element to which to add style class.
   * @param {string} styleClass Style class name to add.
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _addStyleClassName: function(elem, styleClass)
  {
    $(elem).addClass(styleClass);
  },

  /** 
   * Remove a style class name from an element.
   * @param {Object} elem Element from which to remove style class.
   * @param {string} styleClass Style class name to remove.
   * @return {void}
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _removeStyleClassName: function(elem, styleClass)
  {
    $(elem).removeClass(styleClass);
  },
  
  /** 
   * Determine whether the given style class name is applied to the given
   * element.
   * @param {Object} elem Element to check for style class name.
   * @param {string} styleClass Style class name for which to look.
   * @returns {boolean} true if style class name is applied, false if not
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _hasStyleClassName: function(elem, styleClass)
  {
    return $(elem).hasClass(styleClass);
  },
  
  /** 
   * Filter the given array of conveyor content elements to remove extraneous elements, like the
   * divs added by the resize listener.
   * @param {Array} arContentElements Array of conveyor content elements.
   * @returns {Array} filtered array of conveyor content elements
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _filterContentElements: function(arContentElements)
  {
    var ret = [];
    for (var i = 0; i < arContentElements.length; i++)
    {
      var contentElem = arContentElements[i];
      if (!this._hasStyleClassName(contentElem, "oj-helper-detect-expansion") &&
          !this._hasStyleClassName(contentElem, "oj-helper-detect-contraction"))
      {
        ret.push(contentElem);
      }
    }
    return ret;
  },
  
  /**
   * Add a busy state to the busy context.
   *
   * @param {string} description Additional information about busy state.
   * @returns {Function} Resolve function called by the registrant when the busy state completes.
   *          The resultant function will throw an error if the busy state is no longer registered.
   * @memberof oj.ojConveyorBelt
   * @instance
   * @private
   */
  _addBusyState: function(description)
  {
    var element = this.element;
    var context = oj.Context.getContext(element[0]);
    var busyContext = context.getBusyContext();
    
    var desc = "ConveyorBelt";
    var id = element.attr("id");
    desc += " (id='" + id + "')";
    desc += ": " + description;
    
    var busyStateOptions = {"description" : desc};
    return busyContext.addBusyState(busyStateOptions);
  },
  
  // @inheritdoc
  getNodeBySubId: function(locator)
  {
    if (locator == null)
    {
      return this.element ? this.element[0] : null;
    }
    
    var subId = locator['subId'];
    if (subId === "oj-conveyorbelt-start-overflow-indicator") {
      return this.widget().find(".oj-conveyorbelt-overflow-indicator.oj-start")[0];
    }
    if (subId === "oj-conveyorbelt-end-overflow-indicator") {
      return this.widget().find(".oj-conveyorbelt-overflow-indicator.oj-end")[0];
    }
    if (subId === "oj-conveyorbelt-top-overflow-indicator") {
      return this.widget().find(".oj-conveyorbelt-overflow-indicator.oj-top")[0];
    }
    if (subId === "oj-conveyorbelt-bottom-overflow-indicator") {
      return this.widget().find(".oj-conveyorbelt-overflow-indicator.oj-bottom")[0];
    }
    
    // Non-null locators have to be handled by the component subclasses
    return null;
  },

  // @inheritdoc
  getSubIdByNode: function(node) {
    var startIndicator = this.getNodeBySubId({'subId':'oj-conveyorbelt-start-overflow-indicator'});
    var endIndicator = this.getNodeBySubId({'subId':'oj-conveyorbelt-end-overflow-indicator'});
    var topIndicator = this.getNodeBySubId({'subId':'oj-conveyorbelt-top-overflow-indicator'});
    var bottomIndicator = this.getNodeBySubId({'subId':'oj-conveyorbelt-bottom-overflow-indicator'});
    var currentNode = node;
    var elem = this.element[0];
    while (currentNode && currentNode != elem) {
      if (currentNode === startIndicator)
        return {'subId':'oj-conveyorbelt-start-overflow-indicator'};
      else if (currentNode === endIndicator)
        return {'subId':'oj-conveyorbelt-end-overflow-indicator'};
      else if (currentNode === topIndicator)
        return {'subId':'oj-conveyorbelt-top-overflow-indicator'};
      else if (currentNode === bottomIndicator)
        return {'subId':'oj-conveyorbelt-bottom-overflow-indicator'};

      currentNode = currentNode.parentElement;          
    }
    return null;
  }
  
  // start API doc fragments /////////////////////////////////////////////////////
  
  /**
   * <p>Sub-ID for the start overflow indicator of a horizontal ConveyorBelt.</p>
   *
   * @ojsubid oj-conveyorbelt-start-overflow-indicator
   * @memberof oj.ojConveyorBelt
   *
   * @example <caption>Get the start overflow indicator:</caption>
   * var node = myConveyorBelt.getNodeBySubId({'subId': 'oj-conveyorbelt-start-overflow-indicator'} );
   */
  
  /**
   * <p>Sub-ID for the end overflow indicator of a horizontal ConveyorBelt.</p>
   *
   * @ojsubid oj-conveyorbelt-end-overflow-indicator
   * @memberof oj.ojConveyorBelt
   *
   * @example <caption>Get the end overflow indicator:</caption>
   * var node = myConveyorBelt.getNodeBySubId({'subId': 'oj-conveyorbelt-end-overflow-indicator'} );
   */
  
  /**
   * <p>Sub-ID for the top overflow indicator of a vertical ConveyorBelt.</p>
   *
   * @ojsubid oj-conveyorbelt-top-overflow-indicator
   * @memberof oj.ojConveyorBelt
   *
   * @example <caption>Get the top overflow indicator:</caption>
   * var node = myConveyorBelt.getNodeBySubId({'subId': 'oj-conveyorbelt-top-overflow-indicator'} );
   */
  
  /**
   * <p>Sub-ID for the bottom overflow indicator of a vertical ConveyorBelt.</p>
   *
   * @ojsubid oj-conveyorbelt-bottom-overflow-indicator
   * @memberof oj.ojConveyorBelt
   *
   * @example <caption>Get the bottom overflow indicator:</caption>
   * var node = myConveyorBelt.getNodeBySubId({'subId': 'oj-conveyorbelt-bottom-overflow-indicator'} );
   */

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Gesture</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>ConveyorBelt</td>
   *       <td><kbd>Swipe</kbd></td>
   *       <td>Transition to an adjacent logical page of child items.</td>
   *     </tr>
   *     <tr>
   *       <td>Navigation Arrow</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Transition to an adjacent logical page of child items.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojConveyorBelt
   */

  /**
   * <p>ConveyorBelt is for layout only and does not directly support keyboard 
   * interaction. It is up to the application to provide keyboard support. 
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojConveyorBelt
   */
  
  // end API doc fragments ///////////////////////////////////////////////////////
  
}); // end of oj.__registerWidget

// start static members and functions //////////////////////////////////////////

//FIX : log warning message when "disabled" property set
/**
* Warning message when "disabled" property is set.
* @const
* @private
* @type {string}
*/
var _WARNING_DISABLED_OPTION = "JET ConveyorBelt: 'disabled' property not supported";

//make sure the collapseEventTimeout param is less than the one used in the unit tests
//in order to ensure that the filmStrip listener gets the resize event before the unit test
/**
* Timeout in milliseconds for collapse event.
* @const
* @private
* @type {number}
*/
var _RESIZE_LISTENER_COLLAPSE_EVENT_TIMEOUT = 25;
    
// end static members and functions ////////////////////////////////////////////

}()); // end of ConveyorBelt wrapper function
(function() {
var ojConveyorBeltMeta = {
  "properties": {
    "contentParent": {
      "type": "string"
    },
    "orientation": {
      "type": "string",
      "enumValues": ["horizontal", "vertical"]
    }
  },
  "extension": {
    _WIDGET_NAME: "ojConveyorBelt"
  }
};
oj.CustomElementBridge.registerMetadata('oj-conveyor-belt', 'baseComponent', ojConveyorBeltMeta);
oj.CustomElementBridge.register('oj-conveyor-belt', {'metadata': oj.CustomElementBridge.getMetadata('oj-conveyor-belt')});
})();

/**
 * This class contains all utility methods used by the data grid collection model.
 * @export
 */
oj.DataCollectionEditUtils = {};

/**
 * @export
 * @param {Object} event
 * @param {Object|null=} ui
 */
oj.DataCollectionEditUtils.basicHandleEditEnd = function(event, ui)
{
    var input, cancel, widgetConstructor;
    if (ui == null || ui['cellContext'] == null)
    {
        ui = event.detail;
    }
    input = $(ui['cellContext']['parentElement']).find('.oj-component-initnode')[0];
    widgetConstructor = oj.Components.__GetWidgetConstructor(input);
    cancel = ui['cancelEdit'];
    if (cancel)
    {
        // an escape key was pressed so reset the data before we lose focus on the input
        widgetConstructor('reset');
    }     
    else                        
    {   
        widgetConstructor('validate');
        if (!widgetConstructor('isValid'))
        {
            // not valid so do not allow edit to end
            event.preventDefault();
        }
    }
};

/**
 * 
 * @param {Event} event <code class="prettyprint">jQuery</code> event object
 * @param {Object} ui Parameters
 * @property {Object} ui.rowContext the row context
 * @property {number} ui.cancelEdit whether the edit was canceled
 * @return {boolean} Whether to veto the event
 * @export
 * @expose
 * @memberof! oj.DataCollectionEditUtils
 * @instance
 */
oj.DataCollectionEditUtils.basicHandleRowEditEnd = function(event, ui)
{
    var i, cancel, widgetConsructor ;
    var inputComponents = $(ui['rowContext']['parentElement']).find('.oj-component-initnode');
    for (i = 0; i < inputComponents.length; i++)
    {
      widgetConsructor = oj.Components.__GetWidgetConstructor(inputComponents[i]);
      cancel = ui['cancelEdit'];
      try
      {
        if (cancel)
        {
            // an escape key was pressed so reset the data before we lose focus on the input
            widgetConsructor('reset');
        }     
        else                        
        {   
            widgetConsructor('validate');
            if (!widgetConsructor('isValid'))
            {
                // not valid so do not allow edit to end
                return false;
            }
        }
      }
      catch(e)
      {
      }
    }
    return true;
};
/**
 * @protected
 * @ignore
 */
(function () {
  oj.__KO_CUSTOM_BINDING_PROVIDER_INSTANCE.addPostprocessor
    (
      {
        'nodeHasBindings': function (node, wrappedReturn) {
          return wrappedReturn || (node.nodeType === 1 && 'oj-defer' === node.nodeName.toLowerCase());
        },
        'getBindingAccessors': function (node, bindingContext, wrappedReturn) {
          if (node.nodeType === 1 && 'oj-defer' === node.nodeName.toLowerCase()) {
            wrappedReturn = wrappedReturn || {};
            wrappedReturn['_ojDefer_'] = function () {}
          }
          return wrappedReturn;
        }
      }
    );
})();

ko['bindingHandlers']['_ojDefer_'] =
  {
    'init': function (element, valueAccessor, allBindingsAccessor, viewModel, bindingContext) 
    {
      // _shown is set if subtreeShown was called before bound. Therefore, if _shown was set
      // we can directly call ko.applyBindingsToDescendants() without waiting for subtreeShown
      // because it was already called.
      if (!element['_shown'])
      {
        // stash away the children
        if (!element['_savedChildNodes'])
        {
          var fragment = document.createDocumentFragment();
          var childNodes = element.childNodes;
          while (childNodes.length > 0)
          {
            fragment.appendChild(childNodes[0]);
          }
          element['_savedChildNodes'] = fragment;
        }
        // this _activateDescedantBindings function will be called from the element's
        // _activate API.
        Object.defineProperty(element, '_activateDescedantBindings', {
          'value': function () 
          {
            ko.applyBindingsToDescendants(bindingContext, element);
            delete element['_activateDescedantBindings'];
          },
          'configurable': true
        });
      }
      else
      {
        ko.applyBindingsToDescendants(bindingContext, element);
      }
      return {'controlsDescendantBindings': true};
    }
  };

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * @ojstatus preview
 * @since 4.0.0
 * @ojcomponent oj.ojDefer
 * @ojshortdesc Used to delay applying bindings to its children until it is activated. Child elements are disconnected from the DOM tree until the parent component activates its subtree.
 * @ojsignature {target: "Type", value:"class ojDefer extends JetElement<ojDeferSettableProperties>"}
 * @classdesc
 * The oj-defer custom element delays applying bindings to its children until it is activated.
 * It works by disconnecting child elements from the DOM tree until the parent component activates its subtree. In addition, the tag will
 * defer binding execution and disconnect children within hidden subtrees of the components which support it.
 * There are several components which support oj-defer:<br>
 * <ul>
 * <li>Collapsible</li>
 * <li>Dialog</li>
 * <li>Film Strip</li>
 * <li>Off Canvas</li>
 * <li>Popup</li>
 * <li>MasonryLayout</li>
 * <li>Composite Component Slots</li>
 * </ul>
 * Note: For composite component slots, the oj-defer element could have a slot attribute specified directly on it,
 * i.e. &lt;oj-defer slot="something">, or &lt;oj-defer> could appear within a child subtree of the element with a slot attribute. Also,  
 * the current implementation may allow bindings to be applied
 * to the content within &lt;oj-defer> prematurely if the tag is used in one of the
 * 'hiding' components that is nested within another 'hiding' component. That
 * limitation will be removed in the future.
 */

(function () 
{
  oj.DeferElement = {};

  oj.DeferElement.register = function () {
    var deferElementProto = Object.create(HTMLElement.prototype);
    // define a non-public _activate API which will only be called
    // by subtreeAttach
    Object.defineProperty(deferElementProto, '_activate',
      {
        'value': function () 
        {
          if (!this['_activateDescedantBindings'])
          {
            // if the _activateDescedantBindings function is not there then
            // that means we have not been bound yet. Set a flag to activate when bound
            Object.defineProperty(this, '_shown', {
              'configurable': false,
              'value': true
            });
          }
          else
          {
            // if we have stashed away children, put them back
            if (this['_savedChildNodes'])
            {
              this.appendChild(this['_savedChildNodes']);
              delete this['_savedChildNodes'];
            }
            // if the _activateDescedantBindings function is there then just call it
            this['_activateDescedantBindings']();
          }
        },
        'writable': false
      });
      var constructorFunc = function()
      { 
        var reflect = window['Reflect'];
        var ret;
        if (typeof reflect !== 'undefined')
        {
          ret = reflect['construct'](HTMLElement, [], this['constructor']);
        }
        else
        {
          ret = HTMLElement.call(this);
        }
        return ret;
      };
      Object.defineProperty(deferElementProto,  'constructor', { 'value': constructorFunc, 'writable': true, 'configurable': true });
      constructorFunc.prototype = deferElementProto;
      Object.setPrototypeOf(constructorFunc, HTMLElement);
      customElements.define('oj-defer', constructorFunc);
  };
  oj.DeferElement.register();
})();

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

///////////// ConversionDiagramDataSource //////////////////

/**
 * Internal implementation of the DiagramDataSource dedicated to convert
 * 'nodes', 'links' and 'childNodes' options into DiagramDataSource object.
 * @class oj.ConversionDiagramDataSource
 * @extends oj.DiagramDataSource
 * @classdesc JSON implementation of the oj.DiagramDataSource
 * @param {Object} data JSON data object with following properties:
 * @property {Array|Promise|Function=} nodes an array of nodes
 * @property {Array|Promise|Function=} links an array of links
 * @param {Object=} options the options set on this data source
 * @param {Function=} options.childData Function callback to retrieve nodes and links for the specified parent.
 *                      Function will return an array of nodes or
 *                      a Promise that will resolve with an array of nodes
 * @constructor
 * @ignore
 */
oj.ConversionDiagramDataSource = function(data, options) {
    this.childDataCallback = options ? options['childData'] : null;
    oj.ConversionDiagramDataSource.superclass.constructor.call(this, data);
};

// Subclass from oj.DiagramDataSource
oj.Object.createSubclass(oj.ConversionDiagramDataSource, oj.DiagramDataSource, "oj.ConversionDiagramDataSource");

/**
 * Returns child data for the given parent.
 * The data include all immediate child nodes along with links whose endpoints
 * both descend from the current parent node. 
 * If all the links are available upfront, they can be returned as part of the
 * top-level data (since all nodes descend from the diagram root).
 * If lazy-fetching links is desirable, the most
 * optimal way to return links is as part of the data of the
 * nearest common ancestor of the link's endpoints.
 *
 * @param {Object|null} parentData An object that contains data for the parent node.
 *                     If parentData is null, the method retrieves data for top level nodes.
 * @return {Promise|IThenable} Promise resolves to a component object with the following structure:
 * <ul>
 *  <li>{Array<Object>} nodes An array of objects for the child nodes for the given parent</li>
 *  <li>{Array<Object>} links An array of objects for the links for the given parent</li>
 * </ul>
 * @export
 * @method
 * @name getData
 * @memberof! oj.ConversionDiagramDataSource
 * @instance
 * @ignore
 */
oj.ConversionDiagramDataSource.prototype.getData = function(parentData) {
  if (parentData) { //retrieve child data
    var childData = parentData['nodes'];
    if (childData === undefined && this.childDataCallback) {
      var childNodes = this.childDataCallback(parentData);
      return Promise.resolve(childNodes).then(
          function(values) {
            return Promise.resolve({'nodes':values});
          },
          function(reason) {
            return Promise.resolve({'nodes':[]});
          }
        );
    }
    else {
      return Promise.resolve({'nodes': childData});
    }
  }
  else { // retrieve top level data
    if (this.data) {
      var nodes = this.data['nodes'], 
          links = this.data['links'];
      if (nodes instanceof Function) {
        nodes = nodes();
      }
      if (links instanceof Function) {
        links = links();
      }
      return Promise.all([nodes, links]).then(
        function(values) {
          return Promise.resolve({'nodes':values[0],'links':values[1]});
        },
        function(reason) {
          return Promise.resolve({'nodes':[],'links':[]});
        }
      );
    }
    else {
      return Promise.resolve(null);
    }
  }
};

/**
 * Retrieves number of child nodes
 * @param {Object} nodeData A data object for the node in question.
 *                          See node properties section.
 * @return {number} Number of child nodes if child count is available.
 *                  The method returns 0 for leaf nodes.
 *                  The method returns -1 if the child count is unknown
 *                  (e.g. if the children have not been fetched).
 * @export
 * @method
 * @name getChildCount
 * @memberof! oj.ConversionDiagramDataSource
 * @instance
 * @ignore
 */
oj.ConversionDiagramDataSource.prototype.getChildCount= function(nodeData) {
  if (nodeData) {
    var childData = nodeData['nodes'];
    var count = Array.isArray(childData) ? childData.length :
                childData === undefined && this.childDataCallback ? -1 :
                0;
    return count;
  }
  return -1;
};

/**
 * Indicates whether the specified object contains links
 * that should be discovered in order to display promoted links.
 *
 * @param {Object} nodeData A data object for the container node in question.
 *                          See node properties section.
 * @return {string} the valid values are "connected", "disjoint", "unknown"
 * @export
 * @method
 * @name getDescendantsConnectivity
 * @memberof! oj.ConversionDiagramDataSource
 * @instance
 * @ignore
 */
oj.ConversionDiagramDataSource.prototype.getDescendantsConnectivity = function(nodeData){
  return "unknown";
};

/**
 * @class 
 * @name oj.DiagramUtils
 * @ojtsignore
 * 
 * @classdesc
 * <h3>Diagram Layout Utilities</h3>
 *
 * <p> DiagramUtils is a helper object that provides a function to generate a layout callback for ojDiagram out of JSON object. 
 * A JSON object contains positions for the nodes, paths for the links and properties for positioning a label for a node and a link.
 * See object details {@link oj.DiagramUtils.getLayout}
 *
 * <h3> Usage : </h3>
 * <pre class="prettyprint">
 * <code>
 * // create JSON object that contains positions for the nodes and SVG paths for the links
 * // the nodes and links are identified by ids
 * var data = {
 *  "nodes":[
 *   {"id":"N0", "x":100, "y":0},
 *   {"id":"N1", "x":200, "y":100},
 *   {"id":"N2", "x":100, "y":200},
 *   {"id":"N3", "x":0, "y":100}
 * ],
 * "links":[
 *   {"id":"L0", "path":"M120,20L220,120"},
 *   {"id":"L1", "path":"M220,120L120,220"},
 *   {"id":"L2", "path":"M120,220L20,120"},
 *   {"id":"L3", "path":"M20,120L120,20"}
 * ]
 * };
 * //generate the layout callback function using data and the oj.DiagramUtils
 * // pass the generated function to the oj.ojDiagram as the 'layout' option
 * var layoutFunc = oj.DiagramUtils.getLayout(data);
 * </code></pre>
 * @export
 * @constructor
 * @since 3.0
 */
oj.DiagramUtils = function() {
};

/**
 * The complete label layout object used to position node and link label
 * @typedef {Object} oj.DiagramUtils.LabelLayout
 * @property {number} x x-coordinate for the label
 * @property {number} y y-coordinate for the label
 * @property {number} rotationPointX x-coordinate for label rotation point
 * @property {number} rotationPointY y-coordinate for label rotation point
 * @property {number} number angle of rotation for the labelLayout
 * @property {string} halign horizontal alignment for the label. Valid values are "left", "right" or "center"
 * @property {string} valign vertical alignment for the label. Valid values are "top", "middle", "bottom" or "baseline". 
 *                           The default value is <code class="prettyprint">"top"</code>
 */

/**
 * A function that generates the layout callback function for the ojDiagram component.
 * @param {Object} obj JSON object that defines positions of nodes, links paths and label layouts. The object supports the following properties.
 * @property {Array<Object>} obj.nodes An array of objects with the following properties that describe a position for the diagram node and a layout for the node's label
 * @property {number} obj.nodes.x x-coordinate for the node
 * @property {number} obj.nodes.y y-coordinate for the node
 * @property {Object} obj.nodes.labelLayout An object that defines label layout for the node. See {@link oj.DiagramUtils.LabelLayout} object. 
 *                                          The object defines absolute coordinates for label position.
 * @property {Array<Object>} obj.links An array of objects with the following properties that describe a path for the diagram link and a layout for the link's label.
 * @property {string} obj.links.path A string that represents an SVG path for the link.
 * @property {string} obj.links.coordinateSpace The coordinate container id for the. If specified the link points will be applied relative to that container. 
 *                                              If the value is not set, the link points are in the global coordinate space.
 * @property {Object} obj.links.labelLayout An object that defines label layout for the link. See {@link oj.DiagramUtils.LabelLayout} object.
 *
 * @property {Object} obj.nodeDefaults An object that defines the default layout of the node label
 * @property {Object|Function} obj.nodeDefaults.labelLayout An object that defines default label layout for diagram nodes.
 *                         See {@link oj.DiagramUtils.LabelLayout} object. The object defines relative coordinates for label position.
 *                         E.g. if all the node labels should be positioned with a certain offset relative to the node, 
 *                         a label position can be defined using an object in node defaults.
 *                         <p>Alternatively a label layout can be defined with a function. The function will receive the following parameters:
 *                           <ul>
 *                             <li>{DvtDiagramLayoutContext} - layout context for the diagram</li>
 *                             <li>{DvtDiagramLayoutContextNode} - layout context for the current node</li>
 *                           </ul>
 *                           The return value of the function is a label object with the following properties : {@link oj.DiagramUtils.LabelLayout}. 
 *                           The object defines absolute coordinates for label position.
 *                          </p>
 * @property {Object} obj.linkDefaults An object that defines a function for generating a link path and a default layout for the link label
 * @property {Function} obj.linkDefaults.path a callback function that will be used to generate a link path. The function will receive the following parameters:
 *                      <ul>
 *                        <li>{DvtDiagramLayoutContext} - layout context for the diagram</li>
 *                        <li>{DvtDiagramLayoutContextLink} - layout context for the current link</li>
 *                      </ul>
 *                      The return value of the function is a string that represents an SVG path for the link 
 * @property {Function} obj.linkDefaults.labelLayout a function that defines default label layout for diagram links. The function will receive the following parameters:
 *                      <ul>
 *                        <li>{DvtDiagramLayoutContext} - layout context for the diagram</li>
 *                        <li>{DvtDiagramLayoutContextLink} - layout context for the current link</li>
 *                      </ul>
 *                      The return value of the function is a label object with the following properties {@link oj.DiagramUtils.LabelLayout}
 * @property {Object|Function} obj.viewport An object with the following properties that defines diagram viewport.
 *                         <p>Alternatively a viewport can be defined with a function. The function will receive the following parameters:
 *                           <ul>
 *                             <li>{DvtDiagramLayoutContext} - layout context for the diagram</li>
 *                           </ul>
 *                           The return value of the function is a viewport object with the properties defined below. 
 *                          </p>
 * @property {number} obj.viewport.x x-coordinate
 * @property {number} obj.viewport.y y-coordinate 
 * @property {number} obj.viewport.w width
 * @property {number} obj.viewport.h height
 * @returns {Function} layout callback function
 * @export
 */
oj.DiagramUtils.getLayout = function(obj) {
  var layoutFunc = function(layoutContext) {
    
    // position nodes and node labels
    if (obj['nodes'] && layoutContext.getNodeCount() > 0) {
      var nodesDataMap = oj.DiagramUtils._dataArrayToMap(obj['nodes']);
      var defaultLabelLayout = obj['nodeDefaults'] && obj['nodeDefaults']['labelLayout'] ? obj['nodeDefaults']['labelLayout'] : null;
      for (var ni = 0;ni < layoutContext.getNodeCount();ni++) {
        var node = layoutContext.getNodeByIndex(ni);
        var nodeData = nodesDataMap[node.getId()];
        oj.DiagramUtils._positionChildNodes(node.getChildNodes(), nodeData ? nodeData['nodes'] : null, layoutContext, defaultLabelLayout);
        oj.DiagramUtils._positionNodeAndLabel(node, nodeData, layoutContext, defaultLabelLayout);
      }
    }
    
    // position links and link labels
    if (obj['links'] && layoutContext.getLinkCount() > 0) {
      var linksDataMap = oj.DiagramUtils._dataArrayToMap(obj['links']);
      var defaultPath = obj['linkDefaults'] && obj['linkDefaults']['path'] ? obj['linkDefaults']['path'] : null;
      var defaultLabelLayout = obj['linkDefaults'] && obj['linkDefaults']['labelLayout'] ? obj['linkDefaults']['labelLayout'] : null;
      for (var li = 0;li < layoutContext.getLinkCount();li++) {
        var link = layoutContext.getLinkByIndex(li);
        var linkData = linksDataMap[link.getId()];
        if (linkData && linkData['path']) {
          link.setPoints(linkData['path']);
        }
        else if (defaultPath && defaultPath instanceof Function) {
          link.setPoints(defaultPath(layoutContext, link));
        }
        if (linkData && linkData['coordinateSpace']) {
          link.setCoordinateSpace(linkData['coordinateSpace']);
        }        
        //position label if it exists
        if (linkData && linkData['labelLayout']) {
          oj.DiagramUtils._setLabelPosition(link, linkData['labelLayout']);
        }
        else if (defaultLabelLayout && defaultLabelLayout instanceof Function ) {
          oj.DiagramUtils._setLabelPosition(link, defaultLabelLayout(layoutContext, link));
        }
      }
    }
    if (obj['viewport']){
      var viewport = obj['viewport'];
      if (viewport instanceof Function) {
        layoutContext.setViewport(viewport(layoutContext));
      }
      else {
        layoutContext.setViewport(viewport);
      }
    }
  };
  return layoutFunc;
};

/**
 * Converts a data array of nodes or links to a map
 * @param {Array} dataArray data array of node or links
 * @return {Object} a map of nodes or links
 * @private
 * @instance
 * @memberof oj.DiagramUtils
 */
oj.DiagramUtils._dataArrayToMap = function(dataArray) {
  var m = {};
  if (dataArray) {
    for (var i = 0; i < dataArray.length; i++) {
      m[dataArray[i]['id']] = dataArray[i];
    }
  }
  return m;
};

/**
 * Positions child nodes and their labels
 * @param {Array} nodes An array of diagram nodes
 * @param {Array} nodesData An array of objects that describe a position for a diagram node and a layout for the node's label
 * @param {Object} layoutContext Layout context for diagram
 * @param {Object|Function} defaultLabelLayout Default label layout defined as an object or a function
 * @private
 */
oj.DiagramUtils._positionChildNodes =  function(nodes, nodesData, layoutContext, defaultLabelLayout){
  if (nodes && nodesData) {
    var nodesDataMap = oj.DiagramUtils._dataArrayToMap(nodesData);
    for (var ni = 0;ni < nodes.length;ni++) {
      var node = nodes[ni];
      var nodeData = nodesDataMap[node.getId()];
      oj.DiagramUtils._positionChildNodes(node.getChildNodes(), nodeData ? nodeData['nodes'] : null, layoutContext, defaultLabelLayout);
      oj.DiagramUtils._positionNodeAndLabel(node, nodeData, layoutContext, defaultLabelLayout);
    }
  }
};

/**
 * Position a diagram nodes and its label
 * @param {Object} node A node to position
 * @param {Object} nodeData An object that defines a position for the node and a layout for the node's label
 * @param {Object} layoutContext Layout context for diagram
 * @param {Object|Function} defaultLabelLayout Default label layout defined as an object or a function
 * @private
 */
oj.DiagramUtils._positionNodeAndLabel = function(node, nodeData, layoutContext, defaultLabelLayout) {
  if (node && nodeData) {
    node.setPosition({'x': nodeData['x'], 'y': nodeData['y']});
    //node has a label - position it
    if (nodeData['labelLayout']) { 
      //layout should be an object - expect absolute positions
      oj.DiagramUtils._setLabelPosition(node, nodeData['labelLayout']);
    }
    else if (defaultLabelLayout && defaultLabelLayout instanceof Function) {
      oj.DiagramUtils._setLabelPosition(node, defaultLabelLayout(layoutContext, node));
    }
    else if (defaultLabelLayout) {
      //layout should be an object - expect relative positions
      oj.DiagramUtils._setLabelPosition(node, defaultLabelLayout, node.getPosition());
    }
  }
};

/**
 * Sets label position for a link or a node
 * @param {Object} obj layout context for node or link
 * @param {Object} labelLayout an object with the following properties for the label layout
 * @property {number} x x-coordinate for the label position
 * @property {number} y y-coordinate for the label position
 * @property {number} rotationPointX x-coordinate for the rotation point
 * @property {number} rotationPointY y-coordinate for the rotation point
 * @property {number} angle angle for the angle of rotation
 * @property {string} halign horizontal alignment for the label
 * @property {string} valign vertical alignment for the label 
 * @param {Object=} offset an object with the following properties for the label offset
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @private
 * @instance
 * @memberof oj.DiagramUtils
 */
oj.DiagramUtils._setLabelPosition = function(obj, labelLayout, offset) {
  offset = offset ? offset : {'x':0,'y':0};
  obj.setLabelPosition({'x': labelLayout['x'] + offset['x'], 'y': labelLayout['y'] + offset['y']});
  var rotationPointX = labelLayout['rotationPointX'], 
      rotationPointY = labelLayout['rotationPointY'];
  if (!isNaN(rotationPointX) && !isNaN(rotationPointY)) {
    obj.setLabelRotationPoint({'x':rotationPointX + offset['x'], 'y':rotationPointY + offset['y']});
  }
  obj.setLabelRotationAngle(labelLayout['angle']);
  obj.setLabelHalign(labelLayout['halign']);
  obj.setLabelValign(labelLayout['valign']);
};

// Copyright (c) 2008, 2014, Oracle and/or its affiliates. All rights reserved.
/**
 * <p>Pluggable layout code must conform to the pluggable layout contract. The following steps outline a simple pluggable layout.</p>
 * <ol>
 * <li> Diagram calls layout engine with layout context.</li>
 * <li> Layout engine loops over nodes using getNodeCount and getNodeByIndex()</li>
 * <ul>
 *  <li> layout engine gets dimensions of node by calling getBounds()</li>
 *  <li> layout engine sets position on node by calling setPosition()</li>
 * </ul>
 * <li> Layout engine loops over links using getLinkCount and getLinkByIndex()</li>
 * <ul>
 *  <li> layout engine gets start and end nodes by calling getStartId()/getEndId() on a link object</li>
 *  <li> layout engine analyzes nodes position and dimensions by calling getPosition() and getBounds() on each node and finds connection points for the link start/end</li>
 *  <li> layout engine creates a link path by calling setPoints() on a link object</li>
 * </ul>
 * <li> Diagram applies node positions from layout context</li>
 * </ol>
 *
 * <p>The DvtDiagramLayoutContext interface defines the context for a layout call.</p>
 * @export
 * @interface DvtDiagramLayoutContext
 * @since 3.0
 * @ojtsignore
 * @memberof oj
 */

/**
 * Get a node context by id.
 * @method getNodeById
 * @instance
 * @param {string} id id of node context to get
 * @return {oj.DvtDiagramLayoutContextNode}
 * @memberof oj.DvtDiagramLayoutContext 
 * @export
 */

/**
 * Get a node context by index.
 * @method getNodeByIndex
 * @instance 
 * @param {number} index index of node context to get
 * @return {oj.DvtDiagramLayoutContextNode}
 * @memberof oj.DvtDiagramLayoutContext 
 * @export
 */

/**
 * Get the number of nodes to layout.
 * @method getNodeCount
 * @instance  
 * @return {number}
 * @memberof oj.DvtDiagramLayoutContext 
 * @export
 */

/**
 * Get a link context by id.
 * @method getLinkById
 * @instance   
 * @param {string} id id of link context to get
 * @return {oj.DvtDiagramLayoutContextLink}
 * @memberof oj.DvtDiagramLayoutContext 
 * @export
 */

/**
 * Get a link context by index.
 * @method getLinkByIndex
 * @instance 
 * @param {number} index index of link context to get
 * @return {oj.DvtDiagramLayoutContextLink}
 * @memberof oj.DvtDiagramLayoutContext
 * @export
 */

/**
 * Get the number of links to layout.
 * @method getLinkCount
 * @instance 
 * @return {number}
 * @memberof oj.DvtDiagramLayoutContext 
 * @export
 */

/**
 * Get whether the reading direction for the locale is right-to-left.
 * @method isLocaleR2L
 * @instance 
 * @return {boolean}
 * @memberof oj.DvtDiagramLayoutContext  
 * @export
 */

/**
 * Get the size of the Diagram.
 * @method getComponentSize
 * @instance 
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height
 * @return {Object} An object containing properties of the diagram size
 * @memberof oj.DvtDiagramLayoutContext  
 * @export
 */

/**
 * Set the viewport the component should use after the layout, in the layout's coordinate system.
 * @method setViewport
 * @instance 
 * @param {Object} viewport An object containing properties of the viewport that the component should use after the layout
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height
 * @memberof oj.DvtDiagramLayoutContext
 * @export
 */

/**
 * Get the viewport the component should use after the layout, in the layout's coordinate system.
 * @method getViewport
 * @instance 
 * @return {Object} An object containing properties of the viewport
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height
 * @memberof oj.DvtDiagramLayoutContext  
 * @export
 */

/**
 * Get the current viewport used by the component in the layout's coordinate system for the diagram
 * @method getCurrentViewport
 * @instance  
 * @return {Object} An object containing properties of the current viewport
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height 
 * @memberof oj.DvtDiagramLayoutContext  
 * @export
 */
 
 /**
 * The function retrieves nearest common ancestor container for two nodes.
 * @method getCommonContainer
 * @instance  
 * @return {string} Id for the first common ancestor container or null for top level diagram
 * @param {string} nodeId1 first node id
 * @param {string} nodeId2 second node id
 * @memberof oj.DvtDiagramLayoutContext  
 * @export
 */
 
 /**
 * Gets event data object. Values can be retrieved from the object using 'type' and 'data' keys.
 * @method getEventData
 * @instance 
 * @return {Object} event data object
 * @property {string} type Event type - "add", "remove" or "change".
 * @property {Object} data Event payload object for the event - 
 *            see <a href="oj.DiagramDataSource.html#EventType">EventType</a> for event details. 
 * @memberof oj.DvtDiagramLayoutContext
 * @since 4.0.0
 * @export
 */

// Copyright (c) 2008, 2014, Oracle and/or its affiliates. All rights reserved.
/**
 * The DvtDiagramLayoutContextLink interface defines the link context for a layout call.
 * @export
 * @interface DvtDiagramLayoutContextLink
 * @since 3.0
 * @ojtsignore
 * @memberof oj
 */

/**
 * Get the id of the link.
 * @method getId
 * @instance
 * @return {string}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the id of the start node of this link.
 * @method getStartId
 * @instance
 * @return {string}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the id of the end node of this link.
 * @method getEndId
 * @instance
 * @return {string}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Set the points to use for rendering this link.  The given array can contain
 * coordinates, like [x1, y1, x2, y2, ..., xn, yn], or SVG path commands, like
 * ["M", x1, y1, "L", x2, y2, ..., "L", xn, yn].  The points are in the
 * coordinate system of the link's container.
 * @method setPoints
 * @instance
 * @param {array} points array of points to use for rendering this link
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the points to use for rendering this link.  The returned array can
 * contain coordinates, like [x1, y1, x2, y2, ..., xn, yn], or SVG path
 * commands, like ["M", x1, y1, "L", x2, y2, ..., "L", xn, yn].  The points
 * are in the coordinate system of the link's container.
 * @method getPoints
 * @instance
 * @return {array}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Set the position of the link label.  The position is in the coordinate system of the link's container.
 * The position represents the upper-left corner for locales with left-to-right reading
 * direction and the upper-right corner for locales with right-to-left reading
 * direction in the default alignment.
 * @method setLabelPosition
 * @instance
 * @param {Object} pos An object with the following properties for the position of the link label
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the position of the link label.  The position is in the coordinate system of the link's container.
 * The position represents the upper-left corner for locales with left-to-right reading
 * direction and the upper-right corner for locales with right-to-left reading
 * direction in the default alignment.
 * @method getLabelPosition
 * @instance
 * @return {Object} An object with the following properties for the position of the link label
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Set the horizontal alignment of the link label.
 * The default value is left for locales with left-to-right reading direction and right for locales with
 * right-to-left reading direction.
 * @method setLabelHalign
 * @instance
 * @param {string} halign A string that specifies the horizontal label alignment
 * @ojvalue {string} "left"
 * @ojvalue {string} "center"
 * @ojvalue {string} "right"
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the horizontal alignment of the link label.
 * The default value is left for locales with left-to-right reading direction and right for locales with
 * right-to-left reading direction.
 * @method getLabelHalign
 * @instance
 * @return {string} A string that specifies the horizontal label alignment
 * @ojvalue {string} "left"
 * @ojvalue {string} "center"
 * @ojvalue {string} "right"
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Set the vertical alignment of the link label.
 * @method setLabelValign
 * @instance
 * @param {string} valign A string that specifies the vertical label alignment. The "baseline" value
 *                        is not supported when applied to multiline labels. If "baseline" is applied
 *                        to a multiline label, the text element will default to "top" alignment.
 * @default <code class="prettyprint">"top"</code>
 * @ojvalue {string} "top"
 * @ojvalue {string} "middle"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "baseline"
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the vertical alignment of the link label.
 * @method getLabelValign
 * @instance
 * @return {string} A string that specifies the vertical label alignment
 * @default <code class="prettyprint">"top"</code>
 * @ojvalue {string} "top"
 * @ojvalue {string} "middle"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "baseline"
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the label bounds.  The bounds are in the coordinate system of the label.
 * @method getLabelBounds
 * @instance
 * @return {Object} An object with the following properties for the label bound
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the offset of the start connector.  This is the amount of space that the
 * link should leave between its starting point and the node for the connector
 * to be drawn.
 * @method getStartConnectorOffset
 * @instance
 * @return {number}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the offset of the end connector.  This is the amount of space that the
 * link should leave between its ending point and the node for the connector
 * to be drawn.
 * @method getEndConnectorOffset
 * @instance
 * @return {number}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the width of this link.
 * @method getLinkWidth
 * @instance
 * @return {number}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the corresponding object from the links option array.
 * @method getLayoutAttributes
 * @instance
 * @return {object}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @ojdeprecated {since:"2.1", description:'This method is deprecated. Please use the <a href="#getData">getData</a> method instead'}
 * @export
 */

/**
 * Determine whether this link is selected.
 * @method getSelected
 * @instance
 * @return {boolean}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */


/**
 * Set the angle of rotation of the link label, relative to the label
 * rotation point, in radians.
 * @method setLabelRotationAngle
 * @instance
 * @param {number} angle angle of rotation
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the angle of rotation of the link label, relative to the label
 * rotation point, in radians.
 * @method getLabelRotationAngle
 * @instance
 * @return {number}
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Set the point about which to rotate the link label, in the coordinate
 * system of the label.
 * @method setLabelRotationPoint
 * @instance
 * @param {Object} point An object with the following properties for the label rotation point
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the point about which to rotate the link label, in the coordinate
 * system of the label.
 * @method getLabelRotationPoint
 * @instance
 * @return {Object} An object with the following properties for the label rotation point
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */
 
/**
 * Set coordinate space for the link.
 * If the coordinate container id is specified, the link points will be applied relative to that container.
 * If the coordinate container id is not specified, the link points are in the global coordinate space.
 * The coordinate space is global by default.
 * The coordinate space can be any container id that is an ancestor of both the start and end nodes.
 * @method setCoordinateSpace
 * @instance
 * @param {string} containerId  coordinate container id for the link
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get coordinate space for the link.
 * If the coordinate container id is specified, the link points will be applied relative to that container.
 * If the coordinate container id is not specified, the link points are in the global coordinate space.
 * The coordinate space is global by default.
 * The coordinate space can be any container id that is an ancestor of both the start and end nodes.
 * @method getCoordinateSpace
 * @instance
 * @return {string} coordinate container id for the link
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Get the corresponding data for the diagram link.
 * @method getData
 * @instance
 * @return {object|array} a data object for the link if the link is not promoted or 
 * an array of data objects for each link that is represented by the promoted link if the link is promoted
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */

/**
 * Determine whether this link is promoted.
 * @method isPromoted
 * @instance
 * @return {boolean} true if the link is promoted
 * @memberof oj.DvtDiagramLayoutContextLink
 * @export
 */
// Copyright (c) 2008, 2014, Oracle and/or its affiliates. All rights reserved.

/**
 * The DvtDiagramLayoutContextNode interface defines the node context for a layout call.
 * @export
 * @interface DvtDiagramLayoutContextNode
 * @since 3.0
 * @ojtsignore
 * @memberof oj
 */

/**
 * Get the id of the node.
 * @method getId
 * @instance
 * @return {string}
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the bounds of the node.  Currently the method returns the same value as getContentBounds().
 * The bounds are in the coordinate system of the node.
 * @method getBounds
 * @instance
 * @return {Object} An object with the following properties for the bounds of the node
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the bounds of the node content.  Currently the method returns the same value as getBounds().
 * The bounds are in the coordinate system of the node.
 * @method getContentBounds
 * @instance
 * @return {Object} An object with the following properties for the bounds of the node content
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Set the position of the node.  The position is in the coordinate system of
 * the node's container.
 * @method setPosition
 * @instance
 * @param {Object} pos An object with the following properties for the position of the node
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the position of the node.  The position is in the coordinate system of
 * the node's container.
 * @method getPosition
 * @instance
 * @return {Object} An object with the following properties for the position of the node
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Set the position of the node label.  The position is in the coordinate system of the node's container.
 * The position represents the upper-left corner for locales with left-to-right reading
 * direction and the upper-right corner for locales with right-to-left reading
 * direction.
 * @method setLabelPosition
 * @instance
 * @param {Object} pos An object with the following properties for position of the node label
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the position of the node label.  The position is in the coordinate system of the node's container.
 * The position represents the upper-left corner for locales with left-to-right reading
 * direction and the upper-right corner for locales with right-to-left reading
 * direction.
 * @method getLabelPosition
 * @instance
 * @return {Object} An object with the following properties for the position of the node label
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Set the horizontal alignment of the node label.
 * The default value is left for locales with left-to-right reading direction and right for locales with
 * right-to-left reading direction.
 * @method setLabelHalign
 * @instance
 * @param {string} halign A string that specifies the horizontal label alignment
 * @ojvalue {string} "left"
 * @ojvalue {string} "center"
 * @ojvalue {string} "right"
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the horizontal alignment of the node label.
 * The default value is left for locales with left-to-right reading direction and right for locales with
 * right-to-left reading direction.
 * @method getLabelHalign
 * @instance
 * @return {string} A string that specifies the horizontal label alignment
 * @ojvalue {string} "left"
 * @ojvalue {string} "center"
 * @ojvalue {string} "right"
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Set the vertical alignment of the node label.
 * @method setLabelValign
 * @instance
 * @param {string} valign A string that specifies the vertical label alignment. The "baseline" value
 *                        is not supported when applied to multiline labels. If "baseline" is applied
 *                        to a multiline label, the text element will default to "top" alignment.
 * @default <code class="prettyprint">"top"</code>
 * @ojvalue {string} "top"
 * @ojvalue {string} "middle"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "baseline"
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the vertical alignment of the node label.
 * @method getLabelValign
 * @instance
 * @return {string} A string that specifies the vertical label alignment
 * @default <code class="prettyprint">"top"</code>
 * @ojvalue {string} "top"
 * @ojvalue {string} "middle"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "baseline"
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */
/**
 * Get the label bounds.  The bounds are in the coordinate system of the label.
 * @method getLabelBounds
 * @instance
 * @return {Object} An object with the following properties for the label bounds
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @property {number} w width
 * @property {number} h height
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the corresponding object from the nodes option array.
 * @method getLayoutAttributes
 * @instance
 * @return {Object}
 * @memberof oj.DvtDiagramLayoutContextNode
 * @ojdeprecated {since:"2.1", description:'This method is deprecated. Please use the <a href="#getData">getData</a> method instead'}
 * @export
 */

/**
 * Determine whether this node is selected.
 * @method getSelected
 * @instance
 * @return {boolean}
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Set the angle of rotation of the node label, relative to the label
 * rotation point, in radians.
 * @method setLabelRotationAngle
 * @instance
 * @param {number} angle angle of rotation
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the angle of rotation of the node label, relative to the label
 * rotation point, in radians.
 * @method getLabelRotationAngle
 * @instance
 * @return {number}
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Set the point about which to rotate the node label, in the coordinate
 * system of the label.
 * @method setLabelRotationPoint
 * @instance
 * @param {Object} point An object with the following properties for label rotation point
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the point about which to rotate the node label, in the coordinate
 * system of the label.
 * @method getLabelRotationPoint
 * @instance
 * @return {Object} An object with the following properties for the label rotation point
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */
 
/**
 * Determine whether this container is disclosed.
 * @method isDisclosed
 * @instance
 * @return {boolean} true if the node is a disclosed container
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get visible child nodes for the disclosed container.
 * @method getChildNodes
 * @instance
 * @return {array} array of DvtDiagramLayoutContextNode objects
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * The method returns a position of the node relative to the specified ancestor container. 
 * If the container id is null, the method returns global position of the node.
 * If the container id is not an ancestor id for the node, the method returns null.
 * @method getRelativePosition
 * @instance
 * @param {string} containerId ancestor id
 * @return {Object} An object with the following properties for the position of the node relative to the specified ancestor container
 * @property {number} x x-coordinate
 * @property {number} y y-coordinate
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Get the corresponding data object for the diagram node.
 * @method getData
 * @instance
 * @return {object} a data object for the node 
 * @memberof oj.DvtDiagramLayoutContextNode
 * @export
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

 /**
 * @ojcomponent oj.ojDiagram
 * @augments oj.dvtBaseComponent
 * @since 1.1.0
 * @ojstatus preview
 * @ojrole application 
 * @ojtsignore 
 * @ojshortdesc Displays a set of nodes and the links between them. The node positions and link paths are specified by an application-provided layout function.
 *
 * @classdesc
 * <h3 id="diagramOverview-section">
 *   JET Diagram
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#diagramOverview-section"></a>
 * </h3>
 *
 * <p>Diagrams are used to display a set of nodes and the links between them. The node positions 
 * and link paths are specified by an application-provided layout function
 * (see <a href="oj.ojDiagram.html#diagramLayout-section">JET Diagram Layout</a>).
 * </p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-diagram
 *    layout = '{{layoutFunc}}'
 *    data = '{{dataSource}}'>
 * &lt;/oj-diagram>
 * </code>
 * </pre>
 *
 * <h3 id="diagramLayout-section">
 *   JET Diagram Layout
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#diagramLayout-section"></a>
 * </h3>
 *
 * <p>In order to create JET Diagram component, applications are required to provide a layout callback function for 
 * positioning nodes and links. The component does not deliver a default layout. 
 * However there is a set of demo layouts that are delivered with the Cookbook application. 
 * The demo layouts can be reused by the application, but in most cases we expect that the 
 * application will want to create their own layout. The layout code must conform to the pluggable layout contract. 
 * See {@link oj.ojDiagram#layout} for additional information on layout API.</p>
 *
 * <p>In the case when the node positions are known in advance or derived from an external layout engine, 
 * the layout can be generated using [layout helper utility]{@link oj.DiagramUtils}.</p>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets.
 * Alternate visualizations should be considered if identifying data changes is important,
 * since all nodes will generally move and resize on any data change.
 * </p>
 *
 * <h4>Data Set Size</h4>
 * <p>Applications should avoid setting very large data densities on this element.
 * Applications can aggregate small nodes to reduce the displayed data set size.
 * </p>
 * 
 * <h4>Styling</h4>
 * <p>Use the highest level property available. For example, consider setting styling properties on
 *    <code class="prettyprint">styleDefaults.nodeDefaults</code> or
 *    <code class="prettyprint">styleDefaults.linkDefaults</code>, instead of styling properties
 *    on the individual nodes and links. The diagram can take advantage of these higher level properties to apply the style properties on
 *    containers, saving expensive DOM calls.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojDiagram', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * Specifies the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">animation-on-data-change</code> attribute specified:</caption>
     * &lt;oj-diagram animation-on-data-change='auto'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">animationOnDataChange</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.animationOnDataChange;
     * 
     * // setter
     * myDiagram.animationOnDataChange="auto";
     */
    animationOnDataChange: "none",
    
    /**
     * Specifies the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">animation-on-display</code> attribute specified:</caption>
     * &lt;oj-diagram animation-on-display='auto'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">animationOnDisplay</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.animationOnDisplay;
     * 
     * // setter
     * myDiagram.animationOnDisplay="auto";
     */
    animationOnDisplay: "none",

    /**
     * Provides support for HTML5 Drag and Drop events. Please refer to <a href="https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/Drag_and_drop">third party documentation</a> on HTML5 Drag and Drop to learn how to use it. 
     * @ojshortdesc Used to customize the drag and drop features.
     * @expose
     * @name dnd
     * @memberof oj.ojDiagram
     * @instance
     * @type {Object}
     * @default {"drag": null, "drop": null}
     *
     * @example <caption>Initialize the diagram with the
     * <code class="prettyprint">dnd</code> attribute specified:</caption>
     *
     * <!-- Using dot notation -->
     * &lt;oj-diagram dnd.drop.background.data-types = '["text/shape"]' dnd.drop.background.drop = '[[onDropFunc]]'>&lt;/oj-diagram>
     * 
     * &lt;oj-diagram dnd='[[dndObject]]'>&lt;/oj-diagram>
     *
     * @example <caption>Get or set the <code class="prettyprint">dnd</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myDiagram.dnd.drag.nodes.dataTypes;
     * 
     * // Get all
     * var values = myDiagram.dnd;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myDiagram.setProperty('dnd.drop.nodes.drop', onDropFunc);
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myDiagram.dnd=dndObject;
     */
     dnd: {
      /**
       * An object that describes drag functionality.
       * @ojshortdesc Used to customize drag functionality.
       * @expose
       * @name dnd.drag
       * @memberof! oj.ojDiagram
       * @instance
       * @type {Object}
       * @default null
       */
       drag: null,
        /**
         * If this object is specified, the diagram will initiate drag operation when the user drags on diagram nodes. 
         * @ojshortdesc Used to customize the drag features for nodes.
         * @expose
         * @name dnd.drag.nodes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         */
        /**
         * The MIME types to use for the dragged data in the dataTransfer object. This can be a string if there is only one type, or an array of strings if multiple types are needed. For example, if selected employee data items are being dragged, dataTypes could be "application/employees+json". Drop targets can examine the data types and decide whether to accept the data. For each type in the array, dataTransfer.setData will be called with the specified type and the data. The data is an array of the dataContexts of the selected data items. The dataContext is the JSON version of the dataContext that we use for "tooltip" option, excluding componentElement and parentElement. This property is required unless the application calls setData itself in a dragStart callback function. 
         * @ojshortdesc Specifies MIME type for dragged data.
         * @expose
         * @name dnd.drag.nodes.dataTypes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string|Array.<string>}
         */
        /**
         * An optional callback function that receives the "drag" event as argument. 
         * @ojshortdesc Optional handler for drag event.
         * @expose
         * @name dnd.drag.nodes.drag
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event)}
         */
        /**
         * An optional callback function that receives the "dragend" event as argument. 
         * @ojshortdesc Optional handler for dragend event.
         * @expose
         * @name dnd.drag.nodes.dragEnd
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event)}
         */
        /**
         * An optional callback function that receives the "dragstart" event and context information as arguments. 
         * The context information is as follows: 
         * <ul> 
         *   <li>nodes {Array.(object)}: An array of dataContexts of the dragged data nodes. 
         *      The dataContext for the node has the following properties: 
         *      <ul>
         *        {@ojinclude "name":"dataContext"}
         *        <li>nodeOffset: An object with x,y properties, that reflects node offset from the upper left corner of the bounding box for the dragged content.</li>
         *      </ul>
         *   </li>
         * </ul>
         * This function can set its own data and drag image as needed. When this function is called, event.dataTransfer is already populated with the default data and drag image. 
         * @ojshortdesc Optional handler for dragstart event.
         * @expose
         * @name dnd.drag.nodes.dragStart
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * If this object is specified, the diagram will initiate link creation when the user starts dragging from a port. 
         * @ojshortdesc Defines link creation functionality.
         * @expose
         * @name dnd.drag.ports
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         * @default {}
         */
        /**
         * An optional callback function for customizing link feedback based on a starting node and a port. 
         * If the function is not specified the link feedback will use default link styles.
         * The function will take a single parameter - a context object with the following properties: 
         * <ul>
         *   <li>dataContext The dataContext object of the link start node.
         *    <ul>{@ojinclude "name":"dataContext"}</ul>
         *   </li>
         *   <li>portElement DOM element recognized as a port that received drag event.</li>
         * </ul>
         * The function should return one of the following: 
         * <ul> 
         *   <li>an object with the following properties: 
         *     <ul>
         *       <li> svgStyle : Inline stlye object to be applied on the link feedback </li>
         *       <li> svgClassName : A name of a style class to be applied on the link </li> 
         *     </ul> 
         *   </li>
         *   <li>null or undefined : the default link styles will be used for the link feedback </li>
         * </ul> 
         * @ojshortdesc Specifies link creation feedback.
         * @expose
         * @name dnd.drag.ports.linkStyle
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Object)}
         */
        /**
         * A string, containing a selector expression, that will be used to identify the descendant DOM element in a diagram node that can be used for link creation. This property is requred. 
         * @ojshortdesc Specifies DOM element selector used to start link creation.
         * @expose
         * @name dnd.drag.ports.selector
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         */
        /**
         * The MIME types to use for the dragged data in the dataTransfer object. This can be a string if there is only one type, or an array of strings if multiple types are needed. and parentElement. This property is required unless the application calls setData itself in a dragStart callback function. 
         * @ojshortdesc Specifies MIME type for dragged data.
         * @expose
         * @name dnd.drag.ports.dataTypes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string|Array.<string>}
         */
        /**
         * An optional callback function that receives the "drag" event as argument. 
         * @ojshortdesc Optional handler for drag event.
         * @expose
         * @name dnd.drag.ports.drag
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event)}
         */
        /**
         * An optional callback function that receives the "dragend" event as argument. 
         * @ojshortdesc Optional handler for dragend event.
         * @expose
         * @name dnd.drag.ports.dragEnd
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event)}
         */
        /**
         * An optional callback function that receives the "dragstart" event and context information as arguments.
         * The context information is as follows: 
         * <ul> 
         *   <li> ports {object}: An object with the following properties: 
         *     <ul> 
         *       <li> dataContext : The dataContext object of the link start node with the following properties:
         *        <ul>{@ojinclude "name":"dataContext"}<ul>
         *       </li>
         *       <li> portElement : DOM element recognized as a port that received drag event. </li>
         *     </ul>
         *   </li>
         * </ul>
         * @ojshortdesc Optional handler for dragstart event.
         * @expose
         * @name dnd.drag.ports.dragStart
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
      /**
       * An object that describes drop functionality.
       * @expose
       * @name dnd.drop
       * @memberof! oj.ojDiagram
       * @instance
       * @type {Object}
       * @default null
       */
       drop: null
        /**
         * Allows dropping on the diagram background.
         * @expose
         * @name dnd.drop.background
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         */
        /**
         * An array of MIME data types the Diagram background can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events. 
         * @ojshortdesc Specifies MIME types of objects that can be dropped on the Diagram background.
         * @expose
         * @name dnd.drop.background.dataTypes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string|Array.<string>}
         */
        /**
         * An optional callback function that receives the "dragenter" event and context information as arguments. 
         * The context information is as follows: {@ojinclude "name":"backgroundDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable.
         * @ojshortdesc Optional handler for dragenter event.
         * @expose
         * @name dnd.drop.background.dragEnter
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragover" event and context information as arguments.
         * The context information is as follows: {@ojinclude "name":"backgroundDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable.
         * @ojshortdesc Optional handler for dragover event.
         * @expose
         * @name dnd.drop.background.dragOver
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragleave" event and context information as arguments.
         * The context information is as follows: {@ojinclude "name":"backgroundDropContext"}
         * @ojshortdesc Optional handler for dragleave event.
         * @expose
         * @name dnd.drop.background.dragLeave
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * A callback function that receives the "drop" event and context information as arguments. 
         * The context information is as follows: {@ojinclude "name":"backgroundDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
         * @ojshortdesc Handler for drop event.
         * @expose
         * @name dnd.drop.background.drop
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * Allows dropping on diagram nodes.
         * @expose
         * @name dnd.drop.nodes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         */
        /**
         *  An array of MIME data types the Diagram nodes can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events. 
         * @ojshortdesc Specifies MIME types of objects that can be dropped on the Diagram nodes.
         * @expose
         * @name dnd.drop.nodes.dataTypes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string|Array.<string>}
         */
        /**
         * An optional callback function that receives the "dragenter" event and context information as arguments. 
         * The context information is as follows: {@ojinclude "name":"nodeDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable.
         * @ojshortdesc Optional handler for dragenter event.
         * @expose
         * @name dnd.drop.nodes.dragEnter
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragover" event and context information as arguments.
         * The context information is as follows: {@ojinclude "name":"nodeDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
         * @ojshortdesc Optional handler for dragover event.
         * @expose
         * @name dnd.drop.nodes.dragOver
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragleave" event and context information as arguments.
         * The context information is as follows: {@ojinclude "name":"nodeDropContext"}
         * @ojshortdesc Optional handler for dragleave event.
         * @expose
         * @name dnd.drop.nodes.dragLeave
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * A callback function that receives the "drop" event and context information as arguments. 
         * The context information is as follows: {@ojinclude "name":"nodeDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
         * @ojshortdesc Handler for drop event.
         * @expose
         * @name dnd.drop.nodes.drop
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * Allows dropping on diagram links.
         * @expose
         * @name dnd.drop.links
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         */
        /**
         * An array of MIME data types the Diagram links can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events. 
         * @ojshortdesc Specifies MIME types of objects that can be dropped on the Diagram links.
         * @expose
         * @name dnd.drop.links.dataTypes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string|Array.<string>}
         */
        /**
         * An optional callback function that receives the "dragenter" event and context information as arguments.
         * The context information is as follows: {@ojinclude "name":"linkDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
         * @ojshortdesc Optional handler for dragenter event.
         * @expose
         * @name dnd.drop.links.dragEnter
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragover" event and context information as arguments.
         * The context information is as follows: {@ojinclude "name":"linkDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
         * @ojshortdesc Optional handler for dragover event.
         * @expose
         * @name dnd.drop.links.dragOver
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragleave" event and context information as arguments.
         * The context information is as follows: {@ojinclude "name":"linkDropContext"}
         * @ojshortdesc Optional handler for dragleave event.
         * @expose
         * @name dnd.drop.links.dragLeave
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * A callback function that receives the "drop" event and context information as arguments. 
         * The context information is as follows: {@ojinclude "name":"linkDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
         * @ojshortdesc Handler for drop event.
         * @expose
         * @name dnd.drop.links.drop
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * Allows dropping a link end on a port.
         * @expose
         * @name dnd.drop.ports
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         */
        /**
         * A string, containing a selector expression, that will be used to identify the descendant DOM element in a diagram node that can be used for link creation. This property is requred. 
         * @ojshortdesc Specifies DOM element selector used for link creation end.
         * @expose
         * @name dnd.drop.ports.selector
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         */
        /**
         *  An array of MIME data types the Diagram ports can accept. This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events. 
         * @ojshortdesc Specifies MIME types of objects that can be dropped on the Diagram ports.
         * @expose
         * @name dnd.drop.ports.dataTypes
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string|Array.<string>}
         */
        /**
         * An optional callback function that receives the "dragenter" event and context information as arguments. 
         * The context information is as follows: * The context information is as follows: {@ojinclude "name":"portDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
         * @ojshortdesc Optional handler for dragenter event.
         * @expose
         * @name dnd.drop.ports.dragEnter
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragover" event and context information as arguments.
         * The context information is as follows: * The context information is as follows: {@ojinclude "name":"portDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted. Otherwise, dataTypes will be matched against the drag data types to determine if the data is acceptable. 
         * @ojshortdesc Optional handler for dragover event.
         * @expose
         * @name dnd.drop.ports.dragOver
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * An optional callback function that receives the "dragleave" event and context information as arguments.
         * The context information is as follows: * The context information is as follows: {@ojinclude "name":"portDropContext"}
         * @ojshortdesc Optional handler for dragleave event.
         * @expose
         * @name dnd.drop.ports.dragLeave
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
        /**
         * A callback function that receives the "drop" event and context information as arguments. 
         * The context information is as follows: * The context information is as follows: {@ojinclude "name":"portDropContext"}
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
         * @ojshortdesc Handler for drop event.
         * @expose
         * @name dnd.drop.ports.drop
         * @memberof! oj.ojDiagram
         * @instance
         * @type {function(Event, Object)}
         */
     },

    /**
     * Specifies the key set containing the ids of diagram nodes that should be expanded on initial render.
     * Use the <a href="ExpandedKeySet.html">ExpandedKeySet</a> class to specify nodes to expand.
     * Use the <a href="ExpandAllKeySet.html">ExpandAllKeySet</a> class to expand all nodes. 
     * @ojshortdesc Specifies the key set of ids for expanded diagram nodes.
     * @expose
     * @name expanded
     * @memberof oj.ojDiagram
     * @instance
     * @type {KeySet}
     * @default new ExpandedKeySet()
     * @ojwriteback
     *
     * @example <caption>Initialize the diagram with specific items expanded:</caption>
     * myDiagram.expanded = new ExpandedKeySet(['N0', 'N00']);
     *
     * @example <caption>Initialize the diagram with all items expanded:</caption>
     * myDiagram.expanded = new ExpandAllKeySet();
     */
    expanded: new oj.ExpandedKeySet(),
    /**
     * An array containing the ids of the selected nodes and links.
     * @expose
     * @name selection
     * @memberof oj.ojDiagram
     * @instance
     * @type {Array.<string>}
     * @default null
     * @ojwriteback
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">selection</code> attribute specified:</caption>
     * &lt;oj-diagram selection='["N1", "N2", "L0"]'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">selection</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myDiagram.selection[0];
     * 
     * // Get all
     * var values = myDiagram.selection;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myDiagram.selection=["N1", "N2", "L0"];
     */
    selection: null,
    /**
     * Specifies the selection mode.
     * @expose
     * @name selectionMode
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">selection-mode</code> attribute specified:</caption>
     * &lt;oj-diagram selection-mode='multiple'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.selectionMode;
     * 
     * // setter
     * myDiagram.selectionMode="multiple";
     */
    selectionMode: "none",
    /**
     * Specifies whether panning is allowed in Diagram.
     * @expose
     * @name panning
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">panning</code> attribute specified:</caption>
     * &lt;oj-diagram panning='auto'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">panning</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.panning;
     * 
     * // setter
     * myDiagram.panning="auto";
     */
    panning: "none",
    /**
     * Specifies if panning allowed in horizontal and vertical directions.
     * @expose
     * @name panDirection
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "x"
     * @ojvalue {string} "y"
     * @ojvalue {string} "auto"
     * @default "auto"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">pan-direction</code> attribute specified:</caption>
     * &lt;oj-diagram pan-direction='x'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">panDirection</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.panDirection;
     * 
     * // setter
     * myDiagram.panDirection="y";
     */
    panDirection: "auto",
    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojDiagram
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">tooltip</code> attribute specified:</caption>
     * <!-- Using dot notation -->
     * &lt;oj-diagram tooltip.renderer='[[tooltipFun]]'>&lt;/oj-diagram>
     * 
     * &lt;oj-diagram tooltip='[[{"renderer": tooltipFun}]]'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">tooltip</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myDiagram.tooltip.renderer;
     * 
     * // Get all
     * var values = myDiagram.tooltip;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myDiagram.setProperty('tooltip.renderer', tooltipFun);
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myDiagram.tooltip={'renderer': tooltipFun};     
     */
    tooltip: {
      /**
       * A function that returns a custom tooltip. The function takes a dataContext argument, 
       * provided by the diagram, with the following properties:
       * <ul>
       *   <li>parentElement The tooltip element. The function can directly modify or append content to this element.</li>
       *   {@ojinclude "name":"dataContext"}
       * </ul>
       * @ojshortdesc Specifies a function for a custom tooltip.
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojDiagram
       * @instance
       * @type {function(Object):Object|null}
       * @default null
       */
       renderer: null
     },
    /**
     * Specifies whether zooming is allowed in Diagram.
     * @expose
     * @name zooming
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">zooming</code> attribute specified:</caption>
     * &lt;oj-diagram panDirection='auto'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">zooming</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.zooming;
     * 
     * // setter
     * myDiagram.zooming="auto";
     */
    zooming: "none",
    /**
     * Specifies the minimum zoom level for this diagram. If minZoom is greater than zero, it indicates 
     * the minimum point to which Diagram objects can be scaled. A value of 0.1 implies that 
     * the Diagram can be zoomed out until Nodes appear at one-tenth their natural size.
     * By default, minZoom is set to zoom-to-fit level based on the currently visible Nodes and Links.
     * @ojshortdesc Specifies the minimum zoom level for this diagram.
     * @expose
     * @name minZoom
     * @memberof oj.ojDiagram
     * @instance
     * @type {number}
     * @default 0.0
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">min-zoom</code> attribute specified:</caption>
     * &lt;oj-diagram min-zoom='0.5'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">minZoom</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.minZoom;
     * 
     * // setter
     * myDiagram.minZoom=1.0;
     */
    minZoom: 0.0,
    /**
     * Specifies the maximum zoom level for this diagram. This can be any number greater than zero 
     * which indicates the maximum point to which Diagram objects can be scaled. 
     * A value of 2.0 implies that the Diagram can be zoomed in until Nodes appear at twice their natural size. 
     * By default, maxZoom is 1.0.
     * @ojshortdesc Specifies the maximum zoom level for this diagram.
     * @expose
     * @name maxZoom
     * @memberof oj.ojDiagram
     * @instance
     * @type {number}
     * @default 1.0
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">max-zoom</code> attribute specified:</caption>
     * &lt;oj-diagram max-zoom='10.0'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">maxZoom</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.maxZoom;
     * 
     * // setter
     * myDiagram.maxZoom=5.0;
     */
    maxZoom: 1.0,
    /**
     * An array of category strings used for category filtering. Diagram nodes and links with a category in hiddenCategories will be filtered.
     * @ojshortdesc Specifies categories used for filtering.
     * @expose
     * @name hiddenCategories
     * @memberof oj.ojDiagram
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">hidden-categories</code> attribute specified:</caption>
     * &lt;oj-diagram hidden-categories='["green", "blue"]'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">hiddenCategories</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myDiagram.hiddenCategories[0];
     * 
     * // Get all
     * var values = myDiagram.hiddenCategories;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myDiagram.hiddenCategories=["green", "blue"];
     */
    hiddenCategories: [],
    /**
     * Defines the behavior applied when hovering over diagram nodes and links.
     * @expose
     * @name hoverBehavior
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "dim"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">hover-behavior</code> attribute specified:</caption>
     * &lt;oj-diagram max-zoom='dim'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">hoverBehavior</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.hoverBehavior;
     * 
     * // setter
     * myDiagram.hoverBehavior='dim';
     */
    hoverBehavior: "none",
    /**
     * An array of category strings used for category highlighting. Diagram nodes and links with a category in highlightedCategories will be highlighted.
     * @ojshortdesc Specifies categories used for highlighting.
     * @expose
     * @name highlightedCategories
     * @memberof oj.ojDiagram
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     *
     * 
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">highlighted-categories</code> attribute specified:</caption>
     * &lt;oj-diagram highlighted-categories='["green", "blue"]'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">highlightedCategories</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myDiagram.highlightedCategories[0];
     * 
     * // Get all
     * var values = myDiagram.highlightedCategories;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myDiagram.highlightedCategories=["green", "blue"];
     */
    highlightedCategories: [],
    /**
     * The matching condition for the highlightedCategories option. By default, highlightMatch is 'all' and only items whose categories match all of the values specified in the highlightedCategories array will be highlighted. If highlightMatch is 'any', then items that match at least one of the highlightedCategories values will be highlighted.
     * @ojshortdesc Specifies matching condition used for category highlighting.
     * @expose
     * @name highlightMatch
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "any"
     * @ojvalue {string} "all"
     * @default "all"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">highlight-match</code> attribute specified:</caption>
     * &lt;oj-diagram highlight-match='any'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">highlightMatch</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.highlightMatch;
     * 
     * // setter
     * myDiagram.highlightMatch="any";
     */
    highlightMatch: "all",
    /**
     * Defines node highlighting mode.
     * @expose
     * @name nodeHighlightMode
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "nodeAndIncomingLinks"
     * @ojvalue {string} "nodeAndOutgoingLinks"
     * @ojvalue {string} "nodeAndLinks"
     * @ojvalue {string} "node"
     * @default "node"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">node-highlight-mode</code> attribute specified:</caption>
     * &lt;oj-diagram node-highlight-mode='nodeAndLinks'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">nodeHighlightMode</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.nodeHighlightMode;
     * 
     * // setter
     * myDiagram.nodeHighlightMode="nodeAndLinks";
     */
    nodeHighlightMode: "node",
    /**
     * Defines link highlighting mode.
     * @expose
     * @name linkHighlightMode
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "linkAndNodes"
     * @ojvalue {string} "link"
     * @default "link"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">link-highlight-mode</code> attribute specified:</caption>
     * &lt;oj-diagram link-highlight-mode='linkAndNodes'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">linkHighlightMode</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.linkHighlightMode;
     * 
     * // setter
     * myDiagram.linkHighlightMode="linkAndNodes";
     */
    linkHighlightMode: "link",
    /**
     * A callback function - a custom renderer - that will be used for initial node rendering. 
     * The function takes a single argument, provided by the component, with the following properties:
     * {@ojinclude "name":"rendererContext"}
     *
     * The function should return an Object with the following property: 
     * <ul>
     *   <li>insert: SVGElement - An SVG element, which will be used as content of a Diagram node.</li>
     * </ul>
     *
     * @ojshortdesc Specifies custom renderer for the diagram nodes used for initial rendering.
     * @expose
     * @name renderer
     * @memberof oj.ojDiagram
     * @instance
     * @type {null|function(Object):Object}
     * @default null
     *
     * @example <caption>Initialize the diagram with the <code class="prettyprint">renderer</code> attribute specified:</caption>
     * &lt;oj-diagram renderer='{{rendererFunc}}'>&lt;/oj-diagram>
     *
     * @example <caption>Get or set the <code class="prettyprint">renderer</code> property after initialization:</caption>
     * // getter
     * var renderer = myDiagram.renderer;
     *
     * // setter
     * myDiagram.renderer = rendererFunc;
     */
    renderer: null,
    /**
     * An optional callback function to update the node in response to changes in hover state. 
     * The function takes a single argument, provided by the component, with the following properties:
     * {@ojinclude "name":"rendererContext"}
     *
     * The function should return one of the following: 
     * <ul>
     *   <li>An Object with the following property:
     *     <ul><li>insert: SVGElement - An SVG element, which will be used as content of a Diagram node.</li></ul>
     *   </li>
     *   <li>undefined: Indicates that the existing DOM has been directly modified and no further action is required.</li>
     * </ul>
     *
     * @ojshortdesc Specifies custom renderer for the diagram nodes used for hover updates.
     * @expose
     * @name hoverRenderer
     * @memberof oj.ojDiagram
     * @instance
     * @type {null|function(Object):Object}
     * @default null
     *
     * @example <caption>Initialize the diagram with the <code class="prettyprint">hover-renderer</code> attribute specified:</caption>
     * &lt;oj-diagram hover-renderer='{{rendererFunc}}'>&lt;/oj-diagram>
     *
     * @example <caption>Get or set the <code class="prettyprint">hoverRenderer</code> property after initialization:</caption>
     * // getter
     * var renderer = myDiagram.hoverRenderer;
     *
     * // setter
     * myDiagram.hoverRenderer = rendererFunc;
     */
    hoverRenderer: null,
    /**
     * An optional callback function to update the node in response to changes in selection state. 
     * The function takes a single argument, provided by the component, with the following properties:
     * {@ojinclude "name":"rendererContext"}
     *
     * The function should return one of the following: 
     * <ul>
     *   <li>An Object with the following property:
     *     <ul><li>insert: SVGElement - An SVG element, which will be used as content of a Diagram node.</li></ul>
     *   </li>
     *   <li>undefined: Indicates that the existing DOM has been directly modified and no further action is required.</li>
     * </ul>
     *
     * @ojshortdesc Specifies custom renderer for the diagram nodes used for selection updates.
     * @expose
     * @name selectionRenderer
     * @memberof oj.ojDiagram
     * @instance
     * @type {null|function(Object):Object}
     * @default null
     *
     *
     * @example <caption>Initialize the diagram with the <code class="prettyprint">selection-renderer</code> attribute specified:</caption>
     * &lt;oj-diagram selection-renderer='{{rendererFunc}}'>&lt;/oj-diagram>
     *
     * @example <caption>Get or set the <code class="prettyprint">selectionRenderer</code> property after initialization:</caption>
     * // getter
     * var renderer = myDiagram.selectionRenderer;
     *
     * // setter
     * myDiagram.selectionRenderer = rendererFunc;
     */
    selectionRenderer: null,
    /**
     * An optional callback function to update the node in response to changes in keyboard focus state. 
     * The function takes a single argument, provided by the component, with the following properties:
     * {@ojinclude "name":"rendererContext"}
     *
     * The function should return one of the following: 
     * <ul>
     *   <li>An Object with the following property:
     *     <ul><li>insert: SVGElement - An SVG element, which will be used as content of a Diagram node.</li></ul>
     *   </li>
     *   <li>undefined: Indicates that the existing DOM has been directly modified and no further action is required.</li>
     * </ul>
     *
     * @ojshortdesc Specifies custom renderer for the diagram nodes used for focus updates.
     * @expose
     * @name focusRenderer
     * @memberof oj.ojDiagram
     * @instance
     * @type {null|function(Object):Object}
     * @default null
     *
     * @example <caption>Initialize the diagram with the <code class="prettyprint">focus-renderer</code> attribute specified:</caption>
     * &lt;oj-diagram focus-renderer='{{rendererFunc}}'>&lt;/oj-diagram>
     *
     * @example <caption>Get or set the <code class="prettyprint">focusRenderer</code> property after initialization:</caption>
     * // getter
     * var renderer = myDiagram.focusRenderer;
     *
     * // setter
     * myDiagram.focusRenderer = rendererFunc;
     */
    focusRenderer: null,
    /**
     * An optional callback function to update the node in response to changes in zoom level. 
     * The function takes a single argument, provided by the component, with the following properties:
     * {@ojinclude "name":"rendererContext"}
     *
     * The function should return one of the following: 
     * <ul>
     *   <li>An Object with the following property:
     *     <ul><li>insert: SVGElement - An SVG element, which will be used as content of a Diagram node.</li></ul>
     *   </li>
     *   <li>undefined: Indicates that the existing DOM has been directly modified and no further action is required.</li>
     * </ul>
     *
     * @ojshortdesc Specifies custom renderer for the diagram nodes used for zoom updates.
     * @expose
     * @name zoomRenderer
     * @memberof oj.ojDiagram
     * @instance
     * @type {null|function(Object):Object}
     * @default null
     *
     * @example <caption>Initialize the diagram with the <code class="prettyprint">zoom-renderer</code> attribute specified:</caption>
     * &lt;oj-diagram zoom-renderer='{{rendererFunc}}'>&lt;/oj-diagram>
     *
     * @example <caption>Get or set the <code class="prettyprint">zoomRenderer</code> property after initialization:</caption>
     * // getter
     * var renderer = myDiagram.zoomRenderer;
     *
     * // setter
     * myDiagram.zoomRenderer = rendererFunc;
     */
    zoomRenderer: null,
    /**
     * The data source for the Diagram element. See <a href="oj.DiagramDataSource.html">oj.DiagramDataSource</a for details> 
     * @ojshortdesc Specifies data source for the component.
     * @expose
     * @name data
     * @memberof oj.ojDiagram
     * @instance
     * @type {Object}
     * @default null
     *
     * @example <caption>Initialize the diagram with the <code class="prettyprint">data</code> attribute specified:</caption>
     * &lt;oj-diagram data='{{myDataSource}}'>&lt;/oj-diagram>
     *
     * @example <caption>Get or set the <code class="prettyprint">data</code> property after initialization:</caption>
     * // getter
     * var dataValue = myDiagram.data;
     *
     * // setter
     * myDiagram.data = myDataSource;

     */
    data: null,
    /**
     * Optional callback that provides a way to customize an appearance of the specific link, the function maps link data into link styles.
     * The function takes a data object for the link provided by the diagram.
     * The following properties are supported on the return object:
     * @property {object} labelStyle The CSS style object defining the style of the link label. The CSS max-width property can be used to truncate labels.
     * @property {string} color Link color.
     * @property {object} svgStyle The SVG CSS style object defining link style. The style class and style object will be applied directly on the link and override any other styling specified through the properties.
     * @property {string} svgClassName The SVG CSS style class defining link style. The style class and style object will be applied directly on the link and override any other styling specified through the properties.
     * @property {number} width Link width in pixels.
     * @property {string} startConnectorType Specifies the type of start connector on the link. <br/>Supported values are "arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none". <br/>Default value is <code class="prettyprint">"none"</code>.
     * @property {string} endConnectorType Specifies the type of end connector on the link. <br/>Supported values are "arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none". <br/>Default value is <code class="prettyprint">"none"</code>.
     * 
     * @example <caption>Customizing link color using <code class="prettyprint">customColor</code> property defined on the link data object</caption>
     * &lt;oj-diagram
     *    layout = '{{layoutFunc}}'
     *    data = '{{dataSource}}'
     *    node-properties = '[[function(data){return {color:data.customColor};}]]'>
     * &lt;/oj-diagram>
     * 
     * @ojshortdesc Optional callback for customizing link appearance based on link data.
     * @expose
     * @name linkProperties
     * @memberof oj.ojDiagram
     * @instance
     * @type {null|function(Object):Object}
     * @default null
     */
    linkProperties: null,
    /**
     * Optional callback that provides a way to customize an appearance of the specific node, the function maps node data into node styles.
     * The function takes a data object for the node provided by the diagram. 
     * The following properties are supported on the return object:
     * @property {object} labelStyle The CSS style object defining the style of the link label. The CSS max-width property can be used to truncate labels.
     * @property {string} color Link color.
     * @property {object} svgStyle The SVG CSS style object defining link style. The style class and style object will be applied directly on the link and override any other styling specified through the properties.
     * @property {string} svgClassName The SVG CSS style class defining link style. The style class and style object will be applied directly on the link and override any other styling specified through the properties.
     * @property {number} width Link width in pixels.
     * @property {string} startConnectorType Specifies the type of start connector on the link. <br/>Supported values are "arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none". <br/>Default value is <code class="prettyprint">"none"</code>.
     * @property {string} endConnectorType Specifies the type of end connector on the link. <br/>Supported values are "arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none". <br/>Default value is <code class="prettyprint">"none"</code>.
     * 
     * @example <caption>Customizing node icon color using <code class="prettyprint">customColor</code> property defined on the node data object</caption>
     * &lt;oj-diagram
     *    layout = '{{layoutFunc}}'
     *    data = '{{dataSource}}'
     *    node-properties = '[[function(data){return {icon:{color:data.customColor}};}]]'>
     * &lt;/oj-diagram>
     * 
     * @ojshortdesc Optional callback for customizing node appearance based on node data.
     * @expose
     * @name nodeProperties
     * @memberof oj.ojDiagram
     * @instance
     * @type {null|function(Object):Object}
     * @default null
     */
    nodeProperties: null,
    /**
     * Defines promoted link behavior for the component. <br/> If the value is set to none, the diagram will not retrieve additional data to resolve promoted links and will not display promoted links.<br/> If the value is set to lazy, the diagram will retrieve additional data to resolve promoted links if the data is already available and will display available promoted links. The component will not retrieve additional data if the data is not available yet. <br/> If the value is set to full, the diagram will retrieve additional data to resolve all promoted links and will display promoted links. 
     *
     * @ojshortdesc Defines promoted link behavior for the component.
     * @expose
     * @name promotedLinkBehavior
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "full"
     * @ojvalue {string} "lazy"
     * @default "lazy"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">promoted-link-behavior</code> attribute specified:</caption>
     * &lt;oj-diagram promoted-link-behavior='none'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">promotedLinkBehavior</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.promotedLinkBehavior;
     * 
     * // setter
     * myDiagram.promotedLinkBehavior="none";
     */
    promotedLinkBehavior: "lazy",
    /**
     * An object defining the style defaults for this diagram.
     * @expose
     * @name styleDefaults
     * @memberof oj.ojDiagram
     * @instance
     * @type {Object}
     * @default {"hoverBehaviorDelay":200, "nodeDefaults":{"labelStyle":{}, "showDisclosure":"on", "icon":{"pattern":"none", "shape":"circle", "width":10, "height":10, "svgClassName":""}}, "linkDefaults":{"width":1.0, "labelStyle":{}, "startConnectorType":"none", "endConnectorType":"none", "svgClassName":""}, "promotedLink":{"width":1.0, "startConnectorType":"none", "endConnectorType":"none", "svgClassName":""}}
     *
     * @example <caption>Get or set the <code class="prettyprint">styleDefaults</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myDiagram.styleDefaults.animationDuration;
     * 
     * // Get all
     * var values = myDiagram.styleDefaults;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myDiagram.setProperty('styleDefaults.nodeDefaults', {'icon': {width: 125, height: 125}});
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myDiagram.styleDefaults={'nodeDefaults': {'icon': {width: 125, height: 125}}};
     */
    styleDefaults: {
      /**
       * We recommend using the component CSS classes to set component wide styling. This API should be used
       * only for styling a specific instance of the component. The default values come from the CSS classess and 
       * varies based on theme. The duration of the animations in milliseconds.
       * @ojshortdesc Defines animation duration in milliseconds.
       * @expose
       * @name styleDefaults.animationDuration
       * @memberof! oj.ojDiagram
       * @instance
       * @type {number}
       * @ojunits milliseconds
       */
      /**
       * Specifies initial hover delay in ms for highlighting data items.
       * @expose
       * @name styleDefaults.hoverBehaviorDelay
       * @memberof! oj.ojDiagram
       * @instance
       * @type {number}
       * @default 200
       * @ojunits milliseconds
       */
       hoverBehaviorDelay: 200,
      /**
       * Default node styles
       * @expose
       * @name styleDefaults.nodeDefaults
       * @memberof! oj.ojDiagram
       * @instance
       * @type {Object}
       */
       nodeDefaults: {
        /**
         * The CSS style object defining the style of the node label. Supports color, 
         * fontFamily, fontSize, fontStyle, fontWeight, textDecoration, cursor, maxWidth,
         * backgroundColor, borderColor, borderRadius, and borderWidth properties.
         * @ojshortdesc Specifies CSS styles for the node label.
         * @expose
         * @name styleDefaults.nodeDefaults.labelStyle
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         * @default {}
         */
         labelStyle: {},
        /**
         * Determines when to display the expand/collapse button.
         * @expose
         * @name styleDefaults.nodeDefaults.showDisclosure
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         * @ojvalue {string} "off"
         * @ojvalue {string} "on"
         * @default "on"
         */
         showDisclosure: "on",
        /**
         * Default style for the node icon.
         * @expose
         * @name styleDefaults.nodeDefaults.icon
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         * @default {"width": 10, "height": 10, "fillPattern": "none", "shape": "circle"}
         */
         icon: {
          /**
           * Default border color of the icon.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.borderColor
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           */
          /**
           * The default border radius of the icon. CSS border-radius values accepted. Note that non-% values (including unitless) get interpreted as 'px'.
           * @ojshortdesc Specifies default border radius of the icon.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.borderRadius
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           */
          /**
           * Default border width of the icon in pixels.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.borderWidth
           * @memberof! oj.ojDiagram
           * @instance
           * @type {number}
           * @ojunits pixels
           */
          /**
           * Default color of the icon.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.color
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           */
          /**
           * Default fill pattern of the icon.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.pattern
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           * @ojvalue {string} "smallChecker"
           * @ojvalue {string} "smallCrosshatch"
           * @ojvalue {string} "smallDiagonalLeft"
           * @ojvalue {string} "smallDiagonalRight"
           * @ojvalue {string} "smallDiamond"
           * @ojvalue {string} "smallTriangle"
           * @ojvalue {string} "largeChecker"
           * @ojvalue {string} "largeCrosshatch"
           * @ojvalue {string} "largeDiagonalLeft"
           * @ojvalue {string} "largeDiagonalRight"
           * @ojvalue {string} "largeDiamond"
           * @ojvalue {string} "largeTriangle"
           * @ojvalue {string} "none"
           * @default "none"
           */
           pattern: "none",
          /**
           * Default shape of the icon. Can take the name of a built-in shape or the svg path commands for a custom shape.
           * @ojshortdesc Specifies default shape of the icon.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.shape
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           * @ojvalue {string} "ellipse"
           * @ojvalue {string} "square"
           * @ojvalue {string} "plus"
           * @ojvalue {string} "diamond"
           * @ojvalue {string} "triangleUp"
           * @ojvalue {string} "triangleDown"
           * @ojvalue {string} "human"
           * @ojvalue {string} "rectangle"
           * @ojvalue {string} "star"
           * @ojvalue {string} "circle"
           * @default "circle"
           */
           shape: "circle",
          /**
           * The URI of the node image
           * @expose
           * @name styleDefaults.nodeDefaults.icon.source
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           */
          /**
           * The optional URI of the node hover image. If not defined, the source image will be used.
           * @ojshortdesc The optional URI of the node hover image.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.sourceHover
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           */
          /**
           * The optional URI of the selected image on hover. If not defined, the sourceSelected image will be used. If the sourceSelected image is not defined, the source image will be used.
           * @ojshortdesc The optional URI of the selected image on hover.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.sourceHoverSelected
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           */
          /**
           * The optional URI of the selected image. If not defined, the source image will be used.
           * @ojshortdesc The optional URI of the selected image.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.sourceSelected
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           */
          /**
           * Default icon width.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.width
           * @memberof! oj.ojDiagram
           * @instance
           * @type {number}
           * @default 10
           * @ojunits pixels
           */
           width: 10,
          /**
           * Default icon height.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.height
           * @memberof! oj.ojDiagram
           * @instance
           * @type {number}
           * @default 10
           * @ojunits pixels
           */
           height: 10,
          /**
           * The default SVG CSS style object defining the style of the icon.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.svgStyle
           * @memberof! oj.ojDiagram
           * @instance
           * @type {Object}
           * @default {}
           */
          /**
           * The SVG CSS style class to apply to the node icon.
           * @expose
           * @name styleDefaults.nodeDefaults.icon.svgClassName
           * @memberof! oj.ojDiagram
           * @instance
           * @type {string}
           * @default ""
           */
           svgClassName: ""
         }
       },
      /**
       * Default link styles
       * @expose
       * @name styleDefaults.linkDefaults
       * @memberof! oj.ojDiagram
       * @instance
       * @type {Object}
       * @default {"labelStyle: {}, ""startConnectorType": "none", "endConnectorType": "none", "width": 1.0}
       */
       linkDefaults: {
        /**
         * Default link color. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name styleDefaults.linkDefaults.color
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         */
        /**
         * The default style object represents the SVG CSS style of the link. User defined custom SVG CSS Styles will be applied directly on the link. 
         * @ojshortdesc Specifies SVG CSS styles for the link.
         * @expose
         * @name styleDefaults.linkDefaults.svgStyle
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         * @default {}
         */
        /**
         * The default SVG CSS style class to apply to the link.
         * @expose
         * @name styleDefaults.linkDefaults.svgClassName
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         * @default ""
         */
         svgClassName: "",
        /**
         * Default link width in pixels.
         * @expose
         * @name styleDefaults.linkDefaults.width
         * @memberof! oj.ojDiagram
         * @instance
         * @type {number}
         * @ojunits pixels
         */
         width: 1.0,
        /**
         * The CSS style object defining the style of the link label. Supports color, 
         * fontFamily, fontSize, fontStyle, fontWeight, textDecoration, cursor, maxWidth,
         * backgroundColor, borderColor, borderRadius, and borderWidth properties.
         * @ojshortdesc Specifies CSS styles for the link label.
         * @expose
         * @name styleDefaults.linkDefaults.labelStyle
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         */
         labelStyle: {},
        /**
         * Specifies the type of start connector on the link.
         * @expose
         * @name styleDefaults.linkDefaults.startConnectorType
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         * @ojvalue {string} "arrowOpen"
         * @ojvalue {string} "arrow"
         * @ojvalue {string} "arrowConcave"
         * @ojvalue {string} "circle"
         * @ojvalue {string} "rectangle"
         * @ojvalue {string} "rectangleRounded"
         * @ojvalue {string} "none"
         */
         startConnectorType: "none",
        /**
         * Specifies the type of end connector on the link.
         * @expose
         * @name styleDefaults.linkDefaults.endConnectorType
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         * @ojvalue {string} "arrowOpen"
         * @ojvalue {string} "arrow"
         * @ojvalue {string} "arrowConcave"
         * @ojvalue {string} "circle"
         * @ojvalue {string} "rectangle"
         * @ojvalue {string} "rectangleRounded"
         * @ojvalue {string} "none"
         */
         endConnectorType: "none"
       },
      /**
       * Promoted link styles
       * @expose
       * @name styleDefaults.promotedLink
       * @memberof! oj.ojDiagram
       * @instance
       * @type {Object}
       * @default {""startConnectorType": "none", "endConnectorType": "none", "width": 1.0}
       */
       promotedLink: {
        /**
         * Default promoted link color. The default value varies based on theme.
         * @expose
         * @name styleDefaults.promotedLink.color
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         */
         color: "#778999",
        /**
         * The promoted style object represents the CSS style of the link. User defined custom CSS Styles will be applied directly on the link. 
         * @expose
         * @name styleDefaults.promotedLink.svgStyle
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         * @default {}
         */
        /**
         * The SVG CSS style class to apply to the promoted link. 
         * @expose
         * @name styleDefaults.promotedLink.svgClassName
         * @memberof! oj.ojDiagram
         * @instance
         * @type {Object}
         * @default ""
         */
         svgClassName: "",
        /**
         * Default link width in pixels.
         * @expose
         * @name styleDefaults.promotedLink.width
         * @memberof! oj.ojDiagram
         * @instance
         * @type {number}
         * @ojunits pixels
         */
         width: 1.0,
        /**
         * Specifies the type of start connector on the promoted link.
         * @expose
         * @name styleDefaults.promotedLink.startConnectorType
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         * @ojvalue {string} "arrowOpen"
         * @ojvalue {string} "arrow"
         * @ojvalue {string} "arrowConcave"
         * @ojvalue {string} "circle"
         * @ojvalue {string} "rectangle"
         * @ojvalue {string} "rectangleRounded"
         * @ojvalue {string} "none"
         */
         startConnectorType: "none",
        /**
         * Specifies the type of end connector on the promoted link.
         * @expose
         * @name styleDefaults.promotedLink.endConnectorType
         * @memberof! oj.ojDiagram
         * @instance
         * @type {string}
         * @ojvalue {string} "arrowOpen"
         * @ojvalue {string} "arrow"
         * @ojvalue {string} "arrowConcave"
         * @ojvalue {string} "circle"
         * @ojvalue {string} "rectangle"
         * @ojvalue {string} "rectangleRounded"
         * @ojvalue {string} "none"
         */
         endConnectorType: "none"
       }
    },
    /**
     * Data visualizations require a press and hold delay before triggering tooltips and rollover effects on mobile devices to avoid interfering with page panning, but these hold delays can make applications seem slower and less responsive. For a better user experience, the application can remove the touch and hold delay when data visualizations are used within a non scrolling container or if there is sufficient space outside of the visualization for panning. If touchResponse is touchStart the component will instantly trigger the touch gesture and consume the page pan events if the component does not require an internal feature that requires a touch start gesture like panning or zooming. If touchResponse is auto, the component will behave like touchStart if it determines that it is not rendered within scrolling content and if component panning is not available for those components that support the feature. 
     *
     * @ojshortdesc Specifies touch response behavior.
     * @expose
     * @name touchResponse
     * @memberof oj.ojDiagram
     * @instance
     * @type {string}
     * @ojvalue {string} "touchStart"
     * @ojvalue {string} "auto"
     * @default "auto"
     *
     * @example <caption>Initialize the diagram with the 
     * <code class="prettyprint">touch-response</code> attribute specified:</caption>
     * &lt;oj-diagram touch-response='touchStart'>&lt;/oj-diagram>
     * 
     * @example <caption>Get or set the <code class="prettyprint">touchResponse</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myDiagram.touchResponse;
     * 
     * // setter
     * myDiagram.touchResponse="touchStart";
     */
     touchResponse: "auto",
    /**
     * A custom JavaScript client layout method - a custom code developed by a customer used to position Diagram nodes and links. The layout code must conform to the pluggable layout contract.
     * @ojshortdesc Specifies layout callback used to position nodes and links.
     * @expose
     * @name layout
     * @memberof oj.ojDiagram
     * @instance
     * @type {function(DvtDiagramLayoutContext)}
     * @see <a href="oj.DvtDiagramLayoutContext.html">DvtDiagramLayoutContext</a>
     * @see <a href="oj.DvtDiagramLayoutContextLink.html">DvtDiagramLayoutContextLink</a>
     * @see <a href="oj.DvtDiagramLayoutContextNode.html">DvtDiagramLayoutContextNode</a>
     */
    /**
     * Triggered immediately before any container node in the diagram is expanded.
     *
     * @ojshortdesc Event handler for when a node is about to expand.
     * @ojcancelable
     * @expose
     * @property {string} nodeId the id of the expanding object
     * @event
     * @memberof oj.ojDiagram
     * @instance
     */
    beforeExpand: null,
    /**
     * Triggered when a node has been expanded. The ui object contains one property, "nodeId", which is the id of the node that has been expanded.
     *
     * @ojshortdesc Event handler for after a node has expanded.
     * @expose
     * @property {string} nodeId the id of the expanded object 
     * @event
     * @memberof oj.ojDiagram
     * @instance
     */    
    expand: null,
    /**
     * Triggered immediately before any container node in the diagram is collapsed.
     *
     * @ojshortdesc Event handler for when a node is about to collapse.
     * @ojcancelable
     * @expose
     * @property {string} data.nodeId the id of the collapsing object 
     * @event
     * @memberof oj.ojDiagram
     * @instance
     */
    beforeCollapse: null,
    /**
     * Triggered when a node has been collapsed.
     *
     * @ojshortdesc Event handler for after a node has collapsed.
     * @expose
     * @property {string} data.nodeId the id of the collapsed object 
     * @event
     * @memberof oj.ojDiagram
     * @instance
     */
    collapse: null
  },

  //** @inheritdoc */
  _InitOptions : function(originalDefaults, constructorOptions) {
    this._super(originalDefaults, constructorOptions);

    // styleDefaults subproperty defaults are dynamically generated
    // so we need to retrieve it here and override the dynamic getter by
    // setting the returned object as the new value.
    var styleDefaults = this.options["styleDefaults"];
    this.options["styleDefaults"] = styleDefaults;
  },

  //** @inheritdoc */
  _ProcessOptions: function() {
    this._super();
    this.options['_logger'] = oj.Logger;
    if (this.options['_templateFunction']) {
      this.options['renderer'] = this._GetTemplateDataRenderer(this.options['_templateFunction'], 'node');
    }
    if (this.options['renderer']) {
      this.options['_contextHandler'] = this._getContextHandler();
    }
    //convert nodes, links and childNodes options to DiagramDataSource
    if (this.options['nodes']) {
      this.options['nodeProperties'] = this.options['nodeProperties'] ? 
              this.options['nodeProperties'] : function(data){return data};
      this.options['linkProperties'] = this.options['linkProperties'] ? 
              this.options['linkProperties'] : function(data){return data};
      this.options['data'] = 
        new oj.ConversionDiagramDataSource(
          {'nodes': this.options['nodes'],'links': this.options['links']},
          {'childData': this.options['childNodes']});
    }
    // if expanded not declared, pass default empty expanded key set to the toolkit
    if (!this.options['expanded'])
      this.options['expanded'] = new oj.ExpandedKeySet();
    if (!this.options['dnd']['drag']){
      this.options['dnd']['drag'] = {
        'nodes':{}, 'ports':{}
      };
    }
    if (!this.options['dnd']['drop']){
      this.options['dnd']['drop'] = {
        'background':{}, 'nodes':{}, 'links':{}, 'ports':{}
      };
    }
  },
  
  //** @inheritdoc */
  _IsDraggable: function() {
    var dragObj = this.options['dnd'] ? this.options['dnd']['drag'] : null;
    if (!dragObj)
      return false;
    return (dragObj['nodes'] && Object.keys(dragObj['nodes']).length > 0) || 
           (dragObj['ports'] && Object.keys(dragObj['ports']).length > 0);
  },
  
  //** @inheritdoc */
  _GetComponentRendererOptions: function() {
    return ['tooltip/renderer','renderer','focusRenderer','hoverRenderer','selectionRenderer','zoomRenderer'];
  },
 
  /**
   * Creates a callback function that will be used by DvtDiagramNode to populate context for the custom renderer
   * @return {Function} context handler callback used to create context for a custom renderer
   * @private
   * @instance
   * @memberof oj.ojDiagram
   */
  _getContextHandler: function() {
    var thisRef = this;
    var contextHandlerFunc = function (parentElement, rootElement, childContent, data, state, previousState) {
      var context = {
        'component': oj.Components.__GetWidgetConstructor(thisRef.element),
        'parentElement': parentElement,
        'rootElement': rootElement,
        'content' : childContent,
        'data': data,
        'state': state,
        'previousState' : previousState,
        'id' : data['id'],
        'type' :  'node',
        'label' : data ['label']
      };
      if (thisRef._IsCustomElement()) {
        context['renderDefaultHover'] = thisRef.renderDefaultHover.bind(thisRef, context);
        context['renderDefaultSelection'] = thisRef.renderDefaultSelection.bind(thisRef, context);
        context['renderDefaultFocus'] = thisRef.renderDefaultFocus.bind(thisRef, context);
      }
      return thisRef._FixRendererContext(context);
    }
    return contextHandlerFunc;
  },

  /**
   * Renders default hover effect for the diagram node
   * @param {Object} context - property object with the following fields
   * <ul>
   *  <li>{Element} componentElement - Diagram element</li>
   *  <li>{Object} data - a data object for the node</li>
   *  <li>{SVGElement} parentElement  - a parent group element that takes a custom SVG fragment as the node content. Used for measurements and reading properties.
   *                Modifications of the parentElement are not supported</li>
   *  <li>{SVGElement} rootElement  - an SVG fragment created as a node content passed for subsequent modifications</li>
   *  <li>{Object} state  - property object with the following boolean properties: hovered, selected, focused, zoom</li>
   *  <li>{Object} previousState  - property object with the following boolean properties: hovered, selected, focused, zoom</li>
   *  <li>{string} id - node id</li>
   *  <li>{string} type - object type - node</li>
   *  <li>{string} label - object label</li>
   * </ul>
   * @expose
   * @ignore
   * @instance
   * @memberof oj.ojDiagram
   */
  renderDefaultHover: function (context) {
    if (!context['previousState'] || context['state']['hovered'] != context['previousState']['hovered']) {
      var comp = this._GetDvtComponent(this.element);
      comp.processDefaultHoverEffect(context['id'], context['state']['hovered']);
    }
  },

  /**
   * Renders default selection effect for the diagram node
   * @param {Object} context - property object with the following fields
   * <ul>
   *  <li>{Element} componentElement - Diagram element</li>
   *  <li>{Object} data - a data object for the node</li>
   *  <li>{SVGElement} parentElement  - a parent group element that takes a custom SVG fragment as the node content. Used for measurements and reading properties.
   *                Modifications of the parentElement are not supported</li>
   *  <li>{SVGElement} rootElement  - an SVG fragment created as a node content passed for subsequent modifications</li>
   *  <li>{Object} state  - property object with the following boolean properties: hovered, selected, focused, zoom</li>
   *  <li>{Object} previousState  - property object with the following boolean properties: hovered, selected, focused, zoom</li>
   *  <li>{string} id - node id</li>
   *  <li>{string} type - object type - node</li>
   *  <li>{string} label - object label</li>
   * </ul>
   * @expose
   * @ignore
   * @instance
   * @memberof oj.ojDiagram
   */
  renderDefaultSelection: function (context) {
    if (!context['previousState'] || context['state']['selected'] != context['previousState']['selected']) {
      var comp = this._GetDvtComponent(this.element);
      comp.processDefaultSelectionEffect(context['id'], context['state']['selected']);
    }
  },

  /**
   * Renders default focus effect for the diagram node
   * @param {Object} context - property object with the following fields
   * <ul>
   *  <li>{Element} componentElement - Diagram element</li>
   *  <li>{Object} data - a data object for the node</li>
   *  <li>{SVGElement} parentElement  - a parent group element that takes a custom SVG fragment as the node content. Used for measurements and reading properties.
   *                Modifications of the parentElement are not supported</li>
   *  <li>{SVGElement} rootElement  - an SVG fragment created as a node content passed for subsequent modifications</li>
   *  <li>{Object} state  - property object with the following boolean properties: hovered, selected, focused, zoom</li>
   *  <li>{Object} previousState  - property object with the following boolean properties: hovered, selected, focused, zoom</li>
   *  <li>{string} id - node id</li>
   *  <li>{string} type - object type - node</li>
   *  <li>{string} label - object label</li>
   * </ul>
   * @expose
   * @ignore
   * @instance
   * @memberof oj.ojDiagram
   */
  renderDefaultFocus: function (context) {
    if (!context['previousState'] || context['state']['focused'] != context['previousState']['focused']) {
      var comp = this._GetDvtComponent(this.element);
      comp.processDefaultFocusEffect(context['id'], context['state']['focused']);
    }
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    return dvt.Diagram.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-diagram-link') {
      // link[index]
      subId = 'link[' + locator['index'] + ']';
    }
    else if(subId == 'oj-diagram-node') {
      // node[index]
      subId = 'node[' + locator['index'] + ']';
    }
    else if(subId == 'oj-diagram-tooltip') {
      subId = 'tooltip';
    }

    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId.indexOf('link') == 0) {
      // link[index]
      locator['subId'] = 'oj-diagram-link';
      locator['index'] = this._GetFirstIndex(subId);
    }
    else if(subId.indexOf('node') == 0) {
      // node[index]
      locator['subId'] = 'oj-diagram-node';
      locator['index'] = this._GetFirstIndex(subId);
    }
    else if(subId == 'tooltip') {
      locator['subId'] = 'oj-diagram-tooltip';
    }

    return locator;
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-diagram');
    return styleClasses;
  },

  //** @inheritdoc */
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-dvtbase oj-diagram'] = {'path': 'styleDefaults/animationDuration', 'property': 'ANIM_DUR'};
    styleClasses['oj-diagram-node-label'] = {'path': 'styleDefaults/nodeDefaults/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-diagram-node oj-selected'] = {'path': 'styleDefaults/nodeDefaults/selectionColor', 'property': 'border-color'};
    styleClasses['oj-diagram-node oj-hover'] = [
      {'path': 'styleDefaults/nodeDefaults/hoverOuterColor', 'property': 'border-top-color'},
      {'path': 'styleDefaults/nodeDefaults/hoverInnerColor', 'property': 'border-bottom-color'}
    ];
    styleClasses['oj-diagram-link'] = {'path': 'styleDefaults/linkDefaults/color', 'property': 'color'};
    styleClasses['oj-diagram-link-label'] = {'path': 'styleDefaults/linkDefaults/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-diagram-link oj-selected'] = {'path': 'styleDefaults/linkDefaults/selectionColor', 'property': 'border-color'};
    styleClasses['oj-diagram-link oj-hover'] = [
      {'path': 'styleDefaults/linkDefaults/hoverOuterColor', 'property': 'border-top-color'},
      {'path': 'styleDefaults/linkDefaults/hoverInnerColor', 'property': 'border-bottom-color'}
    ];
    return styleClasses;
  },

  //** @inheritdoc */
  _GetEventTypes : function() {
    return ['optionChange', 'beforeExpand', 'beforeCollapse', 'expand', 'collapse'];
  },
  
  //** @inheritdoc */
  _HandleEvent: function(event) {
    var type = event['type'];
    if (type === 'beforeExpand') {
      this.expand(event['id'], true);
    }    
    else if (type === 'beforeCollapse') {
      this.collapse(event['id'], true);
    }
    else if (type === 'expand' || type === 'collapse') {
      this._trigger(type, null, {'nodeId': event['id']});
    }
    else {
      this._super(event);
    }
  },

  //** @inheritdoc */
  _setOptions : function(options, flags) {
    if (options['expanded']) {
      this._component.clearDisclosedState();
    }
    // Call the super to update the property values
    this._superApply(arguments);
  },

  //** @inheritdoc */
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtDiagramBundle.PROMOTED_LINK'] = translations['promotedLink'];
    ret['DvtDiagramBundle.PROMOTED_LINKS'] = translations['promotedLinks'];
    ret['DvtDiagramBundle.PROMOTED_LINK_ARIA_DESC'] = translations['promotedLinkAriaDesc'];    
    ret['DvtUtilBundle.DIAGRAM'] = translations['componentName'];
    return ret;
  },

  //** @inheritdoc */
  _LoadResources: function() {
    // Ensure the resources object exists
    if (this.options['_resources'] == null)
      this.options['_resources'] = {};

    var resources = this.options['_resources'];
    if (oj.DomUtils.getReadingDirection() === "rtl") {
      resources['collapse_ena'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-collapse-button-ena_rtl.svg'), 'width':20, 'height':20};
      resources['collapse_ovr'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-collapse-button-ovr_rtl.svg'), 'width':20, 'height':20};
      resources['collapse_dwn'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-collapse-button-dwn_rtl.svg'), 'width':20, 'height':20};
      resources['expand_ena'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-expand-button-ena_rtl.svg'), 'width':20, 'height':20};
      resources['expand_ovr'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-expand-button-ovr_rtl.svg'), 'width':20, 'height':20};
      resources['expand_dwn'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-expand-button-dwn_rtl.svg'), 'width':20, 'height':20};      
    }
    else { //ltr
      resources['collapse_ena'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-collapse-button-ena.svg'), 'width':20, 'height':20};
      resources['collapse_ovr'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-collapse-button-ovr.svg'), 'width':20, 'height':20};
      resources['collapse_dwn'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-collapse-button-dwn.svg'), 'width':20, 'height':20};
      resources['expand_ena'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-expand-button-ena.svg'), 'width':20, 'height':20};
      resources['expand_ovr'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-expand-button-ovr.svg'), 'width':20, 'height':20};
      resources['expand_dwn'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/diagram/container-expand-button-dwn.svg'), 'width':20, 'height':20};      
    }
  },

  //** @inheritdoc */
  _GetComponentNoClonePaths : function() {
    var noClonePaths = this._super();
    noClonePaths['data'] = true;
    return noClonePaths;
  },
  
  /**
   * Collapses an expanded node. When vetoable is set to false, beforeExpand event will still be fired but the event cannot be veto.
   * @param {String} nodeId The id of the node to collapse
   * @param {boolean} vetoable Whether the event should be vetoable
   * @ignore
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   */
  collapse: function(nodeId, vetoable) {
    var result = this._trigger("beforeCollapse", null, {'nodeId': nodeId});
    if (!vetoable || result !== false) {
      this._NotReady();
      this._component.collapse(nodeId);
    }
  },

  /**
   * Expands a collapsed parent node. When vetoable is set to false, beforeExpand event will still be fired but the event cannot be veto.
   * @param {String} nodeId The id of the node to expand
   * @param {boolean} vetoable Whether the event should be vetoable
   * @ignore
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   */
  expand: function(nodeId, vetoable) {
    var result = this._trigger("beforeExpand", null, {'nodeId': nodeId});
    if (!vetoable || result !== false) {
      this._NotReady();
      this._component.expand(nodeId);
    }
  },

  /**
   * Returns number of diagram nodes
   * @return {Number} The number of nodes
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   *
   * @example <caption>Invoke the <code class="prettyprint">getNodeCount</code> method:</caption>
   * var count = myDiagram.getNodeCount();
   */
  getNodeCount: function() {
    var auto = this._component.getAutomation();
    return auto.getNodeCount();
  },

  /**
   * Returns an object with the following properties for automation testing verification of the diagram node at the
   * specified index.
   *
   * @param {String} nodeIndex Node index
   * @property {Object|null} icon The icon for the node, or null if none exists.
   * @property {string} icon.color The color of the icon
   * @property {string} icon.shape The shape of the icon
   * @property {string} label Node label
   * @property {boolean} selected The selected state of the node
   * @property {string} tooltip Node tooltip
   * @return {Object|null} An object containing properties for the node at the given index, or null if none exists.
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   *
   * @example <caption>Invoke the <code class="prettyprint">getNode</code> method:</caption>
   * var node = myDiagram.getNode(3);
   */
  getNode: function(nodeIndex) {
    var auto = this._component.getAutomation();
    return auto.getNode(nodeIndex);
  },

  /**
   * Returns number of diagram links
   * @return {Number} The number of links
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   *
   * @example <caption>Invoke the <code class="prettyprint">getLinkCount</code> method:</caption>
   * var count = myDiagram.getLinkCount();
   */
  getLinkCount: function() {
    var auto = this._component.getAutomation();
    return auto.getLinkCount();
  },

  /**
   * Returns an object with the following properties for automation testing verification of the diagram link at the
   * specified index.
   *
   * @param {number} linkIndex Link index
   * @property {string} color Link color
   * @property {string} label Link label
   * @property {string} endConnectorType The type of end connector on the link
   * @property {string} endNode The id of the end node.
   * @property {boolean} selected The selected state of the link
   * @property {string} startConnectorType The type of start connector on the link
   * @property {string} startNode The id of the start node.
   * @property {string} style Link style
   * @property {string} tooltip Link tooltip
   * @property {number} width Link width
   * @return {Object|null} An object containing properties for the link at the given index, or null if none exists.
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   *
   * @example <caption>Invoke the <code class="prettyprint">getLink</code> method:</caption>
   * var link = myDiagram.getLink(3);
   */
  getLink: function(linkIndex) {
    var auto = this._component.getAutomation();
    return auto.getLink(linkIndex);
  },
  
  /**
   * Returns an object with the following properties for automation testing verification of the promoted link between 
   * specified nodes.
   *
   * @param {number} startNodeIndex Start node index
   * @param {number} endNodeIndex End node index
   * @property {string} color Link color
   * @property {string} endConnectorType The type of end connector on the link
   * @property {string} endNode The id of the end node.
   * @property {boolean} selected The selected state of the link
   * @property {string} startConnectorType The type of start connector on the link
   * @property {string} startNode The id of the start node.
   * @property {string} style Link style
   * @property {string} tooltip Link tooltip
   * @property {number} width Link width
   * @property {number} count Number of links it represents
   * @return {Object|null} An object containing properties for the link at the given index, or null if none exists.   
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   */
  getPromotedLink: function(startNodeIndex, endNodeIndex) {
    var auto = this._component.getAutomation();
    return auto.getPromotedLink(startNodeIndex, endNodeIndex);
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojDiagram
   */
  getContextByNode: function(node) {
    // context objects are documented with @ojnodecontext
    var context = this.getSubIdByNode(node);
    if (context && context['subId'] !== 'oj-diagram-tooltip')
      return context;

    return null;
  }
});

// Conditionally set the defaults for custom element vs widget syntax since we expose different APIs
oj.Components.setDefaultOptions({
  'ojDiagram': {
    'styleDefaults': oj.Components.createDynamicPropertyGetter(function(context) {
      if (context['isCustomElement'])
        return {
          'linkDefaults': {'svgStyle': {}},
          'nodeDefaults': {'icon': {'svgStyle': {}}},
          'promotedLink': {'svgStyle': {}}
        };
      return {};
    }),
  }
});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *    <tr>
 *       <td rowspan="3">Node or Link</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojDiagram
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>+</kbd></td>
 *       <td>Zoom in one level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>-</kbd></td>
 *       <td>Zoom out one level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>0 (zero)</kbd></td>
 *       <td>Zoom to fit.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Alt + 0 (zero)</kbd></td>
 *       <td>Zoom and center.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>PageUp or PageDown</kbd></td>
 *       <td>Pan up / down.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + PageUp or PageDown</kbd></td>
 *       <td>Pan left/right (RTL: Pan right/left).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow or RightArrow</kbd></td>
 *       <td>When focus is on a node, move focus and selection to nearest node left/right.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow or DownArrow</kbd></td>
 *       <td>When focus is on a node, move focus and selection to nearest node up/down.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Alt + &lt; or Alt + &gt;</kbd></td>
 *       <td>Move focus from the node to a link.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow or DownArrow</kbd></td>
 *       <td>When focus is on a link, navigate between links clockwise or counter clockwise.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow or RightArrow</kbd></td>
 *       <td>When focus is on a link, move focus from a link to a start or end node.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Space</kbd></td>
 *       <td>Select focused node / link.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Space</kbd></td>
 *       <td>Multi-select node / link with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + &lt;node or link navigation shortcut&gt;</kbd></td>
 *       <td>Move focus and multi-select a node or a link.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + &lt;node or link navigation shortcut&gt;</kbd></td>
 *       <td>Move focus to a node or a link but do not select.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojDiagram
 */

// COMMONLY USED FRAGMENTS, that might be converted into TS typedefs
 
/**
 * <ul>
 *   <li>componentElement {Element}: The Diagram element.</li> 
 *   <li>parentElement {Element}: A parent group element that takes a custom SVG fragment as the node content.
 *   Modifications of the parentElement are not supported. </li>
 *   <li>data {Object}: The data object for the node. </li>
 *   <li>content {Object}: An object that describes child content. The object has the following properties
 *     <ul>
 *       <li>element {Element}: an SVG group element that contains child nodes for the container.</li>
 *       <li>width {number}: width of the child content.</li>
 *       <li>height {number}: height of the child content.</li>
 *     </ul> 
 *   </li>
 *   <li>state {Object}: An object that reflects the current state of the diagram node. The object has the following properties 
 *     <ul>
 *       <li>hovered {boolean}: hovered state.</li>
 *       <li>selected {boolean}: selected state.</li>
 *       <li>focused {boolean}: focused state.</li>
 *       <li>zoom {number}: zoom state.</li>
 *       <li>expanded {boolean}: expanded state.</li>
 *     </ul>
 *   </li>
 *   <li>previousState {Object}: An object that reflects the previous state of the diagram node. The object has the following properties
 *     <ul>
 *       <li>hovered {boolean}: hovered state.</li>
 *       <li>selected {boolean}: selected state.</li>
 *       <li>focused {boolean}: focused state.</li>
 *       <li>zoom {number}: zoom state.</li>
 *       <li>expanded {boolean}: expanded state.</li>
 *     </ul>
 *   </li>
 *   <li>id {string}: Node id. </li>
 *   <li>type {string}: Object type = node. </li>
 *   <li>renderDefaultFocus {function()}: Function for rendering default focus effect for the diagram node. </li>
 *   <li>renderDefaultHover {function()}: Function for rendering default hover effect for the diagram node. </li>
 *   <li>renderDefaultSelection {function()}: Function for rendering default selection effect for the diagram node. </li>
 * </ul>
 * @ojfragment rendererContext - renderer context used by custom node renderers, such as renderer, focusRenderer, hoverRenderer, selectionRenderer, zoomRenderer.
 * @memberof oj.ojDiagram
 */
 
/**
 * <li>componentElement {Element}: The diagram element.</li>
 * <li>id {string}: The id of the diagram object</li>
 * <li>type {string}: The type of the diagram object - "link", "promotedLink" or "node".</li>
 * <li>label {string}: The label of the diagram object.</li>
 * <li>data {Object}: Relevant data for the object
 *   <ul>
 *     <li>data object for the node, if the object type is "node"</li>
 *     <li>data object for the link, if the object type is "link"</li>
 *     <li>an array of data objects that correspond to links represented by the promoted link</li>
 *   </ul>
 * </li>
 * @ojfragment dataContext - data context object
 * @memberof oj.ojDiagram
 */
 
 /**
  * <ul> 
  *   <li>x {number}: x-coordinate value of the drop in the component coordinate system.</li>
  *   <li>y {number}: y-coordinate value of the drop in the component coordinate system.</li>
  * </ul> 
  * <i> Note: </i>When the dropped items are originated from Diagram, the x, y coordinates represent the upper left corner of the dropped content.
  * @ojfragment backgroundDropContext - context object for the background drop used by dnd.drop.background callbacks
  * @memberof oj.ojDiagram
  */
  
 /**
  * <ul>
  *   <li>x {number}: x-coordinate value of the drop in the component coordinate system.</li>
  *   <li>y {number}: y-coordinate value of the drop in the component coordinate system.</li>
  *   <li>nodeX {number}: x-coordinate value of the drop in the target node coordinate system.</li>
  *   <li>nodeY {number}: y-coordinate value of the drop in the target node coordinate system.</li>
  *   <li>nodeContext {Object}: The JSON version of the data context for the target node.</li>
  * </ul>
  * <i> Note: </i>When the dropped items are originated from Diagram, the x, y coordinates represent the upper left corner of the dropped content.
  * @ojfragment nodeDropContext - context object for the node drop used by dnd.drop.node callbacks
  * @memberof oj.ojDiagram
  */
  
 /**
  * <ul>
  *   <li>x {number}: x-coordinate value of the drop in the component coordinate system.</li>
  *   <li>y {number}: y-coordinate value of the drop in the component coordinate system.</li>
  *   <li>linkContext {Object}: the JSON version of the data context for the target link.</li>
  * </ul> 
  * <i> Note: </i>When the dropped items are originated from Diagram, the x, y coordinates represent the upper left corner of the dropped content.
  * @ojfragment linkDropContext - context object for the node drop used by dnd.drop.link callbacks
  * @memberof oj.ojDiagram
  */
  
 /**
  * <ul>
  *   <li>dataContext {Object}: the JSON version of the data context for the link end node.</li>
  *   <li>portElement {Element}: DOM element that represents a port that received drop event.</li>
  * </ul>
  * @ojfragment portDropContext - context object for the node drop used by dnd.drop.link callbacks
  * @memberof oj.ojDiagram
  */
  
// KEEP FOR WIDGET SYNTAX

/**
 * A callback function for fetching child nodes. This function will be called to retrieve the children of each expanded node. The function takes a single argument, provided by the component, with the following properties: <ul> <li>id: The id of the parent node </li> <li>type : Object type - "node" </li> <li>label: The parent node label </li> <li>component: ojDiagram widget constructor </li> <li>data : The parent node data </li> </ul> The function should return one of the following: <ul> <li>An array of nodes.</li> <li>Promise: a Promise that will resolve with an array of nodes.</li> </ul> See the documentation for the <a href="#nodes[].nodes">nodes[].nodes</a> option for additional information. 
 * @ignore
 * @name childNodes
 * @memberof oj.ojDiagram
 * @instance
 * @type {function(Object)}
 * @default null
 * @deprecated This attribute is deprecated, use the data attribute instead.
 */
/**
 * An array of objects with the following properties that defines the data for the nodes. Also accepts a Promise or callback function for deferred data rendering. The function should return one of the following: <ul> <li>Promise: A Promise that will resolve with an array of data items. No data will be rendered if the Promise is rejected.</li> <li>Array: An array of data items.</li> </ul> 
 * @ignore
 * @name nodes
 * @memberof oj.ojDiagram
 * @instance
 * @type {Array.<Object>|Promise|function()}
 * @default null
 * @deprecated This attribute is deprecated, use the data attribute instead.
 */
/**
 * The id of this node.
 * @ignore
 * @name nodes[].id
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An array of objects with properties for the child nodes. Also a Promise that will resolve with an array of data items.<br/> Due to the fact that Promises begin resolving immediately, the recommended way to load child nodes on demand is via the callback function (which can return a Promise) using <i>childNodes </i> option in the component. In the case when all container nodes are expanded, the best approach is to pass a top-level Promise that resolves to a fully realized node hierarchy.<br/> Note that the callback function will only be called for this node if the nodes option is undefined. Any other value (including null or an empty array) will take precedence over the callback function. 
 * @ignore
 * @name nodes[].nodes
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Array.<Object>|Promise}
 * @default null
 */
/**
 * The CSS style object defining the style of the expanded container node.
 * @ignore
 * @name nodes[].containerStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class defining the style of the expanded container node.
 * @ignore
 * @name nodes[].containerClassName
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Determines when to display the expand/collapse button.
 * @ignore
 * @name nodes[].showDisclosure
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * Specifies whether or not the node will be selectable.
 * @ignore
 * @name nodes[].selectable
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The description of the node. This is used for accessibility and also for customizing the tooltip text.
 * @ignore
 * @name nodes[].shortDesc
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Primary label for the node.
 * @ignore
 * @name nodes[].label
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style string defining the style of the primary label.
 * @ignore
 * @name nodes[].labelStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style string/object defining the style of the node background.
 * @ignore
 * @name nodes[].backgroundStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string|Object}
 * @default null
 */
/**
 * The CSS style class defining the style of the node background.
 * @ignore
 * @name nodes[].backgroundClassName
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional array of additional category strings corresponding to this data item. This enables highlighting and filtering of individual data items through interactions with other components. Defaults to value of node id if unspecified.
 * @ignore
 * @name nodes[].categories
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * Specifies an icon to be used as a graphical element for the node
 * @ignore
 * @name nodes[].icon
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * Horizontal alignment for the icon relative to the given background.
 * @ignore
 * @name nodes[].icon.halign
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "left"
 * @ojvalue {string} "right"
 * @ojvalue {string} "center"
 * @default "center"
 */
/**
 * Vertical alignment for the icon relative to the given background.
 * @ignore
 * @name nodes[].icon.valign
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "top"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "center"
 * @default "center"
 */
/**
 * The x coordinate of the icon top left corner relative to the given background - pixels or %. The option takes precedence over halign option
 * @ignore
 * @name nodes[].icon.positionX
 * @memberof! oj.ojDiagram
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The y coordinate of the icon top left corner relative to the given background - pixels or %. The option takes precedence over valign option
 * @ignore
 * @name nodes[].icon.positionY
 * @memberof! oj.ojDiagram
 * @instance
 * @type {number|string}
 * @default null
 */
/**
 * The border color of this icon.
 * @ignore
 * @name nodes[].icon.borderColor
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border width in pixels.
 * @ignore
 * @name nodes[].icon.borderWidth
 * @memberof! oj.ojDiagram
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The fill color of this icon.
 * @ignore
 * @name nodes[].icon.color
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The fill pattern of this icon.
 * @ignore
 * @name nodes[].icon.pattern
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The opacity of this icon.
 * @ignore
 * @name nodes[].icon.opacity
 * @memberof! oj.ojDiagram
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The shape of this icon. Can take the name of a built-in shape or the svg path commands for a custom shape.
 * @ignore
 * @name nodes[].icon.shape
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "ellipse"
 * @ojvalue {string} "square"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "human"
 * @ojvalue {string} "rectangle"
 * @ojvalue {string} "star"
 * @ojvalue {string} "circle"
 * @default "circle"
 */
/**
 * The URI of the node image.
 * @ignore
 * @name nodes[].icon.source
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The optional URI of the node hover image. If not defined, the source image will be used.
 * @ignore
 * @name nodes[].icon.sourceHover
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The optional URI of the selected image on hover. If not defined, the sourceSelected image will be used. If the sourceSelected image is not defined, the source image will be used.
 * @ignore
 * @name nodes[].icon.sourceHoverSelected
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The optional URI of the selected image. If not defined, the source image will be used.
 * @ignore
 * @name nodes[].icon.sourceSelected
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of this icon.
 * @ignore
 * @name nodes[].icon.width
 * @memberof! oj.ojDiagram
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The height of this icon.
 * @ignore
 * @name nodes[].icon.height
 * @memberof! oj.ojDiagram
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The CSS style object defining the style of the icon.
 * @ignore
 * @name nodes[].icon.style
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class defining the style of the icon.
 * @ignore
 * @name nodes[].icon.className
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An array of objects with the following properties that defines the data for the links. Also accepts a Promise or callback function for deferred data rendering. The function should return one of the following: <ul> <li>Promise: A Promise that will resolve with an array of data items. No data will be rendered if the Promise is rejected.</li> <li>Array: An array of data items.</li> </ul> 
 * @ignore
 * @name links
 * @memberof oj.ojDiagram
 * @instance
 * @type {Array.<Object>|Promise|function()}
 * @default null
 * @deprecated This attribute is deprecated, use the data attribute instead.
 */
/**
 * The id of this link.
 * @ignore
 * @name links[].id
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Specifies whether or not the link will be selectable.
 * @ignore
 * @name links[].selectable
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The description of the link. This is used for accessibility and also for customizing the tooltip text.
 * @ignore
 * @name links[].shortDesc
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Primary label for the link.
 * @ignore
 * @name links[].label
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style string defining the style of the primary label.
 * @ignore
 * @name links[].labelStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Link color.
 * @ignore
 * @name links[].color
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The link style attribute can be string or object. The style string represents Link style type with following values: solid(default), dash, dot, dashDot. The style string representation has been deprecated. Consider specifying stroke-dasharray in the style object. The style object represents the CSS style of the link. User defined custom CSS Styles will be applied directly on the link. 
 * @ignore
 * @name links[].style
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string|Object}
 * @default null
 */
/**
 * The CSS style class defining the style of the link.
 * @ignore
 * @name links[].className
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Link width in pixels.
 * @ignore
 * @name links[].width
 * @memberof! oj.ojDiagram
 * @instance
 * @type {number}
 * @default 1.0
 */
/**
 * The id of the start node.
 * @ignore
 * @name links[].startNode
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The id of the end node.
 * @ignore
 * @name links[].endNode
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Specifies the type of start connector on the link.
 * @ignore
 * @name links[].startConnectorType
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "arrowOpen"
 * @ojvalue {string} "arrow"
 * @ojvalue {string} "arrowConcave"
 * @ojvalue {string} "circle"
 * @ojvalue {string} "rectangle"
 * @ojvalue {string} "rectangleRounded"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * Specifies the type of end connector on the link.
 * @ignore
 * @name links[].endConnectorType
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "arrowOpen"
 * @ojvalue {string} "arrow"
 * @ojvalue {string} "arrowConcave"
 * @ojvalue {string} "circle"
 * @ojvalue {string} "rectangle"
 * @ojvalue {string} "rectangleRounded"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * An optional array of additional category strings corresponding to this data item. This enables highlighting and filtering of individual data items through interactions with other components. Defaults to value of link id if unspecified.
 * @ignore
 * @name links[].categories
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The CSS style string/object defining the style of the node background.
 * @ignore
 * @name styleDefaults.nodeDefaults.backgroundStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string|Object}
 * @deprecated This attribute is deprecated, use the backgroundSvgStyle attribute instead.
 */
/**
 * The SVG CSS style object defining the style of the node background.
 * @ignore
 * @name styleDefaults.nodeDefaults.backgroundSvgStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 */
/**
 * The SVG CSS style class to apply to the background shape.
 * @ignore
 * @name styleDefaults.nodeDefaults.backgroundSvgClassName
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 */
/**
 * The CSS style object defining the style of the expanded container node.
 * @ignore
 * @name styleDefaults.nodeDefaults.containerStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 * @deprecated This attribute is deprecated, use the containerSvgStyle attribute instead.
 */
/**
 * The SVG CSS style object defining the style of the expanded container node.
 * @ignore
 * @name styleDefaults.nodeDefaults.containerSvgStyle
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 */
/**
 * The SVG CSS style class to apply to the expanded container shape.
 * @ignore
 * @name styleDefaults.nodeDefaults.containerSvgClassName
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 */
/**
 * Default horizontal alignment for the icon relative to the given background.
 * @ignore
 * @name styleDefaults.nodeDefaults.icon.halign
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "left"
 * @ojvalue {string} "right"
 * @ojvalue {string} "center"
 * @default "center"
 */
/**
 * Default vertical alignment for the icon relative to the given background.
 * @ignore
 * @name styleDefaults.nodeDefaults.icon.valign
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string}
 * @ojvalue {string} "top"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "center"
 * @default "center"
 */
/**
 * The default style object defining the style of the icon.
 * @ignore
 * @name styleDefaults.nodeDefaults.icon.style
 * @memberof! oj.ojDiagram
 * @instance
 * @type {Object}
 * @default null
 * @deprecated This attribute is deprecated, use the svgStyle attribute instead.
 */
/**
 *  The default link style attribute can be string or object. The default style string represents Link style type with following values: solid(default), dash, dot, dashDot. The default style string representation has been deprecated. Consider specifying stroke-dasharray in the default style object. The default style object represents the CSS style of the link. User defined custom CSS Styles will be applied directly on the link. 
 * @ignore
 * @name styleDefaults.linkDefaults.style
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string|Object}
 * @default null
 * @deprecated This attribute is deprecated, use the svgStyle attribute instead.
 */
/**
 *  The default promoted link style attribute can be string or object. The promoted style string represents Link style type with following values: solid, dash, dot(default), dashDot. The promoted style string representation has been deprecated. Consider specifying stroke-dasharray in the promoted style object. The promoted style object represents the CSS style of the link. User defined custom CSS Styles will be applied directly on the link. 
 * @ignore
 * @name styleDefaults.promotedLink.style
 * @memberof! oj.ojDiagram
 * @instance
 * @type {string|Object}
 * @default null
 * @deprecated This attribute is deprecated, use the svgStyle attribute instead.
 */
 

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for diagram node at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojsubid oj-diagram-node
 * @memberof oj.ojDiagram
 *
 * @example <caption>Get the first diagram node:</caption>
 * var nodes = $( ".selector" ).ojDiagram( "getNodeBySubId", {'subId': 'oj-diagram-node', 'index': 0} );
 */

 /**
 * <p>Sub-ID for diagram link at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojsubid oj-diagram-link
 * @memberof oj.ojDiagram
 *
 * @example <caption>Get the first diagram link:</caption>
 * var nodes = $( ".selector" ).ojDiagram( "getNodeBySubId", {'subId': 'oj-diagram-link', 'index': 0} );
 */

/**
 * <p>Sub-ID for the the diagram tooltip.</p>
 *
 * @ojsubid oj-diagram-tooltip
 * @memberof oj.ojDiagram
 *
 * @example <caption>Get the tooltip object of the diagram, if displayed:</caption>
 * var nodes = $( ".selector" ).ojDiagram( "getNodeBySubId", {'subId': 'oj-diagram-tooltip'} );
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for diagram node at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojnodecontext oj-diagram-node
 * @memberof oj.ojDiagram
 */

/**
 * <p>Context for diagram link at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojnodecontext oj-diagram-link
 * @memberof oj.ojDiagram
 */

/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojDiagramMeta = {
  "properties": {
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto","none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto","none"]
    },
    "data": {},
    "dnd": {
      "type": "object",
      "properties": {
        "drag": {
          "type": "object",
          "properties":{
            "nodes":{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "drag": {},
                "dragEnd": {},
                "dragStart": {}
              }
            },
            "ports":{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "linkStyle": {},
                "selector": {"type":"string"},
                "drag": {},
                "dragEnd": {},
                "dragStart": {}
              }
            }
          }
        },
        "drop": {
          "type": "object",
          "properties":{
            "background" :{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }              
            },
            "nodes" :{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            },
            "links" :{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            },
            "ports" :{
              "type":"object",
              "properties": {
                "dataTypes": {
                  "type": "Array<string>"
                },
                "feedbackStyle": {},
                "selector": {"type":"string"},
                "dragEnter": {},
                "dragLeave": {},
                "dragOver": {},
                "drop": {}
              }
            }
          }
        }
      }
    },
    "expanded" : {
      "writeback": true
    },
    "focusRenderer": {},
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string",
      "enumValues": ["any", "all"]
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "hoverRenderer": {},
    "layout": {},
    "linkHighlightMode": {
      "type": "string",
      "enumValues": ["linkAndNodes", "link"]
    },
    "linkProperties":{},
    "maxZoom": {
      "type": "number"
    },
    "minZoom": {
      "type": "number"
    },
    "nodeHighlightMode": {
      "type": "string",
      "enumValues": ["nodeAndIncomingLinks", "nodeAndOutgoingLinks", "nodeAndLinks", "node"]
    },
    "nodeProperties": {},
    "panDirection": {
      "type": "string",
      "enumValues": ["x", "y", "auto"]
    },
    "panning": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "promotedLinkBehavior":{
      "type": "string",
      "enumValues": ["none", "lazy", "full"]
    },
    "renderer": {},
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["none", "single", "multiple"]
    },
    "selectionRenderer": {},
    "styleDefaults": {
      "type": "object",
      "properties": {
        "animationDuration": {
          "type": "number"
        },
        "hoverBehaviorDelay": {
          "type": "number"
        },
        "linkDefaults": {
          "type": "object",
          "properties": {
            "color": {
              "type": "string"
            },
            "endConnectorType": {
              "type": "string",
              "enumValues": ["arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none"]
            },
            "labelStyle": {
              "type": "object"
            },
            "startConnectorType": {
              "type": "string",
              "enumValues": ["arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none"]
            },
            "svgStyle": {
              "type": "object"
            },
            "width": {
              "type": "number"
            }
          }
        },
        "nodeDefaults": {
          "type": "object",
          "properties": {
            "icon": {
              "type": "object",
              "properties": {
                "borderColor": {
                  "type": "string"
                },
                "borderWidth": {
                  "type": "number"
                },
                "color": {
                  "type": "string"
                },
                "height": {
                  "type": "number"
                },
                "pattern": {
                  "type": "string",
                  "enumValues": ["smallChecker", "smallCrosshatch", "smallDiagonalLeft", 
                                 "smallDiagonalRight", "smallDiamond", "smallTriangle",
                                 "largeChecker", "largeCrosshatch", "largeDiagonalLeft",
                                 "largeDiagonalRight", "largeDiamond", "largeTriangle", "none"]
                },
                "shape": {
                  "type": "string"
                },
                "source": {
                  "type": "string"
                },
                "sourceHover": {
                  "type": "string"
                },
                "sourceHoverSelected": {
                  "type": "string"
                },
                "sourceSelected": {
                  "type": "string"
                },
                "svgStyle": {
                  "type": "object"
                },
                "width": {
                  "type": "number"
                }
              }
            },
            "labelStyle": {
              "type": "object"
            },
            "showDisclosure": {
              "type": "string",
              "enumValues": ["on", "off"]
            }
          }
        },
        "promotedLink": {
          "type": "object",
          "properties": {
            "color": {
              "type": "string"
            },
            "endConnectorType": {
              "type": "string",
              "enumValues": ["arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none"]
            },
            "startConnectorType": {
              "type": "string",
              "enumValues": ["arrowOpen", "arrow", "arrowConcave", "circle", "rectangle", "rectangleRounded", "none"]
            },
            "svgStyle": {
              "type": "object"
            },
            "width": {
              "type": "number"
            }
          }
        }
      }
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "touchResponse": {
      "type": "string",
      "enumValues": ["touchStart", "auto"]
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Diagram"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "promotedLink": {
          "type": "string",
          "value": "{0} link"
        },
        "promotedLinkAriaDesc": {
          "type": "string",
          "value": "Indirect"
        },
        "promotedLinks": {
          "type": "string",
          "value": "{0} links"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    },
    "zooming": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "zoomRenderer": {}
  },
  "events": {
    "beforeCollapse": {},
    "beforeExpand": {},
    "collapse": {},
    "expand": {}
  },
  "methods": {
    "getContextByNode": {},
    "getLink": {},
    "getLinkCount": {},
    "getNode": {},
    "getNodeCount": {},
    "getPromotedLink": {}
  },
  "extension": {
    _WIDGET_NAME: "ojDiagram"
  }
};
oj.CustomElementBridge.registerMetadata('oj-diagram', 'dvtBaseComponent', ojDiagramMeta);
// Get the combined meta of superclass which contains a shape parse function generator
var dvtMeta = oj.CustomElementBridge.getMetadata('oj-diagram');
oj.CustomElementBridge.register('oj-diagram', {
  'metadata': dvtMeta,
  'parseFunction': dvtMeta['extension']._DVT_PARSE_FUNC({'style-defaults.node-defaults.icon.shape': true})
});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

// retrieved from https://raw.github.com/jquery/jquery-ui/1-10-stable/ui/jquery.ui.dialog.js on 09/03/2013, and then modified

//
// Note that one of the main differences between JET Dialog and the jQueryUI dialog
// is the reparenting approach:
//
//   - JET Dialog reparents to the body on OPEN
//   - jQueryUI dialog reparents to the appendTo() container on CREATE
//

// Notes:
//  - $.uiBackCompat has been deprecated
//

/*!
 *
 * Depends:
 *	jquery.ui.core.js
 *	jquery.ui.widget.js
 *	jquery.ui.draggable.js
 *	jquery.ui.mouse.js
 *	jquery.ui.position.js
 */


(function() {

  // class name constants
  var /** @const */ OJD_BODY = 'oj-dialog-body';
  var /** @const */ OJD_CONTENT = 'oj-dialog-content';
  var /** @const */ OJD_FOOTER = 'oj-dialog-footer';
  var /** @const */ OJD_HEADER = 'oj-dialog-header';
  var /** @const */ OJD_HEADER_CLOSE = 'oj-dialog-header-close';
  var /** @const */ OJD_HEADER_CLOSE_WRAPPER = 'oj-dialog-header-close-wrapper';
  var /** @const */ OJD_OPTION_DEFAULTS = 'oj-dialog-option-defaults';

  var /** @const */ OJ_RESIZABLE_N = 'oj-resizable-n';
  var /** @const */ OJ_RESIZABLE_E = 'oj-resizable-e';
  var /** @const */ OJ_RESIZABLE_S = 'oj-resizable-s';
  var /** @const */ OJ_RESIZABLE_W = 'oj-resizable-w';
  var /** @const */ OJ_RESIZABLE_SE = 'oj-resizable-se';
  var /** @const */ OJ_RESIZABLE_SW = 'oj-resizable-sw';
  var /** @const */ OJ_RESIZABLE_NE = 'oj-resizable-ne';
  var /** @const */ OJ_RESIZABLE_NW = 'oj-resizable-nw';

  var /** @const */ OJ_DRAGGABLE = 'oj-draggable';

  /**
   * @typedef {Object} oj.ojDialog.PositionAlign
   * @property {"top"|"bottom"|"center"} [vertical] Vertical alignment.
   * @property {"start"|"end"|"left"|"center"|"bottom"} [horizontal] Horizontal alignment. <p>
   * <ul>
   *  <li><b>"start"</b> evaluates to "left" in LTR mode and "right" in RTL mode.</li>
   *  <li><b>"end"</b> evaluates to "right" in LTR mode and "left" in RTL mode.</li>
   * </ul>
   * 
   */

  /**
   * @typedef {Object} oj.ojDialog.PositionPoint
   * @property {number} [x] Horizontal aligment offset.
   * @property {number} [y] Vertical alignment offset.
   */

  /**
   * @typedef {Object} oj.ojDialog.Position
   * @property {oj.ojDialog.PositionAlign} [my] Defines which edge on the popup to align with the target ("of") element.
   * @property {oj.ojDialog.PositionAlign} [at] Defines which position on the target element ("of") to align the positioned element
   *                                  against.
   * @property {oj.ojDialog.PositionPoint} [offset] Defines a point offset in pixels from the ("my") alignment.
   * @property {string|oj.ojDialog.PositionPoint} [of] Which element to position the popup against. 
   *
   * If the value is a string, it should be a selector or the literal string value
   * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
   * is used, the values are relative to the whole document.  Page horizontal and vertical
   * scroll offsets need to be factored into this point - see UIEvent
   * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
   * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
   * 
   * @property {"flip"|"fit"|"flipfit"|"none"} [collision] Rule for alternate alignment. <p>
   * <ul>
   *  <li><b>"flip"</b> the element to the opposite side of the target and the
   *             collision detection is run again to see if it will fit. Whichever side
   *             allows more of the element to be visible will be used. </li>
   * <li><b>"fit"</b> shift the element away from the edge of the window. </li>
   * <li><b>"flipfit"</b> first applies the flip logic, placing the element
   *  on whichever side allows more of the element to be visible. Then the fit logic
   *  is applied to ensure as much of the element is visible as possible.</li>
   * <li><b>"none"</b> no collision detection.</li>
   * </ul>
   */

  /**
   * @ojcomponent oj.ojDialog
   * @augments oj.baseComponent
   * @since 0.6
   * @ojrole dialog
   * @ojdisplayname Dialog
   * @ojshortdesc Displays a popup window that provides information and gathers input from the application user.
   * @ojstatus preview
   *
   * @classdesc
   * <h3 id="dialogOverview-section">
   *   JET Dialog Component
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#dialogOverview-section"></a>
   * </h3>
   * <p>Description: Themeable, WAI-ARIA-compliant dialog component.
   * A dialog is a floating window that typically contains a title bar and a content area.
   * The dialog window can be moved by dragging on the title area, and closed with the 'x' icon (by default). Dialogs can also be resized by dragging on edges or corners of the dialog component. </p>
   *
   *<p>If the content length exceeds the maximum height, a scrollbar will automatically appear.</p>
   *
   *<p>A bottom button bar and semi-transparent modal overlay layer are common options that can be added.</p>
   *
   * <h3 id="styling-section">
   *   Styling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"stylingDoc"}
   *
   * <h3 id="focus-section">
   *   Focus
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#focus-section"></a>
   * </h3>
   *
   *<p>Upon opening a dialog, focus is automatically moved to the first item that matches the following:</p>
   *<ol>
   *  <li>The first element within the dialog with the <code>autofocus</code> attribute</li>
   *  <li>The first <code>:tabbable</code> element within the dialog body</li>
   *  <li>The first <code>:tabbable</code> element within the dialog footer</li>
   *  <li>The dialog's close button</li>
   *  <li>The dialog itself</li>
   *</ol>
   *<p>While open, the dialog widget ensures that tabbing cycles focus between elements within the dialog itself, not elements outside of it. Modal dialogs additionally prevent mouse users from clicking on elements outside of the dialog.</p>
   *
   *<p>Upon closing a dialog, focus is automatically returned to the element that had focus when the dialog was opened.</p>
   * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.</p>
   *
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   *
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDoc"}
   *
   *<p>
   *<h3>Sizing</h3>
   *
   * <p> Dialog dimensions, including <code class="prettyprint"> height, width, min-width, max-width, min-height</code> and <code class="prettyprint">max-height</code> are defined with css variables. The default dialog dimensions are the following:
   *
   *<ul>
   *  <li> <code class="prettyprint">height: auto</code> </li>
   *  <li> <code class="prettyprint">width: 300px</code> </li>
   *  <li> <code class="prettyprint">min-width: 200px</code> </li>
   *</ul>
   *
   * In most cases, you will want to use the default <code class="prettyprint">height:auto</code>, since this will automatically adjust the height based on the content.
   * Users can change the dialog dimensions using style attributes:
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-dialog id="wideDialog" title="Wide Dialog" style="width: 400px; min-width: 100px; max-width 500px;"&gt;
   *    &lt;div slot="body"&gt;
   *       &lt;p&gt; Dialog Text
   *    &lt;/div&gt;
   * &lt;/oj-dialog&gt;
   * </code></pre>
   *
   * <h3 id="accessibility-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
   * </h3>
   * <h4> role </h4>
   * By default, the role will be set to dialog.
   * This can be observed by inspecting the DOM:
   *
   * <pre class="prettyprint">
   * <code>
   *  &lt;div class="ojdialog ..." role="dialog"&gt;
   * </code></pre>
   *
   * This can be changed using the role attribute. WAI-ARIA recommends that role="dialog" be used if the dialog expects input (such as text input),
   * otherwise, use the role attribute to assign role="alertdialog".
   *
   * <h4> aria-labelledby </h4>
   *
   * For both default and user-defined headers, the dialog component takes care of aria-labelledby for you.
   * The <code class="prettyprint">aria-labelledby</code> attribute is generated automatically (and set to the id of the header's title).
   * For user-defined headers, the title div is identified by the div that has the <code class="prettyprint">oj-dialog-title</code> class.
   * Note that user-defined headers must have a title div (in order to meet accesibility requirements).
   *
   * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.</p>
   *
   * <h3 id="reparenting-section">
   *   Reparenting
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#reparenting-section"></a>
   * </h3>
   *
   *  <p id="reparenting-strategy">
   *     When dialogs are open, they will be reparented into a common container in the
   *     document body and reparented back when closed.  Within this container in the body,
   *     dialogs will always be top rooted but other types of dialogs used within an open
   *     dialog will be reparented within the dialog's layer. The dialog's layer defines its
   *     z-index weight "stacking context" and marked by the "oj-dialog-layer" style.
   *     The goal of this design is to maintain as much of the page author's document structure
   *     while avoiding most of the clipping and positioning issues of a completely inline design.
   *     Dialogs are assigned the same z-index values  The layering between dialog peers reflect the
   *     opening order.  In addition, the dialog that has active focus will be assigned a greater z-index
   *     by way of the "oj-focus-within" pseudo selector applied with "oj-dialog-layer" selector.
   *     The page author has control over z-index weights by way of the "oj-dialog-layer" selector.
   *  </p>
   *  <p>
   *     There are known caveats with this design. However, these scenarios are considered "bad use"
   *     based on our JET dialog/popup strategy.
   *  </p>
   *  <ol>
   *    <li>Events raised within the dialog will not bubble up to the dialog's original ancestors.  Instead, listeners for menu events should
   *        be applied to either the dialog's root element, or the document.</li>
   *    <li>Likewise, developers should not use CSS descendant selectors, or similar logic, that assumes that the dialog will remain a child
   *        of its original parent.</li>
   *    <li>Dialogs containing iframes are problematic.  The iframe elements "may" fire a HTTP GET request for its src attribute
   *        each time the iframe is reparented in the document.</li>
   *    <li>In some browsers, reparenting a dialog that contains elements having overflow, will cause these overflow elements to
   *        reset their scrollTop.</li>
   *  </ol>
   *
   * <h3 id="rtl-section">
   *   Reading direction
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
   * </h3>
   *
   * <p> Setting the reading direction (LTR or RTL) is supported by setting the <code class="prettyprint">"dir"</code> attribute on the
   * <code class="prettyprint">&lt;html></code> element of the page.  As with any JET component, in the unusual case that the reading direction
   * is changed post-init, the dialog must be <code class="prettyprint">refresh()</code>ed, or the page must be reloaded.
   *
   * <h3 id="pseudos-section">
   *   Pseudo-selectors
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#pseudos-section"></a>
   * </h3>
   *
   * <p>The <code class="prettyprint">:oj-dialog</code> pseudo-selector can be used in jQuery expressions to select JET Dialogs.  For example:
   *
   * <pre class="prettyprint">
   * <code>$( ":oj-dialog" ) // selects all JET Dialogs on the page
   * $myEventTarget.closest( ":oj-dialog" ) // selects the closest ancestor that is a JET Dialog
   * </code></pre>
   *
   *<h3>Additional Examples</h3>
   *
   * <p> The following defines a basic dialog, with an ok button in the footer:
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-dialog id="dialogWithFooter" title="Dialog with Footer" style="width: 400px; min-width: 100px; max-width 500px;"&gt;
   *    &lt;div slot="body"&gt;
   *       &lt;p&gt; Dialog Text
   *    &lt;/div&gt;
   *    &lt;div slot="footer"&gt;
   *       &lt;oj-button id="okButton"&gt; OK &lt;/oj-button&gt;
   *    &lt;/div&gt;
   * &lt;/oj-dialog&gt;
   *
   * </code></pre>
   *
   * Note that you will need to define your own event handlers for the ok and close buttons (see the demos for examples of this).
   *
   * <p> A dialog with user-defined header is shown next. Arbitrary header content can be defined using a user-defined header.
   *
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-dialog id="dialog" title="Dialog Title"&gt;
   *    &lt;div slot="header"&gt;
   *       &lt;span id="dialog-title-id" class="oj-dialog-title"&gt; User Defined Header&lt;/span&gt;
   *    &lt;/div&gt;
   *    &lt;div slot="body"&gt;
   *       &lt;p&gt; Dialog Text
   *    &lt;/div&gt;
   * &lt;/oj-dialog&gt;
   * </code></pre>
   */

  oj.__registerWidget("oj.ojDialog", $['oj']['baseComponent'], {
    version: "1.0.0",
    widgetEventPrefix: "oj",
    options:
      {
        /**
         * Specifies the cancel behavior of the dialog. The following are valid values:
         *
         * <ul>
         * <li>
         * <code class="prettyprint">"icon"</code> - (the default) (a) a close icon will automatically be created, and (b) the dialog will close when it has focus and user presses the escape (ESC) key.
         * </li>
         * <li>
         * <code class="prettyprint">"none"</code> - no actions will be associated with the escape key.
         * </li>
         * <li>
         * <code class="prettyprint">"escape"</code> -  the dialog will close when it has focus and user presses the escape (ESC) key. A close icon will not automatically be created.
         * </li>
         * </ul>
         *
         * Note that the cancelBehavior applies to both automatic and user-defined headers. So by default, a user-defined header will have a system generated close icon.
         * @expose
         * @memberof oj.ojDialog
         * @instance
         * @type {string}
         * @default "icon"
         *
         * @example <caption>Initialize the dialog to disable the default <code class="prettyprint">cancelBehavior</code></caption>
         * &lt;oj-dialog cancel-behavior="none" &gt;&lt;/oj-dialog&gt;
         *
         * @example <caption>Get or set the <code class="prettyprint">cancelBehavior</code> property, after initialization:</caption>
         * // getter
         * var cancelBehavior = myDialog.cancelBehavior;
         *
         * // setter
         * myDialog.cancelBehavior = "none";
         *
         * @example <caption>Set the default in the theme (SCSS) :</caption>
         * $dialogCancelBehaviorOptionDefault: none !default;
         *
         */
        cancelBehavior: "icon",
        /**
         * Specifies the drag affordance.
         * If set to <code class="prettyprint">"title-bar"</code> (the default) the dialog will be draggable by the title bar.
         * If <code class="prettyprint">"none"</code>, the dialog will not be draggable.
         *
         * @expose
         * @memberof oj.ojDialog
         * @instance
         * @type {string}
         * @default "title-bar"
         *
         * @example <caption>Initialize the dialog to disable dragging <code class="prettyprint">dragAffordance</code></caption>
         * &lt;oj-dialog drag-affordance="none" &gt;&lt;/oj-dialog&gt;
         *
         * @example <caption>Get or set the <code class="prettyprint">dragAffordance</code> property, after initialization:</caption>
         * // getter
         * var dragAffordance = myDialog.dragAffordance;
         *
         * // setter
         * myDialog.dragAffordance = "none";
         */
        dragAffordance: "title-bar",
        /**
         * <p> Set the initial visibility of the dialog.
         * If set to <code class="prettyprint">"show"</code>, the dialog will automatically open upon initialization.
         * If <code class="prettyprint">"hide"</code>, the dialog will stay hidden until
         * the <a href="#method-open"><code class="prettyprint">open()</code></a> method is called.
         *
         * @expose
         * @memberof oj.ojDialog
         * @instance
         * @type {string}
         * @default "hide"
         *
         * @example <caption>Initialize the dialog with the <code class="prettyprint">initialVisibility</code> property:</caption>
         * &lt;oj-dialog initial-visibility="show" &gt;&lt;/oj-dialog&gt;
         *
         * @example <caption>Get or set the <code class="prettyprint">initialVisibility</code> property, after initialization:</caption>
         * // getter
         * var initialVisibility = myDialog.initialVisibility;
         *
         * // setter
         * myDialog.initialVisibility = "show";
         */
        initialVisibility: "hide",
        /**
         *
         * The modality of the dialog. Valid values are:
         * <ul>
         * <li>
         * <code class="prettyprint">"modal"</code> - (the default) The dialog is modal. Interactions with other page elements are disabled. Modal dialogs overlay other page elements.
         * </li>
         * <li>
         * <code class="prettyprint">"modeless"</code> - defines a modeless dialog.
         * </li>
         * </ul>
         *
         * @expose
         * @memberof oj.ojDialog
         * @instance
         * @default "modal"
         * @type {string}
         *
         * @example <caption>Initialize the dialog to a specific modality <code class="prettyprint">modality</code></caption>
         * &lt;oj-dialog modality="modeless" &gt;&lt;/oj-dialog&gt;
         *
         * @example <caption>Get or set the <code class="prettyprint">modality</code> property, after initialization:</caption>
         * // getter
         * var modality = myDialog.modality;
         *
         * // setter
         * myDialog.modality = "modeless";
         */
        modality: "modal",
        /**
           * <p>Position object is used to establish the location the dialog will appear relative to
           * another element. {@link oj.ojDialog.Position} defines "my" alignment "at" the alignment
           * "of" some other thing which can be "offset" by so many pixels.</p>
           *
           * <p>The "my" and "at" properties defines aligment points relative to the dialog and other
           * element.  The "my" property represents the dialog's alignment where the "at" property
           * represents the other element that can be identified by "of". The values of these properties
           * describe horizontal and vertical alignments.</p>
           *
           * 
           * <b>Deprecated v3.0.0 jQuery UI position syntax; Use of a percent unit with
           * "my" or "at" is not supported.</b>
           *
           * @expose
           * @memberof oj.ojDialog
           * @default {"my":{"vertical" :"center", "horizontal": "center"}, "at": {"vertical": "center", "horizontal": "center"}, "of": "window", "collision": "fit", "offset": {"x": 0, "y": "0"}}
           * @instance
           * @type {Object}
           * @ojsignature { target: "Type",
           *                value: "oj.ojDialog.Position",
           *                jsdocOverride: true}
           * @name position
           * @example <caption>Initialize the dialog with <code class="prettyprint">position</code>
           *           property specified:</caption>
           * &lt;oj-dialog position.my.horizontal="left"
           *           position.my.vertical="top"
           *           position.at.horizontal="right"
           *           position.at.vertical="top" &gt;&lt;/oj-dialog&gt;
           *
           * @example <caption>Get or set the <code class="prettyprint">position</code> property,
           *          after initialization:</caption>
           * // getter
           * var position = myDialog.position;
           *
           * // setter
           * myDialog.position =
           *    {"my": {"horizontal": "start", "vertical": "bottom"},
           *     "at": {"horizontal": "end", "vertical": "top" },
           *     "offset": {"x": 0, "y":5}};
           */
        position :
            {
              /**
               * Defines which edge on the dialog to align with the target ("of") element.
               *
               * @expose
               * @memberof! oj.ojDialog
               * @instance
               * @alias position.my
               * @name position.my
               * @type {{horizontal:string, vertical:string}}
               */
              my : {
                /**
                 * @expose
                 * @memberof! oj.ojDialog
                 * @instance
                 * @alias position.my.horizontal
                 * @name position.my.horizontal
                 * @type {string}
                 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
                 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
                 * @ojvalue {string} "left"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "right"
                 */
                horizontal: 'center',
                /**
                 * @expose
                 * @memberof! oj.ojDialog
                 * @instance
                 * @alias position.my.vertical
                 * @name position.my.vertical
                 * @type {string}
                 * @ojvalue {string} "top"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "bottom"
                 */
                vertical: 'center'
              },
              /**
               * Defines a point offset in pixels from the ("my") alignment.
               * @expose
               * @memberof! oj.ojDialog
               * @instance
               * @alias position.offset
               * @name position.offset
               * @type {{x:number, y:number}}
               */
              offset: {
                /**
                 * Horizontal aligment offset.
                 * @expose
                 * @memberof! oj.ojDialog
                 * @instance
                 * @alias position.offset.x
                 * @name position.offset.x
                 * @type {number}
                 */
                x: 0,
                /**
                 * Vertical alignment offset.
                 * @expose
                 * @memberof! oj.ojDialog
                 * @instance
                 * @alias position.offset.y
                 * @name position.offset.y
                 * @type {number}
                 */
                y: 0
              },
              /**
               * Defines which position on the target element ("of") to align the positioned element
               * against.
               *
               * @expose
               * @memberof! oj.ojDialog
               * @instance
               * @alias position.at
               * @name position.at
               * @type {{horizontal:string, vertical:string}}
               */
              at : {
                /**
                 * @expose
                 * @memberof! oj.ojDialog
                 * @instance
                 * @alias position.at.horizontal
                 * @name position.at.horizontal
                 * @type {string}
                 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
                 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
                 * @ojvalue {string} "left"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "right"
                 */
                horizontal: 'center',
                /**
                 * @expose
                 * @memberof! oj.ojDialog
                 * @instance
                 * @alias position.at.vertical
                 * @name position.at.vertical
                 * @type {string}
                 * @ojvalue {string} "top"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "bottom"
                 */
                vertical: 'center'
              },
              /**
               * Which element to position the dialog against.
               * If the value is a string, it should be a selector or the literal string value
               * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
               * is used, the values are relative to the whole document.  Page horizontal and vertical
               * scroll offsets need to be factored into this point - see UIEvent
               * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
               * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
               *
               * @example <caption>Finding the point for an svg element:</caption>
               * var rect = svgDom.getBoundingClientRect();
               * var position = {of:{x:rect.left + window.pageXOffset, y:rect.top + window.pageYOffset}};
               *
               * @expose
               * @memberof! oj.ojDialog
               * @instance
               * @alias position.of
               * @name position.of
               * @type {string|{x: number, y: number}}
               */
              of : 'window',
              /**
               * Rule for alternate alignment.
               *
               * @expose
               * @memberof! oj.ojDialog
               * @instance
               * @alias position.collision
               * @name position.collision
               * @type {string}
               * @ojvalue {string} "flip" the element to the opposite side of the target and the
               *  collision detection is run again to see if it will fit. Whichever side
               *  allows more of the element to be visible will be used.
               * @ojvalue {string} "fit" shift the element away from the edge of the window.
               * @ojvalue {string} "flipfit" first applies the flip logic, placing the element
               *  on whichever side allows more of the element to be visible. Then the fit logic
               *  is applied to ensure as much of the element is visible as possible.
               * @ojvalue {string} "none" no collision detection.
               */
              collision : 'fit',
              // Ensure the titlebar is always visible
              using: function(pos) {
                  var topOffset = $(this).css(pos).offset().top;
                  if (topOffset < 0) {
                    $(this).css("top", pos.top - topOffset);
                  }
               }
            },
        /**
         *
         * The resizeBehavior of the dialog. "resizable" (default) makes the dialog resizable.
         * "none" disables dialog resizability.
         *
         * @expose
         * @memberof oj.ojDialog
         * @instance
         * @type {string}
         * @default "resizable"
         *
         * @example <caption>Initialize the dialog to a specific resizeBehavior <code class="prettyprint">resizeBehavior</code></caption>
         * &lt;oj-dialog resize-behavior="none" &gt;&lt;/oj-dialog&gt;
         *
         * @example <caption>Get or set the <code class="prettyprint">resizeBehavior</code> property, after initialization:</caption>
         *
         * // getter
         * var resizeBehavior = myDialog.resizeBehavior;
         *
         * // setter
         * myDialog.resizeBehavior = "none";
         * @example <caption>Set the default in the theme (SCSS) :</caption>
         * $dialogResizeBehaviorOptionDefault: none !default;
         */
        resizeBehavior: "resizable",
        /**
         *
         * The WAI-ARIA role of the dialog. By default, role="dialog" is added to the generated HTML markup that surrounds the dialog.
         * When used as an alert dialog, the user should set role to "alertdialog".
         *
         * @expose
         * @memberof oj.ojDialog
         * @instance
         * @type {string}
         * @default "dialog"
         *
         * @example <caption>Initialize the dialog with the <code class="prettyprint">role</code></caption> property specified:</caption>
         * &lt;oj-dialog role="alertdialog" &gt;&lt;/oj-dialog&gt;
         *
         * @example <caption>Get or set the <code class="prettyprint">role</code> property, after initialization:</caption>
         * // getter
         * var role = myDialog.role;
         *
         * // setter
         * myDialog.role = "alertdialog";
         */
        role: "dialog",
        /**
         *
         * Specify the title of the dialog. null is the default.
         *
         * @expose
         * @ignore
         * @memberof oj.ojDialog
         * @instance
         * @type {string|null}
         *
         * @example <caption>Initialize the dialog to a specific title <code class="prettyprint">title</code></caption>
         * &lt;oj-dialog title="Title of Dialog" &gt;&lt;/oj-dialog&gt;
         *
         * @example <caption>Get or set the <code class="prettyprint">title</code> property, after initialization:</caption>
         * // getter
         * var title = myDialog.title;
         *
         * // setter
         * myDialog.title = "Title of Dialog";
         */
        title: null,
        ///////////////////////////////////////////////////////
        // events
        ///////////////////////////////////////////////////////

          /**
           * Triggered before the dialog is dismissed via the
           * <code class="prettyprint">close()</code> method. The close can be cancelled by calling
           * <code class="prettyprint">event.preventDefault()</code>.
           *
           * @expose
           * @event
           * @memberof oj.ojDialog
           * @instance
           * @ojcancelable
           * @ojbubbles
           * @property {Event} event a custom event
           */
        beforeClose: null,
          /**
           * Triggered before the dialog is launched via the <code class="prettyprint">open()</code>
           * method. The open can be cancelled by calling
           * <code class="prettyprint">event.preventDefault()</code>.
           *
           * @expose
           * @event
           * @memberof oj.ojDialog
           * @instance
           * @ojcancelable
           * @ojbubbles
           * @property {Event} event a custom event
           */
        beforeOpen: null,
          /**
           * Triggered after the dialog is dismissed via the
           * <code class="prettyprint">close()</code> method.
           *
           * @expose
           * @event
           * @memberof oj.ojDialog
           * @instance
           * @ojbubbles
           * @property {Event} event a custom event
           */
        close: null,
          /**
           * Triggered after focus has been transfered to the dialog.
           *
           * @expose
           * @event
           * @memberof oj.ojDialog
           * @instance
           * @ojbubbles
           * @property {Event} event a custom event
           */
        focus: null,
          /**
           * Triggered after the dialog is launched via the <code class="prettyprint">open()</code>
           * method.
           *
           * @expose
           * @event
           * @memberof oj.ojDialog
           * @instance
           * @ojcancelable
           * @ojbubbles
           * @property {Event} event a custom event
           */
        open: null,
        /**
         * Triggered when the dialog is being resized.
         *
         * @expose
         * @event
         * @memberof oj.ojDialog
         * @instance
         * @ojbubbles
         * @property {Event} event a custom event

         */
        resize: null,
        /**
         * Triggered when the user starts resizing the dialog.
         *
         * @expose
         * @event
         * @memberof oj.ojDialog
         * @instance
         * @ojcancelable
         * @ojbubbles
         * @property {Event} event a custom event
         *
         * <ul>
         * <li>
         * <div><strong>event</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Event">Event</a>
         * </div>
         * <div></div>
         * </li>
         * <li>
         * <div><strong>ui</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div></div>
         * <ul>
         * <li>
         * <div><strong>originalPosition</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The CSS position of the dialog prior to being resized.</div>
         * </li>
         * <li>
         * <div><strong>position</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The current CSS position of the dialog.</div>
         * </li>
         * <li>
         * <div><strong>originalSize</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The size of the dialog prior to being resized.</div>
         * </li>
         * <li>
         * <div><strong>size</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The current size of the dialog.</div>
         * </li>
         * </ul>
         * </li>
         * </ul>
         */
        resizeStart: null,
        /**
         * Triggered when the user stops resizing the dialog.
         *
         * @expose
         * @event
         * @memberof oj.ojDialog
         * @instance
         * @ojcancelable
         * @ojbubbles
         * @property {Event} event a custom event
         *
         * <ul>
         * <li>
         * <div><strong>event</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Event">Event</a>
         * </div>
         * <div></div>
         * </li>
         * <li>
         * <div><strong>ui</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div></div>
         * <ul>
         * <li>
         * <div><strong>originalPosition</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The CSS position of the dialog prior to being resized.</div>
         * </li>
         * <li>
         * <div><strong>position</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The current CSS position of the dialog.</div>
         * </li>
         * <li>
         * <div><strong>originalSize</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The size of the dialog prior to being resized.</div>
         * </li>
         * <li>
         * <div><strong>size</strong></div>
         * <div>Type: <a href="http://api.jquery.com/Types/#Object">Object</a>
         * </div>
         * <div>The current size of the dialog.</div>
         * </li>
         * </ul>
         * </li>
         * </ul>
         *
         */
        resizeStop: null,
          /**
           * Triggered when a default animation is about to start, such as when the component is
           * being opened/closed or a child item is being added/removed. The default animation can
           * be cancelled by calling <code class="prettyprint">event.preventDefault</code>.
           *
           * @expose
           * @event
           * @memberof oj.ojDialog
           * @instance
           * @ojcancelable
           * @ojbubbles
           * @property {"open"|"close"} action The action that is starting the animation.
           *            The number of actions can vary from component to component.
           *            Suggested values are:
           *                    <ul>
           *                      <li>"open" - when a dialog component is opened</li>
           *                      <li>"close" - when a dialog component is closed</li>
           *                    </ul>
           * @property {!Element} element target of animation
           * @property {!function():void} endCallback If the event listener calls
           *            event.preventDefault to cancel the default animation, It must call the
           *            endCallback function when it finishes its own animation handling and any
           *            custom animation has ended.
           *
           *
           * @example <caption>Bind an event listener to the
           *          <code class="prettyprint">onOjAnimateStart</code> property to override the default
           *          "close" animation:</caption>
           * myDialog.onOjAnimateStart = function( event )
           *   {
           *     // verify that the component firing the event is a component of interest and action
           *      is close
           *     if (event.detail.action == "close") {
           *       event.preventDefault();
           *       oj.AnimationUtils.slideOut(event.detail.element).then(event.detail.endCallback);
           *   };
           *
           * @example <caption>The default open and close animations are controlled via the theme
           *          (SCSS) :</caption>
           * $dialogOpenAnimation: ((effect: "zoomIn"), "fadeIn")  !default;
           * $dialogCloseAnimation: ((effect: "zoomOut", persist: "all"), "fadeOut")  !default;
           */

        animateStart: null,
          /**
           * Triggered when a default animation has ended, such as when the component is being
           * opened/closed or a child item is being added/removed. This event is not triggered if
           * the application has called preventDefault on the animateStart
           * event.
           *
           * @expose
           * @event
           * @memberof oj.ojDialog
           * @instance
           * @ojcancelable
           * @ojbubbles
           * @property {!Element} element target of animation
           * @property {"open"|"close"} action The action that is ending the animation.
           *                   The number of actions can vary from component to component.
           *                   Suggested values are:
           *                    <ul>
           *                      <li>"open" - when a dialog component is opened</li>
           *                      <li>"close" - when a dialog component is closed</li>
           *                    </ul>
           *
           * @example <caption>Bind an event listener to the
           *          <code class="prettyprint">onOjAnimateEnd</code> property to listen for the "close"
           *          ending animation:</caption>
           * myDialog.onOjAnimateEnd = function( event )
           *   {
           *     // verify that the component firing the event is a component of interest and action
           *      is close
           *     if (event.detail.action == "close") {}
           *   };
           *
           * @example <caption>The default open and close animations are controlled via the theme
           *          (SCSS) :</caption>
           * $dialogOpenAnimation: (effect: "zoomIn", fade: true)  !default;
           * $dialogCloseAnimation: (effect: "zoomOut", fade: true)  !default;
           */
        animateEnd: null
      },
      /**
       * @memberof oj.ojDialog
       * @instance
       * @protected
       * @override
       * @return {void}
       */
    _ComponentCreate: function() {
      this._super();

      this.originalCss = {
        display: this.element[0].style.display,
        width: this.element[0].style.width,
        height: this.element[0].style.height
      };
      this.originalPosition = {
        parent: this.element.parent(),
        index: this.element.parent().children().index(this.element)
      };

      // pull the title attribute from the root element moving to an option
      this.originalTitle = this.element.attr("title");
      this.options.title = this.options.title || this.originalTitle;
      this.element.removeAttr("title");

      this.element.hide();
      this.element.uniqueId();
      this.element.addClass("oj-dialog oj-component");
      this.element.attr({
          // Setting tabIndex makes the div focusable
          'tabIndex': -1
        });

      if (!this._IsCustomElement() || !this.element[0].hasAttribute["role"])
        this.element.attr("role", this.options.role)

     this._on(this.element, {"keydown": this._keydownHandler.bind(this)});

     // fixup references to header, body and footer.  assumption is they will be immediate children
     // of the root node.
     this.userDefinedDialogHeader = false;

     if (!this._IsCustomElement()) {

        var children = this.element.children();
        for (var i = 0; i < children.length; i++) {
          var child = $(children[i]);
          if (child.is(".oj-dialog-header")) {
            this.userDefinedDialogHeader = true;
            this._userDefinedHeader = child;
            this._userDefinedHeaderDiv = children[i];
          }
          else if (child.is(".oj-dialog-body")) {

            this._createContentDiv();
            this._uiDialogContent = $(this._contentDiv);
            //
            // insert content after the body, e.g.
            // <div class='oj-dialog-body'>
            // <div class='oj-dialog-content'>
            //
            this.element[0].insertBefore(this._contentDiv, children[i]); // @HTMLUpdateOK
            oj.Components.subtreeAttached(this._contentDiv);
            //
            // Then make content the parent of body, e.g.
            // <div class='oj-dialog-content'>
            //   <div class='oj-dialog-body'>
            //
            this._contentDiv.appendChild(children[i]); // @HTMLUpdateOK
            oj.Components.subtreeAttached(children[i]);

            this._uiDialogBody = child;
            this._uiDialogBodyDiv = children[i];
          }
          else if (child.is(".oj-dialog-footer")) {
            this._uiDialogFooter = child;
            this._uiDialogFooterDiv = children[i];
          }
        }
     }

     if (this._IsCustomElement()) {
       this._processSlottedChildren();
     }

     // fixup dialog header
     if (this.userDefinedDialogHeader) {
       this._userDefinedTitleDiv = this._userDefinedHeaderDiv.querySelector('.oj-dialog-title');
       this._userDefinedTitle = $(this._userDefinedTitleDiv);

       if (this.options.cancelBehavior === "icon") {
         this._createCloseButton(this._userDefinedHeaderDiv);
       }

       if (this._userDefinedTitleDiv != null) {
         // create an id for the user-defined title
         this._userDefinedTitle.uniqueId();
         // to meet accessibility requirements for user-defined headers,
          // associate the title id with the .oj-dialog aria-labelledby.
          this.element.attr({"aria-labelledby": this._userDefinedTitle.attr("id")});
        }
     } else {
       this._createTitlebar();
     }

      if (this.options.dragAffordance === "title-bar" && $.fn.draggable) {
        this._makeDraggable();
      }

      // body was not provided. insert the content between the header and footer
      if (!this._uiDialogContent) {

        this._createContentDiv();
        this._uiDialogContent = $(this._contentDiv);
        var content =  $(this._contentDiv);

        if (this._userDefinedHeader) {
          this.element[0].insertBefore(this._contentDiv, this._userDefinedHeaderDiv); // @HTMLUpdateOK
        } else if (this._uiDialogTitlebar) {
          this.element[0].insertBefore(this._contentDiv, this._uiDialogTitlebarDiv); // @HTMLUpdateOK
        } else if (this._uiDialogFooter) {
          this.element[0].insertBefore(this._uiDialogFooterDiv, this._contentDiv); // @HTMLUpdateOK
        } else {
          this.element[0].appendChild(this._contentDiv); // @HTMLUpdateOK
        }
        oj.Components.subtreeAttached(this._contentDiv);
      }

      this._setupFocus(this.element);

      // fixup the position option set via the widget constructor
      var options = this.options;
      options["position"] = oj.PositionUtils.coerceToJet(options["position"]);
    },


    // Create the header slot element
    _createHeaderSlot: function() {

      if (this._userDefinedHeader) return;

      this._headerSlot = document.createElement("div");
      this._headerSlot.classList.add(OJD_HEADER);

      this.element[0].appendChild(this._headerSlot); // @HTMLUpdateOK
      oj.Components.subtreeAttached(this._headerSlot);

      this.userDefinedDialogHeader = true;
      this._userDefinedHeaderDiv = this._headerSlot;
      this._userDefinedHeader = $(this._headerSlot);

    },

    // Create the footer slot element.
    _createFooterSlot: function() {

      if (this._uiDialogFooter) return;

      this._footerSlot = document.createElement("div");

      this.element[0].appendChild(this._footerSlot); // @HTMLUpdateOK
      oj.Components.subtreeAttached(this._footerSlot);
      this._uiDialogFooterDiv = this._footerSlot;
      this._uiDialogFooter = $(this._footerSlot);

    },

    _createContentDiv: function() {
      this._contentDiv = document.createElement("div");
      this._contentDiv.classList.add(OJD_CONTENT, "oj-dialog-default-content");
    },

    // Create the body slot element
    _createBodySlot: function() {

      // only create the wrapper once.
      if (this._uiDialogBody) return;

      this._createContentDiv();

      this.element[0].appendChild(this._contentDiv);  // @HTMLUpdateOK
      oj.Components.subtreeAttached(this._contentDiv);

      this._bodySlot = document.createElement("div");

      this._contentDiv.appendChild(this._bodySlot); // @HTMLUpdateOK
      this._uiDialogContent = $(this._contentDiv);

      this._uiDialogBodyDiv = this._bodySlot;
      this._uiDialogBody = $(this._bodySlot);
    },


    // Process any slotted children and move them into the correct location
    _processSlottedChildren: function() {

      if (this._footerSlot != null) {
        this.element[0].removeChild(this._footerSlot);
      }
      if (this._headerSlot != null) {
        this.element[0].removeChild(this._headerSlot);
      }
      if (this._bodySlot != null) {
        this.element[0].removeChild(this._bodySlot);
      }

      var slotMap = oj.BaseCustomElementBridge.getSlotMap(this.element[0]);
      for (var slot in slotMap) {
        if (slotMap.hasOwnProperty(slot)) {
          if (slot != "header" && slot != "footer" && slot != "body" && slot != "") {
            // silently remove as per custom component slot behavior
            slotMap[slot].parentNode.removeChild(slotMap[slot]);
          }
        }
      }

      if (slotMap.hasOwnProperty("header")) {
        this._createHeaderSlot();
      }

      // Note that the default slot is the body slot.
      if (slotMap.hasOwnProperty("body") || slotMap.hasOwnProperty("")) {
        this._createBodySlot();
      }

      if (slotMap.hasOwnProperty("footer")) {
        this._createFooterSlot();
      }

      var slotParent = this._bodySlot;
      for (var slot in slotMap) {
        if (slotMap.hasOwnProperty(slot)) {
          switch (slot) {
          case "header":
            // Note - the header is wrapped with the title for accessibility,
            // so we add the oj-dialog-header class during wrap process.
            slotParent = this._headerSlot;
            break;
          case "footer":
            slotParent = this._footerSlot;
            break;
          case "body":
          case "":
            slotParent = this._bodySlot;
            break;
          }

          var slotElements = slotMap[slot];
          if (slotElements != null) {
            for (var i = 0; i < slotElements.length; i++) {
              slotParent.appendChild(slotElements[i]); // @HTMLUpdateOK
              switch (slot) {
              case "header":
                break;
              case "footer":
                slotParent = this._footerSlot;
                slotMap[slot][i].classList.add(OJD_FOOTER);
                break;
              case "body":
              case "":
                slotParent = this._bodySlot;
                slotMap[slot][i].classList.add(OJD_BODY);
                break;
              }
            }
          }
        }
      }
    },

    /**
     * @memberof oj.ojDialog
     * @instance
     * @protected
     * @override
     */
    _AfterCreateEvent: function() {

      if (this.options.initialVisibility === "show") {
        this.open();
      }

    },
    /**
     * @memberof oj.ojDialog
     * @instance
     * @protected
     * @override
     */
    _destroy: function() {

      this._off(this.element, "keydown");

      if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
        this._closeImplicitly();

      this._setWhenReady("none");

      var isDraggable = this.element.hasClass(OJ_DRAGGABLE);

      if (this._resizableComponent) {
        if (this._resizableComponent("instance"))
          this._resizableComponent("destroy");
        this._resizableComponent = null;
      }

      this._destroyCloseButton();

      if (this.userDefinedDialogHeader) {
        // remove any unique id from the user-defined header's title
        this._userDefinedTitle.removeUniqueId();
      }

      if (this.uiDialogTitle) {
        this.uiDialogTitle.remove();
        this.uiDialogTitle = null;
      }

      if (this._uiDialogContent) {
        if (this._uiDialogBody) {
          // unwrap the dialog body from the content element.
          this._uiDialogBody.insertAfter(this._uiDialogContent);  // @HTMLUpdateOK safe manipulation
        }
        this._uiDialogContent.remove();
        this._uiDialogBody = this._uiDialogContent = null;
      }

      this.element
        .removeUniqueId()
        .removeClass("oj-dialog oj-component")
        .css(this.originalCss);

      this.element.stop(true, true);

      if (this.originalTitle) {
        this.element.attr("title", this.originalTitle);
      }

      // causes testing problems.

      if (this._uiDialogTitlebar) {
        this._uiDialogTitlebar.remove();
        this._uiDialogTitlebar = null;
      }

      delete this._popupServiceEvents;
      this._super();
    },
    disable: $.noop,
    enable: $.noop,
    /**
     * Closes the dialog.
     *
     * @expose
     * @name oj.ojDialog#close
     * @method
     * @memberof oj.ojDialog
     * @instance
     * @return {void}
     * @fires oj.ojDialog#beforeClose
     * @fires oj.ojDialog#close
     * @fires oj.ojDialog#ojAnimationStart
     * @fires oj.ojDialog#ojAnimationEnd
     *
     * @example <caption>Invoke the <code class="prettyprint">close</code> method:</caption>
     * myDialog.close();
     */
    close: function(event) {
      if (this._isOperationPending("close", [event]))
        return;

      if (oj.ZOrderUtils.getStatus(this.element) !== oj.ZOrderUtils.STATUS.OPEN)
        return;

      if (this._trigger("beforeClose", event) === false && !this._ignoreBeforeCloseResultant) {
        return;
      }

      this._setWhenReady("close");
      this._focusedElement = null;

      if (!this.opener.filter(":focusable").focus().length) {
        // Hiding a focused element doesn't trigger blur in WebKit
        // so in case we have nothing to focus on, explicitly blur the active element
        // https://bugs.webkit.org/show_bug.cgi?id=47182
        $(this.document[0].activeElement).blur();
      }

      // if dialog modality is modal, check if we need
      // to restore the disabled accesskey attributes
      if (this.options.modality === 'modal'){
         var forEach = Array.prototype.forEach;
         // Find elements within dialog that have accesskey and remove marker added during open
         var elementsInDialogWithAccesskey = this.element[0].getElementsByClassName('oj-helper-element-in-dialog-with-accesskey');
         forEach.call(elementsInDialogWithAccesskey, function(element){
            element.classList.remove('oj-helper-element-in-dialog-with-accesskey');
         });
         //Find elements with oj-helper-element-with-accesskey class, get accesskey value from data attr, set accesskey attr, remove class
         var elementsInDOMWithAccesskey = document.getElementsByClassName('oj-helper-element-with-accesskey');
         forEach.call(elementsInDOMWithAccesskey, function(element){
            element.setAttribute('accesskey', element.getAttribute('data-ojAccessKey'));
            element.removeAttribute('data-ojAccessKey');
            element.classList.remove('oj-helper-element-with-accesskey');
         });
      }

      /** @type {!Object.<oj.PopupService.OPTION, ?>} */
      var psOptions = {};
      psOptions[oj.PopupService.OPTION.POPUP] = this.element;
      psOptions[oj.PopupService.OPTION.CONTEXT] = {"closeEvent" : event};
      oj.PopupService.getInstance().close(psOptions);
    },
    /**
     * Before callback is invoked while the dialog is still visible and still parented in the zorder container.
     * Close animation is performed here.
     * @memberof oj.ojDialog
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for closing the popup
     * @return {Promise|void}
     */
    _beforeCloseHandler : function (psOptions)
    {
      var rootElement = psOptions[oj.PopupService.OPTION.POPUP];

      var animationOptions = (oj.ThemeUtils.parseJSONFromFontFamily(OJD_OPTION_DEFAULTS)
        || {})["animation"];
      if (!this._ignoreBeforeCloseResultant && animationOptions && animationOptions["close"])
      {
        var style = rootElement.attr("style");
        /** @type {?} */
        var promise = oj.AnimationUtils.startAnimation(rootElement[0], "close",
          animationOptions["close"], this).then(function ()
        {
          rootElement.attr("style", style);
          rootElement.hide();
        });
        return promise;
      }
      else
      {
        rootElement.hide();
        return void(0);
      }
    },
    /**
     * Close finalization callback.
     *
     * @memberof oj.ojDialog
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for closing the popup
     * @return {void}
     */
    _afterCloseHandler : function (psOptions)
    {
      var context = psOptions[oj.PopupService.OPTION.CONTEXT];
      var event;
      if (context)
        event = context["closeEvent"];

      this._trigger("close", event);
    },
    /**
     * <p>Returns true if the dialog is currently open. This method does not accept any arguments.
     * </p>
     *
     * The "open" state reflects the period of time the dialog is visible, including open and
     * close animations.
     *
     * @expose
     * @method
     * @name oj.ojDialog#isOpen
     * @memberof oj.ojDialog
     * @instance
     * @return {boolean} <code>true</code> if the dialog is open.
     *
     * @example <caption>Invoke the <code class="prettyprint">isOpen</code> method:</caption>
     * var isOpen = myDialog.isOpen();
     */
    isOpen: function() {
      var status = oj.ZOrderUtils.getStatus(this.element);
      return (status === oj.ZOrderUtils.STATUS.OPENING ||
              status === oj.ZOrderUtils.STATUS.OPEN ||
              status === oj.ZOrderUtils.STATUS.CLOSING);
    },
    /**
     * Opens the dialog.
     *
     * @expose
     * @method
     * @name oj.ojDialog#open
     * @memberof oj.ojDialog
     * @instance
     * @return {void}
     * @fires oj.ojDialog#beforeOpen
     * @fires oj.ojDialog#open
     * @fires oj.ojDialog#ojAnimationStart
     * @fires oj.ojDialog#ojAnimationEnd
     *
     * @example <caption>Invoke the <code class="prettyprint">open</code> method:</caption>
     * var open = myDialog.open();
     */
    open: function(event) {
      if (this._isOperationPending("open", [event]))
        return;

      // this.$element.on('click.ojDialog', $.proxy(this.uiDialog.hide, this));

      if (this._trigger("beforeOpen", event) === false) {
        return;
      }

      if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN) {
        this._focusTabbable();
        return;
      }

      this._setWhenReady("open");

      this.opener = $(this.document[0].activeElement);

      if (this.options.resizeBehavior === "resizable") {
        this._makeResizable();
      }

      // normalize alignments, so that start and end keywords work as expected.
      var isRtl = this._GetReadingDirection() === "rtl";
      var position = oj.PositionUtils.coerceToJqUi(this.options["position"]);
      position = oj.PositionUtils.normalizeHorizontalAlignment(position, isRtl);

      // if modality is set to modal, prevent accesskey events
      // from being triggered while dialog is open
      if (this.options.modality === 'modal'){
         var forEach = Array.prototype.forEach
         // Mark elements within the dialog that have an accesskey attr. Those shouldn't have accesskey attr removed
         var elementsInDialogWithAccesskey = this.element[0].querySelectorAll('[accesskey]');
         forEach.call(elementsInDialogWithAccesskey, function(element){
            element.classList.add('oj-helper-element-in-dialog-with-accesskey');
         });
         // Mark elements with accesskey attr, move accesskey value to data attr, remove accesskey attr from elements
         var elementsInDOMWithAccesskey = document.querySelectorAll('[accesskey]');
         forEach.call(elementsInDOMWithAccesskey, function(element){
            if (!element.classList.contains("oj-helper-element-in-dialog-with-accesskey")){
               element.classList.add('oj-helper-element-with-accesskey');
               element.setAttribute('data-ojAccessKey', element.getAttribute('accesskey'));
               element.removeAttribute('accesskey');
            }
         });
      }

      /** @type {!Object.<oj.PopupService.OPTION, ?>} */
      var psOptions = {};
      psOptions[oj.PopupService.OPTION.POPUP] = this.element;
      psOptions[oj.PopupService.OPTION.LAUNCHER] = this.opener;
      psOptions[oj.PopupService.OPTION.POSITION] = position;
      psOptions[oj.PopupService.OPTION.MODALITY] = this.options.modality;
      psOptions[oj.PopupService.OPTION.EVENTS] = this._getPopupServiceEvents();
      psOptions[oj.PopupService.OPTION.LAYER_SELECTORS] = "oj-dialog-layer";
      psOptions[oj.PopupService.OPTION.LAYER_LEVEL] = oj.PopupService.LAYER_LEVEL.TOP_LEVEL;
      psOptions[oj.PopupService.OPTION.CUSTOM_ELEMENT] = this._IsCustomElement();
      oj.PopupService.getInstance().open(psOptions);
    },
    /**
     * Before open callback is called after the dialog has been reparented into the
     * zorder container. Open animation is performed here.
     * @memberof oj.ojDialog
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for opening the popup
     * @return {Promise|void}
     */
    _beforeOpenHandler : function (psOptions)
    {
      var rootElement = psOptions[oj.PopupService.OPTION.POPUP];
      var position = psOptions[oj.PopupService.OPTION.POSITION];

      rootElement.show();
      rootElement["position"](position);

      // We add .oj-animate-open when the dialog is animating on open.
      // This supports maintaing the visibility of a nested dialog during animation open.
      rootElement.parent().addClass('oj-animate-open');

      var animationOptions = (oj.ThemeUtils.parseJSONFromFontFamily(OJD_OPTION_DEFAULTS) ||
        {})["animation"];
      if (animationOptions && animationOptions["open"])
      {
        return oj.AnimationUtils.startAnimation(rootElement[0], "open",
          animationOptions["open"], this);
      }
      else
      {
        return void(0);
      }
    },
    /**
     * Called after the dialog is shown. Perform open finalization.
     * @memberof oj.ojDialog
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for opening the popup
     * @return {void}
     */
    _afterOpenHandler : function (psOptions)
    {
      var rootElement = psOptions[oj.PopupService.OPTION.POPUP];
      rootElement.parent().removeClass('oj-animate-open');
      this._trigger("open");
      this._focusTabbable();
    },
    /**     
     * Refresh the dialog.
     * Typically used after dynamic content is added to a dialog.
     * @expose
     * @method
     * @name oj.ojDialog#refresh
     * @memberof oj.ojDialog
     * @instance     
     * @return {void}          
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myDialog.refresh();
     */
    refresh: function()
    {
      this._super();
    },
    _focusTabbable: function() {

      var hasFocus = this._focusedElement;
      if (hasFocus && hasFocus.length > 0)
      {
        // if dialog already has focus then return
          if (oj.DomUtils.isAncestorOrSelf(this.element[0], hasFocus[0])) {
              return;
          }
      }

      // Set focus to the first match:
      // 1. First element inside the dialog matching [autofocus]
      // 2. Tabbable element inside the content element
      // 3. Tabbable element inside the footer
      // 4. The close button
      // 5. The dialog itself

      if (!hasFocus) {
        hasFocus = this.element.find("[autofocus]");
      }
      if (!hasFocus.length) {
        hasFocus = this._uiDialogContent.find(":tabbable");
      }
      if (!hasFocus.length) {
        if (this._uiDialogFooter && this._uiDialogFooter.length) {
          hasFocus = this._uiDialogFooter.find(":tabbable");
        }
      }
      if (!hasFocus.length) {
        if (this.closeButton)
          hasFocus = this.closeButton.filter(":focusable");
      }
      if (!hasFocus.length) {
        hasFocus = this.element;
      }

      if (hasFocus.length > 0) {
        hasFocus.eq(0).focus();
        this._trigger("focus");
      }
    },

    '_keepFocus': function(event) {
      function checkFocus() {
        var activeElement = this.document[0].activeElement,
            // isActive = this._uiDialogContent[0] === activeElement || $.contains(this._uiDialogContent[0], activeElement);
          isActive = this.element === activeElement || $.contains(this.element, activeElement);
        if (!isActive) {
          this._focusTabbable();
        }
      }
      event.preventDefault();
      checkFocus.call(this);
    },

    _isNumber: function(value) {
      return !isNaN(parseInt(value, 10));
    },

    _keydownHandler: function (event)
    {

      if (this.options.cancelBehavior !== "none" && !event.isDefaultPrevented() && event.keyCode &&
          event.keyCode === $.ui.keyCode.ESCAPE) {
        event.preventDefault();
        event.stopImmediatePropagation();
        this.close(event);
        return;
      }

      if (event.keyCode !== $.ui.keyCode.TAB) {
        return;
      }

      // prevent tabbing out of dialogs
      // var tabbables = this._uiDialogContent.find(":tabbable"),
      var tabbables = this.element.find(":tabbable"),
          first = tabbables.filter(":first"),
          last = tabbables.filter(":last");

      var index;

      if (!event.shiftKey)
      {

        // if (event.target === last[0] || event.target === this._uiDialogContent[0]) {

        // Check document.activeElement instead of event.target since descendant
        // elements such as ojTable may change focus when handling Tab key.
        // This aligns with browser behavior because it determines next tabstop
        // based on activeElement.
        if (document.activeElement === last[0] || document.activeElement === this.element[0]) {
          first.focus();
          event.preventDefault();
        } else {

          //
          // Make sure the first dialog tabbable (the header icon)
          // does not tab out of the dialog.
          //
          index = tabbables.index(document.activeElement);

          if (index === 0) {
            if (tabbables[1]) {
              tabbables[1].focus();
              event.preventDefault();
            }
          }
        }
      }
      else if (event.shiftKey) {
        //
        // For SHIFT-TAB, we reverse the tab order.
        //

        // Check document.activeElement instead of event.target since descendant
        // elements such as ojTable may change focus when handling Tab key.
        // This aligns with browser behavior because it determines next tabstop
        // based on activeElement.
        if (document.activeElement === first[0] || document.activeElement === this.element[0]) {
          last.focus();
          event.preventDefault();
        } else {
          //
          // Make sure the second dialog tabbable tabs back to the header
          //
          index = tabbables.index(document.activeElement);

          if (index === 1) {
            if (tabbables[0]) {
              tabbables[0].focus();
              event.preventDefault();
            }
          }
        }
      }
    },

    //
    // Invoke focusable on the passed element.
    // Called on two distinct elements - the outer dialog,
    // and the closeWrapper (assuming that there is an x-icon in the dialog)
    //
    _setupFocus: function(elem) {

        var self = this;
        this._focusable({
            'applyHighlight': true,
            'setupHandlers': function( focusInHandler, focusOutHandler) {
                self._on( elem, {
                    focus: function( event ) {
                        focusInHandler($( event.currentTarget ));
                    },
                    blur: function( event ) {
                        focusOutHandler($( event.currentTarget ));
                    }
                });
            }
        });
    },

    _destroyCloseButton: function() {

      if (this.closeButtonDiv != null) {
        if (this.closeButtonDiv.parentElement) {
          oj.Components.subtreeDetached(this.closeButtonDiv);
          this.closeButtonDiv.parentElement.removeChild(this.closeButtonDiv); // @HTMLUpdateOK
        }

        this.closeButton = null;
      }

      if (this._headerSlot)
        this._headerSlot.classList.remove(OJD_HEADER_CLOSE);
      if (this._uiDialogTitlebarDiv)
        this._uiDialogTitlebarDiv.classList.remove(OJD_HEADER_CLOSE);

    },

    //
    // Create a close button.
    // Needed for user-defined headers.
    //
    _createCloseButton: function(divParentElement) {

      // use oj-button for custom element implementations
      if (this._IsCustomElement()) {

        this.closeButtonDiv = document.createElement('oj-button');
        this.closeButtonDiv.classList.add(OJD_HEADER_CLOSE_WRAPPER);
        this.closeButtonDiv.setAttribute('data-oj-binding-provider', 'none')
        this.closeButtonDiv.setAttribute('display', 'icons')
        this.closeButtonDiv.setAttribute('chroming', 'half')

        var closeButtonLabel = document.createElement('span');
        closeButtonLabel.textContent = this.getTranslatedString('labelCloseIcon');

        var closeButtonStartIcon = document.createElement('span');
        closeButtonStartIcon.className = 'oj-fwk-icon oj-fwk-icon-cross';
        closeButtonStartIcon.setAttribute('slot', 'startIcon');

        this.closeButtonDiv.appendChild(closeButtonStartIcon);
        this.closeButtonDiv.appendChild(closeButtonLabel);

        divParentElement.appendChild(this.closeButtonDiv); // @HTMLUpdateOK
        oj.Components.subtreeAttached(this.closeButtonDiv);

        this.closeButton = $(this.closeButtonDiv);

      }

      if (!this._IsCustomElement()) {

        this.closeButton = $("<button><\button>")
          .addClass(OJD_HEADER_CLOSE_WRAPPER);

        this.closeButton.ojButton(
          {display: 'icons',
           chroming: 'half',
           label: this.getTranslatedString('labelCloseIcon'),
           icons: {start: 'oj-component-icon oj-fwk-icon-cross'}})
          .attr("tabindex", "1")
          .appendTo(divParentElement); // @HTMLUpdateOK

        this.closeButtonDiv = this.closeButton[0];

      }

      this._on(this.closeButton, {
        click: function(event) {
          event.preventDefault();
          event.stopImmediatePropagation();
          this.close(event);
        },
      });

      // When the close button is present, add a class to the title bar.
      // This is used to conditionally apply style changes to the title div.
      if (this._headerSlot)
        this._headerSlot.classList.add(OJD_HEADER_CLOSE);
      if (this._uiDialogTitlebarDiv)
        this._uiDialogTitlebarDiv.classList.add(OJD_HEADER_CLOSE);

      // no need to do this - buttons handle focus on their own.
      // var hasFocus = this.closeButton;
      // this._setupFocus(hasFocus);

    },

    _createTitlebar: function() {

      this._uiDialogTitlebarDiv = document.createElement("div");
      this._uiDialogTitlebarDiv.classList.add(OJD_HEADER);
      this.element[0].insertBefore(this._uiDialogTitlebarDiv, this.element[0].firstChild);  // @HTMLUpdateOK
      oj.Components.subtreeAttached(this._uiDialogTitlebarDiv);

      this._uiDialogTitlebar = $(this._uiDialogTitlebarDiv);

      this._on(this._uiDialogTitlebar, {
        mousedown: function(event) {
          // Don't prevent click on close button (#8838)
          // Focusing a dialog that is partially scrolled out of view
          // causes the browser to scroll it into view, preventing the click event
          // Note that the implementation will search for the dialog close wrapper class 
          // in parent, grandparent, etc., in order to handle custom element vs. widget syntax,
          // as well as handle the differences in browser event implementions (e.g. IE reports
          // a shallower button stack for button click events).
          //
          var isCloseButton = oj.DomUtils.isAncestorOrSelf(this.closeButtonDiv, event.target);
          if (!isCloseButton) {
            // Set focus to the dialog if we are dragging by the header
            this.element.focus();
          }
        }
      });

      var uiDialogTitleDiv = document.createElement("div");
      uiDialogTitleDiv.classList.add("oj-dialog-title");
      $(uiDialogTitleDiv).uniqueId();
      this._uiDialogTitlebarDiv.appendChild(uiDialogTitleDiv) // @HTMLUpdateOK
      oj.Components.subtreeAttached(uiDialogTitleDiv);

      this._title(uiDialogTitleDiv);

      this.element.attr({
        "aria-labelledby": uiDialogTitleDiv.id
      });

      if (this.options.cancelBehavior === "icon") {
        this._createCloseButton(this._uiDialogTitlebarDiv);
      }
    },

    _title: function(title) {
      if (!this.options.title) {
        title.innerHTML = "&#160;";  // @HTMLUpdateOK
      }
      title.textContent = this.options.title;
    },

    _makeDraggable: function() {
      var that = this,
        options = this.options;

      function filteredUi(ui) {
        return {
          position: ui.position,
          offset: ui.offset
        };
      }

      this.element.draggable({
        addClasses: false,
        handle: ".oj-dialog-header",
        containment: "document",
        start: function(event, ui) {
          $(this).addClass("oj-dialog-dragging");
          that._positionDescendents();
          that._trigger("dragStart", event, filteredUi(ui));
        },
        drag: function(event, ui) {
          //
          // call positionDescendents so that any descendents,
          // such as a pulldown menu, will be repositioned as the dialog is dragged.
          //
          that._positionDescendents();
          that._trigger("drag", event, filteredUi(ui));
        },
        stop: function(event, ui) {
          var left = ui.offset.left - that.document.scrollLeft();
          var top = ui.offset.top - that.document.scrollTop();

          options.position = {
            "my": {"horizontal": "left", "vertical": "top"},
            "at": {"horizontal": "left", "vertical": "top"},
            "offset": {"x": left >= 0 ? left : 0, "y": top >= 0 ? top : 0},
            "of": window
          };

          $(this).removeClass("oj-dialog-dragging");
          that._positionDescendents();
          that._trigger("dragStop", event, filteredUi(ui));
        }
      });

      this.element.addClass(OJ_DRAGGABLE);

    },
    _makeResizable: function() {

      var that = this,
        options = this.options,
        // handles = options.resizable,
        position = this.element.css("position"),
        // resizeHandles = typeof handles === "string" ? handles : "n,e,s,w,se,sw,ne,nw";

        resizeHandles = "n,e,s,w,se,sw,ne,nw";

      function filteredUi(ui) {
        return {
          'originalPosition': ui.originalPosition,
          'originalSize': ui.originalSize,
          position: ui.position,
          size: ui.size
        };
      }

      this._resizableComponent = this.element['ojResizable'].bind(this.element);

      this._resizableComponent({
        cancel: ".oj-dialog-content",
        containment: "document",
        handles: resizeHandles,
        start: function(event, ui) {

          that._isResizing = true;

          $(this).addClass("oj-dialog-resizing");
          // fire resizestart
          that._trigger("resizeStart", event, filteredUi(ui));

        },
        resize: function(event, ui) {
          that._trigger("resize", event, filteredUi(ui));
        },
        stop: function(event, ui) {

          that._isResizing = false;

          $(this).removeClass("oj-dialog-resizing");
          that._trigger("resizeStop", event, filteredUi(ui));
        }
      });

    },
    _position: function() {

      //
      // Extended position objects with better names to support RTL.
      //
      var isRtl = this._GetReadingDirection() === "rtl";
      var position = oj.PositionUtils.coerceToJqUi(this.options["position"]);
      position = oj.PositionUtils.normalizeHorizontalAlignment(position, isRtl);
      this.element.position(position);

      this._positionDescendents();
    },
    _positionDescendents: function() {

      // trigger refresh of descendents
      oj.PopupService.getInstance().triggerOnDescendents(this.element, oj.PopupService.EVENT.POPUP_REFRESH);
    },
    _setOption: function(key, value, flags) {
      /*jshint maxcomplexity:15*/
      var isDraggable, isResizable;

      // don't allow a dialog to be disabled.
      if (key === "disabled") {
        return;
      }

      this._super(key, value, flags);

      switch (key) {
      case "dragAffordance":

        // isDraggable = uiDialog.is(":data(oj-draggable)");
        isDraggable = this.element.hasClass(OJ_DRAGGABLE);

        if (isDraggable && value === "none") {
          this.element.draggable("destroy");
          this.element.removeClass(OJ_DRAGGABLE);
        }

        if (!isDraggable && value === "title-bar") {
          this._makeDraggable();
        }

        break;

      case "position":
        // convert to the internal position format and reevaluate the position.
        var options = this.options;
        options["position"] = oj.PositionUtils.coerceToJet(value, options["position"]);
        this._position();

        // setting the option is handled here.  don't call on super.
        return;

      case "resizeBehavior":

        isResizable = false;
        if (this._resizableComponent)
          isResizable = true;

        // currently resizable, becoming non-resizable
        if (isResizable && value !== "resizable") {
          // uiDialog._resizableComponent("destroy");
          if (this._resizableComponent("instance"))
            this._resizableComponent("destroy");
          this._resizableComponent = null;
        }

        // currently non-resizable, becoming resizable
        if (!isResizable && value === "resizable") {
          this._makeResizable();
        }

        break;

      case "title":

        if (this.userDefinedDialogHeader) {
          this._title(this._userDefinedHeaderDiv.querySelector('.oj-dialog-title'));
        } else {
          this._title(this._uiDialogTitlebarDiv.querySelector(".oj-dialog-title"));
        }
        break;

      case "role":
        this.element.attr("role", value);
        break;

      case "modality":
        if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
        {
          /** @type {!Object.<oj.PopupService.OPTION, ?>} */
          var psOptions = {};
          psOptions[oj.PopupService.OPTION.POPUP] = this.element;
          psOptions[oj.PopupService.OPTION.MODALITY] = value;
          oj.PopupService.getInstance().changeOptions(psOptions);
        }
        break;

      case "cancelBehavior":

        if (value === "none" || value === "escape") {

          // we may need additional code here
          // if (this.userDefinedDialogHeader) {   }

          this._destroyCloseButton();

        }
        else if (value === "icon") {

          if (this.userDefinedDialogHeader) {

            this._destroyCloseButton();
            this._createCloseButton(this._userDefinedHeaderDiv);

            //
            // Insert oj-dialog-title between oj-dialog-header and oj-dialog-header-close-wrapper
            //
            this._userDefinedTitleDiv = this._userDefinedHeaderDiv.querySelector('.oj-dialog-title');
            this._userDefinedTitle = $(this._userDefinedTitleDiv);

          } else {

            this._destroyCloseButton();
            this._createCloseButton(this._uiDialogTitlebarDiv);

            this.standardTitleDiv = this._uiDialogTitlebarDiv.querySelector(".oj-dialog-title");
            this.standardTitle = $(this.standardTitleDiv);

          }
        }
        break;
      }
    },

    getNodeBySubId: function(locator) {
      if (locator === null) {
        return this.element ? this.element[0] : null;
      }

      function _escapeId(id) {
        var targetId = [];
        var regex = /\w|_|-/;

        for (var i = 0; i < id.length; i++) {
          var c = id.substring(i, i + 1);
          if (regex.test(c))
            targetId.push(c);
          else
            targetId.push("\\" + c);
        }
        return targetId.join("");
      }

      var subId = locator['subId'];

      //
      // Use slot structure to return body and footer subids.
      //
      if (this._IsCustomElement() && (subId === OJD_FOOTER || subId === OJD_BODY)) {
        if (subId === OJD_BODY) {
          return this._uiDialogBodyDiv.querySelector('.oj-dialog-body');
        }
        else if (subId === OJD_FOOTER) {
          return this._uiDialogFooterDiv.querySelector('.oj-dialog-footer');
        }
      } else {

        // General case
        switch (subId) {

        case OJD_HEADER:
        case OJD_FOOTER:
        case OJD_CONTENT:
        case OJ_RESIZABLE_N:
        case OJ_RESIZABLE_E:
        case OJ_RESIZABLE_S:
        case OJ_RESIZABLE_W:
        case OJ_RESIZABLE_SE:
        case OJ_RESIZABLE_SW:
        case OJ_RESIZABLE_NE:
        case OJ_RESIZABLE_NW:

          var selector = this.element[0].nodeName + '[id="' + _escapeId(this.element.attr("id")) + '"] > ';
          selector +=  "." + subId;
          var node = this.element.parent().find(selector);
          if (!node || node.length === 0)
            return null;

          return (node[0]);
          break;

          // "oj-dialog-close-icon" is deprecated as of 1.2
          // use "oj-dialog-close" instead.
          // "oj-dialog-close" is deprecated as of 2.1.*
        case "oj-dialog-close-icon":
        case "oj-dialog-close":
          return null;
          break;

          // "oj-dialog-body" is deprecated as of 1.2
        case OJD_BODY:
          var selector = this.element[0].nodeName + '[id="' + _escapeId(this.element.attr("id")) + '"] > ';
          selector += ".oj-dialog-content > ";
          selector +=  "." + subId;
          var node = this.element.parent().find(selector);
          if (!node || node.length === 0)
            return null;

          return (node[0]);
          break;

        case OJD_HEADER_CLOSE_WRAPPER:
          var selector = this.element[0].nodeName + '[id="' + _escapeId(this.element.attr("id")) + '"] > ';
          selector += ".oj-dialog-header > ";
          selector +=  "." + subId;
          var node = this.element.parent().find(selector);
          if (!node || node.length === 0)
            return null;

          return (node[0]);
          break;

        }
      }

      // Non-null locators have to be handled by the component subclasses
      return null;
    },
    //** @inheritdoc */
    getSubIdByNode: function(node)
    {

      if (node != null) {

        var nodeCached = $(node);

        if (nodeCached.hasClass(OJD_HEADER))
          return {'subId': OJD_HEADER};
        else if (nodeCached.hasClass(OJD_FOOTER))
          return {'subId': OJD_FOOTER};
        else if (nodeCached.hasClass(OJD_CONTENT))
          return {'subId': OJD_CONTENT};
        else if (nodeCached.hasClass(OJD_HEADER_CLOSE_WRAPPER))
          return {'subId': OJD_HEADER_CLOSE_WRAPPER};
        else if (nodeCached.hasClass(OJ_RESIZABLE_N))
          return {'subId': OJ_RESIZABLE_N};
        else if (nodeCached.hasClass(OJ_RESIZABLE_E))
          return {'subId': OJ_RESIZABLE_E};
        else if (nodeCached.hasClass(OJ_RESIZABLE_S))
          return {'subId': OJ_RESIZABLE_S};
        else if (nodeCached.hasClass(OJ_RESIZABLE_W))
          return {'subId': OJ_RESIZABLE_W};
        else if (nodeCached.hasClass(OJ_RESIZABLE_SE))
          return {'subId': OJ_RESIZABLE_SE};
        else if (nodeCached.hasClass(OJ_RESIZABLE_SW))
          return {'subId': OJ_RESIZABLE_SW};
        else if (nodeCached.hasClass(OJ_RESIZABLE_NE))
          return {'subId': OJ_RESIZABLE_NE};
        else if (nodeCached.hasClass(OJ_RESIZABLE_NW))
          return {'subId': OJ_RESIZABLE_NW};
      }

      return null;
    },
    /**
     * @memberof oj.ojDialog
     * @instance
     * @private
     * @return {void}
     */
    _surrogateRemoveHandler: function() {
      var element = this.element;
      element.remove();
    },
    /**
     * @memberof oj.ojDialog
     * @instance
     * @private
     * @return {!Object.<oj.PopupService.EVENT, function(...)>}
     */
    _getPopupServiceEvents: function() {
      if (!this._popupServiceEvents) {
        /** @type {!Object.<oj.PopupService.EVENT, function(...)>} **/
        var events = this._popupServiceEvents = {};
        events[oj.PopupService.EVENT.POPUP_CLOSE] = this._closeImplicitly.bind(this);
        events[oj.PopupService.EVENT.POPUP_REMOVE] = this._surrogateRemoveHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_REFRESH] = this._positionDescendents.bind(this);
        events[oj.PopupService.EVENT.POPUP_BEFORE_OPEN] = this._beforeOpenHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_AFTER_OPEN] = this._afterOpenHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_BEFORE_CLOSE] = this._beforeCloseHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_AFTER_CLOSE] = this._afterCloseHandler.bind(this);
      }
      return this._popupServiceEvents;
    },
    /**
     * @memberof oj.ojDialog
     * @instance
     * @private
     */
    _closeImplicitly: function() {
      this._ignoreBeforeCloseResultant = true;
      this.close();
      delete this._ignoreBeforeCloseResultant;
    },

    /**
     * Creates a Promise exposed by the {@link oj.ojDialog#whenReady} method.
     *
     * @param {string} operation valid values are "open", "close" or "none"
     * @memberof oj.ojDialog
     * @instance
     * @private
     */
    _setWhenReady : function (operation) {

      /** @type {oj.PopupWhenReadyMediator} */
      var mediator = this._whenReadyMediator;
      if (mediator)
      {
        mediator.destroy();
        delete this._whenReadyMediator;
      }

      // operation === none
      if (["open", "close"].indexOf(operation) < 0)
        return;

      this._whenReadyMediator = new oj.PopupWhenReadyMediator(this.element, operation, "ojDialog",
        this._IsCustomElement());
    },

    /**
     * Checks to see if there is a pending "open" or "close" operation.  If pending and it
     * is the same as the requested operation, the request silently fails.  If the current
     * operation is the inverse operation, we queue the current operation after the pending
     * operation is resolved.
     *
     * @memberof oj.ojDialog
     * @instance
     * @private
     * @param {string} operation currently requested
     * @param {Array} args passed to a queue operation
     * @returns {boolean} <code>true</code> if a "close" or "open" operation is pending completion.
     */
    _isOperationPending: function (operation, args) {
      /** @type {oj.PopupWhenReadyMediator} **/
      var mediator = this._whenReadyMediator;
      if (mediator)
        return mediator.isOperationPending(this, operation, operation, args);
      else
        return false;
    },
    /**
     * Notifies the component that its subtree has been removed from the document
     * programmatically after the component has been created.
     *
     * @memberof oj.ojDialog
     * @instance
     * @protected
     * @override
     */
     _NotifyDetached: function()
     {
       // detaching an open popup results in implicit dismissal
       if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
         this._closeImplicitly();

       this._super();
     }

    /**
     * <p>The default slot is the dialog's body. The <code class="prettyprint">&lt;oj-dialog></code>
     * element accepts DOM nodes as children for the default slot.
     * The default slot can also be named with "body".
     * For styling, the default body slot will be rendered with the <code class="prettyprint">oj-dialog-body</code> class.
     *
     * @ojchild Default
     * @memberof oj.ojDialog
     * @since 4.0.0
     *
     * @example <caption>Initialize the Dialog with body content:</caption>
     * &lt;oj-dialog>
     *   &lt;div>Dialog Content&lt;/div>
     * &lt;/oj-dialog>
     *
     * @example <caption>Initialize the Dialog with body content, explicitly naming the body slot:</caption>
     * &lt;oj-dialog>
     *   &lt;div slot="body">Dialog Content&lt;/div>
     * &lt;/oj-dialog>
     */

    /**
     * <p>The <code class="prettyprint">footer</code> slot is for the dialog's footer area.
     * The <code class="prettyprint">&lt;oj-dialog></code> element accepts DOM nodes as children
     * with the footer slot.
     * For styling, the footer body slot will be rendered with the <code class="prettyprint">oj-dialog-footer</code> class.
     *
     * @ojslot footer
     * @memberof oj.ojDialog
     * @since 4.0.0
     *
     * @example <caption>Initialize the Dialog with body and footer content:</caption>
     * &lt;oj-dialog>
     *   &lt;div>Dialog Content&lt;/div>
     *   &lt;div slot='footer'>Footer Content&lt;/div>
     * &lt;/oj-dialog>
     */

    /**
     * <p>The <code class="prettyprint">header</code> slot is for the dialog's header area.
     * The  <code class="prettyprint">&lt;oj-dialog></code> element accepts DOM nodes as children
     * with the header slot.
     * For styling, the header slot will be rendered with the <code class="prettyprint">oj-dialog-header</code> class.
     * </p>
     * If a header slot is not specified by the user, a header will automatically be created.
     * The automatically generated header will contain a close button, and the header title will be set
     * to the dialog title.
     * @ojslot header
     * @memberof oj.ojDialog
     * @since 4.0.0
     *
     * @example <caption>Initialize the Dialog with header and body content:</caption>
     * &lt;oj-dialog>
     *   &lt;div slot='header'>Header Content&lt;/div>
     *   &lt;div>Dialog Content&lt;/div>
     * &lt;/oj-dialog>
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Dialog Close Icon</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Close the dialog.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojDialog
     */

    /**
     * The JET Dialog can be closed with keyboard actions:
     *
     * <p>
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Dialog</td>
     *       <td><kbd>Esc</kbd></td>
     *       <td>Close the dialog.</td>
     *     </tr>
     *     <tr>
     *       <td>Dialog Close Icon</td>
     *       <td><kbd>Enter</kbd> or <kbd>Space</kbd></td>
     *       <td>Close the dialog.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojDialog
     */

    /**
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>Class(es)</th>
     *       <th>Description</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-dialog-header</td>
     *       <td><p> Class automatically generated on the header slot.</td>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-dialog-title</td>
     *       <td><p> Class used to format the title. Automatically created headers use <code class="prettyprint"> oj-dialog-title </code> to format the title. For user-defined headers, you may want to use the <code class="prettyprint"> oj-dialog-title </code> so that the title in your user-defined header is stylistically similar to a default title. </td>
     *     </tr>
     *     <tr>
     *       <td>oj-dialog-body</td>
     *       <td><p> Class automatically generated on the default (body) slot.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-dialog-footer</td>
     *       <td><p> Class automatically generated on the footer slot.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-dialog-footer-separator</td>
     *       <td><p>A separator between the dialog body and the dialog footer can be added by using a second style class ( <code class="prettyprint"> oj-dialog-footer-separator </code>) in the footer. So use:
     *           <ul>
     *             <li>oj-dialog-footer oj-dialog-footer-separator</li>
     *           </ul>
     *      to add a footer separator to the dialog.
     *        Note that for themes that have a built-in footer separator (specifically the iOS theme),
     *        this class has no effect.
     *     <p>See the demo section for a live example of the footer separator. </td>
     *     </tr>
     *     <tr>
     *       <td>oj-progress-bar-embedded</td>
     *       <td><p> Optional markup. Used to format a progress bar embedded in the dialog header.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-focus-highlight</td>
     *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
     *     </tr>
     *
     *   </tbody>
     * </table>
     *
     * <p> Note that the dialog component wraps additional divs around the user's content and also performs other DOM manipulations.
     * Thus, the user should be careful if they wish to engage in advanced coding approaches.
     * In general, it is best to target DOM elements by id or class name
     * (e.g., developers should not rely on relative positioning of dialog DOM elements).
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojDialog
     */

    //////////////////     SUB-IDS     //////////////////
    /**
     * <p>Sub-ID for the dialog header.</p>
     *
     * @ojsubid oj-dialog-header
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-dialog-header'});
     */

    /**
     * <p>Sub-ID for the dialog footer.</p>
     *
     * @ojsubid oj-dialog-footer
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog footer:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-dialog-footer'});
     */

    /**
     * <p>Sub-ID for the dialog body.</p>
     *
     * @ojsubid oj-dialog-body
     * @memberof oj.ojDialog
     * @ojdeprecated {since:"1.2", description:"This sub-ID is not needed.  Since the application supplies this element, it can supply a unique ID by which the element can be accessed."}
     *
     * @example <caption>Get the node for the dialog body:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-dialog-body'});
     */

    /**
     * <p>Sub-ID for the dialog content.</p>
     *
     * @ojsubid oj-dialog-content
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog content:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-dialog-content'});
     */

    /**
     * <p>Sub-ID for the dialog header-close-wrapper.</p>
     *
     * @ojsubid oj-dialog-header-close-wrapper
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header-close-wrapper:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-dialog-header-close-wrapper'});
     */

    /**
     * <p>Sub-ID for the dialog close-icon.</p>
     *
     * @ojsubid oj-dialog-close-icon
     * @memberof oj.ojDialog
     * @ojdeprecated {since: "1.2", description: "This sub-ID is deprecated."}
     *
     * @example <caption>Get the node for the dialog close-icon:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-dialog-close-icon'});
     */

    /**
     * <p>Sub-ID for the dialog close affordance.</p>
     *
     * @ojsubid oj-dialog-close
     * @memberof oj.ojDialog
     * @deprecated {since: "2.1.0", description: "This sub-ID is deprecated."}
     *
     * @example <caption>Get the node for the dialog close affordance:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-dialog-close'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the north location.</p>
     *
     * @ojsubid oj-resizable-n
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-n'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the south location.</p>
     *
     * @ojsubid oj-resizable-s
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-s'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the east location.</p>
     *
     * @ojsubid oj-resizable-e
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-e'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the west location.</p>
     *
     * @ojsubid oj-resizable-w
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-w'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the northeast location.</p>
     *
     * @ojsubid oj-resizable-ne
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-ne'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the northwest location.</p>
     *
     * @ojsubid oj-resizable-nw
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-nw'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the southwest location.</p>
     *
     * @ojsubid oj-resizable-sw
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-sw'});
     */

    /**
     * <p>Sub-ID for the dialog resizable handle at the southeast location.</p>
     *
     * @ojsubid oj-resizable-se
     * @memberof oj.ojDialog
     *
     * @example <caption>Get the node for the dialog header:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-resizable-se'});
     */
  });

  oj.Components.setDefaultOptions(
    {
      'ojDialog':
        {
          'resizeBehavior': oj.Components.createDynamicPropertyGetter(
            function()
            {
                return (oj.ThemeUtils.parseJSONFromFontFamily(OJD_OPTION_DEFAULTS) || {})["resizeBehavior"];
            }),
          'cancelBehavior': oj.Components.createDynamicPropertyGetter(
            function()
            {
                return (oj.ThemeUtils.parseJSONFromFontFamily(OJD_OPTION_DEFAULTS) || {})["cancelBehavior"];
            }),
          'dragAffordance': oj.Components.createDynamicPropertyGetter(
            function()
            {
                return (oj.ThemeUtils.parseJSONFromFontFamily(OJD_OPTION_DEFAULTS) || {})["dragAffordance"];
            })
        }
    });

}());

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

// retrieved from https://raw.github.com/jquery/jquery-ui/1-10-stable/ui/jquery.ui.resizable.js on 04/09/2014, and then modified


/*
 *
 *   - This widget is NOT EXPOSED.
 *     ojResizable is made available only to dialog and other components that need to call resize functionality.
 *   Changes:
 *    - Options minWidth, minHeight, maxWidth, and maxHeight have been deleted
 *    - Removed zIndex option
 *    - Removed css write of zIndex (this is supported in style sheets)
 *
 */

(function() {
$.widget("oj.ojResizable", {
    version: "1.0.0",
    widgetEventPrefix: "oj",
    options: {
      /////////////////////////////////////////////////////////////////////////////////////
      //
      // Mouse Options (copied)
      //
      /////////////////////////////////////////////////////////////////////////////////////

      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      cancel: "input,textarea,button,select,option",
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      distance: 1,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      delay: 0,
      /////////////////////////////////////////////////////////////////////////////////////
      //
      // Resize Options
      //
      /////////////////////////////////////////////////////////////////////////////////////


      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      maxHeight: null,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      maxWidth: null,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      minHeight: 10,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      minWidth: 10,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      alsoResize: false,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      animate: false,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      animateDuration: "slow",
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      animateEasing: "swing",
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      containment: false,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      ghost: false,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      grid: false,
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      handles: "e,s,se",
      /**
       *
       * @private
       * @expose
       * @memberof! oj.ojResizable
       * @instance
       *
       */
      helper: false,
      // See #7960
      // zIndex: 90,

      /////////////////
      // callbacks
      /////////////////


      /**
       * Triggered when the ojResizable is resized.
       *
       * @private
       * @expose
       * @event
       * @name resize
       * @memberof! oj.ojResizable
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Empty object included for consistency with other events
       *
       * @example <caption>Initialize the resizable with the <code class="prettyprint">resize</code> callback specified:</caption>
       * $( ".selector" ).ojResizable({
       *     "resize": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojresize</code> event:</caption>
       * $( ".selector" ).on( "ojresize", function( event, ui ) {} );
       */
      resize: null,
      /**
       * Triggered on the start of a resize operation.
       *
       * @private
       * @expose
       * @event
       * @name start
       * @memberof! oj.ojResizable
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Empty object included for consistency with other events
       *
       * @example <caption>Initialize the resizable with the <code class="prettyprint">start</code> callback specified:</caption>
       * $( ".selector" ).ojResizable({
       *     "start": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojstart</code> event:</caption>
       * $( ".selector" ).on( "ojstart", function( event, ui ) {} );
       */
      // note - jqui doc has .on("resizestart"
      start: null,
      /**
       * Triggered on the end of a resize operation.
       *
       * @private
       * @expose
       * @event
       * @name stop
       * @memberof! oj.ojResizable
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Empty object included for consistency with other events
       *
       * @example <caption>Initialize the resizable with the <code class="prettyprint">stop</code> callback specified:</caption>
       * $( ".selector" ).ojResizable({
       *     "stop": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojstop</code> event:</caption>
       * $( ".selector" ).on( "ojstop", function( event, ui ) {} );
       */
      // note - jqui doc has .on("resizestop"
      stop: null
    },
    /////////////////////////////////////////////////////////////////////////////////////
    //
    // Original Resize Functions
    //
    /////////////////////////////////////////////////////////////////////////////////////

    _num: function(value) {
      return parseInt(value, 10) || 0;
    },
    _isNumber: function(value) {
      return !isNaN(parseInt(value, 10));
    },
    _hasScroll: function(el, a) {

      if ($(el).css("overflow") === "hidden") {
        return false;
      }

      var scroll = (a && a === "left") ? "scrollLeft" : "scrollTop",
        has = false;

      if (el[ scroll ] > 0) {
        return true;
      }

      // TODO: determine which cases actually cause this to happen
      // if the element doesn't have the scroll set, see if it's possible to
      // set the scroll
      el[ scroll ] = 1;
      has = (el[ scroll ] > 0);
      el[ scroll ] = 0;
      return has;
    },
    /**
     * Triggered when the ojResizable is created.
     *
     * @private
     * @expose
     * @event
     * @name create
     * @memberof! oj.ojResizable
     * @instance
     * @property {Event} event <code class="prettyprint">jQuery</code> event object
     * @property {Object} ui Empty object included for consistency with other events
     *
     * @example <caption>Initialize the resizable with the <code class="prettyprint">create</code> callback specified:</caption>
     * $( ".selector" ).ojResizable({
     *     "create": function( event, ui ) {}
     * });
     *
     * @example <caption>Bind an event listener to the <code class="prettyprint">ojcreate</code> event:</caption>
     * $( ".selector" ).on( "ojcreate", function( event, ui ) {} );
     */
    // note - jqui has on("resizecreate", ... need to verify if we need some form of "ojcreate".
    _create: function()
    {
      this._super();

      var n, i, handle, axis, hname,
        that = this, o = this.options;

      //
      // Create an instance of the 3rd party jqueryui mouse widget.
      //

      var mouseConstructor = this.element['mouse'].bind(this.element);
      mouseConstructor();
      this.mouse = mouseConstructor('instance');

      //
      // Because we aggregating the mouse widget (and not extending it),
      // we override the protected methods of this mouse instance.
      //

      this.mouse['_mouseCapture'] = function(event) {
        return that._mouseCapture(event);
      };

      this.mouse['_mouseStart'] = function(event) {
        return that._mouseStart(event);
      };

      this.mouse['_mouseDrag'] = function(event) {
        return that._mouseDrag(event);
      };

      this.mouse['_mouseStop'] = function(event) {
        if (this.element) {
          this.element.focus();
        }
        return that._mouseStop(event);
      };

      this.element.addClass("oj-resizable");

      $.extend(this, {
        originalElement: this.element,
        _proportionallyResizeElements: [],
        // _helper: o.helper || o.ghost || o.animate ? o.helper || "oj-resizable-helper" : null
        _helper: null
      });

      this._initialResize = true;

      this.handles = o.handles || (!$(".oj-resizable-handle", this.element).length ? "e,s,se" : {n: ".oj-resizable-n", e: ".oj-resizable-e", s: ".oj-resizable-s", w: ".oj-resizable-w", se: ".oj-resizable-se", sw: ".oj-resizable-sw", ne: ".oj-resizable-ne", nw: ".oj-resizable-nw"});
      if (this.handles.constructor === String) {

        if (this.handles === "all") {
          this.handles = "n,e,s,w,se,sw,ne,nw";
        }

        n = this.handles.split(",");
        this.handles = {};

        for (i = 0; i < n.length; i++) {

          handle = $.trim(n[i]);
          hname = "oj-resizable-" + handle;
          axis = $("<div class='oj-resizable-handle " + hname + "'></div>");

          // axis.css({ zIndex: o.zIndex });

          // Todo: refine for alta styles
          //
          // if ("se" === handle) {
          // axis.addClass("ui-icon ui-icon-gripsmall-diagonal-se");
          //}

          this.handles[handle] = ".oj-resizable-" + handle;
          this.element.append(axis);   // @HTMLUpdateOK
        }
      }

      this._renderAxis = function(target) {

        var i, axis, padPos, padWrapper;

        target = target || this.element;

        for (i in this.handles) {

          if (this.handles[i].constructor === String) {
            this.handles[i] = this.element.children(this.handles[ i ]).first().show();
          }

        }
      };

      // TODO: make renderAxis a prototype function
      this._renderAxis(this.element);

      this._handles = $(".oj-resizable-handle", this.element);

      this._handles.mouseover(function() {
        if (!that.resizing) {
          if (this.className) {
            axis = this.className.match(/oj-resizable-(se|sw|ne|nw|n|e|s|w)/i);
          }
          that.axis = axis && axis[1] ? axis[1] : "se";
        }
      });

      this.mouse['_mouseInit']();
    },
    /**
     * Remove the ojResizable functionality completely.
     * This will return the element back to its pre-init state.
     *
     * <p>This method does not accept any arguments.
     *
     * @private
     * @expose
     * @method
     * @name oj.ojResizable#destroy
     * @memberof! oj.ojResizable
     * @instance
     *
     * @example <caption>Invoke the <code class="prettyprint">destroy</code> method:</caption>
     * var destroy = $( ".selector" ).ojResizable( "destroy" );
     */

    _destroy: function() {

      if (this.mouse) {
        this.mouse['_mouseDestroy']();
      }

      try {
        this.mouse['destroy']();
        this.mouse = null;
      } catch (e) {
      }

      var wrapper,
        _destroy = function(exp) {
          $(exp).removeClass("oj-resizable oj-resizable-disabled oj-resizable-resizing")
            .removeData("resizable").removeData("oj-resizable").unbind(".resizable").find(".oj-resizable-handle").remove();
        };

      _destroy(this.originalElement);

      return this;
    },
    _mouseCapture: function(event) {
      var i, handle,
        capture = false;

      for (i in this.handles) {
        handle = $(this.handles[i])[0];
        if (handle === event.target || $.contains(handle, event.target)) {
          capture = true;
        }
      }

      return !this.options.disabled && capture;
    },
    _mouseStart: function(event) {

      var curleft, curtop, cursor,
        o = this.options,
        iniPos = this.element.position(),
        el = this.element;

      this.resizing = true;

      // Bugfix for http://bugs.jqueryui.com/ticket/1749
      if ((/absolute/).test(el.css("position"))) {
        el.css({position: "absolute", top: el.css("top"), left: el.css("left")});
      } else if (el.is(".oj-draggable")) {
        el.css({position: "absolute", top: iniPos.top, left: iniPos.left});
      }

      this._renderProxy();

      curleft = this._num(this.helper.css("left"));
      curtop = this._num(this.helper.css("top"));

      if (o.containment) {
        curleft += $(o.containment).scrollLeft() || 0;
        curtop += $(o.containment).scrollTop() || 0;
      }

      this.offset = this.helper.offset();
      this.position = {left: curleft, top: curtop};
      this.size = {width: el.width(), height: el.height()};
      this.originalSize = {width: el.width(), height: el.height()};
      this.originalPosition = {left: curleft, top: curtop};
      this.sizeDiff = {width: el.outerWidth() - el.width(), height: el.outerHeight() - el.height()};
      this.originalMousePosition = {left: event.pageX, top: event.pageY};

      this.aspectRatio = (this.originalSize.width / this.originalSize.height) || 1;

      // cursor = $(".oj-resizable-" + this.axis).css("cursor");
      cursor = /** @type string */ ($(".oj-resizable-" + this.axis).css("cursor"));
      $("body").css("cursor", cursor === "auto" ? this.axis + "-resize" : cursor);

      el.addClass("oj-resizable-resizing");

      this._propagate("start", event);

      this._alsoresize_start(event);
      this._containment_start(event);

      return true;
    },
    _mouseDrag: function(event) {

      var data,
        el = this.helper, props = {},
        smp = this.originalMousePosition,
        a = this.axis,
        dx = (event.pageX - smp.left) || 0,
        dy = (event.pageY - smp.top) || 0,
        trigger = this._change[a];

      this.prevPosition = {
        top: this.position.top,
        left: this.position.left
      };
      this.prevSize = {
        width: this.size.width,
        height: this.size.height
      };

      if (!trigger) {
        return false;
      }

      data = trigger.apply(this, [event, dx, dy]);

      this._updateVirtualBoundaries(event.shiftKey);
      if (event.shiftKey) {
        data = this._updateRatio(data, event);
      }

      data = this._respectSize(data, event);

      this._updateCache(data);

      this._propagate("resize", event);

      this._alsoresize_resize(event, this.ui());
      this._containment_resize(event, this.ui());

      if (this.position.top !== this.prevPosition.top) {
        props.top = this.position.top + "px";
      }
      if (this.position.left !== this.prevPosition.left) {
        props.left = this.position.left + "px";
      }
      if (this.size.width !== this.prevSize.width) {
        props.width = this.size.width + "px";
      }
      if (this.size.height !== this.prevSize.height) {
        props.height = this.size.height + "px";
      }
      el.css(props);

      if (!this._helper && this._proportionallyResizeElements.length) {
        this._proportionallyResize();
      }

      if (!$.isEmptyObject(props)) {
        this._trigger("resize", event, this.ui());
      }

      return false;
    },
    _mouseStop: function(event) {

      this.resizing = false;
      $("body").css("cursor", "auto");

      this.element.removeClass("oj-resizable-resizing");

      this._propagate("stop", event);

      this._alsoresize_stop(event);
      this._containment_stop(event);

      return false;

    },
    _updateVirtualBoundaries: function(forceAspectRatio) {
      var pMinWidth, pMaxWidth, pMinHeight, pMaxHeight, b,
        o = this.options;

      b = {
        minWidth: this._isNumber(o.minWidth) ? o.minWidth : 0,
        maxWidth: this._isNumber(o.maxWidth) ? o.maxWidth : Infinity,
        minHeight: this._isNumber(o.minHeight) ? o.minHeight : 0,
        maxHeight: this._isNumber(o.maxHeight) ? o.maxHeight : Infinity

          /*
           minWidth: 0,
           maxWidth: Infinity,
           minHeight: 0,
           maxHeight: Infinity
           */

      };

      if (forceAspectRatio) {
        pMinWidth = b.minHeight * this.aspectRatio;
        pMinHeight = b.minWidth / this.aspectRatio;
        pMaxWidth = b.maxHeight * this.aspectRatio;
        pMaxHeight = b.maxWidth / this.aspectRatio;

        if (pMinWidth > b.minWidth) {
          b.minWidth = pMinWidth;
        }
        if (pMinHeight > b.minHeight) {
          b.minHeight = pMinHeight;
        }
        if (pMaxWidth < b.maxWidth) {
          b.maxWidth = pMaxWidth;
        }
        if (pMaxHeight < b.maxHeight) {
          b.maxHeight = pMaxHeight;
        }
      }
      this._vBoundaries = b;
    },
    _updateCache: function(data) {
      this.offset = this.helper.offset();
      if (this._isNumber(data.left)) {
        this.position.left = data.left;
      }
      if (this._isNumber(data.top)) {
        this.position.top = data.top;
      }
      if (this._isNumber(data.height)) {
        this.size.height = data.height;
      }
      if (this._isNumber(data.width)) {
        this.size.width = data.width;
      }
    },
    _updateRatio: function(data) {

      var cpos = this.position,
        csize = this.size,
        a = this.axis;

      if (this._isNumber(data.height)) {
        data.width = (data.height * this.aspectRatio);
      } else if (this._isNumber(data.width)) {
        data.height = (data.width / this.aspectRatio);
      }

      if (a === "sw") {
        data.left = cpos.left + (csize.width - data.width);
        data.top = null;
      }
      if (a === "nw") {
        data.top = cpos.top + (csize.height - data.height);
        data.left = cpos.left + (csize.width - data.width);
      }

      return data;
    },
    _respectSize: function(data) {

      var o = this._vBoundaries,
        a = this.axis,
        ismaxw = this._isNumber(data.width) && o.maxWidth && (o.maxWidth < data.width), ismaxh = this._isNumber(data.height) && o.maxHeight && (o.maxHeight < data.height),
        isminw = this._isNumber(data.width) && o.minWidth && (o.minWidth > data.width), isminh = this._isNumber(data.height) && o.minHeight && (o.minHeight > data.height),
        dw = this.originalPosition.left + this.originalSize.width,
        dh = this.position.top + this.size.height,
        cw = /sw|nw|w/.test(a), ch = /nw|ne|n/.test(a);
      if (isminw) {
        data.width = o.minWidth;
      }
      if (isminh) {
        data.height = o.minHeight;
      }
      if (ismaxw) {
        data.width = o.maxWidth;
      }
      if (ismaxh) {
        data.height = o.maxHeight;
      }

      if (isminw && cw) {
        data.left = dw - o.minWidth;
      }
      if (ismaxw && cw) {
        data.left = dw - o.maxWidth;
      }
      if (isminh && ch) {
        data.top = dh - o.minHeight;
      }
      if (ismaxh && ch) {
        data.top = dh - o.maxHeight;
      }

      // Fixing jump error on top/left 
      if (!data.width && !data.height && !data.left && data.top) {
        data.top = null;
      } else if (!data.width && !data.height && !data.top && data.left) {
        data.left = null;
      }

      return data;
    },
    _proportionallyResize: function() {

      if (!this._proportionallyResizeElements.length) {
        return;
      }

      var i, j, borders, paddings, prel,
        element = this.helper || this.element;

      for (i = 0; i < this._proportionallyResizeElements.length; i++) {

        prel = this._proportionallyResizeElements[i];

        if (!this.borderDif) {
          this.borderDif = [];
          borders = [prel.css("borderTopWidth"), prel.css("borderRightWidth"), prel.css("borderBottomWidth"), prel.css("borderLeftWidth")];
          paddings = [prel.css("paddingTop"), prel.css("paddingRight"), prel.css("paddingBottom"), prel.css("paddingLeft")];

          for (j = 0; j < borders.length; j++) {
            this.borderDif[ j ] = (parseInt(borders[ j ], 10) || 0) + (parseInt(paddings[ j ], 10) || 0);
          }
        }

        prel.css({
          height: (element.height() - this.borderDif[0] - this.borderDif[2]) || 0,
          width: (element.width() - this.borderDif[1] - this.borderDif[3]) || 0
        });

      }

    },
    _renderProxy: function() {

      var el = this.element, o = this.options;
      this.elementOffset = el.offset();

      this.helper = this.element;

    },
    _change: {
      "e": function(event, dx) {
        return {width: this.originalSize.width + dx};
      },
      "w": function(event, dx) {
        var cs = this.originalSize, sp = this.originalPosition;
        return {left: sp.left + dx, width: cs.width - dx};
      },
      "n": function(event, dx, dy) {
        var cs = this.originalSize, sp = this.originalPosition;
        return {top: sp.top + dy, height: cs.height - dy};
      },
      "s": function(event, dx, dy) {
        return {height: this.originalSize.height + dy};
      },
      "se": function(event, dx, dy) {
        return $.extend(this._change["s"].apply(this, arguments), this._change["e"].apply(this, [event, dx, dy]));
      },
      "sw": function(event, dx, dy) {
        return $.extend(this._change["s"].apply(this, arguments), this._change["w"].apply(this, [event, dx, dy]));
      },
      "ne": function(event, dx, dy) {
        return $.extend(this._change["n"].apply(this, arguments), this._change["e"].apply(this, [event, dx, dy]));
      },
      "nw": function(event, dx, dy) {
        return $.extend(this._change["n"].apply(this, arguments), this._change["w"].apply(this, [event, dx, dy]));
      }
    },
    _propagate: function(n, event) {

      //
      // Propage resizeStart and resizeStop events.
      // (resize is propagated internally by drag)
      //

      // $.ui.plugin.call(this, n, [event, this.ui()]);
      (n !== "resize" && this._trigger(n, event, this.ui()));
    },
    //////////////////////////////////////////////////////////////////////////////////
    //
    // Code block that implements functionality formerly in defined as a plugin.
    // (note: plugin code is deprecated)
    //
    // The alsoResize functionality "also resizes" the dialog body.
    // This approach allows the footer area to remain at a fixed height
    // the dialog is resized.
    //
    // $.ui.plugin.add("resizable", "alsoResize", {
    //
    /////////////////////////////////////////////////////////////////////////////////

    _alsoresize_start: function() {

      //var that = $(this).resizable( "instance" ),
      // var that = $(this).data("oj-resizable"), // w
      var that = this;
      var o = that.options;
      // var initialR = that._initialResize;

      var _store = function(exp) {
        $(exp).each(function() {
          var el = $(this);

          el.data("oj-resizable-alsoresize", {
            width: parseInt(el.width(), 10), height: parseInt(el.height(), 10),
            left: parseInt(el.css("left"), 10), top: parseInt(el.css("top"), 10)
          });
        });
      };

      if (typeof (o.alsoResize) === "object" && !o.alsoResize.parentNode) {
        if (o.alsoResize.length) {
          o.alsoResize = o.alsoResize[0];
          _store(o.alsoResize);
        }
        else {
          $.each(o.alsoResize, function(exp) {
            _store(exp);
          });
        }
      } else {
        _store(o.alsoResize);
      }
    },
    _alsoresize_resize: function(event, ui) {

      // var that = $(this).resizable( "instance" ),
      // var that = $(this).data("oj-resizable"), // v
      var that = this;

      var o = that.options;
      var os = that.originalSize;
      var op = that.originalPosition;

      var delta = {
        height: (that.size.height - os.height) || 0, width: (that.size.width - os.width) || 0,
        top: (that.position.top - op.top) || 0, left: (that.position.left - op.left) || 0
      },
      _alsoResize = function(exp, c) {
        $(exp).each(function() {
          var el = $(this), start = $(this).data("oj-resizable-alsoresize"), style = {},
            css = c && c.length ? c : el.parents(ui.originalElement[0]).length ? ["width", "height"] : ["width", "height", "top", "left"];

          $.each(css, function(i, prop) {
            var sum = (start[prop] || 0) + (delta[prop] || 0);
            if (sum && sum >= 0) {
              style[prop] = sum || null;
            }
          });

          el.css(style);
        });
      };

      if (typeof (o.alsoResize) === "object" && !o.alsoResize.nodeType) {
        $.each(o.alsoResize, function(exp, c) {
          _alsoResize(exp, c);
        });
      } else {
        _alsoResize(o.alsoResize, null);
      }
    },
    _alsoresize_stop: function() {
      // $(this).removeData("resizable-alsoresize");
      $(this).removeData("oj-resizable-alsoresize");
    },
    /////////////////////////////////////////////////////////////////////////////////
    //
    // Code block for containment functionality (formerly defined as a plugin)
    //
    // $.ui.plugin.add( "resizable", "containment", {
    //
    /////////////////////////////////////////////////////////////////////////////////

    _containment_start: function() {

      var element, p, co, ch, cw, width, height;

      // var that = $(this).data("oj-resizable");
      var that = this;

      var o = that.options,
        el = that.element,
        oc = o.containment,
        ce = (oc instanceof $) ? oc.get(0) : (/parent/.test(oc)) ? el.parent().get(0) : oc;

      if (!ce) {
        return;
      }

      that.containerElement = $(ce);

      if (/document/.test(oc) || oc === document) {
        that.containerOffset = {
          left: 0,
          top: 0
        };
        that.containerPosition = {
          left: 0,
          top: 0
        };

        that.parentData = {
          element: $(document),
          left: 0,
          top: 0,
          width: $(document).width(),
          height: $(document).height() || document.body.parentNode.scrollHeight
        };
      } else {
        element = $(ce);
        p = [];
        $(["Top", "Right", "Left", "Bottom"]).each(function(i, name) {
          p[ i ] = that._num(element.css("padding" + name));
        });

        that.containerOffset = element.offset();
        that.containerPosition = element.position();
        that.containerSize = {
          height: (element.innerHeight() - p[ 3 ]),
          width: (element.innerWidth() - p[ 1 ])
        };

        co = that.containerOffset;
        ch = that.containerSize.height;
        cw = that.containerSize.width;
        width = (that._hasScroll(ce, "left") ? ce.scrollWidth : cw);
        height = (that._hasScroll(ce) ? ce.scrollHeight : ch);

        that.parentData = {
          element: ce,
          left: co.left,
          top: co.top,
          width: width,
          height: height
        };
      }
    },
    _containment_resize: function(event, ui) {
      var woset, hoset, isParent, isOffsetRelative;

      // var that = $(this).data("oj-resizable");
      var that = this;

      var o = that.options,
        co = that.containerOffset,
        cp = that.position,
        pRatio = event.shiftKey,
        cop = {
          top: 0,
          left: 0
        },
      ce = that.containerElement,
        continueResize = true;

      if (ce[ 0 ] !== document && (/static/).test(ce.css("position"))) {
        cop = co;
      }

      if (cp.left < (that._helper ? co.left : 0)) {
        that.size.width = that.size.width + (that._helper ? (that.position.left - co.left) : (that.position.left - cop.left));
        if (pRatio) {
          that.size.height = that.size.width / that.aspectRatio;
          continueResize = false;
        }
        that.position.left = o.helper ? co.left : 0;
      }

      if (cp.top < (that._helper ? co.top : 0)) {
        that.size.height = that.size.height + (that._helper ? (that.position.top - co.top) : that.position.top);
        if (pRatio) {
          that.size.width = that.size.height * that.aspectRatio;
          continueResize = false;
        }
        that.position.top = that._helper ? co.top : 0;
      }

      that.offset.left = that.parentData.left + that.position.left;
      that.offset.top = that.parentData.top + that.position.top;

      woset = Math.abs((that._helper ? that.offset.left - cop.left : (that.offset.left - co.left)) + that.sizeDiff.width);
      hoset = Math.abs((that._helper ? that.offset.top - cop.top : (that.offset.top - co.top)) + that.sizeDiff.height);

      isParent = that.containerElement.get(0) === that.element.parent().get(0);
      isOffsetRelative = /relative|absolute/.test(that.containerElement.css("position"));

      if (isParent && isOffsetRelative) {
        woset -= Math.abs(that.parentData.left);
      }

      if (woset + that.size.width >= that.parentData.width) {
        that.size.width = that.parentData.width - woset;
        if (pRatio) {
          that.size.height = that.size.width / that.aspectRatio;
          continueResize = false;
        }
      }

      if (hoset + that.size.height >= that.parentData.height) {
        that.size.height = that.parentData.height - hoset;
        if (pRatio) {
          that.size.width = that.size.height * that.aspectRatio;
          continueResize = false;
        }
      }

      if (!continueResize) {
        that.position.left = ui.prevPosition.left;
        that.position.top = ui.prevPosition.top;
        that.size.width = ui.prevSize.width;
        that.size.height = ui.prevSize.height;
      }
    },
    _containment_stop: function() {

      // var that = $(this).data("oj-resizable"),
      var that = this,
        o = that.options,
        co = that.containerOffset,
        cop = that.containerPosition,
        ce = that.containerElement,
        helper = $(that.helper),
        ho = helper.offset(),
        w = helper.outerWidth() - that.sizeDiff.width,
        h = helper.outerHeight() - that.sizeDiff.height;

      if (that._helper && !o.animate && (/relative/).test(ce.css("position"))) {
        $(this).css({
          left: ho.left - cop.left - co.left,
          width: w,
          height: h
        });
      }

      if (that._helper && !o.animate && (/static/).test(ce.css("position"))) {
        $(this).css({
          left: ho.left - cop.left - co.left,
          width: w,
          height: h
        });
      }
    },
    ui: function() {
      return {
        originalElement: this.originalElement,
        element: this.element,
        helper: this.helper,
        position: this.position,
        size: this.size,
        originalSize: this.originalSize,
        originalPosition: this.originalPosition,
        prevSize: this.prevSize,
        prevPosition: this.prevPosition
      };
    }

  });

}());

(function () {
  var ojDialogMeta = {
    "properties" : {
      "cancelBehavior" : {
        "type" : "string",
        "enumValues" : ["icon", "escape", "none"]
      },
      "dragAffordance" : {
        "type" : "string",
        "enumValues" : ["title-bar", "none"]
      },
      "initialVisibility" : {
        "type" : "string",
        "enumValues" : ["hide", "show"]
      },
      "modality" : {
        "type" : "string",
        "enumValues" : ["modal", "modeless"]
      },
      "position" : {
        "type" : "object",
        "properties" : {
          "my" : {
            "type" : "object|string",
            "properties" : {
              "horizontal" : {
                "type" : "string",
                "enumValues" : ["start", "end", "left", "center", "right"]
              },
              "vertical" : {
                "type" : "string",
                "enumValues" : ["top", "center", "bottom"]
              }
            }
          },
          "at" : {
            "type" : "object|string",
            "properties" : {
              "horizontal" : {
                "type" : "string",
                "enumValues" : ["start", "end", "left", "center", "right"]
              },
              "vertical" : {
                "type" : "string",
                "enumValues" : ["top", "center", "bottom"]
              }
            }
          },
          "offset" : {
            "type" : "object",
            "properties" : {
              "x" : {
                "type" : "number"
              },
              "y" : {
                "type" : "number"
              }
            }
          },
          "of" : {
            "type" : "string|{x:number, y:number}"
          },
          "collision" : {
            "type" : "string",
            "enumValues" : ["flip", "fit", "flipfit", "none"]
          }
        }
      },
      "resizeBehavior" : {
        "type" : "string",
        "enumValues" : ["resizable", "none"]
      },
      "translations": {
        "type": "Object",
        "properties": {
          "labelCloseIcon": {
            "type": "string",
            "value": "Close"
          }
        }
      }
    },
    "events" : {
      "animateEnd" : {},
      "animateStart" : {},
      "beforeClose" : {},
      "beforeOpen" : {},
      "close" : {},
      "open" : {},
      "focus" : {},
      "resize" : {},
      "resizeStart" : {},
      "resizeStop" : {},
      "drag" : {},
      "dragStart" : {},
      "dragStop" : {}
    },
    "methods" : {
      "close" : {},
      "isOpen" : {},
      "open" : {},
      "refresh" : {}
    },
    "extension" : {
      _WIDGET_NAME : "ojDialog"
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-dialog', 'baseComponent', ojDialogMeta);
  oj.CustomElementBridge.register('oj-dialog', {'metadata' : oj.CustomElementBridge.getMetadata('oj-dialog')});
})();

/*jslint browser: true*/
/*
** Copyright (c) 2004, 2012, Oracle and/or its affiliates. All rights reserved.
*/

/**
 * @ignore
 * @class oj.DomScroller
 * @classdesc Adds implicit highwatermark scrolling to DOM element
 * @param {Object} element Scrollable DOM element
 * @param {Object} dataprovider dataprovider
 * @param {Object=} options Options for the DomScroller<p>
 *                  <b>asyncIterator</>the iterator for the dataprovider
 *                  <b>success</b>: a user callback called when a fetch has completed successfully after scroll to bottom. Also called with maxCount information if maxCount is reached after scroll to bottom.<br>
 *                  <b>error</b>: a user callback function called if the fetch fails. The callback is called with the failed fetch content.<br>
 *                  <b>fetchSize</b>: the fetch size. Default is 25.<br>
 *                  <b>maxCount</b>: max row count. DomScroller will not exceed this max count. Default is 500.<br>
 *                  <b>initialRowCount</b>: initial row count. DomScroller will be initialized with this row count. Default is 0.<br>
 *                  <b>fetchTrigger</b>: how close should the scroll position be relative to the maximum scroll position before a fetch is triggered. Default is 1 pixel.<br>
 * @constructor
 */
oj.DomScroller = function(element, dataprovider, options)
{
  options = options || {};
  this._data = dataprovider;
  this._asyncIterator = options['asyncIterator'];
  this._element = $(element)[0];
  this._fetchSize = options['fetchSize'];
  this._fetchSize = this._fetchSize > 0 ? this._fetchSize : 25;
  this._maxCount = options['maxCount'];
  this._maxCount = this._maxCount > 0 ? this._maxCount : 500;
  this._rowCount = options['initialRowCount'] > 0 ? options['initialRowCount'] : 0;
  this._successCallback = options['success'];
  this._requestCallback = options['request'];
  this._errorCallback = options['error'];
  this._registerDataSourceEventListeners();
  this._fetchTrigger = options['fetchTrigger'];
  if (this._fetchTrigger == null || isNaN(this._fetchTrigger))
    this._fetchTrigger = 0;
  this._initialScrollTop = this._element.scrollTop;
  
  $(this._getScrollEventElement()).on('scroll.domscroller', function(event) {
    var target = this._element;
    var scrollTop = this._getScrollTop(target);
    var maxScrollTop = target.scrollHeight - target.clientHeight;
    if (maxScrollTop > 0)
    {
      this._handleScrollerScrollTop(scrollTop, maxScrollTop);
    }
  }.bind(this));
};

/**
 * Update value for fetch trigger
 */
oj.DomScroller.prototype.setFetchTrigger = function(fetchTrigger)
{
  if (fetchTrigger != null && !isNaN(fetchTrigger) && fetchTrigger >= 0)
    this._fetchTrigger = fetchTrigger;
};

/**
 * Retrieve the element where the scroll listener is registered on.
 * @private
 */
oj.DomScroller.prototype._getScrollEventElement = function()
{
  // if scroller is the body, listen for window scroll event.  This is the only way that works consistently across all browsers.
  if (this._element == document.body || this._element == document.documentElement)
  {
    return window;
  }
  else
  {
    return this._element;
  }
};

/**
 * Helper method to calculate the offsetTop from element to ancestor
 * @param {Element} ancestor the ancestor element
 * @param {Element} element the element 
 * @return {number} the distance between the specified element and ancestor
 */
oj.DomScroller.calculateOffsetTop = function(ancestor, element)
{
  var offset = 0, current = element;
  while (current && current != ancestor && $.contains(ancestor, current))
  {
    offset = offset + current.offsetTop;
    current = current.offsetParent;
  }

  return offset;
};

/**
 * Gets the scroll top of the element
 * @param {Element} element the element
 * @return {number} scroll top
 * @private
 */
oj.DomScroller.prototype._getScrollTop = function(element)
{
  var scrollTop = this._fetchTrigger;

  if (element == document.documentElement)
  {
    // to ensure it works across all browsers.  See https://bugs.webkit.org/show_bug.cgi?id=106133
    // for firefox we should use documentElement.scrollTop, for Chrome and IE use body.scrollTop
    // detect this by checking initial scrollTop is the same as current scrolltop, if it's the same then the scrollTop is not
    // returning the correct value and we should use body.scrollTop
    if (this._useBodyScrollTop === undefined)
      this._useBodyScrollTop = (this._initialScrollTop == element.scrollTop) ? true : false;

    if (this._useBodyScrollTop)
      return scrollTop + document.body.scrollTop;
  }

  return scrollTop + element.scrollTop;
};

/**
 * Destroys the dom scroller, unregister any event handlers.
 * @export
 * @expose
 * @memberof! oj.DomScroller
 * @instance
 */ 
oj.DomScroller.prototype.destroy = function()
{
  this._unregisterDataSourceEventListeners();
  $(this._getScrollEventElement()).off('scroll.domscroller');
};

/**
 * Check the viewport to see if a fetch needs to be done to fill it. Fetch if it does.
 * @return {Promise} Return a Promise which contains either the content of the fetch
 *                   or maxCount information if it has reached maxCount. Promise resolves to null if no fetch was done.
 * @throws {Error}
 * @export
 * @expose
 * @memberof! oj.DomScroller
 * @instance
 */
oj.DomScroller.prototype.checkViewport = function()
{
  if (this._asyncIterator && this._element.clientHeight > 0 && 
      !this._checkOverflow())
  {
    return this._fetchMoreRows();
  }
  return Promise.resolve(null);
};

/**
 * Handle scrollTop on scroller
 * @private	
 */
oj.DomScroller.prototype._handleScrollerScrollTop = function(scrollTop, maxScrollTop)
{
  if (maxScrollTop - scrollTop <= 1 && scrollTop > this._fetchTrigger && !this._fetchPromise)
  {
    if (this._asyncIterator)
    {
      var self = this;
      this._fetchMoreRows().then(function(result) 
      {
        if (self._successCallback)
        {
          self._successCallback(result);
        }
      }, function(reason) 
      { 
        if (self._errorCallback)      
        {
          self._errorCallback(reason);
        } 
      });
    }
    else
    {
      if (this._errorCallback != null)
      {
        this._errorCallback();
      }
    }
  }
};

/**
 * Check whether the scroll DOM has overflowed
 * @private	
 */
oj.DomScroller.prototype._checkOverflow = function()
{
  var element = this._element;
  
  if (element.scrollHeight > element.clientHeight + this._fetchTrigger)
  {
    return true;
  }
  return false;
};

/**
 * Try to fetch more rows
 * @private	
 */
oj.DomScroller.prototype._fetchMoreRows = function()
{
  if (!this._fetchPromise)
  {
    // make sure we don't exceed maxCount
    var remainingCount = this._maxCount - this._rowCount;
    if (remainingCount > 0)
    {
      var pageSize = this._fetchSize;
      var self = this;
      if (remainingCount < this._fetchSize)
      {
        pageSize = remainingCount;
      }
      if (this._asyncIterator)
      {
        if (this._requestCallback != null )
        {
          this._requestCallback();
        }
        this._fetchPromise = this._asyncIterator.next().then(function(result)
        {
          self._fetchPromise = null;

          if (result != null &&
            result.value != null)
          {
            if (result.value['data'].length > 0)
            {
              self._rowCount = self._rowCount + result.value['data'].length;

              if (remainingCount < self._fetchSize)
              {
                result['maxCount'] = self._maxCount;
                result['maxCountLimit'] = true
                
                if (result.value['data'].length > remainingCount) 
                {
                  result.value['data'] = result.value['data'].slice(0, remainingCount);
                  result.value['metadata'] = result.value['metadata'].slice(0, remainingCount);
                  if (result.value['fetchParameters'] != null) 
                  {
                    result.value['fetchParameters']['size'] = remainingCount;
                  }
                }
              }
            }

            // we have exhausted the iterator, discard so we won't attempt to fetch from it again
            if (result.done)
            {
              self._asyncIterator = null;
            }
          }
          return Promise.resolve(result);
        });
        return this._fetchPromise;
      }
    }
     // we need to indicate that we've hit maxCount
    return Promise.resolve({'maxCount': this._maxCount, 'maxCountLimit': true});
  }
  else
  {
    return this._fetchPromise;
  }
};

oj.DomScroller.prototype._handleDataRowMutateEvent = function(event)
{
  var mutationType = null;
  var eventDetail = null;
  
  if (event['detail']['remove'] != null)
  {
    mutationType = 'remove';
    eventDetail = event['detail']['remove'];
  }
  if (event['detail']['add'] != null)
  {
    mutationType = 'add';
    eventDetail = event['detail']['add'];
  }
  
  if (mutationType != null)
  {
    var eventIndexes = eventDetail['indexes'];
    var i, eventIndexesCount = eventIndexes.length;
    for (i = 0; i < eventIndexesCount; i++)
    {
      var rowIdx = eventIndexes[i];

      // we only care if the row is in our range
      if (rowIdx !== undefined &&
          this._rowCount > 0 &&
          rowIdx <= this._rowCount)
      {
        if (mutationType == 'add')
        {
          this._rowCount = this._rowCount + 1;
        }
        else if (mutationType == 'remove')
        {
          this._rowCount = this._rowCount - 1;
        }
      }
    }
  };
};

/**
 * Register event listeners which need to be registered datasource. 
 * @private
 */
oj.DomScroller.prototype._registerDataSourceEventListeners = function()
{
  // register the listeners on the datasource
  var data = this._data;
  if (data != null)
  {
    this._unregisterDataSourceEventListeners();

    this._dataProviderEventHandlers = [];
    this._dataProviderEventHandlers.push({'eventType': 'mutate', 'eventHandler': this._handleDataRowMutateEvent.bind(this)});

    var i;
    var ev;
    for (i = 0; i < this._dataProviderEventHandlers.length; i++) {
      ev = data.addEventListener(this._dataProviderEventHandlers[i]['eventType'], this._dataProviderEventHandlers[i]['eventHandler']);
      if (ev) {
          this._dataProviderEventHandlers[i]['eventHandler'] = ev;
      }
    }
  }
};

/**
 * Unregister event listeners which are registered on datasource. 
 * @private
 */
oj.DomScroller.prototype._unregisterDataSourceEventListeners = function()
{
  var data = this._data;
  // unregister the listeners on the datasource
  if (this._dataProviderEventHandlers != null && data != null)
  {
    var i;
    for (i = 0; i < this._dataProviderEventHandlers.length; i++)
      data.removeEventListener(this._dataProviderEventHandlers[i]['eventType'], this._dataProviderEventHandlers[i]['eventHandler']);
  }
};


// Copyright (c) 2017, Oracle and/or its affiliates. All rights reserved.
/**
 * File Upload Transport Interface
 *
 * This interface defines the API to communicate/send data to the server asynchronously. It is
 * designed to reflect the needs of the Jet File Upload UI as opposed to being some general
 * File Upload API. Application must implement the FileUploadTransport interface to use the JET
 * File Upload Component. The implementation of this interface should handle the details of features
 * supported by the server such as multipart upload, chunking, concurrent upload, queuing files,
 * abort pause/resume etc.
 *
 * <p>
 * See {@link oj.ProgressItem}
 * </p>
 * @ojstatus preview
 * @since 4.0.0
 * @export
 * @interface FileUploadTransport
 * @memberof oj
 * @ojtsimport ojprogresslist
 */

/**
 * This method queues up files for upload. It returns an array of <code class="prettyprint">ProgressItem</code> objects.
 * Each file corresponds to a <code class="prettyprint">ProgressItem</code> object.
 * These items can be uploaded by calling the flush method.
 *
 * @method
 * @name queue
 * @memberof oj.FileUploadTransport
 * @instance
 * @param {FileList} fileList A list contains the File objects to be queued.
 * @returns {Array.<oj.ProgressItem>} an array of <code class="prettyprint">ProgressItem</code> objects.
 *
 * @example <caption>Upload files</caption>
 * var uploadTransport = new oj.FileUploadTransport();
 * uploadTransport.queue(fileList);
 *
 */

/**
 * This method uploads all files that are currently in the queue to the specified server.
 *
 * @method
 * @name flush
 * @memberof oj.FileUploadTransport
 * @instance
 * @returns {void}
 *
 * @example <caption>Upload all queued files</caption>
 *
 * var uploadTransport = new oj.FileUploadTransport();
 * uploadTransport.queue(fileList);
 * uploadTransport.flush();
 *
 */


/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/* global ko */

/**
 * @ojcomponent oj.ojFilePicker
 * @ojtsimport ojcomposite
 * @since 4.0.0
 * @ojdisplayname File Picker
 * @ojshortdesc Displays a clickable dropzone for selecting files from the device storage.
 * @ojstatus preview
 * @class oj.ojFilePicker
 * @ojsignature {target: "Type", value:"class ojFilePicker extends JetElement<ojFilePickerSettableProperties>"}
 *
 * @classdesc
 * <h3 id="filePickerOverview-section">
 *   JET FilePicker
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#filePickerOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>By default the file picker shows a clickable dropzone for selecting files for upload. However, it can be replaced with any clickable element like a button. After the files are selected, the FilePicker fires a "select" event with the selected files. Application has to specify the listener in order to do the actual upload</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-file-picker on-oj-select='[[listener]]'>
 * &lt;/oj-file-picker>
 * </code>
 * </pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 * <p> The following CSS classes can be applied by the page author as needed.</p>
 *
 * {@ojinclude "name":"stylingDoc"}
 */

/**
 * An array of strings of allowed MIME types or file extensions that can be uploaded. If not specified, accept all file types
<p>Note: If accept is specified, files with empty string type will be rejected if no match found in the "accept" value.
 * @member
 * @name accept
 * @memberof oj.ojFilePicker
 * @instance
 * @type {Array.<string>|null}
 * @default null
 * @ojshortdesc An array of strings of allowed MIME types or file extensions that can be uploaded. If not specified, accept all file types
 *
 * @example <caption>Initialize the file picker with the <code class="prettyprint">accept</code> attribute specified:</caption>
 * &lt;oj-file-picker accept='image/*'>&lt;/oj-file-picker>
 *
 * @example <caption>Get or set the <code class="prettyprint">accept</code> property after initialization:</caption>
 * // getter
 * var acceptValue = myFilePicker.accept;
 *
 * // setter
 * myFilePicker.accept = 'image/*';
 */

/**
 * Whether to allow single or multiple file selection.
 * @member
 * @name selectionMode
 * @memberof oj.ojFilePicker
 * @instance
 * @type {string}
 * @ojvalue {string} "multiple" multiple file selection
 * @ojvalue {string} "single" single file selection
 * @default "multiple"
 * @ojshortdesc Whether to allow single or multiple file selection.
 *
 * @example <caption>Initialize the file picker with the <code class="prettyprint">selection-mode</code> attribute specified:</caption>
 * &lt;oj-file-picker selection-mode='single'>&lt;/oj-file-picker>
 *
 * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> property after initialization:</caption>
 * // getter
 * var selectionModeValue = myFilePicker.selectionMode;
 *
 * // setter
 * myFilePicker.selectionMode = 'single';
 */

/**
 * The type of event to select the files.
 * @member
 * @name selectOn
 * @memberof oj.ojFilePicker
 * @instance
 * @type {string}
 * @ojvalue {string} "auto" either click or drag and drop to select the files
 * @ojvalue {string} "click" click to select the files
 * @ojvalue {string} "drop" drag and drop the files
 * @ojvalue {string} "clickAndDrop" either click or drag and drop to select the files
 * @default "auto"
 * @ojshortdesc The type of event to select the files.
 *
 * @example <caption>Initialize the file picker with the <code class="prettyprint">select-on</code> attribute specified:</caption>
 * &lt;oj-file-picker select-on='click'>&lt;/oj-file-picker>
 *
 * @example <caption>Get or set the <code class="prettyprint">selectOn</code> property after initialization:</caption>
 * // getter
 * var selectOnValue = myFilePicker.selectOn;
 *
 * // setter
 * myFilePicker.selectOn = 'click';
 */


/**
 * Triggered after the files are selected
 * @member
 * @name select
 * @memberof oj.ojFilePicker
 * @instance
 * @event
 * @ojshortdesc Triggered after the files are selected
 * @property {FileList} files The files that were just selected.
 */

/**
 * <p>The <code class="prettyprint">trigger</code> slot is set on the custom content of the file picker.</p>
 *
 * @ojslot trigger
 * @memberof oj.ojFilePicker
 *
 * @example <caption>Display an upload button instead of the default dropzone</caption>
 * &lt;oj-file-picker class='oj-filepicker-custom'>
 *   &lt;oj-button slot='trigger'>
 *      &lt;span slot='startIcon' class='oj-fwk-icon oj-fwk-icon-arrowbox-n'>&lt;/span>
 *      Upload
 *   &lt;/oj-button>
 * &lt;/oj-file-picker>
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Clickable element</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Launch the browser's file picker.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojFilePicker
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Clickable element</td>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Launch the browser's file picker.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojFilePicker
 */

/**
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Class</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-filepicker-custom</td>
 *       <td>Apply to a custom file picker if the entire dropzone is replaced with another clickable element like button or menu item. Note that the oj-filepicker-custom class doesn't have to be specified in order to change the dropzone text.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-filepicker-dropzone</td>
 *       <td>Apply to the dropzone of the file picker.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-filepicker-text</td>
 *       <td>Apply to the dropzone text of the file picker.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
 * @memberof oj.ojFilePicker
 */

var pickerMetadata =
  {
    properties: {
      accept: {
        type: 'Array'
      },
      selectOn: {
        type: 'string',
        enumValues: ['auto', 'click', 'drop', 'clickAndDrop'],
        value: 'auto'
      },
      selectionMode: {
        type: 'string',
        enumValues: ['multiple', 'single'],
        value: 'multiple'
      }
    },
    events: {
      select: {}
    }
  };

var pickerView =
  "<input type='file' style='display:none'" +
  "       data-bind=\"attr:{multiple: $properties.selectionMode == 'multiple'," +
  '                         accept: acceptStr}">' +
  '</input>' +
  "<div class='oj-filepicker-clickable'>" +
  "  <oj-bind-slot name='trigger'>" +
  "    <div tabindex='0'" +
  "         class='oj-filepicker-dropzone'>" +
  "      <p class='oj-filepicker-text' data-bind='text: defDropzoneText'></p>" +
  '    </div>' +
  '  </oj-bind-slot>' +
  '</div>';

function pickerViewModel(context) {
  var self = this;
  var element = context.element;
  var props = context.properties;
  var $dropzone;
  var inputElem;
  var selecting = false;

  self.acceptStr = ko.pureComputed(function () {
    var acceptProp = props.accept;
    if (acceptProp && acceptProp.length) {
      return acceptProp.join(',');
    }
    return null;
  }, self);

  // add property changed listeners
  element.addEventListener('selectOnChanged', function (event) {
    setSelectOn(event.detail.value);
  });

  function setSelectOn(selectOn) {
    var dropzone = $dropzone[0];
    var clickable = $(element).find('.oj-filepicker-clickable')[0];

    removeDragAndDropListeners(dropzone);
    removeClickListeners(clickable);

    switch (selectOn) {
      case 'click':
        addClickListeners(clickable);
        break;
      case 'drop':
        addDragAndDropListeners(dropzone);
        // for accessbility
        if (clickable) {
          clickable.addEventListener('keypress', selectingFiles, false);
        }
        break;
      case 'auto':
      case 'clickAndDrop':
      default:
        addClickListeners(clickable);
        addDragAndDropListeners(dropzone);
        break;
    }
  }

  self.defDropzoneText = oj.Translations.getTranslatedString('oj-ojFilePicker.dropzoneText');

  self.bindingsApplied = function () {
    var $elem = $(element);
    $elem.addClass('oj-filepicker');
    $dropzone = $elem.find('.oj-filepicker-dropzone');
    setSelectOn(props.selectOn);

    // add a change listener on the <input> element
    var $inputElem = $elem.find('input');
    if ($inputElem.length) {
      inputElem = $inputElem[0];
      inputElem.addEventListener('change', handleFileSelected, false);
    }

    // suppress focus ring unless accessed by keyboard
    // makeFocusable doesn't work because file picker is launched by the browser
    // that changes focus
    /*
      oj.DomUtils.makeFocusable({
      'element': $elem.find(".oj-filepicker-dropzone"),
      'applyHighlight': true
      });
    */

    if ($dropzone.length) {
      $dropzone[0].addEventListener('focus', function () {
        if (selecting) {
          return;
        }

        if (oj.DomUtils.recentPointer()) {
          $dropzone.removeClass('oj-focus-highlight');
        } else {
          $dropzone.addClass('oj-focus-highlight');
        }
      });

      $dropzone[0].addEventListener('focusout', function () {
        if (selecting) {
          return;
        }

        $dropzone.removeClass('oj-focus-highlight');
      });
    }
  };

  // delegate the click to the internal input element to select files
  function selectingFiles(event) {
    selecting = true;
    //  - form submit event is triggered when the file upload button is clicked
    event.preventDefault();
    inputElem.click();
    return true;
  }

  function handleFileSelected(event) {
    event.preventDefault();
    event.stopPropagation();

    // if user cancelled out the file picker dialog, don't add files to upload queue
    var files = event.target.files;
    if (files.length > 0) {
      handleFilesAdded(files, event);
    }

    selecting = false;
  }

  function acceptFile(file) {
    var acceptProp = props.accept;
    if (!acceptProp || (acceptProp.length === 0) || !file) {
      return true;
    }

    var accept;
    for (var i = 0; i < acceptProp.length; i++) {
      accept = oj.StringUtils.trim(acceptProp[i]);
      if (!accept) {
        return true;
      } else if (accept.startsWith('.', 0)) {
        // when dragover, only MIME type is available, file name is undefined
        if (!file.name || (file.name && file.name.endsWith(accept))) {
          return true;
        }
      } else if (!file.type) {
        return false;
      } else if (accept === 'image/*') {
        if (file.type.startsWith('image/', 0)) {
          return true;
        }
      } else if (accept === 'video/*') {
        if (file.type.startsWith('video/', 0)) {
          return true;
        }
      } else if (accept === 'audio/*') {
        if (file.type.startsWith('audio/', 0)) {
          return true;
        }
      } else if (file.type === accept) {
        return true;
      }
    }
    return false;
  }

  function getAccepted(files) {
    if (files) {
      for (var i = 0; i < files.length; i++) {
        if (!acceptFile(files[i])) {
          return false;
        }
      }
    }
    return true;
  }

  function addClickListeners(clickable) {
    if (clickable) {
      clickable.addEventListener('click', selectingFiles, false);
      clickable.addEventListener('keypress', selectingFiles, false);
    }
  }

  function addDragAndDropListeners(dropzone) {
    if (dropzone) {
      dropzone.addEventListener('dragenter', handleDragEnter, false);
      dropzone.addEventListener('dragover', handleDragOver, false);
      dropzone.addEventListener('dragleave', handleDragLeave, false);
      dropzone.addEventListener('dragend', handleDragLeave, false);
      dropzone.addEventListener('drop', handleFileDrop, false);
    }
  }

  function removeClickListeners(clickable) {
    if (clickable) {
      clickable.removeEventListener('click', selectingFiles, false);
      clickable.removeEventListener('keypress', selectingFiles, false);
    }
  }

  function removeDragAndDropListeners(dropzone) {
    if (dropzone) {
      dropzone.removeEventListener('dragenter', handleDragEnter, false);
      dropzone.removeEventListener('dragover', handleDragOver, false);
      dropzone.removeEventListener('dragleave', handleDragLeave, false);
      dropzone.removeEventListener('dragend', handleDragLeave, false);
      dropzone.removeEventListener('drop', handleFileDrop, false);
    }
  }

  // check the selection and accept option to filter invalid file number or types
  function checkDroppable(files) {
    var validCnt = element.selectionMode !== 'single' || files.length === 1;
    return validCnt ? getAccepted(files) : validCnt;
  }

  // don't add "oj-valid-drop" class here because
  // hover effect is lost when dragging over the upload text
  function handleDragEnter(event) {
    event.preventDefault();
    event.stopPropagation();
  }


  function handleDragOver(event) {
    event.preventDefault();
    event.stopPropagation();

    // NOTE: dragged files not available
    // event.dataTransfer.files = null (firefox)
    // event.dataTransfer.files.length = 0 (chrome, IE, Edge and Safari)
    // use dataTransfer.items.type instead (works in chrome, firefox and Edge)
    // event.dataTransfer.items = undefined (not work in IE and safari, just don't display ghost buster)

    //  - drag and drop to ojfilepicker fails on safari
    var dataTransfer = event.dataTransfer;
    var droppable = (!dataTransfer.items) || checkDroppable(dataTransfer.items);

    if (droppable) {
      // Explicitly show this is a copy.
      dataTransfer.dropEffect = 'copy';
      $dropzone.addClass('oj-valid-drop');
    } else {
      // This is an invalid drop, show don't drop here cursor,
      // NOTE: dropEffect doesn't work in Edge
      dataTransfer.dropEffect = 'none';
    }
    return droppable;
  }

  function handleDragLeave(event) {
    event.preventDefault();
    event.stopPropagation();

    $dropzone.removeClass('oj-valid-drop');
  }

  function getFileNames(files) {
    var buf = '';
    if (files) {
      for (var i = 0; i < files.length; i++) {
        buf += files[i].name + ' ';
      }
    }
    return buf;
  }

  function handleFileDrop(event) {
    var files = event.dataTransfer.files;
    var droppable = checkDroppable(files);
    if (droppable) {
      // add files to upload queue
      handleFilesAdded(files, event);
    } else {
      oj.Logger.warn('oj-file-picker: Files ' + getFileNames(files) +
                     ' are not acceptable.');
    }

    handleDragLeave(event);
  }

  function handleFilesAdded(files, oEvent) {
    //  - oj_file_picker does not fire ojselect event
    var event = new CustomEvent('ojSelect',
      { detail: { files: files,
        originalEvent: oEvent } });
    element.dispatchEvent(event);
  }
}

oj.Composite.register('oj-file-picker',
  {
    view: pickerView,
    viewModel: pickerViewModel,
    metadata: pickerMetadata
  });

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 *
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 *
 * @expose
 * @memberof oj.ojFilePicker
 * @instance
 *
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 *
 * @expose
 * @memberof oj.ojFilePicker
 * @instance
 *
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 *
 * @expose
 * @memberof oj.ojFilePicker
 * @instance
 *
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*jslint browser: true*/
/**
 * Implementation of PagingModel used by FilmStrip.
 * @class oj.FilmStripPagingModel
 * @classdesc Implementation of PagingModel used by FilmStrip. 
 * @extends oj.EventSource
 * @implements oj.PagingModel
 * @constructor
 * @ignore
 * @ojtsignore
 */
oj.FilmStripPagingModel = function()
{
  this.Init();
};

// Subclass from oj.EventSource 
oj.Object.createSubclass(oj.FilmStripPagingModel, oj.EventSource, "oj.FilmStripPagingModel");

/**
 * Initialize the instance.
 * @return {void}
 * @memberof oj.FilmStripPagingModel
 * @instance
 * @export
 */
oj.FilmStripPagingModel.prototype.Init = function()
{
  oj.FilmStripPagingModel.superclass.Init.call(this);
  
  this._page = -1;
  this._totalSize = 0;
  this._pageSize = -1;
};

/**
 * Set the total size.
 * @param {number} totalSize The total size.
 * @return {void}
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.setTotalSize = function(totalSize)
{
  this._totalSize = totalSize;
};

/**
 * Get the page size.
 * @returns {number} The page size.
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.getPageSize = function()
{
  return this._pageSize;
};

// start oj.PagingModel interface methods //////////////////////////////////////

/**
 * Get the current page
 * @returns {number} The current page
 * @export
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.getPage = function()
{
  return this._page;
};

/**
 * Set the current page
 * @param {number} page The current page
 * @param {Object=} options Options<p>
 *                  pageSize: The page size.<p>
 * @returns {Promise} promise object triggering done when complete
 * @export
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.setPage = function(page, options)
{
  //ensure that the page index is an actual number, because the paging control
  //seems to set a string when you click on one of the page dots
  page = parseInt(page, 10);
  
  try
  {
    var prevPageCount = this.getPageCount();
    var prevPage = this._page;
    var prevPageSize = this._pageSize;
    var pageSize = prevPageSize;
    if (options && options.pageSize)
      pageSize = options.pageSize;
    
    //FIX : if the paging model hasn't been initialized yet, 
    //just return a resolved promise instead of throwing the invalid page
    //error below
    if (this._totalSize === 0 && pageSize === -1)
      return Promise.resolve();
    
    var newPageCount = Math.ceil(this._totalSize / pageSize);
    if (page < 0 || page > newPageCount - 1)
      throw new Error("JET FilmStrip: Invalid 'page' set: " + page);
    
    var bFiredBeforePageEvent = false;
    if (page != prevPage || pageSize != prevPageSize)
    {
      var bSuccess = this.handleEvent("beforePage", {"page": page, "previousPage": prevPage});
      if (bSuccess === false)
        return Promise.reject();
      bFiredBeforePageEvent = true;
    }
    
    this._page = page;
    this._pageSize = pageSize;
    var pageCount = this.getPageCount();
    var self = this;

    return new Promise(function(resolve, reject)
      {
        if (prevPageCount != pageCount)
          self.handleEvent("pageCount", {"pageCount": pageCount, "previousPageCount": prevPageCount});
        if (bFiredBeforePageEvent)
          self.handleEvent("page", {"page": page, "previousPage": prevPage});
        resolve(null);
      });
  }
  catch (e)
  {
    return Promise.reject(e);
  }
};

/**
 * Get the current page start index
 * @returns {number} The current page start index
 * @export
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.getStartItemIndex = function()
{
  //FIX : if there is no data yet, explicitly return -1 for 
  //start index, because otherwise the calculated value would incorrectly
  //be 1
  if (this._page === -1 && this._pageSize === -1)
    return -1;
  
  return this._page * this._pageSize;
};

/**
 * Get the current page end index
 * @returns {number} The current page end index
 * @export
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.getEndItemIndex = function()
{
  return Math.min(this.getStartItemIndex() + this._pageSize, this._totalSize) - 1;
};

/**
 * Get the page count
 * @returns {number} The total number of pages
 * @export
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.getPageCount = function()
{
  return Math.ceil(this._totalSize / this._pageSize);
};

/**
 * Return the total number of items. Returns -1 if unknown.
 * @returns {number} total number of items
 * @export
 * @memberof oj.FilmStripPagingModel
 * @instance
 */
oj.FilmStripPagingModel.prototype.totalSize = function()
{
  return this._totalSize;
};

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @export
 * @expose
 * @memberof oj.FilmStripPagingModel
 * @instance 
 */
oj.FilmStripPagingModel.prototype.totalSizeConfidence = function()
{ 
  return "actual";
};

// end oj.PagingModel interface methods ////////////////////////////////////////

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojFilmStrip
 * @augments oj.baseComponent
 * @since 1.1.0
 * @ojstatus preview
 * @ojshortdesc Container element that lays out its children in a single row or column across logical pages and allows navigating through them.
 * @ojrole region
 * @class oj.ojFilmStrip
 * @ojtsimport ojpagingtabledatasource
 * 
 * @classdesc
 * <h3 id="filmStripOverview-section">
 *   JET FilmStrip
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#filmStripOverview-section"></a>
 * </h3>
 * 
 * <p>Description: Container element that lays out its children in a single 
 * row or column across logical pages and allows navigating through them.
 * 
 * <p>A JET FilmStrip can contain multiple direct child elements and  
 * all child elements are expected to be the same size.
 * <p>FilmStrip will layout the child items across multiple logical pages and 
 * allow for changing between logical pages.  When the element is resized, 
 * relayout will occur and the number of pages and number of items shown per 
 * page may change.
 * 
 * <pre class="prettyprint"><code>
 * &lt;oj-film-strip>
 *   &lt;div class='my-filmstrip-item'>Alpha&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Beta&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Gamma&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Delta&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Epsilon&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Zeta&lt;/div>
 * &lt;/oj-film-strip>
 * </code></pre>
 * 
 * <p>JET FilmStrip and ConveyorBelt look similar, but are intended to be used
 * for different purposes.
 * <p>Use FilmStrip when you want to:
 * <ul>
 * <li>layout a set of items across discrete logical pages</li>
 * <li>control which and how many items are shown</li>
 * <li>hide items outside the current viewport from tab order and screen reader</li>
 * </ul>
 * <p>Use ConveyorBelt when you want to:
 * <ul>
 * <li>handle overflow without showing a scrollbar</li>
 * <li>keep all items accessible via tabbing and readable by a screen reader</li>
 * </ul>
 *
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 * 
 * 
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 * 
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * 
 * <h3 id="keyboard-appdev-section">
 *   Keyboard Application Developer Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-appdev-section"></a>
 * </h3>
 * 
 * <p>FilmStrip is for layout only.  Providing keyboard support for the child 
 * items in the filmstrip is the responsibility of the application developer, if
 * the items do not already support keyboard interaction.  This could be done, 
 * for example, by specifying <code class="prettyprint">tabindex</code> on each 
 * item to enable tab navigation among them.  
 *
 * <p>Only the child items of the FilmStrip that are visible in the current 
 * logical page can be tabbed to.  Items outside the current logical page are
 * hidden from the tab order.  
 * 
 * 
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>FilmStrip is for layout only.  It is the responsibility of the 
 * application developer to make the items in the filmstrip accessible.  Sighted 
 * keyboard-only users need to be able to access the items in the layout just by
 * using the keyboard.  It is up to the child items of the FilmStrip to 
 * support keyboard navigation.  
 *
 * <p>Only the child items of the FilmStrip that are visible in the current 
 * logical page can be read by a screen reader.  Items outside the current 
 * logical page are hidden from a screen reader.  
 * 
 * <p>FilmStrip assigns itself the WAI-ARIA 
 * <code class="prettyprint">role</code> of 
 * <code class="prettyprint">"region"</code>.  The application <b>must</b> 
 * specify a WAI-ARIA label for the FilmStrip element so that screen reader
 * users will understand the purpose of the FilmStrip.  
 * 
 * <p>The application <b>must</b> provide accessible controls for navigating 
 * through the logical pages of the FilmStrip for use by screen reader users, 
 * because the arrow keys may be intercepted by the screen reader itself.  The
 * navigation arrows built into FilmStrip can be used, or if they are hidden,
 * then alternative means, such as a paging control, must be provided.  
 * 
 * 
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 * 
 * <h4>Lazy Rendering</h4>
 * <p>If there are many items in the FilmStrip or complex content in each item,
 * it is recommended to implement lazy rendering of item content.  The application
 * should keep track of which items have rendered their content.  Items that are
 * initially visible in the FilmStrip should render their content from the start.
 * The application should then listen to events from the FilmStrip's 
 * <code class="prettyprint">PagingModel</code> in order to render content for
 * items that are paged into view later.  
 * 
 * 
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 * 
 * <p>As with any JET element, in the unusual case that the directionality 
 * (LTR or RTL) changes post-init, the FilmStrip must be 
 * <code class="prettyprint">refresh()</code>ed.
 */
/**
 * <p>The &lt;oj-film-strip> element lays out its children in a single row or column across logical pages and allows navigating through them.<p>
 *
 * @ojchild Default
 * @memberof oj.ojFilmStrip
 *
 * @example <caption>Initialize the film strip with child content specified:</caption>
 * &lt;oj-film-strip>
 *   &lt;div class='my-filmstrip-item'>Alpha&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Beta&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Gamma&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Delta&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Epsilon&lt;/div>
 *   &lt;div class='my-filmstrip-item'>Zeta&lt;/div>
 * &lt;/oj-film-strip>
 */
(function()
{
oj.__registerWidget("oj.ojFilmStrip", $['oj']['baseComponent'],
{
  defaultElement: "<div>",
  widgetEventPrefix: "oj",

  options:
  {
    /** 
     * Specify the maximum number of child items to show in a logical page.
     * A value of 0 (the default) means that FilmStrip will show as many items
     * per logical page as will fit based on the element and item sizes.
     * 
     * <p>FilmStrip may show fewer than the specified number of items when 
     * <code class="prettyprint">max-items-per-page</code> is set to a value other 
     * than 0 if the element size is smaller than what would be required to 
     * show that many items.  
     *
     * @expose 
     * @memberof oj.ojFilmStrip
     * @instance
     * @type {number}
     * @default 0
     * @see #getItemsPerPage
     * @ojshortdesc Specifies the maximum number of child items shown in a logical page.
     * @ojmin 0
     * @ojunits items
     *
     * @example <caption>Initialize the FilmStrip with the 
     * <code class="prettyprint">max-items-per-page</code> attribute specified:</caption>
     * &lt;oj-film-strip max-items-per-page='3'>
     * &lt;/oj-film-strip>
     * 
     * @example <caption>Get or set the <code class="prettyprint">maxItemsPerPage</code> 
     * property after initialization:</caption>
     * // getter
     * var maxItemsPerPage = myFilmStrip.maxItemsPerPage;
     * 
     * // setter
     * myFilmStrip.maxItemsPerPage = 3;
     */
    maxItemsPerPage: 0,
    
    /** 
     * Specify the orientation of the FilmStrip.
     *
     * @expose 
     * @memberof oj.ojFilmStrip
     * @instance
     * @type {string}
     * @ojvalue {string} "horizontal" Orient the FilmStrip horizontally.
     * @ojvalue {string} "vertical" Orient the FilmStrip vertically.
     * @default "horizontal"
     * @ojshortdesc Specifies the orientation of the FilmStrip.
     *
     * @example <caption>Initialize the FilmStrip with the 
     * <code class="prettyprint">orientation</code> attribute specified:</caption>
     * &lt;oj-film-strip orientation='vertical'>
     * &lt;/oj-film-strip>
     * 
     * @example <caption>Get or set the <code class="prettyprint">orientation</code> 
     * property after initialization:</caption>
     * // getter
     * var orientation = myFilmStrip.orientation;
     * 
     * // setter
     * myFilmStrip.orientation = 'vertical';
     */
    orientation: "horizontal",
    
    /** 
     * Specify the child item whose logical page should be displayed.  The
     * position of the item on the logical page is not guaranteed.
     * 
     * <p>This attribute can be set to an object containing either string id of the item 
     * or numeric 0-based index of the item or both.  If the object contains both 
     * string id and numeric index, string id takes precedence.
     * 
     * <p>FilmStrip will automatically update the value of this attribute when the 
     * logical page is changed to be the first item on the new logical page.  
     * When the value is updated automatically, it will be an object containing 
     * numeric index and string id, if available.
     * 
     * <p>When the element is resized, FilmStrip will preserve the value of 
     * this attribute to show the new logical page on which the item is located.  
     *
     * @property {string} [id] string id of the item 
     * @property {number} [index] numeric 0-based index of the item
     *
     * @expose 
     * @memberof oj.ojFilmStrip
     * @instance
     * @type {Object}
     * @ojsignature {target: "Type", value: "{id?: string, index?: number}", jsdocOverride: true}
     * @default {"index" : 0}
     * @ojwriteback
     * @ojshortdesc Specifies the child item whose logical page should be displayed.
     *
     * @example <caption>Initialize the FilmStrip with the 
     * <code class="prettyprint">current-item</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-film-strip current-item.index='3' current-item.id='thirdItemId'>
     * &lt;/oj-film-strip>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-film-strip current-item='{"index" : 3, "id" : "thirdItemId"}'>
     * &lt;/oj-film-strip>
     * 
     * @example <caption>Get or set the <code class="prettyprint">currentItem</code> 
     * property after initialization:</caption>
     *
     * // Get one
     * var index = myFilmStrip.currentItem.index;
     *
     * // Set one, leaving the others intact. Use the setProperty API for 
     * // subproperties so that a property change event is fired.
     * myFilmStrip.setProperty('currentItem.index', 1);
     * 
     * // Get all
     * var currentItem = myFilmStrip.currentItem;
     * 
     * // Set all
     * myFilmStrip.currentItem = {'index' : 3, 'id' : 'thirdItemId'};
     */
    currentItem: {"index" : 0},
    
    /**
     * Specify the placement of the navigation arrows.
     *
     * @expose 
     * @memberof oj.ojFilmStrip
     * @instance
     * @type {string}
     * @ojvalue {string} "adjacent" Arrows are outside, adjacent to the filmStrip 
     *          content.  The arrows are still inside the bounds of the filmStrip
     *          element itself.
     * @ojvalue {string} "overlay" Arrows are inside, overlaying the filmStrip content.
     * @default "adjacent"
     * @ojshortdesc Specifies the placement of the navigation arrows.
     *
     * @example <caption>Initialize the FilmStrip with the 
     * <code class="prettyprint">arrow-placement</code> attribute specified:</caption>
     * &lt;oj-film-strip arrow-placement='overlay'>
     * &lt;/oj-film-strip>
     * 
     * @example <caption>Get or set the <code class="prettyprint">arrowPlacement</code> 
     * property after initialization:</caption>
     * // getter
     * var arrowPlacement = myFilmStrip.arrowPlacement;
     * 
     * // setter
     * myFilmStrip.arrowPlacement = 'overlay';
     */
    arrowPlacement: "adjacent",
    
    /** 
     * Specify the visibility of the navigation arrows.
     *
     * @expose 
     * @memberof oj.ojFilmStrip
     * @instance
     * @type {string}
     * @ojvalue {string} "visible" Arrows are visible.
     * @ojvalue {string} "hidden" Arrows are hidden.
     * @ojvalue {string} "hover" Arrows are visible when the mouse is over the 
     *          filmStrip, and hidden otherwise.
     * @ojvalue {string} "auto" Behaves as if the value were 
     *          <code class="prettyprint">visible</code> when the 
     *          <code class="prettyprint">arrow-placement</code> 
     *          attribute is set to <code class="prettyprint">adjacent</code>, and
     *          <code class="prettyprint">hover</code> when the 
     *          <code class="prettyprint">arrow-placement</code> 
     *          attribute is set to <code class="prettyprint">overlay</code>.
     * @default "auto"
     * @ojshortdesc Specifies the visibility of the navigation arrows.
     *
     * @example <caption>Initialize the FilmStrip with the 
     * <code class="prettyprint">arrow-visibility</code> attribute specified:</caption>
     * &lt;oj-film-strip arrow-visibility='visible'>
     * &lt;/oj-film-strip>
     * 
     * @example <caption>Get or set the <code class="prettyprint">arrowVisibility</code> 
     * property after initialization:</caption>
     * // getter
     * var arrowVisibility = myFilmStrip.arrowVisibility;
     * 
     * // setter
     * myFilmStrip.arrowVisibility = 'visible';
     */
    arrowVisibility: "auto"
    
    /**
     * FilmStrip inherits the <code class="prettyprint">disabled</code> 
     * attribute from its superclass, but does not support it in order to avoid 
     * tight coupling between a FilmStrip and its contents.  Setting this
     * attribute on FilmStrip has no effect.
     * 
     * <p><b>WARNING:</b> Applications should not depend on this behavior 
     * because we reserve the right to change it in the future in order to 
     * support <code class="prettyprint">disabled</code> and propagate it to 
     * child elements of FilmStrip.  
     * 
     * @member
     * @name disabled
     * @memberof oj.ojFilmStrip
     * @instance
     * @type {boolean}
     * @default false
     * @ignore
     */
    // disabled attribute declared in superclass, but we still want the above API doc

    // Events

    /**
     * Triggered when the filmStrip is created.
     *
     * @event
     * @name create
     * @memberof oj.ojFilmStrip
     * @instance
     * @ignore
     */
    // create event declared in superclass, but we still want the above API doc
  },

  /**
   * After _ComponentCreate and _AfterCreate, 
   * the widget should be 100% set up. this._super should be called first.
   * @return {void}
   * @override
   * @protected
   * @instance
   * @memberof oj.ojFilmStrip
   */
  _ComponentCreate: function() // Override of protected base class method.  
  {
    //call superclass first
    this._super();
    
    var elem = this.element;  
    elem
      .addClass("oj-filmstrip oj-component")
      .attr("tabindex", 0)
      .attr("role", "region");
    
    //ensure a unique id for use with aria-labelledby on navigation arrows
    elem.uniqueId();
    
    //log warning message when "disabled" property set
    var options = this.options;
    if (options.disabled)
      oj.Logger.warn(_WARNING_DISABLED_OPTION);
    
    //throw error if "orientation" property set to invalid value
    if (options.orientation !== _HORIZONTAL && options.orientation !== _VERTICAL)
      throw new Error(_ERROR_INVALID_ORIENTATION + options.orientation);
    
    //throw error if "arrowPlacement" property set to invalid value
    if (options.arrowPlacement !== _ADJACENT && 
        options.arrowPlacement !== _OVERLAY)
      throw new Error(_ERROR_INVALID_NAV_ARROW_PLACEMENT + options.arrowPlacement);
    
    //throw error if "arrowVisibility" property set to invalid value
    if (options.arrowVisibility !== _VISIBLE && 
        options.arrowVisibility !== _HIDDEN && 
        options.arrowVisibility !== _HOVER && 
        options.arrowVisibility !== _AUTO)
      throw new Error(_ERROR_INVALID_NAV_ARROW_VISIBILITY + options.arrowVisibility);
    
    this.touchEventNamespace = this.eventNamespace + "Touch";
    this.mouseEventNamespace = this.eventNamespace + "Mouse";
    this.keyEventNamespace = this.eventNamespace + "Key";
    this.navArrowHoverableEventNamespace = this.eventNamespace + "NavArrowHoverable";
    
    //Make sure currentItem is an object of (id, index)
    options.currentItem = this._convertItemToObj(options.currentItem);
    
    this._setup(true);
    
    //update the currentItem object in options.
    this._populateItemObj(options.currentItem);
    this.option(_CURRENT_ITEM, options.currentItem, {'_context': {internalSet: true, writeback: true}});
  },

  // This method currently runs at create, init, and refresh time (since refresh() is called by _init()).
  /**
   * Refreshes the visual state of the FilmStrip. JET elements require a 
   * <code class="prettyprint">refresh()</code> after the DOM is 
   * programmatically changed underneath the element.
   * 
   * <p>This method does not accept any arguments.
   * @return {void}
   * 
   * @expose 
   * @memberof oj.ojFilmStrip
   * @instance
   * @ojshortdesc Refreshes the visual state of the FilmStrip.
   * 
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   * myFilmStrip.refresh();
   */
  refresh: function() // Override of public base class method.  
  {
    this._super();
    
    this._setup(false);
  },
  
  /**
   * Get the actual number of items shown per logical page.  
   * <p>The value returned by this method may be different from the value of the
   * <code class="prettyprint">maxItemsPerPage</code> property.
   * @returns {number} The actual number of items shown per logical page.
   * 
   * @expose 
   * @memberof oj.ojFilmStrip
   * @instance
   * @see #maxItemsPerPage
   * @ojshortdesc Get the actual number of items shown per logical page.
   * 
   * @example <caption>Invoke the <code class="prettyprint">getItemsPerPage</code> method:</caption>
   * var itemsPerPage = myFilmStrip.getItemsPerPage();
   */
  getItemsPerPage: function()
  {
    return this._itemsPerPage;
  },
  
  /**
   * Get the PagingModel created and used by the FilmStrip.  The PagingModel
   * provides information about the FilmStrip's logical pages and a way to
   * programmatically change pages.  
   * @returns {Object} The instance of the PagingModel created and used by the FilmStrip.
   * @ojtsignore
   * @ojsignature {target: "Type", 
   *               value: "oj.PagingModel",
   *               for: "returns"}
   * 
   * @expose 
   * @memberof oj.ojFilmStrip
   * @instance
   * @ojshortdesc Get the PagingModel created and used by the FilmStrip.
   * 
   * @example <caption>Invoke the <code class="prettyprint">getPagingModel</code> method:</caption>
   * var pagingModel = myFilmStrip.getPagingModel();
   */
  getPagingModel: function()
  {
    return this._pagingModel;
  },
  
  /**
   * Notifies the component that its subtree has been made visible 
   * programmatically after the component has been created.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @protected 
   * @override
   */
  _NotifyShown: function()
  {
    this._super();
    //perform a deferred layout
    if (this._needsSetup)
      this._setup(this._needsSetup[0]);
    else
    {
      //explicitly handle a resize in case we don't get a notification when shown
      this._handleResize();
    }
  },
  
  /**
   * Notifies the component that its subtree has been connected to the document
   * programmatically after the component has been created.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @protected 
   * @override
   */
  _NotifyAttached: function()
  {
    this._super();
    //perform a deferred layout
    if (this._needsSetup)
      this._setup(this._needsSetup[0]);
    else
    {
      //explicitly handle a resize in case we don't get a notification when attached
      this._handleResize();
    }
  },

  // isInit is true for init (create and re-init), false for refresh
  /** 
   * Setup the filmStrip.
   * @param {boolean} isInit True if _setup is called from _init(), false
   *        if called from refresh().
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _setup: function(isInit) // Private, not an override (not in base class).  
  {
    var self = this;
    
    //FIX : always create paging model, even if filmstrip is
    //detached or hidden, so that it's available for an assocated paging
    //control
    //(Note: Check whether paging model is null, because if the filmstrip
    //was created while detached or hidden, then when it's attached or
    //shown, _setup will be called again with isInit=true, and we don't
    //want to create a different instance of the paging model.)
    if (isInit && !this._pagingModel)
      this._pagingModel = new oj.FilmStripPagingModel();
    
    //FIX : when using jquery find() in the filmStrip, filter out any results from
    //nested filmStrips
    if (isInit && !this._filterNestedFilmStripsFunc)
    {
      this._filterNestedFilmStripsFunc = function(index, elem)
      {
        return $(elem).closest(".oj-filmstrip")[0] === self.element[0];
      };
    }
    
    //if filmStrip is detached or hidden, we can't layout correctly, so 
    //defer layout until filmStrip is attached or shown
    if (!this._canCalculateSizes())
    {
      //want a true value of isInit to take precedence over a false value
      var oldIsInit = false;
      if (this._needsSetup)
      {
        oldIsInit = this._needsSetup[0];
      }
      this._needsSetup = [isInit || oldIsInit];
      return;
    }
    this._needsSetup = null;
    
    this._bRTL = (this._GetReadingDirection() === "rtl");
    this._bTouchSupported = oj.DomUtils.isTouchSupported();
    var elem = this.element;
    if (isInit)
    {
      this._itemsPerPage = 0;
      this._handlePageFunc = function(event) {self._handlePage(event);};
      this._componentSize = 0;
      this._itemSize = -1;
      this._handleTransitionEndFunc = function(event) {self._handleTransitionEnd();};
      this._handleResizeFunc = function(width, height) {self._handleResize();};
      
      if (this._bTouchSupported)
      {
        this._handleTouchStartFunc = function(event) {self._handleTouchStart(event);};
        this._handleTouchMoveFunc = function(event) {self._handleTouchMove(event);};
        this._handleTouchEndFunc = function(event) {self._handleTouchEnd(event);};
        this._addTouchListeners();
      }
      this._handleMouseDownFunc = function(event) {self._handleMouseDown(event);};
      this._handleMouseMoveFunc = function(event) {self._handleMouseMove(event);};
      this._handleMouseUpFunc = function(event) {self._handleMouseUp(event);};
      this._addMouseListeners();
      this._handleKeyDownFunc = function(event) {self._handleKeyDown(event);};
      this._addKeyListeners();
    }
    else
      this._destroyInternal();
    
    //notify the original children of the filmStrip that they were detached from
    //the DOM BEFORE actually detaching so that components can save state
    var originalItems = elem.children();
    for (var i = 0; i < originalItems.length; i++)
      oj.Components.subtreeDetached(originalItems[i]);
    
    var pagingModel = this._pagingModel;
    if (isInit)
    {
      //register the page change listener
      pagingModel.on("page", this._handlePageFunc);
    }
    //initialize the paging model with the total number of items
    //FIX : always set total size on paging model in case
    //items have been appended and filmstrip has been refreshed
    pagingModel.setTotalSize(originalItems.length);
    
    //wrap the original child DOM BEFORE adding the resize listener
    this._wrapChildren();
    
    //FIX : only need to calculate layout if the filmstrip is
    //not empty
    if (originalItems.length > 0)
    {
      this._adjustSizes();

      //notify the original children of the filmStrip that they were attached
      //again after wrapping them
      for (var i = 0; i < originalItems.length; i++)
        oj.Components.subtreeAttached(originalItems[i]);

      oj.DomUtils.addResizeListener(elem[0], this._handleResizeFunc, 
        _RESIZE_LISTENER_COLLAPSE_EVENT_TIMEOUT);
    }
  },

  /** 
   * Destroy the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @override
   * @protected
   */
  _destroy: function() // Override of protected base class method.  
  {
    if (this._bTouchSupported)
    {
      this._removeTouchListeners();
      this._handleTouchStartFunc = null;
      this._handleTouchMoveFunc = null;
      this._handleTouchEndFunc = null;
    }
    this._removeMouseListeners();
    this._handleMouseDownFunc = null;
    this._handleMouseMoveFunc = null;
    this._handleMouseUpFunc = null;
    this._removeKeyListeners();
    this._handleKeyDownFunc = null;
    
    this._destroyInternal();
    this._pagingModel.off("page", this._handlePageFunc);
    this._handlePageFunc = null;
    this._pagingModel = null;
    this._handleResizeFunc = null;
    this._handleTransitionEndFunc = null;
    this._filterNestedFilmStripsFunc = null;
    
    var elem = this.element;
    elem
      .removeClass("oj-filmstrip oj-component " + _OJ_FILMSTRIP_HOVER)
      .removeAttr("tabindex role");
    
    //remove a generated unique id
    elem.removeUniqueId();
    
    this._super();
  },
  
  /** 
   * Destroy the internal structure of the filmStrip that will be recreated
   * during a refresh.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _destroyInternal: function()
  {
    //FIX : clear the flag to handle a deferred resize
    this._deferredHandleResize = false;
    
    //FIX : resolve busy state when destroying
    this._resolveBusyState();
    
    var elem = this.element;
    oj.DomUtils.removeResizeListener(elem[0], this._handleResizeFunc);
    //reset item size so it will be recalculated at the next layout
    this._itemSize = -1;
    if (this._queuedHandleResize)
    {
      clearTimeout(this._queuedHandleResize);
      this._queuedHandleResize = null;
    }
    
    var originalItems = this._getItems();
    //notify the original children of the filmStrip that they were detached from
    //the DOM BEFORE actually detaching so that components can save state
    for (var i = 0; i < originalItems.length; i++)
      oj.Components.subtreeDetached(originalItems[i]);
    
    this._clearCalculatedSizes();
    var itemContainers = this._getItemContainers();
    //remove logical page containers here instead of in _unwrapChildren because 
    //they're added in _adjustSizes, not in _wrapChildren
    itemContainers.unwrap();
    this._unwrapChildren();
    
    //notify the original children of the filmStrip that they were attached 
    //again after unwrapping them
    for (var i = 0; i < originalItems.length; i++)
      oj.Components.subtreeAttached(originalItems[i]);
  },

  /** 
   * Set an option on the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @override
   * @protected
   */
  _setOption: function(key, value, flags) // Override of protected base class method.
                                   // Method name needn't be quoted since is in externs.js.
  {
    var options = this.options;
    var bRefresh = false;
    var newPageIndex = -1;
    var pagingModel = this._pagingModel;
    var oldPageIndex = pagingModel.getPage();
    switch (key) 
    {
      case "disabled":
        //log warning message when "disabled" property set
        oj.Logger.warn(_WARNING_DISABLED_OPTION);
        break;
      case "orientation":
        //throw error if "orientation" property set to invalid value
        if (value !== _HORIZONTAL && value !== _VERTICAL)
          throw new Error(_ERROR_INVALID_ORIENTATION + value);
        bRefresh = (options.orientation != value);
        break;
      case "maxItemsPerPage":
        bRefresh = (options.maxItemsPerPage != value);
        break;
      case "arrowPlacement":
        //throw error if "arrowPlacement" property set to invalid value
        if (value !== _ADJACENT && value !== _OVERLAY)
          throw new Error(_ERROR_INVALID_NAV_ARROW_PLACEMENT + value);
        bRefresh = (options.arrowPlacement != value);
        break;
      case "arrowVisibility":
        //throw error if "arrowVisibility" property set to invalid value
        if (value !== _VISIBLE && value !== _HIDDEN && 
            value !== _HOVER && value !== _AUTO)
          throw new Error(_ERROR_INVALID_NAV_ARROW_VISIBILITY + value);
        bRefresh = (options.arrowVisibility != value);
        break;
      case _CURRENT_ITEM:
        //Make sure currentItem value is an object of (id, index)
        value = this._convertItemToObj(value);
        this._populateItemObj(value);
        var currentItem = options.currentItem;
        if (currentItem && value && 
              (currentItem.id != value.id || currentItem.index != value.index))
        {
          newPageIndex = this._findPage(value);
          //throw error if item not found
          if (newPageIndex < 0 || newPageIndex >= pagingModel.getPageCount())
            throw new Error(_ERROR_CURRENT_ITEM_NOT_FOUND + value);
        }
        break;
    }
    
    this._super(key, value, flags);
    
    switch (key)
    {
      case _CURRENT_ITEM:
        if (newPageIndex > -1 && newPageIndex != oldPageIndex)
          pagingModel.setPage(newPageIndex);
        break;
    }
    
    if (bRefresh)
      this._setup(false);
  },
  
  /** 
   * Handle a component resize.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleResize: function()
  {
    //FIX : if we're handling a page change, defer handling the resize
    if (this._busyStateResolveFunc)
    {
      this._deferredHandleResize = true;
    }
    //handle the resize if not already handling one
    else if (!this._bHandlingResize)
    {
      this._bHandlingResize = true;
      this._adjustSizes(true);
      this._bHandlingResize = false;
    }
    //if already handling a resize, queue another one
    else if (!this._queuedHandleResize)
    {
      var self = this;
      this._queuedHandleResize = setTimeout(function()
        {
          self._queuedHandleResize = null;
          self._handleResize();
        }, 0);
    }
  },
  
  /** 
   * Determine whether the filmStrip is oriented horizontally or vertically.
   * @returns {boolean} True if horizontal, false if vertical.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _isHorizontal: function()
  {
    var options = this.options;
    return (options.orientation !== _VERTICAL);
  },
  
  /** 
   * Get the CSS position attribute to use.
   * @returns {string} CSS position attribute name.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _getCssPositionAttr: function()
  {
    var bHorizontal = this._isHorizontal();
    return bHorizontal ? (this._bRTL ? "right" : "left") : "top";
  },
  
  /** 
   * Get the CSS size attribute to use.
   * @returns {string} CSS size attribute name.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _getCssSizeAttr: function()
  {
    var bHorizontal = this._isHorizontal();
    return bHorizontal ? "width" : "height";
  },

  /** 
   * Determine whether the filmStrip can calculate sizes (when it is 
   * attached to the page DOM and not hidden).
   * @returns {boolean} True if sizes can be calculated, false if not.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _canCalculateSizes: function() 
  {
    var div = document.createElement("div");
    var style = div.style;
    //need to set absolute position in order to get correct offsetwidth when 
    //flexbox layout is applied; otherwise the offsetwidth becomes 0
    style.position = "absolute";
    style.width = "10px";
    style.height = "10px";
    var elem = this.element[0];
    elem.appendChild(div); //@HTMLUpdateOK; div is created locally at the beginning of this function
    var bCanCalcSizes = false;
    try
    {
      bCanCalcSizes = div.offsetWidth > 0 && div.offsetHeight > 0;
    }
    catch (e)
    {
      //do nothing
    }
    elem.removeChild(div);
    return bCanCalcSizes;
  },
  
  /** 
   * Wrap the original child items of the filmStrip in internal layout structure.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _wrapChildren: function()
  {
    var elem = this.element;
    var bHorizontal = this._isHorizontal();
    var originalItems = elem.children();
    
    originalItems
      .addClass(_OJ_FILMSTRIP_ITEM)
      .wrap("<div class='" + _OJ_FILMSTRIP_CONTAINER + " " + _OJ_FILMSTRIP_ITEM_CONTAINER + "'></div>"); // @HTMLUpdateOK
    
    //need to initially specify the position on the pagesWrapper so that we can
    //always get the value later
    var cssAttr = this._getCssPositionAttr();
    var pagesWrapper = elem.children()
      .wrapAll("<div class='" + _OJ_FILMSTRIP_CONTAINER + " " + _OJ_FILMSTRIP_PAGES_CONTAINER + "' style='" + cssAttr + ": 0px;'></div>") // @HTMLUpdateOK
      .parent();
    this._pagesWrapper = pagesWrapper;
    
    var options = this.options;
    if (options.arrowVisibility !== _HIDDEN &&
        options.arrowPlacement === _ADJACENT)
    {
      //FIX : add the oj-filmstrip-container class to the content
      //container so that it is a flexbox layout
      this._contentWrapper = pagesWrapper
        .wrap("<div class='" + _OJ_FILMSTRIP_CONTAINER + " oj-filmstrip-content-container'></div>") // @HTMLUpdateOK
        .parent();
    }
    
    elem.addClass(_OJ_FILMSTRIP_CONTAINER);
    if (!bHorizontal)
      elem.addClass(_OJ_FILMSTRIP_VERTICAL);
    
    //FIX : only need to create nav buttons if the filmstrip 
    //is not empty
    if (options.arrowVisibility !== _HIDDEN &&
        originalItems.length > 0)
    {
      this._prevButton = this._createPrevNavArrow();
      this._nextButton = this._createNextNavArrow();

      if (this._navArrowsShownOnHover())
        this._setupNavArrowsHoverable();
    }
  },
  
  /** 
   * Unwrap the internal layout structure around the original child items of the 
   * filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _unwrapChildren: function()
  {
    var elem = this.element;
    var originalItems = this._getItems();
    
    //unregister elem as hoverable
    this._tearDownNavArrowsHoverable();
    
    //remove navigation arrow containers
    if (this._prevButton)
    {
      this._UnregisterChildNode(this._prevButton);
      this._prevButton = null;
    }
    if (this._nextButton)
    {
      this._UnregisterChildNode(this._nextButton);
      this._nextButton = null;
    }
    var navArrowContainers = elem.children(_PERIOD + _OJ_FILMSTRIP_ARROW_CONTAINER);
    if (navArrowContainers)
      navArrowContainers.remove();
    
    //the individual page containers that were added in _adjustSizes are not
    //removed here; they are expected to have been removed before calling
    //_unwrapChildren
    originalItems
      .removeClass(_OJ_FILMSTRIP_ITEM) //original children
      .unwrap()                        //remove item containers
      .unwrap();                       //remove pages container
    this._pagesWrapper = null;
    if (this._contentWrapper)
    {
      originalItems.unwrap();          //remove content container
      this._contentWrapper = null;
    }
    
    elem.removeClass(_OJ_FILMSTRIP_CONTAINER + " " + _OJ_FILMSTRIP_VERTICAL);
  },
  
  /** 
   * Setup events to only show the navigation arrows when the mouse hovers over
   * the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _setupNavArrowsHoverable: function()
  {
    //Note: Use oj-filmstrip-hover marker instead of standard oj-hover marker
    //because if we use the standard one, then when it's added to the filmStrip
    //root node, all hoverables inside the filmStrip appear as if the mouse
    //were hovering over them.  
    //Note: keep this logic updated if superclass _hoverable method changes.
    var elem = this.element;
    elem
      .on("mouseenter" + this.navArrowHoverableEventNamespace, function(event) {
        if (!$(event.currentTarget).hasClass( "oj-disabled" ))
          $(event.currentTarget).addClass(_OJ_FILMSTRIP_HOVER);
      })
      .on("mouseleave" + this.navArrowHoverableEventNamespace, function(event) {
        $(event.currentTarget).removeClass(_OJ_FILMSTRIP_HOVER);
      });
  },
  
  /** 
   * Tear down events to only show the navigation arrows when the mouse hovers 
   * over the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _tearDownNavArrowsHoverable: function()
  {
    var elem = this.element;
    elem.off(this.navArrowHoverableEventNamespace);
  },
  
  /** 
   * Determine whether navigation arrows should only be shown on hover.
   * @returns {boolean} True if nav arrows should only be shown on hover, false
   *          otherwise.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _navArrowsShownOnHover: function()
  {
    var options = this.options;
    var arrowVisibility = options.arrowVisibility;
    return  (arrowVisibility === _HOVER || 
             (arrowVisibility === _AUTO && options.arrowPlacement === _OVERLAY));
  },
  
  /** 
   * Determine whether there is a previous logical page.
   * @returns {boolean} True if there is a previous logical page, false otherwise.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _hasPrevPage: function()
  {
    var pagingModel = this._pagingModel;
    return pagingModel.getPage() > 0;
  },
  
  /** 
   * Determine whether there is a next logical page.
   * @returns {boolean} True if there is a next logical page, false otherwise.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _hasNextPage: function()
  {
    var pagingModel = this._pagingModel;
    return pagingModel.getPage() < pagingModel.getPageCount() - 1;
  },
  
  /** 
   * Go to the previous logical page.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _prevPage: function()
  {
    if (this._hasPrevPage())
    {
      var pagingModel = this._pagingModel;
      pagingModel.setPage(pagingModel.getPage() - 1);
    }
  },
  
  /** 
   * Go to the next logical page.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _nextPage: function()
  {
    if (this._hasNextPage())
    {
      var pagingModel = this._pagingModel;
      pagingModel.setPage(pagingModel.getPage() + 1);
    }
  },
  
  /** 
   * Determine whether there is a next logical page.
   * @param {boolean} bShow True to show the navigation arrow, false to hide it.
   * @param {jQuery} jqNavArrow Navigation arrow jQuery object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _displayNavigationArrow: function(bShow, jqNavArrow)
  {
    var options = this.options;
    var navArrowPlacement = options.arrowPlacement;
    if (navArrowPlacement === _ADJACENT)
      jqNavArrow.css("visibility", bShow ? "" : _HIDDEN);
    else
      jqNavArrow.parent().css("display", bShow ? "" : _NONE);
  },
  
  /** 
   * Update whether the navigation arrows are shown.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _updateNavigationArrowsDisplay: function()
  {
    var options = this.options;
    var navArrowVisibility = options.arrowVisibility;
    if (navArrowVisibility !== _HIDDEN)
    {
      var pagingModel = this._pagingModel;
      var pageIndex = pagingModel.getPage();
      var pageCount = pagingModel.getPageCount();
      this._displayNavigationArrow(pageIndex !== 0, this._prevButton);
      this._displayNavigationArrow(pageIndex !== pageCount - 1, this._nextButton);
    }
  },
  
  /** 
   * Create the previous navigation arrow.
   * @returns {jQuery} Navigation arrow jQuery object.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _createPrevNavArrow: function()
  {
    var elem = this.element;
    var bHorizontal = this._isHorizontal();
    var locationMarker = bHorizontal ? _OJ_START : _OJ_TOP;
    var container = this._createNavigationArrowContainer(locationMarker);
    
    var options = this.options;
    var navArrowPlacement = options.arrowPlacement;
    //need to append prev button when overlay so that it is in front of the
    //filmstrip items in z-order
    if (navArrowPlacement === _OVERLAY)
      elem.append(container); // @HTMLUpdateOK
    else
      elem.prepend(container); // @HTMLUpdateOK
    
    var label = _escapeHtml(this.getTranslatedString("labelAccArrowPreviousPage"));
    var tooltip = _escapeHtml(this.getTranslatedString("tipArrowPreviousPage"));
    var navArrow = this._createNavigationArrow(container, locationMarker, label, tooltip);
    var self = this;
    navArrow.on("click", function() {self._prevPage();});
    
    return navArrow;
  },
  
  /** 
   * Create the next navigation arrow.
   * @returns {jQuery} Navigation arrow jQuery object.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _createNextNavArrow: function()
  {
    var elem = this.element;
    var bHorizontal = this._isHorizontal();
    var locationMarker = bHorizontal ? _OJ_END : _OJ_BOTTOM;
    var container = this._createNavigationArrowContainer(locationMarker);
    elem.append(container); // @HTMLUpdateOK
    
    var label = _escapeHtml(this.getTranslatedString("labelAccArrowNextPage"));
    var tooltip = _escapeHtml(this.getTranslatedString("tipArrowNextPage"));
    var navArrow = this._createNavigationArrow(container, locationMarker, label, tooltip);
    var self = this;
    navArrow.on("click", function() {self._nextPage();});
    
    return navArrow;
  },
  
  /** 
   * Create the container around a navigation arrow.
   * @param {string} locationMarker Marker class for the location of the 
   *        navigation arrow: oj-top, oj-bottom, oj-start, or oj-end.
   * @returns {jQuery} jQuery object for the container around a navigation arrow.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _createNavigationArrowContainer: function(locationMarker)
  {
    var container = $(document.createElement("div"));
    container.addClass(_OJ_FILMSTRIP_ARROW_CONTAINER + " " + locationMarker);
    var options = this.options;
    var navArrowPlacement = options.arrowPlacement;
    if (navArrowPlacement === _OVERLAY)
    {
      container.addClass("oj-filmstrip-arrow-container-overlay");
      
      if (this._navArrowsShownOnHover())
        container.addClass(_OJ_FILMSTRIP_ARROW_TRANSITION);
    }
    return container;
  },
  
  /** 
   * Create a navigation arrow.
   * @param {jQuery} parentElem Parent DOM element of the navigation arrow.
   * @param {string} locationMarker Marker class for the location of the 
   *        navigation arrow: oj-top, oj-bottom, oj-start, or oj-end.
   * @param {string} label Accessible label for the navigation arrow.
   * @param {string} tooltip Tooltip for the navigation arrow.
   * @returns {jQuery} Navigation arrow jQuery object.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _createNavigationArrow: function(parentElem, locationMarker, label, tooltip)
  {
    var str = "<div class='" + _OJ_FILMSTRIP_ARROW + " oj-default oj-enabled " + locationMarker + "' role='button' tabindex='-1'";
    str += "><span class='oj-filmstrip-arrow-icon " + locationMarker + " oj-component-icon'></span></div>";
    parentElem.append(str); // @HTMLUpdateOK
    var elem = this.element;
    var elemId = elem.attr("id");
    var arrowElem = parentElem.children(_PERIOD + _OJ_FILMSTRIP_ARROW).eq(0);
    arrowElem.uniqueId();
    var arrowId = arrowElem.attr("id");
    if (label)
      arrowElem.attr("aria-label", label);
    if (tooltip)
      arrowElem.attr("title", tooltip);
    arrowElem.attr("aria-labelledby", arrowId + " " + elemId);

    //  - filmstrip: next/previous oj-hover colors doesn't go away in touch device
    this._AddHoverable(arrowElem);
    this._AddActiveable(arrowElem);
    
    var options = this.options;
    var navArrowPlacement = options.arrowPlacement;
    if (navArrowPlacement === _ADJACENT)
    {
      if (this._navArrowsShownOnHover())
        arrowElem.addClass(_OJ_FILMSTRIP_ARROW_TRANSITION);
    }
    
    return arrowElem;
  },
  
  /** 
   * Get the containers around the original child items.
   * @returns {jQuery} The containers around the original child items.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _getItemContainers: function()
  {
    var pagesWrapper = this._pagesWrapper;
    var items = pagesWrapper.find(_PERIOD + _OJ_FILMSTRIP_ITEM_CONTAINER)
      .filter(this._filterNestedFilmStripsFunc);
    return items;
  },
  
  /** 
   * Get the original child items.
   * @returns {jQuery} The original child items.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _getItems: function()
  {
    var pagesWrapper = this._pagesWrapper;
    var originalItems = pagesWrapper.find(_PERIOD + _OJ_FILMSTRIP_ITEM)
      .filter(this._filterNestedFilmStripsFunc);
    return originalItems;
  },
  
  /** 
   * Get the logical page containers.
   * @returns {jQuery} The logical page containers.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _getPages: function()
  {
    var pagesWrapper = this._pagesWrapper;
    var pages = pagesWrapper.children(_PERIOD + _OJ_FILMSTRIP_PAGE);
    return pages;
  },
  
  /** 
   * Clear the calculated sizes set on the internal layout DOM structure.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _clearCalculatedSizes: function()
  {
    var pagesWrapper = this._pagesWrapper;
    var pages = this._getPages();
    pages
      .css(_FLEX_BASIS, _EMPTY_STRING)
      .css(_WEBKIT_FLEX_BASIS, _EMPTY_STRING);
    var items = this._getItemContainers();
    items
      .css(_FLEX_BASIS, _EMPTY_STRING)
      .css(_WEBKIT_FLEX_BASIS, _EMPTY_STRING);
    pagesWrapper.css(this._getCssSizeAttr(), _EMPTY_STRING);
  },
  
  /** 
   * Calculate sizes and set them on the internal layout DOM structure.
   * @param {boolean} bNotifyAttach True to notify children when they're 
   *        attached to/detached from the DOM, false to not notify them (when 
   *        the notification will happen outside of this method).
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _adjustSizes: function(bNotifyAttach)
  {
    //clear previously calculated values before recalculating
    this._clearCalculatedSizes();
    
    var options = this.options;
    var bHorizontal = this._isHorizontal();
    var itemsPerPage = options.maxItemsPerPage;
    var bCalcItemsPerPage = (itemsPerPage < 1);
    var elem = this.element;
    
    var items = this._getItemContainers();
    //if we haven't saved the item size yet, do it now
    if (this._itemSize < 0)
    {
      //use the size of the item specified in the options
      var optionItemIndex = this._getItemIndex(options.currentItem);
      var optionItemContainer = $(items[optionItemIndex]);
      
      //unhide the item, in case the app has initially hidden it
      var optionItem = optionItemContainer.children(_PERIOD + _OJ_FILMSTRIP_ITEM);
      optionItem.css(_DISPLAY, _EMPTY_STRING);
      //notify the item that it's being shown
      oj.Components.subtreeShown(optionItem[0]);
      
      this._itemSize = bHorizontal ? optionItemContainer.width() : optionItemContainer.height();
    }
    
    //get component size after itemSize, in case item needed to be unhidden
    var componentSize = bHorizontal ? elem.width() : elem.height();
    if (options.arrowVisibility !== _HIDDEN &&
        options.arrowPlacement === _ADJACENT)
    {
      var arrowContainers = elem.children(_PERIOD + _OJ_FILMSTRIP_ARROW_CONTAINER);
      var firstArrowContainer = arrowContainers.eq(0);
      var arrowSize = bHorizontal ? firstArrowContainer.width() : firstArrowContainer.height();
      componentSize -= 2 * arrowSize;
    }
    this._componentSize = componentSize;
    
    //if there are a fixed number of items per page, but the number specified
    //won't fit, then reduce the number to what will fit
    if (!bCalcItemsPerPage)
    {
      //use min of 1 to prevent browser crash
      var calcFitCount = Math.max(Math.floor(componentSize / this._itemSize), 1);
      if (calcFitCount < itemsPerPage)
        itemsPerPage = calcFitCount;
    }
    
    //if calculating fitCount, use min of 1 to prevent browser crash
    var fitCount = bCalcItemsPerPage ? Math.max(Math.floor(componentSize / this._itemSize), 1) : itemsPerPage;
    var fitItemSize = componentSize / fitCount;
    items
      .css(_FLEX_BASIS, fitItemSize + _PX)
      .css(_WEBKIT_FLEX_BASIS, fitItemSize + _PX);
    
    var newPageCount = Math.ceil(items.length / fitCount);
    //wrap items in logical page containers
    var pages = this._getPages();
    var bCreatePages = false;
    //need to create logical pages if page layout changed, or if we haven't 
    //yet created logical pages
    var pagingModel = this._pagingModel;
    if (pagingModel.getPageCount() != newPageCount ||
        this._itemsPerPage != fitCount ||
        !pages || pages.length < 1)
    {
      bCreatePages = true;
      if (bNotifyAttach)
      {
        //notify the original children of the filmStrip that they were detached from
        //the DOM BEFORE actually detaching so that components can save state
        for (var i = 0; i < items.length; i++)
          oj.Components.subtreeDetached(items[i]);
      }
      
      if (pages && pages.length > 0)
      {
        //remove old logical page containers
        items.unwrap();
      }
      
      //create new logical page containers
      for (var i = 0; i < items.length; i += fitCount)
      {
        var itemsOnPage = items.slice(i, i + fitCount);
        //initially hide the page container
        itemsOnPage.wrapAll("<div class='" + _OJ_FILMSTRIP_CONTAINER + " " + _OJ_FILMSTRIP_PAGE + "' style='" + _DISPLAY + ": " + _NONE + ";' " + _ARIA_HIDDEN + "='true'></div>"); // @HTMLUpdateOK
      }

      if (bNotifyAttach)
      {
        //notify the original children of the filmStrip that they were attached
        //again after wrapping them
        for (var i = 0; i < items.length; i++)
          oj.Components.subtreeAttached(items[i]);
      }
    }
    //always need to update size of each page, even if page count doesn't change
    pages = this._getPages();
    pages
      .css(_FLEX_BASIS, componentSize + _PX)
      .css(_WEBKIT_FLEX_BASIS, componentSize + _PX);
    
    //always need to update the pages container size
    var pagesWrapper = this._pagesWrapper;
    var contentWrapper = this._contentWrapper;
    //only show a single page at a time
    pagesWrapper.css(this._getCssSizeAttr(), componentSize);
    if (contentWrapper)
      contentWrapper.css(this._getCssSizeAttr(), componentSize);
    
    var newPageIndex = 0;
    if (options.currentItem)
      newPageIndex = this._findPage(options.currentItem, fitCount);
    
    if (pagingModel.getPageCount() != newPageCount || 
        this._itemsPerPage != fitCount ||
        pagingModel.getPage() != newPageIndex)
    {
      pagingModel.setPage(newPageIndex, {"pageSize": fitCount});
    }
    //if the page layout didn't change, but we recreated the logical page
    //containers, then simply go to the current logical page to make sure that
    //pages and items are hidden or shown as appropriate
    else if (bCreatePages)
    {
      var currPage = pagingModel.getPage();
      this._handlePage({"previousPage": currPage, "page": currPage});
    }
  },
  
  /** 
   * Handle a 'page' event from the PagingModel.
   * @param {Object} event Event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handlePage: function(event)
  {
    var pageIndex = event["page"];
    var prevPageIndex = event["previousPage"];
    var pagesWrapper = this._pagesWrapper;
    var pages = this._getPages();
    var pagingModel = this._pagingModel;
    var pageSize = pagingModel.getPageSize();
    var bImmediate = prevPageIndex < 0 || prevPageIndex == pageIndex ||
                     this._itemsPerPage != pageSize;
    //update _itemsPerPage AFTER using it to initialize bImmediate above
    this._itemsPerPage = pageSize;
    
    //get the old page, if there is one
    var oldPage = null;
    if (!bImmediate)
      oldPage = $(pages[prevPageIndex]);
    
    var cssAttr = this._getCssPositionAttr();
    
    //unhide the new page
    var newPage = $(pages[pageIndex]);
    var bPageHidden = newPage.is(_HIDDEN_SELECTOR);
    if (bPageHidden)
      this._unhidePage(newPage);
    //defer the scroll if we're dragging so that we can animate it
    var bDeferScroll = this._bDragInit;
    if (prevPageIndex > -1 && !bImmediate)
    {
      bDeferScroll = true;
      //set size of pages wrapper to show two pages at a time because both of
      //them will temporarily be visible
      pagesWrapper.css(this._getCssSizeAttr(), 2 * this._componentSize);
      //if going to the previous page, initially scroll to show the current page
      //because the previous page will be at the current scroll position
      if (pageIndex < prevPageIndex)
      {
        //only adjust scroll if the page was hidden (in the case of dragging,
        //the page will already be unhidden and scrolled correctly)
        if (bPageHidden)
          pagesWrapper.css(cssAttr, -this._componentSize + _PX);
      }
      
      //set initial transition states on the pages
      if (pageIndex > prevPageIndex)
      {
        oldPage.addClass(_OJ_FILMSTRIP_TRANSITION_NEXT_OLDPAGE_FROM);
        newPage.addClass(_OJ_FILMSTRIP_TRANSITION_NEXT_NEWPAGE_FROM);
      }
      else
      {
        oldPage.addClass(_OJ_FILMSTRIP_TRANSITION_PREV_OLDPAGE_FROM);
        newPage.addClass(_OJ_FILMSTRIP_TRANSITION_PREV_NEWPAGE_FROM);
      }
    }
    
    //FIX : add busy state before animating a scroll
    this._busyStateResolveFunc = this._addBusyState("scrolling");
    
    if (bDeferScroll)
    {
      var self = this;
      var bDragInit = this._bDragInit;
      //if bouncing back because a drag scroll didn't cross the threshold,
      //add the transition class before the timeout because the transforms
      //are removed and reapplied in the timeout
      if (bDragInit && prevPageIndex < 0)
      {
        var visiblePages = pages.filter(_VISIBLE_SELECTOR);
        visiblePages.addClass(_OJ_FILMSTRIP_TRANSITION);
      }
      //FIX : In Safari on Mac OS X, sometimes when changing pages
      //the new page appears hidden until after the transition.  It seems to be
      //a timing issue where the new page has been made visible in the DOM, but
      //Safari hasn't processed that change before the transition starts.  
      //Changing the timeout delay from 0 to 25 seems to help.
      setTimeout(function()
        {
          self._finishHandlePage(pageIndex, prevPageIndex, bImmediate, bDragInit);
        }, 25);
    }
    else
      this._finishHandlePage(pageIndex, prevPageIndex, bImmediate);
  },
  
  /** 
   * Finish handling a 'page' event from the PagingModel.
   * @param {number} pageIndex 0-based page index to go to.
   * @param {number} prevPageIndex 0-based old page index.
   * @param {boolean} bImmediate True to change pages immediately with no 
   *        transition, false to transition them over time.
   * @param {boolean} bDragInit True if we're currently drag scrolling, false
   *        otherwise.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _finishHandlePage: function(pageIndex, prevPageIndex, bImmediate, bDragInit)
  {
    var pagesWrapper = this._pagesWrapper;
    if (!bImmediate)
    {
      this._bPageChangeTransition = true;
      pagesWrapper
        .on("transitionend" + this.eventNamespace + " webkitTransitionEnd" + this.eventNamespace, 
            this._handleTransitionEndFunc);
    }

    if (bImmediate)
      this._handleTransitionEnd();
    else
    {
      var pages = this._getPages();
      //if we're currently drag scrolling, remove the transforms that we used
      //to scroll while we were under the drag threshold
      if (bDragInit)
        _removeTransform(pages);
      
      //changing pages
      if (prevPageIndex > -1)
      {
        //remove initial transition states and set destination states, and
        //transition between them
        var bNext = pageIndex > prevPageIndex;
        var oldPage = $(pages[prevPageIndex]);
        var newPage = $(pages[pageIndex]);
        oldPage.addClass(_OJ_FILMSTRIP_TRANSITION);
        newPage.addClass(_OJ_FILMSTRIP_TRANSITION);
        if (bNext)
        {
          oldPage.removeClass(_OJ_FILMSTRIP_TRANSITION_NEXT_OLDPAGE_FROM);
          newPage.removeClass(_OJ_FILMSTRIP_TRANSITION_NEXT_NEWPAGE_FROM);
          oldPage.addClass(_OJ_FILMSTRIP_TRANSITION_NEXT_OLDPAGE_TO);
          newPage.addClass(_OJ_FILMSTRIP_TRANSITION_NEXT_NEWPAGE_TO);
        }
        else
        {
          oldPage.removeClass(_OJ_FILMSTRIP_TRANSITION_PREV_OLDPAGE_FROM);
          newPage.removeClass(_OJ_FILMSTRIP_TRANSITION_PREV_NEWPAGE_FROM);
          oldPage.addClass(_OJ_FILMSTRIP_TRANSITION_PREV_OLDPAGE_TO);
          newPage.addClass(_OJ_FILMSTRIP_TRANSITION_PREV_NEWPAGE_TO);
        }
      }
      //bouncing back to current page after drag scrolling under the threshold
      else if (bDragInit)
      {
        var visiblePages = pages.filter(_VISIBLE_SELECTOR);
        _applyTransform(visiblePages, "translate3d(0, 0, 0)");
      }
    }
  },
  
  /** 
   * Handle the end of a page change transition.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleTransitionEnd: function()
  {
    this._bPageChangeTransition = false;
    var pagesWrapper = this._pagesWrapper;
    var cssAttr = this._getCssPositionAttr();
    pagesWrapper
      .off(this.eventNamespace)
      .css(this._getCssSizeAttr(), this._componentSize)
      .css(cssAttr, "0px");
    
    //if the filmStrip contains the focus element, we may need to focus a new
    //element if the old one is on a page that's being hidden
    var focusElem = null;
    if (oj.FocusUtils.containsFocus(pagesWrapper[0]) ||
        (this._nextButton && oj.FocusUtils.containsFocus(this._nextButton[0])) ||
        (this._prevButton && oj.FocusUtils.containsFocus(this._prevButton[0])))
      focusElem = document.activeElement;
    
    var pagingModel = this._pagingModel;
    var pageIndex = pagingModel.getPage();
    
    //hide all pages except for current one
    var pages = this._getPages();
    for (var i = 0; i < pages.length; i++)
    {
      if (i != pageIndex)
        this._hidePage($(pages[i]));
    }
    //remove transition classes
    pages
      .removeClass(_OJ_FILMSTRIP_TRANSITION + " " + 
                   _OJ_FILMSTRIP_TRANSITION_NEXT_OLDPAGE_TO + " " + 
                   _OJ_FILMSTRIP_TRANSITION_NEXT_NEWPAGE_TO + " " + 
                   _OJ_FILMSTRIP_TRANSITION_PREV_OLDPAGE_TO + " " + 
                   _OJ_FILMSTRIP_TRANSITION_PREV_NEWPAGE_TO);
    //remove transforms left over from drag scrolling
    _removeTransform(pages);
    
    //update display of the navigation arrows
    this._updateNavigationArrowsDisplay();
    
    //if the old focus element is being hidden, transfer focus to something visible
    if (focusElem && $(focusElem).is(_HIDDEN_SELECTOR))
    {
      var elem = this.element;
      //focus an element in the new logical page if possible, otherwise focus
      //the filmStrip itself
      var firstTabStop = oj.FocusUtils.getFirstTabStop(pages[pageIndex]);
      if (firstTabStop)
        oj.FocusUtils.focusElement(firstTabStop);
      else
        oj.FocusUtils.focusElement(elem[0]);
    }
    
    //update currentItem property if it's not on the current page
    var options = this.options;
    var currItemPage = this._findPage(options.currentItem);
    if (currItemPage != pageIndex)
    {
      var newFirstItem = this._getFirstItemOnPage(pageIndex);
      //FIX : only update currentItem property if the filmstrip is not empty
      if (newFirstItem)
        this.option(_CURRENT_ITEM, newFirstItem, {'_context': {writeback: true}});
    }
    
    //FIX : if we've deferred handling a resize during a page change, handle it now
    if (this._deferredHandleResize)
    {
      this._deferredHandleResize = false;
      
      this._handleResize();
    }
    
    //FIX : resolve busy state after animating a scroll
    this._resolveBusyState();
  },
  
  /** 
   * Get the 0-based index of the specified item.
   * @param {Object} item The item object.
   * @returns {number} The 0-based index of the specified item, or -1 if not found.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _getItemIndex: function(item)
  {
    var itemIndex = -1;
    if (item) {
      var items = this._getItems();
      // If the item contains both id and index, item id takes precedence.
      if (item.id && oj.DomUtils.isValidIdentifier(item.id)) {
        for (var i = 0; i < items.length; i++)
        {
          var itemElem = items[i];
          var itemId = itemElem.id;
          if (itemId && itemId.length > 0 && itemId === item.id)
          {
            itemIndex = i;
            break;
          }
        }
      }
      else if (item.index != null && item.index >= 0 && item.index < items.length) {
        itemIndex = item.index;
      }
    }
    return itemIndex;
  },
  
   /** 
   * Convert any 0-based numeric item index or string item id to an item object of (id, index).
   * @param {(string|number|Object)} item String item id or numeric item index or an item object of (id, index).
   * @returns {Object} The item object of (id, index).
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _convertItemToObj: function(item)
  {
    var itemObj = null;
    if (typeof item === "object") {
      itemObj = item;
    }
    else if (typeof item === "number") {
      itemObj = {"index": item};
    }
    else if (typeof item === "string") {
      itemObj = {"id": item};
    }
    return itemObj;
  },
   
   /** 
   * Populates the item object with valid id and index
   * @param {Object} item The item object to populate.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _populateItemObj: function(item)
  {
    if (item && this._pagingModel.getPage() >= 0) {
      var index = this._getItemIndex(item);
      item["index"] = index;
      if (item["id"] == null && index !== -1) {
        var items = this._getItems();
        item["id"] = items[index].id;
      }
    }
  },
  
  /** 
   * Find the logical page containing the specified item.
   * @param {Object} item The item object.
   * @param {?number} itemsPerPage The number of items on each logical page.
   * @returns {number} The 0-based index of the logical page containing the item, 
   *          or -1 if not found.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _findPage: function(item, itemsPerPage)
  {
    var itemIndex = this._getItemIndex(item);
    var pageIndex = -1;
    if (itemIndex > -1)
    {
      if (itemsPerPage === undefined)
        itemsPerPage = this._itemsPerPage;
      pageIndex = Math.floor(itemIndex / itemsPerPage);
    }
    return pageIndex;
  },
  
  /** 
   * Get the first item on the specified logical page.
   * @param {number} pageIndex 0-based index of the logical page.
   * @param {?number} pageCount Number of logical pages.
   * @param {?number} itemsPerPage Number of items on each logical page.
   * @returns {Object} The first item object.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _getFirstItemOnPage: function(pageIndex, pageCount, itemsPerPage)
  {
    var pagingModel = this._pagingModel;
    if (pageCount === undefined)
      pageCount = pagingModel.getPageCount();
    if (pageIndex >= 0 && pageIndex < pageCount)
    {
      var items = this._getItems();
      if (itemsPerPage === undefined)
        itemsPerPage = this._itemsPerPage;
      var itemIndex = pageIndex * itemsPerPage;
      if (itemIndex < items.length)
      {
        var firstItemOnPage = items[itemIndex];
        var firstId = firstItemOnPage.id;
        var firstItem = {"id" : firstId, "index" : itemIndex};
        return firstItem;
      }
    }
    return null;
  },
  
  /** 
   * Hide the logical page.
   * @param {jQuery} page Page to hide.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _hidePage: function(page)
  {
    //notify the page that it's being hidden BEFORE actually hiding it so that
    //components can save state
    oj.Components.subtreeHidden(page[0]);
    
    page
      .css(_DISPLAY, _NONE)
      .attr(_ARIA_HIDDEN, "true");
    
    //hide the items explicitly; unhiding will unhide them explicitly
    var items = page.find(_PERIOD + _OJ_FILMSTRIP_ITEM).filter(this._filterNestedFilmStripsFunc);
    items.css(_DISPLAY, _NONE);
  },
  
  /** 
   * Unhide the logical page.
   * @param {jQuery} page Page to unhide.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _unhidePage: function(page)
  {
    page
      .css(_DISPLAY, _EMPTY_STRING)
      .removeAttr(_ARIA_HIDDEN);
    
    //unhide the items explicitly because the app may have initially hidden them  
    var items = page.find(_PERIOD + _OJ_FILMSTRIP_ITEM).filter(this._filterNestedFilmStripsFunc);
    items.css(_DISPLAY, _EMPTY_STRING);
    
    //notify the page that it was shown
    oj.Components.subtreeShown(page[0]);
  },
  
  /** 
   * Add key listeners on the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _addKeyListeners: function()
  {
    var elem = this.element;
    elem.on("keydown" + this.keyEventNamespace, this._handleKeyDownFunc);
  },
  
  /** 
   * Remove key listeners from the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _removeKeyListeners: function()
  {
    var elem = this.element;
    elem.off(this.keyEventNamespace);
  },
  
  /** 
   * Add mouse listeners on the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _addMouseListeners: function()
  {
    var elem = this.element;
    elem
      .on("mousedown" + this.mouseEventNamespace, this._handleMouseDownFunc)
      .on("mousemove" + this.mouseEventNamespace, this._handleMouseMoveFunc)
      .on("mouseup" + this.mouseEventNamespace, this._handleMouseUpFunc);
  },
  
  /** 
   * Remove mouse listeners from the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _removeMouseListeners: function()
  {
    var elem = this.element;
    elem.off(this.mouseEventNamespace);
  },
  
  /** 
   * Add touch listeners on the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _addTouchListeners: function()
  {
    var elem = this.element;
    elem
      .on("touchstart" + this.touchEventNamespace, this._handleTouchStartFunc)
      .on("touchmove" + this.touchEventNamespace, this._handleTouchMoveFunc)
      .on("touchend" + this.touchEventNamespace, this._handleTouchEndFunc)
      .on("touchcancel" + this.touchEventNamespace, this._handleTouchEndFunc);
  },
  
  /** 
   * Remove touch listeners from the filmStrip.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _removeTouchListeners: function()
  {
    var elem = this.element;
    elem.off(this.touchEventNamespace);
  },
  
  /** 
   * Handle a keydown event.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleKeyDown: function(event)
  {
    var pagingModel = this._pagingModel;
    var pageIndex = pagingModel.getPage();
    var newPageIndex = -1;
    switch (event.keyCode)
    {
      case $.ui.keyCode.RIGHT:
        if (this._bRTL)
          newPageIndex = pageIndex - 1;
        else
          newPageIndex = pageIndex + 1;
        break;
      case $.ui.keyCode.LEFT:
        if (this._bRTL)
          newPageIndex = pageIndex + 1;
        else
          newPageIndex = pageIndex - 1;
        break;
      case $.ui.keyCode.DOWN:
        newPageIndex = pageIndex + 1;
        break;
      case $.ui.keyCode.UP:
        newPageIndex = pageIndex - 1;
        break;
      case $.ui.keyCode.HOME:
        newPageIndex = 0;
        break;
      case $.ui.keyCode.END:
        newPageIndex = pagingModel.getPageCount() - 1;
        break;
      default :
        return;
    }
    
    if (newPageIndex > -1 && newPageIndex < pagingModel.getPageCount())
      pagingModel.setPage(newPageIndex);
    
    event.preventDefault();
  },
  
  /** 
   * Handle a mousedown event.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleMouseDown: function(event)
  {
    var originalEvent = event.originalEvent;
    this._dragScrollStart(originalEvent);
  },
  
  /** 
   * Handle a mousemove event.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleMouseMove: function(event)
  {
    var originalEvent = event.originalEvent;
    this._dragScrollMove(event, originalEvent);
  },
  
  /** 
   * Handle a mouseup event.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleMouseUp: function(event)
  {
    this._dragScrollEnd();
  },
  
  /** 
   * Handle a touchstart event.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleTouchStart: function(event)
  {
    var originalEvent = event.originalEvent;
    var eventTouches = originalEvent.touches;
    if (eventTouches.length === 1)
    {
      var firstTouch = eventTouches[0];
      this._dragScrollStart(firstTouch);
    }
  },
  
  /** 
   * Handle a touchmove event.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleTouchMove: function(event)
  {
    var originalEvent = event.originalEvent;
    var eventTouches = originalEvent.touches;
    var firstTouch = eventTouches[0];
    this._dragScrollMove(event, firstTouch);

    //FIX : if the drag scroll was started, prevent the page
    //from scrolling also
    //FIX : do this for touchMove events instead of touchStart so 
    //that it doesn't prevent clicks/taps from being detected on nav arrows and
    //item content
    if (this._bTouch || this._scrolledForThisTouch)
      event.preventDefault();
  },
  
  /** 
   * Handle a touchend event.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _handleTouchEnd: function(event)
  {
    this._dragScrollEnd();
  },
  
  /** 
   * Start a drag scroll.
   * @param {Event} coordsObj Object that has pageX and pageY properties.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _dragScrollStart: function(coordsObj)
  {
    var pagingModel = this._pagingModel;
    if (pagingModel.getPageCount() > 1 && !this._bPageChangeTransition)
    {
      this._bTouch = true;
      this._bDragInit = false;
      
      var bHorizontal = this._isHorizontal();
      this._touchStartCoord = bHorizontal ? coordsObj.pageX : coordsObj.pageY;
      this._touchStartCoord2 = bHorizontal ? coordsObj.pageY : coordsObj.pageX;
    }
  },
  
  /** 
   * Initialize a drag scroll.
   * Called after a drag scroll has started if the drag has passed an initial
   * threshold.
   * @param {Object} coordsObj Object that has pageX and pageY properties.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _initDragScroll: function(coordsObj)
  {
    //save off some initial information at the start of a swipe
    var bHorizontal = this._isHorizontal();
    this._touchStartCoord = bHorizontal ? coordsObj.pageX : coordsObj.pageY;
    this._touchStartCoord2 = bHorizontal ? coordsObj.pageY : coordsObj.pageX;

    var cssAttr = this._getCssPositionAttr();
    var pagesWrapper = this._pagesWrapper;
    var pagingModel = this._pagingModel;

    //unhide adjacent pages (unhide both because we don't know which way user 
    //will scroll)
    var pages = this._getPages();
    var pageCountToShow = 1;
    if (pagingModel.getPage() > 0)
    {
      this._unhidePage($(pages[pagingModel.getPage() - 1]));

      //when unhiding previous page, need to adjust current scroll position
      //to continue showing current page
      pagesWrapper.css(cssAttr, -this._componentSize + _PX);

      //increment number of pages we need to show
      pageCountToShow++;
    }
    if (pagingModel.getPage() < pagingModel.getPageCount() - 1)
    {
      this._unhidePage($(pages[pagingModel.getPage() + 1]));

      //increment number of pages we need to show
      pageCountToShow++;
    }

    if (pageCountToShow > 1)
      pagesWrapper.css(this._getCssSizeAttr(), pageCountToShow * this._componentSize);

    this._touchStartScroll = parseInt(pagesWrapper.css(cssAttr), 10);
  },

  /** 
   * Process a move during drag scrolling.
   * @param {Event} event <code class="prettyprint">jQuery</code> event object.
   * @param {Object} coordsObj Object that has pageX and pageY properties.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _dragScrollMove: function(event, coordsObj)
  {
    //do nothing if we haven't started the scroll
    if (!this._bTouch)
      return;
    
    var bHorizontal = this._isHorizontal();
    var touchCoord = bHorizontal ? coordsObj.pageX : coordsObj.pageY;
    var diff = touchCoord - this._touchStartCoord;
    
    // - cannot scroll vertically in filmstrip component
    var touchCoord2 = bHorizontal ? coordsObj.pageY : coordsObj.pageX;
    var diff2 = touchCoord2 - this._touchStartCoord2;

    if (Math.abs(diff2) > Math.abs(diff)) {
      //don't scroll again for this same swipe
      this._bTouch = false;

      //set a flag indicating we don't scroll for this touch event
      this._scrolledForThisTouch = false;
    }

    //only initialize the drag once we've passed the initial threshold
    if (!this._bDragInit)
    {
      if (Math.abs(diff) > _DRAG_SCROLL_INIT_THRESHOLD)
      {
        this._initDragScroll(coordsObj);
        this._bDragInit = true;
      }
      //return here even if we just initialized the drag so that we'll start
      //processing the drag move on the next event when we have different mouse
      //coords
      return;
    }
    
    //in non-RTL, if swiping left or up, scroll next; otherwise scroll prev
    //in RTL, if swiping right or up, scroll next; otherwise scroll prev
    var bNext = (bHorizontal && this._bRTL) ? (diff > 0) : (diff < 0);
    //determine whether the filmStrip can be scrolled in the direction of the swipe
    var pagingModel = this._pagingModel;
    var pageIndex = pagingModel.getPage();
    var canScrollInSwipeDirection = (bNext && pageIndex < (pagingModel.getPageCount() - 1)) ||
                                    (!bNext && pageIndex > 0);
    //only need to do something if we can scroll in the swipe direction
    if (canScrollInSwipeDirection)
    {
      //only scroll next/prev if the swipe is longer than the threshold; if it's
      //less, then just drag the items with the swipe
      var elem = this.element[0];
      var threshold = Math.min(_DRAG_SCROLL_THRESHOLD * (bHorizontal ? elem.offsetWidth : elem.offsetHeight), _DRAG_SCROLL_MAX_THRESHOLD);
      var cssAttr = this._getCssPositionAttr();
      var pagesWrapper = this._pagesWrapper;
      var pages = this._getPages();

      //if swiping beyond the threshold, scroll to the next/prev page
      if (Math.abs(diff) >= threshold)
      {
        var newPageIndex = bNext ? pageIndex + 1 : pageIndex - 1;
        
        //hide the page that we're not scrolling to
        var pageToHide = bNext ? pageIndex - 1 : pageIndex + 1;
        if (pageToHide > -1 && pageToHide < pagingModel.getPageCount())
          this._hidePage($(pages[pageToHide]));
        
        //when hiding previous page, need to adjust current scroll position
        //to continue showing current page
        if (bNext && pageToHide > -1)
        {
          var currScroll = parseInt(pagesWrapper.css(cssAttr), 10);
          pagesWrapper.css(cssAttr, (currScroll + this._componentSize) + _PX);
        }
        
        //update size to show two pages instead of three
        pagesWrapper.css(this._getCssSizeAttr(), 2 * this._componentSize);

        //don't scroll again for this same swipe
        this._bTouch = false;
        
        pagingModel.setPage(newPageIndex);
      }
      //if swiping under the threshold, just move the conveyor with the swipe
      else
      {
        var scrollVal = (bHorizontal && this._bRTL) ? -diff : diff;
        var transform = bHorizontal ? "translate3d(" + scrollVal + "px, 0, 0)" :
                                      "translate3d(0, " + scrollVal + "px, 0)";
        _applyTransform(pages.filter(_VISIBLE_SELECTOR), transform);
      }

      //set a flag indicating we've scrolled for this touch event
      this._scrolledForThisTouch = true;
    }

    //if we've scrolled for this touch event, consume the event
    //so that the page doesn't also scroll
    if (this._scrolledForThisTouch)
    {
      event.preventDefault();
      event.stopPropagation();
    }
  },

  /** 
   * Process the end of a drag scroll.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _dragScrollEnd: function()
  {
    //if a full page swipe hasn't happened, scroll back to the original position
    if (this._bTouch && this._bDragInit)
    {
      var pagingModel = this._pagingModel;
      var pageIndex = pagingModel.getPage();
      this._handlePage({"previousPage": pageIndex, "page": pageIndex});
    }
    this._bTouch = false;
    this._bDragInit = false;
    //reset the flag indicating if we've scrolled for this touch event
    this._scrolledForThisTouch = false;
  },
  
  /**
   * Add a busy state to the busy context.
   *
   * @param {string} description Additional information about busy state.
   * @returns {Function} Resolve function called by the registrant when the busy state completes.
   *          The resultant function will throw an error if the busy state is no longer registered.
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _addBusyState: function(description)
  {
    var element = this.element;
    var context = oj.Context.getContext(element[0]);
    var busyContext = context.getBusyContext();
    
    var desc = "FilmStrip";
    var id = element.attr("id");
    desc += " (id='" + id + "')";
    desc += ": " + description;
    
    var busyStateOptions = {"description" : desc};
    return busyContext.addBusyState(busyStateOptions);
  },
  
  /**
   * Resolve an outstanding busy state.
   * @return {void}
   * @memberof oj.ojFilmStrip
   * @instance
   * @private
   */
  _resolveBusyState: function()
  {
    if (this._busyStateResolveFunc)
    {
      this._busyStateResolveFunc();
      this._busyStateResolveFunc = null;
    }
  },
  
  // @inheritdoc
  getNodeBySubId: function(locator)
  {
    if (locator == null)
    {
      return this.element ? this.element[0] : null;
    }
    
    var subId = locator['subId'];
    if (subId === "oj-filmstrip-start-arrow") {
      return this.widget().find(_PERIOD + _OJ_FILMSTRIP_ARROW + _PERIOD + _OJ_START)
        .filter(this._filterNestedFilmStripsFunc)[0];
    }
    if (subId === "oj-filmstrip-end-arrow") {
      return this.widget().find(_PERIOD + _OJ_FILMSTRIP_ARROW + _PERIOD + _OJ_END)
        .filter(this._filterNestedFilmStripsFunc)[0];
    }
    if (subId === "oj-filmstrip-top-arrow") {
      return this.widget().find(_PERIOD + _OJ_FILMSTRIP_ARROW + _PERIOD + _OJ_TOP)
        .filter(this._filterNestedFilmStripsFunc)[0];
    }
    if (subId === "oj-filmstrip-bottom-arrow") {
      return this.widget().find(_PERIOD + _OJ_FILMSTRIP_ARROW + _PERIOD + _OJ_BOTTOM)
        .filter(this._filterNestedFilmStripsFunc)[0];
    }
    
    // Non-null locators have to be handled by the component subclasses
    return null;
  },

  // @inheritdoc
  getSubIdByNode: function(node) {
    var startArrow = this.getNodeBySubId({'subId':'oj-filmstrip-start-arrow'});
    var endArrow = this.getNodeBySubId({'subId':'oj-filmstrip-end-arrow'});
    var topArrow = this.getNodeBySubId({'subId':'oj-filmstrip-top-arrow'});
    var bottomArrow = this.getNodeBySubId({'subId':'oj-filmstrip-bottom-arrow'});
    var currentNode = node;
    var elem = this.element[0];
    while (currentNode && currentNode != elem) {
      if (currentNode === startArrow)
        return {'subId':'oj-filmstrip-start-arrow'};
      else if (currentNode === endArrow)
        return {'subId':'oj-filmstrip-end-arrow'};
      else if (currentNode === topArrow)
        return {'subId':'oj-filmstrip-top-arrow'};
      else if (currentNode === bottomArrow)
        return {'subId':'oj-filmstrip-bottom-arrow'};

      currentNode = currentNode.parentElement;          
    }
    return null;
  },

  // @inheritdoc
  _CompareOptionValues: function(option, value1, value2) 
  {
    if (option == 'currentItem')
      return oj.Object.compareValues(value1, value2);
    else
      return this._super(option, value1, value2);
  }
  
  // start API doc fragments /////////////////////////////////////////////////////
  
  /**
   * <p>Sub-ID for the start navigation arrow of a horizontal FilmStrip.</p>
   *
   * @ojsubid oj-filmstrip-start-arrow
   * @memberof oj.ojFilmStrip
   *
   * @example <caption>Get the start navigation arrow:</caption>
   * var node = myFilmStrip.getNodeBySubId({'subId': 'oj-filmstrip-start-arrow'} );
   */
  
  /**
   * <p>Sub-ID for the end navigation arrow of a horizontal FilmStrip.</p>
   *
   * @ojsubid oj-filmstrip-end-arrow
   * @memberof oj.ojFilmStrip
   *
   * @example <caption>Get the end navigation arrow:</caption>
   * var node = myFilmStrip.getNodeBySubId({'subId': 'oj-filmstrip-end-arrow'} );
   */
  
  /**
   * <p>Sub-ID for the top navigation arrow of a vertical FilmStrip.</p>
   *
   * @ojsubid oj-filmstrip-top-arrow
   * @memberof oj.ojFilmStrip
   *
   * @example <caption>Get the top navigation arrow:</caption>
   * var node = myFilmStrip.getNodeBySubId({'subId': 'oj-filmstrip-top-arrow'} );
   */
  
  /**
   * <p>Sub-ID for the bottom navigation arrow of a vertical FilmStrip.</p>
   *
   * @ojsubid oj-filmstrip-bottom-arrow
   * @memberof oj.ojFilmStrip
   *
   * @example <caption>Get the bottom navigation arrow:</caption>
   * var node = myFilmStrip.getNodeBySubId({'subId': 'oj-filmstrip-bottom-arrow'} );
   */

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Gesture</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>FilmStrip</td>
   *       <td><kbd>Swipe</kbd></td>
   *       <td>Transition to an adjacent logical page of child items.</td>
   *     </tr>
   *     <tr>
   *       <td>Navigation Arrow</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Transition to an adjacent logical page of child items.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojFilmStrip
   */

  /**
   * <p>FilmStrip itself is a tabstop and supports the following keyboard interactions:
   * 
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>FilmStrip</td>
   *       <td><kbd>UpArrow or LeftArrow</kbd> (<kbd>RightArrow</kbd> in RTL)</td>
   *       <td>Transition to the previous logical page of child items.</td>
   *     </tr>
   *     <tr>
   *       <td>FilmStrip</td>
   *       <td><kbd>DownArrow or RightArrow</kbd> (<kbd>LeftArrow</kbd> in RTL)</td>
   *       <td>Transition to the next logical page of child items.</td>
   *     </tr>
   *     <tr>
   *       <td>FilmStrip</td>
   *       <td><kbd>Home</kbd></td>
   *       <td>Transition to the first logical page of child items.</td>
   *     </tr>
   *     <tr>
   *       <td>FilmStrip</td>
   *       <td><kbd>End</kbd></td>
   *       <td>Transition to the last logical page of child items.</td>
   *     </tr>
   *     <tr>
   *       <td>Navigation Arrow</td>
   *       <td><kbd>Enter</kbd> or <kbd>Space</kbd></td>
   *       <td>Transition to an adjacent logical page of child items.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojFilmStrip
   */
  
  // end API doc fragments ///////////////////////////////////////////////////////
  
}); // end of oj.__registerWidget

// start static members and functions //////////////////////////////////////////

//_tmpVar is simply to trick the closure compiler so that it doesn't issue a
//warning about the below real vars declared using a single "var" statement:
//WARNING - declaration of multiple variables with shared type information
var _tmpVar;

var _ADJACENT = "adjacent",
    //aria-hidden DOM attribute
    _ARIA_HIDDEN = "aria-hidden",
    _AUTO = "auto",
    //display CSS attribute
    _DISPLAY = "display",

    //threshold to initialize a drag scroll (pixels)
    _DRAG_SCROLL_INIT_THRESHOLD = 3,
    //drag scroll threshold (percentage of filmStrip size)
    _DRAG_SCROLL_THRESHOLD = .33,
    //drag scroll threshold (max distance, pixels)
    _DRAG_SCROLL_MAX_THRESHOLD = 100,
    
    _EMPTY_STRING = "",
    //throw an error when invalid "currentItem" property set
    _ERROR_CURRENT_ITEM_NOT_FOUND = "JET FilmStrip: Value of 'currentItem' property not found: ",
    //throw an error when "orientation" property set to invalid value
    _ERROR_INVALID_ORIENTATION = "JET FilmStrip: Unsupported value set as 'orientation' property: ",
    //throw an error when "arrowPlacement" property set to invalid value
    _ERROR_INVALID_NAV_ARROW_PLACEMENT = "Unsupported value set as 'arrowPlacement' property: ",
    //throw an error when "arrowVisibility" property set to invalid value
    _ERROR_INVALID_NAV_ARROW_VISIBILITY = "Unsupported value set as 'arrowVisibility' property: ",
    _FLEX_BASIS = "flex-basis",
    _HIDDEN = "hidden",
    //jQuery hidden selector
    _HIDDEN_SELECTOR = ":hidden",
    _HORIZONTAL = "horizontal",
    _HOVER = "hover",
    _CURRENT_ITEM = "currentItem",
    _MS_TRANSFORM = "-ms-transform",
    _NONE = "none",
    
    _OJ_BOTTOM = "oj-bottom",
    _OJ_END = "oj-end",
    _OJ_FILMSTRIP_ARROW = "oj-filmstrip-arrow",
    _OJ_FILMSTRIP_ARROW_CONTAINER = "oj-filmstrip-arrow-container",
    _OJ_FILMSTRIP_ARROW_TRANSITION = "oj-filmstrip-arrow-transition",
    _OJ_FILMSTRIP_CONTAINER = "oj-filmstrip-container",
    _OJ_FILMSTRIP_HOVER = "oj-filmstrip-hover",
    _OJ_FILMSTRIP_ITEM = "oj-filmstrip-item",
    _OJ_FILMSTRIP_ITEM_CONTAINER = "oj-filmstrip-item-container",
    _OJ_FILMSTRIP_PAGE = "oj-filmstrip-page",
    _OJ_FILMSTRIP_PAGES_CONTAINER = "oj-filmstrip-pages-container",
    _OJ_FILMSTRIP_TRANSITION = "oj-filmstrip-transition",
    _OJ_FILMSTRIP_TRANSITION_NEXT_NEWPAGE_FROM = "oj-filmstrip-transition-next-newpage-from",
    _OJ_FILMSTRIP_TRANSITION_NEXT_OLDPAGE_FROM = "oj-filmstrip-transition-next-oldpage-from",
    _OJ_FILMSTRIP_TRANSITION_PREV_NEWPAGE_FROM = "oj-filmstrip-transition-prev-newpage-from",
    _OJ_FILMSTRIP_TRANSITION_PREV_OLDPAGE_FROM = "oj-filmstrip-transition-prev-oldpage-from",
    _OJ_FILMSTRIP_TRANSITION_NEXT_NEWPAGE_TO = "oj-filmstrip-transition-next-newpage-to",
    _OJ_FILMSTRIP_TRANSITION_NEXT_OLDPAGE_TO = "oj-filmstrip-transition-next-oldpage-to",
    _OJ_FILMSTRIP_TRANSITION_PREV_NEWPAGE_TO = "oj-filmstrip-transition-prev-newpage-to",
    _OJ_FILMSTRIP_TRANSITION_PREV_OLDPAGE_TO = "oj-filmstrip-transition-prev-oldpage-to",
    _OJ_FILMSTRIP_VERTICAL = "oj-filmstrip-vertical",
    _OJ_START = "oj-start",
    _OJ_TOP = "oj-top",
    
    _OVERLAY = "overlay",
    _PERIOD = ".",
    _PX = "px",
    //make sure the collapseEventTimeout param is less than the one used in the unit tests
    //in order to ensure that the filmStrip listener gets the resize event before the unit test
    _RESIZE_LISTENER_COLLAPSE_EVENT_TIMEOUT = 25,
    _TRANSFORM = "transform",
    _VERTICAL = "vertical",
    _VISIBLE = "visible",
    //jQuery visible selector
    _VISIBLE_SELECTOR = ":visible",
    _WEBKIT_FLEX_BASIS = "-webkit-flex-basis",
    _WEBKIT_TRANSFORM = "-webkit-transform",
    
    //log warning message when "disabled" property set
    _WARNING_DISABLED_OPTION = "JET FilmStrip: 'disabled' property not supported",
    
    /**
     * Apply a CSS transform to the given object.
     * @param {jQuery} jqObj Object to apply transform to.
     * @param {string} transform The transform to apply.
     * @memberof oj.ojFilmStrip
     * @private
     */
    _applyTransform = function(jqObj, transform)
    {
      jqObj
        .css(_WEBKIT_TRANSFORM, transform)
        .css(_MS_TRANSFORM, transform)
        .css(_TRANSFORM, transform);
    },
    
    /**
     * Remove a CSS transform from the given object.
     * @param {jQuery} jqObj Object to apply transform to.
     * @memberof oj.ojFilmStrip
     * @private
     */
    _removeTransform = function(jqObj)
    {
      jqObj
        .css(_WEBKIT_TRANSFORM, _EMPTY_STRING)
        .css(_MS_TRANSFORM, _EMPTY_STRING)
        .css(_TRANSFORM, _EMPTY_STRING);
    },

    /**
     * Escape an html fragment/text.
     * @param {string} text Text to escape.
     * @returns {string} Escaped text.
     * @memberof oj.ojFilmStrip
     * @private
     */
    _escapeHtml = function(text)
    {
      //let jQuery escape the text
      var jqDiv = $("<div></div>");
      jqDiv.text(text); // @HTMLUpdateOK
      return jqDiv[0].innerHTML;
    };
    
// end static members and functions ////////////////////////////////////////////

}()); // end of FilmStrip wrapper function
(function() {
var ojFilmStripMeta = {
  "properties": {
    "arrowPlacement": {
      "type": "string",
      "enumValues": ["adjacent", "overlay"]
    },
    "arrowVisibility": {
      "type": "string",
      "enumValues": ["visible", "hidden", "hover", "auto"]
    },
    "currentItem": {
      "type": "object",
      "writeback": true,
      "properties": {
        "id": {
          "type": "string"
        },
        "index": {
          "type": "number"
        }
      }
    },
    "maxItemsPerPage": {
      "type": "number"
    },
    "orientation": {
      "type": "string",
      "enumValues": ["horizontal", "vertical"]
    },
    "translations": {
      "type": "Object",
      "properties": {
        "labelAccArrowNextPage": {
          "type": "string",
          "value": "Next Page"
        },
        "labelAccArrowPreviousPage": {
          "type": "string",
          "value": "Previous Page"
        },
        "tipArrowNextPage": {
          "type": "string",
          "value": "Next"
        },
        "tipArrowPreviousPage": {
          "type": "string",
          "value": "Previous"
        }
      }
    }
  },
  "methods": {
    "getItemsPerPage": {},
    "getPagingModel": {},
  },
  "extension": {
    _WIDGET_NAME: "ojFilmStrip"
  }
};
oj.CustomElementBridge.registerMetadata('oj-film-strip', 'baseComponent', ojFilmStripMeta);
oj.CustomElementBridge.register('oj-film-strip', {'metadata': oj.CustomElementBridge.getMetadata('oj-film-strip')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * @class oj.FlattenedTreeCellSet
 * @classdesc A CellSet represents a collection of cells.  The CellSet is an object returned by the success callback
 * of the fetchCells method on DataGridDataSource.  The FlattenedTreeCellSet is a FlattenedDataGridDataSource specific 
 * implementation of methods on CellSet. 
 * @param {number} startRow the start row index of the cell set
 * @param {number} endRow the end row index of the cell set
 * @param {number} startColumn the start column index of the cell set
 * @param {number} endColumn the end column index of the cell set
 * @param {Object} nodeSet the node set in which this cell set wraps around
 * @param {Array|null} columns the set of column keys
 * @constructor
 * @since 1.0
 * @export
 * @ojtsignore
 * @see oj.FlattenedDataGridDataSource
 */
oj.FlattenedTreeCellSet = function(startRow, endRow, startColumn, endColumn, nodeSet, columns)
{
    oj.Assert.assertArrayOrNull(columns);

    this.m_startRow = startRow;
    this.m_endRow = endRow;
    this.m_startColumn = startColumn;
    this.m_endColumn = endColumn;
    this.m_nodeSet = nodeSet;
    this.m_columns = columns;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available 
 * 2) the index specified is out of bounds. 
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} the data object for the specified index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeCellSet 
 */
oj.FlattenedTreeCellSet.prototype.getData = function(indexes)
{
    var relIndex, row, column, columnKey, rowData, returnObj, getter, setter;

    // convert to relative index
    relIndex = this._getRelIndexes(indexes);
    if (relIndex == null)
    {
        return null;
    }

    row = relIndex[0];
    column = relIndex[1];

    // make sure index are valid
    oj.Assert.assert(row < this.m_nodeSet.getStart()+this.m_nodeSet.getCount() && column < this.m_columns.length);

    columnKey = this.m_columns[column];
    rowData = this.m_nodeSet.getData(row);
    
    if (rowData != null)
    {
        returnObj = {};
        if (rowData.get)
        {
            getter = function(){ return rowData.get(columnKey);};
        }
        else
        {
            getter = function(){ return rowData[columnKey]; };
        }
        
        if (rowData.set)
        {
            setter = function(newValue){ return rowData.set(columnKey, newValue);};
        }
        else
        {
            setter = function(newValue){ rowData[columnKey] = newValue; };
        }

        Object.defineProperty(returnObj, 'data', {
            enumerable: true,
            get: getter,
            set: setter
        });
        
        return returnObj;
    }

    
    return null;
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available 
 * 2) the index specified is out of bounds. 
 * @param {Object} indexes the index of each axis in which we want to retrieve the metadata from.
 * @param {number} indexes.row the index of the row axis.
 * @param {number} indexes.column the index of the column axis.
 * @return {Object} the metadata object for the specific index.  The metadata that the DataGrid supports are: 
 *         1) keys - the key (of each axis) of the cell.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeCellSet 
 */
oj.FlattenedTreeCellSet.prototype.getMetadata = function(indexes)
{
    var relIndex, row, column, columnKey, metadata, rowKey;

    // convert to relative index
    relIndex = this._getRelIndexes(indexes);
    if (relIndex == null)
    {
        return null;
    }

    row = relIndex[0];
    column = relIndex[1];

    // make sure index are valid
    oj.Assert.assert(row < this.m_nodeSet.getStart()+this.m_nodeSet.getCount() && column < this.m_columns.length);

    columnKey = this.m_columns[column];

    metadata = this.m_nodeSet.getMetadata(row);
    rowKey = metadata['key'];

    metadata['keys'] = {"row": rowKey, "column": columnKey};

    return metadata;
};

/**
 * Helper method to validate and retrieve the relative indexes.
 * @param {Object} indexes the row and column index
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} the relative indexes
 * @private
 */
oj.FlattenedTreeCellSet.prototype._getRelIndexes = function(indexes)
{
    var row, column;

    oj.Assert.assertObject(indexes);

    if (this.m_nodeSet == null || this.m_nodeSet.length == 0)
    {
        return null;
    }

    // map to the index in nodeSet
    row = indexes['row'] - this.m_startRow + this.m_nodeSet.getStart();
    column = indexes['column'];

    // make sure index are valid
    oj.Assert.assertNumber(row, null);
    oj.Assert.assertNumber(column, null);
    oj.Assert.assert(row >= 0 && column >= 0); 

    return [row, column];
};

/**
 * Gets the start index of the result set for the specified axis.
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 *        Valid values are "row" and "column".
 * @return {number} the start of the index of the result set for the specified axis.  
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeCellSet 
 */
oj.FlattenedTreeCellSet.prototype.getStart = function(axis)
{
    if (axis === "row")
    {
        return this.m_startRow;
    }

    if (axis === "column")
    {
        return this.m_startColumn;
    }

    return 0;
};

/**
 * Gets the actual count of the result set for the specified axis. 
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 *        Valid values are "row" and "column".
 * @return {number} the actual count of the result set for the specified axis.  
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeCellSet 
 */
oj.FlattenedTreeCellSet.prototype.getCount = function(axis)
{
    if (axis === "row")
    {
        return Math.min(this.m_endRow - this.m_startRow, this.m_nodeSet.getCount());
    }
    
    if (axis === "column")
    {
        return this.m_endColumn - this.m_startColumn;
    }

    return 0;
};

/**
 * Gets the extent of a particular row/column index within the context of the cellSet.
 * Extent is defined as the number of indexes along the appropriate axis spanned by the cell.
 * If the extent extends beyond the start and end of the requested cell range the extent should be trimmed to the edge of the requested cell range and the object for {'more': {'before', 'after'}} should have the value appropriate boolean set.
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} an object containing two properties row and column. Each of those properties has two sub properties:
 *              extent: the number of absolute indexes spanned by the cell at this index
 *                      bounded by the edges of the result set for the specified axis.
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is available in the cellSet
 * @example <caption>In this example the cell spans 5 row indexes and 2 column indexes and there are more column indexes spanned by the cell that
 *              aren't included in this cellSet:</caption>
 * {
 *  'row': {'extent':5, 'more': {'before':false, 'after':false}},
 *  'column': {'extent':2, 'more': {'before':false, 'after':true}}
 * }
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeCellSet
 * @since 4.0.0
 */
oj.FlattenedTreeCellSet.prototype.getExtent = function(indexes)
{
    return {'row': {'extent':1, 'more': {'before':false, 'after':false}}, 
            'column': {'extent':1, 'more': {'before':false, 'after':false}}};
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * The DataGrid specific implementation of the FlattenedTreeDataSource class.
 * @param {Object} treeDataSource the instance of TreeDataSource to flattened
 * @param {Object=} options the options set on this data source
 * @property {(Array.<*>|string)=} options.expanded an array of the initial row keys that should be expanded, if all rows are expanded to start, specify the string 'all'
 * @property {Array.<*>=} options.columns an array of columns to return as column headers
 * @property {*} options.rowHeader the key of the attribute designated as the row header
 * @constructor
 * @since 1.0
 * @export
 * @extends oj.FlattenedTreeDataSource
 * @ojtsignore
 */
oj.FlattenedTreeDataGridDataSource = function(treeDataSource, options)
{
    oj.FlattenedTreeDataGridDataSource.superclass.constructor.call(this, treeDataSource, options);
};

// Subclass from oj.FlattenedTreeDataSource
oj.Object.createSubclass(oj.FlattenedTreeDataGridDataSource, oj.FlattenedTreeDataSource, "oj.FlattenedTreeDataGridDataSource");

/**
 * Initializes the data source.
 * @instance
 * @override
 * @protected
 */
oj.FlattenedTreeDataGridDataSource.prototype.Init = function()
{
    oj.FlattenedTreeDataGridDataSource.superclass.Init.call(this);

    this.m_columns = oj.FlattenedTreeDataGridDataSource.superclass.getOption.call(this, 'columns');
    this.m_rowHeader = oj.FlattenedTreeDataGridDataSource.superclass.getOption.call(this, 'rowHeader');
};

/**
 * Returns whether the total count returned in getCount function is an actual or an estimate.
 * @param {string} axis the axis in which we inquire whether the total count is an estimate.
 * @return {string} "exact" if the count returned in getCount function is the actual count, "estimate" if the 
 *         count returned in getCount function is an estimate.  The default value is "exact".
 * @export
 * @method
 * @instance
 */
oj.FlattenedTreeDataGridDataSource.prototype.getCountPrecision = function(axis)
{
    // always returns estimate row count to ensure high watermark scrolling is used.
    if (axis === "row")
    {
        return "estimate";
    }

    return "actual";
};

/**
 * Returns the total number of rows or columns.  If the value return is not >= 0 then it is automatically assumed
 * that the total count is unknown.
 * @param {string} axis the axis in which we inquire for the total count.
 * @return {number} the total number of rows/columns.
 * @export
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 * @method 
 */
oj.FlattenedTreeDataGridDataSource.prototype.getCount = function(axis)
{
    // always returns -1 to ensure high watermark scrolling is used.
    if (axis === "row")
    {
        return -1;
    }

    if (axis === "column")
    {
        return this.m_columns.length;
    }

    return 0;
};

/**
 * Fetch a range of headers from the data source.
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count.
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column".
 * @property {number} headerRange.start the start index of the range in which the header data are fetched.
 * @property {number} headerRange.count the size of the range in which the header data are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch headers operation is completed.
 * @property {function({startHeaderSet: Object}, {headerRange: Object}, {endHeaderSet: Object}):undefined} callbacks.success the callback to invoke when fetch headers completed successfully.
 *        The function takes three parameters: HeaderSet object representing start headers, headerRange object passed into the original fetchHeaders call,
 *        and a HeaderSet object representing the end headers along the axis.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 * @return {undefined} 
 */
oj.FlattenedTreeDataGridDataSource.prototype.fetchHeaders = function(headerRange, callbacks, callbackObjects)
{
    var axis, headerSet, columnCount;

    axis = headerRange['axis'];
    if (axis === "column")
    {
        columnCount = headerRange['start'] + headerRange['count'];
        if (columnCount > this.getCount('column'))
        {
            columnCount = this.getCount('column') - headerRange['start'];
        }
        headerSet = new oj.FlattenedTreeHeaderSet(headerRange['start'], columnCount, this.m_columns);
    }
    else if (axis === "row")
    {
        if (this.m_rowHeader != null)
        {
            // assumes that a fetch header request is immediately followed by a fetch cells request
            // avoid doing two fetch requests for the same set of data, wait until fetch cells request is available
            // before handling the header request
            // since the two fetches are converge on fetchCell, the range should always be in sync
            this.m_fetchHeaderRequest = {'range': headerRange, 'callbacks': callbacks, 'callbackObjects': callbackObjects};
            return;
        }
    }

    if (callbacks != null && callbacks['success'] != null)
    {
        // todo: get rid of callbackObjects
        if (callbackObjects == null)
        {
            callbackObjects = {};
        }	            
        callbacks['success'].call(callbackObjects['success'], headerSet, headerRange, null);
    }
};

/**
 * Fetch a range of cells from the data source.
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched. 
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.
 * @property {function({cellSet: Object}, {cellRanges: Array.<Object>}):undefined} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 * @return {undefined}
 */
oj.FlattenedTreeDataGridDataSource.prototype.fetchCells = function(cellRanges, callbacks, callbackObjects)
{
    var i, cellRange, rowStart, rowCount;

    // extract row range information needed to make the fetchRows call
    for (i=0; i<cellRanges.length; i++)
    {
        cellRange = cellRanges[i];   
        if (cellRange['axis'] == "row")
        {
            rowStart = cellRange['start'];
            rowCount = cellRange['count'];
            break;
        }
    }

    oj.FlattenedTreeDataGridDataSource.superclass.fetchRows.call(this, {'start': rowStart, 'count': rowCount}, {"success": function(nodeSet){this._handleFetchRowsSuccess(nodeSet, cellRanges, callbacks, callbackObjects, 0);}.bind(this), "error": function(status){this._handleFetchRowsError(status, {'start': rowStart, 'count': rowCount}, callbacks, callbackObjects);}.bind(this)}); 
};

/**
 * Returns the keys based on the indexes. 
 * @param {Object} indexes the index for each axis
 * @property {Object} indexes.row the index for the row axis
 * @property {Object} indexes.column the index for the column axis
 * @return {Promise.<Object>} a Promise object which upon resolution will pass in an object containing the keys for each axis,
 *                   or null if not found
 * @export
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 */
oj.FlattenedTreeDataGridDataSource.prototype.keys = function(indexes)
{
    var rowIndex, colIndex;

    rowIndex = indexes['row'];
    colIndex = indexes['column'];

    return new Promise(function(resolve, reject) {
        // if it hasn't been fetched yet or invalid column index, return null
        if (rowIndex > oj.FlattenedTreeDataGridDataSource.superclass.getFetchedRange.call(this)['end'] || colIndex > this.m_columns.length)
        {
            resolve(null);
        }
        else
        {
            resolve({"row": oj.FlattenedTreeDataGridDataSource.superclass.getKey.call(this, rowIndex), "column": this.m_columns[colIndex]});
        }
    }.bind(this));
};

/**
 * Returns the row and column index based on the keys.
 * @param {Object} keys the key for each axis
 * @param {*} keys.row the key for the row axis
 * @param {*} keys.column the key for the column axis
 * @return {Promise.<Object>} a promise object containing the index for each axis, or null if not found
 * @export
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 */
oj.FlattenedTreeDataGridDataSource.prototype.indexes = function(keys)
{
    var rowIndex, colIndex, rowKey, colKey, i;

    rowKey = keys['row'];
    colKey = keys['column'];

    return new Promise(function(resolve, reject) {
        // call helper method to find the flattened index
        rowIndex = oj.FlattenedTreeDataGridDataSource.superclass.getIndex.call(this, rowKey);
        // for column index, just search through the column keys array
        for (i=0; i<this.m_columns.length; i++)
        {
            if (this.m_columns[i] === colKey)
            {
                colIndex = i;
                break;
            }
        }

        // make sure at least one of the indexes are valid
        // (at least one, since caller might only have specified one axis)
        if (rowIndex >= 0 || colIndex >= 0)
        {
            resolve({"row": rowIndex, "column": colIndex});
        }
        else
        {
            // can't find it (should it be reject?)
            resolve(null);
        }
    }.bind(this));
};

/**
 * Performs a sort on the data source.
 * @param {Object|null} criteria the sort criteria. 
 * @property {string} criteria.axis The axis in which the sort is performed, valid values are "row", "column"
 * @property {*} criteria.key The key that identifies which header to sort
 * @property {string} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the sort operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.  
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error * @export
 * @export
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 * @return {undefined}
 */
oj.FlattenedTreeDataGridDataSource.prototype.sort = function(criteria, callbacks, callbackObjects)
{
    var dataSource = oj.FlattenedTreeDataGridDataSource.superclass.getWrappedDataSource.call(this);

    // delegates to the underlying TreeDataSource but intercept the success callback so that we can clear the cache
    return dataSource.sort(criteria, {"success": function(){this._handleSortSuccess(callbacks, callbackObjects);}.bind(this), "error": callbacks['error']});
};

/**
 * Handles sort success callback.
 * @param {Object} callbacks the original callbacks for the sort operation
 * @param {Object} callbackObjects the original callbackObjects for the sort operation
 * @private
 */
oj.FlattenedTreeDataGridDataSource.prototype._handleSortSuccess = function(callbacks, callbackObjects)
{
    // reset state
    this.refresh();

    // invoke original sort success callback
    if (callbacks['success'])
    {
        // todo: get rid of callbackObjects
        if (callbackObjects == null)
        {
            callbackObjects = {};
        }	            
        callbacks['success'].call(callbackObjects['success']);
    }
};

/**
 * Moves a row from one location to another (different position within the same parent or a completely different parent)
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  Valid values are: "before", "after".
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the move operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @export
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 * @return {undefined}
 */
oj.FlattenedTreeDataGridDataSource.prototype.move = function(rowToMove, referenceRow, position, callbacks)
{
    var dataSource = oj.FlattenedTreeDataGridDataSource.superclass.getWrappedDataSource.call(this);

    // delegates to the underlying TreeDataSource.  TreeDataSource is responsible for firing the appropriate model change
    // event so that the FlattenedTreeDataSource state are updated correctly.
    dataSource.move(rowToMove, referenceRow, position, callbacks);
};

/**
 * Determines whether this DataGridDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none", "row", "column".  
 *         Returns null if the feature is not recognized.
 * @export
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeDataGridDataSource 
 */
oj.FlattenedTreeDataGridDataSource.prototype.getCapability = function(feature)
{
    var dataSource = oj.FlattenedTreeDataGridDataSource.superclass.getWrappedDataSource.call(this);

    // just delegates to the underlying TreeDataSource
    if (dataSource.getCapability(feature) === 'default')
    {
        return 'column';
    }
    else
    {
        return 'none';
    }
};

/**
 * A hook for FlattenedTreeDataSource to inject additional metadata into the NodeSet
 * @param {*} key the row key identifying the row
 * @param {Object} metadata the existing metadata to inject into
 * @protected
 */
oj.FlattenedTreeDataGridDataSource.prototype.insertMetadata = function(key, metadata)
{
    // just call super
    oj.FlattenedTreeDataGridDataSource.superclass.insertMetadata.call(this, key, metadata);
};

/**
 * Callback method to handle success callback for fetchRows operation on FlattenedTreeDataSource.
 * @param {Object} nodeSet the result node set from the fetchRows called.
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched. 
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function(Object)} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.  
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @private
 */
oj.FlattenedTreeDataGridDataSource.prototype._handleFetchRowsSuccess = function(nodeSet, cellRanges, callbacks, callbackObjects)
{
    var i, cellRange, rowStart, rowCount, columnStart, columnCount, headerRange, cellSet;

    // extract range information
    for (i=0; i<cellRanges.length; i++)
    {
        cellRange = cellRanges[i];   
        if (cellRange['axis'] == "row")
        {
            rowStart = cellRange['start'];
            rowCount = cellRange['count'];
            
        }
        else if (cellRange['axis'] == "column")
        {
            columnStart = cellRange['start'];
            columnCount = cellRange['count'];
            if (columnStart + columnCount > this.getCount('column'))
            {
                columnCount = this.getCount('column') - columnStart;
            }
        }
    }

    // checks whether there is an outstanding fetch header request with the same range
    if (this.m_fetchHeaderRequest)
    {
        headerRange = this.m_fetchHeaderRequest['range'];
        if (headerRange['start'] == rowStart && headerRange['count'] == rowCount)
        {
            // handle row header request
            this._handleRowHeaderFetchSuccess(nodeSet, headerRange, this.m_fetchHeaderRequest['callbacks'], this.m_fetchHeaderRequest['callbackObjects']);
			// do not clear unless it is the correct range
            this.m_fetchHeaderRequest = null;
        }
    }

    // create wrapper
    cellSet = new oj.FlattenedTreeCellSet(rowStart, rowStart+rowCount, columnStart, columnStart+columnCount, nodeSet, this.m_columns);
    // invoke success callback
    if (callbacks['success'])
    {
        // todo: get rid of callbackObjects
        if (callbackObjects == null)
        {
            callbackObjects = {};
        }            
        callbacks["success"].call(callbackObjects['success'], cellSet, cellRanges);
    }
};

/**
 * Callback method to handle error callback for fetchRows operation on FlattenedTreeDataSource.
 * @param {Object} status the error status.
 * @param {Object} range Information about the row range.  
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function(Object)} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.  
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @private
 */
oj.FlattenedTreeDataGridDataSource.prototype._handleFetchRowsError = function(status, range, callbacks, callbackObjects)
{
    var headerRange, headerCallbacks, headerCallbackObjects;

    // checks whether there is an outstanding fetch header request with the same range
    if (this.m_fetchHeaderRequest)
    {
        headerRange = this.m_fetchHeaderRequest['range'];
        if (headerRange['start'] == range['start'] && headerRange['count'] == range['count'])
        {
            // invoke error callback on fetch header
            headerCallbacks = this.m_fetchHeaderRequest['callbacks'];
            headerCallbackObjects = this.m_fetchHeaderRequest['callbackObjects'];

            if (headerCallbacks['error'])
            {
                // todo: get rid of callbackObjects
                if (headerCallbackObjects == null)
                {
                    headerCallbackObjects = {};
                }            
                headerCallbacks['error'].call(headerCallbackObjects['error'], status);
            }
        }
        this.m_fetchHeaderRequest = null;
    }

    // invoke error callback
    if (callbacks['error'])
    {
        // todo: get rid of callbackObjects
        if (callbackObjects == null)
        {
            callbackObjects = {};
        }            
        callbacks["success"].call(callbackObjects['error'], status);
    }
};

/**
 * Handles header fetch success request based on results from fetch cell operation.
 * @param {Object} nodeSet the result node set from the fetchRows called.
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count.
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column".
 * @property {number} headerRange.start the start index of the range in which the header data are fetched.
 * @property {number} headerRange.count the size of the range in which the header data are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch headers operation is completed.  The valid callback
 *        types are "success" and "error".
 * @property {function(Object)} callbacks.success the callback to invoke when fetch headers completed successfully.
 * @property {function({status: Object})} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.  This is optional.  
 *        You can specify the callback object for each callbacks using the "success" and "error" keys.
 * @private
 */
oj.FlattenedTreeDataGridDataSource.prototype._handleRowHeaderFetchSuccess = function(nodeSet, headerRange, callbacks, callbackObjects)
{
    // create wrapper
    var headerSet = new oj.FlattenedTreeHeaderSet(headerRange['start'], headerRange['start']+headerRange['count'], this.m_columns, nodeSet, this.m_rowHeader);
    // invoke success callback
    if (callbacks['success'])
    {
        // todo: get rid of callbackObjects
        if (callbackObjects == null)
        {
            callbackObjects = {};
        }            
        callbacks['success'].call(callbackObjects['success'], headerSet, headerRange, null);
    }
};

/**
 * Implementation of abstract method to insert a set of rows into the DataGrid
 * @param {number} insertAtIndex the flattened index of the node where the rows are inserted.
 * @param {*} insertAtRowKey the key of the node where the rows are inserted (the parent key)
 * @param {Object} nodeSet the node set containing data/metadata of inserted rows
 * @protected
 */
oj.FlattenedTreeDataGridDataSource.prototype.insertRows = function(insertAtIndex, insertAtRowKey, nodeSet)
{
    var cellSet, headerSet, event, i, keys, indexes;

    headerSet = null;
    if(this.m_rowHeader)
    {
        headerSet = new oj.FlattenedTreeHeaderSet(insertAtIndex, insertAtIndex+nodeSet.getCount(), this.m_columns, nodeSet, this.m_rowHeader);
    }
    // create a CellSet that wraps around a RowSet 
    cellSet = new oj.FlattenedTreeCellSet(insertAtIndex, insertAtIndex+nodeSet.getCount(), 0, this.m_columns.length, nodeSet, this.m_columns);

    keys = [];
    indexes = [];    
    for (i = 0; i < nodeSet.getCount(); i++)
    {
        keys.push({"row": this._getEntry(insertAtIndex + i)['key']});
        indexes.push({"row": insertAtIndex + i, "column": -1});               
    }    

    // construct model insert event with a set of rows to insert
    event = {};
    event['source'] = this;
    event['operation'] = 'insert';
    event['result'] = cellSet;
    if(headerSet)
    {
        event['header'] = headerSet;
    }
    event['keys'] = keys;
    event['indexes'] = indexes;
    
    oj.FlattenedTreeDataGridDataSource.superclass.handleEvent.call(this, "change", event);
};

/**
 * Implementation of abstract method to remove the specified rows in the DataGrid
 * @param {Array.<*>} rowKeys an array of keys of the rows to be remove.
 * @protected
 */
oj.FlattenedTreeDataGridDataSource.prototype.removeRows = function(rowKeys)
{
    var keys, i, event, indexes;
    
    // extract the keys
    keys = [];
    indexes = [];
    for (i=0; i<rowKeys.length; i++)
    {
        keys.push({"row": rowKeys[i]['key']});
        indexes.push({"row": rowKeys[i]['index'], "column": -1});        
    }

    // construct model delete event with a set of row keys to delete
    event = {};
    event['source'] = this;
    event['operation'] = 'delete';
    event['keys'] = keys;
    event['indexes'] = indexes;

    oj.FlattenedTreeDataGridDataSource.superclass.handleEvent.call(this, "change", event);
};

/**
 * Handles the case when the maximum number of rows have been reached
 * @param {Object} range the range of the fetch request that cause the max count to be reached
 * @property {number} range.start the start index of the range
 * @property {number} range.count the count of the range
 * @protected
 */
oj.FlattenedTreeDataGridDataSource.prototype.handleMaxCountReached = function(range, callbacks)
{
    var empty = new oj.EmptyNodeSet(null, range['start']);
    callbacks["success"].call(null, empty);
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * @class oj.FlattenedTreeHeaderSet
 * @classdesc A HeaderSet represents a collection of headers.  The HeaderSet is an object returned by the success callback
 * of the fetchHeaders method on DataGridDataSource.  This is an flattened tree specific implementation of the HeaderSet.
 * @param {number} start the start index of header set.
 * @param {number} end the end index of the header set.
 * @param {Array} headers the array of headers.  Required for column headers.
 * @param {Object=} nodeSet the node set containing data about the row header.  Required for row headers.
 * @param {string=} rowHeader the id of the row header column.  Required for row headers.
 * @constructor
 * @since 1.0
 * @export
 * @ojtsignore
 * @see oj.FlattenedTreeDataGridDataSource
 */
oj.FlattenedTreeHeaderSet = function(start, end, headers, nodeSet, rowHeader)
{
    oj.Assert.assertArrayOrNull(headers);

    this.m_start = start;
    this.m_end = end;
    this.m_headers = headers;
    this.m_nodeSet = nodeSet;
    this.m_rowHeader = rowHeader;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the absolute index of the header in which we want to retrieve the header from.  
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {*} the data object for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeHeaderSet 
 */
oj.FlattenedTreeHeaderSet.prototype.getData = function(index, level)
{
    var rowData;

    // make sure index are valid
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds'); 

    // row or column header
    if (this.m_rowHeader != null && this.m_nodeSet != null)
    {
        rowData = this.m_nodeSet.getData(index - this.m_start + this.m_nodeSet.getStart());
        if (rowData != null)
        {
            if (rowData.get)
            {
                return rowData.get(this.m_rowHeader);
            }
            else
            {
                return rowData[this.m_rowHeader];
            }
        }
        else
        {
            return null;
        }
    }
    else
    {
        return this.m_headers[index];
    }
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source can optionally return are:
 *  1) sortDirection - the initial sort direction of the header.  Valid values are "ascending" and "descending".
 *  2) key - the key of the row/column header.
 * @param {number} index the absolute index of the header in which we want to retrieve the metadata from.  
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} the metadata object for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeHeaderSet 
 */
oj.FlattenedTreeHeaderSet.prototype.getMetadata = function(index, level)
{
    var data;
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds'); 
    
    if (this.m_rowHeader != null && this.m_nodeSet != null)
    {
        return this.m_nodeSet.getMetadata(index - this.m_start + this.m_nodeSet.getStart());
    }
    else
    {
        data = this.getData(index);
        return {'key': data};
    }
};

/**
 * Gets the actual count of the result set. The total indexes spanned along the innermost header.
 * @return {number} the actual count of the result set.  
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeHeaderSet 
 */
oj.FlattenedTreeHeaderSet.prototype.getCount = function()
{
    if (this.m_rowHeader != null && this.m_nodeSet != null)
    {
        return Math.min(this.m_nodeSet.getCount(), this.m_end - this.m_start);
    }
    else
    {
        return Math.max(0, this.m_end - this.m_start);
    }
};

/**
 * Gets the actual number of levels of the result set for the specified axis. The levels
 * are the counted from the outermost header indexed at 0, and moving inwards toward the 
 * databody would increment the level by 1.
 * @return {number} the number of levels of the result set
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeHeaderSet 
 */
oj.FlattenedTreeHeaderSet.prototype.getLevelCount = function()
{
    if (this.getCount() > 0)
    {
        return 1;
    }
    return 0;
};
 
/**
 * Gets the extent of an index on a particular level within the context of the headerSet. Extent is defined as the 
 * number of indexes covered by the header. If the extent extends beyond the start and end of the requested 
 * range the extent should be trimmed to the edge of the requested range and the object for {'more': {before, after}}
 * should have the value appropriate boolean set. For innermost headers the extent will always be 1.
 * @param {number} index the absolute index along the innermost header of the extent to get, 0 is the first header in the data source
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} an object containing two values
 *              extent: the number of absolute indexes spanned by the header at this index 
 *                      bounded by the edges of the result set for the specified axis. 
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is in the headerSet
 * @example <caption>In this example the header spans 5 indexes and there are more indexes to cover after the request that
 *              aren't included in this headerSet:</caption>                     
 * {'extent':5, 'more': {'before':false, 'after':true}}
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeHeaderSet
 */
oj.FlattenedTreeHeaderSet.prototype.getExtent = function(index, level)
{
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds'); 
    
    return {'extent': 1, 'more':{'before': false, 'after':false}};
};
 
/**
 * Gets the depth of an index starting at a particular level. The depth is the number 
 * of levels spanned by the header.
 * @param {number} index the absolute index of the depth to get
 * @param {number=} level the level of the header, 0 is the outermost header
 * @return {number} the number of levels spanned by the header at the specified position
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.FlattenedTreeHeaderSet 
 */
oj.FlattenedTreeHeaderSet.prototype.getDepth = function(index, level)
{
    oj.Assert.assert(index <= this.m_end && index >= this.m_start, 'index out of bounds'); 
    oj.Assert.assert(level == null || level == 0, 'level out of bounds'); 
    
    return 1;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @export
 * @class oj.FlattenedTreeTableDataSource
 * @extends oj.TableDataSource
 * @ojtsignore
 * @classdesc Object representing data used by Table with RowExpander.<br><br>
 *            See the <a href="../jetCookbook.html?component=rowExpander&demo=tableRowExpander">Row Expander - Table</a> demo for an example.<br><br>
 *            Refer to {@link oj.TableDataSource} for other data sources that represent tabular data.
 * @param {Object} data
 * @param {Object|null} options Options for the FlattenedTreeTableDataSource
 * @param {string} options.startFetch Control whether to start initial fetch when the TableDataSource is bound to a component.  Valid values are:<br><br>
 *                                    <b>"enabled"</b> (default) - Start initial fetch automatically when the TableDataSource is bound to a component.<br>
 *                                    <b>"disabled"</b> - Do not start initial fetch automatically.  Application will call the <a href="#fetch">fetch()</a> method to
 *                                                        start the first fetch.
 * @constructor
 * @since 1.0
 */
oj.FlattenedTreeTableDataSource = function(data, options)
{
  // Initialize
  options = options || {};

  if (!(data instanceof oj.FlattenedTreeDataSource))
  {
    var errSummary = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_SUMMARY'];
    var errDetail = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_DETAIL'];
    throw new Error(errSummary + '\n' + errDetail);
  }

  this._data = data;
  this._eventHandlers = [];
  this._startIndex = 0;
  this._nodeSetList = [];
  this._hasMore = true;
  
  // override the fetchSize with -1 if not already specified.
  if (this._data.getOption('fetchSize') == null)
  {
    this._data.getFetchSize = function()
    {
      return -1;
    };
  }
  var self = this;
  // override the insert/removeRows function
  this._data.insertRows = function(insertAtIndex, insertAtKey, nodeSet)
  {
    var i, j, row, rowIdx, rowKey;
    var rowArray = [];
    var keyArray = [];
    var indexArray = [];
    for (i = 0; i < nodeSet.getCount(); i++)
    {
      row = nodeSet.getData(i);
      rowKey = nodeSet.getMetadata(i)['key'];
      rowIdx = insertAtIndex + i;
      self._nodeSetList.splice(rowIdx, 0, {});
      self._nodeSetList[rowIdx]['nodeSet'] = nodeSet;
      self._nodeSetList[rowIdx]['startIndex'] = insertAtIndex;
      
      // update the startIndex of the shifted rows
      for (j = rowIdx + 1; j < self._nodeSetList.length; j++)
      {
        self._nodeSetList[j]['startIndex'] = self._nodeSetList[j]['startIndex'] + 1;
      }
      rowArray.push(self._wrapWritableValue(row));
      keyArray.push(rowKey);
      indexArray.push(rowIdx);
      self._rows['data'].splice(rowIdx, 0, row);
      self._rows['keys'].splice(rowIdx, 0, rowKey);
      self._rows['indexes'].splice(rowIdx, 0, rowIdx);
    }
    self._realignRowIndices();
    self._hasMore = true;
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['ADD'], {'data': rowArray, 'keys': keyArray, 'indexes': indexArray});
  };
  this._data.removeRows = function(rowKeys)
  {
    var i, j, row, rowIdx;
    var rowArray = [];
    var keyArray = [];
    var indexArray = [];
    for (i = rowKeys.length - 1; i >= 0; i--)
    {
      rowIdx = rowKeys[i]['index'];
      // just create a dummy row for deletion
      rowArray.push('');
      keyArray.push(rowKeys[i]['key']);
      indexArray.push(rowIdx);
      self._nodeSetList.splice(rowIdx, 1);
      // update the startIndex of the shifted rows
      for (j = rowIdx; j < self._nodeSetList.length; j++)
      {
        self._nodeSetList[j]['startIndex'] = self._nodeSetList[j]['startIndex'] - 1;
      }
      self._rows['data'].splice(rowIdx, 1);
      self._rows['keys'].splice(rowIdx, 1);
      self._rows['indexes'].splice(rowIdx, 1);
    }
    // The index array must be sorted ascending
    function sortNumber(a, b) {
      return a - b;
    }
    indexArray = indexArray.sort(sortNumber);
    self._realignRowIndices();
    self._hasMore = true;
    oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['REMOVE'], {'data': rowArray, 'keys': keyArray, 'indexes': indexArray});
  };
  
  this.Init();
  
  if ((options != null && (options['startFetch'] == 'enabled' || options['startFetch'] == null))
    || options == null)
  {
    this._startFetchEnabled = true;
  }
};

// Subclass from oj.DataSource 
oj.Object.createSubclass(oj.FlattenedTreeTableDataSource, oj.TableDataSource, "oj.FlattenedTreeTableDataSource");

/**
 * Initializes the instance.
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 * @override
 * @protected
 */
oj.FlattenedTreeTableDataSource.prototype.Init = function()
{
  oj.FlattenedTreeTableDataSource.superclass.Init.call(this);
};

/**
 * Determines whether this FlattenedTreeTableDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none".  
 *         Returns null if the feature is not recognized.
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.getCapability = function(feature)
{
    return 'full';
};

/**
 * Retrieves the underlying DataSource.
 * @return {Object} the underlying oj.DataSource.
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.getWrappedDataSource = function()
{
    return this._data;
};

/**
 * Fetch the row data.
 * @param {Object=} options Options to control fetch
 * @param {number} options.startIndex The index at which to start fetching records.
 * @param {boolean} options.silent If set, do not fire a sync event.
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects, an array of ids, and the startIndex triggering done when complete.<p>
 *         The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>startIndex</b></td><td>The startIndex for the returned set of rows</td></tr>
 * </tbody>
 * </table>  
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.fetch = function(options)
{
  options = options || {};
  var self = this;
  var fetchType = options['fetchType'];
  
  if (fetchType == 'init' && !this._startFetchEnabled)
  {
      return Promise.resolve();
  }

  return self._fetchInternal(options);
};




/**** start delegated functions ****/

/**
 * Return the row data found at the given index.
 * 
 * @param {number} index Index for which to return the row data. 
 * @param {Object=} options Options to control the at.
 * @return {Promise} Promise resolves to a compound object which has the structure below. If the index is out of range, Promise resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.at = function(index, options)
{
  var row;

  if (index < 0 || index >= this._rows['data'].length)
  {
    row = null;
  }
  else
  {
    row = {'data': this._rows['data'][index], 'index': index, 'key': this._rows['keys'][index]};
  }

  return new Promise(function(resolve, reject) {
    resolve(row);
  });
};

/**
 * Collapse the specified row.
 * @param {Object} rowKey the key of the row to collapse
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.collapse = function(rowKey)
{
  this._data.collapse(rowKey);
};


/**
 * Expand the specified row.
 * @param {Object} rowKey the key of the row to expand
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.expand = function(rowKey)
{
  this._data.expand(rowKey);
};

/**
 * Return the first row data whose id value is the given id
 * @param {string} id ID for which to return the row data, if found. 
 * @param {Object=} options Options to control the get.
 * @return {Promise} Promise which resolves to a compound object which has the structure below where the id matches the given id. If none are found, resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.get = function(id, options)
{
  // only works for expanded keys
  var rowIdx = this._data.getIndex(Object(id));
  var row = this._rows['data'][rowIdx];
  var wrappedRow = this._wrapWritableValue(row);
  var result = {'data': wrappedRow, 'key': id, 'index': rowIdx};
  
  return Promise.resolve(result);
};

/**
 * Attach an event handler to the datasource
 * 
 * @param {string} eventType eventType supported by the datasource
 * @param {function(Object)} eventHandler event handler function
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.on = function(eventType, eventHandler)
{
  if (eventType == 'expand' ||
      eventType == 'collapse')
  {
    // expand/collapse listeners should be passed through to the FlattenedTreeDatasource
    this._data.on(eventType, eventHandler);
  }
  else
  {
    oj.FlattenedTreeTableDataSource.superclass.on.call(this, eventType, eventHandler);
  }
};

/**
 * Detach an event handler from the datasource
 * 
 * @param {string} eventType eventType supported by the datasource
 * @param {function(Object)} eventHandler event handler function
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.off = function(eventType, eventHandler)
{
  if (eventType == 'expand' ||
      eventType == 'collapse')
  {
    // expand/collapse listeners should be passed through to the FlattenedTreeDatasource
    this._data.off(eventType, eventHandler);
  }
  else
  {
    oj.FlattenedTreeTableDataSource.superclass.off.call(this, eventType, eventHandler);
  }
};

/**
 * Performs a sort on the data source.
 * @param {Object} criteria the sort criteria.
 * @param {Object} criteria.key The key that identifies which field to sort
 * @param {string} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @return {Promise} promise object triggering done when complete.
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.sort = function(criteria)
{
  if (criteria == null)
  {
    criteria = this['sortCriteria'];
  }
  else
  {
    this['sortCriteria'] = criteria;
  }
  
  var self = this;
  criteria['axis'] = 'column';
  return new Promise(function(resolve, reject) 
  {
    self._data.getWrappedDataSource().sort(criteria,
      {"success": function(nodeSet)
        {
          setTimeout(function()
          {
            self._data.refresh();
            self._rows = null;
            var result = {'header': criteria['key'], 'direction': criteria['direction']};
            // We need to dispatch a RESET event because after a sort the data needs to be completely
            // refetched and we need to start from page 1 again if paging.
            oj.TableDataSource.superclass.handleEvent.call(self, oj.TableDataSource.EventType['RESET'], null);
            resolve(result);
          }, 0);
        }.bind(this),
        "error": function(status)
        {
          //this._handleFetchRowsError(status, {'start': rowStart, 'count': rowCount}, callbacks, callbackObjects);
          reject(status);
        }.bind(this)
      });
  });
};

/**
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance
 */
oj.FlattenedTreeTableDataSource.prototype.totalSize = function()
{
  if (!this._hasMore)
  {
    return this._rows['data'].length;
  }
  return -1;
};

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @export
 * @expose
 * @memberof oj.FlattenedTreeTableDataSource
 * @instance 
 */
oj.FlattenedTreeTableDataSource.prototype.totalSizeConfidence = function()
{ 
  if (!this._hasMore)
  {
    return 'actual';
  }
  return "unknown";
};

/**** end delegated functions ****/

oj.FlattenedTreeTableDataSource.prototype._getMetadata = function(index)
{
  var nodeSetStart = this._nodeSetList[index]['nodeSet'].getStart();
  return this._nodeSetList[index]['nodeSet'].getMetadata(nodeSetStart + index - this._nodeSetList[index]['startIndex']);
}

oj.FlattenedTreeTableDataSource.prototype._fetchInternal = function(options)
{
  options = options || {};
  this._startFetch(options);
  this._startIndex = options['startIndex'] == null ? this._startIndex : options['startIndex'];
  
  // if pageSize is not specified then fetch everything
  var rangeCount = Number.MAX_VALUE;
  
  this._pageSize = options['pageSize'] == null ? this._pageSize : options['pageSize'];
  if (this._pageSize != null)
  {
    rangeCount = this._pageSize;
  }
  
  var startIndex = this._startIndex;
  if (this._rows != null)
  {
    if (this._pageSize != null)
    {
      var endIndex = this._rows['data'].length - 1;

      if (this._startIndex + this._pageSize - 1 <= endIndex)
      {
        endIndex = oj.FlattenedTreeTableDataSource._getEndIndex(this._rows, this._startIndex, this._pageSize);
        var rowArray = [];
        var keyArray = [];
        var i, key;
        for (i = this._startIndex; i <= endIndex; i++)
        {
          key = this._rows['keys'][i];
          rowArray[i - this._startIndex] = this._wrapWritableValue(this._rows['data'][i]);
          keyArray[i - this._startIndex] = key;
        }
        var result = {'data': rowArray, 'keys': keyArray, 'startIndex': this._startIndex};
        this._endFetch(options, result, null);
        return Promise.resolve(result);
      }
      else if (this._startIndex <= endIndex)
      {
        startIndex = endIndex + 1;
      }
    }
    else
    {
      // we are fetching everything again so clear out the existing cached rows
      this._data.refresh();
      this._rows = null;
    }
  }
  else
  {
    // if we don't have any cached rows then we always need to fetch from the start
    startIndex = 0;
  }
  
  var rangeOption = {'start': startIndex, 'count':  rangeCount};
  var self = this;
  return new Promise(function(resolve, reject) 
  {
    self._data.fetchRows(rangeOption,
    {
      "success": function(nodeSet)
      {
        self._handleFetchRowsSuccess(nodeSet);
        options['refresh'] = true;
        var endIndex = oj.FlattenedTreeTableDataSource._getEndIndex(self._rows, self._startIndex, self._pageSize);
        var rowArray = [];
        var keyArray = [];
        var i, key;
        for (i = self._startIndex; i <= endIndex; i++)
        {
          key = self._rows['keys'][i];
          rowArray[i - self._startIndex] = self._wrapWritableValue(self._rows['data'][i]);
          keyArray[i - self._startIndex] = key;
        }
        // if there are results then we potentially have more
        if (rowArray.length > 0)
        {
          if (self._pageSize != null &&
              rowArray.length < self._pageSize)
          {
            // if we are paged and we return less than the page size
            // then we don't have any more rows
            self._hasMore = false;
          }
          else
          {
            self._hasMore = true;
          }
        }
        else
        {
          self._hasMore = false;
        }
        var result = {'data': rowArray, 'keys': keyArray, 'startIndex': self._startIndex};
        self._endFetch(options, result, null);
        resolve(result);
      }.bind(this),
      "error": function(error)
      {
        self._endFetch(options, null, error);
        reject(error);
      }.bind(this)
    });
  });
};

oj.FlattenedTreeTableDataSource.prototype._handleFetchRowsSuccess = function(nodeSet)
{
  var startIndex = nodeSet.getStart();
  var i, rowIdx;
  for (i = 0; i < nodeSet.getCount(); i++)
  {
    rowIdx = startIndex + i;
    this._nodeSetList[rowIdx] = {};
    this._nodeSetList[rowIdx]['nodeSet'] = nodeSet;
    this._nodeSetList[rowIdx]['startIndex'] = startIndex;
  }

  if (!this._rows)
  {
    this._rows = {};
    this._rows['data'] = [];
    this._rows['keys'] = [];
    this._rows['indexes'] = [];
  }
  oj.FlattenedTreeTableDataSource._getRowArray(nodeSet, this, this._rows, startIndex);
};

/**
 * Indicate starting fetch
 * @param {Object} options
 * @private
 */
oj.FlattenedTreeTableDataSource.prototype._startFetch = function(options)
{
  if (!options['silent'])
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['REQUEST'], {'startIndex' : options['startIndex']});
  }
};

/**
 * Indicate ending fetch
 * @param {Object} options
 * @param {Object} result Result object
 * @param {Object} error Error
 * @private
 */
oj.FlattenedTreeTableDataSource.prototype._endFetch = function(options, result, error)
{
  if (error != null)
  {
    oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['ERROR'], error);
  }
  else
  {
    if (!options['silent'])
    {
      oj.TableDataSource.superclass.handleEvent.call(this, oj.TableDataSource.EventType['SYNC'],  result);
    }
  }
};

oj.FlattenedTreeTableDataSource._getEndIndex = function(rows, startIndex, pageSize)
{
  var endIndex = rows['data'].length - 1;
  
  if (pageSize > 0)
  {
    endIndex = startIndex + pageSize - 1;
    endIndex = endIndex > rows['data'].length - 1 ? rows['data'].length - 1 : endIndex;
  }
  
  return endIndex;
};

oj.FlattenedTreeTableDataSource._getRowArray = function(nodeSet, rowSet, rows, startIndex)
{
  var i;
  for (i = 0; i < nodeSet.getCount(); i++)
  {
    var row = nodeSet.getData(nodeSet.getStart() + i);
    rows['data'][startIndex + i] = row;
    rows['keys'][startIndex + i] = nodeSet.getMetadata(nodeSet.getStart() + i)['key'];
    rows['indexes'][startIndex + i] = startIndex + i;
  }
};

// Realign all the indices of the rows (after sort for example)
oj.FlattenedTreeTableDataSource.prototype._realignRowIndices = function()
{
  for (var i = 0; i < this._rows['data'].length; i++)
  {
    this._rows['indexes'][i] = i;
  }
};

oj.FlattenedTreeTableDataSource.prototype._wrapWritableValue = function(m)
{
  var clonedRow = $.extend(true, {}, m);
  var i, props = Object.keys(m);
  
  for (i = 0; i < props.length; i++)
  {
    oj.FlattenedTreeTableDataSource._defineProperty(clonedRow, m, props[i]);
  }
  
  return clonedRow;
};

oj.FlattenedTreeTableDataSource._defineProperty = function(row, m, prop)
{
  Object.defineProperty(row, prop,
    {
      get: function()
      {
        return m[prop];
      },
      set: function(newValue)
      {
        m[prop] = newValue;
      },
      enumerable: true
    });
};
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojFormLayout
 * @since 4.1.0
 * @ojshortdesc Used to group child input controls in an organized layout that can be optimized for multiple display sizes.
 * @ojstatus preview
 * @ojsignature {target: "Type", value:"class ojFormLayout extends JetElement<ojFormLayoutSettableProperties>"}
 *
 * @classdesc
 * <h3 id="optionOverview-section">
 *   JET FormLayout
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#optionOverview-section"></a>
 * </h3>
 * <p>The oj-form-layout element is used to group custom element children in an organized layout that
 * can be optimized for multiple display sizes.  Legal child components are: oj-checkboxset, oj-color-palette
 * oj-color-spectrum, oj-input-date, oj-input-date-time, oj-input-time, oj-input-number, oj-input-text,
 * oj-text-area, oj-input-password, oj-combobox-one, oj-combobox-many, oj-radioset, oj-select-one, oj-select-many,
 * oj-slider, oj-switch
 * 
 * <p>For example:
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-form-layout max-columns='2' label-edge='start' label-width="50%">
 *   &lt;oj-input-text id="inputcontrol" required value="text" label-hint="input 1">&lt;/oj-input-text>
 *   &lt;oj-text-area id="textareacontrol" value='text' rows="6" label-hint="textarea">&lt;/oj-text-area>
 *   &lt;oj-input-text id="inputcontrol2" value="text" label-hint="input 2">&lt;/oj-input-text>
 *   &lt;oj-input-text id="inputcontrol3" value="text" label-hint="input 3 longer label">&lt;/oj-input-text>
 * &lt;/oj-form-layout> 
 * </code></pre>
 *
 * <p>The oj-form-layout element currently supports custom element children that support the label-hint 
 * and help-hints attributes (oj-input-text, oj-text-area, etc.).
 * For each custom element child with label-hint, FormLayout will generate an oj-label element and pair
 * them together in the layout.
 * </p>
 */

/**
 * @member
 * @name direction
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * @type {string}
 * @default <code class="prettyprint">"column"</code>
 * @ojvalue {string} "column" Components are laid out in columns
 * @ojvalue {string} "row" Components are laid out in rows
 * @desc Specifies the layout direction of the form layout children.
 *
 * @example <caption>Initialize the oj-form-layout with the <code class="prettyprint">direction</code> attribute specified:</caption>
 * &lt;oj-form-layout direction="row">
 *   &lt;oj-input-text id="inputcontrol" required value="text" label-hint="input 1">&lt;/oj-input-text>
 *   &lt;oj-text-area id="textareacontrol" value='text' rows="6" label-hint="textarea">&lt;/oj-text-area>
 * &lt;/oj-form-layout>
 *
 * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
 * // getter
 * var dir = myFormLayout.direction;
 *
 * // setter
 * myFormLayout.direction = "column";
 */

/**
 * @member
 * @name labelEdge
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * @type {string}
 * @default "top"
 * @ojvalue {string} "start" Label is inline with the start of its editable value component
 * @ojvalue {string} "top" Label is on top of its editable value component
 * @desc Specifies how the label is aligned with its editable value component.
 *
 * @example <caption>Initialize the oj-form-layout with the <code class="prettyprint">label-edge</code> attribute specified:</caption>
 * &lt;oj-form-layout label-edge="top">
 *   &lt;oj-input-text id="inputcontrol" required value="text" label-hint="input 1">&lt;/oj-input-text>
 *   &lt;oj-text-area id="textareacontrol" value='text' rows="6" label-hint="textarea">&lt;/oj-text-area>
 * &lt;/oj-form-layout>
 *
 * @example <caption>Get or set the <code class="prettyprint">labelEdge</code> property after initialization:</caption>
 * // getter
 * var edge = myFormLayout.labelEdge;
 *
 * // setter
 * myFormLayout.labelEdge = 'start';
 */

/**
 * @member
 * @name labelWidth
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * @type {string}
 * @default "33%"
 * @desc Specifies the label width.
 * <p>This specifies the width of the oj-label elements.  This can be any legal <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/width">CSS width</a>.</p>
 *
 * @example <caption>Initialize the oj-form-layout with the <code class="prettyprint">label-width</code> attribute specified:</caption>
 * &lt;oj-form-layout label-width="50%">
 *   &lt;oj-input-text id="inputcontrol" required value="text" label-hint="input 1">&lt;/oj-input-text>
 *   &lt;oj-text-area id="textareacontrol" value='text' rows="6" label-hint="textarea">&lt;/oj-text-area>
 * &lt;/oj-form-layout>
 *
 * @example <caption>Get or set the <code class="prettyprint">labelWidth</code> property after initialization:</caption>
 * // getter
 * var width = myFormLayout.labelWidth;
 *
 * // setter
 * myFormLayout.labelWidth = '60px';
 */

/**
 * @member
 * @name labelWrapping
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * @type {string}
 * @default <code class="prettyprint">"wrap"</code>
 * @ojvalue {string} "truncate" Label will trunctate if needed
 * @ojvalue {string} "wrap" Label will wrap if needed
 * @desc Specifies if the label text should wrap or truncate.
 *
 * @example <caption>Initialize the oj-form-layout with the <code class="prettyprint">label-wrapping</code> attribute specified:</caption>
 * &lt;oj-form-layout label-wrapping="truncate">
 *   &lt;oj-input-text id="inputcontrol" required value="text" label-hint="input 1">&lt;/oj-input-text>
 *   &lt;oj-text-area id="textareacontrol" value='text' rows="6" label-hint="textarea">&lt;/oj-text-area>
 * &lt;/oj-form-layout>
 *
 * @example <caption>Get or set the <code class="prettyprint">labelWrapping</code> property after initialization:</caption>
 * // getter
 * var wrap = myFormLayout.labelWrapping;
 *
 * // setter
 * myFormLayout.labelWrapping = 'wrap';
 */

/**
 * @member
 * @name maxColumns
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * @type {number}
 * @default 1
 * @desc Specifies the maximum number of columns.
 *
 * @example <caption>Initialize the oj-form-layout with the <code class="prettyprint">max-columns</code> attribute specified:</caption>
 * &lt;oj-form-layout max-columns="2">
 *   &lt;oj-input-text id="inputcontrol" required value="text" label-hint="input 1">&lt;/oj-input-text>
 *   &lt;oj-text-area id="textareacontrol" value='text' rows="6" label-hint="textarea">&lt;/oj-text-area>
 * &lt;/oj-form-layout>
 *
 * @example <caption>Get or set the <code class="prettyprint">maxColumns</code> property after initialization:</caption>
 * // getter
 * var maxCol = myFormLayout.maxColumns;
 *
 * // setter
 * myFormLayout.maxColumns = 2;
 */

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 * 
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 * 
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * 
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */ 
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * 
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * 
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */ 
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 * @expose
 * @memberof oj.ojFormLayout
 * @instance
 * 
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */ 
/**
 * Refreshes the visual state of the component.
 * 
 * @function refresh
 * 
 * @expose
 * @memberof oj.ojFormLayout
 * @return {void}
 * @instance
 */ 

/**
 * @ignore
 */
var ojFormLayoutMeta = {
  "properties": {
    "direction": {
      "type": "string",
      "enumValues": ["column", "row"],
      "value": "column"
    },
    "labelEdge": {
      "type": "string",
      "enumValues": ["start", "top"],
      "value": "top"
    },
    "labelWidth": {
      "type": "string",
      "value": "33%"
    },
    "labelWrapping": {
      "type": "string",
      "enumValues": ["truncate", "wrap"],
      "value": "wrap"
    },
    "maxColumns": {
      "type": "number",
      "value": 1
    }
  },
  "extension": {
    _CONSTRUCTOR: ojFormLayout
  }
};
Object.freeze(ojFormLayoutMeta);

// counter for the generation of unique ids.
var _uidCounter = 0;


/**
 * The _ojFormLayout constructor function.  This function adds a wrapper div with
 * an oj-form class.
 * 
 * @constructor
 * @private
 */
function ojFormLayout(context) {
  var self = this;
  var element = context.element;
  var LABEL_ELEMENT_ID = "-labelled-by";
  // We need to mark every element we add so that we can remove it during a refresh/re-render.
  var BONUS_DOM_ATTR = "data-oj-formlayout-bonus-dom";
  var BONUS_DOM_SELECTOR = "["+BONUS_DOM_ATTR+"]";
  var ojFormReadyResolveFunc;
  var readyResolveFunc;
  var ojForm;
  var isInitialRender = true;
  var unresolvedChildren;
  var labelWidth;

  // Our version of GCC has a bug where the second param of MutationObserver.observe must be of 
  // type MutationObserverInit which isn't a real class that we can instantiate. Work around is to
  // create the MutationObserver on an alias of 'this' and call observe in a different function.
  // TODO Cleanup when we replace GCC with uglify in 5.0.0.
  /**
   * If the dom is mutated, call refresh to allow for added or deleted editable
   * value components.
   * 
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  self._rootElementMutationObserver = new MutationObserver(function (mutations) {
    var needsRefresh = false;
    
    // if the oj-form-layout has been removed from the body element's subtree,
    // no need to remove event listeners or refresh.  Just disconnect the observer.
    // This can happen when you press apply in the cookbook demo.
    if (document.body.contains(element))
    {
      // Check attribute mutations to see if any of our unresolved children have finish upgrading.
      // If so, add the associated oj-label and then add the bonus dom around the oj-label/component
      // pair.
      _processUnresolvedChildren(mutations);
      // Move any added direct children to the ojForm div.
      _moveNewDirectChildrenToOjFormDiv(mutations);
      // Ignore mutations that aren't direct children of bonus dom div elements as those
      // are not generated by application program logic and the should't require
      // a refresh. Examples of mutations that we care about: an application
      // adding a new editable value element; removing an existing original
      // child element.  Example of mutations that we don't care about: dom 
      // elements added or removed inside of the editable value child elements 
      // or oj-label elements.
      // In theory, we shouldn't get mutaion events if we wait until a component is ready
      // but are are still seeing mutations sometimes even though the we waited on the busyContext
      // for the component. See 
      if (!_ignoreMutations(mutations))
      {
        _removeEventListenersFromRemovedChildren(mutations); 
        element.refresh();
      }
    }
    else
    {
      this.disconnect();
    }
  });

  this.createDOM = function() 
  {

    element.classList.add("oj-form-layout");

    // Create wrapper div
    ojForm = document.createElement("div");
    ojForm.classList.add("oj-form");
    ojForm.setAttribute("data-oj-context","");
    ojForm.setAttribute("data-oj-internal","");
    ojForm.setAttribute(BONUS_DOM_ATTR,"");

    // wrap the children with the div
    // we use .firstChild and not .firstElementChild so that comment
    // nodes will be copied
    while(element.firstChild) {
      ojForm.appendChild(element.firstChild); // @HTMLUpdateOK reparenting child nodes
    }
    
    element.appendChild(ojForm); // @HTMLUpdateOK appending internally created DOM element
  };

  /**
   * The main render function.  This function gets called on initial render,
   * when oj-form-layout attributes are modified, and when mutations occur
   * (child nodes are added or deleted).  If the render is not the initial
   * render, all of the bonus dom elements are removed.  Then the bonus dom
   * elements are added (for both initial and non-initial cases).  The bonus dom
   * are the elements like oj-flex/oj-flex-item DIVs and oj-label elements that
   * get added and these are marked with "data-oj-formlayout-bonus-dom".
   * 
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  this.updateDOM = function() 
  {
    _notReady();

    unresolvedChildren = []; // start with an empty list
    
    labelWidth = element["labelEdge"] === "start" ? element["labelWidth"] : "100%";

    // Wait until child elements have finished upgrading before performing DOM 
    // manipulations so we can access child properties, like label-hint and help-hints.
    // The new oj-form div element has a data-oj-context attribute so we can get 
    // the correctly scoped busy context  for the oj-form div subtree. When custom 
    // elements are reparented, they reregister their busy state if they haven't 
    // finished upgrading. Also, if the oj-form itself is being rendered, wait for 
    // that as well.
    var busyContext = oj.Context.getContext(ojForm).getBusyContext();
    busyContext.whenReady().then(function()
    {
      // add a busy state for the oj-form div so we don't have multiple threads
      // modifying the oj-form div at the same time.
      _ojFormNotReady();
      // we don't want the observer being called when we are making modifications to dom subtree elements.
      self._rootElementMutationObserver.disconnect();

      _updateOjFormDiv();

      if (!isInitialRender)
        _removeAllBonusDomElements();

      _addLabelsFromHints();
      _addAllFlexDivs();

      // We are done making modifications to dom subtree elements so we should start paying attention
      // to mutations of the oj-form DIV subtree.  The mutations we care about are added elements and
      // removed elements. For the added elements, we should only see them added next to an existing
      // original child element, which is always a child of a bonus dom DIV element, so we ignore all
      // other added elements. For removed elements, we only care about removed original children.
      // All other removed elements are ignored by the observer.
      // 
      // We also need to track custom element upgrade status. We do this by observing attribute
      // mutations and if the 'class' attribute gets "oj-complete" added, we know the custom element
      // is finished upgrading.
      self._rootElementMutationObserver.observe(element, {"childList": true, "subtree": true, "attributes": true});
      _ojFormMakeReady();
      _makeReady();

      if (isInitialRender)
        isInitialRender = false;
    });
  };

  /*
   * Called to set the component busy during rendering
   */
  function _notReady()
  {
    if (!readyResolveFunc) {
      var busyContext = oj.Context.getContext(element).getBusyContext();
      var options = {'description' : "The oj-form-layout component with id = '"+element["id"]+"' is being rendered."};
      readyResolveFunc = busyContext.addBusyState(options);
    }
  }
  
  /*
   * Called to set the component not busy after rendering is finished
   */
  function _makeReady()
  {
    if (readyResolveFunc) {
      readyResolveFunc();
      readyResolveFunc = null;
    }
  }
  
  /*
   * Called to make the oj-form div busy during rendering
   */
  function _ojFormNotReady()
  {
    if (!ojFormReadyResolveFunc) {
      var busyContext = oj.Context.getContext(ojForm).getBusyContext();
      var options = {'description' : "The oj-form div for oj-form-layout component with id = '"+element["id"]+"' is being rendered."};
      ojFormReadyResolveFunc = busyContext.addBusyState(options);
    }
  }
  
  /*
   * Called to make the oj-form div not busy after rendering is finished
   */
  function _ojFormMakeReady()
  {
    if (ojFormReadyResolveFunc) {
      ojFormReadyResolveFunc();
      ojFormReadyResolveFunc = null;
    }
  }
  
  /**
   * Updates the oj-form div with the appropriate styles and column count.
   *
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _updateOjFormDiv()
  {
    var maxCols = element["maxColumns"];
    
    if (element["labelEdge"] === "start")
    {
      // ie11 doesn't support adding multiple classes so we add these one at a time.
      ojForm.classList.add("oj-form-cols-labels-inline");
      ojForm.classList.add("oj-formlayout-labels-inline");
      ojForm.classList.remove("oj-form-cols");
    }
    else
    {
      ojForm.classList.add("oj-form-cols");
      ojForm.classList.remove("oj-form-cols-labels-inline");
      ojForm.classList.remove("oj-formlayout-labels-inline");
    }

    if (element["labelWrapping"] === "truncate")
      ojForm.classList.add("oj-formlayout-labels-nowrap");
    else
      ojForm.classList.remove("oj-formlayout-labels-nowrap");

    // When direction === "row", we need to set the columns to 1, as we use the
    // oj-flex and oj-flex-item divs to control the number of columns.
    if (element["direction"] === "row")
      maxCols = 1;
    else
      maxCols = element["maxColumns"];

    ojForm["style"]["columnCount"] = maxCols;
    ojForm["style"]["webkitColumnCount"] = maxCols;
    ojForm["style"]["MozColumnCount"] = maxCols;    
  }
  
  /**
   * This method goes through and removes all of the bonus dom elements, those that
   * have been marked with the 'data-oj-formlayout-bonus-dom' attribute. For oj-label bonus
   * dom elements, we just remove the element as we don't need to preserve the children.
   * For the oj-flex and oj-flex-item DIVs, we need to preserve the children as they
   * contain the original child elements that we don't want removed. 
   *
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _removeAllBonusDomElements()
  {
    var bonusDomElems = ojForm.querySelectorAll(BONUS_DOM_SELECTOR);
    var length = bonusDomElems.length;
    
    for (var i = 0; i < length; ++i)
    {
      var bonusDomElem = bonusDomElems[i];
      
      if (bonusDomElem.tagName === "OJ-LABEL")
      {
        // For the oj-label elements we create, we can just remove them safely
        // as none of their children are the original child elements we are preserving
        bonusDomElem.parentElement.removeChild(bonusDomElem);
      }
      else
      {
        // for all other elements, remove the element preserving the children
        _removeElementAndReparentChildren(bonusDomElem);
      }
    }
  }
  
  /**
   * For each editable value child with a label hint, add an oj-label.
   * 
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _addLabelsFromHints()
  {
    var child = ojForm.firstElementChild;
    var childCnt = 0;
    var directionIsColumn = element["direction"] === "column";

    // for all of the direct children, add a label for any child element supports "labelHint"
    while (child) {
      // check for custom element
      if (child.tagName.indexOf("-") !== -1)
      {
        if (child.tagName.toLowerCase() === "oj-label")
        {
          // if we find an oj-label component, then skip it and the next sibling element.
          // If there is no next sibling element, throw an error.  We don't need to know
          // if they are completely upgraded as we are excepting all element types for the child
          // that follows an oj-label.  The assumption is that the app developer has correctly
          // associated the oj-label with and element of some kind.
          var label = child;
          child = child.nextElementSibling; // move to the next child element

          // an oj-label should always be followed by an element
          if (!child)
          {
            _ensureUniqueId(element);
            _ensureUniqueId(label);

            // make the component ready before throwing an error so the page won't hang
            _ojFormMakeReady();
            _makeReady();

            throw new Error("oj-form-layout component with id='"+element.id+"' has an oj-label child element with id='"+label.id+
                    "' but has no next sibling element that it is associated with.");
          }
        }
        else if (child.classList.contains("oj-complete"))
        {
          _addLabelFromHint(child);
        }
        else
        {
          // we need to tell this child if it needs an oj-flex div created for it.
          if (directionIsColumn || childCnt%element["maxColumns"] === 0)
          {
            child.setAttribute("data-oj-needs-oj-flex-div", "");
          }
          // This custom element child hasn't been upgraded so add it to the unresolved children
          unresolvedChildren.push(child);
        }
      }
      else
        _unsupportedChildType(child);
      
      childCnt++;
      child = child.nextElementSibling; // move to the next child element
    }   
  }
  
  function _unsupportedChildType(child)
  {
    // not a valid element with either label-hint or help-hints, so throw an exception
    _ensureUniqueId(element);
    _ensureUniqueId(child);

    // make the component ready before throwing an error so the page won't hang
    _ojFormMakeReady();
    _makeReady();

    throw new Error("oj-form-layout component with id='"+element["id"]+"' has a child element "+child.tagName+" with id='"+child["id"]+
            "' that is not a supported child type.  See oj-form-layout tag documentation for more information.");
  }

  /**
   * For one child, add an oj-label if it supports label-hint.
   * 
   * @param {EventTarget} child The element that may need an oj-label created
   * @memberof oj.ojFormLayout
   * @returns {Element|null} the oj-label element or null if no label was created
   * @instance
   * @private
   */
  function _addLabelFromHint(child)
  {
    var ojLabel = null;
    
    if (child instanceof Element && "labelHint" in child)
    {
      _ensureUniqueId(child);
      ojLabel = _createOjLabelAndInitialize(child);

      // the label should preceed the input element it is associated with
      child.parentElement.insertBefore(ojLabel, child); // @HTMLUpdateOK insert oj-label containing trusted content.

      _updateLabelHelpAndShowRequired(child, ojLabel);
      // JET's custom element's property change events, 
      // e.g., labelHintChanged, 
      // do not bubble and component events do bubble. 
      // Therefore, we can't listen on the 
      // oj-form-layout for child property change events.
      // We add them to the component themselves.
      _addChildEventListeners(child);
    }
    else
      _unsupportedChildType(child);
        
    return ojLabel;
  }
      
  /**
   * create the oj-label and initialize it
   * 
   * @param {Element} editableElem the element associated with the oj-label element
   * @returns {Element} the oj-label element
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _createOjLabelAndInitialize(editableElem)
  {
    var ojLabel = document.createElement("oj-label");
    ojLabel.setAttribute(BONUS_DOM_ATTR,"");
    ojLabel.setAttribute("data-oj-internal","");

    // programmatically created elements not managed by a binding stratagy like knockout
    // needs this attribute to signal the component should be created.
    ojLabel.setAttribute("data-oj-binding-provider", "none");
    ojLabel.setAttribute("data-oj-context","");
        
    // Note: the hint might be null, but that is fine, we still want a label for this case to hang the required and help icons off of
    // and allow for programatic changes to label-hint, help-hints, required.
    var span = document.createElement("span");
    span["id"] = editableElem["id"]+"|hint";
    span["textContent"] = editableElem["labelHint"];
    ojLabel.appendChild(span); // @HTMLUpdateOK append span containing trusted content.
    
    _linkLabelAndElement(ojLabel, editableElem);
    
    return ojLabel;
  }
  
  /**
   * for editable value components that have a labelled-by attribute, we need to
   * give the oj-label an ID and then set the labelled-by to that ID. Otherwise
   * we set the for attribute of the oj-label to the ID of the editable value component
   * 
   * @param {Element} ojLabel
   * @param {Element} editableElem element to link with the oj-label
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _linkLabelAndElement(ojLabel, editableElem)
  {
    if ("labelledBy" in editableElem)
    {
      ojLabel["id"] = editableElem.id+LABEL_ELEMENT_ID;
      editableElem["labelledBy"] = ojLabel.id;
    }
    else
      ojLabel.setAttribute("for", editableElem.id);
  }
  
  /**
   * listener for labelHintChanged events.
   * 
   * @param {Event} event the change event object
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _childLabelHintChanged(event)
  {
    var editable = event.target;
    // Get the hint span and update its innerText
    var span = document.getElementById(editable["id"]+"|hint");

    if (span)
    {
      span["innerText"] = event.detail.value;
    }
  }

  /**
   * listener for helpHintsChanged events.
   * 
   * @param {Event} event the change event object
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _childHelpHintsChanged(event)
  {
    var editable = event.target;
    var ojLabel = _getOjLabelForChildElement(editable);

    if (ojLabel)
    {
      ojLabel["help"] = event.detail.value;
    }
  }

  /**
   * listener for requiredChanged events.
   * 
   * @param {Event} event the change event object
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _childRequiredChanged(event) {
    var editable = event.target;
    var ojLabel = _getOjLabelForChildElement(editable);

    if (ojLabel)
    {
      ojLabel["showRequired"] = event.detail.value;
    }
  }

  /**
   * If the child element doesn't have listeners yet, add them.
   * Only the component events bubble and not the property change events, 
   * so it doesn't work to listen on the oj-form-layout for 
   * these property change events. Instead we
   * put them on the child component themselves, and we share the listener.
   * 
   * @param {Element} child
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _addChildEventListeners(child)
  {
      child.addEventListener("labelHintChanged", _childLabelHintChanged);
      child.addEventListener("helpHintsChanged", _childHelpHintsChanged);
      child.addEventListener("requiredChanged", _childRequiredChanged);
    }
  
  /**
   * This function moves any new direct children under the ojForm node, which is where the
   * oj-form-layout component expects these children to be.
   * @param {Array.<MutationRecord>} mutations the mustation list passed in to the mutation observer function
   * @returns {undefined}
   */
  function _moveNewDirectChildrenToOjFormDiv(mutations)
  {
    var mutationsLength = mutations.length;
    
    for (var i = 0; i < mutationsLength; i++)
    {
      var mutation = mutations[i];
      
      if (mutation.type === "childList")
      {
        var addedNodesLength = mutation.addedNodes.length;

        for (var j = 0; j < addedNodesLength; j++)
        {
          var child = mutation.addedNodes[j];

          // check to see if it was added as a direct child
          if (child.parentNode === element)
          {
            ojForm.appendChild(child);
  }
        }
      }
  
    }
  }
  
  /**
   * Process any component that has been upgraded and was placed on the unresolved children list
   * during a previous render adding the label and flex divs.
   * @param {Array.<MutationRecord>} mutations the mustation list passed in to the mutation observer function
   * @returns {undefined}
   */
  function _processUnresolvedChildren(mutations)
  {
    var mutationsLength = mutations.length;
    
    for (var i = 0; i < mutationsLength; i++)
    {
      var mutation = mutations[i];
      
      if (mutation.type === "attributes" && mutation.attributeName === "class")
      {
        var index = unresolvedChildren.indexOf(mutation.target);
        if (index !== -1 && mutation.target.classList.contains("oj-complete"))
        {
          unresolvedChildren.splice(index, 1);
          var ojLabel = _addLabelFromHint(mutation.target);
          // If a label was created, add the flex divs
          if (ojLabel)
          {
            _addFlexDivs(ojLabel, mutation.target, mutation.target.hasAttribute("data-oj-needs-oj-flex-div"));
          }
        }
      }
    }
  }
  
  /**
   * if the removed child nodes have event listeners we've added, remove them
   * Only the component events bubble and not the property change events, 
   * so it doesn't work to listen on the oj-form-layout 
   * for these property change events. We add/remove the listener from the 
   * child JET component itself.
   * 
   * @param {Array.<MutationRecord>} mutations the array of MutationRecords from the observer
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _removeEventListenersFromRemovedChildren(mutations)
  {
    var mutationsLength = mutations.length;
    
    for (var i = 0; i < mutationsLength; i++)
    {
      var mutation = mutations[i];
      
      if (mutation.type === "childList")
      {
      var removedNodesLength = mutation.removedNodes.length;

      for (var j = 0; j < removedNodesLength; j++)
      {
        var child = mutation.removedNodes[j];

          if (child.nodeType === 1)
        {
          child.removeEventListener("labelHintChanged", _childLabelHintChanged);
          child.removeEventListener("helpHintsChanged", _childHelpHintsChanged);
          child.removeEventListener("requiredChanged", _childRequiredChanged);
        }
      }
      }
    }    
  }
  
  /**
   * Return the oj-label element for a component. This may be the oj-label with a for attribute
   * for the component, or may be the oj-label that this component's labelled-by points to.
   * 
   * @param {EventTarget} child The child element whose label we want to return
   * @memberof oj.ojFormLayout
   * @instance
   * @return {Element} The oj-label element.
   * @private
   */
  function _getOjLabelForChildElement(child)
  {
    var ojLabel;
    // for editable value components that have a labelled-by attribute, we need to
    // retrieve the oj-label by the labelledBy ID. Otherwise
    // we query for it by the for attribute.
    if ("labelledBy" in child) {
      // For some reason, element is undefined in this case, so using document
      ojLabel = document.getElementById(child["labelledBy"]);
    } else {
      ojLabel = element.querySelector('oj-label[for="'+child["id"]+'"]');
    }
    
    return ojLabel;
  }
  
  /**
   * Update the help and showRequired from the form component's helpHints and
   * required attributes.  This needs to be done after the label is no longer
   * busy.
   * 
   * @param {Element} comp
   * @param {Element} ojLabel
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _updateLabelHelpAndShowRequired(comp, ojLabel)
  {
    var busyContext = oj.Context.getContext(ojLabel).getBusyContext();
    busyContext.whenReady().then(function()
    {
      var helpHints = comp["helpHints"];
  
      if (helpHints)
      {
        // if either of these has a value, set the help property of ojLabel.
        // This is a work around for the fact that oj-label doesn't correctly
        // handle the default value for help.definition or help.source (i.e. when
        // the value is "").  If we have the default value, null or undefined
        // for both of them, we don't set the help attribute on the oj-label.
        if (helpHints.definition || helpHints.source)
          ojLabel["help"] = helpHints;
      }

      ojLabel["showRequired"] = comp["required"];
    });
  }
    
  /**
   * Add an oj-flex div for each label/input pair and add an oj-flex-item div
   * for each label and each input.
   * 
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _addAllFlexDivs()
  {
    var childArray = []; // copy array
    var label;
    var elem;
    var arrayLength = ojForm.children.length;
    var directionIsColumn = element["direction"] === "column";
    var j = 0;
    var pairCnt = 0; // counter of label/element count
    _appendChildrenToArray(ojForm.children, childArray);
    
    // each iteration should process a label/input pair
    while(j < arrayLength) {
      label = childArray[j];
      
      // if the current child element is on the unresolvedChildren list, skip it
      if (unresolvedChildren.indexOf(label) === -1)
      {
        j++; // skip to the next element
        elem = childArray[j];
        
        // for direction === "row", we only render the oj-flex div once per row.
        _addFlexDivs(label, elem, directionIsColumn || pairCnt%element["maxColumns"] === 0);
      }
      pairCnt++;
      j++;
    }   
  }

  function _addFlexDivs(label, elem, createOjFlex)
  {
    var ojFlex;
    var labelOjFlexItem;
    var elementOjFlexItem;

    if (createOjFlex)
      ojFlex = _createDivElement("oj-flex");
    else
      ojFlex = label.previousElementSibling; // This will be the oj-flex div for a row

    labelOjFlexItem = _createDivElement("oj-flex-item");
    ojFlex.appendChild(labelOjFlexItem); // @HTMLUpdateOK append div containing trusted content.

    labelOjFlexItem.appendChild(label); // @HTMLUpdateOK append oj-label containing trusted content.

    // set the width of the label flex item
    labelOjFlexItem.style.webkitFlex = "0 1 "+labelWidth;
    labelOjFlexItem.style.flex = "0 1 "+labelWidth;
    labelOjFlexItem.style.maxWidth = labelWidth;
    labelOjFlexItem.style.width = labelWidth;

    // create the component flex-item and append the element as a child
    elementOjFlexItem = _createDivElement("oj-flex-item");
    ojFlex.appendChild(elementOjFlexItem); // @HTMLUpdateOK append div containing trusted content.
    elem.parentElement.insertBefore(ojFlex, elem); // @HTMLUpdateOK insert div containing trusted content.
    elementOjFlexItem.appendChild(elem); // @HTMLUpdateOK append element containing trusted content.
  }

  /**
   * Append a children list to an array.
   * 
   * @param {NodeList} children the child element list to copy
   * @param {Array.<Element>} childArray the array to copy to
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _appendChildrenToArray(children, childArray)
  {
    var i = children.length;
    while (i--) childArray[i] = children[i];
  }

  /**
   * Create a div element and initialize it.
   * 
   * @param {string} className the style class name of the div
   * @returns {Element} the created div element
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _createDivElement(className)
  {
    var div = document.createElement("div");
    div.setAttribute(BONUS_DOM_ATTR,"");
    div.setAttribute("data-oj-internal","");
    div.classList.add(className);
    return div;
      }

  /**
   * The target element of a mutation should be a DIV with the bonus dom attribute
   * set on it.  Otherwise ignore the mutations.  We never have to worry about bonus
   * dom being inserted or removed.  The only time bonus dom elements are added
   * or removed, the mutation observer is disconnected.  The application developer
   * should never manipulate generated dom elements.  What this is checking for
   * is if there are any nodes being added or removed from our bonus dom divs while
   * the observer is active.  Specific examples would be adding a new
   * oj-input-text element, or removing an existing editable value element. These
   * should always be add to, or removed from one of the bonus dom DIV elements. 
   * Any set of mutations that doesn't contain at least one node added or removed 
   * from a bonus dom div will be ignored.
   * 
   * @param {Array.<MutationRecord>} mutations the array of MutationRecords from the observer
   * @private
   * @returns {boolean} true if we should ignore these mutations
   */
  function _ignoreMutations(mutations)
  {
    var ignore = true;
    var mutationsLength = mutations.length;

    for (var i = 0; i < mutationsLength; i++)
    {
      var mutation = mutations[i];
      
      if (mutation.type === "childList" && _isBonusDomDivOrSelf(mutation.target)
                                        && _isMutationOfThisFormLayout(mutation.target))
      {
        ignore = false;
        break;
      }
    }

    return ignore;
  }
  
  /**
   * Checks to see if the node is a DIV with our bonus dom attribute on it or if it is the
   * oj-form-layout itself.
   * 
   * @param {Node} node Node to check.
   * @returns {boolean|null} true if we have a parent DIV with the bonus dom attribute
   */
  function _isBonusDomDivOrSelf(node)
  {
    return node === element || (node &&
                                node.tagName === "DIV" &&
                                node.hasAttribute(BONUS_DOM_ATTR));
  }
  /**
   * Checks to make sure that this mutation belongs to this oj-form-layout rather than a nested -oj-form-layout
   * @param {Node} node
   * @returns {boolean}
   */
  function _isMutationOfThisFormLayout(node)
  {
    var result = true;
    
    while (node !== element)
    {
      if (node.tagName === "OJ-FORM-LAYOUT")
      {
        result = false;
        break;
      }
      
      node = node.parentElement;

      // must be a deleted node so ignore it
      if (node == null)
      {
        result = false;
        break;
      }
    }
    
    return result;
  }
  /**
   * Remove a dom element preserving all of its children
   * 
   * @param {Element} elem
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _removeElementAndReparentChildren(elem)
  {
    while (elem.firstElementChild)
    {
      elem.parentNode.insertBefore(elem.firstElementChild, elem); // @HTMLUpdateOK reparenting existing element.
    }

    elem.parentNode.removeChild(elem);    
  }
  
  /**
   * If an element doesn't have an ID, generate a unique ID for it.
   * 
   * @param {Element} elem
   * @memberof oj.ojFormLayout
   * @instance
   * @private
   */
  function _ensureUniqueId(elem) {
    if (!elem["id"]) {
      elem["id"] = 'oflId_' + _uidCounter++;
    }
  }
}

oj.CustomElementBridge.registerMetadata('oj-form-layout', null, ojFormLayoutMeta);
oj.CustomElementBridge.register('oj-form-layout', {'metadata': oj.CustomElementBridge.getMetadata('oj-form-layout')});

/**
 * Copyright (c) 2016, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojGantt
 * @augments oj.dvtTimeComponent
 * @since 2.1.0
 * @ojstatus preview
 * @ojshortdesc Displays scheduling information graphically, making it easier to plan, coordinate, and track various tasks and resources.
 * @ojrole application
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="GanttOverview-section">
 *   JET Gantt
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#GanttOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET Gantt is a themable, WAI-ARIA compliant element that illustrates the start and finish dates of tasks.</p>
 *
 * {@ojinclude "name":"warning"}
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-gantt
 *   start='2017-01-01T05:00:00.000Z'
 *   end='2017-12-31T05:00:00.000Z'
 *   major-axis='{"scale": "months"}'
 *   minor-axis='{"scale": "weeks"}'
 *   rows='[[data]]'>
 * &lt;/oj-gantt>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="formats-section">
 *   Date and Time Formats
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#formats-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"formatsDoc"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets.</p>
 *
 * <h4>Data Set Size</h4>
 * <p>Gantt has been tested to render up to 5000 tasks in 1 second.  In general, applications 
 *    should avoid displaying a very large number of task bars at the same time.  Applications 
 *    could for example use a filter to only show task bars for a specific set of time ranges.</p>
 *
 * <h4>Gantt Span</h4>
 *
 * <p>It's recommended that applications limit the number of time intervals that are
 *    rendered by the Gantt chart. For example, a Gantt chart spanning one year with a scale
 *    of hours will display (365 * 24) 8,760 intervals. Rendering this many intervals
 *    can cause severe performance degradation when interacting with the element
 *    (scrolling and zooming) regardless of the number of task bars present.
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojGantt', $['oj']['dvtTimeComponent'],
{
  widgetEventPrefix : "oj",
  options: 
  {
    /**
     * The position of the major and minor axis.
     * @expose
     * @name axisPosition
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @ojvalue {string} "bottom"
     * @ojvalue {string} "top"
     * @default "top"
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">axis-position</code> attribute specified:</caption>
     * &lt;oj-gantt axis-position='bottom'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">axisPosition</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.axisPosition;
     *
     * // setter
     * myGantt.axisPosition = 'bottom';
     */
    axisPosition: "top",
    /**
     * Defines the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">animation-on-data-change</code> attribute specified:</caption>
     * &lt;oj-gantt animation-on-data-change='auto'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">animationOnDataChange</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.animationOnDataChange;
     *
     * // setter
     * myGantt.animationOnDataChange = 'auto';
     */
    animationOnDataChange: "none",
    /**
     * Defines the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">animation-on-display</code> attribute specified:</caption>
     * &lt;oj-gantt animation-on-display='auto'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">animationOnDisplay</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.animationOnDisplay;
     *
     * // setter
     * myGantt.animationOnDisplay = 'auto';
     */
    animationOnDisplay: "none",
    /**
     * An array of objects that defines dependencies between tasks. Also accepts a Promise that will resolve with an array for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name dependencies
     * @memberof oj.ojGantt
     * @instance
     * @type {?(Array.<Object>|Promise)}
     * @ojsignature {target: "Accessor", value: {GetterType: "Promise<Array<oj.ojGantt.Dependency>>|null", SetterType: "Array<oj.ojGantt.Dependency>|Promise<Array<oj.ojGantt.Dependency>>|null"}, jsdocOverride: true}
     * @default null
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">dependencies</code> attribute specified:</caption>
     * &lt;oj-gantt dependencies='[[myDependencies]]'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">dependencies</code> property after initialization:</caption>
     * // Get all (The dependencies getter always returns a Promise so there is no "get one" syntax)
     * var values = myGantt.dependencies;
     *
     * // Set all (There is no permissible "set one" syntax.)
     * myGantt.dependencies = [
     *     {
     *         "id": "d1",
     *         "predecessorTaskId": "task1",
     *         "successorTaskId": "task2",
     *         "svgStyle": {"stroke": "red"},
     *         "type": "startFinish"
     *     },
     *     {
     *         "id": "d2",
     *         "predecessorTaskId": "task2",
     *         "successorTaskId": "task3"
     *     }
     * ];
     */
    dependencies: null,
    /**
     * The end time of the Gantt. A valid value is required in order for the Gantt to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name end
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">end</code> attribute specified:</caption>
     * &lt;oj-gantt end='2017-12-31T05:00:00.000Z'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">end</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.end;
     *
     * // setter
     * myGantt.end = '2017-12-31T05:00:00.000Z';
     */
    end: "",
    /**
     * An object specifying whether to display or hide the horizontal and vertical grid lines.
     * @expose
     * @name gridlines
     * @memberof oj.ojGantt
     * @instance
     * @type {Object}
     * @default {"horizontal": "auto", "vertical": "auto"}
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">gridlines</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-gantt gridlines.horizontal='auto' gridlines.vertical='auto'>&lt;/oj-gantt>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-gantt gridlines='{"horizontal": "auto", "vertical": "auto"}'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">gridlines</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.gridlines.horizontal;
     * 
     * // Set one, leaving the others intact.
     * myGantt.setProperty('gridlines.horizontal', 'auto');
     * 
     * // Get all
     * var values = myGantt.gridlines;
     *
     * // Set all. Must list every resource key, as those not listed are lost.
     * myGantt.gridlines = {
     *     "horizontal": "auto",
     *     "vertical": "auto"
     * };
     */
    gridlines: {
      /**
       * Horizontal gridlines. The default value is "auto", which means Gantt will decide whether the grid lines should be made visible or hidden.
       * <br></br>See the <a href="#gridlines">gridlines</a> attribute for usage examples.
       * @expose
       * @name gridlines.horizontal
       * @memberof! oj.ojGantt
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "hidden"
       * @ojvalue {string} "visible"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      horizontal: "auto",
      /**
       * Vertical gridlines. The default value is "auto", which means Gantt will decide whether the grid lines should be made visible or hidden.
       * <br></br>See the <a href="#gridlines">gridlines</a> attribute for usage examples.
       * @expose
       * @name gridlines.vertical
       * @memberof! oj.ojGantt
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "hidden"
       * @ojvalue {string} "visible"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
       vertical: "auto"
    },
    /**
     * An object with the following properties, used to define the minor time axis. This is required in order for the Gantt to properly render.
     * @expose
     * @name minorAxis
     * @memberof oj.ojGantt
     * @instance
     * @type {Object}
     * @default {"converter": {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}}
     *
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">minor-axis</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-gantt minor-axis.converter="[[myConverterObject]]" minor-axis.scale="weeks" minor-axis.zoom-order='["quarters", "months", "weeks", "days"]'>&lt;/oj-gantt>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-gantt minor-axis='{"scale": "weeks", "zoomOrder": ["quarters", "months", "weeks", "days"]}'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">minorAxis</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.minorAxis.scale;
     *
     * // Set one, leaving the others intact.
     * myGantt.setProperty('minorAxis.scale', 'weeks');
     *
     * // Get all
     * var values = myGantt.minorAxis;
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myGantt.minorAxis = {
     *     "converter": myConverterObject,
     *     "scale": "weeks",
     *     "zoomOrder": ["quarters", "months", "weeks", "days"]
     * };
     */
    minorAxis: {
      /**
       * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the labels of the minor axis for all 'scale' values, or 
       * an object literal whose keys are 'scale' values that map specific converters for scale specific formatting (see {@link oj.ojTimeAxis.Converter}). 
       * See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
       * <br></br>See the <a href="#minorAxis">minor-axis</a> attribute for usage examples.
       * @expose
       * @name minorAxis.converter
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?(oj.ojTimeAxis.Converter|oj.Converter<string>)", jsdocOverride: true}  
       * @default {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}
       */
      converter: {
        "default": null,
        "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}),
        "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}),
        "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}),
        "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})
      },
      /**
       * The time scale used for the minor axis. This is required in order for the Gantt to properly render.
       * <br></br>See the <a href="#minorAxis">minor-axis</a> attribute for usage examples.
       * @expose
       * @name minorAxis.scale
       * @memberof! oj.ojGantt
       * @instance
       * @type {?string}
       * @ojvalue {string} "seconds"
       * @ojvalue {string} "minutes"
       * @ojvalue {string} "hours"
       * @ojvalue {string} "days"
       * @ojvalue {string} "weeks"
       * @ojvalue {string} "months"
       * @ojvalue {string} "quarters"
       * @ojvalue {string} "years"
       * @default null
       */
      scale: null,
      /**
       * An array of strings containing the names of scales used for zooming from longest to shortest. If not specified, the 'scale' specified on the axis will be used at all zoom levels.
       * <br></br>See the <a href="#minorAxis">minor-axis</a> attribute for usage examples.
       * @expose
       * @name minorAxis.zoomOrder
       * @memberof! oj.ojGantt
       * @instance
       * @type {?Array.<string>}
       * @ojsignature {target: "Type", value: "?"}
       * @default null
       */
      zoomOrder: null
    },
    /**
     * An object with the following properties, used to define the major time axis. If not specified, no major time axis is shown.
     * @expose
     * @name majorAxis
     * @memberof oj.ojGantt
     * @instance
     * @type {?Object}
     * @default {"converter": {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}}
     *
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">major-axis</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-gantt major-axis.converter="[[myConverterObject]]" major-axis.scale="months" major-axis.zoom-order='["quarters", "months", "weeks", "days"]'>&lt;/oj-gantt>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-gantt major-axis='{"scale": "months", "zoomOrder": ["quarters", "months", "weeks", "days"]}'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">majorAxis</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.majorAxis.scale;
     *
     * // Set one, leaving the others intact.
     * myGantt.setProperty('majorAxis.scale', 'months');
     *
     * // Get all
     * var values = myGantt.majorAxis;
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myGantt.majorAxis = {
     *     "converter": myConverterObject,
     *     "scale": "months",
     *     "zoomOrder": ["quarters", "months", "weeks", "days"]
     * };
     */
    majorAxis: {
      /**
       * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the labels of the major axis for all 'scale' values, or 
       * an object literal whose keys are 'scale' values that map specific converters for scale specific formatting (see {@link oj.ojTimeAxis.Converter}). 
       * See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
       * <br></br>See the <a href="#majorAxis">major-axis</a> attribute for usage examples.
       * @expose
       * @name majorAxis.converter
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?(oj.ojTimeAxis.Converter|oj.Converter<string>)", jsdocOverride: true}
       * @default {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}
       */
      converter: {
        "default": null,
        "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}),
        "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}),
        "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}),
        "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})
      },
      /**
       * The time scale used for the major axis.
       * <br></br>See the <a href="#majorAxis">major-axis</a> attribute for usage examples.
       * @expose
       * @name majorAxis.scale
       * @memberof! oj.ojGantt
       * @instance
       * @type {?string}
       * @ojvalue {string} "seconds"
       * @ojvalue {string} "minutes"
       * @ojvalue {string} "hours"
       * @ojvalue {string} "days"
       * @ojvalue {string} "weeks"
       * @ojvalue {string} "months"
       * @ojvalue {string} "quarters"
       * @ojvalue {string} "years"
       * @default null
       */
      scale: null,
      /**
       * An array of strings containing the names of scales used for zooming from longest to shortest. If not specified, the 'scale' specified on the axis will be used at all zoom levels.
       * <br></br>See the <a href="#majorAxis">major-axis</a> attribute for usage examples.
       * @expose
       * @name majorAxis.zoomOrder
       * @memberof! oj.ojGantt
       * @instance
       * @type {?Array.<string>}
       * @ojsignature {target: "Type", value: "?"}
       * @default null
       */
      zoomOrder: null
    },
    /**
     * The array of reference objects associated with the gantt. For each reference object, a line is rendered at the specified value. Currently only the first reference object in the array is supported. Any additional objects supplied in the array will be ignored.
     * @expose
     * @name referenceObjects
     * @memberof oj.ojGantt
     * @instance
     * @type {Array.<Object>}
     * @ojsignature {target: "Type", value: "Array<oj.ojGantt.ReferenceObject>", jsdocOverride: true}
     * @default []
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">reference-objects</code> attribute specified:</caption>
     * &lt;oj-gantt reference-objects='[{"value": "2017-04-15T04:00:00.000Z"}]'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">referenceObjects</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.referenceObjects[0];
     * 
     * // Get all
     * var values = myGantt.referenceObjects;
     *
     * // Set all (There is no permissible "set one" syntax.)
     * myGantt.referenceObjects = [{
     *     "value": "2017-04-15T00:00:00.000Z",
     *     "svgStyle": {"stroke": "red"}
     * }];
     */
    referenceObjects: [],
    /**
     * An object defining properties for the row labels region.
     * @expose
     * @name rowAxis
     * @memberof oj.ojGantt
     * @instance
     * @type {Object}
     * @default {"rendered": "off", "maxWidth": "none"}
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">row-axis</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-gantt row-axis.rendered='on' row-axis.max-width='50px'>&lt;/oj-gantt>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-gantt row-axis='{"rendered": "on", "maxWidth": "50px"}'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">rowAxis</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.rowAxis.rendered;
     *
     * // Set one, leaving the others intact
     * myGantt.setProperty('rowAxis.rendered', 'on');
     * 
     * // Get all
     * var values = myGantt.rowAxis;
     *
     * // Set all (any value not set will be ignored)
     * myGantt.rowAxis = {
     *     rendered: "on",
     *     maxWidth: "50px"
     * };
     */
    rowAxis: {
      /**
       * Defines whether row labels are rendered.
       * <br></br>See the <a href="#rowAxis">row-axis</a> attribute for usage examples.
       * @expose
       * @name rowAxis.rendered
       * @memberof! oj.ojGantt
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "on"
       * @ojvalue {string} "off"
       * @default "off"
       */
      rendered: "off",
      /**
       * Defines the maximum width of the region in pixels (e.g. '50px') or percent (e.g. '15%') of the element width. If 'none' is specified, then the width has no maximum value. Labels will truncate to fit.
       * <br></br>See the <a href="#rowAxis">row-axis</a> attribute for usage examples.
       * @expose
       * @name rowAxis.maxWidth
       * @memberof! oj.ojGantt
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default "none"
       */
      maxWidth: "none"
    },
    /**
     * An array of objects with the following properties, used to define rows and tasks within rows. Also accepts a Promise that will resolve with an array for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name rows
     * @memberof oj.ojGantt
     * @instance
     * @type {?(Array.<Object>|Promise)}
     * @ojsignature {target: "Accessor", value: {GetterType: "Promise<Array<oj.ojGantt.Row>>|null", SetterType: "Array<oj.ojGantt.Row>|Promise<Array<oj.ojGantt.Row>>|null"}, jsdocOverride: true}
     * @default null
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">rows</code> attribute specified:</caption>
     * &lt;oj-gantt rows='[[myRows]]'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">rows</code> property after initialization:</caption>
     * // Get all (The rows getter always returns a Promise so there is no "get one" syntax)
     * var values = myGantt.rows;
     *
     * // Set all (There is no permissible "set one" syntax.)
     * myGantt.rows = [
     *     {
     *         "id": "r1",
     *         "label": "Row 1",
     *         "tasks": [
     *             {
     *                 "id": "task1_1",
     *                 "start": "2017-01-04T17:00:00.000Z",
     *                 "end": "2017-01-10T17:00:00.000Z",
     *                 "label":"Label 1-1"
     *             },
     *             {
     *                 "id": "task1_2",
     *                 "start": "2017-02-04T17:00:00.000Z",
     *                 "end": "2017-02-10T17:00:00.000Z",
     *                 "label":"Label 1-2"
     *             }
     *         ]
     *     },
     *     {
     *         "id": "r2",
     *         "label": "Row 2",
     *         "tasks": [
     *             {
     *                 "id": "task2_1",
     *                 "start": "2017-01-10T17:00:00.000Z",
     *                 "end": "2017-01-24T17:00:00.000Z",
     *                 "label":"Label 2-1"
     *             },
     *             {
     *                 "id": "task2_2",
     *                 "start": "2017-02-10T17:00:00.000Z",
     *                 "end": "2017-02-27T17:00:00.000Z",
     *                 "label":"Label 2-2"
     *             }
     *         ]
     *     }
     * ];
     */
    rows: null,
    /**
     * An array of strings containing the ids of the initially selected tasks.
     * @expose
     * @name selection
     * @memberof oj.ojGantt
     * @instance
     * @type {Array.<string>}
     * @ojwriteback
     * @default []
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">selection</code> attribute specified:</caption>
     * &lt;oj-gantt selection='["taskID1", "taskID2", "taskID3"]'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">selection</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.selection[0];
     * 
     * // Get all
     * var values = myGantt.selection;
     *
     * // Set all (There is no permissible "set one" syntax.)
     * myGantt.selection = ["taskID1", "taskID2", "taskID3"];
     */
    selection: [],
    /**
     * The type of selection behavior that is enabled on the Gantt. If 'single' is specified, only a single task can be selected at once. If 'multiple', any number of tasks can be selected at once. Otherwise, selection is disabled.
     * @expose
     * @name selectionMode
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">selection-mode</code> attribute specified:</caption>
     * &lt;oj-gantt selection-mode='multiple'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.selectionMode;
     * 
     * // setter
     * myGantt.selectionMode = 'multiple';
     */
    selectionMode: "none",
    /**
     * The start time of the Gantt. A valid value is required in order for the Gantt to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name start
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">start</code> attribute specified:</caption>
     * &lt;oj-gantt start='2017-01-01T05:00:00.000Z'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">start</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.start;
     *
     * // setter
     * myGantt.start = "2017-01-01T05:00:00.000Z";
     */
    start: "",
    /**
     * An object with the following properties, used to define default styling for tasks in the Gantt. Properties specified on this object may 
     * be overridden by specifications on individual tasks.
     * @expose
     * @name taskDefaults
     * @memberof oj.ojGantt
     * @instance
     * @type {Object}
     * @default {"borderRadius": "0", "labelPosition": ["end", "innerCenter", "start", "max"], "height": null, "svgClassName": "", "svgStyle": {}, "type": "auto", "progress": {"borderRadius": "0", "height": "100%", "svgClassName": "", "svgStyle": {}}, "baseline": {"borderRadius": 0, "height": null, "svgClassName": "", "svgStyle": {}}}
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">task-defaults</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-gantt task-defaults.border-radius='5px' task-defaults.label-position='["end"]' task-defaults.progress.height="50%">&lt;/oj-gantt>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-gantt task-defaults='{"borderRadius": "5px", "labelPosition": ["end"], "progress": {"height": "50%"}}'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">taskDefaults</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.taskDefaults.height;
     *
     * // Get all
     * var values = myGantt.taskDefaults;
     *
     * // Set one, leaving the others intact.
     * myGantt.setProperty('taskDefaults.height', 30);
     *
     * // Set all. Must list every resource key, as those not listed are lost.
     * myGantt.taskDefaults = {
     *     "borderRadius": "5px",
     *     "labelPosition": ["end"],
     *     "height": 30,
     *     "progress": {"height": "50%"}
     * };
     */
    taskDefaults: {
      /**
       * The border radius of the task. Accepts values allowed in CSS border-radius attribute.
       * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @expose
       * @name taskDefaults.borderRadius
       * @memberof! oj.ojGantt
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default "0"
       */
      borderRadius: "0",
      /**
       * The position of the label relative to the task. An array of values is also supported. If an array is specified, then the values are traversed until a position that can fully display the label is found. If 'max' is specified in the array, then of all the positions evaluated up to that point of the traversal, the one with the largest space is used (label is truncated to fit). Naturally, 'max' is ignored if it's specified as the first value of the array. If the last value of the array is reached, but the label cannot be fully displayed, then the label is placed at that position, truncated to fit. Due to space constraints in the milestone and task with progress cases, the inner positions will exhibit the following behaviors: <ul> <li> For milestones, specifying 'innerStart', 'innerEnd', or 'innerCenter' would be equivalent to specifying 'start', 'end', and 'end' respectively. </li> <li> For tasks with progress, 'innerCenter' means the label will be aligned to the end of the progress bar, either placed inside or outside of the progress, whichever is the larger space. 'innerStart' and 'innerEnd' positions are honored when there is enough space to show the label at those positions. Otherwise, the aforementioned 'innerCenter' behavior is exhibited. </li> </ul>
       * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @ojshortdesc The position of the label relative to the task, or a priority order of label positions for the component to automatically choose from.
       * @expose
       * @name taskDefaults.labelPosition
       * @memberof! oj.ojGantt
       * @instance
       * @type {string|Array.<string>}
       * @ojsignature {target: "Type", value: "?(string|Array<string>)"}
       * @ojvalue {string} "start"
       * @ojvalue {string} "innerCenter"
       * @ojvalue {string} "innerStart"
       * @ojvalue {string} "innerEnd"
       * @ojvalue {string} "end"
       * @ojvalue {string} "none"
       * @default ["end", "innerCenter", "start", "max"]
       */
      labelPosition: ["end", "innerCenter", "start", "max"],
      /**
       * The height of the task in pixel. If not specified, a default height is used depending on the task type, and whether the baseline is specified.
       * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @expose
       * @name taskDefaults.height
       * @memberof! oj.ojGantt
       * @instance
       * @type {number}
       * @ojsignature {target: "Type", value: "?"}
       * @ojunits pixels
       * @default null
       */
      height: null,
      /**
       * A space delimited list of CSS style classes defining the style of the task.
       * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @expose
       * @name taskDefaults.svgClassName
       * @memberof! oj.ojGantt
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @default ""
       */
      svgClassName: "",
      /**
       * The CSS style defining the style of the task.
       * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @expose
       * @name taskDefaults.svgStyle
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {},
      /**
       * Defines the task type to be rendered.<br></br>If "milestone", and if <a href="#rows[].tasks[].start">start</a> and <a href="#rows[].tasks[].end">end</a> values are specified and unequal, the <a href="#rows[].tasks[].start">start</a> value is used to evaluate position.<br></br>If "auto", the type is inferred from the data:<ul> <li>If <a href="#rows[].tasks[].start">start</a> and <a href="#rows[].tasks[].end">end</a> values are specified and unequal, "normal" type is assumed.</li> <li>Otherwise, "milestone" type is assumed.</li></ul>
       * See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @ojshortdesc The type of task to be rendered.
       * @expose
       * @name taskDefaults.type
       * @memberof! oj.ojGantt
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "normal"
       * @ojvalue {string} "milestone"
       * @ojvalue {string} "summary"
       * @ojvalue {string} "auto"
       * @default "auto"
       */
      type: "auto",
      /**
       * An object with the following properties, used to define default styling for progress bars on non-milestone tasks.
       * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @expose
       * @name taskDefaults.progress
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"borderRadius": "0", "height": "100%", "svgClassName": "", "svgStyle": {}}
       */
      progress: {
        /**
         * The border radius of the progress bar. Accepts values allowed in CSS border-radius attribute.
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.progress.borderRadius
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @default "0"
         */
        borderRadius: "0",
        /**
         * Specifies the height of the progress bar in pixels (e.g. '50px') or percent of the associated task bar (e.g. '15%').
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.progress.height
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @default "100%"
         */
        height: "100%",
        /**
         * A space delimited list of CSS style classes to apply to the progress bar. Note that only CSS style applicable to SVG elements can be used.
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.progress.svgClassName
         * @memberof! oj.ojGantt
         * @instance
         * @type {?string}
         * @ojsignature {target: "Type", value: "?"}
         * @default ""
         */
        svgClassName: "",
        /**
         * The CSS inline style to apply to the progress bar. Only CSS style applicable to SVG elements can be used.
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.progress.svgStyle
         * @memberof! oj.ojGantt
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         * @default {}
         */
        svgStyle: {}
      },
      /**
       * An object with the following properties, used to define default styling for task baseline elements.
       * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
       * @expose
       * @name taskDefaults.baseline
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"borderRadius": 0, "svgClassName": "", "svgStyle": {}}
       */
      baseline: {
        /**
         * The border radius of the baseline. Accepts values allowed in CSS border-radius attribute.
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.baseline.borderRadius
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @default "0"
         */
        borderRadius: "0",
        /**
         * The height of the baseline in pixel. If not specified, a default height is used depending on the baseline type.
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.baseline.height
         * @memberof! oj.ojGantt
         * @instance
         * @type {number}
         * @ojsignature {target: "Type", value: "?"}
         * @ojunits pixels
         * @default null
         */
        height: null,
        /**
         * A space delimited list of CSS style classes defining the style of the baseline.
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.baseline.svgClassName
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @default ""
         */
        svgClassName: "",
        /**
         * The CSS style defining the style of the baseline.
         * <br></br>See the <a href="#taskDefaults">task-defaults</a> attribute for usage examples.
         * @expose
         * @name taskDefaults.baseline.svgStyle
         * @memberof! oj.ojGantt
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         * @default {}
         */
        svgStyle: {}
      }
    },
    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojGantt
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">tooltip</code> attribute specified:</caption>
     * &lt;oj-gantt tooltip.renderer='[[tooltipFun]]'>&lt;/oj-gantt>
     *
     * &lt;oj-gantt tooltip='[[{"renderer": tooltipFun}]]'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">tooltip</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.tooltip.renderer;
     * 
     * // Set one, leaving the others intact.
     * myGantt.setProperty('tooltip.renderer', tooltipFun);
     *
     * // Get all
     * var values = myGantt.tooltip;
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myGantt.tooltip = {'renderer': tooltipFun};
     */
    tooltip: {
      /**
       * A function that returns a custom tooltip. Note that the default is for a tooltip to be displayed.
       * <br></br>See the <a href="#tooltip">tooltip</a> attribute for usage examples.
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojGantt
       * @instance
       * @type {?(function(Object):Object)}
       * @ojsignature {target: "Type", value: "((context: oj.ojGantt.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))", jsdocOverride: true}
       * @default null
       */
      renderer: null
    },
    /**
     * An object specifying value formatting and tooltip behavior, whose keys generally correspond to task properties.
     * @expose
     * @name valueFormats
     * @memberof oj.ojGantt
     * @instance
     * @type {Object}
     * @default {"rows": {"tooltipDisplay": "auto"}, "start": {"converter": null, "tooltipDisplay": "auto"}, "end": {"converter": null, "tooltipDisplay": "auto"}, "date": {"converter": null, "tooltipDisplay": "auto"}, "label": {"tooltipDisplay": "auto"}, "progress": {"converter": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER).createConverter({style: 'percent'}), "tooltipDisplay": "auto"}, "baselineStart": {"converter": null, "tooltipDisplay": "auto"}, "baselineEnd": {"converter": null, "tooltipDisplay": "auto"}, "baselineDate": {"converter": null, "tooltipDisplay": "auto"}}
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">value-formats</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-gantt value-formats.row.tooltip-label="Employee" value-formats.label.tooltip-display="off">&lt;/oj-gantt>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-gantt value-formats='{"row": {"tooltipLabel": "Employee"}, "label": {"tooltipDisplay": "off"}}'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">valueFormats</code> property after initialization:</caption>
     * // Get one
     * var value = myGantt.valueFormats.row.tooltipLabel;
     *
     * // Set one, leaving the others intact
     * myGantt.setProperty('valueFormats.row.tooltipLabel', 'Employee');
     * 
     * // Get all
     * var values = myGantt.valueFormats;
     *
     * // Set all. Must list every resource key, as those not listed are lost.
     * myGantt.valueFormats = {
     *     "row": {"tooltipLabel": "Employee"},
     *     "label": {"tooltipDisplay": "off"}
     * };
     */
    valueFormats: {
      /**
       * Specifies tooltip behavior for the row value.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.row
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"tooltipDisplay": "auto"}
       */
      row: {
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelRow}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.row.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.row.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the start value.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.start
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"converter": null, "tooltipDisplay": "auto"}
       */
      start: {
        /**
         * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the label. If not specified, a default converter depending on the axes scale is used. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.start.converter
         * @memberof! oj.ojGantt
         * @instance
         * @type {?Object}
         * @ojsignature {target: "Type", value: "?(oj.Converter<string>)", jsdocOverride: true}
         * @default null
         */
        converter: null,
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelStart}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.start.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.start.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the end value.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.end
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"converter": null, "tooltipDisplay": "auto"}
       */
      end: {
        /**
         * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the label. If not specified, a default converter depending on the axes scale is used. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.end.converter
         * @memberof! oj.ojGantt
         * @instance
         * @type {?Object}
         * @ojsignature {target: "Type", value: "?(oj.Converter<string>)", jsdocOverride: true}
         * @default null
         */
        converter: null,
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelEnd}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.end.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.end.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the date value of a milestone task.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.date
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"converter": null, "tooltipDisplay": "auto"}
       */
      date: {
        /**
         * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the label. If not specified, a default converter depending on the axes scale is used. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.date.converter
         * @memberof! oj.ojGantt
         * @instance
         * @type {?Object}
         * @ojsignature {target: "Type", value: "?(oj.Converter<string>)", jsdocOverride: true}
         * @default null
         */
        converter: null,
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelDate}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.date.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.date.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the label value.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.label
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"tooltipDisplay": "auto"}
       */
      label: {
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelLabel}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.label.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.label.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the progress value.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.progress
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"converter": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER).createConverter({style: 'percent'}), "tooltipDisplay": "auto"}
       */
      progress: {
        /**
         * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the label. See {@link oj.NumberConverterFactory} for details on creating built-in number converters.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.progress.converter
         * @memberof! oj.ojGantt
         * @instance
         * @type {?Object}
         * @ojsignature {target: "Type", value: "?(oj.Converter<number>)", jsdocOverride: true}
         * @default oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER).createConverter({style: 'percent'})
         */
        converter: oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER).createConverter({style: 'percent'}),
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelProgress}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.progress.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.progress.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the start value of the baseline.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.baselineStart
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"converter": null, "tooltipDisplay": "auto"}
       */
      baselineStart: {
        /**
         * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the label. If not specified, a default converter depending on the axes scale is used. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineStart.converter
         * @memberof! oj.ojGantt
         * @instance
         * @type {?Object}
         * @ojsignature {target: "Type", value: "?(oj.Converter<string>)", jsdocOverride: true}
         * @default null
         */
        converter: null,
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelBaselineStart}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineStart.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineStart.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the end value of the baseline.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.baselineEnd
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"converter": null, "tooltipDisplay": "auto"}
       */
      baselineEnd: {
        /**
         * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the label. If not specified, a default converter depending on the axes scale is used. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineEnd.converter
         * @memberof! oj.ojGantt
         * @instance
         * @type {?Object}
         * @ojsignature {target: "Type", value: "?(oj.Converter<string>)", jsdocOverride: true}
         * @default null
         */
        converter: null,
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelBaselineEnd}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineEnd.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineEnd.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      },
      /**
       * Specifies tooltip behavior for the date value of the milestone baseline.
       * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
       * @expose
       * @name valueFormats.baselineDate
       * @memberof! oj.ojGantt
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"converter": null, "tooltipDisplay": "auto"}
       */
      baselineDate: {
        /**
         * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the label. If not specified, a default converter depending on the axes scale is used. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineDate.converter
         * @memberof! oj.ojGantt
         * @instance
         * @type {?Object}
         * @ojsignature {target: "Type", value: "?(oj.Converter<string>)", jsdocOverride: true}
         * @default null
         */
        converter: null,
        /**
         * A string representing the label that is displayed before the value in the tooltip. The default value comes from {@link oj.ojGantt.translations.labelBaselineDate}.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineDate.tooltipLabel
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojtranslatable
         */
        tooltipLabel: undefined,
        /**
         * Whether the value is displayed in the tooltip.
         * <br></br>See the <a href="#valueFormats">value-formats</a> attribute for usage examples.
         * @expose
         * @name valueFormats.baselineDate.tooltipDisplay
         * @memberof! oj.ojGantt
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        tooltipDisplay: "auto"
      }
    },
    /**
     * The end time of the Gantt's viewport. If not specified or invalid, this will default to a value determined by the initial 'scale' of the minor axis and the width of the Gantt. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name viewportEnd
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">viewport-end</code> attribute specified:</caption>
     * &lt;oj-gantt viewport-end='2017-12-31T05:00:00.000Z'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">viewportEnd</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.viewportEnd;
     *
     * // setter
     * myGantt.viewportEnd = '2017-12-31T05:00:00.000Z';
     */
    viewportEnd: "",
    /**
     * The start time of the Gantt's viewport. If not specified or invalid, this will default to a value determined by the initial 'scale' of the minor axis and the width of the Gantt. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name viewportStart
     * @memberof oj.ojGantt
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the Gantt with the <code class="prettyprint">viewport-start</code> attribute specified:</caption>
     * &lt;oj-gantt viewport-start='2017-01-01T05:00:00.000Z'>&lt;/oj-gantt>
     *
     * @example <caption>Get or set the <code class="prettyprint">viewportStart</code> property after initialization:</caption>
     * // getter
     * var value = myGantt.viewportStart;
     *
     * // setter
     * myGantt.viewportStart = '2017-01-01T05:00:00.000Z';
     */
    viewportStart: "",
    /**
     * Triggered after the viewport is changed due to a zoom or scroll operation.
     *
     * @property {string} viewportStart the start of the new viewport on a gantt chart
     * @property {string} viewportEnd the end of the new viewport on a gantt chart
     * @property {string} majorAxisScale the time scale of the majorAxis
     * @property {string} minorAxisScale the time scale of the minorAxis
     *
     * @expose
     * @event
     * @memberof oj.ojGantt
     * @instance
     * @ojbubbles
     */
    viewportChange: null
  },

  // @inheritdoc
  _CreateDvtComponent : function(context, callback, callbackObj) 
  {
    context['styleClasses'] = this._getComponentStyleMap();
    return dvt.Gantt.newInstance(context, callback, callbackObj);
  },

  // @inheritdoc
  _GetComponentStyleClasses : function() 
  {
    var styleClasses = this._super();
    styleClasses.push('oj-gantt');
    return styleClasses;
  },

  /**
   * @private
   */
  _getComponentStyleMap: function()
  {
    var map = new Object();
    map["databody"] = "oj-gantt-container";
    map['dependencyLine'] = "oj-gantt-dependency-line";
    map['dependencyLineConnector'] = "oj-gantt-dependency-line-connector";
    map["nodata"] = "oj-gantt-no-data-message";
    map["hgridline"] = "oj-gantt-horizontal-gridline";
    map["vgridline"] = "oj-gantt-vertical-gridline";
    map["majorAxis"] = "oj-gantt-major-axis";
    map["majorAxisTicks"] = "oj-gantt-major-axis-separator";
    map["majorAxisLabels"] = "oj-gantt-major-axis-label";
    map["minorAxis"] = "oj-gantt-minor-axis";
    map["minorAxisTicks"] = "oj-gantt-minor-axis-separator";
    map["minorAxisLabels"] = "oj-gantt-minor-axis-label";
    map['row'] = "oj-gantt-row";
    map['rowLabel'] = "oj-gantt-row-label";
    map['task'] = "oj-gantt-task";
    map['taskBar'] = "oj-gantt-task-bar";
    map['taskMilestone'] = "oj-gantt-task-milestone";
    map['taskSummary'] = "oj-gantt-task-summary";
    map['baseline'] = "oj-gantt-baseline";
    map['baselineBar'] = "oj-gantt-baseline-bar"
    map['baselineMilestone'] = "oj-gantt-baseline-milestone";
    map['taskLabel'] = "oj-gantt-task-label";
    map['taskProgress'] = "oj-gantt-task-progress";
    map['tooltipLabel'] = "oj-gantt-tooltip-label";
    map['tooltipValue'] = "oj-gantt-tooltip-value";
    map['tooltipTable'] = "oj-gantt-tooltip-content";
    map['referenceObject'] = "oj-gantt-reference-object";
    map['selected'] = "oj-selected";
    map['hover'] = "oj-hover";
    map['focus'] = "oj-focus";

    return map;
  },

  /**
   * @override
   * @instance
   * @memberof oj.ojGantt
   * @protected
   */
  _ConvertLocatorToSubId : function(locator) 
  {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-gantt-taskbar') 
    {
      // taskbar[rowIndex][index]
      subId = 'taskbar[' + locator['rowIndex'] + '][' + locator['index'] + ']';
    }
    else if (subId == 'oj-gantt-row-label')
    {
      // rowLabel[rowIndex]
      subId = 'rowLabel[' + locator['index'] + ']';
    }
    else if (subId == 'oj-gantt-tooltip') {
      subId = 'tooltip';
    }

    // Return the converted result or the original subId if a supported locator wasn't recognized.
    return subId;
  },

  /**
   * @override
   * @instance
   * @memberof oj.ojGantt
   * @protected
   */
  _ConvertSubIdToLocator : function(subId) 
  {
    var locator = {};

    if (subId.indexOf('taskbar') == 0) 
    {
      // taskbar[rowIndex][index]
      var indexPath = this._GetIndexPath(subId);

      locator['subId'] = 'oj-gantt-taskbar';
      locator['rowIndex'] = indexPath[0];
      locator['index'] = indexPath[1];
    }
    else if (subId.indexOf('rowLabel') == 0)
    {
      // rowLabel[rowIndex]
      var indexPath = this._GetIndexPath(subId);

      locator['subId'] = 'oj-gantt-row-label';
      locator['index'] = indexPath[0];
    }
    else if (subId == 'tooltip') {
      locator['subId'] = 'oj-gantt-tooltip';
    }

    return locator;
  },

  // @inheritdoc
  _GetChildStyleClasses: function()
  {
    var styleClasses = this._super();

    // animation duration
    styleClasses['oj-gantt'] = {'path': '_resources/animationDuration', 'property': 'ANIM_DUR'};

    // Zoom Control Icons
    styleClasses['oj-gantt-zoomin-icon'] = [
      {'path': '_resources/zoomIn_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-gantt-zoomin-icon oj-hover'] = [
      {'path': '_resources/zoomIn_h_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_h_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-gantt-zoomin-icon oj-active'] = [
      {'path': '_resources/zoomIn_a_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_a_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-gantt-zoomin-icon oj-disabled'] = [
      {'path': '_resources/zoomIn_d_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_d_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-gantt-zoomout-icon'] = [
      {'path': '_resources/zoomOut_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-gantt-zoomout-icon oj-hover'] = [
      {'path': '_resources/zoomOut_h_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_h_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-gantt-zoomout-icon oj-active'] = [
      {'path': '_resources/zoomOut_a_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_a_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-gantt-zoomout-icon oj-disabled'] = [
      {'path': '_resources/zoomOut_d_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_d_bc', 'property': 'border-color'}
    ];

    // Axes labels
    styleClasses['oj-gantt-major-axis-label'] = {'path': '_resources/majorAxisLabelFontProp', 'property': 'TEXT'};
    styleClasses['oj-gantt-minor-axis-label'] = {'path': '_resources/minorAxisLabelFontProp', 'property': 'TEXT'};

    // chart border
    styleClasses['oj-gantt-container'] = {'path': '_resources/chartArea/strokeWidth', 'property': 'stroke-width'};

    // horizontal gridline width
    styleClasses['oj-gantt-horizontal-gridline'] = {'path': '_resources/horizontalGridlineWidth', 'property': 'stroke-width'};

    // task label properties
    styleClasses['oj-gantt-task-label'] = {'path': '_resources/taskLabelFontProp', 'property': 'TEXT'};

    // row label properties
    styleClasses['oj-gantt-row-label'] = {'path': '_resources/rowLabelFontProp', 'property': 'TEXT'};

    return styleClasses;
  },

  // @inheritdoc
  _GetTranslationMap: function() 
  {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtUtilBundle.GANTT'] = translations['componentName'];
    ret['DvtUtilBundle.ZOOM_IN'] = translations['tooltipZoomIn'];
    ret['DvtUtilBundle.ZOOM_OUT'] = translations['tooltipZoomOut'];

    return ret;
  },

  // @inheritdoc
  _LoadResources: function() {
    this._super();

    var resources = this.options['_resources'];

    // zoom control icon images
    resources['zoomIn'] = 'oj-gantt-zoomin-icon';
    resources['zoomIn_h'] = 'oj-gantt-zoomin-icon oj-hover';
    resources['zoomIn_a'] = 'oj-gantt-zoomin-icon oj-active';
    resources['zoomIn_d'] = 'oj-gantt-zoomin-icon oj-disabled';
    resources['zoomOut'] = 'oj-gantt-zoomout-icon';
    resources['zoomOut_h'] = 'oj-gantt-zoomout-icon oj-hover';
    resources['zoomOut_a'] = 'oj-gantt-zoomout-icon oj-active';
    resources['zoomOut_d'] = 'oj-gantt-zoomout-icon oj-disabled';

    // progress value converter for task tooltip
    var converterFactory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER);
    resources['percentConverter'] = converterFactory.createConverter({style: 'percent'})
  },

  // @inheritdoc
  _HandleEvent: function(event)
  {
    var type = event['type'];
    if(type === 'viewportChange')
    {
      var viewportStart = new Date(event['viewportStart']).toISOString();
      var viewportEnd = new Date(event['viewportEnd']).toISOString();
      var majorAxisScale = event['majorAxisScale'];
      var minorAxisScale = event['minorAxisScale'];
      var viewportChangePayload = {
        'viewportStart': viewportStart,
        'viewportEnd': viewportEnd,
        'majorAxisScale': majorAxisScale,
        'minorAxisScale': minorAxisScale
      };

      this._UserOptionChange('viewportStart', viewportStart);
      this._UserOptionChange('viewportEnd', viewportEnd);
      this._UserOptionChange('majorAxis.scale', majorAxisScale);
      this._UserOptionChange('minorAxis.scale', minorAxisScale);
      this._trigger('viewportChange', null, viewportChangePayload);
    }
    else
    {
      this._super(event);
    }
  },

  // @inheritdoc
  _GetComponentNoClonePaths: function() {
    var noClonePaths = this._super();

    // Date time options as of 3.0.0 only support number and string types
    // e.g. Date object type is not supported. However,
    // during the options cloning,
    // Date objects are automatically converted to number by default.
    // We want to specify that they are to remain Date objects so that
    // we can handle them in our code later on. Note that data paths are not
    // cloned (see _GetComponentDeferredDataPaths)
    noClonePaths['start'] = true;
    noClonePaths['end'] = true;
    noClonePaths['viewportStart'] = true;
    noClonePaths['viewportEnd'] = true;
    noClonePaths['referenceObjects'] = {'value': true};
    return noClonePaths;
  },

  // @inheritdoc
  _GetComponentDeferredDataPaths : function() {
    return {'root': ['rows', 'dependencies']};
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   * @ojsignature {target:"Type", value:"{subId: 'oj-gantt-row-label', index: number} | {subId: 'oj-gantt-taskbar', rowIndex: number, index: number} | null", jsdocOverride: true, for:"returns"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojGantt
   */
  getContextByNode: function(node)
  {
    // context objects are documented with @ojnodecontext
    var context = this.getSubIdByNode(node);
    if (context && context['subId'] !== 'oj-gantt-tooltip')
      return context;

    return null;
  }
});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *    <tr>
 *       <td rowspan="3">Task</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selection-mode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="3">Chart Area</td>
 *       <td><kbd>Drag</kbd></td>
 *       <td>Paning: navigate forward and backward in time in horizontal/vertical orientation.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Pinch-close</kbd></td>
 *       <td>Zoom Out.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Spread-open</kbd></td>
 *       <td>Zoom In.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2">Zoom Control</td>
 *       <td><kbd>Tap on "+" element</kbd></td>
 *       <td>Zoom In.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Tap on "-" element</kbd></td>
 *       <td>Zoom Out.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojGantt
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>= or +</kbd></td>
 *       <td>Zoom in one level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>- or _</kbd></td>
 *       <td>Zoom out one level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>PageUp or PageDown</kbd></td>
 *       <td>Pan up / down.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + PageUp or PageDown</kbd></td>
 *       <td>Pan left/right (RTL: Pan right/left).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>When focus is on a task, move focus and selection to the task on the left within the same row.  In LTR reading direction, if this is the first task within the row, then move focus and selection to the last task in the previous row. In RTL reading direction, if this is the last task within the row, then move focus and selection to the first task in the next row.
 *           <br>When focus is on a dependency line (see <kbd>Alt + &lt;</kbd> and <kbd>Alt + &gt;</kbd>), move focus to the predecessor task (RTL: successor task).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>When focus is on a task, move focus and selection to the task on the right within the same row.  In LTR reading direction, if this is the last task within the row, then move focus and selection to the first task in the next row. In RTL reading direction, if this is the first task within the row, then move focus and selection to the last task in the previous row.
 *           <br>When focus is on a dependency line (see <kbd>Alt + &lt;</kbd> and <kbd>Alt + &gt;</kbd>), move focus to the successor task (RTL: predecessor task).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>When focus is on a task, move focus and selection to first task in the previous row.
 *           <br>When focus is on a dependency line (see <kbd>Alt + &lt;</kbd> and <kbd>Alt + &gt;</kbd>), move focus to the previous dependency line with the same
 *           predecessor/successor.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>When focus is on a task, move focus and selection to first task in the next row.
 *           <br>When focus is on a dependency line (see <kbd>Alt + &lt;</kbd> and <kbd>Alt + &gt;</kbd>), move focus to the next dependency line with the same
 *           predecessor/successor.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Space</kbd></td>
 *       <td>Multi-select task with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + &lt;task navigation shortcut&gt;</kbd></td>
 *       <td>Move focus and multi-select a task.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + &lt;task navigation shortcut&gt;</kbd></td>
 *       <td>Move focus to a task but do not select.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Alt + &lt;</kbd></td>
 *       <td>Move focus from a task to an associated dependency line connecting to a predecessor task (RTL: successor task). Note that the dependency line must have been 
 *        created referencing the task's ID in its predecessor/successorTask objects for an association to exist. Also note that when focus is on a dependency line, the <kbd>UpArrow</kbd> and <kbd>DownArrow</kbd> keys are used to move focus to the next dependency line with the same predecessor/successor.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Alt + &gt;</kbd></td>
 *       <td>Move focus from a task to an associated dependency line connecting to a successor task (RTL: predecessor task). Note that the dependency line must have been 
 *        created referencing the task's ID in its predecessor/successorTask objects for an association to exist. Also note that when focus is on a dependency line, the <kbd>UpArrow</kbd> and <kbd>DownArrow</kbd> keys are used to move focus to the next dependency line with the same predecessor/successor.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Mousewheel Up</kbd></td>
 *       <td>Zoom In.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Mousewheel Down</kbd></td>
 *       <td>Zoom Out.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojGantt
 */

/**
 *<p>The Gantt supports a simplified version of the ISO 8601 extended date/time format. The format is as follows: <font color="#4B8A08">YYYY-MM-DDTHH:mm:ss.sssZ</font></p>
 *<table  class="keyboard-table">
 *<thead>
 *<tr>
 *<th>Symbol</th>
 *<th>Description</th>
 *<th>Values</th>
 *<th>Examples</th>
 *</tr>
 *</thead>
 *<tbody>
 *<tr>
 *<td><font color="#4B8A08">-, :, .,T</font></td><td>Characters actually in the string. T specifies the start of a time.</td><td></td><td></td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">YYYY</font></td><td>Year</td><td></td><td rowspan="3">2013-03-22<br>2014-02</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">MM</font></td><td>Month</td><td>01 to 12</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">DD</font></td><td>Day of the month</td><td>01 to 31</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">HH</font></td><td>Hours</td><td>00 to 24</td><td rowspan="3">2013-02-04T15:20Z<br>2013-02-10T15:20:45.300Z</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">mm</font></td><td>Minutes</td><td>00 to 59</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">ss</font></td><td>Seconds. The seconds and milliseconds are optional if a time is specified.</td><td>00 to 59</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">sss</font></td><td>Milliseconds</td><td>00 to 999</td><td></td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">Z</font></td><td>The value in this position can be one of the following. If the value is omitted, character 'Z' should be used to specify UTC time.<br><ul><li><b>Z</b> indicates UTC time.</li><li><b>+hh:mm</b> indicates that the input time is the specified offset after UTC time.</li><li><b>-hh:mm</b> indicates that the input time is the absolute value of the specified offset before UTC time.</li></ul></td><td></td><td>2013-02-04T15:20:00-07:00<br>2013-02-04T15:20:00+05:00<br>2013-02-04T15:20:00Z</td>
 *</tr>
 *</tbody>
 *</table>
 *<p>The ISO format support short notations where the string must only include the date and not time, as in the following formats: YYYY, YYYY-MM, YYYY-MM-DD.</p>
 *<p>The ISO format does not support time zone names. You can use the Z position to specify an offset from UTC time. If you do not include a value in the Z position, UTC time is used. The correct format for UTC should always include character 'Z' if the offset time value is omitted. The date-parsing algorithms are browser-implementation-dependent and, for example, the date string '2013-02-27T17:00:00' will be parsed differently in Chrome vs Firefox vs IE.</p>
 *<p>You can specify midnight by using 00:00, or by using 24:00 on the previous day. The following two strings specify the same time: 2010-05-25T00:00Z and 2010-05-24T24:00Z.</p>
 *
 * @ojfragment formatsDoc
 * @memberof oj.ojGantt
 */

// PROPERTY TYPEDEFS

/**
 * @typedef {Object} oj.ojGantt.Dependency
 * @property {string} id The identifier for the dependency line. This must be unique across all dependency lines in Gantt.
 * @property {string} predecessorTaskId The identifier for the predecessor task. This must reference a task in Gantt.
 * @property {string=} shortDesc The description of the dependency line. This is used for accessibility.
 * @property {string} successorTaskId The identifier for the successor task. This must reference a task in Gantt.
 * @property {string=} svgClassName A space delimited list of class name set on the dependency line.
 * @property {Object=} svgStyle The CSS style defining the style of the dependency line.
 * @property {"startStart"|"startFinish"|"finishFinish"|"finishStart"} [type="finishStart"] The type of dependency. The following values are supported: <ul> <li>finishStart: predecessor task must finish before successor task can start.</li> <li>finishFinish: predecessor task must finish before successor task can finish.</li> <li>startStart: predecessor task must start before successor task can start.</li> <li>startFinish: predecessor task must start before successor task can finish.</li> </ul>
 */
/**
 * @typedef {Object} oj.ojGantt.ReferenceObject
 * @property {string=} svgClassName A space delimited list of CSS style classes defining the style of the reference object. Note that only CSS style applicable to SVG elements can be used.
 * @property {Object=} svgStyle The CSS style defining the style of the reference object.
 * @property {string=} value The time value of this reference object. If not specified, no reference object will be shown. See <a href="#formats-section">Date and Time Formats</a> for more details on required string formats.
 */
/**
 * @typedef {Object} oj.ojGantt.Row
 * @property {string=} id The identifier for the row. Optional if the row contains only one task. This must be unique across all rows in Gantt.
 * @property {string=} label The label associated with the row.
 * @property {Object=} labelStyle The CSS style defining the style of the label. Only CSS style applicable to SVG elements can be used.
 * @property {Array.<oj.ojGantt.RowTask>} [tasks] An array of tasks. If not specified, no data will be shown. When only one of 'start' or 'end' value is specified, or when 'start' and 'end' values are equal, the task is considered a milestone task. Note that values of analogous properties from <a href="#taskDefaults">task-defaults</a> are used for any unspecified properties on the task, and values of any specified properties would override those from <a href="#taskDefaults">task-defaults</a>.
 */
/**
 * @typedef {Object} oj.ojGantt.RowTask
 * @property {string=} borderRadius The border radius of the task. Accepts values allowed in CSS border-radius attribute. The default value comes from <a href="#taskDefaults.borderRadius">task-defaults.border-radius</a>.
 * @property {string=} end The end time of this task. Optional if task is a single date event like Milestone. Either start or end has to be defined in order for the task to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
 * @property {number=} height The height of the task in pixel. The default value comes from <a href="#taskDefaults.height">task-defaults.height</a>.
 * @property {string} id The identifier for the task. This must be unique across all tasks in the Gantt, and is required in order for the Gantt to properly render.
 * @property {string=} label The label associated with the task.
 * @property {(string|Array.<string>)=} labelPosition The position of the label relative to the task. An array of values is also supported. If an array is specified, then the values are traversed until a position that can fully display the label is found. If 'max' is specified in the array, then of all the positions evaluated up to that point of the traversal, the one with the largest space is used (label is truncated to fit). Naturally, 'max' is ignored if it's specified as the first value of the array. If the last value of the array is reached, but the label cannot be fully displayed, then the label is placed at that position, truncated to fit. Due to space constraints in the milestone and task with progress cases, the inner positions will exhibit the following behaviors: <ul> <li> For milestones, specifying 'innerStart', 'innerEnd', or 'innerCenter' would be equivalent to specifying 'start', 'end', and 'end' respectively. </li> <li> For tasks with progress, 'innerCenter' means the label will be aligned to the end of the progress bar, either placed inside or outside of the progress, whichever is the larger space. 'innerStart' and 'innerEnd' positions are honored when there is enough space to show the label at those positions. Otherwise, the aforementioned 'innerCenter' behavior is exhibited. </li> </ul> The default value comes from <a href="#taskDefaults.labelPosition">task-defaults.label-position</a>.
 * @property {Object=} labelStyle The CSS style defining the style of the label. Only CSS style applicable to SVG elements can be used.
 * @property {string=} start The start time of this task. Optional if task is a single date event like Milestone. Either start or end has to be defined in order for the task to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
 * @property {string=} shortDesc The description of the task. This is used for accessibility and for customizing the tooltip text.
 * @property {string=} svgClassName A space delimited list of CSS style classes defining the style of the task. The default value comes from <a href="#taskDefaults.svgClassName">task-defaults.svg-class-name</a>.
 * @property {Object=} svgStyle The CSS style defining the style of the task. The default value comes from <a href="#taskDefaults.svgStyle">task-defaults.svg-style</a>.
 * @property {("normal"|"milestone"|"summary"|"auto")=} type Defines the task type to be rendered.<br></br>If "milestone", and if 'start' and 'end' values are specified and unequal, the 'start' value is used to evaluate position.<br></br>If "auto", the type is inferred from the data:<ul> <li>If 'start' and 'end' values are specified and unequal, "normal" type is assumed.</li> <li>Otherwise, "milestone" type is assumed.</li></ul> The default value comes from <a href="#taskDefaults.type">task-defaults.type</a>.
 * @property {Object=} progress Specifies the progress of the task. This property is ignored if the task is a milestone. The default value comes from <a href="#taskDefaults.progress">task-defaults.progress</a>.
 * @property {string=} progress.borderRadius The border radius of the progress bar. Accepts values allowed in CSS border-radius attribute. The default value comes from <a href="#taskDefaults.progress.borderRadius">task-defaults.progress.border-radius</a>.
 * @property {number=} progress.height Specifies the height of the progress bar in pixels (e.g. '50px') or percent of the associated task bar (e.g. '15%'). The default value comes from <a href="#taskDefaults.progress.height">task-defaults.progress.height</a>.
 * @property {string=} progress.svgClassName A space delimited list of CSS style classes to apply to the progress bar. Note that only CSS style applicable to SVG elements can be used. The default value comes from <a href="#taskDefaults.progress.svgClassName">task-defaults.progress.svg-class-name</a>.
 * @property {Object=} progress.svgStyle The CSS inline style to apply to the progress bar. Only CSS style applicable to SVG elements can be used. The default value comes from <a href="#taskDefaults.progress.svgStyle">task-defaults.progress.svg-style</a>.
 * @property {number=} progress.value The value of the progress between 0 and 1 inclusive. If not specified or invalid, no progress will be shown.
 * @property {Object=} baseline Specifies the baseline of the task. When only one of 'start' or 'end' value is specified, or when 'start' and 'end' values are equal, the baseline is considered a milestone baseline. The default value comes from <a href="#taskDefaults.baseline">task-defaults.baseline</a>.
 * @property {string=} baseline.borderRadius The border radius of the baseline. Accepts values allowed in CSS border-radius attribute. The default value comes from <a href="#taskDefaults.baseline.borderRadius">task-defaults.baseline.border-radius</a>.
 * @property {string=} baseline.end The end time of the baseline. Optional if baseline is a milestone. Either start or end has to be defined in order for the baseline to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
 * @property {number=} baseline.height The height of the baseline in pixel. The default value comes from <a href="#taskDefaults.baseline.height">task-defaults.baseline.height</a>.
 * @property {string=} baseline.start The start time of the baseline. Optional if baseline is a milestone. Either start or end has to be defined in order for the baseline to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
 * @property {string=} baseline.svgClassName A space delimited list of CSS style classes defining the style of the baseline. The default value comes from <a href="#taskDefaults.baseline.svgClassName">task-defaults.baseline.svg-class-name</a>.
 * @property {Object=} baseline.svgStyle The CSS style defining the style of the baseline. The default value comes from <a href="#taskDefaults.baseline.svgStyle">task-defaults.baseline.svg-style</a>.
 */
/**
 * @typedef {Object} oj.ojGantt.TooltipContext
 * @property {Element} parentElement The tooltip element. This can be used to change the tooltip border or background color.
 * @property {Object} data The data object of the hovered task.
 * @property {Object} rowData The data for the row the hovered task belongs to.
 * @property {Element} componentElement The gantt element.
 * @property {string} color The color of the hovered task.
 */

// KEEP FOR WIDGET SYNTAX

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for Gantt task (including milestone) at a specified index.</p>
 *
 * @property {number} rowIndex
 * @property {number} index
 *
 * @ojsubid oj-gantt-taskbar
 * @memberof oj.ojGantt
 *
 * @example <caption>Get the second task of the first row:</caption>
 * var nodes = myGantt.getNodeBySubId({'subId': 'oj-gantt-taskbar', 'rowIndex': 0, 'index': 1});
 */

 /**
 * <p>Sub-ID for Gantt row label at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojsubid oj-gantt-row-label
 * @memberof oj.ojGantt
 *
 * @example <caption>Get the label of the first row:</caption>
 * var nodes = myGantt.getNodeBySubId({'subId': 'oj-gantt-row-label', 'index': 0});
 */

 /**
 * <p>Sub-ID for the the Gantt tooltip.</p>
 *
 * @ojsubid oj-gantt-tooltip
 * @memberof oj.ojGantt
 *
 * @example <caption>Get the tooltip object of the gantt, if displayed:</caption>
 * var nodes = myGantt.getNodeBySubId({'subId': 'oj-gantt-tooltip'});
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for Gantt task at a specified index.</p>
 *
 * @property {number} rowIndex
 * @property {number} index
 *
 * @ojnodecontext oj-gantt-taskbar
 * @memberof oj.ojGantt
 */

/**
 * <p>Context for Gantt row label at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojnodecontext oj-gantt-row-label
 * @memberof oj.ojGantt
 */
/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojGanttMeta = {
  "properties": {
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "axisPosition": {
      "type": "string",
      "enumValues": ["bottom", "top"]
    },
    "dependencies": {
      "type": "Array<object>|Promise"
    },
    "end": {
      "type": "string"
    },
    "gridlines": {
      "type": "object",
      "properties": {
        "horizontal": {
          "type": "string",
          "enumValues": ["hidden", "visible", "auto"]
        },
        "vertical": {
          "type": "string",
          "enumValues": ["hidden", "visible", "auto"]
        }
      }
    },
    "majorAxis": {
      "type": "object",
      "properties": {
        "converter": {
          "type": "object",
          "properties": {
            "days": {},
            "default": {},
            "hours": {},
            "minutes": {},
            "months": {},
            "quarters": {},
            "seconds": {},
            "weeks": {},
            "years": {}
          }
        },
        "scale": {
          "type": "string",
          "enumValues": ["seconds", "minutes", "hours", "days", "weeks", "months", "quarters", "years"]
        },
        "zoomOrder": {
          "type": "Array<string>"
        }
      }
    },
    "minorAxis": {
      "type": "object",
      "properties": {
        "converter": {
          "type": "object",
          "properties": {
            "days": {},
            "default": {},
            "hours": {},
            "minutes": {},
            "months": {},
            "quarters": {},
            "seconds": {},
            "weeks": {},
            "years": {}
          }
        },
        "scale": {
          "type": "string",
          "enumValues": ["seconds", "minutes", "hours", "days", "weeks", "months", "quarters", "years"]
        },
        "zoomOrder": {
          "type": "Array<string>"
        }
      }
    },
    "referenceObjects": {
      "type": "Array<object>"
    },
    "rowAxis": {
      "type": "object",
      "properties": {
        "maxWidth": {
          "type": "string"
        },
        "rendered": {
          "type": "string",
          "enumValues": ["on", "off"]
        }
      }
    },
    "rows": {
      "type": "Array<object>|Promise"
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["single", "multiple", "none"]
    },
    "start": {
      "type": "string"
    },
    "taskDefaults": {
      "type": "object",
      "properties": {
        "borderRadius": {
          "type": "string"
        },
        "height": {
          "type": "number"
        },
        "labelPosition": {
          "type": "Array<string>|string",
          "enumValues": ["start", "innerCenter", "innerStart", "innerEnd", "none", "end"]
        },
        "svgClassName": {
          "type": "string"
        },
        "svgStyle": {
          "type": "object"
        },
        "progress": {
          "type": "object",
          "properties": {
            "borderRadius": {
              "type": "string"
            },
            "height": {
              "type": "number"
            },
            "svgClassName": {
              "type": "string"
            },
            "svgStyle": {
              "type": "object"
            }
          }
        },
        "baseline": {
          "type": "object",
          "properties": {
            "borderRadius": {
              "type": "string"
            },
            "height": {
              "type": "number"
            },
            "svgClassName": {
              "type": "string"
            },
            "svgStyle": {
              "type": "object"
            }
          }
        },
        "type": {
          "type": "string",
          "enumValues": ["normal", "milestone", "summary", "auto"]
        }
      }
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "accessibleDependencyInfo": {
          "type": "string",
          "value": "Dependency type {0}, connects {1} to {2}"
        },
        "accessiblePredecessorInfo": {
          "type": "string",
          "value": "{0} predecessors"
        },
        "accessibleSuccessorInfo": {
          "type": "string",
          "value": "{0} successors"
        },
        "accessibleTaskTypeMilestone": {
          "type": "string",
          "value": "Milestone"
        },
        "accessibleTaskTypeSummary": {
          "type": "string",
          "value": "Summary"
        },
        "componentName": {
          "type": "string",
          "value": "Gantt"
        },
        "finishFinishDependencyAriaDesc": {
          "type": "string",
          "value": "finish to finish"
        },
        "finishStartDependencyAriaDesc": {
          "type": "string",
          "value": "finish to start"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelBaselineDate": {
          "type": "string",
          "value": "Baseline Date"
        },
        "labelBaselineEnd": {
          "type": "string",
          "value": "Baseline End"
        },
        "labelBaselineStart": {
          "type": "string",
          "value": "Baseline Start"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelDate": {
          "type": "string",
          "value": "Date"
        },
        "labelEnd": {
          "type": "string",
          "value": "End"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelLabel": {
          "type": "string",
          "value": "Label"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "labelProgress": {
          "type": "string",
          "value": "Progress"
        },
        "labelRow": {
          "type": "string",
          "value": "Row"
        },
        "labelStart": {
          "type": "string",
          "value": "Start"
        },
        "startFinishDependencyAriaDesc": {
          "type": "string",
          "value": "start to finish"
        },
        "startStartDependencyAriaDesc": {
          "type": "string",
          "value": "start to start"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        },
        "tooltipZoomIn": {
          "type": "string",
          "value": "Zoom In"
        },
        "tooltipZoomOut": {
          "type": "string",
          "value": "Zoom Out"
        }
      }
    },
    "valueFormats": {
      "type": "object",
      "properties": {
        "row": {
          "type": "object",
          "properties": {
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "start": {
          "type": "object",
          "properties": {
            "converter": {},
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "end": {
          "type": "object",
          "properties": {
            "converter": {},
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "date": {
          "type": "object",
          "properties": {
            "converter": {},
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "label": {
          "type": "object",
          "properties": {
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "progress": {
          "type": "object",
          "properties": {
            "converter": {},
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "baselineStart": {
          "type": "object",
          "properties": {
            "converter": {},
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "baselineEnd": {
          "type": "object",
          "properties": {
            "converter": {},
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        },
        "baselineDate": {
          "type": "object",
          "properties": {
            "converter": {},
            "tooltipDisplay": {
              "type": "string",
              "enumValues": ["auto", "off"]
            },
            "tooltipLabel": {
              "type": "string"
            }
          }
        }
      }
    },
    "viewportEnd": {
      "type": "string"
    },
    "viewportStart": {
      "type": "string"
    }
  },
  "methods": {
    "getContextByNode": {}
  },
  "events": {
    "viewportChange": {}
  },
  "extension": {
    _WIDGET_NAME: "ojGantt"
  }
};
oj.CustomElementBridge.registerMetadata('oj-gantt', 'dvtTimeComponent', ojGanttMeta);
oj.CustomElementBridge.register('oj-gantt', {'metadata': oj.CustomElementBridge.getMetadata('oj-gantt')});
})();
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Implementation of the IndexerModel and TreeDataSource based on an array of data set.
 * This should be used with the Indexer and its associated ListView.
 * By default, this adapter groups the data based on the first letter of the data and the alphabet of the current locale.
 * @class oj.IndexerModelTreeDataSource
 * @classdesc TreeDataSource and IndexerModel implementation that represents hierachical data available from an array of JSON objects.  This data source can be used by [Indexer]{@link oj.ojIndexer} and 
 *            associated [ListView]{@link oj.ojListView}.<br><br> 
 *            See the <a href="../jetCookbook.html?component=indexer&demo=characterIndexer">Indexer - Basic</a> demo for an example.<br><br>
 *            Refer to {@link oj.TreeDataSource} for other data sources that represent hierarachical data.
 * @param {Array.<*>} data an array of data used for Indexer and ListView 
 * @param {string} idAttribute the id attribute of the data
 * @param {function(Object)|function(string)} listener a callback function that handles when a section becomes current (user clicks on the section in the Indexer).
 *                   the function takes the current section and must return a Promise which when resolve returns the section in which the ListView
 *                   actually scrolls to.
 * @param {Object=} options the options set on this IndexerModelTreeDataSource
 * @param {(Array.<string>|Array.<Object>)=} options.sections the set of sections to use with the Indexer.  If not specified, then the sections are derived from the alphabet of the current locale.
 * @param {string=} options.groupingAttribute the attribute of the data where grouping is based on, mandatory if no groupingStrategy is specified.
 * @param {function(Object)=} options.groupingStrategy a callback function that takes a data and returns the section that the data belongs to.  If no groupingStrategy is specified,
 *                   then the default grouping strategy based on the first letter of the data is used.
 * @param {function(Object, Object)=} options.sortComparatorFunction a comparator function that is used to sort data within a section.  If not specified, then 
 *                   the items within a section will not be sorted.
 * @ojsignature {target: "Type",
 *               value: "(section: string|object)=> Promise<string|object>",
 *               for: "listener"}
 * @ojsignature {target: "Type",
 *               value: "(data: any)=> string|object",
 *               for: "options.groupingStrategy"}
 * @ojsignature {target: "Type",
 *               value: "(section1: string|object, section2: string|object)=> number",
 *               for: "options.sortComparatorFunction"}
 * @constructor
 * @since 3.0
 * @export
 * @ojtsignore
 * @extends oj.TreeDataSource
 * @implements oj.IndexerModel
 */
oj.IndexerModelTreeDataSource = function(data, idAttribute, listener, options)
{
    this.data = data;
    this.idAttr = idAttribute;
    this.listener = listener;
    this.options = options;

    this.Init();
}

// Subclass from oj.TreeDataSource
oj.Object.createSubclass(oj.IndexerModelTreeDataSource, oj.TreeDataSource, "oj.IndexerModelTreeDataSource");

/**
 * Initializes the instance.
 * @return {void}
 * @export
 * @expose
 * @memberof oj.IndexerModelTreeDataSource
 * @ojtsignore
 * @instance
 */
oj.IndexerModelTreeDataSource.prototype.Init = function()
{
    var sections, resource, strategy, field, content, char, sortComparator, section1, section2, index1, index2, diff, pos, current, i, section;

    // call super
    oj.IndexerModelTreeDataSource.superclass.Init.call(this);

    sections = this.options['sections'];
    if (sections == null)
    {
        resource = oj.Translations.getTranslatedString('oj-ojIndexer.indexerCharacters');
        sections = resource.split('|');
    }
    sections.push(oj.IndexerModel.SECTION_OTHERS);

    strategy = this.options['groupingStrategy'];
    if (strategy == null)
    {
        // if no grouping strategy is specified, use the default which groups based on the first charactor 
        // of the data
        field = this.options['groupingAttribute'];
        strategy = function(value) 
        {
            content = value[field] ? value[field] : value;
            char = content.toString().toUpperCase()[0];
            return sections.indexOf(char) > -1 ? char : oj.IndexerModel.SECTION_OTHERS;
        };
    }

    sortComparator = this.options['sortComparatorFunction'];
    this.data = this.data.sort(function(a, b) 
    {
        section1 = strategy(a);
        section2 = strategy(b);

        index1 = sections.indexOf(section1);
        index2 = sections.indexOf(section2);

        diff = index1 - index2; 
        // used specified sort comparator function to sort within section
        if (diff == 0 && sortComparator)
        {
            return sortComparator(a, b);
        }
        else
        {
            return diff;
        }        
    });

    // figure out the position of the buckets for quick lookup
    for (i=0; i<this.data.length; i++)
    {
        section = strategy(this.data[i]);
        if (current != section)
        {
            current = section;
            this._set(section, i);
        }
    }

    this.sections = sections;
};

// TODO: use Map when we can use it for all supported platforms
/**
 * @private
 */
oj.IndexerModelTreeDataSource.prototype._set = function(key, value)
{
    if (this.pos == null)
    {
        this.pos = [];
    }

    this.pos.push({key: key, value: value});
};

/**
 * @private
 */
oj.IndexerModelTreeDataSource.prototype._get = function(key)
{
    var i;

    for (i=0; i<this.pos.length; i++)
    {
        if (this.pos[i].key == key)
        {
            return this.pos[i].value;
        }
    }
    return null;
};

/****************** IndexerModel *******************/
/**
 * Returns an array of objects each representing a section in the associated ListView.  
 * @return {Array.<string>|Array.<Object>} an array of all indexable sections 
 * @export
 * @expose
 * @memberof oj.IndexerModelTreeDataSource
 * @instance
 */
oj.IndexerModelTreeDataSource.prototype.getIndexableSections = function()
{
    // remove other sections from this.sections
    return this.sections.slice(0, this.sections.length-1);
};

/**
 * Returns an array of objects each representing a section that does not have a corresponding section in the associated ListView.  
 * It must be a subset of the return value of <code>getIndexableSections</code>.  Return null or undefined if there's nothing missing.
 * @return {Array.<string>|Array.<Object>} an array of missing sections
 * @export
 * @expose
 * @memberof oj.IndexerModelTreeDataSource
 * @instance
 */
oj.IndexerModelTreeDataSource.prototype.getMissingSections = function()
{
    var missing, i, section;

    if (this.missing == null)
    {
        missing = [];
        // figure out what's missing, skip the others section since it's always available
        for (i=0; i<this.sections.length-1; i++)
        {
            section = this.sections[i];
            if (this._get(section) == null)
            {
                missing.push(section);
            }
        }

        this.missing = missing;
    }

    return this.missing;
};

/**
 * Make a section current in the Indexer.  The implementation should scroll the associated ListView so that the section becomes visible.
 * @param {string|Object} section the current section
 * @return {Promise.<string>|Promise.<Object>} a Promise which when resolved will return the section that the associated ListView actually scrolls to.  
 *                   For example, the implementation could choose to scroll to the next available section in ListView if no data 
 *                   exists for that section.  
 * @export
 * @expose
 * @memberof oj.IndexerModelTreeDataSource
 * @instance
 */
oj.IndexerModelTreeDataSource.prototype.setSection = function(section)
{
    return this.listener.call(this, section);
};

/****************** TreeDataSource *******************/
/**
 * Returns the number of children for a specified parent.  If the value returned is not >= 0 then it is automatically assumed
 * that the child count is unknown.
 * @param {*} parentKey the parent key.  Specify null if inquiring child count of the root.
 * @return {number} the number of children for the specified parent.
 * @export
 * @expose
 * @memberof oj.IndexerModelTreeDataSource
 * @instance
 */
oj.IndexerModelTreeDataSource.prototype.getChildCount = function(parentKey)
{
    var count, pos, index, next, nextPos;

    if (parentKey == null)
    {
        count = this.sections.length - this.getMissingSections().length;
        // exclude other section if there are no entries in it
        return this._get(oj.IndexerModel.SECTION_OTHERS) != null ? count : count - 1;
    }
    else
    {
        pos = this._get(parentKey);
        if (pos != null)
        {
            index = this.sections.indexOf(parentKey);
            // if it's the last section
            if (index == this.sections.length - 1)
            {
                return this.data.length - pos;
            }
            else
            {
                index++;
                next = this.sections[index];
                nextPos = this._get(next);
                while (nextPos == null && index < this.sections.length)
                {
                    index++;
                    next = this.sections[index];
                    nextPos = this._get(next);
                }

                if (isNaN(nextPos) || nextPos == null)
                {
                    nextPos = this.data.length;
                }

                return nextPos - pos;
            }
        }
    }

    // should not get here (ex: unexpected parentKey)
    return 0;
};

/**
 * Fetch the children
 * @param {*} parentKey the parent key.  Specify null if fetching children from the root.
 * @param {Object} range information about the range, it must contain the following properties: start, count.
 * @param {number} range.start the start index of the range in which the children are fetched.
 * @param {number} range.count the size of the range in which the children are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch children operation is completed.  The valid callback
 *        types are "success" and "error".
 * @param {function(oj.NodeSet)} callbacks.success the callback to invoke when fetch completed successfully.
 * @param {function({status: *})=} callbacks.error the callback to invoke when fetch children failed.
 * @param {Object=} options optional parameters for this operation.
 * @param {boolean=} options.queueOnly true if this fetch request is to be queued and not execute yet.  The implementation must maintain 
 *        the order of the fetch operations.  When queueOnly is false/null/undefined, any queued fetch operations are then
 *        flushed and executed in the order they are queued.  This flag is ignored if the datasource does not support batching.
 * @return {void}
 * @export
 * @expose
 * @memberof oj.IndexerModelTreeDataSource
 * @instance
 */
oj.IndexerModelTreeDataSource.prototype.fetchChildren = function(parentKey, range, callbacks, options)
{
    var available, nodeSet, self = this;

    if (parentKey == null)
    {
        // root case
        available = this.sections.filter(function(section) 
        {
            return self._get(section) != null;
        });

        nodeSet = {
            "getParent": function() {
                return parentKey;
            },
            "getStart": function() {
                return 0;
            },
            "getCount": function() {
                return available.length;
            },
            "getData": function(index) {
                return available[index];
            },
            "getMetadata": function(index) {
                return {"key": available[index], "leaf": false, "depth": 0};
            }
        };
    }
    else
    {
        nodeSet = {
            "getParent": function() {
                return parentKey;
            },
            "getStart": function() {
                return 0;
            },
            "getCount": function() {
                return self.getChildCount(parentKey);
            },
            "getData": function(index) {
                return self.data[self._get(parentKey) + index];
            },
            "getMetadata": function(index) {
                var data = self.data[self._get(parentKey) + index];
                return {"key": self.idAttr != null ? data[self.idAttr] : data, "leaf": true, "depth": 1};
            }
        };
    }

    // invoke original success callback
    if (callbacks != null && callbacks['success'] != null)
    {
        callbacks['success'].call(null, nodeSet);
    }
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

// jeanne retrieved from https://raw.github.com/jquery/jquery-ui/1-10-stable/ui/jquery.ui.spinner.js on 6/2013, and then modified

/*!
 * JET InputNumber @VERSION
 *
 *
 * Depends:
 *  jquery.ui.widget.js
 */
(function() // inputNumber wrapper function, to keep "private static members" private
{
/**
 * @ojcomponent oj.ojInputNumber 
 * @augments oj.editableValue
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojInputNumber extends editableValue<number, ojInputNumberSettableProperties, number, string>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojInputNumberSettableProperties extends editableValueSettableProperties<number, number, string>",
 *                for: "SettableProperties"
 *               }
 *              ]
 * @since 0.6
 * @ojshortdesc Provides basic support for specifying a number value.
 * @ojrole textbox
 * @ojrole spinbutton
 * @ojstatus preview
 * @ojtsimport ojvalidation-base
 * @ojtsimport ojvalidation-number
 * @classdesc
 * <h3 id="inputNumberOverview-section">
 *   JET InputNumber Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#inputNumberOverview-section"></a>
 * </h3>
 * <p>Description: The oj-input-number component enhances a browser input element
 * into one that holds numbers and it has a spinbox to quickly increment or
 * decrement the number. The <code class="prettyprint">value</code> attribute must be a number and must
 * be within the <code class="prettyprint">min</code> and <code class="prettyprint">max</code> range.
 * </p>
 * <p>A step mismatch is when
 * the value is not a multiple of <code class="prettyprint">step</code>,
 * starting at the <code class="prettyprint">min</code>
 * else initial vlaue if no <code class="prettyprint">min</code> is set, else 0.
 * A step mismatch will not be flagged as a validation error by default, but
 * the step up and step down feature will change the value to be a step match
 * if it isn't already.
 * </p>
 *
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * <p>
 * The component is accessible; it sets and maintains the appropriate aria- attributes,
 * like aria-valuenow, aria-valuemax, aria-valuemin and aria-valuetext.
 * </p>
 * <p>Disabled content: JET supports an accessible luminosity contrast ratio,
 * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>,
 * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which
 * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user
 * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio
 * required of enabled content, it cannot be used to convey meaningful information.<p>
 * 
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 *
 * <h3 id="label-section">
 *   Label and InputNumber
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate the oj-label to the oj-input-number component.
 * For accessibility, you should associate a oj-label element with the oj-input-number component
 * by putting an <code class="prettyprint">id</code> on the oj-input-number element, and then setting the
 * <code class="prettyprint">for</code> attribute on the oj-label to be the component's id.
 * </p>
 *
 * @example <caption>Declare the oj-input-number component with no attributes specified:</caption>
 * &lt;oj-input-number>&lt;/oj-input-number>
 *
 * @example <caption>Initialize the component with some attributes:</caption>
 * &lt;oj-input-number id="numberId" max="100" min="0" step="2">&lt;/oj-input-number>
 *
 * @example <caption>Initialize a component attribute via component binding:</caption>
 * &lt;oj-input-number id="numberId" value="{{currentValue}}">&lt;/oj-input-number>
 */
oj.__registerWidget("oj.ojInputNumber", $['oj']['editableValue'],
{
  version: "1.0.0",
  defaultElement: "<input>",
  widgetEventPrefix: "oj",
  
  /**
   * @private
   */
  _ALLOWED_TYPES : ['number', 'text'],

  options:
  {
    /** 
     * Dictates component's autocomplete state. 
     * This attribute indicates whether the value of the control can be automatically 
     * completed by the browser.
     * 
     * @example <caption>Initialize component with <code class="prettyprint">autocomplete</code> attribute:</caption>
     * &lt;oj-input-number autocomplete = "on">&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">autocomplete</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.autocomplete;
     *
     * // setter
     * myComp.autocomplete = "on";
     * 
     * @expose 
     * @type {string|undefined}
     * @ojvalue {string} "on" enable autofill
     * @ojvalue {string} "off" disable autofill
     * @alias autocomplete
     * @default "off"
     * @instance
     * @access public
     * @memberof oj.ojInputNumber
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    autocomplete: undefined,
     /** 
     * Autofocus is a Boolean that reflects the autofocus attribute, If it is set to true 
     * then the associated component  will get input focus when the page is loaded.
     * Setting this property doesn't set the focus to the component: 
     * it tells the browser to focus to it when the element is inserted in the document. 
     * 
     * @example <caption>Initialize component with <code class="prettyprint">autofocus</code> attribute:</caption>
     * &lt;oj-input-number autofocus>&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">autofocus</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.autofocus;
     *
     * // setter
     * myComp.autofocus = false;
     * 
     * @expose 
     * @type {boolean}
     * @alias autofocus
     * @access public
     * @default false
     * @instance
     * @memberof oj.ojInputNumber
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    autofocus: false,
    // TODO: revisit
    // It's expensive to create a default converter ahead of time when a page author can set a custom
    // one for if they do this will be promptly discarded.

    /**
     * A number converter instance that duck types {@link oj.NumberConverter}. Or an object literal
     * containing the properties listed below.
     * When no converter is specified, the default converter will be used,
     * and default option of "numeric" is used. 
     * <p>
     * When <code class="prettyprint">converter</code> property changes due to programmatic
     * intervention, the component performs various tasks based on the current state it is in. </br>
     * When initialized with no options, the default options for the current locale are assumed. </br>
     *
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>Any cached converter instance is cleared and new converter created. The converter hint is
     * pushed to messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when <code class="prettyprint">converter</code> property changes, the
     * display value is refreshed.</li>
     * <li>if component is invalid and is showing messages when
     * <code class="prettyprint">converter</code> property changes, then all messages generated by the
     * component are cleared and full validation run using its current display value.
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code>
     *   property is not updated, and the errors are shown. The display value is not refreshed in this case. </li>
     *   <li>if no errors result from the validation, <code class="prettyprint">value</code>
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code>
     *   event to clear custom errors. The display value is refreshed with the formatted
     *   value provided by converter.</li>
     * </ul>
     * </li>
     * <li>if component is invalid and has deferred messages when
     * <code class="prettyprint">converter</code> property changes, then the display value is
     * refreshed with the formatted value provided by converter.</li>
     * </ul>
     * </p>
     *
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>messages created by
     * the component  are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared. Page authors can
     * choose to clear it explicitly when setting the converter property.</li>
     * </ul>
     * </p>
     *
     * @property {string} type - the converter type registered with the oj.ConverterFactory.
     * Usually 'number'. See {@link oj.NumberConverterFactory} for details. <br/>
     * E.g., <code class="prettyprint">{converter: {type: 'number'}</code>
     * @property {Object=} options - optional Object literal of options that the converter expects.
     * See {@link oj.IntlNumberConverter} for options supported by the jet number converter.
     * E.g., <code class="prettyprint">{converter: {type: 'number', options: {style: 'decimal'}}</code>
     *
     * @example <caption>Initialize component to use default converter</caption>
     * &lt;oj-input-number value="25000">&lt;/oj-input-number>
     *
     * @example <caption>Initialize the component with a number converter instance:</caption>
     * // Initialize converter instance using currency options
     * var options = {style: 'currency', 'currency': 'USD', maximumFractionDigits: 0};
     * var numberConverterFactory = oj.Validation.converterFactory("number");
     * var salaryConverter = numberConverterFactory.createConverter(options);<br/>
     * // set converter instance using converter attribute
     * &lt;oj-input-number value="25000" converter="[[salaryConverter]]">&lt;/oj-input-number>
     *
     * @example <caption>Initialize the component with converter object literal:</caption>
     * &lt;oj-input-number value="25000" converter='
     *   {
     *     "type": "number",
     *     "options" : {
     *       "style": "currency",
     *       "currency": "USD",
     *       "maximumFractionDigits": "0"
     *     }
     *   }'>&lt;/oj-input-number>
     *
     * @example <caption>Get or set the <code class="prettyprint">converter</code> property after initialization:</caption>
     * // Getter
     * var convtr = myComponent.converter;
     * 
     * // Setter
     * myComponent.converter = salaryConverter;
     *
     *
     * @default oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_NUMBER).createConverter()
     *
     * @expose
     * @access public
     * @instance
     * @memberof oj.ojInputNumber
     * @ojsignature {
     *    target: "Type",
     *    value: "oj.Converter<number>|oj.Validation.FactoryRegisteredValidatorOrConverter"}
     * @type {Object}
     */
    converter: oj.Validation.converterFactory(
            oj.ConverterFactory.CONVERTER_TYPE_NUMBER).createConverter(),

    /**
     * The maximum allowed value. This number is used in the range validator; if the
     * <code class="prettyprint">value</code> is greater than the <code class="prettyprint">max</code>,
     * then the range validator flags an error to the user. The up arrow
     *  is disabled when the maximum value is reached.
     * <p>
     *  <code class="prettyprint">Max</code> must be a
     *  <code class="prettyprint">number</code> or <code class="prettyprint">null</code>;
     *  <code class="prettyprint">null</code> indicates no maximum.
     * <p>
     * The <code class="prettyprint">max</code> must not be less than the
     * <code class="prettyprint">min</code>, else an Error is thrown during initialization.
     * @expose
     * @memberof oj.ojInputNumber
     * @instance
     * @type {?number}
     * @default null
     * @example <caption>Initialize the inputNumber with the
     * <code class="prettyprint">max</code> attribute specified:</caption>
     * &lt;oj-input-number max="100">&lt;/oj-input-number>
     * 
     * @example <caption>Change the <code class="prettyprint">max</code> property to a float:</caption>
     * myComponent.max = 100.5;
     * 
     * @example <caption>To remove the maximum range restriction from inputNumber:</caption>
     * myComponent.max =  null;
     * 
     * @example <caption>Get or set the <code class="prettyprint">max</code> property after initialization:</caption>
     * // Getter
     * var max = myComponent.max;
     * 
     * // Setter
     * myComponent.max = 100;
     */
    max: null,

    /**
     * The minimum allowed value. This number is used in the range validator; if the
     * <code class="prettyprint">value</code> is less than the <code class="prettyprint">min</code>,
     * then the range validator flags an error to the user. The down arrow
     *  is disabled when the minimum value is reached.
     *  <p>
     *  <code class="prettyprint">Min</code> must be a <code class="prettyprint">number</code> or <code class="prettyprint">null</code>;
     *  <code class="prettyprint">null</code> indicates no minimum.
     * <p>
     * The <code class="prettyprint">max</code> must not be less than the
     * <code class="prettyprint">min</code>, else an Error is thrown during initialization.
     * @expose
     * @memberof oj.ojInputNumber
     * @instance
     * @type {?number}
     * @default null
     * @example <caption>Initialize the inputNumber with the
     * <code class="prettyprint">min</code> attribute specified:</caption>
     * &lt;oj-input-number min="0">&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">min</code> property after initialization:</caption>
     * // Getter
     * var min = myComponent.min;
     * 
     * // Setter
     * myComponent.min = 0;
     * 
     * @example <caption>Change the <code class="prettyprint">min</code> property to a float:</caption>
     * myComponent.min = 10.5;
     * 
     * @example <caption>To remove the minimum range restriction from inputNumber:</caption>
     * myComponent.min =  null;
     * 
     */
    min: null,
     /** 
     * It indicates the name of the component. 
     * 
     * @example <caption>Initialize component with <code class="prettyprint">name</code> attribute:</caption>
     * &lt;oj-input-number name="myName">&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">name</code> property after initialization:</caption>
     * // getter
     * var ro = myComp.name;
     *
     * // setter
     * myComp.name = "myName";
     * 
     * @expose 
     * @type {string}
     * @alias name
     * @access public
     * @instance
     * @default ""
     * @memberof oj.ojInputNumber
     * @ojextension {_COPY_TO_INNER_ELEM: true}
     */
    name: "",
    /**
     * The placeholder text to set on the element.
     *
     * @example <caption>Initialize the component with the <code class="prettyprint">placeholder</code> attribute:</caption>
     * &lt;oj-input-number placeholder="Enter a number">&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">placeholder</code> property after initialization:</caption>
     * // Getter
     * var placeholder = myComponent.placeholder;
     * 
     * // Setter
     * myComponent.placeholder = "Address";
     *
     * If the attribute is not set then the default can be a converter hint. See display-options for
     * details.
     *
     * @expose
     * @instance
     * @default ""
     * @memberof oj.ojInputNumber
     * @type {string|null}
     */
    placeholder: "",
    /**
     * <p>The <code class="prettyprint">rawValue</code> is the read-only property for retrieving
     * the current value from the input field in string form. The main consumer of 
     * <code class="prettyprint">rawValue</code> is a converter.</p>
     * <p>
     * The <code class="prettyprint">rawValue</code> updates on the 'input' javascript event,
     * so the <code class="prettyprint">rawValue</code> changes as the value of the input is changed.
     * If the user types in '1,200' into the field, the rawValue will be '1', then '1,', then '1,2',
     * ..., and finally '1,200'. Then when the user blurs or presses Enter
     * the <code class="prettyprint">value</code> property gets updated.
     * </p>
     * <p>This is a read-only property so page authors cannot set or change it directly.</p>
     * @expose
     * @access public
     * @instance
     * @memberof oj.ojInputNumber
     * @type {string|undefined}
     * @ojsignature {target:"Type", value:"string"}
     * @since 1.2.0
     * @readonly
     * @ojwriteback
     */
    rawValue: undefined,
    /**
     * Whether the component is readonly.
     *
     * @example <caption>Initialize component with <code class="prettyprint">readonly</code> attribute:</caption>
     * &lt;oj-input-number readonly>&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">readonly</code> property after initialization:</caption>
     * // Getter
     * var readonly = myComponent.readonly;
     * 
     * // Setter
     * myComponent.readonly = false;
     *
     * @default false
     * @access public
     * @expose
     * @type {?boolean}
     * @alias readonly
     * @instance
     * @memberof oj.ojInputNumber
     */
    readOnly: false,
    /** 
     * Whether the component is required or optional. When required is set to true, an implicit 
     * required validator is created using the validator factory - 
     * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED).createValidator()</code>.
     * 
     * Translations specified using the <code class="prettyprint">translations.required</code> attribute 
     * and the label associated with the component, are passed through to the options parameter of the 
     * createValidator method. 
     * 
     * <p>
     * When <code class="prettyprint">required</code> property changes due to programmatic intervention, 
     * the component may clear messages and run validation, based on the current state it's in. </br>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when required is set to true, then it runs deferred validation on 
     * the value property. This is to ensure errors are not flagged unnecessarily.
     * </li>
     * <li>if component is invalid and has deferred messages when required is set to false, then 
     * component messages are cleared but no deferred validation is run.
     * </li>
     * <li>if component is invalid and currently showing invalid messages when required is set, then 
     * component messages are cleared and normal validation is run using the current display value. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   property is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code> 
     *   event on the component to clear custom errors.</li>
     * </ul>
     * </li>
     * </ul>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the component are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * 
     * </p>
     * 
     * @ojvalue {boolean} false - implies a value is not required to be provided by the user. 
     * This is the default.
     * @ojvalue {boolean} true - implies a value is required to be provided by user and the 
     * input's label will render a required icon. Additionally a required validator - 
     * {@link oj.RequiredValidator} - is implicitly used if no explicit required validator is set. 
     * An explicit required validator can be set by page authors using the validators attribute. 
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">required</code> attribute:</caption>
     * &lt;oj-input-number required>&lt;/oj-input-number><br/>
     * 
     * @example <caption>Customize messages and hints used by implicit required validator when 
     * <code class="prettyprint">required</code> attribute is set:</caption> 
     * &lt;oj-input-number required translations='{"required": {
     *                 "hint": "custom: enter at least 3 alphabets",
     *                 "messageSummary": "custom: \'{label}\' is Required", 
     *                 "messageDetail": "custom: please enter a valid value for \'{label}\'"}}'>
     * &lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
     * // getter
     * var rq = myComp.required;
     *
     * // setter
     * myComp.required = false;
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof oj.ojInputNumber
     * @type {boolean}
     * @default false
     * @since 0.7
     * @see #translations
     */
    required: false, 
    /**
     * The size of the step to take when spinning via buttons or via the
     * <code class="prettyprint">stepUp()</code>/<code class="prettyprint">stepDown()</code> methods.
     * Step must be a <code class="prettyprint">number</code>
     * greater than 0, otherwise an exception is thrown. It defaults to <code class="prettyprint">1</code>.
     * <p>
     * The step up and step down feature will change the value to be a step match if it isn't already.
     * A step match is when the value is a multiple of step, starting at the
     * <code class="prettyprint">min</code>, and if min is not set, then starting at the initial value, 
     * and if neither <code class="prettyprint">min</code> or initial value are set, 
     * then starting at <code class="prettyprint">0</code>. For example, if the value is 5, min is 0,
     * and step is 100, stepUp will change value to be 100. Now if the value is 5, min is -20,
     * and step is 100, stepUp will change the value to be 80.
     * If the min is not set, and the initial value is 5 and the step is 100, then the stepUp will change
     * the value to be 105.
     * </p>
     * <p> 
     * A value can be a step mismatch; if the <code class="prettyprint">value</code> is set
     * to be a step mismatch, it will not be flagged as a validation error.
     * @expose
     * @instance
     * @type {?number}
     * @default 1
     * @memberof oj.ojInputNumber
     * @access public
     * @example <caption>Initialize the inputNumber with the
     * <code class="prettyprint">step</code> attribute specified:</caption>
     * &lt;oj-input-number step="2">&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">step</code> property after initialization:</caption>
     * // Getter
     * var step = myComponent.step;
     * 
     * // Setter
     * myComponent.step = 5;
     * */
    step: 1,
    /** 
     * List of validators used by component when performing validation. Each item is either an 
     * instance that duck types {@link oj.Validator}, or is an Object literal containing the 
     * properties listed below. Implicit validators created by a component when certain attributes 
     * are present (e.g. <code class="prettyprint">required</code> attribute), are separate from 
     * validators specified through this attribute. At runtime when the component runs validation, it 
     * combines the implicit validators with the list specified through this attribute. 
     * <p>
     * Hints exposed by validators are shown in the notewindow by default, or as determined by the 
     * 'validatorHint' property set on the <code class="prettyprint">display-options</code> 
     * attribute. 
     * </p>
     * 
     * <p>
     * When <code class="prettyprint">validators</code> property changes due to programmatic 
     * intervention, the component may decide to clear messages and run validation, based on the 
     * current state it is in. </br>
     * 
     * <h4>Steps Performed Always</h4>
     * <ul>
     * <li>The cached list of validator instances are cleared and new validator hints is pushed to 
     * messaging. E.g., notewindow displays the new hint(s).
     * </li>
     * </ul>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when validators changes, component does nothing other than the 
     * steps it always performs.</li>
     * <li>if component is invalid and is showing messages when 
     * <code class="prettyprint">validators</code> changes then all component messages are cleared 
     * and full validation run using the display value on the component. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   property is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   property is updated; page author can listen to the <code class="prettyprint">valueChanged</code> 
     *   event to clear custom errors.</li>
     * </ul>
     * </li>
     * <li>if component is invalid and has deferred messages when validators changes, it does 
     * nothing other than the steps it performs always.</li>
     * </ul>
     * </p>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the component are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> property is not cleared.</li>
     * </ul>
     * </p>
     * 
     * @property {string} type - the validator type that has a {@link oj.ValidatorFactory} that can 
     * be retrieved using the {@link oj.Validation} module. For a list of supported validators refer 
     * to {@link oj.ValidatorFactory}. <br/>
     * E.g., <code class="prettyprint">{validators: [{type: 'regExp'}]}</code>
     * @property {Object=} options - optional Object literal of options that the validator expects. 
     * <br/>
     * E.g., <code class="prettyprint">{validators: [{type: 'regExp', options: {pattern: '[a-zA-Z0-9]{3,}'}}]}</code>

     * 
     * @example <caption>Initialize the component with validator object literal:</caption>
     * &lt;oj-input-number validators='[{"type": "regExp", "options": {
     *                     "pattern": "[a-zA-Z0-9]{3,}", 
     *                     "messageDetail": "You must enter at least 3 letters or numbers"}}]'>
     * &lt;/oj-input-number>      
     * 
     * NOTE: oj.Validation.validatorFactory('numberRange') returns the validator factory that is used 
     * to instantiate a range validator for numbers.
     * 
     * @example <caption>Initialize the component with multiple validator instances:</caption>
     * var validator1 = new MyCustomValidator({'foo': 'A'}); 
     * var validator2 = new MyCustomValidator({'foo': 'B'});
     * var validators = [validator1,validator2];<br/>
     * ...
     * &lt;oj-input-number validators='[[validators]]'>
     * &lt;/oj-input-number>      
     * 
     * @example <caption>Get or set the <code class="prettyprint">validators</code> property after initialization:</caption>
     * // getter
     * var validators = myComp.validators;
     *
     * // setter
     * myComp.validators = myValidators;
     * 
     * @expose 
     * @access public
     * @instance
     * @memberof oj.ojInputNumber
     * @ojsignature  { target: "Type", value: "Array<oj.Validator<number>|oj.Validation.FactoryRegisteredValidatorOrConverter>|null"}
     * @type {Array.<Object>|undefined}
     */
    
    validators: undefined,    
    /**
     * The value of the component. Value must be a number or null.
     * 
     * <p>
     * When <code class="prettyprint">value</code> property changes due to programmatic 
     * intervention, the component always clears all messages
     * including <code class="prettyprint">messagesCustom</code>, runs deferred validation, and 
     * always refreshes UI display value.</br>
     * 
     * <h4>Running Validation</h4>
     * <ul>
     * <li>component always runs deferred validation; if there is a validation error the 
     * <code class="prettyprint">valid</code> property is updated.</li>
     * </ul>
     * </p>
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">value</code> attribute specified:</caption>
     * &lt;oj-input-number value='10'>&lt;/oj-input-number>
     * @example <caption>Get or set <code class="prettyprint">value</code> attribute, after initialization:</caption>
     * // Getter: returns 10
     * var val = myComp.value;
     * // Setter: sets 20
     * myComp.value = 20;
     * 
     * @expose 
     * @access public
     * @instance
     * @default null
     * @ojwriteback
     * @memberof oj.ojInputNumber
     * @type {?number}
     */
    value: null,
    /**
     * The type of virtual keyboard to display for entering value on mobile browsers.  This attribute has no effect on desktop browsers.
     *
     * @example <caption>Initialize the component with the <code class="prettyprint">virtual-keyboard</code> attribute:</caption>
     * &lt;oj-input-number virtual-keyboard="number">&lt;/oj-input-number>
     * 
     * @example <caption>Get or set the <code class="prettyprint">virtualKeyboard</code> property after initialization:</caption>
     * // Getter
     * var virtualKeyboard = myComp.virtualKeyboard;
     * 
     * // Setter
     * myComp.virtualKeyboard = "number";
     *
     * @expose
     * @instance
     * @memberof oj.ojInputNumber
     * @type {string}
     * @ojvalue {string} "auto" The component will determine the best virtual keyboard to use.
     *                          <p>This is always "text" for this release but may change in future
     *                          releases.</p>
     * @ojvalue {string} "number" Use a virtual keyboard for entering number.
     *                            <p>Note that on Android and Windows Mobile, the "number" keyboard does
     *                            not contain the minus sign.  This value should not be used on fields that
     *                            accept negative values.</p>
     * @ojvalue {string} "text" Use a virtual keyboard for entering text.
     * @default "auto"
     * @since 5.0.0
     */
    virtualKeyboard: "auto"

    // Events

    /**
     * Triggered when the ojInputNumber is created.
     *
     * @event
     * @name create
     * @memberof oj.ojInputNumber
     * @instance
     * @property {Event} event event object
     * @property {Object} ui Currently empty
     * @ignore
     *
     * @example <caption>Initialize the ojInputNumber with the <code class="prettyprint">create</code> callback specified:</caption>
     * $( ".selector" ).ojInputNumber({
     *     "create": function( event, ui ) {}
     * });
     *
     * @example <caption>Bind an event listener to the <code class="prettyprint">ojcreate</code> event:</caption>
     * $( ".selector" ).on( "ojcreate", function( event, ui ) {} );
     */
    // create event declared in superclass, but we still want the above API doc
  },
  // P U B L I C    M E T H O D S

  getNodeBySubId: function(locator)
  {
    var node = this._superApply(arguments), subId;
    if (!node)
    {
      subId = locator['subId'];
      if (subId === "oj-inputnumber-up") {
        node = this.widget().find(".oj-inputnumber-up")[0];
      }
      if (subId === "oj-inputnumber-down") {
        node = this.widget().find(".oj-inputnumber-down")[0];
      }
      if (subId === "oj-inputnumber-input") {
        node = this.widget().find(".oj-inputnumber-input")[0];
      }
    }
    // Non-null locators have to be handled by the component subclasses
    return node || null;
 },
  getSubIdByNode: function(node)
  {
    var subId = null;

    if (node != null)
    {
      if(node === this.widget().find(".oj-inputnumber-up")[0])
      {
        subId = {'subId': "oj-inputnumber-up"};
      }
      else if(node === this.widget().find(".oj-inputnumber-down")[0])
      {
        subId = {'subId': "oj-inputnumber-down"};
      }
      else if(node === this.widget().find(".oj-inputnumber-input")[0])
      {
        subId = {'subId': "oj-inputnumber-input"};
      }
    }

    return subId || this._superApply(arguments);
  },
  /**
   * Refreshes the inputNumber component
   * <p>A <code class="prettyprint">refresh()</code> or re-init is required
   * when an inputNumber is changed in a non-option way, like in the following circumstances:
   * <ul>
   *   <li>Button translations change.</li>
   * </ul>
   * @expose
   * @memberof oj.ojInputNumber
   * @access public
   * @instance
   * @return {void}
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   * myComp.refresh();
   */
  refresh: function()
  {
    this._super();
    this._setup();
  },
  /**
   * <p>Decrements the value by the specified number of steps.
   * Without the parameter, a single step is decremented.</p>
   <p>If the resulting value is above the max, below the min,
   or results in a step mismatch, the value will be adjusted to the closest valid value.</p>
   * @param {number=} steps - Number of steps to decrement, defaults to 1.
   * @return {void}
   * @expose
   * @instance
   * @memberof oj.ojInputNumber
   * @access public
   * @example <caption>Invoke the <code class="prettyprint">stepDown</code> method:</caption>
   * myComp.stepDown();
   */
  stepDown: function(steps)
  {
    this._step(steps, false);
  },
   /**
   * <p>Increments the value by the specified number of steps.
   * Without the parameter, a single step is incremented.</p>
   <p>If the resulting value is above the max, below the min,
   or results in a step mismatch, the value will be adjusted to the closest valid value.</p>
   * @param {number=} steps - Number of steps to increment, defaults to 1.
   * @return {void}
   * @expose
   * @instance
   * @memberof oj.ojInputNumber
   * @access public
   * @example <caption>Invoke the <code class="prettyprint">stepUp</code> method:</caption>
   * myComp.stepUp();
  */
  stepUp: function(steps)
  {
    this._step(steps, true);
  },

  /**
   * Returns a jQuery object containing the element visually representing the inputnumber.
   *
   * <p>This method does not accept any arguments.
   *
   * @expose
   * @memberof oj.ojInputNumber
   * @instance
   * @return {jQuery} the inputnumber
   * @access public
   * @ignore
   */
  widget: function()
  {
    return this.uiInputNumber;
  },

  // P R O T E C T E D    C O N S T A N T S   A N D   M E T H O D S

  // *********** START WIDGET FACTORY METHODS (they retain _camelcase naming convention) **********


  /**
  * Called at component create time primarily to initialize options, often using DOM values. This
  * method is called before _ComponentCreate is called, so components that override this method
  * should be aware that the component has not been rendered yet. The element DOM is available and
  * can be relied on to retrieve any default values. <p>
  * @param {!Object} originalDefaults - original default options defined on the widget and its ancestors.
  * For example, if we have the step attribute defaulted to 12 in this class, it will be 12. (unless there
  * is some global default on the component ?)
  * @param {?Object} constructorOptions - options passed into the widget constructor
  *
  * @memberof oj.ojInputNumber
  * @instance
  * @protected
  */
  _InitOptions: function(originalDefaults, constructorOptions)
  {
    var opts = this.options;
    var self = this;;

    // call super class with arguments: originalDefaults and constructorOptions. It will set
    // the this.options.
    this._superApply(arguments);

    if (!this._IsCustomElement())
    {
      // 'props' is the list of properties that I need to get from the dom if they aren't already defined in the options.
      // There is no need to list defaults here like we used to do
      // since the defaults are in originalDefaults and they are merged in with the options in the this._superApply call.

      // attribute below is the html-5 dom attribute name. If 'option' is different, like in the case of
      // readonly (readonly html vs readOnly (camelcase) component option), specify both.
      var props =
        [{attribute: "disabled", validateOption: true},
        {attribute: "placeholder"},
        {attribute: "value"}, // don't coerce here. I do it myself
        {attribute: "readonly", option: "readOnly", validateOption: true},
        {attribute: "required", coerceDomValue: true, validateOption: true},
        {attribute: "title"},
        {attribute: "min"},
        {attribute: "max"},
        {attribute: "step"}];


      oj.EditableValueUtils.initializeOptionsFromDom(props, constructorOptions, this,
        // post-process callback
        function (initializedOptions)
        {
        // coerce regardless of where the option value came from - dom/constructor
          var toParse = ['value', 'step', 'min', 'max'];

          for (var i=0; i<toParse.length; i++)
          {
            var opt = toParse[i];
            var value = (opt in initializedOptions) ? initializedOptions[opt] : opts[opt];
            if (value != null)
            {
              if (opt === 'step')
                initializedOptions[opt] =  self._parseStep(value);
              else
                initializedOptions[opt] =  self._parse(opt, value);
            }
          }
        }
      );
    }

    // The custom element bridge framework coerces the options according to the type in the
    // metadata file (componentRegister.js), so there is no need to do the coercion in _InitOptions
    // for custom elements like we do above for non-custom-elements.
    // Check the 'step' to make sure it's in the correct range.
    if (this._IsCustomElement())
    {
      var optValue = opts['step'];
      if (optValue != null)
      {
        // this will coerce with a + and throw an error if it is < 0.
        // since the bridge frameworkd code already coerced step to a number before _InitOptions
        // was called, all we care about is throwing an error if step < 0.
        self._parseStep(optValue);
      }
    }


    if (opts['value'] === undefined)
    {
      throw new Error("ojInputNumber has no value");
    }
    
    this.initialValue = opts['value'];

    // now make sure min < max, else throw an Error
    if (opts['min'] != null && opts['max'] != null)
    {
      if (opts['max'] < opts['min'])
      {
        throw new Error("ojInputNumber's max must not be less than min");
      }
    }

  },
  /**
   * After _ComponentCreate and _AfterCreate,
   * the widget should be 100% set up. this._super should be called first.
   * @override
   * @protected
   * @instance
   * @memberof oj.ojInputNumber
   */
  _ComponentCreate: function()
  {
    this._super();

    this._draw();

    this._inputNumberDefaultValidators = {};
    this._setup();
    // I want everything set up before I turn on events, since the events affect the component,
    // like keydown will update the value and update the buttons.
    this._on(this._events);
    
    this._focusable( this.uiInputNumber );
  },
  /**
   * Performs post processing after _SetOption() is called. Different options when changed perform
   * different tasks. See _AfterSetOption[OptionName] method for details.
   *
   * @param {string} option
   * @param {Object|string=} previous
   * @param {Object=} flags
   * @protected
   * @memberof oj.ojInputNumber
   * @instance
   * @override
   */
  _AfterSetOption : function (option, previous, flags)
  {
    this._superApply(arguments);
    switch (option)
    {
      case "min":
      case "max":
        this._Refresh(option, this.options[option]);
        break;           
      case "readOnly":
        this._AfterSetOptionDisabledReadOnly(option, oj.EditableValueUtils.readOnlyOptionOptions);
        break;
      case "required":
        this._AfterSetOptionRequired(option);
        break;
      case "validators":
        this._AfterSetOptionValidators(option);
        break;
      case "converter":
        this._AfterSetOptionConverter(option);
        break;           
      case "virtualKeyboard":
        this._SetInputType(this._ALLOWED_TYPES);
        break;        
      default:
        break;
    }

  },
  /**
   * Whether the a value can be set on the component. For example, if the component is 
   * disabled or readOnly then setting value on component is a no-op. 
   * 
   * @see #_SetValue
   * @return {boolean}
   * @memberof oj.ojInputNumber
   * @override
   * @instance
   * @protected
   */
  _CanSetValue: function ()
  {
    var readOnly;
    var superCanSetValue = this._super();
    
    if (!superCanSetValue)
      return false;

    readOnly = this.options['readOnly'] || false;
    return (readOnly) ? false : true;

    
  },
  /**
   * Performs post processing after value option changes by taking the following steps.
   * This method piggybacks on the super's method by using the 'doNotClearMessages' flag to
   * determine if this was a direct value option programmatic change.
   * 
   * - Calls super<br/>
   * - if setOption was from programmatic intervention, <br/>
   * &nbsp;&nbsp;- reset the this.initialValue which is used to determine the stepBase; <br/>
   * 
   * @param {string} option
   * @param {Object=} flags
   * 
   * @protected
   * @memberof oj.ojInputNumber
   * @instance
   * 
   */
  _AfterSetOptionValue : function(option, flags)
  {
    this._superApply(arguments);
    var context = flags ? flags['_context'] : null;
    var isUIValueChange;
    var doNotClearMessages;

    if (context)
    {
      isUIValueChange = context.originalEvent ? true : false;
      doNotClearMessages = context.doNotClearMessages || false;
    }

    if (!isUIValueChange)
    {
      // value option can be updated directly (i.e., programmatically or through user interaction) 
      // or updated indirectly as a result of some other option changing - e.g., converter, 
      // validators, required etc. See _updateValue() method for details.
      // When value changes directly due to programatic intervention (usually page author does this) 
      // then update this.initialValue.
      if (!doNotClearMessages) 
      {
        this.initialValue = this.options["value"];
      }

    }

  },
  /**
   * Whether the component is required.
   * 
   * @return {boolean} true if required; false
   * 
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   * @override
   */
  _IsRequired : function () 
  {
    return this.options['required'];
  },
  /**
   * Performs post processing after required option is set by taking the following steps.
   * 
   * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
   * run full validation with UI value (we don't know if the UI error is from a required validator 
   * or something else);<br/>
   * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
   * updated<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to optionChange(value) to clear custom errors.<br/>
   * 
   * - if component is invalid and has messagesHidden -> required: false -> clear component 
   * errors; no deferred validation is run.<br/>
   * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
   * errors unnecessarily)<br/>
   * - messagesCustom is never cleared<br/>
   * 
   * @param {string} option
   * 
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   */
  _AfterSetOptionRequired : oj.EditableValueUtils._AfterSetOptionRequired,
  /**
   * When validators option changes, take the following steps.
   * 
   * - Clear the cached normalized list of all validator instances. push new hints to messaging.<br/>
   * - if component is valid -> validators changes -> no change<br/>
   * - if component is invalid has messagesShown -> validators changes -> clear all component 
   * messages and re-run full validation on displayValue. if there are no errors push value to 
   * model;<br/>
   * - if component is invalid has messagesHidden -> validators changes -> do nothing; doesn't change 
   * the required-ness of component <br/>
   * - messagesCustom is not cleared.<br/>
   * 
   * NOTE: The behavior applies to any option that creates implicit validators - min, max, pattern, 
   * etc. Components can call this method when these options change.
   * 
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   */
  _AfterSetOptionValidators : oj.EditableValueUtils._AfterSetOptionValidators,
  /**
   * Performs post processing after converter option changes by taking the following steps.
   * 
   * - always push new converter hint to messaging <br/>
   * - if component has no errors -> refresh UI value<br/>
   * - if component is invalid has messagesShown -> clear all component errors and run full 
   * validation using display value. <br/>
   * &nbsp;&nbsp;- if there are validation errors, value is not pushed to model; messagesShown is 
   * updated.<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to optionChange(value) to clear custom errors.<br/>
   * - if component is invalid has messagesHidden -> refresh UI value. no need to run deferred 
   * validations. <br/>
   * - messagesCustom is never cleared<br/>
   * 
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   */
  _AfterSetOptionConverter : oj.EditableValueUtils._AfterSetOptionConverter,  
  /**
   * Clears the cached converter stored in _converter and pushes new converter hint to messaging.
   * Called when convterer option changes 
   * @memberof oj.ojInputNumber
   * @instance
   * @protected 
   */
  _ResetConverter : oj.EditableValueUtils._ResetConverter,
 
    // *********** END WIDGET FACTORY METHODS **********
  /**
   * Need to override since we allow users to set the converter to null, undefined, and etc and when
   * they do we use the default converter
   *
   * @return {Object} a converter instance or null
   *
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   * @override
   */
  _GetConverter : function ()
  {
    return this.options['converter'] ?
        this._getConverter() :
        $["oj"]["ojInputNumber"]["prototype"]["options"]["converter"];
  },
  /**
   * This returns an array of all validators 
   * normalized from the validators option set on the component. <br/>
   * @return {Array} of validators. 
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   */
  _GetNormalizedValidatorsFromOption : oj.EditableValueUtils._GetNormalizedValidatorsFromOption,
  
  /**
   * Called to find out if aria-required is unsupported.
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   */
  _AriaRequiredUnsupported : function()
  {
    return false;
  },
  /**
   * Handles options specific to inputnumber.
   * Note that _setOption does not get called during create in the super class.
   * It only gets called when the component has already been created.
   * However, we do call _setOption in _draw for certain attributes
   * (disabled)
   * @override
   * @protected
   * @memberof oj.ojInputNumber
   */
  _setOption: function(key, value, flags)
  {
    var coercedValue;

    if (!this._IsCustomElement() && (key === "value" || key === "max" || key === "min"))
    {
      // we only have to coerce for non-custom-elements since the frameworkd coerces for us for
      // custom elements.
      coercedValue = this._parse(key, value);
    }
    else if (key === "step")
    {
      coercedValue = this._parseStep(value);
    }
    else
      coercedValue = value;


    // the superclass calls _Refresh. Our _Refresh calls _updateButtons
    // and _refreshAriaMinMaxValue.
    // call _super with the newly coerced 'value' property.
    this._super(key, coercedValue, flags);

    if (key === "max" || key === "min")
    {
      // since validators are immutable, they will contain min + max as local values.
      // Because of this will need to recreate
      this._createRangeValidator();
      this._AfterSetOptionValidators();
    }

    // when a dom element supports disabled, use that, and not aria-disabled.
    // having both is an error.
    // having aria-disabled on root dom element is ok (if it is added in base class)
    if (key === "disabled")
    {
      // force it to be a boolean. this is what/how ojbutton, EditableValue does.
      this.element.prop("disabled", !!value);
    }
    // when a dom element supports readonly, use that, and not aria-readonly.
    // having both is an error
    if (key === "readOnly")
    {
      this.element.prop("readonly", !!value);
      this._refreshStateTheming("readOnly", this.options.readOnly);
      this._refreshRoleSpinbutton("readOnly", this.options.readOnly);
    }
  },
  /**
   * Override of protected base class method.
   * Method name needn't be quoted since is in externs.js.
   * @protected
   * @memberof oj.ojInputNumber
   * @instance
   */
  _destroy: function()
  {
    var ret = this._super();

    // destroy the buttonset
    this.buttonSet.ojButtonset("destroy");
    this.buttonSet.remove();
    this.upButton = null;
    this.downButton = null;
    this.buttonSet = null;
    this.initialValue = null;

    //  - DomUtils.unwrap() will avoid unwrapping if the node is being destroyed by Knockout
    oj.DomUtils.unwrap(this.element, this.uiInputNumber);
    clearTimeout(this.timer);
    return ret;
  },
      
  /**
   * Validates the component's display value using the converter and all validators registered on 
   * the component and updates the <code class="prettyprint">value</code> option by performing the 
   * following steps. 
   * 
   * <p>
   * <ol> 
   * <li>All messages are cleared, including custom messages added by the app. </li>
   * <li>If no converter is present then processing continues to next step. If a converter is 
   * present, the UI value is first converted (i.e., parsed). If there is a parse error then 
   * the messages are shown.</li>
   * <li>If there are no validators setup for the component the <code class="prettyprint">value</code> 
   * option is updated using the display value. Otherwise all 
   * validators are run in sequence using the parsed value from the previous step. The implicit 
   * required validator is run first if the component is marked required. When a validation error is 
   * encountered it is remembered and the next validator in the sequence is run. </li>
   * <li>At the end of validation if there are errors, the messages are shown. 
   * If there were no errors, then the 
   * <code class="prettyprint">value</code> option is updated.</li>
   * </ol>
   * 
   * @example <caption>Validate component using its current value.</caption>
   * // validate display value and shows messages if there are any to be shown.
   * myComp.validate();
   * @example <caption>Validate component and use the Promise's resolved state.</caption>
   * myComp.validate().then(
   *  function(result) {
   *    if(result === "valid")
   *    {
   *      submitForm();
   *    }
   *  }); 
   * @return {Promise.<string>} Promise resolves to "valid" if there were no converter parse errors and
   * the component passed all validations. 
   * The Promise resolves to "invalid" if there were converter parse errors or 
   * if there were validation errors.
   * 
   * 
   * @method
   * @access public
   * @expose
   * @instance
   * @memberof oj.ojInputNumber
   * @since 4.0.0
   * @ojstatus preview
   */
  validate : oj.EditableValueUtils.validate,
  /**
   * Used for explicit cases where the component needs to be refreshed
   * (e.g., when the value option changes or other UI gestures).
   * @override
   * @protected
   * @memberof oj.ojInputNumber
   */
  _Refresh: function(name, value, forceDisplayValueRefresh)
  {
    var valuenow;
    
    this._superApply(arguments);
    
    switch (name)
    {
      case "disabled":
      case "max":
      case "min": 
      case "value":
        valuenow = this._getConvertedDisplayValue();
        // disables or enables both the up and down buttons depending upon what the value
        // is on the screen after conversion plus what min/max are set to.
        this._updateButtons(valuenow); 
        // do not break;
      case "max":
      case "min": 
      case "value":
        // refreshes the aria-valuenow, aria-min, aria-max, etc.
        this._refreshAriaMinMaxValue(valuenow);
        break;
        
      case "converter":
        valuenow = this._getConvertedDisplayValue();
        this._refreshAriaText(valuenow);
        break;
     
      case "required":
        this._refreshRequired(value);
        break;
        
      default:
        break;
        
    }
  },
  /**
   * @memberof oj.ojInputNumber
   * @instance
   * @private
   */
  _refreshRequired : oj.EditableValueUtils._refreshRequired,


  /**
   * Sets up the default numberRange validators if there is a min or max.
   *
   * @ignore
   * @protected
   * @override
   * @instance
   * @memberof oj.ojInputNumber
   */
  _GetImplicitValidators: function()
  {
    var ret = this._superApply(arguments);
    if (this.options['min'] != null || this.options['max'] != null)
    {
      this._createRangeValidator();
    }

    return $.extend(this._inputNumberDefaultValidators, ret);
  },
  /**
   * Returns the default styleclass for the component.
   *
   * @return {string}
   * @memberof oj.ojInputNumber
   * @override
   * @protected
   */
  _GetDefaultStyleClass: function()
  {
    return "oj-inputnumber";
  },
  _events:
  {
    'input': function(event)
    {
      this._SetRawValue(this.element.val(), event);
    },
    'keydown': function(event)
    {
      var keyCode = $.ui.keyCode;
      if (event.keyCode === keyCode.ENTER)
      {
        this._blurEnterSetValue(event);
        event.preventDefault();
      }
      else if (this._start() && this._keydown(event))
      {
        event.preventDefault();
      }
    },
    'keyup': function(event)
    {
      this._stop(event);
    },
    'blur': function(event)
    {
      this._blurEnterSetValue(event);
    },
    "touchstart .oj-inputnumber-button.oj-enabled": function(event)
    {
      this._start();
      
      this._repeat(null, $(event.currentTarget).hasClass("oj-inputnumber-up") ? 1 : -1, event);
    },
    "touchend .oj-inputnumber-button": function(event)
    {
      this._stop(event);
    },
    "touchcancel .oj-inputnumber-button": function(event)
    {
      this._stop(event);
    },
    "mousedown .oj-inputnumber-button.oj-enabled": function(event)
    {
      if (this._isRealMouseEvent(event))
      {
        this._start();

        this._repeat(null, $(event.currentTarget).hasClass("oj-inputnumber-up") ? 1 : -1, event);
      }
    },
    "mouseup .oj-inputnumber-button": function(event)
    {
      if (this._isRealMouseEvent(event))
      {
        this._stop(event);
      }
    },
    "mouseenter .oj-inputnumber-button.oj-enabled": function(event)
    {
      // button will add oj-active if mouse was down while mouseleave and kept down
      if (!$(event.currentTarget).hasClass("oj-active"))
      {
        return;
      }
      if (this._isRealMouseEvent(event))
      {
        this._start();

        this._repeat(null, $(event.currentTarget).hasClass("oj-inputnumber-up") ? 1 : -1, event);
      }
    },
    // TODO: do we really want to consider this a stop?
    // shouldn't we just stop the repeater and wait until mouseup before
    // we trigger the stop event?
    "mouseleave .oj-inputnumber-button": function(event)
    {
      if (this._isRealMouseEvent(event))
      {
        this._stop(event);
      }
    }
  },
  // I N T E R N A L   P R I V A T E   C O N S T A N T S    A N D   M E T H O D S
  // 
  // Subclasses should not override or call these methods
  /**
   * @private
   * @const
   * @memberof oj.ojInputNumber
   */
  _BUNDLE_KEY:
  {
    _TOOLTIP_DECREMENT: 'tooltipDecrement',
    _TOOLTIP_INCREMENT: 'tooltipIncrement'
  },
  /**
   * when below listed options are passed to the component, corresponding CSS will be toggled
   * @private
   * @const
   * @memberof oj.ojInputNumber
   * @type {Object}
   */
  _OPTION_TO_CSS_MAPPING:
  {
    "readOnly": "oj-read-only"
  },

  /**
   * _setup is called on create and refresh.
   * @private
   * @memberof oj.ojInputNumber
   */
  _setup: function()
  {

    // add/update translated strings to buttons
    var incrementString =
            this.getTranslatedString(this._BUNDLE_KEY._TOOLTIP_INCREMENT);
    var decrementString =
            this.getTranslatedString(this._BUNDLE_KEY._TOOLTIP_DECREMENT);
    var valuenow = this._getConvertedDisplayValue();
    this.upButton.ojButton({label: incrementString});
    this.downButton.ojButton({label: decrementString});
    this._refreshAriaMinMaxValue(valuenow);
    this._updateButtons(valuenow);
    // update element DOM for readOnly
    if (typeof this.options['readOnly'] === "boolean")
    {
      this.element.prop("readonly", this.options['readOnly']);
    }
    this._refreshStateTheming("readOnly", this.options.readOnly);
    this._refreshRoleSpinbutton("readOnly", this.options.readOnly);
    
    this._refreshRequired(this.options['required']);
  },
  // Mark internal JET components for automation support. The automation
  // support needs to know while traversing the nodes that the JET button/buttonset 
  // is not the root JET component, but an internal node to a JET component.
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _markInternalComponents: function ()
  {
     this.upButton.attr('data-oj-internal', '');
     this.downButton.attr('data-oj-internal', '');
     this.buttonSet.attr('data-oj-internal', '');
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _createOjButtonset: function()
  {
    var $upButton = this.uiInputNumber.find(".oj-inputnumber-up");
    var $downButton = this.uiInputNumber.find(".oj-inputnumber-down");
    var buttonsetDiv = $upButton[0].parentNode;
    this.upButton = $upButton.ojButton({display: 'icons',
      icons: {start: 'oj-component-icon oj-inputnumber-up-icon'}})
    this.downButton = $downButton.ojButton({display: 'icons',
      icons: {start: 'oj-component-icon oj-inputnumber-down-icon'}})
    this.buttonSet = $(buttonsetDiv).ojButtonset({focusManagement: 'none'});
    this._markInternalComponents();
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _draw: function()
  {
    var element = this.element;
    var widgetId;
    
    this.uiInputNumber = element.addClass("oj-inputnumber-input")         
            .wrap(this._uiInputNumberHtml()).parent() // @HTMLUpdateOK
            // add buttons          
            .append(this._buttonHtml()); // @HTMLUpdateOK
    if (this.OuterWrapper) 
    {
      this.uiInputNumber = $(this.OuterWrapper).append(this.uiInputNumber); // @HTMLUpdateOK
      this.uiInputNumber.addClass("oj-inputnumber oj-component");
    }
    else 
    {
      this.uiInputNumber = this.uiInputNumber.wrap("<div class='oj-inputnumber oj-component'></div>").parent(); // @HTMLUpdateOK
    }
    
    if (this._IsCustomElement())
    {
      // if it is a custom element, then set the sub-id on the input so if they have a oj-label
      // pointing to it with the 'for' attrbiute, JAWS will read the label.
      widgetId = this.widget().attr("id");
      if (widgetId)
        oj.EditableValueUtils.setSubIdForCustomLabelFor(this._GetContentElement()[0], widgetId);
    }
    //
    // TODO: need to save off attributes and reset on destroy generically.
    this.saveType = element.prop("type");
    this._SetInputType(this._ALLOWED_TYPES);


    // As the buttons are not in the keyboard sequence at first
    // we decided it makes sense to add aria-hidden="true" to them
    // and rely on the up/down arrow keys. However, now we decided to remove aria-hidden because
    // in voiceover there is no way to access the up/down buttons otherwise.
    // Still, voiceover is broken due to this webkit bug
    //  - ios: input number doesn't support vo because of webkit bug 
    this.uiInputNumber.find(".oj-inputnumber-button")
        .attr("tabIndex", "-1");
    this._createOjButtonset();

  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _keydown: function(event)
  {
    var keyCode = $.ui.keyCode;

    switch (event.keyCode)
    {
      // keeping the up/down pressed repeats
      // using the up and down arrows will adjust the value so that it is
      // a multiple of step and it is in min/max, same as if you used the
      // up and down buttons
      case keyCode.UP:
        this._repeat(null, 1, event);
        return true;
      case keyCode.DOWN:
        this._repeat(null, -1, event);
        return true;
    }

    return false;
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _uiInputNumberHtml: function()
  {
    // need to wrap the input+buttons with a span so that inline messaging
    // can come after this span, and we can position popups on this span.
    return "<span class='oj-inputnumber-wrapper'></span>";
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _buttonHtml: function()
  {
    return "<div class='oj-buttonset-width-auto'>" +
      "<button type='button' class='oj-inputnumber-button oj-inputnumber-down'></button>" +
      "<button type='button' class='oj-inputnumber-button oj-inputnumber-up'></button></div>";
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _start: function()
  {
    this.spinning = true;
    return true;
  },
  /**
   * Calls _spin to increment or decrement the number. It is called in a loop until
   * this.timer is cleared (see this._stop(event) or min/max is reached. 
   * @private
   * @memberof oj.ojInputNumber
   */
  _repeat: function(i, steps, event)
  {
    var stopRepeat = false;
    
    // if steps is > 0, it is going up, else it is going down.
    // need to check if min/max is reached, and if so, stop the repeat.
    // we do a quick css check to see if it is disabled.
    if (steps > 0)
    {
      if (this.upButton.hasClass("oj-disabled"))
        stopRepeat = true;
    }
    else
    {
      if (this.downButton.hasClass("oj-disabled"))
        stopRepeat = true;
    }
    // repeat spinning as long as the key is down and min/max isn't reached
    i = i || 500;

    clearTimeout(this.timer);
    // this.timer will be cleared elsewhere, like when the user stops holding down the up/down
    // arrows. See this._stop
    this.timer = this._delay(function()
    {
      if (!stopRepeat)
      {

        this._repeat(40, steps, event);
      }
    },
            i);

    this._spin(steps * this.options.step, event);
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   * @param {Number} step - Number of steps to increment.
   * @param {Object=} event an optional event if this was a result of ui interaction.
   */
  _spin: function(step, event)
  {
    // When the component's 'value' changes, the displayValue is automatically updated.
    // So reading the component's display value should always give you the element's value
    var value = this._getConvertedDisplayValue();
    var options = this.options;
    var minOpt = options.min;
    var maxOpt = options.max;
    var stepOpt = options.step;
    var initialValue = this.initialValue;
    
    // get the max precision. e.g., min=2.4, initialValue=3.444, maxPrecision is 3.
    var precision = this._precision(minOpt, stepOpt, initialValue);
    
    value = this._adjustValue(value, step, minOpt, maxOpt, stepOpt, precision, initialValue);
    
    // Show the user what is going to be validated. We are making it so that clicking on the
    // up/down button is the same as if the user typed in a number and blurred.
    if (this._CanSetValue())
    {
      // update the input's val which is what the user sees.
      this.element.val(value);
      // keep aria-valuenow in sync with the input's val
      this._refreshAriaMinMaxValue(value);
      // keep up/down buttons disabled state in sync with the input's val
      this._updateButtons(value);
      // keep rawValue in sync with the input's val
      // use this.element.val() because the raw value should always be a string
      this._SetRawValue(this.element.val(), event);
    }

    // now validate and set value and all of that.
    this._SetValue(value, event, {'validationMode': this._VALIDATION_MODE.VALIDATORS_ONLY});
  },
  /**
   * called from _adjustValue
   * @private
   * @memberof oj.ojInputNumber
   */
  _precision: function(minOpt, stepOpt, value)
  {
    var precision = this._precisionOf(stepOpt);
    
    if (minOpt != null)
    {
      precision = Math.max(precision, this._precisionOf(minOpt));
    }
    
    if (value != null)
      precision = Math.max(precision, this._precisionOf(value));
    
    return precision;
  },
  /**
   * return the number of digits after the '.'
   * called from _adjustValue->_precision
   * @private
   * @memberof oj.ojInputNumber
   * @param {Number} num - Number from which to calculate the precision
   */
  _precisionOf: function(num)
  {
    var str = num.toString(), decimal = str.indexOf(".");
    return decimal === -1 ? 0 : str.length - decimal - 1;
  },
  /**
   * adjust the value to be "valid".
   * The logic follows that of HTML-5's input number.
   * http://www.w3.org/TR/html5/forms.html#dom-input-stepup
   * A valid value is one that is a multiple of
   * step starting at stepBase, where stepBase is min (if present),
   * else initial value (if present),
   * else (if type == number) 0
   * If max is not a valid value, stepUp/stepDown will never go to max. It
   * will go to the calculated valid max (one that is the largest value
   * that is an integral multiple of the step, and that is less than or equal
   * to the maximum.
   * @private
   * @memberof oj.ojInputNumber
   * @param {number} value - the current value
   * @param {number} step - the step you want to adjust the value by
   * @param {number} minOpt - the min option
   * @param {number} maxOpt - the max option
   * @param {number} stepOpt - the step option
   * @param {number} precision - the precision @see _precision
   * @param {number} initialValue - the value when the component was created
   * @returns {number} - the new value after it has been adjusted
   */
  _adjustValue: function(value, step, minOpt, maxOpt, stepOpt, precision, initialValue)
  {
    var newValue;
    var stepBase; 
    var aboveMin;
    var valueWithFraction;
    
    if (precision > 0)
    {
      valueWithFraction = 
        this._adjustValueForFractions(value, step, minOpt, maxOpt, stepOpt, precision, initialValue);
      return valueWithFraction;
    }

    // make sure we're at a valid step when we step up or down.
    // - find out where we are relative to the base.
    // follow these rules. use min, else use initial value, else use 0.
    // https://www.w3.org/TR/html5/forms.html#concept-input-min-zero
    stepBase = minOpt != null ? minOpt : initialValue;
    if (stepBase == null)
      stepBase = 0;

    
    // From http://www.w3.org/TR/html5/forms.html#dom-input-stepup:
    // If value subtracted from the step base is not an integral multiple
    // of the step, then set value to the nearest value that, when subtracted
    // from the step base, is an integral multiple of the allowed value step,
    // and that is less than value if the method invoked was stepDown() and
    // more than value if the method invoked was stepUp().

    // is value-stepBase an integral multiple of step?
    try 
    {
      value = parseFloat(value.toFixed(precision));
    }
    catch (e)
    {
      if (e instanceof TypeError)
      {
        // I've only seen this fail if they set a null converter.
        oj.Logger.warn("inputNumber's value after conversion is not a number. \n\
                      The converter must convert the value to a Number. coercing using +");
        value = +value; // coerce
      }
    }
    aboveMin = value - stepBase;
    
    var rounded = Math.round(aboveMin / stepOpt) * stepOpt;
    rounded = parseFloat(rounded.toFixed(precision));
    var multiple = (rounded === aboveMin);

    if (!multiple)
    {
      if (step < 0)
        aboveMin = Math.ceil(aboveMin / stepOpt) * stepOpt;
      else
      {
        aboveMin = Math.floor(aboveMin / stepOpt) * stepOpt;
      }
      // rounding is based on 0, so adjust back to our base
      newValue = stepBase + aboveMin + step;
    }
    else
    {
      newValue = value + step;
    }

    // fix precision from bad JS floating point math
    // toFixed returns the newValue with a specific # of digits after the
    // decimal point (this_precision() looks at max of step/min/value's # of
    // digits.
    newValue = parseFloat(newValue.toFixed(precision));
    
    if (minOpt != null && newValue < minOpt) 
    {
      return minOpt;
    }

    if (maxOpt!= null && newValue > maxOpt)
    {
      var validMax = (Math.floor((maxOpt - stepBase) / stepOpt) *
              stepOpt) + stepBase;
      // fix precision from bad JS floating point math
      validMax = parseFloat(validMax.toFixed(precision));
      return validMax;
    }

    return newValue;
  },
  /**
   * Call this from _adjustValue when you have numbers with precision > 0. This method
   * multiples everything by Math.pow(10,precision), calls _adjustValue with these numbers so that
   * the math works, then divides the result by Math.pow(10,precision) to get it back to fractions.
   * This is to work around the issue with Javascript's binary floating-point numbers not being 
   * great about adding decimal fractions. e.g., 0.1 + 0.2 is not equal to 0.3. It is 
   * an intentional consequence of the IEEE Standard for Binary Floating-Point Arithmetic (IEEE 754)
   * @private
   * @memberof oj.ojInputNumber
   * @param {number} value - the current value
   * @param {number} step - the step you want to adjust the value by
   * @param {number} minOpt - the min option
   * @param {number} maxOpt - the max option
   * @param {number} stepOpt - the step option
   * @param {number} precision - the precision @see _precision
   * @param {number} initialValue - the value when the component was created
   * @returns {number}
   */
  _adjustValueForFractions: function(value, step, minOpt, maxOpt, stepOpt, precision, initialValue)
  {
    // don't call this function if precision is 0
    oj.Assert.assert(precision > 0);
    var power = Math.pow(10,precision);
    // if minOpt, maxOpt, stepOpt are undefined, keep them that way
    // when we adjust the values to make them decimals, they should be whole numbers.
    // javascript sometimes gives them fractions (
    // e.g., 10000000.45*100=1000000044.9999999), so everywhere here we multiply
    // by power, round the value to make it a whole number.
    var minOptPower = minOpt != null ? Math.round(minOpt*power) : minOpt;
    var maxOptPower = maxOpt != null ? Math.round(maxOpt*power) : maxOpt;
    var stepOptPower = stepOpt != null ? Math.round(stepOpt*power) : stepOpt;
    
    
    var adjustValuePower =   
      this._adjustValue(Math.round(value*power), 
                        Math.round(step*power), 
                        minOptPower, 
                        maxOptPower, 
                        stepOptPower, 
                        0, 
                        Math.round(initialValue*power));
    return adjustValuePower/power;
    
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _stop: function(event)
  {
    if (!this.spinning)
    {
      return;
    }
    clearTimeout(this.timer);
    this.spinning = false;
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   * @return {boolean} true if there is no touch detected within the last 500 ms
   */
  _isRealMouseEvent: function(event)
  {
    return ! oj.DomUtils.recentTouchEnd();
  },
  /**
   * disables or enables both the up and down buttons depending upon what the value
   * is on the screen after conversion plus what min/max are set to.
   * e.g., $10.00 is on the screen, valuenow is 10
   * @private
   * @memberof oj.ojInputNumber
   */
  _updateButtons: function(valuenow)
  {
    var options = this.options;
    var maxOpt = options.max;
    var minOpt = options.min;
    var downButton = this.downButton;
    var upButton = this.upButton;
    var downButtonDisabledAlready;
    var upButtonDisabledAlready;
    var isMaxOptNonNull = maxOpt != null;
    var isMinOptNonNull = minOpt != null;

    if (!this.uiInputNumber)
      return;
    
    if (!downButton && !upButton)
      return;

    // to prevent the overhead of disabling a button that is already disabled, check to see
    // if it is already disabled already.
    downButtonDisabledAlready = downButton.hasClass("oj-disabled");
    upButtonDisabledAlready = upButton.hasClass("oj-disabled");
      
    if (options.disabled || valuenow === undefined || 
       (isMaxOptNonNull && isMinOptNonNull && maxOpt === minOpt && valuenow === maxOpt))
    {
      if (!downButtonDisabledAlready)
        downButton.ojButton("disable");
      if (!upButtonDisabledAlready)
        upButton.ojButton("disable");
    }
    else if (isMaxOptNonNull && valuenow >= maxOpt)
    {
      if (downButtonDisabledAlready)
        downButton.ojButton("enable");
      if (!upButtonDisabledAlready)
        upButton.ojButton("disable");
    }
    else if (isMinOptNonNull && valuenow <= minOpt)
    {
      if (!downButtonDisabledAlready)
        downButton.ojButton("disable");
      if (upButtonDisabledAlready)
        upButton.ojButton("enable");
    }
    else
    {
      if (downButtonDisabledAlready)
        downButton.ojButton("enable");
      if (upButtonDisabledAlready)
        upButton.ojButton("enable");
    }
    

  },
  /**
   * Returns the normalized converter instance.
   * 
   * @return {Object} a converter instance or null
   * @memberof oj.ojInputNumber
   * @instance
   * @private 
   */
  _getConverter : oj.EditableValueUtils._GetConverter,   
  /**
   * Returns the converted display value.
   * This function gets the display value (or 0 if no display value), then parses it using
   * the converter.
   * The converted display value is used for 'valuenow'. For example,
   * the display value might be $6.00, but we want valuenow to be 6.
   * @private
   * @memberof oj.ojInputNumber
   */
  _getConvertedDisplayValue: function()
  {
    var value;
    var displayValue;
    try
    {
      displayValue = this._GetDisplayValue() || 0;
      // if displayValue is not parseable, say it is 'abc',
      // then _parseValue throws an error. catch it and move on.
      // TODO: Make _parseValue protected in EditableValue();
      value = this._parseValue(displayValue);
    }
    catch (e)
    {
      // catch the error, set value to undefined, and continue to update
      // the buttons. both the up/down buttons will be disabled in this case.
      value = undefined;
    }
    return value;
  },
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _blurEnterSetValue: function(event)
  {
    var val = this.element.val();
    var valuenow;
    
    this._stop();
    valuenow = this._getConvertedDisplayValue();
    this._refreshAriaMinMaxValue(valuenow);
    this._updateButtons(valuenow);
    // _SetValue triggers valuechange event
    this._SetValue(val, event);
  },
  /**
   * Create the NumberRangeValidator. Use the translations['numberRange'] options if they exist,
   * otherwise use the NumberRangeValidator default strings for
   *  hints, messageSummary and messageDetail.
   * Use the 'min' and 'max' options.
   * @private
   * @memberof oj.ojInputNumber
   */
  _createRangeValidator: function()
  {
    var options = this.options;
    var minOpt = options['min'];
    var maxOpt = options['max'];
    var newMin = minOpt != null ? minOpt : undefined;
    var newMax = maxOpt != null ? maxOpt : undefined;
    var translations = options['translations'];
    var translationsOptionNumberRange = translations ? translations['numberRange'] || {}: {};   
    var hintMin;
    var hintMax;
    var hintInRange;
    var hintExact;
    var messageDetailRangeOverflow;
    var messageDetailRangeUnderflow;
    var messageDetailExact;
    var messageSummaryRangeOverflow;
    var messageSummaryRangeUnderflow;
    var translationsHint = translationsOptionNumberRange['hint'] || {};
    var translationsMessageDetail = translationsOptionNumberRange['messageDetail'] || {};
    var translationsMessageSummary = translationsOptionNumberRange['messageSummary'] || {};
    var numberRangeValidatorOptions;

    // First check if the translations hint/messageDetail/messageSummary options are set, and if
    // so, use them.
    // This is how the app dev could use inputNumber's translations option.
    // var element = $("#spin").ojInputNumber(
    //      {value: 10, min: -50,
    //        translations:
    //        {numberRange:
    //          {hint:
    //            {min: 'Translations Option Test: Please enter a number greater than or equal to {min}',
    //             max: 'Translations Option Test: Please enter a number less than or equal to {max}',
    //             inRange: 'Translations Option Test: Please enter a number between {min} and {max}'
    //            },
    //            messageDetail:
    //            {rangeUnderflow: 
    //            'Translations Option Test: The number {value} must be greater than or equal to {min}',
    //              rangeOverflow: 
    //              'Translations Option Test: The number {value} must be less than or equal to {max}'
    //            }
    //          }
    //        } // end numberRange
    if (translationsHint !== null)
    {
      hintMin = translationsHint['min'] || null;
      hintMax = translationsHint['max'] || null;
      hintInRange = translationsHint['inRange'] || null;
      hintExact = translationsHint['exact'] || null;
    }
    if (translationsMessageDetail !== null)
    {
      messageDetailRangeOverflow = translationsMessageDetail['rangeOverflow'] || null;
      messageDetailRangeUnderflow = translationsMessageDetail['rangeUnderflow'] || null;
      messageDetailExact = translationsMessageDetail['exact'] || null;
    }
    if (translationsMessageSummary !== null)
    {
      messageSummaryRangeOverflow = translationsMessageSummary['rangeOverflow'] || null;
      messageSummaryRangeUnderflow = translationsMessageSummary['rangeUnderflow'] || null;
    }

    // unless the translations options for the numberrange were set, the hints and messageDetails
    // and messageSummaries will be null and we'll pick up the default strings from the 
    // NumberRangeValidator.
    numberRangeValidatorOptions = {
      'min': newMin,
      'max': newMax,
      'hint' : {'min':hintMin || null,
                'max':hintMax || null,
                'inRange': hintInRange || null,
                'exact': hintExact || null
               },
      'messageDetail' : {
                'rangeOverflow':messageDetailRangeOverflow || null,
                'rangeUnderflow': messageDetailRangeUnderflow || null,
                'exact': messageDetailExact || null},
      'messageSummary' : {
                'rangeOverflow':messageSummaryRangeOverflow || null,
                'rangeUnderflow': messageSummaryRangeUnderflow || null},
      'converter': this._GetConverter()};

    this._inputNumberDefaultValidators[oj.ValidatorFactory.VALIDATOR_TYPE_NUMBERRANGE] =
            oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_NUMBERRANGE)
            .createValidator(numberRangeValidatorOptions);
  },

  // The user can clear out min/max by setting the option to null, so we
  // do not coerce null.
  /**
   * @param {string} option name of the option. this will show up in the error if thrown
   * @param val value to parse
   * @throws {Error} if option value is invalid
   * @private
   * @memberof oj.ojInputNumber
   */
  _parse: function(option, val)
  {
    var returnValue;
    if (val !== null)
      returnValue = +val;
    else
      returnValue = val;

    // isNaN(null) returns false, which is what we want
    if (isNaN(returnValue))
      throw new Error("ojInputNumber's " + option + " option is not a number");

    return returnValue;
  },
  /**
   * parse the step's value
   * We are following the behavior of HTML-5 the best we can. According
   * to the spec, it says step must be a number greater than 0.
   * Chrome defaults it to 1 if it is not.
   * @throws {Error} if option value is invalid
   * @private
   * @memberof oj.ojInputNumber
   */
  _parseStep: function(val)
  {
    var defaultStep = 1;
    var parsedStep;
    if (val === null)
      return defaultStep;
    parsedStep = this._parse("step", val);
    if (parsedStep <= 0)
    {
      // throw an exception
      throw new Error("Invalid step for ojInputNumber; step must be > 0");
    }
    // DEFAULT to 1 if it isn't > 0
    if (parsedStep === null || parsedStep <= 0)
      parsedStep = defaultStep;
    return parsedStep;
  },
  /**
   * Toggles css selector on the widget. E.g., when readOnly option changes,
   * the oj-read-only selector needs to be toggled.
   * @param {string} option
   * @param {Object|string|boolean} value
   * @private
   * @memberof oj.ojInputNumber
   */
  _refreshStateTheming: function(option, value)
  {
    if (Object.keys(this._OPTION_TO_CSS_MAPPING).indexOf(option) != -1)
    {
      // value is a boolean
      this.widget().toggleClass(this._OPTION_TO_CSS_MAPPING[option], !!value);
    }
  },
  /**
   * When readOnly option changes,
   * the role spinbutton needs to be toggled We don't have role spinbutton
   * on readOnly inputNumber.
   * @param {string} option
   * @param {Object|string|boolean} readOnly
   * @private
   * @memberof oj.ojInputNumber
   */
  _refreshRoleSpinbutton: function(option, readOnly)
  {
    readOnly = !!readOnly;
    // if readonly is true, remove role spinbutton
    // if readonly is false, add role spinbutton
    if (readOnly)
      this.element.removeAttr("role");
    else
      this.element.attr("role", "spinbutton");
  },
  /* updates the aria-value information */
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _refreshAriaMinMaxValue: function(valuenow)
  {
    var element = this.element;

    element.attr(
            {
              "aria-valuemin": this.options.min, "aria-valuemax": this.options.max,
              "aria-valuenow": valuenow
            });
    this._refreshAriaText(valuenow);
  },
  /* updates the aria-text if needed */
  /**
   * @private
   * @memberof oj.ojInputNumber
   */
  _refreshAriaText: function(valuenow)
  {
    var element = this.element;
    var valuetext = element.val();

    if (!this._CompareOptionValues("value", "" + valuenow, valuetext))
      element.attr({"aria-valuetext": valuetext});
  },
  /**
   * step the inputnumber value up or down
   * @private
   * @memberof oj.ojInputNumber
   * @param {Number} steps - Number of steps to increment.
   * @param {boolean} up If true step up, else step down.
   */
  _step: function(steps, up)
  {
    this._start();
    if (up)
      this._spin((steps || 1) * this.options.step);
    else
      this._spin((steps || 1) * -this.options.step);
    this._stop();
  },
  /**
   * Set the type of the input element based on virtualKeyboard option.
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   * @ignore
   */
  _SetInputType: oj.EditableValueUtils._SetInputType,
  /**
   * the validate method from v3.x that returns a boolean
   * @memberof oj.ojInputNumber
   * @instance
   * @protected
   * @ignore
   */
  _ValidateReturnBoolean: oj.EditableValueUtils._ValidateReturnBoolean,

  // API doc for inherited methods with no JS in this file:

  /**
   * Removes the inputNumber functionality completely.
   * This will return the element back to its pre-init state.
   *
   * <p>This method does not accept any arguments.
   *
   * @method
   * @name oj.ojInputNumber#destroy
   * @memberof oj.ojInputNumber
   * @instance
   * @ignore
   *
   * @example <caption>Invoke the <code class="prettyprint">destroy</code> method:</caption>
   * myComp.destroy();
   */

      // Fragments:

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Up Button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Increment the number.</td>
     *     </tr>
     *     <tr>
     *       <td>Down Button</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Decrement the number.</td>
     *     </tr>
     *     <tr>
     *       <td>Input</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Set focus to the input. If hints, help.instruction or messages exists in a notewindow,
     *       pop up the notewindow.</td>
     *     </tr>
     *     <tr>
     *       <td>Elsewhere on Page</td>
     *       <td><kbd>Touch</kbd></td>
     *       <td>Submit the value you typed in the input field.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojInputNumber
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan="4">Input</td>
     *       <td><kbd>Enter</kbd> or <kbd>Tab</kbd></td>
     *       <td>Submit the value you typed in the input field.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Tab In</kbd></td>
     *       <td>Set focus to input. If hints, help.instruction or messages exist in a notewindow,
     *       pop up the notewindow.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>Increment the number.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Decrement the number.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojInputNumber
     */
    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     * 
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-form-control-text-align-right</td>
     *       <td>Aligns the text to the right regardless of the reading direction,
                 this is normally used for right aligning numbers 
     *       </td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojInputNumber
     */
});

// -----------------------------------------------------------------------------
// "private static members" shared by all inputNumbers
// -----------------------------------------------------------------------------

//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the inputNumber's Up arrow.</p>
 *
 * @ojsubid oj-inputnumber-up
 * @memberof oj.ojInputNumber
 *
 * @example <caption>Get the node for the Up arrow:</caption>
 * var node = myComp.getNodeBySubId('oj-inputnumber-up');
 */

/**
 * <p>Sub-ID for the inputNumber's Down arrow.</p>
 *
 * @ojsubid oj-inputnumber-down
 * @memberof oj.ojInputNumber
 *
 * @example <caption>Get the node for the Down arrow:</caption>
 * var node = myComp.getNodeBySubId('oj-inputnumber-down);
 */

/**
 * <p>Sub-ID for the inputNumber's input element.</p>
 * @ojsubid oj-inputnumber-input
 * @memberof oj.ojInputNumber
 *
 * @example <caption>Get the node for the input element:</caption>
 * var node = myComp.getNodeBySubId('oj-inputnumber-input');
 */
}() ); // end of inputNumber wrapper function
(function() {
var ojInputNumberMeta = {
  "properties": {
    "autocomplete": {
      "type": "string",
      "enumValues": ["off", "on"],
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "autofocus": {
      "type": "boolean",
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "converter": {
      "type": "Object"
    },
    "max": {
      "type": "number"
    },
    "min": {
      "type": "number"
    },
    "name": {
      "type": "string",
      "extension": {
        _COPY_TO_INNER_ELEM: true
      }
    },
    "placeholder": {
      "type": "string"
    },
    "rawValue": {
      "type": "string",
      "writeback": true,
      "readOnly": true
    },
    "readonly": {
      "type": "boolean"
    },
    "required": {
      "type": "boolean"
    },    
    "step": {
      "type": "number"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "numberRange": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "Object",
              "properties": {
                "exact": {
                  "type": "string"
                },
                "inRange": {
                  "type": "string"
                },
                "max": {
                  "type": "string"
                },
                "min": {
                  "type": "string"
                }
              }
            },
            "messageDetail": {
              "type": "Object",
              "properties": {
                "exact": {
                  "type": "string"
                },
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            },
            "messageSummary": {
              "type": "Object",
              "properties": {
                "rangeOverflow": {
                  "type": "string"
                },
                "rangeUnderflow": {
                  "type": "string"
                }
              }
            }
          }
        },
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        },
        "tooltipDecrement": {
          "type": "string",
          "value": "Decrement"
        },
        "tooltipIncrement": {
          "type": "string",
          "value": "Increment"
        }
      }
    },
    "validators": {
      "type": "Array"
    },
    "value": {
      "type": "number",
      "writeback": true
    },
    "virtualKeyboard": {
      "type": "string",
      "enumValues": ["auto", "number", "text"]
    }
  },
  "methods": {
    "stepDown": {},
    "stepUp": {},
    "validate": {}
  },
  "extension": {
    _ALIASED_PROPS: {"readonly": "readOnly"},
    _INNER_ELEM: 'input',
    _WIDGET_NAME: "ojInputNumber",
    _GLOBAL_TRANSFER_ATTRS: ["accesskey", "aria-label", "tabindex"]
  }
};
oj.CustomElementBridge.registerMetadata('oj-input-number', 'editableValue', ojInputNumberMeta);
oj.CustomElementBridge.register('oj-input-number', {'metadata': oj.CustomElementBridge.getMetadata('oj-input-number')});
})();
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

// Forked from Version 2.0.0 of https://github.com/hammerjs/jquery.hammer.js, retrieved 2/4/2015.

if (Hammer) {
    /** @export */
    $.fn.ojHammer = function(options) {
        switch (options) {
            // $(...).ojHammer("instance") returns the Hammer instance previously instantiated on this element via $(...).ojHammer(options), or undefined if none.
            case "instance":
                return this.data("ojHammer");

            // $(...).ojHammer("destroy") destroys the Hammer instance and removes it from the node's data store.
            case "destroy":
                return this.each(function() {
                    var $el = $(this);
                    var hammer = $el.data("ojHammer");
                    if (hammer) {
                        hammer.destroy();
                        $el.removeData("ojHammer");
                    }
                });

            default:
                // $(...).ojHammer(options) instantiates Hammer on this element, and stores it for later retrieval via $(...).ojHammer("instance")
                return this.each(function() {
                    var $el = $(this);
                    if(!$el.data("ojHammer")) {
                        $el.data("ojHammer", new Hammer["Manager"]($el[0], options));
                    }
                });
        }
    };

    // extend the emit method to also trigger jQuery events
    Hammer["Manager"]["prototype"]["emit"] = (function(originalEmit) {
        return function(type, data) {
            originalEmit.call(this, type, data);
            $(this.element).trigger({
                "type": type,
                "gesture": data
            });
        };
    })(Hammer["Manager"]["prototype"]["emit"]);
    
} else {
    oj.Logger.warn("Hammer jQuery extension loaded without Hammer.");
}

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

///////////// JsonDiagramDataSource //////////////////   

/**
 * @export
 * @class oj.JsonDiagramDataSource
 * @extends oj.DiagramDataSource
 * @ojtsignore
 * @classdesc JSON implementation of the oj.DiagramDataSource
 * @param {Object} data JSON data object with following properties:
 * <table>
 * <tbody>
 * <tr><td><b>nodes</b></td><td>An optional array of nodes. See {@link oj.DiagramDataSource.NodeObject} section.</td></tr>
 * <tr><td><b>links</b></td><td>An optional array of links. See {@link oj.DiagramDataSource.LinkObject} section.</td></tr>
 * </tbody>
 * </table>
 * @param {Object=} options the options set on this data source
 * @param {Function=} options.childData Function callback to retrieve nodes and links for the specified parent.
 *                      Function will return a Promise that resolves into an object with the following structure:</p>
 * <table>
 * <tbody>
 * <tr><td><b>nodes</b></td><td>An array of objects for the child nodes for the given parent. 
 *                              See {@link oj.DiagramDataSource.NodeObject} section.</td></tr>
 * <tr><td><b>links</b></td><td>An array of objects for the links for the given parent. 
 *                              See {@link oj.DiagramDataSource.LinkObject} section.</td></tr>
 * </tbody>
 * </table>
 * @constructor
 * @since 3.0
 */
oj.JsonDiagramDataSource = function(data, options) {
    this.childDataCallback = options ? options['childData'] : null;
    this._nodesMapById = {};
    oj.JsonDiagramDataSource.superclass.constructor.call(this, data);
};

// Subclass from oj.DiagramDataSource
oj.Object.createSubclass(oj.JsonDiagramDataSource, oj.DiagramDataSource, "oj.JsonDiagramDataSource");

/**
 * Returns child data for the given parent.
 * The data include all immediate child nodes along with links whose endpoints
 * both descend from the current parent node. 
 * If all the links are available upfront, they can be returned as part of the
 * top-level data (since all nodes descend from the diagram root).
 * If lazy-fetching links is desirable, the most
 * optimal way to return links is as part of the data of the
 * nearest common ancestor of the link's endpoints.
 *
 * @param {Object|null} parentData An object that contains data for the parent node.
 *                     See {@link oj.DiagramDataSource.NodeObject} section.
 *                     If parentData is null, the method retrieves data for top level nodes.
 * @return {Promise} Promise resolves to a component object with the following structure:<p>
 * <table>
 * <tbody>
 * <tr><td><b>nodes</b></td><td>An array of objects for the child nodes for the given parent.
 *                              See {@link oj.DiagramDataSource.NodeObject} section.</td></tr>
 * <tr><td><b>links</b></td><td>An array of objects for the links for the given parent.
 *                              See {@link oj.DiagramDataSource.LinkObject} section.</td></tr>
 * </tbody>
 * </table>
 * @export
 * @method
 * @name getData
 * @memberof! oj.JsonDiagramDataSource
 * @instance
 */
oj.JsonDiagramDataSource.prototype.getData = function(parentData) {
  if (parentData) { //retrieve child data
    var childData = parentData['nodes'];
    if (childData === undefined && this.childDataCallback) {
      childData = this.childDataCallback(parentData);
      this._updateLocalData(this._nodesMapById[parentData['id']], childData);
      return childData;
    }
    else {
      this._updateNodesMap({'nodes': childData});
      return Promise.resolve({'nodes': childData});
    }
  }
  else { // retrieve top level data
    if (this.data) {
      this._updateNodesMap(this.data);
      return Promise.resolve(this.data);
    }
    else if (this.childDataCallback){
      var childData = this.childDataCallback();
      this._updateLocalData(null, childData);
      return childData;
    }
    else {
      return Promise.resolve(null);
    }
  }
};

/**
 * Internal method that updates a map of nodes that is used to improve performace
 * @param {Object} data data that might contain diagram nodes
 * @private
 */
oj.JsonDiagramDataSource.prototype._updateNodesMap = function(data) {
  // if callback for additional data is not specified - no need to keep map of nodes
  if (!this.childDataCallback || !data['nodes'])
    return;
  var nodes = data['nodes'];
  for (var i = 0; i < nodes.length; i++) {
    this._nodesMapById[nodes[i]['id']] = nodes[i];
  }
};

/**
 * Internal method that updates local data copy
 * @param {Object} parentData parent node that has to be updated
 * @param {Promise} childData
 * @private
 */
oj.JsonDiagramDataSource.prototype._updateLocalData = function(parentData, childData) {
  if (childData instanceof Promise) {
    var thisRef = this;
    childData.then(
      function(data) {
        if (!parentData && !thisRef.data) {
          //top level data were initially set to null, 
          //create an object before setting its nodes and links properties
          thisRef.data = {};
        }
        if (Array.isArray(data['nodes'])) {
          if (parentData) {
            parentData['nodes'] = data['nodes'];
          }
          else {
            thisRef.data['nodes'] = data['nodes'];
          }
          thisRef._updateNodesMap(data);
        }
        if (Array.isArray(data['links'])) {
          thisRef.data['links'] = Array.isArray(thisRef.data['links']) ? 
                                  thisRef.data['links'].concat(data['links']) : data['links'];
        }
        thisRef.handleEvent('ADD', {'data':data, 
                                    'parentId':parentData ? parentData['id'] : null,
                                    'index':0});
      },
      function(reason) {}
    );
  }
};

/**
 * Retrieves number of child nodes
 * @param {Object} nodeData A data object for the node in question.
 *                          See {@link oj.DiagramDataSource.NodeObject} section.
 * @return {number} Number of child nodes if child count is available.
 *                  The method returns 0 for leaf nodes.
 *                  The method returns -1 if the child count is unknown
 *                  (e.g. if the children have not been fetched).
 * @export
 * @method
 * @name getChildCount
 * @memberof! oj.JsonDiagramDataSource
 * @instance
 */
oj.JsonDiagramDataSource.prototype.getChildCount= function(nodeData) {
  if (nodeData) {
    var childData = nodeData['nodes'];
    var count = Array.isArray(childData) ? childData.length :
                childData === undefined && this.childDataCallback ? -1 :
                0;
    return count;
  }
  return -1;
};

/**
 * Indicates whether the specified object contains links
 * that should be discovered in order to display promoted links.
 *
 * @param {Object} nodeData A data object for the container node in question.
 *                          See {@link oj.DiagramDataSource.NodeObject} section.
 * @return {string} the valid values are "connected", "disjoint", "unknown"
 * @export
 * @method
 * @name getDescendantsConnectivity
 * @memberof! oj.JsonDiagramDataSource
 * @instance
 */
oj.JsonDiagramDataSource.prototype.getDescendantsConnectivity = function(nodeData){
  return "unknown";
};

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * A JsonNodeSet represents a collection of nodes.  The JsonNodeSet is an object returned by the success callback
 * of the fetchChildren method on TreeDataSource.  
 * @implements oj.NodeSet
 * @constructor
 * @since 1.0
 * @ojtsignore
 * @param {number} startNode the index of the first node in this NodeSet relative to its parent
 * @param {number} endNode the index of the last node in this NodeSet
 * @param {Object} data the JSON data
 * @param {*} currKey the key of the parent node
 * @param {number} depth the depth of the nodes in this NodeSet
 * @export
 */
oj.JsonNodeSet = function(startNode, endNode, data, currKey, depth)
{
    // assert startNode/endNode are number
    oj.Assert.assertNumber(startNode, null);
    oj.Assert.assertNumber(endNode, null);

    this.m_depth = depth;
    this.m_key = currKey;
    this.m_startNode = startNode;
    this.m_endNode = endNode;
    this.m_nodes = data;
};

/**
 * Gets the parent key for this result set.  
 * @return {*} the parent key for this result set. 
 * @export
 */
oj.JsonNodeSet.prototype.getParent = function()
{
    return this.m_key;
};

/**
 * Gets the start index of the result set.  
 * @return {number} the start index of the result set.
 * @export	
 */
oj.JsonNodeSet.prototype.getStart = function()
{
    return this.m_startNode;
};

/**
 * Gets the actual count of the result set.  
 * @return {number} the actual count of the result set.
 * @export	
 */
oj.JsonNodeSet.prototype.getCount = function()
{
    return Math.max(0, this.m_endNode - this.m_startNode);
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds. 
 * @param {number} index the index of the node/row in which we want to retrieve the data from.  
 * @return {*} the data for the specified index.  oj.RowData should be returned for data that represents a row
 *         with a number of columns.
 * @export
 */
oj.JsonNodeSet.prototype.getData = function(index)
{
    // make sure index are valid
    oj.Assert.assert(index <= this.m_endNode && index >= this.m_startNode);

    // adjust to relative index
    index = index - this.m_startNode;

    if (this.m_nodes[index])
        return this.m_nodes[index].attr;
    else
        return null;
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and 
 * 2) the index specified is out of bounds. 
 * The metadata that the data source must return are:
 *  1) key - <any>, the key of the node/row.
 *  2) leaf - boolean, true if it's a leaf, false otherwise. 
 *  3) depth - number, the depth of the node/row. 
 * @param {number} index the index of the node/row in which we want to retrieve the metadata from.  
 * @return {{key: *, leaf: boolean, depth: number}} the metadata of the item
 * @export
 */
oj.JsonNodeSet.prototype.getMetadata = function(index)
{
    var metadata = {"leaf": false, "depth": -1};

    // make sure index are valid
    oj.Assert.assert(index <= this.m_endNode && index >= this.m_startNode);

    // adjust to relative index
    index = index - this.m_startNode;

    metadata["key"] = this.m_nodes[index].id ? this.m_nodes[index].id : this.m_nodes[index].attr.id;
    metadata["leaf"] = this.m_nodes[index].leaf;
    metadata["depth"] = this.m_nodes[index].depth;

    if(metadata["leaf"] == null)
    {
        if (this.m_nodes[index].children && this.m_nodes[index].children.length > 0)
        {
            metadata["leaf"] = false;
        }
        else
        {
            metadata["leaf"] = true;
        }
    }

    return metadata;
};

/**
 * Helper method to update the node's depth recursively with its children.
 * @param {Object} currChild the node to update.
 * @param {number} offset the difference between current and updated depth values.
 * @private
 */
oj.JsonNodeSet.prototype._updateDepth = function (currChild, offset)
{
    var i;

    offset++;
    currChild.depth = offset;

    if (currChild.children && currChild.children.length != 0)
    {
        for (i = 0; i < currChild.children.length; i++)
	{
            this._updateDepth(currChild.children[i], offset);
	}
    }
};

/**
 * Gets the node set child of the specified index.
 * @param {number} index the index of the node/row in which we want to retrieve the child node set
 * @return {oj.JsonNodeSet|null} the child node set representing the child tree data.
 * @export
 */
oj.JsonNodeSet.prototype.getChildNodeSet = function(index) {

    var results, key, depth, i;

    // make sure index are valid
    oj.Assert.assert(index <= this.m_endNode && index >= this.m_startNode);

    // adjust to relative index
    index = index - this.m_startNode;

    depth = this.m_nodes[index].depth;
    results = this.m_nodes[index].children;
    if(results == null || results.length == 0)
    {
        return null;
    }
    key = this.m_nodes[index].id ? this.m_nodes[index].id : this.m_nodes[index].attr.id;
    for(i = 0; i < results.length; i++)
    {
        this._updateDepth(results[i], depth);
    }

    return new oj.JsonNodeSet(0, results.length, results, key, 0);
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
//////////////////// _JsonTreeNodeDataSource ///////////////////////////////////

/**
 * Helper class to implement sort recursive features for tree.
 * @constructor
 * @private
 */
oj._JsonTreeNodeDataSource = function()
{
    this.id = null;
    this.depth = 0;
    this.parent = null;
    /** @type {Array} */
    this.children = null;
    this.title = null;
    this.attr = null;
    this.leaf = null;
};

/**
 * Helper comparer method for ascending sort.
 * @private
 */
oj._JsonTreeNodeDataSource.prototype._ascending = function(key)
{
    return function(a, b) 
    {
        if (a.attr != null && b.attr != null) 
        {
            if (a.attr[key] != null && b.attr[key] != null)
            {
                return a.attr[key] < b.attr[key] ? -1 : a.attr[key] === b.attr[key] ? 0 : 1;
            }
        }
        return a[key] < b[key] ? -1 : a[key] === b[key] ? 0 : 1;
    };
};

/**
 * Helper comparer method for descending sort.
 * @private
 */
oj._JsonTreeNodeDataSource.prototype._descending = function(key)
{
    return function(a, b) 
    {
        if (a.attr != null && b.attr != null) 
        {
            if (a.attr[key] != null && b.attr[key] != null)
            {
                return a.attr[key] < b.attr[key] ? 1 : a.attr[key] === b.attr[key] ? 0 : -1;
            }
        }
        return a[key] < b[key] ? 1 : a[key] === b[key] ? 0 : -1;
    };
};

/**
 * Helper method for recursive sort.
 * @param {Object} criteria the sort criteria.
 * @param {Object} criteria.key the key identifying the attribute to sort on
 *        {string} criteria.direction the sort direction, valid values are "ascending", "descending".
 * @private
 */
oj._JsonTreeNodeDataSource.prototype._sortRecursive = function(criteria)
{
    var key = criteria['key'];
    if (this.children == null)
    {
        return this;
    }

    if (criteria['direction'] === 'ascending')
    {
        this.children.sort(this._ascending(key));
    }
    else if (criteria['direction'] === 'descending')
    {
        this.children.sort(this._descending(key));
    }
    for (var i = 0, l = this.children.length; i < l; i++)
    {
        this.children[i]._sortRecursive(criteria);
    }

    return this;
};

///////////// JsonTreeDataSource //////////////////   

/**
 * @class oj.JsonTreeDataSource
 * @classdesc TreeDataSource implementation that represents hierarchical data available from an array of JSON objects.  This data source can be used by [ListView]{@link oj.ojListView}, 
 *            [NavigationList]{@link oj.ojNavigationList}, and [TreeView]{@link oj.ojTreeView}.<br><br>
 *            See the <a href="../jetCookbook.html?component=treeView&demo=json">Tree View - Data Source: JSON</a> demo for an example.<br><br>
 *            Refer to {@link oj.TreeDataSource} for other data sources that represent hierarchical data.
 * @param {Object} data An array of JSON objects that represent the root nodes.
 *                      <p>Each node object can contain the following properties:</p>
 *                      <p>attr - an object of name-value pairs that represents data for the node.</p>
 *                      <p>children - an array of JSON objects that represent child nodes.</p> 
 * @constructor
 * @since 1.0
 * @export
 * @ojtsignore
 * @extends oj.TreeDataSource
 * @example
 * // First initialize the tree data.  This can be defined locally or read from file.
 * var treeData = [
 *                  {"attr": {"id": "dir1", "title": "Directory 1"},
 *                   "children": [
 *                     {"attr": {"id": "subdir1", "title": "Subdirectory 1"},
 *                      "children": [
 *                        {"attr": {"id": "file1", "title": "File 1"}},
 *                        {"attr": {"id": "file2", "title": "File 2"}},
 *                        {"attr": {"id": "file3", "title": "File 3"}}
 *                      ]}
 *                   ]},
 *                  {"attr": {"id": "dir2", "title": "Directory 2"},
 *                   "children": [
 *                     {"attr": {"id": "file4", "title": "File 4"}},
 *                     {"attr": {"id": "file5", "title": "File 5"}},
 *                   ]}
 *                ];
 *
 * // Then create a JsonTreeDataSource object with the tree data
 * var dataSource = new oj.JsonTreeDataSource(treeData);
 */
oj.JsonTreeDataSource = function(data)
{
    var tree;

    tree = new oj._JsonTreeNodeDataSource(); // that's the root node

    if (data.id == null)
    {
        tree.id = "root";
    }

    this.data = this._createTreeDataSource({count:0}, tree, data);

    oj.JsonTreeDataSource.superclass.constructor.call(this, tree);
};

// Subclass from oj.TreeDataSource
oj.Object.createSubclass(oj.JsonTreeDataSource, oj.TreeDataSource, "oj.JsonTreeDataSource");

/**
 * Initial the json object based data source.
 * @return {undefined}
 * @export
 * @ojtsignore
 */
oj.JsonTreeDataSource.prototype.Init = function()
{
    // call super
    oj.JsonTreeDataSource.superclass.Init.call(this);
};

/**
 * Returns tree based structure/object from json data.
 * @param {Object} c an object to keep track of the count
 * @param {Object} target the final tree structure. 
 * @param {Object} source the json object.
 * @param {number=} depth used recursively for depth calculation.
 * @return target
 * @private
 */
oj.JsonTreeDataSource.prototype._createTreeDataSource = function(c, target, source, depth)
{
    var children, node, child, prop, propr, prp, j;

    if (depth == null)
    {
        depth = 0;
    }

    for (prop in source)
    {
        if (prop == "children" || (depth == 0 && source instanceof Array))
        {
            if (depth == 0 && source instanceof Array)
            {
                children = source;
            }
            else
            {
                children = source[prop];
            }

            target.children = [];

            depth++;
            for (j = 0; j < children.length; j++)
            {
                child = children[j];
                node = new oj._JsonTreeNodeDataSource();
                if (child.id == null) 
                {
                    c.count++;
                    if (child.attr == null) 
                    {
                        node.id = 'rid_' + c.count;
                    } 
                    else if (child.attr.id == null) 
                    {
                        child.attr.id = 'rid_' + c.count;
                    }
                }
                for (propr in child)
                {
                    for (prp in node)
                    {
                        if (propr == prp && propr != "children")
                        {    
                            node[prp] = child[propr];
                        }
                        if (prp == "depth")
                        {
                            node[prp] = depth;
                        }
                    }
                }
                target.children.push(node);
                for (prp in child)
                {
                    if (prp == "children")
                    {
                        this._createTreeDataSource(c, target.children[j], child, depth);
                    }
                }
            }
        }
    }
    return target;
};

/**
 * Returns the number of children for a specified parent.  If the value returned is not >= 0 then it is automatically assumed
 * that the child count is unknown.
 * @param {*} parentKey the parent key.  Specify null if inquiring child count of the root.
 * @return {number} the number of children for the specified parent.
 * @export
 */
oj.JsonTreeDataSource.prototype.getChildCount = function(parentKey)
{
    var parent;

    if (parentKey == null)
    {
        parentKey = this.data.id;
    }

    parent = this._searchTreeById(this.data, parentKey);

    if (parent.children)
    {
        return parent.children.length;
    }
    else
    {
        return 0;
    }
};

/**
 * Fetch the children
 * @param {*} parentKey the parent key.  Specify null if fetching children from the root.
 * @param {Object} range information about the range, it must contain the following properties: start, count.
 * @param {number} range.start the start index of the range in which the children are fetched.
 * @param {number} range.count the size of the range in which the children are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch children operation is completed.  The valid callback
 *        types are "success" and "error".
 * @param {function(oj.JsonNodeSet)} callbacks.success the callback to invoke when fetch completed successfully.
 * @param {function({status: *})=} callbacks.error the callback to invoke when fetch children failed.
 * @param {Object=} options optional parameters for this operation.
 * @param {boolean=} options.queueOnly true if this fetch request is to be queued and not execute yet.  The implementation must maintain 
 *        the order of the fetch operations.  When queueOnly is false/null/undefined, any queued fetch operations are then
 *        flushed and executed in the order they are queued.  This flag is ignored if the datasource does not support batching.
 * @return {void}
 * @export
 */
oj.JsonTreeDataSource.prototype.fetchChildren = function(parentKey, range, callbacks, options)
{
    var i, childStart, childEnd, nodeSet, results, parent, childCount, node;

    childStart = 0;
    childEnd = 0;
    results = [];

    if (parentKey == null)
    {
        parentKey = this.data.id;
    }

    parent = this._searchTreeById(this.data, parentKey);

    // should never be null but this prevents blowing up if someone tries to call fetchChildren on a leaf
    childCount = parent.children != null ? parent.children.length : 0;

    if (!range)
    {
        range = [];
        range['start'] = 0;
        range['count'] = childCount;
    }

    if (!range['count'])
    {
        range['count'] = childCount;
    }

    if (!range['start'])
    {
        range['start'] = 0;
    }

    childStart = range['start'];
    childEnd = Math.min(childCount, childStart + range['count']);

    // now populate results from data array
    for (i = childStart; i < childEnd; i += 1)
    {
        node = new oj._JsonTreeNodeDataSource();
        if(parent.children[i].attr != null)
        {
            node.attr = parent.children[i].attr;
        }
        if(parent.children[i].id != null)
        {
            node.id = parent.children[i].id;
        }
        if(parent.children[i].depth != null)
        {
            node.depth = parent.children[i].depth;
        }
        if(parent.children[i].title != null)
        {
            node.title = parent.children[i].title;
        }
        if(parent.children[i].parent != null)
        {
            node.parent = parent.children[i].parent;
        }
        if(parent.children[i].children != null)
        {
            node.leaf = false;
        }
        else 
        {
            node.leaf = true;
        }
        results.push(node);
    }

    // invoke callback
    nodeSet = new oj.JsonNodeSet(childStart, childEnd, results, parentKey, parent.depth);

    // invoke original success callback
    if (callbacks != null && callbacks['success'] != null)
    {
        callbacks['success'].call(null, nodeSet);
    }
};

/**
 * Fetch all children and their children recursively from a specified parent.
 * @param {*} parentKey the parent key.  Specify null to fetch everything from the root (i.e. expand all)
 * @param {Object} callbacks the callbacks to be invoke when fetch children operation is completed.  The valid callback
 *        types are "success" and "error".
 * @param {function(oj.JsonNodeSet)} callbacks.success the callback to invoke when fetch completed successfully.
 * @param {function({status: *})=} callbacks.error the callback to invoke when fetch children failed.
 * @param {Object=} options optional parameters for this operation
 * @param {number=} options.start the index related to parent in which to begin fetching descendants from.  If this is not specified, then value zero will be used
 * @param {number=} options.maxCount the maximum number of children to fetch.  If a non-positive number is specified, then the value is ignored and
 *        there is no maximum fetch count
 * @return {void}
 * @export
 */
oj.JsonTreeDataSource.prototype.fetchDescendants = function(parentKey, callbacks, options)
{
    var range, i, childStart, childEnd, nodeSet, results, parent, childCount;

    childStart = 0;
    childEnd = 0;
    results = [];

    if (parentKey == null)
    {
        parentKey = this.data.id;
    }

    parent = this._searchTreeById(this.data, parentKey);

    // should never be null but this prevents blowing up if someone tries to call fetchDescendants on a leaf
    childCount = parent.children != null ? parent.children.length : 0;

    range = [];
    range['start'] = 0;
    range['count'] = childCount;

    childStart = range['start'];
    childEnd = Math.min(childCount, childStart + range['count']);

    // now populate results from data array
    for (i = childStart; i < childEnd; i += 1)
    {       
        if (parent.children[i].children != null)
        {
            parent.children[i].leaf = false;
        }
        else 
        {
            parent.children[i].leaf = true;
        }
        results.push(parent.children[i]); 
    }

    // invoke callback
    nodeSet = new oj.JsonNodeSet(0, results.length, results, parentKey, parent.depth);

    // invoke original success callback
    if (callbacks != null && callbacks['success'] != null)
    {
        callbacks['success'].call(null, nodeSet);
    }
};

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {number|string} position The position of the moved row relative to the reference row.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference row's current children.
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @ojsignature {target: "Type",
 *               value: "'valid'|'invalid'",
 *               for: "returns"}
 * @export
 */ 
oj.JsonTreeDataSource.prototype.moveOK = function(rowToMove, referenceRow, position)
{
    return "valid";
};

/**
 * Moves a node from one location to another (different position within the same parent or a completely different parent)
 * @param {*} nodeToMove the key of the node to move
 * @param {*} referenceNode the key of the reference node which combined with position are used to determine 
 *        the destination of where the node should moved to.
 * @param {number|string} position The position of the moved node relative to the reference node.  
 *        This can be a string: "before", "after", "inside", "first", "last", or the zero based index to position 
 *        the element at a specific point among the reference node's current children.
 * @param {Object} callbacks the callbacks for the move function
 * @param {function():void} callbacks.success the callback to invoke when the move completed successfully.  
 * @param {function({status: *})=} callbacks.error the callback to invoke when move failed.
 * @return {void}
 * @export
 */
oj.JsonTreeDataSource.prototype.move = function(nodeToMove, referenceNode, position, callbacks)
{
    var moveNode, refNode, moveNodeKey, refNodeKey, pos, parent, index;

    pos = position;
    moveNodeKey = nodeToMove;
    refNodeKey = referenceNode;

    if ((refNodeKey == null || refNodeKey == this.data.id))
    {
        if (pos != "inside")
        {
            oj.Logger.log("Error: root can not be the reference node if position equals to " + pos);
            return;
        }
        else
        {
            if (!refNodeKey)
            {
                refNodeKey = this.data.id;
            }
        }
    }

    //get node to move;
    moveNode = this._searchTreeById(null, moveNodeKey);
    //if the moveNode doesn't contain the reference node as its sub-tree the action is allowed
    if (!this._searchTreeById(moveNode, refNodeKey))
    {
        refNode = this._searchTreeById(null, refNodeKey);

        parent = this._getParentById(refNodeKey);
        //remove moveNode from the original position;
        this._removeFromTree(moveNode);
        if (pos == "inside")
        {
            this._updateDepth(moveNode, moveNode.depth - (refNode.depth + 1));
            if (refNode.children == null)
            {
                refNode.children = [];
            }
            refNode.children.push(moveNode);
        }
        else if (pos == "before")
        {
            this._updateDepth(moveNode, moveNode.depth - refNode.depth);
            index = parent.children.indexOf(refNode);
            if (index > -1)
            {
                if (index != 0)
                {
                    parent.children.splice(index, 0, moveNode);
                }
                else
                {
                    parent.children.unshift(moveNode);
                }
            }
        }
        else if (pos == "after")
        {
            this._updateDepth(moveNode, moveNode.depth - refNode.depth);
            index = parent.children.indexOf(refNode);
            if (index > -1)
            {
                parent.children.splice(index + 1, 0, moveNode);
            }
        }
        else if (pos == "first")
        {
            this._updateDepth(moveNode, moveNode.depth - refNode.depth);
            parent.children.unshift(moveNode);
        }
        else if (pos == "last")
        {
            //update depth recursively
            this._updateDepth(moveNode, moveNode.depth - refNode.depth);
            parent.children.push(moveNode);
        }

        // invoke original success callback
        if (callbacks != null && callbacks['success'] != null)
        {
            callbacks['success'].call(null, this.data);
        }
    }
    else
    {
        oj.Logger.log("Error: the node to move contains the reference node as its sub-tree.");
    }
};

/**
 * Performs a sort operation on the tree data.
 * @param {Object} criteria the sort criteria.  It must contain the following properties: key, direction
 * @param {*} criteria.key the key identifying the attribute (column) to sort on
 * @param {'ascending'|'descending'|'none'} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object} callbacks the callbacks for the move function
 * @param {function():void} callbacks.success the callback to invoke when the sort completed successfully.  
 * @param {function({status: *})=} callbacks.error the callback to invoke when sort failed.
 * @return {void}
 * @export
 */
oj.JsonTreeDataSource.prototype.sort = function(criteria, callbacks)
{
    var parent, parentKey;

    parentKey = this.data.id;

    parent = this._searchTreeById(this.data, parentKey);

    parent._sortRecursive(criteria);

    // invoke original success callback
    if (callbacks != null && callbacks['success'] != null)
    {
        callbacks['success'].call(null, parent);
    }
};

/**
 * Returns the current sort criteria of the tree data.
 * @return {Object} the current sort criteria.  It should contain the following properties: key, direction where
 *         criteria.key the key identifying the attribute (column) to sort on.  Value is null if it's not sorted.
 *         criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @ojsignature {target: "Type",
 *               value: "{key: any, direction: 'ascending'|'descending'|'none'}",
 *               for: "returns"}
 * @export
 */
oj.JsonTreeDataSource.prototype.getSortCriteria = function()
{
    return {'key': null, 'direction': 'none'};
};

/**
 * @param {*} refNodeKey
 * @param {Object=} currNode
 * @return {Object|null} the node with required key value.
 * @private
 */
oj.JsonTreeDataSource.prototype._getParentById = function(refNodeKey, currNode)
{
    var i, parent = null, parentIsFound = false;

    if ((refNodeKey == this.data.id))
    {
        return null;
    }

    if (currNode == null)
    {
        currNode = this.data;
    }

    if (currNode.children && currNode.children.length > 0)
    {
        for (i = 0; i < currNode.children.length; i++)
        {
            if ((currNode.children[i].id && currNode.children[i].id == refNodeKey) || (currNode.children[i].attr && currNode.children[i].attr.id == refNodeKey))
            {
                parentIsFound = true;
                return currNode;
                break;
            }
        }
        if (!parentIsFound)
        {
            for (i = 0; i < currNode.children.length; i++)
            {
                parent = this._getParentById(refNodeKey, currNode.children[i]);
                if (parent)
                {
                    parentIsFound = true;
                    return parent;
                    break;
                }
            }

        }
    }
    return parent;
};

/**
 * Helper method to traverse through the tree and return the node with required key.
 * @param {Object|null} currChild the start tree node.
 * @param {*|null} parentKey the node key for search.
 * @return {Object|null} the node with required key value.
 * @private
 */
oj.JsonTreeDataSource.prototype._searchTreeById = function(currChild, parentKey)
{
    var i, result = null;

    if (currChild == null)
    {
        currChild = this.data;
    }

    if ((currChild.id && currChild.id == parentKey) || (currChild.attr && currChild.attr.id == parentKey))
    {
        return currChild;
    }
    else if (currChild.children != null)
    {
        for (i = 0; i < currChild.children.length; i++)
        {
            if (result)
            {
                return result;
                break;
            }
            if ((currChild.children[i].id && currChild.children[i].id == parentKey) || (currChild.children[i].attr && currChild.children[i].attr.id == parentKey))
            {
                result = currChild.children[i];
            }
            else
            {
                result = this._searchTreeById(currChild.children[i], parentKey);
            }
        }
        return result;
    }
    return result;
};

/**
 * Helper method to update the node's depth alongside with its children.
 * @param {Object} currChild the node to update.
 * @param {number} offset the difference between current and updated depth values.
 * @private
 */
oj.JsonTreeDataSource.prototype._updateDepth = function(currChild, offset)
{
    var i;

    currChild.depth = currChild.depth - offset;

    if (currChild.children && currChild.children.length != 0)
    {
        for (i = 0; i < currChild.children.length; i++)
        {
            this._updateDepth(currChild.children[i], offset);
        }
    }
};


/**
 * Helper method to remove node from the tree (based on depth value).
 * @param {Object} currChild the node to remove.
 * @private
 */
oj.JsonTreeDataSource.prototype._removeFromTree = function(currChild)
{
    var parent, index, key;

    if (currChild.id != null)
    {
        key = currChild.id;
    }
    else if (currChild.attr != null)
    {
        key = currChild.attr.id;
    }

    parent = this._getParentById(key);
    if (!parent)
    {
        parent = this.data;
    }
    index = parent.children.indexOf(currChild);
    if (index > -1)
    {
        parent.children.splice(index, 1);
    }
};

/**
 * Determines whether this TreeDataSource supports the specified feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the valid features "sort", 
 *        "move", "fetchDescendants", "batchFetch"
 * @return {string|null} the name of the feature.  Returns null if the feature is not recognized.
 *         For "sort", the valid return values are: "default", "none".  
 *         For "fetchDescendants", the valid return values are: "enable", "disable", "suboptimal".  
 *         For "move", the valid return values are: "default", "none".  
 *         For "batchFetch", the valid return values are: "enable", "disable".  
 * @export
 */
oj.JsonTreeDataSource.prototype.getCapability = function(feature)
{
    if (feature === 'fetchDescendants')
    {
        return "enable";
    }
    else if (feature === 'sort')
    {
        return "default";
    }
    else if (feature === 'batchFetch')
    {
        return "disable";
    }
    else if (feature === 'move')
    {
        return "full";
    }
    else
    {
        return null;
    }
};
/**
 * Create an observable version of a KeySet.
 *
 * @export
 * @param {(ExpandedKeySet|ExpandAllKeySet)=} initialValue The KeySet to observe.
 * @class ObservableExpandedKeySet
 * @classdesc Observable implementation of ExpandedKeySet that keeps track of mutation of KeySet. 
 * @ojtsimport ojkeyset
 * @ojsignature [{target: "Type", value: "class ObservableExpandedKeySet<K>"},
 *               {target: "Type", value: "ExpandedKeySet<K>|ExpandAllKeySet<K>", for:"initialValue"}]
 */
var ObservableExpandedKeySet = function (initialValue) 
{
    // by default if initialValue wasn't specified then we create an ExpandedKeySet
    initialValue = initialValue ? initialValue : new oj.ExpandedKeySet();

    var result = ko.observable(initialValue);
    Object.setPrototypeOf(result, ObservableExpandedKeySet.proto);
    return result;
};

ObservableExpandedKeySet.proto = Object.create(ko.observable['fn']);

// mutation functions
ko.utils.arrayForEach(["add", "addAll", "clear", "delete"], function (methodName) 
{
    ObservableExpandedKeySet.proto[methodName] = function () 
    {
        // Use "peek" to avoid creating a subscription in any computed that we're executing in the context of
        // (for consistency with mutating regular observables)
        var underlyingKeySet = this.peek();
        var methodCallResult = underlyingKeySet[methodName].apply(underlyingKeySet, arguments);
        // this should call valueWillMutate, update latestValue, valueHasMutate
        this(methodCallResult);

        // the mutation methods always return a new KeySet so we should return the ObservableExpandedKeySet itself
        return this;
    };
});

/**
 * Updates the observable with a KeySet that includes the specified keys.
 *
 * @param {Set|Array} keys a set of keys to add to this KeySet.
 * @return {ObservableExpandedKeySet} this observable KeySet.
 * @method
 * @name add
 * @memberof ObservableExpandedKeySet
 * @instance
 * @ojsignature [{target: "Type", value: "Set<K>|Array<K>", for:"keys"},
 *               {target: "Type", value: "ObservableExpandedKeySet<K>", for: "returns"}]
 */

/**
 * Updates the observable with a KeySet that has all keys. 
 *
 * @return {ObservableExpandedKeySet} this observable KeySet.
 * @method
 * @name addAll
 * @memberof ObservableExpandedKeySet
 * @instance
 * @ojsignature {target: "Type", value: "ObservableExpandedKeySet<K>", for: "returns"}
 */

/**
 * Updates the observable with a KeySet that contains no keys.
 *
 * @return {ObservableExpandedKeySet} this observable KeySet.
 * @method
 * @name clear
 * @memberof ObservableExpandedKeySet
 * @instance
 * @ojsignature {target: "Type", value: "ObservableExpandedKeySet<K>", for: "returns"}
 */

/**
 * Updates the observable with a KeySet that excludes the specified keys. 
 *
 * @param {Set|Array} keys a set of keys to remove from this KeySet.
 * @return {ObservableExpandedKeySet} this observable KeySet.
 * @method
 * @name delete
 * @memberof ObservableExpandedKeySet
 * @instance
 * @ojsignature [{target: "Type", value: "Set<K>|Array<K>", for:"keys"},
 *               {target: "Type", value: "ObservableExpandedKeySet<K>", for: "returns"}]
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.KnockoutUtils
 * @classdesc Utility methods for blending Knockout observables with the data model
 * @export
 * @since 1.0
 * @hideconstructor
 * @ojtsimport knockout
 */
oj.KnockoutUtils = function () {};

oj.KnockoutUtils.internalObjectProperty = "oj._internalObj";
oj.KnockoutUtils.underUpdateProp = "oj._underUpdate";
oj.KnockoutUtils.collUpdatingProp = "oj.collectionUpdating";
oj.KnockoutUtils.subscriptionProp = "oj.collectionSubscription";
oj.KnockoutUtils.updatingCollectionFunc = "oj.collectionUpdatingFunc";

/**
 * Turns the attributes in a model object (or the attributes of all models in a collection object) into Knockout observables for use with components' view models.  
 * @param {Object} m The model or collection containing the attributes to be converted to Knockout observables. 
 * @param {function(Object)=} callback Called once per model so that a caller can add their own Knockout computed observables to the model. 
 * @param {boolean=} array Should function return an observable array if m is an oj.Collection, vs. an array of observables?
 * 
 * @return {Object|undefined} array of Knockout observables or an observable array
 * @ojsignature {target: "Type", for: "returns", value: "Array<KnockoutObservable<any>>|KnockoutObservableArray<any>"}
 * @export
 */
oj.KnockoutUtils.map = function (m, callback, array) 
{
  var koObject, i, updateCollection, updateObservableArrayRemove, updateObservableArrayAdd,
      data, prop, converted, updateModel, updateObservable, updateObservableArrayReset, updateObservableArraySort;

  function _makeUpdateModel(argProp) {
      return function(value) {
          //arguments.callee refers to the callback function itself, so we can use the stored '_prop' object (see below)
          if (!koObject[oj.KnockoutUtils.underUpdateProp]) {
              // Make sure we don't circuluarly fire the event
              m.set(argProp, value);
          }
        };
      }
      
  if (m instanceof oj.Collection)
  {
    var prealloc = new Array(m._getLength());
    koObject = array ? ko.observableArray(prealloc) : prealloc;
    
    // Need access to original wrapped collection
    oj.KnockoutUtils._storeOriginalObject(koObject, m);
    
    var index;
    if (array) {
        for (i = 0; i < m._modelIndices.length; i++) 
        {
            index = m._modelIndices[i];
            koObject()[index] = oj.KnockoutUtils.map(m._atInternal(index, null, true, false), callback);
        }
    }
    else {
        for (i = 0; i < m._modelIndices.length; i++) 
        {
            index = m._modelIndices[i];
            koObject[index] = oj.KnockoutUtils.map(m._atInternal(index, null, true, false), callback);
        }
    }            
    // Subscribe to the observable Array, if it is one
    updateCollection = function(changes)
    {
       var modArray, i, len;
       try {
            if (!koObject[oj.KnockoutUtils.underUpdateProp]) {
                koObject[oj.KnockoutUtils.collUpdatingProp] = true;

                // Process change information
                for (i = 0; i < changes.length; i++) {
                    var index = changes[i]['index'];
                    var model = oj.KnockoutUtils._getModel(changes[i]['value']);
                    var status = changes[i]['status'];
                    if (status === 'added') {
                        if (index >= m.length-1) {
                            m.add(model);
                        }
                        else {
                            m.add(model, {'at':index});
                        }
                    }                    
                    else if (status === 'deleted') {
                        m._removeInternal(model, index);
                    }
                }
                if (m['comparator']) {
                    koObject[oj.KnockoutUtils.underUpdateProp] = true;
                    // These could have been resorted--need to alter the observable array
                   koObject.sort(function(a, b) { 
                                        return oj.KnockoutUtils._callSort(a, b, m['comparator'], m, this);
                                    });       
                    koObject[oj.KnockoutUtils.underUpdateProp] = false;
                }
            }
       }
       catch (e) {
           throw e;
       }
       finally {
           koObject[oj.KnockoutUtils.collUpdatingProp] = false;
       }           
    };    
    if (array && koObject['subscribe']) {
        koObject[oj.KnockoutUtils.updatingCollectionFunc] = updateCollection;
        koObject[oj.KnockoutUtils.subscriptionProp] = koObject['subscribe'](updateCollection, null, 'arrayChange');
    }
    
     updateObservableArrayRemove = function(model, collection, options) {
         var index;
         try
         {
            if (koObject[oj.KnockoutUtils.collUpdatingProp]) {
                return;
            }
            if (collection instanceof oj.Collection) {
                koObject[oj.KnockoutUtils.underUpdateProp] = true;
                index = options['index'];
                koObject.splice(index, 1);
            }
         }
         catch (e) {
             throw e;
         }
         finally {
            koObject[oj.KnockoutUtils.underUpdateProp] = false;         
         }
     };
     updateObservableArrayAdd = function(model, collection, options) {
         var index, newObservable, len;
         try
         {
            if (koObject[oj.KnockoutUtils.collUpdatingProp]) {
                return;
            }
            if (collection instanceof oj.Collection) {
                koObject[oj.KnockoutUtils.underUpdateProp] = true;
                index = collection._localIndexOf(model);
                if (index !== undefined && index > -1) {
                    // Make sure to map with original callback from closure, if any
                    newObservable = oj.KnockoutUtils.map(model, callback);
                    if (options['fillIn']) {
                        // First, make sure there's enough room, that index actually exists in koObject...
                        var currLen = Array.isArray(koObject) ? koObject.length : koObject().length;
                        for (var i = currLen; i < index; i++) {
                            koObject.splice(i, 0, oj.KnockoutUtils.map(collection._atInternal(i, null, true, false), callback));
                        }
                        // If we're just filling in on a virtual collection, for example, then just set don't add
                        koObject.splice(index, 1, newObservable);
                    }
                    else {
                        koObject.splice(index, 0, newObservable);
                    }
                }                   
            }
         }
         catch (e) {
             throw e;
         }
         finally {
            koObject[oj.KnockoutUtils.underUpdateProp] = false;         
         }
     };     
     updateObservableArrayReset = function(collection, options) {
         try
         {
            if (koObject[oj.KnockoutUtils.collUpdatingProp]) {
                return;
            }
            if (collection instanceof oj.Collection) {
                koObject[oj.KnockoutUtils.underUpdateProp] = true;
               // koObject.splice(0, koObject().length);
                    if (ko.isObservable(koObject)) {
                        if (koObject[oj.KnockoutUtils.subscriptionProp]) {
                            koObject[oj.KnockoutUtils.subscriptionProp]['dispose']();
                        }
                        koObject.removeAll();
                        if (koObject[oj.KnockoutUtils.updatingCollectionFunc]) {
                            koObject['subscribe'](koObject[oj.KnockoutUtils.updatingCollectionFunc], null, 'arrayChange');
                        }
                    }
                    else {
                        koObject = [];
                    }
            }
         }
         catch (e) {
             throw e;
         }
         finally {
            koObject[oj.KnockoutUtils.underUpdateProp] = false;         
         }
     };     
     
     updateObservableArraySort = function(collection, options) {
         try
         {
            if (koObject[oj.KnockoutUtils.collUpdatingProp]) {
                return;
            }
            if (collection instanceof oj.Collection) {
                koObject[oj.KnockoutUtils.underUpdateProp] = true;
                // Redo the knockout observable array

                koObject.sort(function(a, b) { 
                                return oj.KnockoutUtils._callSort(a, b, m['comparator'], collection, this);
                             });   
            }
         }
         catch (e) {
             throw e;
         }
         finally {
            koObject[oj.KnockoutUtils.underUpdateProp] = false;         
         }
     };
     
     // Register these, ignoring the silent flags from normal API calls--knockout always needs to be updated
     m.OnInternal(oj.Events.EventType['ADD'], updateObservableArrayAdd, undefined, undefined, true);    
     m.OnInternal(oj.Events.EventType['REMOVE'], updateObservableArrayRemove, undefined, undefined, true);    
     m.OnInternal(oj.Events.EventType['RESET'], updateObservableArrayReset, undefined, undefined, true);
     m.OnInternal(oj.Events.EventType['SORT'], updateObservableArraySort, undefined, undefined, true);
  }
  else
  {
    if (m === undefined) {
        return;
    }
    
    koObject = {};
    
    data = m.attributes;
    prop = null;
    
    for (prop in data)
    {
      if (data.hasOwnProperty(prop)) {
        converted = ko.observable(m.get(prop));
        koObject[prop] = converted;

        updateModel = _makeUpdateModel(prop);
        updateModel._prop = prop;

        // Subscribe to any changes pushed by Knockout into the observable
        if (converted['subscribe']) {
            converted['subscribe'](updateModel);
        }
      }
     }

     updateObservable = function(model, options) {
         var attrs, prop;
         try
         {
            koObject[oj.KnockoutUtils.underUpdateProp] = true;
            attrs = model.changedAttributes();
            for (prop in attrs) {
                if (attrs.hasOwnProperty(prop)) {
                 koObject[prop](model.get(prop));
                }
            }
         }
         catch (e) {
             throw e;
         }
         finally {
            koObject[oj.KnockoutUtils.underUpdateProp] = false;         
         }
     };
     m.OnInternal(oj.Events.EventType['CHANGE'], updateObservable, undefined, undefined, true);
     // Need access to original wrapped model
    oj.KnockoutUtils._storeOriginalObject(koObject, m);

    // Activate user callback if specified, to allow user computed observables, etc.
    if (callback) {
        callback(koObject);
    }
  }
  
  return koObject;     
};

oj.KnockoutUtils._getModel = function(val) {
    if (val instanceof oj.Model) {
        return val;
    }
    
    if (val.hasOwnProperty(oj.KnockoutUtils.internalObjectProperty)) {
        return val[oj.KnockoutUtils.internalObjectProperty];
    }
    return val;
};

oj.KnockoutUtils._callSort = function(a, b, comparator, collection, caller) {
    return oj.Collection.SortFunc(oj.KnockoutUtils._getModel(a), oj.KnockoutUtils._getModel(b), comparator, collection, caller);
};
    

// Attempt to hide original object from enumeration of properties
oj.KnockoutUtils._storeOriginalObject = function(object, value) {
    // Store any callback along with model for use in event-driven mapping of new additions
       Object.defineProperty(object, oj.KnockoutUtils.internalObjectProperty, {value: value, enumerable:false});
};

/*oj.KnockoutUtils._createArray = function(collection, getCallback) {
    if (collection._isVirtual()) {
        return oj.KnockoutUtils._augment(collection.length, getCallback);
    }
    return [];
};

oj.KnockoutUtils._augment = function(len, getCallback) {
    var array;
    if (len.constructor === Array) {
        array = new Array(len.length);
        array.storage = len;
    }
    else {
        array = new Array(len);
        array.storage = new Array(len);
    }
    function makePropDef(arr) {
        return function (index, callback) {
                Object.defineProperty(arr, index.toString(), {
                    configurable:true,
                    get: function() { 
                        // Someone is accessing this location: if not set, set it using the callback before returning
                        if (!arr.storage[index]) {
                            arr.storage[index] = callback.call(this, index);
                        }
                        return arr.storage[index];
                    },
                    set: function(val) {
                        arr.storage[index] = val;
                    }
                });
            };
    }

    array.getCallback = getCallback;
        
    function setup(arr) {
        var i;
        arr.length = arr.storage.length;
        for (i = 0; i < arr.storage.length; i++) {
            makePropDef(arr)(i, arr.getCallback);
        }
    }
    setup(array);
    
    array.length = array.storage.length;
    
    array['pop'] = function() {
        var retVal = this.storage.pop();
        this.length = this.storage.length;
        return retVal;
    };
    array['push'] = function(val) {
        var index = this.length;
        this.storage.push(val);
        Object.defineProperty(this, index.toString(), {
                configurable:true,
                get: function() { 
                        // Someone is accessing this location: if not set, set it using the callback before returning
                        if (!this.storage[index]) {
                            this.storage[index] = this.getCallback.call(this, index);
                        }
                        return this.storage[index];
                },
                set: function(val) {
                    this.storage[index] = val;
                }
            });    
        return this.length;
    };
    array['shift'] = function() {
        var self = this;
        var retVal = self.storage.shift();
        setup(self);
        return retVal;
    };
    array['unshift'] = function() {
        var args = Array.prototype.slice.call(arguments), i;
        for (i = 0; i < args.length; i++) {
            this.storage.unshift(args[i]);
        }
        setup(this);
        return this.length;
    };
    array['reverse'] = function() {
        this.storage = this.storage.reverse();
        setup(this);
        return this;
    };
    array['slice'] = function(start, end) {
        var newArr = oj.KnockoutUtils._augment(this.storage.slice(start, end), this.getCallback);
        return newArr;
    };
    array['splice'] = function() {
        var args = Array.prototype.slice.call(arguments);    
        // Just redo the array after modifying storage
        Array.prototype.splice.apply(this.storage, args);
        setup(this);
    };
    return array;
};*/
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

(function ()
{
  /**
  * String used in the id on the span that surrounds the help icon.
  * @const
  * @private
  * @type {string}
  */
  var _HELP_ICON_ID = "_helpIcon";
  /**
  * String used in the id on the span that surrounds the required icon.
  * @const
  * @private
  * @type {string}
  */
  var _REQUIRED_ICON_ID = "_requiredIcon";
  /**
  * String used to append to the 'for' attribute on the label element
  * @const
  * @private
  * @type {string}
  */
  var _FORM_INPUT_SUFFIX = "|input";
  /**
  * aria-describedby
  * @const
  * @private
  * @type {string}
  */
  var _ARIA_DESCRIBEDBY = "aria-describedby";
 /**
  * described-by
  * @const
  * @private
  * @type {string}
  */
  var _DESCRIBED_BY = "described-by";  
  /**
  * aria-labelledby
  * @const
  * @private
  * @type {string}
  */
  var _ARIA_LABELLEDBY = "aria-labelledby";
 /**
  * labelled-by
  * @const
  * @private
  * @type {string}
  */
  var _LABELLED_BY = "labelled-by";   
  /*!
   * JET oj-label. @VERSION
   */
  /**
   * <p>
   * The oj-label component decorates the label text with a required icon and help icon. The user
   * can interact with the help icon (on hover, on focus, etc) to display help description text 
   * or to navigate to an url for more help information.
   * </p>
   * <p>For accessibility reasons, you need to associate the oj-label component to its
   * JET form component. For most JET form components you do this
   * using the oj-label's <code class="prettyprint">for</code> attribute and 
   * the JET form component's <code class="prettyprint">id</code> attribute.
   * For a few JET form components
   * (oj-radioset, oj-checkboxset, oj-color-palette, and oj-color-spectrum)
   * you associate the oj-label component to its JET form component using
   * the oj-label's <code class="prettyprint">id</code> attribute and the 
   * JET form component's <code class="prettyprint">labelled-by</code> attribute. For more examples,
   * see the label demos which show all the JET form controls with oj-label or refer to the JET
   * form component's API jsdoc.
   * <pre class="prettyprint">
   * <code>
   * &lt;oj-label for="inputtextid" show-required="[[isRequired]]" 
   * help.definition="[[helpDef]]" help.source="[[helpSource]]">input label&lt;/oj-label>
   * &lt;oj-input-text id="inputtextid" required="[[isRequired]]">&lt;//oj-input-text>
   * 
   * &lt;oj-label id="radiosetlabel" show-required="[[isRequired]]">radioset&lt;/oj-label>
   * &lt;oj-radioset required="[[isRequired]]" labelled-by="radiosetlabel">
   *   &lt;oj-option name="color" value="red">Red&lt;/oj-option>
   *   &lt;oj-option name="color" value="blue">Blue&lt;/oj-option>
   * &lt;/oj-radioset>
   * 
   * &lt;!--  You can bind the text as a child comment node or on a span element, but not on the 
   * oj-label element. The knockout text binding is not supported on a JET custom element; -->
   * &lt;oj-label for="input2">&lt;!--ko text: input2Label -->&lt;!--/ko-->&lt;/oj-label>
   * </code>
   * </pre>
   * <p>
   * </p>
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   * 
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDoc"}
   * @ojstatus preview
   * @ojcomponent oj.ojLabel
   * @ojshortdesc Provides support for 'required' and 'help' icons on form field labels.
   * @since 4.0.0
   * @augments oj.baseComponent
   */
  oj.__registerWidget("oj.ojLabel", $['oj']['baseComponent'],
  {
    version: "1.0.0",
    defaultElement: "<label>",
    widgetEventPrefix: "oj",
    options:
    {
      /**
       * <p>
       * The <code class="prettyprint">for</code> attribute refers to the id of the 
       * element this oj-label element is associated with.
       * Some JET form components support using oj-label's <code class="prettyprint">for</code>
       *  attribute to point to its <code class="prettyprint">id</code> attribute
       * (e.g., oj-input-text, oj-slider), and others do not (e.g., oj-checkboxset). 
       * <p>
       * For the oj-radioset, oj-checkboxset, oj-color-palette, and oj-color-spectrum components, 
       * instead of the <code class="prettyprint">for</code> attribute, use the 
       * <code class="prettyprint">id</code> attribute on oj-label 
       * with JET's form element's <code class="prettyprint">labelled-by</code> attribute.
       * </p>
       * <p>
       * Refer to the JET's form element's documentation/demos for more examples showing the 
       * use of <code class="prettyprint">for</code>/<code class="prettyprint">id</code> 
       * and the use of <code class="prettyprint">id</code>/<code class="prettyprint">labelled-by</code>.
       * </p>
       *
       * @expose
       * @memberof oj.ojLabel
       * @instance
       * @type {string|null}
       * @default null
       * @since 4.0.0
       *
       * @example <caption>Associate oj-label to the oj-input-text using the 
       * <code class="prettyprint">for</code> attribute on oj-label and the 
       * <code class="prettyprint">id</code> attribute on oj-input-text during initialization.</caption>
       * &lt;oj-label for="inputId">Name:&lt;/oj-label>
       * &lt;oj-input-text id="inputId">&lt;/oj-input-text>
       *
       */
      for: null,
      /**
       * The help information that goes on the oj-label.  The help attributes are:
       * <ul>
       * <li><code class="prettyprint">definition</code> - this is the help definition text. 
       * It is what shows up
       * when the user hovers over the help icon, or tabs into the help icon, or press
       * and holds the help icon on a mobile device. No formatted text is available for
       * help definition attribute.</li>
       * <li><code class="prettyprint">source</code> - this is the help source url.
       * If present, a help icon will
       * render next to the label. For security reasons
       *  we only support urls with protocol http: or https:.
       * If the url doesn't comply we ignore it and throw an error.
       * Pass in an encoded URL since we do not encode the URL.
       * </ul>
       *
       * @expose
       * @memberof oj.ojLabel
       * @instance
       * @type {Object.<string, string>|null}
       * @default {'definition' :null, 'source': null}
       * @since 4.0.0
       *
       * @example <caption>Initialize the label with the help definition and help source:</caption>
       * &lt;!-- Using dot notation -->
       * &lt;oj-label help.definition="some help definition" help.source="some external url">Name:&lt;/oj-label>
       * &lt;!-- Using JSON notation -->
       * &lt;oj-label help='{"definition":"some value", "source":"someurl"}'>Name:&lt;/oj-label>
       * @example <caption>Set the <code class="prettyprint">help</code> attribute, after initialization:</caption>
       *
       * // Set one, leaving the others intact. Use the setProperty API for 
       * // subproperties so that a property change event is fired.
       * myComponent.setProperty('help.definition', 'some new value');
       * // Get all
       * var values = myComponent.help;
       * // Set all.  Must list every key, as those not listed are lost.
       * myComponent.help = {
       *   definition: 'some new value',
       *   source: 'some new url'
       * };
       * @property {(string|null)=} definition help definition text
       * @property {(string|null)=} source help source url
       */
      help:
      {
        definition: null,
        source: null
      },
      /**
       * <code class="prettyprint">label-id</code> sets the <code class="prettyprint">id</code> 
       * attribute on the internal label element.  The use case where this may be needed
       * is if you are using a <code class="prettyprint">&lt;div aria-labelledby></code> 
       * and for accessibility reasons you need to point to the 
       * oj-label's <code class="prettyprint">label</code>
       * element. This should be a corner case. Most often you'd use oj-label's 
       * <code class="prettyprint">for</code> attribute to
       * associate with a form component's id attribute or use oj-label's 
       * <code class="prettyprint">id</code> attribute to 
       * associate with a JET form component's <code class="prettyprint">labelled-by</code> attribute.
       * @expose
       * @type {string|null}
       * @default null
       * @public
       * @instance
       * @since 4.0.0 
       * @memberof oj.ojLabel
       * 
       * @example <caption>Initialize the label with the 
       * <code class="prettyprint">label-id</code> attribute:</caption>
       * &lt;oj-label label-id="labelId">Name:&lt;/oj-label>
       * @example <caption>Set the attribute, after initialization:</caption>
       * // getter
       * var labelid = myOjLabel.labelId;
       * // setter
       * myOjLabel.labelId = "myLabelId";
       */
      labelId: null,
      /**
       * Whether this label should have a required icon. It is recommended that you
       * bind the <code class="prettyprint">show-required</code> attribute to the same binding 
       * as the <code class="prettyprint">required</code> attribute on the
       * associated JET form component to make sure they are in sync.
       * @expose
       * @type {?boolean}
       * @default false
       * @public
       * @instance
       * @memberof oj.ojLabel
       * @since 4.0.0
       * 
       * @example <caption>Initialize the oj-label with the 
       * <code class="prettyprint">show-required</code> attribute, binding form component's
       * <code class="prettyprint">required</code> attribute to the same value.</caption>
       * &lt;oj-label show-required="[[isRequired]]">Name:&lt;/oj-label>
       * &lt;oj-input-text required="[[isRequired]]">&lt;/oj-input-text>
       * @example <caption>Set the attribute, after initialization:</caption>
       * // getter
       * var showRequired = myOjLabel.showRequired;
       * // setter
       * myOjLabel.showRequired = false;
       */
      showRequired: false,
      /**
       * Allows you to set certain attributes on the root dom element.
       * For ojLabel, we use 'class' only. The input components (via
       * EditableValue) set a styleclass on the ojLabel's root in case
       * component-specific label styling is needed. For example, ojradioset
       * would pass class: 'oj-radioset-label'. ojinputtext would pass class:
       * 'oj-inputtext-label'.
       *
       * @example <caption>Initialize root dom element with the set of
       * <code class="prettyprint">rootAttributes</code>:</caption>
       * $(".selector").ojLabel("option", "rootAttributes", {
       *   'class': 'oj-inputtext-label'
       * });
       *
       * @expose
       * @access public
       * @memberof oj.ojLabel
       * @instance
       * @type {Object}
       * @default {'id': null, 'class': null, 'style':null}
       * @ignore
       */
      rootAttributes: null
    },
    /**
     * @private
     * @memberof oj.ojLabel
     * @const
     */
    _BUNDLE_KEY:
    {
      _TOOLTIP_HELP: 'tooltipHelp',
      _TOOLTIP_REQUIRED: 'tooltipRequired'
    },
    /**** start Public APIs ****/

    /**
     * Returns a jQuery object containing the root dom element of the label
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof oj.ojLabel
     * @instance
     * @return {jQuery} the label
     * @ignore
     */
    widget: function ()
    {
      return this.uiLabel;
    },
    /**
     * Refreshes the component with all the current attributes.
     * <p>
     * Call refresh if the <code class="prettyprint">id</code> changes, though 
     * changing the <code class="prettyprint">id</code> shouldn't be needed.
     * </p>
     * <p>
     * Also, call refresh after <code class="prettyprint">required</code>
     * or <code class="prettyprint">help</code> changes.
     * The locale could have changed in the meantime, and refresh is needed to update the 
     * <code class="prettyprint">required</code> and 
     * <code class="prettyprint">help</code> tooltips.
     * </p>
     * There should be no need to call refresh for other attribute changes.
     *     
     * @example document.getElementById("label1").setProperty("id","label2");
     * document.getElementById("label2").refresh();
     * @access public
     * @instance
     * @return {void}
     * @expose
     * @memberof oj.ojLabel
     */
    refresh: function ()
    {
      this._super();
      this._refreshRequired();
      this._refreshHelp();
      this._addIdsToDom();
    },
    /**** end Public APIs ****/

    /**** start internal widget functions ****/
    /**
     * Overridden to make sure describedById option is set
     *
     * @memberof oj.ojLabel
     * @instance
     * @protected
     */
    _InitOptions: function (originalDefaults, constructorOptions)
    {
      this._super(originalDefaults, constructorOptions);
      this._isCustomElement = this._IsCustomElement();
      this._checkRequiredOption();
    },
    /**
     * After _ComponentCreate and _AfterCreate,
     * the widget should be 100% set up. this._super should be called first.
     * @override
     * @protected
     * @instance
     * @memberof oj.ojLabel
     */
    _ComponentCreate: function ()
    {
      this._super();

      this._touchEatClickNamespace = this.eventNamespace + "TouchEatClick";
      this._touchEatContextMenuNamespace = this.eventNamespace + "TouchEatContextMenu";
      this._helpDefPopupNamespace = this.eventNamespace + "HelpDefPopup";
      this._bTouchSupported = oj.DomUtils.isTouchSupported();
      


      this._drawOnCreate();
    },
    /**
     * <p>Save only the 'class' attribute since that is what
     * we manipulate. We don't have to save all the attributes.
     * N/A for custom element oj-label, so if custom element, this returns right away.
     *
     * @param {Object} element - jQuery selection to save attributes for
     * @protected
     * @memberof oj.ojLabel
     * @instance
     * @override
     */
    _SaveAttributes: function (element)
    {
      // _SaveAttributes is called before _InitOptions, so we can't use our global _isCustomElement
      // variable here.
      if (this._IsCustomElement())
        return;
      
      this._savedClasses = element.attr("class");
    },
    /**
     * <p>Restore what was saved in _SaveAttributes.      
     * N/A for custom element oj-label, so if custom element, this returns right away.
     * </p>
     *
     * @protected
     * @memberof oj.ojLabel
     * @instance
     * @override
     */
    _RestoreAttributes: function ()
    {
      if (this._IsCustomElement())
        return;
      
      // restore the saved "class" attribute. Setting class attr to undefined is a no/op, so
      // if this._savedClasses is undefined we explicitly remove the 'class' attribute.
      if (this._savedClasses)
        this.element.attr("class", this._savedClasses);
      else
        this.element.removeAttr("class");
    },
    /**
     * Notifies the component that its subtree has been removed from the document programmatically
     * after the component has been created
     * @memberof! oj.ojLabel
     * @instance
     * @protected
     */
    _NotifyDetached: function ()
    {
      this._superApply(arguments);
      this._handleCloseHelpDefPopup();
    },
    /**
     * Notifies the component that its subtree has been made hidden programmatically
     * after the component has been created
     * @memberof! oj.ojLabel
     * @instance
     * @protected
     */
    _NotifyHidden: function ()
    {
      this._superApply(arguments);
      this._handleCloseHelpDefPopup();
    },
    /**
     * The translation section name for the private ojLabel() is oj-ojLabel. 
     * In 4.0.0 and the introduction of the oj-label custom element,
     *  the translation section name is oj-ojLabel.
     * @protected
     * @override
     * @instance
     * @memberof oj.ojLabel
     */
    _GetTranslationsSectionName: function()
    {
      if (!this._IsCustomElement())
        return "oj-ojLabel";
      else
        return "oj-ojLabel";
    },
    /**
     * set up dom and styles on create
     * @private
     * @memberof oj.ojLabel
     */
    _drawOnCreate: function ()
    {
      var helpSpan = null;
      var labelElementId;
      var forTargetElement;
      
      // custom element is <oj-label>. this.element is the <label> element.
      // <oj-label class="oj-label oj-component">
      //    <div class="oj-label-group">(helpandrequired spans)<label></label>
      // </oj-label>
      if (this.OuterWrapper) 
      {
        this.uiLabel = $(this.OuterWrapper).append(
          this.element.wrap(this._createOjLabelGroupDom()).parent()); // @HTMLUpdateOk     
        this.uiLabel.addClass("oj-label oj-component");
      }
      else 
      {
         // wrap the label with a root dom element (oj-label) and its child
        // (oj-label-group). Then point this.uiLabel to the root dom element <div class=oj-label>
        this.uiLabel = this.element.wrap(this._createRootDomElement()) // @HTMLUpdateOK
          .closest(".oj-component");
      }
      
      // adds id to the label element from label-id or id or generates one (in that order).
      // used for both ojLabel and oj-label.
      this._addIdsToDom();

      // use the label element's id to create the help span and required span ids
      labelElementId = this.element[0].id;
      this.helpSpanId = labelElementId + _HELP_ICON_ID;
      this.requiredSpanId = labelElementId + _REQUIRED_ICON_ID;

      // move any oj-label styles off of this.element, and put on the
      // root dom element. They are restored in _destroy
      if (!this._isCustomElement)
        this._moveLabelStyleClassesToRootDom();
      
      // we put a span with an id on it around the help icon and 
      // a span with an id on it around the required icon so that
      // the input's aria-describedby can point to it, if needed. Then the screen reader will
      // read the aria-label on the images when focus is on the input, so the user knows
      // that there is help and/or required icons.
      if (this.options.showRequired)
      {
        this._createRequiredIconSpanDom();
        
        // adds described-by option with required icon span id onto ojradioset and ojcheckboxset
        // since they don't support aria-required.
        if (this._isCustomElement)
          this._addRequiredDescribedByOnCustomFormElement();
      }

      if (this._isCustomElement)
      {
        // if oj-label has the 'for' option, then add the 'for' onto the label element. It may
        // be a for sub-id depending upon if the target element is a oj- element or not.
        this._refreshFor();
      }
      
      if (this._needsHelpIcon())
      {
        helpSpan = this. _createIconSpan(this.helpSpanId, true);
        this._createHelp(helpSpan);
        
        // adds described-by option with help icon span id onto the JET form component,
        // or add aria-describedby with help icon span id onto native html element.
        if (this._isCustomElement)
          this._addHelpSpanIdOnTarget(this.helpSpanId);
      }
    },
    /**  
    * If this is an oj-label, then it can point to a regular HTML element or a custom JET component,
    * like oj-input-text. This returns the custom JET component or the regular HTML element
    * that is linked to the oj-label.
    * @private
    * @instance
    * @memberof oj.ojLabel
    * @return {Element|null} forTargetElement is a reference to an Element object, 
    * or null if the element with id equal to the label's for attribute is not found.
    * 
    */
    _getCustomLabelFormElement: function()
    {
      var id;
      var targetElement = null;
      
      if (!this._isCustomElement)
        return null;
      
      // targetElement is the element with id === oj-label's 'for' option
      targetElement = this._getForTargetElement();
     
      if (targetElement == null)
      {
        // if there isn't a 'for' option or the target wasn't found, 
        // then if there is an id option, find the 
        // element with labelled-by set, and add describedby on it.
        // else, if there is a label-id option, find the element
        // with aria-labelledby pointing to it
        if (this.OuterWrapper)
          id = this.OuterWrapper.id;

        targetElement = this._getTargetElementFromLabelledAttr(_LABELLED_BY, id);
        
        if (targetElement == null)
        {
          targetElement = 
            this._getTargetElementFromLabelledAttr(_ARIA_LABELLEDBY, this.options.labelId);
          }
      }
      
      return targetElement;
    },
    /**  
    * For all form controls except radioset
    * and checkboxset, aria-required on the form control reads required. So no need to
    * put aria-describedby on those.
    * If oj-label, then add described-by for the required icon on the custom form element.
    * This will be oj-checkboxset or oj-radioset only.
    * @private
    * @memberof oj.ojLabel
    * @instance
    * 
    */
    _addRequiredDescribedByOnCustomFormElement : function()
    {
      var id;
      var targetElement;

      if (this.OuterWrapper)
        id = this.OuterWrapper.id;
      
      targetElement = this._getTargetElementFromLabelledAttr(_LABELLED_BY, id);
      
      if (targetElement && this._isElementCustomElementAriaRequiredUnsupported(targetElement))
        this._addElementAttribute(targetElement, this.requiredSpanId, _DESCRIBED_BY);
    },
    /**  
    * Get's the oj-label's id, then finds the target element where the labelled-by option is set
    * to the oj-label's id. Then it checks that this is an oj-radioset or an oj-checkboxset, since
    * those are the only two JET form controls that have the required option and don't support
    * the aria-required attribute, and therefore needs the aria-describedby pointing to the 
    * requiredIcon span.
    * @private
    * @instance
    * @memberof oj.ojLabel
    * 
    */
    _removeRequiredDescribedByOnCustomFormElement : function()
    {
      var id;
      var targetElement;

      if (this.OuterWrapper)
        id = this.OuterWrapper.id;
      targetElement = this._getTargetElementFromLabelledAttr(_LABELLED_BY, id);
      
      if (targetElement && this._isElementCustomElementAriaRequiredUnsupported(targetElement))
        this._removeElementAttribute(targetElement, this.requiredSpanId, _DESCRIBED_BY);
    },
   /**  
    * Call for custom oj-label element only. Adds aria-describedby on native form element or 
    * described-by on JET form element with the help icon span's id or aria-labelledby on
    * div role='group'..
    * @private
    * @instance
    * @memberof oj.ojLabel
    * @param {string} helpSpanId The id of the help span
    * 
    */
    _addHelpSpanIdOnTarget : function(helpSpanId)
    {
      var attributeName;
      var targetElement = this._getCustomLabelFormElement();

      if (targetElement)
      { 
        attributeName = this._getAriaAttributeForTarget(targetElement);
        this._addElementAttribute(targetElement, helpSpanId, attributeName);
      }
    },
   /**  
    * Call for custom oj-label element only. Removes aria-describedby on native form element or 
    * described-by on JET form element with the help icon span's id or aria-labelledby on
    * div role='group'.
    * @private
    * @memberof oj.ojLabel
    * @instance
    */
    _removeHelpSpanIdOnTarget : function(helpSpanId)
    {
      var attributeName;
      var targetElement = this._getCustomLabelFormElement();
      
      if (targetElement)
      {  
        attributeName = this._getAriaAttributeForTarget(targetElement);
        this._removeElementAttribute(targetElement, helpSpanId, attributeName);
      }
    },
   /**
    * Based on the targetElement, this figures out which attribute we need to write onto 
    * the target to make it be able to read out the help definition. E.g., described-by for
    * JET form control, aria-describedby for input, aria-labelled-by for div role='group'
    * @param {Element} targetElement
    * @return {string|null} the attribute, aria-describedby, aria-labelledby, or described-by
    * @private
    * @memberof oj.ojLabel
    * @instance
    */
    _getAriaAttributeForTarget : function(targetElement)
    {
      var attributeName;
      
      if(!this._isElementCustomElement(targetElement))
      {
        // We handle the one label to role='group' differently. 
        // If a <div aria-labelledby role='group'> points to the
        // <oj-label>'s label-id, then add/remove aria-labelledby, not aria-describedby, since
        // screenreaders don't read aria-describedby on divs.
        if (targetElement.getAttribute("role") === "group")
          attributeName = _ARIA_LABELLEDBY;
        else
          attributeName = _ARIA_DESCRIBEDBY;
      }
      else 
      {
        attributeName = _DESCRIBED_BY;
      }
  
      return attributeName;
    },    
    /**  
    * Return the target element of the 'for' option. Called for custom element oj-label.
    * @private
    * @instance
    * @memberof oj.ojLabel
    * @return {Element|null} is a reference to an Element object, 
    * or null if the element with id equal to the label's for attribute is not found.
    * 
    */
    _getForTargetElement : function()
    {
      return document.getElementById(this.options.for);
    },
   /**  
    * Return the element with by using a querySelector to search for the attribute with the id.
    * e.g., aria-labelledby attribute set to labelId.
    * @private
    * @instance
    * @memberof oj.ojLabel
    * @param attrName the attribute name to look for. e.g., labelled-by or aria-labelledby
    * @param {string} id
    * @return {Element|null} is a reference to an Element object, 
    * or null if the element is not found.
    * 
    */
    _getTargetElementFromLabelledAttr : function(attrName, id)
    {
      var attributeSearchString;
      
      if (id)
      {
        // The ~ means to look to see if the id is in a list of whitespace-separated
        // values, one of which is exactly equal to id, and labelled-by could have a
        // list of whitespace-separated ids.
        attributeSearchString = "["+attrName+"~='" + id + "']";
        return document.querySelector(attributeSearchString);
      }
      else
        return null;
    },
    /**  
    * If targetElement's tagName has a "-", return true, else return false.
    * @param {Element} targetElement the HTML element the 'for' attribute is pointing to
    * @private
    * @memberof oj.ojLabel
    * @instance
    * @return {boolean} return true if targetElement's tagName contains "-". Else return false.
    * 
    */
    _isElementCustomElement : function(targetElement)
    {
      oj.Assert.assertDomElement(targetElement);
      return (targetElement.tagName).indexOf('-') !== -1;
    },
    /**  
    * Checks the targetElement's tagname against oj-checkboxset or oj-radioset, because these
    * are the only components that do not support aria-required.
    * @param {Element} targetElement the HTML element the 'for' attribute is pointing to
    * @private
    * @memberof oj.ojLabel
    * @instance
    * @return {boolean} true if targetElement's tagName is oj-radioset or oj-checkboxset.
    * 
    */
    _isElementCustomElementAriaRequiredUnsupported : function(targetElement)
    {
      oj.Assert.assertDomElement(targetElement);
      
      var componentName;
      var componentNames = ["oj-radioset", "oj-checkboxset"];
      var length = componentNames.length;
      var tagName;
      var found = false;

      if (targetElement)
      {
        tagName = targetElement.tagName.toLowerCase();
        
        for(var i=0; i < length && !found; i++)
        {
          componentName = componentNames[i];
          if (tagName.indexOf(componentName) === 0)
            found = true;
        }
      }
      
      return found;
    },
    /**  
    * Generic function to add a value to the element's attribute. For example, you can
    * add "foo" to the elem's aria-describedby attribute. It will not overwrite
    * any current aria-describedby values already there.
    * 
    * @param {Element} elem the HTML element on which to add the attribute
    * @param {string} id the id to add to the attribute
    * @param {string} attr the attribute to add to. e.g., "aria-describedby"
    * @private
    * @memberof oj.ojLabel
    * @instance
    * 
    */
   _addElementAttribute : function (elem, id, attr)
   {
      var index;
      var currentAttributeValue = elem.getAttribute(attr);
      var newAttributeValue;
      var tokens;

      tokens = currentAttributeValue ? currentAttributeValue.split(/\s+/) : [];
      // Get index that id is in the tokens, if at all.
      index = $.inArray(id, tokens);
      // add id if it isn't already there
      if (index === -1)
        tokens.push(id);
      newAttributeValue = $.trim(tokens.join(" "));
      elem.setAttribute(attr, newAttributeValue);
   },
  /**
   * Generic function to remove a value to the element's attribute. For example, you can
   * remove "foo" from the elem's aria-describedby attribute. It will not remove
   * any other aria-describedby values already there.
   * @param {Element} elem the HTML element
   * @param {string} id the id to remove from the attr
   * @param {string} attr the attribute to remove from. e.g., "aria-describedby"
   * @private
   * @memberof oj.ojLabel
   * @instance
   */
  _removeElementAttribute : function (elem, id, attr)
  {
    var currentAttributeValue = elem.getAttribute(attr);
    var newAttributeValue;
    var index;
    var tokens;

    // get aria-describedby that is on the element
    // split into tokens
    tokens = currentAttributeValue ? currentAttributeValue.split(/\s+/) : [];
    // Get index that id is in the tokens, if at all.
    index = $.inArray(id, tokens);
    // remove that from the tokens array
    if (index !== -1)
      tokens.splice(index, 1);
    // join the tokens back together and trim whitespace
    newAttributeValue = $.trim(tokens.join(" "));
    if (newAttributeValue)
      elem.setAttribute(attr, newAttributeValue);
    else
      elem.setAttribute(attr, "");

   },
   /**
    * Get the 'for' attribute from oj-label and copy it to the label element, modifying it if needed.
    * Do not call this if it isn't a custom element and the forTargetElement is null
    * <p>
    * If the form element is an oj- element, then set the 'for' attribute on the label element by
    * taking oj-label's 'for' attribute and appending _FORM_INPUT_SUFFIX to it.
    * else, copy the 'for' attribute to the label element untouched. Both of these need to work:
    * @example &lt;oj-label for="nameId">Name:&lt;/oj-label>
    *       &lt;oj-input-text id="nameId">&lt;/oj-input-text>
    *       
    *       
    * @example &lt;oj-label for="nameId">Name:&lt;/label>
    *       &lt;input id="nameId">&lt;/input>
    *       
    * </p>
    * @private
    * @memberof oj.ojLabel
    * @instance
    * @param {Element} forTargetElement
    */
    _addForOnLabel: function (forTargetElement)
    {
      var labelElement = this.element[0];;
      var forOption = this.options.for;

      // only add the _FORM_INPUT_SUFFIX if we are pointing to a JET custom element,
      // otherwise copy it as is.
      if (this._isElementCustomElement(forTargetElement))
        labelElement.setAttribute("for", forOption + _FORM_INPUT_SUFFIX);
      else
        labelElement.setAttribute("for", forOption);

    },
    /**
     * Create help.
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _createHelp: function (helpSpan)
    {
      var helpDef = this.options.help['definition'];
      var helpSource = this.options.help['source'];
      var helpIconAnchor = this._createHelpIconAnchorDomElement(helpDef, helpSource);
      // .prepend: Insert content, specified by the parameter, to the beginning of each element
      $(helpSpan).prepend(// @HTMLUpdateOK
        helpIconAnchor);
      this._attachHelpDefToIconAnchor();
      this._focusable({'element': helpIconAnchor, 'applyHighlight': true});
    }, 
    /**
     * Create required icon span dom and its contents
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _createRequiredIconSpanDom: function ()
    {
      //_createIconSpan creates the span and adds it in the oj-label-group dom
      var requiredSpan = this. _createIconSpan(this.requiredSpanId, false);      
      requiredSpan.appendChild(this._createRequiredIconDomElement()); // @HTMLUpdateOK

    },
    /**
     * @throws error if showRequired is not a boolean
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _checkRequiredOption: function ()
    {
      var showRequired = this.options.showRequired;

      if (showRequired !== null && typeof showRequired !== "boolean")
        throw new Error("Option 'showRequired' has invalid value set: " + showRequired);
    },
    /**
     * Add an id to the label element if there isn't one already there. 
     * Called during init and refresh.
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _addIdsToDom: function ()
    {
      var labelElementId;
      
      // Use label-id to set label element's id. If label-id doesn't exist, this uses oj-label's
      // id to generate a sub-id on label.
      if (this._isCustomElement)
      {
        this._refreshLabelId();
      }

      // if no id on the label element, generate one.
      // this will be used to wrap the helpIcon and the requiredIcon and 
      // then for the aria-describedby on the input.
      labelElementId = this.element.attr("id");
      if (labelElementId == null)
      {
        this.element.uniqueId(); 
      }

    },
    /**
     * Called if NOT a custom element.
     * move oj-label* classes from label element onto the root dom element
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _moveLabelStyleClassesToRootDom: function ()
    {
      var arrayOfClasses;
      var classes = this.element.attr("class");
      var className;
      var numClasses;

      if (classes)
      {
        arrayOfClasses = classes.split(/\s+/);
        numClasses = arrayOfClasses.length;

        for (var i = 0; i < numClasses; i++)
        {
          className = arrayOfClasses[i];
          // if class name has -label- in it, then move it
          // (e.g., oj-label, oj-label-inline, oj-md-label-nowrap,
          // oj-md-labels-inline)
          if (className.indexOf("-label") > 0)
          {
            this.uiLabel.addClass(className);
            this.element.removeClass(className);
          }
        }
      }
    },
    /**
     * create and return the span with an id that we'll use to put around the help
     * icon or required icon. The order of the dom of oj-label-group's children should always be:
     * <helpIcon span/><requiredIcon span/><label element>
     * @param {string} id the id of the span
     * @param {boolean} isHelp if true, prepend icon span to oj-label-group, else
     * add before the label element. 
     * This is needed because we always want the help dom to come before the required dom and 
     * we always want the label element to come last.
     * @returns {Element}
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _createIconSpan: function (id, isHelp)
    {
      var ojLabelGroupDom;
      var span = document.createElement("span");
      span.setAttribute("id", id);

      if (isHelp)
      {
        ojLabelGroupDom = this.uiLabel.find(".oj-label-group");
        ojLabelGroupDom.prepend(span); // @HTMLUpdateOK
      }
      else
      {
        this.element.before(span); // @HTMLUpdateOK
      }
      return span;
    },
    /**
     * For ojLabel component only, not oj-label.
     * return the dom node for the root dom element
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _createRootDomElement: function ()
    {
      var inputLabelClass;
      var rootAttributes = this.options.rootAttributes;
      var rootDomNode;
      var rootDomNodeClasses = "oj-label oj-component";
      
      if (rootAttributes)
      {
        inputLabelClass = rootAttributes['class'];
      }
      if (inputLabelClass)
      {
        rootDomNodeClasses = rootDomNodeClasses + " " + inputLabelClass;
      }

      //rootDomNode =
      //  $("<div class='oj-label oj-component'><div class='oj-label-group'></div></div>",
      //     this.document[0]);
      rootDomNode = document.createElement("div");
      rootDomNode.className = rootDomNodeClasses;
      rootDomNode.appendChild(this._createOjLabelGroupDom()); // @HTMLUpdateOk

      return rootDomNode;
    },
   /**
     * return the dom node for the oj-label-group node
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _createOjLabelGroupDom: function ()
    {
      var labelGroupNode;

      labelGroupNode = document.createElement("div");
      labelGroupNode.className = "oj-label-group";

      return labelGroupNode;
    },
    /**
     * return the dom node for the span with oj-label-required-icon
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _createRequiredIconDomElement: function ()
    {
      var requiredTooltip = this.getTranslatedString(this._BUNDLE_KEY._TOOLTIP_REQUIRED);
      var requiredDom = document.createElement("span");

      requiredDom.className = "oj-label-required-icon oj-component-icon";
      requiredDom.setAttribute("role", "img");
      requiredDom.setAttribute("title", requiredTooltip);
      // title isn't being read by the screen reader. this is only needed for radioset/checkboxset.
      requiredDom.setAttribute("aria-label", requiredTooltip);
      return requiredDom;
    },
    /**
     * return the dom node for the help icon anchor.
     * if (_needsHelpIcon) , show help icon
     * if (helpSource), add href
     * if (helpDef), add 'aria-label'=helpDef on help icon.
     * @private
     * @memberof oj.ojLabel
     * @instance
     */
    _createHelpIconAnchorDomElement: function (helpDef, source)
    {
      var helpIconAnchor;
      // construct the help html
      // if source (external url) or helpDef, then render a clickable help icon

      // From our Accessibility expert - You must not put role of img on a link.
      // This will make it so it is not a link any more to AT.
      // It is ok to leave it off the the <a> tag and do the following.
      //helpIconAnchor =
      //  $( "<a tabindex='0' target='_blank' class='oj-label-help-icon-anchor oj-label-help-icon oj-component-icon oj-clickable-icon-nocontext'></a>",
      //  this.document[0] );

      // The above is not reading anything when it has focus if it doesn't have an href. So if
      // it doesn't have an href, it needs some kind of role on it.

      helpIconAnchor = document.createElement("a");
      helpIconAnchor.setAttribute("tabindex", "0");
      helpIconAnchor.setAttribute("target", "_blank");
      helpIconAnchor.className =
      "oj-label-help-icon-anchor oj-label-help-icon oj-component-icon oj-clickable-icon-nocontext";
      if (source)
      {
        try
        {
          oj.DomUtils.validateURL(source);
          helpIconAnchor.setAttribute("href", source);
        }
        catch (e)
        {
          throw new Error(e + ". The source option (" + source + ") is invalid.");
        }

      }
      else
      {
        // if there is no href, then we need a role that the screen reader/voiceover will read.
        helpIconAnchor.setAttribute("role", "img");
      }

      if (helpDef)
        helpIconAnchor.setAttribute("aria-label", helpDef);
      else
        helpIconAnchor.setAttribute("aria-label", this.getTranslatedString(this._BUNDLE_KEY._TOOLTIP_HELP));

      return helpIconAnchor;
    },
    /**
     * To accomodate keyboard and touch users,
     * show a popup on hover, on tabbing in or touch press that shows the
     * help definition text on the help icon.
     *
     * press is recognized when the
     * pointer is down for x ms without any movement. In other words, you press with your finger and
     * don't let up and then after x ms the Help Def window shows up. You can then let go of your finger
     * the help def window stays up.
     * On Android, when you press and hold you see touchstart. When you finally let up, you see touchend
     * On ios, when you press and hold you see touchstart. When you finally let up, sometimes you
     * see touchend only. Other times, you see touchend mousedown mouseup click all
     * consecutively.
     * On ios, a quick tap shows touchstart touchend mousedown mouseup click (I only register those events)
     * all right after another.
     * @private
     * @memberof oj.ojLabel
     */
    _attachHelpDefToIconAnchor: function ()
    {
      var $helpDefPopupDiv;
      var $helpIcon;
      var position;
      var self = this;

      $helpIcon = this.widget().find(".oj-label-help-icon-anchor");

      // before we do any of this work, make sure there is a help icon
      if ($helpIcon.length === 0)
        return;

      // Create the popup div where we will display the help def text, add a unique id onto it and save
      // that id so we can use it to popup.

      // 1. Build display:none div with help definition text. This will be our popup.
      // 2. Register click/touch event on label which will call a callback to open the popup on
      // PRESS

      // create popup's div
      $helpDefPopupDiv = this._createHelpDefPopupDiv();

      if (this._bTouchSupported)
      {
        // this is code to be extra careful: Check if the _eatClickOnHelpIconListener exists.
        // If it does exist, call 'off'. We don't want this click listener
        // that eats clicks lying around.
        if (this._eatClickOnHelpIconListener)
        {
          this.widget().off(this._touchEatClickNamespace);
        }
        // The pressHold gesture also fires a click event on iphone on touchend.  Prevent that here.
        // This event is added to the widget on click in the function _handleOpenPopupForHelpDef
        this._eatClickOnHelpIconListener = function (event)
        {
          // changing colors is a good way to debug if the handler is being called. this changes
          // the label color.
//        if (this.style.color === "aqua")
//          this.style.color = "yellow";
//        else
//         this.style.color = "aqua";
          return false;
        };

        // The pressHold gesture also fires a contextmenu event on android.  Prevent that here.
        this._eatContextMenuOnHelpIconListener = function (event)
        {
          return false;
        };

        $helpIcon.on("contextmenu" + this._touchEatContextMenuNamespace,
        this._eatContextMenuOnHelpIconListener);
      }

      // For touch device, press with finger on helpIcon to show the help def in a popup.
      // If there is no help source, you can also tap with finger on the helpIcon to show the help
      // def in a popup.
      // For keyboard users, tab in to helpIcon to show the help def in a popup.
      // For mouse users, hovering on helpIcon shows the help def in a popup.
      // ------------------------------------------------------------------------------------


      // ENTERING CALLBACK TO OPEN THE POPUP IF NEEDED
      // (focusin from tab, not mouse, OR press from touch)

      if (!this._openPopupForHelpDefCallbackListener)
      {
        this._openPopupForHelpDefCallbackListener = function (event)
        {
          self._handleOpenHelpDefPopup(event, $helpDefPopupDiv, $helpIcon);
        };
      }
      // END CALLBACK TO OPEN POPUP
      //


      // CALLBACK TO CLOSE POPUP
      if (!this._closePopupForHelpDefCallbackListener)
      {
        this._closePopupForHelpDefCallbackListener = function (event)
        {
          self._handleCloseHelpDefPopup();
        };
      }
      // END CALLBACK TO CLOSE POPUP


      // Add event handlers to open the help definition popup
      this._addShowHelpDefinitionEventHandlers($helpIcon);

      position =
      {
        'my': 'start bottom',
        'at': 'end top',
        'collision': 'flipcenter',
        'of': $helpIcon
      };
      $helpDefPopupDiv.ojPopup(
      {"position": position,
        "modality": "modeless",
        "animation": {"open": null, "close": null}});
    },
    /**
     * Create the div that will be used as the popup for the help definition.
     * @private
     * @memberof oj.ojLabel
     */
    _createHelpDefPopupDiv: function ()
    {
      var bodyDom;
      var contentDiv;
      var $contentDiv;
      var helpDef = this.options.help['definition'];
      var helpDefPopupDiv;
      var $helpDefPopupDiv;
      var helpDefText;


      if (helpDef)
        helpDefText = helpDef;
      else
        helpDefText = this.getTranslatedString(this._BUNDLE_KEY._TOOLTIP_HELP);
      
      if (!this._helpDefPopupDivId)
      {
        // create a root node to bind to the popup
        helpDefPopupDiv = document.createElement("div");
        helpDefPopupDiv.className = "oj-help-popup";
        helpDefPopupDiv.style.display = "none";
        $helpDefPopupDiv = $(helpDefPopupDiv);
        $helpDefPopupDiv.uniqueId();
        this._helpDefPopupDivId = $helpDefPopupDiv.prop("id");

        // create a content node
        contentDiv = document.createElement("div");
        contentDiv.className = "oj-help-popup-container";
        helpDefPopupDiv.appendChild(contentDiv); // @HTMLUpdateOk created contentDiv ourselves
        $contentDiv = $(contentDiv);

        $contentDiv.text(helpDefText);
        bodyDom = document.getElementsByTagName("body")[0];
        bodyDom.appendChild(helpDefPopupDiv); // @HTMLUpdateOK
      }
      else
      {
        // Find the div with the id, and then update the text of it.
        $helpDefPopupDiv = $(document.getElementById(this._helpDefPopupDivId));
        if ($helpDefPopupDiv)
        {
          $contentDiv = $helpDefPopupDiv.find(".oj-help-popup-container").first();
          $contentDiv.text(helpDefText);
        }
      } 
      return $helpDefPopupDiv;

    },
    /**
     * Add the event listeners to show the helpDefinition text in a popup
     * @param {jQuery} $helpIcon
     * @returns {undefined}
     * @instance
     * @memberof oj.ojLabel
     * @private
     */
    _addShowHelpDefinitionEventHandlers: function ($helpIcon)
    {
      var hammerOptions;
   
      // Open the popup on focusin and mousenter.
      // *I have logic in the listener to ignore these when these trigger as a result of 
      // the user touching the screen.
      $helpIcon.on("focusin" + this._helpDefPopupNamespace +
                   " mouseenter" + this._helpDefPopupNamespace, 
                   this._openPopupForHelpDefCallbackListener);
      $helpIcon.on("mouseleave" + this._helpDefPopupNamespace,
                   this._closePopupForHelpDefCallbackListener);
      if (this._bTouchSupported)
      {
        // And if touch is supported, the user can also open the popup on press or,
        //  if no help source, on tap.
        if (this.options.help['source'])
        {
          hammerOptions = {
            "recognizers": [
              [Hammer.Press, {time: oj.DomUtils.PRESS_HOLD_THRESHOLD}]
            ]};
          $helpIcon.ojHammer(hammerOptions);
          // JET components are encouraged to use JQUI's _on() method, giving all the conveniences
          // of the _on method, like automatic cleanup.
          this._on($helpIcon,
            {"press": this._openPopupForHelpDefCallbackListener});
        }
        else
        {
          hammerOptions = {
            "recognizers": [
              [Hammer.Tap],
              [Hammer.Press, {time: oj.DomUtils.PRESS_HOLD_THRESHOLD}]

            ]};
          $helpIcon.ojHammer(hammerOptions);
          this._on($helpIcon,
            {"press": this._openPopupForHelpDefCallbackListener,
             "tap": this._openPopupForHelpDefCallbackListener});

        }
        
      }
    },
    /**
     * Handle open popup for help definition.
     * @instance
     * @memberof oj.ojLabel
     * @private
     */
    _handleOpenHelpDefPopup: function (event, helpDefPopupDiv, helpIcon)
    {
      var isOpen = helpDefPopupDiv.ojPopup("isOpen");   

      if (isOpen)
        return;

      // touch supported does not mean only touch. It could be a touch-enabled laptop like Windows10
      if (this._bTouchSupported)
      {
        // For a press, we want to show the popup with the help def, 
        // but we do not want to navigate to the source url. So we eat the click.
        if (event.type === "press")
        {
          var widget = this.widget();
          widget.on("click" + this._touchEatClickNamespace, this._eatClickOnHelpIconListener);
          var self = this;
          helpDefPopupDiv.on("ojclose",
          function (event, ui)
          {
            widget.off(self._touchEatClickNamespace);
          });
        }
        else
        {
          helpDefPopupDiv.off("ojclose");
        }

        // Open the popup if I get a 'press' event, a 'tap' event, a 'focusin' event if
        // it wasn't a touch, and a 'mouseenter' event if it wasn't a touch.
        // I look for a recent touchstart event and use this to filter out 
        // the focusin and mouseevent events. I use touchstart and not touchend because while
        // pressing I get the touchstart, but I don't get the touchend until the finger lets up.
        if (event.type === "press" ||  
            event.type === "tap" || 
            (!oj.DomUtils.recentTouchStart() &&
            (event.type ==="focusin" || event.type ==="mouseenter")))
        {
          helpDefPopupDiv.ojPopup("open", helpIcon);
        }
      } // end touch code
      else
      {
        // non-touch devices. focusin/mouseenter are the only ways to open the popup.
        helpDefPopupDiv.ojPopup("open", helpIcon);
      }

    },
    /**
     * Close helpDef popup. This is called from _NotifyDetached and _NotifyHidden and
     * as a callback for this._closePopupForHelpDefCallbackListener.
     * @private
     * @memberof oj.ojLabel
     */
    _handleCloseHelpDefPopup: function ()
    {
      var $helpDefPopupDiv;
      
      if (this._helpDefPopupDivId != null)
      {
        $helpDefPopupDiv = $(document.getElementById(this._helpDefPopupDivId));
        $helpDefPopupDiv.ojPopup("close");
      }
    },
    /**
     * Remove the event listeners for opening a popup on the help def icon and for eating
     * the clicks on the 'press' event. Called from destroy and when we remove the help icon.
     * @private
     * @memberof oj.ojLabel
     */
    _removeHelpDefIconEventListeners: function (helpIcon)
    {
      if (this._bTouchSupported)
      {
        this.widget().off(this._touchEatClickNamespace);
        helpIcon.off(this._touchEatContextMenuNamespace);
        this._eatClickOnHelpIconListener = null;
        this._eatContextMenuOnHelpIconListener = null;
        // helpIcon is same element on which we originally called ojHammer()
        // the listeners are automatically removed since we used jqueryui's _on
        helpIcon.ojHammer("destroy");
      }
      helpIcon.off(this._helpDefPopupNamespace);
      this._openPopupForHelpDefCallbackListener = null;
      this._closePopupForHelpDefCallbackListener = null;
    },
    /**
     * removes the help def popup dom and variables
     * @returns {undefined}
     * @private
     * @memberof oj.ojLabel
     */
    _removeHelpDefPopup: function ()
    {
      var $helpDefPopupDiv;

      if (this._helpDefPopupDivId != null)
      {
        $helpDefPopupDiv = $(document.getElementById(this._helpDefPopupDivId));
        if ($helpDefPopupDiv)
        {
          $helpDefPopupDiv.ojPopup("destroy");
          $helpDefPopupDiv.remove();
        }
        this._helpDefPopupDivId = null;
      }
    },
    /**
     * @private
     * @memberof oj.ojLabel
     * @returns {boolean}
     */
    _needsHelpIcon: function ()
    {  
      var options = this.options;
      var helpDef;
      var helpSource = options.help['source'];
      var needsIcon;
      
      // "", or null, or undefined all mean help.source is not specified.
      needsIcon = (helpSource !== "" && helpSource != null);
      if (!needsIcon)
      {
        // Now check helpDef: "", or null, or undefined all mean help.definition is not specified.
        helpDef = options.help['definition'];
        needsIcon = (helpDef !== "" && helpDef != null);
      }
      
      return needsIcon;
    },
    /**
     * refresh the help dom --
     * find the help root dom node and remove it if it is there
     * and add back the help html. Helpful if a help option changed.
     * @private
     * @memberof oj.ojLabel
     */
    _refreshHelp: function ()
    {
      var helpSpanId = this.helpSpanId;
      var helpSpan;
      var $helpIcon;
      var needsHelpIcon;
      
      // remove the help info if it is there.
      $helpIcon = this.uiLabel.find(".oj-label-help-icon");

      if ($helpIcon.length === 1)
      {     
        // remove things we added in _attachHelpDefToIconAnchor
        this._removeHelpDefIconEventListeners($helpIcon);
        this._removeHelpDefPopup();
        $helpIcon.remove();
      }
      helpSpan = document.getElementById(helpSpanId);
      needsHelpIcon = this._needsHelpIcon();
      
      // ok, we removed the helpIcon (but not the span) at the start of this method, 
      // so we need to add it back if we needHelpIcon
      if(needsHelpIcon)
      {
        if (helpSpan == null)
        {
          helpSpan = this._createIconSpan(helpSpanId, true);  
        }
        this._createHelp(helpSpan);
        if(this._isCustomElement)
        {
          this._addHelpSpanIdOnTarget(helpSpanId);
        }
      }
      else if (helpSpan !== null)
      {
        helpSpan.parentNode.removeChild(helpSpan);
        if(this._isCustomElement)
        {
          this._removeHelpSpanIdOnTarget(helpSpanId); 
        }       
      }

    },
    /**
     * refresh the required dom --
     * if required is true, then add the required dom if it isn't already there
     * if required is false, remove the required dom if it is there.
     * Helpful if the required option changed.
     * @private
     * @memberof oj.ojLabel
     */
    _refreshRequired: function ()
    {
      var $requiredDom;
      var requiredSpanId = this.requiredSpanId;
      var requiredSpan;
      var requiredTooltip;


      requiredSpan = document.getElementById(requiredSpanId);

      if (this.options.showRequired)
      {
        // add required if it wasn't already there
        if (!requiredSpan)
        {
          this. _createRequiredIconSpanDom();

          if (this._isCustomElement)
            this._addRequiredDescribedByOnCustomFormElement();
        }
        else
        {
          // required is there, so we need to refresh the translated value in case it changed.
          requiredTooltip = this.getTranslatedString(this._BUNDLE_KEY._TOOLTIP_REQUIRED);
          $requiredDom = this.uiLabel.find(".oj-label-required-icon");
          $requiredDom.attr("title", requiredTooltip);
        }
      }
      else
      {
        // not required, so remove it
        requiredSpan = document.getElementById(requiredSpanId);
        if (requiredSpan !== null)
        {
          requiredSpan.parentNode.removeChild(requiredSpan);
        }

        if (this._isCustomElement)
          this._removeRequiredDescribedByOnCustomFormElement();
      } 
    },
    /**
     * If this is a custom element (oj-label), it gets the 'for' target element, 
     * and if it is found, it adds the 'for' onto the label element (as a sub-id or directly, 
     * depending upon what the 'for' target element is.
     * @private
     * @memberof oj.ojLabel
     */
    _refreshFor: function ()
    {
      var forTargetElement;
      if (this._isCustomElement)
      {
        forTargetElement = this._getForTargetElement();
        if (forTargetElement)
          this._addForOnLabel(forTargetElement);
      }
    },
    /**
     * refresh label-id which is only for oj-label custom element.
     * Set the label element's id to label-id. If label-id is not set,
     * then use the <oj-label>'s id attribute to set a sub-id on the label element.
     * @private
     * @memberof oj.ojLabel
     */
    _refreshLabelId: function ()
    {
      var labelIdOption;
      var customElementId;
      
      if (this._isCustomElement)
      {
        labelIdOption = this.options.labelId;
        if (labelIdOption)
            this.element.attr("id", labelIdOption);
        else
        {
          // for custom elements, this.uiLabel is <oj-label>.
          customElementId = this.uiLabel.attr("id");
          if (customElementId)
          {
            // create sub-id on the label element
            this.element.attr("id", customElementId + "|label");
          }
        }
      }
    },
   
    /**
     * Note that _setOption does not get called during create. it only gets called
     * when the component has already been created.
     * @override
     * @protected
     * @memberof oj.ojLabel
     * @instance
     */
    _setOption: function (key, value)
    {
      this._superApply(arguments);

      switch(key)
      {
        case "showRequired":
          this._refreshRequired();
          break;
        case "help":
          this._refreshHelp();
          break;
        case "for":
          if (this._isCustomElement)
          {
            // refreshes for, which is only for oj-label custom element, not ojLabel.
            // targetElement is the element with id === oj-label's for option
            this._refreshFor();
          }
          break;
        case "labelId":
          if (this._isCustomElement)
          {
            // refreshes label-id, which is only for oj-label custom element, not ojLabel.
            this._refreshLabelId();
          }
          break;
        default:
          break;
      }
    },
    // @inheritdoc
    getNodeBySubId: function (locator)
    {
      var node;
      var subId;

      node = this._super(locator);
      if (!node)
      {
        subId = locator['subId'];
        if (subId === "oj-label-help-icon")
        {
          node = this.widget().find(".oj-label-help-icon")[0];
        }
      }
      // Non-null locators have to be handled by the component subclasses
      return node || null;
    },
    // @inheritdoc
    getSubIdByNode: function (node)
    {
      var subId = null;

      if (node != null)
      {
        if (node === this.widget().find(".oj-label-help-icon")[0])
        {
          subId = {'subId': "oj-label-help-icon"};
        }
      }

      return subId || this._superApply(arguments);
    },
    /**
     *
     * @override
     * @protected
     * @memberof oj.ojLabel
     * @instance
     */
    _destroy: function ()
    {
      // remove things we added in _attachHelpDefToIconAnchor
      var helpIcon = this.uiLabel.find(".oj-label-help-icon");
      this._removeHelpDefIconEventListeners(helpIcon);
      this._removeHelpDefPopup();
      this.helpSpanId = null;
      this.requiredSpanId = null;
      this._isCustomElement = null;
      // DomUtils.unwrap() will avoid unwrapping if the node is being destroyed by Knockout
      oj.DomUtils.unwrap(this.element, this.uiLabel);

      return this._super();
    }

    /**** end internal widget functions ****/

    /**
     * Removes the label functionality completely.
     * This will return the element back to its pre-init state.
     *
     * <p>This method does not accept any arguments.
     *
     * @method
     * @name oj.ojLabel#destroy
     * @memberof oj.ojLabel
     * @instance
     * @ignore
     *
     * @example <caption>Invoke the <code class="prettyprint">destroy</code> method:</caption>
     * $( ".selector" ).ojLabel( "destroy" );
     */
   /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan=2>Help Icon</td>
     *       <td><kbd>Tap and Hold</kbd></td>
     *       <td>Show the help definition in a popup</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Tap</kbd></td>
     *       <td>If no help source, show the help definition in a popup. 
     *       If help source, navigate to the url.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * 
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojLabel
     */
   /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan=2>Help Icon</td>
     *       <td><kbd>Enter</kbd></td>
     *       <td>If there is an url associated with help icon, navigate to the url.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Tab In</kbd></td>
     *       <td>Show the help definition in a popup.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojLabel
     */
    
    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     * 
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-focus-highlight</td>
     *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
     *     </tr>
     *     <tr>
     *       <td>oj-label-accesskey</td>
     *       <td>Use this in a span around a single text character in the oj-label's text. 
     *       It styles the character in a way that indicates to the 
     *       user that this character is the accesskey. Use this in conjunction with
     *       the HTML accesskey attribute on the oj-label element.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojLabel
     */
  });

//////////////////     SUB-IDS     //////////////////

  /**
   * <p>Sub-ID for the label's help icon.</p>
   *
   * @ojsubid oj-label-help-icon
   * @memberof oj.ojLabel
   *
   * @example <caption>Get the node for the help icon:</caption>
   * var node = myComponent.getNodeBySubId({'subId': 'oj-label-help-icon'});
   */

}());
(function() {
var ojLabelMeta = {
  "properties": {
    "help": {
      "type": "Object",
      "properties": {
        "definition": {
          "type": "string"
        },
        "source": {
          "type": "string"
        }
      }
    },
    "showRequired": {
      "type": "boolean"
    },
    "labelId": {
      "type": "string"
    },
    "for": {
      "type": "string"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "tooltipHelp": {
          "type": "string",
          "value": "Help"
        },
        "tooltipRequired": {
          "type": "string",
          "value": "Required"
        }
      }
    }
    
  },
  "extension": {
    _INNER_ELEM: 'label',
    _WIDGET_NAME: "ojLabel",
    _GLOBAL_TRANSFER_ATTRS: ["accesskey"]
  }
};
oj.CustomElementBridge.registerMetadata('oj-label', 'baseComponent', ojLabelMeta);
oj.CustomElementBridge.register('oj-label', {'metadata': oj.CustomElementBridge.getMetadata('oj-label')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojLegend
 * @augments oj.dvtBaseComponent
 * @since 0.7
 * @ojstatus preview
 * @ojshortdesc Displays an interactive description of symbols, colors, etc., used in graphical information representations.
 * @ojrole application
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="legendOverview-section">
 *   JET Legend
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#legendOverview-section"></a>
 * </h3>
 *
 * <p>A themable, WAI-ARIA compliant legend for JET.</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-legend
    orientation='vertical'
    sections='[{"items": [
                {"text": "Database"},
                {"text": "Middleware"},
                {"text": "Application"}
              ]}]'
   >
 * &lt;/oj-legend>
 * </code>
 * </pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojLegend', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     *  Whether drilling is enabled on all legend items. Drillable objects will show a pointer cursor on hover and fire <code class="prettyprint">ojDrill</code> event on click. To enable or disable drilling on individual legend item, use the drilling attribute in each legend item. 
     * @expose
     * @name drilling
     * @memberof oj.ojLegend
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "off"
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">drilling</code> attribute specified:</caption>
     * &lt;oj-legend drilling="on">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">drilling</code> property after initialization:</caption>
     * // getter 
     * var drillingValue = myLegend.drilling;
     * 
     * // setter 
     * myLegend.drilling = 'on';
     */
    drilling: 'off',

    /**
     * Defines the horizontal alignment of the legend contents.
     * @expose
     * @name halign
     * @memberof oj.ojLegend
     * @instance
     * @type {string}
     * @ojvalue {string} "center"
     * @ojvalue {string} "end"
     * @ojvalue {string} "start"
     * @default "start"
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">halign</code> attribute specified:</caption>  
     * &lt;oj-legend halign="center">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">halign</code> property after initialization:</caption>
     * // getter 
     * var halignValue = myLegend.halign;
     *
     * // setter 
     * myLegend.halign = "center";
     * 
     */
    halign: 'start',

    /**
     * An array of categories that will be hidden.
     * @expose
     * @name hiddenCategories
     * @memberof oj.ojLegend
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">hidden-categories</code> attribute specified:</caption>
     * &lt;oj-legend hidden-categories='["Apples", "Bananas"]'>&lt;/oj-legend>
     *
     * @example <caption>Get or set the <code class="prettyprint">hiddenCategories</code> property after initialization:</caption>
     * // Get one
     * var value = myLegend.hiddenCategories[0];
     *
     * // Get all
     * var hiddenCategoriesValue = myLegend.hiddenCategories;
     *
     * // setter
     * myLegend.hiddenCategories = ["Apples", "Bananas"];
     */
    hiddenCategories: [],

    /**
     * Defines whether the legend can be used to initiate hide and show behavior on referenced data items.
     * @expose
     * @name hideAndShowBehavior
     * @memberof oj.ojLegend
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "off"
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">hide-and-show-behavior</code> attribute specified:</caption>
     * &lt;oj-legend hide-and-show-behavior="on">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">hideAndShowBehavior</code> property after initialization:</caption>
     * // getter 
     * var hideAndShowValue = myLegend.hideAndShowBehavior;
     *
     * // setter 
     * myLegend.hideAndShowBehavior = 'on';
     */
    hideAndShowBehavior: 'off',

    /**
     * An array of categories that will be highlighted.
     * @expose
     * @name highlightedCategories
     * @memberof oj.ojLegend
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     * 
     * @example <caption>Initialize the Legend with the <code class="prettyprint">highlighted-categories</code> attribute specified:</caption>
     * &lt;oj-legend highlighted-categories='["Bananas", "Apples"]'>&lt;/oj-legend>
     *
     * @example <caption>Get or set the <code class="prettyprint">highlightedCategories</code> property after initialization:</caption>
     * // Get one
     * var value = myLegend.highlightedCategories[0];
     *
     * // getter
     * var highlightedCategoriesValue = myLegend.highlightedCategories;
     *
     * // setter
     * myLegend.highlightedCategories = ["Bananas", "Apples"];
     */
    highlightedCategories: [],

    /**
     * Defines the behavior applied when hovering over a legend item.
     * @expose
     * @name hoverBehavior
     * @memberof oj.ojLegend
     * @instance
     * @type {string}
     * @ojvalue {string} "dim"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">hover-behavior</code> attribute specified:</caption>
     * &lt;oj-legend hover-behavior="dim">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">hoverBehavior</code> property after initialization:</caption>
     * // getter 
     * var hoverBehaviorValue = myLegend.hoverBehavior;
     * 
     * // setter 
     * myLegend.hoverBehavior = 'dim';
     */
    hoverBehavior: 'none',

    /**
     * Specifies initial hover delay in ms for highlighting items in legend.
     * @expose
     * @name hoverBehaviorDelay
     * @memberof oj.ojLegend
     * @instance
     * @type {number}
     * @ojunits milliseconds
     * @ojmin 0
     * @default 200
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">hover-behavior-delay</code> attribute specified:</caption>
     * &lt;oj-legend hover-behavior-delay="150">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">hoverBehaviorDelay</code> property after initialization:</caption>
     * // getter 
     * var delayValue = myLegend.hoverBehaviorDelay;
     *  
     * // setter 
     * myLegend.hoverBehaviorDelay = 150;
     */
    hoverBehaviorDelay: 200,

    /**
     * Defines the orientation of the legend, which determines the direction in which the legend items are laid out.
     * @expose
     * @name orientation
     * @memberof oj.ojLegend
     * @instance
     * @type {string}
     * @ojvalue {string} "horizontal"
     * @ojvalue {string} "vertical"
     * @default "vertical"
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">orientation</code> attribute specified:</caption>
     * &lt;oj-legend orientation="horizontal">&lt;/oj-legend>
     *
     * @example <caption>Get or set the <code class="prettyprint">orientation</code> property after initialization:</caption>
     * // getter 
     * var orientationValue = myLegend.orientation;
     *
     * // setter 
     * myLegend.orientation = "horizontal";
     */
    orientation: 'vertical',

    /**
     * Defines whether scrolling is enabled for the legend.
     * @expose
     * @name scrolling
     * @memberof oj.ojLegend
     * @instance
     * @type {string}
     * @ojvalue {string} "off"
     * @ojvalue {string} "asNeeded"
     * @default "asNeeded"
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">scrolling</code> attribute specified:</caption>
     * &lt;oj-legend scrolling="off">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">scrolling</code> property after initialization:</caption>
     * // getter 
     * var scrollingValue = myLegend.scrolling;
     *
     * // setter 
     * myLegend.scrolling = 'off';
     *
     */
    scrolling: 'asNeeded',

    /**
     * An array of objects with the following properties defining the legend sections.
     * @expose
     * @name sections
     * @memberof oj.ojLegend
     * @instance
     * @type {Array.<Object>|null}
     * @default null
     * 
     * @example <caption>Initialize the legend with the 
     * <code class="prettyprint">sections</code> attribute specified:</caption>
     * &lt;oj-legend sections='[{"title": "Brand", "expanded": "on", "collapsible": "on", 
     *                        items: [{"color": "red", "text": "Coke", "id": "Coke"}, 
     *                                {"color": "blue", "text": "Pepsi", "id": "Pepsi"}, 
     *                                {"color": "yellow", "text": "Snapple", "id": "Snapple"},
     *                                {"color": "brown", "text": "Nestle", "id": "Nestle"}]}]'>
     * &lt;/oj-legend>
     * 
     * &lt;oj-legend sections='[[sectionsPromise]]'>&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">sections</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myLegend.sections[0];
     *
     * // Get all (The items getter always returns a Promise so there is no "get one" syntax)
     * var values = myLegend.sections;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myLegend.sections=[{title: "Brand", expanded: "on", collapsible: "on", 
     *                     items: [{color: "red", text: "Coke", id: "Coke"}, 
     *                             {color: "blue", text: "Pepsi", id: "Pepsi"}, 
     *                             {color: "yellow", text: "Snapple", id: "Snapple"},
     *                             {color: "brown", text: "Nestle", id: "Nestle"}]}];
     */
    sections: null,

    /**
     * The height of the legend symbol (line or marker) in pixels. If the value is 0, it will take the same value as symbolWidth. If both symbolWidth and symbolHeight are 0, then it will use a default value that may vary based on theme.
     * @expose
     * @name symbolHeight
     * @memberof oj.ojLegend
     * @instance
     * @type {number}
     * @ojunits pixels
     * @default 0
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">symbol-height</code> attribute specified:</caption>
     * &lt;oj-legend symbol-height="20">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">symbolHeight</code> property after initialization:</caption>
     * // getter 
     * var symbolHeightValue = myLegend.symbolHeight;
     * 
     * // setter 
     * myLegend.symbolHeight = 20;
     */
    symbolHeight: 0,

    /**
     * The width of the legend symbol (line or marker) in pixels. If the value is 0, it will take the same value as symbolWidth. If both symbolWidth and symbolHeight are 0, then it will use a default value that may vary based on theme.
     * @expose
     * @name symbolWidth
     * @memberof oj.ojLegend
     * @instance
     * @type {number}
     * @ojunits pixels
     * @default 0
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">symbol-width</code> attribute specified:</caption>
     * &lt;oj-legend symbol-width="15">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">symbolWidth</code> property after initialization:</caption>
     * // getter 
     * var symbolWidthValue = myLegend.symbolWidth;
     * 
     * // setter 
     * myLegend.symbolWidth = 15;
     */
    symbolWidth: 0,

    /**
     * The CSS style object defining the style of the legend item text. The default value comes from the CSS and varies based on theme.
     * @expose
     * @name textStyle
     * @memberof oj.ojLegend
     * @instance
     * @type {Object}
     * 
     * @example <caption>Initialize the Legend with the <code class="prettyprint">text-style</code> attribute specified:</caption>
     * &lt;oj-legend text-style='{"fontSize":"12px"}'>&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">textStyle</code> property after initialization:</caption>
     * // getter 
     * var textStyleValue = myLegend.textStyle;
     * 
     * // setter 
     * myLegend.textStyle = {"fontSize" : "12px"};
     */
    textStyle: undefined,

    /**
     * Defines the vertical alignment of the legend contents.
     * @expose
     * @name valign
     * @memberof oj.ojLegend
     * @instance
     * @type {string}
     * @ojvalue {string} "middle"
     * @ojvalue {string} "bottom"
     * @ojvalue {string} "top"
     * @default "top"
     *
     * @example <caption>Initialize the Legend with the <code class="prettyprint">valign</code> attribute specified:</caption>
     * &lt;oj-legend valign="middle">&lt;/oj-legend>
     * 
     * @example <caption>Get or set the <code class="prettyprint">valign</code> property after initialization:</caption>
     * // getter 
     * var valignValue = myLegend.valign;
     *  
     * // setter 
     * myLegend.valign = "middle";
     */
    valign: 'top',

    // EVENTS

    /**
     * Triggered during a drill gesture (single click on the legend item).
     * @property {string} id the id of the drilled object
     *
     * @expose
     * @event
     * @memberof oj.ojLegend
     * @instance
     */
    drill: null
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    return dvt.Legend.newInstance(context, callback, callbackObj);
  },


  //** @inheritdoc */
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    if(subId == 'oj-legend-item') {
      // section[sectionIndex0][sectionIndex1]...[sectionIndexN]:item[itemIndex]
      subId = 'section' + this._GetStringFromIndexPath(locator['sectionIndexPath']);
      subId += ':item[' + locator['itemIndex'] + ']';
    }
    else if(subId == 'oj-legend-tooltip') {
        subId = 'tooltip';
    }
    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId.indexOf(':item') > 0) {
      // section[sectionIndex0][sectionIndex1]...[sectionIndexN]:item[itemIndex]
      var itemStartIndex = subId.indexOf(':item');
      var sectionSubstr = subId.substring(0, itemStartIndex);
      var itemSubstr = subId.substring(itemStartIndex);

      locator['subId'] = 'oj-legend-item';
      locator['sectionIndexPath'] = this._GetIndexPath(sectionSubstr);
      locator['itemIndex'] = this._GetFirstIndex(itemSubstr);
    }
    else if(subId == 'tooltip') {
        locator['subId'] = 'oj-legend-tooltip';
    }
    return locator;
  },

  //** @inheritdoc */
	_GetComponentStyleClasses: function() {
		var styleClasses = this._super();
		styleClasses.push('oj-legend');
		return styleClasses;
	},

  //** @inheritdoc */
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-legend'] = {'path' : 'textStyle', 'property' : 'TEXT'};
    styleClasses['oj-legend-title'] = {'path' : 'titleStyle', 'property' : 'TEXT'};
    styleClasses['oj-legend-section-title'] = {'path' : '_sectionTitleStyle', 'property' : 'TEXT'};
    return styleClasses;
  },

  //** @inheritdoc */
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtUtilBundle.LEGEND'] = translations['componentName'];
    return ret;
  },

  //** @inheritdoc */
  _GetEventTypes : function() {
    return ['drill'];
  },

  //** @inheritdoc */
  _HandleEvent : function(event) {
    var type = event['type'];
    if (type === 'drill') {
      this._trigger('drill', null, {'id': event['id']});
    }
    else {
      this._super(event);
    }
  },

  //** @inheritdoc */
  _LoadResources: function() {
    // Ensure the resources object exists
    if (this.options['_resources'] == null)
      this.options['_resources'] = {};

    var resources = this.options['_resources'];

    // Add images
    resources['closedEnabled'] = 'oj-legend-section-close-icon';
    resources['closedOver'] = 'oj-legend-section-close-icon oj-hover';
    resources['closedDown'] = 'oj-legend-section-close-icon oj-active';
    resources['openEnabled'] = 'oj-legend-section-open-icon';
    resources['openOver'] = 'oj-legend-section-open-icon oj-hover';
    resources['openDown'] = 'oj-legend-section-open-icon oj-active';
  },

  //** @inheritdoc */
  _Render: function() {
    this._super();
  },

  /**
   * Returns the legend title for automation testing verification.
   * @return {String} The legend title
   * @expose
   * @instance
   * @memberof oj.ojLegend
   * @ignore
   */
  getTitle : function() {
    var auto = this._component.getAutomation();
    return auto.getTitle();
  },

  /**
   * Returns an object with the following properties for automation testing verification of the legend section with
   * the specified subid path.
   *
   * @param {Array} subIdPath The array of indices in the subId for the desired legend section.
   * @property {string} title
   * @property {function(number)} getSection Returns the section with the specified index.
   * @property {string} getSection.title
   * @return {Object|null} An object containing properties for the legend section at the given subIdPath, or null if
   *   none exists.
   * @expose
   * @instance
   * @memberof oj.ojLegend
   */
  getSection : function(subIdPath) {
    var ret = this._component.getAutomation().getSection(subIdPath);
    if(ret) {
      // Support for getSection(sectionIndex)
      ret['getSection'] = function(sectionIndex) {
        var section = ret['sections'] ? ret['sections'][sectionIndex] : null;
        return section;
      }

      // Support for getSection(itemIndex)
      ret['getItem'] = function(itemIndex) {
        var item = ret['items'] ? ret['items'][itemIndex] : null;
        return item;
      }
    }
    return ret;
  },

  /**
   * Returns an object with the following properties for automation testing verification of the legend item with
   * the specified subid path.
   *
   * @param {Array} subIdPath The array of indices in the subId for the desired legend item.
   * @property {string} text
   * @return {Object|null} An object containing properties for the legend item at the given subIdPath, or null if
   *   none exists.
   * @expose
   * @instance
   * @memberof oj.ojLegend
   */
  getItem : function(subIdPath) {
    return this._component.getAutomation().getItem(subIdPath);
  },

  /**
   * Returns the preferred size of the legend, given the available width and height. A re-render must be triggered
   * by calling <code class="prettyprint">refresh</code> after invoking this function.
   * @param {Number} width The available width.
   * @param {Number} height The available height.
   * @return {Object} An object containing the preferred width and height.
   * @expose
   * @instance
   * @memberof oj.ojLegend
   */
  getPreferredSize : function(width, height) {
    // Check if the options has a promise.
    var hasPromise = false;
    var legendSections = this.options['sections'] ? this.options['sections'] : [];
    for(var i=0; i<legendSections.length; i++) {
      var items = legendSections[i]['items'];
      if(items && items.then) 
        hasPromise = true;
    }
    
    // If the options has a promise, then use the last options to be rendered rather 
    // than passing in a promise that can't be dealt with here. This won't work if the
    // data is provided via a promise and not yet rendered, but this problem will go 
    // away once we have flowing layout.
    var options = hasPromise ? null : this.options;
    var dims = this._component.getPreferredSize(options, width, height);
    return {'width': dims.getWidth(), 'height': dims.getHeight()};
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojLegend
   */
  getContextByNode: function(node)
  {
    // context objects are documented with @ojnodecontext
   var context = this.getSubIdByNode(node);
   if (context && context['subId'] !== 'oj-legend-tooltip')
     return context;

   return null;
  },

  //** @inheritdoc */
  _GetComponentDeferredDataPaths : function() {
    return {'sections': ['items']};
  },

  //** @inheritdoc */
  _GetComponentNoClonePaths : function() {
    return {'sections': {'items': true}};
  }
});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="2">Legend Item</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Filter when <code class="prettyprint">hideAndShowBehavior</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Highlight when <code class="prettyprint">hoverBehavior</code> is enabled.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojLegend
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus to previous item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus to next item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus to previous item (on left).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus to next item (on right).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Hides or unhides the data associated with the current item.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojLegend
 */

/**
 * The title of the legend section.
 * @expose
 * @name sections[].title
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The horizontal alignment of the section title. If the section is collapsible or nested, only start alignment is supported.
 * @expose
 * @name sections[].titleHalign
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "center"
 * @ojvalue {string} "end"
 * @ojvalue {string} "start"
 * @default <code class="prettyprint">"start"</code>
 */
/**
 * The CSS style object defining the style of the section title.
 * @expose
 * @name sections[].titleStyle
 * @memberof! oj.ojLegend
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * Whether the section is collapsible. Only applies if the legend orientation is vertical.
 * @expose
 * @name sections[].collapsible
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */
/**
 * Whether the section is initially expanded. Only applies if the section is collapsible.
 * @expose
 * @name sections[].expanded
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * An array of nested legend sections.
 * @expose
 * @name sections[].sections
 * @memberof! oj.ojLegend
 * @instance
 * @type {Array.<Object>}
 * @default null
 */
/**
 * An array of objects with the following properties defining the legend items. Also accepts a Promise for deferred data rendering. No data will be rendered if the Promise is rejected.
 * @expose
 * @name sections[].items
 * @memberof! oj.ojLegend
 * @instance
 * @type {Array.<Object>|Promise}
 * @default null
 */
/**
 * The id of the legend item, which is provided as part of the context for events fired by the legend. If not specified, the id defaults to the text of the legend item.
 * @expose
 * @name sections[].items[].id
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The legend item text.
 * @expose
 * @name sections[].items[].text
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An array of categories for the legend item. Legend items currently only support a single category. If no category is specified, this defaults to the id or text of the legend item.
 * @expose
 * @name sections[].items[].categories
 * @memberof! oj.ojLegend
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The type of legend symbol to display.
 * @expose
 * @name sections[].items[].symbolType
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "line"
 * @ojvalue {string} "lineWithMarker"
 * @ojvalue {string} "image"
 * @ojvalue {string} "marker"
 * @default <code class="prettyprint">"marker"</code>
 */
/**
 * The URI of the image of the legend symbol.
 * @expose
 * @name sections[].items[].source
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The color of the legend symbol (line or marker). When symbolType is "lineWithMarker", this attribute defines the line color and the markerColor attribute defines the marker color.
 * @expose
 * @name sections[].items[].color
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border color of the marker. Only applies if symbolType is "marker" or "lineWithMarker".
 * @expose
 * @name sections[].items[].borderColor
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The pattern used to fill the marker. Only applies if symbolType is "marker" or "lineWithMarker".
 * @expose
 * @name sections[].items[].pattern
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The line style. Only applies when the symbolType is "line" or "lineWithMarker".
 * @expose
 * @name sections[].items[].lineStyle
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "dotted"
 * @ojvalue {string} "dashed"
 * @ojvalue {string} "solid"
 * @default <code class="prettyprint">"solid"</code>
 */
/**
 * The line width in pixels. Only applies when the symbolType is "line" or "lineWithMarker".
 * @expose
 * @name sections[].items[].lineWidth
 * @memberof! oj.ojLegend
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The CSS style class to apply to the legend item. The style class and inline style will override any other styling specified through the options. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @expose
 * @name sections[].items[].svgClassName
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the legend item. The style class and inline style will override any other styling specified through the options. For tooltips and hover interactivity, it's recommended to also pass a representative color to the color attribute.
 * @expose
 * @name sections[].items[].svgStyle
 * @memberof! oj.ojLegend
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class to apply to the marker. The style class and inline style will override any other styling specified through the options. For tooltips and hover interactivity, it's recommended to also pass a representative color to the markerColor attribute.
 * @expose
 * @name sections[].items[].markerSvgClassName
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the marker. The style class and inline style will override any other styling specified through the options. For tooltips and hover interactivity, it's recommended to also pass a representative color to the markerColor attribute.
 * @expose
 * @name sections[].items[].markerSvgStyle
 * @memberof! oj.ojLegend
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The shape of the marker. Only applies if symbolType is "marker" or "lineWithMarker". Can take the name of a built-in shape or the svg path commands for a custom shape. Does not apply if a custom image is specified.
 * @expose
 * @name sections[].items[].markerShape
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "circle"
 * @ojvalue {string} "ellipse"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "human"
 * @ojvalue {string} "rectangle"
 * @ojvalue {string} "star"
 * @ojvalue {string} "square"
 * @default <code class="prettyprint">"square"</code>
 */
/**
 * The color of the marker, if different than the line color. Only applies if the symbolType is "lineWithMarker".
 * @expose
 * @name sections[].items[].markerColor
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines whether the legend item corresponds to visible data items. A hollow symbol is shown if the value is "hidden".
 * @expose
 * @name sections[].items[].categoryVisibility
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "hidden"
 * @ojvalue {string} "visible"
 * @default <code class="prettyprint">"visible"</code>
 */
/**
 *  Whether drilling is enabled on the legend item. Drillable objects will show a pointer cursor on hover and fire <code class="prettyprint">ojDrill</code> event on click. To enable drilling for all legend items at once, use the drilling attribute in the top level. 
 * @expose
 * @name sections[].items[].drilling
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default <code class="prettyprint">"inherit"</code>
 */
/**
 * The description of this legend item. This is used for accessibility and for customizing the tooltip text.
 * @expose
 * @name sections[].items[].shortDesc
 * @memberof! oj.ojLegend
 * @instance
 * @type {string}
 * @default null
 */

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for legend items indexed by their section and item indices.</p>
 *
 * @property {Array} sectionIndexPath The array of numerical indices for the section.
 * @property {number} itemIndex The index of the item within the specified section.
 *
 * @ojsubid oj-legend-item
 * @memberof oj.ojLegend
 *
 * @example <caption>Get the second item in the first section:</caption>
 * var nodes = $( ".selector" ).ojLegend( "getNodeBySubId", {'subId': 'oj-legend-item', sectionIndexPath: [0], itemIndex: 1} );
 */

/**
 * <p>Sub-ID for the the legend tooltip.</p>
 * 
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 * 
 * @ojsubid
 * @member
 * @name oj-legend-tooltip
 * @memberof oj.ojLegend
 * @instance
 * 
 * @example <caption>Get the tooltip object of the legend, if displayed:</caption>
 * var nodes = $( ".selector" ).ojLegend( "getNodeBySubId", {'subId': 'oj-legend-tooltip'} );
 */
// Node Context Objects ********************************************************


/**
 * <p>Context for legend items indexed by their section and item indices.</p>
 *
 * @property {Array} sectionIndexPath The array of numerical indices for the section.
 * @property {number} itemIndex The index of the item within the specified section.
 *
 * @ojnodecontext oj-legend-item
 * @memberof oj.ojLegend
 */

/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojLegendMeta = {
  "properties": {
    "drilling": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "halign": {
      "type": "string",
      "enumValues": ["center", "end", "start"]
    },
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "hideAndShowBehavior": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "hoverBehaviorDelay": {
      "type": "number"
    },    "orientation": {
      "type": "string",
      "enumValues": ["horizontal", "vertical"]
    },
    "scrolling": {
      "type": "string",
      "enumValues": ["asNeeded", "off"]
    },
    "sections": {
      "type": "Array<object>"
    },
    "symbolHeight": {
      "type": "number"
    },
    "symbolWidth": {
      "type": "number"
    },
    "textStyle": {
      "type": "object"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Legend"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    },
    "valign": {
      "type": "string"
    }
  },
  "events": {
    "drill": {}
  },
  "methods": {
    "getContextByNode": {},
    "getItem": {},
    "getPreferredSize": {},
    "getSection": {}
  },
  "extension": {
    _WIDGET_NAME: "ojLegend"
  }
};
oj.CustomElementBridge.registerMetadata('oj-legend', 'dvtBaseComponent', ojLegendMeta);
oj.CustomElementBridge.register('oj-legend', {'metadata': oj.CustomElementBridge.getMetadata('oj-legend')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * Class which provides list based optimizations
 */
var ListDataProviderView = /** @class */ (function () {
    function ListDataProviderView(dataProvider, options) {
        this.dataProvider = dataProvider;
        this.options = options;
        this._KEY = 'key';
        this._KEYS = 'keys';
        this._DATA = 'data';
        this._STARTINDEX = 'startIndex';
        this._SORT = 'sort';
        this._SORTCRITERIA = 'sortCriteria';
        this._FILTERCRITERION = 'filterCriterion';
        this._METADATA = 'metadata';
        this._FROM = 'from';
        this._OFFSET = 'offset';
        this._REFRESH = 'refresh';
        this._MUTATE = 'mutate';
        this._SIZE = 'size';
        this._FETCHPARAMETERS = 'fetchParameters';
        this._VALUE = 'value';
        this._DONE = 'done';
        this._DATAMAPPING = 'dataMapping';
        this._MAPFIELDS = 'mapFields';
        this._MAPSORTCRITERIA = 'mapSortCriteria';
        this._MAPFILTERCRITERION = 'mapFilterCriterion';
        this._UNMAPSORTCRITERIA = 'mapSortCriteria';
        this._RESULTS = 'results';
        this._CONTAINSPARAMETERS = 'containsParameters';
        this._DEFAULT_SIZE = 25;
        this._CONTAINSKEYS = 'containsKeys';
        this._FETCHBYKEYS = 'fetchByKeys';
        this._FETCHBYOFFSET = 'fetchByOffset';
        this._FETCHFIRST = 'fetchFirst';
        this._ADDEVENTLISTENER = 'addEventListener';
        this.AsyncIterable = /** @class */ (function () {
            function class_1(_parent, _asyncIterator) {
                this._parent = _parent;
                this._asyncIterator = _asyncIterator;
                this[Symbol.asyncIterator] = function () {
                    return this._asyncIterator;
                };
            }
            return class_1;
        }());
        this.AsyncIterator = /** @class */ (function () {
            function class_2(_parent, _nextFunc, _params) {
                this._parent = _parent;
                this._nextFunc = _nextFunc;
                this._params = _params;
            }
            class_2.prototype['next'] = function () {
                var result = this._nextFunc(this._params);
                return Promise.resolve(result);
            };
            return class_2;
        }());
        this.AsyncIteratorResult = /** @class */ (function () {
            function class_3(_parent, value, done) {
                this._parent = _parent;
                this.value = value;
                this.done = done;
                this[_parent._VALUE] = value;
                this[_parent._DONE] = done;
            }
            return class_3;
        }());
        this.FetchListResult = /** @class */ (function () {
            function class_4(_parent, fetchParameters, data, metadata) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.data = data;
                this.metadata = metadata;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._DATA] = data;
                this[_parent._METADATA] = metadata;
            }
            return class_4;
        }());
        this.Item = /** @class */ (function () {
            function class_5(_parent, metadata, data) {
                this._parent = _parent;
                this.metadata = metadata;
                this.data = data;
                this[_parent._METADATA] = metadata;
                this[_parent._DATA] = data;
            }
            return class_5;
        }());
        this.ItemMetadata = /** @class */ (function () {
            function class_6(_parent, key) {
                this._parent = _parent;
                this.key = key;
                this[_parent._KEY] = key;
            }
            return class_6;
        }());
        this.FetchListParameters = /** @class */ (function () {
            function class_7(_parent, size, sortCriteria, filterCriterion) {
                this._parent = _parent;
                this.size = size;
                this.sortCriteria = sortCriteria;
                this.filterCriterion = filterCriterion;
                this[_parent._SIZE] = size;
                this[_parent._SORTCRITERIA] = sortCriteria;
                this[_parent._FILTERCRITERION] = filterCriterion;
            }
            return class_7;
        }());
        this.FetchByOffsetParameters = /** @class */ (function () {
            function class_8(_parent, offset, size, sortCriteria, filterCriterion) {
                this._parent = _parent;
                this.offset = offset;
                this.size = size;
                this.sortCriteria = sortCriteria;
                this.filterCriterion = filterCriterion;
                this[_parent._SIZE] = size;
                this[_parent._SORTCRITERIA] = sortCriteria;
                this[_parent._OFFSET] = offset;
                this[_parent._FILTERCRITERION] = filterCriterion;
            }
            return class_8;
        }());
        this.FetchByKeysResults = /** @class */ (function () {
            function class_9(_parent, fetchParameters, results) {
                this._parent = _parent;
                this.fetchParameters = fetchParameters;
                this.results = results;
                this[_parent._FETCHPARAMETERS] = fetchParameters;
                this[_parent._RESULTS] = results;
            }
            return class_9;
        }());
        this.ContainsKeysResults = /** @class */ (function () {
            function class_10(_parent, containsParameters, results) {
                this._parent = _parent;
                this.containsParameters = containsParameters;
                this.results = results;
                this[_parent._CONTAINSPARAMETERS] = containsParameters;
                this[_parent._RESULTS] = results;
            }
            return class_10;
        }());
        this[this._FROM] = this.options == null ? null : this.options[this._FROM];
        this[this._OFFSET] = this.options == null ? 0 : this.options[this._OFFSET] > 0 ? this.options[this._OFFSET] : 0;
        this[this._SORTCRITERIA] = this.options == null ? null : this.options[this._SORTCRITERIA];
        this[this._DATAMAPPING] = this.options == null ? null : this.options[this._DATAMAPPING];
        this._addEventListeners(dataProvider);
    }
    ListDataProviderView.prototype.containsKeys = function (params) {
        var self = this;
        if (this.dataProvider[self._CONTAINSKEYS]) {
            return this.dataProvider[self._CONTAINSKEYS](params);
        }
        else {
            return this.fetchByKeys(params).then(function (fetchByKeysResult) {
                var results = new Set();
                params[self._KEYS].forEach(function (key) {
                    if (fetchByKeysResult[self._RESULTS].get(key) != null) {
                        results.add(key);
                    }
                });
                return Promise.resolve(new self.ContainsKeysResults(self, params, results));
            });
        }
    };
    ListDataProviderView.prototype.fetchByKeys = function (params) {
        var self = this;
        if (this.dataProvider[self._FETCHBYKEYS]) {
            return this.dataProvider[self._FETCHBYKEYS](params).then(function (value) {
                var resultMap = value[self._RESULTS];
                var mappedResultMap = new Map();
                resultMap.forEach(function (value, key) {
                    var mappedItem = self._getMappedItems([value]);
                    mappedResultMap.set(key, mappedItem[0]);
                });
                return new self.FetchByKeysResults(self, params, mappedResultMap);
            });
        }
        else {
            var options = {};
            options[this._SIZE] = this._DEFAULT_SIZE;
            var resultMap = new Map();
            var dataProviderAsyncIterator = this.dataProvider[self._FETCHFIRST](options)[Symbol.asyncIterator]();
            return this._fetchNextSet(params, dataProviderAsyncIterator, resultMap).then(function (resultMap) {
                var mappedResultMap = new Map();
                resultMap.forEach(function (value, key) {
                    var mappedItem = self._getMappedItems([value]);
                    mappedResultMap.set(key, mappedItem[0]);
                });
                return new self.FetchByKeysResults(self, params, mappedResultMap);
            });
        }
    };
    ListDataProviderView.prototype.fetchFirst = function (params) {
        // this fetchFirst applies the offset and from properties on the this.
        // If fetchByOffset is supported by the underlying dataprovider then that is used for offset.
        // Otherwise, fetches are made in chunks until from and offset are fulfilled.
        var cachedData = {};
        cachedData[this._DATA] = [];
        cachedData[this._KEYS] = [];
        cachedData[this._DONE] = false;
        cachedData[this._STARTINDEX] = 0;
        var size = params != null ? params[this._SIZE] : null;
        var sortCriteria = params != null ? params[this._SORTCRITERIA] : null;
        if (sortCriteria == null) {
            sortCriteria = this[this._SORTCRITERIA];
        }
        var mappedSortCriteria = this._getMappedSortCriteria(sortCriteria);
        var filterCriterion = params != null ? params[this._FILTERCRITERION] : null;
        var mappedFilterCriterion = this._getMappedFilterCriterion(filterCriterion);
        var self = this;
        if (self[this._FROM] == null &&
            self[this._OFFSET] > 0 &&
            oj.DataProviderFeatureChecker.isFetchByOffset(self.dataProvider)) {
            var offset_1 = self[this._OFFSET];
            return new this.AsyncIterable(this, new this.AsyncIterator(this, function (cachedData) {
                return function () {
                    var updatedParams = new self.FetchByOffsetParameters(self, offset_1, size, mappedSortCriteria, mappedFilterCriterion);
                    return self.dataProvider[self._FETCHBYOFFSET](updatedParams).then(function (result) {
                        var results = result['results'];
                        var data = results.map(function (value) {
                            return value[self._DATA];
                        });
                        var metadata = results.map(function (value) {
                            return value[self._METADATA];
                        });
                        var keys = results.map(function (value) {
                            return value[self._METADATA][self._KEY];
                        });
                        offset_1 = offset_1 + keys.length;
                        var mappedResult = self._getMappedDataAndKeys(keys, data);
                        self._cacheResult(cachedData, mappedResult[self._DATA], mappedResult[self._KEYS]);
                        cachedData[self._DONE] = result[self._DONE];
                        var resultFetchParams = result[self._FETCHPARAMETERS];
                        var resultSortCriteria = resultFetchParams != null ? resultFetchParams[self._SORTCRITERIA] : null;
                        var mappedResultSortCriteria = self._getUnmappedSortCriteria(resultSortCriteria);
                        var resultParams = new self.FetchByOffsetParameters(self, self[self._OFFSET], size, mappedResultSortCriteria);
                        // if the dataprovider supports fetchByOffset then we use that to do an offset based fetch
                        return Promise.resolve(new self.AsyncIteratorResult(self, new self.FetchListResult(self, resultParams, mappedResult[self._DATA], mappedResult[self._METADATA]), cachedData[self._DONE]));
                    });
                };
            }(cachedData), params));
        }
        else {
            var updatedParams = new this.FetchListParameters(this, size, mappedSortCriteria, mappedFilterCriterion);
            var cachedAsyncIterator = this.dataProvider[self._FETCHFIRST](updatedParams)[Symbol.asyncIterator]();
            return new this.AsyncIterable(this, new this.AsyncIterator(this, function (cachedData, cachedAsyncIterator) {
                return function () {
                    return cachedAsyncIterator.next().then(function (result) {
                        var data = result[self._VALUE][self._DATA];
                        var metadata = result[self._VALUE][self._METADATA];
                        var keys = metadata.map(function (value) {
                            return value[self._KEY];
                        });
                        var items = data.map(function (value, index) {
                            return new self.Item(self, metadata[index], data[index]);
                        });
                        // apply any mapping defined in the DataMapping parameter
                        var mappedResult = self._getMappedDataAndKeys(keys, data);
                        self._cacheResult(cachedData, mappedResult[self._DATA], mappedResult[self._KEYS]);
                        cachedData[self._DONE] = result[self._DONE];
                        var size = params != null ? params[self._SIZE] : null;
                        var offset = params != null ? params[self._OFFSET] : null;
                        var resultFetchParams = result[self._VALUE][self._FETCHPARAMETERS];
                        var resultSortCriteria = resultFetchParams != null ? resultFetchParams[self._SORTCRITERIA] : null;
                        var mappedResultSortCriteria = self._getUnmappedSortCriteria(resultSortCriteria);
                        var resultParams = new self.FetchListParameters(self, size, mappedResultSortCriteria);
                        return self._fetchUntilKey(resultParams, self[self._FROM], cachedData, cachedAsyncIterator).then(function () {
                            return self._fetchUntilOffset(resultParams, self[self._OFFSET] + cachedData[self._STARTINDEX], data.length, cachedData, cachedAsyncIterator);
                        });
                    });
                };
            }(cachedData, cachedAsyncIterator), params));
        }
    };
    ListDataProviderView.prototype.getCapability = function (capabilityName) {
        return this.dataProvider.getCapability(capabilityName);
    };
    ListDataProviderView.prototype.getTotalSize = function () {
        return this.dataProvider.getTotalSize();
    };
    ListDataProviderView.prototype.isEmpty = function () {
        return this.dataProvider.isEmpty();
    };
    /**
     * Fetches the next block
     */
    ListDataProviderView.prototype._fetchNextSet = function (params, dataProviderAsyncIterator, resultMap) {
        var self = this;
        return dataProviderAsyncIterator.next().then(function (result) {
            var value = result[self._VALUE];
            var data = value[self._DATA];
            var metadata = value[self._METADATA];
            var keys = metadata.map(function (metadata) {
                return metadata[self._KEY];
            });
            var foundAllKeys = true;
            params[self._KEYS].forEach(function (findKey) {
                if (!resultMap.has(findKey)) {
                    keys.map(function (key, index) {
                        if (oj.Object.compareValues(key, findKey)) {
                            resultMap.set(findKey, new self.Item(self, metadata[index], data[index]));
                        }
                    });
                }
                if (!resultMap.has(findKey)) {
                    foundAllKeys = false;
                }
            });
            if (!foundAllKeys && !result[self._DONE]) {
                return self._fetchNextSet(params, dataProviderAsyncIterator, resultMap);
            }
            else {
                return resultMap;
            }
        });
    };
    /**
     * Fetches until we find the key
     */
    ListDataProviderView.prototype._fetchUntilKey = function (params, key, cachedData, cachedAsyncIterator) {
        var self = this;
        if (key != null) {
            // first check if the key is in our cache
            var resultKeys = cachedData[this._KEYS].filter(function (resultKey) {
                if (key == resultKey) {
                    return true;
                }
            });
            if (resultKeys.length > 0) {
                // if the key is in our cache, then trim the cache so that it starts from the key
                var keyIndex = cachedData[this._KEYS].indexOf(resultKeys[0]);
                cachedData[this._KEYS] = cachedData[this._KEYS].slice(keyIndex, cachedData[this._KEYS].length);
                cachedData[this._DATA] = cachedData[this._DATA].slice(keyIndex, cachedData[this._DATA].length);
            }
            else if (!cachedData[self._DONE]) {
                // if the key is not in our cache and we are not done then fetch the next block and call _fetchUntilKey again.
                return cachedAsyncIterator.next().then(function (nextResult) {
                    var data = nextResult[self._VALUE][self._DATA];
                    var keys = nextResult[self._VALUE][self._METADATA].map(function (value) {
                        return value[self._KEY];
                    });
                    var mappedResult = self._getMappedDataAndKeys(keys, data);
                    self._cacheResult(cachedData, mappedResult[self._DATA], mappedResult[self._KEYS]);
                    cachedData[self._DONE] = nextResult[self._DONE];
                    return self._fetchUntilKey(nextResult[self._FETCHPARAMETERS], mappedResult[self._KEYS], cachedData, cachedAsyncIterator);
                });
            }
            else {
                // if we are done then this means that the key is not in the entire data set
                cachedData[this._DATA] = [];
                cachedData[this._KEYS] = [];
            }
        }
        return Promise.resolve(null);
    };
    /**
     * Fetches until we fulfill the offset
     */
    ListDataProviderView.prototype._fetchUntilOffset = function (params, offset, resultSize, cachedData, cachedAsyncIterator) {
        var self = this;
        var fetchSize = params != null ? params[this._SIZE] > 0 ? params[this._SIZE] : resultSize : resultSize;
        offset = offset > 0 ? offset : 0;
        var keys = cachedData[this._KEYS].slice(offset, offset + fetchSize);
        var data = cachedData[this._DATA].slice(offset, offset + fetchSize);
        var metadata = keys.map(function (value) {
            return new self.ItemMetadata(self, value);
        });
        if (data.length < fetchSize) {
            if (!cachedData[self._DONE]) {
                return cachedAsyncIterator.next().then(function (nextResult) {
                    var data = nextResult[self._VALUE][self._DATA];
                    var keys = nextResult[self._VALUE][self._METADATA].map(function (value) {
                        return value[self._KEY];
                    });
                    var mappedResult = self._getMappedDataAndKeys(keys, data);
                    self._cacheResult(cachedData, mappedResult[self._DATA], mappedResult[self._KEYS]);
                    cachedData[self._DONE] = nextResult[self._DONE];
                    return self._fetchUntilOffset(nextResult[self._VALUE][self._FETCHPARAMETERS], offset, data.length, cachedData, cachedAsyncIterator);
                });
            }
            else {
                cachedData[this._STARTINDEX] = cachedData[this._STARTINDEX] + data.length;
                return Promise.resolve(new self.AsyncIteratorResult(self, new self.FetchListResult(self, params, data, metadata), true));
            }
        }
        else {
            cachedData[this._STARTINDEX] = cachedData[this._STARTINDEX] + data.length;
            return Promise.resolve(new self.AsyncIteratorResult(self, new self.FetchListResult(self, params, data, metadata), cachedData[self._DONE]));
        }
    };
    /**
     * Cache the data and keys
     */
    ListDataProviderView.prototype._cacheResult = function (cachedData, data, keys) {
        var self = this;
        data.map(function (value) {
            cachedData[self._DATA].push(value);
        });
        keys.map(function (value) {
            cachedData[self._KEYS].push(value);
        });
    };
    /**
     * Apply DataMapping to the items
     */
    ListDataProviderView.prototype._getMappedItems = function (items) {
        var self = this;
        if (this[this._DATAMAPPING] != null) {
            var mapFields_1 = this[this._DATAMAPPING][this._MAPFIELDS];
            if (mapFields_1 != null) {
                if (items != null && items.length > 0) {
                    var mappedItems = new Array();
                    mappedItems = items.map(function (value) {
                        return mapFields_1.bind(self)(value);
                    });
                    return mappedItems;
                }
            }
        }
        return items;
    };
    /**
     * Apply DataMapping to the keys and data
     */
    ListDataProviderView.prototype._getMappedDataAndKeys = function (keys, data) {
        var self = this;
        var items = data.map(function (value, index) {
            return new self.Item(self, new self.ItemMetadata(self, keys[index]), data[index]);
        });
        var mappedItems = this._getMappedItems(items);
        var mappedData = mappedItems.map(function (value) {
            return value[self._DATA];
        });
        var mappedKeys = mappedItems.map(function (value) {
            return value[self._METADATA][self._KEY];
        });
        var mappedMetadata = mappedItems.map(function (value) {
            return value[self._METADATA];
        });
        var result = {};
        result[this._DATA] = mappedData;
        result[this._KEYS] = mappedKeys;
        result[this._METADATA] = mappedMetadata;
        return result;
    };
    /**
     * Apply mapping to the filterCriterion
     */
    ListDataProviderView.prototype._getMappedFilterCriterion = function (filterCriterion) {
        if (this[this._DATAMAPPING] != null) {
            var mappedFilterCriterion = this[this._DATAMAPPING][this._MAPFILTERCRITERION];
            if (mappedFilterCriterion != null) {
                if (filterCriterion != null) {
                    return mappedFilterCriterion(filterCriterion);
                }
            }
        }
        return filterCriterion;
    };
    /**
     * Apply mapping to the sortCriteria
     */
    ListDataProviderView.prototype._getMappedSortCriteria = function (sortCriteria) {
        if (this[this._DATAMAPPING] != null) {
            var mapSortCriteria = this[this._DATAMAPPING][this._MAPSORTCRITERIA];
            if (mapSortCriteria != null) {
                if (sortCriteria != null && sortCriteria.length > 0) {
                    return mapSortCriteria(sortCriteria);
                }
            }
        }
        return sortCriteria;
    };
    /**
     * Unmapping the sortCriteria
     */
    ListDataProviderView.prototype._getUnmappedSortCriteria = function (sortCriteria) {
        if (this[this._DATAMAPPING] != null) {
            var unmapSortCriteria = this[this._DATAMAPPING][this._UNMAPSORTCRITERIA];
            if (unmapSortCriteria != null) {
                if (sortCriteria != null && sortCriteria.length > 0) {
                    return unmapSortCriteria(sortCriteria);
                }
            }
        }
        return sortCriteria;
    };
    /**
     * Add event listeners
     */
    ListDataProviderView.prototype._addEventListeners = function (dataprovider) {
        var self = this;
        dataprovider[self._ADDEVENTLISTENER](this._REFRESH, function (event) {
            self.dispatchEvent(event);
        });
        dataprovider[self._ADDEVENTLISTENER](this._MUTATE, function (event) {
            self.dispatchEvent(event);
        });
    };
    return ListDataProviderView;
}());
oj['ListDataProviderView'] = ListDataProviderView;
oj.ListDataProviderView = ListDataProviderView;
oj.EventTargetMixin.applyMixin(ListDataProviderView);
oj['FetchByOffsetMixin'].applyMixin(ListDataProviderView);

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @class oj.ListDataProviderView
 * @implements oj.DataProvider
 * @classdesc Provides list based optimizations for oj.DataProvider. Supports fetchFirst starting at arbitrary key or index offset, sortCriteria,
 * and field mapping. Please see the select demos for examples of DataMapping [Select]{@link oj.ojSelect}
 * @param {oj.DataProvider} dataProvider the DataProvider.
 * @param {Object=} options Options for the ListDataProviderView
 * @param {Object=} options.from key to start fetching from. This will be applied first before offset is applied.
 * @param {number=} options.offset offset to start fetching from.
 * @param {Array.<oj.SortCriterion>=} options.sortCriteria {@link oj.sortCriteria} to apply to the data.
 * @param {oj.DataMapping=} options.dataMapping mapping to apply to the data.
 * @ojsignature [{target: "Type",
 *               value: "class ListDataProviderView<K, D, Kin, Din> implements DataProvider<K, D>"},
 *               {target: "Type",
 *               value: "DataProvider<K, D>", 
 *               for: "dataProvider"},
 *               {target: "Type",
 *               value: "SortCriterion<D>[]", 
 *               for: "options.sortCriteria"},
 *               {target: "Type",
 *               value: "DataMapping<K, D, Kin, Din>", 
 *               for: "options.dataMapping"}]
 */

/**
 * Check if there are rows containing the specified keys
 *
 * @ojstatus preview
 * @since 4.1.0
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.ContainsKeysResults>} Promise which resolves to {@link oj.ContainsKeysResults}
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name containsKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<ContainsKeysResults<K>>"}
 */

/**
 * Fetch rows by keys
 *
 * @ojstatus preview
 * @since 4.1.0
 * @param {oj.FetchByKeysParameters} params Fetch by keys parameters
 * @return {Promise.<oj.FetchByKeysResults>} Promise which resolves to {@link oj.FetchByKeysResults}
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name fetchByKeys
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByKeysParameters<K>): Promise<FetchByKeysResults<K, D>>"}
 */

/**
 * Fetch rows by offset
 *
 * @ojstatus preview
 * @since 4.2.0
 * @param {oj.FetchByOffsetParameters} params Fetch by offset parameters
 * @return {Promise.<oj.FetchByOffsetResults>} Promise which resolves to {@link oj.FetchByOffsetResults}
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name fetchByOffset
 * @ojsignature {target: "Type",
 *               value: "(params: FetchByOffsetParameters<D>): Promise<FetchByOffsetResults<K, D>>"}
 */

/**
 * Fetch the first block of data.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @param {oj.FetchListParameters=} params Fetch parameters
 * @return {AsyncIterable.<oj.FetchListResult>} AsyncIterable with {@link oj.FetchListResult}
 * @see {@link https://github.com/tc39/proposal-async-iteration} for further information on AsyncIterable.
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name fetchFirst
 * @ojsignature {target: "Type",
 *               value: "<F extends FetchListResult<K, D>>(params?: FetchListParameters<D>): AsyncIterable<F>"}
 */

/**
 * Determines whether this DataProvider supports certain feature.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @param {string} capabilityName capability name. Supported capability names
 *                  are determined by the underlying dataprovider.
 * @return {Object} capability information or null if unsupported
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name getCapability
 * @ojsignature {target: "Type",
 *               value: "(capabilityName: string)"}
 */

/**
 * Return the total number of rows in this dataprovider
 * 
 * @ojstatus preview
 * @return {Promise.<number>} Returns a Promise which resolves to the total number of rows. -1 is unknown row count.
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name getTotalSize
 */

/**
 * Return a string that indicates if this data provider is empty
 * 
 * @ojstatus preview
 * @return {"yes"|"no"|"unknown"} a string that indicates if this data provider is empty. Valid values are:
 *                  "yes": this data provider is empty.
 *                  "no": this data provider is not empty.
 *                  "unknown": it is not known if this data provider is empty until a fetch is made.
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name isEmpty
 */

/**
 * Optional key to start fetching from.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @name from
 * @type {any}
 */

/**
 * Optional offset to start fetching from. Should be greater than or equal to zero.
 * If a negative offset is used then it will be treated as zero.
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @name offset
 * @type {number}
 */

/**
 * Optional sortCriteria to apply
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @name sortCriteria
 * @ojsignature {target: "Type",
 *               value: "SortCriterion<D>[]"}
 */

/**
 * Optional dataMapping to apply
 * 
 * @ojstatus preview
 * @since 4.1.0
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @name dataMapping
 * @ojsignature {target: "Type",
 *               value: "DataMapping<K, D, Kin, Din>"}
 */

/**
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name addEventListener
 * @ojsignature {target: "Type",
 *               value: "(eventType: string, listener: EventListener): void"}
 */

/**
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name removeEventListener
 * @ojsignature {target: "Type",
 *               value: "(eventType: string, listener: EventListener): void"}
 */

/**
 * @ojstatus preview
 * @export
 * @expose
 * @memberof oj.ListDataProviderView
 * @instance
 * @method
 * @name dispatchEvent
 * @ojsignature {target: "Type",
 *               value: "(evt: Event): boolean"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @export
 * @class oj.ListViewDndContext
 * @classdesc Drag and Drop Utils for ojListView
 * @param {Object} listview the ListView instance
 * @constructor
 */
oj.ListViewDndContext = function(listview)
{
  this.listview = listview;
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.ListViewDndContext, oj.Object, "oj.ListViewDndContext");

oj.ListViewDndContext.C_KEY = 67;
oj.ListViewDndContext.V_KEY = 86;
oj.ListViewDndContext.X_KEY = 88;

oj.ListViewDndContext.CUT_COMMAND = "cut";
oj.ListViewDndContext.COPY_COMMAND = "copy";
oj.ListViewDndContext.PASTE_COMMAND = "paste";
oj.ListViewDndContext.PASTE_BEFORE_COMMAND = "pasteBefore";
oj.ListViewDndContext.PASTE_AFTER_COMMAND = "pasteAfter";

/**
 * Clears the internal of dnd context.  Called by _resetInternal in ListView.
 */
oj.ListViewDndContext.prototype.reset = function()
{
    // this cache might not be clear if dnd never did happened
    this._unsetSelectionDraggable();

    this.m_itemsDragged = null;
    this.m_dragImage = null;
    this.m_currentDragItem = null;
    this.m_dragItems = null;
    this.m_contextMenuItem = null;
};

/******************************** common helpers ***********************************************/

/**
 * Retrieve options for drag or drop
 * @param {string} op either drag or drop 
 * @return {Object} options for drag or drop
 * @private
 */
oj.ListViewDndContext.prototype._getDndOptions = function(op)
{
    var dnd = this.listview.GetOption("dnd");
    if (dnd != null && dnd[op])
    {
        return dnd[op]['items'];
    }

    return null;
};

/**
 * Retrieve options for drag
 * @return {Object} options for drag
 * @private
 */
oj.ListViewDndContext.prototype._getDragOptions = function()
{
    return this._getDndOptions('drag');
};

/**
 * Retrieve options for drop
 * @return {Object} options for drop
 * @private
 */
oj.ListViewDndContext.prototype._getDropOptions = function()
{
    return this._getDndOptions('drop');
};

/**
 * Return true if reordering enabled.
 * @return {boolean} whether item reordering is enable
 * @protected
 */
oj.ListViewDndContext.prototype.IsItemReOrdering = function()
{
    var option = this._getDndOptions('reorder');
    return (option == "enabled");
};

/**
 * Gets the default drag affordance marker class
 * @return {string} affordance marker class
 * @protected
 */
oj.ListViewDndContext.prototype.GetDragAffordanceClass = function()
{
    return "oj-listview-drag-handle";
};

/**
 * Gets the drag Image class. Need to be override by navlist.
 * @return {string} drag image class
 * @protected
 */
oj.ListViewDndContext.prototype.GetDragImageClass = function()
{
    return "oj-listview-drag-image";
};

/**
 * Gets the drag Item class.  Need to be override by navlist.
 * @return {string} drag Item class
 * @protected
 */
oj.ListViewDndContext.prototype.GetDragItemClass = function()
{
    return "oj-listview-drag-item";
};

/**
 * Gets the cut command style class.  Need to be override by navlist.
 * @return {string} cut command style class
 * @protected
 */
oj.ListViewDndContext.prototype.GetCutStyleClass = function()
{
    return "oj-listview-cut";
};

/**
 * Gets the style class prefix.  Need to be override by navlist.
 * @return {string} style class prefix
 * @protected
 */
oj.ListViewDndContext.prototype.GetCommandPrefix = function()
{
    return "oj-listview-";
};

oj.ListViewDndContext.prototype._findItem = function(target)
{
    return this.listview.FindItem(target);
};

/**
 * Gets selected items
 * @param {boolean=} excludeContextMenuItem true if to exclude the the item user right clicked on, false or undefined otherwise.
 * @return {Array.<Element>} an array of selected item's elements.
 * @private
 */
oj.ListViewDndContext.prototype._getSelectedItems = function(excludeContextMenuItem)
{
    var items, selection, i, elem;

    items = [];

    if (this.listview._isSelectionEnabled())
    {
        // if it's item then use selection
        selection = this.listview.GetOption("selection");
        for (i=0; i<selection.length; i++)
        {
            elem = this.listview.FindElementByKey(selection[i]);
            // make sure item is focusable also
            if (elem != null && !this.listview.SkipFocus($(elem)))
            {     
                items.push(elem);
            }
        }
    }
    else
    {
        // if there's no selection, use the active item
        elem = this._getActiveItem();
        if (elem != null)
        {
            items.push(elem);
        }
    }

    // include the item right click on, if it's not in the selection already
    if (!excludeContextMenuItem && this.m_contextMenuItem != null && this.m_contextMenuItem.length > 0)
    {
        // verify contextMenuItem is still valid 
        if (!this.listview.element.get(0).contains(this.m_contextMenuItem.get(0)))
        {
            this.m_contextMenuItem = null;
        }
        else if (items.indexOf(this.m_contextMenuItem.get(0)) == -1)
        {
            items.push(this.m_contextMenuItem.get(0));
        }
    }

    return items;
};

/**
 * @return {Element}
 * @private
 */
oj.ListViewDndContext.prototype._getActiveItem = function()
{
    if (this.listview.m_active == null)
    {
        return null;
    }
    else
    {
        return this.listview.m_active['elem'][0];
    }
};

/**
 * Gets the key and index from an item
 * @param {jQuery} item
 * @private
 */
oj.ListViewDndContext.prototype._getKeyAndIndex = function(item)
{
    var key, items, index;

    key = this.listview.GetKey(item.get(0));
    items = this.listview._getItemsCache();
    index = items.index(item); 

    return {key: key, index: index};
};

/**
 * Called by content handler once the content of an item is rendered
 * @param {Element} elem the item element
 * @param {Object} context the context object used for the item
 */
oj.ListViewDndContext.prototype.itemRenderComplete = function(elem, context)
{
    var dragHandle, ariaLabelledBy;

    // if it's a group, use the div instead
    if (!$(elem).hasClass(this.listview.getItemStyleClass()))
    {
        elem = elem.firstElementChild;
    }

    dragHandle = $(elem).find("."+this.GetDragAffordanceClass());
    if (dragHandle != null && dragHandle.length > 0)
    {
        ariaLabelledBy = dragHandle.attr("aria-labelledby");        
        if (ariaLabelledBy == null)
        {
            dragHandle.attr("aria-labelledby", this.listview._createSubId('instr'));
        }
        else
        {
            dragHandle.attr("aria-labelledby", ariaLabelledBy + ' ' + this.listview._createSubId('instr'));
        }

        // for touch draggable needs to be set prior to touch interaction
        if (this.listview._isTouchSupport())
        {
            dragHandle.attr("draggable", "true");
        }
    }
};
/******************************** Mouse down/up, touch start/end helpers ***********************************************/
oj.ListViewDndContext.prototype._unsetSelectionDraggable = function()
{
    if (this.m_draggableSelection)
    {
        $.each(this.m_draggableSelection, function(index, elem)
        {
            $(elem).removeClass("oj-draggable");
        });
    }
};

/** 
 * Make every item in the current selection draggable.
 * Invoke by listview when selection has changed
 * @private
 */
oj.ListViewDndContext.prototype.setSelectionDraggable = function()
{
    var elems = [], selection, i, elem;

    // clear previous draggable selections
    this._unsetSelectionDraggable();

    // make current selection draggable
    selection = this.listview.GetOption("selection");
    for (i=0; i<selection.length; i++)
    {
        elem = this.listview.FindElementByKey(selection[i]);
        // make sure item is focusable also
        if (elem != null && !this.listview.SkipFocus($(elem)))
        {     
            elems.push(elem);
            $(elem).addClass("oj-draggable");
        }
    }

    this.m_draggableSelection = elems;
};

/** 
 * Sets draggable style class on item
 * @param {jQuery} item the listview item
 * @return {boolean} true if no style class has applied, false otherwise
 * @private
 */
oj.ListViewDndContext.prototype._setItemDraggable = function(item)
{
    var cls, dragHandle;

    // if the item contains the drag affordance, then bail
    cls = this.GetDragAffordanceClass();
    dragHandle = item.find("."+cls);
    if (dragHandle != null && dragHandle.length > 0)
    {
        return true;
    }

    item.addClass("oj-draggable");
    return false;
};

/** 
 * Unsets draggable style class on item
 * @param {jQuery} item the listview item
 * @private
 */
oj.ListViewDndContext.prototype._unsetItemDraggable = function(item)
{
    // would be no-op if there's an affordance
    item.removeClass("oj-draggable");
};

/**
 * Sets draggable attribute on either the item or affordance.
 * See HandleMouseDownOrTouchStart method
 * @param {jQuery} target target from mouse down or touch start event
 * @private
 */
oj.ListViewDndContext.prototype._setDraggable = function(target)
{
    var cls, item, dragger, skipped, selectedItems;

    if (this._getDragOptions() != null || this.IsItemReOrdering())
    {
        cls = this.GetDragAffordanceClass();
        if (target.hasClass(cls))
        {
            dragger = $(target);
        }
        else
        {
            item = this._findItem(target);
            if (!this.shouldDragCurrentItem())
            {
                if (item != null)
                {
                    skipped = this._setItemDraggable(item);
                    if (skipped)
                    {
                        // contains affordance, bail out.
                        return;
                    }
                }

              // if dragging an item, it must be already the current item or one of the selected item
              selectedItems = this._getSelectedItems(true);
              if (selectedItems.length > 0)
              {
                  // in order to initiate drag the item must be the current active item
                  if (item != null && selectedItems.indexOf(item[0]) > -1)
                  {
                      dragger = item;
                  }
                  else
                  {
                      // the active item would change, so remove oj-draggable now
                      // note for multiple selection case see setSelectionDraggable
                      $(selectedItems[0]).removeClass("oj-draggable");
                  }   
              }
            } 
            else 
            {
                dragger = item;
            }
        }

        if (dragger != null)
        {
            dragger.attr("draggable", true);
        }
    }
};

/**
 * Unsets draggable attribute on either the item or affordance.
 * See HandleMouseUpOrPanMove method
 * @param {jQuery} target target from mouse up or touch end event
 * @private
 */
oj.ListViewDndContext.prototype._unsetDraggable = function(target)
{
    var cls, dragger;
    if (this._getDragOptions() != null || this.IsItemReOrdering())
    {
        cls = this.GetDragAffordanceClass();
        if (target.hasClass(cls))
        {
            dragger = $(target);
        }
        else
        {
            dragger = this._findItem(target);
        }

        if (dragger != null)
        {
            dragger.removeAttr("draggable");
        }
    }
};
/*********************************** Drag and drop handler ***********************************************/
/**
 * Invoke user callback function specified in a drag or drop option
 * @param {string} dndType  the dnd option type ('drag' or 'drop')
 * @param {string} callbackType  the callback type such as 'dragStart'
 * @param {Event} event  the jQuery Event object from drag and drop event
 * @param {Object} [ui]  additional properties to pass to callback function
 * @return {boolean|undefined|number} the return value from the callback function, returns -1 if there's no callback function
 * @private
 */
oj.ListViewDndContext.prototype._invokeDndCallback = function(dndType, callbackType, event, ui)
{
    var options, callback, returnValue;

    options = dndType === 'drag' ? this._getDragOptions() : this._getDropOptions();
    if (options)
    {
        // First let the callback decide if data can be accepted
        callback = options[callbackType];          
        if (callback && typeof callback == 'function')
        {
            try
            {
                if (this.listview.ojContext._IsCustomElement())
                {
                    // For custom element, pass original DOM event and ignore return value.
                    callback(event.originalEvent, ui);

                    // preventDefault is called on the original event
                    if (event.originalEvent.defaultPrevented)
                    {
                        event.preventDefault();
                    }
                }
                else
                {
                    // Hoist dataTransfer object from DOM event to jQuery event
                    event.dataTransfer = event.originalEvent.dataTransfer;
        
                    // Invoke callback function
                    returnValue = callback(event, ui);
                }
            }
            catch (e)
            {
                oj.Logger.error('Error: ' + e);
            }
        }
        else
        {
            returnValue = -1;
        }
    }
    else
    {
        returnValue = -1;
    }

    return returnValue;
};

/**
 * Sets the data and drag image on the data transfer object
 * @param {Event} event  the jQuery Event object from drag and drop event
 * @param {Array.<string>} dataTypes the datatypes specified in option
 * @param {Array.<Element>} items an array of item elements
 * @private
 */
oj.ListViewDndContext.prototype._setDragItemDataTransfer = function(event, dataTypes, items)
{
    var isStatic, i, itemDataArray = [], data, originalEvent;

    for (i=0; i<items.length; i++)
    {
        data = this.listview._getDataForItem(items[i]);
        if (data)
        {
            // for Elements we'll use the the inner HTML in data transfer
            if (data.innerHTML && data.tagName && data.tagName == "LI") // @HTMLUpdateOK
            {
                itemDataArray.push(data.innerHTML); // @HTMLUpdateOK
            }
            else
            {
                itemDataArray.push(data);
            }
        }
    }

    // Use the row data array as drag data and create drag image
    if (itemDataArray.length > 0)
    {
        this._setDragItemData(event.originalEvent, dataTypes, itemDataArray);
        this.SetDragItemImage(event.originalEvent, items);

        return {'items': itemDataArray};
    }

    return null;
};

/**
 * Set the data of the selected rows into the dataTransfer object
 * @param {Event} nativeEvent  DOM event object 
 * @param {string | Array.<string>} dataTypes  a data type or array of data types
 * @param {Array.<Object>} itemDataArray  array of row data
 * @private
 */
oj.ListViewDndContext.prototype._setDragItemData = function(nativeEvent, dataTypes, itemDataArray)
{
    var dataTransfer, jsonStr, i;

    dataTransfer = nativeEvent.dataTransfer;
    jsonStr = JSON.stringify(itemDataArray);
  
    if (typeof dataTypes == 'string')
    {
        dataTransfer.setData(dataTypes, jsonStr);
    }
    else if (dataTypes)
    {
        for (i = 0; i < dataTypes.length; i++)
        {
            dataTransfer.setData(dataTypes[i], jsonStr);
        }
    }

    // tag datatransfer so we could determine whether this is dnd within same listview
    dataTransfer.setData(this.GetDragSourceType(), this.listview.element.get(0).id);
};

/**
 * Create and sets the drag image into the dataTransfer object
 * @param {Event} nativeEvent  DOM event object 
 * @param {Array.<Element>} items array of row data
 * @protected
 */
oj.ListViewDndContext.prototype.SetDragItemImage = function(nativeEvent, items)
{
    var target, dragImage, i, minTop = Number.MAX_VALUE, offsetTop, offsetWidth, clone, container, left = 0, top = 0;

    target = nativeEvent.target;

    if (items.length > 1)
    {
        dragImage = $(document.createElement("ul"));
        dragImage.get(0).className = this.listview.element.get(0).className;
        dragImage.addClass(this.GetDragImageClass())
                 .css({"width":this.listview.element.css("width"),
                       "height":this.listview.element.css("height")});
        
        // need to figure out which is the top item
        for (i=0; i<items.length; i++)
        {
            minTop = Math.min(minTop, items[i].offsetTop);
        }

        // now construct the drag image
        for (i=0; i<items.length; i++)
        {
            offsetTop = items[i].offsetTop - minTop;
            offsetWidth = items[i].offsetWidth;
            clone = $(items[i].cloneNode(true));  
            clone.removeClass("oj-selected oj-focus oj-hover")
                 .css({"position":"absolute",
                       "top":offsetTop,
                       "width":offsetWidth});            
            dragImage.append(clone); //@HTMLUpdateOK
        }
    }
    else
    {
        // calculate offset if drag on affordance
        if ($(target).hasClass(this.GetDragAffordanceClass()))
        {
            offsetTop = 0;
            if ($.contains(items[0], target.offsetParent))
            {
                offsetTop = target.offsetTop;
            }

            left = Math.max(0, target.offsetLeft - items[0].offsetLeft) + target.offsetWidth/2;
            top = offsetTop + target.offsetHeight/2;
        }
        else
        {
            left = Math.max(0, nativeEvent.offsetX);
            top = Math.max(0, nativeEvent.offsetY);
        }

        clone = $(items[0].cloneNode(true));
        clone.removeClass("oj-selected oj-focus oj-hover")
             .addClass("oj-drag");

        dragImage = $(document.createElement("ul"));
        dragImage.get(0).className = this.listview.element.get(0).className;
        dragImage.addClass(this.GetDragImageClass())
                 .css({"width": this.GetDragImageWidth(items[0]),
                       "height":items[0].offsetHeight * 2})
                 .append(clone); //@HTMLUpdateOK
    }

    // copying class might not be sufficient since it could be on the custom element root
    if (this.listview.isCardLayout())
    {
        dragImage.addClass("oj-listview-card-layout");
    }

    $("body").append(dragImage); //@HTMLUpdateOK
    this.m_dragImage = dragImage;
    nativeEvent.dataTransfer.setDragImage(dragImage.get(0), left, top);
};

/**
 * Returns the Drag image width. Need to be overriden by navlist.
 * @return {string} width
 * @protected
 */
oj.ListViewDndContext.prototype.GetDragImageWidth = function(item)
{
    return this.listview.element.css("width");
};

/**
 * Returns the default data type. Need to be overriden by navlist
 * @return {string} the default data type
 * @protected
 */
oj.ListViewDndContext.prototype.GetDefaultDataType = function()
{
    return "text/ojlistview-items-data";
};

/**
 * Drag start event handler
 * @param {Event} event  jQuery event object
 * @return {boolean|number|undefined} a value passed back to jQuery.  Returning false will
 *         cause jQuery to call event.preventDefault and event.stopPropagation.
 *         Returning true or other values has no side effect.
 *         In the case of dragstart, returning false cancels the drag operation.
 * @private
 */
oj.ListViewDndContext.prototype._handleDragStart = function(event)
{
    var options, dataTypes, items, ui, ret;

    options = this._getDragOptions();
    if (options != null || this.IsItemReOrdering())
    {
        if (options != null)
        {
            dataTypes = options['dataTypes'];
        }
        else
        {
            dataTypes = this.GetDefaultDataType();
        }

        if ($(event.target).hasClass(this.GetDragAffordanceClass()) || this.shouldDragCurrentItem())
        {
            // if it's affordance then use event target
            items = [];
            items.push(this._findItem(event.target)[0]);
        }
        else
        {
            items = this._getSelectedItems(true);
        }

        if (items.length > 0)
        {
            this.m_dragItems = items;
            this.m_currentDragItem = $(items[0]);

            ui = this._setDragItemDataTransfer(event, dataTypes, items);
            if (ui)
            {
                ret = this._invokeDndCallback('drag', 'dragStart', event, ui);

                // dragStart is optional
                if (ret === -1)
                {
                    return;
                }
                else
                {
                    // but if it's specified, it could abort drag by returning false
                    return ret;
                }
            }
            else
            {
                // no data, abort drag
                return false;
            }
        }
    }

    // don't return anything since apps can register their own custom event handler
};

/**
 * Drag event handler
 * @param {Event} event  jQuery event object
 * @private
 */
oj.ListViewDndContext.prototype._handleDrag = function(event)
{
    // invoke callback
    return this._invokeDndCallback('drag', 'drag', event);
};

/**
 * Delete the drag image
 * @private
 */
oj.ListViewDndContext.prototype._destroyDragImage = function()
{
    if (this.m_dragImage != null)
    {
        this.m_dragImage.remove();
        this.m_dragImage = null;
    }
};

/**
 * Drag end event handler
 * @param {Event} event  jQuery event object
 * @private
 */
oj.ListViewDndContext.prototype._handleDragEnd = function(event)
{
    var i;

    // remove css class and make sure it's visible again
    if (this.m_currentDragItem != null && this.m_dragItems != null)
    {
        // for drag affordance case
        this.m_currentDragItem.find("."+this.GetDragAffordanceClass())
                              .removeAttr("draggable");

        this.m_currentDragItem.removeClass("oj-drag oj-draggable")
                              .removeAttr("draggable");

        for (i=0; i<this.m_dragItems.length; i++)
        {
            $(this.m_dragItems[i]).removeClass(this.GetDragItemClass())
                                  .css("display", "");           
        }
    }

    // remove drop target, if any
    this._cleanupDropTarget();

    // clean up drag image
    this._destroyDragImage();

    // remove style classes from selected items that were draggable
    this._unsetSelectionDraggable();

    // invoke callback
    this._invokeDndCallback('drag', 'dragEnd', event);

    // save dragItems for potential drop within same listview (reordering)
    this.m_itemsDragged = this.m_dragItems;

    // reset drag variables (except this.m_dragItems)
    this.m_dragImage = null;
    this.m_currentDragItem = null;
    this.m_dragItems = null;

    // focus should now be at document body, since blur event was skipped, do the tabindex change here
    this.listview._doBlur();
};

/**
 * Return true if the data types from dnd event match one of the values in an array
 * @param {Event} event  jQuery event object for a drag and drop event
 * @return {boolean} true if one of the types in dragDataTypes and allowedTypes matches       
 * @private
 */
oj.ListViewDndContext.prototype._matchDragDataType = function(event)
{
    var dragDataTypes, options, allowedTypes, allowedTypeArray, i;

    options = this._getDropOptions();
    if (options && options['dataTypes'])
    {
        allowedTypes = options['dataTypes'];
        allowedTypeArray = (typeof allowedTypes == 'string') ? [allowedTypes] : allowedTypes;
    
        // dragDataTypes can be either an array of strings (Chrome) or a 
        // DOMStringList (Firefox and IE).  For cross-browser compatibility, use its 
        // length and index to traverse it.
        dragDataTypes = event.originalEvent.dataTransfer.types;
        for (i = 0; i < dragDataTypes.length; i++)
        {
            if (allowedTypeArray.indexOf(dragDataTypes[i]) >= 0)
            {
                return true;
            }
        }
    }

    return false;
};

/**
 * Invoke user callback function specified in a drop option
 * @param {string} callbackType  the callback type such as 'dragStart'
 * @param {Event} event  the jQuery Event object from drag and drop event
 * @param {Object} [ui]  additional properties to pass to callback function
 * @return {boolean|undefined|number} the return value from the callback function, returns -1 if there is no callback function
 * @private
 */
oj.ListViewDndContext.prototype._invokeDropCallback = function(callbackType, event, ui)
{
    var returnValue = this._invokeDndCallback('drop', callbackType, event, ui);
    if (returnValue === undefined || returnValue === -1)
    {
        if (this._matchDragDataType(event))
        {
            event.preventDefault();
        }
    }

    return returnValue;
};

/**
 * Create drop target based on item
 * @param {jQuery} item the item to create drop target based on
 * @return {jQuery} the drop target
 * @private
 */
oj.ListViewDndContext.prototype._createDropTarget = function(item)
{
    var dropTarget;

    if (this.m_dropTarget == null)
    {
        dropTarget = $(item.get(0).cloneNode(false));
        dropTarget.addClass("oj-drop")
                  .removeClass("oj-drag oj-draggable oj-hover oj-focus oj-selected")
                  .css({"display": "block", "height": item.outerHeight(), "width": item.outerWidth()});
        
        this.m_dropTarget = dropTarget;
    }

    return this.m_dropTarget;
};

/**
 * Clean up group item drop target artifacts
 * @private
 */
oj.ListViewDndContext.prototype._cleanupGroupItem = function()
{
    // for group item
    if (this.m_currentDropItem != null && this.m_dropTargetIndex === -1)
    {
        this.m_currentDropItem.children("."+this.listview.getGroupItemStyleClass()).removeClass("oj-drop");
    }
};

/**
 * Clean up for the case of empty list
 * @private
 */
oj.ListViewDndContext.prototype._cleanupEmptyList = function()
{
    // if it's drop on an empty list
    if (this.m_currentDropItem != null && this.m_currentDropItem.hasClass(this.listview.getEmptyTextStyleClass()))
    {
        this.m_currentDropItem.removeClass("oj-drop");
        this.m_currentDropItem.get(0).textContent = this.listview._getEmptyText();        
    }
};

/**
 * Clean up drop target artifacts
 * @private
 */
oj.ListViewDndContext.prototype._cleanupDropTarget = function()
{
    if (this.m_dropTarget != null)
    {
        this.m_dropTarget.css("height", "0");
        this.m_dropTarget.remove();
        this.m_dropTarget = null;
    }

    this._cleanupEmptyList();
    this._cleanupGroupItem();
};

/**
 * Drag enter event handler
 * @param {Event} event  jQuery event object
 * @private
 */
oj.ListViewDndContext.prototype._handleDragEnter = function(event)
{
    var item, returnValue;

    item = this._findItem(event.target);
    returnValue = this._invokeDropCallback('dragEnter', event, {'item': item.get(0)});
    if (returnValue != -1)
    {
        return returnValue;
    }
};

/**
 * Sets the current drop item
 * @param {jQuery} item the item to set as current drop item
 * @private
 */
oj.ListViewDndContext.prototype._setCurrentDropItem = function(item)
{
    if (this.m_currentDropItem != null)
    {
        this.m_currentDropItem.removeClass("oj-valid-drop oj-invalid-drop");
    }

    this.m_currentDropItem = item;
    this.m_currentDropItem.addClass("oj-valid-drop");
};

/**
 * Sets the accessible text according to the drop position
 * @param {jQuery} item the drop target reference
 * @param {string} position either 'before' or 'after'
 * @private
 */
oj.ListViewDndContext.prototype._setAccInfo = function(item, position)
{
    var label, key, msg;

    // we'll see if there is a label associated with the item, if not, look for text inside the item
    label = item.attr("aria-label");
    if (label == null)
    {
        label = item.text();
    }

    key = "accessibleReorder" + position.charAt(0).toUpperCase() + position.substr(1) + "Item";
    msg = this.listview.ojContext.getTranslatedString(key, {"item": label});
    this.listview._setAccInfoText(msg);
};

/**
 * Adjust the max height of group item
 * @private
 */
oj.ListViewDndContext.prototype._adjustGroupItemStyle = function()
{
    // for touch we'll need to remove max height set
    if (this.m_maxHeightAdjusted == null && this.listview._isTouchSupport())
    {
        this.listview.element.find("ul."+this.listview.getGroupStyleClass()).each(function(index)
        {
            $(this).attr("oldMaxHeight", $(this).css("maxHeight").toString());
            $(this).css("maxHeight", 10000);
        });        

        this.m_maxHeightAdjusted = "adjusted";
    }
};

/**
 * Reverse of _adjustGroupItemStyle
 * @private
 */
oj.ListViewDndContext.prototype._restoreGroupItemStyle = function()
{
    if (this.listview._isTouchSupport())
    {
        this.listview.element.find("ul."+this.listview.getGroupStyleClass()).each(function(index)
        {
            $(this).css("maxHeight", parseInt($(this).attr("oldMaxHeight"), 10));
            $(this).removeAttr("oldMaxHeight");
        });        
    }

    this.m_maxHeightAdjusted = null;
};

/**
 * Drag over event handler
 * @param {Event} event  jQuery event object
 * @private
 */
oj.ListViewDndContext.prototype._handleDragOver = function(event)
{
    var item, dropTarget, i, returnValue, index, emptyItem;

    // do any neccessary group item
    this._adjustGroupItemStyle();

    // checks whether this is initially dragging over the drag item 
    if (this.m_dragItems != null && $(this.m_dragItems[0]).css("display") != "none")
    {
        // take out the current drag item and create drop target
        item = $(this.m_dragItems[0]);

        // this will check for matching data types
        returnValue = this._invokeDropCallback('dragOver', event, {'item': item.get(0)});

        // note drop is allowed in the case where reordering is enabled, but only if there's no dragOver callback
        // to prevent the drop
        if ((returnValue === -1 && this.IsItemReOrdering()) || returnValue === false || event.isDefaultPrevented())
        {
            dropTarget = this._createDropTarget(item);

            for (i=0; i<this.m_dragItems.length; i++)
        {
            // we have to override inline style instead of doing it inside style class since
            // custom style class could override it
            $(this.m_dragItems[i]).addClass(this.GetDragItemClass())
                                  .css("display", "none");           
        }

            dropTarget.insertBefore(item); //@HTMLUpdateOK
            this.m_dropTargetIndex = dropTarget.index();
        }
    }
    else
    {
        // if drag over callback allows drop or if it's item reordering
        item = this._findItem(event.target);
        if (item != null && item.length > 0)
        {
            // this will check for matching data types
            returnValue = this._invokeDropCallback('dragOver', event, {'item': item.get(0)});

            // note drop is allowed in the case where reordering is enabled, but only if there's no dragOver callback
            // to prevent the drop
            if ((returnValue === -1 && this.IsItemReOrdering()) || returnValue === false || event.isDefaultPrevented())
            {
                // if it is non-group item
                if (item.hasClass(this.listview.getItemStyleClass()))
                {
                    this._cleanupGroupItem();

                    // if a drop target has not been created for this item
                    if (!item.hasClass("oj-drop"))
                    {
                        dropTarget = this._createDropTarget(item);

                        index = item.index();
                        if (this.m_dropTargetIndex == null || this.m_dropTargetIndex < index)
                        {
                            dropTarget.insertAfter(item); //@HTMLUpdateOK
                            this.m_dropPosition = "after";
                        }
                        else
                        {
                            dropTarget.insertBefore(item); //@HTMLUpdateOK
                            this.m_dropPosition = "before";
                        }

                        this._setAccInfo(item, this.m_dropPosition);
                        this._setCurrentDropItem(item);
                        this.m_dropTargetIndex = dropTarget.index();
                    }
                }
                else
                {
                    // there won't be a drop target placeholder, remove any previous ones
                    this._cleanupDropTarget();

                    // for group item
                    item.children("."+this.listview.getGroupItemStyleClass()).addClass("oj-drop");
                    this._setCurrentDropItem(item);
                    this.m_dropTargetIndex = -1;
                    this.m_dropPosition = "inside";
                    this._setAccInfo(item, this.m_dropPosition);
                }
                event.preventDefault();
            }
            else
            { 
                // don't remove the drop target if we have inside the empty area of the list of a group item
                // this happens mostly in card layout view
                if (!$(event.target).hasClass(this.listview.getGroupStyleClass()))
                {
                    // not a valid drop target
                    item.addClass("oj-invalid-drop");

                    // remove any drop target placeholder
                    this._cleanupDropTarget();
                }
            }
        }
        else
        {
            // drop on an empty list
            emptyItem = this.listview.element.children("."+this.listview.getEmptyTextStyleClass());
            if (emptyItem != null && emptyItem.length > 0)
            {
                emptyItem.addClass("oj-drop");
                emptyItem.get(0).textContent = "";

                this._setCurrentDropItem(emptyItem);
                event.preventDefault();
            }
        }
    }

    return returnValue;
};

/**
 * Return true if the mouse/touch point of a dnd event is in an element.
 * This is identical to _isDndEventInElement in TableDndContext.js
 * @param {Event} event  jQuery event object
 * @param {EventTarget} element  DOM element
 * @return {boolean} whether mouse/touch point is in an element
 */
oj.ListViewDndContext.prototype._isDndEventInElement = function(event, element)
{
    var rect, domEvent;

    rect = element.getBoundingClientRect();
    domEvent = event.originalEvent;
  
    // clientX and clientY are only available on the original DOM event
    return (domEvent.clientX >= rect.left && domEvent.clientX < rect.right &&
            domEvent.clientY >= rect.top && domEvent.clientY < rect.bottom);
};

/**
 * Drag leave event handler
 * @param {Event} event  jQuery event object
 * @private
 */
oj.ListViewDndContext.prototype._handleDragLeave = function(event)
{
    var item, returnValue;

    if (this.m_currentDropItem == null)
    {
        return;
    }

    item = this._findItem(event.target);
    if (item != null && item.length > 0)
    {
        item.removeClass("oj-valid-drop oj-invalid-drop");

        returnValue = this._invokeDropCallback('dragLeave', event, {'item': item.get(0)});

        // Remove the drop target indicator if we are no longer in listview since
        // this may be the last dnd event we get.
        if (!this._isDndEventInElement(event, event.currentTarget) && item.hasClass('oj-drop'))
        {
            this._cleanupDropTarget();
            this._restoreGroupItemStyle();
        }
    }
    else
    {
        // empty list case
        if (!this._isDndEventInElement(event, event.currentTarget))
        {
            this._cleanupEmptyList();
        }
    }

    if (returnValue != -1)
    {
        return returnValue;
    }
};

/**
 * Drop event handler
 * @param {Event} event  jQuery event object
 * @private
 */
oj.ListViewDndContext.prototype._handleDrop = function(event)
{
    var source, ui, returnValue, item, key, dataSource;

    if (this.m_currentDropItem == null)
    {
        return;
    }

    source = event.originalEvent.dataTransfer.getData(this.GetDragSourceType());

    // invoke callback
    if (this.m_currentDropItem.hasClass(this.listview.getEmptyTextStyleClass()))
    {
        ui = {};
    }
    else
    {
        ui = {'item': this.m_currentDropItem.get(0), 'position': this.m_dropPosition};
    }

    // add the reorder param
    if (this.IsItemReOrdering() && source === this.listview.element.get(0).id)
    {
        ui['reorder'] = true;
    }
    else
    {
        ui['reorder'] = false;
    }

    // cleanup artifacts before firing events
    if (this.m_currentDropItem != null)
    {
        this.m_currentDropItem.removeClass("oj-valid-drop");
    }
    this._cleanupDropTarget();
    this._restoreGroupItemStyle();
    this._destroyDragImage();

    // fire drop event
    returnValue = this._invokeDropCallback('drop', event, ui);

    // fire reorder event
    if (ui['reorder'])
    {
        // the order in which dragEnd and drop event is called is inconsistent, hence the check
        //Always we should check for this.m_dragItems as it will be null only if dragend already invoked 
        //and items are copied to this.m_itemsDragged.
        ui['items'] = this.m_dragItems == null ? this.m_itemsDragged : this.m_dragItems;
        
        this.listview.Trigger("reorder", event, this.CreateReorderPayload(ui['items'], ui['position'], ui['item']));

        // the drop should be complete regardless the value of callback
        event.preventDefault();
    }

    // reset drop variables
    this.m_currentDropItem = null;
    this.m_dropTargetIndex = -1;
    this.m_dropPosition = null;
    this.m_itemsDragged = null;

    if (returnValue === -1)
    {
        return;
    }
    else  
    { 
        return returnValue;
    }
};

/**
 * Returns payload object for reorder event. Navlist overrides this.
 * @param {Array} items, items to be moved. 
 * @param {string} position, the drop position relative to the reference item.
 * @param {Element} reference, the item where the moved items are drop on.
 * @returns {Object} payload object
 * @protected
 */
oj.ListViewDndContext.prototype.CreateReorderPayload = function (items, position, reference) {
  return {
    'items': items,
    'position': position,
    'reference': reference
  };
};

/*********************************** Context menu ***********************************************/
/**
 * Prepares the context menu before it is opened.  Invoked by notifyContextMenuGesture.
 * @param {Element} contextMenu the context menu root element
 */
oj.ListViewDndContext.prototype.prepareContextMenu = function(contextMenu)
{
    var self = this, menuContainer, listItems, menuItemsSet;

    // only add context menu if item reordering is enabled
    if (!this.IsItemReOrdering())
    {
        return;
    }

    menuContainer = $(contextMenu);
    if (this.m_contextMenu != contextMenu)
    {
        this.m_contextMenu = contextMenu;

        if (contextMenu.tagName === "OJ-MENU")
        {
            contextMenu.addEventListener("ojBeforeOpen", this._handleContextMenuBeforeOpen.bind(this));
            contextMenu.addEventListener("ojAction", this._handleContextMenuSelect.bind(this));
        }
        else
        {
            menuContainer.on("ojbeforeopen", this._handleContextMenuBeforeOpen.bind(this));
            menuContainer.on("ojselect", this._handleContextMenuSelect.bind(this));
        }
    }

    menuItemsSet = this._getCommands(contextMenu, function(menuItem, command)
    {
        var newListItem = self._buildContextMenuItem(command, menuItem.tagName);
        if (menuItem.tagName === "OJ-OPTION")
        {
            menuItem.innerHTML = newListItem.get(0).innerHTML; // @HTMLUpdateOK
            $(menuItem).attr('data-oj-command', newListItem.attr('data-oj-command'));
        }
        else
        {
            newListItem.get(0).className = $(menuItem).get(0).className;
            $(menuItem).replaceWith(newListItem); //@HTMLUpdateOK                        
        }        
    });

    // this keeps track of which menu items were generated by the listview dynamically
    // this way on a refresh we know to recreate them in case there was a locale or
    // translations change
    this.m_menuItemsSet = menuItemsSet;

    if (menuItemsSet.length > 0)
    {
        if (menuContainer.data('oj-ojMenu'))
        {
            if (contextMenu.tagName === "OJ-MENU")
            {
                contextMenu.refresh();
            }
            else
            {
                $(contextMenu).ojMenu('refresh');
            }
        }
    }
};

/**
 * Returns selector to extract only the context menu items with valid commands. 
 * Navlist has remove menu item which needs to be ignored by listview. 
 * @returns {string} jquery selector string. 
 */
oj.ListViewDndContext.prototype._getDndContextMenuItemSelector = function()
{
    var self = this, query = '', commands = ['cut', 'copy', 'paste', 'paste-before', 'paste-after', 'pasteBefore', 'pasteAfter'];
    
    if (!this.m_dndMenuItemSelector) 
    {
        commands.forEach(function (command, index) 
        {
            query += '[data-oj-command=' + self.GetCommandPrefix() + command + '],';
            query += '[data-oj-command=' + command + ']';
            if (index < commands.length - 1) 
            {
                query += ',';
            }
        });
        this.m_dndMenuItemSelector = query;
    }
    return this.m_dndMenuItemSelector;
};

/**
 * Retrieves a list of commands from the context menu.
 * @param {Element} contextMenu the context menu element
 * @param {function(Element, string)=} callback the function to invoke on each menu item found
 * @return {Array} a list of commands
 * @private
 */
oj.ListViewDndContext.prototype._getCommands = function(contextMenu, callback)
{
    var capabilities, listItems, self;
    self = this;
    capabilities = [];
    listItems = $(contextMenu).find(this._getDndContextMenuItemSelector());
    listItems.each(function() {
        var command, anchor, newListItem;
        anchor = $(this).children('a');
        if (anchor.length === 0)
        {
            if ($(this).attr('data-oj-command').indexOf(self.GetCommandPrefix()) == 0)
            {
                command = $(this).attr('data-oj-command').substring(self.GetCommandPrefix().length);
                if (callback)
                {
                    callback(this, command);
                }
            }
        }
        else
        {
            command = $(this).attr('data-oj-command');

            // mapping for paste
            if (command == oj.ListViewDndContext.PASTE_BEFORE_COMMAND)
            {
                command = "paste-before";
            }
            else if (command == oj.ListViewDndContext.PASTE_AFTER_COMMAND)
            {
                command = "paste-after";
            }
        }

        if (command)
        {
            capabilities.push(command);
        }
    });

    return capabilities;
};

/**
 * Builds a menu for a command, takes care of submenus where appropriate
 * @param {string} command the command that the listview should build a menu item for
 * @param {string} tagName to use to create the menu item
 * @private
 */
oj.ListViewDndContext.prototype._buildContextMenuItem = function(command, tagName)
{
    if (command === 'paste-before')
    {
        return this._buildContextMenuListItem(oj.ListViewDndContext.PASTE_BEFORE_COMMAND, tagName);
    }
    else if (command === 'paste-after')
    {
        return this._buildContextMenuListItem(oj.ListViewDndContext.PASTE_AFTER_COMMAND, tagName);
    }
    else
    {
        return this._buildContextMenuListItem(command, tagName);
    }
};

/**
 * Builds a context menu list item from a command
 * @param {string} command the string to look up command value for as well as translation
 * @param {string} tagName to use to create the menu item
 * @return {Object} a jQuery object with HTML containing a list item
 * @private
 */
oj.ListViewDndContext.prototype._buildContextMenuListItem = function(command, tagName)
{
    var listItem = $(document.createElement(tagName));
    listItem.attr('data-oj-command', command);
    listItem.append(this._buildContextMenuLabel(command, tagName === "OJ-OPTION")); //@HTMLUpdateOK
    return listItem;
};

/**
 * Builds a context menu label by looking up command translation
 * @param {string} command the string to look up translation for
 * @param {boolean=} useOjOption whether oj-option tag should be used
 * @return {jQuery|string} a jQuery object with HTML containing a label
 * @private
 */
oj.ListViewDndContext.prototype._buildContextMenuLabel = function(command, useOjOption)
{
    // convert to the translation key convention
    var key = 'label' + command.charAt(0).toUpperCase() + command.slice(1);
    var textNode = document.createTextNode(this.listview.ojContext.getTranslatedString(key));
    
    // for custom elements, no <a> tag is required
    if (useOjOption)
        return textNode;
    else
        return $('<a href="#"></a>').append(textNode); //@HTMLUpdateOK
};

/**
 * Handles cut action
 * @param {Event} event jQuery event
 * @private
 */
oj.ListViewDndContext.prototype._handleCut = function(event)
{
    var items;

    // first restore style of any previously cut items
    if (this.m_clipboard != null)
    {
        $(this.m_clipboard).removeClass(this.GetCutStyleClass());
    }
    
    items = this.GetCutItems(event);
    //focus should be moved back to listview after context menu is closed
    this.listview.ojContext.element.focus();
    $(items).addClass(this.GetCutStyleClass());
    this.m_clipboard = items;
    
    // fire cut event
    this.listview.Trigger('cut', event, {'items': items});
};

/**
 * Returns cut items. Navlist overrides this.
 * @param {Event} event jQuery event
 * @returns {Array} array of items to be moved
 */
oj.ListViewDndContext.prototype.GetCutItems = function(event)
{
    return this._getSelectedItems();
};

/**
 * Handles copy action
 * @param {Event} event jQuery event
 * @private
 */
oj.ListViewDndContext.prototype._handleCopy = function(event)
{
    var items;

    // first restore style of any previously cut items
    if (this.m_clipboard != null)
    {
        $(this.m_clipboard).removeClass(this.GetCutStyleClass());
    }

      items = this._getSelectedItems();
    this.m_clipboard = items;

    // fire cut event
    this.listview.Trigger('copy', event, {'items': items});
};

/**
 * Handles paste action
 * @param {Event} event jQuery event
 * @param {jQuery} item the reference item of the paste
 * @param {string} position either before/after/inside
 * @private
 */
oj.ListViewDndContext.prototype._handlePaste = function(event, item, position)
{
    // fire paste event
    this.listview.Trigger("paste", event, {'item': item.get(0)});
    
    // restore cut item style
    $(this.m_clipboard).removeClass(this.GetCutStyleClass());

    // fire reorder event
    this.listview.Trigger("reorder", event, this.CreateReorderPayload(this.m_clipboard, position, /** @type {Element} */ (item.get(0))));

    // clear the clipboard
    this.m_clipboard = null;
};

/**
 * Select handler for context menu items
 * @param {Event} event jQuery event object
 * @param {Object=} ui additional info from event
 * @private
 */
oj.ListViewDndContext.prototype._handleContextMenuSelect = function(event, ui)
{
    var item;

    // should never happen
    if (this.m_contextMenuItem == null)
    {
        return;
    }

    item = ui ? ui.item : $(event.target);
    switch(item.attr("data-oj-command")) 
    {
        case oj.ListViewDndContext.CUT_COMMAND:
            this._handleCut(event);
            break;
        case oj.ListViewDndContext.COPY_COMMAND:
            this._handleCopy(event);
            break;
        case oj.ListViewDndContext.PASTE_COMMAND:
            var inside = true;
        case oj.ListViewDndContext.PASTE_BEFORE_COMMAND:
            var before = true;
        case oj.ListViewDndContext.PASTE_AFTER_COMMAND:
            var position = "after";
            if (inside)
            {
                position = "inside";
            }
            else if (before)
            {
                position = "before";
            }

            this._handlePaste(event, this.m_contextMenuItem, position);

            // cleanup
            this.m_contextMenuItem = null;

            break;
    }
};

/**
 * Append or show menu item based on command
 * @param {jQuery} menuContainer the menu container
 * @param {string} command the command of the menu item
 * @private
 */
oj.ListViewDndContext.prototype._appendToMenuContainer = function(menuContainer, command)
{
    // whether default context menu is used
    if (this.m_menuItemsSet != null)
    {
        if (command == "paste-before")
        {
            command = oj.ListViewDndContext.PASTE_BEFORE_COMMAND;
        }
        else if (command == "paste-after")
        {
            command = oj.ListViewDndContext.PASTE_AFTER_COMMAND;
        }

        // show the menu item
        menuContainer.find("[data-oj-command='"+command+"']")
                     .removeClass("oj-disabled");
    }
}

/**
 * Before open handler so that ListView can customize content of context menu based on item
 * @param {Event} event jQuery event object
 * @param {Object=} ui ui object
 * @private
 */
oj.ListViewDndContext.prototype._handleContextMenuBeforeOpen = function(event, ui)
{
    var menuContainer, item;

    menuContainer = $(event.target);

    // disable all menu items first, needs to be done even if there's no default menu items since
    // there could be one from before refresh
    menuContainer.find(this._getDndContextMenuItemSelector())
                 .addClass("oj-disabled");

    item = ui ? ui['openOptions']['launcher'] : event['detail']['openOptions']['launcher'];
    if (item == null || this.m_menuItemsSet == null || this.m_menuItemsSet.length == 0)
    {
        // refresh to take effect
        if (menuContainer.get(0).tagName != "OJ-MENU")
        {
            menuContainer.ojMenu("refresh");
        }
        return;
    }

    // now add menu items
    if (item.children().first().hasClass(this.listview.getGroupItemStyleClass()))
    {
        // check if cut action was performed
        if (this.m_clipboard != null)
        {
            this._appendToMenuContainer(menuContainer, "paste");
        }
    }
    else
    {
        this._appendToMenuContainer(menuContainer, "cut");
        this._appendToMenuContainer(menuContainer, "copy");
        // check if cut action was performed
        if (this.m_clipboard != null)
        {
            this._appendToMenuContainer(menuContainer, "paste-before");
            this._appendToMenuContainer(menuContainer, "paste-after");
        }
    }

    // refresh to take effect
    if (menuContainer.get(0).tagName != "OJ-MENU")
    {
        menuContainer.ojMenu("refresh");
    }

    this.m_contextMenuItem = item;
};

/**
 * Handles key down event
 * @param {Event} event the keydown event
 * @return {boolean} true if key event is handled, false otherwise
 * @private
 */
oj.ListViewDndContext.prototype.HandleKeyDown = function(event)
{
    var keyCode, contextMenu, commands, active, position;

    if (event.ctrlKey || event.metaKey)
    {
        keyCode = event.keyCode;
        // quickly short circuit it if it's not one of the supported keys
        if (keyCode === oj.ListViewDndContext.X_KEY || keyCode === oj.ListViewDndContext.C_KEY || oj.ListViewDndContext.V_KEY)
        {
            // only if item reordering is enabled
            if (!this.IsItemReOrdering())
            {
                return false;
            }

            // capabilities depends on what's specified in context menu
            contextMenu = this.listview.ojContext._GetContextMenu();
            if (contextMenu == null)
            {
                return false;
            }

            // no clipboard commands found
            commands = this._getCommands(contextMenu);
            if (commands.length == 0)
            {
                return false;
            }

            if (keyCode === oj.ListViewDndContext.X_KEY && commands.indexOf("cut") > -1)
            {
                this._handleCut(event);
                return true;
            }
            else if (keyCode === oj.ListViewDndContext.C_KEY && commands.indexOf("copy") > -1)
            {
                this._handleCopy(event);
                return true;
            }
            else if (keyCode === oj.ListViewDndContext.V_KEY)
            {
                if (this.m_clipboard != null)
                {
                    active = $(this._getActiveItem());
                    if (active.children().first().hasClass(this.listview.getGroupItemStyleClass()))
                    {
                        if (commands.indexOf("paste") > -1)
                        {
                            position = "inside";
                        }
                    }
                    else
                    {
                        if (commands.indexOf("paste-before") > -1)
                        {  
                            position = "before";
                        }
                        else if (commands.indexOf("paste-after") > -1)
                        {  
                            position = "after";
                        }
                    }
     
                    if (position != null)
                    {   
                        this._handlePaste(event, active, position);
                        return true;
                    }
                }
            }
        }
    }

    return false;
};

/**
 * Return true to drag current item. Need to be overriden by navlist.
 * @return {boolean} true to drag current item.
 * @private
 */
oj.ListViewDndContext.prototype.shouldDragCurrentItem = function()
{
    return false;
};

/**
 * Returns drag source type. Need to be overriden by navlist.
 * @return {string} drag source type.
 * @private
 */
oj.ListViewDndContext.prototype.GetDragSourceType = function()
{
    return "text/ojlistview-dragsource-id";
};
/**
 * Copyright (c) 2014, 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*
** Important:
** - This file is designed to be shared verbatim among the ADFui products.
** - Do not add framework-specific dependencies in this file (it must be self-contained).
** - Do not change this file without testing it in other ADFui products (ADF Faces, JET, etc.).
*/

/**
 * Constructor.
 * @param {Object} elem DOM element associated with the masonry layout
 * @param {boolean} rtl True if the reading direction is right-to-left, otherwise false
 * @param {boolean} automationEnabled True if automation mode is enabled, otherwise false
 * @param {Object} selectors Map of properties for the following selector information:
 *  - tiles: Selector for child tiles.
 * @param {Object} styles Map of properties for the following style classes:
 *  - transitionShowFromStyleClass: Initial state for transition to show a child tile,
 * @param {Object} callbackInfo Map of properties for the following callback functions:
 *  - addStyleClassName: Add a style class to a DOM element,
 *  - removeStyleClassName: Remove a style class from a DOM element,
 *  - getSizeStyleClassName: Get the style class for the tile size,
 *  - getTileSpan: Get the tile span,
 *  - showTileOnEndFunc: Called after a tile is shown,
 *  - hideTileOnEndFunc: Called after a tile is hideden,
 *  - startAnimationFunc: Called to animate a tile,
 *  - layoutOnEndFunc: Called after tiles are positioned,
 *  - layoutCycleOnStartFunc: Called before entire layout cycle begins,
 *  - layoutCycleOnEndFunc: Called after entire layout cycle is done,
 *  - sortTilesOriginalOrderFunc: Sort tile DOM elements into their original order,
 *  - subtreeAttached: Called after a tile is attached to the DOM,
 *  - subtreeDetached: Called after a tile is detached from the DOM,
 *  - addBusyState: Add a busy state to the busy context.
 * @class MasonryLayoutCommon
 * @memberof MasonryLayoutCommon
 * @protected
 * @constructor
 * @ignore
 * @ojtsignore
 */
function MasonryLayoutCommon(
  elem, 
  rtl, 
  automationEnabled,
  selectors, 
  styles, 
  callbackInfo)
{
  //this agent initialization concept copied from adf.shared.impl.animationUtils.transition()
  var mlcClass = MasonryLayoutCommon;
  if (mlcClass._agentTypeAndVersion == null)
  {
    // Do a 1-time agent initialization
    mlcClass._agentTypeAndVersion = mlcClass._getAgentTypeAndVersion(navigator.userAgent);
  }
  
  this._elem = elem;
  this._rtl = rtl;
  this._automationEnabled = automationEnabled;
  if (selectors)
  {
    if (selectors.tiles)
      this._tilesSelector = selectors.tiles;
  }
  if (styles)
  {
    if (styles.transitionShowFromStyleClass)
      this._transitionShowFromStyleClass = styles.transitionShowFromStyleClass;
  }
  if (callbackInfo)
  {
    //need to check for existence of properties on callbackInfo before assigning 
    //them to vars because advanced closure compiler will otherwise complain if 
    //they're not defined
    if (callbackInfo.addStyleClassName)
      this._addStyleClassNameFunc = callbackInfo.addStyleClassName;
    if (callbackInfo.removeStyleClassName)
      this._removeStyleClassNameFunc = callbackInfo.removeStyleClassName;
    if (callbackInfo.getSizeStyleClassName)
      this._getSizeStyleClassNameFunc = callbackInfo.getSizeStyleClassName;
    if (callbackInfo.getTileSpan)
      this._getTileSpanFunc = callbackInfo.getTileSpan;
    if (callbackInfo.showTileOnEndFunc)
      this._showTileOnEndFunc = callbackInfo.showTileOnEndFunc;
    if (callbackInfo.hideTileOnEndFunc)
      this._hideTileOnEndFunc = callbackInfo.hideTileOnEndFunc;
    if (callbackInfo.startAnimationFunc)
      this._startAnimationFunc = callbackInfo.startAnimationFunc;
    if (callbackInfo.layoutOnEndFunc)
      this._layoutOnEndFunc = callbackInfo.layoutOnEndFunc;
    //FIX : need to know when layout cycle is starting in order
    //to preserve focus across layout DOM order changes
    if (callbackInfo.layoutCycleOnStartFunc)
      this._layoutCycleOnStartFunc = callbackInfo.layoutCycleOnStartFunc;
    if (callbackInfo.layoutCycleOnEndFunc)
      this._layoutCycleOnEndFunc = callbackInfo.layoutCycleOnEndFunc;
    //FIX : need to be able to sort the tiles into their original 
    //DOM order before running layout
    if (callbackInfo.sortTilesOriginalOrderFunc)
      this._sortTilesOriginalOrderFunc = callbackInfo.sortTilesOriginalOrderFunc;
    if (callbackInfo.subtreeAttached)
      this._subtreeAttachedFunc = callbackInfo.subtreeAttached;
    if (callbackInfo.subtreeDetached)
      this._subtreeDetachedFunc = callbackInfo.subtreeDetached;
    if (callbackInfo.addBusyState)
      this._addBusyStateFunc = callbackInfo.addBusyState;
  }
  
  //create a non-absolutely positioned div to define the size of the 
  //infolet layout, because the absolutely positioned infolets are not 
  //part of the flow and will not define the size of the component div
  var sizeDivWrapper = document.createElement("div");
  var style = sizeDivWrapper.style;
  style.display = "inline-block";
  style.overflow = "hidden";
  style.visibility = "hidden";
  var sizeDiv = document.createElement("div");
  style = sizeDiv.style;
  style.display = "inline-block";
  sizeDivWrapper.appendChild(sizeDiv); // @HtmlUpdateOK
  elem.insertBefore(sizeDivWrapper, elem.firstChild); // @HTMLUpdateOK
  this._sizeDivWrapper = sizeDivWrapper;
  this._sizeDiv = sizeDiv;

  var self = this;
  this._hideTilesFunc = function() {self._hideTiles();};
  this._resolveBusyStateFunc = function() {self._resolveBusyState();};
};

/**
 * Setup the masonry layout.
 * @param {boolean} init True for initialization, false for refresh
 * @param {boolean} reorder True for reordering, false otherwise
 * @return {boolean} True if layout changed, false if not
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.setup = function(init, reorder)
{
  var ret = false;
  if (init)
  {
    //layout without animating
    ret = this._layout() ? true : false;
    //FIX : reorder the tile DOM elements to match the visual layout order
    this._reorderTilesForLayout();
  }
  else
  {
    //FIX : add busy state before animating a layout change
    if (this._addBusyStateFunc)
      this._busyStateResolveFunc = this._addBusyStateFunc("performing layout (setup)");
    
    //FIX : notify the peer that a layout cycle is starting so that
    //it can save state if it wants
    if (this._layoutCycleOnStartFunc)
      this._layoutCycleOnStartFunc();
    
    //if this is a refresh, add the transition class BEFORE doing the next layout
    this._transitionStart(reorder);
    ret = this._transitionLayout();
  }
  return ret;
};

/**
 * Destroy the masonry layout.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.destroy = function()
{
  //FIX : resolve busy state when destroying
  this._resolveBusyState();
  
  var elem = this._elem;
  
  //remove layout positions from the children
  var children = this._getTileChildren();
  for (var i = 0; i < children.length; i++)
  {
    var child = children[i];
    var style = child.style;
    if (this._rtl)
      style.right = "";
    else
      style.left = "";
    style.top = "";
  }
  
  elem.removeChild(this._sizeDivWrapper);
  this._sizeDivWrapper = null;
  this._sizeDiv = null;
  
  this._hideTilesFunc = null;
  this._resolveBusyStateFunc = null;
  
  this._arMovedInfolets = null;
  this._arInfoletsToResize = null;
  this._arInfoletsToShow = null;
  this._arInfoletsToHide = null;
  this._arFireHideOnEnd = null;
  
  this._elem = null;
  this._addStyleClassNameFunc = null;
  this._removeStyleClassNameFunc = null;
  this._getSizeStyleClassNameFunc = null;
  this._getTileSpanFunc = null;
  this._showTileOnEndFunc = null;
  this._hideTileOnEndFunc = null;
  this._startAnimationFunc = null;
  this._layoutOnEndFunc = null;
  this._layoutCycleOnStartFunc = null;
  this._layoutCycleOnEndFunc = null;
  this._sortTilesOriginalOrderFunc = null;
  this._subtreeAttachedFunc = null;
  this._subtreeDetachedFunc = null;
  this._addBusyStateFunc = null;
};

/**
 * Resize a child tile.
 * @param {string} selector Selector for the tile to resize
 * @param {string} sizeStyleClass New size style class 
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.resizeTile = function(selector, sizeStyleClass)
{
  var elem = this._elem;
  var infolet = elem.querySelector(selector);
  if (infolet)
  {
    //add the transition class immediately, but defer the layout call using a 
    //timeout so that the app can set new sizes on the children before the
    //layout call happens so that the size changes will be animated along with
    //the relayout
    if (!this._arInfoletsToResize)
      this._arInfoletsToResize = [];
    var arInfoletsToResize = this._arInfoletsToResize;
    arInfoletsToResize.push(infolet);
    arInfoletsToResize.push(sizeStyleClass);
    this._resizingInfolet = true;
    
    this._queueRelayout();
  }
};

/**
 * Insert a tile into the masonryLayout.
 * @param {Object} tileDomElem Tile to insert
 * @param {number} index Index at which to insert
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.insertTileDomElem = function(tileDomElem, index)
{
  var arChildren = this._getTileChildren();
  //FIX : need to sort the tiles into their original DOM order 
  //before inserting because the index is relative to the original DOM order,
  //not the current layout order
  if (this._sortTilesOriginalOrderFunc)
    this._sortTilesOriginalOrderFunc(arChildren);
  var currChildAtIndex = null;
  if (index >= 0 && index < arChildren.length)
    currChildAtIndex = arChildren[index];
  var elem = this._elem;
  elem.insertBefore(tileDomElem, currChildAtIndex); // @HTMLUpdateOK
  
  this._queueRelayout();
};

/**
 * Show a hidden tile.
 * @param {string} selector Selector for the tile to show
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.showTile = function(selector)
{
  var elem = this._elem;
  var infolet = elem.querySelector(selector);
  if (infolet)
  {
    if (!this._arInfoletsToShow)
      this._arInfoletsToShow = [];
    var arInfoletsToShow = this._arInfoletsToShow;
    arInfoletsToShow.push(infolet);
    this._showingInfolets = true;
    
    //don't queue another layout if we're already doing layout and in a phase
    //before the SHOW phase
    var mlcClass = MasonryLayoutCommon;
    if (this._layoutPhase !== mlcClass._PHASE_HIDE &&
        this._layoutPhase !== mlcClass._PHASE_LAYOUT)
    {
      this._queueRelayout();
    }
    else
    {
      //if not queueing a relayout, need to explicitly set this flag to false
      //now so that future component resizes will be processed in resizeNotify()
      this._showingInfolets = false;
    }
  }
};

/**
 * Hide a tile.
 * @param {string} selector Selector for the tile to hide
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.hideTile = function(selector)
{
  var elem = this._elem;
  var infolet = elem.querySelector(selector);
  if (infolet)
  {
    if (!this._arInfoletsToHide)
      this._arInfoletsToHide = [];
    var arInfoletsToHide = this._arInfoletsToHide;
    arInfoletsToHide.push(infolet);
    this._hidingInfolets = true;
    
    this._queueRelayout();
  }
};

/**
 * Notify the MasonryLayoutCommon that the masonryLayout component has been resized.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.resizeNotify = function()
{
  //don't respond to resize events when we're already doing layout
  if (!this._resizingInfolet && 
      !this._hidingInfolets && 
      !this._showingInfolets)
  {
    //FIX : add busy state before animating a layout change
    if (this._addBusyStateFunc)
      this._busyStateResolveFunc = this._addBusyStateFunc("performing layout (resizeNotify)");
    
    //FIX : notify the peer that a layout cycle is starting so that
    //it can save state if it wants
    if (this._layoutCycleOnStartFunc)
      this._layoutCycleOnStartFunc();
    
    this._transitionStart(false);
    this._transitionLayout();
  }
};

/**
 * Get the size of a unit cell.
 * @return {Object} object specifying unit cell size with properties w and h
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.getCellSize = function()
{
  return this._cellSize;
};

/**
 * Determine whether animation is enabled.
 * @return {boolean} true if animation is enabled, false if not
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.isAnimationEnabled = function()
{
  if (this._temporarilyDisableAnimation)
    return false;
  
  if (!this._cachedAnimationEnabled)
  {
    if (this._automationEnabled)
    {
      this._animationEnabled = false;
    }
    else
    {
      var mlcClass = MasonryLayoutCommon;
      var agentType = mlcClass._agentTypeAndVersion[0];
      var agentVersion = mlcClass._agentTypeAndVersion[1];
    
      //check whether the agent supports CSS transitions
      this._animationEnabled = mlcClass._isMinimumAgentMet(agentType, agentVersion, 
                                                           ["gecko", 16, 
                                                            "trident", 6, 
                                                            "webkit", 533.1]);
    }
    this._cachedAnimationEnabled = true;
  }
  return this._animationEnabled;
};

/**
 * Determine whether a layout cycle is currently underway.
 * @return {boolean} true if in a layout cycle, false otherwise
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.isInLayoutCycle = function()
{
  //FIX : when drag and drop reordering, the layoutPhase doesn't
  //get set because we're only running layout itself, so also check whether
  //we're moving tiles
  return (this._layoutPhase != null || 
          (this._arMovedInfolets != null && this._arMovedInfolets.length > 0));
};

/**
 * Finish the current layout cycle, aborting and/or skipping any animations.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 */
MasonryLayoutCommon.prototype.finishLayoutCycle = function()
{
  //FIX : when we get multiple masonry events in quick succession, we need 
  //to abort and skip animation due to processing the previous event in order to 
  //start processing the latest event immediately
  this._temporarilyDisableAnimation = true;
  
  //remove transition classes for animation
  this._removeStyleClassFromTiles(this._transitionShowFromStyleClass);
  
  //remove transition listeners
  var mlcClass = MasonryLayoutCommon;
  
  var tileChildren = this._getTileChildren();
  for (var i = 0; i < tileChildren.length; i++)
  {
    var child = tileChildren[i];
    if (child._afrCellSize)
      delete child._afrCellSize;    
  }
  
  //now that we've stopped and blocked animation, continue processing the layout cycle
  //based on the phase we're currently in
  if (this._hideTilesInternalTimeout)
  {
    clearTimeout(this._hideTilesInternalTimeout);
    this._hideTilesInternalTimeout = null;
    
    this._handleHideTransitionEnd(null);
  }
  else if (this._hideTilesAnimationStarted)
  {
    this._hideTilesAnimationStarted = false;
    this._handleHideTransitionEnd(null);
  }
  else if (this._showTilesTimeout)
  {
    clearTimeout(this._showTilesTimeout);
    this._showTilesTimeout = null;
    
    this._showTiles();
  }
  //FIX : when drag and drop reordering, the layoutPhase doesn't
  //get set because we're only running layout itself, so also check whether
  //we're moving tiles
  else if (this._layoutPhase === mlcClass._PHASE_LAYOUT ||
           (this._arMovedInfolets != null && this._arMovedInfolets.length > 0))
  {
    this._handleTransitionEnd(null);
  }
  else if (this._layoutPhase === mlcClass._PHASE_SHOW)
  {
    this._handleShowTransitionEnd(null);
  }
  
  //reenable animation
  this._temporarilyDisableAnimation = false;
};

/**
 * Get the size of the given DOM element (including margins).
 * @param {Object} elem DOM element
 * @return {Object} Object specifying size of element with properties w and h
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._getElemSize = function(elem)
{
  var mlcClass = MasonryLayoutCommon;
  var computedStyle = mlcClass._getComputedStyle(elem);
  var extraWidth = mlcClass._getCSSLengthAsInt(computedStyle.marginLeft) + 
                   mlcClass._getCSSLengthAsInt(computedStyle.marginRight);
  var extraHeight = mlcClass._getCSSLengthAsInt(computedStyle.marginTop) + 
                    mlcClass._getCSSLengthAsInt(computedStyle.marginBottom);
  return {w: elem.offsetWidth + extraWidth, h: elem.offsetHeight + extraHeight};
};

/**
 * Get the insets of a given DOM element (padding and border).
 * @param {Object} elem DOM element
 * @return {Object} Object specifying size of insets with properties:
 *         paddingLeft, paddingRight, paddingTop, paddingBottom,
 *         bodrerLeftWidth, borderRightWidth, borderTopWidth, borderBottomWidth
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._getElemInsets = function(elem)
{
  var mlcClass = MasonryLayoutCommon;
  var computedStyle = mlcClass._getComputedStyle(elem);
  return {paddingLeft:       mlcClass._getCSSLengthAsInt(computedStyle.paddingLeft), 
          paddingRight:      mlcClass._getCSSLengthAsInt(computedStyle.paddingRight), 
          paddingTop:        mlcClass._getCSSLengthAsInt(computedStyle.paddingTop), 
          paddingBottom:     mlcClass._getCSSLengthAsInt(computedStyle.paddingBottom),
          borderLeftWidth:   mlcClass._getCSSLengthAsInt(computedStyle.borderLeftWidth), 
          borderRightWidth:  mlcClass._getCSSLengthAsInt(computedStyle.borderRightWidth),
          borderTopWidth:    mlcClass._getCSSLengthAsInt(computedStyle.borderTopWidth), 
          borderBottomWidth: mlcClass._getCSSLengthAsInt(computedStyle.borderBottomWidth)};
};

/**
 * Get the computed style of the given DOM element.
 * @param {Object} elem DOM element
 * @return {Object} Computed style for the element
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._getComputedStyle = function(elem)
{
  var elemOwnerDoc = elem.ownerDocument;
  var defView = elemOwnerDoc.defaultView;
  var computedStyle = null;
  if (defView)
  {
    //this line copied from AdfAgent.getComputedStyle()
    computedStyle = defView.getComputedStyle(elem, null);
  }
  else
  {
    //this line copied from AdfIEAgent.getComputedStyle()
    computedStyle = elem.currentStyle;
  }
  return computedStyle;
};

/**
 * Get the int value of a CSS length.
 * @param {string} cssLength cssLength as a string
 * @return {number} cssLength as an int
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._getCSSLengthAsInt = function(cssLength)
{
  //this function copied from AdfAgent.getCSSLengthAsInt
  if ((cssLength.length) > 0 && (cssLength != 'auto'))
  {
    var intLength = parseInt(cssLength, 10);

    if (isNaN(intLength))
      intLength = 0;

    return intLength;
  }
  return 0;
};

/**
 * Get the index of an item in an array.
 * @param {Array} array Array to search
 * @param {Object} item Item to search for
 * @return {number} index of item in array, or -1 if not found
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._arrayIndexOf = function(array, item)
{
  if (array)
  {
    for (var i = 0; i < array.length; i++)
    {
      if (array[i] == item)
        return i;
    }
  }
  return -1;
};

/**
 * Gets whether the specified agent minimum requirements are met or exceeded.
 * Every 2 arguments past actualAgentType and actualAgentVersion must correspond
 * to a minimum required agent type and floating point version number.
 * @param {string} actualAgentType the actual agent type ("trident", "webkit", "gecko")
 * @param {number} actualAgentVersion the actual agent version number as a floating point number
 * @param {Array} agentRequirements the required agent types and versions
 * @return {boolean} whether the specified agent minimums are met
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._isMinimumAgentMet = function(actualAgentType, actualAgentVersion, agentRequirements)
{
  //This function was copied from adf.shared.impl.animationUtils
  
  var argCount = agentRequirements.length;
  if (argCount % 2 == 0) // even number
  {
    // Loop through each requirement pair to see if we match one
    for (var i = 0; i <= argCount - 2; i += 2)
    {
      var requirementType = agentRequirements[i];
      if (actualAgentType == requirementType)
      {
        // We found an agent type match so now see if the actual version is greater than or equal
        // to the requirement version number:
        var requirementVersion = agentRequirements[1+i];
        if (actualAgentVersion >= requirementVersion)
          return true; // met requirement
        else
          return false; // failed requirement
      }
    }
  }
  return false; // no agent type match found; failed requirement
};

/**
 * Gets the agent type and version.
 * @param {string} givenUserAgentString the navigator's userAgent property
 * @return {Array} with 2 members, a string for the agent 
 *         type ("trident", "webkit", "gecko") and a Float for the agent version
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._getAgentTypeAndVersion = function(givenUserAgentString)
{
  //This function was copied from adf.shared.impl.animationUtils
  
  var mlcClass = MasonryLayoutCommon;
  var versionParser = mlcClass._parseFloatVersion;
  var agentType = null;
  var agentVersion = -1;
  var userAgent = givenUserAgentString.toLowerCase();
  if (userAgent.indexOf("msie") != -1 || userAgent.indexOf("trident") != -1)
  {
    agentType = "trident";
    var possibleVersion = versionParser(userAgent, /trident\/(\d+[.]\d+)/);
    if (possibleVersion != -1)
    {
      // 6.0 = IE10
      // 5.0 = IE9
      // 4.0 = IE8
      agentVersion = possibleVersion;
    }
    else
    {
      possibleVersion = versionParser(userAgent, /msie (\d+\.\d+);/);
      if (possibleVersion == -1)
        possibleVersion = versionParser(userAgent, /msie (\d+\.\d+)b;/); // expression for betas
      agentVersion = possibleVersion - 4; // Trident versions are 4 behind IE numbers
    }
    if (document.documentMode != null)
    {
      // If a documentMode is provided, it would be an IE number and Trident versions are 4 behind IE numbers.
      // The actual Trident version in use would be the smaller of the 2 numbers:
      agentVersion = Math.min(agentVersion, document.documentMode - 4);
    }
  }
  else if (userAgent.indexOf("applewebkit") != -1)
  {
    agentType = "webkit";
    // 536.26.17 = Mac Desktop Safari 6.0.2
    // 535.1 = Chrome 13.0.782.1
    // 534.46 = Safari 5.1 or iOS 5
    // 525.18 = Mac/Windows Desktop Safari 3.1.1
    // 420.1 = iOS 3
    agentVersion = versionParser(userAgent, /applewebkit\/(\d+([.]\d+)*)/);
  }
  else if (userAgent.indexOf("gecko/")!=-1)
  {
    agentType = "gecko";
    // rv:5 = Firefox 5
    // rv:2 = Firefox 4
    // rv:1.9 = Firefox 3
    // rv:1.8.1 = Firefox 2
    // rv:1.8 = Firefox 1.5
    agentVersion = versionParser(userAgent, /rv:(\d+[.]\d+)/);
  }
  return [ agentType, agentVersion ];
};

/**
 * Parses the float version out of of the specified agent string using
 * a regular expression to identify the version portion of the string.
 * @param {string} userAgent the lowercase navigator user agent string
 * @param {RegExp} versionNumberPattern the regular expression pattern 
 *        used to extract a number that will be parsed into a float
 * @return {number} version number
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._parseFloatVersion = function(userAgent, versionNumberPattern)
{
  //This function was copied from adf.shared.impl.animationUtils
  
  var matches = userAgent.match(versionNumberPattern);
  if (matches)
  {
    var versionString = matches[1];
    if (versionString)
      return parseFloat(versionString);
  }
  return -1;
};

/**
 * Comparison function used to sort tiles according to their layout positions.
 * @param {Object} pos1 Object that defines the startCol and startRow for a tile
 * @param {Object} pos2 Object that defines the startCol and startRow for another tile
 * @return {number} -1 if pos1 should be sorted before pos2, 1 if pos2 should be sorted 
 *         before pos1, or 0 if the positions are the same (which should never happen
 *         in practice)
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._compareTilePositions = function(pos1, pos2)
{
  //FIX : sort by rows first, because if the rows are different then 
  //the columns don't matter
  if (pos1.startRow > pos2.startRow)
    return 1;
  if (pos1.startRow < pos2.startRow)
    return -1;
  
  //sort by columns only if the rows are the same
  if (pos1.startCol > pos2.startCol)
    return 1;
  if (pos1.startCol < pos2.startCol)
    return -1;
  
  //this shouldn't happen in practice, but if the positions are the same, don't
  //sort the tiles
  return 0;
};

/**
 * Queue a relayout.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._queueRelayout = function()
{
  if (!this._hideTilesTimeout)
  {
    //FIX : if we're not in a layout cycle, set the timeout to start one;
    //otherwise, if we are in a layout cycle, set the flag indicating we need to queue
    //another one
    if (!this._layoutPhase)
    {
      //FIX : add busy state before animating a layout change
      if (this._addBusyStateFunc)
        this._busyStateResolveFunc = this._addBusyStateFunc("performing layout");
      
      this._hideTilesTimeout = setTimeout(this._hideTilesFunc, 0); // @HtmlUpdateOK
    }
    else if (!this._queuedRelayout)
      this._queuedRelayout = true;
  }
};

/**
 * Get the rendered child tiles.
 * @return {Array} Array of rendered child tiles
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._getTileChildren = function()
{
  var elem = this._elem;
  var children = elem.querySelectorAll(this._tilesSelector);
  var arChildren = [];
  for (var i = 0; i < children.length; i++)
  {
    var child = children[i];
    var childStyle = child.style;
    if (child.offsetWidth > 0 && child.offsetHeight > 0 && childStyle.visibility != "hidden")
      arChildren.push(child);
  }
  return arChildren;
};

/**
 * Layout and animate with a transition.  This is the second phase of the
 * <hide, layout, show> sequence.
 * @return {boolean} true if layout changed, otherwise false
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._transitionLayout = function()
{
  var oldMovedInfolets = this._arMovedInfolets;
  var newMovedInfolets = this._layout();
  
  //an infolet should be considered to "move" if it's resized, even if it doesn't change position,
  //because it will still go through a transition
  if (this._arInfoletsToResize)
  {
    var arInfoletsToResize = this._arInfoletsToResize;
    if (!newMovedInfolets)
      newMovedInfolets = [];
    var mlcClass = MasonryLayoutCommon;
    for (var i = 0; i < arInfoletsToResize.length; i += 2)
    {
      var resizedInfolet = arInfoletsToResize[i];
      if (mlcClass._arrayIndexOf(newMovedInfolets, resizedInfolet) < 0)
      {
        newMovedInfolets.push(resizedInfolet);
      }
    }
  }
  
  var calledHandleTransitionEnd = false;
  if (!newMovedInfolets || newMovedInfolets.length < 1)
  {
    if (!oldMovedInfolets || oldMovedInfolets.length < 1)
    {
      this._arMovedInfolets = null;
      this._handleTransitionEnd(null);
      calledHandleTransitionEnd = true;
    }
  }
  else
  {
    this._arMovedInfolets = newMovedInfolets;
  }
  
  var ret = (newMovedInfolets != null && newMovedInfolets.length > 0);
  
  if (!this.isAnimationEnabled() && !calledHandleTransitionEnd)
    this._handleTransitionEnd(null);
  
  return ret;
};

/**
 * Layout.
 * @return {Array} Array of tiles that were moved
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._layout = function()
{
  var elem = this._elem;
  var children = this._getTileChildren();
  //FIX : need to sort the tiles into their original DOM order 
  //before layout so that layout always starts from the intended order of the 
  //children, not from the current layout order
  if (this._sortTilesOriginalOrderFunc)
    this._sortTilesOriginalOrderFunc(children);
  //always recalculate cell size in case the app has specified media queries
  //to change CSS cell sizes based on screen or component size
  this._cellSize = null;
  var cellSize = null;
  this._cols = 0;
  this._rows = 1;
  this._occupancyMap = null;
  var arMovedInfolets = [];
  var arOldPositions = [];
  var arNewPositions = [];
  var arCols = [];
  var rtl = this._rtl;
  //FIX : need to take container border and padding into account for layout
  var mlcClass = MasonryLayoutCommon;
  var insets = mlcClass._getElemInsets(elem);
  //FIX : keep track of maxColSpan to define width of layout
  var maxColSpan = 0;
  //FIX : keep track of tile positions so that tiles can later be 
  //sorted into layout order
  var arTilePositions = [];
  this._arTilePositions = arTilePositions;
  for (var i = 0; i < children.length; i++)
  {
    var child = children[i];

    //if we've saved cell size on the infolet, we need to use it because
    //that refers to pre-transition cell size based on old size style class.
    //get the cell size if we've saved it on the infolet being resized, 
    //and then delete it from the infolet
    if (child._afrCellSize) {
      this._cellSize = child._afrCellSize;
      delete child._afrCellSize;
    }
    
    var childSpan = this._getTileSpanFunc(child);
    if (!this._cellSize)
    {
      this._cellSize = this._calcCellSize(child, childSpan);
    }
    cellSize = this._cellSize;

    //keep track of which cells are occupied
    if (!this._occupancyMap)
    {
      //make sure we have at least 1 column to work with
      //FIX : width available for layout is the offsetWidth minus
      //padding and borders
      var effectiveWidth = elem.offsetWidth - insets.paddingLeft - insets.paddingRight -
                           insets.borderLeftWidth - insets.borderRightWidth;
      this._cols = Math.max(Math.floor(effectiveWidth / cellSize.w), 1);
      this._initOccupancyMap(this._cols, this._rows);
      //FIX : initialize maxColSpan to define width of layout based 
      //on the calculated number of layout columns
      maxColSpan = this._cols;
    }
    
    //FIX : keep track of maxColSpan to define width of layout
    //(must do this using original value of colSpan BEFORE potentially decreasing 
    //colSpan based on number of _cols below)
    if (childSpan.colSpan > maxColSpan)
      maxColSpan = childSpan.colSpan;

    //handle case where childSpan is bigger than the number of cols by treating
    //childSpan as if it equaled the number of cols (can only modify childSpan
    //AFTER using original value to calculate cellSize above)
    if (childSpan.colSpan > this._cols)
      childSpan.colSpan = this._cols;

    var next = false;
    for (var r = 0; r < this._rows; r++)
    {
      for (var c = 0; c < this._cols; c++)
      {
        if (this._fits(c, r, childSpan))
        {
          //save old position before applying new one so that we can use it below to 
          //figure out which infolets actually moved
          var childStyle = child.style;
          var oldPosition = {top: childStyle.top};
          //FIX : in RTL, position tiles using "right" instead of "left"
          if (rtl)
            oldPosition.right = childStyle.right;
          else
            oldPosition.left = childStyle.left;
          arOldPositions.push(oldPosition);
          var newPosition = this._calcPosition(child, c, r, childSpan, cellSize, insets);
          arNewPositions.push(newPosition);
          
          if (rtl)
            arCols.push(c);
          next = true;
          //FIX : keep track of tile positions so that tiles can later be 
          //sorted into layout order
          arTilePositions.push({startCol: c, startRow: r, tile: child});
          break;
        }
      }
      if (next)
        break;
      if (r === this._rows - 1)
      {
        this._addRow();
      }
    }
  }

  //only need to adjust sizeDiv and adjust layout for rtl if there are child tiles
  if (cellSize)
  {
    var sizeDiv = this._sizeDiv;
    var style = sizeDiv.style;
    
    if (style)
    {
      //FIX : define width based on the maxColSpan
      var width = (maxColSpan * cellSize.w);
      var height = (this._rows * cellSize.h);

      //get the current sizeDiv width and height
      var currWidth = style.width ? MasonryLayoutCommon._getCSSLengthAsInt(style.width.toString()) : 0;
      var currHeight = style.height ? MasonryLayoutCommon._getCSSLengthAsInt(style.height.toString()) : 0;
      if (currWidth > width || currHeight > height) {
        //temporarily store the new width/height of the sizeDiv
        //this dimensions will be applied only after the layout transition is complete,
        //because the new dimensions are smaller than current sizeDiv dimensions
        this._afrSizeDivDims = {};
        this._afrSizeDivDims.width = width + "px";
        this._afrSizeDivDims.height = height + "px";
      }
      //temporarily set the max dimensions on sizeDiv until the layout transition is complete
      //this will avoid any appearance of scrollbar during the layout transition.
      style.width = (currWidth > width ? currWidth : width) + "px";
      style.height = (currHeight > height ? currHeight : height) + "px";
    }
  }
  
  //figure out which infolets actually moved (need to do this AFTER adjusting for RTL above)
  for (var i = 0; i < children.length; i++)
  {
    var child = children[i];
    var childStyle = child.style;
    var newPosition = arNewPositions[i];
    var oldPosition = arOldPositions[i];
    
    //if this is a new tile that masonryLayout didn't know about, then
    //don't mark it as being moved because it won't actually transition
    if (oldPosition.top != "")
    {
      //round to int because fractional value differences in Firefox don't actually result in transition,
      //and it really doesn't make much difference to the user whether they're animated because 
      //they're so small
      //FIX : in RTL, position tiles using "right" instead of "left"
      if ((rtl && (parseInt(newPosition.right, 10) !== parseInt(oldPosition.right, 10))) ||
          (!rtl && (parseInt(newPosition.left, 10) !== parseInt(oldPosition.left, 10))) || 
          (parseInt(newPosition.top, 10) !== parseInt(oldPosition.top, 10)))
      {
        arMovedInfolets.push(child);
        if (this._layoutTransition)
        {
          if (mlcClass._arrayIndexOf(this._arInfoletsToResize, child) < 0)
          {
            // This applies the transition styles via the animation framework and this 
            // has to be called before positioning the tiles for the animation to work correctly.
            // Add transition only if child is not also being resized.  Otherwise
            // the resize code would have added the transition.
            this._animateLayout(child, this._reorderTransitionStarted ? "reorder" : "move");
          }
        }
      }
    }
    
    // FIX  - MASONRY LAYOUT: LAYOUT DOES NOT ANIMATE FOR THE FIRST TIME 
    // Position the child tiles after setting the layout transition animation.
    childStyle.top = newPosition.top;
    if (this._rtl)
    {
      childStyle.right = newPosition.right;
    }
    else
    {
      childStyle.left = newPosition.left;
    }
  }

  if (arMovedInfolets.length < 1)
  {
    arMovedInfolets = null;
  }
  return arMovedInfolets;
};

/**
 * Update sizeDiv dimensions after the layout transition is complete
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._updateSizeDivOnEnd = function()
{
  //if new sizeDiv dimension is available, set that on sizeDiv
  if (this._afrSizeDivDims)
  {
    var style = this._sizeDiv.style;
    style.width = this._afrSizeDivDims.width;
    style.height = this._afrSizeDivDims.height;
    delete this._afrSizeDivDims;
  }
}

/**
 * Reorder tiles in the DOM to match the visual layout order so that tab order
 * and screen reader reading order match the visual layout order.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._reorderTilesForLayout = function()
{
  //FIX : loop through the DOM order and layout order, and where
  //the tiles at a given index are different, rearrange the DOM order to 
  //match the layout order
  var arTilePositions = this._arTilePositions;
  this._arTilePositions = null;
  var mlcClass = MasonryLayoutCommon;
  arTilePositions = arTilePositions.sort(mlcClass._compareTilePositions);
  var children = this._getTileChildren();
  for (var i = 0; i < children.length; i++)
  {
    var child = children[i];
    var pos = arTilePositions[i];
    var posTile = pos.tile;
    //if the tiles are different, need to rearrange the DOM
    if (child != posTile)
    {
      //notify the element that it's been detached from the DOM BEFORE actually
      //detaching it so that components can save state
      this._subtreeDetachedFunc(posTile);
      //insert the tile from the layout order into the DOM at this index
      var parentNode = child.parentNode;
      parentNode.insertBefore(posTile, child); // @HTMLUpdateOK
      //notify the element that it's been attached to the DOM
      this._subtreeAttachedFunc(posTile);
      
      //rearrange the array of children to match the DOM reorder above
      var posTileIndex = mlcClass._arrayIndexOf(children, posTile);
      if (posTileIndex > i)
      {
        children.splice(posTileIndex, 1);
        children.splice(i, 0, posTile);
      }
    }
  }
};

/**
 * Initialize the map of which cells are occupied.
 * @param {number} cols Number of columns in the layout
 * @param {number} rows Number of rows in the layout
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._initOccupancyMap = function(cols, rows)
{
  this._occupancyMap = [];
  var occupancyMap = this._occupancyMap;
  for (var row = 0; row < rows; row++)
  {
    var arCols = [];
    occupancyMap.push(arCols);
    for (var col = 0; col < cols; col++)
      arCols[col] = false;
  }
};

/**
 * Add a row to the layout.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._addRow = function()
{
  this._rows++;

  var arCols = [];
  var occupancyMap = this._occupancyMap;
  occupancyMap.push(arCols);
  for (var col = 0; col < this._cols; col++)
    arCols[col] = false;
};

/**
 * Get whether a tile of the given size fits at the given location.
 * @param {number} col Column index
 * @param {number} row Row index
 * @param {Object} childSpan Object specifying tile span with colSpan and rowSpan properties
 * @return {boolean} True if the tile fits, false otherwise
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._fits = function(col, row, childSpan)
{
  var colSpan = childSpan.colSpan;
  var rowSpan = childSpan.rowSpan;
  for (var r = row; r < row + rowSpan; r++)
  {
    if (r >= this._rows)
      this._addRow();
    for (var c = col; c < col + colSpan; c++)
    {
      if (c >= this._cols)
        return false;
      if (this._occupancyMap[r][c])
        return false;
    }
  }
  return true;
};

/**
 * Calculates the child tile position.
 * @param {Object} child Tile to position
 * @param {number} col Column index
 * @param {number} row Row index
 * @param {Object} childSpan Object specifying child span with colSpan and rowSpan properties
 * @param {Object} cellSize Object specifying unit cell size with w and h properties
 * @param {Object} insets The insets of the layout container
 * @return {Object} Tile position
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._calcPosition = function(child, col, row, childSpan, cellSize, insets)
{
  var colSpan = childSpan.colSpan;
  var rowSpan = childSpan.rowSpan;
  var occupancyMap = this._occupancyMap;
  for (var r = row; r < row + rowSpan; r++)
  {
    for (var c = col; c < col + colSpan; c++)
      occupancyMap[r][c] = true;
  }
  //FIX : take padding into account
  var style = {};
  var newTop = insets.paddingTop + (row * cellSize.h);
  style.top = newTop + "px";
  //FIX : in RTL, position tiles using "right" instead of "left"
  if (this._rtl)
  {
    var newRight = insets.paddingRight + (col * cellSize.w);
    style.right = newRight + "px";
  }
  else
  {
    var newLeft = insets.paddingLeft + (col * cellSize.w);
    style.left = newLeft + "px";
  }
  return style;
};

/**
 * Remove a style class from all of the rendered tile children.
 * @param {string} styleClassName Style class to remove
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._removeStyleClassFromTiles = function(styleClassName)
{
  var children = this._getTileChildren();
  for (var i = 0; i < children.length; i++)
    this._removeStyleClassNameFunc(children[i], styleClassName);
};

/**
 * Animate layout
 * @param {Element} element The element to animate
 * @param {string} action The element action
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._animateLayout = function(element, action)
{
  if (this.isAnimationEnabled())
  {
    var self = this;
    self._startAnimationFunc(element, action).then(function() {
      if (self._layoutTransition)
      {
        // Only need to do this once per layout cycle.
        // _handleTransitionEnd will set _layoutTransition to false when 
        // at least one tile is moved or no tile needs moving.
        self._handleTransitionEnd(element);
      }
    });
  }
};

/**
 * Start a layout transition.
 * @param {boolean} reorder True if this layout is due to a reorder, false otherwise
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._transitionStart = function(reorder)
{
  if (!this._layoutTransition)
  {
    this._reorderTransitionStarted = reorder;
    this._layoutTransition = true;
  }
};

/**
 * Handle a layout transition end.
 * @param {Object} target Target element
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._handleTransitionEnd = function(target)
{
  var doneTransitioning = true;
  if (this._arMovedInfolets)
  {
    var arMovedInfolets = this._arMovedInfolets;
    if (target)
    {
      for (var i = 0; i < arMovedInfolets.length; i++)
      {
        if (target === arMovedInfolets[i])
        {
          //as soon as we get a transition end for one moved tile, clear all the moved tiles
          this._arMovedInfolets = [];
          arMovedInfolets = this._arMovedInfolets;
          break;
        }
      }
    }
    else if (!this.isAnimationEnabled())
    {
      //if animation is not enabled, clear all the moved infolets at the same time
      this._arMovedInfolets = [];
      arMovedInfolets = this._arMovedInfolets;
    }
    //if there are still tiles for which we haven't gotten a transition end event yet,
    //then we're not done yet
    if (arMovedInfolets.length > 0)
      doneTransitioning = false;
  }
  if (doneTransitioning)
  {
    if (this._arInfoletsToResize)
    {
      var arInfoletsToResize = this._arInfoletsToResize;
      this._arInfoletsToResize = null;

      if (this.isAnimationEnabled())
      {
        for (var i = 0; i < arInfoletsToResize.length; i += 2)
        {
          var infolet = arInfoletsToResize[i];
        }
      }
    }

    if (this._reorderTransitionStarted)
    {
      this._reorderTransitionStarted = false;
    }
    var mlcClass = MasonryLayoutCommon;
    this._layoutTransition = false;
    this._resizingInfolet = false;
    this._hidingInfolets = false;
    this._showingInfolets = false;

    //FIX : reorder the tile DOM elements to match the visual layout order
    this._reorderTilesForLayout();

    if (this._layoutOnEndFunc)
      this._layoutOnEndFunc();
    
    //update sizeDiv after the layout transition is complete
    this._updateSizeDivOnEnd();
    
    if (this._layoutPhase === mlcClass._PHASE_LAYOUT)
    {
      if (this.isAnimationEnabled())
      {
        //Need to use a timeout here because we don't keep track of all the expected
        //transitionend events from the layout phase.  We listened for the first one
        //and removed the listener and transition style classes.  But, if we 
        //immediately _showTiles, the transitionend listener we add in there
        //may receive some leftover layout transitionend events.  By using the
        //timeout, we don't inadvertently get those leftover events.  
        var self = this;
        this._showTilesTimeout = setTimeout(function() {
          self._showTiles();
        }, 0);
      }
      else
      {
        this._showTiles();
      }
    }
    else if (!this._layoutPhase)
    {
      //FIX : resolve busy state after laying out
      if (this._layoutCycleOnEndFunc)
        this._layoutCycleOnEndFunc(this._resolveBusyStateFunc);
    }
  }
};

/**
 * Calculate the unit cell size.
 * @param {Object} child Tile to use for calculating cell size
 * @param {Object} childSpan Object specifying child span with colSpan and rowSpan properties
 * @return {Object} Object specifying unit cell size with w and h properties
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._calcCellSize = function(child, childSpan)
{
  var mlcClass = MasonryLayoutCommon;
  var childSize = mlcClass._getElemSize(child);
  return {w: childSize.w / childSpan.colSpan, h: childSize.h / childSpan.rowSpan};
};

/**
 * Hide any tiles waiting to be hidden.  This is the first phase of the
 * <hide, layout, show> sequence.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._hideTiles = function()
{
  if (this._hideTilesTimeout)
  {
    clearTimeout(this._hideTilesTimeout);
    this._hideTilesTimeout = null;
  }
  
  //FIX : notify the peer that a layout cycle is starting so that
  //it can save state if it wants
  if (this._layoutCycleOnStartFunc)
    this._layoutCycleOnStartFunc();
  
  var mlcClass = MasonryLayoutCommon;
  this._layoutPhase = mlcClass._PHASE_HIDE;
  
  if (this._arInfoletsToHide && this.isAnimationEnabled())
  {
    var arInfoletsToHide = this._arInfoletsToHide;
    var self = this;
    this._hideTilesAnimationStarted = true;
    for (var i = 0; i < arInfoletsToHide.length; i++)
    {
      var infolet = arInfoletsToHide[i];
      this._startAnimationFunc(infolet, "remove").then(function() {
        self._handleHideTransitionEnd(infolet);
      });
    }
  }
  else
  {
    this._handleHideTransitionEnd(null);
  }
};

/**
 * Handle the end of the hide transition.
 * @param {Element} target The target element
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._handleHideTransitionEnd = function(target)
{
  //FIX : clear this timeout at the end of the hide transition 
  //instead of when the _hideTilesInternalTimeout closure is called so that 
  //finishLayoutCycle() can accurately detect when we're in the hide phase
  if (this._hideTilesInternalTimeout)
  {
    clearTimeout(this._hideTilesInternalTimeout);
    this._hideTilesInternalTimeout = null;
  }
  else if (this._hideTilesAnimationStarted)
  {
    this._hideTilesAnimationStarted = false;
  }
  
  var mlcClass = MasonryLayoutCommon;
  if (target)
  {
    var infolet = target;
    var arInfoletsToHide = this._arInfoletsToHide;
    if (arInfoletsToHide)
    {
      for (var i = 0; i < arInfoletsToHide.length; i++)
      {
        var tmpInfolet = arInfoletsToHide[i];
        if (tmpInfolet === infolet)
        {
          arInfoletsToHide.splice(i, 1);
          if (!this._arFireHideOnEnd)
            this._arFireHideOnEnd = [];
          var arFireHideOnEnd = this._arFireHideOnEnd;
          arFireHideOnEnd.push(infolet);
          break;
        }
      }
      if (arInfoletsToHide.length < 1)
        this._arInfoletsToHide = null;
    }
  }
  else if (!this.isAnimationEnabled())
  {
    //if animation is not enabled, process all the hides at the same time
    var arInfoletsToHide = this._arInfoletsToHide;
    if (arInfoletsToHide)
    {
      for (var i = 0; i < arInfoletsToHide.length; i++)
      {
        var tmpInfolet = arInfoletsToHide[i];
        if (!this._arFireHideOnEnd)
          this._arFireHideOnEnd = [];
        var arFireHideOnEnd = this._arFireHideOnEnd;
        arFireHideOnEnd.push(tmpInfolet);
      }
      this._arInfoletsToHide = null;
    }
  }
  
  if (!this._arInfoletsToHide)
  {
    if (this._arFireHideOnEnd)
    {
      var arFireHideOnEnd = this._arFireHideOnEnd;
      for (var i = 0; i < arFireHideOnEnd.length; i++)
      {
        var infolet = arFireHideOnEnd[i];
        
        //remove layout position from the child
        var style = infolet.style;
        if (this._rtl)
          style.right = "";
        else
          style.left = "";
        style.top = "";
        
        if (this._hideTileOnEndFunc)
          this._hideTileOnEndFunc(infolet);
      }
      this._arFireHideOnEnd = null;
    }
    
    this._layoutPhase = mlcClass._PHASE_LAYOUT;
    
    this._transitionStart(false);
    
    //update size style classes AFTER starting transition, but BEFORE doing layout
    if (this._arInfoletsToResize)
    {
      var arInfoletsToResize = this._arInfoletsToResize;
      for (var i = 0; i < arInfoletsToResize.length; i += 2)
      {
        var infolet = arInfoletsToResize[i];
        var newSizeStyleClass = arInfoletsToResize[i + 1];
        var oldSizeStyleClass = this._getSizeStyleClassNameFunc(infolet);
        //if we add a _transitionResizeFromStyleClass in the future, we'll have
        //to add it here immediately and then do the below in a timeout because
        //it would need to happen AFTER the _transitionResizeFromStyleClass is 
        //applied
        if (this.isAnimationEnabled())
        {
          //temporarily store current cell size on the infolet before the transition begins, 
          //because the element size still corresponds to the old size style class before
          //the transition runs
          infolet._afrCellSize = this._cellSize;
          
          // This must be called before the style class change in order to work
          // consistently.
          this._animateLayout(infolet, "resize");
        }
        this._removeStyleClassNameFunc(infolet, oldSizeStyleClass);
        this._addStyleClassNameFunc(infolet, newSizeStyleClass);
      }
    }
    
    this._transitionLayout();
  }
};

/**
 * Show any tiles that were inserted.  This is the third phase of the
 * <hide, layout, show> sequence.
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._showTiles = function()
{
  if (this._showTilesTimeout)
  {
    clearTimeout(this._showTilesTimeout);
    this._showTilesTimeout = null;
  }
  
  var mlcClass = MasonryLayoutCommon;
  this._layoutPhase = mlcClass._PHASE_SHOW;
  
  if (this._arInfoletsToShow && this.isAnimationEnabled())
  {
    var self = this;
    var arInfoletsToShow = this._arInfoletsToShow;
    for (var i = 0; i < arInfoletsToShow.length; i++)
    {
      var infolet = arInfoletsToShow[i];
      this._removeStyleClassNameFunc(infolet, this._transitionShowFromStyleClass);
      this._startAnimationFunc(infolet, "insert").then(function() {
        self._handleShowTransitionEnd(infolet);
      });
    }
  }
  else
  {
    //FIX : make sure we remove the transitionShowFrom class that may have been
    //added by the peer
    if (this._arInfoletsToShow)
    {
      var arInfoletsToShow = this._arInfoletsToShow;
      for (var i = 0; i < arInfoletsToShow.length; i++)
      {
        var infolet = arInfoletsToShow[i];
        this._removeStyleClassNameFunc(infolet, this._transitionShowFromStyleClass);
      }
    }
    this._handleShowTransitionEnd(null);
  }
};

/**
 * Handle the end of the show transition.
 * @param {Element} target The target element
 * @return {void}
 * @memberof MasonryLayoutCommon
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._handleShowTransitionEnd = function(target)
{
  if (target)
  {
    var infolet = target;
    var arInfoletsToShow = this._arInfoletsToShow;
    if (arInfoletsToShow)
    {
      for (var i = 0; i < arInfoletsToShow.length; i++)
      {
        var tmpInfolet = arInfoletsToShow[i];
        if (tmpInfolet === infolet)
        {
          arInfoletsToShow.splice(i, 1);
          if (this._showTileOnEndFunc)
            this._showTileOnEndFunc(infolet);
          break;
        }
      }
      if (arInfoletsToShow.length < 1)
        this._arInfoletsToShow = null;
    }
  }
  else if (!this.isAnimationEnabled())
  {
    //if animation is not enabled, process all the shows at the same time
    var arInfoletsToShow = this._arInfoletsToShow;
    if (arInfoletsToShow)
    {
      for (var i = 0; i < arInfoletsToShow.length; i++)
      {
        var tmpInfolet = arInfoletsToShow[i];
        if (this._showTileOnEndFunc)
          this._showTileOnEndFunc(tmpInfolet);
      }
      this._arInfoletsToShow = null;
    }
  }
  
  //only advance the layout phase if there are no more infolets to show
  //still transitioning
  if (!this._arInfoletsToShow)
  {
    this._layoutPhase = null;
    //FIX : notify the callback that the layout cycle is done
    //FIX : resolve busy state after laying out
    if (this._layoutCycleOnEndFunc)
      this._layoutCycleOnEndFunc(this._resolveBusyStateFunc);
    //FIX : if we got another layout request while we were in the layout cycle,
    //queue another layout now
    if (this._queuedRelayout)
    {
      this._queuedRelayout = false;
      this._queueRelayout();
    }
  }
};

/**
 * Resolve an outstanding busy state.
 * @memberof MasonryLayoutCommon
 * @return {void}
 * @instance
 * @private
 */
MasonryLayoutCommon.prototype._resolveBusyState = function()
{
  if (this._busyStateResolveFunc)
  {
    this._busyStateResolveFunc();
    this._busyStateResolveFunc = null;
  }
};

/**
 * Layout phase: hide deleted tiles.
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._PHASE_HIDE = 1;
/**
 * Layout phase: layout tiles.
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._PHASE_LAYOUT = 2;
/**
 * Layout phase: show inserted tiles.
 * @memberof MasonryLayoutCommon
 * @private
 */
MasonryLayoutCommon._PHASE_SHOW = 3;
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojMasonryLayout
 * @augments oj.baseComponent
 * @since 1.1.0
 * @ojstatus preview
 * @ojshortdesc Container element that lays out its children in a grid of tiles.
 * @class oj.ojMasonryLayout
 * 
 * @classdesc
 * <h3 id="masonryLayoutOverview-section">
 *   JET MasonryLayout
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#masonryLayoutOverview-section"></a>
 * </h3>
 * 
 * <p>Description: A grid of tiles.
 * 
 * <p>A JET MasonryLayout can contain multiple direct child elements that can be sized and positioned.
 * <p>Each direct child element must be styled using one of the predefined 
 * <code class="prettyprint">oj-masonrylayout-tile-CxR</code> style classes to specify
 * the size of that tile.  A tile can span multiple columns and/or rows.  The
 * predefined size style classes are:
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Class</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-masonrylayout-tile-1x1</td>
 *       <td>A tile that spans 1 column and 1 row.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-masonrylayout-tile-1x2</td>
 *       <td>A tile that spans 1 column and 2 rows.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-masonrylayout-tile-1x3</td>
 *       <td>A tile that spans 1 column and 3 rows.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-masonrylayout-tile-2x1</td>
 *       <td>A tile that spans 2 columns and 1 row.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-masonrylayout-tile-2x2</td>
 *       <td>A tile that spans 2 columns and 2 rows.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-masonrylayout-tile-2x3</td>
 *       <td>A tile that spans 2 columns and 3 rows.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-masonrylayout-tile-3x1</td>
 *       <td>A tile that spans 3 columns and 1 row.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-masonrylayout-tile-3x2</td>
 *       <td>A tile that spans 3 columns and 2 rows.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * <p>The number of columns in the layout is determined by the width of the 
 * element and the width of a 1x1 tile.  The number of rows is determined by
 * the number of columns and the number and sizes of tiles to be laid out.  When
 * the element is resized, relayout will occur and the number of columns and
 * rows may change.  
 * <p>When performing layout, tiles are processed in the order in which they
 * originally appeared in the DOM.  Cells in the grid are processed in 
 * left-to-right, top-to-bottom order (or right-to-left, top-to-bottom order 
 * when the reading direction is right-to-left).  A tile will be positioned in 
 * the first empty cell in which it fits.  This can result in empty cells in the
 * layout.  Subsequent tiles may fill those earlier gaps if they fit.  
 * 
 * <pre class="prettyprint"><code>
 * &lt;oj-masonry-layout>
 *   &lt;div class='oj-masonrylayout-tile-2x1'>Alpha&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x2'>Beta&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Gamma&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Delta&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Epsilon&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Zeta&lt;/div>
 * &lt;/oj-masonry-layout>
 * </code></pre>
 *
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 * 
 * 
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 * 
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * 
 * <h3 id="keyboard-appdev-section">
 *   Keyboard Application Developer Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-appdev-section"></a>
 * </h3>
 * 
 * <p>Providing keyboard support for the tiles in the masonry layout is the 
 * responsibility of the application developer, if the tiles do not already
 * support keyboard interaction.  This could be done, for example, by specifying 
 * <code class="prettyprint">tabindex</code> on each tile to enable tab
 * navigation.  Alternatively, this could be done by adding a keyboard listener
 * and responding to key events, like pressing the arrow keys.
 * 
 * 
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>MasonryLayout is for layout only.  It is the responsibility of the 
 * application developer to make the tiles in the layout accessible.  Sighted 
 * keyboard-only users need to be able to access the tiles in the layout just by
 * using the keyboard.  It is up to the child tiles of the MasonryLayout to 
 * support keyboard navigation.  The MasonryLayout reorders the tile DOM
 * elements to match the visual layout order so that tab order and screen reader
 * reading order will match the visual layout order.
 * 
 * 
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 * 
 * <p>As with any JET element, in the unusual case that the directionality 
 * (LTR or RTL) changes post-init, the MasonryLayout must be 
 * <code class="prettyprint">refresh()</code>ed.
 * 
 * 
 * <h3 id="contextmenu-section">
 *   Context Menu
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#contextmenu-section"></a>
 * </h3>
 * <p>Masonry Layout supports context menu for performing cut and paste operations on tiles as 
 * an accessible alternative to drag and drop reordering.
 * When defining a context menu, masonry layout will provide built-in 
 * behavior for "cut" and "paste" if the following format for menu &lt;oj-option&gt; items is used:
 * <ul><li> &lt;oj-option data-oj-command="oj-masonrylayout-['commandname']" >&lt;/oj-option&gt;</li></ul>
 * The available translated text will be applied to menu items defined this way.
 *
 * <p>The supported commands:</p>
 * <table class="keyboard-table">
 *   <thead>
 *      <tr>
 *       <th>Default Function</th>
 *       <th>data-oj-command value</th>
 *      </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Cut</kbd></td>
 *       <td>oj-masonrylayout-cut</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Paste Before</kbd></td>
 *       <td>oj-masonrylayout-paste-before</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Paste After</kbd></td>
 *       <td>oj-masonrylayout-paste-after</td>
 *     </tr>
 * </tbody></table>
 *
 * @example <caption>Define a context menu in the masonry layout for performing cut and paste operation:</caption>
 * &lt;oj-masonry-layout&gt;
 *   &lt;oj-menu slot="contextMenu">
 *     &lt;oj-option data-oj-command="oj-masonrylayout-cut" >&lt;/oj-option>
 *     &lt;oj-option data-oj-command="oj-masonrylayout-paste-before" >&lt;/oj-option>
 *     &lt;oj-option data-oj-command="oj-masonrylayout-paste-after" >&lt;/oj-option>
 *   &lt;/oj-menu&gt;
 * &lt;/oj-masonry-layout>
 */
/**
 * <p> &lt;oj-masonry-layout> accepts arbitrary direct child elements that can be
 * sized and positioned, which it will lay out as a grid of tiles.  Each direct
 * child element must be styled using one of the predefined <code
 * class="prettyprint">oj-masonrylayout-tile-CxR</code> style classes to specify
 * the size of that tile.  A tile can span multiple columns and/or rows.</p>
 *
 * @ojchild Default
 * @memberof oj.ojMasonryLayout
 *
 * @example <caption>Initialize the masonry layout with child content specified:</caption>
 * &lt;oj-masonry-layout>
 *   &lt;div class='oj-masonrylayout-tile-2x1'>Alpha&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x2'>Beta&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Gamma&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Delta&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Epsilon&lt;/div>
 *   &lt;div class='oj-masonrylayout-tile-1x1'>Zeta&lt;/div>
 * &lt;/oj-masonry-layout>
 */
(function()
{
oj.__registerWidget("oj.ojMasonryLayout", $['oj']['baseComponent'],
{
  defaultElement: "<div>",
  widgetEventPrefix: "oj",

  options:
  {
    /** 
     * Specify the selector of the descendant DOM element in a <code class="prettyprint">oj-masonry-layout</code> 
     * child that can be used to reorder the child by drag-and-drop.  
     * 
     * <p>This attribute value is <code class="prettyprint">null</code> by default, 
     * meaning that <code class="prettyprint">oj-masonry-layout</code> children 
     * cannot be reordered.  If each child that can be reordered has an element 
     * with style class <code class="prettyprint">'my-reorder-handle'</code>, 
     * then the <code class="prettyprint">reorder-handle</code> attribute would be specified as 
     * <code class="prettyprint">'.my-reorder-handle'</code>.
     * 
     * <p>When specifying a <code class="prettyprint">reorder-handle</code>, 
     * the application should also specify a context menu with actions to cut
     * and paste tiles as an accessible alternative to drag-and-drop reordering.
     * <p>To specify a context menu, see the documentation of <a href="oj.ojMenu.html">oj-menu</a>.</p>
     *
     * @expose 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @type {?string}
     * @default null
     * @ojshortdesc Specifies the selector of the child DOM element that can be used to reorder the child.
     *
     * @example <caption>Initialize the masonryLayout with the 
     * <code class="prettyprint">reorder-handle</code> attribute specified:</caption>
     * &lt;oj-masonry-layout reorder-handle='.my-reorder-handle'>
     *   &lt;div id='child1'>
     *     &lt;div class='my-reorder-handle'/>
     *   &lt;/div>
     *   &lt;div id='child2'>
     *     &lt;div class='my-reorder-handle'/>
     *   &lt;/div>
     *   &lt;!-- additional child elements... -->
     * &lt;/oj-masonry-layout>
     * 
     * @example <caption>Get or set the <code class="prettyprint">reorderHandle</code> 
     * property after initialization:</caption>
     * // getter
     * var contentParent = myMasonryLayout.reorderHandle;
     * 
     * // setter
     * myMasonryLayout.reorderHandle = '.my-reorder-handle';
     */
    reorderHandle: null,
    
    /**
     * MasonryLayout inherits the <code class="prettyprint">disabled</code> 
     * attribute from its superclass, but does not support it in order to avoid 
     * tight coupling between a MasonryLayout and its contents.  Setting this
     * attribute on MasonryLayout has no effect.
     * 
     * <p><b>WARNING:</b> Applications should not depend on this behavior 
     * because we reserve the right to change it in the future in order to 
     * support <code class="prettyprint">disabled</code> and propagate it to 
     * child elements of MasonryLayout.  
     * 
     * @member
     * @name disabled
     * @memberof oj.ojMasonryLayout
     * @instance
     * @type {boolean}
     * @default false
     * @ignore
     */
    // disabled attribute declared in superclass, but we still want the above API doc

    //event callbacks
    /**
     * Triggered when a default animation is about to start on the element.
     *
     * <p>The default animation can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>, 
     * followed by a call to <code class="prettyprint">event.detail.endCallback()</code>. <code class="prettyprint">event.detail.endCallback()</code> 
     * should be called immediately after <code class="prettyprint">event.preventDefault()</code> if the application merely 
     * wants to cancel animation, or it should be called when the custom animation ends if the application is invoking another animation function. 
     * Failure to call <code class="prettyprint">event.detail.endCallback()</code> may prevent the element from working properly.</p>
     *
     * <p>For more information on customizing animations, see the documentation of 
     * <a href="oj.AnimationUtils.html">oj.AnimationUtils</a>.</p>
     *
     * <caption>The default animations are controlled via the theme (SCSS) :</caption>
     * <pre class="prettyprint"><code>
     * $masonryLayoutInsertAnimation: ((effect: "zoomIn", duration: $masonryLayoutAnimationDuration, timingFunction: "ease-in-out"), "fadeIn") !default;
     * $masonryLayoutRemoveAnimation: ((effect: "zoomOut", duration: $masonryLayoutAnimationDuration, timingFunction: "ease-in-out"), "fadeOut") !default;
     * $masonryLayoutMoveAnimation: (effect: "addTransition", duration: $masonryLayoutAnimationDuration, timingFunction: "ease-in-out", transitionProperties: ('top', 'left', 'right')) !default;
     * $masonryLayoutResizeAnimation: (effect: "addTransition", duration: $masonryLayoutAnimationDuration, timingFunction: "ease-in-out", transitionProperties: ('width', 'height', 'top', 'left', 'right')) !default;
     * $masonryLayoutReorderAnimation: (effect: "addTransition", duration: $masonryLayoutAnimationDurationFast, timingFunction: "ease-in-out", transitionProperties: ('width', 'height', 'top', 'left', 'right')) !default;
     * </code></pre>
     *
     * @ojshortdesc Triggered when a default animation is about to start, such as when the component is
     * being opened/closed or a child item is being added/removed. The default animation can
     * be cancelled by calling event.preventDefault.
     *
     * @expose
     * @event
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojcancelable
     * @ojshortdesc Event triggered when the default animation of a particular action is about to start.
     * @property {string} action The action that triggers the animation.
     *            Supported values are:
     *                    <ul>
     *                      <li>"insert" - when a tile is inserted</li>
     *                      <li>"move" - when a tile is moved</li>
     *                      <li>"remove" - when a tile is removed</li>
     *                      <li>"resize" - when a tile is resized</li>
     *                      <li>"reorder" - when a tile is reordered</li>
     *                    </ul>
     *                    <br>Note that some animation effects may not look appropriate for a 
     *                    given action.
     * @property {Element} element The element being animated.
     * @property {function():void} endCallback If the event listener calls
     *            event.preventDefault to cancel the default animation, it must call the
     *            endCallback function after it finishes its own animation handling and any
     *            custom animation has ended.
     *
     * @example <caption>Specify an <code class="prettyprint">onOjAnimateStart</code> listener 
     *          to override the default "insert" animation with a custom animation:</caption>
     *   myMasonryLayout.onOjAnimateStart = function( event )
     *   {
     *     if (!$(event.target).is("oj-masonry-layout"))
     *       return;
     *     var ui = event.detail;
     *     // Change the insert animation
     *     if (ui.action == "insert") {
     *       event.preventDefault();
     *       // Invoke new animation and call event.detail.endCallback when it ends
     *       oj.AnimationUtils.fadeIn(ui.element).then(ui.endCallback);
     *     }
     *   }
     *
     * @example <caption>Specify an <code class="prettyprint">onOjAnimateStart</code> listener 
     *          to prevent the default "remove" animation:</caption>
     *   myMasonryLayout.onOjAnimateStart = function( event )
     *   {
     *     if (!$(event.target).is("oj-masonry-layout"))
     *       return;
     *     var ui = event.detail;
     *     // Prevent the remove animation
     *     if (ui.action == "remove") {
     *       event.preventDefault();
     *       ui.endCallback();
     *     }
     *   }
     */
    animateStart : null,

    /**
     * Triggered when a default animation has ended.
     *
     * @expose
     * @event
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojshortdesc Event triggered when the default animation of a particular action has ended.
     * @property {string} action The action that triggers the animation.
     *            Supported values are:
     *                    <ul>
     *                      <li>"insert" - when a tile is inserted</li>
     *                      <li>"move" - when a tile is moved</li>
     *                      <li>"remove" - when a tile is removed</li>
     *                      <li>"resize" - when a tile is resized</li>
     *                      <li>"reorder" - when a tile is reordered</li>
     *                    </ul>
     * @property {Element} element The element being animated.
     */
    animateEnd : null,

    /**
     * Triggered immediately before a tile is inserted.
     * This event can be cancelled by calling 
     * <code class="prettyprint">event.preventDefault()</code>.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojcancelable
     * @ojshortdesc Event triggered before a tile is inserted.
     * @property {Element} tile The tile that is about to be inserted.
     * @property {number} index The 0-based index into the set of rendered
     *                                       <code class="prettyprint">oj-masonry-layout</code> children where 
     *                                       the tile will be inserted.
     */
    beforeInsert: null,
    
    /**
     * Triggered immediately after a tile is inserted.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojshortdesc Event triggered after a tile is inserted.
     * @property {Element} tile  The tile that was inserted.
     * @property {number} index The 0-based index into the set of rendered
     *                                       <code class="prettyprint">oj-masonry-layout</code> children where 
     *                                       the tile was inserted.
     */
    insert: null,
    
    /**
     * Triggered immediately before a tile is removed.
     * This event can be cancelled by calling 
     * <code class="prettyprint">event.preventDefault()</code>.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojcancelable
     * @ojshortdesc Event triggered before a tile is removed.
     * @property {Element} tile  The tile that will be removed.
     */
    beforeRemove: null,
    
    /**
     * Triggered immediately after a tile is removed.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojshortdesc Event triggered after a tile is removed.
     * @property {Element} tile  The tile that was removed.
     */
    remove: null,
    
    /**
     * Triggered immediately before a tile is resized.
     * This event can be cancelled by calling 
     * <code class="prettyprint">event.preventDefault()</code>.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojcancelable
     * @ojshortdesc Event triggered before a tile is resized.
     * @property {Element} tile  The tile that will be resized.
     * @property {string} previousSizeStyleClass The previous size style
     *                                           class applied to the tile.
     * @property {string} sizeStyleClass The new size style class that will
     *                                   be applied to the tile.
     */
    beforeResize: null,
    
    /**
     * Triggered immediately after a tile is resized.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojshortdesc Event triggered after a tile is resized.
     * @property {Element} tile  The tile that was resized.
     * @property {string} previousSizeStyleClass The previous size style
     *                                           class applied to the tile.
     * @property {string} sizeStyleClass The new size style class applied to 
     *                                   to the tile.
     */
    resize: null,
    
    /**
     * Triggered immediately before a tile is reordered.
     * This event can be cancelled by calling 
     * <code class="prettyprint">event.preventDefault()</code>.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojcancelable
     * @ojshortdesc Event triggered before a tile is reordered.
     * @property {Element} tile  The tile that will be reordered.
     * @property {number} fromIndex The 0-based index into the set of rendered
     *                                           <code class="prettyprint">oj-masonry-layout</code> children from 
     *                                           which the tile will be reordered. 
     */
    beforeReorder: null,
    
    /**
     * Triggered immediately after a tile is reordered.
     *
     * @expose 
     * @event 
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ojshortdesc Event triggered after a tile is reordered.
     * @property {Element} tile  The tile that was reordered.
     * @property {number} fromIndex The 0-based index into the set of rendered
     *                                           <code class="prettyprint">oj-masonry-layout</code> children from 
     *                                           which the tile was reordered.
     * @property {number} toIndex   The 0-based index into the set of rendered
     *                                           <code class="prettyprint">oj-masonry-layout</code> children to 
     *                                            which the tile was reordered.  
     */
    reorder: null

    /**
     * Triggered when the masonryLayout is created.
     *
     * @event
     * @name create
     * @memberof oj.ojMasonryLayout
     * @instance
     * @ignore
     */
    // create event declared in superclass, but we still want the above API doc
  },

  /**
   * After _ComponentCreate and _AfterCreate, 
   * the widget should be 100% set up. this._super should be called first.
   * @return {void}
   * @override
   * @protected
   * @instance
   * @memberof oj.ojMasonryLayout
   */
  _ComponentCreate: function() // Override of protected base class method.  
  {
    //call superclass first
    this._super();
    
    var elem = this.element;  
    elem.addClass("oj-masonrylayout oj-component");
    
    //log warning message when "disabled" property set
    var options = this.options;
    if (options.disabled)
    {
      oj.Logger.warn(_WARNING_DISABLED_OPTION);
    }
    
    this.reorderHandleEventNamespace = this.eventNamespace + "ReorderHandle";
    
    this._menu = {};
    this._menu.usermenu = false;
    this._menu.$container = false;
    this._menu.$elemCut = false;
    this._menu.$elemPasteBefore = false;
    this._menu.$elemPasteAfter = false;

    this._initMenu();
    this._applyMenu();
    
    this._setup(true);
  },

  // This method currently runs at create, init, and refresh time (since refresh() is called by _init()).
  /**
   * Refreshes the visual state of the masonryLayout. JET elements require a 
   * <code class="prettyprint">refresh()</code> after the DOM is 
   * programmatically changed underneath the element.
   * 
   * <p>This method does not accept any arguments.
   * @return {void}
   * 
   * @expose 
   * @memberof oj.ojMasonryLayout
   * @instance
   * @ojshortdesc Refreshes the visual state of the masonryLayout.
   * 
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   * myMasonryLayout.refresh();
   */
  refresh: function() // Override of public base class method.  
  {
    this._super();
    
    //if RTL has changed, just destroy and recreate the MasonryLayoutCommon
    var bRTL = (this._GetReadingDirection() === "rtl");
    var bRecreate = (bRTL !== this._bRTL);
    if (bRecreate)
      this._destroyMLCommon();
    this._setup(bRecreate);
  },
  
  /**
   * Notifies the component that its subtree has been made visible 
   * programmatically after the component has been created.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @protected 
   * @override
   */
  _NotifyShown: function()
  {
    this._super();
    //perform a deferred layout
    if (this._needsSetup)
      this._setup(this._needsSetup[0]);
    else
    {
      //explicitly relayout in case we don't get a resize notification when shown
      var mlCommon = this._mlCommon;
      mlCommon.setup(true);
    }
  },
  
  /**
   * Notifies the component that its subtree has been connected to the document
   * programmatically after the component has been created.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @protected 
   * @override
   */
  _NotifyAttached: function()
  {
    this._super();
    //perform a deferred layout
    if (this._needsSetup)
      this._setup(this._needsSetup[0]);
    else
    {
      //explicitly relayout in case we don't get a resize notification when attached
      var mlCommon = this._mlCommon;
      mlCommon.setup(true);
    }
  },

  /**
   * Notifies the component that its context menu is opening.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @protected 
   * @override
   */
  _NotifyContextMenuGesture: function(menu, event, eventType)
  {
    //prepare the context menu here instead of in a context menu "beforeOpen" event
    this._prepareContextMenuBeforeOpen(event);
    
    //FIX : only allow reordering if context menu was 
    //launched from reorder handle, so only show context menu if it 
    //contains items that are not hidden (i.e. not cut/paste)
    var $menuContainer = this._menu.$container;
    var $menuItems = $menuContainer.children().not(".oj-helper-hidden");
    if ($menuItems && $menuItems.length > 0)
      this._OpenContextMenu(event, eventType, {"launcher": $(event.target).closest(":tabbable")});
  },

  // isInit is true for init (create and re-init), false for refresh
  /** 
   * Setup the masonryLayout.
   * @param {boolean} isInit true if _setup is called from _init(), false
   *        if called from refresh()
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _setup: function(isInit) // Private, not an override (not in base class).  
  {
    //if masonryLayout is detached or hidden, we can't layout correctly, so 
    //defer layout until masonryLayout is attached or shown
    if (!this._canCalculateSizes())
    {
      //want a true value of isInit to take precedence over a false value
      var oldIsInit = false;
      if (this._needsSetup)
      {
        oldIsInit = this._needsSetup[0];
      }
      this._needsSetup = [isInit || oldIsInit];
      return;
    }
    this._needsSetup = null;
    
    this._bRTL = (this._GetReadingDirection() === "rtl");
    this._bTouchSupported = oj.DomUtils.isTouchSupported();
    var elem = this.element;  
    var options = this.options;
    if (isInit)
    {
      var self = this;
      this._showTileOnEndFunc = function(elem) {self._showTileOnEnd(elem);};
      this._hideTileOnEndFunc = function(elem) {self._hideTileOnEnd(elem);};
      this._startAnimationFunc = function(elem, action) {return self._startAnimation(elem, action);};
      this._layoutOnEndFunc = function() {self._layoutOnEnd();};
      this._layoutCycleOnStartFunc = function() {self._layoutCycleOnStart();};
      this._layoutCycleOnEndFunc = 
        function(resolveBusyState) {self._layoutCycleOnEnd(resolveBusyState);};
      if (!this._mlCommon)
      {
        var selectors = {};
        selectors.tiles = _TILE_SELECTOR;
        var styles = {};
        styles.transitionShowFromStyleClass = _OJ_MASONRYLAYOUT_TILE_TRANSITION_SHOW_FROM_CLASS;
        var callbackInfo = {};
        callbackInfo.addStyleClassName = _addStyleClassName;
        callbackInfo.removeStyleClassName = _removeStyleClassName;
        callbackInfo.getSizeStyleClassName = _getSizeStyleClassName;
        callbackInfo.getTileSpan = _getTileSpan;
        callbackInfo.showTileOnEndFunc = this._showTileOnEndFunc;
        callbackInfo.hideTileOnEndFunc = this._hideTileOnEndFunc;
        callbackInfo.startAnimationFunc = this._startAnimationFunc;
        callbackInfo.layoutOnEndFunc = this._layoutOnEndFunc;
        callbackInfo.layoutCycleOnStartFunc = this._layoutCycleOnStartFunc;
        callbackInfo.layoutCycleOnEndFunc = this._layoutCycleOnEndFunc;
        callbackInfo.sortTilesOriginalOrderFunc = _sortTilesOriginalOrder;
        callbackInfo.subtreeAttached = oj.Components.subtreeAttached;
        callbackInfo.subtreeDetached = oj.Components.subtreeDetached;
        callbackInfo.addBusyState = 
          function(description) { return self._addBusyState(description); };
  
        //apply the oj-masonrylayout-tile style class to child tiles
        this._applyTileStyleClass();
        
        //save the original order of the tiles, because the DOM order may 
        //change due to layout and we want to always use the original order
        //for running subsequent layouts
        this._saveTilesOriginalOrder();
        
        this._mlCommon = new MasonryLayoutCommon(
            elem[0], 
            this._bRTL, 
            oj.Config.getAutomationMode() === "enabled",
            selectors, 
            styles, 
            callbackInfo);
      }
      
      this._handleDragStartFunc = function(event) {self._handleDragStart(event);};
      this._handleDragEnterFunc = function(event) {self._handleDragEnter(event);};
      this._handleDragOverFunc = function(event) {self._handleDragOver(event);};
      this._handleDragLeaveFunc = function(event) {self._handleDragLeave(event);};
      this._handleDragEndFunc = function(event) {self._handleDragEnd(event);};
      this._handleDropFunc = function(event) {self._handleDrop(event);};
      if (options.reorderHandle)
        this._setupReorderHandles();
    }
    else
    {
      //apply the oj-masonrylayout-tile style class to child tiles that are missing it
      this._applyTileStyleClass();
      
      //FIX : tear down and setup the reorder handles for the
      //tiles again in case something changed, for example if there are 
      //new tiles
      var children = elem.children();
      this._tearDownReorderHandlesForElem(children);
      this._setupReorderHandlesForElem(children);
      
      //make sure that any new tiles have an assigned index in the 
      //original DOM ordering
      this._checkTilesOriginalOrder();
    }
    
    var mlCommon = this._mlCommon;
    mlCommon.setup(isInit);
    
    if (isInit)
    {
      this._handleResizeFunc = function(width, height) {self._handleResize();};
      oj.DomUtils.addResizeListener(elem[0], this._handleResizeFunc, 
        _RESIZE_LISTENER_COLLAPSE_EVENT_TIMEOUT);
    }
  },

  /** 
   * Destroy the masonryLayout.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @override
   * @protected
   */
  _destroy: function() // Override of protected base class method.  
  {
    this._clearMenu();
    
    var elem = this.element;
    oj.DomUtils.removeResizeListener(elem[0], this._handleResizeFunc);
    this._handleResizeFunc = null;
    
    //restore tiles to their original order and delete stored indices
    this._restoreTilesOriginalOrder();
    var children = this._getTileElements();
    var numChildren = children.length;
    for (var i = 0; i < numChildren; i++)
    {
      var child = children[i];
      delete child._jetDataMasonryOriginalOrder;
    }
    
    this._destroyMLCommon();
    
    elem.removeClass("oj-masonrylayout oj-component");
    var options = this.options;
    if (options.reorderHandle)
    {
      this._tearDownReorderHandles();
    }
    
    //remove the oj-masonrylayout-tile style class from child tiles
    var tiles = elem.children("." + _OJ_MASONRYLAYOUT_TILE);
    tiles.removeClass(_OJ_MASONRYLAYOUT_TILE);
    
    this._handleDragStartFunc = null;
    this._handleDragEnterFunc = null;
    this._handleDragOverFunc = null;
    this._handleDragLeaveFunc = null;
    this._handleDragEndFunc = null;
    this._handleDropFunc = null;
    
    this._showTileOnEndFunc = null;
    this._hideTileOnEndFunc = null;
    this._startAnimationFunc = null;
    this._layoutOnEndFunc = null;
    this._layoutCycleOnStartFunc = null;
    this._layoutCycleOnEndFunc = null;
    
    this._arTilesToInsert = null;
    
    this._super();
  },

  /** 
   * Set an option on the masonryLayout.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @override
   * @protected
   */
  _setOption: function(key, value, flags) // Override of protected base class method.
                                   // Method name needn't be quoted since is in externs.js.
  {
    var elem = this.element;
    var options = this.options;
    var bReorderHandle = false;
    switch (key) 
    {
      case "reorderHandle":
        this._tearDownReorderHandles();
        bReorderHandle = true;
        break;
      case "disabled":
        //log warning message when "disabled" property set
        oj.Logger.warn(_WARNING_DISABLED_OPTION);
        break;
      case "contextMenu":
        if (!oj.DomUtils.isTouchSupported())
        {
          this._clearMenu();
          if (value)
            this._initMenu(value);
        }
        break;
    }
    this._super(key, value, flags);
    if (bReorderHandle && value)
    {
      this._setupReorderHandles();
    }
  },
  
  /**
   * Resize a tile in the masonryLayout.
   * @param {string} selector Selector identifying the tile to resize.
   * @param {string} sizeStyleClass New size style class to apply to the tile.
   * @return {void}
   * 
   * @expose 
   * @memberof oj.ojMasonryLayout
   * @instance
   * @ojshortdesc Resize a tile.
   * 
   * @example <caption>Invoke the <code class="prettyprint">resizeTile</code> method:</caption>
   * myMasonryLayout.resizeTile('#tileSelector', 'oj-masonrylayout-tile-2x1');
   */
  resizeTile: function(selector, sizeStyleClass)
  {
    var mlCommon = this._mlCommon;
    //FIX : if we're already in a layout cycle, finish it immediately
    //before processing the resize
    if (mlCommon.isInLayoutCycle())
      mlCommon.finishLayoutCycle();

    var jqElem = $(selector);
    var elem = jqElem[0];
    var prevSizeStyleClass = _getSizeStyleClassName(elem);
    
    //FIX : if the style class isn't changing, don't do anything;
    //this prevents the layout from preparing for a resize and then waiting for
    //the end of a resize animation that won't happen
    //FIX : throw error if style class isn't changing
    if (sizeStyleClass == prevSizeStyleClass)
    {
      throw new Error("JET MasonryLayout: Unable to resize child " + selector +
                      " to style class " + sizeStyleClass + " because " + 
                      sizeStyleClass + " is already applied.");
    }
    
    //fire beforeResize event
    var eventData = {
      /** @expose */
      tile: jqElem,
      /** @expose */
      previousSizeStyleClass: prevSizeStyleClass,
      /** @expose */
      sizeStyleClass: sizeStyleClass
    };
    var bSuccess = this._trigger("beforeResize", null, eventData);
    
    if (bSuccess !== false)
    {
      if (!this._arResizingTiles)
        this._arResizingTiles = [];
      var arResizingTiles = this._arResizingTiles;
      arResizingTiles.push(elem, prevSizeStyleClass, sizeStyleClass);

      mlCommon.resizeTile(selector, sizeStyleClass);
    }
  },
  
  /**
   * Insert a tile into the masonryLayout.
   * @param {string} selector Selector identifying the tile to insert.  The tile
   *        does not need to be a child of the 
   *        <code class="prettyprint">oj-masonry-layout</code> when this method is
   *        called.  This method will reparent the tile to the 
   *        <code class="prettyprint">oj-masonry-layout</code>.
   * @param {number} index The 0-based index into the set of rendered
   *           <code class="prettyprint">oj-masonry-layout</code> children where 
   *           the tile will be inserted.
   * @return {void}
   * 
   * @expose 
   * @memberof oj.ojMasonryLayout
   * @instance
   * @ojshortdesc Insert a tile.
   * 
   * @example <caption>Invoke the <code class="prettyprint">insertTile</code> method:</caption>
   * myMasonryLayout.insertTile('#tileSelector', 2);
   */
  insertTile: function(selector, index)
  {
    var mlCommon = this._mlCommon;
    //FIX : if we're already in a layout cycle, finish it immediately
    //before processing the insert
    if (mlCommon.isInLayoutCycle())
      mlCommon.finishLayoutCycle();

    if (isNaN(index))
      index = -1;
    
    var jqElem = $(selector);
    var elem = jqElem[0];
    
    //add the oj-masonrylayout-tile style class to the inserted tile
    jqElem.addClass(_OJ_MASONRYLAYOUT_TILE);
    
    //fire beforeInsert event
    var eventData = {
      /** @expose */
      tile: jqElem,
      /** @expose */
      index: index
    };
    var bSuccess = this._trigger("beforeInsert", null, eventData);
    
    if (bSuccess !== false)
    {
      elem._jetDataTileInsertIndex = index;
      
      //only need to initially hide inserts if they will be animated later,
      //otherwise we can just leave them shown here
      if (mlCommon.isAnimationEnabled())
      {
        //immediately hide insert in order to show it later with a transition
        jqElem.addClass(_OJ_MASONRYLAYOUT_TILE_TRANSITION_SHOW_FROM_CLASS);
      }

      var style = elem.style;
      style.top = "-1px";
      //FIX : in RTL, position tiles using "right" instead of "left"
      if (this._bRTL)
        style.right = "-1px";
      else
        style.left = "-1px";
      
      this._insertTileOriginalOrder(elem, index);
      mlCommon.insertTileDomElem(elem, index);
      //notify the element that it's been attached to the DOM
      oj.Components.subtreeAttached(elem);

      if (!this._arTilesToInsert)
        this._arTilesToInsert = [];
      var arTilesToInsert = this._arTilesToInsert;
      arTilesToInsert.push(selector);
    }
  },
  
  /**
   * Remove a tile from the masonryLayout.
   * @param {string} selector Selector identifying the tile to remove.
   * @return {void}
   * 
   * @expose 
   * @memberof oj.ojMasonryLayout
   * @instance
   * @ojshortdesc Remove a tile.
   * 
   * @example <caption>Invoke the <code class="prettyprint">removeTile</code> method:</caption>
   * myMasonryLayout.removeTile('#tileSelector');
   */
  removeTile: function(selector)
  {
    var mlCommon = this._mlCommon;
    //FIX : if we're already in a layout cycle, finish it immediately
    //before processing the delete
    if (mlCommon.isInLayoutCycle())
      mlCommon.finishLayoutCycle();

    //if the infolet being deleted contains the focus, remember that so we can 
    //try to restore focus to the next infolet after layout
    var jqInfolet = $(selector);
    var infolet = jqInfolet[0];
    if (oj.FocusUtils.containsFocus(infolet))
    {
      var children = this._getTileElements(true);
      var index = children.indexOf(infolet);
      //save the previous infolet, because this one is being deleted
      if (index > 0)
        this._deletingTileWithFocusPrev = children[index - 1];
    }
    
    //fire beforeRemove event
    var eventData = {
      /** @expose */
      tile: jqInfolet
    };
    var bSuccess = this._trigger("beforeRemove", null, eventData);
    
    if (bSuccess !== false)
      mlCommon.hideTile(selector);
  },
  
  /** 
   * Handle a component resize.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleResize: function()
  {
    if (!this._bDragging && !this._inLayoutCycle)
    {
      var mlCommon = this._mlCommon;
      mlCommon.resizeNotify();
    }
  },
  
  /** 
   * Callback to run after a tile is shown.
   * @param {Object} elem Tile that was shown.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _showTileOnEnd: function(elem)
  {
    var jqElem = $(elem);
    var index = elem._jetDataTileInsertIndex;
    delete elem._jetDataTileInsertIndex;
    
    //FIX : setup the reorder handle for the inserted tile
    var options = this.options;
    if (options.reorderHandle)
      this._setupReorderHandlesForElem(jqElem);
    
    //fire insert event
    var eventData = {
      /** @expose */
      tile: jqElem,
      /** @expose */
      index: index
    };
    this._eventsToFire.push("insert", eventData);
  },
  
  /** 
   * Callback to run after a tile is hidden.
   * @param {!Element} elem Tile that was hidden.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _hideTileOnEnd: function(elem)
  {
    var jqElem = $(elem);
    
    //FIX : tear down the reorder handle for the removed tile
    var options = this.options;
    if (options.reorderHandle)
      this._tearDownReorderHandlesForElem(jqElem);
    
    //remove the oj-masonrylayout-tile style class from the hidden tile
    jqElem.removeClass(_OJ_MASONRYLAYOUT_TILE);
    
    //notify the element that it's been detached from the DOM BEFORE actually
    //detaching it so that components can save state
    oj.Components.subtreeDetached(elem);
    
    var parentNode = elem.parentNode;
    parentNode.removeChild(elem);
    //remove the tile from the original order
    this._removeTileOriginalOrder(elem);
    
    //fire remove event
    var eventData = {
      /** @expose */
      tile: jqElem
    };
    this._eventsToFire.push("remove", eventData);
  },

  /** 
   * Callback to run to animate a tile.
   * @param {!Element} elem Tile to be animated.
   * @param {!string} action Action that starts animation.
   * @return {Promise} A promise that resolves when the animation ends.
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _startAnimation: function(elem, action)
  {
    var defaults = (oj.ThemeUtils.parseJSONFromFontFamily('oj-masonrylayout-option-defaults') || {})["animation"]; 
    var animation = defaults ? defaults[action] : null;
    
    return oj.AnimationUtils.startAnimation(elem, action, animation, this);
  },

  /** 
   * Callback to run after layout is done.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _layoutOnEnd: function()
  {
    if (this._arTilesToInsert)
    {
      var mlCommon = this._mlCommon;
      var arTilesToInsert = this._arTilesToInsert;
      for (var i = 0; i < arTilesToInsert.length; i++)
      {
        var selector = arTilesToInsert[i];
        mlCommon.showTile(selector);
      }
      this._arTilesToInsert = null;
    }
    
    if (this._arResizingTiles)
    {
      var arResizingTiles = this._arResizingTiles;
      for (var i = 0; i < arResizingTiles.length; i += 3)
      {
        var elem = arResizingTiles[i];
        var prevSizeStyleClass = arResizingTiles[i + 1];
        var sizeStyleClass = arResizingTiles[i + 2];
        
        //fire resize event
        var eventData = {
          /** @expose */
          tile: $(elem),
          /** @expose */
          previousSizeStyleClass: prevSizeStyleClass,
          /** @expose */
          sizeStyleClass: sizeStyleClass
        };
        this._eventsToFire.push("resize", eventData);
      }
      this._arResizingTiles = null;
    }
    
    if (this._bDragging)
    {
      if (this._bDragMoveTransition)
      {
        this._handleDragMoveTransitionEnd();
      }
      else if (this._bDragEndTransition)
      {
        this._handleDragEndTransitionEnd();
      }
    }
  },
  
  /**
   * Callback to run before whole layout cycle starts.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _layoutCycleOnStart: function()
  {
    this._inLayoutCycle = true;
    this._layoutStartActiveDomElem = null;
    
    //store events to fire when the layout cycle ends
    this._eventsToFire = [];

    var activeDomElem = document.activeElement;
    //if the currently focused element is a child of the  masonryLayout, 
    //remember that so we can try to restore focus after layout, in case the 
    //child DOM order changes
    if (activeDomElem)
    {
      if (oj.DomUtils.isAncestor(this.element[0], activeDomElem))
        this._layoutStartActiveDomElem = activeDomElem;
    }
  },

  /**
   * Callback to run after whole layout cycle is done.
   * @param {function():void} resolveBusyState Resolve function for busy state.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _layoutCycleOnEnd: function(resolveBusyState)
  {
    this._inLayoutCycle = false;
    
    var elem = this.element[0];
    //if we saved the focus element before layout, try to restore focus now
    if (this._layoutStartActiveDomElem)
    {
      var layoutStartActiveDomElem = this._layoutStartActiveDomElem;
      this._layoutStartActiveDomElem = null;
      //after deleting a tile, focus the tile that took its position in the layout
      if (this._deletingTileWithFocusPrev)
      {
        var tile = this._deletingTileWithFocusPrev;
        this._deletingTileWithFocusPrev = null;
        if (tile && oj.DomUtils.isAncestor(elem, tile))
        {
          var children = this._getTileElements(elem, true);
          var index = children.indexOf(tile);
          //because we saved the id of the prev tile, try to restore focus
          //to the tile after it, which would have been where the deleted
          //tile was positioned
          if (index >= 0 && index < children.length - 1)
            oj.FocusUtils.focusFirstTabStop(children[index + 1]);
          else
            oj.FocusUtils.focusFirstTabStop(tile);
        }
      }
      else
      {
        //restore focus to the element that had focus before layout, or to
        //the first tabstop in the masonryLayout
        if (oj.DomUtils.isAncestor(elem, layoutStartActiveDomElem))
          oj.FocusUtils.focusElement(layoutStartActiveDomElem);
        else
          oj.FocusUtils.focusFirstTabStop(elem);
      }
    }
    
    //FIX : resolve busy state after laying out, but BEFORE firing events so that when
    //an app gets an event it doesn't then have to wait for the busy context to be resolved
    if (resolveBusyState)
      resolveBusyState();
    
    //fire stored events at end of layout cycle
    if (this._eventsToFire)
    {
      var eventsToFire = this._eventsToFire;
      this._eventsToFire = null;
      
      for (var i = 0; i < eventsToFire.length; i += 2)
      {
        var eventType = eventsToFire[i];
        var eventData = eventsToFire[i + 1];
        this._trigger(eventType, null, eventData);
      }
    }
  },

  /** 
   * Destroy the MasonryLayoutCommon.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _destroyMLCommon: function()
  {
    var mlCommon = this._mlCommon;
    if (mlCommon)
      mlCommon.destroy();
    this._mlCommon = null;
  },

  /** 
   * Determine whether the masonryLayout can calculate sizes (when it is 
   * attached to the page DOM and not hidden).
   * @returns {boolean} true if sizes can be calculated, false if not
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _canCalculateSizes: function() 
  {
    var div = document.createElement("div");
    var style = div.style;
    style.width = "10px";
    style.height = "10px";
    var elem = this.element[0];
    elem.appendChild(div); // @HtmlUpdateOK
    var bCanCalcSizes = false;
    try
    {
      bCanCalcSizes = div.offsetWidth > 0 && div.offsetHeight > 0;
    }
    catch (e)
    {
      //do nothing
    }
    elem.removeChild(div);
    return bCanCalcSizes;
  },
  
  /**
   * Get the child tile elements.
   * @param {boolean} excludeDropSite True to exclude the dropsite, false to 
   *        include it.
   * @returns {Array} Array of child tile elements.
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _getTileElements: function(excludeDropSite)
  {
    var elem = this.element;
    var children = elem.children(_TILE_SELECTOR);
    var numChildren = children.length;
    var arChildren = [];
    for (var i = 0; i < numChildren; i++)
    {
      var child = children[i];
      if (!excludeDropSite || (excludeDropSite && child !== this._dropSite))
      {
        var style = child.style;
        if (style.visibility !== _HIDDEN && style.display !== _NONE)
          arChildren.push(child);
      }
    }
    return arChildren;
  },
  
  /**
   * Save the original DOM order of the tiles.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _saveTilesOriginalOrder: function()
  {
    var arTiles = this._getTileElements();
    if (arTiles)
    {
      for (var i = 0; i < arTiles.length; i++)
      {
        var tile = arTiles[i];
        //only assign original index to tile if it hasn't already been assigned,
        //so that repeated calls to _setup(true) won't overwrite the original
        //values with current layout order values
        if (!tile._jetDataMasonryOriginalOrder)
        {
          //make original order index 1-based so that we can easily check for
          //the existence of the attribute on an element
          //(e.g. "if (tile._jetDataMasonryOriginalOrder)")
          tile._jetDataMasonryOriginalOrder = i + 1;
        }
      }
    }
  },
  
  /**
   * Check that all tiles have an assigned index in the original DOM 
   * order, in case new tiles have been inserted outside of the 
   * insertTile API and the masonryLayout has been refreshed.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _checkTilesOriginalOrder: function()
  {
    var arTiles = this._getTileElements();
    if (arTiles)
    {
      for (var i = 0; i < arTiles.length; i++)
      {
        var tile = arTiles[i];
        //insert the tile at its current index
        if (!tile._jetDataMasonryOriginalOrder)
          this._insertTileOriginalOrder(tile, i);
      }
    }
  },
  
  /**
   * Restore the original DOM order of the tiles.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _restoreTilesOriginalOrder: function()
  {
    var children = this._getTileElements();
    var sortedChildren = this._getTileElements();
    _sortTilesOriginalOrder(sortedChildren);
    for (var i = 0; i < children.length; i++)
    {
      var child = children[i];
      var sortedChild = sortedChildren[i];
      //if the tiles are different, need to rearrange the DOM
      if (child != sortedChild)
      {
        //notify the element that it's been detached from the DOM BEFORE actually
        //detaching it so that components can save state
        oj.Components.subtreeDetached(sortedChild);
        //insert the tile from the layout order into the DOM at this index
        var parentNode = child.parentNode;
        parentNode.insertBefore(sortedChild, child); // @HTMLUpdateOK
        //notify the element that it's been attached to the DOM
        oj.Components.subtreeAttached(sortedChild);

        //rearrange the array of children to match the DOM reorder above
        var sortedChildIndex = children.indexOf(sortedChild);
        if (sortedChildIndex > i)
        {
          children.splice(sortedChildIndex, 1);
          children.splice(i, 0, sortedChild);
        }
      }
    }
  },
  
  /**
   * Insert a tile into the original DOM order at the given index.
   * @param {Object} insertedTile Tile to insert.
   * @param {number} index Index at which to insert.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _insertTileOriginalOrder: function(insertedTile, index)
  {
    var arTiles = this._getTileElements();
    if (index < 0)
      index = arTiles.length;
    if (arTiles)
    {
      for (var i = 0; i < arTiles.length; i++)
      {
        var tile = arTiles[i];
        //increment the index of each shifted tile
        if (tile._jetDataMasonryOriginalOrder && tile._jetDataMasonryOriginalOrder >= index + 1)
          tile._jetDataMasonryOriginalOrder++;
      }
    }
    //save the index on the inserted tile AFTER the above loop so that its index 
    //didn't get incremented
    insertedTile._jetDataMasonryOriginalOrder = index + 1;
  },
  
  /**
   * Remove a tile from the original DOM order.
   * @param {Object} removedTile Tile to remove.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _removeTileOriginalOrder: function(removedTile)
  {
    if (removedTile._jetDataMasonryOriginalOrder)
    {
      var arTiles = this._getTileElements();
      if (arTiles)
      {
        for (var i = 0; i < arTiles.length; i++)
        {
          var tile = arTiles[i];
          //decrement the index of each shifted tile
          if (tile._jetDataMasonryOriginalOrder && 
              tile._jetDataMasonryOriginalOrder > removedTile._jetDataMasonryOriginalOrder)
            tile._jetDataMasonryOriginalOrder--;
        }
      }
      //delete the stored index from the tile
      delete removedTile._jetDataMasonryOriginalOrder;
    }
  },

  /**
   * Apply the "oj-masonrylayout-tile" style class to any child tile that is missing it.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _applyTileStyleClass: function()
  {
    var elem = this.element;
    var children = elem.children();
    children = children.not("." + _OJ_MASONRYLAYOUT_TILE);
    children = children.filter(function(index, childElem) 
    {
      var childClasses = childElem.className;
      return childClasses.match(_OJ_MASONRYLAYOUT_TILE_SIZE_CLASS_PATTERN);
    })
    children.addClass(_OJ_MASONRYLAYOUT_TILE);
  },
  
  // start functions for context menu reordering ///////////////////////////////
  
  /**
   * Initialize the context menu.  This is called on startup, or on option
   * "contextMenu" change.
   * @param {Object=} newVal True if called because of an option change.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _initMenu: function(newVal)
  {
    var t = null;
    var menu = this._GetContextMenu();

    // check for contextmenu attribute on the root element
    if ((!newVal) && (!menu) && !this._IsCustomElement())
    {
      menu = this.element.attr("contextmenu");
      if (menu)
        this.options["contextMenu"] = "#" + menu;
    }

    if ((!newVal) && (!menu))
      return;

    menu = newVal || menu;
    t = $.type(menu);
    if (t == "function") {
      try
      {
        // call user's method to get the context menu
        menu = menu();
      }
      catch (e)
      {
        menu = null;
      }
      t = $.type(menu);
    }

    if (!menu)
      return;

    // get the user's <ul> list  
    var $m = $(menu); 
    if ($m)
    {
      // ensure it's not visible
      $m.css("display", _NONE);
      var dm = this._menu;
      if (!dm)
        return;

      dm.$container = $m;
      // have a context menu
      dm.usermenu   = true;
    }

    // if we have a context menu
    if (this._menu.usermenu)
    {
      // and it is being changed
      if (newVal)
      {
        // complete menu creation/attachment
        this._applyMenu();
      }
    }

    //  If not a new val from options, Menu will be noted at the end of initialization in _start()
  },

  /**
   * Replace built in shortcut reorder context menu items with real items and
   * add listeners to the context menu.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _applyMenu: function()
  {
    if (!this._menu || !this._menu.usermenu || !this.options.reorderHandle)
      return;

    // Add our listeners so that we can handle build-in cut/paste, etc
    var $menuContainer = this._menu.$container;
    var self = this;
    
    // Check if the menu container tag name is 'OJ-MENU' for custom element context menu
    if ($menuContainer[0].tagName === "OJ-MENU")
    {
      // For custom element context menu, attach action listener to handle selection
      $menuContainer[0].addEventListener("ojAction", this._handleContextMenuSelect.bind(this));
    }
    else
    {
      // For data-bind widget context menu, attach selection listener to handle selection
      $menuContainer.on("ojselect", this._handleContextMenuSelect.bind(this));
    }

    // If there are any ojMasonryLayout built in menu item ids, construct the menu items
    var listItems = $menuContainer.find("[data-oj-command]");
    var bChanged  = false;

    listItems.each(function()
      {
        if ($(this).children('a').length === 0)
        {
          var command = $(this).attr('data-oj-command').slice("oj-masonrylayout-".length);
          var tagName = ($menuContainer[0].tagName === "OJ-MENU") ? this.tagName : "li";
          var newMenuItem = self._buildContextMenuItem(command, tagName);
          
          $(this).attr('id', newMenuItem.attr('id'));
          if (this.tagName === "OJ-OPTION")
          {
            // Just update the content of OJ-OPTION and don't replace that with menu item,
            // because replacing OJ-OPTION would trigger a detach/attach of custom element
            this.innerHTML = newMenuItem.get(0).innerHTML;
            $(this).attr('data-oj-command', newMenuItem.attr('data-oj-command'));
          }
          else
          {
            // replace the list item with the new menu item
            newMenuItem.get(0).className = $(this).get(0).className;
            $(this).replaceWith(newMenuItem); //@HTMLUpdateOK                        
          }
          bChanged = true;
        }
      });

    if (bChanged) {
      if ($menuContainer[0].tagName === "OJ-MENU")
      {
        $menuContainer[0].refresh();
      }
      else
      {
        $menuContainer.ojMenu('refresh');
      }
    }
    
    this._menu.$elemCut = $menuContainer.find("#" + _OJMASONRYLAYOUTCUT);
    this._menu.$elemPasteBefore = $menuContainer.find("#" + _OJMASONRYLAYOUTPASTEBEFORE);
    this._menu.$elemPasteAfter = $menuContainer.find("#" + _OJMASONRYLAYOUTPASTEAFTER);
  },

  /**
   * Clear out any contextMenu data.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _clearMenu: function()
  {
    var menu = this._menu;
    if (menu && menu.usermenu)
    {
      menu.usermenu = false;
      if (menu.$container[0].tagName === "OJ-MENU")
      {
        menu.$container[0].removeEventListener("ojAction");
      }
      else
      {
        menu.$container.off("ojselect");
      }
      menu.$container = null;
    }
  },
  
  /**
   * Prepare the context menu before it's opened.
   * @param {Object} e jQuery event object.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _prepareContextMenuBeforeOpen: function(e)
  {
    var elem = this.element;
    
    //FIX : only allow reordering if context menu was 
    //launched from reorder handle
    var target = e.originalEvent.target;
    var options = this.options;
    var reorderHandle = options.reorderHandle;
    var reorderAllowed = false;
    if (reorderHandle)
    {
      var closestHandle = $(target).closest(reorderHandle);
      if (closestHandle && closestHandle.length > 0)
        reorderAllowed = true;
    }
    
    var tile = _findContainingTile(target, elem[0]);
    this._menu.tile = tile;

    //set menu items disabled state
    if (this._menu.usermenu)
    {
      var cutTile = this._menu.cutTile;
      var bRefreshMenu = false;
      
      var elemCut = this._menu.$elemCut;
      if (elemCut)
      {
        //FIX : only allow reordering if context menu was 
        //launched from reorder handle
        var cutHidden = elemCut.hasClass("oj-helper-hidden");
        if (!reorderAllowed && !cutHidden)
        {
          elemCut.addClass("oj-helper-hidden");
          bRefreshMenu = true;
        }
        else if (reorderAllowed && cutHidden)
        {
          elemCut.removeClass("oj-helper-hidden");
          bRefreshMenu = true;
        }
        
        var cutDisabled = elemCut.hasClass(_OJ_DISABLED);
        var bDisable = false;
        //disable "cut" if this tile has already been cut
        if (cutTile && tile === cutTile)
          bDisable = true;
        
        if (bDisable && !cutDisabled)
        {
          elemCut.addClass(_OJ_DISABLED);
          bRefreshMenu = true;
        }
        else if (!bDisable && cutDisabled)
        {
          elemCut.removeClass(_OJ_DISABLED);
          bRefreshMenu = true;
        }
      }
      
      var elemPasteBefore = this._menu.$elemPasteBefore;
      if (elemPasteBefore)
      {
        //FIX : only allow reordering if context menu was 
        //launched from reorder handle
        var pasteBeforeHidden = elemPasteBefore.hasClass("oj-helper-hidden");
        if (!reorderAllowed && !pasteBeforeHidden)
        {
          elemPasteBefore.addClass("oj-helper-hidden");
          bRefreshMenu = true;
        }
        else if (reorderAllowed && pasteBeforeHidden)
        {
          elemPasteBefore.removeClass("oj-helper-hidden");
          bRefreshMenu = true;
        }
        
        var pasteBeforeDisabled = elemPasteBefore.hasClass(_OJ_DISABLED);
        var bDisable = false;
        //disable "pasteBefore" if this is the tile that was cut or if the cut 
        //tile is the one before this tile
        if (!cutTile ||
            (tile === cutTile || tile === _getNextElement(cutTile)))
        {
          bDisable = true;
        }
        
        if (bDisable && !pasteBeforeDisabled)
        {
          elemPasteBefore.addClass(_OJ_DISABLED);
          bRefreshMenu = true;
        }
        else if (!bDisable && pasteBeforeDisabled)
        {
          elemPasteBefore.removeClass(_OJ_DISABLED);
          bRefreshMenu = true;
        }
      }
      
      var elemPasteAfter = this._menu.$elemPasteAfter;
      if (elemPasteAfter)
      {
        //FIX : only allow reordering if context menu was 
        //launched from reorder handle
        var pasteAfterHidden = elemPasteAfter.hasClass("oj-helper-hidden");
        if (!reorderAllowed && !pasteAfterHidden)
        {
          elemPasteAfter.addClass("oj-helper-hidden");
          bRefreshMenu = true;
        }
        else if (reorderAllowed && pasteAfterHidden)
        {
          elemPasteAfter.removeClass("oj-helper-hidden");
          bRefreshMenu = true;
        }
        
        var pasteAfterDisabled = elemPasteAfter.hasClass(_OJ_DISABLED);
        var bDisable = false;
        //disable "pasteAfter" if this is the tile that was cut or if the cut 
        //tile is the one after this tile
        if (!cutTile ||
            (cutTile === tile || cutTile === _getNextElement(tile)))
        {
          bDisable = true;
        }
        
        if (bDisable && !pasteAfterDisabled)
        {
          elemPasteAfter.addClass(_OJ_DISABLED);
          bRefreshMenu = true;
        }
        else if (!bDisable && pasteAfterDisabled)
        {
          elemPasteAfter.removeClass(_OJ_DISABLED);
          bRefreshMenu = true;
        }
      }

      if (bRefreshMenu) 
      {
        var $menuContainer = this._menu.$container;
        
        if ($menuContainer[0].tagName === "OJ-MENU")
        {
          $menuContainer[0].refresh();
        }
        else
        {
          $menuContainer.ojMenu('refresh');
        }
      }
    }
  },

  /**
   * Build a context menu item for a cut/paste command.
   * @param {string} command Command to execute for item.
   * @param {string} tagName to use to create the menu item
   * @returns {jQuery} jQuery list item object.
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _buildContextMenuItem: function(command, tagName)
  {
    var id = _MENU_CMD_MAP[command];
    var item = $(document.createElement(tagName));
    item.attr('id', id);
    item.attr('data-oj-command', command);
    item.append(this._buildContextMenuLabel(command, tagName === "OJ-OPTION")); //@HTMLUpdateOK
    return item;
  },
  
  /**
   * Builds a context menu label by looking up command translation
   * @param {string} command the string to look up translation for
   * @param {boolean=} useOjOption whether oj-option tag should be used
   * @return {jQuery|string} a jQuery object with HTML containing a label
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _buildContextMenuLabel: function(command, useOjOption)
  {
    // convert to the translation key convention
    var transKey = _MENU_TRANSLATION_MAP[command];
    var textNode = document.createTextNode(this.getTranslatedString(transKey));
    
    // for custom elements, no <a> tag is required
    if (useOjOption)
      return textNode;
    else
      return $('<a href="#"></a>').append(textNode); //@HTMLUpdateOK
  },

  /**
   * Execute a "cut" command from the context menu.
   * @param {Object} obj Tile element to cut.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _menuCut: function(obj)
  {
    if (obj)
      this._menu.cutTile = obj;
  },

  /**
   * Execute a "paste" command from the context menu.
   * @param {Object} obj Tile element at paste location.
   * @param {boolean} pasteBefore True to paste before the given tile, false to
   *        paste after the given tile.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _menuPaste: function(obj, pasteBefore)
  { 
    if (obj && this._menu.cutTile)
    {
      var cutTile = this._menu.cutTile;
      this._menu.cutTile = false;
      this._doPaste(cutTile, obj, pasteBefore);
    }
  },
  
  /**
   * Paste the cut tile before or after the given paste tile.
   * @param {Object} cutTile Tile element to paste.
   * @param {Object} pasteTile Tile element at paste location.
   * @param {boolean} pasteBefore True to paste before the given tile, false to
   *        paste after the given tile.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _doPaste: function (cutTile, pasteTile, pasteBefore)
  {
    var fromIndex = cutTile._jetDataMasonryOriginalOrder - 1;
    var jqCutTile = $(cutTile);
    
    //fire beforeReorder event
    var beforeEventData = {
      /** @expose */
      tile: jqCutTile,
      /** @expose */
      fromIndex: fromIndex
    };
    var bSuccess = this._trigger("beforeReorder", null, beforeEventData);
    
    if (bSuccess !== false)
    {
      //remove the cutTile from the original ordering before getting the toIndex
      this._removeTileOriginalOrder(cutTile);
      var toIndex = pasteTile._jetDataMasonryOriginalOrder - 1;
      if (!pasteBefore)
        toIndex++;

      var elem = this.element[0];
      if (pasteBefore)
      {
        this._insertTileOriginalOrder(cutTile, toIndex);
        elem.insertBefore(cutTile, pasteTile); // @HTMLUpdateOK
      }
      else
      {
        var nextElem = _getNextElement(pasteTile);
        this._insertTileOriginalOrder(cutTile, toIndex);
        elem.insertBefore(cutTile, nextElem); // @HTMLUpdateOK
      }

      var mlCommon = this._mlCommon;
      mlCommon.setup(true);

      //fire reorder event
      var eventData = {
        /** @expose */
        tile: jqCutTile,
        /** @expose */
        fromIndex: fromIndex,
        /** @expose */
        toIndex: toIndex
      };
      this._trigger("reorder", null, eventData);
    }
  },

  /**
   * Handle a context menu select event.
   * @param {Event|CustomEvent} event event for context menu item selection
   * @param {Object=} ui an object containing the menu item that was selected
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleContextMenuSelect: function(event, ui)
  {
    // get the selected menu item from the ui object in the old data-bind syntax widget case,
    // or from the event target in the custom element case.
    var item = ui ? ui.item : $(event.target);
    var id = item.attr("id");

    if (id === _OJMASONRYLAYOUTCUT)
      this._menuCut(this._menu.tile);
    else if (id === _OJMASONRYLAYOUTPASTEBEFORE)
      this._menuPaste(this._menu.tile, true);
    else if (id === _OJMASONRYLAYOUTPASTEAFTER)
      this._menuPaste(this._menu.tile, false);
  },
  
  // end functions for context menu reordering /////////////////////////////////
  
  // start functions for drag-and-drop reordering //////////////////////////////
  
  /**
   * Get the index of the given child tile.
   * @param {Object} tile Tile for which to get the index.
   * @returns {number} Index of the tile.
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _getTileIndex: function(tile)
  {
    //index should be with respect to "real" children, not including the dropsite,
    //so exclude the dropsite from the children array
    var children = this._getTileElements(true);
    //need to sort the children because the index is with respect to the original order
    _sortTilesOriginalOrder(children);
    var numChildren = children.length;
    for (var i = 0; i < numChildren; i++)
    {
      if (children[i] === tile)
        return i;
    }
    return -1;
  },
  
  /**
   * Setup drag and drop on the reorder handles.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _setupReorderHandles: function()
  {
    var elem = this.element;
    var children = elem.children();
    this._setupReorderHandlesForElem(children);
    elem
      .on("dragstart" + this.reorderHandleEventNamespace, this._handleDragStartFunc)
      .on("dragenter" + this.reorderHandleEventNamespace, this._handleDragEnterFunc)
      .on("dragover" + this.reorderHandleEventNamespace, this._handleDragOverFunc)
      .on("dragleave" + this.reorderHandleEventNamespace, this._handleDragLeaveFunc)
      .on("dragend" + this.reorderHandleEventNamespace, this._handleDragEndFunc)
      .on("drop" + this.reorderHandleEventNamespace, this._handleDropFunc);
  },
  
  /**
   * Setup drag and drop on the reorder handles.
   * @param {jQuery} jqElem Elem for which to setup drag and drop on
   *        the reorder handles.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _setupReorderHandlesForElem: function(jqElem)
  {
    var options = this.options;
    //setup for tiles themselves
    var tiles = jqElem.filter(options.reorderHandle);
    tiles
      .attr(_DRAGGABLE, "true")
      .addClass(_OJ_DRAGGABLE);
    //setup for descendants of tiles
    var reorderHandles = jqElem.find(options.reorderHandle);
    reorderHandles
      .attr(_DRAGGABLE, "true")
      .addClass(_OJ_DRAGGABLE);
  },
  
  /**
   * Tear down drag and drop on the reorder handles.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _tearDownReorderHandles: function()
  {
    var elem = this.element;
    var children = elem.children();
    //tear down reorder handles on the children of the root elem in case
    //the root elem itself is draggable
    this._tearDownReorderHandlesForElem(children);
    elem.off(this.reorderHandleEventNamespace);
  },
  
  /**
   * Tear down drag and drop on the reorder handles.
   * @param {jQuery} jqElem Elem for which to tear down drag and drop on
   *        the reorder handles.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _tearDownReorderHandlesForElem: function(jqElem)
  {
    var options = this.options;
    //tear down for tiles themselves
    var tiles = jqElem.filter(options.reorderHandle);
    tiles
      .removeAttr(_DRAGGABLE)
      .removeClass(_OJ_DRAGGABLE);
    //tear down for descendants of tiles
    var reorderHandles = jqElem.find(options.reorderHandle);
    reorderHandles
      .removeAttr(_DRAGGABLE)
      .removeClass(_OJ_DRAGGABLE);
  },
  
  /**
   * Handle a dragstart event.
   * @param {Event} event jQuery Event
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDragStart: function(event)
  {
    var options = this.options;
    
    //FIX : if drag didn't start on reorder handle, don't
    //reorder tile
    if (options.reorderHandle)
    {
      var target = event.originalEvent.target;
      var closestHandle = $(target).closest(options.reorderHandle);
      if (!closestHandle || closestHandle.length < 1)
        return;
    }
    
    //don't try to start another drag while there's already one happening
    //(in case the previous transition is still running)
    if (options.reorderHandle && !this._bDragging)
    {
      var target = event.target;
      var elem = this.element;
      var tile = _findContainingTile(target, elem[0]);
      if (tile)
      {
        var index = this._getTileIndex(tile);
        tile._jetDataMasonryDragSourceIndex = index;
        
        //fire beforeReorder event
        var eventData = {
          /** @expose */
          tile: $(tile),
          /** @expose */
          fromIndex: index
        };
        var bSuccess = this._trigger("beforeReorder", null, eventData);
        
        if (bSuccess !== false)
        {
          var originalEvent = event.originalEvent;
          this._dragStart(tile, originalEvent.pageX, originalEvent.pageY, originalEvent.dataTransfer);
        }
      }
    }
  },
  
  /**
   * Handle a dragenter event.
   * @param {Event} event jQuery Event
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDragEnter: function(event)
  {
    //FIX : only process drag and drop event if masonryLayout
    //started the drag
    if (!this._bDragging)
      return;
    
    var originalEvent = event.originalEvent;
    var relatedTarget = originalEvent.relatedTarget;
    var elem = this.element[0];
    var enteringMasonryLayout = false;
    if (relatedTarget)
    {
      enteringMasonryLayout = (elem != relatedTarget) && !oj.DomUtils.isAncestor(elem, relatedTarget);
    }
    //chrome 40.0.2214.111 doesn't set a relatedTarget, so need to use other means
    //to figure out if we're re-entering the masonryLayout
    else if (this._dragLeftMasonryLayout)
    {
      var elemUnderPoint = document.elementFromPoint(originalEvent.clientX, originalEvent.clientY);
      enteringMasonryLayout = elemUnderPoint && 
                              (elemUnderPoint == elem ||
                               oj.DomUtils.isAncestor(elem, elemUnderPoint));
    }
    if (enteringMasonryLayout)
    {
      this._dragLeftMasonryLayout = false;
      if (!this._draggedTile)
      {
        var dataTransfer = originalEvent.dataTransfer;
        dataTransfer.dropEffect = 'none';
      }
      else if (this._dropSite)
      {
        $(this._dropSite).css("display", "");
        var mlCommon = this._mlCommon;
        mlCommon.setup(false, true);
      }
    }
  },
  
  /**
   * Handle a dragover event.
   * @param {Event} event jQuery Event
   * @returns {boolean} False
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDragOver: function(event)
  {
    //FIX : only process drag and drop event if masonryLayout
    //started the drag
    if (!this._bDragging)
      return false;
    
    var originalEvent = event.originalEvent;
    var dataTransfer = originalEvent.dataTransfer;
    dataTransfer.dropEffect = 'move';
    
    this._dragMove(originalEvent.pageX, originalEvent.clientX, originalEvent.clientY);

    //necessary to allow a drop
    event.preventDefault();
    return false;
  },
  
  /**
   * Handle a dragleave event.
   * @param {Event} event jQuery Event
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDragLeave: function(event)
  {
    //FIX : only process drag and drop event if masonryLayout
    //started the drag
    if (!this._bDragging)
      return;
    
    var originalEvent = event.originalEvent;
    var relatedTarget = originalEvent.relatedTarget;
    var elem = this.element[0];
    var leavingMasonryLayout = false;
    if (relatedTarget)
    {
      leavingMasonryLayout = (elem != relatedTarget) && !oj.DomUtils.isAncestor(elem, relatedTarget);
    }
    //chrome 40.0.2214.111 doesn't set a relatedTarget, so need to use other means
    //to figure out if we're exiting the masonryLayout
    else
    {
      var elemUnderPoint = document.elementFromPoint(originalEvent.clientX, originalEvent.clientY);
      leavingMasonryLayout = elemUnderPoint && 
                             elemUnderPoint != elem &&
                             !oj.DomUtils.isAncestor(elem, elemUnderPoint);
    }
    if (leavingMasonryLayout)
    {
      this._dragLeftMasonryLayout = true;
      if (this._dropSite)
      {
        $(this._dropSite).css("display", _NONE);
        var mlCommon = this._mlCommon;
        mlCommon.setup(false, true);
      }
    }
  },
  
  /**
   * Clear the timeout used to hide the tile at the start of a drag.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _clearDragStartHideTileTimeout: function()
  {
    if (this._dragStartHideTileTimeout)
    {
      clearTimeout(this._dragStartHideTileTimeout);
      this._dragStartHideTileTimeout = null;
      
      var draggedTile = this._draggedTile;
      if (draggedTile)
        $(draggedTile).removeClass("oj-drag");
    }
  },
  
  /**
   * Handle a dragend event.
   * @param {Event} event jQuery Event
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDragEnd: function(event)
  {
    //FIX : if the drag and drop was a very short duration,
    //the hide tile timeout may not have fired yet, so clear it now
    this._clearDragStartHideTileTimeout();
    
    //FIX : only process drag and drop event if masonryLayout
    //started the drag
    if (!this._bDragging)
      return;
    
    //only process the dragend if we're not currently processing a drop (for 
    //example if the mouse button was released outside of the masonryLayout)
    if (!this._bDropping)
    {
      var draggedTile = this._draggedTile;
      if (draggedTile && this._dropSite)
      {
        var dropSite = this._dropSite;
        var elem = this.element[0];
        //tile was not moved outside
        if (oj.DomUtils.isAncestor(elem, draggedTile))
        {
          $(dropSite).css("display", "");
          this._removeTileOriginalOrder(dropSite);
          var parent = dropSite.parentNode;
          parent.removeChild(dropSite);
          $(draggedTile).css("display", "");
          this._insertTileOriginalOrder(draggedTile, draggedTile._jetDataMasonryOriginalOrder - 1);
          var mlCommon = this._mlCommon;
          mlCommon.setup(false, true);
        }

        delete draggedTile._jetDataMasonryDragSourceIndex;
      }
      this._draggedTile = null;
      this._dropSite = null;
      this._prevTileUnderPoint = null;
      this._prevCellUnderPoint = null;
      this._bDragMoveTransition = false;
      this._bMouseMoved = false;
      this._dragOffset = null;
      this._bDragEndTransition = false;
      this._bDragging = false;
    }
  },
  
  /**
   * Handle a drop event.
   * @param {Event} event jQuery Event
   * @returns {boolean} False
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDrop: function(event)
  {
    //FIX : only process drag and drop event if masonryLayout
    //started the drag
    if (!this._bDragging)
      return false;
    
    //FIX : if the drop happens before the drag move layout 
    //animation is done, finish that layout cycle immediately before processing
    //the drop
    var mlCommon = this._mlCommon;
    if (mlCommon.isInLayoutCycle())
      mlCommon.finishLayoutCycle();
    
    //FIX : if the drag and drop was a very short duration,
    //the hide tile timeout may not have fired yet, so clear it now
    this._clearDragStartHideTileTimeout();
    
    var originalEvent = event.originalEvent;
    this._drop(this._draggedTile, originalEvent.pageX, originalEvent.pageY);
    
    //stop the browser from redirecting
    event.stopPropagation();
    return false;
  },

  /** 
   * Start a drag.
   * @param {!Element} tile Tile being dragged.
   * @param {number} pageX Page coordinate at which the drag starts.
   * @param {number} pageY Page coordinate at which the drag starts.
   * @param {Object} dataTransfer Drag and drop DataTransfer object.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _dragStart: function(tile, pageX, pageY, dataTransfer)
  {
    this._bDragging = true;
    this._bDropping = false;
    this._bMouseMoved = false;
    this._bDragStartTileHidden = false;
    this._draggedTile = tile;
    var elem = this.element[0];

    var sizeClass = _getSizeStyleClassName(tile);
    this._dropSite = document.createElement("div");
    var dropSite = this._dropSite;
    dropSite._jetDataMasonryOriginalOrder = tile._jetDataMasonryOriginalOrder;
    dropSite.className = sizeClass + " oj-drop " + _OJ_MASONRYLAYOUT_TILE;
    var style = dropSite.style;
    var tileStyle = tile.style;
    style.top = tileStyle.top;
    //FIX : in RTL, position tiles using "right" instead of "left"
    if (this._bRTL)
      style.right = tileStyle.right;
    else
      style.left = tileStyle.left;
    //need to get the relative position of the tile BEFORE replacing it with
    //the dropsite and adding it to the elemParent below
    var offset = _getRelativePosition(tile);
    elem.insertBefore(dropSite, tile); // @HTMLUpdateOK

    var dragOffset = {left: pageX - offset.left, top: pageY - offset.top};
    this._dragOffset = dragOffset;
    
    //add the oj-drag class BEFORE creating the drag feedback image
    $(tile).addClass("oj-drag");
    
    dataTransfer.effectAllowed = 'move';
    dataTransfer.setData('text/html', tile.outerHTML);
    dataTransfer.setDragImage(tile, dragOffset.left, dragOffset.top);
    
    //hide the original tile in a timeout AFTER creating the drag feedback image
    var self = this;
    this._dragStartHideTileTimeout = setTimeout(function() {
        self._bDragStartTileHidden = true;
        tileStyle.display = _NONE;
        $(tile).removeClass("oj-drag");
        self._dragStartHideTileTimeout = null;
        //notify the element that it's been hidden
        oj.Components.subtreeHidden(tile);
      }, 0);
  },

  /** 
   * Drag a tile.
   * @param {number} pageX Page coordinate of the drag move.
   * @param {number} clientX Client coordinate of the drag move.
   * @param {number} clientY Client coordinate of the drag move.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _dragMove: function(pageX, clientX, clientY)
  {
    this._bMouseMoved = true;
    
    //FIX : if the dragged tile hasn't been hidden yet, don't
    //process the drag move because it may layout the dragged tile, which
    //will leave a visual gap in the layout once the tile is hidden
    if (!this._bDragStartTileHidden)
      return;

    if (!this._bDragMoveTransition)
    {
      var elem = this.element[0];
      var elemUnderPoint = document.elementFromPoint(clientX, clientY);
      var tileUnderPoint = _findContainingTile(elemUnderPoint, elem);
      
      //FIX : only allow reordering if tile has reorder 
      //handle
      var tileHasReorderHandle = false;
      if (tileUnderPoint)
      {
        var $tileUnderPoint = $(tileUnderPoint);
        var options = this.options;
        var reorderHandle = options.reorderHandle;
        if (reorderHandle)
        {
          //FIX : the tile has a reorder handle if: 1) the tile 
          //itself is the reorder handle, or 2) the reorder handle is a
          //descendant of the tile
          if ($tileUnderPoint.is(reorderHandle))
          {
            tileHasReorderHandle = true;
          }
          else {
            var handleInTile = $tileUnderPoint.find(reorderHandle);
            tileHasReorderHandle = handleInTile && handleInTile.length > 0;
          }
        }
      }
      
      // FIX :  if the tileUnderPoint is a valid tile and is different than previous
      // tileUnderPoint, then the reorder animation have shifted a different tile under the mouse.
      // Check if the cellUnderPoint before the reorder animation is same as after the animation.
      // If they are same, then skip re-ordering of tiles to prevent recursion.
      // A different tile under the same cell during a drag move indicates that the reorder animation 
      // have shifted a different tile under the mouse.
      var bSkipReorder = false;
      if (tileUnderPoint == null || 
          tileUnderPoint === this._dropSite ||
          tileUnderPoint === this._draggedTile)
      {
        this._prevTileUnderPoint = null;
        this._prevCellUnderPoint = null;
      }
      else if (this._prevTileUnderPoint && this._prevCellUnderPoint &&
               this._prevTileUnderPoint != tileUnderPoint) {
        var cellUnderPoint = this._getCellPosition(clientX, clientY);
        if (cellUnderPoint &&
            cellUnderPoint.row == this._prevCellUnderPoint.row &&
            cellUnderPoint.col == this._prevCellUnderPoint.col) {
          bSkipReorder = true;
        }
      }

      if (tileUnderPoint && !bSkipReorder &&
          tileUnderPoint !== this._dropSite && 
          tileUnderPoint !== this._draggedTile &&
          tileHasReorderHandle)
      {
        var offset = _getRelativePosition(elem);
        var relX = pageX - offset.left;
        var oldNextSibling = _getNextElement(this._dropSite);
        var bRightSide = relX >= tileUnderPoint.offsetLeft + tileUnderPoint.offsetWidth * .5;
        this._removeTileOriginalOrder(this._dropSite);
        if ((bRightSide && !this._bRTL) || (!bRightSide && this._bRTL))
        {
          var nextElem = _getNextElement(tileUnderPoint);
          if (nextElem)
          {
            //insert before the nextElem
            this._insertTileOriginalOrder(this._dropSite, nextElem._jetDataMasonryOriginalOrder - 1);
            elem.insertBefore(this._dropSite, nextElem); // @HTMLUpdateOK
          }
          else
          {
            //append to the end
            this._insertTileOriginalOrder(this._dropSite, tileUnderPoint._jetDataMasonryOriginalOrder);
            elem.appendChild(this._dropSite); // @HtmlUpdateOK
          }
        }
        else
        {
          //insert before the tileUnderPoint
          this._insertTileOriginalOrder(this._dropSite, tileUnderPoint._jetDataMasonryOriginalOrder - 1);
          elem.insertBefore(this._dropSite, tileUnderPoint); // @HTMLUpdateOK
        }
        var newNextSibling = _getNextElement(this._dropSite);
        //only need to layout if the dropSite has actually changed order
        if (oldNextSibling !== newNextSibling)
        {
          //only record the start of a drag move transition if the layout 
          //actually changed
          var mlCommon = this._mlCommon;
          // save the tileUnderPoint and cellUnderPoint of the current layout
          // before the call to mlCommon.setup() that computes the relayout.
          var tmpTileUnderPoint = tileUnderPoint;
          var tmpCellUnderPoint = this._getCellPosition(clientX, clientY);
          
          this._bDragMoveTransition = mlCommon.setup(false, true);
          // if the tile has moved, save tmpTileUnderPoint and tmpCellUnderPoint
          if(this._bDragMoveTransition) {
            this._prevTileUnderPoint = tmpTileUnderPoint;
            this._prevCellUnderPoint = tmpCellUnderPoint;
          }
        }
      }
    }
  },

  /** 
   * Handle the end of a layout transition while dragging a tile.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDragMoveTransitionEnd: function()
  {
    this._bDragMoveTransition = false;
  },

  /** 
   * Drop a tile.
   * @param {!Element} tile Tile being dragged.
   * @param {number} pageX Page coordinate of the drop.
   * @param {number} pageY Page coordinate of the drop.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _drop: function(tile, pageX, pageY)
  {
    this._bDropping = true;
    var elem = this.element[0];
    var dropSite = this._dropSite;
    this._dropSite = null;
    this._prevTileUnderPoint = null;
    this._prevCellUnderPoint = null;
    //notify the element that it's been detached from the DOM BEFORE actually
    //detaching it so that components can save state
    oj.Components.subtreeDetached(tile);
    //FIX : calling DOM replaceChild() seems to break in IE11 on Win7, so instead call
    //equivalent insertBefore() and removeChild()
    elem.insertBefore(tile, dropSite);
    elem.removeChild(dropSite);
    //notify the element that it's been attached to the DOM
    oj.Components.subtreeAttached(tile);
    tile._jetDataMasonryOriginalOrder = dropSite._jetDataMasonryOriginalOrder;
    var style = tile.style;
    style.display = "";
    //notify the element that it's been shown
    oj.Components.subtreeShown(tile);

    //immediately transform the page absolute coords to ones relative to the 
    //ojmasonrylayout
    var offset = _getRelativePosition(elem);
    var dragOffset = this._dragOffset;
    style.top = (pageY - dragOffset.top - offset.top) + _PX;
    //FIX : in RTL, position tiles using "right" instead of "left"
    var newLeft = pageX - dragOffset.left - offset.left;
    if (this._bRTL)
    {
      style.right = (elem.offsetWidth - (newLeft + $(tile).outerWidth(true))) + _PX;
      style.left = "";
    }
    else
      style.left = newLeft + _PX;

    this._dragOffset = null;

    if (this._bMouseMoved)
    {
      //transition the tile to its permanent place in the layout
      var mlCommon = this._mlCommon;
      this._bDragEndTransition = mlCommon.setup(false, true);
    }
    else
    {
      this._handleDragEndTransitionEnd();
    }
  },

  /** 
   * Handle the end of a layout transition when a tile is dropped.
   * @return {void}
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _handleDragEndTransitionEnd: function()
  {
    this._bDragEndTransition = false;
    this._bDragging = false;
    this._bDropping = false;
    this._bMouseMoved = false;
    this._bDragStartTileHidden = false;
    
    var tile = this._draggedTile;
    this._draggedTile = null;
    var fromIndex = tile._jetDataMasonryDragSourceIndex;
    delete tile._jetDataMasonryDragSourceIndex;
    var toIndex = this._getTileIndex(tile);
    
    //fire reorder event
    var eventData = {
      /** @expose */
      tile: $(tile),
      /** @expose */
      fromIndex: fromIndex,
      /** @expose */
      toIndex: toIndex
    };
    this._trigger("reorder", null, eventData);
  },
  
  /** 
   * Get the grid cell position for the given client coordinates.
   * Returns an object with col, row properties, or null if it can't figure out.
   * This will return invalid cell positions if clientX/clientY is outside the component.
   * So this method should be called only for valid coordinates inside the component. 
   *
   * @param {number} clientX Client X coordinate
   * @param {number} clientY Client Y coordinate
   * @returns {Object|null} cell position
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _getCellPosition: function(clientX, clientY)
  {
    var mlCommon = this._mlCommon;
    var cellSize = mlCommon.getCellSize();
    var position = null;
    if (cellSize) {
      position = {col: parseInt(clientX / cellSize.w, 10), 
                  row: parseInt(clientY / cellSize.h, 10)};
    }
    return position;
  },
  
  // end functions for drag-and-drop reordering ////////////////////////////////
  
  /**
   * Add a busy state to the busy context.
   *
   * @param {string} description Additional information about busy state.
   * @returns {Function} Resolve function called by the registrant when the busy state completes.
   *          The resultant function will throw an error if the busy state is no longer registered.
   * @memberof oj.ojMasonryLayout
   * @instance
   * @private
   */
  _addBusyState: function(description)
  {
    var element = this.element;
    var context = oj.Context.getContext(element[0]);
    var busyContext = context.getBusyContext();
    
    var desc = "MasonryLayout";
    var id = element.attr("id");
    desc += " (id='" + id + "')";
    desc += ": " + description;
    
    var busyStateOptions = {"description" : desc};
    return busyContext.addBusyState(busyStateOptions);
  },
  
  // @inheritdoc
  getNodeBySubId: function(locator)
  {
    return this._super(locator);
  },
  
  // @inheritdoc
  getSubIdByNode: function(node)
  {
    return this._super(node);
  }
  
  // start API doc fragments /////////////////////////////////////////////////////
  
  /**
   * <p>This element does not expose any subIds.</p>
   *
   * @ojsubid None
   * @memberof oj.ojMasonryLayout
   */

  /**
   * <p>MasonryLayout is for layout only and does not directly support touch 
   * interaction. It is up to the application to provide touch support. 
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojMasonryLayout
   */

  /**
   * <p>MasonryLayout is for layout only and does not directly support keyboard 
   * interaction. It is up to the application to provide keyboard support. 
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojMasonryLayout
   */
  
  // end API doc fragments ///////////////////////////////////////////////////////
  
}); // end of oj.__registerWidget

// start static members and functions //////////////////////////////////////////

//_tmpVar is simply to trick the closure compiler so that it doesn't issue a
//warning about the below real vars declared using a single "var" statement:
//WARNING - declaration of multiple variables with shared type information
var _tmpVar;

var _PX        = "px",
    _HIDDEN    = "hidden",
    _NONE      = "none",
    _DRAGGABLE = "draggable",
    
    _OJ_DISABLED  = "oj-disabled",
    _OJ_DRAGGABLE = "oj-draggable",

    _OJ_MASONRYLAYOUT_TILE = "oj-masonrylayout-tile",
    _OJ_MASONRYLAYOUT_TILE_SIZE_CLASS_PATTERN = /oj-masonrylayout-tile-[0-9]+x[0-9]+/,

    _TILE_SELECTOR =   "." + _OJ_MASONRYLAYOUT_TILE,
    
    _OJ_MASONRYLAYOUT_TILE_TRANSITION_SHOW_FROM_CLASS = "oj-masonrylayout-tile-transition-show-from",
    
    //log warning message when "disabled" property set
    _WARNING_DISABLED_OPTION = "JET MasonryLayout: 'disabled' property not supported",

    //Context Menu: menu item ids
    _OJMASONRYLAYOUTCUT = "ojmasonrylayoutcut",
    _OJMASONRYLAYOUTPASTEBEFORE = "ojmasonrylayoutpastebefore",
    _OJMASONRYLAYOUTPASTEAFTER = "ojmasonrylayoutpasteafter",
    _MENU_CMD_MAP = {"cut"          : _OJMASONRYLAYOUTCUT,
                     "paste-before" : _OJMASONRYLAYOUTPASTEBEFORE,
                     "paste-after"  : _OJMASONRYLAYOUTPASTEAFTER
                    },
    _MENU_TRANSLATION_MAP = {"cut"          : "labelCut",
                             "paste-before" : "labelPasteBefore",
                             "paste-after"  : "labelPasteAfter"
                            },
    //make sure the collapseEventTimeout param is less than the one used in the unit tests
    //in order to ensure that the masonry listener gets the resize event before the unit test
    _RESIZE_LISTENER_COLLAPSE_EVENT_TIMEOUT = 25,

    /** 
     * Get the position of an element relative to the document.
     * @param {Object} elem Element for which to get position.
     * @returns {Object} Object containing properties: top, left.
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _getRelativePosition = function(elem)
    {
      return $(elem).offset();
    },

    /** 
     * Add a style class name to an element.
     * @param {Object} elem Element to which to add style class.
     * @param {string} styleClass Style class name to add.
     * @return {void}
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _addStyleClassName = function(elem, styleClass)
    {
      $(elem).addClass(styleClass);
    },

    /** 
     * Remove a style class name from an element.
     * @param {Object} elem Element from which to remove style class.
     * @param {string} styleClass Style class name to remove.
     * @return {void}
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _removeStyleClassName = function(elem, styleClass)
    {
      $(elem).removeClass(styleClass);
    },

    /** 
     * Get the masonry tile size style class applied to an element.
     * @param {Object} elem Element for which to get size style class name.
     * @returns {?string} Style class name.
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _getSizeStyleClassName = function(elem)
    {
      var str = null;
      var tile = $(elem);
      var tileClass = tile.attr('class');
      var matches = tileClass.match(_OJ_MASONRYLAYOUT_TILE_SIZE_CLASS_PATTERN);

      if (matches)
        str = matches[0];

      return str;
    },

    /** 
     * Get the masonry tile span of an element.
     * @param {Object} elem Element for which to get span.
     * @returns {Object} Object containing properties: colSpan, rowSpan.
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _getTileSpan = function(elem)
    {
      var span = null;
      var tile = $(elem);
      var tileClass = tile.attr('class');
      var matches = tileClass.match(_OJ_MASONRYLAYOUT_TILE_SIZE_CLASS_PATTERN);
      
      if (matches) {
        var styleClass = matches[0];
        var colRowSpanStr = styleClass.substring(_OJ_MASONRYLAYOUT_TILE.length + 1);
        var xIndex = colRowSpanStr.indexOf('x');
        var columnIndex = colRowSpanStr.substring(0, xIndex);
        var rowIndex = colRowSpanStr.substring(xIndex + 1);
        span = {colSpan: MasonryLayoutCommon._getCSSLengthAsInt(columnIndex),
                rowSpan: MasonryLayoutCommon._getCSSLengthAsInt(rowIndex)};
      }

      return span;
    },
    
    /**
     * Sort tiles based on their original DOM order.
     * @param {Array} arTiles Array of tiles.
     * @returns {Array} The given array of tiles, sorted.
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _sortTilesOriginalOrder = function(arTiles)
    {
      if (arTiles)
        arTiles.sort(_compareTilesOriginalOrder);
      
      return arTiles;
    },
    
    /**
     * Compare tiles for sorting based on the original DOM order.
     * @param {Object} tile1 A tile.
     * @param {Object} tile2 Another tile.
     * @returns {number} 1 if tile2 should be sorted before tile1, -1 if tile1
     *          should be sorted before tile2, or 0 if the tile order doesn't
     *          need to change
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _compareTilesOriginalOrder = function(tile1, tile2)
    {
      if (tile2._jetDataMasonryOriginalOrder < tile1._jetDataMasonryOriginalOrder)
        return 1;
      if (tile1._jetDataMasonryOriginalOrder < tile2._jetDataMasonryOriginalOrder)
        return -1;
      return 0;
    },

    // start static functions for drag-and-drop reordering /////////////////////

    /** 
     * Find the masonry tile that contains the given element.
     * @param {Object} elem Descendant element of a masonry tile.
     * @param {Object} rootElem Root of the masonryLayout.
     * @returns {Object} Masonry tile element.
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _findContainingTile = function(elem, rootElem)
    {
      var currElem = elem;
      while (currElem)
      {
        var style = currElem.style;
        if (style && (style.visibility === _HIDDEN || style.display === _NONE))
          break;
        var parentElem = currElem.parentNode;
        if (parentElem === rootElem)
          return currElem;
        currElem = parentElem;
      }
      return null;
    },

    /** 
     * Get the next visible sibling element.
     * @param {Object} elem Element for which to get next visible sibling.
     * @returns {Object} Next visible sibling element.
     * @memberof oj.ojMasonryLayout
     * @private
     */
    _getNextElement = function(elem)
    {
      var currElem = elem;
      while (currElem)
      {
        var nextElem = currElem.nextSibling;
        var bVisible = true;
        if (nextElem)
        {
          var style = nextElem.style;
          if (style && (style.visibility === _HIDDEN || style.display === _NONE))
            bVisible = false;
        }
        if (nextElem && nextElem.nodeType === 1 && bVisible)
          return nextElem;
        currElem = nextElem;
      }
      return null;
    };

    // end static functions for drag-and-drop reordering ///////////////////////
    
// end static members and functions ////////////////////////////////////////////

}()); // end of MasonryLayout wrapper function
(function() {
var ojMasonryLayoutMeta = {
  "properties": {
    "reorderHandle": {
      "type": "string"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "labelCut": {
          "type": "string",
          "value": "Cut"
        },
        "labelPasteAfter": {
          "type": "string",
          "value": "Paste After"
        },
        "labelPasteBefore": {
          "type": "string",
          "value": "Paste Before"
        }
      }
    }
  },
  "methods": {
    "insertTile": {},
    "removeTile": {},
    "resizeTile": {}
  },
  "events": {
    "animateEnd": {},
    "animateStart": {},
    "beforeInsert": {},
    "beforeRemove": {},
    "beforeReorder": {},
    "beforeResize": {},
    "insert": {},
    "remove": {},
    "reorder": {},
    "resize": {},
  },
  "extension": {
    _WIDGET_NAME: "ojMasonryLayout"
  }
};
oj.CustomElementBridge.registerMetadata('oj-masonry-layout', 'baseComponent', ojMasonryLayoutMeta);
oj.CustomElementBridge.register('oj-masonry-layout', {'metadata': oj.CustomElementBridge.getMetadata('oj-masonry-layout')});
})();
/**
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

(function() {

/**
 * @typedef {Object} oj.ojMenu.PositionAlign
 * @property {"top"|"bottom"|"center"} [vertical] Vertical alignment.
 * @property {"start"|"end"|"left"|"center"|"bottom"} [horizontal] Horizontal alignment. <p>
 * <ul>
 *  <li><b>"start"</b> evaluates to "left" in LTR mode and "right" in RTL mode.</li>
 *  <li><b>"end"</b> evaluates to "right" in LTR mode and "left" in RTL mode.</li>
 * </ul>
 *
 */

/**
 * @typedef {Object} oj.ojMenu.PositionPoint
 * @property {number} [x] Horizontal aligment offset.
 * @property {number} [y] Vertical alignment offset.
 */

/**
 * @typedef {Object} oj.ojMenu.Position
 * @property {oj.ojMenu.PositionAlign} [my] Defines which edge on the menu to align with the target ("of") element.
 * @property {oj.ojMenu.PositionAlign} [at] Defines which position on the target element ("of") to align the positioned element
 *                                  against.
 * @property {oj.ojMenu.PositionPoint} [offset] Defines a point offset in pixels from the ("my") alignment.
 * @property {string|oj.ojMenu.PositionPoint} [of] Which element to position the menu against.  The default is the
 * <code class="prettyprint">launcher</code> argument passed to the
 * <code class="prettyprint">open</code> method. <p>
 *
 * If the value is a string, it should be a selector or the literal string value
 * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
 * is used, the values are relative to the whole document.  Page horizontal and vertical
 * scroll offsets need to be factored into this point - see UIEvent
 * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
 * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
 *
 * @property {"flip"|"fit"|"flipfit"|"none"} [collision] Rule for alternate alignment. <p>
 * <ul>
 *  <li><b>"flip"</b> the element to the opposite side of the target and the
 *             collision detection is run again to see if it will fit. Whichever side
 *             allows more of the element to be visible will be used. </li>
 * <li><b>"fit"</b> shift the element away from the edge of the window. </li>
 * <li><b>"flipfit"</b> first applies the flip logic, placing the element
 *  on whichever side allows more of the element to be visible. Then the fit logic
 *  is applied to ensure as much of the element is visible as possible.</li>
 * <li><b>"none"</b> no collision detection.</li>
 * </ul>
 */

/**
 * @typedef {Object} oj.ojMenu.OpenOptions
 * @property {string} [display] Determines whether the menu is displayed as a dropDown menu or a sheet menu.
 * @property {string} [initialFocus] Determines focus behavior when the menu is initially opened.
 * @property {string|Element} [launcher] The DOM node (which may or may not be a JET element) that launches this menu.
 * @property {oj.ojMenu.Position} [position] Determines the position of a dropDown menu when launched. Ignored for sheet menus.
 */

/**
 * @ojcomponent oj.ojMenu
 * @ojdisplayname Menu
 * @augments oj.baseComponent
 * @ojrole menu
 * @since 0.6
 * @ojshortdesc Displays a popup menu with support for touch, mouse, and keyboard interactions.
 * @ojstatus preview
 *
 * @classdesc
 * <h3 id="menuOverview-section">
 *   JET Menu
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#menuOverview-section"></a>
 * </h3>
 *
 * <p>Description: Themeable, WAI-ARIA-compliant popup menu with touch, mouse and keyboard interactions for navigation.
 *
 * <p>A JET Menu is created using an ( <code class="prettyprint">&lt;oj-menu></code> ) tag with an ( <code class="prettyprint">&lt;oj-option></code> ) tag representing each menu item:
 *
 * <pre class="prettyprint">
 * <code>&lt;oj-menu id="menu" style="display:none" aria-label="Order Edit">
 *   &lt;oj-option>Item 1&lt;/oj-option>
 *   &lt;oj-option>Item 2&lt;/oj-option>
 *   &lt;oj-option>Item 3
 *     &lt;oj-menu id="submenu">
 *       &lt;oj-option>Item 3-1&lt;/oj-option>
 *       &lt;oj-option>Item 3-2&lt;/oj-option>
 *       &lt;oj-option>Item 3-3&lt;/oj-option>
 *       &lt;oj-option>Item 3-4&lt;/oj-option>
 *       &lt;oj-option>Item 3-5&lt;/oj-option>
 *     &lt;/oj-menu>
 *   &lt;/oj-option>
 *   &lt;oj-option>Item 4&lt;/oj-option>
 *   &lt;oj-option>Item 5&lt;/oj-option>
 * &lt;/oj-menu>
 * </code></pre>
 *
 * <p>JET Menus are not intended to be scrollable, as large, unwieldy menus are not good UX.  Ideally menus should have a manageable number of items; if this is not
 * possible, then it is preferable to organize contents into submenus, rather than introducing scrolling.
 *
 *
 * <h3 id="popup-section">
 *   Popup Menus
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#popup-section"></a>
 * </h3>
 *
 * <p>JET Menu is a popup component, for use with [context menu]{@link oj.baseComponent#contextMenu}, [menu button]{@link oj.ojButton#menu},
 * or similar functionality.  It is not intended to sit inline on the page.  See also the [JET NavigationList]{@link oj.ojNavigationList} component.
 *
 * <p>For this reason, the component is automatically hidden until it is opened.  However, this styling is not applied until the component is initialized.
 * To avoid a FOUC (flash of unstyled content), applications are encouraged to apply <code class="prettyprint">style="display:none"</code> to the menu markup,
 * as shown in the above code sample.
 *
 *
 * <h3 id="submenus-section">
 *   Submenus
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#submenus-section"></a>
 * </h3>
 * 
 * <p>Submenus can be created by specifying nested <code class="prettyprint">oj-menu</code> elements under the desired <code class="prettyprint">oj-option</code> elements.
 * 
 * <p>When a submenu is present, a default submenu icon will be automatically added to the parent menu item (see <a href="#itemIcons-section">Menu Item Icons</a>).
 * 
 * <p>Sheet menus are not appropriate when submenus are present. Therefore, submenus and there parent menus are always displayed as a dropDown regardless of the
 * <code class="prettyprint">open-options.display</code> attribute's value.
 * 
 * 
 * <h3 id="dividers-section">
 *   Dividers
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#dividers-section"></a>
 * </h3>
 *
 * <p>Divider elements can be created by including menu items that contain only spaces and/or dashes, or nothing at all:
 *
 * <pre class="prettyprint">
 * <code>&lt;oj-menu id="menu" style="display:none" aria-label="Order Edit">
 *   &lt;oj-option>Item 1&lt;/oj-option>
 *   &lt;oj-option>---&lt;/oj-option>
 *   &lt;oj-option>Item 2&lt;/oj-option>
 * &lt;/oj-menu>
 * </code></pre>
 *
 * <p>For WAI-ARIA compliance, JET automatically adds <code class="prettyprint">role="separator"</code> to the divider element.
 *
 *
 * <h3 id="itemIcons-section">
 *   Menu Item Icons
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#itemIcons-section"></a>
 * </h3>
 *
 * <p>Menu items currently support the rendering of start and end icons. Submenu icons are inserted automatically. To replace the default
 * submenu icon with a custom icon, the <code class="prettyprint">endIcon</code> slot should be specified. To render additional start
 * or end icons for a menu item, the <code class="prettyprint">startIcon</code> or <code class="prettyprint">endIcon</code> slot of the
 * <code class="prettyprint">oj-option</code> should be specified. See the <code class="prettyprint">oj-option</code> doc for details about
 * accepted children and slots.</p>
 * 
 * 
 * <h3 id="dismissal-section">
 *   Dismissal
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#dismissal-section"></a>
 * </h3>
 *
 * <p>JET Menus auto-dismiss in the expected cases, such as focus loss and menu item selection.  In addition, Sheet Menus
 * offer the following optional dismissal affordances:
 *
 * <ul>
 * <li>A "Cancel" menu item is displayed for Sheet Menus if the
 *     <code class="prettyprint">$menuSheetCancelAffordance</code> SASS variable is set to
 *     <code class="prettyprint">"menuItem"</code>.  See its
 *     <a href="#translations.labelCancel">translation</a> and <a href="#oj-menu-cancel-command">subId</a>.</li>
 * <li>The user can dismiss Sheet Menus via a downward swipe on the menu if the
 *     <code class="prettyprint">$menuSheetSwipeDownBehavior</code> SASS variable is set to
 *     <code class="prettyprint">"dismiss"</code>.</li>
 * </ul>
 *
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>The app should supply either an <code class="prettyprint">aria-label</code> or <code class="prettyprint">aria-labelledby</code>
 * attribute on the menu's root element, except possibly for menu buttons as discussed below. These attributes should not be supplied
 * for submenus, which are labeled automatically.
 *
 * <p>If a menu is shared by different launchers, and should have a different label for each launcher, then a
 * <a href="#event:ojBeforeOpen">ojBeforeOpen</a> listener can be used to set a different label per launch.
 *
 * <p>For a menu launched exclusively by one or more [menu buttons]{@link oj.ojButton#menu}, these attributes are optional.  When the
 * menu is opened via the menu button UI, if neither attribute is present after all <a href="#event:ojBeforeOpen">ojBeforeOpen</a>
 * listeners have been called, then <code class="prettyprint">aria-labelledby</code> will be set on the menu, referencing the menu
 * button, and will be removed when the menu is closed.  This approach provides a useful default label, while allowing the app to
 * supply a different label if desired, and while allowing the menu to be shared by several menu buttons and/or other launchers.
 *
 * <p>Disabled content: JET supports an accessible luminosity contrast ratio,
 * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>,
 * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which
 * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user
 * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio
 * required of enabled content, it cannot be used to convey meaningful information.<p>
 *
 *
 * <h3 id="reparenting-section">
 *   Reparenting
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#reparenting-section"></a>
 * </h3>
 *
 *  <p id="reparenting-strategy">
 *     When a menu is opened, it will be reparented in the document and reparented back when closed.
 *     The goal of this design is to maintain as much of the page author's document structure as possible, while
 *     avoiding most of the clipping and positioning issues of a completely inline design.
 *  </p>
 *  <p>
 *     If opened from another popup, the menu will be reparented to the nearest parent popup.
 *     Otherwise, the menu will be reparented to a container in the document body.
 *  </p>
 *  <p>
 *     The context of opening is defined by the resolved <code class="prettyprint">openOptions.launcher</code> value,
 *     which can be set via the <a href="#openOptions.launcher">attribute</a>, via the argument to the <a href="#open">open()</a>
 *     method, or via a <a href="#event:ojBeforeOpen">ojBeforeOpen</a> listener.
 *  <p>
 *     All menus are assigned the same z-index values. The layering between peer popups reflects the opening order.
 *     In addition, the page author has control over z-index weights by way of the menu's layer.
 *     The menu's layer defines the "stacking context" and assignd the "oj-menu-layer" style.
 *  </p>
 *  <p>
 *     Some notable consequences of this design:
 *  </p>
 *  <ul>
 *    <li>Events raised within the menu will not bubble up to the menu's original ancestors.  Instead, listeners for menu events should
 *        be applied to either the menu's root element, or the document.</li>
 *    <li>Likewise, developers should not use CSS descendant selectors, or similar logic, that assumes that the menu will remain a child
 *        of its original parent.</li>
 *  </ul>
 *
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <p>If a menu launcher (such as a [menu button]{@link oj.ojButton#menu} or item with a [context menu]{@link oj.baseComponent#contextMenu})
 * is stamped inside a table, dataGrid, or other container, the resulting set of launchers should share a single menu defined outside the container.
 *
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>The only supported way to set the reading direction (LTR or RTL) is to set the <code class="prettyprint">"dir"</code> attribute on the
 * <code class="prettyprint">&lt;html></code> element of the page.  As with any JET component, in the unusual case that the reading direction
 * is changed post-init, the menu must be <code class="prettyprint">refresh()</code>ed, or the page must be reloaded.
 *
 *
 * <h3 id="binding-section">
 *   Declarative Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#binding-section"></a>
 * </h3>
 *
 * <p>For components like Menu and Buttonset that contain a number of like items, applications may wish to use a <code class="prettyprint">foreach</code> Knockout binding
 * to stamp out the contents.  This binding cannot live on the same node as the JET <code class="prettyprint">ojComponent</code> binding, and must instead live on a nested
 * virtual element as follows:
 *
 * <pre class="prettyprint">
 * <code>&lt;oj-menu id="menu" style="display:none" aria-label="Order Edit">
 *     &lt;!-- ko foreach: menuItems -->
 *         &lt;oj-option data-bind="attr: {id: id, disabled: disabled}">
 *             &lt;span data-bind="text: label">&lt;/span>
 *         &lt;/oj-option>
 *     &lt;!-- /ko -->
 * &lt;/oj-menu>
 * </code></pre>
 *
 *
 * <!-- - - - - Above this point, the tags are for the class.
 *              Below this point, the tags are for the constructor (initializer). - - - - - - -->
 *
 */
oj.__registerWidget("oj.ojMenu", $['oj']['baseComponent'], {
    defaultElement: "<ul>", // added to externs.js, since this is an override of a superclass member.  (That's the rule for public methods, what about protected fields?)  TODO: Would @override do the job and be better than externing?
    delay: 300, // Doesn't get renamed even when unquoted and not in (our) externs.js file, so I'm leaving it unquoted for now.  TBD: This is private, but do NOT rename to _delay since there's an inherited instance method by that name, so rename so something else prefixed with _.
    role: "menu", // private.  I moved from options to here since no longer public option.  Leave unquoted so gets renamed by GCC as desired.
    widgetEventPrefix : "oj",
    options: { // options is in externs.js, so no need for quotes
        /**
         * Disables the menu if set to <code class="prettyprint">true</code>.
         *
         * @member
         * @name disabled
         * @memberof oj.ojMenu
         * @instance
         * @type {boolean}
         * @default false
         *
         * @example <caption>Initialize the menu with the <code class="prettyprint">disabled</code> attribute specified:</caption>
         * &lt;oj-menu disabled='true'>&lt;/oj-menu>
         *
         * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
         * // getter
         * var disabledValue = myMenu.disabled;
         *
         * // setter
         * myMenu.disabled = true;
         */
        // disabled option declared in superclass, but we still want the above API doc

        // Deprecated in release 2.1.0.  Should be removed when that release End-of-Lifes.
        // At that time, update the "if not ul" check in _ComponentCreate per comment there, and update non-ul unit tests to ensure we throw in that case.
        /**
         * Selector for the elements that serve as the menu container, including submenus.
         *
         * <p>Note: The <code class="prettyprint">menuSelector</code> attribute should not be changed after initialization. Existing submenus will not be updated.
         *
         * @expose
         * @memberof oj.ojMenu
         * @instance
         * @ignore
         * @type {string}
         * @default "ul"
         * @deprecated 2.1.0 Menus should always be created from an unordered list ( <code class="prettyprint">&lt;ul></code> ).
         *   This API will be removed in a future release.
         */
        menuSelector: "ul",

        /**
         * <p>A collection of settings impacting the launch of a menu.  These <code class="prettyprint">openOptions</code>
         * may be accessed and overridden individually or collectively, as seen in the examples.
         *
         * <p>The values set here can be overridden on a per-launch basis by passing the corresponding params into the
         * <a href="#open">open</a> method.  Those per-launch values can be further customized by a
         * <a href="#event:ojBeforeOpen">ojBeforeOpen</a> listener.
         *
         * <p>The built-in [menu button]{@link oj.ojMenuButton} and [context menu]{@link oj.baseComponent#contextMenu} functionality
         * overrides some of the Menu's <code class="prettyprint">openOptions</code>, for WAI-ARIA compliance and other reasons.
         * Thus, if the app really wants to customize
         * those values, it must do so in a <code class="prettyprint">ojBeforeOpen</code> listener.  If the built-in menu button
         * or context menu functionality is modified in this way, it is the app's responsibility to ensure that the result is
         * both correct and accessible.
         *
         * @expose
         * @memberof oj.ojMenu
         * @instance
         * @type {Object}
         * @ojsignature { target: "Type",
         *                value: "oj.ojMenu.OpenOptions",
         *                jsdocOverride: true }
         *
         * @example <caption>Initialize the menu, setting some <code class="prettyprint">openOptions</code> values.</caption>
         * &lt;oj-menu open-options.initial-focus='true' open-options.launcher='myLauncher'>&lt;/oj-menu>
         *
         * @example <caption>Get or set the <code class="prettyprint">openOptions</code> attribute, after initialization:</caption>
         * // Get one
         * var value = myMenu.openOptions.launcher;
         *
         * // Get all
         * var values = myMenu.openOptions;
         *
         * // Set one, leaving the others intact
         * myMenu.setProperty('openOptions.initialFocus', 'none');
         *
         * // Set many.  Any existing openOptions not listed are lost
         * myMenu.openOptions = { 'launcher': 'myLauncher',
         *                        'initialFocus': 'firstItem',
         *                        'position': myPositionObj };
         */
        openOptions: {
            /**
             * <p>Determines whether the menu is displayed as a dropDown menu or a sheet menu.
             *
             * <p>The default value is <code class="prettyprint">"auto"</code>, in which case the behavior is a function of the
             * screen width and the <code class="prettyprint">$menuDropDownThresholdWidth</code> SASS variable.  For example,
             * if that variable is set to 768px, then for screen widths of 768px and larger, the menu will display as a dropDown,
             * and for screen widths less than 768px, the menu will display as a sheet.
             *
             * <p>To avoid disorienting the user, if the screen width changes while the menu is already open (e.g. due to a device
             * rotation), the display may not change until the next launch.
             * 
             * <p>Sheet menus are not appropriate when submenus are present.  Thus, menus having submenus are always displayed as
             * a dropDown, regardless of the values of this option and the SASS variable.
             *
             * <p>If the SASS variable is set to 0 or a huge value such as 99999px, then all menus with
             * <code class="prettyprint">display</code> set to <code class="prettyprint">"auto"</code> will always display as a
             * dropDown or sheet, respectively.
             *
             * @expose
             * @alias openOptions.display
             * @memberof! oj.ojMenu
             * @instance
             * @since 2.1.0
             *
             * @type {string}
             * @default "auto"
             * @ojvalue {string} "auto" Displays the menu as a sheet or dropDown, depending on the screen width.
             * @ojvalue {string} "dropDown" Displays the menu as a dropDown.
             * @ojvalue {string} "sheet" Displays the menu as a sheet.
             *
             * @example <caption>Initialize the menu with the <code class="prettyprint">openOptions.display</code> sub-option specified:</caption>
             * &lt;oj-menu open-options.display='dropDown'>&lt;/oj-menu>
             *
             * @example <caption>Get or set the <code class="prettyprint">openOptions.display</code> sub-option, after initialization:</caption>
             * // getter
             * var display = myMenu.openOptions.display;
             *
             * // setter:
             * myMenu.setProperty('openOptions.display', 'sheet');
             */
            display : "auto",

            /**
             * Determines focus behavior when the menu is initially opened.
             *
             * @expose
             * @alias openOptions.initialFocus
             * @memberof! oj.ojMenu
             * @instance
             * @type {string}
             * @default "menu"
             * @ojvalue {string} "none" Leaves focus where it is, e.g. on the launching component.  The application must verify that the result is accessible.
             * @ojvalue {string} "menu" Focuses the menu itself, with no menu item focused (e.g. typical Context Menu behavior).
             * @ojvalue {string} "firstItem" Focuses the first menu item (e.g. MenuButton <kbd>DownArrow</kbd> behavior).
             *
             * @example <caption>Initialize the menu with the <code class="prettyprint">openOptions.initialFocus</code> sub-option specified:</caption>
             * &lt;oj-menu open-options.initial-focus='firstItem'>&lt;/oj-menu>
             *
             * @example <caption>Get or set the <code class="prettyprint">openOptions.initialFocus</code> sub-option, after initialization:</caption>
             * // getter
             * var initialFocus = myMenu.openOptions.initialFocus;
             *
             * // setter:
             * myMenu.setProperty('openOptions.initialFocus', 'none');
             */
            initialFocus : "menu",

            /**
             * <p>The DOM node (which may or may not be a JET element) that launches this menu.
             * This node must be focusable, as focus is returned to it upon menu dismissal.
             *
             * <p>The launcher must either be specified in this component option, or on each menu launch -- see <a href="#open">open()</a>
             * and <a href="#event:ojBeforeOpen">ojBeforeOpen</a>.
             *
             * @expose
             * @alias openOptions.launcher
             * @memberof! oj.ojMenu
             * @instance
             * @type {string|Element}
             * @default null
             *
             * @example <caption>Initialize the menu with the <code class="prettyprint">openOptions.launcher</code> sub-option specified:</caption>
             * &lt;oj-menu open-options.launcher='myLauncher'>&lt;/oj-menu>
             *
             * @example <caption>Get or set the <code class="prettyprint">openOptions.launcher</code> sub-option, after initialization:</caption>
             * // getter
             * var launcher = myMenu.openOptions.launcher;
             *
             * // setter:
             * myMenu.setProperty('openOptions.launcher', 'myLauncher');
             */
            launcher: null,

            /**
             * <p>Determines the position of a dropDown menu when launched via the <code class="prettyprint">open()</code> method or via menu button or
             * context menu functionality.  Ignored for sheet menus.
             *
             * <p>The "my" and "at" properties define aligment points relative to the menu and other element.  The "my" property represents the menu's
             * alignment where the "at" property represents the other element that can be identified by "of" or defauts to the launcher when the menu
             * opens.  The values of these properties describe horizontal and vertical alignments.</p>
             *
             * <ul>
             * <li>JET supports <code class="prettyprint">start</code> and <code class="prettyprint">end</code> values wherever <code class="prettyprint">left</code>
             * and <code class="prettyprint">right</code> are supported.  The <code class="prettyprint">start</code> value means "left in LTR; right in RTL",
             * while the <code class="prettyprint">end</code> value means "right in LTR; left in RTL."</li>
             * </ul>
             *
             * <p>Menu also supports the following extended syntax for the <code class="prettyprint">of</code> field:
             *
             * <ul>
             * <li>The <code class="prettyprint">"event"</code> keyword means "position the menu relative to the UI event that opened the menu."</li>
             * <li>The <code class="prettyprint">"launcher"</code> keyword means "position the menu relative to the launcher element."</li>
             * </ul>
             *
             * <p>By default, when the <code class="prettyprint">of</code> field is not set, the menu is positioned relative to the launcher.</p>
             * 
             * <p>The default position value varies between menus and submenus as follows:
             * <ul>
             *   <li>Top level menu default: <code class="prettyprint">{ my: { horizontal: "start", vertical: "top" }, at: { horizontal: "start", vertical: "bottom" }, offset: { x: 0, y: 0 }, collision: "flipfit" }</code></li>
             *   <li>Submenu default: <code class="prettyprint">{ my: { horizontal: "start", vertical:  "top" }, at: { horizontal: "end", vertical: "top" }, offset: { x: 0, y: 0 }, collision: "flipfit" }</code></li>
             * </ul>
             * 
             * <b>Deprecated v5.0.0 jQuery UI position syntax; Use of a percent unit with "my" or "at" is not supported.</b>
             *
             * @expose
             * @alias openOptions.position
             * @memberof! oj.ojMenu
             * @instance
             * @type {Object}  
             *
             * @example <caption>Initialize the menu with the <code class="prettyprint">openOptions.position</code> option specified:</caption>
             * &lt;oj-menu open-options.position.my.horizontal='start'>&lt;/oj-menu>
             *
             * @example <caption>Get or set the <code class="prettyprint">openOptions.position</code> sub-option, after initialization:</caption>
             * // Get one field of position object
             * var position = myMenu.openOptions.position.my;
             *
             * // Get entire position object
             * var position = myMenu.openOptions.position;
             *
             * // Set one field of position object, leaving the others intact
             * myMenu.setProperty('openOptions.position.at.horizontal', 'right');
             *
             * // Set entire position object. Any fields not listed are lost.
             * myMenu.setProperty('openOptions.position', {"my": {"horizontal": "start", "vertical": "bottom"},
             *                                             "at": {"horizontal": "end", "vertical": "top" },
             *                                             "offset": {"x": 0, "y":5}});
             */
            position: {
              /**
               * Defines which edge on the menu to align with the target ("of") element.
               *
               * @expose
               * @memberof! oj.ojMenu
               * @instance
               * @alias openOptions.position.my
               * @name openOptions.position.my
               * @type {{horizontal:string, vertical:string}}
               */
              'my' : {
                /**
                 * Defines the horizontal alignment of the menu.
                 * @expose
                 * @memberof! oj.ojMenu
                 * @instance
                 * @alias openOptions.position.my.horizontal
                 * @name openOptions.position.my.horizontal
                 * @type {string}
                 * @default 'start'
                 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
                 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
                 * @ojvalue {string} "left"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "right"
                 */
                'horizontal': 'start',
                /**
                 * Defines the vertical alignment of the menu.
                 * @expose
                 * @memberof! oj.ojMenu
                 * @instance
                 * @alias openOptions.position.my.vertical
                 * @name openOptions.position.my.vertical
                 * @type {string}
                 * @default 'top'
                 * @ojvalue {string} "top"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "bottom"
                 */
                'vertical': 'top'
              },
              /**
               * Defines a point offset in pixels from the ("my") alignment.
               * @expose
               * @memberof! oj.ojMenu
               * @instance
               * @alias openOptions.position.offset
               * @name openOptions.position.offset
               * @type {{x:number, y:number}}
               */
              'offset': {
                /**
                 * Horizontal aligment offset.
                 * @expose
                 * @memberof! oj.ojMenu
                 * @instance
                 * @alias openOptions.position.offset.x
                 * @name openOptions.position.offset.x
                 * @type {number}
                 * @default 0
                 */
                'x': 0,
                /**
                 * Vertical alignment offset.
                 * @expose
                 * @memberof! oj.ojMenu
                 * @instance
                 * @alias openOptions.position.offset.y
                 * @name openOptions.position.offset.y
                 * @type {number}
                 * @default 0
                 */
                'y': 0
              },
              /**
               * Defines which position on the target element ("of") to align the positioned element
               * against.
               *
               * @expose
               * @memberof! oj.ojMenu
               * @instance
               * @alias openOptions.position.at
               * @name openOptions.position.at
               * @type {{horizontal:string, vertical:string}}
               */
              'at' : {
                /**
                 * Defines the horizontal alignment of what the menu is aligned to. For top-level menus, the default value is "start". For submenus, the default value is "end".
                 * @expose
                 * @memberof! oj.ojMenu
                 * @instance
                 * @alias openOptions.position.at.horizontal
                 * @name openOptions.position.at.horizontal
                 * @type {string}
                 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
                 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
                 * @ojvalue {string} "left"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "right"
                 */
                'horizontal': 'start',
                /**
                 * Defines the vertical alignment of what the menu is aligned to. For top-level menus, the default value is "bottom". For submenus, the default value is "top".
                 * @expose
                 * @memberof! oj.ojMenu
                 * @instance
                 * @alias openOptions.position.at.vertical
                 * @name openOptions.position.at.vertical
                 * @type {string}
                 * @ojvalue {string} "top"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "bottom"
                 */
                'vertical': 'bottom'
              },
              /**
               * Which element to position the menu against.  The default is the
               * <code class="prettyprint">launcher</code> argument passed to the
               * <code class="prettyprint">open</code> method.
               *
               * If the value is a string, it should be a selector or the literal string value
               * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
               * is used, the values are relative to the whole document.  Page horizontal and vertical
               * scroll offsets need to be factored into this point - see UIEvent
               * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
               * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
               *
               * @example <caption>Finding the point for an svg element:</caption>
               * var rect = svgDom.getBoundingClientRect();
               * var position = {of:{x:rect.left + window.pageXOffset, y:rect.top + window.pageYOffset}};
               *
               * @expose
               * @memberof! oj.ojMenu
               * @instance
               * @alias openOptions.position.of
               * @name openOptions.position.of
               * @type {string|{x: number, y: number}| undefined}
               */
              'of' : undefined,
              /**
               * Rule for alternate alignment.
               *
               * @expose
               * @memberof! oj.ojMenu
               * @instance
               * @alias openOptions.position.collision
               * @name openOptions.position.collision
               * @type {string}
               * @default 'flipfit'
               * @ojvalue {string} "flip" the element to the opposite side of the target and the
               *  collision detection is run again to see if it will fit. Whichever side
               *  allows more of the element to be visible will be used.
               * @ojvalue {string} "fit" shift the element away from the edge of the window.
               * @ojvalue {string} "flipfit" first applies the flip logic, placing the element
               *  on whichever side allows more of the element to be visible. Then the fit logic
               *  is applied to ensure as much of the element is visible as possible.
               * @ojvalue {string} "none" no collision detection.
               */
              // : Ensure menu stays onscreen (hence no autoscrolling/jumping the page to move it back onscreen), even when when
              // the screen is very small (phones, small desktop browser windows).  If the menu height is less than the viewport height, but
              // greater than the viewport height y-above and y-below the launcher, then with the default "flip" policy, the menu will appear
              // y-above or y-below the launcher, and the window will autoscroll vertically to display the menu, while with the "flipfit" policy, the
              // menu will appear z-above the launcher as needed to stay onscreen, so the window does not need to autoscroll vertically. Likewise horizontally.
              'collision' : 'flipfit'
            }
        },

        // Omitting the usual verbiage about whether the "other" sub-options are clobbered by a given setter syntax,
        // since only one sub-option currently.
        // TBD: restore that verbiage (copy from openOptions) if gain 2nd sub-option.
        /**
         * <p>A collection of settings impacting the launch of submenus.
         *
         * <p>This option affects submenus, while the similar <code class="prettyprint">openOptions</code> affects the top-level menu.
         *
         * @expose
         * @memberof oj.ojMenu
         * @instance
         * @ignore
         * @type {Object}
         */
        submenuOpenOptions: {
            /**
             * <p>Determines the position of submenus.
             *
             * <p>Please refer to the jQuery UI [Position]{@link http://api.jqueryui.com/position/} utility for details about the various choices.
             * In addition to that syntax, note that JET supports the following reading direction-aware extended syntax in the
             * <code class="prettyprint">my</code> and <code class="prettyprint">at</code> fields:
             *
             * <ul>
             * <li>JET supports <code class="prettyprint">start</code> and <code class="prettyprint">end</code> values wherever <code class="prettyprint">left</code>
             * and <code class="prettyprint">right</code> are supported.  The <code class="prettyprint">start</code> value means "left in LTR; right in RTL",
             * values in the <code class="prettyprint">my</code> and <code class="prettyprint">at</code> fields wherever <code class="prettyprint">left</code>
             * and <code class="prettyprint">right</code> are supported.  The <code class="prettyprint">start</code> value means "left in LTR; right in RTL",
             * while the <code class="prettyprint">end</code> value means "right in LTR; left in RTL."</li>
             *
             * <li>Similarly, JET supports <code class="prettyprint">></code> and <code class="prettyprint">&lt;</code> operators wherever <code class="prettyprint">+</code>
             * and <code class="prettyprint">-</code> are supported.  The <code class="prettyprint">></code> value means "+ in LTR; - in RTL",
             * while the <code class="prettyprint">&lt;</code> value means "- in LTR; + in RTL."  E.g. a <code class="prettyprint">my</code> value
             * of <code class="prettyprint">"start>40"</code> shifts the submenu 40px "endward," while a <code class="prettyprint">my</code> value
             * of <code class="prettyprint">"start&lt;40"</code> shifts the submenu 40px "startward."</li>
             * </ul>
             *
             * <p>By default, the submenu is positioned relative to the parent menu item, but if a value is set on
             * the <code class="prettyprint">of</code> field, then the submenu is positioned relative to that element or position instead.
             *
             * @expose
             * @alias submenuOpenOptions.position
             * @memberof! oj.ojMenu
             * @instance
             * @ignore
             * @type {Object}
             * @default { "my": "start top", "at": "end top", "collision": "flipfit" }
             */
            position: {
                /** @expose */
                my: "start top",
                /** @expose */
                at: "end top",
                "collision": "flipfit" // see comments on openOptions.position.collision
            }
        },

        // Events

        /**
         * Triggered when a default animation is about to start, such as when the component is
         * being opened/closed or a child item is being added/removed. The default animation can
         * be cancelled by calling <code class="prettyprint">event.preventDefault</code>.
         *
         * <caption>The default animations are controlled via the theme (SCSS) :</caption>
         * <pre class="prettyprint"><code>
         * // dropdown menu
         * $menuDropDownOpenAnimation: (effect: "zoomIn", transformOrigin: "#myPosition", duration: $animationDurationShort) !default;
         * $menuDropDownCloseAnimation: (effect: "none") !default;
         *
         * // sheet menu
         * $menuSheetOpenAnimation: (effect: "slideIn", direction: "top", duration: $animationDurationShort) !default;
         * $menuSheetCloseAnimation: (effect: "slideOut", direction: "bottom", duration: $animationDurationShort) !default;
         *
         * </code></pre>
         * @ojshortdesc Triggered when a default animation is about to start, such as when the component is
         * being opened/closed or a child item is being added/removed. The default animation can
         * be cancelled by calling event.preventDefault.
         *
         * @expose
         * @event
         * @memberof oj.ojMenu
         * @instance
         * @ojcancelable
         * @ojbubbles
         * @property {"open"|"close"} action The action that is starting the animation.
         *            The number of actions can vary from element to element.
         *            Suggested values are:
         *                    <ul>
         *                      <li>"open" - when a menu element is opened</li>
         *                      <li>"close" - when a menu element is closed</li>
         *                    </ul>
         * @property {!Element} element target of animation
         * @property {!function():void} endCallback If the event listener calls
         *            event.preventDefault to cancel the default animation, It must call the
         *            endCallback function when it finishes its own animation handling and any
         *            custom animation has ended.
         *            
         * @example <caption>Bind an event listener to the
         *          <code class="prettyprint">onOjAnimateStart</code> property to override the default
         *          "open" animation:</caption>
         * myMenu.onOjAnimateStart = function( event )
         *   {
         *     // verify that the component firing the event is a component of interest and action
         *      is open
         *     if (event.detail.action == "open") {
         *       event.preventDefault();
         *       oj.AnimationUtils.fadeIn(event.detail.element).then(event.detail.endCallback);
         *   };
         *   
         * @example <caption>Bind an event listener to the
         *          <code class="prettyprint">onOjAnimateStart</code> property to override the default
         *          "close" animation:</caption>
         * myMenu.onOjAnimateStart = function( event )
         *   {
         *     // verify that the component firing the event is a component of interest and action
         *      is close
         *     if (event.detail.action == "close") {
         *       event.preventDefault();
         *       oj.AnimationUtils.fadeOut(event.detail.element).then(event.detail.endCallback);
         *   };
         */
        animateStart : null,

        /**
         * Triggered when a default animation has ended, such as when the component is being
         * opened/closed or a child item is being added/removed. This event is not triggered if
         * the application has called preventDefault on the animateStart
         * event.
         *
         * @expose
         * @event
         * @memberof oj.ojMenu
         * @instance
         * @ojcancelable
         * @ojbubbles
         * @property {!Element} element target of animation
         * @property {"open"|"close"} action The action that is ending the animation.
         *                   The number of actions can vary from element to element.
         *                   Suggested values are:
         *                    <ul>
         *                      <li>"open" - when a menu element is opened</li>
         *                      <li>"close" - when a menu element is closed</li>
         *                    </ul>
         *                    
         * @example <caption>Bind an event listener to the
         *          <code class="prettyprint">onOjAnimateEnd</code> property to listen for the "open"
         *          ending animation:</caption>
         * myMenu.onOjAnimateEnd = function( event )
         *   {
         *     // verify that the component firing the event is a component of interest and action
         *      is open
         *     if (event.detail.action == "open") {}
         *   };
         * 
         * @example <caption>Bind an event listener to the
         *          <code class="prettyprint">onOjAnimateEnd</code> property to listen for the "close"
         *          ending animation:</caption>
         * myMenu.onOjAnimateEnd = function( event )
         *   {
         *     // verify that the component firing the event is a component of interest and action
         *      is close
         *     if (event.detail.action == "close") {}
         *   };
         */
        animateEnd : null,

        // Benefit of making openOptions live is this:
        //
        // - For MenuButton and ContextMenu, the app doesn't control the call to Menu.open().
        // - Our internal call to this method may pass in overrides to things like "initialFocus", in cases where the Right Thing for (say) MenuButtons differs
        //   from Menu's default option value (which may be tailored to, say, contextMenus).
        //     - This way, we don't have to rely on the app to set these things correctly, and we don't have to permanently set the Menu's options, which may be
        //       shared between (say) a MenuButton, a ContextMenu, and some custom app usage of the menu.
        // - The remaining piece of the puzzle is to give the app a way to override the values set in our internal call to open().  A live payload field handles this.
        //
        // If ever needed, we can add a "submenuOpenOptions" payload field alongside the "openOptions" field.
        /**
         * <p>Triggered before this menu is launched via the <a href="#open">open</a> method or via menu button or context menu functionality.
         * The launch can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>.
         *
         * <p>The <code class="prettyprint">ui.openOptions</code> payload field contains the settings being used for this menu launch,
         * resulting from merging the <code class="prettyprint">openOptions</code> passed to <code class="prettyprint">open()</code>, if any,
         * with the <code class="prettyprint">openOptions</code> component option.
         *
         * <p>This field is "live", meaning that the listener can alter fields such as <code class="prettyprint">position</code> to affect this launch without
         * affecting the component option.  Since these changes are applied to the merged object, they supersede both the <code class="prettyprint">openOptions</code>
         * passed to <code class="prettyprint">open()</code> and the <code class="prettyprint">openOptions</code> component option.
         *
         * <p>If any of the above techniques are used to alter the built-in [menu button]{@link oj.ojButton#menu} or [context menu]{@link oj.baseComponent#contextMenu}
         * functionality, it is the app's responsibility to ensure that the result is both correct and accessible.
         *
         * @expose
         * @event
         * @memberof oj.ojMenu
         * @instance
         * @ojcancelable
         * @ojbubbles
         */
        beforeOpen: null,

        /**
         * <p>Triggered after this menu is closed.
         *
         * @expose
         * @event
         * @memberof oj.ojMenu
         * @instance
         * @since 2.0.0
         * @property {Event} event a custom event
         */
        close: null,

        /**
         * Triggered when the menu is created.
         *
         * @event
         * @name create
         * @memberof oj.ojMenu
         * @instance
         * @ignore
         * @property {Event} event <code class="prettyprint">jQuery</code> event object
         * @property {Object} ui Currently empty
         */
        // create event declared in superclass, but we still want the above API doc

        /**
         * <p>Triggered when the active menu item changes.  Private; do not use.
         *
         * <p>Internal notes:
         *
         * <p>We've replaced JQUI's focus/blur events with this internal event, and made their focus/blur methods internal.
         * It's been agreed with the architects that if we ever need any of this API
         * to be public, we'll have a focusedItem option, read-only or R/W, with an optionChange event, instead
         * of the removed API.  (Exact name TBD, but they favored focusedRow, with "ed",  for Table if it had an option rather than
         * a method.)  If for some reason we keep a separate event instead of an optionChange event, do NOT call this event
         * "focusedItem", since that will prevent ever having an "focusedItem" option since same namespace.  Instead, call this
         * "focusedItemChange" in that case.
         *
         * <p>In the meantime, we'll keep firing this private event, since it's used so extensively and usefully in the unit tests to make sure
         * other stuff works, and since keeping this working and tested means that we can just change the name to optionChange if
         * we ever need the public event.
         *
         * <p>The difference between this method and JQUI's focus event is that it fires for blurs too, it doesn't fire if the old
         * and new active item are the same, and we fire a single event, not both a blur and focus, when the active state moves from item A to
         * item B.
         *
         * @event
         * @name _activeItem
         * @memberof oj.ojMenu
         * @instance
         * @private
         * @property {Event} event <code class="prettyprint">jQuery</code> event object
         * @property {Object} ui Parameters
         * @property {jQuery} ui.previousItem the previously focused menu item
         * @property {jQuery} ui.item the currently focused menu item
         *
         * @example <caption>Initialize the menu with the <code class="prettyprint">_activeItem</code> callback specified:</caption>
         * $( ".selector" ).ojMenu({
         *     "_activeItem": function( event, ui ) {}
         * });
         *
         * @example <caption>Bind an event listener to the <code class="prettyprint">oj_activeitem</code> event:</caption>
         * // $( ".selector" ) must select either the menu root, or the document, due to reparenting
         * $( ".selector" ).on( "oj_activeitem", function( event, ui ) {} );
         */

        /**
         * <p>Triggered after this menu is launched via the <a href="#open">open</a> method or via menu button or context menu functionality.
         *
         * @expose
         * @event
         * @memberof oj.ojMenu
         * @instance
         * @since 2.0.0
         * @property {Event} event a custom event
         */
        open: null,

        /**
         * <p>Triggered when a menu item (other than the built-in <a href="#dismissal-section">"Cancel"</a> item) is selected.
         *
         * <p>To ensure keyboard accessibility, the only correct, supported way to react to the selection of a menu item is to listen
         * for this event.  Click listeners and <code class="prettyprint">href</code> navigation should not be used.
         * 
         * <p>To find the value of the menu item that triggered an action event, the event.target.value should be used.
         *
         * @expose
         * @event
         * @memberof oj.ojMenu
         * @instance
         * @ojcancelable
         * @ojbubbles
         * @since 4.0.0
         * @example <caption>Find the value of the selected menu item:</caption>
         *   myMenu.onOjAction = function( event )
         *   {
         *     // the target of the action event is the selected &lt;oj-option> element
         *     var itemSelected = event.target;
         *     
         *     // find the 'value' of the selected &lt;oj-option> element
         *     var selectedValue = itemSelected.value;
         *   }
         */
        action: null
    },

    /**
     * @memberof oj.ojMenu
     * @instance
     * @protected
     * @override
     */
    _ComponentCreate: function() { // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
        this._super();

        // Create aliases, that won't be renamed, for the private methods that are called by unit tests.  These unit tests come
        // from JQUI, in which these methods were actually public.  With these aliases, we don't have to @expose private method names
        // (which prevents renaming and bloats minified code), and our internal calls to these methods can be this._focus rather than this["_focus"].
        // TBD: perhaps the unit tests could simulate keyboard events rather than calling these methods.
        this["_focusForTesting"] = this._focus;
        this["_nextForTesting"] = this._next;
        this["_selectForTesting"] = this._select;

        // Cancel menu item is supported for custom elements, so we only need the following check for the old syntax
        if (!this._IsCustomElement())
        {
            // Cancel menu item supported only for <ul>-based menus, since it's a new feature introduced after the
            // deprecation of non-<ul> menus.  Per arch. discussion, throw in this case rather than just turning off cancel item.
            if (_SHEETS_HAVE_CANCEL_ITEM && this.element[0].tagName.toLowerCase() !== "ul")
                throw new Error("Cancel item supported for <ul> menus only.");
            
            // TODO: When support for non-<ul> menus is pulled after the deprecation period, replace the above check with this one:
//          if (this.element[0].tagName.toLowerCase() !== "ul")
//              throw new Error("Menu must be based on a <ul> element.");

            this._createAsTopLevelMenu();
        }
        else
        {
            //check if this is a submenu
            var parentOption = this.element.parent();
            if (parentOption.is( "oj-option" ))
            {
                var parentMenu = parentOption.parent();
                if (parentMenu.is( "oj-menu" ))
                {
                    this._isSubmenu = true;
                    this._createAsSubmenu();
                }
            }

            if (!this._isSubmenu)
                this._createAsTopLevelMenu();
            
            // fixup the position option if custom element menu or submenu
            var options = this.options;
            options["openOptions"]["position"] = oj.PositionUtils.coerceToJet(options["openOptions"]["position"]);
        }
    },
    
    _createAsTopLevelMenu: function()
    {
        var self = this;
        this.activeMenu = this.element;
        
        // flag used to prevent firing of the click handler as the event bubbles up through nested menus
        this.mouseHandled = false;
        
        this._setupSwipeBehavior();

        this.element
            .uniqueId()
            .addClass( "oj-menu oj-component" )
            .hide()
            .attr({
                "role": this.role,
                "tabIndex": "0"
            });

        // pass true to catch these events on all menus, not just enabled menus
        this._on(true, {
            //Required to stick the focus on disabled menu.
            "mousedown .oj-menu-item": function(event) {
                if (this.options.disabled) {
                    event.preventDefault();
                }
            },
            "click": function(event) {
                if (this.options.disabled) {
                    event.preventDefault();
                }
            },
            //On Esc key focus should be shifted to launcher and dismiss menu.
            //the TAB key should also implicitly dismiss the menu
            "keydown": function(event) {
                if (this.options.disabled) {
                    if (event.keyCode === $.ui.keyCode.ESCAPE || event.keyCode === $.ui.keyCode.TAB) {

                        if (event.keyCode === $.ui.keyCode.TAB) {
                          event.preventDefault();
                        }

                        if (this._launcher) { // if menu is currently open
                            this._focusLauncherAndDismiss(event);
                        }
                    }
                }
            }
        });

        // needed since _setOption() is not automatically called at create time.
        // TBD: Would be a little better to toggle these 2 things rather than only setting them if true, as in superclass _setOption().
        if ( this.options.disabled ) {
            this.element
                .addClass( "oj-disabled" )
                .attr( "aria-disabled", "true" );
        }

        var handleMouseEnterMenuItem = function( event ) {
            // the focusHandled var ensures that this handler only runs for the target
            // menu item, not for the parent menu items to which the event bubbles. Without it, submenu item becomes
            // non-selectable if mouse outside menu on way from parent item to sub (), and flakiness
            // where sometimes a tap on a menu item inside the submenu doesn't "take" in touch scenarios.
            if ( this.focusHandled )
                return;
            this.focusHandled = true;

            var target = $( event.currentTarget );

            try {
                this._focusIsFromPointer = true;
                this._focus( event, target );
            } finally {
                this._focusIsFromPointer = false;
            }
        }.bind(this);

        var handleMouseLeave = function( event ) {
            // Only handle mouseleave if the mouse is leaving the menu, not if
            // the menu is disappearing out from under the mouse.  The latter
            // happens when the mouse is over a submenu, and the submenu is closed
            // via the keyboard (e.g. leftArrow, Esc).  If the callee is called at
            // that time, then a timer clears this.active, which breaks KB nav.
            if (event && event.target && !$(event.target).is(":visible"))
                return;
            this._collapse( event, "eventSubtree" );
        }.bind(this);

        this._on({
            // Prevent focus from sticking to links inside menu after clicking
            // them (focus should always stay on UL during navigation).
            "mousedown .oj-menu-item > a": function( event ) {
                event.preventDefault();
            },
            "click .oj-disabled > a": function( event ) {
                event.preventDefault();
            },
            "click": function(event) {
                // when the click event bubbles out of the root menu element, we're done with it, so
                // reset this flag to its initial value of false in preparation for the next click
                this.mouseHandled = false;
            },
            "touchstart": function(event) {
                // when the touchstart event bubbles out of the root menu element, we're done with it, so
                // reset this flag to its initial value of false in preparation for the next touch
                this.focusHandled = false;
            },
            "mouseover": function(event) {
                // when the mouseover event bubbles out of the root menu element, we're done with it, so
                // reset this flag to its initial value of false in preparation for the next mouse hover.
                // Note that this flag is reset in mouseover event and not in a mouseenter event; mouseover
                // bubbles to the root element, but mouseenter does not.
                this.focusHandled = false;
            },
            "click .oj-menu-item:has(a)": function( event ) {
                var target = $( event.target ).closest( ".oj-menu-item" );
                // the mouseHandled var ensures that the click is handled only for the originally clicked
                // menu item, not for the parent menu items to which it bubbles.
                if ( !this.mouseHandled && target.not( ".oj-disabled" ).length ) {
                    this.mouseHandled = true;

                    // prevent page scrolling and appending # to page URL, which can interfere with routing, etc.
                    // Do this before the bailout so these things are prevented when user clicks a 2nd time on parent menu item.
                    // No need to additionally do this for Enter/Space handler, because menu root, not the anchor, has browser focus
                    // in that case, so anchor click behavior doesn't happen, so doesn't need to be prevented.
                    event.preventDefault();

                    if (this.active && this.active.closest(target).length && this.active.get(0) != target.get(0)) {
                        //If current active menu item  is decendent of (and not equal to) target menu item then
                        //sub menu of the curent target is already open and hence no need to
                        //1. expand the sub menu
                        //2. as current target is a menu item having sub menu no need to invoke this._select(event).
                        return;
                    }
                    // Open submenu on click
                    if ( target.has( ".oj-menu" ).length ) {
                        this._expand( event );
                    }
                    else {
                        //Invoke _select() only for leaf menu items
                        this._select(event);
                        if (!this.element.is(":focus")) {
                            // Redirect focus to the menu
                            this.element.trigger("focus", [true]);

                            // If the active item is on the top level, let it stay active.
                            // Otherwise, blur the active item since it is no longer visible.
                            if (this.active && this.active.parents(".oj-menu").length === 1) {
                                this._clearTimer && this._clearTimer();
                            }
                        }
                    }
                }
            },
            "mouseenter .oj-menu-item": handleMouseEnterMenuItem,

            // : Bad touch device behavior because the JQUI code relies on the above mouseenter handler to call _focus(),
            // but for parent menu items on touch devices, mouseenter is called only if the previous tap was somewhere outside the
            // parent menu item, not if it was in the submenu.  So call that mouseenter handler on touchstart:
            "touchstart .oj-menu-item": handleMouseEnterMenuItem,

            "mouseleave": handleMouseLeave,
            "mouseleave .oj-menu": handleMouseLeave,
            "focus": function( event, keepActiveItem ) {
                if ( !keepActiveItem ) {
                    // If there's already an active item, keep it active
                    // If not, make the first item active
                    // TBD: is there a reason that JQUI needed to redundantly call _focus() on this.active when this.active was already set?
                    //      Or should we only call it when it's not set and we're calling it on the first menu item?
                    var item = this.active || this.element.children( ".oj-menu-item" ).eq( 0 );
                    this._focus( event, item );
                }
            },
            "keydown": this._keydown,
            "keyup": function( event ) {
                if (event.keyCode == $.ui.keyCode.ENTER || event.keyCode == $.ui.keyCode.SPACE)
                    this.__spaceEnterDownInMenu = false;
            }
        });

        this._focusable({
            // suppress focus ring for Mac Safari due to platform repainting bug in which previous item's outline is not fully erased
            'applyHighlight': !_IS_MAC_SAFARI,
            'recentPointer' : function() {
                return self._focusIsFromPointer;
            },
            'setupHandlers': function( focusInHandler, focusOutHandler) {
                self._focusInHandler = focusInHandler;
                self._focusOutHandler = focusOutHandler;
            }
        });

        // callback that overrides the position['using'] for auto dismissal when aligning element is cropped.
        this._usingCallback = $.proxy(this._usingHandler, this);
        this._setup();
    },
    
    _createAsSubmenu: function()
    {
        // set special container name as this should have submenu defaults
        this.element.attr(oj.Components._OJ_CONTAINER_ATTR, this['widgetName']);
        
        this.element
            .uniqueId()
            .addClass( "oj-menu oj-component oj-menu-submenu oj-menu-dropdown" ) // submenus are always dropdown
            .hide()
            .attr({
                "role": this.role,
                "tabIndex": "0"
            });
        this._setup();
    },

    // Resolves a Mobile Safari issue that occurs because mousedown fires after the touchend.
    // To be called only by baseComponent's contextMenu logic, which explains the issue more fully.
    // Gets/sets a static var, since the listener that needs to know whether to bail out is static
    // (shared by all menu instances).
    __contextMenuPressHoldJustEnded: function(val) {
        if (arguments.length)
            _contextMenuPressHoldJustEnded = val;
        else
            return _contextMenuPressHoldJustEnded;
    },
    
    _processOjOptions: function() {
        var self = this;
        
        self._maxEndIconCount = 0;
        self._maxStartIconCount = 0;
        self._startIconWidth = 0;
        self._endIconWidth = 0;
        
        var ojOptions = this.element.find("oj-option");
        $.each(ojOptions, function(i, option) {
            option["customOptionRenderer"] = self._customOptionRenderer.bind(self);
        });
    },
    
    _customOptionRenderer: function(option) {
        // Implement custom rendering here...
        var ojOption = $(option);
        var self = this;
        
        // clear out any roles or classes that we may have previously set
        this._clearOption(ojOption);
        
        // if 'a' tag already exists, we need to remove it and re-render content
        var anchors = ojOption.children('a');
        for (var i = 0; i < anchors.length; i++)
        {
            var customAnchor = $(anchors[i]);
            if (customAnchor.attr('ojmenu') === 'opt')
            {
                // reset existing slots, and remove the custom anchor
                customAnchor.children()
                    .removeClass('oj-menu-item-icon')
                    .removeClass('oj-menu-item-end-icon');

                customAnchor.replaceWith(customAnchor.contents());
                break;
            }
        }
        
        // test to see if this is a divider
        if (!/[^\-\u2014\u2013\s]/.test( ojOption.text() )) // hyphen, em dash, en dash
            this._initDividers(ojOption);
        else
        {
            this._initMenuItems(ojOption);
            
            // create 'a' tag
            var a = document.createElement('a');
            a.setAttribute('href', '#');

            // tag with ojmenu attribute to identify it later
            a.setAttribute('ojmenu', 'opt');
            customAnchor = $(a);

            customAnchor.uniqueId()
                   .attr({
                       "tabIndex": "-1",
                       "role": this._itemRole()
                   });

            ojOption.prepend(a); // @HTMLUpdateOK append trusted new DOM to menu item
            
            var supportedSlots = ["startIcon", "", "endIcon"];
            var slots = oj.BaseCustomElementBridge.getSlotMap(option);

            // reparent the slots, and make sure any necessary styling is applied
            $.each(supportedSlots, function(i, slotName) {
                if (slots[slotName]) {
                    if (slotName === "")
                    {
                        $.each(slots[slotName], function(j, node) {
                            // handle the submenu case
                            if (($(node).is("oj-menu")))
                            {
                                self._setupSubmenu(customAnchor, $(node));
                                this._hasSubmenus = true;
                            }
                            else
                                customAnchor.append(node); // @HTMLUpdateOK reparent trusted child DOM in menu item
                        });
                    }
                    else if (slotName === "startIcon")
                    {
                        var startIconCount = slots[slotName].length;
                        self._maxStartIconCount = Math.max(self._maxStartIconCount, startIconCount);

                        $.each(slots[slotName], function(j, node) {
                            $(node).addClass('oj-menu-item-icon');
                            customAnchor.append(node); // @HTMLUpdateOK reparent trusted child DOM in menu item

                            // positioning logic doesn't need to run if there is only 1 start icon
                            if (startIconCount > 1)
                                self._positionStartIcon(node, j, startIconCount);
                        });
                    }
                    else if (slotName === "endIcon")
                    {
                        var endIconCount = slots[slotName].length;
                        self._maxEndIconCount = Math.max(self._maxEndIconCount, endIconCount);

                        $.each(slots[slotName], function(j, node) {
                            $(node).addClass('oj-menu-item-end-icon');
                            customAnchor.append(node); // @HTMLUpdateOK reparent trusted child DOM in menu item

                            // positioning logic doesn't need to run if there is only 1 end icon
                            if (endIconCount > 1)
                                self._positionEndIcon(node, j, endIconCount);
                        });
                    }
                }
            });
            
            // check for disabled state in case all we need to do is update disabled attribute
            if (option['disabled'] == true)
            {
                ojOption.addClass('oj-disabled');
                customAnchor.attr( "aria-disabled", "true" );
            }
            else if (option['disabled'] == false)
            {
                ojOption.removeClass('oj-disabled');
                customAnchor.removeAttr( "aria-disabled" );
            }
        }
        
        var children = this.element.children();
        var dividers = children.filter( ".oj-menu-divider" );
        
        // ensure "before/after-divider" classes are applied iff appropriate
        this._initDividerNeighbors(children, dividers);
    },
    
    // Helper method to position start icons
    _positionStartIcon: function(node, index, count) {
        if (this.isRtl)
            var marginProperty = 'margin-right';
        else
            marginProperty = 'margin-left';
        
        var margin = parseInt($(node).css( marginProperty ), 10);
        // margins are negative for start icons
        this._startIconWidth = -1 * margin;
        $(node).css( marginProperty, margin * (count - index) + 'px' );
    },
    
    // Helper method to position end icons
    _positionEndIcon: function(node, index, count) {
        if (this.isRtl)
        {
            var marginProperty = 'margin-left';
            var widthProperty = 'margin-right';
        }
        else
        {
            marginProperty = 'margin-right';
            widthProperty = 'margin-left';
        }
        
        var margin = parseInt($(node).css( marginProperty ), 10);
        // margins are negative for end icons
        this._endIconWidth = -1 * parseInt($(node).css( widthProperty ), 10);
        $(node).css( marginProperty, (margin + this._endIconWidth * (count - index - 1)) + 'px' );
    },
    
    // Helper method to handle menu item padding updates
    _updateMenuPadding: function($menu) {
        
        var anchors = $menu.children()  // <li>s
                          .children(); // <a>s
        var iconCount = anchors.children(".oj-menu-item-icon:not(.oj-menu-cancel-icon)").length; // icons other than cancel item's icon
        $menu.toggleClass( "oj-menu-icons", !!iconCount )
            .toggleClass( "oj-menu-text-only", !iconCount );

        if (this._maxStartIconCount && this._maxStartIconCount > 1)
            this._applyAnchorIconPadding(anchors, this._startIconWidth, this._maxStartIconCount, true);

        var endIconCount = anchors.children(".oj-menu-item-end-icon").length;
        $menu.toggleClass( "oj-menu-end-icons", !!endIconCount );

        if (this._maxEndIconCount && this._maxEndIconCount > 1)
            this._applyAnchorIconPadding(anchors, this._endIconWidth, this._maxEndIconCount, false);
    },
    
    // Helper method to apply icon padding to menu item anchors
    _applyAnchorIconPadding: function(anchors, iconWidth, count, isStart) {
        if ((this.isRtl && isStart) || (!this.isRtl && !isStart))
            var paddingProperty = 'padding-right';
        else
            paddingProperty = 'padding-left';
        
        anchors.each(function() {
            var padding = parseInt($(this).css( paddingProperty ), 10);
            $(this).css( paddingProperty, (padding + iconWidth * (count - 1)) + 'px' );
        });
    },
    
   /**
    * @instance
    * @private
    * @param {!jQuery.Event|Event} event
    */
    _clickAwayHandler: function(event) {
                //Focus event needs to be captured because, in case of menu button (where focus is still on menu button instead of open menu), if user does
                //a taboff to another element then menu should be closed. With this we also no need to have additional "blur" handler on menu to close the menu/submenus.

                //Despite of focus/mousedown, still keydown listener is required for contextmenu events especially for menubutton with browser default context menu
                //and user pressed contextmenu keyboard key(not right mouse click).

                //Checking event.KeyCode along with event.which as currently event created by jquery-simulate.js is setting only event.keyCode for chrome/IE.
                //This avoids test failures. This can be removed after jquery simulates event properly.

                if (event.type === "focus" || event.type === "mousedown" || event.type === "touchstart" || event.which == 93 || (event.which == 121 && event.shiftKey) || event.keyCode == 93) { // Windows contextMenu key (93) or Shift-F10 (121)
                    // Resolves a Mobile Safari issue that occurs because mousedown fires after the touchend.
                    // baseComponent's contextMenu logic explains the issue more fully.
                    if (event.type === "mousedown" && _contextMenuPressHoldJustEnded) {
                        return;
                    }

                    var self = this;
                    
                    //Clone _openPopupMenus as __dismiss() will remove the open menu from _openPopupMenus list
                    var openPopupMenus = _openPopupMenus.slice(0, _openPopupMenus.length);
                    $.each(openPopupMenus , function(index, menu) {
                        // This logic dismisses/collapses the menu if event is outside the menu and any of the following are true:
                        // - Event is a touchstart (which may or may not become a pressHold) or left/middle mousedown, unless (event is in launcher and this is not a context menu).  "Unless" clause needed for menuButton.
                        // - Event is focusing something outside of both the menu and launcher.
                        // - Event is a context-menu-launching event other than pressHold.  (i.e. right-click or CM keys)
                        if (!$(event.target).closest(menu.element).length &&  // if event target is outside of menu element AND one of the following is true then close the menu.
                            (event.type === "keydown" || (event.type === "mousedown" && event.which === 3) ||  // 1. if it's a context-menu-launching event other than pressHold (see event.which on outer if)
                             !$(event.target).closest(menu._launcher).length ||  // 2. When focus is moved on to other than launcher or left/middle mousedown or touchstart on element other than launcher
                             (menu._launcherClickShouldDismiss && ((event.type === "mousedown" && event.which !== 3) || event.type === "touchstart" ))))  // 3. If event is a (left/middle-mousedown or touchstart) on launcher and current menu is contextmenu (see )
                        {
                          // Don't do it again if the menu is already being dismissed
                          if (!menu._dismissEvent)
                          {
                            menu._dismissEvent = event;

                            var promise = menu._collapse(event, "eventSubtree"); // "eventSubtree" is effectively "all" since we check that event is outside menu.  "all" would be clearer, but just in case, leaving it as is.

                            // Wait for subtree to be collapsed before dismissing self.
                            // There is no default close animation, but this allows app to define
                            // cascading close animation if it wants to.
                            self._runOnPromise(promise, function() {
                              // The menu could have been opened by a different launcher by
                              // the time its submenus finish collapsing, so double-check if we 
                              // still need to dismiss it.
                              if (menu._dismissEvent)
                              {
                                if (menu._launcher)
                                  menu.__dismiss(event);
                                
                                menu._dismissEvent = null;
                              }
                            });
                          }
                        }
                    });
                };
    },

    /**
     * @memberof oj.ojMenu
     * @instance
     * @protected
     * @param {string} key option name
     * @param {?Object} value of the target option identified by the key
     * @override
     */
    _setOption: function( key, value ) { // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
        this._superApply( arguments ); // TBD: consider calling super at end, so that optionChange (fired at end of super) is fired at very end

        switch (key) {
            case "translations.labelCancel":
            case "translations":
                // no refresh() needed to just change text of existing inited menu item
                this._cancelAnchor && this._cancelAnchor.text(this.options['translations']['labelCancel']);
                break;
        }
    },

    /**
     * @memberof oj.ojMenu
     * @instance
     * @protected
     * @override
     */
    _destroy: function() { // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
        if (this.element.is(":visible"))
            this.__dismiss();

        this._setWhenReady("none");

        this._clearTimer && this._clearTimer();
        delete this._clearTimer;

        // Destroy (sub)menus
        this.element
            .removeAttr( "aria-activedescendant" )
            .removeClass( "oj-component" )
            .find( ".oj-menu" ).addBack()
                .removeClass( "oj-menu oj-menu-submenu oj-menu-icons oj-menu-end-icons oj-menu-text-only" )
                .removeAttr( "role" )
                .removeAttr( "tabIndex" )
                .removeAttr( "aria-labelledby" )
                .removeAttr( "aria-hidden" )
                .removeAttr( "aria-disabled" )
                .removeUniqueId()
                .show();

        // Destroy menu items
        this.element.find( ".oj-menu-item" )
            .removeClass( "oj-menu-item" )
            .removeAttr( "role" )
            .children( "a" )
                .removeAttr( "aria-disabled" )
                .removeUniqueId()
                .removeClass( "oj-hover" )
                .removeAttr( "tabIndex" )
                .removeAttr( "role" )
                .removeAttr( "aria-haspopup" )
                .children().each( function() {
                    var elem = $( this );
                    if ( elem.data( "oj-ojMenu-submenu-icon" ) ) {
                        elem.remove();
                    }
                });

        // Destroy anchors
        this.element.find( "a" ).removeAttr( "aria-expanded" );

        // Destroy menu dividers
        this.element.find( ".oj-menu-divider" )
            .removeClass( "oj-menu-divider" )
            .removeAttr( "role" );

        // Remove the menu from openPopupMenus list if it is still added.
        if(_openPopupMenus.indexOf(this) >= 0)
            _openPopupMenus.splice(_openPopupMenus.indexOf(this),1);

        delete this._popupServiceEvents;
        delete this._usingCallback;

        var clearCloseDelayTimer = this._clearCloseDelayTimer;
        delete this._clearCloseDelayTimer;
        clearCloseDelayTimer && clearCloseDelayTimer();

        this._cancelDom && this._cancelDom.remove();
        this.element.ojHammer("destroy");

        this._super();
    },

    _keydown: function( event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        /*jshint maxcomplexity:20*/
        var match, prev, character, skip, regex,
            preventDefault = true;

        function escape( value ) {
            return value.replace( /[\-\[\]{}()*+?.,\\\^$|#\s]/g, "\\$&" );
        }

        switch ( event.keyCode ) {
        case $.ui.keyCode.HOME:
            this._move( "first", "first", event );
            break;
        case $.ui.keyCode.END:
            this._move( "last", "last", event );
            break;
        case $.ui.keyCode.UP:
            this._previous( event );
            break;
        case $.ui.keyCode.DOWN:
            this._next( event );
            break;
        case $.ui.keyCode.LEFT:
        case $.ui.keyCode.RIGHT:
            var isExpand = (event.keyCode === $.ui.keyCode.RIGHT) ^ this.isRtl;
            if (isExpand) {
                if ( this.active && !this.active.is( ".oj-disabled" ) ) {
                    this._expand( event );
                }
            } else {
                this._collapse( event, "active" );
            }
            break;
        case $.ui.keyCode.ENTER:
        case $.ui.keyCode.SPACE:
            this._handleEnterSpace( event );

            this.__spaceEnterDownInMenu = true;
            var self = this;

            // The spaceEnterDelay and __spaceEnterDownInMenu code addresses an issue where closing a menu, from within the menu via
            // Space or Enter, can immediately reopen the menu, because the keyUp can happen after focus has jumped to the button,
            // which clicks the button, which reopens the menu.  Repros most readily (only??) in Firefox.
            // TODO: try calling preventDefault() on the event in Menu (which is good practice anyway since it's handling the event), and
            // checking isDefaultPrevented() in Button.  If works, should be cleaner / more reliable than this existing fix.
            var spaceEnterDelay = 100; // 1 not enough in FF; 100 seems to do it.  If continued problems, try increasing this value.

            setTimeout(function () {
                self.__spaceEnterDownInMenu = false;
            }, spaceEnterDelay);

            break;
        // tab within a menu will implicitly dismiss
        case $.ui.keyCode.TAB:
          event.preventDefault();
          if (this._launcher) {
            this._focusLauncherAndDismiss( event );
          }
          break;
        // this handles enabled menus.  For disabled menus, see this handler: this._on(true, {...});
        case $.ui.keyCode.ESCAPE:
            if (this._launcher) { // if menu is currently open
                var activeItemId = this.element.attr("aria-activedescendant"); // <a> or nothing.  Always the same as this.active now that we change them in lockstep.
                var topLevelAnchorSelector = "#" + this.element.attr("id") + ">*>a"; // * is typically <li>
                var submenuOpen = activeItemId && !$("#" + activeItemId).is( topLevelAnchorSelector );

                if (submenuOpen)
                    this._collapse( event, "active" );
                else
                    this._focusLauncherAndDismiss( event );
            } else {
                this._collapse( event, "active" );
            }
            break;
        default:
            preventDefault = false;
            prev = this.previousFilter || "";
            character = String.fromCharCode( event.keyCode );
            skip = false;

            clearTimeout( this.filterTimer );

            if ( character === prev ) {
                skip = true;
            } else {
                character = prev + character;
            }

            regex = new RegExp( "^" + escape( character ), "i" );
            match = this.activeMenu.children( ".oj-menu-item" ).filter(function() {
                return regex.test( $( this ).children( "a" ).text() );
            });
            match = skip && match.index( this.active.next() ) !== -1 ?
                this.active.nextAll( ".oj-menu-item" ) :
                match;

            // If no matches on the current filter, reset to the last character pressed
            // to move down the menu to the first item that starts with that character
            if ( !match.length ) {
                character = String.fromCharCode( event.keyCode );
                regex = new RegExp( "^" + escape( character ), "i" );
                match = this.activeMenu.children( ".oj-menu-item" ).filter(function() {
                    return regex.test( $( this ).children( "a" ).text() );
                });
            }

            if ( match.length ) {
                this._focus( event, match );
                if ( match.length > 1 ) {
                    this.previousFilter = character;
                    this.filterTimer = setTimeout(function() {
                        delete this.previousFilter;
                    }.bind(this), 1000 );
                } else {
                    delete this.previousFilter;
                }
            } else {
                delete this.previousFilter;
            }
        }

        if ( preventDefault ) {
            event.preventDefault();
        }
    },

    /*
     * Called for Space and Enter
     */
    _handleEnterSpace: function( event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        if ( this.active && !this.active.is( ".oj-disabled" ) ) {
            if ( this.active.children( "a[aria-haspopup='true']" ).length ) {
                this._expand( event );
            } else {
                this._select( event );
            }
        }
    },

    /**
     * Refreshes the visual state of the menu. JET elements require a <code class="prettyprint">refresh()</code> after the DOM is
     * programmatically changed underneath the component.  For Menu, this includes:
     *
     * <ul>
     *   <li>After menu items are added or removed.</li>
     *   <li>After a change to a menu item's disabled status.
     *   <li>After the reading direction (LTR vs. RTL) changes.</li>
     * </ul>
     *
     * @expose
     * @memberof oj.ojMenu
     * @instance
     * @return {void}
     *
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myMenu.refresh();
     */
    refresh: function() { // Override of public base class method (unlike JQUI).  Method name needn't be quoted since is in externs.js.
        this._super();
        this._setup();


        var element = this.element;
        if (!element.is(":visible"))
          return;

        // reevaluate open menu positions
        var position = element.data(_POSITION_DATA);

        // Avoid repositioning it if the position.of element has gone missing, e.g. due to
        // a responsive layout change.  (Per architects, we don't want to do anything fancier
        // for this case.)  Per JQUI doc, position.of accepts Selector, Element, jQuery, or Event.
        // If it's window or event, we're happy.  Else, is Selector, Element, or jQuery, all of
        // which are accepted by $().
        if (!(position && (position.of instanceof $.Event || position.of instanceof Window || $(position.of).is(":visible"))))
          return;

        element.position(position);

        // Do the same for open submenus.  Don't bother with the position.of check this time, since
        // their position.of is essentially always the parent menu, not some other thing on the page.
        var subMenus = element.find(".oj-menu");
        subMenus.each(function()
        {
            var menu = $(this);
            if (menu.is(":visible"))
            {
              position = menu.data(_POSITION_DATA);
              if (position)
                menu['position'](position);
            }
        });
    },

    _setup: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        this.isRtl = this._GetReadingDirection() === "rtl";
        
        if (!this._IsCustomElement())
            this._setupWidgetMenus();
        else
        {
            this._processOjOptions();
            
            if (this._isSubmenu)
                this.options.openOptions.launcher = this.element.parent();
            
            // ensure menu padding is updated
            this._updateMenuPadding(this.element);
        }

        // If the active item has been removed, blur the menu
        if ( this.active && !$.contains( this.element[ 0 ], this.active[ 0 ] ) ) {
            this._blur();
        }
    },
    
    _setupWidgetMenus: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var self=this;
        
        var submenus = this.element.find( this.options.menuSelector ); // <ul>'s except root <ul>
        
        var menus = submenus.add( this.element ),                      // <ul>'s including root <ul>
            children = menus.children();                               // <li>'s in root menu and submenus

        this._hasSubmenus = !!submenus.length;

        // Anything that used to be a divider, but now has an "a", should become a menu element.
        children.filter( ".oj-menu-divider" ).has("a")
            .removeClass("oj-menu-divider oj-menu-item") // remove oj-menu-item if somehow present to ensure that it enters following block
            .removeAttr("role");

        // Don't refresh list items that are already adapted
        // TBD: .has prob has better perf than :has
        var uninitedItems = children.filter( ":not(.oj-menu-item):has(a)" );
        var uninitedAnchors = uninitedItems.children( "a" );

        this._initMenuItems(uninitedItems);
        this._initAnchors(uninitedAnchors);

        var dividers = children.filter(function(index, item) {
            // menu items without anchors containing spaces and/or dashes only
            // this test relies on the fact that _initMenuItems() has already been called
            var $item = $(item);
            return $item.is(":not(.oj-menu-item)")
                && !/[^\-\u2014\u2013\s]/.test( $item.text() ); // hyphen, em dash, en dash
        });
        this._initDividers(dividers);

        // ensure "before/after-divider" classes are applied iff appropriate
        this._initDividerNeighbors(children, dividers);

        // Add aria-disabled to any disabled menu item, and remove it from any recently enabled menu item
        children.filter( ".oj-disabled" ).children( "a" ).attr( "aria-disabled", "true" );
        children.filter( ":not(.oj-disabled)" ).children( "a" ).removeAttr( "aria-disabled" );

        // Initialize nested menus
        submenus.filter( ":not(.oj-menu)" )
            .addClass( "oj-menu oj-menu-submenu oj-menu-dropdown" ) // submenus are always dropdown
            .hide()
            .attr({
                "role": this.role,
                "aria-hidden": "true"
            })
            .each(function() {
                var menu = $( this ); // <ul>
                self._setupSubmenu(self._getSubmenuAnchor( menu ), menu);
            });

        menus.each(function() {
            // For each menu incl. submenus, apply either "oj-menu-icons" or "oj-menu-text-only" to that menu, depending on whether that menu
            // (excluding its submenus) contains at least one menu item having an icon.
            // This facilitates leaving space for a "column" of icons iff at least one icon is present, and doing so for each menu/submenu independently.
            // We exclude the cancel item's icon from the count, as that icon is always present if the cancel item is, but it's themed to show up iff
            // oj-menu-icons is present, i.e. iff at least one other icon is present.
            self._updateMenuPadding($(this));
        });
    },
    
    // helper method to create and add a submenu icon to the item provided
    _setupSubmenu: function(item, submenu) {
        var submenuIcon = $( "<span>" );
        
        submenuIcon   // separate stmt rather than chaining, since GCC can't tell that this is the setter overload of .data().
            .addClass( "oj-menu-submenu-icon oj-component-icon" )
            .data( "oj-ojMenu-submenu-icon", true ); // TODO: can't we just look for the class at destroy time rather than adding this data?
    
        item
            .attr( "aria-haspopup", "true" )
            .attr( "aria-expanded", "false" ) // per a11y team, live on <a>, not <ul> like JQUI
            .append( submenuIcon ); // @HTMLUpdateOK append trusted new DOM to menu item
    
        submenu.attr( "aria-labelledby", item.attr( "id" ));
    },

    /**
     * @private
     * @param {jQuery} items - non-divider <li>'s that haven't already been inited.
     */
    _initMenuItems: function(items) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        items.addClass( "oj-menu-item" )
             .attr( "role", "presentation" );

    },

    /**
     * @private
     * @param {jQuery} anchors - <a>'s that haven't already been inited.
     */
    _initAnchors: function(anchors) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        anchors.uniqueId()
               .attr({
                   "tabIndex": "-1",
                   "role": this._itemRole()
               });
    },

    /**
     * @private
     * @param {jQuery} dividers - divider <li>'s.
     */
    _initDividers: function(dividers) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        dividers.addClass( "oj-menu-divider" )
                .attr( "role", "separator" );
    },

    /**
     * @private
     * @param {jQuery} dividers - divider <li>'s.
     */
    _initDividerNeighbors: function(items, dividers) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        items.removeClass( "oj-menu-item-before-divider oj-menu-item-after-divider" );
        dividers.prev().addClass( "oj-menu-item-before-divider" );
        dividers.next().addClass( "oj-menu-item-after-divider" );
    },
    
    /**
     * @private
     */
    _clearOption: function(option) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        option.removeClass( "oj-menu-item-before-divider oj-menu-item-after-divider oj-menu-divider oj-menu-item" )
              .removeAttr( "role" );
    },

    /*
     * Given a list of one or more submenus (typically <ul>'s), finds the <a>'s that are their labels.
     */
    _getSubmenuAnchor: function(submenu) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        return submenu.prev( "a" );
    },

    _getSubmenuWidget: function(submenu) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var constructor = oj.Components.__GetWidgetConstructor(submenu, "ojMenu");
        return constructor && constructor("instance");
    },

    _itemRole: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        return "menuitem";
//        {
//            "menu": "menuitem",
//            "listbox": "option"
//        }[ this.role ];
    },

    // given a menu item, returns JQ object with any adjacent group dividers and optionally, that item
    _getAdjacentDividers: function( menuItem, includeItem ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var result = menuItem.prev(".oj-menu-divider")
                         .add(menuItem.next(".oj-menu-divider"));

        if (includeItem)
            result = result.add(menuItem);

        return result;
    },

    /**
     * Focuses the specified menu item and triggers the menu's <code class="prettyprint">_activeItem</code> event.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Event} event - What triggered the menu item to gain focus.  May be <code class="prettyprint">null</code>, but may not be omitted.
     * @param {!jQuery} item - The menu item to focus.  Its containing submenu, if any, must already be expanded. Must not be null or length 0.
     */
    _focus: function( event, item ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        // JQUI called blur() here.  This "if blah clear timer" is the only thing from that call that we (presumably) still want to do here.
        if ( !(event && event.type === "focus") )
            this._clearTimer && this._clearTimer();

        item = item.first(); // li.  Length 1.
        this._makeActive(item, event);

        var containingMenu = item.parent(); // ul.  Length-1.  Might be top or submenu.
        var parentMenuItem = containingMenu.closest( ".oj-menu-item" ); // li. Length 0 iff item is in top menu.

        // Remove oj-focus-ancestor class from all menu items and group dividers in the newly focused menu item's menu and submenus of that menu
        containingMenu.find( ".oj-focus-ancestor" ).removeClass( "oj-focus-ancestor" );

        // Highlight active parent menu item, if any, and adjacent group dividers, if any
        this._getAdjacentDividers(parentMenuItem, true).addClass( "oj-focus-ancestor" );

        if ( event && event.type === "keydown" ) {
            this._close();
        } else {
            this._clearTimer = this._setTimer(function() {
                delete this._clearTimer;
                this._close();
            }, this._getSubmenuBusyStateDescription("closing"), this.delay );
        }

        var nested = item.children( ".oj-menu" ); // immediately nested submenu.  length 0 or 1.
        if ( nested.length && event && ( /^mouse/.test( event.type ) && !this.active.hasClass( "oj-disabled" ) ) ) {
            this._startOpening(event, nested);
        }
        this.activeMenu = item.parent();
    },

    /*
     * Sets this.active (<li>), aria-activedescendant (<a>), and oj-focus (<li> and adjacent group dividers) in lockstep.
     * Never set those things outside of _makeActive() and _removeActive(), so they stay in synch!
     *
     * param item length-1 JQ object containing the <li> to focus
     */
    _makeActive: function( item, event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        // don't need to check for "both items null/empty", and don't need to null-check item, since item required to be length-1 JQ object
        var same = item.is(this.active);
        if (same) {
            return;
        }

        var previousItem = this.active ? this.active : $();
        var anchor = item.children( "a" );

        this.active = item;
        this.element.attr( "aria-activedescendant", anchor.attr( "id" ) );

        this._focusOutHandler(previousItem);
        this._focusInHandler(item);
        this._getAdjacentDividers(previousItem).removeClass( "oj-focus" );
        this._getAdjacentDividers(item).addClass( "oj-focus" );

        // see private API doc on the private _activeItem event declaration in this file
        this._trigger( "_activeItem", event, { "previousItem": previousItem , "item": item, "privateNotice": "The _activeItem event is private.  Do not use." } );
    },

    /*
     * Unsets this.active (<li>), aria-activedescendant (<a>), and oj-focus (<li> and adjacent group dividers) in lockstep.
     * Never set those things outside of _makeActive() and _removeActive(), so they stay in synch!
     *
     * Don't call this if you are immediately going to call _makeActive, to avoid firing the event twice (and redundant work).
     */
    _removeActive: function(event) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        if ( this.active ) { // then there is definitely a change, from "something focused" to "nothing focused", so proceed.
            var previousItem = this.active; // non-null, so don't need null-check

            this.active = null;
            this.element.removeAttr( "aria-activedescendant");

            this._focusOutHandler(previousItem);
            this._getAdjacentDividers(previousItem).removeClass( "oj-focus" );

            // see private API doc on the private _activeItem event declaration in this file
            this._trigger( "_activeItem", event, { "previousItem": previousItem, "item": $(), "privateNotice": "The _activeItem event is private.  Do not use." } );
        }
    },

    /**
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Event=} event - What triggered the menu item to blur.  May be <code class="prettyprint">null</code> or omitted.
     */
    _blur: function( event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        this._clearTimer && this._clearTimer();
        this._removeActive(event);
    },

    /*
     * param {Event} event - What triggered the menu to close. Payload for select (if applicable) and close events.
     * param {Object} selectUi - Payload for select event.  Non-null iff close caused by a menu item selection.
     */
    _focusLauncherAndDismiss: function( event, selectUi ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        // if the focus fails because the launcher has disappeared, e.g. due to a responsive layout change,
        // no warnings are logged, and the document body winds up focused during the subsequent reparenting,
        // which per the architects is the right thing (i.e. don't introduce any fancy handling for this case).
        this._launcher.focus();

        this.__dismiss( event, selectUi );
    },

    /*
     * Internal method, e.g. called by Button for MenuButton functionality.
     * Could make it public if ever needed.
     * param {Event} event - What triggered the menu to close. Payload for select (if applicable) and close events.
     * param {Object} selectUi - Payload for select event.  Non-null iff close caused by a menu item selection.
     */
    __dismiss: function(event, selectUi) { // Internal visibility; called by Button's MenuButton functionality.  Not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        if (this._isOperationPending("close", "__dismiss", [event, selectUi]))
          return;

        var isOpen = this.element.is(":visible");
        this._setWhenReady("close");

        /** @type {!Object.<oj.PopupService.OPTION, ?>} */
        var psOptions = {};
        psOptions[oj.PopupService.OPTION.POPUP] = this.element;

        // capture local state in a context used by the after close callback
        psOptions[oj.PopupService.OPTION.CONTEXT] = {"event": event,
                                                     "selectUi": selectUi,
                                                     "isOpen": isOpen};

        oj.PopupService.getInstance().close(psOptions);
    },

    /**
     * Get the default animation for a menu
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {string} menuType - The menu type (dropdown or sheet)
     * @param {string} action - The action to animate (open, close)
     * @return {Object} The default animation for the menu type
     */
    _getDefaultAnimation: function(menuType, action)
    {
      var defaults;
      var animation;
      
      defaults = (oj.ThemeUtils.parseJSONFromFontFamily('oj-menu-option-defaults') || {})["animation"];
      if (defaults && defaults[menuType])
      {
        animation = defaults[menuType][action];
      }
      
      return animation;
    },

    /**
     * Return a boolean to indicate if animation is disabled
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @return {boolean} true if animation is disabled; false otherwise.
     */
    _isAnimationDisabled: function()
    {
      // Disable animation if this is not custom element, or the menu is being
      // dismissed and immediately reopened.
      return !this._IsCustomElement() || this._disableAnimation;
    },

    /**
     * Replace animation options with runtime property values
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {*} effects - The animation options
     * @param {Object} propertyMap - The runtime property values
     * @return {*} The resolved animation effects
     */
    _replaceAnimationOptions: function(effects, propertyMap)
    {
      if (propertyMap && effects && typeof effects != 'string')
      {    
        var effectsAsString = JSON.stringify(effects);

        for (var key in propertyMap)
        {
          effectsAsString = effectsAsString.replace(new RegExp(key, 'g'), propertyMap[key]);
        }

        effects = JSON.parse(effectsAsString);
      }
      
      return effects;
    },

    /**
     * Utility method to run a task when a promise is resolved, or run it 
     * immediately if there is no promise.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Promise|null} promise - The promise to wait on
     * @param {function()} task - The task to run
     * @return {IThenable|null} A new promise or null
     */
    _runOnPromise: function(promise, task)
    {
      if (promise)
      {
        return promise.then(task);
      }
      else
      {
        return task();
      }
    },

    /**
     * Before callback is invoked while the menu is still visible and still parented in the zorder
     * container. Close animation is performed here.
     * @memberof! oj.ojMenu
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for closing the menu
     * @return {Promise|void}
     */
    _beforeCloseHandler : function (psOptions)
    {
      var rootElement = psOptions[oj.PopupService.OPTION.POPUP];
      
      // For custom element:
      // Fire action event before menu closed, so that app action handlers can do their thing
      // without waiting for the animation to finish.
      if (this._IsCustomElement())
      {
        var context = psOptions[oj.PopupService.OPTION.CONTEXT];
        var selectUi = context["selectUi"];

        //trigger action event on the oj-option element
        if (selectUi && selectUi["item"].length)
        {
            // the following event construction is taken from the '_triggerCustomEvent' logic, but we can't
            // use that method since the child <oj-option> is not a full fledged JET element.
            var detail = {};
            var eventName = "ojAction";
            var itemElement = selectUi["item"][0];
            var event = context["event"];
            if (event)
              detail['originalEvent'] = event instanceof $.Event ? event.originalEvent : event;

            var params = {'detail': detail};
            params['cancelable'] = true;
            params['bubbles'] = true;
            var customEvent = new CustomEvent(eventName, params);
            itemElement.dispatchEvent(customEvent);
            context["event"] = customEvent; // Use the action event as the close event's originalEvent
        }
      }

      if (this._isAnimationDisabled())
      {
        rootElement.hide();
        return void(0);
      }
      else
      {
        var animationOptions = this._getDefaultAnimation(this._sheetMenuIsOpen ? 'sheet' : 'dropdown', 'close');
        var promise = oj.AnimationUtils.startAnimation(rootElement[0], "close", 
          oj.PositionUtils.addTransformOriginAnimationEffectsOption(rootElement, animationOptions), this);

        promise.then(function() {
          rootElement.hide();
        });

        return promise;
      }
    },

    /**
     * Close finalization callback.
     *
     * @memberof! oj.ojMenu
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for closing the menu
     * @return {void}
     */
    _afterCloseHandler : function (psOptions)
    {
        //restore local variable state from #__dismiss
        var context = psOptions[oj.PopupService.OPTION.CONTEXT];
        var event = context["event"];
        var selectUi = context["selectUi"];
        var isOpen = context["isOpen"];

        this.element.removeData(_POSITION_DATA);

        this._launcher = undefined;
        this._sheetMenuIsOpen = false;

        // Preserve original logic for old-style component:
        // Fire select event after menu closed, so that app select handlers can do their thing
        // without worrying about the fact that the menu is still sitting there.
        // Fire select event before close event, because logical, and so that it can be the close event's
        // originalEvent.
        if (!this._IsCustomElement() && selectUi) {
            var selectResults = this._trigger2( "select", event, selectUi );
            event = selectResults['event']; // Use the select event as the close event's originalEvent
        }

        // just in case it's possible for __dismiss() to get called when menu is already closed, avoid firing spurious event:
        if (isOpen)
            this._trigger( "close", event, {} );

        this._currentOpenOptions = null;

        //Remove menu from openPopupMenus list
        if(_openPopupMenus.indexOf(this) >= 0)
            _openPopupMenus.splice(_openPopupMenus.indexOf(this),1);
    },

    /**
     * <p>Returns a copy of the <code class="prettyprint">openOptions</code> object applicable to the current launch, or the <a href="#openOptions">option</a>
     * value otherwise.
     *
     * <p>If the menu is shared among several launchers, this API can be used to find out what element launched the menu, as seen in the example below.
     *
     * <p>Detailed semantics:
     *
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Menu state</th>
     *       <th>Value</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Menu is open, or transitioning between open and closed, including when this method is called from an <a href="#event:open">open</a>,
     *           <a href="#event:action">action</a>, or <a href="#event:close">close</a>  listener. (For <a href="#event:ojBeforeOpen">ojBeforeOpen</a>, see next row.)</td>
     *       <td>A copy of the object used for the most recent launch is returned.  See the <a href="#openOptions">openOptions</a>
     *           option, the <a href="#open">open()</a> method, and the <a href="#event:ojBeforeOpen">ojBeforeOpen</a> event for details on how that
     *           object is constructed.</td>
     *     </tr>
     *     <tr>
     *       <td>This method is called from a <a href="#event:ojBeforeOpen">ojBeforeOpen</a> listener.</td>
     *       <td>A copy of the merged object "so far" is returned. The object ultimately used for the launch may differ if it is changed by
     *           a <code class="prettyprint">ojBeforeOpen</code> listener after this method is called.  Unlike the original copy passed to the
     *           <code class="prettyprint">ojBeforeOpen</code> listener, the copy returned by this method is not "live" and cannot be used to affect the launch.</td>
     *     </tr>
     *     <tr>
     *       <td>Menu is closed.  (All states not listed above.)</td>
     *       <td>A copy of the <a href="#openOptions">option</a> value is returned.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @expose
     * @memberof oj.ojMenu
     * @instance
     * @ignore
     * @since 2.0.0
     *
     * @return {!Object} the <code class="prettyprint">openOptions</code> object
     */
    getCurrentOpenOptions: function() { // Public, not an override (not in base class), so use @expose with unquoted method name.
        return $.extend(true, {}, this._currentOpenOptions || this.options.openOptions); // return a deep copy
    },

    /**
     * <p>Launches this menu after firing the <a href="#event:ojBeforeOpen">ojBeforeOpen</a> event.  Listeners to that event can cancel the launch
     * via <code class="prettyprint">event.preventDefault()</code>.  If the launch is not canceled, then the the <a href="#event:open">open</a> event
     * is fired after the launch.
     *
     * <p>This method's optional <code class="prettyprint">openOptions</code>param can be used to specify per-launch values for the settings in the
     * corresponding component options, without altering those options.  Those per-launch values can be further customized by a
     * <code class="prettyprint">ojBeforeOpen</code> listener.
     *
     * <p>Menus launched manually (as opposed to those launched by built-in functionality such as the [menu button]{@link oj.ojButton#menu} and
     * [context menu]{@link oj.baseComponent#contextMenu} functionality) must be launched via this API, not by simply unhiding the Menu DOM (such as
     * via jQuery's <code class="prettyprint">show()</code> API.
     *
     * @expose
     * @memberof oj.ojMenu
     * @instance
     * @return {void}
     * @ojsignature { target: "Type",
     *                value: "oj.ojMenu.OpenOptions",
     *                for: "openOptions",
     *                jsdocOverride: true }
     * @fires oj.ojMenu#ojBeforeOpen
     * @fires oj.ojMenu#ojAnimationStart
     * @fires oj.ojMenu#ojAnimationEnd
     * @fires oj.ojMenu#ojOpen
     *
     * @param {Object=} event What triggered the menu launch.  May be <code class="prettyprint">null</code>.  May be omitted if subsequent params are omitted.
     * @param {Object=} openOptions Options to merge with the <code class="prettyprint">openOptions</code> option.  May be <code class="prettyprint">null</code>.  May be omitted if subsequent params are omitted.
     * 
     * @example <caption>Invoke the <code class="prettyprint">open</code> method:</caption>
     * // override the launcher for this launch only, without affecting the other
     * // openOptions, and without affecting the component's openOptions option
     * myMenu.open(myEvent, {'launcher': 'myLauncher'});
     */
    open: function( event, openOptions ) { // Public, not an override (not in base class), so use @expose with unquoted method name.
      var submenuOpenOptions = arguments[2];
      if (this._isOperationPending("open", "open", [event, openOptions, submenuOpenOptions]))
        return;

        // If _clickAwayHandler has determined that we need to dismiss the menu,
        // do it first so that ojclose event is triggered for the old launcher
        // before ojbeforeopen event is triggered for the new launcher        
        if (this._dismissEvent)
        {
           // Disable animation since we'll be reopening the menu
          this._disableAnimation = true;
          
          // Use the event from _clickAwayHandler as the close event's originalEvent
          this.__dismiss(this._dismissEvent);
          
          // In case the menu is being opened by a different launcher, we don't
          // want the _clickAwayHandler for the old launcher to dismiss it.
          this._dismissEvent = null;
        }
        
        //
        // Important:  Merge [submenu]openOptions *before* calling _trigger(), and don't use the merged values until *after* the call.
        // Reason:  Per doc on open() and beforeOpen event, we pass the merged openOptions to beforeOpen listeners as a "live" object,
        // so the listener can both read and write the values used for this launch.  We may eventually pass submenuOpenOptions too, either to
        // beforeOpen or to beforeSubmenuOpen, if we ever have that.
        //
        // Merge needs 2 steps:
        // 1) Shallow merge (i.e. don't pass true as first arg to extend) of the 2 openOptions objects, into a new object.  Shallow so that the per-launch position object completely overrides the
        // component option's position object rather than merging with it.
        // 2) Then a deep copy of all object-valued fields in the merged object.  Position is the only such field, and it doesn't contain any objects of its own,
        // so this is actually just a shallow copy of position.  This is so that if beforeOpen listener mutates the position object, the position object in the component option remains unchanged.
        // Step 2 isn't needed for submenuOptions, since it isn't passed to beforeOpen.
        // $.fn.position copies the object passed to it before modifying it, so Step 2 isn't needed for that reason.
        openOptions = $.extend({}, this.options.openOptions, openOptions);
        openOptions.position = $.extend({}, openOptions.position);
        if (this._IsCustomElement())
            this._setPosition(openOptions.position);
        
        submenuOpenOptions = $.extend({}, this.options.submenuOpenOptions, submenuOpenOptions);

        // getCurrentOpenOptions() returns a deep copy of this._currentOpenOptions if set.  Put the live copy in the ivar, and have that method make the copy, so that the method picks up
        // beforeOpen listeners' changes to the live copy.  The old value of the ivar is non-null iff the menu is already open from a previous launch.  Grab the old value so we can restore it
        // if this (new) launch is cancelled, in which case the old launch stays up and subsequent calls to the method should return the old value.
        var oldOpenOptions = this._currentOpenOptions;
        this._currentOpenOptions = openOptions;

        oj.PositionUtils._normalizeEventForPosition(event); // see callee doc

        // Hack:  __openingContextMenu is set and unset by baseComponent._OpenContextMenu(), since Menu needs to know whether the
        // menu is open as a context menu vs. some other kind of menu including menu button,
        // as this affects whether subsequent mousedown/touchstart on launcher should dismiss menu.  IIRC, the upcoming Popup Fmwk
        // will address this need, but if not, fix it separately, perhaps by adding a new openOptions sub-option so it can be passed to menu.open().
        this._launcherClickShouldDismiss = this.element[0].__openingContextMenu;

        // TBD: if we ever pass submenuOpenOptions to a listener, must copy its position object first like we do for openOptions, above.
        var beforeOpenResults = this._trigger2( "beforeOpen", event, {openOptions: openOptions});

        if (!beforeOpenResults['proceed']) {
            this._currentOpenOptions = oldOpenOptions; // see comment above
            this._disableAnimation = false;
            return;
        }

        // Close menu if already open
        if (this.element.is(":visible")) {
            // Disable animation since we'll be reopening the menu
            this._disableAnimation = true;
            
            // if getCurrentOpenOptions() is called during the close event marking the end of the previous launch,
            // then it should return the details for the old launch
            this._currentOpenOptions = oldOpenOptions;

            // Use the beforeOpen event as the close event's originalEvent
            this.__dismiss(beforeOpenResults['event']); // sets this._currentOpenOptions to null
            
            // In case the menu is being opened by a different launcher, we don't
            // want the _clickAwayHandler for the old launcher to dismiss it.
            this._dismissEvent = null;

            this._currentOpenOptions = openOptions; // restore this launch's value
        }

        var launcher = openOptions.launcher;
        if (!this._IsCustomElement())
        {
            launcher = $.type(launcher) === "string"
                       ? $(launcher)
                       : launcher;
        }
        else
        {
            launcher = $.type(launcher) === "string"
                       ? $(document.getElementById(launcher))
                       : $(launcher);
        }

        if (!launcher || !launcher.length) {
            // need launcher so can return focus to it.
            oj.Logger.warn("When calling Menu.open(), must specify openOptions.launcher via the component option, method param, or beforeOpen listener.  Ignoring the call.");
            this._currentOpenOptions = null;
            this._disableAnimation = false;
            return;
        }

        var isDropDown = this._isDropDown(openOptions.display);
        this._toggleCancelDom(isDropDown);
        var position, modality;

        if (isDropDown) { // always true if there are submenus
            // .oj-menu-dropdown already added to any submenus in _setup, since dropdown/sheet status can't
            // vary by launch when there are submenus.
            this.element.addClass("oj-menu-dropdown").removeClass("oj-menu-sheet");
            modality = _DROPDOWN_MODALITY;
            
            // convert the position option back to JQuery format if custom element menu or submenu
            if (this._IsCustomElement())
            {
                // fixup the position option if custom element menu or submenu
                openOptions.position = oj.PositionUtils.coerceToJet(openOptions.position, this.options["openOptions"]["position"]);
                position = oj.PositionUtils.coerceToJqUi(openOptions.position);
            }
            else
                position = openOptions.position;
            position = oj.PositionUtils.normalizeHorizontalAlignment(position, this.isRtl);
            position.of = oj.PositionUtils.normalizePositionOf(position.of, launcher, event);

            // since already checked for null launcher above, this is only possible if "of" was "event" but the event was null.  Caller error.
            if (position.of == null) {
                oj.Logger.warn("position.of passed to Menu.open() is 'event', but the event is null.  Ignoring the call.");
                this._currentOpenOptions = null;
                this._disableAnimation = false;
                return;
            }
        } else { // sheet menu, implying no submenus
            this.element.addClass("oj-menu-sheet").removeClass("oj-menu-dropdown");
            modality = _SHEET_MODALITY;

            position = {
                "my": "bottom",
                "at": _SHEET_POSITION_AT,
                "of": window,
                "collision": "flipfit"
            };
        }

        //Close all other open menus
        var currentMenu = this.element[0];
        //Clone _openPopupMenus as __dismiss() will remove the open menu from _openPopupMenus list
        var openPopupMenus = _openPopupMenus.slice(0, _openPopupMenus.length);
        $.each(openPopupMenus , function(index, menu) {
            if (menu.element[0] !== currentMenu) {
                menu._collapse( event, "eventSubtree"); // TBD: should this be "all"?
                if (menu._launcher)
                    menu.__dismiss(event);
            }
        });

        // cache the merged value for use while the (outer) menu is still open
        this._submenuPosition = oj.PositionUtils.normalizeHorizontalAlignment(submenuOpenOptions.position, this.isRtl);

        var usingCallback = this._usingCallback;

        // if they provided a using function that is not our callback, stash it
        // away so that we can delegate to it in our proxy.
        if ($.isFunction(position["using"]) && position["using"] !== usingCallback)
          position["origUsing"] = position["using"];

        // override with our proxy to handle positioning of the tail
        position["using"] = usingCallback;

        this.element.data(_POSITION_DATA, position);

        this._setWhenReady("open");

        /** @type {!Object.<oj.PopupService.OPTION, ?>} */
        var psOptions = {};
        psOptions[oj.PopupService.OPTION.POPUP] = this.element;
        psOptions[oj.PopupService.OPTION.LAUNCHER] = launcher;
        psOptions[oj.PopupService.OPTION.POSITION] = position;
        psOptions[oj.PopupService.OPTION.EVENTS] = this._getPopupServiceEvents();
        psOptions[oj.PopupService.OPTION.LAYER_SELECTORS] = "oj-menu-layer";
        psOptions[oj.PopupService.OPTION.MODALITY] = modality;

        // local variables passed to the before and after open callbacks.
        psOptions[oj.PopupService.OPTION.CONTEXT] = {"event" : event,
                                                     "initialFocus": openOptions.initialFocus,
                                                     "launcher": launcher,
                                                     "isDropDown": isDropDown
                                                    };
        psOptions[oj.PopupService.OPTION.CUSTOM_ELEMENT] = this._IsCustomElement();

        oj.PopupService.getInstance().open(psOptions);
        
        this._disableAnimation = false;
    },
    
    /**
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {?Object} position object set as an option or passed as an argument to the open
     *                  method.
     */
    _setPosition : function (position)
    {
      var options = this.options; 
      
      // new position extends the existing object
      // covert to jet internal position format
      if (position)
        options["openOptions"]["position"] = oj.PositionUtils.coerceToJet(position, options["openOptions"]["position"]);
    },

    /**
     * Before open callback is called after the menu has been reparented into the
     * zorder container. Open animation is performed here.
     * @memberof! oj.ojMenu
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for opening the menu
     * @return {Promise|void}
     */
    _beforeOpenHandler : function (psOptions)
    {
      var promise;
      var rootElement = psOptions[oj.PopupService.OPTION.POPUP];
      var position = psOptions[oj.PopupService.OPTION.POSITION];
      var isDropDown = psOptions[oj.PopupService.OPTION.CONTEXT]['isDropDown'];

      rootElement.show();
      rootElement["position"](position);

      if (!this._isAnimationDisabled())
      {
        var animationOptions = this._getDefaultAnimation(isDropDown ? 'dropdown' : 'sheet', 'open');
        promise = oj.AnimationUtils.startAnimation(rootElement[0], "open", 
          oj.PositionUtils.addTransformOriginAnimationEffectsOption(rootElement, animationOptions), this);

        if (isDropDown)
        {
          // For dropdown, no need to return a promise from the open animation 
          // since the menu is ready for interaction as soon as it's visible,
          // and we want mouseenter events to be handled while animating.
          //
          // For sheet, we need to return the promise because it's coming in
          // from the bottom and afterOpenHandler set focus to it, which can
          // cause the layout to move around if the menu is not yet in view.
          promise = undefined;
        }
      }

      return promise;
    },

    /**
     * Called after the menu is shown. Perform open finalization.
     * @memberof! oj.ojMenu
     * @instance
     * @private
     * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for opening the menu
     * @return {Promise|void}
     */
    _afterOpenHandler : function (psOptions)
    {
      // From the context passed from the open restore local variable state.
      var context = psOptions[oj.PopupService.OPTION.CONTEXT];
      var event = context["event"];
      var initialFocus = context["initialFocus"];
      var launcher = context["launcher"];
      var isDropDown = context["isDropDown"];

      var focusFirstItem = initialFocus === "firstItem";
      var focusMenu = focusFirstItem || initialFocus === "menu" ;

      if (focusMenu)
          this.element.focus();

      if (focusFirstItem) {
          this._focus(event, this.element.children().first());
      } else {
          this._blur(event);
      }

      // store launcher so we can return focus to it, e.g. if Esc pressed.  Ivar is non-null iff
      // menu is currently open.
      this._launcher = launcher;
      this._sheetMenuIsOpen = !isDropDown;

      //Add current menu to openPopupMenus so that it will be closed on focus lost/click away.
      _openPopupMenus.push(this);

      this._trigger( "open", event, {});
    },

    _startOpening: function( event, submenu ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        this._clearTimer && this._clearTimer();

        // Don't open if already open fixes a Firefox bug that caused a .5 pixel
        // shift in the submenu position when mousing over the submenu icon
        if ( submenu.attr( "aria-hidden" ) !== "true" ) {
            return;
        }

        this._clearTimer = this._setTimer(function() {
            delete this._clearTimer;
            this._close();
            this._open( event, submenu );
        }, this._getSubmenuBusyStateDescription("closing and opening"), this.delay );
    },

    // opens a *sub*menu
    _open: function( event, submenu ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.

        // Don't open if already open.
        // Calling position() again when the submenu is animating will mess up
        // the calculation and make the submenu appear in different positions.
        if ( submenu.attr( "aria-hidden" ) !== "true" ) {
            return;
        }
        
        if (this._IsCustomElement())
        {
            var submenuWidget = this._getSubmenuWidget(submenu);
            var beforeOpenResults = submenuWidget._trigger2( "beforeOpen", event);
            
            if (!beforeOpenResults['proceed'])
                return;
            
            var position = oj.PositionUtils.coerceToJqUi(submenuWidget.options.openOptions.position);
            position = oj.PositionUtils.normalizeHorizontalAlignment(position, this.isRtl);
        }
        else
            position = this._submenuPosition;
        
        position = $.extend( {"of": this.active}, position); // normalizeHorizontalAlignment() was already called on the ivar

        this._clearTimer && this._clearTimer();
        this.element.find( ".oj-menu" ).not( submenu.parents( ".oj-menu" ) )
            .hide()
            .attr( "aria-hidden", "true" )
            .removeData(_POSITION_DATA);

        submenu
            .show()
            .removeAttr( "aria-hidden" )
            .position( position )
            .data(_POSITION_DATA, position);

        this._getSubmenuAnchor(submenu).attr( "aria-expanded", "true" );

        if (!this._launcher && _openPopupMenus.indexOf(this) < 0) {
            _openPopupMenus.push(this);
        }
        
        if (!this._isAnimationDisabled())
        {
          var animation = this._getDefaultAnimation('submenu', 'open');
          animation = this._replaceAnimationOptions(animation, {'#myPosition': position.my});
          oj.AnimationUtils.startAnimation(submenu[0], 'open', animation, this);
        }
    },

    /*
     * Same as calling _collapse(event, "eventSubtree") or _collapse(event, "all"), except that, if delay param is not passed, it collapses the menu immediately.
     */
    __collapseAll: function(event, all, delay) {
        var promise;

        this._clearTimer && this._clearTimer();
        var self = this;
        var collapseMenu = function() {
            delete self._clearTimer;
            // If we were passed an event, look for the submenu that contains the event
            var currentMenu = all ? self.element :
                $(event && event.target).closest(self.element.find(".oj-menu"));

            // If we found no valid submenu ancestor, use the main menu to close all sub menus anyway
            if (!currentMenu.length) {
                currentMenu = self.element;
            }

            var closePromise = self._close(currentMenu);

            closePromise = self._runOnPromise(closePromise, function() {
              self._blur(event);
              self.activeMenu = currentMenu;
            });

            return closePromise;
        };
        if (delay) {
          if (this._isAnimationDisabled())
          {
            self._clearTimer = self._setTimer(collapseMenu, self._getSubmenuBusyStateDescription("closing"), delay);
          }
          else
          {
            promise = new Promise(function(resolve, reject) {
              self._clearTimer = self._setTimer(collapseMenu, self._getSubmenuBusyStateDescription("closing"), delay, function() {
                resolve(true);
              });
            });
          }
        } else {
            promise = collapseMenu();
        }

        if (promise)
        {
          // Need to add busy state since submenu animation doesn't go through PopupService events
          var resolveBusyState = oj.Context.getContext(this.element[0]).getBusyContext().addBusyState({"description": "closing submenus"});

          // IMPORTANT: Do not change promise to the one returned by then().  Doing 
          // so will introduce an additional delay and disrupt the continuity of
          // busy state with any subsequent operation.
          promise.then(resolveBusyState);
        }
        
        return promise;
    },

    // With no arguments, closes the currently active menu - if nothing is active
    // it closes all menus.  If passed an argument, it will search for menus BELOW
    _close: function( startMenu ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.

        var closePromise;

        //TODO: Consider refatoring _close moving into the __dismiss logic.  The _close logic
        //      will hide levels of nested menus.  The __dismiss knocks down the root
        //      menu.  Both _close and _dismiss are called from _closeAll which closes
        //      all submenus and the main menu.

        if ( !startMenu ) {
            startMenu = this.active ? this.active.parent() : this.element;
        }

        var self = this;
        var defaultAnimation = self._getDefaultAnimation('submenu', 'close');
        var menus = startMenu.find( ".oj-menu" );
        var hideSubmenus = function(submenus) {
          submenus.hide()
                  .attr( "aria-hidden", "true" )
                  .removeData(_POSITION_DATA);
          self._getSubmenuAnchor( submenus ).attr( "aria-expanded", "false" );
        };

        if (this._isAnimationDisabled())
        {
          // If animation is hard-disabled, just hide all submenus at once
          hideSubmenus(menus);
          startMenu.find( ".oj-focus-ancestor" ).removeClass( "oj-focus-ancestor" );
        }
        else
        {
          // If there is animation, recursively hide submenus level by level,
          // starting from the innermost level.
          // There is no default close animation, but this allows app to define
          // cascading close animation if it wants to.
          
          // Define a recurive function that close all submenus of a menu
          var closeSubmenus = function(currentMenu)
          {
            var masterPromise = null;
            
            // Get <li> child elements of the current menu.  Submenus are rendered as
            // <ul> children of <li> elements.
            var items = currentMenu.children();
            
            // Find all the immediate child menus and iterate through them
            var childMenus = items.children( ".oj-menu" );
            childMenus.each(function(index, submenu) {
              var jSubmenu = $(submenu);
              
              // Define a function that animate the closing and hiding of the iterated menu
              var animateMenuClose = function() {
                var position = jSubmenu.data(_POSITION_DATA);
                var animation = self._replaceAnimationOptions(defaultAnimation, {'#myPosition': position ? position.my : null});
                return oj.AnimationUtils.startAnimation(submenu, 'close', animation, self).then(function() {
                  hideSubmenus(jSubmenu);
                });
              };

              if (jSubmenu.is(':visible'))
              {
                // If the iterated menu is visible, try to close its child menus first
                var promise = closeSubmenus(jSubmenu);

                // Wait for child menus to close before closing the iterated menu.
                //
                // Keep track of the closing promise for each menu that is visible
                // There should be at most one visible submenu at each level
                masterPromise = self._runOnPromise(promise, animateMenuClose);
              }
              else
              {
                // If the iterated menu is not visible, just hide it and set other attributes
                hideSubmenus(jSubmenu);
              }
            });

            // After iterating through all child menus, return a master promise
            return masterPromise;
          };

          // Start calling the recursive function from the outermost menu
          closePromise = closeSubmenus(startMenu);

          closePromise = this._runOnPromise(closePromise, function() {
            startMenu.find( ".oj-focus-ancestor" ).removeClass( "oj-focus-ancestor" );
          });
        }

        if (!this._launcher) {
            // If the current menu is not a popup menu and it's submenu is already open then remove the menu from _openPopupMenus
            // while closing the submenus of top level menu.
            if (_openPopupMenus.indexOf(this) >= 0) {
                if (startMenu === this.element) {
                    _openPopupMenus.splice(_openPopupMenus.indexOf(this), 1);
                }
            }
        }
        
        return closePromise;
    },

    /**
     * Closes one or more open submenus.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Event=} event - What triggered the menu to collapse.  May be <code class="prettyprint">null</code>.
     *                         May be omitted if the <code class="prettyprint">which</code> parameter is omitted.
     * @param {string=} which - Optional; defaults to <code class="prettyprint">"active"</code>.  Values are the following <code class="prettyprint">string</code>s:
     *     <ul>
     *       <li><code class="prettyprint">"active"</code>: Closes the currently active submenu.</li>
     *       <li><code class="prettyprint">"all"</code>: Closes all submenus.</li>
     *       <li><code class="prettyprint">"eventSubtree"</code>: Closes submenus below but not including the menu that is or contains the target of the triggering event.</li>
     *     </ul>
     */
    _collapse: function( event, which ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var promise;
        
        if (which == null || which === "active") {
            var newItem = this.activeMenu &&
                this.activeMenu.closest( ".oj-menu-item", this.element );
            if ( newItem && newItem.length ) {
                var self = this;
                promise = this._close();
                promise = this._runOnPromise(promise, function() {
                  self._focus( event, newItem );
                });
            }
        } else if ( which === "all" || which === "eventSubtree") {
            promise = this.__collapseAll(event, which === "all", this.delay);
        } else {
            oj.Logger.warn("Invalid param " + which + " passed to Menu._collapse().  Ignoring the call.");
        }

        return promise;
    },

    /**
     * Opens the submenu below the currently focused item, if one exists.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Event=} event - What triggered the menu to expand.  May be <code class="prettyprint">null</code> or omitted.
     */
    _expand: function( event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var newItem = this.active &&
            this.active
                .children( ".oj-menu " )
                .children( ".oj-menu-item" )
                .first();

        if ( newItem && newItem.length ) {
            this._open( event, newItem.parent() );

            this._clearTimer && this._clearTimer();

            // Delay so Firefox will not hide activedescendant change in expanding submenu from AT
            this._clearTimer = this._setTimer(function() {
                delete this._clearTimer;
                this._focus( event, newItem );
            }, this._getBusyStateDescription("focusing an item"));
        }
    },

    /**
     * Focuses the next menu item, wrapping at the bottom, as if <kbd>DownArrow</kbd> had been pressed.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Event=} event - What triggered the focus to move.  May be <code class="prettyprint">null</code> or omitted.
     */
    _next: function( event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        this._move( "next", "first", event );
    },

    /**
     * Focuses the previous menu item, wrapping at the top, as if <kbd>UpArrow</kbd> had been pressed.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Event=} event - What triggered the focus to move.  May be <code class="prettyprint">null</code> or omitted.
     */
    _previous: function( event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        this._move( "prev", "last", event );
    },

    _isFirstItem: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        return this.active && !this.active.prevAll( ".oj-menu-item" ).length;
    },

    _isLastItem: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        return this.active && !this.active.nextAll( ".oj-menu-item" ).length;
    },

    _move: function( direction, filter, event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var next;
        if ( this.active ) {
            if ( direction === "first" || direction === "last" ) {
                next = this.active
                    [ direction === "first" ? "prevAll" : "nextAll" ]( ".oj-menu-item" )
                    .eq( -1 );
            } else {
                next = this.active
                    [ direction + "All" ]( ".oj-menu-item" )
                    .eq( 0 );
            }
        }
        if ( !next || !next.length || !this.active ) {
            next = this.activeMenu.children( ".oj-menu-item" )[ filter ]();
        }

        this._focus( event, next );
    },

    /* TODO: update JSdoc to be something like this revised version, once todo's in code are resolved.
     * Let selectItem be the currently focused menu item if any, else the menu item containing the target of the supplied event if any, else null.
     *
     * If selectItem is non-null, this method selects that item, collapses all submenus, and triggers the menu's
     * <code class="prettyprint">select</code> event.
     *
     * Internally, this method should not be invoked for parent menu items or disabled menu items. But still there is a chance of
     * invoking _select() externally. (Not anymore now that it's private.) In that case, if focused menu item is a disabled or parent menu item then a warning message will be logged.
     */
    /**
     * Selects the currently focused menu item, collapses all submenus and triggers the menu's <code class="prettyprint">select</code> event.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {Event=} event - What triggered the selection.  May be <code class="prettyprint">null</code> or omitted.
     */
    _select: function( event ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        // If no menu item is focused, then infer one from the event if possible.  TBD: still need this now that not public?  Or is this.active always set?
        if (!this.active && event && event.target ) {
            var menuItem = $( event.target ).closest( ".oj-menu-item" );
            if (menuItem.closest( this.element ).length)
                this._makeActive(menuItem, event);
        }

        if (!this.active) {
            // If we reach here, this must be an external call to the method.
            oj.Logger.warn("Menu._select() called when no menu item is focused and no menu item can be inferred from event param.");
            return;
        }

        // now we know this.active is non-null
        if (this.active.has( ".oj-menu" ).length || this.active.is( ".oj-disabled" )) {
            // If we reach here, this must be an external call to the method.
            oj.Logger.warn("Selecting a disabled menu item or parent menu item is not allowed.");
            return;
        }

        // payload for select event
        var ui = this.active.is(this._cancelItem)
            ? undefined // don't fire select for Cancel item
            : { "item": this.active }; // must grab this.active before calling __collapseAll, which clears this.active

        // The menu item has been selected, so we can collapse all menus immediately with no timeout via __collapseAll.
        // If we call the version with a timeout, _collapse(event, "all"), then mouseleave event handler will invoke _collapse(event, "eventSubtree") on event.target
        // which will clear our scheduled _collapse(event, "all") on this.element, so that submenu will not be collapsed,
        // which means that when the menu is later re-launched, the submenu is already open.
        var promise = this.__collapseAll( event, true );

        // if menu is currently open
        if (this._launcher) {
            var self = this;
            this._runOnPromise(promise, function() {
              self._focusLauncherAndDismiss(event, ui);
            });
        }
    },

   /**
    * @instance
    * @private
    */
    _surrogateRemoveHandler: function()
    {
      var element = this.element;
      element.remove();
    },

   /**
    * @instance
    * @private
    * @return {!Object.<oj.PopupService.EVENT, function(...)>}
    */
    _getPopupServiceEvents: function()
    {
      if (!this._popupServiceEvents)
      {
        /** @type {!Object.<oj.PopupService.EVENT, function(...)>} **/
        var events = this._popupServiceEvents = {};
        events[oj.PopupService.EVENT.POPUP_CLOSE] = this._closeAll.bind(this);
        events[oj.PopupService.EVENT.POPUP_REMOVE] = this._surrogateRemoveHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_REFRESH] = this.refresh.bind(this);
        events[oj.PopupService.EVENT.POPUP_AUTODISMISS] = this._clickAwayHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_BEFORE_OPEN] = this._beforeOpenHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_AFTER_OPEN] = this._afterOpenHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_BEFORE_CLOSE] = this._beforeCloseHandler.bind(this);
        events[oj.PopupService.EVENT.POPUP_AFTER_CLOSE] = this._afterCloseHandler.bind(this);

      }
      return this._popupServiceEvents;
    },

   /**
    * @instance
    * @private
    */
    _closeAll: function()
    {
      // TODO: Don't want to animate when force close a menu. This flag is not
      // hooked in the beforeCloseHandler.  There is only a custom element check. Doesn't
      // currently matter because the menu doesn't support custom element syntax yet.
      this._disableAnimation = true;

      this._close(this.element);
      this.__dismiss(null);
      // Forced menu dismissal doesn't queue the close event for some reason. The busy
      // state is resolved on the close event from the mediator.  Force the busy state to release.
      this._setWhenReady("none");
    },

   /**
    * @private
    * @param {Object} pos "my" element associated with the position object
    * @param {Object} props directions as to where the element should be moved
    */
    _usingHandler: function(pos, props)
    {
      var rootMenu = props["element"]["element"];
      rootMenu.css(pos);

      // Capture the position data so that we can set transform-origin later on
      oj.PositionUtils.captureTransformOriginAnimationEffectsOption(rootMenu, props);

      // call on the original using
      var position = rootMenu.data(_POSITION_DATA);
      if (position)
      {
        var origUsing = position["origUsing"];
        if (origUsing)
          origUsing(pos, props);
      }

      // implicitly dismiss the menu when the position.of is clipped in an overflow container.
      if (oj.PositionUtils.isAligningPositionClipped(props))
        this._clearCloseDelayTimer = this._setTimer(this._closeAll, this._getSubmenuBusyStateDescription("closing"), 1);
    },

    // @inheritdoc
    getNodeBySubId: function(locator) {
        switch (locator && locator['subId']) {
            case _SUBID_CANCEL:
                // only return it if it's currently in the DOM, never when it's detached
                return this._cancelDomAttached ? this._cancelItem[0] : null;
            default:
                return this._super(locator);
        }
    },

    /**
     * {@ojinclude "name":"getSubIdByNodeDesc"}
     *
     * @expose
     * @ignore
     * @memberof oj.ojMenu
     * @instance
     * @since 2.1.0
     *
     * @param {!Element} node {@ojinclude "name":"getSubIdByNodeNodeParam"}
     * @return {Object|null} {@ojinclude "name":"getSubIdByNodeReturn"}
     *
     * @example <caption>{@ojinclude "name":"getSubIdByNodeCaption"}</caption>
     * {@ojinclude "name":"getSubIdByNodeExample"}
     */
    getSubIdByNode: function(node) {
        return this._cancelItem && this._cancelItem.is($(node).parents().addBack(node))
            ? {'subId':_SUBID_CANCEL}
            : this._super(node);
    },

    /**
     * Called on menu open.
     * @private
     * @param {string} display - the display value for the current launch, before resolving "auto"
     */
    _isDropDown: function(display) {
        if (this._hasSubmenus)
            return true;

        switch (display) {
            case "dropDown":
                return true;
            case "sheet":
                return false;
            case "auto":
                return _DISPLAY_QUERY.matches;
            default:
                throw new Error("Invalid value for Menu openOptions.display: " + display);
        }
    },

    /**
     * Called on menu open. Adds or removes the Cancel menu item and its divider, as needed.
     * @private
     * @param {boolean} isDropDown
     */
    _toggleCancelDom: function(isDropDown) {
        if (!_SHEETS_HAVE_CANCEL_ITEM)
            return; // shouldn't add cancel DOM, and no need to remove it since it could never have been added

        if (isDropDown) {
            if (this._cancelDomAttached) {
                this._getCancelDom().detach()
                    .eq(0).prev().removeClass( "oj-menu-item-before-divider" );
                this._cancelDomAttached = false;
            }
        } else {
            // if detached, adds it.  If attached, ensures it's at the end of the menu where it belongs,
            // even if app or component has appended menu items.
            this._getCancelDom().appendTo(this.element) // @HTMLUpdateOK trusted string per annotations in callee
                .eq(0).prev().addClass( "oj-menu-item-before-divider" );
            this._cancelDomAttached = true;
        }
    },

    /**
     * Called on menu open when menu needs to add a cancel item for this launch.
     * @private
     * @return {jQuery} JQ object with Cancel divider and Cancel menu item in correct order
     */
    _getCancelDom: function() {
        if (!this._cancelDom) {
            var divider = $("<li></li>", this.document[0]); // @HTMLUpdateOK trusted string
            var a = $("<a href='#'></a>", this.document[0]) // @HTMLUpdateOK trusted string
                .text(this.options['translations']['labelCancel']);
            $("<span class='oj-menu-item-icon oj-component-icon oj-menu-cancel-icon'></span>", this.document[0])
                .prependTo(a); // @HTMLUpdateOK trusted string
            var li = $("<li></li>", this.document[0])
                .addClass( "oj-menu-item-cancel oj-menu-item-after-divider" )
                .append(a); // @HTMLUpdateOK trusted string

            this._initDividers(divider);
            this._initAnchors(a);
            this._initMenuItems(li);

            this._cancelAnchor = a;
            this._cancelItem = li;
            this._cancelDom = $([divider[0], li[0]]); // need array-of-elem syntax to guarantee order
        }
        return this._cancelDom;
    },

    /**
     * Called at create. Sets up Hammer swipe-down-to-dismiss-menu listener if enabled via SASS var.
     * @private
     */
    _setupSwipeBehavior: function() {
        if (!_SHEETS_HAVE_SWIPE_DOWN_TO_DISMISS)
            return;

        this.element.ojHammer(_HAMMER_OPTIONS);
        this._on({
            "swipedown": function(event) {
                // important to check "sheet menu currently open", not "last launch was sheet menu",
                // since a single swipe can fire 2 swipe events, and the 2nd one finds the menu already
                // closed and NPE's if it enters the "if" block
                if (this._sheetMenuIsOpen && event["gesture"]["pointerType"] === "touch") { // Hammer events fire for mouse too
                    this.__collapseAll( event, true );
                    this._focusLauncherAndDismiss(event);
                }
            }
        });
    },

    /**
     * Creates a Promise exposed by the {@link oj.ojMenu#whenReady} method.
     *
     * @param {string} operation valid values are "open", "close" or "none"
     * @memberof oj.ojMenu
     * @instance
     * @private
     */
    _setWhenReady : function (operation)
    {
      /** @type {oj.PopupWhenReadyMediator} */
      var mediator = this._whenReadyMediator;
      if (mediator)
      {
        mediator.destroy();
        delete this._whenReadyMediator;
      }

      // operation === none
      if (["open", "close"].indexOf(operation) < 0)
        return;

      this._whenReadyMediator = new oj.PopupWhenReadyMediator(this.element, operation, "ojMenu",
        this._IsCustomElement());
    },

    /**
     * Checks to see if there is a pending "open" or "close" operation.  If pending and it
     * is the same as the requested operation, the request silently fails.  If the current
     * operation is the inverse operation, we queue the current operation after the pending
     * operation is resolved.
     *
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {string} operation currently requested
     * @param {string} methodName to invoke to trigger the target operation
     * @param {Array} methodArgs passed to a queue operation
     * @return {boolean} <code>true</code> if a close or open operation is pending completion.
     */
    _isOperationPending: function (operation, methodName, methodArgs)
    {
      /** @type {oj.PopupWhenReadyMediator} **/
      var mediator = this._whenReadyMediator;
      if (mediator)
        return mediator.isOperationPending(this, operation, methodName, methodArgs);
      else
        return false;
    },

    /**
     * Adds a busy state with the specified description.
     *
     * Asynchronously, after the specified delay in ms, calls callback with "this" bound to this menu instance, 
     * and then resolves the busy state.
     *
     * Returns a "cancel" function that if called:
     * - Cancels the timer, so that the callback is never called (unless it has already been called).  
     * - Resolves the busy state.
     * 
     * @memberof oj.ojMenu
     * @instance
     * @private
     * @param {function()} callback
     * @param {string} description
     * @param {number=} delay in ms. Defaults to 0.
     * @param {function()=} notifier - A function to notify the caller of several conditions so that it can clean up:
     *                                 1. The timer is cancelled before "callback" is called
     *                                 2. "callback" is called and doesn't return a promise
     *                                 3. "callback" is called, returns a promise, and the promise is resolved
     * @return {function()} a "cancel" function as described above
     */
    _setTimer: function (callback, description, delay, notifier) {
        // Call this line each time rather than caching busyContext
        var resolve = oj.Context.getContext(this.element[0]).getBusyContext().addBusyState({"description": description});

        // resolve() bombs if called a 2nd time, so prevent that possibility by wrapping it in a function that can't call 
        // it twice, and never calling resolve() directly. If that "bombs 2nd time" behavior is removed from the BusyContext 
        // framework, can remove this "resolveOnce" wrapper var, and just call resolve() below.
        var resolveOnce = function() {
            if (resolve) {
                resolve();
                resolve = null;
                notifier && notifier();
            }
        };

        var self = this;

        var id = setTimeout(function() {
            var result = callback.bind(self)();
            if (result && result instanceof Promise)
            {
              // If the callback returns a promise, resolve busy state when the promise resolves.
              result.then(resolveOnce);
            }
            else
            {
              // If the callback doesn't return a promise, just resolve the busy state
              resolveOnce();
            }
        }, delay || 0);

        return function() {
            clearTimeout(id);
            resolveOnce();
        };
    },

    // action is "focusing an item", ...
    _getBusyStateDescription: function (action) {
        return "Menu with id '" + this.element.attr("id") + "' is busy " + action + ".";
    },

    // action is "opening", "closing", "closing and opening", ...
    _getSubmenuBusyStateDescription: function (action) {
        return this._getBusyStateDescription(action + " a submenu");
    },
   /**
    * Notifies the component that its subtree has been removed from the document
    * programmatically after the component has been created.
    *
    * @memberof oj.ojMenu
    * @instance
    * @protected
    * @override
    */
    _NotifyDetached: function()
    {
      // detaching an open menu results in implicit dismissal
      if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
        this._closeAll();

      this._super();
    }
    
    // Override contextMenu slot definition to remove it from the jsdoc as it is not supported for menus
    
    /**
     * @ojslot contextMenu
     * @memberof oj.ojMenu
     * @ignore
     */

    // API doc for inherited methods with no JS in this file:

    /**
     * Returns a <code class="prettyprint">jQuery</code> object containing the root element of the Menu component.
     *
     * @method
     * @name oj.ojMenu#widget
     * @memberof oj.ojMenu
     * @instance
     * @ignore
     * @return {jQuery} the root element of the component
     */

    /**
     * Removes the menu functionality completely. This will return the element back to its pre-init state.
     *
     * @method
     * @name oj.ojMenu#destroy
     * @memberof oj.ojMenu
     * @instance
     * @ignore
     */

    // Fragments:

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Menu Item</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Invoke the menu item's action.</td>
     *     </tr>
     *     <tr>
     *       <td>Menu</td>
     *       <td><kbd>Swipe Down</kbd></td>
     *       <td>Dismiss the menu, if "swipe to dismiss" is enabled by the application.</td>
     *     </tr>
     *     <tr>
     *       <td>JET Component or HTML Element having a JET Context Menu</td>
     *       <td><kbd>Press & Hold</kbd></td>
     *       <td>Open the context menu.</td>
     *     </tr>
     *     <tr>
     *       <td>Outside of Menu</td>
     *       <td><kbd>Touch</kbd></td>
     *       <td>Close the menu.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>Disabled items do not allow any touch interaction.
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojMenu
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td rowspan = "6">Menu Item</td>
     *       <td><kbd>Enter</kbd> or <kbd>Space</kbd></td>
     *       <td>Invoke the focused menu item's action.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>Move focus to the previous menu item, wrapping around at the top.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Move focus to the next menu item, wrapping around at the bottom.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>Home</kbd></td>
     *       <td>Move focus to the first menu item.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>End</kbd></td>
     *       <td>Move focus to the last menu item.</td>
     *     </tr>
     *     <tr>
     *       <td>Menu Item in Top-level Menu</td>
     *       <td><kbd>Esc</kbd></td>
     *       <td>Close the menu and move focus to the launcher.</td>
     *     </tr>
     *     <tr>
     *       <td>JET Component or HTML Element having a JET Context Menu</td>
     *       <td><kbd>Shift + F10</kbd></td>
     *       <td>Open the context menu.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>* RTL refers to pages written in a right-to-left language such as Arabic.
     *
     * <p>Typing a letter moves focus to the first item whose title starts with that character. Repeating the same character cycles through matching items.
     * Typing more characters within the one second timer matches those characters.
     *
     * <p>Note that the "Search for text when I start typing" feature in Firefox can interfere with web content that accepts keystrokes, such as this "type a letter" feature of JET Menu.
     *
     * <p>Disabled items can receive keyboard focus, but do not allow any other interaction.
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojMenu
     */
});

oj.Components.setDefaultOptions(
{
  'ojMenu': // properties for all menu components 
  {
    'openOptions': oj.Components.createDynamicPropertyGetter(function(context){
      return {
        'position': (oj.BaseCustomElementBridge.getRegistered(context['element'].tagName) != null) ?
          (context['containers'].indexOf('ojMenu') >= 0 ?
            { my : { horizontal : "start", vertical : "top"}, at : { horizontal : "end", vertical : "top" }, offset : { x : 0, y : 0 }, collision : "flipfit" } :
            { my : { horizontal : "start", vertical : "top"}, at : { horizontal : "start", vertical : "bottom" }, offset : { x : 0, y : 0 }, collision : "flipfit" }) :
          { my : "start top", at : "start bottom", collision : "flipfit" }
      };
    })
  }
});

//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the <a href="#dismissal-section">"Cancel"</a> menu item.</p>
 *
 * @ojsubid oj-menu-cancel-command
 * @memberof oj.ojMenu
 * @since 2.1.0
 *
 * @example <caption>Get the node for the "Cancel" menu item:</caption>
 * var node = myMenu.getNodeBySubId( {'subId': 'oj-menu-cancel-command'} );
 */

// -----------------------------------------------------------------------------
// "private static members" shared by all menus
// -----------------------------------------------------------------------------

// Array to track all opened menu popups. All the menus opened by Menu Buttons/ Context Menu/using menu.open() and standalone menus having open submenus, will be added to list
// and later will be removed on __dismiss()/_close() on menu popup/standalone menu.
var _openPopupMenus = new Array();

// See usage for explanation.  Can be boolean (doesn't need to be re-entrant int), since
// baseComponent.touchendMousedownThreshold is much less than baseComponent.pressHoldThreshold.
var _contextMenuPressHoldJustEnded = false

/**
 * Key used to store the menu's position object as a jQuery data property.
 * @const
 * @private
 * @type {string}
 */
var _POSITION_DATA = "oj-menu-position";

// Used to suppress focus ring for Mac Safari due to platform repainting bug.
// This returns true for Mac Safari, but not for desktop Chrome, FF, IE11, Edge;
// Mac Chrome, FF; iOS Safari; or Android Chrome.
// Using "Mac" instead of "Macintosh" in this check would return true for Mac
// Safari and iOS Safari, but none of the others.
var _IS_MAC_SAFARI = navigator.userAgent.indexOf("Macintosh") > -1
    && navigator.userAgent.indexOf("Safari") > -1
    && navigator.userAgent.indexOf("Chrome") === -1;

var _config = oj.ThemeUtils.parseJSONFromFontFamily('oj-menu-config') || {};
var _SHEETS_HAVE_CANCEL_ITEM = _config['sheetCancelAffordance'] === "menuItem";
var _SHEETS_HAVE_SWIPE_DOWN_TO_DISMISS = _config['sheetSwipeDownBehavior'] === "dismiss";
var _DROPDOWN_MODALITY = _config['dropDownModality'] || "modeless"; // backward-compatible value
var _SHEET_MODALITY = _config['sheetModality'] || "modal";

// "bottom-0", "bottom-12", or "bottom-10%", per standard JQUI position utility syntax. Put the
// minus on position's "at", not "my", so that %'s (which we're not using, but themers might) are relative to window,
// not menu, thus closer to behavior of using % margin in CSS.  SCSS code comment on our $var says we use "at".
var _SHEET_POSITION_AT = "bottom-" + (_config['sheetMarginBottom'] || 0);

var _HAMMER_OPTIONS = _SHEETS_HAVE_SWIPE_DOWN_TO_DISMISS && {
    "recognizers": [
        [Hammer.Swipe,{ "direction": Hammer["DIRECTION_DOWN"] }]
    ]
};

// media query for when display=="auto"
var _DISPLAY_QUERY = function() {
    // If "display" option is "auto" (the default), then at or above this screen width
    // the option resolves to "dropDown", and below this screen width it resolves to "sheet".
    var dropDownThresholdWidth = _config['dropDownThresholdWidth'];
    if (dropDownThresholdWidth == null)
        dropDownThresholdWidth = "768px"; // min width of "medium" screen size
    var queryString = "(min-width: " + dropDownThresholdWidth + ")";
    return window.matchMedia(queryString);
}();

var _SUBID_CANCEL = 'oj-menu-cancel-command';

}());

(function() {
var ojMenuMeta = {
  "properties": {
    "disabled": {
      "type": "boolean"
    },
    "openOptions": {
      "type": "object",
      "properties": {
        "display": {
          "type": "string",
          "enumValues": ["auto", "dropDown", "sheet"]
        },
        "initialFocus": {
          "type": "string",
          "enumValues": ["firstItem", "menu", "none"]
        },
        "launcher": {
          "type": "string"
        },
        "position": {
          "type": "object",
          "properties": {
            "my": {
              "type": "string|object",
              "properties": {
                "horizontal": {
                  "type": "string",
                  "enumValues": ["start", "end", "left", "center", "right"]
                },
                "vertical": {
                  "type": "string",
                  "enumValues": ["top", "center", "bottom"]
                }
              }
            },
            "at": {
              "type": "string|object",
              "properties": {
                "horizontal": {
                  "type": "string",
                  "enumValues": ["start", "end", "left", "center", "right"]
                },
                "vertical": {
                  "type": "string",
                  "enumValues": ["top", "center", "bottom"]
                }
              }
            },
            "offset": {
              "type": "object",
              "properties": {
                "x": {
                  "type": "number"
                },
                "y": {
                  "type": "number"
                }
              }
            },
            "of": {
              "type": "string|object"
            },
            "collision": {
              "type": "string",
              "enumValues": ["flip", "fit", "flipfit", "flipcenter", "none"]
            }
          }
        },
        "submenuPosition": {
          "type": "object",
          "properties": {
            "my": {
              "type": "object",
              "properties": {
                "horizontal": {
                  "type": "string",
                  "enumValues": ["start", "end", "left", "center", "right"]
                },
                "vertical": {
                  "type": "string",
                  "enumValues": ["top", "center", "bottom"]
                }
              }
            },
            "at": {
              "type": "object",
              "properties": {
                "horizontal": {
                  "type": "string",
                  "enumValues": ["start", "end", "left", "center", "right"]
                },
                "vertical": {
                  "type": "string",
                  "enumValues": ["top", "center", "bottom"]
                }
              }
            },
            "offset": {
              "type": "object",
              "properties": {
                "x": {
                  "type": "number"
                },
                "y": {
                  "type": "number"
                }
              }
            },
            "collision": {
              "type": "string",
              "enumValues": ["flip", "fit", "flipfit", "flipcenter", "none"]
            }
          }
        }
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "labelCancel": {
          "type": "string",
          "value": "Cancel"
        }
      }
    }
  },
  "methods": {
    "getSubIdByNode": {},
    "open": {},
    "refresh": {}
  },
  "events": {
    "animateStart" : {},
    "animateEnd" : {},
    "beforeOpen": {},
    "close": {},
    "open": {}
  },
  "extension": {
    _WIDGET_NAME: "ojMenu"
  }
};
oj.CustomElementBridge.registerMetadata('oj-menu', 'baseComponent', ojMenuMeta);
oj.CustomElementBridge.register('oj-menu', {'metadata' : oj.CustomElementBridge.getMetadata('oj-menu')});
})();
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojMessage
 * @since 5.0.0
 * @ojdisplayname Message
 * @ojshortdesc Displays a message
 * @ojstatus preview
 * @ojsignature {target: "Type", value:"class ojMessage extends JetElement<ojMessageSettableProperties>"}
 *
 * @classdesc
 * <h3 id="messageOverview-section">
 *   JET Message
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#messageOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>Displays a message. All <code class="prettyprint">oj-message</code> elements in a page must
 * have an <code class="prettyprint">oj-messages</code> element as its parent.</p>
 *
 * <pre class="prettyprint">
 * <code>&lt;oj-message id="simpleMessage" message="[[messageData]]">
 * &lt;/oj-message>
 * </code></pre>
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>The {@link oj.ojMessage#message.sound} property is an accessibility feature for playing a
 * sound when a message is opened. This property defaults to "none", and can be enabled by setting
 * it to "defaults" or by providing URL to an audio file of a format that the browser supports. An 
 * accessible application must provide a way for users to enable sound on a settings or preferences 
 * page. Some browsers will have auto-play disabled by default, enabling it may require adjusting 
 * the browser settings.</p>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Message</td>
 *       <td><kbd>Swipe</kbd></td>
 *       <td>Close the message</td>
 *     </tr>
 *     <tr>
 *       <td>Message Close Icon</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Close the message</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojMessage
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Message</td>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Close the message</td>
 *     </tr>
 *     <tr>
 *       <td>Message Close Icon</td>
 *       <td><kbd>Enter or Space</kbd></td>
 *       <td>Close the message</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojMessage
 */

// Attributes
/////////////

/**
 * <p>Structured data specifying essential information to display a message on the UI.
 * @member
 * @name message
 * @memberof! oj.ojMessage
 * @instance
 * @since 5.0.0
 * @type {Object}
 * @default {"icon": "", 'category': "", "severity": "none", "summary": "", "detail": "", "autoTimeout": -1, "closeAffordance": "defaults", "sound": "none"}
 * @ojsignature { target: "Type",
 *                value: "oj.ojMessage.Message",
 *                jsdocOverride: true}
 *
 * @example <caption>Initialize the element with <code class="prettyprint">message</code> attribute:</caption>
 * &lt;!-- Binding message attribute to a structured object -->
 * &lt;oj-message message="[[myMessageData]]">&lt;/oj-message>
 *
 * &lt;!-- Setting message using JSON notation -->
 * &lt;oj-message message='{"severity": "error", "summary": "Some summary", "detail": "Some detail"}'>&lt;/oj-message>
 *
 * &lt;!-- Setting the message sub-attributes -->
 * &lt;oj-message message.severity="error" message.summary="Some summary" message.detail="Some detail">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">message</code> property after initialization:</caption>
 * // getter
 * var message = myMessage.message;
 *
 * // setter - sets all sub-properties (missing sub-properties are defaulted)
 * myMessage.message = {
 *     severity: "error",
 *     summary: "Some summary",
 *     detail: "Some detail"
 * };
 */

/**
 *<p>Specifies the URL for the custom image to be used as an icon representing the message.</p>
 * <p>The icon will be rendered as background image inside a container that is set to size of 16px*16px
 * in alta-web theme and 10px*20px for all other themes, therefore, the icon chosen must fit this
 * dimensions.</p>
 * <p>If this attribute is not specified, a suitable icon corresponding to value of
 * <code class="prettyprint">message.severity</code> will be rendered.</p>
 *
 * @example <caption>Initialize the element with <code class="prettyprint">message.icon</code> attribute:</caption>
 * &lt;oj-message message.icon="images/emailIcon.png" message.summary="Some summary">&lt;/oj-messages>
 *
 * @example <caption>Get or set the <code class="prettyprint">icon</code> property after initialization:</caption>
 * // getter
 * var icon = myMessage.getPropoerty("message.icon");
 *
 * // setter
 * myMessage.setProperty("message.icon", "images/emailIcon.png");
 *
 * @expose
 * @type {string}
 * @name message.icon
 * @default ""
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 */

/**
 *<p>Specifies text representing the message category which is shown next to the message icon</p>
 * <p>If this attribute is not specified, a translated text corresponding to value of
 * <code class="prettyprint">message.severity</code> attribute will be rendered.</p>
 *
 * @example <caption>Initialize the element with <code class="prettyprint">message.category</code> attribute:</caption>
 * &lt;oj-message message.category="Service Request" message.summary="New SR - Escalated">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">message.category</code> property after initialization:</caption>
 * // getter
 * var category = myMessage.getProperty("message.category");
 *
 * // setter
 * myMessage.setProperty("message.category", "Service Request");
 *
 * @expose
 * @type {string}
 * @name message.category
 * @default ""
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 */

/**
 *<p>Specifies the severity of message.</p>
 *
 *@example <caption>Initialize the element with <code class="prettyprint">message.severity</code> attribute:</caption>
 * &lt;oj-message message.severity="error" message.summary="Some summary">&lt;/oj-message>
 *
 *@example <caption>Get or set the <code class="prettyprint">message.severity</code> property after initialization:</caption>
 * // getter
 * var severity = myMessage.getProperty("message.severity");
 *
 * // setter
 * myMessage.setProperty("message.severity", "error");
 *
 * @expose
 * @type {string}
 * @name message.severity
 * @ojvalue {string} "error" error level message
 * @ojvalue {string} "warning" warning level message
 * @ojvalue {string} "confirmation" confirmation message
 * @ojvalue {string} "info" informational message
 * @ojvalue {string} "none" message status level not applicable
 * @default "none"
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 */

/**
 *<p>Specifies timestamp for the message to be displayed in the message header. </p>
 *<p> timestamp could represent the date and time at which the message was created, or otherwise
 *could pertain to the event for which the message was created. For example, a timestamp for an
 *upcoming meeting could be set in the future, whereas a timestamp for a missed message could be set
 *in the past.</p>
 *<p> This specified value must be an ISOString. A default converter is used to convert and format
 *the value suitable for displaying in the message. This default convertor used such will be an
 *implementation detail and could change in future.</p>
 *
 * @example <caption>Initialize the element with <code class="prettyprint">message.timestamp</code>
 * attribute:</caption>
 * &lt;oj-message message.summary="Some summary" message.timestamp="2018-03-09T13:10:47+14:00">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">message.timestamp</code> property
 * after initialization:</caption>
 * // getter
 * var timestamp = myMessage.getProperty("message.timestamp");
 *
 * // setter
 * myMessage.setProperty("message.timestamp", "2018-03-09T13:10:47+14:00");
 *
 * @expose
 * @type {string}
 * @name message.timestamp
 * @default ""
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 */

/**
 *<p>Specifies summary text for the message.
 *
 * @example <caption>Initialize the element with <code class="prettyprint">message.summary</code> attribute:</caption>
 * &lt;oj-message message.summary="Some summary">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">message.summary</code> property after initialization:</caption>
 * // getter
 * var summary = myMessage.getProperty("message.summary");
 *
 * // setter
 * myMessage.setProperty("message.summary", "Some summary");
 *
 * @expose
 * @type {string}
 * @name message.summary
 * @default ""
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 */

/**
 *<p>Specifies detail text for the message.</p>
 *
 * @example <caption>Initialize the element with <code class="prettyprint">message.detail</code> attribute:</caption>
 * &lt;oj-message message.detail="Some detail" message.summary="Some summary">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">message.detail</code> property after initialization:</caption>
 * // getter
 * var detail = myMessage.getProperty("message.detail");
 *
 * // setter
 * myMessage.setProperty("message.detail", "Some detail");
 *
 * @expose
 * @type {string}
 * @name message.detail
 * @default ""
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 */

/**
 * <p>Specifies the number of milliseconds for which duration the message will be shown before it is 
 * closed automatically.</p>
 * <p>This attribute can take the number of millisecond or special number values -1 and 0. If the 
 * value is set to "-1", auto-close will be turned off. If the value is set to "0", application wide 
 * value for autoTimeout as specified in a theming variable will be used.
 * 
 * @example <caption>Initialize the element with
 * <code class="prettyprint">message.auto-timeout</code> attribute:</caption>
 * &lt;oj-message message.auto-timeout="3000" message.summary="Some summary">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">message.autoTimeout</code>
 * property after initialization:</caption>
 * // getter
 * var autoTimeout = myMessage.getProperty("message.autoTimeout");
 *
 * // setter
 * myMessage.message.autoTimeout = 3000;
 *
 * @expose
 * @type {number}
 * @name message.autoTimeout
 * @default -1
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 *
 * @example <caption>The default auto-timeout value can be set at application level using this
 *          theme variable (SCSS) :</caption>
 * $messageAutoTimeoutOptionDefault: 5000 !default;
 */

/**
 *<p>Specifies the UI affordance provided to end users to be able to close the message.</p>
 * @example <caption>Initialize the element with
 * <code class="prettyprint">message.close-affordance</code> attribute:</caption>
 * &lt;oj-message message.close-affordance="none" message.summary="Some summary">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">message.closeAffordance</code>
 * property after initialization:</caption>
 * // getter
 * var closeAffordance = myMessage.getProperty("message.closeAffordance");
 *
 * // setter
 * myMessage.setProperty("message.closeAffordance", "none");
 *
 * @expose
 * @type {string}
 * @name message.closeAffordance
 * @ojvalue {string} "none" no UI affordance is provided to close the message. Application has to 
 * call the close() method to dismiss the message
 * @ojvalue {string} "defaults" use implicit affordance to best suit the native theme and screen
 * touch capabilities. See keyboard and touch end user information sections in this document.
 * @default "defaults"
 * @instance
 * @access public
 * @memberof! oj.ojMessage
 * @since 5.0.0
 */

/**
 * <p>Specifies the sound to be played when a message is opened. Sound is an accessibility feature 
 * required for low vision users who view a zoomed section of the UI. Because messages may be shown 
 * outside of the zoomed section, such users require sound to be played to notify of new messages.</p>
 * 
 * <p>This attribute can take a URL of the audio file for the custom sound to be played. The 
 * supported formats are mp3, wav and ogg. Browser support should also be considered while choosing 
 * the format of the audio file. Literal string values 
 * <code class="prettyprint">"defaults"</code> and <code class="prettyprint">"none"</code> can also 
 * be used for this attribute. If the value is set to "none", then the sound will be disabled. If 
 * the value is set to "defaults", then a default sound is played.<p>
 * 
 * <p>The default sound uses Web Audio APIs, which is not yet supported by some browsers, default 
 * sound will not be played in such browsers. Sound will not be played in browsers where auto play 
 * is not enabled. Some of the browsers do not allow auto play, while other browsers may provide a 
 * user preference to enable it.</p>
 *
 *
 * @example <caption>Initialize the element with <code class="prettyprint">message.sound</code> attribute:</caption>
 * &lt;!-- Use an URL -->
 * &lt;oj-message message.sound="resources/audio/newEmail.wav" message.summary="Some summary">&lt;/oj-message>
 *
 * &lt;!-- Use defaults -->
 * &lt;oj-message message.sound="defaults" message.summary="Some summary">&lt;/oj-message>
 *
 * &lt;!-- Turn off sound -->
 * &lt;oj-message message.sound="none" message.summary="Some summary">&lt;/oj-message>
 *
 * @example <caption>Get or set the <code class="prettyprint">sound</code> property after initialization:</caption>
 * // getter
 * var sound = myMessage.getProperty("message.sound");
 *
 * // setter
 * myMessage.setProperty("message.sound", "resources/audio/newEmail.wav");
 *
 * @expose
 * @type {string}
 * @name message.sound
 * @default "none"
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessage
 */

/**
 * <p>A collection of translated resources from the translation bundle, or
 * <code class="prettyprint">null</code> if this component has no resources. Resources may be
 * accessed and overridden individually or collectively, as seen in the examples.
 *
 * <p>If this component has translations, their documentation immediately follows this doc entry.
 *
 * @member
 * @name translations
 * @memberof! oj.ojMessage
 * @instance
 * @ojtranslatable
 * @since 5.0.0
 * @type {Object}
 *
 * @example <caption>Initialize the element, overriding some translated resources and leaving the
 * others intact:</caption>
 * &lt;!-- Using dot notation -->
 * &lt;oj-some-element translations.some-key='Some value' translations.some-other-key='Some other value'>&lt;/oj-some-element>
 *
 * &lt;!-- Using JSON notation -->
 * &lt;oj-some-element translations='{"someKey":"Some value", "someOtherKey":"Some other value"}'>&lt;/oj-some-element>
 *
 * @example <caption>Get or set the <code class="prettyprint">translations</code> property after
 * initialization:</caption>
 * // Get one
 * var value = myComponent.getProperty("translations.someKey");
 *
 * // Set one, leaving the others intact. Always use the setProperty API for
 * // subproperties rather than setting a subproperty directly.
 * myComponent.setProperty('translations.someKey', 'some value');
 *
 * // Get all
 * var values = myComponent.translations;
 *
 * // Set all.  Must list every resource key, as those not listed are lost.
 * myComponent.translations = {
 *     someKey: 'some value',
 *     someOtherKey: 'some other value'
 * };
 *
 */

// Events
/////////

/**
 * Triggered after the message is closed through user interaction or due to calling
 * <code class="prettyprint">close()</code> method.
 *
 * @expose
 * @event
 * @name close
 * @memberof oj.ojMessage
 * @instance
 * @since 5.0.0
 * @ojcancelable
 * @ojbubbles
 * @property {Object} message instance of type {@link oj.ojMessage#message} corresponding to the message that was closed
 */

/**
 * Triggered when the default animation is about to start for the open or close actions of the message .
 * The default animation can be cancelled by calling
 * <code class="prettyprint">event.preventDefault</code>.
 *
 * <p>In order to override open animation, the ojAnimateStart event listener needs to be
 * registered before the component is created. This means that <code>onOjAnimateStart</code>
 * cannot be used to register listener for 'open' action because it doesn't exist
 * until the component is upgraded or created at which time it is open. For the 'open' action, use 
 * the <code>addEventListener</code> method on the associated oj-message elements or
 * <code>on-oj-animate-start</code> attribute to register a listener instead.</p>
 *
 * @expose
 * @event
 * @memberof oj.ojMessage
 * @name animateStart
 * @instance
 * @since 5.0.0
 * @ojcancelable
 * @ojbubbles
 * @property {!Element} element target of animation
 * @property {"open"|"close"} action The action that is starting the animation.
 *            Suggested values are:
 *                    <ul>
 *                      <li>"open" - when a message is opened</li>
 *                      <li>"close" - when a message is closed</li>
 *                    </ul>
 * @property {!function():void} endCallback If the event listener calls
 *            event.preventDefault to cancel the default animation, it must call the
 *            endCallback function when it finishes its own animation handling and any
 *            custom animation has ended.
 *
 * @example <caption>Bind an event listener to the
 *          <code class="prettyprint">onOjAnimateStart</code> property to override the default
 *          "close" animation:</caption>
 * myMessage.onOjAnimateStart = function(event) {
 *   // verify that the component firing the event is a component of interest and action is close
 *   if (event.detail.action == "close") {
 *     event.preventDefault();
 *     oj.AnimationUtils.slideOut(event.detail.element, {"persist":  "all"}).then(event.detail.endCallback);
 *   }
 * };
 *
 * @example <caption>The default open and close animations are controlled via the theme
 *          (SCSS) :</caption>
 * // Applies for both 'inline' and 'overlay' messages
 * $messageGeneralOverlayOpenAnimation: ((effect: "zoomIn"), "fadeIn")  !default;
 * $messageGeneralOverlayCloseAnimation: ((effect: "zoomOut", persist: "all"), "fadeOut")  !default;
 * // Applies for 'notification' messages
 * $messageNotificationOverlayOpenAnimation: ((effect: "zoomIn"), "fadeIn")  !default;
 * $messageNotificationOverlayCloseAnimation: ((effect: "zoomOut", persist: "all"), "fadeOut")  !default;
 */

/**
 * Triggered when the default animation is about to end for the open or close actions of the message.
 * The default animation can be cancelled by calling
 * <code class="prettyprint">event.preventDefault</code>.
 *
 * @expose
 * @event
 * @memberof oj.ojMessage
 * @name animateEnd
 * @instance
 * @since 5.0.0
 * @ojcancelable
 * @ojbubbles
 *
 * @property {!Element} element target of animation
 * @property {"open"|"close"} action The action that is ending the animation.
 *            The number of actions can vary from component to component.
 *            Suggested values are:
 *                    <ul>
 *                      <li>"open" - when a message is opened</li>
 *                      <li>"close" - when a message is closed</li>
 *                    </ul>
 *
 * @example <caption>Bind an event listener to the
 *          <code class="prettyprint">onOjAnimateEnd</code> property to listen for the "close"
 *          ending animation:</caption>
 * myMessage.onOjAnimateEnd = function(event) {
 *   // verify that the component firing the event is a component of interest and action is close
 *   if (event.detail.action == "close") {}
 * };
 *
 * @example <caption>The default open and close animations are controlled via the theme
 *          (SCSS) :</caption>
 * // Applies for both 'inline' and 'overlay' messages
 * $messageGeneralOverlayOpenAnimation: (effect: "zoomIn", fade: true)  !default;
 * $messageGeneralOverlayCloseAnimation: (effect: "zoomOut", fade: true)  !default;
 * // Applies for 'notification' messages
 * $messageNotificationOverlayOpenAnimation: (effect: "zoomIn", fade: true)  !default;
 * $messageNotificationOverlayCloseAnimation: (effect: "zoomOut", fade: true)  !default;
 */

// Methods
//////////

/**
 * Closes the message.
 *
 * @expose
 * @function close
 * @memberof! oj.ojMessage
 * @instance
 * @since 5.0.0
 * @return {void}
 * @fires oj.ojMessage#ojAnimationStart
 * @fires oj.ojMessage#ojAnimationEnd
 * @fires oj.ojMessage#ojClose
 *
 *
 * @example <caption>Invoke the <code class="prettyprint">close</code> method:</caption>
 * myMessage.close();
 */

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 *
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @expose
 * @return {void}
 * @memberof! oj.ojMessage
 * @instance
 * @since 5.0.0
 *
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */

/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * @expose
 * @memberof! oj.ojMessage
 * @instance
 * @since 5.0.0
 *
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */

/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 * @expose
 * @memberof! oj.ojMessage
 * @instance
 * @since 5.0.0
 *
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */

// Type Defs
////////////

/**
 * @typedef {Object} oj.ojMessage.Message
 * @property {string} [icon] Defines the icon representing the message.
 * @property {string} [category] Defines category text of the message.
 * @property {"error"|"warning"|"confirmation"|"info"|"none"} [severity] Defines severity of the message.
 * @property {string} [timestamp] Defines timestamp of the message.
 * @property {string} [summary] Defines summary text of the message.
 * @property {string} [detail] Defines detail text of the message.
 * @property {number} [autoTimeout] Defines the time after which the message is to be closed automatically.
 * @property {"none"|"defaults"} [closeAffordance] Defines UI affordance provided to close the message.
 * @property {string} [sound] Defines the sound to be played when message is open.
 */


var _MESSAGE_METADATA =
{
  "name": "oj-message",
  "displayName": "oj-message",
  "version": "1.0.0",
  "jetVersion": "^5.0.0",
  "properties": {
    "message": {
      "type": "object",
      "writeback": false,
      "value": {
        "icon": "", "category": "",
        "severity": "none", "summary": "",
        "detail": "", "autoTimeout": -1,
        "closeAffordance": "defaults", "sound": "none"},
      "properties": {
        "icon": {
          "type": "string",
          "value": ""
        },
        "category": {
          "type": "string",
          "value": ""
        },
        "severity": {
          "type": "string",
          "enumValues" : ["error", "warning", "confirmation", "info", "none"],
          "value": "none"
        },
        "timestamp": {
          "type": "string",
          "value": ""
        },
        "summary": {
          "type": "string",
          "value": ""
        },
        "detail": {
          "type": "string",
          "value": ""
        },
        "autoTimeout": {
          "type": "number",
          "translatable": false,
          "value": -1
        },
        "closeAffordance": {
          "type": "string",
          "enumValues" : ["none", "defaults"],
          "value": "defaults"
        },
        "sound": {
          "type": "string",
          "value": "none"
        }
      }
    },

    "translations": {
      "type": "object",
      "writeback": false,
      "value": {"labelCloseIcon": ""},
      "properties": {
        "labelCloseIcon": {
          "type": "string",
          "translatable": true
        }
      }
    }
  },

  "methods": {
    "close": {
    }
  },

  "events": {
      "ojAnimateEnd" : {
          "bubbles" : true,
          "cancelable" : false,
          "detail" : {
            "action" : {"type" : "string",
                        "enumValues" : ["open", "close"]
                       },
            "element" : {"type" : "Element"}
          }
      },
      "ojAnimateStart" : {
          "bubbles" : true,
          "cancelable" : false,
          "detail" : {
            "action" : {"type" : "string",
                        "enumValues" : ["open", "close"]
                       },
            "element" : {"type" : "Element"},
            "endCallback" : {"type" : "function"}
          }
      },
      "ojClose" : {
          "bubbles" : true,
          "cancelable" : false,
          "detail" : {
            "message" : {"type" : "object"}
          }
      },
      "ojOpen" : {
          "bubbles" : true,
          "cancelable" : false,
          "detail" : {
            "message" : {"type" : "object"}
          }
      },
  },

  "slots": {
  }
};

var _MESSAGE_VIEW =
  '<div :id="[[containerId]]" class="oj-message-container" on-keydown="[[handleKeydown]]">' +
  '  <div class="oj-message-header">' +
  '    <div class="oj-message-leading-header" :title="[[computedCategory]]">' +
  '      <oj-bind-if test="[[computedIconStyle]]">' +
  '        <div class="oj-component-icon oj-message-status-icon oj-message-custom-icon" role="img" ' +
  '         :style.background="[[computedIconStyle]]">' +
  '        </div>' +
  '      </oj-bind-if>' +
  '      <oj-bind-if test="[[computedIconClass]]">' +
  '        <div role="img" ' +
  '          :class="[[computedIconClass]]"> ' +
  '        </div>' +
  '      </oj-bind-if>' +
  '      <div class="oj-message-category" tabindex="-1">' +
  '        <h1 :title="[[computedCategory]]">' +
  '          <oj-bind-text value="[[computedCategory]]"></oj-bind-text> ' +
  '        </h1>' +
  '      </div>' +
  '    </div>' +
  '    <div class="oj-message-trailing-header">' +
  '      <oj-bind-if test="[[computedTimestamp]]">' +
  '        <div class="oj-message-timestamp">' +
  '          <oj-bind-text value="[[computedTimestamp]]"></oj-bind-text> ' +
  '        </div>' +
  '      </oj-bind-if>' +
  '      <oj-bind-if test="[[hasCloseAffordance]]">' +
  '        <div class="oj-message-close">' +
  '          <oj-button display="icons" chroming="half" ' +
  '                     on-click="[[handleCloseIcon]]" :title="[[computedLabelCloseIcon]]">' +
  '            <span slot="startIcon" class="oj-fwk-icon oj-fwk-icon-cross"></span>' +
  '          </oj-button>' +
  '        </div>' +
  '      </oj-bind-if>' +
  '    </div>  ' +
  '  </div>' +
  '  <div class="oj-message-body">' +
  '    <div class="oj-message-summary">' +
  '      <oj-bind-text value="[[computedSummary]]"></oj-bind-text>' +
  '    </div>' +
  '    <div class="oj-message-detail">' +
  '      <oj-bind-text value="[[computedDetail]]"></oj-bind-text>' +
  '    </div>' +
  '  <div>' +
  '</div>';

function MessageViewModel(context)
{

  this._composite = context['element'];

  // anything used by the view bindings can't be obfuscated and is why it's defined in quotes.
  this._messagesContainerId = this["containerId"] = [context['unique'], "mc"].join("_");
  this["handleCloseIcon"] = this._handleCloseIcon.bind(this);
  this["close"] = this._closeMessage.bind(this);
  this["bindingsApplied"] = this._bindingsApplied.bind(this);
  this["disconnected"] = this._disconnected.bind(this);
  this["connected"] = this._connected.bind(this);
  this["propertyChanged"] = this._propertyChanged.bind(this);
  this["messageCreatedTime"] = new Date().toLocaleTimeString([],
                                      {"hour": "2-digit", "minute": "2-digit"});
  this["handleKeydown"] = this._handleKeydown.bind(this);

  // Store a reference to the properties for any later use
  this._properties = context['properties'];


  this._initDefaultTranslations();
  this._createObservables();
};

MessageViewModel.prototype._connected = function(element)
{

  // reattached when open (not pending open or close animation), check to see if we need to
  // reschedule the auto close timeout
  var mediator = this._operationMediator;
  if (mediator && mediator.getPendingOperation() === "none" && this._isMessageOpen())
  {
    this._scheduleAutoClose();
  }

};

MessageViewModel.prototype._disconnected = function(element)
{
  // Node disconnected and open (not pending open or close operations), clear the auto close timeout
  // since it might not be reattached.
  var mediator = this._operationMediator;
  if (mediator && mediator.getPendingOperation() === "none")
  {
    this._clearAutoClose();
  }
};

MessageViewModel.prototype._bindingsApplied = function(context)
{
  this._openMessage();
};

MessageViewModel.prototype._propertyChanged = function (detail)
{

  function subPropertyChanged(detail, topPropName, subPropName)
  {
    if ("external" === detail["updatedFrom"] && detail["property"] === topPropName)
    {
      var subproperty = detail["subproperty"];
      if (subproperty)
      {
        //set a subproperty via attribute or setProperty("message.xxxx", value)
        var path = [topPropName, subPropName].join(".");
        return (path === subproperty["path"]);
      }
      else
      {
        // set the entire message object
        var previousValue = detail["previousValue"][subPropName];
        var value = detail["value"][subPropName];
        return value !== previousValue;
      }
    }

    return false;
  }

  // workaround for 

  var message = this._properties["message"];

  if (subPropertyChanged(detail, "message", "autoTimeout"))
  {
    this._clearAutoClose();
    this._scheduleAutoClose();
  }

  if (subPropertyChanged(detail, "message", "closeAffordance"))
  {
    this["hasCloseAffordance"](this._computeCloseAffordance() === "defaults");
    this._unregisterSwipeHandler();
    this._registerSwipeHandler();
  }

  if (subPropertyChanged(detail, "message", "icon") ||
      subPropertyChanged(detail, "message", "severity") ||
      subPropertyChanged(detail, "message", "category"))
  {
    this["computedSeverity"](this._computeSeverity());
    this["computedCategory"](this._computeCategory());
    this["computedIconStyle"](this._computeIconStyle());
    this["computedIconClass"](this._computeIconClass());
  }

  if (subPropertyChanged(detail, "message", "timestamp"))
  {
    this["computedTimestamp"](this._computeTimestamp());
  }

  if (subPropertyChanged(detail, "message", "summary"))
  {
    this["computedSummary"](this._computeSummary());
  }

  if (subPropertyChanged(detail, "message", "detail"))
  {
    this["computedDetail"](this._computeDetail());
  }

  if (subPropertyChanged(detail, "message", "sound"))
  {
    message["sound"] = this._computeSound();
  }

  if (subPropertyChanged(detail, "translations", "labelCloseIcon"))
  {
    this["computedLabelCloseIcon"](this._computeLabelCloseIcon());
  }
};

MessageViewModel.prototype._registerSwipeHandler = function()
{
  if (oj.DomUtils.isTouchSupported() && this._computeCloseAffordance() === "defaults")
  {
    var messageContainerElement = $(document.getElementById(this._messagesContainerId));
    var options = {
                    'recognizers': [
                      [Hammer.Swipe, {'direction': Hammer['DIRECTION_HORIZONTAL']}]
                      ]
                  };

    var type = oj.DomUtils.getReadingDirection() === "rtl" ? "swipeleft" : "swiperight";
    this._hammerSwipe = messageContainerElement.ojHammer(options).on(type, function(event)
    {
      event.preventDefault();
      this._closeMessage();
    }.bind(this));
  }
};

MessageViewModel.prototype._unregisterSwipeHandler = function()
{
  if (oj.DomUtils.isTouchSupported() && this._hammerSwipe)
  {
    var type = oj.DomUtils.getReadingDirection() === "rtl" ? "swipeleft" : "swiperight";
    this._hammerSwipe.off(type);
    delete this._hammerSwipe;
  }
};

MessageViewModel.prototype._scheduleAutoClose = function()
{
  // Schedule auto-close if applicable
  if (this._computeAutoTimeout() > -1 )
  {
    this._autoCloseTimer = window.setTimeout(this._closeMessage.bind(this), this._computeAutoTimeout());
  }
}

MessageViewModel.prototype._clearAutoClose = function()
{
  if (!isNaN(this._autoCloseTimer))
  {
    window.clearTimeout(this._autoCloseTimer);
    delete this._autoCloseTimer;
  }
}

MessageViewModel.prototype._isMessageOpen = function ()
{
  var messageContainerElement = document.getElementById(this._messagesContainerId);
  return (messageContainerElement && $(messageContainerElement).is(":visible") ? true : false);
};

MessageViewModel.prototype._closeMessage = function(event)
{
  var mediator = this._operationMediator;
  // unregister the swipe handler
  this._unregisterSwipeHandler();

  // if we are pending the open operation, the close will be invoked after the open is done
  if (mediator.isOperationPending("close", this._closeMessage.bind(this, event)))
    return;
  else
    mediator.destroy();

  this._clearAutoClose();
  this._operationMediator = new OperationMediator(this._composite, "close");

  var messageContainerElement = document.getElementById(this._messagesContainerId);
  var action = "close";
  var options = this._getAnimateOptionDefaults(action);
  oj.AnimationUtils.startAnimation(messageContainerElement, action, options,
                                   this._composite).then( function ()
  {
    // We will just hide the message container, but not discard it, the ojClose listener
    // could discard it if needed.
    $(messageContainerElement).hide();
    oj.Components.subtreeHidden(messageContainerElement);

    var eventParams = {
        "bubbles" : true,
        "cancelable" : false,
        "detail" : {
            "message" : this._properties['message']
        }
    };

    var closeEvent = new CustomEvent("ojClose", eventParams);
    if (event)
      Object.defineProperty(closeEvent, "_originalEvent", {"value": event, "writable": false});

    this._composite.dispatchEvent(closeEvent);

  }.bind(this));
};

MessageViewModel.prototype._computeSeverity = function()
{
  var message = this._properties['message'];

  if (oj.StringUtils.isEmptyOrUndefined(message["severity"]))
    message["severity"] = MessageViewModel._getMessageDefault("severity");

  return message['severity'];
};

MessageViewModel.prototype._computeTimestamp = function()
{
  var message = this._properties["message"];

  if (oj.StringUtils.isEmptyOrUndefined(message["timestamp"]))
  {
    // ensure we adhere to default value as specified in API
    message["timestamp"] = MessageViewModel._getMessageDefault("timestamp");
    return null;
  }

  try
  {
    // get the default converter, use default format as in UX specs
    var converter = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).
                      createConverter({pattern: "MM/dd/yy, hh:mm a"});

    var formattedTimestamp = converter.format(message["timestamp"]);

    // not so-clean implementation about how we avoid displaying date part when it is same date,
    //  should be solved better when we provide API to set custom converter in near future
    if (this._isDateToday(message["timestamp"]))
    {
      formattedTimestamp = formattedTimestamp.split(", ")[1];
    }

    return formattedTimestamp;
  }
  // expect oj.ConverterError if supplied value is not valid
  catch(e)
  {
    oj.Logger.info(["JET oj-message id='", MessageViewModel._toSelector(this._composite),
                    "': failed to parse or format the supplied value to message.timestamp='",
                    message['timestamp'], "'."].join(""));
    return null;
  }
};

MessageViewModel.prototype._isDateToday = function(isoDate)
{
  var todayDate = new Date();
  var suppliedDate = oj.IntlConverterUtils.isoToLocalDate(isoDate);

  return todayDate.getUTCFullYear() === suppliedDate.getUTCFullYear() &&
         todayDate.getUTCMonth() === suppliedDate.getUTCMonth() &&
         todayDate.getUTCDate() === suppliedDate.getUTCDate();
};

MessageViewModel.prototype._computeCategory = function()
{
  var message = this._properties['message'];

  var mappedCategory = oj.Translations.getComponentTranslations("oj-ojMessage")["categories"];
  message["category"] = mappedCategory[this._computeSeverity()];

  if (oj.StringUtils.isEmptyOrUndefined(message["category"]))
    message["category"] = MessageViewModel._getMessageDefault("category");

  return message['category'];
};

MessageViewModel.prototype._computeAutoTimeout = function()
{
  var message = this._properties['message'];
  if (isNaN(message['autoTimeout']))
    message['autoTimeout'] = MessageViewModel._getMessageDefault("autoTimeout");
  else if (message['autoTimeout'] === 0)
    message["autoTimeout"] = this._getThemedAutoTimeoutDefault();

  return message['autoTimeout'];
};

MessageViewModel.prototype._computeIconStyle = function()
{
  var message = this._properties['message'];

  if (oj.StringUtils.isEmptyOrUndefined(message["icon"]))
    message["icon"] = MessageViewModel._getMessageDefault("icon");

  if (oj.StringUtils.isEmptyOrUndefined(message["icon"]))
    return undefined;
  else
    return ["url('", message["icon"], "') no-repeat"].join("");
};

MessageViewModel.prototype._computeIconClass = function()
{
  var message = this._properties['message'];

  if (oj.StringUtils.isEmptyOrUndefined(message["icon"]))
    message["icon"] = MessageViewModel._getMessageDefault("icon");

  var severity = this._computeSeverity();
  if (!oj.StringUtils.isEmptyOrUndefined(message["icon"]) || severity === "none")
    return undefined;
  else
  {
    var selectors = ["oj-component-icon", "oj-message-status-icon"];
    selectors.push(["oj", "message", severity, "icon"].join("-"));
    return selectors.join(" ");
  }
};

MessageViewModel.prototype._computeCloseAffordance = function()
{
  var message = this._properties['message'];

  if (oj.StringUtils.isEmptyOrUndefined(message["closeAffordance"]))
    message["closeAffordance"] = MessageViewModel._getMessageDefault('closeAffordance');

  return message['closeAffordance'];
};

MessageViewModel.prototype._computeSound = function()
{
  var message = this._properties['message'];

  if (oj.StringUtils.isEmptyOrUndefined(message["sound"]))
    message["sound"] = MessageViewModel._getMessageDefault('sound');

  return message['sound'];
};

MessageViewModel.prototype._computeLabelCloseIcon = function()
{
  var translations = this._properties['translations'];
  if (oj.StringUtils.isEmptyOrUndefined(translations['labelCloseIcon']))
  {
    var label = oj.Translations.getTranslatedString("oj-ojMessage.labelCloseIcon");
    translations['labelCloseIcon'] = label;
  }

  return translations['labelCloseIcon'];
};

MessageViewModel.prototype._computeSummary = function()
{
  var message = this._properties['message'];

  if (oj.StringUtils.isEmptyOrUndefined(message["summary"]))
    message["summary"] = MessageViewModel._getMessageDefault("summary");

  return message["summary"];
};

MessageViewModel.prototype._computeDetail = function()
{
  var message = this._properties['message'];

  if (oj.StringUtils.isEmptyOrUndefined(message["detail"]))
    message["detail"] = MessageViewModel._getMessageDefault("detail");

  return message["detail"];
};

MessageViewModel.prototype._handleCloseIcon = function(event)
{
  this._closeMessage(event);
}

MessageViewModel.prototype._openMessage = function()
{
  this._operationMediator = new OperationMediator(this._composite, "open");

  // Invoke next micro-tick to allow the parent to apply bindings before opening the message.
  // When the parent is an oj-messages, it will need to subscribe to message events:
  // ojBeforeOpen, ojOpen and ojAnimateStart. These events are  queued by the message open
  // besides resolving :id bindings defining its content.
  Promise.resolve().then(function ()
  {
    // notify the oj-messages to show the messages container so initial animation
    // will be seen.
    var eventParams = {
      "bubbles" : true,
      "cancelable" : false,
      "detail" : {
        "message" : this._properties['message']
      }
    };

    this._composite.dispatchEvent(new CustomEvent("ojBeforeOpen", eventParams));

    var messageContainerElement = document.getElementById(this._messagesContainerId);
    var action = "open";
    var options = this._getAnimateOptionDefaults(action);
    oj.AnimationUtils.startAnimation(messageContainerElement, action, options,
                                   this._composite).then( function ()
    {
      eventParams = {
        "bubbles" : true,
        "cancelable" : false,
        "detail" : {
          "message" : this._properties['message']
        }
      };

      if (this._computeSound() !== "none")
      {
        // Initialize the AudioContext if the user agent supports Web Audio API
        this._initAudioContext();
        this._playSound();
      }
      // register the swipe handler for touch devices
      this._registerSwipeHandler();
      this._scheduleAutoClose();
      this._composite.dispatchEvent(new CustomEvent("ojOpen", eventParams));
    }.bind(this));
  }.bind(this));
};

MessageViewModel.prototype._handleKeydown = function(event)
{
  // if the event has been prevented or closeAffordance property is not "defaults", no-opt.
  if (event.defaultPrevented || this._computeCloseAffordance() !== "defaults")
    return;

  if (event.keyCode === $.ui.keyCode.ESCAPE)
  {
    event.preventDefault()
    this._closeMessage(event);
  }
};

MessageViewModel.prototype._playSound = function()
{
  var sound = this._computeSound();

  // Custom URL was specified for the sound, using <audio> element is simple and best for this case.
  if (sound !== "defaults")
  {
    var audio = document.createElement("AUDIO");
    audio.src = sound;

    // Handle any error due to possible invalid URL.
    audio.addEventListener("error", function(e)
    {
      oj.Logger.info(["JET oj-message id='", MessageViewModel._toSelector(this._composite),
                      "': failed to load or play media file for URL in property message.sound='",
                      sound, "'."].join(""));
    }.bind(this));

    audio.play();
  }
  // User agent does not support web audio API
  else if(window["audioContext"] === undefined)
  {
    oj.Logger.info(["JET oj-message id='", MessageViewModel._toSelector(this._composite),
                   "': failed to load or play default sound for message because user agent does  \n\
                       not support Web Audio API'"].join(""));
  }
  else
  {
    // Default gain value will be 100% of speaker volume which is fine
    var gainNode = window["audioContext"]["createGain"]();

    // Default destination is system speakers
    gainNode["connect"](window["audioContext"]["destination"]);

    // Use default oscillator node type 'sine' and frequency '440 Hz', suffices for simple beep
    var oscillatorNode = window["audioContext"]["createOscillator"]();
    oscillatorNode["connect"](gainNode);
    oscillatorNode.start();

    // Play for 10 ms
    oscillatorNode.stop(window["audioContext"]["currentTime"] + 0.01);
  }
};

MessageViewModel.prototype._initAudioContext = function()
{
  if (window["audioContext"] === undefined)
  {
    try
    {
      // Create a page level AudioContext that all messages would use in this page. Browsers have
      //  limit on number of instances (usually 6), and AudioContext is designed to be shared per
      //  page. Webkit has a different constructor.
      window["audioContext"] = new (window["AudioContext"] || window["webkitAudioContext"])();
    }
    catch(e)
    {
      // no op
    }
  }
};

MessageViewModel.prototype._createObservables = function()
{
  // Computed observables
  // anything used by the view bindings can't be obfuscated and is why it's defined in quotes.
  //this["computedSeverity"] = ko.pureComputed(this._computeSeverity.bind(this), this);
  //this["computedCategory"] = ko.pureComputed(this._computeCategory.bind(this), this);
  //this["computedLabelCloseIcon"] = ko.pureComputed(this._computeLabelCloseIcon.bind(this), this);
  //this["computedCloseAffordance"] = ko.pureComputed(this._computeCloseAffordance.bind(this), this);
  // workaround for 

  this["hasCloseAffordance"] = ko.observable(this._computeCloseAffordance() === "defaults");
  this["computedIconStyle"] = ko.observable(this._computeIconStyle());
  this["computedIconClass"] = ko.observable(this._computeIconClass());
  this["computedSeverity"] = ko.observable(this._computeSeverity());
  this["computedCategory"] = ko.observable(this._computeCategory());
  this["computedTimestamp"] = ko.observable(this._computeTimestamp());
  this["computedLabelCloseIcon"] = ko.observable(this._computeLabelCloseIcon());
  this["computedSummary"] = ko.observable(this._computeSummary());
  this["computedDetail"] = ko.observable(this._computeDetail());

};

MessageViewModel.prototype._initDefaultTranslations = function()
{
  var properties = this._properties;

  if (oj.StringUtils.isEmptyOrUndefined(properties['translations']['labelCloseIcon']))
  {
    properties['translations']['labelCloseIcon'] = this._computeLabelCloseIcon();
  }
};

MessageViewModel._getMessageDefault = function (propName)
{
  return MessageViewModel._DEFAULTS["message"][propName];
}

MessageViewModel._DEFAULTS = {
  "autoTimeout": 4000,
  "animation": {
    "open": {"effect": "fadeIn", "duration":"300ms"},
    "close": {"effect": "fadeOut", "duration": "300ms"},
  },
  "message": {
    "icon": "",
    "category": "",
    "severity": "none",
    "timestamp": "",
    "summary": "",
    "detail": "",
    "autoTimeout": -1,
    "closeAffordance": "defaults",
    "sound": "none"
  }
};

MessageViewModel.prototype._getThemedAutoTimeoutDefault = function ()
{
  var themedDefaults = oj.ThemeUtils.parseJSONFromFontFamily('oj-message-option-defaults');
  if (themedDefaults && themedDefaults["autoTimeout"])
    return themedDefaults["autoTimeout"];
  else
    return MessageViewModel._DEFAULTS["autoTimeout"];
};

MessageViewModel.prototype._getAnimateOptionDefaults = function (action)
{
  return MessageViewModel._DEFAULTS["animation"][action];
};

MessageViewModel._toSelector = function (node)
  {
    var selector = "";
    if (node)
    {
      if (node.id && node.id.length > 0)
        selector = "#" + node.id;
      else
      {
        selector = node.nodeName;
        var clazz = node.getAttribute("class");
        if (clazz)
          selector += "." + clazz.split(" ").join(".");

        if (node.parentNode)
          return MessageViewModel._toSelector(node.parentNode) + " > " + selector;
      }
    }

    return selector;
  }

/**
 * Coordinate communications between an event being fulfilled and one or more promises
 * being resolved.  The window of time between the instance creation and the associated event
 * triggered is guarded by the {@link oj.BusyContext}.
 * @private
 * @constructor
 * @ignore
 * @param {Element} element to subscribe on the event type triggered on completion of the operation
 * @param {string} operation that completion will resolve one or more promises
 * @ojtsignore
 */
function OperationMediator (element, operation)
{
  this._element = element;
  this._operation = operation;

  this._init();
};

/**
 * Registers an event handler on the element associated with the target operation.
 * The event handler will resolve one or more pending promises.  The convention
 * is the operation will raise a "oj" + operation event upon completion. The
 * event hander is one and done - unregistered after first delivery.
 * @instance
 * @private
 */
OperationMediator.prototype._init = function ()
{
  this._resolvedQueue = [];
  this._callback = this._eventHandler.bind(this);

  var operation = this._operation;
  var tokens = ["oj"];
  tokens.push(operation.charAt(0).toUpperCase());
  tokens.push(operation.slice(1));

  var eventType = this._eventType = tokens.join("");
  this._element.addEventListener(eventType, this._callback);

  // Add a busy state for the pending operation.  The busy state resolver will
  // be invoked when the resolved queue is delivered (operation completes).
  var busyContext = oj.Context.getContext(this._element).getBusyContext();
  var options = {"description" : this._getBusyStateDescription.bind(this, this._element,
    this._operation)};
  var resolve = busyContext.addBusyState(options);
  this.addPromiseExecutor(function (callback)
  {
    // Resolve busyness next-tick.  The ojOpen/ojClose events bubble
    // up to oj-messages All event processing will be at the parent level.
    // Keep the busy state until the parent has had a chance to act on
    // the open/close event.
    window.setImmediate(function ()
    {
      callback();
    });

  }.bind(this, resolve));

};

/**
 * @private
 * @param {Element} element to subscribe on the event type triggered on completion of the operation
 * @param {string} operation that completion will resolve one or more promises
 * @returns {string} description of the busy state animation operation.
 */
OperationMediator.prototype._getBusyStateDescription = function (element, operation)
{
  return [element.nodeName, " identified by '", MessageViewModel._toSelector(element),
    "' is busy animating on the '", operation, "' operation."].join("");
};

/**
 * Resolves the pending promises.
 *
 * @private
 * @param {string=} operation override sent to the resolverdQueue
 */
OperationMediator.prototype._deliverResolved = function (operation)
{
  // Critical section - the registered resolve queue is disconnect from the
  // instance state so that a race condition will not occur - resolve promoise
  // adding new operations to the same queue.

  var resolvedQueue = this._resolvedQueue;
  this._resolvedQueue = [];

  operation = !operation ? this._operation : operation;
  this._operation = null;

  for (var i = 0; i < resolvedQueue.length; i++)
  {
    resolvedQueue[i](operation);
  }

};

/**
 * Force delivery of unresolved promises.
 */
OperationMediator.prototype.destroy = function ()
{
  // If the promise is swapped (component is destroyed)
  // before the event is fired, resolve with a "none" operation.
  this._deliverResolved("none");

  this._callback = null;
  this._element = null;
  this._operation = null;
  this._eventType = null;
};

/**
 * Event handler associated with completion of the target operation.
 * @private
 * @param {Event} event
 */
OperationMediator.prototype._eventHandler = function (event)
{
  if (event.target === this._element)
  {
    this._element.removeEventListener(event.type, this._callback);
    this._deliverResolved();
    this._callback = null;
  }
};

/**
 * @return {string} Returns the pending operation
 */
OperationMediator.prototype.getPendingOperation = function ()
{
  return this._operation ? this._operation : "none";
};

/**
 * A function that will be passed to other functions via the arguments resolve and reject.
 * The resolve function will be invoked when the event associated with completion of the
 * target operation is delivered to the target element.
 *
 * @public
 * @param {Function} resolve resultant function that will resovle a promise executor
 * @param {Function=} reject (not interested in the reject)
 */
OperationMediator.prototype.addPromiseExecutor = function (resolve, reject)
{
  this._resolvedQueue.push(resolve);
};

/**
 * Checks to see if there is a pending "open" or "close" operation.  If pending and it
 * is the same as the requested operation, the request silently fails.  If the current
 * operation is the inverse operation, we queue the current operation after the pending
 * operation is resolved.
 *
 * @param {string} operation currently requested
 * @param {Function} callback that should be invoked when the operation is the
 *                 inverse of the pending operation
 * @returns {boolean} <code>true</code> if a "close" or "open" operation is pending completion.
 */
OperationMediator.prototype.isOperationPending = function (operation, callback)
{
  // if destroyed return false
  if (!this._element)
    return false;

  var isPending = false;
  var pendingOperation = this.getPendingOperation();
  if (operation === pendingOperation)
  {
    // Same request is already pending. Silently fail.
    oj.Logger.info(["An oj-message id='", MessageViewModel._toSelector(this._element),
           "' invoked a '", operation,
           "' operation while pending animation of the same type of operation. ",
           "The second request will be ignored."].join(""));
    isPending = true;
  }
  else if ("none" !== pendingOperation)
  {
    oj.Logger.info(["An oj-message id='", MessageViewModel._toSelector(this._element),
                   "' invoked a '", operation,
                   "' operation while pending animation of a '", pendingOperation,
                   "' operation. The second request will be invoked after the pending ",
                   "operation completes."].join(""));

    // Queue the operation after the pending operation has completed.
    this.addPromiseExecutor(callback);

    isPending = true;
  }
  return isPending;
};

oj.Composite.register('oj-message',
{
  "view": _MESSAGE_VIEW,
  "viewModel": MessageViewModel,
  "metadata": _MESSAGE_METADATA
});

/*jslint browser: true*/
/*
 ** Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
 */

/**
 * @ojcomponent oj.ojMessages
 * @since 5.0.0
 * @ojdisplayname Messages
 * @ojshortdesc Manages layout and display of messages.
 * @ojstatus preview
 * @ojsignature {target: "Type", value:"class ojMessages extends JetElement<ojMessagesSettableProperties>"}
 * @ojtsimport ojmessage
 *
 * @classdesc
 * <h3 id="messageOverview-section">
 *   JET Messages
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#messageOverview-section"></a>
 * </h3>
 * <p>Description:
 * <p>Manages the layout and display of one or more messages. The only supported component child
 * element of <code class="prettyprint">oj-messages</code> is
 * <code class="prettyprint">oj-message</code>.
 * 
 * <h4 id="messages-layout-section">Layouts
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#messages-layout-section"></a></h4>
 * <p>Messages can be shown in these layouts
 * <ul>
 * <li>Inlined to the page content</li>
 * <li>Overlayed on the page content, positioned to the top of the page by default</li>
 * <li>As notifications, positioned to the top-end corner of the page by default</li>
 * </ul>
 * Inlined messages will push the page contents when a new message is shown and could scroll out of 
 * the view when user scrolls the page. Overlayed or notification style messages are usually 
 * positioned relative to the window, will assume a fixed position and will remain 
 * visible when the user scrolls the page. Overlayed or notification messages can be positioned
 * relative to another element in the page, in which case the messages follow that element upon
 * page scroll. Inline and overlay layouts are suitable for messages pertaining to the page or 
 * region of the page, or relating to the task the user performed on the page. Notification layout is
 * suitable for messages that arrive asynchronously to the application or to communicate alerts to 
 * users. Examples below shows the three layouts.
 * 
 * <pre class="prettyprint">
 * <code>
 * &lt;!-- Inline: Do not specify the 'position', set 'display' to 'general' -->
 * &lt;oj-messages id="inlineMessages" messages="[[serviceRequestMessages]]" display="general">
 * &lt;/oj-messages>
 * 
 * &lt;!-- Overlay: Set 'position' to an empty or a structured object, set 'display' to 'general' -->
 * &lt;oj-messages id="overlayMessages" messages="[[serviceRequestMessages]]" position="{}" display="general">
 * &lt;/oj-messages>
 * 
 * &lt;!-- Notification: Set 'position' to an empty or a structured object, set 'display' to 'notification' -->
 * &lt;oj-messages id="notificationMessages" messages="[[emailMessages]]" position="{}" display="notification">
 * &lt;/oj-messages>
 * </code>
 * </pre>
 * 
 * <h4 id="messages-syntax-section">Syntaxes
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#messages-syntax-section"></a></h4>
 * <p>Messages can be defined as below followed by examples
 * <ul>
 * <li>Include an oj-message child for every message to be shown. This is suitable for case when the
 * messages are static and in small numbers</li>
 * <li>Include oj-message child, wrap it inside of a for-each binding. This is suitable for case
 * where one or more message sources need to be displayed in a single messages element</li>
 * <li>Use the simple syntax to directly bind a collection to the 'messages' attribute of oj-messages.
 * This is a convenient syntax and suitable when a single message source is used</li>
 * </ul>
 * 
 * <pre class="prettyprint">
 * <code>
 * &lt;!-- Including oj-message children -->
 * &lt;oj-messages id="inlineMessages">
 *   &lt;oj-message message='{"summary": "Some summary", "detail": "Some detail", "autoTimeout": 5000}'>&lt;/oj-message>
 *   &lt;oj-message message="[[surveyInstructions]]">&lt;/oj-message>
 *   &lt;oj-message message="[[surveySubmitConfirmation]]">&lt;/oj-message>
 * &lt;/oj-messages>
 *
 * &lt;!-- For-each bound oj-message children -->
 * &lt;oj-messages id="pageOverlayMessages" position="{}" display="notification">
 *   &lt;oj-bind-for-each data="[[serviceRequestStatusUpdateMessages]]">
 *     &lt;template>
 *       &lt;oj-message message="[[$current.data]]">&lt;/oj-message>
 *     &lt;/template>
 *   &lt;/oj-bind-for-each>
 *   &lt;oj-bind-for-each data="[[criticalIncidentMessages]]">
 *     &lt;template>
 *       &lt;oj-message message="[[$current.data]]">&lt;/oj-message>
 *     &lt;/template>
 *   &lt;/oj-bind-for-each>
 * &lt;/oj-messages>
 *
 * &lt;!-- Collection bound simple syntax, no need to specify oj-message children -->
 * &lt;oj-messages id="notificationMessages" messages="[[emailMessages]]" position="{}" display="notification">
 * &lt;/oj-messages>
 * </code>
 * </pre>
 *
 * <h3 id="reparenting-section">
 *   Reparenting
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#reparenting-section"></a>
 * </h3>
 *
 *  <p id="reparenting-strategy">
 *     When the messages region is disclosed as a popup (position property defined), it will be
 *     reparented in the document and reparented back when all contained messages are closed.
 *     The location of the messages within the document will be in context of where it is defined.
 *     This is different than other types of popups where their reparenting location depends on
 *     the context they are used.  The stacking context of the messages region will be defined
 *     by the "oj-messages-layer" style class. The messages region that has active focus will be
 *     assigned a greater z-index value. This is applied to the messages's layer by way of
 *     the "oj-focus-within" pseudo selector applied with "oj-messages-layer" selector.
 *     The page author has control over z-index weights by way of the "oj-messages-layer" selector.
 *  </p>
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 * <p> The following CSS classes can be applied by the page author as needed.</p>
 *
 * {@ojinclude "name":"stylingDoc"}
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan = "3">Focus within Messages</td>
 *       <td><kbd>Tab</kbd> or <kbd>Shift + Tab</kbd></td>
 *       <td>Navigate the content of the messages region.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>F6</kbd></td>
 *       <td>Moves focus back to the last focused element outside the messages region.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Esc</kbd></td>
 *       <td>Moves focus back to the last focused element outside the messages region.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan = "1">Focus outside Messages</td>
 *       <td><kbd>F6</kbd></td>
 *       <td>Move focus to the first message within the more recently disclosed messages region.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojMessages
 */

/**
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Class</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-messages-inline-remove-bottom-border</td>
 *       <td>Inline messages will include a bottom border so that the messages section is demarcated
 *            from the contents below it. If this border is not desirable for certain page layouts, 
 *            it can be removed by setting this marker class on oj-messages.
  *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
 * @memberof oj.ojMessages
 */

// Attributes
/////////////

/**
 *<p> Specifies the array of structured message data used to display the individual messages.
 * Instead of providing multiple oj-message as children, this property can be used to conveniently
 * specify the required data as a single collection. Individual message will be automatically
 * created based on this data. See {@link oj.ojMessage.Message} for message values.</p>
 *<p> More information about the structured 'Message' data can be found in documentation for 
 *'message' attribute of <code class="prettyprint">oj-message</code> element.</p>
 *
 * @example <caption>Initialize component with <code class="prettyprint">messages</code> attribute:</caption>
 * &lt;!-- emailNotifications is an array of messages, with each entry being of 'Message' type -->
 * &lt;oj-messages messages="[[emailNotifications]]">&lt;/oj-messages>
 *
 * @example <caption>Get or set the <code class="prettyprint">messages</code> property after initialization:</caption>
 *
 * // getter
 * var messages = myMessages.messages;
 *
 * // setter
 * myMessages.messages = [{"severity": "error", "summary": "Some summary 1", "detail": "Some detail 1"},
 *                        {"severity": "warning", "summary": "Some summary 2", "detail": "Some detail 2"}];
 *
 * @expose
 * @type {null | Array.<Object>}
 * @name messages
 * @default null
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessages
 * @ojsignature { target: "Type",
 *                value: "Array<oj.ojMessage.Message>|null",
 *                jsdocOverride: true}
**/

/**
 * <p>Specifies the display purpose of messages. The valid values for display are
 * <code class="prettyprint">general</code> and <code class="prettyprint">notification</code>.
 * General messages are commonly rendered at the page level or section level, relating to the task
 * the user performed. Notification messages are typically used for reporting asynchronous events, 
 * or to communicate some background activity.</p>
 *
 * The presentation of the message is inline by default.  However, when a
 * {@link oj.ojMessages#position} property is provided, the presentation will be an overlay.
 * The alignment of the overlay will default based on the <code>display</code> property.  The
 * defaults are defined by the theme.
 *
 * @example <caption>Initialize component with <code class="prettyprint">display</code> attribute:</caption>
 * &lt;oj-messages display="notification" position="{}">&lt;/oj-messages>
 *
 * @example <caption>Get or set the <code class="prettyprint">display</code> property after initialization:</caption>
 * // getter
 * var display = myMessages.display;
 *
 * // setter
 * myMessages.display = "notification";
 *
 * @expose
 * @type {string}
 * @name display
 * @default "general"
 * @ojvalue {string} "general" messages pertaining to the page or region of the application
 * @ojvalue {string} "notification" often used for communicating alerts arriving asynchronously
 * @instance
 * @since 5.0.0
 * @access public
 * @memberof! oj.ojMessages
 **/

/**
 * <p>The position property defines the presentation style.  The default presentation is inline,
 * defined by a <code>null</code> position property value.  When a value is provide for the
 * property, the presentation style will be an overaly "popup".  The aligment of the overaly
 * is defined by the position sub-properties.</p>
 *
 * Default position sub-properites are extended by the provided value.  Defaults vary
 * based on the <code>display</code> property and provided by theme. The position
 * property is used to establish the location where the messages popup overlay will appear
 * relative to another element.</p>
 *
 * <p>The "my" and "at" properties defines aligment points relative to the popup and other
 * element.  The "my" property represents the popups alignment where the "at" property
 * represents the other element that can be identified by "of" or defauts to the launcher
 * when the popup opens.  The values of these properties describe horizontal and
 * vertical alignments.</p>
 *
 * @example <caption>Initialize the popup with <code class="prettyprint">position</code>
 *           attribute specified:</caption>
 * &lt;oj-messages position.my.horizontal="left"
 *              position.my.vertical="top"
 *              position.at.horizontal="right"
 *              position.at.vertical="top">
 * &lt;/oj-messages>;
 *
 * @example <caption>Get or set the <code class="prettyprint">position</code> property,
 *          after initialization:</caption>
 * // getter
 * var position = myMessages.position;
 *
 * // setter
 * myMessages.position =
 *    {"my": {"horizontal": "start", "vertical": "bottom"},
 *     "at": {"horizontal": "end", "vertical": "top" },
 *     "offset": {"x": 0, "y":5}};
 *
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @type {Object|null}
 * @default null
 * @name position
 * @ojsignature { target: "Type",
 *                value: "oj.ojMessages.Position|null",
 *                jsdocOverride: true}
 */

/**
 * Defines which edge on the messages overlay to align with the target ("of") element.
 *
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @default {"horizontal" : "center", "vertical" : "top"}
 * @alias position.my
 * @name position.my
 * @type {Object}
 */

/**
 * Defines the horizontal alignment of the messages overlay.
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.my.horizontal
 * @name position.my.horizontal
 * @type {string}
 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
 * @ojvalue {string} "left"
 * @ojvalue {string} "center"
 * @ojvalue {string} "right"
 */

/**
 * Defines the vertical alignment of the messages overlay.
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.my.vertical
 * @name position.my.vertical
 * @type {string}
 * @ojvalue {string} "top"
 * @ojvalue {string} "center"
 * @ojvalue {string} "bottom"
 */

/**
 * Defines a point offset in pixels from the ("my") alignment.
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.offset
 * @name position.offset
 * @type {Object}
 */

/**
 * Horizontal aligment offset.
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.offset.x
 * @name position.offset.x
 * @type {number}
 */

/**
 * Vertical alignment offset.
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.offset.y
 * @name position.offset.y
 * @type {number}
 */

/**
 * Defines which position on the target element ("of") to align the positioned element
 * against.
 *
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.at
 * @name position.at
 * @type {Object}
 */

/**
 * Defines the horizontal alignment of what the messges overlay is aligned to.
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.at.horizontal
 * @name position.at.horizontal
 * @type {string}
 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
 * @ojvalue {string} "left"
 * @ojvalue {string} "center"
 * @ojvalue {string} "right"
 */

/**
 * Defines the vertical alignment of what the messages overlay is aligned to.
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.at.vertical
 * @name position.at.vertical
 * @type {string}
 * @ojvalue {string} "top"
 * @ojvalue {string} "center"
 * @ojvalue {string} "bottom"
 */

/**
 * Which element to position the messages overlay against.
 *
 * If the value is a string, it should be a selector or the literal string value
 * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
 * is used, the values are relative to the whole document.  Page horizontal and vertical
 * scroll offsets need to be factored into this point - see UIEvent
 * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
 * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
 *
 * @example <caption>Finding the point for an svg element:</caption>
 * var rect = svgDom.getBoundingClientRect();
 * var position = {of:{x:rect.left + window.pageXOffset, y:rect.top + window.pageYOffset}};
 *
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.of
 * @name position.of
 * @type {string}
 */

/**
 * Rule for alternate alignment.
 *
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @alias position.collision
 * @name position.collision
 * @type {string}
 * @ojvalue {string} "flip" the element to the opposite side of the target and the
 *  collision detection is run again to see if it will fit. Whichever side
 *  allows more of the element to be visible will be used.
 * @ojvalue {string} "fit" shift the element away from the edge of the window.
 * @ojvalue {string} "flipfit" first applies the flip logic, placing the element
 *  on whichever side allows more of the element to be visible. Then the fit logic
 *  is applied to ensure as much of the element is visible as possible.
 * @ojvalue {string} "none" no collision detection.
 */

/**
 * <p>A collection of translated resources from the translation bundle, or <code class="prettyprint">null</code> if this
 * component has no resources.  Resources may be accessed and overridden individually or collectively, as seen in the examples.
 *
 * <p>If this component has translations, their documentation immediately follows this doc entry.
 *
 *
 * @example <caption>Initialize the component, overriding some translated resources and leaving the others intact:</caption>
 * &lt;!-- Using dot notation -->
 * &lt;oj-some-element translations.some-key='some value' translations.some-other-key='some other value'>&lt;/oj-some-element>
 *
 * &lt;!-- Using JSON notation -->
 * &lt;oj-some-element translations='{"someKey":"some value", "someOtherKey":"some other value"}'>&lt;/oj-some-element>
 *
 * @example <caption>Get or set the <code class="prettyprint">translations</code> property after initialization:</caption>
 * // Get one
 * var value = myComponent.translations.someKey;
 *
 * // Set one, leaving the others intact. Always use the setProperty API for
 * // subproperties rather than setting a subproperty directly.
 * myComponent.setProperty('translations.someKey', 'some value');
 *
 * // Get all
 * var values = myComponent.translations;
 *
 * // Set all.  Must list every resource key, as those not listed are lost.
 * myComponent.setProperty("translations.someKey" 'some value');
 *
 * @expose
 * @type {Object}
 * @name translations
 * @memberof! oj.ojMessages
 * @ojtranslatable
 * @member
 * @instance
 * @since 5.0.0
 **/


// Methods
//////////

/**
 * Closes the specified message regardless of the {@link oj.ojMessage#message.autoTimeout} or
 * {@link oj.ojMessage#message.closeAffordance} properties. The <code>message</code> argument
 * is a required argument, must be of type {@link oj.ojMessage#message}, and must be the same object 
 * instance that was used to create and show the oj-message.
 *
 * <p>Closing a message changes the visibility to hidden. If the message is defined by an instance
 * in the {@link oj.ojMessages#messages} collection, the close operation will not remove the item
 * from the backing model. Application logic needs to listen for the 
 * {@link oj.ojMessage#event:close} event bubbling up from the underlying oj-message child to remove 
 * from the backing collection.</p>
 *
 * @expose
 * @function close
 * @memberof! oj.ojMessages
 * @instance
 * @since 5.0.0
 * @param {Object} associated message instance of type {@link oj.ojMessage#message}
 * @return {void}
 * @ojsignature { target: "Type",
 *                value: "oj.ojMessage.Message",
 *                for: "message",
 *                jsdocOverride: true}
 *
 * @example <caption>Invoke the <code class="prettyprint">close</code> method:</caption>
 * myMessages.close(myMessage.message);
 */

/**
 * Closes child messages matching the <code>closeFilter</code> callback criteria.
 * The <code>closeFilter</code> callback is an optional argument. If not specified, all child
 * messages will be closed. If filter is specified, object of type {@link oj.ojMessage#message} 
 * corresponding to each child will be passed to the <code>closeFilter</code> function. A return 
 * value of <code>true</code> will result in calling {@link oj.ojMessage#close} for the message. 
 * Otherwise, the message will remain in its current state.
 *
 * <p>Closing a message changes the visibility to hidden. If the message is defined by an instance
 * in the {@link oj.ojMessages#messages} collection, the close operation will not remove the item
 * from the backing model. Application logic needs to listen for the 
 * {@link oj.ojMessage#event:close} event bubbling up from the underlying oj-message child to remove 
 * from the backing collection.</p>
 *
 * @expose
 * @function closeAll
 * @memberof! oj.ojMessages
 * @instance
 * @since 5.0.0
 * @param {function} [closeFilter] an optional callback function that will be passed an instance
 *                   of type {@link oj.ojMessage#message} for each child
 *                   oj-message.  If <code>closeFilter</code> returns <code>true</code>, the
 *                   associated oj-message will be closed.  Returning <code>false</code> will
 *                   exclude the child message from closure.  If a <code>closeFilter</code> is not
 *                   passed, all child messages will be closed.
 * @return {void}
 * @ojsignature {target: "Type",
 *               value: "(message: oj.ojMessage.Message) => boolean",
 *               for: "closeFilter",
 *               jsdocOverride: true}
 *
 * @example <caption>Invoke the <code class="prettyprint">closeAll</code> method:</caption>
 * myMessages.closeAll(function (message)
 * {
 *   // close all messages of error severity
 *   return "error" === message.severity;
 * });
 */

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 *
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @since 5.0.0
 *
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */

/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @since 5.0.0
 *
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */

/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 * @expose
 * @memberof! oj.ojMessages
 * @instance
 * @since 5.0.0
 *
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 **/


// Type Defs
////////////

/**
 * @typedef {Object} oj.ojMessages.PositionAlign
 * @property {"top"|"bottom"|"center"} [vertical] Vertical alignment.
 * @property {"start"|"end"|"left"|"center"|"bottom"} [horizontal] Horizontal alignment. <p>
 * <ul>
 *  <li><b>"start"</b> evaluates to "left" in LTR mode and "right" in RTL mode.</li>
 *  <li><b>"end"</b> evaluates to "right" in LTR mode and "left" in RTL mode.</li>
 * </ul>
 *
 */

/**
 * @typedef {Object} oj.ojMessages.PositionPoint
 * @property {number} [x] Horizontal aligment offset.
 * @property {number} [y] Vertical alignment offset.
 */

/**
 * @typedef {Object} oj.ojMessages.Position
 * @property {oj.ojMessages.PositionAlign} [my] Defines which edge on the popup to align with the target ("of") element.
 * @property {oj.ojMessages.PositionAlign} [at] Defines which position on the target element ("of") to align the positioned element
 *                                  against.
 * @property {oj.ojMessages.PositionPoint} [offset] Defines a point offset in pixels from the ("my") alignment.
 * @property {string|oj.ojMessages.PositionPoint} [of] Which element to position the popup against.
 *
 * If the value is a string, it should be a selector or the literal string value
 * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
 * is used, the values are relative to the whole document.  Page horizontal and vertical
 * scroll offsets need to be factored into this point - see UIEvent
 * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
 * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
 *
 * @property {"flip"|"fit"|"flipfit"|"none"} [collision] Rule for alternate alignment. <p>
 * <ul>
 *  <li><b>"flip"</b> the element to the opposite side of the target and the
 *             collision detection is run again to see if it will fit. Whichever side
 *             allows more of the element to be visible will be used. </li>
 * <li><b>"fit"</b> shift the element away from the edge of the window. </li>
 * <li><b>"flipfit"</b> first applies the flip logic, placing the element
 *  on whichever side allows more of the element to be visible. Then the fit logic
 *  is applied to ensure as much of the element is visible as possible.</li>
 * <li><b>"none"</b> no collision detection.</li>
 * </ul>
 */

var _MESSAGES_METADATA =
  {
    "name" : "oj-messages",
    "displayName" : "oj-messages",
    "version" : "1.0.0",
    "jetVersion" : "^5.0.0",
    "properties" : {
      "messages" : {
        "type" : "Array<object>",
        "writeback" : false,
        "value": null
      },

      "display" : {
        "type" : "string",
        "enumValues" : ["general", "notification"],
        "translatable" : false,
        "value": "general"
      },

      "position" : {
        "type" : "object",
        "translatable" : false,
        "properties" : {
          "my" : {
            "type" : "object|string",
            "properties" : {
              "horizontal" : {
                "type" : "string",
                "enumValues" : ["start", "end", "left", "center", "right"]
              },
              "vertical" : {
                "type" : "string",
                "enumValues" : ["top", "center", "bottom"]
              }
            }
          },
          "at" : {
            "type" : "object|string",
            "properties" : {
              "horizontal" : {
                "type" : "string",
                "enumValues" : ["start", "end", "left", "center", "right"]
              },
              "vertical" : {
                "type" : "string",
                "enumValues" : ["top", "center", "bottom"]
              }
            }
          },
          "offset" : {
            "type" : "object",
            "properties" : {
              "x" : {
                "type" : "number"
              },
              "y" : {
                "type" : "number"
              }
            }
          },
          "of" : {
            "type" : "string|object"
          },
          "collision" : {
            "type" : "string",
            "enumValues" : ["flip", "fit", "flipfit", "flipcenter", "none"]
          }
        }
      },

      "translations" : {
        "type" : "object",
        "writeback" : false,
        "value": {"labelLandmark": ""},
        "properties" : {
          "labelLandmark" : {
            "type" : "string",
            "translatable" : true
          }
        }
      }
    },

    "methods" : {
      "close" : {
        "params" : [
          {
            "name" : "message",
            "type" : "object"
          }
        ]
      },
      "closeAll" : {
        "params" : [
          {
            "name" : "filter",
            "type" : "function"
          }
        ]
      }
    }

  };

var _MESSAGES_VIEW =
  '<div role="presentation" aria-hidden="true" ' +
  '     :id="[[containerId]]" :class="[[containerSelectors]]" ' +
  '     on-oj-open="[[handleOpen]]" on-oj-close="[[handleClose]]" ' +
  '     on-oj-animate-start="[[handleAnimateStart]]">' +
  '  <oj-bind-slot>' +
  '  </oj-bind-slot>' +
//  - switch back to jet when observable inserts are recognized
//  '  <oj-bind-for-each data="[[$props.messages]]" as="item">' +
//  '    <template>' +
//  '      <oj-message message="{{item.data}}">' +
//  '      </oj-message>' +
//  '    </template>' +
//  '  </oj-bind-for-each>' +
  '    <!-- ko foreach: $props.messages -->' +
  '    <oj-message message="[[$data]]">' +
  '    </oj-message>' +
  '    <!-- /ko -->' +
  '</div>';

function MessagesViewModel(context)
{
  this._composite = context['element'];

  // Anything used by the view bindings or component methods can't be obfuscated
  // and is why it's defined in quotes.
  this._messagesContainerId = this["containerId"] = [context['unique'], "mc"].join("_");
  this["handleOpen"] = this._handleOpen.bind(this);
  this["handleClose"] = this._handleClose.bind(this);
  this["handleAnimateStart"] = this._handleAnimateStart.bind(this);
  this["bindingsApplied"] = this._bindingsApplied.bind(this);
  this["disconnected"] = this._disconnected.bind(this);
  this["connected"] = this._connected.bind(this);
  this["close"] = this._close.bind(this);
  this["closeAll"] = this._closeAll.bind(this);
  this["propertyChanged"] = this._propertyChanged.bind(this);

  this._properties = context['properties'];
  this._createObservables();
  this._initDefaultTranslations();
};

MessagesViewModel.prototype._bindingsApplied = function (context)
{
  // detects F6 navigation to the body content
  var messagesContainerDiv = document.getElementById(this._messagesContainerId);
  messagesContainerDiv.addEventListener("ojFocus", this._navigationEventListener.bind(this), false);

  // Shows the message container before animation of the first disclosed message.  The ojOpen
  // event is fired after animation.
  messagesContainerDiv.addEventListener("ojBeforeOpen", this._handleBeforeOpen.bind(this), false);

};

MessagesViewModel.prototype._disconnected = function (element)
{
  MessagesViewModel.NAVIGATION_TRACKER.remove(this._messagesContainerId);

  // detaching an open message overlay results in implicit dismissal
  if (oj.ZOrderUtils.getStatus(this._composite) === oj.ZOrderUtils.STATUS.OPEN)
  {
    this._closeOverlay();
  }
}

MessagesViewModel.prototype._connected = function (element)
{
  MessagesViewModel.NAVIGATION_TRACKER.add(this._messagesContainerId);
}

MessagesViewModel.prototype._closeAll = function (closeFilter)
{
  if (this._isMessagesShown())
  {
    var messageChildren = this._getDefaultSlotMessageElements();
    for (var i = 0; i < messageChildren.length; i++)
    {
      var message = messageChildren[i]["message"];
      var closeable = true;
      if (closeFilter)
      {
        closeable = closeFilter(message);
      }
      if (closeable)
      {
        messageChildren[i]['close']();
      }
    }
  }
};

MessagesViewModel.prototype._propertyChanged = function (detail)
{
  if ("external" === detail["updatedFrom"] && detail["property"] === "position")
  {
    if (detail.previousValue && detail.value)
    {
      // re-evaluate overlay position
      this._refresh();
    }
    else if (!detail.previousValue && detail.value)
    {
      // inline to overlay
      if (this._getDefaultSlotMessageElements().length > 0)
      {
        if (this._isMessagesShown())
          this._hideMessages();

        this._openOverlay();
      }
    }
    else if (detail.previousValue && !detail.value)
    {
      // overlay to inline
      if (this._getDefaultSlotMessageElements().length > 0)
      {
        if (this._isOverlayOpen())
          this._closeOverlay();

        this._showMessages();
      }
    }
  }
  else if ("external" === detail["updatedFrom"] && detail["property"] === "display")
  {
    if (this._getDefaultSlotMessageElements().length > 0)
    {
      if (this._isOverlayOpen())
      {
        this._closeOverlay();
        this._openOverlay();
        this._refresh();
      }
      else if (this._isMessagesShown())
      {
        this._hideMessages();
        this._showMessages();
      }
    }
  }
  else if ("external" === detail["updatedFrom"] && detail["property"] === "translations")
  {
    this._updateLandmark();
  }
};

MessagesViewModel.prototype._close = function (message)
{
  if (message && this._isMessagesShown())
  {
    var messageChildren = this._getDefaultSlotMessageElements();
    for (var i = 0; i < messageChildren.length; i++)
    {
      /** @type {?} */
      var childMessage = $(messageChildren[i]).prop("message");
      if (childMessage === message)
      {
        messageChildren[i]["close"]();
      }
    }
  }
};

MessagesViewModel.prototype._isEventPertaining = function (event)
{
  // verify the event is from a child message
  var target = event.target;
  var messagesContainerDiv = document.getElementById(this._messagesContainerId);
  if (target.nodeName !== "OJ-MESSAGE" || !oj.DomUtils.isAncestor(messagesContainerDiv, target))
    return false;
  else
    return true;
};

MessagesViewModel.prototype._handleBeforeOpen = function (event)
{
  if (!event.defaultPrevented && this._isEventPertaining(event) && !this._isMessagesShown())
  {
    if (!this._isPresentationInline())
      this._openOverlay();
    else
      this._showMessages();
  }
};

MessagesViewModel.prototype._handleOpen = function (event)
{
  // verify the event is from a child message
  if (event.defaultPrevented || !this._isEventPertaining(event))
    return;

  /** @type {{icon: string, category: string, severity:string, timestamp:string, summary:string,
      detail:string, autoTimeout:number, closeAffordance: string, sound: string}} */
  var message = event.detail["message"];

  var translations = oj.Translations.getComponentTranslations("oj-ojMessage")["categories"];
  var category = !message['category'] ? translations[message['severity']] : message['category'];
  var options = {"category": category, "summary": message['summary']};

  var liveRegion = this._getLiveRegion();
  var text = oj.Translations.getTranslatedString("oj-ojMessages.ariaLiveRegion.newMessage",
                 options);
  liveRegion.announce(text);

  this._refresh();  // re-evaluate the position as the overlay size can change.
};

MessagesViewModel.prototype._handleClose = function (event)
{
  // verify the event is from a child message
  if (event.defaultPrevented || !this._isEventPertaining(event))
    return;

  // look for next focus
  var closeMessageElement = event.target;
  var nextFocusElement;

  // If the event was triggered from the close icon, the originalEvent will be populated
  // Next focus is only needed when closure was from the close icon.  autoTimeout closure
  // shouldn't insist on focus handling.
  if (event["_originalEvent"])
    nextFocusElement = this._getNextFocus(closeMessageElement);

  if (nextFocusElement)
    nextFocusElement.focus();

  // remove the oj-message from DOM. Note that we are not calling 'oj.Components.subtreeHidden'
  //  because it is not required if we use jquery remove() as per the doc there
  $(closeMessageElement).remove();

  // if we do not have any "default" slot children, then the message being closed is the last one,
  //  close/hide the containers hence
  if (this._getDefaultSlotMessageElements().length === 0)
  {
    MessagesViewModel.NAVIGATION_TRACKER.togglePreviousFocus(this._messagesContainerId);

    if (this._isOverlayOpen())
      this._closeOverlay();
    else
      this._hideMessages();
  }
};

MessagesViewModel.prototype._getNextFocus = function (target)
{
  var messageElements = this._getDefaultSlotMessageElements();
  var n = messageElements.indexOf(target);
  var nextFocusMessage;
  if (n - 1 > -1)
    nextFocusMessage = messageElements[n - 1];
  else if (n + 1 <= messageElements.length - 1)
    nextFocusMessage = messageElements[n + 1];

  var nextFocusElement;
  if (nextFocusMessage)
    nextFocusElement = nextFocusMessage.querySelector('.oj-message-category[tabindex="-1"]');

  return nextFocusElement;
};

MessagesViewModel.prototype._handleAnimateStart = function (event)
{

  if (event.defaultPrevented || !this._isEventPertaining(event))
    return; // animateStart already overridden on the individual oj-message.

  // eat the event bubbling up from a child oj-message component
  event.preventDefault();
  //event.stopPropagation();

  // override animation
  var messageElement = event.detail["element"];
  var action = event.detail["action"];
  var display = this._isPresentationInline() ? "general" : this._computeDisplay();
  var options = this._getThemedAnimateOptions(display, action);
  //var component = this._composite;
  var endCallback = event.detail["endCallback"];

  // oj-messages doesn't publish animateStart/animateEnd so use the simpler syntax for now and allow
  // the event to bubble.
  //
  // oj.AnimationUtils.startAnimation(messageElement, action, options, component).then(endCallback);
  oj.AnimationUtils[options['effect']](messageElement, options).then(endCallback);
};

MessagesViewModel._DEFAULTS = {
  "general": {
    "animation": {
      "open":  {"effect":"expand","duration":"300ms"},
      "close": {"effect":"collapse","duration":"300ms"}
    },
    "position": {
      "my" : {"horizontal" : "center", "vertical" : "top"},
      "at" : {"horizontal" : "center", "vertical" : "top"},
      "of" : "window",
      "collision": "none"
    }
  },
  "notification": {
    "animation": {
      "open":  {"effect":"slideIn","duration":"300ms"},
      "close": {"effect":"slideOut","duration":"300ms", "direction": "end"}
    },
    "position": {
      "my" : {"horizontal" : "end", "vertical" : "top"},
      "at" : {"horizontal" : "end", "vertical" : "top"},
      "of" : "window",
      "collision": "none"
    }
  }
};

MessagesViewModel.prototype._getThemedAnimateOptions = function (display, action)
{
  var themedDefaults = oj.ThemeUtils.parseJSONFromFontFamily('oj-messages-option-defaults');
  if (themedDefaults && themedDefaults[display] && themedDefaults[display]["animation"] &&
      themedDefaults[display]["animation"][action])
    return themedDefaults[display]["animation"][action];
  else
    return MessagesViewModel._DEFAULTS[display]["animation"][action];
};

MessagesViewModel.prototype._computeDisplay = function ()
{
  var display = this._properties['display'];
  return display;
};

MessagesViewModel.prototype._isPresentationInline = function ()
{
  return !this._properties['position'] ? true : false;
};

MessagesViewModel.prototype._computeContainerSelectors = function ()
{
  var display = this._computeDisplay();
  var composite = $(this._composite);

  // something about the layout changed, take this opp. to update classes on the root element
  composite.removeClass("oj-messages-general oj-messages-notification oj-messages-inline");

  if (this._isPresentationInline())
    composite.addClass("oj-messages-inline")
  else
    composite.addClass(["oj-messages", display].join("-"));

  // a bogus binding on the view, and we return a literal here, but provides a convenient way to be
  //  able to set classes on the root element when the underlying observables change
  return "oj-messages-container";
};

MessagesViewModel.prototype._getThemedPosition = function()
{
  var display = this._computeDisplay();
  var themedDefaults = oj.ThemeUtils.parseJSONFromFontFamily('oj-messages-option-defaults');
  if (themedDefaults[display] && themedDefaults[display]["position"])
    return themedDefaults[display]["position"];
  else
    return MessagesViewModel._DEFAULTS[display]["position"];
};

 MessagesViewModel.prototype._getPositionAsJqUi =  function ()
 {
    var position = oj.PositionUtils.coerceToJqUi(this._computePosition());
    var isRtl = oj.DomUtils.getReadingDirection() === "rtl";
    position = oj.PositionUtils.normalizeHorizontalAlignment(position, isRtl);
    return position;
};

MessagesViewModel.prototype._computePosition = function ()
{
  var position = this._properties['position'];
  return oj.PositionUtils.coerceToJet(position, this._getThemedPosition());
};

MessagesViewModel.prototype._getDefaultSlotMessageElements = function ()
{
  function toSelector(node)
  {
    var selector = "";
    if (node)
    {
      if (node.id && node.id.length > 0)
        selector = "#" + node.id;
      else
      {
        selector = node.nodeName;
        var clazz = node.getAttribute("class");
        if (clazz)
          selector += "." + clazz.split(" ").join(".");

        if (node.parentNode)
          return toSelector(node.parentNode) + " > " + selector;
      }
    }

    return selector;
  }

  var element = document.getElementById(this._messagesContainerId);

  var slotMap = oj.BaseCustomElementBridge.getSlotMap(element);
  var messageElements = [];

  // we just need to deal with "default" slot
  var body = slotMap[""];  // default slot
  for (var i = 0; body && i < body.length; i++)
  {
    // any element child node of oj-messages must be a oj-message. Do this validation at times.
    if (body[i].nodeName !== "OJ-MESSAGE")
    {
      oj.Logger.error(["JET oj-messages id='", toSelector(this._composite),
                          "': can contain only oj-message children in its default slot. ",
                          "Found a child element id='", toSelector(body[i]), "'."].join(""));
      continue;
    }
    else
    {
      messageElements.push(body[i]);
    }
  }

  return messageElements;
};

MessagesViewModel.prototype._isMessagesShown = function ()
{
  return $(this._composite).is(":visible");
};

MessagesViewModel.prototype._showMessages = function()
{
  if (!this._isMessagesShown())
  {
    // unhide the oj-messages root and notify
    $(this._composite).show();
    oj.Components.subtreeShown($(this._composite));

    var messagesContainerDiv = document.getElementById(this._messagesContainerId);
    messagesContainerDiv.removeAttribute("aria-hidden");
    
    MessagesViewModel.NAVIGATION_TRACKER.add(this._messagesContainerId);
    this._announceNavigation();
  }
};

MessagesViewModel.prototype._hideMessages = function()
{
  if (this._isMessagesShown())
  {
    var messagesContainerDiv = document.getElementById(this._messagesContainerId);
    messagesContainerDiv.setAttribute("aria-hidden", "true");

    // hide the oj-messages node root and notify
    $(this._composite).hide();
    oj.Components.subtreeHidden(this._composite);
    
    MessagesViewModel.NAVIGATION_TRACKER.remove(this._messagesContainerId);
    if (this._liveRegion)
    {
      this._liveRegion.destroy();
      delete this._liveRegion;
    }
  }
};

MessagesViewModel.prototype._openOverlay = function()
{
  var composite = $(this._composite);
  var psOptions = {};
  psOptions[oj.PopupService.OPTION.POPUP] = composite;
  psOptions[oj.PopupService.OPTION.LAUNCHER] = composite;
  psOptions[oj.PopupService.OPTION.POSITION] = this._getPositionAsJqUi();
  psOptions[oj.PopupService.OPTION.EVENTS] = this._getPopupServiceEvents();

  psOptions[oj.PopupService.OPTION.LAYER_SELECTORS] = ["oj","messages","layer"].join("-");
  psOptions[oj.PopupService.OPTION.MODALITY] = oj.PopupService.MODALITY.MODELESS;
  psOptions[oj.PopupService.OPTION.CUSTOM_ELEMENT] = true;
  oj.PopupService.getInstance().open(psOptions);

  this._showMessages();

  this._overlayEventsCallback = MessagesViewModel._overlayEventsListener.bind(this, composite);

  // snap on tab key handler
  composite[0].addEventListener("keydown", this._overlayEventsCallback, false);
};

MessagesViewModel.prototype._closeOverlay = function()
{
  this._hideMessages();

  var composite = $(this._composite);

  /** @type {!Object.<oj.PopupService.OPTION, ?>} */
  var psOptions = {};
  psOptions[oj.PopupService.OPTION.POPUP] = composite;
  oj.PopupService.getInstance().close(psOptions);

  // remove tab key handler
  var overlayEventsCallback = this._overlayEventsCallback;
  delete this._overlayEventsCallback;
  composite[0].removeEventListener("keydown", overlayEventsCallback, false);
};

MessagesViewModel.prototype._isOverlayOpen = function()
{
  var composite = this._composite;
  var status = oj.ZOrderUtils.getStatus(composite);
  return (status === oj.ZOrderUtils.STATUS.OPENING ||
          status === oj.ZOrderUtils.STATUS.OPEN ||
          status === oj.ZOrderUtils.STATUS.CLOSING);
};

MessagesViewModel._overlayEventsListener = function(element, event)
{
  if (event.defaultPrevented)
    return;

  if (event.keyCode === $.ui.keyCode.TAB)
  {
    var target = event.target;
    var nodes = element.find(":tabbable");
    if (nodes.length > 0)
    {
      var firstNode = nodes[0];
      var lastNode = nodes[nodes.length - 1];
      if (firstNode === lastNode && target === firstNode)
      {
        // only one tabstop and key event is on the last/first, eat the event
        event.preventDefault();
      }
      else if (firstNode === target && event.shiftKey)
      {
        //tabbing backwards, cycle focus to last node
        event.preventDefault();
        lastNode.focus();  //tabbing backwards, cycle focus to last node
      }
      else if (lastNode === target && !event.shiftKey)
      {
        event.preventDefault();
        firstNode.focus(); //tabbing forwards, cycle to the first node
      }
    }
    else
    {
      // no tab stops, eat tab event
      event.preventDefault();
    }
  }
};

/**
 * Listens for the ojFocus event fired on the messages container by the navigation
 * tracker. Signals the navigation tracker has forced focus to the messages container
 * in response to F6 keypress from outside the messages container.
 *
 * @private
 * @instance
 * @param {Event|CustomEvent} event
 */
MessagesViewModel.prototype._navigationEventListener = function (event)
{
  // The extra id check to to make sure the event is not from ojDialog or ojPopup
  // The dialog and popup raise ojFocus events but the ojFocus raised from the
  // messages navigation tracker is dispatched on the container and doesn't bubble.
  if (event.target.id === this._messagesContainerId)
  {
    event.preventDefault();
    this._announceNavigation(true);
  }
};

/**
 * Announces F6 navigation commands depending on if focus is within the message region
 * and the platform (mobile, desktop).
 *
 * @private
 * @instance
 * @param {boolean=} isFocusWithin true if focus navigation is to the messages container
 */
MessagesViewModel.prototype._announceNavigation = function (isFocusWithin)
{
  var isVOSupported = (oj.AgentUtils.getAgentInfo()['os'] === oj.AgentUtils.OS.IOS ||
                       oj.AgentUtils.getAgentInfo()['os'] === oj.AgentUtils.OS.ANDROID);
  var key;
  if (isFocusWithin)
    key = isVOSupported ? undefined : "ariaLiveRegion.navigationFromKeyboard";
  else
    key = isVOSupported ? "ariaLiveRegion.navigationToTouch" : "ariaLiveRegion.navigationToKeyboard";

  var liveRegion = this._getLiveRegion();
  var message = oj.Translations.getTranslatedString(["oj-ojMessages", key].join("."));
  liveRegion.announce(message);

};

MessagesViewModel.prototype._getLiveRegion = function ()
{
  var id = this._messagesContainerId;
  if (!this._liveRegion)
    this._liveRegion = new LiveRegion(id);

  return this._liveRegion;
};

MessagesViewModel.prototype._getPopupServiceEvents = function ()
{
  var events = {};
  events[oj.PopupService.EVENT.POPUP_CLOSE] = this._closeOverlay.bind(this);
  events[oj.PopupService.EVENT.POPUP_REMOVE] = this._surrogateRemoveHandler.bind(this);
  events[oj.PopupService.EVENT.POPUP_REFRESH] = this._refresh.bind(this);

  return events;
};

MessagesViewModel.prototype._refresh = function ()
{
  if (this._isOverlayOpen())
  {
    var position = this._getPositionAsJqUi();

    var composite = $(this._composite);
    composite["position"](position);
  }
};

MessagesViewModel.prototype._surrogateRemoveHandler = function ()
{
 this._closeOverlay();
 this._composite.parentElement.removeChild(this._composite);
};

MessagesViewModel.prototype._createObservables = function()
{
  // this can't be obfuscated by the closure compiler
  this["containerSelectors"] = ko.pureComputed(this._computeContainerSelectors.bind(this), this);
};

MessagesViewModel.prototype._initDefaultTranslations = function()
{
  var properties = this._properties;

  if (oj.StringUtils.isEmptyOrUndefined(properties['translations']['labelLandmark']))
  {
    var label = oj.Translations.getTranslatedString("oj-ojMessages.labelLandmark");
    properties['translations']['labelLandmark'] = label;
  }

  this._updateLandmark();
};

MessagesViewModel.prototype._updateLandmark = function()
{
  var labelLandmark = this._properties['translations']['labelLandmark'];
  this._composite.setAttribute("aria-label", labelLandmark);
  this._composite.setAttribute("role", "complementary");
}

/**
 * Tracks all visible instances of oj-messages and handles F6 navigation to the the most
 * recently disclosed instance.  In addtion, F6 navigation from the messages back to the
 * last focus element.
 *
 * @const
 * @type {Object}
 */
MessagesViewModel.NAVIGATION_TRACKER = {
  /**
   * @type {Array.<string>}
   */
  _messagesContainerIds: [],
  /**
   * @type {Object}
   */
  _priorFocusCache: {},
  /**
   * @param {string} id
   */
  add: function (id)
  {
    // adds tracking of the messages container when it becomes disclosed.

    this.remove(id);
    this._messagesContainerIds.push(id);
    this._start(id);
  },
  /**
   * @param {string} id
   */
  remove: function (id)
  {
    // disabled tracking the message container when it's not disclosed

    var messagesContainerIds = this._messagesContainerIds;
    var i = messagesContainerIds.indexOf(id);
    if (i > -1)
      messagesContainerIds.splice(i, 1);

    this._stop(id);
  },
  /**
   * @param {string} id
   */
  togglePreviousFocus: function(id)
  {
    // Restore focus from the messages container to what had focus prior to establishing
    // focus within the messages container. Invoked from press F6 within the messages
    // container.

    var priorFocusCache = this._priorFocusCache;

    /** @type {Element} */
    var target =  priorFocusCache[id];

    if (target && $(target).is(":visible") && oj.ZOrderUtils.isAboveTopModalLayer(target))
    {
      target.focus();
      delete priorFocusCache[id];
      return true;
    }
    return false;
  },
  /**
   * @param {string} id
   * @param {Element|undefined} target
   */
  _addPriorFocusCache: function (id, target)
  {
    // Captures the element that had focus prior to navigation to a messages container.
    // navigation can be via F6 or mouse.

    this._priorFocusCache[id] = target;
  },
  /**
   * @param {string} id
   */
  _start: function (id)
  {
    // Establishes listeners on the message container.  Adds the document listeners if
    // it's the first messages container shown.

    var messagesContainerDiv = document.getElementById(id);
    if (!messagesContainerDiv)
      return;

    // establish capture focus and keydown bubble listener on the container
    var messageContainerCallback = this._messageContainerListener.bind(this, id);
    messagesContainerDiv.addEventListener("focus", messageContainerCallback, true);
    messagesContainerDiv.addEventListener("keydown", messageContainerCallback, false);
    messagesContainerDiv.addEventListener("click", messageContainerCallback, false);
    $(messagesContainerDiv).data("oj_messages_nmtl", messageContainerCallback);

    // return if already listening
    if (this._documentCallback)
      return;

    this._documentCallback = this._documentListener.bind(this);
    var docElement = document.documentElement;
    docElement.addEventListener("keydown", this._documentCallback, false);
    docElement.addEventListener("blur", this._documentCallback, true);
  },
  /**
   * @param {string} id
   */
  _stop: function (id)  {

    // Removes message container listeners when the messages container is hidden.
    // Hidding the last message container will remove the listeners on the document.

    var messagesContainerDiv = document.getElementById(id);
    if (messagesContainerDiv)
    {
      /** @type {?} */
      var messageContainerCallback = $(messagesContainerDiv).data("oj_messages_nmtl");
      if (messageContainerCallback)
      {
        messagesContainerDiv.removeEventListener("focus", messageContainerCallback, true);
        messagesContainerDiv.removeEventListener("keydown", messageContainerCallback, false);
        messagesContainerDiv.removeEventListener("click", messageContainerCallback, false);
      }
    }

    if (!this._documentCallback || this._messagesContainerIds.length > 0)
      return;

    var docElement = document.documentElement;
    docElement.removeEventListener("keydown", this._documentCallback, false);
    docElement.removeEventListener("blur", this._documentCallback, true);
    delete this._documentCallback;
  },
  /**
   * @param {!Element} target
   * @return {number}
   */
  _indexOfFocusWithin: function (target)
  {
    // checks if the target of the F6 keydown is within a messages container.
    var messagesContainerIds = this._messagesContainerIds;
    for (var i = 0; i < messagesContainerIds.length; i++)
    {
      var messagesContainerDiv = document.getElementById(messagesContainerIds[i]);
      if (messagesContainerDiv &&
          oj.DomUtils.isAncestorOrSelf(messagesContainerDiv, target))
        return i;
    }

    return -1;
  },
  /**
   * @param {Event} event
   */
  _documentListener: function (event)
  {
    if (event.defaultPrevented)
      return;

    // Listeners for F6 navigation to the messages region from any element in the document.  The
    // listener is a bubble keydown.  If the event has been prevented, it's ignored.  If
    // there are multiple oj-messages components in the page, the selection is based on the
    // most recently disclosed order.

    var messagesContainerIds = this._messagesContainerIds;
    // F6 keypress
    if ("keydown" === event.type && event.keyCode === 117 && messagesContainerIds.length > 0)
    {
      // Look to see if the event target is already within a message area.  If focus is within
      // do nothing as the F6 listener on the mesage area will handle.
      var s = this._indexOfFocusWithin(event.target);
      if (s > -1)
        return;  // @see #_messageContainerListener
      else
        s = messagesContainerIds.length - 1;

      // target the most recently disclosed
      for (var i = s; i > -1; i++)
      {
        var messagesContainerDiv = document.getElementById(messagesContainerIds[i]);

        // if the container doesn't exist or is not visible continue to the next most recently used
        if (!messagesContainerDiv || !$(messagesContainerDiv).is(":visible") ||
            !oj.ZOrderUtils.isAboveTopModalLayer(messagesContainerDiv))
          continue;

        var element = messagesContainerDiv.querySelector('.oj-message-category[tabindex="-1"]');

        event.preventDefault();

        // captrue what had focus prior to pressing F6
        this._addPriorFocusCache(messagesContainerIds[i], event.target);

        element.focus();

        // Use a custom message to signal navigation to the messages container. The event doesn't
        // bubble. The event is used to announce navigation instructions.
        var customEvent = new CustomEvent("ojFocus", {"bubbles" : false, "cancelable" : true});
        messagesContainerDiv.dispatchEvent(customEvent);
        break;
      }
    }
    else if ("blur" === event.type)
    {
      // capture blur listener used to keep track of what last had focus
      this._prevActiveElement = event.target;
    }
  },
  /**
   * @param {string} id
   * @param {Event} event
   */
  _messageContainerListener: function (id, event)
  {
    // capture focus and bubble keydown event listener attached to visible message containers

    if (event.defaultPrevented)
      return;

    if ("focus" === event.type || "click" === event.type)
    {
       // if prior focus was outside the messgaes container and new focus is within, capture
       // what had prior focus. This is the best attempt to capture F2 navigation when using
       // mouse or touch to navigate to the message container versus F2 keyboard.

       /** @type {?} */
       var messagesContainerDiv = document.getElementById(id);
       var prevActiveElement = this._prevActiveElement;
       if (prevActiveElement && messagesContainerDiv &&
           !oj.DomUtils.isAncestorOrSelf(messagesContainerDiv, prevActiveElement))
       {
        // attached to visible message components.  Tracks previous focus.
        this._addPriorFocusCache(id, prevActiveElement);
       }
    }
    else if ("keydown" === event.type &&
            (event.keyCode === 117 || event.keyCode === $.ui.keyCode.ESCAPE))
    {
      // F6 or ESC keypress from within the content of the messages container will toggle focus back
      // to to what had prior focus. An attempt is made to fixup the navigate even if arriving
      // using the mouse or touch.
      if (this.togglePreviousFocus(id))
        event.preventDefault();
    }
  }
};


/**
 * Utility for handling voice over messages sent to a aria live region.  Messages
 * announced will not be deleted until the live region utility instance is destroyed.
 * They are not cleared due to the chattiness and quick timming messages can be added
 * as it could be too quick to read.
 * @private
 * @constructor
 * @class LiveRegion
 * @ignore
 * @ojtsignore
 * @param {string} id of the messages container
 */
function LiveRegion (id)
{
  this.Init(id);
};

/**
 * Adds one to the reference counter instance.
 * @instance
 * @protected
 */
LiveRegion.prototype.Init = function (id)
{
  this._id = id;
};

/**
 * Decrements the reference counter destroying the assocaited shared DOM aria
 * live region element when there are no longer any messages using it.
 * @instance
 * @public
 */
LiveRegion.prototype.destroy = function ()
{
  var liveRegion = $(document.getElementById(LiveRegion._LIVE_REGION_ID));

  //remove all the messages per messages container
  var id = this._id;
  delete this._id;

  liveRegion.find('div[data-container-id="' + id + '"]' ).remove();

  if (liveRegion.children("div").length < 1)
    liveRegion.remove();
};

/**
 * Sends a message to the aria live region for voice over mode.
 * @instance
 * @public
 * @param {string} message to be announce in the live region
 */
LiveRegion.prototype.announce = function (message)
{
  var liveRegion = LiveRegion._getLiveRegion();
  var id = this._id;
  $("<div>").attr("data-container-id", id).text(message).appendTo(liveRegion);
};


/**
 * Creates or returns an existing aria live region used by messages.
 * @returns {jQuery} aria live region
 * @private
 */
LiveRegion._getLiveRegion = function ()
{
  var liveRegion = $(document.getElementById(LiveRegion._LIVE_REGION_ID));
  if (liveRegion.length === 0)
  {
    // only tracks additions
    liveRegion = $("<div>");
    liveRegion.attr({"id" : LiveRegion._LIVE_REGION_ID, "role" : "log", "aria-live"
        : "polite", "aria-relevant" : "additions"});
    liveRegion.addClass("oj-helper-hidden-accessible");
    liveRegion.appendTo(document.body);
  }
  return liveRegion;
};

LiveRegion._LIVE_REGION_ID = "__oj_messages_arialiveregion";

oj.Composite.register("oj-messages",
  {
    "view" : _MESSAGES_VIEW,
    "viewModel" : MessagesViewModel,
    "metadata" : _MESSAGES_METADATA
  });

/*
** Copyright (c) 2008, 2015, Oracle and/or its affiliates. All rights reserved.
**
**34567890123456789012345678901234567890123456789012345678901234567890123456789
*/


/**
 * ojModule Conventions
 * @namespace
 * @since 1.1
 */
oj.ModuleBinding = {};

/**
 * Default configuration values
 * ojModule binding's conventions may be overridden for the entire application after the ojs/ojmodule
 * module is loaded. For example:
 * <p><code class="prettyprint">
 * oj.ModuleBinding.defaults.modelPath = 'models/';
 * </code></p>
 * Note that the default names of the {@link oj.ModuleBinding.ConventionMethods optional lifecycle methods} 
 * on the ViewModel are different from their counterparts on the {@link ojModule.LifecycleListener LifecycleListener} 
 * interface
 * @property {string} viewPath default View path. Defaults to 'text!views/'
 * @property {string} viewSuffix default View suffix. Defaults to '.html'
 * @property {string} modelPath default Model suffix. Defaults to 'viewModels/'
 * @property {string} initializeMethod name of the initialialization method
 * (see {@link oj.ModuleBinding.ConventionMethods#initialize definition})
 * @property {string} disposeMethod name of the dispose method 
 * (see {@link oj.ModuleBinding.ConventionMethods#dispose definition})
 * @property {string} activatedHandler name of the 'activated' event handler
 * (see {@link oj.ModuleBinding.ConventionMethods#handleActivated definition})
 * @property {string} attachedHandler name of the 'attached' event handler 
 * (see {@link oj.ModuleBinding.ConventionMethods#handleAttached definition})
 * @property {string} detachedHandler name of the 'detached' event handler
 * (see {@link oj.ModuleBinding.ConventionMethods#handleDetached definition})
 * @property {string} bindingsAppliedHandler name of the 'bindingsApplied' event handler
 * (see {@link oj.ModuleBinding.ConventionMethods#handleBindingsApplied definition})
 * @property {string} deactivatedHandler name of the 'deactivated' event handler
 * (see {@link oj.ModuleBinding.ConventionMethods#handleDeactivated definition})
 * @property {string} transitionCompletedHandler name of the 'transitionCompleted' event handler
 * (see {@link oj.ModuleBinding.ConventionMethods#handleTransitionCompleted definition})
 * @export
 * @memberof oj.ModuleBinding
 */
oj.ModuleBinding.defaults = 
{
  'viewPath': 'text!views/',
  'viewSuffix': '.html',
  'modelPath': 'viewModels/',
  'initializeMethod': 'initialize',
  'disposeMethod': 'dispose',
  'activatedHandler': 'handleActivated',
  'attachedHandler': 'handleAttached',
  'detachedHandler': 'handleDetached',
  'bindingsAppliedHandler': 'handleBindingsApplied',
  'deactivatedHandler': 'handleDeactivated',
  'transitionCompletedHandler': 'handleTransitionCompleted'
};

/**
 * @ignore
 */
oj.ModuleBinding._EMPTY_MODULE = "oj:blank";

(function ()
{
  ko['bindingHandlers']['ojModule'] = 
  {
    'init' : function (element, valueAccessor, allBindingsAccessor, viewModel, bindingContext)
    {
      var currentViewModel;
      var currentAnimationPromise;
      var cache = {};
      var currentCacheKey;
      var pendingViewId = -1;
      var cacheHolder;
      var endCommentNode;
      var busyStateResolver;
      var currentEmpty;
      var currentCleanupMode;
      var isCustomElement = element.parentNode && element.parentNode.nodeName === "OJ-MODULE"; //custom element check

      var legacyViewModelMethodFunc = isCustomElement ? function(){}: _invokeViewModelMethod;
      var legacyLifecycleListenerFunc = isCustomElement ? function(){}: _invokeLifecycleListener;
      var viewModelMethodFunc = isCustomElement ? _invokeViewModelMethodOnElement : function(){};
      var dispatchLifecycleEventFunc = isCustomElement ? _dispatchLifecycleEvent : function(){};
      
      function resolveBusyState()
      {
        if (busyStateResolver)
        {
          busyStateResolver();
          busyStateResolver = null;
        }
      }

      var invokeModelDispose = function(model)
      {
        legacyViewModelMethodFunc(model, 'disposeMethod', [element, valueAccessor]);
      };
      
      var disposeAssociatedComponentViewModel = function ()
      {
        invokeModelDispose(currentViewModel);
      };
      
      var nodeDispose = function()
      {
        disposeAssociatedComponentViewModel();
        
        // call dispose() on cached models
        var keys = Object.keys(cache);
        for (var i=0; i<keys.length; i++)
        {
          var model = cache[keys[i]].model;
          invokeModelDispose(model);
        }
        resolveBusyState();
        
        // Knockout will call ko.cleanNode on all child nodes including the cacheHolder 
      };

      ko.utils.domNodeDisposal.addDisposeCallback(element, nodeDispose);
      
      var _IGNORE_PROMISE = new Error("Promise cancelled because ojModule is fetching new View and ViewModel");
      
      // This function is used to interrupt Promise chains when the new view/viewModel combination is being loaded
      var checkPendingId = function(id)
      {
        if (id != pendingViewId)
        {
          throw _IGNORE_PROMISE;
        }
      };
      
      var initCacheHolder = function()
      {
        if (!cacheHolder)
        {
          cacheHolder = document.createElement('div');
          cacheHolder.className = "oj-helper-module-cache";
          // it is Ok to insert the cache holder as the first element because
          // all current children of the element will be moved to the cache holder
          ko.virtualElements.prepend(element, cacheHolder); //@HTMLUpdateOK; cacheHolder is constructed above
        }
      };
      
      var contextElement = element;
      
      if (element.nodeType === 8) //comment
      {
        contextElement = element.parentNode;
        ko.virtualElements.setDomNodeChildren(element, []); // remove all child nodes of the virtual element
        endCommentNode = element.nextSibling;
      }

      ko.computed(function ()
      {
        pendingViewId++;
        
        if (!busyStateResolver)
        {
          
          var busyContext = oj.Context.getContext(contextElement).getBusyContext();
          busyStateResolver = busyContext.addBusyState(
                  {'description' : "ojModule binding on a node with the Id " + element.id + 
                   "is loading the module. Binding evaluator: " + valueAccessor.toString()});

        }
        
        var isInitial = pendingViewId === 0;
        
        var unwrap = ko.utils.unwrapObservable;
        var value = unwrap(valueAccessor());

        var moduleName;
        var viewName;
        var params;
        var modelFactory;
        var viewFunction;
        var cacheKey;
        var lifecycleListener;
        var animation;
        var requireFunc;
        var modelPath;
        var viewPath;
        var view;
        var viewModel;
        var cleanupMode;

        if (typeof value === 'string')
        {
          moduleName = value;
        }
        else 
        {
          moduleName = unwrap(value['name']);
          viewName = unwrap(value['viewName']);
          params = unwrap(value['params']);
          modelFactory = unwrap(value['viewModelFactory']);
          viewFunction = unwrap(value['createViewFunction']);
          cacheKey = unwrap(value['cacheKey']);
          lifecycleListener = unwrap(value['lifecycleListener']);
          animation = unwrap(value['animation']);
          view = unwrap(value['view']);
          viewModel = unwrap(value['viewModel']);
          requireFunc = unwrap(value['require']);
          cleanupMode = unwrap(value['cleanupMode']);
        }
        
        if (requireFunc != null && !(requireFunc instanceof Function))
        {
          viewPath = requireFunc['viewPath'];
          modelPath = requireFunc['modelPath'];
          requireFunc = requireFunc['instance'];
        }
        
        viewName = viewName == null ? moduleName : viewName;
        
        var empty = (oj.ModuleBinding._EMPTY_MODULE === viewName);
        
        var attachPromise = legacyLifecycleListenerFunc(lifecycleListener, 'activated', [element, valueAccessor]);
        dispatchLifecycleEventFunc(contextElement, 'ojTransitionStart', [viewModel]);
        
        var viewPromise;
        var modelPromise;
        
        if (empty) 
        {
          viewPromise = Promise.resolve([]);
          modelPromise = Promise.resolve(null);    
        }
        else
        {
          var cached = cacheKey == null ? null : cache[cacheKey];
          
          if (cached != null)
          {
            delete cache[cacheKey]; 
            viewPromise = Promise.resolve(cached.view);
            modelPromise = Promise.resolve(cached.model);
          }
        }
        
        if (viewPromise == null && view != null)
        {
          viewPromise = Promise.resolve(view);
        }
        
        if (modelPromise == null)
        {
          if (viewModel != null)
          {
            modelPromise = Promise.resolve(viewModel);
          }
          else if (modelFactory != null)
          {
            modelPromise = ko.ignoreDependencies(modelFactory['createViewModel'], modelFactory, [params, valueAccessor]);
          }
          
          if (modelPromise == null && moduleName != null)
          {
            if (modelPath == null)
            {
              modelPath = oj.ModuleBinding.defaults['modelPath'];
            }
            modelPromise = _getRequirePromise(requireFunc, modelPath + moduleName);
          }
          
          if (modelPromise != null)
          {
            // Wrap model promise to perform initialization
            modelPromise = modelPromise.then(
              function(id, viewModel)
              {
                checkPendingId(id);
                
                if (typeof viewModel === 'function')
                {
                  viewModel = new viewModel(params);
                }
                else 
                {
                  // If the function returns a value, use it as the new model instance
                  viewModel = legacyViewModelMethodFunc(viewModel, 'initializeMethod', [element, valueAccessor]) || viewModel;
                }
                
                return viewModel;
              }.bind(null, pendingViewId)
            );
            
            // Handle the case where the Model is responsible for creating a View
            if (viewPromise == null && viewFunction != null)
            {
              viewPromise = modelPromise.then(
                function(id, model)
                {
                  checkPendingId(id);
                  
                  if (model == null)
                  {
                    resolveBusyState();
                    throw "createViewFunction option cannot be used when the ViewModel is null";
                  }
                  
                  var viewMethod = model[viewFunction];
                  
                  if (viewMethod == null)
                  {
                    resolveBusyState();
                    throw "function specified by the createViewFunction option was not found on the ViewModel";
                  }
                  
                  return  viewMethod.call(model);
                }.bind(null, pendingViewId)
              ); 
            }  

          }
          if (viewPromise == null)
          {
            if (viewName != null)
            {
              if (viewPath == null)
              {
                viewPath = oj.ModuleBinding.defaults['viewPath'];
              }
              viewPromise = _getRequirePromise(requireFunc, viewPath + viewName + oj.ModuleBinding.defaults['viewSuffix']);
            }
            else
            {
              resolveBusyState();
              throw new Error('View name or view instance must be specified');
            }
          }
          
        }
        
        if (viewPromise == null)
        {
          resolveBusyState();
          throw new Error('ojModule is missing a View');
        }
        
        var modelAttachPromise;
        
        if (modelPromise != null)
        {
          modelAttachPromise = modelPromise.then(
              function(id, viewModel)
              {
                checkPendingId(id);
                return legacyViewModelMethodFunc(viewModel, 'activatedHandler', [element, valueAccessor]);
                
              }.bind(null, pendingViewId)
          );
          
        }
       
        // We will ahve to wait for any module animation that has already started. Allowing a new View to be loaded while
        // animation is in progress would potentially leave the DOM in 
        var masterPromise = Promise.all([viewPromise, modelPromise, attachPromise, modelAttachPromise, currentAnimationPromise]);
        
        masterPromise.then(
          function(id, values)
          {
            if (id != pendingViewId)
            {
              return;
            }
            
            var view = values[0];
            
            if (view == null)
            {
              resolveBusyState();
              throw "The module's View was resolved to null";
            }
            
            var nodes = _getDomNodes(view, resolveBusyState);
            var model = values[1];
            
            var saveInCache = false;
            var cachedNodeArray;
            
            var oldDomNodes = _getContainedNodes(element, cacheHolder, endCommentNode);
            var oldKoNodes = _getKoNodes(element, cacheHolder);
            
            
            if (currentCacheKey != null && !currentEmpty)
            {
              saveInCache = true;
              cachedNodeArray = oldDomNodes;
              initCacheHolder();
            }
            else if(isCustomElement) // need view nodes for the event payload
            {
              cachedNodeArray = oldDomNodes;
            }

            
            var oldNodesRemoved = false;
            
            var removeOldDomNodes = function(oldViewParent)
            {
              if (oldNodesRemoved)
              {
                return;
              }
              
              oldNodesRemoved = true;
              
              if (saveInCache) // legacy cache case - move the node into cacheHolder, then detach
              {
                // Keep the cached nodes connected as a workaround for the Knockout removing observable subscriptions when
                // they fire on disconnected nodes
                _moveDomNodes(oldDomNodes, cacheHolder);
                _detachOldView(element, oldViewParent || element, cacheHolder);
              }
              else if (isCustomElement && currentCleanupMode === 'none') // external cache case
              {
                // detach from DOM by simply removing to preserve applied binding on the nodes
                for (var i = 0; i < cachedNodeArray.length; i++) 
                {
                  var cachedNode = cachedNodeArray[i];
                  cachedNode.parentNode.removeChild(cachedNode);
                }
              }
              else // cache is not involved - clean up and detach
              {
                oldKoNodes.forEach(
                  function(n)
                  {
                    ko.cleanNode(n);
                  }
                );
                _detachOldView(element, oldViewParent || element, cacheHolder);
              }
              
              if (!isInitial) // ensure that this is not the very first view displayed by the binding
              {
                legacyLifecycleListenerFunc(lifecycleListener, 'detached', [element, valueAccessor, currentViewModel, cachedNodeArray]);
                legacyViewModelMethodFunc(currentViewModel, 'detachedHandler', [element, valueAccessor, cachedNodeArray]);
                viewModelMethodFunc(currentViewModel, 'disconnected');
                dispatchLifecycleEventFunc (contextElement, 'ojViewDisconnected', [currentViewModel, cachedNodeArray]);
                
                legacyLifecycleListenerFunc(lifecycleListener, 'deactivated', [element, valueAccessor, currentViewModel]);
                legacyViewModelMethodFunc(currentViewModel, 'deactivatedHandler', [element, valueAccessor]);
              }
              
              if (saveInCache)
              {
                _invokeOnSubtree(cachedNodeArray, oj.Components? oj.Components.subtreeHidden: null);
                cache[currentCacheKey] = {model: currentViewModel, view: cachedNodeArray};
              }
              else if (! (isCustomElement && currentCleanupMode === 'none'))
              {
                disposeAssociatedComponentViewModel();
              }
              currentViewModel = model;
              currentCacheKey = cacheKey;
              currentEmpty = empty;
              currentCleanupMode = cleanupMode;
            };
            
            
            var insertAndActivateNewNodes = function(targetElement)
            {
              targetElement = targetElement || element;
              
              _insertNodes(targetElement, nodes);
              
              var fromCache = cached != null;
              
              if (fromCache)
              {
                _invokeOnSubtree(nodes, oj.Components? oj.Components.subtreeShown: null);
              }
              
              legacyLifecycleListenerFunc(lifecycleListener, 'attached', [targetElement, valueAccessor, model, fromCache]);
              legacyViewModelMethodFunc(model, 'attachedHandler', [targetElement, valueAccessor, fromCache]);
              viewModelMethodFunc(model, 'connected');
              dispatchLifecycleEventFunc(contextElement, 'ojViewConnected', [model]);
              
              var bindingApplied = isCustomElement && _isBindingApplied(nodes, model);
              if (!fromCache && !bindingApplied)
              {
                var childBindingContext = bindingContext['createChildContext'](model,
                undefined, function (ctx)
                {
                  ctx['$module'] = model;
                  ctx['$params'] = params;
                });
                
                if (isCustomElement) {
                  // nulling out the composite binding context props from the contained module
                  childBindingContext['$parent'] = undefined;
                  childBindingContext['$parents'] = undefined;
                  childBindingContext['$parentContext'] = undefined;
                  childBindingContext['$props'] = undefined;
                  childBindingContext['$properties'] = undefined;
                  childBindingContext['$slotNodeCounts'] = undefined;
                  childBindingContext['$slotCounts'] = undefined;
                  childBindingContext['$unique'] = undefined;
                  childBindingContext['$uniqueId'] = undefined;
                }
                
                _applyBindingsToNodes(targetElement, nodes[0], childBindingContext, cacheHolder);
                
                legacyLifecycleListenerFunc(lifecycleListener, 'bindingsApplied', [targetElement, valueAccessor, model]);
                legacyViewModelMethodFunc(model, 'bindingsAppliedHandler', [targetElement, valueAccessor]);
              }
              
            };
            
            var transitionComplete = function()
            {
              legacyLifecycleListenerFunc(lifecycleListener, 'transitionCompleted', [element, valueAccessor, model]);
              legacyViewModelMethodFunc(model, 'transitionCompletedHandler', [element, valueAccessor]);
              viewModelMethodFunc(model, 'transitionCompleted');
              dispatchLifecycleEventFunc(contextElement, 'ojTransitionEnd', [model]);
              resolveBusyState();
            };
            
            
            if (animation != null)
            {
              var actx = _createAnimationContext(element, valueAccessor, isInitial, currentViewModel, model, isCustomElement);
              var promise = _animate(actx, animation, element, oldDomNodes, 
                                                 insertAndActivateNewNodes, removeOldDomNodes, transitionComplete);
              // wrap currentAnimationPromise is a promise that never gets rejected, so that the ojModule can still
              // navigate to a new View
              currentAnimationPromise = _createNoFailPromise(promise);
             
            }
            else
            {
              currentAnimationPromise = undefined;
            }
            
            if (!currentAnimationPromise)
            {
              removeOldDomNodes(null);
              insertAndActivateNewNodes(null);
              transitionComplete();
            }
             
          }.bind(null, pendingViewId),
          
          /* reject callback */
          function(id, reason)
          {
            if (reason === _IGNORE_PROMISE)
            {
              return;
            }
           
            if (reason != null)
            {
              resolveBusyState();
              oj.Logger.error(reason);
              // Additionally log the stack trace for the original error
              /*if (reason instanceof Error)
              {
                oj.Logger.error(reason.stack);
              }*/
            }
            
          }.bind(null, pendingViewId)
        
        );
        
      },
      null, 
      {'disposeWhenNodeIsRemoved' : element}
      );

      return {'controlsDescendantBindings' : true};
    }
  };
  
  // Allow ojModule binding on virtual elements (comment nodes) 
  ko.virtualElements.allowedBindings['ojModule'] = true;
  
  
  
  /**
   * @ignore
   */
  function _animate(actx, animation, element, oldDomNodes, insertAndActivateFunc, removeOldDomNodes, transitionComplete)
  {
    var canAnimateFunc = animation['canAnimate'];
    var animating = canAnimateFunc == null || canAnimateFunc.call(animation, actx);
    
    if (!animating)
    {
      return;
    }
         
    var newViewParent;
    var oldViewParent;
    
    var settings = animation['prepareAnimation'].call(animation, actx);
    if (settings)
    {
      newViewParent = settings['newViewParent'];
      oldViewParent = settings['oldViewParent'];
    }
    
    var targetElement = newViewParent || element;
    
    if (oldViewParent && oldViewParent !== element)
    {
      _moveDomNodes(oldDomNodes, oldViewParent);
    }
    else if (targetElement === element)
    {
      removeOldDomNodes(null);
    }
    // remove all nodes from the target element if it is not the main element managed by ojModule (that one is cleaned
    // separately, and it also contains cached views
    if (targetElement !== element)
    {
      ko.virtualElements.setDomNodeChildren(targetElement, []);
    }
    
    insertAndActivateFunc(targetElement);
    
    var newDomNodes = Array.prototype.slice.call(targetElement.childNodes);
    
    var viewInserted = false;
    
    var insertNewView = function()
    {
      if (viewInserted)
      {
        return;
      }
      viewInserted = true;
      if (element !== targetElement)
      {
        _insertNodes(element, newDomNodes);
        
        var isCustomElement = element.parentNode && element.parentNode.nodeName === "OJ-MODULE"; //custom element check
        if (oj.Components && !isCustomElement)
        {
           // The subtree just got moved to a new parent, so notify components
           // of the 'detach' imeddiately followed by the 'attach'
           _invokeOnSubtree(newDomNodes, oj.Components.subtreeDetached);
           _invokeOnSubtree(newDomNodes, oj.Components.subtreeAttached);
        }
      }
    }
    
    var removeOldView = removeOldDomNodes.bind(null, oldViewParent); // pass oldViewParent as the first parameter
    
    actx['newViewParent'] = newViewParent;
    actx['oldViewParent'] = oldViewParent;
    actx['oldViewNodes'] = oldDomNodes;
    actx['removeOldView'] = removeOldView;
    actx['insertNewView'] = insertNewView;
    
    var postAnimation  = function()
    {
      removeOldView();
      insertNewView();
      transitionComplete();
    }
    
    var animationPromise = animation['animate'].call(animation, actx).then(
      postAnimation, // success
      function(reason) // failure
      {
        postAnimation();
        oj.Logger.error("ojModule animation promise was rejected");
      }
    );
    
    return animationPromise;
  }

  /**
   * @ignore 
   */
  function _detachOldView(element, oldViewParent, cacheHolder)
  {
    oldViewParent = oldViewParent || element;
    var empty = [];
    // If the node being emptied is the ojModule host node, ensure that the cacheHolder stays
    if (cacheHolder && element === oldViewParent)
    {
      // Remove cacheHolder to avoid bindoings being deactivated
      cacheHolder.parentNode.removeChild(cacheHolder);
      empty.push(cacheHolder);
    }
    ko.virtualElements.setDomNodeChildren(oldViewParent, empty);
  }
  
  /**
   * @ignore
   */
  function _moveDomNodes(nodes, target)
  {
    nodes.forEach(
      function(n)
      {
        target.appendChild(n); //@HTMLUpdateOK; child nodes are module view
      }
    );
  }
  
  /**
   * @ignore
   */
  function _isBindingApplied (nodes, model) 
  {
    var bindingContext;
    for (var i = 0; i < nodes.length; i++) 
    {
      bindingContext = ko.contextFor(nodes[i]);
      if (bindingContext) 
        break;
    }
    return bindingContext && bindingContext['$module'] && bindingContext['$module'] == model;
  }
  
  /**
   * @ignore
   */
  function _isDocumentFragment(content) 
  {
    if (window['DocumentFragment'])
    {
      return content instanceof DocumentFragment;
    } 
    else 
    {
      return content && content.nodeType === 11;
    }
  };
  
  /**
   * @ignore
   */
  function _dispatchLifecycleEvent (element, eventName, params)
  {
    var detail = {};
    if (params[0])
      detail['viewModel'] = params[0];
    if (params[1])
      detail['view'] = params[1];
    var customEvent = new CustomEvent(eventName, {'detail': detail});
    element.dispatchEvent(customEvent);
  };
  
  /**
   * @ignore
   */
  function _invokeLifecycleListener(listener, method, params)
  {
    if (listener && listener[method])
    {
      var data = {'element': params[0], 'valueAccessor': params[1]};
      if (params.length > 2)
      {
        data['viewModel'] = params[2];
        if (params.length > 3)
        {
          data[typeof params[3] === 'boolean' ? 'fromCache': 'cachedNodes'] = params[3];
        }
      }
      // : suspend dependency detection while listeners are invoked
      var ret = ko.ignoreDependencies(listener[method], listener, [data]);
      return ret;
    }
  };
  
  /**
   * @ignore
   * @param {?Object} model 
   * @param {string} key 
   * @param {Array=} params
   * @return {Object|undefined}
   */
  function _invokeViewModelMethod(model, key, params)
  {
    if (model == null)
    {
      return;
    }
    var name = oj.ModuleBinding.defaults[key];
    if (name != null && model)
    {
      var handler = model[name];
      if (typeof handler === 'function')
      {
        
        var data = undefined;
        if (params)
        {
          data = {'element': params[0], 'valueAccessor': params[1]};
          if (params.length > 2)
          {
            data[typeof params[2] === 'boolean' ? 'fromCache': 'cachedNodes'] = params[2];
          }
        }
        // : suspend dependency detection while listeners are invoked
        var ret = ko.ignoreDependencies(handler, model, [data]);
        return ret;
      }
    }
  };
  
  /**
   * @ignore
   * @param {?Object} model 
   * @param {string} name
   */
  function _invokeViewModelMethodOnElement(model, name)
  {
    if(model && name)
    {
      var handler = model[name];
      if (typeof handler === 'function')
      {
        //suspend dependency detection while listeners are invoked
        ko.ignoreDependencies(handler, model);
      }
    }
  };
  
  /**
   * @ignore
   */
  function _getContainedNodes(container, cacheHolder, endCommentNode)
  {    
    var childList = [];
    var firstChild = ko.virtualElements.firstChild(container);
    
    for(var node= firstChild; node != null && node != endCommentNode; node = node.nextSibling)
    {
      if (node != cacheHolder)
      {
        childList.push(node);
      }
    }
    
    return childList;
  }
  
  /**
   * @ignore
   */
  function _getKoNodes(container, cacheHolder)
  {    
    var nodes = [];
    var firstChild = ko.virtualElements.firstChild(container);
    
    _koNodesForEach(firstChild, cacheHolder,
      function(node)
      {
        nodes.push(node);
      }
    );
    
    return nodes;
  }
  
  
  /**
   * @ignore
   */
  function _koNodesForEach(first, cacheHolder, callback)
  {
    var node = first;
    while(node != null)
    {
      var next = ko.virtualElements.nextSibling(node)
      var type = node.nodeType;
      if (node !== cacheHolder && (type === 1 || type === 8))
      {
        callback(node);
      }
      node = next;
    }
  }
  
  /**
   * @ignore 
   */
  function _applyBindingsToNodes(parentNode, first, bindingContext, cacheHolder)
  {
    // Workaround for KO not calling preprocessNode() on the node where .applyBindings() is called
    var provider = ko.bindingProvider['instance'];
    var preprocessNode = provider['preprocessNode'];
    
    if (preprocessNode)
    {
      _koNodesForEach(first, cacheHolder,
        function(node)
        {
          preprocessNode.call(provider, node);
        }
      );
      
      // Get the new first node since a new node could have been inserted at the very top
      first =  ko.virtualElements.firstChild(parentNode);
      
    }
    
    _koNodesForEach(first, cacheHolder,
      function(node)
      {
        ko.applyBindings(bindingContext, node);
      }
    );
  }
  
  /**
   * @ignore
   */
  function _insertNodes(container, nodes)
  {
    var nodeCount = nodes.length;
    for (var i=nodeCount-1; i>=0; i--)
    {
      ko.virtualElements.prepend(container, nodes[i]); //@HTMLUpdateOK; nodes are the module view
    }
  }
  
  /**
   * @ignore
   */
  function _createAnimationContext(node, valueAccessor, isInitial, oldViewModel, newViewModel, isCustomElement)
  {
    return {'node': isCustomElement ? node.parentNode : node, 
            'valueAccessor': isCustomElement ?  null : valueAccessor, 
            'isInitial' : isInitial,
            'oldViewModel': oldViewModel, 'newViewModel': newViewModel};
  }
  
  /**
   * @ignore
   */
  function _invokeOnSubtree(nodeArray, method)
  { 
    if (method)
    {
      for (var i = 0; i<nodeArray.length; i++)
      {
        var node = nodeArray[i];
        if (node.nodeType === 1/*element*/)
        {
          method(node);
        }
      }
    }
  }
  
  /**
   * @ignore 
   */
  function _getDomNodes(content, resolveBusyState)
  {
    if (typeof content === 'string')
    {
      content = ko.utils.parseHtmlFragment(content);
    }
    else if (_isDocumentFragment(content))
    {
      content = ko.utils.arrayPushAll([], content.childNodes);
    }
    else if (Array.isArray(content))
    {
      content = ko.utils.arrayPushAll([], content);
    }
    else
    {
      resolveBusyState();
      throw "The View (" + content + ") has an unsupported type";
    }
    return content;
  }
  
  /**
   * @ignore
   */
  function _getRequirePromise(requireFunc, module)
  {
    requireFunc = requireFunc ? requireFunc : require;
    return new Promise(
      function(resolve, reject)
      {
        requireFunc([module],
          function(loaded)
          {
            resolve(loaded);
          },
          function(reason)
          {
            oj.Logger.error("ojModule failed to load " + module);
            reject(reason);
          }
        
        );
      }
    );
  };
  
  /**
   * @ignore 
   */
  function _createNoFailPromise(promise)
  {
    if (!promise)
    {
      return promise;
    }
    var noFail  = new Promise(
      function(resolve, reject)
      {
        promise.then(
          resolve,
          resolve
        )
      }
    
    );
    return noFail;
  }
  

})();

/**
 * A duck-typing interface that defines optional ViewModel methods that the ojModule binding will invoke by convention
 * @interface
 * @name ConventionMethods
 * @memberof oj.ModuleBinding
 * @ojtsimport knockout
 */

/**
 * This method may be implemented on the ViewModel to perform initialization tasks. The method will be invoked only
 * if the ViewModel factory or the AMD module returns a ViewModel instance (as opposed to a constructor function).
 * Note that this method will be invoked only once when the ViewModel is created, i.e. it will not be called when the
 * View is being brought from cache
 * 
 * @method initialize
 * @instance
 * @param {Object} info  - an object with the following key-value pairs:
 * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
 * @param {Function} info.valueAccessor binding's value accessor
 * To retrieve the object passed as the 'params' option on the binding, use  
 * <code>ko.utils.unwrapObservable(info.valueAccessor()).params</code> if the entire binding value is an observable or
 * <code>ko.utils.unwrapObservable(info.valueAccessor().params)</code> if the bidning value is a plain object literal. 
 * @return {Object|undefined} - optional return value that will be used as the new ViewModel instance
 * @memberof oj.ModuleBinding.ConventionMethods
 */

/**
 * This method may be implemented by the ViewModel to perform cleanup tasks. Note that this method will be invoked only 
 * once before all binding's refernces to the ViewModel are released.
 * @method dispose
 * @instance
 * @param {Object} info  - an object with the following key-value pairs:
 * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
 * @param {Function} info.valueAccessor binding's value accessor
 * @return {void}
 * @memberof oj.ModuleBinding.ConventionMethods
 */
 
/**
  * Optional ViewModel method invoked when this ViewModel is about to be used for the View transition
  * @method handleActivated
  * @instance
  * @param {Object} info  - an object with the following key-value pairs:
  * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
  * @param {Function} info.valueAccessor binding's value accessor
  * @return {Promise|undefined} - If the callback returns a Promise, the next phase (attaching DOM) will be delayed until
  * the promise is resolved
  * @memberof oj.ModuleBinding.ConventionMethods
  */
 
 /**
  * Optional ViewModel method invoked after the View is inserted into the document DOM
  * @method handleAttached
  * @instance
  * @param {Object} info  - an object with the following key-value pairs:
  * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
  * @param {Function} info.valueAccessor binding's value accessor
  * @param {boolean} info.fromCache - a boolean indicating whether the module was retrieved from cache
  * @return {void}
  * @memberof oj.ModuleBinding.ConventionMethods
  */
 
 /**
  * Optional ViewModel method invoked after the bidnings are applied on this View. 
  * If the current View is retrieved from cache, the bindings will not be re-applied, 
  * and this callback will not be invoked.
  * @method handleBindingsApplied
  * @instance
  * @param {Object} info  - an object with the following key-value pairs:
  * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
  * @param {Function} info.valueAccessor binding's value accessor
  * @return {void}
  * @memberof oj.ModuleBinding.ConventionMethods
  */
  
  /**
  * Optional ViewModel method invoked after transition to the new View is complete. That includes any possible animation
  * between the old and the new View
  * @method handleTransitionCompleted
  * @instance
  * @param {Object} info  - an object with the following key-value pairs:
  * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
  * @param {Function} info.valueAccessor binding's value accessor
  * @return {void}
  * @memberof oj.ModuleBinding.ConventionMethods
  */

 /**
  * Optional ViewModel method invoked after the View is removed from the document DOM
  * @method handleDetached
  * @instance
  * @param {Object} info  - an object with the following key-value pairs:
  * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
  * @param {Function} info.valueAccessor binding's value accessor
  * @param {Array} info.cachedNodes an Array containing cached nodes for the View (if the cache is enabled)
  * @return {void}
  * @memberof oj.ModuleBinding.ConventionMethods
  */

 /**
  * Optional ViewModel method invoked when the View/ViewModel combination becomes inactive
  * @method handleDeactivated
  * @instance
  * @param {Object} info  - an object with the following key-value pairs:
  * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
  * @param {Function} info.valueAccessor binding's value accessor
  * @return {void}
  * @memberof oj.ModuleBinding.ConventionMethods
  */
  
  /**
   * <p>ojModule binding for Knockout.js manages content replacement within a particular region of the page. The binding 
   * can be used to implement navigation within a region of a single-page application.</p>
   * 
   * <p>
   * <h2>Terminology</h2>
   * <ul>
   * <li>View - the HTML markup to be displayed within the page fragment. The markup can contain Knockout bindings, which 
   * will be automatically activated after the View is attached to its container.</li>
   * <li>ViewModel - optional ViewModel that will be applied by Knockout to the View</li>
   * <li>Module - a combination of View and ViewModel</li>
   * </ul>
   * </p>
   * 
   * <p>
   * <h2>Supported Features</h2>
   * <ul>
   * <li>ViewModel and View definitions loaded using require.js</li>
   * <li>Ability to provide a ViewModel by supplying a factory instance</li>
   * <li>View Caching</li>
   * <li>Ability for a ViewModel to create a View</li>
   * <li>Notifications about the lifecycle events (Module activation and deactivation, attaching and detaching DOM, 
   * applying bindings)</li>
   * </ul>
   * </p>
   * 
   * <p>
   * <h2>Examples</h2>
   * <pre class="prettyprint">
   * &lt;!-- Simple module binding with the currentCenter observable containing the name of the ViewModel, 
   *      which is also used as the name of the View -->
   * &lt;div data-bind="ojModule: currentCenter">&lt;/div>
   * 
   * &lt;!-- Module binding on the comment node -->
   * &lt;!-- ko data-bind="ojModule: currentCenter" -->&lt!-- /ko -->
   *
   * &lt;!-- Module binding with a LifecycleListener -->
   * &lt;div data-bind="ojModule: {name: currentCenter, lifecycleListener: moduleLifecycle}">&lt;/div>
   *
   * &lt;!-- Module binding with caching -->
   * &lt;div data-bind="ojModule: {name: currentName, cacheKey: currentName}">&lt;/div>
   * 
   * &lt;!-- View-only module binding -->
   * &lt;div data-bind="ojModule: {viewName: currentName}">&lt;/div>
   * </pre>
   * </p>
   * 
   * <p>
   * <h2>Conventions</h2>
   * <ul>
   * <li>ViewModel modules will have .js extension, and Views will have .html extension</li>
   * <li>ViewModel modules are expected to be found under the 'viewModels/' folder, and Views are expected to be found 
   * under the 'views/' folder. Both viewModels and /views folders are expected to be in the 'base' RequireJs folder 
   * (usually js/)</li>
   * <li>Optional 'initialize' and 'dispose' functions will be invoked on the ViewModel after it is created or before 
   * it is destroyed respectively. See {@link oj.ModuleBinding.ConventionMethods ConventionMethods} for method definitions</li>
   * <li>Optional lifecycle event handlers can be implemented by the ViewModel as well. 
   * See {@link oj.ModuleBinding.ConventionMethods  ConventionMethods} for method definitions</li>.
   * </ul>
   * For information to modifying the convention defaults see {@link oj.ModuleBinding}
   * </p>
   * 
   * <p>
   * <h2>ViewModel's Lifecycle</h2>
   * The instance of the ViewModel is established as follows:
   * <ol>
   * <li>If the viewModelFactory is provided, the 'starter' object is determined by resolving a Promise returned 
   * by the factory's createViewModel() method. Otherwise, the 'starter' object will be assigned the return value 
   * of the AMD module specified by the 'name' option on the binding (if present)</li>
   * <li>If the 'starter' object is a function, it will be treated as a constructor that will be invoked to create the 
   * ViewModel instance, and the 'params' object specified by the binding's 'params' option will be passed into the constructor 
   * function as a parameter. Otherwise, the {@link oj.ModuleBinding.ConventionMethods#initialize initialize convention method}
   * will be invoked on the 'starter' instance. If the method does not return anything, the 'starter' object will be treated
   * as the ViewModel instance. Otherwise, the return value will become the ViewModel instance.</li>
   * </ol>
   * ViewModel's {@link oj.ModuleBinding.ConventionMethods#dispose dispose convention method} will be invoked before binding's
   * references to the ViewModel instance are released.
   * The following lifecycle listeners are supported on a ViewModel instance as convention methods:
   * <ul>
   * <li>{@link oj.ModuleBinding.ConventionMethods#handleActivated handleActivated}</li>
   * <li>{@link oj.ModuleBinding.ConventionMethods#handleDetached handleDetached}</li>
   * <li>{@link oj.ModuleBinding.ConventionMethods#handleDeactivated handleDeactivated}</li>
   * <li>{@link oj.ModuleBinding.ConventionMethods#handleAttached handleAttached}</li>
   * <li>{@link oj.ModuleBinding.ConventionMethods#handleBindingsApplied handleBindingsApplied}</li>
   * <li>{@link oj.ModuleBinding.ConventionMethods#handleTransitionCompleted handleTransitionCompleted}</li>
   * </ul>
   * </p>
   * 
   * <p>
   * <h2>Acessing ViewModel Data</h2>
   * The instance of the ViewModel will be applied to the View, i.e. it will serve as View's default data object. The same
   * instance will also be available as $module within Knockout binding expressions (this may be useful for accessing 
   * module's ViewModel in foreach iterators).
   * For the View-only modules, the default data object will be null. Parent ViewModel will be available as $parent. 
   * </p>
   * 
   * <p>
   * <h2>Avoiding Mutiple re-renders</h2>
   * This binding keeps track of changes to each of its options. If an observable holding an option value is mutated,
   * the view will be re-rendered. One of the following approaches may be used to avoid unnecessary re-renders:
   * <ul>
   * <li>Use a single observable to store the entire binding value with all attributes:
   * <p>
   * <pre class="prettyprint">
   * this.moduleValue = ko.observable({name: 'acc_details', cacheKey:'acc'});
   * ...
   * &lt;div data-bind="ojModule: moduleValue">&lt;/div>
   * </pre>
   * </p>
   * </li>
   * <li>Use the same observable for the name and cacheKey attributes:
   * <p>
   * <pre class="prettyprint">
   * this.moduleName = ko.observable('acc_details');
   * ...
   * &lt;div data-bind="ojModule: {name: moduleName, cacheKey: moduleName}}">&lt;/div> 
   * </pre>
   * </p>
   * </li>
   * <li>Use rate-limiting for the observables referenced by the binding:
   * <p>
   * <pre class="prettyprint">
   * this.moduleName = ko.observable('acc_details').extend({ rateLimit: 1 });
   * this.cKey = ko.observable('acc').extend({ rateLimit: 1 });
   * this.vName = ko.observable('accView').extend({ rateLimit: 1 });
   * ...
   * &lt;div data-bind="ojModule: {name: moduleName, cacheKey: cKey, viewName: vName}}">&lt;/div>
   * ...
   * // Mutating multiple observables independently is OK when they use a rate limit extender
   * this.moduleName('preferences');
   * this.cKey('prefs');
   * this.vName('prefsView);
   * </pre>
   * </p>
   * </li>
   * </ul>
   * </p>
   * 
   * <p>
   * <h2>Id Uniqueness for View Content</h2>ß
   * ojModule does not modify any HTML Ids specified within the Views. The application is responsible for avoiding Id 
   * conflicts. Possible cases to consider are:
   * <ol>
   * <li>Conflicts with Ids used on the hosting page</li>
   * <li>Conflicts caused by multiple instances of the same View within the HTML document</li>
   * <li>Conflicts during animated View transitions (both Views are attached to the document for the duration of transition)</li>
   * <li>Conflicts with the cached View content when ojModule's 'cacheKey' option is set. ojModule currently keeps cached Views
   * in the hidden subtree within the same HTML document</li>
   * </ol>
   * All cases above except case (2) can be solved by adopting a convention-based prefix for each View. For example, all
   * Ids in the 'Accounts' View would start with 'ac_', all Ids in the 'Preferences' View would start with 'pr_', etc.
   * Case (2) would require the use of Ids that would dynamically bound at runtime. The application could pass a unique 
   * Id prefix to each incarnation of the View using the 'params' option. 
   * </p>
   * 
   * <p>
   * <h2>Using ojModule with Require.js</h2>
   * Request <code>ojs/ojmodule</code> module before Knockout bindings are applied
   * </p>
   * 
   * @namespace
   * @name ojModule
   */
 
   /**
   * @name Options
   * @property {Promise|string|Array<Node>|DocumentFragment} view the View or a Promise for the View.
   * A value has to be a document fragment, an array of DOM nodes, or a string containing the HTML. This option takes
   * precedence over all other ways to load a View. Note that ojModule will not be cloning the document fragment or the node array provided 
   * by the function before using it as the module's View and applying bindings to it. If the application needs to have access to the original 
   * document fragment or node array, it should be setting the 'view' property to a cloned copy.
   * @property {Promise|Function|Object} viewModel the ViewModel(constrcutor or instance) or a Promise for the ViewModel.
   * This option takes precedence over all other ways to load a ViewModel
   * @property {string} name ViewModel name. The name will be
   * used to load an AMD module containing the definition of the ViewModel. If the module returns a function, it will be used
   * as a ViewModel constructor, and any other return value (including null) will be treated as a ViewModel instance.
   * If the entire binding's value is a string, it will be treated as a 'name' option. If the name is set to "oj:blank", 
   * an empty View will be displayed, and no ViewModel will be loaded
   * If you need to create a view-only module, use the <code>viewName</code> option
   * @property {string} viewName View name. If omitted, the name of the View is assumed to be the same as the name of the 
   * VewModel. If the name is set to "oj:blank", an empty View will be displayed, and no ViewModel will be loaded
   * @property {Function|{instance: Function, viewPath: string, modelPath: string}} require an instance of the require()
   * function to be used by this ojModule, or an object optionally defining the require instance, the View path prefix and
   * the ViewModel path prefix. It is recommended that the instance, the viewPath and the modelPath be defined if you are
   * using ojModule within a Composite component, and you are not passing the View and ViewModel instances to ojModule directly.
   * @property {Object} viewModelFactory instance of the factory that implements <code>createViewModel(params, valueAccessor)</code> 
   * method. The method has to a return a Promise that will resolve to the ViewModel instance or constructor.
   * @property {Object} params object that will be passed into the ViewModel constructor or the
   * createViewModel method of the View Model factory as a parameter. The object will be available as $params within the View's
   * Knockout binding expressions.
   * @property {string} createViewFunction name of the ViewModel function used to create a View. If this parameter is 
   * specified, the ViewModel will be responsible for providing the definition of the View. The function has to return 
   * a Promise that will be resolved to document fragment, an array of DOM nodes, or a string containing the HTML. Note that ojModule will not
   * be cloning the document fragment or the node array provided by the function before using it as the module's View and applying bindings to it.
   * If the application needs to have access to the original document fragment or node array, it should be returning a cloned copy.
   * @property {string} cacheKey The key used to cache the View after it is no longer displayed. Setting this parameter will 
   * enable View caching. When the View is about t be cached, its Knockout bindings will not be deactivated when 
   * the View is removed from DOM tree. The cache will be discarded after when the window object is destroyed or 
   * when the ojModule binding instance is deactivated. Since changes to any binding option will re-render the view, 
   * it is very important that cacheKey and the rest of the options are applied atomically. Failure to do so will result in
   * a wrong cached view being retrieved. See the 'Avoiding Multiple Re-renders' section above.
   * @property {Object} lifecycleListener An instance of the object implementing one or more methods defined by the
   * {@link ojModule.LifecycleListener LifecycleListener} duck-typing interface.
   * @property {Object} animation instance of the {@link ojModule.ModuleAnimation ModuleAnimation} duck-typing interface that will
   * manage animation effects during View transitions.
   * 
   * @instance
   * @memberof ojModule
   */
   
   /**
   * A duck-typing interface that defines methods for the ojModule binding's lifecycle listener. 
   * Use 'lifecycleListener' option on the ojModule binding to set the listener.
   * @interface
   * @name LifecycleListener 
   * @memberof ojModule
   * @ojtsignore
   */
  
   /**
    * Invoked when the binding starts loading a new View and ViewModel
    * @function activated
    * @param {Object} info  - an object with the following key-value pairs:
    * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
    * @param {Function} info.valueAccessor binding's value accessor
    * @return {Promise|undefined} - If the callback returns a Promise, the next phase (attaching DOM) will be delayed until
    * the promise is resolved
    * @memberof ojModule.LifecycleListener
    * @instance
    */
    
  
   
   /**
    * Invoked after the View is inserted into the document DOM
    * @method attached
    * @param {Object} info  - an object with the following key-value pairs:
    * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
    * @param {Function} info.valueAccessor binding's value accessor
    * @param {Object} info.viewModel ViewModel for the View being attached
    * @param {boolean} info.fromCache a boolean indicating the view was retrieved from cache
    * @return {void}
    * @memberof ojModule.LifecycleListener
    * @instance
    */
    
   /**
    * Invoked after the bidnings for the new View are applied. 
    * If the current View is retrieved from cache, the bindings will not be re-applied, and this callback will not be invoked.
    * @method bindingsApplied
    * @param {Object} info  - an object with the following key-value pairs:
    * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
    * @param {Function} info.valueAccessor binding's value accessor
    * @param {Object} info.viewMode ViewModel for the new View
    * @return {void}
    * @memberof ojModule.LifecycleListener
    * @instance
    */
  
   /**
    * Invoked after transition to the new View is complete. That includes any possible animation
    * between the old and the new View
    * @method transitionCompleted
    * @param {Object} info  - an object with the following key-value pairs:
    * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
    * @param {Function} info.valueAccessor binding's value accessor
    * @param {Object} info.viewMode ViewModel for the new View
    * @return {void}
    * @memberof ojModule.LifecycleListener
    * @instance
    */
    
  
   /**
    * Invoked after the View is removed from the document DOM
    * @method detached
    * @param {Object} info  - an object with the following key-value pairs:
    * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
    * @param {Function} info.valueAccessor binding's value accessor
    * @param {Object} info.viewModel ViewModel for the View being attached
    * @param {Array} info.cachedNodes an Array containing cached nodes for the View (if the cache is enabled)
    * @return {void}
    * @memberof ojModule.LifecycleListener
    * @instance
    */
 
   /**
    * Invoked when the View/ViewModel combination becomes inactive
    * @method deactivated
    * @param {Object} info  - an object with the following key-value pairs:
    * @param {Node} info.element DOM element or where the binding is attached. This may be a 'virtual' element (comment node)
    * @param {Function} info.valueAccessor binding's value accessor
    * @param {Object} info.viewModel ViewModel for the View being attached
    * @return {void}
    * @memberof ojModule.LifecycleListener
    * @instance
    */
    
  /**
   * A duck-typing interface that defines a contract for managing animations during ojModule View transitions. 
   * Use 'animation' option on the ojModule binding to set ModuleAnimation instance.
   * @interface
   * @name ModuleAnimation 
   * @memberof ojModule
   * @ojtsignore
   */
   
   /**
   * Optional method that determines whether the animated transition should proceed. If the method is not implemented, all
   * transitions will be allowed to proceed
   * @function canAnimate
   *
   * @param {Object} context a context object with the keys detailed below
   * @param {Node} context.node a DOM node where the ojModule binding is attached. This may be a virtual/comment node
   * @param {Function} context.valueAccessor value accessor for the binding
   * @param {boolean} context.isInitial true if the initial View is about to be displayed, false otherwise
   * @param {Object} context.oldViewModel the instance of the ViewModel for the old View
   * @param {Object} context.newViewModel the instance of the ViewModel for the new View
   *
   * @return {boolean} true if animation should proceed, false otherwise
   * @memberof ojModule.ModuleAnimation
   * @instance
   */
   
   /**
   * Prepares animation by designating where the new View should be inserted and optionally specifying where the old View
   * should be moved
   * @function prepareAnimation
   *
   * @param {Object} context a context object with the keys detailed below
   * @param {Node} context.node a DOM node where the ojModule binding is attached. This may be a virtual/comment node
   * @param {Function} context.valueAccessor value accessor for the binding
   * @param {boolean} context.isInitial true if the initial View is about to be displayed, false otherwise
   * @param {Object} context.oldViewModel the instance of the ViewModel for the old View
   * @param {Object} context.newViewModel the instance of the ViewModel for the new View
   *
   * @return {?Object} an object that may contain values for the following keys:
   * <ul>
   * <li>'newViewParent' - a DOM node where the new View should be inserted. If this parameter is not specified, the new View
   * will be inserted into the node associated with the ojModule binding</li>
   * <li>'oldViewParent' - a DOM node where the old View should be moved. If this parameter is not specified, the old View
   * will not be moved</li>
   * </ul>
   * @memberof ojModule.ModuleAnimation
   * @instance
   */
   
   /**
   * Prepares animation by designating where the new View should be inserted and optionally specifying where the old View
   * should be moved
   * @function animate
   *
   * @param {Object} context a context object with the keys detailed below
   * @param {Node} context.node a DOM node where the ojModule binding is attached. This may be a virtual/comment node
   * @param {Function} context.valueAccessor value accessor for the binding
   * @param {boolean} context.isInitial true if the initial View is about to be displayed, false otherwise
   * @param {Object} context.oldViewModel the instance of the ViewModel for the old View
   * @param {Object} context.newViewModel the instance of the ViewModel for the new View
   * @param {Node} context.newViewParent the 'newViewParent' parameter returned by the prepareAnimation() method
   * @param {Node} context.oldViewParent the 'oldViewParent' parameter returned by the prepareAnimation() method
   * @param {Function} context.removeOldView calling this function will remove the DOM nodes representing the old View. If this
   * function is not invoked by the ModuleAnimation implementation, and the old View is still connected when the Promise is 
   * resolved, the old View will be removed by ojModule.
   * @param {Function} context.insertNewView calling this function will insert new View's DOM nodes into the location
   * managed by the ojModule. If this function is not invoked by the ModuleAnimation implementation, and the new View is not at
   * its intended location when the Promise is resolved, the View will be moved by ojModule.
   * @param {Array} context.oldDomNodes an array of DOM nodes representing the old View
   * @return {Promise} - a Promise that should be resolved when the animation, moving/removing of DOM nodes and the
   * cleanup are complete. Note that ojModule will not be able to navigate to a new View until the Promise is resolved.
   *
   * @memberof ojModule.ModuleAnimation
   * @instance
   */
   

   
  
  
   
   
   
  

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojModule
 * @since 4.2.0
 * @ojdisplayname ojModule Element
 * @ojshortdesc Navigational element that manages content replacement within a particular region of the page.
 * @ojstatus preview

 * @classdesc 
 * <h3 id="ojModuleOverview-section">
 *   JET Module
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#ojModuleOverview-section"></a>
 * </h3>
 *
 * <p>
 * The oj-module custom element is used for binding a view and a corresponding view model to an element
 * to provide content replacement within a particular region of the page. In case of a single page application 
 * the element <code>config</code> attribute defines navigation within a region.
 * </p>
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-module config='[[moduleConfig]]' animation='[[moduleAnimation]]'>
 * &lt;/oj-module>
 * </code>
 * </pre>
 * <h2 id="lifecycle">View Model Lifecycle
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#lifecycle"></a>
 * </h2>
 * <p>
 * If a ViewModel is provided as a part of configuration for the oj-module element, the following optional
 * callback methods can be defined on its ViewModel and will be called at each stage of the
 * component's lifecycle. 
 *
 * <h4 class="name">connected<span class="signature">()</span></h4>
 * <div class="description">
 * <p>The optional method may be implemented on the ViewModel and will be invoked after the View is inserted into the DOM.</p>
 * <p>This method might be called multiple times - after the View is created and inserted into the DOM, 
 * after the View is reconnected after being disconnected and after a parent element, oj-module, with attached View 
 * is reconnected to the DOM.</p>
 * </div>
 *
 * <h4 class="name">transitionCompleted<span class="signature">()</span></h4>
 * <div class="description">
 * <p>This optional method may be implemented on ViewModel and invoked after transition to the new View is complete. 
 * That includes any possible animation between the old and the new View.</p>
 * </div>
 *
 * <h4 class="name">disconnected<span class="signature">()</span></h4>
 * <div class="description">
 * <p>This optional method maybe implemented on the ViewModel and will be invoked when the View is disconnected from the DOM.</p>
 * <p>This method might be called multiple times - after the View is disconnected from the DOM and after a parent 
 * element,oj-module, with attached View is disconnected from the DOM.</p>
 * </div>
 */

/**
 * Configuration object that defines a view and a corresponding view model responsible the markup section 
 * under oj-module. An application can also specify a clean up mode for the view as a part of the configuration object.
 * See details for each attribute.
 * 
 * @member
 * @name config
 * @memberof! oj.ojModule
 * @instance
 * @type {Object}
 * @default null
 * @ojshortdesc An array of strings of allowed MIME types or file extensions that can be uploaded. If not specified, accept all file types
 *
 * @example <caption>Initialize the ojModule element with the <code class="prettyprint">config</code> attribute:</caption>
 * &lt;oj-module config='[[moduleConfig]]'>&lt;/oj-module>
 * 
 * @example <caption>Get or set the <code class="prettyprint">accept</code> property after initialization:</caption>
 * // getter
 * var config = myModule.config;
 * 
 * // setter
 * myModule.config = {'view':view,'viewModel':viewModel}; // where view is an Array of DOM nodes and viewModel is a model object for the view
 */
 
/**
 * Defines the view for the ojModule. 
 * <p>Note that oj-module will not be cloning the node array before using it as the module's View 
 * and applying bindings to it. If the application needs to have an access to the original node array, 
 * it should be setting the 'view' property to a cloned copy.</p>
 * @expose
 * @name config.view
 * @ojshortdesc Defines module view.
 * @memberof! oj.ojModule
 * @instance
 * @type {Array.<Node>}
 * @default <code class="prettyprint">null</code>
 */

/**
 * Defines model for the view.
 * <p>The following optional lifecycle methods can be defined on the ViewModel object and will be called 
 * at the corresponding lifecycle stage. See <a href="./oj.ojModule.html#lifecycle">View Model Lifecycle section</a></p>
 * @expose
 * @name config.viewModel
 * @ojshortdesc The ViewModel instance.
 * @memberof! oj.ojModule
 * @instance
 * @type {Object}
 * @default <code class="prettyprint">null</code>
 */

/**
 * Defines the clean up mode for the view.
 * @expose
 * @name config.cleanupMode
 * @ojshortdesc The cleanup mode for the view.
 * @memberof! oj.ojModule
 * @instance
 * @type {string}
 * @default "onDisconnect"
 * @ojvalue {string} "onDisconnect" The View nodes will be destroyed on disconnect.
 * @ojvalue {string} "none" Use the setting to preserve the view, e.g. if the view and view model will be stored in external cache.
 */

/**
 * Instance of the {@link oj.ModuleElementAnimation} duck-typing interface that will manage animation effects during View transitions.
 * @member
 * @name animation
 * @memberof! oj.ojModule
 * @instance
 * @type {Object}
 * @default null
 * @ojshortdesc Defines animation object used for view transitions.
 */
 
/**
 * Triggered before transition to the new View is started - before View is inserted into the DOM.
 * @member
 * @name transitionStart
 * @memberof oj.ojModule
 * @instance
 * @event
 * @ojshortdesc Triggered before transition to the new View is started
 * @property {Object} viewModel ViewModel for the new View.
 */
 
/**
 * Triggered after the View is inserted into the DOM.
 * @member
 * @name viewConnected
 * @memberof oj.ojModule
 * @instance
 * @event
 * @ojshortdesc Triggered after the View is inserted into the DOM
 * @property {Object} viewModel ViewModel for the View being attached to the DOM.
 */

/**
 * Triggered after transition to the new View is complete. That includes any possible animation between the old and the new View.
 * @member
 * @name transitionEnd
 * @memberof oj.ojModule
 * @instance
 * @event
 * @ojshortdesc Triggered after transition to the new View is complete
 * @property {Object} viewModel ViewModel for the new View.
 */
 
/**
 * Triggered after the View is removed from the DOM
 * @member
 * @name viewDisconnected
 * @memberof oj.ojModule
 * @instance
 * @event
 * @ojshortdesc Triggered after the View is removed from the DOM
 * @property {Object} viewModel An array of DOM nodes that represent the disconnected View. The application can use the nodes to store them in cache.
 * @property {Array.<Node>} view ViewModel for the new View.
 */
 
var moduleMetadata =
{
  "properties": 
  {
    "config": 
    {
      "type": "object",
      "properties": 
      {
        "viewModel": {"type": "object"},
        "view": {"type": "array<object>"},
        "cleanupMode": 
        {
          "type":"string",
          "enumValues": ["onDisconnect", "none"],
          "value": "onDisconnect"
        }
      }
    },
    "animation": {"type": "object"}
  },
  "events": 
  {
    "transitionStart": {},
    "transitionEnd": {},
    "viewConnected": {},
    "viewDisconnected": {}
  }
};

function moduleViewModel (context) 
{
  var element = context.element;
  var props = context['properties'];
  
  function isViewAttached(config) 
  {
    var view = config ? config['view'] : null;
    return view && view.length > 0 && element.contains(view[0]);
  };
  
  function invokeViewModelMethod(model, name) 
  {
    var handler = model && model[name];
    if (typeof handler === 'function') 
    {
      ko.ignoreDependencies(handler, model);
    }
  };
  
  this['connected'] = function(context) 
  {
    if(isViewAttached(props && props['config'])) 
    {
      var model = props['config'] ? props['config']['viewModel'] : null;
      invokeViewModelMethod(model, 'connected');
    }
  };
  
  this['disconnected'] = function(context) 
  {
    if(isViewAttached(props && props['config']))
    {
      var model = props['config'] ? props['config']['viewModel'] : null;
      invokeViewModelMethod(model, 'disconnected');
    }
  };
};

var moduleValue = '{\"view\":$properties.config.view, \"viewModel\":$properties.config.viewModel,' + 
                  '\"cleanupMode\":$properties.config.cleanupMode,\"animation\":$properties.animation}';

var moduleView = "<!-- ko ojModule: "+ moduleValue +" --><!-- /ko -->";

oj.Composite.register('oj-module',
{
  "view": moduleView,
  "metadata": moduleMetadata,
  "viewModel": moduleViewModel
});

/**
 * A duck-typing interface that defines a contract for managing animations during the oj-module element View transitions. 
 * Use 'animation' attribute on the [Module]{@link oj.ojModule#animation} to set ModuleElementAnimation instance.
 * @interface ModuleElementAnimation
 * @memberof oj
 * @since 4.2.0
 * @export
 * @ojstatus preview
 */
 
 /**
 * Optional method that determines whether the animated transition should proceed. If the method is not implemented, all
 * transitions will be allowed to proceed
 * @method 
 * @name canAnimate
 *
 * @param {Object} context a context object with the keys detailed below
 * @param {Node} context.node a DOM node for the oj-module element
 * @param {boolean} context.isInitial true if the initial View is about to be displayed, false otherwise
 * @param {Object} context.oldViewModel the instance of the ViewModel for the old View
 * @param {Object} context.newViewModel the instance of the ViewModel for the new View
 *
 * @return {boolean} true if animation should proceed, false otherwise
 * @memberof oj.ModuleElementAnimation
 * @instance
 */
 
 /**
 * Prepares animation by designating where the new View should be inserted and optionally specifying where the old View
 * should be moved
 * @method
 * @name prepareAnimation
 *
 * @param {Object} context a context object with the keys detailed below
 * @param {Node} context.node a DOM node for the oj-module element
 * @param {boolean} context.isInitial true if the initial View is about to be displayed, false otherwise
 * @param {Object} context.oldViewModel the instance of the ViewModel for the old View
 * @param {Object} context.newViewModel the instance of the ViewModel for the new View
 *
 * @return {?Object} an object that may contain values for the following keys:
 * <ul>
 * <li>'newViewParent' - a DOM node where the new View should be inserted. If this parameter is not specified, the new View
 * will be inserted into the node associated with the oj-module element</li>
 * <li>'oldViewParent' - a DOM node where the old View should be moved. If this parameter is not specified, the old View
 * will not be moved</li>
 * </ul>
 * @memberof oj.ModuleElementAnimation
 * @instance
 */
 
 /**
 * Prepares animation by designating where the new View should be inserted and optionally specifying where the old View
 * should be moved
 * @method
 * @name animate
 *
 * @param {Object} context a context object with the keys detailed below
 * @param {Node} context.node a DOM node for the oj-module element
 * @param {boolean} context.isInitial true if the initial View is about to be displayed, false otherwise
 * @param {Object} context.oldViewModel the instance of the ViewModel for the old View
 * @param {Object} context.newViewModel the instance of the ViewModel for the new View
 * @param {Node} context.newViewParent the 'newViewParent' parameter returned by the prepareAnimation() method
 * @param {Node} context.oldViewParent the 'oldViewParent' parameter returned by the prepareAnimation() method
 * @param {Function} context.removeOldView calling this function will remove the DOM nodes representing the old View. If this
 * function is not invoked by the ModuleElementAnimation implementation, and the old View is still connected when the Promise is 
 * resolved, the old View will be removed by the component.
 * @param {Function} context.insertNewView calling this function will insert new View's DOM nodes into the location
 * managed by the component. If this function is not invoked by the ModuleElementAnimation implementation, and the new View is not at
 * its intended location when the Promise is resolved, the View will be moved by the component.
 * @param {Array} context.oldDomNodes an array of DOM nodes representing the old View
 * @return {Promise} - a Promise that should be resolved when the animation, moving/removing of DOM nodes and the
 * cleanup are complete. Note that the component will not be able to navigate to a new View until the Promise is resolved.
 *
 * @memberof oj.ModuleElementAnimation
 * @instance
 */
/**
 * Copyright (c) 2018, Oracle and/or its affiliates.
 * All rights reserved.
 */
 
/**
 * @class 
 * @name ModuleElementUtils
 * @ojstatus preview
 * @classdesc 
 * <h3>Utility methods for oj-module element</h3>
 * <p>ModuleElementUtils is a helper object that provides convenience methods for creating views and view models using RequireJS.
 * In addition to loading a view, the loadView() method also converts loaded data into an array of nodes - format accepted by oj-module.
 * </p>
 *
 * <h3> Usage : </h3>
 * <pre class="prettyprint"><code>
 * //create oj-module config as an observable and initialize it with empty values
 * self.moduleConfig = ko.observable({'view':[],'viewModel':null});
 *
 * //create view and view model, mutate configuration object to update oj-module content
 * var masterPromise = Promise.all([
 *   moduleElementUtils.createView({'viewPath':'views/ojModule-simple/page.html'}),
 *   moduleElementUtils.createViewModel({'viewModelPath':'viewModels/ojModule-simple/page'})
 *  ]);
 *  masterPromise.then(
 *    function(values){
 *      self.moduleConfig({'view':values[0],'viewModel':values[1]});
 *    },
 *    function(reason){}
 *  );
 * </code></pre>
 *
 * @hideconstructor
 * @since 5.0.0
 */
var ModuleElementUtils = function() {};

// Subclass from oj.Object 
oj.Object.createSubclass(ModuleElementUtils, oj.Object, "ModuleElementUtils");

// make it available internally
oj.ModuleElementUtils = ModuleElementUtils;

/** 
 * Utility function for creating a view to be used in configuration object for oj-module.
 * @param {Object} options Options object used to create a view
 * @param {string} options.viewPath The path to the view, relative to the RequireJS baseURL.
 *                                  The text plugin will be used for loading the view.
 * @param {Function=} options.require An optional instance of the require() function to be used for loading the view. 
 *                    By default the path is relative to the baseUrl specified for the application require calls. 
 *
 * @example <caption>Get promise for the view</caption>
 * var viewPromise = moduleElementUtils.createView({'viewPath':'views/ojModule-simple/page.html'});
 *
 * @return {Promise} A promise that resolves into an array of DOM nodes
 * @static
 */
ModuleElementUtils.createView = function(options)
{
  if (!(options && options['viewPath']))
    return Promise.resolve([]);

  var viewPromise = new Promise(function(resolve, reject) {
    var requireFunc = options['require'] ? options['require'] : require;
    requireFunc(['text!' + options['viewPath']], resolve, reject);
  });
  
  var returnPromise = new Promise( function(resolve, reject) {
    viewPromise.then(
      function(value){
        var viewArray = oj.__HtmlUtils.stringToNodeArray(value);
        return resolve(viewArray);
      },
      function(reason){
        return reject(reason);
      }
    );
  });
  return returnPromise;
};
/** 
 * Utility function for creating a view model to be used in configuration object for oj-module.
 * @param {Object} options Options object used to create a view model
 * @param {string} options.viewModelPath The path to the model, relative to the RequireJS baseURL.
 * @param {Function=} options.require An optional instance of the require() function to be used for loading the view. 
 *                    By default the path is relative to the baseUrl specified for the application require calls.
 *
 * @example <caption>Get promise for the model</caption>
 * var modelPromise = moduleElementUtils.createViewModel({'viewModelPath':'viewModels/ojModule-simple/pageModel'});
 *
 * @return {Promise} A promise that contains either model instance or a model constructor. 
 *                  When the promise is resolved into a constructor, the application is responsible 
 *                  for constructing the model instance before passing it to the configuration object on the oj-module.
 * @static
 */
ModuleElementUtils.createViewModel = function(options) 
{
  if (!(options && options['viewModelPath']))
    return Promise.resolve(null);
  
  return new Promise(function(resolve, reject) {
    var requireFunc = options['require'] ? options['require'] : require;
    requireFunc([options['viewModelPath']], resolve, reject);
  });
};

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */


/**
 * A collection of ModuleAnimation implementations that can be specified on the "animation"
 * option of ojModule binding.<br><br>
 * These implementations assume that either the ojModule binding is on a real HTML element
 * or, for ojModule binding on virtual element, each view is contained by a single HTML
 * root element.
 *
 * <h3 id="module-layout-section">
 *   Laying Out ojModule Views
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#module-layout-section"></a>
 * </h3>
 *
 * Guidelines on laying out ojModule views to acheive smoother animation:
 * <ul>
 * <li>The ojModule binding should be placed on a container div that has a dimension
 * independent of its children.  During view switch, new view is inserted and old view
 * is removed.  The container div acts as a placeholder to ensure other elements around
 * the views such as header and navigation controls do not move.
 *
 * <li>All views that will be loaded by the ojModule binding should have the same size.
 * For example by setting their height to 100%.  This will make sure the views fill up
 * the space in the container div.  Some animation effects such as sliding to cover or
 * reveal look best when the old view and the new view have the same size.
 *
 * <li>All views should have a background color.  This will make sure the old view
 * does not show through the new view, unless it is the intent of the application
 * to have a glass pane effect during the animation.
 *
 * <li>Avoid applying css style to any view element that is dependent on the
 * DOM hierarchy outside the view.  This is because during view change the view can be
 * temporarily attached to a different parent.  Having css style that assumes certain
 * hierarchy outside the view can make the desired style not to be applied during animation.
 * </ul>
 *
 * @namespace
 * @ojtsimport ojmodule-element
 * @since 1.2
 * @export
 */
oj.ModuleAnimations = {};

oj.ModuleAnimations._addContainedElements = function(node, roots)
{
  var child = ko.virtualElements.firstChild(node);
  while(child)
  {
    if (child.nodeType == 1)
    {
      roots.push(child);
    }
    else if (child.nodeType == 8)
    {
      this._addContainedElements(child, roots);
    }

    child = ko.virtualElements.nextSibling(child);
  }
};

oj.ModuleAnimations._cacheVirtualViewRoot = function(context, root)
{
  context['_ojOldRoot'] = root;
};

oj.ModuleAnimations._getVirtualViewRoot = function(context)
{
  return context['_ojOldRoot'];
};

oj.ModuleAnimations._defaultCanAnimate = function(context)
{
  // No animation for the initial module display
  if (context['isInitial'])
  {
    return false;
  }

  // We can animate if the module binding is on a real element
  if (context['node'].nodeType == 1)
  {
    return true;
  }

  // If the module binding is on a comment node, we can animate if the view
  // is single-rooted
  if (context['node'].nodeType == 8)
  {
    var children = [];
    oj.ModuleAnimations._addContainedElements(context['node'], children);
    if (children && children.length == 1)
    {
      oj.ModuleAnimations._cacheVirtualViewRoot(context, children[0]);
      return true;
    }
  }

  return false;
};

oj.ModuleAnimations._getOldView = function(context)
{
  var oldView;

  if (context['node'].nodeType == 1)
  {
    oldView = context['node'];
  }
  else if (context['node'].nodeType == 8)
  {
    oldView = oj.ModuleAnimations._getVirtualViewRoot(context);
  }

  return oldView;
};

oj.ModuleAnimations._createViewParent = function(oldView)
{
  var viewport = $(document.createElement('div'));
  var cssStyle = {'position': 'absolute',
                  'height': oldView.offsetHeight + 'px',
                  'width': oldView.offsetWidth + 'px',
                  'left': oldView.offsetLeft + 'px',
                  'top': oldView.offsetTop + 'px'};

  viewport.appendTo(oldView.offsetParent); //@HTMLUpdateOK; viewPort constructed above
  viewport.css(cssStyle);
  viewport.addClass('oj-animation-host-viewport');
  
  var host = document.createElement('div');
  host.className = 'oj-animation-host';

  viewport.append(host); //@HTMLUpdateOK; host is constructed above
  
  return host;
};

/**
  * Create and return a ModuleAnimation implementation that combines base effects from {@link oj.AnimationUtils}
  *
  * @param {null|string|Object} oldViewEffect - an animation effect for the outgoing view.<br><br>
  *                              If this is null, no animation will be applied.<br>
  *                              If this is a string, it should be one of the effect method names in oj.AnimationUtils.<br>
  *                              If this is an object, it should describe the effect:
  * @param {string} oldViewEffect.effect - the name of an effect method in oj.AnimationUtils
  * @param {*=} oldViewEffect.effectOption - any option applicable to the specific animation effect<br><br>
  *                                                 Replace <i>effectOption</i> with the actual option name.  More than one option can be specified.  
  *                                                 Refer to the method description in {@link oj.AnimationUtils} for available options.
  * @param {null|string|Object} newViewEffect - an animation effect for the incoming view.<br><br>
  *                                             This is in the same format as oldViewEffect.
  * @param {boolean} newViewOnTop - specify true to initially create the new view on top of the old view. 
  *                  This is needed for certain effects such as sliding the new view in to cover the old view.
  *                  Default is false.
  * @return {Object} an implementation of the ModuleAnimation interface
  * @export
  *@ojsignature [
  *  { target: "Type", for: "oldViewEffect", value: "{effect: oj.AnimationUtils.AnimationMethods, [key:string]: any}|oj.AnimationUtils.AnimationMethods|null"},
  *  { target: "Type", for: "newViewEffect", value: "{effect: oj.AnimationUtils.AnimationMethods, [key:string]: any}|oj.AnimationUtils.AnimationMethods|null"},
  *  { target: "Type", for: "returns", value: "oj.ModuleElementAnimation"}
  * ]
  * 
  * @example <caption>Create a custom ModuleAnimation that fades out old view by 50% and slides in the new view:</caption>
  * var customAnimation = oj.ModuleAnimations.createAnimation({"effect":"fadeOut", "endOpacity":0.5}, {"effect":"slideIn", "direction":"end"}, true);
  */
oj.ModuleAnimations.createAnimation = function(oldViewEffect, newViewEffect, newViewOnTop)
{
  return {
    'canAnimate': oj.ModuleAnimations._defaultCanAnimate,
    'prepareAnimation': function(context) {
      var viewParents = {};
      var oldView = oj.ModuleAnimations._getOldView(context);
      
      if (newViewEffect && !newViewOnTop)
      {
        viewParents['newViewParent'] = oj.ModuleAnimations._createViewParent(oldView);
      }
      
      if (oldViewEffect)
      {
        viewParents['oldViewParent'] = oj.ModuleAnimations._createViewParent(oldView);
      }

      if (newViewEffect && newViewOnTop)
      {
        viewParents['newViewParent'] = oj.ModuleAnimations._createViewParent(oldView);
      }    

      return viewParents;  
    },
    'animate': function(context) {
      var oldViewHost = context['oldViewParent'];
      var newViewHost = context['newViewParent'];

      var promises = [];

      if (oldViewHost && oldViewEffect)
      {
        promises.push(oj.AnimationUtils.startAnimation(oldViewHost, 'close', oldViewEffect));
      }

      if (newViewHost && newViewEffect)
      {
        promises.push(oj.AnimationUtils.startAnimation(newViewHost, 'open', newViewEffect));
      }

      var animatePromise = Promise.all(promises);

      return animatePromise.then(function(result) {
        oj.ModuleAnimations._postAnimationProcess(context);
      });
    }
  };
};

oj.ModuleAnimations._removeViewParent = function(context, hostProp)
{
  var host = context[hostProp];
  
  if (host)
  {
    var viewport = host.parentNode; 
    if (viewport && viewport.parentNode)
    {
      viewport.parentNode.removeChild(viewport);
    }
  }
};

oj.ModuleAnimations._postAnimationProcess = function(context)
{
  context['removeOldView']();
  context['insertNewView']();

  oj.ModuleAnimations._removeViewParent(context, 'newViewParent');
  oj.ModuleAnimations._removeViewParent(context, 'oldViewParent');
};

oj.ModuleAnimations._getModuleEffect = function(animateName)
{
  if (oj.ModuleAnimations._moduleEffects == null)
  {
    oj.ModuleAnimations._moduleEffects = oj.ThemeUtils.parseJSONFromFontFamily('oj-animation-module-effects');
  }

  if (oj.ModuleAnimations._moduleEffects)
  {
    return oj.ModuleAnimations._moduleEffects[animateName];
  }
  
  return null;
};

oj.ModuleAnimations._getImplementation = function(animateName)
{
  var descriptor = oj.ModuleAnimations._getModuleEffect(animateName);
  if (descriptor)
  {
    return oj.ModuleAnimations.createAnimation(descriptor['oldViewEffect'],
                                               descriptor['newViewEffect'],
                                               descriptor['newViewOnTop']);
  }
  
  return null;
};

oj.ModuleAnimations._getNavigateMethod = function(context, navigationType)
{
  if (oj.ModuleAnimations._navigateMethods == null)
  {
    oj.ModuleAnimations._navigateMethods = oj.ThemeUtils.parseJSONFromFontFamily('oj-animation-navigate-methods');
  }

  if (oj.ModuleAnimations._navigateMethods)
  {
    return oj.ModuleAnimations._navigateMethods[navigationType];
  }
  
  return null;
};

oj.ModuleAnimations._navigateCanAnimate = function(context, navigationType)
{
  var animateName = oj.ModuleAnimations._getNavigateMethod(context, navigationType);
  if (oj.ModuleAnimations[animateName])
  {
    return oj.ModuleAnimations[animateName]['canAnimate'] == null ||
           oj.ModuleAnimations[animateName]['canAnimate'](context);
  }
  
  return false;
};

oj.ModuleAnimations._navigatePrepareAnimation = function(context, navigationType)
{
  var animateName = oj.ModuleAnimations._getNavigateMethod(context, navigationType);
  if (oj.ModuleAnimations[animateName] && oj.ModuleAnimations[animateName]['prepareAnimation'])
  {
    return oj.ModuleAnimations[animateName]['prepareAnimation'](context);
  }
  
  return null;
};

oj.ModuleAnimations._navigateAnimate = function(context, navigationType)
{
  var animateName = oj.ModuleAnimations._getNavigateMethod(context, navigationType);
  if (oj.ModuleAnimations[animateName] && oj.ModuleAnimations[animateName]['animate'])
  {
    return oj.ModuleAnimations[animateName]['animate'](context);
  }
  
  return Promise.resolve();
};

oj.ModuleAnimations._getNavigateImplementation = function(navigationType)
{
  return {
    'canAnimate': function(context) {
      return oj.ModuleAnimations._navigateCanAnimate(context, navigationType);
    },
    'prepareAnimation': function(context) {
      return oj.ModuleAnimations._navigatePrepareAnimation(context, navigationType);
    },
    'animate': function(context) {
      return oj.ModuleAnimations._navigateAnimate(context, navigationType);
    }
  };
};

/**
 * ModuleAnimation implementation for changing views by
 * sliding the new view left to cover the old view.
 * @export
 * @memberof oj.ModuleAnimations
 * @ignore
 */
oj.ModuleAnimations.coverLeft = oj.ModuleAnimations._getImplementation('coverLeft');

/**
 * ModuleAnimation implementation for changing views by
 * sliding the new view left to cover the old view.
 * @export
 * @memberof oj.ModuleAnimations
 * @ignore
 */
oj.ModuleAnimations.coverRight = oj.ModuleAnimations._getImplementation('coverRight');

/**
 * ModuleAnimation implementation for changing views by
 * sliding the old view left to reveal the new view.
 * @export
 * @memberof oj.ModuleAnimations
 * @ignore
 */
oj.ModuleAnimations.revealLeft = oj.ModuleAnimations._getImplementation('revealLeft');

/**
 * ModuleAnimation implementation for changing views by
 * sliding the old view right to reveal the new view.
 * @export
 * @memberof oj.ModuleAnimations
 * @ignore
 */
oj.ModuleAnimations.revealRight = oj.ModuleAnimations._getImplementation('revealRight');

/**
 * ModuleAnimation implementation for changing views by
 * sliding the new view in to cover the old view.<br><br>
 * This will take the text direction of the page into account.  On "ltr" page,
 * the new view will slide to the left.  On "rtl" page, it will slide to the
 * right.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.coverStart = oj.DomUtils.getReadingDirection() === 'rtl' ?
                                 oj.ModuleAnimations.coverRight :
                                 oj.ModuleAnimations.coverLeft;

/**
 * ModuleAnimation implementation for changing views by
 * sliding the old view out to reveal the new view.<br><br>
 * This will take the text direction of the page into account.  On "ltr" page,
 * the old view will slide to the right.  On "rtl" page, it will slide to the
 * left.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.revealEnd = oj.DomUtils.getReadingDirection() === 'rtl' ?
                                oj.ModuleAnimations.revealLeft :
                                oj.ModuleAnimations.revealRight;

/**
 * ModuleAnimation implementation for changing views by
 * sliding the new view up to cover the old view.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.coverUp = oj.ModuleAnimations._getImplementation('coverUp');

/**
 * ModuleAnimation implementation for changing views by
 * sliding the old view down to reveal the new view.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.revealDown = oj.ModuleAnimations._getImplementation('revealDown');

/**
 * ModuleAnimation implementation for changing views by
 * zooming in on the new view.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.zoomIn = oj.ModuleAnimations._getImplementation('zoomIn');

/**
 * ModuleAnimation implementation for changing views by
 * zooming out on the old view.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.zoomOut = oj.ModuleAnimations._getImplementation('zoomOut');

/**
 * ModuleAnimation implementation for changing views by
 * fading in the new view and fading out the old view.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.fade = oj.ModuleAnimations._getImplementation('fade');

/**
 * ModuleAnimation implementation for changing views by
 * having the new view push out the old view towards the start of the reading order.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.pushStart = oj.ModuleAnimations._getImplementation('pushStart');

/**
 * ModuleAnimation implementation for changing views by
 * having the new view push out the old view towards the end of the reading order.
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 */
oj.ModuleAnimations.pushEnd = oj.ModuleAnimations._getImplementation('pushEnd');

/**
 * ModuleAnimation implementation for navigating to child views by
 * using animations that are dependent on the platform theme.
 * <p>The default animations for each platform theme are as follows:</p>
 * <ul>
 *   <li>Web theme: coverStart</li>
 *   <li>iOS theme: coverStart</li>
 *   <li>Android theme: coverUp</li>
 *   <li>Windows theme: zoomIn</li>
 * </ul>
 * <p>The default animation can be changed with the following theme variable:<br>
 * <code class="prettyprint">$animationNavChildDefault</code>: specify
 * the animation for navigating to child views.<br>
 * <p>Valid values are the names of specific animation implementations such as
 * "coverStart".</p>
 * @export
 * @memberof oj.ModuleAnimations
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @example <caption>Set the default in the theme (SCSS) :</caption>
 * $animationNavChildDefault:  coverStart  !default;
 */
oj.ModuleAnimations.navChild = oj.ModuleAnimations._getNavigateImplementation('navChild');

/**
 * ModuleAnimation implementation for navigating to parent views by
 * using animations that are dependent on the platform theme.
 * <p>The default animations for each platform theme are as follows:</p>
 * <ul>
 *   <li>Web theme: revealEnd</li>
 *   <li>iOS theme: revealEnd</li>
 *   <li>Android theme: revealDown</li>
 *   <li>Windows theme: zoomOut</li>
 * </ul>
 * <p>The default animation can be changed with the following theme variable:<br>
 * <code class="prettyprint">$animationNavParentDefault</code>: specify
 * the animation for navigating to parent views.</p>
 * <p>Valid values are the names of specific animation implementations such as
 * "revealEnd".</p>
 * @export
 * @memberof oj.ModuleAnimations
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @example <caption>Set the default in the theme (SCSS) :</caption>
 * $animationNavParentDefault:  revealEnd  !default;
 */
oj.ModuleAnimations.navParent = oj.ModuleAnimations._getNavigateImplementation('navParent');

/**
 * ModuleAnimation implementation for drilling in to views by
 * using animations that are dependent on the platform theme.
 * <p>The default animations for each platform theme are as follows:</p>
 * <ul>
 *   <li>Web theme: coverStart</li>
 *   <li>iOS theme: coverStart</li>
 *   <li>Android theme: coverUp</li>
 *   <li>Windows theme: zoomIn</li>
 * </ul>
 * @export
 * @ignore
 * @memberof oj.ModuleAnimations
 * @deprecated This is deprecated.  Please use <a href="#navChild">navChild</a> instead.
 */
oj.ModuleAnimations.drillIn = oj.ModuleAnimations.navChild;

/**
 * ModuleAnimation implementation for drilling out of views by
 * using animations that are dependent on the platform theme.
 * <p>The default animations for each platform theme are as follows:</p>
 * <ul>
 *   <li>Web theme: revealEnd</li>
 *   <li>iOS theme: revealEnd</li>
 *   <li>Android theme: revealDown</li>
 *   <li>Windows theme: zoomOut</li>
 * </ul>
 * @export
 * @ignore
 * @memberof oj.ModuleAnimations
 * @deprecated This is deprecated.  Please use <a href="#navParent">navParent</a> instead.
 */
oj.ModuleAnimations.drillOut = oj.ModuleAnimations.navParent;

/**
 * ModuleAnimation implementation for navigating to sibling views earlier in the reading order by
 * using animations that are dependent on the platform theme.
 * <p>The default animations for each platform theme are as follows:</p>
 * <ul>
 *   <li>Web theme: none</li>
 *   <li>iOS theme: none</li>
 *   <li>Android theme: pushEnd</li>
 *   <li>Windows theme: pushEnd</li>
 * </ul>
 * <p>The default animation can be changed with the following theme variable:<br>
 * <code class="prettyprint">$animationNavSiblingEarlierDefault</code>: specify
 * the animation for navigating to sibling views earlier in the reading order.</p>
 * <p>Valid values are the names of specific animation implementations such as
 * "pushEnd".</p>
 * @export
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @memberof oj.ModuleAnimations
 * @example <caption>Set the default in the theme (SCSS) :</caption>
 * $animationNavSiblingEarlierDefault:  pushEnd  !default;
 */
oj.ModuleAnimations.navSiblingEarlier = oj.ModuleAnimations._getNavigateImplementation('navSiblingEarlier');

/**
 * ModuleAnimation implementation for navigating to sibling views later in the reading order by
 * using animations that are dependent on the platform theme.
 * <p>The default animations for each platform theme are as follows:</p>
 * <ul>
 *   <li>Web theme: none</li>
 *   <li>iOS theme: none</li>
 *   <li>Android theme: pushStart</li>
 *   <li>Windows theme: pushStart</li>
 * </ul>
 * <p>The default animation can be changed with the following theme variable:<br>
 * <code class="prettyprint">$animationNavSiblingLaterDefault</code>: specify
 * the animation for navigating to sibling views later in the reading order.</p>
 * <p>Valid values are the names of specific animation implementations such as
 * "pushStart".</p>
 * @export
 * @memberof oj.ModuleAnimations
 * @ojsignature {target: "Type", value: "oj.ModuleElementAnimation"}
 * @example <caption>Set the default in the theme (SCSS) :</caption>
 * $animationNavSiblingLaterDefault:  pushStart  !default;
 */
oj.ModuleAnimations.navSiblingLater = oj.ModuleAnimations._getNavigateImplementation('navSiblingLater');

/**
  * Returns an implementation of ModuleAnimation interface that switches between different animation implementations
  *
  * @param {Function} callback - a callback function whose return value should be a string containing one of the animation types
  * supported by oj.ModuleAnimations. The function will be passed a context object with the keys detailed below:
  * <ul>
  * <li>node - a DOM node where the ojModule binding is attached. This may be a virtual/comment node</li>
  * <li>valueAccessor - value accessor for the binding</li>
  * <li>isInitial  - true if the initial View is about to be displayed, false otherwise</li>
  * <li>oldViewModel - the instance of the ViewModel for the old View</li>
  * <li>newViewModel  - the instance of the ViewModel for the new View</li>
  * </ul>
  *
  * @return {Object} switching implementation of the ModuleAnimation interface
  * @ojsignature [
  *    {target:"Type", for: "callback", value: "(param0: oj.ModuleAnimations.SwitcherCallBackParam) => oj.ModuleAnimations.Animations"},
  *    {target:"Type", for: "returns", value: "oj.ModuleElementAnimation"}
  *  ]
  * @export
  */
oj.ModuleAnimations.switcher = function(callback)
{
  /**
   * @constructor
   * @private
   */
  var AnimateProxy = function()
  {
    var _delegate;

    function _getDelegateInvoker(name)
    {
      var invoker = function(context)
      {
        return _delegate[name].call(_delegate, context);
      }
      return invoker;
    };

    var _self  = this;
    var _canAnimate = 'canAnimate';

    this[_canAnimate] = function(context)
    {
      // Get the 'delegate' animation
      var type = callback(context);
      _delegate = (type == null ? null : oj.ModuleAnimations[type]);
      if (!_delegate)
      {
        return false;
      }

      // Define the rest of the methods on the fly if we have a delegate
      var methods = ['prepareAnimation', 'animate'];
      for (var i=0; i<methods.length; i++)
      {
        var method = methods[i];
        _self[method] = _getDelegateInvoker(method);
      }

      return _getDelegateInvoker(_canAnimate)(context);
    };

  };

  return new AnimateProxy();
};

/**
 * @typedef {Object} oj.ModuleAnimations.SwitcherCallBackParam
 * @property {Element} node
 * @property {function() : any} valueAccessor
 * @property {boolean} isInitial
 * @property {any} oldViewModel
 * @property {any} newViewModel
 */
/**
 * All the available animation types supported in oj.ModuleAnimations
 * @typedef {Object} oj.ModuleAnimations.Animations
 * @ojvalue {string} "coverStart"
 * @ojvalue {string} "coverUp"
 * @ojvalue {string} "fade"
 * @ojvalue {string} "navChild"
 * @ojvalue {string} "navParent"
 * @ojvalue {string} "navSiblingEarlier"
 * @ojvalue {string} "navSiblingLater"
 * @ojvalue {string} "pushEnd"
 * @ojvalue {string} "pushStart"
 * @ojvalue {string} "revealDown"
 * @ojvalue {string} "revealEnd"
 * @ojvalue {string} "zoomIn"
 * @ojvalue {string} "zoomOut"
 */

/**
 * @ignore
 * @export
 * @class oj.NavigationListDndContext
 * @classdesc Drag and Drop Utils for ojNavigationList
 * @param {Object} navlist the NavigationList instance
 * @constructor
 */
if (typeof oj.ListViewDndContext != "undefined")
{
  oj.NavigationListDndContext = function (navlist) {
    oj.NavigationListDndContext.superclass.constructor.call(this, navlist);
  };
  // Subclass from oj.ListViewDndContext
  oj.Object.createSubclass(oj.NavigationListDndContext, oj.ListViewDndContext, "oj.NavigationListDndContext");

  /**
   * Gets the default drag affordance marker class
   * @return {string} affordance marker class
   * @override
   * @protected
   */
  oj.NavigationListDndContext.prototype.GetDragAffordanceClass = function ()
  {
    return "oj-tabbar-drag-handle";
  };
  
  /**
   * Gets the drag Image class
   * @return {string} drag image class
   * @override
   * @protected
   */
  oj.NavigationListDndContext.prototype.GetDragImageClass = function () {
    return "oj-tabbar-drag-image";
  };
  
  /**
   * Gets the drag Item class
   * @return {string} drag Item class
   * @override
   * @protected
   */
  oj.NavigationListDndContext.prototype.GetDragItemClass = function () {
    return "oj-tabbar-drag-item";
  };
  
  /**
   * Gets the cut command style class
   * @return {string} cut command style class
   * @override
   * @protected
   */
  oj.NavigationListDndContext.prototype.GetCutStyleClass = function () {
    return "oj-tabbar-cut";
  };
  
  /**
   * Gets the style class prefix
   * @return {string} style class prefix
   * @override
   * @protected
   */
  oj.NavigationListDndContext.prototype.GetCommandPrefix = function () {
    return "oj-tabbar-";
  };
  /**
   * Return true to drag current item.
   * @return {boolean} true to drag current item.
   * @override
   */
  oj.NavigationListDndContext.prototype.shouldDragCurrentItem = function () {
    return true;
  };

  /**
   * Override to return the Drag image width.
   * @return {string} width
   * @protected
   * @override
   */
  oj.NavigationListDndContext.prototype.GetDragImageWidth = function (item) {
    var edge = this.listview.GetOption('edge');
    return (edge === 'top' || edge === 'bottom') ? item.offsetWidth : oj.NavigationListDndContext.superclass.GetDragImageWidth.call(this, item);
  };

  /**
   * overide to return drag source type.
   * @return {string} drag source type.
   * @protected
   * @ignore
   * @override
   */
  oj.NavigationListDndContext.prototype.GetDragSourceType = function () {
    return "text/ojnavigationlist-dragsource-id";
  };

  /**
 * Returns payload object for reorder event. 
 * @param {Array} items, items to be moved. 
 * @param {string} position, the drop position relative to the reference item.
 * @param {Element} reference, the item where the moved items are drop on.
 * @returns {Object} payload object
 * @protected
 * @override
 */
  oj.NavigationListDndContext.prototype.CreateReorderPayload = function (items, position, reference) {
    return {
      'item': items[0],
      'position': position,
      'reference': reference
    };
    //ignore other events
  };

  /**
   * Override to verify whether item reordering enabled or not
   * @protected
   * @override
   */  
  oj.NavigationListDndContext.prototype.IsItemReOrdering = function()
  {
      var option = this.listview.GetOption('reorderable');
      return (option == "enabled");
  };

  /**
   * Returns the default data type. Need to be overriden by navlist
   * @return {string} the default data type
   * @protected
   */
  oj.NavigationListDndContext.prototype.GetDefaultDataType = function ()
  {
    return "text/ojnavigationlist-items-data";
  };

  /**
   * Create and sets the drag image into the dataTransfer object
   * @param {Event} nativeEvent  DOM event object 
   * @param {Array.<Element>} items array of row data
   * @protected
   */
  oj.NavigationListDndContext.prototype.SetDragItemImage = function (nativeEvent, items)
  {
    var target, dragImage, clone, left = 0, top = 0;
    target = nativeEvent.target;
    left = Math.max(0, nativeEvent.offsetX);
    top = Math.max(0, nativeEvent.offsetY);
    clone = $(items[0].cloneNode(true));
    clone.removeClass("oj-selected oj-focus oj-focus-highlight  oj-hover")
            .addClass("oj-drag");
    dragImage = $(document.createElement("div"));
    dragImage.get(0).className = this.listview.ojContext.element.get(0).className;
    dragImage.addClass(this.GetDragImageClass())
            .css({"width": this.GetDragImageWidth(items[0]),
              "height": items[0].offsetHeight,
            })
            .append(clone); //@HTMLUpdateOK
    $("body").append(dragImage); //@HTMLUpdateOK
    this.m_dragImage = dragImage;
    nativeEvent.dataTransfer.setDragImage(dragImage.get(0), left, top);
  };

  /**
   * Returns cut items. Navlist overrides this.
   * @param {Event} event jQuery event
   * @returns {Array} array of items to be moved
   */
  oj.NavigationListDndContext.prototype.GetCutItems = function (event)
  {
    var items = [];
    items.push(this.m_contextMenuItem && this.m_contextMenuItem.length > 0 ? 
                        this.m_contextMenuItem[0] : this._getActiveItem());
    return items;                    
  };

}
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojNBox
 * @augments oj.dvtBaseComponent
 * @since 1.1.0
 * @ojstatus preview
 * @ojrole application
 * @ojshortdesc An interactive data visualization (typically found in Human Capital Management applications) in which employees are grouped and compared across two dimensions.  Each dimension can be split into multiple ranges.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="nBoxOverview-section">
 *   JET NBox
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#nBoxOverview-section"></a>
 * </h3>
 *
 * <p>NBox elements are used in HCM (Human Capital Management) applications
 * to measure employees across two dimensions (e.g. potential and performance). Each dimension
 * can be split into multiple ranges, whose intersections result in distinct cells which
 * employees can be placed into.</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-n-box
 *   columns='[{"id":"low"}, {"id":"high"}]',
 *   rows='[{"id":"low"}, {"id":"high"}]',
 *   nodes='[{"column":"low", "row":"high", "label":"Employee 1"},
 *           {"column":"low", "row":"low", "label":"Employee 2"},
 *           {"column":"high", "row":"high", "label":"Employee 3"},
 *           {"column":"low", "row":"high", "label":"Employee 4"}]'>
 * &lt;/oj-n-box>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets. Alternate visualizations should
 *    be considered if identifying data changes is important, since all nodes will generally move and resize on any data
 *    change.
 * </p>
 *
 * <h4>Data Set Size</h4>
 * <p>Applications should avoid setting very large data densities on this element.
 *    Applications can enable progressive reveal of data through drilling or aggregate small nodes to reduce the
 *    displayed data set size.
 * </p>
 *
 * <h4>Styling</h4>
 * <p>Use the highest level property available. For example, consider setting styling properties on
 *    <code class="prettyprint">styleDefaults.nodeDefaults</code>, instead of styling properties
 *    on the individual nodes. The nbox can take advantage of these higher level properties to apply the style properties on
 *    containers, saving expensive DOM calls.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojNBox', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix : "oj",

  options: {
    /**
     * Specifies the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDataChange: "none",
    
    /**
     * Specifies the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDisplay: "none",

    /**
     * The content the cells will display. "auto" switches between nodes and cell counts based on available space. "counts" forces the NBox to always render cell counts.
     * @expose
     * @name cellContent
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "counts"
     * @ojvalue {string} "auto"
     * @default "auto"
     */
    cellContent: "auto",

    /**
     * Whether or not the cell maximize/de-maximize gestures are enabled.
     * @expose
     * @name cellMaximize
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "off"
     * @ojvalue {string} "on"
     * @default "on"
     */
    cellMaximize: "on",

    /**
     * The list of cells. Also accepts a Promise for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name cells
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @default null
     */
    cells: null,

    /**
     * The list of columns. Also accepts a Promise for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name columns
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @default null
     */
    columns: null,
    
    /**
     * The text for the title on the column edge.
     * @expose
     * @name columnsTitle
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @default ""
     */
    columnsTitle: "",
    
    /**
     * A function that returns custom text for the cell count labels (extra info displayed after primary labels). The function takes a dataContext argument, provided by the NBox, with the following properties: <ul>   <li>row: The row value of the cell.</li>    <li>column: The column value of the cell.</li>    <li>nodeCount: The number of non-hidden nodes in the cell.</li> <li>totalNodeCount: The number of non-hidden nodes in the NBox.</li> <li>highlightedNodeCount: The number of highlighted nodes in the cell.</li> </ul> The custom count label is shown by a cell when its showCount attribute is set "on" or "auto".
     * @expose
     * @ojshortdesc A function that returns custom text for the cell count labels.
     * @name countLabel
     * @memberof oj.ojNBox
     * @instance
     * @type {function(Object)|null}
     * @default null
     */
    countLabel: null,
    
    /**
     * Specifies how nodes should be grouped.
     * @expose
     * @name groupBehavior
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "acrossCells"
     * @ojvalue {string} "none"
     * @ojvalue {string} "withinCell"
     * @default "withinCell"
     */
    groupBehavior: "withinCell",
    
    /**
     * An array of attributes to style the group nodes with. Any attributes not listed will be ignored.
     * @expose
     * @name groupAttributes
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<string>}
     * @ojvalue {string} "color"
     * @ojvalue {string} "indicatorColor"
     * @ojvalue {string} "indicatorIconColor"
     * @ojvalue {string} "indicatorIconPattern"
     * @ojvalue {string} "indicatorIconShape"
     * @default ["color", "indicatorColor", "indicatorIconShape", "indicatorIconColor", "indicatorIconPattern"]
     */
    groupAttributes: ["color", "indicatorColor", "indicatorIconColor", "indicatorIconPattern", "indicatorIconShape"],
    
    /**
     * An array of category strings used for category filtering. Data items with a category in hiddenCategories will be filtered.
     * @expose
     * @name hiddenCategories
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    hiddenCategories: [],
    
    /**
     * Defines the behavior applied when hovering over data items.
     * @expose
     * @name hoverBehavior
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "dim"
     * @ojvalue {string} "none"
     * @default "none"
     */
    hoverBehavior: "none",

    /**
     * An array of category strings used for category highlighting. Data items with a category in highlightedCategories will be highlighted.
     * @expose
     * @name highlightedCategories
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    highlightedCategories: [],
    
    /**
     * The matching condition for the highlightedCategories property. By default, highlightMatch is 'all' and only items whose categories match all of the values specified in the highlightedCategories array will be highlighted. If highlightMatch is 'any', then items that match at least one of the highlightedCategories values will be highlighted.
     * @expose
     * @name highlightMatch
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "any"
     * @ojvalue {string} "all"
     * @default "all"
     */
    highlightMatch: "all",

    /**
     * Determines node label truncation behavior. Labels are always truncated if limited by container (e.g. cell, dialog) width. Optionally, NBox can further truncate node labels to increase the number of nodes visible to the user. "on" allows label truncation to increase number of visible nodes. "ifRequired" only allows truncation when limited by container width.
     * @expose
     * @name labelTruncation
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "ifRequired"
     * @ojvalue {string} "on"
     * @default "on"
     */
    labelTruncation: "on",
    
    /**
     * The id of the column to be maximized.
     * @expose
     * @name maximizedColumn
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @default ""
     * @ojwriteback
     */
    maximizedColumn: "",

    /**
     * The id of the row to be maximized.
     * @expose
     * @name maximizedRow
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @default ""
     * @ojwriteback
     */
    maximizedRow: "",

    /**
     * The list of nodes. Also accepts a Promise for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name nodes
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @default null
     */
    nodes: null,

    /**
     * The color for the "other" group nodes which aggregate any group nodes that fall below the otherThreshold, if specified.  The default value varies based on theme.
     * @expose
     * @ojshortdesc The color for the "other" group nodes.
     * @name otherColor
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     */
    otherColor: "#636363",

    /**
     * A percentage (0-1) of the nodes collection size that determines the value beneath which any groups will be aggregated into an "other" node.
     * @expose
     * @ojshortdesc The threshold for aggregating nodes into an "other" group node.
     * @name otherThreshold
     * @memberof oj.ojNBox
     * @instance
     * @type {number}
     * @default 0
     */
    otherThreshold: 0,

    /**
     * The list of rows. Also accepts a Promise for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name rows
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @default null
     */
    rows: null,

    /**
     * The text for the title on the row edge.
     * @expose
     * @name rowsTitle
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @default ""
     */
    rowsTitle: "",
    
    /**
     * An array containing the ids of the selected nodes.
     * @expose
     * @name selection
     * @memberof oj.ojNBox
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    selection: [],

    /**
     * Specifies the selection mode.
     * @expose
     * @name selectionMode
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @default "multiple"
     */
    selectionMode: "multiple",

    /**
     * An object defining the style defaults for this NBox.
     * @expose
     * @name styleDefaults
     * @memberof oj.ojNBox
     * @instance
     * @type {Object}
     * @default {"cellDefaults": {"labelHalign": "start", "showCount": "auto"}, "hoverBehaviorDelay": 200, "nodeDefaults":{"iconDefaults":{"color":"", "height":0, "opacity":1, "pattern":"none", "shape":"square", "source":"", "width":0}, "indicatorColor":"", "indicatorIconDefaults":{"color":"", "opacity":1, "pattern":"none", "shape":"square", "source":""}}}
     */
    styleDefaults: {
      /**
       * The duration of the animations in milliseconds.  The default value comes from the CSS and varies based on theme.
       * @expose
       * @name styleDefaults.animationDuration
       * @memberof! oj.ojNBox
       * @instance
       * @type {number}
       * @ojunits milliseconds
       */
      animationDuration: undefined,

      /**
       * An object defining the style defaults for cells.
       * @expose
       * @name styleDefaults.cellDefaults
       * @memberof! oj.ojNBox
       * @instance
       * @type {Object}
       * @default {"labelHalign": "start", "showCount": "auto"}
       */
      cellDefaults: {
        /**
         * The halign value for the cell label.
         * @expose
         * @name styleDefaults.cellDefaults.labelHalign
         * @memberof! oj.ojNBox
         * @instance
         * @type {string}
         * @ojvalue {string} "center"
         * @ojvalue {string} "end"
         * @ojvalue {string} "start"
         * @default "start"
         */
        labelHalign: "start",

        /**
         * The CSS style object defining the style of the cell labels.  The default value comes from the CSS and varies based on theme.
         * @expose
         * @name styleDefaults.cellDefaults.labelStyle
         * @memberof! oj.ojNBox
         * @instance
         * @type {Object}
         */
        labelStyle: undefined,
                
        /**
         * Determines when to display the cell count label (extra info displayed after primary cell label). "off" never show the count label. "on" always show the count label. Show countLabel value if specified, otherwise use a simple node count. "auto" show the count label if countLabel attribute is defined.
         * @expose
         * @ojshortdesc Determines when to display the cell count label.
         * @name styleDefaults.cellDefaults.showCount
         * @memberof! oj.ojNBox
         * @instance
         * @type {string}
         * @ojvalue {string} "on"
         * @ojvalue {string} "off"
         * @ojvalue {string} "auto"
         * @default "auto"
         */
        showCount: "auto"
      },      
      
      /**
       * The CSS style object defining the style of the column labels.  The default value comes from the CSS and varies based on theme.
       * @expose
       * @name styleDefaults.columnLabelStyle
       * @memberof! oj.ojNBox
       * @instance
       * @type {Object}
       */
      columnLabelStyle: undefined,
      
      /**
       * The CSS style object defining the style of the columns title.  The default value comes from the CSS and varies based on theme.
       * @expose
       * @name styleDefaults.columnsTitleStyle
       * @memberof! oj.ojNBox
       * @instance
       * @type {Object}
       */
      columnsTitleStyle: undefined,
      
      /**
       * Specifies initial hover delay in ms for highlighting data items.
       * @expose
       * @name styleDefaults.hoverBehaviorDelay
       * @memberof! oj.ojNBox
       * @instance
       * @type {number}
       * @default 200
       * @ojunits milliseconds
       */
      hoverBehaviorDelay: 200,
      
      /**
       * An object defining the style defaults for nodes.
       * @expose
       * @name styleDefaults.nodeDefaults
       * @memberof! oj.ojNBox
       * @instance
       * @type {Object}
       * @default {"iconDefaults":{"color":"", "height":0, "opacity":1, "pattern":"none", "shape":"square", "source":"", "width":0}, "indicatorColor":"", "indicatorIconDefaults":{"color":"", "opacity":1, "pattern":"none", "shape":"square", "source":""}}
       */
      nodeDefaults: {
        /**
         * The default color of the node borders.  The default value varies based on theme.
         * @expose
         * @name styleDefaults.nodeDefaults.borderColor
         * @memberof! oj.ojNBox
         * @instance
         * @type {string}
         */
        borderColor: "",
      
        /**
         * The default width of the node borders.  The default value varies based on theme.
         * @expose
         * @name styleDefaults.nodeDefaults.borderWidth
         * @memberof! oj.ojNBox
         * @instance
         * @type {number}
         * @ojunits pixels
         */
        borderWidth: 0,

        /**
         * The default background color of the nodes.  The default value comes from the CSS and varies based on theme.
         * @expose
         * @name styleDefaults.nodeDefaults.color
         * @memberof! oj.ojNBox
         * @instance
         * @type {string}
         */
        color: undefined,

        /**
         * An object defining the style defaults for the node icons.
         * @expose
         * @name styleDefaults.nodeDefaults.iconDefaults
         * @memberof! oj.ojNBox
         * @instance
         * @type {Object}
         * @default {"color":"", "height":0, "opacity":1, "pattern":"none", "shape":"square", "source":"", "width":0}
         */
        iconDefaults: {
          /**
           * The default border color of the node icons.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.borderColor
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           */
          borderColor: "#000000",
          
          /**
           * The default border radius of the node icons. CSS border-radius values accepted. Note that non-% values (including unitless) get interpreted as 'px'.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.borderRadius
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           */
          borderRadius: 0,
          
          /**
           * The default border width of the node icons.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.borderWidth
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           */
          borderWidth: 0,
          
          /**
           * The default fill color of the node icons.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.color
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @default ""
           */
          color: "",
          
          /**
           * The default height of the node icons. If the value is 0, the height will be automatically based on the remaining node contents.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.height
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           * @default 0
           * @ojunits pixels
           */
          height: 0,
          
          /**
           * The default opacity of the node icons.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.opacity
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           * @default 1
           */
          opacity: 1,
          
          /**
           * The default fill pattern of the node icons.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.pattern
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @ojvalue {string} "smallChecker"
           * @ojvalue {string} "smallCrosshatch"
           * @ojvalue {string} "smallDiagonalLeft"
           * @ojvalue {string} "smallDiagonalRight"
           * @ojvalue {string} "smallDiamond"
           * @ojvalue {string} "smallTriangle"
           * @ojvalue {string} "largeChecker"
           * @ojvalue {string} "largeCrosshatch"
           * @ojvalue {string} "largeDiagonalLeft"
           * @ojvalue {string} "largeDiagonalRight"
           * @ojvalue {string} "largeDiamond"
           * @ojvalue {string} "largeTriangle"
           * @ojvalue {string} "none"
           * @default "none"
           */
          pattern: "none",
                    
          /**
           * The default shape of the node icons. Can take the name of a built-in shape or the svg path commands for a custom shape.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.shape
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @ojvalue {string} "circle"
           * @ojvalue {string} "ellipse"
           * @ojvalue {string} "square"
           * @ojvalue {string} "plus"
           * @ojvalue {string} "diamond"
           * @ojvalue {string} "triangleUp"
           * @ojvalue {string} "triangleDown"
           * @ojvalue {string} "human"
           * @ojvalue {string} "rectangle"
           * @ojvalue {string} "star"
           * @default "square"
           */
          shape: "square",
          
          /**
           * The URL of an image to display by default for the node icons.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.source
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @default ""
           */
          source: "",
          
          /**
           * The default width of the node icons.  If the value is 0, the width will be automatically based on the remaining node contents.
           * @expose
           * @name styleDefaults.nodeDefaults.iconDefaults.width
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           * @default 0
           */
          width: 0
        },
        
        /**
         * The default background color of the node indicator sections.
         * @expose
         * @name styleDefaults.nodeDefaults.indicatorColor
         * @memberof! oj.ojNBox
         * @instance
         * @type {string}
         * @default ""
         */
        indicatorColor: "",
        
        /**
         * An object defining the style defaults for the node indicator icons.
         * @expose
         * @name styleDefaults.nodeDefaults.indicatorIconDefaults
         * @memberof! oj.ojNBox
         * @instance
         * @type {Object}
         * @default {"color":"", "opacity":1, "pattern":"none", "shape":"square", "source":""}
         */
        indicatorIconDefaults: {
          /**
           * The default border color of the node indicator icons.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.borderColor
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           */
          borderColor: "#000000",
          
          /**
           * The default border radius of the node indicator icons. CSS border-radius values accepted. Note that non-% values (including unitless) get interpreted as 'px'.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.borderRadius
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           */
          borderRadius: 0,
          
          /**
           * The default border width of the node indicator icons.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.borderWidth
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           */
          borderWidth: 0,
          
          /**
           * The default fill color of the node indicator icons.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.color
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @default ""
           */
          color: "",

          /**
           * The default height of the node indicator icons.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.height
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           * @ojunits pixels
           */
          height: 10,
          
          /**
           * The default opacity of the node indicator icons.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.opacity
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           * @default 1
           */
          opacity: 1,

          /**
           * The default fill pattern of the node indicator icons.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.pattern
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @ojvalue {string} "smallChecker"
           * @ojvalue {string} "smallCrosshatch"
           * @ojvalue {string} "smallDiagonalLeft"
           * @ojvalue {string} "smallDiagonalRight"
           * @ojvalue {string} "smallDiamond"
           * @ojvalue {string} "smallTriangle"
           * @ojvalue {string} "largeChecker"
           * @ojvalue {string} "largeCrosshatch"
           * @ojvalue {string} "largeDiagonalLeft"
           * @ojvalue {string} "largeDiagonalRight"
           * @ojvalue {string} "largeDiamond"
           * @ojvalue {string} "largeTriangle"
           * @ojvalue {string} "none"
           * @default "none"
           */
          pattern: "none",
          
          /**
           * The default shape of the node indicator icons. Can take the name of a built-in shape or the svg path commands for a custom shape.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.shape
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @ojvalue {string} "circle"
           * @ojvalue {string} "ellipse"
           * @ojvalue {string} "square"
           * @ojvalue {string} "plus"
           * @ojvalue {string} "diamond"
           * @ojvalue {string} "triangleUp"
           * @ojvalue {string} "triangleDown"
           * @ojvalue {string} "human"
           * @ojvalue {string} "rectangle"
           * @ojvalue {string} "star"
           * @default "square"
           */
          shape: "square",
          
          /**
           * The URL of an image to display by default for the node indicator icons.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.source
           * @memberof! oj.ojNBox
           * @instance
           * @type {string}
           * @default null
           */
          source: "",
          
          /**
           * The default width of the node indicator icons.  The default value varies based on theme.
           * @expose
           * @name styleDefaults.nodeDefaults.indicatorIconDefaults.width
           * @memberof! oj.ojNBox
           * @instance
           * @type {number}
           * @ojunits pixels
           */
          width: 10        
        },
        
        /**
         * The CSS style object defining the style of the node labels.  The default value comes from the CSS and varies based on theme.
         * @expose
         * @name styleDefaults.nodeDefaults.labelStyle
         * @memberof! oj.ojNBox
         * @instance
         * @type {Object}
         */
        labelStyle: undefined,
        
        /**
         * The CSS style object defining the style of the node secondary labels.  The default value comes from the CSS and varies based on theme.
         * @expose
         * @name styleDefaults.nodeDefaults.secondaryLabelStyle
         * @memberof! oj.ojNBox
         * @instance
         * @type {Object}
         */
        secondaryLabelStyle: undefined       
      },
      
      /**
       * The CSS style object defining the style of the row labels.  The default value comes from the CSS and varies based on theme.
       * @expose
       * @name styleDefaults.rowLabelStyle
       * @memberof! oj.ojNBox
       * @instance
       * @type {Object}
       */
      rowLabelStyle: undefined,
      
      /**
       * The CSS style object defining the style of the rows title.  The default value comes from the CSS and varies based on theme.
       * @expose
       * @name styleDefaults.rowsTitleStyle
       * @memberof! oj.ojNBox
       * @instance
       * @type {Object}
       * @default null
       */
      rowsTitleStyle: undefined  
    },
    
    
    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojNBox
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     */
    tooltip: {
      /**
       * A function that returns a custom tooltip for the NBox nodes. The function takes a dataContext argument,
       * provided by the NBox, with the following properties:
       * <ul>
       *   <li>parentElement: The tooltip element. The function can directly modify or append content to this element.</li>
       *   <li>id: The id of the hovered node.</li>
       *   <li>label: The label of the hovered node.</li>
       *   <li>secondaryLabel: The secondary label of the hovered node.</li>
       *   <li>color: The color of the hovered node.</li>
       *   <li>indicatorColor: The indicator color of the hovered node.</li>
       *   <li>row: The id of the row containing the hovered node.</li>
       *   <li>column: The id of the column containing the hovered node.</li>
       *   <li>componentElement: The NBox element.</li>
       * </ul>
       *  The function should return an Object that contains only one of the two properties:
       *  <ul>
       *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
       *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
       *  </ul>
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojNBox
       * @instance
       * @type {function(Object)|null}
       * @default null
       */
      renderer: null
    },
 
    /**
     * Data visualizations require a press and hold delay before triggering tooltips and rollover effects on mobile devices to avoid interfering with page panning, but these hold delays can make applications seem slower and less responsive. For a better user experience, the application can remove the touch and hold delay when data visualizations are used within a non scrolling container or if there is sufficient space outside of the visualization for panning. If touchResponse is touchStart the element will instantly trigger the touch gesture and consume the page pan events if the element does not require an internal feature that requires a touch start gesture like scrolling. If touchResponse is auto, the element will behave like touchStart if it determines that it is not rendered within scrolling content and if panning is not available for those elements that support the feature. 
     * @expose
     * @name touchResponse
     * @memberof oj.ojNBox
     * @instance
     * @type {string}
     * @ojvalue {string} "touchStart"
     * @ojvalue {string} "auto"
     * @default "auto"
     */
    touchResponse: "auto"    
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    return dvt.NBox.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-nbox-cell') {
      // cell[row,column]
      subId = 'cell[' + locator['row'] + ',' + locator['column'] + ']';
    }
    else if(subId == 'oj-nbox-dialog') {
      subId = 'dialog';
    }
    else if(subId == 'oj-nbox-dialog-close-button') {
      subId = 'dialog#closeButton';
    }
    else if(subId == 'oj-nbox-group-node') {
      // groupNode[groupCategory] or cell[row,column]#groupNode[groupCategory]
      if(locator['row'] && locator['column'])
        subId = 'cell[' + locator['row'] + ',' + locator['column'] + ']#groupNode[';
      else
        subId = 'groupNode[';

      subId += locator['groupCategory'] +']';
    }
    else if(subId == 'oj-nbox-node') {
      var index;
      subId = '';

      var id = locator['id'];
      var auto = this._component.getAutomation();
      index = auto.getNodeIndexFromId(id);
      subId += 'node[' + index + ']';
    }
    else if(subId == 'oj-nbox-overflow') {
      // cell[row,col]#overflow
      subId = 'cell[' + locator['row'] + ',' + locator['column'] + ']#overflow';
    }
    else if(subId == 'oj-nbox-tooltip') {
      subId = 'tooltip';
    }

    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if (subId.indexOf('node') == 0) {
      locator['subId'] = 'oj-nbox-node';

      var index = this._GetFirstIndex(subId);
      var auto = this._component.getAutomation();
      locator['id'] = auto.getNodeIdFromIndex(index);
    }
    else if(subId.indexOf('cell') == 0) {
      // cell[row,column] or cell[row,column]#groupNode[groupCategory] or cell[row,column]#node[index]
      var cellIds = this._GetFirstBracketedString(subId);
      var commaIndex = cellIds.indexOf(',');
      locator['row'] = cellIds.substring(0, commaIndex);
      locator['column'] = cellIds.substring(commaIndex + 1);

      var poundIndex = subId.indexOf('#');
      if(subId.indexOf('#groupNode') > 0) {
        locator['subId'] = 'oj-nbox-group-node';
        locator['groupCategory'] = this._GetFirstBracketedString(subId.substring(poundIndex));
      }
      else if(subId.indexOf('#overflow') > 0) {
        locator['subId'] = 'oj-nbox-overflow';
      }
      else
        locator['subId'] = 'oj-nbox-cell';
    }
    else if(subId.indexOf('dialog') == 0) {
      if(subId.indexOf('#closeButton') > 0) {
        // dialog#closeButton
        locator['subId'] = 'oj-nbox-dialog-close-button';
      }
      else
        locator['subId'] = 'oj-nbox-dialog';
    }
    else if(subId.indexOf('groupNode') == 0) {
      // groupNode[groupCategory] or cell[row,column]#groupNode[groupCategory]
      locator['subId'] = 'oj-nbox-group-node';
      locator['groupCategory'] = this._GetFirstBracketedString(subId);
    }
    else if(subId == 'tooltip') {
      locator['subId'] = 'oj-nbox-tooltip';
    }

    return locator;
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-nbox');
    return styleClasses;
  },

  //** @inheritdoc */
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-dvtbase oj-nbox'] = {'path': 'styleDefaults/animationDuration', 'property': 'ANIM_DUR'};
    styleClasses['oj-nbox-columns-title'] = {'path': 'styleDefaults/columnsTitleStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-rows-title'] = {'path': 'styleDefaults/rowsTitleStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-column-label'] = {'path': 'styleDefaults/columnLabelStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-row-label'] = {'path': 'styleDefaults/rowLabelStyle', 'property': 'TEXT'};

    styleClasses['oj-nbox-cell'] = {'path': 'styleDefaults/cellDefaults/_style', 'property': 'BACKGROUND'};
    styleClasses['oj-nbox-cell oj-minimized'] = {'path': 'styleDefaults/cellDefaults/minimizedStyle', 'property': 'BACKGROUND'};
    styleClasses['oj-nbox-cell oj-maximized'] = {'path': 'styleDefaults/cellDefaults/maximizedStyle', 'property': 'BACKGROUND'};

    styleClasses['oj-nbox-cell-label'] = {'path': 'styleDefaults/cellDefaults/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-cell-countlabel'] = {'path': 'styleDefaults/cellDefaults/bodyCountLabelStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-cell-countlabel oj-nbox-cell-header'] = {'path': 'styleDefaults/cellDefaults/countLabelStyle', 'property': 'TEXT'};

    styleClasses['oj-nbox-node'] = {'path': 'styleDefaults/nodeDefaults/color', 'property': 'background-color'};
    styleClasses['oj-nbox-node oj-hover'] = {'path': 'styleDefaults/nodeDefaults/hoverColor', 'property': 'border-color'};
    styleClasses['oj-nbox-node oj-selected'] = {'path': 'styleDefaults/nodeDefaults/selectionColor', 'property': 'border-color'};

    styleClasses['oj-nbox-node-label'] = {'path': 'styleDefaults/nodeDefaults/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-node-secondarylabel'] = {'path': 'styleDefaults/nodeDefaults/secondaryLabelStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-node-categorylabel'] = {'path': 'styleDefaults/_categoryNodeDefaults/labelStyle', 'property': 'TEXT'};

    styleClasses['oj-nbox-dialog'] = [{'path': 'styleDefaults/_drawerDefaults/background', 'property': 'background-color'},
                                      {'path': 'styleDefaults/_drawerDefaults/borderColor', 'property': 'border-color'}];
    styleClasses['oj-nbox-dialog-label'] = {'path': 'styleDefaults/_drawerDefaults/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-nbox-dialog-countlabel'] = {'path': 'styleDefaults/_drawerDefaults/countLabelStyle', 'property': 'TEXT'};

    return styleClasses;
  },

  //** @inheritdoc */
  _GetEventTypes : function() {
    return ['optionChange'];
  },

  //** @inheritdoc */
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarantees a new map is returned
    var ret = this._super();
    ret['DvtUtilBundle.NBOX'] = translations['componentName'];
    ret['DvtNBoxBundle.HIGHLIGHTED_COUNT'] = translations['highlightedCount'];
    ret['DvtNBoxBundle.OTHER'] = translations['labelOther'];
    ret['DvtNBoxBundle.GROUP_NODE'] = translations['labelGroup'];
    ret['DvtNBoxBundle.SIZE'] = translations['labelSize'];
    ret['DvtNBoxBundle.ADDITIONAL_DATA'] = translations['labelAdditionalData'];
    return ret;
  },

  //** @inheritdoc */
  _HandleEvent : function(event) {
    var type = event['type'];
    if (type === 'adfPropertyChange') {
      var properties = event['properties'];
      for (var key in properties) {
        var value = properties[key];
        if (key == '_drawer')
          this.options[key] = value ? {'id': value} : null;
        else if (key == 'maximizedRow' || key == 'maximizedColumn')
          this._UserOptionChange(key, value);
      }
    }
    else {
      this._super(event);
    }
  },

  //** @inheritdoc */
  _LoadResources: function() {
    // Ensure the resources object exists
    if (this.options['_resources'] == null)
      this.options['_resources'] = {};

    var resources = this.options['_resources'];
    resources['overflow_dwn'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/nbox/overflow_dwn.png'), 'width':34, 'height':9};
    resources['overflow_ovr'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/nbox/overflow_ovr.png'), 'width':34, 'height':9};
    resources['overflow_ena'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/nbox/overflow_ena.png'), 'width':34, 'height':9};
    resources['overflow_dis'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/nbox/overflow_dis.png'), 'width':34, 'height':9};
    resources['close_dwn'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/nbox/close_dwn.png'), 'width':16, 'height':16};
    resources['close_ovr'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/nbox/close_ovr.png'), 'width':16, 'height':16};
    resources['close_ena'] = {src: oj.Config.getResourceUrl('resources/internal-deps/dvt/nbox/close_ena.png'), 'width':16, 'height':16};
  },

  //** @inheritdoc */
  getNodeBySubId : function(locator) {
    return this._super(locator);
  },

  //** @inheritdoc */
  getSubIdByNode:function(node) {
    return this._super(node);
  },

  /**
   * Get the NBox rows title.
   * @return {String} NBox rows title.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getRowsTitle: function() {
    var auto = this._component.getAutomation();
    return auto.getData('rowsTitle');
  },

  /**
   * Get the NBox row count.
   * @return {Number} NBox row count.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getRowCount: function() {
    var auto = this._component.getAutomation();
    return auto.getData('rowCount');
  },

  /**
   * Returns an object with the following properties for automation testing verification of the NBox row at the
   * specified value.
   *
   * @param {string} rowValue The id of the row.
   * @property {string} label The label of the row.
   * @return {Object|null} An object containing properties for the row, or null if none exists.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getRow: function(rowValue) {
    var auto = this._component.getAutomation();
    return auto.getData('row', rowValue);
  },


  /**
   * Get the NBox columns title.
   * @return {String} NBox columns title.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getColumnsTitle: function() {
    var auto = this._component.getAutomation();
    return auto.getData('columnsTitle');
  },

  /**
   * Get the NBox column count.
   * @return {Number} NBox column count.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getColumnCount: function() {
    var auto = this._component.getAutomation();
    return auto.getData('columnCount');
  },

  /**
   * Returns an object with the following properties for automation testing verification of the NBox column at the
   * specified value.
   *
   * @param {string} columnValue The id of the column.
   * @property {string} label The label of the column.
   * @return {Object|null} An object containing properties for the column, or null if none exists.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getColumn: function(columnValue) {
    var auto = this._component.getAutomation();
    return auto.getData('column', columnValue);
  },

  /**
   * Returns an object with the following properties for automation testing verification of the NBox cell at the
   * specified row and column values.
   *
   * @param {string} rowValue The id of the containing row.
   * @param {string} columnValue The id of the containing column.
   * @property {string} background The background of the cell.
   * @property {string} label The label of the cell.
   * @property {Function(string)} getGroupNode A function taking a group category string and returning the corresponding group node.
   * @property {string} getGroupNode.color The color of the group node.
   * @property {string} getGroupNode.indicatorColor The color of the group node indicator section.
   * @property {Object} getGroupNode.indicatorIcon The indicator marker for the group node.
   * @property {string} getGroupNode.indicatorIcon.color The color of the indicator marker.
   * @property {string} getGroupNode.indicatorIcon.shape The shape of the indicator marker.
   * @property {boolean} getGroupNode.selected Whether or not the group node is selected.
   * @property {number} getGroupNode.size The number of nodes the group node represents.
   * @property {string} getGroupNode.tooltip The tooltip of the group node.
   * @property {Function(number)} getNode A function taking the node index that returns an object with properties for the specified node, or null if none exists.
   * @property {string} getNode.color The color of the node.
   * @property {Object} getNode.icon The icon marker for the node.
   * @property {string} getNode.icon.color The color of the icon marker.
   * @property {string} getNode.icon.shape The shape of the icon marker.
   * @property {string} getNode.indicatorColor The color of the node indicator section.
   * @property {Object} getNode.indicatorIcon The indicator marker for the node.
   * @property {string} getNode.indicatorIcon.color The color of the indicator marker.
   * @property {string} getNode.indicatorIcon.shape The shape of the indicator marker.
   * @property {string} getNode.label The label of the node.
   * @property {string} getNode.secondaryLabel The secondary label of the node.
   * @property {boolean} getNode.selected Whether or not the node is selected.
   * @property {string} getNode.tooltip The tooltip of the node.
   * @property {Function} getNodeCount A function that returns the number of nodes in the cell.
   * @return {Object|null} An object containing properties for the cell, or null if none exists.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getCell: function(rowValue, columnValue) {
    var auto = this._component.getAutomation();
    var ret = auto.getCell(rowValue, columnValue);
    if (ret) {
      ret['getGroupNode'] = function(groupMap) {
        return auto.getCellGroupNode(ret, groupMap);
      }
      ret['getNode'] = function(nodeIndex) {
        return auto.getCellNode(ret, nodeIndex);
      }
    }
    return ret;
  },

  /**
   * Get the NBox group behavior.
   * @return {String} group behavior The group behavior of the NBox ('withinCell', 'acrossCells', 'none').
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getGroupBehavior: function() {
    var auto = this._component.getAutomation();
    return auto.getData('groupBehavior');
  },

  /**
   * Returns an object with the following properties for automation testing verification of the NBox group node with the
   * specified group category string.
   *
   * @param {String} groupCategory A string corresponding to the groupCategory value of the nodes represented by this group node.
   * @property {string} color The color of the group node.
   * @property {string} indicatorColor The indicator color of the group node.
   * @property {Object} indicatorIcon The indicator marker for the group node, or null if none exists.
   * @property {string} indicatorIcon.color The color of the indicator marker.
   * @property {string} indicatorIcon.shape The shape of the indicator marker.
   * @property {boolean} selected Whether or not the group node is selected.
   * @property {number} size The number of nodes the group node represents.
   * @property {string} tooltip The tooltip of the group node.
   * @return {Object|null} An object containing properties for the group node, or null if none exists.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getGroupNode: function(groupCategory) {
    var auto = this._component.getAutomation();
    return auto.getGroupNode(groupCategory);
  },

  /**
   * Returns an object with the following properties for automation testing verification of the currently active NBox
   * dialog.
   *
   * @property {string} label The label of the dialog.
   * @property {Function(number)} getNode A function taking the node index that returns an object with properties for the specified node, or null if none exists.
   * @property {boolean} getNode.selected
   * @property {string} getNode.color The color of the node.
   * @property {Object} getNode.icon The icon marker for the node, or null if none exists.
   * @property {string} getNode.icon.color The color of the icon marker.
   * @property {string} getNode.icon.shape The shape of the icon marker.
   * @property {string} getNode.indicatorColor The indicator color of the node.
   * @property {Object} getNode.indicatorIcon The indicator icon for the node, or null if none exists.
   * @property {string} getNode.indicatorIcon.color The color of the indicator icon.
   * @property {string} getNode.indicatorIcon.shape The shape of the indicator icon.
   * @property {string} getNode.label The label of the node.
   * @property {string} getNode.secondaryLabel The secondary label of the node.
   * @property {string} getNode.tooltip The tooltip of the node.
   * @property {Function} getNodeCount A function that returns the number of nodes in the cell.
   * @return {Object|null} An object containing properties for the dialog, or null if none exists.
   * @expose
   * @memberof oj.ojNBox
   * @instance
   */
  getDialog: function() {
    var auto = this._component.getAutomation();
    var ret = auto.getDialog();
    if (ret) {
      ret['getNode'] = function(nodeIndex) {
        return auto.getDialogNode(nodeIndex);
      }
    }
    return ret;
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojNBox
   */
  getContextByNode: function(node) {
    // context objects are documented with @ojnodecontext
    var context = this.getSubIdByNode(node);
    if (context && context['subId'] !== 'oj-nbox-tooltip' &&
                   context['subId'] !== 'oj-nbox-dialog-close-button' &&
                   context['subId'] !== 'oj-nbox-overflow')
      return context;

    return null;
  },

  //** @inheritdoc */
  _GetComponentDeferredDataPaths : function() {
    return {'root': ['cells', 'rows', 'columns', 'nodes']};
  }

});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Cell</td>
 *       <td><kbd>Double Tap</kbd></td>
 *       <td>Maximize/restore cell.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="3">Node</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *     <tr>
 *       <td>Overflow Indicator</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Maximize cell.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="4">Group Node</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Double Tap</kbd></td>
 *       <td>Open group node dialog.</td>
 *     </tr>
  *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *     <tr>
 *       <td>Dialog</td>
 *       <td><kbd>Double Tap</kbd></td>
 *       <td>Close dialog.</td>
 *     </tr>
 *     <tr>
 *       <td>Close Button</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Close dialog.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojNBox
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Maximize focused cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Open dialog for focused group node.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Escape</kbd></td>
 *       <td>Restore maximized cell.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Escape</kbd></td>
 *       <td>Close dialog.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>[</kbd></td>
 *       <td>Move to first node in cell or dialog</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>]</kbd></td>
 *       <td>Move to containing cell or dialog</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Left Arrow</kbd></td>
 *       <td>When cell focused, move to nearest cell to the left.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Right Arrow</kbd></td>
 *       <td>When cell focused, move to nearest cell to the right.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Up Arrow</kbd></td>
 *       <td>When cell focused, move to nearest cell above.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Down Arrow</kbd></td>
 *       <td>When cell focused, move to nearest cell below.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Left Arrow or Up Arrow</kbd></td>
 *       <td>When node focused, move to previous node.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Right Arrow or Down Arrow</kbd></td>
 *       <td>When node focused, move to previous node.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Left Arrow or Ctrl + Up Arrow</kbd></td>
 *       <td>When node focused, move to previous node but do not select.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Right Arrow or Ctrl + Down Arrow</kbd></td>
 *       <td>When node focused, move to previous node but do not select.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Space</kbd></td>
 *       <td>Select or Unselect focused node.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Left Arrow or Shift + Up Arrow</kbd></td>
 *       <td>Move focus and multi-select previous node.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Right Arrow or Down + Up Arrow</kbd></td>
 *       <td>Move focus and multi-select next node.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojNBox
 */

/**
 * The id of the column containing this cell.
 * @expose
 * @name cells[].column
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The id of the row containing this cell.
 * @expose
 * @name cells[].row
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The text for the cell label.
 * @expose
 * @name cells[].label
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The halign value for the cell label.
 * @expose
 * @name cells[].labelHalign
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @ojvalue {string} "center"
 * @ojvalue {string} "end"
 * @ojvalue {string} "start"
 * @default <code class="prettyprint">"start"</code>
 */
/**
 * The CSS style object defining the style of the cell label.
 * @expose
 * @name cells[].labelStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The description of this cell. This is used for accessibility.
 * @expose
 * @name cells[].shortDesc
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Determines when to display the cell count label (extra info displayed after primary cell label). "off" never show the count label. "on" always show the count label. Show countLabel value if specified, otherwise use a simple node count. "auto" show the count label if countLabel attribute is defined.
 * @expose
 * @name cells[].showCount
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The CSS style string/object for this cell. Used for customizing the cell background and border.
 * @ignore
 * @name cells[].style
 * @memberof! oj.ojNBox
 * @instance
 * @type {string|Object}
 * @default null
 */
/**
 * The CSS style class for this cell. Used for customizing the cell background and border.
 * @ignore
 * @name cells[].className
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style string/object for this cell. Used for customizing the maximized cell background and border.
 * @ignore
 * @name cells[].maximizedStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {string|Object}
 * @default null
 */
/**
 * The CSS style class for this cell. Used for customizing the maximized cell background and border.
 * @ignore
 * @name cells[].maximizedClassName
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style string/object for this cell. Used for customizing the minimized cell background and border.
 * @ignore
 * @name cells[].minimizedStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {string|Object}
 * @default null
 */
/**
 * The CSS style class for this cell. Used for customizing the minimized cell background and border.
 * @ignore
 * @name cells[].minimizedClassName
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object for this cell. Used for customizing the cell background and border.
 * @expose
 * @name cells[].svgStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class for this cell. Used for customizing the cell background and border.
 * @expose
 * @name cells[].svgClassName
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object for this cell. Used for customizing the maximized cell background and border.
 * @expose
 * @name cells[].maximizedSvgStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class for this cell. Used for customizing the maximized cell background and border.
 * @expose
 * @name cells[].maximizedSvgClassName
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object for this cell. Used for customizing the minimized cell background and border.
 * @expose
 * @name cells[].minimizedSvgStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class for this cell. Used for customizing the minimized cell background and border.
 * @expose
 * @name cells[].minimizedSvgClassName
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */

/**
 * The id of the column. Used to identify this column.
 * @expose
 * @name columns[].id
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The text for the column label.
 * @expose
 * @name columns[].label
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object defining the style of the column label.
 * @expose
 * @name columns[].labelStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */


/**
 * The id of this node.
 * @expose
 * @name nodes[].id
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The column id for this node.
 * @expose
 * @name nodes[].column
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The row id for this node.
 * @expose
 * @name nodes[].row
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The color of the node border.
 * @expose
 * @name nodes[].borderColor
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of the node border.
 * @expose
 * @name nodes[].borderWidth
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The text for the node label.
 * @expose
 * @name nodes[].label
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The text for the secondary node label.
 * @expose
 * @name nodes[].secondaryLabel
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The description of this node. This is used for accessibility and also for customizing the tooltip text.
 * @expose
 * @name nodes[].shortDesc
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional array of additional category strings corresponding to this data item. This enables highlighting and filtering of individual data items through interactions with other visualization elements. Defaults to node's id if unspecified.
 * @expose
 * @name nodes[].categories
 * @memberof! oj.ojNBox
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The group category this node belongs to. Nodes with the same groupCategory will be grouped together.
 * @expose
 * @name nodes[].groupCategory
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object defining the style of this node.
 * @ignore
 * @name nodes[].style
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class defining the style of this node.
 * @ignore
 * @name nodes[].className
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
 /**
  * The CSS style object defining the style of this node.
  * @expose
  * @name nodes[].svgStyle
  * @memberof! oj.ojNBox
  * @instance
  * @type {Object}
  * @default null
  */
 /**
  * The CSS style class defining the style of this node.
  * @expose
  * @name nodes[].svgClassName
  * @memberof! oj.ojNBox
  * @instance
  * @type {string}
  * @default null
  */
/**
 * The background color of this node.
 * @expose
 * @name nodes[].color
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The background color for the indicator section of this node.
 * @expose
 * @name nodes[].indicatorColor
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines the primary icon for this node.
 * @expose
 * @name nodes[].icon
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The border color of this icon.
 * @expose
 * @name nodes[].icon.borderColor
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border radius of this icon. CSS border-radius values accepted. Note that non-% values (including unitless) get interpreted as 'px'.
 * @expose
 * @name nodes[].icon.borderRadius
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border width of this icon.
 * @expose
 * @name nodes[].icon.borderWidth
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The fill color of this icon.
 * @expose
 * @name nodes[].icon.color
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The fill pattern of this icon.
 * @expose
 * @name nodes[].icon.pattern
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The opacity of this icon.
 * @expose
 * @name nodes[].icon.opacity
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The shape of this icon. Can take the name of a built-in shape or the svg path commands for a custom shape.
 * @expose
 * @name nodes[].icon.shape
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @ojvalue {string} "circle"
 * @ojvalue {string} "ellipse"
 * @ojvalue {string} "square"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "human"
 * @ojvalue {string} "rectangle"
 * @ojvalue {string} "star"
 * @default null
 */
/**
 * The URL of an image to display for this icon.
 * @expose
 * @name nodes[].icon.source
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of this icon.
 * @expose
 * @name nodes[].icon.width
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The height of this icon.
 * @expose
 * @name nodes[].icon.height
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The CSS style object defining the style of this icon.
 * @ignore
 * @name nodes[].icon.style
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class defining the style of this icon.
 * @ignore
 * @name nodes[].icon.className
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object defining the style of this icon.
 * @expose
 * @name nodes[].icon.svgStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class defining the style of this icon.
 * @expose
 * @name nodes[].icon.svgClassName
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Defines the indicator icon for this node.
 * @expose
 * @name nodes[].indicatorIcon
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The border color of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.borderColor
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border radius of this indicator icon. CSS border-radius values accepted. Note that non-% values (including unitless) get interpreted as 'px'.
 * @expose
 * @name nodes[].indicatorIcon.borderRadius
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border width of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.borderWidth
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The fill color of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.color
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The fill pattern of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.pattern
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The opacity of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.opacity
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The shape of this indicator icon. Can take the name of a built-in shape or the svg path commands for a custom shape.
 * @expose
 * @name nodes[].indicatorIcon.shape
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @ojvalue {string} "circle"
 * @ojvalue {string} "ellipse"
 * @ojvalue {string} "square"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "human"
 * @ojvalue {string} "rectangle"
 * @ojvalue {string} "star"
 * @default null
 */
/**
 * The URL of an image to display for this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.source
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The width of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.width
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The height of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.height
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The CSS style object defining the style of this indicator icon.
 * @ignore
 * @name nodes[].indicatorIcon.style
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class defining the style of this indicator icon.
 * @ignore
 * @name nodes[].indicatorIcon.className
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object defining the style of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.svgStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The CSS style class defining the style of this indicator icon.
 * @expose
 * @name nodes[].indicatorIcon.svgClassName
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional horizontal position (as a percentage) to be used in the average position calculation when grouping across cells.
 * @expose
 * @name nodes[].xPercentage
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */
/**
 * An optional vertical position (as a percentage) to be used in the average position calculation when grouping across cells.
 * @expose
 * @name nodes[].yPercentage
 * @memberof! oj.ojNBox
 * @instance
 * @type {number}
 * @default null
 */

/**
 * The id of the row. Used to identify this row.
 * @expose
 * @name rows[].id
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The text for the row label.
 * @expose
 * @name rows[].label
 * @memberof! oj.ojNBox
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style object defining the style of the row label.
 * @expose
 * @name rows[].labelStyle
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */

/**
 * The CSS style object for this cell. Used for customizing the cell background and border.
 * @ignore
 * @name styleDefaults.cellDefaults.style
 * @memberof! oj.ojNBox
 * @instance
 * @type {Object}
 * @default null
 */


// SubId Locators **************************************************************

/**
 * <p>Sub-ID for NBox cell with specified row and column values.</p>
 *
 * @property {string} row The id of the row.
 * @property {string} column The id of the column.
 *
 * @ojsubid oj-nbox-cell
 * @memberof oj.ojNBox
 *
 * @example <caption>Get the cell with row value 'low' and column value 'high':</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-cell', row: 'low', column: 'high'} );
 */

/**
 * <p>Sub-ID for NBox group node dialog.</p>
 *
 * @ojsubid oj-nbox-dialog
 * @memberof oj.ojNBox
 *
 * @example <caption>Get the group node dialog:</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-dialog'} );
 */

/**
 * <p>Sub-ID for NBox group node dialog close button.</p>
 *
 * @ojsubid oj-nbox-dialog-close-button
 * @memberof oj.ojNBox
 *
 * @example <caption>Get the close button for the group node dialog:</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-dialog-close-button'} );
 */

/**
 * <p>Sub-ID for NBox group node with specified groupCategory value. When grouping is enabled within cells rather than
 * across cells, the row and column ids of the cell should be provided.</p>
 *
 * @property {string} row The id of the row of the associated cell, if one exists.
 * @property {string} column The id of the column of the associated cell, if one exists.
 * @property {string} groupCategory The category represented by the returned group node.
 *
 * @ojsubid oj-nbox-group-node
 * @memberof oj.ojNBox
 *
 * @example <caption>Get the group node with groupCategory value of 'group1' when grouping across cells:</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-group-node', groupCategory: 'group1'} );
 *
 * @example <caption>Get the group node with groupCategory value of 'group1' in the specified cell:</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-group-node', row: 'low', column: 'high', groupCategory: 'group1'} );
 */

/**
 * <p>Sub-ID for NBox node with specified id.</p>
 *
 * @property {string} id The id of the node in the specified cell.
 *
 * @ojsubid oj-nbox-node
 * @memberof oj.ojNBox
 *
 * @example <caption>Get the first node in the specified cell:</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-node', 'id': 'employee1'} );
 */

/**
 * <p>Sub-ID for NBox overflow button in specified cell.</p>
 *
 * @property {string} row The id of the row of the containing cell.
 * @property {string} column The id of the column of the containing cell.
 *
 * @ojsubid oj-nbox-overflow
 * @memberof oj.ojNBox
 *
 * @example <caption>Get the overflow button in the specified cell:</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-overflow', row: 'low', column: 'high'} );
 */

/**
 * <p>Sub-ID for the the NBox tooltip.</p>
 *
 * @ojsubid oj-nbox-tooltip
 * @memberof oj.ojNBox
 *
 * @example <caption>Get the tooltip object of the NBox, if displayed:</caption>
 * var nodes = myNBox.getNodeBySubId( {'subId': 'oj-nbox-tooltip'} );
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for NBox cell with specified row and column values.</p>
 *
 * @property {string} row The id of the row.
 * @property {string} column The id of the column.
 *
 * @ojnodecontext oj-nbox-cell
 * @memberof oj.ojNBox
 */

/**
 * <p>Context for NBox group node dialog.</p>
 *
 * @ojnodecontext oj-nbox-dialog
 * @memberof oj.ojNBox
 */

/**
 * <p>Context for NBox group node with specified groupCategory value. When grouping is enabled within cells rather than
 * across cells, the row and column ids of the cell should be provided.</p>
 *
 * @property {string} row The id of the row of the associated cell, if one exists.
 * @property {string} column The id of the column of the associated cell, if one exists.
 * @property {string} groupCategory The category represented by the returned group node.
 *
 * @ojnodecontext oj-nbox-group-node
 * @memberof oj.ojNBox
 */

/**
 * <p>Context for NBox node with specified id.</p>
 *
 * @property {string} id The id of the node.
 *
 * @ojnodecontext oj-nbox-node
 * @memberof oj.ojNBox
 */

/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojNBoxMeta = {
  "properties": {
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "cellContent": {
      "type": "string",
      "enumValues": ["counts", "auto"]
    },
    "cellMaximize": {
      "type": "string",
      "enumValues": ["off", "on"]
    },
    "cells": {
      "type": "Array<object>|Promise"
    },
    "columns": {
      "type": "Array<object>|Promise"
    },
    "columnsTitle": {
      "type": "string"
    },
    "countLabel": {},
    "groupAttributes": {
      "type": "Array<string>"
    },
    "groupBehavior": {
      "type": "string",
      "enumValues": ["acrossCells", "none", "withinCell"]
    },
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string",
      "enumValues": ["any", "all"]
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "labelTruncation": {
      "type": "string",
      "enumValues": ["ifRequired", "on"]
    },
    "maximizedColumn": {
      "type": "string",
      "writeback": true
    },
    "maximizedRow": {
      "type": "string",
      "writeback": true
    },
    "nodes": {
      "type": "Array<object>|Promise"
    },
    "otherColor": {
      "type": "string"
    },
    "otherThreshold": {
      "type": "number"
    },
    "rows": {
      "type": "Array<object>|Promise"
    },
    "rowsTitle": {
      "type": "string"
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["none", "single", "multiple"]
    },
    "styleDefaults": {
      "type": "object",
      "properties": {
        "animationDuration": {
          "type": "number"
        },
        "cellDefaults": {
          "type": "object",
          "properties": {
            "labelHalign": {
              "type": "string",
              "enumValues": ["center", "end", "start"]
            },
            "labelStyle": {
              "type": "object"
            },
            "showCount": {
              "type": "string",
              "enumValues": ["on", "off", "auto"]
            },
            "svgStyle": {
              "type": "object"
            }
          }
        },
        "columnLabelStyle": {
          "type": "object"
        },
        "columnsTitleStyle": {
          "type": "object"
        },
        "hoverBehaviorDelay": {
          "type": "number"
        },
        "nodeDefaults": {
          "type": "object",
          "properties": {
            "borderColor": {
              "type": "string"
            },
            "borderWidth": {
              "type": "number"
            },
            "color": {
              "type": "string"
            },
            "iconDefaults": {
              "type": "object",
              "properties": {
                "borderColor": {
                  "type": "string"
                },
                "borderRadius": {
                  "type": "string"
                },
                "borderWidth": {
                  "type": "number"
                },
                "color": {
                  "type": "string"
                },
                "height": {
                  "type": "number"
                },
                "opacity": {
                  "type": "number"
                },
                "pattern": {
                  "type": "string",
                  "enumValues": ["smallChecker", "smallCrosshatch", "smallDiagonalLeft", "smallDiagonalRight", "smallDiamond", "smallTriangle",
                                 "largeChecker", "largeCrosshatch", "largeDiagonalLeft", "largeDiagonalRight", "largeDiamond", "largeTriangle", "none"]
                },
                "shape": {
                  "type": "string"
                },
                "source": {
                  "type": "string"
                },
                "width": {
                  "type": "number"
                }
              }
            },
            "indicatorColor": {
              "type": "string"
            },
            "indicatorIconDefaults": {
              "type": "object",
              "properties": {
                "borderColor": {
                  "type": "string"
                },
                "borderRadius": {
                  "type": "string"
                },
                "borderWidth": {
                  "type": "number"
                },
                "color": {
                  "type": "string"
                },
                "height": {
                  "type": "number"
                },
                "opacity": {
                  "type": "number"
                },
                "pattern": {
                  "type": "string",
                  "enumValues": ["smallChecker", "smallCrosshatch", "smallDiagonalLeft", "smallDiagonalRight", "smallDiamond", "smallTriangle",
                                 "largeChecker", "largeCrosshatch", "largeDiagonalLeft", "largeDiagonalRight", "largeDiamond", "largeTriangle", "none"]
                },
                "shape": {
                  "type": "string"
                },
                "source": {
                  "type": "string"
                },
                "width": {
                  "type": "number"
                }
              }
            },
            "labelStyle": {
              "type": "object"
            },
            "secondaryLabelStyle": {
              "type": "object"
            },
          }
        },
        "rowLabelStyle": {
          "type": "object"
        },
        "rowsTitleStyle": {
          "type": "object"
        }
      }
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "touchResponse": {
      "type": "string",
      "enumValues": ["touchStart", "auto"]
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "NBox"
        },
        "highlightedCount": {
          "type": "string",
          "value": "{0}/{1}"
        },
        "labelAdditionalData": {
          "type": "string",
          "value": "Additional Data"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelGroup": {
          "type": "string",
          "value": "Group"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "labelOther": {
          "type": "string",
          "value": "Other"
        },
        "labelSize": {
          "type": "string",
          "value": "Size"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    }
  },
  "events": {},
  "methods": {
    "getCell": {},
    "getColumn": {},
    "getColumnCount": {},
    "getColumnsTitle": {},
    "getContextByNode": {},
    "getDialog": {},
    "getGroupBehavior": {},
    "getGroupNode": {},
    "getRow": {},
    "getRowCount": {},
    "getRowsTitle": {}
  },
  "extension": {
    _WIDGET_NAME: "ojNBox"
  }
};
oj.CustomElementBridge.registerMetadata('oj-n-box', 'dvtBaseComponent', ojNBoxMeta);
// Get the combined meta of superclass which contains a shape parse function generator
var dvtMeta = oj.CustomElementBridge.getMetadata('oj-n-box');
oj.CustomElementBridge.register('oj-n-box', {
  'metadata': dvtMeta,
  'parseFunction': dvtMeta['extension']._DVT_PARSE_FUNC({
  	'style-defaults.node-defaults.icon-defaults.shape': true, 
  	'style-defaults.node-defaults.indicator-icon-defaults.shape': true
  })
});
})();
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/* global Hammer Promise */

/**
 * @class oj.OffcanvasUtils
 * @since 1.1.0
 * @export
 * @ojstatus preview
 *
 * @classdesc
 * This class provides functions used for controlling offcanvas regions.  Offcanvas regions can be used in either static (simply displaying and hiding in response to user interactions) or responsive (using media queries to dynamically move application content between the main viewport and offcanvas regions) contexts.  The open, close and toggle methods can be used to directly control the display of an offcanvas region in both the static and responsive cases.  The setupResponsive and tearDownResponsive methods are only needed for responsive usages and are used to add and remove listeners that use the specified media queries to configure the offcanvas in response to changes in browser size.
 * The setupPanToReveal and tearDownPanToReveal methods are used to add and remove listeners neccessary to reveal the offcanvas as user pans on the outer wrapper.
 *
 * <ul>
 * <li>open: show the offcanvas by sliding it into the viewport.</li>
 * <li>close: hide the offcanvas by sliding it out of the viewport.</li>
 * <li>toggle: toggle the offcanvas in or out of the viewport.</li>
 * <li>setupResponsive: setup offcanvas for the responsive layout.</li>
 * <li>tearDownResponsive: remove listeners that were added in setupResponsive.</li>
 * <li>setupPanToReveal: setup offcanvas for pan to reveal.</li>
 * <li>tearDownPantoReveal: remove listeners that were added in setupPanToReveal.</li><br>
 * </ul>
 *
 * <p>Note for performance reasons, if the Offcanvas content is expensive to render, you should wrap it in an <code class="prettyprint">oj-defer</code> element (API doc {@link oj.ojDefer}) to defer the rendering of that content.<br/>
 * See the <a href="../jetCookbook.html?component=offcanvas&demo=deferredRendering">Offcanvas - Deferred Rendering</a> demo for an example.</p>
 *
 * <h3 id="events-section">
 *   Events
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#events-section"></a>
 * </h3>
 *
 *
 * <table class="generic-table events-table">
 *   <thead>
 *     <tr>
 *       <th>Event</th>
 *       <th>Description</th>
 *       <th>Example</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>beforeClose</td>
 *       <td>Triggered immediately before the offcanvas is closed. It can be canceled to prevent the content from closing by returning a false in the event listener.</td>
*       <td>$(".selector").on("ojbeforeclose", function(event, offcanvas) {});</td>
 *     </tr>
 *     <tr>
 *       <td>beforeOpen<br>
 *       <td>Triggered immediately before the offcanvas is open. It can be canceled to prevent the content from opening by returning a false in the event listener.</td>
*       <td>$(".selector").on("ojbeforeopen", function(event, offcanvas) {});</td>
 *     </tr>
 *     <tr>
 *       <td>close<br>
 *       <td>Triggered after the offcanvas has been closed.</td>
*       <td>$(".selector").on("ojclose", function(event, offcanvas) {});</td>
 *     </tr>
 *     <tr>
 *       <td>open<br>
 *       <td>Triggered after the offcanvas has been open (after animation completes).</td>
*       <td>$(".selector").on("ojopen", function(event, offcanvas) {});</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 */
oj.OffcanvasUtils = {};

oj.OffcanvasUtils._DATA_EDGE_KEY = 'oj-offcanvasEdge';
oj.OffcanvasUtils._DATA_OFFCANVAS_KEY = 'oj-offcanvas';
oj.OffcanvasUtils._DATA_MEDIA_QUERY_KEY = 'oj-mediaQueryListener';
oj.OffcanvasUtils._DATA_HAMMER_KEY = 'oj-offcanvasHammer';
oj.OffcanvasUtils._DATA_STYLE_KEY = 'oj-offcanvasStyle';

/**
 * @private
 */
oj.OffcanvasUtils.SELECTOR_KEY = 'selector';
oj.OffcanvasUtils.CONTENT_KEY = 'content';

oj.OffcanvasUtils.EDGE_START = 'start';
oj.OffcanvasUtils.EDGE_END = 'end';
oj.OffcanvasUtils.EDGE_TOP = 'top';
oj.OffcanvasUtils.EDGE_BOTTOM = 'bottom';


/**
 * @private
 */
oj.OffcanvasUtils.DISPLAY_MODE_KEY = 'displayMode';
/**
 * @private
 */
oj.OffcanvasUtils.DISPLAY_MODE_PUSH = 'push';
/**
 * @private
 */
oj.OffcanvasUtils.DISPLAY_MODE_OVERLAY = 'overlay';
/**
 * @private
 */
oj.OffcanvasUtils.DISPLAY_MODE_PIN = 'pin';

/**
 * @private
 */
oj.OffcanvasUtils.MODALITY_KEY = 'modality';
/**
 * @private
 */
oj.OffcanvasUtils.MODALITY_NONE = 'none';
/**
 * @private
 */
oj.OffcanvasUtils.MODALITY_MODAL = 'modal';


/**
 * @private
 */
oj.OffcanvasUtils.DISMISS_HANDLER_KEY = '_dismissHandler';

/**
 * @private
 */
oj.OffcanvasUtils.OPEN_PROMISE_KEY = '_openPromise';

/**
 * @private
 */
oj.OffcanvasUtils.CLOSE_PROMISE_KEY = '_closePromise';

/**
 * @private
 */
oj.OffcanvasUtils.GLASS_PANE_KEY = '_glassPane';

/**
 * @private
 */
oj.OffcanvasUtils.SURROGATE_KEY = '_surrogate';

/**
 * @private
 */
oj.OffcanvasUtils.SURROGATE_ATTR = 'data-oj-offcanvas-surrogate-id';

/**
 * @private
 */
oj.OffcanvasUtils.OUTER_WRAPPER_SELECTOR = 'oj-offcanvas-outer-wrapper';

/**
 * @private
 */
oj.OffcanvasUtils.OPEN_SELECTOR = 'oj-offcanvas-open';

/**
 * @private
 */
oj.OffcanvasUtils.TRANSITION_SELECTOR = 'oj-offcanvas-transition';

/**
 * @private
 */
oj.OffcanvasUtils.PIN_WRAPPER_SELECTOR = 'oj-offcanvas-pin';
/**
 * @private
 */
oj.OffcanvasUtils.PIN_TRANSITION_SELECTOR = 'oj-offcanvas-pin-transition';

/**
 * @private
 */
oj.OffcanvasUtils.GLASSPANE_SELECTOR = 'oj-offcanvas-glasspane';

/**
 * @private
 */
oj.OffcanvasUtils.GLASSPANE_DIM_SELECTOR = 'oj-offcanvas-glasspane-dim';

/**
 * @private
 */
oj.OffcanvasUtils.VETO_BEFOREOPEN_MSG = 'ojbeforeopen veto';
oj.OffcanvasUtils.VETO_BEFORECLOSE_MSG = 'ojbeforeclose veto';

oj.OffcanvasUtils._shiftSelector =
{
  start: 'oj-offcanvas-shift-start',
  end: 'oj-offcanvas-shift-end',
  top: 'oj-offcanvas-shift-down',
  bottom: 'oj-offcanvas-shift-up'
};

oj.OffcanvasUtils._drawerSelector =
{
  start: 'oj-offcanvas-start',
  end: 'oj-offcanvas-end',
  top: 'oj-offcanvas-top',
  bottom: 'oj-offcanvas-bottom'
};

oj.OffcanvasUtils._getDisplayMode = function (offcanvas) {
  var displayMode = offcanvas[oj.OffcanvasUtils.DISPLAY_MODE_KEY];
  if (displayMode !== oj.OffcanvasUtils.DISPLAY_MODE_OVERLAY &&
      displayMode !== oj.OffcanvasUtils.DISPLAY_MODE_PUSH &&
      displayMode !== oj.OffcanvasUtils.DISPLAY_MODE_PIN) {
    // default displayMode in iOS is push and in android and windows are overlay
    displayMode = (oj.ThemeUtils.parseJSONFromFontFamily('oj-offcanvas-option-defaults') || {}).displayMode;
  }

  return displayMode;
};

oj.OffcanvasUtils._getDrawer = function (offcanvas) {
  return $(offcanvas[oj.OffcanvasUtils.SELECTOR_KEY]);
};

oj.OffcanvasUtils._isModal = function (offcanvas) {
  return offcanvas[oj.OffcanvasUtils.MODALITY_KEY] === oj.OffcanvasUtils.MODALITY_MODAL;
};


// Returns whether the drawer is currently open.
oj.OffcanvasUtils._isOpen = function (drawer) {
  return drawer.hasClass(oj.OffcanvasUtils.OPEN_SELECTOR);
};

oj.OffcanvasUtils._getOuterWrapper = function (drawer) {
  return drawer.closest('.' + oj.OffcanvasUtils.OUTER_WRAPPER_SELECTOR);
};

// selector
// displayMode
oj.OffcanvasUtils._getAnimateWrapper = function (offcanvas) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  if (oj.OffcanvasUtils._noInnerWrapper(offcanvas) ||
      offcanvas[oj.OffcanvasUtils.DISPLAY_MODE_KEY] === oj.OffcanvasUtils.DISPLAY_MODE_OVERLAY) {
    return drawer;
  }

  return drawer.parent();
};


oj.OffcanvasUtils._getShiftSelector = function (edge) {
  var selector = oj.OffcanvasUtils._shiftSelector[edge];
  if (!selector) {
    throw new Error('Invalid edge: ' + edge);
  }

  return selector;
};

oj.OffcanvasUtils._isRTL = function () {
  return oj.DomUtils.getReadingDirection() === 'rtl';
};


oj.OffcanvasUtils._setTransform = function (wrapper, transform) {
  wrapper.css({
    '-webkit-transform': transform,
    transform: transform
  });
};

oj.OffcanvasUtils._getTranslationX = function (edge, width, negate) {
  var minus = (edge === oj.OffcanvasUtils.EDGE_END);
  if (oj.OffcanvasUtils._isRTL() || negate) {
    minus = !minus;
  }

  return 'translate3d(' + (minus ? '-' : '') + width + ', 0, 0)';
};

oj.OffcanvasUtils._setTranslationX = function (wrapper, edge, width) {
  oj.OffcanvasUtils._setTransform(wrapper, oj.OffcanvasUtils._getTranslationX(edge, width, false));
};

oj.OffcanvasUtils._getTranslationY = function (edge, height) {
  var minus = (edge === oj.OffcanvasUtils.EDGE_BOTTOM) ? '-' : '';
  return 'translate3d(0, ' + minus + height + ', 0)';
};

oj.OffcanvasUtils._setTranslationY = function (wrapper, edge, height) {
  oj.OffcanvasUtils._setTransform(wrapper, oj.OffcanvasUtils._getTranslationY(edge, height));
};

oj.OffcanvasUtils._getTranslationY2 = function (height, negate) {
  var minus = negate ? '-' : '';
  return 'translate3d(0, ' + minus + height + ', 0)';
};

oj.OffcanvasUtils._setAnimateClass = function (offcanvas, drawer, $main,
  dtranslation, mtranslation) {
  drawer.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);
  oj.OffcanvasUtils._setTransform(drawer, dtranslation);
  $main.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);
  oj.OffcanvasUtils._setTransform($main, mtranslation);
};


oj.OffcanvasUtils._saveEdge = function (offcanvas) {
  var edge = offcanvas.edge;
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);

  if (!edge || !edge.length) {
    if (drawer.hasClass('oj-offcanvas-start')) {
      edge = oj.OffcanvasUtils.EDGE_START;
    } else if (drawer.hasClass('oj-offcanvas-end')) {
      edge = oj.OffcanvasUtils.EDGE_END;
    } else if (drawer.hasClass('oj-offcanvas-top')) {
      edge = oj.OffcanvasUtils.EDGE_TOP;
    } else if (drawer.hasClass('oj-offcanvas-bottom')) {
      edge = oj.OffcanvasUtils.EDGE_BOTTOM;
    } else {
      // default to start edge
      edge = oj.OffcanvasUtils.EDGE_START;
    }
  }
  $.data(drawer[0], oj.OffcanvasUtils._DATA_EDGE_KEY, edge);

  return edge;
};

oj.OffcanvasUtils._getEdge = function (drawer) {
  return $.data(drawer[0], oj.OffcanvasUtils._DATA_EDGE_KEY);
};


// This method is called right before open and after close animation
// selector
// edge
// displayMode
oj.OffcanvasUtils._toggleClass = function (offcanvas, wrapper, isOpen) {
  var displayMode = offcanvas[oj.OffcanvasUtils.DISPLAY_MODE_KEY];
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var drawerClass = oj.OffcanvasUtils.OPEN_SELECTOR;
  var wrapperClass = (displayMode === oj.OffcanvasUtils.DISPLAY_MODE_OVERLAY) ?
                      oj.OffcanvasUtils.TRANSITION_SELECTOR + ' oj-offcanvas-overlay' :
                      oj.OffcanvasUtils.TRANSITION_SELECTOR;

  // toggle offcanvas and inner wrapper classes
  if (isOpen) {
    drawer.addClass(drawerClass);
    wrapper.addClass(wrapperClass);
  } else {
    // remove oj-focus-highlight
    if (offcanvas.makeFocusable) {
      oj.DomUtils.makeFocusable({
        element: drawer,
        remove: true
      });
    }

    // restore the original tabindex
    var oTabIndex = offcanvas.tabindex;
    if (oTabIndex === undefined) {
      drawer.removeAttr('tabindex');
    } else {
      drawer.attr('tabindex', oTabIndex);
    }

    drawer.removeClass(drawerClass);
    wrapper.removeClass(wrapperClass);
  }
};

// Focus is automatically moved to the first item that matches the following:
// The first element within the offcanvas with the autofocus attribute
// The first :tabbable element inside the offcanvas
// The offcanvas itself
oj.OffcanvasUtils._setFocus = function (_offcanvas) {
  var offcanvas = _offcanvas;
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var focusables = drawer.find('[autofocus]');
  var focusNode;

  if (focusables.length === 0) {
    focusables = drawer.find(':tabbable');
  }
  if (focusables.length === 0) {
    var oTabIndex = drawer.attr('tabindex');
    if (oTabIndex !== undefined) {
      // save the original tabindex
      offcanvas.tabindex = oTabIndex;
    }
    // set tabIndex so the div is focusable
    drawer.attr('tabindex', '-1');
    focusNode = drawer;

    oj.DomUtils.makeFocusable({
      element: drawer,
      applyHighlight: true
    });

    offcanvas.makeFocusable = true;
  } else {
    focusNode = focusables[0];
  }

  oj.FocusUtils.focusElement(focusNode);
};

oj.OffcanvasUtils._isAutoDismiss = function (offcanvas) {
  return offcanvas.autoDismiss !== 'none';
};

oj.OffcanvasUtils._calcTransitionTime = function ($elem) {
  var propertyArray = $elem.css('transitionProperty').split(',');
  var delayArray = $elem.css('transitionDelay').split(',');
  var durationArray = $elem.css('transitionDuration').split(',');
  var maxTime = 0;

  for (var i = 0; i < propertyArray.length; i++) {
    var duration = durationArray[i % durationArray.length];
    var durationMs = (duration.indexOf('ms') > -1) ? parseFloat(duration) : parseFloat(duration) * 1000;
    if (durationMs > 0) {
      var delay = delayArray[i % delayArray.length];
      var delayMs = (delay.indexOf('ms') > -1) ? parseFloat(delay) : parseFloat(delay) * 1000;

      maxTime = Math.max(maxTime, delayMs + durationMs);
    }
  }

  return maxTime + 100;
};

oj.OffcanvasUtils._onTransitionEnd = function (target, handler) {
  var endEvents = 'transitionend.oc webkitTransitionEnd.oc';
  var transitionTimer;
  var listener =
    function () {
      if (transitionTimer) {
        clearTimeout(transitionTimer);
        transitionTimer = undefined;
      }
      // remove handler
      target.off(endEvents, listener);

      handler(target);
    };

  // add transition end listener
  target.on(endEvents, listener);

  transitionTimer =
    setTimeout(listener, oj.OffcanvasUtils._calcTransitionTime(target));
};


oj.OffcanvasUtils._closeWithCatch = function (offcanvas) {
  //  - offcanvas: error occurs when you veto the ojbeforeclose event
  oj.OffcanvasUtils.close(offcanvas).catch(function (reason) {
    oj.Logger.warn('Offcancas close failed: ' + reason);
  });
};

// check offcanvas.autoDismiss
// update offcanvas.dismisHandler
oj.OffcanvasUtils._registerCloseHandler = function (_offcanvas) {
  var offcanvas = _offcanvas;
  // unregister the old handler if exists
  oj.OffcanvasUtils._unregisterCloseHandler(offcanvas);

  if (oj.OffcanvasUtils._isAutoDismiss(offcanvas)) {
    var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);

    // save dismisHandler
    var dismisHandler = function (event) {
      var target = event.target;

      // Ignore mouse events on the scrollbar. FF and Chrome, raises focus events on the
      // scroll container too.
      if (oj.DomUtils.isChromeEvent(event) ||
          (event.type === 'focus' && !$(target).is(':focusable'))) {
        return;
      }

      var key = $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY);
      if (key == null) {
        // offcanvas already destroyed, unregister the handler
        oj.OffcanvasUtils._unregisterCloseHandler(offcanvas);
        return;
      }

      // if event target is not the offcanvas dom subtrees, dismiss it
      if (!oj.DomUtils.isLogicalAncestorOrSelf(drawer[0], target)) {
        oj.OffcanvasUtils._closeWithCatch(offcanvas);
      }
    };
    offcanvas[oj.OffcanvasUtils.DISMISS_HANDLER_KEY] = dismisHandler;

    var documentElement = document.documentElement;
    if (oj.DomUtils.isTouchSupported()) {
      documentElement.addEventListener('touchstart', dismisHandler, true);
    }

    documentElement.addEventListener('mousedown', dismisHandler, true);
    documentElement.addEventListener('focus', dismisHandler, true);
  }

  // register swipe handler
  oj.OffcanvasUtils._registerSwipeHandler(offcanvas);
};

// check offcanvas.autoDismiss
// update offcanvas.dismisHandler
oj.OffcanvasUtils._unregisterCloseHandler = function (_offcanvas) {
  var offcanvas = _offcanvas;
  var dismisHandler = offcanvas[oj.OffcanvasUtils.DISMISS_HANDLER_KEY];
  if (dismisHandler) {
    var documentElement = document.documentElement;

    if (oj.DomUtils.isTouchSupported()) {
      documentElement.removeEventListener('touchstart', dismisHandler, true);
    }

    documentElement.removeEventListener('mousedown', dismisHandler, true);
    documentElement.removeEventListener('focus', dismisHandler, true);
    delete offcanvas[oj.OffcanvasUtils.DISMISS_HANDLER_KEY];

    offcanvas[oj.OffcanvasUtils.DISMISS_HANDLER_KEY] = null;
  }

  // unregister swipe handler
  oj.OffcanvasUtils._unregisterSwipeHandler(offcanvas);
};

oj.OffcanvasUtils._registerSwipeHandler = function (_offcanvas) {
  if (oj.DomUtils.isTouchSupported()) {
    var offcanvas = _offcanvas;
    var selector = offcanvas[oj.OffcanvasUtils.SELECTOR_KEY];
    var drawer = $(selector);
    var edge = oj.OffcanvasUtils._getEdge(drawer);
    var swipeEvent;
    var options;
    var drawerHammer;

    if ((edge === oj.OffcanvasUtils.EDGE_START && !oj.OffcanvasUtils._isRTL()) ||
        (edge === oj.OffcanvasUtils.EDGE_END && oj.OffcanvasUtils._isRTL())) {
      options = {
        recognizers: [
          [Hammer.Swipe, { direction: Hammer.DIRECTION_LEFT }]
        ] };

      swipeEvent = 'swipeleft';
    } else if ((edge === oj.OffcanvasUtils.EDGE_START && oj.OffcanvasUtils._isRTL()) ||
             (edge === oj.OffcanvasUtils.EDGE_END && !oj.OffcanvasUtils._isRTL())) {
      options = {
        recognizers: [
          [Hammer.Swipe, { direction: Hammer.DIRECTION_RIGHT }]
        ] };

      swipeEvent = 'swiperight';
    } else if (edge === oj.OffcanvasUtils.EDGE_TOP) {
      options = {
        recognizers: [
          [Hammer.Swipe, { direction: Hammer.DIRECTION_UP }]
        ] };

      swipeEvent = 'swipeup';
    } else if (edge === oj.OffcanvasUtils.EDGE_BOTTOM) {
      options = {
        recognizers: [
          [Hammer.Swipe, { direction: Hammer.DIRECTION_DOWN }]
        ] };

      swipeEvent = 'swipedown';
    }

    drawerHammer = drawer
      .ojHammer(options)
      .on(swipeEvent, function (event) {
        if (event.target === drawer[0]) {
          event.preventDefault();
          oj.OffcanvasUtils._closeWithCatch(offcanvas);
        }
      });

    // keep the hammer in the offcanvas jquery data
    $.data($(selector)[0], oj.OffcanvasUtils._DATA_HAMMER_KEY,
      { event: swipeEvent,
        hammer: drawerHammer
      });
  }
};

oj.OffcanvasUtils._unregisterSwipeHandler = function (offcanvas) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  if (drawer.length > 0) {
    var dHammer = $.data(drawer[0], oj.OffcanvasUtils._DATA_HAMMER_KEY);
    if (dHammer) {
      dHammer.hammer.off(dHammer.event);
    }
  }
};

oj.OffcanvasUtils._isFixed = function (drawer) {
  return oj.OffcanvasUtils._getOuterWrapper(drawer).hasClass('oj-offcanvas-page');
};

oj.OffcanvasUtils._isPin = function (offcanvas) {
  return (offcanvas[oj.OffcanvasUtils.DISPLAY_MODE_KEY] === oj.OffcanvasUtils.DISPLAY_MODE_PIN);
};

oj.OffcanvasUtils._noInnerWrapper = function (offcanvas) {
  return (offcanvas[oj.OffcanvasUtils.CONTENT_KEY] ||
          oj.OffcanvasUtils._isFixed(oj.OffcanvasUtils._getDrawer(offcanvas)) ||
          oj.OffcanvasUtils._isPin(offcanvas));
};

oj.OffcanvasUtils._saveStyles = function (drawer) {
  var style = drawer.attr('style');
  if (style !== undefined) {
    $.data(drawer[0], oj.OffcanvasUtils._DATA_STYLE_KEY, style);
  }
};

oj.OffcanvasUtils._restoreStyles = function (drawer) {
  var style = $.data(drawer[0], oj.OffcanvasUtils._DATA_STYLE_KEY);
  if (style) {
    drawer.attr('style', style);
  } else {
    drawer.removeAttr('style');
  }
};

oj.OffcanvasUtils._toggleOuterWrapper = function (offcanvas, drawer, test) {
  var edge = oj.OffcanvasUtils._getEdge(drawer);
  var shiftSelector = oj.OffcanvasUtils._getShiftSelector(edge);
  var outerWrapper = oj.OffcanvasUtils._getOuterWrapper(drawer);

  oj.Assert.assertPrototype(outerWrapper, $);

  var isOpen = outerWrapper.hasClass(shiftSelector);
  if (!test) {
    outerWrapper.toggleClass(shiftSelector, !isOpen);
  }

  return isOpen;
};

oj.OffcanvasUtils._afterCloseHandler = function (offcanvas) {
  var wrapper = oj.OffcanvasUtils._getAnimateWrapper(offcanvas);
  // bail if pan to reveal is in progress
  if (wrapper.get(0).style.transform !== '') {
    return;
  }

  //  - customsyntax memory leak: offcanvas needs to implement _disconnected
  // unregister dismiss handler
  oj.OffcanvasUtils._unregisterCloseHandler(offcanvas);

  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var isPin = oj.OffcanvasUtils._isPin(offcanvas);

  // validate offcanvas
  var curOffcanvas = null;
  try {
    curOffcanvas = $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY);
  } catch (e) {
    // Throw away error.
  }
  if (curOffcanvas !== offcanvas) {
    return;
  }

  // After animation, set display:none and remove transition class
  if (isPin) {
    drawer.removeClass(oj.OffcanvasUtils.OPEN_SELECTOR + ' ' +
                      oj.OffcanvasUtils.PIN_TRANSITION_SELECTOR);
  } else {
    oj.OffcanvasUtils._toggleClass(offcanvas, wrapper, false);
  }

  // Remove the glassPane if offcanvas is modal
  oj.OffcanvasUtils._removeModality(offcanvas);

  if (isPin) {
    oj.OffcanvasUtils._getOuterWrapper(drawer).removeClass(oj.OffcanvasUtils.PIN_WRAPPER_SELECTOR);

    oj.OffcanvasUtils._restoreStyles(drawer);
  }

  // fire after close event
  drawer.trigger('ojclose', offcanvas);

  // remove data associate with the offcanvas
  $.removeData(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY);
};

// Set whether the offcanvas is fixed inside the viewport
oj.OffcanvasUtils._setVisible = function (selector, _visible, edge) {
  var drawer = $(selector);
  var visible = !!_visible;

  // close the offcanvas without animation if it's open
  if (visible && oj.OffcanvasUtils._isOpen(drawer)) {
    // hide offcanvas without animation
    oj.OffcanvasUtils._close(selector, false);
  }

  // toggle "oj-offcanvas-" + edge class
  drawer.toggleClass(oj.OffcanvasUtils._drawerSelector[edge], !visible);
};


/**
 * Setup offcanvas for the responsive layout.
 * This method adds a listener based on the media query specified in offcanvas.query.
 * When the media query matches the listener is called and offcanvas behavior is removed.
 * When the media query does not match the listener is called and off canvas behavior is added.
 *
 * @export
 * @param {Object} offcanvas An Object contains the properties in the following table.
 * @property {string} offcanvas.selector JQ selector identifying the offcanvas
 * @property {string} offcanvas.edge the edge of the offcanvas, valid values are start, end, top, bottom. This property is optional if the offcanvas element has a "oj-offcanvas-" + <edge> class specified.
 * @property {string} offcanvas.query the media query determine when the offcanvas is fixed inside the viewport.
 * @return {void}
 *
 * @see #tearDownResponsive
 *
 * @example <caption>Setup the offcanvas:</caption>
 *    var offcanvas = {
 *      "selector": "#startDrawer",
 *      "edge": "start",
 *      "query": oj.ResponsiveUtils.getFrameworkQuery(oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.LG_UP)
 *    };
 *
 * oj.OffcanvasUtils.setupResponsive(offcanvas);
 *
 */
oj.OffcanvasUtils.setupResponsive = function (offcanvas) {
  var mqs = offcanvas.query;
  if (mqs !== null) {
    var selector = offcanvas[oj.OffcanvasUtils.SELECTOR_KEY];
    var query = window.matchMedia(mqs);

    // save the edge
    var edge = oj.OffcanvasUtils._saveEdge(offcanvas);
    var mqListener = function (event) {
      // when event.matches=true fix the offcanvas inside the visible viewport.
      oj.OffcanvasUtils._setVisible(selector, event.matches, edge);
    };

    query.addListener(mqListener);
    oj.OffcanvasUtils._setVisible(selector, query.matches, edge);

    // keep the listener in the offcanvas jquery data
    $.data($(selector)[0], oj.OffcanvasUtils._DATA_MEDIA_QUERY_KEY,
      { mqList: query,
        mqListener: mqListener
      });
  }
};

/**
 * Removes the listener that was added in setupResponsive.  Page authors should call tearDownResponsive when the offcanvas is no longer needed.
 *
 * @export
 * @param {Object} offcanvas An Object contains the properties in the following table.
 * @property {string} offcanvas.selector JQ selector identifying the offcanvas
 * @return {void}
 *
 * @see #setupResponsive
 *
 * @example <caption>TearDown the offcanvas:</caption>
 *    var offcanvas = {
 *      "selector": "#startDrawer"
 *    };
 *
 * oj.OffcanvasUtils.tearDownResponsive(offcanvas);
 *
 */
oj.OffcanvasUtils.tearDownResponsive = function (offcanvas) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var mql = $.data(drawer[0], oj.OffcanvasUtils._DATA_MEDIA_QUERY_KEY);
  if (mql) {
    mql.mqList.removeListener(mql.mqListener);
    $.removeData(drawer[0], oj.OffcanvasUtils._DATA_MEDIA_QUERY_KEY);
  }
};


oj.OffcanvasUtils._openPush = function (offcanvas, resolve, reject, edge) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var $main = $(offcanvas[oj.OffcanvasUtils.CONTENT_KEY]);
  oj.Assert.assertPrototype($main, $);

  // since drawer and main are animated seperately,
  // only resolve true when both transitions are ended
  var pending = true;

  var size = offcanvas.size;
  var translation;

  // transition end handler
  var endHandler = function ($elem) {
    // After animation, remove transition class
    $elem.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

    if (pending) {
      pending = false;
    } else {
      //  - opening offcanvas automatically scrolls to the top
      oj.OffcanvasUtils._setFocus(offcanvas);

      // fire after open event
      drawer.trigger('ojopen', offcanvas);

      //  - push and overlay demos don't work in ie11
      // register dismiss handler as late as possible because IE raises focus event
      // on the launcher that will close the offcanvas if autoDismiss is true
      oj.OffcanvasUtils._registerCloseHandler(offcanvas);

      resolve(true);
    }
  };

  // set display block to get size of offcanvas
  drawer.addClass(oj.OffcanvasUtils.OPEN_SELECTOR);

  // set translationX or Y
  window.setTimeout(function () {
    // if size is not specified, outerWidth/outerHeight is used
    if (edge === oj.OffcanvasUtils.EDGE_START || edge === oj.OffcanvasUtils.EDGE_END) {
      if (size === undefined) {
        size = drawer.outerWidth(true) + 'px';
      }

      //  - offcanvas: drawer push animation is incorrect in rtl mode
//      oj.OffcanvasUtils._setTransform(drawer,
//                                      oj.OffcanvasUtils._getTranslationX(edge, size, true));
      translation = oj.OffcanvasUtils._getTranslationX(edge, size, false);
    } else {
      if (size === undefined) {
        size = drawer.outerHeight(true) + 'px';
      }

      oj.OffcanvasUtils._setTransform(drawer,
        oj.OffcanvasUtils._getTranslationY2(size, edge === oj.OffcanvasUtils.EDGE_TOP));

      translation = oj.OffcanvasUtils._getTranslationY2(size, edge !== oj.OffcanvasUtils.EDGE_TOP);
    }

    // before animation
    window.setTimeout(function () {
      // add transition class
      oj.OffcanvasUtils._setAnimateClass(offcanvas, drawer, $main,
                                         'translate3d(0, 0, 0)', translation);

      oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);

      // add transition end listener
      oj.OffcanvasUtils._onTransitionEnd($main, endHandler);
      oj.OffcanvasUtils._onTransitionEnd(drawer, endHandler);
    }, 0); // before animation
  }, 0);    // set translationX or Y


  // insert a glassPane if offcanvas is modal
  oj.OffcanvasUtils._applyModality(offcanvas, drawer);
};

oj.OffcanvasUtils._openOverlay = function (offcanvas, resolve, reject, edge) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);

  // Before animation, remove display:none and add transition class
  oj.OffcanvasUtils._toggleClass(offcanvas, drawer, true);

  var size = offcanvas.size;
  if (size) {
    if (edge === oj.OffcanvasUtils.EDGE_START || edge === oj.OffcanvasUtils.EDGE_END) {
      oj.OffcanvasUtils._setTransform(drawer,
                                      oj.OffcanvasUtils._getTranslationX(edge, size, true));
    } else {
      oj.OffcanvasUtils._setTransform(drawer,
                                      oj.OffcanvasUtils._getTranslationY(edge, size));
    }
  }

  // show the drawer
  window.setTimeout(function () {
    oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);
  }, 20); // chrome is fine with 0ms but FF needs ~10ms or it wont animate

  // insert a glassPane if offcanvas is modal
  oj.OffcanvasUtils._applyModality(offcanvas, drawer);

  // add transition end listener
  oj.OffcanvasUtils._onTransitionEnd(drawer,
    function () {
      // After animation, remove transition class
      drawer.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

      //  - opening offcanvas automatically scrolls to the top
      oj.OffcanvasUtils._setFocus(offcanvas);

      // fire after open event
      drawer.trigger('ojopen', offcanvas);

      //  - push and overlay demos don't work in ie11
      // register dismiss handler as late as possible because IE raises focus event
      // on the launcher that will close the offcanvas if autoDismiss is true
      oj.OffcanvasUtils._registerCloseHandler(offcanvas);

      resolve(true);
    });
};

/*
oj.OffcanvasUtils._openPin = function(offcanvas, resolve, reject, edge)
{
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var $main = $(offcanvas[oj.OffcanvasUtils.CONTENT_KEY]);
  oj.Assert.assertPrototype($main, $);

  var size = offcanvas["size"];

  //set display block to get size of offcanvas
  drawer.addClass(oj.OffcanvasUtils.OPEN_SELECTOR);

  //set translationX
  window.setTimeout(function () {
    //if size is not specified, outerWidth is used
    if (size === undefined)
      size = drawer.outerWidth(true) + "px";

    drawer.addClass(oj.OffcanvasUtils.PIN_TRANSITION_SELECTOR);

    //make the outer wrapper a flex layout
    oj.OffcanvasUtils._getOuterWrapper(drawer).addClass(oj.OffcanvasUtils.PIN_WRAPPER_SELECTOR);

    oj.OffcanvasUtils._saveStyles(drawer);

    //clear transform only work if set style
    oj.OffcanvasUtils._setTransform(drawer, "none");

    //animate on min-width
    window.setTimeout(function () {
      drawer.css("min-width", size);

      oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);
    }, 10);

  }, 0);    //set translationX

  //insert a glassPane if offcanvas is modal
  oj.OffcanvasUtils._applyModality(offcanvas, drawer);

  //add transition end listener
  oj.OffcanvasUtils._onTransitionEnd(drawer,
    function () {
      //After animation, remove transition class
//      drawer.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

      // - opening offcanvas automatically scrolls to the top
      oj.OffcanvasUtils._setFocus(offcanvas);

      //fire after open event
      drawer.trigger("ojopen", offcanvas);

      // - push and overlay demos don't work in ie11
      //register dismiss handler as late as possible because IE raises focus event
      //on the launcher that will close the offcanvas if autoDismiss is true
      oj.OffcanvasUtils._registerCloseHandler(offcanvas);

      resolve(true);
    });
};
*/

oj.OffcanvasUtils._closePush = function (offcanvas, resolve, reject, drawer, animation) {
  var $main = $(offcanvas[oj.OffcanvasUtils.CONTENT_KEY]);
  // since drawer and main are animated seperately,
  // only resolve true when both transitions are ended
  var pending = true;

  //  - issue in ojoffcanvas when used inside ojtabs
  var endHandler = function () {
    if (!pending) {
      // clear transform translation on $main
      $main.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);
      oj.OffcanvasUtils._setTransform($main, '');
      oj.OffcanvasUtils._afterCloseHandler(offcanvas);
      resolve(true);
    }
    pending = false;
  };

  // clear transform
  oj.OffcanvasUtils._setTransform(drawer, '');
  oj.OffcanvasUtils._setTransform($main, '');
  oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);

  // dim glassPane
  if (oj.OffcanvasUtils._isModal(offcanvas)) {
    offcanvas[oj.OffcanvasUtils.GLASS_PANE_KEY]
      .removeClass(oj.OffcanvasUtils.GLASSPANE_DIM_SELECTOR);
  }

  if (animation) {
    // Before animation, add transition class
    $main.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);
    drawer.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

    // add transition end listener
    oj.OffcanvasUtils._onTransitionEnd(drawer, endHandler);
    oj.OffcanvasUtils._onTransitionEnd($main, endHandler);
  } else {
    pending = false;
    endHandler();
  }
};

oj.OffcanvasUtils._closeOverlay = function (offcanvas, resolve, reject, drawer, animation) {
  var endHandler = function () {
    oj.OffcanvasUtils._afterCloseHandler(offcanvas);
    resolve(true);
  };

  // clear transform
  oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);

  // dim glassPane
  if (oj.OffcanvasUtils._isModal(offcanvas)) {
    offcanvas[oj.OffcanvasUtils.GLASS_PANE_KEY]
      .removeClass(oj.OffcanvasUtils.GLASSPANE_DIM_SELECTOR);
  }

  if (animation) {
    drawer.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

    // add transition end listener
    oj.OffcanvasUtils._onTransitionEnd(drawer, endHandler);
  } else {
    endHandler();
  }
};


oj.OffcanvasUtils._openOldDrawer = function (offcanvas, resolve, reject, edge, displayMode) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var wrapper = oj.OffcanvasUtils._getAnimateWrapper(offcanvas);
  oj.Assert.assertPrototype(wrapper, $);

  // Before animation, remove display:none and add transition class
  oj.OffcanvasUtils._toggleClass(offcanvas, wrapper, true);

  var size;
  if (edge === oj.OffcanvasUtils.EDGE_START || edge === oj.OffcanvasUtils.EDGE_END) {
    // if size is missing, outerWidth is used
    size = (size === undefined) ? (drawer.outerWidth(true) + 'px') : size;

    // don't set transform for oj.OffcanvasUtils.DISPLAY_MODE_OVERLAY
    if (displayMode === oj.OffcanvasUtils.DISPLAY_MODE_PUSH) {
      oj.OffcanvasUtils._setTranslationX(wrapper, edge, size);
    }
  } else {
    // if size is missing, outerHeight is used
    size = (size === undefined) ? (drawer.outerHeight(true) + 'px') : size;

    // don't set transform for oj.OffcanvasUtils.DISPLAY_MODE_OVERLAY
    if (displayMode === oj.OffcanvasUtils.DISPLAY_MODE_PUSH) {
      oj.OffcanvasUtils._setTranslationY(wrapper, edge, size);
    }
  }

  // show the drawer
  window.setTimeout(function () {
    oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);
  }, 10); // chrome is fine with 0ms but FF needs ~10ms or it wont animate

  // insert a glassPane if offcanvas is modal
  oj.OffcanvasUtils._applyModality(offcanvas, drawer);

  // add transition end listener
  oj.OffcanvasUtils._onTransitionEnd(wrapper,
    function () {
      // After animation, remove transition class
      wrapper.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

      //  - opening offcanvas automatically scrolls to the top
      oj.OffcanvasUtils._setFocus(offcanvas);

      // fire after open event
      drawer.trigger('ojopen', offcanvas);

      //  - push and overlay demos don't work in ie11
      // register dismiss handler as late as possible because IE raises focus event
      // on the launcher that will close the offcanvas if autoDismiss is true
      oj.OffcanvasUtils._registerCloseHandler(offcanvas);

      resolve(true);
    });
};

/*
oj.OffcanvasUtils._closePin = function(offcanvas, resolve, reject, drawer, animation)
{
  var endHandler = function ()
  {
    oj.OffcanvasUtils._afterCloseHandler(offcanvas);
    resolve(true);
  };

  //clear transform
  oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);

  //dim glassPane
  if (oj.OffcanvasUtils._isModal(offcanvas))
    offcanvas[oj.OffcanvasUtils.GLASS_PANE_KEY].removeClass(oj.OffcanvasUtils.GLASSPANE_DIM_SELECTOR);

  if (animation) {
    //Before animation, add transition class
    drawer.css("min-width", "0");

    //add transition end listener
    oj.OffcanvasUtils._onTransitionEnd(drawer, endHandler);
  }
  else {
    endHandler();
  }
};
*/

oj.OffcanvasUtils._closeOldDrawer = function (offcanvas, resolve, reject, drawer, animation) {
  var displayMode = offcanvas[oj.OffcanvasUtils.DISPLAY_MODE_KEY];
  var wrapper = oj.OffcanvasUtils._getAnimateWrapper(offcanvas);

  var endHandler = function () {
    oj.OffcanvasUtils._afterCloseHandler(offcanvas);
    resolve(true);
  };

  // clear transform
  if (displayMode === oj.OffcanvasUtils.DISPLAY_MODE_PUSH) {
    oj.OffcanvasUtils._setTransform(wrapper, '');
  }
  oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, false);

  // dim glassPane
  if (oj.OffcanvasUtils._isModal(offcanvas)) {
    offcanvas[oj.OffcanvasUtils.GLASS_PANE_KEY]
      .removeClass(oj.OffcanvasUtils.GLASSPANE_DIM_SELECTOR);
  }

  if (animation) {
    // Before animation, add transition class
    wrapper.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

    // add transition end listener
    oj.OffcanvasUtils._onTransitionEnd(wrapper, endHandler);
  } else {
    endHandler();
  }
};


/**
 * Shows the offcanvas by sliding it into the viewport.  This method fire an ojbeforeopen event which can be vetoed by attaching a listener and returning false.  If the open is not vetoed, this method will fire an ojopen event once animation has completed.
 *
 *<p>Upon opening a offcanvas, focus is automatically moved to the first item that matches the following:</p>
 *<ol>
 *  <li>The first element within the offcanvas with the <code>autofocus</code> attribute</li>
 *  <li>The first <code>:tabbable</code> element inside the offcanvas</li>
 *  <li>The offcanvas itself</li>
 *</ol>
 *
 * @export
 * @param {Object} offcanvas An Object contains the properties in the following table.
 * @property {string} offcanvas.selector JQ selector identifying the offcanvas.
 * @property {string} offcanvas.content JQ selector identifying the main content.
 * @property {string} offcanvas.edge the edge of the offcanvas, valid values are start, end, top, bottom. This property is optional if the offcanvas element has a "oj-offcanvas-" + <edge> class specified.
 * @property {string} offcanvas.displayMode how to show the offcanvas, valid values are push or overlay. Default: defined by theme.
 * @property {string} offcanvas.autoDismiss close behavior, valid values are focusLoss and none. If autoDismiss is default(focusLoss) then any click outside the offcanvas will close it.
 * @property {string} offcanvas.size size width or height of the offcanvas: width if edge is start or end and height if edge is to and bottom. Default to the computed width or height of the offcanvas.
 * @property {string} offcanvas.modality The modality of the offcanvas. Valid values are modal and modeless. Default: modeless. If the offcanvas is modal, interaction with the main content area is disabled like in a modal dialog.
 * @returns {Promise} A promise that is resolved when all transitions have completed. The promise is rejected if the ojbeforeopen event is vetoed.
 * @see #close
 * @see #toggle
 *
 * @example <caption>Slide the offcanvas into the viewport:</caption>
 *    var offcanvas = {
 *      "selector": "#startDrawer",
 *      "content": "#mainContent",
 *      "edge": "start",
 *      "displayMode": "push",
 *      "size": "200px"
 *    };
 *
 * oj.OffcanvasUtils.open(offcanvas);
 *
 */
oj.OffcanvasUtils.open = function (offcanvas) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var oldOffcanvas = $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY);
  if (oldOffcanvas) {
    // if we are in the middle of closing, then return the previous saved promise
    if (oldOffcanvas[oj.OffcanvasUtils.CLOSE_PROMISE_KEY]) {
      return oldOffcanvas[oj.OffcanvasUtils.CLOSE_PROMISE_KEY];
    }

    // if we are in the middle of opening, then return the previous saved promise
    if (oldOffcanvas[oj.OffcanvasUtils.OPEN_PROMISE_KEY]) {
      return oldOffcanvas[oj.OffcanvasUtils.OPEN_PROMISE_KEY];
    }
  }

  var resolveBusyState;
  var veto = false;
  var promise = new Promise(function (resolve, reject) {
    oj.Assert.assertPrototype(drawer, $);

    // save the edge
    var edge = oj.OffcanvasUtils._saveEdge(offcanvas);

    // fire before open event
    var event = $.Event('ojbeforeopen');
    drawer.trigger(event, offcanvas);
    if (event.result === false) {
      reject(oj.OffcanvasUtils.VETO_BEFOREOPEN_MSG);
      veto = true;
      return;
    }

    var displayMode = oj.OffcanvasUtils._getDisplayMode(offcanvas);
    var isPin = oj.OffcanvasUtils._isPin(offcanvas);

    // only support horizontal offcanvas for pin
    if (isPin && (edge === oj.OffcanvasUtils.EDGE_TOP || edge === oj.OffcanvasUtils.EDGE_BOTTOM)) {
      displayMode = oj.OffcanvasUtils.DISPLAY_MODE_PUSH;
    }

    // save a copy of offcanvas object in offcanvas jquery data
    var myOffcanvas = $.extend({}, offcanvas);
    myOffcanvas[oj.OffcanvasUtils.DISPLAY_MODE_KEY] = displayMode;
    $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY, myOffcanvas);

    // throw an error if CONTENT_KEY is specified and the html markup contains an inner wrapper.
    if (offcanvas[oj.OffcanvasUtils.CONTENT_KEY]) {
      if (!oj.OffcanvasUtils._noInnerWrapper(offcanvas)) {
        throw new Error("Error: Both main content selector and the inner wrapper <div class='oj-offcanvas-inner-wrapper'> are provided. Please remove the inner wrapper.");
      }

      // Add a busy state for the animation.  The busy state resolver will be invoked
      // when the animation is completed
      var busyContext = oj.Context.getContext(drawer[0]).getBusyContext();
      resolveBusyState = busyContext.addBusyState(
        { description: "The offcanvas selector ='" +
         offcanvas[oj.OffcanvasUtils.SELECTOR_KEY] + "' doing the open animation." });

      if (isPin) {
//        oj.OffcanvasUtils._openPin(myOffcanvas, resolve, reject, edge);
      } else if (displayMode === oj.OffcanvasUtils.DISPLAY_MODE_PUSH) {
        oj.OffcanvasUtils._openPush(myOffcanvas, resolve, reject, edge);
      } else {
        oj.OffcanvasUtils._openOverlay(myOffcanvas, resolve, reject, edge);
      }
    } else {
      oj.OffcanvasUtils._openOldDrawer(myOffcanvas, resolve, reject, edge, displayMode);
    }
  });

  promise = promise.then(function (value) {
    if (resolveBusyState) {
      resolveBusyState();
    }
    return value;
  }, function (error) {
    if (resolveBusyState) {
      resolveBusyState();
    }
    throw error;
  });

  // save away the current promise
  if (!veto) {
    var nOffcanvas = $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY);
    if (nOffcanvas) {
      nOffcanvas[oj.OffcanvasUtils.OPEN_PROMISE_KEY] = promise;

      // notify subtree
      oj.Components.subtreeShown(drawer[0]);
    }
  }

  return /** @type{Promise} */ (promise);
};

/**
 * Hides the offcanvas by sliding it out of the viewport.  This method fires an ojbeforeclose event which can be vetoed by attaching a listener and returning false.  If the close is not vetoed, this method will fire an ojclose event once animation has completed.
 *
 * @export
 * @param {Object} offcanvas An Object contains the properties in the following table.
 * @property {string} offcanvas.selector JQ selector identifying the offcanvas
 *
 * @returns {Promise} A promise that is resolved when all transitions have completed. The promise is rejected if the ojbeforeclose event is vetoed.
 * @see #open
 * @see #toggle
 *
 * @example <caption>Slide the offcanvas out of the viewport:</caption>
 *    var offcanvas = {
 *      "selector": "#startDrawer"
 *    };
 *
 * oj.OffcanvasUtils.close(offcanvas);
 *
 */
oj.OffcanvasUtils.close = function (offcanvas) {
  return oj.OffcanvasUtils._close(offcanvas[oj.OffcanvasUtils.SELECTOR_KEY], true);
};

oj.OffcanvasUtils._close = function (selector, animation) {
  var drawer = $(selector);

  oj.Assert.assertPrototype(drawer, $);

  var offcanvas = $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY);

  // if we are in the middle of closing, then return the previous saved promise
  if (offcanvas && offcanvas[oj.OffcanvasUtils.CLOSE_PROMISE_KEY]) {
    return offcanvas[oj.OffcanvasUtils.CLOSE_PROMISE_KEY];
  }

  var resolveBusyState;
  var veto = false;
  var promise = new Promise(function (resolve, reject) {
    if (oj.OffcanvasUtils._isOpen(drawer)) {
      if (selector !== offcanvas[oj.OffcanvasUtils.SELECTOR_KEY]) {
        resolve(true);
      }

      // if the outer wrapper doesn't have the correct shift selector, we are done
      if (!oj.OffcanvasUtils._toggleOuterWrapper(offcanvas, drawer, true)) {
        resolve(true);
      }

      // fire before close event
      var event = $.Event('ojbeforeclose');
      drawer.trigger(event, offcanvas);
      if (event.result === false) {
        reject(oj.OffcanvasUtils.VETO_BEFORECLOSE_MSG);
        veto = true;
        return;
      }

      // Add a busy state for the animation.  The busy state resolver will be invoked
      // when the animation is completed
      if (animation) {
        var busyContext = oj.Context.getContext(drawer[0]).getBusyContext();
        resolveBusyState = busyContext.addBusyState(
          { description: "The offcanvas selector ='" +
           offcanvas[oj.OffcanvasUtils.SELECTOR_KEY] + "' doing the close animation." });
      }

      var displayMode = offcanvas[oj.OffcanvasUtils.DISPLAY_MODE_KEY];
      if (offcanvas[oj.OffcanvasUtils.CONTENT_KEY]) {
        if (displayMode === oj.OffcanvasUtils.DISPLAY_MODE_PUSH) {
          oj.OffcanvasUtils._closePush(offcanvas, resolve, reject, drawer, animation);
        } else {
          oj.OffcanvasUtils._closeOverlay(offcanvas, resolve, reject, drawer, animation);
        }
      } else {
        oj.OffcanvasUtils._closeOldDrawer(offcanvas, resolve, reject, drawer, animation);
      }
    } else {
      resolve(true);
    }
  });

  promise = promise.then(function (value) {
    if (resolveBusyState) {
      resolveBusyState();
    }
    return value;
  }, function (error) {
    if (resolveBusyState) {
      resolveBusyState();
    }
    throw error;
  });

  // save away the current promise
  if (!veto) {
    offcanvas = $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY);
    if (offcanvas) {
      offcanvas[oj.OffcanvasUtils.CLOSE_PROMISE_KEY] = promise;

      // notify subtree
      oj.Components.subtreeHidden(drawer[0]);
    }
  }

  return /** @type{Promise} */ (promise);
};

/**
 * Toggles the offcanvas in or out of the viewport.  This method simply delegates to the open or close methods as appropriate.
 *
 * @export
 * @param {Object} offcanvas An Object contains the properties in the following table.
 * @property {string} offcanvas.selector JQ selector identifying the offcanvas.
 * @property {string} offcanvas.content JQ selector identifying the main content.
 * @property {string} offcanvas.edge the edge of the offcanvas, valid values are start, end, top, bottom. This property is optional if the offcanvas element has a "oj-offcanvas-" + <edge> class specified.
 * @property {string} offcanvas.displayMode how to show the offcanvas, valid values are push or overlay. Default: defined by theme.
 * @property {string} offcanvas.autoDismiss close behavior, valid values are focusLoss and none. If autoDismiss is default(focusLoss) then any click outside the offcanvas will close it.
 * @property {string} offcanvas.size size width or height of the offcanvas: width if edge is start or end and height if edge is to and bottom. Default to the computed width or height of the offcanvas.
 * @property {string} offcanvas.modality The modality of the offcanvas. Valid values are modal and modeless. Default: modeless. If the offcanvas is modal, interaction with the main content area is disabled like in a modal dialog.
 * @returns {Promise} A promise that is resolved when all transitions have completed
 * @see #open
 * @see #close
 *
 * @example <caption>Toggle the offcanvas in or out of the viewport:</caption>
 *    var offcanvas = {
 *      "selector": "#startDrawer",
 *      "edge": "start",
 *      "displayMode": "overlay"
 *    };
 *
 * oj.OffcanvasUtils.toggle(offcanvas);
 *
 */
oj.OffcanvasUtils.toggle = function (offcanvas) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  oj.Assert.assertPrototype(drawer, $);

  if (oj.OffcanvasUtils._isOpen(drawer)) {
    return oj.OffcanvasUtils.close(offcanvas);
  }

  return oj.OffcanvasUtils.open(offcanvas);
};


/**
 * Creates an overlay div with the oj-offcanvas-glasspane selector
 * append to the end of the drawer's container
 * @param {!jQuery} drawer the drawer
 * @return {jQuery} the overlay div
 * @private
 */
oj.OffcanvasUtils._addGlassPane = function (drawer) {
  var overlay = $('<div>');
  overlay.addClass(oj.OffcanvasUtils.GLASSPANE_SELECTOR);
  overlay.attr('role', 'presentation');
  overlay.attr('aria-hidden', 'true');

  // append glassPane at the end
  overlay.appendTo(drawer.parent()); // @HTMLUpdateOK
  overlay.on('keydown keyup keypress mousedown mouseup mouseover mouseout click focusin focus',
    function (event) {
      event.stopPropagation();
      event.preventDefault();
    });

  return overlay;
};

/**
 * Creates a script element before the target layer bound to the simple jquery UI
 * surrogate component.
 *
 * @param {!jQuery} layer stacking context
 * @return {jQuery}
 * @private
 */
oj.OffcanvasUtils._createSurrogate = function (layer) {
  //  - offcanvas utils use of <script>
  var surrogate = $("<span style='display:none'>");
  surrogate.attr('aria-hidden', 'true');

  var layerId = layer.attr('id');

  var surrogateId;
  if (layerId) {
    surrogateId = [layerId, 'surrogate'].join('_');
    surrogate.attr('id', surrogateId);
  } else {
    surrogateId = surrogate.uniqueId();
  }
  surrogate.insertBefore(layer); // @HTMLUpdateOK

  // loosely associate the offcanvas to the surrogate element
  layer.attr(oj.OffcanvasUtils.SURROGATE_ATTR, surrogateId);

  return surrogate;
};


/**
 * bring the drawer to the front to keep this order:  mainContent, glassPane, drawer
 * so we don't need to use z-index
 * @private
 */
oj.OffcanvasUtils._swapOrder = function (_offcanvas, drawer) {
  var offcanvas = _offcanvas;
  // create a surrogate in front of the mainContent to be used in _restoreOrder
  offcanvas[oj.OffcanvasUtils.SURROGATE_KEY] = oj.OffcanvasUtils._createSurrogate(drawer);

  drawer.appendTo(drawer.parent()); // @HTMLUpdateOK
};


/**
 * restore the order before _swapOrder
 * @private
 */
oj.OffcanvasUtils._restoreOrder = function (offcanvas) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var surrogate = offcanvas[oj.OffcanvasUtils.SURROGATE_KEY];

  if (drawer && surrogate &&
      drawer.attr(oj.OffcanvasUtils.SURROGATE_ATTR) === surrogate.attr('id')) {
    drawer.insertAfter(surrogate); // @HTMLUpdateOK
    // remove link to the surrogate element
    drawer.removeAttr(oj.OffcanvasUtils.SURROGATE_ATTR);
    surrogate.remove(); // @HTMLUpdateOK
  }
};

/**
 * Apply modality
 * If offcanvas is modal, add a glasspane and keep the dom structure in the following order:
 * mainContent, glassPane and drawer so we don't need to apply z-index
 * @private
 */
oj.OffcanvasUtils._applyModality = function (_offcanvas, drawer) {
  var offcanvas = _offcanvas;
  if (oj.OffcanvasUtils._isModal(offcanvas)) {
    // insert glassPane in front of the mainContent
    offcanvas[oj.OffcanvasUtils.GLASS_PANE_KEY] = oj.OffcanvasUtils._addGlassPane(drawer);

    // bring the drawer <div> to the front
    // to keep this order:  mainContent, glassPane, drawer
    oj.OffcanvasUtils._swapOrder(offcanvas, drawer);

    //  - acc: talkback reports on elements on background page whilst navdrawer open
    var $main = $(offcanvas[oj.OffcanvasUtils.CONTENT_KEY]);
    $main.attr('aria-hidden', 'true');

    window.setTimeout(function () {
      offcanvas[oj.OffcanvasUtils.GLASS_PANE_KEY]
        .addClass(oj.OffcanvasUtils.GLASSPANE_DIM_SELECTOR);
    }, 0);
  }
};

/**
 * Remove modality
 * If offcanvas is modal, remove glasspane and restore the dom element orders
 * @private
 */
oj.OffcanvasUtils._removeModality = function (offcanvas) {
  if (oj.OffcanvasUtils._isModal(offcanvas)) {
    offcanvas[oj.OffcanvasUtils.GLASS_PANE_KEY].remove();
    // restore the order
    oj.OffcanvasUtils._restoreOrder(offcanvas);

    //  - acc: talkback reports on elements on background page whilst navdrawer open
    var $main = $(offcanvas[oj.OffcanvasUtils.CONTENT_KEY]);
    $main.removeAttr('aria-hidden');
  }
};


/**
 * Setup offcanvas for pan to reveal.
 * This method adds a touch listener to handle revealing the offcanvas as user pans on the outer wrapper.  The following events are fired by this method:
 * ojpanstart - fired when pan to reveal gesture initiated by the user.  The event includes the direction and distance of the pan.  If it is vetoed
 *              then pan to reveal is terminated
 * ojpanmove  - fired as user continues the pan gesture.  The event includes the direction and distance of the pan.
 * ojpanend   - fired when pan to reveal gesture ends.  The event includes the direction and distance of the pan.  If it is vetoed then the offcanvas
 *              will be closed.
 *
 * @export
 * @param {Object} offcanvas An Object contains the properties in the following table.
 * @property {string} offcanvas.selector JQ selector identifying the offcanvas
 * @property {string=} offcanvas.edge the edge of the offcanvas, valid values are start, end. This property is optional if the offcanvas element has a "oj-offcanvas-" + <edge> class specified.
 * @property {string=} offcanvas.size size width of the offcanvas.  Default to the computed width of the offcanvas.
 * @return {void}
 *
 * @see #tearDownPanToReveal
 *
 * @example <caption>Setup the offcanvas:</caption>
 *    var offcanvas = {
 *      "selector": "#startDrawer"
 *    };
 *
 * oj.OffcanvasUtils.setupPanToReveal(offcanvas);
 *
 */
oj.OffcanvasUtils.setupPanToReveal = function (_offcanvas) {
  var offcanvas = _offcanvas;
  var drawer;
  var size;
  var outerWrapper;
  var wrapper;
  var mOptions;
  var proceed;
  var direction;
  var ui;
  var evt;
  var delta;
  var edge;
  var endEvents;
  var listener;

  if ($(offcanvas).attr('oj-data-pansetup') != null) {
    // already setup
    return;
  }

  // mark as setup
  $(offcanvas).attr('oj-data-pansetup', 'true');

  // pan to reveal only works for push display mode, so enforce it
  offcanvas.displayMode = 'push';

  drawer = oj.OffcanvasUtils._getDrawer(offcanvas);

  outerWrapper = oj.OffcanvasUtils._getOuterWrapper(drawer);

  // use hammer for swipe
  mOptions = {
    recognizers: [
       [Hammer.Pan, { direction: Hammer.DIRECTION_HORIZONTAL }]
    ] };

  // flag to signal whether pan to reveal should proceed
  proceed = false;

  $(outerWrapper)
    .ojHammer(mOptions)
    .on('panstart', function (event) {
      direction = null;

      switch (event.gesture.direction) {
        case Hammer.DIRECTION_LEFT:
                // diagonal case
          if (Math.abs(event.gesture.deltaY) < Math.abs(event.gesture.deltaX)) {
            direction = oj.OffcanvasUtils._isRTL() ? 'end' : 'start';
          }
          break;
        case Hammer.DIRECTION_RIGHT:
                // diagonal case
          if (Math.abs(event.gesture.deltaY) < Math.abs(event.gesture.deltaX)) {
            direction = oj.OffcanvasUtils._isRTL() ? 'start' : 'end';
          }
          break;
        default:
      }

      if (direction === null) {
        return;
      }

      ui = { direction: direction, distance: Math.abs(event.gesture.deltaX) };
      evt = $.Event('ojpanstart');
      drawer.trigger(evt, ui);

      if (!evt.isDefaultPrevented()) {
            // need the size to display the canvas when release
        size = offcanvas.size;
        if (size == null) {
          size = drawer.outerWidth();
          offcanvas.size = size;
        }

            // make sure it's in closed state
        offcanvas._closePromise = null;

            // cancel any close animation transition handler
        wrapper = oj.OffcanvasUtils._getAnimateWrapper(offcanvas);
        wrapper.off('.oc');

            // sets the appropriate offcanvas class
        oj.OffcanvasUtils._toggleClass(offcanvas, wrapper, true);
        proceed = true;

            // stop touch event from bubbling to prevent for example pull to refresh from happening
        event.gesture.srcEvent.stopPropagation();

            // stop bubbling
        event.stopPropagation();
      }
    })
    .on('panmove', function (event) {
        // don't do anything if start is vetoed
      if (!proceed) {
        return;
      }

      delta = event.gesture.deltaX;
      if ((direction === 'start' && delta > 0) || (direction === 'end' && delta < 0)) {
        oj.OffcanvasUtils._setTranslationX(wrapper, 'start', '0px');
        return;
      }

      drawer.css('width', Math.abs(delta));

        // don't do css transition animation while panning
      wrapper.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);
      oj.OffcanvasUtils._setTranslationX(wrapper, 'start', delta + 'px');

      ui = { direction: direction, distance: Math.abs(delta) };
      evt = $.Event('ojpanmove');
      drawer.trigger(evt, ui);

        // stop touch event from bubbling to prevent for example pull to refresh from happening
      event.gesture.srcEvent.stopPropagation();

        // stop bubbling
      event.stopPropagation();
    })
    .on('panend', function (event) {
        // don't do anything if start is vetoed
      if (!proceed) {
        return;
      }

        // reset flag
      proceed = false;

      delta = Math.abs(event.gesture.deltaX);
      ui = { distance: delta };
      evt = $.Event('ojpanend');
      drawer.trigger(evt, ui);

        // stop bubbling
      event.stopPropagation();

      if (!evt.isDefaultPrevented()) {
        edge = offcanvas.edge;
        if (edge == null) {
          if (drawer.hasClass('oj-offcanvas-start')) {
            edge = 'start';
          } else {
            edge = 'end';
          }
        }

        oj.OffcanvasUtils._animateWrapperAndDrawer(wrapper, drawer, edge, size, offcanvas);

        $.data(drawer[0], oj.OffcanvasUtils._DATA_OFFCANVAS_KEY, offcanvas);
        $.data(drawer[0], oj.OffcanvasUtils._DATA_EDGE_KEY, edge);

        oj.OffcanvasUtils._registerCloseHandler(offcanvas);

        return;
      }

        // close the toolbar
      endEvents = 'transitionend webkitTransitionEnd otransitionend oTransitionEnd';
      listener = function () {
            // reset offcanvas class
        oj.OffcanvasUtils._toggleClass(offcanvas, wrapper, false);

        wrapper.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

            // remove handler
        wrapper.off(endEvents, listener);

            // fire close event after completely closed
        drawer.trigger('ojclose', offcanvas);
      };

        // add transition end listener
      wrapper.on(endEvents, listener);

        // restore item position
      wrapper.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);
      oj.OffcanvasUtils._setTranslationX(wrapper, 'start', '0px');
    });
};

// animate both the wrapper and drawer at the same time
oj.OffcanvasUtils._animateWrapperAndDrawer = function (wrapper, drawer, edge, size, offcanvas) {
  var tt = 400;
  var fps = 60;
  var ifps;
  var matrix;
  var values;
  var current;
  var first;
  var final;
  var reqId;
  var inc;
  var lastFrame;
  var func;
  var currentFrame;
  var adjInc;

    // since we can't synchronize two css transitions, we'll have to do the animation ourselves using
    // requestAnimationFrame
    // make sure wrapper animation is off
  wrapper.removeClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

    // ideal ms per frame
  ifps = Math.round(1000 / fps);
  matrix = wrapper.css('transform');
  if (matrix === 'none') {
        // could happen if this method was called after canvas is closed
    return;
  }
  values = matrix.split('(')[1].split(')')[0].split(',');
    // this is the translateX
  current = parseInt(values[4], 10);
  first = current;
    // the final size/destination
  final = edge === 'end' ? 0 - size : size;
    // calculate the increment needed to complete transition in 400ms with 60fps
  inc = Math.max(1, Math.abs(final - current) / (tt / ifps));
  lastFrame = (new Date()).getTime();
  func = function () {
        // check if it got interrupted by close
    if (first !== current && wrapper.get(0).style.transform === '') {
      window.cancelAnimationFrame(reqId);
      return;
    }

    currentFrame = (new Date()).getTime();
        // see how much we'll need to compensate if fps drops below ideal
    adjInc = Math.max(inc, inc * Math.max((currentFrame - lastFrame) / ifps));
    lastFrame = currentFrame;
    if (current < final) {
      current = Math.min(final, current + adjInc);
    } else if (current > final) {
      current = Math.max(final, current - adjInc);
    }

    oj.OffcanvasUtils._setTranslationX(wrapper, edge, Math.abs(current) + 'px');
    drawer.css('width', Math.abs(current) + 'px');

        // make sure to cancel the animation frame if we are done
    if (current === final) {
      window.cancelAnimationFrame(reqId);
      wrapper.addClass(oj.OffcanvasUtils.TRANSITION_SELECTOR);

            // fire after completely open
      drawer.trigger('ojopen', offcanvas);
    } else {
      reqId = window.requestAnimationFrame(func);
    }
  };

  reqId = window.requestAnimationFrame(func);
};

/**
 * Removes the listener that was added in setupPanToReveal.  Page authors should call tearDownPanToReveal when the offcanvas is no longer needed.
 *
 * @export
 * @param {Object} offcanvas An Object contains the properties in the following table.
 * @property {string} offcanvas.selector JQ selector identifying the offcanvas
 * @return {void}
 * @see #setupPanToReveal
 *
 * @example <caption>TearDown the offcanvas:</caption>
 *    var offcanvas = {
 *      "selector": "#startDrawer"
 *    };
 *
 * oj.OffcanvasUtils.tearDownPanToReveal(offcanvas);
 *
 */
oj.OffcanvasUtils.tearDownPanToReveal = function (offcanvas) {
  var drawer = oj.OffcanvasUtils._getDrawer(offcanvas);
  var outerWrapper = oj.OffcanvasUtils._getOuterWrapper(drawer);

    // remove all listeners
  $(outerWrapper).off('panstart panmove panend');
};

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Offcanvas</td>
 *       <td><kbd>Swipe</kbd></td>
 *       <td>Close the offcanvas by swiping in the close direction.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.OffcanvasUtils
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Close button</td>
 *       <td><kbd>Enter</kbd></td>
 *       <td>If offcanvas has a close button, navigate to the button and press Enter to close the offcanvas.</td>
 *     </tr>
 *     <tr>
 *       <td>Outer wrapper</td>
 *       <td><kbd>Tab or Shift+Tab</kbd></td>
 *       <td>When focus is on the outer wrapper, Tab or shift+Tab moves the focus to the next or previous tab stop on the page. If autoDismiss is true, it also closes the offcanvas.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.OffcanvasUtils
 */

/**
 * <p>The following CSS classes can be applied by the page author as needed.</p>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Class</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-offcanvas-outer-wrapper</td>
 *       <td>Applied to the outer most element of the offcanvas.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-offcanvas-page</td>
 *       <td>Applied to the outer wrapper of the page level offcanvas.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-offcanvas-inner-wrapper<br>
 *       <td>Applied to the inner wrapper of the offcanvas. Deprecated, please remove the inner wrapper.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-offcanvas-start<br>
 *       <td>Applied to the offcanvas on the start edge.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-offcanvas-end<br>
 *       <td>Applied to the offcanvas on the end edge.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-offcanvas-top<br>
 *       <td>Applied to the offcanvas on the top edge.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-offcanvas-bottom<br>
 *       <td>Applied to the offcanvas on the bottom edge.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-offcanvas-overlay-shadow</td>
 *       <td>Add this marker class to an overlay offcanvas to show a drop shadow when it is open.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
 * @memberof oj.OffcanvasUtils
 */

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @constructor
 * @private
 */
function ojOptGroup(context) {
  this.updateDOM = function() {
    var customRenderer = context.element["customOptgroupRenderer"];
    if (customRenderer && typeof customRenderer === "function")
      customRenderer(context.element);    
  };
}
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojOptgroup
 * @since 4.0.0
 * @ojdisplayname Option Group
 * @ojshortdesc Supports grouping of child oj-option elements.
 * @ojrole option
 * @ojstatus preview
 *
 * @classdesc
 * <h3 id="optgroupOverview-section">
 *   JET Optgroup
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#optgroupOverview-section"></a>
 * </h3>
 * <p>The oj-optgroup element is used to group [oj-option]{@link oj.ojOption} elements.</p>
 * <p>For example:
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-select-one>
 *   &lt;oj-optgroup label="group1 label">
 *     &lt;oj-option value="option 1">option 1&lt;/oj-option>
 *     &lt;oj-option value="option 2">option 2&lt;/oj-option>
 *   &lt;/oj-optgroup>
 *   &lt;oj-optgroup label="group2 label">
 *     &lt;oj-option value="option 3">option 3&lt;/oj-option>
 *     &lt;oj-option value="option 4">option 4&lt;/oj-option>
 *   &lt;/oj-optgroup>
 * &lt;/oj-select-one> 
 * </code></pre>
 */
 
/**
 * <p>Disables the oj-optgroup if set to <code class="prettyprint">true</code>.
 *
 * @name disabled
 * @ojshortdesc Disables the group if set to true.
 * @expose
 * @memberof oj.ojOptgroup
 * @instance
 * @type {boolean}
 * @default false
 *
 * @example <caption>Initialize the oj-optgroup with the <code class="prettyprint">disabled</code> attribute specified:</caption>
 * &lt;oj-optgroup disabled="[[isDisabled]]">&lt;/oj-optgroup>
 *
 * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
 * // getter
 * var disabledValue = myOptgroup.disabled;
 *
 * // setter
 * myOptgroup.disabled = true;
 */
 
/**
 * <p>Specifies the oj-optgroup's label.
 *
 * @name label
 * @ojshortdesc The group label.
 * @ojrequired
 * @ojtranslatable
 * @expose
 * @memberof oj.ojOptgroup
 * @instance
 * @type {string}
 *
 * @example <caption>Initialize the oj-optgroup with the <code class="prettyprint">label</code> attribute specified:</caption>
 * &lt;oj-optgroup label="group1 label">&lt;/oj-optgroup>
 *
 * @example <caption>Get or set the <code class="prettyprint">label</code> property after initialization:</caption>
 * // getter
 * var labelValue = myOptgroup.label;
 *
 * // setter
 * myOptgroup.label = 'Group 1';
 */
 
/**
 * <p>The oj-optgroup element accepts <code class="prettyprint">oj-option</code>s as children. See the [oj-option]{@link oj.ojOption} doc for details about
 * accepted children and slots.</p>
 *
 * @ojchild Default
 * @memberof oj.ojOptgroup
 *
 * @example <caption>Initialize the oj-optgroup with child content specified:</caption>
 * &lt;oj-optgroup>
 *   &lt;oj-option value="option1">Option label&lt;/oj-option>
 * &lt;/oj-optgroup>
 */    
 
/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 * 
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * 
 * @expose
 * @memberof oj.ojOptgroup
 * @instance
 * 
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */ 
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * 
 * @expose
 * @memberof oj.ojOptgroup
 * @instance
 * 
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */ 
/**
 * Refreshes the visual state of the component.
 * 
 * @function refresh
 * 
 * @expose
 * @memberof oj.ojOptgroup
 * @instance
 */ 
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * 
 * @expose
 * @memberof oj.ojOptgroup
 * @instance
 * 
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */ 
/**
 * @protected
 * @ignore
 */
(function() {
  var ojOptgroupMeta = {
    "properties": {
      "disabled": {
        "type": "boolean",
        "value": false
      }, 
      "label": {
        "type": "string",
        "value": ""
      },
      // not documented
      "customOptgroupRenderer": {}
    },
    "extension": {
      _CONSTRUCTOR: ojOptGroup
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-optgroup', null, ojOptgroupMeta);
  oj.CustomElementBridge.register('oj-optgroup', {'metadata': oj.CustomElementBridge.getMetadata('oj-optgroup')});
})();

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @constructor
 * @private
 */
function ojOption(context) {

  /**
   * Append each node to the element.
   * @private
   */
  function _appendNodes(elem, nodes) {
    $.each(nodes, function(i, node) {
      elem.appendChild(node);
    })
  };

  /**
   * Remove node from the element.
   * @private
   */
  function _removeNodes(elem, nodes) {
    $.each(nodes, function(i, node) {
      elem.removeChild(node);
    })
  };

  /**
   * Rearrange slots in the right order and remove unwanted slots.
   * @private
   */
  function _arrangeSlots(elem) {
    // get the slotMap
    var slots = oj.BaseCustomElementBridge.getSlotMap(elem);
    var supportedSlots = ["startIcon", "", "endIcon"];
    
    // remove unwanted slots
    $.each(slots, function(slotName, nodes) {
      if (supportedSlots.indexOf(slotName) == -1) {
        _removeNodes(elem, nodes);
      }
    })
    
    // rearrange slots
    $.each(supportedSlots, function(i, slotName) {
      if (slots[slotName])
        _appendNodes(elem, slots[slotName]);
    })
  };

  this.updateDOM = function() {
    var customRenderer = context.element["customOptionRenderer"];

    // reorder the slots and remove unwanted slots
    _arrangeSlots(context.element);

    if (customRenderer && typeof customRenderer === "function")
      customRenderer(context.element);    
  };
}

 
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojOption
 * @since 4.0.0
 * @ojshortdesc Declare values for JET elements that display a list of values.
 * @ojrole option
 * @ojstatus preview
 * @ojsignature class ojOption extends JetElement<ojOptionSettableProperties>
 *
 * @classdesc
 * <h3 id="optionOverview-section">
 *   JET Option
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#optionOverview-section"></a>
 * </h3>
 * <p>The oj-option element is used to declare values for JET elements that display list of values.
 * It's supported by the following elements:
 * <ul>
 * <li>oj-buttonset-one</li>
 * <li>oj-buttonset-many</li>
 * <li>oj-checkboxset</li>
 * <li>oj-combobox-one</li>
 * <li>oj-combobox-many</li>
 * <li>oj-menu</li>
 * <li>oj-radioset</li>
 * <li>oj-select-one</li>
 * <li>oj-select-many</li>
 * </ul>
 * </p>
 * <p>For example:
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-select-one>
 *   &lt;oj-option value="option 1">option 1&lt;/oj-option>
 *   &lt;oj-option value="option 2">option 2&lt;/oj-option>
 *   &lt;oj-option value="option 3">option 3&lt;/oj-option>
 *   &lt;oj-option value="option 4">option 4&lt;/oj-option>
 * &lt;/oj-select-one> 
 * </code></pre>
 */
 
/**
 * <p>Disables the oj-option if set to <code class="prettyprint">true</code>.
 *
 * @name disabled
 * @ojshortdesc Disables the option if set to true.
 * @expose
 * @memberof oj.ojOption
 * @instance
 * @type {boolean}
 * @default false
 *
 * @example <caption>Initialize the oj-option with the <code class="prettyprint">disabled</code> attribute specified:</caption>
 * &lt;oj-option disabled="[[isDisabled]]" value="option1">Option1&lt;/oj-option>
 *
 * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
 * // getter
 * var disabledValue = myOption.disabled;
 *
 * // setter
 * myOption.disabled = true;
 */
 
/**
 * <p>Specifies the oj-option's value. The value is associated with the oj-option element whose display value may be different.
 *
 * @name value
 * @ojshortdesc The value of the option.
 * @ojrequired
 * @expose
 * @memberof oj.ojOption
 * @instance
 * @type {*}
 *
 * @example <caption>Initialize the oj-option with the <code class="prettyprint">value</code> attribute specified:</caption>
 * &lt;oj-option value="option1">Option1&lt;/oj-option>
 *
 * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
 * // getter
 * var optionValue = myOption.value;
 *
 * // setter
 * myOption.value = 'option1';
 */

/**
 * <p>Child content for oj-option. This is normally the text node that displays for oj-option.</p>
 *
 * @ojchild Default
 * @ojshortdesc The default slot for the option's content.
 * @memberof oj.ojOption
 *
 * @example <caption>Initialize the oj-option with child content specified:</caption>
 * &lt;oj-option>
 *   Option label
 * &lt;/oj-option>
 */
 
/**
 * <p>Named slot for the oj-option's start icon.</p>
 *
 * @ojslot startIcon
 * @ojshortdesc The slot for the option's start icon.
 * @memberof oj.ojOption
 *
 * @example <caption>Initialize the oj-option with the <code class="prettyprint">startIcon</code> slot specified:</caption>
 * &lt;oj-option>
 *   &lt;span slot='startIcon'>&lt;img src='start.png' alt='Start'>&lt;/span>
 * &lt;/oj-option>
 */
 
/**
 * <p>Named slot for the oj-option's end icon.</p>
 *
 * @ojslot endIcon
 * @ojshortdesc The slot for the option's end icon.
 * @memberof oj.ojOption
 *
 * @example <caption>Initialize the oj-option with the <code class="prettyprint">endIcon</code> slot specified:</caption>
 * &lt;oj-option>
 *   &lt;span slot='endIcon'>&lt;img src='end.png' alt='End'>&lt;/span>
 * &lt;/oj-option>
 */

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 * 
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 * 
 * @expose
 * @memberof oj.ojOption
 * @instance
 * 
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */ 
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * 
 * @expose
 * @memberof oj.ojOption
 * @instance
 * 
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */ 
/**
 * Refreshes the visual state of the component.
 * 
 * @function refresh
 * @return {void}
 * @expose
 * @memberof oj.ojOption
 * @instance
 */ 
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 * 
 * @expose
 * @memberof oj.ojOption
 * @instance
 * 
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */ 
/**
 * @protected
 * @ignore
 */
(function() {  
  var ojOptionMeta = {
    "properties": {
      "disabled": {
        "type": "boolean",
        "value": false
      }, 
      "value": {
        "type": "any",
        "value": null
      },
      // not documented
      "customOptionRenderer": {}
    },
    "events": {
      "action" : {}
    },
    "extension": {
      _CONSTRUCTOR: ojOption
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-option', null, ojOptionMeta);
  oj.CustomElementBridge.register('oj-option', {'metadata': oj.CustomElementBridge.getMetadata('oj-option')});
})();

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.PagingCellSet
 * @classdesc A CellSet represents a collection of cells.  The CellSet is an object returned by the success callback
 * of the fetchCells method on DataGridDataSource.  The PagingCellSet is a paging specific 
 * implementation of methods on CellSet. 
 * @param {Object} cellSet an cellSet object
 * @param {number} startIndex the true start row index of the cellSet
 * @constructor
 * @since 1.1
 * @export
 * @hideconstructor
 * @ojtsignore
 * @see oj.PagingDataGridDataSource
 */
oj.PagingCellSet = function(cellSet, startIndex)
{
    this.m_cellSet = cellSet;
    this.m_startIndex = startIndex;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available 
 * 2) the index specified is out of bounds. 
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} the data object for the specified index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingCellSet 
 */
oj.PagingCellSet.prototype.getData = function(indexes)
{
    var pagedIndexes = {'column': indexes['column'], 'row': indexes['row'] + this.m_startIndex};
    return this.m_cellSet.getData(pagedIndexes);
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available 
 * 2) the index specified is out of bounds. 
 * @param {Object} indexes the index of each axis in which we want to retrieve the metadata from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return the metadata object for the specific index.  The metadata that the DataGrid supports are: 
 *         1) keys - the key (of each axis) of the cell.
 * @export
 * @expose
 * @memberof oj.PagingCellSet
 * @method
 * @instance
 */
oj.PagingCellSet.prototype.getMetadata = function(indexes)
{
    var pagedIndexes = {'column': indexes['column'], 'row': indexes['row'] + this.m_startIndex};
    return this.m_cellSet.getMetadata(pagedIndexes);
};

/**
 * Gets the actual count of the result set for the specified axis.  Valid values are "row" and "column".
 * @param {string} axis the axis in which to inquire the actual count of the result set.
 * @return {number} the actual count of the result set for the specified axis.  
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingCellSet 
 */
oj.PagingCellSet.prototype.getCount = function(axis)
{
    return this.m_cellSet.getCount(axis);
};

/**
 * For internal testing purposes. Gets the underlying cellSet.
 * @return {Object} the underlying cellSet
 * @export
 * @ignore
 */
oj.PagingCellSet.prototype.getCellSet = function()
{
    return this.m_cellSet;
};

/**
 * For internal testing purposes. Gets the start index.
 * @return {number} the start index
 * @export
 * @ignore
 */
oj.PagingCellSet.prototype.getStartIndex = function()
{
    return this.m_startIndex;
};

/**
 * Gets the extent of a particular row/column index within the context of the cellSet.
 * Extent is defined as the number of indexes along the appropriate axis spanned by the cell.
 * If the extent extends beyond the start and end of the requested cell range the extent should be trimmed to the edge of the requested cell range and the object for {'more': {'before', 'after'}} should have the value appropriate boolean set.
 * @param {Object} indexes the index of each axis in which we want to retrieve the data from.
 * @property {number} indexes.row the index of the row axis.
 * @property {number} indexes.column the index of the column axis.
 * @return {Object} an object containing two properties row and column. Each of those properties has two sub properties:
 *              extent: the number of absolute indexes spanned by the cell at this index
 *                      bounded by the edges of the result set for the specified axis.
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is available in the cellSet
 * @example <caption>In this example the cell spans 5 row indexes and 2 column indexes and there are more column indexes spanned by the cell that
 *              aren't included in this cellSet:</caption>
 * {
 *  'row': {'extent':5, 'more': {'before':false, 'after':false}},
 *  'column': {'extent':2, 'more': {'before':false, 'after':true}}
 * }
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingCellSet
 * @since 4.0.0
 */
oj.PagingCellSet.prototype.getExtent = function(indexes)
{
    var pagedIndexes = {'column': indexes['column'], 'row': indexes['row'] + this.m_startIndex};
    return this.m_cellSet.getExtent(pagedIndexes);
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * @export
 * @class oj.PagingDataGridDataSource
 * @classdesc A paging based implementation of the DataGridDataSource.
 * @param {oj.DataGridDataSource|null} dataSource
 * @param {Object|null} options Options for the PagingControlDataSource
 * @extends oj.DataGridDataSource
 * @implements oj.PagingModel
 * @constructor
 * @since 1.0
 * @ojtsignore
 */
oj.PagingDataGridDataSource = function(dataSource, options)
{
    // Initialize
    if (!(dataSource instanceof oj.DataGridDataSource))
    {
        // we only support Array, oj.Collection, or ko.observableArray. To
        // check for observableArray, we can't do instanceof check because it's
        // a function. So we just check if it contains a subscribe function.
        throw new oj.Message('Not a datagridatasource', 'Not a datagridatasource', oj.Message.SEVERITY_LEVEL['ERROR']);
    }
    this.dataSource = dataSource;
    this._startIndex = 0;
    this.Init();
};

// Subclass from oj.DataSource 
oj.Object.createSubclass(oj.PagingDataGridDataSource, oj.DataGridDataSource, "oj.PagingDataGridDataSource");

/**
 * Initializes the instance.
 * @instance
 * @override
 * @protected
 * @memberof oj.PagingDataGridDataSource
 */
oj.PagingDataGridDataSource.prototype.Init = function()
{
    oj.PagingDataGridDataSource.superclass.Init.call(this);
    this._registerEventListeners();
};

/**
 * Register event handlers on the underlying datasource.
 * @private
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype._registerEventListeners = function()
{
    this.dataSource.on("change", this._handleChange.bind(this));
};

/**
 * Get the current page
 * @return {number} The current page
 * @export
 * @expose
 * @instance
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype.getPage = function()
{
  return this._page;
};

/**
 * Set the current page
 * @param {number} value The current page
 * @param {Object=} options Options<p>
 *                  pageSize: The page size.<p>
 * @returns {Promise} resolves if page is set
 * @export
 * @expose
 * @instance
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype.setPage = function(value, options)
{
  options = options || {};
  value = parseInt(value, 10);
  
  try 
  {
    oj.PagingDataGridDataSource.superclass.handleEvent.call(this, oj.PagingModel.EventType['BEFOREPAGE'], {'page' : value, 'previousPage' : this._page});
  }
  catch (err)
  {
    return Promise.reject(null);
  }
  
  this._pageSize = options['pageSize'] != null ? options['pageSize'] : this._pageSize;
  options['startIndex'] = value * this._pageSize;
  var previousPage = this._page;
  this._page = value;
  this._startIndex = options['startIndex'];
  var self = this;
  
  return new Promise(function(resolve, reject)
  {
    self._fetchInternal(options).then(function(result)
    {
      resolve(null);
    }, function (error)
    {
      // restore old page
      self._page = previousPage;
      self._startIndex = self._page * self._pageSize;
      reject(null);  
    });
  });
};

/**
 * Calls fetch on the datasource with paging options.
 * @private
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype._fetchInternal = function(options)
{
    this._initialized = true;
    this._startIndex = options['startIndex'];
    
    var self = this;
    return new Promise(function(resolve, reject) {
        self.handleEvent("change", {'operation': 'sync', 'pageSize': self._pageSize});  
        resolve(undefined);
    });
};

/**
 * Calls fetch on the datasource with paging options.
 * @param {Object=} options Options to control fetch
 * @property {number} options.startIndex The index at which to start fetching records.
 * @property {boolean} options.silent If set, do not fire a sync event.
 * @return {Promise} Promise object resolves when done
 * @export
 * @expose
 * @memberof oj.PagingDataGridDataSource
 * @instance
 */
oj.PagingDataGridDataSource.prototype.fetch = function(options)
{
    this._pageSize = options['pageSize'] + options['startIndex'];
    options['startIndex'] = 0;
    return this._fetchInternal(options);
};

/**
 * Get the current page start index
 * @return {number} The current page start index
 * @export
 * @expose
 * @instance
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype.getStartItemIndex = function()
{
  return this._startIndex;
};

/**
 * Get the current page end index
 * @return {number} The current page end index
 * @export
 * @expose
 * @instance
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype.getEndItemIndex = function()
{
  return this._endIndex;
};

/**
 * Get the page count
 * @return {number} The total number of pages
 * @export
 * @expose
 * @instance
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype.getPageCount = function()
{
  var totalSize = this.totalSize();
  return totalSize == -1 ? -1 : Math.ceil(totalSize / this._pageSize);
};

/**
 * Handle data grid change events
 * @param {Object} options the options associated with the oj.DataGridDataSource event.
 * @private
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype._handleChange = function(options) {
    var operation;
    operation = options['operation'];
    switch(operation)
    {
        case 'refresh':
            this._startIndex = 0;
            this._page = 0;            
            // pass the refresh event through to the data grid and the paging control
            this.handleEvent("change", {'operation': 'sync', 'pageSize': this._pageSize});  
            this.handleEvent(oj.PagingTableDataSource.EventType['REFRESH'], null);
            break;
        case 'reset':
            // the paging control will set a new startIndex and kick off a fetch here
            this.handleEvent(oj.PagingTableDataSource.EventType['RESET'], null);
            break;
        case 'insert':
            this.handleEvent(oj.PagingTableDataSource.EventType['ADD'], {'index':options['indexes']['row']});
            break;
        case 'delete':
            this.handleEvent(oj.PagingTableDataSource.EventType['REMOVE'], null);
            break;
        case 'update':
            options['indexes']['row'] = options['indexes']['row'] - this._startIndex >= 0 ? options['indexes']['row'] - this._startIndex : -1;
            this.handleEvent("change", options);
            break;
        default:
            this.handleEvent('change', options);
            this.handleEvent(oj.PagingTableDataSource.EventType['SYNC'], null);
    }
};

/**** start delegated functions ****/
/**
 * Returns the total number of rows or columns.  If the value return is not >= 0 then it is automatically assumed
 * that the total count is unknown. In the case of paging returns the total number of rows/columns on the page.
 * @param {string} axis the axis in which we inquire for the total count.
 * @return {number} the total number of rows/columns.
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource */
oj.PagingDataGridDataSource.prototype.getCount = function(axis)
{
    var count = this.dataSource.getCount(axis);
    if (axis === 'row' && count >= 0)
    {
        if (this._startIndex + this._pageSize < count)
        {
            return this._pageSize;
        }
        return count - this._startIndex;
    }
    return count;
};

/**
 * Returns whether the total count returned in getCount function is an actual or an estimate.
 * @param {string} axis the axis in which we inquire whether the total count is an estimate.  Valid values are 
 *        "row" and "column".
 * @return {string} "actual" if the count returned in getCount function is the actual count, "estimate" if the 
 *         count returned in getCount function is an estimate.
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource
 */
oj.PagingDataGridDataSource.prototype.getCountPrecision = function(axis)
{
    return this.dataSource.getCountPrecision(axis);
};

/**
 * Fetch a range of headers from the data source.
 * @param {Object} headerRange information about the header range, it must contain the following properties:
 *        axis, start, count.
 * @property {string} headerRange.axis the axis of the header that are fetched.  Valid values are "row" and "column".
 * @property {number} headerRange.start the start index of the range in which the header data are fetched.
 * @property {number} headerRange.count the size of the range in which the header data are fetched.  
 * @param {Object} callbacks the callbacks to be invoke when fetch headers operation is completed.
 * @property {function({startHeaderSet: Object}, {headerRange: Object}, {endHeaderSet: Object}):undefined} callbacks.success the callback to invoke when fetch headers completed successfully.
 *        The function takes three parameters: HeaderSet object representing start headers, headerRange object passed into the original fetchHeaders call,
 *        and a HeaderSet object representing the end headers along the axis.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource 
 * @return {undefined}
 */
oj.PagingDataGridDataSource.prototype.fetchHeaders = function(headerRange, callbacks, callbackObjects)
{
    var headerSet;
    if (this._initialized == null)
    {
        if (callbacks != null && callbacks['success'])
        {
            callbacks['success'].call(callbackObjects['success'], null, headerRange, null);
        }
    }
    else
    {
        if (headerRange['axis'] === 'row')
        {
            headerRange['start'] += this._startIndex;            
            if (headerRange['start'] + headerRange['count'] > this._startIndex + this._pageSize)
            {
                headerRange['count'] = this._pageSize - headerRange['start'];
            }
            this._pendingRowHeaderCallback = {'headerRange': headerRange, 'callbacks': callbacks, 'callbackObjects': callbackObjects};
            this.dataSource.fetchHeaders(headerRange, {success: this._handleRowHeaderFetchSuccess.bind(this), error: this._handleRowHeaderFetchError.bind(this)}, callbackObjects);
        }
        else
        {
            this.dataSource.fetchHeaders(headerRange, callbacks, callbackObjects);
        }
    }
};

/**
 * Handle row headers fetch success by adjusting startIndex back to 0 and passing a PagingHeaderSet
 * @param {Object} headerSet a cellSet object
 * @param {Object} headerRange
 * @param {Object} endHeaderSet a headerSet object
 * @private
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype._handleRowHeaderFetchSuccess = function(headerSet, headerRange, endHeaderSet)
{
    var pagingHeaderSet, pagingEndHeaderSet, callback, callbackObject;
    
    // make sure we only callback when we have the most recent request
    if (this._pendingRowHeaderCallback['headerRange'] != headerRange)
    {
        return;
    }
    
    headerRange['start'] -= this._startIndex;
    // make the datagrid not issue any extra fetches
    headerRange['count'] += 1;
    if (headerSet != null)
    {
        pagingHeaderSet = new oj.PagingHeaderSet(headerSet, this._startIndex);
    }    
    if (endHeaderSet != null)
    {
        pagingEndHeaderSet = new oj.PagingHeaderSet(endHeaderSet, this._startIndex);
    }
    
    // clear callback before calling it because it may issue a refetch
    callback = this._pendingRowHeaderCallback['callbacks']['success'];
    callbackObject = this._pendingRowHeaderCallback['callbackObjects']['success'];
    this._pendingRowHeaderCallback = null;
    callback.call(callbackObject, pagingHeaderSet, headerRange, pagingEndHeaderSet);
};

/**
 * Handle row header fetch error
 * @param {Object} error error
 * @private
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype._handleRowHeaderFetchError = function(error)
{
    var callback, callbackObject;
    // clear callback before calling it because it may issue a refetch
    callback = this._pendingRowHeaderCallback['callbacks']['error'];
    callbackObject = this._pendingRowHeaderCallback['callbackObjects']['error'];
    this._pendingRowHeaderCallback = null;
    callback.call(callbackObject, error);
};

/**
 * Fetch a range of cells from the data source.
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @property {number} cellRanges.count the size of the range for this axis in which the cells are fetched. 
 * @param {Object} callbacks the callbacks to be invoke when fetch cells operation is completed.
 * @property {function({cellSet: Object}, {cellRanges: Array.<Object>}):undefined} callbacks.success the callback to invoke when fetch cells completed successfully.
 * @property {function({status: Object}):undefined} callbacks.error the callback to invoke when fetch cells failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource
 */
oj.PagingDataGridDataSource.prototype.fetchCells = function(cellRanges, callbacks, callbackObjects)
{
    var i; 
    if (this._initialized == null)
    {
        var emptyCellSet = {};
        emptyCellSet['getData'] = function() { return null; };
        emptyCellSet['getMetaData'] = function() { return null; };
        emptyCellSet['getStart'] = function() { return 0; };
        emptyCellSet['getCount'] = function() { return 0; };
        emptyCellSet['getLevelCount'] = function() { return 0; };
        emptyCellSet['getExtent'] = function() { return 0; };
        emptyCellSet['getDepth'] = function() { return 1; };       
        
        if (callbacks != null && callbacks['success'])
        {
            callbacks['success'].call(callbackObjects['success'], emptyCellSet, cellRanges);
        }
    }
    else
    {
        for (i=0; i<cellRanges.length; i+=1)
        {
            if (cellRanges[i]['axis'] === 'row')
            {
                cellRanges[i]['start'] += this._startIndex;
                if (cellRanges[i]['start'] + cellRanges[i]['count'] > this._startIndex + this._pageSize)
                {
                    cellRanges[i]['count'] = this._pageSize - cellRanges[i]['start'];
                }
            }        
        }
        this._pendingCellCallback = {'cellRanges': cellRanges, 'callbacks': callbacks, 'callbackObjects': callbackObjects};
        this.dataSource.fetchCells(cellRanges, {success: this._handleCellsFetchSuccess.bind(this), error: this._handleCellsFetchError.bind(this)}, callbackObjects);
    }
};

/**
 * Handle cell fetch success by adjusting the row startIndex and passing the PagingCellSet
 * @param {Object} cellSet a cellSet object
 * @param {Array.<Object>} cellRanges Information about the cell range.  A cell range is defined by an array 
 *        of range info for each axis, where each range contains three properties: axis, start, count.
 * @property {string} cellRanges.axis the axis associated with this range where cells are fetched.  Valid 
 *        values are "row" and "column".
 * @property {number} cellRanges.start the start index of the range for this axis in which the cells are fetched.
 * @private
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype._handleCellsFetchSuccess = function(cellSet, cellRanges)
{
    var i, pagedCellSet, callback, callbackObject;
    
    // make sure we only callback when we have the most recent request    
    if (!this._pendingCellCallback['cellRanges'] == cellRanges)
    {
        return;
    }
    
    for (i=0; i<cellRanges.length; i+=1)
    {
        if (cellRanges[i]['axis'] === 'row')
        {
            cellRanges[i]['start'] -= this._startIndex;
            // make the datagrid not issue any extra fetches
            cellRanges[i]['count'] += 1;
        }        
    }
    pagedCellSet = new oj.PagingCellSet(cellSet, this._startIndex);
    
    // clear callback before calling it because it may issue a refetch
    callback = this._pendingCellCallback['callbacks']['success'];
    callbackObject = this._pendingCellCallback['callbackObjects']['success'];
    this._pendingCellCallback = null;    
    
    this._endIndex = this._startIndex + cellSet.getCount('row') - 1;
    // tell PC fetchEnd
    this.handleEvent('sync', {'data': new Array(cellSet.getCount('row')), 'startIndex': this._startIndex});
    
    callback.call(callbackObject, pagedCellSet, cellRanges);
};

/**
 * Handle a cell fetch error
 * @param {Object} error error
 * @private
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype._handleCellsFetchError = function(error)
{
    var callback, callbackObject;
    // clear callback before calling it because it may issue a refetch
    callback = this._pendingCellCallback['callbacks']['error'];
    callbackObject = this._pendingCellCallback['callbackObjects']['error'];
    this._pendingCellCallback = null;
    callback.call(callbackObject,error);    
};

/**
 * Returns the keys based on the indexes.
 * @param {Object} indexes the index for each axis
 * @property {number|null} indexes.row the index for the row axis
 * @property {number|null} indexes.column the index for the column axis
 * @return {Promise.<Object>} a Promise object which upon resolution will pass in an object containing the keys for each axis
 * @export
 * @method
 * @instance
 * @memberof oj.ArrayDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype.keys = function(indexes)
{
    var pagedIndexes = {'column': indexes['column'], 'row': indexes['row'] + this._startIndex};
    return this.dataSource.keys(pagedIndexes);
};

/**
 * Returns the row and column index based on the keys. In a paging case returns the 
 * index on the page, not the absolute index in the array.
 * @param {Object} keys the key for each axis
 * @property {*} keys.row the key for the row axis
 * @property {*} keys.column the key for the column axis
 * @return {Promise.<Object>} a promise object containing the index for each axis
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource 
 */
oj.PagingDataGridDataSource.prototype.indexes = function(keys)
{
    var indexes = this.dataSource.indexes(keys);
    if (indexes['row'] != -1)
    {
        indexes['row'] -= this._startIndex;
    }
    return indexes;
};

/**
 * Determines whether this PagingDataGridDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort" and "move"
 * @return {string|null} the name of the feature.  For sort, the valid return values are: "full", "none".  Returns null if the
 *         feature is not recognized.
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource
*/
oj.PagingDataGridDataSource.prototype.getCapability = function(feature)
{
    return this.dataSource.getCapability(feature);
};

/**
 * @export
 * Return the size of the data locally in the dataSource. -1 if an initial fetch has not been
 * done yet.
 * @returns {number} size of data
 * @method
 * @memberof oj.PagingDataGridDataSource
 * @instance
 */
oj.PagingDataGridDataSource.prototype.size = function()
{
    var count;
    if (this._initialized == null)
    {
        return -1;
    }
    count = this.dataSource.getCount('row')
    if (this.dataSource.getCount('row') > this._pageSize)
    {
        return this._pageSize;
    }
    return count;
};

/**
 * Performs a sort on the data source.
 * @param {Object|null} criteria the sort criteria. 
 * @property {string} criteria.axis The axis in which the sort is performed, valid values are "row", "column"
 * @property {*} criteria.key The key that identifies which header to sort
 * @property {string} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the sort operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.  
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource
 * @return {undefined}
 */
oj.PagingDataGridDataSource.prototype.sort = function(criteria, callbacks, callbackObjects)
{
    this.dataSource.sort(criteria, callbacks, callbackObjects);
};

/**
 * @export
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @method
 * @memberof oj.PagingDataGridDataSource
 * @instance
 */
oj.PagingDataGridDataSource.prototype.totalSize = function()
{
    if (this._initialized == null)
    {
        return -1;
    }
    return this.dataSource.getCount('row');
};

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @export
 * @method
 * @memberof oj.PagingDataGridDataSource
 * @instance 
 */
oj.PagingDataGridDataSource.prototype.totalSizeConfidence = function()
{ 
  return "actual";
};

/**
 * Checks whether a move operation is valid.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine
 *        the destination of where the row should moved to.
 * @param {string} position The position of the moved row relative to the reference row.  
 *        Valid values are: "before", "after".
 * @return {string} returns "valid" if the move is valid, "invalid" otherwise.
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource 
*/
oj.PagingDataGridDataSource.prototype.moveOK = function(rowToMove, referenceRow, position)
{
    return this.dataSource.moveOK(rowToMove, referenceRow, position);
};

/**
 * Moves a row from one location to another.
 * @param {*} rowToMove the key of the row to move
 * @param {*} referenceRow the key of the reference row which combined with position are used to determine 
 *        the destination of where the row should moved to.
 * @param {string} position The position of the moved row relative to the reference row.
 *        Valid values are: "before", "after" 
 * @param {Object=} callbacks the callbacks to be invoke upon completion of the move operation.
 * @property {function():undefined=} callbacks.success the callback to invoke when the sort completed successfully.
 * @property {function({status: Object}):undefined=} callbacks.error the callback to invoke when sort failed.
 * @param {Object=} callbackObjects the object in which the callback function is invoked on.
 * @property {Object=} callbackObjects.success
 * @property {Object=} callbackObjects.error
 * @export
 * @method
 * @instance
 * @memberof oj.PagingDataGridDataSource
 * @return {undefined}
 */
oj.PagingDataGridDataSource.prototype.move = function(rowToMove, referenceRow, position, callbacks, callbackObjects)
{
    this.dataSource.move(rowToMove, referenceRow, position, callbacks, callbackObjects);
};
/**** end delegated functions ****/

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.PagingHeaderSet
 * @classdesc A HeaderSet represents a collection of headers. The HeaderSet is an object returned by the success callback
 * of the fetchHeaders method on DataGridDataSource. This is a paging specific implementation of the HeaderSet.
 * @param {Object} headerSet an headerSet object
 * @param {number} startIndex the true startIndex of the headerSet
 * @constructor
 * @since 1.1
 * @hideconstructor
 * @export
 * @ojtsignore
 * @see oj.PagingDataGridDataSource
 */
oj.PagingHeaderSet = function(headerSet, startIndex)
{
    this.m_headerSet = headerSet;
    this.m_startIndex = startIndex;
};

/**
 * Gets the data of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds.
 * @param {number} index the absolute index of the header in which we want to retrieve the header from.
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {*} the data for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof oj.PagingHeaderSet
 */
oj.PagingHeaderSet.prototype.getData = function(index, level)
{
    return this.m_headerSet.getData(index + this.m_startIndex, level);
};

/**
 * Gets the metadata of the specified index.  An error is throw when 1) the range is not yet available and
 * 2) the index specified is out of bounds.
 * The metadata that the data source can optionally return are:
 *  1) sortDirection - the initial sort direction of the header.  Valid values are "ascending" and "descending".
 *  2) key - the key of the row/column header.
 * @param {number} index the absolute index of the header in which we want to retrieve the metadata from.
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} the metadata object for the specific index.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingHeaderSet 
 */
oj.PagingHeaderSet.prototype.getMetadata = function(index, level)
{
    return this.m_headerSet.getMetadata(index + this.m_startIndex, level);
};

/**
 * Gets the actual count of the result set, the total indexes spanned by the headerSet
 * along the innermost header.
 * @return {number} the actual count of the result set.
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingHeaderSet 
 */
oj.PagingHeaderSet.prototype.getCount = function()
{
    return this.m_headerSet.getCount();
};

/**
 * Gets the actual number of levels of the result set for the specified axis. The levels
 * are the counted from the outermost header indexed at 0, and moving inwards toward the
 * databody would increment the level by 1.
 * @return {number} the number of levels of the result set
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingHeaderSet 
 */
oj.PagingHeaderSet.prototype.getLevelCount = function()
{
     return this.m_headerSet.getLevelCount();
};
 
/**
 * Gets the extent of an index on a particular level within the context of the headerSet. Extent is defined as the
 * number of indexes covered by the header. If the extent extends beyond the start and end of the requested
 * range the extent should be trimmed to the edge of the requested range and the object for {'more': {before, after}}
 * should have the value appropriate boolean set. For innermost headers the extent will always be 1.
 * @param {number} index the absolute index along the innermost header of the extent to get, 0 is the first header in the data source
 * @param {number=} level the level of the header, 0 is the outermost header and increments by 1 moving inward
 * @return {Object} an object containing two values
 *              extent: the number of absolute indexes spanned by the header at this index
 *                      bounded by the edges of the result set for the specified axis.
 *              more: object with keys 'before'/'after' and boolean values true/false representing whether
 *                       there are more indexes before/after what is in the headerSet
 * @example <caption>In this example the header spans 5 indexes and there are more indexes to cover after the request that
 *              aren't included in this headerSet:</caption>
 * {'extent':5, 'more': {'before':false, 'after':true}}
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingHeaderSet 
 */
oj.PagingHeaderSet.prototype.getExtent = function(index, level)
{
     return this.m_headerSet.getExtent(index + this.m_startIndex, level);
};
 
/**
 * Gets the depth of an index starting at a particular level. The depth is the number
 * of levels spanned by the header.
 * @param {number} index the absolute index of the depth to get
 * @param {number=} level the level of the header, 0 is the outermost header
 * @return {number} the number of levels spanned by the header at the specified position
 * @export
 * @expose
 * @method
 * @instance
 * @memberof! oj.PagingHeaderSet 
 */
oj.PagingHeaderSet.prototype.getDepth = function(index, level)
{
     return this.m_headerSet.getDepth(index + this.m_startIndex, level);
};

/**
 * For internal testing purposes. Gets the underlying headerSet.
 * @return {Object} the underlying headerSet
 * @export
 * @ignore
 */
oj.PagingHeaderSet.prototype.getHeaderSet = function()
{
    return this.m_headerSet;
};

/**
 * For internal testing purposes. Gets the start index.
 * @return {number} the start index
 * @export
 * @ignore
 */
oj.PagingHeaderSet.prototype.getStartIndex = function()
{
    return this.m_startIndex;
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * The interface for oj.PagingModel which should be implemented by all object instances
 * bound to the data parameter for oj.PagingControl. oj.PagingModel implementations should
 * also support event subscription by extending oj.EventSource or oj.DataSource.
 * @export
 * @interface
 * @since 1.1
 * @ojtsignore
 */
oj.PagingModel = function()
{
};

/**
 * Get the current page
 * @return {number} The current page
 * @export
 * @expose
 * @method
 * @name getPage
 * @memberof oj.PagingModel
 * @instance
 */

/**
 * Set the current page
 * @param {number} value The current page
 * @param {Object=} options Options<p>
 *                  pageSize: The page size.<p>
 * @return {Promise} promise object triggering done when complete..
 * @export
 * @expose
 * @method
 * @name setPage
 * @memberof oj.PagingModel
 * @instance
 */

/**
 * Get the current page start index
 * @return {number} The current page start index
 * @export
 * @expose
 * @method
 * @name getStartItemIndex
 * @memberof oj.PagingModel
 * @instance
 */

/**
 * Get the current page end index
 * @return {number} The current page end index
 * @export
 * @expose
 * @method
 * @name getEndItemIndex
 * @memberof oj.PagingModel
 * @instance
 */

/**
 * Get the page count
 * @return {number} The total number of pages
 * @export
 * @expose
 * @method
 * @name getPageCount
 * @memberof oj.PagingModel
 * @instance
 */

/**
 * @export
 * Return the total number of items. Returns -1 if unknown.
 * @returns {number} total number of items
 * @expose
 * @method
 * @name totalSize
 * @memberof oj.PagingModel
 * @instance
 */

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @export
 * @expose
 * @method
 * @name totalSizeConfidence
 * @memberof oj.PagingModel
 * @instance 
 */

/**
 * @export
 * Event types
 * @memberof oj.PagingModel
 * @enum {string}
 */
oj.PagingModel.EventType =
  {
    /** Triggered before the current page has changed. <p>
     * This event is vetoable.<p>
     * The event payload contains:<p>
     * <b>page</b> The new current page<br>
     * <b>previousPage</b> The old current page
     */
    'BEFOREPAGE': "beforePage",
    /** Triggered when the current page has changed<p>
     * The event payload contains:<p>
     * <b>page</b> The new current page<br>
     * <b>previousPage</b> The old current page
     */
    'PAGE': "page",
    /** Triggered when the page count has changed<p>
     * The event payload contains:<p>
     * <b>pageCount</b> The new page count<br>
     * <b>previousPageCount</b> The old page count
     */
    'PAGECOUNT': "pageCount"
  };


/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @export
 * @class oj.PagingTableDataSource
 * @classdesc Object representing data that supports paging functionality.  This data source can be used by [ListView]{@link oj.ojListView}, [PagingControl]{@link oj.ojPagingControl}, and [Table]{@link oj.ojTable}.
 *            <p>This class implements the {@link oj.PagingModel} interface and 
 *            publish all of the oj.PagingModel event types, which can be
 *            listened to using the [on()]{@link oj.PagingTableDataSource#on}
 *            method.</p>
 *            See the <a href="../jetCookbook.html?component=pagingControl&demo=basicPagingTable">Paging Control - Paging Table</a> demo for an example.<br><br>
 *            Refer to {@link oj.TableDataSource} for other data sources that represent tabular data.
 * @extends oj.TableDataSource
 * @implements oj.PagingModel
 * @param {Object} dataSource
 * @param {Object|null} options Options for the PagingTableDataSource.  No option is currently supported.
 * @constructor
 * @since 1.0
 * @ojtsignore
 */
oj.PagingTableDataSource = function(dataSource, options)
{
  // Initialize
  options = options || {};
  
  if (!(dataSource instanceof oj.TableDataSource))
  {
    // we only support Array, oj.Collection, or ko.observableArray. To
    // check for observableArray, we can't do instanceof check because it's
    // a function. So we just check if it contains a subscribe function.
    var errSummary = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_SUMMARY'];
    var errDetail = oj.TableDataSource._LOGGER_MSG['_ERR_DATA_INVALID_TYPE_DETAIL'];
    throw new Error(errSummary + '\n' + errDetail);
  }
  this.dataSource = dataSource;
  this._startIndex = 0;
  this._endIndex = -1;
  this._dataSourceWrappedEventHandlers = [];
  this.Init();

  // PagingTableDataSource doesn't need its own copy of sortCriteria and can
  // just act as a proxy for the underlying dataSource
  Object.defineProperty(this, 'sortCriteria',
  {
    'configurable': false,
    'enumerable': true,
    'get': function() { return this.dataSource['sortCriteria']; },
    'set': function(newValue) { this.dataSource['sortCriteria'] = newValue; }
  });
};

// Subclass from oj.DataSource 
oj.Object.createSubclass(oj.PagingTableDataSource, oj.TableDataSource, "oj.PagingTableDataSource");

/**
 * Initializes the instance.
 * @memberof oj.PagingTableDataSource
 * @instance
 * @override
 * @protected
 */
oj.PagingTableDataSource.prototype.Init = function()
{
  oj.PagingTableDataSource.superclass.Init.call(this);
};

/**
 * Retrieves the underlying DataSource.
 * @return {Object} the underlying oj.DataSource.
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.getWrappedDataSource = function()
{
    return this.dataSource;
};

/**
 * Get the current page
 * @return {number} The current page
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.getPage = function()
{
  if (this._fetchType == 'loadMore')
  {
    return 0;
  }
  return this._getPageFromStartIndex();
};

/**
 * Set the current page
 * @param {number} value The current page
 * @param {Object=} options Options<p>
 *                  pageSize: The page size.<p>
 * @return {Promise} promise object triggering done when complete..
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.setPage = function(value, options)
{
  options = options || {};
  value = parseInt(value, 10);
  
  try 
  {
    oj.PagingTableDataSource.superclass.handleEvent.call(this, oj.PagingModel.EventType['BEFOREPAGE'], {'page' : value, 'previousPage' : this._getPageFromStartIndex()});
  }
  catch (err)
  {
    return Promise.reject(err);
  }
  var previousPage = this._getPageFromStartIndex();
  this._pageSize = options['pageSize'] != null ? options['pageSize'] : this._pageSize;
  options['pageSize'] = this._pageSize;
  options['startIndex'] = value * this._pageSize;
  this._startIndex = options['startIndex'] == null ? this._startIndex : options['startIndex'];
  this._fetchType = 'page';
  var self = this;
  return new Promise(function(resolve, reject) 
  {
    // we only support paged fetches
    if (self._pageSize > 0)
    {
      self.dataSource.fetch(options).then(function(result)
      {
        result['startIndex'] = 0;
        if (result['data'].length > 0)
        {
          self._updateEndIndex(self._startIndex + result['data'].length - 1, true);
        }
        else
        {
          self._updateEndIndex(-1, true);
        }
        oj.PagingTableDataSource.superclass.handleEvent.call(self, oj.PagingModel.EventType['PAGE'], {'page' : self._getPageFromStartIndex(), 'previousPage' : previousPage});
        resolve(null);
      },
      function(e)
      {
        self._startIndex = previousPage * self._pageSize;
        reject(e); 
      });
    }
    else
    {
      resolve(null);
    }
  });
};

/**
 * Get the current page start index
 * @return {number} The current page start index
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.getStartItemIndex = function()
{
  if (this._fetchType == 'loadMore')
  {
    return 0;
  }
  return this._startIndex;
};

/**
 * Get the current page end index. If there are no rows in the current page then
 * return -1.
 * @return {number} The current page end index
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.getEndItemIndex = function()
{
  return this._endIndex;
};

/**
 * Get the page count
 * @return {number} The total number of pages
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.getPageCount = function()
{
  var totalSize = this.totalSize();
  return totalSize == -1 ? -1 : Math.ceil(totalSize / this._pageSize);
};

/**** start delegated functions ****/

/**
 * Return the row data found at the given index.
 * 
 * @param {number} index Index for which to return the row data. 
 * @param {Object=} options Options to control the at.
 * @param {number} options.fetchSize fetch size to use if the call needs to fetch more records from the server, if virtualized.  Overrides the overall fetchSize setting <p>
 * @return {Promise} Promise resolves to a compound object which has the structure below. If the index is out of range, Promise resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.at = function(index, options)
{
  return this.dataSource.at(index, options);
};

/**
 * Fetch the row data.
 * @param {Object=} options Options to control fetch
 * @param {number} options.startIndex The index at which to start fetching records.
 * @param {boolean} options.silent If set, do not fire a sync event.
 * @return {Promise} Promise object resolves to a compound object which contains an array of row data objects, an array of ids, and the startIndex triggering done when complete.<p>
 *         The structure of the resolved compound object is:<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>An array of raw row data</td></tr>
 * <tr><td><b>keys</b></td><td>An array of key values for the rows</td></tr>
 * <tr><td><b>startIndex</b></td><td>The startIndex for the returned set of rows</td></tr>
 * </tbody>
 * </table>  
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.fetch = function(options)
{
  options = options || {};
  if (options['startIndex'] == null)
  {
    // this is just a refresh call from ojTable. So treat it as a paged fetch
    return this.setPage(this.getPage());
  }
  else
  {
    this._fetchType = 'loadMore';
  }
  this._startIndex = options['startIndex'] == null ? this._startIndex : options['startIndex'];
  var pageSize = options['pageSize'] == null ? this._pageSize : options['pageSize'];
  
  if (this._pageSize == null)
  {
    this._pageSize = pageSize;
  }
  
  options['pageSize'] = pageSize;
  options['startIndex'] = this._startIndex;
  
  var self = this;
  return new Promise(function(resolve, reject) 
  {
    // we only support paged fetches
    if (pageSize > 0)
    {
      self.dataSource.fetch(options).then(function(result)
      {
        if (result['data'].length > 0)
        {
          self._updateEndIndex(self._startIndex + result['data'].length - 1, true);
        }
        else
        {
          self._updateEndIndex(-1, true);
        }
        resolve(result);
      },
      function(e)
      {
        reject(e);
      });
    }
    else
    {
      resolve(null);
    }
  });
};

/**
 * Return the first row data whose id value is the given id
 * @param {string} id ID for which to return the row data, if found. 
 * @param {Object=} options Options to control the get.
 * @return {Promise} Promise which resolves to a compound object which has the structure below where the id matches the given id. If none are found, resolves to null.<p>
 * <table>
 * <tbody>
 * <tr><td><b>data</b></td><td>The raw row data</td></tr>
 * <tr><td><b>index</b></td><td>The index for the row</td></tr>
 * <tr><td><b>key</b></td><td>The key value for the row</td></tr>
 * </tbody>
 * </table>
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.get = function(id, options)
{
  return this.dataSource.get(id, options);
};

/**
 * Determines whether this TableDataSource supports certain feature.
 * @param {string} feature the feature in which its capabilities is inquired.  Currently the only valid feature is "sort".
 * @return {string|null} the name of the feature.  For "sort", the valid return values are: "full", "none".  
 *         Returns null if the feature is not recognized.
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.getCapability = function(feature)
{
  return this.dataSource.getCapability(feature);
};

/**
 * Attach an event handler to the datasource
 * @param {string} eventType eventType supported by the datasource.
 *        <p>In addition to this class's event types, it can also be one of the 
 *           {@link oj.PagingModel} event types.</p>
 * @param {function(Object)} eventHandler event handler function
 * @return {void}
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.on = function(eventType, eventHandler)
{
  var self = this;
  var dataSource = (/** @type {{on: Function}} */ (this.dataSource));

  if (eventType == oj.TableDataSource.EventType['SYNC'])
  {
    var ev = function(event){self._handleSyncEvent(event, eventHandler);}
    this._dataSourceWrappedEventHandlers.push({'eventType': eventType, 'eventHandler': eventHandler, 'wrappedEventHandler': ev});
    dataSource.on(eventType, ev);
  }
  else if (eventType == oj.TableDataSource.EventType['ADD'] ||
           eventType == oj.TableDataSource.EventType['REMOVE'] ||
           eventType == oj.TableDataSource.EventType['CHANGE'])
  {
    var ev = function(event){self._handleRowEvent(event, eventHandler);}
    this._dataSourceWrappedEventHandlers.push({'eventType': eventType, 'eventHandler': eventHandler, 'wrappedEventHandler': ev});
    dataSource.on(eventType, ev);
  }
  else if (eventType == oj.TableDataSource.EventType['REFRESH'] ||
           eventType == oj.TableDataSource.EventType['RESET'])
  {
    var ev = function(event){
      self._startIndex = 0;
      eventHandler(event);
    }
    this._dataSourceWrappedEventHandlers.push({'eventType': eventType, 'eventHandler': eventHandler, 'wrappedEventHandler': ev});
    dataSource.on(eventType, ev);
  }
  else if (eventType == oj.PagingModel.EventType['PAGE'] ||
           eventType == oj.PagingModel.EventType['BEFOREPAGE'] ||
           eventType == oj.PagingModel.EventType['PAGECOUNT'])
  {
    oj.PagingTableDataSource.superclass.on.call(this, eventType, eventHandler);
  }
  else
  {
    dataSource.on(eventType, eventHandler);
  }
};

/**
 * Detach an event handler from the datasource
 * @param {string} eventType eventType supported by the datasource
 * @param {function(Object)} eventHandler event handler function
 * @return {void}
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.off = function(eventType, eventHandler)
{
  if (eventType == oj.PagingModel.EventType['PAGE'] ||
      eventType == oj.PagingModel.EventType['PAGECOUNT'])
  {
    oj.PagingTableDataSource.superclass.off.call(this, eventType, eventHandler);
  }
  var dataSource = (/** @type {{off: Function}} */ (this.dataSource));
  
  if (this._dataSourceWrappedEventHandlers != null)
  {
    var dataSourceWrappedEventHandlersCount = this._dataSourceWrappedEventHandlers.length;
    var i;
    for (i = 0; i < dataSourceWrappedEventHandlersCount; i++) 
    {
      if (this._dataSourceWrappedEventHandlers[i]['eventType'] == eventType &&
          this._dataSourceWrappedEventHandlers[i]['eventHandler'] == eventHandler)
      {
        dataSource.off(eventType, this._dataSourceWrappedEventHandlers[i]['wrappedEventHandler']);
        this._dataSourceWrappedEventHandlers.splice(i, 1);
        break;
      }
    }
  }
  dataSource.off(eventType, eventHandler);
};

/**
 * Performs a sort on the data source.
 * @param {Object} criteria the sort criteria.
 * @param {Object} criteria.key The key that identifies which field to sort
 * @param {string} criteria.direction the sort direction, valid values are "ascending", "descending", "none" (default)
 * @return {Promise} promise object triggering done when complete.
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.sort = function(criteria)
{
  return this.dataSource.sort(criteria);
};

/**
 * Return the total size of data available, including server side if not local.
 * @returns {number} total size of data
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance
 */
oj.PagingTableDataSource.prototype.totalSize = function()
{
  return this.dataSource.totalSize();
};

/**
 * Returns the confidence for the totalSize value. 
 * @return {string} "actual" if the totalSize is the time of the fetch is an exact number 
 *                  "estimate" if the totalSize is an estimate 
 *                  "atLeast" if the totalSize is at least a certain number 
 *                  "unknown" if the totalSize is unknown
 * @ojsignature {target: "Type", for: "returns", value: "'actual'|'estimate'|'atLeast'|'unknown'"}
 * @export
 * @expose
 * @memberof oj.PagingTableDataSource
 * @instance 
 */
oj.PagingTableDataSource.prototype.totalSizeConfidence = function()
{ 
  return this.dataSource.totalSizeConfidence();
};

/**** end delegated functions ****/

oj.PagingTableDataSource.prototype._getPageFromStartIndex = function()
{
  if (this._pageSize > 0)
  {
    return Math.floor(this._startIndex / this._pageSize);
  }
  return 0;
}

oj.PagingTableDataSource.prototype._handleRowEvent = function(event, eventHandler)
{
  var ignoreRows = [];
  var i;
  for (i = 0; i < event['indexes'].length; i++)
  {
    var rowIdx = event['indexes'][i];
    if (rowIdx !== undefined) 
    {
      // adjust by startIndex
      if (this._fetchType == 'page')
      {
        rowIdx = rowIdx - this._startIndex;
      }
      
      if (rowIdx < 0 || rowIdx >= this._startIndex + this._pageSize)
      {
        // ignore if it's not in our page range
        ignoreRows.push(i);
      }
    }
  }
  
  if (ignoreRows.length > 0)
  {
    ignoreRows.sort(function(a, b) 
    { 
      return a - b;
    });
    
    for (i = ignoreRows.length - 1; i >= 0; i--)
    {
      event['data'].splice(ignoreRows[i], 1);
      event['indexes'].splice(ignoreRows[i], 1);
      event['keys'].splice(ignoreRows[i], 1);
    }
  }
  
  if (event['indexes'].length > 0)
  {
    this._updateEndIndex(event['indexes'][event['indexes'].length - 1], false);
  }
  
  event['startIndex'] = this._startIndex;
  eventHandler(event);
};

oj.PagingTableDataSource.prototype._handleSyncEvent = function(event, eventHandler)
{
  var eventStartIndex = event['startIndex'];
  
  if (eventStartIndex != this._startIndex)
  {
    this._startIndex = event['startIndex'];
  }
  
  if (event['data'].length > 0)
  {
    this._updateEndIndex(event['startIndex'] + event['data'].length - 1, true);
  }
  else
  {
    this._updateEndIndex(-1, true);
  }
  
  if (this._fetchType == 'page')
  {
    // For paged fetches, reset the startIndex to zero since we always render
    // starting at 0
    var clonedEvent = {};
    oj.CollectionUtils.copyInto(clonedEvent, event);
    clonedEvent['startIndex'] = 0;
    eventHandler(clonedEvent);
  }
  else
  {
    eventHandler(event);
  }
};

oj.PagingTableDataSource.prototype._updateEndIndex = function(lastRowIdx, reset)
{ 
  if (reset)
  {
    this._endIndex = lastRowIdx;
  }
  else
  {
    // only update if greater
    this._endIndex = lastRowIdx > this._endIndex ? lastRowIdx : this._endIndex;
  }

  var totalSize = this.totalSize();

  if (totalSize > 0)
  {
    this._endIndex = this._endIndex > totalSize - 1 ? totalSize - 1 : this._endIndex;
  }
};

/**
 * @export
 * Event types
 * @enum {string}
 * @memberof oj.PagingTableDataSource
 */
oj.PagingTableDataSource.EventType =
  {
    /** Triggered when a Row is added to a PagingDataSource */
    'ADD': "add",
    /** Triggered when a Row is removed from a PagingDataSource */
    'REMOVE': "remove",
    /** Triggered when a PagingDataSource is reset */
    'RESET': "reset",
    /** Triggered when a PagingDataSource has been updated by a fetch */
    'SYNC': "sync",
    /** Triggered when a PagingDataSource has been refreshed */
    'REFRESH': "refresh",
    /** Triggered when a PagingDataSource has been sorted */
    'SORT': "sort",
    /** Triggered when a Row's attributes have been changed */
    'CHANGE': "change",
     /** Triggered when a PagingDataSource has sent a fetch request. */
    'REQUEST': "request",
    /** Triggered when an error occurred on the PagingDataSource */
    'ERROR': "error"
  };
/**This file is generated. Do not edit directly. Actual file located in 3rdparty/dvt/prebuild.**/
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojPictoChart
 * @augments oj.dvtBaseComponent
 * @since 1.2.0
 * @ojstatus preview
 * @ojshortdesc Displays information using icons to visualize an absolute number or the relative sizes of the different parts of a population.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="pictoChartOverview-section">
 *   JET PictoChart
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#pictoChartOverview-section"></a>
 * </h3>
 *
 * <p>PictoChart uses icons to visualize an absolute number, or the relative sizes of the different parts of a population.</p>
 *
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-picto-chart
 *   items='[{"name": "Red", "shape": "human", "color": "#ed6647", "count": 3},
 *          {"name": "Blue", "shape": "circle", "color": "#267db3", "count": 17}]'>
 * &lt;/oj-picto-chart>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <p>When using colors as a data dimension for PictoChart, the application
 * needs to ensure that they meet minimum contrast requirements. Not all colors
 * in the default value ramp provided by oj.ColorAttributeGroupHandler
 * will meet minimum contrast requirements.</p>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets. Alternate visualizations should
 *    be considered if identifying data changes is important, since all data items will generally move and resize on any data
 *    change.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * <h3 id="layout-section">
 *   Layout
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#layout-section"></a>
 * </h3>
 *
 * <h4>Fixed and Flowing Layout</h4>
 * <p>PictoChart supports both fixed and flowing layout. If the element has a fixed width and height
 * (set by the inline style, style class, etc.), then the pictoChart will use a <i>fixed layout</i>, which means that
 * the shapes will be resized to occupy the given space as much as possible. Otherwise, the pictoChart will use a
 * <i>flowing layout</i>, which means that the shapes are rendered at a constant size and the element will take up as
 * much space as necessary. It is possible to fix just one of the two dimensions, and the pictoChart would still
 * use the flowing layout.</p>
 * <p>If fixed layout is used, please avoid using the <code class="prettyprint">rowHeight</code> and <code class="prettyprint">columnWidth</code>
 * attributes as they may cause the shapes to be dropped if the given space is not large enough.</p>
 *
 * <h4>Layout Orientation and Origin</h4>
 * <p>PictoChart currently supports rectangular layouts with two different orientations (<i>horizontal</i> and <i>vertical</i>)
 * and four different origins (<i>topStart</i>, <i>topEnd</i>, <i>bottomStart</i>, and <i>bottomEnd</i>). Please refer to the
 * <a href="../jetCookbook.html?component=pictoChart&demo=default">cookbook demo</a> to see how these layout attributes work.</p>
 *
 * <h4>Mixed Sizes</h4>
 * <p>PictoChart supports items that are varying in sizes by specifying the <code class="prettyprint">columnSpan</code> and
 * <code class="prettyprint">rowSpan</code> attributes on the items. To ensure the best layout, it is recommended that the
 * bigger items are ordered first because the layout algorithm is greedy and will position items to the first available space.</p>
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojPictoChart', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     *  The duration of the animations, in milliseconds. The default value comes from the CSS and varies based on theme.
     * @expose
     * @name animationDuration
     * @memberof oj.ojPictoChart
     * @instance
     * @type {number}
     * @ojsignature {target: "Type", value: "?"}
     * @ojunits "milliseconds"
     */
     animationDuration: undefined,

     /**
     *  The animation when the data changes. 
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
     animationOnDataChange: "none",

    /**
     *  The animation that is shown on initial display. 
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "popIn"
     * @ojvalue {string} "alphaFade"
     * @ojvalue {string} "zoom"
     * @ojvalue {string} "none"
     * @default "none"
     */
     animationOnDisplay: "none",

    /**
     *  The number of columns that the picto chart has. The number of columns will be automatically computed if not specified. 
     * @expose
     * @name columnCount
     * @memberof oj.ojPictoChart
     * @instance
     * @type {number|null}
     * @default null
     */
     columnCount: null,

    /**
     *  The width of a column in pixels. The width of columns will be automatically computed if not specified.  Setting this property in a fixed layout (when the element width and height are defined) may cause items to be truncated. 
     * @expose
     * @name columnWidth
     * @memberof oj.ojPictoChart
     * @instance
     * @type {number|null}
     * @default null
     * @ojunits "pixels"
     */
     columnWidth: null,

    /**
     * An oj.DataProvider that generates rows of <a href="/jsdocs/oj.ojPictoChart.html#Item">oj.ojPictoChart.Item</a> objects.
     * @ignore
     * @name data
     * @memberof oj.ojPictoChart
     * @instance
     * @type {oj.DataProvider|Promise|null}
     * @ojsignature {target: "Accessor", value: {GetterType: "oj.DataProvider<any, oj.ojPictoChart.Item>", SetterType: "oj.DataProvider<any, oj.ojPictoChart.Item>|null"}}
     * @default null
     */  
     data: null,

    /**
     *  Whether drilling is enabled. Drillable items will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). To enable or disable drilling on individual items, use the drilling attribute in each item. 
     * @expose
     * @name drilling
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "off"
     */
     drilling: "off",

    /**
     *  An array of category strings used for category filtering. Data items with a category in hiddenCategories will be filtered. 
     * @expose
     * @name hiddenCategories
     * @memberof oj.ojPictoChart
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
     hiddenCategories: [],

    /**
     *  An array of category strings used for category highlighting. Data items with a category in highlightedCategories will be highlighted. 
     * @expose
     * @name highlightedCategories
     * @memberof oj.ojPictoChart
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    highlightedCategories: [],

    /**
     *  The matching condition for the highlightedCategories property. By default, highlightMatch is 'all' and only items whose categories match all of the values specified in the highlightedCategories array will be highlighted. If highlightMatch is 'any', then items that match at least one of the highlightedCategories values will be highlighted. 
     * @expose
     * @name highlightMatch
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "any"
     * @ojvalue {string} "all"
     * @default "all"
     */
     highlightMatch: "all",

    /**
     *  The behavior applied when hovering over data items. 
     * @expose
     * @name hoverBehavior
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "dim"
     * @ojvalue {string} "none"
     * @default "none"
     */
     hoverBehavior: "none",

    /**
     *  Specifies initial hover delay in ms for highlighting data items.
     * @expose
     * @name hoverBehaviorDelay
     * @memberof oj.ojPictoChart
     * @instance
     * @type {number}
     * @ojunits "milliseconds"
     * @default 200
     */
     hoverBehaviorDelay: 200,

    /**
     *  An array of objects with the following properties that defines the pictoChart items. 
     * @expose
     * @name items
     * @memberof oj.ojPictoChart
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @default null
     */
     items: null,

    /**
     *  The direction in which the items are laid out. 
     * @expose
     * @name layout
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "vertical"
     * @ojvalue {string} "horizontal"
     * @default "horizontal"
     */
     layout: "horizontal",

    /**
     *  Defines where the first item is rendered. The subsequent items follow the first item according to the layout. 
     * @expose
     * @name layoutOrigin
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "topEnd"
     * @ojvalue {string} "bottomStart"
     * @ojvalue {string} "bottomEnd"
     * @ojvalue {string} "topStart"
     * @default "topStart"
     */
     layoutOrigin: "topStart",

    /**
     *  The number of rows that the picto chart has. The number of rows will be automatically computed if not specified. 
     * @expose
     * @name rowCount
     * @memberof oj.ojPictoChart
     * @instance
     * @type {number|null}
     * @default null
     */
     rowCount: null,

    /**
     *  The height of a row in pixels. The height of rows will be automatically computed if not specified. Setting this property in a fixed layout (when the element width and height are defined) may cause items to be truncated. 
     * @expose
     * @name rowHeight
     * @memberof oj.ojPictoChart
     * @instance
     * @type {number|null}
     * @default null
     * @ojunits "pixels"
     */
     rowHeight: null,

    /**
     *  An array of id strings, used to define the selected objects. 
     * @expose
     * @name selection
     * @memberof oj.ojPictoChart
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
     selection: [],

    /**
     *  The type of selection behavior that is enabled on the picto chart. 
     * @expose
     * @name selectionMode
     * @memberof oj.ojPictoChart
     * @instance
     * @type {string}
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @ojvalue {string} "none"
     * @default "none"
     */
     selectionMode: "none",

    /**
     *  An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojPictoChart
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     * @desc
     * A function that returns a custom tooltip. The function takes a dataContext argument,
     * provided by the picto chart. The function should return an Object that contains only one of the two properties:
     * <ul>
     *   <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
     *   <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
     * </ul>
     */
    tooltip: {
        /**
         * @expose
         * @name tooltip.renderer
         * @memberof! oj.ojPictoChart
         * @instance
         * @type {function(Object):Object|null}
         * @ojsignature {target: "Type", value: "((context: oj.ojPictoChart.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))|null", jsdocOverride: true}
         * @default null
         */
         renderer: null

    },

    /**
     * Triggered during a drill gesture (double click if selection is enabled, single click otherwise).
     *
     * @property {string} id the id of the drilled object
     *
     * @expose
     * @event
     * @memberof oj.ojPictoChart
     * @instance
     */
    drill: null
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    return dvt.PictoChart.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-pictochart-item') {
      // item[index]
      subId = 'item[' + locator['index'] + ']';
    }
    else if(subId == 'oj-pictochart-tooltip') {
      subId = 'tooltip';
    }

    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId.indexOf('item') == 0) {
      // item[index]
      locator['subId'] = 'oj-pictochart-item';
      locator['index'] = this._GetFirstIndex(subId);
    }
    else if(subId == 'tooltip') {
      locator['subId'] = 'oj-pictochart-tooltip';
    }

    return locator;
  },

  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-pictochart');
    return styleClasses;
  },

  //** @inheritdoc */
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-pictochart-item'] = {'path': '_defaultColor', 'property': 'background-color'};
    styleClasses['oj-dvtbase oj-pictochart'] = {'path': 'animationDuration', 'property': 'ANIM_DUR'};
    return styleClasses;
  },

  //** @inheritdoc */
  _GetEventTypes : function() {
    return ['optionChange'];
  },

  //** @inheritdoc */
  _HandleEvent : function(event) {
    var type = event['type'];
    if (type === 'drill') {
      this._trigger('drill', null, {'id': event['id']});
    }
    else {
      this._super(event);
    }
  },

  /**
   * Returns an object with the following properties for automation testing verification of the item at the
   * specified index.

   * @param {number} index The index.
   * @return {Object|null} An object containing data for the item at the given index, or null if none exists.
   * @ojsignature {target: "Type", value: "oj.ojPictoChart.ItemContext|null", jsdocOverride: true, for: "returns"}
   * @expose
   * @memberof oj.ojPictoChart
   * @instance
   */
  getItem: function(index) {
    var auto = this._component.getAutomation();
    return auto.getItem(index);
  },

  /**
   * Returns the number of items in the pictoChart data.
   * @return {number} The number of data items
   * @expose
   * @memberof oj.ojPictoChart
   * @instance
   */
  getItemCount: function() {
    return this._component.getAutomation().getItemCount();
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   * @ojsignature {target: "Type", value: "oj.ojPictoChart.NodeContext|null", jsdocOverride: true, for: "returns"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojPictoChart
   */
  getContextByNode: function(node) {
    // context objects are documented with @ojnodecontext
    var context = this.getSubIdByNode(node);
    if (context && context['subId'] !== 'oj-pictochart-tooltip')
      return context;

    return null;
  },

  //** @inheritdoc */
  _GetComponentDeferredDataPaths : function() {
    return {'root': ['items', 'data']};
  },

  //** @inheritdoc */
  _IsFlowingLayoutSupported : function() {
    return true;
  }
});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="5">Data Item</td>
 *       <td rowspan="2"><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td>Drill when <code class="prettyprint">drilling</code> is enabled and <code class="prettyprint">selectionMode</code> is <code class="prettyprint">none</code>.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Double Tap</kbd></td>
 *       <td>Drill when <code class="prettyprint">drilling</code> is enabled and <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojPictoChart
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus and selection to previous item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus and selection to next item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus and selection to previous item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus and selection to next item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + UpArrow</kbd></td>
 *       <td>Move focus and multi-select previous item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + DownArrow</kbd></td>
 *       <td>Move focus and multi-select next item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + LeftArrow</kbd></td>
 *       <td>Move focus and multi-select previous item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + RightArrow</kbd></td>
 *       <td>Move focus and multi-select next item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + UpArrow</kbd></td>
 *       <td>Move focus to previous item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + DownArrow</kbd></td>
 *       <td>Move focus to next item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + LeftArrow</kbd></td>
 *       <td>Move focus to previous item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + RightArrow</kbd></td>
 *       <td>Move focus to next item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Spacebar</kbd></td>
 *       <td>Multi-select item with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Drill on item when <code class="prettyprint">drilling</code> is enabled.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojPictoChart
 */

// KEEP FOR WIDGET SYNTAX

/**
 *  The id of the item. 
 * @expose
 * @name items[].id
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The name of the item. Used for default tooltip and accessibility. 
 * @expose
 * @name items[].name
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The shape of the item. Can take the name of a built-in shape or the svg path commands for a custom shape. "None" will make the item transparent and can be used to create gaps. Does not apply if custom image is specified. 
 * @expose
 * @name items[].shape
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @ojvalue {string} "ellipse"
 * @ojvalue {string} "square"
 * @ojvalue {string} "circle"
 * @ojvalue {string} "diamond"
 * @ojvalue {string} "triangleUp"
 * @ojvalue {string} "triangleDown"
 * @ojvalue {string} "star"
 * @ojvalue {string} "plus"
 * @ojvalue {string} "human"
 * @ojvalue {string} "none"
 * @ojvalue {string} "rectangle"
 * @default "rectangle"
 */
/**
 *  The color of the item. Does not apply if custom image is specified. 
 * @expose
 * @name items[].color
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The border color of the item. Does not apply if custom image is specified. 
 * @expose
 * @name items[].borderColor
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The border width of the item in pixels. Does not apply if custom image is specified. 
 * @expose
 * @name items[].borderWidth
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {number}
 * @default null
 */
/**
 *  The URI of the custom image. If specified, it takes precedence over shape. 
 * @expose
 * @name items[].source
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the item. The style class and inline style will override any other styling specified with other properties. For tooltip interactivity, it's recommended to also pass a representative color to the item color attribute. Does not apply if custom image is specified.
 * @expose
 * @name items[].svgClassName
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the item. The style class and inline style will override any other styling specified with other properties. For tooltip interactivity, it's recommended to also pass a representative color to the item color attribute. Does not apply if custom image is specified.
 * @expose
 * @name items[].svgStyle
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {Object}
 * @default null
 */
/**
 *  The optional URI for the hover state. If not specified, the source image will be used. 
 * @expose
 * @name items[].sourceHover
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The optional URI for the selected state. If not specified, the source image will be used. 
 * @expose
 * @name items[].sourceSelected
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  The optional URI for the hover selected state. If not specified, the source image will be used. 
 * @expose
 * @name items[].sourceHoverSelected
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  Specifies the number of times that the shape (or custom image) is drawn. Fractional counts (such as 4.5) are supported; however, fractions other than the multiples of 0.5 should be avoided because the fractional rendering ignores the gaps between shapes and the irregularity of the shapes. 
 * @expose
 * @name items[].count
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {number}
 * @default 1
 */
/**
 *  The number of rows each shape (or custom image) spans. Used for creating a pictoChart with mixed item sizes. 
 * @expose
 * @name items[].rowSpan
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {number}
 * @default 1
 */
/**
 *  The number of columns each shape (or custom image) spans. Used for creating a pictoChart with mixed item sizes. 
 * @expose
 * @name items[].columnSpan
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {number}
 * @default 1
 */
/**
 *  Short description string for accessibility users. 
 * @expose
 * @name items[].shortDesc
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @default null
 */
/**
 *  An array of category strings corresponding to this item. If not specified, defaults to the item id or name. This enables highlighting and filtering of individual data items through interactions with other visualization elements. 
 * @expose
 * @name items[].categories
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 *  Whether drilling is enabled for the item. Drillable items will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). To enable drilling for all items at once, use the drilling attribute in the top level. 
 * @expose
 * @name items[].drilling
 * @memberof! oj.ojPictoChart
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default "inherit"
 */

 // PROPERTY TYPEDEFS

/**
 * @typedef {Object} oj.ojPictoChart.Item
 * @property {string=} id The id of the item.
 * @property {string=} name The name of the item. Used for default tooltip and accessibility.
 * @property {"ellipse"|"square"|"circle"|"diamond"|"triangleUp"|"triangleDown"|"star"|"plus"|"human"|"none"|"rectangle"|string} [shape="rectangle"] The shape of the item. Can take the name of a built-in shape or the svg path commands for a custom shape. "None" will make the item transparent and can be used to create gaps. Does not apply if custom image is specified. 
 * @property {string=} color The color of the item. Does not apply if custom image is specified. 
 * @property {string=} borderColor The border color of the item. Does not apply if custom image is specified. 
 * @property {number=} borderWidth The border width of the item in pixels. Does not apply if custom image is specified.
 * @property {string=} source The URI of the custom image. If specified, it takes precedence over shape.
 * @property {string=} svgClassName The CSS style class to apply to the item. The style class and inline style will override any other styling specified with other properties. For tooltip interactivity, it's recommended to also pass a representative color to the item color attribute. Does not apply if custom image is specified.
 * @property {Object=} svgStyle The inline style to apply to the item. The style class and inline style will override any other styling specified with other properties. For tooltip interactivity, it's recommended to also pass a representative color to the item color attribute. Does not apply if custom image is specified.
 * @property {string=} sourceHover The optional URI for the hover state. If not specified, the source image will be used.
 * @property {string=} sourceSelected The optional URI for the selected state. If not specified, the source image will be used.
 * @property {string=} sourceHoverSelected The optional URI for the hover selected state. If not specified, the source image will be used.
 * @property {number=} count Specifies the number of times that the shape (or custom image) is drawn. Fractional counts (such as 4.5) are supported; however, fractions other than the multiples of 0.5 should be avoided because the fractional rendering ignores the gaps between shapes and the irregularity of the shapes. 
 * @property {number=} rowSpan The number of rows each shape (or custom image) spans. Used for creating a pictoChart with mixed item sizes. 
 * @property {number=} columnSpan The number of columns each shape (or custom image) spans. Used for creating a pictoChart with mixed item sizes. 
 * @property {string=} shortDesc Short description string for accessibility users. 
 * @property {Array.<string>=} categories An array of category strings corresponding to this item. If not specified, defaults to the item id or name. This enables highlighting and filtering of individual data items through interactions with other visualization elements.
 * @property {"inherit"|"off"|"on"} [drilling="inherit"] Whether drilling is enabled for the item. Drillable items will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). To enable drilling for all items at once, use the drilling attribute in the top level. 
 */

/**
 * @typedef {Object} oj.ojPictoChart.TooltipContext
 * @property {Element} parentElement The tooltip element. The function can directly modify or append content to this element.
 * @property {string} id The id of the hovered item.
 * @property {string} name The name of the hovered item.
 * @property {number} count The count of the hovered item.
 * @property {string} color The color of the hovered item.
 * @property {Element} componentElement The picto chart HTML element.
 */

 // METHOD TYPEDEFS

/**
 * @typedef {Object} oj.ojPictoChart.ItemContext
 * @property {string} color
 * @property {number} count
 * @property {string} id
 * @property {string} name
 * @property {boolean} selected
 * @property {string} tooltip
 */

 /**
 * @typedef {Object} oj.ojPictoChart.NodeContext
 * @property {string} subId The subId string identify the particular DOM node.
 * @property {number} index The zero based index of the picto chart item.
 */

 // KEEP FOR WIDGET SYNTAX

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for tag cloud items at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojsubid oj-pictochart-item
 * @memberof oj.ojPictoChart
 *
 * @example <caption>Gets the first tag cloud item:</caption>
 * var nodes = myPictoChart.getNodeBySubId({'subId': 'oj-pictochart-item', 'index': 0});
 */

/**
 * <p>Sub-ID for the the tag cloud tooltip.</p>
 *
 * @ojsubid oj-pictochart-tooltip
 * @memberof oj.ojPictoChart
 *
 * @example <caption>Get the tooltip object of the tag cloud, if displayed:</caption>
 * var nodes = myPictoChart.getNodeBySubId({'subId': 'oj-pictochart-tooltip'});
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for tag cloud items at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojnodecontext oj-pictochart-item
 * @memberof oj.ojPictoChart
 */
/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojPictoChartMeta = {
  "properties": {
    "animationDuration": {
      "type": "number"
    },
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "popIn", "alphaFade", "zoom", "none"]
    },
    "columnCount": {
      "type": "number"
    },
    "columnWidth": {
      "type": "number"
    },
    "drilling": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string",
      "enumValues": ["any", "all"]
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "hoverBehaviorDelay": {
      "type": "number"
    },    
    "items": {
      "type": "Array<object>|Promise"
    },
    "layout": {
      "type": "string",
      "enumValues": ["horizontal", "vertical"]
    },
    "layoutOrigin": {
      "type": "string",
      "enumValues": ["topEnd", "bottomStart", "bottomEnd", "topStart"]
    },
    "rowCount": {
      "type": "number"
    },
    "rowHeight": {
      "type": "number"
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["single", "multiple", "none"]
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Picture Chart"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    }
  },
  "events": {
    "drill": {}
  },
  "methods": {
    "getContextByNode": {},
    "getItem": {},
    "getItemCount": {}
  },
  "extension": {
    _WIDGET_NAME: "ojPictoChart"
  }
};
oj.CustomElementBridge.registerMetadata('oj-picto-chart', 'dvtBaseComponent', ojPictoChartMeta);
oj.CustomElementBridge.register('oj-picto-chart', {'metadata': oj.CustomElementBridge.getMetadata('oj-picto-chart')});
})();

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*!
 * jQuery UI Popup @VERSION
 * http://jqueryui.com
 *
 * Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/menu/
 *
 * Depends:
 *  jquery.ui.core.js
 *  jquery.ui.widget.js
 *  jquery.ui.position.js
 */
(function ()
{

  /**
   * List of all pseudo marker selectors that defines rules for where a tail is aligned.
   * @private
   * @const
   */
  var _TAIL_STYLES = ["oj-left", "oj-center", "oj-right", "oj-top", "oj-middle", "oj-bottom"];

  /**
   * Mapping of horizontal-vertical (x,y) positon using alignment to jet tail pseudo marker
   * selectors.
   *
   * horizontal: right, left, center
   * vertical: top, bottom, middle
   *
   * @private
   * @const
   */
  var _TAIL_ALIGN_RULES =
    {
      'right-top' : 'oj-right oj-top',
      'right-middle' : 'oj-right oj-middle',
      'right-bottom' : 'oj-right oj-bottom',
      'left-top' : 'oj-left oj-top',
      'left-middle' : 'oj-left oj-middle',
      'left-bottom' : 'oj-left oj-bottom',
      'center-top' : 'oj-center oj-top',
      'center-middle' : 'oj-left oj-middle',
      'center-bottom' : 'oj-center oj-bottom'
    };

  /**
   * @typedef {Object} oj.ojPopup.PositionAlign
   * @property {"top"|"bottom"|"center"} [vertical] Vertical alignment.
   * @property {"start"|"end"|"left"|"center"|"bottom"} [horizontal] Horizontal alignment. <p>
   * <ul>
   *  <li><b>"start"</b> evaluates to "left" in LTR mode and "right" in RTL mode.</li>
   *  <li><b>"end"</b> evaluates to "right" in LTR mode and "left" in RTL mode.</li>
   * </ul>
   *
   */

  /**
   * @typedef {Object} oj.ojPopup.PositionPoint
   * @property {number} [x] Horizontal aligment offset.
   * @property {number} [y] Vertical alignment offset.
   */

  /**
   * @typedef {Object} oj.ojPopup.Position
   * @property {oj.ojPopup.PositionAlign} [my] Defines which edge on the popup to align with the target ("of") element.
   * @property {oj.ojPopup.PositionAlign} [at] Defines which position on the target element ("of") to align the positioned element
   *                                  against.
   * @property {oj.ojPopup.PositionPoint} [offset] Defines a point offset in pixels from the ("my") alignment.
   * @property {string|oj.ojPopup.PositionPoint} [of] Which element to position the popup against.  The default is the
   * <code class="prettyprint">launcher</code> argument passed to the
   * <code class="prettyprint">open</code> method. <p>
   *
   * If the value is a string, it should be a selector or the literal string value
   * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
   * is used, the values are relative to the whole document.  Page horizontal and vertical
   * scroll offsets need to be factored into this point - see UIEvent
   * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
   * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
   *
   * @property {"flip"|"fit"|"flipfit"|"none"} [collision] Rule for alternate alignment. <p>
   * <ul>
   *  <li><b>"flip"</b> the element to the opposite side of the target and the
   *             collision detection is run again to see if it will fit. Whichever side
   *             allows more of the element to be visible will be used. </li>
   * <li><b>"fit"</b> shift the element away from the edge of the window. </li>
   * <li><b>"flipfit"</b> first applies the flip logic, placing the element
   *  on whichever side allows more of the element to be visible. Then the fit logic
   *  is applied to ensure as much of the element is visible as possible.</li>
   * <li><b>"none"</b> no collision detection.</li>
   * </ul>
   */

  /**
   * @ojcomponent oj.ojPopup
   * @augments oj.baseComponent
   * @ojstatus preview
   * @since 1.1.0
   * @ojdisplayname Popup
   * @ojshortdesc Displays a popup window that can display arbitrary content.
   * @ojrole tooltip
   * @ojrole dialog
   * @ojrole alertdialog
   *
   * @classdesc
   * <h3 id="popupOverview-section">
   *   JET Popup Component
   *  <a class="bookmarkable-link" title="Bookmarkable Link" href="#popupOverview-section"></a></h3>
   *
   * <p>Description: Themeable, WAI-ARIA-compliant popup that can display arbitrary content.</p>
   *
   * <p>A JET popup can be created from custom element
   * <code class="prettyprint">&lt;oj-popup&gt;</code> syntax. This element will become the
   * root - outer chrome of the popup.  The content of the popup will be relocated under an
   * element marked with the <code class="prettyprint">.oj-popup-content</code> selector.
   * Dynamic content can be inserted under the element identified by the
   * <code class="prettyprint">.oj-popup-content</code> selector.  However, page developers
   * are encouraged to create their own content element to anchor dynamic content changes
   * versus using <code class="prettyprint">.oj-popup-content</code> as the marker selector.</p>
   *
   * <pre class="prettyprint">
   * <code>&lt;oj-popup id="popup"&gt;
   *   &lt;span class="mycontent"&gt;
   *     Hello World!
   *   &lt;/span class="mycontent"&gt;
   * &lt;/oj-popup&gt;
   * </code></pre>
   *
   * <p>The following is an example of dynamically changing the content of the popup defined above.
   * </p>
   *
   * <pre class="prettyprint">
   * <code>var content = popup.querySelector( ".mycontent" );
   * content.textContent = "Hello Universe!";
   * </code></pre>
   *
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDoc"}
   *
   * <br/><br/>
   *
   * <h3 id="accessibility-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
   * </h3>
   *
   * <p>For WAI-ARIA compliance, JET automatically adds
   * <code class="prettyprint">role="tooltip"</code> to the root popup dom element if not
   * already specificed. This is not a component property but rather the standard html
   * <a href="https://www.w3.org/TR/wai-aria/roles">role</a> attribute. Depending on how the
   * popup is used in the page, the page developer should choose from the following:
   * <ul>
   *   <li>"tooltip" defines contextual popup that displays a description for an element.</li>
   *   <li>"dialog" defines an application window that is designed to interrupt the current
   *       processing of an application in order to prompt the user to enter information or
   *       require a response.</li>
   *   <li>"alertdialog" defines type of dialog that contains an alert message, where initial focus
   *       goes to an element within the dialog.</li>
   * </ul>
   * The popup also adds the <code class="prettyprint">aria-describedby="popup-id"</code> attribute
   * to the assocaited launcher while the popup is open.
   * </p>
   *
   * On platforms that support voice over mode (VO), the popup injects anchor tags "skip links"
   * into the document for navigation. Skip links are not visible but read in VO mode.
   * Two skip links are injected into the document when a popup is disclosed:
   *  <ul>
   *    <li>A close link is injected as a sibling to the popup's content. Activation of this link
   *        will close the popup.</li>
   *    <li>For cases where the popup doesn't steal focus when it's open, a content navigation
   *        skip link is injected as a sibling to the launcher (first required argument of the open
   *        method). If the launcher selector targets a sub-element of the launcher,
   *        the skip link could be injected under the launcher, which can be problematic for
   *        oj-button as skip link activation will also activate the associated launcher.</li>
   *  </ul>
   *
   * <p>One point often overlooked is making the gestures that launch a popup accessible.
   *   There are no constraints to what events a page developer might choose to trigger opening a
   *   popup.  The choice should be accessible for screen reader users.  Page
   *   developers should take care when using mouse events to trigger opening of a popup.
   *   This is especially important if the content of the popup can't be derived from other
   *   visible areas on the page. In cases that mouseover, mouseout, mouseenter, mouseleave and
   *   hover events are used to launch popups, there needs to be a keyboard functional equivalency.
   * </p>
   *
   * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.
   *
   * <h3 id="reparenting-section">
   *   Reparenting
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#reparenting-section"></a>
   * </h3>
   *
   *  <p id="reparenting-strategy">
   *     When popups are open, they will be reparented in the document and reparented back when
   *     closed. When open, the location of the popup within the document will be in context of
   *     how it's used. Popups open from other popups will be relocated in the document into the
   *     nearest parent popup's layer. The popups layer defines its z-index weight "stacking
   *     context".  The ojPopup's layer is marked with the "oj-popup-layer" style.
   *     The context of opening is defined by the launcher argument passed to the open method.  If
   *     not open from another popup, the popup will be reparented to a container in the document
   *     body. Popups of the same type are assigned the same z-index values.  The layering between
   *     peer popups reflect the opening order. The popup that has active focus will be assigned a
   *     greater z-index value. This is applied to the popup's layer by way of the "oj-focus-within"
   *     pseudo selector applied with "oj-popup-layer" selector. The page author has control over
   *     z-index weights by way of the "oj-popup-layer" selector.
   *  </p>
   *  <p>
   *     There are known caveats with this design. However, these scenarios are considered "bad use"
   *     based on our JET popup strategy.
   *  </p>
   *  <ol>
   *    <li>Events raised within the popup will not bubble up to the popup's original ancestors.
   *        Instead, listeners for popup events should be applied to either the popup's root
   *        element, or the document.</li>
   *    <li>Likewise, developers should not use CSS descendant selectors, or similar logic, that
   *        assumes that the popup will remain a child of its original parent.</li>
   *    <li>Popups containing iframes are problematic.  The iframe elements "may" fire a HTTP GET
   *        request for its src attribute each time the iframe is reparented in the document.</li>
   *    <li>In some browsers, reparenting a popup that contains elements having overflow, will cause
   *        these overflow elements to reset their scrollTop.</li>
   *  </ol>
   *
   * <h3 id="styling-section">
   *   Styling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"stylingDoc"}
   *
   * <h3 id="eventHandling-section">
   *   Event Handling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#eventHandling-section"></a>
   * </h3>
   * <ul>
   *  <li>ojBeforeClose(event) - Triggered before a popup closes. Event can prevent closing the
   *      popup; However, there are cases the framework must veto, such as when the popup is
   *      destroyed.</li>
   *  <li>ojBeforeOpen(event) - Triggered before a popup opens. Event can prevent opening the
   *      popup.</li>
   *  <li>ojClose(event) - Triggered after the popup has closed.</li>
   *  <li>ojFocus(event) - Triggered when initial focus is established on opening, depending on
   *      the value of the initalFocus property, or <kbd>F6</kbd> focus toggle from the associated
   *      launcher.</li>
   *  <li>ojOpen(event) - Triggered after the popup has been made visible.</li>
   * </ul>
   */
  oj.__registerWidget("oj.ojPopup", $['oj']['baseComponent'],
    {
      widgetEventPrefix : "oj",
      options :
        {
          /**
           *
           * @private
           * @memberof oj.ojPopup
           * @instance
           */
          'animation' : null,
          /**
           * Defines conditions that will cause an open popup to auto close dismiss.  A value of
           * <code class="prettyprint">focusLoss</code> defines the dismissal condition where focus
           * has left the content of the popup or from the associated launcher or if what the popup
           * is aligned to is not fully visible within an overflow area.
           *
           * @expose
           * @memberof oj.ojPopup
           * @instance
           * @type {string}
           * @default 'focusLoss'
           * @ojvalue {string} "none" disables auto dismissal behaviors.
           * @ojvalue {string} "focusLoss" defines auto dismissal behavior when focus leaves the
           *   content of the popup or associated launcher.  In addition, if what the popup is
           *   positioned to is not visible within an overflow area, the popup will auto close
           *   dismiss.
           *
           * @example <caption>Initialize the popup with
           *          <code class="prettyprint">auto-dismiss</code> attribute specified:</caption>
           * &lt;oj-popup auto-dismiss="focusLoss" &gt;&lt;/oj-popup&gt;
           *
           * @example <caption>Get or set the <code class="prettyprint">autoDismiss</code> property,
           *          after initialization:</caption>
           * // getter
           * var autoDismiss = myPopup.autoDismiss;
           * // setter
           * myPopup.autoDismiss = "none";
           */
          autoDismiss : 'focusLoss',
          /**
           * Defines the presents of border, shadow and background color of the root popup dom.
           * Value of <code class="prettyprint">none</code> applies the
           * <code class="prettyprint">oj-popup-no-chrome</code> selector defined by the active
           * theme to the root dom of the popup to remove the default chrome.
           *
           * @expose
           * @memberof oj.ojPopup
           * @instance
           * @type {string}
           * @default 'default'
           * @ojvalue {string} "default" describes the popups border, shadow, and background color
           *           defined by the active theme.
           * @ojvalue {string} "none" turns off the outer chrome defined by the active theme.
           *
           * @example <caption>Initialize the popup with <code class="prettyprint">chrome</code>
           *          attribute specified:</caption>
           * &lt;oj-popup chrome="none" &gt;&lt;/oj-popup&gt;
           *
           * @example <caption>Get or set the <code class="prettyprint">chrome</code> property, after
           *          initialization:</caption>
           * // getter
           * var chrome = myPopup.chrome;
           *
           * // setter
           * myPopup.chrome = "none";
           */
          chrome : 'default',
          /**
           * Determines if the popup should steal focus to its content when initially open. A value
           * of <code class="prettyprint">none</code> prevents the popup from grabbing focus when
           * open.
           *
           * @expose
           * @memberof oj.ojPopup
           * @instance
           * @type {string}
           * @default 'auto'
           * @ojvalue {string} "auto" is derived from the values of the modality and
           *          autoDismiss properties
           * @ojvalue {string} "none" prevents the popup from stealing focus when open.
           * @ojvalue {string} "firstFocusable" defines that a popup should grab focus to the first
           *          focusable element within the popup's content.
           * @ojvalue {string} "popup" focus to the root popup container (good choice for touch
           *          platforms).
           *
           * @example <caption>Initialize the popup with
           *           <code class="prettyprint">initial-focus</code> attribute specified:</caption>
           * &lt;oj-popup initial-focus="none" &gt;&lt;/oj-popup&gt;
           *
           * @example <caption>Get or set the <code class="prettyprint">initialFocus</code> property,
           *          after initialization:</caption>
           * // getter
           * var initialFocus = myPopup.initialFocus;
           *
           * // setter
           * myPopup.initialFocus = "none";
           */
          initialFocus : 'auto',


          /**
           * <p>Position property is used to establish the location the popup will appear relative to
           * another element. {@link oj.ojPopup.Position} defines "my" alignment "at" the alignment
           * "of" some other thing which can be "offset" by so many pixels.</p>
           *
           * <p>The "my" and "at" properties defines aligment points relative to the popup and other
           * element.  The "my" property represents the popups alignment where the "at" property
           * represents the other element that can be identified by "of" or defauts to the launcher
           * when the popup opens.  The values of these properties describe horizontal and
           * vertical alignments.</p>
           *
           * <b>Deprecated v3.0.0 jQuery UI position syntax; Use of a percent unit with
           * "my" or "at" is not supported.</b>
           *
           * @expose
           * @memberof oj.ojPopup
           * @instance
           * @type {Object}
           * @default {"my":{"vertical" :"start", "horizontal": "top"}, "at": {"vertical": "start", "horizontal": "bottom"}, "collision": "flip"}
           * @ojsignature { target: "Type",
           *                value: "oj.ojPopup.Position",
           *                jsdocOverride: true}
           * @name position
           * @example <caption>Initialize the popup with <code class="prettyprint">position</code>
           *           attribute specified:</caption>
           * &lt;oj-popup position.my.horizontal="left"
           *           position.my.vertical="top"
           *           position.at.horizontal="right"
           *           position.at.vertical="top" &gt;&lt;/oj-popup&gt;
           *
           * @example <caption>Get or set the <code class="prettyprint">position</code> property,
           *          after initialization:</caption>
           * // getter
           * var position = myPopup.position;
           *
           * // setter
           * myPopup.position =
           *    {"my": {"horizontal": "start", "vertical": "bottom"},
           *     "at": {"horizontal": "end", "vertical": "top" },
           *     "offset": {"x": 0, "y":5}};
           */
          position :
            {
              /**
               * Defines which edge on the popup to align with the target ("of") element.
               *
               * @expose
               * @memberof! oj.ojPopup
               * @instance
               * @alias position.my
               * @name position.my
               * @type {{horizontal:string, vertical:string}}
               */
              'my' : {
                /**
                 * Defines the horizontal alignment of the popup.
                 * @expose
                 * @memberof! oj.ojPopup
                 * @instance
                 * @alias position.my.horizontal
                 * @name position.my.horizontal
                 * @type {string}
                 * @default 'start'
                 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
                 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
                 * @ojvalue {string} "left"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "right"
                 */
                'horizontal': 'start',
                /**
                 * Defines the vertical alignment of the popup.
                 * @expose
                 * @memberof! oj.ojPopup
                 * @instance
                 * @alias position.my.vertical
                 * @name position.my.vertical
                 * @type {string}
                 * @default 'top'
                 * @ojvalue {string} "top"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "bottom"
                 */
                'vertical': 'top'
              },
              /**
               * Defines a point offset in pixels from the ("my") alignment.
               * @expose
               * @memberof! oj.ojPopup
               * @instance
               * @alias position.offset
               * @name position.offset
               * @type {{x:number, y:number}}
               */
              'offset': {
                /**
                 * Horizontal aligment offset.
                 * @expose
                 * @memberof! oj.ojPopup
                 * @instance
                 * @alias position.offset.x
                 * @name position.offset.x
                 * @type {number}
                 * @default 0
                 */
                'x': 0,
                /**
                 * Vertical alignment offset.
                 * @expose
                 * @memberof! oj.ojPopup
                 * @instance
                 * @alias position.offset.y
                 * @name position.offset.y
                 * @type {number}
                 * @default 0
                 */
                'y': 0
              },
              /**
               * Defines which position on the target element ("of") to align the positioned element
               * against.
               *
               * @expose
               * @memberof! oj.ojPopup
               * @instance
               * @alias position.at
               * @name position.at
               * @type {{horizontal:string, vertical:string}}
               */
              'at' : {
                /**
                 * Defines the horizontal alignment of what the popup is aligned to.
                 * @expose
                 * @memberof! oj.ojPopup
                 * @instance
                 * @alias position.at.horizontal
                 * @name position.at.horizontal
                 * @type {string}
                 * @default 'start'
                 * @ojvalue {string} "start" evaluates to "left" in LTR mode and "right" in RTL mode.
                 * @ojvalue {string} "end" evaluates to "right" in LTR mode and "left" in RTL mode.
                 * @ojvalue {string} "left"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "right"
                 */
                'horizontal': 'start',
                /**
                 * Defines the vertical alignment of what the popup is aligned to.
                 * @expose
                 * @memberof! oj.ojPopup
                 * @instance
                 * @alias position.at.vertical
                 * @name position.at.vertical
                 * @type {string}
                 * @default 'top'
                 * @ojvalue {string} "top"
                 * @ojvalue {string} "center"
                 * @ojvalue {string} "bottom"
                 */
                'vertical': 'bottom'
              },
              /**
               * Which element to position the popup against.  The default is the
               * <code class="prettyprint">launcher</code> argument passed to the
               * <code class="prettyprint">open</code> method.
               *
               * If the value is a string, it should be a selector or the literal string value
               * of <code class="prettyprint">window</code>.  Otherwise, a point of x,y.  When a point
               * is used, the values are relative to the whole document.  Page horizontal and vertical
               * scroll offsets need to be factored into this point - see UIEvent
               * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageX">pageX</a>,
               * <a href="https://developer.mozilla.org/en-US/docs/Web/API/UIEvent/pageY">pageY</a>.
               *
               * @example <caption>Finding the point for an svg element:</caption>
               * var rect = svgDom.getBoundingClientRect();
               * var position = {of:{x:rect.left + window.pageXOffset, y:rect.top + window.pageYOffset}};
               *
               * @expose
               * @memberof! oj.ojPopup
               * @instance
               * @alias position.of
               * @name position.of
               * @type {string|{x: number, y: number}| undefined}
               */
              'of' : undefined,
              /**
               * Rule for alternate alignment.
               *
               * @expose
               * @memberof! oj.ojPopup
               * @instance
               * @alias position.collision
               * @name position.collision
               * @type {string}
               * @default 'flip'
               * @ojvalue {string} "flip" the element to the opposite side of the target and the
               *  collision detection is run again to see if it will fit. Whichever side
               *  allows more of the element to be visible will be used.
               * @ojvalue {string} "fit" shift the element away from the edge of the window.
               * @ojvalue {string} "flipfit" first applies the flip logic, placing the element
               *  on whichever side allows more of the element to be visible. Then the fit logic
               *  is applied to ensure as much of the element is visible as possible.
               * @ojvalue {string} "none" no collision detection.
               */
              'collision' : 'flip'
            },
          /**
           * Determines if a decoration will be displayed from the popup that points to the element
           * the popup is aligned to. The <code class="prettyprint">simple</code> value enables the
           * tail defined by the current theme.  In addtion, the
           * <code class="prettyprint">oj-popup-tail-simple</code> selector will be applied to the
           * root dom element.  This is to allow the box-shadow, z-index and other chrome styling to
           * vary per tail decoration.
           *
           * @expose
           * @memberof oj.ojPopup
           * @instance
           * @type {string}
           * @default 'none'
           * @ojvalue {string} "none" no decoration will be displayed from the popup pointing to the
           *          launcher.
           * @ojvalue {string} "simple" enables showing the tail defined by the current theme.
           *
           * @example <caption>Initialize the popup with <code class="prettyprint">tail</code>
           *          attribute specified:</caption>
           * &lt;oj-popup tail="simple" &gt;&lt;/oj-popup&gt;
           *
           * @example <caption>Get or set the <code class="prettyprint">tail</code> property, after
           *          initialization:</caption>
           * // getter
           * var tail = myPopup.tail;
           *
           * // setter
           * myPopup.tail = "simple";
           */
          tail : 'none',
          /**
           * Determines if the popup should block user input of the page behind with a blocking
           * overlay pane.
           *
           * <p>The default modality varies by theme.  Each theme can set its default by setting
           * <code class="prettyprint">$popupModalityOptionDefault</code>.
           *
           * @expose
           * @memberof oj.ojPopup
           * @instance
           * The modality of the popup. Valid values are:
           * @ojvalue {string} "modeless" defines a modeless popup.
           * @ojvalue {string} "modal" The popup is modal. Interactions with other page elements are
           *          disabled. Modal popups overlay other page elements.
           * @type {string}
           * @default 'modeless'
           * @example <caption>Initialize the popup to have modality
           *          <code class="prettyprint">modality</code></caption>
           * &lt;oj-popup modality="modal" &gt;&lt;/oj-popup&gt;
           *
           * @example <caption>Get or set the <code class="prettyprint">modality</code> property,
           *          after initialization:</caption>
           * // getter
           * var modality = myPopup.modality;
           *
           * // setter
           * myPopup.modality = "modal";
           *
           * @example <caption>Set the default in the theme (SCSS) :</caption>
           * $popupModalityOptionDefault: modal !default;
           */
          modality : "modeless",
          /**
           * @private
           * @memberof oj.ojPopup
           * @instance
           * @type {string}
           */
          'role' : "tooltip",
          // Events
          /**
           * Triggered before the popup is launched via the <code class="prettyprint">open()</code>
           * method. The open can be cancelled by calling
           * <code class="prettyprint">event.preventDefault()</code>.
           *
           * @expose
           * @event
           * @memberof oj.ojPopup
           * @instance
           * @ojcancelable
           * @ojbubbles
           */
          beforeOpen : null,
          /**
           * Triggered after the popup is launched via the <code class="prettyprint">open()</code>
           * method.
           *
           * @expose
           * @event
           * @memberof oj.ojPopup
           * @instance
           * @ojbubbles
           */
          open : null,
          /**
           * Triggered before the popup is dismissed via the
           * <code class="prettyprint">close()</code> method. The close can be cancelled by calling
           * <code class="prettyprint">event.preventDefault()</code>.
           *
           * @expose
           * @event
           * @memberof oj.ojPopup
           * @instance
           * @ojcancelable
           * @ojbubbles
           */
          beforeClose : null,
          /**
           * Triggered after the popup is dismissed via the
           * <code class="prettyprint">close()</code> method.
           *
           * @expose
           * @event
           * @memberof oj.ojPopup
           * @instance
           * @ojbubbles
           */
          close : null,
          /**
           * Triggered after focus has been transfered to the popup. This will occur after the
           * <code class="prettyprint">open()</code> method is called, depending on the value
           * of the <code class="prettyprint">initialFocus</code> property.  It's also triggered
           * when using the <kbd>F6</kbd> key to toggle focus from the associated launcher element
           * to the content of the popup.
           *
           * @expose
           * @event
           * @memberof oj.ojPopup
           * @instance
           * @ojbubbles
           */
          focus : null,
          /**
           * Triggered when a default animation is about to start, such as when the component is
           * being opened/closed or a child item is being added/removed. The default animation can
           * be cancelled by calling <code class="prettyprint">event.preventDefault</code>.
           *
           * @expose
           * @event
           * @memberof oj.ojPopup
           * @instance
           * @ojcancelable
           * @ojbubbles
           * @property {"open"|"close"} action The action that is starting the animation.
           *            The number of actions can vary from component to component.
           *            Suggested values are:
           *                    <ul>
           *                      <li>"open" - when a popup component is opened</li>
           *                      <li>"close" - when a popup component is closed</li>
           *                    </ul>
           * @property {!Element} element target of animation
           * @property {!function():void} endCallback If the event listener calls
           *            event.preventDefault to cancel the default animation, It must call the
           *            endCallback function when it finishes its own animation handling and any
           *            custom animation has ended.
           *
           * @example <caption>Bind an event listener to the
           *          <code class="prettyprint">onOjAnimateStart</code> property to override the default
           *          "close" animation:</caption>
           * myPopup.onOjAnimateStart = function( event )
           *   {
           *     // verify that the component firing the event is a component of interest and action
           *      is close
           *     if (event.detail.action == "close") {
           *       event.preventDefault();
           *       oj.AnimationUtils.slideOut(event.detail.element).then(event.detail.endCallback);
           *   };
           *
           * @example <caption>The default open and close animations are controlled via the theme
           *          (SCSS) :</caption>
           * $popupOpenAnimation: ((effect: "zoomIn"), "fadeIn")  !default;
           * $popupCloseAnimation: ((effect: "zoomOut", persist: "all"), "fadeOut")  !default;
           */
          animateStart : null,
          /**
           * Triggered when a default animation has ended, such as when the component is being
           * opened/closed or a child item is being added/removed. This event is not triggered if
           * the application has called preventDefault on the animateStart
           * event.
           *
           * @expose
           * @event
           * @memberof oj.ojPopup
           * @instance
           * @ojcancelable
           * @ojbubbles
           * @property {!Element} element target of animation
           * @property {"open"|"close"} action The action that is ending the animation.
           *                   The number of actions can vary from component to component.
           *                   Suggested values are:
           *                    <ul>
           *                      <li>"open" - when a popup component is opened</li>
           *                      <li>"close" - when a popup component is closed</li>
           *                    </ul>
           *
           * @example <caption>Bind an event listener to the
           *          <code class="prettyprint">onOjAnimateEnd</code> property to listen for the "close"
           *          ending animation:</caption>
           * myPopup.onOjAnimateEnd = function( event )
           *   {
           *     // verify that the component firing the event is a component of interest and action
           *      is close
           *     if (event.detail.action == "close") {}
           *   };
           *
           * @example <caption>The default open and close animations are controlled via the theme
           *          (SCSS) :</caption>
           * $popupOpenAnimation: (effect: "zoomIn", fade: true)  !default;
           * $popupCloseAnimation: (effect: "zoomOut", fade: true)  !default;
           */
          animateEnd : null
        },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @protected
       * @override
       */
      _ComponentCreate : function ()
      {
        this._super();

        var rootStyle = this._getRootStyle();
        var element = this.element;
        element.hide().addClass(rootStyle).attr("aria-hidden", "true");
        element.addClass("oj-component");

        // Creates a content element and moves the children of the root to the content element
        // and then appends the content element to the root element.
        var content = $("<div>");
        content.addClass([rootStyle, "content"].join("-"));
        content.attr("role", "presentation");
        content.append(element[0].childNodes);    //@HTMLUpdateOK; move app defined children to content wrapper
        content.appendTo(element);   //@HTMLUpdateOK; attach programmaticly generated node
        this._content = content;

        this._setChrome();
        this._setupFocus(element);

        // fixup the position option set via the widget constructor
        var options = this.options;
        options["position"] = oj.PositionUtils.coerceToJet(options["position"]);
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @protected
       * @override
       */
      _AfterCreate: function ()
      {
        // first apply rootAttributes that might define an id if unspecified
        this._super();

        // aria-describedby needs an id.  Assign an id to the popup root if not already defined.
        this.element.uniqueId();

        // create the tail with a subid
        this._createTail();
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @protected
       * @override
       */
      _destroy : function ()
      {
        if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
          this._closeImplicitly();

        this._setWhenReady("none");
        this._destroyTail();
        delete this._popupServiceEvents;

        // make sure the root element is hidden
        var element = this.element;
        element.hide().attr("aria-hidden", "true").removeUniqueId();

        // Move the children back under the root node removing the content node.
        var content = this._content;
        delete this._content;
        element.append(content[0].childNodes);  //@HTMLUpdateOK; destructor move children back under root
        content.remove();

        var closeDelayTimer = this._closeDelayTimer;
        if (closeDelayTimer)
        {
          delete this._closeDelayTimer;
          closeDelayTimer();
        }

        this._super();
      },
      /**
       * Opens the popup. This method accepts two arguments. The first argument "launcher" is
       * required but the second "position" is optional.
       *
       * @expose
       * @method
       * @name oj.ojPopup#open
       * @memberof oj.ojPopup
       * @instance
       * @param {!(string|Element)} launcher selector or dom element that is associated with the
       *        popup. Defines the context of how the popup is used. The argument is required.
       * @param {Object=} position {@link oj.ojPopup.Position} an element relative to another
       * @return {void}
       * @ojsignature { target: "Type",
       *                value: "oj.ojPopup.Position",
       *                for: "position",
       *                jsdocOverride: true}
       * @fires oj.ojPopup#ojBeforeOpen
       * @fires oj.ojPopup#ojAnimationStart
       * @fires oj.ojPopup#ojAnimationEnd
       * @fires oj.ojPopup#ojOpen
       *
       * @example <caption>Invoke the <code class="prettyprint">open</code> method:</caption>
       * var open = myPopup.open("#launcher");
       */
      open : function (launcher, position)
      {
        if (this._isOperationPending("open", [launcher, position]))
          return;

        if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
          this._closeImplicitly();

        this._setLauncher(launcher);

        var element = this.element;
        launcher = this._launcher;

        if (this._trigger("beforeOpen") === false)
          return;

        this._setWhenReady("open");

        var options = this.options;
        position = position ? position : options["position"];
        if (!position["of"])
        {
          this._hasPositionOfLauncherOverride = true;
          position["of"] = launcher;
        }

        this._setPosition(position);

        this._setAutoDismiss(options["autoDismiss"]);
        this._addDescribedBy();

        if (!this._IsCustomElement() || !element[0].hasAttribute["role"])
          element.attr("role", options["role"]);

        // convert to the jquery ui position format
        position = this._getPositionAsJqUi();

        // build layer class selectors applied to the popup layer
        var rootStyle = this._getRootStyle();
        var layerClass = [rootStyle, "layer"].join("-");
        var tailDecoration = options['tail'];
        if ("none" !== tailDecoration)
          layerClass += " " + [rootStyle, "tail", tailDecoration].join("-");

        /** @type {!Object.<oj.PopupService.OPTION, ?>} */
        var psOptions = {};
        psOptions[oj.PopupService.OPTION.POPUP] = element;
        psOptions[oj.PopupService.OPTION.LAUNCHER] = launcher;
        psOptions[oj.PopupService.OPTION.POSITION] = position;
        psOptions[oj.PopupService.OPTION.EVENTS] = this._getPopupServiceEvents();
        psOptions[oj.PopupService.OPTION.LAYER_SELECTORS] = layerClass;
        psOptions[oj.PopupService.OPTION.MODALITY] = options["modality"];
        psOptions[oj.PopupService.OPTION.CUSTOM_ELEMENT] = this._IsCustomElement();
        oj.PopupService.getInstance().open(psOptions);

      },
      /**
       * Before open callback is called after the popup has been reparented into the
       * zorder container. Open animation is performed here.
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for opening the popup
       * @return {Promise|void}
       */
      _beforeOpenHandler : function (psOptions)
      {
        var element = psOptions[oj.PopupService.OPTION.POPUP];
        var position = psOptions[oj.PopupService.OPTION.POSITION];

        element.show();
        element["position"](position);

        // TODO might want to add fadeIn for the modal overlay in the future.
        var animationOptions = this.options["animation"];
        if (animationOptions && animationOptions["open"])
        {
          var actionPrefix = animationOptions["actionPrefix"];
          var action = actionPrefix ? [actionPrefix, "open"].join("-") : "open";

          return oj.AnimationUtils.startAnimation(element[0], action,
            oj.PositionUtils.addTransformOriginAnimationEffectsOption(element,
            animationOptions["open"]), this);
        }
        else
        {
          return void(0);
        }
      },
      /**
       * Called after the popup is shown. Perform open finalization.
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for opening the popup
       * @return {void}
       */
      _afterOpenHandler : function (psOptions)
      {
        var element = psOptions[oj.PopupService.OPTION.POPUP];
        var launcher = psOptions[oj.PopupService.OPTION.LAUNCHER];

        this._initVoiceOverAssist();

        this._trigger("open");

        this._intialFocus();

        this._on(element, {'keydown' : this._keyHandler, 'keyup' : this._keyHandler});
        if (launcher && launcher.length > 0)
          this._on(launcher, {'keydown' : this._keyHandler, 'keyup' : this._keyHandler});
      },
      /**
       * Closes the popup. This method does not accept any arguments.
       *
       * @expose
       * @method
       * @name oj.ojPopup#close
       * @memberof oj.ojPopup
       * @instance
       * @return {void}
       * @fires oj.ojPopup#ojBeforeClose
       * @fires oj.ojPopup#ojAnimationStart
       * @fires oj.ojPopup#ojAnimationEnd
       * @fires oj.ojPopup#ojClose
       *
       * @example <caption>Invoke the <code class="prettyprint">close</code> method:</caption>
       * myPopup.close();
       */
      close : function ()
      {
        if (this._isOperationPending("close", []))
          return;

        if (oj.ZOrderUtils.getStatus(this.element) !== oj.ZOrderUtils.STATUS.OPEN)
          return;

        if (this._trigger("beforeClose") === false && !this._ignoreBeforeCloseResultant)
          return;

        this._setWhenReady("close");

        // if the content has focus, restore the the launcher
        this._restoreFocus();

        var launcher = this._launcher;
        var element = this.element;

        this._off(element, "keydown keyup");
        if (launcher && launcher.length > 0)
          this._off(launcher, "keydown keyup");

        //clean up voice over assist
        this._destroyVoiceOverAssist();

        /** @type {!Object.<oj.PopupService.OPTION, ?>} */
        var psOptions = {};
        psOptions[oj.PopupService.OPTION.POPUP] = element;
        oj.PopupService.getInstance().close(psOptions);

      },
      /**
       * Before callback is invoked while the popup is still visible and still parented in the
       * zorder container. Close animation is performed here.
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for closing the popup
       * @return {Promise|void}
       */
      _beforeCloseHandler : function (psOptions)
      {
        var element = psOptions[oj.PopupService.OPTION.POPUP];

        //TODO might want to add fadeOut for the modal overlay in the future.
        var animationOptions = this.options["animation"];

        if (!this._ignoreBeforeCloseResultant && animationOptions && animationOptions["close"])
        {
          var style = element.attr("style");
          var actionPrefix = animationOptions["actionPrefix"];
          var action = actionPrefix ? [actionPrefix, "close"].join("-") : "close";

          /** @type {?} */
          var promise = oj.AnimationUtils.startAnimation(element[0], action,
            oj.PositionUtils.addTransformOriginAnimationEffectsOption(element,
              animationOptions["close"]), this)
            .then(function ()
            {
              element.attr("style", style);
              element.hide();
            });
          return promise;
        }
        else
        {
          element.hide();
          return void(0);
        }
      },
      /**
       * Close finalization callback.
       *
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {!Object.<oj.PopupService.OPTION, ?>} psOptions property bag for closing the popup
       * @return {void}
       */
      _afterCloseHandler : function (psOptions)
      {
        this._removeDescribedBy();
        this._setAutoDismiss();

        delete this._launcher;
        this._trigger("close");

        if (this._hasPositionOfLauncherOverride)
        {
          var options = this.options;
          options["position"]["of"] = null;
          delete this._hasPositionOfLauncherOverride;
        }
      },
      /**
       * <p>Returns the state of whether the popup is currently open. This method does not accept
       * any arguments.</p>
       *
       * The "open" state reflects the period of time the popup is visible, including open and
       * close animations.
       *
       * @expose
       * @method
       * @name oj.ojPopup#isOpen
       * @memberof oj.ojPopup
       * @instance
       * @return {boolean} <code>true</code> if the popup is open.
       *
       * @example <caption>Invoke the <code class="prettyprint">isOpen</code> method:</caption>
       * var isOpen = myPopup.isOpen();
       */
      isOpen : function ()
      {
        var status = oj.ZOrderUtils.getStatus(this.element);
        return (status === oj.ZOrderUtils.STATUS.OPENING ||
                status === oj.ZOrderUtils.STATUS.OPEN ||
                status === oj.ZOrderUtils.STATUS.CLOSING);
      },
      /**
       * Causes the popup to reevaluate its position.  Call this function after
       * the content of the popup has dynamically changed while the popup is open.
       *
       * <p>This method does not accept any arguments.</p>
       *
       * @expose
       * @method
       * @name oj.ojPopup#refresh
       * @memberof oj.ojPopup
       * @instance
       * @return {void}
       *
       * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
       * myPopup.refresh();
       */
      refresh : function ()
      {
        this._super();

        if (oj.ZOrderUtils.getStatus(this.element) !== oj.ZOrderUtils.STATUS.OPEN)
          return;

        if (!this._reposition())
          return;

        // trigger refresh of descendents if reposition was successful
        var element = this.element;
        oj.PopupService.getInstance().triggerOnDescendents(element,
          oj.PopupService.EVENT.POPUP_REFRESH);
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @protected
       * @param {string} key option name
       * @param {?Object} value of the target option identified by the key
       * @override
       */
      _setOption : function (key, value)
      {

        var options = this.options;
        switch (key)
        {
          case "tail":
            if (value !== options["tail"])
            {
              this._setTail(value);
            }
            break;
          case "chrome":
            if (value !== options["chrome"])
              this._setChrome(value);
            break;
          case "position":
            this._setPosition(value);
            this.refresh();
            // don't call super because setPosition sets the option after creating a new
            // instance.  This prevents the same position instance from getting registered
            // with multiple component instances.
            return;
          case "autoDismiss":
            if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN && value !== options["autoDismiss"])
              this._setAutoDismiss(value);
            break;
          case "modality":
            if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
            {
              var element = this.element;
              /** @type {!Object.<oj.PopupService.OPTION, ?>} */
              var psOptions = {};
              psOptions[oj.PopupService.OPTION.POPUP] = element;
              psOptions[oj.PopupService.OPTION.MODALITY] = value;
              oj.PopupService.getInstance().changeOptions(psOptions);
            }
            break;
        }

        this._superApply(arguments);
      },
      /**
       * Returns the root selector prefix for the popup components.  In the future if the popup is
       * subcassed, this method can be made protected and override to change the root selector names
       * for the target component.
       *
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @return {string}
       */
      _getRootStyle : function ()
      {
        return "oj-popup";
      },
      /**
       * Handles setting up the target tail.
       *
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {string} tail option value
       */
      _setTail : function (tail)
      {
        this._destroyTail();
        this._createTail(tail);
        this._reposition();
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {string} tail option value
       */
      _createTail : function (tail)
      {
        var tailDecoration = tail ? tail : this.options['tail'];
        if ("none" === tailDecoration)
          return;

        var rootStyle = this._getRootStyle();
        var tailMarkerStyle = [rootStyle, "tail"].join("-");
        var tailStyle = [tailMarkerStyle, tailDecoration].join("-");

        var tailDom = $("<div>").hide();
        tailDom.addClass(tailMarkerStyle).addClass(tailStyle);
        tailDom.attr("role", "presentation");

        // id over "marker style" due to nesting popups in popups
        this._tailId = tailDom.attr("id", this._getSubId("tail")).attr("id");
        var element = this.element;
        tailDom.appendTo(element);  //@HTMLUpdateOK; attach programmaticly generated node

        // tail "value" style is applied to the root dom for shadow and z-index adjustments
        element.addClass(tailStyle);

        // The tail can change the z-index of the layer that defines the stacking context
        // of the popup.  If the popup is open, update the layers class.
        if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
        {
          var layerClass = [rootStyle, "layer"].join("-");
          layerClass += " " + tailStyle;

          /** @type {!Object.<oj.PopupService.OPTION, ?>} */
          var options = {};
          options[oj.PopupService.OPTION.POPUP] = element;
          options[oj.PopupService.OPTION.LAYER_SELECTORS] = layerClass;
          oj.PopupService.getInstance().changeOptions(options);
        }

      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @return {jQuery}
       */
      _getTail : function ()
      {
        var tailId = this._tailId;
        if (!tailId)
          return null;

        return $(document.getElementById(tailId));
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _destroyTail : function ()
      {

        var tail = this._getTail();
        if (tail)
          tail.remove();

        delete this._tailId;

        var tailDecoration = this.options['tail'];
        var rootStyle = this._getRootStyle();
        var tailStyle = [rootStyle, "tail", tailDecoration].join("-");

        var element = this.element;
        element.removeClass(tailStyle);

        // if the popup is open, reseed the layer class removing the
        // tail style.
        if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
        {
          var layerClass = [rootStyle, "layer"].join("-");
          /** @type {!Object.<oj.PopupService.OPTION, ?>} */
          var options = {};
          options[oj.PopupService.OPTION.POPUP] = element;
          options[oj.PopupService.OPTION.LAYER_SELECTORS] = layerClass;
          oj.PopupService.getInstance().changeOptions(options);
        }

      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {string} chrome option value
       */
      _setChrome : function (chrome)
      {
        var chromeDecoration = (chrome ? chrome : this.options["chrome"]);
        var noChromeStyle = [this._getRootStyle(), "no-chrome"].join("-");
        var element = this.element;

        if ("default" === chromeDecoration && element.hasClass(noChromeStyle))
          element.removeClass(noChromeStyle);
        else if ("none" === chromeDecoration && !element.hasClass(noChromeStyle))
          element.addClass(noChromeStyle);
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {string|Node|jQuery|null} launcher provided when the popup is open
       */
      _setLauncher : function (launcher)
      {
        if (!launcher)
          launcher = $(document.activeElement);
        else if ($.type(launcher) === "string")//id jquery selector
          launcher = $(launcher);
        else if (launcher.nodeType === 1)//dom element
          launcher = $(launcher);

        // if a jquery collection, select the first dom node not in the popups content
        if (launcher instanceof $ && launcher.length > 1)
        {
          var element = this.element;

          for (var i = 0; i < launcher.length; i++)
          {
            var target = launcher[0];
            if (!oj.DomUtils.isAncestorOrSelf(element[0], target))
            {
              launcher = $(target);
              break;
            }
          }
        }
        else if (!(launcher instanceof $) || //object is not a jq
          ((launcher instanceof $) && launcher.length === 0))// empty jq collection
          launcher = $(document.activeElement);

        this._launcher = launcher;
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {?Object} position object set as an option or passed as an argument to the open
       *                  method.
       */
      _setPosition : function (position)
      {
        var options = this.options;

        // new position extends the existing object
        // covert to jet internal position format
        if (position)
        {
          options["position"] = oj.PositionUtils.coerceToJet(position, options["position"]);
        }
      },

      /**
       * Returns a jQuery UI position object from the internal object.
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @returns {Object}
       */
      _getPositionAsJqUi: function ()
      {
        var options = this.options;
        var position = oj.PositionUtils.coerceToJqUi(options["position"]);
        var isRtl = this._GetReadingDirection() === "rtl";
        position = oj.PositionUtils.normalizeHorizontalAlignment(position, isRtl);

        var origUsing = position["using"];
        origUsing = $.isFunction(origUsing) ? origUsing : null;

        // override with our proxy to handle positioning of the tail
        // overload the callback arguments forcing the original using as the first argument
        position["using"] = this._usingHandler.bind(this, origUsing);

        return position;
      },

      /**
       * Resolves a busy state blocking delayed call to implicitly close
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {Function} busyStateResolver
       */
      _resolveBusyStateAndCloseImplicitly: function(busyStateResolver)
      {
        busyStateResolver();
        delete this._closeDelayTimer;
        this._closeImplicitly();
      },
      /**
       * Cancels the delayed implicit closure.
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {number} timer
       * @param {Function} busyStateResolver
       */
      _resolveBusyStateAndCancelDelayedClosure: function (timer, busyStateResolver)
      {
        window.clearTimeout(timer);
        busyStateResolver();
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {Function} origUsing position using hander if already provided
       * @param {Object} pos "my" element associated with the position object
       * @param {Object} props directions as to where the element should be moved
       */
      _usingHandler : function (origUsing, pos, props)
      {
        var element = props["element"]["element"];

        // do nothing if the position is the same
        if (pos["top"] === element.css("top") && pos["left"] === element.css("left"))
          return;

        var tail = this._getTail();
        if (!tail)
          element.css(pos);
        else
        {
          tail.hide();
          for (var i = 0; i < _TAIL_STYLES.length; i++)
          {
            tail.removeClass(_TAIL_STYLES[i]);
            element.removeClass(_TAIL_STYLES[i]);
          }
          tail.removeAttr("style");

          // 
          // Check if "of" alignment is to a x,y versus a dom element.  The horizontal
          // rule returned from jquery UI position defaults to "left" or "right".  The height and
          // width of the target will be zero when the "of" is not a DOM element.
          // Use the position.my alignment rules over what jquery returns when aligned to a point.
          //
          // jQuery 3.1 made changes to the offset.  The basic refactor was to switch over to using
          // "Element.getBoundingClientRect()".  Part of this refactor will return a offset rect of
          // {top: 0, left: 0} for any element that doesn't have a bounding box
          // "Element.getClientRects()". Many types of SVG elements such as <g> fall into this
          // category.  The popup will appear in the top left of the browser.
          //
          if (props["target"] && props["target"]["height"] === 0 && props["target"]["width"] === 0)
          {
            var isRtl = this._GetReadingDirection() === "rtl";
            var position = oj.PositionUtils.normalizeHorizontalAlignment(this.options["position"], isRtl);
            var myrule = position["my"];
            if (!oj.StringUtils.isEmptyOrUndefined(myrule))
            {
              // If the original horizontal rule is center, use it; otherwise, use the calculated
              // hint. The left/right rules reflect the actual positioning but center is never
              // represented correctly aligned to a point even though the alignment is correct.
              var suggestedHrule = "center" === myrule["horizontal"] ? myrule["horizontal"] : props["horizontal"];
              var suggestedVrule = ("center" === myrule["vertical"]) ? "middle" : myrule["vertical"];
              props['horizontal'] = suggestedHrule;
              props['vertical'] = suggestedVrule;
            }
          }

          var alignMnemonic = [props["horizontal"], props["vertical"]].join("-");
          var tailStyle = _TAIL_ALIGN_RULES[alignMnemonic];
          tail.addClass(tailStyle);
          element.addClass(tailStyle);
          tail.show();

          // adjust the vertical and horizontal positioning to account for the tail
          // so that the page developer doesn't have to factor that in
          var borderFactor = 2; // factor in a little extra so the borders overlap
          if ("left" === props["horizontal"])
          {
            var tailHOffset = tail.outerWidth();
            tailHOffset -= (tailHOffset + oj.DomUtils.getCSSLengthAsInt(tail.css("left")));
            pos["left"] = pos["left"] + (tailHOffset - borderFactor);
          }
          else if ("right" === props["horizontal"])
          {
            var tailHOffset = tail.outerWidth();
            tailHOffset -= (tailHOffset + oj.DomUtils.getCSSLengthAsInt(tail.css("right")));
            pos["left"] = pos["left"] - (tailHOffset - borderFactor);
          }

          // tail adjustments when the offset of the image is not the total size of the image
          if ("top" === props["vertical"])
          {
            var tailVOffset = tail.outerHeight();
            tailVOffset -= (tailVOffset +
              oj.DomUtils.getCSSLengthAsInt(tail.css(props["vertical"])));
            pos["top"] = pos["top"] + (tailVOffset - borderFactor);
          }
          else if ("bottom" === props["vertical"])
          {
            var tailVOffset = tail.outerHeight();
            tailVOffset -= (tailVOffset +
              oj.DomUtils.getCSSLengthAsInt(tail.css(props["vertical"])));
            pos["top"] = pos["top"] - (tailVOffset - borderFactor);
          }
          element.css(pos);

          // adjustments to the vertical or horizontal centering.  The 50% alignment is from
          // the edge of the tail versus the center of the image.  The tail can't be located
          // at "center, middle". In this case (dead center), horizintal center looks better
          // on small viewports (_TAIL_ALIGN_RULES["center-middle"] === 'oj-left oj-middle')
          if ("center" === props["horizontal"] && "middle" !== props["vertical"])
          {
            var rootWidth = element.width();
            var leftPercent = Math.round((((rootWidth / 2) - (tail.outerWidth() / 2))
              / rootWidth) * 100);
            tail.css(
              {
                left : leftPercent + '%'
              });
          }
          else if ("middle" === props["vertical"])
          {
            var rootHeight = element.height();
            var topPercent = Math.round((((rootHeight / 2) - (tail.outerHeight() / 2))
              / rootHeight) * 100);
            tail.css(
              {
                top : topPercent + '%'
              });
          }
        }

        oj.PositionUtils.captureTransformOriginAnimationEffectsOption(element, props);

        // call on the original using regardless of the tail
        if (origUsing)
          origUsing(pos, props);

        var options = this.options;

        // The "origUsing" could alter the positon.  This check needs to be last.
        // When focusLoss auto dismissal is enabled, implicitly close the popup when the
        // position.of is clipped in an overflow container.
        if ("focusLoss" === options["autoDismiss"])
        {
          if (oj.PositionUtils.isAligningPositionClipped(props))
          {
            // Ignore focus back to what had focus before the popup was open. Focus
            // restore could fight scroll if the popup was closed due to the aligning
            // element being clipped.
            this._ignoreRestoreFocus = true;

            // operation needs to happen in the next stacking frame and guarded by a
            // busy state.
            var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
            var bsOptions = {description: ["ojPopup identified by '", this.element.attr('id'),
                                         "' is pending implicit closure."].join("")};
            var resolver = busyContext.addBusyState(bsOptions);
            var delayTimer = window.setTimeout(this._resolveBusyStateAndCloseImplicitly
              .bind(this, resolver), 0);

            this._closeDelayTimer = this._resolveBusyStateAndCancelDelayedClosure.bind(this,
              delayTimer, resolver);
          }
        }
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @return {boolean} <code>false</code> if the position was skipped
       */
      _reposition : function ()
      {
        var element = this.element;
        var position = this._getPositionAsJqUi();

        // verify selector is valid; otherwise, skip the reposition
        if (oj.StringUtils.isString(position["of"]))
        {
          var jqOf = $(position["of"]);
          if (jqOf.length === 0)
            return false;

          position["of"] = jqOf;
        }

        element.position(position);
        return true;
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {boolean=} waiAriaAssisted focus established via keyboard or voice over versus from
       *        open API
       */
      _intialFocus : function (waiAriaAssisted)
      {
        var initialFocus = this._deriveInitialFocus();

        // We are toggling focus into the popup due to F6 or voice over skip link.
        if (waiAriaAssisted && "none" === initialFocus)
          initialFocus = "popup";

        if ("firstFocusable" === initialFocus)
        {
          var content = this._content;
          var nodes = content.find(":focusable");
          var first;
          for (var i = 0; i < nodes.length; i++)
          {
            if (oj.FocusUtils.isFocusable(nodes[i]))
            {
              first = nodes[i];
              break;
            }
          }

          if (first)
          {
            first.focus();
            this._trigger("focus");
          }
          else  // nothing to set focus to, default to "popup"
            initialFocus = "popup";
        }

        // Establish focus to the root element of the popup.  It's not a natural focus stop
        if ("popup" === initialFocus)
        {
          var element = this.element;
          element.attr("tabindex", "-1");

          var closeSkipLink = this._closeSkipLink;
          if (closeSkipLink)
          {
            var linkElement = closeSkipLink.getLink();
            linkElement.focus();
          }
          else
          {
            element.focus();
          }

          this._trigger("focus");
        }
      },
      /**
       * @memberof oj.ojPopup
       * @private
       * @return {string} derives the target initialFocus option when the default is auto
       */
      _deriveInitialFocus : function ()
      {
        var options = this.options;
        var initialFocus = options["initialFocus"];

        if ("auto" === initialFocus)
        {
          var modality = options["modality"];
          if (modality === "modal")
          {
            if (oj.DomUtils.isTouchSupported())
              initialFocus = "popup";
            else
              initialFocus = "firstFocusable";
          }
          else
            initialFocus = "none";
        }

        return initialFocus;
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {Element} activeElement from the event being handled
       * @param {boolean!} includeChildren when true the focus test will include the scope of any
       *                   child popups.
       * @return {boolean} <code>true</code> if the active element is within the content of the
       *                   popup
       */
      _isFocusInPopup : function (activeElement, includeChildren)
      {
        if (!activeElement)
          activeElement = document.activeElement;

        // added to avoid automation issues where an active element is not established
        if (!activeElement)
          return false;

        var element = this.element;

        // popups that are children are siblings to the parent popup within the
        // layer that defines the stacking context.
        if (includeChildren)
          element = element.parent();

        return oj.DomUtils.isAncestorOrSelf(element[0], activeElement);
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {Element} activeElement from the event being handled
       * @return {boolean} <code>true</code> if the active element the launcher or a decedent of the
       *         launcher
       */
      _isFocusInLauncher : function (activeElement)
      {
        if (!activeElement)
          activeElement = document.activeElement;

        var launcher = this._launcher;
        return oj.DomUtils.isAncestorOrSelf(launcher[0], activeElement);
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _restoreFocus : function ()
      {
        if (this._ignoreRestoreFocus)
        {
          delete this._ignoreRestoreFocus;
          return;
        }

        // extend the focus check to include popups that are children
        if (this._isFocusInPopup(null, true))
        {
          var launcher = this._launcher;
          launcher.focus();
        }
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {jQuery.Event|Event} event keydown
       */
      _keyHandler : function (event)
      {
        if (event.isDefaultPrevented())
          return;

        var eventType = event.type;
        var content = this._content;

        /** @type {?} */
        var target = event.target;
        if ("keyup" === eventType && event.keyCode === $.ui.keyCode.ESCAPE &&
          (this._isFocusInPopup(target) || this._isFocusInLauncher(target)))
        {
          event.preventDefault();
          this.close();
        }
        else if ("keydown" === eventType && event.keyCode === 117)
        {
          //F6 - toggle focus to launcher or popup
          if (this._isFocusInPopup(target))
          {
            // If this is a modeless popup, toggle focus to the launcher;
            // otherwise, close the popup as we can't set focus under the
            // modal glass pane.
            var options = this.options;
            if ("modeless" === options['modality'])
            {
              event.preventDefault();
              var launcher = this._launcher;
              launcher.focus();
            }
            else
              this.close();
          }
          else if (this._isFocusInLauncher(target))
          {
            event.preventDefault();
            this._intialFocus(true);
          }
        }
        else if ("keydown" === eventType && event.keyCode === $.ui.keyCode.TAB &&
          this._isFocusInPopup(target))
        {
          // TAB within popup
          var nodes = content.find(":tabbable");
          if (nodes.length > 0)
          {
            var firstNode = nodes[0];
            var lastNode = nodes[nodes.length - 1];
            var element = this.element;

            if ((firstNode === target || element[0] === target) && event.shiftKey)
            {
              //tabbing backwards, cycle focus to last node
              event.preventDefault();
              // If the first and last tab stops are the same,
              // force focus to the root popup dom.  This will
              // cause the blur to fire on any input components.
              // If we are back tabbing on the popup dom, jump to the
              // last tab stop.
              if (firstNode === lastNode && firstNode === target)
              {
                element.attr("tabindex", "-1");
                element.focus();
              }
              else
                $(lastNode).focus();  //tabbing backwards, cycle focus to last node
            }
            else if (lastNode === target && !event.shiftKey)
            {
              event.preventDefault();
              // If the first and last tab stops are the same,
              // force focus to the root popup dom.  This will
              // cause the blur to fire on any input components.
              if (lastNode === firstNode)
              {
                element.attr("tabindex", "-1");
                element.focus();
              }
              else
                $(firstNode).focus(); //tabbing forwards, cycle to the first node
            }
          }
          else
          {
            event.preventDefault();
            var options = this.options;
            if ("modeless" === options['modality'])
            {
              // if there is nothing in the popup that is tabbable, handle as a F6
              // toggle to the launcher
              var launcher = this._launcher;
              launcher.focus();
            }
            else
            {
              // Modal popup can't set focus to something under the overlay,
              // implicitly close.
              this.close();
            }
          }
        }
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {string|null} autoDismiss option value
       */
      _setAutoDismiss : function (autoDismiss)
      {

        // unregister any existing handlers, might need to add mouseOut in the future
        var focusLossCallback = this._focusLossCallback;
        var events = this._getPopupServiceEvents();
        if (focusLossCallback)
        {
          delete events[oj.PopupService.EVENT.POPUP_AUTODISMISS];
          delete this._focusLossCallback;
        }

        if ("focusLoss" === autoDismiss)
        {
          focusLossCallback = this._focusLossCallback = this._dismissalHandler.bind(this);
          events[oj.PopupService.EVENT.POPUP_AUTODISMISS] = focusLossCallback;
        }

        if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
        {
          var element = this.element;
          /** @type {!Object.<oj.PopupService.OPTION, ?>} */
          var options = {};
          options[oj.PopupService.OPTION.POPUP] = element;
          options[oj.PopupService.OPTION.EVENTS] = events;
          oj.PopupService.getInstance().changeOptions(options);
        }
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {Event} event native doc
       */
      _dismissalHandler : function (event)
      {
        if (oj.ZOrderUtils.getStatus(this.element) !== oj.ZOrderUtils.STATUS.OPEN)
          return;

        var launcher = this._launcher;
        var element = this.element;

        // child popups are siblings to the parent in the layer.
        var layer = element.parent();

        /** @type {?} */
        var target = event.target;

        // if the target is on the focus skip link next to the launcher, ignore.
        var focusSkipLink = this._focusSkipLink;
        if (focusSkipLink)
        {
          var link = focusSkipLink.getLink();
          if (link && oj.DomUtils.isAncestorOrSelf(link[0], target))
            return;
        }

        // if event target is not under the laucher or popup root dom subtrees, dismiss
        if (!oj.DomUtils.isAncestorOrSelf(launcher[0], target) &&
          !oj.DomUtils.isAncestorOrSelf(layer[0], target))
        {
          if (oj.FocusUtils.isFocusable(target))
          {
            // If the dismissal event target can take focus and the
            // event type is a mousedown or touchstart, wait for the focus event
            // to trigger dismissal.  This allows the blur to happen
            // on input components which triggers validation.
            if ("mousedown" === event.type || "touchstart" === event.type)
              return;

            this._ignoreRestoreFocus = true;
          }

          // invoke standard close dismissal that can be canceled via the beforeclose
          // event.
          this.close();
        }
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _addDescribedBy : function ()
      {
        var launcher = this._launcher;
        var element = this.element;

        var popupId = element.attr("id");
        var describedby = launcher.attr("aria-describedby");
        var tokens = describedby ? describedby.split(/\s+/) : [];
        tokens.push(popupId);
        describedby = $.trim(tokens.join(" "));
        launcher.attr("aria-describedby", describedby);
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _removeDescribedBy : function () {
        var launcher = this._launcher;
        var element = this.element;

        if (!launcher || launcher.length === 0)
          return;

        var popupId = element.attr("id");
        var describedby = launcher.attr("aria-describedby");
        var tokens = describedby ? describedby.split(/\s+/) : [];
        var index = $.inArray(popupId, tokens);
        if (index !== -1)
          tokens.splice(index, 1);

        describedby = $.trim(tokens.join(" "));
        if (describedby)
          launcher.attr("aria-describedby", describedby);
        else
          launcher.removeAttr("aria-describedby");
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _initVoiceOverAssist : function ()
      {
        var isVOSupported = (oj.AgentUtils.getAgentInfo()['os'] === oj.AgentUtils.OS.IOS ||
                             oj.AgentUtils.getAgentInfo()['os'] === oj.AgentUtils.OS.ANDROID);
        var liveRegion = this._liveRegion;
        if (!liveRegion)
          liveRegion = this._liveRegion = new oj.PopupLiveRegion();

        var message;
        var initialFocus = this._deriveInitialFocus();
        if (isVOSupported)
          message = this.getTranslatedString("none" === initialFocus ?
            "ariaLiveRegionInitialFocusNoneTouch" :
            "ariaLiveRegionInitialFocusFirstFocusableTouch");
        else
          message = this.getTranslatedString("none" === initialFocus ?
            "ariaLiveRegionInitialFocusNone" :
            "ariaLiveRegionInitialFocusFirstFocusable");
        liveRegion.announce(message);

        if (isVOSupported)
        {
          var focusSkipLinkId = this._getSubId("focusSkipLink");
          var launcher = this._launcher;
          var callback = this._intialFocus.bind(this, true);
          message = this.getTranslatedString("ariaFocusSkipLink");
          this._focusSkipLink = new oj.PopupSkipLink(launcher, message, callback, focusSkipLinkId);

          var content = this._content;
          var closeSkipLinkId = this._getSubId("closeSkipLink");
          callback = this._closeImplicitly.bind(this);
          message = this.getTranslatedString("ariaCloseSkipLink");
          this._closeSkipLink = new oj.PopupSkipLink(content, message, callback, closeSkipLinkId);
        }
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _destroyVoiceOverAssist : function ()
      {

        var liveRegion = this._liveRegion;
        liveRegion.destroy();
        delete this._liveRegion;

        var focusSkipLink = this._focusSkipLink;
        if (focusSkipLink)
        {
          focusSkipLink.destroy();
          delete this._focusSkipLink;
        }

        var closeSkipLink = this._closeSkipLink;
        if (closeSkipLink)
        {
          closeSkipLink.destroy();
          delete this._closeSkipLink;
        }
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {string} sub id that will become a composite id prefixed with the components uuid
       * @return {string}
       */
      _getSubId : function (sub)
      {
        /** @type {?} */
        var id = this.element.attr("id");
        if (oj.StringUtils.isEmptyOrUndefined(id))
          id = this["uuid"];
        return [id, sub].join("_");
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _surrogateRemoveHandler : function ()
      {
        var element = this.element;
        element.remove();
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @return {!Object.<oj.PopupService.EVENT, function(...)>}
       */
      _getPopupServiceEvents : function ()
      {
        if (!this._popupServiceEvents)
        {
          /** @type {!Object.<oj.PopupService.EVENT, function(...)>} **/
          var events = this._popupServiceEvents = {};
          events[oj.PopupService.EVENT.POPUP_CLOSE] = this._closeImplicitly.bind(this);
          events[oj.PopupService.EVENT.POPUP_REMOVE] = this._surrogateRemoveHandler.bind(this);
          events[oj.PopupService.EVENT.POPUP_REFRESH] = this.refresh.bind(this);
          events[oj.PopupService.EVENT.POPUP_BEFORE_OPEN] = this._beforeOpenHandler.bind(this);
          events[oj.PopupService.EVENT.POPUP_AFTER_OPEN] = this._afterOpenHandler.bind(this);
          events[oj.PopupService.EVENT.POPUP_BEFORE_CLOSE] = this._beforeCloseHandler.bind(this);
          events[oj.PopupService.EVENT.POPUP_AFTER_CLOSE] = this._afterCloseHandler.bind(this);
        }
        return this._popupServiceEvents;
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _closeImplicitly : function ()
      {
        this._ignoreBeforeCloseResultant = true;
        this.close();
        delete this._ignoreBeforeCloseResultant;
      },
      /**
       * Creates a Promise exposed by the {@link oj.ojPopup#whenReady} method.
       *
       * @param {string} operation valid values are "open", "close" or "none"
       * @memberof oj.ojPopup
       * @instance
       * @private
       */
      _setWhenReady : function (operation)
      {
        /** @type {oj.PopupWhenReadyMediator} */
        var mediator = this._whenReadyMediator;
        if (mediator)
        {
          mediator.destroy();
          delete this._whenReadyMediator;
        }

        // operation === none
        if (["open", "close"].indexOf(operation) < 0)
          return;

        this._whenReadyMediator = new oj.PopupWhenReadyMediator(this.element, operation, "ojPopup",
          this._IsCustomElement());
      },

      /**
       * Checks to see if there is a pending "open" or "close" operation.  If pending and it
       * is the same as the requested operation, the request silently fails.  If the current
       * operation is the inverse operation, we queue the current operation after the pending
       * operation is resolved.
       *
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {string} operation currently requested
       * @param {Array} args passed to a queue operation
       * @returns {boolean} <code>true</code> if a close or open operation is pending completion.
       */
      _isOperationPending : function (operation, args)
      {
        /** @type {oj.PopupWhenReadyMediator} **/
        var mediator = this._whenReadyMediator;
        if (mediator)
          return mediator.isOperationPending(this, operation, operation, args);
        else
          return false;
      },
      /**
       * @memberof oj.ojPopup
       * @instance
       * @private
       * @param {jQuery} elem to manage the focus ring on
       */
      _setupFocus : function (elem) {

        var self = this;
        this._focusable({
          'applyHighlight' : true,
          'setupHandlers' : function (focusInHandler, focusOutHandler) {
            self._on(elem, {
              focus : function (event) {
                focusInHandler($(event.currentTarget));
              },
              blur : function (event) {
                focusOutHandler($(event.currentTarget));
              }
            });
          }
        });
      },
      /**
       * Notifies the component that its subtree has been removed from the document
       * programmatically after the component has been created.
       *
       * @memberof oj.ojPopup
       * @instance
       * @protected
       * @override
       */
      _NotifyDetached: function()
      {
        // detaching an open popup results in implicit dismissal
        if (oj.ZOrderUtils.getStatus(this.element) === oj.ZOrderUtils.STATUS.OPEN)
          this._closeImplicitly();

        this._super();
      }
    });

    // sets the default modality option from the current theme
    oj.Components.setDefaultOptions(
      {
        'ojPopup' :
          {
            'modality' : oj.Components.createDynamicPropertyGetter(
              function ()
              {
                return (oj.ThemeUtils.parseJSONFromFontFamily('oj-popup-option-defaults')
                  || {})["modality"];
              }),
            'animation' : oj.Components.createDynamicPropertyGetter(
              function ()
              {
                return (oj.ThemeUtils.parseJSONFromFontFamily('oj-popup-option-defaults')
                  || {})["animation"];
              })
          }
      });
}());

// Fragments:

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Gesture</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>Outside popup or launcher</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Close the popup.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * <p>Disabled items do not allow any touch interaction.
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojPopup
   */

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td rowspan = "3">Focus within Popup</td>
   *       <td><kbd>Tab</kbd> or <kbd>Shift + Tab</kbd></td>
   *       <td>Navigate the content of the popup. Close the open popup if there are no tab stops in
   *           the popup.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>F6</kbd></td>
   *       <td>Move focus to the launcher for a popup with modeless modality.  Close the open popup
   *           if the modality is modal.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>Esc</kbd></td>
   *       <td>Close the open popup.</td>
   *     </tr>
   *     <tr>
   *       <td rowspan = "1">Popup Launcher</td>
   *       <td><kbd>F6</kbd></td>
   *       <td>Move focus to the first tab stop within the open popup.  If there is not a tab stop
   *           within the content, focus is established on the popup.</td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojPopup
   */

  /**
   * {@ojinclude "name":"ojStylingDocIntro"}
   *
   * <table class="generic-table styling-table">
   *   <thead>
   *     <tr>
   *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
   *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
   *       <th>{@ojinclude "name":"ojStylingDocExampleHeader"}</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>oj-focus-highlight</td>
   *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
   *       <td>
   * <pre class="prettyprint">
   * <code>&lt;oj-popup class="oj-focus-highlight">
   *  &lt;input id="blueopt" type="checkbox" value="blue">
   *  &lt;label for="blueopt">Blue&lt;/label>
   * &lt;/oj-popup>
   * </code></pre>
   *       </td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
   * @memberof oj.ojPopup
   */

(function () {
  var ojPopupMeta = {
    "properties" : {
      "autoDismiss" : {
        "type" : "string",
        "enumValues" : ["focusLoss", "none"]
      },
      "chrome" : {
        "type" : "string",
        "enumValues" : ["default", "none"]
      },
      "initialFocus" : {
        "type" : "string",
        "enumValues" : ["auto", "none", "firstFocusable", "popup"]
      },
      "modality" : {
        "type" : "string",
        "enumValues" : ["modeless", "modal"]
      },
      "position" : {
        "type" : "object",
        "properties" : {
          "my" : {
            "type" : "object|string",
            "properties" : {
              "horizontal" : {
                "type" : "string",
                "enumValues" : ["start", "end", "left", "center", "right"]
              },
              "vertical" : {
                "type" : "string",
                "enumValues" : ["top", "center", "bottom"]
              }
            }
          },
          "at" : {
            "type" : "object|string",
            "properties" : {
              "horizontal" : {
                "type" : "string",
                "enumValues" : ["start", "end", "left", "center", "right"]
              },
              "vertical" : {
                "type" : "string",
                "enumValues" : ["top", "center", "bottom"]
              }
            }
          },
          "offset" : {
            "type" : "object",
            "properties" : {
              "x" : {
                "type" : "number"
              },
              "y" : {
                "type" : "number"
              }
            }
          },
          "of" : {
            "type" : "string|object"
          },
          "collision" : {
            "type" : "string",
            "enumValues" : ["flip", "fit", "flipfit", "flipcenter", "none"]
          }
        }
      },
      "tail" : {
        "type" : "string",
        "enumValues" : ["simple", "none"]
      }
    },
    "events" : {
      "beforeOpen" : {},
      "open" : {},
      "beforeClose" : {},
      "close" : {},
      "focus" : {},
      "animateStart" : {},
      "animateEnd" : {}
    },
    "methods" : {
      "close" : {},
      "isOpen" : {},
      "open" : {}
    },
    "extension" : {
      _WIDGET_NAME : "ojPopup"
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-popup', 'baseComponent', ojPopupMeta);
  oj.CustomElementBridge.register('oj-popup', {'metadata' : oj.CustomElementBridge.getMetadata('oj-popup')});
})();
/*jslint browser: true*/
/*
 ** Copyright (c) 2004, 2012, Oracle and/or its affiliates. All rights reserved.
 */

/**
 * Internal framework service for managing popup instances.
 *
 * @extends {oj.Object}
 * @protected
 * @constructor
 * @since 1.1.0
 * @class oj.PopupService
 * @ignore
 * @ojtsignore
 */
oj.PopupService = function ()
{
  this.Init();
};

oj.Object.createSubclass(oj.PopupService, oj.Object, "oj.PopupService");

/**
 * @override
 * @instance
 * @protected
 */
oj.PopupService.prototype.Init = function ()
{
  oj.PopupService.superclass.Init.call(this);
};

/**
 * @param {Object=} options used by the factory method for service instantiation
 * @return {!oj.PopupService} singleton instance of the manager
 * @public
 */
oj.PopupService.getInstance = function (options)
{
  // in the future we might need a JET Island impl
  if (!oj.PopupService._popupService)
  {
    oj.PopupService._popupService = new oj.PopupServiceImpl();
  }
  return oj.PopupService._popupService;
};

/**
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag for opening the popup
 * @return {void}
 * @instance
 * @public
 */
oj.PopupService.prototype.open = function (options)
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag for closing the popup
 * @return {void}
 * @instance
 * @public
 */
oj.PopupService.prototype.close = function (options)
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag to change
 * the state of the target popup.
 * @return {void}
 * @instance
 * @public
 */
oj.PopupService.prototype.changeOptions = function (options)
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * @param {!jQuery} popup to target triggering the event on decendents
 * @param {!oj.PopupService.EVENT} event to trigger
 * @param {Array=} argsArray to pass to the associated callback for the event
 * @return {void}
 * @instance
 * @public
 */
oj.PopupService.prototype.triggerOnDescendents = function (popup, event, argsArray)
{
  oj.Assert.failedInAbstractFunction();
};

/**
 * @return {void}
 * @instance
 * @public
 */
oj.PopupService.prototype.destroy = function ()
{
  oj.PopupService._popupService = null;
};

/**
 * Dialog modality states.
 * @enum {string}
 * @public
 * @ojtsignore
 */
oj.PopupService.MODALITY =
  {
    /** Type of popup that doesn't support modality */
    NONE : "none",
    /** Dialog that that blocks user input of the primary window.*/
    MODAL : "modal",
    /** Type of dialog that doesn't block user input of the primary window */
    MODELESS : "modeless"
  };

/**
 * Event names used to identify the {@link oj.PopupService.OPTION.EVENT} option
 * property.
 * @enum {string}
 * @public
 * @ojtsignore
 */
oj.PopupService.EVENT =
  {
    /**
     * Event called by the popup service when the surrogate is removed in the document
     * resulting in the popup getting implicitly closed and associated bound element
     * is removed */
    POPUP_REMOVE : "ojPopupRemove",
    /**
     * Event called when a parent popup is closed causing implicit closure of
     * descendent popups.
     */
    POPUP_CLOSE : "ojPopupClose",
    /**
     * Event called on when a parent popup is refreshed triggering a cascade
     * refresh on children.
     */
    POPUP_REFRESH : "ojPopupRefresh",
    /**
     * Event called to enforce auto dismissal rules specific to each popup category.
     */
    POPUP_AUTODISMISS : "ojPopupAutoDismiss",
    /**
     * Event called before the popup is open but after it has been reparented into the
     * zorder container.  The callback should be used for implementing open animation.
     * The callback function should return a Promise if animation is reqired or void/undefined
     * if no animation is necessary. The callback is passed the open options
     * {!Object.<oj.PopupService.OPTION, ?>} as the only argument.
     * @since 3.0.0
     */
    POPUP_BEFORE_OPEN : "ojPopupBeforeOpen",
    /**
     * Event called after the popup is open. The callback should implement component
     * open finalization actions such as triggering an open event.  The resultant of
     * the callback function is void.  The callback is passed the
     * open options {!Object.<oj.PopupService.OPTION, ?>} as the only argument.
     */
    POPUP_AFTER_OPEN : "ojPopupAfterOpen",
    /**
     * Event called before the popup is close.  When invoked, the popup dom is still
     * located within the zorder container.  The callback is a good place to implement
     * close animation. The callback should return a Promise if animation is required
     * or void/undefined if no animation is required.  The callback is passed the
     * close options {!Object.<oj.PopupService.OPTION, ?>} as the only argument.
     * @since 3.0.0
     */
    POPUP_BEFORE_CLOSE : "ojPopupBeforeClose",
    /**
     * Event called after the popup is closed and reparented back into its original
     * location within the document.  The callback is for close finalization logic.
     * It's a good places to trigger a component close event. The resultant of the
     * callback is expected to be void. The callback is passed the
     * close options {!Object.<oj.PopupService.OPTION, ?>} as the only argument.
     * @since 3.0.0
     */
    POPUP_AFTER_CLOSE : "ojPopupAfterClose"
  };

/**
 * Layer level used to identify the {@link oj.PopupService.OPTION.LAYER_LEVEL} option
 * property.
 * @enum {string}
 * @public
 * @ojtsignore
 */
oj.PopupService.LAYER_LEVEL =
  {
    /**
     * Option used by dialogs.  Dialogs are always top rooted.
     */
    TOP_LEVEL : "topLevel",
    /**
     * The default layer option.  Popups will be reparented to the nearest ancestor layer defined
     * relative to the associated {@link oj.PopupService.OPTION.LAUNCHER}.
     */
    NEAREST_ANCESTOR : "nearestAncestor"
  };

/**
 * Property names in the options property bag passed to popup service api.
 * @enum {string}
 * @public
 * @see oj.PopupService#close
 * @see oj.PopupService#open
 * @ojtsignore
 */
oj.PopupService.OPTION =
  {
    /**
     * Parameter holding the jQuery element that is the root of the popup.  This
     * element is reparented into the zorder container when open. It is a required
     * option.
     */
    POPUP : "popup",
    /**
     * Map of event names to callbacks.  The event names are defined by the
     * {@link oj.PopupService.EVENT} enumerated type.
     */
    EVENTS : "events",
    /**
     * Defines the modal state of a popup.  The value of this attribute is defined
     * by the {@link oj.PopupService.MODALITY} enumeration.
     */
    MODALITY : "modality",
    /**
     * The jQuery element that is associated with the popup being open.  The launcher
     * is used to find the target popups reparented location within the zorder
     * container when open.  This is an optional parameter.  Dialogs don't require a launcher.
     */
    LAUNCHER : "launcher",
    /**
     * The jQuery UI position object that defines where the popup should be aligned.
     * This is an optional parameter.
     */
    POSITION : "position",
    /**
     * The CSS selector names applied to the layer for the target type of popup.  These
     * selectors will define the stacking context for the popup and its children.  Multiple
     * selector names should be delimited by a space similar to the syntax for the jquery
     * addClass API.  This option is required.
     */
    LAYER_SELECTORS : "layerSelectors",
    /**
     * The initial level that the popup will be reparented to when open.  Dialogs are reparented
     * into the top level.  Other types of popups will be parented to their nearest ancestor layer.
     * The values of this attribute are defined by {@link oj.PopupService.LAYER_LEVEL}.
     */
    LAYER_LEVEL : "layerLevel",
    /**
     * General purpose context that the "open" and "close" can add to the options and it will be
     * passed thru to the corresponding "before" and "after" operations. Use this context to pass
     * variables declared locally in the open to the associated before and after
     * callbacks.
     * @since 3.0.0
     */
    CONTEXT : "context",

    /**
     * Indicates the component with the associated popup was instantiated as a custom element.
     * This switch determines how the associated surrogate will be created.
     * @since 4.0.0
     */
    CUSTOM_ELEMENT: "customElement"
  };

// -----------------------------------------------------------------------------

/**
 * @extends {oj.PopupService}
 * @public
 * @constructor
 * @since 1.1.0
 * @ignore
 * @ojtsignore
 */
oj.PopupServiceImpl = function ()
{
  this.Init();
};

oj.Object.createSubclass(oj.PopupServiceImpl, oj.PopupService, "oj.PopupServiceImpl");

/**
 * Establishes a popup to be open and managed by the framework.  Managed popups will
 * by reparented to the zorder container appended to the document body.  The
 * location within that container is determined by the launcher or modality options.
 *
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag for opening the popup
 * @return {void}
 * @instance
 * @public
 * @override
 */
oj.PopupServiceImpl.prototype.open = function (options)
{
  oj.Assert.assertObject(options);

  /** @type {!jQuery} */
  var popup = options[oj.PopupService.OPTION.POPUP];
  oj.Assert.assertPrototype(popup, $);

  // Trying to open a popup that is either already opening, open or closing.
  // Evaulate if global dom listeners are still needed and ignore the request.
  // This is generally and indicator something wacky has happened that needs recovery.
  var status = oj.ZOrderUtils.getStatus(popup);
  if (status !== oj.ZOrderUtils.STATUS.UNKNOWN &&
      status !== oj.ZOrderUtils.STATUS.CLOSE)
  {
    this._assertEventSink();
    return;
  }

  /** @type {jQuery} */
  var launcher = options[oj.PopupService.OPTION.LAUNCHER];
  oj.Assert.assertPrototype(launcher, $);

  /** @type {Object} */
  var position = options[oj.PopupService.OPTION.POSITION];
  oj.Assert.assertObjectOrNull(position);

  /** @type {!Object.<oj.PopupService.EVENT, function(...)>} **/
  var events = options[oj.PopupService.OPTION.EVENTS];
  oj.Assert.assertObject(events);

  var modality = options[oj.PopupService.OPTION.MODALITY];
  if (!modality ||
    !(oj.PopupService.MODALITY.MODELESS === modality ||
      oj.PopupService.MODALITY.MODAL === modality))
    modality = oj.PopupService.MODALITY.NONE;

  var layerClass = options[oj.PopupService.OPTION.LAYER_SELECTORS];
  oj.Assert.assertString(layerClass);

  var isCustomElement = options[oj.PopupService.OPTION.CUSTOM_ELEMENT];

  var layerLevel = options[oj.PopupService.OPTION.LAYER_LEVEL];
  if (!layerLevel ||
    !(oj.PopupService.LAYER_LEVEL.TOP_LEVEL === layerLevel ||
      oj.PopupService.LAYER_LEVEL.NEAREST_ANCESTOR === layerLevel))
    layerLevel = oj.PopupService.LAYER_LEVEL.NEAREST_ANCESTOR;

  var beforeOpenCallback = events[oj.PopupService.EVENT.POPUP_BEFORE_OPEN];
  if (!beforeOpenCallback || !$.isFunction(beforeOpenCallback))
    beforeOpenCallback = oj.PopupServiceImpl._defaultBeforeOpenCallback;

  var afterOpenCallback = events[oj.PopupService.EVENT.POPUP_AFTER_OPEN];

  oj.ZOrderUtils.setStatus(popup, oj.ZOrderUtils.STATUS.OPENING);

  //set logical parent
  oj.DomUtils.setLogicalParent(popup, launcher);

  oj.ZOrderUtils.addToAncestorLayer(popup, launcher, modality, layerClass, layerLevel,
    isCustomElement);

  var _finalize = function ()
  {
    try
    {
      popup.removeAttr("aria-hidden");

      this._assertEventSink();
      oj.Components.subtreeShown(popup[0]);

    }
    catch (e)
    {
      oj.Logger.error("Error opening popup:\n%o", e);
    }
    finally
    {
      // invoke the after open callback if one is provided.
      if (afterOpenCallback)
        afterOpenCallback(options);

      // delay activating event callbacks until after open is resolved
      // preventing a race condition
      var layer = oj.ZOrderUtils.getFirstAncestorLayer(popup);
      oj.Assert.assertPrototype(layer, $);
      oj.ZOrderUtils.applyEvents(layer, events);
      oj.ZOrderUtils.setStatus(popup, oj.ZOrderUtils.STATUS.OPEN);
    }
  };
  _finalize = _finalize.bind(this);

  var resultant;
  try
  {
    resultant = beforeOpenCallback(options);
  }
  catch (e)
  {
    oj.Logger.error("Error before open popup:\n%o", e);
  }
  finally
  {
    if (resultant && resultant instanceof Promise)
      resultant.then(_finalize);
    else
      _finalize();
  }
};

/**
 * Default {@link oj.PopupService.EVENT.POPUP_BEFORE_OPEN} if one is not provided.
 * @private
 * @since 3.0.0
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag for closing the popup
 * @return {Promise|void}
 */
oj.PopupServiceImpl._defaultBeforeOpenCallback = function (options)
{
  /** @type {!jQuery} */
  var popup = options[oj.PopupService.OPTION.POPUP];
  oj.Assert.assertPrototype(popup, $);

  /** @type {Object} */
  var position = options[oj.PopupService.OPTION.POSITION];

  popup.show();
  if (position)
    popup["position"](position);

  return void(0);
};

/**
 * Closes a open popup managed by the framework.  The popup element is reparented
 * to its original location within the document.  Any open descendent popups are
 * implicitly closed.
 *
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag for closing the popup
 * @return {void}
 * @instance
 * @public
 * @override
 */
oj.PopupServiceImpl.prototype.close = function (options)
{
  oj.Assert.assertObject(options);

  /** @type {!jQuery} */
  var popup = options[oj.PopupService.OPTION.POPUP];
  oj.Assert.assertPrototype(popup, $);

  /** @type {!jQuery} */
  var layer = oj.ZOrderUtils.getOpenPopupLayer(popup);
  oj.Assert.assertPrototype(layer, $);

  /** @type {!Object.<oj.PopupService.EVENT, function(...)>} **/
  var events = options[oj.PopupService.OPTION.EVENTS];
  if (!events)
    events = options[oj.PopupService.OPTION.EVENTS] = oj.ZOrderUtils.getEvents(layer);
  else
    events = $.extend(oj.ZOrderUtils.getEvents(layer), events);

  // Popup is not in a open status or there are no events registered for the popup,
  // then it is opening, closing or already closed. Evaluate if the document level
  // dom listeners are still needed and ignore the request.
  if (oj.ZOrderUtils.getStatus(popup) !== oj.ZOrderUtils.STATUS.OPEN || !events)
  {
    this._assertEventSink();
    return;
  }

  var beforeCloseCallback = events[oj.PopupService.EVENT.POPUP_BEFORE_CLOSE];
  if (!beforeCloseCallback || !$.isFunction(beforeCloseCallback))
    beforeCloseCallback = oj.PopupServiceImpl._defaultBeforeCloseCallback;

  var afterCloseCallback = events[oj.PopupService.EVENT.POPUP_AFTER_CLOSE];

  oj.ZOrderUtils.setStatus(popup, oj.ZOrderUtils.STATUS.CLOSING);
  // Unregister events during before close callback
  oj.ZOrderUtils.applyEvents(layer, {});

  var _finalize = function ()
  {
    try
    {
      popup.hide();
      popup.attr("aria-hidden", "true");
      //reset position units
      popup.css({"top" : "auto", "bottom" : "auto", "left" : "auto", "right" : "auto"});

      oj.ZOrderUtils.removeFromAncestorLayer(popup);

      //remove the logical parent
      oj.DomUtils.setLogicalParent(popup, null);

      this._assertEventSink();
      oj.Components.subtreeHidden(popup[0]);
    }
    catch (e)
    {
      oj.Logger.error("Error closing popup:\n%o", e);
    }
    finally
    {
      if (afterCloseCallback && $.isFunction(afterCloseCallback))
        afterCloseCallback(options);

      oj.ZOrderUtils.setStatus(popup, oj.ZOrderUtils.STATUS.CLOSE);
    }

  };
  _finalize = _finalize.bind(this);

  var resultant;
  try
  {
    resultant = beforeCloseCallback(options);
  }
  catch (e)
  {
    oj.Logger.error("Error before close popup:\n%o", e);
  }
  finally
  {
    if (resultant && resultant instanceof Promise)
      resultant.then(_finalize);
    else
      _finalize();
  }
};

/**
 * Default {@link oj.PopupService.EVENT.POPUP_BEFORE_CLOSE} if one is not provided.
 *
 * @private
 * @since 3.0.0
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag for closing the popup
 * @return {Promise|void} undefined is returned indicating no animation (sync operation)
 */
oj.PopupServiceImpl._defaultBeforeCloseCallback = function (options)
{
  /** @type {!jQuery} */
  var popup = options[oj.PopupService.OPTION.POPUP];
  oj.Assert.assertPrototype(popup, $);

  popup.hide();
  return void(0);
};

/**
 * Applies a new {@link oj.PopupService.OPTION.EVENTS} callback linkage or
 * applies changes a popup dialogs {@link oj.PopupService.OPTION.MODALITY},
 * {@link oj.PopupService.OPTION.EVENTS}, or {@link oj.PopupService.OPTION.LAYER_SELECTORS}.
 * @param {!Object.<oj.PopupService.OPTION, ?>} options property bag including target popup
 * @return {void}
 * @instance
 * @public
 * @override
 */
oj.PopupServiceImpl.prototype.changeOptions = function (options)
{
  oj.Assert.assertObject(options);

  /** @type {!jQuery} */
  var popup = options[oj.PopupService.OPTION.POPUP];
  oj.Assert.assertPrototype(popup, $);

  if (oj.ZOrderUtils.getStatus(popup) !== oj.ZOrderUtils.STATUS.OPEN)
    return;

  /** @type {!jQuery} */
  var layer = oj.ZOrderUtils.getOpenPopupLayer(popup);
  oj.Assert.assertPrototype(layer, $);

  /** @type Object.<oj.PopupService.EVENT, function(...)> */
  var events = options[oj.PopupService.OPTION.EVENTS];
  if (events)
    oj.ZOrderUtils.applyEvents(layer, events);

  /** @type {oj.PopupService.MODALITY} */
  var modality = options[oj.PopupService.OPTION.MODALITY];
  if (modality)
    oj.ZOrderUtils.applyModality(layer, modality);

  /** @type {?} */
  var layerClass = options[oj.PopupService.OPTION.LAYER_SELECTORS];
  if (!oj.StringUtils.isEmptyOrUndefined(layerClass))
    layer.attr("class", layerClass);

};

/**
 * Triggers the target event defined on all open descendent popups.
 * @param {!jQuery} popup to target triggering the event on decendents
 * @param {!oj.PopupService.EVENT} event to trigger
 * @param {Array=} argsArray to pass "apply" to the associated callback for the event
 * @return {void}
 * @instance
 * @override
 * @public
 */
oj.PopupServiceImpl.prototype.triggerOnDescendents = function (popup, event, argsArray)
{
  // if the popup is not open, there are not descendents
  if (!oj.ZOrderUtils.isPopupOpen(popup))
    return;

  var context = {};
  context['event'] = event;
  context['argsArray'] = argsArray;

  /** @type {!jQuery} */
  var layer = oj.ZOrderUtils.getFirstAncestorLayer(popup);
  oj.ZOrderUtils.postOrderVisit(layer, this._triggerOnDescendentsVisitCallback, context);
};

/**
 * The {@link oj.ZOrderUtils.postOrderVisit} callback implementation for
 * {@link oj.PopupServiceImpl#triggerOnDescendents}.
 *
 * @param {!jQuery} layer
 * @param {!Object} context
 * @instance
 * @return {oj.ZOrderUtils.VISIT_RESULT}
 * @private
 * @see oj.PopupServiceImpl#triggerOnDescendents
 */
oj.PopupServiceImpl.prototype._triggerOnDescendentsVisitCallback = function (layer, context)
{
  var event = context['event'];
  var argsArray = context['argsArray'];

  var events = oj.ZOrderUtils.getEvents(layer);
  if (events && $.isFunction(events[event]))
    events[event].apply(this, argsArray);

  return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
};

/**
 * Depending on if popups are open, adds or removes event listeners redistributed
 * to open popups.
 * @return {void}
 * @instance
 * @private
 */
oj.PopupServiceImpl.prototype._assertEventSink = function ()
{

  var hasPopupsOpen = oj.ZOrderUtils.hasPopupsOpen();
  var callbackEventFilter = this._callbackEventFilter;

  if (!hasPopupsOpen && callbackEventFilter)
  {
    window.removeEventListener("resize", oj.PopupServiceImpl._refreshCallback, true);
    window.removeEventListener("scroll", oj.PopupServiceImpl._refreshCallback, true);

    var docElement = document.documentElement;
    docElement.removeEventListener("mousewheel", oj.PopupServiceImpl._refreshCallback, true);
    docElement.removeEventListener("DOMMouseScroll", oj.PopupServiceImpl._refreshCallback, true);

    this._callbackEventFilter = null;
    for (var i = 0; i < oj.PopupServiceImpl._REDISTRIBUTE_EVENTS.length; i++)
    {
      var event = oj.PopupServiceImpl._REDISTRIBUTE_EVENTS[i];
      docElement.removeEventListener(event, callbackEventFilter, true);
    }

    var simpleTapRecognizer = this._simpleTapRecognizer;
    if (simpleTapRecognizer)
    {
      simpleTapRecognizer.destroy();
      this._simpleTapRecognizer = null;
    }
  }
  else if (hasPopupsOpen && !callbackEventFilter)
  {
    window.addEventListener("resize", oj.PopupServiceImpl._refreshCallback, true);
    window.addEventListener("scroll", oj.PopupServiceImpl._refreshCallback, true);

    var docElement = document.documentElement;
    docElement.addEventListener("mousewheel", oj.PopupServiceImpl._refreshCallback, true);
    docElement.addEventListener("DOMMouseScroll", oj.PopupServiceImpl._refreshCallback, true);

    callbackEventFilter = this._callbackEventFilter = this._eventFilterCallback.bind(this);
    for (var i = 0; i < oj.PopupServiceImpl._REDISTRIBUTE_EVENTS.length; i++)
    {
      var event = oj.PopupServiceImpl._REDISTRIBUTE_EVENTS[i];
      docElement.addEventListener(event, callbackEventFilter, true);
    }

    if (oj.DomUtils.isTouchSupported())
      this._simpleTapRecognizer = new oj.SimpleTapRecognizer(callbackEventFilter);
  }
};

/**
 * Event callback for events defined by {@link oj.PopupServiceImpl._REDISTRIBUTE_EVENTS}.
 * This callback handles applying the ".oj-focus-within" selector to the popup that
 * is active.  It also handles redistributing events to open popups.
 *
 * @param {Event} event from document capture listeners
 * @return {void}
 * @instance
 * @private
 */
oj.PopupServiceImpl.prototype._eventFilterCallback = function (event)
{
  var target = $(event.target);

  var hasPopupsOpen = oj.ZOrderUtils.hasPopupsOpen();
  if (!hasPopupsOpen)
  {
    this._assertEventSink();
    return;
  }

  // Ignore mouse events on the scrollbar. FF and Chrome, raises focus events on the
  // scroll container too.
  if (oj.DomUtils.isChromeEvent(event) || ("focus" === event.type && !target.is(":focusable")))
    return;

  var defaultLayer = oj.ZOrderUtils.getDefaultLayer();
  if ("keydown" === event.type && oj.ZOrderUtils.hasModalDialogOpen() &&
    !oj.DomUtils.isAncestor(defaultLayer[0], target[0]))
  {
    // Inexpensive check to make sure that if a modal dialog is open,
    // we prevent a keydown outside the zorder layer that contains all
    // popups.  This handles the scenario where focus is placed in the
    // location bar and you start tabbing.  The browser will try to tab to
    // the first tabstop in the document.  Eat this event if it's under the
    // modal glass (not within the zorder container) and don't redistribute.

    oj.ZOrderUtils.eatEvent(event);
    return;
  }

  var targetWitinLayer = oj.ZOrderUtils.getFirstAncestorLayer(target);

  // toggle the oj-focus-within pseudo state
  if (defaultLayer[0] !== targetWitinLayer[0])
  {
    if (!targetWitinLayer.hasClass(oj.PopupServiceImpl._FOCUS_WITHIN_SELECTOR))
    {
      var lastFocusLayer = this._lastFocusLayer;
      if (lastFocusLayer)
        lastFocusLayer.removeClass(oj.PopupServiceImpl._FOCUS_WITHIN_SELECTOR);
      targetWitinLayer.addClass(oj.PopupServiceImpl._FOCUS_WITHIN_SELECTOR);
      this._lastFocusLayer = targetWitinLayer;
    }
  }
  else
  {
    // focus relinquished outside any managed popup
    var lastFocusLayer = this._lastFocusLayer;
    if (lastFocusLayer)
    {
      lastFocusLayer.removeClass(oj.PopupServiceImpl._FOCUS_WITHIN_SELECTOR);
      this._lastFocusLayer = null;
    }
  }

  // Don't redistribute a focus event targeted for an element that doesn't normally take focus.
  // Clicking on the scrollbars sometimes targets focus on containers with a -1 tabindex.
  // However, we still need to process the focus-within logic for this scenario.
  if ("focus" === event.type && "-1" === target.attr("tabindex"))
    return;

  // redistribute events for auto dismissal
  var context = {};

  // Capture all interesting event properties.  Similar to jQuery.event.fix.
  var props = {};
  for (var key in event)
  {
    if (oj.PopupServiceImpl._COPY_SAFE_EVENT_PROPERTIES[key] && !$.isFunction(event[key]))
      props[key] = event[key];
  }

  // Wrap a native event in a jQuery.Event
  context["event"] = $.Event(event, props);
  oj.ZOrderUtils.postOrderVisit(defaultLayer, oj.PopupServiceImpl._redistributeVisitCallback,
    context);
};

/**
 * The {@link oj.ZOrderUtils.postOrderVisit} callback for redistributing
 * {@link oj.PopupService.EVENT.POPUP_AUTODISMISS} events to open popups.
 * @param {!jQuery} layer
 * @param {!Object} context
 * @return {oj.ZOrderUtils.VISIT_RESULT}
 * @instance
 * @private
 * @see oj.PopupServiceImpl#_eventFilterCallback
 */
oj.PopupServiceImpl._redistributeVisitCallback = function (layer, context)
{
  var events = oj.ZOrderUtils.getEvents(layer);
  var event = context["event"];
  var eventType = event.type;

  if (events && $.isFunction(events[oj.PopupService.EVENT.POPUP_AUTODISMISS]))
    events[oj.PopupService.EVENT.POPUP_AUTODISMISS](event);

  return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
};

/**
 * Event listener added to the window for the resize and scroll events.
 *
 * @param {Event} event resize,scroll or mousewheel event
 * @return {void}
 * @private
 */
oj.PopupServiceImpl._refreshCallback = function (event)
{

  var refreshTimerId = oj.PopupServiceImpl._refreshTimerId;
  if (!isNaN(refreshTimerId))
    return;

  // Throttle redistributing the refresh listener to intervals of ? ms.
  // This will help performance for chatty events such as scroll.

  oj.PopupServiceImpl._refreshTimerId = window.setTimeout(function ()
  {
    oj.PopupServiceImpl._refreshTimerId = Number.NaN;
    var defaultLayer = oj.ZOrderUtils.getDefaultLayer();

    if ($.isFunction(window.requestAnimationFrame))
    {
      oj.PopupServiceImpl._afRequestId = window.requestAnimationFrame(function ()
      {
        oj.PopupServiceImpl._afRequestId = null;
        oj.ZOrderUtils.postOrderVisit(defaultLayer, oj.PopupServiceImpl._refreshVisitCallback);
      });
    }
    else
    {
      oj.ZOrderUtils.postOrderVisit(defaultLayer, oj.PopupServiceImpl._refreshVisitCallback);
    }
  }, oj.PopupServiceImpl._REFRESH_DELAY);
};

/**
 * The {@link oj.ZOrderUtils.postOrderVisit} callback for invoking the
 * {@link oj.PopupService.EVENT.POPUP_REFRESH} function for open popups.
 *
 * @param {!jQuery} layer
 * @param {!Object} context
 * @return {oj.ZOrderUtils.VISIT_RESULT}
 * @private
 * @see oj.PopupServiceImpl._refreshCallback
 */
oj.PopupServiceImpl._refreshVisitCallback = function (layer, context)
{
  // Only need to call on the first level of popups as they will recursively
  // call on children.
  var level = context["level"];
  if (level > 0)
    return oj.ZOrderUtils.VISIT_RESULT.REJECT;

  var events = oj.ZOrderUtils.getEvents(layer);
  if (events && $.isFunction(events[oj.PopupService.EVENT.POPUP_REFRESH]))
    events[oj.PopupService.EVENT.POPUP_REFRESH]();

  return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
};

/**
 * @return {void}
 * @instance
 * @public
 * @override
 */
oj.PopupServiceImpl.prototype.destroy = function ()
{
  oj.PopupServiceImpl.superclass.destroy.call(this);
};

/**
 * The pseudo select name applied to the popup that has focus.
 *
 * @const
 * @private
 * @type {string}
 */
oj.PopupServiceImpl._FOCUS_WITHIN_SELECTOR = "oj-focus-within";

/**
 * Array of events that are redistributed to open popups.
 *
 * @const
 * @private
 * @type {Array.<string>}
 */
oj.PopupServiceImpl._REDISTRIBUTE_EVENTS = ["focus", "mousedown", "keydown"];

/**
 * Map of event properties that are interesting to capture when creating a jQuery.Event
 * wrapper from a native event.  This is similar to what is done in jQuery.event.fix.
 * This list is needed to prevent warning messages for event properties that have been
 * deprecated.
 *
 * @const
 * @private
 * @type {Object}
 */
oj.PopupServiceImpl._COPY_SAFE_EVENT_PROPERTIES = {'altKey' : true, 'bubbles' : true,
  'cancelable' : true, 'ctrlKey' : true,
  'currentTarget' : true, 'eventPhase' : true, 'metaKey' : true,
  'relatedTarget' : true, 'shiftKey' : true, 'target' : true, 'timeStamp' : true,
  'view' : true, 'which' : true, 'button' : true, 'buttons' : true, 'clientX' : true,
  'clientY' : true, 'offsetX' : true, 'offsetY' : true, 'pageX' : true,
  'pageY' : true, 'screenX' : true, 'screenY' : true, 'toElement' : true,
  'char' : true, 'charCode' : true, 'key' : true, 'keyCode' : true};
/**
 * Milliseconds that is used to throttle processing of native resize, scroll and mousewheel
 * events. Dispatching refresh events to open popups will happen after the delay.
 * This is to guard against chatty events.
 *
 * @const
 * @private
 * @type {number}
 */
oj.PopupServiceImpl._REFRESH_DELAY = 10;

// -----------------------------------------------------------------------------

/**
 * Utilities used by the popup framework.
 * @since 1.1.0
 * @ignore
 * @ojtsignore
 */
oj.ZOrderUtils = {};

/**
 * Operation status for a target popup.
 * @enum {number}
 * @public
 * @see oj.ZOrderUtils.getStatus
 * @see oj.ZOrderUtils.setStatus
 * @ojtsignore
 */
oj.ZOrderUtils.STATUS =
{
  /** Node has not interacted with the popup service. */
  UNKNOWN : 0,
  /** Popup is in the process of opening. */
  OPENING : 1,
  /** Popup is currently open. */
  OPEN : 2,
  /** Popup is in the process of closing */
  CLOSING : 3,
  /** Popup previously open is now closed **/
  CLOSE : 4
};

/**
 * Key to store the current operation status of the popup.
 * @const
 * @private
 * @type {string}
 * @see oj.ZOrderUtils.getStatus
 * @see oj.ZOrderUtils.setStatus
 */
oj.ZOrderUtils._STATUS_DATA = "oj-popup-status";

/**
 * Returns the current operation status of the target popup element.
 * @param {jQuery|Element} popup
 * @returns {oj.ZOrderUtils.STATUS}
 * @see oj.ZOrderUtils.setStatus
 */
oj.ZOrderUtils.getStatus = function (popup)
{
  if (popup instanceof Element)
    popup = $(popup);

  /** @type {?} */
  var status = popup.data(oj.ZOrderUtils._STATUS_DATA);
  if (isNaN(status))
    return oj.ZOrderUtils.STATUS.UNKNOWN;
  else
    return status;
};

/**
 * Sets the current operational status of the popup element.
 * @param {jQuery|Element} popup
 * @param {oj.ZOrderUtils.STATUS} status
 * @see oj.ZOrderUtils.getStatus
 */
oj.ZOrderUtils.setStatus = function (popup, status)
{
  if (popup instanceof Element)
    popup = $(popup);

  if (status >= 1 && status <= 4)
    popup.data(oj.ZOrderUtils._STATUS_DATA, status);
};

/**
 * Accepts a launcher associated with the target popup being open.  The
 * resultant is the nearest layer or default zorder container that the
 * popup should be reparented to when open.  It can also return the
 * layer "stacking context" of a popup that is already open.
 *
 * @param {jQuery=} launcher associated with the target popup
 * @return {!jQuery}
 * @public
 */
oj.ZOrderUtils.getFirstAncestorLayer = function (launcher)
{
  // dialogs will not have launchers and will be top rooted
  if (!launcher)
    return oj.ZOrderUtils.getDefaultLayer();

  var parent = launcher;
  while (parent && parent.length > 0
    && parent.attr(oj.ZOrderUtils._SURROGATE_ATTR) !== oj.ZOrderUtils._DEFAULT_LAYER_ID)
  {
    if (oj.ZOrderUtils._hasSurrogate(parent[0]))
      return parent;
    else
      parent = parent.parent();
  }

  return oj.ZOrderUtils.getDefaultLayer();
};

/**
 * Returns the jQuery DIV element that represents the zorder container prepended to the
 * document body.  All open popups will be a descendent of the resultant node.
 *
 * @return {!jQuery}
 * @public
 */
oj.ZOrderUtils.getDefaultLayer = function ()
{
  /** @type {jQuery} */
  var defaultLayer = $(document.getElementById(oj.ZOrderUtils._DEFAULT_LAYER_ID));
  if (defaultLayer.length > 0)
    return defaultLayer;

  defaultLayer = $("<div>");
  defaultLayer.attr("role", "presentation");
  defaultLayer.attr("id", oj.ZOrderUtils._DEFAULT_LAYER_ID);
  defaultLayer.prependTo($(document.body)); //@HTMLUpdateOK; attach programmatic generated node

  return defaultLayer;
};

/**
 * Adds the target popup to the nearest ancestor layer within the zorder container.
 * A surrogate script element will be added to mark where the popup root element
 * was prior to reparenting.
 *
 * @param {!jQuery} popup root widget
 * @param {?jQuery} launcher associated with a popup
 * @param {!oj.PopupService.MODALITY} modality of the popup being open
 * @param {string} layerClass selector that defines the stacking context "z-index" of the popup
 * @param {oj.PopupService.LAYER_LEVEL} layerLevel defines where the popup will be reparented
 * @param {boolean} isCustomElement indicates if the owning component is a custom element
 * @return {void}
 * @public
 */
oj.ZOrderUtils.addToAncestorLayer = function (popup, launcher, modality, layerClass, layerLevel,
  isCustomElement)
{
  var popupDom = popup[0];
  if (oj.ZOrderUtils._hasSurrogate(popupDom.parentNode))
    throw new Error("JET Popup is already open - id: " + popupDom.getAttribute("id"));

  var ancestorLayer = oj.ZOrderUtils.getFirstAncestorLayer(layerLevel ===
    oj.PopupService.LAYER_LEVEL.TOP_LEVEL ? null : launcher);

  var layer = $("<div>");

  /** @type {?} */
  var popupId = popup.attr("id");
  if (oj.StringUtils.isEmptyOrUndefined(popupId))
    layer.uniqueId();
  else
    layer.attr("id", [popupId, "layer"].join("_"));

  layer.attr("role", "presentation");
  layer.addClass(layerClass);
  popup.after(layer);  //@HtmlUpdateOk

  var surrogate = oj.ZOrderUtils._createSurrogate(layer, isCustomElement);

  oj.Components.subtreeDetached(popupDom);
  popup.appendTo(layer);  //@HtmlUpdateOk

  // link the popup to the layer @see oj.ZOrderUtils.getOpenPopupLayer
  popup.data(oj.ZOrderUtils._LAYER_ID_DATA, layer.attr("id"));

  layer.appendTo(ancestorLayer);  //@HtmlUpdateOk
  oj.Components.subtreeAttached(popupDom);

  oj.ZOrderUtils.applyModality(layer, modality);
};

/**
 * Replaces the event callback map associated with an open popup.  The event
 * callbacks are used for auto dismissal or handling implicit dismissal when
 * the surrogate element associated with a layer is removed from the document.
 *
 * @param {!jQuery} layer of the target popup
 * @param {!Object.<oj.PopupService.EVENT, function(...)>} events map of event name to callback
 * @param {?Object=} surrogate saves position within the document where popup is
 *             defined
 * @return {void}
 * @public
 */
oj.ZOrderUtils.applyEvents = function (layer, events, surrogate)
{
  if (!surrogate)
  {
    /** @type {?} */
    var surrogateId = layer.attr(oj.ZOrderUtils._SURROGATE_ATTR);
    if (surrogateId)
      surrogate = $(document.getElementById(surrogateId));
  }

  layer.data(oj.ZOrderUtils._EVENTS_DATA, events);

  if (surrogate && surrogate.length > 0 &&
    events && $.isFunction(events[oj.PopupService.EVENT.POPUP_REMOVE]))
  {
    // if the surrogate script element gets replaced in the dom it will trigger closure of the
    // popup.
    oj.Components.setComponentOption(surrogate[0], "beforeDestroy",
      events[oj.PopupService.EVENT.POPUP_REMOVE]);
  }
};

/**
 * Returns the map of event callbacks associated with an open popup.
 *
 * @param {!jQuery} layer of an open popup
 * @return {!Object.<oj.PopupService.EVENT, function(...)>}
 * @public
 */
oj.ZOrderUtils.getEvents = function (layer)
{
  /** @type {?} */
  var events = layer.data(oj.ZOrderUtils._EVENTS_DATA);
  return events;
};

/**
 * Creates a script element before the target layer bound to the simple jquery UI
 * surrogate component.  Links this element to the layer by attribute named
 * {@link oj.ZOrderUtils._SURROGATE_ATTR}.
 *
 * @param {!jQuery} layer stacking context
 * @param {boolean} isCustomElement
 * @return {jQuery}
 * @private
 * @see oj.ZOrderUtils.addToAncestorLayer
 */
oj.ZOrderUtils._createSurrogate = function (layer, isCustomElement)
{
  var nodeName = "script";
  if (isCustomElement)
    nodeName = "oj-surrogate";

  /** @type {?} */
  var surrogate = $(document.createElement(nodeName));  //@HtmlUpdateOk;

  /** @type {?} */
  var layerId = layer.attr("id");
  if (!oj.StringUtils.isEmptyOrUndefined(layerId))
    surrogate.attr("id", [layerId, "surrogate"].join("_"));

  if (isCustomElement)
  {
    // programmatically created elements not managed by a binding stratagy like knockout
    // needs this attribute to signal the component should be created.
    surrogate.attr("data-oj-binding-provider", "none");
  }

  surrogate.insertBefore(layer);  //@HtmlUpdateOk

  if (!isCustomElement)
  {
    // create the jquery ui component bound to the script node
    surrogate["ojSurrogate"]();
  }

  /** @type {?} */
  var surrogateId = surrogate.attr('id');
  // loosely associate the popup to the surrogate element
  layer.attr(oj.ZOrderUtils._SURROGATE_ATTR, surrogateId);

  return surrogate;
};

/**
 * Reparents the layer after the associated surrogate script element removing
 * the surrogate script element.
 *
 * @param {!jQuery} layer stacking context of the popup
 * @return {void}
 * @private
 * @see oj.ZOrderUtils.removeFromAncestorLayer
 */
oj.ZOrderUtils._removeSurrogate = function (layer)
{
  /** @type {?} */
  var surrogateId = layer.attr(oj.ZOrderUtils._SURROGATE_ATTR);
  layer.removeAttr(oj.ZOrderUtils._SURROGATE_ATTR);

  /** @type {jQuery} */
  var surrogate = $(document.getElementById(surrogateId));
  layer.insertAfter(surrogate);  //@HtmlUpdateOk
  if (surrogate.length > 0)
    oj.Components.setComponentOption(surrogate[0], "beforeDestroy", null);

  surrogate.remove();
};

/**
 * Returns the layer associated with a popup.  The layer should be the immediate parent of an open
 * popup unless it was disconnected from the document.
 *
 * @param {!jQuery} popup
 * @returns {!jQuery} open popup's layer
 * @public
 */
oj.ZOrderUtils.getOpenPopupLayer = function (popup)
{
  /** @type {?} */
  var layer = popup.parent();
  if (!layer || layer.length === 0)
  {
    // the open popup has been detached from the layer before it was closed
    // use the backup pointer for better cleanup
    /** @type {?} */
    var layerId = popup.data(oj.ZOrderUtils._LAYER_ID_DATA);
    layer = $(document.getElementById(layerId));
  }

  return layer;
};

/**
 * Closes a open popup by reparenting it back to its original location within
 * the document marked by the surrogate.  Recursively calls the
 * {@link oj.PopupService.EVENT.POPUP_CLOSE} event callback on descendent popups.
 * @param {!jQuery} popup root widget
 * @return {void}
 * @public
 */
oj.ZOrderUtils.removeFromAncestorLayer = function (popup)
{
  var layer = oj.ZOrderUtils.getOpenPopupLayer(popup);

  oj.ZOrderUtils.preOrderVisit(layer, oj.ZOrderUtils._closeDescendantPopupsCallback);

  oj.ZOrderUtils._removeOverlayFromAncestorLayer(layer);

  layer.removeData(oj.ZOrderUtils._EVENTS_DATA);
  layer.removeData(oj.ZOrderUtils._MODALITY_DATA);
  popup.removeData(oj.ZOrderUtils._LAYER_ID_DATA);

  var popupDom = popup[0];
  oj.Components.subtreeDetached(popupDom);
  oj.ZOrderUtils._removeSurrogate(layer);

  // if the popup is not orphaned
  if (popupDom && popupDom.parentElement)
    oj.DomUtils.unwrap(popup, layer);
  else
    layer.remove();

  oj.Components.subtreeAttached(popupDom);
};

/**
 * The {@link oj.ZOrderUtils.preOrderVisit} callback that invokes the
 * {@link oj.PopupService.EVENT.POPUP_CLOSE} function for descendent
 * popups.
 * @param {!jQuery} layer to be dismissed
 * @param {!Object} context for visit tree
 * @return {oj.ZOrderUtils.VISIT_RESULT}
 * @private
 * @see oj.ZOrderUtils.removeFromAncestorLayer
 */
oj.ZOrderUtils._closeDescendantPopupsCallback = function (layer, context)
{
  var level = context['level'];
  // Only need to visit the immediate children as the children will recursively
  // close any open child popups.
  if (level > 0)
    return oj.ZOrderUtils.VISIT_RESULT.REJECT;

  var events = layer.data(oj.ZOrderUtils._EVENTS_DATA);
  if (events && $.isFunction(events[oj.PopupService.EVENT.POPUP_CLOSE]))
    events[oj.PopupService.EVENT.POPUP_CLOSE]();

  return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
};

/**
 * Handles adding or removing a sibling overlay blocking pane before the dialog
 * based on the modality option.  The overlay pane is associated with the dialog
 * vai a attribute pointing to the id of the overlay.
 *
 * @param {!jQuery} layer for the target popup
 * @param {oj.PopupService.MODALITY} modality
 * @return {void}
 * @public
 */
oj.ZOrderUtils.applyModality = function (layer, modality)
{
  /** @type {?} */
  var currModality = layer.data(oj.ZOrderUtils._MODALITY_DATA);
  layer.data(oj.ZOrderUtils._MODALITY_DATA, modality);

  if (oj.StringUtils.isEmptyOrUndefined(currModality))
  {
    if (oj.PopupService.MODALITY.MODAL === modality)
      oj.ZOrderUtils._addOverlayToAncestorLayer(layer);
    else
      oj.ZOrderUtils._removeOverlayFromAncestorLayer(layer);
  }
  else if (currModality === modality)
    return;
  else if (modality !== currModality && modality === oj.PopupService.MODALITY.MODAL)
    oj.ZOrderUtils._addOverlayToAncestorLayer(layer);
  else
    oj.ZOrderUtils._removeOverlayFromAncestorLayer(layer);
};

/**
 * @return {boolean} <code>true</code> if one or more modal dialogs are open
 * @public
 */
oj.ZOrderUtils.hasModalDialogOpen = function ()
{
  var defaultLayer = oj.ZOrderUtils.getDefaultLayer();
  var children = defaultLayer.children();
  var childrenCount = children.length;
  for (var i = childrenCount - 1; i > -1; i--)
  {
    var child = $(children[i]);
    if (child.hasClass(oj.ZOrderUtils._OVERLAY_SELECTOR))
      return true;
  }

  return false;
};

/**
 * Creates a overlay div assigned the {@link oj.ZOrderUtils._OVERLAY_SELECTOR}
 * selector inserted before the dialog layer.  The layer is associated to the
 * overlay by the {@link oj.ZOrderUtils._OVERLAY_ATTR} assigned to the root layer
 * element.
 * @param {!jQuery} layer root widget
 * @return {void}
 * @private
 * @see oj.ZOrderUtils.applyModality
 */
oj.ZOrderUtils._addOverlayToAncestorLayer = function (layer)
{
  /** @type {jQuery} */
  var overlay = $("<div>");
  overlay.addClass(oj.ZOrderUtils._OVERLAY_SELECTOR);
  overlay.addClass(layer[0].className);
  overlay.attr("role", "presentation");

  /** @type {?} */
  var layerId = layer.attr("id");
  if (oj.StringUtils.isEmptyOrUndefined(layerId))
    overlay.uniqueId();
  else
    overlay.attr("id", [layerId, "overlay"].join("_"));

  layer.before(overlay);  //@HtmlUpdateOk

  /** @type {?} */
  var overlayId = overlay.attr("id");
  layer.attr(oj.ZOrderUtils._OVERLAY_ATTR, overlayId);

};

/**
 * Removes the overlay associated with a modal dialog and removes the the attribute
 * that associates the popup with the overlay.
 *
 * @param {!jQuery} layer root widget
 * @return {void}
 * @private
 * @see oj.ZOrderUtils.applyModality
 */
oj.ZOrderUtils._removeOverlayFromAncestorLayer = function (layer)
{
  /** @type {?} */
  var overlayId = layer.attr(oj.ZOrderUtils._OVERLAY_ATTR);

  if (!oj.StringUtils.isEmptyOrUndefined(overlayId))
  {
    layer.removeAttr(oj.ZOrderUtils._OVERLAY_ATTR);
    var overlay = $(document.getElementById(overlayId));
    overlay.remove();
  }
};

/**
 * Resultant enumerated type used to control the popup visit (pre/post order)
 * traversal.
 * @enum {number}
 * @public
 * @see oj.ZOrderUtils.postOrderVisit
 * @see oj.ZOrderUtils.preOrderVisit
 * @ojtsignore
 */
oj.ZOrderUtils.VISIT_RESULT =
  {
    /** Continue to descend into current subtree. */
    ACCEPT : 0,
    /** Halt processing the subtree but contine visiting */
    REJECT : 1,
    /** Halt tree visit **/
    COMPLETE : 2
  };

/**
 * Defines the visit traversal type.
 * @enum {number}
 * @private
 */
oj.ZOrderUtils._VISIT_TRAVERSAL =
  {
    /** The callback is invoked on the target popup before any children. */
    PRE_ORDER : 0,
    /** The callback is invoked on the target popup after first visiting all descendents. */
    POST_ORDER : 1
  };

/**
 * Visits all open popups invoking the callback function on the target popup after first
 * visiting all children in order of last open.
 * @param {!jQuery} layer to begin searching for popups
 * @param {function(!jQuery, !Object) : oj.ZOrderUtils.VISIT_RESULT} callback invoked for each
 *        child popup
 * @param {Object=} context passed to the visit
 * @return {void}
 * @public
 */
oj.ZOrderUtils.postOrderVisit = function (layer, callback, context)
{
  if (!context)
    context = {};

  context["level"] = 0;
  context["type"] = oj.ZOrderUtils._VISIT_TRAVERSAL.POST_ORDER;
  oj.ZOrderUtils._visitTree(layer, callback, context);
};

/**
 * Visits all open popups invoking the callback on the target popup before any
 * popups that are descendents.
 * @param {!jQuery} layer to begin searching for popups
 * @param {function(!jQuery, !Object) : oj.ZOrderUtils.VISIT_RESULT} callback invoked for each child
 *         popup
 * @param {Object=} context passed to the visit
 * @return {void}
 * @public
 */
oj.ZOrderUtils.preOrderVisit = function (layer, callback, context)
{
  if (!context)
    context = {};

  context["level"] = 0;
  context["type"] = oj.ZOrderUtils._VISIT_TRAVERSAL.PRE_ORDER;
  oj.ZOrderUtils._visitTree(layer, callback, context);
};

/**
 * Visits popups in order defined by {@link oj.ZOrderUtils._VISIT_TRAVERSAL.PRE_ORDER}
 * invoking the callback for each popup.
 * @param {!jQuery} layer to begin searching for popups
 * @param {function(!jQuery, !Object) : oj.ZOrderUtils.VISIT_RESULT} callback
 *        invoked for each child popup
 * @param {!Object} context passed to the visit
 * @return {oj.ZOrderUtils.VISIT_RESULT} instructions on how to proceed
 * @private
 * @see oj.ZOrderUtils.preOrderVisit
 * @see oj.ZOrderUtils.postOrderVisit
 */
oj.ZOrderUtils._visitTree = function (layer, callback, context)
{
  // Patterned from the RC visit APIs
  var level = context["level"];

  var children = layer.children();
  var childrenCount = children.length;
  for (var i = childrenCount - 1; i > -1; i--)
  {
    var child = $(children[i]);
    if (!oj.ZOrderUtils._hasSurrogate(child[0]))
      continue;

    var vrtn;

    // handle visit pre-order
    if (context["type"] === oj.ZOrderUtils._VISIT_TRAVERSAL.PRE_ORDER)
    {
      vrtn = callback(child, context);
      if (vrtn === oj.ZOrderUtils.VISIT_RESULT.COMPLETE)
        return vrtn;
      else if (vrtn === oj.ZOrderUtils.VISIT_RESULT.REJECT)
        break;
    }

    // visit children
    context["level"] = level + 1;
    vrtn = oj.ZOrderUtils._visitTree(child, callback, context);
    context["level"] = level;
    if (vrtn === oj.ZOrderUtils.VISIT_RESULT.COMPLETE)
      return vrtn;

    // handle visit post-order
    if (context["type"] === oj.ZOrderUtils._VISIT_TRAVERSAL.POST_ORDER)
    {
      vrtn = callback(child, context);
      if (vrtn === oj.ZOrderUtils.VISIT_RESULT.COMPLETE)
        return vrtn;
      else if (vrtn === oj.ZOrderUtils.VISIT_RESULT.REJECT)
        break;
    }
  }

  return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
};


/**
 * Determines the target element is an open popup by checking for the
 * {@link oj.ZOrderUtils._SURROGATE_ATTR} attribute assigned to open popup layers.
 *
 * @param {!Element} element to check for a stand-in component
 * @return {boolean} <code>true</code> if the element is associated with a placeholder element
 * @private
 */
oj.ZOrderUtils._hasSurrogate = function (element)
{
  if (element && element.nodeType === 1 && element.hasAttribute(oj.ZOrderUtils._SURROGATE_ATTR))
    return true;
  else
    return false;
};

/**
 * @return {boolean} <code>true</code> if one or more popups are open
 * @public
 */
oj.ZOrderUtils.hasPopupsOpen = function ()
{
  var defaultLayer = oj.ZOrderUtils.getDefaultLayer();
  return defaultLayer.children().length > 0;
};

/**
 * @return {number} total number of open popups
 * @public
 */
oj.ZOrderUtils.getOpenPopupCount = function ()
{
  var context = {};
  context["popupCount"] = 0;

  var defaultLayer = oj.ZOrderUtils.getDefaultLayer();
  oj.ZOrderUtils.preOrderVisit(defaultLayer, oj.ZOrderUtils._openPopupCountCallback, context);

  return context["popupCount"];
};

/**
 * The {@link oj.ZOrderUtils.preOrderVisit} callback for counting the total number of
 * open popups.
 * @param {!jQuery} layer
 * @param {!Object} context for visit tree
 * @return {oj.ZOrderUtils.VISIT_RESULT}
 * @private
 * @see oj.ZOrderUtils.getOpenPopupCount
 */
oj.ZOrderUtils._openPopupCountCallback = function (layer, context)
{
  context["popupCount"] = context["popupCount"] + 1;
  return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
};

/**
 * Returns a jQuery set of all open popup layer dom elements.  Popups open last will appear at the
 * end of the set. Used by automated testing.
 * @return {!jQuery} set of all open popup root elements managed by the popup service
 * @public
 */
oj.ZOrderUtils.findOpenPopups = function ()
{
  var context = {};

  /** @type {Array.<Element>} */
  var popups = [];

  context["popups"] = popups;
  var defaultLayer = oj.ZOrderUtils.getDefaultLayer();
  oj.ZOrderUtils.preOrderVisit(defaultLayer, oj.ZOrderUtils._openPopupsCallback, context);
  popups = context["popups"];

  return $(popups);
};

/**
 * The {@link oj.ZOrderUtils.preOrderVisit} callback that collects a set of open
 * popups.
 * @param {!jQuery} layer
 * @param {!Object} context for visit tree
 * @return {oj.ZOrderUtils.VISIT_RESULT}
 * @private
 * @see oj.ZOrderUtils.findOpenPopups
 */
oj.ZOrderUtils._openPopupsCallback = function (layer, context)
{
  /** @type {Array.<Element>} */
  var popups = context["popups"];
  popups.push(layer[0]);
  return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
};

/**
 * @public
 * @param {!Element} element to determine if it's above the top modal popup
 * @returns {boolean} returns <code>true</code> if the element is above the top modal popup
 */
oj.ZOrderUtils.isAboveTopModalLayer = function(element)
{
  /**
   * @return {Element|undefined}
   */
  function getTopLayer()
  {
    // traverses the first level of popups looking for the popup with the highest
    // stacking context.
    // pre-order traversal
    function callback(layer, context)
    {
      var level = context['level'];

      // first level traversal only
      if (level > 0)
        return oj.ZOrderUtils.VISIT_RESULT.REJECT;

      var prevLayer = context['topLayer'];
      if (prevLayer)
      {
        // if the current layer has a higher context than the prev, it becomes the top
        if (oj.ZOrderUtils.compareStackingContexts($(layer), $(prevLayer)) > 0)
          context['topLayer'] = layer;
      }
      else
        context['topLayer'] = layer;
    }

    var context = {"topLayer": null};
    var defaultLayer = oj.ZOrderUtils.getDefaultLayer();
    oj.ZOrderUtils.preOrderVisit(defaultLayer, callback, context);
    if (context["topLayer"])
      return context["topLayer"][0];
    else
      return undefined;
  }

  /**
   *
   * @param {!Element} topLayer
   * @return {Element|undefined}
   */
  function getTopModalLayer(topLayer)
  {
    // post-order traversal starting at the topLayer.  Retuns the modal popup at the
    // highest deepest level.
    function callback(layer, context)
    {
      if (layer[0].hasAttribute(oj.ZOrderUtils._OVERLAY_ATTR))
      {
        context["topModalPopup"] = layer;
        return oj.ZOrderUtils.VISIT_RESULT.COMPLETE;
      }
      else
        return oj.ZOrderUtils.VISIT_RESULT.ACCEPT;
    }

    var context = {"topModalPopup": null};
    if (topLayer.hasAttribute(oj.ZOrderUtils._OVERLAY_ATTR))
      context["topModalPopup"] = $(topLayer);

    oj.ZOrderUtils.postOrderVisit($(topLayer), callback, context);

    if (context["topModalPopup"])
      return context["topModalPopup"][0];
    else
      return undefined;
  }

  // inexpensive prerequisite check
  if (!element || !oj.ZOrderUtils.hasPopupsOpen())
    return true;

  // Popups closest to the body will have the greatest stacking context weight.
  // Find the top first level layer.
  var topLayer = getTopLayer();
  if (!topLayer)
    return true;

  // look for the last modal popup open starting from the top layer
  var topModalLayer = getTopModalLayer(topLayer);
  if (!topModalLayer)
    return true;

  // Returns true if the target element is a child of the top most modal layer.
  return oj.DomUtils.isAncestorOrSelf(topModalLayer, element) ||
         oj.ZOrderUtils.compareStackingContexts($(topModalLayer), $(element)) < 0;
};

/**
 * Utility used for testing. Compares two jquery singleton wappered elements
 * determining which element has the greatest stacking context.
 * @public
 * @param {jQuery} el1 first element to compare
 * @param {jQuery} el2 second element to compare
 * @return {number} 0 if elements have the same stacking context;
 *                  1 if the first element has a greater stacking context;
 *                 -1 when the second element has a greater stacking context;
 */
oj.ZOrderUtils.compareStackingContexts = function (el1, el2)
{
  oj.Assert.assertPrototype(el1, $);
  oj.Assert.assertPrototype(el2, $);

  function describeStackingContext (element, allLevels) {
    var positions = ['absolute', 'relative', 'fixed'];
    var parents = element.parents();

    var tmp = [];
    for (var i = parents.length - 1; i > - 1; i--)
      tmp.push($(parents[i]));
    parents = tmp;

    parents.push(element);

    var stack = [];
    var level = 0;
    for (var i = 0; i < parents.length; i++)
    {
      var parent = parents[i];
      var position = parent.css("position");
      var opacity = oj.DomUtils.getCSSLengthAsFloat(parent.css("opacity"));
      var zindex = oj.DomUtils.getCSSLengthAsInt(parent.css("z-index"));
      var order = $.inArray(parent[0], parent.parent().children());
      if ($.inArray(position, positions) > -1 && zindex > 0)
        stack.push({'weight' : [level++, zindex, order], 'order' : [order]});
      else if (opacity < 1)
        stack.push({'weight' : [level++, 1, order], 'order' : [order]});
      else
      {
        if (!allLevels)
          continue;
        else
          stack.push({'weight' : [0, 0, order], 'order' : [order]});
      }
    }
    return stack;
  }

  function compareSets (n1, n2) {
    var maxLen = Math.max(n1.length, n2.length);
    for (var i = 0; i < maxLen; i++)
    {
      var e1 = i < n1.length ? n1[i] : 0;
      var e2 = i < n2.length ? n2[i] : 0;
      if (e1 === e2)
        continue;
      else if (e1 < e2)
        return -1;
      else
        return 1;
    }
    return 0;
  }
  ;

  var n1 = describeStackingContext(el1, false);
  var n2 = describeStackingContext(el2, false);

  var maxLen = Math.max(n1.length, n2.length);
  for (var i = 0; i < maxLen; i++)
  {
    var e1 = i < n1.length ? n1[i]['weight'] : [-1];
    var e2 = i < n2.length ? n2[i]['weight'] : [-1];

    var c = compareSets(e1, e2);
    if (c === 0)
      continue;
    else
      return c;
  }

  // include all elements for tie breaker
  n1 = describeStackingContext(el1, true);
  n2 = describeStackingContext(el2, true);

  maxLen = Math.max(n1.length, n2.length);
  // tie breaker based on document order
  for (var i = 0; i < maxLen; i++)
  {
    e1 = i < n1.length ? n1[i]['order'] : [-1];
    e2 = i < n2.length ? n2[i]['order'] : [-1];

    var c = compareSets(e1, e2);
    if (c === 0)
      continue;
    else
      return c;
  }

  return 0;
};

/**
 * Event listener that will stop propagation and prevent default on the event.
 * @param {jQuery.Event|Event} event
 * @return {void}
 * @public
 */
oj.ZOrderUtils.eatEvent = function (event)
{
  event.stopPropagation();
  event.preventDefault();
};

/**
 * @public
 * @param {jQuery} popup jQuery element
 * @return {boolean} <code>true</code> if the popup is reparented into the zorder container
 */
oj.ZOrderUtils.isPopupOpen = function (popup)
{
  // only open popups will be associated with a surrogate via parent layer
  var parent = popup.parent();
  if (parent && parent.length === 1 && oj.ZOrderUtils._hasSurrogate(parent[0]))
    return true;
  else
    return false;
};

/**
 * Key used to store the popup event callbacks on the popup layer
 * as a jQuery data property.
 * @const
 * @private
 * @type {string}
 */
oj.ZOrderUtils._EVENTS_DATA = "oj-popup-events";

/**
 * Key used to store the modality of a popup layer as a jQuery data property.
 * @const
 * @private
 * @type {string}
 */
oj.ZOrderUtils._MODALITY_DATA = "oj-popup-modality";

/**
 * The id assigned to the zorder container that will house all open popups.
 * The zorder container will be appended to the document body.
 * @const
 * @private
 * @type {string}
 */
oj.ZOrderUtils._DEFAULT_LAYER_ID = "__oj_zorder_container";

/**
 * The attribute name assigned to the associated layer of open popups.  The value of the
 * attribute will point to the script element that holds the place in the document
 * that the popup originated from prior to being open.
 * @const
 * @private
 * @type {string}
 */
oj.ZOrderUtils._SURROGATE_ATTR = "data-oj-surrogate-id";

/**
 * This key that captures the popups layer id.  The layer should always be the immediate parent
 * of the popup after open but if the popup is disconnected from the layer, this is a secondary
 * link to the layer for better cleanup.
 *
 * @const
 * @private
 * @type {string}
 * @see oj.ZOrderUtils.getOpenPopupLayer
 */
oj.ZOrderUtils._LAYER_ID_DATA = "oj-popup-layer-id";

/**
 * The attribute name assigned to the popup layer for open dialogs that have a
 * modality state of "modal".  The value of the attribute points to the associated
 * overlay element.
 * @const
 * @private
 * @type {string}
 */
oj.ZOrderUtils._OVERLAY_ATTR = "data-oj-overlayid";

/**
 * The CSS selector name assigned to the modal overlay DIV.
 * @const
 * @private
 * @type {string}
 */
oj.ZOrderUtils._OVERLAY_SELECTOR = "oj-component-overlay";

// -----------------------------------------------------------------------------

oj.__registerWidget("oj.ojSurrogate", $['oj']['baseComponent'],
  {
    version: "1.0.0",
    widgetEventPrefix : "oj",
    options :
      {
        'beforeDestroy' : null
      },
    _ComponentCreate : function ()
    {
      this._super();
      this.element.uniqueId();
    },
    _invokeBeforeDestroy: function ()
    {
      var callback = this.options["beforeDestroy"];
      this.options["beforeDestroy"] = null;
      if (callback)
        callback();
    },
    _destroy : function ()
    {

      this._invokeBeforeDestroy();
      this.element.removeUniqueId();
      this._super();
    },
    _NotifyDetached: function()
    {
      this._invokeBeforeDestroy();
      this._super();
    }
  });

  var ojSurrogateMeta = {
   "properties" : {
      "beforeDestroy" : {
        "type" : "function"
      }
   },
   "extension" : {
      _WIDGET_NAME : "ojSurrogate"
   }
  };
  oj.CustomElementBridge.registerMetadata('oj-surrogate', 'baseComponent', ojSurrogateMeta);
  oj.CustomElementBridge.register('oj-surrogate', {'metadata' :
      oj.CustomElementBridge.getMetadata('oj-surrogate')});

/**
 * Invokes the callback function with the touchstart event if the touch sequence
 * resulted in a "Tap".  The goal is to distinguish a touchstart that doesn't result
 * in scroll.
 * @extends {oj.Object}
 * @public
 * @constructor
 * @class oj.SimpleTapRecognizer
 * @since 1.1.0
 * @param {function(!Event)} tapCallback function invoked when a Tap is detected
 * @ignore
 * @ojtsignore
 */
oj.SimpleTapRecognizer = function (tapCallback)
{
  this._tapCallback = tapCallback;
  this.Init();
};

oj.Object.createSubclass(oj.SimpleTapRecognizer, oj.Object, "oj.SimpleTapRecognizer");

/**
 * Sets up the touch listeners ready to fire the callback.
 * @override
 * @instance
 * @protected
 */
oj.SimpleTapRecognizer.prototype.Init = function ()
{
  oj.SimpleTapRecognizer.superclass.Init.call(this);
  var eventHandlerCallback = this._eventHandlerCallback = this._eventHandler.bind(this);

  var docElement = document.documentElement;
  for (var i = 0; i < oj.SimpleTapRecognizer._TOUCHEVENTS.length; i++)
    docElement.addEventListener(oj.SimpleTapRecognizer._TOUCHEVENTS[i], eventHandlerCallback, true);
};

/**
 * Keeps reference to the last touchstart event.  If at touchend is encountered before a
 * touchmove or touchcancel, then fire the tap callback with the touchstart event.
 * @private
 * @param {Event} event native touch event
 */
oj.SimpleTapRecognizer.prototype._eventHandler = function (event)
{
  var tapCallback = this._tapCallback;
  var eventType = event.type;
  if ("touchstart" === eventType)
  {
    this._touchStartEvent = event;
    this._touchStartEvent._tapStart = new Date().getTime();
  }
  else if ("touchmove" === eventType || "touchcancel" === eventType)
  {
    this._touchStartEvent = null;
  }
  else if ("touchend" === eventType)
  {
    if (this._touchStartEvent)
    {
      var tapStart = this._touchStartEvent._tapStart;
      if (!isNaN(tapStart))
      {
        var now = new Date().getTime();
        // if the period of ms between touchstart and touchend is less than the long touch
        // thresshold, invoke the callback
        if (now - tapStart < oj.SimpleTapRecognizer._PRESSHOLDTHRESSHOLD)
          tapCallback(this._touchStartEvent);
      }
      else
        tapCallback(this._touchStartEvent);
    }

    this._touchStartEvent = null;
  }
};

/**
 * Unregisters touch listeners and deletes references to callbacks.
 * @public
 */
oj.SimpleTapRecognizer.prototype.destroy = function ()
{
  this._tapCallback = null;

  var eventHandlerCallback = this._eventHandlerCallback;
  this._eventHandlerCallback = null;

  var docElement = document.documentElement;
  for (var i = 0; i < oj.SimpleTapRecognizer._TOUCHEVENTS.length; i++)
    docElement.removeEventListener(oj.SimpleTapRecognizer._TOUCHEVENTS[i],
      eventHandlerCallback, true);
};

/**
 * Touch events that we are interested in listening for.
 *
 * @const
 * @private
 * @type {Array.<string>}
 */
oj.SimpleTapRecognizer._TOUCHEVENTS = ["touchstart", "touchmove", "touchcancel", "touchend"];

/**
 * Period of milliseconds for determining a long tap.  The normal threshold is 750ms.
 * The auto dismissal listeners for capture events versus bubble by the rest of the
 * framework.  Make the window for determining a long tap shorter than normal
 *
 * @const
 * @private
 * @type {number}
 */
oj.SimpleTapRecognizer._PRESSHOLDTHRESSHOLD = 700;

/**
 * Utility for handling popup voice over messages sent to a aria live region.
 * @extends {oj.Object}
 * @public
 * @constructor
 * @since 1.1
 * @class oj.PopupLiveRegion
 * @ignore
 * @ojtsignore
 */
oj.PopupLiveRegion = function () {
  this.Init();
};

oj.Object.createSubclass(oj.PopupLiveRegion, oj.Object, "oj.PopupLiveRegion");

/**
 * Adds one to the reference counter instance.
 * @override
 * @instance
 * @protected
 */
oj.PopupLiveRegion.prototype.Init = function ()
{
  oj.PopupLiveRegion.superclass.Init.call(this);
  if (isNaN(oj.PopupLiveRegion._refCounter))
    oj.PopupLiveRegion._refCounter = 1;
  else
    ++oj.PopupLiveRegion._refCounter;
};

/**
 * Decrements the reference counter destroying the assocaited shared DOM aria
 * live region element when there are no longer any popups using it.
 * @instance
 * @public
 */
oj.PopupLiveRegion.prototype.destroy = function ()
{
  if (!isNaN(oj.PopupLiveRegion._refCounter))
  {
    --oj.PopupLiveRegion._refCounter;
    if (oj.PopupLiveRegion._refCounter < 1)
    {
      var liveRegion = $(document.getElementById(oj.PopupLiveRegion._POPUP_LIVE_REGION_ID));
      if (liveRegion.length > 0)
        liveRegion.remove();
    }
  }
};

/**
 * Sends a message to the aria live region for voice over mode.
 * @instance
 * @public
 * @param {string} message to be announce in the live region
 */
oj.PopupLiveRegion.prototype.announce = function (message)
{
  if (!oj.StringUtils.isEmpty(message))
  {
    var liveRegion = oj.PopupLiveRegion._getLiveRegion();
    liveRegion.children().remove();
    $("<div>").text(message).appendTo(liveRegion);  //@HtmlUpdateOk; the "messsage" comes from a
                                                    // translated string that can be overridden by
                                                    // an option on the ojPopup.  The jquery "text"
                                                    // function will escape script.
  }
};

/**
 * Creates or returns an existing aria live region used by popups.
 * @returns {jQuery} aria live region
 * @private
 */
oj.PopupLiveRegion._getLiveRegion = function ()
{
  var liveRegion = $(document.getElementById(oj.PopupLiveRegion._POPUP_LIVE_REGION_ID));
  if (liveRegion.length === 0)
  {
    liveRegion = $("<div>");
    liveRegion.attr({'id' : oj.PopupLiveRegion._POPUP_LIVE_REGION_ID, 'role' : 'log', 'aria-live'
        : 'polite', 'aria-relevant' : 'additions'});
    liveRegion.addClass("oj-helper-hidden-accessible");
    liveRegion.appendTo(document.body);  //@HtmlUpdateOk
  }
  return liveRegion;
};

/**
 * Id assigned to the popup aria live region dom element.
 * @const
 * @private
 * @type {string}
 */
oj.PopupLiveRegion._POPUP_LIVE_REGION_ID = "__oj_popup_arialiveregion";

/**
 * Utility that injects a hidden link relative to another for voice support
 * @class oj.PopupSkipLink
 * @extends {oj.Object}
 * @public
 * @constructor
 * @since 1.1
 * @ignore
 * @param {jQuery} sibling element to the new skip link element
 * @param {string} message text assigned to the skip link
 * @param {function(!Event)} callback fired for activation of the skip link
 * @param {string=} id assigned to the skiplink component
 * @ojtsignore
 */
oj.PopupSkipLink = function (sibling, message, callback, id)
{
  oj.Assert.assertPrototype(sibling, $);
  oj.Assert.assertString(message);
  oj.Assert.assertFunction(callback);
  oj.Assert.assertStringOrNull(id);

  this._sibling = sibling;
  this._message = message;
  this._callback = callback;
  this._id = id;
  this.Init();
};

oj.Object.createSubclass(oj.PopupSkipLink, oj.Object, "oj.PopupSkipLink");

/**
 * Creates an invisible anchor relative to the sibling and hooks up the activation callack.
 * @override
 * @instance
 * @protected
 */
oj.PopupSkipLink.prototype.Init = function ()
{
  oj.PopupSkipLink.superclass.Init.call(this);
  var sibling = this._sibling;
  var callback = this._callback;
  var message = this._message;
  this._message = null;
  var id = this._id;
  this._id = null;

  var link = $("<a>").attr({'tabindex' : '-1', 'href' : '#', 'role': 'link'});
  link.attr("id", id);
  link.addClass("oj-helper-hidden-accessible");
  link.text(message);
  link.insertAfter(sibling);  //@HtmlUpdateOk

  link.on("click",
      oj.PopupSkipLink._activateHandler.bind(this, callback));

  // This could only happen in some kind of simulator as this skip link is only used on the iOS
  // platform. Prevent click generated from an "enter" key press by eating the event.
  link.on("keydown keyup keypress", oj.PopupSkipLink._keyHandler);

  sibling.data(oj.PopupSkipLink._SKIPLINK_ATTR, link);
};

/**
 * Handles activation of the skiplink.  Cancels the click event.
 * @private
 * @param {?} listener
 * @param {Event} event
 */
oj.PopupSkipLink._activateHandler = function (listener, event)
{
  oj.ZOrderUtils.eatEvent(event);
  window.setImmediate(listener);
};

/**
 * Listener registered on the skip link element to prevent an enter key from generating a click
 * @private
 * @param {jQuery.Event|Event} event
 */
oj.PopupSkipLink._keyHandler = function (event)
{
  if (event.keyCode === $.ui.keyCode.ENTER)
    oj.ZOrderUtils.eatEvent(event);
};

/**
 * Removes the voice over skip link.
 * @instance
 * @public
 */
oj.PopupSkipLink.prototype.destroy = function ()
{
  var sibling = this._sibling;
  delete this._sibling;
  delete this._callback;

  if (sibling)
  {
    var link = sibling.data(oj.PopupSkipLink._SKIPLINK_ATTR);
    sibling.removeData(oj.PopupSkipLink._SKIPLINK_ATTR);
    if (link)
    {
      link.off("click keydown keyup keypress");
      link.remove();
    }
  }
};

/**
 * Returns the skip link jQuery element.
 * @instance
 * @public
 * @return {jQuery} skip link
 */
oj.PopupSkipLink.prototype.getLink = function ()
{
  /** @type {?} */
  var sibling = this._sibling;
  /** @type {jQuery} */
  var link;
  if (sibling)
    link = sibling.data(oj.PopupSkipLink._SKIPLINK_ATTR);
  return link;
};

/**
 * Data attribute name assigned to the sibling element that tracks a
 * reference for the associated skip link.
 * @const
 * @private
 * @type {string}
 */
oj.PopupSkipLink._SKIPLINK_ATTR = "oj-skiplink";

/**
 * Coordinate communications between an event being fulfilled and one or more promises
 * being resolved.  The window of time between the instance creation and the associated event
 * triggered is guarded by the {@link oj.BusyContext}.  The
 * @link{oj.PopupWhenReadyMediator#getWhenReadyPromise} promise will resolve when either the
 * target event is triggered or instance destroyed.
 * @class oj.PopupWhenReadyMediator
 * @extends {oj.Object}
 * @constructor
 * @since 3.0.0
 * @ignore
 * @param {jQuery} element to subscribe on the event type triggered on completion of the operation
 * @param {string} operation that completion will resolve one or more promises
 * @param {string} widgetName component constructor
 * @param {boolean} isCustomElement <code>true</code> if the widget is created as a custom element
 * @ojtsignore
 */
oj.PopupWhenReadyMediator = function (element, operation, widgetName, isCustomElement)
{
  this._element = element;
  this._operation = operation;
  this._widgetName = widgetName;
  this._isCustomElement = isCustomElement ? true : false;

  this.Init();
};

oj.Object.createSubclass(oj.PopupWhenReadyMediator, oj.Object, "oj.PopupWhenReadyMediator");

/**
 * Registers an event handler on the element associated with the target operation.
 * The event handler will resolve one or more pending promises.  The convention
 * is the operation will raise a "oj" + operation event upon completion. The
 * event hander is one and done - unregistered after first delivery.
 *
 * @override
 * @instance
 * @protected
 */
oj.PopupWhenReadyMediator.prototype.Init = function ()
{
  oj.PopupWhenReadyMediator.superclass.Init.call(this);
  this._resolvedQueue = [];
  this._callback = this._eventHandler.bind(this);

  var operation = this._operation;
  var tokens = ["oj"];
  if (this._isCustomElement)
  {
    tokens.push(operation.charAt(0).toUpperCase());
    tokens.push(operation.slice(1));
  }
  else
  {
    tokens.push(operation);
  }

  var eventType = this._eventType = tokens.join("");
  this._element.on(eventType, this._callback);

  // Add a busy state for the pending operation.  The busy state resolver will
  // be invoked when the resolved queue is delivered (operation completes).
  var busyContext = oj.Context.getContext(this._element[0]).getBusyContext();
  var options = {"description" : this._getBusyStateDescription.bind(this, this._element,
      this._operation, this._widgetName)};
  var resolve = busyContext.addBusyState(options);
  this.AddPromiseExecutor(resolve);

  // setup the when ready promise
  this._whenReadyPromise = new Promise(this.AddPromiseExecutor.bind(this));
};

/**
 * @private
 * @param {jQuery} element to subscribe on the event type triggered on completion of the operation
 * @param {string} operation that completion will resolve one or more promises
 * @param {string} widgetName component constructor
 * @returns {string} description of the busy state animation operation.
 */
oj.PopupWhenReadyMediator.prototype._getBusyStateDescription = function (element, operation,
  widgetName)
{
  return widgetName + " identified by '" + element.attr('id') + "' is busy animating on " +
    "the '" + operation + "' operation.";
};

/**
 * Resolves the pending promises.
 *
 * @private
 * @param {string=} operation override sent to the resolverdQueue
 */
oj.PopupWhenReadyMediator.prototype._deliverResolved = function (operation)
{
  // Critical section - the registered resolve queue is disconnect from the
  // instance state so that a race condition will not occur - resolve promoise
  // adding new operations to the same queue.

  var resolvedQueue = this._resolvedQueue;
  this._resolvedQueue = null;

  operation = !operation ? this._operation : operation;
  this._operation = null;

  for (var i = 0; i < resolvedQueue.length; i++)
  {
    try
    {
      resolvedQueue[i](operation);
    }
    catch (e)
    {
      oj.Logger.error("Error resolving whenReady promises:\n%o", e);
    }
  }

  this._whenReadyPromise = Promise.resolve("none");
};

/**
 * Force delivery of unresolved promises.
 */
oj.PopupWhenReadyMediator.prototype.destroy = function ()
{
  // If the promise is swapped (component is destroyed)
  // before the event is fired, resolve with a "none" operation.
  if (this._resolvedQueue)
    this._deliverResolved("none");

  if (this._callback)
  {
    var eventType = this._eventType;
    this._element.off(eventType, this._callback);
  }

  this._callback = null;
  this._element = null;
  this._operation = null;
  this._whenReadyPromise = null;
  this._widgetName = null;
  this._eventType = null;
};

/**
 * @returns {Promise} returns an instance of the current whenReadyPromise
 */
oj.PopupWhenReadyMediator.prototype.getWhenReadyPromise = function ()
{
  return this._whenReadyPromise;
};

/**
 * Event handler associated with completion of the target operation.
 * @private
 * @param {jQuery.Event} event
 */
oj.PopupWhenReadyMediator.prototype._eventHandler = function (event)
{
  if (event.target === this._element[0])
  {
    this._element.off(event.type, this._callback);
    this._deliverResolved();
    this._callback = null;
  }
};

/**
 * @private
 * @return {string} Returns the pending operation
 */
oj.PopupWhenReadyMediator.prototype._getPendingOperation = function ()
{
  return this._operation ? this._operation : "none";
};

/**
 * A function that will be passed to other functions via the arguments resolve and reject.
 * The resolve function will be invoked when the event associated with completion of the
 * target operation is delivered to the target element.
 *
 * @protected
 * @param {Function} resolve resultant function that will resovle a promise executor
 * @param {Function=} reject (not interested in the reject)
 */
oj.PopupWhenReadyMediator.prototype.AddPromiseExecutor = function (resolve, reject)
{
  if (this._resolvedQueue)
    this._resolvedQueue.push(resolve);
};

/**
 * Checks to see if there is a pending "open" or "close" operation.  If pending and it
 * is the same as the requested operation, the request silently fails.  If the current
 * operation is the inverse operation, we queue the current operation after the pending
 * operation is resolved.
 *
 * @param {Object} widgetInstance this mediator is negotiating on behalf of
 * @param {string} operation currently requested
 * @param {string} methodName that should be invoked on the widgetInstance if the operation is the
 *                 inverse of the pending operation
 * @param {Array} methodArgs passed to a queue method invocation
 * @returns {boolean} <code>true</code> if a "close" or "open" operation is pending completion.
 */
oj.PopupWhenReadyMediator.prototype.isOperationPending = function (widgetInstance, operation,
  methodName, methodArgs)
{
  var isPending = false;
  var widgetName = this._widgetName;
  var pendingOperation = this._getPendingOperation();
  if (operation === pendingOperation)
  {
    // Same request is already pending. Silently fail.
    oj.Logger.info("An %s instance invoked a '%s' operation while pending animation of " +
      "the same type of operation.  The second request will be ignored.", widgetName,
      operation);
    isPending = true;
  }
  else if ("none" !== pendingOperation)
  {
    oj.Logger.info("An %s instance invoked a '%s' operation while pending animation of a " +
      "'%s' operation. The second request will be invoked after the pending operation " +
      "completes."
      , widgetName, operation, pendingOperation);

    // Queue the operation after the pending operation has completed
    // register another resolve promise with the mediator that will be
    // call when the pending operation finishes.
    var promise = new Promise(this.AddPromiseExecutor.bind(this));
    promise.then(function ()
    {
      this[methodName].apply(this, methodArgs);
    }.bind(widgetInstance));
    isPending = true;
  }
  return isPending;
};
/*jslint browser: true*/
/*
** Copyright (c) 2004, 2012, Oracle and/or its affiliates. All rights reserved.
*/
/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * Utilities used in conjunction with the jquery positon utility.
 * @ignore
 * @class oj.PositionUtils
 * @ojtsignore
 */
oj.PositionUtils = {};

/**
 * <p>Of the properties on the position object, "my" and "at" are of interest. The base jQuery
 * horizontal alignment mnemonics are "right", "center" and "left". For better JET RTL
 * support we have added "start" and "end".  Depending on the rtl direction, "start" and
 * "end" will be replaced with "left" or "right". The resultant postion object will
 * be a new instance that extends the position passed as the first argument.
 *
 * <p>Likewise, JET supports "<" and ">" operators wherever "-" and "+" are supported.
 * The "<" value means "- in LTR; + in RTL", while the ">" value means "+ in LTR; - in RTL".
 * E.g. a "my" value of "start<40" shifts the menu 40px "startward," while a "my" value of
 * "start>40" shifts the menu 40px "endward."
 *
 * @param {Object} position source position object
 * @param {boolean} isRtl
 * @return {Object} position object that has normalized jquery horizontal mnemonics.
 */
oj.PositionUtils.normalizeHorizontalAlignment = function(position, isRtl)
{
  // This assertion prevents security testing: someone could pass in a bogus position
  //oj.Assert.assertObject(position, "position");
  var target = $.extend({}, position);
  for (var i = 0; i < oj.PositionUtils._ALIGN_RULE_PROPERTIES.length; i++)
  {
    var propName = oj.PositionUtils._ALIGN_RULE_PROPERTIES[i];
    var align = target[propName];
    if (!align)
      continue;

    if (oj.StringUtils.isString(align))
    {
      target[propName] = align.replace("start", (isRtl ? "right" : "left"))
                              .replace("end", (isRtl ? "left" : "right"))
                              .replace("<", (isRtl ? "+" : "-"))
                              .replace(">", (isRtl ? "-" : "+"));
    }
    else
    {
      for (var s = 0; s < oj.PositionUtils._SUB_ALIGN_RULE_PROPERTIES.length; s++)
      {
        var subPropName = oj.PositionUtils._SUB_ALIGN_RULE_PROPERTIES[s];
        var subAlign = align[subPropName];
        if (oj.StringUtils.isString(subAlign))
        {
            align[subPropName] = subAlign.replace("start", (isRtl ? "right" : "left"))
                              .replace("end", (isRtl ? "left" : "right"))
                              .replace("<", (isRtl ? "+" : "-"))
                              .replace(">", (isRtl ? "-" : "+"));
        }
      }
    }
  }

  return target;
};

/**
 * <p>In the jQuery UI [Position]{@link http://api.jqueryui.com/position/} utility,
 * the "of" field specifies the element or event relative to which the popup should
 * be positioned.
 *
 * <p>Popup components like Menu often need to position themselves relative to their
 * launcher (e.g. MenuButton), or the launching event (e.g. right-click event), both
 * of which can vary on a per-launch basis.  To facilitate specifying these policies
 * in advance, in a component option, it's useful to support "launcher" and "event"
 * keywords in the "of" field.  Without these keywords, the app would have to set
 * this field at each popup launch, via either the open() method or a beforeOpen listener,
 * at which point they could supply the actual event or launcher element for this
 * particular launch.
 *
 * <p>This method is intended to be called by the popup component at launch time.  It
 * takes an "of" field possibly containing these keywords, and the event and launcher
 * for the current launch, and returns a new "of" value where those keywords have
 * been replaced by the event or launcher, so that the resulting "position" object
 * is ready to be passed to the JQUI utility.
 *
 * <p>Since callers may have different needs, this method does not log if, say,
 * of==="event" && event==null.  Callers should log or throw if this condition
 * indicates application error.
 *
 * <p>While the code is simple, it's useful to centralize the logic to avoid subtle
 * differences in behavior among JET popup components.
 *
 * @param of the "of" field of a position object.  Defaults to "launcher" if null/undefined.
 * @param launcher launcher element
 * @param event event that opened the popup
 * @return normalized "of" value
 */
oj.PositionUtils.normalizePositionOf = function(of, launcher, event)
{
  return (of === "event")
    ? event
    : (of == null || of === "launcher")
      ? launcher
      : of;
};

// TODO: file a JQ bug and link to it here.
/**
 * On iOS and Android, the JQ Event object wrapping touch* events lacks pageX and pageY properties, which is contrary to the
 * contract [1].  This breaks JQ's position() API [2], which assumes that the contract is obeyed.  Specifically, it
 * relies on the pageX/Y fields of the Event object passed as the "of" field (and publicly docs that it does so).
 *
 * Per W3C [3], pageX/Y are found in originalEvent.touches[i] or originalEvent.changedTouches[i], where originalEvent is
 * the native (not JQ) event, and i is 0 for us.  In practice, for touchstart at least, iOS7 and 8 Mobile Safari, but
 * apparently not Android Chrome, also put pageX and pageY on the top-level native event, and the values seem to be the
 * same as those in the touches array.  We'll use the cross-platform W3C location.
 *
 * To workaround the JQ bug, popup components like Menu can call this method at open() time.  This method copies the properties
 * to the wrapper JQ event.
 *
 * [1] http://api.jquery.com/category/events/event-object/
 * [2] http://api.jqueryui.com/position/
 * [3] http://www.w3.org/TR/touch-events/#touch-interface et. seq.
 *
 * @private
 * @param event
 */
oj.PositionUtils._normalizeEventForPosition = function(event)
{
  $.each(["pageX", "pageY"], function (index, pagePos)
  {
    if (event && event[pagePos] === undefined && event.originalEvent)
    {
      var originalEvent = event.originalEvent;
      var type = originalEvent.type;
      var touchList = (type === "touchstart" || type === "touchmove")
          ? "touches"
          : (type === "touchend") ? "changedTouches" : null;

      if (touchList)
      {
        var firstTouch = originalEvent[touchList][0];
        if (firstTouch)
          event[pagePos] = firstTouch[pagePos];
      }
    }
  });
};

/**
 * @private
 * @const
 */
oj.PositionUtils._ALIGN_RULE_PROPERTIES = ['my', 'at'];

/**
 * @private
 * @const
 */
oj.PositionUtils._SUB_ALIGN_RULE_PROPERTIES = ['vertical', 'horizontal'];

/**
 * A common utilty that is designed be be called for a jquery ui position "using" callback
 * that will check to see if the target aligning "of" position is clipped in an overflow
 * container.  Used by popups that should auto dismiss when what they are aligning to is
 * no longer visible.  The aligning position can be either an element, event or a rect.
 *
 * @param {Object} props second argument to the jquery ui position "using" callback.
 * @returns {boolean} <code>true</code> if the point aligned is not totally visible in and overflow container
 */
oj.PositionUtils.isAligningPositionClipped = function(props)
{
  // Alignment can be to an element, event or a rect but we only care to make this
  // check if alignment is to an element - .
  if (props["target"] && props["target"]["height"] > 0 && props["target"]["width"] > 0)
  {
    // if the target has a width and height greater than zero then it's an element
    /** @type {jQuery} */
    var positionOf = props["target"]["element"];
    return !oj.PositionUtils.isWithinViewport(positionOf);
  }
  else
    return false;
};


/**
 * Returns <code>true</code> if the jquery element is visible within overflow an
 * overflow container. The check only considers statically positioned elements and
 * stops short of the body.
 *
 * The first positioned ancestor is treated as the root viewport. Positioned elements need be
 * compared with the window viewport versus its ancestors.  Visibility of positioned
 * elements also need to compare stacking contexts within the document to determine
 * what is on top "visible" - thus excluded from this check.
 *
 * @param {jQuery} element jquery element to test
 * @returns {boolean}
 */
oj.PositionUtils.isWithinViewport = function(element)
{

  function isVisible(alignBox, containerBox)
  {
    if (["hidden", "scroll", "auto"].indexOf(containerBox["overflowY"]) > -1)
    {
      // 1px fudge factor for rounding errors
      if ((alignBox["bottom"] - containerBox["top"]) < -1)
        return false;

      // find horizontal scrollbar size. always present when "scroll", or when "auto" and scrollWidth > innerWidth
      var scrollBarWidth = ((containerBox["overflowX"] === "auto" && containerBox["scrollWidth"] > containerBox["innerWidth"]) ||
              containerBox["overflowX"] === "scroll") ? oj.DomUtils.getScrollBarWidth() : 0;
      if ((containerBox["bottom"] - scrollBarWidth) - alignBox["top"] < 1)
        return false;
    }

    if (["hidden", "scroll", "auto"].indexOf(containerBox["overflowX"]) > -1)
    {
      // find vertical scrollbar width. always present when "scroll", or when "auto" and scrollHeight > innerHeight
      scrollBarWidth = ((containerBox["overflowY"] === "auto" && containerBox["scrollHeight"] > containerBox["innerHeight"]) || 
              containerBox["overflowY"] === "scroll") ? oj.DomUtils.getScrollBarWidth() : 0;
      
      // depending on ltr vs rtl, the vertical scrollbar can be on either side of the container, so only include the side its on
      if (((alignBox["right"] - (containerBox["left"] + (oj.DomUtils.getReadingDirection() === "rtl" ? scrollBarWidth : 0))) < -1) ||
          ((alignBox["left"] - (containerBox["right"] - (oj.DomUtils.getReadingDirection() === "ltr" ? scrollBarWidth : 0))) > -1))
        return false;
    }

    return true;
  };

  function hasOverflow(element)
  {
    return "visible" !== element.css("overflow-x") ||
           "visible" !== element.css("overflow-y");
  };

  function getRect(element)
  {
    var domElement = element[0];
    if (domElement.nodeType === 1)
    {
      var rec = $.extend({}, domElement.getBoundingClientRect());
      rec["overflowX"] = element.css("overflow-x");
      rec["overflowY"] = element.css("overflow-y");
      rec["innerHeight"] = element.innerHeight();
      rec["innerWidth"] = element.innerWidth();
      rec["scrollHeight"] = domElement["scrollHeight"];
      rec["scrollWidth"] = domElement["scrollWidth"];
      return rec;
    }
    return {'height': 0, 'width': 0};
  };

  function isPositioned(element)
  {
    return ["fixed", "absolute", "relative"].indexOf(element.css("position")) > -1 &&
           (Math.abs(oj.DomUtils.getCSSLengthAsInt(element.css("top"))) > 0 ||
            Math.abs(oj.DomUtils.getCSSLengthAsInt(element.css("bottom"))) > 0 ||
            Math.abs(oj.DomUtils.getCSSLengthAsInt(element.css("left"))) > 0 ||
            Math.abs(oj.DomUtils.getCSSLengthAsInt(element.css("right"))) > 0);
  };

  if (!element)
    return false;
  else if ($.isWindow(element[0]) || isPositioned(element))
    return true;

  var alignBox = getRect(element);

  // check that the element is not hidden in overflow
  var isWithinViewPort = true;
  var parent = element.parent();
  while (isWithinViewPort && parent && parent.length > 0 &&  "BODY" !== parent[0].nodeName && parent[0].nodeType === 1 && !isPositioned(parent))
  {
    if (hasOverflow(parent))
    {
      var parentBox = getRect(parent);
      // ignore elements with empty border-boxes
      if (parentBox['height'] > 0 && parentBox['width'] > 0)
      {
        isWithinViewPort = isVisible(alignBox, parentBox);
      }
    }
    parent = parent.parent();
  }

  return isWithinViewPort;
};

/**
 * Mapping of horizontal-vertical (x,y) alignment positon to the corresponding css
 * "transform-origin" attribute.
 *
 * horizontal: right, left, center
 * vertical: top, bottom, middle
 *
 * @private
 * @const
 */
oj.PositionUtils._ANIMATION_TRANSFORM_ORIGIN_RULES =
    {
      'right-top' : 'right top',
      'right-middle' : 'right center',
      'right-bottom' : 'right bottom',
      'left-top' : 'left top',
      'left-middle' : 'left center',
      'left-bottom' : 'left bottom',
      'center-top' : 'center top',
      'center-middle' : 'center center',
      'center-bottom' : 'center bottom'
    };

/**
 * Data attribute key used to store the alignment of the popup relative
 * to the aligning element - position.of
 *
 * @private
 * @const
 * @type {string}
 */
 oj.PositionUtils._ALIGN_MNEMONIC_DATA = "oj-popup-align-mnemonic"

/**
 * Pass the root popup element and the second argument of the jquery ui position utils using
 * callback.  Stashes away the alignment hints returned by the position utility as a data property
 * on the target jquery element.  The alignment hints are used to define the transform-origin
 * of animation effects.
 *
 * @see oj.PositionUtils.addTransformOriginAnimationEffectsOption
 * @param {jQuery} element popup root node
 * @param {Object} props second argument to the jquery ui position "using" callback.
 * @return {void}
 */
 oj.PositionUtils.captureTransformOriginAnimationEffectsOption =  function(element, props)
 {
   var alignMnemonic = [props["horizontal"], props["vertical"]].join("-");
   element.data(oj.PositionUtils._ALIGN_MNEMONIC_DATA, alignMnemonic);
 };

/**
 * Pass "open" or "close" animation effects.  Replaces occurances of
 * "transformOrigin":"#myPositon" with a value that represents the popups alignment.
 *
 * @see oj.PositionUtils.captureTransformOriginAnimationEffectsOption
 * @param {jQuery} element popup root node
 * @param {string|Object|Array} effects animation instructions
 * @returns {string|Object|Array} effects with the transformOrign property resolved
 */
oj.PositionUtils.addTransformOriginAnimationEffectsOption = function (element, effects)
{
  var effectsAsString;
  var isEffectsTypeofString;

  if (!oj.StringUtils.isString(effects))
  {
    isEffectsTypeofString = false;
    effectsAsString = JSON.stringify(effects);
  }
  else
  {
    isEffectsTypeofString = true;
    effectsAsString = effects;
  }

  var exp = /#myPosition/g;
  if (effectsAsString.match(exp))
  {
    var alignMnemonic = /** @type {string} */ (element.data(oj.PositionUtils._ALIGN_MNEMONIC_DATA));
    if (oj.StringUtils.isEmptyOrUndefined(alignMnemonic))
      alignMnemonic = "center-middle";

    var transformOrigin = oj.PositionUtils._ANIMATION_TRANSFORM_ORIGIN_RULES[alignMnemonic];

    effectsAsString = effectsAsString.replace(exp, transformOrigin);

    effects = isEffectsTypeofString ? effectsAsString :
      /** @type {Object} */ (JSON.parse(effectsAsString));
  }

  return effects;
};


/**
 * Splits the jquery UI vertical mnemonic into 3 groups.
 * @private
 * @constant {RegExp}
 */
oj.PositionUtils._JQUI_MNEMONIC_GRP_REGX = /^(\w+)(\+|\-)?(\d+)?/;

/**
 * Verify vertical mnemonic.
 * @private
 * @constant {RegExp}
 */
oj.PositionUtils._VERTICAL_ENUM_TST_REGX = /^top$|^center$|^bottom$/;

/**
 * Verify horizontal mnemonic.
 * @private
 * @constant {RegExp}
 */
oj.PositionUtils._HORIZONTAL_ENUM_TST_REGX = /^start$|^left$|^center$|^end$|^right$/;

/**
 * Verify collision mnemonic.
 * @private
 * @constant {RegExp}
 */
oj.PositionUtils._COLLISION_ENUM_TST_REGX = /^none$|^flip$|^flipfit$|^fit$|^flipcenter$/;

/**
 * @private
 * @param {string} token containing a position alignment rule
 * @param {RegExp} testRegX regular expression to verify the token enumerations
 * @returns {Array} array of two values [alignment, offset].
 */
oj.PositionUtils._parsePositionNmnemonic = function(token, testRegX)
{
  var data = [null, Number.NaN];
  var groups = oj.PositionUtils._JQUI_MNEMONIC_GRP_REGX.exec(token);

  if (groups[1] && testRegX.test(groups[1]))
  {
    data[0] = groups[1];

    // has an offset prefiex by +|-
    if (groups[2])
    {
      var offset = parseInt(groups[3], 10);
      if (!isNaN(offset))
      {
        offset *= (groups[2] === "-" ? -1 : 1);
        data[1] = offset;
      }
    }
  }
  return data;
};

/**
 *
 * @private
 * @param {?} value that might be a json string
 * @returns {?} returns an object if the value is a json string; otherwise,
 *          a null value is returned.
 */
oj.PositionUtils._parseJSON = function(value)
{
  if (oj.StringUtils.isString(value) && /^{/.test(value) && /}$/.test(value))
  {
    try
    {
      return JSON.parse(value);
    }
    catch (e) {}
  }

  return null;
}

/**
 * Converts a source "position.my" or "position.at" into a suitable state held by jet components.
 *
 * @param {string} type "my" or "at"
 * @param {Object} source postion.my or position.at to shape into a Jet position object
 * @param {Object=} offsetSource position.offset
 * @param {Object=} sourceDefault default values
 * @returns {Object} internal position impl
 * @private
 */
oj.PositionUtils._coerceMyAtToJet = function (type, source, offsetSource, sourceDefault)
{
  var obj = oj.PositionUtils._parseJSON(source);
  if (obj)
    source = obj;

  obj = oj.PositionUtils._parseJSON(offsetSource);
  if (obj)
    offsetSource = obj;

  if (!sourceDefault)
    sourceDefault = {};

  var target = $.extend({}, sourceDefault);
  var offsetTarget = {"x": 0, "y": 0};
  if (offsetSource && "x" in offsetSource && "y" in offsetSource)
  {
    offsetTarget["x"] = oj.DomUtils.getCSSLengthAsInt(offsetSource["x"]);
    offsetTarget["y"] = oj.DomUtils.getCSSLengthAsInt(offsetSource["y"]);
  }

  var groups;

  if (oj.StringUtils.isString(source))  // jquery ui
  {
    // split horizontal and vertical tokens
    var tokens = source.split(/\s/);

    // parse horizontal
    if (tokens.length > 0 && !oj.StringUtils.isEmpty(tokens[0]))
    {
      groups = oj.PositionUtils._parsePositionNmnemonic(tokens[0],
        oj.PositionUtils._HORIZONTAL_ENUM_TST_REGX);

      // verify horizontal enum
      if (groups[0])
      {
        target["horizontal"] = groups[0];
        if (!isNaN(groups[1]))
          offsetTarget["x"] = groups[1];
      }
    }

    // parse vertical
    if (tokens.length > 1 && !oj.StringUtils.isEmpty(tokens[1]))
    {
      groups = oj.PositionUtils._parsePositionNmnemonic(tokens[1],
        oj.PositionUtils._VERTICAL_ENUM_TST_REGX);

      // verify vertical enum
      if (groups[0])
      {
        target["vertical"] = groups[0];
        if (!isNaN(groups[1]))
          offsetTarget["y"] = groups[1];
      }
    }
  }
  else if (source)
  {
    // my is is in the jet position format
    if ("horizontal" in source)
    {
      groups = oj.PositionUtils._parsePositionNmnemonic(source["horizontal"],
        oj.PositionUtils._HORIZONTAL_ENUM_TST_REGX);

      if (groups[0])
      {
        target["horizontal"] = groups[0];
        if (!isNaN(groups[1]))
          offsetTarget["x"] = groups[1];
      }
    }

    if ("vertical" in source)
    {
      groups = oj.PositionUtils._parsePositionNmnemonic(source["vertical"],
        oj.PositionUtils._VERTICAL_ENUM_TST_REGX);

      if (groups[0])
      {
        target["vertical"] = groups[0];
        if (!isNaN(groups[1]))
          offsetTarget["y"] = groups[1];
      }
    }
  }

  var targetPosition = {};
  targetPosition[type] = target;
  targetPosition["offset"] = offsetTarget;
  return targetPosition;
};

/**
 * Converts a source "position.collision" into a suitable state held by jet components.
 *
 * @param {string} collisionSource postion.collision to shape into a Jet position object
 * @param {string=} collisionDefault default values
 * @returns {Object} internal position impl
 * @private
 */
oj.PositionUtils._coerceCollisionToJet = function (collisionSource, collisionDefault)
{
  var collisionTarget = collisionDefault;

  if (oj.PositionUtils._COLLISION_ENUM_TST_REGX.test(collisionSource))
  {
    collisionTarget = collisionSource;
  }

  return {"collision": collisionTarget};
};


/**
 * Converts a source "position.of" into a suitable state held by jet components.
 *
 * @param {Object} ofSource position.of
 * @param {Object=} ofDefault default value
 * @return {Object} internal postion object
 * @private
 */
oj.PositionUtils._coerceOfToJet = function (ofSource, ofDefault)
{
  function _escapeId(id)
  {
    var targetId = [];
    var regex = /\w|_|-/;

    for (var i = 0; i < id.length; i++)
    {
      var c = id.substring(i, i + 1);
      if (regex.test(c))
        targetId.push(c);
      else
        targetId.push("\\" + c);
    }
    return targetId.join("");
  }

  var obj = oj.PositionUtils._parseJSON(ofSource);
  if (obj)
    ofSource = obj;

  var targetOf = ofDefault;

  if (oj.StringUtils.isString(ofSource))
  {
    targetOf = ofSource;          // assume a valid selector
  }
  else if ($.isWindow(ofSource))
  {
    targetOf = "window";
  }
  else if (ofSource instanceof Element || ofSource instanceof $)
  {
    ofSource = $(ofSource);
    ofSource.uniqueId();
    var id = ofSource.attr("id");
    targetOf = "#" + _escapeId(id);
  }
  else if (ofSource instanceof Event || ofSource instanceof $.Event)
  {
    if ("pageX" in ofSource || "pageY" in ofSource)
    {
      targetOf = {};
      targetOf["x"] = oj.DomUtils.getCSSLengthAsFloat(ofSource["pageX"]);
      targetOf["y"] = oj.DomUtils.getCSSLengthAsFloat(ofSource["pageY"]);
    }
  }
  else if (ofSource)
  {
    if ("x" in ofSource || "y" in ofSource)
    {
      targetOf = {};
      targetOf["x"] = oj.DomUtils.getCSSLengthAsFloat(ofSource["x"]);
      targetOf["y"] = oj.DomUtils.getCSSLengthAsFloat(ofSource["y"]);
    }
  }

  return {"of": targetOf};
};

/**
 * Converts a source "position" into a suitable state held by jet components.
 *
 * @param {Object} source position
 * @param {Object=} defaults for target position
 * @return {Object} internal postion object
 */
oj.PositionUtils.coerceToJet = function (source, defaults)
{
  if (!source)
    source = {};

  var obj = oj.PositionUtils._parseJSON(source);
  if (obj)
    source = obj;

  if (!defaults)
    defaults = {};

  function _coerceUsingToJet(usingSource, usingDefault)
  {
    var targetUsing = $.isFunction(usingSource) ? usingSource : usingDefault;
    return {"using": targetUsing};
  }

  var myDefault = defaults["my"];
  var atDefault = defaults["at"];
  var collisionDefault = defaults["collision"];
  var ofDefault = defaults["of"];
  var usingDefault = undefined;  // to dangerous to inherit

  var targetMy = oj.PositionUtils._coerceMyAtToJet("my", source["my"], source["offset"], myDefault);
  var targetAt = oj.PositionUtils._coerceMyAtToJet("at", source["at"], null, atDefault);

  // sum the "at" and "my" offsets
  var targetOffset = {"offset":
      {"x": targetMy["offset"]["x"] + targetAt["offset"]["x"],
       "y": targetMy["offset"]["y"] + targetAt["offset"]["y"],
    }};
  delete targetMy["offset"];
  delete targetAt["offset"];

  var target = $.extend({},
                 targetMy,
                 targetAt,
                 targetOffset,
                 oj.PositionUtils._coerceCollisionToJet(source["collision"], collisionDefault),
                 oj.PositionUtils._coerceOfToJet(source["of"], ofDefault),
                 _coerceUsingToJet(source["using"], usingDefault));

  return target;
};


/**
 * Converts the jet position object into the jQuery UI object used by the position utility.
 *
 * @param {Object} source internal position object
 * @return {Object} jQuery UI position Object
 */
oj.PositionUtils.coerceToJqUi = function (source)
{
  function alignToJqUi(align, direction)
  {
    var tokens = [];
    if (source[align][direction])
      tokens.push(source[align][direction]);
    else
      tokens.push("center");

    if ("my" === align && source["offset"])
    {
      var offsetDirection = ("horizontal" === direction ? "x" : "y");
      var offset = source["offset"][offsetDirection];
      if (!isNaN(offset) && offset !== 0)
      {
        tokens.push(offset > 0 ? "+" : "");
        tokens.push(Math.floor(offset).toString());
      }
    }

    return tokens.join("");
  }


  var target = {};

  // convert my and at
  ["my","at"].forEach(function (align)
    {
      if (source[align])
      {
         var tokens = [];
         tokens.push(alignToJqUi(align, "horizontal"));
         tokens.push(" ");
         tokens.push(alignToJqUi(align, "vertical"));
         target[align] = tokens.join("");
      }
    });

    // convert of
    var ofSource = source["of"];
    if (oj.StringUtils.isString(ofSource))
    {
      if ("window" === ofSource)
        target["of"] = window;
      else
        target["of"] = ofSource;
    }
    else if (ofSource && !oj.StringUtils.isString(ofSource) && "x" in ofSource && "y" in ofSource)
    {
      var x = ofSource["x"];
      var y = ofSource["y"];
      var nativeEvent = document.createEvent("MouseEvents");
      nativeEvent.initMouseEvent("click", true, true, window, 1, x, y, x, y,
                                 false, false, false, false, 0, null);
      target["of"] = $.Event(nativeEvent, {"pageX": x, "pageY": y});
    }
    else
    {
      target["of"] = ofSource;
    }

    if (source["collision"])
      target["collision"] = source["collision"];

    // convert using
    if (source["using"])
      target["using"] = source["using"];

  return target;
};

/**
 * Custom jquery UI position collision rule that will first apply the "flip" rule and follow with "center" alignment.
 * @ojtsignore
 */
$.ui.position["flipcenter"] =
{
  /**
   * @param {{top: number, left: number}} position
   * @param {{targetWidth: number,
   *         targetHeight: number,
   *         elemWidth: number,
   *         elemHeight: number,
   *         collisionPosition: {marginLeft: number, marginTop: number},
   *         collisionWidth: number,
   *         collisionHeight: number,
   *         offset: Array.<number>,
   *         my: Array.<string>,
   *         at: Array.<string>,
   *         within: {element: jQuery, isWindow: boolean, isDocument: boolean, offset: {left: number, top: number}, scrollLeft: number, scrollTop: number, width: number, height: number},
   *         elem: jQuery
   *        }} data
   * @returns {undefined}
   */
  "left": function (position, data)
  {
    // stash away the initial position calculated from the at alignment
    var posLeft = position["left"];

    // call on the flip rule
    $.ui.position["flip"]["left"].call(this, position, data);

    // These calcs were taken from the "fit" rule.
    var within = data["within"];
    var withinOffset = within["isWindow"] ? within["scrollLeft"] : within["offset"]["left"];
    var outerWidth = within["width"];
    var collisionPosLeft = position["left"] - data["collisionPosition"]["marginLeft"];
    var overLeft = withinOffset - collisionPosLeft;
    var overRight = collisionPosLeft + data["collisionWidth"] - outerWidth - withinOffset;

    // if popup is not within, center align it
    if (overLeft > 0 || overRight > 0)
    {
      // find the center of the target element
      if ("right" === data["at"][0])
        posLeft -= data["targetWidth"] /2;
      else if ("left" === data["at"][0])
        posLeft += data["targetWidth"] /2;

      var isRTL = oj.DomUtils.getReadingDirection() === "rtl";
      var dirFactor = isRTL ? -1 : 1;

      // factor in half the width of the popup
      posLeft -= dirFactor * (data["elemWidth"] / 2);

      // Force the popup start to be within the viewport.
      // This collision rule is only used by input components internally. The notewindow will auto dismiss when
      // what it is aligned to is hidden in a scroll container.
      position["left"] = Math.max(0, posLeft);
    }
  },
  /**
   * @param {{top: number, left: number}} position
   * @param {{targetWidth: number,
   *         targetHeight: number,
   *         elemWidth: number,
   *         elemHeight: number,
   *         collisionPosition: {marginLeft: number, marginTop: number},
   *         collisionWidth: number,
   *         collisionHeight: number,
   *         offset: Array.<number>,
   *         my: Array.<string>,
   *         at: Array.<string>,
   *         within: {element: jQuery, isWindow: boolean, isDocument: boolean, offset: {left: number, top: number}, scrollLeft: number, scrollTop: number, width: number, height: number},
   *         elem: jQuery
   *        }} data
   * @returns {undefined}
   */
  "top": function (position, data)
  {
    // stash away the initial position calculated from the at alignment
    var posTop = position["top"];

    $.ui.position["flip"]["top"].call(this, position, data);

    // These calcs were taken from the "fit" rule.
    var within = data["within"];
    var withinOffset = within["isWindow"] ? within["scrollTop"] : within["offset"]["top"];
    var outerHeight = data["within"]["height"];
    var collisionPosTop = position["top"] - data["collisionPosition"]["marginTop"];
    var overTop = withinOffset - collisionPosTop;
    var overBottom = collisionPosTop + data["collisionHeight"] - outerHeight - withinOffset;

    if (overTop > 0 || overBottom > 0)
    {
      // find the center of the target element
      if ("top" === data["at"][1])
        posTop += data["targetHeight"] /2;
      else if ("bottom" === data["at"][1])
        posTop -= data["targetHeight"] /2;

      // factor in half the height of the popup
      posTop += data["elemHeight"] / 2;

      // Force the popup top to be within the viewport.
      // This collision rule is only used by input components internally. The notewindow will auto dismiss when
      // what it is aligned to is hidden in a scroll container.
      position["top"] = Math.max(0, posTop);
    }
  }
};

/**
 * Forked the jquery UI "flip" position collision rule in version 1.11.4.
 * The jquery version doesn't consider the best fit in terms of
 * top/bottom. The rule favors top when there is no fit versus choosing the
 * best fit, lesser of two evils. The new JET spin of this rule will pick
 * the better fit versus favoring top when there is no fit.
 *
 * Outside of making the code closure compiler friendly, there is only
 * a single line difference.  It's noted with a bug number.
 * @private
 */
var _origLeftFlipCollisionRule = $.ui.position["flip"]["left"];  //stash away the original left flip rule

/**
 * @ojtsignore
 */
$.ui.position["flip"] = {
  "left": _origLeftFlipCollisionRule.bind(this),
  /**
   * @param {{top: number, left: number}} position
   * @param {{targetWidth: number,
   *         targetHeight: number,
   *         elemWidth: number,
   *         elemHeight: number,
   *         collisionPosition: {marginLeft: number, marginTop: number},
   *         collisionWidth: number,
   *         collisionHeight: number,
   *         offset: Array.<number>,
   *         my: Array.<string>,
   *         at: Array.<string>,
   *         within: {element: jQuery, isWindow: boolean, isDocument: boolean, offset: {left: number, top: number}, scrollLeft: number, scrollTop: number, width: number, height: number},
   *         elem: jQuery
   *        }} data
   * @returns {undefined}
   */
  "top": function(position, data)
  {
    var within = data["within"],
    withinOffset = within["offset"]["top"] + within["scrollTop"],
    outerHeight = within["height"],
    offsetTop = within["isWindow"] ? within["scrollTop"] : within["offset"]["top"],
    collisionPosTop = position["top"] - data["collisionPosition"]["marginTop"],
    overTop = collisionPosTop - offsetTop,
    overBottom = collisionPosTop + data["collisionHeight"] - outerHeight - offsetTop,
    top = data["my"][ 1 ] === "top",
    myOffset = top ? -data["elemHeight"] :
                data["my"][ 1 ] === "bottom" ? data["elemHeight"] : 0,
    atOffset = data["at"][ 1 ] === "top" ? data["targetHeight"] :
                data["at"][ 1 ] === "bottom" ? -data["targetHeight"] : 0,
    offset = -2 * data["offset"][ 1 ],
    newOverBottom,
    newOverTop;
    if (overTop < 0)
    {
      newOverBottom = position["top"] + myOffset + atOffset + offset + data["collisionHeight"] - outerHeight - withinOffset;
      if (newOverBottom < 0 || newOverBottom < Math.abs(overTop))
      {
        //  - only flip up if there is more "over" on top than bottom
        if (overBottom < 0 && overTop > overBottom)
          position["top"] += myOffset + atOffset + offset;
      }
    } else if (overBottom > 0)
    {
      newOverTop = position["top"] - data["collisionPosition"]["marginTop"] + myOffset + atOffset + offset - offsetTop;
      if (newOverTop > 0 || Math.abs(newOverTop) < overBottom)
      {
        position["top"] += myOffset + atOffset + offset;
      }
    }
  }
};
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */


/**
 * @ojcomponent oj.ojProgress
 * @ojstatus preview
 * @since 1.0.0
 * @augments oj.baseComponent
 * @ojshortdesc Displays progress of an operation in a rectangular horizontal meter or in a circular meter.
 * @classdesc
 * <h3 id="progressOverview-section">
 *   JET Progress
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#progressOverview-section"></a>
 * </h3>
 * The JET Progress element allows a user to display progress of an operation in a rectangular horizontal or circular meter.
 * If a developer does not wish to display the exact value, a value of '-1' can be passed in to display an indeterminate value.
 *
 * <pre class="prettyprint"><code>&lt;oj-progress value='{{progressValue}}'>&lt;/oj-progress></code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>If this element is being used to describe the loading process of a particular region of a page, then the <code class="prettyprint">aria-describedby</code>
 *    attribute must point to the id of the oj-progress and <code class="prettyprint">aria-busy = "true"</code> must be added to the region until the loading is complete.</p>
 *
 *
 */

(function() {
   /*
	* <h3 id="markup-section">
    *   HTML Markup and Style Classes
    *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#markup-section"></a>
    * </h3>
    *
    * <p>To create the start label for the progress bar wrap the start label text in a div with the  <code class="prettyprint"> oj-progress-bar-start-label </code> class. </p>
	  *
	  * <p>To create the end label for the progress bar wrap the end label text in a div with the  <code class="prettyprint"> oj-progress-bar-end-label </code> class. </p>
    *
	*/
oj.__registerWidget("oj.ojProgressbar",  $['oj']['baseComponent'], {
  version: "1.0.0",
  defaultElement : "<div>",
  widgetEventPrefix : "oj",
  options: {
    /**
     * The maximum allowed value. The element's max attribute is used if it
     * is provided, otherwise the default value of 100 is used.
     * @ojshortdesc The maximum allowed value.
     * @expose
     * @public
     * @type {number}
     * @instance
     * @memberof oj.ojProgress
     * @default 100
     * @ojmin 0
     * @example <caption>Initialize the Progress with the <code class="prettyprint">max</code> attribute specified</caption>
     * &lt;oj-progress max='220'>&lt;/oj-progress>
     * @example <caption>Get or set the <code class="prettyprint">max</code> property after initialization</caption>
     * //Get progress max
     * var max = myProgress.max;
     *
     * //Set progress max
     * myProgress.max = 200;
     */
    max: 100,
    /**
     * The value of the Progress. The element's value attribute is used if it
     * is provided, otherwise the default value of 0 is used. For indeterminate Progress, set value to -1.
     * @ojshortdesc The value of the Progress.
     * @expose
     * @public
     * @type {number}
     * @instance
     * @memberof oj.ojProgress
     * @default 0
     * @ojmin -1
     * @ojwriteback
     * @example <caption>Initialize the Progress with the <code class="prettyprint">value</code> attribute specified</caption>
     * &lt;oj-progress value='{{progressVal}}'>&lt;/oj-progress>
     * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization</caption>
     * //Get progress value
     * var value = myProgress.value;
     *
     * //Set progress value
     * myProgress.value = 10;
     */
    value: 0,
    /**
     * The shape of the Progress.
     * @ojshortdesc The shape of the Progress.
     * @expose
     * @instance
     * @type {string}
     * @memberof oj.ojProgress
     * @since 3.0
     * @ojvalue {string} "bar" displays progress in a rectangular horizontal meter
     * @ojvalue {string} "circle" displays progress in a circular meter
     * @default "bar"
     * @example <caption>Initialize the Progress with the <code class="prettyprint">type</code> attribute specified</caption>
     * &lt;oj-progress type='circle'>&lt;/oj-progress>
     * @example <caption>Get or set the <code class="prettyprint">type</code> property after initialization</caption>
     * //Get progress type
     * var type = myProgress.type;
     *
     * //Set progress type
     * myProgress.type= 'bar';
     */
    type: "bar",
    /**
     * If disabled is set to true, then the Progress will not change if a new value is passed in.
     * By default this parameter is set to false.
     * @ojshortdesc If disabled is set to true, then the Progress will not change if a new value is passed in.
     * @expose
     * @public
     * @type {boolean}
     * @ignore
     * @instance
     * @memberof oj.ojProgress
     * @default false
     * @example <caption>Initialize the Progress with the <code class="prettyprint">disabled</code> attribute specified</caption>
     * &lt;oj-progress disabled='true'>&lt;/oj-progress>
     * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization</caption>
     * //Get progress disabled
     * var disabled = myProgress.disabled;
     *
     * //Set progress disabled
     * myProgress.disabled= true;
     */
    disabled: false
  },
  // The min value is a constant and 0 is the value set for it.
  min: 0,

   /**
     * Variable used to indicate that the value is indeterminate
     *
     * @override
     * @memberof oj.ojProgress
     * @private
     */
  _indeterminate: false,
  /**
   * Create the Progress
   * @override
   * @memberof oj.ojProgress
   * @return {void}
   * @protected
   */
  _ComponentCreate: function() {
    this._super();

    // Constrain initial value
    this.oldValue = this.options.value = this._constrainedValue();

    this.element.attr({
                    // Only set static values, aria-valuenow and aria-valuemax are
                    // set inside _refreshValue()
                    "role": "progress",
                    "aria-valuemin": this.min
            });
    this.element.addClass("oj-component");
    this._setUpProgressType();
  },
  /**
   * Sets up train based on type
   * @memberof oj.ojProgress
   * @return {void}
   * @private
   */
  _setUpProgressType: function() {
    if (this.options.type === "circle") {
      this.element.addClass("oj-progress-circle")
      this._setupCircleSVG();
    }
    else {
      this.element.addClass( "oj-progress-bar" )
      this.valueDiv = $( "<div class='oj-progress-bar-value'></div>" )
            .appendTo( this.element ); // @HTMLUpdateOK
    }

    this._refreshValue();
  },

  /**
   * <p>Initialize the options</p>
   * @protected
   * @param {Object} originalDefaults
   * @param {Object} constructorOptions
   * @override
   * @memberof oj.ojProgress
   * @return {void}
   */
  _InitOptions : function (originalDefaults, constructorOptions)
  {
    var element = this.element;

    this._super(originalDefaults, constructorOptions);

    if (constructorOptions['max'] === undefined)
    {
      // get from dom. if still undefined, get from originalDefaults.
      var dom = element.attr("max") || undefined;
      if (dom != null)
        this.options['max'] = dom;
    }
  },

  /**
   * Check that value is valid and within the correct bounds. A value of -1 indicates an indeterminate value.
   * @param {number} newValue The new value of the progress being passed in.
   * @override
   * @private
   * @memberof oj.ojProgress
   * @return {number}
   */
  _constrainedValue: function( newValue ) {
    if ( newValue === undefined ) {
      newValue = this.options.value;
    }

    // Indicates that the value is indeterminate.
    this._indeterminate = (newValue == -1 );

    // sanitize value
    if ( typeof newValue !== "number" ) {
      newValue = isNaN(newValue) ? 0 : Number(newValue);
    }

    return this._indeterminate ? -1 :
            Math.min( this.options.max, Math.max( this.min, newValue ) );
  },
  /**
   * Set the value
   * @param {Object} options The options being set
   * @override
   * @private
   * @memberof oj.ojProgress
   * @return {void}
   */
  _setOptions: function( options, flags ) {
    // Ensure "value" option is set after other values (like max)
    var value = options.value;
    delete options.value;

    this._super( options, flags );
    if( !this.options.disabled) {
        this.options.value = this._constrainedValue( value );
        this._refreshValue();
    }
  },

  /**
   * Setup the svg for the determinate progress circle. The determinate progress circle is composed of
   * two circles, one for the value and one for the track. The strokes of the circles are used for the visual rendering.
   * @private
   * @memberof oj.ojProgress
   * @return {void}
   */
  _setupCircleSVG : function ()
  {
    // Clear existing SVG
    if(this.svg)
      this.svg.remove();

    if(!this._indeterminate) {

      var namespace = "http://www.w3.org/2000/svg";
      var svg = document.createElementNS(namespace, "svg")
      var bgCircle = document.createElementNS(namespace, "circle");
      var circle = document.createElementNS(namespace, "circle");
      var radius = Math.min(this.element.outerWidth(), this.element.outerHeight()) * .5;

      svg.setAttribute("width", radius * 2);
      svg.setAttribute("height", radius * 2);
      svg.setAttribute("class", "oj-progress-circle-transform");

      // Track Circle
      bgCircle.setAttribute("r", radius * .9);
      bgCircle.setAttribute("cx", radius);
      bgCircle.setAttribute("cy", radius);
      bgCircle.setAttribute("class", "oj-progress-circle-base");

      // Value Circle
      circle.setAttribute("r", radius * .9);
      circle.setAttribute("cx", radius);
      circle.setAttribute("cy", radius);
      circle.setAttribute("class", "oj-progress-circle-base oj-progress-circle-value");
      var circumference = 2 * radius * .9 * Math.PI;
      var offset = circumference - ((this._percentage() / 100) * circumference);
      circle.setAttribute("style", "stroke-dasharray:" + circumference.toString()
  + ";stroke-dashoffset:" + offset);

      svg.appendChild(bgCircle);
      svg.appendChild(circle);

      this.svg = $(svg);

      this.element[0].appendChild(svg);
    }
  },

  /**
   * Check that the max value is not less than the min
   * @param {string} key The key for the option being set
   * @param {string|number} value The value being set
   * @override
   * @private
   * @memberof oj.ojProgress
   * @return {void}
   */
   _setOption: function( key, value, flags ) {
     if ( key === "max" ) {
       // Don't allow a max less than min
       value = Math.max( this.min, value );
     }
     //If changing the type of progress, want to remove the old progress
     //and rerender a new one
     if ( key === "type" ) {
       if (this.options.type === "circle") {
         this.element.removeClass("oj-progress-circle");
         if(this.svg)
           this.svg.remove();
       }
       else {
         this.element
               .removeClass( "oj-progress-bar" )
         this.valueDiv.remove();
       }

       if(this.overlayDiv) {
         this.overlayDiv.remove();
         this.overlayDiv = null;
       }
     }
     this._super( key, value, flags );
     //If changing the type of progress, want to remove the old progress
     //and rerender a new one
     if ( key === "type" ) {
       this._setUpProgressType();
     }
   },

  /**
   * Calculates the percentage of the progress that has been loaded based on min, max, and value.
   *
   * @override
   * @private
   * @memberof oj.ojProgress
   * @return {number}
   */
    _percentage: function() {
      return Math.min(this._indeterminate ? 100 : 100 * ( this.options.value - this.min ) / ( this.options.max - this.min ), 100);
    },

   /**
    * This function is used to update the value when the value has changed.
    *
    * @override
    * @private
    * @memberof oj.ojProgress
    * @return {void}
    */
    _refreshValue: function() {
      var value = this.options.value;
      var percentage = this._percentage();
      var isCircle = this.options.type === "circle";
      if (isCircle) {
        this._setupCircleSVG();
      }
      else if(!isCircle){
        this.valueDiv
                .toggle( this._indeterminate || value > this.min )
                .width( percentage.toFixed(0) + "%" );

        this.element.toggleClass( "oj-progress-bar-indeterminate", this._indeterminate );
      }

      if ( this._indeterminate) {
        this.element.attr({"aria-valuetext": this.getTranslatedString("ariaIndeterminateProgressText")});
        this.element.removeAttr( "aria-valuenow" );
        this.element.removeAttr( "aria-valuemin" );
        this.element.removeAttr( "aria-valuemax" );
        if ( !this.overlayDiv ) {
          if (isCircle) {
            this.overlayDiv = $( "<div class='oj-progress-circle-overlay'></div>" ).appendTo( this.element ); // @HTMLUpdateOK
            var diameter = Math.min(this.element.outerWidth(), this.element.outerHeight());
            this.overlayDiv.css("width", diameter);
            this.overlayDiv.css("height", diameter);
          }
          else {
            this.overlayDiv = $( "<div class='oj-progress-bar-overlay'></div>" ).appendTo( this.valueDiv ); // @HTMLUpdateOK
          }
          this.overlayDiv.addClass("oj-indeterminate");
        }
      }
      else {
        this.element.attr({"aria-valuemax": this.options.max, "aria-valuenow": value});
        if ( this.overlayDiv ) {
          this.overlayDiv.remove();
          this.overlayDiv = null;
        }
      }
    },

  /**
   * Overide the destory function to remove appropriate class and attributes.
   *
   * @override
   * @private
   * @memberof oj.ojProgress
   * @return {void}
   */
   _destroy: function() {
     if (this.options.type === "circle") {
       this.element.removeClass("oj-progress-circle oj-component");
       if(this.svg)
         this.svg.remove();
     }
     else {
       this.element
             .removeClass( "oj-progress-bar oj-component" );

       this.valueDiv.remove();
       if(this._indeterminate)
         this.element.removeClass( "oj-progress-bar-indeterminate" )
     }
     this.element
                  .removeAttr( "role" )
                  .removeAttr( "aria-valuemin" )
                  .removeAttr( "aria-valuemax" )
                  .removeAttr( "aria-valuenow" )
     if(this.overlayDiv)
       this.overlayDiv.remove();
     this._super();
   }

  // Fragments:

  /**
   * <p>This element has no touch interaction.  </p>
   *
   *
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojProgress
   */

  /**
   * <p>This element has no keyboard interaction.  </p>
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojProgress
   */

/**
 * {@ojinclude "name":"ojStylingDocIntro"}
 *
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
 *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
 *       <th>{@ojinclude "name":"ojStylingDocExampleHeader"}</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-progress-bar-embedded</td>
 *       <td> Optional class that can be set on a oj-progress bar element to style an embedded progress bar within a web application or dialog.</td>
 *       <td>
 *       <pre class="prettyprint">
 *<code>&lt;div class='oj-web-applayout-page'>
 *  &lt;header class='oj-web-applayout-header'>&lt;/header>
 *  &lt;oj-progress class='oj-progress-bar-embedded' value='{{loadingValue}}'>
 *  &lt;/oj-progress>
 *&lt;/div>
 *</code></pre>
 *       </td>
 *     </tr>
 *     <tr>
 *       <td> oj-progress-bar-start-label</td>
 *       <td> Optional class that can be set on a div after the oj-progress element.  This div contains the start text to display underneath the oj-progress bar.</td>
 *       <td>
 *       <pre class="prettyprint">
 *<code>&lt;div id='progress-container'>
 *  &lt;oj-progress value='{{loadingValue}}'>&lt;/oj-progress>
 *  &lt;div class='oj-progress-bar-start-label'>0%&lt;/div>
 *  &lt;div class='oj-progress-bar-end-label'>100%&lt;/div>
 *&lt;/div>
 *</code></pre>
 *       </td>
 *     </tr>
 *    <tr>
 *       <td>oj-progress-bar-end-label</td>
 *       <td> Optional class that can be set on a div after the oj-progress element.  This div contains the end text to display underneath the oj-progress bar. </td>
 *       <td>
 *       <pre class="prettyprint">
 *<code>&lt;div id='progress-container'>
 *  &lt;oj-progress value='{{loadingValue}}'>&lt;/oj-progress>
 *  &lt;div class='oj-progress-bar-start-label'>0%&lt;/div>
 *  &lt;div class='oj-progress-bar-end-label'>100%&lt;/div>
 *&lt;/div>
 *</code></pre>
 *       </td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
 * @memberof oj.ojProgress
 */

});

}( ));

(function() {
var ojProgressMeta = {
  "properties": {
    "max": {
      "type": "number"
    },
    "type": {
      "type": "string",
      "enumValues": ["bar", "circle"]
    },
    "radius": {
      "type": "number"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "ariaIndeterminateProgressText": {
          "type": "string",
          "value": "In Progress"
        }
      }
    },
    "value": {
      "type": "number",
      "writeback": true
    }
  },
  "methods": {},
  "extension": {
    _WIDGET_NAME: "ojProgressbar"
  }
};
oj.CustomElementBridge.registerMetadata('oj-progress', 'baseComponent', ojProgressMeta);
oj.CustomElementBridge.register('oj-progress', {'metadata': oj.CustomElementBridge.getMetadata('oj-progress')});
})();

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/* jslint browser: true,devel:true*/
/**
 * This interface defines the API for oj.ProgressItem.
 * It can be implemented in order to the track progress and status of an arbitrary task
 * (e.g. a file being uploaded)
 *
 * @export
 * @interface ProgressItem
 * @memberof oj
 * @since 4.0.0
 * @ojstatus preview
 */
oj.ProgressItem = function () {
};

/**
 * Attach an event handler
 *
 * @method
 * @name addEventListener
 * @memberof oj.ProgressItem
 * @instance
 * @param {oj.ProgressItem.EventType} eventType eventType
 * @param {EventListener} listener The event listener to add.
 * @returns {void}
 *
 * @export
 */

/**
 * Detach an event handler
 *
 * @method
 * @name removeEventListener
 * @memberof oj.ProgressItem
 * @instance
 * @param {oj.ProgressItem.EventType} eventType eventType
 * @param {EventListener} listener The event listener to remove.
 * @returns {void}
 *
 * @export
 */

/**
 * ProgressItem status
 *
 * @export
 * @enum {string}
 * @memberof oj.ProgressItem
 */
oj.ProgressItem.Status = {
  /**
   * initial state before any progress events
   */
  QUEUED: 'queued',
  /**
   * upload is in progress
   */
  LOADSTARTED: 'loadstarted',
  /**
   * upload aborted
   */
  ABORTED: 'aborted',
  /**
   * upload failed
   */
  ERRORED: 'errored',
  /**
   * upload timeout
   */
  TIMEDOUT: 'timedout',
  /**
   * upload is completed
   */
  LOADED: 'loaded'
};

/**
 * ProgressItem event types
 *
 * @export
 * @enum {string}
 * @memberof oj.ProgressItem
 */
oj.ProgressItem.EventType = {
  /**
   * Triggered when the progress start
   */
  LOADSTART: 'loadstart',
  /**
   * Triggered for upload progress events.
   */
  PROGRESS: 'progress',
  /**
   * Triggered when an upload has been aborted.
   */
  ABORT: 'abort',
  /**
   * Triggered when an upload failed.
   */
  ERROR: 'error',
  /**
   * Triggered when an upload succeeded.
   */
  LOAD: 'load',
  /**
   * Triggered when timeout has passed before upload completed
   */
  TIMEOUT: 'timeout',
  /**
   * Triggered when an upload completed (success or failure).
   */
  LOADEND: 'loadend'
};

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/* global ko */

/**
 * @ignore
 * @ojtsignore
 * @ojcomponent oj.ojProgressItem
 * @since 4.0.0
 * @classdesc Display a progress and status icon when ended. The property data is a <code class="prettyprint">ProgressItem</code>.
 */

/**
 * @member
 * @name data
 * @memberof oj.ojProgressItem
 * @instance
 * @type {null|oj.ProgressItem}
 * @default null
 * @ojshortdesc Data used by the oj.ojProgressList
 */

/**
 * @ignore
 * @ojtsignore
 */
var progressItemMetadata =
  {
    properties: {
      data: {
        type: 'object'
      }
    }
  };

var progressItemView =
  "    <div class='oj-flex oj-sm-justify-content-space-between'>" +
  "      <div class='oj-flex-item oj-flex oj-sm-flex-direction-column'>" +
  "        <span data-bind='text: $properties.data.name' class='oj-progresslist-name'></span>" +
  "        <div data-bind='text: message' class='oj-progresslist-error-message'></div>" +
  '      </div>' +
  "      <div class='oj-flex oj-sm-align-items-center'>" +
  "        <oj-bind-slot name='itemInfo'>" +
  "          <div class='oj-flex-item oj-flex oj-progresslist-info'>" +
  "            <span data-bind='text: $data.getSizeInBKMGT($properties.data.size)'></span>" +
  '          </div>' +
  '        </oj-bind-slot>' +
  "        <div class='oj-flex-item oj-flex'>" +
  "          <oj-progress-status status='{{status}}'" +
  "                              progress='{{progress}}'>" +
  '          </oj-progress-status>' +
  '        </div>' +
  '      </div>' +
  '    </div>';


function progressItemViewModel(context) {
  var self = this;
  var handleUploadProgress;
  var handleUploadFail;
  var handleUploadDone;

  // Triggered for upload progress events. The event payload contains:
  // total: {integer} total number of bytes for the upload
  // loaded: {integer} number of bytes are loaded
  handleUploadProgress = function (event) {
    var item = event.target;
    if (item && event.lengthComputable) {
      self.status(oj.ProgressItem.Status.LOADSTARTED);
      self.progress(parseInt((event.loaded / event.total) * 100, 10));
    }
  };

  // Triggered when the upload fails (error, timeout or aborted).
  handleUploadFail = function (event) {
    var item = event.target;
    if (item) {
      self.message(event.error.message);
      self.status(oj.ProgressItem.Status.ERRORED);
      removeListeners(item);
    }
  };

  // Triggered when the upload is successfully completed.
  handleUploadDone = function (event) {
    var item = event.target;
    if (item) {
      self.status(oj.ProgressItem.Status.LOADED);
      removeListeners(item);
    }
  };

  self.status = ko.observable(oj.ProgressItem.Status.QUEUED);
  self.progress = ko.observable(-1);
  self.message = ko.observable('');

  // return size in B, KB, MB or GB
  var units = ['B', 'KB', 'MB', 'GB', 'TB'];
  self.getSizeInBKMGT = function (_bytes) {
    var bytes = _bytes;
    if (isNaN(bytes)) {
      return null;
    }

    var i;
    for (i = 0; bytes >= 1024 && i < 4; i++) {
      bytes /= 1024;
    }
    return bytes.toFixed(2) + units[i];
  };

  function addListeners(item) {
    item.addEventListener(oj.ProgressItem.EventType.PROGRESS, handleUploadProgress);
    item.addEventListener(oj.ProgressItem.EventType.ERROR, handleUploadFail);
    item.addEventListener(oj.ProgressItem.EventType.TIMEOUT, handleUploadFail);
    item.addEventListener(oj.ProgressItem.EventType.ABORT, handleUploadFail);
    item.addEventListener(oj.ProgressItem.EventType.LOAD, handleUploadDone);
  }

  function removeListeners(item) {
    item.removeEventListener(oj.ProgressItem.EventType.PROGRESS, handleUploadProgress, false);
    item.removeEventListener(oj.ProgressItem.EventType.ERROR, handleUploadFail, false);
    item.removeEventListener(oj.ProgressItem.EventType.TIMEOUT, handleUploadFail, false);
    item.removeEventListener(oj.ProgressItem.EventType.ABORT, handleUploadFail, false);
    item.removeEventListener(oj.ProgressItem.EventType.LOAD, handleUploadDone, false);
  }

  // Add listeners after the listeners have been defined
  addListeners(context.properties.data);
}

oj.Composite.register('oj-progress-item',
  {
    view: progressItemView,
    viewModel: progressItemViewModel,
    metadata: progressItemMetadata
  });


/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojProgressList
 * @ojtsimport ojcomposite
 * @ojtsimport ojlistview
 * @ojtsimport ojprogress
 * @since 4.0.0
 * @ojdisplayname Progress List
 * @ojshortdesc Displays a ListView where data rows are ProgressItems.
 * @ojstatus preview
 * @class oj.ojProgressList
 * @ojsignature {target: "Type", value:"class ojProgressList extends JetElement<ojProgressListSettableProperties>"}
 *
 * @classdesc
 * <h3 id="progressListOverview-section">
 *   JET ProgressList
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#progressListOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>Display a <code class="prettyprint">ListView</code> where the data is a <code class="prettyprint">DataProvider</code> or
 * <code class="prettyprint">TableDataSource</code> and data rows are <code class="prettyprint">ProgressItem</code>s.
 * <p>Please see the <code class="prettyprint">ListView</code> for Touch and Keyboard End User Information</p>
 *
 * <p>See {@link oj.ojListView}</p>
 * <p>See {@link oj.DataProvider}</p>
 * <p>See {@link oj.TableDataSource}</p>
 * <p>See {@link oj.ProgressItem}</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-progress-list data='{{dataProvider}}'>
 * &lt;/oj-progress-list>
 * </code>
 * </pre>
 */

/**
 * Data used by the ProgressList.
 * @member
 * @name data
 * @memberof oj.ojProgressList
 * @instance
 * @ojshortdesc Data used by the ProgressList.
 * @type {oj.DataProvider|oj.TableDataSource|null}
 * @ojsignature {target: "Type", value: "oj.DataProvider<any, any>|null"}
 *
 * @default null
 *
 * @example <caption>Initialize the progress list with the <code class="prettyprint">data</code> attribute specified:</caption>
 * &lt;oj-progress-list data='{{dataProvider}}'>&lt;/oj-progress-list>
 *
 * @example <caption>Get or set the <code class="prettyprint">data</code> property after initialization:</caption>
 * // getter
 * var dataValue = myProgressList.data;
 *
 * // setter
 * myProgressList.data = dataProvider;
 */

/**
 * @ignore
 */
var progressListMetadata =
  {
    properties: {
      data: {
        type: 'Object'
      }
    }
  };

var progressListView =
  "<oj-list-view data-bind=\"attr: {id: $unique + '_list'}\"" +
  "              aria-label='list using array'" +
  "              data='{{$properties.data}}'" +
  "              item.renderer='{{renderer()}}'" +
  "              selection-mode='single'>" +
  '</oj-list-view>' +
  '' +
  "<script type='text/html' data-bind=\"attr: {id: tempId}\">" +
  "  <li class='oj-progresslist-item' tabindex='0'>" +
  "    <oj-progress-item data='{{$data}}'>" +
  '    </oj-progress-item>' +
  '  </li>' +
  '</script>';

function progressListViewModel(context) {
  var self = this;
  self.tempId = context.unique + '_templ';

  self.renderer = function () {
    return oj.KnockoutTemplateUtils.getRenderer(self.tempId, true);
  };
}

oj.Composite.register('oj-progress-list',
  {
    view: progressListView,
    viewModel: progressListViewModel,
    metadata: progressListMetadata
  });

/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 *
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 *
 * @expose
 * @memberof oj.ojProgressList
 * @instance
 *
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 *
 * @expose
 * @memberof oj.ojProgressList
 * @instance
 *
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 *
 * @expose
 * @memberof oj.ojProgressList
 * @instance
 *
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ignore
 * @ojtsignore
 * @ojcomponent oj.ojProgressStatus
 * @since 4.0.0
 * @classdesc Display a progress and status icon when ended. The property data is a <code class="prettyprint">ProgressItem</code>.
 */

/**
 * @member
 * @name data
 * @memberof oj.ojProgressStatus
 * @instance
 * @type {null|oj.ProgressItem}
 * @default null
 * @ojshortdesc Data used by the file upload progress status.
 */

/**
 * @ignore
 * @ojtsignore
 */
var progressStatusMetadata =
  {
    properties: {
      status: {
        type: 'string'
      },
      progress: {
        type: 'number'
      }
    }
  };

var progressStatusView =
  "  <div data-bind='visible: ($properties.status == \"queued\")'" +
  "       class='oj-progressstatus-cell'>" +
  "    <div class='oj-component-icon oj-progressstatus-cancel-icon'" +
  "         role='img' title='cancel'></div>" +
  '  </div>' +
  "  <div data-bind='visible: ($properties.status == \"loadstarted\")'" +
  "       class='oj-progressstatus-cell'>" +
  "    <oj-progress type='circle' value='{{$properties.progress}}'>" +
  '    </oj-progress>' +
  '  </div>' +
  "  <div data-bind='visible: ($properties.status == \"loaded\")'" +
  "       class='oj-progressstatus-cell'>" +
  "    <div class='oj-component-icon oj-progressstatus-done-icon' role='img' title='done'></div>" +
  '  </div>' +
  "  <div data-bind='visible: ($properties.status == \"errored\" || $properties.status == \"timedout\" || $properties.status == \"aborted\")'" +
  "       class='oj-progressstatus-cell'>" +
  "    <div class='oj-component-icon oj-progressstatus-error-icon' role='img' title='error'></div>" +
  '  </div>';

oj.Composite.register('oj-progress-status',
  {
    view: progressStatusView,
    metadata: progressStatusMetadata
  });


/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.PullToRefreshUtils
 * @since 1.2.0
 * @export
 * @hideconstructor
 * @ojstatus preview
 *
 * @classdesc
 * This class provides functions for adding pull to refresh functionality to any container element which hosts refreshable content.
 * By default this class will generate a default panel, which consists of a refresh icon, a primary text, and a secondary text.  The application
 * can use the callback to provide their own panel.  When the release happens, the refresh function will be invoked and application should
 * use this to execute any logic required to refresh the content inside the panel.  For example, fetching new content for the ListView inside
 * the container.  The application must resolve or reject the Promise so that this class can do the necessary cleanup.
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 */
oj.PullToRefreshUtils = {};

/**
 * Setup content container for pull to refresh capability.  This method adds touch listeners on the content container.  The following events are fired by this method:
 * ojpull is fired when user pulls the content container, the event contains the content element as well as the distance pulled in pixels.
 * ojrelease is fired when user releases the content container, the event contains the content element.
 * ojcomplete is fired when the refresh is done and the panel is completely closed.  The event contains the content element.
 *
 * @export
 * @param {Element} element the DOM element that hosts the content to refresh.  When the content is scrollable, the value of this parameter must be the scrollable element.  
 *                  Specifically, when using this with ListView, the ListView element might not necessarily be the scrollable element, but is one of its ancestors instead.
 * @param {function():Promise.<*>} refreshFunc the function to invoke when refresh is triggered.  It must return a Promise.
 * @param {Object=} options optional values that controls aspects of pull to refresh
 * @param {number=} options.threshold the number of pixels to pull until refresh is triggered.  If not specified, a default value will be calculated based on the height of the panel consisting the refresh icon, primary text, and secondary text.
 * @param {string=} options.primaryText the primary text to display.  The primary text is usually used to describe the pull action.  If not specified then no primary text will be displayed.
 * @param {string=} options.secondaryText the secondary text to display.  The secondary text is used to add supplementary text.  If not specified then no secondary text will be displayed.
 * @return {void}
 *
 * @see #tearDownPullToRefresh
 */
oj.PullToRefreshUtils.setupPullToRefresh = function(element, refreshFunc, options)
{
    var outer, content, panel, checkTolerance, threshold, start, height, movex, icon, iconOffset, lastIconClass, title, ratio, iconClass;

    // make sure we are clean
    oj.PullToRefreshUtils.tearDownPullToRefresh(element);

    outer = $(document.createElement("div")).addClass("oj-pulltorefresh-outer");
    oj.PullToRefreshUtils._renderAccessibleLink(element, outer, refreshFunc, options);

    // create the element containing the content
    content = $(document.createElement("div")).addClass("oj-pulltorefresh-panel");    
    outer.append(content); // @HtmlUpdateOK

    // prepend it to the panel
    panel = $(element);
    panel.prepend(outer); //@HTMLUpdateOK; outer is created by the component

    panel
    .on("touchstart.pulltorefresh", function(event) 
    {
        // checks if we are still refreshing
        if ($.data(content[0], "data-pullstart") != null)
        {
             return;
        }

        // make sure we are at the top
        if (panel[0].scrollTop === 0)
        {
            oj.PullToRefreshUtils._handlePull(event, content, options);

            // extract these info up front instead of in move
            icon = content.find(".oj-pulltorefresh-icon");
            if (icon.length > 0)
            {
                iconOffset = icon.parent().outerHeight(true);
            }

            if (options && !isNaN(options['threshold']))
            {
                threshold = parseInt(options['threshold'], 10);
            }

            if (isNaN(threshold))
            {
                threshold = content.outerHeight(true);
            }
            else
            {
                threshold = Math.max(0, Math.min(threshold, content.outerHeight(true)));
            }

            // hide it
            content.css("height", 0);
            content.removeClass("oj-pulltorefresh-transition");
            $.data(content[0], "data-pullstart", event.originalEvent.touches[0].clientY);
            $.data(content[0], "data-pullstart-horiz", event.originalEvent.touches[0].clientX);

            checkTolerance = true;
        }
    })
    .on("touchmove.pulltorefresh", function(event) 
    {
        // first checks whether if the pull had started
        start = $.data(content[0], "data-pullstart");
        if (start == null)
        {
             return;
        }

        // see how far it's been pull
        height = event.originalEvent.touches[0].clientY - parseInt(start, 10);

        // wrong direction
        if (height < 0)
        {
            return;
        }

        // make sure the page doesn't scrolls
        event.preventDefault();

        // checks if we are in the middle of closing the panel
        if ($.data(content[0], "data-closing") != null)
        {
            return;
        }

        // checks if we are still in the loading stage
        if ($.data(content[0], "data-loading") != null)
        {
            content.css("height", Math.max(height, threshold));
            return;
        }

        if (checkTolerance)
        {
            movex = event.originalEvent.touches[0].clientX - parseInt($.data(content[0], "data-pullstart-horiz"), 10);
            // check if the intention is swipe left, if it is don't show the panel yet
            if (Math.abs(movex) > height)
            {
                return;
            }

            // we only want to check this as long as the panel is not visible yet
            checkTolerance = false;
        }    

        content.css("height", height);

        // fire pull event
        oj.PullToRefreshUtils._fireEvent(event, "pull", content, height);

        if (icon != null && icon.length > 0)
        {
            // remove last class
            lastIconClass = $.data(content[0], "data-lasticonclass");
            if (lastIconClass != null)
            {
                icon.removeClass(lastIconClass);
            }

            // find the appropriate class
            ratio = Math.round((height / threshold) * 10) * 10;
            if (ratio >= 100)
            {
                iconClass = "oj-pulltorefresh-icon-full";
                title = oj.Translations.getTranslatedString('oj-pullToRefresh.titleIconFull');
            }
            else
            {
                iconClass = "oj-pulltorefresh-icon-" + ratio + "-percent";
                title = oj.Translations.getTranslatedString('oj-pullToRefresh.titleIcon' + ratio + 'percent');
            }
            icon.addClass(iconClass);
            icon.attr("title", title);

            $.data(content[0], "data-lasticonclass", iconClass);

            // check whether to show/hide primary/secondary text
            oj.PullToRefreshUtils._showHideDefaultText(content, height > iconOffset);            
        }
    })
    .on("touchcancel.pulltorefresh", function(event) 
    {
        oj.PullToRefreshUtils._cleanup(content);
    })
    .on("touchend.pulltorefresh", function(event) 
    {
        // first checks whether if the pull had started
        start = $.data(content[0], "data-pullstart");
        if (start == null)
        {
            return;
        }

        // checks if we are in the middle of closing the panel
        if ($.data(content[0], "data-closing") != null)
        {
            return;
        }

        // checks if we are still in the loading stage
        if ($.data(content[0], "data-loading") != null)
        {
            height = $.data(content[0], "data-panelheight");
            content.css("height", height);
            return;
        }

        // less than threshold, hide panel and do nothing
        if (content.outerHeight() < threshold)
        {
            content.addClass("oj-pulltorefresh-transition")
                   .css("height", 0);

            oj.PullToRefreshUtils._cleanup(content);
        }
        else
        {
            oj.PullToRefreshUtils._handleRelease(event, element, content, refreshFunc);
        }
    });
};

oj.PullToRefreshUtils._handlePull = function(event, content, options)
{
    var primaryText, secondaryText, icon, iconOffset;

    oj.PullToRefreshUtils._fireEvent(event, "pull", content, 0);

    if (content.children().length == 0)
    {
        if (options)
        {
            primaryText = options['primaryText'];
            secondaryText = options['secondaryText'];
        }

        oj.PullToRefreshUtils._createDefaultContent(content, primaryText, secondaryText);
    }

    content.prev().text(oj.Translations.getTranslatedString('oj-pullToRefresh.ariaRefreshingLink'));

    // set it back to auto so we can get the actual height
    content.css("height", "auto");
    $.data(content[0], "data-panelheight", content.outerHeight());
};

oj.PullToRefreshUtils._handleRelease = function(event, element, content, refreshFunc)
{
    var height, icon, lastIconClass, busyContext, busyStateResolve, listener;

    height = $.data(content[0], "data-panelheight");
    content.addClass("oj-pulltorefresh-transition")
           .css("height", height);

    oj.PullToRefreshUtils._fireEvent(event, "release", content, height);

    // mark that we are in loading stage
    $.data(content[0], "data-loading", true);

    icon = content.find(".oj-pulltorefresh-icon");
    if (icon.length > 0)
    {
        // remove last class
        lastIconClass = $.data(content[0], "data-lasticonclass");
        if (lastIconClass != null)
        {
            icon.removeClass(lastIconClass);
        }
        icon.addClass("oj-pulltorefresh-icon-full");
    }

    // register busy state with host as the context, note you can't pull again until release is done
    busyContext = oj.Context.getContext(element).getBusyContext();
    busyStateResolve = busyContext.addBusyState({'description': 'PullToRefresh:handleRelease'});
    $.data($(element)[0], "data-pulltorefresh-busystate", busyStateResolve);

    refreshFunc().then(function(val)
    {
        listener = function()
        {
            oj.PullToRefreshUtils._fireEvent(event, "complete", content, height);

            // cleanup after everything is complete
            oj.PullToRefreshUtils._cleanup(content);

            content.off("transitionend", listener);                  

            // clear the text, otherwise voice over will allow text to be focusable
            content.prev().text("");

            // clear busy state
            oj.PullToRefreshUtils._resolveBusyState(element);
        }

        // change text to complete
        content.prev().text(oj.Translations.getTranslatedString('oj-pullToRefresh.ariaRefreshCompleteLink'));
        // hide the link
        content.prev().prev().css("position", "");

        // mark that we are in the middle of closing the panel, we want this to complete without interruption
        $.data(content[0], "data-closing", true);

        content.on("transitionend", listener);
        content.css("height", 0);
    }, function(e)
    {
        listener = function()
        {
            // cleanup after everything is complete
            oj.PullToRefreshUtils._cleanup(content);

            content.off("transitionend", listener);                  

            // clear the text, otherwise voice over will allow text to be focusable
            content.prev().text("");

            // clear busy state
            oj.PullToRefreshUtils._resolveBusyState(element);
        }

        // hide the link
        content.prev().prev().css("position", "");

        // mark that we are in the middle of closing the panel, we want this to complete without interruption
        $.data(content[0], "data-closing", true);

        content.on("transitionend", listener);
        content.css("height", 0);
    });
};

/**
 * Removes the listener that was added in setupPullToRefresh.  Page authors should call tearDownPullToRefresh when the content container is no longer needed.
 * 
 * @export
 * @param {Element} element the DOM element that hosts the content to refresh
 * @return {void}
 *
 * @see #setupPullToRefresh
 */
oj.PullToRefreshUtils.tearDownPullToRefresh = function(element)
{
    // remove the content panel
    $(element).children(".oj-pulltorefresh-outer").remove();

    // remove all listeners
    $(element).off(".pulltorefresh");

    // free up busy state if it's not done already
    oj.PullToRefreshUtils._resolveBusyState(element);
};

/**
 * Helper method to resolve busy state
 * @param {Element} element the DOM element that hosts the content to refresh
 * @private
 */
oj.PullToRefreshUtils._resolveBusyState = function(element)
{
    var elem, busyStateResolve;

    elem = $(element)[0];

    busyStateResolve = /** @type {Function} */ ($.data(elem, "data-pulltorefresh-busystate"));
    if (busyStateResolve)
    {
        busyStateResolve(null);
        $.removeData(elem, "data-pulltorefresh-busystate");
    }
};

/**
 * Helper method to fire specific event
 * @private
 */
oj.PullToRefreshUtils._fireEvent = function(originalEvent, key, content, distance)
{
    var event = $.Event("oj"+key);
    event['originalEvent'] = originalEvent;

    content.trigger(event, {'content': content, 'distance': distance});
};

/**
 * Creates default content inside refresh panel if one is not provided by the app through callbacks
 * @private
 */
oj.PullToRefreshUtils._createDefaultContent = function(content, primaryText, secondaryText)
{
    var icon, iconContainer, primary, secondary;

    content.addClass("oj-pulltorefresh-content")
           .attr("aria-hidden", "true");

    icon = $(document.createElement("div"));
    icon.addClass("oj-icon oj-pulltorefresh-icon oj-pulltorefresh-icon-initial");
    iconContainer = $(document.createElement("div"));
    iconContainer.addClass("oj-pulltorefresh-icon-container");

    iconContainer.append(icon); // @HtmlUpdateOK
    $.data(content[0], "data-lasticonclass", "oj-pulltorefresh-icon-initial");
    content.append(iconContainer); // @HtmlUpdateOK

    if (primaryText != null)
    {
        primary = $(document.createElement("div"))
                      .addClass("oj-pulltorefresh-primary-text")
                      .text(primaryText);
        content.append(primary); // @HtmlUpdateOK

        // you can't have secondary text without primary text
        if (secondaryText != null)
        { 
            secondary = $(document.createElement("div"))
                            .addClass("oj-pulltorefresh-secondary-text")
                            .text(secondaryText);
            content.append(secondary); // @HtmlUpdateOK
        }
    }

};

/**
 * Show/hide primary/secondary text.  We don't want to show the text until the refresh icon
 * is fully visible otherwise they will overlap.
 * @private
 */
oj.PullToRefreshUtils._showHideDefaultText = function(content, show)
{
    var primaryText, secondaryText;

    primaryText = content.find(".oj-pulltorefresh-primary-text");
    secondaryText = content.find(".oj-pulltorefresh-secondary-text");

    if (show)
    {
        if (primaryText)
        {
            primaryText.show();
        }
        if (secondaryText)
        {
            secondaryText.show();
        }
    }
    else
    {
        if (primaryText)
        {
            primaryText.hide();
        }
        if (secondaryText)
        {
            secondaryText.hide();
        }
    }
};

/**
 * Renders link to provide accessible way to refresh content
 * @private
 */
oj.PullToRefreshUtils._renderAccessibleLink = function(element, panel, refreshFunc, options)
{
    var link, content, status;

    link = $(document.createElement("a"));
    link.text(oj.Translations.getTranslatedString('oj-pullToRefresh.ariaRefreshLink'));
    link.addClass("oj-helper-hidden-accessible")
        .attr("href", "#")
        .focus(function()
        { 
            // the link should be visible when tab to
            // note if removeClass("oj-helper-hidden-accessible") is used here then it would cause
            // weird behavior in Voiceover where the link will be skip immediately after focus
            link.css("position", "static");
        })
        .blur(function(event)
        {
            // Voiceover focus would immediately cause a blur when making the link visible
            if (event.relatedTarget != null)
            {
                link.css("position", "");
            }
        })
        .click(function(event)
        {
            content = panel.children().last();
            oj.PullToRefreshUtils._handlePull(event, content, options);
            oj.PullToRefreshUtils._handleRelease(event, element, content, refreshFunc);

            refreshFunc();
        });

    status = $(document.createElement("div"));
    status.addClass("oj-helper-hidden-accessible")
          .attr("aria-live", "polite");

    panel.append(link); // @HtmlUpdateOK
    panel.append(status); // @HtmlUpdateOK
};

/**
 * Cleanup internals
 * @private
 */
oj.PullToRefreshUtils._cleanup = function(content)
{
    var icon;

    $.removeData(content[0], "data-pullstart");
    $.removeData(content[0], "data-pullstart-horiz");
    $.removeData(content[0], "data-loading");
    $.removeData(content[0], "data-closing");

    // only clear if it's the default panel, give app full control on custom content
    icon = content.find(".oj-pulltorefresh-icon");
    if (icon.length > 0)
    {
        content.empty();
    }
};

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Pull to refresh panel element</td>
 *       <td><kbd>Pan</kbd></td>
 *       <td>Shows the pull to refresh panel as user pan downward.  Refresh action is triggered if the user pan downward over a specified threshold.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.PullToRefreshUtils
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Pull to refresh panel element</td>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Shift focus to refresh action link.</td>
 *     </tr>
 *     <tr>
 *       <td>Link</td>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Invoke refresh action.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.PullToRefreshUtils
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
/*!
 * JET Radio This component is private. @VERSION
 */
/**
 * The _ojRadio component enhances a browser input element into one that is 
 * of type=radio. This is a private component used by ojRadioset.
 * 
 * <h3>Events:</h3>
 * <ul>
 *   <li>clicked/checked?<p>
 *   Triggered if the checkbox is clicked; or if the checkbox was checked programatically
 *   with the checked option.
 *   </li>
 * </ul>
 * 
 * @ojcomponent oj._ojRadioCheckbox
 * @private
 * @augments oj.baseComponent TODO: Should I extend this? 
 * Pros: it gives me oj-disabled/oj-enabled. (easy to add myself)
 * Cons: It gives me tooltip stuff that I don't want. I want that on the div or on the first checkbox only.
 * Pro/Con?: it rewrites required for me if it is on the dom node, but then it makes it required??? Should I rewrite required or don't care?
 */
oj.__registerWidget("oj._ojRadioCheckbox", $['oj']['baseComponent'],
{
  version : "1.0.0",  
  defaultElement : "<input>", 
  widgetEventPrefix : "oj", 
  options : 
  {	 
    /** 
     * First we look for the disabled option to be explicitly set. If not, then
     * we look if disabled is on the dom. If null, disabled defaults to false.
     * @expose 
     * @type {?boolean|undefined}
     * @default <code class="prettyprint">false</code>
     * @public
     * @instance
     * @memberof oj._ojRadioCheckbox
     */
    disabled: null,
    /** 
     * First we look for the checked option to be explicitly set. If not, then
     * we look if checked is on the dom. If null, checked defaults to false.
     * @expose 
     * @type {?boolean}
     * @public
     * @instance
     * @memberof oj._ojRadioCheckbox */
    checked : null,
    /** 
     * First we look for the disabled option to be explicitly set. If not, then
     * we look if disabled is on the dom. If null, disabled defaults to false.
     * @expose 
     * @type {?boolean}
     * @default <code class="prettyprint">false</code>
     * @public
     * @instance
     * @memberof oj._ojRadioCheckbox
     */
     type: null	
  },
  /**** start Public APIs ****/
  /**
   * 
   * <p>This method does not accept any arguments.
   * 
   * @public
   * @expose
   * @memberof oj.RadioCheckbox 
   * @return {jQuery} the label(s) for the checkbox/radio input
  */
  label : function ()
  {
    if (this.$label === undefined)
    {
      this.$label = this._getLabelsForElement();
    }
    return this.$label;
  },    
  /**
   * @expose 
   * @memberof oj._ojRadioCheckbox
   * @instance
   * @override
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   * $( ".selector" )._ojRadioCheckbox( "refresh" );
   */
  refresh: function() 
  {
    this._super();
    this._setup();
  },  
  /**
   * @expose 
   * @memberof oj._ojRadioCheckbox
   * @instance
   * @override
   * @example <caption>Invoke the <code class="prettyprint">refreshDisabled</code> method:</caption>
   * $( ".selector" )._ojRadioCheckbox( "refreshDisabled" );
   */
  refreshDisabled: function() 
  {
    // this looks at the effectivelyDisabled flag and updates the disabled attributes on the dom.
    this._renderDisabled();
  },  
  /**
   * Set the oj-selected class to the element's checked property.
   * We keep the oj-selected class in sync with the input's checked attribute, not necessarily the 
   * component's checked property. The component's checked property is set after it is validated.
   * if validation doesn't pass, the input may still be checked, but the this.options.checked isn't. 
   * Think of it like an inputText. You can clear it out when it is required, blur,
   * and the display value is an empty field, but the value is the value that was there before.
   * @expose 
   * @memberof oj._ojRadioCheckbox
   * @instance
   * @override
   * @example <caption>Invoke the <code class="prettyprint">setSelectedClass</code> method:</caption>
   * $( ".selector" )._ojRadioCheckbox( "setSelectedClass", true );
   */
  setSelectedClass: function(checked) 
  {
    this.element.toggleClass("oj-selected", checked);
    this.$label.toggleClass("oj-selected", checked);
    this.$choiceItem.toggleClass("oj-selected", checked);
  },
  /**
   * Returns a jQuery object containing the element visually representing the checkbox. 
   * 
   * <p>This method does not accept any arguments.
   * 
   * @expose
   * @memberof oj._ojRadioCheckbox
   * @instance
   * @return {jQuery} the checkbox or radio
  */
  widget : function ()
  {
    return this.uiRadioCheckbox;
  },
          
   /**** end Public APIs ****/         
          
  /**** start internal widget functions ****/   
 /**
  * Called at component create time primarily to initialize options, often using DOM values. This 
  * method is called before _ComponentCreate is called, so components that override this method 
  * should be aware that the component has not been rendered yet. The element DOM is available and 
  * can be relied on to retrieve any default values. <p> 
  * @param {!Object} originalDefaults - original default options defined on the widget and its ancestors
  * @param {?Object} constructorOptions - options passed into the widget constructor
  * 
  * @memberof oj._ojRadioCheckbox
  * @instance
  * @protected
  */
  _InitOptions: function(originalDefaults, constructorOptions)
  {
    var checkedFromDom;
    var disabledFromDom;
    
    this._super(originalDefaults, constructorOptions);
    
    // CHECKED:
    // if options.checked is not set, get it from the element
    // if options.checked is set to a valid value (boolean), set it on the 
    // element to keep the two in sync (we do this in _CreateComponent->_setup)
    // 
    // use DOM value if not in constructorOptions
    if (!('checked' in constructorOptions))
    {
      this.initCheckedFromDom = true;
      checkedFromDom = !!this.element.prop("checked");
      // writeback not needed since "not in constructorOptions" means "not bound"
      this.option('checked', checkedFromDom, {'_context': {internalSet: true}});
    }
    if (typeof this.options.checked !== "boolean")
    {
      throw new Error("checked option must be a boolean");
    }
    // DISABLED:
    // if options.disabled is not set, get it from the element
    // if options.disabled is set to a valid value (boolean), set it on the 
    // element to keep the two in sync (we do this in _CreateComponent->_setup)
    // use DOM value if not in constructorOptions
    if (!('disabled' in constructorOptions))
    {
      // !! ensures it is a boolean
      disabledFromDom =  !!this.element.prop("disabled");
      // writeback not needed since "not in constructorOptions" means "not bound"
      this.option('disabled', disabledFromDom, {'_context': {internalSet: true}});
    }
    if (typeof this.options.disabled !== "boolean")
    {
      throw new Error("disabled option must be a boolean");
    } 
	
    // TYPE:
    // Gets the type which will be either radio or checkbox.
    // writeback not needed since "not in constructorOptions" means "not bound"
    if (!('type' in constructorOptions))
    {
      this.option('type', this.element.prop( "type" ), {'_context': {internalSet: true}});
    }
  },
  /**
   * After _ComponentCreate and _AfterCreate, 
   * the widget should be 100% set up. this._super should be called first.
   * @override
   * @protected
   * @instance
   * @memberof oj._ojRadioCheckbox
   */
  _ComponentCreate: function()
  {
    this._super();

    var type = this.options.type;
    if (type == "checkbox") {
        this.uiRadioCheckbox = this.element.addClass("oj-checkbox oj-component");
        this.$label = this._getLabelsForElement();
        this.$label.addClass("oj-checkbox-label");
    }
    else if (type == "radio") {
        this.uiRadioCheckbox = this.element.addClass("oj-radio oj-component");
        this.$label = this._getLabelsForElement();
        this.$label.addClass("oj-radio-label");
    }
    
    // this will not be null since we create a choice item if there isn't one
    this.$choiceItem = this._getChoiceItem();
  
    var iElem = document.createElement("span"); 
    iElem.setAttribute("class", "oj-radiocheckbox-icon");
    this.element.wrapAll(iElem); //@HTMLUpdateOk iElem constructed locally
      
    var self = this;
    this._focusable( this.element );
    
    this._AddHoverable(this.$choiceItem);
    this._AddActiveable(this.$choiceItem);

    // the input gets focus on keyboard tabbing. It bubbles up, so in case the 
    // input element is hidden (e.g., in the native themes the input is hidden and an image is 
    // shown instead), we need to set the focus selectors on the oj-choice-item so
    // we can style the checked image.
    this._focusable({
      'element': this.$choiceItem,
      'applyHighlight': true
    });


    this._AddHoverable(this.$label);
    this._AddActiveable(this.$label);
    
    // loop through each label to add dom and styles
    $.each(self.$label, function ()
    {   
      // wrap children in span
      $(this.childNodes).wrapAll("<span class='oj-radiocheckbox-label-text'></span>"); //@HTMLUpdateOK

      
    });   
    this._setup();
  },
  /**
   * <p>Save only the 'class' attribute since that is what
   * we manipulate. We don't have to save all the attributes.
   * </p>
   *
   * @param {Object} element - jQuery selection to save attributes for
   * @protected
   * @memberof oj._ojRadioCheckbox
   * @instance
   * @override
   */
  _SaveAttributes : function (element)
  {
    this._savedClasses = element.attr("class");
  },
  /**
   * <p>Restore what was saved in _SaveAttributes
   * </p>
   *
   * @protected
   * @memberof oj._ojRadioCheckbox
   * @instance
   * @override
   */
  _RestoreAttributes : function ()
  {
    // restore the saved "class" attribute. Setting class attr to undefined is a no/op, so
    // if this._savedClasses is undefined we explicitly remove the 'class' attribute.
    if (this._savedClasses)
      this.element.attr("class", this._savedClasses);   
    else
      this.element.removeAttr("class");
  },
  /* _setup called during creation */
  _setup : function() 
  {
    this._renderDisabled();

    if (!this.initCheckedFromDom) 
    {
      // if we got it from the dom in _InitOptions, there is no need to reset it on the dom in _setup
      this._setCheckedOnDom(this.options.checked);     
    }
    if (this.options.checked)
      this.setSelectedClass(this.options.checked);
  },
  _setCheckedOnDom : function(checked)
  {
    // !! to turn checked into a boolean
    checked = !!checked;
  
    this.element.prop("checked", checked);
  },
  _renderDisabled : function()
  {
    // Determines whether this component is effectively disabled, 
    // i.e. it has its 'disabled' option set to true or 
    // it has been disabled by its ancestor component
    var effectivelyDisabled = this._IsEffectivelyDisabled();
    if (effectivelyDisabled)
    {
      // set the dom to show DISABLED, but do NOT change the disabled option!
      // 
      // when a dom element supports disabled, use that, and not aria-disabled.
      // having both is an error.
      this.element.prop("disabled", true).removeAttr( "aria-disabled")
       .removeClass("oj-enabled").addClass("oj-disabled");

      this.$label.removeClass("oj-enabled")
       .addClass("oj-disabled");

      this.$choiceItem.removeClass("oj-enabled").addClass("oj-disabled");
    }
    else // option not set to disabled. nor is parent. On refresh this is ok, since we get it from the option.
    {
      // when a dom element supports disabled, use that, and not aria-disabled.
      // having both is an error.
      this.element.prop("disabled", false)
      .removeAttr( "aria-disabled").removeClass("oj-disabled")
      .addClass("oj-enabled");
      this.$label.addClass("oj-enabled")
      .removeClass("oj-disabled"); 
      this.$choiceItem.addClass("oj-enabled").removeClass("oj-disabled");
    }
  },
  /**
   * @override
   * @private
   */
  _setOption : function (key, value)
  {

    this._superApply(arguments);

    if (key === "disabled")
    {
      value = !!value;
      this._renderDisabled(value);
    }
 
    if (key === "checked")
    {
      this._setCheckedOnDom(value);
      this.setSelectedClass(value);

    }
  }, 
  /**
   * Returns the list of labels for the element. Most likely this will be 
   * one label, not multiple labels.
   * We do not guarantee that the returned list is live
   * We do not guarantee that the returned list is in document order
   * We check a jquery selector query on <label>s with a 'for' id equal to our id starting at the document level
   * and also as a sibling of the input (needed if documentFragment instead of document as the
   * table/datagrid use).
   * We do not support a label wrapping an input, so if we find that, we log an error.
   * NOTE: The .labels DOM property does not work on most browsers, so we don't use it.
   * e.g,
   * <pre>
   * <input id="opt3" type="checkbox" name="rb" value="opt3">
   * <label for="opt3">Checkbox Option 3</label>
   * </pre>
   * @private
   */
  _getLabelsForElement: function() 
  {
    // this looks to see if the label is wrapping the input, which we do not support
    var $labelClosestParent = this.element.closest("label");
    
    var id = this.element.prop("id");
    var labelForQuery = "label[for='" + id + "']";
    var $labelForElems = $(labelForQuery);
    var $labelSibling;
    
    if ($labelClosestParent.length !== 0)
    {
      oj.Logger.error("Found a label that is an input's ancestor." +
      " This is not supported in the ojCheckboxset or ojRadioset component and the component will\n\
      not work correctly. " +
      "Use a label as a sibling to the input and use the label 'for' attribute to tie the two together.");
    }
    
    // make sure that the $labelClosestParent isn't also the one we found with the 'for' which would
    // be weird, but still. If it is, remove it from the jQuery element list so we don't count it as found.
    // e.g., <label for="input"><input id="input"></label> // not supported!
    //.not() method creates a new set and leaves the original set unchanged
    $labelForElems = $labelForElems.not($labelClosestParent); 
    
    if ($labelForElems.length === 0)
    {
      // table and datagrid create their rows using documentFragment. With documentFragment
      // the labelFor query will return []. In that case, look for the label as a sibling of 
      // the element
      $labelSibling = this.element.siblings(labelForQuery);
      //.add() method creates a new set and leaves the original set unchanged
      $labelForElems = $labelForElems.add($labelSibling); 
    } 

    // combine these query results to return the label
    // with the for attribute pointing to the checkbox's id.
    return $labelForElems;
  },
  /**
   * Call this before you wrap the input in a span class='oj-radiocheckbox-icon' because we
   * are looking for the span with oj-choice-item as the parent of the input.
   * If it isn't there, this function will add it. So it never returns null.
   * @private
   * @returns {Object}
   */
  _getChoiceItem: function()
  {
    var choiceItem = null;
    var elementParent;
    var labelSelector;
    var ojChoiceItemSpanString;    
    var siblingLabel;

    elementParent = this.element.parent();
    // oj-choice-row and oj-choice-row-inline have been deprecated on December 7, 2016 in
    // version 3.0.0. Use oj-choice-item instead.
    if (elementParent && (elementParent.hasClass("oj-choice-item") || 
    elementParent.hasClass("oj-choice-row") || elementParent.hasClass("oj-choice-row-inline")))
    {
      choiceItem = elementParent;
    }
    else
    {
      oj.Logger.warn("The radioset/checkboxset's input and label dom should be wrapped in a dom " +
        "node with class 'oj-choice-item'. JET is adding this missing dom to make the component work correctly.");

      // Since we can't find the oj-choice-item, create one. 
      // It needs to wrap the input and its label (if any)
      ojChoiceItemSpanString = "<span class='oj-choice-item oj-choice-item-added'></span>";

      // the most common case is an <input id='foo'><label for='foo'> pair, so look for that first
      labelSelector = "label[for='" + this.element.attr("id") + "']";
      siblingLabel = this.element.siblings().filter( labelSelector );

      if (siblingLabel.length !== 0)
      {
       this.element.add(siblingLabel).wrapAll(ojChoiceItemSpanString); //@HTMLUpdateOk adding empty span for styling
       choiceItem = this.element.parent();
      }
      else
      {
        this.element.wrapAll(ojChoiceItemSpanString); //@HTMLUpdateOk adding empty span for styling
        choiceItem = this.element.parent();
      }   
    }
 
    return choiceItem;
  },
  /**
   * Return the subcomponent node represented by the documented locator attribute values.
   * Test authors should target spinner sub elements using the following names:
   * <ul>
   * <li><b>oj-radiocheckbox-input</b>: the radio/checkbox's input</li>
   * <li><b>oj-radiocheckbox-label</b>: the radio/checkbox's label</li>
   * </ul>
   * 
   * @expose
   * @override
   * @memberof oj._ojRadioCheckbox
   * @instance
   * @param {Object} locator An Object containing at minimum a subId property 
   * whose value is a string, documented by the component, that allows the component to 
   * look up the subcomponent associated with that string.  It contains:
   * <ul>
   * <li>
   * component: optional - in the future there may be more than one component 
   *   contained within a page element
   * </li>
   * <li>
   * subId: the string, documented by the component, that the component expects 
   * in getNodeBySubId to locate a particular subcomponent 
   * </li>
   * </ul>  
   * @returns {Element|null} the subcomponent located by the subId string 
   * passed in locator, if found.
   */
  getNodeBySubId: function(locator)
  {
    var node = this._super(locator);    
    if (!node)
    {
      var subId = locator['subId'];
      if (subId === "oj-radiocheckbox-input") {
        node = this.element[0];
      }
      if (subId === "oj-radiocheckbox-label") {
        // this.label() returns a jquery object. we want to return a dom element
        node = this.label()[0];
      }
    }
    // Non-null locators have to be handled by the component subclasses
    return node || null;
  },  
  
  /**
   * @ignore
   * @protected
   * @override
   */
  _destroy : function ()
  {  
    var ret = this._super();

    this._RemoveHoverable(this.$choiceItem);
    this._RemoveActiveable(this.$choiceItem);

    this._RemoveHoverable(this.$label);
    this._RemoveActiveable(this.$label);
          
    var type = this.options.type;
    
    // this.$label is the label for the checkbox/radio, NOT the label for the radioset/checkboxset.
    // We don't save and restore these attributes in base class, so we need to clean up ourselves
    if (type == "checkbox") 
    {
      this.$label.removeClass("oj-enabled oj-disabled oj-selected oj-checkbox-label");
    }
    else if (type == "radio") 
    {
      this.$label.removeClass("oj-enabled oj-disabled oj-selected oj-radio-label");
    }

    this.$choiceItem.removeClass("oj-enabled oj-disabled oj-selected");
    
    var self = this;

    // loop through each label to remove things we added
    $.each(self.$label, function ()
    {
      var text = this.getElementsByClassName("oj-radiocheckbox-label-text");
      if (text !== undefined)
        $(text[0].childNodes[0]).unwrap();
    });
    
    // remove the oj-radiocheckbox-icon span around the element.
    this.element.unwrap();
    
    // remove the oj-choice-item span only if I added it.
    // I marked it with 'oj-choice-item-added' style.
    if (this.$choiceItem.hasClass("oj-choice-item-added"))
    {
        this.element.unwrap();
    }
    
    this.$choiceItem = null;
    this.$label = null;

    return ret;
  }
  
  /**** end internal widget functions ****/ 
 
});
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */
(function() {

/*!
 * JET Radioset @VERSION
 */
/**
 * @ojcomponent oj.ojRadioset
 * @augments oj.editableValue
 * @ojsignature [{
   *                target: "Type",
   *                value: "class ojRadioset extends editableValue<any, ojRadiosetSettableProperties>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojRadiosetSettableProperties extends editableValueSettableProperties<any>",
   *                for: "SettableProperties"
   *               }
   *              ]
 * @since 0.6
 * @ojshortdesc A grouping of related radio buttons.
 * @ojrole radio
 * @ojrole radiogroup
 * @ojrole option
 * @ojstatus preview
 * @classdesc
 * <h3 id="radiosetOverview-section">
 *   JET Radioset
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#radiosetOverview-section"></a>
 * </h3> 
 * <p>
 * The JET oj-radioset component manages a set of 
 * <code class="prettyprint">oj-option</code> child elements and creates the necessary dom elements for
 * the actual radio buttons. 
 * </p>
 * <p>To use an oj-radioset, add one or more oj-option child elements for each radio button desired.
 * Note, if you add or remove an oj-option after the oj-radioset is rendered, you should call
 * refresh() on the oj-radioset.
 * </p>
 * <p>
 *  You can enable and disable an oj-radioset, 
 *  which will enable and disable all contained radios. 
 * </p>
 * <p>
 *  oj-radioset does not have a readonly attribute since HTML does not support
 *  readonly on radios and checkboxes.
 * </p>
 * <p>
 * The oj-label element is not required. If you don't use an oj-label element,
 * then you need to set <code class="prettyprint">aria-label</code> on the input for accessibility.
 * </p>
 * 
 * {@ojinclude "name":"validationAndMessagingDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * 
 * <p>JET oj-radioset takes care of setting 
 * <code class="prettyprint">role="radiogroup"</code> on the oj-radioset element.  
 * 
 * <p>As shown in the online demos, the application is responsible for applying 
 * <code class="prettyprint">labelled-by</code>
 * to point to the radioset's <code class="prettyprint">oj-label</code> element for the oj-radioset.
 * <p>Disabled content: JET supports an accessible luminosity contrast ratio, 
 * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>, 
 * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which 
 * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user 
 * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio 
 * required of enabled content, it cannot be used to convey meaningful information.<p>
 *  
 * <h3 id="label-section">
 *   Label and Radioset
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * For accessibility, you should associate an <code class="prettyprint">oj-label</code> element with the 
 * radioset by putting an <code class="prettyprint">id</code> on the oj-label, and then setting the 
 * <code class="prettyprint">labelled-by</code> attribute on the oj-radioset to be the oj-label's 
 * <code class="prettyprint">id</code>.
 * </p>
 * 
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 * {@ojinclude "name":"stylingDoc"}
 * 
 * @example <caption>Initialize the radioset with no options specified:</caption>
 * &lt;oj-radioset id="colorRadio" value="{{currentColor}}">
 *   &lt;oj-option value="blue">Blue&lt;/oj-option>
 *   &lt;oj-option value="green">Green&lt;/oj-option>
 * &lt;/oj-radioset>
 * 
 * @example <caption>Initialize component and an associated oj-label component</caption>
 * &lt;label id="grouplabel">Greetings&lt;/label>
 * &lt;oj-radioset id="radioset" aria-labelledby="grouplabel" value="{{currentGreeting}}">
 *   &lt;input id="helloid" value="hello" type="radio" name="greetings"/&gt;
 *   &lt;label for="helloid"/&gt;Hello&lt;/label>
 *   &lt;input id="bonjourid" value="bonjour" type="radio" name="greetings"/&gt;
 *   &lt;label for="bonjourid"/&gt;Bonjour&lt;/label>
 *   &lt;input id="ciaoid" value="ciao" type="radio" name="greetings"/&gt;
 *   &lt;label for="ciaoid"/&gt;Ciao&lt;/label>
 * &lt;oj-radioset>
 * <br/>
 * // set the value to "ciao". (The 'ciao' radio will be selected)
 * myComp.value = "ciao";
 * 
 * @example <caption>Initialize a radioset via the JET component binding:</caption>
 * &lt;oj-label id="grouplabel">Time&lt;/oj-label>
 * &lt;oj-radioset id="radioset" value="{{radiosetValue}}" 
 *                 labelled-by="grouplabel">
 *   &lt;oj-option id="morningid" value="morning">Morning&lt;/oj-option>
 *   &lt;oj-option id="nightid" value="night">Night&lt;/oj-option>
 * &lt;/oj-radioset>
 * <br/>
 */
oj.__registerWidget("oj.ojRadioset", $['oj']['editableValue'],
{
  version : "1.0.0",  
  defaultElement : "<div>", 
  widgetEventPrefix : "oj", 
  options : 
  {
    /** 
     * <p>
     * Disabled <code class="prettyprint">true</code> disables the component and disables all the 
     * inputs/labels. 
     * Disabled <code class="prettyprint">false</code> enables the component, and leaves the 
     * inputs' <code class="prettyprint">disabled</code> property as it is in the dom.
     * <p>
     *  
     * @example <caption>Initialize component with <code class="prettyprint">disabled</code> attribute:</caption>
     * &lt;oj-radioset disabled>
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-radioset>
     *
     * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
     * // getter
     * var disabled = myComp.disabled;
     *
     * // setter
     * myComp.disabled = false;
     * 
     * @expose 
     * @type {boolean|null}
     * @default false
     * @public
     * @instance
     * @memberof oj.ojRadioset
     */
    disabled: false,
     /**
     * It is used to establish a relationship between this component and another element.
     * A common use is to tie the oj-label and the oj-radioset together.
     * The oj-label custom element has an id, and you use the labelled-by attribute
     * to tie the two components together.
     *
     * @example <caption>Initialize component with <code class="prettyprint">labelled-by</code> attribute:</caption>
     * &lt;oj-label id="labelId">Name:&lt;/oj-label>
     * &lt;oj-radioset labelled-by="labelId">
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-radioset>
     *
     * @example <caption>Get or set the <code class="prettyprint">labelledBy</code> property after initialization:</caption>
     * // getter
     * var disabled = myComp.labelledBy;
     *
     * // setter
     * myComp.labelledBy = "labelId";
     *
     * @expose
     * @type {string|null}
     * @public
     * @instance
     * @memberof oj.ojRadioset
     */
    labelledBy: null,   
    /** 
     * Whether the component is required or optional. When required is set to true, an implicit 
     * required validator is created using the validator factory - 
     * <code class="prettyprint">oj.Validation.validatorFactory(oj.ValidatorFactory.VALIDATOR_TYPE_REQUIRED).createValidator()</code>.
     * 
     * Translations specified using the <code class="prettyprint">translations.required</code> option 
     * and the label associated with the component, are passed through to the options parameter of the 
     * createValidator method. 
     * 
     * <p>
     * When <code class="prettyprint">required</code> option changes due to programmatic intervention, 
     * the component may clears message and run validation, based on the current state it's in. </br>
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>if component is valid when required is set to true, then it runs deferred validation on 
     * the property value. This is to ensure errors are not flagged unnecessarily.
     * </li>
     * <li>if component is invalid and has deferred messages when required is set to false, then 
     * component messages are cleared but no deferred validation is run.
     * </li>
     * <li>if component is invalid and currently showing invalid messages when required is set, then 
     * component messages are cleared and normal validation is run using the current display value. 
     * <ul>
     *   <li>if there are validation errors, then <code class="prettyprint">value</code> 
     *   property is not updated and the error is shown. 
     *   </li>
     *   <li>if no errors result from the validation, the <code class="prettyprint">value</code> 
     *   property is updated; page author can listen to the <code class="prettyprint">optionChange</code> 
     *   event to clear custom errors.</li>
     * </ul>
     * </li>
     * </ul>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>Only messages created by the component are cleared.</li>
     * <li><code class="prettyprint">messagesCustom</code> option is not cleared.</li>
     * </ul>
     * 
     * </p>
     * 
     * @ojvalue {boolean} false - implies a value is not required to be provided by the user. 
     * This is the default.
     * @ojvalue {boolean} true - implies a value is required to be provided by user and the 
     * input's label will render a required icon. Additionally a required validator - 
     * {@link oj.RequiredValidator} - is implicitly used if no explicit required validator is set. 
     * An explicit required validator can be set by page authors using the validators option. 
     * 
     * When the option is not set, the element's required property is used as its initial 
     * value if it exists.
     * 
     * @example <caption>Initialize the component with the <code class="prettyprint">required</code> attribute:</caption>
     * &lt;oj-radioset required>
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-radioset>
     * 
     * @example <caption>Customize messages and hints used by implicit required validator when 
     * <code class="prettyprint">required</code> option is set:</caption> 
     * &lt;oj-radioset required value="{{colors}}"
     *                    translations='{"required":
     *                                    {"hint": "custom: check at least one value",
     *                                     "messageSummary": "custom: \'{label}\' is Required",
     *                                     "messageDetail", "custom: please check at least one value for \'{label}\'"}}'>
     *   &lt;oj-option value="blue">Blue&lt;/oj-option>
     * &lt;/oj-radioset>
     *
     * @example <caption>Get or set the <code class="prettyprint">required</code> property after initialization:</caption>
     * // getter
     * var required = myComp.required;
     *
     * // setter
     * myComp.required = false;
     *
     * @expose 
     * @access public
     * @instance
     * @memberof oj.ojRadioset
     * @type {boolean}
     * @default false
     * @since 0.7
     * @see #translations
     */
    required: false,    
    /** 
     * The value of the component. 
     * 
     * <p>
     * When <code class="prettyprint">value</code> option changes due to programmatic 
     * intervention, the component always clears all messages and runs deferred validation, and 
     * always refreshes UI display value.</br>
     * 
     * <h4>Clearing Messages</h4>
     * <ul>
     * <li>All messages are cleared. This includes  
     * <code class="prettyprint">messagesCustom</code> property.</li>
     * </ul>
     * 
     *  
     * <h4>Running Validation</h4>
     * <ul>
     * <li>component always runs deferred validation</li>
     * </ul>
     * </p>
     * 
     * @example <caption>Initialize component with <code class="prettyprint">value</code> attribute:</caption>
     * &lt;oj-radioset value="coffee">
     *   &lt;oj-option value="coffee">Coffee&lt;/oj-option>
     *   &lt;oj-option value="tea">Tea&lt;/oj-option>
     * &lt;/oj-radioset>
     *
     * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
     * // getter
     * var val = myComp.value;
     *
     * // setter
     * myComp.value = "tea";
     *
     * 
     * @expose 
     * @access public
     * @instance
     * @ojwriteback
     * @default null
     * @memberof oj.ojRadioset
     * @type {*}
     */
    value: undefined
  },
  /**** start Public APIs ****/
      
   /**
   * Refreshes the radioset
   * <p>A <code class="prettyprint">refresh()</code> is required 
   * when a radioset is programatically changed, like in the following circumstances:
   * <ul>
   *   <li>After radios are added or removed or modified (without using ojRadioset) in the DOM.</li>
   * </ul>    
   * 
   * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
   * myComp.refresh();
   * 
   * @expose 
   * @memberof oj.ojRadioset
   * @public
   * @return {void}
   * @instance
   */
  refresh: function() 
  {
    // Set _ResetComponentState. It is called first in EditableValue's refresh,
    // and we override it in ojradioset to reset the this.$radios in case some have 
    // been deleted or overridden.
    
    this._super();      
    this._setup();

  },
  /**
   * Returns a jQuery object containing the element visually representing the radioset. 
   * 
   * <p>This method does not accept any arguments.
   * 
   * @expose
   * @memberof oj.ojRadioset
   * @instance
   * @public
   * @return {jQuery} the radio
   * @ignore
  */
  widget : function ()
  {
    return this.uiRadioset;
  },
  /**
   * Validates the component's display value using all validators registered on 
   * the component and updates the <code class="prettyprint">value</code> option by performing the 
   * following steps. 
   * 
   * <p>
   * <ol> 
   * <li>All messages are cleared, including custom messages added by the app. </li>
   * <li>The implicit 
   * required validator is run if the component is marked required.</li>
   * <li>At the end of validation if there are errors, the messages are shown. 
   * If there were no errors, then the 
   * <code class="prettyprint">value</code> property is updated.</li>
   * </ol>
   * 
   * @example <caption>Validate component using its current value.</caption>
   * myComp.validate();
   * @example <caption>Validate component and use the Promise's resolved state.</caption>
   * myComp.validate().then(
   *  function(result) {
   *    if(result === "valid")
   *    {
   *      submitForm();
   *    }
   *  });
   * @return {Promise.<string>} Promise resolves to "valid" if 
   * the component passed all validations. 
   * The Promise resolves to "invalid" if there were validation errors.
   * 
   * @method
   * @access public
   * @expose
   * @instance
   * @memberof oj.ojRadioset
   * @since 4.0.0
   * @ojstatus preview
   */
  validate : oj.EditableValueUtils.validate,
          
   /**** end Public APIs ****/         
          
  /**** start internal widget functions ****/   
  /**
   * @protected
   * @override
   * @instance
   * @memberof oj.ojRadioset
   */
  _InitOptions : function (originalDefaults, constructorOptions)
  {
    var checkedRadio, domValue;
    var props = [{attribute: "disabled", validateOption: true},
                 {attribute: "required", coerceDomValue: true, validateOption: true},
                 {attribute: "title"}
                 // {attribute: "value"} // code below sets value
               ]; 
    
    this._super(originalDefaults, constructorOptions);
    
    if (!this._IsCustomElement())
    {
      oj.EditableValueUtils.initializeOptionsFromDom(props, constructorOptions, this);  

      // component, app, and constructor are merged into this.options.option by the time _InitOptions 
      // is called. Let's take this example:
      // component (widget) default - 'foo'<br/>
      // app default - 'bar'<br/>
      // dom  - 'lucy'<br/>
      // constructorOptions['value'] - undefined<br/>
      // this.options.option is set to 'bar' initially. We don't want to just take this value, because
      // we want DOM value to win over the app and components default if DOM is set.
      // Therefore, the component needs to check if the constructorOptions['value'] is 
      // undefined and if so, set value option to 'lucy' (the DOM value in this example). <br/>
      // 
      // use DOM value if constructorOptions is undefined. if DOM value is undefined, then
      // leave this.options["value"] alone since it's the merged app/widget default at this point.
      if (constructorOptions['value'] === undefined)
      {
        // return the checked values by simply looking at DOM node
        this.$radios = this._findRadiosWithMatchingName();
        checkedRadio = this.$radios.filter(":checked");
        domValue = (checkedRadio.length === 0) ? undefined : checkedRadio.val();
        // we only use the dom if SOMETHING is checked. If nothing is checked, we stay with whatever
        // is in this.options['value'].
        if (domValue !== undefined)
        {
          // when defaulting from DOM we want to trigger optionChange to writeback new value
          this.option("value", domValue, {'_context':{writeback: true, internalSet: true}});
        }
        // widget defaults to null
        if (this.options['value'] === undefined)
          this.options['value'] = null;
      }
    }
  },      
  /**
   * After _ComponentCreate and _AfterCreate, 
   * the widget should be 100% set up. this._super should be called first.
   * @override
   * @protected
   * @memberof oj.ojRadioset
   * @instance
   */
  _ComponentCreate : function ()
  {
    var element = this.element;
    this._super();
    
    // since the oj-option renderer uses the oj-radioset ID for the name
    // attribute of the rendered radio buttons, let's make sure the radioset
    // has an ID
    element.uniqueId();
    this._processOjOptions();
    // The processOjOptions renders input/label from the oj-options, 
    // so now we need to go through and get this.$radios.
    this.$radios = this._findRadiosWithMatchingName();
    
    // first check to see if this.element is NOT a fieldset. If fieldset, throw error.
    if (element.is("fieldset"))
    {
      throw new Error("ojRadioset cannot be bound to a fieldset. Use a div instead.");
    }
    // Turn each radio into ojRadioCheckbox. Since ojRadiooxSet delegates to the _ojRadioCheckbox
    // component, and we need to mark this as an internal node so that oj.Components.getComponentElementByNode
    // knows it is an internal component in this case, not a stand-alone component
    this.$radios._ojRadioCheckbox().attr('data-oj-internal', '');

    // keep the root dom element and slots as is, and add a wrapper dom underneath it. This way we can
    // have one div around all the inputs and labels, and for inline messaging we can have another
    // div around the inline messaging content. And we can style the borders of the two boxes differently.
    this.uiRadioset = element.addClass("oj-radioset oj-component").attr( "role", "radiogroup" );
    // need to grab all elements using childNodes first and then do filter because jquery children will automatically 
    // exclude all comment and text nodes
    $(element.get(0).childNodes).filter(function(){
        return (this.getAttribute && this.getAttribute("slot") === 'contextMenu') ? false : true;
    }).wrapAll("<div class='oj-radioset-wrapper'></div>"); //@HTMLUpdateOK
    
    this._on(this._events);
    this._setup();
  },
    /**
   * Resets this.$radios. This is called at the beginning of a refresh in EditableValue
   * @override
   * @memberof oj.ojRadioset
   * @instance
   * @protected   
   */
  _ResetComponentState : function ()
  { 
    // we could have added, removed, or modified radios, so we need to re-find all the 
    // inputs on refresh and turn the ones that aren't already ojRadioCheckboxes into them.
    this._processOjOptions();
    this.$radios = this._findRadiosWithMatchingName();
    
    // we have a rule for refresh: if we have a public API for it, then the app dev has to use the
    // option, and not expect changing the dom will update the state with refresh.
    // However, ojRadioset does not expose a public API for the individual radios' disabled state
    // to the app developer. Our private ojRadioCheckbox component has a disabled option that
    // our code has access to.
    // For each radio, we need to look at the disabled attribute dom and update the 
    // ojradiocheckbox's disabled option.

    // !! ensures it is a boolean
    // update the private ojradiocheckbox component's disabled option to keep it in sync with the dom
    this.$radios.filter( ".oj-radio" ).each(function() 
    {
        var disabledValue = $( this ).attr("disabled") !== undefined ? 
          !!$( this ).prop("disabled") : false;
        $( this )._ojRadioCheckbox("option", "disabled", disabledValue);
    });
      
    // no need to refresh the ojRadioCheckbox's that exist since we have options for everything.
    // of the type=radio inputs that are not yet ojRadioCheckboxs, make them ojRadioCheckboxes.

    // create ojRadioCheckboxes on any new ones.
    this.$radios.not(".oj-radio")._ojRadioCheckbox();
  },  
  /**
   * @override
   * @memberof oj.ojRadioset
   * @instance
   * @protected
   * @since 5.0.0
   */
  GetFocusElement : function ()
  {
    // We need :focusable here so that we don't try to focus on an element that isn't focusable.
    return this._GetContentElement().filter(":focusable").first()[0];
  },
  /**
   * Whether the component is required.
   * 
   * @return {boolean} true if required; false
   * 
   * @memberof! oj.ojRadioset
   * @instance
   * @protected
   * @override
   */
  _IsRequired : function () 
  {
    return this.options['required'];
  },
  /**
   * Sets the disabled option onto the dom. 
   * This is a no-op for radioset since its root dom element is a div, and disabled is
   * invalid on a div. If we did try to set disabled on the div, then restore attributes doesn't
   * work correctly since it wasn't saved correctly.
   * @param {Object} node - dom node
   * 
   * @memberof oj.ojRadioset
   * @instance
   * @protected
   * @since 1.0.0
   */
  _SetDisabledDom : function(node)
  {
    // no-op
  },
  /**
   * @memberof oj.ojRadioset
   * @instance
   * @private
   */
  _refreshRequired : oj.EditableValueUtils._refreshRequired,
  /**
   * the validate method from v3.x that returns a boolean
   * @memberof oj.ojRadioset
   * @instance
   * @protected
   * @ignore
   */
  _ValidateReturnBoolean: oj.EditableValueUtils._ValidateReturnBoolean,

  /** 
   * This function processes the oj-option children, sets the custom renderer, and 
   * creates input type=radio and label dom from them.
   * 
   * We don't want to rely on the framework calling the customOptionRenderer 
   * as a result of setting options[i]["customOptionRenderer"] = renderer; in this function. 
   * This could lead to timing bugs when data-oj-binding-provider="none". (when not "none",
   * we know the oj-options are created before the oj-radioset gets created, so no timing issue)
   * Therefore we create the input/label not in the customOptionRenderer, 
   * but in a separate function that we call.
   * @private
   * @instance
   */
  _processOjOptions : function ()
  {
    if (this._IsCustomElement())
    {
      // set the custom renderer on oj-option
      var i, len;
      var renderer = this._customOptionRenderer.bind(this);
      var options = this.element.children("oj-option");
      
      for (i = 0, len = options.length; i < len; i++)
      { 
        options[i]["customOptionRenderer"] = renderer;
        this._initInputLabelFromOjOption(options[i]);
      }

      // for oj-option that are inside the wrapper (after refresh)
      options = this.element.children(".oj-radioset-wrapper").find("oj-option");
      if (options.length > 0) 
      {
        for (i = 0, len = options.length; i < len; i++)
        {
          options[i]["customOptionRenderer"] = renderer;
          this._initInputLabelFromOjOption(options[i]);
        }
      }
    }        
  },

  /**
   * Create the input type='radio'/label dom from attributes on oj-option element.
   * oj-radioset is made up of input/labels.
   * This gets called during the oj-radioset _CreateComponent and refresh
   * @param {Element} elem the oj-option element
   * @private
   * @instance
   */
  _initInputLabelFromOjOption : function (elem)
  {
    var span;
    var label;
    var ojoption = elem;

    // let's make sure that each oj-option has an ID so the
    // label element can reference the input element via the 'for' attribute
    // we have tests in place where oj-option doesn't have id and where it does
    // both for the databound case and non-databound case. In the databound case, the 
    // bindings are resolved before we get here, so we will be fine.
    $(ojoption).uniqueId();
    
    var id = ojoption.getAttribute("id");
    var radioId = id+"|rb";
    var radio = document.getElementById(radioId);
    // Was the oj-option already rendered?
    // It is possible that the answer is true;
    // the use case is when we dynamically add an oj-option to the oj-radioset, and in that
    // case we call _processOjOptions (which in turn calls this function) from _ResetComponentState
    var alreadyProcessed = (radio !== null); 

    // check to see if we've already processed the oj-option dom or not
    // in the code below, we use setAttribute() for everything as we want to be
    // setting the initial value for these elements.
    if (!alreadyProcessed)
    {
      span = document.createElement("span");
      radio = document.createElement("input");
      label = document.createElement("label");
      span.setAttribute("class", "oj-choice-item");
      radio.setAttribute("type", "radio");
      radio.setAttribute("id", radioId);
      label.setAttribute("for", radioId);
      ojoption.parentElement.insertBefore(span, ojoption);//@HTMLUpdateOK
      label.appendChild(ojoption); // append the oj-option as a child of label
      span.appendChild(radio);
      span.appendChild(label);
    }

    var name = this.element[0].id; // Use the id of the ojradioset as the name for the oj-options.
    var ariaLabel = ojoption.getAttribute("aria-label");
    var ariaLabelledBy = ojoption.getAttribute("aria-labelledby");
    
    // The value attribute of the radio only supports text, so we need to be
    // able to access the oj-option's value property instead.  This attribute
    // is a link back to the oj-option so that we don't need to use dom traversal
    // to get to the oj-option to get its value.
    radio.setAttribute("data-oj-option-id", id);

    if (name && name != "")
      radio.setAttribute("name", name);

    if (ariaLabel && ariaLabel != "")
      radio.setAttribute("aria-label", ariaLabel);
    else
      radio.removeAttribute("aria-label");

    if (ariaLabelledBy && ariaLabelledBy != "")
      radio.setAttribute("aria-labelledby", ariaLabelledBy);
    else
      radio.removeAttribute("aria-labelledby");

    if (ojoption.disabled)
      radio.setAttribute("disabled", true);
    else
      radio.removeAttribute("disabled");
  },  
  
  // custom oj-option renderer
  // Because we can't rely on this being called when we set the customOptionRenderer property
  // in _processOjOptions we shouldn't do the input.label creation from the oj-option 
  // in this function. (If we did, the _ComponentCreate code that relies on the
  // inputs/labels being created already would not work.)
  // The correct thing to do is to create the input/label in _initInputLabelFromOjOption
  // Then rely on this function being called after the oj-option has been created and we are
  // changing properties on it.
  _customOptionRenderer : function (elem)
  {
    var ojoption = elem;
    var id = ojoption.getAttribute("id");
    var radioId = id+"|rb";
    var radio = document.getElementById(radioId);
    // Was the oj-option already rendered into an _ojRadioCheckbox() in _CreateComponent?
    var radioExists = (radio !== null); 
    var hasOjRadioClass = radioExists && radio.classList.contains('oj-radio');

    // When an oj-option child is disabled (by setting the disabled attribute to
    // true) and it re-renders, the component should refresh automatically rather than requiring the
    // user to call refresh. See .   
    if (hasOjRadioClass)
    {
      $(radio)._ojRadioCheckbox("option", "disabled", ojoption.disabled);
    }
  },  
  /**
   * If custom element, get the labelledBy option, and set this 
   * onto the root dom element as aria-labelledby. We append "|label" so it matches the id that
   * is on the oj-label's label element.
   * @memberof oj.ojRadioset
   * @instance
   * @private
   */
  _updateLabelledBy : oj.EditableValueUtils._updateLabelledBy,  
  /**
   * Returns a jquery object that is a set of elements that are input type radio
   * and have the name of the first radio found.
   * 
   * @return {jQuery} jquery object of all the radios within the root dom element
   * that have the same 'name' attribute as the first radio found.
   * @private
   * @memberof oj.ojRadioset
   */
  _findRadiosWithMatchingName : function ()
  {
    var allradios;
    var element = this.element;
    var $first = element.find("input[type=radio]:first");
    var name;
    var selector;
      
    if ($first.length === 0)
    {
      oj.Logger.warn("Could not find any input type=radio within this element");
    }
    // get the name attribute of the first input radio
    name = $first.attr("name");
    // find all input radios with matching name
    if (name === undefined)
    {
      // search for all radios with no name
      allradios = element.find("input[type=radio]");
      // now loop and find the ones without 'name' attribute
      return allradios.not("[name]");
    }
    else
    {
      // search for all radios with the name
      selector = "input[type=radio][name=" + name + "]";
      return element.find(selector);
      
    }
  },
  // Override to set custom launcher
  _NotifyContextMenuGesture: function(menu, event, eventType)
  {
    // Setting the launcher to the checked radio if any (since that's what's tabbable in mainstream browsers), 
    // else the first enabled radio (when no selection, all enabled radios are tabbable).
    // Component owner should feel free to specify a different launcher if appropriate.
    // See the superclass JSDoc for _OpenContextMenu for tips on choosing a launcher.
    var radios = this.element.find("input[type=radio]");
    var checked = radios.filter(":checked");
    var launcher = checked.length ? checked : radios.filter(":enabled").first();
    this._OpenContextMenu(event, eventType, {"launcher": launcher});
  },
  // Override to set launcher to widget
  _GetMessagingLauncherElement : function ()
  {
    // focus events only get triggered on input, but they do bubble up and we will capture them
    // on the widget.
    // mouseenter events get called once once if the user hovers over for the entire widget. if
    // we put it on the inputs, it gets called every time you leave and enter a new input. Plus,
    // this doesn't work when we hide the input like we do in the native themes.
     return this.widget();                          
  },
  /**
   * _setup is called on create and refresh. Use the disabled option to 
   * update the component. If the component's option is disabled, then
   * leave it alone.
   * @memberof oj.ojRadioset
   * @instance
   * @private
   */
  _setup: function() 
  {
    // at this point we already have this.$radios set to a list of radios for this radioset.
    this._propagateDisabled(this.options.disabled );
    
    // add to the root dom the style class 'oj-choice-direction-column' 
    // if there isn't already a 'oj-choice-direction-row' or 'oj-choice-direction-column' there.
    if (!(this.element.hasClass("oj-choice-direction-column")) && 
        !(this.element.hasClass("oj-choice-direction-row")))
    {
      this.element.addClass("oj-choice-direction-column");
    }
    this._refreshRequired(this.options['required']);
    
    this._updateLabelledBy(null, this.options['labelledBy'], this.widget());
  },  
  _events : 
  {
    'change' : function (event)
    {      
      this._HandleChangeEvent(event);
    }
  },

  /**
   * @param {Event} event DOM event 
   * @override
   * @protected
   * @memberof oj.ojRadioset
   */
  _HandleChangeEvent: function(event)
  {
    // TODO make sure the target is an input radio?
    // TODO any more checks I need to do?
    //alert("XYZ In _changeSetValue target is " + event.target + " And the value of the input is " + event.target.value);

    // should I double check that the event.target is the same as the 'checked'?
    // if (event.target === this.$radios.filter(":checked"))???
    // 
    var submittedValue = this._GetDisplayValue();
    // run full validation. There is no need to check if values have changed
    //  since for checkboxset/radioset if we get into this function we know value has changed.
    // passing in doValueChangeCheck: false will skip the new-old value comparison
    this._SetValue(submittedValue, event, _sValueChangeCheckFalse);
  },
                  
  /**
   * Returns the display value that is ready to be passed to the converter.
   * 
   * @param {Object} value the stored value if available that needs to be formatted for display
   * @override
   * @protected
   * @memberof oj.ojRadioset
   */
  _GetDisplayValue : function (value) 
  {
    // return the value of the 'checked' radio
    return this._GetElementValue();
  },
  /**
   * Called when the display value on the element needs to be updated 
   * as a result of a value change. 
   * ojRadioset stores a String value, and this value matches the value
   * of the currently checked radio. So, if we need to set the display value,
   * what this means is we need to 'check' the radio whose value matches the
   * displayValue.
   * 
   * @param {String} displayValue the value of the radio button that needs to be selected
   * @override
   * @protected
   * @memberof oj.ojRadioset
  */  
  _SetDisplayValue : function (displayValue) 
  {
    var i;
    var length = this.$radios.length;
    var matchIndex = -1;
    var radios = this.$radios.get();
    var radioValues = [];
    var $radio;

    // Find the element matching displayValue
    // first, do an === compare
    for (i = 0; i < length; i++)
    {
      // save the values for the deep compare case
      radioValues[i]= this._GetOptionValue(radios[i]);
        
      if (radioValues[i] === displayValue)
      {
        matchIndex = i;
        break;
      }
    }
    
    // If not found, do a deep compare
    if (matchIndex === -1)
    {
      for (i = 0; i < length; i++)
      {
        if (oj.Object.compareValues(radioValues[i], displayValue))
        {
          matchIndex = i;
          break;
        }
      }
    }

    // go through each _ojRadioCheckbox and see if it needs to be checked or unchecked.
    for (i = 0; i < length; i++)
    {
      $radio = $(this.$radios[i]);
      // does the radio's value match the displayValue?
      var checked = $radio._ojRadioCheckbox("option", "checked"); 
      if (i === matchIndex)
      {
        // yes. this needs to be checked, if it isn't already
        if (!checked)
        {
          $radio._ojRadioCheckbox("option", "checked", true);
        }
      }
      else
      {
        /// no. this needs to be unchecked, if it isn't already
        if (checked)
        {
          $radio._ojRadioCheckbox("option", "checked", false);
        }
      }
    } 
  },
  /**
   * Returns the element's value. Normally, this is a call to this.element.val(),
   * but in the case of ojRadioset, the element's value is really the value
   * of the checked radio in the set.
   * @override
   * @protected
   * @memberof oj.ojRadioset
   * @return {*} The value of the selected radio, or null if no selection.
   */
  _GetElementValue : function () 
  {
    // "input:checked" selects radios that are currently checked as 
    // reflected in their boolean (true or false) checked property, 
    // which is affected when the user clicks the radio for example.
    // for radio, there will be one or none checked; 
    // if none are checked, return null (checkedRadio.val() is undefined if nothing is checked)
    var checkedRadio = this.$radios.filter(":checked");
    if (checkedRadio.length === 0)
      return null;
    else
      return this._GetOptionValue(checkedRadio[0]);
  },
  /**
   * For custom element, we get the value from the oj-option element, otherwise
   * we get the value from the radio element.
   * @override
   * @protected
   * @memberof oj.ojRadioset
   * @return {*} Returns the value property of the associated oj-option, or the value attribute of the radio element.
   */
  _GetOptionValue : function (radioElem)
  {
    var option;
    var val;
    
    if (this._IsCustomElement())
    {
      option = document.getElementById(radioElem.getAttribute("data-oj-option-id"));
    
      if (option)
      {
        val = option.value;
      }
    }
    else
    {
      val = radioElem.value;
    }
    
    return val;
  },
  /**
   * Returns the default styleclass for the component. Currently this is 
   * used to pass to the ojLabel component, which will append -label and 
   * add the style class onto the label. This way we can style the label
   * specific to the input component. For example, for inline labels, the
   * radioset/checkboxset components need to have margin-top:0, whereas all the
   * other inputs need it to be .5em. So we'll have a special margin-top style 
   * for .oj-label-inline.oj-radioset-label
   * All input components must override
   * 
   * @return {string}
   * @memberof oj.ojRadioset
   * @override
   * @protected
   */
  _GetDefaultStyleClass : function ()
  {
    return "oj-radioset";
  },
  /**
   * Returns a jquery object of the elements representing the 
   * content nodes (input type=radio). This is used in EditableValue to add
   * aria-describedby to the input when there is a help icon, to add
   * aria-required and aria-invalid
   * @protected
   * @override
   * @memberof oj.ojRadioset
   */
  _GetContentElement : function ()
  {
    if (this.$radios != null)
      return this.$radios;
    else 
    {
      this.$radios = this._findRadiosWithMatchingName();
      return this.$radios
    }
  }, 
  /**
   * Called to find out if aria-required is unsupported. This is needed for the label.
   * It is not legal to have aria-required on radio/checkboxes, nor on
   * radiogroup/group.
   * If aria-required is not supported, then we wrap the required icon as well as the
   * help icons so that JAWS can read required. We don't do this for form controls that use
   * aria-required because if we did JAWS would read required twice.
   * @memberof oj.ojRadioset
   * @instance
   * @protected
   * @return {boolean}
   */
  _AriaRequiredUnsupported : function()
  {
    return true;
  },
  /**
   * Performs post processing after required option is set by taking the following steps.
   * 
   * - if component is invalid and has messgesShown -> required: false/true -> clear component errors; 
   * run full validation with UI value (we don't know if the UI error is from a required validator 
   * or something else);<br/>
   * &nbsp;&nbsp;- if there are validation errors, then value not pushed to model; messagesShown is 
   * updated<br/>
   * &nbsp;&nbsp;- if no errors result from the validation, push value to model; author needs to 
   * listen to optionChange(value) to clear custom errors.<br/>
   * 
   * - if component is invalid and has messagesHidden -> required: false -> clear component 
   * errors; no deferred validation is run.<br/>
   * - if component has no error -> required: true -> run deferred validation (we don't want to flag 
   * errors unnecessarily)<br/>
   * - messagesCustom is never cleared<br/>
   * 
   * @param {string} option
   * 
   * @memberof oj.ojRadioset
   * @instance
   * @protected
   */
  _AfterSetOptionRequired : oj.EditableValueUtils._AfterSetOptionRequired,
  
  /**
   * @private
   * @memberof oj.ojRadioset
   */
  _propagateDisabled: function( disabled ) {
      disabled = !!disabled;
      this.$radios.each(function() 
      {
        // this is the technique to use to call package-private functions
        // Calling it like this.$radios.ojRadioCheckbox("__setAncestorComponentDisabled",value)
        // gives an error because jquery prevents you from calling functions with an "_"
        // 
        // This is how we handle 'disabled' for radioset. We don't change the radiocheckbox
        // component's disabled option ever since if we do that we've lost what the initial disabled
        // state is (we store the disabled dom value from the radio into its disabled option)
        // and we need that when we refresh. Instead what we do
        // is we mark if its ancestor (the radioset) is disabled or not. Then, when we render
        // out the radios 'disabled' state, like oj-disabled, we look to see if it is 'effectively
        // disabled' (see _IsEffectivelyDisabled call in ojRadioCheckbox), that is if its 
        // option is disabled OR its ancestor (the radioset) is disabled.
        $( this ).data("oj-_ojRadioCheckbox").__setAncestorComponentDisabled(disabled);
      });
 
      this.$radios._ojRadioCheckbox("refreshDisabled");
  },
  /**
   * Note that _setOption does not get called during create in the super class. 
   * It only gets called when the component has already been created.
   * @override
   * @private
   * @memberof oj.ojRadioset
   */
  _setOption : function (key, value, flags)
  {
    var originalValue = this.options["labelledBy"];
 
    this._super(key, value, flags);
    
    switch (key)
    {
      case "disabled":
        this._propagateDisabled(value);
        break;
      case "labelledBy":
        // remove the old one and add the new one
        this._updateLabelledBy(originalValue, value, this.widget());
        break;
      default:
        break;
    }
  },
  /**
   * Performs post processing after _SetOption() is called. Different options when changed perform
   * different tasks. See _AfterSetOption[OptionName] method for details.
   *
   * @param {string} option
   * @param {Object|string=} previous
   * @param {Object=} flags
   * @protected
   * @memberof oj.ojRadioset
   * @instance
   * @override
   */
  _AfterSetOption : function (option, previous, flags)
  {
    this._superApply(arguments);
    switch (option)
    {
      case "required":
        this._AfterSetOptionRequired(option);
        break;
      default:
        break;
    }
  },
  //** @inheritdoc */
   getNodeBySubId: function(locator)
  {
    var node = this._super(locator),
        radios,
        subId,
        value;

    if (!node)
    {
      radios = this.$radios.get();
      subId = locator['subId'];
      
      switch (subId)
      {
        case 'oj-radioset-inputs': // TODO: depricated for a while now, remove this in 4.0.0
          node = radios;
          break;

        case 'oj-radioset-radio':
          // We find these by the value attribute on the input element, which is
          // much more stable that using an index.
          value = locator['value'];

          if (typeof value != 'undefined')
          {
            var arrayLength = radios.length;
            var matchIndex = -1;
            var radioValues = [];
                  
            // Find the index of the match
            for (var i = 0; i < arrayLength; i++)
            {
              // save the values just in case we need to do deep compare
              radioValues[i] = this._GetOptionValue(radios[i]);
              
              if (radioValues[i] === value)
              {
                matchIndex = i;
                break;
              }
            }
            // If not found, do a deep compare
            if (matchIndex === -1)
              for (i = 0; i < arrayLength; i++)
                if (oj.Object.compareValues(radioValues[i], value))
                {
                  matchIndex = i;
                  break;
                }

            if (matchIndex !== -1)
              node = radios[matchIndex];
          }
          break;
      }
    }

    // Non-null locators have to be handled by the component subclasses
    return node || null;
  },

  //** @inheritdoc */
  getSubIdByNode: function(node) {
    var topElem = this._GetContentElement()[0].parentElement.parentElement.parentElement;
    var currentNode = node;
    
    while (currentNode  && currentNode != topElem) {
      if (currentNode.nodeName == "LABEL")
        currentNode = document.getElementById(currentNode.for);
      
      if (currentNode.nodeName == "INPUT")
      {
        return {"subId":"oj-radioset-radio", "value":this._GetOptionValue(currentNode)};
      }

      currentNode = currentNode.parentElement;
    }
    return this._super(node);
  },

  /**
   * @ignore
   * @protected
   * @memberof oj.ojRadioset
   * @override
   */
  _destroy : function ()
  {  
    var ret = this._super();
    var wrapperDom = this.element[0].firstElementChild;
    
    if (this.$radios)
    {
      this.$radios._ojRadioCheckbox( "destroy" );
    }

    // remove the dom we added to wrap the children of this.element, but don't remove the children.
    $(wrapperDom).contents().unwrap();

    return ret;
  }
  /**** end internal widget functions ****/ 
 
  /**
   * Removes the radioset functionality completely. 
   * This will return the element back to its pre-init state.
   * 
   * <p>This method does not accept any arguments.
   * 
   * @method
   * @name oj.ojRadioset#destroy
   * @memberof oj.ojRadioset
   * @instance
   * @ignore
   * 
   * @example <caption>Invoke the <code class="prettyprint">destroy</code> method:</caption>
   * $( ".selector" ).ojRadioset( "destroy" );
   */
});

//////////////////     SUB-IDS     //////////////////


/**
 * <p>Sub-ID for the radioset's radios.</p>
 *
 * @ojsubid oj-radioset-inputs
 * @deprecated 3.0.0 Since the application supplies the input elements, it can supply a unique ID by which the input elements can be accessed.
 * @ignore
 * @memberof oj.ojRadioset
 *
 * @example <caption>Get the nodes for the radios:</caption>
 * var nodes = $( ".selector" ).ojRadioset( "getNodeBySubId", {'subId': 'oj-radioset-inputs'} );
 */

  // Fragments:

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Gesture</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>Input</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Select the input. In some themes, the input is not visible, 
   *       so you will tap on the label.</td>
   *     </tr>
   *     <tr>
   *       <td>Input's label</td>
   *       <td><kbd>Tap</kbd></td>
   *       <td>Select the corresponding input.</td>
   *     </tr> 
   *     <tr>
   *       <td>Input or Label</td>
   *       <td><kbd>Press & Hold</kbd></td>
   *       <td>If hints, help.instruction or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *    </tr> 
   *   </tbody>
   * </table>
   *
   * 
   * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
   * @memberof oj.ojRadioset
   */

  /**
   * <table class="keyboard-table">
   *   <thead>
   *     <tr>
   *       <th>Target</th>
   *       <th>Key</th>
   *       <th>Action</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td rowspan="2">Input</td>
   *       <td><kbd>UpArrow</kbd></td>
   *       <td>Select the previous input in the group.</td>
   *     </tr>
   *     <tr>
   *       <td><kbd>DownArrow</kbd></td>
   *       <td>Select the next input in the group.</td>
   *     </tr>
   *     <tr>
   *       <td>Radioset</td>
   *       <td><kbd>Tab In</kbd></td>
   *       <td>Set focus to the checked radio input. If hints, title or messages exist in a notewindow, 
   *        pop up the notewindow.</td>
   *     </tr> 
   *   </tbody>
   * </table>
   *
   * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
   * @memberof oj.ojRadioset
   */

  /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     * 
   * <table class="generic-table styling-table">
   *   <thead>
   *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocExampleHeader"}</th>
   *     </tr>
   *   </thead>
   *   <tbody>
   *     <tr>
   *       <td>oj-choice-direction-column</td>
     *       <td>This is the default. It lays out the radios in a column.
     *       </td>
     *       <td>
   * <pre class="prettyprint">
   * <code>&lt;oj-radioset id="radiosetId" 
   *  class="oj-choice-direction-column">
   * </code></pre>
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-choice-direction-row</td>
     *       <td>It lays out the radio buttons in a row.
     *       </td>
     *       <td>
   * <pre class="prettyprint">
   * <code>&lt;oj-radioset id="radiosetId" 
   *  class="oj-choice-direction-row">
   * </code></pre>
   *       </td>
   *     </tr>   
   *     <tr>
   *       <td>oj-radioset-no-chrome</td>
   *       <td>Use this styleclass if you don't want the chrome around the set.
   *       </td>
   *       <td>
   * <pre class="prettyprint">
   * <code>&lt;oj-radioset id="radiosetId"
   *  class="oj-radioset-no-chrome">
   * </code></pre>
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-radioset-input-start</td>
   *       <td>Use this styleclass to order the radio at the start and label text at the end 
   *       even if a theme has a different default order.
   *       </td>
   *       <td>
   * <pre class="prettyprint">
   * <code>&lt;oj-radioset id="radiosetId"
   *  class="oj-radioset-input-start">
   * </code></pre>
   *       </td>
   *     </tr>
   *     <tr>
   *       <td>oj-radioset-input-end</td>
   *       <td>Use this styleclass to order the radio at the end and the label text at the start 
   *       even if a theme has a different default order.
   *       </td>
   *       <td>
   * <pre class="prettyprint">
   * <code>&lt;oj-radioset id="radiosetId" 
   *  class="oj-radioset-input-end">
   * </code></pre>
   *       </td>
   *     </tr>
   *   </tbody>
   * </table>
   *
   * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
   * @memberof oj.ojRadioset
   */

// -----------------------------------------------------------------------------
// "private static members" shared by all radiosets
// -----------------------------------------------------------------------------
/**
 * do not do a value change check in _SetValue
 */
var _sValueChangeCheckFalse = {doValueChangeCheck: false};

}());
(function() {
var ojRadiosetMeta = {
  "properties": {
    "disabled": {
      "type": "boolean"
    },
    "labelledBy": {
      "type": "string"
    },    
    "required": {
      "type": "boolean"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "required": {
          "type": "Object",
          "properties": {
            "hint": {
              "type": "string"
            },
            "messageDetail": {
              "type": "string"
            },
            "messageSummary": {
              "type": "string"
            }
          }
        }
      }
    },  
    "value": {
      "type": "any",
      "writeback": true
    }
  },
  "methods": {
    "validate": {}
  },
  "extension": {
    _WIDGET_NAME: "ojRadioset"
  }
};
oj.CustomElementBridge.registerMetadata('oj-radioset', 'editableValue', ojRadiosetMeta);
oj.CustomElementBridge.register('oj-radioset', {'metadata': oj.CustomElementBridge.getMetadata('oj-radioset')});
})();
/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */
/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

(function () {
  /*!
   * JET Slider @VERSION
   *
   *
   * Depends:
   *  jquery.ui.widget.js
   */

  /**
   * @ojcomponent oj.ojSlider
   * @ojdisplayname Slider
   * @augments oj.editableValue
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojSlider extends editableValue<number, ojSliderSettableProperties>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojSliderSettableProperties extends editableValueSettableProperties<number>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * 
   * @ojrole slider
   * @since 0.7
   * @ojshortdesc Displays an interactive slider element.
   * @ojstatus preview
   *
   * @classdesc
   * <h3 id="sliderOverview-section">
   *   JET Slider Component
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#sliderOverview-section"></a>
   * </h3>
   * <p>Description: The oj-slider component enhances an HTML
   * <code class="prettyprint">input</code> element into an interactive slider.
   * </p>
   * The numerical <code class="prettyprint">value</code> attribute determines the 
   * current value of the slider, and thus affects the position of the slider thumb.
   * The value should be between the <code class="prettyprint">min</code> and
   * <code class="prettyprint">max</code> attribute values.
   * </p>
   * The <code class="prettyprint">step</code> attribute of the slider specifies the 
   * interval between thumb stops. For example,
   * if <code class="prettyprint">min</code>  is set to 0 and
   * <code class="prettyprint">max</code>
   * is set to 10, a <code class="prettyprint">step</code> value of 2 would allow the thumb
   * to be positioned at 0, 2, 4, 6, 8, and 10.
   * </p>
   * The <code class="prettyprint">orientation</code> attribute defaults to
   * <code class="prettyprint">"horizontal"</code>.
   * Set <code class="prettyprint">orientation</code> to
   * <code class="prettyprint">"vertical"</code> for a vertical slider (one where the thumb
   * travels along the vertical axis).
   * </p>
   * The <code class="prettyprint">type</code> attribute is used to effect the rendered
   * style of the slider.
   * The <code class="prettyprint">type</code> attribute defaults to
   * <code class="prettyprint">"fromMin"</code>, which will style the value bar from the minimum
   * value to the slider thumb.
   * The <code class="prettyprint">type</code> attribute to either "single" or "fromMax" -
   * this will alter the rendered style of the slider's bar value.
   * </p>
   * Set the <code class="prettyprint">disabled</code> attribute
   * <code class="prettyprint">true</code> to display a slider that displays a value but does
   * not allow interaction.
   * </p>
   * Use <code class="prettyprint">style </code> attributes on the 
   * <code class="prettyprint">oj-slider </code> element to set a horizontal slider's 
   * width or a vertical slider's height.
   * </p>
   * Use the <code class="prettyprint">transient-value</code> attribute to access
   * slider value changes during slider thumb repositioning.
   * </p>
   * Note that the <code class="prettyprint">range</code> value for the
   * <code class="prettyprint">type</code> attribute
   * is not part of the initial (4.0) release of the custom element slider. 
   * </p>
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDoc"}
   *
   * <h3 id="accessibility-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
   * </h3>
   * <p>
   * The slider component is accessible - it sets and maintains the appropriate aria- attributes,
   * including <code class="prettyprint">aria-valuenow</code>,
   * <code class="prettyprint">aria-valuemax</code>,
   * <code class="prettyprint">aria-valuemin</code>
   * and <code class="prettyprint">aria-orientation</code>.
 * <h3 id="label-section">
 *   Label and Slider
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
 * </h3>
 * <p>
 * It is up to the application developer to associate the oj-label to the oj-slider element.
 * For accessibility, you should associate a oj-label element with the oj-slider element
 * by putting an <code class="prettyprint">id</code> on the oj-slider element, and then setting the
 * <code class="prettyprint">for</code> attribute on the oj-label to be the slider element's id.
 * </p>
   * </p>
   * In addition, the slider thumb element can be accessed programmatically.
   * This approach may be necessary to ensure accessibility conformance.
   * For example, if the slider controls another element that is in a remote area of the page,
   * then the <code class="prettyprint">aria-controls</code> attribute for the slider thumb
   * should be set.
   *
   * <p>
   * Consider an example where you may need to set additional attributes for accessibility reasons.
   * Suppose there is another component that is in a remote area of the page
   * that controlled by the slider.
   * Assume that the <code class="prettyprint">id</code> of the remote element is
   * "idOfRemoteElement".
   * Below we show how to access the thumb element in order to set the
   * <code class="prettyprint">aria-controls</code> attribute of the thumb to point to the the
   * id ("idOfRemoteElement") of the remote html element:
   *
   * <pre class="prettyprint">
   * <code>
   *     var thumb0 = myComponent.querySelectorAll('.oj-slider-thumb')[0];
   *     thumb0.setAttribute(aria-controls, "idOfRemoteElement");
   * </code></pre>
   *
   * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.
   *
   * <h3 id="styling-section">
   *   Styling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
   * </h3>
   * {@ojinclude "name":"stylingDoc"}
   *
   * @example <caption>Declare the oj-slider component with no attributes specified:</caption>
   * &lt;oj-slider>&lt;/oj-slider>
   *
   * @example <caption>Initialize the slider with a few attributes:</caption>
   * &lt;oj-slider value=10 max=100 min=0 step=2>&lt;/oj-slider>
   *
   * @example <caption>Initialize a component attribute via component binding:</caption>
   * &lt;oj-slider value="{{currentValue}}">&lt;/oj-slider>
   */
  oj.__registerWidget('oj.ojSlider', $['oj']['editableValue'], {
    defaultElement: '<input>',
    version: '1.0.1',
    widgetEventPrefix: 'oj',

    options: {
      /**
       * The maximum value of the slider.
       * The <code class="prettyprint">max</code> must not be less than the
       * <code class="prettyprint">min</code>, or else an Error is thrown during
       * initialization.
       * @expose
       * @memberof oj.ojSlider
       * @instance
       * @type {?number}
       * @default null
       * @since 0.7
       * @example <caption>Initialize the slider with the
       * <code class="prettyprint">max</code> attribute:</caption>
       * &lt;oj-slider max=100>&lt;/oj-slider>
       * @example <caption>Get or set the <code class="prettyprint">max</code> property after initialization:</caption>
       * // Getter
       * var max = myComponent.max;
       * 
       * // Setter
       * myComponent.max = 100;
       */
      max: 100,

      /**
       * The minimum value of the slider.
       * The <code class="prettyprint">min</code> must not be greater than the
       * <code class="prettyprint">max</code>, or else an Error is thrown during
       * initialization.
       * @expose
       * @memberof oj.ojSlider
       * @instance
       * @type {?number}
       * @default null
       * @since 0.7
       * @example <caption>Initialize the slider with the
       * <code class="prettyprint">min</code> attribute:</caption>
       * &lt;oj-slider min=0>&lt;/oj-slider>
       * @example <caption>Get or set the <code class="prettyprint">min</code> property after initialization:</caption>
       * // Getter
       * var min = myComponent.min;
       * 
       * // Setter
       * myComponent.min = 0;
       * 
       */
      min: 0,

      /**
       * Specify the orientation of the slider.
       *
       * @expose
       * @memberof! oj.ojSlider
       * @instance
       * @type {string}
       * @ojvalue {string} "horizontal" Orient the slider horizontally.
       * @ojvalue {string} "vertical" Orient the slider vertically.
       * @default "horizontal"
       * @since 0.7
       *
       * @example <caption>Initialize the slider with the
       * <code class="prettyprint">orientation</code> attribute:</caption>
        * &lt;oj-slider orientation="vertical">&lt;/oj-slider>
       *
       * @example <caption>Get or set the <code class="prettyprint">orientation</code>
       * property after initialization:</caption>
       * // Getter
       * var orientation = myComponent.orientation;
       * 
       * // Setter
       * myComponent.orientation = "vertical";
       * 
       */
      orientation: 'horizontal',

      /**
       * readOnly is private - more UX design is necessary to support readonly across
       * components.
       * Whether the component is readOnly. The element's
       * <code class="prettyprint">readOnly</code>
       * attribute is used as its initial value if it exists, when the attribute is not explicitly
       * set. When neither is set, <code class="prettyprint">readOnly </code>
       * defaults to false.
       *
       * @example <caption>Initialize component with <code class="prettyprint">readOnly</code>
       * attribute:</caption>
        * &lt;oj-slider readOnly="true">&lt;/oj-slider>
       * @example <caption>Get or set the <code class="prettyprint">readOnly</code> property after initialization:</caption>
       * // Getter
       * var readOnly = myComponent.readOnly;
       * 
       * // Setter
       * myComponent.readOnly = true;
       * 
       *
       * @private
       * @type {?boolean}
       * @default false
       * @instance
       * @memberof oj.ojSlider
       */
      readOnly: false,

      /**
       * Whether the component is disabled. The 
       * <code class="prettyprint">disabled</code> attribute is used as its initial
       * value if it exists, when the attribute is not explicitly set. When neither is set,
       * <code class="prettyprint">disabled </code>
       * defaults to false.
       *
       * @example <caption>Initialize the slider with
       * <code class="prettyprint">disabled</code> attribute:</caption>
        * &lt;oj-slider disabled="true">&lt;/oj-slider>
       * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
       * // Getter
       * var disabled = myComponent.disabled;
       * 
       * // Setter
       * myComponent.disabled = true;
       *
       * @expose
       * @type {?boolean}
       * @default false
       * @since 0.7
       * @instance
       * @memberof oj.ojSlider
       */
      disabled: false,

      /**
       * Determines the size or amount of each interval or step the slider takes
       * between min and max.
       * The full specified value of the range (max - min) should be divisible by step.
       *
       * @expose
       * @instance
       * @type {?number}
       * @default 1
       * @since 0.7
       * @memberof oj.ojSlider
       * @example <caption>Initialize the slider with the
       * <code class="prettyprint">step</code> attribute:</caption>
       * &lt;oj-slider step=10>&lt;/oj-slider>
       * @example <caption>Get or set the <code class="prettyprint">step</code> property after initialization:</caption>
       * // Getter
       * var step = myComponent.step;
       * 
       * // Setter
       * myComponent.step = 10;
       * 
       **/
      step: 1,

      /**
       * The slider type determines whether the slider how the value is represented in the UI.
       *
       * @expose
       * @type {?string}
       * @ojvalue {string} "fromMin" A single-thumb slider where the value bar goes from
       * the slider min to the the slider thumb.
       * @ojvalue {string} "fromMax" A single-thumb slider where the value bar goes from
       * the slider thumb to the the slider max.
       * @ojvalue {string} "single" A single-thumb slider where the value bar has no
       * additional styling.
       * @default "fromMin"
       * @since 0.7
       * @instance
       * @memberof oj.ojSlider
       *
       * @example <caption>Initialize component with <code class="prettyprint">type</code>
       * attribute set to "fromMax":</caption>
        * &lt;oj-slider type="fromMax">&lt;/oj-slider>
       * @example <caption>Get or set the <code class="prettyprint">type</code> property after initialization:</caption>
       * // Getter
       * var type = myComponent.type;
       * 
       * // Setter
       * myComponent.type = "fromMax";
       *
       */
      type: 'fromMin',

      /**
       * The numerical value of the slider.
       *
       * <p> Note that the <code class="prettyprint">value</code> attribute should
       * be compatible with the <code class="prettyprint">type</code> attribute, as
       * described above.
       *
       * @example <caption>Initialize the slider with the
       * <code class="prettyprint">value</code> attribute:</caption>
        * &lt;oj-slider value=55>&lt;/oj-slider>
       * @example <caption>Get or set <code class="prettyprint">value</code> property
       * after initialization:</caption>
       * // Getter
       * var value = myComponent.value;
       * 
       * // Setter
       * myComponent.value = 10;
       * 
       * @expose
       * @access public
       * @instance
       * @default 0
       * @since 0.7
       * @ojwriteback
       * @memberof oj.ojSlider
       * @type {?number}
       */
      value: 0,

      /**
       * <p>The  <code class="prettyprint">transientValue</code> is the read-only attribute for
       * retrieving the transient value from the slider.</p>
       * <p>
       * The <code class="prettyprint">transientValue</code> updates to display the transient
       * changes of the slider thumb value (subject to the step constraints). The difference
       * in behavior is <code class="prettyprint">transientValue</code> will be updated
       * as the thumb is sliding, where as <code class="prettyprint">value</code>
       * is updated only after the thumb is released (or after a key press).
       * </p>
       * <p>This is a read-only attribute so page authors cannot set or change it directly.</p>
       * @expose
       * @alias transientValue
       * @access public
       * @instance
       * @ojwriteback
       * @memberof oj.ojSlider
       * @type {?number}
       * @since 5.0
       * @readonly
       * @ojstatus preview
       */
      rawValue: null

    },

    // number of pages in a slider
    // (how many times can you page up/down to go through the whole range)
    _numPages: 5,
    _defaultElementId: null,

    _sliderDisplayValue: null,

    _isRTL: function () {
      return oj.DomUtils.getReadingDirection() === 'rtl';
    },

    _ComponentCreate: function () {
      this._super();

      // ojSlider could support both <div> and <input> tags
      this._inputtag = false;

      if (this.element.is('INPUT')) {
        this._inputtag = true;

        // Save the input tag style, apply to sliderWrapper later in the code.
        if (this.element[0].style) {
          this._styleFromInputTag = this.element[0].style.cssText;
        }

        this._inputElementOriginalDisplay = this.element.css('display');
        this.element.css('display', 'none');

        if (this.OuterWrapper) {
          this._elementWrapped = $(this.OuterWrapper);
        } else {
          var inputDiv = $(this.element).wrap('<div> </div>'); // @HTMLUpdateOK

          // _elementWrapped is the new wrapped input element.
          this._elementWrapped = inputDiv.parent();
        }
      } else {
        this._elementWrapped = this.element;
      }

      // use 'transientValue' for custom elements,
      // and use 'rawValue' for widget syntax.
      this._transientValueName = this._IsCustomElement() ? "transientValue" : "rawValue";
      this._componentSetup();
    },

     /**
      * Convenience function to set the rawValue option. Overrides EditableValue _SetRawValue.
      * @param {String} val value to set rawValue to
      * @param {Event} event DOM event 
      * @return {void}
      * @memberof oj.ojSlider
      * @instance
      * @override
      * @private
      */
     _SetRawValue : function (val, event) {
       var flags = {};
       flags['_context'] = {originalEvent: event, writeback: true, internalSet: true, readOnly: true};
       if (!oj.Object.compareValues(this.options[this._transientValueName], val)) {
         this.option(this._transientValueName, val, flags);
       }
     },

    //
    // Setup the component based on the current options.
    // Also create DOM elements for thumbs and bars.
    //
    _componentSetup: function () {
      this._newMultiValue = [];
      this._thumbIndex = null;

      var classes = 'oj-slider ';

      if (this._isVertical()) {
        classes += 'oj-slider-vertical';
      } else {
        classes += 'oj-slider-horizontal';
      }

      classes += ' oj-component oj-form-control';

      if (this.options.readonly) classes += ' oj-read-only';
      if (this.options.disabled) classes += ' oj-disabled';

      this._elementWrapped.removeClass();
      this._elementWrapped.addClass(classes);

      if (this.options.type === 'range') {
        this._multipleThumbs = true;
      } else {
        this._multipleThumbs = false;
      }

      this._calculateNewMax();

      this._createSliderContainer();
      this._createBarBackground();
      this._buildValueOption();
      this._createRange();
      this._createThumbs();
      this._updateUI();

      this._setupEvents();
    },

    _copyLabelledbyToThumb: function(labelId) {

        var thumb = this._elementWrapped.find('.oj-slider-thumb');
        thumb.attr('aria-labelledby', labelId);

        if (thumb.length > 1) {
          var thumb2 = thumb[1];
           $(thumb2).attr('aria-labelledby', String(labelId));
        }
    },

    _AfterCreate: function () {
      this._super();

      this._makeDraggable();
      var thumb;

      // For custom elements, set the aria-labelledby on the input element
      // (provided the oj-label is part of the form)
      if (this.OuterWrapper)  {
          var defaultLabelId = this["uuid"] + "_Label";
          var LId = oj.EditableValueUtils.getOjLabelId(this.widget(), defaultLabelId);
          if (LId)
              this.element.attr("aria-labelledby", LId);
      }

      //
      // For custom elements, if the input has an aria-labelledby, then just
      // copy it to the thumb (since EditableValue has already determined that 
      // the slider control is using aria-labelledby)
      //
      if (this.OuterWrapper) {
        var ariaId = this.element.attr('aria-labelledby');
        if (ariaId !== undefined && ariaId !== null) {
          this._copyLabelledbyToThumb(ariaId);
          return;
        }
      }

      var label = this._GetLabelElementLocal();

      //
      // Copy any labelled-by on the <input labelled-by="id"> to the slider thumb.
      //
      if (label) {
        //
        // this id should be on the thumb: aria-labelledby =
        //

        // Set the aria-labelledby attribute of the thumb to the returned id.
        var labelId = label.attr('id');
        if (!labelId) {
          labelId = label.attr('for');
        }

        if (labelId.length > 0) {
          this._copyLabelledbyToThumb(labelId);
        }

      } else {
        //
        // Check if the <input> has aria-label=""
        //

        var ariaLabelString;
          
        if (this.OuterWrapper) {
          ariaLabelString = this._elementWrapped.attr('aria-label');
        } else {
          ariaLabelString = this.element.attr('aria-label');
        }

        if (ariaLabelString) {

          thumb = this._elementWrapped.find('.oj-slider-thumb');

          // Set the aria-labelledby attribute of the thumb to the returned id
          thumb.attr('aria-label', ariaLabelString);
        }
      }
    },

    _GetLabelElementLocal: function () {

      // If <input> has aria-labelledby set, then look for label it is referring to.
      var queryResult = this._getAriaLabelledByElementLocal();
      if (queryResult !== null && queryResult.length !== 0) {
        return queryResult;
      }

      queryResult = this._getAriaLabelForElementLocal();
      if (queryResult !== null && queryResult.length !== 0) {
        return queryResult;
      }

      return null;
    },

    _getAriaLabelForElementLocal: function () {

      var id;
      id = this.element.prop('id');

      if (id !== undefined) {
        var labelQuery = "label[for='" + id + "']";

        var jqLabelQuery = $(labelQuery);
        if (jqLabelQuery.length > 0) return jqLabelQuery;

        var spanQuery = "span[for='" + id + "']";

        if ($(spanQuery).length !== 0) {
          return $(spanQuery);
        }
      }
      return null;
    },

    _getAriaLabelledByElementLocal: function () {
      // look for a label with an id equal to the value of aria-labelledby.
      // .prop does not work for aria-labelledby. Need to use .attr to find
      // aria-labelledby.

      var ariaId = this.element.attr('aria-labelledby');

      // Handle direct labelling case for custom elements
      // (this is not the common case, but still is supported)
      if (ariaId === undefined && this.OuterWrapper) {
          // ariaId = this._elementWrapped.attr('aria-labelledby');
          var ariaIdDirect = this._elementWrapped.attr('aria-labelledby');
          if (ariaIdDirect) {
              if (document.getElementById(ariaIdDirect) !== null) {
                  return $(document.getElementById(ariaIdDirect));
              }
          }
      }

      if (ariaId !== undefined) {
        var jqLabelQuery = $("label[id='" + ariaId + "']");
        if (jqLabelQuery.length > 0) return jqLabelQuery;

        var jqSpanQuery = $("span[id='" + ariaId + "']");
        if (jqSpanQuery.length > 0) {
          return jqSpanQuery;
        }
      }
      return null;
    },

    widget: function () {
      return this._elementWrapped;
    },

    /**
     * Called when the display value on the element needs to be updated. This method updates the
     * (content) element value. Widgets can override this method to update the element
     * appropriately.
     *
     * @param {string} displayValue of the new string to be displayed
     *
     * @memberof oj.ojSlider
     * @instance
     * @protected
     */
    _SetDisplayValue: function (displayValue) {
      this._sliderDisplayValue = displayValue;
    },

    /**
     * Returns the display value that is ready to be passed to the converter.
     *
     * @return {string} usually a string display value
     *
     * @memberof oj.ojSlider
     * @instance
     * @protected
     */
    _GetDisplayValue: function () {
      return this._sliderDisplayValue;
    },

    _getElementId: function () {

      if (this.OuterWrapper) {
        if (!this._elementWrapped[0].id)
          this._elementWrapped.uniqueId()
        return (this._elementWrapped[0].id);
      }
      else {
        if (!this.element[0].id)
          this.element.uniqueId();
        return (this.element[0].id);
      }
    },

    //
    // Return the id of the slider thumb at index.
    //
    _getThumbId: function (index) {
      var elementId = this._getElementId();
      var thumbId = elementId + '-thumb' + index;
      return thumbId;
    },

    //
    // Return the id of the slider bar value.
    //
    _getBarValueId: function () {
      var elementId = this._getElementId();
      var thumbId = elementId + '-barValue';
      return thumbId;
    },

    //
    // Return the id of the slider bar.
    //
    _getBarBackgroundId: function () {
      var elementId = this._getElementId();
      var thumbId = elementId + '-barBack';
      return thumbId;
    },

    //
    // Return the id of the slider bar.
    //
    _getSliderWrapperId: function () {
      var elementId = this._getElementId();
      var sliderWrapperId = elementId + '-sliderWrapper';
      return sliderWrapperId;
    },

    _createThumbs: function () {
      var i;
      var thumbCount;
      var ariaMin = "aria-valuemin = '" + this._valueMin() + "' ";
      var ariaMax = "aria-valuemax = '" + this._valueMax() + "' ";
      var thumb = '';
      var thumbSpanStart = '<span ';
      var thumbClasses = "class='oj-slider-thumb ui-state-default' tabindex='0' role='slider'"
          + ariaMin + ariaMax + '></span>';
      var thumbs = [];

      if (this._multipleThumbs) {
        thumbCount = 2; // limit number of thumbs to 2.
      } else {
        thumbCount = 1;
      }

      //
      // Assign each thumb a unique id based on the elementId and the thumb number.
      //
      for (i = 0; i < thumbCount; i++) {
        var thumbId = "id='" + this._getThumbId(i) + "' ";
        thumb = thumbSpanStart + thumbId + thumbClasses;
        thumbs.push(thumb);
      }

      this._thumbs = $(thumbs.join('')).appendTo(this._sliderContainer);  // @HTMLUpdateOK
      this._thumb = this._thumbs.eq(0);

      this._refreshThumbOptions();
    },

    // 
    // reapply disabled properties to each slider thumb.
    // 
    _refreshThumbOptions: function () {
      var that = this;
      this._thumbs.each(function () {
        $(this).data('oj-slider-thumb-index');

        if (that._isVertical()) {
          $(this).attr('aria-orientation', 'vertical');
        }

        if (that.options.disabled) {
          $(this).attr('aria-disabled', 'true');
          $(this).removeAttr('tabindex');
        } else {
          $(this).removeAttr('aria-disabled');
        }
        // To support read only, we place set title = "read only" on the thumb.
        if (that.options.readOnly) {
          $(this).attr('title', 'read only');
          // $(this).removeAttr("tabindex")
        } else {
          $(this).removeAttr('title');
        }

      });
    },

    //
    // Create a containing div to group all component generated content.
    // This is used in messaging, so that we can apply margins/padding
    // between the inline message div and the pixels that make up the slider.
    //
    _createSliderContainer: function () {
      var sliderWrapperId = this._getSliderWrapperId();
      var existingSliderWrapper = this._elementWrapped.find('#' + sliderWrapperId);

      if (existingSliderWrapper.length) existingSliderWrapper.remove();

      this._sliderContainer = $('<div></div>');
      $(this._sliderContainer).attr('id', sliderWrapperId);
      this._sliderContainer.addClass('oj-slider-container');

      this.element.after(this._sliderContainer); // @HTMLUpdateOK

      // copy all style properties from the input tag to the enclosing
      // slider div (sliderWrapper)
      this._sliderContainer[0].style.cssText = this._styleFromInputTag;
    },

    _createBarBackground: function () {
      var barId = this._getBarBackgroundId();

      var existingBarBack = this._elementWrapped.find('#' + barId);

      if (existingBarBack.length) existingBarBack.remove();

      this._barback = $('<div></div>');

      var classes = 'oj-slider-bar';

      $(this._barback).attr('id', barId);
      this._barback.addClass(classes);

      // Place the background bar element immediately after the hidden input tab.
      this._sliderContainer.append(this._barback);  // @HTMLUpdateOK

      //
      // Clicking on the bar repositions the thumb.
      //
      var that = this;
      this._barback.on('mousedown' + that.eventNamespace, function (event) {
        that._repositionThumb(event);
        that._mouseStop(event);

        var thumb = that._getActiveThumb();
        thumb.focus();
      });
    },

    //
    // Set the options.value to the correct type, based on whether the slider has
    // been set to a range slider (_multipleThumbs is true), or a single-thumbed slider.
    //
    // Handle cases when the values option was never set, or when the values
    // option was incorrectly set.
    //
    _buildValueOption: function () {
      var options = this.options;

      if (options.type) {
        if (this.options.value == null) {
          //
          // If the value options was never set,
          // then initialize the value using valueMin (for a single-thumbbed slider)
          // or [valueMin, valueMax] (for a two-thumbbed slider).
          //
          if (this._multipleThumbs) {
            this.options.value = [this._valueMin(), this._valueMax()];
          } else {
            this.options.value = this._valueMin();
          }

          this.option('value',
                      this.options.value,
                      { '_context': { writeback: true, internalSet: true } });
        } else if (this._multipleThumbs) {
          if (this.options.value.length !== 2) {
            //
            // Transform to an array of two values if the user did not supply
            // the correct # of array values.
            //
            var firstValue;
            if (this.options.value.length > 0) firstValue = this.options.value[0];
            else firstValue = this._valueMin();

            this.options.value = [firstValue, this._valueMax()];
            this.option('value',
                        this.options.value,
                        { '_context': { writeback: true, internalSet: true } });
          }
        }
      }
    },

    _createRange: function () {
      var options = this.options;
      var classes = '';

      if (options.type) {
        //
        // Define the range (value bar) div
        //
        this._range = $('<div></div>');
        // Give the bar an id.
        $(this._range).attr('id', this._getBarValueId());

        this._sliderContainer.append(this._range);  // @HTMLUpdateOK
        classes = 'oj-slider-range oj-slider-bar-value';

        //
        // Like the bar background, clicking on the bar value also repositions the thumb.
        //
        var that = this;
        this._range.on('mousedown' + that.eventNamespace, function (event) {
          that._repositionThumb(event);
          that._mouseStop(event);

          var thumb = that._getActiveThumb();
          // if we add oj-active here it "sticks"
          // thumb.addClass("oj-active").focus();
          thumb.focus();
        });

        this._range = this._sliderContainer.find('#' + this._getBarValueId());

        var newClass = '';
        if (options.type === 'fromMin') newClass = ' oj-slider-range-min';
        else if (options.type === 'fromMax') newClass = ' oj-slider-range-max';

        this._range.addClass(classes + newClass);
      } else {
        if (this._range) {
          this._range.remove();
        }
        this._range = null;
      }
    },

    _setupTouch: function (e) {
      this._touchProxy = oj._TouchProxy.addTouchListeners(e);
    },

    _tearDownTouch: function (e) {
      oj._TouchProxy.removeTouchListeners(e);
    },

    /**
     * Setup events for slider.
     *
     * @protected
     * @memberof oj.ojSlider
     * @instance
     */
    _setupEvents: function () {
      if (this._CanSetValue()) {
        this._AddHoverable(this._elementWrapped);
      }

      this._thumbs.toArray().forEach(

        function (current) {
          var thumb = $(current);

          // setup keyboard events on each thumb.
          this._UnregisterChildNode(thumb);
          this._on(thumb, this._thumbEvents);

          // setup touch events on each thumb
          this._setupTouch(thumb);

          // Each thumb can be focusable.
          this._focusable({
            'element': thumb,
            'applyHighlight': true
          });

          // We should double-check the need for hoverable on thumbs
          // once there is a more consistent cross-component story for hoverable
          // this._hoverable(thumb);
        },
        this
      );
    },

    // This call is necessary in order to implement popup messaging properly.

    /**
     * Returns a jquery object of the launcher element representing the content nodes (slider).
     * @protected
     * @override
     * @memberof oj.ojSlider
     */
    _GetMessagingLauncherElement: function () {
      return this._elementWrapped;
    },

    /**
     * Returns a jquery object of the elements representing the content nodes (slider thumb).
     * @protected
     * @override
     * @memberof oj.ojSlider
     */
    _GetContentElement: function () {
      return this._getActiveThumb();
    },
 
    //
    // Destroy the slider DOM.
    // This is called both by _destroy and during an option change.
    // During option change, we do not unwrap - since we wish to maintain any
    // divs added for messaging, such as oj-messaging-inline-container.
    //
    _destroySliderDom: function () {
      // Tear down touch events for each thumb.
      this._thumbs.toArray().forEach(
        function (current) {
          var thumb = $(current);
          this._tearDownTouch(thumb);
        },
        this
      );

      this._destroyDraggable();

      if (this._range) this._range.remove();
      if (this._sliderContainer) this._sliderContainer.remove();

      if (this.OuterWrapper) {
        this._elementWrapped.removeUniqueId()
        this._RemoveHoverable(this._elementWrapped);
      } else  {
        this.element.removeUniqueId();
        this._RemoveHoverable(this.element);
      }
    },

    //
    // Unwrap the slider.
    // This is only called when we completely destroy the slider (_destroy).
    //
    _unwrapSlider: function () {
      oj.DomUtils.unwrap(this.element, this._elementWrapped); // @HTMLUpdateOK
      this.element.css('display', this._inputElementOriginalDisplay);
      this._RestoreAttributes(this.element);
    },

    /**
     * Override of protected base class method.
     * Method name needn't be quoted since is in externs.js.
     * @protected
     * @memberof oj.ojSlider
     * @instance
     */
    _destroy: function () {
      this._destroySliderDom();
      this._unwrapSlider();

      return this._super();
    },

    //
    // Called when the the user clicks on the bar in order to reposition the thumb.
    // Setup initial positions, distance.
    // The mouse position is used for bar clicks,
    // while the thumb position is used when dragging the thumb.
    //
    // Do not process mouse events if the slider is disabled (or readOnly).
    //
    _repositionThumb: function (event) {
      var position;
      var normValue;
      var distance;
      var o = this.options;
      var index = 0;
      var that = this;

      this._closestThumb = this._thumb;

      if (o.disabled) return false;
      if (o.readOnly) return false;

      //
      // Reposition, since when we clicked on a bar.
      //
      position = { x: event.pageX, y: event.pageY };
      normValue = this._getNormValueFromMouse(position);

      distance = this._valueMax() - this._valueMin() + 1;

      if (this._multipleThumbs) {
        this._thumbs.each(function (i) {
          var thisDistance = Math.abs(normValue - that._getMultiValues(i));
          if ((distance > thisDistance) ||
              (distance === thisDistance &&
               (i === that._lastChangedValueIndex ||
                that._getMultiValues(i) === o.min))) {
            distance = thisDistance;
            this._closestThumb = $(this);
            index = i;
          }
        });
      }

      this._thumbIndex = index;
      if (!this._closestThumb) return true;

      // This call is needed to support 'click-to-reposition' the thumb
      if (!this._thumbs.hasClass('ui-state-hover')) {
        this._slide(event, index, normValue);
      }

      var thumb = this._getActiveThumb();
      thumb.addClass('oj-active').focus();
      // For mobile theming, we need to change the color of the value bar when active.
      this._range.addClass('oj-active');

      return true;
    },

    //
    // Called by draggable start.
    // Ad the oj-active classes, place thumb in focus.
    //
    _initDragging: function (event, thumb) {
      var o = this.options;

      if (o.disabled) return false;
      if (o.readOnly) return false;

      // tabbing could have added oj-focus-highlight to the thumb,
      // if so, remove the class since we are moving the thumb via mouse interaction.
      thumb.removeClass('oj-focus-highlight');
      thumb.addClass('oj-active').focus();
      // For mobile theming, we need to change the color of the value bar when active.
      this._range.addClass('oj-active');

      return true;
    },

    _mouseDragInternal: function (event, thumb) {
      // Mirror the mouse drag with a pct change.

      //
      // Raw value update.
      //
      var normValue = this._getNormValueFromThumb(thumb);
      this._slide(event, this._thumbIndex, normValue, true);

      var pct = this._getFracFromThumb(thumb) * 100;

      if (this._multipleThumbs) this._setRangeMultiThumb(pct, this._thumbIndex);
      else this._setRange(pct);

      return false;
    },

    _mouseStop: function (event, thumb) {
      this._thumbs.removeClass('oj-active');
      this._range.removeClass('oj-active');

      var normValue = this._getNormValueFromThumb(thumb);
      this._slide(event, this._thumbIndex, normValue);
      // _change is needed for click positioning
      this._change(event, this._thumbIndex, false);

      this._thumbIndex = null;

      return false;
    },

    _isVertical: function () {
      return (this.options.orientation === 'vertical');
    },
    //
    // Adjust the fraction for bounds limits and orientation.
    //
    _getOrientationAdjustedFrac: function (frac) {
      var fracReturn = frac;
      if (fracReturn > 1) {
        fracReturn = 1;
      }
      if (fracReturn < 0) {
        fracReturn = 0;
      }
      if (this._isVertical()) {
        fracReturn = 1 - fracReturn;
      }

      return fracReturn;
    },

    //
    // Return a normalized value (trimmed to step increments)
    // based on the passed mouse coordinates.
    //
    _getNormValueFromMouse: function (position) {
      var valueTotal;
      var valueMouse;

      var fracMouse = this._getFracFromMouse(position);

      valueTotal = this._valueMax() - this._valueMin();

      if (this._isRTL() && !this._isVertical()) {
        fracMouse = 1 - fracMouse;
      }

      valueMouse = this._valueMin() + fracMouse * valueTotal;

      return this._trimAlignValue(valueMouse);
    },

    //
    // Return the fraction (between 0 and 1)
    // that represents the bar value.
    // This is based on the mouse position parameter.
    //
    _getFracFromMouse: function (position) {
      var pixelTotal;
      var pixelMouse;
      var fracMouse;

      if (!this._isVertical()) {
        pixelTotal = this._barback.width();
        pixelMouse = position.x - this._barback.offset().left;
      } else {
        pixelTotal = this._barback.height();
        pixelMouse = position.y - this._barback.offset().top;
      }

      if (pixelTotal === 0) return 1;

      fracMouse = (pixelMouse / pixelTotal);
      fracMouse = this._getOrientationAdjustedFrac(fracMouse);

      return fracMouse;
    },

    // Return the active thumb
    _getActiveThumb: function () {
      if (this._multipleThumbs) {
        return ($(this._thumbs[this._thumbIndex]));
      }
      return (this._thumb);
    },

    //
    // Return the fraction (between 0 and 1)
    // that represents the bar value.
    // This is based on the current position of the thumb.
    //
    _getFracFromThumb: function (thumbParam) {
      var pixelTotal;
      var pixelMouse;
      var fracThumb;

      var thumb = thumbParam;
      if (!thumbParam) {
        thumb = this._getActiveThumb();
      }

      var pos;

      if (!this._isVertical()) {
        var halfThumbWidth = thumb.outerWidth() / 2;
        pos = thumb.offset().left + halfThumbWidth;
        pixelTotal = this._barback.width();
        pixelMouse = pos - this._barback.offset().left;
      } else {
        var halfThumbHeight = thumb.outerHeight() / 2;
        pos = thumb.offset().top + halfThumbHeight;
        pixelTotal = this._barback.height();
        pixelMouse = pos - this._barback.offset().top;
      }

      if (pixelTotal === 0) return 1;

      fracThumb = (pixelMouse / pixelTotal);

      fracThumb = this._getOrientationAdjustedFrac(fracThumb);

      return fracThumb;
    },

    _getNormValueFromThumb: function (thumb) {
      var fracThumb;
      var valueTotal;
      var valueMouse;

      fracThumb = this._getFracFromThumb(thumb);

      valueTotal = this._valueMax() - this._valueMin();

      if (this._isRTL() && !this._isVertical()) {
        fracThumb = 1 - fracThumb;
      }

      valueMouse = this._valueMin() + fracThumb * valueTotal;

      var trimmedValue = this._trimAlignValue(valueMouse);

      return trimmedValue;
    },

    // Return the value for the inactive thumb.
    _getOtherThumbValue: function (index) {
      return (this._getMultiValues(index ? 0 : 1));
    },

    //
    // Return the new value, limited by the value of the other thumb.
    // (We ensure that we do not go past the value of the other thumb).
    //
    _getNewThumbValueLimited: function (index, newVal, otherVal) {
      if ((this.options.value.length === 2) &&
          ((index === 0 && newVal > otherVal) || (index === 1 && newVal < otherVal))
         ) {
        return (otherVal);
      }
      return (newVal);
    },

    _slide: function (event, index, newValParam, rawOnly) {

      var otherVal;

      if (this._multipleThumbs) {
        otherVal = this._getOtherThumbValue(index);
        var newVal = this._getNewThumbValueLimited(index, newValParam, otherVal);

        if (newVal !== this._getMultiValues(index)) {
          this._setMultiValue(event, index, newVal, rawOnly);
        }
      } else {
        // This case handles a single value
        if (newValParam !== this._getSingleValue()) {
          // sets slider thumb value
          this._setSingleValue(event, newValParam, rawOnly);

          if (!rawOnly) {
            if (this._inputtag) {
              this.element.val(newValParam);
            }
          }
        }
      }
    },

    _setSingleValue: function (event, newValue, rawOnly) {
      this._newValue = this._trimAlignValue(newValue);
      this._SetRawValue(this._newValue, event);
      if (!rawOnly) {
        this.options[this._transientValueName] = this._newValue;
        this._SetValue(this._newValue, event);
        this._updateUI();
      }

      return;
    },

    _change: function (event, index, rawOnly) {

      if (this._multipleThumbs) {
        // store the last change values for creating draggable containment
        this._lastChangedValues = this._getNewValues(index, this._newMultiValue[index]);
          this._SetRawValue(this._lastChangedValues, event);
        if (!rawOnly) {
          this._SetValue(this._lastChangedValues, event);
        }
      } else {
        this._SetRawValue(this._newValue, event);
        if (!rawOnly) {
          this._SetValue(this._newValue, event);
        }
      }

      // store the last changed value index for reference when thumbs overlap
      this._lastChangedValueIndex = index;
    },

    //
    // Return options.values with the new value copied in the array
    // Used to format the values argument to SetValue (for the case of multiple thumbs)
    //
    _getNewValues: function (index, newValue) {
      var vals;
      var i;

      vals = this.options.value.slice();

      for (i = 0; i < vals.length; i += 1) {
        vals[i] = this._trimAlignValue(vals[i]);
      }

      // assume newValue is trim aligned
      // Assign only if it is the thumb that is actually sliding
      if (index === this._thumbIndex) {
        vals[index] = newValue;
      }

      return vals;
    },

    //
    // Return the value of a single thumbbed slider.
    //
    _getSingleValue: function () {
      return this._getValueAligned();
    },

    //
    // Return the value for the specified thumb.
    //
    _getMultiValues: function (index) {
      //
      // Parameter index is supplied.
      // Return the value for the specified thumb.
      //
      return this._getValuesAligned(index);
    },

    //
    // Internal setter for slider values.
    // Sets the value for the specifed thumb.
    // (index 0 is used for a single thumb,
    // for a range slider index 0 and index 1 are used).
    //
    _setMultiValue: function (event, index, newValue, rawOnly) {
      this._newMultiValue[index] = this._trimAlignValue(newValue);
      this._change(event, index, rawOnly);
      if (!rawOnly) this._updateUI();
      return;
    },

    _setOption: function (key, value, flags) {
      var coercedValue;

      if (key === 'value') {
        if (Array.isArray(value)) {
          if (!isNaN(value[0])) {
            this._multipleThumbs = true;
            coercedValue = value;
            // verify that the array values are all within range.
            for (var index = 0; index < coercedValue.length; index += 1) {
              this._checkValueBounds(coercedValue[index], this._valueMin(), this._valueMax());
            }
          } else {
            //
            // Don't set multipleThumbs if the value is not a number
            // (as would be the case for an error code)
            //
            this._multipleThumbs = false;
            coercedValue = this._parse(key, value[0]);
            // verify that the new value is within range.
            this._checkValueBounds(coercedValue, this._valueMin(), this._valueMax());
          }
        } else {
          this._multipleThumbs = false;
          // Only coerce values for widget syntax,
          if (!this._IsCustomElement()) {
            coercedValue = this._parse(key, value);
          } else {
            coercedValue = value;
          }
          // verify that the new value is within range.
          this._checkValueBounds(coercedValue, this._valueMin(), this._valueMax());
        }
      }

      if (key === 'max' || key === 'min') {

        // Only coerce values for widget syntax,
        // since the framework handles this for custom elements.
        if (!this._IsCustomElement()) {
          coercedValue = this._parse(key, value);
        } else {
          coercedValue = value;
        }

        //
        // Note that coerce the value to be within the 
        // min and max when we option change the min or max.
        //
        if (key === 'min') {
          this._checkMinMax(coercedValue, this._valueMax());
          // Coerce any out of bounds "value" attributes to be within the min bounds
          if (!this._multipleThumbs) {
            if (this._getSingleValue() < coercedValue) {
              this._super('value', coercedValue, flags);
            }
          } else {
            if (this._getMultiValues(0) < coercedValue) {
              this._super('value', [coercedValue, this._getMultiValues(1)], flags);
            }
            if (this._getMultiValues(1) < coercedValue) {
              this._super('value', [this._getMultiValues(0), coercedValue], flags);
            }
          }
        } else if (key === 'max') {
          this._checkMinMax(this._valueMin(), coercedValue);
          // Coerce any out of bounds "value" attributes to be within the max bounds
          if (!this._multipleThumbs) {
            if (this._getSingleValue() > coercedValue) {
              this._super('value', coercedValue, flags);
            }
          } else {
            if (this._getMultiValues(0) > coercedValue) {
              this._super('value', [coercedValue, this._getMultiValues(1)], flags);
            }
            if (this._getMultiValues(1) > coercedValue) {
              this._super('value', [this._getMultiValues(0), coercedValue], flags);
            }
          }
        }
      } else if (key === 'step') {
        coercedValue = this._parseStep(value);
      } else {
        coercedValue = value;
      }

      this._super(key, coercedValue, flags);
      // when a dom element supports readonly, use that, and not aria-readonly.
      // having both is an error

      // Note - for now, readOnly is private.
      if (key === 'readOnly') {
        this.options.readonly = coercedValue;
      }
      if (key === 'disabled') {
        this.options.disabled = coercedValue;
      }

      switch (key) {

      case 'disabled':
        this._refreshThumbOptions();
        // disable the dragging if disabled is true.
        if (this.options.disabled) {
          this._disableDraggable();
        } else {
          this._makeDraggable();
        }
        break;

      case 'value':
        this._updateUI();
        this._makeDraggable();
        break;

      case 'min':
      case 'max':

        this._calculateNewMax();
        this._updateUI();
        this._makeDraggable();
        break;

      case 'orientation':
      case 'readonly':
      case 'step':
      case 'type':
        this._reCreate();
        break;

      default:
        break;
      }
    },

    //
    // Recreate the slider.
    // Destroy's everything except the wrapper.
    // Called after optionChange.
    //
    _reCreate: function () {
      this._destroySliderDom();
      this._componentSetup();
      if (this.OuterWrapper) {
        // When we programmatically recreate a custom element slider
        // (in order to implemement option change), we assert oj-complete.
        this._elementWrapped.addClass('oj-complete');
      }
      this._AfterCreate();
    },

    // internal value getter
    // _getValueAligned() returns value trimmed by min and max, aligned by step
    _getValueAligned: function () {
      var val = this.options.value;
      val = this._trimAlignValue(val);

      return val;
    },

    //
    // Internal values getter
    //
    // _getValuesAligned() returns array of values trimmed by min and max, aligned by step
    // _getValuesAligned( index ) returns single value trimmed by min and max, aligned by step
    //
    _getValuesAligned: function (index) {
      var val = this._trimAlignValue(this.options.value[index]);
      return val;
    },

    //
    // Return the step-aligned value that val is closest to, between (inclusive) min and max
    //
    _trimAlignValue: function (val) {
      if (val <= this._valueMin()) {
        return this._valueMin();
      }
      if (val >= this._valueMax()) {
        return this._valueMax();
      }
      var step = (this.options.step > 0) ? this.options.step : 1;
      var valModStep = (val - this._valueMin()) % step;
      var alignValue = val - valModStep;

      if (Math.abs(valModStep) * 2 >= step) {
        alignValue += (valModStep > 0) ? step : (-step);
      }

      // Since JavaScript has problems with large floats, round
      // the final value to 5 digits after the decimal point (see #4124)
      return parseFloat(alignValue.toFixed(5));
    },

    _calculateNewMax: function () {
      var min = this._valueMin();
      if ((((this.options.max - min) / this.options.step) % 1) !== 0) {
          var remainder = (this.options.max - min) % this.options.step;
          this.max = (this.options.max - remainder) + this.options.step;
      } else {
        this.max = this.options.max;
      }
    },

    _valueMin: function () {
      return this.options.min;
    },

    _valueMax: function () {
      return this.max;
    },

    //
    // Calculate the grid size passed to draggable.
    // This implements the drag to step increments.
    // For a horizontal grid, we would return an array of [pixelInterval, 1],
    // while a vertical grid returns [1, pixelInterval].
    //
    _getGrid: function () {
      var numIntervals;
      if (this.options.step > 0) {
        numIntervals = (this._valueMax() - this._valueMin()) / this.options.step;
      } else {
        numIntervals = 100; // this case should not occur.
      }

      var pixelTotal;

      if (!this._isVertical()) {
        pixelTotal = this._barback.width();
      } else {
        pixelTotal = this._barback.height();
      }

      var pixelInterval = pixelTotal / numIntervals;

      if (pixelInterval < 1) pixelInterval = 1;

      if (!this._isVertical()) {
        return [pixelInterval, 1];
      }
      return [1, pixelInterval];
    },

    _getThumbsValueFrac: function (index) {
      return ((this._getMultiValues(index) - this._valueMin())
              / (this._valueMax() - this._valueMin()));
      // Note - (max - min) is checked in options to make sure that
      // it is non-zero and positive
      // Note - we always use aligned values.
    },


    //
    // Update the UI, reflecting the value.
    //
    _updateUI: function () {
      var valPercent;
      var value;
      var valueMin;
      var valueMax;

      //
      // Multiple thumbs case.
      //
      if (this._multipleThumbs) {
        this._thumbs.toArray().forEach(

          function (current, i) {
            var thumb = $(current);
            valPercent = this._getThumbsValueFrac(i) * 100;

            if (this._isRTL() && !this._isVertical()) valPercent = 100 - valPercent;

            if (!this._isVertical()) {
              thumb['css']({ left: valPercent + '%' });
            } else {
              thumb['css']({ top: (100 - valPercent) + '%' });
            }

            if (!thumb.hasClass('oj-active')) {
              thumb.attr('aria-valuenow', this._getMultiValues(i));
              thumb.attr('aria-valuemin', valueMin);
              thumb.attr('aria-valuemax', valueMax);
            }
            this._setRangeMultiThumb(valPercent, i);
          },
          this
        );
      } else {
        //
        // Scalar value (single thumb)
        //

        // We always want an aligned value here.
        value = this._getValueAligned();

        valueMin = this._valueMin();
        valueMax = this._valueMax();
        valPercent = (valueMax !== valueMin) ?
          (value - valueMin) / (valueMax - valueMin) * 100 : 0;

        if (this._isRTL() && !this._isVertical()) valPercent = 100 - valPercent;

        if (!this._isVertical()) {
          this._thumb['css']({ left: valPercent + '%' });
        } else {
          this._thumb['css']({ top: (100 - valPercent) + '%' });
        }

        //
        // note - we don't want to continuously update aria values,
        // otherwise it causes unwanted screen reader chatter.
        // see .
        //
        if (!$(this._thumb).hasClass('oj-active')) {
          $(this._thumb).attr('aria-valuenow', value);
          $(this._thumb).attr('aria-valuemin', valueMin);
          $(this._thumb).attr('aria-valuemax', valueMax);
        }
        this._setRange(valPercent);
      }
    },

    // Set the range (bar value)
    _setRange: function (val) {

      var oRange = this.options.type;

      if (!this._isVertical()) {
        if (!this._isRTL()) {
          if (oRange === 'fromMin') {
            this._range['css']({ width: val + '%' });
          }
          if (oRange === 'fromMax') {
            this._range['css']({ width: (100 - val) + '%' });
          }
        } else {
          if (oRange === 'fromMin') {
            this._range['css']({ width: (100 - val) + '%' });
          }
          if (oRange === 'fromMax') {
            this._range['css']({ width: val + '%' });
          }
        }
      } else {
        if (oRange === 'fromMin') {
          this._range['css']({ height: val + '%' });
        }
        if (oRange === 'fromMax') {
          this._range['css']({ height: (100 - val) + '%' });
        }
      }
    },

    //
    // set the range for a multi-thumb (range) slider
    //
    _setRangeMultiThumb: function (val, index) {
      var id = this._range.attr('id');

      if (index === 0) {
        var thumb1Pct = this._getThumbsValueFrac(1) * 100;

        switch (this.options.type) {

          case 'fromMin':

            if (!this._isVertical()) {
              this._range['css']({ width: val + '%' });
            } else {
              this._range['css']({ height: val + '%' });
            }
            break;

          case 'range':

            if (!this._isVertical()) {
              if (!this._isRTL()) {
                this._range['css']({ left: val + '%' });
                this._range['css']({ width: (thumb1Pct - val) + '%' });
              } else {
                this._range['css']({ left: (100 - thumb1Pct) + '%' });
                this._range['css']({ width: (thumb1Pct - (100 - val)) + '%' });
              }
            } else {
              this._range['css']({ top: (100 - thumb1Pct) + '%' });
              this._range['css']({ height: (thumb1Pct - val) + '%' });
            }
            break;

          default:
            break;
        }
      } else {
        var thumb0Pct = this._getThumbsValueFrac(0) * 100;

        switch (this.options.type) {

          case 'fromMax':
            if (!this._isVertical()) {
              this._range['css']({ width: (100 - val) + '%' });
            } else {
              this._range['css']({ height: (100 - val) + '%' });
            }
            break;

          case 'range':

            if (!this._isVertical()) {
              if (!this._isRTL()) {
                if (document.getElementById(id)) {
                  var barLeft = parseInt(document.getElementById(id).style.left, 10);
                  this._range['css']({ width: (val - barLeft) + '%' });
                }
              } else {
                if (document.getElementById(id)) {
                  this._range['css']({ left: val + '%' });
                  this._range['css']({ width: (- val + 100 - thumb0Pct) + '%' });
                }
              }
            } else {
              if (document.getElementById(id)) {
                this._range['css']({ top: (100 - val) + '%' });
                this._range['css']({ height: (val - thumb0Pct) + '%' });
              }
            }

            break;
          default:
            break;

        }
      }
    },

    _thumbEvents: {
      keydown: function (event) {
        var curVal;
        var newVal;
        var step;
        var tempVal;
        var index = $(event.target).data('oj-slider-thumb-index');

        this._thumbIndex = index;

        switch (event.keyCode) {
          case $.ui.keyCode.HOME:
          case $.ui.keyCode.END:
          case $.ui.keyCode.PAGE_UP:
          case $.ui.keyCode.PAGE_DOWN:
          case $.ui.keyCode.UP:
          case $.ui.keyCode.RIGHT:
          case $.ui.keyCode.DOWN:
          case $.ui.keyCode.LEFT:
            event.preventDefault();
            //
            // Note - while tabbing into the thumb will retain focus,
            // when clicking on the thumb we do not need to call focus()
            // (per Max's review)
            //
            $(event.target).addClass('oj-active');
            break;
          default:
            break;

        }

        step = this.options.step;

        if (this._multipleThumbs) {
          curVal = newVal = this._getMultiValues(index);
        } else {
          curVal = newVal = this._getSingleValue();
        }

        switch (event.keyCode) {
          case $.ui.keyCode.HOME:
            newVal = this._valueMin();
            break;
          case $.ui.keyCode.END:
            newVal = this._valueMax();
            break;
          case $.ui.keyCode.PAGE_UP:
            newVal = this._trimAlignValue(
              curVal + ((this._valueMax() - this._valueMin()) / this._numPages)
            );
            break;
          case $.ui.keyCode.PAGE_DOWN:
            newVal = this._trimAlignValue(
              curVal - ((this._valueMax() - this._valueMin()) / this._numPages));
            break;

          case $.ui.keyCode.UP:

            // upArrow always increments the value
            if (curVal === this._valueMax()) return;
            tempVal = curVal + step;
            newVal = this._trimAlignValue(tempVal);
            break;

          case $.ui.keyCode.RIGHT:

            if (!this._isRTL() || this._isVertical()) {
              if (curVal === this._valueMax()) return;
              tempVal = curVal + step;
            } else {
              if (curVal === this._valueMin()) return;
              tempVal = curVal - step;
            }

            newVal = this._trimAlignValue(tempVal);
            break;

          case $.ui.keyCode.DOWN:

            // Down arrow always decrements the value.
            if (curVal === this._valueMin()) return;
            tempVal = curVal - step;
            newVal = this._trimAlignValue(tempVal);

            break;

          case $.ui.keyCode.LEFT:

            if (!this._isRTL() || this._isVertical()) {
              if (curVal === this._valueMin()) return;
              tempVal = curVal - step;
            } else {
              if (curVal === this._valueMax()) return;
              tempVal = curVal + step;
            }

            newVal = this._trimAlignValue(tempVal);
            break;
          default:
            break;

        }

        this._slide(event, index, newVal);
      },

      keyup: function (event) {
        switch (event.keyCode) {
          case $.ui.keyCode.HOME:
          case $.ui.keyCode.END:
          case $.ui.keyCode.PAGE_UP:
          case $.ui.keyCode.PAGE_DOWN:
          case $.ui.keyCode.UP:
          case $.ui.keyCode.RIGHT:
          case $.ui.keyCode.DOWN:
          case $.ui.keyCode.LEFT:

            var index = $(event.target).data('oj-slider-thumb-index');
            this._thumbIndex = index;

            this._change(event, index, false);

            $(event.target).removeClass('oj-active');
            this._updateUI(true);

            this._thumbIndex = null;
            break;
          default:
            break;

        }
      }
    },

    // P R O T E C T E D    C O N S T A N T S   A N D   M E T H O D S

    // ***** START WIDGET FACTORY METHODS (they retain _camelcase naming convention) ******

    /**
     * Called at component create time primarily to initialize options, often using DOM values.
     * This method is called before _ComponentCreate is called, so components that override this
     * method should be aware that the component has not been rendered yet. The element DOM is
     * available and can be relied on to retrieve any default values. <p>
     * @param {!Object} originalDefaults - original default options defined on the widget and
     * its ancestors.
     * @param {?Object} constructorOptions - options passed into the widget constructor
     *
     * @memberof oj.ojSlider
     * @instance
     * @protected
     */
    _InitOptions: function (originalDefaults, constructorOptions) {
      var opts = this.options;
      var self = this;

      this._superApply(arguments);
      var props =
          [{ attribute: 'disabled', validateOption: true },
           //                 {attribute: "placeholder"},
           { attribute: 'value' },

           //
           // Once slider supports read-only, uncomment the following line.
           // {attribute: "readonly", option: "readOnly", validateOption: true},
           //

           { attribute: 'title' },
           { attribute: 'min' },
           { attribute: 'max' },
           { attribute: 'step' }];


      oj.EditableValueUtils.initializeOptionsFromDom(
        props, constructorOptions, this,
        // post-process callback
        function (initializedOptions) {
          // coerce regardless of where the option value came from - dom/constructor
          var toParse = ['value', 'step', 'min', 'max'];

          for (var i = 0; i < toParse.length; i++) {
            var opt = toParse[i];
            var value = (opt in initializedOptions) ?
                initializedOptions[opt] : opts[opt];
            if (value != null) {
              if (opt === 'step') {
                initializedOptions[opt] = self._parseStep(value);
              } else if (opt === 'min' || opt === 'max') {
                initializedOptions[opt] = self._parse(opt, value);
              } else if (opt === 'value') {
                if (Array.isArray(value)) {
                  initializedOptions[opt] = value;
                } else {
                  initializedOptions[opt] = self._parse(opt, value);
                }
              }
            }
          }
        }
      );

      if (!this._IsCustomElement()) {
        if (opts['value'] === undefined) {
          throw new Error(this.getTranslatedString('noValue'));
        }
      }

      this._checkMinMax(opts['min'], opts['max']);

      // Make sure value is within min and max
      if (Array.isArray(opts['value'])) {
        for (var index = 0; index < opts['value'].length; index += 1) {
          this._checkValueBounds(opts['value'][index], opts['min'], opts['max']);
        }
      } else {
        this._checkValueBounds(opts['value'], opts['min'], opts['max']);
      }
    },

    // function that will throw an error if the value is not between min and max
    _checkValueBounds: function(value, min, max) {
      if (min != null) {
        if (value < min) {
          throw new Error(this.getTranslatedString('valueRange'));
        }
      }
      if (max != null) {
        if (value > max) {
          throw new Error(this.getTranslatedString('valueRange'));
        }
      }
    },

    // throw an error if min >= max
    _checkMinMax: function(min, max) {
      if (min != null && max != null) {
        if (min >= max) {
          throw new Error(this.getTranslatedString('maxMin'));
        }
      }
    },

    // @inheritdoc
    getNodeBySubId: function(locator) {
      if (locator == null) {
        return this.element ? this.element[0] : null;
      }

      var subId = locator['subId'];

      if (subId === 'oj-slider-thumb-0') {
        return this.widget().find('.oj-slider-thumb')[0];
      } else if (subId === 'oj-slider-thumb-1') {
        return this.widget().find('.oj-slider-thumb')[1];
      } else if (subId === 'oj-slider-bar') {
        return this.widget().find('.' + subId)[0];
      } else if (subId === 'oj-slider-bar-value') {
        return this.widget().find('.' + subId)[0];
      }

      // Non-null locators have to be handled by the component subclasses
      return null;
    },

    // @inheritdoc
    getSubIdByNode: function (node) {
      if (node != null) {
        if (node.id === this._getThumbId(0) && $(node).hasClass('oj-slider-thumb')) {
          return { 'subId': 'oj-slider-thumb-0' };
        } else if (node.id === this._getThumbId(1) && $(node).hasClass('oj-slider-thumb')) {
          return { 'subId': 'oj-slider-thumb-1' };
        } else if ($(node).hasClass('oj-slider-bar')) {
          return { 'subId': 'oj-slider-bar' };
        } else if ($(node).hasClass('oj-slider-bar-value')) {
          return { 'subId': 'oj-slider-bar-value' };
        }
      }

      return null;
    },

    // *********** END WIDGET FACTORY METHODS **********

    /**
     * Returns the default styleclass for the component.
     *
     * @return {string}
     * @memberof oj.ojSlider
     * @override
     * @protected
     */
    _GetDefaultStyleClass: function () {
      return 'oj-slider';
    },

    // The user can clear out min/max by setting the option to null, so we
    // do not coerce null.
    /**
     * @param {string} option name of the option. this will show up in the error if thrown
     * @param val value to parse
     * @throws {Error} if option value is invalid
     * @private
     */
    _parse: function (option, val) {
      var returnValue;
      if (val !== null) {
        returnValue = +val;
      } else {
        returnValue = val;
      }

      if (isNaN(returnValue)) {
        throw new Error(this.getTranslatedString('optionNum', { 'option': option }));
      }

      return returnValue;
    },
    /**
     * parse the step's value
     * We are following the behavior of HTML-5 the best we can. According
     * to the spec, it says step must be a number greater than 0.
     * Chrome defaults it to 1 if it is not.
     * @throws {Error} if option value is invalid
     * @private
     */
    _parseStep: function (val) {
      var defaultStep = 1;
      var parsedStep;
      if (val === null) {
        return defaultStep;
      }
      parsedStep = this._parse('step', val);
      if (parsedStep <= 0) {
        // throw an exception
        throw new Error(this.getTranslatedString('invalidStep'));
      }
      // DEFAULT to 1 if it isn't > 0
      if (parsedStep === null || parsedStep <= 0) {
        parsedStep = defaultStep;
      }
      return parsedStep;
    },

    /////////////////////////////////////////////////////////////////////////////////////////
    //
    // Draggable - used to implement slider thumb dragging
    //
    // JQueryUI draggable is used to implement the dragging of slider thumbs.
    // Several draggable options are used:
    //
    // - Axis constaint
    //   The thumb is constrained to movement along the x-axis (for horizontal sliders)
    //   or movement along the y-axis (for vertical siders)
    //
    // - Range constraints
    //   The thumb is constrained to ranges along the axis using the draggable constraint
    //   option.
    //
    // - Step
    //   The granularity of movement is constrained to the step size using the grid
    //   option.
    //
    /////////////////////////////////////////////////////////////////////////////////////////

    // return the endpoint of the bar
    _getEndInterval: function () {
      return (this._barback.offset().left + this._barback.width());
    },

    // return the startpoint of the bar
    _getStartInterval: function () {
      return (this._barback.offset().left);
    },


    //
    // Set up the draggable with the context, thumb, and containment parameters.
    // Use the axis method to ensure only horizontal or vertical movement.
    //
    _callDraggable: function (thumbParam) {
      var g = this._getGrid();
      var cachedStyle = thumbParam[0].style;

      var axisValue;
      if (!this._isVertical()) axisValue = 'x';
      else axisValue = 'y';

      var that = this;

      thumbParam.draggable({
        axis: axisValue,
        // grid: [8.8,1],
        grid: g,
        disabled: false,
        start: function (event) {
          //
          // Set current thumb
          //

          if (thumbParam[0] === $(that._thumbs)[0]) {
            that._thumbIndex = 0;
          } else if (thumbParam[0] === $(that._thumbs)[1]) {
            that._thumbIndex = 1;
          }

          that._initDragging(event, thumbParam);
        },

        'drag': function (event, ui) {
          //
          // Compensate for a draggable bug.
          // The bug causes the thumb to drift off axis (for some .css thumb sizes).
          // The workaround address the problem by:
          //   When the thumb travels along the x-axis (horizontal slider),
          //   null out modifications made to top.
          // Handle this issue similarly for vertical sliders.
          //
          var pos = ui.position;

          if (!that._isVertical()) {
            cachedStyle.top = '';
            pos.top = '';
          } else {
            cachedStyle.left = '';
            pos.left = '';
          }

          that._mouseDragInternal(event, thumbParam);

          //
          // Enforce constraints (don't allow sliding past the end)
          //
          if (!that._isVertical()) {
            if (pos.left < 0) {
              pos.left = 0;
            }
            if (pos.left > that._barback.width()) {
              pos.left = that._barback.width();
            }
          } else {
            if (pos.top < 0) {
              pos.top = 0;
            }
            if (pos.top > that._barback.height()) {
              pos.top = that._barback.height();
            }
          }

          //
          // For range sliders, ensure that thumbs do not cross.
          //
          if (that._multipleThumbs) {
            var otherThumb;
            // var thumb = that._getActiveThumb();

            if (that._thumbIndex === 0) {
              otherThumb = $(that._thumbs[1]);
            } else {
              otherThumb = $(that._thumbs[0]);
            }

            //
            // parentLeft ensures that the offsets are calculated properly
            // for a slider embedded in a repositioned container (popup or dialog)
            //
            var pos2;
            if (!that._isVertical()) {
              var halfThumbWidth = thumbParam.outerWidth() / 2;
              var parentLeft = that._barback.offsetParent().offset().left;
              // pos1 = thumbParam.offset().left + halfThumbWidth - parentLeft;
              pos2 = otherThumb.offset().left + halfThumbWidth - parentLeft;
            } else {
              var halfThumbHeight = thumbParam.outerHeight() / 2;
              var parentTop = that._barback.offsetParent().offset().top;
              // pos1 = thumbParam.offset().top + halfThumbHeight - parentTop;
              pos2 = otherThumb.offset().top + halfThumbHeight - parentTop;
            }

            if (that._thumbIndex === 0) {
              if (!that._isVertical()) {
                if (!that._isRTL()) {
                  if (pos.left > pos2) pos.left = pos2;
                } else {
                  if (pos.left < pos2) pos.left = pos2;
                }
              } else {
                if (pos.top < pos2) pos.top = pos2;
              }
            } else {
              if (!that._isVertical()) {
                if (!that._isRTL()) {
                  if (pos.left < pos2) pos.left = pos2;
                } else {
                  if (pos.left > pos2) pos.left = pos2;
                }
              } else {
                if (pos.top > pos2) pos.top = pos2;
              }
            }
          }
        },

        stop: function (event) {
          //
          // compensate for a firefox draggable bug.
          // without this code, thumbs with larger active sizes become oval on stop.
          //
          this.style.width = '';
          this.style.height = '';

          that._mouseStop(event, thumbParam);
        }
      });
    },

    //
    // Setup the draggable for each of the thumbs.
    //
    _makeDraggable: function () {
      // Do not allow dragging on a disabled thumb.
      if (this.options.disabled) return;

      if (this._multipleThumbs) {
        this._thumbs.toArray().forEach(

          function (current) {
            var thumb = $(current);
            this._callDraggable(thumb);
          },
          this
        );
      } else {
        this._callDraggable(this._thumb);
      }
    },

    //
    // Call this if we change option to disabled.
    //
    _disableDraggable: function () {
      if (this._multipleThumbs) {
        this._thumbs.toArray().forEach(

          function (current) {
            var thumb = $(current);
            if (thumb.is('.ui-draggable')) {
              thumb['draggable']('disable');
            }
          },
          this
        );
      } else {
        if (this._thumb.is('.ui-draggable')) {
          this._thumb.draggable('disable');
        }
      }
    },

    // 
    // Destroy the draggable that was instantiated on each slider thumb.
    // 
    _destroyDraggable: function () {
      if (this._multipleThumbs) {
        this._thumbs.toArray().forEach(

          function (current) {
            var thumb = $(current);
            if (thumb.is('.ui-draggable')) {
              thumb['draggable']('destroy');
            }
          },
          this
        );
      } else {
        if (this._thumb.is('.ui-draggable')) {
          this._thumb.draggable('destroy');
        }
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////
    // Draggable - end
    /////////////////////////////////////////////////////////////////////////////////////////

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Slider Bar</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Reposition the thumb.</td>
     *     </tr>
     *     <tr>
     *       <td>Slider Thumb</td>
     *       <td><kbd>Swipe</kbd></td>
     *       <td>Reposition the thumb.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc,
     * and standalone gesture doc
     * @memberof oj.ojSlider
     */

    /**
     * The JET slider supports keyboard actions for thumb movement:
     *
     * <p>
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *        <th>Target</th>
     *       <th>Key</th>
     *       <th>Use</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>Tab</kbd></td>
     *       <td> Places focus on the slider component.
     *        If hints, title or messages exist in a notewindow, pop up the notewindow.
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>RightArrow</kbd></td>
     *       <td>Scrolls right on a horizontal slider, scrolls up on a vertical slider.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>LeftArrow</kbd></td>
     *       <td>Scrolls left on a horizontal slider, scrolls down on a vertical slider.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>UpArrow</kbd></td>
     *       <td>Scrolls right on a horizontal slider, scrolls up on a vertical slider.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>DownArrow</kbd></td>
     *       <td>Scrolls left on a horizontal slider,
     *        scrolls down on a vertical slider.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>PageUp</kbd></td>
     *       <td>Scrolls one page right on a horizontal slider,
     *        scrolls one page up on a vertical slider. <br>
     *       A page is defined as 20% of the range of the slider.
     *     </td>
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>PageDown</kbd></td>
     *       <td>Scrolls one page left on a horizontal slider,
     *       scrolls one page down on a vertical slider.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>End</kbd></td>
     *       <td>Scrolls to the right end on a horizontal slider, scrolls to the bottom on a
     * vertical slider.
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>Slider</td>
     *       <td><kbd>Home</kbd></td>
     *       <td>Scrolls to the left end on a horizontal slider, scrolls to the top on a
     * vertical slider.
     *       </td>
     *     </tr>
     * </tbody>
     * </table>
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone
     * gesture doc
     * @memberof oj.ojSlider
     */

    /**
     * The following CSS class can be applied by the page author as needed.
     *
     * <p>
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>Class</th>
     *       <th>Description</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-focus-highlight</td>
     *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojSlider
     */

    /// ///////////////     SUB-IDS     //////////////////

    /**
     * <p>Sub-ID for the slider thumb. Use this id to access the thumb of the  slider. </p>
     *
     * @ojsubid oj-slider-thumb-0
     * @memberof oj.ojSlider
     *
     * @example <caption>Get the node for the slider thumb:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-slider-thumb-0'});
     */

    /**
     * <p>Sub-ID for the slider bar. </p>
     *
     * @ojsubid oj-slider-bar
     * @memberof oj.ojSlider
     *
     * @example <caption>Get the node for the slider bar:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-slider-bar'});
     *
     */

    /**
     * <p>Sub-ID for the slider bar value. </p>
     *
     * @ojsubid oj-slider-bar-value
     * @memberof oj.ojSlider
     *
     * @example <caption>Get the node for the slider bar value:</caption>
     * var node = myComponent.getNodeBySubId({'subId': 'oj-slider-bar-value'});
     *
     */
  });
}());

(function () {
  var ojSliderMeta = {
    'properties': {
      'max': {
        'type': 'number'
      },
      'min': {
        'type': 'number'
      },
      'orientation': {
        'type': 'string',
        'enumValues': ['horizontal', 'vertical']
      },
      'step': {
        'type': 'number'
      },
      'transientValue': {
        'type': 'number',
        'readOnly': true,
        'writeback': true
      },
      'type': {
        'type': 'string',
        'enumValues': ['single', 'fromMin', 'fromMax']
      },
      "translations": {
        "type": "Object",
        "properties": {
          "invalidStep": {
            "type": "string",
            "value": "Invalid step; step must be > 0"
          },
          "maxMin": {
            "type": "string",
            "value": "Max must not be less than min"
          },
          "noValue": {
            "type": "string",
            "value": "ojSlider has no value"
          },
          "optionNum": {
            "type": "string",
            "value": "{option} option is not a number"
          },
          "valueRange": {
            "type": "string",
            "value": "Value must be within min to max range"
          }
        }
      },
      'value': {
        'type': 'number',
        'writeback': true
      }
    },
    'methods': {},
    'extension': {
      _INNER_ELEM: 'input',
      _WIDGET_NAME: 'ojSlider'
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-slider', 'editableValue', ojSliderMeta);
  oj.CustomElementBridge.register('oj-slider', {'metadata': oj.CustomElementBridge.getMetadata('oj-slider')});
}());

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojSunburst
 * @augments oj.dvtBaseComponent
 * @since 0.7
 * @ojstatus preview
 * @ojrole application
 * @ojshortdesc An interactive data visualization in which hierarchical data is represented in concentric rings.  Each ring segment is proportionally sized relative to the other segments at a given level.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="sunburstOverview-section">
 *   JET Sunburst
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#sunburstOverview-section"></a>
 * </h3>
 *
 * <p>Sunbursts are used to display hierarchical data across two dimensions, represented by
 * the size and color of the sunburst nodes.</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-sunburst
 *    nodes='[{"value": 100, "color": "#FFFF00", "label": "Total Sales",
 *             "nodes": [{"value": 75, "color": "#00FF00", "label": "Candy"},
 *                       {"value": 20, "color": "#FFFF00", "label": "Fruit"},
 *                       {"value": 15, "color": "#FF0000", "label": "Vegetables"}]}]'>
 * &lt;/oj-sunburst>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets. Alternate visualizations should
 *    be considered if identifying data changes is important, since all nodes will generally move and resize on any data
 *    change.
 * </p>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it's recommended that applications only set usable data densities on this element.
 *    Applications can enable progressive reveal of data through drilling or aggregate small nodes to reduce the
 *    displayed data set size.
 * </p>
 * 
 * <h4>Styling</h4>
 * <p>Use the highest level property available. For example, consider setting styling properties on
 *    <code class="prettyprint">nodeDefaults</code>, instead of styling properties
 *    on the individual nodes. The sunburst can take advantage of these higher level properties to apply the style properties on
 *    containers, saving expensive DOM calls.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojSunburst', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * Specifies the animation duration in milliseconds. For data change animations with multiple stages, 
     * this attribute defines the duration of each stage. For example, if an animation contains two stages, 
     * the total duration will be two times this attribute's value. The default value comes from the CSS and varies based on theme.
     * @expose
     * @name animationDuration
     * @memberof oj.ojSunburst
     * @instance
     * @type {number}
     * @ojunits milliseconds
     */
    animationDuration: undefined,
    
    /**
     * Specifies the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDataChange: "none",
    
    /**
     * Specifies the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDisplay: "none",
    
    /**
     * The color that is displayed during a data change animation when a node is updated.
     * @expose
     * @name animationUpdateColor
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @default ""
     */
    animationUpdateColor: "",
    
    /**
     * The number of levels of nodes to display. By default all nodes are displayed.
     * @expose
     * @name displayLevels
     * @memberof oj.ojSunburst
     * @instance
     * @type {number}
     * @default Number.MAX_VALUE
     */
    displayLevels: Number.MAX_VALUE,
    
    /**
     * Specifies the key set containing the ids of sunburst nodes that should be expanded on initial render.
     * Use the <a href="ExpandedKeySet.html">ExpandedKeySet</a> class to specify nodes to expand.
     * Use the <a href="ExpandAllKeySet.html">ExpandAllKeySet</a> class to expand all nodes.                 
     * By default, all sunburst nodes are expanded.
     * @expose
     * @name expanded
     * @memberof oj.ojSunburst
     * @instance
     * @type {KeySet}
     * @default new ExpandAllKeySet()
     * @ojwriteback
     */
    expanded: new oj.ExpandAllKeySet(),
    
    /**
     * An array of category strings used for filtering. Nodes with any category matching an item in this array will be filtered.
     * @expose
     * @name hiddenCategories
     * @memberof oj.ojSunburst
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    hiddenCategories: [],
    
    /**
     * An array of category strings used for highlighting. Nodes matching all categories in this array will be highlighted.
     * @expose
     * @name highlightedCategories
     * @memberof oj.ojSunburst
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    highlightedCategories: [],
    
    /**
     * The matching condition for the highlightedCategories property. By default, highlightMatch is 'all' and only items whose categories match all of the values specified in the highlightedCategories array will be highlighted. If highlightMatch is 'any', then items that match at least one of the highlightedCategories values will be highlighted.
     * @expose
     * @name highlightMatch
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "any"
     * @ojvalue {string} "all"
     * @default "all"
     */
    highlightMatch: "all",
    
    /**
     * Defines the behavior applied when hovering over the nodes.
     * @expose
     * @name hoverBehavior
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "dim"
     * @ojvalue {string} "none"
     * @default "none"
     */
    hoverBehavior: "none",
    
    /**
     * Specifies initial hover delay in ms for highlighting nodes.
     * @expose
     * @name hoverBehaviorDelay
     * @memberof oj.ojSunburst
     * @instance
     * @type {number}
     * @default 200
     * @ojunits milliseconds
     */
    hoverBehaviorDelay: 200,
    
    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojSunburst
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     */
    tooltip: {
      /**
       *  A function that returns a custom tooltip. The function takes a dataContext argument,
       *  provided by the sunburst, with the following properties:
       *  <ul>
       *    <li>parentElement: The tooltip element. The function can directly modify or append content to this element.</li>
       *    <li>id: The id of the hovered node.</li>
       *    <li>label: The label of the hovered node.</li>
       *    <li>value: The value of the hovered node.</li>
       *    <li>radius: The radius of the hovered node.</li>
       *    <li>color: The color of the hovered node.</li>
       *    <li>data: The data object of the hovered node.</li>
       *    <li>componentElement: The sunburst element.</li>
       *  </ul>
       *  The function should return an Object that contains only one of the two properties:
       *  <ul>
       *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
       *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
       *  </ul>
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojSunburst
       * @instance
       * @type {function(Object):Object|null}
       * @default null
       */
      renderer: null
    },
    /**
     * Specifies the selection mode.
     * @expose
     * @name selectionMode
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @default "multiple"
     */
    selectionMode: "multiple",
    
    /**
     * Specifies whether client side rotation is enabled.
     * @expose
     * @name rotation
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "off"
     * @ojvalue {string} "on"
     * @default "on"
     */
    rotation: "on",
    
    /**
     * Specifies whether whether the nodes are sorted by size. When sorting is enabled, nodes that have the same parent are sorted in order of descending size.
     * @expose
     * @name sorting
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "off"
     */
    sorting: "off",
    
    /**
     * An object defining custom root node content for the sunburst.
     * @expose
     * @name rootNodeContent
     * @memberof oj.ojSunburst
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     */
    rootNodeContent: {
      /**
       * A function that returns custom root node content. The function takes a dataContext argument,
       * provided by the sunburst, with the following properties:
       * <ul>
       *   <li>outerBounds: Object containing (x, y, width, height) of the rectangle circumscribing the root node area. 
       *   The x and y coordinates are relative to the top, left corner of the element.</li>
       *   <li>innerBounds: Object containing (x, y, width, height) of the rectangle inscribed in the root node area. 
       *   The x and y coordinates are relative to the top, left corner of the element.</li>
       *   <li>id: The id of the root node.</li>
       *   <li>data: The data object of the root node.</li>
       *   <li>componentElement: The sunburst element.</li>
       * </ul>
       * The function should return an Object with the following property: 
       * <ul>
       *    <li>insert: HTMLElement - HTML element, which will be overlaid on top of the sunburst. 
       *    This HTML element will block interactivity of the sunburst by default, but the CSS pointer-events 
       *    property can be set to 'none' on this element if the sunburst's interactivity is desired. 
       *    </li> 
       * </ul>
       * @expose
       * @name rootNodeContent.renderer
       * @memberof! oj.ojSunburst
       * @instance
       * @type {function(Object):Object|null}
       * @default null
       */
       renderer: null
    },
    
    /**
     * Specifies the starting angle of the sunburst. Valid values are numbers between 0 and 360.
     * @expose
     * @name startAngle
     * @memberof oj.ojSunburst
     * @instance
     * @type {number}
     * @default 90
     * @ojunits degrees
     * @ojmin 0
     * @ojmax 360
     * @ojwriteback
     */
    startAngle: 90,
    
    /**
     * Specifies the label describing the color metric of the sunburst. This label will be used in the legend.
     * @expose
     * @name colorLabel
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @default ""
     */
    colorLabel: "",
    
    /**
     * Specifies the label describing the size metric of the sunburst. This label will be used in the legend.
     * @expose
     * @name sizeLabel
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @default ""
     */
    sizeLabel: "",
    
    /**
     * Specifies whether drilling is enabled. Drillable nodes will show a pointer cursor on hover and fire an <code class="prettyprint">ojBeforeDrill</code> and <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). Drilling on a node causes a property change to the rootNode attribute. The displayLevels attribute can be used in conjunction with drilling to display very deep hieracrchies. Use "on" to enable drilling for all nodes. To enable or disable drilling on individual nodes use the drilling attribute in each node.
     * @expose
     * @name drilling
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "off"
     */
    drilling: "off",
    
    /**
     * The id of the root node. When specified, only the root node and children of the root will be displayed.
     * @expose
     * @name rootNode
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @default ""
     */
    rootNode: "",
    
    /**
     * An array of objects with the following properties that defines the data for the nodes. Also accepts a Promise for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name nodes
     * @memberof oj.ojSunburst
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @default null
     */
    nodes: null,
    
    /**
     * An object defining default properties for the nodes. Component CSS classes should be used to set component wide styling. 
     * This API should be used only for styling a specific instance of the component. Properties specified on this object may 
     * be overridden by specifications on the sunburst nodes. Some property default values come from the CSS and varies based on theme.
     * @expose
     * @name nodeDefaults
     * @memberof oj.ojSunburst
     * @instance
     * @type {Object}
     * @default {"borderWidth": 1, "showDisclosure": "off", "labelMinLength": 1, "labelDisplay": "auto", "labelHalign": "center"}
     */
     nodeDefaults: {
       /**
        * The default border color of the nodes.
        * @expose
        * @name nodeDefaults.borderColor
        * @memberof! oj.ojSunburst
        * @instance
        * @type {string}
        */
       borderColor: "rgba(255,255,255,0.3)",
       
       /**
        * The default border width of the nodes.
        * @expose
        * @name nodeDefaults.borderWidth
        * @memberof! oj.ojSunburst
        * @instance
        * @type {number}
        * @default 1
        * @ojunits pixels
        */
       borderWidth: 1,
       
       /**
        * Specifies whether to display the expand/collapse button on hover. If the button is clicked, the expanded attribute is updated with the new array of node ids.
        * @expose
        * @name nodeDefaults.showDisclosure
        * @memberof! oj.ojSunburst
        * @instance
        * @type {string}
        * @ojvalue {string} "on"
        * @ojvalue {string} "off"
        * @default "off"
        */
       showDisclosure: "off",
      
       /**
        * The horizontal alignment for labels displayed within the node. Only applies to rotated text.
        * @expose
        * @name nodeDefaults.labelHalign
        * @memberof! oj.ojSunburst
        * @instance
        * @type {string}
        * @ojvalue {string} "inner"
        * @ojvalue {string} "outer"
        * @ojvalue {string} "center"
        * @default "center"
        */
       labelHalign: "center",
       
       /**
        * The CSS style object defining the style of the label. The default value comes from the CSS and varies based on theme.
        * @expose
        * @name nodeDefaults.labelStyle
        * @memberof! oj.ojSunburst
        * @instance
        * @type {Object}
        */
       labelStyle: undefined,
       
       /**
        * The minimum number of visible characters needed in order to render a truncated label. If the minimum is not met when calculating the truncated label then the label is not displayed.
        * @expose
        * @name nodeDefaults.labelMinLength
        * @memberof! oj.ojSunburst
        * @instance
        * @type {number}
        * @default 1
        */
       labelMinLength: 1,
       
       /**
        * The label display behavior for the nodes. More labels are generally displayed when using rotation, with the trade off of readability. When auto is used, rotated or horizontal labels will be used based on the client browser and platform.
        * @expose
        * @name nodeDefaults.labelDisplay
        * @memberof! oj.ojSunburst
        * @instance
        * @type {string}
        * @ojvalue {string} "horizontal"
        * @ojvalue {string} "rotated"
        * @ojvalue {string} "off"
        * @ojvalue {string} "auto"
        * @default "auto"
        */
       labelDisplay: "auto",
       
       /**
        * The color of the node hover feedback. The default value comes from the CSS and varies based on theme.
        * @expose
        * @name nodeDefaults.hoverColor
        * @memberof! oj.ojSunburst
        * @instance
        * @type {string}
        */
       hoverColor: undefined,
       
       /**
        * The inner color of the node selection feedback. The default value comes from the CSS and varies based on theme.
        * @expose
        * @name nodeDefaults.selectedInnerColor
        * @memberof! oj.ojSunburst
        * @instance
        * @type {string}
        */
       selectedInnerColor: undefined,
       
       /**
        * The outer color of the node selection feedback. The default value comes from the CSS and varies based on theme.
        * @expose
        * @name nodeDefaults.selectedOuterColor
        * @memberof! oj.ojSunburst
        * @instance
        * @type {string}
        */
       selectedOuterColor: undefined  
     },
    
    /**
     * An array containing the ids of the initially selected nodes.
     * @expose
     * @name selection
     * @memberof oj.ojSunburst
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    selection: [],
    
    /**
     * Data visualizations require a press and hold delay before triggering tooltips and rollover effects on mobile devices to avoid interfering with page panning, but these hold delays can make applications seem slower and less responsive. For a better user experience, the application can remove the touch and hold delay when data visualizations are used within a non scrolling container or if there is sufficient space outside of the visualization for panning. If touchResponse is touchStart the element will instantly trigger the touch gesture and consume the page pan events. If touchResponse is auto, the element will behave like touchStart if it determines that it is not rendered within scrolling content and if panning is not available for those elements that support the feature. 
     * @expose
     * @name touchResponse
     * @memberof oj.ojSunburst
     * @instance
     * @type {string}
     * @ojvalue {string} "touchStart"
     * @ojvalue {string} "auto"
     * @default "auto"
     */
    touchResponse: "auto",
     
    /**
     * Triggered during user rotation of the sunburst.
     *
     * @property {number} value the start angle of the sunburst, in degrees
     *
     * @expose
     * @event
     * @memberof oj.ojSunburst
     * @instance
     */
    rotateInput : null,
    /**
     * Triggered immediately before any node in the sunburst is drilled into. The drill event can be vetoed if the beforeDrill callback returns false.
     *
     * @property {string} id the id of the drilled node
     * @property {Object} data the data object of the drilled node
     * 
     * @expose
     * @event
     * @memberof oj.ojSunburst
     * @instance
     */
    beforeDrill: null,
    /**
     * Triggered during a drill gesture (double click if selection is enabled, single click otherwise).
     *
     * @property {string} id the id of the drilled node
     * @property {Object} data the data object of the drilled node
     *
     * @expose
     * @event
     * @memberof oj.ojSunburst
     * @instance
     */
    drill: null,
     /**
     * Triggered immediately before any node in the sunburst is expanded. The expand event can be vetoed if the beforeExpand callback returns false.
     *
     * @property {string} id the id of the node to expand
     * @property {Object} data the data object of the node to expand
     *
     * @expose
     * @event
     * @memberof oj.ojSunburst
     * @instance
     */
    beforeExpand: null,
    /**
     * Triggered when a node has been expanded. The ui object contains one property, "nodeId", which is the id of the node that has been expanded.
     *
     * @property {string} id the id of the expanded node
     * @property {Object} data the data object of the expanded node
     *
     * @expose
     * @event
     * @memberof oj.ojSunburst
     * @instance
     */    
    expand: null,
    /**
     * Triggered immediately before any container node in the sunburst is collapsed. The collapse event can be vetoed if the beforeCollapse callback returns false.
     *
     * @property {string} id the id of the node to collapse
     * @property {Object} data the data object of the node to collapse
     *
     * @expose
     * @event
     * @memberof oj.ojSunburst
     * @instance
     */
    beforeCollapse: null,
    /**
     * Triggered when a node has been collapsed.
     *
     * @property {string} id the id of the collapsed node
     * @property {Object} data the data object of the collapsed node
     *
     * @expose
     * @event
     * @memberof oj.ojSunburst
     * @instance
     */        
    collapse: null
  },

  //** @inheritdoc */
  _CreateDvtComponent : function(context, callback, callbackObj) {
    return dvt.Sunburst.newInstance(context, callback, callbackObj);
  },

  //** @inheritdoc */
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-sunburst-node') {
      // node[index0][index1]...[indexN]
      subId = 'node' + this._GetStringFromIndexPath(locator['indexPath']);
    }
    else if(subId == 'oj-sunburst-tooltip') {
        subId = 'tooltip';
    }

    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  },

  //** @inheritdoc */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};
    if(subId.indexOf('node') == 0) {
      // node[index0][index1]...[indexN]
      locator['subId'] = 'oj-sunburst-node';
      locator['indexPath'] = this._GetIndexPath(subId);
    }
    else if(subId == 'tooltip') {
        locator['subId'] = 'oj-sunburst-tooltip';
    }
    return locator;
  },

  //** @inheritdoc */
  _ProcessOptions: function() {
    this._super();
    var rootNodeContent = this.options['rootNodeContent'];
    if (rootNodeContent && rootNodeContent['_renderer'])
      rootNodeContent['renderer'] = this._GetTemplateRenderer(rootNodeContent['_renderer'], 'rootNodeContent');
    
    // if expanded not declared, pass default expandAll key set to the toolkit  
    if (!this.options['expanded'])
      this.options['expanded'] = new oj.ExpandAllKeySet();
  },

  //** @inheritdoc */
  _GetComponentRendererOptions: function() {
    return ['tooltip/renderer', 'rootNodeContent/renderer'];
  },
  
  //** @inheritdoc */
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-sunburst');
    return styleClasses;
  },

  //** @inheritdoc */
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-dvtbase oj-sunburst'] = {'path': 'animationDuration', 'property': 'ANIM_DUR'};
    styleClasses['oj-sunburst-attribute-type-text'] = {'path' : 'styleDefaults/_attributeTypeTextStyle', 'property' : 'TEXT'};
    styleClasses['oj-sunburst-attribute-value-text'] = {'path' : 'styleDefaults/_attributeValueTextStyle', 'property' : 'TEXT'};
    styleClasses['oj-sunburst-drill-text'] = {'path' : 'styleDefaults/_drillTextStyle', 'property' : 'TEXT'};
    styleClasses['oj-sunburst-current-drill-text'] = {'path' : 'styleDefaults/_currentTextStyle', 'property' : 'TEXT'};
    styleClasses['oj-sunburst-node'] = {'path' : 'nodeDefaults/labelStyle', 'property' : 'TEXT'};
    styleClasses['oj-sunburst-node oj-hover'] = {'path' : 'nodeDefaults/hoverColor', 'property' : 'border-top-color'};
    styleClasses['oj-sunburst-node oj-selected'] = [
      {'path' : 'nodeDefaults/selectedOuterColor', 'property' : 'border-top-color'},
      {'path' : 'nodeDefaults/selectedInnerColor', 'property' : 'border-bottom-color'}
    ];
    return styleClasses;
  },

  //** @inheritdoc */
  _GetEventTypes : function() {
    return ['optionChange', 'rotateInput','beforeDrill', 'drill', 'beforeExpand', 'expand', 'beforeCollapse', 'collapse'];
  },

  //** @inheritdoc */
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtSunburstBundle.COLLAPSE'] = translations['tooltipCollapse'];
    ret['DvtSunburstBundle.COLOR'] = translations['labelColor'];
    ret['DvtSunburstBundle.EXPAND'] = translations['tooltipExpand'];
    ret['DvtSunburstBundle.SIZE'] = translations['labelSize'];
    ret['DvtUtilBundle.SUNBURST'] = translations['componentName'];
    return ret;
  },

  //** @inheritdoc */
  _HandleEvent : function(event) {
    var type = event['type'];
    var eventData = {'id': event['id'], 'data': event['data']};
    if (!this._IsCustomElement())
      eventData['component'] = event['component'];
  
    if(type === 'rotation') {
      if(event['complete'])
        this._UserOptionChange('startAngle', event['startAngle']);
      else
        this._trigger('rotateInput', null, {'value': event['startAngle']});
    }
    else if(type == 'drill') {
      if(event['id'] && this._trigger('beforeDrill', null, eventData)) {
        this._UserOptionChange('rootNode', event['id']);
        this._Render();
        this._trigger('drill', null, eventData);
      }
    }
    else if(type == 'expand') {
      if(event['id'] && this._trigger('beforeExpand', null, eventData)) {
        this._UserOptionChange('expanded', event['expanded']);
        this._Render();
        this._trigger('expand', null, eventData);
      }
    }
    else if(type == 'collapse') {
      if(event['id'] && this._trigger('beforeCollapse', null, eventData)) {
        this._UserOptionChange('expanded', event['expanded']);
        this._Render();
        this._trigger('collapse', null, eventData);
      }
    }
    else {
      this._super(event);
    }
  },

  //** @inheritdoc */
  _LoadResources : function() {
    // Ensure the resources object exists
    if(this.options['_resources'] == null)
      this.options['_resources'] = {};

    var resources = this.options['_resources'];

    // Add cursors
    resources['rotateCursor'] = oj.Config.getResourceUrl('resources/internal-deps/dvt/sunburst/rotate.cur');
    
    resources['expand'] = 'oj-sunburst-expand-icon';
    resources['expandOver'] = 'oj-sunburst-expand-icon oj-hover';
    resources['expandDown'] = 'oj-sunburst-expand-icon oj-active';
    resources['collapse'] = 'oj-sunburst-collapse-icon';
    resources['collapseOver'] = 'oj-sunburst-collapse-icon oj-hover';
    resources['collapseDown'] = 'oj-sunburst-collapse-icon oj-active';
  },

  /**
   * Returns an object with the following properties for automation testing verification of the node with
   * the specified subid path.
   *
   * @param {Array} subIdPath The array of indices in the subId for the desired node
   * @property {string} color
   * @property {string} label
   * @property {boolean} selected
   * @property {number} size
   * @property {string} tooltip
   * @return {Object|null} An object containing properties for the node, or null if none exists.
   * @expose
   * @instance
   * @memberof oj.ojSunburst
   */
  getNode: function(subIdPath) {
    return this._component.getAutomation().getNode(subIdPath);
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojSunburst
   */
  getContextByNode: function(node)
  {
    // context objects are documented with @ojnodecontext
    var context = this.getSubIdByNode(node);
    if (context && context['subId'] !== 'oj-sunburst-tooltip')
      return context;

    return null;
  },

  //** @inheritdoc */
  _GetComponentDeferredDataPaths : function() {
    return {'root': ['nodes']};
  }
});

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="3">Node</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *     <tr>
 *       <td>Outer Edge</td>
 *       <td><kbd>Drag</kbd></td>
 *       <td>Rotate when <code class="prettyprint">rotation</code> is enabled.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojSunburst
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>
 *         Move focus and selection to the first adjacent sector in an inner or outer layer (ring). In
 *         the northern hemisphere of the sunburst, this will move away from the center, while it will move towards the
 *         center in the southern hemisphere of the sunburst.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>
 *         Move focus and selection to the first adjacent sector in an inner or outer layer (ring). In
 *         the northern hemisphere of the sunburst, this will move towards the center, while it will move away from the
 *         center in the southern hemisphere of the sunburst.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus and selection counterclockwise to adjacent sector in the same layer (ring).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus and selection clockwise to adjacent sector in the same layer (ring).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + UpArrow</kbd></td>
 *       <td>
 *         Move focus and extend selection to the first adjacent sector in an inner or outer layer (ring). In
 *         the northern hemisphere of the sunburst, this will move away from the center, while it will move towards the
 *         center in the southern hemisphere of the sunburst.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + DownArrow</kbd></td>
 *       <td>
 *         Move focus and extend selection to the first adjacent sector in an inner or outer layer (ring). In
 *         the northern hemisphere of the sunburst, this will move towards the center, while it will move away from the
 *         center in the southern hemisphere of the sunburst.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + LeftArrow</kbd></td>
 *       <td>Move focus and extend selection counterclockwise to adjacent sector in the same layer (ring).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + RightArrow</kbd></td>
 *       <td>Move focus and extend selection clockwise to adjacent sector in the same layer (ring).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + UpArrow</kbd></td>
 *       <td>
 *         Move focus to the first adjacent sector in an inner or outer layer (ring), without changing the
 *         current selection. In the northern hemisphere of the sunburst, this will move away from the center, while it
 *         will move towards the center in the southern hemisphere of the sunburst.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + DownArrow</kbd></td>
 *       <td>
 *         Move focus to the first adjacent sector in an inner or outer layer (ring), without changing the
 *         current selection. In the northern hemisphere of the sunburst, this will move towards the center, while it
 *         will move away from the center in the southern hemisphere of the sunburst.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + LeftArrow</kbd></td>
 *       <td>Move focus counterclockwise to adjacent sector in the same layer (ring), without changing the current
 *       selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + RightArrow</kbd></td>
 *       <td>Move focus clockwise to adjacent sector in the same layer (ring), without changing the current
 *       selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Spacebar</kbd></td>
 *       <td>Multi-select sectors with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Alt + LeftArrow</kbd></td>
 *       <td>Rotate 5 degrees counterclockwise.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Alt + RightArrow</kbd></td>
 *       <td>Rotate 5 degrees clockwise.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Drill on a node when <code class="prettyprint">drilling</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>+</kbd></td>
 *       <td>Expand a node when <code class="prettyprint">showDisclosure</code> is enabled and the node is expandable.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>-</kbd></td>
 *       <td>Collapse a node when <code class="prettyprint">showDisclosure</code> is enabled and the node is collapsable.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojSunburst
 */

/**
 * The border color of the node.
 * @expose
 * @name nodes[].borderColor
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The border width of the node.
 * @expose
 * @name nodes[].borderWidth
 * @memberof! oj.ojSunburst
 * @instance
 * @type {number}
 * @default null
 */
/**
 * An array of objects with properties for the child nodes.
 * @expose
 * @name nodes[].nodes
 * @memberof! oj.ojSunburst
 * @instance
 * @type {Array.<Object>}
 * @default null
 */
/**
 * The id of the node.
 * @expose
 * @name nodes[].id
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional array of category strings corresponding to this data item. This enables highlighting and filtering of individual data items through interactions with the legend and other visualization elements. The categories array of each node is required to be a superset of the categories array of its parent node. If not specified, the ids of the node and its ancestors will be used.
 * @expose
 * @name nodes[].categories
 * @memberof! oj.ojSunburst
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The relative size of the node.
 * @expose
 * @name nodes[].value
 * @memberof! oj.ojSunburst
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The fill color of the node.
 * @expose
 * @name nodes[].color
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the node. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the node color attribute.
 * @expose
 * @name nodes[].svgClassName
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the node. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the node color attribute.
 * @expose
 * @name nodes[].svgStyle
 * @memberof! oj.ojSunburst
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The label for this node.
 * @expose
 * @name nodes[].label
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The horizontal alignment for labels displayed within the node. Only applies to rotated text.
 * @expose
 * @name nodes[].labelHalign
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @ojvalue {string} "inner"
 * @ojvalue {string} "outer"
 * @ojvalue {string} "center"
 * @default "center"
 */
/**
 * Specifies whether or not the node will be selectable.
 * @expose
 * @name nodes[].selectable
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * The description of this node. This is used for accessibility and also for customizing the tooltip text.
 * @expose
 * @name nodes[].shortDesc
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The pattern used to fill the node.
 * @expose
 * @name nodes[].pattern
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The CSS style object defining the style of the label.
 * @expose
 * @name nodes[].labelStyle
 * @memberof! oj.ojSunburst
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * Specifies whether to display the expand/collapse button on hover for a specific node. If the button is clicked, the expanded attribute is updated with the new array of node ids.
 * @expose
 * @name nodes[].showDisclosure
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default "inherit"
 */
/**
 * The label display behavior for the nodes. More labels are generally displayed when using rotation, with the trade off of readability. When auto is used, rotated or horizontal labels will be used based on the client browser and platform.
 * @expose
 * @name nodes[].labelDisplay
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @ojvalue {string} "horizontal"
 * @ojvalue {string} "rotated"
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * Specifies whether drilling is enabled for the node. Drillable nodes will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). To enable drilling for all nodes at once, use the drilling attribute in the top level.
 * @expose
 * @name nodes[].drilling
 * @memberof! oj.ojSunburst
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default "inherit"
 */
/**
 * The radius of the node relative to the other nodes.
 * @expose
 * @name nodes[].radius
 * @memberof! oj.ojSunburst
 * @instance
 * @type {number}
 * @default 1
 */

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for sunburst nodes indexed by their position in the hierarchy.</p>
 *
 * @property {Array} indexPath The array of numerical indices for the node.
 *
 * @ojsubid oj-sunburst-node
 * @memberof oj.ojSunburst
 *
 * @example <caption>Get the node at index 0 in the first layer, index 1 in the second:</caption>
 * var nodes = mySunburst.getNodeBySubId( {'subId': 'oj-sunburst-node', 'indexPath': [0, 1]} );
 */

/**
 * <p>Sub-ID for the the sunburst tooltip.</p>
 * 
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 * 
 * @ojsubid
 * @member
 * @name oj-sunburst-tooltip
 * @memberof oj.ojSunburst
 * @instance
 * 
 * @example <caption>Get the tooltip object of the sunburst, if displayed:</caption>
 * var nodes = mySunburst.getNodeBySubId( {'subId': 'oj-sunburst-tooltip'} );
 */
// Node Context Objects ********************************************************

/**
 * <p>Context for sunburst nodes indexed by their position in the hierarchy.</p>
 *
 * @property {Array} indexPath The array of numerical indices for the node.
 *
 * @ojnodecontext oj-sunburst-node
 * @memberof oj.ojSunburst
 */
/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojSunburstMeta = {
  "properties": {
    "animationDuration": {
      "type": "number"
    },
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationUpdateColor": {
      "type": "string"
    },
    "colorLabel": {
      "type": "string"
    },
    "displayLevels": {
      "type": "number"
    },
    "drilling": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "expanded" : {
      "writeback": true
    },
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string",
      "enumValues": ["any", "all"]
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "hoverBehaviorDelay": {
      "type": "number"
    },    
    "nodeDefaults": {
      "type": "object",
      "properties": {
        "borderColor": {
          "type": "string"
        },
        "borderWidth": {
          "type": "number"
        },
        "hoverColor": {
          "type": "string"
        },
        "labelDisplay": {
          "type": "string",
          "enumValues": ["auto", "horizontal", "rotated", "off"]
        },
        "labelHalign": {
          "type": "string",
          "enumValues": ["inner", "outer", "center"]
        },
        "labelStyle": {
          "type": "object"
        },
        "selectedInnerColor": {
          "type": "string"
        },
        "selectedOuterColor": {
          "type": "string"
        },
        "showDisclosure": {
          "type": "string",
          "enumValues": ["on", "off"]
        }
      }
    },
    "nodes": {
      "type": "Array<object>|Promise"
    },
    "rootNode": {
      "type": "string"
    },
    "rotation": {
      "type": "string",
      "enumValues": ["off", "on"]
    },
    "rootNodeContent": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["none", "single", "multiple"]
    },
    "sizeLabel": {
      "type": "string"
    },
    "sorting": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "startAngle": {
      "type": "number",
      "writeback": true
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "touchResponse": {
      "type": "string"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Sunburst"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelColor": {
          "type": "string",
          "value": "Color"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "labelSize": {
          "type": "string",
          "value": "Size"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        },
        "tooltipCollapse": {
          "type": "string",
          "value": "Collapse"
        },
        "tooltipExpand": {
          "type": "string",
          "value": "Expand"
        }
      }
    }
  },
  "methods": {
    "getContextByNode": {},
    "getNode": {}
  },
  "events": {
    "rotateInput": {},
    "beforeDrill": {},
    "drill": {},
    "beforeExpand": {},
    "expand": {},
    "beforeCollapse": {},
    "collapse": {}
  },
  "extension": {
    _WIDGET_NAME: "ojSunburst"
  }
};
oj.CustomElementBridge.registerMetadata('oj-sunburst', 'dvtBaseComponent', ojSunburstMeta);
oj.CustomElementBridge.register('oj-sunburst', {'metadata': oj.CustomElementBridge.getMetadata('oj-sunburst')});
})();

/**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @class oj.SwipeToRevealUtils
 * @since 1.2.0
 * @export
 * @hideconstructor
 * @ojstatus preview
 *
 * @classdesc
 * This class provides functions for setting up and handling swipe to reveal on an offcanvas element.  The offcanvas
 * element contains contextual actions that users can perform on the element that user perform the swipe gesture on.
 * This is most commonly found in ListView where user swipes on an item to reveal contextual actions that can be done on the item.
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>Application must ensure that the context menu is available and setup with the
 * equivalent menu items so that keyboard-only users can perform all the swipe actions 
 * just by using the keyboard.  
 */
oj.SwipeToRevealUtils = {};

/**
 * Setup listeners needed for swipe actions capability.  
 *
 * @export
 * @param {Element} elem the DOM element (of the offcanvas) that hosts the swipe actions
 * @param {Object=} options the options to set for swipe actions
 * @param {number=} options.threshold the threshold that triggers default action.  If no default action is found (no item with style  
 *                 "oj-swipetoreveal-default") then this value is ignored.  If percentage value is specified it will be calculated
 *                 based on the width of the element with class "oj-offcanvas-outer-wrapper".  A default value is determined if not specified.
 *                 An "ojdefaultaction" event will be fired when threshold is exceed upon release.
 * @return {void}
 *
 * @see #tearDownSwipeActions
 * @see oj.OffcanvasUtils.html#setupPanToReveal
 */
oj.SwipeToRevealUtils.setupSwipeActions = function(elem, options)
{
    var drawer, direction, offcanvas, outerWrapper, threshold, minimum, drawerShown, evt, checkpoint, defaultAction, distance;

    drawer = $(elem);
    // checks if it's already registered
    if (drawer.hasClass("oj-swipetoreveal"))
    {
        return;
    }

    drawer.addClass("oj-swipetoreveal");

    direction = drawer.hasClass("oj-offcanvas-start") ? "end" : "start";

    offcanvas = {};
    offcanvas["selector"] = drawer;
    oj.OffcanvasUtils.setupPanToReveal(offcanvas);

    outerWrapper = oj.OffcanvasUtils._getOuterWrapper(drawer);

    // the panning triggers a click event at the end (since we are doing translation on move, the relative position has not changed)
    // this is to prevent the click event from bubbling (to list item for example, see )
    drawerShown = false;
    outerWrapper.on("click.swipetoreveal", function(event)
    {
        if (drawerShown)
        {
            event.stopImmediatePropagation();   
            drawerShown = false;
        }
    });

    // However, this does not get trigger in hybrid app, see .  
    // this change ensures that it always get reset
    outerWrapper.on("touchstart.swipetoreveal", function(event)
    {
        drawerShown = false;
    });

    drawer
    .on("ojpanstart", function(event, ui) 
    {
        // if the swipe direction does not match the offcanvas's edge, veto it
        if (ui['direction'] != direction)
        {
            event.preventDefault();
        }
        else
        {
            // figure out the threshold for default action and the minimum distance
            // to keep the offcanvas open
            if (minimum == null)
            {
                if (options != null)
                {
                    threshold = options['threshold'];
                }

                if (threshold != null)
                {
                    threshold = parseInt(threshold, 10);

                    // check if it's percentage value
                    if (/%$/.test(options['threshold']))
                    {
                        threshold = (threshold / 100) * outerWrapper.outerWidth();
                    }
                }
                else
                {
                    // by default it will be 55% of the outer wrapper
                    threshold = outerWrapper.outerWidth() * 0.55;
                }
                // by default the minimum will be the lesser of the width of the offcanvas and half of the outer wrapper
                minimum = Math.min(outerWrapper.outerWidth() * 0.3, drawer.outerWidth());
            }

            // setup default style class
            drawer.children().addClass("oj-swipetoreveal-action");

            // find if there's any default action item specified
            defaultAction = drawer.children(".oj-swipetoreveal-default").get(0);

            // used to determine if it's a quick swipe
            checkpoint = (new Date()).getTime();
        }
    })
    .on("ojpanmove", function(event, ui)
    {
        drawerShown = true;

        // check if pan pass the threshold position, the default action item gets entire space.
        if (defaultAction != null)
        {
            if (ui['distance'] > threshold)
            {
                drawer.children().each(function() {
                    if (this != defaultAction)
                    {
                        $(this).addClass("oj-swipetoreveal-hide-when-full");
                    }                  
                });
            }
            else
            {
                drawer.children().removeClass("oj-swipetoreveal-hide-when-full");
            }
        }
    })
    .on("ojpanend", function(event, ui) 
    {
        distance = ui['distance'];
        if (defaultAction != null && distance > threshold)
        {
            // default action
            evt = $.Event("ojdefaultaction");
            drawer.trigger(evt, offcanvas);
            event.preventDefault();
        }

        // if pan pass the minimum threshold position, keep the toolbar open
        if (distance < minimum)
        {
            // check if this is a swipe, the time should be < 200ms and the distance must be > 10px
            if ((new Date()).getTime() - checkpoint > 200 || distance < 10)
            {
                event.preventDefault();
            }
        }
    });
};

/**
 * Removes the listener that was added in setupSwipeActions.  Page authors should call tearDownSwipeActions when the content container is no longer needed.
 * 
 * @export
 * @param {Element} elem the DOM element (of the offcanvas) that hosts the swipe actions
 * @return {void}
 *
 * @see #setupSwipeActions
 * @see oj.OffcanvasUtils.html#tearDownPanToReveal
 */
oj.SwipeToRevealUtils.tearDownSwipeActions = function(elem)
{
    var drawer, offcanvas, outerWrapper;

    drawer = $(elem);

    drawer.removeClass("oj-swipetoreveal");

    offcanvas = {};
    offcanvas["selector"] = drawer;

    outerWrapper = oj.OffcanvasUtils._getOuterWrapper(drawer);
    if (outerWrapper != null)
    {
        outerWrapper.off(".swipetoreveal");
    }

    oj.OffcanvasUtils.tearDownPanToReveal(offcanvas);
};

/**
 * The following CSS classes can be applied by the page author as needed.
 * <p>
 * <table class="generic-table styling-table">
 *   <thead>
 *     <tr>
 *       <th>Class</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>oj-swipetoreveal-more</td>
 *       <td>Designed for use with an action item that shows more available actions that users can perform.
 *           <p>Is applied to the element that represents the more action item inside the offcanvas element.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-swipetoreveal-flag</td>
 *       <td>Designed for use with an action item that tags the associated item in the host like listview item.
 *           <p>Is applied to the element that represents the flag action item inside the offcanvas element.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-swipetoreveal-alert</td>
 *       <td>Designed for use with an action item that performs an explicit action like deleting the associated listview item.
 *           <p>Is applied to the element that represents the alert action item inside the offcanvas element.</td>
 *     </tr>
 *     <tr>
 *       <td>oj-swipetoreveal-default</td>
 *       <td>Designed for use with an action item that should get all the space when user swipes pass the threshold distance.
 *           <p>Is applied to the element that represents the default action item inside the offcanvas element.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
 * @memberof oj.SwipeToRevealUtils
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Offcanvas wrapper element</td>
 *       <td><kbd>Swipe</kbd></td>
 *       <td>Reveals the offcanvas element.  Depending on the distance relative to the target is swiped, the offcanvas will either be closed (swipe distance too short), opened, or the default action is performed (swipe distance passed the specified threshold).</td>
 *     </tr>
 *     <tr>
 *       <td>Offcanvas wrapper element</td>
 *       <td><kbd>Pan</kbd></td>
 *       <td>Reveals the offcanvas element.  If a default action is specified, the default action will get the entire space of the offcanvas after the user panned past a specified distance threshold.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.SwipeToRevealUtils
 */
 /**
 * Copyright (c) 2015, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

(function () {
  /*!
   * JET Switch @VERSION
   */
  /**
   * @ojcomponent oj.ojSwitch
   * @ojdisplayname Switch
   * @augments oj.editableValue
   * @ojsignature [{
   *                target: "Type",
   *                value: "class ojSwitch extends editableValue<boolean, ojSwitchSettableProperties>"
   *               },
   *               {
   *                target: "Type",
   *                value: "ojSwitchSettableProperties extends editableValueSettableProperties<boolean>",
   *                for: "SettableProperties"
   *               }
   *              ]
   * @since 0.7
   * @ojshortdesc Provides basic support for toggling a boolean value.
   * @ojstatus preview
   *
   * @classdesc
   * <p>
   * the oj-switch component enhances <code class="prettyprint">input</code>
   * element and manages the selection of Boolean values.
   * </p>
   *
   * <h3 id="touch-section">
   *   Touch End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"touchDoc"}
   *
   *
   * <h3 id="keyboard-section">
   *   Keyboard End User Information
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
   * </h3>
   *
   * {@ojinclude "name":"keyboardDoc"}
   *
   *
   * <h3 id="a11y-section">
   *   Accessibility
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
   * </h3>
   *
   * <p>
   * The component is accessible; it sets and maintains the appropriate aria- attributes,
   * like aria-checked and aria-disabled.
   * </p>
   * <p>
   * It is up to the application developer to associate the label to the input element.
   * For Switch, you should put an <code class="prettyprint">id</code> on the input element, and
   * then set the <code class="prettyprint">for</code> attribute on the label to be the
   * input's id.
   * </p>
   *
   * <p>See also the <a href="#styling-section">oj-focus-highlight</a> discussion.
   *
   * <h3 id="styling-section">
   *   Styling
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
   * </h3>
   * {@ojinclude "name":"stylingDoc"}
   *
   * <h3 id="label-section">
   *   Label and Switch
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#label-section"></a>
   * </h3>
   * <p>
   * For accessibility, you should associate a label element with the input by putting an
   * <code class="prettyprint">id</code> on the input, and then setting the
   * <code class="prettyprint">for</code> attribute on the label to be the input's id.
   * </p>
   * <h3 id="state-section">
   *   Setting the Value Attribute
   *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#state-section"></a>
   * </h3>
   * <p>The value attribute should be Boolean. If the value attribute is undefined, then it is
   * <code class="prettyprint">false</code>.
   * </p>
   *
   * @example <caption>Initialize the switch with no attributes specified:</caption>
   * &lt;oj-switch>&lt;/oj-switch>
   *
   * @example <caption>Initialize the switch with some attributes specified:</caption>
   * &lt;oj-switch value=true disabled=false>&lt;/oj-switch>
   */
  oj.__registerWidget('oj.ojSwitch', $['oj']['editableValue'], {
    version: '1.1.0',
    defaultElement: '<input>',
    widgetEventPrefix: 'oj',

    options:
    {
      /**
       * <p>
       * Whether the component is disabled. The element's
       * <code class="prettyprint">disabled</code>
       *  property is used as its initial
       * value if it exists, when the attribute is not explicitly set. When neither is set,
       * <code class="prettyprint">disabled </code>
       * defaults to false.
       *
       * @example <caption>Initialize the switch with
       * <code class="prettyprint">disabled</code> attribute:</caption>
       * &lt;oj-switch disabled="true">&lt;/oj-switch>
       * @example <caption>Get or set the <code class="prettyprint">disabled</code> property after initialization:</caption>
       * // Getter
       * var disabled = myComponent.disabled;
       * 
       * // Setter
       * myComponent.disabled = true;
       * 
       * @expose
       * @type {boolean}
       * @default false
       * @public
       * @instance
       * @memberof oj.ojSwitch
       */
      disabled: false,

      /**
       * Whether the component is readOnly. The element's
       * <code class="prettyprint">readOnly</code>
       *  property is used as its initial
       * value if it exists, when the attribute is not explicitly set. When neither is set,
       * <code class="prettyprint">readOnly </code>
       * defaults to false.
       *
       * @example <caption>Initialize the switch with
       * <code class="prettyprint">readOnly</code> attribute:</caption>
       * &lt;oj-switch readonly="true">&lt;/oj-switch>
       * @example <caption>Get or set the <code class="prettyprint">readonly</code> property after initialization:</caption>
       * // Getter
       * var readonly = myComponent.readonly;
       * 
       * // Setter
       * myComponent.readonly = true;
       * 
       *
       * @expose
       * @type {boolean}
       * @alias readonly
       * @default false
       * @instance
       * @memberof oj.ojSwitch
       */
      readOnly: false,

      /**
       * The state of the switch component (true/false).
       *
       * @example <caption>Initialize component (switch is ON) with
       * <code class="prettyprint">value</code> attribute:</caption>
       * &lt;oj-switch value="true">&lt;/oj-switch>
       * @example <caption>Get or set the <code class="prettyprint">value</code> property after initialization:</caption>
       * // Getter
       * var value = myComponent.value;
       * 
       * // Setter
       * myComponent.value = true;
       * 
       * @expose
       * @type {boolean}
       * @default false
       * @ojwriteback
       * @public
       * @instance
       * @memberof oj.ojSwitch
       */
      value: false
    },

    // P U B L I C    M E T H O D S

    /**
     * Returns a jQuery object containing the element visually representing the switch.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof oj.ojSwitch
     * @instance
     * @return {jQuery} the switch
     * @ignore
     */
    widget: function () {
      return this._element2;
    },

    // @inheritdoc
    getNodeBySubId: function (locator) {
      var node = this._super(locator);
      var subId;
      var rootElement = this.widget();

      if (!node)
      {
        node = (locator == null || locator['subId'] == null) ? rootElement : null;
        if (!node) {
          subId = locator['subId'];
          if (subId === 'oj-switch-thumb' || subId === 'oj-switch-track') {
            return rootElement.find('.' + subId)[0];
          }
        }
      }
      
      return node || null;
    },

    /**
     * Returns the subId locator for the given child DOM node.
     * <p>
     * If DOM node is null then method returns null.
     * </p>
     * <p>
     * If DOM node is not a child of the current component then method returns null.
     * </p>
     *
     * @expose
     * @override
     * @ignore
     * @memberof oj.ojSwitch
     * @instance
     * @param {!Element} node - child DOM node
     * @return {Object|null} The subId for the DOM node, or null when none is found.
     */
    getSubIdByNode: function (node) {
      var originalId = $(this.element).attr('id');
      var nodeId;
      var result = this._super(node);

      if (node != null) {
        if ($(node).hasClass('oj-switch-track') || $(node).hasClass('oj-switch-thumb')) {
          nodeId = $(node).parents('div.oj-switch')
            .find('input.oj-component-initnode')
            .attr('id');
          if (originalId === nodeId && $(node).hasClass('oj-switch-track')) {
            result = { 'subId': 'oj-switch-track' };
          } else if (originalId === nodeId && $(node).hasClass('oj-switch-thumb')) {
            result = { 'subId': 'oj-switch-thumb' };
          }
        }
      }

      return result;
    },

    // P R O T E C T E D    C O N S T A N T S   A N D   M E T H O D S

    /**
     * @private
     * @const
     */
    _BUNDLE_KEY:
    {
      _SWITCH_OFF: 'SwitchOFF',
      _SWITCH_ON: 'SwitchON'
    },

    /**
     * Overridden to set the options.value. When constructorOptions value is undefined,
     * we read the CHECKED, DISABLED, READONLY OPTIONS on the checkbox and
     * build the switch options from that.
     *
     * @memberof oj.ojSwitch
     * @instance
     * @protected
     */
    _InitOptions: function (originalDefaults, constructorOptions) {
      var props;
      var val;

      props = [
        { attribute: 'disabled', validateOption: true },
        { attribute: 'readonly', option: 'readOnly', validateOption: true },
        { attribute: 'checked', option: 'value', validateOption: false,
         coerceDomValue: function (domValue) {
           return domValue ? true : false;
         } },
        { attribute: 'title' }
      ];

      this._super(originalDefaults, constructorOptions);
      
      // Only needed for non new element style
      if (!this._IsCustomElement()) {
        oj.EditableValueUtils.initializeOptionsFromDom(props, constructorOptions, this);
        val = this.option('value');
        this.option({ value: !!val },
                    { '_context': { writeback: true, internalSet: true } });
      }
    },

    /**
     * Create switch component
     *
     * @override
     * @protected
     * @memberof oj.ojSwitch
     * @instance
     */
    _ComponentCreate: function () {
      this._super();

      if (!this.element.is('input')) {
        throw new Error('ojSwitch can be bound to INPUT only.');
      }

      this._inputElementOriginalDisplay = this.element.css('display');
      this.element
        .css('display', 'none')
        .attr('type', 'checkbox')
        .attr('checked', this.option('value'))
        .attr('tabindex', '-1')
        .attr('disabled', this.option('disabled'))
        .attr('readonly', this.option('readOnly'));

      if (this.OuterWrapper) {
        this._element2 = $(this.OuterWrapper);
      } else {
        this._element2 = this.element.wrap('<div></div>').parent();  // @HTMLUpdateOK trusted string
      }
      this._element2.addClass('oj-switch oj-component oj-form-control');
      this._element2.append("<div class='oj-switch-container'><div class='oj-switch-track'><div class='oj-switch-thumb' tabIndex='0'></div></div></div>"); // @HTMLUpdateOK append or prepend trusted new DOM to switch elem

      this.switchThumb = this._element2.find('.oj-switch-thumb');
      this.switchThumb
        .attr('role', this._setSwitchRole());
    },

    _AfterCreate: function () {
      this._super();
    
      // Get aria-label and aria-labelledby attribute values from the component and move them to the thumb which is the dom that gets focus
      var target = this.switchThumb;
      this._SetAriaInfo(target);
    },
		
    /**
     * Set aria label information for the switch.  It will come from either aria-labelledby or aria-label
     * @protected
     * @memberof oj.ojSwitch
     * @ignore
     */
    _SetAriaInfo: function(target) {
      var component;
      
      // Grab the right component
      if (this.OuterWrapper) {
    	component =  this._element2[0]; 
      }	  
      else {
        component = this._element2[0].querySelector("input");
      }
            
      var labelElementId;
      if (this._IsCustomElement()) {
        // Custom element case
          var defaultLabelId = this["uuid"] + "_Label";
          labelElementId = oj.EditableValueUtils.getOjLabelId(this.widget(), defaultLabelId);
      }
      else {
        // Non custom element case
        var label = this._GetLabelElement();
        if (label) {
          labelElementId = label.attr('id');
        }
      }

      // Apply the label to the target
      if (labelElementId)
      {
        // Set the aria-labelledby attribute of the thumb to the returned id
        target.attr('aria-labelledby', labelElementId);
      }
      else
      {
        // Check if the element has aria-label
        var ariaLabelString = component.getAttribute('aria-label');
        if (ariaLabelString)
        {
          // Set the aria-label of the thumb to the returned string
          target.attr('aria-label', ariaLabelString);
          // And remove it from the component
          component.removeAttr('aria-label');
        }
      }
    },

    /**
     * _setup is called on create and refresh.
     *
     * @memberof oj.ojSwitch
     * @instance
     * @private
     */
    _setup: function () {
      var rootElement = $(this.widget());

      this._setupEvents();

      if (rootElement === undefined) return;
      this.element
        .attr('checked', this.option('value')); // Switch vs Input synchonization

      rootElement.removeClass('oj-disabled oj-read-only oj-selected oj-hover oj-active');
      $(this.switchThumb).attr('tabindex', '0');
      $(this.switchThumb).html(''); // @HTMLUpdateOK

      if (this.option('disabled') || this.option('readOnly')) {
        if (this.option('disabled')) {
          rootElement.addClass('oj-disabled');
        } else {
          rootElement.addClass('oj-read-only');
          $(this.switchThumb).html(this._setReadOnlyValue()); // @HTMLUpdateOK internal strings
        }
        $(this.switchThumb).removeAttr('tabindex');
      }

      if (this.option('value')) {
        rootElement.addClass('oj-selected');
      }

      // Aria
      $(this.switchThumb).attr('aria-checked', this.option('value'));

      $(this.switchThumb).removeAttr('aria-disabled');
      rootElement.removeAttr('aria-disabled');
      if (!this._CanSetValue()) {
        $(this.switchThumb).attr('aria-disabled', 'true');
      }
    },

    /**
     * Translate On/Off switch
     *
     * @protected
     * @memberof oj.ojSwitch
     * @instance
     */
    _setReadOnlyValue: function () {
      var strReturn = this._BUNDLE_KEY._SWITCH_OFF;
      if (this.option('value')) {
        strReturn = this._BUNDLE_KEY._SWITCH_ON;
      }
      return this.getTranslatedString(strReturn);
    },

    /**
     * Binding events to the switch component
     *
     * @protected
     * @memberof oj.ojSwitch
     * @instance
     */
    _setupEvents: function () {
      this._off(this._element2, 'keydown keyup mousedown mouseup mouseleave mouseenter touchstart touchend');
      if (this._CanSetValue()) {
        this._on(this._element2, this._switchEvents);
        this._AddHoverable(this._element2);
      }
      this._focusable({
        'element': this.switchThumb,
        'applyHighlight': true
      });
    },

    /**
     * @override
     * @private
     */
    _switchEvents:
    {
      keydown: function (event) {
        // ENTER and SPACE will change the switch
        if (event.which === $.ui.keyCode.ENTER || event.which === $.ui.keyCode.SPACE) {
          $(event.currentTarget).addClass('oj-active');
          event.preventDefault();
        }
      },
      keyup: function (event) {
        // ENTER and SPACE will change the switch
        if (event.which === $.ui.keyCode.ENTER || event.which === $.ui.keyCode.SPACE) {
          this._SetValue(!this.option('value'), event);
        }
      },
      mousedown: function (event) {
        // LEFT MOUSE BUTTON will change the switch
        if (event.which === 1) {
          $(event.currentTarget).addClass('oj-active');
          event.preventDefault();
        }
      },
      mouseup: function (event) {
        // LEFT MOUSE BUTTON will change the switch
        if (event.which === 1) {
          this._SetValue(!this.option('value'), event);
        }
      },
      mouseleave: function (event) {
        if (event.which === 1) {
          $(event.currentTarget).removeClass('oj-active');
        }
      },
      mouseenter: function (event) {
        if (event.which === 1) {
          $(event.currentTarget).addClass('oj-active');
        }
      },
      "touchstart": function (event) {
		$(event.currentTarget).addClass('oj-active');  
          event.preventDefault();
      },
      "touchend": function (event) {
        this._SetValue(!this.option('value'), event);
      }
    },

    /**
     * Returns the default styleclass for the component. Currently this is
     * used to pass to the ojLabel component, which will append -label and
     * add the style class onto the label. This way we can style the label
     * specific to the input component. For example, for inline labels, the
     * checkboxset/checkboxset components need to have margin-top:0, whereas all the
     * other inputs need it to be .5em. So we'll have a special margin-top style
     * for .oj-label-inline.oj-checkboxset-label
     * All input components must override
     *
     * @return {string}
     * @memberof oj.ojSwitch
     * @override
     * @protected
     */
    _GetDefaultStyleClass: function () {
      return 'oj-switch';
    },
  /**
   * Whether the a value can be set on the component. For example, if the component is 
   * disabled or readOnly then setting value on component is a no-op. 
   * 
   * @see #_SetValue
   * @return {boolean}
   * @memberof oj.ojSwitch
   * @override
   * @protected
   */
    _CanSetValue: function ()
    {
      var readOnly;
      var superCanSetValue = this._super();

      if (!superCanSetValue)
        return false;

      readOnly = this.options['readOnly'] || false;
      return (readOnly) ? false : true;
    },
    /**
     * Returns switch role for ARIA
     * ToDo: for IE it should be role="cehckbox"
     *
     * @return {string}
     * @memberof oj.ojSwitch
     * @override
     * @protected
     */
    _setSwitchRole: function () {
      return 'switch checkbox';
    },

    /**
     * @ignore
     * @protected
     * @override
     */
    _destroy: function () {
      if (this._CanSetValue()) {
        this._RemoveHoverable(this._element2);
      }
      this._element2.find('.oj-switch-track').remove();
      oj.DomUtils.unwrap(this.element);
      this._RestoreAttributes(this.element);
      return this._super();
    },

    /**
     * Returns a jquery object of the launcher element representing the content nodes (switch).
     * @protected
     * @override
     * @memberof oj.ojSwitch
     */
    _GetMessagingLauncherElement: function () {
      return this._element2;
    },

    /**
     * Returns a jquery object of the elements representing the content nodes (switch thumb).
     * @protected
     * @override
     * @memberof oj.ojSwitch
     */
    _GetContentElement: function () {
      return this.switchThumb;
    },
    /**
   * Performs post processing after _SetOption() is called. Different options when changed perform
   * different tasks. See _AfterSetOption[OptionName] method for details.
   *
   * @param {string} option
   * @param {Object|string=} previous
   * @param {Object=} flags
   * @protected
   * @override
   * @memberof oj.ojSwitch
   * @instance
   */
  _AfterSetOption : function (option, previous, flags)
  {
    this._superApply(arguments);
    switch (option)
    {        
      case "readOnly":
        this._AfterSetOptionDisabledReadOnly(option, oj.EditableValueUtils.readOnlyOptionOptions);
        break;
      default:
        break;
    }

  },

  /**
   * @override
   * @private
   */
  _setOption: function (key, value, flags) {
    var coercedValue;
    switch (key) {
      case 'disabled':
      case 'readOnly':
      case 'value':
        coercedValue = !!value;
        break;
      default:
        coercedValue = value;
    }
    this._super(key, coercedValue, flags);
  },

  /**
   * Used for explicit cases where the component needs to be refreshed
   * (e.g., when the value option changes or other UI gestures).
   * @override
   * @protected
   * @memberof oj.ojSwitch
   */
  _Refresh : function (name, value, forceDisplayValueRefresh)
  {
    this._superApply(arguments);
    this._setup();
  }


    // Fragments:

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Switch Thumb</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Toggle switch value</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojSwitch
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Switch Thumb</td>
     *       <td><kbd>Enter</kbd> or <kbd>Space</kbd></td>
     *       <td>Toggle switch value</td>
     *     </tr>
     *     <tr>
     *       <td>Switch Thumb</td>
     *       <td><kbd>Tab In</kbd></td>
     *       <td>Set focus to the thumb. If hints, title or messages exist in a notewindow,
     *        pop up the notewindow.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * <p>Disabled items can not receive keyboard focus.
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojSwitch
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     *
     * <p>
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-switch</td>
     *       <td>Top level switch class. </td>
     *     </tr>
     *     <tr>
     *       <td>oj-focus-highlight</td>
     *       <td>{@ojinclude "name":"ojFocusHighlightDoc"}</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojSwitch
     */

  });
}());

/// ///////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for the switch's track.</p>
 *
 * @ojsubid oj-switch-track
 * @memberof oj.ojSwitch
 *
 * @example <caption>Get the node for the track:</caption>
 * var node = myComponent.getNodeBySubId({'subId': 'oj-switch-track'});
 */

/**
 * <p>Sub-ID for the switch's thumb.</p>
 *
 * @ojsubid oj-switch-thumb
 * @memberof oj.ojSwitch
 *
 * @example <caption>Get the node for the thumb:</caption>
 * var node = myComponent.getNodeBySubId({'subId': 'oj-switch-thumb'});
 */

(function() {
var ojSwitchMeta = {
  "properties": {
    "value": {
      "type": "boolean",
      "writeback": true
    },
    "readonly": {
      "type": "boolean"
    }
  },
  "methods": {
  },
  "extension": {
    _ALIASED_PROPS: {"readonly": "readOnly"},
    _INNER_ELEM: 'input',
    _WIDGET_NAME: "ojSwitch"
  }
};
oj.CustomElementBridge.registerMetadata('oj-switch', 'editableValue', ojSwitchMeta);
oj.CustomElementBridge.register('oj-switch', {'metadata': oj.CustomElementBridge.getMetadata('oj-switch')});
})();
/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojSwitcher
 * @ojsignature {
 *                target: "Type",
 *                value: "class ojSwitcher extends JetElement<ojSwitcherSettableProperties>"
 *               }
 * @since 4.0.0
 * @ojstatus preview
 * @ojshortdesc Dynamically decides which child element should be made visible.
 * @classdesc The switcher element dynamically decides which child element should be made visible. It will make a child element visible only if it's <code class="prettyprint">slot</code> attribute's value matches with switcher's <code class="prettyprint">value</code> property.
 * <p> If child content is expensive to render, use <a href="oj.ojDefer.html">oj-defer</a> to defer rendering until child element is made visible. 
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-switcher value="[[selectedValue]]">
 *  &lt;div slot="home">...&lt;div>
 *  &lt;div slot="about">...&lt;div>
 *  &lt;div slot="settings">...&lt;div>
 *  ...
 * &lt;/oj-switcher>
 * </code></pre>
 */


/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 * 
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 * @expose
 * @memberof oj.ojSwitcher
 * @instance
 * 
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */ 
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * 
 * @expose
 * @memberof oj.ojSwitcher
 * @instance
 * 
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */ 
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 * @expose
 * @memberof oj.ojSwitcher
 * @instance
 * 
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */ 
/**
 * Refreshes the visual state of the component.
 * 
 * @function refresh
 * @return {void}
 * @expose
 * @memberof oj.ojSwitcher
 * @instance
 */ 

/**
 * @member
 * @name value
 * @memberof oj.ojSwitcher
 * @instance
 * @type {string}
 * @default undefined
 * @ojshortdesc Gets and Sets value for this switcher.
 * @desc <code class="prettyprint">value</code> of the switcher. Setting <code class="prettyprint">value</code> will make all child elements with matching <code class="prettyprint">slot</code> attribute as visible and hides elements which are not matching.
 * @example <caption>Initialize the Switcher with the <code class="prettyprint">value</code> attribute specified:</caption>
 *  &lt;oj-switcher value='settings'> ... &lt;/oj-switcher>
 * @example <caption>Read value property:</caption>
 * var selectedValue = mySwitcher.value;
 * @example <caption>Change value property:</caption>
 * mySwitcher.value = 'settings';
 */

/**
 * @ignore
 */
var switcherMetadata =
        {
          "properties": {
            "value": {
              "type": "string",
              "value": ""
            }
          },
          "extension": {
            _CONSTRUCTOR: ojSwitcher,
            _CONTROLS_SUBTREE_HIDDEN: true
          }
        };

/**
 * @constructor
 * @private
 */
function ojSwitcher(context) {
  var ATTR_SWITCHER_SLOT = "slot";
  var SWITCHER_VALUE_ATTR = "value";
  var SWITCHER_ORIG_DISPLAY_STYLE = "_ojSwitcher_orig_display_style";
  var self = this;
  var element = context.element;
  var isInitialRender = true;
  var _slotMap = null;
    
  // Our version of GCC has a bug where the second param of MutationObserver.observe must be of 
  // type MutationObserverInit which isn't a real class that we can instantiate. Work around is to
  // create the MutationObserver on an alias of 'this' and call observe in a different function.
  // TODO Cleanup when we replace GCC with uglify in 5.0.0.
  self._caseElementMutationObserver = new MutationObserver(function (mutations) {
    mutations.forEach(function (mutation) {
      if (mutation.type === "childList") {
        if (mutation.addedNodes) {
          //For each newly added child node, apply the switcher value 
          Array.prototype.forEach.call(mutation.addedNodes, function (item) {
            if (item.nodeType === 1) {
              var itemSlotAttributeValue = item.getAttribute(ATTR_SWITCHER_SLOT);
              _applyValueToItem(item, itemSlotAttributeValue);
              //force to re calculate slot map nex time
              _resetSlotMap();
            }
          });
        }
        if (mutation.removedNodes) {
          //For each removed child node restore original display style property
          Array.prototype.forEach.call(mutation.removedNodes, function (item) {
            if (item.nodeType === 1) {
              if (item[SWITCHER_ORIG_DISPLAY_STYLE] !== undefined) {
                item.style.display = item[SWITCHER_ORIG_DISPLAY_STYLE];
              }
              //force to re calculate slot map nex time
              _resetSlotMap();
            }
          });
        }
      }
    });
  });

  function _resetSlotMap() {
    _slotMap = null;
  }

  function _getSlotMap() {
    if (!_slotMap) {
      _slotMap = oj.BaseCustomElementBridge.getSlotMap(element);
    }
    return _slotMap;
  }

  function _hide(item) {
    var isCurrentlyVisible = !(item.style.display === "none");
    if (isCurrentlyVisible) {
      item.style.display = "none";
      oj.Components.subtreeHidden(item);
    }
  }

  function _show(item) {
    var isCurrentlyVisible = !(item.style.display === "none");
    if (!isCurrentlyVisible) {
      item.style.display = "";
      oj.Components.subtreeShown(item, isInitialRender ? {'initialRender': true} : undefined);
    } else if (isInitialRender) {
      oj.Components.subtreeShown(item, {'initialRender': true});
    }
  }

  this.createDOM = function() {
    self._caseElementMutationObserver.observe(element, {"childList": true});
  };

  this.updateDOM = function() {
    var slots = _getSlotMap();
    for (var key in slots) {
      slots[key].forEach(function (item) {
        _applyValueToItem(item, key);
      });
    }
    if (isInitialRender)
      isInitialRender = false;
  };

  function _applyValueToItem(item, itemSlotAttributeValue) {
    var switcherValue = element[SWITCHER_VALUE_ATTR];
    if (item[SWITCHER_ORIG_DISPLAY_STYLE] === undefined) {
      Object.defineProperty(item, SWITCHER_ORIG_DISPLAY_STYLE, {'value': item.style.display});
    }
    if (itemSlotAttributeValue === switcherValue) {
      _show(item);
    } else {
      _hide(item);
    }
  }
}

oj.CustomElementBridge.registerMetadata('oj-switcher', null, switcherMetadata);
oj.CustomElementBridge.register('oj-switcher', {'metadata': oj.CustomElementBridge.getMetadata('oj-switcher')});

 /**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * @ojcomponent oj.ojTabs
 * @ignore
 * @augments oj.baseComponent
 * @since 0.6
 *
 * @classdesc
 * <h3 id="tabsOverview-section">
 *   JET Tabs Component
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#tabsOverview-section"></a>
 * </h3>
 *
 * <p>Description: Themeable, WAI-ARIA-compliant tabs with mouse and keyboard interactions for navigation.
 *
 * <p>A JET Tabs can be created from a <code class="prettyprint">div</code> element as long as the root element has a tab bar in a <code class="prettyprint">ul</code> with individual tabs in <code class="prettyprint">li</code> preceding the content divs. Each <code class="prettyprint">li</code> will be matched by position with its content in the html DOM.
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;div id="tabs">
 *   &lt;ul>
 *     &lt;li>&lt;span>Tab 1&lt;/span>&lt;/li>
 *     &lt;li>&lt;span>Tab 2&lt;/span>&lt;/li>
 *     &lt;li>&lt;span>Tab 3&lt;/span>&lt;/li>
 *   &lt;/ul>
 *   &lt;div id="tab1">
 *     &lt;p>Tab 1 content&lt;/p>
 *     &lt;p>Tab 1 more content&lt;/p>
 *   &lt;/div>
 *   &lt;div id="tab2">
 *     &lt;p>Tab 2 content&lt;/p>
 *   &lt;/div>
 *   &lt;div id="tab3">
 *     &lt;p>Tab 3 content&lt;/p>
 *   &lt;/div>
 * &lt;/div>
 * </code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Lazy Rendering</h4>
 * <p>If an ojTabs has complex content in a tab, it is recommended to implement lazy
 * rendering of tab content. The application should keep track of which tab contents
 * have been rendered. The initial selected tab should render its content from the
 * start. The application should then listen to the ojbeforeselect event from the
 * ojTabs to control when to render the content.
 *
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 *
 * <p>As with any JET component, in the unusual case that the directionality (LTR or RTL) changes post-init, the tabs must be <code class="prettyprint">refresh()</code>ed.
 *
 * <h3 id="pseudos-section">
 *   Pseudo-selectors
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#pseudos-section"></a>
 * </h3>
 *
 * <p>The <code class="prettyprint">:oj-tabs</code> pseudo-selector can be used in jQuery expressions to select JET Tabs.  For example:
 *
 * <pre class="prettyprint">
 * <code>$( ":oj-tabs" ) // selects all JET Tabs on the page
 * $myEventTarget.closest( ":oj-tabs" ) // selects the closest ancestor that is a JET Tabs
 * </code></pre>
 *
 *
 * <h3 id="jqui2jet-section">
 *   JET for jQuery UI developers
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#jqui2jet-section"></a>
 * </h3>
 *
 * <ol>
 *   <ul>
 *   <li>JQUI Tabs expects the tabs titles either in an ordered or unordered list followed by their content elements. Each tab must have an anchor with the href points to its content element.
 *     <pre class="prettyprint">
 *     <code>
 *       &lt;div id="tabs">
 *         &lt;ul>
 *           &lt;li>&lt;a href="#tabs-1">Tab 1 Title&lt;/a>&lt;/li>
 *           &lt;li>&lt;a href="#tabs-2">Tab 2 Title&lt;/a>&lt;/li>
 *         &lt;/ul>
 *         &lt;div id="tabs-1">
 *           &lt;p>Tab 1 content.&lt;/p>
 *         &lt;/div>
 *         &lt;div id="tabs-2">
 *           &lt;p>Tab 2 content.&lt;/p>
 *           &lt;p>More Tab 2 content.&lt;/p>
 *         &lt;/div
 *       &lt;/div>
 *     </code></pre>
 *   </li>
 *
 *   <li>JET Tabs requires a DOM structures like the JQuery Tabs, except the tab header and their content are matched by position. It requires no anchors and pointers to the contents.
 *     <pre class="prettyprint">
 *     <code>
 *       &lt;div id="tabs">
 *          &lt;ul>
 *            &lt;li>&lt;span>Tab 1&lt;/span>&lt;/li>
 *            &lt;li>&lt;span>Tab 2&lt;/span>&lt;/li>
 *          &lt;/ul>
 *          &lt;div id="tab1">
 *            &lt;p>Tab 1 content&lt;/p>
 *          &lt;/div>
 *          &lt;div id="tab2">
 *            &lt;p>Tab 2 content&lt;/p>
 *          &lt;/div>
 *       &lt;/div>
 *     </code></pre>
 *   </li>
 *   </ul>
 *   <li>JET Tabs supports <code class="prettyprint">edge</code> option: to be placed the tab bar at top(default), bottom, start or end</li>
 *   <li>JET Tabs supports <code class="prettyprint">removable</code> option by adding a close icon to each tab header which when clicked remove the tab from the DOM.</li>
 *   <li>JET Tabs supports <code class="prettyprint">reorderable</code> option allow the tab to be reordered by drag and drop within the Tab bar</li>
 * </ol>
 *
 * <p>Also, event names for all JET components are prefixed with "oj", instead of component-specific prefixes like "tabs".
 *
 * <!-- - - - - Above this point, the tags are for the class.
 *              Below this point, the tags are for the constructor (initializer). - - - - - - -->
 *
 * @desc Creates a JET Tabs.
 * @example <caption>Initialize the tabs with no options specified:</caption>
 * $( ".selector" ).ojTabs();
 *
 * @example <caption>Initialize the tabs with some options specified:</caption>
 * $( ".selector" ).ojTabs( { "edge": "start" } );
 *
 * @example <caption>Initialize the tabs via the JET <code class="prettyprint">ojComponent</code> binding:</caption>
 * &lt;div id="tabs" data-bind="ojComponent: { component: 'ojTabs', edge: 'end'}">
 *
 */
(function () {
  var _CLOSE_ICON = 'oj-tabs-close-icon';
  var _CLOSE_ICON_SIZE = 28;
  var _ID_PREFIX = 'ojtabs-id_';
  var _DELETE_KEY = 46;

  // Context Menu: menu item id's
  var /** @const */ _arMenuCmdMap = {
    cut: 'ojtabscut',
    'paste-before': 'ojtabspastebefore',
    'paste-after': 'ojtabspasteafter',
    remove: 'ojtabsremove'
  };

  // Context Menu translation keys
  var _arMenuKeyMap = {
    cut: 'labelCut',
    'paste-before': 'labelPasteBefore',
    'paste-after': 'labelPasteAfter',
    remove: 'labelRemove'
  };

  oj.__registerWidget('oj.ojTabs', $.oj.baseComponent,
    {
      widgetEventPrefix: 'oj',
      delay: 300,
      options:
      {
      /**
       * The id or zero-based index of the tab that is selected.
       * Note: using zero-based index is deprecated, please use id in the future.<p>
       *
       * Setter value: either an id or index.<p>
       * Getter value: id or index. If the selected tab has a page author provided id then that id is returned, otherwise that tab's index will be returned.
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       * @type {string|number}
       * @default <code class="prettyprint">0</code>
       *
       * @example <caption>Initialize the tabs with the
       * <code class="prettyprint">selected</code> option specified:</caption>
       * $( ".selector" ).ojTabs( { "selected": "myTabDiv" } );
       *
       * @example <caption>Get or set the <code class="prettyprint">selected</code>
       * option after initialization:</caption>
       * // getter
       * var selected = $( ".selector" ).ojTabs( "option", "selected" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "selected", "myTabDiv" );
       */
        selected: 0,

      /**
       * Array contains either ids or zero-based indices of the tabs that should be disabled.
       * Note: using zero-based indices is deprecated, please use ids in the future.<p>
       *
       * Setter value: array of either ids or indices.<p>
       * Getter value: array of either ids or indices. If a disabled tab has a page author provided id then that id is returned, otherwise that tab's index will be returned.
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       * @default <code class="prettyprint">null</code>
       * @type {Array|null}
       *
       * @example [ "disableMe" ] would disable the tab with id="disableMe"
       * @example <caption>Initialize the tabs with the
       * <code class="prettyprint">disabledTabs</code> option specified:</caption>
       * $( ".selector" ).ojTabs( { "disabledTabs": ["disableMe"] } );
       *
       */
        disabledTabs: null,

      /**
       * Truncation option applies to the tab titles when there is not enough room to display
       * all tabs.
       * Valid Values: none, progressive and auto.
       *
       * <ul>
       *  <li> none - tabs always take up the space needed by the title texts. When there
       *   is not enough room, the conveyorBelt's navigation arrows are displayed to allow
       *   the title texts be scrolled within the conveyor.</li>
       *
       *  <li> progressive - If not enough space is available to display all of the tabs,
       *   then the width of each tab title is restricted just enough to allow all tabs to fit.
       *   All tab titles that are truncated are displayed with ellipses. However the width
       *   of each tab title will not be truncated below tabLabelMinWidth. If after all truncation
       *   has been applied, there still is not enough room, then the conveyorBelt's navigation
       *   arrows will appear. When the container of the tabs is resized the truncation will
       *   be reevaluated.</li>
       *
       *  <li> auto - same as "progressive".</li>
       * </ul>
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       * @type {string}
       * @default <code class="prettyprint">auto</code>
       *
       * @example <caption>Get or set the <code class="prettyprint">truncation</code> option for
       *      an ojTabs after initialization:</caption>
       * // getter
       * var truncation = $( ".selector" ).ojTabs( "option", "truncation" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "truncation", "none" );
       */
        truncation: 'auto',

      /**
       * The type of event to select the tab.
       * To select a tab on hover, use "mouseover".
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       * @type {string}
       * @default <code class="prettyprint">"click"</code>
       *
       * @example <caption>Get or set the <code class="prettyprint">selectOn</code> option for
       *      an ojTabs after initialization:</caption>
       * // getter
       * var selectOn = $( ".selector" ).ojTabs( "option", "selectOn" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "selectOn", "mouseover" );
       */
        selectOn: 'click',

      /**
       * The orientation of the tab bar.
       * Valid Values: horizontal and vertical
       *
       * @deprecated Use the <a href="#edge">edge</a> option instead.
       * If the tabs is initialized without an <code class="prettyprint">edge</code> specified,
       * the <code class="prettyprint">orientation</code> value is used to convert to an equivalent <code class="prettyprint">edge</code>:
       * <code class="prettyprint">horizontal</code> to <code class="prettyprint">top</code> and <code class="prettyprint">vertical</code> to <code class="prettyprint">start</code>.
       * @expose
       * @memberof! oj.ojTabs
       * @ignore
       * @instance
       * @type {string}
       * @default <code class="prettyprint">"horizontal"</code>
       *
       * @example <caption>Get or set the <code class="prettyprint">orientation</code> option for
       *      an ojTabs after initialization:</caption>
       * // getter
       * var orientation = $( ".selector" ).ojTabs( "option", "orientation" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "orientation", "vertical" );
       */
        orientation: 'horizontal',

      /**
       * The position of the tab bar.
       * Valid Values: top, bottom, start and end.
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       * @type {string}
       * @default <code class="prettyprint">top</code>
       *
       * @example <caption>Get or set the <code class="prettyprint">edge</code> option for
       *      an ojTabs after initialization:</caption>
       * // getter
       * var edge = $( ".selector" ).ojTabs( "option", "edge" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "edge", "end" );
       */
        edge: 'top',

      /**
       * Specifies if the tabs can be removed.
       * If removable is false, all tabs are not removable.
       * If removable is true, all tabs are removable.<p>
       * If an array is specified, the array contains ids of the tabs that are removable.
       * <p>Please note that this option is only supported if the orientation option is horizontal.
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       * @type {boolean|Array}
       * @default <code class="prettyprint">false</code>
       *
       * @example <caption>Initialize an ojTabs with the <code class="prettyprint">removable</code> option.</caption>
       * $( ".selector" ).ojTabs( { "removable": ["removeMe"] } );
       *
       * @example <caption>Get or set the <code class="prettyprint">removable</code> option for
       *      an ojTabs after initialization:</caption>
       * // getter
       * var removable = $( ".selector" ).ojTabs( "option", "removable" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "removable", true );
       */
        removable: false,

      /**
       * Specifies if the tabs can be reordered within the tab bar by drag-and-drop
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       * @type {boolean}
       * @default <code class="prettyprint">false</code>
       *
       * @example <caption>Get or set the <code class="prettyprint">reorderable</code> option for
       *      an ojTabs after initialization:</caption>
       * // getter
       * var reorderable = $( ".selector" ).ojTabs( "option", "reorderable" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "reorderable", true );
       */
        reorderable: false,

      /**
       * <p>The JET Menu should be initialized before the Tabs using it as a
       * context menu.
       *
       * @ojfragment contextMenuInitOrderDoc - Decomped to fragment so Tabs, Tree, and MasonryLayout can convey their init order restrictions.
       * @memberof oj.ojTabs
       */
      /**
       * {@ojinclude "name":"contextMenuDoc"}
       *
       * <p>When defining a contextMenu, ojTabs will provide built-in behavior for "cut" and "paste"
       *  if the following format for menu &lt;li&gt; item's is used (no &lt;a&gt;
       *  elements are required):
       * <ul>
       *     <li> &lt;li data-oj-command="oj-tabs-cut" /&gt;</li>
       *     <li> &lt;li data-oj-command="oj-tabs-paste-before" /&gt;</li>
       *     <li> &lt;li data-oj-command="oj-tabs-paste-after" /&gt;</li>
       *     <li> &lt;li data-oj-command="oj-tabs-remove" /&gt;</li>
       * </ul>
       * The available translated text will be applied to menu items defined this way.
       *
       * @member
       * @name contextMenu
       * @memberof! oj.ojTabs
       * @instance
       * @type {Element|Array.<Element>|string|jQuery|NodeList}
       * @default <code class="prettyprint">null</code>
       *
       * @example <caption>Initialize a JET Tabs with a context menu:</caption>
       * // via recommended HTML5 syntax:
       * &lt;div id="myTabs" contextmenu="myMenu" data-bind="ojComponent: { ... }>
       *
       * // via JET initializer (less preferred) :
       * $( ".selector" ).ojTabs({ "contextMenu": "#myContextMenu"  ... } });
       *
       * @example <caption>Get or set the <code class="prettyprint">contextMenu</code> option for
       *      an ojTabs after initialization:</caption>
       * // getter
       * var menu = $( ".selector" ).ojTabs( "option", "contextMenu" );
       *
       * // setter
       * $( ".selector" ).ojTabs( "option", "contextMenu", "#myContextMenu" );
       */

      // callbacks
      /**
       * Triggered immediately before a tab is selected.<p>
       * beforeSelect can be canceled to prevent the content from selecting by returning a false in the event listener.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.fromTab The tab being navigated from
       * @property {jQuery} ui.fromContent The content being navigated from
       * @property {jQuery} ui.toTab The tab being navigated to
       * @property {jQuery} ui.toContent The content being navigated to
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">beforeSelect</code> callback specified:</caption>
       * $( ".selector" ).ojTabs({
       *     "beforeSelect": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojbeforeselect</code> event:</caption>
       * $( ".selector" ).on( "ojbeforeselect", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        beforeSelect: null,

      /**
       * Triggered after a tab has been selected.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.fromTab The tab being navigated from
       * @property {jQuery} ui.fromContent The content being navigated from
       * @property {jQuery} ui.toTab The tab being navigated to
       * @property {jQuery} ui.toContent The content being navigated to
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">select</code> callback specified:</caption>
       * $( ".selector" ).ojTabs({
       *     "select": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojselect</code> event:</caption>
       * $( ".selector" ).on( "ojselect", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        select: null,

      /**
       * Triggered immediately before a tab is deselected.<p>
       * beforeDeselect can be canceled to prevent the content from deselecting by returning a false in the event listener.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.fromTab The tab being navigated from
       * @property {jQuery} ui.fromContent The content being navigated from
       * @property {jQuery} ui.toTab The tab being navigated to
       * @property {jQuery} ui.toContent The content being navigated to
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">beforeDeselect</code> callback specified:</caption>
       * $( ".deselector" ).ojTabs({
       *     "beforeDeselect": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojbeforedeselect</code> event:</caption>
       * $( ".selector" ).on( "ojbeforedeselect", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        beforeDeselect: null,

      /**
       * Triggered after a tab has been deselected.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.fromTab The tab being navigated from
       * @property {jQuery} ui.fromContent The content being navigated from
       * @property {jQuery} ui.toTab The tab being navigated to
       * @property {jQuery} ui.toContent The content being navigated to
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">deselect</code> callback specified:</caption>
       * $( ".deselector" ).ojTabs({
       *     "deselect": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojdeselect</code> event:</caption>
       * $( ".selector" ).on( "ojdeselect", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        deselect: null,

      /**
       * Triggered immediately before a tab is removed.
       * beforeRemove can be canceled to prevent the content from removeing by returning a false in the event listener.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.tab The tab that is about to be removed.
       * @property {jQuery} ui.content The content that is about to be removed.
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">beforeRemove</code> callback specified:</caption>
       * $( ".selector" ).ojTabs({
       *     "beforeRemove": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojbeforeremove</code> event:</caption>
       * $( ".selector" ).on( "ojbeforeremove", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        beforeRemove: null,

      /**
       * Triggered after a tab has been removed.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.tab The tab that was just removed.
       * @property {jQuery} ui.content The content that was just removed.
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">remove</code> callback specified:</caption>
       * $( ".selector" ).ojTabs({
       *     "remove": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojremove</code> event:</caption>
       * $( ".selector" ).on( "ojremove", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        remove: null,

      /**
       * Triggered immediately before a tab is reordered.
       * beforeReorder can be canceled to prevent the content from reordering by returning a false in the event listener.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.tab The tab that is about to be reordered.
       * @property {jQuery} ui.content The content that is about to be reordered.
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">beforeReorder</code> callback specified:</caption>
       * $( ".selector" ).ojTabs({
       *     "beforeReorder": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojbeforereorder</code> event:</caption>
       * $( ".selector" ).on( "ojbeforereorder", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        beforeReorder: null,

      /**
       * Triggered after a tab has been reordered.
       *
       * @expose
       * @event
       * @memberof! oj.ojTabs
       * @instance
       * @property {Event} event <code class="prettyprint">jQuery</code> event object
       * @property {Object} ui Parameters
       * @property {jQuery} ui.tab The tab that was just reordered.
       * @property {jQuery} ui.content The content that was just reordered.
       *
       * @example <caption>Initialize the tabs with the <code class="prettyprint">reorder</code> callback specified:</caption>
       * $( ".selector" ).ojTabs({
       *     "reorder": function( event, ui ) {}
       * });
       *
       * @example <caption>Bind an event listener to the <code class="prettyprint">ojreorder</code> event:</caption>
       * $( ".selector" ).on( "ojreorder", function( event, ui ) {
       *      // verify that the component firing the event is a component of interest
       *      if ($(event.target).is(".mySelector")) {}
       * } );
       */
        reorder: null

      },

      _ComponentCreate: function () {
        var options = this.options;

        this._super();
        this.running = false;

      // don't fire events during creation
        this._initialRender = true;

        this._setupEdge(options.edge);

      // flag if an old markup is used
        this._isOldMarkup = (this.element.children('ul').length === 0);

        this._createTabbarFromOldMarkup();
        this._processTabs();

      //  - ojtabs touch support issues
      // only handle context menu if not touch device
        this._menu = {};
        this._menu.usermenu = false;
        this._menu.$menuItems = [];
        this._menu.$container = null;
        this._menu.$elemPasteBefore = null;
        this._menu.$elemPasteAfter = null;

        this._initMenu();

      // update disabledTabs to contains ids if provided, index otherwise
//      this._updateDisabledTabs();

        this._refresh();

      //  - for tabs w/ pre 1.1 markup, selected & disabledtabs options doesn't work w/ ids
        var selectedIndex = this._getIndexByTabOrContentId(options.selected);
        if (selectedIndex === undefined) {
          selectedIndex = 0;
        }

      //  - tabs shouldn't implement _init()
        this._fireSelectEvents(selectedIndex);

        this._initialRender = undefined;
      },

      // TODO: selectedIndex
      _fireSelectEvents: function (_selectedIndex, flags) {
        //  - When ojtab first displays beforeSelect & select events are not raised
        //  - ojtabs should not let user set focus on disabled tabs
        // if the selected tab is disabled or undefined, select the next enabled tab
        var selectedIndex = _selectedIndex;
        if (this._isTabDisabled(selectedIndex)) {
          selectedIndex = this._getNextEnabledTab(selectedIndex);
        }
        this._activate((selectedIndex === undefined) ? undefined : $(this.tabs[selectedIndex]),
                       flags);
      },

    /**
      *  Notification that the user has invoked the context menu via the default
      *  gestures: right-click, pressHold, and Shift-F10.
      *  @param {Object} menu       The JET Menu to open as a context menu
      *  @param {Event}  event      Triggering event
      *  @param {string} eventType  Triggering event type (e.g. 'keyboard', 'mouse', 'touch')
      *  @private
      */
      _NotifyContextMenuGesture: function (menu, event, eventType) {
      //  - ojtabs:context menu shouldn't open - right mouse is clicked inside content area
      //  - contextmenu issues: presshold should launch the contextmenu on touch devices
        if (this._isInATab(event.target) && !(this._touchProxy && this._touchProxy.touchMoved())) {
          var keyboard = (eventType === 'keyboard');

          if ((event.type !== 'contextmenu') && (!keyboard) && (eventType !== 'touch')) {
            return;
          }

        // got here either by right mouse click (event.which == 3)
        // or <shift F10> key (event.keyCode = 121 and event.shiftKey = true)
        // or pressHold

        // get the tab acted on
          var tab = $(event.target).closest('li');
          var openOptions = { launcher: tab };
          this._menu.tab = keyboard ? this.active : openOptions.launcher;
          if (!this._menu.tab) {
            event.preventDefault();
            return;
          }

          if (this._menu.$elemRemove) {
          // if tab is disable or not removable and if "remove" is only item,  dont display context menu
            var isDisabled = tab.hasClass('oj-disabled');
            if ((isDisabled || this._getCloseIcons(tab).length === 0) &&
              this._menu.$container.children().length === 1) {
              event.preventDefault();
              return;
            }
            if (isDisabled || !this._isTabRemovable(tab)) {
              this._menu.$elemRemove.addClass('oj-disabled');
            } else {
              this._menu.$elemRemove.removeClass('oj-disabled');
            }
          }

        // Set menu "Paste" disable state, depending on whether there's been a
        // previous "cut"
          if (this._menu.$elemPasteBefore || this._menu.$elemPasteAfter) {
            var disabledState = (!this._menu.cutTab);
            var state = this._menu.$elemPasteBefore.hasClass('oj-disabled');

            if (state !== disabledState) {
              if (disabledState) {
                this._menu.$elemPasteBefore.addClass('oj-disabled');
                this._menu.$elemPasteAfter.addClass('oj-disabled');
              } else {
                this._menu.$elemPasteBefore.removeClass('oj-disabled');
                this._menu.$elemPasteAfter.removeClass('oj-disabled');
              }
              this._menu.$container.ojMenu('refresh');
            }
          }

          this._OpenContextMenu(event, eventType, openOptions);
        }
      },

      _tabKeydown: function (event) {
      /* jshint maxcomplexity:15*/
        if (this._handlePageNav(event)) {
          return;
        }

        var focusedTab = $(this.document[0].activeElement).closest('li');
        var enabledTabs = this._getEnabledTabs();
        var selectedIndex = enabledTabs.index(focusedTab);
        var length = enabledTabs.length;

        switch (event.keyCode) {
          case $.ui.keyCode.RIGHT:
          case $.ui.keyCode.DOWN:
            selectedIndex = (selectedIndex + 1) % length;

            break;
          case $.ui.keyCode.UP:
          case $.ui.keyCode.LEFT:
            selectedIndex = ((selectedIndex === 0 ? length : selectedIndex) - 1) % length;
            break;
          case $.ui.keyCode.END:
            selectedIndex = length - 1;
            break;
          case $.ui.keyCode.HOME:
            selectedIndex = 0;
            break;

          case _DELETE_KEY:
            var tab = this.active;

          // James: remove tab keystroke doesn't seem to work with JAWS.
          // ALT+DEL seems to conflict with a JAWS keystroke. I have raised an issue on the Authoring
          // Practices for this. Could we just use Delete as well or does that sound like a bad idea?
          // simulate a click on the close icon of the current selected header
            if (tab && this._getCloseIcons(tab).length > 0) {
              event.preventDefault();
              this._removeTab(tab, null, event);
            }
            return;

          default :
            return;
        }

      // Focus the appropriate tab, based on which key was pressed
        event.preventDefault();
        clearTimeout(this.activating);

        var selTab = $(enabledTabs[selectedIndex]);
        selTab.focus();

      // Navigating with control key will prevent automatic activation
        if (!event.ctrlKey) {
        // Update aria-selected immediately so that AT think the tab is already selected.
        // Otherwise AT may confuse the user by stating that they need to select the tab,
        // but the tab will already be selected by the time the announcement finishes.
          focusedTab.attr('aria-selected', 'false');
          selTab.attr('aria-selected', 'true');

          var self = this;
          this.activating = this._delay(function () {
          // qunit test failure
            if (!self || !self.tabs) {
              return;
            }

            self._activate(selTab, event);
          },
        this.delay);
        }
      },

      _panelKeydown: function (event) {
      // In the nested tabs case, handle keydown for the closest tabs only
        if ($(event.target).closest('.oj-tabs-selected').attr('id') !==
            this.element.children('.oj-tabs-selected').attr('id')) {
          return;
        }

        if (this._handlePageNav(event)) {
          return;
        }

      // Ctrl+up moves focus to the current tab
        if (event.ctrlKey && event.keyCode === $.ui.keyCode.UP) {
          event.preventDefault();
          this.active.focus();
        }
      },

      // Ctrl+page up/down moves focus to the previous/next tab (and selects)
      _handlePageNav: function (event) {
        var selectedIndex = this._getSelectedIndex();

        if (event.ctrlKey && event.keyCode === $.ui.keyCode.PAGE_UP) {
          this._activate(this._focusNextTab(selectedIndex, false));
          return true;
        }
        if (event.ctrlKey && event.keyCode === $.ui.keyCode.PAGE_DOWN) {
          this._activate(this._focusNextTab(selectedIndex, true));
          return true;
        }

        return false;
      },

      _isTabDisabled: function (index) {
        if (index >= 0 && index < this.tabs.length) {
          return $(this.tabs[index]).hasClass('oj-disabled');
        }
        return false;
      },

    // TODO: return tab
      _focusNextTab: function (index, goingForward) {
        var enabledTabs = this._getEnabledTabs();
        var currentIndex = enabledTabs.index(this.tabs[index]);
        var length = enabledTabs.length;

        if (goingForward) {
          currentIndex = (currentIndex + 1) % length;
        } else {
          currentIndex = ((currentIndex === 0 ? length : currentIndex) - 1) % length;
        }

        var selTab = $(enabledTabs[currentIndex]);
        selTab.focus();
        return selTab;
      },

    // return index
    // TODO: index: tabIndex (not panelId)
      _getNextEnabledTab: function (index) {
        var next = index + 1;
        var lastTabIndex = this.tabs.length - 1;

        while (next <= lastTabIndex) {
          if (!this._isTabDisabled(next)) {
            return next;
          }
          next += 1;
        }
        next = index - 1;
        while (next >= 0) {
          if (!this._isTabDisabled(next)) {
            return next;
          }
          next -= 1;
        }
        return undefined;
      },

      _isHorizontal: function () {
        return (this.options.edge === 'top' || this.options.edge === 'bottom');
      },

      _setOption: function (key, _value, flags) {
        var value = _value;
        if (key === 'selected') {
          value = this._getTab(value);
          if (value === undefined) {
            return;
          }

        // _activate() will update this.options
          this._activate(value);
          return;
        }

        if (key === 'disabledTabs') {
          if (value === null) {
            value = [];
          }

          if (Array.isArray(value)) {
            this._setOjDisabledOnTab(value);

            // in case the selected tab is disabled
            var selectedIndex = this._getSelectedIndex();
            if (this._isTabDisabled(selectedIndex)) { this._fireSelectEvents(selectedIndex); }

            //  - when observable is used to update disabled tabs, tabs do not function right
            this.refresh();
          }
          return;
        }

        if (key === 'removable' || key === 'truncation') {
          if (value !== this.options[key]) {
            this._super(key, value, flags);
            this.refresh();
          }
          return;
        }

      // allow drag and drop a tab within the tab bar
        if (key === 'reorderable') {
          if (value !== this.options.reorderable) {
            this._super(key, value, flags);
//          this._setupReorder();

          //  - if reorderable option is changed to true programmatically,
          // it doesn't take affect
            this.refresh();
          }
          return;
        }

      // change orientation need refresh
        if (key === 'orientation') {
          if (!this._edgeSpecified) {
            var edge = this._orientationToEdge(value);
            if (edge && this._setupEdge(edge)) {
              this.refresh();
            }
          }
          return;
        }

      // change edge need refresh
        if (key === 'edge') {
          if (this._setupEdge(value)) {
            this._edgeSpecified = true;
            this._super(key, value, flags);
            this.refresh();
          }
          return;
        }

        if (key === 'selectOn') {
          this._tearDownEvents(true);
          this._super(key, value, flags);
          this._setupEvents();
          return;
        }

        if (key === 'contextMenu') {
          this._clearMenu();
          if (value) {
            this._initMenu(value);
          }
        }

        this._super(key, value, flags);

      //  - specifying 'translations.removecuetext' as option does not set the aria-label
      // problem is setOption is called after aria-label already set
        if (key === 'translations' && flags && flags.subkey === 'removeCueText' && this.tablist) {
          this._getCloseIcons(this.tablist).attr('aria-label',
                                               value ? value.removeCueText : '');
        }
      },

    /**
     * Refreshes the visual state of the tabs. JET components require a <code class="prettyprint">refresh()</code> or re-init after the DOM is
     * programmatically changed underneath the component.
     *
     * <p>This method does not accept any arguments.
     *
     * @expose
     * @memberof! oj.ojTabs
     * @instance
     *
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * $( ".selector" ).ojTabs( "refresh" );
     */
      refresh: function () {
        this._super();

        //  - open/close of tabs leaks sortable
//      this._destroyCloseIcons();
        this._processTabs();

        this._refresh();

      // set selected if needed
        if (!this.element.children('.oj-tabs-selected').length) {
          var selected = this._getNextEnabledTab(-1);
          if (selected >= 0) {
            this._fireSelectEvents(selected);
          }
        }
      },

      _refresh: function () {
        // check for length avoids error when initializing empty list
        var selectedPanel = this.element.children('.oj-tabs-selected');
        if (selectedPanel.length) {
          this.active = this.tablist.children('.oj-selected');
        } else {
          this.active = $();
        }

        this._createCloseIcons();

        //  - open/close of tabs leaks sortable
        this._setupEvents();

        this.tabs.not(this.active).attr(
          {
            'aria-selected': 'false',
            tabIndex: '-1'
          });
        this.panels.not(this._getPanelForTab(this.active)).hide().attr(
          {
            'aria-expanded': 'false',
            'aria-hidden': 'true'
          });

        // Make sure one tab is in the tab order
        if (this.active.length) {
          this.active.addClass('oj-selected').attr(
            {
              'aria-selected': 'true',
              tabIndex: '0'
            });

          selectedPanel.show()
          .attr('aria-expanded', 'true')
          .removeAttr('aria-hidden');
        } else {
          $(this.tabs[0]).attr('tabIndex', '0');
        }

        if (this._isHorizontal()) {
          if (this.tabs.length > 0) {
            this._truncateBeforeOverflow();
            this._addConveyor();
          } else {
            // no tab in the tab bar
            this._getTabbarWrapper();
          }

          //  - er for placing buttons alongside tabs
          this._addFacets();
        }

        if (this.options.selected === undefined ||
            this._getSelectedIndex() === 0) {
          this.element.addClass('oj-first-child-selected');
        } else {
          this.element.removeClass('oj-first-child-selected');
        }

      //  - After a tab is deleted, reorder does not work
        this._setupReorder();
      },

      _addConveyor: function () {
        var tabsId = this.tablist.uniqueId().attr('id');
        var conveyorDiv = this._getTabbarWrapper().parent();

        this.conveyor = conveyorDiv.ojConveyorBelt(
          {
            orientation: 'horizontal',
            contentParent: '#' + tabsId
          }).attr('data-oj-internal', ''); // mark internal component, used in oj.Components.getComponentElementByNode

        var cparent = this.conveyor.parent();
        if (cparent.hasClass('oj-tabs-conveyorbelt-wrapper')) {
          var flex = '0 1 ' + this._getConveyorWrapperMaxWidth() + 'px';
          cparent.css('flex', flex);
          cparent.css('-webkit-flex', flex);
          cparent.css('-ms-flex', flex);
        }
      },

      _processTabs: function () {
        var self = this;
        var edge = this.options.edge;

        // destroy the old tab bar before creating a new one
        this._destroyTabBar();

        //  - use case where gauges are displayed in tabs title bar does not work
        // create a tab bar
        this._createTabbar();

        // find the tab bar
        this.tablist = this.element.children('ul')
        .addClass('oj-tabs-nav oj-helper-clearfix')
        .attr('role', 'tablist');

        var tabbarIndex = this.tablist.index();
        var startBtn;
        var trailingBtn;

        //  - er for placing buttons alongside tabs
        // add oj-start class to the facets before the tab bar
        this.element.children('.oj-tabs-facet').each(
        function () {
          var facet = $(this);
          if (facet.index() < tabbarIndex) {
            facet.addClass('oj-start');
            startBtn = facet;
          } else {
            if (!trailingBtn) {
              trailingBtn = facet;
            }
            facet.removeClass('oj-start');
          }
        });

        //  - ojtab removing first tab
        // don't insert <ul> between <!-- ko foreach: tabs --> and <!-- /ko -->
        if (edge === 'start' || edge === 'top') {
          // add <ul> after the start button or as the first child of the container
          if (startBtn) {
            startBtn.after(this.tablist); // @HTMLUpdateOK
          } else {
            this.element.prepend(this.tablist); // @HTMLUpdateOK
          }
        } else {
          // add <ul> before the trailing button or as the last child of the container
          if (trailingBtn) { // eslint-disable-line no-lonely-if
            trailingBtn.before(this.tablist); // @HTMLUpdateOK
          } else {
            this.element.append(this.tablist); // @HTMLUpdateOK
          }
        }

      // list of tabs
        this.tabs = $();
      // list of contents
        this.panels = $();

      //  - tabs lose selection when moved (sortable tabs)
        this.tablist.children('li').each(function () {
          var tab = $(this)
          .addClass('oj-tabs-tab')
          .attr({
            role: 'tab',
            tabIndex: '-1'
          })
          .removeAttr('aria-hidden');

          var div = tab.children();
          div.addClass('oj-tabs-tab-content');

          var anchor = div.children();
          anchor.addClass('oj-tabs-anchor')
            .attr({
              role: 'presentation',
              tabIndex: '-1'
            });

          anchor.children()
            .addClass('oj-tabs-title')
            .removeAttr('aria-hidden');

          self.tabs = self.tabs.add(tab);

          var anchorId = anchor.uniqueId().attr('id');
          var panelId = tab.attr('data-content');
          var panel = self.element.find(self._sanitizeIdSelector(panelId));

/*
        if (tab.hasClass("oj-disabled"))
        {
          tab.attr("aria-disabled", "true");
          // - clicking on disabled tab takes user to the top of the page
          anchor.removeAttr("href");
        }
*/
          tab.attr(
            {
              'aria-controls': panelId,
              'aria-labelledby': anchorId
            });

          self.panels = self.panels.add(panel);

          panel.attr('aria-labelledby', anchorId)
             .attr('role', 'tabpanel');
        });

        if (edge === 'start' || edge === 'end') {
        // In vertical, "oj-tabs-nav-root" is on the same element as the element with "oj-tabs-nav"
        // as long as the theming doesn't care whether the root is on the same element or not
          self.tablist.addClass('oj-tabs-nav-root');
        }

      // set oj-disabled on the tabs
        if (this._initialRender) {
          this._setOjDisabledOnTab(this.options.disabledTabs);
        }
      },

      _setupEvents: function () {
        var self = this;

        var events =
          {
            keydown: this._tabKeydown
          };

        var event = this.options.selectOn;
        var selectOnClick = false;
        if (event) {
          $.each(event.split(' '), function (index, eventName) {
          //  - mousedown is a better default for jtab selecton than click
            if (eventName === 'click') {
              selectOnClick = true;
            }

          // security test
            if (oj.DomUtils.isValidIdentifier(eventName)) {
              events[eventName] = self._eventHandler;
            }
          });
        }

        var enabledTabs = this._getEnabledTabs();

        this._on(enabledTabs, events);

      //  - mousedown is a better default for jtab selecton than click
        if (!selectOnClick) {
          this._on(enabledTabs,
            {
              // prevent the default action
              click: function (e) {
                e.preventDefault();
              }
            });
        }

      //  - ojtabs: call jquery "on"/"off" methods instead of jquery ui "_on"/"_off"
        this.panels.on('keydown' + this.eventNamespace, this._panelKeydown.bind(this));

      // add listeners on close icon
        if (this.options.removable) {
          var revents =
            {
              click: this._removeTabHandler
            };
          var closeIcons = this._getCloseIcons(enabledTabs);
          this._on(closeIcons, revents);
          this._AddHoverable(closeIcons);
          this._AddActiveable(closeIcons);
        }

        this._focusable({
          element: enabledTabs,
          applyHighlight: true
        });

        this._AddHoverable(enabledTabs);
        this._AddActiveable({
          element: enabledTabs,
          afterToggle: function (eventtype) {
          // after mousedown, clear the "oj-focus-highlight" class
            if (eventtype === 'mousedown') {
              enabledTabs.filter('.oj-focus-highlight').blur();
            }
          }
        });
      },

      _tearDownEvents: function (noCloseIcon) {
        var enabledTabs = this._getEnabledTabs();
      //  - open/close of tabs leaks sortable
      //  - ojtabs: call jquery "on"/"off" methods instead of jquery ui "_on"/"_off"
        if (!noCloseIcon) {
          this._UnregisterChildNode(this._getCloseIcons(enabledTabs));
        }

        this._UnregisterChildNode(enabledTabs);

        if (this.panels) {
          this.panels.off('keydown' + this.eventNamespace);
        }
      },

      _eventHandler: function (event, flags) {
        var active = this.active;
        var tab = $(event.currentTarget).closest('li');
        var clickedIsActive = (active && tab[0] === active[0]);
        var oToContent = this._getPanelForTab(tab);
        var oFromContent = (active && active.length) ? this._getPanelForTab(active) : $();

        var eventData =
          {
            /** @expose */
            fromTab: active,
            /** @expose */
            fromContent: oFromContent,
            /** @expose */
            toTab: tab,
            /** @expose */
            toContent: oToContent
          };

        event.preventDefault();

        var oEvent = flags || event;
        if (tab.hasClass('oj-disabled') ||

          // can't switch durning an animation
          this.running ||
          // click on active header,
          clickedIsActive ||

          // allow canceling deselect
          //  - ojbeforedeseselect & ojdeselect events shouldn't be raised
          // when tab is rendered initially
          (oFromContent && oFromContent.length &&
           this._trigger('beforeDeselect', oEvent, eventData) === false) ||

          // allow canceling select
          // don't fire initial option change events
          (!this._initialRender && this._trigger('beforeSelect', oEvent, eventData) === false)) {
          return;
        }

      //  - tabs shouldn't implement _init()
      // don't fire initial option change events
        if (this._initialRender) {
          this.options.selected = this._getTabIdOrIndex(tab);
        } else {
          this.option('selected', this._getTabIdOrIndex(tab),
                    { _context: { originalEvent: oEvent, internalSet: true } });
        }

        this.active = tab;

      // add or remove oj-selected on the corresponding panel
        oToContent.addClass('oj-tabs-selected');
        if (oFromContent && oFromContent.length) {
          oFromContent.removeClass('oj-tabs-selected');
        }

        this._toggle(oEvent, eventData);
      },

    // handles show/hide for selecting tabs
      _toggle: function (event, eventData) {
        var self = this;
        var toTab = eventData.toTab.closest('li');
        var toShow = eventData.toContent;
        var toHide = eventData.fromContent;

        this.running = true;

        function complete() {
          self.running = false;

        //  - ojbeforedeseselect & ojdeselect events shouldn't be raised
        // when tab is rendered initially
        // don't fire select event during creation
          if (!self._initialRender) {
            self._trigger('deselect', event, eventData);
            self._trigger('select', event, eventData);
          }
        }

        function show() {
          toTab.addClass('oj-selected');
          if (self._isHorizontal() && self.tabs.length > 0) {
            if (toTab.index() === 0) {
              self.element.addClass('oj-first-child-selected');
            } else {
              self.element.removeClass('oj-first-child-selected');
            }
          }

          toShow.show();
        //  - unable to stretch ojtable in an initially hidden ojtab
          if (toShow.length > 0) {
            oj.Components.subtreeShown(toShow[0], { initialRender: self._initialRender });
          }
          complete();
        }

      // start out by hiding, then showing, then completing
        var fromTab = eventData.fromTab.closest('li')
        .removeClass('oj-selected');

        toHide.hide();
      //  - unable to stretch ojtable in an initially hidden ojtab
        if (toHide.length > 0) {
          oj.Components.subtreeHidden(toHide[0]);
        }
        show();

        toHide.attr(
          {
            'aria-expanded': 'false',
            'aria-hidden': 'true'
          });

        fromTab.attr('aria-selected', 'false');

      // If we're switching tabs, remove the old tab from the tab order.
      // If we're opening from collapsed state, remove the previous tab from the tab order.
        if (toShow.length && toHide.length) {
          fromTab.attr('tabIndex', '-1');
        } else if (toShow.length) {
          this.tabs.filter(function () {
            return $(this).attr('tabIndex') === '0';
          })
          .attr('tabIndex', '-1');
        }

        toShow.attr('aria-expanded', 'true')
        .removeAttr('aria-hidden');

        toTab.attr(
          {
            'aria-selected': 'true',
            tabIndex: '0'
          });
      },

    // TODO: active: active tab, not panelId
      _activate: function (active, flags) {
        if (active === undefined) {
          return;
        }

        var anchor;

      // trying to activate the already active panel
        if (this.active && (active[0] === this.active[0])) {
          return;
        }

      //  - ojtabs should not let user set focus on disabled tabs
      // this.options.selected = panelId;

      // simulate a click on the new active header
        anchor = active.find('.oj-tabs-anchor')[0];
        this._eventHandler(
          {
            target: anchor,
            currentTarget: anchor,
            preventDefault: $.noop
          }, flags);
      },

      _createCloseIcons: function () {
        var removable = this.options.removable;

      //  - for vertical tabs, removable option should be ignored
      // create close icon only if it's horizontal and  not disabled
        if (removable && this._isHorizontal()) {
          var removeCueText = this.getTranslatedString('removeCueText');
          var self = this;

          this._getEnabledTabs().each(function (index) {
            if (self._isTabRemovable($(this))) {
              var div = $(this).find('> :first-child');
              div.addClass('oj-removable');

            // add cue text for removable icon for screen reader users
              var rmId = _ID_PREFIX + 'rm_' + index;
              $(this).attr('aria-describedby', rmId);

              $("<a href='#'>")
              .addClass('oj-tabs-icon oj-component-icon oj-clickable-icon-nocontext ' + _CLOSE_ICON)
              .attr(
                {
                  id: rmId,
                  tabIndex: '-1',
                  'aria-label': removeCueText,
                  role: 'presentation'
                })
              .appendTo(div); // @HTMLUpdateOK
            }
          });
        }
      },

      _getEnabledTabs: function () {
        return $(this.tabs.not('.oj-disabled'));
      },

      _getCloseIcons: function (elem) {
        return elem.find('.' + _CLOSE_ICON);
      },

      _destroyTabBar: function () {
      //  - open/close of tabs leaks sortable
        if (this.tabs) {
        // NOTE: also remove panel event handler
          this._tearDownEvents();
        }

      // remove listener
        if (this._hasResizeListener) {
          oj.DomUtils.removeResizeListener(this.element[0], this._resizeHandler);
          this._hasResizeListener = false;
          this._originalWidth = undefined;
        }

      //  - open/close of tabs leaks sortable
        if (this._sortable) {
        // remove touchListeners
          if (oj.DomUtils.isTouchSupported()) {
            this._tearDownTouchReorder();
          }

        // destroy sortable
          if (this.tablist.sortable('instance')) {
            this.tablist.sortable('destroy');
          }
          this._sortable = undefined;
        }

        var navRoot = this.element.children('.oj-tabs-nav-root');
        var navRootNotUL = !navRoot.hasClass('oj-tabs-nav');

      //  - open/close of tabs leaks sortable
        if (this.tabs) {
          this._getCloseIcons(this.tabs).remove();

        //  - er for placing buttons alongside tabs
          if (navRoot.length) {
            var afterMe = navRoot;
            var noFacet = true;
            var tabbar = this.tablist;

            navRoot.children().each(
            function () {
              var child = $(this);
              if (child.hasClass('oj-tabs-conveyorbelt-wrapper')) {
                child = tabbar;
              } else if (!child.hasClass('oj-tabs-facet')) {
                return;
              }

              noFacet = false;
              afterMe.after(child); // @HTMLUpdateOK
              afterMe = child;
            });

          // move ul and lis
            if (noFacet && navRootNotUL) {
              navRoot.after(tabbar); // @HTMLUpdateOK
            }
          }
        }
        if (this.conveyor) {
          this.conveyor.ojConveyorBelt('destroy');
          this.conveyor.remove();
          this.conveyor = null;
          this.active = null;
        }

        if (navRootNotUL) {
          navRoot.remove();
        }

        this.tabs = null;
        this.tablist = null;
      },

      _destroy: function () {
        this._clearMenu();

        var verticalEdge = !this._isHorizontal();
        if (verticalEdge) {
          this.element.removeClass('oj-tabs oj-component oj-tabs-vertical oj-tabs-start oj-tabs-end oj-helper-clearfix');
        } else {
          this.element.removeClass('oj-tabs oj-component oj-tabs-horizontal oj-tabs-top oj-tabs-bottom oj-first-child-selected');
        }

        var self = this;
        var tab;
        var div;
        var anchor;
        var header;

        this.tabs.each(function (index) {
          tab = $(this);

          tab.removeAttr('tabIndex')
          .removeAttr('aria-selected')
          .removeAttr('aria-labelledby')
          .removeAttr('aria-hidden')
          .removeAttr('aria-controls')
          .removeAttr('aria-disabled')
          .removeAttr('aria-describedby')
          .removeAttr('role')
          .removeAttr('data-content')
          .removeClass('oj-active oj-disabled oj-selected oj-tabs-gen-id oj-tabs-tab')
          .removeUniqueId()
          .css('display', '');

          div = tab.children('div')
          .removeClass('oj-tabs-tab-content');

          anchor = div.children('a')
          .removeClass('oj-tabs-anchor')
          .removeAttr('tabIndex')
          .removeAttr('role');

          header = anchor.children();
          header.removeClass('oj-tabs-title')
          .removeAttr('aria-hidden');

        // if old markup is used, move header back to its own div
          if (self._isOldMarkup) {
            header.prependTo(self.panels.get(index)); // @HTMLUpdateOK
          } else if (div.hasClass('oj-tabs-gen-div') && anchor.hasClass('oj-tabs-gen-a')) {
            header.prependTo(tab); // @HTMLUpdateOK
            div.remove();
          } else if (anchor.hasClass('oj-tabs-gen-a')) {
            header.prependTo(div); // @HTMLUpdateOK
            anchor.remove();
          } else if (div.hasClass('oj-tabs-gen-div')) {
            anchor.prependTo(tab); // @HTMLUpdateOK
            div.remove();
          }
        });

        var tabbar = this.tablist;
        this._destroyTabBar();
        tabbar.removeAttr('tabIndex')
          .removeAttr('role')
          .removeClass('oj-tabs-nav oj-tabs-nav-root oj-helper-clearfix')
          .removeUniqueId();

      // remove ul and lis
        if (this._isOldMarkup) {
          tabbar.remove();
        }

        this.panels.each(function () {
          var panel = $(this);

          panel.removeAttr('tabIndex')
          .removeAttr('aria-expanded')
          .removeAttr('aria-selected')
          .removeAttr('aria-labelledby')
          .removeAttr('aria-hidden')
          .removeAttr('role')
          .removeClass('oj-active oj-tabs-selected oj-tabs-gen-id oj-tabs-panel')
          .removeUniqueId()
          .css('display', '');
        });

        this.element.children('.oj-tabs-facet').removeClass('oj-start');
      },

      _isTabRemovable: function (tab) {
        var removable = this.options.removable;
        return (removable &&
              (!Array.isArray(removable) || removable.indexOf(tab.attr('id')) > -1));
      },

      _removeTab: function (tab, event, flags) {
        if (!this._isTabRemovable(tab)) {
          return;
        }

        var panel = this._getPanelForTab(tab);
        var eventData =
          {
            /** @expose */
            tab: tab,
            /** @expose */
            content: panel
          };

        var oEvent = flags || event ? event :
        {
          target: tab,
          currentTarget: tab,
          preventDefault: $.noop
        };

        // trigger before delete event and only delete if it's not cancelled
        if (tab && this._trigger('beforeRemove', oEvent, eventData) !== false) {
          //  - deleteing a tab scrolls page back to the top
          if (event) {
            event.preventDefault();
          }

          // if tab to be removed is selected, select the next enabled tab
          if (tab.hasClass('oj-selected')) {
            var curIndex = this.tabs.index(tab);
            var nextIndex = this._getNextEnabledTab(curIndex);

            // no enabled tabs left
            if (nextIndex === undefined) {
              this.active = undefined;
              this.option('selected', undefined,
                        { _context: { originalEvent: oEvent, internalSet: true } });
            } else {
              // fire selected event
              this._fireSelectEvents(nextIndex, oEvent);

              // adjust index
              var idOrIndex = this._getTabIdOrIndex($(this.tabs[nextIndex]));
              if (typeof idOrIndex === 'number' && idOrIndex > curIndex) {
                idOrIndex -= 1;
              }

            // fire option change event
              this.option('selected', idOrIndex,
                        { _context: { originalEvent: oEvent, internalSet: true } });
            }
          }

          //  - open/close of tabs leaks sortable
          this._tearDownEvents();
          this.tabs = this.tabs.not(tab);
          tab.remove();
          panel.remove();

          //  - If user closes tab then index for tabs changes and disabled tab becomes enable
          // update indices of disabled tabs
          this._updateDisabledTabs();

          this.refresh();

          //  - tab area grabs focus when tabs are closed
          // set focus on the active
//        if (this.active)
//          this.active.focus();

          this._trigger('remove', oEvent, eventData);
        }
      },

      _removeTabHandler: function (event, flags) {
        this._removeTab($(event.currentTarget).closest('li'), event, flags);
      },

    /**
     * Remove a tab from the tab bar.
     * <p>Please note that the tab must be <a href="#removable">removable</a>
     *
     * @expose
     * @memberof! oj.ojTabs
     * @instance
     *
     * @param {string} tabId The id of the tab that is to be removed.
     *
     * @example <caption>Invoke the <code class="prettyprint">removeTab</code> method:</caption>
     * $( ".selector" ).ojTabs( "removeTab", "removeMe" );
     *
     */
      removeTab: function (tabId) {
        if (typeof tabId !== 'string') {
          throw new Error("'" + tabId + "' is not a tab id");
        }

        this._removeTab(this._getTab(tabId), null, null);
      },

      _wrapLi: function (header, contentId) {
        var tab;
        if (header.prop('tagName').toLowerCase() === 'li') {
          tab = header;
          var div = tab.children('div');
          if (div.length === 1 && div.hasClass('oj-tabs-tab-content')) {
            var anchor = div.children('a');
            if (anchor.length !== 1) {
              tab.wrapInner("<a href='#'></a>"); // @HTMLUpdateOK
              anchor.addClass('oj-tabs-gen-a');
            }
          } else {
            div = tab.wrapInner("<div><a href='#'></a></div>") // @HTMLUpdateOK
            .children();

            div.addClass('oj-tabs-gen-div oj-tabs-tab-content');
            div.children().addClass('oj-tabs-gen-a');
          }
        } else {
          tab = header
          .wrap("<li><div><a href='#'></a></div></li>") // @HTMLUpdateOK
          .parent().parent().parent();

          tab.addClass('oj-tabs-gen-li');
          tab.children().addClass('oj-tabs-gen-div oj-tabs-tab-content');
        }

      // set content id on tab
        if (contentId) {
          tab.attr('data-content', contentId);
        }

        return tab;
      },

      /**
       * Add a tab to the end of the tabs
       *
       * @expose
       * @memberof! oj.ojTabs
       * @instance
       *
       * @param {Object} newTab An Object contains the properties in the following table.
       * @property {jQuery} newTab.tab The new tab
       * @property {jQuery} newTab.content The new content
       * @property {number} newTab.index The index of new tab. Default is -1, newTab is appended to the end
       *
       * @example <caption>Invoke the <code class="prettyprint">addTab</code> method:</caption>
       * $( ".selector" ).ojTabs( "addTab",
       *                          {
       *                            "tab" : $("&lt;h3>New Tab&lt;/h3>"),
       *                            "content" : $("&lt;div>&lt;p>Content of New Tab&lt;/p>&lt;/div>"),
       *                            "index" : 2
       *                          } );
       *
       * Please note that single jQuery object parameter is deprecated
       * $( ".selector" ).ojTabs( "addTab", $("&lt;div>&lt;h3>New Tab&lt;/h3>&lt;p>Content of New Tab&lt;/p>&lt;/div>") );
       */
      addTab: function (newTab) {
        var tab;
        var content;
        var index = -1;

        // check if it is a property bag
        if (newTab.tab && newTab.content) {
          content = newTab.content;
          tab = this._wrapLi(newTab.tab, this._getUniqueId(content));
          if (newTab.index !== undefined) {
            index = newTab.index;
          }
        } else {
          // Backward compatible: move header to tab bar
          content = newTab;
          tab = this._wrapLi($(newTab).find('> :first-child'),
                           this._getUniqueId(content));
        }

        var navRoot = this.element.children('.oj-tabs-nav-root');
        var tabbar;
        // vertical tabs
        if (navRoot.hasClass('oj-tabs-nav')) {
          tabbar = navRoot;
        } else {
          tabbar = navRoot.length ? navRoot.find('.oj-tabs-nav') :
        this.element.children('.oj-tabs-nav');
        }

        var isNew = false;
        if (!tabbar.length) {
          tabbar = $('<ul>');
          isNew = true;
        }
        if (isNew) {
          tabbar.prependTo(this.element); // @HTMLUpdateOK
        }

        // insertion inside
        if (index >= 0 && index < tabbar.children().length) {
          var tabAfter = tabbar.children(':eq(' + index + ')');
          var contentAfter = this.element.children(
          this._sanitizeIdSelector(tabAfter.attr('aria-controls')));

          tabAfter.before(tab); // @HTMLUpdateOK
          contentAfter.before(content); // @HTMLUpdateOK
        } else {
          // insertion at the end
          tab.appendTo(tabbar); // @HTMLUpdateOK
          content.appendTo(this.element); // @HTMLUpdateOK
        }

        this.refresh();

      // if no active selected tab, select the new tab
        if (this.active.length === 0) {
          this._fireSelectEvents(0);
        }

      //  - overflow icons do not appear correctly
      //  - new tabs don't get focus
      //  - tabs flash when adding
        tab[0].scrollIntoView(false);
      },

      _setupTouchReorder: function () {
        this._touchProxy = oj._TouchProxy.addTouchListeners(this.tablist);
      },

      _tearDownTouchReorder: function () {
        oj._TouchProxy.removeTouchListeners(this.tablist);
      },

      _setupReorder: function () {
        // enable sortable
        if (this.options.reorderable) {
          var self = this;

          //  - ojtabs touch support issues
          if (oj.DomUtils.isTouchSupported()) {
            this._setupTouchReorder();
          }

          this.tablist.sortable(
            {
              axis: (self._isHorizontal()) ? 'x' : 'y',
              distance: 10,
              start: function () {
                self.tablist.children('.ui-sortable-placeholder').each(function () {
                  $(this).addClass('oj-sortable-placeholder');
                });
              },
              stop: function (event, ui) {
                // find the element that was moved
                var mvTab = ui.item;

                // cancel reorder
                if (!self._doReorder(event, mvTab, mvTab.prev())) {
                  self.tablist.sortable('cancel');
                }
              }
            });

          //  - open/close of tabs leaks sortable
          this._sortable = true;
        } else {
          // disable sortable
          //  - ojtabs touch support issues
          this._tearDownTouchReorder();

          //  - open/close of tabs leaks sortable
          // this.tablist.sortable({ disabled: true });
          this._sortable = false;
        }
      },

      _InitOptions: function (originalDefaults, constructorOptions) {
        // need to find out if edge is specified
        var edge = constructorOptions.edge;
        this._edgeSpecified =
          (edge === 'top' || edge === 'bottom' || edge === 'start' || edge === 'end');
        this._super(originalDefaults, constructorOptions);
      },

      _orientationToEdge: function (value) {
        if (value === 'horizontal') {
          return 'top';
        } else if (value === 'vertical') {
          return 'start';
        }

        return null;
      },

      _setupEdge: function (_edge) {
        var edge = _edge;
        if ((this._initialRender && !this._edgeSpecified) || !edge) {
          edge = this._orientationToEdge(this.options.orientation);
        }

        var oEdge = this.options.edge;
        this.element.removeClass('oj-tabs-top oj-tabs-bottom oj-tabs-start oj-tabs-end');

        switch (edge) {
          case 'top':
          case 'bottom':
            if (this.panels && (oEdge === 'start' || oEdge === 'end')) {
              this.element.removeClass('oj-tabs-vertical oj-helper-clearfix');
            }
            this.element.addClass('oj-tabs oj-component oj-tabs-horizontal');

        //  - support bottom positioned tabs
            if (edge === 'bottom') {
              this.element.addClass('oj-tabs-bottom');
            } else {
              this.element.addClass('oj-tabs-top');
            }

        // don't fire optionChange event during creation
            if (this._initialRender) {
              this.options.orientation = 'horizontal';
            } else if (this.options.orientation !== 'horizontal') {
              this.option('orientation', 'horizontal',
                      { _context: { internalSet: true }, changed: true });
            }
            break;

          case 'start':
          case 'end':
            if (this.panels && (oEdge === 'top' || oEdge === 'bottom')) {
              this.element.removeClass('oj-tabs-horizontal');
            }
            this.element.addClass('oj-tabs oj-component oj-tabs-vertical oj-helper-clearfix');

        //  - em:support tabs on right
            if (edge === 'end') {
              this.element.addClass('oj-tabs-end');
            } else {
              this.element.addClass('oj-tabs-start');
            }

        // don't fire optionChange event during creation
            if (this._initialRender) {
              this.options.orientation = 'vertical';
            } else if (this.options.orientation !== 'vertical') {
              this.option('orientation', 'vertical',
                      { _context: { internalSet: true }, changed: true });
            }
            break;

          default:
            return false;
        }

      // use "edge" and deprecate "orientation"
      // don't fire optionChange event during creation
        if (this._initialRender) {
          this.options.edge = edge;
        } else if (this.options.edge !== edge) {
          this.option('edge', edge,
                    { _context: { internalSet: true }, changed: true });
        }

        return true;
      },

      _getTabbarWrapper: function () {
        var ulParent = this.tablist.parent();
        if (!ulParent.hasClass('oj-tabs-conveyor')) {
          ulParent = this.tablist
            .wrap('<div>') // @HTMLUpdateOK
            .parent()
            .addClass('oj-tabs-conveyor');

        //  - er for placing buttons alongside tabs
          var newTable;

          if (this.element.children('.oj-tabs-facet').length > 0) {
            newTable = ulParent
            .wrap('<div>') // @HTMLUpdateOK
            .wrap('<div>') // @HTMLUpdateOK
            .parent()
            .parent()
            .addClass('oj-tabs-conveyorbelt-wrapper');
          } else {
            newTable = ulParent;
          }

        //  - truncation not working
        // need to create the "oj-tabs-nav-root" before addConveyor
        // otherwise, all skin selectors wont apply

        // add special class so tabs component can skin the conveyor
        // overflow indicators to addpadding between them and the tabs

        //  - er for placing buttons alongside tabs
          var navRoot = newTable
          .wrap('<div>') // @HTMLUpdateOK
          .parent()
          .addClass('oj-tabs-nav-root');

          navRoot.uniqueId().attr('id');
        }
        return ulParent;
      },

      _addPrefixId: function (elem) {
        if (elem.id.indexOf(_ID_PREFIX) < 0) {
          $(elem).attr('id', _ID_PREFIX + elem.id);
        }
      },

    //* * @inheritdoc */
      getNodeBySubId: function (locator) {
        if (locator == null) {
          return this.element ? this.element[0] : null;
        }

        var subId = locator.subId;
        var index = locator.index;

        if (subId !== 'oj-conveyorbelt' &&
          ((typeof index !== 'number') ||
           index < 0 || index >= this.panels.length)) {
          return null;
        }

        switch (subId) {
          case 'oj-conveyorbelt':
            return this.conveyor ? this.conveyor[0] : null;

          case 'oj-tabs-panel':
            return this._getPanelForTab(this.tabs[index])[0];

          case 'oj-tabs-tab':
            return this.tabs[index];

          case 'oj-tabs-title':
            return $(this.tabs[index]).find('.oj-tabs-title')[0];

          case 'oj-tabs-close-icon':
          case 'oj-tabs-close':
            return $(this.tabs[index]).find('.oj-tabs-close-icon')[0];
          default:
        }

      // Non-null locators have to be handled by the component subclasses
        return null;
      },

    //* * @inheritdoc */
      getSubIdByNode: function (node) {
        var panels = [];
        for (var i = 0; i < this.tabs.length; i++) {
          panels.push(this._getPanelForTab(this.tabs[i])[0]);
        }
        var tabIndex = -1;
        var panelIndex = -1;
        var currentNode = node;
      // Find the tab, panel or conveyor belt this node is a descendent of
        while (currentNode) {
          if (this.conveyor && currentNode === this.conveyor[0]) {
            return { subId: 'oj-conveyorbelt' };
          }

          tabIndex = Array.prototype.indexOf.call(this.tabs, currentNode);
          if (tabIndex !== -1) {
            break;
          }
          panelIndex = panels.indexOf(currentNode);
          if (panelIndex !== -1) {
            return { subId: 'oj-tabs-panel', index: panelIndex };
          }

          currentNode = currentNode.parentElement;
        }
      // If it's a tab, identify the specific subelement
        if (tabIndex !== -1) {
          var title = this.getNodeBySubId({ subId: 'oj-tabs-title', index: tabIndex });
          var closeIcon = this.getNodeBySubId({ subId: 'oj-tabs-close', index: tabIndex });
          currentNode = node;
          while (currentNode) {
            if (currentNode === title) {
              return { subId: 'oj-tabs-title', index: tabIndex };
            } else if (currentNode === closeIcon) {
              return { subId: 'oj-tabs-close', index: tabIndex };
            } else if (currentNode === this.tabs[tabIndex]) {
              return { subId: 'oj-tabs-tab', index: tabIndex };
            }
            currentNode = currentNode.parentElement;
          }
        }
        return null;
      },

      _getConveyorWrapperMaxWidth: function () {
      // add extra 10px for buffer
        return this._originalWidth + 10;
      },

      _getTabsWidth: function () {
      //  - er for placing buttons alongside tabs
        var tabbar = this.element.find('.oj-tabs-conveyorbelt-wrapper');
        return (tabbar.length) ? tabbar[0].clientWidth : this.element[0].clientWidth;
      },

      _isOverflow: function () {
        return (this._originalWidth > this._getTabsWidth());
      },

      _getTabMaxWidth: function () {
        var max = Math.floor(this._getTabsWidth() / this.tabs.length);

        if (this.options.removable) {
          max -= _CLOSE_ICON_SIZE;
        }

        return max;
      },

      _applyTabMaxWidth: function () {
        var maxWidth = this._getTabMaxWidth();

        this.tablist.find('.oj-tabs-title').each(function () {
          $(this).css('max-width', '' + maxWidth + 'px')
               .addClass('oj-tabs-title-overflow');
        });

//      this._logMessage("apply max width");
      },

      _removeTabMaxWidth: function () {
        this.tablist.find('.oj-tabs-title').each(function () {
          $(this).css('max-width', '')
               .removeClass('oj-tabs-title-overflow');
        });

//      this._logMessage("remove max width");
      },

      _logMessage: function (/* msg */) {
//      console.log(msg);
      },

    /* resize handler */
      _handleResize: function (/* width */) {
//      this._logMessage("width " + width + " ulWidth " + this._originalWidth +
//                       " clientWidth " + this._getTabsWidth());

      //  - truncation=none tabs + buttons are not laid out correctly.
      // if no truncation, skip applying tab max width
        if (this._isProgressive()) {
          if (this._isOverflow()) {
//        this._logMessage("overflow");
            this._applyTabMaxWidth();
          } else {
//        this._logMessage("underflow");
            this._removeTabMaxWidth();
          }
        }
      },

      _isProgressive: function () {
        return this.options.truncation === 'auto' ||
        this.options.truncation === 'progressive';
      },

      _truncateBeforeOverflow: function () {
      //  - truncation=none tabs + buttons are not laid out correctly.
      // after adding buttons to the tab bar, the parent of the conveyor is a div with inline-block
      // which prevent the conveyor to get the width correctly.
      // Now tabs needs to add resize listener so it can set size on the conveyor in the size listener
        if (this._isHorizontal() && this.tabs.length > 0) {
          if (this._resizeHandler == null) {
            this._resizeHandler = this._handleResize.bind(this);
          }

          oj.DomUtils.addResizeListener(this.element[0], this._resizeHandler);
          this._hasResizeListener = true;

        // truncation=none and has facet
          this._originalWidth = this._getTabbarWrapper()[0].scrollWidth;

        // don't need to apply tab max width if not progressive
        // handle initial overflow
          if (this._isProgressive()) {
            if (this._isOverflow()) {
              this._applyTabMaxWidth();
            }
          }
        }
      },

    /**
     * @override
     * @protected
     * @instance
     * @memberof oj.ojtabs
     */
      _NotifyShown: function () {
        this._super();
      // 19518432(using tab in popup), tab need not be visible while calculating the margin,
        this.refresh();
      },

    /**
     * @override
     * @protected
     * @instance
     * @memberof oj.ojtabs
     */
      _NotifyAttached: function () {
        this._super();
        this.refresh();
      },

      _buildContextMenuItem: function (cmd) {
        return $('<a>')
        .text(this.getTranslatedString(_arMenuKeyMap[cmd]))
        .attr('href', '#')
        .wrap('<li>')  // @HTMLUpdateOK
        .parent()
        .attr('id', _arMenuCmdMap[cmd])
        .addClass('oj-menu-item');
      },

    /**
     *  Menu "cut" functionality
     *  @private
     */
      _menu_cut: function (obj) {
        if (!obj || !obj.length) {
          return false;
        }

        this._menu.cutTab = obj;
        return true;
      },

    /**
     *  Menu "paste" functionality
     *  @private
     */
      _menu_paste: function (event, obj, pasteBefore) {
        if (!obj || !obj.length) {
          return false;
        }
        if (!this._menu.cutTab) {
          return false;
        }

        var mvTab = this._menu.cutTab;
        this._menu.cutTab = false;
        this._doReorder(event, mvTab, obj, pasteBefore);
        return true;
      },

    /**
     *  Menu "remove" functionality
     *  @private
     */
      _menu_remove: function (event, obj) {
        if (!obj || !obj.length) {
          return false;
        }

        this._removeTab(obj, null, event);
        return true;
      },

    /**
     *   Check menu selected to see if it one of tabs predefined cut/paste id's
     *   @private
     */
      _handleContextMenuSelect: function (ev, ui) {
        var id = ui ? ui.item.attr('id') : undefined;

        if (id === 'ojtabscut') {
          this._menu_cut(this._menu.tab);
        } else if (id === 'ojtabspastebefore') {
          this._menu_paste(ev, this._menu.tab, true);
        } else if (id === 'ojtabspasteafter') {
          this._menu_paste(ev, this._menu.tab, false);
        } else if (id === 'ojtabsremove') {
          this._menu_remove(ev, this._menu.tab);
        }
      },

    /**
     *  Initialize the context menu.  This is called on startup, or on option
     *  "contextMenu" change.
     *  @param {Object=} newVal   true if called because of an option change.
     *  @private
     */
      _initMenu: function (newVal) {
        var menu = newVal || this.options.contextMenu;

        if (menu) {
          var t = $.type(menu);
          if (t === 'function') {
            try {
              menu = menu();             // call user's method to get the context menu
            } catch (e) {
              menu = null;
            }
            t = $.type(menu);
          }

          if (menu) {
            var $m = $(menu);                  // get the user's <ul> list
            if ($m.length) {
              this.options.contextMenu = menu;
            }
          }
        }
        this._menu.usermenu = !!menu;

      // add default context menu
        this._addContextMenu();
      },

      _addIfNotExist: function ($ul, itemList, command) {
        if (itemList.indexOf(command) === -1) {
          var menuItem = this._buildContextMenuItem(command);
          $ul.append(menuItem); // @HTMLUpdateOK
          this._menu.$menuItems.push(menuItem);
        }
      },

    /**
     * Add a default context menu to the tabs if there is none. If there is
     * a context menu set on the tabs options we use that one. Add listeners
     * for context menu before show.
     * @private
     */
      _addContextMenu: function () {
        var $menuContainer = $(this.options.contextMenu);
        if ($menuContainer.length === 0 && !this.options.reorderable && !this.options.removable) {
          return;
        }

        var self = this;
        if ($menuContainer.length === 0) {
          var key = this.options.reorderable ? 'labelReorder' : _arMenuKeyMap.remove;
          var menuId = this.element.uniqueId().attr('id') + 'contextmenu';
          $menuContainer = $('<ul>');
          $menuContainer.css('display', 'none')
          .attr('id', menuId)
          .attr('aria-label', this.getTranslatedString(key));

        // append the menu to body
          $(document.body).append($menuContainer); // @HTMLUpdateOK
          $menuContainer.ojMenu();

        // don't call setOption
          this.options.contextMenu = this._sanitizeIdSelector(menuId);
        }

        var itemList = [];

      // If there are any ojTabs built in menu item ids, construct the menu items
        $menuContainer.find('[data-oj-command]').each(function () {
          if ($(this).children('a').length === 0) {
            var command = $(this).attr('data-oj-command').slice('oj-tabs-'.length);
            $(this).replaceWith(self._buildContextMenuItem(command)); // @HTMLUpdateOK
            $(this).addClass('oj-menu-item');
            itemList.push(command);
          }
        });

      // add build-in cut/paste
        if (this.options.reorderable) {
          this._addIfNotExist($menuContainer, itemList, 'cut');
          this._addIfNotExist($menuContainer, itemList, 'paste-before');
          this._addIfNotExist($menuContainer, itemList, 'paste-after');

          this._menu.$elemPasteBefore = $menuContainer.find('#ojtabspastebefore');
          this._menu.$elemPasteAfter = $menuContainer.find('#ojtabspasteafter');
        }

      // add build-in "remove"
        if (this.options.removable) {
          this._addIfNotExist($menuContainer, itemList, 'remove');
          this._menu.$elemRemove = $menuContainer.find('#ojtabsremove');
        }
        this._menu.$container = $menuContainer;
        $menuContainer.ojMenu('refresh');

      // Add our listeners so that we can handle build-in remove, cut/paste
        $menuContainer.on('ojselect', $.proxy(this._handleContextMenuSelect, this));
      },

    /**
     *  Clear out any contextMenu data.
     *  @private
     */
      _clearMenu: function () {
        var menu = this._menu;
        if (menu && menu.$container) {
          menu.$container.off('ojselect');
          if (!menu.usermenu) {
            menu.$container.ojMenu('destroy');
            menu.$container.remove();
          }
          if (menu.$menuItems) {
            for (; menu.$menuItems.length > 0;) {
              menu.$menuItems.pop().remove();
            }
          }
          menu.$container = null;
        }

        menu.$elemPasteBefore = null;
        menu.$elemPasteAfter = null;
        menu.$elemRemove = null;
      },

    //  - tabs lose selection when moved (sortable tabs)
      _doReorder: function (event, mvTab, prevTab, pasteBefore) {
        var mvContent = this._getPanelForTab(mvTab);

        var eventData = {
        /** @expose */
          tab: mvTab,
        /** @expose */
          content: mvContent
        };

      // fire beforeReorder event and only reorder when there is no veto
        if (this._trigger('beforeReorder', event, eventData) === false) {
          return false;
        }

        var mvInd = this.tabs.index(mvTab);

        if (prevTab.length) {
          if (this.tabs.index(prevTab) === mvInd) {
            return true;
          }

          var prevContent = this._getPanelForTab(prevTab);
          if (pasteBefore) {
            prevTab.before(mvTab); // @HTMLUpdateOK
            prevContent.before(mvContent); // @HTMLUpdateOK
          } else {
            prevTab.after(mvTab); // @HTMLUpdateOK
            prevContent.after(mvContent); // @HTMLUpdateOK
          }
        } else if (this.tabs.length > 0) {
          if (mvInd === 0) {
            return true;
          }
          this.tabs.first().before(mvTab); // @HTMLUpdateOK
          this.panels.first().before(mvContent); // @HTMLUpdateOK
        }

      //  - calling refresh after reordering tabs causes tabs to loose there disabled state
      // update disabled and active
        this._updateDisabledTabs();

      //  - add new tab. move it to first place. add another tab.first tab disappears.
        this.refresh();

      // remove the "oj-focus-highlight" class from the old active tab
        prevTab.blur();
        mvTab.focus();

      // fire reorder event
        this._trigger('reorder', event, eventData);

        return true;
      },

      //  - ojtabs do not work with tabs whose id has '::'
      _sanitizeSelector: function (hash) {
         // eslint-disable-next-line no-useless-escape
        return hash ? hash.replace(/[#!\"$%&'()*+,./:;<=>?@\[\]^`{|}~]/g, '\\$&') : '';
      },

      //  - can not add new tab if existing selected tab has non-blessed jquery id
      // return "#" + sanitized idSelector
      _sanitizeIdSelector: function (idSelector) {
        return idSelector ? '#' + this._sanitizeSelector(idSelector) : '';
      },

      _getPanelForTab: function (tab) {
        return this.element.find(this._sanitizeIdSelector($(tab).attr('aria-controls')));
      },

      _getUniqueId: function (panel) {
        var id = panel.attr('id');
        if (!id) {
          id = panel.uniqueId().attr('id');
          panel.addClass('oj-tabs-gen-id');
        }
        return id;
      },

      _getTab: function (idOrIndex) {
        var index = this._getIndexByTabOrContentId(idOrIndex);
        if (index !== -1) {
          return $(this.tabs[index]);
        }
        return undefined;
      },

    // Note: return -1 if not valid id
      _getIndexByTabOrContentId: function (idOrIndex) {
        var index = -1;

        if (typeof idOrIndex === 'number') {
          if (idOrIndex >= 0 && idOrIndex < this.tabs.length) {
            index = idOrIndex;
          }
        } else if (typeof idOrIndex === 'string') {
        // security test
          var selector = this._sanitizeSelector(idOrIndex);
          if (oj.DomUtils.isValidIdentifier(selector)) {
            var tabOrContent = this.element.find(this._sanitizeIdSelector(selector));

            if (tabOrContent.length) {
              index = this.tabs.index(tabOrContent);
              if (index === -1) {
                index = this.panels.index(tabOrContent);
              }
            }
          }
        }
        return index;
      },

      _getSelectedIndex: function () {
        var selected = this.options.selected;
        if (typeof selected === 'number') {
          return selected;
        }

        return this.tabs.index($(this._sanitizeIdSelector(selected)));
      },

    /*
     * return either page author provided id or index
     */
      _getTabIdOrIndex: function (tab) {
        if (tab) {
          var id = tab.attr('id');
          return id || this.tabs.index(tab);
        }

        return undefined;
      },

      _setOjDisabledOnTab: function (disTabs) {
      // clear oj-disabled that were on the tabs
        var tabbar = this.tablist ? this.tablist : this.element.children('ul');
        var children = tabbar.children('li');
        children.removeClass('oj-disabled')
              .removeAttr('aria-disabled');

        var arr = [];

      // for each disabled tab in otpions.disabledTabs, add oj-disabled to it
        if (disTabs && disTabs.length > 0) {
          var id;
          var tab;

          for (var i = 0; i < disTabs.length; i++) {
            tab = this._getTab(disTabs[i]);
            if (tab) {
              tab.addClass('oj-disabled');

              tab.attr('aria-disabled', 'true');
            //  - clicking on disabled tab takes user to the top of the page
              tab.find('.oj-tabs-anchor').removeAttr('href');

              id = tab.attr('id');
              arr.push(id || children.index(tab));
            }
          }
        }
        this._updateDisabledTabs(arr);
      },

    /*
     * update disabledTabs array
     * entry value is either page author provided id or index
     */
      _updateDisabledTabs: function (_arr) {
        var arr = _arr;
        if (!arr) {
          arr = [];
          var self = this;
          this.tablist.children().each(
          function () {
            var tab = $(this);
            if (tab.hasClass('oj-disabled')) {
              arr.push(self._getTabIdOrIndex(tab));
            }
          });
        }
        if (!oj.Object._compareArrayIdIndexObject(arr, this.options.disabledTabs)) {
        //  - tabs shouldn't implement _init()
        // don't fire initial option change events
          if (this._initialRender) {
            this.options.disabledTabs = arr;
          } else {
            var context = {};
            context.internalSet = true;
            context.writeback = true;
            this.option({ disabledTabs: arr }, { _context: context, changed: true });
          }
        }
      },

      _isInATab: function (element) {
        var found = false;
        this.tabs.each(function () {
          if (this === element || $.contains(this, element)) {
            found = true;
            return false;
          }
          return true;
        });
        return found;
      },

    //  - use case where gauges are displayed in tabs title bar does not work
      _createTabbar: function () {
        var contentIds = this._getContentIds();

        var self = this;
        var tabbar = this.element.children('ul');

      // tabbar exists
        if (tabbar.length > 0) {
          tabbar.children('li').each(function (index) {
            self._wrapLi($(this), contentIds[index]);
          });
        }
      },

    // Backward compatible: move headers to form a tab bar
      _createTabbarFromOldMarkup: function () {
        var tabbar = this.element.children('ul');

        if (tabbar.length === 0) {
          var self = this;
          var contentIds = this._getContentIds();

          tabbar = $('<ul>');

          this.element.children().each(
          function (index) {
            var tab = self._wrapLi($(this).find('> :first-child'), contentIds[index]);

            tab.appendTo(tabbar); // @HTMLUpdateOK
          });

          tabbar.prependTo(this.element); // @HTMLUpdateOK
        }
      },

    //  - er for placing buttons alongside tabs
      _addFacets: function () {
        var navRoot = this.element.children('.oj-tabs-nav-root');
        var self = this;
        var tabbarWrapper = navRoot.children('.oj-tabs-conveyorbelt-wrapper');

        this.element.children('.oj-tabs-facet').each(
        function () {
          var facet = $(this);
          var facetId = self._getUniqueId(facet);

          // if facet doesn't exist add it to the tab bar
          if (navRoot.find(self._sanitizeIdSelector(facetId)).length === 0) {
            // add start facet
            if (facet.hasClass('oj-start')) {
              tabbarWrapper.before(facet); // @HTMLUpdateOK
            } else {
              // add other facets
              facet.appendTo(navRoot); // @HTMLUpdateOK
            }
          }
        });
      },

      _getContentIds: function () {
        var contentIds = [];
        var self = this;

        this.element.children(':not(ul):not(.oj-tabs-facet)').each(
        function () {
          var panel = $(this);
          panel.addClass('oj-tabs-panel');
          contentIds.push(self._getUniqueId(panel));
        });

        return contentIds;
      }

    // Fragments:

    /**
     * <p>Sub-ID for the specified tab.</p>
     *
     * @property {number} index The zero-based index of the tab.
     *
     * @ojsubid oj-tabs-tab
     * @memberof oj.ojTabs
     * @ignore
     * @deprecated This sub-ID is not needed.  Since the application supplies this element, it can supply a unique ID by which the element can be accessed.
     *
     * @example <caption>Get the second tab:</caption>
     * var node = $( ".selector" ).ojTabs( "getNodeBySubId", {'subId': 'oj-tabs-tab', 'index': 1} );
     */

    /**
     * <p>Sub-ID for the title of the specified tab.</p>
     *
     * @property {number} index The zero-based index of the tab.
     *
     * @ojsubid oj-tabs-title
     * @memberof oj.ojTabs
     * @ignore
     * @deprecated This sub-ID is not needed.  Since the application supplies this element, it can supply a unique ID by which the element can be accessed.
     *
     * @example <caption>Get the title of the second tab:</caption>
     * var node = $( ".selector" ).ojTabs( "getNodeBySubId", {'subId': 'oj-tabs-title', 'index': 1} );
     */

    /**
     * <p>Sub-ID for the content corresponding to the specified tab.</p>
     *
     * @property {number} index The zero-based index of the tab.
     *
     * @ojsubid oj-tabs-panel
     * @memberof oj.ojTabs
     * @ignore
     * @deprecated This sub-ID is not needed.  Since the application supplies this element, it can supply a unique ID by which the element can be accessed.
     *
     * @example <caption>Get the content corresponding to the second tab:</caption>
     * var node = $( ".selector" ).ojTabs( "getNodeBySubId", {'subId': 'oj-tabs-panel', 'index': 1} );
     */

    /**
     * <p>Sub-ID for the close icon of the specified tab, if present.</p>
     *
     * @property {number} index The zero-based index of the tab.
     *
     * @ojsubid oj-tabs-close-icon
     * @memberof oj.ojTabs
     * @ignore
     * @deprecated this sub-ID is deprecated, please use oj-tabs-close instead.
     *
     * @example <caption>Get the close icon of the second tab:</caption>
     * var node = $( ".selector" ).ojTabs( "getNodeBySubId", {'subId': 'oj-tabs-close-icon', 'index': 1} );
     */

    /**
     * <p>Sub-ID for the close icon of the specified tab, if present.</p>
     *
     * @property {number} index The zero-based index of the tab.
     *
     * @ojsubid oj-tabs-close
     * @memberof oj.ojTabs
     *
     * @example <caption>Get the close icon of the second tab:</caption>
     * var node = $( ".selector" ).ojTabs( "getNodeBySubId", {'subId': 'oj-tabs-close', 'index': 1} );
     */

    /**
     * <p>Sub-ID for the conveyor belt used for horizontal overflow.</p>
     *
     * @ojsubid oj-conveyorbelt
     * @memberof oj.ojTabs
     *
     * @example <caption>Get the conveyor belt:</caption>
     * var node = $( ".selector" ).ojTabs( "getNodeBySubId", {'subId': 'oj-conveyorbelt'} );
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Select</td>
     *     </tr>
     *     <tr>
     *       <td>Tab Close Icon</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Delete</td>
     *     </tr>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>Press and pan left/right and release</kbd></td>
     *       <td>Reorder</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojTabs
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>Tab</kbd></td>
     *       <td>Only the selected tab is in the tab order.</td>
     *     </tr>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>UpArrow or LeftArrow</kbd> (<kbd>RightArrow</kbd> in RTL)</td>
     *       <td> Move focus to the previous tab and select it.</tr>
     *     </tr>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>DownArrow or RightArrow</kbd> (<kbd>LeftArrow</kbd> in RTL)</td>
     *       <td> Move focus to the next tab and select it.</tr>
     *     </tr>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>Home</kbd></td>
     *       <td> Move focus to the first tabs item.</tr>
     *     </tr>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>End</kbd></td>
     *       <td> Move focus to the last tabs item.</tr>
     *     </tr>
     *     <tr>
     *       <td>Tab</td>
     *       <td><kbd>Delete</kbd></td>
     *       <td> If deletion is allowed, will delete the current tab.</tr>
     *     </tr>
     *     <tr>
     *       <td>Tab content</td>
     *       <td><kbd>Shift+Tab</kbd></td>
     *       <td> move focus to the tab for that tab panel.</tr>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojTabs
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     *
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-tabs-icon-only</td>
     *       <td>Applied to an ojtabs if the tab headers contain only icons.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-tabs-text-icon<br>
     *       <td>Applied to an ojtabs if the tab headers contain both icons and text.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-tabs-tab-icon</td>
     *       <td>Applied to a custom icon of a tab.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-tabs-facet</td>
     *       <td>Applied to an element that is displayed on the tab bar either before or after the horizontal tab headers.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojTabs
     */
    });
}());

/**This file is generated. Do not edit directly. Actual file located in 3rdparty/dvt/prebuild.**/
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojTagCloud
 * @augments oj.dvtBaseComponent
 * @since 1.1.0
 * @ojstatus preview
 * @ojrole application
 * @ojshortdesc An interactive data visualization of textual data, where the importance of each tagged word or phrase is represented by font size or color.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="tagCloudOverview-section">
 *   JET Tag Cloud
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#tagCloudOverview-section"></a>
 * </h3>
 *
 * <p>Tag clouds are used to display text data with
 * the importance of each tag shown with font size or color.</p>
 *
 * {@ojinclude "name":"warning"}
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-tag-cloud 
 *   items='[{"id": "item1", "label": "the", "value": 20}, 
 *           {"id": "item2", "label": "cat", "value": 17}, 
 *           {"id": "item3", "label": "hat", "value": 13}]'>
 * &lt;/oj-tag-cloud>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <p>When using font colors as a data dimension for tag clouds, the application
 * needs to ensure that they meet minimum contrast requirements. Not all colors
 * in the default value ramp provided by oj.ColorAttributeGroupHandler
 * will meet minimum contrast requirements.</p>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets. Alternate visualizations should
 *    be considered if identifying data changes is important, since all data items will generally move and resize on any data
 *    change.
 * </p>
 *
 * <h4>Layout</h4>
 * <p>Rectangular layouts are faster than cloud layouts and are recommended for larger data sets.
 * </p>
 * 
 * <h4>Styling</h4>
 * <p>Use the highest level property available. For example, consider setting styling properties using CSS or
 *    <code class="prettyprint">styleDefaults.svgStyle</code>, instead of styling properties
 *    on the individual data items. The tag cloud can take advantage of these higher level properties to apply the style properties on
 *    containers, saving expensive DOM calls.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojTagCloud', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix : "oj",
  options: {
    /**
     * Defines the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojTagCloud
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">animation-on-data-change</code> attribute specified:</caption>
     * &lt;oj-tag-cloud animation-on-data-change='auto'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">animationOnDataChange</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myTagCloud.animationOnDataChange;
     * 
     * // setter
     * myTagCloud.animationOnDataChange="auto";
     */
    animationOnDataChange: "none",
    /**
     * Defines the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojTagCloud
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">animation-on-display</code> attribute specified:</caption>
     * &lt;oj-tag-cloud animation-on-display='auto'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">animationOnDisplay</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myTagCloud.animationOnDisplay;
     * 
     * // setter
     * myTagCloud.animationOnDisplay="auto";
     */
    animationOnDisplay: "none",
    /**
     * An array of category strings used for category filtering. Data items with a category 
     * in hiddenCategories will be filtered.
     * @expose
     * @name hiddenCategories
     * @memberof oj.ojTagCloud
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">hidden-categories</code> attribute specified:</caption>
     * &lt;oj-tag-cloud hidden-categories='["soda", "water"]'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">hiddenCategories</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myTagCloud.hiddenCategories[0];
     * 
     * // Get all
     * var values = myTagCloud.hiddenCategories;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myTagCloud.hiddenCategories=["soda", "water"];
     */
    hiddenCategories: [],
    /**
     * An array of category strings used for category highlighting. Data items with a 
     * category in highlightedCategories will be highlighted.
     * @expose
     * @name highlightedCategories
     * @memberof oj.ojTagCloud
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">highlighted-categories</code> attribute specified:</caption>
     * &lt;oj-tag-cloud highlighted-categories='["soda", "water"]'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">highlightedCategories</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myTagCloud.highlightedCategories[0];
     * 
     * // Get all
     * var values = myTagCloud.highlightedCategories;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myTagCloud.highlightedCategories=["soda", "water"];
     */
    highlightedCategories: [],
    /**
     * The matching condition for the highlightedCategories option. By default, highlightMatch is 'all' 
     * and only items whose categories match all of the values specified in the highlightedCategories 
     * array will be highlighted. If highlightMatch is 'any', then items that match at least one of 
     * the highlightedCategories values will be highlighted.
     * @expose
     * @name highlightMatch
     * @memberof oj.ojTagCloud
     * @instance
     * @type {string}
     * @ojvalue {string} "any"
     * @ojvalue {string} "all"
     * @default "all"
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">highlight-match</code> attribute specified:</caption>
     * &lt;oj-tag-cloud highlight-match='all'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">highlightMatch</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myTagCloud.highlightMatch;
     * 
     * // setter
     * myTagCloud.highlightMatch="all";
     */
    highlightMatch: "all",
    /**
     * Defines the behavior applied when hovering over data items.
     * @expose
     * @name hoverBehavior
     * @memberof oj.ojTagCloud
     * @instance
     * @type {string}
     * @ojvalue {string} "dim"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">hover-behavior</code> attribute specified:</caption>
     * &lt;oj-tag-cloud hover-behavior='dim'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">hoverBehavior</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myTagCloud.hoverBehavior;
     * 
     * // setter
     * myTagCloud.hoverBehavior="dim";
     */
    hoverBehavior: "none",
    /**
     * An array of objects with the following properties that defines the data items for the tag cloud items. 
     * Also accepts a Promise or callback function for deferred data rendering. The function should return 
     * one of the following: 
     * <ul>
     *   <li>Promise: A Promise that will resolve with an array of data items. No data will be rendered if the Promise is rejected.</li> 
     *   <li>Array: An array of data items.</li> 
     * </ul> 
     * @expose
     * @name items
     * @memberof oj.ojTagCloud
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @ojsignature {target: "Accessor", value: {GetterType: "Promise<Array<oj.ojTagCloud.Item>>|null", SetterType: "Array<oj.ojTagCloud.Item>|Promise<Array<oj.ojTagCloud.Item>>|null"}, jsdocOverride: true}
     * @default null
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">items</code> attribute specified:</caption>
     * &lt;oj-tag-cloud items='[{"id": "item1", "label": "the", "value": 20}, 
     *                          {"id": "item2", "label": "cat", "value": 17}, 
     *                          {"id": "item3", "label": "hat", "value": 13}]'>&lt;/oj-tag-cloud>
     * 
     * &lt;oj-tag-cloud items='[[itemsPromise]]'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">items</code> 
     * property after initialization:</caption>
     * 
     * // Get all (The items getter always returns a Promise so there is no "get one" syntax)
     * var values = myTagCloud.items;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myTagCloud.items=[{"id": "item1", "label": "the", "value": 20}, 
     *                   {"id": "item2", "label": "cat", "value": 17}, 
     *                   {"id": "item3", "label": "hat", "value": 13}];
     */
    items: null,
    /**
     * The layout to use for tag display.
     * @expose
     * @name layout
     * @memberof oj.ojTagCloud
     * @instance
     * @type {string}
     * @ojvalue {string} "cloud"
     * @ojvalue {string} "rectangular"
     * @default "rectangular"
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">layout</code> attribute specified:</caption>
     * &lt;oj-tag-cloud layout='cloud'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">layout</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myTagCloud.layout;
     * 
     * // setter
     * myTagCloud.layout="cloud";
     */
    layout: "rectangular",
    /**
     * An array of id strings, used to define the selected data items.
     * @expose
     * @name selection
     * @memberof oj.ojTagCloud
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">selection</code> attribute specified:</caption>
     * &lt;oj-tag-cloud selection='["area1", "area2", "marker7"]'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">selection</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myTagCloud.selection[0];
     * 
     * // Get all
     * var values = myTagCloud.selection;
     * 
     * // Set all (There is no permissible "set one" syntax.)
     * myTagCloud.selection=["tag2", "tag3", "tag12"];
     */
    selection: [],
    /**
     * The type of selection behavior that is enabled on the tag cloud.
     * @expose
     * @name selectionMode
     * @memberof oj.ojTagCloud
     * @instance
     * @type {string}
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">selection-mode</code> attribute specified:</caption>
     * &lt;oj-tag-cloud selection-mode='multiple'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myTagCloud.selectionMode;
     * 
     * // setter
     * myTagCloud.selectionMode="multiple";
     */
    selectionMode: "none",
    /**
     * An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojTagCloud
     * @instance
     * @type {Object}
     * @default {"renderer": null}
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">tooltip</code> attribute specified:</caption>
     * <!-- Using dot notation -->
     * &lt;oj-tag-cloud tooltip.renderer='[[tooltipFun]]'>&lt;/oj-tag-cloud>
     * 
     * &lt;oj-tag-cloud tooltip='[[{"renderer": tooltipFun}]]'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">tooltip</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myTagCloud.tooltip.renderer;
     * 
     * // Get all
     * var values = myTagCloud.tooltip;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myTagCloud.setProperty('tooltip.renderer', tooltipFun);
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myTagCloud.tooltip={'renderer': tooltipFun};
     */
    tooltip: {
      /**
       * A callback function that returns a custom tooltip. The callback function will be called with a context
       * object containing information about the item that was clicked.
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojTagCloud
       * @instance
       * @type {function(Object):Object|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojTagCloud.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))", jsdocOverride: true}
       * 
       * @example <caption>See the <a href="#tooltip">tooltip</a> attribute for usage examples.</caption>
       */
      renderer: null
    },
    /**
     * Component CSS classes should be used to set component wide styling. This API should be used
     * only for styling a specific instance of the component. Properties specified on this object may 
     * be overridden by specifications on the data item. Some property default values come from the CSS 
     * and varies based on theme.
     * @expose
     * @name styleDefaults
     * @memberof oj.ojTagCloud
     * @instance
     * @type {Object}
     * @default {"hoverBehaviorDelay": 200, "svgStyle": {}}
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">style-defaults</code> attribute specified:</caption>
     * <!-- Using dot notation -->
     * &lt;oj-tag-cloud style-defaults.animation-duration='200'>&lt;/oj-tag-cloud>
     * 
     * <!-- Using JSON notation -->
     * &lt;oj-tag-cloud style-defaults='{"animationDuration": 200, "svgStyle": {"fill": "url(someURL#filterId)"}}'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">styleDefaults</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myTagCloud.styleDefaults.animationDuration;
     * 
     * // Get all
     * var values = myTagCloud.styleDefaults;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myTagCloud.setProperty('styleDefaults.svgStyle', {'fill': 'url(someURL#filterId)'});
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myTagCloud.styleDefaults={'fill': 'url(someURL#filterId)'};
     */
    styleDefaults: {
      /**
       * The duration of the animations in milliseconds. The default value comes from the CSS and varies based on theme.
       * @expose
       * @name styleDefaults.animationDuration
       * @memberof! oj.ojTagCloud
       * @instance
       * @type {number}
       * @ojsignature {target: "Type", value: "?"}
       * 
       * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
       */
      animationDuration: undefined,
      /**
       * Specifies initial hover delay in ms for highlighting data items.
       * @expose
       * @name styleDefaults.hoverBehaviorDelay
       * @memberof! oj.ojTagCloud
       * @instance
       * @type {number}
       * @ojsignature {target: "Type", value: "?"}
       * @default 200
       * 
       * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
       */
      hoverBehaviorDelay: 200,
      /**
       * The CSS style object defining the style of the items. The default value comes from the CSS and varies based on theme.
       * @expose
       * @name styleDefaults.svgStyle
       * @memberof! oj.ojTagCloud
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * 
       * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
       */
    },
    /**
     * Data visualizations require a press and hold delay before triggering tooltips and rollover effects on 
     * mobile devices to avoid interfering with page panning, but these hold delays can make applications seem 
     * slower and less responsive. For a better user experience, the application can remove the touch and hold 
     * \delay when data visualizations are used within a non scrolling container or if there is sufficient space 
     * outside of the visualization for panning. If touchResponse is touchStart the element will instantly trigger 
     * the touch gesture and consume the page pan events. If touchResponse is auto, the element will behave like 
     * touchStart if it determines that it is not rendered within scrolling content and if element panning is not 
     * available for those elements that support the feature. 
     * @expose
     * @name touchResponse
     * @memberof oj.ojTagCloud
     * @instance
     * @type {string}
     * @ojvalue {string} "touchStart"
     * @ojvalue {string} "auto"
     * @default "auto"
     * 
     * @example <caption>Initialize the tag cloud with the 
     * <code class="prettyprint">layout</code> attribute specified:</caption>
     * &lt;oj-tag-cloud touch-response='touchStart'>&lt;/oj-tag-cloud>
     * 
     * @example <caption>Get or set the <code class="prettyprint">touchResponse</code> 
     * property after initialization:</caption>
     * // getter
     * var value = myTagCloud.touchResponse;
     * 
     * // setter
     * myTagCloud.touchResponse="touchStart";
     */
    touchResponse: "auto"
  },

  // @inheritdoc
  _CreateDvtComponent : function(context, callback, callbackObj) {
    return dvt.TagCloud.newInstance(context, callback, callbackObj);
  },

  // @inheritdoc
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-tagcloud-item') {
      // item[index]
      subId = 'item[' + locator['index'] + ']';
    }
    else if(subId == 'oj-tagcloud-tooltip') {
      subId = 'tooltip';
    }

    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  },

  // @inheritdoc
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId.indexOf('item') == 0) {
      // item[index]
      locator['subId'] = 'oj-tagcloud-item';
      locator['index'] = this._GetFirstIndex(subId);
    }
    else if(subId == 'tooltip') {
      locator['subId'] = 'oj-tagcloud-tooltip';
    }

    return locator;
  },

  // @inheritdoc
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-tagcloud');
    return styleClasses;
  },

  // @inheritdoc
  _GetChildStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses['oj-dvtbase oj-tagcloud'] = {'path': 'styleDefaults/animationDuration', 'property': 'ANIM_DUR'};
    styleClasses['oj-tagcloud'] = {'path': 'styleDefaults/svgStyle', 'property': 'TEXT'};
    return styleClasses;
  },

  // @inheritdoc
  _GetEventTypes : function() {
    return ['optionChange'];
  },

  // @inheritdoc
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtUtilBundle.TAG_CLOUD'] = translations['componentName'];
    return ret;
  },
  
  //** @inheritdoc */
  _InitOptions : function(originalDefaults, constructorOptions) {
    this._super(originalDefaults, constructorOptions);

    // styleDefaults subproperty defaults are dynamically generated
    // so we need to retrieve it here and override the dynamic getter by
    // setting the returned object as the new value.
    var styleDefaults = this.options["styleDefaults"];
    this.options["styleDefaults"] = styleDefaults;
  },

  /**
   * Returns an object with the following properties for automation testing verification of the item at the
   * specified index.
   * @param {number} index The index.
   * @return {Object|null} An object containing data for the node at the given index, or null if none exists.
   * @ojsignature {target: "Type", value: "oj.ojTagCloud.ItemContext|null", jsdocOverride: true, for: "returns"}
   * @expose
   * @memberof oj.ojTagCloud
   * @instance
   */
  getItem: function(index) {
    var auto = this._component.getAutomation();
    return auto.getItem(index);
  },
  /**
   * Returns the number of items in the tag cloud data.
   * @return {number} The number of data items
   * @expose
   * @memberof oj.ojTagCloud
   * @instance
   */
  getItemCount: function() {
    return this._component.getAutomation().getItemCount();
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   * @ojsignature {target: "Type", value: "oj.ojTagCloud.NodeContext|null", jsdocOverride: true, for: "returns"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojTagCloud
   */
  getContextByNode: function(node) {
    // context objects are documented with @ojnodecontext
    var context = this.getSubIdByNode(node);
    if (context && context['subId'] !== 'oj-tagcloud-tooltip')
      return context;

    return null;
  },

  // @inheritdoc
  _GetComponentDeferredDataPaths : function() {
    return {'root': ['data', 'items']};
  }
});

// Conditionally set the defaults for custom element vs widget syntax since we expose different APIs
oj.Components.setDefaultOptions({
  'ojTagCloud': {
    'styleDefaults': oj.Components.createDynamicPropertyGetter(function(context) {
      if (context['isCustomElement'])
        return {'svgStyle': {}};
      return {};
    }),
  }
});
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *    <tr>
 *       <td rowspan="4">Data Item</td>
 *       <td rowspan="2"><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *    <tr>
 *       <td>Open a link when the <code class="prettyprint">url</code> for a data item is set.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojTagCloud
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus and selection to previous data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus and selection to next data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus and selection to previous data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus and selection to next data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + UpArrow</kbd></td>
 *       <td>Move focus and multi-select previous data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + DownArrow</kbd></td>
 *       <td>Move focus and multi-select next data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + LeftArrow</kbd></td>
 *       <td>Move focus and multi-select previous data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + RightArrow</kbd></td>
 *       <td>Move focus and multi-select next data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + UpArrow</kbd></td>
 *       <td>Move focus to previous data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + DownArrow</kbd></td>
 *       <td>Move focus to next data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + LeftArrow</kbd></td>
 *       <td>Move focus to previous data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + RightArrow</kbd></td>
 *       <td>Move focus to next data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Spacebar</kbd></td>
 *       <td>Multi-select data item with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Open a link when the <code class="prettyprint">url</code> for a data item is set.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojTagCloud
 */

// PROPERTY TYPEDEFS

/**
 * @typedef {Object} oj.ojTagCloud.Item
 * @property {Array.<string>=} categories An array of category strings corresponding to the tag cloud items. This allows highlighting and filtering of items.
 * @property {string=} color The color of the text. Will be overridden by any color defined in the style option. The default value comes from the CSS and varies based on theme.
 * @property {string=} id The item id.
 * @property {string} label The text of the item.
 * @property {string=} shortDesc The description of the item. This is used for customizing the tooltip text.
 * @property {Object=} svgStyle The CSS style object defining the style of the item text.
 * @property {string=} svgClassName The CSS style class defining the style of the item text.
 * @property {string=} url The url this item references.
 * @property {number} value The value of this item which will be used to scale its font-size within the tag cloud.
 */
/**
 * @typedef {Object} oj.ojTagCloud.TooltipContext
 * @property {string} color The color of the hovered item.
 * @property {Element} componentElement The tag cloud element.
 * @property {string} id The id of the hovered item.
 * @property {string} label The data label of the hovered item.
 * @property {Element} parentElement The tooltip element. The function can directly modify or append content to this element.
 * @property {number} value The value of the hovered item.
 */

// METHOD TYPEDEFS

/**
 * @typedef {Object} oj.ojTagCloud.ItemContext
 * @property {string} color The color of the item at the given index.
 * @property {string} label The data label of the item at the given index.
 * @property {boolean} selected True if the item at the given index is currently selected and false otherwise.
 * @property {string} tooltip The tooltip of the item at the given index.
 * @property {number} value The value of the item at the given index.
 */
/**
 * @typedef {Object} oj.ojTagCloud.NodeContext
 * @property {string} subId The subId string identify the particular DOM node.
 * @property {number} index The zero based index of the tag cloud item.
 */

// KEEP FOR WIDGET SYNTAX

/**
 * The CSS style string/object defining the style of the text.
 * @expose
 * @name items[].style
 * @memberof! oj.ojTagCloud
 * @instance
 * @type {string|Object}
 * @ignore
 */
/**
 * The CSS style class defining the style of the text.
 * @expose
 * @name items[].className
 * @memberof! oj.ojTagCloud
 * @instance
 * @type {string}
 * @ignore
 */
/**
 * The CSS style string/object defining the style of the items.
 * @expose
 * @name styleDefaults.style
 * @memberof! oj.ojTagCloud
 * @instance
 * @type {string|Object}
 * @ignore
 */

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for tag cloud items at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojsubid oj-tagcloud-item
 * @memberof oj.ojTagCloud
 *
 * @example <caption>Gets the first tag cloud item:</caption>
 * var nodes = myTagCloud.getNodeBySubId({'subId': 'oj-tagcloud-item', 'index': 0});
 */

/**
 * <p>Sub-ID for the the tag cloud tooltip.</p>
 *
 * @ojsubid oj-tagcloud-tooltip
 * @memberof oj.ojTagCloud
 *
 * @example <caption>Get the tooltip object of the tag cloud, if displayed:</caption>
 * var nodes = myTagCloud.getNodeBySubId({'subId': 'oj-tagcloud-tooltip'});
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for tag cloud items at a specified index.</p>
 *
 * @property {number} index
 *
 * @ojnodecontext oj-tagcloud-item
 * @memberof oj.ojTagCloud
 */

/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojTagCloudMeta = {
  "properties": {
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string",
      "enumValues": ["any", "all"]
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "items": {
      "type": "Array<object>|Promise"
    },
    "layout": {
      "type": "string",
      "enumValues": ["cloud", "rectangular"]
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["single", "multiple", "none"]
    },
    "styleDefaults": {
      "type": "object",
      "properties": {
        "animationDuration": {
          "type": "number"
        },
        "hoverBehaviorDelay": {
          "type": "number"
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "touchResponse": {
      "type": "string",
      "enumValues": ["touchStart", "auto"]
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Tag Cloud"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    }
  },
  "methods": {
    "getContextByNode": {},
    "getItem": {},
    "getItemCount": {}
  },
  "extension": {
    _WIDGET_NAME: "ojTagCloud"
  }
};
oj.CustomElementBridge.registerMetadata('oj-tag-cloud', 'dvtBaseComponent', ojTagCloudMeta);
oj.CustomElementBridge.register('oj-tag-cloud', {'metadata': oj.CustomElementBridge.getMetadata('oj-tag-cloud')});
})();
/**
 * Default JET Template engine iumplementation
 * @ignore
 */
 function JetTemplateEngine()
 {
   this.execute = function(componentElement, node, properties)
   { 
     var tmpContainer = _createAndPopulateContainer(node);
     
     ko.applyBindingsToDescendants(_getContext(componentElement, properties), tmpContainer);
     
     return Array.prototype.slice.call(tmpContainer.childNodes, 0);
   }
   
   function _createAndPopulateContainer(node)
   {
     var div = document.createElement("div");

     if (node.nodeType === 1 && node.tagName.toLowerCase() === 'template') 
     {
       var content = node.content;
       if (content)
       {
         div.appendChild(document.importNode(content, true /* deep clone*/));
       }
       else
       {
         Array.prototype.forEach.call(node.childNodes,
           function(child)
           {
             div.appendChild(child.cloneNode(true));
           }
         );
       }
     }
     else
     {
       throw "Invalid template node " + node;
     }
     
     return div;
   }
   
   function _getContext(componentElement, properties)
   {
     var bindingContext = ko.contextFor(componentElement);
     
     if (bindingContext)
     {
       return bindingContext['extend'](properties);
     }
     
     return properties;
   }
 }

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojThematicMap
 * @augments oj.dvtBaseComponent
 * @since 0.7
 * @ojstatus preview
 * @ojrole application
 * @ojshortdesc An interactive data visualization that displays data corresponding to geographic locations or regions.  Applications are required to supply a mapProvider for a valid thematic map.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="thematicMapOverview-section">
 *   JET Thematic Map
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#thematicMapOverview-section"></a>
 * </h3>
 *
 * <p>Thematic maps are used to display data corresponding to geographic locations
 * or regions, such as election data for a state or sales by territory for a product. 
 * Applications are required to supply a mapProvider for a valid thematic map.</p>
 *
 * {@ojinclude "name":"warning"}
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-thematic-map mapProvider='[[mapProviderObj]]'
 *   areas='[{"color":"#003366", "location":"FL"},
*            {"color":"#CC3300", "location":"TX"},
*            {"color":"#99CC33", "location":"CA"}]'>
 * &lt;/oj-thematic-map>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Styling</h4>
 * <p>Use the highest level property available. For example, consider setting styling properties on
 *    <code class="prettyprint">styleDefaults.dataAreaDefaults</code> or 
 *    <code class="prettyprint">styleDefaults.dataMarkerDefaults</code>, instead of styling properties
 *    on the individual data items. The thematic map can take advantage of these higher level properties to apply the style properties on
 *    containers, saving expensive DOM calls.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 *
 * <h3 id="mapprovider-section">
 *   Map Rendering
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#mapprovider-section"></a>
 * </h3>
 *
 * <p>
 *   Thematic map supports rendering of GeoJSON formatted geographic data. An application can specify the GeoJSON along with keys
 *   used for determining area IDs and labels by setting the the [mapProvider]{@link oj.ojThematicMap#mapProvider} attribute. 
 *   Currently only GeoJSON objects of "type" Feature or FeatureCollection are supported. Each Feature object contains
 *   the information to render a map area including the area id, coordinates, and optional short and long labels. Only Feature
 *   "geometry" objects of "type" Polygon and MutliPolgyon will be used for defining area boundaries.  All other "type" values
 *   will be skipped.  The Feature "properties" object is where the thematic map will look up area info like id, short
 *   label, and long label using the key mappings provided in the [propertiesKeys]{@link oj.ojThematicMap#mapProvider.propertiesKeys} property.
 *   See the thematic map <a href="../jetCookbook.html?component=thematicMap&demo=mapProvider">Map Provider Demo</a> for an example.
 * </p>
 * <p>
 *   If longitude/latitude coordinate data need to be rendered, the application should use a projection library to
 *   project the coordinates to the map projection before passing as x and y properties to the marker object.
 * </p>
 */
oj.__registerWidget('oj.ojThematicMap', $['oj']['dvtBaseComponent'],
  {
    widgetEventPrefix: "oj",
    options: {
      /**
       * We recommend using the component CSS classes to set component wide styling. This API should be used
       * only for styling a specific instance of the component. The default values come from the CSS classess and 
       * varies based on theme. The duration of the animations in milliseconds.
       * @expose
       * @name animationDuration
       * @memberof oj.ojThematicMap
       * @instance
       * @type {number}
       * @ojunits "milliseconds"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">animation-duration</code> attribute specified:</caption>
       * &lt;oj-thematic-map animation-duration='200'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">animationDuration</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.animationDuration;
       * 
       * // setter
       * myThematicMap.animationDuration=200;
       */
      animationDuration: undefined,
      /**
       * The type of animation to apply when the element is initially displayed.
       * @expose
       * @name animationOnDisplay
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "auto"
       * @ojvalue {string} "none"
       * @default "none"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">animation-on-display</code> attribute specified:</caption>
       * &lt;oj-thematic-map animation-on-display='auto'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">animationOnDisplay</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.animationOnDisplay;
       * 
       * // setter
       * myThematicMap.animationOnDisplay="auto";
       */
      animationOnDisplay: 'none',
      /**
       * An array of objects that define a row of data for an area data layer. 
       * Also accepts a Promise where no data will be rendered if the Promise is rejected. Regardless of the
       * set value type, we will wrap and return a Promise when accessing the areas property.
       * @expose
       * @name areas
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Array.<Object>|Promise|null}
       * @ojsignature {target: "Accessor", value: {GetterType: "Promise<Array<oj.ojThematicMap.Area>>|null", SetterType: "Array<oj.ojThematicMap.Area>|Promise<Array<oj.ojThematicMap.Area>>|null"}, jsdocOverride: true}
       * @default null
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">areas</code> attribute specified:</caption>
       * &lt;oj-thematic-map areas='[{"id": "a1", "color": "red", "location": "FRA"}, 
       *                             ...
       *                             {"id": "a27", "color": "green", "location": "USA"}]'>
       * &lt;/oj-thematic-map>
       * 
       * &lt;oj-thematic-map areas='[[areasPromise]]'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">areas</code> 
       * property after initialization:</caption>
       * 
       * // Get all (The areas getter always returns a Promise so there is no "get one" syntax)
       * var values = myThematicMap.areas;
       * 
       * // Set all (There is no permissible "set one" syntax.)
       * myThematicMap.areas=[{"id": "a1", "color": "red", "location": "FRA"}, 
       *              ...
       *              {"id": "a27", "color": "green", "location": "USA"}];
       */
      areas: null,
      /**
       * An optional callback function to update the data item in response to changes in keyboard focus state.
       * @expose
       * @name focusRenderer
       * @memberof oj.ojThematicMap
       * @instance
       * @type {function(Object):(Object|void)|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojThematicMap.RendererContext) => ({insert: SVGElement}|void))", jsdocOverride: true}
       * @default null
       */
      focusRenderer: null,
      /**
       * An array of category strings used for category filtering. Data items with a category in 
       * hiddenCategories will be filtered.
       * @expose
       * @name hiddenCategories
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Array.<string>}
       * @default []
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">hidden-categories</code> attribute specified:</caption>
       * &lt;oj-thematic-map hidden-categories='["soda", "water"]'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">hiddenCategories</code> 
       * property after initialization:</caption>
       * // Get one
       * var value = myThematicMap.hiddenCategories[0];
       * 
       * // Get all
       * var values = myThematicMap.hiddenCategories;
       * 
       * // Set all (There is no permissible "set one" syntax.)
       * myThematicMap.hiddenCategories=["soda", "water"];
       */
      hiddenCategories: [],
      /**
       * An array of category strings used for category highlighting. Data items with a category in 
       * highlightedCategories will be highlighted.
       * @expose
       * @name highlightedCategories
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Array.<string>}
       * @default []
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">highlighted-categories</code> attribute specified:</caption>
       * &lt;oj-thematic-map highlighted-categories='["soda", "water"]'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">highlightedCategories</code> 
       * property after initialization:</caption>
       * // Get one
       * var value = myThematicMap.highlightedCategories[0];
       * 
       * // Get all
       * var values = myThematicMap.highlightedCategories;
       * 
       * // Set all (There is no permissible "set one" syntax.)
       * myThematicMap.highlightedCategories=["soda", "water"];
       */
      highlightedCategories: [],
      /**
       * The matching condition for the highlightedCategories option. By default, highlightMatch is 
       * 'all' and only items whose categories match all of the values specified in the highlightedCategories 
       * array will be highlighted. If highlightMatch is 'any', then items that match at least one of the 
       * highlightedCategories values will be highlighted.
       * @expose
       * @name highlightMatch
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "any"
       * @ojvalue {string} "all"
       * @default "all"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">highlight-match</code> attribute specified:</caption>
       * &lt;oj-thematic-map highlight-match='any'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">highlightMatch</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.highlightMatch;
       * 
       * // setter
       * myThematicMap.highlightMatch="any";
       */
      highlightMatch: 'all',
      /**
       * Defines the behavior applied when hovering over data items.
       * @expose
       * @name hoverBehavior
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "dim"
       * @ojvalue {string} "none"
       * @default "none"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">hover-behavior</code> attribute specified:</caption>
       * &lt;oj-thematic-map hover-behavior='dim'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">hoverBehavior</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.hoverBehavior;
       * 
       * // setter
       * myThematicMap.hoverBehavior="dim";
       */
      hoverBehavior: 'none',
      /**
       * An optional callback function to update the node in response to changes in hover state.
       * @expose
       * @name hoverRenderer
       * @memberof oj.ojThematicMap
       * @instance
       * @type {function(Object):(Object|void)|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojThematicMap.RendererContext) => ({insert: SVGElement}|void))", jsdocOverride: true}
       * @default null
       */
      hoverRenderer: null,
      /**
       * Specifies whether the map will zoom to fit the data objects on initial render.
       * @expose
       * @name initialZooming
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "auto"
       * @ojvalue {string} "none"
       * @default "none"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">initial-zooming</code> attribute specified:</caption>
       * &lt;oj-thematic-map initial-zooming='auto'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">initialZooming</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.initialZooming;
       * 
       * // setter
       * myThematicMap.initialZooming="auto";
       */
      initialZooming: 'none',
      /**
       * The id for the isolated area of this area data layer. If set, only the isolated area will be displayed.
       * @expose
       * @name isolatedItem
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @default ""
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">isolated-item</code> attribute specified:</caption>
       * &lt;oj-thematic-map isolated-item='a2'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">isolatedItem</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.isolatedItem;
       * 
       * // setter
       * myThematicMap.isolatedItem="a2";
       */
      isolatedItem: '',
      /**
       * Determines how labels for this layer should be displayed.
       * @expose
       * @name labelDisplay
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "on"
       * @ojvalue {string} "off"
       * @ojvalue {string} "auto" Renders the label if it fits within the area bounds.
       * @default "off"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">label-display</code> attribute specified:</caption>
       * &lt;oj-thematic-map label-display='auto'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">labelDisplay</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.labelDisplay;
       * 
       * // setter
       * myThematicMap.labelDisplay="auto";
       */
      labelDisplay: 'off',
      /**
       * Determines which type of map labels to display.
       * @expose
       * @name labelType
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "long"
       * @ojvalue {string} "short"
       * @default "short"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">label-display</code> attribute specified:</caption>
       * &lt;oj-thematic-map label-type='long'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">labelType</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.labelType;
       * 
       * // setter
       * myThematicMap.labelType="long";
       */
      labelType: 'short',
      /**
       * An array of objects that define the data for links.
       * Also accepts a Promise where no data will be rendered if the Promise is rejected. Regardless of the
       * set value type, we will wrap and return a Promise when accessing the links property.
       * @expose
       * @name links
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Array.<Object>|Promise|null}
       * @ojsignature {target: "Accessor", value: {GetterType: "Promise<Array<oj.ojThematicMap.Link>>|null", SetterType: "Array<oj.ojThematicMap.Link>|Promise<Array<oj.ojThematicMap.Link>>|null"}, jsdocOverride: true}
       * @default null
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">links</code> attribute specified:</caption>
       * &lt;oj-thematic-map links='[{"id": "l1", "startLocation": {"id": "m2"}, "endLocation": {"id": "m29"}}, 
       *                             ...
       *                             {"id": "l7", "startLocation": {"id": "m17"}, "endLocation": {"id": "m9"}}]'>
       * &lt;/oj-thematic-map>
       * 
       * &lt;oj-thematic-map links='[[markersPromise]]'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">links</code> 
       * property after initialization:</caption>
       * 
       * // Get all (The links getter always returns a Promise so there is no "get one" syntax)
       * var values = myThematicMap.links;
       * 
       * // Set all (There is no permissible "set one" syntax.)
       * myThematicMap.links=[{"id": "l1", "startLocation": {"id": "m2"}, "endLocation": {"id": "m29"}}, 
       *              ...
       *              {"id": "l7", "startLocation": {"id": "m17"}, "endLocation": {"id": "m9"}}];
       */
      links: null,
      /**
       * An object with the following properties, used to define a custom map.
       * @expose
       * @name mapProvider
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Object}
       * @default {"geo": {}, "propertiesKeys": {"id": "", "longLabel": "", "shortLabel": ""}}
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">map-provider</code> attribute specified:</caption>
       * <!-- Using dot notation -->
       * &lt;oj-thematic-map map-provider.geo='[[geoJsonObj]]'
       *   map-provider.properties-keys='{"id": "country", "shortLabel": "iso_a3", "longLabel": "country"}'>
       * &lt;/oj-thematic-map>
       * 
       * &lt;oj-thematic-map mapProvider='[[{"geo": geoJsonObj, 
       *                                     "propertiesKeys": {"id": "country", 
       *                                                        "shortLabel": "iso_a3", 
       *                                                        "longLabel": "country"}]]'>
       * &lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">mapProvider</code> 
       * property after initialization:</caption>
       * // Get one
       * var value = myThematicMap.mapProvider.geo;
       * 
       * // Get all
       * var values = myThematicMap.mapProvider;
       *
       * // Set one, leaving the others intact. Always use the setProperty API for 
       * // subproperties rather than setting a subproperty directly.
       * myThematicMap.setProperty('mapProvider.geo', geoJsonObj);
       * 
       * // Set all. Must list every resource key, as those not listed are lost.
       * myThematicMap.mapProvider={'geo': geoJsonObj};
       */
      mapProvider: {
        /**
         * The GeoJSON object containing custom area coordinates. Only GeoJSON objects of "type" Feature or 
         * FeatureCollection and Feature "geometry" objects of "type" Polygon or MultiPolygon are currently 
         * supported. Each Feature object will contain a thematic map area and each Feature's "properties" 
         * object will at a minimum need to contain a key, which can be defined in the propertiesKeys object, 
         * that will be used as the ID of the area. 
         * @expose
         * @name mapProvider.geo
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {Object}
         * @default {}
         * 
         * @example <caption>See the <a href="#mapProvider">mapProvider</a> attribute for usage examples.</caption>
         */
        geo: {},
        /**
         * The object specifying the GeoJSON Feature "properties" object keys to use for the custom 
         * area id, short label, and long label mappings.
         * @expose
         * @name mapProvider.propertiesKeys
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {Object}
         * @default {}
         * 
         * @example <caption>See the <a href="#mapProvider">mapProvider</a> attribute for usage examples.</caption>
         */
        propertiesKeys: {
          /**
           * The required name of the "properties" key to use as the location id that will map a data item to a map area.
           * @expose
           * @name mapProvider.propertiesKeys.id
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @default ""
           * 
           * @example <caption>See the <a href="#mapProvider">mapProvider</a> attribute for usage examples.</caption>
           */
          id: "",
          /**
           * The optional name of the "properties" key to use for rendering area labels when labelType is set to "short".
           * @expose
           * @name mapProvider.propertiesKeys.shortLabel
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * @default ""
           * 
           * @example <caption>See the <a href="#mapProvider">mapProvider</a> attribute for usage examples.</caption>
           */
          shortLabel: "",
          /**
           * The optional name of the "properties" key to use for rendering area labels when labelType is set to "long".
           * @expose
           * @name mapProvider.propertiesKeys.longLabel
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * @default ""
           * 
           * @example <caption>See the <a href="#mapProvider">mapProvider</a> attribute for usage examples.</caption>
           */
          longLabel: ""
        }
      },
      /**
       * An array of objects that define a row of data for a data layer.
       * Also accepts a Promise where no data will be rendered if the Promise is rejected. Regardless of the
       * set value type, we will wrap and return a Promise when accessing the markers property.
       * @expose
       * @name markers
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Array.<Object>|Promise|null}
       * @ojsignature {target: "Accessor", value: {GetterType: "Promise<Array<oj.ojThematicMap.Marker>>|null", SetterType: "Array<oj.ojThematicMap.Marker>|Promise<Array<oj.ojThematicMap.Marker>>|null"}, jsdocOverride: true}
       * @default null
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">markers</code> attribute specified:</caption>
       * &lt;oj-thematic-map markers='[{"id": "m1", "color": "red", "shape": "circle", x": 2102, "y": 910}, 
       *                             ...
       *                             {"id": "m27", "color": "green", "shape": "circle", "x": 4820, "y": 277}]'>
       * &lt;/oj-thematic-map>
       * 
       * &lt;oj-thematic-map markers='[[markersPromise]]'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">markers</code> 
       * property after initialization:</caption>
       * 
       * // Get all (The markers getter always returns a Promise so there is no "get one" syntax)
       * var values = myThematicMap.markers;
       * 
       * // Set all (There is no permissible "set one" syntax.)
       * myThematicMap.markers=[{"id": "m1", "color": "red", "shape": "circle", x": 2102, "y": 910}, 
       *                ...
       *                {"id": "m27", "color": "green", "shape": "circle", "x": 4820, "y": 277}];
       */
      markers: null,
      /**
       * Specifies marker behavior on zoom.
       * @expose
       * @name markerZoomBehavior
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "zoom"
       * @ojvalue {string} "fixed"
       * @default "fixed"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">marker-zoom-behavior</code> attribute specified:</caption>
       * &lt;oj-thematic-map marker-zoom-behavior='zoom'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">markerZoomBehavior</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.markerZoomBehavior;
       * 
       * // setter
       * myThematicMap.markerZoomBehavior="zoom";
       */
      markerZoomBehavior: 'fixed',
      /**
       * Specifies the maximum zoom level for this element. This can be any number greater than 1.0 
       * which indicates the maximum point to which the map can be scaled. A value of 2.0 implies that 
       * the map can be zoomed in until it reaches twice the viewport size. A maxZoom of 1.0 indicates 
       * that the user cannot zoom the map beyond the viewport size.
       * @expose
       * @name maxZoom
       * @memberof oj.ojThematicMap
       * @instance
       * @type {number}
       * @default 6.0
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">max-zoom</code> attribute specified:</caption>
       * &lt;oj-thematic-map max-zoom='10'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">maxZoom</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.maxZoom;
       * 
       * // setter
       * myThematicMap.maxZoom=10;
       */
      maxZoom: 6.0,
      /**
       * Determines whether element panning is allowed.
       * @expose
       * @name panning
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "auto"
       * @ojvalue {string} "none"
       * @default "none"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">panning</code> attribute specified:</caption>
       * &lt;oj-thematic-map panning='auto'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">panning</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.panning;
       * 
       * // setter
       * myThematicMap.panning="auto";
       */
      panning: 'none',
      /**
       * A callback function used to stamp custom SVG elements for a data layer.
       * @expose
       * @name renderer
       * @memberof oj.ojThematicMap
       * @instance
       * @type {function(Object):(Object|void)|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojThematicMap.RendererContext) => ({insert: SVGElement}|void))", jsdocOverride: true}
       * @default null
       */
      renderer: null,
      /**
       * An array of id strings, used to define the selected data items.
       * @expose
       * @name selection
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Array.<string>}
       * @default []
       * @ojwriteback
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">selection</code> attribute specified:</caption>
       * &lt;oj-thematic-map selection='["area1", "area2", "marker2"]'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">selection</code> 
       * property after initialization:</caption>
       * // Get one
       * var value = myThematicMap.selection[0];
       * 
       * // Get all
       * var values = myThematicMap.selection;
       * 
       * // Set all (There is no permissible "set one" syntax.)
       * myThematicMap.selection=["area1", "area2", "marker2"];
       */
      selection: [],
      /**
       * The type of selection behavior that is enabled on the thematic map.
       * @expose
       * @name selectionMode
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "single"
       * @ojvalue {string} "multiple"
       * @ojvalue {string} "none"
       * @default "none"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">selection-mode</code> attribute specified:</caption>
       * &lt;oj-thematic-map selection-mode='multiple'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.selectionMode;
       * 
       * // setter
       * myThematicMap.selectionMode="multiple";
       */
      selectionMode: 'none',
      /**
       * An optional callback function to update the data item in response to changes in selection state.
       * @expose
       * @name selectionRenderer
       * @memberof oj.ojThematicMap
       * @instance
       * @type {function(Object):(Object|void)|null}
       * @ojsignature {target: "Type", value: "((context: oj.ojThematicMap.RendererContext) => ({insert: SVGElement}|void))", jsdocOverride: true}
       * @default null
       */
      selectionRenderer: null,
      /**
       * We recommend using the component CSS classes to set component wide styling. This API should be used
       * only for styling a specific instance of the component. Properties specified on this object may 
       * be overridden by specifications on the data item. The default values come from the CSS classess and 
       * varies based on theme.
       * @expose
       * @name styleDefaults
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Object}
       * @default {"areaSvgStyle": {}, "dataAreaDefaults": {}, "dataMarkerDefaults": {"borderWidth": 0.5, "borderStyle": "solid", "height": 8, "labelStyle": {}, "opacity": 1, "shape": "circle", "width": 8},"hoverBehaviorDelay": 200, "labelStyle": {}, "linkDefaults": {"width": 2}}
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">style-defaults</code> attribute specified:</caption>
       * <!-- Using dot notation -->
       * &lt;oj-thematic-map style-defaults.animation-duration='200'>&lt;/oj-thematic-map>
       * 
       * <!-- Using JSON notation -->
       * &lt;oj-thematic-map style-defaults='{"animationDuration": 200, "areaSvgStyle": {"fill": "url(someURL@filterId)"}'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">styleDefaults</code> 
       * property after initialization:</caption>
       * // Get one
       * var value = myThematicMap.styleDefaults.animationDuration;
       * 
       * // Get all
       * var values = myThematicMap.styleDefaults;
       *
       * // Set one, leaving the others intact. Always use the setProperty API for 
       * // subproperties rather than setting a subproperty directly.
       * myThematicMap.setProperty('styleDefaults.areaSvgStyle', {'fill': 'url(someURL#filterId)'});
       * 
       * // Set all. Must list every resource key, as those not listed are lost.
       * myThematicMap.styleDefaults={'fill': 'url("someURL#filterId")'};
       */
      styleDefaults: {
        /**
         * The CSS style object defining the style of the area layer areas. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name styleDefaults.areaSvgStyle
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         * 
         * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
         */
        /**
         * An object defining the default styles for data areas. Properties specified on this object 
         * may be overridden by specifications on the data object.
         * @expose
         * @name styleDefaults.dataAreaDefaults
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         * @default {}
         * 
         * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
         */
        dataAreaDefaults: {
          /**
           * The area stroke color for the area data layer. The default value comes from the CSS and varies based on theme.
           * @expose
           * @name styleDefaults.dataAreaDefaults.borderColor
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          borderColor: undefined,
          /**
           * The hover data area border color. The default value comes from the CSS and varies based on theme.
           * @expose
           * @name styleDefaults.dataAreaDefaults.hoverColor
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          hoverColor: undefined,
          /**
           * The outer selected data area border color. The default value comes from the CSS and varies based on theme.
           * @expose
           * @name styleDefaults.dataAreaDefaults.selectedInnerColor
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          selectedInnerColor: undefined,
          /**
           * The outer selected data area border color. The default value comes from the CSS and varies based on theme.
           * @expose
           * @name styleDefaults.dataAreaDefaults.selectedOuterColor
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          selectedOuterColor: undefined
        },
        /**
         * An object defining the default styles for data markers. Properties specified on this object may be 
         * overridden by specifications on the data object.
         * @expose
         * @name styleDefaults.dataMarkerDefaults
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {Object}
         * @default  {"borderWidth": 0.5, "borderStyle": "solid", "height": 8, "labelStyle": {}, "opacity": 1, "shape": "circle", "width": 8}
         * @ojsignature {target: "Type", value: "?"}
         * 
         * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
         */
        dataMarkerDefaults: {
          /**
           * The border color. The default value comes from the CSS and varies based on theme.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.borderColor
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          borderColor: undefined,
          /**
           * The border width in pixels.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.borderWidth
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {number}
           * @ojsignature {target: "Type", value: "?"}
           * @default 0.5
           * @ojunits "pixels"
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          borderWidth: 0.5,
          /**
           * The border style.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.borderStyle
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojvalue {string} "none"
           * @ojvalue {string} "solid"
           * @default "solid"
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          borderStyle: "solid",
          /**
           * The fill color of a marker. The default value comes from the CSS and varies based on theme.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.color
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          color: undefined,
          /**
           * The default marker pixel height. Note that this option will be ignored if a value is provided to calculate marker sizes.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.height
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {number}
           * @default 8
           * @ojunits "pixels"
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          height: 8,
          /**
           * The CSS style for a marker label.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.labelStyle
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {Object}
           * @default  {}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          labelStyle: {},
          /**
           * The default marker opacity.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.opacity
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {number}
           * @default  1
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          opacity: 1,
          /**
           * The default marker shape. Can take the name of a built-in shape or the svg path commands for a custom shape.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.shape
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @default  "circle"
           * @ojsignature {target: "Type", value: "?'circle'|'diamond'|'ellipse'|'human'|'plus|'rectangle'|'square'|'star'|'triangleDown'|'triangleUp'|string"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          shape: 'circle',
          /**
           * The default marker pixel width. Note that this option will be ignored if a value is provided to calculate marker sizes.
           * @expose
           * @name styleDefaults.dataMarkerDefaults.width
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {number}
           * @default 8
           * @ojunits "pixels"
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          width: 8
        },
        /**
         * Specifies initial hover delay in ms for highlighting data items.
         * @expose
         * @name styleDefaults.hoverBehaviorDelay
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {number}
         * @default 200
         * @ojunits "milliseconds"
         * @ojsignature {target: "Type", value: "?"}
         * 
         * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
         */
        hoverBehaviorDelay: 200,
        /**
         * The CSS style for the area layer labels.
         * @expose
         * @name styleDefaults.labelStyle
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {Object}
         * @default  {}
         * @ojsignature {target: "Type", value: "?"}
         * 
         * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
         */
        labelStyle: {},
        /**
         * An object defining the default styles for data areas. Properties specified on this object may be 
         * overridden by specifications on the data object.
         * @expose
         * @name styleDefaults.linkDefaults
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {Object}
         * @default  {}
         * @ojsignature {target: "Type", value: "?"}
         * 
         * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
         */
        linkDefaults: {
          /**
           * The stroke color for links. The default value comes from the CSS and varies based on theme.
           * @expose
           * @name styleDefaults.linkDefaults.color
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          color: undefined,
          /**
           * The stroke width for links in pixels.
           * @expose
           * @name styleDefaults.linkDefaults.width
           * @memberof! oj.ojThematicMap
           * @instance
           * @type {number}
           * @default  2
           * @ojunits "pixels"
           * @ojsignature {target: "Type", value: "?"}
           * 
           * @example <caption>See the <a href="#styleDefaults">styleDefaults</a> attribute for usage examples.</caption>
           */
          width: 2,
        }
      },
      /**
       * An object containing an optional callback function for tooltip customization. 
       * @expose
       * @name tooltip
       * @memberof oj.ojThematicMap
       * @instance
       * @type {Object}
       * @default {"renderer": null}
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">tooltip</code> attribute specified:</caption>
       * <!-- Using dot notation -->
       * &lt;oj-thematic-map tooltip.renderer='[[tooltipFun]]'>&lt;/oj-thematic-map>
       * 
       * &lt;oj-thematic-map tooltip='[[{"renderer": tooltipFun}]]'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">hiddenCategories</code> 
       * property after initialization:</caption>
       * // Get one
       * var value = myThematicMap.tooltip.renderer;
       * 
       * // Get all
       * var values = myThematicMap.tooltip;
       *
       * // Set one, leaving the others intact. Always use the setProperty API for 
       * // subproperties rather than setting a subproperty directly.
       * myThematicMap.setProperty('tooltip.renderer', tooltipFun);
       * 
       * // Set all. Must list every resource key, as those not listed are lost.
       * myThematicMap.tooltip={'renderer': tooltipFun};
       */
      tooltip: {
        /**
         * A function that returns a custom tooltip.
         * @expose
         * @name tooltip.renderer
         * @memberof! oj.ojThematicMap
         * @instance
         * @type {function(Object):Object|null}
         * @ojsignature {target: "Type", value: "((context: oj.ojThematicMap.TooltipContext) => ({insert: Element|string}|{preventDefault: boolean}))", jsdocOverride: true}
         * @example <caption>See the <a href="#tooltip">tooltip</a> attribute for usage examples.</caption>
         */
        renderer: null
      },
      /**
       * Specifies the tooltip behavior of the thematic map.
       * @expose
       * @name tooltipDisplay
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "auto"
       * @ojvalue {string} "labelAndShortDesc"
       * @ojvalue {string} "none"
       * @ojvalue {string} "shortDesc"
       * @default "auto"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">tooltip-display</code> attribute specified:</caption>
       * &lt;oj-thematic-map tooltip-display='none'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">tooltipDisplay</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.tooltipDisplay;
       * 
       * // setter
       * myThematicMap.tooltipDisplay="none";
       */
      tooltipDisplay: 'auto',
      /**
       * Data visualizations require a press and hold delay before triggering tooltips and 
       * rollover effects on mobile devices to avoid interfering with page panning, but these 
       * hold delays can make applications seem slower and less responsive. For a better user 
       * experience, the application can remove the touch and hold delay when data visualizations 
       * are used within a non scrolling container or if there is sufficient space outside of the 
       * visualization for panning. If touchResponse is touchStart the element will instantly 
       * trigger the touch gesture and consume the page pan events if the element does not require 
       * an internal feature that requires a touch start gesture like panning or zooming. 
       * If touchResponse is auto, the element will behave like touchStart if it determines that 
       * it is not rendered within scrolling content and if element panning is not available for 
       * those elements that support the feature. 
       * @expose
       * @name touchResponse
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "touchStart"
       * @ojvalue {string} "auto"
       * @default "auto"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">touch-response</code> attribute specified:</caption>
       * &lt;oj-thematic-map touch-response='touchStart'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">touchResponse</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.touchResponse;
       * 
       * // setter
       * myThematicMap.touchResponse="touchStart";
       */
      touchResponse: 'auto',
      /**
       * Determines whether element zooming is allowed.
       * @expose
       * @name zooming
       * @memberof oj.ojThematicMap
       * @instance
       * @type {string}
       * @ojvalue {string} "auto"
       * @ojvalue {string} "none"
       * @default "none"
       * 
       * @example <caption>Initialize the thematic map with the 
       * <code class="prettyprint">zooming</code> attribute specified:</caption>
       * &lt;oj-thematic-map zooming='auto'>&lt;/oj-thematic-map>
       * 
       * @example <caption>Get or set the <code class="prettyprint">zooming</code> 
       * property after initialization:</caption>
       * // getter
       * var value = myThematicMap.zooming;
       * 
       * // setter
       * myThematicMap.zooming="auto";
       */
      zooming: 'none'
    },
    _currentLocale : '',
    _loadedBasemaps : [],
    _basemapPath : 'resources/internal-deps/dvt/thematicMap/basemaps/',
    _supportedLocales: {
      'ar' : 'ar',
      'cs' : 'cs',
      'da' : 'da',
      'de' : 'de',
      'el' : 'el',
      'es' : 'es',
      'fi' : 'fi',
      'fr' : 'fr',
      'fr-ca' : 'fr_CA',
      'he' : 'iw',
      'hu' : 'hu',
      'it' : 'it',
      'ja' : 'ja',
      'ko' : 'ko',
      'nl' : 'nl',
      'no' : 'no',
      'pl' : 'pl',
      'pt' : 'pt_BR',
      'pt-pt' : 'pt',
      'ro' : 'ro',
      'ru' : 'ru',
      'sk' : 'sk',
      'sv' : 'sv',
      'th' : 'th',
      'tr' : 'tr',
      'zh-hans' : 'zh_CN',
        'zh-hant' : 'zh_TW'},

    // @inheritdoc
    _ComponentCreate : function() {
      this._super();
      this._checkBasemaps = [];
      this._initiallyRendered = false;
      this._dataLayersToUpdate = [];
    },

    // @inheritdoc
    _CreateDvtComponent: function(context, callback, callbackObj) {
      return dvt.ThematicMap.newInstance(context, callback, callbackObj);
    },

    // @inheritdoc
    _ConvertLocatorToSubId : function(locator) {
      var subId = locator['subId'];

      // Convert the supported locators
      if(subId == 'oj-thematicmap-area') {
        // dataLayerId:area[index]
        subId = this._getDataLayerId(locator['dataLayer'], locator['index'], 'area') + ':area[' + locator['index'] + ']';
      }
      else if(subId == 'oj-thematicmap-marker') {
        // dataLayerId:marker[index]
        subId = this._getDataLayerId(locator['dataLayer'], locator['index'], 'marker') + ':marker[' + locator['index'] + ']';
      }
      else if(subId == 'oj-thematicmap-link') {
        // dataLayerId:link[index]
        subId = this._getDataLayerId(locator['dataLayer'], locator['index'], 'link') + ':link[' + locator['index'] + ']';
      }
      else if(subId == 'oj-thematicmap-tooltip') {
        subId = 'tooltip';
      }
      // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
      // support for the old subId syntax in 1.2.0.
      return subId;
    },

    // @inheritdoc
    _ConvertSubIdToLocator : function(subId) {
      var locator = {};

      if(subId.indexOf(':area') > 0) {
        // dataLayerId:area[index]
        locator['subId'] = 'oj-thematicmap-area';
        if (!this._IsCustomElement())
          locator['dataLayer'] = subId.substring(0, subId.indexOf(':'));
        locator['index'] = this._GetFirstIndex(subId);
      }
      else if(subId.indexOf(':marker') > 0) {
        // dataLayerId:marker[index]
        locator['subId'] = 'oj-thematicmap-marker';
        if (!this._IsCustomElement())
          locator['dataLayer'] = subId.substring(0, subId.indexOf(':'));
        locator['index'] = this._GetFirstIndex(subId);
      }
      else if(subId.indexOf(':link') > 0) {
        // dataLayerId:link[index]
        locator['subId'] = 'oj-thematicmap-link';
        if (!this._IsCustomElement())
          locator['dataLayer'] = subId.substring(0, subId.indexOf(':'));
        locator['index'] = this._GetFirstIndex(subId);
      }
      else if(subId == 'tooltip') {
        locator['subId'] = 'oj-thematicmap-tooltip';
      }

      return locator;
    },

    // @inheritdoc
    _GetComponentStyleClasses: function() {
      var styleClasses = this._super();
      styleClasses.push('oj-thematicmap');
      return styleClasses;
    },

    // @inheritdoc
    _GetChildStyleClasses: function() {
      var styleClasses = this._super();
      styleClasses['oj-dvtbase oj-thematicmap'] = {'path': 'animationDuration', 'property': 'ANIM_DUR'};
      styleClasses['oj-thematicmap-arealayer'] = [
        {'path': 'styleDefaults/areaSvgStyle', 'property': 'BACKGROUND'},
        {'path': 'styleDefaults/labelStyle', 'property': 'TEXT'}
      ];
      styleClasses['oj-thematicmap-area'] = {'path': 'styleDefaults/dataAreaDefaults/borderColor', 'property': 'border-top-color'};
      styleClasses['oj-thematicmap-area oj-hover'] = {'path': 'styleDefaults/dataAreaDefaults/hoverColor', 'property': 'border-top-color'};
      styleClasses['oj-thematicmap-area oj-selected'] = [
        {'path': 'styleDefaults/dataAreaDefaults/selectedInnerColor', 'property': 'border-top-color'},
        {'path': 'styleDefaults/dataAreaDefaults/selectedOuterColor', 'property': 'border-bottom-color'}];
      styleClasses['oj-thematicmap-marker'] = [
        {'path': 'styleDefaults/dataMarkerDefaults/labelStyle', 'property': 'TEXT'},
        {'path': 'styleDefaults/dataMarkerDefaults/color', 'property': 'background-color'},
        {'path': 'styleDefaults/dataMarkerDefaults/opacity', 'property': 'opacity'},
        {'path': 'styleDefaults/dataMarkerDefaults/borderColor', 'property': 'border-top-color'}
      ];
      styleClasses['oj-thematicmap-link'] = {'path': 'styleDefaults/linkDefaults/color', 'property': 'color'};
      styleClasses['oj-thematicmap-link oj-hover'] = {'path': 'styleDefaults/linkDefaults/_hoverColor', 'property': 'color'};
      styleClasses['oj-thematicmap-link oj-selected'] = {'path': 'styleDefaults/linkDefaults/_selectedColor', 'property': 'border-color'};
      return styleClasses;
    },

    // @inheritdoc
    _GetEventTypes : function() {
      return ['optionChange'];
    },

    //** @inheritdoc */
    _InitOptions : function(originalDefaults, constructorOptions) {
      this._super(originalDefaults, constructorOptions);

      // styleDefaults subproperty defaults are dynamically generated
      // so we need to retrieve it here and override the dynamic getter by
      // setting the returned object as the new value.
      var styleDefaults = this.options["styleDefaults"];
      this.options["styleDefaults"] = styleDefaults;
    },

    // @inheritdoc
    _setOptions : function (options, flags) {
      // determine if option change is a data layer update and save data to call data layer update API instead of render in _Render
      var numUpdates = Object.keys(options).length;
      var newAreaLayers = options['areaLayers'];
      var oldAreaLayers = this.options['areaLayers'];
      var pointDataLayers = options['pointDataLayers'];
      if (numUpdates == 1 && newAreaLayers && oldAreaLayers && oldAreaLayers.length > 0) {
        for (var i = 0; i < newAreaLayers.length; i++) {
          var newAreaLayer = newAreaLayers[i];
          var currAreaLayer = oldAreaLayers[i];
          var updateDataLayer = true;
          for (var areaLayerKey in newAreaLayer) {
            // check to see if option update is a data layer update by seeing if any other area layer property is changed
            if (areaLayerKey != 'areaDataLayer' && newAreaLayer[areaLayerKey] != currAreaLayer[areaLayerKey]) {
              updateDataLayer = updateDataLayer && false;
            }
          }
          if (updateDataLayer && !oj.Object.compareValues(currAreaLayer['areaDataLayer'], newAreaLayer['areaDataLayer'])) {
            this._dataLayersToUpdate.push({'options': newAreaLayer['areaDataLayer'], 'parentLayer': newAreaLayer['layer'], 'isADL': true});
          }
        }
      }
      else if (numUpdates == 1 && pointDataLayers && this.options['pointDataLayers'] && this.options['pointDataLayers'].length > 0) {
        for (var i = 0; i < pointDataLayers.length; i++) {
          if (!oj.Object.compareValues(this.options['pointDataLayers'][i], pointDataLayers[i])) {
            this._dataLayersToUpdate.push({'options': pointDataLayers[i], 'parentLayer': pointDataLayers[i]['id'], 'isADL': false});
          }
        }
      }
      this._super(options, flags);
     },

    // @inheritdoc
    _GetComponentRendererOptions: function() {
      return ['tooltip/renderer','_tooltip/renderer','renderer','focusRenderer','hoverRenderer','selectionRenderer'];
    },

    // @inheritdoc
    _ProcessOptions: function() {
      this._super();

      // wrap tooltip function in try catch
      var tooltipObj = this.options['tooltip'];
      var tooltipFun = tooltipObj ? tooltipObj['renderer'] : null;
      if (tooltipFun) {
        var self = this;
        this.options['_tooltip'] = {
          'renderer': function(context) {
            var defaultTooltip = self._IsCustomElement() ? 
                           {'insert' : context['tooltip']} : context['tooltip'];
            try {
              var tooltip = tooltipFun(context);
              return tooltip || defaultTooltip;
            } catch (error) {
              oj.Logger.warn("Showing default tooltip. " + error);
              return defaultTooltip;
            }
          }
        };
      }

      var areaLayers = this.options['areaLayers'];
      // call custom renderers
      if (areaLayers) {
        for (var i = 0; i < areaLayers.length; i++) {
          var areaDataLayer = areaLayers[i]['areaDataLayer'];
          if (areaDataLayer) {
            var renderer = areaDataLayer['_templateRenderer'];
            if (renderer)
              areaDataLayer['renderer'] = this._GetTemplateDataRenderer(renderer, 'area');
          }
        }
      }
      var pointDataLayers = this.options['pointDataLayers'];
      if (pointDataLayers) {
        for (var i = 0; i < pointDataLayers.length; i++) {
          var pointDataLayer = pointDataLayers[i];
          if (pointDataLayer) {
            var renderer = pointDataLayer['_templateRenderer'];
            if (renderer)
              pointDataLayer['renderer'] = this._GetTemplateDataRenderer(renderer, 'point');
          }
        }
      }

      // callback function for getting the context needed for custom renderers
      this.options['_contextHandler'] = this._getContextHandler();
    },

    // @inheritdoc
    _Render: function() {
      this._NotReady();
      
      var areaLayers = this.options['areaLayers'];
      // Don't render unless a basemap and area layer or at least a mapProvider for custom element are provided
      if (this._IsCustomElement()) {
        if (!this.options['mapProvider']['geo']['type']) {
          this._MakeReady();
          return;
        }
      } else {
        if (!this.options['basemap'] || !areaLayers || areaLayers.length < 1) {
          this._MakeReady();
          return;
        }

        // For thematic map, we must ensure that all basemaps are loaded before rendering.  If basemaps are still loading,
        // return and wait for the load listener to call _Render again.
        this._loadBasemap();
        for (var i = 0; i < this._checkBasemaps.length; i++) {
          if (!this._loadedBasemaps[this._checkBasemaps[i]])
            return;
        }
        this._checkBasemaps = [];
      }

      // do data layer updates only if we've already initially rendered the thematic map
      if (this._initiallyRendered && this._dataLayersToUpdate.length > 0) {
        // Fix 18498656: If the component is not attached to a visible subtree of the DOM, rendering will fail because
        // getBBox calls will not return the correct values.
        // Note: Checking offsetParent() does not work here since it returns false for position: fixed.
        if(this._context.isReadyToRender()) {
          for (var i = 0; i < this._dataLayersToUpdate.length; i++) {
            var dl = this._dataLayersToUpdate[i];
            var isAdl = dl['isADL'];
            if (isAdl)
              this._CleanTemplate('area');
            else
              this._CleanTemplate('point');
            this._component.updateLayer(dl['options'], dl['parentLayer'], isAdl);
          }
          this._dataLayersToUpdate = [];
        }
      } else {
        // Delegate to the super to call the shared JS component for actual rendering.
        this._super();
        this._initiallyRendered = true;
      }
    },

    // @inheritdoc
    _RenderComponent : function(options, isResize) {
      // Map old to new APIs right before render so we handle deferred data case as well
      if (this._IsCustomElement()) {
        this._mapCustomElementOptions(options);
      }
      // parse the top level selection map and populate data layer selection option
      this._updateDataLayerSelection(options); 

      this._super(options, isResize);
    },

    /**
     * Creates a callback function that will be used by a data item to populate context for its custom renderer
     * @return {Function} context handler callback used to create context for a custom renderer
     * @private
     * @instance
     * @memberof oj.ojThematicMap
     */
    _getContextHandler: function() {
     var thisRef = this;
     var contextHandlerFunc = function (parentElement, rootElement, data, state, previousState) {
       var context = {
         'component': oj.Components.__GetWidgetConstructor(thisRef.element),
         'parentElement': parentElement,
         'rootElement': rootElement,
         'data': data,
         'state': state,
         'previousState': previousState,
         'id': data['id'],
         'label': data['label'],
         'color': data['color'],
         'location': data['location'],
         'x': data['x'],
         'y': data['y']
       };
       if (thisRef._IsCustomElement()) {
        context['renderDefaultHover'] = thisRef.renderDefaultHover.bind(thisRef, context);
        context['renderDefaultSelection'] = thisRef.renderDefaultSelection.bind(thisRef, context);
        context['renderDefaultFocus'] = thisRef.renderDefaultFocus.bind(thisRef, context);
       }
       return thisRef._FixRendererContext(context);
     }
     return contextHandlerFunc;
    },

    /**
     * Renders the default hover effect for a data item.
     * @ojsignature {target: "Type", value: "oj.ojThematicMap.RendererContext", jsdocOverride: true}
     * @param {Object} context A context object.
     * @return {void}
     * @expose
     * @ignore
     * @instance
     * @memberof oj.ojThematicMap
     */
    renderDefaultHover: function (context) {
      if (!context['previousState'] || context['state']['hovered'] != context['previousState']['hovered']) {
        this._component.processDefaultHoverEffect(context['id'], context['state']['hovered']);
      }
    },

    /**
     * Renders the default selection effect for a data item.
     * @ojsignature {target: "Type", value: "oj.ojThematicMap.RendererContext", jsdocOverride: true}
     * @param {Object} context A context object.
     * @return {void}
     * @expose
     * @ignore
     * @instance
     * @memberof oj.ojThematicMap
     */
    renderDefaultSelection: function (context) {
      if (!context['previousState'] || context['state']['selected'] != context['previousState']['selected']) {
        this._component.processDefaultSelectionEffect(context['id'], context['state']['selected']);
      }
    },

    /**
     * Renders the default focus effect for a data item.
     * @ojsignature {target: "Type", value: "oj.ojThematicMap.RendererContext", jsdocOverride: true}
     * @param {Object} context A context object.
     * @return {void}
     * @expose
     * @ignore
     * @instance
     * @memberof oj.ojThematicMap
     */
    renderDefaultFocus: function (context) {
      if (!context['previousState'] || context['state']['focused'] != context['previousState']['focused']) {
        this._component.processDefaultFocusEffect(context['id'], context['state']['focused']);
      }
    },

    /**
     * Loads the basemaps and resource bundles.
     * @private
     */
    _loadBasemap: function() {
      var basemap = this.options['basemap'];
      if (basemap) {
        var locale = oj.Config.getLocale();
        if (locale !== this._currentLocale) {
          this._currentLocale = locale;
          this._loadedBasemaps = [];
        }

        // Track basemaps that need to be loaded before rendering
        var areaLayers = this.options['areaLayers'];
        // load area layer basemaps
        if (areaLayers) {
          for (var i = 0; i < areaLayers.length; i++) {
            var layer = areaLayers[i]['layer'];
            if (layer) {
              this._loadBasemapHelper(basemap, layer, locale);
            }
          }
        }

        // load city basemap
        var pointDataLayers = this.options['pointDataLayers'];
        // Don't try and load cities basemap if mapProvider is used
        if (!this.options['mapProvider']['geo']['type'] && pointDataLayers && pointDataLayers.length > 0)
          this._loadBasemapHelper(basemap, 'cities', locale);
      }
    },

    /**
     * Utility function for loading resource bundles by url.
     * @param {string} url The url of the resource to load
     * @param {boolean} bRenderOnFail True if we should render even if resource fails to load
     * @private
     */
    _loadResourceByUrl: function(url, bRenderOnFail) {
      // resource is already loaded or function tried to load this resource but failed
      if (this._loadedBasemaps[url])
        return;

      var thisRef = this;
      var renderCallback = function () {
        thisRef._loadedBasemaps[url] = true;
        thisRef._Render();
      }
      // TODO Update to use requirejs for internal resource bundle loading after 2.2.0(?)
      var getScript = $.getScript(oj.Config.getResourceUrl(url), function(script, textStatus) {renderCallback();});
      // Resource bundles might not get included, but this should not stop component rendering
      if (bRenderOnFail) {
        getScript.fail(function(jqxhr, settings, exception) {renderCallback();});
      }
    },

    /**
     * Helper function to load a single layer basemap and resource bundle.
     * @private
     */
    _loadBasemapHelper: function(basemap, layer, locale) {
      var isLoaded = true;
      try {
        // If the basemap is not loaded in DvtBaseMapManager then we'll get an error thrown
        // so we need to catch and load the basemap
        isLoaded = Object.keys(dvt.DvtBaseMapManager.getLayerIds(basemap, layer)).length > 0;
      } catch (err) {
        isLoaded = false;
      }

      if (!isLoaded) {
        // Don't need to wait for loaded basemaps if using mapProvider
        if (!this.options['mapProvider']['geo']['type']) {
          var relativeUrl = this._basemapPath + 'ojthematicmap-' + basemap + '-' + layer + '.js';
          if (this._checkBasemaps.indexOf(relativeUrl) == -1) {
            this._checkBasemaps.push(relativeUrl);
            this._loadResourceByUrl(relativeUrl, false);
          }
        }
      }

      if (locale.indexOf('en') === -1) {
        /*
         * Split locale by subtags and try to load resource bundle that satisfies
         * Locale syntax defined by BCP 47 and should be one of the following formats:
         * 1)language_script_region
         * 2)language_region
         * 3)language_script
         * 4)language
         * 5)default (en)
         */
        var splitLocale = locale.toLowerCase().split('-');
        var localeList = [splitLocale[0]];
        if (splitLocale.length > 1) {
          localeList.unshift(splitLocale[0]+'-'+splitLocale[1]);
        }
        if (splitLocale.length > 2) {
          localeList.unshift(splitLocale[0]+'-'+splitLocale[2], splitLocale[0]+'-'+splitLocale[1]+'-'+splitLocale[2]);
        }

        basemap = basemap.charAt(0).toUpperCase() + basemap.slice(1);
        layer = layer.charAt(0).toUpperCase() + layer.slice(1);
        var bundleName = this._basemapPath + 'resourceBundles/' + basemap + layer + 'Bundle_';
        // Go thru list of supported DVT languages
        for (var i = 0; i < localeList.length; i++) {
          if (this._supportedLocales[localeList[i]]) {
            var bundleUrl = bundleName + this._supportedLocales[localeList[i]] + ".js";
            if (this._checkBasemaps.indexOf(bundleUrl) == -1) {
              this._checkBasemaps.push(bundleUrl);
              this._loadResourceByUrl(bundleUrl, true);
            }
            break;
          }
        }
      }
    },

    // @inheritdoc
    _HandleEvent: function(event) {
      var type = event['type'];
      if (!this._IsCustomElement() && type === 'selection') {
        var selection = {};
        var id = event['clientId'];
        selection[id] = event['selection'];
        if (this.options['selection']) {
          for (var dataLayerId in this.options['selection']) {
            if (id !== dataLayerId)
              selection[dataLayerId] = this.options['selection'][dataLayerId];
          }
        }
        this._UserOptionChange('selection', selection);
      } else {
        this._super(event);
      }
    },

    // @inheritdoc
    _GetTranslationMap: function() {
      // The translations are stored on the options object.
      var translations = this.options['translations'];

      // Safe to modify super's map because function guarentees a new map is returned
      var ret = this._super();
      ret['DvtUtilBundle.THEMATIC_MAP'] = translations['componentName'];
      return ret;
    },

    /**
     * Updates the options object with the current data layer selection states
     * @param {Object} options The options object to update
     * @memberof oj.ojThematicMap
     * @instance
     * @private
     */
    _updateDataLayerSelection: function(options) {
      var selection = options['selection'];
      // Assume ids are unique amongst all data items for custom elements
      // so just same selection array on all data layers
      if (selection) {
        var pdls = options['pointDataLayers'];
        if (pdls) {
          if (this._IsCustomElement() && pdls[0]) {
            pdls[0]['selection'] = selection;
          } else {
            for (var i = 0; i < pdls.length; i++) {
              if (selection[pdls[i]['id']])
                pdls[i]['selection'] = selection[pdls[i]['id']];
            }
          }
        }

        var als = options['areaLayers'];
        if (als && als[0]) {
          // JET thematic map does not support nesting of point data layers within an area layer
          var adl = als[0]['areaDataLayer'];
          if (this._IsCustomElement() && adl) {
            adl['selection'] = selection;
          } else {
            if (adl && selection[adl['id']])
              adl['selection'] = selection[adl['id']];
          }
        }
      }
    },

    /**
     * Returns an object for automation testing verification of an area with
     * the specified index in the areas property.
     * @param {number} index The index of the area in the areas Array.
     * @ojsignature {target: "Type", value: "oj.ojThematicMap.DataContext|null", jsdocOverride: true, for: "returns"}
     * @return {Object|null} An object containing properties for the area, or null if none exists.
     * @expose
     * @instance
     * @memberof oj.ojThematicMap
     */
    getArea : function(index) {
      if (this._IsCustomElement())
        return this._component.getAutomation().getData(this._getDataLayerId(null, index, 'area'), 'area', index);
      else
        return this._component.getAutomation().getData(arguments[0], 'area', arguments[1]);
    },

    /**
     * Returns an object for automation testing verification of a marker with
     * the specified index in the markers property.
     * @param {number} index The index of the marker in the markers Array.
     * @ojsignature {target: "Type", value: "oj.ojThematicMap.DataContext|null", jsdocOverride: true, for: "returns"}
     * @return {Object|null} An object containing properties for the marker, or null if none exists.
     * @expose
     * @instance
     * @memberof oj.ojThematicMap
     */
    getMarker : function(index) {
      if (this._IsCustomElement())
        return this._component.getAutomation().getData(this._getDataLayerId(null, index, 'marker'), 'marker', index);
      else
        return this._component.getAutomation().getData(arguments[0], 'marker', arguments[1]);
    },

    /**
     * Returns an object for automation testing verification of a link with
     * the specified index in the links property.
     * @param {number} index The index of the link in the links Array.
     * @ojsignature {target: "Type", value: "oj.ojThematicMap.DataContext|null", jsdocOverride: true, for: "returns"}
     * @return {Object|null} An object containing properties for the link, or null if none exists.
     * @expose
     * @instance
     * @memberof oj.ojThematicMap
     */
    getLink : function(index) {
      if (this._IsCustomElement())
        return this._component.getAutomation().getData(this._getDataLayerId(null, index, 'link'), 'link', index);
      else
        return this._component.getAutomation().getData(arguments[0], 'link', arguments[1]);
    },

    /**
     * {@ojinclude "name":"nodeContextDoc"}
     * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
     * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
     * @ojsignature {target: "Type", value: "oj.ojThematicMap.NodeContext|null", jsdocOverride: true, for: "returns"}
     *
     * @example {@ojinclude "name":"nodeContextExample"}
     *
     * @expose
     * @instance
     * @memberof oj.ojThematicMap
     */
    getContextByNode: function(node) {
      // context objects are documented with @ojnodecontext
      var context = this.getSubIdByNode(node);
      if (context) {
        if (this._IsCustomElement())
          delete context['dataLayer'];
        if (context['subId'] !== 'oj-thematicmap-tooltip')
          return context;
      }

      return null;
    },

    // @inheritdoc
    _GetComponentDeferredDataPaths : function() {
      if (this._IsCustomElement())
        return {'root': ['areas', 'markers', 'links']};
      else
        return {
          'areaLayers': ['areaDataLayer/areas', 'areaDataLayer/markers', 'areaDataLayer/links'],
          'pointDataLayers': ['markers', 'links']
        };
    },

    // @inheritdoc
    _GetComponentNoClonePaths : function() {
      // For custom elements we want to skip cloning for data and also the mapProvider object so we can 
      // internally do === checks and perform animations when the mapProvider changes
      if (this._IsCustomElement())
        return {'areas': true, 'markers': true, 'links': true, 'mapProvider': true};
      else
        return {'mapProvider': true, 'areaLayers': {'areaDataLayer': {'areas': true, 'markers': true, 'links': true}}, 'pointDataLayers': {'markers': true, 'links': true}};
    },

    // @inheritdoc
    _GetDataContext : function(options) {
      if (this._IsCustomElement()) {
        return this._super();
      } else {
        // Need to use this.options instead of options because the options
        // passed in will actually be an areaLayer or pointDataLayer options object
        var basemap = this.options['basemap'];
        var layer = options['layer'] ? options['layer'] : 'cities';
        return {'basemap': basemap, 'layer' : layer, 'ids': dvt.DvtBaseMapManager.getLayerIds(basemap, layer)};
      }
    },

    /**
     * Map custom element APIs to the widget APIs and return an object containing
     * the new area and point data layers for rendering purposes.
     * @param {string} options The options to map
     * @private
     * @instance
     * @memberof oj.ojThematicMap
     */
    _mapCustomElementOptions: function(options) {
      // AnimationOnMapChange is not exposed, map animationOnDisplay to that property instead
      var animOnDisplay = options['animationOnDisplay'];
      if (animOnDisplay)
        options['animationOnMapChange'] = animOnDisplay;

      // Create areaLayers and pointDataLayers objects for data items
      var areaLayers = [{}];
      options['areaLayers'] = areaLayers;
      var areaLayer = areaLayers[0];
      var adl = {'id': 'adl1'}; // need id for automation lookup
      areaLayer['areaDataLayer'] = adl;

      var pointDataLayers = [{'id': 'pdl1'}]; // need id for automation lookup
      options['pointDataLayers'] = pointDataLayers;
      var pdl = pointDataLayers[0];

      // New top level properties that should be mapped to the areaLayer
      var props = ['labelDisplay', 'labelType'];
      this._mapOptionHelper(options, props, [areaLayer]);

      // New top level properties that should be mapped to the area/pointDataLayers
      props = ['animationOnDataChange', 'focusRenderer', 'hoverRenderer', 'renderer', 'selectionMode', 'selectionRenderer'];
      this._mapOptionHelper(options, props, [adl, pdl]);

      // New top level properties that should be mapped to the areaDataLayer
      props = ['areas', 'isolatedItem'];
      this._mapOptionHelper(options, props, [adl]);

      // Create a map of area/marker ids to data layer ids
      this._idToDataLayerMap = {};
      var areas = options['areas'];
      if (areas) {
        for (var i = 0; i < areas.length; i++) {
          var area = areas[i];
          this._idToDataLayerMap[area.id] = 'adl1';
        }
      }

      // For markers and links we need to check whether they belong to an 
      // area/pointDataLayer by checking if location is defined. 
      // Save the old array index to the new data layer mapping for constructing
      // and deconstructing subIds. Custom element subIds lack the dataLayerId.
      var markers = options['markers'];
      if (markers) {
        this._markerToDataLayerMap = [];
        var areaMarkers = [];
        var pointMarkers = [];
        for (var i = 0; i < markers.length; i++) {
          var marker = markers[i];
          if (marker['location']) {
            areaMarkers.push(marker);
            this._markerToDataLayerMap[i] = 'adl1';
            this._idToDataLayerMap[marker.id] = 'adl1';
          }
          else if (marker['x'] && marker['y']) {
            pointMarkers.push(marker);
            this._markerToDataLayerMap[i] = 'pdl1';
            this._idToDataLayerMap[marker.id] = 'pdl1';
          }
        }

        if (areaMarkers.length > 0)
          adl['markers'] = areaMarkers;
        if (pointMarkers.length > 0)
          pdl['markers'] = pointMarkers;
      }

      var links = options['links'];
      if (links) {
        this._linkToDataLayerMap = [];
        var areaLinks = [];
        var pointLinks = [];
        for (var i = 0; i < links.length; i++) {
          var link = links[i];
          var linkStart = link['startLocation'];
          // Just check the startLocation, further validation of endLocation
          // occurs in toolkit impl
          if (linkStart && linkStart['location']) {
            areaLinks.push(link);
            this._linkToDataLayerMap[i] = 'adl1';
          }
          else if (linkStart && linkStart['x'] && linkStart['y']) {
            pointLinks.push(link);
            this._linkToDataLayerMap[i] = 'pdl1';
          }
          else if (linkStart && linkStart['id']) {
            // If id is specified, find whether it maps to an area or marker.
            // Ids should be unique so we can stop at the first match.
            var dlId = this._idToDataLayerMap[linkStart['id']];
            if (dlId === 'adl1')
              areaLinks.push(link);
            else
              pointLinks.push(link);
            this._linkToDataLayerMap[i] = dlId;
          }
        }

        if (areaLinks.length > 0)
          adl['links'] = areaLinks;
        if (pointLinks.length > 0)
          pdl['links'] = pointLinks;
      }
    },

    /**
     * Helper method to map an custom element API to the old API.
     * @private
     * @instance
     * @memberof oj.ojThematicMap
     */
    _mapOptionHelper: function(options, props, optionObjs) {
      for (var i = 0; i < props.length; i++) {
        var key = props[i];
        var value = options[key];
        if (value) {
          for (var j = 0; j < optionObjs.length; j++)
            optionObjs[j][key] = value;
        }
      }
    },

    /**
     * Returns the dataLayerId used for a subId. Returns empty string for
     * custom elements
     * @param  {string} dataLayer The data layer or null if known
     * @param  {number} index The data item index
     * @param  {string} dataItemType Enum of type area, marker or link
     * @return {string}
     * @private
     */
    _getDataLayerId: function(dataLayer, index, dataItemType) {
      if (this._IsCustomElement()) {
        switch(dataItemType) {
          case 'area':
            return 'adl1';
          case 'marker':
            return this._markerToDataLayerMap[index];
          case 'link':
            return this._linkToDataLayerMap[index];
          default:
            return '';
        }
      } else {
        return dataLayer;
      }
    },
});

// Conditionally set the defaults for custom element vs widget syntax since we expose different APIs
oj.Components.setDefaultOptions({
  'ojThematicMap': {
    'styleDefaults': oj.Components.createDynamicPropertyGetter(function(context) {
      if (context['isCustomElement'])
        return {
          'areaSvgStyle': {}
        };
      return {};
    }),
  }
});
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="3">Data Item</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="3">Element</td>
 *       <td><kbd>Drag</kbd></td>
 *       <td>Pan when <code class="prettyprint">panning</code> is <code class="prettyprint">auto</code>.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Pinch-close</kbd></td>
 *       <td>Zoom out when <code class="prettyprint">zooming</code> is <code class="prettyprint">auto</code>.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Spread-open</kbd></td>
 *       <td>Zoom in when <code class="prettyprint">zooming</code> is <code class="prettyprint">auto</code>.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojThematicMap
 */
 
/**
 *<table class="keyboard-table">
 *  <thead>
 *    <tr>
 *      <th>Key</th>
 *      <th>Action</th>
 *    </tr>
 *  </thead>
 *  <tbody>
 *  <tr>
 *     <td><kbd>Tab</kbd></td>
 *     <td>Move focus to map and then to next element.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Shift + Tab</kbd></td>
 *     <td>Move focus to map and then to previous element.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>= or +</kbd></td>
 *     <td>Zoom in one level if zooming is enabled.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>- or _</kbd></td>
 *     <td>Zoom out one level if zooming is enabled.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>0</kbd></td>
 *     <td>Zoom to fit map if zooming is enabled.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Ctrl + Alt + 0</kbd></td>
 *     <td>Zoom to fit region with focus.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Ctrl + 0</kbd></td>
 *     <td>Zoom to fit selected regions.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Ctrl + Shift + 0</kbd></td>
 *     <td>Reset map.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>PageUp</kbd></td>
 *     <td>Pan up.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>PageDown</kbd></td>
 *     <td>Pan down.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Shift + PageUp</kbd></td>
 *     <td>Pans left in left to right locales.  Pans right in right to left locales.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Shift + PageDown</kbd></td>
 *     <td>Pans right in left to right locales.  Pans left in right to left locales.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>LeftArrow or RightArrow</kbd></td>
 *     <td>Move focus and selection to the left or right nearest data item in the collection (e.g. areas, markers, links) or 
 *         the left link end marker if the focus is on a link and Alt + &lt; or Alt + &gt; was used to move there.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>UpArrow or DownArrow</kbd></td>
 *     <td>Move focus and selection to the above or below nearest data item in the collection (e.g. areas, markers, links) or
 *         to the next link above that is associated with the previous data item,
 *         if the focus is on a link and Alt + &lt; or Alt + &gt; was used to move there.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Shift + LeftArrow or Shift + RightArrow</kbd></td>
 *     <td>Move focus and multi-select the left or right nearest data item in the collection (e.g. areas, markers, links).</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Shift + UpArrow or Shift + DownArrow</kbd></td>
 *     <td>Move focus and multi-select the nearest data item above or below in the collection (e.g. areas, markers, links).</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Ctrl + LeftArrow or Ctrl + RightArrow</kbd></td>
 *     <td>Move focus to the left or right nearest data item in the collection (e.g. areas, markers, links), 
 *         without changing the current selection.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Ctrl + UpArrow or Ctrl + DownArrow</kbd></td>
 *     <td>Move focus to nearest data item above or below in the collection (e.g. areas, markers, links), without changing the current selection.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>] or [</kbd></td>
 *     <td>Move focus and selection to nearest data item in the next data layer above or below.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Shift + ] or Shift + [</kbd></td>
 *     <td>Move focus to nearest data item in the next data layer above or below and multi-select.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Ctrl + ] or Ctrl + [</kbd></td>
 *     <td>Move focus to nearest data item in the next data layer above or below, without changing the current selection.</td>
 *  </tr>
 *  <tr>
 *    <td><kbd>Alt + &lt; or Alt + &gt;</kbd></td>
 *    <td>Move focus from a data item to an associated link. Note that the link must have been created referencing the data item's ID 
 *        in its start/endLocation objects for an association to exist.</td>
 *  </tr>
 *  <tr>
 *     <td><kbd>Ctrl + Spacebar</kbd></td>
 *     <td>Multi-select base map region or marker with focus.</td>
 *  </tr>
 *  </tbody>
 *</table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojThematicMap
 */


// PROPERTY TYPEDEFS

/**
 * @typedef {Object} oj.ojThematicMap.Area
 * @property {Array.<string>=}       categories An array of category strings corresponding to this area. This allows highlighting and filtering of areas.
 * @property {string=}       color The area color.
 * @property {string=}       id The identifier for this area.
 * @property {string=}       label Text used for the area's label.
 * @property {Object=}       labelStyle The CSS style defining the label style for this area.
 * @property {string}        location An identifier corresponding to a Feature provided in the mapProvider geo object that this area is associated with.
 * @property {number=}       opacity The area opacity.
 * @property {"auto"|"off"}  [selectable="auto"] Specifies whether or not the area will be selectable.
 * @property {string=}       shortDesc The text that displays in the area's tooltip.
 * @property {string=}       svgClassName The CSS style class defining the style of the area.
 * @property {Object=}       svgStyle The CSS style object defining the style of the area.
 */
/**
 * @typedef {Object} oj.ojThematicMap.Link
 * @property {Array.<string>=}       categories An array of category strings corresponding to this link. This allows highlighting and filtering of links.
 * @property {string=}       color The link color.
 * @property {Object}       endLocation An object with used to determine the end point of the link.
 * @property {string=}       endLocation.id The marker or area id to be used as the end point.
 * @property {string=}       endLocation.location An identifier corresponding to a Feature provided in the mapProvider geo object to be used as the end point.
 * @property {number=}       endLocation.x The x coordinate which can represent latitude of the end point.
 * @property {number=}       endLocation.y The y coordinate which can represent longitude of the end point.
 * @property {string=}       id The identifier for this link.
 * @property {"auto"|"off"}  [selectable="auto"] Specifies whether or not the link will be selectable.
 * @property {string=}       shortDesc The text that displays in the links's tooltip.
 * @property {Object}       startLocation An object with used to determine the start point of the link.
 * @property {string=}       startLocation.id The marker or area id to be used as the start point.
 * @property {string=}       startLocation.location An identifier corresponding to a Feature provided in the mapProvider geo object to be used as the start point.
 * @property {number=}       startLocation.x The x coordinate which can represent latitude of the start point.
 * @property {number=}       startLocation.y The y coordinate which can represent longitude of the start point.
 * @property {string=}       svgClassName The CSS style class defining the style of the link.
 * @property {Object=}       svgStyle The CSS style object defining the style of the link.
 * @property {number=}       width The link width in pixels.
 */
/**
 * @typedef {Object} oj.ojThematicMap.Marker
 * @property {string=}       borderColor The marker border color.
 * @property {"solid"|"none"}    [borderStyle="solid"] The marker border style.
 * @property {number=}       borderWidth The marker border width in pixels.
 * @property {Array.<string>=}       categories An array of category strings corresponding to this marker. This allows highlighting and filtering of markers.
 * @property {string=}       color The marker color.
 * @property {number=}       height The pixel height for this marker. Note that this attribute will be ignored if a value is provided to calculate marker sizes.
 * @property {string=}       id The identifier for this link.
 * @property {string=}       label Text used for the marker's label.
 * @property {"bottom"|"center"|"top"}       [labelPosition="center"] Determines the label position relative to the marker.
 * @property {Object=}       labelStyle The CSS style object defining the style of the marker.
 * @property {string=}       location An identifier corresponding to a Feature provided in the mapProvider geo object that this marker is associated with.
 * @property {number=}       opacity The marker opacity.
 * @property {number=}       rotation The angle to rotate the marker in clockwise degrees around the marker center.
 * @property {"auto"|"off"}  [selectable="auto"] Specifies whether or not the link will be selectable.
 * @property {"circle"|"diamond"|"ellipse"|"human"|"plus"|"rectangle"|"square"|"star"|"triangleDown"|"triangleUp"|string}       
 *                           [shape="circle"] Specifies the shape of a marker. Can take the name of a built-in shape or the svg path commands for a custom shape.
 * @property {string=}       shortDesc The text that displays in the area's tooltip.
 * @property {string=}       source Specifies an URI specifying the location of the image resource to use for the marker instead of a built-in shape. 
 *                                  The shape attribute is ignored if the source image is defined.
 * @property {string=}       sourceHover An optional URI specifying the location of the hover image resource. If not defined, the source image will be used.
 * @property {string=}       sourceHoverSelected An optional URI specifying the location of the selected image resource on hover. If not defined, the 
 *                                               sourceSelected image will be used. If sourceSelected is not defined, then the source image will be used.
 * @property {string=}       sourceSelected An optional URI specifying the location of the selected image. If not defined, the source image will be used.
 * @property {string=}       svgClassName The CSS style class defining the style of the marker.
 * @property {Object=}       svgStyle The CSS style object defining the style of the marker.
 * @property {number=}       value A data value used to calculate the marker dimensions based on the range of all the data values and the element size. 
 *                                 Markers with negative or zero data values will not be rendered. If specified, this value takes precedence over the 
 *                                 width and height attributes.
 * @property {number=}       width The pixel width for this marker. Note that this attribute will be ignored if a value is provided to calculate marker.
 * @property {number=}       x The x coordinate of the marker transformed using the map projection, which can be null if location is set instead.
 * @property {number=}       y The y coordinate of the marker transformed using the map projection, which can be null if location is set instead.
 */
/**
 * @typedef {Object} oj.ojThematicMap.RendererContext
 * @property {string}       color The color of the data item.
 * @property {Element}      componentElement The thematic map element.
 * @property {Object}       data The data object for a stamped data visualization.
 * @property {string}       id The id of the data item.
 * @property {string}       label The label of the data item.
 * @property {string|null}  location The location of the data item which can be null if x/y are set instead.
 * @property {Element}      parentElement An element that is part of a displayed subtree on the DOM. Modifications of the parentElement are not supported.
 * @property {Object}       previousState An object that reflects the previous state of the data item.
 * @property {boolean}      previousState.hovered True if the data item was previously hovered.
 * @property {boolean}      previousState.selected True if the data item was previously selected.
 * @property {boolean}      previousState.focused True if the data item was previously selected.
 * @property {function():void} renderDefaultFocus Function for rendering default focus effect for the data item
 * @property {function():void} renderDefaultHover Function for rendering default hover effect for the data item
 * @property {function():void} renderDefaultSelection Function for rendering default selection effect for the data item
 * @property {Element|null} root Null on initial rendering or the current data item SVG element.
 * @property {Object}       state An object that reflects the current state of the data item.
 * @property {boolean}      state.hovered True if the data item is currently hovered.
 * @property {boolean}      state.selected True if the data item is currently selected.
 * @property {boolean}      state.focused True if the data item is currently selected.
 * @property {number|null}  x The x coordinate of the data item which can be null if location is set instead.
 * @property {number|null}  y The y coordinate of the data item which can be null if location is set instead.
 */
/**
 * @typedef {Object} oj.ojThematicMap.TooltipContext
 * @property {string|null} color The color of the hovered item or null if the hovered item if not associated with any data.
 * @property {Element} componentElement The thematic map element.
 * @property {Object|null} data The data object of the hovered item or null if the hovered item is not associated with any data.
 * @property {string|null} id The id of the hovered item or null if the hovered item if not associated with any data.
 * @property {string|null} label The data label of the hovered item or null if the hovered item if not associated with any data.
 * @property {string|null} location The location id of the hovered item which can be null if x/y are set instead.
 * @property {string|null} locationName The location name of the hovered item if location id is set.
 * @property {Element} parentElement The tooltip element. The function can directly modify or append content to this element.
 * @property {string} tooltip The default tooltip string constructed by the element if any.
 * @property {number} x The x coordinate of the hovered item which can be null if location is set instead.
 * @property {number} y The y coordinate of the hovered item which can be null if location is set instead.
 */

// METHOD TYPEDEFS

/**
 * @typedef {Object} oj.ojThematicMap.DataContext
 * @property {string} color The color of the item at the given index.
 * @property {string} label The label of the item at the given index.
 * @property {boolean} selected True if the item at the given index is currently selected and false otherwise.
 * @property {string} tooltip The tooltip of the item at the given index.
 */
/**
 * @typedef {Object} oj.ojThematicMap.NodeContext
 * @property {string} subId The subId string identify the particular DOM node.
 * @property {number} index The zero based index of the thematic map item.
 */

// KEEP FOR WIDGET SYNTAX

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for an area in the specified data layer.</p>
 *
 * @property {string} dataLayer The id of the data layer.
 * @property {number} index The index of the area within the specified data layer.
 *
 * @ojsubid oj-thematicmap-area
 * @memberof oj.ojThematicMap
 *
 * @example <caption>Get the first area in the collection (e.g. areas, markers, links) with id 'states':</caption>
 * var nodes = myThematicMap.getNodeBySubId({'subId': 'oj-thematicmap-area', 'dataLayer': 'states', 'index' : 0});
 */

/**
 * <p>Sub-ID for a marker in the specified data layer.</p>
 *
 * @property {string} dataLayer The id of the data layer.
 * @property {number} index The index of the marker within the specified data layer.
 *
 * @ojsubid oj-thematicmap-marker
 * @memberof oj.ojThematicMap
 *
 * @example <caption>Get the first marker in the collection (e.g. areas, markers, links) with id 'states':</caption>
 * var nodes = myThematicMap.getNodeBySubId({'subId': 'oj-thematicmap-marker', 'dataLayer': 'states', 'index' : 0});
 */

/**
 * <p>Sub-ID for a link in the specified data layer.</p>
 *
 * @property {string} dataLayer The id of the data layer.
 * @property {number} index The index of the link within the specified data layer.
 *
 * @ojsubid oj-thematicmap-link
 * @memberof oj.ojThematicMap
 *
 * @example <caption>Get the first link in the collection (e.g. areas, markers, links) with id 'states':</caption>
 * var nodes = myThematicMap.getNodeBySubId({'subId': 'oj-thematicmap-link', 'dataLayer': 'states', 'index' : 0});
 */

/**
 * <p>Sub-ID for the the thematic map tooltip.</p>
 *
 * @ojsubid oj-thematicmap-tooltip
 * @memberof oj.ojThematicMap
 *
 * @example <caption>Get the tooltip object of the thematic map, if displayed:</caption>
 * var nodes = myThematicMap.getNodeBySubId({'subId': 'oj-thematicmap-tooltip'});
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for an area in the specified data layer.</p>
 *
 * @property {number} index The index of the area within the specified data layer.
 *
 * @ojnodecontext oj-thematicmap-area
 * @memberof oj.ojThematicMap
 */

/**
 * <p>Context for a marker in the specified data layer.</p>
 * 
 * @property {number} index The index of the marker within the specified data layer.
 *
 * @ojnodecontext oj-thematicmap-marker
 * @memberof oj.ojThematicMap
 */

/**
 * <p>Context for a link in the specified data layer.</p>
 * 
 * @property {number} index The index of the link within the specified data layer.
 *
 * @ojnodecontext oj-thematicmap-link
 * @memberof oj.ojThematicMap
 */


/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojThematicMapMeta = {
  "properties": {
    "animationDuration": {
      "type": "number"
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues":  ["auto", "none"]
    },
    "areas": {
      "type": "Array<object>|Promise"
    },
    "focusRenderer": {},
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string"
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues":  ["dim", "none"]
    },
    "hoverRenderer": {},
    "initialZooming": {
      "type": "string",
      "enumValues":  ["auto", "none"]
    },
    "isolatedItem": {
      "type": "string"
    },
    "labelDisplay": {
      "type": "string",
      "enumValues":  ["auto", "on", "off"]
    },
    "labelType": {
      "type": "string",
      "enumValues":  ["short", "long"]
    },
    "links": {
      "type": "Array<object>|Promise"
    },
    "mapProvider": {
      "type": "object",
      "properties": {
        "geo": {
          "type": "object"
        },
        "propertiesKeys": {
          "type": "object"
        }
      }
    },
    "markers": {
      "type": "Array<object>|Promise"
    },
    "markerZoomBehavior": {
      "type": "string",
      "enumValues":  ["fixed", "zoom"]
    },
    "maxZoom": {
      "type": "number"
    },
    "panning": {
      "type": "string",
      "enumValues":  ["none", "auto"]
    },
    "renderer": {},
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues":  ["none", "single", "multiple"]
    },
    "selectionRenderer": {},
    "styleDefaults": {
      "type": "object",
      "properties": {
        "areaSvgStyle": {
          "type": "object"
        },
        "dataAreaDefaults": {
          "type": "object",
          "properties": {
            "borderColor": {
              "type": "string"
            },
            "hoverColor": {
              "type": "string"
            },
            "selectedInnerColor": {
              "type": "string"
            },
            "selectedOuterColor": {
              "type": "string"
            }
          }
        },
        "dataMarkerDefaults": {
          "type": "object",
          "properties": {
            "borderColor": {
              "type": "string"
            },
            "borderStyle": {
              "type": "string",
              "enumValues":  ["solid", "none"]
            },
            "borderWidth": {
              "type": "number"
            },
            "color": {
              "type": "string"
            },
            "height": {
              "type": "number"
            },
            "labelStyle": {
              "type": "object"
            },
            "opacity": {
              "type": "number"
            },
            "shape": {
              "type": "string"
            },
            "width": {
              "type": "number"
            }
          }
        },
        "hoverBehaviorDelay": {
          "type": "number"
        },
        "labelStyle": {
          "type": "object"
        },
        "linkDefaults": {
          "type": "object",
          "properties": {
            "color": {
              "type": "string"
            },
            "width": {
              "type": "number"
            }
          }
        }
      }
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "tooltipDisplay": {
      "type": "string",
      "enumValues":  ["auto", "shortDesc", "labelAndShortDesc", "none"]
    },
    "touchResponse": {
      "type": "string",
      "enumValues":  ["auto", "touchStart"]
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Thematic Map"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    },
    "zooming": {
      "type": "string",
      "enumValues":  ["none", "auto"]
    }
  },
  "methods": {
    "getArea": {},
    "getContextByNode": {},
    "getLink": {},
    "getMarker": {}
  },
  "extension": {
    _WIDGET_NAME: "ojThematicMap"
  }
};
oj.CustomElementBridge.registerMetadata('oj-thematic-map', 'dvtBaseComponent', ojThematicMapMeta);
// Get the combined meta of superclass which contains a shape parse function generator
var dvtMeta = oj.CustomElementBridge.getMetadata('oj-thematic-map');
oj.CustomElementBridge.register('oj-thematic-map', {
  'metadata': dvtMeta,
  'parseFunction': dvtMeta['extension']._DVT_PARSE_FUNC({'style-defaults.data-marker-defaults.shape': true})
});
})();
/**
 * Copyright (c) 2016, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojTimeAxis
 * @augments oj.dvtBaseComponent
 * @since 2.1.0
 * @ojstatus preview
 * @ojshortdesc Displays a range of dates based on specified start and end dates and a time scale.
 * @ojrole application
 * @ojtsimport ojvalidation-base
 * @ojtsimport ojvalidation-datetime
 *
 * @classdesc
 * <h3 id="TimeAxisOverview-section">
 *   JET Time Axis
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#TimeAxisOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET Time Axis is a themable, WAI-ARIA compliant element that displays a range of dates based on specified start and end date and time scale.  The Time Axis is intended to be placed in the header of JET DataGrid or Table.</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-time-axis
 *   scale="months"
 *   start="2017-01-01T05:00:00.000Z"
 *   end="2017-12-31T05:00:00.000Z"
 *   style="height:38px"
 *   aria-hidden="true">
 * &lt;/oj-time-axis>
 * </code>
 * </pre>
 *
 * <h3 id="formats-section">
 *   Date and Time Formats
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#formats-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"formatsDoc"}
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * <p>The application is responsible for supplying a meaningful <code class="prettyprint"><span class="pln">aria</span><span class="pun">-</span><span class="pln">label</span></code> to the element.</p>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Time Axis Span</h4>
 *
 * <p>It's recommended that applications limit the number of time intervals that are
 *    rendered by the Time Axis. For example, a Time Axis spanning one year with a scale
 *    of hours will display (365 * 24) 8,760 intervals. Rendering or rerendering this many intervals
 *    can cause severe performance degradation. Depending on the amount of space allocated,
 *    this many intervals can also cause visual clutteredness.
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojTimeAxis', $['oj']['dvtBaseComponent'],
{
  widgetEventPrefix: "oj",
  options: {
    /**
     * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the labels of the time axis for all 'scale' values, or 
     * an object literal whose keys are 'scale' values that map specific converters for scale specific formatting. 
     * See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
     * @expose
     * @name converter
     * @memberof oj.ojTimeAxis
     * @instance
     * @type {Object}
     * @ojsignature {target: "Type", value: "oj.ojTimeAxis.Converter|oj.Converter<string>", jsdocOverride: true}
     * @default {
     *            "default": null,
     *            "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}),
     *            "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}),
     *            "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}),
     *            "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
     *            "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
     *            "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
     *            "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
     *            "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})
     *          }
     * 
     * @example <caption>Initialize the TimeAxis with the <code class="prettyprint">converter</code> attribute specified:</caption>
     * &lt;oj-time-axis converter='[[myConverterObject]]'>&lt;/oj-time-axis>
     *
     * &lt;oj-time-axis converter.days='[[myConverterDays]]'>&lt;/oj-time-axis>
     *
     * @example <caption>Get or set the <code class="prettyprint">converter</code> property after initialization:</caption>
     * // Get one
     * var converterDays = myTimeAxis.converter.days;
     *
     * // Set one, leaving the others intact.
     * myTimeAxis.setProperty('converter.days', converterDays);
     *
     * // Get all
     * var converterObject = myTimeAxis.converter;
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myTimeAxis.converter = {
     *     "default": converterDefault,
     *     "seconds": converterSeconds,
     *     "minutes": converterMinutes,
     *     "hours": converterHours,
     *     "days": converterDays,
     *     "weeks": converterWeeks,
     *     "months": converterMonths,
     *     "quarters": converterQuarters,
     *     "years": converterYears
     * };
     */
    converter: {
      "default": null,
      "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}),
      "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}),
      "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}),
      "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
      "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
      "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
      "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
      "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})
    },
    /**
     * The start time of the time axis. A valid value is required in order for the time axis to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name start
     * @memberof oj.ojTimeAxis
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the TimeAxis with the <code class="prettyprint">start</code> attribute specified:</caption>
     * &lt;oj-time-axis start='2017-01-01T05:00:00.000Z'>&lt;/oj-time-axis>
     *
     * @example <caption>Get or set the <code class="prettyprint">start</code> property after initialization:</caption>
     * // getter
     * var startValue = myTimeAxis.start;
     *
     * // setter
     * myTimeAxis.start = '2017-01-01T05:00:00.000Z';
     */
    start: "",
    /**
     * The end time of the time axis. A valid value is required in order for the time axis to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name end
     * @memberof oj.ojTimeAxis
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the TimeAxis with the <code class="prettyprint">end</code> attribute specified:</caption>
     * &lt;oj-time-axis end='2017-12-31T05:00:00.000Z'>&lt;/oj-time-axis>
     *
     * @example <caption>Get or set the <code class="prettyprint">end</code> property after initialization:</caption>
     * // getter
     * var endValue = myTimeAxis.end;
     *
     * // setter
     * myTimeAxis.end = '2017-12-31T05:00:00.000Z';
     */
    end: "",
    /**
     * The time scale used for the time axis. This is required in order for the time axis to properly render.
     * @expose
     * @name scale
     * @memberof oj.ojTimeAxis
     * @instance
     * @type {?string}
     * @ojvalue {string} "seconds"
     * @ojvalue {string} "minutes"
     * @ojvalue {string} "hours"
     * @ojvalue {string} "days"
     * @ojvalue {string} "weeks"
     * @ojvalue {string} "months"
     * @ojvalue {string} "quarters"
     * @ojvalue {string} "years"
     * @default null
     * 
     * @example <caption>Initialize the TimeAxis with the <code class="prettyprint">scale</code> attribute specified:</caption>
     * &lt;oj-time-axis scale='weeks'>&lt;/oj-time-axis>
     *
     * @example <caption>Get or set the <code class="prettyprint">scale</code> property after initialization:</caption>
     * // getter
     * var scaleValue = myTimeAxis.scale;
     *
     * // setter
     * myTimeAxis.scale = 'weeks';
     */
    scale: null
  },

  // @inheritdoc
  _CreateDvtComponent: function(context, callback, callbackObj)
  {
    return dvt.TimeAxis.newInstance(context, callback, callbackObj);
  },

  // @inheritdoc
  _GetComponentStyleClasses : function() {
    var styleClasses = this._super();
    styleClasses.push('oj-timeaxis');
    return styleClasses;
  },

  // @inheritdoc
  _GetChildStyleClasses: function()
  {
    var styleClasses = this._super();
    styleClasses['oj-timeaxis-label'] = {'path': 'labelStyle', 'property': 'TEXT'};

    return styleClasses;
  },

  // @inheritdoc
  _GetEventTypes : function() {
    return ['optionChange'];
  },

  // @inheritdoc
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtUtilBundle.TIMEAXIS'] = translations['componentName'];
    return ret;
  },
  
  // @inheritdoc
  _GetComponentRendererOptions: function() {
    //the function should be removed if the component will support 'tooltip.renderer' attr
    return [];
  },

  // @inheritdoc
  _ProcessOptions: function() {
    this._super();

    // Date related options support only number | string types
    // TODO: remove deprecated number type support in favor of ISO String in future release
    var self = this;
    var processRootDateOptions = function(key) {
      var optionType = typeof self.options[key];
      if (!(optionType === 'number' || optionType === 'string'))
        self.options[key] = null; // e.g. this will exclude Date object types
    };

    processRootDateOptions('start');
    processRootDateOptions('end');
  },

  // @inheritdoc
  _LoadResources: function()
  {
    // Ensure the resources object exists
    if (this.options['_resources'] == null)
      this.options['_resources'] = {};

    var resources = this.options['_resources'];

    // Create default converters
    var converterFactory = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME);
    var secondsConverter = converterFactory.createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'});
    var minutesConverter = converterFactory.createConverter({'hour': 'numeric', 'minute': '2-digit'});
    var hoursConverter = converterFactory.createConverter({'hour': 'numeric'});
    var daysConverter = converterFactory.createConverter({'month': 'numeric', 'day': '2-digit'});
    var monthsConverter = converterFactory.createConverter({'month': 'long'});
    var yearsConverter = converterFactory.createConverter({'year': 'numeric'});

    var monthsConverterVert = converterFactory.createConverter({'month': 'short'});
    var yearsConverterVert = converterFactory.createConverter({'year': '2-digit'});

    var converter = {
      'seconds': secondsConverter,
      'minutes': minutesConverter,
      'hours': hoursConverter,
      'days': daysConverter,
      'weeks': daysConverter,
      'months': monthsConverter,
      'quarters': monthsConverter,
      'years': yearsConverter
    };
    var converterVert = {
      'seconds': secondsConverter,
      'minutes': minutesConverter,
      'hours': hoursConverter,
      'days': daysConverter,
      'weeks': daysConverter,
      'months': monthsConverterVert,
      'quarters': monthsConverterVert,
      'years': yearsConverterVert
    };

    resources['converterFactory'] = converterFactory;
    resources['converter'] = converter;
    resources['converterVert'] = converterVert;

    // Class names to be set on appropriate svg elements
    resources['axisClass'] = 'oj-timeaxis-container';
    resources['axisLabelClass'] = 'oj-timeaxis-label';
    resources['axisSeparatorClass'] = 'oj-timeaxis-separator';

    // default disable all borders
    resources['borderTopVisible'] = false;
    resources['borderRightVisible'] = false;
    resources['borderBottomVisible'] = false;
    resources['borderLeftVisible'] = false;

    // first day of week; locale specific
    resources['firstDayOfWeek'] = oj.LocaleData.getFirstDayOfWeek();
  }
});
/**
 * <p>The Time Axis is intended to be used inside of a JET Table or DataGrid. All touch interactions are the same as those of the root elements. See the <a href="oj.ojTable.html#touch-section">Table</a> and <a href="oj.ojDataGrid.html#touch-section">DataGrid</a> touch doc for more details.</p>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojTimeAxis
 */

/**
 * <p>The Time Axis is intended to be used inside of a JET Table or DataGrid. All keyboard interactions are the same as those of the root elements. See the <a href="oj.ojTable.html#keyboard-section">Table</a> and <a href="oj.ojDataGrid.html#keyboard-section">DataGrid</a> keyboard doc for more details.</p>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojTimeAxis
 */

/**
 * <p>The Time Axis supports a simplified version of the ISO 8601 extended date/time format. The format is as follows: <font color="#4B8A08">YYYY-MM-DDTHH:mm:ss.sssZ</font></p>
 * <table  class="keyboard-table">
 * <thead>
 * <tr>
 * <th>Symbol</th>
 * <th>Description</th>
 * <th>Values</th>
 * <th>Examples</th>
 * </tr>
 * </thead>
 * <tbody>
 * <tr>
 * <td><font color="#4B8A08">-, :, .,T</font></td><td>Characters actually in the string. T specifies the start of a time.</td><td></td><td></td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">YYYY</font></td><td>Year</td><td></td><td rowspan="3">2013-03-22<br>2014-02</td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">MM</font></td><td>Month</td><td>01 to 12</td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">DD</font></td><td>Day of the month</td><td>01 to 31</td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">HH</font></td><td>Hours</td><td>00 to 24</td><td rowspan="3">2013-02-04T15:20Z<br>2013-02-10T15:20:45.300Z</td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">mm</font></td><td>Minutes</td><td>00 to 59</td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">ss</font></td><td>Seconds. The seconds and milliseconds are optional if a time is specified.</td><td>00 to 59</td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">sss</font></td><td>Milliseconds</td><td>00 to 999</td><td></td>
 * </tr>
 * <tr>
 * <td><font color="#4B8A08">Z</font></td><td>The value in this position can be one of the following. If the value is omitted, character 'Z' should be used to specify UTC time.<br><ul><li><b>Z</b> indicates UTC time.</li><li><b>+hh:mm</b> indicates that the input time is the specified offset after UTC time.</li><li><b>-hh:mm</b> indicates that the input time is the absolute value of the specified offset before UTC time.</li></ul></td><td></td><td>2013-02-04T15:20:00-07:00<br>2013-02-04T15:20:00+05:00<br>2013-02-04T15:20:00Z</td>
 * </tr>
 * </tbody>
 * </table>
 * <p>The ISO format support short notations where the string must only include the date and not time, as in the following formats: YYYY, YYYY-MM, YYYY-MM-DD.</p>
 * <p>The ISO format does not support time zone names. You can use the Z position to specify an offset from UTC time. If you do not include a value in the Z position, UTC time is used. The correct format for UTC should always include character 'Z' if the offset time value is omitted. The date-parsing algorithms are browser-implementation-dependent and, for example, the date string '2013-02-27T17:00:00' will be parsed differently in Chrome vs Firefox vs IE.</p>
 * <p>You can specify midnight by using 00:00, or by using 24:00 on the previous day. The following two strings specify the same time: 2010-05-25T00:00Z and 2010-05-24T24:00Z.</p>
 *
 * @ojfragment formatsDoc
 * @memberof oj.ojTimeAxis
 */

// PROPERTY TYPEDEFS

/**
 * @typedef {Object} oj.ojTimeAxis.Converter
 * @property {oj.Converter.<string>} [default=null] The default converter (an object literal or instance that duck types {@link oj.Converter}) to use for all 'scale' values that do not otherwise have a converter object provided. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [seconds=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'seconds' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [minutes=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'minutes' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [hours=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'hours' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [days=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'days' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [weeks=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'weeks' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [months=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'months' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [quarters=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'quarters' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 * @property {oj.Converter.<string>} [years=oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})] The converter (an object literal or instance that duck types {@link oj.Converter}) used for the 'years' scale. If not specified, the default converter will be used for this scale. See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
 */

/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojTimeAxisMeta = {
  "properties": {
    "converter": {
      "type": "object",
      "properties": {
        "days": {},
        "default": {},
        "hours": {},
        "minutes": {},
        "months": {},
        "quarters": {},
        "seconds": {},
        "weeks": {},
        "years": {}
      }
    },
    "end": {
      "type": "string"
    },
    "scale": {
      "type": "string",
      "enumValues": ["seconds", "minutes", "hours", "days", "weeks", "months", "quarters", "years"]
    },
    "start": {
      "type": "string"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Time Axis"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        }
      }
    }
  },
  "methods": {},
  "extension": {
    _WIDGET_NAME: "ojTimeAxis"
  }
};
oj.CustomElementBridge.registerMetadata('oj-time-axis', 'dvtBaseComponent', ojTimeAxisMeta);
oj.CustomElementBridge.register('oj-time-axis', {'metadata': oj.CustomElementBridge.getMetadata('oj-time-axis')});
})();
/**
 * @ojcomponent oj.ojTimeline
 * @ojdisplayname Timeline
 * @augments oj.dvtTimeComponent
 * @ojrole application
 * @since 1.1.0
 * @ojstatus preview
 * @ojshortdesc An interactive data visualization that displays a series of events in chronological order.
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="timelineOverview-section">
 * JET Timeline
 * <a class="bookmarkable-link" title="Bookmarkable Link" href="#timelineOverview-section"></a>
 * </h3>
 * <p>Description:</p>
 * <p>A JET Timeline is a themable, WAI-ARIA compliant element that displays a set of events in chronological order.</p>
 *
 * {@ojinclude "name":"warning"}
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-timeline
 *   start='{{oj.IntlConverterUtils.dateToLocalIso(new Date("Jan 1, 2016"))}}'
 *   end='{{oj.IntlConverterUtils.dateToLocalIso(new Date("Dec 31, 2016"))}}'
 *   major-axis='{"scale": "months"}'
 *   minor-axis='{"scale": "weeks"}'
 *   series='{{seriesData}}'>
 * &lt;/oj-timeline>
 * </code>
 * </pre>
 *
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"a11yDoc"}
 *
 * <h3 id="formats-section">
 *   Date and Time Formats
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#formats-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"formatsDoc"}
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets.</p>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it's recommended that applications only set usable data
 *    densities on the timeline. For example, applications should limit the number of
 *    overlapping items rendered beyond the height of the timeline. This can be
 *    achieved by increasing the size of the timeline, decreasing the axis scale, or
 *    providing external filters to reduce the amount of data rendered at any given
 *    time. While there are several optimizations within the timeline to deal with
 *    large data sets, it's always more efficient to reduce the data set size as early
 *    as possible. Future optimizations will focus on improving end user experience as
 *    well as developer productivity for common use cases.</p>
 *
 * <h4>Timeline Span</h4>
 *
 * <p>It's recommended that applications limit the number of time intervals that are
 *    rendered by the timeline. For example, a timeline spanning one year with a scale
 *    of hours will display (365 * 24) 8,760 intervals. Rendering this many intervals
 *    can cause severe performance degradation when interacting with the timeline
 *    (scrolling and zooming) regardless of the number of items present.
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojTimeline', $['oj']['dvtTimeComponent'],
{
  widgetEventPrefix: "oj",
  options:
  {
    /**
     * Defines the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">animation-on-data-change</code> attribute specified:</caption>
     * &lt;oj-timeline animation-on-data-change='auto'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">animationOnDataChange</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.animationOnDataChange;
     *
     * // setter
     * myTimeline.animationOnDataChange = 'auto';
     */
    animationOnDataChange: "none",
    /**
     * Defines the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">animation-on-display</code> attribute specified:</caption>
     * &lt;oj-timeline animation-on-display='auto'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">animationOnDisplay</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.animationOnDisplay;
     *
     * // setter
     * myTimeline.animationOnDisplay = 'auto';
     */
    animationOnDisplay: "none",
    /**
     * The end time of the timeline. A valid value is required in order for the timeline to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name end
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @default ""
     *
     * @example <caption>Get or set the <code class="prettyprint">end</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.end;
     *
     * // setter
     * myTimeline.end = '2017-12-31T05:00:00.000Z';
     */
    end: "",
    /**
     * An object with the following properties, used to define a timeline axis. This is required in order for the timeline to properly render.
     * @expose
     * @name minorAxis
     * @memberof oj.ojTimeline
     * @instance
     * @type {Object}
     * @default {"converter": {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}, "svgStyle": {}}
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">minor-axis</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-timeline minor-axis.converter="[[myConverterObject]]" minor-axis.scale="weeks" minor-axis.zoom-order='["quarters", "months", "weeks", "days"]'>&lt;/oj-timeline>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-timeline minor-axis='{"scale": "weeks", "zoomOrder": ["quarters", "months", "weeks", "days"]}'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">minorAxis</code> property after initialization:</caption>
     * // Get one
     * var value = myTimeline.minorAxis.scale;
     *
     * // Set one, leaving the others intact.
     * myTimeline.setProperty('minorAxis.scale', 'weeks');
     *
     * // Get all
     * var values = myTimeline.minorAxis;
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myTimeline.minorAxis = {
     *     "converter": myConverterObject,
     *     "scale": "weeks",
     *     "svgStyle": {"backgroundColor": "red"},
     *     "zoomOrder": ["quarters", "months", "weeks", "days"]
     * };
     */
    minorAxis: {
      /**
       * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the labels of the minor axis for all 'scale' values, or 
       * an object literal whose keys are 'scale' values that map specific converters for scale specific formatting (see {@link oj.ojTimeAxis.Converter}). 
       * See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
       * <br></br>See the <a href="#minorAxis">minor-axis</a> attribute for usage examples.
       * @expose
       * @name minorAxis.converter
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?(oj.ojTimeAxis.Converter|oj.Converter<string>)", jsdocOverride: true}
       * @default {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}
       */
      converter: {
        "default": null,
        "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}),
        "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}),
        "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}),
        "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})
      },
      /**
       * The time scale used for the minor axis. This is required in order for the timeline to properly render.
       * <br></br>See the <a href="#minorAxis">minor-axis</a> attribute for usage examples.
       * @expose
       * @name minorAxis.scale
       * @memberof! oj.ojTimeline
       * @instance
       * @type {?string}
       * @ojvalue {string} "seconds"
       * @ojvalue {string} "minutes"
       * @ojvalue {string} "hours"
       * @ojvalue {string} "days"
       * @ojvalue {string} "weeks"
       * @ojvalue {string} "months"
       * @ojvalue {string} "quarters"
       * @ojvalue {string} "years"
       * @default null
       */
      scale: null,
      /**
       * The CSS style defining any additional styling of the axis. If not specified, no additional styling will be applied.
       * <br></br>See the <a href="#minorAxis">minor-axis</a> attribute for usage examples.
       * @expose
       * @name minorAxis.svgStyle
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {},
      /**
       * An array of strings containing the names of scales used for zooming from longest to shortest. If not specified, the 'scale' specified on the axis will be used at all zoom levels.
       * <br></br>See the <a href="#minorAxis">minor-axis</a> attribute for usage examples.
       * @expose
       * @name minorAxis.zoomOrder
       * @memberof! oj.ojTimeline
       * @instance
       * @type {?Array.<string>}
       * @ojsignature {target: "Type", value: "?"}
       * @default null
       */
      zoomOrder: null
    },
    /**
     * An object with the following properties, used to define a timeline axis. If not specified, no axis labels will be shown above the minor axis or in the overview.
     * @expose
     * @name majorAxis
     * @memberof oj.ojTimeline
     * @instance
     * @type {?Object}
     * @default {"converter": {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}, "svgStyle": {}}
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">major-axis</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-timeline major-axis.converter="[[myConverterObject]]" major-axis.scale="months" major-axis.zoom-order='["quarters", "months", "weeks", "days"]'>&lt;/oj-timeline>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-timeline major-axis='{"scale": "months", "zoomOrder": ["quarters", "months", "weeks", "days"]}'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">majorAxis</code> property after initialization:</caption>
     * // Get one
     * var value = myTimeline.majorAxis.scale;
     *
     * // Set one, leaving the others intact.
     * myTimeline.setProperty('majorAxis.scale', 'months');
     *
     * // Get all
     * var values = myTimeline.majorAxis;
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myTimeline.majorAxis = {
     *     "converter": myConverterObject,
     *     "scale": "months"
     * };
     */
    majorAxis: {
      /**
       * A converter (an object literal or instance that duck types {@link oj.Converter}) used to format the labels of the major axis for all 'scale' values, or 
       * an object literal whose keys are 'scale' values that map specific converters for scale specific formatting (see {@link oj.ojTimeAxis.Converter}). 
       * See {@link oj.DateTimeConverterFactory} for details on creating built-in datetime converters.
       * <br></br>See the <a href="#majorAxis">major-axis</a> attribute for usage examples.
       * @expose
       * @name majorAxis.converter
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?(oj.ojTimeAxis.Converter|oj.Converter<string>)", jsdocOverride: true}
       * @default {"default": null, "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}), "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}), "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}), "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}), "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}), "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})}
       */
      converter: {
        "default": null,
        "seconds": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit', 'second': '2-digit'}),
        "minutes": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric', 'minute': '2-digit'}),
        "hours": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'hour': 'numeric'}),
        "days": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "weeks": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'numeric', 'day': '2-digit'}),
        "months": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "quarters": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'month': 'long'}),
        "years": oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_DATETIME).createConverter({'year': 'numeric'})
      },
      /**
       * The time scale used for the major axis. If not specified, no axis labels will be shown above the minor axis or in the overview.
       * <br></br>See the <a href="#majorAxis">major-axis</a> attribute for usage examples.
       * @expose
       * @name majorAxis.scale
       * @memberof! oj.ojTimeline
       * @instance
       * @type {?string}
       * @ojvalue {string} "seconds"
       * @ojvalue {string} "minutes"
       * @ojvalue {string} "hours"
       * @ojvalue {string} "days"
       * @ojvalue {string} "weeks"
       * @ojvalue {string} "months"
       * @ojvalue {string} "quarters"
       * @ojvalue {string} "years"
       * @default null
       */
      scale: null,
      /**
       * The CSS style defining any additional styling of the axis. If not specified, no additional styling will be applied.
       * <br></br>See the <a href="#majorAxis">major-axis</a> attribute for usage examples.
       * @expose
       * @name majorAxis.svgStyle
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {}
    },
    /**
     * The orientation of the element.
     * @expose
     * @name orientation
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @ojvalue {string} "vertical"
     * @ojvalue {string} "horizontal"
     * @default "horizontal"
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">orientation</code> attribute specified:</caption>
     * &lt;oj-timeline orientation='vertical'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">orientation</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.orientation;
     *
     * // setter
     * myTimeline.orientation = 'vertical';
     */
    orientation: "horizontal",
    /**
     * An object with the following properties, used to define a timeline overview. If not specified, no overview will be shown.
     * @expose
     * @name overview
     * @memberof oj.ojTimeline
     * @instance
     * @type {?Object}
     * @default {"rendered": "off", "svgStyle": {}}
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">overview</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-timeline overview.rendered="on" overview.svg-style='{"height":"50px"}'>&lt;/oj-timeline>
     *
     * &lt;!-- Using JSON notation -->
     * &lt;oj-timeline overview='{"rendered": "on", "svgStyle": {"height":"50px"}}'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">majorAxis</code> property after initialization:</caption>
     * // Get one
     * var value = myTimeline.overview.rendered;
     *
     * // Set one, leaving the others intact.
     * myTimeline.setProperty('overview.rendered', 'on');
     *
     * // Get all
     * var values = myTimeline.rendered;
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myTimeline.majorAxis = {
     *     "rendered": "on",
     *     "svgStyle": {"height":"50px"}
     * };
     */
    overview: {
      /**
       * Specifies whether the overview scrollbar is rendered.
       * <br></br>See the <a href="#overview">overview</a> attribute for usage examples.
       * @expose
       * @name overview.rendered
       * @memberof! oj.ojTimeline
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       * @ojvalue {string} "on"
       * @ojvalue {string} "off"
       * @default "off"
       */
      rendered: "off",
      /**
       * The CSS style defining any additional styling of the overview. If not specified, no additional styling will be applied.
       * <br></br>See the <a href="#overview">overview</a> attribute for usage examples.
       * @expose
       * @name overview.svgStyle
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      svgStyle: {}
    },
    /**
     * The array of reference objects associated with the timeline. For each reference object, a line is rendered at the specified value. Currently only the first reference object in the array is supported. Any additional objects supplied in the array will be ignored.
     * @expose
     * @name referenceObjects
     * @memberof oj.ojTimeline
     * @instance
     * @type {Array.<Object>}
     * @ojsignature {target: "Type", value: "Array<oj.ojTimeline.ReferenceObject>", jsdocOverride: true}
     * @default []
     * 
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">reference-objects</code> attribute specified:</caption>
     * &lt;oj-timeline reference-objects='[{"value": "2017-04-15T04:00:00.000Z"}]'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">referenceObjects</code> property after initialization:</caption>
     * // Get one
     * var value = myTimeline.referenceObjects[0];
     * 
     * // Get all
     * var values = myTimeline.referenceObjects;
     *
     * // Set all (There is no permissible "set one" syntax.)
     * myTimeline.referenceObjects = [{"value": "2017-04-15T00:00:00.000Z"}];
     */
    referenceObjects: [],
    /**
     * An array of strings containing the ids of the initially selected items.
     * @expose
     * @name selection
     * @memberof oj.ojTimeline
     * @instance
     * @type {Array.<string>}
     * @ojwriteback
     * @default []
     * 
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">selection</code> attribute specified:</caption>
     * &lt;oj-timeline selection='["itemID1", "itemID2", "itemID3"]'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">gridlines</code> property after initialization:</caption>
     * // Get one
     * var value = myTimeline.selection[0];
     * 
     * // Get all
     * var values = myTimeline.selection;
     *
     * // Set all (There is no permissible "set one" syntax.)
     * myTimeline.selection = ["itemID1", "itemID2", "itemID3"];
     */
    selection: [],
    /**
     * The type of selection behavior that is enabled on the timeline. If 'single' is specified, only a single item across all series can be selected at once. If 'multiple', any number of items across all series can be selected at once. Otherwise, selection is disabled.
     * @expose
     * @name selectionMode
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @ojvalue {string} "none"
     * @default "none"
     * 
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">selection-mode</code> attribute specified:</caption>
     * &lt;oj-timeline selection-mode='multiple'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">selectionMode</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.selectionMode;
     * 
     * // setter
     * myTimeline.selectionMode = 'multiple';
     */
    selectionMode: "none",
    /**
     * An array of objects with the following properties, used to define a timeline series. Also accepts a Promise that will resolve with an array for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name series
     * @memberof oj.ojTimeline
     * @instance
     * @type {?(Array.<Object>|Promise)}
     * @ojsignature {target: "Accessor", value: {GetterType: "Promise<Array<oj.ojTimeline.Series>>|null", SetterType: "Array<oj.ojTimeline.Series>|Promise<Array<oj.ojTimeline.Series>>|null"}, jsdocOverride: true}
     * @default null
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">series</code> attribute specified:</caption>
     * &lt;oj-timeline series='[[mySeries]]'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">series</code> property after initialization:</caption>
     * // Get all (The series getter always returns a Promise so there is no "get one" syntax)
     * var values = myTimeline.series;
     *
     * // Set all (There is no permissible "set one" syntax.)
     * myTimeline.series = [
     *     {
     *         "id": "s1",
     *         "emptyText": "No Tournaments Played.",
     *         "label": "Rafael Nadal: 75-7",
     *         "items": [
     *             {
     *                 "id": "e1",
     *                 "title":"ATP VTR Open",
     *                 "start": "2013-02-04",
     *                 "description":"Finalist: 3-1"
     *             },
     *             {
     *                 "id": "e2",
     *                 "title":"ATP Brasil Open",
     *                 "start": "2013-02-11",
     *                 "description":"Champion: 4-0"
     *             }
     *         ]
     *     },
     *     {
     *         "id": "s2",
     *         "emptyText": "No Tournaments Played.",
     *         "label": "Novak Djokovic: 74-9",
     *         "items": [
     *             {
     *                 "id": "e101",
     *                 "title":"AUSTRALIAN OPEN",
     *                 "start": "2013-01-14",
     *                 "description":"Champion: 7-0"
     *             },
     *             {
     *                 "id": "e102",
     *                 "title":"Davis Cup World Group Round 1n",
     *                 "start": "2013-02-01",
     *                 "description":"Results: 1-0"
     *             },
     *             {
     *                 "id": "e103",
     *                 "title":"ATP Dubai Duty Free Tennis Championships",
     *                 "start": "2013-02-25",
     *                 "description":"Champion: 5-0"
     *             }
     *         ]
     *     }
     * ];
     */
    series: null,
    /**
     * The start time of the timeline. A valid value is required in order for the timeline to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name start
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @default ""
     * 
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">start</code> attribute specified:</caption>
     * &lt;oj-timeline start='2017-01-01T05:00:00.000Z'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">start</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.start;
     *
     * // setter
     * myTimeline.start = "2017-01-01T05:00:00.000Z";
     */
    start: "",
    /**
     * An object with the following properties, used to define default styling for the timeline.
     * Component CSS classes should be used to set component wide styling. This API should be used
     * only for styling a specific instance of the component. Properties specified on this object may 
     * be overridden by specifications on the data item. Some property default values come from the CSS 
     * and varies based on theme.
     * @expose
     * @name styleDefaults
     * @memberof oj.ojTimeline
     * @instance
     * @type {Object}
     * @default {"item": {}, "minorAxis": {}, "majorAxis": {}, "overview": {"window": {}},  "referenceObject": {}, "series": {"colors": ["#237bb1", "#68c182", "#fad55c", "#ed6647", "#8561c8", "#6ddbdb", "#ffb54d", "#e371b2", "#47bdef", "#a2bf39", "#a75dba", "#f7f37b"]}}
     * 
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">style-defaults</code> attribute specified:</caption>
     * &lt;!-- Using dot notation -->
     * &lt;oj-timeline style-defaults.animation-duration='200'>&lt;/oj-timeline>
     * 
     * &lt;!-- Using JSON notation -->
     * &lt;oj-timeline style-defaults='{"animationDuration": 200, "item": {"backgroundColor": "red"}'>&lt;/oj-timeline>
     * 
     * @example <caption>Get or set the <code class="prettyprint">styleDefaults</code> 
     * property after initialization:</caption>
     * // Get one
     * var value = myTimeline.styleDefaults.animationDuration;
     * 
     * // Get all
     * var values = myTimeline.styleDefaults;
     *
     * // Set one, leaving the others intact. Always use the setProperty API for 
     * // subproperties rather than setting a subproperty directly.
     * myTimeline.setProperty('styleDefaults.borderColor', 'red');
     * 
     * // Set all. Must list every resource key, as those not listed are lost.
     * myTimeline.styleDefaults = {'borderColor': 'red'};
     */
    styleDefaults: {
      /**
       * The duration of the animations, in milliseconds. The default value comes from the CSS and varies based on theme. For data change animations with multiple stages, this attribute defines the duration of each stage. For example, if an animation contains two stages, the total duration will be two times this attribute's value.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.animationDuration
       * @memberof! oj.ojTimeline
       * @instance
       * @type {number}
       * @ojsignature {target: "Type", value: "?"}
       * @ojunits "milliseconds"
       */
      animationDuration: undefined,
      /**
       * The border color of the timeline. The default value comes from the CSS and varies based on theme.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.borderColor
       * @memberof! oj.ojTimeline
       * @instance
       * @type {string}
       * @ojsignature {target: "Type", value: "?"}
       */
      borderColor: undefined,
      /**
       * An object with the following properties, used to define the default styling for the timeline item.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.item
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      item: {
        /**
         * The background color of the timeline items. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.backgroundColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        backgroundColor: undefined,
        /**
         * The border color of the timeline items. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.borderColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        borderColor: undefined,
        /**
         * The CSS style defining the style of the timeline item description text. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.descriptionStyle
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         */
        descriptionStyle: undefined,
        /**
         * The background color of the highlighted timeline items. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.hoverBackgroundColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        hoverBackgroundColor: undefined,
        /**
         * The border color of the highlighted timeline items. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.hoverBorderColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        hoverBorderColor: undefined,
        /**
         * The background color of the selected timeline items. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.selectedBackgroundColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        selectedBackgroundColor: undefined,
        /**
         * The border color of the selected timeline items. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.selectedBorderColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        selectedBorderColor: undefined,
        /**
         * The CSS style defining the style of the timeline item title text. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.item.titleStyle
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         */
        titleStyle: undefined
      },
      /**
       * An object with the following properties, used to define the default styling for the time axis.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.minorAxis
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      minorAxis: {
        /**
         * The background color of the time axis. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.minorAxis.backgroundColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        backgroundColor: undefined,
        /**
         * The border color of the time axis. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.minorAxis.borderColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        borderColor: undefined,
        /**
         * The CSS style defining the style of the time axis label text. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.minorAxis.labelStyle
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         */
        labelStyle: undefined,
        /**
         * The color of the time axis separators. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.minorAxis.separatorColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        separatorColor: undefined
      },
      
      /**
       * An object with the following properties, used to define the default styling for the major time axis.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.majorAxis
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      majorAxis: {
        /**
         * The CSS style defining the style of the major time axis label text. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.majorAxis.labelStyle
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         */
        labelStyle: undefined,
        /**
         * The color of the major time axis separators. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.majorAxis.separatorColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        separatorColor: undefined
      },
      /**
       * An object with the following properties, used to define the default styling for the timeline overview.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.overview
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      overview: {
        /**
         * The background color of the timeline overview. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.overview.backgroundColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        backgroundColor: undefined,
        /**
         * The CSS style defining the style of the timeline overview label text. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.overview.labelStyle
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         */
        labelStyle: undefined,
        /**
         * An object with the following properties, used to define the default styling for the timeline overview window.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.overview.window
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         * @default {}
         */
        window: {
          /**
           * The background color of the timeline overview window. The default value comes from the CSS and varies based on theme.
           * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
           * @expose
           * @name styleDefaults.overview.window.backgroundColor
           * @memberof! oj.ojTimeline
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           */
          backgroundColor: undefined,
          /**
           * The border color of the timeline overview window. The default value comes from the CSS and varies based on theme.
           * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
           * @expose
           * @name styleDefaults.overview.window.borderColor
           * @memberof! oj.ojTimeline
           * @instance
           * @type {string}
           * @ojsignature {target: "Type", value: "?"}
           */
          borderColor: undefined
        }
      },
      /**
       * An object with the following properties, used to define the default styling for the reference objects.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.referenceObject
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {}
       */
      referenceObject: {
        /**
         * The color of the reference objects. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.referenceObject.color
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        color: undefined
      },
      /**
       * An object with the following properties, used to define the default styling for the timeline series.
       * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
       * @expose
       * @name styleDefaults.series
       * @memberof! oj.ojTimeline
       * @instance
       * @type {Object}
       * @ojsignature {target: "Type", value: "?"}
       * @default {"colors": ["#237bb1", "#68c182", "#fad55c", "#ed6647", "#8561c8", "#6ddbdb", "#ffb54d", "#e371b2", "#47bdef", "#a2bf39", "#a75dba", "#f7f37b"]}
       */
      series: {
        /**
         * The background color of the series. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.series.backgroundColor
         * @memberof! oj.ojTimeline
         * @instance
         * @type {string}
         * @ojsignature {target: "Type", value: "?"}
         */
        backgroundColor: undefined,
        /**
         * The array defining the default color ramp for the series items.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.series.colors
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Array.<string>}
         * @ojsignature {target: "Type", value: "?"}
         * @default ["#237bb1", "#68c182", "#fad55c", "#ed6647", "#8561c8", "#6ddbdb", "#ffb54d", "#e371b2", "#47bdef", "#a2bf39", "#a75dba", "#f7f37b"]
         */
        colors: ["#237bb1", "#68c182", "#fad55c", "#ed6647", "#8561c8", "#6ddbdb", "#ffb54d", "#e371b2", "#47bdef", "#a2bf39", "#a75dba", "#f7f37b"],
        /**
         * The CSS style defining the style of the series empty text. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.series.emptyTextStyle
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         */
        emptyTextStyle: undefined,
        /**
         * The CSS style defining the style of the series label text. The default value comes from the CSS and varies based on theme.
         * <br></br>See the <a href="#styleDefaults">style-defaults</a> attribute for usage examples.
         * @expose
         * @name styleDefaults.series.labelStyle
         * @memberof! oj.ojTimeline
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         */
        labelStyle: undefined
      }
    },
    /**
     * The end time of the timeline's viewport. If not specified or invalid, this will default to a value determined by the initial 'scale' of the minor axis and the width of the timeline. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name viewportEnd
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @default ""
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">viewport-end</code> attribute specified:</caption>
     * &lt;oj-timeline viewport-end='2017-12-31T05:00:00.000Z'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">viewportEnd</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.viewportEnd;
     *
     * // setter
     * myTimeline.viewportEnd = '2017-12-31T05:00:00.000Z';
     */
    viewportEnd: "",
    /**
     * The start time of the timeline's viewport. If not specified or invalid, this will default to a value determined by the initial 'scale' of the minor axis and the width of the timeline. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
     * @expose
     * @name viewportStart
     * @memberof oj.ojTimeline
     * @instance
     * @type {string}
     * @default ""
     *
     * @example <caption>Initialize the Timeline with the <code class="prettyprint">viewport-start</code> attribute specified:</caption>
     * &lt;oj-timeline viewport-start='2017-01-01T05:00:00.000Z'>&lt;/oj-timeline>
     *
     * @example <caption>Get or set the <code class="prettyprint">viewportStart</code> property after initialization:</caption>
     * // getter
     * var value = myTimeline.viewportStart;
     *
     * // setter
     * myTimeline.viewportStart = '2017-01-01T05:00:00.000Z';
     */
    viewportStart: "",
    /**
     * Triggered after the viewport is changed due to a zoom or scroll operation.
     *
     * @property {string} viewportStart the start of the new viewport on a timeline
     * @property {string} viewportEnd the end of the new viewport on a timeline
     * @property {string} minorAxisScale the time scale of the minor axis
     *
     * @expose
     * @event
     * @memberof oj.ojTimeline
     * @instance
     * @ojbubbles
     */
    viewportChange: null
  },

  //@inheritdoc
  _CreateDvtComponent: function(context, callback, callbackObj)
  {
    return dvt.Timeline.newInstance(context, callback, callbackObj);
  },

  /**
   * @override
   * @instance
   * @memberof oj.ojTimeline
   * @protected
   */
  _ConvertLocatorToSubId : function(locator) {
    var subId = locator['subId'];

    // Convert the supported locators
    if(subId == 'oj-timeline-item') {
      // timelineItem[seriesIndex][itemIndex]
      subId = 'timelineItem[' + locator['seriesIndex'] + '][' + locator['itemIndex'] + ']';
    }

    // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
    // support for the old subId syntax in 1.2.0.
    return subId;
  },

  /**
   * @override
   * @instance
   * @memberof oj.ojTimeline
   * @protected
   */
  _ConvertSubIdToLocator : function(subId) {
    var locator = {};

    if(subId.indexOf('timelineItem') == 0) {
      // timelineItem[seriesIndex][itemIndex]
      var indexPath = this._GetIndexPath(subId);

      locator['subId'] = 'oj-timeline-item';
      locator['seriesIndex'] = indexPath[0];
      locator['itemIndex'] = indexPath[1];
    }

    return locator;
  },

  //@inheritdoc
  _ProcessStyles: function()
  {
    this._super();
    if (!this.options['styleDefaults'])
      this.options['styleDefaults'] = {};

    if (!this.options['styleDefaults']['series'])
      this.options['styleDefaults']['series'] = {};

    if (!this.options['styleDefaults']['series']['colors'])
    {
      var handler = new oj.ColorAttributeGroupHandler();

      // override default colors with css attribute group colors
      this.options['styleDefaults']['series']['colors'] = handler.getValueRamp();
    }
  },

  //@inheritdoc
  _GetComponentStyleClasses: function()
  {
    var styleClasses = this._super();
    styleClasses.push('oj-timeline');
    return styleClasses;
  },

  //@inheritdoc
  _GetComponentRendererOptions: function() {
    //the function should be removed if the timeline will support 'tooltip.renderer' attr
    return [];
  },

  //@inheritdoc
  _GetChildStyleClasses: function()
  {
    var styleClasses = this._super();
    styleClasses['oj-dvtbase oj-timeline'] = {'path': 'styleDefaults/animationDuration', 'property': 'ANIM_DUR'};
    styleClasses['oj-timeline'] = {'path': 'styleDefaults/borderColor', 'property': 'border-color'};
    styleClasses['oj-timeline-item'] = [
      {'path': 'styleDefaults/item/borderColor', 'property': 'border-color'},
      {'path': 'styleDefaults/item/backgroundColor', 'property': 'background-color'}
    ];
    styleClasses['oj-timeline-item oj-hover'] = [
      {'path': 'styleDefaults/item/hoverBorderColor', 'property': 'border-color'},
      {'path': 'styleDefaults/item/hoverBackgroundColor', 'property': 'background-color'}
    ];
    styleClasses['oj-timeline-item oj-selected'] = [
      {'path': 'styleDefaults/item/selectedBorderColor', 'property': 'border-color'},
      {'path': 'styleDefaults/item/selectedBackgroundColor', 'property': 'background-color'}
    ];
    styleClasses['oj-timeline-item-description'] = {'path': 'styleDefaults/item/descriptionStyle', 'property': 'TEXT'};
    styleClasses['oj-timeline-item-title'] = {'path': 'styleDefaults/item/titleStyle', 'property': 'TEXT'};
    styleClasses['oj-timeline-major-axis-label'] = {'path': 'styleDefaults/majorAxis/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-timeline-major-axis-separator'] = {'path': 'styleDefaults/majorAxis/separatorColor', 'property': 'color'};
    styleClasses['oj-timeline-minor-axis'] = [
      {'path': 'styleDefaults/minorAxis/backgroundColor', 'property': 'background-color'},
      {'path': 'styleDefaults/minorAxis/borderColor', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-minor-axis-label'] = {'path': 'styleDefaults/minorAxis/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-timeline-minor-axis-separator'] = {'path': 'styleDefaults/minorAxis/separatorColor', 'property': 'color'};
    styleClasses['oj-timeline-overview'] = {'path': 'styleDefaults/overview/backgroundColor', 'property': 'background-color'};
    styleClasses['oj-timeline-overview-label'] = {'path': 'styleDefaults/overview/labelStyle', 'property': 'TEXT'};
    styleClasses['oj-timeline-overview-window'] = [
      {'path': 'styleDefaults/overview/window/backgroundColor', 'property': 'background-color'},
      {'path': 'styleDefaults/overview/window/borderColor', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-reference-object'] = {'path': 'styleDefaults/referenceObject/color', 'property': 'color'};
    styleClasses['oj-timeline-series'] = {'path': 'styleDefaults/series/backgroundColor', 'property': 'background-color'};
    styleClasses['oj-timeline-series-empty-text'] = {'path': 'styleDefaults/series/emptyTextStyle', 'property': 'TEXT'};
    styleClasses['oj-timeline-series-label'] = {'path': 'styleDefaults/series/labelStyle', 'property': 'TEXT'};

    // Zoom Control Icons
    styleClasses['oj-timeline-zoomin-icon'] = [
      {'path': '_resources/zoomIn_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-zoomin-icon oj-hover'] = [
      {'path': '_resources/zoomIn_h_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_h_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-zoomin-icon oj-active'] = [
      {'path': '_resources/zoomIn_a_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_a_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-zoomin-icon oj-disabled'] = [
      {'path': '_resources/zoomIn_d_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomIn_d_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-zoomout-icon'] = [
      {'path': '_resources/zoomOut_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-zoomout-icon oj-hover'] = [
      {'path': '_resources/zoomOut_h_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_h_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-zoomout-icon oj-active'] = [
      {'path': '_resources/zoomOut_a_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_a_bc', 'property': 'border-color'}
    ];
    styleClasses['oj-timeline-zoomout-icon oj-disabled'] = [
      {'path': '_resources/zoomOut_d_bgc', 'property': 'background-color'},
      {'path': '_resources/zoomOut_d_bc', 'property': 'border-color'}
    ];

    return styleClasses;
  },

  //@inheritdoc
  _GetTranslationMap: function() {
    // The translations are stored on the options object.
    var translations = this.options['translations'];

    // Safe to modify super's map because function guarentees a new map is returned
    var ret = this._super();
    ret['DvtUtilBundle.TIMELINE'] = translations['componentName'];
    ret['DvtUtilBundle.TIMELINE_SERIES'] = translations['labelSeries'];
    ret['DvtUtilBundle.ZOOM_IN'] = translations['tooltipZoomIn'];
    ret['DvtUtilBundle.ZOOM_OUT'] = translations['tooltipZoomOut'];

    return ret;
  },

  //@inheritdoc
  _LoadResources: function() {
    this._super();

    var resources = this.options['_resources'];
    var converter = resources['converter'];
    var converterFactory = resources['converterFactory'];

    // Create default converters for vertical timeline
    var monthsConverterVert = converterFactory.createConverter({'month': 'short'});
    var yearsConverterVert = converterFactory.createConverter({'year': '2-digit'});

    var converterVert = {
      'seconds': converter['seconds'],
      'minutes': converter['minutes'],
      'hours': converter['hours'],
      'days': converter['days'],
      'weeks': converter['weeks'],
      'months': monthsConverterVert,
      'quarters': monthsConverterVert,
      'years': yearsConverterVert
    };

    resources['converterVert'] = converterVert;

    // Zoom control icons
    resources['zoomIn'] = 'oj-timeline-zoomin-icon';
    resources['zoomIn_h'] = 'oj-timeline-zoomin-icon oj-hover';
    resources['zoomIn_a'] = 'oj-timeline-zoomin-icon oj-active';
    resources['zoomIn_d'] = 'oj-timeline-zoomin-icon oj-disabled';
    resources['zoomOut'] = 'oj-timeline-zoomout-icon';
    resources['zoomOut_h'] = 'oj-timeline-zoomout-icon oj-hover';
    resources['zoomOut_a'] = 'oj-timeline-zoomout-icon oj-active';
    resources['zoomOut_d'] = 'oj-timeline-zoomout-icon oj-disabled';

    // Overview icons
    resources['overviewHandleHor'] = 'oj-timeline-overview-window-handle-horizontal';
    resources['overviewHandleVert'] = 'oj-timeline-overview-window-handle-vertical';
  },

  //@inheritdoc
  _GetComponentDeferredDataPaths : function() {
    return {'root': ['series']};
  },

  /**
   * {@ojinclude "name":"nodeContextDoc"}
   * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
   * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
   * @ojsignature {target: "Type", value: "oj.ojTimeline.NodeContext|null", jsdocOverride: true, for: "returns"}
   *
   * @example {@ojinclude "name":"nodeContextExample"}
   *
   * @expose
   * @instance
   * @memberof oj.ojTimeline
   */
  getContextByNode: function(node)
  {
    // context objects are documented with @ojnodecontext
    return this.getSubIdByNode(node);
  }
});
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td>Timeline Item</td>
 *       <td>Tap</td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2">Timeline Panel</td>
 *       <td>Drag</td>
 *       <td>Paning: navigate forward and backward in time in horizontal/vertical orientation.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan>Pinch-Close/Spread-Open</td>
 *       <td>Zoom In/Out.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2">Zoom Control</td>
 *       <td>Tap on "+" element</td>
 *       <td>Zoom In.</td>
 *     </tr>
 *     <tr>
 *       <td>Tap on "-" element</td>
 *       <td>Zoom Out.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2">Overview</td>
 *       <td>Press on right/left side of window & Hold & Drag in right of left direction</td>
 *       <td>Zoom In/Out (resize overview window).</td>
 *     </tr>
 *     <tr>
 *       <td>Press & Hold on the body of window & Drag in right of left direction</td>
 *       <td>Pan (move overview window).</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojTimeline
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td rowspan="2">Moves focus between series in a Dual Timeline and does nothing in a Single Timeline.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus and selection to previous data item (on left).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus and selection to next data item (on right).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + UpArrow</kbd></td>
 *       <td>Move focus and multi-select previous data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + DownArrow</kbd></td>
 *       <td>Move focus and multi-select next data item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + LeftArrow</kbd></td>
 *       <td>Move focus and multi-select previous data item (on left).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + RightArrow</kbd></td>
 *       <td>Move focus and multi-select next data item (on right).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + UpArrow</kbd></td>
 *       <td>Move focus to previous data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + DownArrow</kbd></td>
 *       <td>Move focus to next data item, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + LeftArrow</kbd></td>
 *       <td>Move focus to previous data item (on left), without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + RightArrow</kbd></td>
 *       <td>Move focus to next data item (on right), without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Spacebar</kbd></td>
 *       <td>Multi-select data item with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>= or +</kbd></td>
 *       <td>Zoom in one level if zooming is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>- or _</kbd></td>
 *       <td>Zoom out one level if zooming is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>PageUp</kbd></td>
 *       <td>Pan up if scrolling is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>PageDown</kbd></td>
 *       <td>Pan down if scrolling is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + PageUp</kbd></td>
 *       <td>Pan left in left-to-right locales. Pan right in right-to-left locales.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + PageDown</kbd></td>
 *       <td>Pan right in left-to-right locales. Pan left in right-to-left locales.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojTimeline
 */

/**
 *<p>The date/time data in the Timeline plays a key role, not only in the representation of events in the order in which they occurred, but also in many other places, such as the time axis, event durations, time markers, size and position calculations for the overview locator window, etc.</p>
 *<p>The Timeline supports a simplified version of the ISO 8601 extended date/time format. The format is as follows: <font color="#4B8A08">YYYY-MM-DDTHH:mm:ss.sssZ</font></p>
 *<table  class="keyboard-table">
 *<thead>
 *<tr>
 *<th>Symbol</th>
 *<th>Description</th>
 *<th>Values</th>
 *<th>Examples</th>
 *</tr>
 </thead>
 <tbody>
 *<tr>
 *<td><font color="#4B8A08">-, :, .,T</font></td><td>Characters actually in the string. T specifies the start of a time.</td><td></td><td></td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">YYYY</font></td><td>Year</td><td></td><td rowspan="3">2013-03-22<br>2014-02</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">MM</font></td><td>Month</td><td>01 to 12</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">DD</font></td><td>Day of the month</td><td>01 to 31</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">HH</font></td><td>Hours</td><td>00 to 24</td><td rowspan="3">2013-02-04T15:20Z<br>2013-02-10T15:20:45.300Z</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">mm</font></td><td>Minutes</td><td>00 to 59</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">ss</font></td><td>Seconds. The seconds and milliseconds are optional if a time is specified.</td><td>00 to 59</td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">sss</font></td><td>Milliseconds</td><td>00 to 999</td><td></td>
 *</tr>
 *<tr>
 *<td><font color="#4B8A08">Z</font></td><td>The value in this position can be one of the following. If the value is omitted, character 'Z' should be used to specify UTC time.<br><ul><li><b>Z</b> indicates UTC time.</li><li><b>+hh:mm</b> indicates that the input time is the specified offset after UTC time.</li><li><b>-hh:mm</b> indicates that the input time is the absolute value of the specified offset before UTC time.</li></ul></td><td></td><td>2013-02-04T15:20:00-07:00<br>2013-02-04T15:20:00+05:00<br>2013-02-04T15:20:00Z</td>
 *</tr>
 *</tbody>
 *</table>
 *<p>The ISO format support short notations where the string must only include the date and not time, as in the following formats: YYYY, YYYY-MM, YYYY-MM-DD.</p>
 *<p>The ISO format does not support time zone names. You can use the Z position to specify an offset from UTC time. If you do not include a value in the Z position, UTC time is used. The correct format for UTC should always include character 'Z' if the offset time value is omitted. The date-parsing algorithms are browser-implementation-dependent and, for example, the date string '2013-02-27T17:00:00' will be parsed differently in Chrome vs Firefox vs IE.</p>
 *<p>You can specify midnight by using 00:00, or by using 24:00 on the previous day. The following two strings specify the same time: 2010-05-25T00:00Z and 2010-05-24T24:00Z.</p>
 *
 * @ojfragment formatsDoc
 * @memberof oj.ojTimeline
 */

/**
 *<p>The application is responsible for populating the shortDesc value in the component options object with meaningful descriptors when the component does not provide a default descriptor. Since component terminology for keyboard and touch shortcuts can conflict with those of the application, it is the application's responsibility to provide these shortcuts, possibly via a help popup.</p>
 *
 * @ojfragment a11yDoc
 * @memberof oj.ojTimeline
 */

// PROPERTY TYPEDEFS

/**
 * @typedef {Object} oj.ojTimeline.ReferenceObject
 * @property {string=} value The time value of this reference object. If not specified, no reference object will be shown. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
 */
/**
 * @typedef {Object} oj.ojTimeline.Series
 * @property {string=} emptyText The text of an empty timeline series.
 * @property {string} id The identifier for the timeline series.
 * @property {"bottomToTop"|"topToBottom"|"auto"} [itemLayout="auto"] The direction in which items are laid out when in a horizontal orientation. This attribute is ignored when in a vertical orientation.
 * @property {string=} label The label displayed on the timeline series. In not specified, no label will be shown.
 * @property {Object=} svgStyle The CSS style defining any additional styling of the series. If not specified, no additional styling will be applied.
 * @property {Array.<oj.ojTimeline.SeriesItem>} [items] An array of items. If not specified, no data will be shown in this series.
 */
/**
 * @typedef {Object} oj.ojTimeline.SeriesItem
 * @property {string=} description The description text displayed on the timeline item. If not specified, no description will be shown.
 * @property {string=} end The end time of this timeline item. If not specified, no duration bar will be shown. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
 * @property {string=} durationFillColor The color applied to the duration bar of the timeline item. If not specified, this will be determined by the color ramp of the series.
 * @property {string} id The identifier for the timeline item. This must be unique across all items in the timeline, and is required in order for the timeline to properly render.
 * @property {string} start The start time of this timeline item. This is required in order for the timeline item to properly render. See <a href="#formats-section">Date and Time Formats</a> for more details on the required string formats.
 * @property {Object=} svgStyle The CSS style defining any additional styling of the item. If not specified, no additional styling will be applied.
 * @property {string=} thumbnail An optional URI specifying the location of an image resource to be displayed on the item. The image will be rendered at 32px x 32px in size. If not specified, no thumbnail will be shown.
 * @property {string=} title The title text displayed on the timeline item. If not specified, no title will be shown.
 */

// METHOD TYPEDEFS

/**
 * @typedef {Object} oj.ojTimeline.NodeContext
 * @property {string} subId The subId string to identify the particular DOM node.
 * @property {number} seriesIndex The zero based index of the timeline series.
 * @property {number} itemIndex The zero based index of the timeline series item.
 */

// KEEP FOR WIDGET SYNTAX

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for timeline series items indexed by series and item indices.</p>
 *
 * @property {number} seriesIndex
 * @property {number} itemIndex
 *
 * @ojsubid oj-timeline-item
 * @memberof oj.ojTimeline
 *
 * @example <caption>Gets the second item from the first series:</caption>
 * var node = myComponent.getNodeBySubId({'subId': 'oj-timeline-item', 'seriesIndex': 0, 'itemIndex': 1});
 */

// Node Context Objects ********************************************************

/**
 * <p>Context for timeline series items indexed by series and item indices.</p>
 *
 * @property {number} seriesIndex
 * @property {number} itemIndex
 *
 * @ojnodecontext oj-timeline-item
 * @memberof oj.ojTimeline
 */
/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojTimelineMeta = {
  "properties": {
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "end": {
      "type": "string"
    },
    "majorAxis": {
      "type": "object",
      "properties": {
        "converter": {
          "type": "object",
          "properties": {
            "days": {},
            "default": {},
            "hours": {},
            "minutes": {},
            "months": {},
            "quarters": {},
            "seconds": {},
            "weeks": {},
            "years": {}
          }
        },
        "scale": {
          "type": "string",
          "enumValues": ["seconds", "minutes", "hours", "days", "weeks", "months", "quarters", "years"]
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "minorAxis": {
      "type": "object",
      "properties": {
        "converter": {
          "type": "object",
          "properties": {
            "days": {},
            "default": {},
            "hours": {},
            "minutes": {},
            "months": {},
            "quarters": {},
            "seconds": {},
            "weeks": {},
            "years": {}
          }
        },
        "scale": {
          "type": "string",
          "enumValues": ["seconds", "minutes", "hours", "days", "weeks", "months", "quarters", "years"]
        },
        "svgStyle": {
          "type": "object"
        },
        "zoomOrder": {
          "type": "Array<string>"
        }
      }
    },
    "orientation": {
      "type": "string"
    },
    "overview": {
      "type": "object",
      "properties": {
        "rendered": {
          "type": "string",
          "enumValues": ["on", "off"]
        },
        "svgStyle": {
          "type": "object"
        }
      }
    },
    "referenceObjects": {
      "type": "Array<object>"
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["single", "multiple", "none"]
    },
    "series": {
      "type": "Array<object>|Promise"
    },
    "start": {
      "type": "string"
    },
    "styleDefaults": {
      "type": "object",
      "properties": {
        "animationDuration": {
          "type": "number"
        },
        "borderColor": {
          "type": "string"
        },
        "item": {
          "type": "object",
          "properties": {
            "item": {
              "backgroundColor": {
                "type": "string"
              },
              "borderColor": {
                "type": "string"
              },
              "descriptionStyle": {
                "type": "object"
              },
              "hoverBackgroundColor": {
                "type": "string"
              },
              "hoverBorderColor": {
                "type": "string"
              },
              "selectedBackgroundColor": {
                "type": "string"
              },
              "selectedBorderColor": {
                "type": "string"
              },
              "titleStyle": {
                "type": "object"
              }
            }
          }
        },
        "majorAxis": {
          "type": "object",
          "properties": {
            "majorAxis": {
              "labelStyle": {
                "type": "object"
              },
              "separatorColor": {
                "type": "string"
              }
            }
          }
        },
        "minorAxis": {
          "type": "object",
          "properties": {
            "minorAxis": {
              "backgroundColor": {
                "type": "string"
              },
              "borderColor": {
                "type": "string"
              },
              "labelStyle": {
                "type": "object"
              },
              "separatorColor": {
                "type": "string"
              }
            }
          }
        },
        "overview": {
          "type": "object",
          "properties": {
            "overview": {
              "backgroundColor": {
                "type": "string"
              },
              "labelStyle": {
                "type": "object"
              },
              "window": {
                "type": "object",
                "properties": {
                  "window": {
                    "backgroundColor": {
                      "type": "string"
                    },
                    "borderColor": {
                      "type": "string"
                    }
                  }
                }
              }
            }
          }
        },
        "referenceObject": {
          "type": "object",
          "properties": {
            "referenceObject": {
              "color": {
                "type": "string"
              }
            }
          }
        },
        "series": {
          "type": "object",
          "properties": {
            "series": {
              "backgroundColor": {
                "type": "string"
              },
              "colors": {
                "type": "Array<string>"
              },
              "emptyTextStyle": {
                "type": "object"
              },
              "labelStyle": {
                "type": "object"
              }
            }
          }
        }
      }
    },
    "translations": {
      "type": "Object",
      "properties": {
        "accessibleItemDesc": {
          "type": "string",
          "value": "Description is {0}."
        },
        "accessibleItemEnd": {
          "type": "string",
          "value": "End time is {0}."
        },
        "accessibleItemStart": {
          "type": "string",
          "value": "Start time is {0}."
        },
        "accessibleItemTitle": {
          "type": "string",
          "value": "Title is {0}."
        },
        "componentName": {
          "type": "string",
          "value": "Timeline"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "labelSeries": {
          "type": "string",
          "value": "Series"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        },
        "tooltipZoomIn": {
          "type": "string",
          "value": "Zoom In"
        },
        "tooltipZoomOut": {
          "type": "string",
          "value": "Zoom Out"
        }
      }
    },
    "viewportEnd": {
      "type": "string"
    },
    "viewportStart": {
      "type": "string"
    }
  },
  "methods": {
    "getContextByNode": {}
  },
  "events": {
    "viewportChange": {}
  },
  "extension": {
    _WIDGET_NAME: "ojTimeline"
  }
};
oj.CustomElementBridge.registerMetadata('oj-timeline', 'dvtTimeComponent', ojTimelineMeta);
oj.CustomElementBridge.register('oj-timeline', {'metadata': oj.CustomElementBridge.getMetadata('oj-timeline')});
})();
/**
  * Copyright (c) 2015, Oracle and/or its affiliates.
  * All rights reserved.
  */

/**
 * @class oj.TimeUtils
 * @since 2.1.0
 * @public
 * @export 
 * @hideconstructor
 * @classdesc
 * This class provides functions related to time to position conversions and vice versa.
 *
 */
oj.TimeUtils = function() {};

/**
 * Calculates the position of a given time point relative to the width of a given time range, e.g.
 * useful for calculating the start position of a task bar based on its start time and the given time axis range and width.
 * @param {(Date|string|number)} time the time point of interest, where rangeStartTime <= time <= rangeEndTime
 * @param {(Date|string |number)} rangeStartTime the lower bound of the time range of interest
 * @param {(Date|string|number)} rangeEndTime the upper bound of the time range of interest
 * @param {number} rangeWidth the width of the time range of interest
 * @return {number} the position of the time relative to the width of the time range.
 * @export
 */
oj.TimeUtils.getPosition = function(time, rangeStartTime, rangeEndTime, rangeWidth)
{
	time = (new Date(time)).getTime();
	rangeStartTime = (new Date(rangeStartTime)).getTime();
	rangeEndTime = (new Date(rangeEndTime)).getTime();

	var numerator = (time - rangeStartTime) * rangeWidth;
	var denominator = (rangeEndTime - rangeStartTime);
	if (numerator == 0 || denominator == 0)
		return 0;
	return numerator / denominator;
};

/**
 * Calculates the distance between two time points relative to the width of a given time range, e.g.
 * useful for calculating the length of a task bar given the start and end time, and the time axis range and width.
 * Expects rangeStartTime <= startTime <= endTime <= rangeEndTime.
 * @param {(Date|string|number)} startTime the start time of interest
 * @param {(Date|string|number)} endTime the end time of interest
 * @param {(Date|string|number)} rangeStartTime the lower bound of the time range of interest
 * @param {(Date|string|number)} rangeEndTime the upper bound of the time range of interest
 * @param {number} rangeWidth the width of the time range of interest
 * @return {number} the distance between the startTime and endTime positions relative to the width of the time range.
 * @export
 */
oj.TimeUtils.getLength = function(startTime, endTime, rangeStartTime, rangeEndTime, rangeWidth)
{
	var startPos, endPos;
	startTime = (new Date(startTime)).getTime();
	endTime = (new Date(endTime)).getTime();
	rangeStartTime = (new Date(rangeStartTime)).getTime();
	rangeEndTime = (new Date(rangeEndTime)).getTime();
	
	startPos = oj.TimeUtils.getPosition(startTime, rangeStartTime, rangeEndTime, rangeWidth);
	endPos = oj.TimeUtils.getPosition(endTime, rangeStartTime, rangeEndTime, rangeWidth);
	return endPos - startPos;
};

/**
 * Calculates the date/time of a given position point relative to the width of a given time range.
 * @param {number} pos the position of interest
 * @param {(Date|string|number)} rangeStartTime the lower bound of the time range of interest
 * @param {(Date|string|number)} rangeEndTime the upper bound of the time range of interest
 * @param {number} rangeWidth the width of the time range of interest
 * @return {number} the date in milliseconds since midnight Jan 1 1970 corresponding to the provided position and time range width.
 * @export
 */
oj.TimeUtils.getDate = function(pos, rangeStartTime, rangeEndTime, rangeWidth)
{
	rangeStartTime = (new Date(rangeStartTime)).getTime();
	rangeEndTime = (new Date(rangeEndTime)).getTime();

	var number = pos * (rangeEndTime - rangeStartTime);
	if (number == 0 || rangeWidth == 0)
		return rangeStartTime;
	return (number / rangeWidth) + rangeStartTime;
};

/*global ojtd:true*/


/**
 * Internal utilities for dealing with timezone data
 * @ignore
 */
oj.TimezoneData = {};

/**
 * Merges timezone data bundle into the LocaleElements bundle
 * @param {Object} timezoneBundle bundle to merge into the LocaleElements bundle
 * @ignore
 */
oj.TimezoneData.__mergeIntoLocaleElements = function(timezoneBundle)
{
  var localeElements = oj.LocaleData.__getBundle();
  oj.CollectionUtils.copyInto(localeElements, timezoneBundle, undefined, true);
}


/**
 * @return {Array.<string>} names of the timezone bundles
 * @ignore
 */
oj.TimezoneData.__getBundleNames = function()
{
  return oj.TimezoneData._bundleNames;
}

/**
 * @param {string} name bundle name
 * @ignore
 */
oj.TimezoneData.__registerBundleName = function(name)
{
  oj.TimezoneData._bundleNames.push(name);
}

/**
 * @ignore
 */
oj.TimezoneData._bundleNames = [];


(function()
{
  oj.TimezoneData.__registerBundleName('/timezoneData');
  oj.TimezoneData.__mergeIntoLocaleElements(typeof ojtd === 'undefined' ? {}: ojtd);
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

(function() { // Toolbar wrapper function, to keep "private static members" private

/**
 * @ojcomponent oj.ojToolbar
 * @ojdisplayname Toolbar
 * @augments oj.baseComponent
 * @ojrole toolbar
 * @since 0.6
 * @ojshortdesc Displays a strip of control elements (icons, buttons, separators, etc.) with support for touch, mouse, and keyboard interactions.
 * @ojstatus preview
 * 
 * @classdesc
 * <h3 id="toolbarOverview-section">
 *   JET Toolbar
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#toolbarOverview-section"></a>
 * </h3>
 * 
 * <p>Description: Themeable, WAI-ARIA-compliant toolbar element.
 * 
 * <pre class="prettyprint"><code>&lt;oj-toolbar id="myToolbar">
 *     &lt;oj-button id="myButton">
 *          &lt;span>My Button&lt;/span>
 *     &lt;/oj-button>
 * &lt;/oj-toolbar>
 * </code></pre>
 * 
 * <p>The JET Toolbar can contain [JET Buttons]{@link oj.ojButton}, [JET Menu Buttons]{@link oj.ojMenuButton}, [JET Buttonsets]{@link oj.ojButtonset}, and non-focusable content 
 * such as separator icons.  Toolbar provides WAI-ARIA-compliant focus management.
 * 
 * <p>A toolbar that contains radios should contain all radios in the radio group.
 * 
 * 
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * 
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 * 
 * 
 * <h3 id="keyboard-appdev-section">
 *   Keyboard Application Developer Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-appdev-section"></a>
 * </h3>
 * 
 * <p>Any buttonsets placed in the toolbar should have <code class="prettyprint">focusManagement</code> set to <code class="prettyprint">"none"</code>, 
 * so as not to compete with the toolbar's focus management.  
 * 
 * <p>The application should not do anything to interfere with the Toolbar's focus management, such as setting the <code class="prettyprint">tabindex</code> 
 * of the buttons.  Also, enabled buttons should remain user-visible, without which arrow-key navigation to the button would cause the focus to 
 * seemingly disappear.
 * 
 * 
 * <h3 id="a11y-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#a11y-section"></a>
 * </h3>
 * 
 * <p>JET Toolbar takes care of focus management, as noted above.  
 * 
 * <p>As shown in the online demos, the application is responsible for applying <code class="prettyprint">aria-label</code> and/or 
 * <code class="prettyprint">aria-controls</code> attributes to the toolbar element, if applicable per the instructions that follow:
 * 
 * <p>If this toolbar is (or might be) placed in context with other toolbars, then the application should apply an 
 * <code class="prettyprint">aria-label</code> to the toolbar element to distinguish it, e.g. an "Edit" toolbar.  The 
 * <code class="prettyprint">aria-label</code> is optional when there is only one toolbar.
 * 
 * <p>If the toolbar is controlling something else on the page, e.g. bold / italic / underline buttons controlling a rich 
 * text editor, then the application should apply an <code class="prettyprint">aria-controls</code> attribute to the toolbar element, 
 * e.g. <code class="prettyprint">aria-controls="myTextEditor"</code>.
 * 
 * <p>See the <a href="#styling-section">Styling</a> section for information on how to create accessible toolbar separators.
 *
 * <p>Disabled content: JET supports an accessible luminosity contrast ratio, 
 * as specified in <a href="http://www.w3.org/TR/WCAG20/#visual-audio-contrast-contrast">WCAG 2.0 - Section 1.4.3 "Contrast"</a>, 
 * in the themes that are accessible.  (See the "Theming" chapter of the JET Developer Guide for more information on which 
 * themes are accessible.)  Note that Section 1.4.3 says that text or images of text that are part of an inactive user 
 * interface component have no contrast requirement.  Because disabled content may not meet the minimum contrast ratio 
 * required of enabled content, it cannot be used to convey meaningful information.</p>
 * 
 * 
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 * 
 * {@ojinclude "name":"stylingDoc"}
 * 
 * 
 * <h3 id="rtl-section">
 *   Reading direction
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
 * </h3>
 * 
 * <p>The only supported way to set the reading direction (LTR or RTL) is to set the <code class="prettyprint">"dir"</code> attribute on the 
 * <code class="prettyprint">&lt;html></code> element of the page.  As with any JET component, in the unusual case that the reading direction
 * is changed post-create, the toolbar must be <code class="prettyprint">refresh()</code>ed, or the page must be reloaded. 
 * 
 * 
 * <h3 id="binding-section">
 *   Declarative Binding
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#binding-section"></a>
 * </h3>
 * 
 * <p>For elements like Toolbar and Menu that contain a number of like items, applications may wish to use a <code class="prettyprint">foreach</code> Knockout binding 
 * to stamp out the contents.  This binding cannot live on the same node as the JET <code class="prettyprint">ojComponent</code> binding, and must instead live on a nested 
 * virtual element as follows:
 * 
 * <pre class="prettyprint">
 * <code>&lt;oj-toolbar id="myToolbar">
 *     &lt;!-- ko foreach: myButtons -->
 *         &lt;oj-button data-bind="attr: {id: id, label: label}">
 *         &lt;/oj-buton>
 *     &lt;!-- /ko -->
 * &lt;/oj-toolbar>
 * </code></pre>
 * 
 * 
 * <!-- - - - - Above this point, the tags are for the class.
 *              Below this point, the tags are for the constructor (initializer). - - - - - - -->
 * 
 */
oj.__registerWidget("oj.ojToolbar", $['oj']['baseComponent'], {
    widgetEventPrefix : "oj", 

    options: { // options is in externs.js.  TODO: same as other prototype fields.
        /** 
         * <p>Indicates in what states the toolbar's buttons and buttonsets have chrome (background and border).
         * 
         * <p>This option only affects buttons and buttonsets that have never had their own <code class="prettyprint">chroming</code> option set.  This allows 
         * individual buttons and buttonsets to opt out of their toolbar's chroming if needed.  
         * 
         * <p>The default chroming varies by theme.  Each theme can set its default by setting 
         * <code class="prettyprint">$toolbarChromingOptionDefault</code> as seen in the example below.
         *
         * <p>Once a value has been set on this option, that value applies regardless of theme.
         * 
         * @expose
         * @memberof oj.ojToolbar
         * @instance
         * @since 1.2.0
         * 
         * @type {string}
         * @ojvalue {string} "full" In typical themes, full-chrome buttons always have chrome.
         * @ojvalue {string} "half" In typical themes, half-chrome buttons acquire chrome only in their hover, active, and selected states. Half-chroming is recommended for buttons in a toolbar.  
         *     (This is the toolbar default in most themes.)
         * @ojvalue {string} "outlined" In typical themes, outlined buttons are similar to half-chrome buttons, but have a border in the default state.
         *
         ** @example <caption>Initialize the Toolbar with the <code class="prettyprint">chroming</code> attribute specified:</caption>
         * &lt;oj-toolbar chroming='half'>&lt;/oj-toolbar>
         *
         * @example <caption>Get or set the <code class="prettyprint">chroming</code> property after initialization:</caption>
         * // getter
         * var chromingValue = myToolbar.chroming;
         *
         * // setter
         * myToolbar.chroming = 'half';
         *
         * @example <caption>Set the default in the theme (SCSS) :</caption>
         * $toolbarChromingOptionDefault: half !default;
         */
        chroming: "half"

        /**
         * <p>JET Toolbar does not support a 
         * <code class="prettyprint">disabled</code> option.  The following 
         * one-liner can be used to disable or enable all buttons in a toolbar:
         * 
         * @member
         * @name disabled
         * @memberof oj.ojToolbar
         * @instance
         * @ignore
         */
        // disabled option declared in superclass, but we still want the above API doc
        
        // Events
        
        /**
         * Triggered when the toolbar is created.
         *
         * @event 
         * @name create
         * @memberof oj.ojToolbar
         * @instance
         * @ignore
         * @property {Event} event <code class="prettyprint">jQuery</code> event object
         * @property {Object} ui Empty object included for consistency with other events
         */
        // create event declared in superclass, but we still want the above API doc
    },

    _InitOptions: function(originalDefaults, constructorOptions) {
        this._super(originalDefaults, constructorOptions);
        
        if ('disabled' in constructorOptions)
            oj.Logger.warn("Caller attempted to set the 'disabled' option on Toolbar, but Toolbar does not support the 'disabled' option.  See API doc.");
    },

    _ComponentCreate: function() {
        this._super();
        
        this.element
            .attr(oj.Components._OJ_CONTAINER_ATTR, this['widgetName'])
            .addClass( "oj-toolbar oj-component" )
            .attr( "role", "toolbar" );

        this._setup();
    },

    // Override to set custom launcher
    _NotifyContextMenuGesture: function(menu, event, eventType) {
        // Set launcher to the current tabbable button
        // For toggle buttons, launcher must be the hidden focusable input, but for Shift-F10 we want the CM aligned to the Button's root element, not that 
        // hidden input.  This is no change from the default for push buttons, since in that case the root element and launcher (input) are the same.
        var currentButton = this.element.find(":oj-button[tabindex=0]");
        this._OpenContextMenu(event, eventType, {
            "launcher": currentButton, 
            "position": {"of": eventType==="keyboard" ? currentButton.ojButton("widget") : event}
        });
    },

    _setOption: function( key, value ) { // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
        if ( key === "disabled" ) {
            oj.Logger.warn("Caller attempted to set the 'disabled' option on Toolbar, but Toolbar does not support the 'disabled' option.  See API doc.  Ignoring the call.");
            return;
        }

        // Call super() after the "disabled" check returns, to avoid super() setting oj-disabled, etc.
        // Call it before handling chroming case, so that when that case calls refresh() on Button/Buttonset, callees see the new value of the option.
        // (Previously called super() at end, so that optionChange (fired at end of super) is fired at very end.)
        this._superApply( arguments );

        if ( key === "chroming" ) {
            // refresh the top-level buttons, and refresh the buttonsets to make them refresh their buttons, so that all toolbar buttons are refreshed.
            // Reason: to make them re-fetch their chroming option, in case it's still set to the default dynamic getter, 
            // which takes its value from the containing buttonset or toolbar if present.
            // TBD: Consider only calling refresh() on children that haven't had their chroming option set, i.e. those still using the dynamic getter.
            this._refreshChildren();
        }
    },

    /**
     * Refreshes the toolbar, including the following:
     * 
     * <ul>
     *   <li>Re-applies focus management / keyboard navigation.
     *   <li>Rechecks the reading direction (LTR vs. RTL).
     * </ul>
     * 
     * 
     * <p>A <code class="prettyprint">refresh()</code> is required in the following circumstances:
     * <ul>
     *   <li>After buttons or buttonsets are added to, removed from, or reordered within the toolbar.</li>
     *   <li>After a change to the [disabled]{@link oj.ojButton#disabled} status of any of the buttons in the toolbar.</li>
     *   <li>After a programmatic change to the <code class="prettyprint">checked</code> status of a radio button in the toolbar
     *       (which should be done via Buttonset's [checked]{@link oj.ojButtonset#checked} option).  This applies only to radios, 
     *       not to checkboxes or push buttons.</li>
     *   <li>After the reading direction (LTR vs. RTL) changes.</li>
     * </ul>
     * 
     * @expose 
     * @memberof oj.ojToolbar
     * @return {void}
     * @instance
     * 
     * @example <caption>Invoke the <code class="prettyprint">refresh</code> method:</caption>
     * myToolbar.refresh();
     */
    refresh: function() {
        this._super();
        this._setup();
    },
    
    _setup: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var self = this;
        this.isRtl = this._GetReadingDirection() === "rtl";
        
        // initialize as an empty JQuery object so we don't fall over if focus hasn't been introduced before a click is made
        this.$enabledButtons = $();
        
        // When toolbar is binding listeners to buttons, use the Toolbar's eventNamespace, not the Button's 
        // eventNamespace, to facilitate later unbinding only the Toolbar listeners.

        // For checkbox/radio, we're binding to inputs, not labels.

        // Put listeners on every button, b/c it's too unreliable to put them on the toolbar node and rely on event bubbling.
        // - E.g. bubbling doesn't work for antonym buttons (is this still true after the refactoring?) -- see comment on Button._setLabelOption().
        // - Likewise, focus mgmt can't just break if app listener stops propagation.
        // - Both of these problems still happen when using the delegation / selector overload of .on(); there is no special JQ bubbling magic.

        if (this._IsCustomElement())
        {
            // defer setting up button-specific event handling until the first focusin event is triggered
            this._focusinListener = function(event) {
                self._handleInitialFocus();
            }
            this.element[0].addEventListener("focusin", this._focusinListener, true);
            
            // find any current supported children to refresh in case they were already initialized and need to update their 'chroming' values
            this.$topLevelChildren = this.element.find( "oj-button" ).add(this.element.find( "oj-menu-button" ))
                    .add(this.element.find( "oj-buttonset-one" )).add(this.element.find( "oj-buttonset-many" ));
            
            // refresh the top-level buttons, and refresh the buttonsets to make them refresh their buttons, so that all toolbar buttons are refreshed.
            // Reason: to make them re-fetch their chroming option, in case it's still set to the default dynamic getter, 
            // which takes its value from the containing buttonset or toolbar if present.
            // TBD: Consider only calling refresh() on children that haven't had their chroming option set, i.e. those still using the dynamic getter.
            this._refreshChildren();
        }
        else
        {
            this.$buttons = this.element.find( ":oj-button" )
                .unbind( "keydown" + this.eventNamespace )
                .bind( "keydown" + this.eventNamespace, function(event) { 
                    self._handleKeyDown(event, $(this)); 
                })
                
                .unbind( "click" + this.eventNamespace )
                .bind( "click" + this.eventNamespace, function(event) {
                    if ( !$(this).ojButton("option", "disabled") ) { 
                        // Normally the button will be tabbable after the click, since (a) if we reach here, the clicked button is enabled, and 
                        // (b) an unchecked radio before the click will normally be checked after the click.  But just in case it's unchecked 
                        // (e.g. due to app listener), we let callee run it thru _mapToTabbable() before using, as usual.
                        self._setTabStop( $(this) );
                    }
                })
                .unbind( "focus" + this.eventNamespace )
                .bind( "focus" + this.eventNamespace, function(event) { 
                    self._handleFocus( $(this) );
                });
            
            // refresh the top-level buttons, and refresh the buttonsets to make them refresh their buttons, so that all toolbar buttons are refreshed.
            // Reason: to make them re-fetch their chroming option, in case it's still set to the default dynamic getter, 
            // which takes its value from the containing buttonset or toolbar if present.
            // TBD: Consider only calling refresh() on children that haven't had their chroming option set, i.e. those still using the dynamic getter.
            this.$buttonsets = this.element.find( ":oj-buttonset" )
                .ojButtonset( "refresh" );
            this.$topLevelButtons = this.$buttons.not( this.$buttonsets.find( ":oj-button" ) )
                .ojButton( "refresh" );
        }
    },
    
    _handleFocus: function( $button ) { 
        if (!this._IsCustomElement() && this.$enabledButtons.length == 0) {
            // the subset of Toolbar buttons that are enabled.  Disabled buttons are not tabbable.
            this.$enabledButtons = this.$buttons.filter(function(index) {
                return !$( this ).ojButton( "option", "disabled" );
            });
            
            this._initTabindexes(this._lastTabStop == null);
            this.$enabledButtons[0].focus();
        }
        else
            this._setTabStop($button);
    },
    
    // For custom element only, we setup this handler for any focusin event on the toolbar. We then remove this handler, and setup the rest of the handlers we need
    // once all of our children have finished being initialized.
    _handleInitialFocus: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var self = this;
        // remove this event handling as we only want to run this setup logic on the first focusin event
        this.element[0].removeEventListener("focusin", this._focusinListener, true);

        var $topLevelButtons = this.element.find( "oj-button" ).add(this.element.find( "oj-menu-button" ));
        var $buttonsets = this.element.find( "oj-buttonset-one" ).add(this.element.find( "oj-buttonset-many" ));
        this.$topLevelChildren = $topLevelButtons.add($buttonsets);
        this.$buttons = $topLevelButtons.add($buttonsets.find( ".oj-button" ))
            .unbind( "keydown" + this.eventNamespace )
            .bind( "keydown" + this.eventNamespace, function(event) {
                var $button = $(this);
                self._handleKeyDown(event, $button); 
            })

            .unbind( "click" + this.eventNamespace )
            .bind( "click" + this.eventNamespace, function(event) {
                var $button = $(this);
                if (!$button.hasClass( "oj-disabled" ))
                {
                    // Normally the button will be tabbable after the click, since (a) if we reach here, the clicked button is enabled, and 
                    // (b) an unchecked radio before the click will normally be checked after the click.  But just in case it's unchecked 
                    // (e.g. due to app listener), we let callee run it thru _mapToTabbable() before using, as usual.
                    self._setTabStop($button);
                }
            })
            .unbind( "focusin" + this.eventNamespace )
            .bind( "focusin" + this.eventNamespace, function(event) { 
                var $button = $(this);
                self._handleFocus($button);
            });
            
            // the subset of Toolbar buttons that are enabled.  Disabled buttons are not tabbable.
            this.$enabledButtons = this.$buttons.filter(function(index) {
                return !($(this).hasClass( "oj-disabled" ));
            });
            
            this._initTabindexes(this._lastTabStop == null);
            this._getButtonFocusElem(this.$enabledButtons[0]).focus();
    },
    
    // Returns the focusable inner element of the button.
    _getButtonFocusElem: function(button) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        if (this._IsCustomElement())
        {
            var $button = $(button);
            if ($button.hasClass( "oj-button-toggle" ))
            {
                // the underlying input element is one of the button's children
                return $button.children( "input" )[0];
            }
            else
            {
                // the underlying button element is the first child of an oj-button / oj-menu-button
                return $button.children( "button" )[0];
            }
        }
        else
            return button;
    },
    
    // For create, make only the first enabled button tabbable.  (We decided to have Shift-Tab go to first, not last, button.)
    // For refreshes, keep the existing tabstop if we can, otherwise proceed as with create.
    // Either way, if that button is a radio and some radio in its group is checked, make that one tabbable instead.
    // If there are no enabled buttons, makes them all untabbable.
    // No return value.
    _initTabindexes: function(isCreate) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        // even for refreshes where we'll wind up keeping the same tabstop, must make everything untabbable first, to ensure any new buttons become untabbable.
        var $last = $(this._lastTabStop);
        this._lastTabStop = undefined;
        
        if (this._IsCustomElement()) {
            for (var i = 0; i < this.$buttons.length; i++) {
                this._getButtonFocusElem(this.$buttons[i]).setAttribute( "tabindex", "-1" );
            }
        }
        else
            this.$buttons.attr( "tabindex", "-1" );
        
        var $newTabStop; // callee might map this to radio groupmate
        
        // TBD: for refreshes when $last is a disabled radio with a checked enabled groupmate and they are in the toolbar, the groupmate would be 
        // a valid tabstop, but this defaults to the first.
        if (isCreate || !$last.is(this.$enabledButtons)) {
            // is create, or is refresh but must treat like create b/c $last is empty, or not enabled, or no longer in the toolbar
            $newTabStop = this.$enabledButtons.first(); // if empty (none enabled), no tabstop will be set
        } else {
            // is a refresh, and $last is non-empty and is an enabled button still in the toolbar.  May be a radio whose groupmate
            // has become checked, in which case callee will map it to that groupmate.
            $newTabStop = $last;
        }
        this._setTabStop( $newTabStop );
    },
    
    // For each button in $button (in our usage always 0-1 button hence $button singular), if that button is an unchecked radio 
    // with a checked groupmate (which means it's not tabbable), then map it to the checked one (checked enabled radios are 
    // tabbable, and we know it's enabled per the argument below).
    // 
    // $button contains 0 or more buttons to map.  Must be enabled since disabled buttons aren't tabbable.
    // Returns the mapped JQ object (which the caller will make the tabstop).
    // 
    // We know that this.$enabledButtons contains all buttons in $button, and all of their potentially checked radio-groupmates, since:
    // - The above "enabled" requirement guarantees that $button's contents are all in $enabledButtons.
    // - The prohibition against radio groupmates that are not in the toolbar, and the prohibition against checked disabled groupmates 
    //   of enabled radios, guarantee that if $button is a radio, then all of its potentially checked groupmates are enabled and thus in 
    //   $enabledButtons.
    // 
    // Firefox browser issue:   (TODO: should we doc this?  File FF and/or JET bug?)
    // 
    // When this method maps an unchecked radio to its checked groupmate, the caller ensures that the former still has focus, but the latter
    // is the tabstop for when the user tabs out and back in.  When tabbing / Shift-Tabbing from the unchecked radio in the direction of the 
    // checked one, Chrome and IE9 are smart enough to tab out of the radio group to the adjacent tabstop as desired.  But in FF, focus goes to 
    // the checked one, which is not what we want.
    // 
    // Reason: Chrome and IE9 are smart enough never to tab within a radio group.  If focus is in the radio group, Tab and Shift-Tab exit the radio group.
    // But in FF, [a radio is reachable via Tab/Shift-Tab from within the group] if [it's enabled, tabindex != -1, and either checked or has no 
    // checked groupmates], i.e. [it would be reachable via Tab/Shift-Tab from outside the group if all its groupmates happened to be untabbable 
    // (e.g. disabled)].  In other words, FF is the only one that fails to distinguish between radios that could be valid tabstops from outside, and
    // those that should be valid tabstops from inside.  
    // 
    // This impl improves on the native behavior.  In FF, in an unchecked radio group, every single radio is a tabstop.  Our use of tabindex=-1 
    // guarantees that we never tab within the group in that case.  It's only a checked groupmate that can be tabbed to from within. (So at 
    // most one unwanted tabstop.)
    // 
    // After much time and effort, the latter issue seems to be infeasible to fix in any robust, non-brittle way.  E.g.:
    // - Clearing all tabstops and restoring on tab-out of toolbar:  FF provides no reliable way to find out whether a blur is exiting the entire 
    //   toolbar.  The obvious setTimeout workaround needed to be unacceptably long (e.g. 250ms) and even then was unreliable.  If we ever fail 
    //   to restore the tabstop, the toolbar becomes untabbable and inaccessible.
    // - Every other approach had similar robustness issues. 
    _mapToTabbable: function( $button ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        var focusElems = [];
        for (var i = 0; i < this.$enabledButtons.length; i++) {
            focusElems.push(this._getButtonFocusElem(this.$enabledButtons[i]));
        }
        var $enabledButtonFocusElems = $(focusElems);
        return $button.map(function(index, elem) {
            // Buttons other than radios, and checked radios, are always tabbable if they're enabled, which this method requires.
            // Radios w/ name="" (incl name omitted) are not in a radio group, not even with other radios with w/ name="".  Radios 
            // with no groupmates are always tabbable, since either they're checked, or they're unchecked with no checked groupmate.
            if (elem.type != "radio" || elem.checked || elem.name == "") { 
                return elem;
            } else {
                // elem is unchecked radio in real (not "") group, which is tabbable iff no groupmate is checked.  Per above doc, we know that 
                // all of its potentially checked groupmates are in $enabledButtons.
                var $checkedRadio = _radioGroup(elem, $enabledButtonFocusElems).filter(":checked");
                return ($checkedRadio.length ? $checkedRadio[0] : elem);
            }
        });
    },
    
    // Set which button is in the tab sequence.
    // $button should contain 0 or 1 button to be made tabbable (since at most one should be tabbable at a time).
    //   If 0 (i.e. no enabled buttons), all will become untabbable.  If 1, it must be tabbable in every way (e.g. enabled) except possibly 
    //   being an unchecked radio with a checked groupmate, which this method will map to its checked groupmate, which 
    //   we know is enabled thus tabbable since we require that checked radios with enabled groupmates not be disabled.  
    // No return value.
    _setTabStop: function( $button ) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
//        if (!window.setTabStopCounter) window.setTabStopCounter=1; // REMOVE, is only for console.log's
//        console.log("in _setTabStop: " + window.setTabStopCounter++ + ".  Orig (premap) button checked: " + $button[0].checked); // + " and is:");
//        console.log($button[0]);
        
        if (this._IsCustomElement())
            $button = this._mapToTabbable($(this._getButtonFocusElem($button[0])));
        else
            $button = this._mapToTabbable( $button );
        
        var button = $button[0]; // button is undefined iff $button is empty iff we need to clear all tabstops b/c there are no enabled buttons to make tabbable
        var last = this._lastTabStop; // last is undefined iff $(last) is empty iff there are no existing tabstops to clear (b/c _initTabindexes just ran 
                                      // or previously there were no enabled buttons to make tabbable)
        
//        console.log("mapped button and last button are:");  console.log(button);  console.log(last);  console.log(".");
        
        // Cases: both are undefined: have no tabstops; want to keep it that way (b/c none enabled), so do nothing
        //        both are node X: X is the tabstop; want to keep it that way, so do nothing
        //        last is node X; button is undefined: X is the tabstop; want to clear it w/o replacing it (b/c none enabled).  This logic does that.
        //        last is undefined; button is node X: no existing tabstop; want to make X the tabstop.  This logic does that.
        //        last is node X; button is node Y: X is the tabstop; want to clear it and make Y the tabstop.  This logic does that.
        if ( button !== last) {
            //console.log("setting tab stop to " + $button.attr("id"));  console.log("$(last).length:");  console.log($(last).length);
            
            $(last).attr( "tabindex", "-1" ); // no-op iff $(last) is empty iff (see comment above)
            $button.attr( "tabindex", "0" ); // no-op iff $button is empty iff (see comment above)
            this._lastTabStop = button;
        }
    },
    
    // No return value.
    _handleKeyDown: function(event, $button) { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        switch (event.which) {
            case $.ui.keyCode.UP:   // up arrow
            case $.ui.keyCode.DOWN: // down arrow
                if ( $button.attr("type")!="radio" )
                    break;
                // fall thru for radio only.  See comments below.
            case $.ui.keyCode.LEFT:  // left arrow
            case $.ui.keyCode.RIGHT: // right arrow
                event.preventDefault();
                
                var $enabledButtons = this.$enabledButtons;
                var length = $enabledButtons.length;
                if (length<2) // nowhere to navigate to; currently focused button is the only enabled one in toolbar
                    break;
                
                var oldIndex = $enabledButtons.index($button);
                var increment = (event.which == $.ui.keyCode.DOWN || ((event.which == $.ui.keyCode.RIGHT) ^ this.isRtl)) ? 1 : -1;
                var newIndex = (oldIndex+increment+length)%length; // wrap around if at start/end of toolbar
                
                // When radios are inside an element with role=toolbar, WAI-ARIA doesn't specify how to reconcile its recommended 
                // Toolbar behavior (left/right arrows move focus w/o selecting) and radio behavior (all 4 arrow keys both move focus
                // and check/select that radio).  A11y office recommended treating radios in a Buttonset or Toolbar like other buttons: 
                // Arrow moves focus without selecting, Spacebar selects, which we prefer too.  
                // Previously we did that for only left/right arrows, and disabled up/down arrows, but since both native and WAI-ARIA-
                // compliant radios support up/down arrows, and since JAWS automatically instructs the user to use up/down arrows even 
                // when the radio group is inside a role=toolbar, we now support up/down arrows for radios via the fall-thru above
                // (but still focus only, not select).
                this._getButtonFocusElem($enabledButtons.eq(newIndex)[0]).focus();
                break;
            
            // Don't need Space/Enter handlers.  For all buttons except already-checked radios in some browsers, Space/Enter fire a click event 
            // (natively or manually), which already calls _setTabStop.  For checked radios (which are focused if they're getting 
            // this key event), _setTabStop has already been called for whichever happened 2nd:  focus (an already checked radio) or
            // check (an already focused radio) via click/Space/Enter.  If checking was done programmatically (via Bset.checked option), we require a refresh().
        }
    },
    
    _destroy: function() { // Override of protected base class method.  Method name needn't be quoted since is in externs.js.
        // this entire stmt can be removed once restoreAttrs code is running.
        this.element
            .removeClass( "oj-toolbar oj-component" )
            .removeAttr(oj.Components._OJ_CONTAINER_ATTR)
            .removeAttr( "role" );
    
        // Since we're not destroying the buttons themselves, the restoreAttrs code will NOT take care of this.
        this.$buttons.attr( "tabindex", "0" ); // bsets in a toolbar should not have focusMgmt turned on, so this is OK, but should revert to orig value, not assume 0.

        // Refresh the top-level buttons, and refresh the buttonsets to make them refresh their buttons, so that all toolbar buttons are refreshed.
        // Reason: to make them re-fetch their chroming option, in case it's still set to the default dynamic getter, 
        // which takes its value from the containing toolbar, which is no longer present.
        // Call refresh *after* removing _OJ_CONTAINER_ATTR, so the buttons/sets no longer detect that they're in a toolbar.
        // TBD: Consider only calling refresh() on children that haven't had their chroming option set, i.e. those still using the dynamic getter.
        this._refreshChildren();
    },
    
    // Refreshes the toolbar's child components.
    _refreshChildren: function() { // Private, not an override (not in base class).  Method name unquoted so will be safely optimized (renamed) by GCC as desired.
        if (!this._IsCustomElement()) {
            this.$buttonsets.ojButtonset( "refresh" );
            this.$topLevelButtons.ojButton( "refresh" );
        }
        else {
            // This relies on the button and buttonset components being implemented with JQUI - this will need to be revisited once that is no longer the case
            for (var i = 0; i < this.$topLevelChildren.length; i++)
            {
                var child = this.$topLevelChildren[i];
                if (child.tagName == "OJ-BUTTON" || child.tagName == "OJ-MENU-BUTTON")
                {
                    // must check to make sure the child button element has been initialized
                    if (oj.Components.__GetWidgetConstructor(this._getButtonFocusElem(child), "ojButton"))
                        child.refresh();
                }
                else if (child.tagName == "OJ-BUTTONSET-ONE" || child.tagName == "OJ-BUTTONSET-MANY")
                {
                    // must check to make sure the child buttonset element has been initialized
                    if (oj.Components.__GetWidgetConstructor(child, "ojButtonset"))
                        child.refresh();
                }
            }
        }
    }
    
    // API doc for inherited methods with no JS in this file:

    /**
     * Returns a <code class="prettyprint">jQuery</code> object containing the root element of the Toolbar component.
     * 
     * @method
     * @name oj.ojToolbar#widget
     * @memberof oj.ojToolbar
     * @instance
     * @ignore
     * @return {jQuery} the root element of the component
     */

    /**
     * Removes the toolbar functionality completely. This will return the element back to its pre-init state, 
     * and remove the toolbar's focus management from the contained buttons.
     * 
     * @method
     * @name oj.ojToolbar#destroy
     * @memberof oj.ojToolbar
     * @instance
     * @ignore
     */
    
    // Fragments:
    
    /**
     * <p>All Toolbar touch interaction is with the individual buttons.  See the [JET Button]{@link oj.ojButton} touch gesture doc.
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojToolbar
     */
    
    /**
     * <p>JET Toolbar is a single tabstop, with arrow-key navigation within the toolbar, as follows:
     * 
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td><kbd>LeftArrow</kbd></td>
     *       <td>Navigate to the previous enabled button on the left, wrapping around at the end.</td>
     *     </tr>
     *     <tr>
     *       <td><kbd>RightArrow</kbd></td>
     *       <td>Navigate to the next enabled button on the right, wrapping around at the end.</td>
     *     </tr>
     *   </tbody>
     * </table>
     * 
     * <p>See also the [JET Button]{@link oj.ojButton} keyboard doc, for details on interacting with 
     * the individual buttons.
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojToolbar
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     * 
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocExampleHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td>oj-toolbar-separator</td>
     *       <td>Separators should be placed around any buttonsets in the toolbar, and anywhere else in the toolbar that 
     *           a separator is desirable.  For accessibility, additionally apply <code class="prettyprint">role</code> 
     *           and <code class="prettyprint">aria-orientation</code> as shown.
     *       </td>
     *       <td>
     * <pre class="prettyprint">
     * <code>&lt;oj-toolbar id="myToolbar" aria-label="Foo" aria-controls="bar">
     *   &lt;oj-button ...>&lt;/oj-button>
     *   &lt;span role="separator" aria-orientation="vertical"
     *         class="oj-toolbar-separator">&lt;/span>
     *   &lt;oj-button ...>&lt;/oj-button>
     * &lt;/oj-toolbar>
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-toolbar-top-border</td>
     *       <td>Applies a top border to the toolbar, or to the <code class="prettyprint">oj-toolbars</code> element, 
     *           in themes not having this border by default.</td>
     *       <td rowspan=3>
     * <pre class="prettyprint">
     * <code>&lt;!-- These classes can be mixed and matched, and can also be applied to the 
     *      oj-toolbars element seen in the next example. -->
     * &lt;oj-toolbar id="myToolbar" aria-label="Foo" aria-controls="bar"
     *      class="oj-toolbar-top-border oj-toolbar-bottom-border oj-toolbar-no-chrome">
     *   &lt;!-- toolbar contents -->
     * &lt;/oj-toolbar>
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-toolbar-bottom-border</td>

     *       <td>Applies a bottom border to the toolbar, or to the <code class="prettyprint">oj-toolbars</code> element, 
     *           in themes not having this border by default.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-toolbar-no-chrome</td>
     *       <td>Removes "chrome" (background and border) from the toolbar(s), in themes having this chrome by default.</td>
     *     </tr>
     *     <tr>
     *       <td>oj-toolbars</td>
     *       <td>An outer element representing a multiple toolbar layout. Contains one or more <code class="prettyprint">oj-toolbar-row</code> elements.</td>
     *       <td rowspan=2>
     * <pre class="prettyprint">
     * <code>&lt;div class="oj-toolbars">
     *   &lt;div class="oj-toolbar-row">
     *     &lt;oj-toolbar id="toolbar1" ...>...&lt;/oj-toolbar>
     *     &lt;oj-toolbar id="toolbar2" ...>...&lt;/oj-toolbar>
     *   &lt;/div>
     *   &lt;div class="oj-toolbar-row">
     *     &lt;oj-toolbar id="toolbar3" ...>...&lt;/oj-toolbar>
     *   &lt;/div>
     * &lt;/div>
     * </code></pre>
     *       </td>
     *     </tr>
     *     <tr>
     *       <td>oj-toolbar-row</td>
     *       <td>Each <code class="prettyprint">oj-toolbar-row</code> element is a row containing one or more toolbars.
     *           These rows go inside an <code class="prettyprint">oj-toolbars</code> element.</td>
     *     </tr>
     *   </tbody>
     * </table>
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojToolbar
     */
});

// -----------------------------------------------------------------------------
// "private static members" shared by all toolbars: 
// -----------------------------------------------------------------------------

/**
 * This is the same as _radioGroup() in Buttonset's code, minus the code branches that toolbar's caller never reaches.
 * In the cases that toolbar's caller reaches, the behavior should be the same.  So if Buttonset and Toolbar ever share code, 
 * keep Buttonset's copy of this function for use by both components.
 * 
 * In all cases, the return value includes only radios that are an :oj-button, i.e. radios that have been buttonized.
 * 
 * Where this method looks for radio groupmates:  This method will only look in $elems, and will not attempt to weed out any 
 * false positives as defined below.  (So return value includes the specified radio iff it's an :oj-button in $elems.)
 * 
 * Radios w/ name="" (incl name omitted) are not in a radio group (i.e. no SelectOne semantics), not even with other radios with 
 * w/ name="".  So if radio is nameless, the return value will include only radio (or nothing at all if it isn't an :oj-button, or 
 * if $elems doesn't include radio).  
 * 
 * False positives: radios with nonempty names that match radio's name, but are actually not groupmates (i.e. no SelectOne 
 * relationship), e.g. because they're from a different form.
 * 
 * 
 * @param {!Element} radio  a radio button.  Not a JQ object, other button or element type, or null.
 * @param {jQuery=} $elems  optional JQ object, containing 0 or more elems that aren't necessarily radios or buttons, in which to look for groupmates.
 *                          E.g. the elements in a buttonset or toolbar.  Must not contain any false positives as defined above.
 * @private
 */
var _radioGroup = function( radio, $elems ) {
    var name = radio.name,
        $radios;
    if ( name ) {
        name = name.replace( /'/g, "\\'" ); // escape single quotes
        var selector = ":radio[name='" + name + "']:oj-button";
        $radios = $elems.filter( selector );
    } else {
        $radios = $elems.filter(radio).filter(":oj-button");
    }
    return $radios;
};

}() ); // end of Toolbar wrapper function


// Set theme-based defaults
oj.Components.setDefaultOptions({
    // same Q as for Button: does this correctly handle the case where the theme has no $var, in which case we want there to effectively be no dynamic getter so that the prototype default is used?
    'ojToolbar': {
        'chroming': oj.Components.createDynamicPropertyGetter( function(context) {
            return (oj.ThemeUtils.parseJSONFromFontFamily('oj-toolbar-option-defaults') || {})["chroming"];
        })
    }
});
(function() {
var ojToolbarMeta = {
  "properties": {
    "chroming": {
      "type": "string",
      "enumValues": ["half", "full", "outlined"]
    }
  },
  "extension": {
    _WIDGET_NAME: "ojToolbar"
  }
};
oj.CustomElementBridge.registerMetadata('oj-toolbar', 'baseComponent', ojToolbarMeta);
oj.CustomElementBridge.register('oj-toolbar', {'metadata': oj.CustomElementBridge.getMetadata('oj-toolbar')});
})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/**
 * @ojcomponent oj.ojTrain
 * @augments oj.baseComponent
 * @ojstatus preview
 * @ojshortdesc Displays a navigation visual element that allows an application user to go between different steps of a process.
 * @since 1.0.0
 * @classdesc
 * <h3 id="trainOverview-section">
 *   JET Train
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#trainOverview-section"></a>
 * </h3>
 *
 * <p>The JET Train element is a navigation visual that allows a user to go between different "steps".
 * Each step can display information about the state of the step("visited", "unvisited", "disabled")
 * and display a message type("error", "confirmation", "warning", "info", "fatal")</p>
 *
 * <pre class="prettyprint"><code>&lt;oj-train
 *   selected-step="{{currentStepValue}}"
 *   steps="{{stepArray}}">
 * &lt;/oj-train></code></pre>
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="styling-section">
 *   Styling
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#styling-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"stylingDoc"}
 */
(function()
{
  oj.__registerWidget("oj.ojTrain", $['oj']['baseComponent'],
  {
    version: "1.0.0",
    defaultElement: "<div>",
    widgetEventPrefix: "oj",
    options: {

      /**
       * @typedef {Object} oj.ojTrain.Step
       * @property {string} id string id of the step
       * @property {string} label string label of the step
       * @property {boolean} [disabled] boolean indicates whether the step is disabled
       * @property {boolean} [visited] boolean indicates whether the step has been visited
       * @property {"info"|"error"|"fatal"|"warning"} [messageType] string the type of message icon displayed on the step
       *
       */

      /**
       * The array of step objects. Each <a href="#Step">oj.ojTrain.Step</a> must have an 'id' and 'label' property.
       *
       * @ojshortdesc The array of step objects.
       * @ojrequired
       * @expose
       * @public
       * @type {Array.<Object>}
       * @ojsignature {target: "Type", value:"Array<oj.ojTrain.Step>", jsdocOverride: true}
       * @instance
       * @memberof! oj.ojTrain
       * @example <caption>Initialize the Train with the <code class="prettyprint">steps</code> attribute specified</caption>
       * &lt;oj-train steps='[{"id":"id1", "label":"Label 1"}, {"id":"id2", "label":"Label 2", "disabled": false, "visited": true, "messageType":"info"}]'>&lt;/oj-train>
       * @example <caption>Get or set the <code class="prettyprint">steps</code> property after initialization</caption>
       * //Get one
       * var step = myTrain.steps[0];
       *
       * //Get all
       * var steps = myTrain.steps;
       *
       * //Set all.  Must list every resource key, as those not listed are lost.
       * myTrain.steps = [{
       *      'id':'id1',
       *      'label': 'Label 1'},
       *     {'id':'id2',
       *      'label': 'Label 2',
       *      "disabled": false,
       *      "visited": true,
       *      "messageType":"info"}];
       */
       steps: [],

      /**
       * The selected variable indicates the id of the current selected step.
       * @expose
       * @public
       * @type {string}
       * @deprecated 3.0.0 use selectedStep
       * @ignore
       * @instance
       * @memberof! oj.ojTrain
       */
      selected: "",

      /**
       * Indicates the id of the current selected step.  Default is the first step in the steps array.
       * @ojshortdesc Indicates the id of the current selected step.
       * @expose
       * @public
       * @type {string}
       * @instance
       * @memberof! oj.ojTrain
       * @example <caption>Initialize the Train with the <code class="prettyprint">selected-step</code> attribute specified</caption>
       * &lt;oj-train selected-step='{{selectedStepValue}}'>&lt;/oj-train>
       * @example <caption>Get or set the <code class="prettyprint">selectedStep</code> property after initialization</caption>
       * //Get train selected step
       * var selectedStep = myTrain.selectedStep;
       *
       * //Set train selected step
       * myTrain.selectedStep = 'stp1';
       */
      selectedStep: "",

      // Events

      /**
       * Triggered immediately before a step is deselected.
       * The ojBeforeDeselect can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>.
       * @ojshortdesc Triggered immediately before a step is deselected.
       * @ojcancelable
       * @expose
       * @event
       * @memberof oj.ojTrain
       * @instance
       * @property {string} toStep The step that is about to be deselected.
       * @property {string} fromStep The step that is about to be selected.
       */
      beforeDeselect: null,

      /**
       * Triggered after a step has been deselected.
       * The ojDeselect can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>.
       * @ojshortdesc Triggered after a step has been deselected.
       * @expose
       * @event
       * @memberof oj.ojTrain
       * @instance
       * @property {string} toStep The step that is about to be deselected.
       * @property {string} fromStep The step that is about to be selected.
       *
       */
      deselect: null,

      /**
       * Triggered immediately before a Step is selected.
       * The ojBeforeSelect can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>.
       * @ojshortdesc Triggered immediately before a Step is selected.
       * @ojcancelable
       * @expose
       * @event
       * @memberof oj.ojTrain
       * @instance
       * @property {string} toStep The step that is about to be deselected.
       * @property {string} fromStep The step that is about to be selected.
       *
       */
      beforeSelect: null,

      /**
       * Triggered after a step has been selected.
       * @ojshortdesc Triggered after a step has been selected.
       * @expose
       * @event
       * @memberof oj.ojTrain
       * @instance
       * @property {string} toStep The step that is about to be deselected.
       * @property {string} fromStep The step that is about to be selected.
       *
       */
      select: null
    },

    /**
     * Variable for storing the number of steps in the train
     *
     * @private
     */
    _stepNum: 0,

    /**
     * Variable for storing the train information about each step in 2D Array form
     *
     * @private
     */
    _stepArray: null,

    /**
     * Create the train
     * @override
     * @memberof oj.ojTrain
     * @return {void}
     * @protected
     */
    _ComponentCreate: function()
    {
      this._super();
      this._setupTrain();
    },

    /**
     * Set up the train component.
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _setupTrain: function() {
      // Constrain initial value
      var options = this.options;
      var steps = options.steps;
      this._stepNum = steps.length;

      // Train Wrapper
      this._wrapper = $("<div class='oj-train-wrapper'></div>");

      this._wrapper.appendTo(this.element);// @HTMLUpdateOK

      this._connectorWrapper = $("<div class='oj-train-connector-wrapper'></div>");

      this._connectorWrapper.appendTo(this._wrapper);// @HTMLUpdateOK

      var styleClass = this.element.attr('class');
      this._stretch = (styleClass != null && styleClass.indexOf("oj-train-stretch") >= 0);
      if (this._stretch)
        this._connectorWrapper.css("padding", "0 " + 100 / (this._stepNum * 2) + "%");

      // Draw the connector bar  for the train
      this._connector = $("<div class='oj-train-connector'></div>");

      this._connector.appendTo(this._connectorWrapper);// @HTMLUpdateOK

      this._stepList = $("<ul>");
      this._stepList.addClass('oj-train-step-list');


      // Initialize the background progress bar object that will be updated to have the correct width based on the current step.
      this._connectorFill = $("<div class='oj-train-connector-fill'></div>");

      this._connectorFill.appendTo(this._connectorWrapper);// @HTMLUpdateOK

      // Setup array that stores train information for each step.
      this._setupArray();
      this._selectedIndex = this._getStepIndex(options.selectedStep || options.selected);
      if (this._selectedIndex === -1 && steps[0] && steps[0]['id'])
      {
        this._selectedIndex = 0;
        options.selectedStep = steps[0]['id'];
        options.selected = steps[0]['id'];
      }

      // Draw each step. Visually each step consists of a background circle, a button, an icon, and a label.
      for (var i = 0; i < this._stepNum; i++)
      {
        // Create a list item to store each step.
        var stepTag = $("<li>").addClass("oj-train-step-list-item").attr({'id': this._stepArray[i][1]})

        // Add message information
        var messageType = this._stepArray[i][4];
        if (messageType === "confirmation")
          stepTag.addClass("oj-confirmation");
        else if (messageType === "info")
          stepTag.addClass("oj-info");
        else if (messageType === "error" || messageType === "fatal")
          stepTag.addClass("oj-invalid");
        else if (messageType === "warning")
          stepTag.addClass("oj-warning");

        stepTag.appendTo(this._stepList);// @HTMLUpdateOK

        this._drawLabel(i);
        this._drawStepFill(i);
        this._drawIndividualConnectorLine(i);
        this._drawButton(i);
        this._drawStepNumberLabel(i);
        this._drawMessageType(i);
        if (this._stretch)
          stepTag.css("width", 100 / (this._stepNum) + "%");
            }

      // Update background progress bar width to show the progress.
      var connectorFillWidth = this._stepNum - 1 === this._selectedIndex ? 100 : (100 / (2 * (this._stepNum - 1)) + this._selectedIndex / (this._stepNum - 1) * 100)
      this._connectorFill.css({'width': connectorFillWidth + "%"});

      this._stepList.appendTo(this._wrapper);// @HTMLUpdateOK

      this.element.addClass("oj-train");
    },

    /**
     * Set up the _stepArray(). The first dimension indicates the step index and the second dimension indicates the step variables.
     * The order of the step variables are: label, id, selection, visited, messageType
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _setupArray: function() {
      var options = this.options;
      this._stepArray = [];
      for (var i = 0; i < this._stepNum; i++) {
        var step = options.steps[i];
        this._stepArray[i] = new Array(5);
        this._stepArray[i][0] = step['label'] ? step['label'] : null;
        this._stepArray[i][1] = step['id'] ? step['id'] : null;
        this._stepArray[i][2] = step['disabled'] ? true : false;
        this._stepArray[i][3] = step['visited'] ? true : false;
        this._stepArray[i][4] = step['messageType'] ? step['messageType'] : null;
      }
    },

    /**
     * Draw the button object for the step based on the index. If a button for that step alread exists remove it and draw the new one.
     * @param {number} index The index of the step for which the button is being drawn.
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _drawButton: function(index)
    {
      var button = $("<div/>")
              .addClass("oj-train-button");
      var scrnRead = $("<span/>");
      var self = this;
      var desc = "";
      if (this._stepArray[index])
      {
        var visited = this._stepArray[index][3];
        var disabled = this._stepArray[index][2];
        if (this._selectedIndex === index) {
          button.addClass("oj-selected");
          desc = " current ";
        }
        else
        {
          if (visited && !disabled) {
            button.addClass("oj-visited");
            desc = " visited ";
          }
          else if (!visited && !disabled) {
            button.addClass("oj-default");
            desc = " not visited ";
          }
          else
            button.addClass("oj-disabled");
        }

        if (!this._stepArray[index][2] && this._selectedIndex !== index) {
          this._AddHoverable(button);
          this._AddActiveable(button);

          button.on("click" + this.eventNamespace, function(event) {
            self._fireSelectedStepOptionChange(this.parentNode.parentNode.id, event);
            self.refresh();
            self._setFocus(this.parentNode.parentNode.id);
          });
        }
        ;
        var stepBackground = this._stepList.children().eq(index).find('.oj-train-button-connector');

        // Check that there are at least 3 items in the list item indicating that a button has already been created for this step.
        // If there is remove it and insert the new button in the same position. If there isn't a preexsisting button, simply add one.
        if (stepBackground.length >= 1) {
          stepBackground.children().remove();
          stepBackground.append(button);// @HTMLUpdateOK
        }
        else
          stepBackground.append(button);// @HTMLUpdateOK
        scrnRead.text(desc);
        scrnRead.addClass("oj-helper-hidden-accessible");

        this._stepList.children().eq(index).find('a').append(scrnRead);// @HTMLUpdateOK
      }
    },

    /**
     * Draw the icon that displays the messageType for the step based on the index. If there already is an icon remove it and draw the new one.
     * @param {number} index The index of the step for which the icon is being drawn.
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _drawMessageType: function(index)
    {
      if (this._stepArray[index] && this._stepArray[index][4])
      {
        var icon = $("<div/>")
                .addClass("oj-train-icon oj-component-icon")
                .attr( "aria-hidden", "true" );
        var scrnRead = $("<span/>");
        var desc = "";
        var self = this;
        var messageType = this._stepArray[index][4];

        if (messageType === "confirmation") {
          icon.addClass("oj-confirmation");
          desc = " Confirmation ";
        }
        else if (messageType === "info") {
          icon.addClass("oj-info");
          desc = " Info ";
        }
        else if (messageType === "error") {
          icon.addClass("oj-error");
          desc = " Error ";
        }
        else if (messageType === "fatal") {
          icon.addClass("oj-error");
          desc = " Error ";
        }
        else if (messageType === "warning") {
          icon.addClass("oj-warning");
          desc = " Warning ";
        }

        // Remove previous messageType
        var button = this._stepList.children().eq(index).find('.oj-train-button');
        if (button.children().length >= 2) {
          button.children()[1].remove();
        }
        // Make icon clickable
        if (!this._stepArray[index][2] && this._selectedIndex !== index) {
          icon.on("click" + this.eventNamespace, function(event) {
            self._fireSelectedStepOptionChange(this.parentNode.parentNode.parentNode.id, event);
            self.refresh();
            self._setFocus(this.parentNode.parentNode.parentNode.id);
          });
        }
        // Add new message
        if (messageType != null) {
          // If there is remove it and insert the new icon in the same position.
          // Add description to span
          scrnRead.text(desc);
          scrnRead.addClass("oj-helper-hidden-accessible");

          this._stepList.children().eq(index).find('a').append(scrnRead);// @HTMLUpdateOK
          button.append(icon); // @HTMLUpdateOK

        }
      }
    },

    /**
     * Fire optionChange event
     * @param {string} previousValue
     * @param {string} value
     * @param {Object} originalEvent
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _fireOptionChange: function(previousValue, value, originalEvent)
    {
      var eventData = {
        "fromStep": this.getStep(previousValue),
        "toStep": this.getStep(value),
        // (originalEvent is non-null) iff (option change is due to user interaction) iff (binding should write back the value)
        "optionMetadata": {'writeback': originalEvent ? "shouldWrite" : "shouldNotWrite"}
      };
      if (this._trigger("beforeDeselect", originalEvent, eventData) === false || this._trigger("beforeSelect", originalEvent, eventData) === false)
        return;

      // Set the previous step to visited
      var stepIndex = this._getStepIndex(previousValue);
      if (stepIndex !== -1) {
        var oldStepProperties = this.options.steps[stepIndex];
        oldStepProperties['visited'] = true;
      }

      this._trigger("deselect", originalEvent, eventData);

      if (this.options.selectedStep)
        this.option('selectedStep', value, {'_context': {originalEvent: originalEvent, internalSet: true}});
      if (this.options.selected)
        this.option('selected', value, {'_context': {originalEvent: originalEvent, internalSet: true}});

      this._trigger("select", originalEvent, eventData);

    },

    /**
     * Draw the number label that is displayed on the button of each step without a message type. This is only used for native mobile themes.
     * @param {number} index The index of the step for which the number label is being drawn.
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _drawStepNumberLabel: function(index) {
      var stepNumberLabel = $("<div/>")
               .addClass("oj-train-button-text");
      stepNumberLabel.append((index + 1).toString());// @HTMLUpdateOK
      var stepBackground = this._stepList.children().eq(index).find('.oj-train-button');
      stepBackground.append(stepNumberLabel);// @HTMLUpdateOK
    },

    /**
     * Draws the individual connector line between each step.
     * @param {number} index The index of the step for which the icon is being drawn.
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _drawIndividualConnectorLine: function(index) {
      if(index != this._stepNum - 1) {
        var stepConnector = $("<div/>")
                 .addClass("oj-train-step-individual-connector");
        if(index < this._selectedIndex) {
          stepConnector.addClass("oj-train-connector-before-selected-step");
        }
        this._stepList.children().eq(index).prepend(stepConnector);// @HTMLUpdateOK
      }
    },

    /**
     * Draw the background circle for the step which is either light or dark base on if the step is before or after the selected step.
     * @param {number} index The index of the step for which the icon is being drawn.
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _drawStepFill: function(index)
    {
      var stepFill = $("<div/>");
      stepFill.addClass("oj-train-button-connector");
      if (this._stepArray[index])
      {
        if (index <= this._selectedIndex)
          stepFill.addClass("oj-train-fill");
        var stepLi = this._stepList.children().eq(index).children();
        stepFill.insertBefore(stepLi);// @HTMLUpdateOK
      }
    },

    /**
     * Draw the label for the step.
     * @param {number} index The index of the step for which the icon is being drawn.
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _drawLabel: function(index)
    {
      var self = this;
      if (this._stepArray[index])
      {
        var labelWrapper = $("<div/>")
                .addClass("oj-train-label-wrapper");
        var label = $("<a>");
        label.text(this._stepArray[index][0]);
        var disabled = this._stepArray[index][2];
        labelWrapper.append(label);// @HTMLUpdateOK
        label.addClass("oj-train-label");
        if (index === this._selectedIndex)
          label.addClass("oj-selected");
        else if (this._stepArray[index][3] && !disabled)
          label.addClass("oj-visited");
        else if (disabled)
          label.addClass("oj-disabled");
        if (!disabled) {
          label.attr("href", "#");
          this._focusable({
            'element': label,
            'applyHighlight': true
          });
          this._AddHoverable(label);
          this._AddActiveable(label);

          label.on("click keydown" + this.eventNamespace,
            /**
             * @suppress {missingProperties}
             */
            function(event) {
              if (event.keyCode === $.ui.keyCode.ENTER || event.type === "click") {
                event.preventDefault();
                self._fireSelectedStepOptionChange(this.parentNode.parentNode.id, event);
                self.refresh();
                self._setFocus(this.parentNode.parentNode.id);
        }
            });
                }
        var stepLi = this._stepList.children().eq(index).children();
        if (stepLi.length >= 2)
          stepLi[1].remove();
        this._stepList.children().eq(index).append(labelWrapper);// @HTMLUpdateOK
      }
    },

    /**
     * Return the index of the step
     * @param {string} id The id of the step whose index is returned.
     * @return {number} index of step. -1 for not valid ids.
     * @memberof oj.ojTrain
     * @private
     */
    _getStepIndex: function(id)
    {
      for (var i = 0; i < this._stepNum; i++) {
        if (this._stepArray[i] && this._stepArray[i][1] === id)
          return i;
      }
      return -1;
    },

    /**
     * Returns the step based on the id passed in. If the step doesn't exist, return null.
     * @ojshortdesc Returns the step based on the id passed in.
     * @public
     * @param {string} id The id of the step.
     * @return {Object | null} step object.
     * @ojsignature { target: "Type",value: "oj.ojTrain.Step|null",for: "returns", jsdocOverride:true}
     * @expose
     * @instance
     * @memberof oj.ojTrain
     */
    getStep: function(id)
    {
      for (var i = 0; i < this._stepNum; i++) {
        if (this._stepArray[i] && this._stepArray[i][1] === id)
          return $.extend({}, this.options.steps[i]);
      }
      return null;
    },

    /**
     * <p>Returns the id of the next selectable step based on the selected id. If the current step is the last selectable step, returns null</p>
     * @public
     * @expose
     * @return {string} next selectable Id
     * @deprecated 3.0.0 Use getNextSelectableStep
     * @ignore
     * @instance
     * @memberof oj.ojTrain
     */
    nextSelectableStep: function()
    {
      return this.getNextSelectableStep();
    },

    /**
     * <p>Returns the id of the previous selectable step based on the selected id. If the current step is the first selectable step, returns null</p>
     * @public
     * @expose
     * @return {string} previous selectable Id
     * @ignore
     * @deprecated 3.0.0 Use getPreviousSelectableStep
     * @instance
     * @memberof oj.ojTrain
     */
    previousSelectableStep: function()
    {
      return this.getPreviousSelectableStep();
    },

    /**
     * Returns the id of the next selectable step based on the current selectedStep. If the current step is the last selectable step, the function returns null
     * @ojshortdesc Returns the id of the next selectable step based on the current selectedStep.
     * @public
     * @return {string | null} next selectable Id
     * @expose
     * @instance
     * @memberof oj.ojTrain
     */
    getNextSelectableStep: function()
    {
      var selectedIndex = this._getStepIndex(this.options.selectedStep || this.options.selected);
      for (var i = selectedIndex; i < this._stepNum; i++) {
        if (i + 1 < this._stepNum && this._stepArray[i + 1] && !this._stepArray[i + 1][2])
        {
          return this._stepArray[i + 1][1];
        }
      }
      return null;
    },

    /**
     * Returns the id of the previous selectable step based on the current selectedStep. If the current step is the first selectable step, the function returns null
     * @ojshortdesc Returns the id of the previous selectable step based on the current selectedStep.
     * @public
     * @return {string | null} previous selectable Id
     * @expose
     * @instance
     * @memberof oj.ojTrain
     */
    getPreviousSelectableStep: function()
    {
      var selectedIndex = this._getStepIndex(this.options.selectedStep || this.options.selected);
      for (var i = selectedIndex; i >= 0; i--) {
        if (this._stepArray[i - 1] && !this._stepArray[i - 1][2])
        {
          return this._stepArray[i - 1][1];
        }
      }
      return null;
    },

    /**
     * <p>Sets the properties for the step. Takes in the object conatining the properties for the step.</p>
     * @public
     * @expose
     * @deprecated 3.0.0 Use updateStep instead
     * @param {Object} stepProperties The property bag to overwrite properties on the step.
     * @ignore
     * @instance
     * @return {void}
     * @memberof oj.ojTrain
     */
    setStep: function(stepProperties)
    {
      if (stepProperties['id'])
      {
        this.updateStep(stepProperties['id'], stepProperties);
      }
    },

    /**
     * Update the step with the specified id with the provided property bag.
     * @ojshortdesc Update the step with the specified id with the provided property bag.
     * @public
     * @param {string} id The id of the step to update
     * @param {Object} stepProperties The set of step properties to update. Will overwrite any previously set values.
     * @expose
     * @param {string} [stepProperties.id] id of step
     * @param {string} [stepProperties.label] label of step
     * @param {boolean} [stepProperties.disabled] whether step is disabled
     * @param {boolean} [stepProperties.visited] whether step has been visited
     * @param {"info"|"error"|"fatal"|"warning"} [stepProperties.messageType] type of message displayed
     * @instance
     * @memberof oj.ojTrain
     * @return {void}
     */
    updateStep: function(id, stepProperties)
    {
      if (id)
      {
        var stepInfo = this.getStep(id);
        var stepIndex = this._getStepIndex(id);
        if (stepIndex !== -1)
        {
          var stepObj = this.options.steps[stepIndex];
          if (stepProperties['label'])
          {
            stepInfo[0] = stepProperties['label'];
            stepObj['label'] = stepProperties['label'];
          }
          if (typeof (stepProperties['id']) === 'string')
          {
            stepInfo[1] = stepProperties['id'];
            stepObj['id'] = stepProperties['id'];
          }
          if (typeof (stepProperties['disabled']) === 'boolean')
          {
            stepInfo[2] = stepProperties['disabled'];
            stepObj['disabled'] = stepProperties['disabled'];
          }
          if (typeof (stepProperties['visited']) === 'boolean')
          {
            stepInfo[3] = stepProperties['visited'];
            stepObj['visited'] = stepProperties['visited'];
          }
          if (stepProperties['messageType'])
          {
            stepInfo[4] = stepProperties['messageType'];
            stepObj['messageType'] = stepProperties['messageType'];
          }
        this.refresh();
      }
      }
    },
    /**
     * <p>Sets the options</p>
     * @protected
     * @param {Object} options
     * @override
     * @memberof oj.ojTrain
     * @return {void}
     */
    _setOptions: function(options) {
      this._super(options);
      this.refresh();
    },
    /**
     * <p>Sets the options</p>
     * @protected
     * @param {string} key
     * @param {Object} value
     * @param {string} flags
     * @override
     * @memberof oj.ojTrain
     * @return {void}
     */
    _setOption: function(key, value, flags)
    {
      if ((key == "selectedStep" || key == "selected") && this._stepArray && this._stepArray[this._selectedIndex])
      {
        var prevSelected = this._stepArray[this._selectedIndex][1];
        if(value == prevSelected)
          return;

        this._fireOptionChange(this._stepArray[this._selectedIndex][1], value, null);
      }
      else
        this._super(key, value, flags);
    },

    /**
     * Refreshes the train.
     *
     * <p>This method does not accept any arguments.</p>
     * @ojshortdesc Refreshes the train.
     * @expose
     * @memberof oj.ojTrain
     * @return {void}
     * @instance
     */
    refresh: function()
    {
      // TODO: Refresh should not call destroy.
      this._super();
      this._destroy();
      this._setupTrain();
    },

    /**
     * @override
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _destroy: function()
    {
      this._stepList.children().each(function() {
        $(this).remove();
      });
      this.element.removeClass("oj-train");
      this.element.find(".oj-train-wrapper").remove();
      this.element.find(".oj-train-connector-wrapper").remove();
      this.element.find(".oj-train-step-list").remove();
      this.element.find(".oj-train-step-list").remove();

      // Call super at the end for destroy.
      this._super();
    },

    /**
     * Fire option change event for selectedStep on click or keyboard action
     * @param {string} event
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _fireSelectedStepOptionChange: function(newSelectedStep, event)
    {
      var prevSelected = this._stepArray[this._selectedIndex][1];
      if(prevSelected != newSelectedStep)
        this._fireOptionChange(prevSelected, newSelectedStep, event);
    },

    /**
     * Set the focus on the step defined by the id.
     * @param {string} id The id of the step to focus
     *
     * @memberof oj.ojTrain
     * @return {void}
     * @private
     */
    _setFocus: function(id)
    {
      var index = this._getStepIndex(id)
      this._stepList.children().eq(index).find('.oj-train-label').focus()
    },

    //** @inheritdoc */
    getNodeBySubId: function(locator)
    {

      if (locator === null)
      {
        return this.element ? this.element[0] : null;
      }

      var subId = locator['subId'],
              index = locator['index'];

      if ((typeof index !== 'number') ||
              index < 0 || index >= this._stepNum)
        return null;

      switch (subId)
      {
        //returns the clickable button
        //QA automated tests need to click on selected step and button is clickable
        //while the whole step is not
        case 'oj-train-step':
          return this._stepList.children().eq(index).find('.oj-train-button')[0];

          // Deprecated sub-id
        case 'oj-train-button':
          return this._stepList.children().eq(index).find('.oj-train-button')[0];

          // Deprecated sub-id
        case 'oj-train-button-connector':
          return this._stepList.children().eq(index).find('.oj-train-button-connector')[0];

          // Deprecated sub-id
        case 'oj-train-connector':
          return this._connector;

          // Deprecated sub-id
        case 'oj-train-connector-fill':
          return this._connectorFill;

          // Deprecated sub-id
        case 'oj-train-icon':
          return this._stepList.children().eq(index).find('.oj-train-icon')[0];

          // Deprecated sub-id
        case 'oj-train-label':
          //return this._stepList.children().eq(index).find('.oj-train-button');

          //  - OJ-TRAIN-LABEL DOES NOT RETURN THE RIGHT NODE
          return this._stepList.children().eq(index).find('.oj-train-label')[0];
      }

      // Non-null locators have to be handled by the component subclasses
      return null;
    },

    //** @inheritdoc */
    getSubIdByNode: function(node)
    {
      var currentNode = node;

      var stepNum = this._stepArray ? this._stepArray.length : 0;
      for (var stepIndex = 0; stepIndex < stepNum; stepIndex++) {
        var stepLocator = {'subId': 'oj-train-step', 'index': stepIndex};
        //subIdNode should be button
        var subIdNode = this.getNodeBySubId(stepLocator);
        //Checking whether currentNode is a descendant of the button or label because
        //those two are only clickable things in the step so we only want to return a valid subIdNode
        //for those cases(and their children)
        if ($(currentNode).closest(subIdNode).length > 0
            || $(currentNode).closest($(subIdNode.parentNode.parentNode).find('.oj-train-label')[0]).length > 0)
          return stepLocator;
      }
      return null;
    }

    // Fragments:
    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Target</th>
     *       <th>Gesture</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *    <tr>
     *       <td>Step</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Move to the targeted step</td>
     *     </tr>
     *     <tr>
     *       <td>Step Label</td>
     *       <td><kbd>Tap</kbd></td>
     *       <td>Move to the targeted step</td>
     *    </tr>
     *   </tbody>
     *  </table>
     *
     *
     * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
     * @memberof oj.ojTrain
     */

    /**
     * <table class="keyboard-table">
     *   <thead>
     *     <tr>
     *       <th>Key</th>
     *       <th>Action</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td><kbd>Tab</kbd></td>
     *       <td> Move focus to the next selectable step. </tr>
     *     <tr>
     *       <td><kbd>Shift+Tab</kbd></td>
     *       <td> Move focus to the previous selectable step.</tr>
     *     <tr>
     *       <td><kbd>Enter</kbd></td>
     *       <td> Select the focused step.</tr>
     *   </tbody>
     *  </table>
     *
     *
     * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
     * @memberof oj.ojTrain
     */

    /**
     * {@ojinclude "name":"ojStylingDocIntro"}
     *
     * <table class="generic-table styling-table">
     *   <thead>
     *     <tr>
     *       <th>{@ojinclude "name":"ojStylingDocClassHeader"}</th>
     *       <th>{@ojinclude "name":"ojStylingDocDescriptionHeader"}</th>
     *     </tr>
     *   </thead>
     *   <tbody>
     *     <tr>
     *       <td> oj-train-stretch</td>
     *       <td> Optional class that may be added to the train div which will stretch the train to cover the full width of the container specified. </td>
     *     </tr>
     *   </tbody>
     * </table>
     * <br>
     *
     * Train step label wrapping is controlled by the $trainLabelTextWrap SASS Variable. $trainLabelTextWrap accepts css white-space values such as normal or nowrap(default).
     *
     * @ojfragment stylingDoc - Used in Styling section of classdesc, and standalone Styling doc
     * @memberof oj.ojTrain
     */

    // Sub_IDs:

    /**
     * <p>Sub-ID for the specified train step.</p>
     *
     * @property {number} index The zero-based index of the train step.
     *
     * @ojsubid oj-train-step
     * @memberof oj.ojTrain
     *
     * @example <caption>Get the second step:</caption>
     * var node = myTrain.getNodeBySubId({'subId': 'oj-train-step', 'index': 1});
     */

    /**
     * <p>Sub-ID for the specified train step button.</p>
     *
     * @property {number} index The zero-based index of the train step.
     *
     * @ojsubid oj-train-button
     * @memberof oj.ojTrain
     *
     * @example <caption>Get the button of the second step:</caption>
     * var node = myTrain.getNodeBySubId({'subId': 'oj-train-button', 'index': 1});
     * @deprecated 1.2.0 This sub-id was deprecated please use oj-train-step instead.
     * @ignore
     */

    /**
     * <p>Sub-ID for the train step button connector background.</p>
     *
     * @property {number} index The zero-based index of the train step.
     *
     * @ojsubid oj-train-button-connector
     * @memberof oj.ojTrain
     *
     * @example <caption>Get the button connector background of the second step:</caption>
     * var node = myTrain.getNodeBySubId({'subId': 'oj-train-button-connector', 'index': 1});
     * @deprecated 1.2.0 This sub-id was deprecated because the returned node is not interactive.
     * @ignore
     */

    /**
     * <p>Sub-ID for the train background connector bar.</p>
     *
     * @ojsubid oj-train-connector
     * @memberof oj.ojTrain
     *
     * @example <caption>Get the background connector bar of the train:</caption>
     * var node = myTrain.getNodeBySubId({'subId': 'oj-train-connector'});
     * @deprecated 1.2.0 This sub-id was deprecated because the returned node is not interactive.
     * @ignore
     */

    /**
     * <p>Sub-ID for the train background connector bar fill inidcating the train progress.</p>
     *
     * @ojsubid oj-train-connector-fill
     * @memberof oj.ojTrain
     *
     * @example <caption>Get the train background connector bar fill:</caption>
     * var node = myTrain.getNodeBySubId({'subId': 'oj-train-connector-fill'});
     * @deprecated 1.2.0 This sub-id was deprecated because the returned node is not interactive.
     * @ignore
     */

    /**
     * <p>Sub-ID for the specified train step icon.</p>
     *
     * @property {number} index The zero-based index of the train step.
     *
     * @ojsubid oj-train-icon
     * @memberof oj.ojTrain
     *
     * @example <caption>Get the icon of the second step:</caption>
     * var node = myTrain.getNodeBySubId({'subId': 'oj-train-icon', 'index': 1});
     * @deprecated 1.2.0 This sub-id was deprecated please use oj-train-step instead.
     * @ignore
     */

    /**
     * <p>Sub-ID for the specified train step label.</p>
     *
     * @property {number} index The zero-based index of the train step.
     *
     * @ojsubid oj-train-label
     * @memberof oj.ojTrain
     *
     * @example <caption>Get the label of the second step:</caption>
     * var node = myTrain.getNodeBySubId({'subId': 'oj-train-label', 'index': 1});
     * @deprecated 1.2.0 This sub-id was deprecated please use oj-train-step instead.
     * @ignore
     */
  });
}());

(function() {
var ojTrainMeta = {
  "properties": {
    "selectedStep": {
      "type": "string",
      "writeback": true
    },
    "steps": {
      "type": "Array<object>"
    }
  },
  "methods": {
    "getStep": {},
    "getNextSelectableStep": {},
    "getPreviousSelectableStep": {},
    "updateStep": {}
  },
  "extension": {
    _WIDGET_NAME: "ojTrain"
  },
  "events": {
    "select": {},
    "deselect": {},
    "beforeDeselect": {},
    "beforeSelect": {}
  },
};
oj.CustomElementBridge.registerMetadata('oj-train', 'baseComponent', ojTrainMeta);
oj.CustomElementBridge.register('oj-train', {'metadata': oj.CustomElementBridge.getMetadata('oj-train')});
})();
/**
  * Copyright (c) 2015, Oracle and/or its affiliates.
  * All rights reserved.
  */
 
/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @class oj.TreeDndContext
 * @classdesc  HTML5 Drag and Drop Support and Utils for ojTree
 * @param {Object} component ojTree instance
 * @constructor
 */
oj.TreeDndContext = function(component)
{
  this.Init(component);
};

// Subclass from oj.Object 
oj.Object.createSubclass(oj.TreeDndContext, oj.Object, "oj.TreeDndContext");

/**
 * Initializes the instance.
 * @export
 */
oj.TreeDndContext.prototype.Init = function(component)
{
  oj.TreeDndContext.superclass.Init.call(this);
  this.component = component;
  this.options   = component['options'];
  this._reset() ;
};


/**
  * Returns true if tree is enabled for drag.
  * @return {boolean}   true if the tree is enabled for drag (or reorder drag).
  */
oj.TreeDndContext.prototype.isDragEnabled = function()
{
  var dnd  = this._dnd ;
  return (dnd.reorder || dnd.dragFromEnabled) ;
};


/**
  * Test if tree is enabled for drop.
  * @return {boolean}   true if the tree is enabled for drop (or reorder drop).
  */
oj.TreeDndContext.prototype.isDropEnabled = function()
{
  var dnd  = this._dnd ;
  return (dnd.reorder || dnd.dropToEnabled) ;
};


/**
  *  Handles changes to the dnd option property. On entry the options "dnd"
  *  property has already been updated in  _setOption().  Here we just blow
  *  away all the old settings and reinit.
  */
oj.TreeDndContext.prototype.handleDnDOptionChange = function(newval)
{
   var dnd     = this._dnd ;
   var cons    = oj.TreeDndContext ;
   var reorder = dnd.reorder ;
   var dragFromEnabled = dnd.dragFromEnabled;

   this._dndCleanAll() ;       // blow away all the old settings
   this.initDnDOpts() ;        // reprocess the options into dnd
   dnd  = this._dnd ;          // _dnd is new so reget

   //  If reorder or dragging ability changed, may need to remove css classes.
   //  (Only need to check for removal - if it was not set previously but is
   //   now added, this will be taken care of _initDnD(). 
   if ((reorder != dnd.reorder) || (dragFromEnabled != dnd.dragFromEnabled))
   {
     if ( (! dnd.reorder) && (! dnd.dragFromEnabled))
     {
       this.component._getContainerList().find("li")
                                         .removeClass(cons._OJ_DRAGGABLE)
                                         .removeClass(cons._OJ_DRAG)
                                         .removeClass(cons._OJ_VALID_DROP)
                                         .removeAttr("draggable") ;
     }
   }

   this.initDnD() ;       // finally re-init DnD support with new options.
};


/**
  *  Process Drag and Drop options.
  *  Note: all relevent DnD listeners have been removed before 
  *  entry in _handleDnDOptionChange()
  */
oj.TreeDndContext.prototype.initDnDOpts = function()
{
  var odnd        = this.options["dnd"] ;
  var dnd         = this._dnd ;
  var cons        = oj.TreeDndContext ;
  var bSetDisable = true ;
  var opts ;
  var ot ;
  var cb ;
  var n ;

  // Handle tree reorder via DnD

  dnd.reorder = false ;
  ot = typeof odnd ;
  if (ot != "object")
  {
    this.options["dnd"] = {"reorder" : "disable"} ;
    return ;
  }

  opts = odnd[cons._DND_REORDER] ;
  if (typeof opts === "string")
  {
    if (opts === "enable")
    {
      dnd.reorder = true ;
      bSetDisable = false ;
    }
    else if (opts == "disable")
    {
      bSetDisable = false ;
    }
  }
  if (bSetDisable)
  {
    this.options["dnd"]["reorder"] = "disable" ;
  }

  // Check dnd  drag options for user app callbacks

  //  Inspect drag option
  opts = odnd[cons._DND_DRAG] ;
  ot   = typeof opts ;
  if (ot == 'object')
  { 
    n = opts[cons._DND_NODE] ;
    if (typeof n == 'object')
    {
      cb = n[cons._DND_DRAGSTART_CB] ;
      if (typeof cb == 'function')
      {
        dnd.dragStartCallback = cb;
        dnd.dragFromEnabled   = true ;
      }
      cb = n[cons._DND_DRAGDRAG_CB] ;
      if (typeof cb == 'function')
      {
        dnd.dragDragCallback = cb;
        dnd.dragFromEnabled  = true ;
      }
      cb = n[cons._DND_DRAGEND_CB] ;
      if (typeof cb == 'function')
      {
        dnd.dragEndCallback = cb;
        dnd.dragFromEnabled = true ;
      }
      cb = n[cons._DND_DATATYPES] ;
      if ($.isArray(cb))
      {
        dnd.dragDataTypes   = cb.slice();
        dnd.dragFromEnabled = true ;
      }
      else if (typeof cb == 'string')
      {
        dnd.dragDataTypes   = [cb];
        dnd.dragFromEnabled = true ;
      }
    }
  }

  //  Inspect drop option
  opts = odnd[cons._DND_DROP] ;
  ot   = typeof opts ;
  if (ot == 'object')
  { 
    n = opts[cons._DND_NODE] ;
    if (typeof n == 'object')
     {
      cb = n[cons._DND_DROPDRAGENTER_CB] ;
      if (typeof cb == 'function')
      {
        dnd.dropDragEnterCallback = cb;
        dnd.dropToEnabled         = true ;
      }
      cb = n[cons._DND_DROPDRAGOVER_CB] ;
      if (typeof cb == 'function')
      {
        dnd.dropDragOverCallback = cb;
        dnd.dropToEnabled        = true ;
      }
      cb = n[cons._DND_DROPDRAGLEAVE_CB] ;
      if (typeof cb == 'function')
      {
        dnd.dropDragLeaveCallback = cb;
        dnd.dropToEnabled         = true ;
      }
      cb = n[cons._DND_DROPDROP_CB] ;
      if (typeof cb == 'function')
      {
        dnd.dropDropCallback = cb;
        dnd.dropToEnabled    = true ;
      }
      cb = n[cons._DND_DATATYPES] ;
      if ($.isArray(cb))
      {
        dnd.dropDataTypes = cb.slice();
        dnd.dropToEnabled = true ;
      }
      else if (typeof cb == 'string')
      {
        dnd.dropDataTypes = [cb];
        dnd.dropToEnabled = true ;
      }
    }
  }

  // Final DnD checks on option values parsed.

  // If no drag dataTypes and no dragStart callback, this tree can't be a drag source
  // unless reorder is permitted..
  if (! dnd.reorder)
  {
    if ((! dnd.dragDataTypes) && (! dnd.dragStartCallback))
    {
      dnd.dragFromEnabled   = false ;
      dnd.dragStartCallback =
      dnd.dragDragCallback  =
      dnd.dragEndCallback   =
      dnd.dragDataTypes     = null ;
    }
    // If no drop callback, this tree can't be a drop target unless reorder is permitted.
    if (! dnd.dropDropCallback)
    {
      dnd.dropToEnabled         = false ;
      dnd.dropDragEnterCallback = 
      dnd.dropDragOverCallback  = 
      dnd.dropDragLeaveCallback = 
      dnd.dropDropCallback      = 
      dnd.dropDataTypes         = null ;
    }
  }

  dnd.pureReorder = (dnd.reorder && (!dnd.dragFromEnabled));

// if (dnd.reorder || dnd.dropToEnabled) {
//   this._applyDefaults(this.options["dnd"], this._dnd.defaults) ;  // not yet exposed
// }

};


/**
  *  Initialize Drag and Drop functionality.
  */
oj.TreeDndContext.prototype.initDnD = function()
{
   var dnd  = this._dnd ;
   var vars = dnd.vars ;
   var cons = oj.TreeDndContext ;

   if ((! dnd.dragFromEnabled) && (! dnd.dropToEnabled) && (! dnd.reorder))
   {
     return ;                                 // No DnD support required
   }

   //  Create drag insertion ptr and line divs if they don't exist.
   if (! oj.TreeDndContext._dragMarker)
   {
     vars.m =                                 // dnd insertion point marker
       $("<div class='oj-tree-drop-marker'><span class='oj-tree-drop-ptr oj-component-icon'>&#160;</span></div>")
         .css("pointer-events", "none")       // prevent mouse events on marker, so that
         .hide()                              // we don't cause bogus dragLeave/dragEnter
         .appendTo("body");                   //@HTMLUpdateOK

     vars.ml = $("<div />")                   // dnd insert point marker line
         .addClass(cons._OJT_DROP_MARKER_LINE)
         .css("pointer-events", "none")       // prevent mouse events on the line, so
         .hide()                              // that we don't cause bogus dragLeave/dragEnter
         .appendTo("body");                   //@HTMLUpdateOK

    oj.TreeDndContext._dragMarker     = vars.m ;
    oj.TreeDndContext._dragMarkerLine = vars.ml;
   }
   else
   {
      vars.m  = oj.TreeDndContext._dragMarker ;
      vars.ml = oj.TreeDndContext._dragMarkerLine ;
   }

   dnd.ml_width = vars.ml.width() ;  // marker line width from style

   // Add Drag/Drop listeners
   var div = this.component.element[0];         // containing div
   var $ul = $(div).children("ul:eq(0)");       // containing UL

   if (dnd.reorder || dnd.dragFromEnabled)
   {
     dnd.fnDragStart = this._dragStart.bind(this) ;
     dnd.fnDragEnd   = this._dragEnd.bind(this) ;
     $(div).on("dragstart", dnd.fnDragStart) ;
     $(div).on("dragend",   dnd.fnDragEnd) ;

     // If here after tree is constructed, it is because of a dnd option change
     if ($ul.length)
     {
       $ul.find("li")
          .addClass(cons._OJ_DRAGGABLE)
          .addClass(cons._OJ_DRAG)
          .addClass(cons._OJ_VALID_DROP)
          .attr("draggable", "true") ;
     }
   }

   if (dnd.reorder || dnd.dropToEnabled)
   {
     dnd.fnDrag      = this._drag.bind(this) ;
     dnd.fnDragEnter = this._dragEnter.bind(this);
     dnd.fnDragOver  = this._dragOver.bind(this) ;
     dnd.fnDragLeave = this._dragLeave.bind(this);
     dnd.fnDragDrop  = this._dragDrop.bind(this) ;
     $(div).on("drag",      dnd.fnDrag) ;
     $(div).on("dragenter", dnd.fnDragEnter) ;
     $(div).on("dragover",  dnd.fnDragOver) ;
     $(div).on("dragleave", dnd.fnDragLeave) ;
     $(div).on("drop",      dnd.fnDragDrop) ;
   }
};



/**
  *  Close down all DnD support.
  */
oj.TreeDndContext.prototype._closedown = function()
{
   this._dndCleanAll() ;
};



//-------------------------------------------------------------------//
//                      Internal methods below                       //
//-------------------------------------------------------------------//

/**
   *  (dragstart event handler).
   *  A drag operation on a node has started..
   *  @private
   */
oj.TreeDndContext.prototype._dragStart =  function(e)
{
   var odt ;
   var odts ;
   var dnd ;
   var vars ;
   var adata ;
   var data ;
   var i ;
   var $targ = $(e.target) ;
   var $node = $targ.closest("li") ;   // target node dragged
   var $nodes ;                        // all nodes participating in the drag
   var dt    = e.originalEvent.dataTransfer ;
   var cons  = oj.TreeDndContext ;
   var ret ;

   dnd  = this._dnd,
   vars = dnd.vars ;

   //this._trace(this._elemId + " _dragStart(" + $node.attr("id") + ") - dragged=" + dnd.nodeDraggedId + 
   //            "  nodeLastEnter=" + 
   //            (dnd.nodeLastEnter? dnd.nodeLastEnter.attr("id") : "null") + " lastLeft=" +
   //            (dnd.nodeLastLeft? dnd.nodeLastLeft.attr("id") : "null"), 1) ;

   vars.o            = null ;      // no dragged node(s) yet
   vars.r            = null ;      // no reference node yet
   dnd.nodeLastEnter = null ;
   dnd.nodeLastLeft  = null ;
   dnd.nodeDraggedId = null ;
   dnd.dragStarted   = false ;

   if ( ((! dnd.reorder) && (!dnd.dragFromEnabled)) || 
        $node.hasClass(TreeUtils._OJ_DISABLED) || this.component._isTreeDisabled())
   {
     e.preventDefault() ;          // drag not allowed.
     return ;
   }

   //  If node types have been defined, check that a move is permitted for this node type.
   if (! this.component._canTypedNodeMove($node, "move")) {
     return false;                 // drag not allowed.
   }

   //  Add dragged node(s) to transfer data.  If dragged node is selected,
   //  then will drag all selected nodes, else just have the one node.
   $nodes = (this.component.isSelected($node))?  this.component._getNode(null, true) :
                                                 $nodes = $node;

   // Set app defined dataTypes to the node data
   if (dnd.dragFromEnabled || dnd.pureReorder)
   {
     this._setDataTypes(dt, $nodes) ;
   }

   //  If only one node is selected, will use the default drag image.
   //  If multiple nodes selected, will create a custom image.
   this._setDragImage(dt, $nodes) ;

   //  Call app callback in drag options
   if (dnd.dragStartCallback)
   {
     adata = this._getDraggedItemArray($nodes) ;
     ret = dnd.dragStartCallback(e, { "item" : adata}) ;
     if (e.isDefaultPrevented() || (typeof ret === 'boolean' && (!ret)))
     {
       return ret ;                             // app vetoed drag
     }

     // If dataTypes was not specified in the drag options, check that callback
     // has created at least one.
     if ((! dnd.dragDataTypes) && (! this._checkAddedDataTypes(dt)) && (! dnd.reorder))
     {
       e.preventDefault() ;
       return false ;                          // dragstart not permitted 
     }
   }     // end if dragStart callback

   // If no callback or callback did not set the effectAllowed, set default value.
   if (dt.effectAllowed === "uninitialized")
   {
     if (dnd.pureReorder)
     {
       dt.effectAllowed  = "move" ; // reorder only & no dragStart callback
     }
     else {
       dt.effectAllowed  = (oj.DomUtils.isMetaKeyPressed(e))? "copy" : "move";
     }
   }

   this._dndStartDrag($nodes, $node, e) ;               // final prep
};


/**
   *  (drag event handler).
   *  A drag operation is continuing (drag event handler).
   *  @private
   */
oj.TreeDndContext.prototype._drag =  function(e)
{
  var dnd  = this._dnd ;

  if (dnd.dragDragCallback)      // check if app callback in drag options?
  {
     dnd.dragDragCallback(e) ;
  }
};


oj.TreeDndContext.prototype._dragEnter =  function(e)
{
   //  Note: due to the composite DOM construction of a node, there
   //  can be several dragEnter calls for the same logical node.

   // If tree is totally empty add a temporary dummy node to act as a reference node.
   if (! this.component._$container_ul[0].firstChild) {
     this._addInternalNode() ;
   }

   var $targ           = $(e.target) ;
   var $node           = this.component._getNode($targ) ;    // <li> of the element entered
   var dnd             = this._dnd ;
   var vars            = dnd.vars ;
   var bNode           = (($node) && ($node.length > 0) && ($node.hasClass("oj-tree-node"))) ;
   var dt              = e.originalEvent.dataTransfer ;
   var cons            = oj.TreeDndContext ;
   var bAllowDataTypes = true ;
   var bForeignTree    = false ;        // true if drag from another tree
   var bClean          = false ;
   var elemId ;
   var pos ;
   var ret ;
   var comp ;
   var comps ;
   var i ;


   // Check if entering the internal placeholder node (created after
   // tree was emptied by previously dragging all nodes out of it.
   if (bNode && $node.hasClass(TreeUtils._OJ_TEMPNODE))
   {
     dnd.bInternalNode = true ;
   }

   // Ignore the UL, only interested in the nodes.
   if (bNode && (e.target.tagName == "UL") && (e.target.getAttribute("role") == "group"))
   {
     return ;   // (prevents bogus insert line under top node)
   }

   // Set dnd.foreign if the drag did not originate from this tree
   dnd.foreign  = (! dnd.dragStarted);  // no preceding _dragStart() so froms another
                                        // component which may or may not be a tree.

   // If over same tree, but re-order not permitted, or from a foreign tree that is
   // defined as reorder only (no transfer data), ignore.
   if ( (! dnd.foreign && (! dnd.reorder)) ||
        (dnd.foreign && this. _isDtType(dt, cons._DND_INTERNAL_DT_REORDER)) )
   {
     this._dndClean() ;
     return ;
   }

   //this._trace(this._elemId + " _dragEnter(" +  e.target.tagName + ", " + 
   //            (bNode? $node.attr("id") : "nonode") + ") - draggedId=" + 
   //            dnd.nodeDraggedId + "  nodeLastEnter=" + " dragStarted=" + dnd.dragStarted + 
   //            (dnd.nodeLastEnter? dnd.nodeLastEnter.attr("id") : "null") + 
   //            " lastLeft=" + (dnd.nodeLastLeft? dnd.nodeLastLeft.attr("id") : "null") +
   //            " vars.o=" + vars.o + " vars.r=" + vars.r + "  dnd.foreign=" + dnd.foreign, 1) ;

   // If foreign object see if it is a tree or non-tree
   if (dnd.foreign)
   {
     bForeignTree = this. _isDtType(dt, cons._DND_INTERNAL_DT) ;
     //this._trace("_dragEnter - " + (bForeignTree? "Foreign Tree" : 
     //                                             "Foreign Object (non-tree)"), 1) ;

     // Not from this tree. If this tree does not allow drops, can't continue.
     if (! dnd.dropToEnabled)
     {
       // Drop of any object except (except for reorder from same tree) not permitted.
       // Also since the dragend event goes to the source component, which in this case
       // is foreign, we must clean-up and hide the drop ui for this component too.
       this._dndClean() ;
       return ;                // vetoed
     }
   }

   vars.r = false  ;           // no reference node yet

   //  Clear last entered node classes
   if (dnd.nodeLastEnter)
   {
     this._clearDropClasses(dnd.nodeLastEnter) ;
   }

   // If we move out of the nodes but are still over the tree container, the
   // reference node is the last node we previously entered.
   if ($targ.hasClass("oj-tree"))
   {
     $node = dnd.nodeLastEnter ;
   }

   //  Check that we have a node and that it is not being dragged over itself
   if ($node && $node.length && ($node.attr("id") == dnd.nodeDraggedId) )
   {
     this._dndHideMarker() ;
     return ;                  // vetoed - drag over self ignored, or don't have a node
   }

   // Handle case where node(s) are being dragged from a foreign tree
   if (bForeignTree && bNode)
   {
     dnd.nodeLastEnter = null ;
     vars.o = 2 ;      // dragged foreign tree node(s) - data is in dataTransfer
   }
   // Handle case where a foreign object (non-tree node) is being dragged to the tree.
   if (dnd.foreign && (! bForeignTree))
   {
     vars.o = 1 ;          // Don't have a source node (non-tree obj)
   }

   vars.r            = $node ;        // the reference node
   dnd.nodeLastEnter = $node ;

   this._dndPlacement(e, $node) ;     // get initial marker placement (in dnd.place)
   if (! this._dndEnter($node))
   {
     // Not a valid placement position
     this._clearDropClasses(vars.r) ;

     vars.r  = null ;                 // no reference node
     this._dndHideMarker() ;
     return ;
   }

   vars.r.removeClass(cons._OJ_INVALID_DROP)
         .addClass(cons._OJ_VALID_DROP)
         .addClass(cons._OJ_ACTIVE_DROP) ; 

   //  Call dragEnter callback in the drop options if defined
   if (dnd.dropDragEnterCallback)
   {
     ret = dnd.dropDragEnterCallback(e, {"item"      : dnd.bInternalNode? null :$node, 
                                         "position"  : dnd.bInternalNode? "first" : dnd.place,
                                         "reference" : dnd.bInternalNode? null : vars.r}) ;
   } 

   if (ret === undefined)
   {
     if (this._isDtAcceptDataTypes(dt))
     {
       e.preventDefault() ;        // accept node or position beside it for a drop
       e.stopPropagation();
     }
     else
     {
       // component rejected by dataTypes.
       // If the source component is foreign, we must clean-up and hide the drop
       // ui for this component too.
       bClean = (dnd.foreign || (! dnd.reorder)) ;
     }
   }
   else
   {
     // callback returned something
     if (ret === false || e.isDefaultPrevented())
     {
       e.preventDefault() ;             // accept node or position beside it for a drop
       e.stopPropagation();
     }
     else if (dnd.foreign || (! dnd.reorder))
     {
       // If the source component is foreign, we must clean-up and hide the drop
       // ui for this component too.
       bClean = (dnd.foreign || (! dnd.reorder)) ;
     }
   }

   if (bClean)
   {
     this._dndClean() ;
   }

   return ret ;
};


/**  (dragover event handler).
  *  In drag mode and mouse has moved over a node (or between nodes - i.e.
  *  still over the tree containing div).
  *  @private
  */
oj.TreeDndContext.prototype._dragOver =  function(e)
{
  var dnd    = this._dnd ;
  var vars   = dnd.vars ;
  var dt     = e.originalEvent.dataTransfer ;
  var bClean = false ;
  var elemId ;
  var $node ;
  var ret;

  // -------Debugging----------
  //var $n = this._getNode($(e.target));
  //this._trace(this._elemId + " _dragOver(" + e.target.tagName + ", " +
  //             (($n && $n.length)? $n.attr("id") : "nonode") + ") - dragged=" + dnd.nodeDraggedId + 
  //              "  nodeLastEnter=" + 
  //              (dnd.nodeLastEnter? dnd.nodeLastEnter.attr("id") : "null") + " lastLeft=" +
  //              (dnd.nodeLastLeft? dnd.nodeLastLeft.attr("id") : "null") +
  //              " vars.o=" + vars.o + " vars.r=" + vars.r + " dnd.foreign=" + dnd.foreign, 1) ;
  // -------End Debugging--------

  // Did _dragEnter() determine node is a drop target?
  if ((! vars.r) || (! vars.r.length))
  {
    this._dndHideMarker() ;
    return ;                 // no.
  }

  $node = vars.r ;           // the reference node set by _dragEnter()

  if (!$node.hasClass("oj-tree-node"))
  {
    this._dndHideMarker() ;
    return ;
  }

  elemId = $node.attr("id") ;
  if (elemId == dnd.nodeDraggedId)    // ignore drag over the dragged node itself
  {
    this._dndHideMarker() ;
    return ;
  }

  this._dndPlacement(e, $node) ;
//this._dndDrag(e) ;

  if (! this._dndPrepare())        // returns "after", "before", "inside" if
  {                                // drop position is acceptable
    this._dndHideMarker() ;
    return ;
  }

  // Call app dragover callback if defined in dnd options and check if veto
  if (dnd.dropDragOverCallback)
  {
    ret = dnd.dropDragOverCallback(e, {"item"      : dnd.bInternalNode? null : $node, 
                                       "reference" : dnd.bInternalNode? null : vars.r,
                                       "position"  : dnd.bInternalNode? "first" : vars.last_pos}) ;
  }

  if (ret === undefined)
  {
    if (this._isDtAcceptDataTypes(dt))
    {
      e.preventDefault() ;        // accept node or position beside it for a drop
      e.stopPropagation();
    }
    else
    {
      // component rejected by dataTypes.
      // If the source component is foreign, we must clean-up and hide the drop
      // ui for this component too.
      bClean = (dnd.foreign || (! dnd.reorder)) ;
    }
  }
  else
  {
     // callback returned something
     if (ret === false || e.isDefaultPrevented())
     {
       e.preventDefault() ;           // accept node or position beside it for a drop
       e.stopPropagation();
     }
     else if (dnd.foreign || (! dnd.reorder))
     {
       // If the source component is foreign, we must clean-up and hide the drop
       // ui for this component too.
       bClean = (dnd.foreign || (! dnd.reorder)) ;
     }
  }

  if (bClean)
  {
    this._dndClean() ;
  }

  return ret ;
};


/**
  *  (drop event handler).
  *  A dragged node has been dropped.
  *  @private
  */ 
oj.TreeDndContext.prototype._dragDrop =  function(e)
{
   var dnd        = this._dnd ;
   var vars       = dnd.vars ;
   var dt         = e.originalEvent.dataTransfer ;
   var foreignObj = (typeof vars.o == 'boolean' && vars.o) ;  // true if not a tree node
   var veto       = false ;
   var ret ;
    
   // -- debugging --
   // var $t = $(e.target);
   // var $n = this._getNode($t);
   // this._trace(this._elemId + " _dragDrop(" + $n.attr("id") + ") - dnd.nodeDraggedId=" +
   //                  dnd.nodeDraggedId + "  dnd.nodeLastEnter=" + 
   //                  (dnd.nodeLastEnter? dnd.nodeLastEnter.attr("id") : "null") + " lastLeft=" +
   //                  (dnd.nodeLastLeft? dnd.nodeLastLeft.attr("id") : "null") +
   //                  " vars.o=" + vars.o + " vars.r=" + vars.r + " dnd.foreign=" + dnd.foreign, 1) ;
   // -- debugging --

   if (vars.r)           // sanity check - ensure we have a ref node drop position
   {
     if (dnd.dropDropCallback)
     {
       ret = dnd.dropDropCallback(e, {"reference" : dnd.bInternalNode? null : vars.r,
                                      "position"  : dnd.bInternalNode? "first" : vars.last_pos,
                                      "reorder"   : ((! dnd.foreign) && dnd.reorder)
                                     });
     }
     else
     {
       // No callback, only permitted if pure reorder on same tree.
       if (dnd.foreign || (! dnd.reorder))
       {
         veto = true ;             // can't allow - shouldn't be able to get here
       }                           // because dragEnter should have rejected operation.
     }

     // If reorder, handle itProceed with reorder if not vetoed
     if ( (! veto) && ((!ret) || e.isDefaultPrevented()))
     {
       this._dndFinishReorder(e) ;   // Handle same tree node movement
     }
   }         // end if (vars.r)

   // Since the dragend event goes to the source component, which could
   // be foreign, we must hide the drop ui for this component too.
   if (dnd.foreign)
   {
     this._dndClean() ;
   }

   if (! veto)
   {
     e.preventDefault() ;
     e.stopPropagation() ;
   }

   dnd.nodeLastLeft = null ;

   return ret ;
};

/**
  *  (dragend event handler).
  *  Drag operation has ended with/without a drop (dragend event handler).
  *  @private
  */ 
oj.TreeDndContext.prototype._dragEnd =  function(e)
{
  var dnd  = this._dnd ;
  var vars = dnd.vars ;
  var ret ; 

  //----debugging
  //this._trace(this._elemId + " _dragEnd(" + ( e?  $(e.target).closest("li").attr("id") : "") + ") - dragged=" + 
  //            dnd.nodeDraggedId + "  lastEnter=" + 
  //            (dnd.nodeLastEnter? dnd.nodeLastEnter.attr("id") : "null") + " lastLeft=" +
  //            (dnd.nodeLastLeft? dnd.nodeLastLeft.attr("id") : "null") +
  //            " vars.o=" + vars.o + " vars.r=" + vars.r + " dnd.foreign=" + dnd.foreign +
  //            " reorder=" + (dnd.reorder && vars.o) + dropEffect=" + 
  //            ((e && e.dataTransfer)? e.dataTransfer.dropEffect : "(event=null)"), 1) ;
  //----debugging

  if (dnd.dragEndCallback)                 // check if app callback in drag options?
  {
    dnd.dragEndCallback(e, {
                              "reorder"  : (dnd.reorder && vars.o === false)
                           }) ;
  }

  this._dndStopDrag() ;      // dragend completion
  this._dndClean() ;         // clean-up/housekeeping
};


/**
  *  Add a dummy node so that the tree is not empty, and there is a target that can
  *  be used as a reference node to be dragged to.
  *  @private
  */
oj.TreeDndContext.prototype._addInternalNode =  function()
{
  if (! this._internalNode)
  {
    this._internalNode = $("<li class='" + TreeUtils._OJ_TEMPNODE + " oj-tree-node oj-tree-leaf oj-valid-drop' id='" + TreeUtils._OJ_TEMPNODE + "'><ins class='oj-tree-icon'></ins><a href='#'><ins class='oj-tree-icon'></ins><span class='oj-tree-title'></span></a></li>");
  }
  this.component._$container_ul.append(this._internalNode) ;       // @HTMLUpdateOK  (no user supplied text involved)
};


/**
  *  (dragleave event handler).
  *  Mouse has moved away from a node while dragging (or esc was used
  *  to cancel the drag.
  *  @private
  */
oj.TreeDndContext.prototype._dragLeave =  function(e)
{
   var dnd    = this._dnd ;
   var vars   = dnd.vars ;
   var $targ  = $(e.target) ;
   var $node  = $targ.closest("li") ;
   var bNode  = ($node && ($node.length > 0) && $node.hasClass("oj-tree-node")) ;
   var elemId = $node.attr("id") ;
   var dt     = e.originalEvent.dataTransfer ;
   var cons   = oj.TreeDndContext ;

   // this._trace(this._elemId + " _dragLeave(" + e.target.tagName + ", " + (bNode? $node.attr("id"): "nonode") +
   //                 ") - dragged=" + dnd.nodeDraggedId + "  nodeLastEnter=" + 
   //                 (dnd.nodeLastEnter? dnd.nodeLastEnter.attr("id") : "null") + " lastLeft=" +
   //                 (dnd.nodeLastLeft? dnd.nodeLastLeft.attr("id") : "null") +
   //                 " vars.o=" + vars.o + " vars.r=" + vars.r + " dnd.foreign=" + dnd.foreign +
   //                 " dropEffect=" + e.dataTransfer.dropEffect, 1) ;

   if (dnd.dropDragLeaveCallback)
   {
     dnd.dropDragLeaveCallback(e, {"item" : dnd.bInternalNode? null : $node}) ;
   }

   // Clean up if leaving the tree, or a non-node, or drag canceled
   if ($targ.hasClass("oj-tree")  || (!bNode) || (dt.dropEffect== "none"))
   {
     this._dndLeave(e) ;            // clean up and hide marker(s)
   }
   else
   {
     dnd.nodeLastLeft = $node ;
   }

   if (vars.r)
   {
     this._clearDropClasses(vars.r) ;
   }

};


/**
  *  Handle node entry - inspect position, and prepare move block.
  *  @return {boolean}  true if placement is suitable, else false.
  *  @private
  */
oj.TreeDndContext.prototype._dndEnter = function (obj)
{
   var dnd  = this._dnd ;
   var vars = dnd.vars ;
   var cons = oj.TreeDndContext ;
// var s    = this.options["dnd"] ;
   var ret ;

   //this._trace("_dndEnter vars.o = " + vars.o + " vars.r = " +  vars.r) ;

   dnd.prepared = false;                       // no prepMoveBlk object yet
   ret = this._dndPrepare(); 

// var ms  = s["openTimeout"] ;     // not publicly exposed
   var ms  = dnd.openTimeoutMs ;    // open timeout value (ms)
   if (ms)
   { 
     if (dnd.openTimer_Id)
     {
       clearTimeout(dnd.openTimerId);
     }
     if (vars.r && vars.r.length && vars.r.hasClass(TreeUtils._OJ_COLLAPSED))
     { 
       // if the node is closed - open it, then recalculate
       dnd.openTimerId = setTimeout(this._dndOpen.bind(this), ms);     // @HTMLUpdateOK
     }
   }

   var t = (typeof ret) ;
   ret = (t == "boolean" && ret) || (t == "string") ;
   return ret ;
};



/**
  *  Expand the target node hovered over while dragging so the user can
  *  drop inside a folder.
  *  @private
  */
oj.TreeDndContext.prototype._dndOpen = function ()
{
   var  dnd =  this._dnd ;
   var  vars = dnd.vars ;

   dnd.openTimerId = false;
   this.component._expand(vars.r, this._dndPrepare.bind(this), true);
};


/**
  *  Drop is complete - do the node creation/insertion/movement if this is a reorder.
  *  @private
  */
oj.TreeDndContext.prototype._dndFinishReorder = function (e)
 {
    var dnd  = this._dnd ;
    var vars = dnd.vars ;

    // For a pure reorder on the same tree, we can just move the nodes.
    if (dnd.dragStarted && dnd.reorder && (typeof vars.o == 'object') && vars.o.length > 0)
    {
      this.component._moveNode(vars.o, vars.r, vars.last_pos, oj.DomUtils.isMetaKeyPressed(e));

      // Don't want dragEnd() to do any node(s) removal since they have just been
      // moved  by tree
      vars.o = false ;
   }
};


/**
  *  Apply selection to node and its children if state was set in __getJson/_parseJson
  *  after _createNode().
  *  @private
  */
oj.TreeDndContext.prototype._dndFinishSelection = function(node)
{
   if (node.hasClass(TreeUtils._OJ_SELECTED))     // is node tagged from the json?
   {
     node.removeClass(TreeUtils._OJ_SELECTED) ;   // remove selected tag
     this.component._select(node, true) ;         // and select it fully.
   }

   var children = this.component.getChildren(node) ;
   var len      = (children && children.length)? children.length : 0 ; 
   var i ;

   for (i = 0; i < len; i++)
   {
     this._dndFinishSelection($(children[i])) ;
   }
};


/**
  *  Decide on the placement relative to the reference node.
  *  Sets this._dnd.place to "after" , "before", or  "inside".
  *  @private
  */
oj.TreeDndContext.prototype._dndPlacement =  function (e, $node)
{
   var dnd   = this._dnd ;

   if (! $node) {
     return ;
   }

   if (! $node.hasClass(TreeUtils._OJ_TEMPNODE)) {
     e         = e.originalEvent ; 
     dnd.off   = $node.offset() ;                // get node coords rel to document
     dnd.place = (e.pageY - (dnd.off.top || 0)) ;
     dnd.place = (dnd.place >= this.component._getNodeHeight())?  "after"  :
                 (dnd.place < 0)?  "before"  : "inside" ;

     // If placement is after an open folder, treat placement as before
     // the first child node
     if (dnd.place === "after") {
       if ($node.hasClass(TreeUtils._OJ_EXPANDED)) {
         dnd.place = "first" ;
       }
     }
   }
   else {
     dnd.place = 'after' ;    // reference node is the dummy placed in an empty tree
   }

};


/**
  *  Mouse is down and we are dragging 
  *  @private 
  */
/*      Currently not used.
oj.TreeDndContext.prototype._dndDrag  =  function (e)
{ 
   var dnd  = this._dnd ;
   var vars = dnd.vars ;

   e    = e.originalEvent ;

// if (Math.abs(e.pageX - ctl.init_x) > 5 || Math.abs(e.pageY - ctl.init_y) > 5)  { 

   // maybe use a scrolling parent element instead of document?
   if (e.type === "dragover")
   {
      // thought of adding scroll in order to move the helper, but mouse position is n/a
      var d = $(document),
          t = d.scrollTop(),
          l = d.scrollLeft();

      if (e.pageY - t < 20)
      { 
        if (vars.sti && vars.dir1 === "down")
        {
//        clearInterval(vars.sti);
          vars.sti = undefined;
        }
        if (! vars.sti)
        {
          vars.dir1 = "up";
//        vars.sti = setInterval(function ()
//                   {
//                     $(document).scrollTop( $(document).scrollTop() - ctl.scroll_spd );
//                   }, 150);
        }
      }
      else
      { 
         if (vars.sti && vars.dir1 === "up")
         {
//         clearInterval(vars.sti);
           vars.sti = undefined;
         }
      }

      if ($(window).height() - (e.pageY - t) < 20)
      {
        if (vars.sti && vars.dir1 === "up")
        {
//        clearInterval(vars.sti);
          vars.sti = undefined;
        }
        if (! vars.sti)
        {
          vars.dir1 = "down";
//        vars.sti = setInterval(function ()
//                  {
//                    $(document).scrollTop( Number($(document).scrollTop()) + ctl.scroll_spd );
//                    // above Number() is for the closure compiler!!
//                  }, 150);
        }
      }
      else
      { 
         if (vars.sti && vars.dir1 === "down")
         {
//         clearInterval(vars.sti);
           vars.sti = undefined;
         }
      }

      if (e.pageX - l < 20)
      {
        if (vars.sli && vars.dir2 === "right")
        {
//         clearInterval(vars.sli);
           vars.sli = undefined;
        }
        if (! vars.sli)
        {
          vars.dir2 = "left";
//        vars.sli = setInterval(function ()
//               {
//                 $(document).scrollLeft( $(document).scrollLeft() - ctl.scroll_spd );
//               }, 150);
        }
      }
      else
      { 
        if (vars.sli && vars.dir2 === "left")
        {
//        clearInterval(vars.sli);
          vars.sli = undefined;
        }
      }

      if ($(window).width() - (e.pageX - l) < 20)
      {
        if (vars.sli && vars.dir2 === "left")
        {
//        clearInterval(vars.sli);
          vars.sli = undefined;
        }
        if (! vars.sli)
        {
          vars.dir2 = "right";
//        vars.sli = setInterval(function ()
//                 {
///                  Number() is for the closure compiler!!
//                   $(document).scrollLeft(Number($(document).scrollLeft()) + ctl.scroll_spd);
//                 }, 150);
        }
      }
      else
      { 
        if (vars.sli && vars.dir2 === "right")
        {
//        clearInterval(vars.sli);
          vars.sli = undefined;
        }
      }

   }
};
*/

/**
  *   The mouse has left a node while in drag mode. Clear the markers.
  *   @private
  */
oj.TreeDndContext.prototype._dndLeave = function (e)
{
   var dnd  = this._dnd ;
   var vars = dnd.vars ;

   dnd.after   = false;
   dnd.before  = false;
   dnd.inside  = false;
   dnd.lastNodeEnter = null ;

   this._dndHideMarker() ;
   if (vars.r && vars.r[0] === e.target.parentNode)
   {
     if (dnd.openTimerId)
     {
       clearTimeout(dnd.openTimerId);
       dnd.openTimerId = false;
     }
   }
};



/**
  *  Cleanup after DnD drag/drop cycle.
  *  @private
  */
oj.TreeDndContext.prototype._dndClean = function()
{
  this._dndHideMarker() ;

  var dnd  = this._dnd ;
  var vars = dnd.vars;

  vars.o        = false ;
  vars.r        = false ;

  if (dnd.openTimerId)
  {
    clearTimeout(dnd.openTimerId);
  }
//if (dnd.i1)  {
//  clearInterval(dnd.i1);
//}
//if (dnd.i2)  {
//  clearInterval(dnd.i2);
//}

  dnd.nodeDraggedId  = null ;
  dnd.foreign        = false ;
  dnd.nodeLastEnter  = null ;
  dnd.nodeLastLeft   = null ;
  dnd.after          = false;
  dnd.before         = false;
  dnd.inside         = false;
  dnd.off            = false;
  dnd.prepared       = false;
  dnd.openTimerId    = false;     
//dnd.i1             = false;
//dnd.i2             = false;
  dnd.foreign        = false;
  dnd.bInternalNode  = false ;
};

/**
  *  Cleanout all DnD settings.
  *  @private
  */
oj.TreeDndContext.prototype._dndCleanAll = function()
{
   var  dnd  = this._dnd ;
   var  $div ;

   //  Clean out the old dnd settings
   $div = $(this.component._getContainer()[0]) ;
   $div.off("dragstart", dnd.fnDragStart) ;
   $div.off("dragend",   dnd.fnDragEnd) ;
   $div.off("drag",      dnd.fnDrag) ;
   $div.off("dragenter", dnd.fnDragEnter) ;
   $div.off("dragover",  dnd.fnDragOver) ;
   $div.off("dragleave", dnd.fnDragLeave) ;
   $div.off("drop",      dnd.fnDragDrop) ;

   this._dndClean() ;
   this._reset() ;
};


/**
  *  Shows the placement marker/marker line. If internal reorder creates a prepMoveBlk object.
  *  Returns placement - "above", "inside", "below", or false
  *  @private
  */
oj.TreeDndContext.prototype._dndPrepare = function ()
{
   var  dnd         = this._dnd ;
   var  vars        = dnd.vars ;
   var  tree        = this.component ;
   var  foreignObj  = false ; 

   if ((! vars.r) || (! vars.r.length))
   {
     return false ;
   }

   // See if from foreign object
   if (typeof vars.o == 'number' && (vars.o === 1 || vars.o === 2))
   {
     foreignObj = true ;     // (1 = foreign tree node, 2 = non-tree node)
   }
   else if (! vars.o)
   {
     return false ;         // no object, so can't handle.
   }

   dnd.off = vars.r.offset();   // get pos of reference node relative to doc

   if (this.component._isRtl)
   {
     // arrow icon is on the right of the marker line
     dnd.off.right  = dnd.off.left + vars.r.width() - vars.r.find(">a").width() -30 ;
   }
   dnd.targ_ml_width = vars.r.find("a").width();     // get the width for the insert line

   if (foreignObj)
   {
     // This is a foreign object (and we are not directly rearranging the
     // Tree DOM as in a reorder, so no internal prepMoveBlk is required).

     dnd.before = (dnd.place == 'before') ;
     dnd.after  = (dnd.place == 'after') ;
     dnd.inside = (dnd.place == 'inside') ;
   }
   else
   {
     // Re-order - set the dnd before/after/inside booleans
     tree._prepare_move(vars.o, vars.r, "before");       // create a prepMoveBlk object
     dnd.before = tree._checkMove();
     tree._prepare_move(vars.o, vars.r, "after");
     dnd.after  = tree._checkMove();

     if (this.component._is_loaded(vars.r))
     {
       tree._prepare_move(vars.o, vars.r, "inside");
       dnd.inside = tree._checkMove();
     }
     else
     {
       dnd.inside = false ;
     }
   }

   dnd.prepared = true;              // prepare has completed
   return this._dndShowMarker();     // returns "after", "before", "inside", or false
};


/**
  *  Positions and shows (or hides) the drag insert position marker and line,
  *  and updates the drag and drop targets with the can/can't drop classes.
  *  Returns "above", "inside", "below" or false
  *  @private
  */
oj.TreeDndContext.prototype._dndShowMarker = function()
{
   var  dnd      = this._dnd ;
   var  isParent = false ;
   var  cons     = oj.TreeDndContext ;
   var  vars     = dnd.vars;
   var  rtl      = this.component._isRtl ;
   var  mLeft ;                     // pos of marker rel to document
   var  mlLeft ;                    // position of marker line rel to doc
   var  lineTop ;                   // top pos of marker line rel to doc
   var  nodeHeight = this.component._getNodeHeight()
   var  r          = false ;
   var  pos ;

   //this._trace("dndShowMarker entry  vars.o = " + vars.o + "  vars.r = " + vars.r + " dnd.place = " + dnd.place, 1) ;
   //this._trace("dnd['after'] = " + dnd['after'] + " dnd['inside'] = " + dnd['inside'] + " dnd['before'] = " + dnd['before']) ;

   if (! dnd.prepared)
   {
     this._dndHideMarker() ;
     return r ;
   }

   // Get placement relative to target node
   if ((dnd.place == "after"  && dnd.after)  ||
       (dnd.place == "inside" && dnd.inside) ||
       (dnd.place == "before" && dnd.before) ||
       (dnd.place == "first"))
   {
     r = dnd.place ;
   }

   if (r === "first")                    // if to be placed as the first child of
   {                                     // a parent, will actually treat as above
      var fc = vars.r.find("li:eq(0)");  // the first child, as far as the marker
      dnd.off = fc.offset() ;            // is concerned.
   }


   pos       = rtl? (dnd.off.right - 18) : (dnd.off.left + 5);
   isParent  = ! this.component.isLeaf(vars.r);
   mLeft     = rtl? (pos + dnd.targ_ml_width + 18) : pos ;  // pos of marker rel to doc
   mlLeft    = rtl? (mLeft - dnd.ml_width) : (pos + 8)      // pos of marker line rel to doc
   
   lineTop   = dnd.off.top - 3 ;       // set line top a small dist above the node for clarity

   switch (r)
   {
      case "before":
      case "first" :
          this._moveMarkerUI(vars.m, {"left" : mLeft, "top"  : (lineTop -7)}) ;
          if (vars.ml)
          {
            this._moveMarkerUI(vars.ml, {"left" : mlLeft, "top"  : lineTop}) ;
          }
          break;

      case "after":
          lineTop +=  nodeHeight ;
          this._moveMarkerUI(vars.m, {"left" : mLeft, "top"  : (lineTop - 2)}) ;
          if (vars.ml)
          {
            this._moveMarkerUI(vars.ml, {"left" : mlLeft, "top" : (lineTop + 6)}) ;
          }
          break;

      case "inside":       // marker only, no line
          this._moveMarkerUI(vars.m, {"left" : mLeft + (rtl? (isParent? -4 : 0) : 4),
                                      "top" : (dnd.off.top + nodeHeight/2 - 8)}) ; 
          if (vars.ml)
          {
            vars.ml.hide();
          }
          break;

      default:
          this._dndHideMarker() ;
          break;
   }

   if (r)
   {
     vars.r.removeClass(cons._OJ_INVALID_DROP)
           .addClass(cons._OJ_VALID_DROP) ;
   }
   else
   {
     vars.r.removeClass(cons._OJ_VALID_DROP)
           .removeClass(cons._OJ_VALID_DROP) ;
   }

   vars.last_pos = r;
   return  r;
};


/**
  *  Hide the drag/drop insertion marker and line
  *  @private
  */
oj.TreeDndContext.prototype._dndHideMarker =  function ()
{
   var vars  = this._dnd.vars ;
   var cons  = oj.TreeDndContext ;
   
   if (vars)
   {
     if (vars.m)
     {
       vars.m.hide() ;
     }
     if (vars.ml)
     {
       vars.ml.hide() ;
     }
     if (vars.r)
     {
       this._clearDropClasses(vars.r) ;
     }
   }
};

/**
  *  Move the specified marker UI element.
  *  @private
  */
oj.TreeDndContext.prototype._moveMarkerUI =  function ($obj, styles)
{
   var vars  = this._dnd.vars ;

   styles["left"] = styles["left"] + "px" ;
   styles["top"]  = styles["top"] + "px" ;

   $obj.css(styles).show() ;
} ;


/**
  *  The drag operation has started, initial prepare.
  *  @private
  */
oj.TreeDndContext.prototype._dndStartDrag  =  function ($nodes, $node, e)
{
   var  dnd  = this._dnd ;
   var  vars = dnd.vars ;
   var  cons = oj.TreeDndContext ;
   var  $cnt ;


   vars.o  = $nodes ;                            // internal note of the node(s) dragged
   vars.o.addClass(cons._OJ_DRAG) ;
   dnd.nodeDraggedId = $node.attr("id") ; ;

   try
   {
       e.currentTarget.unselectable = "on";
       e.currentTarget.onselectstart = function()  { return false; };
       if (e.currentTarget.style)
       {
         e.currentTarget.style.MozUserSelect = "none";
       }
   } catch(err) {}

   if (this.component._isTheme())
   {
     if (vars.m)
     {
       vars.m.addClass("oj-tree-" + this.component._getTheme());
     }
   }

   $cnt     = this.component._getContainer();
   dnd.cof  = $cnt.offset();
   dnd.cw   = parseInt($cnt.width(),10);
   dnd.ch   = parseInt($cnt.height(),10);

   dnd.dragStarted   = true ;
};


/**
  *  Cleanup after end of drag.
  *  @private
  */
oj.TreeDndContext.prototype._dndStopDrag  =  function()
{
   var dnd  = this._dnd ;
   var vars = dnd.vars ;
   var cons = oj.TreeDndContext ;

//      if (vars.sli)  {
//        clearInterval(vars.sli);
//      }
//      if (vars.sti)   {
//        clearInterval(vars.sti);
//      }
//

   if (vars.r)
   {
     this._clearDropClasses(vars.r) ;
   }
   if ((typeof vars.o != 'number') && vars.o) 
   {
     vars.o.removeClass(cons._OJ_DRAG) ;
   }

   // If we created a drag image, remove it from the DOM.
   if (this._dragImg)
   {
     document.body.removeChild(this._dragImg) ;
     this._dragImg = null ;
   }

   dnd.dragStarted  = false ;
};


/**
  *  Set the app defined data types to the dragged node data 
  *  @private
  */
oj.TreeDndContext.prototype._setDataTypes = function(dt, $nodes)
{
   var  dnd  = this._dnd ;
   var  cons = oj.TreeDndContext ;
   var  adata ;
   var  data ;
   var  i ;

   adata = this.component.__getJson($nodes) ;
   if (adata)
   {
     data = JSON.stringify(adata) ;
   }
   if (dnd.dragDataTypes)
   { 
     for (i = 0; i < dnd.dragDataTypes.length; i++)
     {
       dt.setData(dnd.dragDataTypes[i], data) ;
     }
   }
   dt.setData(cons._DND_INTERNAL_DT, "0") ;           // internal id tree data
   if (dnd.pureReorder)
   {
     dt.setData(cons._DND_INTERNAL_DT_REORDER, "0") ; // internal id for reorder
   }
};


/**
  *  Create the drag image and set in the data transfer object
  *  @private
  */
oj.TreeDndContext.prototype._setDragImage = function(dt, $nodes)
{
   var $dragImg, ul, elem, el,
       len = $nodes.length,
       i ;
   
   // If only one node being dragged, use the default drag image
   if (len === 1) 
   {
     this._dragImg = null ;
     return ;
   }

   $dragImg = $("<div></div>");
   ul = $("<ul style='padding:0px;margin:0;'></ul>");
   $dragImg.append(ul);       //@HTMLUpdateOK; no user input involved and 
                              // only creating a drag image.

   for (i = 0; i < len; i++)
   {
     el = $nodes[i].cloneNode(true);
     el.style.marginLeft  = 0;
     el.style.paddingLeft = 0;
     ul.append(el);            // @HTMLUpdateOK; appended elems have already
                               // been sanitized before being added to the DOM,
                               // and are only referenced here to create an image
                               // for dragging.
   }

   elem = $dragImg[0] ;
   document.body.appendChild(elem);    // @HTMLUpdateOK; pre-sanitized elems used to create a drag image
   elem.style.position = "absolute";
   elem.style.top      = "-" + ($dragImg.height() * 2) + "px";
   elem.style.right    = "0";
   this._dragImg       = elem;     // note for removal in _dragend()
   dt.setDragImage(elem, 0, 0);
};


/**
  *  Check if the specified type is in the drop event's dataTransfer
  *  object.
  *  @return {boolean}  true if the specified type is in the drop dataTransfer
  *  object, else false.
  *  @private
  */
oj.TreeDndContext.prototype._isDtType =  function(dt, type)
{
   var aTypes = dt.types ;
   var len    = aTypes.length ;
   var i ;

   for (i = 0; i < len; i++)
   {
     if (type == aTypes[i])
     {
       return true ;
     }
   }
   return false ;
},

/**
  *  Check if there is a list of dataTypes in the dnd drag/drop options
  *  (or the list is not defined), at least one of which matches the
  *  dataTypes defined in the dataTransfer object.
  *  @return {boolean}   true if there is a list of dataTypes in the dnd
  *  drag/drop options (or the list is not defined), at least one of which
  *  matches the dataTypes defined in the dataTransfer object, else false.
  *  @private
  */
oj.TreeDndContext.prototype._isDtAcceptDataTypes =  function(dt)
{
   var dnd   =  this._dnd ;
   var comps =  dnd.dropDataTypes ;
   var ret   = false ;
   var i ;

   if (comps)
   {
     ret = ! (comps.length > 0) ;
     for (i = 0; i < comps.length; i++)
     {
        if (this._isDtType(dt, comps[i]))
        {
          ret = true ;
          break ;
        }
     }
   }
   else {
     ret = true ;
   }

   return ret;
};


/**
  *  Check if the user app has added at least one dataType to the dataTransfer
  *  object.
  *  @return {boolean}   true if the if the user app has added at least one
  *  dataType, else false.
  *  @private
  */
oj.TreeDndContext.prototype._checkAddedDataTypes =  function(dt)
{
   var aTypes = dt.types ;
   var len    = ((aTypes && aTypes.length)? aTypes.length : 0) ;
   var cons   = oj.TreeDndContext ;
   var ret    = false ;
   var type ;
   var i ;

   for (i = 0; i < len; i++)
   {
     type = aTypes[i] ;
     if ((type != cons._DND_INTERNAL_DT) && (type != cons._DND_INTERNAL_DT_REORDER))
     {
       ret = true ;
       break ;
     }
   }
   return ret ;
};


/**
  *  Clear DnD drop classes
  *  @private
  */
oj.TreeDndContext.prototype._clearDropClasses = function($var)
{
   var cons  = oj.TreeDndContext ;

   if ($var) {
     $var.removeClass(cons._OJ_VALID_DROP)
         .removeClass(cons._OJ_INVALID_DROP)
         .removeClass(cons._OJ_ACTIVE_DROP) ;
   }
} ;


/**
  *  Return an array of node elements 
  *  @private
  */
oj.TreeDndContext.prototype._getDraggedItemArray = function($nodes)
{
   var a = [] ;

   if ($nodes)
   {
      $nodes.each(function ()
                  {
                    a.push(this) ;
                  });
   }

   return a ;
} ;


/**
  *  Reset this context's data
  *  @private
  */
oj.TreeDndContext.prototype._reset = function()
{
  var _dnd  = {} ;
  this._dnd = _dnd ;

  _dnd.reorder      = false ;
  _dnd.pureReorder  = false ;        // true if reorder and not a drag source
  _dnd.foreign      = false ;        // dragged obj not from this tree

  _dnd.dragStarted    = false ;      // startDrag received on this tree
  _dnd.nodeDraggedId  = null ;       // id of node being dragged
  _dnd.nodeLastEnter  = null ;       // node last entered
  _dnd.fnDragStart    = null ;       // event listeners bound to 'this'
  _dnd.fnDragEnd      = null ;       //   ..     ..
  _dnd.fnDrag         = null ;       //   ..     ..
  _dnd.fnDragEnter    = null ;       //   ..     ..
  _dnd.fnDragOver     = null ;       //   ..     ..
  _dnd.fnDragLeave    = null ;       //   ..     ..
  _dnd.fnDragDrop     = null ;       //   ..     ..

  // App callbacks and flags
  _dnd.dragFromEnabled       =  false;
  _dnd.dropToEnabled         =  false;
  _dnd.dragStartCallback     =  null ; //  User App callback
  _dnd.dragDragCallback      =  null ; //   ..   ..    ..
  _dnd.dragDragEndCallback   =  null ; //   ..   ..    ..
  _dnd.dropDragEnterCallback =  null ; //   ..   ..    ..
  _dnd.dropDragOverCallback  =  null ; //   ..   ..    ..
  _dnd.dropDragLeaveCallback =  null ; //   ..   ..    ..
  _dnd.dragDropCallback      =  null ; //   ..   ..    ..
  _dnd.dropDataTypes         =  null ;                                             

  _dnd.after      = false ;
  _dnd.inside     = false ;
  _dnd.before     = false ;
  _dnd.off        = false ;
  _dnd.prepared   = false ;       // true when a prepMoveBlk object has been created
  _dnd.openTimerId= false ;       // open folder timeout Id from setTimeout() when dragged over
  _dnd.cof        = false ;
  _dnd.cw         = false ;
  _dnd.ch         = false ;
// _dnd.i1         = false ;      // not currently used
// _dnd.i2         = false ;      // not currently used
  _dnd.ml_width   = 100 ;         // marker line width - updated in _initDnd() from css
  _dnd.targ_ml_width = 100 ;      // marker line width of target node
  _dnd.openTimeoutMs = 500 ;      // open folder timeout when dragged over
// _dnd.defaults = {              // not yet exposed
//                    "openTimeout"  : 500
//                    "checkTimeout" : 100 };
  _dnd.vars          = {}          // drag/drop block of vars 
  _dnd.vars.o        = false ;     // node(s) being dragged, or 
                                   // 1 if foreign obj, or 2 if foreign node(s)
  _dnd.vars.r        = false ;     // reference node 
  _dnd.vars.m        = false ;     // marker pointer div
  _dnd.vars.ml       = false ;     // marker line div
// _dnd.vars.sli      = undefined ;
// _dnd.vars.sti      = undefined ;
// _dnd.vars.dir1     = false ;
// _dnd.vars.dir2     = false ;
// _dnd.vars.last_pos = false ;
};


/**
  *   Convenience function for tracing
  *   @private
  */
oj.TreeDndContext.prototype._trace = function(s)
{
   this.component._trace(s) ;
};




//------------------------------//
//  ojTree DnD CSS class names  //
//------------------------------//
/**
 * @private
 * @const
 * @type {string}
 */
 oj.TreeDndContext._OJ_DRAGGABLE  = "oj-draggable" ;      // node is draggable
/**
 * @private
 * @const
 * @type {string}
 */
 oj.TreeDndContext._OJ_DRAG       = "oj-drag" ;           // node that is being dragged
/**
 * @private
 * @const
 * @type {string}
 */
 oj.TreeDndContext._OJ_ACTIVE_DROP = "oj-active-drop" ;   // over node in process of being dropped on
/**
 * @private
 * @const
 * @type {string}
 */
 oj.TreeDndContext._OJ_VALID_DROP  = "oj-valid-drop" ;    // over node that could be dropped on
/**
 * @private
 * @const
 * @type {string}
  */
 oj.TreeDndContext._OJ_INVALID_DROP = "oj-invalid-drop" ; // over node that can't be dropped on
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._OJ_DROP  = "oj-drop" ;                // nodes that are valid drop targets
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._OJT_DROP_MARKER       = "oj-tree-drop-marker" ;       // insertion point marker
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._OJT_DROP_MARKER_LINE  = "oj-tree-drop-marker-line" ;  // insertion point marker line



//--------------------------//
//    ojTree DnD Options    //
//--------------------------//
/**
  * @private
  * @const
  * @type {string}
 */
oj.TreeDndContext._DND_REORDER = "reorder" ;
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DRAG = "drag" ;
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DROP = "drop" ;
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_NODE  = "node" ;
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DATATYPES  = "dataTypes" ;
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DRAGSTART_CB  = "dragStart" ;      // callbacks
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DRAGDRAG_CB   = "drag" ;           //   . .
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DRAGEND_CB    = "dragEnd" ;        //   . .
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DROPDRAGENTER_CB  = "dragEnter" ;  //   . .
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_DROPDRAGOVER_CB  = "dragOver" ;  //   . .
/**
  * @private
  * @const
  * @type {string}
 */
oj.TreeDndContext._DND_DROPDRAGLEAVE_CB = "dragLeave" ; //   . .
/**
  * @private
  * @const
  * @type {string}
 */
oj.TreeDndContext._DND_DROPDROP_CB  = "drop" ;          //   . .
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_INTERNAL_DT  ="_ojtree" ;    // internal data type for tree data
/**
 * @private
 * @const
 * @type {string}
 */
oj.TreeDndContext._DND_INTERNAL_DT_REORDER ="_ojtreereorder" ;   // internal data type for reorder

/**
  * Copyright (c) 2015, Oracle and/or its affiliates.
  * All rights reserved.
  */

/**
  *  Common tree/node state
  *  @ignore
  *  @export
  *  @classdesc  Common Tree/node state
  *  @protected
  *  @constructor
  */
function TreeUtils()
{
};


// css class names

TreeUtils._OJ_EXPANDED    = "oj-expanded" ;
TreeUtils._OJ_COLLAPSED   = "oj-collapsed" ;
TreeUtils._OJ_HOVER       = "oj-hover" ;
TreeUtils._OJ_SELECTED    = "oj-selected" ;
TreeUtils._OJ_DISABLED    = "oj-disabled" ;
TreeUtils._OJ_DEFAULT     = "oj-default" ;
TreeUtils._OJ_TEMPNODE    = "oj-treenode-temp" ;
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/* This component is based on original code from:
   jsTree 1.0-rc3   http://jstree.com/
   "Copyright (c) 2010 Ivan Bozhanov (vakata.com)
    Licensed same as jquery - under the terms of either the MIT License or
    the GPL Version 2 License
    http://www.opensource.org/licenses/mit-license.php
    http://www.gnu.org/licenses/gpl.html"
*/


/*---------------------------------------------------------
   ojTree     Displays a Hierarchical Tree of Nodes
   Depends:   jquery.ui.core.js
              jquery.ui.widget.js
----------------------------------------------------------*/

(function ()
{
  //  "use strict";   // added by build, not required.

  /*
  function debugObj(o)  {
    var s ;
    try { s = JSON.stringify(o) ; }
    catch (e) { s = "ERROR";}
    return s ;
  };
  */

  //  ojTree class names
  var   /** @const */   OJT_TREE          = "oj-tree",                 // a tree
        /** @const */   OJT_NODE          = "oj-tree-node",            // a tree node (parent or leaf)
        /** @const */   OJT_LEAF          = "oj-tree-leaf",            // leaf node
        /** @const */   OJT_ICON          = "oj-tree-icon",            // <ins> generic icon
        /** @const */   OJT_NICON         = "oj-tree-node-icon",       // <ins> node icon
        /** @const */   OJT_DISC          = "oj-tree-disclosure-icon", // <ins> disclosure icon
        /** @const */   OJT_LAST          = "oj-tree-last",            // last <li> in a <ul>
        /** @const */   OJT_LOADING       = "oj-tree-loading",
        /** @const */   OJT_LAST_SELECTED = "oj-tree-last-selected",
        /** @const */   OJT_TITLE         = "oj-tree-title",           // node text
        /** @const */   OJT_TYPE          = "oj-tree-type",            // node has a user-defined type
        /** @const */   OJT_INACTIVE      = "oj-tree-inactive" ;

  // ojTree disclosure class names
  var  /** @const */   OJ_DISC            = "oj-tree-icon oj-tree-disclosure-icon oj-component-icon oj-clickable-icon-nocontext oj-default";

  //  WAI-ARIA
  var  /** @const */   WA_ROLE              = "role",
       /** @const */   WA_TREE              = "tree",
       /** @const */   WA_TREEITEM          = "treeitem",
       /** @const */   WA_GROUP             = "group",
       /** @const */   WA_SELECTED          = "aria-selected",
       /** @const */   WA_EXPANDED          = "aria-expanded",
       /** @const */   WA_ACTIVEDESCENDANT  = "aria-activedescendant",
       /** @const */   WA_MULTISELECTABLE   = "aria-multiselectable",
       /** @const */   WA_LABELLEDBY        = "aria-labelledby";

  //  Data names
  var  /** @const */   OJT_CHILDREN         = "oj-tree-children";

  //  Data source in use
  var  /** @const */   DS_TREE        =  1,
       /** @const */   DS_COLLECTION  =  2,
       /** @const */   DS_JSON        =  3,
       /** @const */   DS_HTML        =  4,
       /** @const */   DS_NONE        =  0,
       /** @const */   DS_ERROR       = -1;

  // Context Menu item id's
  var  /** @const */  _arMenuCmdMap = { "cut"          : "ojtreecut",
                                        "copy"         : "ojtreecopy",
                                        "paste"        : "ojtreepaste",
                                        "paste-after"  : "ojtreepasteafter",
                                        "paste-before" : "ojtreepastebefore",
                                        "remove"       : "ojtreeremove",
                                        "rename"       : "ojtreerename"
                                      } ;

  // Context Menu translation keys
  var  /** @const */  _arMenuKeyMap = { "cut"          : "labelCut",
                                        "copy"         : "labelCopy",
                                        "paste"        : "labelPaste",
                                        "paste-after"  : "labelPasteAfter",
                                        "paste-before" : "labelPasteBefore",
                                        "remove"       : "labelRemove",
                                        "rename"       : "labelRename"
                                      } ;

  // Misc translation keys
  var  /** @const */  TRANSKEY_LOADING  = "stateLoading" ;
  var  /** @const */  TRANSKEY_NEWNODE  = "labelNewNode" ;
  var  /** @const */  TRANSKEY_MULTISEL = "labelMultiSelection" ;
  var  /** @const */  TRANSKEY_NODATA   = "labelNoData" ;

  var  /** @const */  USER_UL_ID_PREFIX = "oj-tree-existing-markup-" ;

  //  Attr's we ignore when creating a list of user attr's
  var /** @const */    ATTR_IGN          = ["id", "title", "class", "role", "draggable", "style"] ;



  var scrollbar_width, e1, e2;
  $(function() {
     if (/msie/.test(navigator.userAgent.toLowerCase())) {
       e1 = $('<textarea cols="10" rows="2"></textarea>').css({ position: 'absolute', top: -1000, left: 0 }).appendTo('body');  //@HTMLUpdateOK
       e2 = $('<textarea cols="10" rows="2" style="overflow: hidden;"></textarea>').css({ position: 'absolute', top: -1000, left: 0 }).appendTo('body'); //@HTMLUpdateOK
       scrollbar_width = e1.width() - e2.width();
       e1.add(e2).remove();
     }
     else {
       e1 = $('<div />').css({ width: 100, height: 100, overflow: 'auto', position: 'absolute', top: -1000, left: 0 })
                        .prependTo('body').append('<div />').find('div').css({ width: '100%', height: 200 });    //@HTMLUpdateOK
       scrollbar_width = 100 - e1.width();
       e1.parent().remove();
     }
  });


  /**
    * Stack of key handler objects for all trees
    * @private
    */
  var  _aKeyHandlerStack = [] ;

  /**
    * Add key combination handler to stack, and add listener
    * @private
    */
  function  _addKeyFilter(obj)
  {
    _aKeyHandlerStack.push(obj) ;
    $(obj._selector).keydown(_KeyFilterHandler) ;
  };

  /**
    * Handle keystroke from _KeyFilter and dispatch to the
    * relevent accessibility key combination handler.
    * @private
    */
  function  _KeyFilterHandler(e)
  {
    var s = "" ;

    if (oj.DomUtils.isMetaKeyPressed(e)) {
      s += "ctrl+" ;
    }
    else if (e.shiftKey) {
      s += "shift+" ;
    }
    var key = e.which ;
    switch(key) {
      case 32:                      // Space
                s += "space" ;
                break;
      case 37:                      // Left arrow
                s += "left" ;
                break;
      case 38:                      // Up arrow
                s += "up" ;
                break;
      case 39:                      // Right arrow
                s += "right" ;
                break;
      case 40:                      // Down arrow
                s += "down" ;
                break;
      case 46:                      // Delete
                s += "del" ;
                break;
      case 33:                      // Page Up
                s += "pgup" ;
                break;
      case 34:                      // Page Down
                s += "pgdn" ;
                break;
      case 35:                      // End
                s += "end" ;
                break;
      case 36:                      // Home
                s += "home" ;
                break;
      case 56:                      // asterisk
                s = "*" ;            // remove the shift required to get *
                break;
      case 113:                     // F2
      case 121:                     // F10
/*
      case 112:                     // F1
      case 114:                     // F3
      case 115:                     // F4
      case 116:                     // F5
      case 117:                     // F6
      case 118:                     // F7
      case 119:                     // F8
      case 120:                     // F9
      case 122:                     // F11
      case 123:                     // F12
*/
                s += ("f" + ('1' - (112 - key))) ;
                break;
    }
    if (s.length === 0 || s == "shift+" || s == "ctrl+") {
      return true ;
    }
    // If we can find a handler from a tree with focus, dispatch the
    // keystroke event to it.
    var retHandler = null ;
    $.each(_aKeyHandlerStack, function(i, obj) {
        if (obj._this._data.ui.focused) {         // does associated tree have focus
                                                  // yes
          obj._this._data.ui.touchEvent = false ; // ensure touchEvent flag is off
          if (s == "shift+f10") {                 // note the active node if shift+F10
            obj._this._data.menu.activeNode = obj._this._data.ui.hovered ;
            //  - if no hovered node (maybe a click and then mouse off the
            // node), use the last selected node.
            if (! obj._this._data.menu.activeNode) {                               // 
              obj._this._data.menu.activeNode = obj._this._data.ui.lastSelected ;  // 
            }
          }
          else  if (obj._handler[s]) {            // check if matching handler function
              e.preventDefault() ;
              retHandler =  obj._handler[s].call(obj._this, e) ;   // handle the keystroke(s)
          }
          return false ;                          // break out of $.each
        }
    });

    if (retHandler != null) {
      return retHandler ;
    }
  };

  /**
    * Remove key combination handler from stack and remove listener
    * @private
    */
  function  _removeKeyFilter(selector)
  {
    $.each(_aKeyHandlerStack, function(i) {
       if (_aKeyHandlerStack[i]._selector === selector) {
         $(selector).off("keydown") ;
         _aKeyHandlerStack[i] = null ;
         _aKeyHandlerStack.splice(i, 1) ;
         return false ;
       }
    }) ;
  };

  /**
    *  Returns true if a non-inherited object has no properties.
    *  @private
    */
  function  isPureObjEmpty(o)
  {
     var ret = true ;
     for (var prop in o) {
        if (o.hasOwnProperty(prop)) {
          ret = false ;
          break ;
        }
     }
     return ret ;
  };

  /**
    *  Creates and adds a style sheet.
    *  @param {Object=}  opts  The css data.
    *  @param {boolean=} bLink  True if opts.url is to be ignored.
    *  @private
    */
  function  _addSheet(opts, bLink)
  {
    var tmp    = false,
        isNew  = true;

    if (opts.str)  {
      if (opts.title)  {
        tmp = $("style[id='" + opts.title + "-stylesheet']")[0];
      }
      if (tmp)  {
        isNew = false;
      }
      else  {
        tmp = document.createElement("style");
        tmp.setAttribute('type',"text/css");
        if (opts.title) {
          tmp.setAttribute("id", opts.title + "-stylesheet");
        }
      }

      if (tmp.styleSheet)  {
        if (isNew)  {
          document.getElementsByTagName("head")[0].appendChild(tmp);  //@HTMLUpdateOK
          tmp.styleSheet.cssText = opts.str;
        }
        else  {
          tmp.styleSheet.cssText = tmp.styleSheet.cssText + " " + opts.str;
        }
      }
      else  {
        tmp.appendChild(document.createTextNode(opts.str));           //@HTMLUpdateOK
        document.getElementsByTagName("head")[0].appendChild(tmp);    //@HTMLUpdateOK
      }

      return tmp.sheet || tmp.styleSheet;
    }

/*  Not currently used
    if (opts.url) {
      if (bLink) {
        if (document.createStyleSheet)  {
          try  {
                 tmp = document.createStyleSheet(opts.url);    // IE
          }
          catch (e) { }

        }
        else  {
          tmp        = document.createElement('link');
          tmp.rel    = 'stylesheet';
          tmp.type   = 'text/css';
          tmp.media  = "all";
          tmp.href   = opts.url;

          document.getElementsByTagName("head")[0].appendChild(tmp);
          return tmp.styleSheet;
        }
      }
    }
*/
  };


  /**
    * @private
    */
  var _instance   = -1 ;      // current Tree instance id


/**
  * @ojcomponent oj.ojTree
  * @augments oj.baseComponent
  * @since 0.6
  * @ignore
  * @deprecated  4.2.0 Use oj.ojTreeView instead.
  *
  * @classdesc
  * <h3 id="treeOverview-section">
  *   JET Tree Component
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#treeOverview-section"></a>
  * </h3>
  *
  * The ojTree component allows a user to display the hierarchical relationship between the nodes of a tree.<p>
  * The tree contents can be specified in JSON format, or by prepopulating the tree's containing &lt;div&gt; with
  * HTML &lt;ul&gt; list markup.
  * </p></br>
  *
  * <h4 id="treeJSON-section"> JSON Node Format</h4>
  * </br>
  * Each node object typically has a <code class="prettyprint">title</code> and an
  * <code class="prettyprint">attr</code> property. Any node can be defined as a parent by supplying
  * a <code class="prettyprint">children</code> property, which is an array of more node definitions.
  * (Note that if a node has a <code class="prettyprint">children</code> property defined, but no children
  * are actually specified, then ojTree will perform lazy-loading by requesting child node data only
  * when a node is expanded for the first time - refer to <code class="prettyprint">option</code> property
  * <code class="prettyprint">data</code>.
  * <p>Example: Basic JSON Tree definition
  * <pre class="prettyprint">
  * <code>
  *[
  *   {
  *     "title": "Home",
  *     "attr": {"id": "home"},
  *   },
  *   {
  *     "title": "News",
  *     "attr": {"id": "news"}
  *   },
  *   {
  *      "title": "Blogs",
  *      "attr": {"id": "blogs"},
  *      "children": [ {
  *                       "title": "Today",
  *                       "attr": {"id": "today"}
  *                    },
  *                    {
  *                       "title": "Yesterday",
  *                       "attr": {"id": "yesterday"}
  *                    }
  *                  ]
  *   }
  *]
  *</code></pre>
  *</p></br>
  * Whatever attributes are defined for the <code class="prettyprint">attr</code> property are transferred
  * to the associated DOM &lt;li&gt; element. A <code class="prettyprint">metadata</code> attribute can also be
  * defined for arbitrary user-defined data that is to be associated with a node. (This metadata is
  * maintained within the ojTree instance, and is not represented in the DOM.)  A node's metadata can be retrieved
  * using the jQuery .data() method.
  * </p></br>Example: Expanded use of the <code class="prettyprint">attr</code> and <code class="prettyprint">metadata</code> properties
  * <pre class="prettyprint">
  * <code>
  *[
  *  {
  *    "title": "Home",
  *    "attr": {
  *               "id": "home",
  *               "myattr1": "Hello",         &lt;-- additional user-defined attributes
  *               "myattr2": "World"          &lt;-- additional user-defined attributes
  *            },
  *    "metadata": {                          &lt;-- node metadata
  *                  "type": "T123",
  *                  "val": 42,
  *                  "active": true
  *                }
  *  },
  *
  *  . . .
  *  $('#mytree').on("ojoptionchange", function (ev, ui) {
  *                  if (ui.option && ui.option == "selection") {
  *                    // retrieve metadata from (first) selected node
  *                    var meta = $(ui.value[0]).data() ;
  *                  }
  *              }) ;
  *]
  *</code></pre>
  *
  * </p></br>Example: Retrieving node attributes and data
  * <pre class="prettyprint">
  * <code>
  *$("#mytree).on("ojtreehover", function (ev, ui){
  *
  *  // ui.item = node
  *  // ui.item.attr("id")         -  retrieve a node attribute
  *  // ui.item.attr("myattr1")    -    ..
  *  // ui.item.data("active")     -  retrieve the "active" meta-data value from previous example
  *
  *});
  *</code></pre>
  *</p></br>
  * For flexibility, attributes can also be applied to the node's &lt;a&gt; element if required, by specifying
  * the node <code class="prettyprint">data</code> property as an object.
  * </p>Example: Using the data property
  * <pre class="prettyprint">
  * <code>
  *{
  *   "attr" : { "id" : "myid" },                    &lt;-- this is set on the &lt;li&gt;
  *   "data" : {
  *              "attr" : {
  *                         "flags"   : "A-B",       &lt;-- this is set on the &lt;a&gt;
  *                         "title" : "This is a tooltip"
  *                       }
  *             }
  *}
  *</code></pre>
  *</p></br>
  * <h4 id="treeHTML-section"> HTML Node Format</h4>
  * </br>
  * A Tree can be populated via standard HTML markup using a &lt;ul&gt; list structure - refer to
  * <code class="prettyprint">option</code> property <code class="prettyprint">"data"</code>.  In
  * the case where the <code class="prettyprint">"data"</code> option has not been defined, ojTree
  * will use any HTML markup defined in the Tree's containing &lt;div&gt;, and on startup the &lt;ul&gt;
  * the markup will be detached from the containing &lt;div&gt;, saved, and used as a template to create a new
  * tree structure in its place.  When the tree is destroyed, the original markup is restored.  Lazy loading of
  * a node's children (when expanded) is performed if any node indicates that it has children,
  * but its child &lt;ul&gt; list is left empty.
  * </p></br>Example: Using HTML markup to populate a Tree.
  * <pre class="prettyprint">
  * <code>
  * &lt;div id="mytree"&gt;
  *    &lt;ul&gt;
  *       &lt;li id="home"&gt;
  *          &lt;a href="#"&lt;Home&gt;/a&gt;
  *       &lt;/li&gt;
  *       &lt;li id="news"&gt;
  *          &lt;a href="#"&gt;News&lt;/a&gt;
  *       &lt;/li&gt;
  *       &lt;li id="blogs"&gt;
  *            &lt;a href="#"&gt;Blogs&lt;/a&gt;
  *            &lt;ul&gt;
  *              &lt;li id="today"&gt;
  *                 &lt;a href="#"&gt;Today&lt;/a&gt;
  *              &lt;/li>
  *              &lt;li id="yesterday"&gt;
  *                 &lt;a href="#"&gt;Yesterday&lt;/a&gt;
  *              &lt;/li&gt;
  *            &lt;/ul&gt;
  *       &lt;/li&gt;
  * &lt;/div&gt;
  </code></pre>
  * </p></br>
  *
  *
  *
  * <h3 id="touch-section">
  *   Touch End User Information
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
  * </h3>
  *
  * {@ojinclude "name":"touchDoc"}
  * <p>
  *
  * <h3 id="keyboard-section">
  *   Keyboard End User Information
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
  * </h3>
  *
  * {@ojinclude "name":"keyboardDoc"}
  *
  *
  * <br>
  * <h3 id="rtl-section">
  *   Reading direction
  *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#rtl-section"></a>
  * </h3>
  *
  * <p>The only supported way to set the reading direction (LTR or RTL) is to set the <code class="prettyprint">"dir"</code> attribute on the
  * <code class="prettyprint">&lt;html></code> element of the page.  As with any JET component, in the unusual case that the reading direction
  * is changed post-init, the tree must be <code class="prettyprint">refresh()</code>ed, or the page must be reloaded.
  *
  * <br>
  *
  * @desc Creates a JET Tree.
  * @param {Object=} options a map of option-value pairs to set on the component
  *
  * @example <caption>Initialize the Tree with options:</caption>
  * $( ".selector" ).ojTree( {"selectionMode": "single", "data": [JSON objects]} );
  *
  */
  oj.__registerWidget("oj.ojTree", $['oj']['baseComponent'],
  {
     widgetEventPrefix : "oj",
     defaultElement    : "<div>",

     options: {
                /** Disables the tree if set to <code class="prettyprint">true</code>.
                  *
                  * @member
                  * @name disabled
                  * @memberof oj.ojTree
                  * @instance
                  * @type {boolean}
                  * @default <code class="prettyprint">false</code>
                  *
                  * @example <caption>Initialize the tree with the <code class="prettyprint">disabled</code> option specified:</caption>
                  * $( ".selector" ).ojTree( { "disabled": true } );
                  *
                  * @example <caption>Get or set the <code class="prettyprint">disabled</code> option, after initialization:</caption>
                  * // getter
                  * var disabled = $( ".selector" ).ojTree( "option", "disabled" );
                  *
                  * // setter
                  * $( ".selector" ).ojTree( "option", "disabled", true );
                  */
                  // disabled option declared in superclass, but we still want the above API doc

                /**
                 * <p>The JET Menu should be initialized before the Tree using it as a 
                 * context menu. 
                 *
                 * @ojfragment contextMenuInitOrderDoc - Decomped to fragment so Tabs, Tree, and MasonryLayout can convey their init order restrictions.
                 * @memberof oj.ojTree
                 */
                /**
                  * {@ojinclude "name":"contextMenuDoc"}
                  *
                  * <p>When defining a contextMenu, ojTree will provide built-in behavior for "edit" style functionality
                  *  (e.g. cut/copy/paste) if the following format for menu &lt;li&gt; item's is used (no &lt;a&gt;
                  *  elements are required):
                  * <ul><li> &lt;li data-oj-command="oj-tree-cut" /&gt;</li>
                  *     <li> &lt;li data-oj-command="oj-tree-copy" /&gt;</li>
                  *     <li> &lt;li data-oj-command="oj-tree-paste" /&gt;</li>
                  *     <li> &lt;li data-oj-command="oj-tree-paste-after" /&gt;</li>
                  *     <li> &lt;li data-oj-command="oj-tree-paste-before" /&gt;</li>
                  *     <li> &lt;li data-oj-command="oj-tree-remove" /&gt;</li>
                  *     <li> &lt;li data-oj-command="oj-tree-rename" /&gt;</li>
                  * </ul>
                  * The available translated text will be applied to menu items defined this way.
                  *
                  * @member
                  * @name contextMenu
                  * @memberof oj.ojTree
                  * @instance
                  * @type {Element|Array.<Element>|string|jQuery|NodeList}
                  * @default <code class="prettyprint">null</code>
                  *
                  * @example <caption>Initialize a JET Tree with a context menu:</caption>
                  * // via recommended HTML5 syntax:
                  * &lt;div id="myTree" contextmenu="myMenu" data-bind="ojComponent: { ... }>
                  *
                  * // via JET initializer (less preferred) :
                  * $( ".selector" ).ojTree({ "contextMenu": "#myContextMenu"  ... } });
                  *
                  * @example <caption>Get or set the <code class="prettyprint">contextMenu</code> option for
                  *      an ojTree after initialization:</caption>
                  * // getter
                  * var menu = $( ".selector" ).ojTree( "option", "contextMenu" );
                  *
                  * // setter
                  * $( ".selector" ).ojTree( "option", "contextMenu", "#myContextMenu"} );
                  */

                /**
                  * Specifies support for HTML5 Drag and Drop. Please refer to
                  * <a href="https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/Drag_and_drop">third-party documentation</a> for details of the HTML5 Drag and Drop facilities.<p>
                  * The <em>dnd</em> property and its values indicates whether:<ol>
                  * <li>the nodes may be reordered using drag and drop within the same tree
                  * <li>the tree can participate in drag/drop operations with other drag/drop
                  *     enabled components or elements.
                  * </ol>
                  * To make a tree reorderable, specify an object with the property
                  * <code class="prettyprint">reorder</code> set to <code class="prettyprint">'enable'</code>.
                  * Setting the <code class="prettyprint">reorder</code>
                  * property to <code class="prettyprint">"disable"</code>, or omitting the
                  * <code class="prettyprint">reorder</code> property disables reordering support.
                  * <br></br>
                  * <b>Example: Enable drag and drop for tree node reordering:</b>
                  *<pre class="prettyprint">
                  *<code>
                  * dnd : (
                  *         reorder : 'enable'
                  *       }
                  *</code></pre>
                  * </br></br>
                  * To allow tree nodes to act as a drag source, specify the optional property <code class="prettyprint">drag</code>.
                  * The  drag <code class="prettyprint">node</code> property object allows the application to control the drag process.
                  * <br></br>
                  * <b>Example: Enable drag support:</b>
                  *<pre class="prettyprint">
                  *<code>
                  * dnd : (
                  *         drag : {
                  *                  node: {
                  *                          ... drag callback properties
                  *                        }
                  *                }
                  *       }
                  *</code></pre>
                  *</br>
                  * To allow a tree to act as a drop target, specify the optional property <code class="prettyprint">drop</code>.
                  * The properties of the <code class="prettyprint">drop</code> object relate to movement over potential node
                  * drop targets and the drop itself, and permit application control of the drop process.
                  * <br></br>
                  * <b>Example: Enable drop support:</b>
                  *<pre class="prettyprint">
                  *<code>
                  * dnd : (
                  *         drop : {
                  *                  node: {
                  *                          ... drop callback properties
                  *                        }
                  *                }
                  *       }
                  *</code></pre>
                  *</br>
                  * Note: if the neither of the <code class="prettyprint">drag</code>
                  * <code class="prettyprint">dataTypes</code> nor <code class="prettyprint">dragStart</code> properties
                  * are specified, the tree will not be drag enabled.  If the <code class="prettyprint">drop</code> property
                  * is not specifed, the tree will not be drop enabled.  However, if a tree is drag and/or drop disabled and
                  * <code class="prettyprint">reorder: 'enabled'</code> is defined, drag/drop will still be permitted for node
                  * reordering only (within the same tree).</br>
                  *
                  * </br>
                  * @property {string} reorder  Valid values are <code class="prettyprint">'enable'</code> or
                  *                             <code class="prettyprint">'disable'</code>. Default if omitted
                  *                             is <code class="prettyprint">'disable'</code. When reorder support
                  *                             is enabled, the Tree automatically adds the HTML attribute draggable=true
                  *                             to all tree nodes.
                  * @property {Object} drag     Specifies that the tree will act as a drag source.<br>
                  *                             The following drag options allow callback functions
                  *                             to be specified to provide application control over the drag process.</br>
                  *                             (For a discussion of the possible return values from these callback functions,
                  *                             refer to the HTML5 documentation for the <code class="prettyprint">dragstart</code>,
                  *                             <code class="prettyprint">drag</code>, and <code class="prettyprint">dragend</code> events.)
                  * @property {Object} drag.node  Specify an object containing the following optional callback properties.
                  * To permit dragging, at least one of the properties <code class="prettyprint">dataTypes</code> or
                  * <code class="prettyprint">dragStart</code> is required. That is, if <code class="prettyprint">dataTypes</code>
                  * is specified, then <code class="prettyprint">dragStart</code> is not required unless additional control of
                  * the drag start is required. If <code class="prettyprint">dataTypes</code> is not defined, then
                  *<code class="prettyprint">dragStart</code> should be used to assign the node data to a data type in the
                  * event.dataTransfer object.
                  * @property {string | Array.<string>} drag.node.dataTypes  The optional MIME type(s) to use for the dragged
                  *    data in the dataTransfer object. May be specified as a string if there is only one type, or an array of
                  *    strings if multiple types are needed.<p>
                  *    When a drag starts, each data type will be set using <em>dataTransfer.setData()</em> with the stringified
                  *    array of the JSON representation of the dragged tree node(s). (These node objects may be recreated via
                  *    JSON.parse()).  Each array entry is of the same format as used to create/insert nodes using the create() method.)<p>
                  *    If not specified, the app should set its own data type(s) in its <em>dragStart</em> callback, using the node
                  *    data found in the callback's <code class="prettyprint">ui</code> argument.  If the <em>dataTypes</em> property is
                  *    omitted and no data type is set in the <em>dragStart</em> callback, drag is not permitted.
                  * @property {function} drag.node.dragStart  An application callback function that will be called
                  *                                      when a drag operation is initiated on a node.</br></br>
                  *                                      <em>function(event, ui)</em>
                  *              <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                 <tr><td>event</td><td>Event</td><td>the jQuery wrapped native <em>dragstart</em> event object.</td></tr>
                  *                 <tr><td>ui</td><td>Object</td><td>tree parameters.
                  *                     <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                            <tr><td>item</td><td>array of objects</td><td>the node(s) being dragged.</td></tr>
                  *                     </table></td></tr>
                  *              </table>
                  *              The <code class="prettyprint">dragStart</code> callback function may use the dataTransfer
                  *              method <code class="prettyprint">setDragImage()</code> if it wishes to replace the  default
                  *              drag image, and also the <code class="prettyprint">effectAllowed</code>
                  *              property. If the callback does not modify <em>effectAllowed</em>, a default of <em>'copy'</em>
                  *              is used if the <em>ctrl</em> key is used, else it is set to <em>'move'</em>.  If the <em>dataTypes</em> property
                  *              has not been specified, this callback should set a data type using event.dataTransfer.setData() and
                  *              the node data in the <em>ui</em> argument. If the <em>dataTypes</em> property is omitted and no data
                  *              type is set in the dragStart callback, drag is not permitted.
                  *        <p>This function should return <em>true</em> to indicate that the drag is permissible or <em>false</em>
                  *         otherwise.  If this function does not return a value, the drag will be permitted.
                  * @property {function} drag.node.drag  An application callback function that will be called repeatedly as the
                  *                                      drag source is dragged.</br></br>
                  *                                      <em>function(event)</em>
                  *              <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                 <tr><td>event</td><td>Event</td><td>the jQuery wrapped native <em>drag</em> event object.</td></tr>
                  *              </table>
                  * @property {function} drag.node.dragEnd  An application callback function that will be called
                  *                                      when a drag operation completes (regardless of whether
                  *                                      the drop operation ever occurs or ends successfully or not.</br></br>
                  *                                      <em>function(event, ui)</em>
                  *              <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                 <tr><td>event</td><td>Event</td><td>the jQuery wrapped native <em>dragend</em> event object.</td></tr>
                  *                 <tr><td>ui</td><td>Object</td><td>tree parameters.
                  *                   <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                       <tr><td>reorder</td><td>boolean</td><td><em>true</em> if a reorder was just performed, else <em>false</em>.<p>
                  *                            If <em>reorder</em> is false the <em>dragEnd</em> callback should remove any nodes from the
                  *                            tree that were moved (that is, <em>dropEffect</em> is <em>"move"</em>).</td></tr>
                  *                   </table>
                  *                 </td></tr>
                  *              </table>
                  *
                  * @property {Object} drop       Specifies that the tree can act as a drop target.
                  * @property {Object} drop.node  An object allowing callback functions to be specifed to provide drag/drop
                  *                               feedback to the application.</br>
                  *                               (For a discussion of the possible return values from these callback functions,
                  *                               refer to the HTML5 documentation for the <code class="prettyprint">dragenter</code>,
                  *                               <code class="prettyprint">dragover</code>, <code class="prettyprint">drop</code>,
                  *                               and <code class="prettyprint">dragleave</code> events.)
                  * @property {function} drop.node.dragEnter An application callback function that will be called
                  *                                      when a tree node is entered during a drag.</br></br>
                  *                                      <em>function(event, ui)</em>
                  *              <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                 <tr><td>event</td><td>Event</td><td>the jQuery wrapped native <em>dragenter</em> event object.</td></tr>
                  *                 <tr><td>ui</td><td>Object</td><td>tree parameters.
                  *                   <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                     <tr><td>item</td><td>Object</td><td>the jQuery wrapped tree node being entered. This will be null if the tree is empty.</td></tr>
                  *                     <tr><td>position</td><td>string</td><td>the drop position relative to the
                  *                             reference node. Can be <em>'before'</em>, <em>'after'</em>, <em>'inside'</em>, or <em>'first'</em>.</td></tr>
                  *                     <tr><td>reference</td><td>Object</td><td>the jQuery wrapped reference node that <em>ui.position</em> refers to.
                                                         This will be null if the tree is empty.</td></tr>
                  *                   </table>
                  *                 </td></tr>
                  *              </table>
                  *        This function should return <em>false</em> to indicate that the dragged data can be accepted,
                  *        or <em>true</em> otherwise. Any explicit return value will be passed back to jQuery.  Returning
                  *        <em>false</em> will cause event.stopPropagation() and event.preventDefault() to be called.
                  *        (event.preventDefault() is required by HTML5 Drag and Drop to indicate acceptance of the dragged
                  *        data at the potential drop position.)<p>
                  *        If this function does not return a value, <em>dataTypes</em> will be matched against the drag data
                  *        types to determine if the data is acceptable. If <em>dataTypes</em> is not defined, the dragged data will be accepted.
                  *
                  * @property {function} drop.node.dragOver  An application callback function that will be called
                  *                                     when a tree node is dragged over. (Note that there will be
                  *                                     multiple calls to this function as the mouse moves over
                  *                                     the node.</br></br>
                  *                                      <em>function(event, ui)</em>
                  *              <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                 <tr><td>event</td><td>Event</td><td>the jQuery wrapped native event object.</td></tr>
                  *                 <tr><td>ui</td><td>Object</td><td>tree parameters.
                  *                    <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                       <tr><td>item</td><td>Object</td><td>the jQuery wrapped tree node being dragged over. This will be null if the tree is empty.</td></tr>
                  *                       <tr><td>position</td><td>string</td><td>the drop position relative to the
                  *                          reference node Can be <em>'before'</em>, <em>'after'</em>, <em>'inside'</em>, or <em>'first'</em>.</td></tr>
                  *                      <tr><td>reference</td><td>Object</td><td>the reference node that <em>ui.position</em> refers to.  this will be null if the tree is empty.</td></tr>
                  *                    </table>
                  *                 </td></tr>
                  *              </table>
                  *        As for <em>dragEnter</em>, this function should return <em>false</em> to indicate that the dragged data
                  *       can be accepted at this position, or <em>true</em> otherwise.  If this function does not return a value,
                  *       <em>dataTypes</em> will be matched against the drag data types to determine if the data is acceptable.
                  *       If <em>dataTypes</em> is not defined the dragged data will be accepted.
                  *
                  * @property {function} drop.node.drop  An application callback function that will be called
                  *                                     when the dragged object is dropped on a tree.</br></br>
                  *                                      <em>function(event, ui)</em>
                  *
                  *              <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                 <tr><td>event</td><td>Event</td><td>the jQuery wrapped native event object.</td></tr>
                  *                 <tr><td>ui</td><td>Object</td><td>tree parameters.
                  *                    <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                       <tr><td>position</td><td>string</td><td>the drop position relative to the
                  *                           reference node Can be <em>'before'</em>, <em>'after'</em>, <em>'inside'</em>, or <em>'first'</em>.</td></tr>
                  *                       <tr><td>reference</td><td>string</td><td>the reference node that <em>ui.position</em> refers to.</td></tr>
                  *                       <tr><td>reorder</td><td>boolean</td><td><em>true</em> if a reorder was just performed, else <em>false</em>.<p>
                  *                            If <em>reorder</em> is <em>true</em>, the <em>drop</em> callback should not perform any tree
                  *                            restructuring since this is automatically performed as part of the reorder.</td></tr>
                  *                    </table>
                  *                 </td></tr>
                  *              </table>
                  *        This function should return <em>false</em> to indicate that the dragged data is accepted, or <em>true</em> otherwise.
                  *
                  * @property {function} drop.node.dragLeave An application callback function that will be called
                  *                                     when the drag operation moves out of the last entered node.
                  *                                     It is also called if the escape key is pressed while the
                  *                                     cursor is over a drop position, or a dragOver callback
                  *                                     cancels the drag.</br></br>
                  *                                      <em>function(event, ui)</em>
                  *              <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                 <tr><td>event</td><td>Event</td><td>the jQuery wrapped native event object.</td></tr>
                  *                 <tr><td>ui</td><td>Object</td><td>tree parameters.
                  *                    <table><tr><th>Name</th><th>Type</th><th>Description</th></tr>
                  *                       <tr><td>item</td><td>Object</td><td>the jQuery wrapped tree node that was last entered.</td></tr>
                  *                    </table>
                  *                 </td></tr>
                  *              </table>
                  *
                  * @property {string | Array.<string>} drop.node.dataTypes An optional data type string, or array of strings that the
                  *                                     tree will test for in the <code class="prettyprint">event.dataTransfer</code>
                  *                                     object when the object is moved over a potential tree drop position. If
                  *                                     none of the dataTransfer data types are found in this list, a drop over
                  *                                     the current position is not permitted.  If the array is omitted or
                  *                                     empty, the drag operation will continue.<p>
                  *                                     Refer also to <em>dataTypes</em> in the <em>drag</em> property object.
                  *</br></br>
                  *
                  * @example <caption>Allow drag from a tree, do not allow drop on the tree.</caption>
                  * <code>
                  *Note: to permit dragging, at least one of the dataTypes or dragStart properties is required. That is, if
                  *dataTypes is specified, dragStart is not required unless additional tuning of the drag start is needed.
                  *If dataTypes is not defined, then dragStart should be used to assign the node data to a dataType.
                  *
                  *dnd: {
                  *       drag: {
                  *               node : {
                  *                        datatypes: ['xxx'],
                  *                        dragStart: myDragStart,
                  *                        dragEnd:   myDragend
                  *                      }
                  *             }
                  *     }
                  *
                  *self.myDragStart = function(event, ui)
                  *{
                  *   event.preventDefault() ;   // allow drag
                  *   return false ;
                  *}
                  *
                  *self.myDragend = function(event)
                  *{
                  *  var selections ;
                  *
                  *  if (event.originalEvent.dataTransfer.dropEffect === "move") {
                  *    // Remove the moved node(s) selected.
                  *    if (self.$tree.ojTree("isSelected", event.target)) {
                  *      selections =  self.$tree.ojTree("option", "selection") ;
                  *      for (i = 0; i < selections.length; i++) {
                  *         self.$tree.ojTree("remove", selections[i]) ;
                  *      }
                  *    }
                  *    else {
                  *      // Single unselected node moved
                  *      self.$tree.ojTree("remove", event.target) ;
                  *    }
                  *  }
                  *}
                  * </code>
                  *
                  * @example <caption>Handle drop of node(s) from another tree.</caption>
                  * <code>
                  *dnd: {
                  *       drop: {
                  *               node : {
                  *                        dataTypes: ['xxx'],
                  *                        drop:  myDrop
                  *                      }
                  *             }
                  *     }
                  *
                  *self.myDrop = function(event, ui)
                  *{
                  *  var data = JSON.parse(event.originalEvent.dataTransfer.getData('xxx')) ;
                  *
                  *  self.$tree.ojTree("create", ui['reference'][0], ui['position'], data) ;
                  *  return false ;
                  *}
                  * </code>
                  *
                  * @example <caption>Do not allow drag of a particular node.</caption>
                  * <code>
                  *dnd: {
                  *       drag: {
                  *               node : {
                  *                        dragStart:  function(event, ui) {
                  *                             if (ui.item.id != "myTopId") {
                  *                               event.preventDefault() ;     // allow drag
                  *                               return false ;
                  *                             }
                  *                             // else can't drag
                  *                        }
                  *                      }
                  *             }
                  *     }
                  * </code>
                  * @example <caption>Disallow drop above a certain node.</caption>
                  * <code>
                  *dnd: {
                  *       drag: {
                  *               node : {
                  *                        dragEnter:  function(event, ui)  {
                  *                             if ((ui.position != 'above') ||
                  *                                 (ui.item.id) != 'mynode')) {
                  *                               event.preventDefault() ;     // allow drag
                  *                               return false ;
                  *                             }
                  *                             // else can't drop here
                  *                             return true ;
                  *                        }
                  *                      }
                  *             }
                  *     }
                  * </code>
                  *
                  * @type {Object}
                  * @default <code class="prettyprint">{reorder:'disable'}</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                dnd : {reorder:'disable'},

                /**
                  * Specify <span class="code-caption">true</span> if expanding a node programatically should
                  * also expand its parents (i.e all parent nodes down to this node will be expanded).
                  * @type {boolean}
                  * @default <code class="prettyprint">false</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                expandParents : false,

                /**
                  * Specifies whether any nodes should be initially expanded on start-up.   Specify an array
                  * of node id's, or the string <span class="code-caption">"all"</span> if all parent nodes
                  * should be expanded.  The value may optionally be specified as an empty array.
                  * @type {Array | null}
                  * @default <code class="prettyprint">null</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                initExpanded  : null,

                initLoaded    : [],                  // suppressed per design review

                /** An array of node elements that are currently selected. If the array is modified
                  * by an application, the selected node status of the tree is modified to match the array
                  * (nodes may be defined as elements, jQuery wrapped nodes, or selectors pointing to the
                  * elements that should be selected).
                  * @type {Array}
                  * @default <code class="prettyprint">Array</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                selection : [],

                /**
                  * Specifies whether selection is permitted, and whether more than one node
                  * can be selected at a time.  Values are <span class="code-caption">"single"</span>
                  * for single selection, <span class="code-caption">"multiple"</span> to allow multiple
                  * concurrent selections, and <span class="code-caption">"none"</span> to inhibit selection.
                  * @type {string}
                  * @default <code class="prettyprint">"single"</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                selectionMode  :  "single" ,

                /**
                  *  Specifies what action is to be taken when a selected node's parent is collapsed.  Specify
                  *  false if nothing is to be done. Specify <span class="code-caption">"selectParent"</span>
                  *  if the node's closed parent is to be selected, or specify <span class="code-caption">"deselect"</span>
                  *  if the node is to be deselected.
                  * @type {boolean | string}
                  * @default <code class="prettyprint">false</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                selectedParentCollapse  :  false ,

                /**
                  *  Specifies what action is to be taken when a node is programmatically expanded.  Specify
                  *  <span class="code-caption">true</span> if all of the node's closed parents should be opened
                  *  automatically.  If <span class="code-caption">false</span> is specified, the node is selected but will
                  *  remain invisible if its parents are currently collapsed.
                  * @type {boolean}
                  * @default <code class="prettyprint">true</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                selectedParentExpand   :  true ,

                /**
                  * Specifies the action to take when a selected node is deleted.  If set to
                  * <span class="code-caption">true</span>, its previous sibling (or parent, if no previous siblings)
                  * is selected.  If <span class="code-caption">false</span> is specified, no action is taken.
                  * @type {boolean}
                  * @default <code class="prettyprint">false</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                selectPrevOnDelete     :  false ,

                /**
                  * Specifies the data source used to populate the tree. Currently supported data sources are a <code class="prettyprint">JsonTreeDataSource</code>,
                  * or json, or html.</br></br>
                  * The general format of the <code class="prettyprint">data</code> option is one of the following:
                  *</br></br>
                  *<ul>
                  *   <li>data : oj.JsonTreeDataSource</br></br></li>
                  *   <li>data : null    (or omit) - ojTree will look at the containing &lt;div&gt;
                  *                       and use any existing html &lt;ul&gt; markup found</br></br></li>
                  *   <li>data : "  json string  "</br></br></li>
                  *   <li>data : [ array of json objects ]</br></br></li>
                  *   <li>data : "&lt;ul&gt;&lt;li&gt; ...  html markup string  &lt;/ul&gt;"</br></br></li>
                  *   <li>data : { "data" : &nbsp; &nbsp; ... or &nbsp; &nbsp; "ajax" : &nbsp; &nbsp; . . . &nbsp; &nbsp;}  &nbsp; &nbsp; &nbsp; // retrieve json or html</li>
                  * </ul>
                  *</br>
                  * Use of the <code class="prettyprint">"data"</code> property of the <code class="prettyprint">data</code> option,
                  * specifies that the tree is to be populated from JSON or HTML (local or remote).
                  * The <code class="prettyprint">"data"</code> object contains one of two properties:
                  * <ul>
                  *  <li>"data"</li>
                  *  <li>"ajax"</li>
                  * </ul>
                  *  An optional <code class="prettyprint">"dataType"</code> property may also be specified, which can take the
                  *  value <code class="prettyprint">"json"</code> or <code class="prettyprint">"html"</code>, and indicates
                  *  what kind of data is being returned in the <code class="prettyprint">"data"</code> or
                  *  <code class="prettyprint">"ajax"</code> method (default is "json").
                  * </ul>
                  *
                  * When <span class="code-caption">"data"</span> is specified as an object, its <span class="code-caption">"data"</span> property may be specified as a function which
                  * receives two arguments: <span class="code-caption">node</span>, and <span class="code-caption">fn</span>.
                  * </p></br>
                  * Example: Skeleton outline of a <code class="prettyprint">"data"</code> function:
                  *</br>
                  *<pre class="prettyprint">
                  *<code>
                  *data : {
                  *          "data" : function(node, fn) {
                  *                    // node  -  the jQuery wrapped node to be expanded for a lazy load,
                  *                    //          or -1 if it is the initial call to load the tree.
                  *                    // fn    -  a function to call with the JSON to be applied.
                  *
                  *                    fn( new_json_node_data ) ;   // return the JSON
                  *                   }
                  *        }
                  *</code></pre>
                  * </br>
                  * The <code class="prettyprint">"ajax"</code> property of the <code class="prettyprint">"data"</code> option
                  * allows remote JSON to be retrieved. It may be specified as an object (refer to the
                  * jQuery .ajax() settings object). If may also be specified as <code class="prettyprint">false</code> or
                  * omitted, if no AJAX operations are performed.</br></br>
                  * When specified as an object, it should contain the following two properties:
                  * <ul>
                  *  <li>type</li>
                  *  <li>url</li>
                  * </ul>
                  *<pre class="prettyprint">
                  *<code>
                  *"ajax" : {
                  *           "type": "GET",
                  *           "url":   "my_url"      // some url to the content
                  *          }
                  *</code></pre>
                  * <code class="prettyprint">"url"</code> may also be specified as a function which should return
                  * a url string:
                  *</br>
                  *<pre class="prettyprint">
                  *<code>
                  *"ajax" : {
                  *           "type" : "GET",
                  *           "url":   function (node) {
                  *                         ... return a url string ...
                  *                     }
                  *          )
                  *</code></pre>
                  * </br>
                  *  where  <span class="code-caption">node</span> is a parent node (can be used for lazy loading), or -1 to
                  *  indicate the initial tree load.
                  * </br></br>
                  *  Optionally, <span class="code-caption">success</span> and <span class="code-caption">error</span>
                  *  functions may be defined. If the <span class="code-caption">success</span> function returns a
                  *  value, it will be used to populate the tree; this can be useful if there is a need to transform
                  *  the data returned by a server at the client before it is displayed in the tree.
                  * </br></br></br>
                  *
                  * Note: to enable lazy loading of a parent node, specify that it has children but do not define them.
                  * When it is opened, data() or ajax() will be called with the node whose JSON is to be returned.</br></br>
                  *@example <caption>Example 1: Skeleton outline of success and error functions</caption>
                  *<code>
                  *"ajax": {
                  *          "type":"GET",
                  *          "url": myurl    &lt;-- url to full tree JSON
                  *          "success" : function(data, status, obj) {
                  *                        // data   = the JSON data
                  *                        // status = "success"
                  *                        // obj    = the AJAX object.
                  *                        trace("Ajax " + status) ;
                  *                        // return the data, can transform it first if required.
                  *                        // if no return value, the data is used untransformed.
                  *          },
                  *          "error" : function(reason, feedback, obj) {
                  *                        // reason e.g. "parsererror"
                  *                        // feedback.message  e.g. "unexpected string"
                  *                        // obj    = the AJAX object.
                  *                        trace("Ajax error " + reason + " feedback=" + feedback.message) ;
                  *          },
                  * </code>
                  *
                  *@example <caption>Example 2:  Load the complete tree from locally defined JSON.</caption>
                  *<code>
                  *"data" :  [
                  *            {
                  *             "title": "Home",
                  *             "attr": {"id": "home"},
                  *            },
                  *            {
                  *              "title": "News",
                  *              "attr": {"id": "news"}
                  *            },
                  *            {
                  *              "title": "Blogs",
                  *              "attr": {"id": "blogs"},
                  *              "children": [ {
                  *                             "title": "Today",
                  *                             "attr": {"id": "today"}
                  *                            },
                  *                            {
                  *                              "title": "Yesterday",
                  *                              "attr": {"id": "yesterday"}
                  *                            }
                  *                          ]
                  *            }
                  *          ]
                  *</code>
                  *
                  *@example <caption>Example 3:  Load the complete tree with remotely served JSON.</caption>
                  *<code>
                  *"data" : {
                  *            "ajax": {
                  *                     "type":"GET",
                  *                     "url": myurl    <-- url to full tree JSON
                  *                    }
                  *
                  *          }
                  *</code>
                  *
                  *@example <caption>Example 4:  Load the complete tree with remotely served JSON via a function.</caption>
                  *<code>
                  *"data" : {
                  *
                  *           "ajax": {
                  *                     "type":"GET",
                  *                     "url": function() {
                  *                               return (a url) ;
                  *                            }
                  *                   }
                  *
                  *          }
                  * </code>
                  *
                  *@example <caption>Example 5:  Load a partial tree, and retrieve node data when a parent node is expanded and needs to be populated.</caption>
                  *<code>
                  *"data" : {
                  *           "ajax": {
                  *                     "type":"GET",
                  *                     "url": function(node) {
                  *                             if (node === -1) {                       // -1 indicates initial load
                  *                               return (url for for  partial json) ;   // the tree outline with parent nodes empty.
                  *                             }
                  *                             else {
                  *                               var id = node.attr("id") ;
                  *
                  *                               return (a url based on the node id to retrieve just the node children) ;
                  *                             }
                  *                           }
                  *                   }
                  *
                  *          }
                  *</code>
                  *
                  *@example <caption>Example 6:  Transform data received from server before passing to ojTree.</caption>
                  *<code>
                  *"data" : {
                  *           "ajax": {
                  *                     "type":"GET",
                  *                     "url": function(node) {
                  *                              . . .
                  *                            },
                  *                      "success" : function (data)  {
                  *                                    . . .    // transform the received data into node JSON format
                  *
                  *                                    return (transformed data) ;
                  *                                  },
                  *                      "error" : function () {
                  *                                   // ajax call failed.
                  *                                }
                  *                   }
                  *
                  *          }
                  *</code>
                  *
                  * @example <caption>Example 7:  Use own mechanism to load a partial tree and retrieve node data when a parent is expanded.</caption>
                  * <code>
                  * // Sample outline of a tree.  Note that the parent nodes "Node2" and "Node3" have
                  * // their "children" property specifed, but no children are actually defined.
                  *
                  *{
                  *  "title" : Node1",
                  *  "attr" : {"id" : "n1"}
                  *},
                  *{
                  *  "title" : Node2",
                  *  "attr" : {"id" : "n2"},
                  *  "children" : []
                  *},
                  *{
                  *  "title" : Node3",
                  *  "attr" : {"id" : "n3"},
                  *  "children" : []
                  *},
                  *
                  *
                  *"data" : {
                  *           "data": function(node, fn) {
                  *                     // node  =  the node whose children are to be retrieved
                  *                     // fn    =  the function to call with the retrieved node json
                  *
                  *                     if (node === -1) {             // initial tree load
                  *                       fn( acquired node json for the tree) ;
                  *                     }
                  *                     else {                         // node lazy load
                  *                       var id = node.attr("id") ;   // get the node id, will be "n2"
                  *                                                    // or "n3", in this example.
                  *                       fn( acquired node json for the expanded node ) ;
                  *                     }
                  *                  }
                  *
                  *          }
                  *}
                  * </code>
                  * When an <span class="code-caption">option</span> call is made to reset the <span class="code-caption">data</span> property
                  * of a tree, the application does not need to call <span class="code-caption">refresh</span>.
                  * @type {Object | Array | string | null}
                  * @default <code class="prettyprint">null</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                data : null,

                /**
                 * The text to display when there are no data in the Tree. If not specified,
                 * default text is extracted from the resource bundle.  Specify an empty string if
                 * this default behavior is not required.
                 *
                 * @expose
                 * @memberof oj.ojTree
                 * @instance
                 * @type {string|null}
                 * @default <code class="prettyprint">"No data"</code>
                 *
                 * @example <caption>Initialize the tree with text set to 'no data':</caption>
                 * $( ".selector" ).ojTree({ "data":data, "emptyText": "no data" });
                 */
                emptyText: null,

                /**
                  * Specifies whether node icons are to be displayed.  Specify <span class="code-caption">true</span>
                  * to display icons, or <span class="code-caption">false </span> to suppress node icons.
                  * @type {boolean}
                  * @default <code class="prettyprint">true</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                icons : true,

                /*
                 *  Specifies whether hierarchy lines between nodes are displayed.
                 *  // not exposed in V1
                 */
                //dots" : false,


                /**
                  * The <span class="code-caption">'types'</span> option allow nodes to be classified and their appearance
                  * and behavior modified.</br></br>
                  * Typical uses are to define a specific icon for a particular node, or to inhibit certain
                  * operations on a particular type of folder (e.g. the root node cannot be deleted or moved).
                  * <p>
                  * A node <span class="code-caption">type</span> has the following properties:
                  * <ul>
                  *   <li><span class="code-caption">"image"</span> -  specifies the location of the icon to be used
                  *         (optional). May also be specified as <span class="code-caption">false</span> to suppress
                  *         the image.</br></br></li>
                  *   <li><span class="code-caption">"position"</span> - position of sprite in the image in the format
                  *        <span class="code-caption">"left top"</span>, e.g. "-36px -16px".</br>
                  *        Optional - omit if icon is not contained within a multi-sprite image.</br></br></li>
                  *   <li><span class="code-caption">method name</span> - specify a function or a
                  *         boolean. Optional.</br>  Any node operation method (that
                  *         is, takes a node as its first argument) can be redefined (e.g. <span class="code-caption">select</span>,
                  *         <span class="code-caption">expand</span>, <span class="code-caption">collapse</span>, etc).
                  *         Alternatively, the method can be defined as <span class="code-caption">true</span> or
                  *         <span class="code-caption">false</span> to permit or inhibit the operation, or a
                  *         function that returns a boolean value. The default value
                  *         if omitted is <span class="code-caption">true</span> (i.e. the operation is permitted).</li>
                  * </ul>
                  * In the following example, three node types have been defined: <span class="code-caption">"myroot"</span>,
                  * <span class="code-caption">"myfolder"</span>, and <span class="code-caption">"myleaf"</span>.
                  * Any node that does not have one of these types defaults its behavior to the default type
                  * (whose properties can also be redefined).  The default <span class="code-caption">"default"</span>
                  * node type has no restrictions on the operations that can be performed on the node. In the following
                  * example, a modification to the default type properties have been made.  Also, for the
                  * <span class="code-caption">"myroot"</span> node type, the standard <span class="code-caption">select</span>,
                  * <span class="code-caption">remove</span> and <span class="code-caption">move</span> operations return false
                  * which inhibts those operations.
                  * been redefined to be no-ops.
                  * @example <caption>Example 1:  Add custom appearance and node behavior.</caption>
                  * <code>
                  *"types": {
                  *            "myroot" :   {
                  *                            "image"  : baseurl + "/img/root.png",
                  *                            "select" : function() { return false; },
                  *                            "remove" : function() { return false; },
                  *                            "move" :   function() { return false; },
                  *                         },
                  *            "myfolder" : {
                  *                            "image" : baseurl + "/img/folder.png"
                  *                         },
                  *            "myleaf" :   {
                  *                           "image" : "baseurl + "/img/leaf.png"
                  *                         },
                  *            "default" : {   <-- optional redefinition of the default behavior
                  *                           "image" : "baseurl + "/img/leaf.png",
                  *                           "remove" : function() { return false; }
                  *                        }
                  *
                  *          }
                  *}
                  *</code>
                  * User-defined types are specified as an attribute of the node.  The default
                  * node type attribute is <span class="code-caption">"type"</span>, but this could be changed if desired using
                  * the <span class="code-caption">"attr"</span> property. Thus, for the node types in example 1 above, the node
                  * type attribute values in the node definitions could be set as in example 2:
                  * @example <caption>Example 2:  Using node types in the tree JSON.</caption>
                  * <code>
                  *[
                  *   {
                  *     "title": "Root",
                  *     "attr": {
                  *               "id": "root",
                  *               "type": "myroot"                      &lt;--- node type
                  *             },
                  *     "children": [
                  *                   {
                  *                     "title": "Home",
                  *                     "attr": {"id": "home",
                  *                              "type": "myleaf"}      &lt;--- node type
                  *                   },
                  *                   {
                  *                     "title": "News",
                  *                     "attr": {
                  *                               "id": "news",
                  *                               "type": "myleaf"      &lt;--- node type
                  *                             }
                  *                   },
                  *                   {
                  *                     "title": "Blogs",
                  *                     "attr": {
                  *                               "id": "blogs",
                  *                               "type": "myfolder"    &lt;--- node type
                  *                             },
                  *                     "children": [ {
                  *                                     "title": "Today",
                  *                                     "attr": {
                  *                                               "id": "today",
                  *                                               "type": "myleaf"
                  *                                             }
                  *                                   },
                  *                                   {                 &lt;--- default node type
                  *                                     "title": "Yesterday",
                  *                                     "attr": {"id": "yesterday"}
                  *                                   }
                  *                                 ]
                  *                   }
                  *                 ]
                  *  }
                  *]
                  *</code>
                  * As described above, the node type attribute used on the corresponding tree
                  * &lt;li&gt; element defaults to <span class="code-caption">"type"</span>, but this can be redefined using the <span class="code-caption">attr</span>
                  * property as in the following example:
                  * @example <caption>Example 2:  Using node types in the tree JSON.</caption>
                  * <code>
                  *"types": {
                  *           "attr" : "mytype",    &lt;--- node type attribute is now "mytype"
                  *           "types": {
                  *                      "myroot" : {
                  *                                   "image" : . . .
                  *                                    . . .
                  *                                 }
                  *          }
                  *</code>
                  * @type {Object | null}
                  * @default <code class="prettyprint">true</code>
                  * @expose
                  * @instance
                  * @memberof oj.ojTree
                  */
                types: null,


                //---------------------------//
                //   Option Event callbacks  //
                //---------------------------//

                /**
                  * Triggered prior to an event.<p>
                  * The following events can be vetoed during <code class="prettyprint">before</code> event
                  * processing by returning <code>false</code> from the <code class="prettyprint">before</code>
                  * event handler (omitting a return value or returning <code class="prettyprint">true</code>
                  * permits the event processing to continue):
                  * <code class="prettyprint">collapse</code>, <code class="prettyprint">expand</code>,
                  * <code class="prettyprint">hover</code>, <code class="prettyprint">select</code>,
                  * <code class="prettyprint">remove</code>, <code class="prettyprint">rename</code>.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {string} ui.func the event causing this <code class="prettyprint">before</code> event to be triggered.
                  * @property {Object} ui.item the node that is the subject of the event
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">before</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "before": function(event, ui)  {
                  *                     console.log("Before event " + ui.func);
                  *               }
                  * });
                  *
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojbefore</code> event:</caption>
                  * $( ".selector" ).on( "ojbefore", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) {
                  *                          console.log("Before event " + ui.func);
                  *                   } 
                  *                 });
                  */
                before : null,

                /**
                  * Triggered when a tree node is collapsed.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that has been collapsed
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">collapse</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "collapse": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojcollapse</code> event:</caption>
                  * $( ".selector" ).on( "ojcollapse", function(event, ui) {
                  *                    // Verify that component of interest fired the event
                  *                    if ($(event.target).is("#mytree")) { . . . }
                  *                  });
                  */
                collapse : null,

                /**
                  * Triggered when a tree node has been created and added to the tree.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that has been created
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">create</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "create": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojcreate</code> event:</caption>
                  * $( ".selector" ).on( "ojcreate", function(event, ui) {
                  *                    // Verify that component of interest fired the event
                  *                    if ($(event.target).is("#mytree")) { . . . }
                  *                  });
                  */
                create : null,

                /**
                  * Triggered when all nodes of a parent node, or the complete tree, have been collapsed.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node(s) that were collapsed.
                  * @property {Object} ui.targ the node that was targeted for collapseAll, or -1 if the complete tree is collapsed.
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">collapseAll</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "collapseAll": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojcollapseall</code> event:</caption>
                  * $( ".selector" ).on( "ojcollapseall", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                collapseAll : null,

                /**
                  * Triggered when a tree node has been cut from the tree via the context menu.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that was cut
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">cut</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "cut": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojcut</code> event:</caption>
                  * $( ".selector" ).on( "ojcut", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                cut : null,

                /**
                  * Triggered when a tree node is no longer hovered over.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that is no longer hovered over
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">dehover</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "dehover": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojdehover</code> event:</caption>
                  * $( ".selector" ).on( "ojdehover", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                dehover : null,

                /**
                  * Triggered when a tree node has been removed.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that has been removed.
                  * @property {Object} ui.parent the parent of the node that was removed.
                  * @property {Object} ui.prev the previous sibling, or if ui.item is the first child of
                  *                    its parent, the parent node.
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">remove</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "remove": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojremove</code> event:</caption>
                  * $( ".selector" ).on( "ojremove", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                remove : null,

                /**
                  * Triggered when a tree is destroyed.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">destroy</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "destroy": function( event, ui ) {}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojdestroy</code> event:</caption>
                  * $( ".selector" ).on( "ojdestroy", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                destroy : null,

                /**
                  * Triggered when a tree node is expanded.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that has been expanded
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">expand</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "expand": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojexpand</code> event:</caption>
                  * $( ".selector" ).on( "ojexpand", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                expand : null,

                /**
                  * Triggered when all nodes of a parent node, or the complete tree, have been expanded.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node(s) that were expanded.
                  * @property {Object} ui.targ the node that was targeted for expandAll, or -1 if the complete tree is collapsed.
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">expandAll</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "expandAll": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojexpandall</code> event:</caption>
                  * $( ".selector" ).on( "ojexpandall", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                expandAll : null,

                /**
                  * Triggered when a tree node is hovered over.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that is hovered over
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">hover</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "hover": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojhover</code> event:</caption>
                  * $( ".selector" ).on( "ojhover", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                hover : null,

                /**
                  * Triggered when a tree has been loaded and the node data has been applied.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">loaded</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "loaded": function( event, ui ) {}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojloaded</code> event:</caption>
                  * $( ".selector" ).on( "ojloaded", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                loaded : null,

                /**
                  * Triggered when a tree node has been moved within the tree.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that was moved
                  * @property {string} ui.position the moved node's new position relative to the reference node.
                  *                    Can be <em>"before"</em>, <em>"after"</em>, or <em>"inside"</em>.
                  * @property {Object} ui.reference the reference node that ui.position refers to.
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">move</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "move": function(event, ui) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojmove</code> event:</caption>
                  * $( ".selector" ).on( "ojmove", function(event, ui) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                move : null,

                /**
                  * Triggered when one or more tree nodes have been pasted into the tree via the context menu.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Array} ui.item the node(s) pasted
                  * @property {string} ui.position the placement of the nodes relative to the reference
                  *                     node. May be "inside", "before", or "after".
                  * @property {Object} ui.reference the reference node
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">paste</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "paste": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojpaste</code> event:</caption>
                  * $( ".selector" ).on( "ojpaste", function( event, ui ) {
                  *                   // Verify that the component firing the event is the component of interest
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                paste : null,

                /**
                  * Triggered when a tree node, or the complete tree, has been refreshed.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that has been refreshed, or -1 if the whole tree has been refreshed.
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">refresh</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "refresh": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojrefresh</code> event:</caption>
                  * $( ".selector" ).on( "ojrefresh", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                refresh : null,

                /**
                  * Triggered when a tree node has been renamed.
                  *
                  * @expose
                  * @event
                  * @memberof oj.ojTree
                  * @instance
                  * @property {Event} event <code class="prettyprint">jQuery</code> event object
                  * @property {Object} ui Parameters
                  * @property {Object} ui.item the node that has been renamed
                  * @property {string} ui.title the new node text title.
                  * @property {string} ui.prevTitle the node title prior to the rename.
                  *
                  * @example <caption>Initialize the Tree with the <code class="prettyprint">rename</code> callback specified:</caption>
                  * $( ".selector" ).ojTree({
                  *     "rename": function( event, ui ) {. . .}
                  * });
                  *
                  * @example <caption>Bind an event listener to the <code class="prettyprint">ojrename</code> event:</caption>
                  * $( ".selector" ).on( "ojrename", function( event, ui ) {
                  *                   // Verify that component of interest fired the event
                  *                   if ($(event.target).is("#mytree")) { . . .}
                  *                 });
                  */
                rename : null

     },   // end options


     /*---------------*/
     /* Public API's  */
     /*---------------*/

     /** Collapses an expanded node, so that its children are not visible.  Triggers a "collapse" event.
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node, or a selector
       *                        pointing to the element to be collapsed.
       * @param {boolean=} skipAnim - Set to true to suppress node collapse animation (if a non-zero duration
       *                              is defined or defaulted). Default is false.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     collapse : function(node, skipAnim)
     {
        skipAnim = skipAnim || false ;

        // Get css value $treeAnimationDuration.  This is needed because if the value is zero
        // and a transition is started, browsers do not apparently fire the "transitionend"
        // event and so we don't get the opportunity to emit a collapse event.

        var dur  = skipAnim? 0 : this._animDuration ;

        node = this._getNode(node);

        if (!node.length || node === -1 || !node.hasClass(TreeUtils._OJ_EXPANDED) ||
                            this._data.core.locked || node.hasClass(TreeUtils._OJ_DISABLED))  {
          return false;
        }

        var rslt = this._emitEvent({"obj" : node, "func" : "collapse"}, "before") ;
        if (typeof rslt == "boolean" && (!rslt)) {
          return ;
        }

//      if (skipAnim || dur)  {
//        node.children("ul").attr("style","display:block !important");
//      }

        node.removeClass(TreeUtils._OJ_EXPANDED)
            .addClass(TreeUtils._OJ_COLLAPSED)
            .attr(WA_EXPANDED, "false");
        $(node.children()[0])
              .removeClass(TreeUtils._OJ_SELECTED)
              .addClass("oj-default") ;

        this._busyStack.push( {op : 'c', id : node.attr("id")});

        // Add a busy state for the animation.  The busy state resolver will be invoked
        // when the animation is completed
        this._addBusyState("Tree (id='" + this._elemId  + "') : node id='" + node.attr('id') + "' is animating.");

        if ((!skipAnim) && dur ) {
          this._slide(node, true) ;     // slideUp
        }
        else  {
          this._transitionEnd($(node.children("UL")[0]), node) ;
        }
     },

    /**
      *  @private
      */
     _slide : function(node, bSlideUp)
     {
        var $ul, hNow, hFinal ;

        $ul = $(node.children("ul")) ;

        // For the windows safari platform only we have a problem setting "transitionend" events,
        // the transition end event handler does not get called. Also For the iPad, the transition
        // end handler is called multiple times for some unknown reason.
        // ul.on("transitionend", this._proxyTransitionEndHandler) ;    // doesn't work!

        if (this._isSafari) {
          $ul[0].addEventListener("webkitTransitionEnd", this._proxyTransitionEndHandler);
        }
        else {
          $ul[0].addEventListener("transitionend", this._proxyTransitionEndHandler);
        }

        if (bSlideUp) {
          hNow   = $ul[0].offsetHeight ;        // collapse
          hFinal = 0 ;
        }
        else
        {                                      // expand
          hNow   = 0 ;
          hFinal = this._getElemHeight($ul[0]) ;
        }

        $ul.css('max-height', hNow + "px") ;       // set the current height
        this._overflow = $ul.css('overflow') ;     // note for reset in _transitionEnd()
        $ul.css('overflow', 'hidden') ;

        if (! bSlideUp) {
          $ul[0].style.display = 'block' ;
        }

        setTimeout(function() {
          $ul.addClass("oj-tree-transition");       // enable transitioning on max-height
          $ul.css('max-height', hFinal + "px") ;    // start the transition
        }, 20) ;

     },


     /**
       * Compute the height of the ul element
       * @private
       */
     _getElemHeight : function( el)
     {
       var parent, nextSib, h, disp,
           div = $("<div style='position: absolute'>") ;

       parent  = el.parentNode;
       nextSib = el.nextSibling;
       parent.removeChild(el);

       disp = el.style.display ;                   // note display
       el.style.display = "block";                 // set - it might be 'none'
       div.append(el) ;                            //@HTMLUpdateOK
       this._$container[0].appendChild(div[0]) ;   //@HTMLUpdateOK

       h = (el.offsetHeight || el.scrollHeight) ;

       // Restore DOM
       el.style.display = disp;                    // reset orig display
       this._$container[0].removeChild(div[0]) ;
       div[0].removeChild(el) ;
       if (nextSib) {
         parent.insertBefore(el, nextSib);         //@HTMLUpdateOK
       }
       else {
          parent.appendChild(el);                  //@HTMLUpdateOK
       }

//     return h ;
       return h + 10;    // TDO why is this a few pixels too little.  It causes
                         // a small jump at the end of the transition when the
                         // ul style is set to 'block'
     },


     /**
       *  @private
       */
     _transitionEndHandler : function (e)
     {
        var $ul  = $(e.target) ;
        var node = $ul.closest("li") ;

        // For the windows safari platform only we have a problem with "transitionend" events,
        // (see _slide() where the event listener is set up.  This code matches the routine in _slide().
        // $ul.off("transitionend", this._proxyTransitionEndHandler) ;    // doesn't work

        if (this._isSafari) {
          $ul[0].removeEventListener("webkitTransitionEnd", this._proxyTransitionEndHandler);
        }
        else {
          $ul[0].removeEventListener("transitionend", this._proxyTransitionEndHandler);
        }

        $ul.removeClass("oj-tree-transition") ;
        this._transitionEnd($ul, node) ;
     },

    /**
      *  Common code called after a collapse/expand whether there was a transition or not.
      *  @private
      */
     _transitionEnd : function($ul, node)
     {
        var op ;

        if (node.hasClass(TreeUtils._OJ_COLLAPSED)) {   // end collapse
          op = 'c' ;       // collapse op
          $ul[0].style.display = "none";
$ul.css('max-height', '') ;   //JRM
          this._emitEvent({ "obj" : node }, "collapse");
          this["after_close"](node);
        }
        else {                                          // end expand
          op = 'e' ;       // expand op
          $ul[0].style.display = "block";
          $ul.css('max-height', '') ;
          this._emitEvent({ "obj" : node }, "expand");
//        this["after_open"](node);
        }

        if (this._overflow) {
          $ul.css('overflow', this._overflow) ;         // restore, 'hidden' was set in _slide()
          this._overflow = null ;
        }

        this._removeFromBusyStack(op, node.attr("id"));
        this._resolveIfBusyStackEmpty() ;
     },


     /** Collapses a node and all its descendants.  Triggers a "collapseall" event.
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node, or a
       *                       selector pointing to the element whose descendants are to be collapsed.
       *                       If omitted , or set to -1, all nodes in the tree are collapsed.
       * @param {boolean=} anim - Set to true (or omit) if all nodes are to be collapsed with animation
       *                          (if a non-zero duration is defined or defaulted). Default is true.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
     */
     collapseAll : function(node, anim)
     {
        var origTarg = node? node : -1 ;
        var _this = this;

        if (this._data.core.locked) {
          return ;
        }

        node = node ? this._getNode(node) : this._$container;
        if (node && origTarg !== -1) {
          origTarg = node ;
        }
        if (!node || origTarg === -1)  {
          node = this._$container_ul;
        }

        if (node.hasClass(TreeUtils._OJ_DISABLED)) {
          return ;
        }

        var subject ;
        if (origTarg !== -1 && this["isExpanded"](node)) {
          subject = node[0] ;
        }

        var objs  =  node.find("li.oj-expanded") ;     // find child nodes that are open
        if (objs.length)  {
          objs.each(function ()  {
                        _this["collapse"](this, !anim);
                    });
        }

        if (subject) {                               // if subject node is also expanded
          this["collapse"](subject, !anim) ;         // include it in the list
          objs.splice(0, 0, subject) ;
        }

        if (objs.length) {
          this._emitEvent({"obj" : objs, "targ" : origTarg }, "collapseAll");
        }
     },


     /** Expands a collapsed parent node, so that its children are visible. Triggers an "expand" event.
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element to be expanded.
       * @param {boolean=} skipAnim - Set to true to suppress node expansion animation (if a
       *                        non-zero duration is defined or defaulted). Default is false.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     expand : function(node, skipAnim)
     {
        this._expand(node, false, skipAnim) ;
     },

     /** May be used as a getter of setter.  If no argument is supplied, the method returns an
       * array of nodes currently expanded. (An empty array implies that no nodes are expanded.)
       * If an array of nodes is supplied as an argument, the specified nodes are expanded.
       * @param {Array=} nodes - Omit to use as a getter, or specify an array of nodes to be
       *                 expanded.  Nodes may be defined as elements, id strings, jQuery wrapped nodes, or
       *                 selectors pointing to the elements to be expanded.
       * @param {boolean=} skipAnim - Set to true to suppress node expansion animation (if a non-zero
       *                        duration is defined or defaulted). Default is false.


       * @return {Object | null} A jQuery wrapped array of nodes if used as a getter, else null
       *                        if used as a setter.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     expanded : function(nodes, skipAnim)
     {
        var exlr,
            exlen,
            _this = this;

        if (nodes && $.type(nodes) === "array")  {    // setter

          if (this._data.core.locked) {
            return null ;
          }

          exlen = nodes.length ;
          $.each(nodes, function (i, val) {
                          _this._expand(val, false, skipAnim);
          });
          return null ;
        }
        else  {     // getter

          nodes = this._$container_ul.find("li.oj-expanded");
          exlen  = nodes.length ;
          exlr   = [] ;
          for (var n = 0; n < exlen; n++)  {
            exlr.push(nodes[n]);
          };
          return $(exlr) ;
        }
     },


     /** Expands a node and all its descendants.  Triggers an "expandall" event.
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element whose descendants are to be expanded.
       *                        If omitted , or set to -1, all nodes in the tree are expanded.
       * @param {boolean=} anim - Set to true (or omit) if all nodes are to expanded with animation
       *                        (if a non-zero duration is defined or defaulted). Default is true.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     expandAll : function(node, anim)
     {
        this._expandAll(node, anim) ;
     },

     /** Expands a node if collapsed, or collapses a node if expanded. Triggers an "expand" or "collapse" event.
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element to be expanded/collapsed.
       * @param {boolean=} skipAnim - Set to true to suppress node expand/collapse animation (if a
       *                        non-zero duration is defined or defaulted). Default is false.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     toggleExpand : function (node, skipAnim)
     {
        node = this._getNode(node);
        if ((! node) || (! node.length) || (node === -1)) {
          return ;
        }
        if (node.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
          return ;
        }

        if (node.hasClass(TreeUtils._OJ_COLLAPSED))  {
          return this["expand"](node, skipAnim);
        }
        if (node.hasClass(TreeUtils._OJ_EXPANDED))  {
          return this["collapse"](node, skipAnim);
        }
     },


     /** Deselects a node. Triggers an "optionChange" event for options property "selection".
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element to be deselected.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     deselect : function (node)
     {
        var prevSelections =  this.options["selection"].slice(0) ;  // make new array for optionChange event

        this._deselect(node) ;
        this._fireOptionChangeEvent("selection", prevSelections, null, null) ;
     },


     /** Deselects all selected nodes. If optional argument "context" is specified, only the selected
       * nodes within that context will be selected. Triggers an "optionChange" event for options property
       * "selection".
       * @param {HTMLElement | Object | string=} context - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to an element within the tree.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     deselectAll : function(context)
     {
       if (this._data.core.locked) {
         return ;
       }

       var prevSelections =  this.options["selection"].slice(0) ;  // make new array for optionChange event

       this._deselectAll(context) ;
       this._fireOptionChangeEvent("selection", prevSelections, null, null) ;
     },

     /** Selects a node. Triggers an "optionChange event for options property "selection".
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                       or a selector pointing to the element to be selected.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     select : function (node)
     {
        this._select(node, true) ;
     },


     /** Selects a node if deselected, or deselects a node if selected. Triggers an "optionChange"
       * event for options property "selection".
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element to be expanded/collapsed.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     toggleSelect : function (node)
     {
        node = this._getNode(node) ;
        if (! node.length)  {
           return false;
        }

        if (node.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
          return false ;
        }

        if (this._isSelected(node))  {
          this["deselect"](node);
        }
        else  {
          this._select(node, true);
        }
        return true ;         // selection was toggled
     },

     /** Returns true if the node is collapsed, else false.
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @return {boolean} true if the node is collapsed, else false.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     isCollapsed : function(node)
     {
       node = this._getNode(node);
       return node && node !== -1 && node.hasClass(TreeUtils._OJ_COLLAPSED);
     },

     /** Returns true if the node is expanded, else false.
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @return {boolean} true if the node is expanded,  else false.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     isExpanded : function(node)
     {
       node = this._getNode(node);
       return node && node !== -1 && node.hasClass(TreeUtils._OJ_EXPANDED);
     },

     /** Returns true if the node is a leaf node (that is, it has no children), else false.
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @return {boolean} true if the node is a leaf node, else false.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     isLeaf : function(node)
     {
       return this._isLeaf(node) ;
     },

     /** Returns true if the node is selected, else false.
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @return {boolean} true if the node is selected,  else false.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     isSelected : function(node)
     {
       return this._isSelected(node) ;
     },


     /** Creates a new node and adds it to the tree.  Triggers a "create" event.
       *
       * @param {HTMLElement | Object | string} refnode - specifies the node that the new node will be
       *                       placed in, or next to, depending on the "position" argument. Can be a
       *                       DOM element, a jQuery wrapped node, or a selector pointing to the element.
       *                       If there is no reference node (because the tree is empty), specify null or undefined
       *                       (or -1).
       * @param {string | number} position - specifies the position of the newly created node in relation
       *                       to the "refnode" specified by the first argument.  Can be a string : "before",
       *                       "after", "inside", "first", "last", or a zero-based index to position the
       *                       new node at a specific point among the children of "refnode".
       * @param {Object | Array} data - An object or array of objects containing data to create new node(s). The object
       *                       properties are the same as for defining a JSON node:<br>
       *                       "attr" - an object of attribute name/value pairs (at least an "id" property should
       *                       be defined).<br>
       *                       "title" - a string used for the visible text of the node.<br><br>
       * <code>
       * var new Node = { "title" : "My Title", "attr" : { "id": "myid" } };
       * </code>
       * @return {Object} Returns the jQuery wrapped node(s) created from the 'data' argument.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     create : function(refnode, position, data)
     {
       var node, ar, i ;
       ar = $.isArray(data) ;
       if ((!ar) || (data.length == 1)) {
         node = this._createNode(refnode, position, ar? data[0] : data);   // single node
         // Check if this node has a selected tag applied temporarily by dnd dragStart,
         // and if so apply the selection state to the node
         this._getDndContext()._dndFinishSelection(node) ;
         return node ;
       }

       // Handle multiple nodes
       ar = [] ;
       if (position != 'before') {
         // Recreate in reverse order to maintain tree order because of reuse of 'position'.
         for (i = data.length -1; i >= 0; i--) {
            node = this._createNode(refnode, position, data[i]) ;
            ar.unshift(node[0]) ;
            // if parseJson() applied the select class, select the node
            this._getDndContext()._dndFinishSelection(node) ;
         }
       }
       else {
         for (i = 0; i <data.length; i++) {
            node = this._createNode(refnode, position, data[i]) ;
            ar.push(node[0]) ;
            this._getDndContext()._dndFinishSelection(node) ;
         }
       }

       return $(ar) ;
     },

     /**  Removes a node. Triggers a "remove" event.
       *
       *  @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       *  @return {Object | boolean}   The jQuery wrapped node used as an argument.
       *  @expose
       *  @public
       *  @instance
       *  @memberof oj.ojTree
       */
     remove : function(node)
     {
       node = this._getNode(node);

       if ((!node.length) || node.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
         return false;
       }

       var rslt = this._emitEvent({"obj" : node, "func" : "remove"}, "before") ;
       if (typeof rslt == "boolean" && (!rslt)) {
         return false ;
       }

       this.__rollback();
       var p    = this._getParent(node),
           prev = $([]),
           t = this,
           sib = this._getPrev(node) ;

       node.each(function () {
           prev = prev.add(t._getPrev(this));
       });

       node = node.detach();

       if (p !== -1 && p.find("> ul > li").length === 0)  {
         p.removeClass("oj-expanded oj-collapsed")
          .addClass(OJT_LEAF)
          .removeAttr(WA_EXPANDED);
       }

       this._cleanNode(p);
       this._emitEvent({ "obj" : node, "prev" : sib, "parent" : p }, "remove");

       return node ;
     },


     /** Returns the title of the specified node
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                       or a selector pointing to the element.
       * @return {string | boolean} The text string title of the node.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     getText : function(node)
     {
       node = this._getNode(node) ;
       if (!node.length) {
          return false;
       }

       node = node.children("a:eq(0)");
       node = node.find("span:eq(0)") ;
       return node[0].textContent;

     },

     /** Renames a node title.
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @param {string=} text - The new text string.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     rename : function(node, text)
     {
       text = this._escapeHtml(text) ;
       this._renameNode(node, text) ;
     },

     /** Sets the specifed node as the current node of interest (e.g. a mouse-over).  Triggers a "hover" event.
       *
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     hover : function(node)
     {
       this._hover(node) ;
     },


     /** Removes the "hover" state of the currently hovered (i.e. active) node.  Triggers a "dehover" event.
       *
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     dehover : function()
     {
       this._dehover() ;
     },

     /** Returns the full path to a node, either as an array of ID's or node names, depending on
       * the value of the "idMode" argument.<p>
       * e.g. Given a node with Id 'Node1' at the root level, with a child 'Node2' which has a child 'Node3',
       * then:<p>
       * <span class="code-caption">$tree.ojTree("getPath", "#Node3", true) ;</span>
       * <p>
       * will return:<p>
       * <span class="code-caption">["Node1", "Node2", "Node3"]</span>
       * <br>
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                       or a selector pointing to the element.
       * @param {boolean=} idMode - Set to true (or omit) to return ID's from the node attribute
       *                       "id"), or false to return the names (i.e. text titles).  Default is true.
       *
       * @return {Array | boolean} An array of node ID's or names. If the node is not found, false is returned.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     getPath : function(node, idMode)
     {
        var p     = [],
            _this = this;

        node = this._getNode(node);
        if (node === -1 || !node || !node.length)  {
          return false;
        }

        node.parentsUntil(".oj-tree", "li").each(function () {
               p.push( idMode ? this.id : _this["getText"](this) );
        });

        p.reverse();
        p.push( idMode ? node.attr("id") : this["getText"](node) );

        return p;
      },

     /** Returns the jQuery wrapped top outer &lt;ul&gt; element of the tree.
       *
       * @return {Object} The jQuery wrapped &lt;ul&gt; element of the tree.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     getRoot : function()
     {
        return this._$container.children("ul:eq(0)");    // return the top <ul>
     },

     /**  Refreshes the tree or a node.
       *
       *  @param {HTMLElement | Object | string | number=} node - If -1 is specified (or the argument is omitted),
       *                   the whole tree is refreshed.  Alternatively, a specific node to be refreshed can be
       *                   supplied. Can be a DOM element, a jQuery wrapped node, or a selector pointing to the element.
       *  @expose
       *  @public
       *  @instance
       *  @memberof oj.ojTree
       */
     refresh : function(node)
     {
       this._super() ;

        if (this._data.core.locked) {
          return ;
        }

       this._refresh(node? node : -1) ;
     },


     /**  Moves (or copies) a node within a tree, or from one tree to a different tree.
       *
       *  @param {HTMLElement | Object | string | number} node  The node to be moved. Can be a DOM element,
       *            a jQuery wrapped node, or a selector pointing to the element.
       *  @param {HTMLElement | Object | string | number} refnode  The reference node for the move (see "position"
       *            argument below). Can be a DOM element, a jQuery wrapped node, or a selector pointing to the
       *            element. If the receiving tree ie empty and there can be no reference node, null or undefined
       *            (or -1) may be specified.
       *  @param {string | number} position  The position of the moved node relative to the reference node refnode.
       *            Can be "before", "after", "inside", "first", "last", or the zero-based index to position the node at a
       *            specific point among the reference node's current children.
       *  @param {boolean=} iscopy  Omit or specify false for a move operation, or true for a copy.
       *  @expose
       *  @public
       *  @instance
       *  @memberof oj.ojTree
       */
     move : function(node, refnode, position, iscopy)
     {
        this._moveNode(node, refnode, position, iscopy);
     },

     /**
       *  Returns the user classified node type applied to the node in the
       *  <span class="code-caption">"types"</span> option.
       *  @return {string |boolean}  The node's type. If no types have been defined in the tree options,
       *                             false is returned.
       *  @expose
       *  @public
       *  @instance
       *  @memberof oj.ojTree
       */
     getType : function(node)
     {
         return this._getType(node) ;
     },

     /**
       *  Sets the "type" attribute of the node using a type defined in the
       * <span class="code-caption">"types"</span> option.
       *  @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                       or a selector pointing to the element.
       *  @param {string} str - The type.
       *  @return {boolean}  true if the change was successful, else false.
       *  @expose
       *  @public
       *  @instance
       *  @memberof oj.ojTree
       */
     setType : function(node, str)
     {
         return this._setType(node, str) ;
     },

     // @inheritdoc
     getNodeBySubId: function(locator)
     {
        if (! locator)  {
          return this.element ? this.element[0] : null;
        }

        return this._processSubId(locator)
     },

     /**
       * {@ojinclude "name":"getSubIdByNodeDesc"}
       *
       * @expose
       * @memberof oj.ojTree
       * @instance
       * @since 2.1.0
       * @ignore
       *
       * @param {!Element} node {@ojinclude "name":"getSubIdByNodeNodeParam"}
       * @returns {Object|null} {@ojinclude "name":"getSubIdByNodeReturn"}
       *
       * @example <caption>{@ojinclude "name":"getSubIdByNodeCaption"}</caption>
       * {@ojinclude "name":"getSubIdByNodeExample"}
       */
     getSubIdByNode: function(node)
     {
        return this._getSubIdFromNodeElem(node) ;
     },

     /**
       * Returns a context object for the specified tree node. This includes the subid for the 
       * node as the subId property.
       * Additional Tree component information is also included. </br>
       *
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       * @param {HTMLElement | Object | string | number} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to a node element.
       * @return {Object|null} An object containing node context data, or
       *                                <span class="code-caption">null</span> if not found.
       * <table class="params"<tr><th>Property</th><th>Type</th><th>Description</th></tr>
       *  <tr><td>subId</td><td>string</td><td><span class="code-caption">"oj-tree-node"</span> or
       *                                 <span class="code-caption">"oj-tree"</span>.</td></tr>
       *  <tr><td>item</td><td>object</td><td>the tree node element.</td></tr>
       *  <tr><td>node</td><td>object</td><td>the jQuery wrapped node.<br>DEPRECATED - please use <span class="code-caption">item</span>.</td></tr>
       *  <tr><td>leaf</td><td>boolean</td><td><span class="code-caption">true</span> if leaf node, else
       *                                 <span class="code-caption">false</span> if a parent node.</td></tr>
       * </table>
       <p>
       */
     getContextByNode: function(node)
     {
        var  bNode, bTree ;

        node  = (node == null || node == undefined)? 0 : node ;  // don't give getNode() null/undefined
        node  = this._getNode(node) ;  // validate node reference (returns jQ wrapped node, -1, null)
        bNode = (node != -1 && (node.length > 0)) ;
        bTree = (node == -1) ;

        if (bNode || bTree) {
          return {
                   "subId" : bNode ? OJT_NODE : OJT_TREE,
                   "item"  : bNode? node[0] : null,               // <li>
                   "node"  : bNode? node : false,                 // jQuery wrapped <li>
                   "leaf"  : bNode? this._isLeaf(node) : false    // true/false
                 } ;
        }
        return null ;
     },

     /**
       * Returns the parent node of the node specified.
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @return {Object | null}   The jQuery wrapped parent node, or null if <span class="code-caption">node</span>
                                   is a top level node.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     getParent : function (node)
     {
        var l = this._getParent(node) ;
        return (l && l.length > 0)? l : null ;
     },

     /**
       * Returns the previous sibling of the node specified.
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @return {Object | null}   The jQuery wrapped sibling node, or null if there is no previous sibling.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     getPrevSibling : function (node)
     {
        var l = this._getPrev(node, true) ;
        return  (l && l.length > 0)? l : null ;
     },

     /**
       * Returns the next sibling of the node specified.
       * @param {HTMLElement | Object | string} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element.
       * @return {Object | null}   The jQuery wrapped sibling node, or null if there is no next sibling.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     getNextSibling : function (node)
     {
        var l = this._getNext(node, true) ;
        return (l && l.length>  0)? l : null ;
     },

     /**
       * Returns the children of the node specified.
       * @param {HTMLElement | Object | string | number} node - Can be a DOM element, a jQuery wrapped node,
       *                        or a selector pointing to the element. May also be specified as -1 or omitted to
       *                        indicate the tree, in which case the top level children of the tree are returned.
       * @return {Object | null}   The jQuery wrapped array of child nodes, or null if there are no children.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     getChildren : function (node)
     {
        node = node? node : -1 ;
        var c = this._getChildren(node) ;

        c = (c && c.length)?  c.not("." + TreeUtils._OJ_TEMPNODE) : c ;
        return (c && c.length > 0)?  c :  null ;
     },

     /** Scrolls a node into view.
       * @param {Object} obj - An object containing the scroll properties.
       * <table style="border-collapse:collapse;border:1px solid"><tr style="background-color:#eee"><th>Property</th><th>Description</th></tr>
       *  <tr><td><em>"node"</em></td><td>Can be a DOM element, a jQuery wrapped node,
       *                       or a selector pointing to the element to be selected. Any other non-node reference will
       *                       default to the first tree node.</td></tr>
       * </table>
       * @param {string=} alignTo - Specify <span class="code-caption">'top'</span> if the top of the
       *                         <span class="code-caption">node</span> is to be aligned to the top of
       *                         the container. Specify <span class="code-caption">'bottom'</span> if the
       *                         bottom of the <span class="code-caption">node</span> is to be aligned to
       *                         the bottom of the container. Any other value (or if omitted) will be treated
       *                         as <span class="code-caption">'bottom'</span>.<br>
       *                         Note that placement is only honored if feasible. For example, when scrolling a
       *                         node to the top, movement will stop should the bottom node scroll into view -
       *                         placement to the requested top will not be completely honored in this case.
       * @param {boolean=} setActive - If <span class="code-caption">true</span>, makes
       *                         <span class="code-caption">node</span> the active node (i.e. keyboard focus)
       *                         after scrolling into view. If omitted or set to <span class="code-caption">false</span>,
       *                         node is not made active.
       * @expose
       * @public
       * @instance
       * @memberof oj.ojTree
       */
     scrollIntoView : function (obj, alignTo, setActive)
     {
        var node, toTop ;

        if (! obj) {
          node = -1 ;
        }
        else {
          node = obj['node'] ;
          node  = this._getNode(node);
        }
        if (node == -1) {
          node = this._$container.find("> ul > li:first-child");
        }
        if ((!node) || (! node.length)) {
          return ;
        }

        toTop     = (alignTo == 'top') ;     // default is 'bottom'
        setActive = !! setActive ;           // default is false

        if (setActive) {
          this._$container_ul.focus() ;
          this["hover"](node) ;
        }
        node[0].scrollIntoView(toTop) ;

        return this ;
     },

     //-----------------------------------------//
     //     Internal API's and data             //
     //-----------------------------------------//

     /**
       *  Called the first time the widget is called on an element.
       *  @private
       */
     _ComponentCreate: function ()
     {
        this._super() ;

        this._index         = this._newIndex() ;                    // index for this instance
        this._elemId        = this.element.attr("id") ;             // tree element id
        if (this._elemId === undefined) {
          this._elemId = "oj-tree-" + this._getIndex() ;
          this.element.attr("id", this._elemId) ;
        }
        this._$container    = this.element ;                        // the containing <div>
        this._$container_ul = null ;                                // the containing <ul>
        this._data          = {} ;                                  // working data
        this._tds           = null ;                                // Tree DataSource
        this._isRtl         = this._GetReadingDirection() === "rtl";
        this._isTouch       = oj.DomUtils.isTouchSupported() ;
        this._initTree() ;                                          // initialize data structures
        this._animDuration  = this._getAnimDuration() ;             // get css $treeAnimationDuration value
        if (this._animDuration) {
          var ai = oj.AgentUtils.getAgentInfo() ;
          this._isSafari = (ai["browser"] === oj.AgentUtils.BROWSER.SAFARI) ;
          this._proxyTransitionEndHandler = this._transitionEndHandler.bind(this);
        }
        this._busyStack = [] ;
        this._start() ;                                             // build/display the tree
     },

     /**
       *  Widget is being destroyed.
       *  @private
       */
     _destroy : function()
     {
        this._resolveBusyContext();
        this._clearTree() ;    // Clean out the DOM.  After this, the tree markup has
                               // been removed from the div, and all event handlers
                               // removed.

        // If the tree was constructed from existing user markup found in the div,
        // reinstate that markup to reset the div to its original state pre tree create.
        if (this._data.html.markup_ul) {
          this._unTransformElemIds(this._data.html.markup_ul);   // reset to original user id's
          this._$container.append(this._data.html.markup_ul) ;        //@HTMLUpdateOK
          this._data.html.markup_div.remove() ;
          this._data.html.markup_div        = false ;
          this._data.html.useExistingMarkup = false ;
        }

        this._data.ds.type = DS_NONE ;
        this._super() ;
     },


     /**
       *  Handle an option change.
       *  Called by $(selector).ojtree("options", "prop", value)
       *  @private
       */
     _setOption: function (key, newval, flags)
     {
        var _this = this ;
        var prev ;
        var val ;

        if (key === "selection") {
          if (typeof newval == "string") {
            newval = [newval] ;
          }
          else {
            newval = (newval && newval.length)? newval : [] ;
          }
          this._data.core.suppressSelectEvent = true ;    // don't want select/deselect to fire selection change
          this._handleSelectionOptionChange(newval) ;     // apply selection/deselection
          this._data.core.suppressSelectEvent = false ;
          newval = this._getSelected() ;
          // optionChange event will be fired when super is called below.
        }
        else if (key === "selectionMode") {
          if (newval === "none") {
            val = 0 ;
          }
          else if (newval === "single") {
            val = 1;
          }
          else if (newval === "multiple") {
            val = -1 ;
          }
          else {
            val = 0 ;
            newval = "none"
          }
          if (val != _this._data.core.selectMode) {
            _this._data.core.selectMode = val ;
          }
        }
        else if (key === "icons") {
           if ($.type(newval) == "boolean") {
             if (newval != _this._data.themes.icons) {
               _this._data.themes.icons = newval ;
               newval? _this._showIcons() : _this._hideIcons() ;
             }
           }
        }                                // end "core/ui" options
        else if (key === "contextMenu") {
           this._clearMenu() ;
           if (newval) {
             this._initMenuOpts() ;
             this._initMenu(newval) ;
           }
        }                                // end "contextMenu"
        else if (key === "disabled")  {
           this._handleDisabledChanged(newval) ;
        }                            // end "disabled"
        else if (key === "data") {    //TDO
          this._super(key, newval, flags);
          this._initDSOpts() ;
          this._initDataSource();
          this._initExpandedOpts() ;     // reset the nodes to be initially expanded
          this._loadNodes() ;            // start node loading from the datasource
          return ;                       // super has already been called
        }
        else if (key === "dnd") {
          this._super(key, newval, flags);
          this._getDndContext().handleDnDOptionChange(newval) ;
          return ;                       // super has already been called
        }
        else if (key == "emptyText") {
          newval = this._escapeHtml(newval) ;   // escape the text for security
        }
        this._super(key, newval, flags);
     },


     /**
       *  Compare two option value(s) for equality.  Returns false if they are different.
       *  @private
       */
     _CompareOptionValues : function (option, value1, value2)
     {
        if (option === "selection") {
          return  this._compareSelectionValues(value1, value2) ;
        }

        return this._superApply(arguments) ;
     },


     /**
       *  Compare selection values, return true if the same, else false.
       *  @private
       */
     _compareSelectionValues : function(value1, value2)
     {
       var  i, j, id, len, len2 ;
       var  bDiff   = false,
            bInList = false ;

       len  = (value1 && value1.length)? value1.length : 0 ;
       len2 = (value2 && value2.length)? value2.length : 0 ;

       if (len !== len2) {
         bDiff = true ;
       }
       else if (len == 0 && len2 === 0) {
         bDiff = false ;
       }
       else {
         for (i = 0; i < len; i++) {
            id = $(value1[i]).attr("id") ;
            bInList = false ;
            for (j = 0; j < len2; j++) {
               if (id == $(value2[j]).attr("id")) {
                 bInList = true ;
                 break ;
               }
            }

            if (! bInList) {        // if not in list, values are different
              bDiff = true ;
              break ;
            }
         }
       }

       return (! bDiff) ;
     },

     /**
       *  Add animation busyState
       *  @private
       */
     _addBusyState : function(description)
     {
        if (! this._animationResolve) 
        {
          var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
          this._animationResolve = busyContext.addBusyState( {"description" : description} );
        }
     },

     /**
       *  Resolve BusyContext
       *  @private
       */
     _resolveBusyContext : function()
     {
        if (this._animationResolve)
        {
          this._animationResolve() ;
          this._animationResolve = null ;
        }
     },

     /**
       *  Check if entry is on the busy stack. Returns the index, or -1 if not found.
       *  @private
       */
     _isOnBusyStack : function(op, id)
     {
       var a   = this._busyStack,
       len = a.length,
       i,
       ret = -1 ;

       for (i = 0; i < len; i++)
       {
         if ((a[i].id === id) && (a[i].op === op))
         {  
           ret = i ;
           break ;
         }
       }

       return ret ;
     },


     /**
       *  Remove entry from the busy stack. Returns the index, or -1 if not found.
       *  @private
       */
     _removeFromBusyStack : function(op, id)
     {
       var i = this._isOnBusyStack(op, id) ;

       if (i >= 0)
       {
         this._busyStack.splice(i, 1) ;
       }

       return i ;
     },

     /**
       *  If busy stack is empty can resolve busy
       *  @private
       */
    _resolveIfBusyStackEmpty : function()
    {
      if (this._busyStack.length === 0)
      {
        this._resolveBusyContext() ;    
      }
    },

     /**
       *  Clears out the tree dom
       *  @private
       */
     _clearTree : function()
     {
        if (this._TreeDndContext) {
          this._TreeDndContext._closedown() ;   // clear DnD support
          this._TreeDndContext = null ;
        }

        var  n  = this._getIndex();

        this._$container
            .unbind(".oj-tree")
            .undelegate(".oj-tree")
            .removeData("oj-tree-instance-id")
            .find("[class^='oj-tree']")
            .addBack()
            .attr("class", function ()
                    {
                      return this["className"].replace(/oj-tree[^ ]*|$/ig,'');
                    });

        var cl = this._$container.attr("class") ;    // if jQuery has left us with a
        cl = cl.trim();                              // class of blanks, remove it.
        if (cl.length === 0) {
          this._$container.removeAttr("class") ;
        }

        _removeKeyFilter(this._$container_ul) ;      // remove keyboard listener because
                                                     // _$container_ul will be removed
        $(document).unbind(".oj-tree-" + n)
                   .undelegate(".oj-tree-" + n);

        // Remove the constructed tree from the DOM.
        this._$container_ul.remove() ;
        this._$container_ul = null ;
     },


//   /*
//     *  Returns a jQuery wrapped tree node.  obj can be a selector pointing
//     *  to an element within the tree, a DOM node, or a jQuery wrapped node.  If -1 is used
//     *  (indicating the whole tree), -1 is returned.
//     *  @private
//     */
//                                         Not used in V1
//   _getNodeCore : function(obj)
//   {
//       var $obj = $(obj, this._$container);
//
//       if ($obj.is(".oj-tree") || obj == -1)  {
//          return -1;
//       }
//       $obj = $obj.closest("li", this._$container);
//       return $obj.length ? $obj : false;
//   },


     /**
       *  Returns a jQuery wrapped tree node.  obj can be a selector pointing
       *  to an element within the tree, a DOM node, or a jQuery wrapped node.  If obj is -1
       *  (indicating the whole tree), -1 is returned.  If obj is null, all currently selected
       *  are returned.
       *  @private
       */
     _getNode :  function (obj, allowMultiple)
     {
       if (typeof obj === "undefined" || obj === null) {
         return allowMultiple ? this._data.ui.selected : this._data.ui.lastSelected;
       }

       var $obj = $(obj, this._getContainer());

       if ($obj.is(".oj-tree") || obj === -1) {
          return -1;
       }

       $obj = $obj.closest("li", this._getContainer());
       return $obj.length ? $obj : false;
     },

     /**
       *  Returns the node previous to the supplied obj.  obj can be a selector pointing
       *  to an element within the tree, a DOM node, or a jQuery wrapped node.  If -1 is used
       *  (indicating the whole tree), -1 is returned.  If arg strict is true, only immediate
       *  siblings are considered, and false will be returned if there is no previous sibling.
       *  If strict is false or omitted and obj has no previous siblings (i.e it is the first
       *  child of its parent), the parent is returned.
       *  @private
       */
     _getPrev  : function (obj, strict)
     {
        obj = this._getNode(obj);
        if (obj === -1) {
          return this._$container.find("> ul > li:last-child");
        }
        if (!obj.length) {
          return false;
        }
        if (strict) {
         return (obj.prevAll("li").length > 0) ? obj.prevAll("li:eq(0)") : false;
        }

        if (obj.prev("li").length) {
            obj = obj.prev("li").eq(0);
            while (obj.hasClass(TreeUtils._OJ_EXPANDED)) {
               obj = obj.children("ul:eq(0)").children("li:last");
            }
           return obj;
        }
        else {
           var o = obj.parentsUntil(".oj-tree","li:eq(0)");
           return o.length ? o : false;
        }
     },

     /**  Returns the node after the supplied obj.  obj can be a selector pointing
       *  to an element within the tree, a DOM node, or a jQuery wrapped node.  If -1 is used
       *  (indicating the whole tree), -1 is returned.  If arg strict is true, only immediate
       *  siblings are considered, else if the obj has no more siblings (i.e is the last
       *  child of its parent), the parent's next sibling is returned.
       *  @private
       */
     _getNext  : function (obj, strict)
     {
        obj = this._getNode(obj);
        if (obj === -1)  {
          return this._$container.find("> ul > li:first-child");
        }
        if (!obj.length) {
           return false;
         }
        if (strict)  {
           return (obj.nextAll("li").length > 0) ? obj.nextAll("li:eq(0)") : false;
        }

        if (obj.hasClass(TreeUtils._OJ_EXPANDED))  {
          return obj.find("li:eq(0)");
        }
        else if(obj.nextAll("li").length > 0)  {
          return obj.nextAll("li:eq(0)");
        }
        else  {
           return obj.parentsUntil(".oj-tree","li").next("li").eq(0);
        }
     },


     /**
       *  Returns the parent node of the supplied obj.  obj can be a selector pointing
       *  to an element within the tree, a DOM node, or a jQuery wrapped node.
       *  child of its parent), the parent is returned.
       *  @private
       */
     _getParent  : function (obj)
     {
        obj = this._getNode(obj);
        if (obj == -1 || !obj.length)  {
          return false;
        }
        var o = obj.parentsUntil(".oj-tree", "li:eq(0)");
        return o.length ? o : -1;
     },

     /**
       *  Returns the child nodes of the supplied obj, or false if no children or failure.
       *  obj can be a selector pointing to an element within the tree, a DOM node, or a
       *  jQuery wrapped node.  obj can be -1 to indicate the whole tree.
       *  @private
       */
     _getChildren   : function (obj)
     {
        obj = this._getNode(obj);
        if (obj === -1) {
          return this._$container.children("ul:eq(0)").children("li");
        }
        if (!obj.length) {
          return false;
        }
        return obj.children("ul:eq(0)").children("li");
     },

     /**
       *  @private
       */
     _isLeaf : function(node)
     {
       node = this._getNode(node) ;
       return node && node !== -1 && node.hasClass(OJT_LEAF);
     },

     /**
       *  Returns the jQuery wrapped <li> element for a node specification
       *  @private
       */
     _getNodeElem : function(node)
     {
        var el  = this._$container_ul.find(node) ;
        var ret =  false ;

        if (el && el.length && el.length > 0) {
          if ($(el).is("li")) {
            ret = el ;
          }
        }

        return ret ;
     },

     /**
       *  Returns true if the node is selected, else false.
       *  @private
       */
     _isSelected : function(node)
     {
       var n = this._getNode(node) ;
       var r = false ;

       if (n && n.length && this._data.ui.selected) {
         r = (this._data.ui.selected.index(n) >= 0);
       }
       return  r ;
     },

     /**
       *  Sets the widget instance for the supplied node as the member
       *  ot in the supplied object o.
       *  @instance
       *  @private
       */
     _reference : function(node, o)
     {
       //if ((! node) || (! node.length) || (! node.parents) || (typeof node.parents != "function")) {
       //        this._trace("_REFERENCE ERROR " + node) ;
       //}

       var div  = node.parents("div").eq(0) ;
       var ctor = oj.Components.__GetWidgetConstructor(div) ;

       if (ctor && o) {
         ctor("getCI", o) ;
       }
       return this ;
     },

//   /**
//     *  Get the data block for the tree component containing the specified
//     *  node.  Returned as member db of the supplied object.
//     *  @instance
//     *  @private
//     */
// Not used in V1
//   _referenceData : function(node, o)
//   {
//      var div   = node.parents("div").eq(0) ;
//      var ctor  = oj.Components.__GetWidgetConstructor(div) ;
//
//      if (ctor && o) {
//        ctor("getDB", o) ;
//      }
//   },


     /**
       *  Internal use only - return component instance as member ot
       *  in supplied object.
       *  @instance
       *  @private
       */
     "getCI" : function(o)
     {
       o.ot = this ;
     },

//   /**
//     *  Internal use only - return the data block for a tree's this.
//     *  @instance
//     *  @private
//     */
//  Not used in V1
//   "getDB" : function(o)
//   {
//      o.db = this._data ;
//   },
//

     /**
       *  Add default values to options, unless already defined in options.
       *  @private
       */
     _applyDefaults :  function(to, from)
     {
        if (to != undefined && from != undefined)  {
          $.each(from, function(prop, val) {
                                             if (to[prop] == undefined) {
                                               to[prop] = val ;
                                             }
                                            });
        }
     },

     /**
          Handle change of the disabled option
       *  @private
       */
     _handleDisabledChanged  : function(newval)
     {
        var curState ;

        if (typeof newval === "undefined")  {
          return ;
        }

        curState  = this._$container_ul.hasClass(TreeUtils._OJ_DISABLED) ;
        if (! curState) {
          curState = false ;      // in case undefined
        }

        if (curState != newval) {
          if (newval) {
            this._$container_ul.addClass(TreeUtils._OJ_DISABLED) ;
            this._$container_ul.prop("disabled", true);
          }
          else {
            this._$container_ul.removeClass(TreeUtils._OJ_DISABLED) ;
            this._$container_ul.prop("disabled", false);
          }
//        this._lock(newval) ;
          this._treeDisable(newval) ;
        }
     },

     /**
       *  @private
       */
     _treeDisable : function (lstate) {

        lstate = lstate || false ;
        if (lstate) {
          this._data.core.locked = true;
          this._data.ui.opacity  = this._$container.children("ul").css("opacity") ;
          this._$container_ul.addClass(TreeUtils._OJ_DISABLED).css("opacity","0.9");
        }
        else {
          this._data.core.locked = false;
          this._$container_ul.removeClass(TreeUtils._OJ_DISABLED).css("opacity", this._data.ui.opacity);
        }
     },


     /**
       *  @private
       */
     _isTreeDisabled : function ()
     {
        return  this._data.core.locked ;
     },


     /**
       *  Handles a change to the selection option property. Nodes currently selected that are
       *  not in the new array are deselected.  Nodes in the new array become (or stay) selected.
       *  @private
       */
     _handleSelectionOptionChange : function(newSels)
     {
        var sels   = newSels.slice(0) ;                       // working copy
        var cur    = [] ;                                     // xfer valid selection args to this
        var newvals ;
        var _this  = this ;
        var $elem, id, inList, len, bRemoved = false;

        // Convert the new array to jQuery wrapped node elements (thereby changing any
        // refs like "#myid"). If any node reference is invalid, remove it from the
        // selections array.
        newvals = $.grep(sels, function(node, i) {
                                  $elem = _this._getNodeElem(node) ;
                                  if ($elem) {                       // if node elem exists
                                    cur.push($elem) ;                // note it.
                                    sels[i] = $elem[0] ;             // ensure we store an <li>
                                  }                                  // not a string like "#news"
                                  else {
                                   return false ;                    // remove from array
                                  }
                                  return true ;                      // retain in array
                      });

        //  Deselect any currently selected nodes that are not in the new array.
        var aSelected = this._getSelected() ;
        $.each(aSelected, function(i, node) {
                            len    = cur.length ;
                            id     =  $(node).attr("id") ;
                            inList = false ;
                            for (i = 0; i < len; i++) {
                               if (id == cur[i].attr("id")) {
                                 inList = true ;
                                 break ;
                               }
                            }
                            if (! inList) {
                              _this._deselect(node) ;
                            }
                          }) ;

        this._setSelected(cur, null) ;      //  Select all nodes in the new selection array
     },


     /**
       *  Creates a "prepMoveBlk" object containing details of the impending move, and
       *  stores it in this._data.core.prepMoveBlk
       *
       *  The object contains :
       *    .o   the node being moved
       *    .r   the reference node
       *    .p   the destination position relative to the reference node
       *    .np  the new parent
       *    .oc  the original node if there was a copy
       *    .cy  boolean indicating if the move was a copy
       *    .op  the former parent
       *    .or  the node that was previously in the position of the moved node
       *    .ot  the original tree instance this
       *    .rt  the reference tree instance
       *    .cr  same as .np but if a root node is created this is -1
       *
       *  @private
       */
     _prepare_move :  function (o, r, pos, cb, is_cb)
     {
       var p  = {},                          // new prepMoveBlk object
           _t = this;

       r  = ((r === - 1) || (!r)) ? -1 : this._getNode(r);

       // Get tree component's involved
       this._reference(o, p) ;               // for node being moved
       p.rt = this;                          // for reference node

       p.o  = p.ot._getNode(o);
       p.r  = r ;
       p.p  = (typeof pos === "undefined" || pos === false) ? "last" : pos; // TODO: move to a setting

       if (!is_cb && this._data.core.prepMoveBlk.o                &&
                     this._data.core.prepMoveBlk.o[0] === p.o[0]  &&
                     this._data.core.prepMoveBlk.r[0] === p.r[0]  &&
                     this._data.core.prepMoveBlk.p    === p.p)  {
//       this._emitEvent(this._data.core.prepMoveBlk, "prepare_move", true);
         if (cb)  {
           cb.call(this, this._data.core.prepMoveBlk);
         }
         return;
       }

       if (p.r === -1 || !p.r)  {
         p.cr = -1;

         switch(p.p)
         {
           case "first":
           case "before":
           case "inside":
                          p.cp = 0;
                          break;
           case "after":
           case "last":
                          p.cp = p.rt._$container.find(" > ul > li").length;
                          break;
           default:
                          p.cp = p.p;
                          break;
         }
       }
       else {
         if (!/^(before|after)$/.test(p.p) && !this._is_loaded(p.r)) {
           return this._load_node(p.r, function () {
                                           this._prepare_move(o, r, pos, cb, true);
                                 });
         }
         switch(p.p)
         {
           case "before":  p.cp = p.r.index();
                           p.cr = p.rt._getParent(p.r);
                           break;
           case "after":   p.cp = p.r.index() + 1;
                           p.cr = p.rt._getParent(p.r);
                           break;
           case "inside":
           case "first":
                           p.cp = 0;
                           p.cr = p.r;
                           break;
           case "last":
                           p.cp = p.r.find(" > ul > li").length;
                           p.cr = p.r;
                           break;
           default:
                           p.cp = p.p;
                           p.cr = p.r;
                           break;
         }
       }

       p.np  = p.cr == -1 ? p.rt._getContainer() : p.cr;
       p.op  = p.ot._getParent(p.o);
       p.cop = p.o.index();

       if (p.op === -1)  {
         p.op = p.ot ? p.ot._$container : this._$container;
       }
       if (!/^(before|after)$/.test(p.p) && p.op && p.np && p.op[0] === p.np[0] && p.o.index() < p.cp)
       {
         p.cp++;
       }

       //if(p.p === "before" && p.op && p.np && p.op[0] === p.np[0] && p.o.index() < p.cp) { p.cp--; }
       p.or = p.np.find(" > ul > li:nth-child(" + (p.cp + 1) + ")");
       this._data.core.prepMoveBlk = p;
//     this._emitEvent(this._data.core.prepMoveBlk, "prepare_move", true);

       if (cb) {
         cb.call(this, this._data.core.prepMoveBlk, "prepare_move");
       }
     },


     /**
       *  Checks the prepMoveBlk object, and returns true if the move it
       *  reflects is acceptable, else false.
       *  @private
       */
     _checkMove : function ()
     {
       var obj  = this._data.core.prepMoveBlk,
           ret  = true,
           r ;

       r = obj.r === -1 ? this._$container : obj.r;

       if ((!obj) || (!obj.o) || obj.or[0] === obj.o[0]) {
         return false;
       }
       if (!obj.r) {        // tdo, why is r is false when we use jquery simulate drag
         return false ;
       }

       if (!obj.cy) {
         if (obj.op && obj.np && obj.op[0] === obj.np[0] && obj.cp - 1 === obj.o.index())  {
           return false;
         }

         // Get all li ancestors plus the reference node. If the dragged node
         // is found, end loop and set return to false to indicate move is invalid.
         obj.o.each(function ()
                    {
                      if (r.parentsUntil(".oj-tree", "li").addBack().index(this) !== -1)  {
                        ret = false;
                        return false;
                      }
                    });
       }

       return ret;
     },

     /**
       *  Changes the text title of a node. (Text is already expected to have been escaped
       *  prior to this call.)
       *  @private
       */
     _renameNode : function(node, text)
     {
       var  t ;

       node = this._getNode(node);
       this.__rollback();
       t  = this["getText"](node) ;

       if (node && node.length)
       {
          var rslt = this._emitEvent({"obj"  : node,
                                      "func" : "rename",
                                      "title" : text,
                                      "prevTitle": t
                                     }, "before") ;
          if (typeof rslt == "boolean" && (!rslt)) {
            return ;
          }
       }

       if (node && node.length && this._set_text.apply(this, Array.prototype.slice.call(arguments)))
       {
         this._emitEvent({ "obj" : node, "title" : text, "prevTitle" : t}, "rename");
       }
     },

     /**
       *   Moves a node within the tree
       *   @private
       */
     _moveNode : function (obj, ref, position, is_copy, is_prepared, skip_check)
     {
        if (! is_prepared) {
          obj = this._getNode(obj) ;

          var rslt = this._emitEvent({"obj" : obj, "func": "move"}, "before") ;
          if (typeof rslt == "boolean" && (!rslt)) {
            return false ;
          }
        }

        if ((obj.hasClass && obj.hasClass(TreeUtils._OJ_DISABLED)) || this._data.core.locked) {
          return ;
        }

        if (!is_prepared)  {
          return this._prepare_move(obj, ref, position,
                                    function (p) {
                                       // Before the move/copy occurs, if the
                                       // node is being moved (not copied) to
                                       // a different tree, deselect it.
                                       if ((p.ot !== p.rt) && (!p.cy)) {
                                         p.ot["deselect"](p.o) ;
                                       }
                                       this._moveNode(p, false, false,
                                                      is_copy, true, skip_check);
                                    });
        }

        if (is_copy) {
          this._data.core.prepMoveBlk.cy = true;
        }
        if (!skip_check && !this._checkMove())  {
          return false;
        }

        this.__rollback();
        var o = false;

        if (is_copy)  {
           o = obj.o.clone(true);
           o.find("*[id]").addBack().each(function () {
                                             if (this.id) {
                                               this.id = "copy_" + this.id;
                                             }
                                         });
        }
        else   {
          o = obj.o;
        }

        if (obj.or.length)  {
          obj.or.before(o);                             //@HTMLUpdateOK
        }
        else   {
           if (!obj.np.children("ul").length)  {
             $("<ul />").appendTo(obj.np);              //@HTMLUpdateOK
           }
           obj.np.children("ul:eq(0)").append(o);       //@HTMLUpdateOK
        }

        try {
               obj.ot._cleanNode(obj.op);
               obj.rt._cleanNode(obj.np);
               if (!obj.op.find("> ul > li").length)  {
                obj.op.removeClass("oj-expanded oj-collapsed")
                      .removeAttr(WA_EXPANDED)
                      .addClass(OJT_LEAF)
                      .children("ul")
                      .remove();
               }
        } catch (e) { }

        if (is_copy)  {
           this._data.core.prepMoveBlk.cy = true;
           this._data.core.prepMoveBlk.oc = o;
        }
        var d = $.extend(true, {}, this._data.core.prepMoveBlk) ;
        d["obj"] = obj.o ;
        this._emitEvent(d, "move");
        return this._data.core.prepMoveBlk;
     },


      /**
        *  Returns the "prepMoveBlk" object created by a previous _prepare_move()
        *
        *  The object contains :
        *     .o   the node being moved
        *     .r   the reference node
        *     .p   the destination position relative to the reference node
        *     .np  the new parent
        *     .oc  the original node if there was a copy
        *     .cy  boolean indicating if the move was a copy
        *     .op  the former parent
        *     .or  the node that was previously in the position of the moved node
        *     .ot  the original tree instance
        *     .rt  the reference tree instance
        *     .cr  same as .np but if a root node is created this is -1
        *     .cp  index of ref node
        *     .cr  parent of ref node (-1, or node)
        *     .cop index of node
        *
        *  @private
        */
      _getMoveBlk  : function ()
      {
         return this._data.crrm.prepMoveBlk;
      },


      /**
        *  Returns the "type" attribute of the node.  If not found, returns "default"
        *  @private
        */
      _getType : function (node)
      {
        var n = null ;
        var t = this.options["types"];             // are types defined?

        if (t) {
          n = this._getNode(node);
        }
        return (!n || !n.length) ? false : n.attr(this.options["types"]["attr"]) || "default";
      },

      /**
        *  Sets the "type" attribute of the node.
        *  @return {boolean}  true if the change was successful, else false.
        *  @private
        */
      _setType : function (node, str)
      {
         var  s   = this.options["types"],
              tattr,
              types,
              ret = false ;

         node = this._getNode(node);

         if (s && node && (node != -1) && node.length && str) {
           tattr = s["attr"] ;
           types = s["types"] ;
           if (tattr && types && types[str]) {
             node.attr(tattr, str);
             node.addClass(OJT_TYPE) ;
//           this._emitEvent({ "obj" : node, "type" : str}, "settype", true);
             ret = true ;
           }
         }

         return ret;
      },

      /**
        *  Check if a node has been typed and the type definition inhibits moving.
        *  @return {boolean} false if the node is inhibited from being moved, else true.
        *  @private
        */
      _canTypedNodeMove : function($node, func)
      {
         var types, type, ret = true ;

         types = this.options["types"] ;
         if (types) {
           types = types["types"];
           if (types) { 
             type  = this._getType($node);            // "type" attr name for node
             if (((types[type] && typeof types[type][func] !== "undefined") ||
                 (types["default"] && typeof types["default"][func] !== "undefined")) &&
                 (this._check(func, $node) === false)) {
               ret = false;
             }
           }
         }

         return ret ;
      },

      /**
        *  @private
        */
      _check : function (rule, obj, opts)
      {
         obj = this._getNode(obj);

         var v     = false,
             ty    = this._getType(obj),
             d     = 0,
             _this = this,
             s     = this._getOptions()["types"],
             data  = false;

             if (obj === -1) {
               if (!!s[rule])  {
                 v = s[rule];
                }
               else  {
                 return;
               }
             }
             else   {
               if (ty === false)  {
                 return;
               }

               data = this._data.types.defaults["useData"] ? obj.data("oj-tree") : false;
               if (data && data["types"] && typeof data["types"][rule] !== "undefined")  {
                 v = data["types"][rule];
               }
               else if(!!s["types"][ty] && typeof s["types"][ty][rule] !== "undefined") {
                   v = s["types"][ty][rule];
               }
               else if(!!s["types"]["default"] && typeof s["types"]["default"][rule] !== "undefined")  {
                   v = s["types"]["default"][rule]; }
               }

               if ($.isFunction(v))  {
                   v = v.call(this, obj);
               }

/*  Not used in v1
               var md = this._data.types.defaults["maxDepth"] ;

               if ((rule === "maxDepth") && (obj !== -1) && (opts !== false) &&
                   (this._data.types.defaults["maxDepth"] !== -2) && (v !== 0))  {
                 // also include the node itself - otherwise if root node it is not checked
                 obj.children("a:eq(0)").parentsUntil(".oj-tree","li").each(function (i)
                        {
                           // check if current depth already exceeds global tree depth
                           if ((md !== -1) && (md - (i + 1) <= 0))  {
                             v = 0;
                             return false;
                           }

                           d = (i === 0) ? v : _this._check(rule, this, false);

                           // Check if current node max depth is already matched or exceeded
                           if (d !== -1 && d - (i + 1) <= 0)  {
                             v = 0; return false;
                           }

                           // otherwise - set the max depth to the current value minus current depth
                           if (d >= 0 && (d - (i + 1) < v || v < 0) )  {
                             v = d - (i + 1);
                           }

                           // If the global tree depth exists and it minus the nodes calculated
                           // so far is less than `v` or `v` is unlimited
                           if ((md >= 0) && (md - (i + 1) < v || v < 0) )  {
                             v = md - (i + 1);
                           }
                        });
               }
*/
               return v;
     },


      /**
        * Applies/removes all necessary classes/attributes to the node and
        * its descendants. obj can be a node, or -1 to clean the whole tree.
        * @private
        */
      _cleanNode  : function (obj)
      {
         var dndClasses,
             draggableAttr,
             draggableAttrVal,
             dndContext = this._getDndContext(),
             cons       = oj.TreeDndContext ;

         if (dndContext.isDragEnabled()) {
           dndClasses       =  cons._OJ_DRAGGABLE ;
           draggableAttr    =  "draggable" ;
           draggableAttrVal =  true ;
         }
         if (dndContext.isDropEnabled()) {
           dndClasses       += " " + cons._OJ_VALID_DROP ;
         }

         obj = obj && obj != -1 ? $(obj) : this._$container_ul;
         obj = obj.is("li") ? obj.find("li").addBack() : obj.find("li");

         obj.removeClass(OJT_LAST)
            .addClass(OJT_NODE)
            .addClass(dndClasses)
            .attr(draggableAttr, draggableAttrVal)
            .filter("li:last-child")
            .addClass(OJT_LAST)
            .end()
            .filter(":has(li)")
            .not(".oj-expanded")
            .removeClass(OJT_LEAF)
            .addClass(TreeUtils._OJ_COLLAPSED)
            .attr(WA_EXPANDED, "false");

         obj.not(".oj-expanded, .oj-collapsed")     // clean leaf nodes
            .addClass(OJT_LEAF)
            .children("ul")
            .remove();

         var typeAttr = (this.options["types"])? this.options["types"]["attr"] : false ;

         var nodes = obj.find("li"),         // clean the node and disclosure icon <ins> 's
             disc,
             t ;
         $.each(obj, function (i, val) {
                 t = $(this) ;
                 disc = t.find("> ins") ;
                 if (disc.length > 1) {
                   disc = $(disc[0]) ;
                 }
                 if (t.hasClass(OJT_LEAF)) {
                   disc.removeClass(OJ_DISC);
                   disc.addClass(OJT_ICON);
                 }
                 else {
                   disc.addClass(OJ_DISC);
                   disc.removeClass(OJT_NICON);
                 }

                 if (typeAttr && t.attr(typeAttr)) {      // if user has declared a node type
                   t.addClass(OJT_TYPE) ;                 // give the node our type class.
                 }
         }) ;

         //this._emitEvent({ "obj" : obj }, "clean_node", true);
      },


      /**
        *  Creates a new node
        *  @private
        */
      _createNode :  function (obj, position, js, callback, is_loaded)
      {
         obj = obj || -1 ;

         obj = this._getNode(obj);
         if (obj !== -1 && !obj.length) {
           return null ;
         }

         var d,
             s = this.options,
             tmp;

         position = (typeof position === "undefined") ? "last" : position;

         if (!is_loaded && !this._is_loaded(obj)) {
           this._load_node(obj, function ()
                               {
                                 this._createNode(obj, position, js, callback, true);
                               });
           return null;
         }

         this.__rollback();

         //  Parse the JSON
         js = this._parseJson(js) ;
         if (!js) {
           return null ;              // parse failed!
         }
         js = js.children() ;         // remove the wrapping <ul> added by _parseJson()
         d = $(js[0]) ;

         if (obj === -1) {
           obj = this._$container;
           if (position === "before")  {
             position = "first";
           }
           if (position === "after") {
             position = "last";
           }
         }

         switch(position)
         {
           case "before":  obj.before(d);                                      //@HTMLUpdateOK
                           tmp = this._getParent(obj);
                           break;
           case "after" :  obj.after(d);                                       //@HTMLUpdateOK
                           tmp = this._getParent(obj);
                           break;
           case "inside":
           case "first" :  if (!obj.children("ul").length) {
                            obj.append("<ul />");                              //@HTMLUpdateOK
                           }
                           obj.children("ul").prepend(d);                      //@HTMLUpdateOK
                           tmp = obj;
                           break;
           case "last":    if (!obj.children("ul").length) {
                             obj.append("<ul />");                             //@HTMLUpdateOK
                           }
                           obj.children("ul").append(d);                       //@HTMLUpdateOK
                           tmp = obj;
                           break;
           default:        if (!obj.children("ul").length) {
                             obj.append("<ul />");                             //@HTMLUpdateOK
                           }
                           if (!position) {
                             position = 0;
                           }
                           tmp = obj.children("ul").children("li").eq(position);
                           if (tmp.length)  {
                             tmp.before(d);                                    //@HTMLUpdateOK
                           }
                           else  {
                             obj.children("ul").append(d);                     //@HTMLUpdateOK
                           }
                           tmp = obj;
                           break;
         }

         if (tmp === -1 || tmp.get(0) === this._$container.get(0)) {
           tmp = -1;
         }
         this._cleanNode(tmp);
         this._emitEvent({ "obj" : d, "parent" : tmp }, "create");

         if (callback) {
           callback.call(this, d);
         }

         //  If there is now more than one node, remove the dummy reference
         //  node that might have been added by Drag/Drop.
         if (this._$container_ul[0].childNodes.length > 1) {
           this._$container_ul.find("." + TreeUtils._OJ_TEMPNODE).remove() ;
         }

         return d;
      },


      /**
        *  Expands a collapsed node
        *  @private
        */
      _expand : function (obj, callback, skipAnim)
      {
         obj = this._getNode(obj);
         if ((! obj) || (! obj.length)) {
           return false;
         }

         skipAnim = skipAnim || false ;

         if (obj.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
           return ;
         }

         if (! obj.hasClass(TreeUtils._OJ_COLLAPSED))  {
           if (callback)  {
             callback.call();
           }
           return false;
         }

         // Issue a before event, and cancel the expand if the user vetoes it.

         var rslt = this._emitEvent({"obj" : obj, "func" : "expand"}, "before") ;
         if (typeof rslt == "boolean" && (!rslt)) {
           return ;
         }

        // Get css value $treeAnimationDuration. This is needed because if the value is
        // zero and a transition is started, the transitionend event does not fire and
        // we don't get the opportunity to emit an expand event.

         var dur = skipAnim? 0 : this._animDuration,
             t   = this;

         if (! this._is_loaded(obj))
         {
           obj.children("a").addClass(OJT_LOADING);

           // Load the node. This does not return if an Ajax request is made and ends in failure.
           // will downdate stack and resolve busy if stack is empty.
           this._load_node(obj, function () {
                                   t._expand(obj, callback, skipAnim);
                                }, callback) ;
         }
         else
         {
           if (this.options["expandParents"])  {
             obj.parentsUntil(".oj-tree", ".oj-collapsed").each(function ()
                                   {
                                      t._expand(this, false, true);
                                   });
           }

//         if (skipAnim || dur)  {
//           obj.children("ul").css("display","none");
//         }
           obj.removeClass(TreeUtils._OJ_COLLAPSED)
              .addClass(TreeUtils._OJ_EXPANDED)
              .attr(WA_EXPANDED, "true")
              .children("a")
              .removeClass(OJT_LOADING);
           $(obj.children()[0])
                .removeClass(TreeUtils._OJ_SELECTED)
                .addClass("oj-default") ;

           this._busyStack.push( {op : 'e', id : obj.attr("id")});

           // Add a busy state for the animation.  The busy state resolver will be invoked
           // when the animation is completed
           this._addBusyState("Tree (id='" + this._elemId  + "') : node id='" + obj.attr('id') + "' is animating.");

           if ((!skipAnim) && dur)  {
             this._slide(obj, false) ;       // slideDown
           }
           else  {
             this._transitionEnd($(obj.children("UL")[0]), obj) ;
           }

           if (callback)  {
             callback.call();
           }
         }
    },

    /**
      *  Expands all collapsed nodes
      *  @private
      */
    _expandAll  : function (obj, animate, original_obj)
    {
      var origTarg = obj? obj : -1 ;

      obj = obj ? this._getNode(obj) : -1;
      if (!obj || obj === -1) {
        obj = this._$container_ul;
      }
      else {
        origTarg = obj ;
      }

      if (original_obj)  {
        obj = obj.find("li.oj-collapsed");
      }
      else   {
        original_obj = obj;
        if (obj.is(".oj-collapsed"))
        {
          obj = obj.find("li.oj-collapsed").addBack();
        }
        else  {
          obj = obj.find("li.oj-collapsed");
        }
      }

      var _this = this;
      obj.each(function ()
           {
             var __this = this;

             if (!_this._is_loaded(this))   {
               _this["expand"](this, function()
                                     {
                                       _this._expandAll(__this, animate, original_obj);
                                     }, !animate);
             }
             else  {
               _this._expand(this, false, !animate);
             }
           });

      // so that callback is fired AFTER all nodes are open
      if (original_obj.find('li.oj-collapsed').length === 0)  {
        this._emitEvent({ "obj" : obj, "targ" : origTarg }, "expandAll");
      }
    },


    /**
      *  Selects a node. If "e" is not defined, the selection originated from an
      *  app API call and not via mouse/keyboard/touch action.
      *  @return  {boolean} true if selection was actually performed, else false.
      *  @private
      */
    _select : function (node, check, e)
    {
       var  core        = this._data.core,
            ui          = this._data.ui,
            selectMode  = core.selectMode;      // 0 = no select, 1 to n, or -1 = multi-select

       if (selectMode == 0) {
         return false ;                         //  selection not permitted by option
       }

       node = this._getNode(node);

       if ((node == -1) || (!node) || (!node.length))  {
         return false;
       }

       if (node.hasClass(TreeUtils._OJ_DISABLED) || core.locked) {
         return false ;
       }

       if ((! e)|| (e.type != "touchend")) {       // if via api, ensure touchEvent flag is off
         ui.touchEvent = false ;
       }

       var isSelected  = this._isSelected(node) ;

       if (! isSelected) {
         var rslt = this._emitEvent({"obj" : node, "func": "select"}, "before") ;
         if (typeof rslt == "boolean" && (!rslt)) {
           return false ;
         }
       }

       // make copy for potential optionChange event
       var prevSelections =  this.options["selection"].slice(0) ;

       //  Establish whether the select should happen, and deselect where necessary.

       var s              = this.options ;
       var selMultMod     = ui.defaults["selectMultipleModifier"] ;   // to remove - not published
       var selRangeMod    = ui.defaults["selectRangeModifier"] ;      // not yet published
       var disSelChildren = ui.defaults["disableSelectingChildren"] ; // not yet published
       var isMultiple     = (selMultMod == "on" || (selMultMod !== false && e && oj.DomUtils.isMetaKeyPressed(e))),
           isRange        = ((selRangeMod !== false) && (e &&  (e[selRangeMod + "Key"]))  &&
                            ui.lastSelected && this._data.ui.lastSelected[0] !== node[0] &&
                            (ui.lastSelected.parent()[0] === node.parent()[0])),
           proceed        = true,
           t              = this;

       if (check) {
          if (disSelChildren && isMultiple &&
              ((node.parentsUntil(".oj-tree","li").children("a.oj-selected").length) ||
               (node.children("ul").find("a.oj-selected:eq(0)").length)))  {
             return false;
          }
          proceed = false;
          switch(!0)  {
             case (isRange):
                 ui.lastSelected.addClass("oj-tree-last-selected");
                 node = node[ node.index() < ui.lastSelected.index() ? "nextUntil" : "prevUntil" ](".oj-tree-last-selected").addBack();

                 if (selectMode == -1 || node.length < selectMode)  {
                   ui.lastSelected.removeClass("oj-tree-last-selected");
                   ui.selected.each(function ()
                                      {
                                         if (this !== ui.lastSelected[0])  {
                                           t._deselect(this);
                                         }
                                       });
                   isSelected = false;
                   proceed    = true;
                 }
                 else {
                   proceed = false;
                 }
                 break;

             case  (ui.touchEvent && selectMode == -1) :  // touch and multi-select mode
                 ui.touchEvent = false ;                  // prevent race, when we call toggleselect()
                 this["toggleSelect"](node) ;             // call into public API, want event
                 proceed = false;
                 break ;

             case (isSelected && !isMultiple):
                 if (! e) {                      // if via api, nothing to do - node is
                   break ;                       // already selected.
                 }
                 this._deselectAll();
                 if (! ui.spacebar) {            // only kbd toggles selected status
                   isSelected = false;
                 }
                 proceed = true;
                 break;

             case (! isSelected && !isMultiple):
                 if (e) {                                          // if mouse click
                   if (ui.selected && (ui.selected.length == 1)) {
                     this._deselect(ui.selected) ;                 // clear the selection
                   }
                   else {
                     this._deselectAll(ui.selected) ;              // clear all selections
                   }
                 }
                 else   //if (this._data.core.selectMode == -1 || this._data.core.selectMode > 0)  {
                   if (selectMode === 1)  {
                     this._deselect(ui.selected) ;            // clear the selection
                 }
                 else if (selectMode > 1)  {                  // TDO check if the selected count will exceed the
                   this._deselectAll();                       // the selectMode count. Do we veto this select?
                 }
                 proceed = true;
                 break;

             case (isSelected && isMultiple):
                 this["deselect"](node);                      // call into public API, want event
                 break;

             case (! isSelected && isMultiple):
                 if ((selectMode == -1) ||
                     (ui.selected.length + 1 <= selectMode))  {
                   proceed = true;
                 }
                 break;
          }
       }

       // Proceed with the selection

       var bRet = false ;
       if (proceed && !isSelected)  {
          if (!isRange)  {
            ui.lastSelected = node;
          }
          node.children("a").addClass(TreeUtils._OJ_SELECTED);
          node.attr(WA_SELECTED, "true");
          bRet = true ;                      // selection performed

          if (s["selectedParentExpand"])  {
             node.parents(".oj-collapsed").each(function () {
                                                   t._expand(this, false, true);
                                            });
          }

          ui.selected = ui.selected.add(node);
          this._fix_scroll(node.eq(0));

          if (! core.suppressSelectEvent) {
            this._fireOptionChangeEvent("selection", prevSelections, null, e) ;
          }
       }

       return bRet ;
     },


     /**
       *  Deselects a node.
       *  @private
       */
     _deselect : function (node)
     {
       node = this._getNode(node);
       if (!node.length)  {
         return false;
       }
       if (node.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
         return ;
       }

       if (this._isSelected(node))  {
         node.children("a").removeClass(TreeUtils._OJ_SELECTED);
         node.removeAttr(WA_SELECTED);

         this._data.ui.selected = this._data.ui.selected.not(node);

         if (this._data.ui.lastSelected && this._data.ui.lastSelected.length) {
           if (this._data.ui.lastSelected.get(0) === node.get(0))  {
             this._data.ui.lastSelected = this._data.ui.selected.eq(0);
           }
         }
       }

     },


      /**
        *  Deselect all selected nodes, or a specified context node and its selected children.
        *  @private
        */
     _deselectAll : function(context)
     {
        if (this._data.core.locked) {
          return ;
        }

        var origTarg = context? context : -1 ;
        var ret = context ? $(context).find("a.oj-selected").parent() :
                            this._$container.find("a.oj-selected").parent();

        ret.not(".oj-disabled") ;
        if (ret.length === 0) {
          return ;                // nothing selected
        }

        var _this = this ;
        $.each(ret, function() {
            _this._deselect(this) ;
        }) ;
     },


     /** Selects all nodes specified in array.
       * @param {Array}  nodes - an array of nodes to be selected.  Nodes may be defined as elements,
       *                         jQuery wrapped nodes, or selectors pointing to the elements to be selected.
       * @private
       */
     _setSelected : function(nodes, e)
     {
        if (this._data.core.locked) {
          return null ;
        }

        if (nodes && nodes.length > 0)  {
          var _this = this;
          $.each(nodes, function (i, val) {
            if (val)  {
              _this._select(val, true, e);
            }
          });
        }
     },

     /*
      *  Handle tapping or clicking of a node
      *  @private
      */
     _handleNodeTapClick : function(event)
     {
        event.preventDefault();
        event.currentTarget.blur();
        if (! $(event.currentTarget).hasClass(OJT_LOADING))  {
          this._setFocus();
          //  Handle case where tree does not have focus and a node has been selected.
          //  We want lastHovered set to the same node so that focus does not move
          //  hover to first node.
          if (! this._data.ui.focused) {
            var hov = this._getNode(event.currentTarget) ;
            if ((hov.length > 0) && (hov != -1)) {
              this._data.ui.lastHovered = hov ;
            }
            this._$container_ul.focus() ;
          }

          this._select(event.currentTarget, true, event) ;
        }
        this._data.ui.touchEvent = false ;
     },

     /**
       *  Handle hover over a disclosure icon-font
       *  @private
       */
     _disclosureHover : function(elem, bHover)
     {
        elem = $(elem) ;
        if (elem.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
          return ;
        }

        var par     = elem.parent() ;
        var bOpen   = par.hasClass(TreeUtils._OJ_EXPANDED) ;
        var bClosed = par.hasClass(TreeUtils._OJ_COLLAPSED) ;
        if (bOpen || bClosed) {                    // only care about disclosure
          if (bHover) {                            // <ins>'s of a disclosable
            elem.addClass(TreeUtils._OJ_HOVER) ;    // parent node.
            elem.removeClass("oj-default") ;
            elem.removeClass(TreeUtils._OJ_SELECTED) ;
          }
          else {
            elem.removeClass(TreeUtils._OJ_HOVER) ;
            elem.addClass("oj-default") ;
          }
       }
     },


     /**
       *  Handle hover
       *  @private
       */
     _hover : function(node)
     {
       if (this._data.menu.activeNode) {     //TDO. why is this needed for shift-f10 on a node
         return ;                            // near the bottom. A bogus mousenter for a node
       }                                     // near the middle is received

       node = this._getNode(node);
       if ((! node.length) || node.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
         return ;
       }
       if (node.hasClass(TreeUtils._OJ_HOVER))  {
         return ;
       }

       var rslt = this._emitEvent({"obj" : node, "func" : "hover"}, "before") ;
       if (typeof rslt == "boolean" && (!rslt)) {
         return ;
       }

       //if(this.data.ui.hovered && node.get(0) === this.data.ui.hovered.get(0)) { return; }
       if (! node.hasClass(TreeUtils._OJ_HOVER))  {
          this._dehover();
       }

       this._data.ui.hovered = node.children("a").addClass(TreeUtils._OJ_HOVER).parent();
       this._$container_ul.attr(WA_ACTIVEDESCENDANT, this._data.ui.hovered.attr("id")) ;
       this._fix_scroll(node);
       this._emitEvent({ "obj" : node }, "hover");
     },

     /**
       *  Handle dehover
       *  @private
       */
     _dehover : function()
     {
       if (this._data.menu.activeNode) {     //TDO. why is this needed for shift-f10 on a node
         return ;                            // near the bottom. A bogus mousenter for a node
       }                                     // near the middle is received

       var obj = this._data.ui.hovered,
           p;

       if (!obj || !obj.length) {
         return;
       }
       if (obj.hasClass(TreeUtils._OJ_DISABLED) || this._data.core.locked) {
         return ;
       }

       p = obj.find("a.oj-hover") ;
       if (! p.length) {
         p = this._$container_ul.find("a.oj-hover") ;    // final check in case we get out of sync?
         if (! p.length) {
           return ;                                      // no hover, ignore.
         }
       }
       p = p.removeClass(TreeUtils._OJ_HOVER).parent();
       this._$container_ul.removeAttr(WA_ACTIVEDESCENDANT) ;

//       if (this._data.ui.hovered[0] === p[0]) {
         this._data.ui.hovered = null;
//       }

       if (obj.attr("id") != undefined) {
         this._emitEvent({ "obj" : obj }, "dehover");
       }
     },


     /**
       *  Placeholder for a more specific refresh action. It is replaced
       *  by _refresh_json() for json_data, or _refresh_ui() for html_data.
       *
       *  Refreshes the tree - notes all open nodes, reloads them in the tree
       *  and then reopens the noted open nodes.
       *  @private
       */
     _refresh : function(node) {
        this._refresh_core(node) ;
     },


     /**
       *  Base _refresh action
       *  @private
       */
     _refresh_core : function(node)
     {
       var origTarg = node? node : -1 ;
       var _this    = this;

       this._save_opened();
       this._saveSelected() ;
       if (!node) {
          node = -1;
       }
       node = this._getNode(node);
       if (!node)  {
         node = -1;
       }
       else {
         origTarg = node
       }

       if (node !== -1)  {
         node.children("UL").remove();
       }
       else  {
         this._$container_ul.empty();
         this._processExistingMarkup() ;    // if existing markup, reprocess in case it changed
       }
       this._load_node(node, function ()
                           {
                             _this._emitEvent({ "obj" : origTarg}, "refresh");
                             _this._reload_nodes();
                             _this._rehover() ;
                            }
                      );
     },


     /**
       *  Ui _refresh action
       *  @private
       */
     _refresh_ui : function (obj)
     {
        this._saveSelected() ;
        this._refresh_core(obj) ;
     },


//Not used currently
//   /**
//     *  Emits an "after_open" event for internal use only.
//     *  @private
//     */
//   "after_open"  : function (obj)
//   {
//      this._emitEvent({ "obj" : obj }, "after_open", true);
//   },


    /**
      *  Emits an internal "after_close" event for internal use only.
      *  @private
      */
    "after_close" : function (obj)
    {
       this._emitEvent({ "obj" : obj }, "after_close", true);
    },


    /**
      *  Expand any nodes that have been specified to be expanded.
      *  Emits an internal "reopen" event.
      *  @private
      */
    _reopen : function ()
    {
       var _this = this;

       if (this._data.core.toExpand.length) {
         $.each(this._data.core.toExpand, function (i, val)
                                          {
                                             _this._expand(val, false, true);
                                          });
       }
       this._emitEvent({}, "reopen", true);     // this event will also cause selections to be tried.
    },


    /**
      *  If a node was hovered, refind it in the DOM (after refresh).
      *  @private
      */
     _rehover : function()
     {
       var id ;
       var $hovered = this._data.ui.lastHovered ;
       if ($hovered) {
         id = $hovered.attr("id") ;
         if (id) {
           $hovered = this._getNode("#" + id) ;
         }
       }
       this._data.ui.hovered = this._data.ui.lastHovered = $hovered;
     },


     /**
       *  Returns an array of currently selected nodes.
       *  @private
       */
     _getSelected : function (context)
     {
         var sel, ar, i, len ;

         sel = context ? $(context).find("a.oj-selected").parent() : this._data.ui.selected ;
         ar  =[] ;
         len = sel.length ;
         for (i = 0; i < len; i++) {
            ar.push(sel[i]) ;
         }
         return ar ;
     },


     /**
       *  Changes the text title of the node's <a>
       *  @private
       */
     _set_text  : function (obj, val)
     {
        obj = this._getNode(obj);
        if (!obj.length)  {
          return false;
        }

        obj = obj.children("a:eq(0)");
        obj = obj.find("span:eq(0)") ;
        this._emitEvent({ "obj" : obj, "name" : val }, "set_text", true);
        return (obj[0].textContent = val);
     },

     /**
          Load the tree
       *  @private
       */
     _loadNodes : function()
     {
        if (this._data.ds.type !== DS_NONE && this._data.ds.type !== DS_ERROR) {
          // Construct the tree
          this._load_node(-1,
                          function () {                         // success func
                                         this._loaded();        // emit ojloaded event
                                         this._reload_nodes();  // handle initExpanded, etc
                                      });
        }
        else {
           this._applyEmptyText() ;     //  Use emptyText option if defined.
           this._loaded() ;
          }
      },


     /**
       *  @private
       */
     _load_node : function(obj, success_callback, error_callback)     // Dummy function overriden by data methods
     {
        this._emitEvent({ "obj" : obj }, "load_node", true);
     },


     /**
       *  Returns whether a node is current loaded. This is a dummy function
       *  overriden by data methods such as _is_loaded_J(), _is_loaded_html()
       *  or _is_loaded_DS().
       *  @private
       */
     _is_loaded  : function (obj)                 // Dummy function overriden by data methods
     {
       return true;
     },


     /**
       *  TreeDataSource's _load_node
       *  @private
       */
     _load_node_DS : function (obj, s_call, e_call)
     {
        var _this = this;
        this._load_node_tree(obj, function ()  {
                                      _this._emitEvent({ "obj" : _this._getNode(obj) }, "load_node", true);
                                      s_call.call(this);
                                  }, e_call);
     },


     /**
       *   JsonTreeDataSource's _is_loaded()
       *   @return {boolean}
       *   @private
       */
     _is_loaded_DS : function (obj)
     {
//     var s = this.options["data"]["json"];

       obj = this._getNode(obj);

//     return (obj === -1 || !obj ) || (!s["ajax"] && !this._data.ds.progressiveRender && !$.isFunction(s["data"]))
       return (obj === -1 || !obj ) || obj.is(".oj-expanded, .oj-tree-leaf")
                                    || obj.children("ul").children("li").length > 0;
     },


     /**
       *   JsonTreeDataSource's _refreshDS
       *  @private
       */
     _refresh_DS  : function (obj)
     {
        obj = this._getNode(obj);

//      var s = this.options["data"]["json"];
//      if (obj && obj !== -1 && this._data.ds.progressiveUnload && ($.isFunction(s["data"]) || !!s["ajax"]))  {
        if (obj && obj !== -1)  {
          obj.removeData(OJT_CHILDREN);
        }
        return this._refresh_ui(obj);
     },


     /**
       *  JSON_data's _load_node()
       *  @private
       */
     _load_node_J : function (obj, s_call, e_call)
     {
        var _this = this;
        this._loadNodeJson(obj,                             // node or -1 for tree
                             function ()                      // success func
                              {
                                _this._emitEvent({ "obj" : _this._getNode(obj) }, "load_node", true);
                                s_call.call(this);
                              },
                              e_call);                        // error func
     },

     /**
       *   JSON _data's _is_loaded()
       *   @private
       */
     _is_loaded_J : function (obj)
     {
       var s = this.options["data"];

       obj = this._getNode(obj);
       return obj == -1 || !obj || (!s["ajax"] && !this._data.ds.progressiveRender && !$.isFunction(s["data"]))
                        || obj.is(".oj-expanded, .oj-tree-leaf")
                        || obj.children("ul").children("li").length > 0;
     },


     /**
       *   html_data's _load_node
       *   @private
       */
     _load_node_H : function (obj, s_call, e_call)
     {
       var _this = this;
       this._loadNodeHtml(obj, function ()                   // success func
                                {
                                  _this._emitEvent({ "obj" : _this._getNode(obj)}, "load_node", true) ;
                                  s_call.call(this);
                                },
                                e_call);                   // error func
     },

     /**
       *  html_data's _is_loaded()
       *  @private
       */
     _is_loaded_H : function (obj)
     {
       var s    = this.options["data"],
           data = null,
           ajax = null ;

      if (s) {
        data  = s["data"] || null ;
        ajax  = s["ajax"] || null ;
      }

       obj = this._getNode(obj);

       return obj == -1 || !obj || (!ajax && !$.isFunction(data)) ||
                           obj.is(".oj-expanded, .oj-tree-leaf") || obj.children("ul").children("li").length > 0;
     },


     /**
       *  Rselect nodes as of last savedSelected()
       *  @private
       */
     _reselect : function ()
     {
        var _this = this,
            s     = this._data.ui.toSelect;

        s = $.map($.makeArray(s), function (n) {
                           return "#" + n.toString().replace(/^#/,"")
                                                    .replace(/\\\//g,"/").replace(/\//g,"\\\/").replace(/\\\./g,".").replace(/\./g,"\\.").replace(/\:/g,"\\:"); });
        // Clear selection option
        if (s && s.length) {
          this.options["selection"].length = 0 ;  // forget the noted selections
          this._data.ui.selected.length = $() ;
        }

        // Reapply the selections
        $.each(s, function (i, val)
                   {
                      if (val && val !== "#")  {
                        _this["select"](val);
                      }
                   });

        this._data.ui.selected = this._data.ui.selected.filter(function ()
                                           {
                                             return this["parentNode"];
                                           });
//      this._emitEvent(null, "reselect", true);
     },

     /**
       * Save selection state - see also _reselect()
       * @private
       */
     _saveSelected : function ()
     {
        var _this = this,
            ui    =  this._data.ui ;

        ui.toSelect = [];
        ui.selected.each(function ()
                  {
                    if (this.id) {
                      ui.toSelect.push("#" + this.id.toString().replace(/^#/,"").replace(/\\\//g,"/").replace(/\//g,"\\\/").replace(/\\\./g,".").replace(/\./g,"\\.").replace(/\:/g,"\\:"));
                     }
                  });
//     this._emitEvent(ui.toSelect, "savedselected", true);
     },


     /*
      *      Not documented
      */
     "rollback" : function (rb)
     {
        if (rb)  {
          if (!$.isArray(rb))  {
            rb = [ rb ];
          }
//        $.each(rb, function (i, val)
//                   {
//                     instances[val.i]["set_rollback"](val.h, val.d);     //TDO
//                   });
        }
    },


    "get_rollback" : function ()
    {
       this._emitEvent(null, "get_rollback", true);
       return {
                i : this._getIndex(),
                h : this._$container.children("ul").clone(true),
                d : this["data"]                    // TDO ???
              };
    },

/*  Currently not used
    "set_rollback" : function (html, data)
    {
       if (this._$container && this._$container_ul)   {
         this._$container_ul.empty().append(html);
       }

       this["data"] = data;                       // TDO ???
       this._emitEvent(null, "set_rollback", true);
    },
*/

    /**
      *  Load json for a particular node (or the whole tree)
      *  @private
      */
    _load_node_tree : function (obj, s_call, e_call)
    {
       var parentKey = (obj && obj != -1)? obj[0].id : (obj? obj : -1) ;
       var rslt = this._JsonDSToJson(parentKey, obj) ;

       if (! rslt.success) {
         return ;
       }

       var  bTree = ((! obj) || (obj === -1)) ;
       var  s     = this.options["data"] ;
       var  d ;

       if ((!!s["data"] && !s["ajax"]) || (!!s["data"] && !!s["ajax"] && bTree))  {
         if (bTree)  {
           d = this._parseJson(rslt.js, obj) ;
           if (d)  {
             this._$container_ul.empty().append(d.children());      //@HTMLUpdateOK
             this._cleanNode();
           }
           else  {
              if (this._data.ds.correctState) {
                  this._$container_ul.empty();
               }
           }
         }
         if (s_call) {
            s_call.call(this);
         }
       }
       else if  ((!s["data"] && !!s["ajax"]) || (!!s["data"] && !!s["ajax"] && (! bTree))) {
          d = this._parseJson(rslt.js, obj);
          if (d)  {
            if (bTree)  {
              var $u =  this._$container_ul;
              $u.empty().append(d.children());                     //@HTMLUpdateOK
              $u.attr(WA_ROLE, WA_TREE).attr(WA_LABELLEDBY, this._elemId).attr("tabindex", "0").css("outline", "none") ;
              if (this._data.core.selectMode === -1)  {
                $u.attr("aria-multiselectable", true) ;
              }
            }
            else  {
              obj.append(d).children("a.oj-tree-loading").removeClass(OJT_LOADING);   //@HTMLUpdateOK
              obj.removeData("oj-tree-is-loading");
            }

            this._cleanNode(obj);
            if (s_call)  {
              s_call.call(this);
            }
          }
          else  {
            if (bTree) {
              if (this._data.ds.correctState)  {
                this._$container_ul.empty();
                if (s_call) {
                  s_call.call(this);
                }
              }
            }
            else  {
               obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
               obj.removeData("oj-tree-is-loading");
               if (s["correct_state"]) {
                 this._correct_state(obj);
                 if (s_call)  {
                   s_call.call(this);
                 }
               }
            }
          }
       }

     },


    /**
      * Recursively process a JsonTreeDataSource to a Json array ready for parsing.
      * JsonTreeDataSource does not support later data addition, so all nodes
      * must be available from the outset. This implies that there can be no 
      * no lazy-loading.
      * @private
      * @return {Object} contains "success" (boolean) and "js" (the json nodes)
      */
     _JsonDSToJson : function (parentKey, node)
     {
        var arJson = [],
            ds     = this._tds,
            cc,
            range  = {},
            rslt   = {success : false, js : null} ;

        if (parentKey == -1) {
          parentKey = null ;
          range["start"] = 0 ;
        }
        else {
          // TDO
        }

        cc = ds.getChildCount(parentKey) ;
        if (cc > 0) {
          range["count"] = cc ;
          ds.fetchChildren(parentKey, range,              // get the JsonNodeSet
              {                                           // callbacks
                "success" : function(jns) {
                     var c = jns.getCount(),
                         attr,
                         n ;

                     for (var i = 0; i < c; i++) {
                        node = {} ;
                        attr = jns.getData(i) ;           // get the attribute list to be applied
                        if (attr) {
                          node["attr"] = attr ;
                        }
                       node["title"] = jns.m_nodes[i]["title"] ;         // wait for chadwick
                       if (attr.metadata) {
                         node["metadata"] = jns.m_nodes[i]["metadata"] ; // wait for chadwick
                       }

                       var key  = node["attr"]["id"] ;

                       n = ds.getChildCount(key) ;
                       if (n > 0) {
                          var r = this._JsonDSToJson(key, node) ;
                          node["children"] = r.js ;
                       }

                       arJson.push(node) ;
                     }

                     rslt.success = true ;
                     rslt.js      = arJson ;
                   }.bind(this),

                "error" :  function(status) {
                              rslt.success = false ;
                           }
              }) ;
        }
        else {
          rslt.success = true ;   // DS was empty.  Will not treat this as an error,
                                  // because we did not have any failures.
        }
        return rslt ;
     },


     /**
       *  @private
       */
     _refresh_json  : function (obj)
     {
        obj = this._getNode(obj);

        if (this._data.core.locked) {
          return ;
        }

        var bTree = (!obj || obj !== -1 || !obj.length) ;
        if (!bTree && obj.hasClass(TreeUtils._OJ_DISABLED)) {
          return ;
        }

        var s = this.options["data"]["json"];

        if (! bTree && this._data.ds.progressiveUnload && ($.isFunction(s["data"]) || !!s["ajax"]))  {
          obj.removeData(OJT_CHILDREN);
        }
        return this._refresh_ui(obj);
     },


     /**
       *  Load json for a particular node (or -1 for the whole tree). s_call and e_call are
       *  optional success and error callback's.
       *  @private
       */
     _loadNodeJson : function (obj, s_call, e_call)
     {
       var s = this._getOptions()["data"],        // work on a copy of the options to avoid
           d,                                     // endless loop in calling success/error funcs
           error_func   = function () {},
           success_func = function () {},
           data         = (s && s["data"]) || null,
           ajax         = (s && s["ajax"]) || null;

       if (s && !data && !ajax) {
         data = s ;                               // we just have data, no data() ajax() methods defined
       }

       obj = this._getNode(obj);

       if (obj && obj !== -1 && (this._data.ds.progressiveRender || this._data.ds.progressiveUnload)
                             && !obj.is(".oj-expanded, .oj-tree-leaf")
                             && obj.children("ul").children("li").length === 0 && obj.data(OJT_CHILDREN)) {
          d = this._parseJson(obj.data(OJT_CHILDREN), obj);
          if (d)  {
            obj.append(d);                                             //@HTMLUpdateOK
            if (! this._data.ds.progressiveUnload) {
              obj.removeData(OJT_CHILDREN);
            }
          }

          this._cleanNode(obj);
          if (s_call)  {
            s_call.call(this);
          }
          return;
       }

       if (obj && obj !== -1) {
         if (obj.data("oj-tree-is-loading"))  {
            return;
         }
         else  {
            obj.data("oj-tree-is-loading",true);
         }
       }

       switch (!0)
       {
          case (!data && !ajax) : throw "ojTree - neither data nor ajax settings supplied.";

          // Data function supplied by user data definition.
          case ($.isFunction(data)):
                       data.call(this, obj, function (d) { // invoke user function passing a callback
                           d = this._parseJson(d, obj);
                           if (!d) {
                              if (obj === -1 || !obj)  {
                                if (this._data.ds.correctState)  {
                                  this._$container_ul.empty();
                                }
                              }
                              else  {
                                 obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
                                 obj.removeData("oj-tree-is-loading");

                                 if (this._data.ds.correctState) {
                                   this._correct_state(obj);
                                 }
                              }
                              if (e_call) {
                                e_call.call(this);
                              }
                           }
                           else  {
                              if (obj === -1 || !obj)  {
                                 this._$container_ul.empty().append(d.children());    //@HTMLUpdateOK
                              }
                              else  {
                                 obj.append(d).children("a.oj-tree-loading").removeClass(OJT_LOADING);   //@HTMLUpdateOK
                                 obj.removeData("oj-tree-is-loading");
                              }
                              this._cleanNode(obj);
                              if (s_call)  {
                                s_call.call(this);
                              }
                           }
                       }.bind(this));
                       break;

          case (!!data && !ajax) || (!!data && !!ajax && (!obj || obj === -1)):

                       if (!obj || obj == -1)  {
                         d = this._parseJson(data, obj) ;
                         if (d)  {
                           this._$container_ul.empty().append(d.children());    //@HTMLUpdateOK
                           this._cleanNode();
                         }
                         else  {
                            if (this._data.ds.correctState) {
                              this._$container_ul.empty();
                             }
                         }
                       }
                       if (s_call) {
                          s_call.call(this);
                       }
                       break;

          case (!data && !!ajax) || (!!data && !!ajax && obj && obj !== -1):

                       // AJAX error function
                       error_func = function (x, status, e)
                       {
                         var self = this.ctx ;                      // get the tree's this
                         //console.log("Ajax FAIL JSON id='" + this.id + "' obj.id=" +                        
                         //              ((obj === -1)? -1 : obj.attr("id")) + " url=" + this.url);

                         self._noteLoadFailureId(this.id, true) ;   // add id to load failure list
                         self._removeFromBusyStack('j', ((obj === -1)? -1 : obj.attr('id'))) ;
                         self._resolveIfBusyStackEmpty();

                         var ef = self._getOptions()["data"]["ajax"]["error"];  // reget the options
                         if (ef) {                                              // without our updated ajax
                           ef.call(self, status, e, x);                         // changes to avoid forever loop
                         }
                         if (obj != -1 && obj.length)
                         {
                            obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
                            obj.removeData("oj-tree-is-loading");
                            if (status === "success" && self._data.ds.correctState)  {
                              self._correct_state(obj);
                            }
                         }
                         else
                         {
                           if ((status == "error") || (status === "success" && self._data.ds.correctState))  {
                             self._$container_ul.empty();             // remove tree loading icon/node
                             self._applyEmptyText() ;  // no tree data
                           }
                         }
                         if (e_call)  {
                           e_call.call(self);
                         }
                       };

                       // AJAX success function
                       success_func = function (d, status, x)
                       {
                         // 'this' is the context ptr
                         var self = this.ctx ;      // get the tree's this

                         //console.log("Ajax SUCCESS JSON id='" + this.id + "' obj.id=" +                        
                         //              ((obj === -1)? -1 : obj.attr("id")) + " url=" + this.url);

                         self._noteLoadFailureId(this.id, false) ;    // if this id is the failure list, remove it
                         self._removeFromBusyStack('j', this.id) ;

                         // Reget the options without our updated ajax changes to avoid forever loop
                         var sf = self._getOptions()["data"]["ajax"]["success"];
                         if (sf) {
                           d = sf.call(self, d, status, x) || d;  // call app success function
                         }

                         // Check if all whitespace since JSON.parse() will barf on this.
                         if (typeof d == "string") {
                           var tempd = d.replace(/^[\s\n]+$/,"") ;
                           try {
                                 tempd = JSON.parse(tempd) ;
                               } catch(err) {tempd = null;}     // have nothing useful to display

                           if (! tempd) {
                             return error_func.call(this, x, "Bad JSON", "");
                           }
                         }

                         d = self._parseJson(d, obj);        // parse json and get a $(ul)
                         if (d)
                         {
                            if (obj === -1 || !obj)  {
                               var $u =  self._$container_ul;
                               $u.empty().append(d.children());                                         //@HTMLUpdateOK
                               $u.attr(WA_ROLE, WA_TREE).attr(WA_LABELLEDBY, self._elemId)
                                                        .attr("tabindex", "0").css("outline", "none") ;
                               if (self._data.core.selectMode === -1)  {
                                 $u.attr("aria-multiselectable", true) ;
                               }
                             }
                             else   {
                                obj.append(d).children("a.oj-tree-loading").removeClass(OJT_LOADING);   //@HTMLUpdateOK
                                obj.removeData("oj-tree-is-loading");
                             }

                             self._cleanNode(obj);
                             if (s_call)  {
                               s_call.call(self);
                             }
                         }
                         else
                         {                             // parse returned nothing
                           if (obj === -1 || !obj) {
                              if (self._data.ds.correctState)  {
                                 self._$container_ul.empty();
                                 if (s_call) {
                                   s_call.call(self);
                                 }
                              }
                           }
                           else  {
                              obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
                              obj.removeData("oj-tree-is-loading");
                              if (self._data.ds.correctState) {
                                self._correct_state(obj);
                                if (s_call)  {
                                  s_call.call(self);
                                }
                              }
                           }
                         }
                       };

                       //  Prepare for an aynchronous ajax op. (note: we are updating a copy of the options)

                       var nodeId = ((obj === -1)? -1 : obj.attr("id"));
                       this._busyStack.push( {op : 'j', id : nodeId});
                       this._addBusyState("Tree (id='" + this._elemId + "') : node id='" + nodeId + "' is lazyloading.");

                       //console.log("ajax JSON load (" + nodeId + ") . . .  busyStack=" + this._arrayToStr(this._busyStack)) ; 

                       var reqContext = {ctx: this, id: nodeId};

                       s["ajax"]["context"] = reqContext;
                       s["ajax"]["error"]   = error_func;
                       s["ajax"]["success"] = success_func;

                       if (! s["dataType"])  {
                         s["ajax"]["dataType"] = "json";
                       }
                       reqContext.type = s["ajax"]["dataType"] ;

                       if ($.isFunction(s["ajax"]["url"]))
                       {
                         s["ajax"]["url"] = s["ajax"]["url"].call(this, obj);
                         reqContext.url = s["ajax"]["url"] ;
                       }
                       if ($.isFunction(s["ajax"]["data"]))
                       {
                         s["ajax"]["data"] = s["ajax"]["data"].call(this, obj);
                       }
                       $.ajax(s["ajax"]);
                       break;                 // end case

       }   // end switch

     },

     /**
       *  Recursively parses a JSON representation of a tree or tree node(s).
       *  Returns a jQuery wrapped <ul> structure, or false if parse fails.
       *  @private
       */
     _parseJson : function (js, obj, isRecurse)
     {
        var d = false,
            tmp, i, j, ul1, ul2, title;

        if (!js) {
          return d;
        }

        if (this._data.ds.progressiveUnload && obj && obj !== -1)  {
          obj.data(OJT_CHILDREN, d);
        }

        if (typeof js == "string")  {
          try {
                js = JSON.parse(js) ;
              } catch(err) {js = [];}     // have nothing useful to display
        }

        if ($.isArray(js))  {             // array of ojects ?
          d = $('<ul>');
          if (! js.length)  {
            return false;
          }

          for (i = 0, j = js.length; i < j; i++)  {       // recurse until we have an individual node obj
             tmp = this._parseJson(js[i], obj, true);
             if (tmp.length)  {
               d = d.append(tmp);                        //@HTMLUpdateOK
             }
          }

          d = d.children();
        }
        else  {
           //  We have an individual node object

           if (typeof js == "string")  {
             js = { "data" : js };
           }

           title = (typeof js["title"] == "string")?  js["title"] : " " ;

           d = $("<li role='" + WA_TREEITEM + "' />");                  // start a node

           if (js["attr"])  {
             if (this._data.types.defType && (! js["attr"]["type"])) {  // if no type and "default" type
               js["attr"]["type"] = "oj-tree-deftype" ;                 // defined, add default type
               d.addClass(OJT_TYPE) ;
             }
             d.attr(js["attr"]);       // apply attr's to the <li>
           }
           if (js["metadata"])  {      // apply any user defined arbitrary data for the <li>
             d.data(js["metadata"]);
           }

           // js.state     // not published - per Design Review
           // Used internally only for DnD
           if (js["children"] && js["children"].length === 0) {    // length zero means lazy load
             d.addClass(TreeUtils._OJ_COLLAPSED);
           }
           if (js["state"] && (js["state"] === "s")) {
             d.addClass(TreeUtils._OJ_SELECTED);    // Added as a flag, see _dndFinish()
           }

           if (! js["data"]) {
             js["data"] = {"dummy": 0} ;
           }

           var bIns = false,
               sp ;

           tmp = $("<a/>");
           $.each(js["data"], function (i, m)   {
                if ($.isFunction(m)) {
                  m = m.call(this, js);
                }
                if (typeof m === "object") {
                 if (i == "attr") {
                   tmp.attr(m) ;
                 }
                 else if (i == "style") {
                   tmp.css(m) ;
                 }
                  if (i == "language") {
                    tmp.addClass(m);
                  }
                }
//              else if (typeof m == "string")  {
//                tmp.attr('href','#')[ ht? "html" : "text" ](title);
//              }

                if (! bIns) {
                  sp = $("<span class='" + OJT_TITLE + "'>") ;
                  sp[0].textContent = title ;       // per securty post-impl

                  tmp.prepend("<ins class='oj-tree-icon " + OJT_NICON + " oj-component-icon" + "'>&#160;</ins>", sp);    // node icon //@HTMLUpdateOK
                  bIns = true ;
                }
                if (!m["icon"] && js["icon"])  {
                  m["icon"] = js["icon"];
                }
                if (m["icon"])  {
                  if (m["icon"].indexOf("/") === -1)  {
                    tmp.children("ins").addClass(m["icon"]);
                  }
                  else  {
                    tmp.children("ins").css("background","url('" + m["icon"] + "') center center no-repeat");
                  }
                }
           });

           d.append(tmp);            // append the <a> and its children to the <li>                    //@HTMLUpdateOK
           d.prepend("<ins class='" + OJ_DISC + "'>&#160;</ins>");      // potential disclosure icon   //@HTMLUpdateOK

           if (js["children"])  {
             if (this._data.ds.progressiveRender && js["state"] !== "expanded")  {
               d.addClass(TreeUtils._OJ_COLLAPSED)
                .attr(WA_EXPANDED, "false")
                .data(OJT_CHILDREN, js["children"]);
             }
             else  {
               if (this._data.ds.progressiveUnload)  {
                 d.data(OJT_CHILDREN, js["children"]);
               }
               if ($.isArray(js["children"]) && js["children"].length) {
                 tmp = this._parseJson(js["children"], obj, true);
                 if (tmp.length)  {
                   ul2 = $("<ul role='" + WA_GROUP + "' />");
                   ul2.append(tmp);                                     //@HTMLUpdateOK
                   d.append(ul2);                                       //@HTMLUpdateOK
                 }
               }
             }
           }
        }  // end else

        if (! isRecurse)  {
          ul1 = $("<ul />");
          ul1.append(d);                                                //@HTMLUpdateOK
          d = ul1;
        }

        return d;
     },

     /*
      *   Returns a Json representation of a node
      *   @private
      */
     __getJson : function (obj, li_attr, a_attr, isCallback)
     {
       var result = [],
           s      = this.options,
           _this  = this,
           tmp1, tmp2, val, li, a, t, sAttr, title, attr, uattr, n;

       obj = this._getNode(obj);

       if (!obj || obj === -1)  {
         obj = this._$container.find("> ul > li");
       }
       li_attr = $.isArray(li_attr) ? li_attr : [ "id", "class"];

       if (!isCallback && s["types"])  {
         li_attr.push(s["types"]["attr"]);
       }
       a_attr = $.isArray(a_attr) ? a_attr : [ ];

       obj.each(function () {
             li    = $(this);
             tmp1  = {};
             uattr = _this._noteUserAttrs(li) ;   // get list of user added attr's

             if (li_attr.length || (uattr && uattr.length))  {
               tmp1["attr"] = {};
             }

             // Add user added attr's back into the "attr" property
             if (uattr) {
               $.each(uattr, function (i, v)
                 {
                   if (li.attr(v).length > 0) {
                     tmp1["attr"][v] = li.attr(v) ;
                   }
                });
             }

             $.each(li_attr, function (i, v)
               {
                   //  Find/save any non oj classes
                   tmp2 = li.attr(v);
                   if (tmp2 && tmp2.length && tmp2.replace(/oj-tree[^ ]*/ig,'').length)  {
                     tmp2  = (" " + tmp2).replace(/ oj-tree[^ ]*/ig,'')
                                         .replace(/ oj-[^ ]*/ig,'')
                                         .replace(/\s+$/ig," ")
                                         .replace(/^ /,"")
                                         .replace(/ $/,"");
                     if (tmp2.length) {
                       tmp1["attr"][v] = tmp2 ;
                     }
                   }
               });

             // Additional state for DnD only, not valid for user code
             if (_this._getDndContext().isDragEnabled()) {
//             if (li.hasClass(OJ_EXPANDED))  {
//               tmp1["state"] = "e";
//             }
               if (_this._isSelected(li)) {
                 tmp1["state"] = "s";
               }
             }

             //  Check if node json included a metadata attribute
             val = li.data() ;
             if (! isPureObjEmpty(val))  {
               tmp1["metadata"] = val;
             }
             a = li.children("a");
             tmp2 = null ;
             a.each(function ()  {
                 t = $(this);

                 if (a_attr.length ||
                     (t.children("ins").get(0)["style"]["backgroundImage"].length) ||
                     (t.children("ins").get(0)["className"] && t.children("ins").get(0)["className"]
                                                                                .replace(/oj-tree[^ ]*|$/ig,'').length))
                 {
                   title = _this["getText"](t) ;
                   $.each(a_attr, function (k, z) {
                       sAttr = (" " + (t.attr(z) || "")).replace(/ oj-tree[^ ]*/ig,'')
                                                        .replace(/ oj-[^ ]*/ig,'')
                                                        .replace(/\s+$/ig," ")
                                                        .replace(/^ /,"")
                                                        .replace(/ $/,"");
                       if (sAttr.length) {
                         if (! tmp2) {
                           tmp2 = { "attr" : { } };
                         }
                         tmp2["attr"][z] = sAttr ;
                       }
                   });

                   if (t.children("ins").get(0)["className"].replace(/oj-tree[^ ]*|$/ig,'')
                                                            .replace(/^\s+$/ig,"").length) {
                     sAttr = t.children("ins").get(0)["className"]
                                              .replace(/oj-tree[^ ]*|$/ig,'')
                                              .replace(/ oj-[^ ]*/ig,'')
                                              .replace(/\s+$/ig," ")
                                              .replace(/^ /,"")
                                              .replace(/ $/,"");
                     if (sAttr.length) {
                       if (! tmp2) {
                         tmp2 = {} ;
                       }
                       tmp2["icon"] = sAttr ;
                     }
                   }

                   if (t.children("ins").get(0)["style"]["backgroundImage"].length)  {
                     sAttr = t.children("ins").get(0)["style"]["backgroundImage"]
                                              .replace("url(","")
                                              .replace(")","");
                     if (sAttr.length) {
                       if (! tmp2) {
                         tmp2 = {} ;
                       }
                       tmp2["icon"] = sAttr ;
                     }
                   }
                 }
                 else  {
                    title = _this["getText"](t);
                 }

                 if (tmp2 != null)  {
                   if (! tmp1["data"]) {
                     tmp1["data"] = [] ;
                   }
                   if (a.length > 1) {
                     tmp1["data"].push(tmp2);
                   }
                   else  {
                     tmp1["data"] = tmp2;
                   }
                 }
                 tmp1["title"] = title ;
             });

             li = li.find("> ul > li");
             if (li.length)  {
                tmp1.children = _this.__getJson(li, li_attr, a_attr, true);
             }
             result.push(tmp1);
       });

       return result;
     },

     /**
       *  If LI elem contains user attr definitions other than "id", "title" or ones we
       *  have added like "role"/"class", note them so that they can be reinstated for
       *  parseJson().
       *  @private
       *  @return {Array | null} An array of user attr names, else null
       */
     _noteUserAttrs : function($node)
     {
        var attrs, attr ;
        var ar, len ;
        
        attrs = $node[0].attributes ;    // get the LI's attributes

        if (attrs && attrs.length)
        {
          ar  = [];
          len = ATTR_IGN.length ;

          for (var i = 0; i < attrs.length; i++) {
             attr = attrs[i] ;
             if (attr) {
               attr = attr.name ;
               for (var j = 0; j < len; j++) {
                  if ((ATTR_IGN[j] == attr) || (attr.indexOf("aria-") == 0)) {  // closure compiler is not allowing startsWith!
                    break ;                   // ignore if in our excluded list
                  }
               }
               if (j >= len) {
                 ar.push(attr) ;
               }
            }
          }
        }

        return (ar && ar.length > 0)? ar : null ;
     },


     /**
       *  Corrects nodes by setting closed nodes to leaf nodes if no children found.
       *  @private
       */
     _correct_state   : function (obj)
     {
         obj = this._getNode(obj);
         if (!obj || obj === -1) {
           return false;
         }
         obj.removeClass("oj-collapsed oj-expanded")
            .removeAttr(WA_EXPANDED)
            .addClass(OJT_LEAF)
            .children("ul")
            .remove();

         obj.find("ins:first-child").removeClass(OJ_DISC).addClass(OJT_ICON) ;
         this._emitEvent({ "obj" : obj }, "correct_state", true);
     },


     /*
      *  @private
      */
/*                                           // not used in V1
     _core_notify : function (n, data)
     {
        if (data.opened)  {
          this._expand(n, false, true);
        }
     },
*/

     /**
       *  Dummy function to fire after the first load (so that there is a loaded event)
       *  @private
       */
     _loaded  : function ()
     {
        this._emitEvent(null, "loaded");
      },


     /**
       *   Process the user tree <ul> list placed in the tree div, or loaded via ajax.
       *   @private
       */
     _loadNodeHtml : function (obj, s_call, e_call)
     {
       var s = this._getOptions()["data"],        // work on a copy of the options to avoid
           d,                                     // endless loop in calling success/error funcs
           error_func   = function () {},
           success_func = function () {},
           data         = (s && s["data"]) || null,
           ajax         = (s && s["ajax"]) || null;

       obj = this._getNode(obj);
       if (obj && obj !== -1)  {
         if (obj.data("oj-tree-is-loading")) {
           return;
         }
         else  {
           obj.data("oj-tree-is-loading", true);
         }
       }

       switch(!0)
       {
          case ((!data && !ajax) && (s && (typeof s === "string"))) :
                    this._loadHtmlString(s, obj, s_call, e_call) ;
                    break ;

          case $.isFunction(data):
                    data.call(this, obj, function (d)
                              {
                                 this._loadHtmlString(d, obj, s_call, e_call) ;
                              }.bind(this));
                    break;

          case (!data && !ajax):
                    if (!obj || obj == -1 )  {
                      this._$container_ul
                              .empty()
                              .append(this._data.html.cloneMarkup)      //@HTMLUpdateOK
                              .find("li, a")
                              .filter(function () {
                                               return (!this.firstChild || !this.firstChild.tagName ||
                                                       this.firstChild.tagName !== "INS");
                                                   })
                              .prepend("<ins class='oj-tree-icon' >&#160;</ins>").end()          //@HTMLUpdateOK
                              .filter("a").children("ins:first-child")
                              .not(".oj-tree-node-icon")
                              .addClass(OJT_NICON)
                              .addClass(OJT_ICON);

                      this._$container_ul.find("li").children("ins:first-child")
                                                    .addClass(OJ_DISC);

                      //  No point in checking for empty parent <ul> because we don't have a
                      //  "data" option in the first place, so no lazy loading possible.

                      //  Add the <a> text <span> for hover/click styling
                      this._insertHtmlTextSpan(this._$container_ul) ;

                      if (this._data.types.defType) {             // if "default" type defined
                        this._addDefType(this._$container_ul) ;   // apply to nodes with no asooc type
                      }

                      this._cleanNode();
                      this._$container_ul.find("ul").attr(WA_ROLE, WA_GROUP) ;
                      this._$container_ul.find("li").attr(WA_ROLE, WA_TREEITEM) ;
                    }
                    if (s_call)  {
                      s_call.call(this);
                    }
                    break;

          case (!!data && !ajax) || (!!data && !!ajax && (!obj || obj === -1)):
                    if (!obj || obj == -1) {
                           d = $(data);
                           if (!d.is("ul")) {
                             d = $("<ul />").append(d);            //@HTMLUpdateOK
                           }
                           this._$container_ul
                               .empty()
                               .append(d                           //@HTMLUpdateOK
                               .children())
                               .find("li, a")
                               .filter(function ()
                                           {
                                              return (!this.firstChild || !this.firstChild.tagName ||
                                                      this.firstChild.tagName !== "INS");
                                           })
                               .prepend("<ins class='oj-tree-icon'>&#160;</ins>").end()   //@HTMLUpdateOK
                               .filter("a")
                               .children("ins:first-child")
                               .not(".oj-tree-node-icon")
                               .addClass(OJT_NICON)
                               .addClass(OJT_ICON);

                           this._$container_ul.find("li.oj-tree-leaf ins:first-child").removeClass(OJ_DISC)
                                                                                      .removeClass(OJT_NICON)
                                                                                      .addClass(OJT_ICON);

                           //  Add the <a> text <span> for hover/click styling
                           this._insertHtmlTextSpan(this._$container_ul) ;

                           if (this._data.types.defType) {             // if "default" type defined
                             this._addDefType(this._$container_ul) ;   // apply to nodes with no asooc type
                           }

                           this._cleanNode();
                           this._$container_ul.find("ul").attr(WA_ROLE, WA_GROUP) ;
                           this._$container_ul.find("li").attr(WA_ROLE, WA_TREEITEM) ;
                    }
                    if (s_call)  {
                         s_call.call(this);
                    }
                    break;

          // AJAX request
          case (!data && !!ajax) || (!!data && !!ajax && obj && obj !== -1):
                    obj = this._getNode(obj);
                    // AJAX error function
                    error_func = function (x, status, e)
                    {
                      var self = this.ctx ;                      // get the tree's this
                      //console.log("Ajax FAIL HTML id='" + this.id + "' obj.id=" +                        
                      //              ((obj === -1)? -1 : obj.attr("id")) + " url=" + this.url);

                      self._noteLoadFailureId(this.id, true) ;   // add id to load failure list

                      self._removeFromBusyStack('j', ((obj === -1)? -1 : obj.attr('id'))) ;
                      self._resolveIfBusyStackEmpty();

                      var ef = self._getOptions()["data"]["ajax"]["error"];  // reget the options
                      if (ef) {                                              // without our updated ajax
                        ef.call(self, x, status, e);                         // changes to avoid forever loop
                      }

                      if (obj != -1 && obj.length)  {
                        obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
                        obj.removeData("oj-tree-is-loading");
                        if (status === "success" && self._data.ds.correctState) {
                           self._correct_state(obj);
                        }
                      }
                      else  {
                        if (status === "success" && self._data.ds.correctState) {
                          self._$container_ul.empty();
                          self._applyEmptyText() ;  // no tree data
                        }
                      }
                      if (e_call)  {
                        e_call.call(self);
                      }
                    };

                    success_func = function (d, status, x)
                    {
                      var self = this.ctx ;                      // get the tree's this
                      var parent, lazy ;

                      self._noteLoadFailureId(this.id, false) ;  // if this id is in the failure list, remove
                      self._removeFromBusyStack('j', ((obj === -1)? -1 : obj.attr('id'))) ;

                      var sf = self._getOptions()["data"]["ajax"]["success"];  // reget the options
                      if (sf) {                                                // without our updated ajax
                        d = sf.call(self, d, status, x) || d;                       // changes to avoid forever loop
                      }

                      if (d === "" || (d && d.toString && d.toString().replace(/^[\s\n]+$/,"") === "")) {
                        return error_func.call(self, x, status, "");
                      }

                      //console.log("Ajax SUCCESS HTML id='" + this.id + "' obj.id=" +                        
                      //              ((obj === -1)? -1 : obj.attr("id")) + " url=" + this.url);
                      if (d)
                      {
                        d = $(d);
                        if (!d.is("ul")) {
                          d = $("<ul />").append(d);             //@HTMLUpdateOK
                        }
                        if (obj == -1 || !obj)
                        {
                          self._$container_ul.empty()
                                             .append(d.children())               //@HTMLUpdateOK
                                             .find("li, a")
                                             .filter(function ()
                                                {
                                                  return !self.firstChild || !self.firstChild.tagName ||
                                                                             self.firstChild.tagName !== "INS";
                                                })
                                             .prepend("<ins class='oj-tree-icon'>&#160;</ins>")    //@HTMLUpdateOK
                                             .end()
                                             .filter("a")
                                             .children("ins:first-child")
                                             .not(".oj-tree-node-icon")
                                             .addClass(OJT_NICON)
                                             .addClass(OJT_ICON);

                          self._$container_ul.find("li.oj-tree-leaf ins:first-child").removeClass(OJ_DISC)
                                                                                     .removeClass(OJT_NICON)
                                                                                     .addClass(OJT_ICON);
                          parent = self._$container_ul ;
                        }
                        else
                        {
                           obj.children("a.oj-tree-loading")
                              .removeClass(OJT_LOADING);

                           self._removeEmptyUL(obj) ;

                           obj.append(d)                           //@HTMLUpdateOK
                              .children("ul")
                              .find("li, a")
                              .filter(function () {
                                         return (!self.firstChild || !self.firstChild.tagName ||
                                                                     self.firstChild.tagName !== "INS");
                                      })
                              .prepend("<ins class='oj-tree-icon'>&#160;</ins>")     //@HTMLUpdateOK
                              .end()
                              .filter("a")
                              .children("ins:first-child")
                              .not(".oj-tree-node-icon")
                              .addClass(OJT_NICON)
                              .addClass(OJT_ICON);
                           obj.removeData("oj-tree-is-loading");

                           obj.find("li.oj-tree-leaf ins:first-child").removeClass(OJ_DISC)
                                                                      .removeClass(OJT_NICON)
                                                                      .addClass(OJT_ICON);
                           parent = obj ;
                        }

                        //  Look for parents with empty children <ul> list (lazy loading),
                        //  and add the closed class to make it a parent.
                        self._handleHtmlParentNoChildren(parent) ;

                        //  Add the <a> text <span> for hover/click styling
                        self._insertHtmlTextSpan(parent) ;

                        //  Escape all Ajax loaded node text for security
                        var nodes = parent.children("UL") ;
                        if (nodes.length >= 1)
                        {
                          nodes = nodes.first().find("span.oj-tree-title") ;
                          $.each(nodes, function() {
                                         this.textContent = self._escapeHtml(this.textContent) ;
                                         });
                        }

                        // If "default" type defined, apply to nodes with no assoc type
                        if (self._data.types.defType && parent)  {
                          self._addDefType(this._$container_ul) ;
                        }

                        self._cleanNode(obj);
                        if (s_call)  {
                          s_call.call(self);
                        }
                      }
                      else  {
                        if (obj && obj !== -1)  {
                           obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
                           obj.removeData("oj-tree-is-loading");
                           if (self._data.ds.correctState)  {
                              self._correct_state(obj);
                              if (s_call) {
                                s_call.call(self);
                              }
                           }
                          }
                          else  {
                            if (self._data.ds.correctState)  {
                               self._$container_ul.empty();
                               if (s_call) {
                                 s_call.call(self);
                               }
                            }
                          }
                          self._$container_ul.find("ul").attr(WA_ROLE, WA_GROUP) ;
                          self._$container_ul.find("li").attr(WA_ROLE, WA_TREEITEM) ;
                      }
                    };

                    //  Prepare for an asynchronous ajax op. (note: we are updating a copy of the options)

                    var nodeId = ((obj === -1)? -1 : obj.attr("id")) ;
                    this._busyStack.push( {op: 'j', id : nodeId} );
                    this._addBusyState("Tree (id='" + this._elemId + "') : node id='" + nodeId + "' is lazyloading.");
                    //console.log("ajax HTML load (" + nodeId + ") . . . busyStack=" + this._arrayToStr(this._busyStack));

                    var reqContext = {ctx: this, id: nodeId};
                    s["ajax"]["context"] = reqContext;
                    s["ajax"]["error"]   = error_func;
                    s["ajax"]["success"] = success_func;

                    if (!s["ajax"]["dataType"]) {
                      s["ajax"]["dataType"] = "html";
                    }
                    reqContext.type = s["ajax"]["dataType"] ;

                    if ($.isFunction(s["ajax"]["url"]))
                    {
                      s["ajax"]["url"] = s["ajax"]["url"].call(this, obj);
                      reqContext.url   = s["ajax"]["url"]
                    }
                    if ($.isFunction(s["ajax"]["data"]))
                    {
                      s["ajax"]["data"] = s["ajax"]["data"].call(this, obj);
                    }
                    $.ajax(s["ajax"]);
                    break;
       }
     },


     /**
       *   Look for parents with empty children <ul> list (lazy loading), and add the closed
       *   class to make it a parent.
       *   @private
       */
     _handleHtmlParentNoChildren : function (parent)
     {
        var lazy = parent
                     .find(parent.is("ul")?  "li ul" : "ul")
                     .filter(function () {
                                return (!this.firstChild ||
                                        this.childNodes.length == 0 ||
                                        (this.childNodes.length == 1 && this.firstChild.nodeType ==3));
                                       }) ;
        $.each(lazy, function() {
                                  $(this).closest("li")
                                         .addClass(TreeUtils._OJ_COLLAPSED) ;
                                }) ;
     },


     /**
       *   Remove an empty <ul> in a node (lazy loading).
       *   @private
       */
     _removeEmptyUL : function(parent)
     {
        var l = parent.find("ul")
                     .filter(function () {
                                return (!this.firstChild ||
                                        this.childNodes.length == 0 ||
                                        (this.childNodes.length == 1 && this.firstChild.nodeType ==3));
                                       }) ;
       if (l.length > 0) {
         l.remove() ;
       }
     },


     /**
       *   Load an HTML <ul><li>...</ul> markup string
       *   @private
       */
     _loadHtmlString : function (s, obj, s_call, e_call)
     {
        var parent ;

        if (s && s !== "" && s.toString && s.toString().replace(/^[\s\n]+$/,"") !== "") {
          s = $(s);
          if (! s.is("ul")) {
            s = $("<ul />").append(s);                     //@HTMLUpdateOK
          }

          if (obj == -1 || !obj) {
            this._$container_ul.empty()
                               .append(s.children())       //@HTMLUpdateOK
                               .find("li, a")
                               .filter(function () {
                                         return (!this.firstChild || !this.firstChild.tagName ||
                                                 this.firstChild.tagName !== "INS");
                                       })
                               .prepend("<ins class='oj-tree-icon oj-tree-disclosure-icon'>&#160;</ins>")  //@HTMLUpdateOK
                               .end().filter("a")
                               .children("ins:first-child")
                               .not(".oj-tree-node-icon")
                               .addClass(OJT_NICON)
                               .addClass(OJT_ICON);

            this._$container_ul.find("li.oj-tree-leaf ins:first-child").removeClass(OJ_DISC)
                                                                       .removeClass(OJT_NICON)
                                                                       .addClass(OJT_ICON);

            parent = this._$container_ul ;

            //  No point in checking for empty parent <ul> because we don't have a
            //  "data" option in the first place, so no lazy loading posssible.

            //  If the "default" node type has been defined, add the def type to any nodes
            //  that have not been given an explicit type
            this._addDefType(this._$container_ul) ;
          }
          else  {
            obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
            obj.append(s)                                                    //@HTMLUpdateOK
               .children("ul")
               .find("li, a")
               .filter(function ()
                          {
                             return !this.firstChild || !this.firstChild.tagName ||
                                                         this.firstChild.tagName !== "INS";
                          })
               .prepend("<ins class='oj-tree-icon'>&#160;</ins>")         //@HTMLUpdateOK
               .end().filter("a")
               .children("ins:first-child")
               .not(".oj-tree-node-icon")
               .addClass("oj-tree-node-icon")
               .addClass("oj-tree-icon");
            obj.removeData("oj-tree-is-loading");

            parent = obj ;

            //  If the "default" node type has been defined, add the def type to
            //  any nodes that have not been given an explicit type
            this._addDefType(this.obj) ;
          }

          //  Add the <a> text <span> for hover/click styling
          if (parent) {
            this._insertHtmlTextSpan(parent) ;
          }

          //  Escape all node text for security
          var nodes = parent.children("UL") ;
          if (nodes.length >= 1) {
            nodes = nodes.first().find("span.oj-tree-title") ;
            var _this = this ;
            $.each(nodes, function() {
                           this.textContent = _this._escapeHtml(this.textContent) ;
                           });
          }

          this._cleanNode(obj);
          if (s_call)  {
            s_call.call(this);
          }
        }
        else  {
          if (obj && obj !== -1)  {
            obj.children("a.oj-tree-loading").removeClass(OJT_LOADING);
            obj.removeData("oj-tree-is-loading");
            if (this._data.ds.correctState) {
              this._correct_state(obj);
              if (s_call) {
                s_call.call(this);
              }
            }
          }
          else  {
            if (this._data.ds.correctState)
            {
              this._$container_ul.empty();
              if (s_call) {
                s_call.call(this);
              }
            }
          }
        }
     },

     /**
       *  Insert the <a> text <span> for hover/click styling
       *  @private
       */
     _insertHtmlTextSpan : function(elem)
     {
        $.each(elem.find("li a"), function (i, val) {
             var ih = val.innerHTML ;                    //@HTMLUpdateOK
             ih = ih.replace("ins>", "ins><span class='" + OJT_TITLE + "'>") ;
             ih += "</span>" ;
             val.innerHTML = ih ;                        //@HTMLUpdateOK
        });
     },

     /**
       *  Add the "default" type attr to all nodes that do not have an explicit type defined.
       *  @private
       */
     _addDefType : function(obj)
     {
        var s, typeAttr ;
        if (this._data.types.defType) {
          s = this.options["types"];
          typeAttr = s? s["attr"] : this._data.types.defaults["attr"];

          $.each(obj.find("li"), function (i, val) {
                  val = $(val)
                  if (! val.attr(typeAttr)) {
                    val.attr(typeAttr, "oj-tree-deftype").addClass(OJT_TYPE) ;
                  }
          });
        }
     },


     /**
       *  @private
       */
     _save_opened : function ()
     {
        var _this = this;
        this._data.core.toExpand = [];
        this._$container_ul.find("li.oj-expanded").each(function () {
            if (this.id)  {
              _this._data.core.toExpand.push("#" + this.id.toString().replace(/^#/,"").replace(/\\\//g,"/").replace(/\//g,"\\\/").replace(/\\\./g,".").replace(/\./g,"\\.").replace(/\:/g,"\\:"));
            }
        });

        // this._emitEvent(_this._data.core.toExpand, "save_opened", true);
     },


     /**
       *  Reload tree nodes
       *  If this is not a callback request, then if initExpanded is defined,
       *  any specified (or all) collapsed nodes are  expanded (loading if neccesaary
       *  because of lazy-loading).
       *  @param {boolean=} bIsCallback  true if this method has been called from within this method.
       *  @private
       */
      _reload_nodes : function (bIsCallback)
      {
         var _this     = this,
             bDone     = true,
             current   = [],
             remaining = [],
             ea, n, id ;

         if (! bIsCallback)  {
            this._data.core.reopen     = false;
            this._data.core.refreshing = true;
         }

         // If option initExpanded "all", make list of nodes to expand
         if (this._isOptExpandAll())
         {
           this._data.core.toExpand  = [];
           ea = this._$container_ul.find("li.oj-collapsed").each(function()
           {
             id = '#' + $(this).attr("id") ;
             if (! _this["isExpanded"](id))
             {
               if (! _this._isLoadFailureId(id))
               {
                 _this._data.core.toExpand.push(id) ;
               }
             }
           }) ;
         }

         this._data.core.toExpand =
                  $.map(
                         $.makeArray(this._data.core.toExpand),
                             function (n) {
                                return "#" + n.toString().replace(/^#/, "")
                                                         .replace(/\\\//g, "/")
                                                         .replace(/\//g, "\\\/")
                                                         .replace(/\\\./g, ".")
                                                         .replace(/\./g, "\\.")
                                                         .replace(/\:/g, "\\:");
                             }
                       );

         this._data.core.toLoad =
                    $.map(
                           $.makeArray(this._data.core.toLoad),
                               function (n)  {
                                  return "#" + n.toString().replace(/^#/,"")
                                                           .replace(/\\\//g,"/")
                                                           .replace(/\//g,"\\\/")
                                                           .replace(/\\\./g,".")
                                                           .replace(/\./g,"\\.")
                                                           .replace(/\:/g,"\\:");
                         }
                    );

         if (this._data.core.toExpand.length)  {
           this._data.core.toLoad  = this._data.core.toLoad.concat(this._data.core.toExpand);

           // Remove duplicates
           this._data.core.toLoad = this._data.core.toLoad.reduce(function(a, b) {
                       if (a.indexOf(b) < 0 ) a.push(b);
                         return a;
                       }, []);
         }


         // Analyse combined expand/load list and create current and remaining arrays.

         if (this._data.core.toLoad.length) {
            $.each(this._data.core.toLoad, function (i, val) {
                if (val == "#") {
                  return true;
                }
                n = _this._$container.find(val) ;
                if (n.length) {                      // is node in the tree
                  if (! _this["isExpanded"](val)) {  // and not expanded
                    current.push(val);               // yes.
                  }
                }
                else {
                  remaining.push(val);               // no.
                }
            });

            if (current.length) {
               this._data.core.toLoad  =  remaining;
               $.each(current, function (i, val) {
                   if (! _this._is_loaded(val))  {        // if load not loaded,
                      _this._load_node(val,               // load it.
                                       function ()  {_this._reload_nodes(true);},    // success func
                                       function ()  {_this._reload_nodes(true);}     // error func
                                      );
                      bDone = false;        // note we're not complete yet since we just loaded a node
                   }
               });
            }
         }


         //  Expand the nodes referred to.

         if (this._data.core.toExpand.length)  {
            $.each(this._data.core.toExpand, function (i, val) {
                if (! _this["isExpanded"](val)) {
                  _this._expand(val, false, true);      // no animation for initExpand
                }
            });
         }

         if (bDone)  {
            // TODO: find a more elegant approach to syncronizing returning requests

            if (this._data.core.reopen)  {
              clearTimeout(this._data.core.reopen);
            }
            this._data.core.reopen = setTimeout(function ()
            {
              _this._emitEvent({}, "reload_nodes", true);
              _this._resolveIfBusyStackEmpty() ; // there may be nothing to expand
            }, 50);
            this._data.core.refreshing = false;
            this._reopen();                      // check expanded and selections
         }
    },


     /**
       *   Not exposed in V1
       *   @private
       */
    "setTheme" : function (theme_name, theme_url)
    {
       if (! theme_name)  {
         return false;
       }

       if (! theme_url)  {
          theme_url = this._data.themes._themes + theme_name + '/style.css';
       }

       if ($.inArray(theme_url, this._data.themes.themes_loaded) == -1)  {
         _addSheet({ url : theme_url });
         this._data.themes.themes_loaded.push(theme_url);
       }

       if (this._data.themes.theme != theme_name)  {
         this._$container.removeClass('oj-tree-' + this._data.themes.theme);
         this._data.themes.theme = theme_name;
       }

       this._$container.addClass('oj-tree-' + theme_name);

       if (! this._data.themes.dots) {
         this._hideDots();
       }
       else  {
          this._showDots();
       }

       if (! this._data.themes.icons)  {
         this._hideIcons();
       }
       else  {
         this._showIcons();
       }

//     this._emitEvent(null, "set_theme", true);
    },


    /*
     *   Shows the hierarchy lines.  Not used in V1
     *   @private
     */
/*
    "show_dots"  : function ()
    {
      this._showDots() ;
    },
*/
    /*
     *   Hides the hierarchy lines.  Not used in V1
     *   @private
     */
/*
    "hide_dots"   : function ()
    {
      this._hideDots() ;
    },
*/
    /*
     *   Toggles the current hierarchy line state.  Not used in V1
     *   @private
     */
/*
    "toggle_dots" : function ()
    {
       if (this._data.themes.dots) {
         this._hideDots();
       }
       else  {
         this._showDots"]();
       }
    },
*/


     /**
      *   @private
      */
    _isTheme  : function ()
    {
       return this._data.themes != null ;
    },


     /**
      *   @private
      */
    _getTheme  : function ()
    {
       return  this._data.themes.theme;
    },


     /**
      *   Returns true if node icons are currently displayed.  Users can find this state from options.
      *   @private
      */
    "isIcons"  : function ()
    {
       return this._data.themes.icons ;
    },


    /**
      *   Displays node icons. Users can set this state from options.
      *   @private
      */
    _showIcons  : function ()
    {
       this._data.themes.icons = true;
       this._$container.children("ul").removeClass("oj-tree-no-icons");
    },

    /**
     *   Hides node icons. Users can set this state from options.
     *   @private
     */
    _hideIcons  : function ()
    {
       this._data.themes.icons = false;
       this._$container.children("ul").addClass("oj-tree-no-icons");
    },

    /**
     *   Toggles the display state of node icons. User can set this state from options.
     *   @private
     */
    "toggleIcons" : function ()
    {
/*                                      Not used in V1
       if (this.data.themes.icons) {
         this._hideIcons();
       }
       else  {
         this._showIcons();
       }
*/
    },

    /**
      *   Enables keyboard support
      *   @private
      */
    _enableKeys : function ()
    {
       this._data.keys.enabled = true;
    },

//  /**
//    *   Disables keyboard support
//    *   @private
//    */
//         Not used in V1
//  _disableKeys : function ()
//  {
//     this._data.keys.enabled = false;
//  },
//

    /**
      *   Initializes the widget, examining options and setting up
      *   internal data structures.
      *   @private
      */
     _initTree : function()
     {
        this._initData() ;
        this._initCoreOpts() ;
        this._initUIOpts() ;
//      this._initCrrmOpts() ;      // not in V1
//      this._initThemeOpts()       // not in V1
        this._initDSOpts(true) ;
        this._initTypeOpts()
        this._getDndContext().initDnDOpts() ;
        this._initMenuOpts() ;

        this._initCore() ;
        this._initUI() ;
        this._initThemes() ;
        this._initDataSource() ;
//      this._initCrrm() ;          // not in V1
        this._initTypes() ;
        this._getDndContext().initDnD() ;
        this._initKeys() ;
        this._initMenu() ;
     },


     /**
       *  Emit events
       *  @param {Object} data an object containing details about the event.
       *  @param {string} evname the raw event name (e.g. "hover", or "expandAll".
       *                          When the event is fired, it is "ojhover" and "ojexpandall",
       *                          but also the options "hover" and "expandAll" are called.
       *  @param {boolean=} bInternal true if an internal event, or omit or set false for
       *                              a public event.
       *  @private
       */
     _emitEvent : function (data, evname, bInternal)
     {
        if ((! evname) || $.type(evname) !== "string") {
          return ;
        }
        if ((this._data.core.locked === true) && (evname !== "unlock")
                                              && (evname !== "isLocked")
                                              && (evname !== "lock")) {
          return ;
        }

        // console.log("Event " +  evname + (evname === "before"? " (" + data["func"] + ")" : "") +
        //             " public=" + (bInternal? false : true));

        var func,
            args     = Array.prototype.slice.call(arguments),
            rlbk     = false,
            inst     = this._$container,
            isBefore = (evname === "before"),
            isPublic = bInternal? false : true ;

        if (! isPublic)  {                       // internal event?
           evname = "_tree" + evname ;
        }

        // Build the "ui" argument
        var eventdata = {} ;
        eventdata["item"] = (data? data["obj"] : undefined) ;
        eventdata["inst"] = inst ;

        if (isBefore) {
          func               = data["func"] ;      // target method
          eventdata["func"]  = func ;
          eventdata["args"]  = args ;
          if (func === "rename") {
            eventdata["title"]     = data["title"] ;
            eventdata["prevTitle"] = data["prevTitle"] ;
          }
        }
        else  if (isPublic) {
          if (evname == "move") {
            eventdata["position"]  = data.p ;      // position relative to the reference node
            eventdata["reference"] = data.r ;      // the reference node
            eventdata["data"]      = data ;        // (req'd internally)
          }
          else if (evname == "rename") {
            eventdata["title"]     = data["title"] ;      // the new node title to be
            eventdata["prevTitle"] = data["prevTitle"] ;  // the current title
          }
          else if (evname == "remove") {            // node was deleted via context menu
            eventdata["parent"]  = data["parent"] ; // parent node
            eventdata["prev"]    = data["prev"] ;   // the "previous" node
          }
          else if (evname == "delete") {            // node was deleted via context menu
            eventdata["prev"]   = data["prev"] ;    // (req'd internally) - the "previous" node
            eventdata["parent"] = data["parent"] ;  // parent node
          }
          else if (evname === "expandAll" || evname === "collapseAll" || evname === "deselectAll") {
            eventdata["targ"]   = data["targ"] ;    // the target of the op (node or -1)
          }
          else if (evname === "loaded") {
            eventdata["item"] = -1 ;
          }
          else if (evname == "paste") {
            eventdata["position"]  = data.p ;      // position relative to the reference node
            eventdata["reference"] = data.r ;      // the reference node
          }
        }

        if (isPublic) {
          var bContinue = this._trigger(evname, new $.Event("oj" + evname), eventdata) ;

          if (isBefore) {
            //  console.log("Event " + evname + ") " + eventdata["func"] + " returned " + rslt);
            if (typeof bContinue != "undefined") {
              bContinue = bContinue? true : false ;         // returns true/false/undefined
            }
            return bContinue ;
          }
        }
        else {
          this._$container.trigger(evname, eventdata) ;    // internal event
        }

     },


     /**
       *  Fire optionChange event
       *  @param {string}  key             the option key whose property value has been changed.
       *  @param {Object}  prevVal         the previous value prior to the change
       *  @param {Object | null}  newVal   the new value after the change
       *  @param {Event | null} origEvent  false if option change is not due to user interaction.
       *  @private
       */
     _fireOptionChangeEvent : function(key, prevVal, newVal, origEvent)
     {
        if (key === "selection")  {
          if (newVal == null) {
            newVal = this._getSelected() ;
          }

          // Only fire the event if the new and prev vals are different.

          if (! this._compareSelectionValues(prevVal, newVal))  {
            this.option(key, newVal, {'_context' : { originalEvent: origEvent,
                                                     internalSet: true
                                                   },
                                      'changed': true     // don't need comparison check
                                     });
          }    // end if "selection"
        }
     },

     /**
       *  TDO _not currently used in V1
       *  @private
       */
     __rollback : function ()
     {
         var
         rlbk = this["get_rollback"]();
         return rlbk;
     },

     /**
       *  Tree initialization is complete.  Build and render the tree.
       *  @private
       */
     _start : function()
     {
       if (this._isRtl) {
         this._$container.addClass("oj-tree-rtl").css("direction", "rtl");
       }

       //  Create the outer <ul> with a temporary <li> saying "Loading..."

       this._$container.html("<ul role='tree' tabindex='0' class='oj-tree-list' style='outline:none'" +       //@HTMLUpdateOK
                    ((this._data.core.selectMode === -1)? " aria-multiselectable='true'" : "") +
                    " " + WA_LABELLEDBY + "='" + this._elemId + "'" + 
                    "><li class='oj-tree-last oj-tree-leaf'><ins class='oj-tree-icon'>&#160;</ins><a class='oj-tree-loading' href='#'><ins class='oj-tree-icon'>&#160;</ins>" +
                     this._getString(TRANSKEY_LOADING) + "</a></li></ul>");

       this._$container_ul = this._$container.children("ul:eq(0)");
       this._$container.data("oj-tree-instance-id", this._getIndex());

       this._data.core.li_height = this._$container_ul.find("li.oj-collapsed, li.oj-tree-leaf")
                                                      .eq(0)
                                                      .height() || 18;

       if (this._isTouch) {             // for tap expand/collapse
         this._$container
              .delegate(".oj-tree-list ins.oj-tree-disclosure-icon", "touchend.ojtree", 
                     function (event)  {
                        event.preventDefault() ;        // don't want browser click generated after 300ms
                        var trgt = $(event.target);
                        this["toggleExpand"](trgt);
                     }.bind(this));
       }

       this._$container
              .delegate(".oj-tree-list ins.oj-tree-disclosure-icon", "click.ojtree",
                     function (event)  {
                       var trgt = $(event.target);
                       // if(trgt.is("ins") && event.pageY - trgt.offset().top < this.data.core.li_height) { this.toggle_node(trgt); }
                       this["toggleExpand"](trgt);
                     }.bind(this))
              .delegate(".oj-tree-list ins.oj-tree-disclosure-icon", "mousedown",
                    function (event)  {
                       this._data.ui.disclosureClick = true ;  // if no prev hover/selection, this will
                                                               // prevent hover going to 1st node when
                                                               // a disclosure has been clicked as the first action
                                                               // on the tree (see focus event below)
                       var trgt = $(event.target);
                       trgt.removeClass("oj-default")
                           .removeClass(TreeUtils._OJ_HOVER)
                           .addClass(TreeUtils._OJ_SELECTED) ;
                    }.bind(this))
              .delegate(".oj-tree-list ins.oj-tree-disclosure-icon", "mouseup",
                    function (event)  {
                       var trgt = $(event.target);
                       trgt.removeClass(TreeUtils._OJ_SELECTED).addClass("oj-default") ;
                    }.bind(this))
              .bind("mousedown.ojtree", function (event)  {
                      this._setFocus();
                      // . If IE11, need preventDefault() to avoid weird
                      // shift-click highlighting. Not done if HTML5 drag/drop is supported 
                      if (event.shiftKey) {
                        if (! (this._getDndContext() && this._getDndContext().isDragEnabled())) {
                          event.preventDefault() ;
                        }
                      }
                    }.bind(this))
              .bind("dblclick.ojtree", function (event)  {
                     var sel;
                     if (document.selection && document.selection.empty) {
                       document.selection.empty();
                     }
                     else  {
                       if (window.getSelection)  {
                         sel = window.getSelection();
                         try {
                               sel.removeAllRanges();
                               sel["collapse"](document.getElementsByTagName("body")[0], 0);
                             } catch (err) { }
                       }
                     }
                });

       this._$container_ul
               .focus(function(e) {
                        if (this._data.ui.disclosureClick) {         // if very first action is
                          this._data.ui.disclosureClick = false ;    // mousedown on a disclosure icon
                          return false;                              // don't set hover (and prevent jump to top of list)
                        }

                        this._data.ui.focused = true ;
                        var n ;

                        // Tree has gained focus, highlight a node as current.
                        // If no last "current" or selected, will default to the top node. 
                        if (this._data.ui.hovered)    // . Check if mouse is over a node, before
                        {                             // defaulting to another node as described above.
                          n = this._data.ui.hovered ;
                        }
                        else  if (this._data.ui.lastHovered) {
                          this._data.ui.hovered =  this._data.ui.lastHovered ;
                          n = this._data.ui.hovered ;
                        }
                        else if (this._data.ui.lastSelected && (this._data.ui.lastSelected.length > 0)) {
                          this._data.ui.hovered =  this._data.ui.lastSelected ;
                          n = this._data.ui.hovered ;
                        }
                        else {
                          n = this._$container_ul.find("li:first");
                        }
                        if (n) {
                          this["hover"](n);
                          this._data.ui.lastHovered = null ;

                          // want color change for selected inactive to selected
                          this._$container_ul.find("a.oj-selected").removeClass(OJT_INACTIVE) ;
                        }
                        return false ;    // prevent jump to top of list
                      }.bind(this)
               )
               .blur(function(e) {
                        this._data.ui.focused     = false ;
                        this._data.ui.lastHovered = this._data.ui.hovered ;
                        if (this._data.ui.lastHovered) {
                          this["dehover"](this._data.ui.hovered) ;
                        }
                        // want color change for selected active to selected inactive
                        this._$container_ul.find("a.oj-selected").addClass(OJT_INACTIVE) ;
                     }.bind(this)
               );

/*     Not used in V1
       if (this._data.core.load_open)   {
         this._$container
            .bind("_treeload_node", $.proxy(function (e, ui)  {
                 var o = this._getNode(ui["item"]),
                     t = this;
                 if (o === -1)  {
                   o = this._$container_ul;
                 }
                 if (!o.length)  {
                   return;
                 }
                 o.find("li.oj-expanded:not(:has(ul))").each(function ()
                     {
                        this._load_node(this, $.noop, $.noop);
                     });
            }, this));
       }
*/
       this._emitEvent({}, "init", true);  // Tree is now init'd (but nodes have not yet been added)

       this._loadNodes() ;                 // start node loading from the datasource

       //  Tree div is now constructed, note if context menu has been set.
       if (this._data.menu.usermenu) {
         this._applyMenu() ;
       }

       // Add and enable this tree's keystroke handler
       _addKeyFilter( {
                        _handler : this._keyHandler,
                        _selector: this._$container_ul,
                        _this    : this
                      }) ;
       this._enableKeys();
     },


     /**
       *  Initialize the Core section
       *  @private
       */
     _initCore : function()
     {
       this._data.core.locked   = false;

       this._$container.addClass("oj-tree oj-tree-" + this._getIndex()) ;     //TDO

       this._$container.css("outline", "none");
       this._$container.css("MozUserSelect", "none");
       this._$container.css("WebkitTouchCallout", "none");
       this._$container.css("WebkitUserSelect", "none");
       this._$container.css("-ms-user-select", "none");
       this._$container.css("WebkitTapHighlightColor", "rgba(0,0,0,0)");
     },


    /**
      *  Initialize the UI section handlers
      *  @private
      */
     _initUI : function()
     {
         this._data.ui.selected      = $();
         this._data.ui.lastSelected  = false;
         this._data.ui.hovered       = null;
         var a = this.options["selection"] ;

         if (a && ($.type(a) === "array") && a.length > 0) {
           this._data.ui.toSelect    = a ;
           this.options["selection"] = [] ;  // will be updated when the selections are performed
         }

        // Bind to events

        if (this._isTouch) {             // for tap select/deselect
          this._$container
             .delegate(".oj-tree-list a", "touchend.ojtree", $.proxy(function (event)
               {
                 // This becomes the hovered node
//               if (! $(event.currentTarget).hasClass(OJT_LOADING))  {
//                 this["hover"](event.target) ;
//               }

                 this._data.ui.touchEvent = true ;
                 this._handleNodeTapClick(event) ;   // select the node, or toggle if multi-select

                 //  Since this is a tap, no node is hovered
                 if (! $(event.currentTarget).hasClass(OJT_LOADING))  {
                   this["dehover"]();
                 }
               }, this));
        }

        //-----------------------------------------------------------------//
        //  Handle events (particularly post operation events like move()  //
        //  remove(), collapse(), etc.                                     //
        //-----------------------------------------------------------------//

        this._$container
             .delegate(".oj-tree-list a", "click.ojtree",
                        function (event)
                        {
                          this._data.ui.touchEvent = false ;
                          this._handleNodeTapClick(event) ;
                        }.bind(this))
             .delegate(".oj-tree-list a", "mouseenter.ojtree",
                        function (event)
                        {
                           if (! $(event.currentTarget).hasClass(OJT_LOADING))  {
                             this["hover"](event.target);
                           }
                        }.bind(this))
             .delegate(".oj-tree-list a", "mouseleave.ojtree", 
                        function (event)
                        {
                          if (! $(event.currentTarget).hasClass(OJT_LOADING))  {
                            this["dehover"](event.target);
                          }
                        }.bind(this))
             .delegate(".oj-tree-list .oj-tree-disclosure-icon", "mouseenter.ojtree",
                       function (event)
                       {
                         if (! $(event.currentTarget).hasClass(OJT_LOADING))  {
                           this._disclosureHover(event.target, true);
                         }
                       }.bind(this))
             .delegate(".oj-tree-list .oj-tree-disclosure-icon", "mouseleave.ojtree",
                        function (event)
                        {
                          if (! $(event.currentTarget).hasClass(OJT_LOADING))  {
                            this._disclosureHover(event.target, false);
                          }
                        }.bind(this))
             .bind("_treereopen", function ()
                {
                  this._reselect();
                }.bind(this))
             .bind("_treeget_rollback", function ()
                {
                  this["dehover"]();
                  this._saveSelected();
                }.bind(this))
/*   Not used in V1
             .bind("_treeset_rollback", $.proxy(function ()
               {
                  this["reselect"]();
               }, this))
*/
             .bind("ojcollapse", function (event, ui)
              {
                var obj   = this._getNode(ui["item"]),
                    clk   = (obj && obj.length) ? obj.children("ul").find("a.oj-selected") : $(),
                    _this = this;
                if (this.options["selectedParentCollapse"] === false || (! clk.length))  {
                  return;
                }
                clk.each(function ()  {
                               if (_this.options["selectedParentCollapse"] === "selectParent")  {
                                  _this["deselect"](this);
                                  _this["select"](obj);
                               }
                               else if (_this.options["selectedParentCollapse"] === "deselect")  {
                                  _this["deselect"](this);
                               }
                         });
                 }.bind(this))
             .bind("ojremove", function (event, ui)         // node delete
                {
                   var s     = this.options["selectPrevOnDelete"],
                       obj   = this._getNode(ui["item"]),
                       clk   = (obj && obj.length) ? obj.find("a.oj-selected") : [],
                       _this = this,
                       n ;

                   clk.each(function ()
                              {
                                 _this["deselect"](this);

                                 // Remove lastHovered if it refers to the node removed
                                 if (_this._data.ui.lastHovered) {
                                   n = _this._getNode(this) ;
                                   if (n && (_this._data.ui.lastHovered.attr("id") == n.attr("id"))) {
                                     _this._data.ui.lastHovered = null ;
                                   }
                                 }
                              });
                   if (s && clk.length)  {
                     if (ui["prev"]) {
                       ui["prev"].each(function ()
                          {
                            if (this["parentNode"]) {
                              _this["select"](this);
                              return false;   // if return false is removed all prev nodes will be selected
                            }
                          });
                     }
                   }
                 }.bind(this))
             .bind("ojmove", function (event, ui)
                {
                   var data = ui["data"],
                       copy = data.cy,
                       p ;

                   if (copy && data.oc)  {
                     data.oc.find("a.oj-selected").removeClass(TreeUtils._OJ_SELECTED);
                     data.oc.removeAttr(WA_SELECTED);
                   }

                   // If the node being moved (not copied) is from a different
                   // tree, and the node was noted as lastHovered or lastSelected
                   // in the tree when it lost focus, clear that tree's lastHovered/
                   // lastSelected.

                   if ((data.ot !== data.rt) && (!copy)) {     // different trees and a move op?
                     p = data.ot._data.ui ;
                     if (p.lastHovered) {
                       if (data.o.attr("id") == p.lastHovered.attr("id")) {
                         p.lastHovered = null ;
                       }
                       if (p.lastSelected) {
                         if (data.o.attr("id") == p.lastSelected.attr("id")) {
                           p.lastSelected = null ;
                         }
                       }
                     }
                   }
                 }.bind(this));

     },

     /**
       *  Initialize the data source
       *  @private
       */
     _initDataSource : function()
     {
        this._initTreeData() ;
        this._initJsonData() ;
        this._initHtmlData() ;
     },

     /**
       *  Initialize the tree data source
       *  @private
       */
     _initTreeData : function()
     {
        if (this._data.ds.type === DS_TREE) {
          this._tds        = (this.options["data"] || null) ;     // the tree data source
          this._load_node  = this._load_node_DS ;
          this._is_loaded  = this._is_loaded_DS ;
          this._refresh    = this._refresh_DS ;
       }
     },

     /**
       *  Initialize the json_data section if requested
       *  @private
       */
     _initJsonData : function()
     {
        if (this._data.ds.type === DS_JSON) {
          if (this._data.ds.progressiveUnload)  {
            this._$container.bind("_treeafter_close", function (e, ui)  {
                                                         ui["item"].children("ul").remove();
                                                      });
          }

          this._load_node  = this._load_node_J ;
          this._is_loaded  = this._is_loaded_J ;
          this._refresh    = this._refresh_json ;
        }
     },


     /**
       *  Initialize html_data support if requested.
       *  @private
       */
     _initHtmlData : function()
     {
       if (this._data.ds.type !== DS_HTML) {
         return ;
       }

       this._processExistingMarkup() ;         // check for existing markup in the div

       this._load_node = this._load_node_H ;
       this._is_loaded = this._is_loaded_H ;
       this._refresh   = this._refresh_ui ;
     },


     /**
      *  If there is existing html markup in the div, clone it and move it to another div.
      *  @private
      */
     _processExistingMarkup : function ()
     {
       if (! this._data.html.useExistingMarkup) {
         return ;
       }

       if (! this._data.html.markup_ul) {  // first time find the user <ul>
         this._data.html.markup_ul   = this._$container.find(" > ul");          // user's <ul>

         // Move the user <ul> in the tree div to another div so that it is still in the DOM,
         // and can be located by the app, changed, and refresh() called if desired, and so it
         // will not interfere with the <ul> we maintain for the tree in the tree div. If the
         // app wants to change the markup, it must find its <ul> first and then find node id's
         // from it if it needs to, because the <ul> has been cloned and the id's will
         // therefore be duplicated.
         // TDO - we need to have a better way to do this.

       this._data.html.markup      = this._data.html.markup_ul.find(" > li");   // user's <li>'s
       this._data.html.cloneMarkup = this._data.html.markup.clone(true);        // our clone of the <li>'s

       // this used to use html() and clean the whitespace, but this way any attached data was lost
       // remove white space from LI node - otherwise nodes appear a bit to the right
       this._data.html.cloneMarkup.find("li").addBack()
                                             .contents()
                                             .filter(function()
                                                {
                                                  return this.nodeType == 3;
                                                 })
                                             .remove();


         this._transformElemIds(this._data.html.markup_ul) ;     // make user's markup id's unique 

         this._data.html.markup_div  = $("<div id='" +
                                       USER_UL_ID_PREFIX + this._getIndex() +
                                       "' style='display:none'>").append(this._data.html.markup_ul) ;   //@HTMLUpdateOK
         this._$container.after(this._data.html.markup_div) ;                                           //@HTMLUpdateOK
       }
     },


     /**
       *  Transform the supplied element Id attributes to make them unique. The tree index is prepended
       *  withe tree index.
       *  @private
       */
     _transformElemIds : function(markup)
     {
        var $nodes = $(markup).find("li"),
            index  = this._getIndex() ;

        $nodes.each(function(i, node) {
            node.setAttribute("id", index + "_" + node.getAttribute("id")) ;
        });
     },


     /**
       *  Untransform the supplied element Id attributes by removing the prepended tree index
       *  (see _transformElemIds).
       *  @private
       */
     _unTransformElemIds : function(markup)
     {
        var id, n,
            $nodes = $(markup).find("li"),
            index  = this._getIndex() ;

        $nodes.each(function(i, node) {
            id = node.getAttribute("id") ;
            n  = id.indexOf("_") ;
            id = id.substr(n+1) ;             // original user id follows the underscore
            node.setAttribute("id", id) ;
        });
     },


     /**
       *  Initialize the Themes section   Not used in V1
       *  @private
       */
     _initThemes : function()
     {
/*
        // autodetect themes path
        if (this._data.themes._themes === false)  {
           $("script").each(function ()
            {
              if( this.src.toString().match(/jquery\.oj-tree[^\/]*?\.js(\?.*)?$/))  {
                 this._data.themes._themes = this.src.toString().replace(/jquery\.oj-tree[^\/]*?\.js(\?.*)?$/, "") + 'themes/';
                 return false;
              }
            });
        }
*/
        if (this._data.themes._themes === false)    {
          this._data.themes._themes = "themes/";
        }

        this._$container
             .bind("_treeinit", $.proxy(function ()
                {
                   var s = this.options ;

                   this._data.themes.dots  = s["dots"];
                   this._data.themes.icons = s["icons"];
                   this["setTheme"](this._data.themes.theme, this._data.themes.url);

                }, this))

             .bind("ojloaded", $.proxy(function ()
                {
                   // bound here too, as simple HTML tree's won't honor dots & icons otherwise
                   if (! this._data.themes.dots)  {
                     this._hideDots();
                   }
                   else  {
                     this._showDots();
                   }
                   if (!this._data.themes.icons)  {
                     this._hideIcons();
                   }
                   else  {
                     this._showIcons();
                   }
                }, this));

     },


     /**
       *  Initialize the icon type functionality
       *  @private
       */
     _initTypes : function()
     {
        var s = this.options["types"];
        if (! s) {
          return ;
        }

        this._$container
               .bind("_treeinit", $.proxy(function ()
                  {
                    var types     = $.extend(true, {}, s["types"]),
                        attr      = s["attr"] || this._data.types.defaults["attr"],
                        icons_css = "",
                        _this     = this;

                    $.each(types, function (i, tp)
                       {
                         $.each(tp, function (k, v)
                            {
                              if (!/^(maxDepth|maxChildren|icon|validChildren)$/.test(k))  {
                                _this._data.types.attachTo.push(k);
                              }
                            });

                         // For ojTree we allow image and position props to not
                         // have to be in an icon object like jsTree.
                         var ot = (typeof tp["icon"]) ;
                         if (ot === "undefined") {
                           ot = (typeof tp["image"]) ;
                           if ((ot === "boolean") && (!tp["image"])) {
                             tp["image"] = "ojt$none" ;
                           }
                           else if ((! tp["image"]) && (! tp["position"])) {
                             return true;
                           }

                           tp["icon"] = {} ;
                           if (tp["image"])  {
                             tp["icon"]["image"] = tp["image"];
                             delete tp["image"] ;
                           }

                           if (tp["position"] !== undefined)  {
                             tp["icon"]["position"] = tp["position"] ;
                             delete tp["position"] ;
                           }
                         }

                         if ( tp["icon"]["image"] || tp["icon"]["position"])  {
                           if (i == "default")  {
                             _this._data.types.defType = true ;     // note that the "default" type has been defined       icons_css += '.oj-tree-' + _this._getIndex() + ' a > .oj-tree-icon { ';
                             icons_css += '.oj-tree-' + _this._getIndex() + ' .oj-tree-list li.oj-tree-type a > .oj-tree-node-icon { ';
                             icons_css += _this._addTypeCss(tp, icons_css) ;
                             icons_css += '} ';
                             icons_css += '.oj-tree-' + _this._getIndex() + ' .oj-tree-list li[' + attr + '="oj-tree-deftype"].oj-tree-type > a ins.oj-tree-node-icon { ';
                           }
                           else if (tp["icon"]["image"]) {
                             icons_css += '.oj-tree-' + _this._getIndex() + ' .oj-tree-list li[' + attr + '="' + i + '"].oj-tree-type > a > ins.oj-tree-node-icon { ';
                           }

                           icons_css += _this._addTypeCss(tp, icons_css) ;
                           icons_css += '} ';
                         }
                    });

                    if (icons_css !== "")  {
                      _addSheet({ str : icons_css, title : "oj-tree-types" });
                    }
                   }, this))
               .bind("ojbefore", function (e, data)
                  {
                     var s,
                         ty,
                         func = data["func"],
                         item = data["item"],
                         o = this._data.types.defaults["useData"] ? this._getNode(item) : false,
                         d = (o && o !== -1 && o.length)? o.data("oj-tree") : false;

                     if (d && d["types"] && d[func] === false)
                     {
                       e.stopImmediatePropagation();
                       return false;
                     }

                     if ($.inArray(func, this._data.types.attachTo) !== -1)  {
                       if (!data["item"] || (!data["item"]["tagName"] && !data["item"]["jquery"]))  {
                         return;
                       }

                       if (! this._canTypedNodeMove(item, func)) {
                         e.stopImmediatePropagation();
                         return false;
                       }
                     }

                  }.bind(this));

     },

     /**
       *  Add css for the type.
       *  @private
       */
     _addTypeCss : function(tp)
     {
         var css = "" ;

         if (tp["icon"]["image"] !== "ojt$none")  {
           css += ' background-image:url(' + tp["icon"]["image"] + '); ';
         }
         else  {
           css += " background-image:none; ";
         }

         if (tp["icon"]["position"])  {
           css += ' background-position:' + tp["icon"]["position"] + '; ';
         }
         else  {
           css += ' background-position:0 0; ';
         }
         return css ;
     },


     /*
      *  Returns the DnD support context object.
      *  @private
      */
     _getDndContext : function()
     {
        if (! this._TreeDndContext) {
          this._TreeDndContext = new oj.TreeDndContext(this);
        }
        return this._TreeDndContext;
     },


     /*
      *  Initialize the context menu cut/copy/paste/remove/rename support.
      *  @private
      */
/*                                Not currently used in V1
     _initCrrm :  function()
     {
        this._$container
               .bind("ojmove",
                    $.proxy(function (e, data)
                              {
                                //if (this.options["crrm"]["move"]["openOnMove"])
                                if (this._data.crrm.defaults["move"]["openOnMove"]) {
                                  var t = this;
//
//                                data.rslt.np.parentsUntil(".oj-tree").addBack()
//                                                                     .filter(".oj-tree-closed")
//                                                                     .each(function ()
//                                                                       {
//                                                                        t["expand"](this, false, true);
//                                                                       });
//
                                }
                              }, this));
     },
*/

     /**
       *  Initialize keyboard support
       *  @private
       */
     _initKeys  :  function()
     {
     },


     /**
       *  Initialize the context menu.  This is called on startup, or on option
       *  "contextMenu" change.
       *  @param {Object=} newVal   true if called because of an option change.
       *  @private
       */
     _initMenu : function(newVal)
     {
       var  menu, t, html, $html ;

// Temp code for M6 to clear .  Once the new methods
// are in place via Max and Jim, this whole mechanism will change.

if ((! newVal) && (! this.options["contextMenu"])) {
 menu = this._$container.attr("contextmenu") ;   // check for attribute
 if (menu) {
   this.options["contextMenu"] = "#" + menu ;
 }
}

       menu =  newVal || this.options["contextMenu"] ;
       if (!menu) {
         return ;
       }

       var $m = $(menu) ;                    // get the user's <ul> list
       if ($m) {
         $m.css("display", "none") ;         // ensure it's not visible
         var  dm       = this._data.menu ;
         dm.$container = $m ;
         dm.usermenu   = true ;              // have a context menu
       }

       if (this._data.menu.usermenu) {       // if we have a context menu
         if (newVal) {                       // and it is being changed
           this._applyMenu() ;               // complete menu creation/attachnment
         }
       }

       //  If not a new val from options, Menu will be noted at the end of initialization in _start()
     },


     /**
       *   Check menu selected to see if it one of our predefined remove/cut/copy/paste id's
       *   @private
       */
     _handleContextMenuSelect: function(ev, ui)
     {
        //  Make sure its a menu ojselect and not from ojTree
        if (ui["inst"]) {
          return ;
        }

        // There may be multiple trees sharing (and thereby listening to) the same
        // context menu.  Check that this tree is the same as the tree on which the
        // menu was invoked.
        if (this._data.menu.treeDivId != this._elemId) {
          return ;
        }

        var  id = ui? ui.item.attr("id") : undefined ;

        if (id === "ojtreecopy") {
          this._crrm_copy(this._data.menu.node);
        }
        else if (id === "ojtreecut") {
          this._crrm_cut(this._data.menu.node);
        }
        else if ((id === "ojtreepaste") || (id === "ojtreepastebefore") || (id === "ojtreepasteafter")) {
          this._crrm_paste(this._data.menu.node, id);
        }
        else if (id === "ojtreeremove") {
          if (this._isSelected(this._data.menu.node)) {
            this._crrm_remove();
          }
          else {
           this._crrm_remove(this._data.menu.node);
          }
        }
        else if (id === "ojtreerename") {
          this._crrm_rename(this._data.menu.node);
        }
        else if (id === "ojtreecreate") {
          this._crrm_create(this._data.menu.node);
        }
     },

    /**
      *  Notification that the user has invoked the context menu via the default
      *  gestures: right-click, pressHold, and Shift-F10.
      *  We ensure that focus returns to the tree if the menu is dismissed in
      *  some way via the keyboard.
      *  @param {Object} menu       The JET Menu to open as a context menu
      *  @param {Event}  event      Triggering event
      *  @param {string} eventType  Triggering event type (e.g. 'keyboard', 'mouse', 'touch')
      *  @private
      */
    _NotifyContextMenuGesture: function(menu, event, eventType)
     {
        //this._trace("_NotifyContextMenuGesture() eventType=" + eventType + " event.type=" + event.type) ;
        var keyboard = (eventType === 'keyboard'),
            textElem, $nodeElem ;

        if ((event.type != "contextmenu") && (!keyboard) && (eventType != 'touch')) {
          return ;
        }

        //  Get the tree node acted on
        //  - if no hovered node, use the last selected node
        this._data.menu.node       = keyboard? (this._data.ui.hovered || this._data.menu.activeNode) :
                                               $(event.target);
        this._data.menu.activeNode = null ;    // clear keyboard active node

        if (! this._data.menu.node) {          // is there an active (hovered) or last selected node ?
          event.preventDefault() ;
          return ;
        }

        // () Check that right-click/touch is not on whitespace near a node.
        if (!keyboard)
        {
          var $origTarg = $(event.originalEvent.target);
          if ($origTarg.is("LI") || $origTarg.hasClass(OJT_TREE) || $origTarg.is("UL"))
          {
            return ;   // no context menu
          }
        }

        $nodeElem = keyboard? this._data.menu.node : this._data.menu.node.closest("li") ;
        //console.log("_NotifyContextMenuGesture node=" + $nodeElem.attr("id"));

        this._data.menu.treeDivId = this._data.menu.node.closest("div").attr("id") ;

        var openOptions = {"launcher" : this._$container_ul};
        if (keyboard) {
          // Set position relative to the node in the SHIFT+F10 case.
          // Set here to avoid conflicting with user override in "before open" event
          textElem  = this._data.menu.node.find('.oj-tree-title')[0] ;  // the node text <span>
          openOptions["position"] = { "of" : textElem } ;
        }

        // Set submenu "Paste" disable state, depending on whether there's been a
        // previous "cut" or "copy"
        if (this._data.menu.usermenu && (this._data.menu.$elemPaste || this._data.menu.$elemPasteBefore ||
                                                                       this._data.menu.$elemPasteAfter)) {
          var disabledState = (!this._data.crrm.ct_nodes && !this._data.crrm.cp_nodes) ;
          var state ;
          var items = [this._data.menu.$elemPaste, this._data.menu.$elemPasteBefore, this._data.menu.$elemPasteAfter] ;
          var $item, i, refresh = false ;

          for (i = 0; i < items.length; i++) {
             $item = items[i] ;
             if ($item) {
               state  = !! $item.hasClass(TreeUtils._OJ_DISABLED) ;
               if (state != disabledState) {
                 if (disabledState) {
                   $item.addClass(TreeUtils._OJ_DISABLED) ;
                 }
                 else {
                   $item.removeClass(TreeUtils._OJ_DISABLED) ;
                 }
                 refresh = true ;
               }
             }
          }
          if (refresh) {
            this._data.menu.$container.ojMenu("refresh") ;
          }
        }

        this._OpenContextMenu(event, eventType, openOptions);
     },

     /**
       *  Process Core options
       *  @private
       */
     _initCoreOpts : function()
     {
        var val = this.options["selectionMode"] ;

        val = (val == undefined)? "single" : val ;
        if (val === "none") {
          val = 0 ;
        }
        else if (val === "single") {
          val = 1;
        }
        else if (val === "multiple") {
          val = -1 ;
        }
        this._data.core.selectMode = val ;

        this._data.themes.icons = this.options["icons"] ;

        this._initExpandedOpts();             // (updates this._data.coreToExpand)
        this._data.core.toLoad   = this.options["initLoaded"];
     },


     /**
       *  Process UI type options
       *  @private
       */
     _initUIOpts : function()
     {
     },

     /**
       *  Process data source options
       *  @private
       */
     _initDSOpts : function(bInit)
     {
        var s = this.options["data"],
            dt,
            ot ;

        this._data.ds.type = DS_NONE ;
        this._data.html.useExistingMarkup  = false ;
        this._data.html.cloneMarkup        = false ;
        this._loadFailureIds = {} ;

        // _data.html.markup_ul, _data.html.markup and  _data.html.markup_div
        // are left intact for destroy.

        if (s)
        {
          ot = $.type(s) ;
          if (ot === "string") {
           if (this._isHtml(s)) {
             this._data.ds.type = DS_HTML ;          // we have an non-Tree DS html source
           }
           else {
             this._data.ds.type = DS_JSON ;          // we have a non-tree DS json source
             this._initJsonOpts() ;
           }
          }
          else if (ot === "array") {
             //  we have an array of local json objects
             this._data.ds.type = DS_JSON ;          // we have a non-tree DS json source
          }
          else if (ot === "object") {
             try
             {                                           // don't crash if no jsontreedatasource
               if (s instanceof oj.JsonTreeDataSource) {
                 this._data.ds.type = DS_TREE ;          // we have a tree DS source
               }
             }
             catch(e)
             {
                this._data.ds.type = DS_ERROR
             } ;
             if (this._data.ds.type !== DS_TREE) {
               try
               {                                           // don't crash if no data source
                  if (s instanceof oj.CollectionTreeDataSource) {
                    this._data.ds.type = DS_COLLECTION ;   // we have a tree DS source
                  }
               }
               catch(e)
               {
                  this._data.ds.type = DS_ERROR
               } ;
             }
             if ((this._data.ds.type === DS_TREE)   ||
                 (this._data.ds.type === DS_COLLECTION))  {
               this._initTreeDSOpts() ;
             }
             else {
               if (s["data"] || s["ajax"])  {
                 dt = s["dataType"] ;
                 if (dt) {
                   if (dt === "json") {
                     this._data.ds.type = DS_JSON ;    // we have a non-tree DS json source
                     this._initJsonOpts() ;
                   }
                   else if (dt === "html") {
                     this._data.ds.type = DS_HTML ;    // we have a non-tree DS html source
                     this._initHtmlOpts() ;
                   }
                 }
                 else {
                   s["dataType"]      = "json" ;
                   this._data.ds.type = DS_JSON ;      // we have a non-tree DS json source
                   this._initJsonOpts() ;
                 }
               }
             }
          }
        }

        //  If no data defined, see if there is initial user markup in the div
        if (bInit && this._data.ds.type == DS_NONE) {
          if (this._$container.find("ul").length > 0) {
            this._data.ds.type                = DS_HTML ; // note we have an non-Tree DS html source
            this._data.html.useExistingMarkup = true ;
          }
        }
     },

     /**
       *  Process Tree DataSource options
       *  @private
       */
     _initTreeDSOpts : function()
     {
     },

     /**
       *  Process json_data options
       *  @private
       */
     _initJsonOpts : function()
     {
     },

     /**
       *  Process html_data options
       *  @private
       */
     _initHtmlOpts : function()
     {
     },

     /**
       *  Process cut/copy/paste/rename/remove options
       *  @private
       */
//   _initCrrmOpts : function()
//   {
//       if (this.options["crrm"] == undefined)  {
//         //  Will use our defaults until these are made public and can be
//         //  defined in options.
//         this._applyDefaults(this.options["crrm"], this._data.crrm.defaults) ;
//       }
//   },

     /**
       *  Process context menu options
       *  @private
       */
     _initMenuOpts : function()
     {
     },

//   /**
//     *  Process theme options
//     *  @private
//     */
//   _initThemeOpts : function()
//   {
//     //  Themes support not yet published
//   },


     /**
       *  Process node type options
       *  @private
       */
     _initTypeOpts  : function()
     {
       var opts ;
       var o = this.options["types"] ;

//     For V1, there are no default options that we publish
//     if (typeof o === "object") {
//       //  Add our default requirements until these are made public and can be
//       //  defined in options.
//       this._applyDefaults(this["options"]["types"], this._data.types.defaults) ;
//     }
       if (typeof o === "object") {
         this._applyDefaults(o, {"attr" : this._data.types.defaults["attr"]}) ;
       }
     },

     /**
       *  Set/reset the initExpanded opts
       *  @private
       */
     _initExpandedOpts : function()
     {
        this._data.core.toExpand = this._varCopy(this.options, "initExpanded");
        if (this._data.core.toExpand == null) {
          this._data.core.toExpand = [] ;
        }
     },

     /**
       *  Set up tree instance data
       *  @private
       */
     _initData : function()
     {
        //  Establish working values, and define default values

        var data = this._data ;

        //  Core
        data.core  =  {
                        initLoaded     : [],
                        selectMode     : 1,          //  0 = no selection, 1, 2 ..., or -1 for unlimited
                        load_open      : false,
                        li_height      : 0,
                        toExpand       : false,
                        toLoad         : false,
                        prepMoveBlk    : {},         // for the _moveNode function
                        suppressSelectEvent : false, // used for batching up optionChange events
                        strings        : {}          // translated/escaped strings
                      };

        // UI
        data.ui =  {
                      selected         : $(),        // selected node jquery list
                      lastSelected     : false,
                      hovered          : null,       // currently hovered
                      lastHovered      : null,       // last hovered before blur
                      disclosureClick  : false,      // used for 1st time disclosure click
                      toSelect         : null,       // initial selection in options:selection
                      opacity          : 1,          // used by disable/_lock()
                      spacebar         : false,      // true if select caused by keybd (toggles)
                      focused          : false,      // tree has focus
                      animDurDiv       : null,       // div used to get animation duration
                      touchEvent       : false       // true if touch event
                   };

        data.ui.defaults = {                          // default values not yet published or suppressed
                                                      // design committee
                             "selectMultipleModifier"   :  "ctrl" ,      // on, or ctrl, shift, alt
                             "selectRangeModifier"      :  "shift" ,
                             "disableSelectingChildren" :  false
                           } ;


        //  Creating/renaming/removing/moving via context menu

        data.crrm = {} ;
        data.crrm.cp_nodes = false;     // nodes that have been copy'd
        data.crrm.ct_nodes = false;     // node(s) that have been cut
        data.crrm.defaults = {
                               "inputWidthLimit" : 200,
                               "move"            : {
                                                     "alwaysCopy"       : false, // false, true or "multitree"
                                                     "openOnMove"       : true,
                                                     "defaultPosition"  : "last",
                                                     "checkMove"        : function (m) {
                                                                             return true;
                                                                          }
                                                   }
                             };
        data.crrm.prepMoveBlk = {} ;   // for the move node function

        // Data Source

        data.ds = {} ;
        data.ds.progressiveRender = false ;       // options not currently exposed
        data.ds.progressiveUnload = false ;
        data.ds.correctState      = true ;

        data.ds.type              = DS_NONE ;     // type of data source (DS_TREE, DS_JSON, DS_HTML)

        //  json data source

        data.json = {} ;
        data.json.defaults = {
                               "data"  : false,   // `data` can be a function:
                                                  //  accepts two arguments - node being loaded
                                                  //  and a callback to pass the result to,
                                                  //  & ajax won't be supported
                               "ajax"  : false
                             } ;

        // html data source

       data.html = {} ;
       data.html.defaults = {
                              "data"   : false,   // `data` can be a function:
                              "ajax"   : false
                            };

       data.html.useExistingMarkup  = false ;  // true == use existing div markup
       data.html.markup_ul          = false ;  // user's <ul>
       data.html.markup_div         = false ;  // the added div used to store user's <ul>
       data.html.markup             = false ;  // the user's markup ( <li>'s )
       data.html.cloneMarkup        = false ;  // our clone of user's orig markup <li's>


        //  Themes

        data.themes         = {} ;
        data.themes.icons   = true ;
        data.themes.dots    = false ;
        data.themes.theme   = "default" ;
        data.themes.url     = false ;

        data.themes.themes_loaded = [];
        data.themes._themes = false;   // path to themes folder - if false it will be autodetected

        //  Types

        data.types           = {} ;
        data.types.attachTo  = [];
        data.types.defType   = false ;                    // true if "default" type defined
        data.types.defaults  = {                                 //  Options not published in V1
                                 "maxChildren"         : -1,     // defines max number of root nodes
                                                                 // (-1 = unlimited, -2 = disable maxChildren checking)
                                 "maxDepth"            : -1,     //  maximum depth of the tree
                                 "validChildren"       : "all",  // defines valid node types for the root nodes
                                 "useData"             : false,  // whether to use $.data     TDO
                                 "attr"                : "type", // attr name in <li> where type is stored
                                 "types" : {                     // a list of types
                                             "default" : {       // the default type
                                                           "maxChildren"   : -1,
                                                           "maxDepth"      : -1,
                                                           "validChildren" : "all"

                                                           // Bound functions - you can bind any other function here
                                                           // (using boolean or function)
                                                           //"select_node" : true
                                                         }
                                           }
                               } ;

        //  Context menu

        data.menu              = {} ;
        data.menu.usermenu     = false ;   // user has supplied an ojMenu id if true
        data.menu.$container   = false ;   // the menu <ul>
        data.menu.$elemPaste   = false ;   // the menu "Paste" element
        data.menu.$elemPasteAfter  = false ;   // the menu "Paste" After element
        data.menu.$elemPasteBefore = false ;   // the menu "Paste Before" element
        data.menu.node         = false ;   // the tree node the menu was activated on
        data.menu.activeNode   = false ;   // active node for shift-F10


        //  Keyboard support

        data.keys         = {} ;
        data.keys.enabled = true ;
        data.keys.bound   = [];
     },

     /**
       *  @private
       */
     _fix_scroll : function (obj)
     {
        var c = this._$container[0],
            t;

       if (c.scrollHeight > c.offsetHeight)
       {
          obj = this._getNode(obj);
          if (!obj || obj === -1 || !obj.length || !obj.is(":visible"))  {
             return;
          }

          t = obj.offset().top - this._$container.offset().top;

          if (t < 0)  {
            c.scrollTop = c.scrollTop + t - 1;
          }

         if (t + this._data.core.li_height + (c.scrollWidth > c.offsetWidth ? scrollbar_width : 0) > c.offsetHeight)  {
           c.scrollTop = c.scrollTop + (t - c.offsetHeight + this._data.core.li_height + 1 + (c.scrollWidth > c.offsetWidth ? scrollbar_width : 0));
          }
       }
     },

     /**
       *  @private
       */
     _setFocus   : function ()            // undocumented per Design Review
     {
/*                                        // 
       if (this._isFocused()) {
          return;
       }
       var f = this._focused();
       if (f) {
         this._unsetFocus();
       }

       this._$container.addClass("oj-tree-focused");
       if (this._$container_ul) {
         this._$container_ul.focus() ;
       }
       this._data.ui.focused = true ;
       this._emitEvent(null, "setfocus", true);
*/
     },

//                                         // currently unused
//   /**
//      *  Return true if this tree has focus.
//      *  @private
//      */
//    _isFocused  : function ()            // undocumented per Design Review
//    {
//       return this._data.ui.focused ;
//    },

     /**
       *  @private
       */
     _unsetFocus : function ()            // undocumented per Design Review
     {
/*    See setFocus()
        if (this._isFocused()) {
          this._$container.removeClass("oj-tree-focused");
          this._data.ui.focused = false ;
          if (this._$container_ul) {
            this._$container_ul.blur() ;
          }
        }
        this._emitEvent(null, "unsetfocus", true);
*/
     },

//                                          // currently unused
//   /**
//     *  Return the "this of the tree that currently has focus, or null.
//     *  @private
//     */
//   _focused : function ()
//   {
//      var ret = null ;
//      $.each(_aInstances, function(i, _this) {
//          if (_this._data.ui.focused) {
//            ret = _this ;
//            return false ;
//          }
//      });
//
//      return ret ;
//   },

     /**
       *  Returns the next instance index to use.
       *  @private
       */
     _newIndex: function()
     {
       return ++_instance
     },

     /**
       *  Returns the current instance index.
       *  @private
       */
     _getIndex : function()
     {
        return this._index ;
     },

     /**
       *  Returns a copy of the options.
       *  @private
       */
     _getOptions : function()
     {
        return ($.extend(true, {}, this.options)) ;
     },

     /**
       *  Returns the tree containing div
       *  @private
       */
     _getContainer : function()
     {
        return this._$container ;
     },


     /**
       *  Returns the tree outer list UL
       *  @private
       */
     _getContainerList : function()
     {
        return this._$container_ul ;
     },

     /**
       *   Handle keystroke and combination keystrokes.
       *   The return value is passed to the jQuery keydown handler.
       *   @private
       */
     _keyHandler : {

         "up" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected || -1;
            this["hover"](this._getPrev(o));
            return false;
         },

         "ctrl+up" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected || -1;
            this["hover"](this._getPrev(o));
            return false;
         },

         "shift+up" : function (t)
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected || -1;
            this["select"](this._getPrev(o), (this._data.ui.selectMode !== -1));
            return false;
         },

         "down" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected || -1;
            this["hover"](this._getNext(o));
            return false;
         },

         "ctrl+down" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected || -1;
            this["hover"](this._getNext(o));
            return false;
         },

         "shift+down" : function ()
         {
           var o = this._data.ui.hovered || this._data.ui.lastSelected || -1;
           this["select"](this._getNext(o), (this._data.ui.selectMode !== -1));
           return false;
         },

         "left" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected;
            if (o) {
              if (o.hasClass(TreeUtils._OJ_EXPANDED))  {
                this["collapse"](o);
              }
              else  {
                this["hover"](this._getPrev(o));
              }
            }
            return false;
         },

         "ctrl+left" : function ()
         {
           var o = this._data.ui.hovered || this._data.ui.lastSelected;
           if (o)  {
             if (o.hasClass(TreeUtils._OJ_EXPANDED))  {
                this["collapse"](o);
             }
             else  {
               this["hover"](this._getPrev(o));
             }
           }
           return false;
         },

         "shift+left" : function ()
         {
           var o = this._data.ui.hovered || this._data.ui.lastSelected;
            if (o) {
              if (o.hasClass(TreeUtils._OJ_EXPANDED)) {
                this["collapse"](o);
              }
              else  {
                this["hover"](this._getPrev(o));
              }
            }
            return false;
         },

         "right" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected;
            if (o && o.length) {
               if (o.hasClass(TreeUtils._OJ_COLLAPSED))  {
                  this["expand"](o);
               }
               else  {
                  this["hover"](this._getNext(o));
               }
            }
            return false;
         },

         "ctrl+right" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected;
            if (o && o.length) {
               if (o.hasClass(TreeUtils._OJ_COLLAPSED)) {
                 this["expand"](o);
               }
               else   {
                 this["hover"](this._getNext(o));
               }
            }
            return false;
         },

         "shift+right" : function ()
         {
            var o = this._data.ui.hovered || this._data.ui.lastSelected;
            if (o && o.length) {
              if (o.hasClass(TreeUtils._OJ_COLLAPSED))  {
                this["expand"](o);
              }
              else  {
                this["hover"](this._getNext(o));
              }
            }
            return false;
         },

         "space" : function ()              // toggle hovered node select status
         {
            if (this._data.ui.hovered)  {
              this._data.ui.spacebar = true ;      // note keyboard - so click will toggle
              this._data.ui.hovered.children("a:eq(0)").click();
              this._data.ui.spacebar = false ;
            }
            return false;
         },

         "home" : function ()               // move hover to top node
         {
           this["hover"](this._$container_ul.find("li:first"));
           return false;
         },

         "end" : function ()                // move hover to last visible node
         {
           var a = this._$container_ul.find("li.oj-tree-last:visible");
           if (a && a.length) {
             this["hover"](a[a.length-1]) ;
           }
           return false;
         },

         "*" : function ()                  // expand all nodes
         {
           var l = this._$container_ul.find("a") ;
           this._expandAll(-1, false) ;
           return false;
         },

         "ctrl+space" : function (event)    // toggle the select state
         {
           if (this._data.ui.hovered)  {
             var ev = $.Event("click") ;
             ev.ctrlKey = true;
             this._data.ui.hovered.children("a:eq(0)").trigger(ev);
           }
           return false;
         },

         "shift+space" : function (event)   // select to the hovered node
         {
           if (this._data.ui.hovered)  {
             var ev = $.Event("click") ;
             ev.shiftKey = true;
             this._data.ui.hovered.children("a:eq(0)").trigger(ev);
           }
           return false;
         },

         "shift+home" : function (event)   // extend selection to top node
         {
           var prevSelections ;
           var hover  = this._data.ui.hovered;

           if (hover)  {
             var bContinue  = true,
                 _this      = this,
                 nodes      = this._$container_ul.find("li:visible");

             // Batch up the optionChange/selection events so that the final result is
             // fired in the optionChange event (rather than individual events).
             this._data.core.suppressSelectEvent = true  ;
             prevSelections =  this.options["selection"].slice(0) ;  // make new array for optionChange event

             hover = hover[0] ;
             $.each(nodes, function(i, node) {         // select from top
                  if (node == hover) {                 // until we find our
                    bContinue = false ;                // starting node
                  }
                  _this._select(node, true) ;
                  return bContinue ;
             });

             this._data.core.suppressSelectEvent = false ;
             this._fireOptionChangeEvent("selection", prevSelections, null, event) ;
           }

           return false;
         },

         "shift+pgdn" : function (event)      // extend selection to last node
         {
           var prevSelections ;
           var hover  =  this._data.ui.lastSelected || this._data.ui.hovered ;

           if (hover)  {
             var bFound = false,
                 _this  = this,
                 l      = this._$container_ul.find("li:visible");


             // Will batch up the optionChange/selection events so that the final result is fired in
             // the optionChange event (rather than individual events).
             this._data.core.suppressSelectEvent = true  ;
             prevSelections =  this.options["selection"].slice(0) ;  // make new array for optionChange event

             hover = hover[0] ;
             $.each(l, function(i, node) {
                if (! bFound) {                        // select to bottom from
                  bFound = (node == hover) ;           // last selected node
                }
                if (bFound && (! _this._isSelected(node))) {
                  _this._select(node, true) ;
                }
                return true ;
             });

             this._data.core.suppressSelectEvent = false ;
             this._fireOptionChangeEvent("selection", prevSelections, null, event) ;
           }

           return false;
         },

         "f2" : function ()
         {
//         this["rename"](this._data.ui.hovered || this._data.ui.lastSelected);
           this._crrm_rename(this._data.ui.hovered || this._data.ui.lastSelected) ;
           return false ;
         },

         "del" : function ()
         {
           this["remove"](this._data.ui.hovered || this._getNode(null));
           return false ;
         }
     },

     /**
       *   Attach the user menu <ul> list to the tree div, and set up listeners on it.
       *   @private
       */
     _applyMenu : function()
     {
       if (! this._data.menu.usermenu) {
         return ;
       }

       // Add our listeners so that we can handle build-in cut/copy/paste, etc
       var $menuContainer = this._data.menu.$container ;
       var _this          = this ;

       $menuContainer.on("ojselect", this._handleContextMenuSelect.bind(this));

       // If there are any ojTree built in menu item ids, construct the menu items
       var listItems = $menuContainer.find("[data-oj-command]");
       var bChanged  = false ;

       listItems.each(function()  {
              var cmd;
              if ($(this).children('a').length === 0)  {
                cmd = $(this).attr('data-oj-command').split("oj-tree-");                // get cut/paste, etc
                $(this).replaceWith(_this._buildContextMenuItem(cmd[cmd.length-1]));    //@HTMLUpdateOK
                if ($(this).hasClass("oj-menu-divider")) {
                  $(this).removeClass("oj-menu-divider")
                         .addClass("oj-menu-item") ;
                }
                bChanged = true ;
              }
          });

       if (bChanged) {
         $menuContainer.ojMenu('refresh');
       }

       //  Note "paste" element for disabling if no prev cut/copy
       this._data.menu.$elemPaste       = $menuContainer.find("#ojtreepaste") ;
       this._data.menu.$elemPasteAfter  = $menuContainer.find("#ojtreepasteafter") ;
       this._data.menu.$elemPasteBefore = $menuContainer.find("#ojtreepastebefore") ;
     },

     /**
       *  Clear out any contextMenu data.
       *  @private
       */
     _clearMenu : function() {

       var um = this._data.menu ;

       if (um.usermenu)  {
         um.usermenu   = false ;
         um.$container.off("ojselect") ;
         um.$container = null ;
       }
     },

     /**
       * Builds a menu for a command, takes care of submenus where appropriate
       * @param {string} cmd
       * @private
       */
     _buildContextMenuItem : function(cmd)
     {
        return $(this._buildContextMenuListItem(cmd)) ;
     },

     /**
       * Builds a context menu list item from a command
       * @param {string} cmd the string to look up command value for as well as translation
       * @return {string} an HTML string containing a list item and a
       * @private
       */
     _buildContextMenuListItem: function(cmd)
     {
        var id = _arMenuCmdMap[cmd] ;
        return '<li id=' + id + '>' + this._buildContextMenuLabel(cmd) + '</li>';
     },

     /**
       * Builds a context menu anchor element with translated text
       * @param {string} cmd the command string whose label translated label should be used.
       * @return {string} an HTML anchor element string containing the translated label
       * @private
       */
     _buildContextMenuLabel: function(cmd)
     {
         var key = _arMenuKeyMap[cmd] ;
         return '<a href="#">' + this._getString(key) + '</a>';
     },

     /**
       *  Returns true if node id is in the load failure map
       *  @param {string | number}  id   the node id to query
       *  @return {boolean}     true if in list, else false
       *  @private
       */
     _isLoadFailureId : function(id)
     {
       id = (id === -1)? "-1" : id ;
       id = (id.charAt(0) === '#')? id.substr(1) : id ;

       return !! this._loadFailureIds[id] ;
     },

     /**
       *  Add/remove node id from load failure map
       *  @param {string | number}  id     the node id to add/remove
       *  @param {boolean} add    true if id is to be added, false if it is to be removed.
       *  @private
       */
     _noteLoadFailureId: function(id, add)
     {
       id = (id === -1)? "-1" : id ;

       if (add)
       {
         this._loadFailureIds[id] = true ;
       }
       else  if (this._loadFailureIds.hasOwnProperty(id))
       {
         delete this._loadFailureIds[id];
       }
     },

     /**
       *  Menu "cut" functionality
       *  @private
       */
     _crrm_cut : function (obj)
     {
        // If node is selected, then will cut all selected nodes, else just the one node.
        obj = obj.closest("li") ;
        if (this._isSelected(obj)) {
          obj = null ;        // _getNode() will now return all selected nodes
        }

        obj = this._getNode(obj, true);
        if (!obj || !obj.length)  {
          return false;
        }

        this._data.crrm.cp_nodes = false;
        this._data.crrm.ct_nodes = obj;
        this._emitEvent({ "obj" : obj }, "cut");
     },

     /**
       *  Menu "copy" functionality
       *  @private
       */
     _crrm_copy : function (obj)
     {
       // If node is selected, then will copy all selected nodes, else just the one node.
       obj = obj.closest("li") ;
       if (this._isSelected(obj)) {
         obj = null ;        // _getNode() will now return all selected nodes
       }
       obj = this._getNode(obj, true);
       if (!obj || !obj.length)  {
         return false;
       }
       this._data.crrm.ct_nodes = false;
       this._data.crrm.cp_nodes = obj;
       this._emitEvent({ "obj" : obj }, "copy");
     },

     /**
       *  Menu "paste" functionality
       *  @private
       */
     _crrm_paste : function (obj, menuId)
     {
        obj = this._getNode(obj);
        if (!obj || !obj.length) {
          return false;
        }
        var nodes = this._data.crrm.ct_nodes ? this._data.crrm.ct_nodes : this._data.crrm.cp_nodes;

        if (! this._data.crrm.ct_nodes && !this._data.crrm.cp_nodes)  {
           return false;
        }

        var position = "inside" ;
        if (menuId === "ojtreepasteafter") position = "after"
        else if (menuId === "ojtreepastebefore") position = "before" ; 

        // Handle cut nodes
        if (this._data.crrm.ct_nodes)  {
          this._crrm_move_node(this._data.crrm.ct_nodes, obj, position, false, false, true);
          this._data.crrm.ct_nodes = false;
        }
        // Handle copied nodes
        if (this._data.crrm.cp_nodes)  {
          this._crrm_move_node(this._data.crrm.cp_nodes, obj, position, true, false, true);
        }
        this._emitEvent({ "obj" : nodes,
                          p : position,
                          r : obj
                        }, "paste");
     },

     /**
       *  Menu move node support for paste
       *  @private
       */
     _crrm_move_node : function (obj, ref, position, is_copy, is_prepared, skip_check)
     {
        //var s = this.options["crrm"]["move"];
        var s = this._data.crrm.defaults["move"];
        if (! is_prepared)  {
          if (typeof position === "undefined")  {
            position = s["defaultPosition"];
          }
          if (position === "inside" && !s["defaultPosition"].match(/^(before|after)$/))  {
            position = s["defaultPosition"];
          }
          return this._moveNode(obj, ref, position, is_copy, false, skip_check);
        }

        // if the move is already prepared
        if (s["alwaysCopy"] === true || (s["alwaysCopy"] === "multitree" && obj.rt._getIndex() !== obj.ot._getIndex() ))  {
          is_copy = true;
        }
        this._moveNode(obj, ref, position, is_copy, true, skip_check);
     },

     /**
       *  Menu "remove" (i.e. delete) functionality
       *  @private
       */
     _crrm_remove : function (obj)
     {
        obj = this._getNode(obj, true);

//      var p    = this._getParent(obj),
//          prev = this._getPrev(obj);

         this.__rollback();
         obj = this["remove"](obj);
//         if (obj !== false) {
//           this._emitEvent({ "obj" : obj, "prev" : prev, "parent" : p }, "remove");
//         }
     },

     /**
       *  Menu rename a node functionality
       *  @private
       */
     _crrm_rename  :  function (obj)
     {
        var f = this._emitEvent;

        obj = this._getNode(obj);
        this.__rollback();
        this._crrm_showInput(obj, function (obj, new_name, old_name)
                                  {
                                    //  Emit rename event
//                                    f.call(this, {
//                                                   "obj"       : obj,
//                                                   "title"     : new_name,
//                                                   "prevTitle" : old_name
//                                                 },  "rename");
                                  });
     },

     /**
       *  Turn node into an input field to allow the user to rename the node.
       *  @private
       */
     _crrm_showInput : function (obj, callback)
     {
        obj = this._getNode(obj);

        var rtl = this._isRtl,
//          w   = this.options["crrm"]["inputWidthLimit"],       // applyDefaults() not done for V1
            w   = this._data.crrm.defaults["inputWidthLimit"],
            w1  = obj.children("ins").width(),
            w2  = obj.find("> a:visible > ins").width() * obj.find("> a:visible > ins").length,
            t   = this["getText"](obj),
            _this = this,

            h1  = $("<div />", { css : { "position"   : "absolute",
                                         "top"        : "-200px",
                                         "left"       : (rtl ? "0px" : "-1000px"),
                                         "visibility" : "hidden"
                                       }
                               }).appendTo("body"),               //@HTMLUpdateOK

            h2  = obj.css("position", "relative").append(         //@HTMLUpdateOK
                $("<input />", {
                    "value" : t,
                    "class" : "oj-tree-rename-input",
                    // "size" : t.length,
                    "css" : {
                              "padding"  : "0",
                              "border"   : "1px solid silver",
                              "position" : "absolute",
                              "left"     : (rtl ? "auto" : (w1 + w2 + 4) + "px"),
                              "right"    : (rtl ? (w1 + w2 + 4) + "px" : "auto"),
                              "top"      : "0px",
                              "height"   : (this._data.core.li_height - 2) + "px",
                              "lineHeight" : (this._data.core.li_height - 2) + "px",
                              "width"    : "150px" // will be set later below
                             },
                    "blur" : $.proxy(function ()
                    {
                        var i = obj.children(".oj-tree-rename-input"),
                            v = i.val();

                        v = this._escapeHtml(v) ;
                        if (v === "")
                        {
                          v = t;
                        }
                        h1.remove();
                        i.remove();                    // rollback purposes
                        this._set_text(obj, t);        // rollback purposes
                        if (v != t)     // if different node text, perform rename with prior before event
                        {
                          this._renameNode(obj, v);
                        }
                        callback.call(this, obj, v, t);
                        obj.css("position","");
                    }, this),

                    "keyup" : function (event)
                    {
                       var key = event.keyCode || event.which;
                       if(key == 27)        // Esc
                       {
                          this.value = t;
                          this.blur();
                          return;
                       }
                       else if(key == 13)   // Enter
                       {
                          this.blur();
                          return;
                       }
                       else  {
                          h2.width(Math.min(h1.text("pW" + this.value).width(), w));
                       }
                    },

                    "keypress" : function(event)
                    {
                      var key = event.keyCode || event.which;
                      if (key == 13)  {
                        return false;
                      }
                    }
                })
            ).children(".oj-tree-rename-input");
            this._set_text(obj, "");
            h1.css({
                    fontFamily      : h2.css('fontFamily')      || '',
                    fontSize        : h2.css('fontSize')        || '',
                    fontWeight      : h2.css('fontWeight')      || '',
                    fontStyle       : h2.css('fontStyle')       || '',
                    fontStretch     : h2.css('fontStretch')     || '',
                    fontVariant     : h2.css('fontVariant')     || '',
                    letterSpacing   : h2.css('letterSpacing')   || '',
                    wordSpacing     : h2.css('wordSpacing')     || ''
            });
            h2.width(Math.min(h1.text("pW" + h2[0].value).width(),w))[0]["select"]();
     },

     /**
       *  @private
       */
     _crrm_create : function (obj, position, js, callback, skip_rename)
     {
        var t, _this = this;
        obj = this._getNode(obj);
        if (!obj) {
           obj = -1;
        }
        this.__rollback();
        t = this._createNode(obj, position, js, function (t)
             {
                var p = this._getParent(t),
                    pos = $(t).index();

                if (callback)  {
                   callback.call(this, t);
                }
                if (p.length && p.hasClass(TreeUtils._OJ_COLLAPSED))  {
                   this["expand"](p, false, true);
                }
                if (!skip_rename)  {
                  this._crrm_showInput(t, function (obj, new_name, old_name)
                         {
                            _this._emitEvent({ "obj"      : obj,
                                               "name"     : new_name,
                                               "parent"   : p,
                                               "position" : pos
                                              });
                         });
                }
                else  {
                   _this._emitEvent({ "obj"      : t,
                                      "name"     : this["getText"](t),
                                      "parent"   : p,
                                      "position" : pos
                                    });
                }
             });

        return t;
     },

//   // Currently unused
//   /**
//     *  @private
//     */
//   _crrm_check_move : function ()
//   {
//      if (! this.__call_old())  {
//        return false;
//      }
//      //var s = this.options["crrm"]["move"];
//      var s = this._data.defaults.crrm["move"];
//      if (! s["checkMove"].call(this, this._getMoveBlk()))  {
//        return false;
//      }
//      return true;
//   },

     /**
       *  Very brief check to see if string looks like it might be html.
       *  @return {boolean} true if string starts with a "<"
       *  @private
       */
     _isHtml : function(s)
     {
        var ret = false ;

        if (s && s.length >= 3) {
          s = s.trim() ;
          ret = (s.charAt(0) === "<") ;
        }
        return ret ;
     },

     /**
       *  Use emptyText option if defined.
       *  @private
       */
     _applyEmptyText : function()
     {
        var txt = this.options.emptyText ;                   // has app set new text

        if (typeof txt !== "string") {
          txt = this._getString(TRANSKEY_NODATA);            // no, get default value.
        }
        if (txt && txt.length > 0)  {

          var $u =  this._$container_ul;
          var $d  = $("<li class='oj-tree-empty'></li>") ;   // no <a>, not interactive
          $d[0].textContent = txt ;
          $u.empty().append($d);                             //@HTMLUpdateOK
        }
     },

     /**
       *  Return the HTMLElement based on the locator subid property.
       *  @private
       */
     _processSubId : function(locator)
     {
        // Parent node
        // <li role="treeitem" id="blogs" class="oj-tree-node oj-tree-expanded|oj-tree-collapsed">
        //    <ins class="oj-tree-icon oj-tree-disclosure-icon oj-..."> </ins>  <-- disclosure icon
        //    <a tabindex="-1" href="#">
        //       <ins class="oj-tree-icon"> </ins>
        //       <span class="oj-tree-title">Blogs</span>
        //       <ul role="group">
        //          <li . . . child nodes . . .
        //       </ul>
        //    </a>
        // </li>
        //
        //  Leaf node
        // <li role="treeitem" id="home" myattr1="Hello" class="oj-tree-node oj-tree-leaf">
        //    <ins class="oj-tree-icon"> </ins>                   <-- unused disclosure icon
        //    <a tabindex="-1" href="#">
        //       <ins class="oj-tree-icon oj-tree-node-icon oj-component-icon"> </ins>
        //       <span class="oj-tree-title">Home</span>
        //    </a>
        // </li>

        var subId, node, a, sNode, c, sKey = null, i, ret ;

        if (locator) {
          subId = locator["subId"] ;
        }
        if (! subId) {
          return null ;
        }

        a  = subId.split("[") ;
        if (a.length === 3) {
          a[0] = a[0].trim() ;
          if (a[0] === "oj-tree-node") {
            a[1] = a[1].trim() ;
            c = a[1].charAt(0) ;                  // should be ' or "
            i = a[1].indexOf(c, 1) ;
            if (i >= 0) {
              sNode = a[1].substring(1, i).trim() ;

              a[2] = a[2].trim() ;
              c = a[2].charAt(0) ;                // should be ' or "
              i = a[2].indexOf(c, 1) ;
              if (i >= 0) {
                sKey = a[2].substring(1, i).trim() ;
                try {
                      node = this._getNode(sNode) ;
                    }
                catch (e)
                    {
                      node = null ;
                    }
                if (sKey && node && node !== -1) {
                  switch(sKey)
                  {
                    case "icon" :       if (this._data.themes.icons) {
                                          ret = node.find(" > a > ins:eq(0)") ;
                                        }
                                        break;
                    case "link" :       ret = node.find(" > a:eq(0)") ;
                                        break;
                    case "disclosure" : if (! this._isLeaf(node)) {
                                          ret = node.find(" > ins:eq(0)") ;
                                        }
                                        break;
                    case "title" :      ret = node.find(" > a > span") ;
                                        break;
                  }
                }
              }
            }
          }
        }

        return ret? (ret.length? ret[0] : null) : null ;
     },

     /**
       *  Return the subid of a node's DOM child element (or null).
       *  Refer to _processSubId() for returned subId values.
       *  @private
       */
     _getSubIdFromNodeElem : function(elem)
     {
        var node = (elem? this._getNode(elem) : null), // get tree node from the sub-element
            type, parent, parentType,
            request = null, subId = null;

        if (node && (node != -1) && node.length && node.hasClass(OJT_NODE)) {
          if (elem.tagName && elem.parentNode) {
            type       = elem.tagName ;
            parent     = elem.parentNode ;
            parentType = parent.tagName ;

            if (type === "SPAN") {
              request = "title" ;
            }
            else if (type === "A") {
              request = "link" ;
            }
            else if (type === "INS") {
              //  Determine whether subId is a node icon, or a parent disclosure icon
              elem = $(elem) ;
              if ((parentType === "LI") && elem.hasClass(OJT_DISC)) {
                request = "disclosure" ;
              }
              else if ((parentType === "A") && elem.hasClass(OJT_NICON)) {
                request = "icon" ;
              }
            }
          }
        }

        if (request) {
          subId = "oj-tree-node['#" + node.attr("id") + "']['" + request + "']" ;
          return {'subId': subId} ;
        }

        return null ;
     },

     /**
       *  Hide the node connecting lines
       *  @private
       */
     _hideDots   : function ()
     {
       this._data.themes.dots = false;
       this._$container.children("ul").addClass("oj-tree-no-dots");
     },

     /**
       *  Show the node connecting lines
       *  @private
       */
     _showDots   : function ()
     {
       this._data.themes.dots = true;
       this._$container.children("ul").removeClass("oj-tree-no-dots");
     },

     /**
       *  Check the options for initExpanded : ["all"] ;
       *  @private
       */
     _isOptExpandAll : function()
     {
        var s = this.options["initExpanded"] ;
        return  ((s && s == "all") || ($.isArray(s) && (s.length && s[0] == "all"))) ;
     },

     /**
       *   Get translated string and escape it.
       *   @private
       */
     _getString: function(key)
     {
       var oStrings = this._data.core.strings,
           s;

       s = oStrings[key] ;                          // have we already escaped this string?
       if (! s) {                                   // no
         s = this.getTranslatedString(key) ;
         s = this._escapeHtml(s) ;
         oStrings[key] = s ;
       }
       return s;
     },

     /**
       *  Return sanitized html fragment/text for security
       *  @private
       */
     _escapeHtml: function(text)
     {
       var div = document.createElement('div') ;
       $(div).text(text) ;                         // escape the text
       return  div.textContent;
     },

     /**
       *  Returns the height of node li
       *  @private
       */
     _getNodeHeight: function()
     {
        return this._data.core.li_height ;
     },


     /**
       *  Get the duration in ms from the css $treeAnimationDuration value.
       *  (This is currently neccessary because if the duration is zero, the
       *  transition end callback is not called, and the expand event is
       *  never fired.
       *  @private
       */
     _getAnimDuration : function()
     {
        var ret = 0, val, c, s, $div, span ;

        if (window.getComputedStyle) {
          $div = this._data.ui.animDurDiv ;
          if (! $div) {
            $div  = $("<div id='ojtree-comp-animduration'><span class='oj-tree-transition'>dummy</span></div>") ;
            this._data.ui.animDurDiv = $div ;
            $("body").append($div) ;   // must be <body>, doesn't work if _$container used and not attached to the DOM //@HTMLUpdateOK
          }
          span = $div[0].childNodes[0] ;
          val = window.getComputedStyle(span);
          if (val["transitionDuration"]) {
            val = val["transitionDuration"] ;
          }
          else if (val["-webkit-transition-duration"]) {   // need this for windows safari
            val = val["-webkit-transition-duration"] ;
          }
          else {
            val = 0 ;
          }

          $div.detach() ;

          if (typeof val == "string" && val.length > 1) {
            c = val.charAt(val.length -1).toLowerCase()
            if (c == "s") {
              val = val.substring(0, val.length-1) ;
              val = parseFloat(val) ;
              if (! isNaN(val)) {
                ret = val * 1000 ;
              }
            }
            else {
               if (val.length > 2) {
                 s = val.substring(val.length - 2).toLowerCase() ;
                 if (s == "ms") {
                   s = val.substring(0, val.length - 2) ;
                   val = parseFloat(val) ;
                   if (! isNaN(val)) {
                     ret = val ;
                   }
                 }
              }
            }
          }
        }
        return ret ;        // return milliseconds
     },


     /**
       *  Return a copy of a member of an object.
       *  @private
       */
     _varCopy :  function(obj, s)
     {
        var o  = {} ;
        o[s]   = obj[s] ;
        var o2 = $.extend(true, {}, o) ;

        return o2[s] ;
     }


   /**
     * For easier debugging on tablet test page via vpn.  Emits a _treetrace event
     * containing the supplied string.
     * @private
     */
,  _trace : function(s, level)
   {
//     this._emitEvent({ "obj" : {"msg" : s }}, "trace", true);
//   console.log(s) ;
//   if (level === 1) {
//     console.log(s) ;
//   }
   }


    // API doc for inherited methods with no JS in this file:

    /**
     * Removes the Tree from the DOM.  If the tree was constructed from original
     * user &lt;ul&gt; markup defined in the Tree's containing &lt;div&gt;, this
     * markup is reinstated.
     * <p>This method does not accept any arguments.
     *
     * @method
     * @name oj.ojTree#destroy
     * @memberof oj.ojTree
     * @instance
     *
     * @example <caption>Invoke the <code class="prettyprint">destroy</code> method:</caption>
     * $( ".selector" ).ojTree( "destroy" );
     */


    /**
     * Returns a <code class="prettyprint">jQuery</code> object containing the root element of the Tree component.
     * 
     * @method
     * @name oj.ojTree#widget
     * @memberof oj.ojTree
     * @instance
     * @return {jQuery} the root element of the component
     * 
     * @example <caption>Invoke the <code class="prettyprint">widget</code> method:</caption>
     * var widget = $( ".selector" ).ojTree( "widget" );
     */


    //////////////////     SUB-IDS     //////////////////

    /**
     * <p>Sub-ID for ojTree component's node icons.</p>
     *
     * To find the icon DOM node for a Tree node, the locator object should have the following:
     * <ul>
     * <li>{<b> subId</b> : "oj-tree-node['node id']<b>['icon']</b>" }</li>
     * </ul>
     *
     * @ojsubid icon
     * @memberof oj.ojTree
     *
     * @example <caption>Get the node icon for the Tree node with Id "#home":</caption>
     * var node = $( ".selector" ).ojTree("getNodeBySubId", {"subId": "oj-tree-node['#home']['icon]" } );
     */

    /**
     * <p>Sub-ID for the ojTree component's disclosure icons.</p>
     *
     * To find the disclosure (expand/collapse) icon DOM node for a Tree node, the locator object should
     * have the following:
     * <ul>
     * <li>{<b> subId</b> : "oj-tree-node['node id']<b>['disclosure']</b>" }</li>
     * </ul>
     *
     * @ojsubid disclosure
     * @memberof oj.ojTree
     *
     * @example <caption>Get the disclosure icon DOM node for the Tree node with Id "#home":</caption>
     * var node = $( ".selector" ).ojTree("getNodeBySubId", {"subId": "oj-tree-node['#home']['disclosure]" } );
     */

    /**
     * <p>Sub-ID for the ojTree component's node title.</p>
     *
     * To find the node title DOM node for a Tree node, the locator object should have the following:
     * <ul>
     * <li>{<b> subId</b> : "oj-tree-node['node id']<b>['title']</b>" }</li>
     * </ul>
     *
     * @ojsubid title
     * @memberof oj.ojTree
     *
     * @example <caption>Get the text title DOM node for the Tree node with Id "#home":</caption>
     * var node = $( ".selector" ).ojTree("getNodeBySubId", {"subId": "oj-tree-node['#home']['title]" } );
     */

    /**
     * <p>Sub-ID for the ojTree's component's node link element.</p>
     *
     * To find the link DOM element for a Tree node, the locator object should have the following:
     * <ul>
     * <li>{<b> subId</b> : "oj-tree-node['node id']<b>['link']</b>" }</li>
     * </ul>
     *
     * @ojsubid link
     * @memberof oj.ojTree
     *
     * @example <caption>Get the link DOM node for the Tree node with Id "#home":</caption>
     * var node = $( ".selector" ).ojTree("getNodeBySubId", {"subId": "oj-tree-node['#home']['link]" } );
     */


    //////////////////     FRAGMENTS    //////////////////

    /**
      * <table class="keyboard-table">
      *   <thead>
      *     <tr>
      *       <th>Target</th>
      *       <th>Gesture</th>
      *       <th>Action</th>
      *     </tr>
      *   </thead>
      *   <tbody>
      *     <tr>
      *       <td>Node Disclosure Icon</td>
      *       <td><kbd>Tap</kbd></td>
      *       <td>Toggle the node's expanded state.</td>
      *     </tr>
      *     <tr>
      *       <td>Node Text</td>
      *       <td><kbd>Tap</kbd></td>
      *       <td>Select the node. Toggle the select state if multi-select mode.</td>
      *     </tr>
      *     <tr>
      *       <td>Node Text</td>
      *       <td><kbd>Press and Hold</kbd></td>
      *       <td>Open the context menu.</td>
      *     </tr>
      *   </tbody>
      * </table>
      *
      * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
      * @memberof oj.ojTree
      */

     /**
       *
       * <table class="keyboard-table">
       *   <thead>
       *     <tr>
       *       <th>Key</th>
       *       <th>Action</th>
       *     </tr>
       *   </thead>
       *   <tbody>
       *     <tr>
       *       <td><kbd>UpArrow/DownArrow</kbd></td>
       *       <td>Moves between visible nodes.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>LeftArrow</kbd></td>
       *       <td>On an expanded node, collapses the node.<br>On a collapsed or leaf node, moves focus to the node's parent.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>RightArrow</kbd></td>
       *       <td>On a collapsed node, expands the node.<br>On an expanded node, moves to the first first child of the node.<br>On an end node, does nothing.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Spacebar</kbd></td>
       *       <td>Toggles the selected status of the node.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Home</kbd></td>
       *       <td>Moves to the top node of the tree.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>End</kbd></td>
       *       <td>Moves to the last visible node of the tree.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Shift + UpArrow</kbd></td>
       *       <td>Extends selection up one node (assuming multiple selection has been defined).</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Shift + DownArrow</kbd></td>
       *       <td>Extends selection down one node (assuming multiple selection has been defined).</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Shift + Home</kbd></td>
       *       <td>Extends selection up to the top-most node.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Shift + PageDown</kbd></td>
       *       <td>Extends selection to the last node.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Ctrl + Spacebar</kbd></td>
       *       <td>Toggles the selection state of the current node (assuming multiple selection has been defined).</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Shift + Spacebar</kbd></td>
       *       <td>Extends selection to the current node (assuming multiple selection has been defined).</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>Shift + F10</kbd></td>
       *       <td>Invoke Context Menu (if defined) on current node.</td>
       *     </tr>
       *     <tr>
       *       <td><kbd>* (asterisk)</kbd></td>
       *       <td>Expands all nodes.</td>
       *     </tr>
       * </tbody></table>
       *
       * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
       * @memberof oj.ojTree
       */

  }) ;    // end    $.registerWidget("oj.ojTree", ...

})();
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @preserve Copyright 2013 jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */

/*jslint browser: true,devel:true*/
/**
 * @ignore
 * @ojstatus preview
 * @since 5.1.0
 * @export
 * @interface oj.TreeDataProvider
 * @extends oj.DataProvider
 * @ojsignature {target: "Type",
 *               value: "interface TreeDataProvider<K, D> extends DataProvider<K, D>"}
 * @classdesc
 * TreeDataProvider is the basic interface for getting runtime data which JET components that display hierarchical list of items (such as [oj-tree-view]{@link oj.ojTreeView}) can use.
 * <p>
 * This interface extends DataProvider, with the addition of a <code class="prettyprint">getChildDataProvider</code> method.
 * </p><p>
 * Unless otherwise noted, all methods operate on the set of sibling items with the same parent.
 * For example, <code class="prettyprint">fetchFirst</code> returns an iterable for the sibling items with the same parent; <code class="prettyprint">getTotalSize</code> returns a promise that 
 * resolves to the number of items with the same parent.
 * </p><p>
 * To retrieve the children of an item, consumers use the <code class="prettyprint">getChildDataProvider</code> method to get another TreeDataProvider for the children of that item.
 * </p><p>
 * JET provides some implementations of this interface, such as {@link oj.ArrayTreeDataProvider}.
 * </p><p>
 * Applications can also create their own implementations of this interface and use them with JET components that support it.  For example, an application can create a TreeDataProvider implementation
 * that fetches data from a REST endpoint.
 * </p><p>
 * Implementation class must implement all of the interface methods.  It should also fire the events described below when appropriate, so that JET components or other consumers can respond to data change accordingly. 
 * </p>
 *
 * <h3 id="events-section">
 *   Events
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#events-section"></a>
 * </h3>
 * Implementation can fire the following events by creating an instance of the event class and passing the event payload in the constructor.
 * <h4 id="event:DataProviderMutationEvent" class="name">
 *   {@link oj.DataProviderMutationEvent}
 * </h4>
 * This event is fired when items have been added or removed from the data.
 * <p>
 * Event payload should implement the {@link oj.DataProviderMutationEventDetail} interface.
 * </p><p>
 * Consumers can add an event listener for the "mutate" event type on the DataProvider object.
 * </p>
 * <h4 id="event:DataProviderRefreshEvent" class="name">
 *   {@link oj.DataProviderRefreshEvent}
 * </h4>
 * This event is fired when the data has been refreshed and components need to re-fetch the data.
 * <p>
 * This event contains no additional event payload.
 * </p><p>
 * Consumers can add an event listener for the "refresh" event type on the DataProvider object.
 * </p>
 *
 * <i>Example of implementation firing an oj.DataProviderMutationEvent for removed items:</i>
 * <pre class="prettyprint"><code>var removeDetail = {data: removedDataArray,
 *                     indexes: removedIndexArray,
 *                     keys: removedKeySet,
 *                     metadata: removedMetadataArray};
 * this.dispatchEvent(new oj.DataProviderMutationEvent({remove: removeDetail}));
 * </code></pre>
 *
 * <i>Example of consumer listening for the "mutate" event type:</i>
 * <pre class="prettyprint"><code>var listener = function(event) {
 *   if (event.detail.remove) {
 *     var removeDetail = event.detail.remove;
 *     // Handle removed items
 *   }
 * };
 * dataProvider.addEventListener("mutate", listener);
 * </code></pre>
 */
oj.TreeDataProvider = function()
{
};

/**
 * Get the data provider for the children of the row identified by parentKey.
 * 
 * @ojstatus preview
 * @since 5.1.0
 * @param {any} parentKey key of the row to get child data provider for.
 * @return {TreeDataProvider | null} A TreeDataProvider if the row can (but doesn't have to) have children; or null if the row cannot have children.
 *   Use the <code class="prettyprint">isEmpty</code> method on the returned TreeDataProvider to determine if it currently has children.
 * @export
 * @expose
 * @memberof oj.TreeDataProvider
 * @instance
 * @method
 * @name getChildDataProvider
 * @ojsignature {target: "Type",
 *               value: "(any): TreeDataProvider<K, D>"}
 */

/**
 * End of jsdoc
 */
/**
 * Copyright (c) 2014, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojTreemap
 * @augments oj.dvtBaseComponent
 * @since 0.7
 * @ojstatus preview
 * @ojshortdesc An interactive data visualization in which hierarchical data is represented across two dimensions by the size and color of nested rectangular nodes.
 * @ojrole application
 * @ojtsignore
 *
 * @classdesc
 * <h3 id="treemapOverview-section">
 *   JET Treemap
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#treemapOverview-section"></a>
 * </h3>
 *
 * <p>Treemaps are used to display hierarchical data across two dimensions, represented by
 * the size and color of the treemap nodes. Treemaps are generally preferred over sunbursts when emphasizing the data
 * for the leaf nodes.</p>
 *
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-treemap
 *    nodes='[{"value": 100, "color": "#FFFF00", "label": "Total Sales",
 *             "nodes": [{"value": 75, "color": "#00FF00", "label": "Candy"},
 *                       {"value": 20, "color": "#FFFF00", "label": "Fruit"},
 *                       {"value": 15, "color": "#FF0000", "label": "Vegetables"}]}]'>
 * &lt;/oj-treemap>
 * </code>
 * </pre>
 *
 * {@ojinclude "name":"a11yKeyboard"}
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Gesture End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Animation</h4>
 * <p>Animation should only be enabled for visualizations of small to medium data sets. Alternate visualizations should
 *    be considered if identifying data changes is important, since all nodes will generally move and resize on any data
 *    change.
 * </p>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it's recommended that applications only set usable data densities on this element.
 *    Applications can enable progressive reveal of data through drilling or aggregate small nodes to reduce the
 *    displayed data set size.
 * </p>
 * 
 * <h4>Styling</h4>
 * <p>Use the highest level property available. For example, consider setting styling properties on
 *    <code class="prettyprint">nodeDefaults</code>, instead of styling properties
 *    on the individual nodes. The treemap can take advantage of these higher level properties to apply the style properties on
 *    containers, saving expensive DOM calls.
 * </p>
 *
 * {@ojinclude "name":"fragment_trackResize"}
 *
 * {@ojinclude "name":"rtl"}
 */
oj.__registerWidget('oj.ojTreemap', $['oj']['dvtBaseComponent'],
  {
  widgetEventPrefix : "oj",
  options: {
    /**
     * Specifies the animation duration in milliseconds. For data change animations with multiple stages, 
     * this attribute defines the duration of each stage. For example, if an animation contains two stages, 
     * the total duration will be two times this attribute's value. The default value comes from the CSS and varies based on theme.
     * @expose
     * @name animationDuration
     * @memberof oj.ojTreemap
     * @instance
     * @type {number}
     * @ojunits milliseconds
     */
    animationDuration: undefined,
    
    /**
     * Specifies the animation that is applied on data changes.
     * @expose
     * @name animationOnDataChange
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDataChange: "none",
    
    /**
     * Specifies the animation that is shown on initial display.
     * @expose
     * @name animationOnDisplay
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "auto"
     * @ojvalue {string} "none"
     * @default "none"
     */
    animationOnDisplay: "none",
    
    /**
     * The color that is displayed during a data change animation when a node is updated.
     * @expose
     * @name animationUpdateColor
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @default ""
     */
    animationUpdateColor: "",
    
    /**
     * An array of category strings used for filtering. Nodes with any category matching an item in this array will be filtered.
     * @expose
     * @name hiddenCategories
     * @memberof oj.ojTreemap
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    hiddenCategories: [],
    
    /**
     * An array of category strings used for highlighting. Nodes matching all categories in this array will be highlighted.
     * @expose
     * @name highlightedCategories
     * @memberof oj.ojTreemap
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    highlightedCategories: [],
    
    /**
     * The matching condition for the highlightedCategories property. By default, highlightMatch is 'all' and only items whose categories match all of the values specified in the highlightedCategories array will be highlighted. If highlightMatch is 'any', then items that match at least one of the highlightedCategories values will be highlighted.
     * @expose
     * @name highlightMatch
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "any"
     * @ojvalue {string} "all"
     * @default "all"
     */
    highlightMatch: "all",
    
    /**
     * Defines the behavior applied when hovering over the nodes.
     * @expose
     * @name hoverBehavior
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "dim"
     * @ojvalue {string} "none"
     * @default "none"
     */
    hoverBehavior: "none",
    
    /**
     * Specifies initial hover delay in ms for highlighting nodes.
     * @expose
     * @name hoverBehaviorDelay
     * @memberof oj.ojTreemap
     * @instance
     * @type {number}
     * @default 200
     * @ojunits milliseconds
     */
    hoverBehaviorDelay: 200,
    
    /**
     * An object defining custom node content for the treemap. Only leaf nodes with no child nodes will have the custom content rendered.
     * @expose
     * @name nodeContent
     * @memberof oj.ojTreemap
     * @instance
     * @type {Object}
     * @default {"renderer" : null}
     */
    nodeContent: {
      /**
       * A function that returns custom node content. The function takes a dataContext argument, 
       * provided by the treemap, with the following properties: 
       * <ul>
       *   <li>bounds: Object containing (x, y, width, height) of the node area. 
       *   The x and y coordinates are relative to the top, left corner of the element.</li>
       *   <li>id: The id of the node.</li> <li>data: The data object of the node.</li>
       *   <li>componentElement: The treemap element.</li>
       * </ul>
       * The function should return an Object with the following property: 
       * <ul>
       *   <li>insert: HTMLElement - An HTML element, which will be overlaid on top of the treemap. 
       *   This HTML element will block interactivity of the treemap by default, but the CSS pointer-events 
       *   property can be set to 'none' on this element if the treemap's interactivity is desired. 
       *   </li>
       * </ul>
       * @expose
       * @name nodeContent.renderer
       * @memberof! oj.ojTreemap
       * @instance
       * @type {function(Object):Object|null}
       * @default null
       */
       renderer: null
     },
     
    /**
     *  An object containing an optional callback function for tooltip customization. 
     * @expose
     * @name tooltip
     * @memberof oj.ojTreemap
     * @instance
     * @type {Object}
     * @default {"renderer" : null}
     */
    tooltip: {
      /**
       * A function that returns a custom tooltip. The function takes a dataContext argument,
       * provided by the treemap, with the following properties:
       * <ul>
       *   <li>parentElement: The tooltip element. The function can directly modify or append content to this element.</li>
       *   <li>id: The id of the hovered node.</li>
       *   <li>label: The label of the hovered node.</li>
       *   <li>value: The value of the hovered node.</li>
       *   <li>color: The color of the hovered node.</li>
       *   <li>data: The data object of the hovered node.</li>
       *   <li>componentElement: The treemap element.</li>
       * </ul>
       *  The function should return an Object that contains only one of the two properties:
       *  <ul>
       *    <li>insert: HTMLElement | string - An HTML element, which will be appended to the tooltip, or a tooltip string.</li> 
       *    <li>preventDefault: <code>true</code> - Indicates that the tooltip should not be displayed. It is not necessary to return {preventDefault:false} to display tooltip, since this is a default behavior.</li> 
       *  </ul>
       * @expose
       * @name tooltip.renderer
       * @memberof! oj.ojTreemap
       * @instance
       * @type {function(Object):Object|null}
       * @default null
       */
       renderer: null
     },
    
    /**
     * Specifies whether gaps are displayed between groups. Gaps can be useful for drawing attention to the differences between groups.
     * @expose
     * @name groupGaps
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "all"
     * @ojvalue {string} "none"
     * @ojvalue {string} "outer"
     * @default "outer"
     */
    groupGaps: "outer",
     
    /**
     * The number of levels of nodes to display. By default all nodes are displayed.
     * @expose
     * @name displayLevels
     * @memberof oj.ojTreemap
     * @instance
     * @type {number}
     * @default Number.MAX_VALUE
     */
    displayLevels: Number.MAX_VALUE,
     
    /**
     * Specifies the layout of the treemap. The squarified layout results in nodes that are as square as possible, for easier comparison of node sizes. The sliceAndDice layouts are useful for animation, as the ordering of the data is maintained. SliceAndDice layouts are also useful for small form factor treemaps.
     * @expose
     * @name layout
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "sliceAndDiceHorizontal"
     * @ojvalue {string} "sliceAndDiceVertical"
     * @ojvalue {string} "squarified"
     * @default "squarified"
     */
    layout: "squarified",
     
    /**
     * An array of objects with the following properties that defines the data for the nodes. Also accepts a Promise for deferred data rendering. No data will be rendered if the Promise is rejected.
     * @expose
     * @name nodes
     * @memberof oj.ojTreemap
     * @instance
     * @type {Array.<Object>|Promise|null}
     * @default null
     */
    nodes: null,
    
    /**
     * An object defining default properties for the nodes. Component CSS classes should be used to set component wide styling. 
     * This API should be used only for styling a specific instance of the component. Properties specified on this object may 
     * be overridden by specifications on the treemap nodes. Some property default values come from the CSS and varies based on theme.
     * @expose
     * @name nodeDefaults
     * @memberof oj.ojTreemap
     * @instance
     * @type {Object}
     * @default {"labelDisplay": "node", "groupLabelDisplay": "header", "labelHalign": "center", "labelMinLength": 1, "labelValign": "center", "header": {"labelHalign": "start", "isolate": "on", "useNodeColor": "off"}}
     */
    nodeDefaults: {
      /**
       * The CSS style object defining the style of the label. The CSS white-space property can be defined with value "nowrap" to disable default text wrapping.
       * The default value comes from the CSS and varies based on theme.
       * @expose
       * @name nodeDefaults.labelStyle
       * @memberof! oj.ojTreemap
       * @instance
       * @type {Object}
       */
      labelStyle: undefined,
      
      /**
       * The label display behavior for leaf nodes.
       * @expose
       * @name nodeDefaults.labelDisplay
       * @memberof! oj.ojTreemap
       * @instance
       * @type {string}
       * @ojvalue {string} "off"
       * @ojvalue {string} "node"
       * @default "node"
       */
      labelDisplay: "node",
      
      /**
       * The label display behavior for group nodes.
       * @expose
       * @name nodeDefaults.groupLabelDisplay
       * @memberof! oj.ojTreemap
       * @instance
       * @type {string}
       * @ojvalue {string} "node"
       * @ojvalue {string} "off"
       * @ojvalue {string} "header"
       * @default "header"
       */
      groupLabelDisplay: "header",
       
      /**
       * The horizontal alignment for labels displayed within the node.
       * @expose
       * @name nodeDefaults.labelHalign
       * @memberof! oj.ojTreemap
       * @instance
       * @type {string}
       * @ojvalue {string} "start"
       * @ojvalue {string} "end"
       * @ojvalue {string} "center"
       * @default "center"
       */
      labelHalign: "center",
       
      /**
       * The minimum number of visible characters needed in order to render a truncated label. If the minimum is not met when calculating the truncated label then the label is not displayed.
       * @expose
       * @name nodeDefaults.labelMinLength
       * @memberof! oj.ojTreemap
       * @instance
       * @type {number}
       * @default 1
       * @ojunits pixels
       */
      labelMinLength: 1,
       
      /**
       * The vertical alignment for labels displayed within the node.
       * @expose
       * @name nodeDefaults.labelValign
       * @memberof! oj.ojTreemap
       * @instance
       * @type {string}
       * @ojvalue {string} "top"
       * @ojvalue {string} "bottom"
       * @ojvalue {string} "center"
       * @default "center"
       */
      labelValign: "center",
       
      /**
       * The color of the node hover feedback. The default value comes from the CSS and varies based on theme.
       * @expose
       * @name nodeDefaults.hoverColor
       * @memberof! oj.ojTreemap
       * @instance
       * @type {string}
       */
      hoverColor: undefined,
       
      /**
       * The inner color of the node selection feedback. The default value comes from the CSS and varies based on theme.
       * @expose
       * @name nodeDefaults.selectedInnerColor
       * @memberof! oj.ojTreemap
       * @instance
       * @type {string}
       */
      selectedInnerColor: undefined,
      
      /**
       * The outer color of the node selection feedback. The default value comes from the CSS and varies based on theme.
       * @expose
       * @name nodeDefaults.selectedOuterColor
       * @memberof! oj.ojTreemap
       * @instance
       * @type {string}
       */
      selectedOuterColor: undefined,
       
      /**
       * An object defining default properties for the node header. Component CSS classes should be used to set component wide styling. 
       * This API should be used only for styling a specific instance of the component. Some property default values come from the CSS and varies based on theme.
       * @expose
       * @name nodeDefaults.header
       * @memberof! oj.ojTreemap
       * @instance
       * @type {Object}
       * @default {"labelHalign": "start", "isolate": "on", "useNodeColor": "off"}
       */
      header: {
        /**
         * The background color of the node headers. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.backgroundColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        backgroundColor: undefined,
        
        /**
         * The border color of the node headers. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.borderColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        borderColor: undefined,
         
        /**
         * The background color of the node hover feedback. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.hoverBackgroundColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        hoverBackgroundColor: undefined,
         
        /**
         * The inner color of the node hover feedback. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.hoverInnerColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        hoverInnerColor: undefined, 
        /**
         * The outer color of the node hover feedback. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.hoverOuterColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        hoverOuterColor: undefined,
         
        /**
         * The background color of the node selection feedback. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.selectedBackgroundColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        selectedBackgroundColor: undefined,
         
        /**
         * The inner color of the node selection feedback. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.selectedInnerColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        selectedInnerColor: undefined,
         
        /**
         * The outer color of the node selection feedback. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.selectedOuterColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         */
        selectedOuterColor: undefined,
         
        /**
         * The horizontal alignment of the header title.
         * @expose
         * @name nodeDefaults.header.labelHalign
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         * @ojvalue {string} "center"
         * @ojvalue {string} "end"
         * @ojvalue {string} "start"
         * @default "start"
         */
        labelHalign: "start",
         
        /**
         * The CSS style string defining the style of the header title. The default value comes from the CSS and varies based on theme.
         * @expose
         * @name nodeDefaults.header.labelStyle
         * @memberof! oj.ojTreemap
         * @instance
         * @type {Object}
         */
        labelStyle: undefined,
         
        /**
         * Specifies whether isolate behavior is enabled on the node.
         * @expose
         * @name nodeDefaults.header.isolate
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         * @ojvalue {string} "off"
         * @ojvalue {string} "on"
         * @default "on"
         */
        isolate: "on",
         
        /**
         * Specifies whether the node color should be displayed in the header.
         * @expose
         * @name nodeDefaults.header.useNodeColor
         * @memberof! oj.ojTreemap
         * @instance
         * @type {string}
         * @ojvalue {string} "on"
         * @ojvalue {string} "off"
         * @default "off"
         */  
        useNodeColor: "off" 
      }
    },
    
    /**
     * Specifies the visual effect for separating the nodes from each other. This allows for adjacent nodes of the same color to be distinguished.
     * @expose
     * @name nodeSeparators
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "bevels"
     * @ojvalue {string} "gaps"
     * @default "gaps"
     */
    nodeSeparators: "gaps",
     
    /**
     * Specifies the selection mode.
     * @expose
     * @name selectionMode
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "none"
     * @ojvalue {string} "single"
     * @ojvalue {string} "multiple"
     * @default "multiple"
     */
    selectionMode: "multiple",
     
    /**
     * Specifies whether the nodes are sorted by size. When sorting is enabled, nodes that have the same parent are sorted in order of descending size.
     * @expose
     * @name sorting
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "off"
     */
    sorting: "off",
    
    /**
     * Specifies the label describing the color metric of the treemap. This label will be used in the legend.
     * @expose
     * @name colorLabel
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @default ""
     */
    colorLabel: "",
     
    /**
     * Specifies the label describing the size metric of the treemap. This label will be used in the legend.
     * @expose
     * @name sizeLabel
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @default ""
     */
    sizeLabel: "",
     
    /**
     * Specifies whether drilling is enabled. Drillable nodes will show a pointer cursor on hover and fire an 
     * <code class="prettyprint">ojBeforeDrill</code> and <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). Drilling on a node causes a property change to the rootNode attribute. The displayLevels attribute can be used in conjunction with drilling to display very deep hieracrchies. Use "on" to enable drilling for all nodes. To enable or disable drilling on individual nodes use the drilling attribute in each node.
     * @expose
     * @name drilling
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "on"
     * @ojvalue {string} "off"
     * @default "off"
     */
    drilling: "off",
     
    /**
     * The id of the root node. When specified, only the root node and children of the root will be displayed.
     * @expose
     * @name rootNode
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @default ""
     */
    rootNode: "",
     
    /**
     * An array containing the ids of the initially selected nodes.
     * @expose
     * @name selection
     * @memberof oj.ojTreemap
     * @instance
     * @type {Array.<string>}
     * @default []
     * @ojwriteback
     */
    selection: [],
    
    /**
     * The id of the initially isolated node.
     * @expose
     * @name isolatedNode
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @default ""
     * @ojwriteback
     */
    isolatedNode: "",
    
    /**
     * Data visualizations require a press and hold delay before triggering tooltips and rollover effects on mobile devices to avoid interfering with page panning, but these hold delays can make applications seem slower and less responsive. 
     * For a better user experience, the application can remove the touch and hold delay when data visualizations are used within a non scrolling container or if there is sufficient space outside of the visualization for panning. 
     * If touchResponse is touchStart the element will instantly trigger the touch gesture and consume the page pan events. If touchResponse is auto, the element will behave like touchStart if it determines that it is not rendered within scrolling content and if panning is not available for those elements that support the feature. 
     * @expose
     * @name touchResponse
     * @memberof oj.ojTreemap
     * @instance
     * @type {string}
     * @ojvalue {string} "touchStart"
     * @ojvalue {string} "auto"
     * @default "auto"
     */  
    touchResponse: "auto",
     
    /**
     * Triggered immediately before any node in the treemap is drilled into. The drill event can be vetoed if the beforeDrill callback returns false.
     *
     * @property {string} id the id of the drilled object
     * @property {Object} data the data object of the drilled node
     *
     * @expose
     * @event
     * @memberof oj.ojTreemap
     * @instance
     */
    beforeDrill: null,
    /**
     * Triggered during a drill gesture (double click if selection is enabled, single click otherwise).
     *
     * @property {string} id the id of the drilled object
     * @property {Object} data the data object of the drilled node
     *
     * @expose
     * @event
     * @memberof oj.ojTreemap
     * @instance
     */
    drill: null,
  },

    //** @inheritdoc */
    _CreateDvtComponent: function(context, callback, callbackObj) {
      return dvt.Treemap.newInstance(context, callback, callbackObj);
    },

    //** @inheritdoc */
    _ConvertLocatorToSubId : function(locator) {
      var subId = locator['subId'];

      // Convert the supported locators
      if(subId == 'oj-treemap-node') {
        // node[index0][index1]...[indexN]
        subId = 'node' + this._GetStringFromIndexPath(locator['indexPath']);
      }
      else if(subId == 'oj-treemap-tooltip') {
        subId = 'tooltip';
      }

      // Return the converted result or the original subId if a supported locator wasn't recognized. We will remove
      // support for the old subId syntax in 1.2.0.
      return subId;
    },

    //** @inheritdoc */
    _ConvertSubIdToLocator : function(subId) {
      var locator = {};
      if(subId.indexOf('node') == 0) {
        // node[index0][index1]...[indexN]
        locator['subId'] = 'oj-treemap-node';
        locator['indexPath'] = this._GetIndexPath(subId);
      }
      else if(subId == 'tooltip') {
        locator['subId'] = 'oj-treemap-tooltip';
      }
      return locator;
    },

    //** @inheritdoc */
    _GetComponentStyleClasses: function() {
      var styleClasses = this._super();
      styleClasses.push('oj-treemap');
      return styleClasses;
    },

    //** @inheritdoc */
    _GetChildStyleClasses: function() {
      var styleClasses = this._super();
      styleClasses['oj-dvtbase oj-treemap'] = {'path': 'animationDuration', 'property': 'ANIM_DUR'};
      styleClasses['oj-treemap-attribute-type-text'] = {'path': 'styleDefaults/_attributeTypeTextStyle', 'property': 'TEXT'};
      styleClasses['oj-treemap-attribute-value-text'] = {'path': 'styleDefaults/_attributeValueTextStyle', 'property': 'TEXT'};
      styleClasses['oj-treemap-drill-text '] = {'path' : 'styleDefaults/_drillTextStyle', 'property' : 'TEXT'};
      styleClasses['oj-treemap-current-drill-text '] = {'path' : 'styleDefaults/_currentTextStyle', 'property' : 'TEXT'};
      styleClasses['oj-treemap-node'] = {'path': 'nodeDefaults/labelStyle', 'property': 'TEXT'};
      styleClasses['oj-treemap-node oj-hover'] = {'path': 'nodeDefaults/hoverColor', 'property': 'border-top-color'};
      styleClasses['oj-treemap-node oj-selected'] = [
        {'path': 'nodeDefaults/selectedOuterColor', 'property': 'border-top-color'},
        {'path': 'nodeDefaults/selectedInnerColor', 'property': 'border-bottom-color'}
      ];
      styleClasses['oj-treemap-node-header'] = [
        {'path': 'nodeDefaults/header/backgroundColor', 'property': 'background-color'},
        {'path': 'nodeDefaults/header/borderColor', 'property': 'border-top-color'},
        {'path': 'nodeDefaults/header/labelStyle', 'property': 'TEXT'}
      ];
      styleClasses['oj-treemap-node-header oj-hover'] = [
        {'path': 'nodeDefaults/header/hoverBackgroundColor', 'property': 'background-color'},
        {'path': 'nodeDefaults/header/hoverOuterColor', 'property': 'border-top-color'},
        {'path': 'nodeDefaults/header/hoverInnerColor', 'property': 'border-bottom-color'},
        {'path': 'nodeDefaults/header/_hoverLabelStyle', 'property': 'TEXT'}
      ];
      styleClasses['oj-treemap-node-header oj-selected'] = [
        {'path': 'nodeDefaults/header/selectedBackgroundColor', 'property': 'background-color'},
        {'path': 'nodeDefaults/header/selectedOuterColor', 'property': 'border-top-color'},
        {'path': 'nodeDefaults/header/selectedInnerColor', 'property': 'border-bottom-color'},
        {'path': 'nodeDefaults/header/_selectedLabelStyle', 'property': 'TEXT'}
      ];
      return styleClasses;
    },

    //** @inheritdoc */
    _GetEventTypes : function() {
      return ['optionChange', 'drill', 'beforeDrill'];
    },

    //** @inheritdoc */
    _GetTranslationMap: function() {
      // The translations are stored on the options object.
      var translations = this.options['translations'];

      // Safe to modify super's map because function guarentees a new map is returned
      var ret = this._super();
      ret['DvtTreemapBundle.COLOR'] = translations['labelColor'];
      ret['DvtTreemapBundle.ISOLATE'] = translations['tooltipIsolate'];
      ret['DvtTreemapBundle.RESTORE'] = translations['tooltipRestore'];
      ret['DvtTreemapBundle.SIZE'] = translations['labelSize'];
      ret['DvtUtilBundle.TREEMAP'] = translations['componentName'];
      return ret;
    },

    //** @inheritdoc */
    _HandleEvent: function(event) {
      var type = event['type'];
      if (type === 'isolate') {
        // Keep track of all isolated nodes
        var isolatedNodes = this.options._isolatedNodes;
        if (!isolatedNodes) {
          this.options._isolatedNodes = [];
          isolatedNodes = this.options._isolatedNodes;
        }

        // If event has id, it's an isolate.  If null id, then restore.
        var isolateType;
        var isolatedNode = event['id'];
        if (isolatedNode) {
          isolateType = "on";
          isolatedNodes.push(isolatedNode);
          this._UserOptionChange('isolatedNode', isolatedNode);
        }
        else {
          isolateType = "off";
          isolatedNode = isolatedNodes.pop();
          this._UserOptionChange('isolatedNode', (isolatedNodes.length > 0) ? isolatedNodes[isolatedNodes.length] : null);
        }
      }
      else if(type == 'drill') {
        var eventData = {'id': event['id'], 'data': event['data']};
        if (!this._IsCustomElement())
          eventData['component'] = event['component'];
             
        if(event['id'] && this._trigger('beforeDrill', null, eventData)) {
          this._UserOptionChange('rootNode', event['id']);
          this._Render();
          this._trigger('drill', null, eventData);
        }
      }
      else {
        this._super(event);
      }
    },
    
    //** @inheritdoc */
    _GetComponentRendererOptions: function() {
      return ['tooltip/renderer', 'nodeContent/renderer'];
    },

    //** @inheritdoc */
    _ProcessOptions: function() {
      this._super();
      var nodeContent = this.options['nodeContent'];
      if (nodeContent && nodeContent['_renderer'])
        nodeContent['renderer'] = this._GetTemplateRenderer(nodeContent['_renderer'], 'nodeContent');
    },
  
    //** @inheritdoc */
    _LoadResources : function() {
      // Ensure the resources object exists
      if(this.options['_resources'] == null)
        this.options['_resources'] = {};
  
      var resources = this.options['_resources'];

      // Add isolate and restore icons
      resources['isolate'] = 'oj-treemap-isolate-icon';
      resources['isolateOver'] = 'oj-treemap-isolate-icon oj-hover';
      resources['isolateDown'] = 'oj-treemap-isolate-icon oj-active';

      resources['restore'] = 'oj-treemap-restore-icon';
      resources['restoreOver'] = 'oj-treemap-restore-icon oj-hover';
      resources['restoreDown'] = 'oj-treemap-restore-icon oj-active';
    },

    /**
     * Returns an object with the following properties for automation testing verification of the node with
     * the specified subid path.
     *
     * @param {Array} subIdPath The array of indices in the subId for the desired node
     * @property {string} color
     * @property {string} label
     * @property {boolean} selected
     * @property {number} size
     * @property {string} tooltip
     * @return {Object|null} An object containing properties for the node, or null if none exists.
     * @expose
     * @instance
     * @memberof oj.ojTreemap
     */
    getNode: function(subIdPath) {
      return this._component.getAutomation().getNode(subIdPath);
    },

    /**
     * {@ojinclude "name":"nodeContextDoc"}
     * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
     * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
     *
     * @example {@ojinclude "name":"nodeContextExample"}
     *
     * @expose
     * @instance
     * @memberof oj.ojTreemap
     */
    getContextByNode: function(node)
    {
      // context objects are documented with @ojnodecontext
      var context = this.getSubIdByNode(node);
      if (context && context['subId'] !== 'oj-treemap-tooltip')
        return context;

      return null;
    },

    //** @inheritdoc */
    _GetComponentDeferredDataPaths : function() {
      return {'root': ['nodes']};
    }
  });

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="3">Node</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Select when <code class="prettyprint">selectionMode</code> is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td rowspan="2"><kbd>Press & Hold</kbd></td>
 *       <td>Display tooltip.</td>
 *     </tr>
 *     <tr>
 *       <td>Display context menu on release.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojTreemap
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Move focus to next element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + Tab</kbd></td>
 *       <td>Move focus to previous element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Move focus and selection up to the nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Move focus and selection down to the nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>Move focus and selection to the left to nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>Move focus and selection to the right to nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>]</kbd> or <kbd>Alt + UpArrow</kbd></td>
 *       <td>Move focus and selection from tile or header to group header in level above.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>[</kbd> or <kbd>Alt + DownArrow</kbd></td>
 *       <td>Move focus and selection from tile or header to group header in level below.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + UpArrow</kbd></td>
 *       <td>Move focus and extend selection up to the nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + DownArrow</kbd></td>
 *       <td>Move focus and extend selection down to the nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + LeftArrow</kbd></td>
 *       <td>Move focus and extend selection to the left to nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + RightArrow</kbd></td>
 *       <td>Move focus and extend selection to the right to nearest tile or header in same level.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + ]</kbd> or <kbd>Shift + Alt + UpArrow</kbd></td>
 *       <td>Move focus and extend selection from tile or header to group header in level above.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift + [</kbd> or <kbd>Shift + Alt + DownArrow</kbd></td>
 *       <td>Move focus and extend selection from tile or header to group header in level below.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + UpArrow</kbd></td>
 *       <td>Move focus up to the nearest tile or header in same level, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + DownArrow</kbd></td>
 *       <td>Move focus down to the nearest tile or header in same level, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + LeftArrow</kbd></td>
 *       <td>Move focus to the left to nearest tile or header in same level, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + RightArrow</kbd></td>
 *       <td>Move focus to the right to nearest tile or header in same level, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + ]</kbd> or <kbd>Ctrl + Alt + UpArrow</kbd></td>
 *       <td>Move focus from tile or header to group header in level above, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + [</kbd> or <kbd>Ctrl + Alt + DownArrow</kbd></td>
 *       <td>Move focus from tile or header to group header in level below, without changing the current selection.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Spacebar</kbd></td>
 *       <td>Multi-select tiles or headers with focus.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl + Enter</kbd></td>
 *       <td>Maximize/Restore on a group header.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Drill on a node when <code class="prettyprint">drilling</code> is enabled.</td>
 *     </tr>
 *   </tbody>
 * </table>
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojTreemap
 */

/**
 * An array of objects with properties for the child nodes.
 * @expose
 * @name nodes[].nodes
 * @memberof! oj.ojTreemap
 * @instance
 * @type {Array.<Object>}
 * @default null
 */
/**
 * The id of the node.
 * @expose
 * @name nodes[].id
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @default null
 */
/**
 * An optional array of category strings corresponding to this data item. This enables highlighting and filtering of individual data items through interactions with the legend and other visualization elements. The categories array of each node is required to be a superset of the categories array of its parent node. If not specified, the ids of the node and its ancestors will be used.
 * @expose
 * @name nodes[].categories
 * @memberof! oj.ojTreemap
 * @instance
 * @type {Array.<string>}
 * @default null
 */
/**
 * The value of the node. The value determines the relative size of the node.
 * @expose
 * @name nodes[].value
 * @memberof! oj.ojTreemap
 * @instance
 * @type {number}
 * @default null
 */
/**
 * The fill color of the node.
 * @expose
 * @name nodes[].color
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The CSS style class to apply to the node. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the node color attribute.
 * @expose
 * @name nodes[].svgClassName
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The inline style to apply to the node. The style class and inline style will override any other styling specified through the properties. For tooltip interactivity, it's recommended to also pass a representative color to the node color attribute.
 * @expose
 * @name nodes[].svgStyle
 * @memberof! oj.ojTreemap
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The label for this node.
 * @expose
 * @name nodes[].label
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @default null
 */
/**
 * Specifies whether drilling is enabled for the node. Drillable nodes will show a pointer cursor on hover and fire an <code class="prettyprint">ojDrill</code> event on click (double click if selection is enabled). To enable drilling for all nodes at once, use the drilling attribute in the top level.
 * @expose
 * @name nodes[].drilling
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @ojvalue {string} "inherit"
 * @default "inherit"
 */
/**
 * The description of this node. This is used for accessibility and also for customizing the tooltip text.
 * @expose
 * @name nodes[].shortDesc
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @default null
 */
/**
 * The pattern used to fill the node.
 * @expose
 * @name nodes[].pattern
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "smallChecker"
 * @ojvalue {string} "smallCrosshatch"
 * @ojvalue {string} "smallDiagonalLeft"
 * @ojvalue {string} "smallDiagonalRight"
 * @ojvalue {string} "smallDiamond"
 * @ojvalue {string} "smallTriangle"
 * @ojvalue {string} "largeChecker"
 * @ojvalue {string} "largeCrosshatch"
 * @ojvalue {string} "largeDiagonalLeft"
 * @ojvalue {string} "largeDiagonalRight"
 * @ojvalue {string} "largeDiamond"
 * @ojvalue {string} "largeTriangle"
 * @ojvalue {string} "none"
 * @default "none"
 */
/**
 * The CSS style object defining the style of the label. The CSS white-space property can be defined with value "nowrap" to disable default text wrapping.
 * @expose
 * @name nodes[].labelStyle
 * @memberof! oj.ojTreemap
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The label display behavior for leaf nodes.
 * @expose
 * @name nodes[].labelDisplay
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "node"
 * @default "node"
 */
/**
 * The label display behavior for group nodes.
 * @expose
 * @name nodes[].groupLabelDisplay
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "node"
 * @ojvalue {string} "off"
 * @ojvalue {string} "header"
 * @default "header"
 */
/**
 * The horizontal alignment for labels displayed within the node.
 * @expose
 * @name nodes[].labelHalign
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "start"
 * @ojvalue {string} "end"
 * @ojvalue {string} "center"
 * @default "center"
 */
/**
 * The vertical alignment for labels displayed within the node.
 * @expose
 * @name nodes[].labelValign
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "top"
 * @ojvalue {string} "bottom"
 * @ojvalue {string} "center"
 * @default "center"
 */
/**
 * Specifies whether or not the node will be selectable.
 * @expose
 * @name nodes[].selectable
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "auto"
 * @default "auto"
 */
/**
 * An object defining the properties for the node header.
 * @expose
 * @name nodes[].header
 * @memberof! oj.ojTreemap
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * The horizontal alignment of the header title.
 * @expose
 * @name nodes[].header.labelHalign
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "center"
 * @ojvalue {string} "end"
 * @ojvalue {string} "start"
 * @default "start"
 */
/**
 * The CSS style object defining the style of the header title.
 * @expose
 * @name nodes[].header.labelStyle
 * @memberof! oj.ojTreemap
 * @instance
 * @type {Object}
 * @default null
 */
/**
 * Specifies whether isolate behavior is enabled on the node.
 * @expose
 * @name nodes[].header.isolate
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "off"
 * @ojvalue {string} "on"
 * @default "on"
 */
/**
 * Specifies whether the node color should be displayed in the header.
 * @expose
 * @name nodes[].header.useNodeColor
 * @memberof! oj.ojTreemap
 * @instance
 * @type {string}
 * @ojvalue {string} "on"
 * @ojvalue {string} "off"
 * @default "off"
 */

// SubId Locators **************************************************************

/**
 * <p>Sub-ID for treemap nodes indexed by their position in the hierarchy.</p>
 *
 * @property {Array} indexPath The array of numerical indices for the node.
 *
 * @ojsubid oj-treemap-node
 * @memberof oj.ojTreemap
 *
 * @example <caption>Get the node at index 0 in the first layer, index 1 in the second:</caption>
 * var nodes = myTreemap.getNodeBySubId( {'subId': 'oj-treemap-node', 'indexPath': [0, 1]} );
 */

/**
 * <p>Sub-ID for the the treemap tooltip.</p>
 * 
 * <p>See the <a href="#getNodeBySubId">getNodeBySubId</a> and 
 * <a href="#getSubIdByNode">getSubIdByNode</a> methods for details.</p>
 * 
 * @ojsubid
 * @member
 * @name oj-treemap-tooltip
 * @memberof oj.ojTreemap
 * @instance
 * 
 * @example <caption>Get the tooltip object of the treemap, if displayed:</caption>
 * var nodes =  myTreemap.getNodeBySubId( {'subId': 'oj-treemap-tooltip'} );
 */
// Node Context Objects ********************************************************

/**
 * <p>Context for treemap nodes indexed by their position in the hierarchy.</p>
 *
 * @property {Array} indexPath The array of numerical indices for the node.
 *
 * @ojnodecontext oj-treemap-node
 * @memberof oj.ojTreemap
 */
/**
 * Ignore tag only needed for DVTs that have jsDoc in separate _doc.js files.
 * @ignore
 */
(function() {
var ojTreemapMeta = {
  "properties": {
    "animationDuration": {
      "type": "number"
    },
    "animationOnDataChange": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationOnDisplay": {
      "type": "string",
      "enumValues": ["auto", "none"]
    },
    "animationUpdateColor": {
      "type": "string"
    },
    "colorLabel": {
      "type": "string"
    },
    "displayLevels": {
      "type": "number"
    },
    "drilling": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "groupGaps": {
      "type": "string",
      "enumValues": ["all", "none", "outer"]
    },
    "hiddenCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightedCategories": {
      "type": "Array<string>",
      "writeback": true
    },
    "highlightMatch": {
      "type": "string",
      "enumValues": ["any", "all"]
    },
    "hoverBehavior": {
      "type": "string",
      "enumValues": ["dim", "none"]
    },
    "hoverBehaviorDelay": {
      "type": "number"
    },    "isolatedNode": {
      "type": "string",
      "writeback": true
    },
    "layout": {
      "type": "string",
      "enumValues": ["sliceAndDiceHorizontal", "sliceAndDiceVertical", "squarified"]
    },
    "nodeDefaults": {
      "type": "object",
      "properties": {
        "groupLabelDisplay": {
          "type": "string",
          "enumValues": ["node", "off", "header"]
        },
        "header": {
          "type": "object",
          "properties": {
            "backgroundColor": {
              "type": "string"
            },
            "borderColor": {
              "type": "string"
            },
            "hoverBackgroundColor": {
              "type": "string"
            },
            "hoverInnerColor": {
              "type": "string"
            },
            "hoverOuterColor": {
              "type": "string"
            },
            "isolate": {
              "type": "string",
              "enumValues": ["on", "off"]
            },
            "labelHalign": {
              "type": "string",
              "enumValues": ["center", "end", "start"] 
            },
            "labelStyle": {
              "type": "object"
            },
            "selectedBackgroundColor": {
              "type": "string"
            },
            "selectedInnerColor": {
              "type": "string"
            },
            "selectedOuterColor": {
              "type": "string"
            },
            "useNodeColor": {
              "type": "string",
              "enumValues": ["on", "off"]
            }
          }
        },
        "hoverColor": {
          "type": "string"
        },
        "labelDisplay": {
          "type": "string",
          "enumValues": ["off", "node"]
        },
        "labelHalign": {
          "type": "string",
          "enumValues": ["start", "end", "center"]
        },
        "labelStyle": {
          "type": "object"
        },
        "labelValign": {
          "type": "string",
          "enumValues": ["top", "bottom", "center"]
        },
        "selectedInnerColor": {
          "type": "string"
        },
        "selectedOuterColor": {
          "type": "string"
        }
      }
    },
    "nodes": {
      "type": "Array<object>|Promise"
    },
    "nodeContent": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "nodeSeparators": {
      "type": "string",
      "enumValues": ["bevels", "gap"]
    },
    "rootNode": {
      "type": "string"
    },
    "selection": {
      "type": "Array<string>",
      "writeback": true
    },
    "selectionMode": {
      "type": "string",
      "enumValues": ["none", "single", "multiple"]
    },
    "sizeLabel": {
      "type": "string"
    },
    "sorting": {
      "type": "string",
      "enumValues": ["on", "off"]
    },
    "tooltip": {
      "type": "object",
      "properties": {
        "renderer": {}
      }
    },
    "touchResponse": {
      "type": "string"
    },
    "translations": {
      "type": "Object",
      "properties": {
        "componentName": {
          "type": "string",
          "value": "Treemap"
        },
        "labelAndValue": {
          "type": "string",
          "value": "{0}: {1}"
        },
        "labelClearSelection": {
          "type": "string",
          "value": "Clear Selection"
        },
        "labelColor": {
          "type": "string",
          "value": "Color"
        },
        "labelCountWithTotal": {
          "type": "string",
          "value": "{0} of {1}"
        },
        "labelDataVisualization": {
          "type": "string",
          "value": "Data Visualization"
        },
        "labelInvalidData": {
          "type": "string",
          "value": "Invalid data"
        },
        "labelNoData": {
          "type": "string",
          "value": "No data to display"
        },
        "labelSize": {
          "type": "string",
          "value": "Size"
        },
        "stateCollapsed": {
          "type": "string",
          "value": "Collapsed"
        },
        "stateDrillable": {
          "type": "string",
          "value": "Drillable"
        },
        "stateExpanded": {
          "type": "string",
          "value": "Expanded"
        },
        "stateHidden": {
          "type": "string",
          "value": "Hidden"
        },
        "stateIsolated": {
          "type": "string",
          "value": "Isolated"
        },
        "stateMaximized": {
          "type": "string",
          "value": "Maximized"
        },
        "stateMinimized": {
          "type": "string",
          "value": "Minimized"
        },
        "stateSelected": {
          "type": "string",
          "value": "Selected"
        },
        "stateUnselected": {
          "type": "string",
          "value": "Unselected"
        },
        "stateVisible": {
          "type": "string",
          "value": "Visible"
        },
        "tooltipIsolate": {
          "type": "string",
          "value": "Isolate"
        },
        "tooltipRestore": {
          "type": "string",
          "value": "Restore"
        }
      }
    }
  },
  "events": {
    "beforeDrill": {},
    "drill": {}
  },
  "methods": {
    "getContextByNode": {},
    "getNode": {}
  },
  "extension": {
    _WIDGET_NAME: "ojTreemap"
  }
};
oj.CustomElementBridge.registerMetadata('oj-treemap', 'dvtBaseComponent', ojTreemapMeta);
oj.CustomElementBridge.register('oj-treemap', {'metadata': oj.CustomElementBridge.getMetadata('oj-treemap')});
})();

/**
 * @ojcomponent oj.ojTreeView
 * @augments oj.baseComponent
 * @since 4.0.0
 * @ojstatus preview
 * @ojshortdesc Displays hierarchical relationships between the items in a tree.
 * @ojtsignore
 * @ojrole tree
 * @ojsignature [{
 *                target: "Type",
 *                value: "class ojTreeView<K, D> extends baseComponent<ojTreeViewSettableProperties<K,D>>"
 *               },
 *               {
 *                target: "Type",
 *                value: "ojTreeViewSettableProperties<K,D> extends baseComponentSettableProperties",
 *                for: "SettableProperties"
 *               }]
 *
 * @classdesc
 * <h3 id="treeViewOverview-section">
 *   JET TreeView
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#treeViewOverview-section"></a>
 * </h3>
 *
 * <p>The JET TreeView allows a user to display the hierarchical relationship between the items of a tree.</p>
 *
 * <h3 id="data-section">
 *   Data
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#data-section"></a>
 * </h3>
 * <p>The JET TreeView gets its data in two different ways. The first way is from a TreeDataSource. There are two types
 * of TreeDataSource that are available out of the box:</p>
 * <ul>
 * <li><b>oj.JsonTreeDataSource</b> - Use this when the underlying data is a JSON object.
 * See the documentation for <a href="oj.JsonTreeDataSource.html">oj.JsonTreeDataSource</a>
 * for more details on the available options.</li>
 * <li><b>oj.CollectionTreeDataSource</b> - Use this when oj.Collection is the model for each group of data.
 * See the documentation for <a href="oj.CollectionTreeDataSource.html">oj.CollectionTreeDataSource</a>
 * for more details on the available options.</li>
 * </ul>
 *
 * <p>The second way is using static HTML content as data.</p>
 *
 * <p>Example of static content</p>
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-tree-view id="treeview1">
 *   &lt;ul>
 *     &lt;li>
 *       &lt;a id="group1" href="#">Group 1&lt;/a>
 *       &lt;ul>
 *         &lt;li>&lt;a id="item1-1" href="#">Item 1-1&lt;/a>&lt;/li>
 *         &lt;li>&lt;a id="item1-2" href="#">Item 1-2&lt;/a>&lt;/li>
 *       &lt;/ul>
 *     &lt;/li>
 *     &lt;li>
 *       &lt;a id="group2" href="#">Group 2&lt;/a>
 *       &lt;ul>
 *         &lt;li>&lt;a id="item2-1" href="#">Item 2-1&lt;/a>&lt;/li>
 *         &lt;li>&lt;a id="item2-2" href="#">Item 2-2&lt;/a>&lt;/li>
 *       &lt;/ul>
 *     &lt;/li>
 *   &lt;/ul>
 * &lt;/oj-tree-view>
 * </code></pre>
 *
 * <h3 id="touch-section">
 *   Touch End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#touch-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"touchDoc"}
 *
 * <h3 id="keyboard-section">
 *   Keyboard End User Information
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#keyboard-section"></a>
 * </h3>
 *
 * {@ojinclude "name":"keyboardDoc"}
 *
 * <h3 id="context-section">
 *   Item Context
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#context-section"></a>
 * </h3>
 *
 * <p>For item attributes, developers can specify a function as the return value.
 * The function takes a single argument, which is an object that contains contextual
 * information about the particular item. This gives developers the flexibility
 * to return different value depending on the context.</p>
 *
 * <p>The context parameter contains the following keys:</p>
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Key</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>componentElement</kbd></td>
 *       <td>The TreeView element.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>data</kbd></td>
 *       <td>The data object for the item (not available for static content).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>datasource</kbd></td>
 *       <td>A reference to the data source object (not available for static content).</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>depth</kbd></td>
 *       <td>The depth of the item. The depth of the first level children under the invisible root is 1.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>index</kbd></td>
 *       <td>The index of the item relative to its parent, where 0 is the index of the first item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>key</kbd></td>
 *       <td>The key of the item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>leaf</kbd></td>
 *       <td>Whether the item is a leaf item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>parentElement</kbd></td>
 *       <td>The TreeView item element. The renderer can use this to directly append content.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>parentKey</kbd></td>
 *       <td>The key of the parent item. The parent key is null for root item.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * <h3 id="accessibility-section">
 *   Accessibility
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#accessibility-section"></a>
 * </h3>
 *
 * <p>Application must ensure that the context menu is available and setup with the
 * appropriate clipboard menu items so that keyboard-only users are able to reorder items
 * just by using the keyboard.
 *
 * <h3 id="perf-section">
 *   Performance
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#perf-section"></a>
 * </h3>
 *
 * <h4>Data Set Size</h4>
 * <p>As a rule of thumb, it is recommended that applications limit the amount of data to display. Displaying large
 * number of items in TreeView makes it hard for user to find what they are looking for, but affects the load time and
 * scrolling performance as well.</p>
 *
 * <h4>Item Content</h4>
 * <p>TreeView allows developers to specify arbitrary content inside its item. In order to minimize any negative effect on
 * performance, please avoid putting a large number of heavy-weight components inside because as it adds more complexity
 * to the structure, and the effect will be multiplied because there can be many items in the TreeView.</p>
 *
 * <h4>Expand All</h4>
 * <p>While TreeView provides a convenient way to initially expand all parent items in the TreeView, it might have an impact
 * on the initial rendering performance since expanding each parent item might cause a fetch from the server depending on
 * the TreeDataSource. Other factors that could impact performance includes the depth of the tree, and the number of children
 * in each level.</p>
 *
 * <h3 id="animation-section">
 *   Animation
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#animation-section"></a>
 * </h3>
 *
 * <p>Applications can customize animations triggered by actions in TreeView by either listening for <code class="prettyprint">animateStart/animateEnd</code>
 *    events or overriding action specific style classes on the animated item.  See the documentation of <a href="oj.AnimationUtils.html">oj.AnimationUtils</a> 
 *    class for details.</p>
 *
 * <p>The following are actions and their corresponding sass variables in which applications can use to customize animation effects.
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Action</th>
 *       <th>Sass Variable</th>
 *       <th>Description</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>expand</kbd></td>
 *       <td>$treeViewExpandAnimation</td>
 *       <td>When user expands an item.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>collapse</kbd></td>
 *       <td>$treeViewCollapseAnimation</td>
 *       <td>When user collapses an item.</td>
 *     </tr>
 *   </tbody>
 * </table>
 */
(function()
{
  oj.__registerWidget('oj.ojTreeView', $['oj']['baseComponent'],
  {
    version: '1.0.0',
    defaultElement: '<div>',
    widgetEventPrefix: 'oj',
    options: {
      /**
       * The key of the item that has the browser focus.
       * This is a read-only attribute so page authors cannot set or change it directly.
       *
       * @expose
       * @public
       * @type {*}
       * @instance
       * @memberof! oj.ojTreeView
       * @readonly
       * @ojwriteback
       * @ojsignature {target:"Type", value:"K"}
       *
       * @example <caption>Get the current item:</caption>
       * myTreeView.currentItem;
       */
      currentItem: null,

      /**
       * The data source for the TreeView. Accepts an instance of oj.TreeDataSource.
       * See the data source section in the introduction for out of the box data source types.
       * If the data attribute is not specified, the child elements are used as content. If there's no
       * content specified, then an empty list is rendered.
       *
       * @expose
       * @public
       * @type {oj.TreeDataSource}
       * @instance
       * @memberof! oj.ojTreeView
       * @default null
       *
       * @example <caption>Initialize the TreeView with an oj.Collection:</caption>
       * myTreeView.data = new oj.CollectionTableDataSource(collection);
       */
      data: null,

      /**
       * Enable drag and drop functionality.<br><br>
       * JET provides support for HTML5 Drag and Drop events.  Please refer to {@link https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/Drag_and_drop third party documentation} 
       * on HTML5 Drag and Drop to learn how to use it.
       *
       * @expose
       * @memberof! oj.ojTreeView
       * @type {Object}
       * @default {"drag": null, "drop": null}
       * @instance
       */
      dnd: {
        /**
         * @expose
         * @alias dnd.drag
         * @memberof! oj.ojTreeView
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         * @default null
         * @property {Object} items If this object is specified, TreeView will initiate drag operation when the user drags on an item.
         * @property {string|Array.<string>} [items.dataTypes] (optional) The MIME types to use for the dragged data in the dataTransfer object. This can be a string if there is only one
         * type, or an array of strings if multiple types are needed.<br><br>
         * For example, if selected items of employee data are being dragged, dataTypes could be "application/employees+json". Drop targets can examine the data types and decide
         * whether to accept the data. A text input may only accept "text" data type, while a chart for displaying employee data may be configured to accept the "application/employees+json" type.<br><br>
         * For each type in the array, dataTransfer.setData will be called with the specified type and the JSON version of the selected item data as the value. The selected item data 
         * is an array of objects, with each object representing a model object from the underlying data source. For example, if the underlying data is an oj.Collection, then this
         * would be a oj.Model object. Note that when static HTML is used, then the value would be the HTML string of the selected item.<br><br>
         * This property is required unless the application calls setData itself in a dragStart callback function.
         * @property {function(Event, {items: Array.<D>}):void} [items.dragStart] (optional) A callback function that receives the "dragstart" event and context information as arguments:<br><br>
         * <code class="prettyprint">function(event, context)</code><br><br>
         * All of the event payloads listed below can be found under the <code class="prettyprint">context</code> argument.
         * <ul><li><code class="prettyprint">items</code>: An array of objects, with each object representing the data of one selected item.</li></ul><br>
         * This function can set its own data and drag image as needed. If dataTypes is specified, event.dataTransfer is already populated with the default data when this function is invoked.
         * If dataTypes is not specified, this function must call event.dataTransfer.setData to set the data or else the drag operation will be cancelled. In either case, the drag image is
         * set to an image of the dragged items on the TreeView.
         * @property {function(Event):void} [items.drag] (optional) A callback function that receives the "drag" event as argument:<br><br>
         * <code class="prettyprint">function(event)</code><br><br>
         * @property {function(Event):void} [items.dragEnd] (optional) A callback function that receives the "dragend" event as argument:<br><br>
         * <code class="prettyprint">function(event)</code><br><br>
         *
         * @ojsignature {target: "Type",
         *               value: "?((event: Event, context: {items: Array<D>}) => void)",
         *               for: "items.dragStart"}
         *
         * @example <caption>Initialize the TreeView such that only leaf items are focusable:</caption>
         * myTreeView.setProperty('dnd.drag.items', {
         *   'dataTypes': ['application/ojtreeviewitems+json'],
         *   'dragEnd': handleDragEnd
         * });
         */
        drag: null,

        /**
         * @typedef {Object} oj.ojTreeView.ItemsDropOnDropContext
         * @property {Element} item The item being dropped on.
         * @property {'inside'|'before'|'after'|'first'} position The drop position relative to the item being dropped on.
         */
        /**
         * @expose
         * @alias dnd.drop
         * @memberof! oj.ojTreeView
         * @instance
         * @type {Object}
         * @ojsignature {target: "Type", value: "?"}
         * @default null
         * @property {Object} items  An object that specifies callback functions to handle dropping items<br><br>
         * @property {string|Array.<string>} [items.dataTypes] A data type or an array of data types this component can accept.<br><br>
         * This property is required unless dragEnter, dragOver, and drop callback functions are specified to handle the corresponding events.
         * @property {function(Event, {item: Element}):void} [items.dragEnter] (optional) A callback function that receives the "dragenter" event and context information as arguments:<br><br>
         * <code class="prettyprint">function(event, context)</code><br><br>
         * All of the event payloads listed below can be found under the <code class="prettyprint">context</code> argument.
         * <ul><li><code class="prettyprint">item</code>: The item being entered.</li></ul><br>
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
         * Otherwise, dataTypes will be matched against the drag dataTypes to determine if the data is acceptable. If there is a match, <code class="prettyprint">event.preventDefault()</code>
         * will be called to indicate that the data can be accepted.
         * @property {function(Event, {item: Element}):void} [items.dragOver] (optional) A callback function that receives the "dragover" event and context information as arguments:<br><br>
         * <code class="prettyprint">function(event, context)</code><br><br>
         * All of the event payloads listed below can be found under the <code class="prettyprint">context</code> argument.
         * <ul><li><code class="prettyprint">item</code>: The item being dragged over.</li></ul><br>
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.
         * Otherwise, dataTypes will be matched against the drag dataTypes to determine if the data is acceptable. If there is a match, <code class="prettyprint">event.preventDefault()</code>
         * will be called to indicate that the data can be accepted.
         * @property {function(Event, {item: Element}):void} [items.dragLeave] (optional) A callback function that receives the "dragleave" event and context information as arguments:<br><br>
         * <code class="prettyprint">function(event, context)</code><br><br>
         * All of the event payloads listed below can be found under the <code class="prettyprint">context</code> argument.
         * <ul><li><code class="prettyprint">item</code>: The item that was last entered.</li></ul><br>
         * @property {function(Event, oj.ojTreeView.ItemsDropOnDropContext):void} items.drop (required) A callback function that receives the "drop" event and context information as arguments:<br><br>
         * <code class="prettyprint">function(event, context)</code><br><br>
         * All of the event payloads listed below can be found under the <code class="prettyprint">context</code> argument.
         * <ul><li><code class="prettyprint">item</code>: The item being dropped on.</li>
         * <li><code class="prettyprint">position</code>: The drop position relative to the item being dropped on.
         * Valid values are "inside", "before", "after", and "first" (the first child of the item being dropped on).</li></ul><br>
         * This function should call <code class="prettyprint">event.preventDefault()</code> to indicate the dragged data can be accepted.<br><br>
         * If the application needs to look at the data for the item being dropped on, it can use the <code class="prettyprint">getContextByNode</code> method.
         *
         * @ojsignature [{target: "Type",
         *                value: "?((event: Event, context: {item: Element}) => void)",
         *                for: "items.dragEnter"},
         *               {target: "Type",
         *                value: "?((event: Event, context: {item: Element}) => void)",
         *                for: "items.dragOver"},
         *               {target: "Type",
         *                value: "?((event: Event, context: {item: Element}) => void)",
         *                for: "items.dragLeave"}]
         *
         *
         * @example <caption>Initialize the TreeView such that only leaf items are focusable:</caption>
         * myTreeView.setProperty('dnd.drop.items', {
         *   'dataTypes': ['application/ojtreeviewitems+json'],
         *   'drop': handleDrop
         * });
         */
        drop: null
      },

      /**
       * Specifies the key set containing the keys of the TreeView items that should be expanded.
       * Use the <a href="ExpandedKeySet.html">ExpandedKeySet</a> class to specify items to expand.
       * Use the <a href="ExpandAllKeySet.html">ExpandAllKeySet</a> class to expand all items.
       *
       * @expose
       * @memberof! oj.ojTreeView
       * @instance
       * @type {KeySet}
       * @default new ExpandedKeySet()
       * @ojwriteback
       * @ojsignature {target:"Type", value:"KeySet<K>"}
       *
       * @example <caption>Initialize the TreeView with some expanded items:</caption>
       * myTreeView.expanded = new keySet.ExpandedKeySet(['item1', 'item2']);
       */
      expanded: new oj.ExpandedKeySet(),

      /**
       * The item attribute contains a subset of attributes for items.
       *
       * @expose
       * @memberof! oj.ojTreeView
       * @instance
       */
      item: {
        // TODO: Add data provider to item context
        /**
         * @typedef {Object} oj.ojTreeView.ItemContext
         * @property {Element} componentElement The TreeView element.
         * @property {D} [data] The data object of the item (not available for static content).
         * @property {number} depth The depth of the item. The depth of the first level children under the invisible root is 1.
         * @property {number} index The index of the item relative to its parent, where 0 is the index of the first item.
         * @property {K} key The key of the item.
         * @property {boolean} leaf Whether the item is a leaf item.
         * @property {Element} parentElement The TreeView item element. The renderer can use this to directly append content.
         * @property {K} [parentKey] The key of the parent item (not available for root item).
         * @ojsignature {target:"Type", value:"<K,D>", for:"genericTypeParameters"}
         */
        /**
         * A function that returns whether the item is focusable.
         * A item that is not focusable cannot be clicked on or navigated to.
         * See <a href="#context-section">itemContext</a> in the introduction
         * to see the object passed into the focusable function.
         * If no function is specified, then all the items will be focusable.
         *
         * @expose
         * @alias item.focusable
         * @memberof! oj.ojTreeView
         * @instance
         * @type {function(Object):boolean|null}
         * @ojsignature {target: "Type",
         *               value: "?((itemContext: oj.ojTreeView.ItemContext<K,D>) => boolean)",
         *               jsdocOverride: true}
         * @default null
         *
         * @example <caption>Initialize the TreeView such that only leaf items are focusable:</caption>
         * myTreeView.setProperty('item.focusable', function(itemContext)
         * {
         *   return itemContext['leaf'];
         * });
         */
        focusable: null,

        /**
         * The renderer function that renders the contents of the item. See <a href="#context-section">itemContext</a>
         * in the introduction to see the object passed into the renderer function.
         * The function should return one of the following: 
         * <ul>
         *   <li>An Object with the following property:
         *     <ul><li>insert: HTMLElement | string - A string or a DOM element of the content inside the item.</li></ul>
         *   </li>
         *   <li>Nothing: If the developer chooses to manipulate the item element directly, the function should return nothing.</li>
         * </ul>
         *
         * @expose
         * @alias item.renderer
         * @memberof! oj.ojTreeView
         * @instance
         * @type {function(Object):Object|null}
         * @ojsignature {target: "Type",
         *               value: "?((itemContext: oj.ojTreeView.ItemContext<K,D>) => {insert: Element|string}|void)|null",
         *               jsdocOverride: true}
         * @default null
         *
         * @example <caption>Initialize the TreeView with a renderer:</caption>
         * myTreeView.setProperty('item.renderer', function(itemContext)
         * {
         *   return itemContext['data'].get('FIRST_NAME');
         * });
         */
        renderer: null,

        /**
         * A function that returns whether the item can be selected.
         * If selectionMode is set to "none" this attribute is ignored.
         * See <a href="#context-section">itemContext</a> in the introduction
         * to see the object passed into the selectable function. 
         * If no function is specified, then all the items will be selectable.
         *
         * @expose
         * @alias item.selectable
         * @memberof! oj.ojTreeView
         * @instance
         * @type {function(Object):boolean|null}
         * @ojsignature {target: "Type",
         *               value: "?((itemContext: oj.ojTreeView.ItemContext<K,D>) => boolean)",
         *               jsdocOverride: true}
         * @default null
         *
         * @example <caption>Initialize the TreeView such that only leaf items are selectable:</caption>
         * myTreeView.setProperty('item.selectable', function(itemContext)
         * {
         *   return itemContext['leaf'];
         * });
         */
        selectable: null
      },

      /**
       * The current selections in the TreeView. An empty array indicates nothing is selected.
       *
       * @expose
       * @memberof! oj.ojTreeView
       * @instance
       * @type {Array.<*>}
       * @default []
       * @ojwriteback
       * @ojsignature {target:"Type", value:"Array<K>"}
       *
       * @example <caption>Initialize the TreeView with specific selection:</caption>
       * myTreeView.selection = ['item1', 'item2'];
       */
      selection: [],

      /**
       * Specifies whether selection can be made and the cardinality of selection in the TreeView.
       *
       * @expose
       * @memberof! oj.ojTreeView
       * @instance
       * @type {string}
       * @default "none"
       * @ojvalue {string} "none" Selection is disabled.
       * @ojvalue {string} "single" Only one item can be selected at a time.
       * @ojvalue {string} "multiple" Multiple items can be selected at the same time.
       *
       * @example <caption>Initialize the tree view to enable multiple selection:</caption>
       * myTreeView.selectionMode = 'multiple';
       */
      selectionMode: 'none',

      // Events

      /**
       * Triggered when the default animation of a particular action has ended.
       * Note this event will not be triggered if application cancelled the default animation on animateStart.
       *
       * @expose
       * @event
       * @memberof oj.ojTreeView
       * @instance
       * @property {'expand'|'collapse'} action The action that started the animation. See <a href="#animation-section">animation</a> section for a list of actions.
       * @property {Element} element The target of animation.
       */
      animateEnd: null,

      /**
       * Triggered when the default animation of a particular action is about to start.
       * The default animation can be cancelled by calling <code class="prettyprint">event.preventDefault()</code>.
       *
       * @expose
       * @event
       * @memberof oj.ojTreeView
       * @instance
       * @property {'expand'|'collapse'} action The action that starts the animation. See <a href="#animation-section">animation</a> section for a list of actions.
       * @property {Element} element The target of animation.
       * @property {function():void} endCallback If the event listener calls <code class="prettyprint">event.preventDefault()</code> to cancel the default animation, it must call the endCallback function when it finishes its own animation handling and when any custom animation ends.
       */
      animateStart: null,

      /**
       * Triggered before an item is collapsed via the <code class="prettyprint">expanded</code> attribute or via the UI.
       * Call <code class="prettyprint">event.preventDefault()</code> to veto the event, which prevents collapsing the item.
       *
       * @expose
       * @event
       * @ojcancelable
       * @memberof oj.ojTreeView
       * @instance
       * @property {*} key The key of the item to be collapsed.
       * @property {Element} item The item to be collapsed.
       */
      beforeCollapse: null,

      /**
       * Triggered before the current item is changed via the <code class="prettyprint">currentItem</code> attribute or via the UI.
       * Call <code class="prettyprint">event.preventDefault()</code> to veto the event, which prevents changing the current item.
       *
       * @expose
       * @event
       * @ojcancelable
       * @memberof oj.ojTreeView
       * @instance
       * @property {*} previousKey The key of the previous item.
       * @property {Element} previousItem The previous item.
       * @property {*} key The key of the new current item.
       * @property {Element} item The new current item.
       */
      beforeCurrentItem: null,

      /**
       * Triggered before an item is expanded via the <code class="prettyprint">expanded</code> attribute or via the UI.
       * Call <code class="prettyprint">event.preventDefault()</code> to veto the event, which prevents expanding the item.
       *
       * @expose
       * @event
       * @ojcancelable
       * @memberof oj.ojTreeView
       * @instance
       * @property {*} key The key of the item to be expanded.
       * @property {Element} item The item to be expanded.
       */
      beforeExpand: null,

      /**
       * Triggered after an item has been collapsed.
       *
       * @expose
       * @event
       * @memberof oj.ojTreeView
       * @instance
       * @property {*} key The key of the item that was just collapsed.
       * @property {Element} item The item that was just collapsed.
       */
      collapse: null,

      /**
       * Triggered after an item has been expanded.
       *
       * @expose
       * @event
       * @memberof oj.ojTreeView
       * @instance
       * @property {*} key The key of the item that was just expanded.
       * @property {Element} item The item that was just expanded.
       */
      expand: null
    },

    // @inheritdoc
    _ComponentCreate: function()
    {
      this._super();
    },

    // @inheritdoc
    _AfterCreate: function()
    {
      this._super();
      this._initRender();
      this._render();
    },

    /**
     * Initializes the TreeView.
     * @private
     */
    _initRender: function()
    {
      var self = this;

      // Event listeners
      this._on(this.element, {
        'click': function(event)
        {
          self._handleClick(event);
        },
        'mouseover': function(event)
        {
          self._handleMouseOver(event);
        },
        'mouseout': function(event)
        {
          self._handleMouseOut(event);
        },
        'mousedown': function(event)
        {
          self._handleMouseDown(event);
        },
        'mouseup': function(event)
        {
          self._handleMouseUp(event);
        },
        'keydown': function(event)
        {
          self._handleKeyDown(event);
        },
        'dragstart': function(event)
        {
          self._handleDragStart(event);
        },
        'drag': function(event)
        {
          self._handleDragSourceEvent(event, 'drag');
        },
        'dragend': function(event)
        {
          self._handleDragSourceEvent(event, 'dragEnd');
        },
        'dragenter': function(event)
        {
          self._handleDropTargetEvent(event, 'dragEnter');
        },
        'dragover': function(event)
        {
          self._handleDropTargetEvent(event, 'dragOver');
        },
        'dragleave': function(event)
        {
          self._handleDropTargetEvent(event, 'dragLeave');
        },
        'drop': function(event)
        {
          self._handleDropTargetEvent(event, 'drop');
        }
      });

      // Drop marker
      var dropMarkerIcon = $(document.createElement('span'))
        .addClass('oj-treeview-drop-marker-icon oj-component-icon');

      this._dropMarker = $(document.createElement('div'))
        .addClass('oj-treeview-drop-marker')
        .append(dropMarkerIcon) // @HTMLUpdateOK
        .appendTo(this.element); // @HTMLUpdateOK

      this._dropMarkerRect = this._dropMarker[0].getBoundingClientRect();
      this._dropMarker.hide();

      this._dropLine = $(document.createElement('div'))
        .addClass('oj-treeview-drop-line')
        .appendTo(this.element); // @HTMLUpdateOK

      this._dropLineRect = this._dropLine[0].getBoundingClientRect();
      this._dropLine.hide();
    },

    /**
     * Renders the TreeView.
     * @private
     */
    _render: function()
    {
      var self = this;
      this.element.removeClass('oj-complete');
      this._keyList = new Set(); // list of existing node keys

      if (this.options['data'])
      {
        this.element.find('ul').remove();
        this._fetchChildren(null, function(nodeSet)
        {
          self._renderItems(nodeSet, self.element);
          self._resetFocus();
          self.element.addClass('oj-complete');
        });
      }
      else
      {
        this.element.find('ul').addClass('oj-treeview-list').attr('role', 'group');
        this.element.find('li').each(function()
        {
          var item = $(this);
          self._decorateItem(item);
        });
        this._resetFocus();
        this.element.addClass('oj-complete');
      }

      this._decorateTree();
      this._lastSelectedItem = null;
    },

    /**
     * Fetch the children of a parent item from the data source.
     * @param {string} parentKey The key of the parent item.
     * @param {Function} successFunc The function to be called if the fetch is successful.
     * @private
     */
    _fetchChildren: function(parentKey, successFunc)
    {
      var self = this;
      var dataSource = self.options['data'];
      var range = {'start': 0, 'count': dataSource.getChildCount(parentKey)};
      var busyResolve = this._addBusyState('fetching data');

      dataSource.fetchChildren(parentKey, range, {
        'success': function(nodeSet)
        {
          successFunc(nodeSet);
          busyResolve();
        },
        'error': function(status)
        {
          oj.Logger.error(status);
          busyResolve();
        }
      });
    },

    /**
     * Render the TreeView items after the data is fetched.
     * @param {oj.JsonNodeSet|oj.CollectionNodeSet} nodeSet The array of item data returned by the data source.
     * @param {jQuery} parentElem The parent element to append the items to.
     * @private
     */
    _renderItems: function(nodeSet, parentElem)
    {
      var ulElem = $(document.createElement('ul'))
        .addClass('oj-treeview-list')
        .attr('role', 'group')
        .appendTo(parentElem); // @HTMLUpdateOK

      for (var i = 0; i < nodeSet.getCount(); i++)
      {
        this._renderItem(ulElem, nodeSet, i);
      }
    },

    /**
     * Render a TreeView item after the data is fetched.
     * @param {jQuery} ulElem The <ul> to attach the item to.
     * @param {oj.JsonNodeSet|oj.CollectionNodeSet} nodeSet The array of item data returned by the data source.
     * @param {number} index The index of the item.
     * @private
     */
    _renderItem: function(ulElem, nodeSet, index)
    {
      var self = this;

      index += nodeSet.getStart();
      var data = nodeSet.getData(index);
      var metadata = nodeSet.getMetadata(index);
      var key = metadata['key'];

      // Prevent infinite recursion due to duplicated keys ()
      if (this._keyList.has(key)) {
        throw new Error("JET TreeView nodes should not have duplicated keys: " + key);
      }
      this._keyList.add(key);

      var renderer = self.options['item']['renderer'];
      renderer = self._WrapCustomElementRenderer(renderer);

      var liElem = $(document.createElement('li'))
        .attr('id', key)
        .appendTo(ulElem); // @HTMLUpdateOK

      if (renderer)
      {
        // TODO: store index somewhere to return in the context

        var context = {};
        context['parentElement'] = liElem;
        context['index'] = index;
        context['data'] = data;
        context['datasource'] = self.options['data'];
        context['parentKey'] = self._getKey(self._getParentItem(liElem));

        context['component'] = oj.Components.__GetWidgetConstructor(self.element);
        if (self._FixRendererContext)
          context = self._FixRendererContext(context);

        // Merge properties from metadata into item context
        for (var prop in metadata) 
        {
          if (metadata.hasOwnProperty(prop))
            context[prop] = metadata[prop];
        }

        var content = renderer.call(self, context);
        if (content != null)
        {
          // Allow return of document fragment from jQuery create or JS document.createDocumentFragment
          if (content.parentNode === null || content.parentNode instanceof DocumentFragment)
          {
            liElem.append(content); // @HTMLUpdateOK
          }
          else if (content.parentNode != null)
          {
            // Parent node exists, do nothing
          }
          else if (content.toString)
          {
            var textWrapper = document.createElement('span');
            textWrapper.appendChild(document.createTextNode(content.toString())); // @HTMLUpdateOK
            liElem.append(textWrapper); // @HTMLUpdateOK
          }
        }

        // Get the item from root again as template replaces the item element
        liElem = ulElem.children().eq(index);
        context['parentElement'] = liElem;

        // Set id on the liElem if not set by the renderer
        if (liElem.attr('id') == null)
          liElem.attr('id', key);
      }

      // Cache data and metadata for lookup
      liElem.data('data', data).data('metadata', metadata)

      self._decorateItem(liElem);
    },

    /**
     * Adds the necessary attributes to a TreeView root element.
     * @private
     */
    _decorateTree: function()
    {
      var self = this;

      // Keyboard focus and ARIA attributes
      var root = this._getRoot();
      this._focusable({
        'element': root,
        'applyHighlight': true,
        'setupHandlers': function(focusInHandler, focusOutHandler)
        {
          self._focusInHandler = focusInHandler;
          self._focusOutHandler = focusOutHandler;
        }
      });

      root.attr('tabIndex', 0)
        .on('focus', function(event)
        {
          self._focusInHandler(self._getItemContent(self._currentItem));
        })
        .on('blur',function(event)
        {
          self._focusOutHandler(self._getItemContent(self._currentItem));
        })
        .attr('role', 'tree')
        .attr('aria-labelledby', this.element.attr('id'));

      var selectionMode = this.options['selectionMode'];
      if (selectionMode != 'none')
        root.attr('aria-multiselectable', selectionMode == 'multiple' ? 'true' : 'false');
      else
        root.removeAttr('aria-multiselectable');
    },

    /**
     * Adds the necessary attributes to a TreeView item element.
     * @param {jQuery} item The item element to decorate.
     * @private
     */
    _decorateItem: function(item)
    {
      var self = this;
      item.addClass('oj-treeview-item').attr('role', 'treeitem');

      // Create wrapper for item icon and text
      var itemContent = this._getItemContent(item);
      if (itemContent.length == 0)
      {
        // Wrap everything except the subtree
        // Use item.contents() to include text and comment nodes as well
        itemContent = $(document.createElement('div'))
          .addClass('oj-treeview-item-content')
          .append(item.contents()) // @HTMLUpdateOK
          .appendTo(item); // @HTMLUpdateOK

        // Take the subtree out of the itemContent wrapper
        // Selector can't be used to exclude the ul when wrapping item.contents()
        itemContent.children('ul').appendTo(item);

        itemContent.find('.oj-treeview-item-icon').addClass('oj-treeview-icon');
      }

      // Initial selection
      self._select(item);

      // DnD
      var dragOptions = this._getDragOptions();
      itemContent.attr('draggable', Object.keys(dragOptions).length > 0 ? 'true' : 'false');

      // Create disclosure icon or spacer
      var disclosureIcon = item.children('.oj-treeview-spacer');
      if (disclosureIcon.length == 0)
      {
        disclosureIcon = $(document.createElement('ins'))
          .addClass('oj-treeview-icon oj-treeview-spacer')
          .prependTo(item); // @HTMLUpdateOK
      }

      if (self._isLeaf(item))
      {
        item.addClass('oj-treeview-leaf');
      }
      else
      {
        // Expanded option
        if (self._isInitExpanded(item))
          self._expand(item, false);
        else 
          self._collapse(item, false);

        disclosureIcon.addClass('oj-treeview-disclosure-icon oj-component-icon oj-clickable-icon-nocontext oj-default');
      }
    },

    /**
     * Returns all the item elements that belongs to the TreeView.
     * @return {jQuery} All the item elements that belongs to the TreeView.
     * @private
     */
    _getItems: function()
    {
      return this.element.find('.oj-treeview-item');
    },

    /**
     * Returns the key of the provided item.
     * @param {jQuery} item The TreeView item element.
     * @return {jQuery|string|undefined} The key.
     * @private
     */
    _getKey: function(item)
    {
      // Rely on the key in the metadata first. The DOM id stringifies the key,
      // so it won't return the correct key if the key isn't string.
      var metadata = item.data('metadata');
      if (metadata)
        return metadata['key'];
      else
        return item.attr('id');
    },

    /**
     * Returns the item element that is identified by the provided key.
     * @param {string} key The key.
     * @return {jQuery} The item element.
     * @private
     */
    _getItemByKey: function(key)
    {
      return this.element.find('li#' + key);
    },

    /**
     * Returns the content element of the provided item element.
     * @param {jQuery} item The TreeView item element.
     * @return {jQuery} The item content element.
     * @private
     */
    _getItemContent: function(item)
    {
      return item.children('.oj-treeview-item-content');
    },

    /**
     * Returns the child items of the provided item element.
     * @param {jQuery} item The TreeView item element.
     * @return {jQuery} The child item elements.
     * @private
     */
    _getChildItems: function(item)
    {
      return item.children('.oj-treeview-list').children('.oj-treeview-item');
    },

    /**
     * Returns the parent item of the provided item element.
     * @param {jQuery} item The TreeView item element.
     * @return {jQuery} The parent item element.
     * @private
     */
    _getParentItem: function(item)
    {
      return item.parent('.oj-treeview-list').parent('.oj-treeview-item');
    },

    /**
     * Returns the subtree element (ul) of the provided item element.
     * @param {jQuery} item The TreeView item element.
     * @return {jQuery} The subtree element (ul).
     * @private
     */
    _getSubtree: function(item)
    {
      return item.children('.oj-treeview-list');
    },

    /**
     * Returns the root ul of the TreeView.
     * @return {jQuery}
     * @private
     */
    _getRoot: function()
    {
      return this.element.children('.oj-treeview-list');
    },

    /**
     * Returns whether the item is a leaf item.
     * @param {jQuery} item The TreeView item element.
     * @return {boolean}
     * @private
     */
    _isLeaf: function(item)
    {
      var metadata = item.data('metadata');
      var hasChildren = metadata ? !metadata['leaf'] : this._getSubtree(item).length > 0;
      return !hasChildren;
    },

    /**
     * Returns whether item element is initially expanded in the TreeView option.
     * Note that the expanded option is not kept in sync with the actual expanded state of the TreeView.
     * @param {jQuery} item The TreeView item element.
     * @return {boolean}
     * @private
     */
    _isInitExpanded: function(item)
    {
      var key = this._getKey(item);
      var expanded = this.options['expanded'];
      return (expanded && expanded.has) ? expanded.has(key) : false;
    },

    /**
     * Returns whether item element is currently expanded.
     * Note that the expanded option is not kept in sync with the actual expanded state of the TreeView.
     * @param {jQuery} item The TreeView item element.
     * @return {boolean}
     * @private
     */
    _isExpanded: function(item)
    {
      return item.hasClass('oj-expanded');
    },

    /**
     * Expands an item.
     * @param {jQuery} item The TreeView item element.
     * @param {boolean} animate Whether the expand should be animated.
     * @param {Event} event The event that triggers the expand.
     * @param {boolean} vetoable Whether the expand can be vetoed by beforeExpand.
     * @private
     */
    _expand: function(item, animate, event, vetoable)
    {
      var self = this;

      if (this._isExpanded(item) || this._isLeaf(item))
        return;

      if (animate)
      {
        var cancelled = !this._trigger('beforeExpand', event, this._getEventPayload(item));
        if (cancelled && vetoable !== false)
          return;
      }

      this._lastSelectedItem = null;

      var subtree = this._getSubtree(item);
      if (subtree.length == 0)
      {
        return self._fetchChildren(self._getKey(item), function(nodeSet)
        {
          self._renderItems(nodeSet, item);
          self._expandAfterFetch(item, animate, event);
        });
      }
      else
      {
        self._expandAfterFetch(item, animate, event);
      }
    },

    /**
     * Expands an item after its child items have been fetched.
     * @param {jQuery} item The TreeView item element.
     * @param {boolean} animate Whether the expand should be animated.
     * @param {Event} event The event that triggers the expand.
     * @private
     */
    _expandAfterFetch: function(item, animate, event)
    {
      var self = this;

      item.removeClass('oj-collapsed').addClass('oj-expanded').attr('aria-expanded', 'true');

      var subtree = this._getSubtree(item);
      subtree.css('max-height', 'none');

      if (animate)
      {
        var busyResolve = this._addBusyState('animating expand');
        item.addClass('oj-treeview-animated'); // animation flag

        this._startAnimation(subtree, 'expand').then(function()
        {
          item.removeClass('oj-treeview-animated');
          self._trigger('expand', event, self._getEventPayload(item));

          // Update option and fire optionChange
          var expanded = self.options['expanded'];
          var newExpanded = expanded.add([self._getKey(item)]);
          self._userOptionChange('expanded', newExpanded, event);

          busyResolve();
        });
      }
    },

    /**
     * Collapses an item.
     * @param {jQuery} item The TreeView item element.
     * @param {boolean} animate Whether the collapse should be animated.
     * @param {Event} event The event that triggers the collapse.
     * @param {boolean} vetoable Whether the collapse can be vetoed by beforeCollapse.
     * @private
     */
    _collapse: function(item, animate, event, vetoable)
    {
      var self = this;

      if (item.hasClass('oj-collapsed') || this._isLeaf(item))
        return;

      if (animate)
      {
        var cancelled = !this._trigger('beforeCollapse', event, this._getEventPayload(item));
        if (cancelled  && vetoable !== false)
          return;
      }

      item.removeClass('oj-expanded').addClass('oj-collapsed').attr('aria-expanded', 'false');

      this._lastSelectedItem = null;

      var subtree = this._getSubtree(item);
      if (animate)
      {
        var busyResolve = this._addBusyState('animating collapse');
        item.addClass('oj-treeview-animated'); // animation flag

        this._startAnimation(subtree, 'collapse').then(function()
        {
          subtree.css('max-height', 0);
          item.removeClass('oj-treeview-animated');
          self._trigger('collapse', event, self._getEventPayload(item));

          // Update option and fire optionChange
          var expanded = self.options['expanded'];
          var newExpanded = expanded.delete([self._getKey(item)]);
          self._userOptionChange('expanded', newExpanded, event);

          busyResolve();
        });
      }
      else
      {
        subtree.css('max-height', 0);
      }
    },

    /**
     * Starts the animation for the specific action.
     * @param {jQuery} elem The element to animate.
     * @param {string} action The name of the action to animate.
     * @return {Promise} A promise that will be resolved when the animation ends.
     * @private
     */
    _startAnimation: function(elem, action)
    {
      if (!this.defaultOptions)
        this.defaultOptions = oj.ThemeUtils.parseJSONFromFontFamily('oj-treeview-option-defaults');
  
      var effects = (this.defaultOptions['animation'] || {})[action];
      return oj.AnimationUtils.startAnimation(elem[0], action, effects, this);
    },

    /**
     * Returns the default event payload for the provided item element.
     * @param {jQuery} item The TreeView item element.
     * @return {Object}
     * @private
     */
    _getEventPayload: function(item)
    {
      return {'key': this._getKey(item), 'item': item[0]};
    },

    /**
     * Returns whether an action (select or focus) can be performed on the item.
     * @param {jQuery} item The TreeView item element.
     * @param {string} actionName The action name: 'select' or 'focus'.
     * @return {boolean}
     * @private
     */
    _isActionable: function(item, actionName)
    {
      var actionable = this.options['item'][actionName + 'able'];
      if (actionable === false)
        return false;
      else if (typeof actionable === 'function')
      {
        var itemContext = this.getContextByNode(item[0]);
        return actionable(itemContext);
      }
      else
        return true;
    },

    /**
     * Returns whether item element is selected.
     * @param {jQuery} item The TreeView item element.
     * @return {boolean}
     * @private
     */
    _isSelected: function(item)
    {
      var selectionMode = this.options['selectionMode'];
      var selection = this.options['selection'];

      if (selectionMode === 'none')
        return false;
      else if (selectionMode === 'single' && selection.length > 1)
        selection = [selection[0]];

      var key = this._getKey(item);
      return selection.indexOf(key) != -1;
    },

    /**
     * Selects or unselects an item, depending on what triggers the selection.
     * @param {jQuery} item The TreeView item element.
     * @param {Event} event The event that triggers the select.
     * @private
     */
    _select: function(item, event)
    {
      var selectionMode = this.options['selectionMode'];
      if (selectionMode === 'none')
        return;

      // Check whether the item is selectable
      if (!this._isActionable(item, 'select'))
        return;

      var isSelected = this._isSelected(item);

      if (event)
      {
        var isTouch = oj.DomUtils.isTouchSupported();
        var isMetaKey = oj.DomUtils.isMetaKeyPressed(event);
        var isMultiple = selectionMode === 'multiple';
        var isNavigation = event.keyCode == 40 || event.keyCode == 38;
        var key = this._getKey(item);
        var selection = [];

        if (isMultiple && event.shiftKey && !isNavigation)
        {
          // Maintain selection of other items if meta key is pressed
          if (isMetaKey)
            selection = this.options['selection'].slice(0);
          else
            this._clearSelection();

          // Select a range from the last selected item to the current item
          var nextItem = this._lastSelectedItem;
          var getNextItem = (nextItem && nextItem.offset().top < item.offset().top) ?
            this._getNextActionableItem.bind(this) : this._getPreviousActionableItem.bind(this);

          while (nextItem && (nextItem[0] != item[0]))
          {
            var nextKey = this._getKey(nextItem);
            if (selection.indexOf(nextKey) == -1)
            {
              selection.push(nextKey);
              this._setSelected(nextItem);
            }
            nextItem = getNextItem(nextItem, 'select');
          }

          // Select the current item
          isSelected = true;
          selection.push(key);
        }
        else if (isMultiple && (isMetaKey || isTouch || isNavigation))
        {
          // Toggle selection of current item while maintaining the selection of the other items
          isSelected = !isSelected;
          selection = this.options['selection'].slice(0);
          if (isSelected)
            selection.push(key);
          else
            selection.splice(selection.indexOf(key), 1);
        }
        else
        {
          // Clear selection of all other items
          this._clearSelection();

          // On touch or spacebar, toggle the selection of the current item
          // Otherwise, select the current item even if it is already selected
          if ((isTouch || event.keyCode == 32) && isSelected)
          {
            isSelected = false;
            selection = [];
          }
          else
          {
            isSelected = true;
            selection = [key];
          }
        }

        // Update option and fire optionChange
        this._userOptionChange('selection', selection, event);
        this._lastSelectedItem = item;
      }

      if (isSelected)
        this._setSelected(item);
      else
        this._setUnselected(item);
    },

    /**
     * Style the provided item as selected.
     * @param {jQuery} item The TreeView item element.
     * @private
     */
    _setSelected: function(item)
    {
      this._getItemContent(item).addClass('oj-selected');
      item.attr('aria-selected', 'true');
    },

    /**
     * Style the provided item as unselected.
     * @param {jQuery} item The TreeView item element.
     * @private
     */
    _setUnselected: function(item)
    {
      this._getItemContent(item).removeClass('oj-selected');
      item.attr('aria-selected', 'false');
    },

    /**
     * Clears the selection of all items.
     * @private
     */
    _clearSelection: function()
    {
      this._setUnselected(this._getItems());
    },

    /**
     * Sets the focus (current item) on the provided item element.
     * @param {jQuery} item The TreeView item element.
     * @param {Event} event The event that triggers the focus.
     * @private
     */
    _focus: function(item, event)
    {
      // Check whether the item is focusable
      if (!this._isActionable(item, 'focus'))
        return;

      if (event)
      {
        var payload = this._getEventPayload(item);
        if (this._currentItem)
        {
          payload['previousKey'] = this._getKey(this._currentItem);
          payload['previousItem'] = this._currentItem[0];
        }

        var cancelled = !this._trigger('beforeCurrentItem', event, payload);
        if (cancelled)
          return;

        // Update option and fire optionChange
        this._userOptionChange('currentItem', this._getKey(item), event);
      }

      this._focusOutHandler(this._getItemContent(this._currentItem));
      this._focusInHandler(this._getItemContent(item));
      this._setCurrentItem(item);
    },

    /**
     * Resets the focus (current item) of the TreeView.
     * @private
     */
    _resetFocus: function()
    {
      if (this.options['currentItem'])
      {
        var currentItem = this._getItemByKey(this.options['currentItem']);
        if (currentItem.length > 0)
        {
          this._setCurrentItem(currentItem);
          return;
        }
      }

      // CurrentItem not specified, so default to the first item.
      // Update the option so the currentItem attribute is in sync.
      var firstItem = this._getItems().first();
      this._setCurrentItem(firstItem);
      this._userOptionChange('currentItem', this._getKey(this._currentItem), null);
    },

    /**
     * Set the provided item as the current item.
     * @param {jQuery} item The TreeView item element.
     * @private
     */
    _setCurrentItem: function(item)
    {
      this._currentItem = item;

      // Set the item content to be the activedescendant so that the screen reader
      // does not read the child items.
      this._getRoot().attr('aria-activedescendant', this._getKey(item));
    },

    /**
     * Adds a busy state.
     * @param {string} description The description of the busy state.
     * @return {Function} The resolve function of the busy state.
     * @private
     */
    _addBusyState: function(description)
    {
      var busyContext = oj.Context.getContext(this.element[0]).getBusyContext();
      var id = this.element.attr('id');
      return busyContext.addBusyState({
        'description': "The component identified by '" + id + "', " + description
      });
    },

    /**
     * Writes back a user-triggered change into the option.
     * @param {string} key The option name.
     * @param {Object} value The new value of the option.
     * @param {Event} event The event that triggers the change.
     * @private
     */
    _userOptionChange: function(key, value, event)
    {
      this.option(key, value, {'_context': {originalEvent: event, writeback:true, internalSet: true}});
    },

    /**
     * Returns the dnd.drag.items option.
     * @return {Object} The option object. Defaults to {}.
     * @private
     */
    _getDragOptions: function()
    {
      return ((this.options['dnd'] || {})['drag'] || {})['items'] || {};
    },

    /**
     * Returns the dnd.drop.items option.
     * @return {Object} The option object. Defaults to {}.
     * @private
     */
    _getDropOptions: function()
    {
      return ((this.options['dnd'] || {})['drop'] || {})['items'] || {};
    },

    /**
     * Returns the closest item to the element.
     * @param {jQuery} elem The element.
     * @return {jQuery} The item content element.
     * @private
     */
    _getClosestItem: function(elem)
    {
      return elem.closest('.oj-treeview-item');
    },

    /**
     * Returns the closest item content to the element.
     * @param {jQuery} elem The element.
     * @return {jQuery} The item content element.
     * @private
     */
    _getClosestItemContent: function(elem)
    {
      return elem.closest('.oj-treeview-item-content');
    },

    /**
     * Returns the closest disclosure icon to the element.
     * @param {jQuery} elem The element.
     * @return {jQuery} The disclosure icon element.
     * @private
     */
    _getClosestDisclosureIcon: function(elem)
    {
      return elem.closest('.oj-treeview-disclosure-icon');
    },

    /**
     * Handles click event.
     * @param {Event} event The event.
     * @private
     */
    _handleClick: function(event)
    {
      // Clicking on disclosure icon
      var disclosureIcon = this._getClosestDisclosureIcon($(event.target));
      if (disclosureIcon.length > 0)
      {
        var item = disclosureIcon.closest('.oj-treeview-item');
        if (this._isExpanded(item))
          this._collapse(item, true, event);
        else
          this._expand(item, true, event);
        return;
      }

      // Clicking on item content
      var itemContent = this._getClosestItemContent($(event.target));
      if (itemContent.length > 0)
      {
        var item = itemContent.parent();
        this._select(item, event);
        this._focus(item, event);
        return;
      }

      // Clear selection otherwise
      var selectionMode = this.options['selectionMode'];
      if (selectionMode != 'none')
      {
        this._clearSelection();
        this._lastSelectedItem = null;
        this._userOptionChange('selection', [], event);
      }
    },

    /**
     * Handles mouse over event.
     * @param {Event} event The event.
     * @private
     */
    _handleMouseOver: function(event)
    {
      // Don't add hover effect for touch
      if (oj.DomUtils.isTouchSupported())
        return;

      var target = this._getClosestDisclosureIcon($(event.target));

      if (target.length == 0)
        target = this._getClosestItemContent($(event.target));

      // Add hover effect
      target.removeClass('oj-default');
      target.addClass('oj-hover');
    },

    /**
     * Handles mouse out event.
     * @param {Event} event The event.
     * @private
     */
    _handleMouseOut: function(event)
    {
      // Don't add hover effect for touch
      if (oj.DomUtils.isTouchSupported())
        return;

      var target = this._getClosestDisclosureIcon($(event.target));
      target.removeClass('oj-selected');

      if (target.length == 0)
        target = this._getClosestItemContent($(event.target));

      // Remove hover effect
      target.addClass('oj-default');
      target.removeClass('oj-hover');
    },
    
    /**
     * Handles mouse down event.
     * @param {Event} event The event.
     * @private
     */
    _handleMouseDown: function(event)
    {
      var disclosureIcon = this._getClosestDisclosureIcon($(event.target));
      disclosureIcon.addClass('oj-selected');
    },

    /**
     * Handles mouse up event.
     * @param {Event} event The event.
     * @private
     */
    _handleMouseUp: function(event)
    {
      var disclosureIcon = this._getClosestDisclosureIcon($(event.target));
      disclosureIcon.removeClass('oj-selected');
    },

    /**
     * Handles key down event.
     * @param {Event} event The event.
     * @private
     */
    _handleKeyDown: function(event)
    {
      var keyCode = event.keyCode;
      var currentItem = this._currentItem;
      var nextItem;

      if (keyCode == 38 || keyCode == 40) // UP or DOWN
      {
        nextItem = (keyCode == 40) ? this._getNextActionableItem(currentItem, 'focus') :
                                     this._getPreviousActionableItem(currentItem, 'focus');
        if (nextItem)
        {
          event.preventDefault(); // prevent scrolling the page
          if (this._isSelected(currentItem) && event.shiftKey)
          {
            // Shift+Up/Down either extends the selection to the next item or cancels previous Shift+Down/Up
            this._select(this._isSelected(nextItem) ? currentItem : nextItem, event);
          }
          this._focus(nextItem, event);
        }
      }
      else if (keyCode == 37 || keyCode == 39) // LEFT or RIGHT
      {
        var isRTL = this._GetReadingDirection() == "rtl";
        var isEnd = (!isRTL && keyCode == 39) || (isRTL && keyCode == 37);

        if (isEnd && !this._isLeaf(currentItem) && !this._isExpanded(currentItem))
        {
          event.preventDefault(); // prevent scrolling the page
          this._expand(currentItem, true, event);
        }
        else if (!isEnd && !this._isLeaf(currentItem) && this._isExpanded(currentItem))
        {
          event.preventDefault(); // prevent scrolling the page
          this._collapse(currentItem, true, event);
        }
        else
        {
          nextItem = isEnd ? this._getNextActionableItem(currentItem, 'focus') :
                             this._getPreviousActionableItem(currentItem, 'focus');
          if (nextItem)
          {
            event.preventDefault(); // prevent scrolling the page
            this._focus(nextItem, event);
          }
        }
      }
      else if (keyCode == 13 || keyCode == 32) // ENTER or SPACE
      {
        event.preventDefault(); // prevent scrolling the page
        this._select(currentItem, event);
      }
    },

    /**
     * Returns the item below the provided item.
     * @param {jQuery} item The TreeView item element.
     * @return {jQuery} The next item element.
     * @private
     */
    _getNextItem: function(item)
    {
      // If the item is expanded, go to the first child
      if (!this._isLeaf(item) && this._isExpanded(item))
      {
        var firstChild = this._getChildItems(item).first();
        if (firstChild.length > 0)
          return firstChild;
      }

      // Otherwise, go to the next sibling of the item or its ancestors
      var parent = item;
      while (parent.length > 0)
      {
        var nextSibling = parent.next('.oj-treeview-item');
        if (nextSibling.length > 0)
          return nextSibling;

        parent = this._getParentItem(parent);
      }

      // Otherwise, don't go anywhere
      return null;
    },

    /**
     * Returns the closest item below the provided item that can accept the specified action.
     * @param {jQuery} item The TreeView item element.
     * @param {string} actionName The action name: 'select' or 'focus'.
     * @return {jQuery} The next item element.
     * @private
     */
    _getNextActionableItem: function(item, actionName)
    {
      while (item != null)
      {
        item = this._getNextItem(item);
        if (item != null && this._isActionable(item, actionName))
          return item;
      }
      return null;
    },

    /**
     * Returns the item above the provided item.
     * @param {jQuery} item The TreeView item element.
     * @return {jQuery} The previous item element.
     * @private
     */
    _getPreviousItem: function(item)
    {
      // Go to the last expanded child of the previous sibling
      var prevSibling = item.prev('.oj-treeview-item');
      while (prevSibling.length > 0)
      {
        if (!this._isLeaf(prevSibling) && this._isExpanded(prevSibling))
        {
          var lastChild = this._getChildItems(prevSibling).last();
          if (lastChild.length > 0)
            prevSibling = lastChild;
          else
            return prevSibling;
        }
        else
          return prevSibling;
      }

      // Otherwise, go to the parent
      var parent = this._getParentItem(item);
      if (parent.length > 0)
        return parent;

      // Otherwise, don't go anywhere
      return null;
    },

    /**
     * Returns the closest item above the provided item that can accept the specified action.
     * @param {jQuery} item The TreeView item element.
     * @param {string} actionName The action name: 'select' or 'focus'.
     * @return {jQuery} The previous item element.
     * @private
     */
    _getPreviousActionableItem: function(item, actionName)
    {
      while (item != null)
      {
        item = this._getPreviousItem(item);
        if (item != null && this._isActionable(item, actionName))
          return item;
      }
      return null;
    },

    /**
     * Handles DnD dragStart event.
     * @param {Event} event The event.
     * @private
     */
    _handleDragStart: function(event)
    {
      var self = this;
      var isRTL = this._GetReadingDirection() == "rtl";

      var targetItem = this._getClosestItem($(event.target));
      if (targetItem.length == 0)
        return;

      var items = $();
      if (!this._isSelected(targetItem))
      {
        this._select(targetItem, event);
        items = items.add(targetItem);
      }
      else
      {
        var selection = this.options['selection'];
        for (var i = 0; i < selection.length; i++)
        {
          items = items.add(this._getItemByKey(selection[i]));
        }
      }

      // TODO: static HTML needs to pass parent innerHTML for data
      var dragOptions = this._getDragOptions();
      var dataTransfer = event.originalEvent.dataTransfer;

      var dragImage = $(document.createElement('ul')).addClass('oj-treeview-drag-image oj-treeview-list');
      if (oj.AgentUtils.getAgentInfo()['browser'] === oj.AgentUtils.BROWSER.SAFARI)
      {
        // On Safari, the drag image is not rendered if the position is fixed.
        // If we set {position: absolute; top: -10000px}, it should theoretically
        // work for all browsers (as it does for table and listview), but it
        // causes the treeview to blink frantically if you drag an unselected
        // item on Chrome. When an unselected item is dragged, the treeview applies
        // selection effect to it, but Chrome somehow doesn't like modifying the
        // DOM and setting drag image at the same time if the drag image has absolute
        // position and negative top (I tested that positive top is fine, but positive
        // top adds a scrollbar on IE). This issue never happens on table and
        // listview because they don't allow dragging unselected items.
        dragImage.css('position', 'absolute');
      }

      var dragData = [];
      var topmost = Infinity;
      var leftmost = Infinity;

      items.each(function(index)
      {
        var item = $(this);
        dragData.push(item.data('data'));

        // Clone the item for the drag image and match the offset of the original item.
        // For RTL, we have to use the left offset of the item content because the 
        // left offset of the li is always zero due to the li filling up the entire line.
        var offset = item.offset();
        if (isRTL)
          offset.left = self._getItemContent(item).offset().left;

        var clonedItem = item.clone().css({
          'top': offset.top,
          'left': offset.left
        });

        // Don't include children in the drag image
        self._getSubtree(clonedItem).remove();

        // Drag image offset is based on the top left corner of the resulting drag image
        if (offset.top < topmost)
          topmost = offset.top;
        if (offset.left < leftmost)
          leftmost = offset.left;

        dragImage.append(clonedItem); // @HTMLUpdateOK
      });

      // : There's inconsistency in how the drag image offset is computed
      // for the native DnD impl and the polyfill. In the native impl, the offset is
      // relative to the top-left corner of the rendered drag image element. In the
      // polyfill, the offset is relative to the {top: 0, left: 0} position. To
      // reconcile this, we must make sure that the top-left corner of the rendered
      // drag image element is at the {top: 0, left: 0} position.
      dragImage.children().each(function()
      {
        $(this).css({
          'top': parseFloat($(this).css('top')) - topmost,
          'left': parseFloat($(this).css('left')) - leftmost
        });
      });

      var dragDataTypes = dragOptions['dataTypes'] || [];
      for (var i = 0; i < dragDataTypes.length; i++)
      {
        dataTransfer.setData(dragDataTypes[i], JSON.stringify(dragData));
      }

      // Drag image has to be attached to the DOM when being assigned to the dataTransfer.
      // It has to be removed afterwards to prevent leaks.
      $('body').append(dragImage); // @HTMLUpdateOK
      dataTransfer.setDragImage(dragImage[0], event.pageX - leftmost, event.pageY - topmost);
      setTimeout(function()
      {
        dragImage.remove();
      }, 0);

      var callback = dragOptions['dragStart'];
      if (callback)
        callback(event.originalEvent, {'items': dragData});
    },

    /**
     * Handles DnD drag and dragEnd events.
     * @param {Event} event The event.
     * @param {string} eventType The event type (drag or dragEnd).
     * @private
     */
    _handleDragSourceEvent: function(event, eventType)
    {
      var callback = this._getDragOptions()[eventType];
      if (callback)
        callback(event.originalEvent);
    },

    /**
     * Handles DnD dragEnter, dragOver, dragLeave, and drop events.
     * @param {Event} event The event.
     * @param {string} eventType The event type (dragEnter, dragOver, dragLeave, or drop).
     * @private
     */
    _handleDropTargetEvent: function(event, eventType)
    {
      var dropOptions = this._getDropOptions();
      var dropDataTypes = dropOptions['dataTypes'] || [];
      var callback = dropOptions[eventType];
      var targetItem = $(event.target).closest('.oj-treeview-item');

      // Position drop effects based on the spacer (disclosure icon) because it takes up the entire item height
      var spacerRect = targetItem.children('.oj-treeview-spacer')[0].getBoundingClientRect();
      var middleY = (spacerRect.top + spacerRect.bottom) / 2;

      var position = 'inside';
      var relativeY = event.pageY - spacerRect.top;
      if (relativeY < 0.25 * spacerRect.height)
        position = 'before';
      else if (relativeY > 0.75 * spacerRect.height)
        position = this._isExpanded(targetItem) ? 'first' : 'after';

      if (callback)
        callback(event.originalEvent, {'item': targetItem[0], 'position': position});

      for (var i = 0; i < dropDataTypes.length; i++)
      {
        var dataTypes = event.originalEvent.dataTransfer.types;
        if (dataTypes && dataTypes.indexOf(dropDataTypes[i]) >= 0)
        {
          event.preventDefault();
          break;
        }
      }

      if ((eventType == 'dragEnter' || eventType == 'dragOver') && event.originalEvent.defaultPrevented)
      {
        var isRTL = this._GetReadingDirection() == "rtl";

        // Draw the drop target effect on dragEnter and dragOver
        var dropLineTop = middleY;
        var dropLineLeft = isRTL ? (spacerRect.left - this._dropLineRect.width) : spacerRect.right;

        // Align the drop marker with the disclosure icon (spacer)
        var dropMarkerTop = middleY - this._dropMarkerRect.height / 2;
        var dropMarkerLeft = spacerRect.left + spacerRect.width / 2 - this._dropMarkerRect.width / 2;

        if (position == 'before')
        {
          dropMarkerTop -= spacerRect.height / 2;
          dropLineTop -= spacerRect.height / 2;
        }
        else if (position == 'after' || position == 'first')
        {
          dropMarkerTop += spacerRect.height / 2;
          dropLineTop += spacerRect.height / 2;

          if (position == 'first')
          {
            // Align with the child items
            var spacerOffset = (isRTL ? -1 : 1) * spacerRect.width;
            dropMarkerLeft += spacerOffset;
            dropLineLeft += spacerOffset;
          }
        }

        this._dropMarker.css('top', dropMarkerTop + 'px').css('left', dropMarkerLeft + 'px').show();

        if (position != 'inside')
          this._dropLine.css('top', dropLineTop + 'px').css('left', dropLineLeft + 'px').show();
        else
          this._dropLine.hide();
      }
      else
      {
        // Remove the drop target effect on dragLeave and drop
        this._dropMarker.hide();
        this._dropLine.hide();
      }
    },

    // @inheritdoc
    _NotifyContextMenuGesture: function(menu, event, eventType)
    {
      if (eventType === 'keyboard')
      {
        // If launched by Shift+F10, the context menu should be rendered next
        // to the currentItem.
        var launcher = this._currentItem ? this._getItemContent(this._currentItem) : this.element;
        var openOptions = {
          'launcher': launcher,
          'initialFocus': 'menu',
          'position': {'my': 'start top', 'at': 'start bottom', 'of': launcher}
        };
        this._OpenContextMenu(event, eventType, openOptions);
      }
      else
        this._superApply(arguments);
    },

    // @inheritdoc
    refresh: function()
    {
      this._super();

      this._render();
    },

    // @inheritdoc
    getNodeBySubId: function(locator)
    {
      if (locator == null)
        return this.element ? this.element[0] : null;

      var key = locator['key'];
      var subId = locator['subId'];
      var item = this._getItemByKey(key);
      var ret;

      if (subId === 'oj-treeview-disclosure')
        ret = item.children('.oj-treeview-disclosure-icon')[0];
      else if (subId === 'oj-treeview-item')
        ret = item.children('.oj-treeview-item-content')[0];

      // Non-null locators have to be handled by the component subclasses
      return ret || null;
    },

    // @inheritdoc
    getSubIdByNode: function(node)
    {
      if (!$.contains(this.element[0], node))
        return null;

      var subId;
      var disclosureIcon = this._getClosestDisclosureIcon($(node));
      var item = this._getClosestItem($(node));

      if (disclosureIcon.length > 0)
        subId = 'oj-treeview-disclosure';
      else if (item.length > 0)
        subId = 'oj-treeview-item';
      else
        return null;

      return {'subId': subId, 'key': this._getKey(item)};
    },

    /**
     * {@ojinclude "name":"nodeContextDoc"}
     * @param {!Element} node - {@ojinclude "name":"nodeContextParam"}
     * @returns {Object|null} {@ojinclude "name":"nodeContextReturn"}
     *
     * @example {@ojinclude "name":"nodeContextExample"}
     *
     * @expose
     * @instance
     * @memberof oj.ojTreeView
     */
    getContextByNode: function(node)
    {
      if (!$.contains(this.element[0], node))
        return null;

      var item = this._getClosestItem($(node));
      if (item.length < 1)
        return null;

      var context = {};
      context['subId'] = 'oj-treeview-item';
      context['index'] = item.parent().children('.oj-treeview-item').index(item);
      context['parentKey'] = this._getKey(this._getParentItem(item));

      context['component'] = oj.Components.__GetWidgetConstructor(this.element);
      if (this._FixRendererContext)
        context = this._FixRendererContext(context);

      var metadata = item.data('metadata');
      if (metadata)
      {
        context['data'] = item.data('data');
        context['datasource'] = this.options['data'];

        // Merge properties from metadata into item context
        // Contains key, leaf, and depth
        for (var prop in metadata) 
        {
          if (metadata.hasOwnProperty(prop))
            context[prop] = metadata[prop];
        }
      }
      else
      {
        // Static content
        context['key'] = this._getKey(item);
        context['leaf'] = this._isLeaf(item);
        context['depth'] = item.parents('.oj-treeview-list').length;
      }

      return context;
    },

    // @inheritdoc
    _setOption: function(key, value, flags)
    {
      var self = this;

      // Call the super to update the property values
      this._superApply(arguments);

      if (key == 'expanded')
      {
        this._getItems().each(function()
        {
          var item = $(this);
          if (self._isInitExpanded(item))
            self._expand(item, true);
          else 
            self._collapse(item, true);
        });
      }
      else if (key == 'selection')
      {
        this._getItems().each(function()
        {
          var item = $(this);
          self._select(item);
        });
      }
      else if (key == 'currentItem')
      {
        this._resetFocus();
      }
      else
      {
        this.refresh();
        return;
      }
    },

    // @inheritdoc
    _destroy: function()
    {
      // TODO
      // Call super at the end for destroy.
      this._super();
    }
  });
}());

// Fragments
/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Gesture</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td rowspan="2">Item</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Focus on the item. If <code class="prettyprint">selectionMode</code> is enabled, selects the item as well.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Press & Hold</kbd></td>
 *       <td>Display context menu</td>
 *     </tr>
 *     <tr>
 *       <td>Disclosure Icon</td>
 *       <td><kbd>Tap</kbd></td>
 *       <td>Expand or collapse the item.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment touchDoc - Used in touch gesture section of classdesc, and standalone gesture doc
 * @memberof oj.ojTreeView
 */

/**
 * <table class="keyboard-table">
 *   <thead>
 *     <tr>
 *       <th>Target</th>
 *       <th>Key</th>
 *       <th>Action</th>
 *     </tr>
 *   </thead>
 *   <tbody>
 *     <tr>
 *       <td><kbd>Tab</kbd></td>
 *       <td>Navigates to next focusable element on page.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Tab</kbd></td>
 *       <td>Navigates to previous focusable element on page.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>DownArrow</kbd></td>
 *       <td>Moves focus to the item below.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>UpArrow</kbd></td>
 *       <td>Moves focus to the item above.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>LeftArrow</kbd></td>
 *       <td>On an expanded item, collapses the item. Otherwise, move focus to the item above. The action is swapped with <kbd>RightArrow</kbd> in RTL locales.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>RightArrow</kbd></td>
 *       <td>On a collapsed item, expands the item. Otherwise, move focus to the item above. The action is swapped with <kbd>LeftArrow</kbd> in RTL locales.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+DownArrow</kbd></td>
 *       <td>Extends the selection to the item below. Only applicable if the multiple selection is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+UpArrow</kbd></td>
 *       <td>Extends the selection to the item above. Only applicable if the multiple selection is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Space</kbd></td>
 *       <td>Toggles the selection of the current item and deselects the other items.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Enter</kbd></td>
 *       <td>Selects the current item and deselects the other items. No op if the current item is already selected.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Ctrl+Space/Enter</kbd></td>
 *       <td>Toggles the selection of the current item while maintaining previously selected items. Only applicable if the multiple selection is enabled.</td>
 *     </tr>
 *     <tr>
 *       <td><kbd>Shift+Space/Enter</kbd></td>
 *       <td>Selects contiguous items from the last selected item to the current item. Only applicable if the multiple selection is enabled.</td>
 *     </tr>
 *   </tbody>
 * </table>
 *
 * @ojfragment keyboardDoc - Used in keyboard section of classdesc, and standalone gesture doc
 * @memberof oj.ojTreeView
 */

//////////////////     SUB-IDS     //////////////////

/**
 * <p>Sub-ID for TreeView items. See the <a href="#getNodeBySubId">getNodeBySubId</a>
 * method for details.</p>
 *
 * @ojsubid oj-treeview-item
 * @memberof oj.ojTreeView
 *
 * @example <caption>Get the item with key 'foo':</caption>
 * var item = myTreeView.getNodeBySubId({'subId': 'oj-treeview-item', 'key': 'foo'});
 */

/**
 * <p>Sub-ID for TreeView disclosure icons. See the <a href="#getNodeBySubId">getNodeBySubId</a>
 * method for details.</p>
 *
 * @ojsubid oj-treeview-disclosure
 * @memberof oj.ojTreeView
 *
 * @example <caption>Get the disclosure icon for the non-leaf item with key 'foo':</caption>
 * var item = myTreeView.getNodeBySubId({'subId': 'oj-treeview-disclosure', 'key': 'foo'});
 */

/**
 * <p>Context for TreeView items.</p>
 *
 * @property {Element} componentElement The TreeView element.
 * @property {Object} data The data object for the item (not available for static content).
 * @property {oj.TreeDataSource} datasource A reference to the data source object (not available for static content).
 * @property {number} depth The depth of the item. The depth of the first level children under the invisible root is 1.
 * @property {number} index The index of the item relative to its parent, where 0 is the index of the first item.
 * @property {Object} key the key of the item.
 * @property {boolean} leaf Whether the item is a leaf item.
 * @property {Object} parentKey The key of the parent item. The parent key is null for root item.
 *
 * @ojnodecontext oj-treeview-item
 * @memberof oj.ojTreeView
 */

(function() {
  var ojTreeViewMeta = {
    'properties': {
      'currentItem': {
        'type': 'any',
        'readOnly': true,
        'writeback': true
      },
      'data': {},
      'dnd': {
        'type': 'object',
        'properties': {
          'drag': {
            'type': 'object',
            'properties':{
              'items':{
                'type':'object',
                'properties': {
                  'dataTypes': {
                    'type': 'Array<string>'
                  },
                  'drag': {},
                  'dragEnd': {},
                  'dragStart': {}
                }
              }
            }
          },
          'drop': {
            'type': 'object',
            'properties':{
              'items' :{
                'type':'object',
                'properties': {
                  'dataTypes': {
                    'type': 'Array<string>'
                  },
                  'dragEnter': {},
                  'dragLeave': {},
                  'dragOver': {},
                  'drop': {}
                }
              }
            }
          }
        }
      },
      'expanded': {
        'writeback': true
      },
      'item': {
        'type': 'object',
        'properties': {
          'focusable': {},
          'renderer': {},
          'selectable': {}
        }
      },
      'selection': {
        'type': 'Array<object>',
        'writeback': true
      },
      'selectionMode': {
        'type': 'string',
        'enumValues': ['none', 'single', 'multiple']
      }
    },
    'events': {
      'animateEnd': {},
      'animateStart': {},
      'beforeCollapse': {},
      'beforeCurrentItem': {},
      'beforeExpand': {},
      'collapse': {},
      'expand': {}
    },
    'methods': {
      'getContextByNode': {}
    },
    'extension': {
      _WIDGET_NAME: 'ojTreeView'
    }
  };
  oj.CustomElementBridge.registerMetadata('oj-tree-view', 'baseComponent', ojTreeViewMeta);
  oj.CustomElementBridge.register('oj-tree-view', {'metadata': oj.CustomElementBridge.getMetadata('oj-tree-view')});
})();

/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/**
 * @ojcomponent oj.ojValidationGroup
 * @since 4.2.0
 * @ojstatus preview
 * @ojsignature class ojValidationGroup extends JetElement<ojValidationGroupSettableProperties>
 *
 * @classdesc
 * <h3 id="optionOverview-section">
 *   JET ValidationGroup
 *   <a class="bookmarkable-link" title="Bookmarkable Link" href="#optionOverview-section"></a>
 * </h3>
 * <p>The oj-validation-group element tracks the current validity state of its
 * JET custom element descendents that contain the <code class="prettyprint">valid</code>
 *  property.
 * </p>
 * <p>The oj-validation-group element searches all its descendents. 
 * When it finds a component with a <code class="prettyprint">valid</code> property, 
 * it adds it to the list of components 
 * it is tracking. For performance reasons it does not search the tracked component's 
 * children since the tracked component's valid state will  be based on 
 * its children's valid state, if it has children with valid states.

 * Once it finds all the components it needs to track, it calculates its own
 * valid property value based on all the enabled (including hidden) components it tracks.
 * Enabled means not disabled or readonly, so 
 * oj-validation-group's valid state ignores any disabled or readonly components.
 * </p>
 * <p>
 * The most 'invalid' component's
 * <code class="prettyprint">valid</code> property value will be
 * oj-validation-group's <code class="prettyprint">valid</code> property value.
 * When any of the tracked component's valid value changes, oj-validation-group will
 * be notified and will update its own valid value if it has changed.
 * </p>
 * <p>This is an example of the oj-validation-group wrapping the JET form components.
 * All the JET form components have the <code class="prettyprint">valid</code> property. 
 * It is not shown in the HTML markup
 * because it is readonly and cannot be set by the application developer.
 * <pre class="prettyprint">
 * <code>
 * &lt;oj-validation-group>
 *   &lt;oj-form-layout>
 *     &lt;oj-input-text required value="text" label-hint="input 1">&lt;/oj-input-text>
 *     &lt;oj-text-area value='text' rows="6" label-hint="textarea">&lt;/oj-text-area>
 *     &lt;oj-input-text value="text" label-hint="input 2">&lt;/oj-input-text>
 *     &lt;oj-input-text value="text" label-hint="input 3 longer label">&lt;/oj-input-text>
 *   &lt;/oj-form-layout> 
 * &lt;/oj-validation-group> 
 * </code></pre>
 * </p>
 *
 */

/**
 * The current validity state of the  
 * <code class="prettyprint">oj-validation-group</code> element. 
 * <p>
 * The oj-validation-group's 
 * <code class="prettyprint">valid</code> property value is calculated from all the 
 * enabled components it tracks. The most 'invalid' component's
 * <code class="prettyprint">valid</code> property value will be
 * oj-validation-group's <code class="prettyprint">valid</code> property value.
 * For example,
 * if all the components are valid except one is "invalidShown", then oj-validation-group's 
 * valid value will be "invalidShown". If one is "invalidShown" and one is "invalidHidden",
 * then oj-validation-group's valid value will be "invalidShown" since "invalidShown" is more
 * invalid than "invalidHidden".
 * </p>
 * <p>
 * When any of the enabled tracked component's valid value changes, oj-validation-group will
 * be notified and will update its own valid value if it has changed.
 * There is no default value for the valid property 
 * since it is a read-only property that is calculated on initialization
 * and kept updated if any of its tracked component's valid value changes.
 * </p>
 * <p>
 * The oj-validation-group does not filter out components that are hidden.
 * Hidden components will be considered 
 * in oj-validation-group's valid calculation as long as they are enabled. 
 * A hidden enabled component's valid state is no different 
 * than a visible enabled component; if there
 * is an error that isn't a deferred error, its valid value is "invalidShown".
 * You can disable any components you do not want
 * considered in oj-validation-group's valid value.
 * </p> 
 * <p>
 *  Note: New valid states may be added to the list of valid values in future releases.
 *  If so, we will keep the convention where if it is valid, it will start with "valid". 
 *  If it is invalid, it will start with "invalid". 
 *  If it is pending, it will start with "pending".
 * </p>
 * @member
 * @name valid
 * @expose
 * @memberof oj.ojValidationGroup
 * @instance
 * @type {string}
 * @ojvalue {string} "valid" The component is valid
 * @ojvalue {string} "pending" The component is waiting for the validation state to be determined.
 *    The "pending" state is never set in this release of JET. It will be set in a future release.
 * @ojvalue {string} "invalidHidden" The component has invalid messages hidden
 *    and no invalid messages showing. An invalid message is one with severity "error" or higher.
 * @ojvalue {string} "invalidShown" The component has invalid messages showing.
 *  An invalid message is one with severity "error" or higher.
 * @readonly
 * @ojwriteback
 * @since 4.2
 * @example <caption>Get the <code class="prettyprint">valid</code> property after initialization:</caption>
 * // getter
 * var valid = myValidationGroupElement.valid;
 */


/**
 * Sets a property or a single subproperty for complex properties and notifies the component
 * of the change, triggering a [property]Changed event.
 * 
 * @function setProperty
 * @param {string} property - The property name to set. Supports dot notation for subproperty access.
 * @param {*} value - The new value to set the property to.
 * @return {void}
 * 
 * @expose
 * @memberof oj.ojValidationGroup
 * @instance
 * 
 * @example <caption>Set a single subproperty of a complex property:</caption>
 * myComponent.setProperty('complexProperty.subProperty1.subProperty2', "someValue");
 */ 
/**
 * Retrieves a value for a property or a single subproperty for complex properties.
 * @function getProperty
 * @param {string} property - The property name to get. Supports dot notation for subproperty access.
 * @return {*}
 * 
 * @expose
 * @memberof oj.ojValidationGroup
 * @instance
 * 
 * @example <caption>Get a single subproperty of a complex property:</caption>
 * var subpropValue = myComponent.getProperty('complexProperty.subProperty1.subProperty2');
 */ 
/**
 * Performs a batch set of properties.
 * @function setProperties
 * @param {Object} properties - An object containing the property and value pairs to set.
 * @return {void}
 * @expose
 * @memberof oj.ojValidationGroup
 * @instance
 * 
 * @example <caption>Set a batch of properties:</caption>
 * myComponent.setProperties({"prop1": "value1", "prop2.subprop": "value2", "prop3": "value3"});
 */ 


/**
 * @ignore
 */
var ojValidationGroupMeta = {
  "properties": {
    "valid": {
      "type": "string",
      "enumValues": ["valid", "invalidShown", "invalidHidden", "pending"],
      "readOnly": true,
      "writeback": true
    }
  },
  "methods": {
    "focusOn": {},
    "showMessages": {}
  },
  "extension": {
    _CONSTRUCTOR: ojValidationGroup
  }
};
Object.freeze(ojValidationGroupMeta);


/**
 * The _ojValidationGroup constructor function.
 * 
 * @constructor
 * @private
 */
function ojValidationGroup(context) {
  
  var FIRST_INVALID_SHOWN_KEY = "@firstInvalidShown";
    
  var self = this;
  var element = context.element;

  var _trackedComponents = [];
  var _ojDiv;
  


  // Our version of GCC has a bug where the second param of MutationObserver.observe must be of 
  // type MutationObserverInit which isn't a real class that we can instantiate. Work around is to
  // create the MutationObserver on an alias of 'this' and call observe in a different function.
  // TODO Cleanup when we replace GCC with uglify in 5.0.0.
  /**
   * If the dom is mutated, call _refreshTrackedComponents if needed
   * 
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  self._rootElementMutationObserver = new MutationObserver(function (mutations) {
    // mutations is an array of objects, each of type MutationRecord.
    // 
    // if the oj-validation-group has been removed from the body element's subtree,
    // no need to remove event listeners or refresh.  Just disconnect the observer.
    // This can happen when you press apply in the cookbook demo.

    if (document.body.contains(element))
    {
      var addedNodes;
      var ignore;    
      var mutation;
      var mutationsLength = mutations.length;
      var refreshNeeded = false;
      var removedNodes;
      var target;
      
      // figure out if we have to refresh the tracked components list
      // if refreshNeeded is false, 
      // then continue for loop and look at the next mutation to see if that can be ignored
      // or if _refreshTrackedComponents needs to be called.
      for (var i=0; i < mutationsLength && !refreshNeeded; i++)
      {
        mutation = mutations[i];
        // target is the node whose children changed
        target = mutation.target;
        
        // we definitely need to refresh if we are removing a tracked node.
        // if we remove a tracked node, like ojInputTextNode.remove(), the target is the
        // tracked node's parent and the tracked node is in removedNodes.
        removedNodes = mutation.removedNodes;
        addedNodes = mutation.addedNodes;
        refreshNeeded = _isOrContainsTrackedComponent(removedNodes);
        if (!refreshNeeded && addedNodes.length > 0)
        {
          // if refreshNeeded is false, and we are adding something,
          // then look at this mutation; 
          // if it is a child of any of the tracked components,
          // then ignore this mutation.
          ignore = _isNodeInTrackedComponent(target);

          // if ignore is false, see if target or addedNodes needs to be tracked.
          // checking on the addedNodes does not always work. you must also check target
          // If so, set refreshNeeded to true
          if (!ignore && _isAnyNodeTrackable(target, addedNodes))
          {
            refreshNeeded = true;
          }
        }
      }
      // if 'refreshNeeded' is true, 
      // one of the mutations is or contains an element that needs tracking.
      if (refreshNeeded)
      {
        // remove event listeners from removed nodes if they are there
        _removeEventListenersFromRemovedChildren(mutations);
        // refresh the tracked components (clears and updates the _trackedComponents
        // and re-adds event handlers)
        _refreshTrackedComponents();
      }
    }
    else
    {
      this.disconnect();
    }
  }); // end self._rootElementMutationObserver

  // This gets called on initial render.
  // It gets the tracked components. No ui change is made here.
  self.createDOM = function() 
  {
      element.classList.add("oj-validation-group");

    // 1. add a div child of oj-validation-group. 
    //    We call this _ojDiv and add data-oj-context onto it
    //    to be used by the BusyContext. 
    // 2. reparent its children to _ojDiv. With the reparenting the framework 
    //    resolves the old busy context and creates a 
    //    new one for the children when you add them back in the DOM. It doesn't work to 
    //    put the data-oj-context on oj-validation-group since the 
    //    BusyContext needs the data-oj-context attribute needs to be in place 
    //    when the children are added. It would work if the app developer puts the data-oj-context
    //    on oj-validation-group in his html page, 
    //    but we don't want the app dev to have to do that.
    _ojDiv = document.createElement("div");
    _ojDiv.setAttribute("data-oj-context","");

    // wrap the children with the div
    // use firstChild so we can get text/comments/etc. ko bindings are comments, and we don't
    // want to skip those.
    while(element.firstChild) {
      _ojDiv.appendChild(element.firstChild); // @HTMLUpdateOK reparenting child nodes
    }
    element.appendChild(_ojDiv); // @HTMLUpdateOK appending internally created DOM element

  };

  // remove event listeners and refresh tracked components
  // this is not called for property changes since I override handlePropertyChanged
  // and return true from it.
  // This function gets called on initial render and when refresh() is called.
  self.updateDOM = function() 
  {

    // We wait until the child elements have finished upgrading before we get the 
    // tracked components.
    // If we do not do this, we may get a lot of mutations as the children are adding
    // dom themselves; e.g., oj-form-layout  marks itself busy as it manipulates
    // its dom (adds oj-label, bonus dom, etc), and we don't have to get
    // a mutation observer for each of its manipulations if we wait until the BusyContext resolves.
    // 
    // Another benefit of waiting is if an app dev adds an on-valid-changed listener
    // on oj-validation-group that he wants called on page load, 
    // the listener gets called on page load only when we wait on the BusyContext resolving. 
    // We can wait here or we can have the app dev wait. It's best to wait here for these 
    // reasons.
    //
    // However, in the case where knockout is not used and the child components 
    // are registered after the parent so that the parent's logic 
    // gets run first before the children even get a chance to 
    // add their busy states (before they even get registered).
    //  In that case this would immediately get called. It's ok since we will
    // get mutations when the components get upgraded, and they get the valid
    // property.
    var busyContext = oj.Context.getContext(_ojDiv).getBusyContext();
 
    busyContext.whenReady().then(function () {
      
      // if this is the initial render, _trackedComponents.length is 0
      var length = _trackedComponents.length;
      for (var i=0; i < length; i++)
        _removeEventListenersFromNode(_trackedComponents[i]);

      _refreshTrackedComponents();

      // Pay attention to mutations.  The mutations 
      // we care about are added elements and removed elements. 
      // We need to update the trackedComponents and their listeners in these cases.
      self._rootElementMutationObserver.observe(element,
      {"childList": true, "subtree": true});

    });

  };
  

  // Handles property changes for oj-validation-group.
  // handlePropertyChanged is an optional DefinitionalElementBridge method where if defined, 
  // will be called with the property and value that changed
  // letting the component handle a partial update case. The component
  // can return true in this callback to skip a full render call to updateDOM
  // which will be done if this method returns false or undefined. 
  self.handlePropertyChanged  = function(property, value)
  {
    // skip a full render call to updateDOM when we get a property change.
    return true;
  };
  
  /**
   * Takes all deferred messages and shows them on each enabled component. An enabled
   * component is one that is not disabled or readonly.
   * As a result, the valid property may be updated; e.g.,
   * if the valid state was "invalidHidden"
   * before showMessages(), the valid state will become "invalidShown" after showMessages(). 
   * <p>
   * If there were no deferred messages this method simply returns. 
   * </p>
   * @function showMessages
   * @access public
   * @instance
   * @return {void}
   * @example <caption>Display all messages including deferred ones.</caption>
   * validationGroupElem.showMessages();
   * @expose
   * @memberof oj.ojValidationGroup
   * @since 4.2.0
   */
   self.showMessages = function ()
   {
     var component;
     
      // get tracked components and call showMessages() on them
      // as long as they are enabled.
      for (var i = 0; i < _trackedComponents.length; i++)
      {
        component = _trackedComponents[i];
        if (!(component["disabled"] || component["readonly"]))
        {
          if ("showMessages" in component)
            component.showMessages();
        }
      }
   };
    
  /**
   * If the parameter passed in is "@firstInvalidShown", then
   * it sets focus on the first enabled component with invalid messages showing, if any.
   * If nothing is passed in, it will set focus to the first enabled component being tracked, 
   * if any. An enabled
   * component is one that is not disabled or readonly.
   * @function focusOn 
   * @param {string=} key - "@firstInvalidShown" will focus on first invalidShown enabled
   * component in DOM order, if any. When no parameter is passed in, the method will
   * focus on first enabled component regardless of the valid state, if any. An enabled
   * component is one that is not disabled or readonly.
   * @ojsignature {target: "Type", for:"key", value: "'@firstInvalidShown'"}
   * @example <caption>Focus on the first enabled component showing invalid messages.</caption>
   * validationGroupElem.focusOn("@firstInvalidShown");
   * @example <caption>Focus on the first enabled component.</caption>
   * validationGroupElem.focusOn();
   * @access public
   * @instance
   * @return {void}
   * @expose
   * @memberof oj.ojValidationGroup
   * @since 4.2.0
   */
  self.focusOn = function (key) 
  { 
    var firstFocusable = null;
    var firstInvalid = null;

    if (key === FIRST_INVALID_SHOWN_KEY)
    {
      firstInvalid = _getFirstInvalidComponent();

      if (firstInvalid) 
      {
        // If the component has the 'focusOn' method, 
        // call that instead of focus() 
        // to handle nested oj-validation-groups
        if ("focusOn" in firstInvalid)
          firstInvalid.focusOn(FIRST_INVALID_SHOWN_KEY);
        else
          firstInvalid.focus();
      }
    }
    else if (key === undefined)
    {
      // Get the first non readonly or disabled, then call focus
      firstFocusable = _getFirstFocusable();
      
      if (firstFocusable) 
      {
        // If the component has the 'focusOn' method, 
        // call that instead of focus() 
        // to handle nested oj-validation-groups
        if ("focusOn" in firstFocusable)
          firstFocusable.focusOn();
        else
          firstFocusable.focus();
      }
    }
    else
    {
      // it's a no/op if they pass something into focusOn and it's not a key we support
      // which is now only '@firstInvalidShown'.
      // this behaves the same way html's focus() works.
      oj.Logger.info("focusOn's parameter value is not '@firstInvalidShown' or empty, so it's a no-op.");
    }
  };

  /**
   * This function clears out _trackedComponents, finds them all again, and adds
   * the child event listeners on each one. It also updates the 'valid' property.
   * <p>
   * The dom spec for addEventListener states that it will not add duplicate
   * event listeners, so there is no need to track if we already added this.
   * If you want to remove child event listeners, do so before this function is called.
   * </p>
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _refreshTrackedComponents()
  {
    // clear out the tracked components when this is called
    _trackedComponents = [];
    // this sets the _trackedComponents
    _getTrackedComponents(element);

    context.props["valid"] = _consolidateValid();

    // property change events do not bubble, 
    // so add to each child where the property change occurs
    var length = _trackedComponents.length;
    for (var i = 0; i < length; i++)
      _addChildEventListeners(_trackedComponents[i]);
  }

  /**
   * Add the property change event listeners to the tracked component node, aka child.
   * 
   * Only the component events bubble and not the property change events, 
   * so it doesn't work to listen on the oj-validation-group for 
   * these property change events. Instead we
   * put them on the child component themselves, and we share the listener.
   * @param {Element} child
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _addChildEventListeners(child)
  {
    child.addEventListener("validChanged", _handleValidChanged.bind(this));
    child.addEventListener("disabledChanged", _handleDisabledReadonlyChanged.bind(this));
    child.addEventListener("readonlyChanged", _handleDisabledReadonlyChanged.bind(this));
  }
  
  /**
   * remove child event listeners
   * Only the component events bubble and not the property change events, 
   * so it doesn't work to listen on the oj-validation-group 
   * for these property change events. We add/remove the listener from the 
   * child JET component itself.
   * 
   * @param {Array.<MutationRecord>} mutations the array of MutationRecords from the observer
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _removeEventListenersFromRemovedChildren(mutations)
  {
    var mutationsLength = mutations.length;
    for (var i = 0; i < mutationsLength; i++)
    {
      var mutation = mutations[i];
      var removedNodesLength = mutation.removedNodes.length;  
      for (var j = 0; j < removedNodesLength; j++)
      {
        var child = mutation.removedNodes[j];

        if (child.nodeType === 1)
        {
          _removeEventListenersFromNode(child);
        }
      }
    }    
  }
  
  /**
   * Given a node element, remove the event listeners we added
   * @param {Element} node The element from which to remove the event listeners we added.
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _removeEventListenersFromNode(node)
  {
    node.removeEventListener("validChanged", _handleValidChanged);
    node.removeEventListener("disabledChanged", _handleDisabledReadonlyChanged);
    node.removeEventListener("readonlyChanged", _handleDisabledReadonlyChanged);    
  }
  
  /**
   * Handles any validChanged events from the tracked components.
   * This consolidates the valid properties of the tracked components,
   * finding the most invalid state, and setting the oj-validation-group's 
   * 'valid' property to that state.
   * @param {Event} event
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _handleValidChanged(event)
  {
    context.props["valid"] = _consolidateValid();
  }
  
    
  /**
   * Handles any disabledChanged and readonlyChanged
   * events from the tracked components.
   * This consolidates the valid properties of the tracked components,
   * finding the most invalid state, and setting the oj-validation-group's 
   * 'valid' property to that state.
   * We do not want to count the valid state of a non-enabled component
   * in our oj-validation-group valid state.
   * @param {Event} event
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _handleDisabledReadonlyChanged(event)
  {
    var component = event.target;
    var componentValid = component["valid"];
    var componentEnabled = !(component["readonly"] || component["disabled"]);

    // If the component that changed
    // its disabled/readonly state is valid, 
    // then it's a no-op since that won't
    // change oj-validation-group's valid property value.
    if (componentValid === "valid")
      return;
    // invalidShown is the 'most severe' valid state, so if the
    // enabled component's valid state is invalidShown, then no
    // matter what oj-validation-group is already, it will be invalidShown now.
    // if oj-validation-group is already invalidShown, the set property is a no/op.
    else if (componentEnabled && componentValid === "invalidShown")
    {
      context.props["valid"] = "invalidShown";
    }
    else
    {
      // if it went from true to false or vice versa,
      // we need to re-consolidate the valid value since
      // we do not count disabled/readonly components
      // in the valid calculation.
      context.props["valid"] = _consolidateValid();
      
    }
  }

  /**
   * This consolidates the valid properties of the tracked components,
   * finding the most invalid state, and setting the oj-validation-group's 
   * 'valid' property to that state.
   * We track components even when they are disabled or readonly, but
   * we do NOT consider them in the 'valid' state calculation. 
   * We track them because we will consider them in the 'value',
   *  when that property is added to oj-validation-group
   * in a future release.
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   * @return {string} Returns the most invalid state: "invalidShown", "invalidHidden",
   * "pending", or "valid".
   */
  function _consolidateValid()
  {
    var component;
    var disabled;
    var componentValidState;
    var mostInvalid = "valid";
    var readonly;
    
    // The order of validity (least valid to most valid) is: 
    // "invalidShown", "invalidHidden", "pending", "valid"
    //
    // Find the most invalid state of all the tracked components.
    // break out of loop as soon as we see the 
    // most invalid state, "invalidShown"
    var length = _trackedComponents.length;
    for(var i=0; i < length && mostInvalid !== "invalidShown"; i++)
    {
      component = _trackedComponents[i];
      if (_isComponentNotDestroyed(component))
      {
        componentValidState = component["valid"];
        disabled = component["disabled"];
        readonly = component["readonly"];
        // we do not count disabled/readonly components
        // in the valid calculation.
        if (!(disabled || readonly))
        { 
          if (componentValidState === "invalidShown")
            mostInvalid = "invalidShown";
          else if (componentValidState === "invalidHidden")
            mostInvalid = "invalidHidden";
          else if (componentValidState === "pending" && mostInvalid === "valid")
            mostInvalid = "pending";
        }
      }   
    }
    return mostInvalid;
  }
    
  /**
   * Gets all the tracked components within the tracker.
   * Check childNodes of root. If trackable, store in _trackedComponents and 
   * don't go into descendents. If not trackable, store in notTrackableElements, 
   * and do go into the descendents by calling this function recursively.
   * @param {Element} root The element from which to start the search for tracked components.
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _getTrackedComponents(root)
  {
    var child;
    var children;
    var i;
    var length;
    var notTrackableElements = [];
    var trackable;
     
    // get children
    // check each one. if trackable, store in trackedComponents and don't go into descendents.
    // if not trackable, store in notTrackableElements, and do go into the descendents. 
    children = root.childNodes;
    length = children.length;
    for ( i=0; i< length; i++)
    {
      child = children[i];
      trackable = _isTrackable(child);
      if (trackable)
      {
        _trackedComponents.push(child); 
      } 
      else
      {
        notTrackableElements.push(child);
      }
    }
    
    // now loop through the notTrackableElements we found the first pass and check their children.
    length = notTrackableElements.length;
    for (i=0; i < length; i++)
    {
      _getTrackedComponents(notTrackableElements[i]);
    }   
  }

  /**
   * Check the node to see if oj-validation-group should track it.
   * In this first revision of oj-validation-group, we see if it has the 'valid' property.
   * 
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   * @param {Node} elem see if the element is trackable by oj-validation-group
   * @return {boolean} true if the element is one the oj-validation-group must track.
   */
  function _isTrackable(elem)
  {
    var valid;
    
    // only look at elements with a '-' in it.
    if (elem.nodeType === 1 && elem.tagName.indexOf("-") !== -1 && "valid" in elem)
    {
      if (_isComponentNotDestroyed(elem))
      {
        valid = elem["valid"];
        if (valid !== undefined)
          return true;
      }
    }
    return false;
  }
  
  /**
   * Check the target and nodes to see if any of them are trackable. 
   * Useful for our MutationObserver when figuring out if we can ignore a mutation. 
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   * @param {Node} target node to see if trackable
   * @param {NodeList} nodes nodes to see if any are trackable
   * @return {boolean} true if the target or a node from nodelist is trackable, else false
   */
  function _isAnyNodeTrackable(target, nodes)
  {
    var length = nodes.length;
    
    if (_isTrackable(target))
      return true;
    
    // loop through the nodes and see if one is trackable
    for (var i=0; i < length; i++)
    {
      if (_isTrackable(nodes[i]))
        return true;
    }
    return false; 
  }
  
  /**
   * Checks if a _trackedComponents component contains the elem (i.e., a child)
   * Useful for our MutationObserver when figuring out if we can ignore a mutation.
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   * @param {Node} node node to check to see if it is a child of one of the tracked components
   * @return {boolean} true if a _trackedComponents component contains the elem, false otherwise
   * 
   */
  function _isNodeInTrackedComponent(node)
  {
    var length = _trackedComponents.length;
    var trackedComponent;
    
    if (node.nodeType !== 1)
      return false;
    
    // look at this mutation. 
    // if it is a child of any of the tracked components,
    // then ignore this mutation.
    for (var i=0; i < length; i++)
    {
      trackedComponent = _trackedComponents[i];
      if (trackedComponent.contains( node ))
      {
        return true;
      }
    }   
    return false;
  }
  
  /**
   * Returns true if one of the nodes is or contains a tracked component.
   * Useful for our MutationObserver when looking at what nodes are being removed.
   * 
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   * @param {NodeList} nodes nodes to see if any are in the _trackedComponents list or contain
   *  a tracked component
   * @return {boolean} true if any of the nodes are or contain a tracked component.
   */
  function _isOrContainsTrackedComponent(nodes)
  {
    var node;
    var trackedLength = _trackedComponents.length;
    
    for (var i=0; i < nodes.length; i++)
    {
      node = nodes[i];
      if (node.nodeType === 1)
      {
        // check if node is in the _trackedComponents
        if (_trackedComponents.indexOf(node) !== -1)
          return true;
      }
      // now check if the node contains a tracked node
      for (var j=0; j < trackedLength; j++)
      {
        if (node.contains(_trackedComponents[j]))
          return true;
      }
    }
    return false;
  }
  
  /**
   * Gets the first component that isn't readonly or disabled to focus on.
   * 
   * @returns {Element|null} the first focusable element or null if none
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _getFirstFocusable() 
  {

    var trackedComponent = null;
    
    var length = _trackedComponents.length;
    for (var i = 0; i < length; i++)
    {
      trackedComponent = _trackedComponents[i];
      if (!(trackedComponent["readonly"] || trackedComponent["disabled"]))
      {
        return trackedComponent;
      }
    }
    return null;
  }
  
  /**
   * Gets the first invalid component and returns the component to focus on.
   * 
   * @returns {Element|null} the component instance that has focus or null
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _getFirstInvalidComponent() 
  {
    
    var component;
    var invalidComponents = [];
    
    
    if (element["valid"] !== "invalidShown") 
    {
      return null;
    }

    // locate first invalid component and set focus on it
    // _trackedComponents may not be in DOM order if a new tracked component was added not
    // at the end of the DOM order.
    var length = _trackedComponents.length;
    for (var i = 0; i < length; i++) 
    {
      component = _trackedComponents[i];
      // don't try to focus on a disabled/readonly component
      if (!(component["disabled"] || component["readonly"]))
      {
        if (component["valid"] === "invalidShown") {
          invalidComponents.push(component);
        }
      }
    }

    if (invalidComponents.length === 0)
    {
      return null;
    }

    // sort the invalidComponents based on dom order
    invalidComponents.sort(function (elementA, elementB) 
    {
      // If elementA precedes elementB in dom order, return -1
      return (elementA.compareDocumentPosition(elementB) & Node.DOCUMENT_POSITION_FOLLOWING) ? -1 : 1;
    });

    // invalidComponents is sorted now by document order, so return the first one.
    return invalidComponents[0];
  }
  
  /**
   * We need to make sure widgetElem hasn't been destroyed before 
   * accessing its properties, like elem.["valid"].
   * We can remove this code in v5.0 when the framework handles custom elements implemented
   * as widgets's destroy/ko.cleanNode better.
   * In 5.0 the framework will  circumvent 
   * jquery's cleanup altogether for custom elements and children of custom elements
   * 
   * @param {Node} elem the element to check
   * @returns {boolean} true if it is safe to call .valid on the element, false if not.
   * @memberof oj.ojValidationGroup
   * @instance
   * @private
   */
  function _isComponentNotDestroyed(elem) 
  {
    
    var bridge;
    var bridgeWidgetElem;
    var widgetElem;
    
    bridge = oj.BaseCustomElementBridge.getInstance(elem);
    bridgeWidgetElem = bridge._WIDGET_ELEM;
    // bridgetWidgetElem is undefined for oj-validation-group 
    // since it's not implemented using the widget code.
    // bridgetWidgetElem has a value for oj-input-text, oj-input-number, etc., 
    // since they are implemented using the widget code. 
    if (bridgeWidgetElem !== undefined)
      widgetElem = oj.Components.__GetWidgetConstructor(bridgeWidgetElem);

    return (bridgeWidgetElem === undefined || widgetElem) ? true : false;
      
  }
}

oj.CustomElementBridge.registerMetadata('oj-validation-group', null, ojValidationGroupMeta);
oj.CustomElementBridge.register('oj-validation-group', 
  {'metadata': oj.CustomElementBridge.getMetadata('oj-validation-group')});
})();